#--------------------------------------------------------------------------
# $Id: //depot/prod/test/Rsierranevada/ATI/ati_ontapi.pm#1 $
#
# Copyright (c) 1992-2005 Network Appliance, Inc.
# All rights reserved.
#--------------------------------------------------------------------------
#
# PACKAGE: ati_ontapi.pm
#
# AUTHOR: Will McGovern (will@netapp.com)
#
# This package encapsulates all the core functionality for interfacing with
# a NetApp system using the ONTAPI management API over either HTTP or secure
# HTTPS connections. It provides all the flexibility of specification of XML
# requests in an intuitive perl "deep array" format.
#
# All XML requests and responses are validated against an ONTAPI XML DTD.
#
# This package also relies on some general properties of the XML accepted
# and generated by ONTAPI in order to take some shortcuts and provide more
# convenience. It is not intended for generalized XML processing.
#
# TODO:
#
# o Get the ONTAPI DTD fixed as it is broken in many places.
# o Need deep array validator to help catch malformed requests.
# o Validate against an ONTAPI XML schema rather than just a DTD.
#
#--------------------------------------------------------------------------
#
# Exported API interface subroutines:
#
# ati_ontapi_create_context     : Create an ATI context.
# ati_ontapi_call_api           : Execute an ONTAPI API.
# ati_ontapi_call_cli           : Execute a CLI command via ONTAPI.
#
# Exported constants:
#
# ATI_FAILURE                   : Operation failure
# ATI_SUCCESS                   : Operation success
#
# ATI_ONTAPI_TRANSPORT_HTTP     : Use plain HTTP transport protocol
# ATI_ONTAPI_TRANSPORT_HTTPS    : Use secure HTTPS transport protocol
# ATI_ONTAPI_TRANSPORT_D        : Use ONTAPI-D socket protocol
# ATI_ONTAPI_TRANSPORT_DS       : Use ONTAPI-D secure socket protocol
# ATI_ONTAPI_TRANSPORT_DEFAULT  : Default tranport protocol
#
# ATI_ONTAPI_VERSION_1_3        : ONTAPI SDK version 1.3 (6.5.x)
# ATI_ONTAPI_VERSION_1_4        : ONTAPI SDK version 1.4 (7.0.x)
# ATI_ONTAPI_VERSION_1_5        : ONTAPI SDK version 1.5 (7.1.x)
# ATI_ONTAPI_VERSION_1_6        : ONTAPI SDK version 1.6 (7.2.x)
# ATI_ONTAPI_VERSION_DEFAULT    : ONTAPI SDK default version
#
# ATI_ONTAPI_D_VERSION_1_1      : ONTAPI-D SDK version 1.1
# ATI_ONTAPI_D_VERSION_DEFAULT  : ONTAPI-D SDK default version
#
# ATI_ONTAPI_PRIVILEGE_ADMIN    : Administrative privilege level
# ATI_ONTAPI_PRIVILEGE_ADVANCED : Advanced privilege level
# ATI_ONTAPI_PRIVILEGE_DIAG     : Diagnostic privilege level
# ATI_ONTAPI_PRIVILEGE_TEST     : Test privilege level (debug kernels)
# ATI_ONTAPI_PRIVILEGE_DEFAULT  : Default privilege level
#
# ATI_ONTAPI_USER_DEFAULT       : Default user login
# ATI_ONTAPI_PASSWORD_DEFAULT   : Default user password
#
# ATI_ONTAPI_CACHE_DIR_DEFAULT  : Default persistent cache directory
#
# ATI_ONTAPI_CACHE_FILE_DEFAULT : Default persistent cache file name
#
# ATI_ONTAPI_CACHE_ALL_DEFAULT  : Default persistent cache generation mode
#
# ATI_ONTAPI_VALIDATE_OFF       : Do NOT validate against the DTD.
# ATI_ONTAPI_VALIDATE_ON        : Validate XML against the DTD
# ATI_ONTAPI_VALIDATE_DEFAULT   : Default validation
#
# ATI_ONTAPI_DEBUG_DEBUG        : Debug logging output
# ATI_ONTAPI_DEBUG_VERBOSE      : Verbose logging output
# ATI_ONTAPI_DEBUG_TERSE        : Terse logging output
# ATI_ONTAPI_DEBUG_NONE         : Turn off logging output
# ATI_ONTAPI_DEBUG_DEFAULT      : Default logging level
#
#--------------------------------------------------------------------------
#
# ONTAPI XML FORMAT NOTES:
#
# The ONTAPI XML request template looks like this:
#
#   POST URL = 'http://<filer>/servlets/netapp.servlets.admin.XMLrequest_filer'
#
#   <?xml version='1.0' encoding='utf-8' ?>
#   <!DOCTYPE netapp SYSTEM 'file:/etc/netapp_filer.dtd'>
#   <netapp version='1.3' xmlns='http://www.netapp.com/filer/admin'>
#       ----- XML API request content ------
#   </netapp>
#
# The ONTAPI XML success response template looks like this:
#
#   <?xml version="1.0" encoding="UTF-8" ?>
#   <!DOCTYPE netapp SYSTEM '/na_admin/netapp_filer.dtd'>
#   <netapp xmlns='http://www.netapp.com/filer/admin' version="1.1">
#       <results status="passed">
#           ----- XML API response content ------
#       </results>
#   </netapp>
#
# The ONTAPI XML failure response template looks like this:
#
#   <?xml version="1.0" encoding="UTF-8" ?>
#   <!DOCTYPE netapp SYSTEM '/na_admin/netapp_filer.dtd'>
#   <netapp xmlns='http://www.netapp.com/filer/admin' version="1.1">
#       <results status="failed" errno="666" reason="very good reason"/>
#   </netapp>
#--------------------------------------------------------------------------
#
# NOTE TO THOSE WORKING ON THIS MODULE CODE:
#
# I am not a perl guru so if you see anything broken feel free to fix or
# improve the code. If, however, these changes mess up the readability or
# maintainability (by those of us non-perl gurus) with obscure code,
# ego-boosting fancy perl tricks, or unnecessary "optimizations" then I
# will find you and you will suffer greatly.
#
#--------------------------------------------------------------------------
#
package ATI::ati_ontapi;
#
# Always require hard-assed checking
#
use strict;
use warnings;
#
# System packages
#
use Socket;
use IO::Socket;
use XML::LibXML;
use LWP::UserAgent;
use Storable 1.013;
#use Digest::SHA1;
#
# Export symbols for external use
#
use Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(ATI_FAILURE
		    ATI_SUCCESS
		    ATI_ONTAPI_TRANSPORT_HTTP
		    ATI_ONTAPI_TRANSPORT_HTTPS
		    ATI_ONTAPI_TRANSPORT_D
		    ATI_ONTAPI_TRANSPORT_DS
		    ATI_ONTAPI_TRANSPORT_DEFAULT
		    ATI_ONTAPI_VERSION_1_3
		    ATI_ONTAPI_VERSION_1_4
		    ATI_ONTAPI_VERSION_1_5
		    ATI_ONTAPI_VERSION_1_6
		    ATI_ONTAPI_VERSION_DEFAULT
		    ATI_ONTAPI_D_VERSION_1_1
		    ATI_ONTAPI_D_VERSION_DEFAULT
		    ATI_ONTAPI_PRIVILEGE_ADMIN
		    ATI_ONTAPI_PRIVILEGE_ADVANCED
		    ATI_ONTAPI_PRIVILEGE_DIAG
		    ATI_ONTAPI_PRIVILEGE_TEST
		    ATI_ONTAPI_PRIVILEGE_DEFAULT
		    ATI_ONTAPI_USER_DEFAULT
		    ATI_ONTAPI_PASSWORD_DEFAULT
		    ATI_ONTAPI_VALIDATE_OFF
		    ATI_ONTAPI_VALIDATE_ON
		    ATI_ONTAPI_VALIDATE_DEFAULT
		    ATI_ONTAPI_CACHE_DIR_DEFAULT
		    ATI_ONTAPI_CACHE_FILE_DEFAULT
		    ATI_ONTAPI_CACHE_ALL_DEFAULT
		    ATI_ONTAPI_DEBUG_NONE
		    ATI_ONTAPI_DEBUG_TERSE
		    ATI_ONTAPI_DEBUG_VERBOSE
		    ATI_ONTAPI_DEBUG_DEBUG
		    ATI_ONTAPI_DEBUG_DEFAULT
		    ati_ontapi_create_context
		    ati_ontapi_call_api
		    ati_ontapi_call_cli
		   );
#
# PUBLIC: Integer constants for success/failure
#
use constant ATI_FAILURE => 0;   # Operation successful
use constant ATI_SUCCESS => 1;   # Operation failed
#
# PUBLIC: ONTAPI network transport protocols
#
use constant ATI_ONTAPI_TRANSPORT_HTTP    => 'http';
use constant ATI_ONTAPI_TRANSPORT_HTTPS   => 'https';
use constant ATI_ONTAPI_TRANSPORT_D       => 'ONTAPI-D';
use constant ATI_ONTAPI_TRANSPORT_DS      => 'ONTAPI-DS';
use constant ATI_ONTAPI_TRANSPORT_DEFAULT => ATI_ONTAPI_TRANSPORT_HTTP;
#
# PUBLIC: ONTAPI network transport timeout default, in seconds.
#
use constant ATI_ONTAPI_TIMEOUT_DEFAULT => 0;      # Forever
#
# PUBLIC: ONTAPI version numbers
#
use constant ATI_ONTAPI_VERSION_1_3     => '1.3';  # Data ONTAP 6.5.x
use constant ATI_ONTAPI_VERSION_1_4     => '1.4';  # Data ONTAP 7.0.x
use constant ATI_ONTAPI_VERSION_1_5     => '1.5';  # Data ONTAP 7.1.x
use constant ATI_ONTAPI_VERSION_1_6     => '1.6';  # Data ONTAP 7.2.x
use constant ATI_ONTAPI_VERSION_DEFAULT => ATI_ONTAPI_VERSION_1_3;
#
# PUBLIC: ONTAPI-D version numbers
#
use constant ATI_ONTAPI_D_VERSION_1_1     => '1.1';  # Data ONTAPI-D 1.1
use constant ATI_ONTAPI_D_VERSION_DEFAULT => ATI_ONTAPI_D_VERSION_1_1;
#
# PUBLIC: ONTAP administrative privilege levels
#
use constant ATI_ONTAPI_PRIVILEGE_ADMIN    => 'admin';
use constant ATI_ONTAPI_PRIVILEGE_ADVANCED => 'advanced';
use constant ATI_ONTAPI_PRIVILEGE_DIAG     => 'diag';
use constant ATI_ONTAPI_PRIVILEGE_TEST     => 'test';
use constant ATI_ONTAPI_PRIVILEGE_DEFAULT  => ATI_ONTAPI_PRIVILEGE_ADMIN;
#
# PUBLIC: ONTAPI user default
#
use constant ATI_ONTAPI_USER_DEFAULT  => 'root';
#
# PUBLIC: ONTAPI password default
#
use constant ATI_ONTAPI_PASSWORD_DEFAULT  => '';
#
# PUBLIC: ONTAPI XML validation against DTD
#
use constant ATI_ONTAPI_VALIDATE_OFF     => 0;
use constant ATI_ONTAPI_VALIDATE_ON      => 1;
use constant ATI_ONTAPI_VALIDATE_DEFAULT => ATI_ONTAPI_VALIDATE_OFF;
#
# PUBLIC: ONTAPI API/type cache default directory
#
use constant ATI_ONTAPI_CACHE_DIR_DEFAULT => "/tmp";
#
# PUBLIC: ONTAPI API/type cache default file name
#
use constant ATI_ONTAPI_CACHE_FILE_DEFAULT => undef;
#
# PUBLIC: ONTAPI cache all type/api information on context creation
#
use constant ATI_ONTAPI_CACHE_ALL_DEFAULT => 0;
#
# PUBLIC: ONTAPI debug levels
#
use constant ATI_ONTAPI_DEBUG_NONE    => 0;
use constant ATI_ONTAPI_DEBUG_TERSE   => 10;
use constant ATI_ONTAPI_DEBUG_VERBOSE => 20;
use constant ATI_ONTAPI_DEBUG_DEBUG   => 30;
use constant ATI_ONTAPI_DEBUG_DEFAULT => ATI_ONTAPI_DEBUG_NONE;
#
# PRIVATE: Data ONTAP XML response status strings
#
use constant ATI_ONTAPI_STATUS_PASSED => 'passed';
use constant ATI_ONTAPI_STATUS_FAILED => 'failed';
#
# PRIVATE: Data ONTAPI-D server ports
#
use constant ATI_ONTAPI_D_SERVER_PORT  => 951;
use constant ATI_ONTAPI_DS_SERVER_PORT => 952;

### Old ZSM port values, no longer used.
###use constant ATI_ONTAPI_D_SERVER_PORT  => 2005;
###use constant ATI_ONTAPI_DS_SERVER_PORT => 2006;

#
# PRIVATE: Define the base atomic types for ONTAPI
#
my %base_types = (string  => undef,
		  integer => undef,
		  boolean => undef,
		  empty   => undef);

my $Ssl_initialized = 0;

#--------------------------------------------------------------------------
# PRIVATE: validate_filer($filer)
#
# Validate the specified filer dotted IP or host name string.
#
# Arguments:
#
#   string $filer = Filer IP or FQDN (e.g. "192.168.0.1", 'filer.boo.com').
#
# Returns:
#
#   1 if valid.
#
# Exceptions:
#
#   Die if the value is invalid.
#
# Notes:
#
#   This function will perform a host lookup on any $filer argument to
#   resolve a host that is not just a simple IPv4 dotted address.
#
sub validate_filer($)
{
	defined($_[0]) || die "Filer IP/FQDN argument is not defined";
	my $filer = $_[0];
	#
	# Validate as a dotted IP and fall back to a host lookup
	#
	my $regex = '([01]?\d\d?|2[0-4]\d|25[0-5])';
	if (($filer !~ /^$regex\.$regex\.$regex\.$regex$/)) {
		if (!Socket::inet_aton($filer)) {
			die "Invalid filer IP or FQDN: \"$filer\"";
		}
	}
	return 1;
}

#--------------------------------------------------------------------------
# PRIVATE: validate_version($version)
#
# Validate the specified ONTAPI SDK version string.
#
# Arguments:
#
#   string $version = version string (e.g. ATI_ONTAPI_VERSION_1_3).
#
# Returns:
#
#   1 if valid.
#
# Exceptions:
#
#   Die if the value is invalid.
#
sub validate_version($)
{
	defined($_[0]) || die "ONTAPI SDK version argument is not defined";
	my $version = $_[0];
	#
	# Validate the ONTAPI or ONTAPI-D SDK version
	#
	if (($version ne ATI_ONTAPI_VERSION_1_3) &&
	    ($version ne ATI_ONTAPI_VERSION_1_4) &&
	    ($version ne ATI_ONTAPI_VERSION_1_5) &&
	    ($version ne ATI_ONTAPI_VERSION_1_6) &&
	    ($version ne ATI_ONTAPI_D_VERSION_1_1)) {
		die "Invalid ONTAPI SDK version argument: \"$version\"";
	}
	return 1;
}

#--------------------------------------------------------------------------
# PRIVATE: validate_transport($transport)
#
# Validate the specified ONTAPI transport protocol string.
#
# Arguments:
#
#   string $transport = protocol string (e.g. ATI_ONTAPI_TRANSPORT_HTTPS).
#
# Returns:
#
#   1 if valid.
#
# Exceptions:
#
#   Die if the value is invalid.
#
sub validate_transport($)
{
	defined($_[0]) || die "ONTAPI transport protocol argument is not defined";
	my $transport = $_[0];
	#
	# Validate the ONTAPI transport protocol
	#
	if (($transport ne ATI_ONTAPI_TRANSPORT_HTTP) &&
	    ($transport ne ATI_ONTAPI_TRANSPORT_HTTPS) &&
	    ($transport ne ATI_ONTAPI_TRANSPORT_D) &&
	    ($transport ne ATI_ONTAPI_TRANSPORT_DS)) {
		die "Invalid ONTAPI transport protocol argument: \"$transport\"";
	}
	return 1;
}

#--------------------------------------------------------------------------
# PRIVATE: validate_user($user)
#
# Validate the specified filer user name.
#
# Arguments:
#
#   string $user = Filer user name (e.g. 'root').
#
# Returns:
#
#   1 if valid.
#
# Exceptions:
#
#   Die if the value is invalid.
#
sub validate_user($)
{
	defined($_[0]) || die "Filer user name argument is not defined";
	my $user = $_[0];
	#
	# Validate the user name according to the following rules:
	#
	# 1) Length must be at least 1 but not more than 32
	# 2) Whitespace must only be real space characters
	# 3) At least one non-whitespace character
	# 4) No upper case characters
	# 5) Any alphanumeric or punctuation characters EXCEPT
	#    any of "*+,/:;<=>?[\]| or '\0' (null).
	#
	((length($user) > 0) && (length($user) <= 32)) ||
	  die "User name length is invalid: \"$user\"";
	$_[0] =~ s/\s/ /g;
	$user eq $_[0] ||
	  die "User name contains whitespace that are not space characters: \"$user\"";
	$user =~ /[\S]/ ||
	  die "User name must have at least one non-whitespace character: \"$user\"";
	$user !~ /[A-Z]/ ||
	  die "User name contains upper case characters: \"$user\"";
	$user !~ /[\"\*\+\,\/\:\;\<\=\>\?\[\]\|\0]/ ||
	  die "User name contains illegal characters: \"$user\"";
	return 1;
}

#--------------------------------------------------------------------------
# PRIVATE: validate_password($password)
#
# Validate the specified filer password.
#
# Arguments:
#
#   string $password = Filer password (e.g. 'password').
#
# Returns:
#
#   1 if valid.
#
# Exceptions:
#
#   Die if the value is invalid.
#
sub validate_password($)
{
	defined($_[0]) || die "Filer password argument is not defined";
	my $password = $_[0];
	#
	# Validate the volume name according to the following rules:
	#
	# 1) Length must be not more than 16
	# 2) At least one non-whitespace character
	#
	(length($password) <= 16) || die "Password length is invalid: \"$password\"";
	$_[0] =~ s/\s/ /g;
	$password eq $_[0] ||
	  die "Password contains whitespace that are not space characters: \"$password\"";
	return 1;
}

#--------------------------------------------------------------------------
# PRIVATE: validate_privilege($privilege)
#
# Validate a specified Data ONTAP execution privilege level.
#
# Arguments:
#
#   string $privilege = Privilege level (e.g. ATI_ONTAPI_PRIVILEGE_DIAG)
#
# Returns:
#
#   1 if valid.
#
# Exceptions:
#
#   Die if the value is invalid.
#
sub validate_privilege($)
{
	defined($_[0]) || die "Privilege level argument is not defined";
	my $privilege = $_[0];
	#
	# Validate the privilege level
	#
	if (($privilege ne ATI_ONTAPI_PRIVILEGE_ADMIN) &&
	    ($privilege ne ATI_ONTAPI_PRIVILEGE_ADVANCED) &&
	    ($privilege ne ATI_ONTAPI_PRIVILEGE_DIAG) &&
	    ($privilege ne ATI_ONTAPI_PRIVILEGE_TEST)) {
		die "Invalid privilege level argument value: \"$privilege\"";
	}
	return 1;
}

#--------------------------------------------------------------------------
# PRIVATE: validate_timeout($timeout)
#
# Validate a specified transport timeout value in seconds.
#
# Arguments:
#
#   string $timeout = Timeout in seconds (e.g. "180").
#
# Returns:
#
#   1 if valid.
#
# Exceptions:
#
#   Die if the value is invalid.
#
sub validate_timeout($)
{
	defined($_[0]) || die "Timeout argument is not defined";
	my $timeout = $_[0];
	#
	# Validate the timeout integer value
	#
	$timeout =~ /^[0-9]+$/ ||
	  die "Timeout value must be an integer: \"$timeout\"";
	return 1;
}

#--------------------------------------------------------------------------
# PRIVATE: $xml = array_to_xml($xml, $xmlnode, \@tree)
#
# Transcode a deep "tree" array structure into an XML::LibXML::Document.
#
# Arguments:
#
#   XML::LibXML::Document $xml     = XML document to be filled
#   XML::LibXML::Node     $xmlnode = Node in document below which to insert
#   arrayref             \@tree    = Arrayref to deep "tree" array
#
# Returns:
#
#   $xml = filled XML document (XML::LibXML::Document)
#
# Exceptions:
#
#   Die on error.
#
# Example 1: Generate XML for volume-list-info API call without attributes
#
#   my $xml = array_to_xml($xml, $netapp, 0,
#                          ['volume-list-info',
#                              ['volume', 'vol0']
#                          ]);
#
#   Will insert this chunk of XML below the $netapp document element:
#
#     <volume-list-info>
#         <volume>vol0</volume>
#     </volume-list-info>
#
# Notes:
#
#   To insert into an empty document simply use $xmlnode == $xml.
#
sub array_to_xml
{
	my $xml = shift;
	my $xmlnode = shift;
	my @iterator = @{(shift)};
	#
	# Iterate through the entire tree array
	#
	while (@iterator > 0) {
		#
		# Create the start tag for this node
		#
		my $tag = shift @iterator;
		my $element = $xml->createElement($tag);
		#
		# Recurse or write value and end current tag
		#
		my $value = shift @iterator;
		if (ref($value) eq 'ARRAY') {
			array_to_xml($xml, $element, \@{$value});
		} else {
			$element->appendTextNode($value);
		}
		#
		# Add the complete child to the parent node
		#
		$xmlnode->appendChild($element);
	}
	return $xml;
}

#--------------------------------------------------------------------------
# PRIVATE $xml = ati_ontapi_array_to_xml(<named args>)
#
# Create an ONTAPI XML request from a deep "tree" array.
#
# Arguments:
#
#   OPT string    $version = ONTAPI SDK version (def. ATI_ONTAPI_VERSION_DEFAULT)
#   REQ arrayref \@tree    = Reference to deep "tree" array
#
# Returns:
#
#   $xml = XML::LibXML::Document containing XML request
#
# Exceptions:
#
#   Die on error
#
# Example 1:
#
#   ati_ontapi_array_to_xml(tree => ['volume-list-info', [],
#                                      ['volume', [], 'vol0']
#                                   ];
#
#   Will return this XML::LibXML::Docmuent:
#
#     <?xml version="1.0" encoding="UTF-8"?>
#     <!DOCTYPE  SYSTEM "file:/etc/netapp_filer.dtd">
#     <netapp version="1.3" xmlns="http://www.netapp.com/filer/admin">
#         <volume-list-info>
#             <volume>vol0</volume>
#         </volume-list-info>
#     </netapp>
#
sub ati_ontapi_array_to_xml
{
	#
	# Specify all the named arguments and their defaults
	#
	my %args = (
		    version   => ATI_ONTAPI_VERSION_DEFAULT,
		    tree      => undef
		   );
	#
	# Apply the arguments over the top of the default values.
	# Bitch if we run into any arguments we do not recognize.
	#
	while (@_ > 0) {
		my $name = shift;
		my $value = shift;
		if (!exists($args{$name})) {
			die "Unknown argument \"$name\" = \"$value\"";
		}
		$args{$name} = $value;
	}
	#
	# Verify that the required arguments are defined
	#
	defined($args{version}) || die "Required argument \"version\" is undefined";
	defined($args{tree})    || die "Required argument \"tree\" is undefined";
	#
	# Cache function arguments from hash
	#
	my $version   = $args{version};
	my $tree      = $args{tree};
	#
	# Validate the arguments
	#
	validate_version($version);
#	validate_tree($tree);
	#
	# Construct the XML header and document type
	#
	my $xml = XML::LibXML::Document->new('1.0', 'UTF-8');
	$xml->createInternalSubset('netapp', undef, 'file:/etc/netapp_filer.dtd');
	#
	# Create the wrapping "netapp" tag for the XML request
	#
	my $netapp = $xml->createElementNS('http://www.netapp.com/filer/admin', 'netapp');
	$netapp->setAttribute('version', $version);
	$xml->setDocumentElement($netapp);
	#
	# Transcode the array "tree" into the XML document
	#
	return array_to_xml($xml, $netapp, \@{$tree});
}

#--------------------------------------------------------------------------
# PRIVATE $hashref = xml_to_hash_internal($xml, $root)
#
# Convert an XML document into a convenient deep array/hash structure.
#
# Arguments:
#
#   REQ XML::LibXML::Element $xml = XML::LibXML::Node
#   REQ $root                     = non-zero if this is the root XML node.
#
# Returns:
#
#   $ref = hash reference for root of deep array/hash structure.
#
# Exceptions:
#
#   Die on error
#
# Example 1: Convert ONTAPI volume-list-info XML response to a deep array
#
#   my $result = xml_to_hash($xml, 1);
#
#   XML:                               DEEP STRUCTURE:
#                                      {
#   <volumes>                            'volumes' = [
#     <volume-info>                        {
#       <name>vol0</name>                    'name' => 'vol0',
#       <plex-count>1</plex-count>           'plex-count' => 1,
#       <plexes>                             'plexes' = [
#         <plex-info>                          {
#           <name>/vol0/plex0</name>             'name' => '/vol0/plex0',
#           <is-online>true</is-online>          'is-online' => 'true',
#           <raid-groups>                        'raid-groups' = [
#             <raid-group-info>                    {
#               <name>/vol0/plex0/rg0</name>         'name => '/vol0/plex0/rg0',
#               <disks>                              'disks' = [
#                 <disk-info>                          {
#                   <name>2a.16></name>    ==>           'name' => '2a.16',
#                 </disk-info>             ==>         }
#                 <disk-info>              ==>         {
#                   <name>2a.17></name>                  'name' => '2a.17',
#                 </disk-info>                         }
#               </disks>                             ]
#             </raid-group-info>                   }
#           </raid-groups>                       ]
#         </plex-info>                         }
#       </plexes>                            ]
#     </volume-info>                       }
#   </volumes>                           ]
#                                      }
#
# Example 2: Access the name of the second disk in /vol0/plex0/rg0
#
#   $disk = $$result{'volumes'}[0]{'plexes'}[0]{'raid-groups'}[0]{'disks'}[1]{'name'}
#
# Notes:
#
#   It is important to note that this function is tailored specifically to
#   handle XML produced by ONTAPI. In general, it is not possible to convert
#   arbitrary XML to anything but lists of lists because of strict ordering
#   and arbitrary repetition of elements.
#
#   The usefulness of this function lies in the fact that is knows that the
#   XML produced by ONTAPI has these invariants:
#
#   1) Nodes with repeated non-leaf elements are homogenous arrays.
#   2) An element with a single non-leaf child is a homogenous array.
#   3) Non-leaf nodes never contain any PCDATA or CDATA.
#
#   This knowledge allows the function to produce anonymous hashes for the
#   interesting parts of the returned deep structure which enables efficient
#   and convenient lookup syntax as demonstrated in Example 2.
#
sub xml_to_hash_internal
{
	#
	# Check that the expected arguments are defined
	#
	defined($_[0]) || die "XML node reference argument is not defined";
	#
	# Cache function arguments
	#
	my $xml = $_[0];
	#
	# Find the child leaf and child tree nodes
	#
	my @child_leafs = $xml->findnodes("*[not(*)]");
	my @child_trees = $xml->findnodes("*[*]");
	#
	# If all of the child nodes are trees and they all have the
	# same name then we assume that this node is a homogenous
	# array of elements.
	#
	my %hash = ();
	my $child;
	if ((@child_leafs == 0) && (@child_trees > 0)) {
		#
		# Check if all the child tree nodes have matching names
		#
		my $match = 1;
		foreach $child (@child_trees) {
			if ($child->nodeName ne $child_trees[0]->nodeName) {
				$match = 0;
				last;
			}
		}
		if ($match != 0) {
			#
			# Recursively process all child tree nodes
			#
			my @list = ();
			foreach $child (@child_trees) {
				push (@list, xml_to_hash_internal($child));
			}
			return \@list;
		}
	}
	#
	# Add all child leaf and tree nodes to the hash
	#
	foreach $child (@child_leafs) {
		$hash{$child->nodeName} = $child->textContent;
	}
	foreach $child (@child_trees) {
		$hash{$child->nodeName} = xml_to_hash_internal($child, 0);
	}
	return \%hash;
}

#
# Entry point should be called with the parent of the XML root node.
#
# This separate entry point is to ensure that the children of the root
# node go into a hash rather than a list. This relies on the assumption
# that ONTAPI never returns a homogenous array as the root element.
#
sub xml_to_hash
{
	#
	# Check that the expected arguments are defined
	#
	defined($_[0]) || die "XML node reference argument is not defined";
	#
	# Cache function arguments
	#
	my $xml = $_[0];
	#
	# Find the child leaf and child tree nodes
	#
	my @child_leafs = $xml->findnodes("*[not(*)]");
	my @child_trees = $xml->findnodes("*[*]");
	#
	my %hash = ();
	my $child;
	foreach $child (@child_leafs) {
		$hash{$child->nodeName} = $child->textContent;
	}
	foreach $child (@child_trees) {
		$hash{$child->nodeName} = xml_to_hash_internal($child, 0);
	}
	return \%hash;
}

#--------------------------------------------------------------------------
# PRIVATE: @out = ati_ontapi_send_to_filer($context, $xml)
#
# Send an XML encoded ONTAPI request to the filer over the transport
# protocol specified in the ATI connection context.
#
# Arguments:
#
#   REQ \%hash            context = ATI connection context.
#   REQ XML::LibXML::Document xml = XML request to send
#
# Returns:
#
#   $root = XML document element node reference.
#
# Exceptions:
#
#   Die on error.
#
sub ati_ontapi_send_to_filer($$)
{
	#
	# Check that the expected arguments are defined
	#
	defined($_[0]) || die "ATI context argument is not defined";
	defined($_[1]) || die "XML node reference argument is not defined";
	#
	# Cache function arguments
	#
	my $context = $_[0];
	my $xml     = $_[1];
	my $timeout = $context->{timeout};
	#
	# Debug output for XML we are about to send
	#
	if ($context->{debug} >= ATI_ONTAPI_DEBUG_VERBOSE) {
		print "DBG(TERSE): Sending this ONTAPI XML call:\n", $xml->toString(2);
	}
	#
	# Validate the XML request against the DTD
	#
	my $dtd;
	if ($context->{validate} == ATI_ONTAPI_VALIDATE_ON) {
		$dtd = XML::LibXML::Dtd->new('/na_admin/netapp_filer.dtd', 'netapp_filer.dtd');
		$xml->validate($dtd);
	}
	my $response;
	#
	# Select the appropriate transport over which to send the request
	#
	my $transport = $context->{transport};
	if (($transport eq ATI_ONTAPI_TRANSPORT_D) ||
	    ($transport eq ATI_ONTAPI_TRANSPORT_DS)) {
		#
		# Determine if we are using an SSL connection
		#
		my $use_ssl = ($transport eq ATI_ONTAPI_TRANSPORT_DS);
		#
		# Open a connection to the ONTAPI-D server
		#
		my $port = ($use_ssl) ? ATI_ONTAPI_DS_SERVER_PORT : ATI_ONTAPI_D_SERVER_PORT;
		my $socket = IO::Socket::INET->new(PeerAddr => $context->{filer},
						   PeerPort => $port,
						   Proto    => "tcp",
						   Type     => SOCK_STREAM,
						   Timeout  => $timeout)
		  || die "Failed to connect to ONTAPI-D server $context->{filer}";
		$socket->autoflush(1);
		#
		# Fire up OpenSSL for this connection if necessary
		#
		my $ctx;
		my $ssl;
		if ($use_ssl) {
		    # This is here so we can run tests that don't require SSL on clients that do not have SSL.
		    # Tests that require SSL will continue to fail on clients that do not have SSL.
		    if (!$Ssl_initialized) {
			$Ssl_initialized=1;
			require Net::SSLeay;
			Net::SSLeay::load_error_strings();
			Net::SSLeay::SSLeay_add_ssl_algorithms();
			Net::SSLeay::randomize();
		    }

			$ctx = Net::SSLeay::CTX_new() || die "Failed to create a SSL context";
			$ssl = Net::SSLeay::new($ctx) || die "Failed to create SSL $!";
			Net::SSLeay::set_fd($ssl, fileno($socket));
			Net::SSLeay::connect($ssl);
			#
			# For debugging print out the cipher in use
			#
			if ($context->{debug} >= ATI_ONTAPI_DEBUG_DEBUG) {
				print "DBG(VERBOSE): Using cipher " . Net::SSLeay::get_cipher($ssl) . "\n";
			}
		}
		#
		# Generate the ONTAPI-D headers plus the XML request body
		#
		# $xml->toString() terminates with LF which we replace with CRLF
		#
		my $reqxml = $xml->toString();
		chomp($reqxml);
		$reqxml .= "\r\n";
		my $length = length($reqxml);
		my $request = "ONTAPI-D/1.0\r\n" .
		              "User: $context->{user}\r\n" .
		              "Password: $context->{password}\r\n" .
		              "Content-length: $length\r\n" .
			      "\r\n" . $reqxml;
		#
		# For debugging print out the request we are sending
		#
		if ($context->{debug} >= ATI_ONTAPI_DEBUG_VERBOSE) {
			print "DBG(TERSE): Sending this ONTAPI-D request:\n" . $request;
		}
		#
		# Send the request to the ONTAPI-D server
		#
		if ($use_ssl) {
			Net::SSLeay::write($ssl, $request);
			die_if_ssl_error("Failed to send XML request to ONTAPI-D server");
		} else {
			print($socket $request) || die "Failed to send XML request to ONTAP-D server";
		}
		#
		# Wait for the first header line of the response from the server
		#
		local $SIG{ALRM} = sub { die "Timeout reading response line 1 from ONTAPI-D server\n"; };
		alarm $timeout;
		if ($use_ssl) {
			$response = Net::SSLeay::ssl_read_CRLF($ssl);
			die_if_ssl_error("Failed to read response line 1 from ONTAPI-D server");
		} else {
			$response = <$socket> || die "Failed to read response line 1 from ONTAPI-D server";
		}
		alarm 0;
		$response =~ s/\s+$//;
		if ($response !~ /^ONTAPI-D\/1.0 ([0-9]{3}) OK$/) {
			die "Invalid response line 1 from ONTAPI-D server: \"$response\"\n";
		}
		#
		# Wait for the second header line of the response from the server
		#
		local $SIG{ALRM} = sub { die "Timeout reading response line 2 from ONTAPI-D server\n"; };
		alarm $timeout;
		if ($use_ssl) {
			$response = Net::SSLeay::ssl_read_CRLF($ssl);
			die_if_ssl_error("Failed to read response line 1 from ONTAPI-D server");
		} else {
			$response = <$socket> || die "Failed to read response line 1 from ONTAPI-D server";
		}
		alarm 0;
		$response =~ s/\s+$//;
		if ($response !~ /^Content-length: ([0-9]+)$/) {
			die "Invalid response line 2 from ONTAPI-D server: \"$response\"\n";
		}
		$length = $1;
		#
		# Wait for the second third line of the response from the server
		#
		local $SIG{ALRM} = sub { die "Timeout reading response line 3 from ONTAPI-D server\n"; };
		alarm $timeout;
		if ($use_ssl) {
			$response = Net::SSLeay::ssl_read_CRLF($ssl);
			die_if_ssl_error("Failed to read response line 3 from ONTAPI-D server");
		} else {
			$response = <$socket> || die "Failed to read response line 3 from ONTAPI-D server";
		}
		alarm 0;
		$response =~ s/\s+$//;
		if ($response ne '') {
			die "Invalid response line 3 from ONTAPI-D server: \"$response\"\n";
		}
		#
		# Now read the entire response body
		#
		$response = "";
		while ($length > 0) {
			my $t = "<uninitialized>";
			my $rc = undef;
			local $SIG{ALRM} = sub {
				die "Timeout reading content from ONTAPI-D server. $length bytes to go.\n";
			};
			alarm $timeout;
			if ($use_ssl) {
				$t = Net::SSLeay::ssl_read_CRLF($ssl);
				die_if_ssl_error("Failed to read content from ONTAPI-D server: \n\"$t\"\n");
			} else {
				read($socket, $t, $length) || die "Failed to read response body from ONTAPI-D server";
			}
			alarm 0;
			$response .= $t;
			$length -= length($t);
		}
		#
		# Close down the connection
		#
		if ($use_ssl) {
			Net::SSLeay::free($ssl);
			Net::SSLeay::CTX_free($ctx);
		}
		close($socket);
		#
		# Quick sanity check to ensure we did not receive too much data
		#
		($length == 0) || die "Content length mismatch in ONTAPI-D response";

	} elsif (($transport eq ATI_ONTAPI_TRANSPORT_HTTP) ||
		 ($transport eq ATI_ONTAPI_TRANSPORT_HTTPS)) {
		#
		# Send the XML request via HTTP to the server
		#
		my $http_ua = new LWP::UserAgent;
		$http_ua->timeout($timeout);
		my $http_url = "$transport://$context->{filer}/servlets/netapp.servlets.admin.XMLrequest_filer";
		my $http_request = HTTP::Request->new('POST', "$http_url");
		$http_request->authorization_basic($context->{user}, $context->{password});
		my $http_content = $xml->toString();
		$http_request->content($http_content);
		$http_request->content_length(length($http_content));
		my $http_response = $http_ua->request($http_request);
		#
		# Check the HTTP response from the filer
		#
		$http_response->is_success || die "HTTP(S) failure: ", $http_response->status_line;
		$response = $http_response->content();

	} else {
		die "Unknown ONTAPI transport \"$transport\" specified";
	}
	#
	# Parse and XML response from the filer
	#
	my $parser = XML::LibXML->new();
	$parser->expand_entities(0);
	my $xml_response = $parser->parse_string($response);
	#
	# Debug output for XML response we received
	#
	if ($context->{debug} >= ATI_ONTAPI_DEBUG_VERBOSE) {
		print "DBG(TERSE): Got this ONTAPI XML response:\n", $xml_response->toString(2);
	}
	#
	# Validate the XML response against the DTD
	#
	if ($context->{validate} == ATI_ONTAPI_VALIDATE_ON) {
		$xml_response->validate($dtd);
	}
	#
	# We need to set the namespace because XPath queries will not work
	# on the tags in the default namespace. Use "na" for the namespace.
	#
	my $root = $xml_response->getDocumentElement();
	$root->setNamespace('http://www.netapp.com/filer/admin', 'netapp', 0);
	return $root;
}

#--------------------------------------------------------------------------
# PRIVATE: @out = ati_ontapi_execute_xml(<named args>)
#
# Execute an ONTAP API call using XML over HTTP(S).
#
# Arguments:
#
#   REQ \%hash             context = ATI connection context.
#   REQ  XML::LibXML::Document xml = XML request to send
#
# Returns:
#
#   @out = ($status, $result, $errno, $reason)
#
#   $string  status  =  ATI_SUCCESS if successful.
#                    =  ATI_FAILURE if failure.
#   \%hash   result  =  Deep structure for API call result on ATI_SUCCESS.
#   $integer errno   =  Error number on ATI_FAILURE.
#   $string  reason  =  Error reason string on ATI_FAILURE.
#
# Exceptions:
#
#   Die on error.
#
sub ati_ontapi_execute_xml
{
	#
	# Specify all the named arguments and their defaults
	#
	my %args = (context => undef,
		    xml     => undef);
	#
	# Apply the arguments over the top of the default values.
	# Bitch if we run into any arguments we do not recognize.
	#
	while (@_ > 0) {
		my $name = shift;
		my $value = shift;
		exists($args{$name}) || die "Unknown argument \"$name\" = \"$value\"";
		$args{$name} = $value;
	}
	#
	# Verify that the required arguments are defined
	#
	defined($args{context}) || die "Required argument \"context\" is undefined";
	my $context = $args{context};
	defined($args{xml}) || die "Required argument \"xml\" is undefined";
	my $xml = $args{xml};
	#
	# Validate the arguments
	#
#	validate_xml($xml);
	#
	# Send the the request to the filer
	#
	my $root = ati_ontapi_send_to_filer($context, $xml);
	#
	# Get and validate the status and reason attributes of the results tag
	#
	my @nodes = $root->findnodes('/netapp:netapp/netapp:results[@status]');
	if (@nodes != 1) {
		die "No results tag in the XML response from ONTAPI";
	}
	my $status = $nodes[0]->getAttribute('status');
	my $errno  = $nodes[0]->getAttribute('errno');
	my $reason = $nodes[0]->getAttribute('reason');
	#
	# Convert the returned XML result to deep array format
	#
	my $result = undef;
	if ($status eq ATI_ONTAPI_STATUS_PASSED) {
		$status = ATI_SUCCESS;
		$result = xml_to_hash($nodes[0]);
	} elsif ($status eq ATI_ONTAPI_STATUS_FAILED) {
		$status = ATI_FAILURE;
	} else {
		die "Invalid status attribute in the XML response from ONTAPI: \"$status\"";
	}
	#
	# Return the XML response subtree below the results tag
	#
	return ($status, $result, $errno, $reason);
}

#--------------------------------------------------------------------------
# PUBLIC: @out = ati_ontapi_create_context(<named args>)
#
# Create an ATI context structure for use in all interactions with ATI.
#
# This subroutine will connect to the filer and download the API and type
# names via the reflective APIs (when they are available) and automatically
# persistently cache this information. If the reflective APIs are not
# available in the version of ONTAP running on the filer (SDK <= 1.6) then
# automatically fall back to a static pre-generated cache file.
#
# Arguments:
#
#   REQ  $string  filer      = Filer IP or FQDN (e.g. '192.168.0.1').
#   OPT  $string  version    = ONTAPI version (ATI_ONTAPI_VERSION_DEFAULT).
#   OPT  $string  transport  = Transport protocol (ATI_ONTAPI_TRANSPORT_DEFAULT).
#   OPT  $string  timeout    = Timeout value to use, in seconds (ATI_ONTAPI_TIMEOUT_DEFAULT).
#   OPT  $string  user       = Filer login user name (ATI_ONTAPI_USER_DEFAULT).
#   OPT  $string  password   = Filer login password (ATI_ONTAPI_PASSWORD_DEFAULT).
#   OPT  $string  cache_dir  = Cache directory (ATI_ONTAPI_DEFAULT_CACHE_DIR).
#   OPT  $string  cache_file = Cache directory (ATI_ONTAPI_DEFAULT_CACHE_FILE).
#   OPT  $string  cache_all  = Cache everything (ATI_ONTAPI_DEFAULT_CACHE_ALL).
#   OPT  $integer validate   = XML validation against DTD (ATI_ONTAPI_VALIDATE_DEFAULT).
#   OPT  $integer debug      = Debug level (ATI_ONTAPI_DEBUG_NONE).
#
# Returns:
#
#   \%context = Hash reference for ATI context.
#
# Exceptions:
#
#   Die on error
#
# Example:  Create an ATI context for filer nulb.lab.netapp.com using the
#           credentials of user "admin" and password "sp00ge over a secure
#           HTTPS transport with a timeout of 30 seconds (instead of the
#           default action of waiting forever for a reply).
#
#   $context = ati_ontapi_create_context(filer     => 'nulb.lab.netapp.com',
#                                        transport => ATI_ONTAPI_TRANSPORT_HTTPS,
#                                        timeout   => 30,
#                                        user      => 'admin',
#                                        password  => 'sp00ge');
#
sub ati_ontapi_create_context
{
	#
	# Define all the context parameters and their defaults
	#
	my %context = (filer      => undef,
		       version    => undef,
		       transport  => ATI_ONTAPI_TRANSPORT_DEFAULT,
		       timeout    => ATI_ONTAPI_TIMEOUT_DEFAULT,
		       user       => ATI_ONTAPI_USER_DEFAULT,
		       password   => ATI_ONTAPI_PASSWORD_DEFAULT,
		       cache_dir  => ATI_ONTAPI_CACHE_DIR_DEFAULT,
		       cache_file => ATI_ONTAPI_CACHE_FILE_DEFAULT,
		       cache_all  => ATI_ONTAPI_CACHE_ALL_DEFAULT,
		       validate   => ATI_ONTAPI_VALIDATE_DEFAULT,
		       debug      => ATI_ONTAPI_DEBUG_DEFAULT);
	#
	# Apply the arguments over the top of the default values.
	# Bitch if we run into any parameters we do not recognize.
	#
	while (@_ > 0) {
		my $name = shift;
		my $value = shift;
		exists($context{$name}) || die "Unknown context parameter \"$name\" = \"$value\"";
		$context{$name} = $value;
	}
	#
	# The filer context parameter is the only mandatory parameter
	#
	defined($context{filer}) || die "Required context parameter \"filer\" not specified.";
	#
	# If no explicit ONTAPI protocol version number is specified, divine an
	# appropriate default version number from the transport type.
	#
	if (!defined($context{version})) {
		if (($context{transport} eq ATI_ONTAPI_TRANSPORT_HTTP) ||
		    ($context{transport} eq ATI_ONTAPI_TRANSPORT_HTTPS)) {
			$context{version} = ATI_ONTAPI_VERSION_DEFAULT;
		} else {
			$context{version} = ATI_ONTAPI_D_VERSION_DEFAULT;
		}
	}
	#
	# Validate all the base parameters
	#
	validate_filer($context{filer});
	validate_version($context{version});
	validate_transport($context{transport});
	validate_timeout($context{timeout});
	validate_user($context{user});
	validate_password($context{password});
	#
	# Attempt to get the ONTAPI version number. Use this to automatically fall
	# back to a pre-built cached context for pre-reflective API releases. This
	# fallback cache allows us to talk to such releases without relying upon
	# the reflective APIs at all.
	#
	my @out = ati_ontapi_execute(context => \%context,
				     request => ['system-get-ontapi-version', []]);
	($out[0] == ATI_SUCCESS) || die "Failed to get ONTAPI version: $out[2]: $out[3]\n";
	#
	# Do version checking for both ONTAPI-C (1.6) and ONTAPI-D (1.1)
	#
	my $transport = $context{transport};
	my $major = $out[1]->{'major-version'};
	my $minor = $out[1]->{'minor-version'};
	if (((($transport eq ATI_ONTAPI_TRANSPORT_HTTP) ||
	      ($transport eq ATI_ONTAPI_TRANSPORT_HTTPS)) && ($major == 1) && ($minor < 6)) ||
	    ((($transport eq ATI_ONTAPI_TRANSPORT_D) ||
	      ($transport eq ATI_ONTAPI_TRANSPORT_DS)) && ($major == 1) && ($minor < 1))) {
		#
		# If cache_all is specified to force downloading of all API/type
		# information then it is an error to be talking to a filer running
		# this version of ONTAPI as it has no reflective APIs.
		#
		$context{cache_all} && die "Filer must be have ONTAPI V1.6 or later (ONTAP 7.2.x)";
		#
		# Update the cache directory and file to point to the fallback
		# cache file in the ATI library directory for pre-7.0 releases.
		#
		my $path = $INC{'ATI/ati_ontapi.pm'};
		$path =~ s/\/*ati_ontapi\.pm$//;
		$context{cache_dir} = $path;
		$context{cache_file} = "fallback_cache.ati";
		#
		# Verify that the fallback cache file actually exists
		#
		$path .= "/" . $context{cache_file};
		(-f $path) || die "API/type fallback cache not found \[$path\]";
	}
	#
	# Attempt to get the ONTAP version string for the target.
	#
	@out = ati_ontapi_execute(context => \%context,
				  request => ['system-get-version', []]);
	($out[0] == ATI_SUCCESS) || die "Failed to get ONTAP version: $out[2]: $out[3]\n";
	#
	# Compute SHA1 of version and use it for the base name for persistent files:
	#
	# API:   <cachedir>/<sha1_digest>.ati
	# Types: <cachedir>/<sha1_digest>.ati
	#
	# Note that if cache_file is supplied it is used in place of <sha1_digest>.
	#
	# XXX - Until Digest::SHA1 is intalled we have to shell out to get the SHA1.
	#
	my $cache_path = $context{cache_dir} . "/";
	if (defined($context{cache_file})) {
		$cache_path .= $context{cache_file};
	} else {
#		my $digest = Digest::SHA1::sha1_hex($out[1]->{'version'});
		my @digest = split(/ /, `echo \'$out[1]->{'version'}\' | sha1sum`);
		$cache_path .= $digest[0] . ".ati";
	}
	#
	# Attempt to load any cached context except for cache_all case
	#
	my $loaded_from_cache = 0;
	if (!$context{cache_all}) {
		if ($context{debug} >= ATI_ONTAPI_DEBUG_TERSE) {
			print "DBG(TERSE): Attempting to load cached context \[$cache_path\]\n";
		}
		my $cached_context;
		eval { $cached_context = Storable::lock_retrieve($cache_path) };
		if (defined($cached_context)) {
			if ($context{debug} >= ATI_ONTAPI_DEBUG_TERSE) {
				print "DBG(TERSE): Cached context loaded\n";
			}
			$context{type_hash} = $cached_context->{type_hash};
			$context{api_hash} = $cached_context->{api_hash};
			$loaded_from_cache = 1;
		} else {
			if ($context{debug} >= ATI_ONTAPI_DEBUG_TERSE) {
				print "DBG(TERSE): Cached context not found. Regenerating...\n";
			}
		}
	}
	#
	# Try to read the persistently cached type hash for this ONTAP version.
	# If that fails then pull the information from the filer using the
	# reflective API interfaces and build the type hash from scratch.
	#
	if (!defined($context{type_hash})) {
		#
		# Query all supported API types via the reflective interface
		#
		# This must be done AFTER the basic context initialization is complete.
		#
		@out = ati_ontapi_execute(context => \%context,
					  request => ['system-api-list-types', []]);
		($out[0] == ATI_SUCCESS) || die "Failed to query API types: $out[2]: $out[3]\n";
		#
		# Generate a hash table of all API types (key=name, value=coderef)
		#
		# Also create a hash table entry for the array version (key=name[], value=coderef)
		#
		my %type_hash;
		$context{type_hash} = \%type_hash;
		my @type_entry_info_array = @{$out[1]->{'type-entries'}};
		foreach my $type_entry_info (@type_entry_info_array) {
			my $name = $type_entry_info->{'name'};
			my $element_info_array = $type_entry_info->{'type-elements'};
			$type_hash{$name} = [$element_info_array, undef];
			$type_hash{"$name\[\]"} = [$element_info_array, undef];
		}
	}
	#
	# Try to read the persistently cached API hash for this ONTAP version.
	# If that fails then pull the information from the filer using the
	# reflective API interfaces and build the API hash from scratch.
	#
	if (!defined($context{api_hash})) {
		#
		# Query all supported filer APIs via the reflective interface
		#
		# This must be done AFTER the basic context initialization is complete.
		#
		@out = ati_ontapi_execute(context => \%context,
					  request => ['system-api-list', []]);
		($out[0] == ATI_SUCCESS) || die "Failed to query APIs: $out[2]: $out[3]\n";
		#
		# Generate a hash table of all APIs (key=name, value=coderef)
		#
		my %api_hash;
		$context{api_hash} = \%api_hash;
		my @api_info_array = @{$out[1]->{'apis'}};
		foreach my $api_info (@api_info_array) {
			my $name = $api_info->{'name'};
			$api_hash{$name} = [undef, undef];
			#
			# Fetch all API information if we are doing cache_all. Do
			# not compile the code. Just fetch the API element lists.
			#
			if ($context{cache_all}) {
				if ($context{debug} >= ATI_ONTAPI_DEBUG_TERSE) {
					print "DBG(TERSE): Caching API \[$name\]\n";
				}
				build_api_code(\%context, $name, 0);
			}
		}
	}
	#
	# Store the hash persistently in the cache directory only if it was not
	# loaded from the cache on disk. That is, do not overwrite an existing
	# cache file unless we are doing cache_all.
	#
	if (!$loaded_from_cache) {
		if ($context{debug} >= ATI_ONTAPI_DEBUG_TERSE) {
			print "DBG(TERSE): Writing persistent cache file \[$cache_path\]\n";
		}
		eval { Storable::lock_nstore(\%context, $cache_path) };
		($@) && die "Failed to store persistent cache file \[$cache_path\]: $@";
	}
	#
	# Return the hash reference
	#
	return \%context;
}

#--------------------------------------------------------------------------
# PUBLIC: @out = ati_ontapi_execute(<named args>)
#
# Execute an ONTAP API call using XML over HTTP(S).
#
# Arguments:
#
#   REQ \%hash  context = ATI connection context.
#   REQ \@array request = Deep array containing XML request.
#
# Returns:
#
#   @out = ($status, $result, $errno, $reason)
#
#   $string  status  =  ATI_SUCCESS if successful.
#                    =  ATI_FAILURE if failure.
#   \%hash   result  =  Deep structure for API call result on ATI_SUCCESS.
#   $integer errno   =  Error number on ATI_FAILURE.
#   $string  reason  =  Error reason string on ATI_FAILURE.
#
# Exceptions:
#
#   Die on error.
#
# Example:  Get volume information for volume "vol0".
#
#   @out = ati_ontapi_execute(context => $context,
#                             request => ['volume-list-info',
#                                            ['volume', 'vol0']
#                                        ]);
#
sub ati_ontapi_execute
{
	#
	# Specify all the named arguments and their defaults
	#
	my %args = (context => undef,
		    request => undef);
	#
	# Apply the arguments over the top of the default values.
	# Bitch if we run into any arguments we do not recognize.
	#
	while (@_ > 0) {
		my $name = shift;
		my $value = shift;
		exists($args{$name}) || die "Unknown argument \"$name\" = \"$value\"";
		$args{$name} = $value;
	}
	#
	# Verify that the required arguments are defined
	#
	defined($args{context}) || die "Required argument \"context\" is undefined";
	my $context = $args{context};
	defined($args{request}) || die "Required argument \"request\" is undefined";
	my $request = $args{request};
	#
	# Convert the tree array to XML
	#
	my $xml = ati_ontapi_array_to_xml(version => $context->{version},
					  tree    => $request);
	#
	# Execute the XML request on the server
	#
	return ati_ontapi_execute_xml(context => $context,
				      xml     => $xml);
}

#--------------------------------------------------------------------------
# PUBLIC: @out = ati_ontapi_call_cli(<named args>)
#
# Execute a CLI command via the ONTAPI interface over HTTP(S)
#
# Arguments:
#
#   REQ \%hash  context = ATI connection context.
#   REQ \@array request = Deep array containing XML request.
#
# Returns:
#
#   @out = ($status, $output, $errno)
#
#   $string  status = ATI_SUCCESS if XML response indicated pass
#                       = ATI_FAILURE if XML response indicated failure
#   $string  output = String output of CLI command
#   $integer errno  = Status code returned by the CLI (usually 0 #?$!)
#
# Exceptions:
#
#   Die on error.
#
# Example:  Get volume information for volume "vol0".
#
#   @out = ati_ontapi_call_cli(context   => $context,
#                              privilege => $privilege,
#                              command   => 'vol status vol0');
#
sub ati_ontapi_call_cli
{
	#
	# Specify all the named arguments and their defaults
	#
	my %args = (context   => undef,
		    privilege => ATI_ONTAPI_PRIVILEGE_DEFAULT,
		    command   => undef);
	#
	# Apply the arguments over the top of the default values.
	# Bitch if we run into any arguments we do not recognize.
	#
	while (@_ > 0) {
		my $name = shift;
		my $value = shift;
		exists($args{$name}) || die "Unknown argument \"$name\" = \"$value\"";
		$args{$name} = $value;
	}
	#
	# Verify that the required arguments are defined
	#
	defined($args{context}) || die "Required argument \"context\" is undefined";
	my $context = $args{context};
	defined($args{privilege}) || die "Required argument \"privilege\" is undefined";
	my $privilege = $args{privilege};
	defined($args{command}) || die "Required argument \"command\" is undefined";
	my $command = $args{command};
	#
	# Validate the arguments
	#
	validate_privilege($privilege);
	#
	# Slice up the command into argument components for the API
	#
	my $xml = XML::LibXML::Document->new('1.0', 'utf-8');
	$xml->createInternalSubset('netapp', undef, 'file:/etc/netapp_filer.dtd');
	#
	# Construct the XML headers and basic hierarchy for this call
	#
	my $root = $xml->createElementNS('http://www.netapp.com/filer/admin', 'netapp');
	$root->setAttribute('version', $context->{version});
	$xml->setDocumentElement($root);
	my $func = $xml->createElement('system-cli');
	#
	# Add the execution privilege element
	#
	my $element = $xml->createElement('priv');
	$element->appendText($privilege);
	$func->appendChild($element);
	#
	# Create the root element for the argument vector elements
	#
	my $args = $xml->createElement('args');
	$root->appendChild($func);
	$func->appendChild($args);
	#
	# Need to split the command into separate <arg> tags
	#
	my @splitcmd = split(" ", $command);
	while (@splitcmd > 0) {
		my $element = $xml->createElement('arg');
		$element->appendText(shift @splitcmd);
		$args->appendChild($element);
	}
	#
	# Send the the request to the filer
	#
	$root = ati_ontapi_send_to_filer($context, $xml);
	#
	# Get and validate the status and reason attributes of the results tag
	#
	my @nodes = $root->findnodes('/netapp:netapp/netapp:results[@status]');
	if (@nodes != 1) {
		die "No results tag in the XML response from ONTAPI";
	}
	my $status = $nodes[0]->getAttribute('status');
	my $reason = $nodes[0]->getAttribute('reason');
	#
	# Validate the status attribute
	#
	if ($status eq ATI_ONTAPI_STATUS_PASSED) {
		$status = ATI_SUCCESS;
	} elsif ($status eq ATI_ONTAPI_STATUS_FAILED) {
		$status = ATI_FAILURE;
	} else {
		die "Invalid status attribute in the XML response from ONTAPI: \"$status\"";
	}
	#
	# Get the CLI return value from the XML response
	#
	@nodes = $root->findnodes("/netapp:netapp/netapp:results/netapp:cli-result-value");
	if (@nodes != 1) {
		die "No cli-output results tag in the XML response from ONTAPI";
	}
	my $errno = $nodes[0]->textContent();
	#
	# Get the CLI output string from the XML response
	#
	@nodes = $root->findnodes("/netapp:netapp/netapp:results/netapp:cli-output");
	if (@nodes != 1) {
		die "No cli-output results tag in the XML response from ONTAPI";
	}
	my $output = $nodes[0]->textContent();
	return ($status, $output, $errno);
}

#--------------------------------------------------------------------------
# PUBLIC: @out = ati_ontapi_execute_api(<named_args>)
#
# Convenience function to execute an ONTAPI call as well as validate the
# supplied arguments against as list of allowed options.
#
# Arguments:
#
#   hashref  \%context   = ATI connection context.
#   string    $api_call  = API call name. e.g. 'volume-list-info'
#   hashref  \%api_args  = Hash of allowed arguments. e.g. {'volume' => undef}
#   arrayref \@call_args = List of arguments for call. e.g. ['volume', 'vol0']
#
# Returns:
#
#   See ati_ontapi_execute()
#
# Exceptions:
#
#   Die if invalid arguments are encountered.
#
sub ati_ontapi_execute_api_NEW
{	
	#
	# Specify all the named arguments and their defaults
	#
	my %args = (context   => undef,
		    api       => undef,
		    api_args  => undef,
		    call_args => undef);
	#
	# Apply the arguments over the top of the default values.
	# Bitch if we run into any arguments we do not recognize.
	#
	while (@_ > 0) {
		my $name = shift;
		my $value = shift;
		exists($args{$name}) || die "Unknown argument \"$name\" = \"$value\"";
		$args{$name} = $value;
	}
	#
	# Verify that the required arguments are defined
	#
	defined($args{context}) || die "Required argument \"context\" is undefined";
	my $context = $args{context};
	defined($args{api}) || die "Required argument \"api\" is undefined";
	my $api = $args{api};
	defined($args{api_args}) || die "Required argument \"api_args\" is undefined";
	my %api_args = %{$args{api_args}};
	defined($args{call_args}) || die "Required argument \"call_args\" is undefined";
	my @call_args = @{$args{call_args}};
	#
	# Process all the arguments to construct the API request
	#
	my @request = ($api);
	my @params = ();
	my $index;
	for ($index = 0 ; $index < @call_args ; $index += 2) {
		my $key = $call_args[$index];
		my $value = $call_args[$index + 1];
		exists($api_args{$key}) ||
		    die "Bad argument $key=$value api=$api";
		#
		# If a function is provided for processing
		# this argument then call the function.
		#
		if (defined($api_args{$key})) {
			$value = $api_args{$key}->($context, $value);
		}
		push (@params, $key, $value);
	}
	push (@request, \@params);
	#
	# Execute the ONTAPI request
	#
	return ati_ontapi_execute(context => $context,
				  request => \@request);
}

#--------------------------------------------------------------------------
# PUBLIC: @out = ati_ontapi_build_argument(<named_args>)
#
# Convenience function to build a complex API argument that possesses
# sub-arguments. Some ONTAPI API calls require such arguments and can
# be a headache to process inside a regular argument callback function.
#
# This function provides the same level of validation of sub-arguments
# as ati_ontapi_execute_api() does for regular API arguments. You provide
# a hash of allowed sub-arguments and the list of sub-arguments actually
# provided for the argument and it does all the work and returns a simple
# array reference for use by the caller.
#
# This function also allows for callback functions on sub-arguments which
# makes it possible to process complex arguments with arbitrary nesting.
#
# Arguments:
#
#   hashref  \%context  = ATI connection context.
#   hashref  \%sub_args = Hash of allowed sub-arguments.
#   arrayref \@args     = List of provided sub-arguments for call.
#
# Returns:
#
#   Array reference to fully constructed argument.
#
# Exceptions:
#
#   Die if invalid sub-arguments are encountered.
#
sub ati_ontapi_build_argument
{	
	#
	# Specify all the named arguments and their defaults
	#
	my %args = (context   => undef,
		    sub_args  => undef,
		    call_args => undef);
	#
	# Apply the arguments over the top of the default values.
	# Bitch if we run into any arguments we do not recognize.
	#
	while (@_ > 0) {
		my $name = shift;
		my $value = shift;
		exists($args{$name}) || die "Unknown argument \"$name\" = \"$value\"";
		$args{$name} = $value;
	}
	#
	# Verify that the required arguments are defined
	#
	defined($args{context}) || die "Required argument \"context\" is undefined";
	my $context = $args{context};
	defined($args{sub_args}) || die "Required argument \"sub_args\" is undefined";
	my %sub_args = %{$args{sub_args}};
	defined($args{call_args}) || die "Required argument \"call_args\" is undefined";
	my @call_args = @{$args{call_args}};
	#
	# Process all the sub-arguments to construct the argument
	#
	my @params = ();
	my $index;
	for ($index = 0 ; $index < @call_args ; $index += 2) {
		my $key = $call_args[$index];
		my $value = $call_args[$index + 1];
		exists($sub_args{$key}) || die "Bad sub-argument $key = $value";
		#
		# If a function is provided for processing this
		# sub-argument then call the function.
		#
		if (defined($sub_args{$key})) {
			$value = $sub_args{$key}->($context, $value);
		}
		push (@params, $key, $value);
	}
	#
	# Return a reference to the constructed argument list
	#
	return \@params;
}

#--------------------------------------------------------------------------
# PRIVATE: @out = build_type_code(\%context, $name)
#
# Function to generate and compile an type processing subroutine for the
# named type. Can take the type as either a single instance (e.g. "type")
# or as an array of a specific type (e.g. "type[]"). The single and array
# are treated as independent types and generate very different code.
#
# The compiled code reference is inserted into the context type hash table.
#
# This function is driven by build_api_code() to build type processing
# subroutines on demand when APIs subroutines that reference them are
# generated.
#
# Arguments:
#
#   hashref  \%context = ATI connection context.
#   string    $name    = Name of the type (e.g. "volume-info[]")
#   integer   $compile = 0 to skip compilation of generated code (return undef).
#                        1 to compile the generated code and return a coderef.
#
# Returns:
#
#   Code reference to anonymous compiled type processing subroutine.
#
# Exceptions:
#
#   Dies with an internal error if the generated code does not compile.
#
sub build_type_code
{
	my $context   = $_[0];
	my $name      = $_[1];
	my $compile   = $_[2];
	my $type_hash = $context->{'type_hash'};
	my $result    = undef;
	#
	# Verify that the named type exists
	#
	exists($type_hash->{$name}) || die "API type is not defined \[$name\]";
	#
	# Get the element list for this type
	#
	my @elements  = @{$type_hash->{$name}->[0]};
	#
	# Nothing to do if we have already generated code for this type
	#
	return if defined($type_hash->{$name}->[1]);
	#
	# The code generated for array types is different than single instances
	#
	if ($name =~ /\[\]$/) {
		#
		# Strip off the array '[]' trailer
		#
		my $stripped_name = substr($name, 0, -2);
		#
		# Generate array code
		#
		my $code = "sub {\n";
		$code .= "\tmy \$context = \$_\[0\];\n";
		$code .= "\tmy \@out = ();\n";
		$code .= "\tmy \@args = \@\{\$_\[1\]\};\n";
		$code .= "\tforeach my \$arg (\@args) {\n";
		#
		# Skip the stupid typedefs with a single unnamed element
		#
		if ((@elements == 1) && ($elements[0]->{'name'} eq '')) {
			$code .= "\t\tpush (\@out, \'$stripped_name\', \$arg);\n";
			$code .= "\t}\n";
		} else {
			$code .= "\t\tpush(\@out, \'$stripped_name\',\n";
			$code .= "\t\t\tati_ontapi_build_argument(context   => \$context,\n";
			$code .= "\t\t\t                          sub_args  => {";
			#
			# Construct the argument list for this API call
			#
			for (my $i = 0 ; $i < @elements ; ++$i) {
				my $element = $elements[$i];
				#
				# Add comma to end of the last line generated
				#
				$code .= ($i == 0) ? "\n" : ",\n";
				#
				# Get the name and type of this element
				#
				my $name = $element->{'name'};
				my $type = $element->{'type'};
				#
				# Generate argument line for ati_ontapi_build_argument
				#
				$code .= "\t\t\t                              \'$name\' => ";
				if (exists($base_types{$type})) {
					$code .= "undef";
				} else {
					build_type_code($context, $type, $compile);
					$code .= "\$context->{\'type_hash\'}->{\'$type\'}->\[1\]";
				}
			}
			$code .= "},\n";
			$code .= "\t\t\t                          call_args => \$arg));\n";
			$code .= "\t}\n";
		}
		#
		# Generate the common subroutine trailer code
		#
		$code .= "\treturn \\\@out;\n";
		$code .= "}\n";
		$code .= "\n";
		#
		# Print out the actual code to compile if debug is set to verbose
		#
		if ($context->{debug} >= ATI_ONTAPI_DEBUG_DEBUG) {
			print "DBG(VERBOSE): Generated this type subroutine for \[$name\]:\n", $code;
		}
		#
		# Compile the generated subroutine
		#
		if ($compile) {
			$result = eval $code;
			$@ && die "Eval failed for type subroutine \[$name\]: $@";
		}

	} else {
		#
		# Skip the stupid typedefs with a single unnamed element
		#
		if ((@elements != 1) || ($elements[0]->{'name'} ne '')) {
			#
			# Generate the common subroutine header code
			#
			my $code = "sub {\n";
			$code .= "\tmy \$context = \$_\[0\];\n";
			$code .= "\treturn \[\'$name\',\n";
			$code .= "\t        ati_ontapi_build_argument(context   => \$context,\n";
			$code .= "\t                                  sub_args  => {";
			#
			# Construct the argument list for this API call
			#
			for (my $i = 0 ; $i < @elements ; ++$i) {
				my $element = $elements[$i];
				#
				# Add comma to end of the last line generated
				#
				$code .= ($i == 0) ? "\n" : ",\n";
				#
				# Get the name and type of this element
				#
				my $name = $element->{'name'};
				my $type = $element->{'type'};
				#
				# Generate argument line for ati_ontapi_build_argument
				#
				$code .= "\t                                     \'$name\' => ";
				if (exists($base_types{$type})) {
					$code .= "undef";
				} else {
					build_type_code($context, $type, $compile);
					$code .= "\$context->{\'type_hash\'}->{\'$type\'}->\[1\]";
				}
			}
			$code .= "},\n";
			$code .= "\t                                  call_args => \$_[1])];\n";
			$code .= "}\n";
			#
			# Print out the actual code to compile if debug is set to verbose
			#
			if ($context->{debug} >= ATI_ONTAPI_DEBUG_DEBUG) {
				print "DBG(VERBOSE): Generated this type subroutine for \[$name\]:\n", $code;
			}
			#
			# Compile the generated subroutine
			#
			if ($compile) {
				$result = eval $code;
				$@ && die "Eval failed for type subroutine \[$name\]: $@";
			}
		}
	}
	#
	# Store a reference to the compiled code in the context type hash table
	#
	$type_hash->{$name}->[1] = $result;
   	return $result;
}

#--------------------------------------------------------------------------
# PRIVATE: @out = build_api_code(\%context, $name, $compile)
#
# Function to generate and compile an API processing subroutine for the
# named API. Expects to be handed a reference to the elements-info array
# for the API that is returned by the system-api-get-elements API. Any
# needed type processing subroutines will be generated on the fly.
#
# The compiled code reference is inserted into the context API hash table.
#
# Arguments:
#
#   hashref  \%context = ATI connection context.
#   string    $name    = Name of the API (e.g. "volume-list-info")
#   integer   $compile = 0 to skip compilation of generated code (return undef).
#                        1 to compile the generated code and return a coderef.
#
# Returns:
#
#   Code reference to anonymous compiled API processing subroutine.
#
# Exceptions:
#
#   Dies with an internal error if the generated code does not compile.
#
sub build_api_code
{
	my $context  = $_[0];
	my $name     = $_[1];
	my $compile  = $_[2];
	my $api_hash = $context->{api_hash};
	my $result   = undef;
	#
	# Verify that the named API exists
	#
	exists($api_hash->{$name}) || die "API is not defined \[$name\]";
	#
	# Go fetch the API element list if we don't have it already
	#
	if (!defined($api_hash->{$name}->[0])) {
		#
		# Query the API element list via the reflective interface
		#
		my @out = ati_ontapi_execute(context => $context,
					     request => ['system-api-get-elements',
							 ['api-list',
							  ['api-list-info', $name]]]);
		($out[0] == ATI_SUCCESS) ||
		  die "Failed to query API elements \[$name\]: $out[2]: $out[3]\n";
		#
		# Get the API element list or use emtpy list if none is returned
		#
		$api_hash->{$name}->[0] = $out[1]->{'api-entries'}->[0]->{'api-elements'} || [];
	}
	#
	# Get the list of elements for this API
	#
	my $elements = $api_hash->{$name}->[0];
	#
	# Generate the common subroutine header code
	#
	my $code = "sub {\n";
	$code .= "\tmy \$context = \$_\[0\];\n";
	$code .= "\treturn ati_ontapi_execute_api_NEW(context   => \$context,\n";
	$code .= "\t                              api       => \'$name\',\n";
	$code .= "\t                              api_args  => {";
	#
	# Construct the argument list for this API call
	#
	for (my $i = 0 ; $i < @{$elements} ; ++$i) {
		my $element = $elements->[$i];
		#
		# Only process the input arguments
		#
		if (!exists($element->{'is-output'}) || !$element->{'is-output'}) {
			#
			# Add comma to end of the last line generated
			#
			$code .= ($i == 0) ? "\n" : ",\n";
			#
			# Get the name and type of this element
			#
			my $name = $element->{'name'};
			my $type = $element->{'type'};
			#
			# Generate argument line for ati_ontapi_execute_api
			#
			$code .= "\t                                  \'$name\' => ";
			if (exists($base_types{$type})) {
				$code .= "undef";
			} else {
				build_type_code($context, $type, $compile);
				$code .= "\$context->{\'type_hash\'}->{\'$type\'}->\[1\]";
			}
		}
	}
	#
	# Generate the common subroutine trailer code
	#
	$code .= "},\n";
	$code .= "\t                              call_args => \$_[1]);\n";
	$code .= "}\n";
	#
	# Print out the actual code to compile if debug is set to verbose
	#
	if ($context->{debug} >= ATI_ONTAPI_DEBUG_DEBUG) {
		print "DBG(VERBOSE): Generated this API subroutine for \[$name\]:\n", $code;
	}
	#
	# Compile the generated subroutine
	#
	if ($compile) {
		$result = eval $code;
		$@ && die "Eval failed for API subroutine \[$name\]: $@";
	}
	#
	# Store a reference to the compiled code in the context API hash table
	#
	$api_hash->{$name}->[1] = $result;
	return $result;
}

#--------------------------------------------------------------------------
# PUBLIC: @out = ati_ontapi_call_api(<named_args>)
#
# Function to execute an ONTAPI call as well as validate the arguments.
#
# Arguments:
#
#   hashref  \%context = ATI connection context.
#   string    $api     = API name. e.g. 'volume-list-info'
#   arrayref \@args    = List of arguments for call. e.g. ['volume', 'vol0']
#
# Returns:
#
#   See ati_ontapi_execute()
#
# Exceptions:
#
#   Die if invalid arguments are encountered.
#
# Example:
#
#	@out = ati_ontapi_call_api(context => \%context,
#			           api     => "volume-list-info",
#			           args    => ['volume' => 'vol1']);
#
sub ati_ontapi_call_api
{
	#
	# Specify all the named arguments and their defaults
	#
	my %args = (context => undef,
		    api     => undef,
		    args    => []);
	#
	# Apply the arguments over the top of the default values.
	# Bitch if we run into any arguments we do not recognize.
	#
	while (@_ > 0) {
		my $name = shift;
		my $value = shift;
		exists($args{$name}) || die "Unknown argument \"$name\" = \"$value\"";
		$args{$name} = $value;
	}
	#
	# Verify that the required arguments are defined
	#
	defined($args{context}) || die "Required argument \"context\" is undefined";
	my $context = $args{context};
	defined($args{api}) || die "Required argument \"api\" is undefined";
	my $api = $args{api};
	defined($args{args}) || die "Required argument \"args\" is undefined";
	my $api_args = $args{args};
	#
	# Check the specified API exists
	#
	my $api_hash = $context->{api_hash};
	exists($api_hash->{$api}) ||
	    die "Attempt to call a non-existant API \[$api\]";
	#
	# Generate and compile the API execution subroutine code if necessary
	#
	defined($api_hash->{$api}->[1]) || build_api_code($context, $api, 1);
	#
	# Execute the API subroutine
	#
	return $api_hash->{$api}->[1]->($context, $api_args);
}

#
# End of ati_ontapi package
#
1;
