#--------------------------------------------------------------------------
# $Id: //depot/prod/test/Rsierranevada/ATI/ati_util.pm#1 $
#
# Copyright (c) 1992-2005 Network Appliance, Inc.
# All rights reserved.
#--------------------------------------------------------------------------
#
# PACKAGE: ati_util.pm
#
# AUTHOR: Will McGovern (will@netapp.com)
#
# This package implements utility functions to augment the ATI modules.
#
# Ideally, some of these functions represent functionality that should be
# provided by ONTAPI directly.
#
# TODO:
#
# o Module self-test function to verify correct operation.
# o More functions... duh.
#
#--------------------------------------------------------------------------
#
# Exported constants:
#
# ATI_UTIL_AGGR_STATE_CREATING     : Aggregate "creating" state.
# ATI_UTIL_AGGR_STATE_FAILED       : Aggregate "failed" state.
# ATI_UTIL_AGGR_STATE_OFFLINE      : Aggregate "offline" state.
# ATI_UTIL_AGGR_STATE_ONLINE       : Aggregate "online" state.
# ATI_UTIL_AGGR_STATE_PARTIAL      : Aggregate "partial" state.
# ATI_UTIL_AGGR_STATE_RESTRICTED   : Aggregate "restricted" state.
# ATI_UTIL_AGGR_STATE_UNKNOWN      : Aggregate "unknown" state.
#
# ATI_UTIL_VOLUME_STATE_CREATING   : Volume "creating" state.
# ATI_UTIL_VOLUME_STATE_FAILED     : Volume "failed" state.
# ATI_UTIL_VOLUME_STATE_OFFLINE    : Volume "offline" state.
# ATI_UTIL_VOLUME_STATE_ONLINE     : Volume "online" state.
# ATI_UTIL_VOLUME_STATE_PARTIAL    : Volume "partial" state.
# ATI_UTIL_VOLUME_STATE_RESTRICTED : Volume "restricted" state.
# ATI_UTIL_VOLUME_STATE_UNKNOWN    : Volume "unknown" state.
#
# Exported functions:
#
# ati_util_aggr_wait_state   : Wait for aggregate to reach a state.
# ati_util_volume_wait_state : Wait for volume to reach a state.
# ati_util_format_dsid 	     : Format a given number into dsid format.
# ati_util_format_msid 	     : Format a given number into msid format.
#
#--------------------------------------------------------------------------
#
# CODING CONVENTIONS FOR THIS MODULE:
#
#   o All exported functions must take named arguments.
#
#   o No obfuscated or obscure perl code. Use perl's strengths judiciously.
#
#   o Maintain a clear and consistent coding style between functions.
#
#   o Use comments liberally.
#
#--------------------------------------------------------------------------
#
# NOTE TO THOSE WORKING ON THIS MODULE CODE:
#
# I am not a perl guru so if you see anything broken feel free to fix or
# improve the code. If, however, these changes mess up the readability or
# maintainability (by those of us non-perl gurus) with obscure code,
# ego-boosting fancy perl tricks, or unnecessary "optimizations" then I
# will find you and you will suffer greatly.
#
#--------------------------------------------------------------------------
#
package ATI::ati_util;
#
# Always require hard-assed checking
#
use strict;
use warnings;
#
# ATI base packages
#
use ATI::ati_ontapi qw(ATI_FAILURE
		       ATI_SUCCESS
		       ati_ontapi_call_api);
#
# Export symbols for external use
#
use Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(ATI_UTIL_AGGR_STATE_CREATING
		    ATI_UTIL_AGGR_STATE_FAILED
		    ATI_UTIL_AGGR_STATE_OFFLINE
		    ATI_UTIL_AGGR_STATE_ONLINE
		    ATI_UTIL_AGGR_STATE_PARTIAL
		    ATI_UTIL_AGGR_STATE_RESTRICTED
		    ATI_UTIL_AGGR_STATE_UNKNOWN
		    ATI_UTIL_VOLUME_STATE_CREATING
		    ATI_UTIL_VOLUME_STATE_FAILED
		    ATI_UTIL_VOLUME_STATE_OFFLINE
		    ATI_UTIL_VOLUME_STATE_ONLINE
		    ATI_UTIL_VOLUME_STATE_PARTIAL
		    ATI_UTIL_VOLUME_STATE_RESTRICTED
		    ATI_UTIL_VOLUME_STATE_UNKNOWN
		    ati_util_aggr_wait_state
		    ati_util_volume_wait_state
		    ati_util_format_msid
		    ati_util_format_dsid
		   );
#
# PUBLIC: Aggregate states
#
use constant ATI_UTIL_AGGR_STATE_CREATING   => 'creating';
use constant ATI_UTIL_AGGR_STATE_FAILED     => 'failed';
use constant ATI_UTIL_AGGR_STATE_OFFLINE    => 'offline';
use constant ATI_UTIL_AGGR_STATE_ONLINE     => 'online';
use constant ATI_UTIL_AGGR_STATE_PARTIAL    => 'partial';
use constant ATI_UTIL_AGGR_STATE_RESTRICTED => 'restricted';
use constant ATI_UTIL_AGGR_STATE_UNKNOWN    => 'unknown';
#
# PUBLIC: Volume states
#
use constant ATI_UTIL_VOLUME_STATE_CREATING   => 'creating';
use constant ATI_UTIL_VOLUME_STATE_FAILED     => 'failed';
use constant ATI_UTIL_VOLUME_STATE_OFFLINE    => 'offline';
use constant ATI_UTIL_VOLUME_STATE_ONLINE     => 'online';
use constant ATI_UTIL_VOLUME_STATE_PARTIAL    => 'partial';
use constant ATI_UTIL_VOLUME_STATE_RESTRICTED => 'restricted';
use constant ATI_UTIL_VOLUME_STATE_UNKNOWN    => 'unknown';
#
# PRIVATE: Default aggr/volume state wait time-out in seconds
#
use constant ATI_UTIL_WAIT_STATE_TIMEOUT => 3600;  # 1 hour
#
# PRIVATE: ETIMEDOUT error number
#
use constant ETIMEDOUT => 60;

#--------------------------------------------------------------------------
# PUBLIC: @out = ati_util_aggr_wait_state(<named_args>)
#
# Wait for an aggregate to reach the specified state by polling every five
# seconds until either the state is reached, an error is encountered polling
# the filer or the specified timeout is reached.
#
# Arguments in addition to ATI base ati_* arguments:
#
#   REQ  $string  aggregate     = Name of aggregate.
#   OPT  $string  state         = Aggregate state for which to wait.
#                                 default = ATI_AGGR_STATE_ONLINE
#   OPT  $integer timeout       = Time-out duration in seconds.
#                                 default = ATI_UTIL_WAIT_AGGR_STATE_TIMEOUT
#
# Returns standard ATI return values (see ati_ontapi_execute).
#
# Example: Wait for aggregate "plop" to come to an online state with a
#          300 second timeout.
#
# @out = ati_util_aggr_wait_state(context   => $context,
#                                 aggregate => 'plop',
#                                 state     => ATI_UTIL_AGGR_STATE_ONLINE,
#                                 timeout   => 300);
#
sub ati_util_aggr_wait_state
{	
	#
	# Specify all the named arguments and their defaults
	#
	my %args = (context   => undef,
		    aggregate => undef,
		    state     => ATI_UTIL_AGGR_STATE_ONLINE,
		    timeout   => ATI_UTIL_WAIT_STATE_TIMEOUT);
	#
	# Apply the arguments over the top of the default values.
	# Bitch if we run into any arguments we do not recognize.
	#
	while (@_ > 0) {
		my $name = shift;
		my $value = shift;
		exists($args{$name}) || die "Unknown argument \"$name\" = \"$value\"";
		$args{$name} = $value;
	}
	#
	# Verify that the required arguments are defined
	#
	defined($args{context}) || die "Required argument \"context\" is undefined";
	my $context = $args{context};
	defined($args{aggregate}) || die "Required argument \"aggregate\" is undefined";
	my $aggregate = $args{aggregate};
	#
	# Cache arguments from hash for readability
	#
	my $state     = $args{state};
	my $timeout   = $args{timeout};
	#
	# Compute the time at which we should give up waiting
	#
	my $abort_time = time() + $timeout;
	#
	# Keep polling the aggregate state until we have a match
	#
	my @out;
	my $timed_out = 1;
	while (time() < $abort_time) {
		#
		# Query the state of the aggregate via ONTAPI
		#
		@out = ati_ontapi_call_api(context => $context,
					   api     => 'aggr-list-info',
					   args    => ['aggregate' => $aggregate]);
		#
		# Terminate if request failed or we reached the desired state
		#
		if (($out[0] == ATI_FAILURE) ||
		    ($out[1]->{'aggregates'}[0]{'state'} eq $state)) {
			$timed_out = 0;
			last;
		}
		#
		# Sleep for a little while before next poll
		#
		sleep 5;
	}
	#
	# If we timed out then fake up our own error response
	#
	if ($timed_out == 1) {
		@out = (ATI_FAILURE, undef, ETIMEDOUT,
			"Timed out waiting for state $state on aggregate $aggregate");
	}
	#
	# Return the real success or faked up error response to the caller
	#
	return @out;
}

#--------------------------------------------------------------------------
# PUBLIC: @out = ati_util_volume_wait_state(<named_args>)
#
# Wait for a volume to reach the specified state by polling every five
# seconds until either the state is reached, an error is encountered in
# polling the filer, or the specified timeout is reached.
#
# Arguments in addition to ATI base ati_* arguments:
#
#   REQ \%hash    context = ATI connection context.
#   OPT  $string   volume = Name of volume.
#   OPT  $string       id = Id of volume.
#   OPT  $integer   state = Volume state for which to wait.
#                           default = ATI_UTIL_VOLUME_STATE_ONLINE
#   OPT  $integer timeout = Time-out duration in seconds.
#                           default = 90 minutes
#
# Returns standard ATI return values (see ati_ontapi_execute).
# One of volume or id argument is required, but not both.
# volume should be used for Ontap-Classic filers.
# id should be used for Ontap-NG filers.
#
# Example 1: On filer shingles.lab.netapp.com, wait for volume "eatthis"
#            to come to an online state with a 300 second timeout.
#
# @out = ati_util_volume_wait_state(context => $context,
#                                   volume  => 'plop',
#                                   state   => ATI_UTIL_VOLUME_STATE_ONLINE,
#                                   timeout => 300);
#
# Example 2: On filer shingles.lab.netapp.com, wait for id "0x00000400"
#            to come to an online state with a 300 second timeout.
#
# @out = ati_util_volume_wait_state(context => $context,
#                                   id      => '0x00000400',
#                                   state   => ATI_UTIL_VOLUME_STATE_ONLINE,
#                                   timeout => 300);
#
sub ati_util_volume_wait_state
{	
	#
	# Specify all the named arguments and their defaults
	#
	my %args = (context => undef,
		    volume  => undef,
		    id	    => undef,
		    state   => ATI_UTIL_VOLUME_STATE_ONLINE,
		    timeout => ATI_UTIL_WAIT_STATE_TIMEOUT);
	#
	# Apply the arguments over the top of the default values.
	# Bitch if we run into any arguments we do not recognize.
	#
	while (@_ > 0) {
		my $name = shift;
		my $value = shift;
		exists($args{$name}) || die "Unknown argument \"$name\" = \"$value\"";
		$args{$name} = $value;
	}
	#
	# Verify that the required arguments are defined
	#
	defined($args{context}) || die "Required argument \"context\" is undefined";
	my $context = $args{context};
	if ((!defined($args{volume})) && (!defined($args{id}))) {
		die "One of arguments \"volume\" or \"id\" is required";
	}
	my $volume = $args{volume};
	my $id = $args{id};
	#
	# Cache arguments from hash for readability
	#
	my $state   = $args{state};
	my $timeout = $args{timeout};
	#
	# Compute the time at which we should give up waiting
	#
	my $abort_time = time() + $timeout;
	#
	# Keep polling the volume state until we have a match
	#
	my @out;
	my $timed_out = 1;
	my $outstate;
	while (time() < $abort_time) {
		#
		# Query the state of the volume via ONTAPI
		#
		if (defined($id)) {
			@out = ati_ontapi_call_api(context => $context,
						   api     => 'd-volume-list-info',
						   args    => ['id'            => $id,
							       'desired-attrs' => 'state']);
			if ($out[0] != ATI_FAILURE) {
				$outstate = $out[1]->{'volume-attrs'}[0]{'state'};
			}
		} else {
			@out = ati_ontapi_call_api(context => $context,
						   api     => 'volume-list-info',
						   args    => ['volume' => $volume]);
			if ($out[0] != ATI_FAILURE) {
				$outstate = $out[1]->{'volumes'}[0]{'state'};
			}
		}
		#
		# Terminate if request failed or we reached the
		# desired state
		#
		if (($out[0] == ATI_FAILURE) || ($outstate eq $state)) {
			$timed_out = 0;
			last;
		}
		#
		# Sleep for a little while before next poll
		#
		sleep 5;
	}
	#
	# If we timed out then fake up our own error response
	#
	if ($timed_out == 1) {
		@out = (ATI_FAILURE, undef, ETIMEDOUT,
			"Timed out waiting for state $state on volume $volume");
	}
	#
	# Return the real success or faked up error response to the caller
	#
	return @out;
}

#--------------------------------------------------------------------------
# PUBLIC: $out = ati_util_format_dsid(<named_args>)
#
# Returns a formatted dsid string given a number in hex or decimal
#
#   REQ  $string    id = dsid of volume
#
# Example 1: $voldsid = ati_util_format_dsid("0x400") returns 0x00000400
# Example 2: $voldsid = ati_util_format_dsid("1024") returns 0x00000400
#
sub ati_util_format_dsid($)
{
	my $conv = $_[0];

	if (!defined($conv)) {
		return undef;
	} elsif ($conv !~ /^0x/) {
		#
		# If not in hex, return in hex.
		#
		return sprintf("0x%08x", int($conv));
	} else {
		#
		# If in hex, return with filled in zeroes
		#
		return sprintf("0x%08x", int(hex($conv)));
	}
}

#--------------------------------------------------------------------------
# PUBLIC: $out = ati_util_format_msid(<named_args>)
#
# Returns a formatted msid string given a number in hex or decimal
#
#   REQ  $string    id = msid of volume
#
# Example 1: $volmsid = ati_util_format_msid("0x400") returns 0x00000400
# Example 2: $volmsid = ati_util_format_msid("1024") returns 0x00000400
#
sub ati_util_format_msid($)
{
	#
	# For now, Msid is the same as Dsid.
	#
	return ati_util_format_dsid($_[0]);
}

#
# End of ati_util package
#
1;
