#!/usr/local/bin/perl
#--------------------------------------------------------------------------
# $Id: //depot/prod/test/Rsierranevada/ATI/aticli.pl#1 $
#
# Copyright (c) 1992-2004 Network Appliance, Inc.
# All rights reserved.
#--------------------------------------------------------------------------
#
# ONTAP administrative CLI tunnelling over ONTAPI (via HTTP/HTTPS).
#
# AUTHOR: Will McGovern (will@netapp.com)
#
# This code demonstrates the simplicity of using ATI and more specifically
# the ati_ontapi_execute_cli() subroutine to tunnel CLI commands over HTTP
# or HTTPS to a filer. Actually quite a useful tool.
#
#--------------------------------------------------------------------------
# NOTES:
#
# Running: aticli.pl [-s] [-u <user>] [-p <password>] <filer>
#
# Non-terminating commands such as "sysstat" should NOT be executed because
# the ONTAP server side goes off into the weeds waiting for the command to
# terminate and will refuse to accept any subsequent requests.
#
#--------------------------------------------------------------------------
# TO DO:
#
# o Put in guards for non-terminating CLI commands that hose the server.
#
#--------------------------------------------------------------------------
#
# Always require hard-assed checking
#
use strict;
use warnings;
#
# Add our parent directory to the module search path
#
use FindBin;
use lib "$FindBin::Bin/..";
#
# System packages
#
use Getopt::Std;
use Term::ReadLine;
#
# ATI base modules
#
use ATI::ati_ontapi qw(ATI_SUCCESS
		       ATI_ONTAPI_TRANSPORT_HTTP
		       ATI_ONTAPI_TRANSPORT_HTTPS
		       ATI_ONTAPI_USER_DEFAULT
		       ATI_ONTAPI_PASSWORD_DEFAULT
		       ATI_ONTAPI_PRIVILEGE_ADMIN
		       ati_ontapi_create_context
		       ati_ontapi_call_api
		       ati_ontapi_call_cli);

#--------------------------------------------------------------------------
#
# Start of main code
#
#--------------------------------------------------------------------------
#
# Parse the options and arguments
#
my %opts;
my $gotopts   = getopts('su:p:', \%opts);
my $transport = defined($opts{'s'}) ? ATI_ONTAPI_TRANSPORT_HTTPS : ATI_ONTAPI_TRANSPORT_HTTP;
my $user      = $opts{'u'} || ATI_ONTAPI_USER_DEFAULT;
my $password  = $opts{'p'} || ATI_ONTAPI_PASSWORD_DEFAULT;
#
# Sanity check the arguments
#
($gotopts && (@ARGV == 1)) ||
  die "Usage: aticli.pl [-s] [-u <user] [-p <password>] <filer>\n" .
      "       -s            =  use secure HTTPS connection.\n" .
      "       -u <user>     =  login to filer using this user name.\n" .
      "       -p <password> =  login password.\n" .
      "       <filer>       =  filer to connect.\n";
#
# Create an ATI context for the filer connection
#
my $context = ati_ontapi_create_context(filer     => $ARGV[0],
					transport => $transport,
				        user      => $user,
				        password  => $password);
#
# Get the system information for this filer
#
print "\nConnecting...\n\n";
my @out = ati_ontapi_call_api(context => $context, api => 'system-get-info');
($out[0] == ATI_SUCCESS) || die "Error $out[2]: $out[3]\n";
my $sysname  = $out[1]->{'system-info'}->{'system-name'};
my $sysmodel = $out[1]->{'system-info'}->{'system-model'};
#
# Get the Data ONTAP version for the filer
#
@out = ati_ontapi_call_api(context => $context, api => 'system-get-version');
($out[0] == ATI_SUCCESS) || die "Error $out[2]: $out[3]\n";
my $version = $out[1]->{'version'};
$version =~ s/NetApp Release //;
$version =~ s/\/.*\///;
$version =~ s/:.*$//;
#
# Print out the system information
#
print "  System Name: $sysname\n";
print " System Model: $sysmodel\n";
print "ONTAP Version: $version\n\n";
print "*** CAUTION: Do NOT use commands that do not terminate (like sysstat) ***\n\n";
#
# Start tunneling CLI commands
#
my $term = new Term::ReadLine 'ATI Tunneled CLI';
my $priv = ATI_ONTAPI_PRIVILEGE_ADMIN;
while (defined($_ = $term->readline("$sysname($priv)> "))) {
	#
	# Split compound commands and execute serially
	#
	my @commands = split(/;/);
	foreach $_ (@commands) {
		#
		# Ignore blank lines and mirror privilege settings
		#
		/^\s*$/ && next;
		/^priv set (admin|advanced|test|diag)$/ && ($priv = $1);
		#
		# Execute the CLI command on the filer via ONTAPI
		#
		@out = ati_ontapi_call_cli(context   => $context,
					   privilege => $priv,
					   command   => $_);
		if ($out[0] == ATI_SUCCESS) {
			print $out[1];
		} else {
			print "Error $out[2]: $out[3]\n";
		}
	}
}
