#!/usr/local/bin/perl
#--------------------------------------------------------------------------
# $Id: //depot/prod/test/Rsierranevada/ATI/ng_server.pl#1 $
#
# Copyright (c) 1992-2004 Network Appliance, Inc.
# All rights reserved.
#--------------------------------------------------------------------------
#
# PACKAGE: ng_server.pl
#
# AUTHOR: Will McGovern (will@netapp.com)
#
# This script implements a test ONTAPI/NG server for testing against
# the ATI library using the ATI_ONTAPI_TRANSPORT_D transport type. The
# server currently listens on port 12345 and responds to all NG requests
# with a standard ONTAPI error response.
#
# This test server does not do any validation of the ONTAPI requests but
# does verify the ONTAPI/NG headers and content length.
#
# The ONTAPI/NG transport protocol resembles a much cut down HTTP style
# protocol with a protocol ID/version and a set of headers. Currently
# the only supported header is Content-length which gives the length of
# the ONTAPI XML request which follows the blank line after the headers.
#
# Here is a simple template:
#
#    NG/1.0
#    Content-length: <size>
#    <blank line>
#    <ONTAPI XML request>
#
#--------------------------------------------------------------------------
# Running this script:
#
#   perl test_ng_server.pl
#
#--------------------------------------------------------------------------
# TODO:
#
# o Add timeout code to catch data underruns in the client requests.
#
#--------------------------------------------------------------------------
#
# Always require hard-assed checking
#
use strict;
use warnings;
#
# System packages
#
use Getopt::Std;
use IO::Socket;
use Net::SSLeay;
#
# PRIVATE: Data ONTAP/NG server ports
#
use constant ATI_ONTAPI_D_SERVER_PORT  => 2005;
use constant ATI_ONTAPI_DS_SERVER_PORT => 2006;

#--------------------------------------------------------------------------
#
# Start of main driver code for the code generator
#
#--------------------------------------------------------------------------
#
# Parse the options and arguments
#
my %opts;
my $gotopts = getopts('s', \%opts);
my $use_ssl = defined($opts{'s'}) ? 1 : 0;
#
# Sanity check the arguments
#
($gotopts) ||
  die "Usage: ng_server.pl [-s]\n" .
  "       -s = use secure sockets (SSL).\n";
#
# If SSL was specified then initialize SSL and use the right port
#
my $port = ATI_ONTAPI_D_SERVER_PORT;
my $ssl_ctx;
if ($use_ssl) {
	print "Initializing SSL\n";
	$port = ATI_ONTAPI_DS_SERVER_PORT;
	Net::SSLeay::load_error_strings();
	Net::SSLeay::SSLeay_add_ssl_algorithms();
	Net::SSLeay::randomize();
	$ssl_ctx = Net::SSLeay::CTX_new();
	Net::SSLeay::CTX_use_RSAPrivateKey_file($ssl_ctx, 'ng_server_key.pem', &Net::SSLeay::FILETYPE_PEM);
	Net::SSLeay::CTX_use_certificate_file($ssl_ctx, 'ng_server_cert.pem', &Net::SSLeay::FILETYPE_PEM);
}
#
# Open the server socket and start listening for client connections
#
my $server = IO::Socket::INET->new(Proto     => 'tcp',
				   Type      => SOCK_STREAM,
				   LocalPort => $port,
				   Reuse     => 1,
				   Listen    => SOMAXCONN) ||
  die "Failed to open server socket\n";
#
# Loop accepting client connections
#
while (my $client = $server->accept()) {
	$client->autoflush(1);
	#
	# Print out the connection information
	#
	print "Incoming connection from " . inet_ntoa($client->peeraddr()) . "\n";
	#
	# Handle SSL setup if necessary
	#
	my $ssl;
	my $sslsock;
	if ($use_ssl) {
		$ssl = Net::SSLeay::new($ssl_ctx);
		Net::SSLeay::set_fd($ssl, fileno($client));
		Net::SSLeay::accept($ssl) ||
		    die "Failed to accept SSL connection\n";
		print "SSL using cipher " . Net::SSLeay::get_cipher($ssl) . "\n";
	}
	#
	# Parse first line of request: "NG/1.0"
        #
	print "Parsing request headers\n";
	my $request = ($use_ssl) ? Net::SSLeay::ssl_read_CRLF($ssl) : <$client>;
	$request =~ s/\s+$//;
	if ($request !~ /^ONTAPI-D\/1.0$/) {
		die "Invalid request line 1 from ONTAP/NG client: \"$request\"\n";
	}
	#
	# Parse the second header line: "User: <username>\n"
	#
	$request = ($use_ssl) ? Net::SSLeay::ssl_read_CRLF($ssl) : <$client>;
	$request =~ s/\s+$//;
	if ($request !~ /^User: .+$/) {
		die "Invalid request line 2 from ONTAP/NG client: \"$request\"\n";
	}
	#
	# Parse the third header line: "Password: <password>\n"
	#
	$request = ($use_ssl) ? Net::SSLeay::ssl_read_CRLF($ssl) : <$client>;
	$request =~ s/\s+$//;
	if ($request !~ /^Password:*$/) {
		die "Invalid request line 3 from ONTAP/NG client: \"$request\"\n";
	}
	#
	# Parse the fourth header line: "Content-Length: nnnn\n"
	#
	$request = ($use_ssl) ? Net::SSLeay::ssl_read_CRLF($ssl) : <$client>;
	$request =~ s/\s+$//;
	if ($request !~ /^Content-length: ([0-9]+)$/) {
		die "Invalid request line 4 from ONTAP/NG client: \"$request\"\n";
	}
	my $length = $1;
	#
	# Read and dump the blank line separating the headers and the body
	#
	$request = ($use_ssl) ? Net::SSLeay::ssl_read_CRLF($ssl) : <$client>;
	$request =~ s/\s+$//;
	if ($request ne '') {
		die "Invalid request line 5 from ONTAP/NG client: \"$request\"\n";
	}
	#
	# Now read all the bytes from the body
	#
	print "Parsing XML request body\n";
	while ($length > 0) {
		$request = ($use_ssl) ? Net::SSLeay::ssl_read_CRLF($ssl) : <$client>;
		$length -= length($request);
	}
	#
	# Quick sanity check to ensure we did not receive too much data
	#
	($length == 0) || die "Content length mismatch in ONTAP/NG request";
	#
	# Construct an ONTAPI error response for all requests
	#
	my $xml =
          "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" .
	  "<!DOCTYPE netapp SYSTEM \"/na_admin/netapp_filer.dtd\">" .
	  "<netapp xmlns=\"http://www.netapp.com/filer/admin\" version=\"1.1\">" .
	  "<results status=\"failed\" errno=\"666\" reason=\"This is only a test NG server\"/>" .
	  "</netapp>\r\n";
	$length = length($xml);
	my $response = "ONTAPI-D/1.0 200 OK\r\nContent-length: $length\r\n\r\n$xml";
	#
	# Send the error response to the client
	#
	print "Sending XML response\n";
	if ($use_ssl) {
		Net::SSLeay::write($ssl, $response) || die "Failed to send response (SSL)";
	} else {
		print($client $response) || die "Failed to send response";
	}
	#
	# Free the SSL state for this connection
	#
	if ($use_ssl) {
		Net::SSLeay::free($ssl);
	}
	close($client);
}
#
# Clean up and exit
#
if ($use_ssl) {
	Net::SSLeay::CTX_free($ssl_ctx);
}
close($server);

1;
