#  $Id: //depot/prod/test/Rsierranevada/FSR/DR/DRutils.pm#1 $

## @summary Dump/Restore utility library
##
## @description
##	This package provides a common set of library routines
##	useful for the Dump/Restore tests.
##
## @dependencies
##    lib/DataManip
##    lib/QtreeCmds
##    FSR/FSRutils
##
## @bugs
##
## @author marlon@netapp.com

package FSR::DR::DRutils;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(addSlashVol
             TapeRewind
             DumpRestore
             RemoteDump
             RemoteRestore
             DRCompare_tree
	     compareTrees
             getInodeNum
             getInodeNum_local
             linkIt
             linkIt_local
             symlinkIt
             symlinkIt_local
             truncIt
             truncIt_local
             chmodIt
             chmodIt_local
             moveIt
             moveIt_local
             makeDir
             makeDir_local
             deleteIt
             deleteIt_local
             tarIt
             tarIt_local);

use Tharn;
use DataManip;
use QtreeCmds;
use FSR::FSRutils;
use strict;

## @name ::addSlashVol
## @usage addSlashVol($volname)
## @synopsis
##    If necessary, prepend /vol/ to $volname.
##    Returns the string starting with /vol/
##
sub addSlashVol ($) {
    my ($volname) = @_;
    if ($volname =~ /^\/vol\//) {
        return $volname;
    }
    else {
        return "/vol/" . $volname;
    }
}

## @name ::TapeRewind
## @usage TapeRewind($cfiler, $dump_device)
## @synopsis
##	If $dump_device is a tape device then rewind it.
##	Many Dump/Restore tests can use either a real tape device or a
##	a file to hold the dump image. Of course a file does not need
##	to be rewound. Assume that anything starting "/vol" is not a tape device.
sub TapeRewind( $$ ) {
    my ($cfiler, $dump_dev) = @_;
    my $TIMEOUT = 3600;

    if ($dump_dev =~ /^\/vol/) {
        return;
    }

    say($cfiler, "mt -f $dump_dev rewind", -timeout => $TIMEOUT);
}


## @name ::DumpRestore
## @usage DumpRestore($cfiler, $dump_device, $dump_level, $dump_pt, $restore_pt)
## @synopsis
##	Rewind tape device ($dump_device), do a level $dump_level dump
##	of $dump_pt, rewind tape device, do a restore of dump image to
##	$restore_pt
##
sub DumpRestore( $$$$$ ) {

    my ($cfiler, $dump_dev, $dump_level, $dump_pt, $restore_pt) = @_;
    my $TIMEOUT = 3600;
    my $sub_obj;

    TapeRewind($cfiler, $dump_dev);
    $sub_obj = subtest("dumprestore/dump_target.thpl",
            "DUMP_LEVEL=$dump_level",
            "DUMP_OPTIONS=u",
            "DUMP_FILE=$dump_dev",
            "DUMP_TARGET=$dump_pt",
            '-runid', "Dump-$dump_level",
            '-onexit', sub {},
            '-iffatal', '-fatal', "Dump of $dump_pt failed");
    Subtest::wait_finish(subtest => [$sub_obj]);
    subtestResults($sub_obj);
    die $@ if $@;

    TapeRewind($cfiler, $dump_dev);
    $sub_obj = subtest("dumprestore/restore_target.thpl", "RESTORE_FILE=$dump_dev",
            "DUMP_FILE=$dump_dev",
            "RESTORE_TARGET=$restore_pt",
            '-runid', "Restore-$dump_level",
            '-onexit', sub {},
            '-iffatal', '-fatal',"Restore of $restore_pt failed");
    Subtest::wait_finish(subtest => [$sub_obj]);
    subtestResults($sub_obj);
    die $@ if $@;

}

## @name ::RemoteDump
## @usage RemoteDump($dc_filer, $dump_filer, $dump_dev, $dump_level, $dump_pt )
## @synopsis
##	Using the $dc_filer, rewind tape device ($dump_dev),
##	do a level $dump_level dump of $dump_pt.
##
sub RemoteDump( $$$$$ ) {

    my ($dc_filer, $dump_filer, $dump_dev, $dump_level, $dump_pt) = @_;
    my $TIMEOUT = 3600;

    TapeRewind($dc_filer, $dump_dev);
    subtest("dumprestore/dump_target.thpl",
            "FILER=$dump_filer",
            "DUMP_LEVEL=$dump_level",
            "DUMP_OPTIONS=u",
            "DUMP_FILE=$dump_dev",
            "DUMP_TARGET=$dump_pt",
            '-runid', "Dump-$dump_level",
            '-iffatal', '-fatal', "Dump of $dump_pt failed");
}

## @name ::RemoteRestore
## @usage RemoteRestore($rc_filer, $restore_filer, $restore_dev, $restore_pt )
## @synopsis
##	Using the $cc_filer, rewind tape device ($restore_dev),
##	do a restore to $restore_pt.
##
sub RemoteRestore( $$$$ ) {

    my ($rc_filer, $restore_filer, $restore_dev, $restore_pt) = @_;
    my $TIMEOUT = 3600;

    TapeRewind($rc_filer, $restore_dev);
    subtest("dumprestore/restore_target.thpl",
            "FILER=$restore_filer",
            "RESTORE_FILE=$restore_dev",
            "RESTORE_TARGET=$restore_pt",
            '-runid', "Restore",
            '-iffatal', '-fatal',"Restore of $restore_pt failed");

}


## @name ::DRCompare_tree
## @usage DRCompare_tree($client, $src_dir, $dst_dir, @opts)
## @synopsis
##	Calls lib::DataManip(compare_tree()), if $src_dir and $dst_dir match then
##      return 1, if they do not match then return 0.
##	If compare_trees() fails, that failure is passed back to the calling routine
##	in $@, and return value is undef.
##
sub DRCompare_tree( $$$@ ) {

    my ($client, $src_dir, $dst_dir, @opts) = @_;
    my $TIMEOUT = 3600;
    my $matched = 1;

    # We need to convert the @opts array into a hash so that compare_trees
    # doesn't give us a warning about 'using list-style options'
    my %opts_h = map { /([^=]*)=(.*)/ } @opts;
    return eval {
        my ($in1, $in2, $differ) = compare_trees("$client", "$src_dir", "$dst_dir", %opts_h);

        if ($@) {
            die("Error compare_trees($src_dir, $dst_dir): $@");
        } else {
            if (scalar(@$in1)) {
	        logcomment("Files in just $src_dir: " . join("\n", @$in1) . "\n");
		$matched = 0;
            }
            if (scalar(@$in2)) {
	        logcomment("Files in just $dst_dir: " . join("\n", @$in2) . "\n");
		$matched = 0;
            }
            if (scalar(@$differ)) {
	        logcomment("Files that differ: " . join("\n", @$differ) . "\n");
		$matched = 0;
            }
            return $matched;
        }
    } # return eval
}

## @name ::getInodeNum
## @usage getInodeNum($client, $file)
## @synopsis
##	Returns the file's inode number, if the file exists.
##	Otherwise return value is undef and $@ contains error message.
##
sub getInodeNum ( $$ ) {
    my($client, $filepath) = @_;

    my $code = '{ use FSR::DR::DRutils;
		  my $inode = &getInodeNum_local(@_);
		  logresult("FATAL", "EVAL_ERROR=$@") if ($@);
		  logresult("PASS", "INODENUM=$inode");
		 }' ;

    $code =~ s/\n\s*/ /g;

    return eval {
        my @subtest_opts = ('-client', $client,
			    '-iffatal', '-ok',
			    '-ifscript', '-ok',
                            '-onexit', sub {},
			    "--", "$filepath");

        my $obj = subtest($code, @subtest_opts);

        my $rslts = subtestResults($obj, 'INODENUM');
        die $@ if ($@);
        return $rslts->{'INODENUM'} ;
    }
}

# getInodeNum_local - return the inode of the named file. If it can not
#   get the inode then set $@ and return undef
#
sub getInodeNum_local($) {
    my ($filepath) = @_;

    return eval {
        die ("$!\n")
            unless (scalar(lstat "$filepath"));
	return (lstat("$filepath"))[1];
    }
}

## @name ::linkIt
## @usage linkIt($client, $dir, $old_f, $new_f)
## @synopsis
##	Goes to directory $dir and links (hard) $new_f to $old_f (exising) file
##	returns 1 on success, on failure linkIt sets $@ and returns undef.
##
sub linkIt ( $$$$ ) {
    my($client, $dir, $old_f, $new_f) = @_;

    my $code = '{ use FSR::DR::DRutils;
		  &linkIt_local(@_);
		  logresult("FATAL", "EVAL_ERROR=$@") if ($@);
		  logresult("PASS");
		 }' ;

    $code =~ s/\n\s*/ /g;

    return eval {
        my @subtest_opts = ('-client', $client,
			    '-iffatal', '-ok',
			    '-ifscript', '-ok',
                            '-onexit', sub {},
			    "--", $dir, $old_f, $new_f);

        my $obj = subtest($code, @subtest_opts);
        subtestResults($obj);
        die $@ if ($@);
        return 1;
    }
}

sub linkIt_local ($$$) {
    my($dir, $old_f, $new_f) = @_;

    return eval {
        die("$!\n")
          unless (chdir "$dir");

        die("$!\n")
          unless (link($old_f, $new_f));
        return 1;
    }
}


## @name ::symlinkIt
## @usage symlinkIt($client, $dir, $old_f, $new_f)
## @synopsis
##	Goes to directory $dir and symlinks $new_f to $old_f (exising) file
##	returns 1 on success, on failure sets $@ and returns undef.
##
sub symlinkIt ( $$$$ ) {
    my($client, $dir, $old_f, $new_f) = @_;

    my $code = '{ use FSR::DR::DRutils;
		  &symlinkIt_local(@_);
		  logresult("FATAL", "EVAL_ERROR=$@") if ($@);
		  logresult("PASS");
		 }' ;

    $code =~ s/\n\s*/ /g;

    return eval {
        my @subtest_opts = ('-client', $client,
			    '-iffatal', '-ok',
			    '-ifscript', '-ok',
                            '-onexit', sub {},
			    "--", $dir, $old_f, $new_f);

        my $obj = subtest($code, @subtest_opts);
        subtestResults($obj);
        die $@ if ($@);
        return 1;
    }
}

sub symlinkIt_local ($$$) {
    my($dir, $old_f, $new_f) = @_;

    return eval {
        die("$!\n")
          unless (chdir "$dir");

        die("$!\n")
          unless (symlink($old_f, $new_f));
        return 1;
    }
}

## @name ::truncIt
## @usage truncIt($client, $file)
## @synopsis
##	truncates the $file to 0 size
##	returns 1 on success, on failure set $@ and returns undef.
##
sub truncIt ( $$ ) {
    my($client, $object) = @_;

    my $code = '{ use FSR::DR::DRutils;
		  &truncIt_local(@_);
		  logresult("FATAL", "EVAL_ERROR=$@") if ($@);
		  logresult("PASS");
		 }' ;

    $code =~ s/\n\s*/ /g;

    return eval {
        my @subtest_opts = ('-client', $client,
			    '-iffatal', '-ok',
			    '-ifscript', '-ok',
                            '-onexit', sub {},
			    "--", $object);

        my $obj = subtest($code, @subtest_opts);
        subtestResults($obj);
        die $@ if ($@);
        return 1;
    }
}


sub truncIt_local ( $ ) {
    my($object) = @_;
    my $r = "/usr/local/bin/r";

    return eval {
        if (-e "$object" && ! (-f "$object")) {
	    die ("$object is not a file");
	}
	if (system("$r /bin/echo  '' > $object") ) {
	    die("Failed to truncate $object");
	}
    return 1;
    }
}

## @name ::chmodIt
## @usage chmodIt($client, $file, $newmode)
## @synopsis
##	sets the unix mode of $file to $newmode
##	returns 1 on success, on failure it sets $@ and returns undef.
##
sub chmodIt ( $$$ ) {
    my($client, $object, $newmode) = @_;

    my $code = '{ use FSR::DR::DRutils;
		  &chmodIt_local(@_);
		  logresult("FATAL", "EVAL_ERROR=$@") if ($@);
		  logresult("PASS");
		 }' ;

    $code =~ s/\n\s*/ /g;

    return eval {
        my @subtest_opts = ('-client', $client,
			    '-iffatal', '-ok',
			    '-ifscript', '-ok',
                            '-onexit', sub {},
			    "--", $object, $newmode);

        my $obj = subtest($code, @subtest_opts);
        subtestResults($obj);
        die $@ if ($@);
        return 1;
    }
}

sub chmodIt_local ( $$ ) {
    my($object, $newmode) = @_;

    return eval {
        die("$!\n")
          unless chmod $newmode, "$object" ;
        return 1;
    }
}


## @name ::moveIt
## @usage moveIt($client, $object, $target_dir)
## @synopsis
##	moves $object to $target_dir
##	returns 1 on success, on failure it sets $@ and returns undef.
##
sub moveIt ( $$$ ) {
    my($client, $object, $target) = @_;

    my $code = '{ use FSR::DR::DRutils;
		  &moveIt_local(@_);
		  logresult("FATAL", "EVAL_ERROR=$@") if ($@);
		  logresult("PASS");
		 }' ;

    $code =~ s/\n\s*/ /g;

    return eval {
        my @subtest_opts = ('-client', $client,
			    '-iffatal', '-ok',
			    '-ifscript', '-ok',
                            '-onexit', sub {},
			    "--", $object, $target);

        my $obj = subtest($code, @subtest_opts);
        subtestResults($obj);
        die $@ if ($@);
        return 1;
    }
}

sub moveIt_local ( $$ ) {
    my($object, $target) = @_;

    return eval {

	die("Failed to move $object $target: $!")
	  unless (! system("mv $object $target")) ;
	return 1;
    }
}

## @name ::makeDir
## @usage makeDir($client, $dir)
## @synopsis
##	Makes $dir directory
##	returns 1 on success, on failure sets $@ and returns undef.
##
sub makeDir ( $$% ) {
    my($client, $object, %options) = @_;

    my $code = '{ use FSR::DR::DRutils;
		  &makeDir_local(@_);
		  logresult("FATAL", "EVAL_ERROR=$@") if ($@);
		  logresult("PASS");
		 }' ;

    $code =~ s/\n\s*/ /g;

    return eval {
        my @subtest_opts = ('-client', $client,
			    '-iffatal', '-ok',
			    '-ifscript', '-ok',
                            '-onexit', sub {},
			    "--", $object, %options);

        my $obj = subtest($code, @subtest_opts);
        subtestResults($obj);
        die $@ if ($@);
        return 1;
    }
}

sub makeDir_local ( $% ) {
    my($object, %options) = @_;

    return eval {
	umask 0;
	if ($options{SUDO}) {
	  die("could not create directory $object")
	    if system("/usr/local/bin/sudo mkdir -m 777 $object");
	} else {
          my $perm = ($options{PERMISSION}) ? $options{PERMISSION} : 0777;
	  die("$!\n")
            unless (mkdir($object, $perm));
	}
	return 1;
    }
}

## @name ::deleteIt
## @usage deleteIt($client, $dir)
## @synopsis
##	Delete $dir directory, this is different from the lib/DataManip.pm routine
##      called delete_tree(). This routine uses the run-as-root 'r' command to
##	to do the deleting.
##	returns 1 on success, on failure it sets $@ and returns undef.
##
sub deleteIt ( $$ ) {
    my($client, $object) = @_;

    my $code = '{ use FSR::DR::DRutils;
		  &deleteIt_local(@_);
		  logresult("FATAL", "EVAL_ERROR=$@") if ($@);
		  logresult("PASS");
		 }' ;

    $code =~ s/\n\s*/ /g;

    return eval {
        my @subtest_opts = ('-client', $client,
			    '-iffatal', '-ok',
			    '-ifscript', '-ok',
                            '-onexit', sub {},
			    "--", $object);

        my $obj = subtest($code, @subtest_opts);
        subtestResults($obj);
        die $@ if ($@);
        return 1;
    }
}

sub deleteIt_local ( $ ) {
    my($object) = @_;

    my $r = "/usr/local/bin/r";

    return eval {
	if (! ($object =~ /^\/t/)) {
	    die ("Not deleting $object because it does not start with /t");
	}

	if ( -e "$object") {
	    system("rm -rf $object");
	    if ( -e "$object") {
	        system("$r rm -rf $object");
            }
	    if ( -e "$object") {
		die("Failed to delete $object");
	    }
	}
	return 1;
    }
}


## @name ::tarIt
## @usage tarIt($client, $wdir, $cmd)
## @synopsis
##	Go to $wdir directory and run the tar command $cmd.
##	Example: tarIt($CLIENT, "/t/fsr-u32/compat_data", "cf cu.tar compat_unix");
##	returns 1 on success, on failure it sets $@ and return undef.
##
sub tarIt ( $$$ ) {
    my($client, $wdir, $cmd) = @_;

    my $code = '{ use FSR::DR::DRutils;
		  &tarIt_local(@_);
		  logresult("FATAL", "EVAL_ERROR=$@") if ($@);
		  logresult("PASS");
		 }' ;

    $code =~ s/\n\s*/ /g;

    return eval {
        my @subtest_opts = ('-client', $client,
			    '-iffatal', '-ok',
			    '-ifscript', '-ok',
                            '-onexit', sub {},
			    "--", '$wdir', '$cmd');

        my $obj = subtest($code, @subtest_opts);
        subtestResults($obj);
        die $@ if ($@);
        return 1;
    }
}

sub tarIt_local ( $$ ) {
    my($wdir, $cmd) = @_;

    return eval {
	if (! system("cd $wdir && tar $cmd")) {
	    die("Failed tar $cmd");
	}
        return 1;
    }
}
