# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/FSR/FSRutils.pm#1 $

## @summary FSR utilities library
##
## @description
##    This package defines routines which perform some common
##    functions used in FSR tests.  These are the functions
##    that don't quite fit into any general-purpose group, and
##    are somewhat specific to the way the FSR group does automation.
##
## @dependencies
##    lib/mount.pm
##    lib/ExportFS.pm
##    snapmirror/qsm_update.thpl
##    snapmirror/qsm_update_start.thpl
##
## @bugs
##
## @author remi@netapp.com
## @status public

package FSR::FSRutils;

BEGIN {
    require Exporter;
    our @ISA = qw(Exporter);
    our @EXPORT = qw(regx_escape time_id host_component
		 inherit_params bequeath_params
		 hash_to_params params_to_hash
		 contains eject
		 serialize
		 logerror
		 setflag
		 filer_mount filer_umount
		 make_accessible
		 extractResults extractAllResults
		 worstType subtestResults
		 deleteDir
		 split_filer_path split_vol_path
		 snapmirror_sanity
		 load_compat_data
		 is_compat_data
		 get_tape_device
		 translate_tape_device
		 make_connections);
}

use mount;
use ExportFS qw(:nfs);
use Tharn;
use QtreeCmds qw(qtreeSecurity);
use filer     qw(filer_file_exists);
use Tapeinfo  qw(getTapes getTapeStatus);
use strict;

## @name ::regx_escape
## @usage regx_escape($string)
## @synopsis
##    Escapes all the characters in the string which
##    could interfere with regexp matching.
##    Returns the new escaped string.
sub regx_escape ($) {
  my ($arg) = @_;
  $arg =~ s/([\\\|\(\)\[\{\^\$\*\+\?\.])/\\$1/g;
  return $arg;
}

## @name ::time_id
## @usage time_id()
## @synopsis
##    Returns current time as one big human-readable number.
sub time_id () {
  my ($sec,$min,$hour,$mday,$mon,$year) = localtime;
  return sprintf("%04d%02d%02d%02d%02d%02d",$year+1900,$mon+1,$mday,$hour,$min,$sec);
}

## @name ::host_component
## @usage host_component($hostspec)
## @synopsis
##    Returns the shortest hostname which will properly resolve
##    to the address of the host specified by the given hostspec,
##    or an IP address if the host has no DNS mapping.
sub host_component($) {
  my ($hostspec) = @_;
  my $hostobj = host($hostspec);

  # XXX this is a hack.  We really need to do more here, in case
  # the hostspec file doesn't give us a short name.  See burt 86743.
  return $hostobj->id();
}

## @name ::inherit_params
## @usage inherit_params(@params)
## @synopsis
##    Parses all the FIELD=value strings in the given list
##    and adds them as parameters to the environment.
##    Returns success or failure with an error message in $@.
sub inherit_params {
  return eval {
    foreach $_ (@_) {
      die "Error parsing $_\n"
	unless ($_ =~ /^($Tharn::_PAT_IDENT)(\+?)=(.*)$/o);
      if ($2) {
	no strict 'refs';
	param($1, -set, $ {"::$1"} . $3);
	use strict 'refs';
      } else {
	param($1, '-set', $3);
      }
    }
    return 1;
  };
}

## @name ::bequeath_params
## @usage bequeath_params(@params_names)
## @synopsis
##    For each parameter named in the argument list,
##    retrieves the value (if defined) from the environment
##    and returns a list of FIELD=value strings.
sub bequeath_params {
  my @params = ();
  foreach $_ (@_) {
    push @params, $_ . "=" . $ENV{$_}
      if (defined $ENV{$_});

    # Unfortunately, this doesn't work unless the param()
    # call was used to grab the value from the environment.
    #no strict 'refs';
    #push @params, $_ . "=" . $ { "::$_" }
    #  if (defined $ { "::$_" });
    #use strict 'refs';
  }

  return @params;
}

## @name ::params_to_hash
## @usage params_to_hash(@params)
## @synopsis
##    Parses all the FIELD=value strings in the specified
##    list and sets them as keys into an anonymous hash.
##    On success, returns the anonymous hash.
##    On failure, returns false with an error message in $@.
sub params_to_hash {
  my $params = {};

  # This is effectively $Tharn::_PAT_IDENT, but it was requested
  # that we don't go using internal NATE/Tharn variables ourselves.
  my $pat_ident = "[A-Za-z_][A-Za-z0-9_]*";

  return eval {
    foreach $_ (@_) {
      die "Error parsing $_\n"
	unless ($_ =~ /^($pat_ident)(\+?)=(.*)$/so);
      if ($2) {
	no strict 'refs';
	$params->{$1} = $ {"::$1"} . $3;
	use strict 'refs';
      } else {
	$params->{$1} = $3;
      }
    }
    return $params;
  };
}

## @name ::hash_to_params
## @usage hash_to_params($hashref)
## @synopsis
##    Returns all the values stored in the hash specified by
##    the given hash reference as a list of FIELD=value strings.
sub hash_to_params($) {
  my ($hashref) = @_;
  my @params = ();
  my ($key, $value);
  while (($key, $value) = each %$hashref) {
    push @params, $key . "=" . $value;
  }
  return @params;
}

## @name ::contains
## @usage contains($item, @list)
## @synopsis
##    Returns whether the list contains the given item.
sub contains ($@) {
  my $item = shift @_;
  return(scalar(grep { $_ eq $item } @_));
}

## @name ::eject
## @usage eject($item, @list)
## @synopsis
##    Given a list and an item, removes every occurrence of the item from the list.
##    Returns the new list (with the item not present).
sub eject ($@) {
  my $item = shift @_;
  return(grep { $_ ne $item } @_);
}

## @name ::serialize
## @synopsis
##   Given a list or a hash, serializes it into a string.  That string
##   can then be passed to eval to obtain the original list or hash.
sub serialize {
  my @strings;
  foreach my $arg (@_) {
    my $str = $arg; # copy to avoid modifying originals
    if (defined $str) {
      $str =~ s|\\|\\\\|g;
      $str =~ s|\(|\\\(|g;
      $str =~ s|\)|\\\)|g;
      push @strings, "q($str)";
    } else {
      push @strings, 'undef';
    }
  }
  return "( " . join(", ", @strings) . " )";
}

## @name ::logerror
## @usage logerror(%options)
## @synopsis
##   This function is infrastructure to enable subtests to return a failure
##   to their calling scripts.  The current implementation accomplishes
##   this by logging a FAIL result object with a specially-formatted message
##   that the parent script parses and interprets.  However, in the future
##   the underlying mechanism may change, so scripts should not make any
##   assumptions about the way this works.  Basically, make your subtest call
##   this function when you want to "return failure and an error message"
##   to the calling script.
## @param message Optional: error message to abort with (default=$@)
## @param abort Optional: aborts the current script immediately (default=1)
sub logerror(;%) {
  my (%param) = @_;
  $param{'abort'} = 1 unless (exists $param{'abort'});

  my $type = $param{'abort'} ? 'FATAL' : 'FAIL';

  if (exists $param{'message'}) {
    # Throw a script error if someone gives us a null message.
    unless (defined $param{'message'} && $param{'message'} ne "") {
      logresult('SCRIPT', abort => $param{'abort'}, "EVAL_ERROR=no message");
    }
    logresult($type, "EVAL_ERROR=" . $param{'message'});
  } else {
    chomp $@;
    logresult($type, "EVAL_ERROR=$@");
  }
}

## @name ::setflag
## @usage setflag($filerConn, $flag, $value)
## @synopsis
##    Just a wrapper for the filer-native 'setflag' command.
##    (Sorry, but I couldn't stand the thought of importing
##    SPARSE::Resource::Filer::setflag() and the baggage therein.)
sub setflag( $$$ ) {
  my ($filerConn, $flag, $value) = @_;

  require SayWrap;
  import SayWrap qw(sayEOF);

  my $cmd = "setflag $flag $value";
  $cmd = "priv set -q diag ; priv set -q test ; $cmd";

  my $text = sayEOF($filerConn, $cmd);
  $text =~ s/^priv: unrecognized privilege: test(\n|\z)//;
  die "$text\n" if ($text);
  return 1;
}

## @name ::filer_mount
## @usage filer_mount($filerConn, $filername, $path, $CLIENT, %options)
## @synopsis
##    Exports $path on the given filer and tries mounting $filername:$path.
##    Optionally, if $CLIENT hostid is given, performs the mount
##    on the remote client, otherwise defaults to local host.
##    Returns the mount point on success, else fails the script.
##    Options hash is passed through to mount.
sub filer_mount( $$$;$% ) {
  my ($conn, $fname, $path, $CLIENT, %opts) = @_;
  my $mnt;

  logerror() unless export_path($conn, $path);

  if (defined $CLIENT) {
    $opts{client} = $CLIENT;
  }
  $mnt = mount($fname, $path, %opts);
  logerror() unless ($mnt);
  return $mnt;
}

## @name ::filer_umount
## @usage filer_umount($filerConn, $mountpoint, $path, $CLIENT)
## @synopsis
##    Unmounts the mountpoint and unexports $path on the given filer.
##    Optionally, if $CLIENT hostid is given, performs the unmount
##    on the remote client, otherwise defaults to local host.
##    Returns whether or not the path was successfully unmounted.
sub filer_umount($$$;$) {
  my ($conn, $mnt, $path, $CLIENT) = @_;
  my $rc;

  if (defined $CLIENT) {
    $rc = umount($mnt, client => $CLIENT);
  } else {
    $rc = umount($mnt);
  }
  logerror() unless $rc;

  logerror() unless unexport_path($conn, $path);

  return $rc;
}

## @name ::make_accessible
## @usage make_accessible($CLIENT, $path)
## @synopsis
##    Makes the named path, mounted on the given client, completely
##    accessible for read/write access by NetApp engineers.
sub make_accessible( $$ ) {
  my ($CLIENT, $path) = @_;

  my $code = <<'CODE';
  {
    sub sudo( $ ) {
      my $text = `/usr/local/bin/sudo $_[0]`;
      chomp $text;
      logresult("FATAL", "EVAL_ERROR=$text") if ($text);
    }
    my ($path) = @_;
    sudo("chmod ug+rw $path");
    sudo("chgrp engr $path");
    logresult("PASS");
  }
CODE
  $code =~ s/(^\s*|\s*$)//mg;
  $code =~ s/\n/ /g;

  my $obj = subtest($code,
		    '-client', $CLIENT,
		    '-onexit', sub {},
		    "--", $path);
  my $rslts = subtestResults($obj);
  die $@ if ($@);
  return 1;
}

## @name ::extractResults
## @usage extractResults($object, $type, @expected)
## @synopsis
##    Extracts the results from the given result $object (only the
##    results from the subtest itself, not its children).
##    Finds the messages of all the extracted results of the given $type,
##    and puts them into a hash.  Fails if not all the parameters named
##    in the @expected list were present.
##    On success, returns a reference to the hash of the returned results.
##    On failure, returns false and an error message in $@.
sub extractResults($$@) {
  my ($obj, $type, @expected) = @_;
  return extract_results($obj, 1, $type, @expected);
}

## @name ::extractAllResults
## @usage extractAllResults($object, $type, @expected)
## @synopsis
##    Extracts all of the results from the given result $object,
##    including the results of all of its subtests.
##    Finds the messages of all the extracted results of the given $type,
##    and puts them into a hash.  Fails if not all the parameters named
##    in the @expected list were present.
##    On success, returns a reference to the hash of the returned results.
##    On failure, returns false and an error message in $@.
sub extractAllResults($$@) {
  my ($obj, $type, @expected) = @_;
  return extract_results($obj, 0, $type, @expected);
}

# The internal workings of extractResults() and extractAllResults().
# The $own argument specifies to use own_results() or just results().
# Don't call this directly -- call one of the wrappers.
sub extract_results($$$@) {
  my ($obj, $own, $type, @expected) = @_;

  # Get the list of results with the given type
  my @rslts = grep { $_->type() eq $type } ($own ? $obj->own_results() : $obj->results());

  # Construct a list of the messages in those results
  my @msgs = map { $_->msgs() } @rslts;

  return eval {

    # Build a hash of all the result messages
    my $rslts = params_to_hash(@msgs);
    die $@ unless $rslts;

    foreach my $param (@expected) {
      die "Did not get \"${param}\" result message (" . join(", ", @msgs) . ")."
	unless (exists $rslts->{$param});
    }

    return $rslts;
  };
}

## @name ::worstType
## @usage worstType($object)
## @synopsis
##    Determines the type of the worst result returned by the given
##    subtest object, and returns it.  Ignores result types which
##    do not actually indicate the pass/failure status of a test.
sub worstType($) {
  my ($obj) = @_;
  my @rslts = grep { $_->type() !~ /^(?:WARN|INFO|PARAM)$/ } $obj->own_results();

  # No results means it passed.
  return "PASS" if (scalar(@rslts) == 0);

  my $type = (Rslt::worst(@rslts))->type();
  return $type;
}

## @name ::subtestResults
## @usage subtestResults($object, @expected)
## @synopsis
##    This is a convenience function, catered to the special (yet common)
##    case where you have a subtest object, and you just want to extract
##    a certain set of results if the test passed, or the EVAL_ERROR
##    result if the test failed.  It also presumes that you don't care
##    about warnings/comments -- if it didn't fail, you want the results,
##    not the warnings.  Pass in the subtest object and the list of expected
##    results.  If the subtest succeeded, this function returns the results
##    hash as returned by extractResults().  If the subtest failed, this
##    function extracts the EVAL_ERROR results, and returns failure with
##    the EVAL_ERROR result message in $@.  (If any errors are encountered
##    while extracting results, this function returns failure with an error
##    message in $@.)
sub subtestResults($@) {
  my ($obj, @expected) = @_;

  return eval {
    my $rslts;
    my $type = worstType($obj);

    if ($type eq 'PASS') {
      # Subtest passed (or only logged warnings, info, etc.).
      $rslts = extractResults($obj, 'PASS', @expected);
      die $@ if ($@);
    } else {
      # Subtest failed -- extract EVAL_ERROR and die with it.
      $rslts = extractResults($obj, $type, qw(EVAL_ERROR));
      die $@ if ($@);
      die $rslts->{'EVAL_ERROR'} . "\n";
    }

    return $rslts;
  }
}

## @name ::deleteDir
## @usage deleteDir($filerConnection)
## @synopsis
##    Deletes a tree completely from the file system.
##    Returns success or failure, with an error in $@ on failure.
sub deleteDir($) {
    my ($dirpath) = @_;

    return eval {
      my $sudo = "/usr/local/bin/r";
      my $text = `$sudo /bin/rm -rf $dirpath`;
      die "$text\n" if $text;
      return 1;
    };
}

## @name ::split_filer_path
## @usage split_filer_path($filer_path)
## @synopsis
##    Takes a "filer path" argument string, i.e. [<filer>:]<path>,
##    where path can be a filer volume name, a windows-style path,
##    or a normal path.  Parses the path, and returns a tuple of
##    the path and the filer name, if present.  If the filer was
##    not present, the second value return is undef.
sub split_filer_path($) {
  my ($fpath) = @_;

  if ($fpath =~ /^(.+?):(.+)$/) {
    return ($2, $1);
  } else {
    return ($fpath, undef);
  }
}

## @name ::split_vol_path
## @usage split_vol_path($path)
## @synopsis
##    Given a path on a filer, splits out the volume name and
##    returns it.  If the path doesn't look like a path on a filer,
##    then returns undef with an error message in $@.
sub split_vol_path($) {
  my ($path) = @_;

  return eval {
    die "invalid path\n"
      unless ($path =~ m@^/vol/(\w[^/\s]*)(/.+)?$@);
    return $1;
  }
}

## @name ::snapmirror_sanity
## @usage snapmirror_sanity(@params)
## @synopsis
##    Essentially just a sanity check, intended to be run
##    at the start of every snapmirror script.
##    Makes sure the relationship specified by the default
##    snapmirror parameters is idle (waiting if necessary)
##    and heals the mirror if necessary.
sub snapmirror_sanity(@) {
  my @bg_params = @_;
  my @fg_params = grep { $_ ne '-bg' } @bg_params;
  my $running_in_background = (scalar(@fg_params) != scalar(@bg_params));
  my $obj;

  # Sanity check -- try doing a simple update

  # First, start the update
  $obj = subtest("snapmirror/sm_update_start.thpl",
		 @fg_params,
		 '-runid', "SM-UpdateStart",
		 '-iffatal', '-ok');
  logresult('FATAL', "EVAL_ERROR=$@")
    unless subtestResults($obj);

  # Get the results of the update, and extract the base snapshot
  my $rslts = extractResults($obj, 'PASS', 'BASE_SNAPSHOT');
  tstlog('-script', $@) unless $rslts;
  push @bg_params, "BASE_SNAPSHOT=" . $rslts->{'BASE_SNAPSHOT'};

  # Wait for it to finish (passing the base snapshot)
  $obj = subtest("snapmirror/sm_update.thpl",
		 "COMPLETE=1",
		 @bg_params,
		 '-runid', "SM-InitialUpdateComplete",
		 '-iffatal', '-ok');

  # If we ran it in the foreground, check the results.
  unless ($running_in_background) {
    logresult('FATAL', "EVAL_ERROR=$@")
      unless subtestResults($obj);
  }

  return $obj;
}

## @name ::load_compat_data
## @usage load_compat_data($filerName, $path, $compat_dump_gzip)
## @synopsis
##	Takes a 'gzip'ed dump image and restores it at $path.
##	The $compat_dump_gzip argument is optional. If it is not
##	supplied then 
##	/x/eng/http/html/engineering/qa/fsr/dump_image/compat_data.dmp.gz
##	is used. The data, either the default compat_data.dmp.gz or
##	$compat_dump_gzip, is passed through gunzip and then feed
##	to restore on the $filerName filer. The actual command is
##	"gunzip -c <data> | rsh $filerName restore rfD - $path"
##	Uses die() to throw exception if command does not work.
##	
sub load_compat_data($$;$)
{
    my ($filerName, $path, $compat_dump_gzip) = @_;

    my $hfiler = host($filerName);
    my $cfiler = connect("#rsh-socket");
    my $filername = $hfiler->{hostname};
    my $filer_passwd = $cfiler->{password};
    my $filer_user = $cfiler->{username};
    
    if (! $compat_dump_gzip) {
        $compat_dump_gzip = "/x/eng/http/html/engineering/qa/fsr/dump_images/compat_data.dmp.gz";
    }

    # Do a little sanity checking...
    #
    if (! -r $compat_dump_gzip) {
        die("$compat_dump_gzip is not readable!");
    }
    my $filer_cmd = "rsh $filername -l $filer_user:$filer_passwd";
    my $cmd = "gunzip -c $compat_dump_gzip | $filer_cmd restore rfD - $path";
    logcomment("$cmd");
    system($cmd) and die("Command \"$cmd\" failed");
    # Command to remove restore_symboltable created by the restore command that it creates
    # This is to avoid creation of rstab_<file> names in the event restored files are dump and restored again
    $cmd = "priv set -q diag; rm $path/restore_symboltable";
    logcomment($cmd);
    system ("$filer_cmd \"$cmd\"") and die ("Command \"$cmd\" failed");
}

## @name ::is_compat_data
## @usage is_compat_data($filerConn, $vol_name, @check_list)
## @synopsis
##    Returns 1 if $vol_name looks to be a version of DPG's compat_data data set.
##    @check_list is optional. If @check_list is not supplied then is_compat_data will
##    look for these standard items; $vol_name/compat_unix is a unix qtree,
##    $vol_name/compat_ntfs is a ntfs qtree, $vol_name/compat_mixed is a mixed qtree,
##    $vol_name/fscreator/streams/file_with_1_stream exists and
##    $vol_name/fsrceator/aclsdir/root_full exists. 
##    If @check_list is set, is_compat_data will look for the standard
##    items plus whatever is in @check_list. 
##    Names in @check_list list should be relative to $vol_name. 
##    is_compat_data returns 0 if it does not find all the standard objects and
##    whatever is in @check_list.
sub is_compat_data($$;@) {
    my ($filerConn, $vol_name, @check_list) = @_;

    if (QtreeCmds::qtreeSecurity($filerConn, "/vol/$vol_name/compat_unix") ne "unix") {
        logcomment("/vol/$vol_name/compat_unix not a unix qtree: $@");
        return 0;
    }
    if (QtreeCmds::qtreeSecurity($filerConn, "/vol/$vol_name/compat_ntfs") ne "ntfs") {
        logcomment("/vol/$vol_name/compat_ntfs not a ntfs qtree: $@");
        return 0;
    }
    if (QtreeCmds::qtreeSecurity($filerConn, "/vol/$vol_name/compat_mixed") ne "mixed") {
        logcomment("/vol/$vol_name/compat_mixed not a mixed qtree: $@");
        return 0;
    }
    if (!filer_file_exists($filerConn, "/vol/$vol_name/fscreator/streams/file_with_1_stream")) {
        logcomment("/vol/$vol_name/fscreator/streams is missing: $@");
        return 0;
    }
    if (!filer_file_exists($filerConn, "/vol/$vol_name/fscreator/aclsdir/root_full")) {
        logcomment("/vol/$vol_name/fscreator/aclsdir/root_full is missing: $@");
        return 0;
    }

    my $item;
    foreach $item (@check_list) {
        if (!filer_file_exists($filerConn, "\'/vol/$vol_name/$item\'")) {
            logcomment("/vol/$vol_name/$item is missing: $@");
            return 0;
        }
    }

    return 1;
}

## @name ::get_tape_device
## @usage get_tape_device($fc, $device)
## @synopsis
##    If a device is supplied then if it is ready and write-enabled return
##    the tape device, otherwise throw an exception using die().
##
##    If no device is supplied then try to find one that is ready and
##    write-enabled.  Return the device name if one is found, otherwise
##    throw an exception using die().
sub get_tape_device( $;$ );
sub get_tape_device( $;$ ) {
    my ($fc, $tape_dev) = @_;

    # if tape device specified make sure it is ready and write-enabled.
    #
    if ($tape_dev) {

        # if a file name is passed in just return it.
        #
        if ($tape_dev =~ /^\//) {
            return $tape_dev;
        }

        # $tape_dev could be <filer>:<device>
        # In this case we need to connect to <filer> and use
        # that connection when calling getTapeStatus, rather than
        # the connection passed in ($fc)
        if ($tape_dev =~ /:/) {
            my ($remote_fname, $remote_dev) = split(/:/, $tape_dev);
            host($remote_fname);
            my $remote_conn = connect("#rsh-socket");
            get_tape_device($remote_conn, $remote_dev);

            # We get here it means the <filer>:<device> is ready, write-enabled
            # so return it.
            return $tape_dev;
        }

        # It's a local tape device
        #
        my %tape_status = getTapeStatus($fc, $tape_dev);

        if ($tape_status{status} =~ /ready, write enabled/ ) {
            return $tape_dev;
        } else {
            die("Device $tape_dev status is not \'ready, write enable\'");
        }

    # no tape device specified so go find one
    #
    } else {

        # get list of all tape device available, then look for a
        # rewindable one that is ready and write-enabled
        #
        my @avail_tapes = getTapes($fc);
        foreach my $tape (@avail_tapes) {
            if ($tape =~ /^r.*a$/ ) {	# find first high density rewindable device
                my %tape_status = getTapeStatus($fc, $tape);
                if ($tape_status{status} =~ /ready, write enabled/ ) {
                    return $tape;
                }
            }
        }
        die("No tape device specified and none found");
    }
}

## @name ::translate_tape_device
## @usage translate_tape_device($SrcName, $SrcDevice, $DstName)
## @synopsis
##    Given a tape device name ($SrcDevice) from the perspective 
##    of $SrcName (a hostname) translate that tape device name into 
##    the correct name from  $DstName's perspective.
##    On success it returns the translated name, on failure it
##    throws an exception using die().
sub translate_tape_device($$$)
{
    my ($SrcName, $SrcDevice, $DstName) = @_;
    my $tape;

    if ($SrcName eq $DstName) {
        return "$SrcDevice";

    } else {		# SrcName != DstName

        # Since SrcName != DstName the tape device must be a real tape device.
        # It cannot be a file.
        #
        if ($SrcDevice =~ /\// )  {  # name start with a slash, this is wrong
            die("Since $SrcName != $DstName a file name like $SrcDevice cannot be used");

        }

        # SrcDevice has <filer>:<device> format so no translation to do 
        # unless <filer> equals $DstName. In that case we chop off <filer>:
        # return just the device name.
        #
        if ($SrcDevice =~ /:/ ) {
            if ($SrcDevice =~ /$DstName:/) {
                my @dstname = split(/:/, $SrcDevice);
                return $dstname[1];
            }
            return "$SrcDevice";
        }

        return "$SrcName:$SrcDevice";
    }
}


##
## @name ::make_connections
##
## @usage make_connections($FILER_CONN, $FILER, $FILER_SRC, $FILER_DST)
##
## @synopsis
##	This routine handles making filer connections based on $FILER,
##	$FILER_SRC and $FILER_DST in a standard way. The function
##	returns 4 values; source host, source connection, destination host
##	and destination connection. In general the following three
##	argument sequences are allowed:
##
##	FILER	   Both source host and destination host are set to FILER.
##
##	FILER_SRC  The source host is set to FILER_SRC and the destination
##	FILER_DST  host is set to FILER_DST.
##
##	FILER	   The source host is set to FILER and the destination
##	FILER_DST  host is set to FILER_DST.
##
##	FILER and FILER_SRC are allowed if they both contain the same
##	value. Invalid or missing paramters cause an execption using die().
##
##	Obviously the source and destination connections are made to the 
##	appropriate host.
##	Note: FILER_SRC and FILER_DST are optional.
##

sub make_connections($$;$$) {
    my ($FILER_CONN, $FILER, $FILER_SRC, $FILER_DST) = @_;

    die("Both FILER ($FILER) and FILER_SRC ($FILER_SRC) supplied.")
	if (($FILER && $FILER_SRC) && ($FILER ne $FILER_SRC));
    die("Either FILER or FILER_SRC must be supplied.")
	if (!($FILER || $FILER_SRC));
    die("Both FILER_SRC and FILER_DST must be supplied.")
	if ($FILER_SRC && !($FILER_DST));

    if ($FILER) {
	$FILER_SRC = $FILER;
	$FILER_DST = $FILER if (!($FILER_DST));
    }

    my $SrcHost = host($FILER_SRC);
    my $SrcConn = connect($FILER_CONN);
    my $DstHost = host($FILER_DST);
    my $DstConn = connect($FILER_CONN);

    return ($SrcHost, $SrcConn, $DstHost, $DstConn);
}


1;
