# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/FSR/FijiDB.pm#1 $

## @summary Database for Fault Injection Testing using FIJI
##
## @description
##    This is a library database package that consists of:
##    - All PREJOBs including creating file(s) of particualr sizes,
##      directory structure etc.
##    - All the actions/operations that could trigger a test point.
##	Actions include the outcome of hitting a testpoint, which could
##    	be panicing/rebooting etc.
##    - All POSTJOBs that including removing the dataset created during the
##	PREJOB
##    - A verifier function that is common to all the actions
##
## @dependencies
##    FSR/FSRutils.pm
##    FSR/SM/CommonSMutils.pm
##    lib/DataManip.pm
##    snapmirror/sm_init.thpl
##    snapmirror/sm_update.thpl
##    snapmirror/sm_resync.thpl
##    snapmirror/SMutils.pm
##    lib/Volume.pm
##    lib/VolCopy.pm
##    lib/filer.pm
##    lib/SnapCmds.pm
##
## @author neha@netapp.com

package FSR::FijiDB;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(%fiji_prejobs %fiji_postjobs %fiji_transfers fiji_verifier
		    fiji_clear_all);


use strict;
use filer;
use SnapCmds;
use snapmirror::SMutils;
use FSR::FSRutils;
use Volume qw(vol_container);
use fileSystem;
use FSR::SM::CommonSMutils qw(compare_src_dst);
use Tharn;

our (%fiji_prejobs, %fiji_postjobs, %fiji_transfers);  # exported dbs

## @name ::fiji_prejobs
## @synopsis
##    The prejobs include creating the right dataset in order to hit a
##    FIJI testpoint. We create a hash where:
##    - keys are names of prejobs
##    - values are the code that creates the dataset
%fiji_prejobs = (
  create_nfs_data => 'my $create_nfsdata = sub {
			 my $srcvol_mnt = filer_mount($FILERSRC, $FSRCNAME, "/vol/$FILER_SRC_PATH", $CLIENT);
			   die "$@"
			     unless ($srcvol_mnt);

             	         # Generate some data on source
              		 my @param_list = env_options(populate_opts());
		         my @options = bequeath_params(@param_list);
              		 unless (populate_tree($CLIENT, $srcvol_mnt, @options)){
			   return 0;
		         }
			 filer_umount($FILERSRC, $srcvol_mnt, "/vol/$FILER_SRC_PATH", $CLIENT);
			 return 1;
              	       }; $create_nfsdata->();',
);


## @name ::fiji_postjobs
## @synopsis
##    The postjobs include deleting the dataset created during the PREJOB
##    We create a hash where:
##    - keys are names of postjobs
##    - values are strings of code that delete the dataset
%fiji_postjobs = (
  delete_nfs_data => 'my $delete_nfsdata = sub {
			 my $srcvol_mnt = filer_mount($FILERSRC, $FSRCNAME, "/vol/$FILER_SRC_PATH", $CLIENT);
			 die "$@"
			   unless ($srcvol_mnt);

             	         # Delete data on source
 		         my @param_list = grep !/CONTENTS_ONLY/, delete_opts();
			 my @options = bequeath_params(@param_list);
			 push @options, "CONTENTS_ONLY=1";
        	         unless (delete_tree($CLIENT, $srcvol_mnt, @options)) {
			   return 0;
			 }
			 filer_umount($FILERSRC, $srcvol_mnt, "/vol/$FILER_SRC_PATH", $CLIENT);
			 return 1;
              	       }; $delete_nfsdata->();',
);


## @name ::fiji_transfers
## @synopsis
##    This includes all the transfers that can be started in order to hit
##    a testpoint.
##    We create a has where:
##    - keys are names of the transfers
##    - values are strings of code that start and wait for a transfer to finish
%fiji_transfers = (
  snapmirror_initialize => 'my $init_code = sub {
              		       my $iobj = subtest("snapmirror/sm_init.thpl",
                               "COMPLETE=1",
                               "-runid", "sm_init_".$NUM,
                               "-onexit", sub {});
              		       # For eval to return a string in $@, we need
              		       # to propagate any failures from the subtest
			       # to this level and then die with the error
              		       # message obtained from the child subtest.
              		       subtestResults($iobj);
              		       die "$@" if ($@);
              		       return 1;
             		     }; $init_code->();',
 
  volcopy => 'my $volcopy_code = sub {
	         die "$@"
		   unless (snapshot_create($filerSrc, $FILER_SRC_PATH,
					   $VOLCOPY_SNAPSHOT));
                 # Since this code is executed in the calling library (e.g.
		 # FijiDeswizzle.pm), we use the variables which that function
		 # knows about.
		 host($fdstname);
		 my $filerDstConn = connect($FILER_CONN);
	         die "$@"
		   unless volcopy_start($filerDstConn, $FILER_DST_PATH,
					SOURCE => "$fsrcname:$FILER_SRC_PATH",
					SNAPSHOT => $VOLCOPY_SNAPSHOT, COMPLETE => 1,
					RESTRICT => 1, TIMEOUT => $TIMEOUT); 
	         die "$@"
 		   unless (vol_online($filerDst, $FILER_DST_PATH));
		 return 1;
	       }; $volcopy_code->();',
);

## @name ::fiji_verifier
## @usage fiji_verifier($FILERSRC, $FILERDST, src, $dst, $DELAY, $TIMEOUT)
## @synopsis
##    This function uses treecompare to compare the data on source and
##    destination and fails if there is a mismatch. Then it runs wafl iron on
##    the aggregate containing the destination volume to make sure that there
##    is no data corruption.
sub fiji_verifier($$$$$$;$) {
  my ($filerSrc, $filerDst, $src, $dst, $DELAY, $TIMEOUT, $volcopy_ss) = @_;

  my ($fsrcname, $FILER_SRC_PATH) = split/:/, $src;
  my ($fdstname, $FILER_DST_PATH) = split/:/, $dst;

  ##########################################################################
  # Run treecompare between source and destination 
  ##########################################################################
  if ($volcopy_ss) {
    # VolCopy Transfer
    my $src_path = "/vol/$FILER_SRC_PATH/.snapshot/$volcopy_ss";
    my $dst_path = "/vol/$FILER_DST_PATH/.snapshot/$volcopy_ss";
   
    logresult('FATAL', "EVAL_ERROR=treecompare found a mismatch")
      unless (treecompare($filerDst, $dst_path, $src_path, timeout => $TIMEOUT, filer_dst=>$fsrcname, args=>'-V'));
  } else {
    # Snapmirror Transfers
    die "$@"
      unless (compare_src_dst($filerSrc, $filerDst, $FILER_SRC_PATH, $FILER_DST_PATH, $TIMEOUT, $fsrcname, $fdstname));
  }
  ############################################################################
  # Run WAFL Iron on the destination volume
  ############################################################################
  # Run wafl iron on the aggregate containing destination volume
  my $dst_aggr = vol_container($filerDst, $FILER_DST_PATH);
  logresult('FAIL', "EVAL_ERROR=$@")
    if ($@);
  logresult('FAIL', "EVAL_ERROR=Could not fetch the containing aggregate of the destination volume")
    unless ($dst_aggr);

  my $ironObj = subtest("wafl/wafliron/wafliron_aggr.thpl",
                        "FILER=$fdstname", "FILER_AGGR=$dst_aggr",
			"TIMEOUT=$TIMEOUT",
                        '-runid', "wafliron_after_bootup",
                        '-onexit', \&propagate);
}

## @name ::fiji_clear_all
## @usage fiji_clear_all($filerDst)
## @synopsis
##    This function clears all the FIJI testpoints that are set on the
##    filer. It logs a FATAL result in case of a failure and returns 1 on
##    success.
sub fiji_clear_all($) {
  my ($filerDst) = @_;

  my $output = filer_cli_test_say($filerDst, "fiji clear all");
  if (!($output =~ /filter cleared/i)) {
         logresult('FATAL', "EVAL_ERROR=$output");
  }
  return 1;
}

# This is an internal function and should never be called from any script.
# This function is called when a subtest exists. It will look through all the
# results logged by the subtest and will log a failure, if one is found.
sub propagate($) {
  my ($obj) = @_;

  my $t = worstType($obj);
  my @rslts = grep { $_->type eq $t } $obj->own_results();
  return unless (scalar(@rslts));
  if ($rslts[0]->severity() >= 4) {
    logerror(message => join(" ", $rslts[0]->msgs()));
  }
}

1;
