# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/FSR/rpl_cmds.pm#1 $

## @summary Replica (SnapMirror/SnapVault) commands library
##
## @description
##    This package defines internal routines used by SnapMirror
##    and SnapVault libraries.  Much of the ONTAP code for these
##    products is shared, and the interfaces are similar, so there
##    is little cause to separate the common library functionality.
##
##    For snapmirror verifier, there are different levels of verification
##    that a user can select. Please look at the documentation in SMutils.pm
##    for details.
##
## @dependencies
##    Module/Common.pm
##    lib/SayWrap.pm
##
## @bugs
##
## @author dl-dpgqa-automation@netapp.com
## @status Private

package FSR::rpl_cmds;

use warnings;
use strict;

BEGIN {
  use Module::Common;
  Module::Common::setup_export_tags(
	error_codes => [ qw($RPL_SUCCESS $RPL_NO_DESTINATIONS) ],
	parse => [ qw(rpl_cmd_from_opts) ],
	xfer => [ qw(rpl_start_xfer rpl_finish_xfer
		     rpl_start_snap rpl_finish_snap
		     rpl_start_op rpl_finish_op) ],
	internal => [ qw(rpl_strip_header
			 rpl_status_internal rpl_details_internal
			 rpl_destinations_internal rpl_dst_snap_internal
			 rpl_abort_internal rpl_release_internal
			 rpl_convert_internal rpl_break_internal) ]
  );
}

use Carp;
use SayWrap;
use Tharn;
use strict;

our $RPL_SUCCESS = "Transfer started";
our $RPL_NO_DESTINATIONS = "no known destinations";

# These options take no arguments.
my %cmd_flags = (
		   WAIT => 'w',
		   RESTART => 'r',
		   FORCE => 'f',
		   NOACTION => 'n',
		   ALL_BLOCKS => 'a',
		   CKSUM => 'z',
		   VERIFIER_QUICK => 'Q',
		   VERIFIER_FIX => 'F',
	        );

# These options all take arguments.
my %cmd_options = (
		     SNAPSHOT => 's',
		     CREATE_SNAPSHOT => 'c',

		     # Ideally, we would just use the lower-case versions which
		     # match the option names, depending upon other libraries.
		     THROTTLE => 'k',
		     kbs => 'k',
		     TRIES_COUNT => 't',
		     tries => 't',
	             wsize => 'b',
		     connection_mode => 'p', 
		     SOURCE => 'S',
		     source => 'S',
		     VERIFIER_SNAPSHOT => 'd',
		     VERIFIER_CKSUM => 'L',
		     VERIFIER_IGNORE => 'i',
		     # what is M for ? It needs release agnostic switches too..
		     # NEWOPTION => 'M',
	          );

## @name ::rpl_cmd_from_opts
## @usage rpl_cmd_from_opts($cmdname, $subcmd, $dst, %options)
## @synopsis
##    Given the command name, subcommand name, destination path, and
##    hash reference of options, constructs and returns the replica-based
##    command to send to the filer, with all appropriate options and
##    priviledge settings..
sub rpl_cmd_from_opts($$$;%) {
    my ($cmdname, $subcmd, $dst, %opts) = @_;
    my $cmd = "$cmdname $subcmd";

    # XXX Someday, maybe we can make this unneccessary.
    if ((exists $opts{THROTTLE} && exists $opts{kbs}) ||
	(exists $opts{TRIES_COUNT} && exists $opts{tries}) ||
	(exists $opts{SOURCE} && exists $opts{source})) {
      croak "mixed option types\n";
    }

    while (my ($opt, $letter) = each %cmd_flags) {
      if ($opts{$opt}) {
	$cmd .= " -${letter}";
      }
    }

    while (my ($opt, $letter) = each %cmd_options) {
      if ($opts{$opt}) {
	$cmd .= sprintf(" -%s %s", $letter, $opts{$opt});
      }
    }

    $cmd .= " $dst";

    # Some options only work in higher priviledge modes.
    if ($opts{CKSUM}) {
      $cmd = "priv set -q test ; " . $cmd;
    } elsif ($opts{ALL_BLOCKS} || $opts{wsize}) {
      $cmd = "priv set -q diag ; " . $cmd;
    } elsif ($cmdname =~ /^snap(mirror|vault) check$/) {
      $cmd = "priv set -q diag ; " . $cmd;
    }

    return $cmd;
}


## @name ::rpl_start_xfer
## @usage rpl_start_xfer($connobj, $cmdname, $cmd, %options)
## @synopsis
##    Basically just an operation-specific wrapper for rpl_start_op(),
##    where that operation is a transfer.  Uses the given command name
##    to construct the approprate waitstring and calls rpl_start_op().
sub rpl_start_xfer( $$$;% ) {
  my ($conn, $cmdname, $cmd, %opts) = @_;
  my $command = '[Ss]nap' . ($cmdname eq 'snapvault' ? 'vault' : 'mirror');

  # This is a hardcoded regexp which corresponds to the ONTAP code.
  # The parentheses also determine how we determine errors below.
  # Don't change these unless you know exactly what you're doing.


  # Case 1, transfer successfully started.
  my $waitstring = "(?m)" .
    "^(?:Transfer|$command check) started\\.\\s+" .
    "Monitor progress with '$cmdname status(?: -l)?' " .
      "or the snap(?:mirror|vault) log\\.\\s*(?:\\n|\$)";

  # Case 2, transfer queued.  (Failure unless ALLOW_DEFER.)
  $waitstring .= "|^" . ($opts{ALLOW_DEFER} ? "(" : "") .
    "Transfer not initially successful, retrying\\." .
    ($opts{ALLOW_DEFER} ? ")" : "") . "\\s+" .
    "Monitor progress with '$cmdname status(?: -l)?' " .
      "or the snapmirror log\\.\\s*(?:\\n|\$)";

  # Case 3, transfer failed.
  $waitstring .= "|^Transfer aborted: (.+?)\\.\\s*(?:\\n|\$)";

  # Case 4, resync successfully started
  $waitstring .= "|^".
	"(?:NOTE: Destination volume .+? is already a replica\\.)?\\s*".
	"(?:NOTE: Resync will not need to revert the volume\\.)?\\s*".
	"The resync base snapshot will be: .+?\\s+" .
    "(?:These older snapshots have already been deleted from the source\\s+" .
    "and will be deleted from the destination:\\s+.*?)?\\s*" .
	"(?:\\s*.+?\\s*){0,}?".
    "(?:Volume .+? will be briefly unavailable before coming back " .
    "online\\.)?\\s*" .
    "(?:Revert to resync base snapshot was successful\\.)?\\s*" .
    "Transfer started\\.\\s+" .
    "Monitor progress with '$cmdname status(?: -l)?' " .
      "or the snapmirror log\\.\\s*(?:\\n|\$)";

  return rpl_start_op($conn, $cmd, $waitstring, %opts);
}

## @name ::rpl_finish_xfer
## @usage rpl_finish_xfer($connobj, $cmdname, %options)
## @synopsis
##    Basicaly just an operation-specific wrapper for rpl_finish_op(),
##    where that operation is a transfer.  Uses the given command name
##    to construct the approprate waitstring and calls rpl_finish_op().
sub rpl_finish_xfer($$;%) {
  my ($conn, $cmdname, %opts) = @_;
  my $command = '[Ss]nap' . ($cmdname eq 'snapvault' ? 'vault' : 'mirror');

  # This is a hardcoded regexp which corresponds to the ONTAP code.
  # The parentheses also determine how we determine errors below.
  # Don't change these unless you know exactly what you're doing.
  my $waitstring = "(?m)" .
    "^Transfer (?:successful|failed: (.+?))\\.\\s*(?:\\n|\$)|" .
    "^(?:$command check completed successfully|" .
      "($command check failed: .+?))\\.\\s*(?:\\n|\$)|" .
    "^($command check completed with mismatches detected)\\.\\s+" .
    "Look at the snapmirror log for report of detected mismatches\\.\\s*(?:\\n|\$)";

  return rpl_finish_op($conn, $waitstring, %opts);
}

## @name ::rpl_start_snap
## @usage rpl_start_snap($connobj, $cmdname, $cmd, %options)
## @synopsis
##    Basically just an operation-specific wrapper for rpl_start_op(),
##    where that operation is a snapshot creation.  Uses the given command
##    name to construct the approprate waitstring and calls rpl_start_op().
sub rpl_start_snap($$$;%) {
  my ($conn, $cmdname, $cmd, %opts) = @_;

  # This is a hardcoded regexp which corresponds to the ONTAP code.
  # The parentheses also determine how we determine errors below.
  # Don't change these unless you know exactly what you're doing.
  my $waitstring = "(?m)" .
    "^Snapshot creation started\\.\\s+" .
    "Monitor progress with '$cmdname status -s'\\.\\s*(?:\\n|\$)|" .
    "^Snapshot creation not initially successful, retrying\\.\\s+" .
    "Monitor progress with '$cmdname status -s'\\.\\s*(?:\\n|\$)|" .
    "^Snapshot creation aborted: (.+?)\\.\\s*(?:\\n|\$)";

  return rpl_start_op($conn, $cmd, $waitstring, %opts);
}

## @name ::rpl_finish_snap
## @usage rpl_finish_snap($connobj, $cmdname, %options)
## @synopsis
##    Basicaly just an operation-specific wrapper for rpl_finish_op(),
##    where that operation is a snapshot creation.  Uses the given command
##    name to construct the approprate waitstring and calls rpl_finish_op().
sub rpl_finish_snap($$;%) {
  my ($conn, $cmdname, %opts) = @_;

  # This is a hardcoded regexp which corresponds to the ONTAP code.
  # The parentheses also determine how we determine errors below.
  # Don't change these unless you know exactly what you're doing.
  my $waitstring = "(?m)" .
    "^Snapshot creation (?:successful|failed: (.+?))\\.\\s*(?:\\n|\$)";

  return rpl_finish_op($conn, $waitstring, %opts);
}

## @name ::rpl_start_op
## @usage rpl_start_op($connobj, $cmd, $waitstring, %options)
## @synopsis
##    Given the connection to send to, raw command text to start an operation
##    with, command-specific waitstring, and a reference to the options list
##    for sayWAIT(), tries starting the desired operation.  The waitstring
##    passed in must match, without storing any regx-clusters, if the operation
##    was successfully started.  Optionally, if the waitstring matches and
##    does store a regx-cluster, that is interpreted as a specific failure
##    message to cull out.  This function returns whatever non-waitstring text
##    was seen (to be identified by the caller), and if the operation failed to
##    start, the error message is returned in $@.
sub rpl_start_op($$$;%) {
  my ($conn, $cmd, $waitstring, %opts) = @_;

  return eval {
    my ($matched, $text) = sayWAIT($conn, $cmd, $waitstring, %opts);
    die $@ if ($@); # sayWAIT failed
    unless ($matched) {

      # waitstring never matched
      die "$text\n" if ($text);
      die "unexpected end of output\n";
    }

    # Re-match the matched string to determine success/failure.
    die "waitstring did not match returned string\n"
      unless ($matched =~ /$waitstring/);

    # Get out with the error message if the transfer failed.
    die "$1\n" if (defined $1);
    die "$2\n" if (defined $2);
    die "$3\n" if (defined $3);
    die "$4\n" if (defined $4);
    die "$5\n" if (defined $5);

    # Transfer was successful, return any other observed text.
    return $text;
  }
}

## @name ::rpl_finish_op
## @usage rpl_finish_op($connobj, $waitstring, %options)
## @synopsis
##    Given the connection to listen on and a command-specific waitstring,
##    waits for the text that indicates an existing operation completed.
##    Logs a PASS or FATAL result to indicate whether or not the operation
##    completed successfully.  If the operation failed, then the FATAL
##    result message (with key EVAL_ERROR) contains the error message.
##    Note, this function is primarily intended to run as a backgrounded
##    subtest which reports the success/failure of the operation with
##    NATE results, but does not pass those results back to the parent.
sub rpl_finish_op($$;%) {
  my ($conn, $waitstring, %opts) = @_;
  my $text = rpl_start_op($conn, "", $waitstring, %opts);
  chomp $@;
  logresult('FATAL', "EVAL_ERROR=$@") if ($@);
  logresult('FATAL', "EVAL_ERROR=$text") if ($text);
  logresult('PASS');
}


# Strip the header off a Snap(Mirror|Vault) status command's output.
# Will die without catching the exception if an error occurs.
# This should be caught by the immediate parent.
sub rpl_strip_header($$) {
  my ($cmd, $text) = @_;

  my $header;
  if ($cmd eq "snapmirror") {
    $header = "Snapmirror is (on|off)\\.";
  } elsif ($cmd eq "snapvault") {
    # Allowed primary|secondary keyword to be missing
    # in the case where they are the same filer
    $header = "Snapvault (primary |secondary |)is (ON|OFF)\\.";
  } else {
    die "Invalid command type \"$cmd\".\n";
  }

  die "Did not receive string \"$header\".\n"
    unless ($text =~ s/^$header\s*(\n|\z)//);

  return $text;
}

sub rpl_status_internal($$$;$) {
  my ($cmd, $filer, $dst, $src) = @_;
  $src = "" unless($src); # Prevents warning "Use of uninitialized value..."

  return eval {
    my $text = sayEOF($filer, "$cmd status $dst $src", TIMEOUT => 180);
    $text = rpl_strip_header($cmd, $text);

    # No status entry for the given dst or dst/src
    my @ret = ();
    return @ret if ($text eq "");

    die "Malformed $cmd status output\n"
      unless ($text =~ s/^.*Source\s+Destination\s+State\s+Lag\s+Status\s*\n//s);

    if ($src) {
      if ($text =~ /^($src)\s+($dst)\s+([\w\-]+)\s+([\w\:\-]+)\s+(.+?)\s*$/m) {
	@ret = ($1, $2, $3, $4, $5);
      } # else no status entry with the given src/dst
    } else {
      if ($text =~ /^(\S+)\s+($dst)\s+([\w\-]+)\s+([\w\:\-]+)\s+(.+?)\s*$/m) {
	@ret = ($1, $2, $3, $4, $5);
      } # else no status entry with $dst as a destination
    }

    return @ret;
  }
}

sub rpl_details_internal($$;$$) {
  my ($cmd, $filer, $dst, $src) = @_;
  $src = "" unless($src); # Prevents warning "Use of uninitialized value..."
  $dst = "" unless($dst); # Prevents warning "use of uninitialized value..."

  return eval {
    my $text = sayEOF($filer, "$cmd status -l $dst $src", TIMEOUT => 180);
    $text = rpl_strip_header($cmd, $text);

    if ( $dst ) {
      # Check src only if dst is passed
      if (!$src && ($text =~ s/^.*?(Source:\s+?(\S+)\s+?Destination:\s+?$dst\s.*?Last Transfer From:\s+\S+).*$/$1/s)) {
        $src = $2; # set source
      } elsif ($src && ($text =~ s/^.*?(Source:\s+?$src\s+?Destination:\s+?$dst\s.*?Last Transfer From:\s+\S+).*$/$1/s)) {
      # do nothing, source already set
      } else {
        return undef;
      }
    }

    my @list;
    my %entry;
    # If no dst passed, then return all statuses in an array of hashes
    # otherwise return a pointer to a single hash
    foreach my $line (split(/\n/,$text)) {
      next unless ( $line or $dst ); # Skip blank lines if no dst specified
      if ($line =~ /^(\S.+):\s+(.+)$/) {
	my $field = $1;
	my $data = $2;
	$data =~ s/^\s*(.*?)\s*$/$1/;
	$data = "" if ($data eq "-");
	$entry{$field} = $data;
	unless ( $dst ) {
	  if ( $field =~ /last transfer from/i ) {
	    my %data = %entry;
	    push @list, \%data if $field =~ /last transfer from/i;
	    %entry = {};
	  }
	}
      } else {
	die "Unparsable line in $cmd status output: $line\n";
      }
    }
    return ($dst ? \%entry : @list);
  }
}

sub rpl_destinations_internal($$;$) {
  my ($cmd, $filer, $src) = @_;
  $src = "" unless($src); # Prevents warning "Use of uninitialized value..."
  $src =~ s/^\S+:(\S+)$/$1/; # Chop off filer name if given

  return eval {
    my $text = sayEOF($filer, "$cmd destinations $src");

    # Return nothing if there are no destinations
    if ($text =~ /^$cmd destination.*?: $RPL_NO_DESTINATIONS$/) {
      if ($src) {
	my @destlist = ();
	return @destlist;
      } else{
	my %desthash = ();
	return(\%desthash);
      }
    }

    die "Malformed $cmd destinations output\n"
      unless ($text =~ s/^Path\s+Destination\s*\n//s);

    my $dst_pattern = '^(\S+)\s+(\S+)\s*' . "\$";

    if ($src) {
      # Source was specified.
      # The command should have only printed destinations for the given source.
      # Return a list of those destinations.
      my @destlist = ();

      foreach my $line (split(/\n/,$text)) {
	next if ($line =~ /^\s*$/);
	die "Error parsing line: $line\n"
	  unless ($line =~ /$dst_pattern/o);
	my ($tmp_src, $tmp_dst) = ($1, $2);

	die "Received unexpected source: $tmp_src\n"
	  unless ($tmp_src eq $src);
	push @destlist, $tmp_dst;
      }

      return(@destlist);

    } else {
      # Source was not specified.
      # Return a hash of sources to destinations.
      my %desthash = ();

      foreach my $line (split(/\n/,$text)) {
	next if ($line =~ /^\s*$/);
	die "Error parsing line: $line\n"
	  unless ($line =~ /$dst_pattern/o);
	my ($tmp_src, $tmp_dst) = ($1, $2);

	my $listref = $desthash{$tmp_src};
	if (defined $listref) {
	  push @$listref, $tmp_dst;
	} else {
	  $desthash{$tmp_src} = [ $tmp_dst ];
	}
      }

      return(\%desthash);

    }
  }
}

sub rpl_dst_snap_internal($$$$) {
  my ($cmd, $filer, $src, $dst) = @_;
  $src =~ s/^\S+:(\S+)$/$1/; # Chop off filer name if given
  my $srcexp = quotemeta($src);
  my $dstexp = quotemeta($dst);

  return eval {
    my $text = sayEOF($filer, "$cmd destinations -s $src");

    # Return nothing if there are no destinations
    return undef if ($text =~ /^$cmd destination(s): $RPL_NO_DESTINATIONS$/);

    die "Malformed $cmd destinations -s output\n"
      unless ($text =~ s/^Path\s+Snapshot\s+Destination\s*\n//s);

    if ($text =~ /^$srcexp\s+(\S+)\s+$dstexp\s*$/m) {
      my $snap = $1;
      die "Encountered \"Unknown\" snapshot.\n"
	if ($snap eq "Unknown");
      # Found the src/dst entry -- return the snapshot name
      return $snap;
    } else {
      # No entry for this src/dst combination
      return undef;
    }
  }
}

sub rpl_abort_internal($$$;%) {
  my ($cmd, $filer, $dst, %opts) = @_;

  return eval {
    my $text;
    if ($opts{'HARD'}) {
      $text = sayEOF($filer, "$cmd abort -h $dst");
    } else {
      $text = sayEOF($filer, "$cmd abort $dst");
    }
    die "$text\n"
      if ($text);
    return 1;
  }
}

sub rpl_release_internal($$$$) {
  my ($cmd, $filer, $src_targ, $dst) = @_;

  return eval {
    my $text = sayEOF($filer, "$cmd release $src_targ $dst");
    die "$text\n" if ($text);
    return 1;
  }
}

sub rpl_convert_internal($$$) {
  my ($cmd, $filer, $dst) = @_;

  return eval {
    my $dstexp = quotemeta($dst);
    my $text = sayEOF($filer, "priv set -q diag; $cmd convert $dst");
    die "$text\n"
      unless ($text =~ /^Qtree $dstexp has been successfully converted\.?$/);
    return 1;
  }
}

sub rpl_break_internal {
  my ($cmd, $filer, $dst, %opts) = @_;

  my $break_info_msg = "Volume size is being retained for potential snapmirror resync.  If you would like to grow the volume and do not expect to resync, set vol option fs_size_fixed to off.";

  return eval {

    my $text;
    if ($opts{'HARD'}) {
      $text = sayEOF($filer, "priv set -q diag; $cmd break -h $dst");
    } else {
      $text = sayEOF($filer, "$cmd break $dst");
    }

    # Strip off the expected string.
    my $cmd_exp = quotemeta($cmd);
    my $dst_exp = quotemeta($dst);
    $dst_exp =~ s/.*://g; # remove the name of filer if present
    unless ($text =~ s/^$cmd_exp break: Destination $dst_exp is now writable\.\s*//) {
      die "$text\n";
    }

    # Strip the informational message that gets printed when you break volumes.
    my $info_exp = quotemeta($break_info_msg);
    $text =~ s/^$info_exp\s*//;

    # Anything left is an error message.
    die "$text\n" if ($text);

    return 1;
  }
}


1;
