# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/FSR/rpl_log.pm#1 $

## @summary Utilities for parsing and verifying replica logs
## @author remi@netapp.com
## @status public
## @pod here

=head1 NAME

rpl_log - Utilities for parsing and verifying replica logs

=head1 DESCRIPTION

This module defines routines for parsing and verifying replica logs.
These include the SnapMirror and SnapVault log, F</etc/log/snapmirror>,
and the robust WORM logs such as the transferred-file log.

This module is not and should not be made dependent on NATE, so that
command-line wrappers for these tools can be easily maintained and run.

=head1 FUNCTIONS

=cut

package FSR::XferEntry;
use warnings;
use strict;

my @Fields = qw(epoch operation path stream from);
my @field_names = map { __PACKAGE__ . "::" . $_ } @Fields;
use fields @Fields;

use Carp;
use DataManip::cmd_utils;

=head2 FSR::XferEntry::new

This is the constructor for a transferred-file log entry object.  The
constructor must be passed a complete, ordered list of all the entry
values, or a serialized version of an xfer entry object.

External users would typically use the C<parse_xfer_entry()>
function below to build new XferEntry objects, or might
possibly pass a serialized object to this constructor.

=cut

sub new {
  my $class = shift;
  my @vals = ();

  if (scalar(@_) == 1) {
    # Create from a serialized log entry.
    @vals = eval $_[0];
    if ($@) { croak $@; }

  } else {
    # Create from parsed entry fields.
    @vals = @_;
  }

  my @from = splice(@vals, 4);
  unless (scalar(@vals) == 4) {
    croak "invalid arguments to " . __PACKAGE__ . " constructor\n";
  }

  if (scalar(@from)) {
    # Recursively build the second corresponding entry first.
    unless ((scalar(@from) == 1) || (scalar(@from) == 4)) {
      croak "invalid arguments to " . __PACKAGE__ . " constructor\n";
    }
    if ((scalar(@from) == 1) && (ref($from[0]) eq __PACKAGE__)) {
      push @vals, @from;
    } else {
      push @vals, new FSR::XferEntry(@from);
    }
  }
  my $self = fields::phash(\@field_names, \@vals);
  return bless($self, $class);
}

=head2 FSR::XferEntry I<accessors>

Each xfer log entry object has accessor methods for the entry values
C<epoch>, C<operation>, C<path>, and C<stream>.

=head2 FSR::XferEntry::from

Some xfer log operations (currently just "Rename To") have a corresponding
second xfer log (i.e. "Rename From").  Once parsed, this second entry is
associated with the original entry and can be retrieved through this accessor.

=cut

# Create a closure for each accessor method.
foreach my $field (@Fields) {
  my $slot = __PACKAGE__ . "::$field";
  my $method = sub {
    my $self = shift;
    $self->{$slot} = shift if (scalar(@_));
    return $self->{$slot};
  };
  no strict 'refs';
  *$field = $method;
  use strict 'refs';
}

=head2 FSR::XferEntry::serialize

Flattens the xfer log entry object into a scalar string,
which can be passed around and stored more easily.

Later, this serialized version can be passed to a constructor
to get back the original xfer log entry object.

=cut

sub serialize {
  my $self = shift;

  # Must use map(), not values(), in order to preserve order.
  my @vals = map { $self->{$_} } @field_names;

  # Get values from the second corresponding entry.
  my $struct = pop @vals;
  if (defined $struct) {
    my $type = ref($struct);
    unless ($type eq __PACKAGE__) {
      croak "unexpected nested object type: $type";
    }
    my @from_vals = map { $struct->{$_} } @field_names;
    my $from_from = pop @from_vals;
    if (defined $from_from) {
      croak "structures nested too deep";
    }
    push @vals, @from_vals;
  }

  my $str = DataManip::cmd_utils::serialize(@vals);
  return $str;
}

=head2 FSR::XferEntry::datestr

Returns the date and time of the entry as a human-readable string,
instead of the number of seconds since the epoch.

=cut

sub datestr {
  my $self = shift;

  my $slot = __PACKAGE__ . "::epoch";
  my $datestr = gmtime($self->{$slot});
  unless ($datestr =~ s/(\ \d{4})\z/ GMT$1/) {
    croak "error constructing date string";
  }
  return $datestr;
}

=head2 FSR::XferEntry::sprint

Converts the entire entry to a human-readable string and returns it.

=cut

sub sprint {
  my $self = shift;

  my ($epoch, $operation, $path, $stream, $from) =
    map { $self->{$_} } @field_names;

  my $str = "";
  if (defined $from) {
    $str .= $from->sprint();
  }
  $str .= sprintf("%s %s %s%s\n", $self->datestr(), $operation,
		  $path, ((defined $stream) ? ":$stream" : ""));
  return $str;
}

package FSR::rpl_log;
use warnings;
use strict;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw();
our @EXPORT_OK = qw(verify_xfer_log);

use Carp;
use File::Spec;
use DataManip::cmd_utils qw(serialize);
use DataManip::manip_log qw(parse_datestring read_path_extent
			    parse_manipulation);
use chain_db;

# Predeclare so we can override it.
sub gribbit;
# Override gribbit() with croak().
# We'll override it again later if necessary.
*gribbit = \&croak;

# Capitalizes the given word and returns it.
sub capitalize( $ ) {
  my ($word) = @_;
  unless ($word =~ /\A(.)/s) {
    croak "cannot captilize zero-length word\n";
  }
  substr($word, 0, 1, uc($1));
  return $word;
}

# Given numbers A and B, returns whether they are within
# some delta of each other (default delta=0).
sub within( $$;$ ) {
  my ($a, $b, $delta) = @_;
  $delta = 0 unless (defined $delta);
  return (abs($a - $b) <= $delta);
}

=head2 parse_xfer_entry

Given an open filehandle for a transferred-files log, reads off
exactly one entry and parses it, taking into account the fact tha
paths can contain newlines.  However, since every xfer log entry ends
in a newline, this should only read and parse some integer number of
lines from the filehandle, with no extra characters to spare
(i.e. nothing should be ignored).

Returns a false value if EOF is hit while reading from the filehandle.
Otherwise, returns a new XferEntry object, initialized from the
parsed xfer log entry.  Croaks on any error.

=cut

sub read_xfer_log_entry( $ ); # prototype first

sub parse_xfer_entry( $ ) {
  my ($fh) = @_;

  # Get the entry values.
  my @vals = read_xfer_log_entry($fh);

  # Check for EOF.
  return $vals[0] if (scalar(@vals) < 4);

  # If the first entry has a corresponding second entry,
  # get another set of entry values.
  if ($vals[1] eq "Rename From") {
    my @data = read_xfer_log_entry($fh);
    if (scalar(@data) < 4) {
      croak "EOF after Rename From\n";
    }
    unless ($data[1] eq "Rename To") {
      croak "Rename From followed by $data[1]\n";
    }
    unshift @vals, @data;
  }

  return new FSR::XferEntry(@vals);
}

# The core of the parse_xfer_entry(), separated out so that we can
# call it twice for entries with corresponding second entries.
sub read_xfer_log_entry( $ ) {
  my ($fh) = @_;

  # Get the next line, but don't immediately chomp it,
  # since the entry path could span lines.
  my $entry = <$fh>;
  unless ($entry) {
    # Get out when we hit EOF.
    return $entry;
  }

  # Strip off and parse the date.
  unless ($entry =~ s/\A(.{28})//) {
    croak "invalid entry date: \"$entry\"\n";
  }
  my $epoch = parse_datestring($1);

  # Grab the action and path lengths.
  unless ($entry =~ s/\A\       # a space
	  ([A-Za-z ]+?)\s+      # Action
	  (\d+)(?:\:(\d+))?\s+  # Path length, and maybe stream name length
	  //x) {
    croak "invalid entry action: \"$entry\"\n";
  }
  my ($action, $path_length, $stream_length) = ($1, $2, $3);

  # Get the path, reading extra lines as necessary.
  # Make sure that the path starts with a './'.
  my $path = read_path_extent($entry, $fh, $path_length);
  unless ($path =~ /\A\.\//) {
    croak "invalid entry path: $path\n";
  }

  # If necessary, get the stream, and make sure it's separated by a ':';
  my $stream = undef;
  if (defined $stream_length) {
    $stream = read_path_extent($entry, $fh,
			       $stream_length + 1); # 1 extra for colon
    unless ($stream =~ s/\A\://) {
      croak "entry stream not colon-delimited: $stream\n";
    }
  }

  # Now we can chomp the entry, after which there shouldn't be anything left.
  chomp $entry;
  if (length($entry) > 0) {
    croak "extra cruft at end of entry: $entry\n";
  }

  return ($epoch, $action, $path, $stream);
}

sub xfer_db_remove_verify( $$$$ ) {
  my ($db, $pathkey, $xle, $delta) = @_;

  # Lookup and remove the entry from the database
  # indexed under the same path with the same operation.
  my $entry = $db->remove($pathkey,
			  sub { $_[0]->operation eq $xle->operation } );
  unless (defined $entry) {
    gribbit "missing expected entry: " . $xle->sprint();
    return;
  }
  unless (ref($entry) eq 'FSR::XferEntry') {
    croak "removed unexpected object type: " . ref($entry) . "\n";
  }

  # Make sure the removed entry really matches.
  unless (within($entry->epoch, $xle->epoch, $delta)) {
    gribbit sprintf("timestamp mismatch, expected \"%s\": %s",
		    $xle->datestr(), $entry->sprint());
    return;
  }
}

=head2 verify_xfer_log($manip_log, $xfer_log, %options)

Given paths to a log file of manipulations made to a SnapVault primary
filesystem (as produced by C<DataManip::modify_tree()>) and the
corresponding transferred-file log of the received changes from the
SnapVault secondary, verifies the contents of the transferred-file log
to ensure that it correctly recorded all changes.

Returns true on success, indicating that the $xfer_log correctly
represented all manipulations listed in the $manip_log.

Dies on any error.

Available options:

=over 4

=item DELTA

Specifies a number of seconds (default=5), which is the maximum amount
by which timestamps on corresponding entries are allowed to differ by.

=item PERSEVERE

Attempt to run to completion, regardless of any encountered errors.
Instead, report all errors at the end, instead of when the first
is encountered.

=item DEBUG

Produce debugging output.  Force complete stack traces on all errors.
(Unfortunately, does this globally, so it can affect other functions
in this module and submodules.)

=back

=cut

sub verify_xfer_log( $$% ) {
  my ($manip_log, $xfer_log, %opts) = @_;

  # Import the non-existent Carp symbol to turn on stack traces.
  if ($opts{DEBUG}) {
    require Carp;
    import Carp qw(verbose);
  }

  unless (exists $opts{DELTA}) {
    $opts{DELTA} = 5;
  }

  # If we're trying to continue in the face of errors, we accumulate
  # errors in a string.  At the end, we'll call gribbit() without any
  # arguments to croak() with all the accumulated errors.
  if ($opts{PERSEVERE}) {
    my $errors = "";
    undef *gribbit;
    *gribbit = sub {
      if (scalar(@_)) {
	$errors .= join("", @_);
      } else {
	undef *gribbit;
	*gribbit = \&croak;
	return unless ($errors);
	croak $errors;
      }
    };
  }

  unless (-f $manip_log) {
    croak "could not find manipulations log file\n";
  }
  unless (-f $xfer_log) {
    croak "could not find transferred-files log file\n";
  }

  # Create and open a database file.
  my $xfer_db = new chain_db(OBJECT_PACKAGE => 'FSR::rpl_log',
			     OBJECT_CLASS => 'FSR::XferEntry');

  # Run through transferred file log, one entry at a time.
  # Put each entry into a disk-backed database, indexed by file path.

  my $fh; # empty filehandle reference
  open($fh, "<$xfer_log") or croak "$!\n";

  while (my $xle = parse_xfer_entry($fh)) {

    # Store stream paths as subdirectories.
    my $pathkey = $xle->path;
    if (defined $xle->stream) {
      $pathkey .= '/' . $xle->stream;
    }

    $xfer_db->store($pathkey, $xle);
  }

  close($fh) or croak "$!\n";

  # Create a database of the manipulations log, just for lookup purposes.
  my $manip_db = new chain_db(OBJECT_PACKAGE => 'DataManip::manip_log',
			      OBJECT_CLASS => 'DataManip::ManipEntry');

  $fh = undef; # empty filehandle reference
  open($fh, "<$manip_log") or croak "$!\n";

  while (my $mle = parse_manipulation($fh)) {
    $manip_db->store($mle->path, $mle);
  }

  close($fh) or croak "$!\n";

  # Run through the manipulations log, one entry at a time.
  # For each manipulation entry, determine which corresponding xfer-log entries
  # should have resulted.  Remove each such entry from the database, or croak.

  $fh = undef; # empty filehandle reference
  open($fh, "<$manip_log") or croak "$!\n";

  # Track all the directories which were modified by creating/removing
  # entries in them.  We have to track this across all iterations and check
  # it at the end, because multiple entries can modify the same directory.
  my %modified_dirs = ();
  my $track_modified_parent = sub {
    my ($xle) = @_;

    # Just for paranoia's sake.
    if ($xle->stream) {
      croak "no code to track parents of streams";
    }

    my @entries = ($xle);
    if (defined $xle->from) {
      push @entries, $xle->from;
    }

    foreach my $entry (@entries) {
      # Get the parent directory of the newly created/deleted file.
      my ($vol, $dirs, $file) = File::Spec->splitpath($entry->path);
      my $parent_dir = File::Spec->catpath($vol, $dirs, "");
      $parent_dir =~ s/(?<!\A\.)\/+\z//; # strip trailing slashes

      my $foo = './dir0/dir0/dir0/';
      if ($parent_dir eq $foo) {
	print "Parent directory $parent_dir: (really $vol|SEP|$dirs) ";
      }

      # Lookup the parent directory in the manipulations database.
      if ($manip_db->lookup($parent_dir,
			   sub { (($_[0]->action eq 'create') &&
				  ($_[0]->type eq 'directory')) } )) {
	# The directory was just created,
	# so there shouldn't be a modify for it.

	if ($parent_dir eq $foo) {
	  print "created, not storing.\n";
	}

	next;
      }

      # Only record the latest timestamp for this directory modification.
      if (exists $modified_dirs{$parent_dir} &&
	  $entry->epoch <= $modified_dirs{$parent_dir}) {
	# This timestamp is older, so ignore it.
	if ($parent_dir eq $foo) {
	  print "old timestamp, " . $entry->epoch . "\n";
	}
      } else {
	$modified_dirs{$parent_dir} = $entry->epoch;
	if ($parent_dir eq $foo) {
	  print "storing, timestamp is " . $entry->epoch . "\n";
	}
      }
    }
  };

  while (my $mle = parse_manipulation($fh)) {

    # Construct the corresponding entry that should be in the database.
    my $xle = new FSR::XferEntry($mle->epoch, undef,
				 $mle->path_without_stream,
				 $mle->stream_name);

    if (($mle->action eq "modify") ||
	($mle->action eq "grow") ||
	($mle->action eq "shrink")) {
      if ($mle->type eq "ACL") {
	$xle->operation("Modify Attr");
      } else {
	$xle->operation("Modify File");
      }

    } elsif (($mle->action eq "create") ||
	     ($mle->action eq "delete")) {
      my $act = capitalize($mle->action);
      if ($mle->type eq "directory") {
	$xle->operation("$act Dir");
	$track_modified_parent->($xle);
      } elsif ($mle->type eq "ACL") {
	$xle->operation("Modify Attr");
      } else {
	$xle->operation("$act File");
	$track_modified_parent->($xle);
      }

    } elsif ($mle->action eq "rename") {
      my $from = new FSR::XferEntry($mle->epoch, "Rename From",
				    $mle->path2_without_stream,
				    $mle->stream2_name);
      $xle->operation("Rename To");
      $xle->from($from);
      $track_modified_parent->($xle);

    } elsif ($mle->action eq "hardlink") {
      croak "unimplemented\n";

    } elsif ($mle->action eq "chmod") {
      croak "unimplemented\n";

    } else {
      croak "invalid action: " . $mle->action . "\n";
    }

    xfer_db_remove_verify($xfer_db, $mle->path, $xle, $opts{DELTA});
  }

  # Make sure all the modified directories were reflected in the xfer log.
  while (my ($dirpath, $epoch) = each %modified_dirs) {
    # Make sure we look for the corresponding directory modfication.
    my $xle = new FSR::XferEntry($epoch, "Modify Attr",
				 $dirpath, undef);

    xfer_db_remove_verify($xfer_db, $dirpath, $xle, $opts{DELTA});
  }

  close($fh) or croak "$!\n";

  # Any entries left in the database are extra log cruft.
  my @entries = ();
  while (my $pathkey = $xfer_db->each) {
    foreach my $xle ($xfer_db->lookup($pathkey)) {
      push @entries, $xle->sprint();
    }
  }
  if (scalar(@entries) > 0) {
    gribbit "cruft entries found in xfer log:\n" . join("", @entries);
  }

  if ($opts{PERSEVERE}) {
    gribbit();
  }
}

=head1 BUGS

Similarly, C<XferEntry::datestr()> always returns a string in the GMT
timezone, instead of the local time zone.

=cut


1;
