# $Id$
#
#  Tharndoc comments:
#

## @summary Module used for Antivirus testing 
## @status reviewed
## @author Michael Eddy
## @pod here

=head1 NAME

    AVUtils.pm

=head1 DESCRIPTION

    This will provide AV routines needed for FIT testing

=head1 EXAMPLES

    TODO

=head1 COPYRIGHT

    "Copyright 2008, Network Appliance Inc., All Rights Reserved."
          
=head1 STATUS 
    Shared
         
=head1 AUTHOR
         
    Michael Eddy
             
=head1 ROUTINES

=cut

# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDE_VOLUMEMODIFY}) {
    my $newModule = $ENV{OVERRIDE_VOLUMEMODIFY};
    delete $ENV{OVERRIDE_VOLUMEMODIFY};
    print "The main VolumeModify module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package AVUtils;
use TracePkg;
use Data::Dumper;
use Server;
use Params::Common;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF UNDEF);
my $Trace = new TracePkg();
my $Server = new Server();
my $self;

# Class Data, only accessible from this class
{
    my $refCount = 0;

    sub incRef {
        $refCount++;
    }
    sub getRef {
        return $refCount;
    }
}

sub new
{
    $Trace->entry();
    my $class = shift(@_);

    # Check to see if this class has already been declared. If so, then we can
    # just return its reference.

    if ( getRef() ) {
        $Trace->exit(msg => "Package has already been declared. Returning...");
        return $self;
    }

    # Bless and return
    bless ($self, $class);
    incRef();

    $Trace->exit();
    return $self;

} 


#
# AVUtils::find_avoa_policies
#
#
#
=head2 find_avoa_policies

    AVUtils::find_avoa_policies(search_params=>$params, wantarray=>0);

    This method will return a string or an array of avoa policies that match search_params.

=over

=item search_params
    
    (Required) : The regular expression to use when searching for avoa policies

=item wantarray

    (Required) : If true an array of avoa policies will be returned. If false a string of comma separated avoa names will be returned.

=back

=cut

sub find_avoa_policies
{
    $Trace->entry();
    my %params = Params::Validate::validate (
	@_,
	{   
	    wantarray => {type => SCALAR},
	    search_params => {type => ARRAYREF},
	}
    );
    $Trace->debug(msg=>"Dumping AVUtils::find_avoa_policies params:\n".Dumper(\%params));


    my @policies;
    my $server = new Server; 
    my $Topology = $server->discover_topology(cmd_nodes => [qw(avoa_policy)]);
    $Trace->debug(msg=>"Dumping topology:\n".Dumper($Topology));
    my @search_results = $server->search_topology(
	topology => $Topology,
	cmd_node => "avoa_policy",
	key_pairs => $params{search_params},
    );
    $Trace->debug(msg=>"Dumping discovered avoa policies:\n".Dumper(\@policies));
    foreach my $unnamed (@search_results) {
	#$Trace->debug(msg=>"Dumping unnamed:\n".Dumper($unnamed));
	my $pol = $$unnamed{name};
	push @policies, $pol;
	$Trace->debug(msg=>"Added $pol to list of policies");
    }
    my @sorted_policies = sort @policies;
    my $to_return;
    if ($params{wantarray} == 0) {
	$to_return = join ',', @sorted_policies;
	$Trace->debug(msg=>"Our Policies: $to_return");
    } else {
	$to_return = \@sorted_policies;
	$Trace->debug(msg=>"Our Policies:\n ".Dumper($to_return));
    }

    $Trace->exit();
    return $to_return;
} # end find_avoa_policies

1;
