# this is basically -*- perl -*-
# $Id: Disk.pm,v 1.19 2008/08/22 22:15:36 emartinez Exp $
#
## @summary	Disk - library routines to support disk commands
## @author	dl-cf-qa@netapp.com
## @status	private
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Disk - library routines to support disk commands

=head1 SYNOPSIS

  use cluster::Disk;

  # Run the 'disk show' command
  $disk_h = disk_show(conn => $conn);

  # Assign a disk to the specified owner
  $result = disk_assign(conn => $conn, disk => '0a.19', owner => 'cfo-u32');

  # Assign the disks to the specified sysid
  $result = disk_assign(
      conn => $conn,
      disk => ['0a.19', '0a.20', '0a.21'],
      sysid => $sysid,
  );

  # Assign all unowned disks to the current filer
  $result = disk_assign(conn => $conn, disk => 'all');

  # Auto assign all unowned disks
  $result = disk_assign(conn => $conn, disk => 'auto');

  # Re-assign all the disks owned by 'cfo-u32' to a new sysid
  $result = disk_reassign(
      conn => $conn,
      old_owner => 'cfo-u32',
      new_sysid => $sysid,
  );

  # Enable software disk ownership on the filer
  $result = disk_upgrade_ownership(conn => $conn);

  # XXX need some examples
  $result = disk_remove_ownership();

  # Enable partner mode
  partner_mode(enable => 1);

  # Enable maintenance mode
  maint_mode(enable => 1);

  # Get notes from the last function call
  $notes = get_notes();

=head1 DESCRIPTION

library routines to support disk commands. It supports the 'show', 'assign',
'reassign', 'upgrade_ownership' and 'remove_ownership'. Each of these functions,
unless otherwise noted, support both the partner and maintenance mode. When
running in partner mode, either by globally enabling it using the 'partner_mode'
function or by setting the 'partner' option on the function call, the 'disk'
command is prefixed with the 'partner' keyword when sent to the filer.
If the filer is in maintenance mode, the 'maint' option must be set or globally
enable it by using the 'maint_mode' function, in order to expect the correct
maintenance mode prompt.

  # Example: for both partner_mode and maint_mode, all subsequent calls
  # to any disk command function will be done in that mode:
  partner_mode(enable => 1);
  # Return disk information in partner mode:
  $disk_h = disk_show(conn => $conn);

  # Return disk information in normal mode, regardless if partner mode
  # is enabled globally
  $disk_h = disk_show(conn => $conn, partner => 0);

The 'get_notes' function is used to retrieve the notes or info messages given
by some commands which otherwise will be discarded.

  # Example: the following gets the disk information and calling
  # get_notes will return any info messages returned by the call to 'disk show'
  # which if there are any unowned disks then $notes will have the string:
  # "NOTE: Currently 1 disks are unowned. Use 'disk show -n' ..."
  $disk_h = disk_show(conn => $conn);
  $notes = get_notes();

=head1 STATUS

  Private

=head1 AUTHOR

  dl-cf-qa@netapp.com
  $Id: Disk.pm,v 1.19 2008/08/22 22:15:36 emartinez Exp $
  Last revised $Date: 2008/08/22 22:15:36 $

=head1 COPYRIGHT

  Copyright 2008, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package cluster::Disk;

use strict;
use warnings;

use base qw(Exporter);

use Carp qw(croak);
use SayWrap qw(sayEOF);
use boot qw(transit_to);
use Params::Validate qw(validate SCALAR OBJECT ARRAYREF);

our @EXPORT_OK = qw(
    disk_show
    disk_assign
    disk_reassign
    disk_upgrade_ownership
    disk_remove_ownership
    partner_mode
    maint_mode
    get_notes
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '1.0';

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#---------------------------------------------------------------------
# Local variables
#---------------------------------------------------------------------
my $disk_cmd_note;
my $partner_mode = 0;
my $maint_mode = 0;
my $maint_mode_prompt = '\\*>';

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub disk_show(%;);
sub disk_assign(%;);
sub disk_reassign(%;);
sub disk_upgrade_ownership(%;);
sub disk_remove_ownership(%;);
sub partner_mode(%;);
sub maint_mode(%;);
sub get_notes();
sub _check_for_note(\$;);
sub _say($$$$;@);
sub _chop_messages($;);

#---------------------------------------------------------------------

=head2 disk_show

    $disk_h = disk_show(conn => $conn
                        [, all => 0|1 ]
                        [, owner => $owner ]
                        [, sysid => $sysid ]
                        [, all_owned => 0|1 ]
                        [, not_owned => 0|1 ]
                        [, reservations => 0|1 ]
                        [, partner => 0|1 ]
                        [, maint => 0|1 ]
                        [, options => $options ]);

This function returns the list of all disks and the owner of the disks
which are part of Sanown configuration.
With no arguments, it returns all disks which the filer has access to.

The following named arguments are accepted:

=over

=item Arguments

=over

=item conn

(Required) : Connection object to a filer.

=item all

(Optional) : Return all disks which the filer has access to,
including any current ownership.
Default is 0.

=item owner

(Optional) : Return only the disks owned by the named node given
by the hostname.

=item sysid

(Optional) : Return only the disks owned by the named node given
by the sysid.

=item all_owned

(Optional) : Return all disks which have an owner.
Default is 0.

=item not_owned

(Optional) : Return all disks which don't have an owner.
Default is 0.

=item reservations

(Optional) : Include the reservation holder for each disk.
Default is 0.

=item partner

(Optional) : Run command in partner mode.

=item maint

(Optional) : Command is being run in maintenance mode.

=item options

(Optional) : Options to pass to the 'disk show' command as is,
it will override any other options.

=back

=item Returns

Returns hash reference where the disk name is the key or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub disk_show(%;) {
    my (%opts) = validate (
        @_,
        {
            conn         => { type => OBJECT },
            all          => { type => SCALAR,  default => 0,
                                                 regex => qr/^[01]$/ },
            owner        => { type => SCALAR, optional => 1 },
            sysid        => { type => SCALAR, optional => 1 },
            all_owned    => { type => SCALAR,  default => 0,
                                                 regex => qr/^[01]$/ },
            not_owned    => { type => SCALAR,  default => 0,
                                                 regex => qr/^[01]$/ },
            reservations => { type => SCALAR,  default => 0,
                                                 regex => qr/^[01]$/ },
            partner      => { type => SCALAR, optional => 1,
                                                 regex => qr/^[01]$/ },
            maint        => { type => SCALAR, optional => 1,
                                                 regex => qr/^[01]$/ },
            options      => { type => SCALAR, optional => 1 },
        }
    );
    my (%ret, $disk, @list, @keys, %hash, $key, $out);
    my @owner_keys = qw(owner sysid);
    my @home_keys  = qw(home home_sysid);
    my @other_keys = qw(pool serial_number);
    my $options = '';

    if ($opts{options}) {
        # Override all other options
        $options = $opts{options};
    } else {
        # Set any option specified
        $options .= ' -v' if $opts{all};
        $options .= " -o $opts{owner}" if defined $opts{owner};
        $options .= " -s $opts{sysid}" if defined $opts{sysid};
        $options .= ' -a' if $opts{all_owned};
        $options .= ' -n' if $opts{not_owned};
        $options .= ' -z' if $opts{reservations};
    }

    # Send command to filer
    $out = _say($opts{conn}, "disk show $options", $opts{maint}, $opts{partner});

    # Remove all "NOTE:" messages
    _check_for_note($out);

    # Check for proper header information and remove it
    if ($out && !($out =~ s/^.*\bDISK\s.*\s*\n--[-\s]+//ms)) {
        croak $out;
    }

    # Process one line at a time
    foreach (split /\n/, $out) {
        # Get the disk name
        s/^(\S+)\s+//;
        $disk = $1;

        # Get optional entries
        %hash = (/([^\s=]+)=\s*(\S+)/g);
        s/\s+[^\s=]+=\s*\S+//g;
        $ret{$disk}{$_} = $hash{$_} foreach keys %hash;

        # Get owner and home of disk
        if (s/\s*Not Owned\s*//) {
            @list = ('Not Owned');
        } else {
            @list = (/([^\s(]*)\s*\(([^)]*)\)/g);
            s/[^\s(]*\s*\([^)]*\)\s*//g;
        }
        @keys = %hash ? @home_keys : ();
        @{$ret{$disk}}{@owner_keys, @keys} = @list;

        # Get the pool and serial number
        @list = (/(\S+)(?:\s+(\S+))?/);
        @{$ret{$disk}}{@other_keys} = @list;

        # Set all undefined values to the empty string or -1 for sysid keys
        foreach $key (keys %{$ret{$disk}}) {
            if (!defined $ret{$disk}{$key}) {
                $ret{$disk}{$key} = $key =~ /sysid/ ? -1 : '';
            }
        }
    }

    return \%ret;
}

#---------------------------------------------------------------------

=head2 disk_assign

    $result = disk_assign(conn => $conn
                          [, disk => $disk|\@disks ]
                          [, owner => $owner ]
                          [, sysid => $sysid ]
                          [, pool => $pool ]
                          [, count => $count ]
                          [, force => 0|1 ]
                          [, partner => 0|1 ]
                          [, maint => 0|1 ]);

This function assigns an owner to the specified disks.

The following named arguments are accepted:

=over

=item Arguments

=over

=item conn

(Required) : Connection object to a filer.

=item disk

(Optional) : Disk or list of disks to assign owner to.
Default is 0.

=item owner

(Optional) : Assign the disk(s) to this owner given by the hostname.

=item sysid

(Optional) : Assign the disk(s) to this owner given by the sysid.

=item pool

(Optional) : Assign the disk(s) to this pool number.

=item count

(Optional) : Assign ownership to this number of disks.

=item force

(Optional) : Ownership assignment is forced.
Default is 0.

=item partner

(Optional) : Run command in partner mode.

=item maint

(Optional) : Command is being run in maintenance mode.

=back

=item Returns

Returns 1 on success or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub disk_assign(%;) {
    my (%opts) = validate (
        @_,
        {
            conn    => { type => OBJECT },
            disk    => { type => SCALAR|ARRAYREF, optional => 1 },
            owner   => { type => SCALAR, optional => 1 },
            sysid   => { type => SCALAR, optional => 1 },
            pool    => { type => SCALAR, optional => 1,
                                            regex => qr/^\d+$/ },
            count   => { type => SCALAR, optional => 1,
                                            regex => qr/^\d+$/ },
            force   => { type => SCALAR,  default => 0,
                                            regex => qr/^[01]$/ },
            partner => { type => SCALAR, optional => 1,
                                            regex => qr/^[01]$/ },
            maint   => { type => SCALAR, optional => 1,
                                            regex => qr/^[01]$/ },
            shelf   => {type => SCALAR, optional => 1 },
                                            
            type    => {type => SCALAR, optional => 1 },
        }
    );
    my $out;
    my $disks = '';
    my $options = '';

    # Process the disk option
    if (ref $opts{disk}) {
        $disks .= join(' ', @{$opts{disk}});
    } elsif (defined $opts{disk}) {
        $disks .= $opts{disk};
    }

    # Process options
    $options .= " -o $opts{owner}" if defined $opts{owner};
    $options .= " -s $opts{sysid}" if defined $opts{sysid};
    $options .= " -p $opts{pool}"  if defined $opts{pool};
    $options .= " -n $opts{count}" if defined $opts{count};
    $options .= " -shelf $opts{shelf}" if defined $opts{shelf};
    $options .= " -T $opts{type}" if defined $opts{type};
    $options .= " -f" if $opts{force};

    # Send command to filer
    $out = _say($opts{conn}, "disk assign $disks $options", $opts{maint}, $opts{partner});

    # Remove all "NOTE:" messages
    _check_for_note($out);

    # Remove information messages from output
    $out =~ s/^.*run 'disk assign' with -o option to specify name.*\r*\n*//img;
    $out =~ s/^.*Please type disk show.*\r*\n*//img;
    $out =~ s/^.*Cannot do auto-rescan in maint mode.*\r*\n*//img;
    $out =~ s/^.*Disk assigned but unable to obtain owner name.*\r*\n*//img;

    # Make sure not to croak on an empty line
    $out =~ s/^\s+$//;

    croak $out if $out;

    return 1;
}

#---------------------------------------------------------------------

=head2 disk_reassign

    $result = disk_reassign(conn => $conn
                            [, old_owner => $old_owner ]
                            [, old_sysid => $old_sysid ]
                            [, new_owner => $new_owner ]
                            [, new_sysid => $new_sysid ]
                            [, maint => 0|1 ]);

This function provides the interface of the disk reassign command. It
initiates the reassign ownership from the previous system to the
new system. The routine supports both maint and normal mode.

The following named arguments are accepted:

=over

=item Arguments

=over

=item conn

(Required) : Connection object to a filer.
Must use a persistent connection.

=item old_owner

(Optional) : Name of the current owner.

=item old_sysid

(Optional) : Sysid of the current owner.

=item new_owner

(Optional) : Name of the new owner.

=item new_sysid

(Optional) : Sysid of the new owner.

=item maint

(Optional) : Command is being run in maintenance mode.

=back

=item Returns

Returns 1 on success or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub disk_reassign(%;) {
    my (%opts) = validate (
        @_,
        {
            conn      => { type => OBJECT },
            old_owner => { type => SCALAR, optional => 1 },
            old_sysid => { type => SCALAR, optional => 1 },
            new_owner => { type => SCALAR, optional => 1 },
            new_sysid => { type => SCALAR, optional => 1 },
            maint     => { type => SCALAR, optional => 1,
                                              regex => qr/^[01]$/ },
        }
    );
    my ($cmd, $priv, $prompt, $out);
    my $options = '';
    my $cont_prompt = 'Would you like to continue \(y/n\)\?';
    my $maint_prompt = 'Are you sure partner is NOT in takeover mode \(y/n\)\?';

    # Check for persistent connection
    croak "disk_reassign: Must use a persistent connection"
      if $opts{conn}->is_persay;

    # Process options
    $options .= " -o $opts{old_owner}" if defined $opts{old_owner};
    $options .= " -s $opts{old_sysid}" if defined $opts{old_sysid};
    $options .= " -n $opts{new_owner}" if defined $opts{new_owner};
    $options .= " -d $opts{new_sysid}" if defined $opts{new_sysid};

    # Send command to filer
    $cmd = "disk reassign $options";
    $priv = "priv set -q advanced;";
    if (defined $opts{maint} ? $opts{maint} : $maint_mode) {
        ($out, $prompt) = (Tharn::say($opts{conn}, $cmd, '-swallow',
                                      '-prompt', $maint_prompt))[2,3];
        $out =~ s/\r*//g;
        croak $out if ($prompt !~ /$maint_prompt/);
        ($out, $prompt) = (Tharn::say($opts{conn}, "y",
                                      '-prompt', $cont_prompt))[2,3];
        $out =~ s/\r*//g;
        croak $out if ($prompt !~ /$cont_prompt/);
        $out = (Tharn::say($opts{conn}, "y",
                          '-prompt', $maint_mode_prompt))[2];
    } else {
        ($out, $prompt) = (Tharn::say($opts{conn}, "$priv $cmd", '-swallow',
                                      '-prompt', $cont_prompt))[2,3];
        $out =~ s/\r*//g;
        croak $out if ($prompt !~ /$cont_prompt/);
        $out = (Tharn::say($opts{conn}, "y"))[2];
    }

    # Remove echo of command from output
    $out =~ s/.*^\s*y\r*\n?//ms;

    # Remove information messages
    $out =~ s/^.*ownership will be updated.*\r*\n*//img;

    # Remove syslog messages from output
    _chop_messages($out);

    # Remove all "NOTE:" messages
    _check_for_note($out);

    # Make sure not to croak on an empty line
    $out =~ s/^\s+$//;

    croak $out if $out;

    return 1;
}

#---------------------------------------------------------------------

=head2 disk_upgrade_ownership

    $result = disk_upgrade_ownership(conn => $conn);

This function provides the interface of the disk upgrade ownership command. It
upgrades disks ownership to the sanown software based ownership. This routine
supports only maint mode.

The following named arguments are accepted:

=over

=item Arguments

=over

=item conn

(Required) : Connection object to a filer.

=back

=item Returns

Returns 1 on success or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub disk_upgrade_ownership(%;) {
    my (%opts) = validate (
        @_,
        {
            conn    => { type => OBJECT },
        }
    );
    my ($cmd, $out);

    # Send command to filer
    transit_to($opts{conn}->hostp(), 'MAINT')
        or croak("This command only works in maintenance mode:$@");

    $cmd = "disk upgrade_ownership";
    $out = _say($opts{conn}, $cmd, 1, 0, (-prompt, 'continue'));
    $out = _say($opts{conn}, 'y',  1, 0);

    # Remove information messages from output
    $out =~ s/^.*disk upgrade_ownership.*\r*\n*//mg;
    $out =~ s/^.*changing ownership for disk.*\r*\n*//mg;
    $out =~ s/^.*being changed to use software based disk ownership.*\r*\n*//mg;
    $out =~ s/^\s+$//;

    croak $out if $out;

    eval {
        disk_show(conn => $opts{conn}, maint => 1);
    };
    croak "Sanown error after ownership upgrade: $@" if $@;

    return 1;
}

#---------------------------------------------------------------------

=head2 disk_remove_ownership

    $result = disk_remove_ownership(conn => $conn
                                    [, disk => $disk|\@disks ]
                                    [, all => 0|1 ]
                                    [, sysid => $sysid ]
                                    [, partner => 0|1 ]
                                    [, maint => 0|1 ]
                                    [, options => $options ]);

The routine provides the interface of "disk remove_ownership" CLI.
The command is available in both maint and normal mode.

The routine takes the following parameters.

=over

=item Arguments

=over

=item conn

(Required) : Connection object to a filer.

=item disk

(Optional) : Disk or list of disks to remove ownership from.

=item all

(Optional) : Remove ownership from the all the disks.
This option is only allowed in maintenance mode.

=item sysid

(Optional) : Remove ownership from the disk(s) to this owner given by the sysid.
This option is only allowed in maintenance mode.

=item partner

(Optional) : Run command in partner mode.

=item maint

(Optional) : Command is being run in maintenance mode.

=item options

(Optional) : Options to pass to the 'disk remove_ownership' command as is,
it will override any other options.

=back

=item Returns

Returns 1 on success or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub disk_remove_ownership (%;) {
    my (%opts) = validate (
        @_,
        {
            conn    => { type => OBJECT },
            disk    => { type => SCALAR|ARRAYREF, optional => 1 },
            all     => { type => SCALAR,  default => 0,
                                            regex => qr/^[01]$/ },
            sysid   => { type => SCALAR, optional => 1 },
            partner => { type => SCALAR, optional => 1,
                                            regex => qr/^[01]$/ },
            maint   => { type => SCALAR, optional => 1,
                                            regex => qr/^[01]$/ },
            options => { type => SCALAR, optional => 1 },
        }
    );
    my ($cmd, %ret, $disk, @list, @keys, %hash, $key, $out);
    my $options = '';
    my $disks = '';

    # Process the disk option
    if (ref $opts{disk}) {
        $disks .= join(' ', @{$opts{disk}});
    } elsif (defined $opts{disk}) {
        $disks .= $opts{disk};
    }

    # Process options
    if ($opts{options}) {
        # Override all other options
        $options = $opts{options};
    } else {
        # Set any option specified
        $options .= ' all' if $opts{all};
        $options .= " -s $opts{sysid}" if defined $opts{sysid};
        $options .= " -f";
    }

    # Send command to filer
    if ($opts{maint}) {
        $cmd = "disk remove_ownership $disks $options";
    } elsif ($opts{partner} || $partner_mode) {
        $cmd = "priv set -q advanced;partner disk remove_ownership ".
               "$disks $options";
    } else {
        $cmd = "priv set -q advanced;disk remove_ownership $disks $options";
    }
    $out = _say($opts{conn}, $cmd, $opts{maint}, $opts{partner});

    # Remove all "NOTE:" messages
    _check_for_note($out);

    # Remove syslog messages from output
    $out =~ s/^.*ownership\s+(?:information\s+)?removed.*\r*\n*//img;

    # Make sure not to croak on an empty line
    $out =~ s/^\s+$//;

    croak $out if $out;

    return 1;
}

#---------------------------------------------------------------------

=head2 partner_mode

    partner_mode([ enable => 0|1 ]);

This function enables partner mode for subsequent calls to any disk
command function.

The following named arguments are accepted:

=over

=item Arguments

=over

=item enable

(Optional) : Set to 1 to enable partner mode or 0 to disable it.
Default is 1.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub partner_mode(%;) {
    my (%opts) = validate (
        @_,
        {
            enable => { type => SCALAR, default => 1, regex => qr/^[01]$/ },
        }
    );
    $partner_mode = $opts{enable};
    return;
}

#---------------------------------------------------------------------

=head2 maint_mode

    maint_mode([ enable => 0|1 ]);

This function enables maintenance mode for subsequent calls to any disk
command function.

The following named arguments are accepted:

=over

=item Arguments

=over

=item enable

(Optional) : Set to 1 to enable maintenance mode or 0 to disable it.
Default is 1.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub maint_mode(%;) {
    my (%opts) = validate (
        @_,
        {
            enable => { type => SCALAR, default => 1, regex => qr/^[01]$/ },
        }
    );
    $maint_mode = $opts{enable};
    return;
}

#---------------------------------------------------------------------

=head2 get_notes

    $notes = get_notes();

This function returns the note (or info) messages from the last disk
command function.

=over

=item Returns

The note messages from the last disk command function.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_notes() {
    return defined $disk_cmd_note ? $disk_cmd_note : '';
}

#---------------------------------------------------------------------
# Internal function
# Pre-process output of command to remove all '\r' characters
# and remove all "NOTE:" messages while saving them in the package
# variable for later use
#---------------------------------------------------------------------
sub _check_for_note(\$;) {
    my $out = shift;
    my @notes;

    # Remove '\r'
    $$out =~ s/\r+//g;

    # Remove all "NOTE:" messages saving them in the package variable
    while ($$out =~ s/^(NOTE:.*)\n?//m) {
        push @notes, $1;
    }

    # Save notes into package variable
    $disk_cmd_note = join("\n", @notes);

    return;
}

#---------------------------------------------------------------------
# Internal function
# This is a wrapper to either sayEOF or say depending on the
# third parameter 'maint' (maintenance mode). If the maint parameter
# is true it uses Tharn say because sayEOF sends a 'echo' command
# after the command and in maintenance mode the 'echo' command is
# not supported
#---------------------------------------------------------------------
sub _say($$$$;@) {
    my ($conn, $cmd, $maint, $partner, @options) = @_;
    my ($out, $qcmd);

    # Prefix partner to the command
    $cmd = "partner $cmd" if (defined $partner ? $partner : $partner_mode);

    if (defined $maint ? $maint : $maint_mode) {
        # Maintenance mode, make sure to use the maintenance prompt
        # and swallow all previous output before sending the command
        unshift @options, (
            '-prompt' => $maint_mode_prompt,
            '-swallow',
        );
        $out = (Tharn::say($conn, $cmd, @options))[2];
    } else {
        # Normal mode
        # Make sure to swallow all previous output before sending the command
        Tharn::say($conn, '', '-nocr', '-swallow', '-blind');
        $out = sayEOF($conn, $cmd, @options);
    }

    # Remove echo of command from output
    $qcmd = quotemeta($cmd);
    $out =~ s/.*^\s*$qcmd\r*\n*//ms;

    # Chop console messages
    _chop_messages($out);

    return $out;
}

#---------------------------------------------------------------------
# Internal function
# This function chops all console messages having the following format:
#   Day Mon day# hh:mm:ss TMZ [any message]: description
#---------------------------------------------------------------------
sub _chop_messages($;) {
    # Remove syslog messages
    $_[0] =~ s{
               ^(?:Mon|Tue|Wed|Thu|Fri|Sat|Sun)\s+
                (?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s+
                \d+\s+
                \d+:\d+:\d+\s+
                \S{3}\s+
                \[[^\]]+\]:\s
                .*\r*\n*
              }{}mgx;

    # Remove DBG messages
    $_[0] =~ s/^DBG:.*\r*\n*//mg;

    return;
}

1;

