# this is basically -*- perl -*-
# $Id: Utils.pm,v 1.48 2008/09/05 20:48:58 emartinez Exp $
#
## @summary	Utils - library routines to support CFO test cases
## @author	dl-cf-qa@netapp.com
## @status	private
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Utils - library routines to support CFO test cases

=head1 SYNOPSIS

  use cluster::Utils;

  # Disable disk inventory on the filer
  cf_disable_disk_inventory(fc => $fc);

  # Enable disk inventory on the filer
  cf_enable_disk_inventory(fc => $fc);

  # Get the cluster fmdisk inventory for a filer
  my $href = cf_fmdisk_inventory(fc => $fc);

  # Sample output:
  $href = {
            'partner_pending' => {
                                   'last_update' => '0',
                                   'ndisks' => '0',
                                   'seq' => '0',
                                   'gen' => '0'
                                 },
            'force_send' => '0',
            'current_time' => '1219273346',
            'inventory_version' => 'ok',
            'last_send_status' => '0',
            'missing' => [
                           {
                             'sim0' => '0'
                           },
                           {
                             'sim1' => '0'
                           }
                         ],
            'problem_time' => '0',
            'local' => {
                         'last_update' => '0',
                         'ndisks' => '0',
                         'seq' => '0',
                         'gen' => '0'
                       },
            'fmdisk_inventory' => 'enabled',
            'autosupport_time' => '0',
            'local_pending' => {
                                 'disks' => [
                                              {
                                                'v1.27' => {
                                                             'guid' => '4E4...',
                                                             'idx' => '0',
                                                             'host' => '0'
                                                           }
                                              },
                                              {
                                                'v0.29' => {
                                                             'guid' => '4E5...',
                                                             'idx' => '1',
                                                             'host' => '0'
                                                           }
                                              },
                                              {
                                                'v0.19' => {
                                                             'guid' => '4E6...',
                                                             'idx' => '2',
                                                             'host' => '0'
                                                           }
                                              },
                                            ],
                                 'last_update' => '0',
                                 'ndisks' => '3',
                                 'seq' => '0',
                                 'gen' => '0'
                               },
            'partner' => {
                           'last_update' => '0',
                           'ndisks' => '0',
                           'seq' => '0',
                           'gen' => '0'
                         }
          };

  my $state = cf_monitor(fc => $fc);

  if ($state =~ /UP/i) {
      my $href = cf_mailbox(fc => $fc);
  }

  # Sample output:
  $href = {
            'mailbox' => {
                           'local' => [
                                        '0b.21',
                                        '0b.18'
                                      ],
                           'partner' => [
                                          '0b.20',
                                          '0b.19'
                                        ]
                         }
          };

  # Special note when using CFO setup and cleanup:
  # The 'recycle-trad' option will take a snapshot of the trad
  # volume on the filer before hammer starts. Setup will create
  # the trad volume on the filer if it does not already exist. The
  # volume snapshot is then restored at cleanup so that the next
  # thpl can re-use the pristine trad volume. This will save time
  # when spare disks are limited. The trad volume should start out
  # empty for the best results.

  # CFO setup
  cfo_setup(filer => $FILER, setup_h  => \%set_h);

  # CFO cleanup
  cfo_cleanup(cleanup => $CLEANUP, setup_h => \%set_h);

  # Check setup for FM tests and their requirements
  check_fm_config(
      fc  => $fc,
  );

  # Check setup for FM tests and their requirements
  my $log = LogReader->new($fc);
  check_msgs(
      log_obj     => $log,
      log_timeout => 120,
      msg         => 'This is the expected message in /etc/messages on '.
                     'the filer',
  );

  # Simulate a disk pull on a simulator, get a list of pulled disks,
  # and push them back in on the local filer.

  use Storage::Disk qw(disk_simpull);
  disk_simpull(host => $fh, disk => 'v0.29');

  $href = get_simpush_disks(fc => $fc);

  foreach $disk (@{$hrefB->{local}}) {
      disk_simpush_local_disk(fc => $fcB, disk => $disk);
  }

  # Filer 'cf takeover' command for negative testing. It does not bring ensure
  # filer states. It only issues the command on the filer and croaks on an
  # unexpected output.
  cfo_takeover(fc => $fc);

  # Filer 'cf giveback' command for negative testing. It does not bring ensure
  # filer states. It only issues the command on the filer and croaks on an
  # unexpected output.
  cfo_giveback(fc => $fc);

  # This function returns a formated string.
  my $output = fmt_exp_act_results(
      [prefix    => 'This is a prefix to be printed before expected results',]
      actual     => 'Reported incorrect results',
      expected   => 'Should have seen these results',
      [postfix   => 'This is a postfix to be printed after actual results',]
  );

  # License cluster function that supports 7G and BR releases. Croaks on an
  # error.
  license_cluster(fc => $fc);

  # Function to test new style commands in either maintenance mode or partner
  # mode.
  maint_mode([ enable => 0|1 ]);
  partner_mode([ enable => 0|1 ]);

  # Setup an aggregate on a filer
  setup_aggr(
      fc        => $fc,
      aggr_name => 'my_aggr',
      aggr_size => 2,
      raidtype  => 'raid4',
      timeout   => 7200,
      force     => 1,
  );

  # Setup a flex volume on a filer
  setup_vol(
      fc        => $conn,
      aggr_name => 'my_aggr',
      vol_name  => 'my_vol,
      flex_size => '30m',
  );

  # Set the following default cluster options on one filer
  # cf.giveback.auto.cifs.terminate.minutes         => 5,
  # cf.giveback.auto.enable                         => off,
  # cf.giveback.auto.terminate.bigjobs              => on,
  # cf.giveback.check.partner                       => on,
  # cf.quickloop.enable                             => false,
  # cf.takeover.change_fsid                         => on,
  # cf.takeover.detection.seconds                   => 15,
  # cf.takeover.on_disk_shelf_miscompare            => off,
  # cf.takeover.on_failure                          => on,
  # cf.takeover.on_network_interface_failure        => off,
  # cf.takeover.on_network_interface_failure.policy => all_nics,
  # cf.takeover.on_panic                            => on,
  # cf.takeover.on_short_uptime                     => on,

  set_default_options(fc => $fc);

  # Get the system disk information using 'sysconfig -r'
  # This routine is used to get disk information for the system.
  # It returns a hash reference having as keys the following:
  # volumes    => Hash reference of trad volumes
  # aggregates => Hash reference of aggregates
  # spares     => Hash reference of spare disks
  # partner    => Hash reference of partner disks
  # broken     => Hash reference of broken disks

  my $disk_info = system_disk_info(fc => $fc);

  # This routine returns a hash for the system id and the partner id info
  my $sysid_info = get_sysid_info(fc => $fc);
  Sample output:
  $sysid_info = {
        System => {
            sysid => 1234567,
            name  => sim0,
        },
        Partner => {
            sysid => 1234568,
            name  => sim1,
        },
    };

  # Get root volume for a filer
  my $vol_name = get_root_vol(fc => $fc);

  # Get root volume disks and their information
  my $root_disk_info = get_root_disks(fc => $fc);

  # Get disk information for the specified volume
  my $vol_disk_info = get_vol_disks(fc  => $fc,
                                    vol => $vol);

  # Zero the disks before install the filer
  zero_disks_install(fc => $fc);

=head1 DESCRIPTION

General purpose library routines to support CFO test cases.

=head1 STATUS

  Private

=head1 AUTHOR

  dl-cf-qa@netapp.com
  $Id: Utils.pm,v 1.48 2008/09/05 20:48:58 emartinez Exp $
  Last revised $Date: 2008/09/05 20:48:58 $

=head1 COPYRIGHT

  Copyright 2008, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package cluster::Utils;

use strict;
use warnings;

use Tharn;
use LogReader;
use Carp qw(croak);
use base qw(Exporter);
use SayWrap qw(sayEOF);
use boot qw(reboot transit_to);
use Network::Utils qw(run_ping);
use nfs::Utils qw(setup_cluster);
use cluster::Disk qw(disk_assign);
use install qw(newfs basic_setup);
use nic::Utils qw(filer_netstat_rn);
use Network::Dns qw(dns_get_options);
use cluster qw(cf_enable get_cf_state);
use Network::Utils qw(get_ifconfig_info);
use SPARSE::Util::Logger qw(logStep logWarn);
use Hammer qw(nfs_hammer_start nfs_hammer_stop);
use Filer::Version qw(:releases get_wafl_version);
use Admin::License qw(is_licensed license unlicense);
use FS::Util::Utilities qw(get_firmware_env set_firmware_env);
use Volume qw(vol_create vol_create_flex vol_exists vol_purge vol_status);
use SnapCmds qw(snapshot_create snapshot_exists snaprestore_volume);
use Params::Validate qw(ARRAYREF HASHREF OBJECT SCALAR UNDEF validate);
use Aggregate qw(aggr_create aggr_exists aggr_online aggr_status aggr_purge);
use filer qw(filer_options_get filer_options_set filer_sysstat_get
             filer_vol_snap_delete_all filer_cluster_license
             filer_cluster_is_licensed filer_sysconfig_a_check);

our @EXPORT_OK = qw(
    cf_disable_disk_inventory
    cf_enable_disk_inventory
    cf_fmdisk_inventory
    cf_mailbox
    cf_monitor
    cfo_setup
    cfo_cleanup
    cfo_giveback
    cfo_takeover
    check_fm_config
    check_msgs
    disk_simpush_local_disk
    fmt_exp_act_results
    get_notes
    get_simpush_disks
    license_cluster
    maint_mode
    partner_mode
    sanown_ownership_cleanup
    setup_aggr
    setup_vol
    set_default_options
    system_disk_info
    get_sysid_info
    get_root_vol
    get_vol_disks
    get_root_disks
    zero_disks_install
    get_disk_of_same_type
    get_disk_from_same_carrier
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '1.0';

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#---------------------------------------------------------------------
# Local variables
#---------------------------------------------------------------------
# Default supported systems or failover monitor tests:
my %supported_systems = (
    FAS20XX => {
        bios_version => '3.0.0',
    },
    FAS30XX => {
        bios_version => '2.1.0',
    },
    FAS31XX => {
        bios_version => '4.0.0',
    },
    FAS60XX => {
        bios_version => '1.5.0',
    },
);

# Default cf options:
my %default_options = (
    'cf.giveback.auto.cifs.terminate.minutes'         => 5,
    'cf.giveback.auto.enable'                         => 'off',
    'cf.giveback.auto.terminate.bigjobs'              => 'on',
    'cf.giveback.check.partner'                       => 'on',
    'cf.quickloop.enable'                             => 'false',
    'cf.takeover.change_fsid'                         => 'on',
    'cf.takeover.detection.seconds'                   => 15,
    'cf.takeover.on_disk_shelf_miscompare'            => 'off',
    'cf.takeover.on_failure'                          => 'on',
    'cf.takeover.on_network_interface_failure'        => 'off',
    'cf.takeover.on_network_interface_failure.policy' => 'all_nics',
    'cf.takeover.on_panic'                            => 'on',
    'cf.takeover.on_short_uptime'                     => 'on',
);

# Firmware configuration variables
my %firmware_env = (
    'gdb-on-panic?' => 'false',
    'floppy-boot?'  => 'false',
    'AUTOBOOT'      => 'true',
);

# Wait time to sleep in seconds
my $wait_time = 5;

# Ping wait time in seconds before timing out
my $ping_timeout = 300;

my $disk_cmd_note;
my $partner_mode = 0;
my $maint_mode = 0;
my $maint_mode_prompt = '\\*>';

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub cf_disable_disk_inventory(%;);
sub cf_enable_disk_inventory(%;);
sub cf_fmdisk_inventory(%;);
sub cf_mailbox(%;);
sub cf_monitor(%;);
sub cfo_setup(%;);
sub cfo_setup_a_filer(%;);
sub cfo_cleanup(%;);
sub cfo_cleanup_a_filer(%;);
sub cfo_giveback(%;);
sub cfo_takeover(%;);
sub check_fm_config(%;);
sub check_msgs(%;);
sub disk_simpush_local_disk(%;);
sub fmt_exp_act_results(%;);
sub get_notes();
sub get_simpush_disks(%;);
sub license_cluster(%;);
sub maint_mode(%;);
sub partner_mode(%;);
sub sanown_ownership_cleanup(%;);
sub set_default_options(%;);
sub setup_aggr(%;);
sub setup_vol(%;);
sub system_disk_info(%;);
sub get_sysid_info(%;);
sub get_root_vol(%;);
sub get_vol_disks(%;);
sub get_root_disks(%;);
sub zero_disks_install(%;);
sub _parse_disk_info($;);
sub _check_for_note(\$;);
sub _say($$$$;@);
sub _chop_messages($;);
sub _search_disk_serial($$;);

#---------------------------------------------------------------------

=head2 cf_disable_disk_inventory

    cf_disable_disk_inventory(fc => $fc);

This routine issues the 'cf disable diskInventory' and searches for the message
'cluster disk inventory disabled' in /etc/messages as confirmation. The filer
command does not return any output but messages are logged.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cf_disable_disk_inventory(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );

    my ($log, $msg, $out);
    $msg = 'cluster\s+disk\s+inventory\s+disabled';

    $log = LogReader->new($opts{fc});
    $log->start(['messages']);

    $out = sayEOF($opts{fc}, 'priv set -q test;cf disable diskInventory');

    eval {
        check_msgs(
            log_obj     => $log,
            log_timeout => 120,
            msg         => $msg,
        );
    };
    croak ("Could not find message in /etc/messages confirming ".
           "cluster disk inventory disabled") if $@;

    return;
}

#---------------------------------------------------------------------

=head2 cf_enable_disk_inventory

    cf_enable_disk_inventory(fc => $fc);

This routine issues the 'cf enable diskInventory' and searches for the message
'cluster disk inventory enabled' in /etc/messages as confirmation. The filer
command does not return any output but messages are logged.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cf_enable_disk_inventory(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );

    my ($log, $msg, $out);
    $msg = 'cluster\s+disk\s+inventory\s+enabled';

    $log = LogReader->new($opts{fc});
    $log->start(['messages']);

    $out = sayEOF($opts{fc}, 'priv set -q test;cf enable diskInventory');

    eval {
        check_msgs(
            log_obj     => $log,
            log_timeout => 120,
            msg         => $msg,
        );
    };
    croak ("Could not find message in /etc/messages confirming ".
           "cluster disk inventory enabled") if $@;

    return;
}

#---------------------------------------------------------------------

=head2 cf_fmdisk_inventory

    my $href = cf_fmdisk_inventory(fc => $fc);

This routine parses the 'cf fmdisk inventory' disk information on a
cluster and returns a hash reference to the various information returned.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

Returns hash reference or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cf_fmdisk_inventory(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );

    my (%hash, $int, $key, $line, @lines, $group, @groups, $out);
    $out = sayEOF($opts{fc}, 'priv set -q advanced;cf fmdisk inventory');

    # Die if cf fmdisk usage error
    if ($out=~ /(cf:)|(Usage:\s+cf\s+fmdisk)|(cf\s+fmdisk:\s+usage\s+error)/i) {
        die $out;
    }

    @lines = split (/\n/i, $out);
    $int = 0;

    while (@lines) {
        while ($line = shift @lines) {
            $groups[$int] .= "$line\n";
        }
        $int++;
    }

    while ($group = shift @groups) {
        @lines = split (/\n/i, $group);
        if ($group =~ /Number\s+of\s+Missing/i) {
            foreach $line (@lines) {
                if ($line =~ /Number\s+of\s+Missing\s+Disks\s+on\s+host\((.*)\)\:\s+(\d+)/i) {
                    push @{$hash{'missing'}} , { $1 => $2};
                }
                if ($line =~ /fmdisk\s+inventory\s+(\w+)/i) {
                    $hash{'fmdisk_inventory'} = $1;
                }
                if ($line =~ /inventory_version\s+(\w+)/i) {
                    $hash{'inventory_version'} = $1;
                }
                if ($line =~ /partner\s+IC\s+version\s+(\d+)/i) {
                    $hash{'partner_ic_version'} = $1;
                }
                if ($line =~ /currentTime\s+(\d+)/i) {
                    $hash{'current_time'} = $1;
                }
                if ($line =~ /ProblemTime\s+(\d+)/i) {
                    $hash{'problem_time'} = $1;
                }
                if ($line =~ /AutosupportTime\s+(\d+)/i) {
                    $hash{'autosupport_time'} = $1;
                }
                if ($line =~ /lastSendStatus\s+(\d+)/i) {
                    $hash{'last_send_status'} = $1;
                }
                if ($line =~ /forceSend\s+(\d+)/i) {
                    $hash{'force_send'} = $1;
                }
            }
        } elsif ($group =~ /^(.*):/i) {
            $key = $1;
            $key =~ s/^\s+//;
            $key =~ s/\s+/_/;
            $key =~ s/\(//;
            $key =~ s/\)//;

            if ($group =~ /gen\s+(\d+)/i) {
                $hash{$key}{'gen'} = $1;
            }
            if ($group =~ /seq\s+(\d+)/i) {
                $hash{$key}{'seq'} = $1;
            }
            if ($group =~ /lastUpdate\s+(\d+)/i) {
                $hash{$key}{'last_update'} = $1;
            }
            if ($group =~ /ndisks\s+(\d+)/i) {
                $hash{$key}{'ndisks'} = $1;
            }
            foreach $line (@lines) {
                if ($line =~ m{\[(\d+)\]\s+(\S+)\s+(\d+)\s+(.*)}img) {
                    push @{$hash{$key}{'disks'}} , { $2 => {idx  => $1,
                                                            host => $3,
                                                            guid => $4,
                                                           }
                                                   };
                }
            }
        }
    }
    return (\%hash);
}

#---------------------------------------------------------------------

=head2 cf_mailbox

    my $href = cf_mailbox(fc => $fc);

This routine parses the 'cf status all' disk information on a
cluster and returns a hash reference to the various information returned.
So far, only the mail box information is returned that is required for
the CFO tests.

  i.e.
    Sample output:
    $href = {
              'mailbox' => {
                             'local' => [
                                          '0b.21',
                                          '0b.18'
                                        ],
                             'partner' => [
                                            '0b.20',
                                            '0b.19'
                                          ]
                           }
            };

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

Returns hash reference or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cf_mailbox(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );

    my (@entries, %hash, $mailbox, @mailbox_info, $out, %results);
    $out = sayEOF(
        $opts{fc},
        "priv set -q test;cf monitor all",
    );

    if ($out =~ /Cluster\s+disabled/i) {
        croak "Clustering must be enabled for this command to work: $out";
    } elsif ($out =~ /not\s+allowed/i) {
        croak $out;
    } else {
        $out =~ s/^\s*\n//;
        $out =~ s/^\s+//mg;
        @entries = split /^(.*):\n/m,$out;
        shift @entries;
        %hash = @entries;

        # We are parsing out the mailbox information so far for the tests.
        # There is more information that is needed to be parsed out but for
        # initial testing this is all we need for the HA Mailbox test.
        @mailbox_info = split /\n/m,$hash{'mailbox disks'};
        foreach $mailbox (@mailbox_info) {
            if ($mailbox =~
                /Disk\s+(\S+)\s+is\s+a\s+(\w+)\s+mailbox\s+disk/i) {
                push @{$results{mailbox}{$2}}, $1;
            }
        }
    }
    return (\%results);
}

#---------------------------------------------------------------------

=head2 cf_monitor

    my $state = cf_monitor(fc => $fc);

This routine parses the 'UP|TAKEOVER|DEAD' information on a filer in a
cluster and returns the parsed information. If there is an error or the state
cannot be parsed, the state is reported as 'UNREACH'

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

Returns a scalar value of 'UP|TAKEOVER|DEAD|UNREACH'

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cf_monitor(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );

    my ($out, $cur_state);
    eval {
        $out = sayEOF(
            $opts{fc},
            "priv set -q test;cf monitor",
        );
    };

    if (!defined($out)) {
        return 'UNREACH';
    }

    if ($out =~ /Cluster disabled/i) {
        croak "Clustering must be enabled for this command to work: $out";
    } else {
        if ($out =~ /(UP|TAKEOVER|DEAD)\s+\d+/) {
            $cur_state = $1;
        } else {
            $cur_state = 'UNREACH';
        }
    }

    return $cur_state;
}

#---------------------------------------------------------------------

=head2 cfo_setup

  cfo_setup(
            filers                => [ $FILERA, $FILERB ],
            filer_conn            => $FILER_CONN,
            client                => $CLIENT,
            cmd_timeout           => $CMDTIMEOUT,
            conn_timeout          => $TIMEOUT,
            setup_h               => \%set_h
            [, aggr_name          => $FILER_AGGR ]
            [, aggr_size          => $AGGR_SIZE ]
            [, aggr_timeout       => $AGGR_TIMEOUT ]
            [, force_aggr_create  => 0|1 ]
            [, hammer_bin         => $HAMMER_BIN ]
            [, hammer_panic       => $PANIC_ON_CORRUPTION ]
            [, hammer_time        => -1 ]
            [, nfs_wait           => $NFS_WAIT ]
            [, ping_filer         => $PING_FILER ]
            [, raid_type          => $RAID_TYPE ]
            [, recycle_trad       => $RECYCLE_TRAD ]
            [, start_hammer       => 0|1 ]
            [, trad_size          => $TRAD_SIZE ]
            [, vol_prefix         => $VOL_PREFIX ]
            [, install_ontap      => $FRESH_INSTALL ]
            [, install_timeout    => $INSTALL_TIMEOUT ]
            [, install_zero_disks => $ZERO_DISKS ]
            [, ontapdir           => $ONTAPDIR ]
            [, ontap_releases_dir => $FILER_ONTAP_RELEASES_DIR ]
            [, set_default_ops    => $SET_DEFAULT_OPS ]
            [, set_firmware_ops   => $SET_FIRMWARE_OPS ]
  );

Perform generic setup for CF test cases. This includes creating a trad and a
flex volume. Optionally the trad can be recycled and the default value of cf
options can be set.

The following named arguments are accepted:

=over

=item Arguments

=over

=item filers

(Required) : List of hostspec of cluster filers.

=item filer_conn

(Required) : Filer connection type.

=item client

(Required) : Hostspec of Client.

=item cmd_timeout

(Required) : NATE say timeout for issued commands.

=item conn_timeout

(Required) : NATE connection setup timeout.

=item aggr_name

(Optional) : Name of aggregate to use when creating a volume. Required if
option hammer_start is set.

=item aggr_size

(Optional) : Size of aggregate. Required if option hammer_start is set.

=item aggr_timeout

(Optional) : Timeout of filer aggr create. Required if option hammer_start
is set.

=item flex_size

(Optional) : The size of the volume to create. Required if option
hammer_start is set.

=item force_aggr_create

(Optional) : Force aggregate creation flag: 0 or 1.

=item hammer_bin

(Optional) : Path to hammer executable on 'client' argument. Required if option
hammer_start is set.

=item hammer_time

(Optional) : Amount of time to run hammer. Required if option hammer_start
is set.

=item nfs_wait

(Optional) : Number of seconds to wait for NFS traffic to start. Required if
option hammer_start is set.

=item ping_filer

(Optional) : Ping from the client to the filer: 0 or 1.
Default is 0.

=item hammer_panic

(Optional) : Hammer panic on corruption flag: 'yes' or 'no'.
Default is 'yes'.

=item raid_type

(Optional) : Raid type to create: 'raid_dp or 'raid4'. Required if option
hammer_start is set.

=item recycle_trad

(Optional) : Recycle trad volumes flag: 0 or 1. This routine get a snapshot of
the trad volume to be restored during cleanup. The trad volume should be empty
so that it can be reused for other tests.
Default is 0.

=item start_hammer

(Optional) : Start hammer flag: 0 or 1.
Default is 0.

=item trad_size

(Optional) : The number of disks to use for the trad volume. Required if
option hammer_start is set.

=item vol_prefix

(Optional) : Volume name prefix. Required if option hammer_start is set.

=item install_ontap

(Optional) : Fresh install of the OS flag: 1 or 0.
Default is 0.

=item install_timeout

(Optional) : Timeout for a fresh install of ONTAP.
Default is 3600 (1 hr).

=item install_zero_disks

(Optional) : Zero disks at 'fresh install' flag: 1 or 0.
Default is 0.

=item ontapdir

(Optional) : The path of the top of an ONTAP directory from which to install.
Required if option install_tap is set.

=item ontap_releases_dir

(Optional) : The name of the directory to use under ontapdir argument.
Required if option install_ontap is set.

=item set_default_ops

(Optional) : Set default value of cf options flag: 0 or 1.
Defaults to 0.

=item set_firmware_ops

(Optional) : Set default value of firmware options flag: 0 or 1.
Defaults to 0.

=back

=item Returns

Nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cfo_setup(%;) {
    my (%opts) = validate (
        @_,
        {
            filers             => { type => ARRAYREF },
            filer_conn         => { type => SCALAR },
            client             => { type => SCALAR },
            cmd_timeout        => { type => SCALAR, regex => qr/^\d+$/ },
            conn_timeout       => { type => SCALAR, regex => qr/^\d+$/ },
            setup_h            => { type => HASHREF },
            aggr_name          => { type => SCALAR|UNDEF, optional => 1 },
            aggr_size          => { type => SCALAR|UNDEF, optional => 1 },
            aggr_timeout       => { type => SCALAR|UNDEF, optional => 1 },
            flex_size          => { type => SCALAR|UNDEF, optional => 1 },
            force_aggr_create  => { type => SCALAR, regex => qr/^[01]$/,
                                      default => 1 },
            hammer_bin         => { type => SCALAR|UNDEF, optional => 1 },
            hammer_panic       => { type => SCALAR, regex => qr/^(yes|no)$/,
                                      default => 'yes' },
            hammer_time        => { type => SCALAR|UNDEF, optional => 1 },
            nfs_wait           => { type => SCALAR|UNDEF, optional => 1 },
            ping_filer         => { type => SCALAR, regex => qr/^[01]$/,
                                      default => 0 },
            raid_type          => { type => SCALAR|UNDEF, optional => 1 },
            recycle_trad       => { type => SCALAR, regex => qr/^[01]$/,
                                      default => 0 },
            start_hammer       => { type => SCALAR, regex => qr/^[01]$/,
                                      default => 0 },
            trad_size          => { type => SCALAR|UNDEF, optional => 1 },
            vol_prefix         => { type => SCALAR|UNDEF, optional => 1 },
            install_ontap      => { type => SCALAR, regex => qr/^[01]$/,
                                      default => 0 },
            install_timeout    => { type => SCALAR, regex => qr/^\d+$/,
                                      default => 3600 },
            install_zero_disks => { type => SCALAR, regex => qr/^[01]$/,
                                      default => 0, },
            ontapdir           => { type => SCALAR|UNDEF, optional => 1 },
            ontap_releases_dir => { type => SCALAR|UNDEF, optional => 1 },
            set_default_ops    => { type => SCALAR, regex => qr/^[01]$/,
                                      default => 0 },
            set_firmware_ops   => { type => SCALAR, regex => qr/^[01]$/,
                                      default => 0 },
        }
    );
    my ($end_time, $fc, $fh, $filer, $hostA, $hostB, $id, $key, $msgs,
        $nfs_hamid, $ping, %result, $res_obj, $set_h, $sobj, @sobjs,
        @vols, $vol);

    my $filera= $opts{filers}[0];
    my $filerb= $opts{filers}[1];
    my $cnt = 0;
    my $setup_h = $opts{setup_h};

    # 2 filers are required
    croak("cfo_setup: Two filers are required for setup")
      if (scalar @{$opts{filers}} != 2);

    # Validate parameter dependencies

    # Check option start_hammer
    if ($opts{start_hammer} == 1) {
        if (!defined($opts{aggr_name}) || !defined($opts{aggr_size}) ||
            !defined($opts{aggr_timeout}) || !defined($opts{client}) ||
            !defined($opts{flex_size}) || !defined($opts{hammer_bin}) ||
            !defined($opts{hammer_time}) || !defined($opts{nfs_wait}) ||
            !defined($opts{raid_type}) || !defined($opts{trad_size}) ||
            !defined($opts{vol_prefix})) {
            croak ('cfo_setup: When using option start_hammer the following '.
                   'options must be defined: aggr_name, aggr_size, '.
                   'aggr_timeout, client, flex_size, hammer_bin, '.
                   'hammer_time, nfs_wait, raid_type, trad_size, and '.
                   'vol_prefix');
        }
    }

    # Check option install_ontap
    if ($opts{install_ontap} == 1) {
        if (!defined($opts{ontapdir}) || !defined($opts{ontap_releases_dir})) {
            croak ('cfo_setup: When using option install_ontap the '.
                   'following options must be defined: ontapdir and '.
                   'ontap_releases_dir');
        }
    }

    # Check option install_zero_disks
    if ($opts{install_zero_disks} == 1) {
        if (!defined($opts{ontapdir}) || !defined($opts{ontap_releases_dir}) ||
            !$opts{install_ontap}) {
            croak ('cfo_setup: When using option install_zero_disks the '.
                   'following options must be defined: install_ontap, '.
                   'ontapdir and ontap_releases_dir');
        }
    }

    # Zero the disks before installing ontap
    if ($opts{install_zero_disks} == 1) {
        logStep("cfo_setup: Zero out the filers' disks");
        $id = 0;
        foreach $filer (@{$opts{filers}}) {
            push @sobjs, subtest(
                sub {
                    zero_disks_install(
                        filer        => $filer,
                        filer_conn   => $opts{filer_conn},
                        cmd_timeout  => $opts{cmd_timeout},
                        conn_timeout => $opts{install_timeout},
                    );
                },
                '-bg',
                '-onexit', sub {},
                '-runid', "zero_disks_$id",
            );
            $id++;
        }

        # Wait for all zero subtests to finish before croaking
        Subtest::wait_finish(subtests => \@sobjs);

        # Now that each subtest has finished, check the
        foreach $sobj (@sobjs) {
            # Subtest results processing
            $res_obj = NATE::Result::worst($sobj->own_results);
            if ($res_obj && $res_obj->severity > 2) {
                # Get error messages (severity > 2)
                $msgs = join("\n", $res_obj->msgs);
                croak ("cfo_setup: $msgs") if ($msgs);
            }
        }
    }

    # Fresh install ONTAP build
    # $ONTAPDIR defaults to the value of the ONTAPDIR field in the host
    # record for filer
    if ($opts{install_ontap}) {

        foreach $filer (@{$opts{filers}}) {
            # Connect to each filer and license NFS
            host($filer);
            $fc = connect($opts{filer_conn}, -timeout, $opts{conn_timeout});
            $fc->timeout($opts{cmd_timeout});

            logStep("cfo_setup: License NFS for filer ($filer) install");
            # License snap restore if needed
            if (!is_licensed($fc, 'nfs')) {
                license($fc, 'nfs');
                croak ("cfo_setup: $@") if $@;
            }
        }

        logStep('cfo_setup: Install fresh ONTAP build on both filers');
        $sobj = subtest(
            'Filer/install.thpl',
            "FILER_CONN=$opts{filer_conn}",
            "FILERA=$filera",
            "FILERB=$filerb",
            "ONTAPDIR=$opts{ontapdir}",
            "FILER_ONTAP_RELEASES_DIR=$opts{ontap_releases_dir}",
            "CLIENT=$opts{client}",
            "TIMEOUT=$opts{install_timeout}",
            '-runid', "install",
            '-nobg',
            '-onexit', sub {},
        );

        # Subtest results processing
        $res_obj = NATE::Result::worst($sobj->own_results);
        if ($res_obj && $res_obj->severity > 2) {
            # Get error messages (severity > 2)
            $msgs = join("\n", $res_obj->msgs);
            croak("cfo_setup: $msgs") if ($msgs);
        }
    }

    # Setup each filer
    foreach $filer (@{$opts{filers}}) {
        $set_h = {};
        cfo_setup_a_filer(
            filer              => $filer,
            filer_conn         => $opts{filer_conn},
            client             => $opts{client},
            cmd_timeout        => $opts{cmd_timeout},
            conn_timeout       => $opts{conn_timeout},
            setup_h            => $set_h,
            aggr_name          => $opts{aggr_name},
            aggr_size          => $opts{aggr_size},
            aggr_timeout       => $opts{aggr_timeout},
            flex_size          => $opts{flex_size},
            force_aggr_create  => $opts{force_aggr_create},
            raid_type          => $opts{raid_type},
            recycle_trad       => $opts{recycle_trad},
            start_hammer       => $opts{start_hammer},
            trad_size          => $opts{trad_size},
            vol_prefix         => $opts{vol_prefix},
            set_default_ops    => $opts{set_default_ops},
            set_firmware_ops   => $opts{set_firmware_ops},
        );

        # Load the return hash
        if (!$cnt) {
            $setup_h->{filerA} = $set_h;
        } else {
            $setup_h->{filerB} = $set_h;
        }

        # Increment counter
        $cnt++;
    }

    # Get the hostnames for both filers
    $hostA = $setup_h->{filerA}->{fh}->hostname();
    $hostB = $setup_h->{filerB}->{fh}->hostname();

    # Check each partner
    logStep("cfo_setup: Verify cluster partner's nodename");
    croak("cfo_setup: Filer ($hostA) cluster partner's nodename should be ".
          "filer ($hostB). Found: $setup_h->{filerA}{cf_partner}.")
      if ($setup_h->{filerA}{cf_partner} !~ /$hostB/);

    croak("cfo_setup: Filer ($hostB) cluster partner's nodename should be ".
          "filer ($hostA). Found: $setup_h->{filerB}{cf_partner}.")
      if ($setup_h->{filerB}{cf_partner} !~ /$hostA/);
=head
    if ($opts{ping_filer}) {
        # Ping each filer
        # Extra time may be needed to ensure connection
        foreach $key (qw(filerA filerB)) {
            $fh = $setup_h->{$key}->{fh};
            # Get the current filer name from the fc connection
            $filer = $fh->hostname();
            logStep("cfo_setup: Ping from client ($opts{client}) to ".
                    "filer ($filer)");
            $ping = 0;
            $end_time = time() + $ping_timeout;
            while (time() < $end_time) {
                if (run_ping($setup_h->{$key}->{cc}, $fh->default_ip(),
                    'confirm_only' => 1) > 0) {
                    $ping = 1;
                    last;
                } else {
                    sleep($wait_time);
                }
            }
            # Connection should be established
            croak("cfo_setup: Unable to ping from client ($opts{client}) to ".
                  "filer ($filer)")
              if !$ping;
        }
    }
=cut
    # Start NFS hammer
    if ($opts{start_hammer}) {
        # Start hammer traffic on both filers on each volume
        # Get the volume names, they will be same on both filers
        @vols = @{$setup_h->{filerA}->{vol_names}};

        # Get NFS hammer started
        foreach $key (qw(filerA filerB)) {
            $fc = $setup_h->{$key}->{fc};
            # Get the current filer name from the fc connection
            $filer = $fc->hostp()->hostname();

            ## @step Starting NFS hammer on filer on each volume
            logStep("cfo_setup: Starting NFS hammer on filer ($filer) on ".
                    "each volume");
            foreach $vol (@vols) {
                # Subtest NFS hammer background
                $nfs_hamid = nfs_hammer_start(
                    fc          => $fc,
                    client      => $opts{client},
                    hammer_bin  => $opts{hammer_bin},
                    hammer_time => $opts{hammer_time},
                    vol_name    => $vol,
                    panic       => $opts{hammer_panic},
                );

                # Save the hammer ids for cleanup
                push @{$setup_h->{$key}->{nfs_hamid}}, $nfs_hamid;
            }
        }

        # Verify that NFS hammer traffic started
        foreach $fc ($setup_h->{filerA}->{fc}, $setup_h->{filerB}->{fc}) {
            # Get the current filer name from the fc connection
            $filer = $fc->hostp()->hostname();

            logStep("cfo_setup: Verify that NFS hammer traffic started ".
                    "on filer ($filer)");
            $end_time = time() + $opts{nfs_wait};
            do {
                # Collect one minute worth of stats every 10 secs
                %result = filer_sysstat_get(
                            $fc, {'FILER_SYSSTAT_COUNT'    => 6,
                                  'FILER_SYSSTAT_INTERVAL' => 10,
                                 }
                );
            } while (($result{NFS}{Avg} == 0) && (time() < $end_time));

            croak("cfo_setup: Initial NFS hammer traffic did ".
                              "not start up on filer ($filer)")
              if ($result{NFS}{Avg} == 0);
        }
    }

    return;
}

#---------------------------------------------------------------------

=head2 cfo_setup_a_filer

  cfo_setup_a_filer(
            filer                 => $filer,
            client                => $CLIENT,
            cmd_timeout           => $CMDTIMEOUT,
            conn_timeout          => $TIMEOUT,
            filer_conn            => $FILER_CONN,
            setup_h               => \%set_h
            [, aggr_name          => $FILER_AGGR ]
            [, aggr_size          => $AGGR_SIZE ]
            [, aggr_timeout       => $AGGR_TIMEOUT ]
            [, force_aggr_create  => 0|1 ]
            [, raid_type          => $RAID_TYPE ]
            [, recycle_trad       => $RECYCLE_TRAD ]
            [, start_hammer       => 0|1 ]
            [, trad_size          => $TRAD_SIZE ]
            [, vol_prefix         => $VOL_PREFIX ]
            [, set_default_ops    => $SET_DEFAULT_OPS ]
            [, set_firmware_ops   => $SET_FIRMWARE_OPS ]
  );

Perform generic setup for CF test cases. This includes creating a trad and a
flex volume. Optionally the trad can be recycled and the default value of cf
options can be set.

The following named arguments are accepted:

=over

=item Arguments

=over

=item filer

(Required) : Hostspec of cluster filer.

=item filer_conn

(Required) : Filer connection type.

=item client

(Required) : Hostspec of Client.

=item cmd_timeout

(Required) : NATE say timeout for issued commands.

=item conn_timeout

(Required) : NATE connection setup timeout.

=item setup_h

(Required) : Hash reference where setup state is passed back to test.

=item aggr_name

(Optional) : Name of aggregate to use when creating a volume. Required if
option hammer_start is set.

=item aggr_size

(Optional) : Size of aggregate. Required if option hammer_start is set.

=item aggr_timeout

(Optional) : Timeout of filer aggr create. Required if option hammer_start
is set.

=item flex_size

(Optional) : The size of the volume to create. Required if option
hammer_start is set.

=item force_aggr_create

(Optional) : Force aggregate creation flag: 0 or 1.
Default is 1.

=item raid_type

(Optional) : Raid type to create: raid_dp or raid4. Required if hammer_start
is set.

=item start_hammer

(Optional) : Start hammer flag: 0 or 1.
Default is 0.

=item recycle_trad

(Optional) : Recycle trad volumes flag: 0 or 1. This routine get a snapshot of
the trad volume to be restored during cleanup. The trad volume should be empty
so that it can be reused for other tests.

Default is 0.

=item trad_size

(Optional) : The number of disks to use for the trad volume. Required if
option hammer_start is set.

=item vol_prefix

(Optional) : Volume name prefix. Required if option hammer_start is set.

=item set_default_ops

(Optional) : Set default value of cf options flag: 0 or 1.
Defaults to 0.

=item set_firmware_ops

(Optional) : Set default value of firmware options flag: 0 or 1.
Defaults to 0.

=back

=item Returns

Nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cfo_setup_a_filer(%;) {
    my (%opts) = validate (
        @_,
        {
            filer              => { type => SCALAR },
            filer_conn         => { type => SCALAR },
            client             => { type => SCALAR },
            cmd_timeout        => { type => SCALAR, regex => qr/^\d+$/ },
            conn_timeout       => { type => SCALAR, regex => qr/^\d+$/ },
            setup_h            => { type => HASHREF },
            aggr_name          => { type => SCALAR|UNDEF, optional => 1 },
            aggr_size          => { type => SCALAR|UNDEF, optional => 1 },
            aggr_timeout       => { type => SCALAR|UNDEF, optional => 1 },
            flex_size          => { type => SCALAR|UNDEF, optional => 1 },
            force_aggr_create  => { type => SCALAR, regex => qr/^[01]$/,
                                      default => 1 },
            raid_type          => { type => SCALAR|UNDEF, optional => 1 },
            recycle_trad       => { type => SCALAR, regex => qr/^[01]$/,
                                      default => 0 },
            start_hammer       => { type => SCALAR, regex => qr/^[01]$/,
                                      default => 0 },
            trad_size          => { type => SCALAR|UNDEF, optional => 1 },
            vol_prefix         => { type => SCALAR|UNDEF, optional => 1 },
            set_default_ops    => { type => SCALAR, regex => qr/^[01]$/,
                                      default => 0 },
            set_firmware_ops   => { type => SCALAR, regex => qr/^[01]$/,
                                      default => 0 },
        }
    );
    my ($aggr_exist, $cc, $cmd, $fc, $fconn, $fh, $results, $key, %save_env,
        %set_env, $vol_name_flex, $vol_name_trad);
    my $clean_snapshot = '__clean__';
    my $snap_licensed = 1;
    my $set_h = $opts{setup_h};

    # Save filer connections options
    $set_h->{filer} = $opts{filer};
    $set_h->{filer_conn} = $opts{filer_conn};
    $set_h->{cmd_timeout} = $opts{cmd_timeout};
    $set_h->{conn_timeout} = $opts{conn_timeout};
    $set_h->{start_hammer} = $opts{start_hammer};

    # Get a session to the filer
    $fh = host($opts{filer});
    $set_h->{fh} = $fh;
    transit_to($fh, 'UP') or croak ("cfo_setup: $@");
=head
    # Set firmware configuration variables if required for testing
    if ($opts{set_firmware_ops}) {
        logStep("cfo_setup: Get a console connection to filer ($opts{filer})");
        host($opts{filer});
        $fconn = connect('#console', -timeout, $opts{conn_timeout});
        $fconn->timeout($opts{cmd_timeout});

        # Check firmware configuration variables
        %save_env = ();
        foreach $key (keys %firmware_env) {
            $cmd = "priv set -q test; bootargs get $key";
            $results = sayEOF($fconn, $cmd);
            if ($results !~ /$firmware_env{$key}/i) {
                $save_env{$key} = '';
                $set_env{$key} = $firmware_env{$key};
            }
        }

        # Save original firmware configuration variables
        if (%save_env) {
            get_firmware_env($opts{filer}, env_hash => \%save_env,
                             final_state=>'FIRMWARE') or croak ("cfo_setup: $@");
            $set_h->{firmware} = \%save_env;

            # Set firmware configuration variables
            set_firmware_env(
                             $opts{filer},
                             env_hash    => \%set_env,
                             final_state => 'UP')
              or croak ("cfo_setup: $@");
        }
    }
=cut
    logStep("cfo_setup: Get a $opts{filer_conn} connection to filer ".
            "($opts{filer})");
    host($opts{filer});
    $fc = connect($opts{filer_conn}, -timeout, $opts{conn_timeout});
    $fc->timeout($opts{cmd_timeout});
    $set_h->{fc} = $fc;

    # Save miscellaneous variables
    $set_h->{wait_time} = $wait_time;
    $set_h->{ping_timeout} = $ping_timeout;

    # Make sure that the filer is setup properly for clustering. Note that
    # if this is the first node being setup that clustering might not get
    # setup successfully until the 2nd node has been configured. Therefore,
    # ignore any errors from setup_cluster.
    logStep("cfo_setup: Make sure that filer ($opts{filer}) is setup ".
            "properly for clustering");
	 #commented since check for licensing is not need

    logStep("cfo_setup: Get cluster partner's nodename");
    $cmd = 'cf partner';
    $results = sayEOF($fc, $cmd);

    # Save the nodename
    $set_h->{cf_partner} = $results;

    # Set all of the cf options to their default values
    if ($opts{set_default_ops}) {
        logStep("cfo_setup: Set default value of cf options on ".
                "filer ($opts{filer})");
        $results = set_default_options(fc => $fc);
        croak("cfo_setup: Unable to reset options to their defaults")
          if $results;
    }

    # Save the default value of all of the cf options
    $set_h->{default_options} = \%default_options;

    if ($opts{start_hammer}) {
        # Create aggregate and volumes if needed
        $vol_name_flex = $opts{vol_prefix} ."_flex";
        $vol_name_trad = $opts{vol_prefix} ."_trad";

        # The volume names will be returned in an array in $setup_h{vol_names}[]
        logStep("cfo_setup: Create one flex volume and one trad volume on ".
                "filer ($opts{filer})");
        # Flex
        push @{$set_h->{vol_names}}, $vol_name_flex;
        setup_aggr(
            fc        => $fc,
            aggr_name => $opts{aggr_name},
            aggr_size => $opts{aggr_size},
            raidtype  => $opts{raid_type},
            timeout   => $opts{aggr_timeout},
            force     => $opts{force_aggr_create},
        );

        setup_vol(
            fc        => $fc,
            aggr_name => $opts{aggr_name},
            vol_name  => $vol_name_flex,
            flex_size => $opts{flex_size}
        );

        # Trad Volume
        $aggr_exist = 0;

        # If the recycle flag is on then check the existence of the aggregate
        if ($opts{recycle_trad}) {
            $aggr_exist = aggr_exists($fc, $vol_name_trad);
        }

        # Create a trad volume if the recycle flag is off or if the trad
        # volume does not exist
        push @{$set_h->{vol_names}}, $vol_name_trad;
        if (!$aggr_exist) {
            vol_create(
                $fc,
                $vol_name_trad,
                $opts{trad_size},
                force    => $opts{force_aggr_create},
                raidtype => $opts{raid_type},
                timeout  => $opts{aggr_timeout},
            );
            croak ("cfo_setup: $@") if $@;
        }

        # Recycle trad volumes. This routine gets a snapshot of the trad volume
        # to be restored during cleanup. The trad volume should be empty so that
        # it can be reused for other tests.
        if ($opts{recycle_trad}) {
            logStep("cfo_setup: Recycle trad volume on filer ($opts{filer})");

            # Save recycle option for cleanup
            $set_h->{recycle_trad} = $opts{recycle_trad};

            # Save the name of the recycle volume
            $set_h->{recycle_vol} = $vol_name_trad;

            # Get the original state of snaprestore license and
            # save the state for cleanup
            $snap_licensed = is_licensed($fc, 'snaprestore');
            croak ("cfo_setup: $@") if $@;
            $set_h->{snap_licensed} = $snap_licensed;

            logStep("cfo_setup: License snaprestore on filer ".
                    "($opts{filer}) if needed");
            # License snap restore if needed
            if (!$snap_licensed) {
                license($fc, "snaprestore");
                croak ("cfo_setup: $@") if $@;
            }

            # Create snapshot of the trad volume on the filer and save the
            # snapshot name for cleanup
            logStep("cfo_setup: Create snapshot of volume ($vol_name_trad) ".
                    "on filer ($opts{filer})");
            # Get rid of old snapshot if it exists
            if (snapshot_exists($fc, $vol_name_trad, $clean_snapshot)) {
                filer_vol_snap_delete_all($fc, $set_h->{recycle_vol});
            }

            # Create a fresh snapshot
            snapshot_create($fc, $vol_name_trad, $clean_snapshot);
            croak ("cfo_setup: $@") if $@;
            $set_h->{clean_snapshot} = $clean_snapshot;
        }
    }
=head
    # Get a connection to the client
    logStep('cfo_setup: Connect to client');
    host($opts{client});
    $cc = connect('#rsh', -timeout, $set_h->{conn_timeout});
    $cc->timeout($set_h->{cmd_timeout});
    $set_h->{client} = $opts{client};
    $set_h->{cc} = $cc;
=cut
    return;
}

#---------------------------------------------------------------------

=head2 cfo_cleanup

  cfo_cleanup(
               setup_h            => \%set_h,
               cleanup            => 'test'
               [, set_default_ops => 1 ]
  );

Perform generic cleanup for CFO test cases.

The following named arguments are accepted:

=over

=item Arguments

=over

=item setup_h

(Required) : Hash reference where setup state is passed back to test.

=item cleanup

(Required) : Cleanup to be performed after each test using one of the
             following parameters:
                 none - no cleanup at all.
                 vol/test - only test specific volumes are destroyed.
                 all - all volumes & aggrs are destroyed

=item set_default_ops

(Optional) : Set default value of cf options 0|1.
Default is 0.

=back

=item Returns

Nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cfo_cleanup (%;) {
    my (%opts) = validate (
        @_,
        {
            cleanup         => { type => SCALAR,
                                   regex => qr/^(none|vol|test|all)$/ },
            setup_h         => { type => HASHREF },
            set_default_ops => { type => SCALAR, default => 0}
        }
    );
    my ($fh, $key, $nfs_hamid);
    my $set_h = $opts{setup_h};

    # If cfo_setup() has not run then return instead of running cfo_cleanup()
    return if !%{$set_h};

    # Catch any errors to log a warning
    eval {
        # Make sure filers are up
        foreach $key (qw(filerA filerB)) {
            $fh = $set_h->{$key}->{fh};
            if (defined ($fh)) {
                transit_to($fh, 'UP') or logWarn("cfo_cleanup: $@");
            }
        }

        # Stop all processes
        foreach $key (qw(filerA filerB)) {
            if (defined @{$set_h->{$key}->{nfs_hamid}}) {
                logStep('cleanup: Stop all the NFS hammer tests');
                while (scalar(@{$set_h->{$key}->{nfs_hamid}}) > 0) {
                    $nfs_hamid = pop @{$set_h->{$key}->{nfs_hamid}};
                    eval { nfs_hammer_stop(id => $nfs_hamid) };
                    logWarn("cfo_cleanup: $@") if $@;
                }
            }
        }

        if ($opts{cleanup} ne 'none') {
            foreach $key (keys %{$set_h}) {
                # Run cleanup the environment
                cfo_cleanup_a_filer(
                    cleanup         => $opts{cleanup},
                    setup_h         => $set_h->{$key},
                    set_default_ops => $opts{set_default_ops},
                );
            }
        }
    };
    logWarn("cfo_cleanup: $@") if $@;
}

#---------------------------------------------------------------------

=head2 cfo_cleanup_a_filer

  cfo_cleanup_a_filer(
                      setup_h            => \%set_h,
                      cleanup            => 'test'
                      [, set_default_ops => 1 ]
  );

Perform generic cleanup for CFO test cases.

The following named arguments are accepted:

=over

=item Arguments

=over

=item setup_h

(Required) : Hash reference where setup state is passed back to test.

=item cleanup

(Required) : Cleanup to be performed after each test using one of the
             following parameters:
                 none - no cleanup at all.
                 vol/test - only test specific volumes are destroyed.
                 all - all volumes & aggrs are destroyed

=item set_default_ops

(Optional) : Set default value of cf options 0|1.
Default is 0.

=back

=item Returns

Nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cfo_cleanup_a_filer (%;) {
    my (%opts) = validate (
        @_,
        {
            cleanup         => { type => SCALAR,
                                   regex => qr/^(none|vol|test|all)$/ },
            setup_h         => { type => HASHREF },
            set_default_ops => { type => SCALAR, default => 0},
            filerA  => { type => SCALAR, default => $main::FILERA },
            filerB  => { type => SCALAR, default => $main::FILERB },
        }
    );
    my ($cc, $fc, $filer, $results, $vol);
    my $set_h = $opts{setup_h};
    my ($filerA, $connA);
    my ($filerB, $connB);
    my $stateA = '';
    my $stateB = '';
    my ($key, $after);    
    # Create connection handles for both filers
    $filerA = host($opts{filerA});
    $connA  = connect($main::FILER_CONN, '-timeout', $main::TIMEOUT);
    $connA->timeout($main::CMDTIMEOUT);

    $filerB = host($opts{filerB});
    $connB  = connect($main::FILER_CONN, '-timeout', $main::TIMEOUT);
    $connB->timeout($main::CMDTIMEOUT);



    # If filer is not defined in the setup hash then cfo_setup never ran
    return if (!defined($set_h->{filer}));

    # Take care of cleaning up the filer in this eval
    eval {
        # Reconnect to the filer
        host($set_h->{filer});
        $fc = connect($set_h->{filer_conn}, -timeout,
                      $set_h->{conn_timeout});
        $fc->timeout($set_h->{cmd_timeout});

        if ($opts{cleanup} ne 'none') {
            # Reset the original firmware variables
            if (defined($set_h->{firmware})) {
                logStep("cfo_cleanup: Restore original firmware variables");
                set_firmware_env($set_h->{filer},
                                 env_hash    => $set_h->{firmware},
                                 final_state => 'UP')
                  or logWarn("cfo_cleanup: $@");
            }
    #Added to make the filer to ha mode

    # Changing to HA part is added
     my $status_nodeA = filer_options_get($connA, 'cf.mode');
    if (($status_nodeA =~ /non_ha/i)){
    ##########################
    ##Change to HA mode
    ##########################

    logStep("CHANGING TO HA");
    $key = 'cf.mode';
    filer_options_set($connA, $key, "ha");
    filer_options_set($connB, $key, "ha");

    logStep("REBOOT both nodes after changing cf.mode options");
    say ($connA, "priv set -q diag; bootargs set AUTOBOOT TRUE; result -l", '-eval_error');
    say ($connA, "reboot -t 1", '-eval_error');

    say ($connB, "priv set -q diag; bootargs set AUTOBOOT TRUE; result -l", '-eval_error');
    say ($connB, "reboot -t 1", '-eval_error');
    sleep (300); #need time to reboot


    my $status_nodeA = filer_options_get($connA, $key);
    my $status_nodeB = filer_options_get($connB, $key);
    if (($status_nodeA =~ /ha/i) && ($status_nodeB =~ /ha/i)) {
         tstlog(-comment, "After REBOOT the filers are in HA mode on both nodes");
    } else {
         logresult("FATAL", "Both nodes are still in NON-HA mode");
    }

    cf_enable($filerA);
    sleep 10;

    (undef, undef, $after) = say ($connA, "cf status", '-eval_error');
    if ($after =~ /Controller Failover enabled/) {
        tstlog(-comment, "Controller Failover Enabled");
    }else {
        logresult("FATAL", "Controller Failover is NOT Enabled");
    }

    ## @step Execute 'cf enable' on one filer
     cf_enable($filerB);
    sleep 10;

    (undef, undef, $after) = say ($connB, "cf status", '-eval_error');
    if ($after =~ /Controller Failover enabled/) {
        tstlog(-comment, "Controller Failover Enabled");
    }else {
        logresult("FATAL", "Controller Failover is NOT Enabled");
    }
    }
    #Added to make the filer to ha mode
            # Make sure that the filer is setup properly for clustering

=head
            # License filer because a test had unlicensed clustering and died
            if (!filer_cluster_is_licensed($fc)) {
                # Cluster Failover has been unlicensed by the test

                # Get the filer name
                $filer = $fc->hostp()->hostname();

                logStep("cfo_cleanup: License cluster ($filer)");
                # We don't know if the filer has been rebooted since
                # clustering was unlicensed so reboot first to start at
                # a known state.
                reboot($fc->hostp);
                logWarn("cfo_cleanup: $@") if $@;

                # License filer
                # Ignore licensing errors since the filer must be rebooted
                eval { filer_cluster_license($fc) };

                # Filer must be rebooted before re-licensing takes
                reboot($fc->hostp);
                logWarn("cfo_cleanup: $@") if $@;
                if (!filer_cluster_is_licensed($fc)) {
                    logWarn("Could not license cluster ($filer): $@") if $@;
                }
            }
=cut

            # Both "cf status" and "cf monitor" are used to determine the state
            # of the filer with respect to clustering
            # Some of these steps will not 'take' until the second filer has
            # been through 'cleanup'
            $results = get_cf_state($set_h->{filer});
            if ($results !~ /UP/i) {
                logStep("cfo_cleanup: Setup cluster");

                # Cluster Failover has been unlicensed. Filer must be
                # rebooted before re-licensing Cluster Failover.
                reboot($fc->hostp);
                logWarn("cfo_cleanup: $@") if $@;

                transit_to($fc->hostp, 'UP');
 logStep("cfo_cleanup: cf_enable");
                # Enable cluster if possible
                cf_enable($set_h->{filer});
=head
                # Setup cluster if possible
                eval { setup_cluster() };
=cut
            }

            if ($set_h->{start_hammer}) {
                # Remove any volumes created on the filer
                # Do not remove the trad volume if the recycle trad flag is on
                foreach $vol (@{$set_h->{vol_names}}) {
                    if ((!$set_h->{recycle_trad}) ||
                        $vol !~ /$set_h->{recycle_vol}/) {
                        logStep("cfo_cleanup: Remove volume $vol on filer ".
                                "($set_h->{filer})");
                        vol_purge($fc, $vol) if (vol_exists($fc, $vol));
                        logWarn("cfo_cleanup: $@") if $@;
                    }
                }

                # Set all of the cf options to their default values
                if ($opts{set_default_ops}) {
                    logStep("cfo_cleanup: Set default value of cf options on ".
                            "filer ($set_h->{filer})");
                    $results = set_default_options(fc => $fc);
                    logWarn("cfo_cleanup: Unable to reset options to their ".
                            "defaults")
                      if $results;
                }

                # Recycle trad volume
                if ($set_h->{recycle_trad}) {
                    if (defined($set_h->{clean_snapshot})) {
                        logStep("cfo_cleanup: Recycle trad volume on filer ".
                                "($set_h->{filer})");

                        # Restore the snapshot
                        logStep("cfo_cleanup: Restore the snapshot ".
                                "($set_h->{clean_snapshot})");
                        snaprestore_volume(
                            $fc,
                            $set_h->{recycle_vol},
                            $set_h->{clean_snapshot},
                        );
                        logWarn("cfo_cleanup: $@") if $@;

                        # Un-License snap restore if needed
                        if (!$set_h->{snap_licensed}) {
                            unlicense($fc, "snaprestore");
                            logWarn("cfo_cleanup: $@") if $@;
                        }
                    }

                    # Delete all volume snapshots
                    eval {
                        filer_vol_snap_delete_all($fc, $set_h->{recycle_vol});
                    };
                    logWarn("cfo_cleanup: $@") if $@;
                }
            }
        }
        if ($opts{cleanup} eq 'all') {
            if (defined($set_h->{aggr_name})) {
                aggr_purge($fc, $set_h->{aggr_name}) or
                  logWarn("cfo_cleanup: $@");
            }
        }
    };
    logWarn("cfo_cleanup: $@") if $@;
    return;
}

#---------------------------------------------------------------------

=head2 cfo_takeover

    cfo_takeover(fc => $fc);

This routine issues the 'cf takeover [-f | -n]' on the filer connection and
croaks if a 'cf: ....' error is returned. It does not help the partner boot if
the partner is a simulator. It only issues the command on the filer.
The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=item partner_mode

(Optional) : Run the command with 'partner' prefix.

=item force

(Optional) : Connection object to a filer.

=item non_disruptive

(Optional) : Run the non disruptive cluster takeover.

=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cfo_takeover(%;) {
    my (%opts) = validate (
        @_,
        {
            fc             => { type => OBJECT },
            non_disruptive => { type => SCALAR, regex    => qr/^[01]$/,
                                                default  => 0 },
            partner        => { type => SCALAR, optional => 1,
                                                regex    => qr/^[01]$/ },
            options        => { type => SCALAR, optional => 1 },
            force          => { type => SCALAR, regex    => qr/^[01]$/,
                                                default  => 1 },

        }
    );

    my ($cmd, $out);
    my $options = '';

    # Process options
    if ($opts{options}) {
        # Override all other options
        $options = $opts{options};
    } else {
        # Set any option specified
        $options .= ' -f' if $opts{force};
        $options .= ' -n' if $opts{non_disruptive};
    }

    # Send command to filer
    $cmd = "cf takeover $options";
    $out = _say($opts{fc}, $cmd, 0, $opts{partner});

    # Remove all "NOTE:" messages
    _check_for_note($out);

    # Remove syslog messages from output
    $out =~ s/^.*cf:\s+takeover\s+initiated\s+by\s+operator*\r*\n*//img;

    # Make sure not to croak on an empty line
    $out =~ s/^\s+$//;

    croak $out if $out;

    return 1;

}

#---------------------------------------------------------------------

=head2 cfo_giveback

    cfo_giveback(fc => $fc);

This routine issues the 'cf giveback [-f]' on the filer connection and
croaks if a 'cf: ....' error is returned. It does not help the partner boot if
the partner is a simulator. It only issues the command on the filer.
The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : connection object to a filer.

=item partner

(Optional) : run the command with 'partner' prefix.

=item force

(Optional) : connection object to a filer.

=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cfo_giveback(%;) {
    my (%opts) = validate (
        @_,
        {
            fc      => { type => OBJECT },
            partner => { type => SCALAR, optional => 1,
                                         regex    => qr/^[01]$/ },
            options => { type => SCALAR, optional => 1 },
            force   => { type => SCALAR, regex    => qr/^[01]$/,
                                                default  => 1 },

        }
    );

    my ($cmd, $out);
    my $options = '';

    # Process options
    if ($opts{options}) {
        # override all other options
        $options = $opts{options};
    } else {
        # set any option specified
        $options .= ' -f' if $opts{force};
    }

    # Send command to filer
    $cmd = "cf giveback $options";
    $out = _say($opts{fc}, $cmd, 0, $opts{partner});

    # Remove all "NOTE:" messages
    _check_for_note($out);

    # Remove syslog messages from output
    $out =~ s/^.*cf:\s+giveback\s+initiated\s+by\s+operator*\r*\n*//img;

    # Make sure not to croak on an empty line
    $out =~ s/^\s+$//;

    croak $out if $out;

    return 1;

}

#---------------------------------------------------------------------

=head2 check_fm_config

    check_fm_config(
        fc  => $fc,
    );

This routine verifies that the ontap version is at least ironcity, the filer
model is supported along with the bios version.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

Returns nothing or croaks on error

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub check_fm_config(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );

    my ($act_bios, $expected_bios, $filer, $min_bios, $ontap_ver, $sysconfig,
        $supported, $tmp, @bios_tmp, $value, $int);

    my $model   = '';
    my $bios    = '';
    my $model_supported = 0;

    logStep("Check filer is setup for failover/monitor tests");
    logStep("Verify filer is not a simulator");
    # Filer cannot be a simulator
    if ($opts{fc}->hostp()->hostname() =~ /simulator/i) {
        croak("Filer cannot be a simulator for these tests");
    }

    logStep("Verify filer OS is at least IRONCITY 7.3 loaded");
    # Check for at least Ironcity build 7.3
    $ontap_ver = get_wafl_version($opts{fc}->hostp());
    if ($ontap_ver < WV_IRONCITY_MIN) {
        croak("The filer must be running running at least IRONCITY 7.3");
    }

    logStep("Verify filer model and bios");
    # Get model and bios
    $sysconfig = filer_sysconfig_a_check($opts{fc});
    if ($sysconfig =~ /Model\s+Name\:\s+(\S+)/i) {
        $model = $1;
    }
    if ($sysconfig =~ /BIOS\s+version\:\s+(\S+)/i) {
        $bios = $1;
    } elsif ($sysconfig =~ /Firmware\s+release\:\s+CFE\s+(\S+)/i) {
        $bios = $1;
    }

    foreach $filer (sort keys %supported_systems) {
        $tmp = $filer;
        $tmp =~ s/XX//;
        if ($model =~ /^$tmp/i) {
            $model_supported = $filer;
            last;
        }
    }

    if (!$model_supported) {
        $supported = join (',', keys %supported_systems);
        croak("This model is not supported. Supported model types are: ".
              "($supported) and tests are running on model ($model)");
    } else {

        # Convert version numbers into base ten number comparisons
        # Compute expected values
        $expected_bios   = $supported_systems{$model_supported}{bios_version};

        # Compute expected minimum filer values
        @bios_tmp = split /\./,$expected_bios;

        $int = 100000;
        while (@bios_tmp) {
            $value = shift @bios_tmp;
            $min_bios += $value * $int;
            $int /= 100;
        }

        # Compute given filer values
        $bios =~ s/X/\./;
        @bios_tmp = split /\./,$bios;

        $int = 100000;
        while (@bios_tmp) {
            $value = shift @bios_tmp;
            $act_bios += $value * $int;
            $int /= 100;
        }

        # compare versions on a base ten comparison
        if ($act_bios < $min_bios) {
            croak("Correct model version ($model) but incorrect bios ".
                  "version. Found ($bios) and expected at least ".
                  "($expected_bios)\n");
        }
    }

    return;
}

#---------------------------------------------------------------------

=head2 check_msgs

    check_msgs(
        log_obj     => $log,
        log_timeout => 120,
        msg         => $msg,
    );

This routine searches for the given message from a LogReader object. If the
message is not found, the routine will croak with an error saying the message
cannot be found.

The following named arguments are accepted:

=over

=item Arguments

=over

=item log_obj

(Required) : LogReader object from a filer

=item log_timeout

(Optional) : Maximum time to wait for the message to appear in /etc/messages

=item msg

(Required) : Message to search for in /etc/messages. Regex are allowed.

=back

=item Returns

Returns 1 on success or croaks on error

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub check_msgs(%;) {
    my (%opts) = validate (
        @_,
        {
            log_obj     => { type => OBJECT },
            log_timeout => { type => SCALAR, default => 300 },
            msg         => { type => SCALAR },
        }
    );
    my ($end_time, $log, $text, $timeout);

    $log = $opts{log_obj};

    $timeout = 1;
    $end_time = time() + $opts{log_timeout};
    while (time() < $end_time) {
        $log->stop();
        $text = $log->{'logs'}{'messages'}->to_string;

        if ($text !~ /$opts{msg}/i) {
            sleep(30);
            next;
        } else {
            $timeout = 0;
            last;
        }
    }
    if ($timeout) {
        croak "\nCannot find '$opts{msg}' in messages log";
    } else {
        return 1;
    }
}

#---------------------------------------------------------------------

=head2 disk_simpush_local_disk

    disk_simpush_local_disk(fc => $fc, disk => 'v0.29');

    disk_simpush_local_disk(
        fc   => $fc,
        disk => 'v0.29:NETAPP__:VD-100MB-FZ-520_:30042012:261248',
    );

This routine sets the default cf options.

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=item disk

(Required) : Simulator disk to push in

=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub disk_simpush_local_disk(%;) {
    my (%opts) = validate (
        @_,
        {
            fc   => { type => OBJECT },
            disk => { type => SCALAR },
        }
    );
    my ($after, $cmd, $disk, @entries, $before, $out, $push_disk);

    if ($opts{fc}->hostp()->hosttype() !~ /simulator/i) {
        croak "This command only works on simulators";
    }
    $before = get_simpush_disks(fc => $opts{fc});

    foreach $disk (@{$before->{local}}) {
        if ($disk =~ /$opts{disk}/i) {
            $push_disk = $disk;
        }
    }
    if (!defined($push_disk)) {
        croak "Cannot push in $opts{disk}";
    }

    $cmd = "disk simpush $push_disk";
    $out = sayEOF(
        $opts{fc},
        $cmd,
    );

    $after = get_simpush_disks(fc => $opts{fc});
    foreach $disk (@{$after->{local}}) {
        if ($disk =~ /$opts{disk}/i) {
            croak "Pushed in disk but still listed as needing to be pushed in";
        }
    }
    return;
}

#---------------------------------------------------------------------

=head2 fmt_exp_act_results

    my $output = fmt_exp_act_results(
        actual     => 'Reported incorrect results',
        expected   => 'Should have seen these results',
        [prefix    => 'This is a prefix to be printed before expected results',]
        [postfix   => 'This is a postfix to be printed after actual results',]
    );

This routine formats the output to print expected and actual results of the
test.

=over

=item Arguments

=over

=item actual

(Required) : String to be passed into as the actual results of the test

=item expected

(Required) : String to be passed into as the expected results of the test

=item prefix

(Optional) : String to be passed into as the prefix before the expected
             results of the test

=item postfix

(Optional) : String to be passed into as the postfix after the actual
             results of the test

=back

=item Returns

Returns a formated string.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub fmt_exp_act_results(%;) {
    my (%opts) = validate (
        @_,
        {
            prefix   => { type => SCALAR|UNDEF, optional => 1 },
            expected => { type => SCALAR|ARRAYREF },
            actual   => { type => SCALAR|ARRAYREF },
            postfix  => { type => SCALAR|UNDEF, optional => 1 },
        }
    );
    my ($actual, $expected, $prefix, $postfix, $format, $msg);

    # If we have a prefix use it
    if (defined $opts{prefix}) {
        $prefix = $opts{prefix};
    } else {
        $prefix = '';
    }

    # If we have a postfix use it
    if (defined $opts{postfix}) {
        $postfix = "$opts{postfix}\n";
    } else {
        $postfix = '';
    }

    # Determine if the expected messages are in an array or scalar
    if ( ref $opts{expected} eq 'ARRAY' ) {
        foreach $msg (@{$opts{expected}}) {
            $expected .= "$msg\n";
        }
    } else {
        $expected = "$opts{expected}\n";
    }

    # Determine if the actual messages are in an array or scalar
    if ( ref $opts{actual} eq 'ARRAY' ) {
        foreach $msg (@{$opts{actual}}) {
            $actual .= "$msg\n";
        }
    } else {
        $actual = "$opts{actual}\n";
    }

    # Format the results
    $format = "$prefix\n".
              "Expected Results:\n$expected\n".
              "Actual Results:\n$actual\n".
              "$postfix";

    return $format;
}
#---------------------------------------------------------------------

=head2 get_simpush_disks

    my $href = get_simpush_disks(fc => $fc);

    my @local_disks = @{$href->{local}};

This routine sets the default cf options.

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

Returns a hash reference of simulator disks that have been pulled
or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_simpush_disks(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );
    my ($cmd, %disks, $line, @local_disks, $out, @output,
        @partner_disks, $search);

    if ($opts{fc}->hostp()->hosttype() !~ /simulator/i) {
        croak "This command only works on simulators";
    }

    $cmd = 'disk simpush -l';
    $out = sayEOF(
        $opts{fc},
        $cmd,
    );

    @output = split /\n/, $out;
    foreach my $line (@output) {
        if ($line =~
            /There\s+are\s+no\s+local\s+disks\s+available\s+for\s+pushing/i) {
            @local_disks = ();
        } elsif ($line =~
            /There\s+are\s+no\s+partner\s+disks\s+available\s+for\s+pushing/i) {
            @partner_disks = ();
        } elsif ($line =~ /The\s+following\s+local\s+pulled\s+disks\s+are\s+available\s+for\s+pushing/i) {
            shift @output;
            while (($search = shift @output)) {
                $search =~ s/^\s+//g;
                if (($search !~ /partner/i) &&
                    ($search =~ /\w+\.\d+\:\S+\:\S+\:\d+\:\d+$/i)) {
                    push (@local_disks, $search);
                }
            }
        } elsif ($line =~ /The\s+following\s+partner\s+pulled\s+disks\s+are\s+available\s+for\s+pushing/i) {
            shift @output;
            shift @output;
            if (!@local_disks) {
                shift @output;
            }
            while (($search = shift @output)) {
                $search =~ s/^\s+//g;
                if ($search =~ /\w+\.\d+\:\S+\:\S+\:\d+\:\d+$/i) {
                    push (@partner_disks, $search);
                }
            }
        }
    }

    $disks{local} = \@local_disks;
    $disks{partner} = \@partner_disks;
    return(\%disks);
}

#---------------------------------------------------------------------

=head2 license_cluster

    license_cluster(fc => $fc);

This routine is an interface for licensing cluster on filers in 7-mode and
10-mode filers.

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=item Returns

Returns 1 or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub license_cluster(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );
    my ($fc, $hostname);

    $fc = $opts{fc};
    $hostname = $fc->hostp()->hostname();

    # Make sure both filers are licensed for cluster
    if (!filer_cluster_is_licensed($fc)) {
        # Ignore licensing errors since the filer must be rebooted
        eval { filer_cluster_license($fc) };
        eval { reboot($fc->hostp()) };
        croak "Could not reboot after licensing cluster on ".
              "filer ($hostname): $@" if $@;
        if (!filer_cluster_is_licensed($fc)) {
            croak "Could not license clustering on ".
                  "filer ($hostname): $@" if $@;
        }
    }
    return 1;
}

#---------------------------------------------------------------------

=head2 set_default_options

    set_default_options(fc => $fc);

This routine sets the default value cf options.

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub set_default_options(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type  => OBJECT },
        }
    );
    my ($key, $results);
    my $err = 0;

    foreach $key (keys %default_options) {
        $results = filer_options_get($opts{fc}, $key);
        if ($results ne $default_options{$key}) {
            eval { filer_options_set($opts{fc}, $key, $default_options{$key}, 0) };
            $err = 1 if $@;
        }
    }
    return $err;
}

#---------------------------------------------------------------------
=head2 sanown_ownership_cleanup

    $orig_sanown = {
             'FILERA' => 'v1.21' => {
                                      'serial_number' => '29732704',
                                      'owner' => 'FILERA',
                                      'PRkey' => '0',
                                      'duid' => '4E455441:50502020',
                                      'sysid' => '000000001',
                                      'UpCnt' => '47',
                                      'home_sysid' => '000000001',
                                      'pool' => 'Pool0',
                                      'home' => 'FILERA'
                                    },
                         'v4.20' => {
                                      'serial_number' => '29732501',
                                      'owner' => 'FILERB',
                                      'PRkey' => '0',
                                      'duid' => '4E455441:50502020',
                                      'sysid' => '000000002',
                                      'UpCnt' => '9',
                                      'home_sysid' => '000000002',
                                      'pool' => 'Pool0',
                                      'home' => 'FILERB'
                                    },
             'FILERB' => (...),
    }

    $new_sanown = {
             'FILERA' => 'v1.21' => {
                                      'serial_number' => '29732704',
                                      'owner' => '',
                                      'PRkey' => '0',
                                      'duid' => '4E455441:50502020',
                                      'sysid' => '000000003',
                                      'UpCnt' => '47',
                                      'home_sysid' => '000000003',
                                      'pool' => 'Pool0',
                                      'home' => ''
                                    },
                         'v4.20' => {
                                      'serial_number' => '29732501',
                                      'owner' => 'FILERA',
                                      'PRkey' => '0',
                                      'duid' => '4E455441:50502020',
                                      'sysid' => '000000001',
                                      'UpCnt' => '9',
                                      'home_sysid' => '000000001',
                                      'pool' => 'Pool0',
                                      'home' => 'FILERA'
                                    },
             'FILERB' => (...),
    }

    sanown_ownership_cleanup(
        fcA => $fcA,
        fcB => $fcB,
        orig_sanown => \%orig_sanown,
        new_sanown  => \%new_sanown,
    );

Usage: setup_aggr(fcA         => $connA,
                  fcB         => $connB,
                  orig_sanown => \%orig_sanown,,
                  new_sanown  => \%new_sanown);

This routine restores the sanown environment back to the original state
provided by the orig_sanown hash.

=over

=item Arguments

=over

=item fcA

(Required) : Connection object to a filer.

=item fcB

(Required) : Connection object to a filer.

=item orig_sanown

(Required) : Hash reference to original sanown ownership keyed by each filer.

=item new_sanown

(Required) : Hash reference to new sanown ownership keyed by each filer.

=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#-------------------------------------------------------------------------------
sub sanown_ownership_cleanup(%;) {
    my (%opts) = validate (
        @_,
        {
            fcA         => { type => OBJECT },
            fcB         => { type => OBJECT },
            sysidA      => { type => SCALAR },
            sysidB      => { type => SCALAR },
            orig_sanown => { type => HASHREF },
            new_sanown  => { type => HASHREF },
        }
    );

    my ($conn, $disk, $disk2assign, $filerA, $filerB,
        $sn1, $sn2, $sysidA, $sysidB, $sysid,
        $old_A, $old_B, $new_A, $new_B, $disk_assign);

    $filerA = $opts{fcA}->hostp()->hostname();
    $filerB = $opts{fcB}->hostp()->hostname();

    if (defined $opts{orig_sanown}->{$filerA}
     && defined $opts{orig_sanown}->{$filerB}
     && defined $opts{new_sanown}->{$filerA}
     && defined $opts{new_sanown}->{$filerB}) {
        $old_A = $opts{orig_sanown}->{$filerA};
        $old_B = $opts{orig_sanown}->{$filerB};
        $new_A = $opts{new_sanown}->{$filerA};
        $new_B = $opts{new_sanown}->{$filerB};
    } else {
        croak("Cannot restore with the given hash. The hash must contain ".
              "keys as each filers hostname");
    }

    # Get system IDs for cleanup
    $sysidA = $opts{sysidA};
    $sysidB = $opts{sysidB};

    # FILERA
    foreach $disk (keys %{$old_A}) {
        $sn1 = $old_A->{$disk}->{serial_number};
        $sn2 = $new_A->{$disk}->{serial_number};
        $sysid = $old_A->{$disk}->{sysid};

        # Make sure the serial number match to assign the correct disk
        if ($sn1 ne $sn2) {
          $disk2assign = _search_disk_serial($new_A, $sn1);
        } else {
          $disk2assign = $disk;
        }

        # Check owner and restore ownership
        if (defined $old_A->{$disk}->{sysid}
         && defined $new_A->{$disk2assign}->{sysid}
         && ($old_A->{$disk}->{sysid} != $new_A->{$disk2assign}->{sysid})) {
            if ($new_A->{$disk2assign}->{sysid} == $sysidA) {
                eval {
                    disk_assign(
                        conn  => $opts{fcA},
                        disk  => $disk2assign,
                        sysid => $sysid,
                        force => 1,
                    );
                };
                logWarn("cleanup: $@") if $@;

            } elsif ($new_A->{$disk2assign}->{sysid} == $sysidB) {
                $disk_assign = _search_disk_serial($new_B, $sn1);
                eval {
                    disk_assign(
                        conn  => $opts{fcB},
                        disk  => $disk_assign,
                        sysid => $sysid,
                        force => 1,
                    );
                };
                logWarn("cleanup: $@") if $@;
            } elsif ($new_A->{$disk2assign}->{pool} eq 'NONE'
                  && $old_A->{$disk}->{sysid} == $sysidA) {
                eval {
                    disk_assign(
                        conn  => $opts{fcA},
                        disk  => $disk2assign,
                        sysid => $sysidA,
                        force => 1,
                    );
                };
                logWarn("cleanup: $@") if $@;
            } elsif ($new_A->{$disk2assign}->{pool} eq 'NONE'
                  && $old_A->{$disk}->{sysid} == $sysidB) {
                $disk_assign = _search_disk_serial($new_B, $sn1);
                eval {
                    disk_assign(
                        conn  => $opts{fcB},
                        disk  => $disk_assign,
                        sysid => $sysidB,
                        force => 1,
                    );
                };
                logWarn("cleanup: $@") if $@;
            } elsif ($new_A->{$disk2assign}->{sysid} != $sysidA
                  && $new_A->{$disk2assign}->{sysid} != $sysidB
                  && $old_A->{$disk}->{sysid} == $sysidA) {
                eval {
                    disk_assign(
                        conn  => $opts{fcA},
                        disk  => $disk2assign,
                        sysid => $sysidA,
                        force => 1,
                    );
                };
                logWarn("cleanup: $@") if $@;
            } elsif ($new_A->{$disk2assign}->{sysid} != $sysidA
                  && $new_A->{$disk2assign}->{sysid} != $sysidB
                  && $old_A->{$disk}->{sysid} == $sysidB) {
                $disk_assign = _search_disk_serial($new_B, $sn1);
                eval {
                    disk_assign(
                        conn  => $opts{fcB},
                        disk  => $disk_assign,
                        sysid => $sysidB,
                        force => 1,
                    );
                };
                logWarn("cleanup: $@") if $@;
            }
        }
    }
}

#---------------------------------------------------------------------

=head2 setup_aggr

    setup_aggr(fc        => $fc,
               aggr_name => $aggr,
               aggr_size => $AGGR_SIZE,
               raidtype  => 'raid4'|'raid_dp',
               timeout   => 7200
               [,force   => 0|1 ]);

Usage: setup_aggr(fc        => $conn,
                  aggr_name => 'my_aggr',
                  aggr_size => 3,
                  raidtype  => 'raid4',
                  timeout   => 7200,
                  force     => 1);

This routine checks if the aggr_name exists and verifies that it is online.
If the aggregate does not exists, create it creates the aggregate.

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=item aggr_name

(Required) : Name of aggregate to create

=item aggr_size

(Required) : Number of disks to use for the aggregate

=item raidtype

(Required) : Specify the raid type. The value must be either
             'raid4' or 'raid_dp'

=item timeout

(Required) : Specify the timeout for aggregate creation.

=item force

(Optional) : Force aggregate creation
Defaults to 1.

=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub setup_aggr(%;) {
    my (%opts) = validate (
        @_,
        {
            fc         => { type  => OBJECT },
            aggr_name  => { type  => SCALAR },
            aggr_size  => { type  => SCALAR },
            raidtype   => { type  => SCALAR },
            timeout    => { type  => SCALAR },
            force      => { type  => SCALAR, default => 1 },
        }
    );
    my ($status_obj);

    if (aggr_exists($opts{fc}, $opts{aggr_name})) {
        $status_obj = aggr_status($opts{fc}, $opts{aggr_name});
        if ($status_obj->state() ne 'online') {
            aggr_online($opts{fc}, $opts{aggr_name});
        }
    } else {
        logStep("Creating aggregate $opts{aggr_name}");
        aggr_create(
            $opts{fc},
            $opts{aggr_name},
            $opts{aggr_size},
            force    => $opts{force},
            raidtype => $opts{raidtype},
            timeout  => $opts{timeout},
        ) or croak $@;
    }
    return;
}

#---------------------------------------------------------------------

=head2 setup_vol

    setup_vol(fc        => $fc,
              aggr_name => $aggr_name,
              vol_name  => $vol_name,
              flex_size => $flex_size);

Usage: setup_vol(fc        => $conn,
                 aggr_name => 'my_aggr',
                 vol_name  => 'my_vol,
                 flex_size => '30m');

This routine checks to see if the given volume exists and purges it if it
does. The routine then creates a flex_size volume on the given aggregate.

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=item aggr_name

(Required) : Name of aggregate to use when creating a volume.

=item vol_name

(Required) : Name of volume to create.

=item flex_size

(Required) : Size of volume to create.

=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub setup_vol(%;) {
    my (%opts) = validate (
        @_,
        {
            fc         => { type  => OBJECT },
            aggr_name  => { type  => SCALAR },
            vol_name   => { type  => SCALAR },
            flex_size  => { type  => SCALAR },
        }
    );
    my ($vol_exist);

    $vol_exist = vol_exists($opts{fc}, $opts{vol_name});
    if ($vol_exist) {
        vol_purge($opts{fc}, $opts{vol_name});
        croak $@ if $@;
        $vol_exist = 0;
    }
    if (!$vol_exist) {
        logStep("Creating volume $opts{vol_name}");

        vol_create_flex(
            $opts{fc},
            $opts{vol_name},
            $opts{aggr_name},
            $opts{flex_size},
        ) or croak $@;
    }
    return;
}

#---------------------------------------------------------------------

=head2 system_disk_info

    my $disk_info = system_disk_info(fc => $fc);

This routine is used to get disk information for the system.
It returns a hash reference having as keys the following:

    volumes    => Hash reference of trad volumes
    aggregates => Hash reference of aggregates
    spares     => Hash reference of spare disks
    partner    => Hash reference of partner disks
    broken     => Hash reference of broken disks

The hash for both volumes and aggregates have information about
the plex, raid_group, options and disks.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

A hash reference for the disk information for the system
or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub system_disk_info(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );
    my (@sections, $section, $type, $name, $disks, $disk, $vol_h, $out);
    my (@plex_sections, $plex_section, $plex, $plex_h, $raid_group, $rgroup_h);
    my %ret = (
        volumes    => {},
        aggregates => {},
        spares     => {},
        partner    => {},
        broken     => {},
    );

    # Send command to filer
    $out = sayEOF($opts{fc}, 'sysconfig -r');

    # Remove extra new lines within sections
    if (!($out =~ s/\n\n\s*(RAID Disk)/\n$1/g)) {
        croak $out;
    }

    # Include multiple plexes within sections
    $out =~ s/\n\n\s*(Plex\s)/\n$1/g;

    # Split sections divided by an empty line
    @sections = split /\n{2,}/, $out;

    foreach $section (@sections) {
        if ($section =~ /^(Volume|Aggregate)\s+(\S+)\s+\(([^\)]+)\)\s+\(([^\)]+)\)/) {
            # This is a trad volume or an aggregate
            $type = $1 eq 'Volume' ? 'volumes' : 'aggregates';
            $name = $2;
            # Get the volume/aggregate options
            $ret{$type}{$name} = {
                options => [split(/,\s*/, $3), split(/,/, $4)],
            };
            $vol_h = $ret{$type}{$name};

            # Get the sections for each of the plexes
            @plex_sections = split /\s*Plex\s+/, $section;
            # Remove volume/aggregate information
            shift @plex_sections;

            # Process all plexes for this volume/aggregate
            foreach $plex_section (@plex_sections) {
                $plex_section =~ /(\S+)\s+\(([^\(]+)\)/;
                $plex = $1;

                # Get the plex and its options
                $vol_h->{plexes}{$plex}{options} = [split(/,\s*/, $2)];
                $plex_h = $vol_h->{plexes}{$plex};

                # Get the raid group and its options
                $plex_section =~ /\sRAID group\s+(\S+)\s+\(([^\(]+)\)/;
                $raid_group = $1;
                $plex_h->{raid_group}{$raid_group} = {
                    options => [split(/,\s*/, $2)],
                };
                $rgroup_h = $plex_h->{raid_group}{$raid_group};

                # Get disks for this volume/aggregate
                $disks = _parse_disk_info($plex_section);
                foreach $disk (keys %$disks) {
                    $rgroup_h->{disk}{$disk} = $disks->{$disk};
                    $vol_h->{disks}{$disk} = $disks->{$disk};
                }
            }
        } elsif ($section =~ /^(Pool\d|Spare disks)/) {
            # This is the spare disks section
            $section =~ s/Spare disks.*\n//;
            $ret{spares} = { %{$ret{spares}}, %{_parse_disk_info($section)} };

        } elsif ($section =~ /^(Partner|Broken disks)/) {
            # This is either the partner or broken disks section
            $type = $1 eq 'Partner' ? 'partner' : 'broken';
            $ret{$type} = _parse_disk_info($section);
        }
    }

    return \%ret;
}

#---------------------------------------------------------------------

=head2 get_sysid_info

    my $sysid_info = get_sysid_info(fc => $fc);

This routine returns a hash reference of the sysid for
both the system and the partner.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

A hash reference of the sysid for both the system and the partner
or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_sysid_info(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );
    my (@list, $out);

    # Send command to filer
    $out = sayEOF($opts{fc}, 'sysconfig');

    # Parse the sysid information
    @list = ($out =~ /(?:System|partner) ID:\s+(\d+)\s+\(([^\)]+)\)/g);

    return {
        System => {
            sysid => $list[0],
            name  => $list[1],
        },
        Partner => {
            sysid => $list[2],
            name  => $list[3],
        },
    };
}

#---------------------------------------------------------------------

=head2 get_root_vol

    my $vol_name = get_root_vol(fc => $fc);

This routine returns the name of the root volume.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

The name of the root volume or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_root_vol(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );
    my ($vol_name, $out);

    # Send command to filer
    $out = sayEOF($opts{fc}, 'df -V -k /');

    # Get the name of root volume and check for errors
    ($vol_name) = ($out =~ m{^/vol/([^\s\/]+)}m);
    croak $out if !$vol_name;

    return $vol_name;
}

#---------------------------------------------------------------------

=head2 get_vol_disks

    my $vol_disk_info = get_vol_disks(fc => $fc,
                                      vol => $vol);

This routine returns a hash reference of the disks
associated with the given volume.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=item vol

(Required) : Volume name to get the disks for.

=back

=item Returns

A hash reference of the disks associated with the given volume
or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_vol_disks(%;) {
    my (%opts) = validate (
        @_,
        {
            fc  => { type => OBJECT },
            vol => { type => SCALAR },
        }
    );
    my ($disk_info, $vol_obj, $ret);
    my $vol = $opts{vol};

    # Get the volume type and container if a flex vol
    $vol_obj = vol_status($opts{fc}, $vol);

    # Get the system disk info
    $disk_info = system_disk_info(fc => $opts{fc});

    if ($vol_obj->type eq 'flex') {
        # Flex volume: get the disks of aggregate containing the volume
        $vol = $vol_obj->parent;
        $ret = $disk_info->{aggregates}{$vol}{disks};
    } else {
        # Trad volume: get the disks of the volume
        $ret = $disk_info->{volumes}{$vol}{disks};
    }

    return $ret;
}

#---------------------------------------------------------------------

=head2 get_root_disks

    my $root_disk_info = get_root_disks(fc => $fc);

This routine returns a hash reference of the disks
associated with the root volume.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

A hash reference of the disks associated with the root volume
or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_root_disks(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );
    my ($root_vol, $disk_info, $vol_obj, $ret);

    # Get the root volume
    $root_vol = get_root_vol(fc => $opts{fc});

    # Get the disks for the root volume
    return get_vol_disks(fc => $opts{fc}, vol => $root_vol);
}

#---------------------------------------------------------------------
# Internal function
# It parses the disk info from the output of 'sysconfig -r'
#---------------------------------------------------------------------
sub _parse_disk_info($;) {
    my $section = shift;
    my (%ret, @list, @used, @phys, $status);

    # Verify this is a disk section and remove the heading
    if ($section =~ s/^.*---\s*\n//s) {
        # Process one line at a time
        foreach (split /\n/, $section) {
            # Remove disk copy tag if any
            s/->//;
            # Get the status and remove it from the section
            $status = s/\s*\(([^\)]+)\)// ? $1 : '';

            # Get all fields
            @list = split ' ';
            # Deal with space separated words in the 'RAID Disk' field
            # by verifying the format on the 'Device' field and checking
            # for the right number of fields
            while (($list[1] !~ /^\S+\.\d+$/) && ($list[1] !~ /^\S+\.\d+\.\S+$/)) {
                last if @list < 11;
                $list[0] = shift(@list) . ' ' . $list[0];
            }
            next if @list < 11;

            # Get the MB and blocks off the used and phys fields
            @used = split /\//, $list[9];
            @phys = split /\//, $list[10];

            # Information for this disk
            $ret{$list[1]} = {
                raid_disk => $list[0],
                ha        => $list[2],
                shelf     => $list[3],
                bay       => $list[4],
                channel   => $list[5],
                pool      => $list[6],
                type      => $list[7],
                rpm       => $list[8],
                used_mb   => $used[0],
                used_blks => $used[1],
                phys_mb   => $phys[0],
                phys_blks => $phys[1],
                status    => $status,
            };
        }
    }

    return \%ret;
}

#---------------------------------------------------------------------

=head2 zero_disks_install

    zero_disks_install(fc => $fc);

This routine zeros out he disks on a filer and then creates a newfs on the
filer. This routine boots to the special boot menu and intializes all disks
and runs basic setup.

The following named arguments are accepted:

=over

=item Arguments

=over

=item filer

(Required) : Hostspec of cluster filer.

=item filer_conn

(Required) : Filer connection type.

=item cmd_timeout

(Required) : NATE say timeout for issued commands.

=item conn_timeout

(Required) : NATE connection setup timeout.

=item location

(Optional) : Location of the filer to use for basic_setup.

=item vol_type

(Optional) : Type of volume to be created for the root volume.
             Can be 'flex' | 'trad' and defaults to 'trad'.
=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub zero_disks_install(%;) {
    my (%opts) = validate (
        @_,
        {
            filer        => { type => SCALAR },
            filer_conn   => { type => SCALAR },
            cmd_timeout  => { type => SCALAR, regex    => qr/^\d+$/ },
            conn_timeout => { type => SCALAR, regex    => qr/^\d+$/ },
            location     => { type => SCALAR, default  => ''},
            vol_type     => { type => SCALAR, default  => 'trad',
                                regex => qr/^(trad|flex)$/ },
        }
    );

    my ($dns_options, $fc, %filer_ifcs, $fh, $ifc, $ifconfig,
        $netstat, $option, $out, %partner_ips, $root_vol_type, $route,
        @servers, $value);

    my %global_ops = (
        'nis.domainname' => undef,
        'nis.servers'    => undef,
        'dns.domainname' => undef,
    );

    my $filer_zero_option = 4;

    if ($opts{vol_type} eq 'trad') {
        $root_vol_type = 't';
    } else {
        $root_vol_type = 'f';
    }

    # Get connection to filer
    $fh = host($opts{filer});
    $fc = connect($opts{filer_conn}, -timeout, $opts{conn_timeout});
    $fc->timeout($opts{cmd_timeout});

    # Get sysconfig information because it helps for debugging failed installs
    logStep('Get sysconfig information');
    filer_sysconfig_a_check($fc);

    # Get ip information from the filer
    logStep('Get ip information from the filer');
    $ifconfig = get_ifconfig_info($fc);

    foreach $ifc (keys %{$ifconfig}) {
        if ($ifc eq 'lo') {
            next;
        }
        $filer_ifcs{"IFC_$ifc"} = $ifc;
        if (defined($ifconfig->{$ifc}->{ip_addr})) {
            $filer_ifcs{"IP_$ifc"} = $ifconfig->{$ifc}->{ip_addr};
        }
        if (defined($ifconfig->{$ifc}->{netmask})) {
            $filer_ifcs{"MASK_$ifc"} = $ifconfig->{$ifc}->{netmask};
        }
        if (defined($ifconfig->{$ifc}->{partner_ip_addr})) {
            $partner_ips{"PARTNER_$ifc"} = $ifconfig->{$ifc}->{partner_ip_addr};
        }
    }

    # Get basic DNS and NIS options
    logStep('Get basic DNS and NIS options');
    foreach $option (keys %global_ops) {
        $value = filer_options_get($fc, $option);
        $global_ops{$option} = $value;
    }

    # DNS parameters
    if (defined($global_ops{'dns.domainname'})) {
        $global_ops{'dns'} = 'y';
    } else {
        $global_ops{'dns'}            = 'n';
        $global_ops{'dns.domainname'} = '';
    }

    # NIS parameters
    if (defined($global_ops{'nis.domainname'})) {
        $global_ops{'nis'} = 'y';
    } else {
        $global_ops{'nis'}            = 'n';
        $global_ops{'nis.domainname'} = '';
    }

    # Get DNS server(s)
    # Only works on IPV4
    logStep('Get DNS server(s)');
    $dns_options = dns_get_options(conn => $fc);

    if ($dns_options->{resolv_conf} =~ /nameserver\s+(\d+\.\d+\.\d+\.\d+)/i) {
        @servers =
        ($dns_options->{resolv_conf} =~ /nameserver\s+(\d+\.\d+\.\d+\.\d+)/ig);
        $global_ops{'dns.servers'} = join ",",@servers;
    } else {
        $global_ops{'dns.servers'} = '';
    }

    # Get filer timezone
    logStep('Get filer timezone');
    $out = sayEOF($fc, 'timezone');
    if ($out =~ /Current\s+time\s+zone\s+is\s+(\S+)/i) {
        $global_ops{'timezone'} = $1;
    } else {
        $global_ops{'timezone'} = 'GMT';
    }

    # Get gateway and default ifc
    logStep('Get gateway and default ifc');
    # SNMP parameters
    $netstat = filer_netstat_rn(fc => $fc);
    foreach $ifc (keys %{$netstat}) {
        foreach $route (@{$netstat->{$ifc}}) {
            if ($route->{destination} eq 'default') {
                $global_ops{'gateway'} = $route->{gateway};
                last;
            }
        }
        if (defined($global_ops{'gateway'})) {
            last;
        }
    }

    # If there is no gateway setup, we have no way of communicating with
    # the filer
    if (!defined($global_ops{'gateway'})) {
        croak("Configure filer with a gateway before running ".
              "zero_disks_install");
    }

    # use console connection for setup
    $fh = host($opts{filer});
    $fc = connect('#console', -timeout, $opts{conn_timeout});
    $fc->timeout($opts{cmd_timeout});

    # Create new file system on the filer
    logStep('Create new file system on the filer');
    newfs(
        $fh,
        root_vol_type => $root_vol_type,
        prefix        => "FILER",
        command       => $filer_zero_option,
        timeout       => 1800,
    ) or croak("newfs failed on filer: $@");

    # Begin basic setup
    logStep('Begin basic setup');
    basic_setup(
        $fc,
        prefix          =>"FILER",
        issue_setup_cmd => 0,
        timeout         => 1800,
        setup           => {
            GATEWAY         => $global_ops{'gateway'},
            LOCATION        => $opts{location},
            DNS             => $global_ops{'dns'},
            DNS_DOMAIN      => $global_ops{'dns.domainname'},
            DNS_SERVER_LIST => $global_ops{'dns.servers'},
            NIS             => $global_ops{'nis'},
            NIS_DOMAIN      => $global_ops{'nis.domainname'},
            NIS_SERVER_LIST => $global_ops{'nis.servers'},
            TIMEZONE        => $global_ops{'timezone'},
            %filer_ifcs,
            %partner_ips },
    ) or croak("basic_setup failed on filer: $@");

    # Make sure the filer is up
    $fh = host($opts{filer});
    transit_to($fh, 'UP');
    croak("Could not bring FILER up after zero disk install: $@") if ($@);
}

#---------------------------------------------------------------------

=head2 partner_mode

    partner_mode([ enable => 0|1 ]);

This function enables partner mode for subsequent calls to any disk
command function.

The following named arguments are accepted:

=over

=item Arguments

=over

=item enable

(Optional) : Set to 1 to enable partner mode or 0 to disable it.
Default is 1.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub partner_mode(%;) {
    my (%opts) = validate (
        @_,
        {
            enable => { type => SCALAR, default => 1, regex => qr/^[01]$/ },
        }
    );
    $partner_mode = $opts{enable};
    return;
}

#---------------------------------------------------------------------

=head2 maint_mode

    maint_mode([ enable => 0|1 ]);

This function enables maintenance mode for subsequent calls to any disk
command function.

The following named arguments are accepted:

=over

=item Arguments

=over

=item enable

(Optional) : Set to 1 to enable maintenance mode or 0 to disable it.
Default is 1.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub maint_mode(%;) {
    my (%opts) = validate (
        @_,
        {
            enable => { type => SCALAR, default => 1, regex => qr/^[01]$/ },
        }
    );
    $maint_mode = $opts{enable};
    return;
}

#---------------------------------------------------------------------

=head2 get_notes

    $notes = get_notes();

This function returns the note (or info) messages from the last disk
command function.

=over

=item Returns

The note messages from the last disk command function.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_notes() {
    return defined $disk_cmd_note ? $disk_cmd_note : '';
}

#---------------------------------------------------------------------
# Internal function
# Pre-process output of command to remove all '\r' characters
# and remove all "NOTE:" messages while saving them in the package
# variable for later use
#---------------------------------------------------------------------
sub _check_for_note(\$;) {
    my $out = shift;
    my @notes;

    # Remove '\r'
    $$out =~ s/\r+//g;

    # Remove all "NOTE:" messages saving them in the package variable
    while ($$out =~ s/^(NOTE:.*)\n?//m) {
        push @notes, $1;
    }

    # Save notes into package variable
    $disk_cmd_note = join("\n", @notes);

    return;
}

#---------------------------------------------------------------------
# Internal function
# This is function used to restore the sanown environment during cleanup
#---------------------------------------------------------------------
sub _search_disk_serial($$;) {
    my ($href, $serial) = @_;
    my ($disk);

    foreach $disk (keys %{$href}) {
        if ($href->{$disk}->{serial_number} eq $serial) {
            return $disk;
        }
    }
}

#---------------------------------------------------------------------
# Internal function
# This is a wrapper to either sayEOF or say depending on the
# third parameter 'maint' (maintenance mode). If the maint parameter
# is true it uses Tharn say because sayEOF sends a 'echo' command
# after the command and in maintenance mode the 'echo' command is
# not supported
#---------------------------------------------------------------------
sub _say($$$$;@) {
    my ($conn, $cmd, $maint, $partner, @options) = @_;
    my ($out, $qcmd);

    # Prefix partner to the command
    $cmd = "partner $cmd" if (defined $partner ? $partner : $partner_mode);

    if (defined $maint ? $maint : $maint_mode) {
        # Maintenance mode, make sure to use the maintenance prompt
        # and swallow all previous output before sending the command
        unshift @options, (
            '-prompt' => $maint_mode_prompt,
            '-swallow',
        );
        $out = (Tharn::say($conn, $cmd, @options))[2];
    } else {
        # Normal mode
        # Make sure to swallow all previous output before sending the command
        Tharn::say($conn, '', '-nocr', '-swallow', '-blind');
        $out = sayEOF($conn, $cmd, @options);
    }

    # Remove echo of command from output
    $qcmd = quotemeta($cmd);
    $out =~ s/.*^\s*$qcmd\r*\n*//ms;

    # Chop console messages
    _chop_messages($out);

    return $out;
}

#---------------------------------------------------------------------
# Internal function
# This function chops all console messages having the following format:
#   Day Mon day# hh:mm:ss TMZ [any message]: description
#---------------------------------------------------------------------
sub _chop_messages($;) {
    # Remove syslog messages
    $_[0] =~ s{
               ^(?:Mon|Tue|Wed|Thu|Fri|Sat|Sun)\s+
                (?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s+
                \d+\s+
                \d+:\d+:\d+\s+
                \S{3}\s+
                \[[^\]]+\]:\s
                .*\r*\n*
              }{}mgx;

    # Remove DBG messages
    $_[0] =~ s/^DBG:.*\r*\n*//mg;

    return;
}

=head2 get_disk_of_same_type 

   $type_disks = get_disk_of_same_type($disks, $disk_type);

This function combines the given type of disks in an array and 
returns the final array to user.

The following named arguments are accepted:

=over

=item Arguments

=over

=item disks 

(Required) : Array ref of disks obtained from system_disk_info().

=item disk type

(Required) : Type of the disks ( SATA, MSATA, FCAL ...)

=back

=item Returns

Array of disks of same type.

=back

=for html <hr />

=cut

sub get_disk_of_same_type {
    my $disks = shift;
    my $disk_type = shift;

    my @disks_list ;

    foreach my $disk ( @{$disks} ) {
        foreach my $key ( keys %{$disk} ) {
            my $type  = $disk->{$key}->{type};
            if ( $type eq $disk_type ) {
                my %hash ;
                $hash{$key} = $disk->{$key};
                push ( @disks_list , \%hash);
            }
         }
    }

    return \@disks_list

}
=head2 get_disk_of_same_type

   $boab_disks = get_disk_from_same_carrier($disks);

This function combines the disks belong to same carrirer in an array
returns the final array to user.

The following named arguments are accepted:

=over

=item Arguments

=over

=item disks

(Required) : Array ref of disks obtained from system_disk_info().

=back

=item Returns

Array of disks hash.

=back

=for html <hr />

=cut

sub get_disk_from_same_carrier {
    my $disks = shift;

    my @array_of_disks = @{$disks};
    my $array_length = scalar @array_of_disks;
    my $i ;
    my $j;
    my $key1;
    my $key2;
    my $disk1;
    my $disk2;
    my @same_carrier;

    for ( $i = 0 ; $i < $array_length ; $i++  ) {
         my @list1 = keys %{$array_of_disks[$i]};
         $key1 = $list1[0];
         $disk1 = $array_of_disks[$i];
        for ( $j= $i+ 1 ; $j <= ( $array_length -1 ); $j++ ) {
           my @list2 = keys %{$array_of_disks[$j]};
           $key2 = $list2[0];
           $disk2 = $array_of_disks[$j];
           my $bay = $disk2->{$key2}->{bay};
           if ( ( $disk1->{$key1}->{bay} == $disk2->{$key2}->{bay} ) && ( $disk1->{$key1}->{ha} eq $disk2->{$key2}->{ha} ) && ( $disk1->{$key1}->{shelf} == $disk2->{$key2}->{shelf} )){
               my %hash;
               my $tmp = $array_of_disks[$i+1];
               $array_of_disks[$i+1] = $array_of_disks[$j];
               $array_of_disks[$j] = $tmp;
               if ( $key1 =~ /L1$/ ) {
                   $hash{"L1"} = $key1 ;
                   $hash{"L2"} = $key2 ;
               }else {
                    $hash{"L2"} = $key1;
                    $hash{"L1"} = $key2;
              }
               push (@same_carrier , \%hash );

            }
        }
    }
    return \@same_carrier;

}

1;

