# this is basically -*- perl -*-
# $Id: VolUtils.pm,v 1.2 2008/05/16 15:27:16 emartinez Exp $
#
## @summary	VolUtils - volume and aggregate library routines to support
##		CFO test cases
## @author	dl-cf-qa@netapp.com
## @status	private
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

VolUtils - volume library routines to support CFO test cases

=head1 SYNOPSIS

  use cluster::VolUtils qw(vol_mirror aggr_mirror);

=head1 DESCRIPTION

VolUtils - volume and aggregate library routines to support CFO test cases

=head1 STATUS

  Private

=head1 AUTHOR

  dl-cf-qa@netapp.com
  $Id: VolUtils.pm,v 1.2 2008/05/16 15:27:16 emartinez Exp $
  Last revised $Date: 2008/05/16 15:27:16 $

=head1 COPYRIGHT

  Copyright 2008, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package cluster::VolUtils;

use strict;
use warnings;

use Tharn;
use LogReader;
use Carp qw(croak);
use base qw(Exporter);
use SayWrap qw(sayEOF);
use Volume qw(vol_status);
use Aggregate qw(aggr_status);
use SPARSE::Util::Logger qw(logStep);
use Admin::License qw(is_licensed license);
use Params::Validate qw(validate OBJECT SCALAR);

our @EXPORT_OK = qw(
    vol_mirror
    aggr_mirror
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '1.0';

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#---------------------------------------------------------------------
# Local variables
#---------------------------------------------------------------------

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub vol_mirror(%);
sub aggr_mirror(%);

#---------------------------------------------------------------------

=head2 vol_mirror

    my $plex = vol_mirror(fc => $fc, vol => '/testvol_trad');

This routine issues the 'vol mirror <volname>' command on the filer and waits
for the volume to be mirrored. This only works for trad volumes. For flex
volumes 'aggr mirror' must be used.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=item vol

(Required) : Volume to be mirrored.

=item timeout

(Optional) : Timeout to wait for volume to be mirrored.

=back

=item Returns

Returns mirrored plex name or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub vol_mirror(%) {
    my (%opts) = validate (
        @_,
        {
            fc      => { type => OBJECT },
            vol     => { type => SCALAR },
            timeout => { type => SCALAR , default => '7200'},
        }
    );
    my ($endtime, $fc, $log, $out, $status_obj,
        $text, $timeout, $vol);

    $fc = $opts{fc};
    $vol = $opts{vol};
    $timeout = $opts{timeout};

    $log = LogReader->new($fc);
    $log->start(['messages']);

    $status_obj = vol_status($fc, $vol);
    if ($status_obj->type() eq 'flex') {
        croak "'vol mirror' only works with trad volumes.";
    }

    if (!is_licensed($fc, 'syncmirror_local')) {
        license($fc, 'syncmirror_local');
        croak($@) if $@;
    }

    $out = sayEOF(
        $opts{fc},
        "vol mirror $vol -f",
    );

    if ($out =~ /vol\s+mirror:\s+/i) {
        croak "vol mirror error: $out";
    }

    $endtime = time() + $opts{timeout};
    while (time() < $endtime) {
        $status_obj = vol_status($fc, $vol);
        if ($status_obj->status() =~ /mirror\s+degraded|resyncing/i) {
            logStep("Sleeping 120s and waiting for volume to be mirrored");
            sleep(120);
        } elsif ($status_obj->status() =~ /mirrored/i) {
            last;
        } else {
            croak "Did not get expected status information for $vol: ".
                  $status_obj->status();
        }
    }

    $log->stop();
    $text = $log->{'logs'}{'messages'}->to_string;
    if ($text =~ /(\/$vol\/plex\d+)\s+\(PLEX\):\s+raid\s+state\s+change/i) {
        return $1;
    } else {
        croak "Could not get mirrored plex name from logs";
    }
}

#---------------------------------------------------------------------

=head2 aggr_mirror

    my $plex = aggr_mirror(fc => $fc, aggr => '/testaggr');

This routine issues the 'aggr mirror <aggr>' command on the filer and waits
for the aggregate to be mirrored.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=item aggr

(Required) : Volume to be mirrored.

=item timeout

(Optional) : Timeout to wait for volume to be mirrored.

=back

=item Returns

Returns mirrored plex name or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub aggr_mirror(%) {
    my (%opts) = validate (
        @_,
        {
            fc      => { type => OBJECT },
            aggr    => { type => SCALAR },
            timeout => { type => SCALAR , default => '7200'},
        }
    );
    my ($aggr, $endtime, $fc, $log, $out, $status_obj,
        $text, $timeout);

    $fc = $opts{fc};
    $aggr = $opts{aggr};
    $timeout = $opts{timeout};

    $log = LogReader->new($fc);
    $log->start(['messages']);

    if (!is_licensed($fc, 'syncmirror_local')) {
        license($fc, 'syncmirror_local');
        croak($@) if $@;
    }

    $out = sayEOF(
        $opts{fc},
        "aggr mirror $aggr -f",
    );

    if ($out =~ /aggr\s+mirror:\s+/i) {
        croak "aggr mirror error: $out";
    }

    $endtime = time() + $opts{timeout};
    while (time() < $endtime) {
        $status_obj = aggr_status($fc, $aggr);
        if ($status_obj->status() =~ /mirror\s+degraded|resyncing/i) {
            logStep("Sleeping 120s and waiting for aggregate to be mirrored");
            sleep(120);
        } elsif ($status_obj->status() =~ /mirrored/i) {
            last;
        } else {
            croak "Did not get expected status information for $aggr: ".
                  $status_obj->status();
        }
    }

    $log->stop();
    $text = $log->{'logs'}{'messages'}->to_string;
    if ($text =~ /(\/$aggr\/plex\d+)\s+\(PLEX\):\s+raid\s+state\s+change/i) {
        return $1;
    } else {
        croak "Could not get mirrored plex name from logs";
    }
}
1;
