# $Id$
#
#  Tharndoc comments:
#

## @summary Module used for modification of GX/BR/RR volumes
## @status reviewed
## @author Michael Eddy
## @pod here

=head1 NAME

    VolumeModify.pm

=head1 DESCRIPTION

    This will provide volume modification routines for GX/BR/RR.

=head1 EXAMPLES

    TODO

=head1 COPYRIGHT

    "Copyright 2008, Network Appliance Inc., All Rights Reserved."

=head1 STATUS
    Shared

=head1 AUTHOR

    Michael Eddy

=head1 ROUTINES

=cut

# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDE_VOLUMEMODIFY}) {
    my $newModule = $ENV{OVERRIDE_VOLUMEMODIFY};
    delete $ENV{OVERRIDE_VOLUMEMODIFY};
    print "The main VolumeModify module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package VolumeModify;
use strict;
use TracePkg;
use Data::Dumper;
use Server;
use Params::Common;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF UNDEF);
my $Trace = new TracePkg();
my $Server = new Server();
my $self;

# Class Data, only accessible from this class
{
    my $refCount = 0;

    sub incRef {
        $refCount++;
    }
    sub getRef {
        return $refCount;
    }
}

sub new
{
    $Trace->entry();
    my $class = shift(@_);

    # Check to see if this class has already been declared. If so, then we can
    # just return its reference.

    if ( getRef() ) {
        $Trace->exit(msg => "Package has already been declared. Returning...");
        return $self;
    }
    $self->{Volume_info};   # a place to store some global info about our volume

    # Bless and return
    bless ($self, $class);
    incRef();

    $Trace->debug(msg=>"Dumping the init'ed VolumeModify object config:\n".Dumper($self));
    $Trace->exit();
    return $self;
}

#
# VolumeModify::state
#
#

=head2 state

   VolumeModify::state(vserver=>$vserver, volume=>$volume, state=>$state);

   This method will change the state of the given volume to the specified value.
   
=over

=item state

   (Required) : The value to which the volume state will be set [online restricted offline force-online];

=item vserver

   (Required) : The vserver on which the volume lives

=item volume

   (Required if no uuid given) : The name of the volume who's state is to be modified

=item uuid

   (Required if no volume given) : The uuid of the volume who's state is to be modified

=back

=cut

sub state
{
    $Trace->entry();
    my $orig_state;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $output;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::state params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::state requires either a volume or uuid param!");
    }

    # Get the current "state" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_state = _current_value($self, uuid=>$volume_uuid, param_name=>"state");
    $Trace->debug(msg=>"The current \"state\" of $volume_uuid is: $orig_state");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_state;
	$Trace->trace(msg=>"Saving the orig \"state\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
    }

    # Modify the volume
    _change_state($self, uuid=>$volume_uuid, state=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"state");
    # the state of our volume will be returned as "online" if we set it to "force-online"
    if ($params{argument} eq "force-online") { 
	$params{argument} = "online";
	$Trace->trace(msg=>"Reset \$params{argument} to $params{argument} since we set our volume to \"force-online\"");
    }
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the state of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the state of $volume_uuid to $post_mod_state");
    }
    
    $Trace->exit();
    return;
} # end state

sub group
{
    $Trace->entry();
    my $orig_group;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::group params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::group requires either a volume or uuid param!");
    }

    # Get the current "group" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_group = _current_value($self, uuid=>$volume_uuid, param_name=>"group");
    $Trace->debug(msg=>"The current \"group\" of $volume_uuid is: $orig_group");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_group;
	$Trace->trace(msg=>"Saving the orig \"group\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
    }

    # Modify the volume
    _change_group($self, uuid=>$volume_uuid, group=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"group");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the group of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the group of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end group


sub user
{
    $Trace->entry();
    my $orig_user;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::user params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::user requires either a volume or uuid param!");
    }

    # Get the current "user" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_user = _current_value($self, uuid=>$volume_uuid, param_name=>"user");
    $Trace->debug(msg=>"The current \"user\" of $volume_uuid is: $orig_user");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_user;
	$Trace->trace(msg=>"Saving the orig \"user\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
    }

    # Modify the volume
    _change_user($self, uuid=>$volume_uuid, user=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"user");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the user of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the user of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end user


sub percent_snapshot_space
{
    $Trace->entry();
    my $orig_pct;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::percent_snapshot_space params:\n".Dumper(\%params));

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::percent_snapshot_space requires either a volume or uuid param!");
    }

    # Get the current "percent-snapshot-space" of our volume. If it hasn't already been saved to
    # our object, then save it as orig_value
    $orig_pct = _current_value($self, uuid=>$volume_uuid, param_name=>"percent-snapshot-space");
    $Trace->debug(msg=>"The current \"percent-snapshot-space\" of $volume_uuid is: $orig_pct");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_pct;
	$Trace->trace(msg=>"Saving the orig \"percent-snapshot-space\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
    }

    # If we have a DELTA, add DELTA to orig value to get our new value
    if ($params{arg_type} eq "DELTAS") {
	$Trace->trace(msg=>"Using DELTA value to determine new value.");
	# Remove the '%' from the value returned by server and from the DELTA so we
	# can do the addition
	my $o = rm_perc($orig_pct);
	my $d = rm_perc($params{argument});
	$Trace->trace(msg=>"Adding delta, $d, to current value, $o");
	# Reset $params{argument} to the new value
	$params{argument} = $d + $o;
    }

    # Append a '%' to our value 
    $params{argument} = add_perc($params{argument});
    if ($params{argument} !~ /%$/) {
	$params{argument} = $params{argument}.'%';
	$Trace->trace(msg=>"Appended \'%\' to our value: $params{argument}");
    }

    # Modify the volume
    _change_percent_snapshot_space($self, uuid=>$volume_uuid, "percent-snapshot-space"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"percent-snapshot-space");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the percent-snapshot-space of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the percent-snapshot-space of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end percent_snapshot_space 


sub max_autosize
{
    $Trace->entry();
    my $orig_max;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::max_autosize params:\n".Dumper(\%params));

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::max_autosize requires either a volume or uuid param!");
    }

    # Get the current "max-autosize" of our volume. If it hasn't already been saved to
    # our object, then save it as orig_value
    $orig_max = _current_value($self, uuid=>$volume_uuid, param_name=>"max-autosize");
    $Trace->debug(msg=>"The current \"max-autosize\" of $volume_uuid is: $orig_max");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_max;
	$Trace->trace(msg=>"Saving the orig \"max-autosize\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
    }

    # Convert argument and orig value to a common base (Bytes)
    $params{argument} = conv2bytes($self, value => $params{argument});
    $orig_max = conv2bytes($self, value => $orig_max);
    
    # If we have a DELTA, add DELTA to orig value to get our new value
    if ($params{arg_type} eq "DELTAS") {
	$Trace->trace(msg=>"Adding DELTA, $params{argument}, to current value, $orig_max");
	# Reset $params{argument} to the new value
	$params{argument} = $params{argument} + $orig_max;
	$Trace->trace(msg=>"Based on DELTA value, reset \$params{argument} to $params{argument}");
    }

    # Modify the volume
    _change_max_autosize($self, uuid=>$volume_uuid, "max-autosize"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"max-autosize");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the max-autosize of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the max-autosize of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end max_autosize 


sub stripe_width
{
    $Trace->entry();
    my $orig_sw;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::stripe_width params:\n".Dumper(\%params));

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::stripe_width requires either a volume or uuid param!");
    }

    my $vol_style = _get_vol_style($self, uuid=>$params{volume_uuid});
    if ($vol_style ne "striped") {
	$Trace->warn(msg=>"Can not modify the stripe width of a flex volume");
    }

    # Get the current "stripe-width" of our volume. If it hasn't already been saved to
    # our object, then save it as orig_value
    $orig_sw = _current_value($self, uuid=>$volume_uuid, param_name=>"stripe-width");
    $Trace->debug(msg=>"The current \"stripe-width\" of $volume_uuid is: $orig_sw");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_sw;
	$Trace->trace(msg=>"Saving the orig \"stripe-width\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
    }

    # Convert argument and orig value to a common base (Bytes)
    $params{argument} = conv2bytes($self, value => $params{argument});
    $orig_sw = conv2bytes($self, value => $orig_sw);
    
    # If we have a DELTA, add DELTA to orig value to get our new value
    if ($params{arg_type} eq "DELTAS") {
	$Trace->trace(msg=>"Adding DELTA, $params{argument}, to current value, $orig_sw");
	# Reset $params{argument} to the new value
	$params{argument} = $params{argument} + $orig_sw;
	$Trace->trace(msg=>"Based on DELTA value, reset \$params{argument} to $params{argument}");
    }

    # Modify the volume
    _change_stripe_width($self, uuid=>$volume_uuid, "stripe-width"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"stripe-width");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the stripe-width of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the stripe-width of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end stripe_width 


sub size
{
    $Trace->entry();
    my $orig_size;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::size params:\n".Dumper(\%params));

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::size requires either a volume or uuid param!");
    }

    # Get the current "size" of our volume. If it hasn't already been saved to
    # our object, then save it as orig_value
    $orig_size = _current_value($self, uuid=>$volume_uuid, param_name=>"size");
    $Trace->debug(msg=>"The current \"size\" of $volume_uuid is: $orig_size");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_size;
	$Trace->trace(msg=>"Saving the orig \"size\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
    }

    # Convert argument and orig value to a common base (Bytes)
    $params{argument} = conv2bytes($self, value => $params{argument});
    $orig_size = conv2bytes($self, value => $orig_size);
    
    # If we have a DELTA, add DELTA to orig value to get our new value
    if ($params{arg_type} eq "DELTAS") {
	$Trace->trace(msg=>"Adding DELTA, $params{argument}, to current value, $orig_size");
	# Reset $params{argument} to the new value
	$params{argument} = $params{argument} + $orig_size;
	$Trace->trace(msg=>"Based on DELTA value, reset \$params{argument} to $params{argument}");
    }

    # Modify the volume
    _change_size($self, uuid=>$volume_uuid, "size"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"size");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the size of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the size of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end size 


sub maxdir_size
{
    $Trace->entry();
    my $orig_maxdir_size;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::maxdir_size params:\n".Dumper(\%params));

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::maxdir_size requires either a volume or uuid param!");
    }

    # Get the current "maxdir-size" of our volume. If it hasn't already been saved to
    # our object, then save it as orig_value
    $orig_maxdir_size = _current_value($self, uuid=>$volume_uuid, param_name=>"maxdir-size");
    $Trace->debug(msg=>"The current \"maxdir-size\" of $volume_uuid is: $orig_maxdir_size");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_maxdir_size;
	$Trace->trace(msg=>"Saving the orig \"maxdir-size\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
    }

    # Convert argument and orig value to a common base (Bytes)
    $params{argument} = conv2bytes($self, value => $params{argument});
    $orig_maxdir_size = conv2bytes($self, value => $orig_maxdir_size);
    
    # If we have a DELTA, add DELTA to orig value to get our new value
    if ($params{arg_type} eq "DELTAS") {
	$Trace->trace(msg=>"Adding DELTA, $params{argument}, to current value, $orig_maxdir_size");
	# Reset $params{argument} to the new value
	$params{argument} = $params{argument} + $orig_maxdir_size;
	$Trace->trace(msg=>"Based on DELTA value, reset \$params{argument} to $params{argument}");
    }

    # Modify the volume
    _change_maxdir_size($self, uuid=>$volume_uuid, "maxdir-size"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"maxdir-size");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the maxdir-size of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the maxdir-size of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end maxdir_size 


sub autosize_increment
{
    $Trace->entry();
    my $orig_ai;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::autosize_increment params:\n".Dumper(\%params));

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::autosize_increment requires either a volume or uuid param!");
    }

    # Get the current "autosize-increment" of our volume. If it hasn't already been saved to
    # our object, then save it as orig_value
    $orig_ai = _current_value($self, uuid=>$volume_uuid, param_name=>"autosize-increment");
    $Trace->debug(msg=>"The current \"autosize-increment\" of $volume_uuid is: $orig_ai");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_ai;
	$Trace->trace(msg=>"Saving the orig \"autosize-increment\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
    }

    # Convert argument and orig value to a common base (Bytes)
    $params{argument} = conv2bytes($self, value => $params{argument});
    $orig_ai = conv2bytes($self, value => $orig_ai);
    
    # If we have a DELTA, add DELTA to orig value to get our new value
    if ($params{arg_type} eq "DELTAS") {
	$Trace->trace(msg=>"Adding DELTA, $params{argument}, to current value, $orig_ai");
	# Reset $params{argument} to the new value
	$params{argument} = $params{argument} + $orig_ai;
	$Trace->trace(msg=>"Based on DELTA value, reset \$params{argument} to $params{argument}");
    }

    # Modify the volume
    _change_autosize_increment($self, uuid=>$volume_uuid, "autosize-increment"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"autosize-increment");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the autosize-increment of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the autosize-increment of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end autosize_increment


sub files
{
    my $orig_files;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::files params:\n".Dumper(\%params));


    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::files requires either a volume or uuid param!");
    }

    # Get the current "files" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_files = _current_value($self, uuid=>$volume_uuid, param_name=>"files");
    $Trace->debug(msg=>"The current \"files\" of $volume_uuid is: $orig_files");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_files;
	$Trace->trace(msg=>"Saving the orig \"files\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # If we have a DELTA, add DELTA to orig value to get our new value
    if ($params{arg_type} eq "DELTAS") {
	$Trace->trace(msg=>"Adding DELTA, $params{argument}, to current value, $orig_files");
	# Reset $params{argument} to the new value
	$params{argument} = $params{argument} + $orig_files;
	$Trace->trace(msg=>"Based on DELTA value, reset \$params{argument} to $params{argument}");
    }

    # Modify the volume
    _change_files($self, uuid=>$volume_uuid, "files"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"files");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the files of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the files of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end files


sub max_write_alloc_blocks
{
    $Trace->entry();
    my $orig_mb;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::max_write_alloc_blocks params:\n".Dumper(\%params));

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::max_write_alloc_blocks requires either a volume or uuid param!");
    }

    # Get the current "max-write-alloc-blocks" of our volume. If it hasn't already been saved to
    # our object, then save it as orig_value
    $orig_mb = _current_value($self, uuid=>$volume_uuid, param_name=>"max-write-alloc-blocks");
    $Trace->debug(msg=>"The current \"max-write-alloc-blocks\" of $volume_uuid is: $orig_mb");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_mb;
	$Trace->trace(msg=>"Saving the orig \"max-write-alloc-blocks\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
    }

    ## Convert argument and orig value to a common base (Bytes)
    #$params{argument} = conv2bytes($self, value => $params{argument});
    #$orig_mb = conv2bytes($self, value => $orig_mb);
    
    # If we have a DELTA, add DELTA to orig value to get our new value
    if ($params{arg_type} eq "DELTAS") {
	$Trace->trace(msg=>"Adding DELTA, $params{argument}, to current value, $orig_mb");
	# Reset $params{argument} to the new value
	$params{argument} = $params{argument} + $orig_mb;
	$Trace->trace(msg=>"Based on DELTA value, reset \$params{argument} to $params{argument}");
    }

    # Modify the volume
    _change_max_write_alloc_blocks($self, uuid=>$volume_uuid, "max-write-alloc-blocks"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"max-write-alloc-blocks");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the max-write-alloc-blocks of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the max-write-alloc-blocks of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end max_write_alloc_blocks 


sub snapshot_space_used
{
    $Trace->entry();
    my $orig_pct;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::snapshot_space_used params:\n".Dumper(\%params));

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::snapshot_space_used requires either a volume or uuid param!");
    }

    # Get the current "snapshot-space-used" of our volume. If it hasn't already been saved to
    # our object, then save it as orig_value
    $orig_pct = _current_value($self, uuid=>$volume_uuid, param_name=>"snapshot-space-used");
    $Trace->debug(msg=>"The current \"snapshot-space-used\" of $volume_uuid is: $orig_pct");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_pct;
	$Trace->trace(msg=>"Saving the orig \"snapshot-space-used\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
    }

    # If we have a DELTA, add DELTA to orig value to get our new value
    if ($params{arg_type} eq "DELTAS") {
	$Trace->trace(msg=>"Using DELTA value to determine new value.");
	# Remove the '%' from the value returned by server and from the DELTA so we
	# can do the addition
	my $o = rm_perc($orig_pct);
	my $d = rm_perc($params{argument});
	$Trace->trace(msg=>"Adding delta, $d, to current value, $o");
	# Reset $params{argument} to the new value
	$params{argument} = $d + $o;
    }

    # Append a '%' to our value 
    $params{argument} = add_perc($params{argument});
    if ($params{argument} !~ /%$/) {
	$params{argument} = $params{argument}.'%';
	$Trace->trace(msg=>"Appended \'%\' to our value: $params{argument}");
    }

    # Modify the volume
    _change_snapshot_space_used($self, uuid=>$volume_uuid, "snapshot-space-used"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"snapshot-space-used");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the snapshot-space-used of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the snapshot-space-used of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end snapshot_space_used


sub security_style
{
    $Trace->entry();
    my $orig_ss;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::security_style params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::security_style requires either a volume or uuid param!");
    }

    # Get the current "security-style" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_ss = _current_value($self, uuid=>$volume_uuid, param_name=>"security-style");
    $Trace->debug(msg=>"The current \"security-style\" of $volume_uuid is: $orig_ss");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_ss;
	$Trace->trace(msg=>"Saving the orig \"security-style\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
    }

    # Modify the volume
    _change_security_style($self, uuid=>$volume_uuid, "security-style"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"security-style");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the security-style of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the security-style of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end security_style


sub junction_active
{
    my $orig_ja;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::junction_active params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::junction_active requires either a volume or uuid param!");
    }

    # Get the current "junction-active" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_ja = _current_value($self, uuid=>$volume_uuid, param_name=>"junction-active");
    $Trace->debug(msg=>"The current \"junction-active\" of $volume_uuid is: $orig_ja");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_ja;
	$Trace->trace(msg=>"Saving the orig \"junction-active\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_junction_active($self, uuid=>$volume_uuid, "junction-active"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"junction-active");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the junction-active of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the junction-active of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end junction_active


sub comment
{
    my $orig_comment;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::comment params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::comment requires either a volume or uuid param!");
    }

    # Get the current "comment" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_comment = _current_value($self, uuid=>$volume_uuid, param_name=>"comment");
    $Trace->debug(msg=>"The current \"comment\" of $volume_uuid is: $orig_comment");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_comment;
	$Trace->trace(msg=>"Saving the orig \"comment\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_comment($self, uuid=>$volume_uuid, "comment"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"comment");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the comment of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the comment of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end comment


sub policy
{
    my $orig_policy;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::policy params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::policy requires either a volume or uuid param!");
    }

    # Get the current "policy" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_policy = _current_value($self, uuid=>$volume_uuid, param_name=>"policy");
    $Trace->debug(msg=>"The current \"policy\" of $volume_uuid is: $orig_policy");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_policy;
	$Trace->trace(msg=>"Saving the orig \"policy\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_policy($self, uuid=>$volume_uuid, "policy"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"policy");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the policy of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the policy of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end policy


sub antivirus_on_access_policy
{
    my $orig_policy;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::antivirus_on_access_policy params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::antivirus_on_access_policy requires either a volume or uuid param!");
    }

    # Get the current "antivirus-on-access-policy" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_policy = _current_value($self, uuid=>$volume_uuid, param_name=>"antivirus-on-access-policy");
    $Trace->debug(msg=>"The current \"antivirus-on-access-policy\" of $volume_uuid is: $orig_policy");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_policy;
	$Trace->trace(msg=>"Saving the orig \"policy\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_antivirus_on_access_policy($self, uuid=>$volume_uuid, "antivirus-on-access-policy"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"antivirus-on-access-policy");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the antivirus-on-access-policy of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the antivirus-on-access-policy of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end antivirus_on_access_policy


sub unix_permissions
{
    my $orig_unix_permissions;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::unix-permissions params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::unix-permissions requires either a volume or uuid param!");
    }

    # Get the current "unix-permissions" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_unix_permissions = _current_value($self, uuid=>$volume_uuid, param_name=>"unix-permissions");
    $Trace->debug(msg=>"The current \"unix-permissions\" of $volume_uuid is: $orig_unix_permissions");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_unix_permissions;
	$Trace->trace(msg=>"Saving the orig \"unix-permissions\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_unix_permissions($self, uuid=>$volume_uuid, "unix-permissions"=>$params{argument});

    # Verify the modification
    #   Get our current value
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"unix-permissions");
    #   If we were passed octal notation, the symbolic notation returned by server for our
    #   new value will be translated to octal before comparison.
    #
    #   If we were passed symbolic notation, we must handle the 4th "group" that is always
    #   returned by server.
    if ($params{argument} =~ /\d{4}/) {
	# we have to translate the unix notation to octal since user passed in 4 digit octal
	$Trace->trace(msg=>"User passed in 4 digit octal notation. Translating...");
	$post_mod_state = unixnotation2octal($self, perm=>$post_mod_state, fourdigit=>"true");
    } elsif ($params{argument} =~ /\d{3}/) {
	# we have to translate the unix notation to octal since user passed in 3 digit octal
	$Trace->trace(msg=>"User passed in 3 digit octal notation. Translating...");
	$post_mod_state = unixnotation2octal($self, perm=>$post_mod_state, fourdigit=>"");
    } elsif ($params{argument} =~ /^[-rwx]{9}$/) {
	# we have to account for the special group that server will return by removing
	# the leading '---'
	$post_mod_state =~ s/^-{3}//;
	$Trace->trace(msg=>"User passed in standard symbolic notation, ignoring special group returned by server.");
    } elsif ($params{argument} =~ /^[-rwx]{12}$/) {
	# no adjustment needed - this is the format that server will return
	$Trace->trace(msg=>"User passed in 4 position symbolic notation.");
    } else {
	# we shouldn't get here
	$Trace->fatal(msg=>"Permission format is not recognized: $params{argument}");
    }
    $Trace->debug(msg=>"Before verification, \$post_mod_state is: $post_mod_state");
    #  Compare current value with what it SHOULD be
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the unix-permissions of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the unix-permissions of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end unix_permissions


sub autosize
{
    my $orig_autosize;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::autosize params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::autosize requires either a volume or uuid param!");
    }

    # Get the current "autosize" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_autosize = _current_value($self, uuid=>$volume_uuid, param_name=>"autosize");
    $Trace->debug(msg=>"The current \"autosize\" of $volume_uuid is: $orig_autosize");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_autosize;
	$Trace->trace(msg=>"Saving the orig \"autosize\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_autosize($self, uuid=>$volume_uuid, autosize=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"autosize");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the autosize of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the autosize of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end autosize


sub space_guarantee
{
    my $orig_sg;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::space_guarantee params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::space_guarantee requires either a volume or uuid param!");
    }

    # Get the current "space-guarantee" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_sg = _current_value($self, uuid=>$volume_uuid, param_name=>"space-guarantee");
    $Trace->debug(msg=>"The current \"space-guarantee\" of $volume_uuid is: $orig_sg");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_sg;
	$Trace->trace(msg=>"Saving the orig \"space-guarantee\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_space_guarantee($self, uuid=>$volume_uuid, "space-guarantee"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"space-guarantee");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the space-guarantee of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the space-guarantee of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end space_guarantee



sub min_readahead
{
    my $orig_mr;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::min_readahead params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::min_readahead requires either a volume or uuid param!");
    }

    # Get the current "min-readahead" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_mr = _current_value($self, uuid=>$volume_uuid, param_name=>"min-readahead");
    $Trace->debug(msg=>"The current \"min-readahead\" of $volume_uuid is: $orig_mr");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_mr;
	$Trace->trace(msg=>"Saving the orig \"min-readahead\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_min_readahead($self, uuid=>$volume_uuid, "min-readahead"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"min-readahead");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the min-readahead of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the min-readahead of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end min_readahead


sub atime_update
{
    my $orig_atime;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::atime_update params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::atime_update requires either a volume or uuid param!");
    }

    # Get the current "atime-update" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_atime = _current_value($self, uuid=>$volume_uuid, param_name=>"atime-update");
    $Trace->debug(msg=>"The current \"atime-update\" of $volume_uuid is: $orig_atime");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_atime;
	$Trace->trace(msg=>"Saving the orig \"atime-update\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_atime_update($self, uuid=>$volume_uuid, "atime-update"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"atime-update");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the atime-update of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the atime-update of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end atime_update


sub snapdir_access
{
    my $orig_snapdir_access;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::snapdir_access params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::snapdir_access requires either a volume or uuid param!");
    }

    # Get the current "snapdir-access" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_snapdir_access = _current_value($self, uuid=>$volume_uuid, param_name=>"snapdir-access");
    $Trace->debug(msg=>"The current \"snapdir-access\" of $volume_uuid is: $orig_snapdir_access");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_snapdir_access;
	$Trace->trace(msg=>"Saving the orig \"snapdir-access\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_snapdir_access($self, uuid=>$volume_uuid, "snapdir-access"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"snapdir-access");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the snapdir-access of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the snapdir-access of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end snapdir_access


sub lookup_cache
{
    my $orig_lookup_cache;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::lookup_cache params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }
   
    # Warn if this is not a striped volume
    my $vol_style = _get_vol_style($self, uuid=>$params{volume_uuid});
    if ($vol_style ne "striped") {
	$Trace->warn(msg=>"Can not modify the lookup-cache of a flex volume");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::lookup_cache requires either a volume or uuid param!");
    }

    # Get the current "lookup_cache" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_lookup_cache = _current_value($self, uuid=>$volume_uuid, param_name=>"lookup-cache");
    $Trace->debug(msg=>"The current \"lookup-cache\" of $volume_uuid is: $orig_lookup_cache");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_lookup_cache;
	$Trace->trace(msg=>"Saving the orig \"lookup-cache\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_lookup_cache($self, uuid=>$volume_uuid, "lookup-cache"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"lookup-cache");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the lookup-cache of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the lookup-cache of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end lookup-cache 


sub concurrency
{
    my $orig_concurrency;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::concurrency params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # Warn if this is not a striped volume
    my $vol_style = _get_vol_style($self, uuid=>$params{volume_uuid});
    if ($vol_style ne "striped") {
	$Trace->warn(msg=>"Can not modify the concurrency of a flex volume");
    }


    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::concurrency requires either a volume or uuid param!");
    }

    # Get the current "concurrency" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_concurrency = _current_value($self, uuid=>$volume_uuid, param_name=>"concurrency");
    $Trace->debug(msg=>"The current \"concurrency\" of $volume_uuid is: $orig_concurrency");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_concurrency;
	$Trace->trace(msg=>"Saving the orig \"concurrency\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_concurrency($self, uuid=>$volume_uuid, "concurrency"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"concurrency");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the concurrency of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the concurrency of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end concurrency 


sub index_dir_enabled
{
    my $orig_index_dir_enabled;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::index_dir_enabled params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::index_dir_enabled requires either a volume or uuid param!");
    }

    # Get the current "index-dir-enabled" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_index_dir_enabled = _current_value($self, uuid=>$volume_uuid, param_name=>"index-dir-enabled");
    $Trace->debug(msg=>"The current \"index-dir-enabled\" of $volume_uuid is: $orig_index_dir_enabled");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_index_dir_enabled;
	$Trace->trace(msg=>"Saving the orig \"index-dir-enabled\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_index_dir_enabled($self, uuid=>$volume_uuid, "index-dir-enabled"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"index-dir-enabled");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the index-dir-enabled of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the index-dir-enabled of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end index-dir-enabled 


sub optimize
{
    my $orig_optimize;
    my $volume_name;
    my $volume_uuid;
    my $post_mod_state;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    argument => {type => SCALAR},
	    arg_type => {type => SCALAR},
	    vserver => {type => SCALAR, optional => 1},
	    volume => {type => SCALAR, optional => 1, depends => ['vserver']},
	    "volume_uuid" => {type => SCALAR, optional => 1},
	}
    );
    $Trace->debug(msg=>"Dumping VolumeModify::optimize params:\n".Dumper(\%params));

    # This method does not support arg_type of DELTA
    if ($params{arg_type} ne "VALUES") {
	$Trace->fatal(msg=>"This method only supports an arg_type of VALUES; arg_type is $params{arg_type}");
    }

    # Warn if this is not a striped volume
    my $vol_style = _get_vol_style($self, uuid=>$params{volume_uuid});
    if ($vol_style ne "striped") {
	$Trace->warn(msg=>"Can not modify the optimize value of a flex volume");
    }


    # We want to deal with volume uuid since the name can change so if we were passed the
    # name, get the uuid.
    if ($params{"volume_uuid"}) {
	$volume_uuid = $params{"volume_uuid"};
    } elsif ($params{volume}) {
	$volume_name = $params{volume};
	$volume_uuid = volume_uuid_lookup($self, name=>$volume_name, vserver=>$params{vserver});
    } else {
	# Ensure that we have been passed either a volume or uuid
	$Trace->fatal(msg=>"VolumeModify::optimize requires either a volume or uuid param!");
    }

    # Get the current "optimize" of our volume. If it hasn't already been saved to our object,
    # then save it as orig_value
    $orig_optimize = _current_value($self, uuid=>$volume_uuid, param_name=>"optimize");
    $Trace->debug(msg=>"The current \"optimize\" of $volume_uuid is: $orig_optimize");
    if (!defined $self->{Volume_info}{orig_value}){
	$self->{Volume_info}{orig_value} = $orig_optimize;
	$Trace->trace(msg=>"Saving the orig \"optimize\" of $volume_uuid: $self->{Volume_info}{orig_value} to our object");
	$Trace->debug(msg=>"Dumping \$self->Volume_info:\n".Dumper($self->{Volume_info}));
    }

    # Modify the volume
    _change_optimize($self, uuid=>$volume_uuid, "optimize"=>$params{argument});

    # Verify the modification
    $post_mod_state = _current_value($self, uuid=>$volume_uuid, param_name=>"optimize");
    if ($post_mod_state ne $params{argument}) {
	$Trace->fatal(msg=>"Failed to set the optimize of $volume_uuid to $params{argument}! It is currently $post_mod_state");
    } else {
	$Trace->comment(msg=>"Successfully set the optimize of $volume_uuid to $post_mod_state");
    }

    $Trace->exit();
    return;
} # end optimize 


sub volume_name_lookup
{
    $Trace->entry();
    my $name;
    my $vserver;
    my @info;
    my $output;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR,},
	}
    );
    
    $Trace->trace(msg=>"Looking up the volume name and vserver of $params{uuid}");
    $output = $Server->volume_show(uuid=>$params{uuid});
    $Trace->debug(msg=>"Dumping output of volume_show:\n".Dumper($output));
    # Get our volume name
    $name = $$output[0]{volume};
    if (!$name) {
	$Trace->fatal(msg=>"Failed to find the volume name of $params{uuid}!");
    }
    # Get our vserver name
    $vserver = $$output[0]{vserver};
    if (!$vserver) {
	$Trace->fatal(msg=>"Failed to find the vserver for $params{uuid}!");
    }
    $Trace->comment(msg=>"The vserver:volume name for $params{uuid} is: $vserver:$name");

    # Build up our array to be returned
    @info = ($vserver, $name);
    $Trace->debug(msg=>"Dumping \@info:".Dumper(\@info));
    
    $Trace->exit();
    return \@info;
} # end volume_name_lookup


sub volume_uuid_lookup 
{
    $Trace->entry();
    my $uuid;
    my $output;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    vserver => {type => SCALAR},
	    name => {type => SCALAR},
	}
    );
  
    # Get our uuid
    $output = $Server->volume_show(vserver=>$params{vserver}, volume=>$params{name});
    $uuid = $$output[0]{uuid};

    # Check that we have a valid hex uuid
    if ($uuid !~ /^[a-f0-9]+/) {
	$Trace->fatal(msg=>"Failed to find a valid uuid for volume: $params{name}!: $uuid"); 
    }
    
    $Trace->comment(msg=>"$params{vserver} $params{name} has a uuid of: $uuid");

    $Trace->exit();
    return $uuid;
} # end volume_uuid_lookup

sub _current_value
{
    $Trace->entry();
    my $self = shift;
    my $value;
    my $output;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    param_name => {type => SCALAR},
	}
    );
    
    # Get the current value of our param_name
    $output = $Server->volume_show(uuid=>$params{uuid});
    $value = $$output[0]{$params{param_name}};

    $Trace->debug(msg=>"The value of $params{param_name} for $params{uuid} is: $value");
    $Trace->exit();
    return $value;
} # end _current_value;


sub _change
{
    $Trace->entry();
    my $self = shift;
    my $vserver;
    my $volume_name;
    my $output;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    attribute => {type => SCALAR},
	    value => {type => SCALAR},
	}
    );
    $Trace->debug(msg=>"Dumping _change params:\n".Dumper(\%params));

    ## TRYING EXTENDED QUERIES instead of looking up the vserver/name
    ## Lookup our vserver and volume name
    #$Trace->trace(msg=>"Looking up volume name of: $params{uuid}");
    #my $info = volume_name_lookup($self, uuid=>$params{uuid});
    #$vserver = shift @$info;
    #$volume_name = shift @$info;
    #if ((!$vserver) || (!$volume_name)) {
#	$Trace->fatal(msg=>"Failed to find the vserver or volume name of $params{uuid}!");
#    }

    # Modify the volume's attribute
    $Trace->trace(msg=>"Issuing volume modify $params{attribute}=>$params{value} for $params{uuid}");
    $Server->cli(cmd=>"volume modify {-uuid $params{uuid}} -$params{attribute} $params{value}");
    #$Trace->trace(msg=>"Issuing volume modify vserver=>$vserver, volume=>$volume_name, $params{attribute}=>$params{value}");
    #$Server->volume_modify(vserver=>$vserver, volume=>$volume_name, $params{attribute}=>$params{value});

    $Trace->exit();
    return;
} # end _change


sub _change_state
{
    # This internal subroutine change's the volume's state.  The main point of this being
    # a seperate routine is so that we can call it with the uuid and look up the vserver 
    # and volume name at the last minute. This minimizes the window during which the 
    # names can be changed out from under us.
    $Trace->entry();
    my $self = shift;
    my $vserver;
    my $volume_name;
    my $output;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    state => {type => SCALAR},
	}
    );

    ## Lookup our vserver and volume name
    #$Trace->trace(msg=>"Looking up volume name of: $params{uuid}");
    #my $info = volume_name_lookup($self, uuid=>$params{uuid});
    #$vserver = shift @$info;
    #$volume_name = shift @$info;
    #if ((!$vserver) || (!$volume_name)) {
#	$Trace->fatal(msg=>"Failed to find the vserver or volume name of $params{uuid}!");
#    }

    # We must unmount the volume before taking it offline or restricted
    if (($params{state} eq "offline") || ($params{state} eq "restricted")) {
	$Trace->trace(msg=>"Volume must be unmounted before taking it $params{state}");
	_unmount($self, uuid=>$params{uuid});
    }

    # Modify the volume's state
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"state", value=>$params{state});
    #$Trace->trace(msg=>"Issuing volume modify vserver=>$vserver, volume=>$volume_name, state=>$params{state}");
    #$Server->volume_modify(vserver=>$vserver, volume=>$volume_name, state=>$params{state});

    # Remount the volume if we are bringing it online and it had been unmounted
    $Trace->debug(msg=>"Dumping Volume_info:\n".Dumper($self->{Volume_info}));
    if (($params{state} =~ /online/) && ($self->{Volume_info}{mounted} eq "false")) {
	$Trace->comment(msg=>"Volume was previously unmounted, remounting now.");
	_mount($self, uuid=>$params{uuid});
    } else {
	$Trace->trace(msg=>"Volume does not need to be remounted");
    }
    
    $Trace->exit();
    return;
} # end _change_state


sub _change_group
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    group => {type => SCALAR},
	}
    );

    # Modify the volume's group
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"group", value=>$params{group});

    $Trace->exit();
    return;
} # end _change_group


sub _change_user
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    user => {type => SCALAR},
	}
    );

    # Modify the volume's user
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"user", value=>$params{user});

    $Trace->exit();
    return;
} # end _change_user


sub _change_percent_snapshot_space
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "percent-snapshot-space" => {type => SCALAR},
	}
    );

    # Modify the volume's percent-snapshot-space
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"percent-snapshot-space", value=>$params{"percent-snapshot-space"});

    $Trace->exit();
    return;
} # end _change_percent_snapshot_space


sub _change_snapshot_space_used
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "snapshot-space-used" => {type => SCALAR},
	}
    );

    # Modify the volume's snapshot-space-used 
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"snapshot-space-used", value=>$params{"snapshot-space-used"});

    $Trace->exit();
    return;
} # end _change_snapshot_space_used


sub _change_max_autosize
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "max-autosize" => {type => SCALAR},
	}
    );

    # Modify the volume's max-autosize
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"max-autosize", value=>$params{"max-autosize"});

    $Trace->exit();
    return;
} # end _change_max_autosize


sub _change_autosize_increment
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "autosize-increment" => {type => SCALAR},
	}
    );

    # Modify the volume's 
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"autosize-increment", value=>$params{"autosize-increment"});

    $Trace->exit();
    return;
} # end _change_autosize_increment



sub _change_stripe_width
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "stripe-width" => {type => SCALAR},
	}
    );

    # Modify the volume's stripe-width
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"stripe-width", value=>$params{"stripe-width"});

    $Trace->exit();
    return;
} # end _change_stripe_width


sub _change_size
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "size" => {type => SCALAR},
	}
    );

    # Modify the volume's size
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"size", value=>$params{"size"});

    $Trace->exit();
    return;
} # end _change_size


sub _change_maxdir_size
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "maxdir-size" => {type => SCALAR},
	}
    );

    # Modify the volume's maxdir-size
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"maxdir-size", value=>$params{"maxdir-size"});

    $Trace->exit();
    return;
} # end _change_maxdir_size


sub _change_max_write_alloc_blocks
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "max-write-alloc-blocks" => {type => SCALAR},
	}
    );

    # Modify the volume's max-write-alloc-blocks 
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"max-write-alloc-blocks", value=>$params{"max-write-alloc-blocks"});

    $Trace->exit();
    return;
} # end _change_max_write_alloc_blocks


sub _change_files
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    files => {type => SCALAR},
	}
    );

    # Modify the volume's files
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"files", value=>$params{files});

    $Trace->exit();
    return;
} # end _change_files



sub _change_security_style
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "security-style" => {type => SCALAR},
	}
    );

    # Modify the volume's security-style
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"security-style", value=>$params{"security-style"});

    $Trace->exit();
    return;
} # end _change_security_style


sub _change_junction_active
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "junction-active" => {type => SCALAR},
	}
    );

    # Modify the volume's junction-active
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"junction-active", value=>$params{"junction-active"});

    $Trace->exit();
    return;
} # end _change_junction_active


sub _change_comment
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    comment => {type => SCALAR},
	}
    );

    # Modify the volume's comment
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"comment", value=>$params{comment});

    $Trace->exit();
    return;
} # end _change_comment


sub _change_policy
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    policy => {type => SCALAR},
	}
    );

    # Modify the volume's policy
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"policy", value=>$params{policy});

    $Trace->exit();
    return;
} # end _change_policy


sub _change_antivirus_on_access_policy
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "antivirus-on-access-policy" => {type => SCALAR},
	}
    );

    # Modify the volume's policy
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"antivirus-on-access-policy", value=>$params{"antivirus-on-access-policy"});

    $Trace->exit();
    return;
} # end _change_antivirus_on_access_policy


sub _change_unix_permissions
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "unix-permissions" => {type => SCALAR},
	}
    );

    # Modify the volume's unix-permissions
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"unix-permissions", value=>$params{"unix-permissions"});

    $Trace->exit();
    return;
} # end _change_unix_permissions


sub _change_autosize
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    autosize => {type => SCALAR},
	}
    );

    # Modify the volume's autosize
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"autosize", value=>$params{autosize});

    $Trace->exit();
    return;
} # end _change_autosize


sub _change_space_guarantee
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "space-guarantee" => {type => SCALAR},
	}
    );

    # Modify the volume's "space-guarantee"
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"space-guarantee", value=>$params{"space-guarantee"});

    $Trace->exit();
    return;
} # end _change_space_guarantee


sub _change_min_readahead
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "min-readahead" => {type => SCALAR},
	}
    );

    # Modify the volume's "min-readahead"
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"min-readahead", value=>$params{"min-readahead"});

    $Trace->exit();
    return;
} # end _change_min_readahead


sub _change_atime_update
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "atime-update" => {type => SCALAR},
	}
    );

    # Modify the volume's "atime-update"
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"atime-update", value=>$params{"atime-update"});

    $Trace->exit();
    return;
} # end _change_atime_update


sub _change_snapdir_access
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "snapdir-access" => {type => SCALAR},
	}
    );

    # Modify the volume's "snapdir-access"
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"snapdir-access", value=>$params{"snapdir-access"});

    $Trace->exit();
    return;
} # end _change_snapdir_access


sub _change_lookup_cache
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "lookup-cache" => {type => SCALAR},
	}
    );

    # Modify the volume's "lookup-cache"
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"lookup-cache", value=>$params{"lookup-cache"});

    $Trace->exit();
    return;
} # end _change_lookup_cache


sub _change_concurrency
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "concurrency" => {type => SCALAR},
	}
    );

    # Modify the volume's "concurrency"
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"concurrency", value=>$params{"concurrency"});

    $Trace->exit();
    return;
} # end _change_concurrency


sub _change_optimize
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "optimize" => {type => SCALAR},
	}
    );

    # Modify the volume's "optimize"
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"optimize", value=>$params{"optimize"});

    $Trace->exit();
    return;
} # end _change_optimize


sub _change_index_dir_enabled
{
    $Trace->entry();
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    "index-dir-enabled" => {type => SCALAR},
	}
    );

    # Modify the volume's "optimize"
    $Trace->trace(msg=>"Calling _change");
    _change($self, uuid=>$params{uuid}, attribute=>"index-dir-enabled", value=>$params{"index-dir-enabled"});

    $Trace->exit();
    return;
} # end _change_index_dir_enabled


sub _unmount
{
    $Trace->entry();
    my $orig_mountpoint;
    my $vserver;
    my $volume_name;
    my $output;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	}
    );

    # Lookup our mountpoint
    $output = $Server->volume_show(uuid=>$params{uuid});
    $orig_mountpoint = $$output[0]{"junction-path"};
    if (!$orig_mountpoint) {
	$Trace->fatal(msg=>"Failed to find the mountpoint of $params{uuid}!");
    }
    
    # Store the original mount point in our global %Volume_info if it's not already
    if (!defined $self->{Volume_info}{orig_mountpoint}) {
	$self->{Volume_info}{orig_mountpoint} = $orig_mountpoint;
	$Trace->trace(msg=>"Storing mountpoint $self->{Volume_info}{orig_mountpoint}");
    } else {
	$Trace->trace(msg=>"We already have a mountpoint saved: $self->{Volume_info}{orig_mountpoint}");
    }
    $Trace->debug(msg=>"Dumping Volume_info:\n".Dumper($self->{Volume_info}));

    # Look up our vserver/volume name
    my $info = volume_name_lookup($self, uuid=>$params{uuid});
    $vserver = shift @$info;
    $volume_name = shift @$info;
    
    # Unmount our volume
    $Trace->comment(msg=>"Unmounting $vserver $volume_name from $orig_mountpoint");
    $Server->volume_unmount(vserver=>$vserver, volume=>$volume_name);
    
    # Update Volume_info
    $self->{Volume_info}{mounted} = "false";

    $Trace->exit();
} # end _unmount


sub _mount {
    $Trace->entry();
    my $mountpoint;
    my $vserver;
    my $volume_name;
    my $output;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	    mountpoint => {type => SCALAR, optional => 1},
	}
    );

    # TODO take active and perhaps policy-override options too!

    # Determine what our mountpoint should be
    if (!defined $params{mountpoint}) {
	$Trace->debug(msg=>"Dumping our Volume_info:\n".Dumper($self->{Volume_info}));
	$mountpoint = $self->{Volume_info}{orig_mountpoint};
	$Trace->trace(msg=>"Using \$self->{Volume_info}{orig_mountpoint} for our mountpoint.");
    } else {
	$mountpoint = $params{mountpoint};
	$Trace->trace(msg=>"Using mountpoint value passed to _mount for our mountpoint.");
    }

    # Check our mountpoint
    if ((!$mountpoint) || ($mountpoint eq "-")){
	$Trace->fatal(msg=>"We have an invalid mountpoint of: $mountpoint");
    }
   
    # Mount doesn't take extended queries so we must look up our vserver/volume name
    my $info = volume_name_lookup($self, uuid=>$params{uuid});
    $vserver = shift @$info;
    $volume_name = shift @$info;

    # Mount our volume
    $Trace->comment(msg=>"Mounting $vserver $volume_name at $mountpoint");
    $Server->volume_mount(vserver=>$vserver, volume=>$volume_name, "junction-path"=>$mountpoint);

    # Update Volume_info
    $self->{Volume_info}{mounted} = "true";

    $Trace->exit();
} # end _mount


sub unixnotation2octal
{
    $Trace->entry();
    my @groups;
    my @oct;
    my $oct;
    my $other;
    my $user;
    my $group;
    my $world;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    perm => {type => SCALAR},
	    fourdigit => {type => SCALAR, optional => 1, default => 0},
	}
    );

    # Split up our permission string
    if ($params{perm} =~ /([-rwx]{3})([-rwx]{3})([-rwx]{3})([-rwx]{3})/) {
	$other = $1;
	$user = $2;
	$group = $3;
	$world = $4;
    } else {
	$Trace->fatal(msg=>"un2oct method did not match the perm string as expected!: $params{perm}");
    }
  
    # Server is going to return "4 digit" notation, but user may only care about std 3 digit
    if ($params{fourdigit}) {
	$Trace->debug(msg=>"Using 4 digit octal notation.");
	@groups = ($other, $user, $group, $world);
    } else {
	$Trace->debug(msg=>"Using 3 digit octal notation.");
	@groups = ($user, $group, $world);
    }

    # Translate each notation to octal
    foreach my $id (@groups) {
	push @oct, _un2oct($id);
	$Trace->debug(msg=>"Dumping \@oct:".Dumper(\@oct));
    }

    # Build the octal notation string
    my $oct = join '', @oct;
    $Trace->comment(msg=>"$params{perm} translated to $oct");
    
    $Trace->exit();
    return $oct;
} # end unixnotation2octal


sub _un2oct
{
    $Trace->entry();
    my $oct = undef;
    my $value = shift;
    $Trace->debug(msg=>"Translating $value to octal");
    if (!$value) {
	$Trace->fatal(msg=>"_un2oct requires a value!");
    }

    $oct = 0 if ($value =~ /---/);
    $oct = 1 if ($value =~ /--x/);
    $oct = 2 if ($value =~ /-w-/);
    $oct = 3 if ($value =~ /-wx/);
    $oct = 4 if ($value =~ /r--/);
    $oct = 5 if ($value =~ /r-x/);
    $oct = 6 if ($value =~ /rw-/);
    $oct = 7 if ($value =~ /rwx/);

    if ($oct !~ /\d/) {
	$Trace->fatal(msg=>"Failed to match $value as expected!");
    }

    $Trace->exit();
    return $oct;
} # end _un2oct


sub rm_perc
{
    $Trace->entry();
    my $perc = shift;
    # Remove the '%' from the value provided
    if ($perc =~ /(\d+)%/) {
	$perc = $1;
	$Trace->trace(msg=>"Removed \'%\' from our value: $perc");
    } else {
	$Trace->comment(msg=>"Percentage did not have a \"%\" at the end");
    }

    $Trace->exit();
    return $perc;
}# end rm_perc


sub add_perc
{
    $Trace->entry();
    my $perc = shift;
    # Append a '%' to our value 
    if ($perc !~ /\d+%$/) {
	$perc = $perc.'%';
	$Trace->trace(msg=>"Appended \'%\' to our value: $perc");
    } else {
	$Trace->comment(msg=>"Percentage already has a \"%\" at the end");
    }

    $Trace->exit();
    return $perc;
} # end add_perc


sub conv2bytes
{
    $Trace->entry();
    my $bytes;
    my $factor;
    my $digits;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    value => {type => SCALAR},
	}
    );

    my $input = $params{value};
    $Trace->debug(msg=>"conv2bytes was passed $input");
   
    if ($input =~ /(\d+)/) {
	$digits = $1;
    }
    $factor = 1 if ($input =~ /\d+[k,K][b,B]?/);
    $factor = 2 if ($input =~ /\d+[m,M][b,B]?/);
    $factor = 3 if ($input =~ /\d+[g,G][b,B]?/);
    $factor = 4 if ($input =~ /\d+[t,T][b,B]?/);
    $factor = 5 if ($input =~ /\d+[p,P][b,B]?/);
    return $input if ($input =~ /\d+$/);  # no conv needed if it's just a number
    
    if (!$factor) {
	$Trace->fatal(msg=>"Failed to match the input!; $input");
    }

    # do the conversion
    $bytes = $digits * (1024**$factor);
    $Trace->trace(msg=>"$input was converted to $bytes bytes");

    $Trace->exit();
    return $bytes;
} # end conv2bytes


sub _get_vol_style
{
    $Trace->entry();
    my $style;
    my $self = shift;
    my %params = Params::Validate::validate (
	@_,
	{
	    uuid => {type => SCALAR},
	}
    );

    # Get our volume-style
    my $output = $Server->volume_show(uuid=>$params{uuid});
    $style = $$output[0]{"volume-style"};

    if (!$style) {
	$Trace->fatal(msg=>"Failed to get a volume-style for $params{uuid}!\n".Dumper($output));
    }

    $Trace->exit();
    return $style;
} # end _get_vol_style

1;
