## $Id: //depot/prod/test/Rsierranevada/dp/TestArch/lib/StorageConfig.pm#1 $
##
## @summary
##  Library to tie scripts into SnapMirror Test Architecture
##
## @description
##  Primarily convenience accessor methods for Storage Configurator created volumes
##
## @dependencies
##  To be useful, call these methods for systems which have been configured using
##  the storage configurator layer of the SnapMirror Test Architecture.
##  Needs LOGDIR to _not_ be stdout ie LOGDIR=<dir> and not LOGDIR=-
##
## @author
##  dl-dpsas-svl
##
###########################################################################################

=head1 NAME

Snapmirror::TestArch::StorageConfig.pm

=head1 AUTHOR

dl-dpsas-svl

=head1 DESCRIPTION

C<StorageConfig.pm> will handle fetching of precreated resources, specifically
precreated volumes by the storage configurator, sm_storage_config.thpl

=head1 SYNOPSIS

C<StorageConfig.pm> provides convenience getter methods to tie into
the snapmirror test architecture

=head2 METHODS

=over 12

=item C<getSrcVols>

Gets storage configured volumes as NACL::C::Volume objects and
places a lock on these resources on a per vserver level until ungetSrcVols is called.

Blocking call ie waits if vserver on which volumes reside is locked by another test.

=item C<ungetSrcVols>

Releases the lock on a particular vserver

=back

=cut

#package StorageConfig;
package dp::TestArch::lib::StorageConfig;
use base 'Exporter';

use strict;
use Params::Validate qw(:all);
use Fcntl qw(:DEFAULT :flock :seek);
use Tie::File;
use Tharn;

use NACL::C::Node;
use NACL::C::Vserver;
use NACL::C::Volume;
use NACL::C::Aggregate;
use NACL::STask::VolumeSnapshot;

use dp::TestArch::lib::Constants qw(:names);

use Exporter qw(import);

our @EXPORT_OK = qw( getVserver getSrcVols getDstVols isStorageConfigSrc ungetSrcVols );

sub getVserver
{
  my($node,$vs) = shift || die "Must pass node object";
  $node = ref($node) =~ /NODE|HASH/i ?
            $node : NACL::C::Node->new(name=>$node);
 
  $vs = VS_PREFIX . lc($node->node());
  $vs=~s/-/_/g;

  NACL::C::Vserver->find(command_interface=>$node,
                        filter=> { vserver=>$vs},
                        allow_empty=>1) ? return $vs : return '';
}

=head2 ungetSrcVols

Takes as required inputs a node name or node componenet object
and volume component object as reference to list or as scalar.

    ungetSrcVols(node=>'excal-h3',volumes=>\@cvols);
    ungetSrcVols(node=>'excal-h3',volumes=>$cvol1);

Releases the lock on these volume component objects.
These would usually be objects returned by getSrcVols.

=cut

sub ungetSrcVols
{
  my($node,@cvols,$lockdir);
  
  $lockdir = param('TOP_LOGDIR');

  my %args = validate_with ( params => \@_,
                             spec => {
                                 node => { type=>HASHREF|SCALAR },
                                 volumes => { type=>HASHREF|ARRAYREF,optional=>0 }
                                        } );

  $node = ref($args{'node'}) =~ /NODE|HASH/i ?
            $args{'node'} : NACL::C::Node->new(name=>$args{'node'});

  @cvols = ref($args{'volumes'}) =~ /HASH|VOLUME/i ?
            ($args{'volumes'}) : @{$args{'volumes'}};

  my @locked_vols;

  my %free_us;

  foreach my $cvol (@cvols)
  {
    my($vs,$vol) = ($cvol->get_one_state_attribute('vserver'),$cvol->get_one_state_attribute('volume'));
    if( exists $free_us{$vs} ) { push @{$free_us{$vs}},$vol }
    else { $free_us{$vs} = [ $vol ] }
  }

  foreach my $vs (keys %free_us)
  {
    my @ungetvols = @{$free_us{$vs}};

    my $o = tie @locked_vols, 'Tie::File', "$lockdir/$vs" or die $!;
    $o->flock(LOCK_EX) || die "Unable to acquire lock for unget $vs: $!";
    (tied @locked_vols)->flush;

    print "FLOCK: \t*acquired* lock for unget: " . time ."\n";
    print "FLOCK: Lock file has " . scalar(@locked_vols) . " locked vols:\n";
    print "FLOCK: " . join(",",@locked_vols) . "\n";

    @locked_vols = grep { my $vol = $_; !grep { /^\Q$vol\E$/ } @ungetvols; } @locked_vols;

    (tied @locked_vols)->flush;

    $o->flock(LOCK_UN) || die "Unable to release lock for unget $vs: $!";

    print "FLOCK: UNPOPULATED lock file with " . scalar(@ungetvols) . " unlocked vols\n";
    print "FLOCK: Lock file now contains " . scalar(@locked_vols) . " locked vols\n";
    print "FLOCK: " . join(",",@locked_vols) . "\n";
    print "FLOCK: \t*released* lock for unget: " . time ."\n";

    untie @locked_vols;
    undef $o;
  }
}

=head2 getSrcVols

my @cvols = getSrcVols(node=>'excal-h3',vserver=>'sc_vs_excal_h3',vol_count=>3)

Takes as input a node name or node componenet object and vserver name and optional
vol_names, vol_count and block_type.

vol_names can be specified if storage configured volume names are known.
If vol_names is passed as a reference to an empty list or set to undef
then the first available storage configured volume is returned.

Valid values for vol_count are positive numeric values.
vol_count is honored only if it falls within the count of vols found.
If not i.e. if vol_count is negative, zero, or greater than number of
vols found then all available volumes are returned.

Valid values for block_type are 32 or 64 for 32-bit and 64-bit.

Returns a list of storage configured volume component objects
if called in list context or the first available
volume object in scalar context.

Returns undef when called in scalar context and
no storage configured volumes exist.

If flock fails to acquire a lock on the lock file then
returns an empty list if list context is desired and undef
if scalar.

If all storage configured volumes are locked by other tests
then return an empty list if list context is desired and undef
if scalar.

=cut

sub getSrcVols
{
  my($node,$vs,$lockdir,@locked_vols,@cvols,$blk,@aggr,@ret);

  my %args = validate_with ( params => \@_,
                             spec => {  
                                 node        => { type=>HASHREF|SCALAR },
                                 vserver     => { type=>SCALAR,optional=>0 },
                                 vol_names   => { type=>ARRAYREF|UNDEF,optional=>1 },
                                 vol_count   => { type=>SCALAR,optional=>1 },
                                 block_type  => { type=>SCALAR,optional=>1 },
                                        } );

  $node = ref($args{'node'}) =~ /NODE|HASH/i ?
            $args{'node'} : NACL::C::Node->new(name=>$args{'node'});

  $vs = $args{'vserver'} || getVserver($node) || return undef;

  $lockdir = param('TOP_LOGDIR');

  $blk = $args{'block_type'};
  unless(($blk==32)||($blk==64)) { undef $blk }

  if($blk) {
      my $caggr = NACL::C::Aggregate->find(command_interface=>$node,
                                            filter=>
                                            {
                                                aggregate=>AGGR_PREFIX.'*',
                                                nodes=>[$node->node()],
                                                'block-type'=>$blk.'-bit'
                                            },
                                            allow_empty=>1);
      if($caggr) { @aggr = ('aggregate', $caggr->get_one_state_attribute('aggregate')) };
  }

  @cvols = NACL::C::Volume->find(
            'command_interface'=>$node,
            filter=> {
                vserver=>$vs,
                volume=>VOL_SRC_PREFIX.'*',
                state=>'online',@aggr},
            allow_empty=>1 );

  my $o = tie @locked_vols, 'Tie::File', "$lockdir/$vs" or die $!;
  unless($o->flock(LOCK_EX)) { return wantarray ? () : undef }
  (tied @locked_vols)->flush;

  print "FLOCK: \t*acquired* lock for get: " . time ."\n";

  print "FLOCK: Checking " . scalar(@cvols) ." vols\n";
  print "FLOCK: " . scalar(@locked_vols) . " vols already locked\n";
  print "FLOCK: " . join(',',@locked_vols) . "\n";

  @cvols = grep {
                    my $vol = $_->get_one_state_attribute('volume');
                    !grep { /^\Q$vol\E$/ } @locked_vols;
                } @cvols;

  print "FLOCK: " . scalar(@cvols) . " vols available\n";

  if(!@cvols)
  {
    (tied @locked_vols)->flush;
    $o->flock(LOCK_UN) || die "Unable to release lock: $!";
    print "FLOCK: \t*released* lock for get: 0 free vols: " . time ."\n\n";

    untie @locked_vols;
    undef $o;
    return wantarray ? () : undef;
  }

  if(ref($args{'vol_names'}) eq 'ARRAY')
  {
    if(@{$args{'vol_names'}}==0) {
      @ret = $cvols[0];
    }
    else {
      @ret = map {
                my $wanted = $_;
                my ($ret) =
                  grep {
                      !$wanted ||
                      ($_->get_one_state_attribute('volume') =~ /^\s*\Q$wanted\E\s*$/)
                  } @cvols;

                $ret ? $ret : undef
             } @{$args{'vol_names'}};

      if((@ret==1) && !defined($ret[0])) { undef @ret };
    }
  }
  elsif($args{'vol_count'})
  {
    if($args{'vol_count'}==0) { @ret = @cvols }
    elsif($args{'vol_count'}<0) { @ret = @cvols }
    elsif(($args{'vol_count'}>0)
            &&
         ($args{'vol_count'}<=@cvols))
            { @ret = @cvols[0..$args{'vol_count'}-1] }
    elsif($args{'vol_count'}>@cvols) { @ret = @cvols }
  }
  else
  {
    @ret = $cvols[0]
  }
  _restoreSrc($node,\@ret);

  foreach (@ret) { push @locked_vols, $_->get_one_state_attribute('volume') }
  (tied @locked_vols)->flush;

  print "FLOCK: Locking " . scalar(@ret) . " more vols\n";
  print "FLOCK: " . scalar(@locked_vols) . " total vols now locked\n";
  print "FLOCK: " . join(',',@locked_vols) . "\n";

  $o->flock(LOCK_UN) || die "Unable to release lock:$!";
  print "FLOCK: \t*released* lock for get: " . time ."\n\n";

  untie @locked_vols;
  undef $o;


  if(@ret) { return wantarray ? @ret : $ret[0] }
  else { return wantarray ? () : undef }
}

sub _restoreSrc
{
  my($n,$cvols,@reset_snaps) = @_;

  foreach my $cvol (@$cvols)
  {
    @reset_snaps = NACL::STask::VolumeSnapshot->find(command_interface => $n,
                                          filter => { $cvol->get_primary_keys_hash(),
                                                      snapshot=>VOL_SNAP_PREFIX.'*' },
                                          allow_empty=>1);

    if(@reset_snaps) { $reset_snaps[-1]->restore(force=>'true') }
  }
}

sub getDstVols
{
  my($node,$vs,@cvols,$blk,@aggr,@ret);

  my %args = validate_with ( params => \@_,
                             spec => {  node => { type=>HASHREF|SCALAR },
                                        vol_names => { type=>ARRAYREF|UNDEF,optional=>1 },
                                        vol_count => { type=>SCALAR,optional=>1 },
                                        block_type => { type=>SCALAR,optional=>1 },
                                        } );

  $node = ref($args{'node'}) =~ /NODE|HASH/i ? $args{'node'} : NACL::C::Node->new(name=>$args{'node'});

  $vs = getVserver($node) || return undef;

  $blk = $args{'block_type'};
  unless(($blk==32)||($blk==64)) { undef $blk }

  if($blk) {
      my $caggr = NACL::C::Aggregate->find(
                            command_interface=>$node,
                            filter=>{
                                aggregate=>AGGR_PREFIX.'*',
                                nodes=>[$node->node()],
                                'block-type'=>$blk.'-bit'
                            },
                            allow_empty=>1);

      if($caggr) { @aggr = ('aggregate', $caggr->get_one_state_attribute('aggregate')) };
  }
  @cvols = NACL::C::Volume->find( 'command_interface'=>$node,
                                    filter=>{
                                        vserver=>$vs,
                                        volume=>VOL_DST_PREFIX.'*',
                                        state=>'online',@aggr }, allow_empty=>1 );

  if(!@cvols) { return wantarray ? () : undef };
  
  if(ref($args{'vol_names'}) eq 'ARRAY')
  {


  }
  elsif($args{'vol_count'})
  {
    if($args{'vol_count'}==0) { @ret = @cvols }
    elsif($args{'vol_count'}<0) { @ret = @cvols }
    elsif(($args{'vol_count'}>0) && ($args{'vol_count'}<=@cvols))
            { @ret = @cvols[0..$args{'vol_count'}-1] }
    elsif($args{'vol_count'}>@cvols) { @ret = @cvols }
  }
  else
  {
    @ret = $cvols[0];
  }
  if(@ret) { return wantarray ? @ret : $ret[0] } else { return wantarray ? () : undef }

}

sub isStorageConfigSrc
{
  my($n,$vol,$vs,$match) = @_;
  $vs = getVserver($n) || return 0;
  $match = VOL_SRC_PREFIX;

  if(($vol =~ /^\Q$match\E\d+$/)
      &&
        NACL::C::Volume->find('command_interface'=>$n,
                                filter=>{
                                    vserver=>$vs,
                                    volume=>$vol,
                                    state=>'online' }, allow_empty=>1)
      &&
        NACL::STask::VolumeSnapshot->find(command_interface => $n,
                                          filter => {
                                                vserver=>$vs,
                                                volume=>$vol,
                                                snapshot=>VOL_SNAP_PREFIX.'*'
                                          },
                                          allow_empty=>1))
  { return 1 }
  else { return 0 }
}

1;
