# -*- perl -*-
# $Id: TestWrapper.pm,v 1.6 2005/07/27 19:29:16 tapl Exp $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
## @summary TestWrapper - Eval Wrapper Functions to simplify 
##                        testscript writing
##
## @author dl-dpgqa-automation@netapp.com
## @status Private
## @pod here

=head1 NAME

TestWrapper - Eval Wrapper Functions to simplify testscript writing

=head1 SYNOPSIS

 use dp::TestWrapper qw(pretest testcase posttest testscript);

 Example 1:

   eval {
     pretest(func => \&func, params => ['arg1', 'arg2']);
   };

   # $@ is a Result object
   logresult($@) if ($@);

 Example 2:

   testscript(pretest  => {func => \&func1},
              testcase => {func => \&func2, params => ['arg1']},
              teardown => {func => \&func3, params => ['arg2', 'arg3']});

 Example 3. Function to call uses positional parameters.

   sub call_me_last {
     my ($first, $second, $third) = @_;
     ...
   }

   eval {
     posttest(func   => \&call_me_last, 
              params => [$first, $second, $third]);
   };
   logresult($@) if ($@);

Example 4. Function to call uses named parameters.

   # usage: main_test(first=>$a, $second=>$b, third=>$c);
   sub main_test {
     my (%args) = @_;
     ...
   }

   eval {
     testcase(func   => \&main_test, 
              params => [first => a, second => $b, third => $c]);
   };
   logresult($@) if ($@);

=head1 DESCRIPTION

 Generic eval wrappers that throws the appropriate Result object
 depending on what wrapper was called.

 pretest() and posttest() wrappers will always throw a Nate::Result::Fatal
 object while the testcase() will always throw a Nate::Result::Fail object.

 testscript() is a convenient wrapper that calls the wrappers
 in the following order, pretest(), testcase(), posttest(),
 teardown callback, and automatically report to Nate the correct
 error type thrown.

=head1 ROUTINES

=cut

package dp::TestWrapper;

use strict;
use warnings;
use Tharn;
use Carp qw(croak);
use Data::Dumper qw(Dumper);
use Params::Validate qw(validate SCALAR ARRAYREF BOOLEAN CODEREF HASHREF);


BEGIN {
    Params::Validate::validation_options(
         on_fail => sub { my @msgs = @_; chomp(@msgs);
			  Carp::confess(Tharn::script_result(@msgs)) }
    );
}

use base qw(Exporter);
our @EXPORT_OK = qw(
    posttest
    pretest
    testcase
    run_test
    testscript
    testwrapper_testme
);


=head2 pretest

Example 1:

 push(@args_list, $arg1, $arg2, $arg3, ...);
 pretest(func => \&func, params => \@args_list);
 logresult($@) if ($@);

Example 2:

 pretest(func => \&func, params => ['arg1', 'arg2', 'arg3', ...]);
 logresult($@) if ($@);

 Wrapper to trap all unexpected operations that take place before the
 testcase verification phase. This wrapper will always throw a
 NATE::Result::Fatal object if something unexpected takes place.

=over

=item func

 (Required) : Code reference to the function

=item params

 (Optional) : Array reference to the list of parameters to passed to the
 function.

=back

=cut

sub pretest {
    my (%args) = validate( @_, { func => { type => CODEREF },
				 params => { type => ARRAYREF,
					     optional => 1 } } );

    $args{exit_status} = "FATAL";
    _generic_callback (%args);
}


=head2 posttest

Example 1:

 push(@args_list, $arg1, $arg2, $arg3, ...);
 posttest(func => \&func, params => \@args_list);
 logresult($@) if ($@);

Example 2:

 posttest(func => \&func, params => ['arg1', 'arg2', 'arg3', ...]);
 logresult($@) if ($@);

 Wrapper to trap all unexpected operations that take place after the
 testcase verification phase. This wrapper will always throw a
 NATE::Result::Fatal object if something unexpected takes place.

=over

=item func

 (Required) : Code reference to the function

=item params

 (Optional) : Array reference to the list of parameters to passed to the
 function.

=back

=cut

sub posttest {
    my (%args) = validate( @_, { func	=> { type => CODEREF },
				 params => { type => ARRAYREF,
					     optional => 1 } } );
    
    $args{exit_status} = "FATAL";
    _generic_callback (%args);
}


=head2 testcase

Example 1:

 push(@args_list, $arg1, $arg2, $arg3, ...);
 testcase(func => \&func, params => \@args_list);
 logresult($@) if ($@);

Example 2:

 testcase(func => \&func, params => ['arg1', 'arg2', 'arg3', ...]);
 logresult($@) if ($@);

 Wrapper to trap all unexpected operations that take place during the
 testcase verification phase. This wrapper will always report a
 NATE::Result::Fail object if something unexpected that invalidate 
 the testcase takes place.

=over

=item func

(Required) : Code reference to the function

=item params

 (Optional) : Array reference to the list of parameters to passed to the
 function.

=back

=cut

sub testcase {
    my (%args) = validate( @_, { func	=> { type => CODEREF },
				 params => { type => ARRAYREF,
					     optional => 1 } } );

    $args{exit_status} = "FAIL";
    _generic_callback (%args);
}

# Generic routine for pretest, testcase and posttest callbacks
# Additional parameter:
#	exit_status: FAIL or FATAL, fail status of callbacks.

sub _generic_callback {
    my (%args) = validate( @_, { exit_status => { type => SCALAR,
										regex => qr/^FAIL|FATAL$/ },
                                 func	=> { type => CODEREF },
				 params => { type => ARRAYREF,
					     optional => 1 } } );

    my $success = 0;
    eval {
        $args{func}->( @{$args{params}} );
        $success = 1;
    };

    unless ($success) {
        if ($@) {
	    my $err_msg = $@;
            logcomment("$args{exit_status}: $err_msg");
            logresult("INFO", "Error detected for NOTE_SPARE_ALL_ON_ERROR");
            croak ($args{exit_status} eq "FATAL"
                  ? fatal_result($err_msg)
                  : fail_result($err_msg));
        } else {
            my $msg = "$args{exit_status}: failed to complete callback or "
                      ."callback 'die' with empty \$\@";
            logcomment("$args{exit_status}:", $msg);
            logresult("INFO", "Error detected for NOTE_SPARE_ALL_ON_ERROR");
            croak ($args{exit_status} eq "FATAL"
                  ? fatal_result($msg)
                  : fail_result($msg));
        }
    }
}

=head2 run_test

Function that invokes testscript with standard functions in ::main
do_setup, do_test and do_teardown.

=cut

sub run_test {
    # use standard functions do_setup, do_test and do_teardown
    testscript(
	pretest  => { func => \&main::do_setup },
	testcase => { func => \&main::do_test },
	teardown => { func => \&main::do_teardown }
    );
}

=head2 testscript

Example:

 push(@args_list1, arg1, arg2, arg3, ...);
 push(@args_list4, arg1 => $a, arg2 => $b, arg3 => $c, ...); 

 testscript(pretest  => { func => \&func1, params => \@args_list1 },
            testcase => { func => \&func2, params => [c => $c, d => $d] },
            posttest => { func => \&func3, params => [$e, $f] },
            teardown => { func => \&func4, params => \@args_list4 });
 
 Catch-all wrapper to trap all unexpected operations. Depending on which
 phase (pretest, testcase, posttest) the code is in, it will
 throw the appropriate error type for that phase. This wrapper will
 execute code based on this order: pretest, testcase, posttest, teardown.

=over

=item pretest

 (Optional) : Hashref of parameters to be passed to pretest().

=item testcase

 (Optional) : Hashref of parameters to be passed to testcase().

=item posttest

 (Optional) : Hashref of parameters to be passed to posttest().

=item teardown

 (Optional) : Hashref of parameters. The coderef to the function
 will be executed with the list of parameters passed in (if any).

=back

=cut

sub testscript {
    my (%args) = validate( @_, { pretest  => { type => HASHREF,
					       optional => 1 },
				 testcase => { type => HASHREF,
					       optional => 1 },
				 posttest => { type => HASHREF,
					       optional => 1 },
				 teardown => { type => HASHREF,
					       optional => 1 },
				 # the following two are for testing
				 # purposes. do not document!
				 no_report => { type => BOOLEAN,
						default => 0,
						optional => 1 },
			         abortive  => { type => BOOLEAN,
						default => 0,
						optional => 1 } } );

    eval {
        # pass references of functions to these wrappers
        pretest($args{pretest})   if (exists $args{pretest});
        testcase($args{testcase}) if (exists $args{testcase});
        posttest($args{posttest}) if (exists $args{posttest});
    };

    my $rsltobj = $@;

#    logcomment($rsltobj->type() . ": " . ($rsltobj->messages())[0])
#      if ($rsltobj);

    # for testing purpose
    if (exists $args{no_report} && $args{no_report} && $rsltobj) {
	croak $rsltobj;
    }

    # call the teardown callback function.
    # XXX/tapl may need to fix this
    ($args{teardown}->{func})->( @{$args{teardown}->{params}} )
      if (exists $args{teardown});

    # for testing purpose
    if ($rsltobj && 
	exists $args{abortive} && !$args{abortive}) {
	# set the abortive level of the Result object
	$rsltobj->abortive($args{abortive});
    }

    # at this point, we're guaranteed that the
    # the $rsltobj is a Result Object since it 
    # was thrown 'correctly' by pretest(), 
    # testcase(), or posttest().
    logresult($rsltobj) if ($rsltobj);
}


=head2 testwrapper_testme

 testwrapper_testme

 Unit test function for all the test wrappers exported by this
 Perl module.

=cut

sub testwrapper_testme {
    # pretest() unit test
    eval {
 	pretest(func=>\&_die_testme, params=>['pretest()', '9876']);
    };

    my $err = $@;

    if ($err =~ /9876/) {
	logresult('PASS', 'Passed pretest() unit test');
    } else {
	logresult('FAIL', "Failed pretest(): $err");
    }

    # testcase() unit test
    eval {
	my @params;
	push(@params, 'testcase()', '1369');
	testcase(func=>\&_die_testme, params=>\@params);
    };

    $err = $@;

    if ($err =~ /1369/) {
	logresult('PASS', 'Passed testcase() unit test');
    } else {
	logresult('FAIL', "Failed testcase(): $err");
    }

    # posttest() unit test
    eval {
	posttest(func=>\&_die_testme, params=>['posttest()', '2468']);
    };

    $err = $@;

    if ($err =~ /2468/) {
	logresult('PASS', 'Passed posttest() unit test');
    } else {
	logresult('FAIL', "Failed posttest(): $err");
    }

    # testscript() unit test 1
    eval {
	testscript(testcase=>{func=>\&_die_testme,
			      params=>['testscript()', 
				       "10111213"]},
		   no_report=>1
		  );
    };

    $err = $@;
    
    if ($err =~ /10111213/) {
	logresult('PASS', 'Passed testscript() unit test');
    } else {
	logresult('FAIL', "Failed testscript(): $err");
    }

    # testscript unit test 2
    logcomment("Visually inspect that testscript() throws " .
	       "a FATAL with msg 'testscript arg: 112233'");
    testscript(pretest=>{func=>\&_die_testme,
			 params=>['testscript arg', 
				   "112233"]},
	       abortive=>0
	      );

    # testscript unit test 3
    logcomment("Visually inspect that testscript() throws " .
	       "a FAIL with msg 'testcase arg: 33445566'");
    testscript(testcase=>{func=>\&_die_testme,
			  params=>['testcase arg', 
				   "33445566"]},
	       abortive=>0
	      );

    # testscript unit test 4
    logcomment("Visually inspect that testscript() throws " .
	       "a FATAL with msg 'posttest arg: 99887766'");
    testscript(posttest=>{func=>\&_die_testme,
			  params=>['posttest arg', 
				   "99887766"]},
	       abortive=>0
	      );


    # testscript unit test 5
    eval {
	testscript(teardown=>{func=>\&_warn_testme,
			      params=>['teardown arg', 
				       "446644"]}
		  );
    };

    $err = $@;

    if ($err =~ /446644/) {
	logresult('PASS', "Passed testscript()'s teardown arg");
    } else {
	logresult('FAIL', "Failed testscript(): $err");
    }

    # testscript unit test 6
    # ensure that you can transition from one phase to another
    logcomment("Verifying that testscript() can transition from one phase" .
    " to another");
    testscript(testcase=>{func=>\&_dummy_testme},
	       pretest=>{func=>\&_dummy_testme});
}

# dummy stub for testwrapper_testme() above
sub _die_testme {
    my ($func, $magic_num) = @_;
    logresult('INFO', "Testing $func...");
    die "$func: $magic_num";
}

sub _warn_testme {
    my ($func, $magic_num) = @_;
    die (warn_result("$func: $magic_num"));
}

sub _print_testme {
    print "Hello\n";
}

sub _dummy_testme {
}

=head1 STATUS

 Private

=head1 ATTRIBUTES

 ONTAP_COMMANDS = NA

 ONTAP_ZAPI_APIS = NA

=head1 AUTHOR

 DP QA (dl-dpgqa-automation)
 $Id: TestWrapper.pm,v 1.6 2005/07/27 19:29:16 tapl Exp $
 Last revised $Date: 2005/07/27 19:29:16 $

=head1 COPYRIGHT

 Copyright 2005, Network Appliance Inc., All Rights Reserverd.

=cut




1;
