# -*- perl -*-

## $Id$
##
## @summary
##        Dump restore utility routines
##
## @description
##        List of common routines to be used by dump / restore tests
##
## @author
##        dl-dpgqa-automation@netapp.com
##

package dp::dumprestore::DumpRestore;

use strict;
use Carp;
use IO::File;
use Tharn;
use Module::Common;
use FSR::FSRutils qw (load_compat_data);
use Aggregate qw (aggr_create aggr_purge);
use Volume qw (vol_create vol_create_flex vol_purge);
use QtreeCmds qw (qtreeCreate);
use ExportFS qw (export_path unexport_path);
use SnapCmds qw (snapshot_create);
use filer qw (filer_file_destroy);
use fileSystem qw (treecompare);
use mount qw (mount umount);
use Tapeinfo qw( getTapeStatus );

require Exporter;
our @ISA = qw (Exporter);
our @EXPORT = qw ();
our @EXPORT_OK = qw (
        create_dump_vol
        drop_dump_vol
        create_dump_subdir
        create_dump_qtree

        drop_dump_localfile

        create_data
        dump_data
        restore_data

        compare_dump_data
);

#---------------------------------------------------------------
## @name ::create_dump_vol
## @synopsis subroutine to create a volume for dump-restore test
##        $fh - filer connection
##        $vol_name - Volume name
##        $aggr_name - aggregate for flexible volume,
##                   create traditional volume if not passed
##        $disk_cnt - Number of disk to create aggr/flex vol, defaults to 10
##        $vol_size - Size of the volume to create, defaults to 180m
#---------------------------------------------------------------

sub create_dump_vol {
    my ($fh, $vol_name, $aggr_name, $disk_cnt, $vol_size) = @_;

    $disk_cnt = 10 unless $disk_cnt;
    $vol_size = 180 unless $vol_size;
    if ( $aggr_name ) {
        logcomment("DumpRestore: creating aggr $aggr_name");
        aggr_create($fh, $aggr_name, "$disk_cnt") or die $@;
        logcomment("DumpRestore: create flexible vol $vol_name "
                  ."on aggr $aggr_name");
        vol_create_flex ($fh, $vol_name, $aggr_name, $vol_size."m");
    } else {
        logcomment("DumpRestore: create traditional vol $vol_name, "
                  ."disk cnt $disk_cnt");
        vol_create($fh, $vol_name, $disk_cnt) or die $@;
    }
}

#---------------------------------------------------------------
## @name ::drop_dump_vol
## @synopsis subroutine to drop an aggr created for dump-restore test
##        $fh - filer connection
##        $vol_name - Volume name
##        $aggr_name - aggregate name for flexible volume,
##                     otherwise it is a traditional volume
#---------------------------------------------------------------

sub drop_dump_vol {
    my ($fh, $vol_name, $aggr_name) = @_;

    if ( $aggr_name ) {
        logcomment("DumpRestore: purge aggr $aggr_name "
                  ."and flexible vol $vol_name");
        aggr_purge($fh, $aggr_name) or die $@;
    } else {
        logcomment("DumpRestore: destroy traditional vol $vol_name");
        vol_purge($fh, $vol_name) or die $@;
    }
}

#---------------------------------------------------------------
## @name ::create_dump_subdir
## @synopsis subroutine to create a subdir for dump-restore test
##                        creates the volume if is does not exists
##        $fh - filer connection
##        $subdir - sub-directory name of the form /vol/<vol name>/<dir name>
##
#---------------------------------------------------------------

sub create_dump_subdir {
    my ($fh, $subdir) = @_;

    logcomment("DumpRestore: Create sub-directory $subdir");
    my (undef, $vol_prefix, $vol_name, $dir) = split "/", $subdir;
    my $vol_root = "/" . $vol_prefix . "/" . $vol_name;
    logcomment("DumpRestore: Exporting volume $vol_root");
    export_path ($fh, $vol_root) or die $@;
    my $mt_pt = mount($fh->hostname(), $vol_root, chmod => 777) or die $@;
    if ( $mt_pt ) {
        logcomment("DumpRestore: Creating dir $dir on mount point $mt_pt");
        my $status = system ("mkdir $mt_pt/$dir") >> 8;
        umount($mt_pt) or die $@;
        unexport_path($fh, $vol_root) or die $@;
        croak("Failed to create subdir $mt_pt/$dir") if $status;
    } else {
        croak("Failed to create mount for $vol_root");
    }
}

#---------------------------------------------------------------
## @name ::create_dump_qtree
## @synopsis subroutine to create a qtree for dump-restore test
##                        creates the volume if it does not exists
##        $fh - filer connection
##        $qtree - qtree pathname of the form /vol/<vol name>/<qtree name>
#---------------------------------------------------------------

sub create_dump_qtree {
    my ($fh, $qtree) = @_;

    logcomment("DumpRestore: Create qtree $qtree");
    qtreeCreate($fh, $qtree) or die $@;
}

#---------------------------------------------------------------
## @name ::drop_dump_localfile
## @synopsis subroutine to remove local dump file
##        $fh - filer connection
##        $localfile - local dump file name
#---------------------------------------------------------------

sub drop_dump_localfile {
    my ($fh, $localfile) = @_;

    logcomment("DumpRestore: remove localfile used to hold dump");
    filer_file_destroy($fh, $localfile);
}

#---------------------------------------------------------------
## @name ::create_data
## @synopsis subroutine to create standard set of files on given directory
##        $fh - filer connection
##        $path - directory to create the data on
#---------------------------------------------------------------

sub create_data {
    my ($filer, $path) = @_;

    logcomment("DumpRestore: load compat-data to dir $path");
    load_compat_data($filer, $path);
}

#---------------------------------------------------------------
## @name ::dump_data
## @synopsis subroutine used to perform a dump
##        $fh - filer connection
##        $level - dump level, 0 - 9
##        $device - Device name of device to do dump or a localfile
##        $files - directory path or list of files to dump
##        $opts - Other optional parameters 
#---------------------------------------------------------------

sub dump_data {
    my ($fh, $level, $device, $files, $opts) = @_;

    _dev_ready($fh, $device);
    logcomment("DumpRestore: Level $level dump to device $device "
              ."for files $files");
	my $after;
    eval {
        (undef, undef, $after) = say($fh, "dump $level" . $opts . "f $device $files");
    };
    return $after =~ /DUMP IS DONE/;
}

#---------------------------------------------------------------
## @name ::restore_data
## @synopsis subroutine use to restore a dump from tape to a directory
##        $fh - filer connection
##        $device - Device name of device to load data from or a localfile
##        $dst_dir - destination pathname to restore dump to
#---------------------------------------------------------------

sub restore_data {
    my ($fh, $device, $dst_dir) = @_;

    _dev_ready($fh, $device);
    logcomment("DumpRestore: Load files from device $device to dst $dst_dir");
	my $after;
    eval {
        (undef, undef, $after) = say($fh, "restore rfD $device $dst_dir");
    };
    return $after =~ /RESTORE IS DONE/;
}

#---------------------------------------------------------------
## @name ::compare_dump_data
## @synopsis subroutine compare two directories for dump-restore test
##           returns 1 for matched and 0 otherwise
##        $fh - filer connection
##        $src_path - Source path
##        $dst_path - Destination path
##        $skip_qtree - Inidicates that the comparison should
##                      ignore qtree differences
#---------------------------------------------------------------

sub compare_dump_data {
    my ($fh, $src_path, $dst_path, $skip_qtree) = @_;

    my $src_snappath = _dump_snap ($fh, $src_path);
    my $dst_snappath = _dump_snap ($fh, $dst_path);
    my $cmt = "DumpRestore: Comparing src $src_snappath, dst $dst_snappath";
    # treecompare croaks if there are mismatches, and returns 1 if match
    # the eval is to catch the exception
	eval {
        if ($skip_qtree) {
            logcomment($cmt . ", Skip Qtree differences");
            treecompare($fh, $src_snappath, $dst_snappath,
                        operation=>"dr", skip=>["Found qtree "]) or die $@;
        } else {
            logcomment($cmt);
            treecompare($fh, $src_snappath, $dst_snappath,
                        operation=>"dr") or die $@;
        }
    };
    # translate the result to 1 for matched and 0 for mismatches found
    return ! $@;
}

#---------------------------------------------------------------
## @name ::_dump_snap
## @synopsis subroutine to create snapshot at root of volume and return snapshot pathname
##        $fh - filer connection
##        $path - pathname containing the volume where snapshot is to be created
#---------------------------------------------------------------

sub _dump_snap {
    my ($fh, $path) = @_;

    my $snap_name = "snapDR";
    my (undef, $vol_prefix, $vol_name, @name) = split "/", $path;
    my $base = (join "/", @name); 

    logcomment ("DumpRestore: creating snapshot on volume $vol_name");
    snapshot_create($fh, $vol_name, $snap_name) or die $@;

    my $snap_path = "/" . $vol_prefix . "/" . $vol_name
                  . "/.snapshot/" . $snap_name . "/" . $base;
    logcomment("DumpRestore: snapshot pathname is $snap_path");
    return $snap_path;
}

sub _dev_ready {
    my ($fh, $tape_device) = @_;

    if ( $tape_device !~ /^\// ) {
	# Assume tape device
	my %info = getTapeStatus($fh, $tape_device);
	croak "Tape device $tape_device does not exists or not ready"
		unless $info{status} =~ /ready/i;
    }
    return 1;
}

