# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/dp/dumprestore/snapdiff/SnapdiffUtil.pm#1 $
#
## @summary SnapdiffUtil - Utilities for snapdiff tests
## @author      dl-dpgqa-automation@netapp.com
## @status	private
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

SnapdiffUtil - Utilities for snapdiff tests

=head1 SYNOPSIS

  use dp::dumprestore::snapdiff::SnapdiffUtil;

  @file_list = create_files(path     => $mntpt,
                            name     => $basename,
                            contents => $contents,
                            count    => 1200);

  %dir_list = create_dirs(path       => $mntpt,
                          name       => $basename,
                          contents   => $contents,
                          dir_count  => $num_dirs,
                          file_count => 100);

  @files = recreate_dir(path       => "$mntpt/$dir",
                        contents   => $contents,
                        file_count => 1);

  ($created, $deleted, $modified) =
         modify_files(path      => $mntpt,
                      file_list => \@file_list,
                      contents  => $contents,
                      count     => 200);

=head1 DESCRIPTION

Provides file manipulation routines specifically for snapdiff tests
under dp/dumprestore/snapdiff. These routines create files and
directories, and manipulate files and directories.

The main purpose of these routines is to enable keeping track of
changes made to a file system between snapshots, for comparison
with output from the snapdiff requests. That is why the existing
population libraries such as DataManip or TargetedData were
not used.

=head1 STATUS

  Private

=head1 AUTHOR

  dl-dpgqa-automation@netapp.com
  $Id: //depot/prod/test/Rsierranevada/dp/dumprestore/snapdiff/SnapdiffUtil.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2007, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package dp::dumprestore::snapdiff::SnapdiffUtil;

use strict;
use warnings;

use Carp qw(croak);
use Params::Validate qw(validate SCALAR ARRAYREF);
use FileManip qw(write_file remove_file setattr_file);
use waflobjCreator qw(create_directory delete_object);

use vars ('@ISA','@EXPORT');
require Exporter;
our @ISA = ('Exporter');
our @EXPORT = qw(
  create_files
  create_dirs
  recreate_dir
  modify_files
);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub create_files(%);
sub create_dirs(%);
sub _create_dir($$$);
sub recreate_dir(%);
sub modify_files(%);

#---------------------------------------------------------------------

=head2 create_files

  @file_list = create_files(path     => $mntpt,
                            name     => $basename,
                            contents => $contents,
                            count    => 1200);

Create a fixed number of files on a given path on localhost using
the specified contents.

The following named arguments are accepted:

=over

=item Arguments

=over

=item path

(Required) : A path on the localhost where files will be created.

=item name

(Optional) : Base file name to use for created files. Defaults
to 'file'. Name will be appended with a numeric value.

=item contents

(Optional) : A data string that will be written to each of the
created files. Defaults to 'A' x 1024.

=item count

(Optional) : The number of files to be created. Defaults to 100.

=back

=item Returns

An list of the filenames created.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub create_files(%) {
    my (%opts) = validate(
        @_,
        {
            path     => { type => SCALAR },
            name     => { type => SCALAR, default => 'file' },
            contents => { type => SCALAR, default => 'A' x 1024 },
            count    => { type => SCALAR, default => 100 },
        }
    );

    my @file_list;

    foreach (1..$opts{count}) {
        my $file = $opts{name}."_$_";
        eval {
            write_file(file => $opts{path}."/$file",
                       contents => $opts{contents},
                       mode => '0644');
        };
        croak "Could not create file set in ".$opts{path}.": $@" if $@;
        push @file_list, $file;
    }

    return @file_list;
}

#---------------------------------------------------------------------

=head2 create_dirs

  %dir_list = create_dirs(path       => $mntpt,
                          name       => $basename,
                          contents   => $contents,
                          dir_count  => $num_dirs,
                          file_count => 100);

Create a fixed number of directories on a given path on localhost.
Each directory will contain a fixed number of files populated using
the specified contents.

The following named arguments are accepted:

=over

=item Arguments

=over

=item path

(Required) : A path on the localhost where directories will be created.

=item name

(Optional) : Base file name to use for created directories. Defaults
to 'directory'. Name will be appended with a numeric value.

=item contents

(Optional) : A data string that will be written to each of the
created files. Defaults to 'A' x 1024. Files will be named
'file' appended with a number.

=item dir_count

(Optional) : The number of directories to be created.
Defaults to 20.

=item file_count

(Optional) : The number of files to be created in each directory.
Defaults to 100.

=back

=item Returns

A hash with directory names as keys, and array references of the
filenames created under each directory as the values.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub create_dirs(%) {
    my (%opts) = validate(
        @_,
        {
            path       => { type => SCALAR },
            name       => { type => SCALAR, default => 'directory' },
            contents   => { type => SCALAR, default => 'A' x 1024 },
            dir_count  => { type => SCALAR, default => 20 },
            file_count => { type => SCALAR, default => 100 }
        }
    );

    my %dir_list;

    foreach (1..$opts{dir_count}) {
        my @file_list;
        my $dir = $opts{name}."_$_";
        my $dir_path = $opts{path}."/$dir";
        @file_list = _create_dir($dir_path,
                                 $opts{contents},
                                 $opts{file_count});
        $dir_list{$dir} = \@file_list;
    }

    return %dir_list;
}

#---------------------------------------------------------------------
# Internal function to handle creating one directory. Used by
# create_dirs and recreate_dir.
#
# Creates the directory then populates it using create_files.
#---------------------------------------------------------------------
sub _create_dir($$$) {
    my ($path, $contents, $file_count) = @_;
    my @file_list;

    create_directory($path)
      or croak "Could not create directory $path: $@";
    @file_list = create_files(path     => $path,
                              contents => $contents,
                              count    => $file_count);

    return @file_list;
}

#---------------------------------------------------------------------

=head2 recreate_dir

  @files = recreate_dir(path       => "$mntpt/$dir",
                        contents   => $contents,
                        file_count => 1);

Creates a single directory on a given path, first removing the
existing directory. Populates the directory with a specified
number of files, each containing the given contents.

The following named arguments are accepted:

=over

=item Arguments

=over

=item path

(Required) : A path on the localhost to the directory to
be recreated.

=item contents

(Optional) : A data string that will be written to each of the
created files. Defaults to 'B' x 1024. Files will be named
'file' appended with a number.

=item file_count

(Optional) : The number of files to be created in the directory.
Defaults to 100.

=back

=item Returns

A list of filenames that were created in the directory

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub recreate_dir(%) {
    my (%opts) = validate(
        @_,
        {
            path       => { type => SCALAR },
            contents   => { type => SCALAR, default => 'B' x 1024 },
            file_count => { type => SCALAR, default => 100 }
        }
    );

    my @file_list;

    delete_object($opts{path}, purge => 1)
      or croak "Could not remove directory ".$opts{path}.": $@";
    @file_list = _create_dir($opts{path},
                             $opts{contents},
                             $opts{file_count});

    return @file_list;
}

#---------------------------------------------------------------------

=head2 modify_files

  ($created, $deleted, $modified) =
         modify_files(path      => $mntpt,
                      file_list => \@file_list,
                      contents  => $contents,
                      count     => 200);

Modifies files located in a specified path. New files are created,
existing files are deleted, file contents are modified, file
attributes are modified, and existing files are deleted and
recreated with the same name.

The following named arguments are accepted:

=over

=item Arguments

=over

=item path

(Required) : A path on the localhost to the location of the files
to be modified.

=item file_list

(Required) : An array reference to a list of the existing files
located in 'path'.

=item contents

(Optional) : A data string that will be written to each of the
created files. Defaults to 'C' x 1024. Files will be named
'file' appended with a number.

=item count

(Optional) : The number of files in the directory to be modified
by each method. E.g. 'count' files will be created, 'count' existing
files will be deleted, 'count' files will have their contents
modified, etc. Defaults to 100. The number of files in 'file_list'
must be at least 4 times 'count' for the process to work correctly,
so that no file is changed more than one way.

=back

=item Returns

A list of array references that point to the created filenames,
deleted filenames, and modified filenames.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub modify_files(%) {
    my (%opts) = validate(
        @_,
        {
            path       => { type => SCALAR },
            file_list  => { type => ARRAYREF },
            contents   => { type => SCALAR, default => 'C' x 1024 },
            count => { type => SCALAR, default => 100 }
        }
    );

    my (@created, @deleted, $file, @modified);

    croak "modify_files cannot change more files than exist in directory"
      if (scalar @{$opts{file_list}} < 4 * $opts{count});

    foreach (1..$opts{count}) {
        # Create a new file
        # Add the filename onto the file list
        $file = "file_created_$_" . time;
        eval {
            write_file(file => $opts{path}."/$file",
                       contents => $opts{contents},
                       mode => '0644');
        };
        croak "Could not create file in ".$opts{path}.": $@" if $@;
        push @created, $file;
        push @{$opts{file_list}}, $file;

        # Delete a file
        $file = shift @{$opts{file_list}};
        eval {
            remove_file(file => $opts{path}."/$file");
        };
        croak "Error removing file in ".$opts{path}.": $@" if $@;
        push @deleted, $file;

        # Modify a file's data
        # Move the filename to the end of the file list
        $file = shift @{$opts{file_list}};
        eval {
            write_file(file => $opts{path}."/$file",
                       contents => $opts{contents},
                       mode => '0644');
        };
        croak "Could not modify file data in ".$opts{path}.": $@" if $@;
        push @modified, $file;
        push @{$opts{file_list}}, $file;

        # Modify a file's attributes
        # Move the filename to the end of the file list
        $file = shift @{$opts{file_list}};
        eval {
            setattr_file(file => $opts{path}."/$file",
                         mode => '0777');
        };
        croak "Could not modify file attributes in ".$opts{path}.": $@" if $@;
        push @modified, $file;
        push @{$opts{file_list}}, $file;

        # Delete a file then recreate it
        # Move the filename to the end of the file list
        $file = shift @{$opts{file_list}};
        eval {
            remove_file(file => $opts{path}."/$file");
        };
        croak "Error removing file in ".$opts{path}.": $@" if $@;
        push @deleted, $file;

        eval {
            write_file(file => $opts{path}."/$file",
                       contents => $opts{contents},
                       mode => '0644');
        };
        croak "Could not create file in ".$opts{path}.": $@" if $@;
        push @created, $file;
        push @{$opts{file_list}}, $file;
    }

    return \@created, \@deleted, \@modified;
}

1;
