# $Id: //depot/prod/test/Rsierranevada/dp/filer_profiler/FilerProfiler.pm#1 $
#!/usr/software/bin/perl -w
package FilerProfiler;
use strict;
use warnings;

# NetApp Libraries
use Server;
use SysconfigParser;
use FilerProfilerUtils qw(parse_key join_hash);
use FilerProfilerDBUtils;

# Standard/Third Party Libraries
use DBI;
use Data::Dumper;

sub new {
    my ($class, %args) = @_;
    my $self; 

    my $dbh = DBI->connect("dbi:SQLite:dbname=filer_profiler.sqlite", "", "");
    my $db = FilerProfilerDBUtils->new(dbh => $dbh); 
    my $server = Server->new(
                            nodes => @{$args{'filers'}}[0],
                            nodes_on_demand => 1
                        );
    foreach (@{$args{'filers'}}) {
        eval {
            $server->add_nodes(
                        nodes => $_,
                        nodes_on_demand => 1
                    );
        }
    }
    $self = {
                filers => $args{'filers'}, 
                server => $server,
                dbh => $dbh,
                db => $db,
                profiles => {}
            };

    bless($self, $class);

    return $self;
}

# Populate filers by either running the generated CDEF or
# other helper methods to parse the output of a specific
# command.
sub populate_profile {
    my $profile;
    my $class = shift;
    $profile->{'name'} = shift;

    # Gathers general information about the system by running
    # `system node show -node <filer>`. Information includes:
    # location, system id, and uptime.
    eval {
        $profile->{'sys_show'} = $class->{'server'}->system_node_show(
                                    -Server => $profile->{'name'},
                                    node => $profile->{'name'});
        $profile->{'systemid'} = $profile->{'sys_show'}[0]->{'systemid'};                           
    };
    if ($@) {
        print "$@\n";
    }

    # Runs the command `run local -node <filer> sysconfig -v` and parses
    # the output of the sysconfig using the Sysconfig Parser Library
    eval {
        my $sysconfig = $class->{'server'}->system_node_run(
                            -Server => $profile->{'name'},
                            node => $profile->{'name'},
                            command => 'sysconfig -v');
        $profile->{'sysconfig'} = SysconfigParser->parse_sysconfig($sysconfig);
    };
    if ($@) {
        print "$@\n";
    }

    # Gathers information about all the disks that belong to the filer,
    # with the assumption that the disks that belong to filer are only located
    # on odd ports (0a, 0c, 0e, ...)
    eval {
        $profile->{'disks'} = $class->{'server'}->storage_disk_show(
                                -Server => $profile->{'name'},
                                owner => $profile->{'name'},
                                disk => "$profile->{'name'}:0a*,"
                                            ."$profile->{'name'}:0c*"
                            );
    };
    if ($@) {
        print "$@\n";
    }

    # Gathers information about the network interfaces that belong solely
    # to the filer by running the following cmd:
    # `network interface show -curr-node <filer> -home-node <filer>`
    eval {
        $profile->{'net_int'} = $class->{'server'}->network_interface_show(
                                    -Server => $profile->{'name'},
                                    'curr-node' => $profile->{'name'},
                                    'home-node' => $profile->{'name'}
                                );
    };
    if ($@) {
        print "$@\n";
    }

    # Gathers infomation about all the shelves the filer is connected to.
    # Since some of these shelves loop, there will be reundancies and
    # these still need to be accounted for accuracy purposes.
    eval {
        $profile->{'shelves'} = get_shelves($class, $profile->{'name'});
        if (defined($profile->{'sysconfig'})) {
            my $shelf_id;
            foreach $shelf_id (keys(%{$profile->{'sysconfig'}->{'esh'}})) {
                my $channel;
                my $esh = $profile->{'sysconfig'}->{'esh'}->{$shelf_id};
                my $shelf = $profile->{'shelves'}->{$shelf_id};
                foreach $channel (keys(%$esh)) {
                    $shelf->{$channel}->{'esh_version'} = $esh->{$channel}->{'esh_version'};
                    $shelf->{$channel}->{'esh_a_firmware'} = $esh->{$channel}->{'esh_a_firmware'};
                    $shelf->{$channel}->{'esh_b_firmware'} = $esh->{$channel}->{'esh_b_firmware'};
                }
            }
            # Remove redundant data about shelves from the sysconfig hash
            delete($profile->{'sysconfig'}->{'esh'});
        }
    };
    if ($@) {
        print "$@\n";
    }

    # Collects the cluster health information of the entire cluster.
    eval {
        $profile->{'cluster'} = $class->{'server'}->cluster_show(-Server => $profile->{'name'});
    };
    if ($@) {
        print "$@\n";
    }

    # Gets the remote management module that is found on every filer,
    # which is a standalone system that physically connected to the board to
    # monitor the filer's hardware statuses.
    eval {
        $profile->{'rmm'} = get_remote_management($class, $profile->{'name'});
    };
    if ($@) {
        print "$@\n";
    }

    eval {
        $profile->{'stats'} = get_statistics($class, $profile->{'name'});
        $profile->{'stats'}->{'fid'} = $profile->{'systemid'};
    };
    if ($@) {
        print "$@\n";
    }

    return $profile;
}

sub save_run {
    my $class = shift;
    my $runid = shift;
    my $profile;

    $class->{'db'}->{'runid'} = $runid;

    foreach $profile (values(%{$class->{'profiles'}->{$runid}})) {
        $class->{'db'}->insert_filer($profile);
        $class->{'db'}->insert_network_interface($profile);
        $class->{'db'}->insert_sysconfig($profile);
        $class->{'db'}->insert_rmm($profile);
        $class->{'db'}->insert_disks($profile);
        $class->{'db'}->insert_shelves($profile);
        $class->{'db'}->insert_statistics($profile->{'stats'});
    }
}

sub save_profiles {
    my $class = shift;
    my $runid = shift;
    my $profile;

    $class->{'db'}->{'runid'} = $runid;

    foreach $profile (values(%{$class->{'profiles'}->{$runid}})) {
        $class->{'db'}->insert_filer($profile);
        $class->{'db'}->insert_network_interface($profile);
        $class->{'db'}->insert_sysconfig($profile);
        $class->{'db'}->insert_rmm($profile);
        $class->{'db'}->insert_disks($profile);
        $class->{'db'}->insert_shelves($profile);
        $class->{'db'}->insert_statistics($profile->{'stats'});
    }
}

sub save_profile {
    my $class = shift;
    my $runid = shift;
    my $profile = shift;

    $class->{'db'}->{'runid'} = $runid;
    $class->{'db'}->insert_filer($profile);
    $class->{'db'}->insert_network_interface($profile);
    $class->{'db'}->insert_sysconfig($profile);
    $class->{'db'}->insert_rmm($profile);
    $class->{'db'}->insert_disks($profile);
    $class->{'db'}->insert_shelves($profile);
    $class->{'db'}->insert_statistics($profile->{'stats'});
}

sub get_remote_management {
    my $class = shift;
    my $filer_name = shift; 
    my $rmm;
    my $rmm_out;
    my $rlm_missing = 'There is no RLM \(Remote LAN Module\) in this system';

    $rmm_out = $class->{'server'}->system_node_run(
                                                    -Server => $filer_name,
                                                    node => $filer_name,
                                                    command => 'rlm status'
                                                );
    if ($rmm_out =~ $rlm_missing) {
        $rmm_out = $class->{'server'}->system_node_run(
                                                        -Server => $filer_name,
                                                        node => $filer_name,
                                                        command => 'sp status'
                                                    );
    }

    $rmm_out =~ s/(Remote\s+LAN\s+Module)|(Service\s+Processor)//;

    $rmm->{'type'} = $1 || $2; 

    foreach (split(/\n/, $rmm_out)) {
        /^\s+(\S.*):\s+(\S.*)$/;
        $rmm->{parse_key($1)}= $2; 
    }
    return $rmm;
}

sub get_statistics {
    my (@usage, @dashes, @lh, @uh) = [];
    my $class = shift;
    my $filer = shift;
    my $stats;

    my $stats_out = $class->{'server'}->statistics_periodic(
                                    -Server => $filer, 
                                    node => $filer,
                                    interval => 5,
                                    iterations => 1
                                );

    $stats_out = [split(/\n/, $stats_out)];

    foreach(@$stats_out) {
        s/^\s+//;
    }

    @usage = split(/\s+/, pop(@$stats_out));
    @dashes = split(/\s+/, pop(@$stats_out));
    @lh = split(/\s+/, pop(@$stats_out));
    @uh = split(/(\s+)/, pop(@$stats_out));

    foreach (@lh) {
        s/-/_/g;
        if (defined($uh[0]) && $uh[0] =~ /\S+/) {
            $stats->{shift(@uh) . "_$_"} = shift(@usage);
        } else { 
            if (defined($uh[0]) && length($uh[0]) > length($dashes[0])) {
                $uh[0] = substr($uh[0], 0, length($dashes[0]) * -1);
                $stats->{$_} = shift(@usage);
            } elsif (defined($uh[0])) {
                shift(@uh);
                $stats->{shift(@uh) . "_$_"} = shift(@usage);
            } else {
                $stats->{$_} = shift(@usage);
            }
        }
        shift(@dashes);
    }
    return $stats;
}

sub get_shelves {
    my $class = shift;
    my $filer = shift;
    my $shelf_out;
    my $shelves = {};
    my $shelves_out = $class->{'server'}->system_node_run(
                        -Server => $filer,
                        node    => $filer,
                        command => 'storage show shelf -a'
                    );
    $shelves_out = [split(/Shelf name:/, $shelves_out)];

    foreach (@$shelves_out) {
        $shelf_out = [split(/\n/, $_)]; 
        my $shelf_name = shift(@$shelf_out);
        my $shelf;

        # Skip shelves that are not owned by the filer
        if (!defined($shelf_name) || $shelf_name =~ /PARTNER/i) {
            next;
        }

        # Trim excess whitespace
        $shelf_name =~ s/^\s+//;

        foreach (@$shelf_out) {
            if (/^(\S.*?):\s+(\S.*?)\s*$/) {
                $shelf->{parse_key($1)} = $2;
            }
            # Still need to handle the output for each individual disk.
        }
        if ($shelf_name =~ /\w+\.shelf(\d+)/) {
            $shelves->{$1}->{$shelf->{'channel'}}=  $shelf;
        }
    }
    return $shelves;
}

1;
