# $Id: //depot/prod/test/Rsierranevada/dp/fmc/FcMetroclusterUtils.pm#2 $

## @summary  General methods to ease FMC automation
##
## @description
##      FcMetroclusterUtils.pm functions
##
## @status public
##
## @author vvatsal@netapp.com
## @author amit3@netapp.com
## @author dl-bijaya@netapp.com
##
## @keywords FcMetroclusterUtils
=head1 NAME

FcMetroclusterUtils

=head1 DESCRIPTION

This package provides methods for helping FMC automation

=head1 METHODS

=cut

package FcMetroclusterUtils;
use strict;
use base qw(Exporter);
use Cwd qw(getcwd);
use Hammer qw(nfs_hammer_start nfs_hammer_stop);
use Carp qw(croak);
use Tharn;
use Params::Validate qw(validate validate_pos SCALAR OBJECT HASHREF ARRAYREF);
use NOTE::Vfiler;
use GenericMount qw(mount umount);
use filer qw(filer_sysstat_get);

BEGIN
{
  our @EXPORT = qw(
      restart_hammer
	  vfiler_sio
	  run_sio
	  nfs_traffic
	  ssl_setup
	  rpower_devices
  );
}


=head2 ::restart_hammer

Checks if the started hammer is still running, if not then restart the nfs hammer again. Returns the hammer id for the currently running hammer

=over

=item Arguments

B<fc> => I<Filer conn object>

I<Required.>
Filer conn object for the filer on which hammer is already running

B<hammer_id> => I<Hammer id of already running hammer>

I<Required.>
Hammer id of the already running hammer


B<vol_name> => I<Volume name on which hammer is to be run>

I<Required.>
Volume name on which hammer is to be run.


B<client> => I<Client through which hammer is to be run>

I<Optional.>
Defaults to local host



=item Examples

=back


use dp::fmc::FcMetroclusterUtils;
FcMetroclusterUtils::restart_hammer(fc =>  $Filer1->conn , hammer_id =>  $nfs_hammer_id1 , vol_name => $Volume1->name );


=cut


sub restart_hammer(%) {
	my (%args) = @_;

	croak "Filer_conn not passed " if(!defined($args{fc})); 
        my $filer_conn =$args{fc}; 
	croak "Hammer id not passed " if(!defined($args{hammer_id})); 
        my $ham_id=$args{hammer_id};
        my $client= defined $args{client} ? $args{client}:'localhost';
        croak "Volume name not passed" if(!defined($args{vol_name}));
        my $vol=$args{vol_name};
        if(!($$ham_id{subtest}->is_alive())){
            my $nfs_hammer_id1 = nfs_hammer_start(
      		fc          => $filer_conn,
      		client      => $client,
      		hammer_time => -1,
      		vol_name    => $vol,
   		);
           NOTE::logcomment(" Hammer is not running, restarting hammer ");

      	    return $nfs_hammer_id1;
       }
       else {
	   NOTE::logcomment(" Hammer is still running ");
           return $ham_id ;
	}
}

=head2 ::vfiler_sio

Provide SIO support for vfilers. See:
http://perf-web.corp.netapp.com/twiki/bin/view/Perfweb/SIO
for full details of SIO functionality.

The SIO process is continuously 'churned' until terminated by call with 'stop' set
or automatically at Teardown.

The SIO program is expected to be located at /usr/local/test/bin/sio

Returns the SIO object is successful, or 1 if stopping SIO

=over

=item Arguments

B<vfiler_obj> => I<vfiler object>

vfiler on which to run SIO.

B<vol_name> => I<volume name>

Name of the volume on vfiler on which to run SIO.

B<client> => I<hostspec>

Client to run SIO from. Default localhost or set by a previous call. Requires a Linux client

B<sio_hash> => I<hash reference>

Reference to the hash for holding SIO related information for this function

B<cycle> => I<integer>

How long to run sio before repeating, in seconds. A special value of zero indicates
only create data files and not to start sio process. Default 900

B<read> => I<percent>

Percent of read access vs write access. Default 100

B<rand> => I<percent>

Percent of random access vs sequential access. Default 100

B<blksize> => I<IO block size>

Size of IO. Default 5k

B<start_byte> => I<integer>

Lower bound for access. Default 0

B<count> => I<integer>

If specified determines number of files to create for sio to use. Default 5

B<file_size> => I<size>

File size of each file used by sio. Default 20m

B<numthreads> => I<integer>

Number of IO generators. Default 1

B<stop> => I<1|0>

Stop the SIO process if set. Default 0

B<restart> => I<integer>

Flag to indicate to start sio with parameters from previous invocation.

B<timeout> => I<integer>

Timeout in seconds for Tharn::say calls in the function.

=item Examples

=back

	#Library path to be included
	use dp::fmc::FcMetroclusterUtils;
	
	#For starting SIO on vfiler
	FcMetroclusterUtils::vfiler_sio(vfiler_obj => $vfiler_obj, vol_name => $vfiler_nfs_vol_name );

	#For stopping running SIO on vfiler
	FcMetroclusterUtils::vfiler_sio( vfiler_obj => $vfiler_obj, vol_name => $vfiler_nfs_vol_name,  stop => 1, sio_hash => $sio_hash );
	
	#For restarting SIO on vfiler
	FcMetroclusterUtils::vfiler_sio( vfiler_obj => $vfiler_obj, vol_name => $vfiler_nfs_vol_name,  restart => 1, sio_hash => $sio_hash );
	
	#Additional parameters for SIO are configurable and can be passed through function call. Example
	# Example: Tell sio to use listed files on client nam9
	FcMetroclusterUtils::vfiler_sio(vfiler_obj => $vfiler_obj, vol_name => $vfiler_nfs_vol_name, file_list => "f1 f2");

	# Perform sio with (25% read, 75% write) 
	FcMetroclusterUtils::vfiler_sio(vfiler_obj => $vfiler_obj, vol_name => $vfiler_nfs_vol_name, read => 25);

=cut

sub vfiler_sio {
	my $sio_hash; 
	my %sio_hash1;
	my (%args) = validate (@_, {
				vfiler_obj => {type=>OBJECT, optional=>0},
				vol_name => { type => SCALAR, optional => 0 },
				client => {type=>SCALAR, optional=>1},
				sio_hash => {type=>HASHREF, optional=>1},
				cycle => {type=>SCALAR, optional=>1},
				read => {type=>SCALAR, optional=>1, default=>50, regex=>qr/\d/},
				rand => {type=>SCALAR, optional=>1, default=>50, regex=>qr/\d/},
				blksize => {type=>SCALAR, optional=>1, default=>"32k"},
				start_byte => {type=>SCALAR, optional=>1, default=>0, regex=>qr/\d/},
				count => {type=>SCALAR, optional=>1, default=>3, regex=>qr/\d/},
				file_size => {type=>SCALAR, optional=>1},
				numthreads => {type=>SCALAR, optional=>1, default=>1, regex=>qr/\d/},
				stop => {type=>SCALAR, optional=>1, default=>0},
				umount => {type=>SCALAR, optional=>1, default=>0},
				restart => {type=>SCALAR, optional=>1, default=>0},
				timeout => {type=>SCALAR, optional=>1, default=>300},
                proto          => { type => SCALAR, default => 'tcp',
                                    regex => qr/^(udp|tcp)$/ },
                rsize          => { type => SCALAR, default => 32768,
                                    regex => qr/^\d+$/ },
                version        => { type => SCALAR, default => 3 },
                wsize          => { type => SCALAR, default => 32768,
                                    regex => qr/^\d+$/ },
				
			});
	if (defined $args{sio_hash}){
		$sio_hash = $args{sio_hash};
	}
	else {
		$sio_hash = \%sio_hash1;
	}
	croak "sio 'stop' and 'restart' options are mutually exclusive"
		if $args{stop} && $args{restart};
	
	###Validating few input parameters
	$args{client} = $args{client} || $sio_hash->{_client} || "localhost";
	$args{file_size}= "1024"	unless defined $args{file_size};
	croak "numthreads must be > 0"
		unless $args{numthreads} > 0;

	my $host  = Tharn::host( $args{client} );
    my $lnxclnt = Tharn::connect('#rsh');

	###Let's stop SIO if requested
	if ( $args{stop} ) {
		if ( $sio_hash->{_sio} ) {
			NOTE::logcomment("Stopping SIO on volume $args{vol_name}");
			$sio_hash->{_sio}->raise("TERM");
			_sio_stop_file($sio_hash, "create", $lnxclnt);
			# Lets wait a little to give sio chance to exit
			sleep 60; 
			my $retry = 0;
			while ( $sio_hash->{_sio}->is_alive ) {
				croak "Unable to stop SIO on volume $args{vol_name} after $retry attempts. Aborting..."
					unless $retry++ < 30;
				NOTE::logcomment("Send kill signal for SIO on volume $args{vol_name}");
				$sio_hash->{_sio}->kill;
				sleep(15);
			}
			$sio_hash->{_sio}->wait_finish(timeout => 120);
	    	FSR::FSRutils::subtestResults($sio_hash->{_sio}) or croak $@;
			delete $sio_hash->{_sio};
			_sio_stop_file($sio_hash, "remove", $lnxclnt);
			sleep(10);
			NOTE::logcomment("SIO process terminated");
			if($args{umount}){
			    $sio_hash->{umount} = GenericMount::umount( $sio_hash->{mountpoint}, force => "1");
			}
			
		} 
		else {
			croak "Cannot stop SIO. No SIO process running";
		}
		return 1;
	}	
	###We're here means SIO stop was not requested. Let's proceed
	NOTE::logcomment("Checking mount volume $args{vol_name} exists on vfiler...");
	my @vfiler_volumes;
	my $found = 0;
	@vfiler_volumes = $args{vfiler_obj}->vfiler_volumes;
	foreach (@vfiler_volumes){
		if (/$args{vol_name}/){
			$found = 1;
			last;
		}
	}
	croak ("Volume $args{vol_name} does not exists on vfiler\n") unless ($found);
	
	###Performing volume attach to mount volume. Needed for mounting the volume through vfiler
	my $basefiler = $args{vfiler_obj}->basefiler;
	$sio_hash->{vol_obj} = NOTE::Volume->attach(
		 Filer => $basefiler,
		 name  => $args{vol_name},
	);
	
		
	###Let's restore SIO hash with old parameters if SIO restart is requested
	if ( $args{restart} && exists $sio_hash->{_sio_args} ) {
		NOTE::logcomment("SIO restart requested. Restoring previous SIO parameters");
		delete $args{restart};
		foreach my $key (keys %{$sio_hash->{_sio_args}}) {
			$args{$key} = $sio_hash->{_sio_args}{$key};
		}
		NOTE::logcomment("Completed restoring previous SIO parameters needed for SIO restart\n");
	}
	%{$sio_hash->{_sio_args}} = %args;

	# Mount for SIO access
	NOTE::logcomment("Mounting the vfiler volume $args{vol_name} now");	
    $args{vfiler_obj}->exportfs( container => $sio_hash->{vol_obj} );
    eval{
    	$sio_hash->{mountpoint} = GenericMount::mount(
                                        $args{vfiler_obj}->defaultip,
                                        $sio_hash->{vol_obj}->mount_path,
                                        client => $args{client},
                                        chmod  => '0777',
                                        mntopts => "vers=$args{version},proto=$args{proto},noac,timeo=600," .
                                                   "rsize=$args{rsize},wsize=$args{wsize}",
                                    );
    };
    Carp::croak( "Error in mounting volume:\n " . $@ ) if ($@);
    croak("Mountpoint is NULL. Mount failed\n")if (!defined $sio_hash->{mountpoint});
	my $rw_mnt = undef;
	my $fname;
	my $idx = 1;
	NOTE::logcomment ("Creating placeholder data files for running SIO\n");
	do {
		my $sio_file = "sio_".$idx."_".$$;
		my $fullpath = $sio_hash->{mountpoint}."/".$sio_file;
		my @cmdoutput =   Tharn::say( $lnxclnt,
                            		"sudo /bin/dd if=/dev/urandom of=" 
                            		. $fullpath . " bs=" 
                            		. $args{file_size}
                            		. "M count=1; echo __D0NE__", 
                            		"__D0NE__", 
                            		-noprompt, 
                            		-timeout, 1800 
                            		);
		if ($cmdoutput[1] !~ /1\+0 records in([\r\n])*1\+0 records out/ ){
			croak("Unable to create data files for SIO\n");
		}
		Tharn::say( $lnxclnt, "sudo chmod 777 $fullpath; echo __D0NE__", 
		            "__D0NE__", 
		            -noprompt, 
		            -timeout, $args{timeout}
		           );
		#Append the full path of each data file to a single string
		$fname .= "$sio_hash->{mountpoint}/" .$sio_file. " ";
	} while $idx++ < $args{count};
	NOTE::logcomment ("Done creating placeholder data files for running SIO\n");
	return 1 if defined $args{cycle} && $args{cycle} <= 0; # Create and return data files only, do not start sio
	
	#Creating stop file to stop SIO when requested
	#Note: SIO -stop file option only works for linux host, so use short 60 cycle
	#     unless this is the case
	my $stop_cmd = "";
	$stop_cmd = "-stop -"._sio_stop_file($sio_hash, "set",$lnxclnt)
		if Tharn::host($args{client})->hosttype =~ /linux/i;
	$args{cycle} = ($stop_cmd ? 900 : 60) # Default to long cycle time if able to use stop file
		unless defined $args{cycle};
	
	###Create the SIO command
	my $cmdsio = "sudo /usr/software/test/bin/sio"
			." $args{read} $args{rand} $args{blksize} $args{start_byte}"
			." $args{file_size}m $args{cycle} $args{numthreads} $fname $stop_cmd; 
			; echo __D0NE__";
	
	NOTE::logcomment ("Calling a subtest now for starting SIO threads\n");
	my $ret = Tharn::subtest(sub {
					close STDIN;
					my ($rc, $run) = (1, 1);
					#Configure to stop and exit on receiving a terminate signal
					$SIG{TERM} = sub { 
							NOTE::logcomment("Received a TERM signal. Stopping SIO on $args{vol_name}");
							$run = 0;
						};
					#Run continuous SIO loop until stopped by parent
					NOTE::logcomment("Starting SIO on $args{vol_name}");
					my $timeout = $args{cycle} + 180;
					do {
						NOTE::logcomment("Starting SIO on $args{vol_name} yet again, run: $run");
						my @cmdoutput = Tharn::say( $lnxclnt, $cmdsio, "__D0NE__", -noprompt, -timeout, $timeout);
						if ($cmdoutput[1] =~ /FAIL/i || $cmdoutput[1] =~ /ERROR/i){
							$rc = 0; #configure to stop if found any issue
						}
						else{
							$rc++; #Everything looks good. Let's continue
						}
					} while $run && $rc;
					Tharn::logresult("FATAL", "FAIL: Attempt at running SIO on $args{vol_name} failed at iteration = $rc")
						if $rc < 1;
					NOTE::logcomment("SIO stopped on $args{vol_name}");
				}, '-bg', '-onexit', sub {});
	croak "Failed to start SIO: $@"
		if $@;
	$sio_hash->{_sio} = $ret;
	NOTE::logcomment("Completed running SIO successfully\n");
	# Schedule SIO to be stopped during teardown
	NOTE::reference( sub { vfiler_sio( vfiler_obj => $args{vfiler_obj}, vol_name => $args{vol_name}, client => $args{client}, stop => 1, sio_hash => $sio_hash) if exists $sio_hash->{_sio} } );
	return $sio_hash;
}

sub _sio_stop_file {
	my ($sio_hash, $mode, $lnxclnt) = @_;
	my ($cmd, $cmdoutput);

	if ( $mode =~ /set/i ) {
		$sio_hash->{_sio_stop_file} = $sio_hash->{vol_name}."_sio_stop_".$$;
	} else {
		return undef unless defined $sio_hash->{_sio_stop_file};

		my $fullpath = $sio_hash->{mountpoint}."/" .$sio_hash->{_sio_stop_file};
		if ( $mode =~ /create/i ) {
			# Don't abort if stop file creation failed
			$cmd = "sudo touch $fullpath; sudo chmod 777 $fullpath; echo __D0NE__";
			$cmdoutput = Tharn::say( $lnxclnt, $cmd, 
			                          "__D0NE__", 
                                     -noprompt,
                                     -timeout, 300);
			NOTE::logcomment("SIO stop file ".$sio_hash->{_sio_stop_file}." created.\n")
				if -e $fullpath;
		}
	
		if ( $mode =~ /remove/i && -e $fullpath ) {
			# Attempt to remove stop file if exists
			if ( -e $fullpath ) {
				$cmd = "sudo rm -f $fullpath; echo __D0NE__";
            $cmdoutput = Tharn::say( $lnxclnt, $cmd, 
                                     "__D0NE__", 
                                     -noprompt,
                                     -timeout, 300);
				croak ("Failed to removed stop file ".$sio_hash->{_sio_stop_file})
					if -e $fullpath;
			}
			delete $sio_hash->{_sio_stop_file}; # No need to save stop after
		}
	}

	return $sio_hash->{_sio_stop_file};
}

=head2 ::run_sio

Run SIO on given mount path of a client. See:
http://perf-web.corp.netapp.com/twiki/bin/view/Perfweb/SIO
for full details of SIO functionality.

The SIO process is continuously 'churned' until terminated by call with 'stop' set
or automatically at Teardown.

The SIO program is expected to be located at /usr/local/test/bin/sio

Returns the SIO object is successful, or 1 if stopping SIO

=over

=item Arguments

B<mount_path> => I<mount path on client>

vMount path on the client where to run SIO.

B<client> => I<hostspec>

Client to run SIO from. Default localhost or set by a previous call. Requires a Linux client

B<sio_hash> => I<hash reference>

Reference to the hash for holding SIO related information for this function

B<cycle> => I<integer>

How long to run sio before repeating, in seconds. A special value of zero indicates
only create data files and not to start sio process. Default 900

B<read> => I<percent>

Percent of read access vs write access. Default 100

B<rand> => I<percent>

Percent of random access vs sequential access. Default 100

B<blksize> => I<IO block size>

Size of IO. Default 5k

B<start_byte> => I<integer>

Lower bound for access. Default 0

B<count> => I<integer>

If specified determines number of files to create for sio to use. Default 5

B<file_size> => I<size>

File size of each file used by sio. Default 20m

B<numthreads> => I<integer>

Number of IO generators. Default 1

B<stop> => I<1|0>

Stop the SIO process if set. Default 0

B<restart> => I<integer>

Flag to indicate to start sio with parameters from previous invocation.

B<timeout> => I<integer>

Timeout in seconds for Tharn::say calls in the function.

=item Examples

=back

	#Library path to be included
	use dp::fmc::FcMetroclusterUtils;
	
	#For starting SIO 
	FcMetroclusterUtils::run_sio(mount_path => $mount_path, client => $client );

	#For stopping running SIO 
	FcMetroclusterUtils::run_sio( mount_path => $mount_path,  stop => 1, sio_hash => $sio_hash );
	
	#For restarting SIO on vfiler
	FcMetroclusterUtils::run_sio( mount_path => $mount_path,  restart => 1, sio_hash => $sio_hash );
	
	#Additional parameters for SIO are configurable and can be passed through function call. Example
	# Example: Tell sio to use listed files on client nam9
	FcMetroclusterUtils::run_sio(mount_path => $mount_path, file_list => "f1 f2", client => $client);

	# Perform sio with (25% read, 75% write) 
	FcMetroclusterUtils::run_sio(mount_path => $mount_path, read => 25);

=cut

sub run_sio {
	my $sio_hash; 
	my %sio_hash1;
	my (%args) = validate (@_, {
	            mount_path => {type=>SCALAR, optional=>0},
				client => {type=>SCALAR, optional=>1},
				sio_hash => {type=>HASHREF, optional=>1},
				cycle => {type=>SCALAR, optional=>1},
				read => {type=>SCALAR, optional=>1, default=>50, regex=>qr/\d/},
				rand => {type=>SCALAR, optional=>1, default=>50, regex=>qr/\d/},
				blksize => {type=>SCALAR, optional=>1, default=>"32k"},
				start_byte => {type=>SCALAR, optional=>1, default=>0, regex=>qr/\d/},
				sio_threads => {type=>SCALAR, optional=>1, default=>3, regex=>qr/\d/},
				file_size => {type=>SCALAR, optional=>1},
				numthreads => {type=>SCALAR, optional=>1, default=>1, regex=>qr/\d/},
				stop => {type=>SCALAR, optional=>1, default=>0},
				umount => {type=>SCALAR, optional=>1, default=>0},
				restart => {type=>SCALAR, optional=>1, default=>0},
				timeout => {type=>SCALAR, optional=>1, default=>300},
                proto          => { type => SCALAR, default => 'tcp',
                                    regex => qr/^(udp|tcp)$/ },
                rsize          => { type => SCALAR, default => 32768,
                                    regex => qr/^\d+$/ },
                version        => { type => SCALAR, default => 3 },
                wsize          => { type => SCALAR, default => 32768,
                                    regex => qr/^\d+$/ },
				
			});
	if (defined $args{sio_hash}){
		$sio_hash = $args{sio_hash};
	}
	else {
		$sio_hash = \%sio_hash1;
	}
	croak "sio 'stop' and 'restart' options are mutually exclusive"
		if $args{stop} && $args{restart};
	
	###Validating few input parameters
	$args{client} = $args{client} || $sio_hash->{_client} || "localhost";
	$args{file_size}= "1024"	unless defined $args{file_size};
	croak "numthreads must be > 0"
		unless $args{numthreads} > 0;

	my $host  = Tharn::host( $args{client} );
    my $lnxclnt = Tharn::connect('#rsh');

	###Let's stop SIO if requested
	if ( $args{stop} ) {
		if ( $sio_hash->{_sio} ) {
			NOTE::logcomment("Stopping SIO on $args{mount_path}");
			$sio_hash->{_sio}->raise("TERM");
			_sio_stop_file($sio_hash, "create", $lnxclnt);
			# Lets wait a little to give sio chance to exit
			sleep 60; 
			my $retry = 0;
			while ( $sio_hash->{_sio}->is_alive ) {
				croak "Unable to stop SIO on $args{mount_path} after $retry attempts. Aborting..."
					unless $retry++ < 30;
				NOTE::logcomment("Send kill signal for SIO on $args{mount_path}");
				$sio_hash->{_sio}->kill;
				sleep(15);
			}
			$sio_hash->{_sio}->wait_finish(timeout => 120);
	    	if( !FSR::FSRutils::subtestResults($sio_hash->{_sio})){
			if($@ !~ /Test killed by TERM signal/i){
				croak $@;
			}
		}
			delete $sio_hash->{_sio};
			_sio_stop_file($sio_hash, "remove", $lnxclnt);
			sleep(10);
			NOTE::logcomment("SIO process terminated");
			if($args{umount}){
			    $sio_hash->{umount} = GenericMount::umount( $sio_hash->{mountpoint}, force => "1");
			}
			
		} 
		else {
			croak "Cannot stop SIO. No SIO process running";
		}
		return 1;
	}	
	###We're here means SIO stop was not requested. Let's proceed
	
	$sio_hash->{mountpoint} = $args{mount_path};	
	###Let's restore SIO hash with old parameters if SIO restart is requested
	if ( $args{restart} && exists $sio_hash->{_sio_args} ) {
		NOTE::logcomment("SIO restart requested. Restoring previous SIO parameters");
		delete $args{restart};
		foreach my $key (keys %{$sio_hash->{_sio_args}}) {
			$args{$key} = $sio_hash->{_sio_args}{$key};
		}
		NOTE::logcomment("Completed restoring previous SIO parameters needed for SIO restart\n");
	}
	%{$sio_hash->{_sio_args}} = %args;

	my $rw_mnt = undef;
	my $fname;
	my $idx = 1;
	if (!defined($sio_hash->{sio_threads_name})){
    	NOTE::logcomment ("Creating placeholder data files for running SIO\n");
    	do {
    		my $sio_file = "sio_".$idx."_".$$;
    		my $fullpath = $sio_hash->{mountpoint}."/".$sio_file;
    #        my $fullpath = $args{mount_path}."/".$sio_file;
    		my @cmdoutput =   Tharn::say( $lnxclnt,
                                		"sudo /bin/dd if=/dev/urandom of=" 
                                		. $fullpath . " bs=" 
                                		. $args{file_size}
                                		. "M count=1; echo __D0NE__", 
                                		"__D0NE__", 
                                		-noprompt, 
                                		-timeout, 1800 
                                		);
    		if ($cmdoutput[1] !~ /1\+0 records in([\r\n])*1\+0 records out/ ){
    			croak("Unable to create data files for SIO\n");
    		}
    		Tharn::say( $lnxclnt, "sudo chmod 777 $fullpath; echo __D0NE__", 
    		            "__D0NE__", 
    		            -noprompt, 
    		            -timeout, $args{timeout}
    		           );
    		#Append the full path of each data file to a single string
    		$fname .= "$sio_hash->{mountpoint}/" .$sio_file. " ";
    	} while $idx++ < $args{sio_threads};
    	$sio_hash->{sio_threads_name} = $fname;    
	}
	else{
	    NOTE::logcomment ("Reusing existing SIO thread files for running SIO\n");
	    $fname = $sio_hash->{sio_threads_name};
	    NOTE::logcomment ("SIO thread files: $fname ");
	}
	NOTE::logcomment ("Done creating placeholder data files for running SIO\n");
	return 1 if defined $args{cycle} && $args{cycle} <= 0; # Create and return data files only, do not start sio
	
	#Creating stop file to stop SIO when requested
	#Note: SIO -stop file option only works for linux host, so use short 60 cycle
	#     unless this is the case
	my $stop_cmd = "";
	$stop_cmd = "-stop -"._sio_stop_file($sio_hash, "set",$lnxclnt)
		if Tharn::host($args{client})->hosttype =~ /linux/i;
	$args{cycle} = ($stop_cmd ? 900 : 60) # Default to long cycle time if able to use stop file
		unless defined $args{cycle};
	
	###Create the SIO command
	my $cmdsio = "sudo /usr/software/test/bin/sio"
			." $args{read} $args{rand} $args{blksize} $args{start_byte}"
			." $args{file_size}m $args{cycle} $args{numthreads} $fname $stop_cmd; 
			; echo __D0NE__";
	
	NOTE::logcomment ("Calling a subtest now for starting SIO threads\n");
	my $ret = Tharn::subtest(sub {
					close STDIN;
					my ($rc, $run) = (1, 1);
					#Configure to stop and exit on receiving a terminate signal
					$SIG{TERM} = sub { 
							NOTE::logcomment("Received a TERM signal. Stopping SIO on $args{vol_name}");
							$run = 0;
						};
					#Run continuous SIO loop until stopped by parent
					NOTE::logcomment("Starting SIO on $sio_hash->{mountpoint}");
					my $timeout = $args{cycle} + 180;
					do {
						NOTE::logcomment("Starting SIO on $sio_hash->{mountpoint} yet again, run: $run");
						my @cmdoutput = Tharn::say( $lnxclnt, $cmdsio, "__D0NE__", -noprompt, -timeout, $timeout);
						if ($cmdoutput[1] =~ /FAIL/i || $cmdoutput[1] =~ /ERROR/i){
							$rc = 0; #configure to stop if found any issue
						}
						else{
							$rc++; #Everything looks good. Let's continue
						}
					} while $run && $rc;
					Tharn::logresult("FATAL", "FAIL: Attempt at running SIO on $sio_hash->{mountpoint} failed at iteration = $rc")
						if $rc < 1;
					NOTE::logcomment("SIO stopped on $sio_hash->{mountpoint}");
				}, '-bg', '-onexit', sub {});
	croak "Failed to start SIO: $@"
		if $@;
	$sio_hash->{_sio} = $ret;
	NOTE::logcomment("Completed running SIO successfully\n");
	# Schedule SIO to be stopped during teardown
	NOTE::reference( sub { run_sio( mount_path => $args{mount_path}, client => $args{client}, stop => 1, sio_hash => $sio_hash) if exists $sio_hash->{_sio} } );
	return $sio_hash;
}

=head2 ::nfs_traffic

Reads NFS operations running on a filer at specified interval and returns the average NFS Ops

=over

=item Arguments

B<filer_conn> => I<Filer conn object>

I<Required.>
Filer conn object for the filer to read NFS ops

B<filer_sysstat_count> => I<Numer of iterations to read for calulating average NFS Ops>

I<Optional.>
Defaults to 10


B<filer_sysstat_interval> => I<interval in seconds between each iteration>

I<Optional.>
Defaults to 2


=item Examples

=back


use dp::fmc::FcMetroclusterUtils;
$avg_nfs_ops = FcMetroclusterUtils::nfs_traffic( filer_conn => $mc->me->conn);


=cut


sub nfs_traffic(%) {    
    my (%args) = validate (@_, {
            filer_conn => {type => OBJECT, optional => 0},
            filer_sysstat_count => { type => SCALAR, optional => 1, default => 10 },
            filer_sysstat_interval => {type => SCALAR, optional => 1, default => 2 },
        });
    my %sysstat = filer_sysstat_get(
         $args{filer_conn},
         {
            'FILER_SYSSTAT_COUNT'    => $args{filer_sysstat_count},
            'FILER_SYSSTAT_INTERVAL' => $args{filer_sysstat_interval},
         }
      );
      my $avg_nfs_ops = $sysstat{'NFS'}{'Avg'};
      NOTE::logcomment("Average NFS OPs: $avg_nfs_ops\n");
      return $avg_nfs_ops;    
}

=head2 ::ssl_setup

Setup SSL on filer with default options [Needed for FMC-DataMotion]

=over

=item Arguments

B<filer_conn> => I<Filer conn object>

I<Required.>
Filer conn object for the filer

=item Examples

=back


use dp::fmc::FcMetroclusterUtils;
FcMetroclusterUtils::ssl_setup( filer_conn => $mc->me->conn);


=cut

sub ssl_setup(%) {
    my (%args) = @_;
    croak ("Insuffient Parameter. Provide filer connection") unless exists ($args{filer_conn});
    my $filer_conn = $args{filer_conn};
    my $ctrl_C = sprintf( "%c", 003 );    # [Ctrl]+C char;
    # Create a ReplyTable.
    my %reply_table = (
        "SSL Setup has already been done before. Do you want to proceed?" =>
          "yes",
        "Country Name \\(2 letter code\\)"              => "",
        "State or Province Name \\(full name\\)"        => "",
        "Locality Name \\(city, town, etc\\.\\)"        => "",
        "Organization Name \\(company\\)"               => "",
        "Organization Unit Name \\(division\\)"         => "",
        "Common Name \\(fully qualified domain name\\)" => "",
        "Administrator email"                           => "",
        "Days until expires"                            => "",
        "Key length \\(bits\\)"                         => "",
    );

    # wait for "secureadmin setup ssl" prompts in %ReplyTable, and act on
    # them, until the prompt coming back is one of the built-in
    # say prompts (i.e. the filer's normal prompt, showing that
    # the setup is done)

    say( $filer_conn, " secureadmin setup ssl", '-blind' );
    my ( $tries, $prompt, $reply, $after );
    for ( $tries = 50 ; $tries ; $tries-- ) {
        ( undef, undef, $after, $prompt ) =
          say( $filer_conn, "", '-prompt', [ keys %reply_table ] );

        last if ( matchstrings( $prompt, [ $filer_conn->prompts() ] ) );
        if ( !matchstrings( $prompt, [ keys %reply_table ] ) ) {
            say( $filer_conn, "\r", '-nocr', '-blind' )
              ;    # if we don't have a match, just
                   # take the default
        }
        foreach my $reg_exp ( keys %reply_table ) {
            next unless ( matchstrings( $prompt, $reg_exp ) );
            if ( $tries <= 1 ) {

                # if got stuck, try ^C to terminate the setup.
                $reply = $ctrl_C;
            }
            else {
                $reply = $reply_table{$reg_exp};
            }
            $reply .= "\r";
            say( $filer_conn, $reply, '-nocr', '-blind' );
        }
    }

    # If we are out of tries, something is wrong.
    if ( !$tries ) {
        die "ssl_setup(): too many prompts during \"secureadmin setup ssl\"\n";
    }

    # Enable SSL related options
    my @output = say( $filer_conn, "options ssl.enable on" );
    if ( $output[2] =~ /SSL is not configured./ ) {
        die "ssl_setup() unsuccessful : \"$output[2]\"\n";
    }
    @output = say( $filer_conn, "options httpd.admin.ssl.enable on" );
    if ( $output[2] =~ /SSL is not configured./ ) {
        die "ssl_setup() unsuccessful : \"$output[2]\"\n";
    }
    return 0;
                                                                
}


=head2 ::rpower_devices

Performs rpower ON, OFF or RESET operations on a provided list of rpower names

=over

=item Arguments

B<rpower_list> => I<list of rpower names>

I<Required.>
List of rpower names of devices for performing rpower operation (ON, OFF or RESET)

B<rpower_cmd> => I<Absolute path to rpower command>

I<Optional.>
Defaults to '/usr/software/utils/bin/rpower'

B<client> => I<Client from which to issue rpower command>

I<Optional.>
Defaults to 'localhost'

B<on> => I<Flag to be set to 1 for rpower ON operation>

I<Optional.>
Defaults to 1

B<off> => I<Flag to be set to 1 for rpower OFF operation>

I<Optional.>
Defaults to 0

B<reset> => I<Flag to be set to 1 for rpower RESET operation>

I<Optional.>
Defaults to 0


=item Examples

=back


use dp::fmc::FcMetroclusterUtils;
FcMetroclusterUtils::rpower_devices( rpower_list => \@rpower_list);

OR

use dp::fmc::FcMetroclusterUtils;
FcMetroclusterUtils::rpower_devices( rpower_list => \@rpower_list, off => 1);


=cut


sub rpower_devices{
    my (%args) = validate (@_, {
                rpower_list => { type => ARRAYREF, optional => 0 },
                rpower_cmd => { type => SCALAR, optional => 1, default => '/usr/software/utils/bin/rpower' },
                client => {type=>SCALAR, optional=>1},
                on => {type=>SCALAR, optional=>1, default=>1, regex=>qr/\d/},
                off => {type=>SCALAR, optional=>1, default=>0, regex=>qr/\d/},
                reset => {type=>SCALAR, optional=>1, default=>0, regex=>qr/\d/},                
            });
    my (@subtest_obj, $rpower_task);
    my $rpower_list_ref = $args{rpower_list};
    my @rpower_list = @$rpower_list_ref;

    if ($args{reset}){
        $rpower_task = '';
    }
    elsif ($args{off}){
        $rpower_task = 'off';
    }
    else{
        $rpower_task = 'on';
    }
    my $client = $args{client} || 'localhost';
    #Create an object for client to issue rpower command
    my $host    = Tharn::host($client);
    my $client_conn = Tharn::connect('#rsh');
    
    #Setting the time when subtests will do rpower off of filer/shelves
    my $rpower_sync_time = Tharn::timeout2time(3);
    NOTE::logcomment("RPower Device List: @rpower_list \n");                        
    #Calling subtests to do rpower off of shelves & filer                          
    foreach my $rpower_name (@rpower_list){            
        my $subtest = Tharn::subtest(sub{
                           while ( time() < $rpower_sync_time){
                               sleep(1);
                           }
                           my $retry = 4; #Rpower cannot handle many commands at same instance so let's retry if needed
                           RETRY:
                           NOTE::logcomment("$rpower_name: Issuing rpower $rpower_task command");
                           my $output = SayWrap::sayEOF($client_conn, "$args{rpower_cmd} $rpower_name $rpower_task" );
                           if( $output =~ /rpower: Command not found/ ) { 
                               croak ("Unable to locate rpower command at $args{rpower_cmd}\n");
                           }
                           elsif( $output =~ /Error during expect sequence/i ) { 
                               if ($retry){
                                    NOTE::logcomment("$rpower_name: Rpower command failed. Let's retry once more!\n");
                                    $retry--;
                                    goto RETRY; 
                               }
                               else{
                                    croak ("Rpower internal error! \n$rpower_name: Unable to do rpower $rpower_task\n");
                               } 
                           }
                           elsif($output =~ /sending '$rpower_task' request to sentry CDU/i) {
                               NOTE::logcomment("$rpower_name: Rpower command succeeded! rpower $rpower_task in progress\n"); 
                               sleep(2);                             
                           }
                           else{
                               croak ("FAILED: Unable to do rpower $rpower_task for $rpower_name");
                           }
                           NOTE::logcomment("Subtest Completed Successfully");
                       },'-bg', '-onexit', sub {});
        croak "Subtest failed for $rpower_name with error: $@" if($@);  
        push(@subtest_obj, $subtest);  
    }
    sleep(2);
    #Waiting for subtests to finish
    $subtest_obj[$#subtest_obj]->wait_finish(timeout => 60);
    foreach my $subtest(@subtest_obj){
        $subtest->wait_finish(timeout => 60) if($subtest->is_alive); 
        FSR::FSRutils::subtestResults($subtest);
        if($@){
            logresult("FAIL","Subtest(s) failed with error: $@");
            foreach my $rpower_name (@rpower_list){
                SayWrap::sayEOF($client_conn, "$args{rpower_cmd} $rpower_name on" );
            }
            croak("Failed to do rpower $rpower_task for device(s)");
        }           
    }        
    NOTE::logcomment("Completed rpower $rpower_task for device(s).\n");         
 }

    
1;
