# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/dp/lockvault/Utils.pm#1 $
##
## @summary lockvault::Utils library
##	Utils methods for LockVault test scripts
##
## @author sydneyw@netapp.com, dl-dpgqa-automation@netapp.com
##

=head1 NAME

lockvault::Utils - Utility methods for creating a LockVault environment using NOTE objects for the purposes of simplifying test script writing.

=head1 SYNOPSIS

use dp::lockvault::Utils;

=cut

package dp::lockvault::Utils;

use strict;
use Carp;
use Tharn;

use Data::Dumper;
use NOTE;

BEGIN {
    Params::Validate::validation_options(
	on_fail => sub { my @msgs = @_; chomp(@msgs);
			Carp::confess(Tharn::script_result(@msgs)) }
    );
}

=head2 ::value

Returns value for the key-value pair stored within this object.
The keys are typivally the command line arguments for the lockvault
test

=over

=item key

The key for which the value is requested. The keys corresponds
to the command line parameters. The current list is as follows:

FILER_SRC = Required parameter
FILER_DST = Defaults to FILER_SRC
FILER = Defaults to FILER_SRC
FILER_SRC_AGGR = Existing named Src Aggr. Defaults to undef
FILER_DST_AGGR = Existing named dst Aggr. Defaults to undef
FILER_SRC_QTREE = Existing named src qtree. Defaults to undef
FILER_DST_QTREE = Existing named dst qtree. Defaults to undef
FILER_LOG_VOL = Optional. Existing named log volume. Defaults to undef
VOL_TYPE = flex or trad. Defaults to flex
VOL_SIZE = Defaults to 10g for src and dst and 5g for log
LOCKVAULT = Optional. compliance or enterprise. Defaults to enterprise
MIN_PERIOD = Minimum retention period. Defaults to 60s
MAX_PERIOD = Maximum retention period. Defaults to MIN_PERIOD
DEF_PERIOD = Default retention period. Defaults to MIN_PERIOD

=item Examples

=back

     $utils->value('FILER_SRC', 'fsr-u37');

     my $val = $utils->value('VOL_SIZE');

=cut

sub value {
    my ($self, $key) = @_;

    $self->_get_param()
	unless $self->{_inited};
    my $param = lc $key;
    if ( exists $self->{$param} ) {
	return $self->{$param};
    } else {
	return undef;
    }
}

sub _get_param {
    my ($self) = @_;

    return if $self->{_init};
    $self->{_inited} = 1;

    $self->{filer} = param("FILER", -default, undef);
    $self->{filer_src} = param("FILER_SRC", -default, undef);
    croak "FILER or FILER_SRC must be set"
	unless ($self->{filer} or $self->{filer_src});
    $self->{filer_src} = $self->{filer} unless $self->{filer_src};
    $self->{filer_dst} = param("FILER_DST", -default, $self->{filer_src});
    $self->{filer_log} = $self->{filer_dst};

    $self->_param("FILER_SRC_AGGR", undef);

    $self->_param("FILER_DST_AGGR", undef);

    $self->_param("FILER_SRC_QTREE", undef);
    $self->_param("FILER_DST_QTREE", undef);
    $self->_param("FILER_LOG_VOL", undef); 

    $self->{vol_type} = param("VOL_TYPE", -default, "flex");
    $self->{vol_size} = param("VOL_SIZE", -default, undef);
    $self->{lockvault} = param("LOCKVAULT", -default, "enterprise");
    croak "LOCKVAULT must be one of 'compliance' or 'enterprise'"
	unless $self->{lockvault} =~ /compliance|enterprise/;
    $self->{min_period} = param("MIN_PERIOD", -default, "120s");
    $self->{max_period} = param("MAX_PERIOD", -default, $self->{min_period});
    $self->{def_period} = param("DEF_PERIOD", -default, "min");
}

sub _param {
    my ($self, $key, $def) = @_;

    my $val = param($key, -default, $def);
    $self->{lc($key)} = $val if $val;
}

sub _set_lv_objects {
    my ($self, %args) = @_;

    $self->{src} = (exists $args{src} ? $args{src} : $self->make_lv_object('src'));
    $self->{dst} = (exists $args{dst} ? $args{dst} : $self->make_lv_object('dst'));
    $self->{log} = (exists $args{log} ? $args{log} : $self->make_lv_volume('log'));
}

=head2 ::new

Method to create a Utils object. It first processes the command line arguments to form
key-vault pairs with this object, and then it initialises the src, dst and log vault
objects.

=over

=item %args

Hash of arguments for:

B<src> => I<NOTE::Qtree object>

A qtree object used for the lockvault soruce. Optional. If not supplied one is created.

B<dst> => I<NOTE::Qtree object>

A qtree object used for the lockvault destination. Optional. If not supplied one is created.

B<log> => I<NOTE:Volume object>

A Volume object used for the lockvault log volume. Optional. If not supplied one is created.

=item Examples

=back

     my $utils = dp::lockvault->new(src=>$qt_src, dst=>$qt_dst);

     my $utils = dp::lockvault->new();

=cut

sub new {
    my ($class, %args) = @_;
    my $self = {};

    bless $self, $class;
    $self->_get_param();
    $self->_set_lv_objects(%args);

    return $self;
}

=head2 ::create_vault

Creates a lockvault object by calling the LockVault::start method with the appropriate
src, dst and log objects together with the retention times.

=over

=item %args 

Hash of arguments

B<src> => I<NOTE::Qtree object>

A qtree object used for the lockvault soruce. Optional. If not supplied one is created.

B<dst> => I<NOTE::Qtree object>

A qtree object used for the lockvault destination. Optional. If not supplied one is created.

B<log> => I<NOTE:Volume object>

A Volume object used for the lockvault log volume. Optional. If not supplied one is created.

=item Examples

=back

      my $lv = dp::lockvault::create_vault();

      my $lv = dp::lockvault::create_vault(src=>$qt_src, dst=>$qt_dst, log=>$log_vol);

=cut

sub create_vault {
    my ($class, %args) = @_;
    my $self = {};

    if ( ref $class eq '' ) {
	$self = dp::lockvault::Utils->new(%args);
    } else {
	$self = $class;
	$args{src} = $self->{src} unless exists $args{src};
	$args{dst} = $self->{dst} unless exists $args{dst};
	$args{log} = $self->{log} unless exists $args{log};
	$self->_set_lv_objects(%args);
    }

    my %opts;
    $opts{src} = $self->{src};
    $opts{dst} = $self->{dst};
    $opts{log} = $self->{log};
    $opts{snaplock_minimum_period} = $self->{min_period} if exists $self->{min_period};
    $opts{snaplock_maximum_period} = $self->{max_period} if exists $self->{max_period};
    $opts{snaplock_default_period} = $self->{def_period} if exists $self->{def_period};

    return NOTE::LockVault->start(%opts);
}

=head2 ::make_lv_object

Creates the lockvault src, dst or log object

=over

=item tag

Determines which object to create, src, dst or log

=item Examples

=back

     my $src_qt = $utils->make_lv_object("src");

     my $log_vol = $utils->make_lv_object("log");

=cut

sub make_lv_object {
    my ($self, $tag) = @_;

    $self->_get_param()
	unless $self->{_inited};

    my $id = "filer_".$tag."_qtree";
    my $qtree;
    if  ( $self->{$id} ) {
	$qtree = NOTE::Qtree->attach(name=>$self->{$id});
    } else {
	my $vol = $self->make_lv_volume($tag);
	if ( $tag eq "src" ) {
		$qtree = NOTE::Qtree->create(Volume=>$vol, tag=>$tag);
	} else {
		$qtree = NOTE::Qtree->new(Volume=>$vol, tag=>$tag);
	}
    }
    return $qtree;
}

=head2 ::make_lv_volume

Create a volume with the lockvault settings.

=over

=item tag

Create a specific lockvault object, src, dst or log

=item Example

=back

      my $log_vol = $utils->make_lv_volume("log");

=cut

sub make_lv_volume {
    my ($self, $tag) = @_;

    $self->_get_param()
	unless $self->{_inited};

    my %args;

    $args{tag} = $tag;
    $args{name} = $self->{"vol_".$tag};
    $args{Filer}{host} = $self->{"filer_".$tag};
    $args{type} = $self->{vol_type};
    $args{recycle} = 1;
    if ( $args{type} =~ /flex/i ) {
	$args{Aggregate} =
		{ recycle => 1, disks => 3, name => $self->{"filer_".$tag."_aggr" }};
	$args{size} = $self->{vol_size} || ($tag =~ /log/ ? "5g" : "10g");
	$args{guarantee} = "none";
    } else {
	$args{size} = 2;
	$args{zeroed} = 1;
    }
    if ( $tag =~ /dst|log/i ) {
	$args{snaplock} = $self->{lockvault};
	$args{Aggregate}{snaplock} = $args{snaplock} if exists $args{Aggregate};
	$args{recycle} = ($args{snaplock} eq "enterprise"); # Compliance Volumes cannot be recycle
    }

    return NOTE::Volume->create(%args);
}

=head2 ::wait_for_cache_to_flush

The method causes to script to pause until 10 secs pass the next minute, as the
top of each minute is when cached logs are flushed due to lockvault activities every
minute. The added 10 secs is to give time for the cached to be flushed fully.

=over

=item filer

The filer's system time to used to determine how long to wait

=item Example

=back

     $utils->wait_for_cache_to_flush($lv->dst->Filer);

=cut

sub wait_for_cache_to_flush {
    my ($self, $filer) = @_;

    my ($sec, $min, $hr);
    ($sec, $min, $hr) = gmtime($filer->time);
    my $pause = 70 - $sec;
    NOTE::logcomment("Current time: ".sprintf("%02d:%02d:%02d", $hr, $min, $sec)
			.", pausing $pause secs");
    while ( $pause > 0 ) {
    	$pause -= sleep($pause);
    }
    ($sec, $min, $hr) = gmtime($filer->time);
    NOTE::logcomment("Current time: ".sprintf("%02d:%02d:%02d", $hr, $min, $sec)
			.", cached log(s) should have been flushed");
}

sub _search_log {
    my ($self, $log_path, $recovery_only, @expected) = @_;

    # Read in relevant log entries
    open F, "<$log_path";
    my $logs;
    my $in_recovery = 0;
    while (<F>) {
	# Return last recovery section or the full log file
	if ( $_ =~ /WORM_LOG_FAILURE_RECOVER/ ) {
		if ( $recovery_only ) {
			$in_recovery = ( $_ =~ /START/ );
			$logs = "" if $in_recovery; # Capture last section only
		}
		$logs .= $_;
	} else { 
		$logs .= $_
			if ($in_recovery or ! $recovery_only);
	}
    }
    close F;
    NOTE::logcomment(($recovery_only ? "Recovery" : "All")." log entries:\n$logs");
    # Search log entries for expected entries
    my @missing;
    foreach my $entry (@expected) {
	if ( $logs =~ /$entry/ ) {
		NOTE::logcomment("Found expected entry: $entry")
	} else {
		push @missing, $entry;
	}
    }
    NOTE::logcomment("Missing expected entries:\n".(join "\n", @missing))
	if @missing;

    return @missing;
}

=head2 ::search_log

Search the whole of a given log file for a list of expected entries

Returns 1 if all entries found, and 0 otherwise

=over

=item log_path

Full path of the log file to search for

= item expected

An array of expected entries

=item Example

=back

     my $result = $utils->search_log($hist_log_path, @expected_entries);

=cut

sub search_log {
    my ($self, $log_path, @expected) = @_;

    my $all_log = 0;
    return ! $self->_search_log($log_path, $all_log, @expected);
}

=head2 ::search_recovery_log

Search the last recovery section for the given expected entries.
If the given log file contains multiple recovery entries, it will
only search the latest one.

Returns an array of entries that is it did not find.

=over

=item log_path

Full path of the log to search locate the latest recovery section.

=item expected

An array of expected entries.

=item Example

=back

      my $result = $utils->search_recovery_log($hist_log_path, @expected_entries);

=cut

sub search_recovery_log {
    my ($self, $log_path, @expected) = @_;

    my $recovery_only = 1;
    return $self->_search_log($log_path, $recovery_only, @expected);
}

1;
