# -*-perl-*-
# $Id: //depot/prod/test/maindp/snapdiff/zapi/lib/Setup.pm#1 $

##
## @synopsis 
##   Common setup needed for majority of the Snapdiff zapi test cases.
##   The high level wrapper functions snapdiff_zapi_setup and snapdiff_zapi_cleanup
##   should be typically invoked by scripts but other functions
##   The argument to these functions is a reference to the hash e.g. \%set_h
##   The hash is populated in by the functions in this library and returned to the
##   caller (i.e. test script).
##   The naming convention for hash indices are as follows
##   Key names starting with CAPs are NACL objects
##   E.g. "Node" is used denote the index for a node object and the node object
##   can be accessed from the script as $set_h{Node}
##   Key names starting with small letters are names
##   E.g. "node" is used to denote the index for the name of an node and the node name
##   can be accessed from the script as $set_h{node}
##   Key names ending with 's' denotes array/list
##   E.g. "Nodes" is used to denote the index for the array of node objects and can
##   can be accessed from the script as $set_h{Nodes}
##   The list of valid indices used by this module are
##   Node related
##   a) Nodes - Array of NACL Node objects
##   b) Node  - Single NACL Node object
##   c) nodes - Array of node names
##   d) node  - Node name (as returned by NACL libs)
##   Aggregate related
##   a) Aggregate - Single NACL Aggregate object
##   b) 
##   Vserver related
##   a) Vserver          - Single NACL Vserver object
##   b) vserver          - Vserver name
##   c) vserver_validate - flag locally used to check if vserver is created or not
##   d) vserver_security_style - security style
##   e) vserver_language       - For the vserver language setting
##   Volume related
##   a) volume           - Volume name
##   b) volume_path      - Volume path (used for 7mode mainly)
##   Flags and other params
##   a) is_repository    - same as $main::REPOSITORY
##   b) enable_snapdiff  - same as $main::ENABLE_SNAPDIFF
##   c) mode             - 7mode or Cmode
##   d) Zapi             - ZAPI object 
##   e) username         - username used for ZAPI access
##   f) password         - password used for ZAPI access
##   TODO: Need to validate the hash indices when the caller invokes any of the functions
##   in this library
##
## @author  susheel@netapp.com burra@netapp.com
##
## @status  public
##

package dp::snapdiff::zapi::lib::Setup;

use strict;
use warnings;
use base qw(Exporter);

use Tharn;
use TCD;

use Net::Ping;
use Data::Dumper;

use Params::Validate qw(validate_with UNDEF HASHREF SCALAR HASHREF);
use NATE::BaseException qw{:try};

use NACL::C::Volume;
use NACL::C::StorageDisk;
use NACL::C::Vserver;
use NACL::C::VserverServicesUnixUser;
use NACL::C::VserverServicesUnixGroup;
use NACL::C::VserverServicesNisDomain;
use NACL::STask::Node;
use NACL::STask::Aggregate;
use NACL::STask::VolumeSnapshot;
use NACL::STask::Volume;
use NACL::MTask::VserverSetup;
use NACL::Exceptions::NoElementsFound qw{:try};
use NACL::STask::SystemServicesFirewallPolicy;
use NACL::Exceptions::SetupFailure::VserverSetupFailure qw{:try};
use NACL::C::Exceptions::SystemServicesFirewallPolicy::AlreadyExists qw{:try};
use NACL::C::Exceptions::SystemServicesFirewallPolicy::DoesNotExist qw{:try};

use dp::snapdiff::zapi::lib::ZapiSD;
use dp::snapdiff::zapi::lib::SDutils qw(do_mount create_dataset modify_dataset);
use dp::snapdiff::zapi::lib::CommonParams;

our $VERSION = '1.0';

our @EXPORT_OK = qw(
    aggr_exists
    aggr_create
    vserver_exists
    vserver_create
    vserver_delete
    vserver_add_nis
    vol_exists
    vol_create
    create_snapshot
    snapshots_setup
    snapdiff_zapi_setup
    snapdiff_zapi_cleanup
);


#--------------------------------------------------
# Prototypes
#--------------------------------------------------
sub aggr_exists;
sub aggr_create;
sub vserver_exists;
sub vserver_create;
sub vserver_delete;
sub vserver_add_nis;
sub vol_exists;
sub vol_create;
sub create_snapshot;
sub snapshots_setup;
sub snapdiff_zapi_setup;
sub snapdiff_zapi_cleanup;
sub get_ipaddr;

#--------------------------------------------------
# Package local variables
#--------------------------------------------------
my $Node;
my $Vserver;
my $Volume;
my $Aggregate;
my $Vserver_obj;
my $allow_pol = 0;


##
## internal function: picked up from CR/commonSetup.pm
##
sub _bg_job_helper {
    my %opt = validate_with(
        params => \@_,
        spec   => {
            'job'      => { type => HASHREF },
            'wait_msg' => { type => SCALAR | UNDEF, optional => 1 },
            'warn_msg' => { type => SCALAR | UNDEF, optional => 1 },
            'erro_msg' => { type => SCALAR | UNDEF, optional => 1 },
        },
    );
    my $job      = $opt{'job'};
    my $wait_msg = $opt{'wait_msg'};
    my $warn_msg = $opt{'warn_msg'};
    my $erro_msg = $opt{'erro_msg'};
    my $job_state;
    $Log->comment($wait_msg) if ($wait_msg);
    {
        try {
            $job_state = $job->get_one_state_attribute('state');
        }
        catch NACL::Exceptions::NoElementsFound with {
            $Log->warn(
                "Job disappear from job list, may be it's changing state, retry again"
            );
            $Log->comment(
                "Job disappear from job list, may be it's changing state, retry again"
            );
            $job_state = $job->get_one_state_attribute('state');
        };
        sleep(5) if ( $job_state =~ m/queued/i );
    };
    if ( $job_state !~ m/(running|queued)/i && $warn_msg ) {
        $Log->warn($warn_msg);
    }
    $job->watch_progress('method-timeout' => 20);
    if (   $job->get_one_state_attribute('state') !~ m/success/i
        && $erro_msg )
    {
        NACL::APISet::Exceptions::ResponseException->throw(
            text            => $erro_msg,
            response_object => $job
        );
    }
}


##
## @name vserver_exists
##
## @description
##   Function to check if the specified vserver exists
##
## @arg $opts   Required: Reference to the hash that contains info as specified
##              in the synopsis
##              
## @return
##   $TCD::PASS on Success, $TCD::FAIL on Failure
##   Sets flag vserver_validate to 0/1 based on the validation
##
sub vserver_exists {

    my $opts = shift;
    my $is_repository;

    if($opts->{mode} ne "7Mode")  {
        $Vserver  = NACL::C::Vserver->find(
                            command_interface => $opts->{Node},
                            canned_filters => [ qw(data_vservers) ],
                            filter         => {  vserver => $opts->{vserver}},
                            allow_empty       => 1,
                    );
        if (!$Vserver) {
            $Log->comment('Unable to find specified vserver! need to create one');
            return $TCD::FAIL;
        }
        # Validate REPOSITORY param against the vserver specified
        if ($Vserver->get_one_state_attribute('is-repository') eq 'true') {
            $Log->comment('Vserver is repository');
            $is_repository = 'true';  
        } else {
            $is_repository = 'false';  
        }
        if ($opts->{is_repository} ne $is_repository) {
            $Log->comment('Vserver specified is ' . $opts->{vserver});
            $Log->comment('REPOSITORY param passed is ' . $opts->{is_repository});
            $Log->comment('Vserver repository attribute is' . 
                            $Vserver->get_one_state_attribute('is-repository'));
            $Log->comment('Vserver does not match the repository param');
            $opts->{vserver_validate} = 0;
            return $TCD::FAIL;
        }
    }
    $opts->{Vserver} = $Vserver;

    # Check a few things before confirming that the vserver exists
    # Check if network interface is created

    return $TCD::PASS;
}

##
## @name vserver_create
##
## @description
##
## @arg $opts   Required: Reference to the hash that contains info as specified
##              in the synopsis
##
## @return
##   $TCD::PASS on Success, $TCD::FAIL on Failure
##   Sets flag vserver_validate to 0/1 based on the validation
##
sub vserver_create {

    my $opts = shift;

    $Log->comment('Create vserver which is of type REPOSITOY= ' . 
                    $opts->{is_repository});
	my $repo_flg = $opts->{is_repository};

    if (!aggr_exists($opts)) {
        $Log->comment('Aggregate does not exist, create it');
        aggr_create($opts);
    }

    # Populate all vserver params
    if (!defined($opts->{vserver_security_style})) {
        # Use mixed security style by default
        $opts->{vserver_security_style} = "mixed";
    }
    if (!defined($opts->{vserver_language})) {
        # Use C language as default
        $opts->{vserver_language} = "C";
    }
    my $root_vol = $opts->{vserver} . '_rootvol';
    # TODO: use these for now, need to change later
    # In future, take ns-switch and nm-switch in options
    my @ns_switch  = ['file', 'nis']; 
    my @nm_switch  = ['file'];
    my @data_proto = ['nfs', 'cifs', 'fcache'];

    # vserver params
    my $vserver_params = {
            vserver                      => $opts->{vserver},
            aggregate                    => $opts->{aggregate},
            "rootvolume"                 => $opts->{vserver} . '_rootvol',
            "ns-switch"                  => @ns_switch,
            "nm-switch"                  => @nm_switch,
            "rootvolume-security-style"  => $opts->{vserver_security_style},
            "language"                   => $opts->{vserver_language},
            'snapshot-policy'            => 'default',
            'is-repository'              => $opts->{is_repository},
    };
    # aggregate params
    my $aggr_params = {
            aggregate         => $opts->{aggregate},
            raidtype          => "raid4",
            diskcount         => 3,
            if_exists         => 'reuse',
    };

    # network interface params
    # TODO: Need to change this later
    #my @nodes = NACL::STask::Node->find();
    my $host    = $opts->{Node}->hostrec();
    my $free_ip = get_ipaddr();

    if ($free_ip !~ /^([\d]+)\.([\d]+)\.([\d]+)\.([\d]+)$/) {
        $Log->error("Unable to find IP addr for data LIF creation");
        return $TCD::FAIL;
    }
 
    my $netmask = $host->{"DATA_MASK1"};

    my $network_interface_params = {
            lif               => 'data_' . $host->id(),
            vserver           => $opts->{vserver},
            role              => 'data',
            'home-node'       => $host->id(),
            'home-port'       => 'e0c',
            address           => $free_ip,
            netmask           => $netmask,
    };

    my %protocols = (
        'v3'        => 'enabled',
        'v4.0'      => 'enabled',
        'v4.1'      => 'enabled',
        'v4.1-pnfs' => 'enabled',
    );
    my %nfs_options = (
        'v4.0-acl'              => 'enabled',
        'v4.0-read-delegation'  => 'enabled',
        'v4.0-write-delegation' => 'enabled',
        'v4-id-domain'          => $main::NIS_DOMAIN,
    );
    my $routing_group = get_routing_grp($host->{'DATA_GATEWAY1'});
	$Log->comment("routing group :". Dumper($routing_group));
	
    # NFS params
    # TODO: Add all other options for NFS
    my $nfs_params = {
            vserver       => $opts->{vserver},
            if_exists     => 'die',
            %protocols,
            %nfs_options,
    };

    # export policy
    my $export_policy_params = {
            vserver           => $opts->{vserver},
            policyname        => 'default',
    };

    # export policy rule
    my $export_policy_rule_params = {
            vserver           => $opts->{vserver},
            policyname        => 'default',
            clientmatch       => '0.0.0.0/0',
            'rorule'          => ['any'],
            'rwrule'          => ['any'],
            'superuser'       => ['any'],
            'anon'            => '0',
    };
    # security login params
    my $security_login_params = {
	        vserver     => $opts->{vserver},
	        username    => 'vsadmin',
	        force       => '1',
	        'oldpwd'    => 'netapp1!',
	        'newpwd'    => 'netapp1!',
	        newpwd2     => 'netapp1!',
    };
	
	my $routing_grp_params = {
                'vserver'         => $opts->{vserver},
				'routing-group'   => $routing_group,
                'destination'     => '0.0.0.0/0',
                'gateway'         => $host->{'DATA_GATEWAY1'},
    };

    # Create vserver
    try {
        $Vserver  = NACL::MTask::VserverSetup->setup(
            command_interface           => $opts->{Node},
            vserver_params              => $vserver_params,
            aggregate_params            => $aggr_params,
            network_interface_params    => $network_interface_params,
            nfs_params                  => $nfs_params,
			routing_grp_params          => $routing_grp_params,
            export_policy_params        => $export_policy_params,
            export_policy_rule_params   => $export_policy_rule_params,
            security_login_params       => $security_login_params,
        );
    }
    catch NACL::Exceptions::SetupFailure::VserverSetupFailure with {
        $Log->error("Error: Vserver create failed");
        return $TCD::FAIL;
    }

    $opts->{Vserver} = $Vserver->{vserver_obj};
    # TODO: More testing required - need to see if this step is required
	if(!defined($opts->{flex_vol})){
    if ($repo_flg eq 'true') {
        $Log->comment ('This is a repository Vserver, create aggrs on all nodes');
        if (!repos_vserver_add_aggrs($opts)) {
            $Log->comment("Adding aggr list to vserver failed");
            $Log->comment("Cleanup: Delete the vserver we created");
            vserver_delete($opts);
            return $TCD::FAIL;
        }
    }
	}
	# my $routing_group;
	# my @data_objs = @{$Vserver->{data_lif_obj}};
	# foreach my $data_obj(@data_objs){
	    # if ($data_obj){
		 # $routing_group = $data_obj->get_one_state_attribute('routing-group');
		 # $Log->comment("Fetch routing group name to create a routing group");
         # $Log->comment("Routing group name is " . $routing_group);
		# last;
		# }
	  # }
    #my $routing_group = $Vserver->{data_lif_obj}->get_one_state_attribute('routing-group');
     

    # Create route for this routing group
    # $Log->comment("Create route for this routing group");
    # my $route = NACL::C::Route->create(
                # command_interface => $opts->{Node},
                # 'vserver'         => $opts->{vserver},
                # 'routing-group'   => $routing_group,
                # 'destination'     => '0.0.0.0/0',
                # 'gateway'         => $host->{'DATA_GATEWAY1'},
    # );
    # if (!$route) {
        # $Log->comment("Route create failed");
        # $Log->comment("Cleanup: Delete the vserver we created");
        # vserver_delete($opts);
        # return $TCD::FAIL;
    # }

    # No checks here - expected to pass always?
    vserver_add_nis($opts);
	# $Log->comment("Setting $opts->{vserver} password");
        # NACL::C::SecurityLogin->password(
            # command_interface => $opts->{Node},
            # username          => 'vsadmin',
            # vserver           => $opts->{vserver},
            # newpwd =>         =>'netapp1!',
        # );
        # $Log->comment("Unlock $opts->{vserver} account");
        # NACL::C::SecurityLogin->unlock(
            # command_interface => $opts->{Node},
            # username          => 'vsadmin',
            # vserver           => $opts->{vserver},
        # );
    

    return $TCD::PASS;
}



##
## internal function: 
## Gets the routing-group
## TODO: Currently, we just look at DATA_IP1 and DATA_IP2 
## in the tharnhost file
## d10.225.164.0/22
##  
sub get_routing_grp{
    my $gate_way = shift;
	chop($gate_way);
	$gate_way = "d".$gate_way."0/22"; 
	$Log->comment("gate_way :". Dumper($gate_way));
	return $gate_way;
}
##
## @name vserver_add_nis
##
## @description
##   Function to configure NIS for the specified vserver
##
## @arg $opts   Required: Reference to the hash that contains info as specified
##              in the synopsis
##
## @return
##   $TCD::PASS on Success, $TCD::FAIL on Failure
##
    # Assume that the local users and groups are configured 
    # through /etc/passwd and /etc/groups
sub vserver_add_nis {
    my $opts = shift;
    my $nis_domain;
    my @nis_servers;
	
    # TODO: Works only for SVL - need to enhance this later 
    $nis_domain  = $main::NIS_DOMAIN;
    @nis_servers = $main::NIS_SERVERS;

    NACL::C::VserverServicesNisDomain->create(
        command_interface => $opts->{Node},
        'domain'          => $nis_domain,
        'servers'         => \@nis_servers,
        'vserver'         => $opts->{vserver},
        'active'          => "true",
    );
    return $TCD::PASS;

}

##
## internal function: 
## Gets the first available IP address in a range  
## TODO: Currently, we just look at DATA_IP1 and DATA_IP2 
## in the tharnhost file
##  
sub get_ipaddr {

    # TODO: For now, try DATA_IP1 and DATA_IP2
    # This needs to be enhanced later
    my @nodes = NACL::STask::Node->find();
    my $p = Net::Ping->new();
    my @ipaddrs;
	
    # Construct IP addresses list
    foreach my $node (@nodes) {
        if ((!defined $node->hostrec()->{"DATA_IP1"}) &&
            (!defined $node->hostrec()->{"DATA_IP2"}))  {
			 $Log->comment("Data IPs not in tharnhost file");
        } else {
            push (@ipaddrs, ($node->hostrec()->{"DATA_IP1"}, 
                       $node->hostrec()->{"DATA_IP2"}));
        } 
    }
    # check for available IP addresses and return the first available
    my $flg = 0;
    foreach my $ip (@ipaddrs) {
        if ($ip =~ /^([\d]+)\.([\d]+)\.([\d]+)\.([\d]+)$/) {
            $Log->comment("Found valid IP Address");
            if ( $p->ping($ip) ) {
                $Log->comment("$ip answered");
            } else {
                $Log->comment("$ip did not answer");
                return $ip;
            }
        }
    }
    $Log->error("No IP addresses available for creating data LIF");
    return $TCD::FAIL;
}


##
## @name repos_vserver_add_aggrs
##
## @description
##   Add aggregates to the repository vserver
##
## @arg $opts   Required: Reference to the hash that contains info as specified
##              in the synopsis
##
## @return
##   $TCD::PASS on Success, $TCD::FAIL on Failure
##
sub repos_vserver_add_aggrs {
    my $opts = shift;

    my @nodes; 
    my @aggr_list;
    # Initially aggr_list will be '-' need to excluded
    #@aggr_list = $Vserver->get_one_state_attribute('aggr-list');

    my $index  = 1;
    #@nodes = $opts->{nodes};
    @nodes = NACL::STask::Node->find();
    #print Dumper @nodes;
    # Create one extra aggr for ns volume 
    my $actual_aggr = $opts->{aggregate};
    foreach my $n (@nodes, $nodes[0]) {
        my $aggr_name = $opts->{vserver} . "_aggr" . $index++;
        push @aggr_list, $aggr_name;
        # Check if this aggregate exists, if yes, pass
        #$opts->{aggregate} = $aggr_name;
        #next if aggr_exists();

        my $node = $n->{node};
        $Log->comment("Create $aggr_name on node $node");
        my $Aggr = NACL::STask::Aggregate->create(
                            command_interface => $n,
                                    aggregate => $aggr_name,
                                    if_exists => 'reuse',
                                        nodes => [$node],
                                     raidtype => 'raid4',
                                    diskcount => 3,
        );
        if (!$Aggr){
            $Log->error("Failed to create Aggr $aggr_name");
            return $TCD::FAIL;
        } 
        #$Log->comment("Aggr " . $Aggr->aggregate . "created successfully");
    }
    $opts->{Vserver}->modify('aggr-list'=>\@aggr_list);
    my @vs_aggr_list = $opts->{Vserver}->get_one_state_attribute('aggr-list');
    $Log->comment("Vserver aggr-list after adding aggr @vs_aggr_list ");

    return $TCD::PASS;
}

##
## @name vserver_delete
## @description
##   Delete the specified vserver  
##
## @arg $opts   Required: Reference to the hash that contains info as specified
##              in the synopsis
##
## @return
##   $TCD::PASS on Success, $TCD::FAIL on Failure
##
sub vserver_delete {

    my $opts = shift;
    
	if(!defined($opts->{flex_vol})){
    my @repos = NACL::C::Volume->find(
        command_interface => $opts->{Node},
        allow_empty       => 1,
        filter            => {
            'volume-style' => 'infinitevol',
            'vserver'      => $opts->{vserver}
        }
    );
    foreach my $repo (@repos) {
        if ($repo->get_one_state_attribute('junction-active') =~ m/true/i ) {
            $Log->comment("Unmount infinivol");
            $repo->unmount();
        }
        unless ( $repo->get_one_state_attribute('state') =~ m/(-|offline)/i )
        {
            my $job;
            $Log->comment("Offline infinivol using async command");
            $repo->offline(
                'foreground'    => 'false',
                'job_component' => \$job
            );
            _bg_job_helper(
                'job'      => $job,
                'wait_msg' => 'Waiting for infinivol to go offline',
                'erro_msg' => 'Volume offline job fail',
                'warn_msg' => "Offline doesn't seem to be a background job",
            ) if $job;
        }
        my $job;
        $Log->comment("Delete infinivol using async command");
        $repo->delete( 'job_component' => \$job, 'foreground' => 'false' );
        _bg_job_helper(
            'job'      => $job,
            'wait_msg' => 'Waiting for infinivol to finish delete job',
            'erro_msg' => 'Volume delete job fail',
            'warn_msg' => "Delete doesn't seem to be a background job",
        ) if $job;
    }
	}
    # Find all volumes
    my @volumes = NACL::C::Volume->find(
                    command_interface => $opts->{Node},
                    filter            => { vserver => $opts->{vserver} },
                    allow_empty       => 1
                  );
    $Log->comment('Stop vserver');
	
    $Vserver_obj  = NACL::C::Vserver->find(
                            command_interface => $opts->{Node},
                            filter         => {  vserver => $opts->{vserver}},
                            allow_empty       => 1,
                    );
        if (!$Vserver_obj) {
            $Log->comment('Unable to find specified vserver! need to create one');
            return $TCD::FAIL;
        }
    $Vserver_obj->stop()
        if ( $Vserver_obj->get_one_state_attribute('admin-state') ne 'stopped' );

    my $root_vol = undef;
    foreach my $volume (@volumes) {
        if ($Vserver_obj->get_one_state_attribute('rootvolume') ne
            $volume->get_one_state_attribute('volume') )
        {
            $Log->comment( "Unmount, offline and delete volume: "
                    . $volume->get_one_state_attribute('volume') );
            $volume->unmount()
                if ( $volume->get_one_state_attribute('junction-active')
                =~ m/true/i );
            $volume->offline()
                if (
                $volume->get_one_state_attribute('state') !~ m/offline/i );
            $volume->delete("method-timeout" => 20);
        } else {
            $root_vol = $volume;
        }
    }
    if (defined $root_vol) {
        $Log->comment("Deleting root volume: "
                . $root_vol->get_one_state_attribute('volume'));
        $root_vol->offline()
            if ($root_vol->get_one_state_attribute('state') !~ m/offline/i);
        $root_vol->delete("method-timeout" => 20);
    }
    $Log->comment("Deleting vserver: " . $opts->{vserver});
    {# Code to work around issue in burt 483819
        try {
            $Vserver_obj->delete("method-timeout" => 20);
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            $Vserver_obj->start();
            $Vserver_obj->delete("method-timeout" => 20);
            $Log->warn( "Apply work around for BURT 483819 to delete vserver "
                    . $opts->{vserver});
        }
    };
    return $TCD::PASS;
}

##
## @name  aggr_exists
##
## @description
##   Check if the aggregate exists 
##
## @arg $opts   Required: Reference to the hash that contains info as specified
##              in the synopsis
##
## @return
##   $TCD::PASS on Success, $TCD::FAIL on Failure
##
sub aggr_exists {

    my $opts = shift;

    $Log->comment("Check to see if aggregate named " . 
                    $opts->{aggregate} . " exists");
    try 
    {
        $Aggregate = 
            NACL::STask::Aggregate->find( 
                command_interface => $opts->{Node},
                'filter'          => { 'aggregate' => $opts->{aggregate}, },
                );
        $Log->comment('Using Aggregate ' . $opts->{aggregate});
        return $TCD::PASS;
    }
    catch NACL::Exceptions::NoElementsFound with {
        $Log->comment('Aggregate ' . $opts->{aggregate} . ' not found');
        return $TCD::FAIL;
    }
}

##
## @name aggr_create
##
## @description
##   Create an aggregate
##
## @arg $opts   Required: Reference to the hash that contains info as specified
##              in the synopsis
##
## @return
##   $TCD::PASS on Success, $TCD::FAIL on Failure
##
sub aggr_create {

    my $opts = shift;

    my $aggr_created;
    my $res = 0;

    try 
    {
        $Aggregate = 
            NACL::STask::Aggregate->find( 
                    command_interface => $opts->{Node},
                                filter=> {
                                    aggregate => $opts->{aggregate},
                                },
                );
        $Log->comment('Using Aggregate ' . $Aggregate->aggregate);
        $res = 1;
    }
    catch NACL::Exceptions::NoElementsFound with {
        # TODO: remove hard coding # of Disks
        try {
            $Log->comment('Create/fetch Aggregate');
            $Aggregate = NACL::STask::Aggregate->create(
                                aggregate => $opts->{aggregate},
                        command_interface => $opts->{Node},
                                if_exists => 'reuse',
                             _was_created => \$aggr_created,
                            nacltask_wait => 1,
                                diskcount => 3,
                                 );
            
            if ($aggr_created){
                $Log->comment('Aggregate ' . $Aggregate->aggregate .' created');
            } else {
                $Log->comment('Using Aggregate ' . $Aggregate->aggregate);
            }
            $opts->{is_aggr_created}   = $aggr_created;
            $res = 1;
        }
        catch NATE::BaseException with {
            my $eobj = shift;
            $Log->error("Failed while creating Aggregate Error: " .
                        $eobj->text());
            $res = 0;
        };
    }
    catch NATE::BaseException with {
        my $eobj = shift;
        $Log->error("Failed while creating Aggregate Error: " .
                        $eobj->text());
        $res = 0;
    };
    return $res;
}

## @name vol_exists
##
## @description
##   Check if the volume exists 
##
## @arg $opts   Required: Reference to the hash that contains info as specified
##              in the synopsis
##
## @return
##   $TCD::PASS on Success, $TCD::FAIL on Failure
##
sub vol_exists {

    my $opts = shift;
	my $repo2 = $opts->{is_repository};

    my @params = ( vserver => $opts->{Vserver});

    if ($main::VOLUME) {
        push @params, (volume  => $opts->{volume});
    }
    if ($repo2 eq 'true') {
        push @params, ('volume-style' => 'infinitevol');
    }
    # If volume specified or if it is a repository specified
    if (($main::VOLUME) || ($opts->{is_repository} eq 'true')) {
        try {
            $Volume = NACL::STask::Volume->find(
                    command_interface => $opts->{Node},
                    filter => { @params }
                  );
        }
        catch NATE::BaseException with {
            my $eobj = shift;
            $Log->comment("Volume $opts->{volume} not found: " .
                        $eobj->text());
                return $TCD::FAIL;
        };

        if ($Volume) {
            $Log->comment('Found volume, it can be reused : ' . 
                    $Volume->volume);
            # Store the new volume in hash
            $opts->{volume} = $Volume->volume;
            $opts->{Volume} = $Volume;

            # Check enable-snapdiff flag for IV
            check_enable_sd_flag($Volume) 
                if ($opts->{is_repository} eq 'true');
            return $TCD::PASS;
        }
    } else {
        return $TCD::FAIL;
    } 
}

## @name check__flag
##
## @description 
##   Check and enable snapdiff flag for infinite volume. 
##
## @arg $opts Required: Volume object 
##
## @return
##   $TCD::PASS on Success, $TCD::FAIL on Failure
##
sub check_enable_sd_flag {
    my ($Volume) = @_;

    try {
        # check and update enable-snapdiff flag to 'true' 
        if ($Volume->get_one_state_attribute('enable-snapdiff') ne 'true') {
            $Log->comment('updating enable-snapdiff flag to true');
            $Volume->modify('enable-snapdiff' => 'true');
        }
    } catch NATE::BaseException with {
        my $eobj = shift;
        $Log->comment("Volume " . $Volume->volume() . "not found: " .
                        $eobj->text());
        return $TCD::FAIL;
    };

    return $TCD::PASS;
}

## @name vol_create
##
## @description 
##   Create volume, vol type can be flex or infinite volume
##
## @arg $opts   Required: Reference to the hash that contains info as specified
##              in the synopsis
##
## @return
##   $TCD::PASS on Success, $TCD::FAIL on Failure
##
sub vol_create {

    my $opts = shift;

    # No aggr check needed for Infinite volume 
    if ($opts->{is_repository} eq 'false') {
        if (!aggr_exists($opts)) {
            $Log->comment('Aggregate does not exist, create it');
            aggr_create($opts);
        }
    }
    my $vol_size;
    if (!defined $opts->{volume_size}) {
        $vol_size = $main::VOL_SIZE;
    } else {
        $vol_size = $opts->{volume_size}; 
    }

   my @params;
    if ($opts->{is_repository} eq 'false') {
        @params = (aggregate   => $opts->{aggregate});
    }
  
    $Volume  = NACL::STask::Volume->create(
        command_interface   => $opts->{Node},
		volume              => $opts->{volume},
        'junction-path'     => $opts->{volume_path},
        vserver             => $opts->{Vserver},
        size                => $vol_size,
		nacltask_wait       => 1,
        'space-guarantee'   => 'none',
        'unix-permissions'  => '777',
        'security-style'    => 'mixed',
        'enable-snapdiff'   => $opts->{enable_snapdiff},
        @params,
    );
    # If we are in this function, it means that we created the volume
    $opts->{is_volume_created} = 1;

    # Call vol_exists to see if we created the volume successfully
    #my $rc = vol_exists($opts);

	if ($Volume) {
            # Store the new volume in hash
            $opts->{volume} = $Volume->volume;
            $opts->{Volume} = $Volume;
			if ( $opts->{mode} eq '7Mode' ) {
        # 7-mode set export flag anon=0
        my $apiset = $opts->{Node}->get_7m_or_nodescope_apiset();
        $apiset->exportfs(
                          persistent => 1,
                'persistent-options' => 'sec=sys,rw,anon=0',
                                path => $opts->{volume_path},
        );
    }
    # Save the new params in the hash
        $opts->{volume} = $Volume->volume;
        $opts->{Volume} = $Volume;
           
    }
    
    return $TCD::PASS;;
}

##
## @name create_snapshot
###
## @description 
##   Create snapshot and return return the snapshot object
##
## @arg $opts   Required: Reference to the hash that contains info as specified
##              in the synopsis
##
## @return
##   snapshot object
##

sub create_snapshot {

    my $opts = shift;

    $opts->{foreground} = 'true';

    if (!$opts->{snapshot}) {
        $opts->{snapshot} = $opts->{snapshot} || 
            'SD_'. $$ . int(rand(99999)) . '_snapshot';
    }

    my $si  = NACL::STask::VolumeSnapshot->create(
      command_interface     => $opts->{Node},
      volume                => $opts->{volume},
      vserver               => $opts->{vserver},
      snapshot              => $opts->{snapshot}, 
      nacltask_if_exists    => 'reuse',
      nacltask_wait         => 'true',
    );

    $Log->comment('Snapshot: '.$si->snapshot().' created');

    return $si;
}


##
## @name snapshots_setup
##
## @description
##   Setup routine to mount the test volume, create base snapshot,
##   populate data and then create the diff snapshot
##
## @arg $opts   Required: Reference to the hash that contains info as specified
##              in the synopsis
##
## @return
##   $TCD::PASS or $TCD::FAIL
##
sub snapshots_setup {

    my $set_h = shift;

    my $Dataset_info;
    my $diff_snapshot;
    my $base_snapshot;
    my $Mount;
    my $Client;

    my %snap_opts;
    $snap_opts{Node}    = $set_h->{Node};
    $snap_opts{vserver} = $set_h->{vserver} if ($Node->mode() ne '7Mode');
    $snap_opts{volume}  = $set_h->{volume};
    # Create snapshots for test
    if (($main::DIFF_SNAPSHOT && !$main::BASE_SNAPSHOT)) {
        # As diff snapshot specifed, without base snapshot
        # just do base line snapdiff
        $Log->comment('Only Diff snapshot specified');
            if (!$main::USE_DIFFSNAPSHOT_ONLY) {
                $Log->comment('Fetch diff snapshot');
                $snap_opts{snapshot} = $main::DIFF_SNAPSHOT;
                $diff_snapshot = create_snapshot(\%snap_opts);
            } else {
                return $TCD::FAIL; 
            }
        # TODO: If specified diff snapshot not present, return error
    }
    my $sn = 0;
    if ((!$main::DIFF_SNAPSHOT && !$main::BASE_SNAPSHOT)) {
        $Log->comment('Base and Diff Snapshots not specified');

    } else {
        $Log->comment('Both Base and Diff Snapshots are specified');
        $sn = 1;
    }
    # The sequence is 
    # 1) Mount volume
    # 2) Create base snapshot
    # 3) Populate dataset
    # 4) Create diff snapshot

    # Mount volume 
    $Log->comment('Mount volume on client');
    ($Client, $Mount) = SDutils::do_mount($set_h->{Volume});
    if (!$Mount) {
        $Log->comment('Failed to mount volume '. $set_h->{volume});
        return $TCD::FAIL;
    }

    # Create dataset
    $Log->comment('Create dataset');
    $Dataset_info = SDutils::create_dataset($Client, $Mount);

    # Create base snapshot
    if ($sn) { 
        # Both base and diff given
        $snap_opts{snapshot} = $main::BASE_SNAPSHOT;
    } else {
        $snap_opts{snapshot} = "";
    }
    $base_snapshot = create_snapshot(\%snap_opts);

    # Populate data
    $Log->comment('Populate data');
    my $rc = SDutils::modify_dataset($Client, $Mount, $Dataset_info);

    # Create diff snapshot
    if ($sn) { 
        # Both base and diff given
        $snap_opts{snapshot} = $main::DIFF_SNAPSHOT;
    } else {
        $snap_opts{snapshot} = "";
    }
    $diff_snapshot = create_snapshot(\%snap_opts);

    # Save all values in the hash
    $set_h->{Basesnapshot}  = $base_snapshot;
    $set_h->{basesnapshot}  = $base_snapshot->snapshot();
    $set_h->{Diffsnapshot}  = $diff_snapshot;
    $set_h->{diffsnapshot}  = $diff_snapshot->snapshot();
    $set_h->{Mount}         = $Mount;
    $set_h->{Client}        = $Client;

    return $TCD::PASS;
}


## @name snapdiff_zapi_setup
## @description
##   High level wrapper subroutine to create aggregates, volumes, snapshots, 
##   mount points and perform other setup procedures required for the SnapDiff 
##   tests
##
## @arg $opts   Required: Reference to the hash that contains info as specified
##              in the synopsis
##
## @return
##   $TCD::PASS on Success, $TCD::FAIL on Failure
##
sub snapdiff_zapi_setup {

    my $set_h = shift;

    my $aggr_name  = $main::AGGR_NAME;
    my $volume_name = $main::VOLUME || 'SD_' . int(rand(9999)) . '_volume';
    my $vol_path = "/$volume_name";
    my $username;
    my $password;
    my $base_snapshot;
    my $diff_snapshot;
    my $Zapi;
    my @nodes;

    $Node           = NACL::STask::Node->find();
    $set_h->{Node}  = $Node;
    @nodes          = NACL::STask::Node->find();
    $set_h->{Nodes} = \@nodes;
    $set_h->{mode}  = $Node->mode();

    $vol_path = "/$volume_name" if ($set_h->{mode} ne '7Mode');

    if (!$Node) {
        $Log->error("Unable to find Node");
        return $TCD::FAIL;
    }
	
    try {
        my @iplist = qw(0.0.0.0/0);

        my $SystemServicesFirewallPolicy
            = NACL::STask::SystemServicesFirewallPolicy->create(
            command_interface => $Node,
            policy            => 'data',
            action            => 'allow',
            service           => 'http',
            'ip-list'         => \@iplist
            );
    }
	catch NACL::C::Exceptions::SystemServicesFirewallPolicy::AlreadyExists with {
        $Log->comment('Firewall policy allow already exists');
		$allow_pol = 1;
    };
	if(!$allow_pol){
	    try{
	        my $SystemServicesFirewallPolicy = NACL::STask::SystemServicesFirewallPolicy->purge(
                                    command_interface => $Node,
                                    policy            => 'data',
                                    action            => 'deny',
                                    service           => 'http',
                                    );
		}
	    catch NACL::C::Exceptions::SystemServicesFirewallPolicy::DoesNotExist with{
		     $Log->comment('Firewall policy deny doesnt exists');
		   
        };
	}
	
    $set_h->{vserver}           = $main::VSERVER;
    $set_h->{aggregate}         = $main::AGGR_NAME;
    $set_h->{vserver_validate}  = 1;
    $set_h->{is_repository}     = ($main::REPOSITORY) ? 'true': 'false';
	my $repo = $set_h->{is_repository};
	
    $set_h->{enable_snapdiff}   = $main::ENABLE_SNAPDIFF;

    $Log->comment('Check if vserver exists');
    if (!vserver_exists($set_h)) {
        if ($set_h->{vserver_validate}) {
            $Log->comment('Vserver does not exist, creating vserver');
            if (!vserver_create($set_h)) {
                return $TCD::FAIL;
            } 
        } else {
            $Log->comment('Vserver/repository param validation failed!');
            return $TCD::FAIL;
        }
    }
    # Save vol info in the hash
    $set_h->{volume}        = $volume_name;
    $set_h->{volume_path}   = $vol_path;

    # Volume for the tests
    $Log->comment('Check if vol exists');
    if (!vol_exists($set_h)) {
        $Log->comment('Volume does not exist, creating volume for test');
        if (!vol_create($set_h)) {
            return $TCD::FAIL; 
        }
    }
    # Create snapshots for the tests
    if (!snapshots_setup($set_h)) {
        return $TCD::FAIL;
    }

    # TODO: remove hard-coding here 
    $username = $main::USERNAME || 
          (($Node->mode() eq '7Mode') ? 'root' : 'admin');
    $password = $main::PASSWORD || 'netapp1!';

    $set_h->{username} = $username;
    $set_h->{password} = $password;
    # TODO decide which ip based on mode & vol info
    my @param = (   filer => $set_h->{Node},
                  vserver => $set_h->{Vserver}, 
                 username => $username,
                 password => $password,);

    $Log->comment ("Param to snapdiff zapi @param");

    $Zapi = ZapiSD->new(@param);
    if (!$Zapi) {
        $Log->comment('Snapdiff Zapi object init failed');
        return $TCD::FAIL;
    }
    $set_h->{Zapi}      = $Zapi;

    # Print hash values
    if ($main::DEBUG) {
        $Log->comment('$$ Printing the contents of the hash $$');
        $Log->comment("Key is " . $_ . "\n" . "Value is " . "$set_h->{$_}")
        foreach (keys%$set_h) ;
    }
    return $TCD::PASS;
}


##
## @name snapdiff_zapi_cleanup
##
## @description
##   Cleanup routine to delete snapshots, mount points that were
##   created by/for the test
##
## @arg  %set_h  Required: Reference to the hash that contains: 
##               Mount: NACL Mount object
##               Volume: NACL Volume object
##               Basesnapshot: NACL object for base-snapshot
##               Diffsnapshot: NACL object for diff-snapshot
##               Aggregate: NACL aggregate object
##               is_aggr_created: to check if aggregate was created by test
##               is_volume_created: to check if volume was created by test
##               Look at the synopsis for full list of all params
##  
sub snapdiff_zapi_cleanup {

    my $set_h = shift;

    if ($main::DEBUG) {
        $Log->comment('$$ Printing the contents of the hash $$');
        $Log->comment("Key is " . $_ . "\n" . "Value is " . "$set_h->{$_}")
        foreach (keys%$set_h) ;
    }
	
	if (!$main::DIFF_SNAPSHOT) {
        foreach my $si ($set_h->{Basesnapshot}, $set_h->{Diffsnapshot}){
            next unless $si;
            $Log->comment('Removing snapshot: ' . $si->snapshot());
            $si->delete(foreground => 'true');
        }
    }
    # Cleanup
    if ($set_h->{Mount}) {
        $Log->comment("Unmount " . "$set_h->{Mount}->mount_point") ;
        $set_h->{Mount}->purge(force => 'true'); 
    }
    $Log->comment('Cleanup if aggregate, volume, snapshots were created');
    if ($set_h->{is_aggr_created} && $main::CLEANUP) {
        $set_h->{Aggregate}->purge();
        NACL::C::StorageDisk->zerospares(command_interface => $Node);
        $Log->comment('Deleting aggregate - it was created by test');
    }
    # If volume was created by the test
    # Delete Flex vol always and Infinivol only if CLEANUP=1
    if ($set_h->{is_volume_created} && ($main::CLEANUP || !$main::REPOSITORY)) {
        $set_h->{Volume}->purge(); 
        $Log->comment('Deleting volume - it was created by test');
    }
    # No clenaup if snapshots passed as argument
    
    return $TCD::PASS;
}

1;
