#! /usr/local/bin/perl
#
# $Id: //depot/prod/test/Rsierranevada/dp/snapmirror/Setup.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
## @summary dp::snapmirror::Setup - Simple scaffolding to suck in
##          parameters and automatically create the containers
##          required for snapmirror testing
##
## @author dl-dpgqa-automation@netapp.com
## @status Private
## @pod here

=head1 NAME

  dp::snapmirror::Setup - Simple scaffolding to suck in parameters
  and automatically create the containers required for snapmirror
  testing

=head1 SYNOPSIS

  use dp::snapmirror::Setup;

  EXAMPLE:

  use NOTE;
  use dp::Testscript qw(testscript);
  use dp::snapmirror::Setup;

  my $Setup = dp::snapmirror::Setup->new;
  my ($sm, $vol_src, $vol_dst, $qtree_src, $qtree_dst, $qtree);

  testscript(pretest  => { func => \&do_pretest  },
	     testcase => { func => \&do_testcase },
	     teardown => { func => \&do_teardown } );

  sub do_pretest {
      $Setup->setup;

      # NOTE::Volume objects
      $vol_src = $Setup->vol_src;
      $vol_dst = $Setup->vol_dst;

      # NOTE::Qtree objects
      $qtree_src = $Setup->qtree_src;
      $qtree_dst = $Setup->qtree_dst;
      
      my $sm = NOTE::SnapMirror->initialize(
          src=>$qtree_src,
	  dst=>$qtree_dst
      );
      
      # not an anonymouse NOTE::Qtree object
      $qtree = NOTE::Qtree:->create(
          Volume => $vol_src,
          name   => 'my_qtree'
      );
      ...
  }

  sub do_testcase {
      $vol_src->offline;
      ...
      $sm->compare if ($Setup->do_sm_compare);
      $vol_dst->wafliron_start if ($Setup->do_wafliron);
  }

  sub do_teardown {
      if ($Setup->do_cleanup) {
	  # remove resources created by this testscript
          $qtree->delete if ($qtree);
      }

      # call cleanup() method *OUTSIDE* of the if statement above
      $Setup->cleanup;
  }

=head1 DESCRIPTION

  SnapMirror specified scaffolding that automatically suck in all the
  predefined NATE parameters. Volumes and qtrees will be automatically 
  created (based on the parameters passed) and testscripts will now be
  shorter since the Setup/Cleanup code is performed by the library.

  It is the responsibility of the testscript to check the type of volume
  created by this library based on the parameters passed in by the user. 
  Certain operations can be performed on a flexvol and not on a trad 
  volume, e.g. $Volume->size;. In such scenarios, the testscript has to
  have different code paths for flexvols or trad volumes.

=head1 ROUTINES

=cut

package dp::snapmirror::Setup;
use strict;
use Carp qw(croak);
use Data::Dumper;
use Tharn;
use boot qw(transit_to);
use config_base qw(host_is_alive);
use NOTE;

=head2 new

 Constructor

 my $Setup = dp::snapmirror::Setup->new;

=cut

sub new {
    my ($class, %args) = @_;
    my $self = {};

    $self->{_owner} = $$;

    bless $self, $class;
    return $self;
}

=head2 get_parameters

 Initialize and return the hash reference of parameters.

 my $params = $Setup->get_parameters();

=cut

sub get_parameters {
    my ($self) = @_;

    $self->_init_parameters;
    # has to be called *before* _init_src_parameters
    $self->_init_dst_parameters; 
    $self->_init_src_parameters;

    return $self->{params};
}

=head2 setup

 Method to suck in all predefined NATE parameters. It also
 creates volumes (and aggregates depending on parameters used)
 on both source and destination filers. Qtrees will also be
 created inside the volumes by default.

 $Setup->setup;

=cut

sub setup {
    my ($self) = @_;

    $self->_init_parameters;
    $self->_init_setup_cleanup;

    # has to be called *before* _init_src_parameters
    $self->_init_dst_parameters; 
    $self->_init_src_parameters;

    $self->_init_2_filers;
}

=head2 cleanup

 Wrapper method that will invoke NOTE::teardown() method. This
 method will check CLEANUP parameter to determine if a cleanup
 is requested by the user.

 sub testscript_cleanup {
     ....
     if ($Setup->do_cleanup) {
	 # remove resources created by this testscript
     }

     $Setup->cleanup;
 }

=cut

sub cleanup {
    my ($self) = @_;

    # no cleanup for stray object(s) from a fork()
    return if ($self->_owner != $$);

    # it's do_cleanup() and *NOT* cleanup() which results in infinite
    # recursion. once bitten, twice shy...
    if ($self->do_cleanup) {
	logcomment("CLEANUP performing now");
	NOTE::teardown();

	if (!$self->do_setup) {
	    # XXX/tapl for now, just handle 2 filers until the need arises
	    foreach my $node qw(vol_dst vol_src) {
		my $vol = $self->{$node};

		next if (!defined($vol));

		# online volumes in case they're offline
		eval { $vol->online };

		# revert the attached volumes if it were setup by user
		# since these are attached volumes, need to do the cleanup
		# manually
		if ($vol->snapshot_exists($self->_snapshot)) {
		    logcomment("CLEANUP: Reverting volume ".
			       $vol->name . " " . $vol->tag);
		    $vol->snapshot_restore($self->_snapshot);
		    $vol->snapshot_delete($self->_snapshot);
		}
	    }
	}

    } else {
	# do nothing
	logcomment("Not performing CLEANUP");
    }
}


=head2 do_setup

 Method that returns the value of SETUP. It's mainly used in testscripts
 to determine if a certain setup portion inside a testscript can be
 controlled by the SETUP parameter.

 sub testscript_setup {
     my $qtree;

     # user can pre-create these resources if he/she does not want the
     # script to do so
     if ($Setup->do_setup) {
	 # anonymous NOTE::Qtree object
	 $qtree = NOTE::Qtree->create(
	      Volume => $Setup->vol_src,
	      name   => 'my_qtree';
         );
     }
     ...
 }

=head2 do_cleanup

 Method that returns the value of CLEANUP. It's mainly used in testscripts
 to determine if a certain setup portion inside a testscript can be
 controlled by the SETUP parameter.

 sub testscript_cleanup {
     ....
     if ($Setup->do_cleanup) {
	 # remove resources created by this testscript
	 $qtree->delete;
     }

     $Setup->cleanup;
 }

=head2 do_wafliron

 Method to determine if walfiron should be run on the aggregate/trad volume 
 if specified by the user. Ths method does *not* start wafliron.

 $Volume->wafliron_start if ($Setup->do_wafliron);

=head2 do_sm_compare

 Method to determine if snapmirror checker/treecompare should be run on the 
 snapmirror relationship if specified by the user. Ths method does *not* start 
 snapmirror checker/treecompare.

 # explicitly run snapmirror checker if specified by user
 $SnapMirror->check if ($Setup->do_sm_compare);

 # explicitly run treecompare if specified by user
 $SnapMirror->treecompare if ($Setup->do_sm_compare);

 # explicitly run snapmirror checker or treecompare (depending on src and dst
 # filers' version) if specified by user
 $SnapMirror->compare if ($Setup->do_sm_compare);

=head2 params

 Returns a hash containing all the NATE parameter that were sucked in.

=head2 vol_src

 Returns the source NOTE::Volume object.

=head2 vol_dst

 Returns the destination NOTE::Volume object.

=head2 qtree_src

 Returns the source NOTE::Qtree object.

=head2 qtree_dst

 Returns the destination NOTE::Qtree object.

=cut

sub do_setup      { return shift->{params}{SETUP};      }
sub do_cleanup    { return shift->{params}{CLEANUP};    }
sub do_wafliron   { return shift->{params}{WAFLIRON};   }
sub do_sm_compare { return shift->{params}{SM_COMPARE}; }
sub do_sv_compare { return shift->{params}{SV_COMPARE}; }
sub params        { return shift->{params};    }
sub vol_src       { return shift->{vol_src};   }
sub vol_dst       { return shift->{vol_dst};   }
sub qtree_src     { return shift->{qtree_src}; }
sub qtree_dst     { return shift->{qtree_dst}; }


#######################################################################
#
# Private Routines
#
#######################################################################


sub _snapshot { return shift->{_snapshot}; }
sub _owner    { return shift->{_owner};    }

sub _init_setup_cleanup {
    my ($self) = @_;
    $self->{_snapshot} = "__setup_cleanup_" . time . "__";
}

sub _init_parameters {
    my ($self) = @_;

    # Global private parameters. **NOT** to be used in testscripts
    # These can be overriden if we need to tune the aggregate or volume
    # parameters for testscripts to run on either simulators or hardware
    # filers.
    param("AGGR_SIZE",                -default, undef);
    param("CIFS_DOMAIN",              -default, "lab.netapp.com");
    param("CIFS_DOMAIN_ADMIN",        -default, "labclient");
    param("CIFS_DOMAIN_ADMIN_PASSWD", -default, "sundance");
    param("CIFS_PASSWD",              -default, "sundance");
    param("CIFS_USER",                -default, "Administrator");
    param("CLEANUP",                  -default, 1);
    param("CMDTIMEOUT",               -default, 120); # 2 mins
    param("FILER_CONN",               -default, "#rsh-socket");
    param("FLEX_SIZE",                -default, undef);
    param("SETUP",                    -default, 1);
    param("SM_COMPARE",               -default, 1);
    param("SV_COMPARE",               -default, 1);
    param("SM_TYPE",                  -default, "all");
    param("TIMEOUT",                  -default, 600); # 10 mins
    param("TRAD_SIZE",                -default, undef);
    param("TRAD_ZEROED",              -default, 0);
    param("VOL_TYPE",                 -default, "flex");
    param("WAFLIRON",                 -default, 1);

    # presize the hash buckets
    $self->{params} = {};
    keys(%{$self->{params}}) = 64;

    # XXX/tapl if anyone has a better idea on how to do this, let me
    # know
    $self->{params}{AGGR_SIZE}                = $::AGGR_SIZE;
    $self->{params}{CIFS_DOMAIN}              = $::CIFS_DOMAIN;
    $self->{params}{CIFS_DOMAIN_ADMIN}        = $::CIFS_DOMAIN_ADMIN;
    $self->{params}{CIFS_DOMAIN_ADMIN_PASSWD} = $::CIFS_DOMAIN_ADMIN_PASSWD;
    $self->{params}{CIFS_PASSWD}              = $::CIFS_PASSWD;
    $self->{params}{CIFS_USER}                = $::CIFS_USER;
    $self->{params}{CLEANUP}                  = $::CLEANUP;
    $self->{params}{CMDTIMEOUT}               = $::CMDTIMEOUT;
    $self->{params}{FILER_CONN}               = $::FILER_CONN;
    $self->{params}{FLEX_SIZE}                = $::FLEX_SIZE;
    $self->{params}{SETUP}                    = $::SETUP;
    $self->{params}{SM_COMPARE}               = $::SM_COMPARE;
    $self->{params}{SV_COMPARE}               = $::SV_COMPARE;
    push(@{$self->{params}{SM_TYPE}}, $self->_get_sm_types($::SM_TYPE));
    $self->{params}{TIMEOUT}                  = $::TIMEOUT;
    $self->{params}{TRAD_SIZE}                = $::TRAD_SIZE;
    $self->{params}{TRAD_ZEROED}              = $::TRAD_ZEROED;
    $self->{params}{VOL_TYPE}                 = $::VOL_TYPE;
    $self->{params}{WAFLIRON}                 = $::WAFLIRON;

    # Set the parameters for FILER

    param("FILER",          -default, undef);
    param("FILERA",          -default, undef);
    param("FILERB",          -default, undef);
    param("FILER_AGGR",     -default, undef);
    param("FILER_VOL",      -default, undef);
    param("FILER_VOL_SIZE", -default, undef);
    param("FILER_VOL_TYPE", -default, $self->{params}{VOL_TYPE});
    param("FILER_QTREE",    -default, undef);

    $self->{params}{FILER}          = $::FILER;
    $self->{params}{FILERA}          = $::FILERA;
    $self->{params}{FILERB}          = $::FILERB;
    $self->{params}{FILER_AGGR}     = $::FILER_AGGR;
    $self->{params}{FILER_VOL}      = $::FILER_VOL;
    $self->{params}{FILER_VOL_SIZE} = $::FILER_VOL_SIZE;
    $self->{params}{FILER_VOL_TYPE} = $::FILER_VOL_TYPE;
    $self->{params}{FILER_QTREE}    = $::FILER_QTREE;

    # NOTE: TO AVOID THE FOLLOWING PERL BUG IN MEMORY ALLOCATION...
    #
    #   Out of memory during request for XXX bytes, total sbrk() is YYYY bytes!
    #
    # VERIFY THAT THE KEY TO THE HASH IS DEFINED FIRST BEFORE ACCESSING THE
    # VALUE OF THE KEY.

    # determine size of volume (and aggregate) on destination filer
    if ( defined($self->{params}{FILER_VOL_TYPE}) &&
	 $self->{params}{FILER_VOL_TYPE} eq 'flex' ) {

	if ( defined($self->{params}{FLEX_SIZE}) ) {
	    param("FILER_VOL_SIZE",  -default, $self->{params}{FLEX_SIZE});
	    $self->{params}{FILER_VOL_SIZE}  = $::FILER_VOL_SIZE;
	}

	if ( defined($self->{params}{AGGR_SIZE}) ) {
	    param("FILER_AGGR_SIZE", -default, $self->{params}{AGGR_SIZE});
	    $self->{params}{FILER_AGGR_SIZE} = $::FILER_AGGR_SIZE;
	}

	if ( defined($self->{params}{FILER_VOL_SIZE} ) &&
	    $self->{params}{FILER_VOL_SIZE} !~ /^\d+[k|m|g|t]$/) {
	    croak "Filer flexvol size must be <size>[k|m|g|t]";
	}

    } elsif ( defined($self->{params}{FILER_VOL_TYPE} ) &&
	     $self->{params}{FILER_VOL_TYPE} eq 'trad') {

	if ( defined($self->{params}{TRAD_SIZE}) ) {
	    param("FILER_VOL_SIZE",  -default, $self->{params}{TRAD_SIZE});
	    $self->{params}{FILER_VOL_SIZE} = $::FILER_VOL_SIZE;
	}

	if ( defined($self->{params}{FILER_VOL_SIZE} ) &&
	    $self->{params}{FILER_VOL_SIZE} !~ /^\d+$/) {
	    croak "Filer trad vol size must be in numeric (no. of disks)";
	}

    } else {
	croak "Unknown FILER_VOL_TYPE" . 
	  defined( $self->{params}{FILER_VOL_TYPE} ) 
	    ? ": " . $self->{params}{FILER_VOL_TYPE} : 0;
    }
}

sub _init_dst_parameters {
    my ($self) = @_;
    my $params = $self->params;

    # exported testscript parameters


    
    if (!defined $params->{FILER_DST}) {
        if (!defined $params->{FILER} && !defined $params->{FILERB}) {
           param("FILER_DST", -mesg,    'FILER or FILERB or FILER_DST is required');
        } elsif (defined $params->{FILER}) {
           param("FILER_DST", -default, $params->{FILER});
        } else {
	   param("FILER_DST", -default, $params->{FILERB});
	}
    }
    param("FILER_DST_CONN",     -default, $params->{FILER_CONN});
    param("FILER_DST_AGGR",     -default, $params->{FILER_AGGR});
    param("FILER_DST_VOL",      -default, $params->{FILER_VOL});
    param("FILER_DST_VOL_SIZE", -default, $params->{FILER_VOL_SIZE});
    param("FILER_DST_VOL_TYPE", -default, $params->{VOL_TYPE});
    param("FILER_DST_QTREE",    -default, $params->{FILER_QTREE});

    $params->{FILER_DST}          = $::FILER_DST;
    $params->{FILER_DST_CONN}     = $::FILER_DST_CONN;
    $params->{FILER_DST_AGGR}     = $::FILER_DST_AGGR;
    $params->{FILER_DST_VOL}      = $::FILER_DST_VOL;
    $params->{FILER_DST_VOL_SIZE} = $::FILER_DST_VOL_SIZE;
    $params->{FILER_DST_VOL_TYPE} = $::FILER_DST_VOL_TYPE;
    $params->{FILER_DST_QTREE}    = $::FILER_DST_QTREE;

    # NOTE: TO AVOID THE FOLLOWING PERL BUG IN MEMORY ALLOCATION...
    #
    #   Out of memory during request for XXX bytes, total sbrk() is YYYY bytes!
    #
    # VERIFY THAT THE KEY TO THE HASH IS DEFINED FIRST BEFORE ACCESSING THE
    # VALUE OF THE KEY.

    # determine size of volume (and aggregate) on destination filer
    if ( defined($params->{FILER_DST_VOL_TYPE}) &&
	 $params->{FILER_DST_VOL_TYPE} eq 'flex' ) {

	if ( defined($params->{FLEX_SIZE}) ) {
	    param("FILER_DST_VOL_SIZE",  -default, $params->{FLEX_SIZE});
	    $params->{FILER_DST_VOL_SIZE}  = $::FILER_DST_VOL_SIZE;
	}

	if ( defined($params->{AGGR_SIZE}) ) {
	    param("FILER_DST_AGGR_SIZE", -default, $params->{AGGR_SIZE});
	    $params->{FILER_DST_AGGR_SIZE} = $::FILER_DST_AGGR_SIZE;
	}

	if ( defined($params->{FILER_DST_VOL_SIZE} ) &&
	    $params->{FILER_DST_VOL_SIZE} !~ /^\d+[k|m|g|t]$/) {
	    croak "Dst flexvol size must be <size>[k|m|g|t]";
	}

    } elsif ( defined($params->{FILER_DST_VOL_TYPE} ) &&
	     $params->{FILER_DST_VOL_TYPE} eq 'trad') {

	if ( defined($params->{TRAD_SIZE}) ) {
	    param("FILER_DST_VOL_SIZE",  -default, $params->{TRAD_SIZE});
	    $params->{FILER_DST_VOL_SIZE} = $::FILER_DST_VOL_SIZE;
	}

	if ( defined($params->{FILER_DST_VOL_SIZE} ) &&
	    $params->{FILER_DST_VOL_SIZE} !~ /^\d+$/) {
	    croak "Dst trad vol size must be in numeric (no. of disks)";
	}

    } else {
	croak "Unknown FILER_DST_VOL_TYPE" . 
	  defined( $params->{FILER_DST_VOL_TYPE} ) 
	    ? ": " . $params->{FILER_DST_VOL_TYPE} : 0;
    }
}

# XXX/tapl if anyone has a better idea on how to refactor this function and
# the above (_init_dst_parameters()) into one function, pls let me know.
# tried a foreach loop with '_DST_' and '_SRC_' keys in each iteration and
# it doesn't work when you try to populate a hash key with the parameter, e.g.
#
#  $params->{"FILER_" . $key . "_VOL"} = $::"FILER_" . $key . "_VOL";
#
sub _init_src_parameters {
    my ($self) = @_;
    my $params = $self->params;

    # sanity check. this function must always be called *after*
    # _init_dst_parameters()
    if (! exists($params->{FILER_DST}) ) {
	croak "_init_dst_parameters() must be called before calling " .
	  "_init_src_parameters";
    }

    # exported testscript parameters
    if (!defined ($params->{FILER_SRC}) ) {
	    if (defined ($params->{FILERA}) ) {
                  param("FILER_SRC", -default, $params->{FILERA});
	    }
	    else {
		  param("FILER_SRC", -default, $params->{FILER_DST});
	    }
    }
    param("FILER_SRC_CONN",     -default, $params->{FILER_CONN});
    param("FILER_SRC_AGGR",     -default, undef);
    param("FILER_SRC_VOL",      -default, undef);
    param("FILER_SRC_VOL_SIZE", -default, undef);
    param("FILER_SRC_VOL_TYPE", -default, $params->{VOL_TYPE});
    param("FILER_SRC_QTREE",    -default, undef);

    $params->{FILER_SRC}          = $::FILER_SRC;
    $params->{FILER_SRC_CONN}     = $::FILER_SRC_CONN;
    $params->{FILER_SRC_AGGR}     = $::FILER_SRC_AGGR;
    $params->{FILER_SRC_VOL}      = $::FILER_SRC_VOL;
    $params->{FILER_SRC_VOL_SIZE} = $::FILER_SRC_VOL_SIZE;
    $params->{FILER_SRC_VOL_TYPE} = $::FILER_SRC_VOL_TYPE;
    $params->{FILER_SRC_QTREE}    = $::FILER_SRC_QTREE;

    # NOTE: TO AVOID THE FOLLOWING PERL BUG IN MEMORY ALLOCATION...
    #
    #   Out of memory during request for XXX bytes, total sbrk() is YYYY bytes!
    #
    # VERIFY THAT THE KEY TO THE HASH IS DEFINED FIRST BEFORE ACCESSING THE
    # VALUE OF THE KEY.

    # determine size of volume (and aggregate) on destination filer
    if ( defined($params->{FILER_SRC_VOL_TYPE}) &&
	 $params->{FILER_SRC_VOL_TYPE} eq 'flex' ) {

	if ( defined($params->{FLEX_SIZE}) ) {
	    param("FILER_SRC_VOL_SIZE",  -default, $params->{FLEX_SIZE});
	    $params->{FILER_SRC_VOL_SIZE}  = $::FILER_SRC_VOL_SIZE;
	}

	if ( defined($params->{AGGR_SIZE}) ) {
	    param("FILER_SRC_AGGR_SIZE", -default, $params->{AGGR_SIZE});
	    $params->{FILER_SRC_AGGR_SIZE} = $::FILER_SRC_AGGR_SIZE;
	}

	if ( defined($params->{FILER_SRC_VOL_SIZE} ) &&
	    $params->{FILER_SRC_VOL_SIZE} !~ /^\d+[k|m|g|t]$/) {
	    croak "Dst flexvol size must be <size>[k|m|g|t]";
	}

    } elsif ( defined($params->{FILER_SRC_VOL_TYPE} ) &&
	     $params->{FILER_SRC_VOL_TYPE} eq 'trad') {

	if ( defined($params->{TRAD_SIZE}) ) {
	    param("FILER_SRC_VOL_SIZE",  -default, $params->{TRAD_SIZE});
	    $params->{FILER_SRC_VOL_SIZE} = $::FILER_SRC_VOL_SIZE;
	}

	if ( defined($params->{FILER_SRC_VOL_SIZE} ) &&
	    $params->{FILER_SRC_VOL_SIZE} !~ /^\d+$/) {
	    croak "Dst trad vol size must be in numeric (no. of disks)";
	}

    } else {
	croak "Unknown FILER_SRC_VOL_TYPE" . 
	  defined( $params->{FILER_SRC_VOL_TYPE} ) 
	    ? ": " . $params->{FILER_SRC_VOL_TYPE} : 0;
    }
}

sub _init_2_filers {
    my ($self) = @_;
    my $params = $self->params;

    # loop through for all parameters with FILER_DST* and FILER_SRC*
    # prefixes
    foreach my $node ( qw(dst src) ) {
	my $key = "FILER_" . uc($node);
	my $vol   = undef;
	my $qtree = undef;
	
	# user did not indicate that they did the setup
	if ($self->do_setup) {

	    # create flex vol and aggr
	    if ($params->{$key.'_VOL_TYPE'} eq 'flex') {
		$vol = NOTE::Volume->create(
		    Filer     => { host  => $params->{$key} },
		    Aggregate => { name  => $params->{$key.'_AGGR'}, 
				   disks => $params->{$key.'_AGGR_SIZE'},
				   recycle => 1,
				 },
		    name      => $params->{$key.'_VOL'},
		    type      => $params->{$key.'_VOL_TYPE'},
		    size      => $params->{$key.'_VOL_SIZE'},
		    tag       => $node . ' vol',
	        );

	    } else {
		$vol = NOTE::Volume->create(
		    Filer     => { host => $params->{$key} },
		    name      => $params->{$key.'_VOL'},
		    type      => $params->{$key.'_VOL_TYPE'},
		    disks     => $params->{$key.'_VOL_SIZE'},
		    zeroed    => $params->{TRAD_ZEROED},
		    recycle   => 1,
		    tag       => $node . ' vol',
		);

		# check and make sure dst vol is as large or larger in
		# capacity than src vol
		if ($node eq 'src') {
		    while (exists($self->{'vol_dst'}) &&
			   $self->{'vol_dst'}->total < $vol->total) {
			$self->{'vol_dst'}->add;
		    }
		}
	    }

	    $qtree = NOTE::Qtree->create(
		 Volume => $vol,
		 name   => $params->{$key.'_QTREE'},
		 tag    => $node . ' qtree',
	    );

	} else {
	    # setup has been done by user

	    if (defined($params->{$key.'_VOL'})) {
		$vol = NOTE::Volume->attach(
	            Filer     => { host => $params->{$key} },
		    name      => $params->{$key.'_VOL'},
		    tag       => $node . ' vol',
	        );
	    } else {
		croak $key . "_VOL must be specified if SETUP=0";
	    }
	    
	    # if user specified qtree also, attach that as well
	    if (defined($params->{$key.'_QTREE'})) {
		$qtree = NOTE::Qtree->attach(
		    Volume => $vol,
		    name   => $params->{$key.'_QTREE'},
		    tag    => $node . ' qtree',
		);
	    }

	    eval { $vol->online };

	    # create a snapshot to 'protect' this volume
	    $vol->snapshot_create($self->_snapshot);
	}

	# store these objects into $self
	$self->{'vol_' . $node}  = $vol;
	$self->{'qtree_'. $node} = $qtree;

	my $filer = $vol->Filer;
	    
	$filer->conn()->timeout($params->{CMDTIMEOUT});

	# if simulator is down, bring it up
	if ($filer->hostrec()->hosttype() eq 'filer-simulator' &&
	    !host_is_alive($filer->hostrec()->default_ip(), 10)) {
	    transit_to($filer->hostrec(), 'UP') 
	      or croak("transit_to(): $@");
	}
    }
}

sub _get_sm_types {
    my ($self, $types) = @_;
    my @types;

    my @split = split(/,| /, $types);
    foreach my $type (@split) {
	if ($type =~ /^vsm$/i) {
	    push(@types, lc($type));

	} elsif ($type =~ /^qsm$/i) {
	    push(@types, lc($type));

	} elsif ($type =~ /^rsm$/i) {
	    push(@types, lc($type));

	} elsif ($type =~ /^all$/i) {
	    @types = ();
	    push(@types, 'qsm', 'vsm', 'rsm');
	    last;

	} else {
	    croak "Unrecognized SnapMirror type: $type";
	}
    }

    return @types;
}

1;
