# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/dp/snapmirror/SetupCifs.pm#2 $
#
## @summary SetupCifs - CIFS setup class
## @author	corbin@netapp.com, dl-dpgqa-automation@netapp.com
## @status	private
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

SetupCifs - CIFS setup class

=head1 SYNOPSIS

  use dp::snapmirror::SetupCifs;

  $cifs = dp::snapmirror::SetupCifs->new(
      filer  => $filer,
      client => $client,
      params => $params,
  );
  $out = $cifs->is_minwin_client();
  $out = dp::snapmirror::SetupCifs::is_minwin_client(client => $client);
  $drive = $cifs->create_and_map_share(qtree => $qtree_obj);
  $drive = $cifs->drive();
  $share = $cifs->share();
  $cifs->teardown();

=head1 DESCRIPTION

Provides an object interface for the CIFS setup and cleanup.

=head1 STATUS

  Private

=head1 AUTHOR

  corbin@netapp.com, dl-dpgqa-automation@netapp.com
  $Id: //depot/prod/test/Rsierranevada/dp/snapmirror/SetupCifs.pm#2 $
  Last revised $Date: 2013/04/24 $

=head1 COPYRIGHT

  Copyright 2006, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package dp::snapmirror::SetupCifs;

use strict;
use warnings;

use Carp qw(croak);
use CIFS::Setup qw(cifs_setup);
use CLI::FileIO qw(rdfile wrfile);
use ExportFS qw(share_add share_delete);
use TestEnv::MinWin qw(minwin_compatible);
use Tharn qw(host connect logresult logcomment);
use MapShare qw(map_cifs_share unmap_cifs_share);
use Params::Validate qw(validate SCALAR OBJECT HASHREF);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub new($%);
sub is_minwin_client($%);
sub create_and_map_share($%);
sub teardown($;);
sub drive;
sub share;

#---------------------------------------------------------------------
# Package variables
#---------------------------------------------------------------------
my $usermap = '/etc/usermap.cfg';
my %minwin_client;

#---------------------------------------------------------------------

=head2 new

  $cifs = dp::snapmirror::SetupCifs->new(filer  => $filer,
                                         client => $client,
                                         params => \%params);

CIFS setup object constructor. It checks if the client is minwin
compatible. CIFS is always setup on the filer regardless if it is
already running.

The following named arguments are accepted:

=over

=item Arguments

=over

=item filer

(Required) : Hostspec of filer.

=item client

(Required) : Hostspec of Windows client.

=item params

(Required) : Reference to a hash of global parameters as returned by
the get_parameters method of the dp::snapmirror::Setup object.
The parameters of interest are:

  {
      CIFS_USER
      CIFS_PASSWD
      CIFS_DOMAIN
      CIFS_DOMAIN_ADMIN
      CIFS_DOMAIN_ADMIN_PASSWD
      FILER_CONN
      TIMEOUT
      CMDTIMEOUT
  }

=back

=item Returns

SetupCifs object.

=item Example

=back

     # Instantiate the Setup object to get the global parameters
     $setup = dp::snapmirror::Setup->new();
     $params = $setup->get_parameters();

     $cifs = dp::snapmirror::SetupCifs->new(
         filer  => $filer,
         client => $client,
         params => $params,
     );

=for html <hr />

=cut

#---------------------------------------------------------------------
sub new($%) {
    my $class = shift;
    my (%opts) = validate (
        @_,
        {
            filer  => { type => SCALAR },
            params => { type => HASHREF },
            client => { type => SCALAR },
            setup  => { type => SCALAR, default => 1 },
        }
    );
    my ($fh, $fc, $text, $qtext, $domain_user);
    my $params = $opts{params};

    if ( ! $minwin_client{$opts{client}} ) {
        # Check if client is minwin compatible
        is_minwin_client(client => $opts{client});
        $minwin_client{$opts{client}} = 1;
    }

    if ( $opts{setup} ) {
        logcomment('Setup CIFS on filer');
        $domain_user = $params->{CIFS_DOMAIN}.'\\'.$params->{CIFS_DOMAIN_ADMIN};
        cifs_setup(host         => $opts{filer},
                   cifs_pass    => $params->{CIFS_PASSWD},
                   domain       => $params->{CIFS_DOMAIN},
                   domain_user  => $domain_user,
                   domain_pass  => $params->{CIFS_DOMAIN_ADMIN_PASSWD},
                   passwd_group => 'yes');
    }

    # Add user mapping on filer
    $fh = host($opts{filer});
    $fc = connect($params->{FILER_CONN}, '-timeout', $params->{TIMEOUT});
    $fc->timeout($params->{CMDTIMEOUT});
    $opts{_filer_name} = uc($fh->hostname());
    $opts{_usermap_text} = rdfile($fc, $usermap) or croak $@;
    $text = "\n" . $opts{_filer_name} . "\\$params->{CIFS_USER} == *\n";
    $qtext = quotemeta($text);
    if ( $opts{_usermap_text} !~ /$qtext/s ) {
        $text = $opts{_usermap_text} . $text;
        wrfile($fc, $usermap, $text) or croak $@;
    } else {
        $opts{_usermap_text} = undef;
    }

    # Return object
    bless { %opts }, $class;
}

#---------------------------------------------------------------------

=head2 is_minwin_client

  $out = $cifs->is_minwin_client();
  $out = dp::snapmirror::SetupCifs::is_minwin_client(client => $client);

Checks if client is minwin compatible.
This method may be called as a class or instance method.
When called as an instance method it always returns 1
since the constructor already has checked the client.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client

(Optional) : Hostspec of Windows client.
This argument is required when called as a class method.
It is ignored when called as an instance method.

=back

=item Returns

1 if it succeeds.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub is_minwin_client($%) {
    my $self = shift if ((@_ % 2) == 1);
    my (%opts) = validate (
        @_,
        {
            client => { type => SCALAR, optional => 1 },
        }
    );

    # Always return 1 when called as an
    # instance method
    return 1 if defined($self);

    logcomment('Check if client is minwin compatible');

    # Check if client is a CIFS client
    my $fh = host($opts{client});
    if ( $fh->hosttype() !~ /windows/ ) {
        croak "Client $opts{client} is not a Windows client";
    }

    # Check if client is minwin compatible
    minwin_compatible($opts{client}) or croak $@;

    return 1;
}

#---------------------------------------------------------------------

=head2 create_and_map_share

  $drive = $cifs->create_and_map_share(qtree => $qtree_obj
                                       [, name => $sharename ]
                                       [, drive => $drive ]);

Creates a CIFS share for the qtree directory on the filer
and maps it on the Windows client.

The following named arguments are accepted:

=over

=item Arguments

=over

=item qtree

(Required) : NOTE::Qtree or NOTE::Volume object.

=item name

(Optional) : Name of share to use. It defaults to the qtree object's
tag or the name of the object if the tag does not exist.

=item drive

(Optional) : Drive letter to use on the client.
Defaults to '*' (let the client choose the drive letter).

=back

=item Returns

Drive letter where share is mapped.

=item Example

=back

     # Create a Qtree object
     $qtree_obj = NOTE::Qtree->create(Volume => $vol_obj);

     $drive = $cifs->create_and_map_share(qtree => $qtree_obj);

=for html <hr />

=cut

#---------------------------------------------------------------------
sub create_and_map_share($%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            qtree => { type => OBJECT },
            name  => { type => SCALAR, optional => 1 },
            drive => { type => SCALAR,  default => '*' },
        }
    );

    # Get Filer object from qtree object
    $self->{_Filer} = $opts{qtree}->Filer;
    # Get full path of share on filer
    $self->{_mount_path} = $opts{qtree}->mount_path;

    # Check that qtree is in the same filer as the one on setup
    my $fh = host($self->{filer});
    if ( $fh->hostname() ne $self->{_Filer}->host() ) {
        croak 'Qtree ' . $self->{_Filer}->host . ':' . $self->{_mount_path} .
              ' does not belong to filer ' . $fh->hostname();
    }

    # Get share name
    $self->{_share} = defined($opts{name}) ? $opts{name} :
                      defined($opts{qtree}->tag) ? $opts{qtree}->tag :
                      $opts{qtree}->name;

    # Create share on filer
    logcomment('Create share on filer');
    share_add($self->{_Filer}->conn, $self->{_share}, $self->{_mount_path});

    # Map share on client
    logcomment('Map share on client');
    $self->{_drive} = map_cifs_share(
      server   => $self->{_Filer}->conn->connhost,
      client   => $self->{client},
      drive    => $opts{drive},
      share    => $self->{_share},
      username => $self->{params}{CIFS_USER},
      password => $self->{params}{CIFS_PASSWD},
    ) or croak $@;

    return $self->{_drive};
}

#---------------------------------------------------------------------

=head2 teardown

  $cifs->teardown();

Cleans up both the filer and client. It unmaps the share on the
client and deletes the share on the filer. It only cleans up exactly
what this object setup in the first place.
It does not terminate CIFS on the filer to avoid disconnecting other
users.

This method does not croak, instead it logs a warning in order
to try to do all cleanup steps.

=over

=item Returns

Nothing.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub teardown($;) {
    my $self = shift;
    my $fc;
    my $params = $self->{params};

    if ( $self->{_drive} ) {
        # Unmap share on client
        logcomment('Unmap share on client');
        unmap_cifs_share(
            client => $self->{client},
            drive  => $self->{_drive},
        ) or logresult('WARN', $@);
        $self->{_drive} = undef;
    }

    eval {
        # Re-connect to filer
        host($self->{filer});
        $fc = connect($params->{FILER_CONN}, '-timeout', $params->{TIMEOUT});
        $fc->timeout($params->{CMDTIMEOUT});
    };
    if ( $@ ) {
        logresult('WARN', $@) if $@;
        # Return now since the following cleanup
        # steps require the filer connection
        return;
    }

    if ( exists($self->{_share}) ) {
        # Delete share on filer
        logcomment('Delete share on filer');
        eval { share_delete($fc, $self->{_share}) };
        logresult('WARN', $@) if $@;
        $self->{_share} = undef;
    }

    if ( defined($self->{_usermap_text}) ) {
        wrfile($fc, $usermap, $self->{_usermap_text}) or logresult('WARN', $@);
        $self->{_usermap_text} = undef;
    }

    return;
}

#---------------------------------------------------------------------

=head2 drive

  $cifs->drive();

Returns the drive letter where the share is mapped on the client.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub drive { shift->{_drive} }

#---------------------------------------------------------------------

=head2 share

  $cifs->share();

Returns the share name of the qtree.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub share { shift->{_share} }

1;

