# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/dp/snapmirror/Utils.pm#1 $
##
## @summary NOTE::Utils library
##	Utils methods for NOTES objects
##
## @pod here
##
## @author sydneyw@netapp.com, dl-dpgqa-automation@netapp.com
#

=head1 NAME

NOTE::Utils - Utility methods for NOTE objects

=head1 SYNOPSIS

CLI parameters process by init_param and made available by param_value routines are as follows:

I<FILER>          - Filer to use in test. Defaults to FILER_SRC.

I<FILER_SRC>      - Snapmirror source filer. Defaults to FILER.

I<FILER_SRC_AGGR> - Specify aggregate to use for flex source volume if set.

I<FILER_SRC_VOL>  - Specify source volume on FILER_SRC. Creates named volume if not found.

I<FILER_DST>      - Snapmirror destination filer. Defaults to FILER_SRC.

I<FILER_DST_AGGR> - Specify aggregate to use for flex destination volume if set.

I<FILER_DST_VOL>  - Specify destination volume on FILER_DST. Create named volume if not found.

I<FILER_CAS>      - Snapmirror cascading filer, 3rd filer in cascade relationships. Defaults to FILER_SRC.

I<FILER_CAS_AGGR> - Specify aggregate to use for flex cascade volume if set.

I<FILER_CAS_VOL>  - Specify destination volume on FILER_CAS. Create named volume if not found.

I<AGGR_SRC_BITS>  - Source aggregate indirect types, 32 or 64 if defined.

I<AGGR_DST_BITS>  - Destination aggregate indirect types, 32 or 64 if defined.

I<AGGR_CAS_BITS>  - Cascade aggregate indirect types, 32 or 64 if defined.

I<VOL_SIZE>       - Flex volume size. Default: RSM: 10g, VSM or QSM 5g.

I<DISK_CNT>       - Number of disks to use per aggregate / trad volume.

I<QT_SIZE>        - Qtree size to use in calculating Volume size when creating QSM volumes. Default 500m.

I<VOL_TYPE>       - Volume type, one of flex or trad. Default flex.

I<SM_TYPE>        - Snapmirror test, one of all, qsm, rsm or vsm. Default: if test type is common: all, otherwise set to test type.

I<DUMP_FILE>      - Dump file to use for tape device. Default /tapefile.

I<WAFLIRON>       - Determines if wafliron will be executed or not. If set to 0, then simulate a wafliron run rather then actually running wafliron. Default 1

I<WRECK_CMD>      - Specify comma separated list of wreck commands to use

B<use dp::snapmirror::Utils;>

init_params(test_type=>{'common'|'qsm'|'rsm'|'vsm'}, multi_tests=>{1|0}, def_vol_type=>{'flex'|'trad'});

my $lvalue = param_value($name, $new_value);

my @volumes = make_volumes(src_tag=>"src", dst_tag=>"dst", cas_tag=>"cas", mount=>1);

my $volume = make_volume(tag=>$tag, filer=>$hostspec, aggr=>$aggr_name, vol=>$vol_name);

vol_action(vol=>$vol_obj, action=>"fill|clear");

my $note_vol = fill_aggr($aggr);

sm_action(title=>$label, sm=>$note_sm_obj, skip_error=>$error, timeout=>$timeout_in_sec);

my @sm_list = make_mirror(count=>5, type=>"qsm", qt_per_src=>1, qt_per_dst=>10, dst_per_src=2);

my @proc_list = do_mirror($sm_action, \@sm_list, %sm_opts);

monitor_mirror($sm_label, @proc_list);

wait_mirror($filer, $expected_status, @sm_list);

my @status_list = stats_mirror($filer, @sm_list);

mount_object(object=>$vol_or_qtree, exportfs=>$exportfs, mode=>$mode);

sm_fiji(filer=>$filer, action=>$action, dst_path=>$path, state=>$state, fiji_action=>$fij_action);

check_DUMP_FILE(filer=>$filer);

my @lun_list = lun_type_list(filer=>$filer);

my @ics_list = get_ics_list();

my $hostpath = hostpath_from_icspath($ics_src);

clr_ics_src(@ics_list);

my $status = check_volumes(@volume_list);

=cut

package dp::snapmirror::Utils;

use strict;
use Carp;
use Tharn;

use Data::Dumper;
use NOTE;

use SnapMirror::Helper qw( sm_set_permissions );
use Filer::Version qw( get_wafl_version WV_ANCHORSTEAM_MIN WV_ANCHORSTEAM_MAX );
use SayWrap qw( sayEOF );
use CLI::FileIO qw( rdfile );

use Params::Validate qw( validate validation_options SCALAR BOOLEAN HASHREF );
use base qw( Exporter );

use vars qw(
	$FILER
	$FILER_SRC $FILER_SRC_AGGR $FILER_SRC_VOL
	$FILER_DST $FILER_DST_AGGR $FILER_DST_VOL
	$AGGR_SRC_BITS $AGGR_DST_BITS $AGGR_CAS_BITS
	$FILER_CAS $FILER_CAS_AGGR $FILER_CAS_VOL
	$DISK_CNT $VOL_SIZE $QT_SIZE $VOL_TYPE $SM_TYPE
	$DUMP_FILE $WAFLIRON $WRECK_CMD
	%pvalue
);

my $need_init = 0;
# Hash to keep track of the volumes created for populating aggregates
# used in aggr copy
my %aggr_pop_vols;

BEGIN {
    Params::Validate::validation_options(
	on_fail => sub { my @msgs = @_; chomp(@msgs);
			Carp::confess(Tharn::script_result(@msgs)) }
    );

    our @EXPORT_OK = qw(
	init_params
	param_value
	sm_fiji
	check_DUMP_FILE
	lun_type_list
	get_ics_list
	hostpath_from_icspath
	clr_ics_list
	make_volumes
	make_volume
	vol_action
	fill_aggr
	sm_action
	wreck_test
	make_mirror
	do_mirror
	monitor_mirror
	wait_mirror
	stats_mirror
	mount_object
	check_volumes
	get_pop_vol
        get_messages
	get_sm_id
    );
}

use constant PRIV_TEST_CMD => "priv set test; ";

=head2 init_params

Initializes and verify command line parameters according to test case script types.

=over

=item Arguments

B<test_type> => I<common|qsm|rsm|vsm>

Determines type of snapmirror test script. Default: I<common>

B<multi_tests> => I<1|0>

Determines whether the test script contains more then one test case. Default: 0

B<def_vol_type> => I<flex|trad>

Defaults for VOL_TYPE, flex or trad. This is only needed for multi tests scripts that
do not use VOL_TYPE to determine test cases.

=item Examples

=back

      init_params(test_type=>'vsm', def_vol_type=>'flax');

=cut

sub init_params {
    my (%opts) = validate(@_, {
			test_type => {
				type => SCALAR,
				optional => 1,
				default => "common",
				regex => qr/common|qsm|rsm|vsm|other/ },
			multi_tests => {
				type => BOOLEAN,
				optional => 1,
				default => 0 },
			def_vol_type => {
				type => SCALAR,
				optional => 1,
				regex => qr/flex|trad/ }
		});
		
    $need_init = 1;

    # Set default value
    $pvalue{VOL_TYPE} = $opts{def_vol_type} if $opts{def_vol_type};
    $pvalue{SM_TYPE} = ($opts{test_type} eq 'common' ? "all" : lc($opts{test_type}));
    $pvalue{WAFLIRON} = 1;

    my @param_spec = (
	[ 'VOL_TYPE', "-default" ],
	[ 'SM_TYPE', "-default" ],
        [ 'WAFLIRON', "-default" ],
	[ 'WRECK_CMD', "-default" ],
	[ 'DUMP_FILE' ], # Default: /tapefile

	[ 'DISK_CNT' ],
	[ 'VOL_SIZE' ],
        [ 'QT_SIZE' ],

	[ 'FILER' ],
	[ 'FILER_SRC' ],
	[ 'FILER_SRC_AGGR' ],
	[ 'FILER_SRC_VOL' ],

	[ 'FILER_DST' ],
	[ 'FILER_DST_AGGR' ],
	[ 'FILER_DST_VOL' ],

	[ 'FILER_CAS' ],
	[ 'FILER_CAS_AGGR' ],
	[ 'FILER_CAS_VOL' ],

	[ 'AGGR_SRC_BITS' ],
	[ 'AGGR_DST_BITS' ],
	[ 'AGGR_CAS_BITS' ],
    );
    foreach my $pspec (@param_spec) {
	my @ps = @$pspec;
	# take default from $pvalue
	$ps[2] = $pvalue{$ps[0]} if $ps[1] eq "-default";
	push @ps, "-default", undef unless @ps > 1;
	$pvalue{$ps[0]} = param(@ps);
    }
    croak "FILER or FILER_SRC must be set"
 	unless $pvalue{FILER} or $pvalue{FILER_SRC};
    $pvalue{FILER_SRC} = $pvalue{FILER} unless $pvalue{FILER_SRC};
    $pvalue{FILER} = $pvalue{FILER_SRC} unless $pvalue{FILER};

    $pvalue{FILER_DST} = $pvalue{FILER_SRC} unless $pvalue{FILER_DST};
    $pvalue{FILER_CAS} = $pvalue{FILER_SRC} unless $pvalue{FILER_CAS};

    $pvalue{DUMP_FILE} = "/tapefile" unless $pvalue{DUMP_FILE};

    unless ( $opts{multi_tests} ) {
	# The default test case volume type unless specified
	$pvalue{VOL_TYPE} = 'flex' unless $pvalue{VOL_TYPE};
    }

    unless ( $opts{test_type} =~ /other/i ) {
	if ( $opts{test_type} =~ /common/i ) {
		# Set the $SM_TYPE according to script type
		$pvalue{SM_TYPE} = lc($pvalue{SM_TYPE});
		$pvalue{SM_TYPE} = "rsm,vsm,qsm"
			if $pvalue{SM_TYPE} =~ /all/i;
	}

	$pvalue{VOL_SIZE} = ($pvalue{SM_TYPE} =~ /rsm/i ? "10g" : "5g")
		unless $pvalue{VOL_SIZE};
	$pvalue{QT_SIZE} = $pvalue{VOL_SIZE} unless $pvalue{QT_SIZE};

	croak "FILER_SRC and FILER_DST must be different for rsm test with flex vols"
		if ($opts{test_type} =~ /rsm/i || $pvalue{SM_TYPE} =~ /rsm/i)
		and $pvalue{VOL_TYPE} =~ /flex/i
		and $pvalue{FILER_SRC} eq $pvalue{FILER_DST};
    }
}

=head2 param_value

Sets or returns the current value of the command line parameter

=over

=item Arguments

B<name> = I<cli parameter>

Name of the command line parameter. e.g. FILER_SRC

B<new_value> = I<new value>

Sets command line parameter to a new value

=item Examples

=back

      $value = param_value('FILER_SRC'); # Retrieved CLI parameter value

      param_value('VOL_TYPE', 'trad');   # Change CLI parameter value

=cut

sub param_value {
    my ($name, $new_value) = @_;

    $pvalue{$name} = $new_value
	if defined $new_value;

    return $pvalue{$name};
}


=head2 make_volumes

Calls make_volume to create the requested volumes and mounts them if specified.

Returns array of
L<C<NOTE::Volume>|lib-NOTE-Volume-pm>
objects tagged in the following order, where applicable: src, dst, cas

=over

=item Arguments

B<src_tag> => I<src vol tag>

I<Required>
Create src volume and tag it with src_tag

B<dst_tag> => I<dst vol tag>

I<Optional>
Create dst volume and tag it with dst_tag.

B<cas_tag> => I<cas vol tag>

I<Optional>
Create cas(cade) volume and tag it with cas_tag.

B<vol_size> => I<vol_size>

I<Optional>
Create volume(s) of this size, acts as an override to the default size set by VOL_SIZE

B<mount> = I<1|0>

I<Optional>
Export and mount the created volumes. exportfs = "rw,nosuid,anon=0", mount mod = 777,
Default: I<0>

=back

=cut

sub make_volumes {
    my (%opts) = validate (@_, {
			src_tag => { type => SCALAR, optional => 0 },
			dst_tag => { type => SCALAR, optional => 1 },
			cas_tag => { type => SCALAR, optional => 1 },
			vol_size => { type => SCALAR, optional => 1 },
			mount => { type => BOOLEAN, optional => 1, default => 0 }
		});

    croak "init_params needs to be call first"
	unless $need_init;

    my ($src_size, @vols);
    foreach my $type ('src', 'dst', 'cas') {
	if ( $opts{$type."_tag"} ) {
		my $vol = make_volume(_vol_opts($type, $opts{vol_size}));
		mount_object(object=>$vol) if $opts{mount};
		push @vols, $vol;
		if ( $pvalue{VOL_TYPE} =~ /trad/i ) {
			# Guarantee that trad src is smaller or equal to the others
			if ( $type =~ /src/ ) {
				$src_size = $vol->size;
			} else {
				if ( $src_size ) {
					while ( $src_size > $vol->size ) {
						NOTE::logcomment("Add disk to "
							."$type volume (size ".$vol->size.") "
							."is smaller then src (size $src_size)");
						$vol->add;
					}
				}
			}
		}
	}
    }

    return @vols;
}


sub _vol_opts {
    my ($tag, $vol_size) = @_;

    croak "Invalid tag $tag, must be one of 'src', 'dst' or 'cas'"
	unless $tag =~ /(src|dst|cas)/i;

    my %args;
    $args{tag} = $tag;
    $args{vol_size} = ($vol_size ? $vol_size : $pvalue{VOL_SIZE});
    my $filer = "FILER_".uc($tag);
    $args{filer} = $pvalue{$filer};
    $args{aggr} = $pvalue{$filer."_AGGR"} if $pvalue{$filer."_AGGR"};
    $args{vol} = $pvalue{$filer."_VOL"} if $pvalue{$filer."_VOL"};
    $args{aggr_bits} = $pvalue{"AGGR_".uc($tag)."_BITS"} if $pvalue{"AGGR_".uc($tag)."_BITS"};

    return %args;
}

=head2 make_volume

Create a single L<C<NOTE::Volume>|lib-NOTE-Volume-pm> object according to parameter(s).
Volume type will be determined by VOL_TYPE.
For flex vol the size will be 10g for RSM and 300m for VSM/QSM,
and for trad vol it will use 2 disks.

Returns L<C<NOTE::Volume>|lib-NOTE-Volume-pm> object if successful, otherwise croaks.

=over

=item Arguments

B<tag> => I<vol tag>

I<Required>
Tag for the volume.

B<filer> => I<Filer>

I<Required>
Filer where the volume will be created.

B<aggr> => I<aggregate>

I<Optional>
Aggr name of existing flex aggr to use.

B<vol> => I<volume>

I<Optional>
Volume name of existing traditional volume to use.

B<vol_size> => I<size>

I<Optional>
Create volume(s) of this size

B<aggr_bits> => I<32|64>

I<Optional>
Aggregate indirect type, 32 or 64. If not set than use regular aggregate. Default undef

=item Examples

=back

     my $vol = make_volume(tag=>'src', filer=><fsr-test-filer>, vol_size=>'500m');

=cut

sub make_volume {
    my (%opts) = validate(@_, {
			tag => { type => SCALAR, optional => 0 },
			filer => { type => SCALAR, optional => 0 },
			aggr => { type => SCALAR, optional => 1, default => undef },
			vol => { type => SCALAR, optional => 1, default => undef },
			vol_type => { type => SCALAR, optional => 1, default => undef },
			vol_size => { type => SCALAR, optional => 1, default => undef },
			aggr_bits => { type => SCALAR, optional => 1, default => undef }
		});

    croak "init_params needs to be call first"
	unless $need_init;

    my %args;

    my $filer = NOTE::Filer->connect(host=>$opts{filer});
    $args{Filer} = $filer;
    $args{type} = $opts{vol_type} || $pvalue{VOL_TYPE};
    if ( get_wafl_version($filer->hostrec) < WV_ANCHORSTEAM_MIN ) {
	logcomment("Force ".$opts{tag}." Vol to traditional type for pre-Anchorsteam filer");
    	$opts{vol_type} = "trad";
    }
    my $disk_cnt = $pvalue{DISK_CNT}
		|| (($args{type} =~ /flex/i ? 3 : 2)
			# Add more disk for simulators
			+ ( $filer->is_a_simulator ? 3 : 0));
    $disk_cnt = 3 if $args{type} =~ /trad/ && $disk_cnt < 3; # change due to new raid-dp default on filers

    $args{tag} = $opts{tag};
    $args{name} = $opts{vol};
    # Use recycle vol only if trad vol or named volume - The named volume part is due to a bug in
    # NOTE::Volume which does not handle named volume properly if recycle is not set
    $args{recycle} = ($pvalue{VOL_TYPE} =~ /trad/i || $args{name});
    if ( $args{type} eq 'flex' ) {
	$args{Aggregate} = { recycle => 1, disks => $disk_cnt, name => $opts{aggr} };
	$args{Aggregate}{indirect_type} = $opts{aggr_bits} if defined $opts{aggr_bits};
	$args{size} = $opts{vol_size} || $pvalue{VOL_SIZE};
	$args{guarantee} = "none";
    } else {
    	$args{disks} = $disk_cnt;
	$args{zeroed} = 1;
    };

    return NOTE::Volume->create(%args);
}


=head2 ::vol_action

Fills or clears the given NOTE::Volume.
When filling, it creates 1G LUNs until it fails, and completely any residual space by
creating a data file.
When clearing it removes the LUNs and data file it created previously

=over

=item Arguments

B<vol> => I<NOTE::Volume object>

I<C<NOTE::Volume>|lib-NOTE-Volume-pm>>
object to fill/clear.

B<action> => I<fill|clear>

Fills or clears the specify volume

=item Examples

=back

      vol_action(vol=>$root_vol, action=>"fill");

=cut

sub vol_action {
    my (%opts) = validate(@_, {
			vol => { type => HASHREF, optional => 0 },
			action => { type => SCALAR, optional => 0, regex => qr/fill|clear/ }
			});

    croak "vol must be a NOTE::Volume object."
	unless $opts{vol}->isa("NOTE::Volume");

    my $client = NOTE::Client->new(storage=>$opts{vol});
    my $lun_name = "lun_space_".$$."_";
    my $idx = 1;
    my ($lun_action, %args);
    if ( $opts{action} =~ /fill/ ) {
	NOTE::logcomment("Fill up volume ".$opts{vol}->name." with 1G LUNs until it fails...");
	$lun_action = "create_lun";
	$args{size} = "1g";
    } else {
	NOTE::logcomment("Removed LUNs from volume ".$opts{vol}->name." until no more");
	$lun_action = "destroy_lun";
    }
    do {
	eval {
		$client->$lun_action(name=>$lun_name.$idx++, %args);
	};
    } while ! $@;
    $opts{vol}->fillfile(name=>"full_file_".$$);
    if ( $opts{action} =~ /fill/ ) {
	NOTE::logcomment("Fill the rest of the space in volume ".$opts{vol}->name." with data file");
	$opts{vol}->fill(proto=>"local", percent=>100);
    } else {
	NOTE::logcomment("Remove fill file from volume ".$opts{vol}->name);
	$opts{vol}->unfill;
    };
    $opts{vol}->disk_free;
}

=head2 ::fill_aggr

Creates a dummy volume within the aggregate with guarantee option set to 'volume'. Then takes up
all the aggregate's space by incrementally increment the size of the volume until there are no
more space.

Routine returns a NOTE::Volume object that is used to 'fill' the aggregate.

=over

=item Argument

B<aggr> = 
I<C<NOTE::Aggregate>|lib-NOTE-Aggregate-pm>>
object to fill.

=item Examples

=back

    my $fill_vol = fill_aggr($note_aggr_obj);

=cut

sub fill_aggr {
    my ($aggr) = @_;

    croak "Requires a NOTE::Aggregate object."
	unless $aggr->isa("NOTE::Aggregate");

    # Create a dummy volume to take up all the space in the aggr
    my $aggr_space = NOTE::Volume->create(tag=>"aggr_space",
				Filer=>$aggr->Filer, Aggregate=>$aggr, guarantee=>"volume");
    # Repeatedly increase size of vol until no more space in aggregate
    foreach my $increment ("+10g", "+1g", "+100m", "+10m", "+1m", "+100k", "+4k") {
	undef $@; # clear any errors
	my $title = "Inflate vol ".$aggr_space->tag." by $increment:";
	do {
		eval {
			$aggr_space->size($increment);
		};
	} while ! $@;
	if ( $@ =~ /(insufficient space|not enough space in the aggregate)/i ) {
		NOTE::logcomment("$title failed. Error: $@");
	} else {
		die "$title encountered unexpected error: $@";
	}
    }
    return $aggr_space;
}

=head2 sm_action

Retry snapmirror action, skipping specified error, for $timeout sec(s).
If error persist, panices the src filer to provide debugging information to Dev.

=over

=item Arguments

B<sm> => I<NOTE::SnapMirror object>

I<C<NOTE::SnapMirror>|lib-NOTE-SnapMirror-pm>>
object to issue the snapmirror update command.

B<title> => I<string>

Label string to prefix comments in routine. Default: SM-update

B<action> => I<action string>

Snapmirror action to perform. Default: update

B<skip_error> => I<error string>

Error string to skip. Default: 'source volume may be busy'

B<itmeout> => I<integer>

Timeout in sec(s). Default: 7

=item Example

=back

    sm_action(title=>"test", sm=>$sm, timeout=>3);

=cut

sub sm_action {
    my (%opts) = validate(@_, {
			sm => { type => HASHREF, optional => 0 },
			title => { type => SCALAR, optional => 1, default => "SM"},
			action => { type => SCALAR, optional => 1, default => "update"},
			skip_error => { type => SCALAR, optional => 1,
					default => "source volume may be busy"},
			timeout => { type => SCALAR, optional => 1, default => 7 }
			});

    croak "'sm' must be a NOTE::SnapMirror object"
	unless $opts{sm}->isa("NOTE::SnapMirror");

    my $title = $opts{title}."-".$opts{action}.":";
    my $action = $opts{action};
    my $abort_time = time + $opts{timeout};
    my $retry = 0;
    my $skip_err = $opts{skip_error};
    do {
	NOTE::logcomment("$title attempt ".(++$retry));
	eval {
		$opts{sm}->$action;
	};
    } while $@ =~ /$skip_err/i && time < $abort_time;
    if ( $@ =~ /$skip_err/i ) {
	NOTE::logcomment("$title failed after $retry attempts, timed out after "
			.$opts{timeout}." sec(s)");
	NOTE::logcommen ("$title Panic-ed src filer to collect information for Dev");
	$opts{sm}->src->Filer->panic;
    };
    die "$title failed. Error: $@" if $@;
    NOTE::logcomment("$title succeed after $retry attempt(s)");
}

=head2 wreck_test

Utility script to inject error in a VSM dst and to execute the test_fn provided to test the corruption.
The handles 4 types of named corruptions or allow for the specification of specific wreck commands
from the WRECK_CMD cli parameter.

The 4 named types of corruptions are:

1. data - corrupt data blocks using the wreck 'block_pattern' command.
2. block - corrupt vbn/vvbn entries using the wreck 'ind_vol_bno' and 'ind_vol_vbno' cmds respectively.
3. mapping - corrupt pvbn mappings for a flex vol using the wreck 'ind_vol_bno' command.
4. pvbn_unkown - causes pvbn_unknown invariant corruptions using the wreck 'ind_vol_bno' command.

For each of the named types, 3 corruptions points are injected, one for the 1st, middle and last 
byte in the data block or index in the indirect block.

The default is to run 'data' and 'block' corruptions together. For flex volumes, this is followed by
'mapping' and 'pvbn_unkown' corruptions in sequentially.

After injection of each corruption type, the supplied test function is called.

It is possible to override this default by setting WRECK_CMD cli parameter, examples:

	# To execute 'data' corruptions only
	WRECK_CMD=data 

	# To execute 'block' corruptions only and direct all 3 corruptions to a single block
	WRECK_CMD=block:bno=1

	# To execute 'data' and 'mapping' corruptions only
	WRECK_CMD=data,mapping

	# To execute a sinlge 'data' corruption
	WRECK_CMD=data:bno=1:start=10:end=17:fill=123

	# To execute a specific wreck command
	WRECK_CMD=block_ascii:inum=1:bno=2:start=3:end=4:fill=5

=over

=item Arguments

B<sm> = I<C<NOTE::SnapMirror>|lib-NOTE-SnapMirror-pm>>
object to corrupt.

B<test_fn> = A pointer to a test function to call after injection of corruptions.

B<args> = Hash of arguments to pass to the test function 'test_fn'

=item Examples

=back

    wreck_test($sm, \&test_fn);

=cut

sub wreck_test {
    my ($sm, $test_fn, %args) = @_;

    ## @step Create file to corrupt 
    my $size = ($sm->dst->Filer->is_a_simulator ? "10m" : "100m");
    my $inode = $sm->src->nfs->create->fill(proto=>'local', max=>$size)->inode;
    $sm->update;

    my $vol_type = $sm->dst->type;
    my $spec = param_value("WRECK_CMD") || "data|block,mapping,pvbn_unkown";
    NOTE::logcomment("Running wreck_test with $spec");
    foreach my $token (split /,/, $spec) {
	my ($title, $script, $extra);
	foreach my $entry (split /\|/, $token) {
		my ($item, @param) = split ":", $entry;
		$extra = " ".(join " ", @param) if @param;
		$title .= ", " if $title; $title .= $item;
		$script .= "," if $script;
		if ( $item =~ /^(data|block|mapping|pvbn_unkown)/ ) {
			croak "Corruption type $item does not apply for trad vol type"
				if $item =~ /^(mapping|pvbn_unkown)/ && $vol_type =~ /trad/;
			my ($state, $status) = $sm->dst->Aggregate->status; 
			my $ind_cnt = ( $status =~ /64-bit/ ? fileSystem::FLEX_INDIRECT_COUNT_64 : fileSystem::FLEX_INDIRECT_COUNT );
			$script .= _chk_spec ($vol_type, $ind_cnt, $item, $item.":inum=".$inode, $extra);
		} else {
			$script .= $item.(join " ", @param);
		}
	}
	$script = _cvt_name($sm->dst->name, $vol_type, $script);
	NOTE::logcomment("For type: $title, script:\n$script");
	$sm->dst->wreck(inject_script=>$script);
	eval { $test_fn->(%args); };
	croak "wreck_test for type $token found error: $@" if $@;
	NOTE::logcomment("wreck_test for type $title completed");
    }
}

sub _chk_spec {
    my ($vol_type, $flex_ind_cnt, $type, $prefix, $postfix) = @_;

    # Expand a named corruption type into 3 named commands.
    if ( $type =~ /data/ ) {
	my $max = fileSystem::BLOCK_SIZE - 2;
	my $mid = int ($max / 2);
	# Corrupt 1st, middle and last data byte in a block
	return    "$prefix bno=1 start=0 end=1".$postfix
		.",$prefix bno=2 start=$mid end=".($mid+1).$postfix
		.",$prefix bno=3 start=$max end=".($max+1).$postfix;
    } else {
	my $min_idx = 0;
	my $max_idx = ( $vol_type =~ /flex/
			? $flex_ind_cnt
			: fileSystem::TRAD_INDIRECT_COUNT
			) - 1;
	my $mid_idx = int ($max_idx / 2);
	my $bno_key = _bno_key($vol_type);
	# Corrupt 1st, middle and last index in an indirect block
	return    "$prefix bno=1 new_$bno_key=0x12345676 index=$min_idx".$postfix
		.",$prefix bno=511 new_$bno_key=0x12345677 index=$mid_idx".$postfix
		.",$prefix bno=1023 new_$bno_key=0x12345678 index=$max_idx".$postfix;
    }
}

sub _bno_key {
    my ($vol_type) = @_;

    return ($vol_type =~ /flex/ ? "vbno" : "bno");
}

sub _cvt_name {
    my ($vol_name, $vol_type, $script) = @_;

    # Converts list of named corruption type into wreck command
    my %wreck_cmd;
    foreach my $spec (split ",", $script) {
    	if ( $spec =~ /^data:/i ) {
		# Corruption type: data block mismatch
		$wreck_cmd{_cvt_spec ($vol_name, "block_pattern", $spec, "inum",
				"bno=1", "start=501", "end=508", "fill=255")} = 1;
		next;
	}
	if ( $spec =~ /^block:/ ) {
		# Corruption type: indirect block mismatch (vbn/vvbn incorrect)
		$wreck_cmd{_cvt_spec ($vol_name, "ind_vol_"._bno_key($vol_type), $spec,
				"inum", "level=1", "bno=2",
				"new_"._bno_key($vol_type)."=9997", "index=13")} = 1;
		next;
	}
	if ( $spec =~ /^mapping:/ ) {
		# Corruption type: indirect block with inconsistant pvbn mapping (pvbn mismatch)
		croak "mapping corruption type applies to flex vol type only"
			unless $vol_type =~ /flex/;
		$wreck_cmd{_cvt_spec ($vol_name, "ind_vol_bno", $spec,
				"inum", "level=1", "bno=3", "new_bno=1001", "index=17")} = 1;
		next;
	}
	if ( $spec =~ /^pvbn_unkown:/ ) {
		# Corruption type: pvbn_unkown invaraint (pvbn mismatch)
		croak "pvbn_unkown corruption type applies to flex vol type only"
			unless $vol_type =~ /flex/;
		$wreck_cmd{_cvt_spec ($vol_name, "ind_vol_bno", $spec,
				"inum", "level=1", "bno=4", "index=501")." new_bno=-5"} = 1;
		next;
	}
	$wreck_cmd{$spec} = 1;
    };
    return join "\n", sort keys %wreck_cmd;
}

sub _cvt_spec {
    my ($name, $cmd, $spec, @def) = @_;

    # Compose the wreck command from spec and defaults
    my ($type, $cli) = split /:/, $spec;
    my %val = map { split /=/, $_ } (split / /, $cli);
    my $script = $cmd;
    foreach my $token (@def) {
	my ($key, $data) = split /=/, $token;
	$data = $val{$key} if exists $val{$key};
	croak "Missing $key for $type corruption type..."
		unless (defined $data);
	$script .= " ".$key."=".$data;
    }
    NOTE::logcomment("Name: $name, corruption type $type\nspec: $spec, inject: ".$script);
    return $script;
}

=head2 get_ics_list

Creates ICS connection entries in the snapmirror conf file from the available interface cards
for the FILER_SRC and FILER_DST filers. Each ICS connection entry will be made up of 2 ip
addresses. The ip addresses will be read from /etc/hosts file

=over

=item Arguments

I<None.>

=item Example

=back

    my @list = get_ics_list();

=cut

sub get_ics_list {
    my $src_host = $pvalue{'FILER_SRC'};
    my $dst_host = $pvalue{'FILER_DST'};

    my $src_hostrec = NOTE::Filer->connect(host=>$src_host)->hostrec;
    my $reg = NOTE::Filer->connect(host=>$dst_host)->registry;
    my @src_list = _get_host_list($src_host);
    my @dst_list = _get_host_list($dst_host);
    my @ics_list;
    while ( @src_list > 1 && @dst_list > 1 ) {
	my $name = $src_host."-ics_src".(@ics_list + 1);
	NOTE::logcomment("Setting Registry entry for ics $name");
    	my $sm_conn = "options.files.snapmirror.connection.$name.";
	$reg->{$sm_conn."mode"} = "multi";
	foreach my $i (1, 2) {
		my $dst = shift @dst_list;
		sm_set_permissions(host=>$src_hostrec, allowed_host=>$dst);

		$reg->{$sm_conn."ips".$i} = shift @src_list;
		$reg->{$sm_conn."ipd".$i} = $dst;
	}
	push @ics_list, $name;
    }
    return @ics_list;
}

=head2 hostpath_from_icspath


Returns the passed in valued striped of ics_src portion if present.

=over

=item Arguments

B<ics_path> - ics_src name consisting of <filer><ics portion>:<path>

=item Examples

=back

    $hostname = hostpath_from_icspath("filer-ics1:/vol/src"); # Returns 'filer:/vol/src'
    $hostname = hostpath_from_icspath("filer:/vol/src"); # Returns 'filer:/vol/src'

=cut

sub hostpath_from_icspath {
    my ($path) = @_;

    $path =~ s/-ics_src\d+//;
    return $path;
}


=head2 clr_ics_list

Removes ICS connection entries from snapmirror.conf created by get_ics_list () routine

=over

=item Arguments

I<None.>

=item Example

=back

    clr_ics_list();

=cut

sub clr_ics_list {
    my (@ics_list) = @_;

    my $reg = NOTE::Filer->connect(host=>$pvalue{'FILER_DST'})->registry;
    foreach my $ics ( @ics_list ) {
	delete $reg->{"options.files.snapmirror.connection.$ics"};
    }
}

sub _get_host_list {
    my ($filer) = @_;

    my @list;
    my $f = NOTE::Filer->connect(host=>$filer);
    my $text = rdfile($f->conn, "/etc/hosts");
    croak "Error reading /etc/hosts file: $@" if $@;
    foreach my $line (split "\n", $text) {
	next if $line =~ /^#/;
	my ($ip, @term) = split " ", $line;
	if ( (join " ", @term) =~ /$filer-e\d/i ) {
		unshift @list, $ip;
	}
    }
    NOTE::logcomment("ics list for $filer: ".(join ", ", @list));
    return @list;
}

=head2 make_mirror

Create the specified number and type of
L<C<NOTE::SnapMirror>|lib-NOTE-SnapMirror-pm>,
L<C<NOTE::SnapVault>|lib-NOTE-SnapVault-pm>,
L<C<NOTE::Volume::Copy>|lib-NOTE-Volume-Copy-pm>,
L<C<NOTE::Aggregate::Copy>|lib-NOTE-Aggregate-Copy-pm>,
dump_src, dump_dst, restore_src, restore_dst, smstore_src, smstore_dst,
smretrieve_src or smretrieve_dst
objects.

The src and dst volumes or qtrees are created on the respective filers,
FILER_SRC and FILER_DST. Qtrees are created one per volume, unless indicated
otherwise. Dst qtrees are named but not created.
Src qtrees are created unless indicated otherwise

The snapmirror relationships are not initialized. So when calling
initialize for rsm list, need to include "mode=>'sync'" parameter.

Returns an array of
L<C<NOTE::SnapMirror>|lib-NOTE-SnapMirror-pm>,
L<C<NOTE::SnapVault>|lib-NOTE-SnapVault-pm>,
L<C<NOTE::Volume::Copy>|lib-NOTE-Volume-Copy-pm>,
L<C<NOTE::Aggregate::Copy>|lib-NOTE-Aggregate-Copy-pm>,
dump_src, dump_dst, restore_src, restore_dst, smstore_src, smstore_dst,
smretrieve_src or smretrieve_dst
objects.

=over

=item Arguments

B<count> => I<number>

Optional.

The number of L<C<NOTE::Snapmirror>|lib-NOTE-SnapMirror-pm>,
L<C<NOTE::SnapVault>|lib-NOTE-SnapVault-pm>,
L<C<NOTE::Volume::Copy>|lib-NOTE-Volume-Copy-pm>,
L<C<NOTE::Aggregate::Copy>|lib-NOTE-Aggregate-Copy-pm>,
dump_src, dump_dst, restore_src, restore_dst, smstore_src, smstore_dst,
smretrieve_src or smretrieve_dst objects to create. Default 1

B<type> => I<qsm|rsm|vsm|sv|vc|ac|dump_src|dump_dst|restore_src|restore_dst|smstore_src|smstore_dst|smretrieve_src|smretrieve_dst>

Optional.

The type of snapmirror/snapvault/volcopy/aggrcopy/dump/restore/smstore/smretrieve relationship to create with the corresponding
L<C<NOTE::Volume>|lib-NOTE-Volume-pm>
L<C<NOTE::Qtree>|lib-NOTE-Qtree-pm>
or
L<C<NOTE::Aggregate>|lib-NOTE-Aggregate-pm>
objects.
Default vsm

B<dst_per_src> => I<number>

Optional.

Number of dst volumes/qtrees per src volume/qtree. Default 1

B<create_qt_src> => I<number>

Optional.

Create qtree src unless indicated otherwise. Default: 1

B<qt_src> => I<number>

Optional.

A L<C<NOTE::Qtree>|lib-NOTE-Qtree-pm>
object to be used as all QSM source.

B<qt_per_src> => I<number>

Optional.

Determines how many qtree to create per src volume. Default 1

B<qt_per_dst> => I<number>

Optional.

Determines how many qtree to create per dst volume. Default 1

B<src_aggr_prefix> => I<string[:offset]>

Optional.

Specifies pre-defined prefix to src aggregate names.
Name will be created using 'string' followed by number starting from 'offset'
which defaults to 1 if not supplied.

B<src_vol_prefix> => I<string[:offset]>

Optional.

Specifies pre-defined prefix to src volume names.
Name will be created using 'string' followed by number starting from 'offset'
which defaults to 1 if not supplied.

B<src_qt_prefix> => I<string[:offset]>

Optional.

Specified pre-defined prefix to src qtree names.
Name will be created using 'string' followed by number starting from 'offset'
which defaults to 1 if not supplied.

B<devices> => I<array of tape devices>

Optional.

Array reference of tape devices to use.
This option is required for dump_src, dump_dst, restore_src,
restore_dst, smstore_src, smstore_dst, smretrieve_src or smretrieve_dst.
Also, the number of devices must be equal or greater than count.

=item Examples

=back

     my @sm_list = make_mirrors(count=>5, type=>"qsm", qt_src=>$qt_obj);

     my @sm_list = make_mirrors(count=>10, type=>"qsm", qt_per_src=>1, qt_per_dst=>10);

     my @sm_list = make_mirrors(count=>512, type=>"vsm");

=cut

sub make_mirror {
    my (%opts) = @_;

    my $dev_index = 0;
    my $cnt = $opts{count} || 1;
    my $sm_type = "vsm";
    if ( exists $opts{type} ) {
	croak "Invalid snapmirror type ".$opts{type}.". Must be vsm, qsm, ".
              "rsm, sv, vc, ac, dump_src, dump_dst, restore_src, restore_dst, ".
              "smstore_src, smstore_dst, smretrieve_src or smretrieve_dst."
	      unless $opts{type} =~ /^([qvr]sm|sv|vc|ac|(dump|restore|smstore|smretrieve)_(src|dst))$/i;
	$sm_type = $opts{type};
    }

    if ( $opts{type} =~ /^(dump|restore|smstore|smretrieve)_(src|dst)$/i ) {
        croak "Option 'devices' is required for $opts{type}"
          unless exists $opts{devices};
        croak "The number of devices must be equal or greater than count"
          if (ref $opts{devices} ne 'ARRAY' || @{$opts{devices}} < $cnt);
    }

    $opts{dst_per_src} = 1 unless exists $opts{dst_per_src};
    croak "Invalid dst_per_src: $opts{dst_per_src}. Must be > 0"
   	 if $opts{dst_per_src} < 1;
    if ( $sm_type =~ /^(qsm|sv)$/i ) {
	$opts{create_qt_src} = 1 unless exists $opts{create_qt_src};
	$opts{qt_per_src} = 1 unless $opts{qt_per_src};
	croak "Invalid qt_per_src of $opts{qt_per_src}, must be > 0"
		unless $opts{qt_per_src} > 0;

	$opts{qt_per_dst} = 1 unless $opts{qt_per_dst};
	croak "Invalid qt_per_dst of $opts{qt_per_dst}, must be > 0"
		unless $opts{qt_per_dst} > 0;
    }

    NOTE::logcomment("Creating $cnt $sm_type Relationships");
    my $vol_tag = "_".$opts{type}."_vol_";
    my $aggr_tag = "_".$opts{type}."_aggr_";
    my (@list, $src_vol, $dst_vol, $qt_tag, $qt_src_size, $qt_dst_size);
    if ( $sm_type =~ /^(qsm|sv)$/i ) {
	$qt_tag = "_".$opts{type}."_qt_";
	my $qt_size = NOTE::units($pvalue{'QT_SIZE'});
	$qt_src_size = ( $opts{qt_per_src}*$qt_size > NOTE::units($pvalue{'VOL_SIZE'})
			? ($opts{qt_per_src}*$qt_size)."k" : $pvalue{'VOL_SIZE'} );
	$qt_dst_size = ($opts{qt_per_dst}*$qt_size)."k";
	$qt_dst_size = $qt_src_size if NOTE::units($qt_src_size) > NOTE::units($qt_dst_size);
    }
    my ($src_vol_cnt, $dst_vol_cnt, $src_qt_cnt, $dst_qt_cnt) = (1, 1, 0, 0);
    my $src_aggr_cnt = 1;
    ($opts{src_vol_prefix}, $src_vol_cnt) = split ":", $opts{src_vol_prefix}
	if $opts{src_vol_prefix} =~ /:/;
    ($opts{src_aggr_prefix}, $src_aggr_cnt) = split ":", $opts{src_aggr_prefix}
	if $opts{src_aggr_prefix} =~ /:/;
    my $qt_idx = 0;
    if ( $opts{src_qt_prefix} =~ /:/ ) {
	($opts{src_qt_prefix}, $qt_idx) = split ":", $opts{src_qt_prefix};
	$qt_idx--;
    }

    my ($entry, $src) = (0, undef);
    my $ics_idx = 0;
    do {
	my $dst;
	my $need_src = ! ($entry++ % $opts{dst_per_src} && $src);
	if ( $sm_type =~ /^(qsm|sv)$/i ) {
		if ( $opts{qt_src} ) {
			NOTE::logcomment("Using ".$opts{qt_src}->description." for qtree src");
			$src = $opts{qt_src};
		} else {
			if ( $need_src ) {
				$src_vol = _mk_sm_vol('src', $opts{src_vol_prefix},
						$vol_tag, $src_vol_cnt++, $qt_src_size)
					unless ( ($src_qt_cnt++ % $opts{qt_per_src}) && $src_vol );
				$src = _mk_sm_qt($opts{create_qt_src}, 'src',
						$opts{src_qt_prefix},
						$qt_tag, $qt_idx + $src_qt_cnt, $src_vol);
			}
		}
		$dst_vol = _mk_sm_vol('dst', undef, $vol_tag, $dst_vol_cnt++, $qt_dst_size)
			unless ( ($dst_qt_cnt++ % $opts{qt_per_dst}) && $dst_vol );
		$dst = _mk_sm_qt(0, 'dst', undef, $qt_tag, $dst_qt_cnt, $dst_vol);
	} elsif ( $sm_type =~ /^ac$/i ) {
		$src = _mk_sm_aggr('src', $opts{src_aggr_prefix}, $aggr_tag, $src_aggr_cnt) if $need_src;
		$dst = _mk_sm_aggr('dst', undef, $aggr_tag, $entry);
	} elsif ( $sm_type =~ /^(dump|smstore)_(src|dst)$/i ) {
                # For dump and smstore, $src is the volume object to be
                # dumped/smstored and is located on either the source or
                # destination filer according to suffix of sm_type
                my ($ftype) = ($sm_type =~ /_(.+)/);
		$src = _mk_sm_vol($ftype, undef, $vol_tag, $entry);
	} elsif ( $sm_type =~ /^(restore|smretrieve)_(src|dst)$/i ) {
                # For restore and smretrieve, $dst is the volume object to be
                # restored/smretrieved and is located on either the source or
                # destination filer according to suffix of sm_type
                my ($ftype) = ($sm_type =~ /_(.+)/);
		$dst = _mk_sm_vol($ftype, undef, $vol_tag, $entry);
	} else {
		$src = _mk_sm_vol('src', $opts{src_vol_prefix}, $vol_tag, $entry) if $need_src;
		$dst = _mk_sm_vol('dst', undef, $vol_tag, $entry);
	}
	NOTE::logcomment("Create $sm_type object $entry");
	# Use ics_list if supplied
	my (%args, $sm);
	$args{preprocess} = 1;
	$args{qt_src_exists} = $opts{create_qt_src};
	if ( $opts{ics_list} ) {
		# Rotate ics connection if ics_list is supplied
		$args{ics_src} = $opts{ics_list}->[$ics_idx++];
		$ics_idx = 0 if $ics_idx >= @{$opts{ics_list}};
	}
	if ( $sm_type =~ /^sv$/i ) {
		$sm = NOTE::SnapVault->new(src=>$src, dst=>$dst, %args);
	} elsif ( $sm_type =~ /^vc$/i ) {
		$sm = NOTE::Volume::Copy->new(src=>$src, dst=>$dst, %args);
	} elsif ( $sm_type =~ /^ac$/i ) {
		$sm = NOTE::Aggregate::Copy->new(src=>$src, dst=>$dst, %args);
	} elsif ( $sm_type =~ /^(dump|restore|smstore|smretrieve)_(src|dst)$/i ) {
                # The src and dst hash entries for these types do not
                # necessarily mean volumes in the src and dst filers
                # respectively, but the source and destination of the
                # transfer itself which is done either at the source or
                # destination filer according to suffix of sm_type
                my ($src_path, $dst_path);
                my $device = $opts{devices}[$dev_index++];
                if ( $sm_type =~ /(dump|smstore)/ ) {
                    $dst_path = $src->Filer->host . ":$device";
                    $sm = {
                        type     => $sm_type,
                        src      => $src,
                        dst      => undef,
                        device   => $device,
                        desc     => "\n    +---" . $src->description .
                                    "\n    +-> " . $dst_path,
                        src_path => $src->hostpath,
                        dst_path => $dst_path,
                    };
                } else {
                    $src_path = $dst->Filer->host . ":$device";
                    $sm = {
                        type     => $sm_type,
                        src      => undef,
                        dst      => $dst,
                        device   => $device,
                        desc     => "\n    +---" . $src_path .
                                    "\n    +-> " . $dst->description,
                        src_path => $src_path,
                        dst_path => $dst->hostpath,
                    };
                }
                # These transfer types do not have a mechanism to get the
                # status of the transfer are blessed as an object in the
                # current library to make these transfers consistent with
                # snapmirror, snapvault and vol/aggr copy objects
                bless $sm;
	} else {
		$sm = NOTE::SnapMirror->new(src=>$src, dst=>$dst, %args);
	}
	# Manually add object to NOTE's stack for teardown
	NOTE::reference( $sm ) if $sm_type =~ /^([qvr]sm|sv|[va]c)$/;
	push @list, $sm;
    } while @list < $cnt;
    NOTE::logcomment("Created $cnt $sm_type Relationships");

    return @list;
}

sub _mk_sm_qt {
    my ($create, $type, $prefix, $tag, $cnt, $vol) = @_;

    my $qt_label = $tag.$cnt;

    my %opts;
    $opts{name} = $prefix.$cnt if $type =~ /src/ && $prefix;
    $opts{tag} = $type.$qt_label;
    $opts{Volume} = $vol;
    if ( $create ) {
	NOTE::logcomment("Create sm $opts{tag} in vol ".$vol->description);
	return NOTE::Qtree->create(%opts)
    } else {
	NOTE::logcomment("Map sm $opts{tag} in vol ".$vol->description);
	return NOTE::Qtree->new(%opts);
    }
}

sub _mk_sm_vol {
    my ($type, $prefix, $tag, $cnt, $size) = @_;

    $pvalue{'FILER_SRC_VOL'} = $prefix.$cnt if $type =~ /src/i && $prefix;
    my $label = $tag.$cnt;
    my $tag = $type.$label;
    NOTE::logcomment("Create sm volume $tag");
    my $vol = make_volume(_vol_opts($type, $size));
    $vol->tag($tag);
    return $vol;
}

sub _mk_sm_aggr {
    my ($type, $prefix, $tag, $cnt, $size) = @_;
    my %args;

    $args{name} = $prefix.$cnt if $type =~ /src/i && $prefix;
    $args{disks} = $size if $size;
    my $label = $tag.$cnt;
    my $tag = $type.$label;
    my $filer = "FILER_".uc($type);

    NOTE::logcomment("Create sm aggregate $tag");
    return NOTE::Aggregate->create(
        Filer => { host => $pvalue{$filer} },
        tag   => $tag,
        share => 0,
        recycle => 1,
        %args,
    );
}

=head2 do_mirror

Perform the specified action on the supplied list of objects.
The actions are perfrom in the background in parallel.

Note that the array of 
L<C<NOTE::SnapMirror>|lib-NOTE-SnapMirror-pm>,
L<C<NOTE::SnapVault>|lib-NOTE-SnapVault-pm>,
L<C<NOTE::Volume::Copy>|lib-NOTE-Volume-Copy-pm>,
L<C<NOTE::Aggregate::Copy>|lib-NOTE-Aggregate-Copy-pm>,
dump_src, dump_dst, restore_src, restore_dst, smstore_src, smstore_dst,
smretrieve_src or smretrieve_dst
objects are not initialized, so when calling the 'initialize' action, need to set "mode=>'sync'"
parameter

Croaks on detection of failures.

=over

=item Arguments

B<action> => I<sm actions|DataManip action>

Action to perform. One of:
populate, modify (on src) or any of the snapmirror operations such as initialize, update etc

B<sm_list> => I<array of sm objects>

A reference or pointer to an array of
L<C<NOTE::SnapMirror>|lib-NOTE-SnapMirror-pm>,
L<C<NOTE::SnapVault>|lib-NOTE-SnapVault-pm>,
L<C<NOTE::Volume::Copy>|lib-NOTE-Volume-Copy-pm>,
L<C<NOTE::Aggregate::Copy>|lib-NOTE-Aggregate-Copy-pm>,
dump_src, dump_dst, restore_src, restore_dst, smstore_src, smstore_dst,
smretrieve_src or smretrieve_dst
objects

B<opts> => I<hash of options>

A hash of options to pass to the
L<C<NOTE::Snapmirror>|lib-NOTE-Snapmirror-pm>
L<C<NOTE::SnapVault>|lib-NOTE-SnapVault-pm>,
L<C<NOTE::Volume::Copy>|lib-NOTE-Volume-Copy-pm>,
L<C<NOTE::Aggregate::Copy>|lib-NOTE-Aggregate-Copy-pm>,
dump_src, dump_dst, restore_src, restore_dst, smstore_src, smstore_dst,
smretrieve_src or smretrieve_dst
objects, such as:

B<background> => I<1> will returns a list of subtest running the
snapmirror commands. Default: 0

B<client> => I<hostspec>. Client to perform 'src' action on.
If this point to a list then assign in round robin style.

B<ipc> => C<0|1>

This is used to transfer the information of the transfer identifier to the
calling process via ipc. The transfer identifier is needed if calling process
needs to check the status of transfers for the following types: vc, ac,
dump_src, dump_dst, restore_src, restore_dst, smstore_src, smstore_dst,
smretrieve_src, smretrieve_dst. Default is 0.

All other parameters will be passed to the specified action.

=item Examples

=back

     do_mirror('populate', \@sm_list, 'FILE_DEV'=>100);

     do_mirror('initialize', \@sm_list, 'throttle'=>'512k');

     do_mirror('initialize', \@rsm_list, 'mode'=>'sync'); # Initializing RSM list

     my $proc_list = do_mirror('populate', \@sm_list, 'background' => 1);

=cut

sub do_mirror {
    my ($action, $sm_list, %opts) = @_;

    my %src_op = (	'populate' => 1
		,	'modify' => 1
		,	'fill' => 1
		,	'sio' => 1
	);

    NOTE::logcomment("Performing $action on ".@{$sm_list}." mirror(s)");
    my (@list, @ipc_opt);
    my $run_at_src = $src_op{$action};
    my $bkgd = $opts{background};
    delete $opts{background};

    # Use the ipc option on the subtest
    @ipc_opt = qw(-ipc) if $opts{ipc};

    my ($idx, @client_list) = (0, ());
    if ( exists $opts{client} ) {
	if  ( ref $opts{client} eq "ARRAY" ) {
		@client_list = @{$opts{client}};
	} else {
		push @client_list, $opts{client};
	}
	delete $opts{client};
    }
    $opts{parallel} = 1 if $action =~ /initialize|resync|update/i; # Force parallel mode
    foreach my $obj (@{$sm_list}) {
        next if ($run_at_src && $obj->{type} =~ /restore|smretrieve/);
	my $client = (@client_list ? $client_list[$idx++ % @client_list] : "localhost");
	if ( $obj->isa ('NOTE::SnapMirror')
		# Exclude RSM initialize
		&& ! ( $action =~ /initialize/ && $opts{mode} eq "sync" )
		# Exclude NOTE::SnapMirror destroy,wait_sync method
		&& $action !~ /destroy|wait_sync/ ) {
		# Selected Snapmirror commands in a remote script
		delete $opts{mode} if $action =~ /initialize/; # delete mode as it is only used in Helper.pm
		my $title = "SM: ".$obj->src->description." -> ".$obj->dst->description;
		my $script = _remote_sm_script($title, $action, $obj, %opts);
		NOTE::logcomment($title."\nAction $action".", Client ".$client.", Remote script:\n$script");
		push @list, subtest ( $script, '-bg', '-client', $client, '-onexit', sub {});
	} else {
		# All other methods in a subtest
		$opts{client} = $client unless $client eq "localhost";
		push @list, subtest ( sub { _do_action($action, $run_at_src, $obj, %opts) },
					'-bg', @ipc_opt );
	}
    }
    if ( $bkgd ) {
	return @list;
    } else {
	my $rc =  monitor_mirror($action, @list);
	if ( $rc and $opts{parallel} ) {
		$rc =  wait_mirror($sm_list->[0]->dst->Filer,
				($opts{mode} eq "sync" ? "in-sync" : "idle"),
				@{$sm_list})
	}
	return $rc;
    }
}

sub _remote_sm_script {
    my ($label, $action, $sm, %opts) = @_;

    my $title = $sm->src->description." -> ".$sm->dst->description;
    my $cmd = "snapmirror_".$action;
    my $script = "{
use Tharn;
use SnapMirror;

my %opts = (
";
    my $cnt = 0;
    foreach my $key (keys %opts) {
	$opts{$key} = '"'.$opts{$key}.'"' unless $opts{$key} =~ /\d+/;
	$script .= ",\n" if $cnt++;
	$script .= "	$key => ".$opts{$key};
    }
    $script .= ");
\$opts{host} = Tharn::host('"
.($action =~ /migrate|release/ ? $sm->src->Filer->host : $sm->dst->Filer->host)."');
";
    if ( $action =~ /abort|check/ ) {
	$script .= "\$opts{dst} = '".$sm->dst_path."';";
    }
    if ( $action =~ /initialize|migrate|resync|update|destroy/ ) {
	$script .= "\$opts{src} = '".$sm->src_path."';
\$opts{dst} = '".$sm->dst_path."';";
    }
    if ( $action =~ /break|quiesce|resume/ ) {
	$script .= "\$opts{dstpath} = '".$sm->dst_path."';";
    }
    if ( $action =~ /release/ ) {
	$script .= "\$opts{srcpath} = '".$sm->src_name."';
\$opts{dst} = '".$sm->dst_path."';";
    }
    $script .= "
eval {
    SnapMirror::$cmd(%opts);
};
die \"$title Failed: \$\@\"
    if \$\@;
}";
    return $script;
}

=head2 monitor_mirror

Monitors and returns 'worse' result from list of processes returned by 'do_mirror'
Note for actions such as initialize and update, where it is performed in 'parallel' mode,
the result signifies that the action has started in the background and may not have
completed. Use wait_mirror to verify when action completes.

=over

=item Arguments

B<label> => I<text>

A label used to describe the list of processes being passed, typically it is the
snapmirror action being performed by the list of processes.

B<proc_list> => I<array of subtests>

Array of subtests returned by 'do_mirror'

=item Example

=back

     my $rc = monitor_mirror('update', @list);

=cut

sub monitor_mirror {
    my ($label, @proc_list) = @_;

    my $bad = 0;
    foreach my $proc (@proc_list) {
	$proc->wait_finish();
	my $worstResult = NATE::Result::worst($proc->results());
	$bad += ($worstResult->severity() >= NATE::Result::Fail->severity())
			if $worstResult;
    }
    croak ("Error detected in parallel mirror $label action") if $bad;
    NOTE::logcomment("Parallel mirror $label action completed successfully");

    return 1;
}

sub _do_action {
    my ($action, $on_src, $sm, %opts) = @_;
    my $ipc = 0;

    if ( exists $opts{ipc} ) {
        $ipc = $opts{ipc};
        delete $opts{ipc};
    }

    my $title = "SM $action ".$sm->description;
    NOTE::logcomment("Starting ".$title);
    eval {
	if ( $on_src ) {
            # Get proper volume/qtree to populate
            my $obj = get_pop_vol(obj => $sm->src);
            $obj->$action(%opts);
	} else {
            if ( ($action eq 'initialize' || $action eq 'update')
              && ($sm->isa('NOTE::Volume::Copy')
               || $sm->isa('NOTE::Aggregate::Copy')) ) {
                # The vol/aggr copy objects use the start method
                $sm->start(%opts, background=>1);
                if ( $ipc ) {
                    # Send operation number to the parent process
                    my $msg = $sm->description . " (id = " . $sm->op_num . ")";
                    NOTE::logcomment("Send to parent process: $msg");
                    Subtest::ipc->write_message($msg);
                }
                # Wait for transfer to finish
                $sm->wait;
            } else {
                if ( $action eq 'initialize' && $sm->isa('NOTE::SnapVault') ) {
                    # The SnapVault object uses the start method
                    $action = 'start';
                }

                if ( $ipc && $sm->isa('dp::snapmirror::Utils') ) {
                    # Send process id to the parent process
                    $sm->{pid} = $$;
                    my $msg = $sm->description . " (id = $$)";
                    NOTE::logcomment("Send to parent process: $msg");
                    Subtest::ipc->write_message($msg);
                }

                $sm->$action(%opts);
            }
	}
    };
    if ( $@ ) {
	logresult("FAIL", "$title Failed. Error: $@");
    } else {
	logresult("PASS", "$title succeed");
    }
}

=head2 wait_mirror

Given an array of 
L<C<NOTE::SnapMirror>|lib-NOTE-SnapMirror-pm>
objects, will wait until all relationships is in the specified end_status or failed.

=over

=item Arguments

B<filer> => I<dst filer>

Filer to query snapmirror statuses from

B<end_status> => I<text>

Expected state,status combination for list of snapmirror relationships

B<sm_list> => I<array of L<C<NOTE::SnapMirror>|lib-NOTE-SnapMirror-pm>>

An array of L<C<NOTE::SnapMirror>|lib-NOTE-SnapMirror-pm> objects

=item Example

    wait_mirror($dst_filer, "idle", @sm_list);
    wait_mirror($dst_filer, "snapmirrored:idle", @sm_list);

=back

=cut

sub wait_mirror {
    my ($filer, $end_status, @sm_list) = @_;

    my $end_status = lc($end_status);

    my $error;
    # Default to 5 secs, and add a sec per 2 extra mirrors, cutoff at 60 secs
    my $interval = 5;
    $interval += int ((@sm_list - 5) / 2) if @sm_list > 5;
    $interval = 60 if $interval > 60;
    while ( @sm_list ) {
	NOTE::logcomment("Pausing $interval secs in between status check of ".@sm_list." transfers...");
	sleep $interval; # pause in between checks
	my @stats = stats_mirror($filer, @sm_list);
	my $idx = 0;
	foreach my $status (@stats) {
		$status = lc ($status);
		if ( $status =~ /$end_status/ ) {
			splice @sm_list, $idx, 1;
		} else {
			if ( lc($status) =~ /transferring|quiescing/i ) {
				$idx++;
			} else {
				my $bad_sm = splice @sm_list, $idx, 1;
				$error .= "\n" if $error;
				$error .= $bad_sm->description." status: $status";
			}
		}
	}
    }
    croak "Encountered following error(s):\n$error" if $error;
    return 1;
}

=head2 stats_mirror

Takes a list of
L<C<NOTE::SnapMirror>|lib-NOTE-SnapMirror-pm>,
L<C<NOTE::SnapVault>|lib-NOTE-SnapVault-pm>,
L<C<NOTE::Volume::Copy>|lib-NOTE-Volume-Copy-pm>,
L<C<NOTE::Aggregate::Copy>|lib-NOTE-Aggregate-Copy-pm>,
dump_src, dump_dst, restore_src, restore_dst, smstore_src, smstore_dst,
smretrieve_src and/or smretrieve_dst
objects, and returns their 'state:status' of mirrors in the same order
'current-transfer-error' is also added if found

=over

=item Arguments

B<filer> => I<filer>

A L<C<NOTE::Filer>|lib-NOTE-Filer-pm>
object to get statuses from

B<sm_list> => I<array of sm objects>

An array of
L<C<NOTE::SnapMirror>|lib-NOTE-SnapMirror-pm>,
L<C<NOTE::SnapVault>|lib-NOTE-SnapVault-pm>,
L<C<NOTE::Volume::Copy>|lib-NOTE-Volume-Copy-pm>,
L<C<NOTE::Aggregate::Copy>|lib-NOTE-Aggregate-Copy-pm>,
dump_src, dump_dst, restore_src, restore_dst, smstore_src, smstore_dst,
smretrieve_src and/or smretrieve_dst objects.

=item Example

=back

    my @status_list = stats_mirror($filer, @sm_list);

=cut

sub stats_mirror {
    my ($filer, @sm_list) = @_;

    NOTE::logcomment("Checking statuses for ".@sm_list." mirror(s)");
    my (@stats, @list, %type_h, $sm, $type, $out, $function);
    # Find out which type of objects the status is needed for
    foreach $sm ( @sm_list ) {
        if ( $sm->isa('NOTE::SnapMirror') ) {
            $type_h{sm} = 'snapmirror_status';
        } elsif ( $sm->isa('NOTE::SnapVault') ) {
            $type_h{sv} = 'snapvault_status';
        } elsif ( $sm->isa('NOTE::Volume::Copy') ) {
            $type_h{vc} = 'volcopy_status';
        } elsif ( $sm->isa('NOTE::Aggregate::Copy') ) {
            $type_h{ac} = 'aggrcopy_status';
        } elsif ( $sm->isa('dp::snapmirror::Utils') ) {
            # Objects blessed in this library
            $type_h{$sm->type} = 1;
        }
    }
    foreach $type ( keys %type_h ) {
        if ( $type =~ /^(dump|restore|smstore|smretrieve)_(src|dst)$/ ) {
            # For transfers which there is no good way to get the status
            # just check the subtest running the transfer. If subtest is
            # running then it is still transferring otherwise set the
            # status to 'idle'
            foreach my $mirror (@sm_list) {
                next if $mirror->{type} ne $type;
                push @$out, {
                    status                 => 'idle',
                    'source-location'      => $mirror->src_path,
                    'destination-location' => $mirror->dst_path,
                };
                if ( $mirror->{pid} && kill 0, $mirror->{pid} ) {
                    # Transfer subtest is still alive
                    $out->[-1]{status} = 'transferring';
                }
                $out->[-1]{state} = $type;
            }
        } else {
            # Call the appropriate function to get the status for all
            # transfers of this type on the given filer
            $function = $type_h{$type};
            $out = $filer->$function;
        }
        if ( $type eq 'sv' ) {
            # Change the src/dst location for the SnapVault status to
            # match the status on SnapMirror
            foreach my $s ( @$out ) {
                $s->{'source-location'} = $s->{'source-system'} .
                  ':' . $s->{'source-path'};
                $s->{'destination-location'} = $s->{'destination-system'} .
                  ':' . $s->{'destination-path'};
            }
        }
        push @list, @$out;
    }

    my $err_label = "current-transfer-error";
    foreach my $mirror (@sm_list) {
	my $status = "no status";
	foreach my $s (@list) {
            if ( $mirror->isa('NOTE::Volume::Copy')
              || $mirror->isa('NOTE::Aggregate::Copy') ) {
                if ( $mirror->op_num == $s->{op_num} ) {
                    $status = $s->{progress} ? 'transferring' : 'idle';
                    last;
                }
            } else {
                if ( $mirror->src_path eq $s->{'source-location'}
                  && $mirror->dst_path eq $s->{'destination-location'} ) {
                    $status = $s->{'state'} . ':' . $s->{'status'} .
				# ignore error due to scheduled transfers - work-around for RSM
                              (	$s->{'current-transfer-type'} !~ /scheduled/
				&& exists $s->{$err_label} ?
                              ", $err_label: " . $s->{$err_label} : '');
                    last;
                }
            }
	}
	push @stats, $status;
    }
    return @stats;
}

=head2 mount_object

Mount the specified
L<C<NOTE::Volume>|lib-NOTE-Volume-pm>
or
L<C<NOTE::Qtree>|lib-NOTE-Qtree-pm>
object. Returns 1 on success and croaks upon failure.

=over

=item Arguments

B<object> => I<vol or qtree obj>

A L<C<NOTE::Volume>|lib-NOTE-Volume-pm>
or
L<C<NOTE::Qtree>|lib-NOTE-Qtree-pm>
object.

B<exportfs> => I<exportfs string>

Exportfs parameter. Default "rw,nosuid,anon=0".

B<mode> => I<unix mode>

Chmod mode. Default 777.

=item Example

=back

     mount_object(object=>$vol_obj, exportfs=>"rw,anon=0", mode=> 755);

=cut

sub mount_object {
    my (%opts) = validate (@_, {
			object => { type => HASHREF, optional => 0 },
			exportfs => {
				type => SCALAR,
				optional => 1,
				default => "rw,nosuid,anon=0" },
			mode => {
				type => SCALAR,
				optional => 1,
				default => '777' }
		});
			

    $opts{object}->exportfs($opts{exportfs});
    $opts{object}->mount(chmod=>$opts{mode});
    return 1;
}

=head2 sm_fiji

Perform various snapmirror fiji actions

=over

=item Arguments

B<filer> = I<L<C<NOTE::Qtree>|lib-NOTE-Qtree-pm>>

A L<C<NOTE::Qtree>|lib-NOTE-Qtree-pm>
object to received the fiji command

B<action> => I<set|pause|resume>

Action to perform on the fiji

B<dst_path> => I<dst hostpath>

B<state. => I<state>

B<fiji_action> => I<fiji action>

Optional fiji action.

=item Example

=back

      sm_fiji($dst_filer, action=>"set", state=>"FIJI state to set");

=cut

sub sm_fiji {
    my (%opts) = validate (@_, {
			filer => { type => HASHREF, optional => 0 },
			action => { type => SCALAR, optional => 0,
					regex => qr/set|pause|resume/ },
			dst_path => { type => SCALAR, optional => 0 },
			state => {type => SCALAR, optional => 0 },
			fiji_action => { type => SCALAR, optional => 1 }
		});

    if ( $opts{action} =~ /set/ ) {
	croak "sm_fiji 'set' requires 'fiji_action' parameter"
		unless exists $opts{fiji_action};
	my $text = sayEOF($opts{filer}->conn, PRIV_TEST_CMD
		. "fiji filter \"lrep.at.state "
		. "(&& (eq \@dstpath ".$opts{dst_path}.") (== \@state ".$opts{state}.")) "
		. "(".$opts{fiji_action}.")\"");
	croak "Failed to setup fiji. Error: $text"
		unless $text =~ /filter set/i;
    } else {
	sayEOF($opts{filer}->conn, PRIV_TEST_CMD
		. "fiji $opts{action} $opts{dst_path} $opts{state}"
		. ($opts{action} =~ /pause/ ? " 1" : "")
		, TIMEOUT=>120);
    }
}

=head2 check_DUMP_FILE

Verify that the tape_device specified by DUMP_FILE cli parameter exists and ready.
Dies otherwise

=over

=item Argument

B<filer> => I<filer>

A L<C<NOTE::Filer>|lib-NOTE-Filer-pm>
object that contains the tape device

=item Example

=back

      check_DUMP_FILE($dst_filer);

=cut

sub check_DUMP_FILE {
    my (%opts) = validate(@_, {
			filer => { type => HASHREF, optional => 0 },
		});

    if ( $pvalue{DUMP_FILE} !~ /^\// ) {
	croak "Tape device ".$pvalue{DUMP_FILE}." does not exists or not ready"
		unless $opts{filer}->tape_ready(tape_device=>$pvalue{DUMP_FILE});
    };
}

=head2 lun_type_list

Returns a list of agnostic lun specs for test scripts to create

=over

=item Argument

B<filer> => I<filer>

A L<C<NOTE::Filer>|lib-NOTE-Filer-pm>
object to determine the filer version

=item Example

=back

      my @list = lun_type_list($src_filer);

=cut

sub lun_type_list {
    my (%opts) = validate(@_, {
			filer => { type => HASHREF, optional => 0 },
		});

    my @lun_list = (
	[ "aix", "4m" ],
	[ "hpux", "8m" ],
	[ "solaris", "12m" ],
	[ "windows", "32m" ]
    );

    if ( get_wafl_version($opts{filer}->hostrec) > WV_ANCHORSTEAM_MAX ) {
	# post anchorsteam supported lun type
	push @lun_list, [ "netware", "4m" ];
	push @lun_list, [ "vmware", "16m" ];
    }

    return @lun_list;
}

=head2 check_volumes

Perform parallel wafliron checks on a supplied list of
L<C<NOTE::Volume>|lib-NOTE-Volume-pm>.
objects.
For flex volumes, it will run wafliron only once per aggregate / filer.

Croaks on the first wafliron error detected, otherwise returns 1.

=over

=item Argument

B<vols> = I<array of L<C<NOTE::Volume>|lib-NOTE-Volume-pm>>

List of volumes to be checked  with wafliron in parallel.

=back

=cut

sub check_volumes {
    my (@vol) = @_;

    my @run_chk = (1) x @vol;
    my $idx = 0;
    do {
	if ( $vol[$idx]->type eq 'flex' ) {
		my ($new_filer, $new_aggr) = ($vol[$idx]->Filer->hostspec,
			$vol[$idx]->Aggregate->name);
		# Verify for flex vols that it does not exists on the same
		# aggregate as another volume, as wafliron is done on an
		# aggregate basis for flex vols
		my $src = 0;
		while ( $src < $idx ) {
			if ( $new_filer eq $vol[$src]->Filer->hostspec
			&& $new_aggr eq $vol[$src]->Aggregate->name ) {
				$run_chk[$idx] = 0;
				last;
			} else {
				$src++;
			}
		};
	}
	if ( $run_chk[$idx] ) {
		$vol[$idx]->wafliron_start(wait=>0);
	}
    } while ++$idx < @vol;

    $idx = 0;
    do {
	if ( $run_chk[$idx] ) {
		my $label = "wafliron of ".$vol[$idx]->name
				." (".$vol[$idx]->tag
				.") on Filer ".$vol[$idx]->Filer->hostspec;
		logcomment("Wait for $label to complete:");
		my ($bad, $msg) = $vol[$idx]->wafliron_wait();
		if ( $bad ) {
			croak "$label found corruptions: $msg";
		} else {
			logcomment("$label completed with no errors");
		}
	}
    } while ++$idx < @vol;

    return 1;
}

=head2 get_pop_vol

This function is used to get the appropriate NOTE object to populate.
For an aggregate object, which does not have a populate method, a volume
is created on that aggregate and the volume object is returned. For all
other objects which are either a volume or qtree, that object is returned.

If the object is an aggregate, a volume is created the first time this function
is called, subsequent calls will reuse the previously created volume.

=over

=item Arguments

B<obj> => I<NOTE volume, qtree or aggregate object>

A L<C<NOTE::Volume>|lib-NOTE-Volume-pm>
L<C<NOTE::Qtree>|lib-NOTE-Qtree-pm>,
or
A L<C<NOTE::Aggregate>|lib-NOTE-Aggregate-pm>
object.

=back

=cut

sub get_pop_vol {
    my %opts = @_;
    my $obj = $opts{obj};

    if ( $obj->isa('NOTE::Aggregate') ) {
        if ( ! $aggr_pop_vols{$obj->hostpath} ) {
            # Create volume on aggregate
            $aggr_pop_vols{$obj->hostpath} = NOTE::Volume->create(
                Filer     => $obj->Filer,
                Aggregate => $obj,
                size      => $pvalue{VOL_SIZE},
                tag       => 'populate volume for ' . $obj->hostpath,
            );
        }
        # The volume on this aggregate has been created already
        $obj = $aggr_pop_vols{$obj->hostpath};
    }

    return $obj;
}

=head2 get_messages

Get all pending messages from child processes. It returns all messages
received in the allowed timeout period and sends them all to the parent
process if the propagate option is set.

=over

=item Arguments

B<subtests> => I<array of subtest objects>

Array of subtest objects.

B<propagate> => C<0|1>

Propagate all the messages received to the parent process. Default is 0.

B<timeout> => I<number>

Timeout in seconds before returning. Default is 0.

=back

=cut

sub get_messages {
    my (%opts) = @_;
    my (@list, %filenos, $sobj, $message, $rset, $read, $sock, $etime);
    my $timeout   = $opts{timeout} || 0;
    my $propagate = $opts{propagate} || 0;

    # Create select object
    $rset = IO::Select->new;

    # Add all subtest sockets to select
    foreach $sobj ( @{$opts{subtests}} ) {
        next unless $sobj->is_alive;
        $rset->add($sobj->ipc);
        # Add table entry to translate file number of socket
        # to proper subtest object
        $filenos{$sobj->ipc->fileno} = $sobj;
    }

    # Set expiration time
    $etime = time + $timeout;

    do {
        # Wait for sockets to be ready for read
        $read = (IO::Select->select($rset, undef, undef, $timeout))[0];

        # Read all sockets which are ready
        foreach $sock ( @$read ) {
            # Get subtest object from socket file number
            $sobj = $filenos{$sock->fileno};
            next unless $sobj->is_alive;

            # Get message from child process
            $message = eval { $sobj->ipc->read_message };
            next unless $message;
            NOTE::logcomment("Received from " . $sobj->runid . ": $message");
            push @list, $message;

            if ( $propagate ) {
                # Send message to parent process
                NOTE::logcomment("Send to parent process: $message");
                Subtest::ipc->write_message($message);
            }
        }
    } while ( $etime > time );

    return \@list;
}

=head2 get_sm_id

Get the transfer id from the message and finds the appropriate transfer
object for that id.

=over

=item Arguments

B<sm_list> => I<array of sm objects>

An array of
L<C<NOTE::SnapMirror>|lib-NOTE-SnapMirror-pm>,
L<C<NOTE::SnapVault>|lib-NOTE-SnapVault-pm>,
L<C<NOTE::Volume::Copy>|lib-NOTE-Volume-Copy-pm>,
L<C<NOTE::Aggregate::Copy>|lib-NOTE-Aggregate-Copy-pm>,
dump_src, dump_dst, restore_src, restore_dst, smstore_src, smstore_dst,
smretrieve_src and/or smretrieve_dst objects.

B<message> => I<string>

One of the elements returned from L<C<get_messages>|/get_messages>.

=back

=cut

sub get_sm_id {
    my (%opts) = @_;
    my ($sm, $desc, $id);

    foreach $sm ( @{$opts{sm_list}} ) {
        $desc = quotemeta($sm->description);
        if ( ($id) = ($opts{message} =~ /$desc.*\(id = (\d+)\)/s) ) {
            # Message matched one relationship
            if ( $sm->isa('NOTE::Volume::Copy')
              || $sm->isa('NOTE::Aggregate::Copy') ) {
                # For vol/aggr copy the id is the operation number
                $sm->{_op_num} = $id;
            } else {
                # For all others the id is the process id
                $sm->{pid} = $id;
            }
        }
    }

    return 1;
}

# Internal function to dump the contents of a volume
sub _filer_volume_dump ($$$;$) {
    my ($fc, $volume, $device, $level) = @_;
    my ($cmd);

    # Options for dump
    my $opts = 'f';

    $cmd = "dump $level$opts $device $volume";
    # Dump the contents of volume to device
    my $text = sayEOF($fc, $cmd);

    # Make sure the dump is done
    if ( $text !~ /DUMP IS DONE/ ) {
        croak("Volume dump failed");
    }
    return 1;
}

# Internal function to restore contents of dump to a volume
sub _filer_volume_restore ($$$;) {
    my ($fc, $device, $dir) = @_;
    my ($cmd);

    # Options for dump
    my $opts = 'rf';

    # If the directory to restore in is specified then add the option
    if ($dir ne '') {
        $opts .= 'D';
    }
    $cmd = "restore $opts $device $dir";
    # Restore the contents of volume from device
    my $text = sayEOF($fc, $cmd);

    # Make sure the restore is done
    if ( $text !~ /RESTORE IS DONE/ ) {
        croak("Volume restore failed");
    }
    return 1;
}

#===============================================================================
# Methods for objects blessed in this library to emulate NOTE most basic methods
#
# initialize
#   Start transfer according to type: dump, restore, smstore and smretrieve
#   Returns until the transfer finishes
sub initialize {
    my ($self, %opts) = @_;
    my %sm_args = ( timeout => $opts{timeout} ) if $opts{timeout};

    if ( $self->{type} =~ /dump/ ) {
        # Dump source volume to dumpfile
        # The source volume is created on either the source or destination
        # filer according to suffix of type so for dump_src the source volume
        # is located at the source filer and for dump_dst the source volume is
        # located at the destination filer
        local $filer::FILER_CLI_SAY_TIMEOUT = $opts{timeout} || 900;
        _filer_volume_dump(
            $self->src->Filer->conn,
            $self->src->mount_path,
            $self->device,
            0,
        );
    } elsif ( $self->{type} =~ /restore/ ) {
        # Restore from dumpfile to destination volume
        # The destination volume is created on either the source or destination
        # filer according to suffix of type so for restore_src the destination
        # volume is located at the source filer and for restore_dst the
        # destination volume is located at the destination filer
        local $filer::FILER_CLI_SAY_TIMEOUT = $opts{timeout} || 900;
        _filer_volume_restore(
            $self->dst->Filer->conn,
            $self->device,
            $self->dst->mount_path,
        );
    } elsif ( $self->{type} =~ /smstore/ ) {
        # SnapMirror to tape -- store from source volume to dumpfile
        $self->src->store(device => $self->device, %sm_args);
    } elsif ( $self->{type} =~ /smretrieve/ ) {
        # SnapMirror to tape -- retrieve from dumpfile to destination volume
        $self->dst->retrieve(device => $self->device, %sm_args);
    } else {
        croak "Trying to start an unknown transfer type <$self->{type}>";
    }
}
sub update      { return shift->initialize(@_) }
sub description { return shift->{desc} }
sub src_path    { return shift->{src_path} }
sub dst_path    { return shift->{dst_path} }
sub src         { return shift->{src} }
sub dst         { return shift->{dst} }
sub device      { return shift->{device} }
sub type        { return shift->{type} }
sub destroy     { 1 };

1;
