#
# $Id: //depot/prod/test/Rsierranevada/dp/snapmirror/common/SmScalability.pm#1 $
#
## @summary  SM Scalability class
## @status   Review
## @pod      here
## @author   dl-dpgqa-automation@netapp.com

=head1 NAME

SM_Scalability

=head1 DESCRIPTION

This is a class for working with SM relationships in a scalability context.

=head1 METHODS

=cut

use strict;
use warnings;

package dp::snapmirror::common::SmScalability;

use NOTE;
#XXX Waiting for max transfers library
#use AspenLib;
use Carp qw(croak);
use SayWrap qw(sayEOF);
use Storable qw(dclone);
use nfs::Utils qw(setup_cluster);
use Network::Utils qw(get_ifconfig_info);
use Params::Validate qw(validate SCALAR);
use dp::snapmirror::Utils qw(get_sm_id get_messages hostpath_from_icspath);
use cluster qw(cf_takeover cf_giveback cf_disable wait_state);
use filer qw(filer_options_get filer_options_set);

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
# Parameters which can be defined multiple times
my %max_xfers;
my $parallel_sm = 'dp/snapmirror/common/parallel_sm.thpl';
my $ipfw_en     = 'ipfw.enable';
my $FILL_DATA   = '__SM_FILL_DATA__';
my $FILL_SPARSE = '__SM_FILL_SPARSE__';
my $obj_id = 0;
my $drop_conn_rule_id = 62300;
my $FILER_CONN = $main::FILER_CONN ? $main::FILER_CONN : '#rsh';
my $TIMEOUT    = $main::TIMEOUT    ? $main::TIMEOUT    : 300;
my $CMDTIMEOUT = $main::CMDTIMEOUT ? $main::CMDTIMEOUT : 120;
my $TESTDEF    = "$main::TOP_LOGDIR_UNIX/testdef" if $main::TOP_LOGDIR_UNIX;

# Parameters which can be specified multiple times
my %dup_h = (
    EXEC   => 1,
    ERRINJ => 1,
);

# Parameters which can have modifiers
my %mod_h = (
    EXEC   => 1,
    ERRINJ => 1,
);

# Precedence of main relationship type to use where there is an ambiguity
my @type_list  = qw(VSM QSM RSM SV VC AC
                    DUMP_SRC DUMP_DST RESTORE_SRC RESTORE_DST
                    SMSTORE_SRC SMSTORE_DST SMRETRIEVE_SRC SMRETRIEVE_DST);
my @qtype_list = qw(QSM SV);

# List of valid type modifiers
my %type_h  = map { $_ => 1 } @type_list;
my %qtype_h = map { $_ => 1 } @qtype_list;

# List of parameters to exclude from parallel_sm
my %psm_exclude = (
    %type_h,
    FILER_SRC           => 1,
    FILER_DST           => 1,
    CLIENTFILE          => 1,
    CFO_TIMEOUT         => 1,
    RUN_TIME            => 1,
    RESERVED_XFER_SPACE => 1,
);

# Injection errors that need a relationship from parallel_sm
my %getsm_h = (
    ABORT      => 1,
    BURSTY     => 1,
    AGGR_SPACE => 1,
    VOL_SPACE  => 1,
);

# Temporal modifiers other than transfer progress
my %tmod_h = (
    TEST_BEG => 1,
    TEST_END => 1,
    XFER_END => 1,
);

# Testcase key words
my %test_keywords = (
    TESTCASE        => 1,
    CLONE_TC        => 1,
    ERRINJ          => 1,
    EXEC            => 1,
    EXP_RES         => 1,
    SM_MODE         => 1,
    MULT_FILER_MODE => 1,
);

# Error injection table
my %errinj_table = (
    ABORT => {
        mode => 'one',
        fail => {
            XFER => 1,
        },
    },
    BURSTY => {
        mode => 'one',
        fail => {
            XFER => 0,
        },
    },
    AGGR_SPACE => {
        mode => 'filer_dst',
        fail => {
            TEST_BEG => 1,
            XFER     => 1,
        },
    },
    VOL_SPACE => {
        mode => 'volume',
        fail => {
            TEST_BEG => 1,
            XFER     => 1,
        },
        option => {
            ROOT => {
                mode => 'filer',
                fail => {
                    TEST_BEG => 1,
                    XFER     => 1,
                },
            },
        },
    },
    DROP_CONN => {
        mode => 'filer',
        fail => {
            TEST_BEG => 1,
            XFER     => 1,
        },
    },
    CFO => {
        mode => 'filer',
        fail => {
            XFER => 1,
        },
    },
    PANIC => {
        mode => 'filer',
        fail => {
            XFER => 1,
        },
    },
);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub new($%;);
sub start($;);
sub results($;);
sub _process_testdef($;);
sub _process_value($$;);
sub _process_line_quotes($;);
sub _process_modifiers($$;);
sub _start_on_state($$$;);
sub _process_client_file($$;);
sub _get_test_vars($$;);
sub _allocate_xfers($$;);
sub _process_max($$;);
sub _run_cmd($$$;);
sub _bursty($$;);
sub _get_volume($$$;);
sub _process_results($$$$;);

#---------------------------------------------------------------------
# Methods
#---------------------------------------------------------------------

=head2 ::new

Constructor for the SM scalability class.

=over

=item Arguments

B<testdef> = I<file> | I<string>

I<Required.>
Test definition file name or test definition string.

=item Examples

=back

     # Create object using 'testdef.file'
     $sm = SmScalability->new(testdef => 'testdef.file');

     # Create object using string format
     $test_def = q(
         FILER_SRC=myfiler_src
         FILER_DST=myfiler_dst
         { TESTCASE=sm001 QSM=1 ERRINJ=ABORT }
     );
     $sm = SmScalability->new(testdef => $test_def);

=for html <hr />

=cut

sub new($%;) {
    my $class = shift;
    my %opts = validate(
        @_,
        {
            testdef => { type => SCALAR },
            obj_id  => { type => SCALAR, default => $obj_id++ },
        }
    );

    # Bless the object
    my $self = bless { %opts }, $class;

    $self->{drop_conn_rule_id} = $drop_conn_rule_id + 100*$self->{obj_id};

    # Process the test definition
    $self->_process_testdef;

    return $self;
}

=head2 ::start

Starts the test by creating the SM relationships it needs for the
error injection and the rest are created by calling parallel_sm.thpl.
It injects the errors at the appropriate times and checks for errors.
All testcases in the test definition are executed sequentially.

=over

=item Examples

=back

     # Start test
     $sm->start;

=for html <hr />

=cut

sub start($;) {
    my $self = shift;
    my ($test, $mode, $type, $item, $fc, $sobj, $sm_h, @errors, $cmd);
    my (@param_list, @inj_list, $vars, $filer, $partner, %aggr, $run_psm);
    my (%exp_fail, $rule, $index, %psm_params, $sm_ret, $stop, $msg, $msgs);
    my ($etime, $lctype, $runid);
    my $psm_index = 1;

    foreach $test ( @{$self->{tests}} ) {
        eval {
            # Clear any data from the previous testcase
            @errors = ();
            $sm_h = undef;
            %exp_fail = ();
            @param_list = ();
            %psm_params = ();
            $stop = 0;

            Tharn::logcomment("Start testcase $test->{TESTCASE}");
            # Get all variables for this testcase
            $vars = $test->{_vars};
            $mode = $test->{SM_MODE};
            $type = $test->{_sm_type};
            $lctype = lc($type);

            # Get list of all the items to inject
            @inj_list = map { $test->{$_} ? @{$test->{$_}} : () }
              qw(_beg_list _xfer_list _xfer_end_list _end_list);

            # Setup according to type of error injection items to execute
            foreach $item ( @inj_list ) {
                # Get filer where action will be executed on
                $filer = $item->{where};
                $filer = $test->{_vars}{FILER_SRC} if $filer eq 'SRC';
                $filer = $test->{_vars}{FILER_DST} if $filer eq 'DST';

                if ( $item->{type} eq 'CFO' ) {
                    # Setup cluster pair -- make sure cluster is enabled
                    # and both filers are in the 'UP' state

                    # Setup cluster pair if it has not been setup yet
                    if ( ! $test->{_cfo_partners}{$filer} ) {
                        # Get partner of filer specified in the CFO
                        # injection type
                        Tharn::host($filer);
                        $fc = Tharn::connect($FILER_CONN, '-timeout', $TIMEOUT);
                        $fc->timeout($CMDTIMEOUT);
                        $partner = sayEOF($fc, "cf partner");

                        # Save filer and partner into the partner hash
                        $test->{_cfo_partners}{$filer} = $partner;
                        $test->{_cfo_partners}{$partner} = $filer;

                        # Setup cluster pair
                        Tharn::logcomment("Setup cluster pair " .
                                          "$filer <-> $partner");
                        setup_cluster(
                            filerA  => $filer,
                            filerB  => $partner,
                            timeout => $vars->{CFO_TIMEOUT},
                        );
                    }
                } elsif ( $item->{type} eq 'AGGR_SPACE' ) {
                    # For filling up the aggregate space the volume created in
                    # this aggregate must have the guarantee option set to none
                    $aggr{$item->{where}} = 1;
                } elsif ( $item->{type} eq 'DROP_CONN' ) {
                    $filer = NOTE::Filer->connect(host => $filer);
                    $test->{_filer} = $filer;
                    $test->{_ipfw_opt} = filer_options_get($filer->conn,
                                                           $ipfw_en);
                } elsif ( $item->{type} eq 'EXEC' ) {
                    foreach ( @type_list ) {
                        if ( $item->{$_} && $test->{$_} ) {
                            $test->{$_} -= $item->{$_};
                            $test->{$_} = 0 if $test->{$_} < 0;
                        }
                    }
                }
            }

            # This is the beginning of the test, inject all items for this state
            $self->_start_on_state($test, 'TEST_BEG');

            # Create a list of parameters for parallel_sm
            foreach $item ( grep { ! $psm_exclude{$_} } keys %$vars ) {
                $psm_params{$item} = $vars->{$item};
            }

            # Parameters that names do not map
            $psm_params{TEST}    = $mode;
            $psm_params{AUTO}    = 2;
            $psm_params{TIMEOUT} = 0;

            # Add list of clients to the parameters
            if ( $vars->{CLIENTFILE} ) {
                $psm_params{CLIENT} =
                  join(',', @{$self->{_client_file}{$vars->{CLIENTFILE}}});
            }

            # Convert RUN_TIME to TIMEOUT
            if ( $vars->{RUN_TIME} ) {
                $psm_params{TIMEOUT} = $vars->{RUN_TIME};
            }

            # Check if we need to run parallel_sm
            $run_psm = grep { $test->{_vars}{$_} } @type_list;

            eval {
                if ( $run_psm ) {
                    foreach $item ( @{$test->{_filer_alloc}} ) {
                        Tharn::logcomment("Start parallel_sm: " .
                                          $item->{FILER_SRC} . " -> " .
                                          $item->{FILER_DST});

                        # Get parallel_sm parameters ready for subtest
                        @param_list = map { "$_=$item->{$_}" } keys %$item;
                        push @param_list, map { "$_=$psm_params{$_}" }
                                                keys %psm_params;

                        # Skip if no transfers are defined
                        next unless grep { $item->{$_} } @type_list;

                        $runid = sprintf('%04d_', $psm_index++) .
                          Tharn::host($item->{FILER_SRC})->hostname . '__' .
                          Tharn::host($item->{FILER_DST})->hostname;

                        push @{$test->{_psm_objs}}, Tharn::subtest(
                            $parallel_sm, @param_list,
                            qq(FILER_SRC=$item->{FILER_SRC}),
                            qq(FILER_DST=$item->{FILER_DST}),
                            '-runid', $runid,
                            '-bg', '-ipc', '-onexit', sub {}
                        );

                        Tharn::logcomment("Wait for parallel_sm " .
                                          "to finish setup");
                        $stop = 1;
                        $sm_ret = $test->{_psm_objs}[-1]->ipc->read_obj;
                        foreach ( keys %$sm_ret ) {
                            push @{$sm_h->{$_}}, @{$sm_ret->{$_}};
                        }
                    }

                    if ( $type ) {
                        Tharn::logcomment("Attach $type to monitor");
                        if ( $type eq 'SV' ) {
                            # SnapVault relationship
                            $test->{_sm} = NOTE::SnapVault->attach(
                                src => $sm_h->{sv}[0]{src_path},
                                dst => $sm_h->{sv}[0]{dst_path},
                            );
                        } elsif ( $type eq 'VC' ) {
                            # Volume Copy relationship
                            $test->{_sm} = NOTE::Volume::Copy->attach(
                                src => $sm_h->{vc}[0]{src_path},
                                dst => $sm_h->{vc}[0]{dst_path},
                            );
                        } elsif ( $type eq 'AC' ) {
                            # Aggregate Copy relationship
                            $test->{_sm} = NOTE::Aggregate::Copy->attach(
                                src => $sm_h->{ac}[0]{src_path},
                                dst => $sm_h->{ac}[0]{dst_path},
                            );
                        } else {
                            # SnapMirror relationship
                            $test->{_sm} = NOTE::SnapMirror->attach(
                                srcpath  => hostpath_from_icspath(
                                    $sm_h->{$lctype}[0]{src_path}
                                ),
                                ics_path => $sm_h->{$lctype}[0]{src_path},
                                dstpath  => $sm_h->{$lctype}[0]{dst_path},
                            );
                        }

                        # Save source and destination objects
                        $test->{_src} = $test->{_sm}->src;
                        $test->{_dst} = $test->{_sm}->dst;

                        # Set the tag for source and destination
                        $test->{_src}->tag($sm_h->{$lctype}[0]{src_tag});
                        $test->{_dst}->tag($sm_h->{$lctype}[0]{dst_tag});
                    }

                    # Send parallel_sm the 'go' to start the transfers
                    Tharn::logcomment("Start all transfers");
                    foreach $sobj ( @{$test->{_psm_objs}} ) {
                        $sobj->ipc->write_message('GO') if $sobj->is_alive;
                    }

                    # All subtests have been notified, no need to notify
                    # them again
                    $stop = 0;

                    if ( $type && $type =~ /^[VA]C$/ ) {
                        Tharn::logcomment("Get operation number for $type");
                        $etime = time + 60;
                        do {
                            $msgs = get_messages(
                                subtests => $test->{_psm_objs},
                                timeout  => 5
                            );
                            foreach $msg ( @$msgs ) {
                                get_sm_id(
                                    sm_list => [ $test->{_sm} ],
                                    message => $msg,
                                );
                                last if defined $test->{_sm}->op_num;
                            }
                        } while ( $etime > time &&
                                  ! defined $test->{_sm}->op_num );
                        croak "Could not get operation number for $type"
                          unless defined $test->{_sm}->op_num;
                    }
                }

                if ( $mode eq 'initialize' ) {
                    # This is the initial transfer, inject all items for
                    # this state
                    $self->_start_on_state($test, 'I<xfer>');
                } else {
                    # This is the incremental transfer, inject all items
                    # for this state
                    $self->_start_on_state($test, 'U<xfer>');
                }
            };
            push(@errors, $@) if $@;

            # There has been an error so do not start transfers
            if ( $@ && $stop ) {
                Tharn::logcomment("Stopping test due to errors");
                foreach $sobj ( @{$test->{_psm_objs}} ) {
                    eval { $sobj->ipc->write_message('STOP') }
                      if $sobj->is_alive;
                }
            }

            eval {
                if ( $test->{_psm_objs} ) {
                    Tharn::logcomment("Waiting for parallel_sm to finish");
                    foreach $sobj ( @{$test->{_psm_objs}} ) {
                        $sobj->wait_finish;
                        push @{$test->{_psm_results}}, $sobj->results;
                    }
                }
            };
            push(@errors, "Error getting results from parallel_sm: $@") if $@;

            # This is the end of the transfer, inject all items for this state
            eval { $self->_start_on_state($test, 'XFER_END') unless @errors };
            push(@errors, $@) if $@;

            Tharn::logcomment("Wait for all EXEC's to finish")
              if $test->{_exec_sobjs};
            foreach $sobj ( @{$test->{_exec_sobjs}} ) {
                eval {
                    $sobj->wait_finish;
                    push @{$test->{_exec_results}}, [ $sobj->results ];
                };
                push(@errors, "Error getting results from exec: $@") if $@;
            }

            # This is the end of the test, inject all items for this state
            eval { $self->_start_on_state($test, 'TEST_END') unless @errors };
            push(@errors, $@) if $@;

            $item = 0;
            foreach $filer ( keys %{$test->{_cfo_partners}} ) {
                next if ($item++)%2;
                eval {
                    $partner = $test->{_cfo_partners}{$filer};
                    Tharn::logcomment("Reset cluster pair $filer <-> $partner");
                    setup_cluster(
                        filerA  => $filer,
                        filerB  => $partner,
                        timeout => $vars->{CFO_TIMEOUT},
                    );
                    cf_disable($filer) or croak $@;
                };
                push(@errors, $@) if $@;
            }

            if ( $test->{_ipfw_opt} && $test->{_drop_conn} ) {
                eval {
                    filer_options_set($test->{_filer}->conn, $ipfw_en,
                                      $test->{_ipfw_opt});
                };
                push(@errors, $@) if $@;

                # Remove ipfw rules
                foreach $index ( 1..$test->{_drop_conn} ) {
                    $rule = $self->{drop_conn_rule_id} + $index - 1;
                    eval {
                        $cmd = "priv set -q test; ipfw delete $rule";
                        Tharn::say($test->{_filer}->conn, $cmd);
                    };
                    push(@errors, $@) if $@;
                }
            }
        };
        push(@errors, $@) if $@;

        $test->{_results} = [ @errors ] if @errors;

        _process_results($test, \@inj_list, $sm_h, $type);
    }

    return;
}

=head2 ::results

Returns an array of hashes containing the results from each
of the test cases. Each result hash contains the testcase
identifier and the result object for that testcase.

=over

=item Examples

=back

     # Get results
     @results = $sm->results;

=for html <hr />

=cut

sub results($;) {
    my $self = shift;
    my @results;

    foreach my $test ( @{$self->{tests}} ) {
        my %result_h;
        $result_h{testcase} = $test->{TESTCASE};
        $result_h{result} = $test->{_test_result};
        push @results, \%result_h;
    }

    return @results;
}

#---------------------------------------------------------------------
# Private functions
#---------------------------------------------------------------------

# Function to process the test definition file or string
sub _process_testdef($;) {
    my $self = shift;
    my ($testdata, $line, $block, $test_line, $test, $key, $value, %tc_h);
    my (@inj_list, $info, $item);
    my $msg = '';

    if ( $self->{testdef} =~ /\{/ ) {
        # Test definition is a string
        $testdata = $self->{testdef};
    } else {
        # Open definition file and read it
        open(FH, "<$self->{testdef}")
          or croak "Could not open test definition file $self->{testdef}: $!";
        $testdata = join('', <FH>);
        close(FH);
    }

    if ( $TESTDEF ) {
        # Save the test definition file in the log directory
        open(FH, ">$TESTDEF")
          or croak "Could not open test definition save file $TESTDEF: $!";
        print FH $testdata;
        close(FH);
    }

    # Remove leading spaces
    $testdata =~ s/^\s*//gm;
    # Remove trailing spaces
    $testdata =~ s/\s*$//gm;
    # Remove comments
    $testdata =~ s/^#.*//gm;
    # Remove empty lines
    $testdata =~ s/\n+/\n/g;

    # Process each line at a time
    $block = 0;
    $test_line = '';
    foreach ( split /\n/, $testdata ) {
        $line = _process_line_quotes($_);
        if ( $line =~ /^\{/ ) {
            # Mark the beginning of a testcase block
            croak "The following testcase is not properly closed:\n$test_line"
              if $block;
            $block = 1;
        }
        if ( $block ) {
            # Inside a testcase block, keep appending each line
            # so each testcase block is processed as a whole
            $test_line .= " $line";
        } else {
            # Outside a block -- global variable
            $line =~ /(\S+)\s*=\s*(.*)/;
            if ( defined $1 ) {
                $self->{vars}{$1} = $self->_process_value($2);
            }
            $test_line = '';
        }

        if ( $line =~ /\}$/ ) {
            # This is the end of the testcase
            # Process whole testcase
            $block = 0;

            # Remove curly braces from testcase definition
            $test_line =~ s/^\s*{\s*//;
            $test_line =~ s/\s*}$//;

            # Save all fields of testcase into a hash
            $test = {};
            while ( $test_line =~ /(\S+)\s*=\s*(\S+)/g ) {
                $key   = $1;
                $value = $2;
                if ( $dup_h{$key} ) {
                    # Parameter can be defined multiple times so push
                    # each value into an array
                    if ( $mod_h{$key} ) {
                        # Parameter can have modifiers
                        push @{$test->{$key}}, [ split /:/, $value ];
                    } else {
                        # Parameter with no modifiers
                        push @{$test->{$key}}, $value;
                    }
                } else {
                    # Parameter can only be defined once
                    croak "Parameter '$key' can only be defined once"
                      if defined $test->{$key};
                    if ( $mod_h{$key} ) {
                        # Parameter can have modifiers
                        push @{$test->{$key}}, split /:/, $value;
                    } else {
                        # Parameter with no modifiers
                        $test->{$key} = $value;
                    }
                }
            }

            # Process all fields in testcase: quotes and variable interpolation
            $self->{_vars} = $self->_get_test_vars($test);
            foreach $key ( keys %$test ) {
                if ( ref $test->{$key} eq 'ARRAY' ) {
                    # Process parameters with modifiers/multiple instances
                    foreach $value ( @{$test->{$key}} ) {
                        if ( ref $value eq 'ARRAY' ) {
                            # Process all modifiers for each instance
                            # of the parameter
                            $value = [
                                map { $self->_process_value($_) } @$value
                            ];
                        } else {
                            # Process each modifier
                            $value = $self->_process_value($value);
                        }
                    }
                } else {
                    # Process parameter: single instance and no modifiers
                    $test->{$key} = $self->_process_value($test->{$key});
                }
            }
            $test->{_vars} = $self->{_vars};
            delete $self->{_vars};

            # Check for required parameters
            croak "Parameter 'TESTCASE' must be defined"
              unless defined $test->{TESTCASE};

            if ( ! grep { $test->{$_} } @type_list ) {
                croak "At least one of the following parameters must be " .
                      "specified: " . join(', ', @type_list);
            }

            # Save testcase and get ready for next
            $self->_process_modifiers($test);
            push @{$self->{tests}}, $test;
            $test_line = '';
        }
    }

    if ( ! $self->{tests} ) {
        croak "There must be at least one testcase in the test definition file";
    }

    if ( $block ) {
        croak "The following testcase is not properly closed:\n$test_line"
    }

    # Process the CLIENTFILE global variable
    if ( $self->{vars}{CLIENTFILE} ) {
        $self->_process_client_file($self->{vars}{CLIENTFILE});
    }

    # Get list of all testcases -- use first occurrence
    %tc_h = map { $_->{TESTCASE} => $_ } reverse @{$self->{tests}};

    # Defaults for global variables
    $self->{vars}{RICHDATA}            ||= 0;
    $self->{vars}{RESERVED_XFER_SPACE} ||= '100M';
    $self->{vars}{CFO_TIMEOUT}         ||= '300';

    # Process the CLONE_TC option on all testcases
    foreach $test ( @{$self->{tests}} ) {
        $key = $test->{CLONE_TC};
        if ( $key ) {
            if ( ! exists $tc_h{$key} ) {
                $msg .= "Testcase '$test->{TESTCASE}' references an " .
                        "undefined testcase '$key' on CLONE_TC\n";
            } else {
                # Clone testcase and overwrite parameters
                $test = { %{dclone($tc_h{$key})}, %$test };
            }
            delete $test->{CLONE_TC};
        }

        # Get test variables
        $test->{_vars} = eval { $self->_get_test_vars($test) };
        $msg .= $@ if $@;

        # Defaults for testcase variables
        $test->{SM_MODE} ||= $self->{vars}{SM_MODE} ?
                             $self->{vars}{SM_MODE} : 'update';
        $test->{MULT_FILER_MODE} ||= $self->{vars}{MULT_FILER_MODE} ?
                                     $self->{vars}{MULT_FILER_MODE} : 'fan';

        # Allocate SM transfers for multiple filers
        eval { $self->_allocate_xfers($test) };
        $msg .= $@ if $@;

        # Process the CLIENTFILE test variable
        if ( $test->{CLIENTFILE} ) {
            eval { $self->_process_client_file($test->{CLIENTFILE}) };
            $msg .= $@ if $@;
        }

        # Get list of all the items to inject
        @inj_list = map { $test->{$_} ? @{$test->{$_}} : () }
          qw(_beg_list _xfer_list _xfer_end_list _end_list);

        if ( @inj_list ) {
            # Get the type of any item to be injected at transfer
            $info = (grep { ! $tmod_h{$_->{tmod}} } @inj_list)[0];
            if ( $info ) {
                # If type of transfer is ambiguous then decide according
                # to precedence. The precedence order is given by @type_list
                $test->{_sm_type} = $info->{mode} ||
                                    (grep {$test->{$_}} @type_list)[0];
                if ( ! $test->{$test->{_sm_type}} ) {
                    $msg .= "$test->{TESTCASE}: " .
                            "There must be at least one transfer for error " .
                            "injection target type of <$test->{_sm_type}>\n";
                }
            }
        }

        # Validate testcase
        foreach $item ( @inj_list ) {
            if ( ! $errinj_table{$item->{type}}
              && $item->{type} ne 'FILER_CMD' ) {
                $msg .= "$test->{TESTCASE}: Error injection $item->{type} " .
                        "is not supported\n";
            } elsif ( ($item->{type} eq 'BURSTY' && $test->{_sm_type} eq 'SV')
              || ($item->{type} eq 'AGGR_SPACE' && $test->{_sm_type} eq 'AC')
              || $test->{_sm_type} =~ /^(DUMP|RESTORE)_(SRC|DST)$/
              || $test->{_sm_type} =~ /^SM(STORE|RETRIEVE)_(SRC|DST)$/
              || ($item->{type} eq 'VOL_SPACE'
                  && !($item->{option} && $item->{option} eq 'ROOT')
                  && ($test->{_sm_type} =~ /^([VR]SM|[VA]C)$/)) ) {
                $msg .= "$test->{TESTCASE}: Error injection $item->{type} " .
                        "is not supported for $test->{_sm_type}\n";
            }
        }
    }

    # Report any errors on testcase
    croak "\n$msg" if $msg;

    return;
}

# Function to process the input string and interpolate variables if necessary
sub _process_value($$;) {
    my ($self, $value_in) = @_;
    my ($value, $var, $key, %var_list);
    my $vars = $self->{_vars} || $self->{vars};

    $value = '';
    # Process each sub-string separately
    foreach ( split /\./, $value_in ) {
        if ( s/^'// ) {
            # Single quotes
            s/'\s*$//;
        } else {
            # Double quotes or no quotes
            s/^"//;
            s/"\s*$//;

            # Expand all variables in string
            # Use a hash to get only unique variables
            %var_list = map { $_ => 1 } /\$({?\w+}?)/g;
            foreach $var ( keys %var_list ) {
                ($key) = ($var =~ /{?(\w+)}?/);
                croak "Variable '$key' is not defined"
                  unless defined $vars->{$key};
                s/\$$var\b/$vars->{$key}/g;
            }

            if ( /\\/ ) {
                # Interpolate string
                $_ = eval qq("$_");
                croak $@ if $@;
            }
        }

        # Restore special characters that were renamed in order not
        # to process them as operators
        # Space and Tab characters to make it easier to process key=value
        s/--<SP>--/ /g;
        s/--<TB>--/\t/g;
        # Backslash is the escape character
        s/--<BS>--/\\/g;
        # Quotes in the strings to differentiate them between the quote
        # operators to delimit strings
        s/--<SQ>--/'/g;
        s/--<DQ>--/"/g;
        # Escaped dollar sign to differentiate with variable interpolation
        s/--<DS>--/\$/g;
        # Period to differentiate with concatenate operator
        s/--<PD>--/\./g;
        # Colon to differentiate with field separator
        s/--<CL>--/\:/g;
        # Number sign to differentiate with in-line comments
        s/--<NS>--/\#/g;
        # Equals sign to make it easier to process key=value
        s/--<EQ>--/\=/g;
        $value .= $_;
    }

    return $value;
}

# Function to process all quotes in the input string, also convert any
# special characters into markers (to convert them back) to make it easier
# to process each testcase
sub _process_line_quotes($;) {
    my $line = shift;
    my (@chars, $quote);

    # Rename escaped characters so they do not get processed
    $line =~ s/\\\\/--<BS>--/g;
    $line =~ s/\\\'/--<SQ>--/g;
    $line =~ s/\\\"/--<DQ>--/g;
    $line =~ s/\\\$/--<DS>--/g;

    if ( $line =~ /['"]/ ) {
        # Process single and double quote strings
        # one character at a time
        @chars = split(//, $line);
        $quote = '';
        foreach ( @chars ) {
            if ( /['"]/ ) {
                if ( ! $quote ) {
                    # Mark the start of quote
                    $quote = $_;
                } elsif ( $quote eq $_ ) {
                    # Mark the end of quote
                    $quote = '';
                }
            }

            if ( $quote ) {
                # Change all special characters inside quotes
                s/ /--<SP>--/;
                s/\t/--<TB>--/;
                s/\./--<PD>--/;
                s/\:/--<CL>--/;
                s/\#/--<NS>--/;
                s/\=/--<EQ>--/;
            }
        }
        # Recreate testcase line
        $line = join('', @chars);
    }

    # Remove in-line comments
    $line =~ s/\s*#.*//;
    # Remove spacing around concatenate operator '.'
    $line =~ s/\s*\.\s*/\./g;
    # Remove spacing around modifier operator ':'
    $line =~ s/\s*\:\s*/\:/g;

    return $line;
}

# Function to process modifiers
#
# <tmod> := TEST_BEG|TEST_END|XFER_END|<xfer>
# <xfer> := Until the number of bytes have been transferred xxx[k|m|g|t]
#
# ERRINJ:
# ABORT [: QSM|VSM|RSM|SV|VC|AC] [: HARD] [: SRC|DST] [: <xfer>]
# BURSTY [: QSM|VSM|RSM|SV|VC|AC] [: SRC|DST]
# AGGR_SPACE [: SRC|DST] [: <tmod>]
# VOL_SPACE [: QSM|SV] [:ROOT] [: SRC|DST] [: <xfer>]
# DROP_CONN : 1|all [: SRC|DST] [: <xfer>]
# CFO [: TAKEOVER|GIVEBACK] [: SRC|DST|<filer_var>] [: <tmod>]
# PANIC [: SRC|DST] [: <tmod>]
# FILER_CMD : <command> [: EXP_RES=<regex>] [: SRC|DST] [: <tmod>]
#
# EXEC:
# <exec_name> [: +(QSM|VSM|RSM|SV|VC|AC)=<number>] [: <param>=<value>] [: <tmod>]
sub _process_modifiers($$;) {
    my ($self, $test) = @_;
    my ($info, @list, $item, @run_list);
    my $vars = $test->{_vars};

    if ( $test->{ERRINJ} ) {
        # Error injection
        foreach $item ( @{$test->{ERRINJ}} ) {
            $info = {};
            # Copy list of modifiers for item
            @list = @$item;
            # Type is required
            $info->{type} = shift @list;

            if ( $info->{type} eq 'FILER_CMD' ) {
                # Get the filer command to run and expected output of command
                $info->{cmd} = shift @list;
                $info->{out} = '';
                if ( @list && $list[0] =~ /^EXP_RES/ ) {
                    (shift @list) =~ /EXP_RES\s*=\s*(.*)/;
                    $info->{out} = $1;
                }
            } elsif ( $info->{type} eq 'DROP_CONN' ) {
                # Get value
                $info->{option} = shift @list;
            } elsif ( @list && $info->{type} =~ /^(ABORT|BURSTY)$/
              && $type_h{$list[0]} ) {
                $info->{mode} = shift @list;
            } elsif ( $info->{type} eq 'VOL_SPACE' ) {
                # Get type of SM
                if ( @list && $qtype_h{$list[0]} ) {
                    $info->{mode} = shift @list;
                }

                # Set a default value according to precedence
                $info->{mode} ||= (grep {$test->{$_}} @qtype_list)[0];

                # Get ROOT option
                if ( @list && $list[0] eq 'ROOT' ) {
                    $info->{option} = shift @list;
                }
            } elsif ( $info->{type} eq 'CFO' ) {
                if ( @list && $list[0] =~ /^(TAKEOVER|GIVEBACK)$/ ) {
                    $info->{action} = shift @list;
                } else {
                    $info->{action} = 'TAKEOVER';
                }
            }

            if ( $info->{type} eq 'ABORT' ) {
                # Get abort flag
                if ( @list && $list[0] eq 'HARD' ) {
                    $info->{hard} = 1;
                    shift @list;
                } else {
                    $info->{hard} = 0;
                }
            }

            # Get filer modifier
            if ( @list && $list[0] =~ /^(SRC|DST)$/ ) {
                $info->{where} = shift @list;
            } else {
                if ( @list && %$vars && grep { $_ eq $list[0] } keys %$vars ) {
                    $info->{where} = $vars->{shift @list};
                } else {
                    $info->{where} = 'DST';
                }
            }

            # Get temporal modifier
            $info->{tmod} = shift @list;

            if ( !$info->{tmod} ) {
                # Set default temporal modifier
                if ( $info->{type} =~ /^(ABORT|DROP_CONN|VOL_SPACE)$/ ) {
                    $info->{tmod} = '10M';
                } else {
                    $info->{tmod} = 'TEST_BEG';
                }
            }

            # Burst on transfer only
            if ( $info->{type} eq 'BURSTY' ) {
                $info->{tmod} = '1M';
            }

            push @run_list, $info;
        }
    }

    if ( $test->{EXEC} ) {
        foreach $item ( @{$test->{EXEC}} ) {
            @list = @$item;
            $info = {
                type => 'EXEC',
                cmd  => shift @list,
            };

            # Get number of SM's being used by exec
            while ( @list && grep { $list[0] =~ /^\+$_\s*=/ } @type_list ) {
                (shift @list) =~ /\+(\S+)\s*=\s*(\S+)/;
                $info->{$1} = $2;
            }

            # Get parameters for exec
            while ( @list && $list[0] =~ /=/ ) {
                (shift @list) =~ /(\S+)\s*=\s*(\S+)/;
                $info->{params}{$1} = $2;
            }

            # Get temporal modifier
            $info->{tmod} = shift @list || 'TEST_BEG',

            push @run_list, $info;
        }
    }

    # Create a list for each of the valid states
    foreach $item ( @run_list ) {
        if ( $item->{tmod} eq 'TEST_BEG' ) {
            push @{$test->{_beg_list}}, $item;
        } elsif ( $item->{tmod} eq 'TEST_END' ) {
            push @{$test->{_end_list}}, $item;
        } elsif ( $item->{tmod} eq 'XFER_END' ) {
            push @{$test->{_xfer_end_list}}, $item;
        } else {
            $item->{_tmod} = NOTE::units($item->{tmod});
            push @{$test->{_xfer_list}}, $item;
        }
    }

    # Sort according to temporal modifier
    if ( defined $test->{_xfer_list} ) {
        @{$test->{_xfer_list}} = sort { $a->{_tmod} <=> $b->{_tmod} }
                                        @{$test->{_xfer_list}};
    }

    return;
}

# Function to read the clients from the clientfile
sub _process_client_file($$;) {
    my ($self, $file) = @_;

    # Check if the clientfile has already been read
    return if $self->{_client_file}{$file};

    # Open the clientfile and read the clients into an array
    open(FH, "<$file") or croak "Could not open client file $file: $!";
    while ( <FH> ) {
        chomp;
        push @{$self->{_client_file}{$file}}, $_;
    }
    close(FH);

    return;
}

# Function to exec item at proper state
sub _start_on_state($$$;) {
    my ($self, $test, $state) = @_;
    my ($item, $progress, $list, $idx, $status);

    if ( $state eq 'TEST_BEG' ) {
        foreach $item ( @{$test->{_beg_list}} ) {
            $self->_run_cmd($test, $item);
        }
    } elsif ( $state eq 'TEST_END' ) {
        foreach $item ( @{$test->{_end_list}} ) {
            $self->_run_cmd($test, $item);
        }
    } elsif ( $state eq 'XFER_END' ) {
        foreach $item ( @{$test->{_xfer_end_list}} ) {
            $self->_run_cmd($test, $item);
        }
    } elsif ( $state =~ /<xfer>/ ) {
        if ( ! defined $test->{_xfer_list}
          || ($test->{SM_MODE} eq 'initialize' && $state !~ /^I/)
          || ($test->{SM_MODE} eq 'update' && $state !~ /^U/) ) {
            return;
        }

        $idx = 0;
        $list = $test->{_xfer_list};
        while ( $idx < @$list ) {
            # Get status
            $status = $test->{_sm}->status;
            if ( $status->{'current-transfer-error'} ) {
                croak $status->{'current-transfer-error'};
            }
            if ( $status->{status} ne 'transferring'
              && $status->{'last-transfer-type'} eq $test->{SM_MODE} ) {
                croak "Relationship is not transferring: " . $status->{status};
            }

            # Get progress and process all items that need to be run
            $progress = $status->{'transfer-progress'} || 0;
            while ( $idx < @$list
              && $progress >= ($item = $list->[$idx])->{_tmod} ) {
                $idx++;
                $self->_run_cmd($test, $item);
            }
            sleep 1;
        }
    }

    return;
}

# Function to get variables for given test
# Returns a hash reference of all variables
# Global variables are overwritten by test variables
sub _get_test_vars($$;) {
    my ($self, $test) = @_;
    my (%ret, $key);

    # Get global variables
    %ret = %{$self->{vars}} if $self->{vars};

    # Overwrite globals with test variables
    foreach $key ( keys %{$test} ) {
        if ( $key !~ /^_/ && ! $test_keywords{$key} ) {
            $ret{$key} = $test->{$key};
        } elsif ( $test_keywords{$key} && exists $ret{$key} ) {
            delete $ret{$key};
        }
    }

    # Get multiple filers for source
    if ( ! $test->{_filer_src} ) {
        croak "At least one source filer must be specified"
          unless $ret{FILER_SRC};
        push @{$test->{_filer_src}}, split(/\s*,\s*/, $ret{FILER_SRC});
    }

    # Get multiple filers for destination
    if ( ! $test->{_filer_dst} ) {
        croak "At least one destination filer must be specified"
          unless $ret{FILER_DST};
        push @{$test->{_filer_dst}}, split(/\s*,\s*/, $ret{FILER_DST});
    }

    # Set FILER_SRC to first on the list
    $ret{FILER_SRC} = $test->{_filer_src}[0];

    # Set FILER_DST to first on the list
    $ret{FILER_DST} = $test->{_filer_dst}[0];

    return \%ret;
}

# Function to allocate transfers based on multiple filers
sub _allocate_xfers($$;) {
    my ($self, $test) = @_;
    my ($filer_a, $filer_b, $min_set, $max_set, $fname_a, $fname_b, $n);
    my (%sms, %remain, %xfers, %rcount, $max, $type, $item, $index, $set);
    my $nsrc = @{$test->{_filer_src}};
    my $ndst = @{$test->{_filer_dst}};

    # Re-calculate file allocation in case this is a clone
    delete $test->{_filer_xfers};
    delete $test->{_filer_alloc};

    # Process the 'MAX' token
    $self->_process_max($test);

    # Get the minimum number of filers between source and destination
    if ( $nsrc < $ndst ) {
        $min_set = $test->{_filer_src};
        $max_set = $test->{_filer_dst};
        $fname_a = 'FILER_SRC';
        $fname_b = 'FILER_DST';
        $max     = $ndst;
    } else {
        $min_set = $test->{_filer_dst};
        $max_set = $test->{_filer_src};
        $fname_a = 'FILER_DST';
        $fname_b = 'FILER_SRC';
        $max     = $nsrc;
    }

    # Get number of transfers for each type
    foreach $type ( grep { $test->{_vars}{$_} } @type_list ) {
        $xfers{$type} = int($test->{_vars}{$type} / $max);
        $remain{$type} = $test->{_vars}{$type} % $max;
    }

    # Allocate transfers on all the filers
    $index = 0;
    foreach $filer_a ( @$min_set ) {
        %rcount = %remain;
        $n = 0;
        foreach $filer_b ( @$max_set ) {
            # Get total number of transfers to/from max subset filers
            %sms = %xfers;

            if ( $index == $n ) {
                # Allocate any left over transfers
                $set = 0;
                foreach $type ( keys %rcount ) {
                    if ( $rcount{$type}-- > 0 ) {
                        $sms{$type}++;
                        $set = 1;
                    }
                }
                if ( $set ) {
                    $index++;
                    $index = 0 if $index == $max;
                }
            }

            # Allocate transfers for this filer pair
            push @{$test->{_filer_alloc}}, {
                $fname_a => $filer_a,
                $fname_b => $filer_b,
                %sms,
            };
            $n++;
        }
    }

    # Create a hash containing the number of source and destination
    # transfers on a per filer basis
    foreach $item ( @{$test->{_filer_alloc}} ) {
        foreach $type ( keys %xfers ) {
            $test->{_filer_xfers}{$item->{FILER_SRC}}{src}{$type} +=
              $item->{$type} if $type !~ /_DST$/;
            $test->{_filer_xfers}{$item->{FILER_DST}}{dst}{$type} +=
              $item->{$type} if $type !~ /_SRC$/;
        }
    }

    return;
}

# Function to process the 'MAX' token found on the SM type parameters
sub _process_max($$;) {
    my ($self, $test) = @_;
    my ($sm_type, $value, $filer, @list, %params, $model, $conn, $fh, $out);
    my ($max_src, $max_dst, $sm_app, $max);
    my $vars = $test->{_vars};

    if ( ! (@list = grep { $test->{$_} && $test->{$_} =~ /MAX/ } @type_list) ) {
        # Token does not exist
        return;
    }
    croak "XXX MAX is not yet supported: waiting for max transfers library";

    # Get max transfers for both source and destination filers
    foreach $filer ( $vars->{FILER_SRC}, $vars->{FILER_DST} ) {
        if ( ! $max_xfers{$filer} ) {
            # Get the model from the hostspec
            $fh = Tharn::host($filer);
            %params = $fh->param_fields;
            $model = $params{MODEL};

            if ( ! $model ) {
                # Get the model by querying the filer
                $conn = Tharn::connect($FILER_CONN, '-timeout', $TIMEOUT);
                $conn->timeout($CMDTIMEOUT);
                $out = sayEOF($conn, 'sysconfig -v 0');
                ($model) = ($out =~ /^\s+Model Name:\s+(\S+)/m);
            }
            die "Could not determine model for filer $filer" unless $model;

            $max_xfers{$filer}{model} = $model;
        }
    }

    # Get max transfers for both source and destination filers
    foreach $sm_type ( @list ) {
        $sm_app = $sm_type . '_SRC';
        if ( ! $max_xfers{$vars->{FILER_SRC}}{$sm_app} ) {
            # Model of source filer
            $model = $max_xfers{$vars->{FILER_SRC}}{model};
            # Get streams available of type on the source
#XXX Waiting for max transfers library
#            $max_xfers{$vars->{FILER_SRC}}{$sm_app} =
#              AspenLib::get_filer_streams($model, 0, $sm_app);
        }

        # Maximum for source
        $max_src = $max_xfers{$vars->{FILER_SRC}}{$sm_app};

        $sm_app = $sm_type . '_DST';
        if ( ! $max_xfers{$vars->{FILER_DST}}{$sm_app} ) {
            # Model of destination filer
            $model = $max_xfers{$vars->{FILER_DST}}{model};
            # Get streams available of type on the destination
#XXX Waiting for max transfers library
#            $max_xfers{$vars->{FILER_DST}}{$sm_app} =
#              AspenLib::get_filer_streams($model, 0, $sm_app);
        }

        # Maximum for destination
        $max_dst = $max_xfers{$vars->{FILER_DST}}{$sm_app};

        # The maximum number of transfers between source and destination
        # is the minimum of both maximum streams between them
        $max = $max_dst < $max_src ? $max_dst : $max_src;

        # Evaluate value containing MAX token
        $value = $test->{$sm_type};
        $value =~ s/MAX/$max/g;
        $value = eval $value;
        croak "Could not parse $sm_type value <$test->{$sm_type}> " .
              "for testcase $test->{TESTCASE}" if $@;
        $test->{$sm_type} = int($value);
    }
}

# Function to run the appropriate ERRINJ or EXEC
sub _run_cmd($$$;) {
    my ($self, $test, $item) = @_;
    my ($avail, $filer, $ofiler, $aggr, $vol, $out, $delta, $status, @params);
    my ($mtpoint, $size, $fill_file, $cmd, $reserved_space, $timeout, $err);
    my (@ip_a, @ip_b, $dev, $ip_addr_a, $ip_addr_b, $rule, $filer_list, $conn);
    my $vars = $test->{_vars};

    if ( $item->{type} eq 'EXEC' ) {
        $vars = { %$vars, %{$item->{params}} };
        Tharn::logcomment("Running <$item->{cmd}>");
        @params = map { "$_=$vars->{$_}" } keys %$vars;
        push @{$test->{_exec_sobjs}}, Tharn::subtest(
            $item->{cmd}, @params, '-bg', '-onexit', sub {}
        );
    } else {
        Tharn::logcomment("Error injection: <$item->{type}>");
        if ( $getsm_h{$item->{type}} ) {
            ($aggr, $vol) = $self->_get_volume($test, $item);
        }
        $filer = $item->{where};
        if ( $filer eq 'SRC' ) {
            $filer      = $vars->{FILER_SRC};
            $filer_list = $test->{_filer_dst};
        } elsif ( $filer eq 'DST' ) {
            $filer      = $vars->{FILER_DST};
            $filer_list = $test->{_filer_src};
        }

        # Save filer where error will be injected
        $item->{_filer} = $filer;

        if ( $item->{type} eq 'ABORT' ) {
            # Abort the transfer
            $test->{_sm}->abort(hard => $item->{hard});
        } elsif ( $item->{type} eq 'CFO' ) {
            # Cluster takeover/giveback
            $timeout = $vars->{CFO_TIMEOUT};
            if ( $item->{action} eq 'GIVEBACK' ) {
                cf_giveback($filer, partner => $test->{_cfo_partners}{$filer})
                  or croak $@;
                wait_state($filer, 'UP', $timeout) or croak $@;
                wait_state($test->{_cfo_partners}{$filer}, 'UP', $timeout)
                  or croak $@;
            } else {
                cf_takeover($filer) or croak $@;
                wait_state($filer, 'TAKEOVER', $timeout) or croak $@;
            }
        } elsif ( $item->{type} eq 'FILER_CMD' ) {
            # Run filer command
            $filer = NOTE::Filer->connect(host => $filer);
            $out = sayEOF($filer->conn, $item->{cmd});
            if ( $item->{out} ne '' && $out !~ qr/$item->{out}/ ) {
                croak "FILER_CMD output does not match expected results";
            }
        } elsif ( $item->{type} eq 'AGGR_SPACE' ) {
            # Fill up aggregate
            # Get reserved space in KB
            $reserved_space = NOTE::units($vars->{RESERVED_XFER_SPACE});
            $size = int(2 * $reserved_space / 1024);
            $size = 100 if $size < 100;
            $item->{_aggr} = $aggr->name;

            # Create a volume for precise control on space
            $test->{_aggr_fill_0} = NOTE::Volume->create(
                Filer => $aggr->Filer,
                tag   => 'aggr_fill_0',
                type  => 'flex',
                flex  => { size      => $size . 'm',
                           Aggregate => $aggr },
                recycle => 0,
                guarantee => 'none',
            );

            # Create volume using up most of the space in the aggregate
            # Leave 100M to avoid race condition during transfers
            $avail = $aggr->avail - $reserved_space;
            if ( $avail > 51200 ) {
                $test->{_aggr_fill_1} = NOTE::Volume->create(
                    Filer => $aggr->Filer,
                    tag   => 'aggr_fill_1',
                    type  => 'flex',
                    flex  => { size      => $avail,
                               Aggregate => $aggr },
                    recycle => 0,
                );
            }

            # Fill up the aggregate using the non-guarantee precision volume
            $test->{_aggr_fill_0}->fill(percent => 100, proto => 'local');
        } elsif ( $item->{type} eq 'VOL_SPACE' ) {
            # Fill up volume
            # Get reserved space in KB
            $reserved_space = NOTE::units($vars->{RESERVED_XFER_SPACE});
            # Get size of unsparse file to create
            $size = int(($vol->avail - $reserved_space) / 1024);
            $size = 10 if $size < 10;
            $item->{_vol} = $vol->name;
            $mtpoint = $vol->mount;

            eval {
                # Create unsparse file on filer for speed
                $vol->fillfile(name => $FILL_SPARSE, proto => 'local');
                $vol->fill(max => $size, proto => 'local', type => 'sparse');
                sleep 5;

                $delta = int(8 * $size / 10240);
                $delta = 1 if $delta < 1;

                # Wait until unsparse file gets created
                $size = 0;
                $fill_file = "$mtpoint/$FILL_SPARSE";
                do {
                    $out = $size;
                    sleep $delta;
                    $size = -e $fill_file ? -s $fill_file : 0;
                } while ( $size != $out );

                # Create regular file on filer for accuracy
                $vol->fillfile(name => $FILL_DATA, proto => 'local');
                $vol->fill(percent => 100, proto => 'local', type => 'data');
            };
            $err = $@;
            eval { $vol->umount };
            die $err if $err;
        } elsif ( $item->{type} eq 'PANIC' ) {
            $filer = NOTE::Filer->connect(host => $filer);
            eval { $filer->panic };
            $err = $@;
            $filer->transit_to('UP');
            die $err if $err;
        } elsif ( $item->{type} eq 'BURSTY' ) {
            # Burst the aggregate transfer if it is an aggrcopy
            $vol = $aggr if $test->{_sm}->isa('NOTE::Aggregate::Copy');
            Tharn::subtest( sub { _bursty($test, $vol) },
                            '-bg', '-onexit', sub {} );
        } elsif ( $item->{type} eq 'DROP_CONN' ) {
            $filer = NOTE::Filer->connect(host => $filer);

            if ( $item->{option} eq 'all' ) {
                # Get all ip addresses and aliases of main filer
                $out = get_ifconfig_info($filer->conn);
                foreach ( keys %$out ) {
                    $dev = $out->{$_};
                    next if $dev->{ip_addr} eq '127.0.0.1';
                    push @ip_a, $dev->{ip_addr} if $dev->{ip_addr};
                    push @ip_a, keys %{$dev->{alias}} if $dev->{alias};
                }
            } else {
                # Just drop the default connection
                @ip_a = ( $filer->hostrec->default_ip );
            }

            foreach $ofiler ( @$filer_list ) {
                Tharn::host($ofiler);
                $conn = Tharn::connect($FILER_CONN, '-timeout', $TIMEOUT);
                $conn->timeout($CMDTIMEOUT);

                # Get all ip addresses and aliases of secondary filer
                $out = get_ifconfig_info($conn);
                foreach ( keys %$out ) {
                    $dev = $out->{$_};
                    next if $dev->{ip_addr} eq '127.0.0.1';
                    push @ip_b, $dev->{ip_addr} if $dev->{ip_addr};
                    push @ip_b, keys %{$dev->{alias}} if $dev->{alias};
                }
            }

            $test->{_drop_conn} = 0;
            foreach $ip_addr_a ( @ip_a ) {
                foreach $ip_addr_b ( @ip_b ) {
                    $rule = $self->{drop_conn_rule_id} + $test->{_drop_conn};
                    $test->{_drop_conn}++;
                    $cmd = "priv set -q test; ipfw add $rule deny ip from " .
                           "$ip_addr_b to $ip_addr_a";
                    Tharn::say($filer->conn, $cmd);
                }
            }

            # Enable rules
            filer_options_set($filer->conn, $ipfw_en, 'on');
        }

        # Error has been injected
        $item->{_done} = 1;
    }

    return;
}

# Function to speed up and slow down the transfer repeatedly
sub _bursty($$;) {
    my ($test, $vol) = @_;
    my (@tvalues, $status, $function, $throttle);
    my $delta = 10;

    if ( $test->{_sm}->isa('NOTE::Volume::Copy') ) {
        # Throttle values for vol copy are from 1(10%) through 10(100%)
        @tvalues = (1, 10);
        $function = \&Volume::Copy::volcopy_throttle;
    } elsif ( $test->{_sm}->isa('NOTE::Aggregate::Copy') ) {
        # Throttle values for vol copy are from 1(10%) through 10(100%)
        @tvalues = (1, 10);
        $function = \&Aggregate::Copy::aggrcopy_throttle;
    } else {
        # Throttle values are in kb
        @tvalues = (20, 1000000);
    }

    $throttle = $tvalues[0];

    do {
        # Throttle the transfer and wait for a few seconds
        if ( $function ) {
            # VolCopy
            &$function($vol->Filer->conn, $test->{_sm}->op_num, $throttle);
        } else {
            # SnapMirror
            sayEOF($vol->Filer->conn, "snapmirror throttle $throttle " .
                   $test->{_sm}->dst_path);
        }

        sleep $delta;

        # Get status of transfer
        $status = $test->{_sm}->status;
        if ( $status->{'current-transfer-error'} ) {
            croak $status->{'current-transfer-error'};
        }

        # Get next throttle value
        if ( $throttle == $tvalues[0] ) {
            $throttle = $tvalues[1];
            $delta    = 2;
        } else {
            $throttle = $tvalues[0];
            $delta    = 10;
        }
    } while ( $status->{status} eq 'transferring' );
}

# Function to get the proper volume to use for error injection
sub _get_volume($$$;) {
    my ($self, $test, $item) = @_;
    my ($aggr, $vol, $filer, $var);

    # Get volume for SRC or DST
    if ( $item->{where} eq 'SRC' ) {
        $vol = $test->{_src};
        $filer = 'FILER_SRC';
    } else {
        $vol = $test->{_dst};
        $filer = 'FILER_DST';
    }

    if ( $vol ) {
        if ( $vol->isa('NOTE::Aggregate') ) {
            # Src or dst is already an aggregate
            return ($vol, undef);
        }

        if ( $vol->isa('NOTE::Qtree') ) {
            # Get volume from the qtree object
            $vol = $vol->Volume;
        }

        if ( $item->{option} && $item->{option} eq 'ROOT' ) {
            # Use root volume of SRC/DST instead
            $vol = NOTE::Volume->attach(
                Filer => $vol->Filer,
                name  => $vol->Filer->vol_root,
            );
        }
        $aggr = $vol->Aggregate if $vol->type eq 'flex';
    } else {
        $var = $filer . '_AGGR';
        if ( $test->{_vars}{$var} ) {
            eval {
                $aggr = NOTE::Aggregate->attach(
                    Filer   => { host => $test->{_vars}{$filer} },
                    name    => $test->{_vars}{$var},
                    recycle => 1,
                );
            };
        }
        if ( ! $test->{_vars}{$var} || $@ ) {
            $aggr = NOTE::Aggregate->create(
                Filer   => { host => $test->{_vars}{$filer} },
                recycle => 1,
            );
            $test->{_vars}{$var} = $aggr->name;
        }
    }

    return ($aggr, $vol);
}

# Function to step through results of a given test and validate
# pass/fail status of relationships and return a PASS/FAIL result object
sub _process_results($$$$;) {
    my ($test, $inj_list, $sm_h, $type) = @_;
    my ($fail, $fail_dst, $failure, $result, %found_fail);
    my ($item, $sm_type, $sm_rel, $itype, $ientry, %exp_fail);
    my ($filer, $filer_name, $vol_name, $aggr_name, $tmod);
    my ($msg_exec, $msg_psm, $msg_tc, $msg, @tc_results, $res, $err);
    $type = lc($type);

    $fail = 0;
    $msg = "\n";

    $res = eval {
        # Mark the relationships expected to fail
        foreach $item ( @$inj_list ) {
            # Skip error not injected
            next unless $item->{_done};

            # Find where error was injected
            if ( $tmod_h{$item->{tmod}} ) {
                $tmod = $item->{tmod};
            } else {
                $tmod = 'XFER';
            }

            # Get error injection table entry
            $itype = $item->{type};
            $ientry = $errinj_table{$itype};
            if ( defined $ientry ) {
                if ( $item->{option} && $ientry->{$item->{option}} ) {
                    $ientry = $ientry->{$item->{option}};
                }
            } else {
                croak "Error injection <$itype> does not have " .
                      "an entry in the results table";
            }

            if ( ! exists $ientry->{fail} || ! $ientry->{fail}{$tmod} ) {
                # Not expecting an error
                next;
            }

            # Get filer where action was executed on
            $filer = $item->{_filer};

            # Get volume or aggregate name
            $vol_name  = $item->{_vol};
            $aggr_name = $item->{_aggr};

            if ( $filer ) {
                $filer_name = Tharn::host($item->{_filer})->id;
            }

            # Process all relationships
            foreach $sm_type ( keys %$sm_h ) {
                foreach $sm_rel ( @{$sm_h->{$sm_type}} ) {
                    if ( $ientry->{mode} eq 'one' && $sm_type eq $type ) {
                        $exp_fail{$sm_rel->{dst_path}} = 1;
                        last;
                    } elsif ( $ientry->{mode} eq 'volume' &&
                      $sm_rel->{dst_path} =~ /^$filer_name:\/vol\/$vol_name\//o ) {
                        $exp_fail{$sm_rel->{dst_path}} = 1;
                    } elsif ( $ientry->{mode} eq 'filer_dst' &&
                      $sm_rel->{dst_path} =~ /^$filer_name:/o ) {
                        $exp_fail{$sm_rel->{dst_path}} = 1;
                    } elsif ( $ientry->{mode} eq 'filer' ) {
                        if ( $sm_rel->{src_path} =~ /^$filer_name:/o ) {
                            $exp_fail{$sm_rel->{src_path}} = 1;
                        }
                        if ( $sm_rel->{dst_path} =~ /^$filer_name:/o ) {
                            $exp_fail{$sm_rel->{dst_path}} = 1;
                        }
                    }
                }
            }
        }

        # Create a message string that contains the testcase processing
        # errors, any script execution results, and any errors found
        # in parallel_sm results.

        # Check _exec_results
        ## script execution results
        if (exists $test->{_exec_results}) {
            $msg_exec = '';
            foreach my $exec_rslt (@{$test->{_exec_results}}) {
                # A PASS result is OK, otherwise we need to
                # report a failure
                $fail = 1 if ($exec_rslt->severity > 2);
                $msg_exec .= $exec_rslt->type() . "  " .
                  join("\n", $exec_rslt->messages()) . "\n\n";
            }
        }

        # If we are expecting a FATAL result from parallel_sm.thpl,
        # verify that is what was received.
        $msg_psm = '';
        if (exists $exp_fail{fatal}) {
            foreach $result (@{$test->{_psm_results}}) {
                # Only examine the result for parallel_sm
                next if ($result->runid() !~ /_parallel_sm$/);
                if ($result->type() eq 'FATAL') {
                    # We wanted a FATAL, the test case passed
                    $msg_psm .= "Expected and received FATAL result from " .
                                "parallel_sm.thpl\n\n";
                    return;
                } elsif ($result->severity() <= 2) {
                    # PASSed or received WARN, INFO, etc.
                    $msg_psm .= "parallel_sm.thpl expected a FATAL result, " .
                                "but passed\n\n";
                    $fail = 2;
                    return;
                } else {
                    # Test failed, but with incorrect result type
                    $msg_psm .= "parallel_sm failed for unexpected reason. " .
                                "Returned a " . $result->type() .
                                " result, expected a FATAL result\n\n";
                    $fail = 2;
                    return;
                }
            }
        }

        # If we are expecting a FILER_CMD failure, check the results
        # and verify at least one failure takes place.
        if (exists $exp_fail{filer_cmd_fail}) {
            foreach $result (@{$test->{_psm_results}}) {
                if ($result->severity > 2) {
                    # We found a failure, test case passes
                    $msg_psm .= "FILER_CMD testcase " . $test->{TESTCASE} .
                                " reports at least one failure as expected\n\n";
                    return;
                }
            }
            $msg_psm .= "FILER_CMD testcase " . $test->{TESTCASE} .
                        " expected FAIL result, but none was found\n\n";
            $fail = 2;
            return;
        }

        # If we are expecting normal results with one or more failures then
        # verify that the correct relationships passed or failed.
        foreach $result (@{$test->{_psm_results}}) {
            # If the result is a FAIL, check the messages
            # against the expected list of failure destinations
            if ($result->type() eq 'FAIL') {
                $failure = ($result->messages)[0];
                if ($failure =~ /\+\-\> (\S+)/) {
                    # The failure message should contain a NOTE
                    # representation of a relationship
                    $fail_dst = $1;
                    if (!exists $exp_fail{$fail_dst}) {
                        # If the destination in the failure message is not
                        # a destination in the expected failures, then first
                        # make sure we haven't already found this destination
                        # in a previous failure message. If not, report the
                        # unexpected error.
                        if (!exists $found_fail{$fail_dst}) {
                            $fail = 2;
                            $msg_psm .= "Destination $fail_dst reported an " .
                                        "unexpected error: $failure\n\n";
                        }
                    } else {
                        # The failure destination was in the expected failures,
                        # remove it from the expected failures and put it into
                        # the found failures.
                        delete $exp_fail{$fail_dst};
                        $found_fail{$fail_dst} = 1;
                    }
                } else {
                    # The failure message did not contain a NOTE
                    # formatted relationship
                    $fail = 2;
                    $msg_psm .= "Unexpected error message found: $failure\n\n";
                }
            } elsif ($result->severity() > 2) {
                # Result is INSPECT, FATAL, CONFIG, or SCRIPT
                # Something bad happened, we don't worry about
                # the other results anymore. The test is invalid.
                return $result;
            }
        }

        foreach $fail_dst (keys %exp_fail) {
            # The remaining keys in exp_fail are expected failures that were
            # not found in the failure messages
            $fail = 2;
            $msg_psm .= "Destination $fail_dst was expected to fail but " .
                        "did not report an error\n\n";
        }

        # If we have reached this point without finding a failure
        # in the parallel_sm results (fail = 2), report a success
        if ($fail < 2) {
            $msg_psm .= "Expected results received for parallel_sm.thpl\n";
        }

        return;
    };
    $err = $@;

    if ( $res ) {
        $test->{_test_result} = $res;
        return;
    }

    eval {
        # Put testcase processing errors into the message string.
        ## testcase processing results
        if (exists $test->{_results}) {
            # Do not report socket read errors if parallel_sm failed
            @tc_results = ($fail == 2) ?
              grep { ! /Unexpected EOF on socket read/ } @{$test->{_results}} :
              @{$test->{_results}};
            $fail = 1;
            return unless @tc_results;
            $msg_tc .= join("\n", @tc_results) . "\n\n";
        }
    };
    if ($@) {
        $err = $err ? "$err\n$@" : $@;
    }

    if ($err) {
        $fail = 1;
        $msg .= "++++++++++ RESULTS PROCESSING ++++++++++\n\n" . $err;
    }

    $msg .= "++++++++++ TESTCASE PROCESSING ++++++++++\n\n$msg_tc" if $msg_tc;
    $msg .= "++++++++++ SCRIPT EXECUTION RESULTS ++++++++++\n\n$msg_exec"
      if $msg_exec;
    $msg .= "++++++++++ PARALLEL_SM RESULTS ++++++++++\n\n$msg_psm" if $msg_psm;

    # Return a FAIL result if any errors discovered, else return a PASS result
    if ($fail) {
        $test->{_test_result} = NATE::Result::Fail->new($msg);
    } else {
        $test->{_test_result} = NATE::Result::Pass->new($msg);
    }

    return;
}

1;

