# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Execution Report system library code
## @author rjray@netapp.com
## @status shared
## @pod here

package ExecutionReport;

use strict;
use warnings;
use lib '/usr/software/test/noarch/share/nate/current/lib'; # For NATE libs

use File::Basename qw(dirname basename);
use File::Find;
use File::Spec;
use IO::Handle;
use Time::Local 'timelocal';

use Clone 'clone';
use Config::General qw(ParseConfig);
use Params::Validate qw(validate validation_options SCALAR OBJECT);

use ExecutionReportData;
use ExecutionReportWriter;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;

# Set the on_fail behavior for Params::Validate to throw an exception (of the
# NATE::Exceptions::Argument class) on errors instead of a plain croak().
validation_options(
    on_fail => sub {
        NATE::Exceptions::Argument->throw(shift);
    }
);

# These are spec-fragments that are used in multiple places in various calls
# to validate() and/or validate_with().

# Plain optional scalar, no real validation
my $OPTIONAL_SCALAR = {
    type     => SCALAR,
    optional => 1,
};

# Note that all of these elements are all named as private attributes, thus
# ensuring that the methods C::MM generates are also named as private. We
# will define our own public accessors that are read-only for users to call.
# Also, I will be providing an explicit new() constructor, so that users can
# use the "visible" accessor names as options, while the data is stored under
# the private names.
use Class::MethodMaker [
    scalar => '_config',
    scalar => '_workflows',
    scalar => '_streams',
];

# These are the publically-visible accessors, all read-only.

# Scalar accessors:
for my $method (qw(config)) {
    ## no critic(ProhibitNoStrict)
    no strict 'refs';
    my $private = "_$method";
    *{$method} = sub {
        my $self = shift;

        return $self->$private();
    };
}
# Compound data structure accessors:
for my $method (qw(workflows streams)) {
    ## no critic(ProhibitNoStrict)
    no strict 'refs';
    my $private = "_$method";
    *{$method} = sub {
        my $self = shift;

        return clone($self->$private());
    };
}

=head1 NAME

ExecutionReport - Generate and write execution reports based on WFC

=head1 SYNOPSIS

    my $reporter = ExecutionReport->new(config => $config_file);
    my $data     = $reporter->get_data(
        name   => $project_name,
        logdir => $log_dir
    );
    $reporter->write_report(
        format        => 'wiki',
        data          => $data,
        document_base => 'Execution_Reports_%Y%m%d/%N',
        file          => 'report_%Y%m%d_%H%M%S',
        title         => '%N Execution for %C',
        username      => $MW_username,
        password      => $MW_password
    );

=head1 DESCRIPTION

This is the main class for the Execution Reports system. This class is used
to retrieve data from logs for a single execution, and then can use the
B<ExecutionReportWriter> class to generate and publish a report.

=head1 ACCESSORS

The following property accessors are defined for this class. Each of these is
a read-only accessor that cannot change the value of the underlying property.

=over

=item B<config>

A string value, the file name of the configuration file that was read when
this object was created. See L</CONFIGURATION FILE>, below, for detail on the
format of this file.

=item B<workflows>

An array reference of hash references, the workflow definitions from the
configuration file. See L</DATA STRUCTURE>, below, for more detail.

=item B<streams>

An array reference of hash references, the stream definitions from the
configuration file. See L</DATA STRUCTURE>, below, for more detail.

=back

The B<streams> and B<workflows> accessors return deep clones of their data, so
that any changes made by the user do not affect the object itself.

=head1 METHODS

The class provides the following methods. No subroutines are exported by this
package.

=head2 new

    my $reporter = ExecutionReport->new(config => $config_file);

Create a new object of this class and return it to the caller.

The following option is recognized by B<new>:

=over

=item B<config>

Required, the name of the configuration file to read the definitions of
workflows and streams from. See L</CONFIGURATION FILE>, below, for details
on the file content and layout.

=back

The following exceptions may be thrown by B<new>:

=over

=item B<NATE::Exceptions::Argument>

This exception will be thrown if any unknown parameter is passed in, or if the
value of the C<config> parameter is invalid.

=item B<NATE::BaseException>

This exception will be thrown if anything else goes wrong, such as file access
issues on the configuration file, or problems parsing the file.

=back

=cut

###############################################################################
#
#   Sub Name:       new
#
#   Description:    Class constructor.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $class    in      scalar    Class to bless into
#                   @opts     in      list      Additional options
#
#   Returns:        Success:    new object
#                   Failure:    throws exception
#
###############################################################################
sub new {
    my ($class, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            config   => { type => SCALAR },
        }
    );
    my $self = bless {}, $class;
    $self->_read_validate_config($opts{config});

    return $self;
}

=head2 get_data

    my $data = $reporter->get_data(
        name   => $project_name,
        logdir => $log_dir
    );

Perform the analysis of a tree of log files rooted at the directory given by
B<logdir> and return an B<ExecutionReportData> object containing the data.
Throws an exception if something goes wrong.

The following options are recognized by B<get_data>:

=over

=item B<name>

Required, the name of the project for which this execution was done. This is a
string that is used later in report generation. However, it has no direct
effect on the analysis, other than being a part of the data object that is
returned.

=item B<logdir>

Required, the directory path that should be analyzed for WFC-based workflow
test results. Should generally be an absolute path, but if it is relative the
current working directory will be prepended to it, and the path will be made
canonical. The returned data object contains the (final) path used as one of
its attributes.

=back

The following exceptions may be thrown by B<get_data>:

=over

=item B<NATE::Exceptions::Argument>

If an unknown parameter is passed, or if the value of one the defined
parameters is invalid, this exception is thrown.

=item B<NATE::BaseException>

Any other error will throw this exception. Such cases include being unable
to extract a time-stamp from the execution logs, file access errors, etc.

=back

=cut

###############################################################################
#
#   Sub Name:       get_data
#
#   Description:    Retrieve the execution data from a given directory.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional arguments
#
#   Returns:        Success:    ExecutionReportData instance
#                   Failure:    throws exception
#
###############################################################################
sub get_data {
    my ($self, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            name   => { type => SCALAR },
            logdir => { type => SCALAR },
        }
    );

    # Get the logdir, and make sure it's an absolute path.
    my $logdir = $opts{logdir};
    if ($logdir !~ m{^/}) {
        $logdir = File::Spec->canonpath(File::Spec->rel2abs($logdir));
    }
    # Get the timestamp for this run. If the logdir name does not contain it,
    # this method will look for the first time-stamped line in the top-most
    # ntest.log file in the directory.
    my $timestamp = $self->_get_timestamp($logdir);
    # Get all the *.log files for the run. Filtering/pruning of the list
    # happens later.
    my @logfiles = $self->_get_logfiles($logdir);

    # Now iterate over the streams in this configuration, getting the relevant
    # log files and searching them for test results from the set of
    # workflows.
    my @data;
    for my $stream (@{$self->_streams}) {
        my $path = $self->_derive_path_re($stream);
        my @files = grep { /$path/ } @logfiles;
        my @results;
        # Iterate over the workflows, looking for results for each one, one
        # at a time. Keep the ordering consistent so that @data is a
        # consistent structure.
        for my $workflow (@{$self->_workflows}) {
            push(@results, $self->_analyze_workflow($workflow, \@files));
        }
        push(@data, \@results);
    }

    my $data = ExecutionReportData->new(
        name      => $opts{name},
        datetime  => $timestamp,
        logdir    => $logdir,
        streams   => $self->_streams,
        workflows => $self->_workflows,
        data      => \@data
    );

    return $data;
}

=head2 write_report

    $reporter->write_report(
        format        => 'wiki',
        data          => $data,
        document_base => 'Execution_Reports_%Y%m%d/%N',
        file          => 'report_%Y%m%d_%H%M%S',
        title         => '%N Execution for %C',
        username      => $MW_username,
        password      => $MW_password
    );

Write a report out, using the given data from a previous call to B<get_data>.
This is a front-end to the B<ExecutionReportWriter> class, to simplify things
for the user of this class.

Returns the calling object upon success, throws an exception on error.

The following parameters are recognized by B<write_report>:

=over

=item B<format>

Required, a string, one of C<wiki> or C<html>. Specifies the format of the
report to write, either to wikid or to a raw HTML file.

=item B<data>

Required, an instance of B<ExecutionReportData>. The data object returned from
a previous call to B<get_data>.

=item B<document_base>

Optional, a string. If given, this is the base location for HTML files or
wikid pages. See L<ExecutionReportWriter/new> for a more detailed explanation
of this parameter.

=item B<file>

Required, a string. The file/page to write the report to. Should I<not> include
a suffix, as the report-writer will add one if/when needed (wikid does not need
a suffix, HTML will add C<.html> for you).

=item B<title>

Optional, a string. The report title, used in the page itself and in the case
of HTML used to provide content for the C<< <title> >> tag. If not given, a
simple default title is used.

=item B<username>

Optional, a string. Only required if B<format> is C<wiki>. The username to use
when connecting to the MediaWiki publishing API.

=item B<password>

Optional, a string. Only required if B<format> is C<wiki>. The password to use
when connecting to the MediaWiki publishing API.

=item B<wiki_url>

Optional, a string. If passed, overrides the default URL for the MediaWiki
publishing API.

=back

The B<document_base>, B<file> and B<title> parameters are all subject to
token expansion, to allow the project name and elements of the time-stamp to
be included in the report location, file name and title. See
L<ExecutionReportWriter/TOKEN EXPANSION> for a detailed description of the
tokens that can be inserted into these values.

The following exception may be thrown by B<write_report>:

=over

=item B<NATE::Exceptions::Argument>

If an unknown parameter is passed, or an invalid value for a recognized
parameter, this exception will be thrown.

=back

Additionally, any exceptions thrown by the B<new> or B<write_report> methods
of B<ExecutionReportWriter> may be seen here as well.

=cut

###############################################################################
#
#   Sub Name:       write_report
#
#   Description:    Write a report for the given data, in the specified format.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional arguments
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub write_report {
    my ($self, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            format        => { type => SCALAR },
            data          => { type => OBJECT, isa => 'ExecutionReportData' },
            document_base => { type => SCALAR, default => q{} },
            file          => { type => SCALAR },
            title         => $OPTIONAL_SCALAR,
            username      => $OPTIONAL_SCALAR,
            password      => $OPTIONAL_SCALAR,
            wiki_url      => $OPTIONAL_SCALAR,
        }
    );

    my $writer = ExecutionReportWriter->new(
        format => $opts{format},
        document_base => $opts{document_base},
        $opts{username} ? (username => $opts{username}) : (),
        $opts{password} ? (password => $opts{password}) : (),
        $opts{wiki_url} ? (wiki_url => $opts{wiki_url}) : ()
    );
    $writer->write_report(
        data => $opts{data},
        file => $opts{file},
        ($opts{title}) ? (title => $opts{title}) : ()
    );

    return $self;
}

###############################################################################
#
# Private methods
#
###############################################################################

###############################################################################
#
#   Sub Name:       _read_validate_config
#
#   Description:    Read and validate the given configuration file.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $file     in      scalar    Config file to read
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub _read_validate_config {
    my ($self, $file) = @_;
    my %hash;

    # If $file is not absolute, assume it is relative to this module:
    if ($file !~ m{^/}) {
        my $dir = dirname __FILE__;
        $file = "$dir/$file";
    }

    # Parse the given file:
    my $retval = eval {
        %hash = ParseConfig(
            -ConfigFile     => $file,
            -LowerCaseNames => 1,
            -AutoTrue       => 1
        );
        1;
    };
    if (! $retval) {
        NATE::BaseException->throw("Error parsing configuration $file: $@");
    }

    # OK, we've parsed it successfully. Now validate the content of it.

    # Check <streams>. It must be a hash with one key, whose value is an array
    # with 1+ elements.
    if ($hash{streams} && (ref($hash{streams}) eq 'HASH') &&
        (ref($hash{streams}->{stream}) eq 'ARRAY') &&
        (@{$hash{streams}->{stream}} >= 1)) {
        # Every stream must have these elements, and all must be non-null
        my $count = 1;
        for my $stream (@{$hash{streams}->{stream}}) {
            if (! ($stream->{name} && $stream->{variant} && $stream->{path} &&
                       $stream->{io})) {
                NATE::BaseException->throw(
                    "Stream block #$count is missing one or more fields"
                );
            }
            $count++;
        }
        $self->_streams($hash{streams}->{stream});
    } else {
        NATE::BaseException->throw(
            'The "<streams>" block is either missing or has invalid structure'
        );
    }
    # Same for <workflows>.
    if ($hash{workflows} && (ref($hash{workflows}) eq 'HASH') &&
        (ref($hash{workflows}->{workflow}) eq 'ARRAY') &&
        (@{$hash{workflows}->{workflow}} >= 1)) {
        # Every workflow must have these elements, and all must be non-null
        my $count = 1;
        for my $workflow (@{$hash{workflows}->{workflow}}) {
            if (! ($workflow->{name} && $workflow->{script} &&
                       $workflow->{testcase})) {
                NATE::BaseException->throw(
                    "Workflow block #$count is missing one or more fields"
                );
            }
            $count++;
        }
        $self->_workflows($hash{workflows}->{workflow});
    } else {
        NATE::BaseException->throw(
            'The "<workflows>" block is either missing or has invalid structure'
        );
    }

    # Save the config file used, just in case it's needed
    $self->_config($file);

    return $self;
}

###############################################################################
#
#   Sub Name:       _get_timestamp
#
#   Description:    Get the time-stamp associated with this test run. Get it
#                   either from the name of the logdir itself, or from the
#                   top-most ntest.log file in the directory.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $logdir   in      scalar    Log directory to refer to
#
#   Returns:        Success:    integer time value
#                   Failure:    throws exception
#
###############################################################################
sub _get_timestamp {
    my ($self, $logdir) = @_;
    my ($year, $month, $day, $hour, $min, $sec);

    if ($logdir =~ /(\d{4})(\d\d)(\d\d)_(\d\d)(\d\d)(\d\d)/) {
        $year  = $1;
        $month = $2;
        $day   = $3;
        $hour  = $4;
        $min   = $5;
        $sec   = $6;
    } else {
        if (opendir(my $dh, $logdir)) {
            # Take everything in $logdir that looks like an ntest log file
            my @files = sort grep { /^0+_ntest[.]log$/ } readdir($dh);
            closedir $dh;
            if (@files) {
                if (open my $fh, '<', "$logdir/$files[0]") {
                    my @lines = <$fh>;
                    # Close and make sure nothing fishy happened as we were
                    # closing:
                    if (! close($fh)) {
                        NATE::BaseException->throw(
                            "Error closing $files[0]: $!"
                        );
                    }
                    chomp(@lines);
                    # Find the first line with a time-stamp
                    for my $line (@lines) {
                        if ($line =~
                                /^(\d{4})(\d\d)(\d\d) (\d\d)(\d\d)(\d\d)/) {
                            $year  = $1;
                            $month = $2;
                            $day   = $3;
                            $hour  = $4;
                            $min   = $5;
                            $sec   = $6;
                            last;
                        }
                    }
                    # Did we find such a line?
                    if (! $year) {
                        NATE::BaseException->throw(
                            "No time-stamped lines found in $files[0], " .
                            'cannot continue'
                        );
                    }
                } else {
                    NATE::BaseException->throw(
                        "Error opening $files[0] for reading: $!"
                    );
                }
            } else {
                NATE::BaseException->throw(
                    "No ntest.log files found in $logdir, cannot continue"
                );
            }
        } else {
            NATE::BaseException->throw("Error opening $logdir for reading: $!");
        }
    }

    return timelocal($sec, $min, $hour, $day, $month - 1, $year - 1900);
}

###############################################################################
#
#   Sub Name:       _get_logfiles
#
#   Description:    Scan the given directory for *.log files using File::Find.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $logdir   in      scalar    Directory to scan
#
#   Returns:        Success:    sorted list of files
#                   Failure:    throws an exception
#
###############################################################################
sub _get_logfiles {
    my ($self, $logdir) = @_;

    # According to docs, File::Find::find can produce warnings in certain
    # weird situations. Disable just the warnings for that module.
    ## no critic(ProhibitNoWarnings)
    no warnings 'File::Find';

    my @filelist;
    find(sub { /[.]log$/ && push(@filelist, $File::Find::name) }, $logdir);

    return (sort @filelist);
}

###############################################################################
#
#   Sub Name:       _derive_path_re
#
#   Description:    Turn a "path" string in a stream block definition into a
#                   Regexp object that can be used to prune the full list of
#                   files down to just those that are relevant to the stream.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $stream   in      hashref   Stream definition block
#
#   Returns:        regexp object
#
###############################################################################
sub _derive_path_re {
    my ($self, $stream) = @_;

    my $path = $stream->{path};
    my $name = $stream->{name};

    my %substitutions = (
        '<STEST>' => '\d+_stest',
        '<NAME>'  => $name,
        '<N>'     => '\d+_',
        '<NNAME>' => "\\d+_$name",
    );

    my $tokens = join(q{|}, keys %substitutions);
    $path =~ s/($tokens)/$substitutions{$1}/eg;

    return qr{$path};
}

###############################################################################
#
#   Sub Name:       _analyze_workflow
#
#   Description:    Analyze the files passed in for data relevant to the given
#                   workflow.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $workflow in      hashref   Workflow spec from config
#                   $files    in      arrayref  Files from the stream being
#                                                 analyzed that are relevant
#
#   Returns:        Success:    one of ('pass', 'fail', 'script', undef)
#                   Failure:    throws exception
#
###############################################################################
sub _analyze_workflow {
    my ($self, $workflow, $files) = @_;
    my $result = undef;

    # We'll need the testcase name, to search for in the logs. But to isolate
    # the relevant logs, we'll also need the script name, stripped of '.thpl'
    # and appended with '.log'.
    my $test = $workflow->{testcase};
    my $test_re = qr{
        (pass|fail):\s+         # This is the result I'll pull from the line
        testcase\s+complete:\s+ # Helps distinguish from other lines
        $test[(][)]             # The test-case we're specifically looking for
    }ix;
    my $script = basename $workflow->{script};
    $script =~ s/[.]thpl$//i;
    $script .= '.log';
    my @possible_files = grep { /$script/ } @{$files};

    # If this list is empty, $result just stays undef which is what we want.
    if (@possible_files) {
        for my $file (@possible_files) {
            ## no critic(RequireBriefOpen)
            if (open my $fh, '<', $file) {
                # Normally I'd read the whole file, to avoid the critic-warning
                # I just suppressed. But some of these log files can be HYOOGE,
                # so not this time.
                while (defined(my $line = <$fh>)) {
                    if ($line =~ /$test_re/) {
                        $result = lc $1;
                        last;
                    }
                }
                # Close and make sure nothing fishy happened as we were closing:
                if (! close($fh)) {
                    NATE::BaseException->throw("Error closing $file: $!");
                }
            } else {
                NATE::BaseException->throw(
                    "Error opening $file for reading: $!"
                );
            }

            # Did this file give us a result? If so, we can stop.
            last if (defined $result);
        }

        # Do we have a result at this point? If we had files, but none produced
        # a result, assume there was a script or config error.
        $result ||= 'script';
    }

    return $result;
}

=head1 DATA STRUCTURE

The attributes B<workflows> and B<streams> return compound data structures. Each
is an array reference of hash references. This section details those hash
structures.

=head2 workflows

The array of workflows preserves the order in which the workflows were defined
in the configuration file. Each element of the array is a hash reference with
the same set of keys:

=over

=item B<name>

This is the "human-readable" workflow name, used for column headings in the
report, etc.

=item B<script>

The path of the script itself, relative to the top of a Perforce workspace.
This is used to prune down the full set of log files to just those related to
the specific workflow.

=item B<testcase>

This is the name of the testcase itself, as seen in the logs and as reported
to ALM. This is used to identify failure/success results in the logs themselves.

=back

Note that in the data returned from analyzing logs, that some streams may not
run all the workflows that are defined. As detailed in
L<ExecutionReportData/DATA FORMAT>, these cases are stored as C<undef> in
the matrix of results that gets created.

=head2 streams

Like workflows, the array of streams also preserves the order in which the
streams themselves were defined in the configuration file. Each element of the
array is a hash reference with the same set of keys:

=over

=item B<name>

The name of the stream, should be distinct from all other streams in the
array.

=item B<path>

The relative path element(s) from the log-directory down to the start of the
log files that are specific to this stream. This value may have tokens that
are translated into regular expressions, see the definition of this field in
L</CONFIGURATION FILE>, below.

=item B<variant>

The name of the variant being tested in this stream. Should correspond to a
variant declared in the project's configuration for WFC itself. That is, it
should match the name of a variant in the declaration of a C<< <variant> >>
block, see L<RRS_Config/CONFIGURATION FORMAT>.

=item B<io>

The type of I/O being done by this stream, one of C<nfs>, C<cifs>, C<san> or
C<san-windows>. This is used primarily to distinguish when a given variant
appears more than once in an execution.

=back

When analysis is done on the log files, streams are examined in the order they
are declared in the configuration file. This may not be the order in which the
execution actually ran them, but it provides a clear, predictable ordering for
the sake of report generation.

=head1 CONFIGURATION FILE

When creating an object of this class, it is necessary to provide a
configuration file to be read and parsed. This file provides the declaration
of the workflows and streams, which are read into the structures described by
the previous section. This section describes the format of this configuration
file.

=head2 Comments

Lines which are blank are ignored. Lines that start with a C<#> character are
considered comments, and are also ignored. The C<#> character can also start
a comment within a line (known as "side comments"), unless it is preceded by a
backslash character (C<\>).

Comments are encouraged, and the sample configuration file for FLO is heavily
commented in order to make the contents clearer to anyone reading it.

=head2 The C<workflows> Section

The first section of the file is the section for workflows. This section
declares one or more workflows using a mix of pseudo-XML and "key = value"
syntax. It is closest in syntax to the Apache web server configuration file.

There must be a single top-level C<< <workflows> >> block, which contains one
or more C<< <workflow> >> blocks. Each workflow block contains three
declarations: B<name>, B<script> and B<testcase>. These are defined as:

=over

=item B<name>

The human-friendly name for the workflow, such as "LS & DP" or "Cascade". It is
this name that will be used for the column headings in any reports generated.

=item B<script>

The actual THPL script that constitutes this workflow, expressed as a relative
path within any given Perforce repository. The name of the script (the last
element of the path) is used to identify log files that are specific to this
workflow.

=item B<testcase>

The testcase name that is reported for this workflow to ALM. This string is
what is ultimately searched for through the logs, to determine whether this
workflow passed or failed.

=back

A sample block might look like this:

    <workflows>
      <workflow>
        name     = 1 to 1
        script   = dp/snapmirror/rapid/SM_lifecycle_1x1-RRS.thpl
        testcase = FLO_workflow_Basic
      </workflow>
      <workflow>
        name     = Fanout
        script   = dp/snapmirror/rapid/SM_lifecycle_fanout-RRS.thpl
        testcase = FLO_workflow_Fanout
      </workflow>
      <workflow>
        name     = LS & DP
        script   = dp/snapmirror/rapid/SM_lifecycle_LSandDP-RRS.thpl
        testcase = FLO_workflow_LSandDP
      </workflow>
      <workflow>
        name     = Cascade
        script   = dp/snapmirror/rapid/SM_lifecycle_cascade-RRS.thpl
        testcase = FLO_workflow_Cascade
      </workflow>
    </workflows>

This defines four workflows. Each has a distinct name and distinct testcase
string. Each defines the script as the relative path that ends in a C<*.thpl>
file.

=head2 The C<streams> Section

The second section of the file is the section for streams. It uses the same
basic syntax that the workflows block uses, except that the outer block is
called C<< <streams> >>, and the inner blocks are called C<< <stream> >>.

Each stream declaration contains four declarations: B<name>, B<path>,
B<variant> and B<io>. These are defined as:

=over

=item B<name>

The name of the stream. This may be used in the construction of regular
expressions for the B<path> attribute (described the next). The name must
be unique, or else different streams will try to use the same directory for
log files.

=item B<path>

The (relative) directory path within the log directory, that leads to the log
files that are specific to this stream.

This value may have several tokens in it that get turned into regular expression
snippets when the value is actually used. This allows the user to not have to
worry about things such as the numbering of steps that B<ntest> puts as a
prefix to file and directory names. See below for a full description of the
tokens that are available for use

=item B<variant>

The name of the variant, per the WFC configuration file, that this stream is
using for its tests. At present, this value is not checked directly against
that configuration. So the name does not have to match. However, future
functionality might require matching this name to the that configuration, so it
is recommended that they match correctly.

This value is used, along with B<io> if needed, to uniquely identify an
execution stream. When two executions are compared, this value is what is used
to match streams for comparison purposes (as the stream name can be easily
changed, or streams re-ordered). At present, the B<ExecutionReport> system
does not support having the same variant appear more than one, unless the value
of B<io> is different.

=item B<io>

The type of I/O being utilized in this stream, one of C<nfs>, C<cifs>, C<san>
or C<san-windows>. This value is mainly used to disambiguate cases where a
B<variant> value appears in more than one stream. A B<variant> can appear more
than once only if the value of B<io> is different for each stream.

=back

The value of B<path>, as explained above, can have one or more tokens in the
string which will be turned into regular expression fragments when the path
value is used to filter the list of log files down to just those relevant to
the given stream.

The available tokens are:

=over

=item B<< <NAME> >>

If present, this token will be replaced with the value of the B<name> attribute
for this stream. This is not actually a regular expression fragment, but it is
handled by the same part of the library.

=item B<< <N> >>

This token is replaced with C<\d+_>, a pattern that matches a seqence of digits
followed by an underscore. This is used to represent the numbering of steps
that B<ntest> performs, and allows for path elements such as C<< <N>step_name
>> that are not hard-coded to specific step-number.

=item B<< <NNAME> >>

This is a shortcut for specifying C<< <N><NAME> >>. It combines the C<< <N> >>
token with the name token, allowing for cases where the stream name might be
prefixed with a number due to B<ntest>.

=item B<< <STEST> >>

This is a shortcut for specifying C<< <N>stest >>, since B<stest> is generally
always being used by executions that use the WFC elements.

=back

To illustrate all of this, consider the following example C<< <streams> >>
block:

    <streams>
      <stream>
        name    = stream_01
        path    = <STEST>/<N>RRS_block_01/<NAME>
        variant = rw_230ss
        io      = nfs
      </stream>
      <stream>
        name    = stream_02
        path    = <STEST>/<N>RRS_block_01/<NAME>
        variant = rw_128ss
        io      = nfs
      </stream>
      <stream>
        name    = stream_03
        path    = <STEST>/<N>RRS_block_01/<NAME>
        variant = dp2rw_230ss
        io      = nfs
      </stream>
      <stream>
        name    = stream_04
        path    = <STEST>/<N>RRS_block_01/<NAME>
        variant = dp2rw_230ss
        io      = cifs
      </stream>
    </streams>

This block defines four streams. Each stream has the same value of B<path>,
because the token expansion handles the adjustment of the value for each
stream. For example, the first stream's value for B<path> will be expanded to:

    \d+_stest/\d+_RRS_block_01/stream_01

This will match across executions, even if the numbering of the B<stest>
step changes, or that of the next element, etc.

Note that the third and fourth streams use the same B<variant>. This is
acceptable in this configuration because each has a different value for B<io>.
If two executions are compared, it will be possible to distinguish these two
streams even if they are in a different order between executions.

=head1 AUTHOR

Randy J. Ray E<lt>rjray@netapp.comE<gt>

=cut

1;
