# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Execution Report data instance code
## @author rjray@netapp.com
## @status shared
## @pod here

package ExecutionReportData;

use strict;
use warnings;
use lib '/usr/software/test/noarch/share/nate/current/lib'; # For NATE libs

use Clone 'clone';
use Params::Validate qw(validate validation_options SCALAR ARRAYREF);

use NATE::Exceptions::Argument;

# Set the on_fail behavior for Params::Validate to throw an exception (of the
# NATE::Exceptions::Argument class) on errors instead of a plain croak().
validation_options(
    on_fail => sub {
        NATE::Exceptions::Argument->throw(shift);
    }
);

# Note that all of these elements are all named as private attributes, thus
# ensuring that the methods C::MM generates are also named as private. We
# will define our own public accessors that are read-only for users to call.
# Also, I will be providing an explicit new() constructor, so that users can
# use the "visible" accessor names as options, while the data is stored under
# the private names.
use Class::MethodMaker [
    scalar => '_name',
    scalar => '_datetime',
    scalar => '_logdir',
    scalar => '_workflows',
    scalar => '_streams',
    scalar => '_data',
];

# These are the publically-visible accessors, all read-only.

# Scalar accessors:
for my $method (qw(name datetime logdir)) {
    ## no critic(ProhibitNoStrict)
    no strict 'refs';
    my $private = "_$method";
    *{$method} = sub {
        my $self = shift;

        return $self->$private();
    };
}
# Compound data structure accessors:
for my $method (qw(workflows streams data)) {
    ## no critic(ProhibitNoStrict)
    no strict 'refs';
    my $private = "_$method";
    *{$method} = sub {
        my $self = shift;

        return clone($self->$private());
    };
}

=head1 NAME

ExecutionReportData - Encapsulate the data for one execution report

=head1 SYNOPSIS

    use ExecutionReportData;

    my $exec_data = ExecutionReportData->new(
        name      => $project_name,
        datetime  => $timestamp,
        logdir    => $dir,
        workflows => $workflows,
        streams   => $streams,
        data      => $data
    );
    # This is usually done by the ExecutionReport class

=head1 DESCRIPTION

This class encapsulates the data from a single execution report. It is
generally instantiated from within the B<get_data> method of the
B<ExecutionReport> class. The purpose of this class it to provide clean
access to the data in a way that can be extensible in the future if/when
some sort of database backing is added to the system.

An object of this class provides read-only accessors to all the underlying
attributes of the object. Once an instance of this object is created, the
content cannot be modified (a new object must be created with the differing
data).

=head1 ACCESSORS

The following property accessors are defined for this class. Each of these is a
read-only accessor that cannot change the value of the underlying property.
Those properties that are compound data structures (C<workflows>, C<streams>
and C<data>) return clones of the internal data, to prevent direct manipulation
of references.

=over

=item B<name>

The name of the "project" that this execution is a part of. For example, this
code is initially written to support the "FLO" project. The name is completely
arbitrary, but should be reasonably unique as some other parts of the
execution report system use this name in the reports that get generated.

=item B<datetime>

The time-stamp of when this execution was run, based on either the name of
the logs directory (logging directories at NetApp generally include the
start-time in the name of the directory) or if the name does not contain that
data, the first time-stamped line in the execution's C<ntest.log> file. The
value is an integer, a UNIX C<time> value.

=item B<logdir>

The log directory that was scanned to generate this data. This is kept for the
sake of uniquely indexing execution runs, even though the physical directory
itself may be deleted to free up space.

=item B<workflows>

An array reference of hash references that describe the workflows configured
for this execution run. See L<ExecutionReport|ExecutionReport> for details on
the structure of these hash references, as this property is (usually) copied
directly from the B<ExecutionReport> instance that created this object. The
value returned by this accessor is a deep clone of the internal data, so no
changes to any of the reference contents will affect this object.

=item B<streams>

As above, an array reference of hash references. This structure describes the
streams themselves. Again, see See L<ExecutionReport|ExecutionReport> for
details on the structure of these hash references, as this property is also
usually copied directly from the B<ExecutionReport> instance that created
this object. The value returned by this accessor is a deep clone of the
internal data, so no changes to any of the reference contents will affect this
object.

=item B<data>

An array reference of array references that contains the actual data that
results from scanning and analyzing the execution. The top-level reference
will have the same number of elements as C<streams> has. Each element within
this reference will have the same number of elements as C<workflows> has. The
values at each location in this matrix will represent the run status of the
corresponding workflow in the corresponding stream. See L</DATA FORMAT> below
for more detail on this.

=back

=head1 METHODS

Currently, the class only provides a constructor method. No subroutines are
exported by this package.

=head2 new

    my $exec_data = ExecutionReportData->new(
        name      => $project_name,
        datetime  => $timestamp,
        logdir    => $dir,
        workflows => $workflows,
        streams   => $streams,
        data      => $data
    );

Create a new data object and return it to the caller.

The following options are recognized by B<new>:

=over

=item B<name>

A string, the project name for the execution run that was analyzed.

=item B<datetime>

An integer, the UNIX "time" value of when the execution was started.

=item B<logdir>

A string, the logging directory that was analyzed to produce this data.

=item B<workflows>

An array reference of hash references, the description of the workflows used
in this execution.

=item B<streams>

An array reference of hash references, the description of the streams used in
this execution.

=item B<data>

An array reference of array references, the data from the analysis of the
log contents from the execution.

=back

The compound data structures (C<workflows>, C<streams> and C<data>) are cloned
when saved to the new object, so that changes to the orginial references will
not affect the new object.

The following exception may be thrown by B<new>:

=over

=item B<NATE::Exceptions::Argument>

This exception will be thrown if any of the options are given invalid values,
or if any unknown option is passed. It will also be thrown if the
dimensionality of C<workflows>, C<streams> and C<data> do not match.

=back

=cut

###############################################################################
#
#   Sub Name:       new
#
#   Description:    Class constructor
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $class    in      scalar    Class to bless into
#                   @opts     in      list      Additional options/data
#
#   Returns:        Success:    new object
#                   Failure:    throws exception
#
###############################################################################
sub new {
    my ($class, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            name      => { type => SCALAR },
            datetime  => { type => SCALAR, regex => qr/^\d+$/ },
            logdir    => { type => SCALAR },
            workflows => { type => ARRAYREF },
            streams   => { type => ARRAYREF },
            data      => { type => ARRAYREF },
        }
    );

    # Validate the content of workflows, streams and data:
    my $workflows_len = @{$opts{workflows}};
    my $streams_len   = @{$opts{streams}};
    if (! ($workflows_len && $streams_len)) {
        NATE::Exceptions::Argument->throw(
            'Both "workflows" and "streams" must be non-empty arrayrefs'
        );
    }
    if ($streams_len == @{$opts{data}}) {
        for my $elt (@{$opts{data}}) {
            if (! ((ref($elt) eq 'ARRAY') && (@{$elt} == $workflows_len))) {
                NATE::Exceptions::Argument->throw(
                    'An element of "data" is either not an arrayref, or is ' .
                    'the wrong length'
                );
            }
        }
    } else {
        NATE::Exceptions::Argument->throw(
            '"data" arrayref must have same number of elements as "streams"'
        );
    }

    # OK, we're good:
    my $self = bless {}, $class;
    $self->_name($opts{name});
    $self->_datetime($opts{datetime});
    $self->_logdir($opts{logdir});
    $self->_workflows(clone($opts{workflows}));
    $self->_streams(clone($opts{streams}));
    $self->_data(clone($opts{data}));

    return $self;
}

=head1 DATA FORMAT

The data format for an execution report is fairly simple. The C<data> member
of the object is essentially a matrix of test results, where the X axis is the
stream and the Y axis is the workflow. Hence, when the object is created the
dimensionality of C<streams> is checked against the first axis of C<data>, and
C<workflows> is checked against the second axis.

Each element of the matrix is either a string value or C<undef>. If the given
workflow did not run for the given stream at all, then the value is C<undef>.
Otherwise, it is one of the following three strings:

=over

=item C<pass>

The test reported a passing result.

=item C<fail>

The test reported a failing result.

=item C<script>

The test either reported no result, or reported an "unknown" result string. It
is therefore assumed that there was a script error of some sort, whether in the
workflow itself or in the configuration. This is intentionally distinguished
from the workflow explicitly reporting a C<fail>.

=back

For example, if there were two workflows and five streams, the structure might
look something like this:

        0           1
    0   'pass'      'pass'
    1   'pass'      'fail'
    2   'pass'      'pass'
    3   'fail'      'fail'
    4   'script'    undef

Here, element [4][1] is C<undef>, indicating that the second workflow did not
run in the fifth stream.

For each row in the data, the elements will always be in the same order as the
list of workflows in the C<workflows> member. This is regardless of what order
the workflows themselves ran in. And likewise, the rows themselves will be in
the same order as the list of streams in the C<streams> member.

For explanation of the content and structure of the C<streams> and C<workflows>
members, see L<ExecutionReport/DATA FORMAT>.

=head1 AUTHOR

Randy J.Ray E<lt>rjray@netapp.comE<gt>

=cut

1;
