# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Execution Report system report writer code
## @author rjray@netapp.com
## @status shared
## @pod here

package ExecutionReportWriter;

use strict;
use warnings;
use lib '/usr/software/test/noarch/share/nate/current/lib'; # For NATE libs

use Date::Format 'time2str';
use MediaWiki::API;
use Text::Template;
use Params::Validate qw(validate validation_options SCALAR OBJECT BOOLEAN);

use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;

# Set the on_fail behavior for Params::Validate to throw an exception (of the
# NATE::Exceptions::Argument class) on errors instead of a plain croak().
validation_options(
    on_fail => sub {
        NATE::Exceptions::Argument->throw(shift);
    }
);

# These are spec-fragments that are used in multiple places in various calls
# to validate() and/or validate_with().

# Plain optional scalar, no real validation
my $OPTIONAL_SCALAR = {
    type     => SCALAR,
    optional => 1,
};

# This is the default URL for interacting with wikid:
my $WIKI_API_URL = 'http://wikid.netapp.com/wikid/api.php';

# This will hold all the compiled templates for use by all instances:
my %TEMPLATES = ();

# The colors to use in the cells of the tables, based on the result being
# reported:
my %COLORS = (
    PASS   => '#aaffaa',
    FAIL   => '#ff8888',
    SCRIPT => '#ffcc99',
    'N/A'  => '#777777',
);

# This is a look-up table of writer functions for the formats we support:
my %WRITERS = (
    wiki => \&_write_wiki_report,
    html => \&_write_html_report,
);

# Note that all of these elements are all named as private attributes, thus
# ensuring that the methods C::MM generates are also named as private. We
# will define our own public accessors that are read-only for users to call.
# Also, I will be providing an explicit new() constructor, so that users can
# use the "visible" accessor names as options, while the data is stored under
# the private names.
use Class::MethodMaker [
    scalar => '_format',
    scalar => '_document_base',
    scalar => '_username',
    scalar => '_password',
    scalar => '_wiki_url',
];

# These are the publically-visible accessors, all read-only:
for my $method (qw(format document_base username password wiki_url)) {
    ## no critic(ProhibitNoStrict)
    no strict 'refs';
    my $private = "_$method";
    *{$method} = sub {
        my $self = shift;

        return $self->$private();
    };
}

=head1 NAME

ExecutionReportWriter - Write out execution reports from data

=head1 SYNOPSIS

    my $writer = ExecutionReportWriter->new(
        format        => 'wiki',
        document_base => $base,
        username      => $wikid_user,
        password      => $wikid_pass
    );
    $writer->write_report(
        data  => $data,
        file  => $file,
        title => '%N: Execution report for %C'
    );

=head1 DESCRIPTION

This class provides the report-writing functionality for the Execution Report
system. It takes a data object of the ExecutionReportData class and formats the
content into an HTML table, either writing this as an HTML file or uploading it
to wikid.

An object of this class does not keep any data instances associated with it.
Also, all of its attributes are read-only once the object is created. To have
a writer object with different attributes, you will have to create a new
object.

=head1 ACCESSORS

The following property accessors are defined for this class. Each of these is
a read-only accessor that cannot change the value of the underlying property.

=over

=item B<format>

The format of output that this writer generates, will be one of C<wiki> or
C<html>. The tables created are in fact identical, the only difference between
the wiki and pure HTML formats is the rest of the "page" surrounding the
table.

=item B<document_base>

The document base is the base directory/location where reports are written. When
B<format> is C<wiki>, this is the leading part of the wikid document location.
When B<format> is C<html>, this is just the leading directory part of the path
to the final file being written. May be null if no document base is needed.

=item B<username>

=item B<password>

=item B<wiki_url>

When B<format> is C<wiki>, these three values control the interface with the
wikid MediaWiki instance. B<username> and B<password> are the user/password
pair that are used to log in to the publishing API. B<wiki_url> is the URL of
the publishing API for the wikid server.

=back

=head1 METHODS

The class provides the following methods. No subroutines are exported by this
package.

=head2 new

    my $writer = ExecutionReportWriter->new(
        format        => 'wiki',
        document_base => $base,
        username      => $wikid_user,
        password      => $wikid_pass
    );

Create a new report-writer object and return it to the caller.

The following options are recognized by B<new>:

=over

=item B<format>

Required. A string value that selects the format of reports to write. Must be
one of C<wiki> or C<html>. Case is not important.

=item B<document_base>

Optional. A string/path value that tells the writer where to write the files
or wiki pages that are requested. If B<format> is C<wiki>, this cannot be an
absolute path, that is, it cannot start with a slash (C</>). There is no such
restriction for C<html>. This value may contain expandable tokens that start
with C<%>, followed by a single character. These are used to insert time/date
information into the value of the string at the time of writing a report. See
L</TOKEN EXPANSION>, below.

=item B<username>

Only required if B<format> is C<wiki>. The username to use when connecting to
the MediaWiki publishing API.

=item B<password>

Only required if B<format> is C<wiki>. The password to use when connecting to
the MediaWiki publishing API.

=item B<wiki_url>

Optional. If passed, overrides the default URL for the MediaWiki publishing
API.

=back

The following exception may be thrown by B<new>:

=over

=item B<NATE::Exceptions::Argument>

This exception will be thrown if any of the known options are given invalid
values, or if any unknown option is passed. It will also be thrown if the
value of B<format> is not a known format, or if B<document_base> is not valid
for the format of the writer.

=back

=cut

###############################################################################
#
#   Sub Name:       new
#
#   Description:    Create a new ExecutionReportWriter instance. Validates and
#                   stores the parameters given for later use.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $class    in      scalar    Class this constructor was
#                                                 called in
#                   @opts     in      list      Additional arguments to ctor
#
#   Globals:        %TEMPLATES
#                   %WRITERS
#
#   Returns:        Success:    new object
#                   Failure:    throws exception
#
###############################################################################
sub new {
    my ($class, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            format        => { type => SCALAR },
            document_base => $OPTIONAL_SCALAR,
            username      => $OPTIONAL_SCALAR,
            password      => $OPTIONAL_SCALAR,
            wiki_url      => $OPTIONAL_SCALAR,
        }
    );
    my $self = bless {}, $class;
    my $format = lc $opts{format};
    if (! $WRITERS{$format}) {
        NATE::Exceptions::Argument->throw("Format '$format' is not known");
    }
    if ($opts{document_base}) {
        if (($format ne 'html') && ($opts{document_base} =~ m{^/})) {
            NATE::Exceptions::Argument->throw(
                q('document_base' can only be an absolute path when 'format' ) .
                'is "html"'
            );
        }
        $self->_verify_path($opts{document_base});
    }

    # If this is the first time through, pre-load and pre-compile the templates:
    if (! %TEMPLATES) {
        $self->_compile_templates;
    }

    $self->_format($format);
    $self->_document_base($opts{document_base} || q{});
    $self->_username($opts{username} || q{});
    $self->_password($opts{password} || q{});
    $self->_wiki_url($opts{wiki_url} || $WIKI_API_URL);

    return $self;
}

=head2 write_report

    $writer->write_report(
        data  => $data,
        file  => $file,
        title => '%N: Execution report for %C'
    );

Write a report for the given data to the file (or wikid location) given.

Formats the content of the data into an HTML table, then wraps it in the
needed additional content for either a raw HTML file or a MediaWiki page.
For MediaWiki, also uploads the content to the page.

Returns the calling object upon success, or throws an exception in case of
errors.

The following options are recognized by B<write_report>:

=over

=item B<data>

The data to generate the report from. Must be an object of the
B<ExecutionReportData> class.

=item B<file>

The file (or wikid location) to write the report to. This may contain directory
elements, but must always be a relative path. If an absolute path is needed for
raw HTML files, use B<document_base> in the constructor to specify the leading
path components. This value may contain expandable tokens that start with "%",
followed by a single character. These are used to insert time/date information
into the value of the string at the time of writing a report. See L</TOKEN
EXPANSION>, below.

=item B<title>

The title of the report. Used in the page, and for raw HTML used to set the
C<title> element of the header as well. May also contain expandable tokens,
as with B<file>.

=back

The following exceptions may be thrown by <write_report>:

=over

=item B<NATE::Exceptions::Argument>

This exception will be thrown if an unknown option is passed, or if a valid
option is passed but with an invalid value.

=item B<NATE::BaseException>

This exception will be thrown if anything else happens, such as file-access
issues, etc.

=back

=cut

###############################################################################
#
#   Sub Name:       write_report
#
#   Description:    Write a single report, using the ExecutionReportData
#                   instance passed in.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional arguments
#
#   Globals:        %WRITERS
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub write_report {
    my ($self, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            data  => { type => OBJECT, isa => 'ExecutionReportData' },
            file  => { type => SCALAR },
            title => $OPTIONAL_SCALAR,
        }
    );

    my $file = $opts{file};
    if ($file =~ m{^/}) {
        NATE::Exceptions::Argument->throw(
            "'file' argument ($file) must be a relative path"
        );
    } else {
        # Even if it's relative, still need to check for /../, etc.
        $self->_verify_path($file);
    }
    my $data = $opts{data};
    my $table = $self->_create_table($data);
    my $timestamp = $data->datetime;
    my $title = $opts{title};
    if (! $title) {
        $title = '%N Execution Report: %C';
    }
    my $base_dir = $self->_document_base;
    my $format = $self->_format;

    my $name = $data->name;
    for ($title, $file, $base_dir) {
        # "%N" is a special marker for the project name, expand it before
        # handing off to time2str:
        s/%N/$name/g;
        $_ = time2str($_, $timestamp);
    }

    if (my $writer = $WRITERS{$format}) {
        $self->$writer($title, $file, $base_dir, $table);
    } else {
        NATE::BaseException->throw(
            "Format '$format' is not known to the report writer"
        );
    }

    return $self;
}

###############################################################################
#
# Private methods
#
###############################################################################

###############################################################################
#
#   Sub Name:       _compile_templates
#
#   Description:    Load all templates from the <DATA> handle and compile
#                   them into the global %TEMPLATES.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      any       Ref or class name, not used.
#
#   Globals:        %TEMPLATES
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub _compile_templates {
    my ($self) = @_;
    my (@content, $format, $name);

    while (defined(my $line = <DATA>)) {
        if ($line =~ /^TEMPLATE\s+format=(.*?)\s+name=(.*)$/) {
            # If we match this line, this is the start of a template. Save the
            # match values, then if there is existing content create a template
            # from it and save it under the current format/name. Then clear the
            # @content accumulator and assign new $format and $name from the
            # matched values.

            my $new_format = $1;
            my $new_name   = $2;
            chomp($new_name);

            if (@content) {
                $TEMPLATES{$format}->{$name} = Text::Template->new(
                    TYPE       => 'ARRAY',
                    SOURCE     => [ @content ],
                    DELIMITERS => [ q({%), q(%}) ]
                );
            }

            @content = ();
            $format  = $new_format;
            $name    = $new_name;
        } else {
            # Otherwise it's just a line in the current template
            push(@content, $line);
        }
    }
    # Get the last template in, if there was any content to it:
    if (@content) {
        $TEMPLATES{$format}->{$name} = Text::Template->new(
            TYPE       => 'ARRAY',
            SOURCE     => [ @content ],
            DELIMITERS => [ q({%), q(%}) ]
        );
    }

    return;
}

###############################################################################
#
#   Sub Name:       _create_table
#
#   Description:    Create the HTML table from the passed-in data object.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $data     in      ref       ExecutionReportData instance
#
#   Globals:        %COLORS
#
#   Returns:        Success:    table text
#                   Failure:    throws exception
#
###############################################################################
sub _create_table {
    my ($self, $data) = @_;

    my @rows;
    my %variants;
    my $hdr_row   = '<th></th>';
    my $workflows = $data->workflows;
    my $streams   = $data->streams;
    my $run_data  = $data->data;

    my %html_ent = (
        q{&} => '&amp;',
        q{<} => '&lt;',
        q{>} => '&gt;',
    );

    for my $workflow (@{$workflows}) {
        (my $name = $workflow->{name}) =~ s/([&<>])/$html_ent{$1}/ge;
        $name =~ s/\s+/&nbsp;/g;
        $hdr_row .= "<th>$name</th>";
    }
    push(@rows, $hdr_row);

    # This will tell us whether a given variant appears in the table more than
    # once. Those that do will get the I/O added to their label, to tell them
    # apart.
    for my $stream (@{$streams}) {
        $variants{$stream->{variant}}++;
    }

    for my $idx (0 .. $#{$run_data}) {
        my $row;
        my $label;

        # Do the variant-label in the first column:
        if ($variants{$streams->[$idx]->{variant}} > 1) {
            $label = "$streams->[$idx]->{variant} ($streams->[$idx]->{io})";
        } else {
            $label = $streams->[$idx]->{variant};
        }
        $label =~ s/([&<>])/$html_ent{$1}/ge;
        $label =~ s/\s+/&nbsp;/g;
        $row = qq(<th style="text-align:left">$label</th>);

        # Now do the workflow columns
        for my $wf_idx (0 .. $#{$workflows}) {
            $label = uc $run_data->[$idx]->[$wf_idx] || 'N/A';
            my $color = $COLORS{$label};
            $row .=
                qq(<td style="text-align:center;background:$color">$label</td>);
        }

        # Put it on the list
        push(@rows, $row);
    }

    # Turn @rows into actual <tr>'s
    for (@rows) {
        $_ = "<tr>$_</tr>";
    }

    # Build in the rest of the table
    unshift(
        @rows,
        '<table border="1" cellpadding="5" cellspacing="0">',
        '<tbody>'
    );
    push(@rows, '</tbody>', '</table>');

    return join("\n", @rows);
}

###############################################################################
#
#   Sub Name:       _write_wiki_report
#
#   Description:    Write a MediaWiki report, using the data.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $title    in      scalar    Title of the report
#                   $file     in      scalar    File name (location within wiki)
#                   $dir      in      scalar    Section name for wiki
#                   $table    in      scalar    Pre-formatted table data for
#                                                 the report
#
#   Globals:        %TEMPLATES
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub _write_wiki_report {
    my ($self, $title, $file, $dir, $table) = @_;

    my $template = $TEMPLATES{wiki}->{main};
    if (! $template) {
        NATE::BaseException->throw(
            'No template found for creating a wiki report, cannot continue'
        );
    }
    if ($dir =~ m{^/}) {
        NATE::BaseException->throw(
            "Value for 'document_base' ($dir) is not valid for wiki format." .
            ' Value cannot not be an absolute path.'
        );
    }
    my $path = $dir ? "$dir/$file" : $file;
    if (! ($self->_username && $self->_password)) {
        NATE::BaseException->throw(
            'Writing a report to the wiki requires both username and password'
        );
    }
    my $text = $template->fill_in(
        HASH => {
            title => $title,
            table => $table,
        }
    );
    if (! $text) {
        NATE::BaseException->throw(
            "Error rendering wiki template 'main': $Text::Template::ERROR"
        );
    }

    my $mw = MediaWiki::API->new;
    $mw->{config}->{api_url} = $self->_wiki_url;
    $mw->{config}->{on_error} = sub {
        NATE::BaseException->throw(
            "MediaWiki error: Code $mw->{error}->{code}, details " .
            $mw->{error}->{details}
        );
    };
    $mw->login({
        lgname     => $self->_username,
        lgpassword => $self->_password,
    }) or NATE::BaseException->throw(
        "MediaWiki login error: $mw->{error}->{code}: $mw->{error}->{details}"
    );

    # Try to fetch the page. If it exists, we still are completely overwriting
    # it. But we'll want the current timestamp for the edit() call, to prevent
    # conflicts.
    my $pageref = $mw->get_page({ title => $path });

    # Create/update the page
    $mw->edit({
        action => 'edit',
        title => $path,
        basetimestamp => $pageref->{timestamp},
        text => $text,
    }) or NATE::BaseException->throw(
        "MediaWiki edit error: $mw->{error}->{code}: $mw->{error}->{details}"
    );

    # That's all there is to it...
    $mw->logout;

    return $self;
}

###############################################################################
#
#   Sub Name:       _write_html_report
#
#   Description:    Write an HTML report to a file, using the data
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $title    in      scalar    Title of the report
#                   $file     in      scalar    File name
#                   $dir      in      scalar    Dir into which to write $file
#                   $table    in      scalar    Pre-formatted table data for
#                                                 the report
#
#   Globals:        %TEMPLATES
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub _write_html_report {
    my ($self, $title, $file, $dir, $table) = @_;

    my $template = $TEMPLATES{html}->{main};
    if (! $template) {
        NATE::BaseException->throw(
            'No template found for creating a HTML report, cannot continue'
        );
    }
    my $path = $dir ? "$dir/$file.html" : "$file.html";
    my $text = $template->fill_in(
        HASH => {
            title => $title,
            table => $table,
        }
    );
    if (! $text) {
        NATE::BaseException->throw(
            "Error rendering HTML template 'main': $Text::Template::ERROR"
        );
    }

    if (open(my $fh, '>', $path)) {
        print {$fh} $text;
        if (! close($fh)) {
            NATE::BaseException->throw("Error closing $path: $!");
        }
    } else {
        NATE::BaseException->throw("Error opening $path for writing: $!");
    }

    return $self;
}

###############################################################################
#
#   Sub Name:       _verify_path
#
#   Description:    Validate the given path, to make sure it does not start
#                   with a slash or contain any "/../" sequences, etc.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref/sc    Object or class name
#                   $path     in      scalar    Path to check
#
#   Returns:        Success:    path value
#                   Failure:    throws exception
#
###############################################################################
sub _verify_path {
    my ($self, $path) = @_;

    if ($path =~ m{^[.]{1,2}/}) {
        NATE::Exceptions::Argument->throw(
            q(Path must not start with a '../' or './' sequence)
        );
    }
    if ($path =~ m{/[.]{1,2}/}) {
        NATE::Exceptions::Argument->throw(
            q(Path must not contain '/../' or '/./' sequences)
        );
    }

    return $path;
}

=head1 TOKEN EXPANSION

The C<document_base> parameter to B<new> and the C<file> and C<title>
parameters to B<write_report> all allow for token expansion in their values.
This allows the location of reports, the report names and the titles within
the reports to all be responsive to the time and date of the execution being
reported on.

The B<ExecutionReportWriter> class uses the B<Date::Format> module to do the
expansion of these date/time tokens into values taken from the timestamp on
the data itself. Each token is of the form:

    %x

where C<%> is the percent sign, always present, and C<x> is a single letter.
The various tokens are described under the heading of "Conversion
Specification" in the B<Date::Format> manual page.

For example, a C<document_base> value of:

    Execution_Reports_%Y%m%d

will result in a directory component that always has the current date
embedded in it. For February 28, 2013, it would be expanded as:

    Execution_Reports_20130228

There is one addition to the formats offered by B<Date::Format>: a token of
C<%N> will be replaced by the project name, allowing the name to be
incorporated into the base, name or title. In fact, the default title string
is expressed this way, as the string:

    %N Execution Report: %C

This yields a title with the name, and with the full UNIX-style date string. If
the project name were "FLO", and the full time and date were 2:39:33PM on
2/28/2013, then the title would be:

    FLO Execution Report: Thu Feb 28 14:39:33 EST 2013

(This assumes that the timezone for the machine running the report script is
in the Eastern time zone.)

=head1 AUTHOR

Randy J.Ray E<lt>rjray@netapp.comE<gt>

=cut

1;

__DATA__
TEMPLATE format=wiki name=main
'''<span style="font-size: 180%">{% $title %}</span>'''

<html>
<div style="margin-top:2em">
{% $table %}
</div>
</html>
