# $Id: //depot/prod/test/Rsierranevada/dp/snapmirror/paloma/rapid/RRS_IO.pm#3 $
#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

## @summary Rapid Regression Suite I/O operations
## @author rjray@netapp.com
## @status shared
## @pod here

package RRS_IO;

use strict;
use warnings;

use Params::Validate qw(
    validate validate_with validation_options SCALAR HASHREF OBJECT BOOLEAN
);
use Readonly;
use List::MoreUtils 'any';

use Tharn qw($Log);
use NACL::STask::VolumeFile;
use NACL::C::NetworkInterface;
use NACL::C::Slist;
use NACL::C::Client;
use NACL::C::Client::Sanlun;
use NACL::STask::Client::Mount;
use NACL::STask::Lun;
use NACL::STask::VserverCifsShare;
use NACL::MTask::Data;
use NACL::MTask::Traffic;
use NACL::ComponentUtils qw(_verify_invocation);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;

use RRS_Config;

# The default size of populate data, used unless the user specifies an
# explicit value.
Readonly my $DEFAULT_POPULATE_SIZE => '1g';

Readonly my %FILE_SIZES => (
    'k' => 1024,
    'm' => 1024 * 1024,
    'g' => 1024 * 1024 * 1024,
);

# Params::Validate fragments to avoid re-typing the same specs over and over:
Readonly my $OPTIONAL_SCALAR => {
    type     => SCALAR,
    optional => 1,
};
Readonly my $OPTIONAL_HASHREF => {
    type     => HASHREF,
    optional => 1,
};
Readonly my $OPTIONAL_BOOLEAN => {
    type     => BOOLEAN,
    optional => 1,
};

# Note that all of these elements are all named as private attributes, thus
# ensuring that the methods C::MM generates are also named as private. We
# will define our own public accessors that are read-only for users to call.
# Also, I will be providing an explicit new() constructor, so that users can
# use the "visible" accessor names as options, while the data is stored under
# the private names.
use Class::MethodMaker [
    scalar => [
        { -type => 'NACL::C::Client' },
        '_client',
    ],
    scalar => [
        { -type => 'NACL::STask::Volume' },
        '_volume',
    ],
    scalar => [
        {
            -type    => 'NACL::STask::Client::Mount',
            -forward => [ 'mount_point' ],
        },
        '_mount',
    ],
    scalar => [
        {
            -type => 'NACL::STask::Lun'
        },
        '_lun',
    ],
    scalar => '_mount_access',
    scalar => '_traffic',
    scalar => '_nfs',
    scalar => '_cifs',
    scalar => '_san',
];

# These are the publically-visible accessors, all read-only:
for my $method (qw(client volume mount mount_access nfs cifs san)) {
    ## no critic(ProhibitNoStrict)
    no strict 'refs';
    my $private = "_$method";
    *{$method} = sub {
        $Log->enter();
        my $self = shift;

        $Log->exit();
        return $self->$private();
    };
}

# Use this to store a persistent RRS_Config instance, so that we only have to
# read the config file once, even if new() is called multiple times:
my $CONFIG;

# Set the on_fail behavior for Params::Validate to throw an exception (of the
# NATE::Exceptions::Argument class) on errors instead of a plain croak().
validation_options(
    on_fail => sub {
        NATE::Exceptions::Argument->throw(shift);
    }
);

=head1 NAME

RRS_IO - Data I/O generation for the Rapid Regression Suite

=head1 SYNOPSIS

    use RRS_IO;

    # Create a handle on I/O generation:
    $io_handle = RRS_IO->new(
        volume  => $volume,
        client  => $client,
        variant => $variant
    );

    # Create the mount-point:
    $io_handle->create_mount(
        nfs_options => $opts,
        rw          => 1
    );

    # Populate some data:
    $io_handle->populate(size => '3g');

    # Unmount the volume:
    $io_handle->delete_mount();

=head1 DESCRIPTION

The B<RRS_IO> class is used to do data population on volumes within the Rapid
Regression Suite lifecycle scripts. It uses data from the variant specification
and client specification to determine whether to use NFS, CIFS or SAN protocols
for the data. Ideally, the lifecycle script should have no inherent knowledge
of whether it is generating traffic via NFS, CIFS or SAN.

=head1 ACCESSORS

This class has the following property accessors. Each of these is a read-only
accessor, that cannot change the value of the underlying property. All of the
properties are read-only; to operate on a different volume or with a different
client, you will have to create a new B<RRS_IO> object.

=over

=item client

The B<NACL::C::Client> object that represents the client this object was
instantiated with. If this object was only passed a client name, this property
will still return an object referent.

=item volume

The B<NACL::STask::Volume> object that represents the volume this object was
instantiated with.

=item mount

The B<NACL::STask::Client::Mount> object that represents the current
mount-point for the client/volume pair contained within this object. If there
is no current mount-point, this accessor returns C<undef>.

=item nfs

A boolean value that tells whether the mount-point and I/O are done with NFS
(as opposed to CIFS).

=item cifs

A boolean value that tells whether the mount-point and I/O are done with CIFS
(as opposed to NFS).

=item san

A boolean value that tells whether the volume being used is using SAN protocol
or not.

=back

=head1 METHODS

This class provides the following methods. No subroutines are exported by this
package.

=head2 new

    $io_handle = RRS_IO->new(
        volume  => $volume_obj,
        client  => $client_name_or_obj,
        variant => $variant_name_or_hashref
    );

(Must be called as a class method)

Create a new I/O handle object and return it to the caller. Does not create any
mount-points or CIFS shares, SAN LUNs, etc. Derives from the arguments whether
the eventual data population will be via NFS, CIFS, etc.

All failures in B<new> throw an exception. See C<Exceptions>, below.

=over

=item Options

The following options are recognized by B<new>:

=over

=item C<< volume =E<gt> $volume_object >>

(Required) A volume object that represents the volume that will receive the
generated data. This must be either a B<NACL::C::Volume> or a
B<NACL::STask::Volume> instance. Additionally, it should already be mounted on
the filer itself. If the volume is a source volume created by the B<RRS_Volume>
package, it will have been mounted for you. If it was created outside of
B<RRS_Volume>, or if it is a volume that was previously a DP volume that has
been converted to RW by means of a C<snapmirror break> operation, then you will
have to mount it explicitly prior to passing to this constructor.

=item C<< client =E<gt> $client_object_or_name >>

(Required) A client object or the name of a client. If a name is passed, a
client object will be created for you. If an object is passed, it must be
either a B<NACL::C::Client> instance or a B<NACL::STask::Client> instance.
This represents the client that will mount the volume and perform the actual
data population.

=item C<< variant =E<gt> $variant_name_or_hashref >>

(Required) The name of the variant that data will be generated for. The name is
used to retrieve a record from the configuration table encapsulated by the
B<RRS_Config> class. If the value passed in is not found in the configuration
table an exception is thrown.

Alternately, the value may be the data returned by a previous call to the
B<get_variant> method of B<RRS_Config>. If the passed-in value is a
hash-reference it is assumed to be the result of such a call and is used
directly.

=item C<< configuration_file =E<gt> $file_name >>

(Optional) If the B<RRS_Config> class should read the configuration from some
place other than the default, pass that file name here. This value is not saved
on the B<RRS_IO> object, it is only used to retrieve the B<RRS_Config>
object. This value is subject to the same usage and consideration as the
C<file> option to the the B<new> method in B<RRS_Config>. In fact, it is passed
as the value to that option. This means that if it is not an absolute path, it
will be treated as relative to the the directory that these modules reside in.

=back

=item Exceptions

The following exceptions may be thrown by B<new> in cases of failure:

=over

=item B<NATE::Exceptions::Argument>

If a required parameter is missing, or if the value of a paramter is invalid,
this exception will be thrown.

=item B<NATE::BaseException>

If there is a problem in loading the configuration file, or if the variant
passed in is not present in the configuration file, this exception will be
thrown.

This exception may also be thrown in cases where the variant specifies SAN
I/O, but the client is of a type other than Linux. At present, SAN I/O is
only supported with Linux clients.

=back

=back

=cut

###############################################################################
#
#   Sub Name:       new
#
#   Description:    Create a handle object to be used for the other method
#                   calls.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $class    in      scalar    Class we're blessing into
#                   @opts     in      list      Additional options
#
#   Globals:        $CONFIG
#
#   Returns:        Success:    object
#                   Failure:    throws exception
#
###############################################################################
sub new {
    $Log->enter();
    my ($class, @opts) = @_;
    # This must be a static method invocation:
    $class->_verify_invocation(style => 'static_only');
    my %opts = validate(
        @opts,
        {
            volume => {
                type => OBJECT,
                isa  => 'NACL::C::Volume',
            },
            client => {
                type      => SCALAR | OBJECT,
                callbacks => {
                    'correct object type' => sub {
                        my $obj = shift;
                        # We know from type-validation that this is either a
                        # plain scalar or a blessed ref:
                        return (ref $obj) ? $obj->isa('NACL::C::Client') : 1;
                    },
                },
            },
            variant => { type => SCALAR | HASHREF },
            configuration_file => $OPTIONAL_SCALAR,
        }
    );
    my $self = bless {}, $class;
    if (! ref($opts{client})) {
        $opts{client} = NACL::C::Client->new(name => $opts{client});
    }
    $self->_volume($opts{volume});
    $self->_client($opts{client});

    # Variant is either a scalar or a hashref.
    if (ref $opts{variant}) {
        # If it is a hashref, it should be the result of a get_variant() call
        # on a RRS_Config object. Then, just access the "san" key out of it
        # directly.
        $self->_san($opts{variant}->{san});
    } else {
        # If not a ref, then it's a scalar. We have to use it to pull a record
        # out of the config.

        # Read the config if we haven't already:
        if (! $CONFIG) {
            my @conf_opts = $opts{configuration_file} ?
                (file => $opts{configuration_file}) : ();
            $CONFIG = RRS_Config->new(@conf_opts);
        }

        # Verify the value of $opts{variant}. This will throw an exception if
        # the value is bad:
        my $variant_data = $CONFIG->get_variant(variant => $opts{variant});
        $self->_san($variant_data->{san});
    }
    if ($opts{client}->is_windows) {
        $self->_cifs(1);
        $self->_nfs(0);
    } else {
        $self->_cifs(0);
        $self->_nfs(1);
    }

    # Sanity check: At present (9/26/2012), we cannot do SAN on Windows, only
    # on Linux.
    if ($self->_san && $opts{client}->type !~ /linux/) {
        $Log->exit();
        NATE::BaseException->throw(
            'SAN I/O can currently only be done on linux clients'
        );
    }

    # Init the _traffic listref.
    $self->_traffic([]);

    $Log->exit();
    return $self;
}

=head2 create_mount

    $io_handle->create_mount(
        nfs_options  => $nfs_params_hash,
        cifs_options => $cifs_params_hash,
        san_options  => $san_params_hash,
        access       => 'RW',
        is_ls        => 0
    );

(Must be called as an instance method)

Create the mount-point for the combination of volume and client that are
encapsulated in this object. Whether the mount point is for NFS or CIFS is
dependent on whether the client is a Windows host or not.

If there is an existing mount for this client/volume pair, it is removed and
a new one created. This is to prevent mounts that are read-write or read-only
from interfering in an attempt to create the opposite type.

If the operation is successful, the calling object is returned. Otherwise, an
exception is thrown.

=over

=item Options

The following options are recognized by this method. Note that certain options
will be silently ignored if they are not applicable; for example, the
C<nfs_options> value is ignored when creating a CIFS mount. This allows all
options to be passed all the time, without causing errors.

=over

=item C<< nfs_options =E<gt> $hashref >>

(Optional) Must be a hash-reference of the NFS-related options from the
B<NACL::C::Client::Mount::create()> method. These options are:

    proto ('tcp')
    vers (3)
    mountopts ('hard,intr')

Any of these three that are provided will be passed unchanged to the call to
create a NFS mount. Any keys other than these three will be ignored. Each of
these three has a default value (shown with the option, above). If the option
is not passed at all, the default value will be used. If you wish to suppress
the default value without providing a new value in the process, pass C<undef>
as a value:

    $io_handle->create_mount(
        # Other two NFS options will use their defaults:
        nfs_options => { mountopts => undef },
        ...
    );

=item C<< cifs_options =E<gt> $hashref >>

(Optional) Options to be used in the creation and mounting of a CIFS share
prior to mounting on a Windows client. This option is only used when the client
is a Windows machine. Must be a hash-reference. Unlike C<nfs_options>, above,
the different keys are used at different stages. The recognized keys are:

    user ('labclient')
    password ('sundance')
    properties (no default)

Each parameter has a default value shown above in parentheses (except for
C<properties>, which has no default). Any keys other than these three will be
ignored.

The value of the C<properties> option is passed directly as the
C<share-properties> option to B<NACL::C::VserverCifsShare::create()>, but only
if the share does not already exist (in which case the call to create is not
made). The C<user> and C<password> options are used when actually mounting the
share on the Windows client.

=item C<< san_options =E<gt> $hashref >>

Extra options for the mounting of an existing SAN LUN as a filesystem for
data population. The LUN creation should have been done by B<RRS_Volume> when
the volume was initially created. The only option used by B<create_mount>
is:

    fstype ('ext2')

This specifies the type of the filesystem that was created when the volume
and LUN were initially created. It defaults to C<ext2>.

Note that you may still need to pass C<nfs_options> or C<cifs_options>, to
control the actual mounting of the filesystem (depending on whether the
client is Windows or UNIX/Linux).

=item C<< access =E<gt> 'RW'|'RO' >>

(Optional) Determines whether the mount is created as read-only or read-write.
The default is C<RW> (read-write), and the value is treated as case-insensitive
so C<rw> is also acceptable.

=item C<< is_ls =E<gt> 0|1 >>

(Optional) A boolean value that tells whether the volume being mounted is
part of a LS mirror relationship. This affects the path used to mount the
volume when mounting it as read-write. The default is C<0> (false), meaning
that the volume is assumed to B<not> be part of a LS relationship.

=back

=item Exceptions

If an invalid value is passed for any of the options, an exception of the
class B<NATE::Exceptions::Argument> will be thrown.

Any of the operations for creating a CIFS share, SAN LUN, or mounting any of
these (including NFS) may throw exceptions. See the related manual pages for
the exceptions they can throw. If you need to catch an exception from this
method, it is probably best to catch B<NATE::BaseException> as an umbrella for
anything that could be thrown by the various libraries being used.

=back

=cut

###############################################################################
#
#   Sub Name:       create_mount
#
#   Description:    Create a mount-point for this object, based on the volume
#                   and client stored within. May also involve creating a
#                   CIFS share for the volume, if one does not already exist.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      array     Additional options, converted
#                                                 to a hash by validate()
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub create_mount {
    $Log->enter();
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate(
        @opts,
        {
            nfs_options  => $OPTIONAL_HASHREF,
            cifs_options => $OPTIONAL_HASHREF,
            san_options  => $OPTIONAL_HASHREF,
            is_ls        => $OPTIONAL_BOOLEAN,
            access       => {
                type      => SCALAR,
                default   => 'RW',
                callbacks => {
                    value => sub {
                        my $value = shift;
                        return ($value =~ /^r[wo]$/i);
                    },
                },
            },
        }
    );

    # These are the default options for NFS. Anything the user passes
    # overrides these, they do not augment these. SAN may add to these.
    my $nfs_options = {
        proto     => 'tcp',
        vers      => 3,
        mountopts => 'hard,intr',
    };
    # These are the default options for CIFS. Anything the user passes
    # overrides these, they do not augment these. SAN may (eventually) add
    # to these (when SAN on Windows is supported).
    my $cifs_options = {
        user       => 'labclient',
        password   => 'sundance',
        properties => q{},
    };

    # This does nothing if the volume is already mounted. If it isn't, it
    # creates a junction-point based on the volume's name and mounts it:
    $self->_volume->mount;

    # For SAN, we have to set up some additional arguments and framework
    # before the actual call to mount or share is made.
    if ($self->_san) {
        if (! $self->_lun) {
            # The LUN has not been "found" and cached on $self yet, so do so.
            # This is the filer-side LUN object, not the client-side.
            $self->_find_lun(options => $opts{san_options});
        }

        my $lun     = $self->_lun;
        my $lunpath = $lun->path;
        my $vserver = $self->_volume->vserver;

        if ($self->_nfs) {
            # Determine the device that this LUN is using on the client, so
            # we can mount as device instead of straight NFS. Do this by
            # using properties of the filer-side LUN object to locate the
            # matching client-side LUN and getting the device_filename from it.
            my $client_lun = NACL::C::Client::Sanlun->find(
                command_interface => $self->_client,
                filter            => {
                    controller   => $vserver,
                    lun_pathname => $lunpath,
                }
            );
            $nfs_options->{path}        =
                $client_lun->get_one_state_attribute('device_filename');
            $nfs_options->{type}        = 'device';
            $nfs_options->{mountopts}   = q{}; # Null for device mounts
            $nfs_options->{fstype}      =
                $opts{san_options}->{fstype} || 'ext2';
            $nfs_options->{mount_point} = $self->_derive_lun_mountpoint(
                type => 'nfs',
                lun  => $lun
            );
        } else {
            $Log->exit();
            NATE::BaseException->throw(
                'SAN I/O not yet supported for CIFS clients'
            );
        }
    }

    # Delegate to a private sub based on NFS or CIFS
    if ($self->_nfs) {
        # Copy over the NFS opts we recognize, discarding anything else:
        for my $opt (qw(proto vers mountopts)) {
            if (exists $opts{nfs_options}->{$opt}) {
                $nfs_options->{$opt} = $opts{nfs_options}->{$opt};
            }
        }

        $self->_create_nfs_mount(
            options => $nfs_options,
            access  => lc $opts{access},
            is_ls   => $opts{is_ls}
        );
    } elsif ($self->_cifs) {
        # Copy over the CIFS opts we recognize, discarding anything else:
        for my $opt (qw(user password properties)) {
            if (exists $opts{cifs_options}->{$opt}) {
                $cifs_options->{$opt} = $opts{cifs_options}->{$opt};
            }
        }

        $self->_create_cifs_mount(
            options => $cifs_options,
            access  => lc $opts{access},
            is_ls   => $opts{is_ls}
        );
    } else {
        $Log->exit();
        NATE::BaseException->throw(
            'Sanity failure! Object somehow has neither nfs or cifs set!'
        );
    }
    # If we didn't throw an exception by this point, then _mount is set. Also
    # set _mount_access, as well.
    $self->_mount_access(lc $opts{access});

    $Log->exit();
    return $self;
}

=head2 delete_mount

    # Delete an existing mount
    $io_handle->delete_mount;

    # Delete with the "-f" (force) option passed:
    $io_handle->delete_mount(force => 1);

(Must be called as an instance method)

Delete a mount that was created by B<create_mount>. Removes the mount-point
on the client, but does not do an "unmount" of the volume itself at the filer
level.

If the operation is successful, the calling object is returned. Otherwise, an
exception is thrown.

=over

=item Options

The following option is the only recognized option to B<delete_mount>:

=over

=item C<< force =E<gt> 0|1 >>

(Optional) A boolean value that determines whether to "force" the unmount.
This maps to the C<force> option to B<NACL::C::Client::Mount::delete()>,
which is used for both NFS and CIFS mounts.

=back

=item Exceptions

If an invalid value is passed for C<force>, or if an unknown option is passed,
an exception of the type B<NATE::Exceptions::Argument> will be thrown. If
no mount-point exists currently (i.e., B<create_mount> has not been called) an
exception of the type B<NATE::BaseException> will be thrown.

Additionally, any exceptions that may be thrown by
B<NACL::STask::Client::Mount> may be thrown here, as they will be passed along
unchanged if they occur.

=back

=cut

###############################################################################
#
#   Sub Name:       delete_mount
#
#   Description:    Remove a mount-point that exists for the volume/client
#                   stored within this object. Remove even if we didn't create
#                   it ourselves.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      array     Additional options, converted
#                                                 to a hash by validate()
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub delete_mount {
    $Log->enter();
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate(
        @opts,
        {
            force        => $OPTIONAL_BOOLEAN,
            stop_traffic => $OPTIONAL_BOOLEAN,
        }
    );
    if (! $self->_mount_isset) {
        $Log->exit();
        NATE::BaseException->throw(
            'Cannot delete_mount: no mount-point has been created yet'
        );
    }

    # Can't delete the mount-point if we're currently using it for traffic.
    if ($opts{stop_traffic}) {
        $self->traffic_stop;
    } else {
        if (@{$self->_traffic}) {
            $Log->exit();
            NATE::BaseException->throw(
                'Cannot delete_mount: traffic is currently running'
            );
        }
    }

    $self->_mount->purge($opts{force} ? (force => 1) : ());
    $self->_mount_reset;
    $self->_mount_access_reset;

    $Log->exit();
    return $self;
}

=head2 populate

    # Populate 2Gb of data:
    $io_handle->populate(
        size => '2g',
        path => 'RRS'
    );

    # Start a population but return immediately with a data-handle:
    my $data_handle = $io_handle->populate(
        size  => '2g',
        path  => 'RRS',
        async => 1,
    );
    # Do some stuff, like taking periodic snapshots...
    # Wait for background populate to finish (with 15-min timeout):
    $data_handle->wait_for_completion(timeout => 900);

(Must be called as an instance method)

Populate data on the volume contained by this object, using the client also
contained by this object. The population may be done synchronously or
asynchronously, depending on the user's needs.

The return value is an object reference: either the calling object (when
populating synchronously) or an instance of B<NACL::MTask::Data> (when
populating asynchronously). In case of an error, an exception is thrown.

=over

=item Options

The following options are recognized:

=over

=item C<< size =E<gt> $string >>

(Optional) Specify the size of data to populate. The string should be of the
form, C<D+u>, where "D+" is one or more digits, and "u" is a single-letter unit
representing gigabytes, etc. At present, only C<g> (gigabytes) and C<m>
(megabytes) are supported for the units. The default is C<1g> (one gigabyte).

=item C<< path =E<gt> $string >>

(Optional) Specify the path from the root of the mounted volume where data
population should be rooted. The value is appended to the mount-point of the
volume. The default is no path (an empty string), which populates starting at
the root of the volume.

=item C<< async =E<gt> 0|1 >>

(Optional) Specify whether the population should be done asynchronously or
synchronously, that is, whether it should be backgrounded or not. If the
population is I<not> backgrounded (C<async> = 0), B<populate> will wait until
it is complete before returning. If it I<is> backgrounded (C<async> = 1),
B<populate> will return immediately with an object of the B<NACL::MTask::Data>
class that can be used to wait for its completion or kill it. The default is
C<0>, to do data population synchronously.

=back

=item Exceptions

The following exceptions may be thrown by B<populate>:

=over

=item B<NATE::Exceptions::Argument>

If an unknown argument is passed, or if a bad value for one of the known
arguments is passed, this exception will be thrown.

=item B<NATE::BaseException>

If B<populate> cannot start the population (for example, if you have not yet
called B<create_mount>) this exception will be thrown.

=back

In addition, any exceptions thrown by B<NACL::MTask::Data::populate()> may be
thrown here, as well.

=back

=cut

###############################################################################
#
#   Sub Name:       populate
#
#   Description:    Populate some amount of data on the volume encapsulated by
#                   this object, using the client encapsulated by this object.
#                   There must be an existing mount between the two or this
#                   will throw an exception. If this is done in an async manner
#                   then the return value is a NACL::MTask::Data handle,
#                   otherwise $self is returned.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      array     Additional options, converted
#                                                 to a hash by validate()
#
#   Returns:        Success:    $self or a NACL::MTask::Data instance
#                   Failure:    throws exception
#
###############################################################################
sub populate {
    $Log->enter();
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate(
        @opts,
        {
            size  => {
                type    => SCALAR,
                default => $DEFAULT_POPULATE_SIZE,
            },
            path  => $OPTIONAL_SCALAR,
            async => $OPTIONAL_BOOLEAN,
        }
    );

    # We must have a valid mount-point in order to populate:
    if (! $self->_mount_isset) {
        $Log->exit();
        NATE::BaseException->throw(
            'Cannot call populate() without a valid mount-point set'
        );
    }

    # Start up the data population. The hash of options specifies a 1G and a
    # 1M dataset. Based on the unit of the value passed in, we copy the set we
    # need, then multiply the desired size out on the "file_avg" key, which is
    # the number of files created.
    my %populate_args = ();
    my %populate_args_base = (
        'g' => {
            create_directory => 1,
            file_size_avg    => 65536,
            file_size_dev    => 0,
            depth_avg        => 4,
            depth_dev        => 0,
            dir_avg          => 6,
            dir_dev          => 0,
            file_avg         => 9,
            file_dev         => 0,
            link_num         => 0,
            link_pct         => 0,
            slink_num        => 0,
            slink_pct        => 0,
        },
        'm' => {
            create_directory => 1,
            file_size_avg    => 1048576,
            file_size_dev    => 0,
            depth_avg        => 0,
            depth_dev        => 0,
            dir_avg          => 0,
            dir_dev          => 0,
            file_avg         => 1,
            file_dev         => 0,
            link_num         => 0,
            link_pct         => 0,
            slink_num        => 0,
            slink_pct        => 0,
        },
    );
    if (lc $opts{size} =~ /^(\d+)([mg])b?$/i) {
        # First copy the set of parameters for the appropriate units given:
        %populate_args = %{$populate_args_base{lc $2}};
        # Scale out the size of the dataset by increasing the number of files:
        $populate_args{file_avg} *= $1;
    } else {
        $Log->exit();
        NATE::Exceptions::Argument->throw(
            "Value for 'size' ($opts{size}) is invalid or in units other " .
            'than megabytes (m) or gigabytes (g)'
        );
    }

    # Determine the path to populate at. Start with the mount-point, add "path"
    # if given.
    my $path = $self->mount_point;
    if ($opts{path}) {
        if ($opts{path} !~ m{^/}) {
            $path .= q{/};
        }
        $path .= $opts{path};
    }

    # Start the data population:
    my $data_handle = NACL::MTask::Data->populate(
        command_interface => $self->_client,
        path              => $path,
        async             => $opts{async} || 0,
        template          => 'none',
        %populate_args
    );

    $Log->exit();
    return $opts{async} ? $data_handle : $self;
}

=head2 traffic_start

    # With options to control the traffic:
    my $id = $io_handle->traffic_start(
        traffic_file => $path,
        traffic_tool => 'sio',
        %additional_options
    );

    # Or with all defaults:
    my $id = $io_handle->traffic_start;

(Must be called as an instance method)

Start a traffic stream using the mount-point and client encapsulated by this
IO object. The return value is an integer ID for the stream that was created,
or an exception is thrown in case of error.

This is a wrapper around the B<NACL::MTask::Traffic> library. In theory it
supports all the tools that are supported by that library. However, at the
moment only C<sio> is supported, as there are not defaults defined for the
other tools. This should be augmented at a later date.

=over

=item Options

The following options are recognized:

=over

=item C<< traffic_file =E<gt> string >>

(Optional) The name of the file to use for traffic. If no name is passed, a
random name is generated.

=item C<< traffic_tool =E<gt> string >>

(Optional) The name of the traffic tool to use. Currently defaults to C<sio>,
which is currently the only supported tool.

=back

In addition, any options other than the above are passed directly to the
B<NACL::MTask::Traffic::start> method unchanged. There are some reasonable
defaults defined for each supported tool. Any options passed by the user
will override the default value of the same name.

=item Exceptions

The following exceptions may be thrown by this method:

=over

=item B<NATE::Exceptions::Argument>

If either of the options specified above has a non-scalar value, this
exception will be thrown. It will also be thrown if an unknown value
for C<traffic_tool> is passed in.

=item B<NATE::BaseException>

If the options to the traffic tool require read/write access, but the mount
is read-only, then this exception will be thrown.

=back

In addition, any exceptions thrown by B<NACL::MTask::Traffic::start> will be
allowed to pass through uncaught.

=back

=cut

###############################################################################
#
#   Sub Name:       traffic_start
#
#   Description:    Start a traffic stream, based on the options given. Return
#                   the ID of the new stream so that the caller can later
#                   stop the traffic stream.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional arguments
#
#   Returns:        Success:    integer ID
#                   Failure:    throws exception
#
###############################################################################
sub traffic_start {
    $Log->enter();
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate_with(
        params      => \@opts,
        allow_extra => 1,
        spec        => {
            traffic_file => $OPTIONAL_SCALAR,
            traffic_tool => $OPTIONAL_SCALAR,
        }
    );
    my $traffic = $self->_traffic;
    my $id      = @{$traffic};
    my $tool    = delete $opts{traffic_tool} || 'sio';
    my $file    = delete $opts{traffic_file} ||
        sprintf('wfc_traffic_%d_%d', $$, time);

    # We must have a valid mount-point in order to run traffic:
    if (! $self->_mount_isset) {
        $Log->exit();
        NATE::BaseException->throw(
            'Cannot call traffic_start() without a valid mount-point set'
        );
    }

    # Default options for the supported tools:
    my %default_options = (
        sio => {
            read           => 50,
            rand           => 100,
            block_size     => '4k',
            start_byte     => 0,
            file_size      => '40m',
            num_of_threads => 2,
        },
    );
    if (! $default_options{$tool}) {
        $Log->exit();
        NATE::Exceptions::Argument->throw(
            "Traffic tool $tool not (yet) supported. Supported tools are: " .
            join(', ' => sort keys %default_options)
        );
    }

    # Make sure that the user doesn't try to sneakily override any of these
    # three options when we call NACL::MTask::Traffic::start() later:
    delete @opts{qw(command_interface paths tool)};
    my %traffic_opts = (%{$default_options{$tool}}, %opts);
    # Make sure the mount is OK with the type of traffic we're planning to
    # start:
    my $read_write_traffic =
        ($traffic_opts{read} eq 'random') || ($traffic_opts{read} < 100);
    if ($read_write_traffic && ($self->_mount_access ne 'rw')) {
        $Log->exit();
        NATE::BaseException->throw(
            'Cannot start read/write traffic: mount is read-only'
        );
    }
    # Normalize the value of $traffic_opts{file_size}:
    $traffic_opts{file_size} = lc $traffic_opts{file_size};
    $traffic_opts{file_size} =~ s/b$//;

    if (my $size = $self->_validate_traffic_file(file => $file)) {
        # If the file exists and there's an explicit requested size, the
        # file needs to be at least that large. If there is no explicit
        # size, then use the file's size at the value to be passed to the
        # tool by overwriting $traffic_opts{file_size}.
        if ($opts{file_size}) {
            # If there was a particular size given in the original opts, then
            # convert it to a straight number and make sure that the current
            # file size is adequate. This is the tricky branch.
            if ($opts{file_size} =~ /^(\d+)([kmg])b?$/i) {
                my $req_size = $1 * $FILE_SIZES{lc $2};
                if ($req_size > $size) {
                    NATE::BaseException->throw(
                        "File '$file' exists on the volume already, but " .
                        'is too small for the traffic request. Cannot ' .
                        'use this file for traffic.'
                    );
                }
            } else {
                NATE::Exceptions::Argument->throw(
                    "Value for 'file_size' ($opts{file_size}) is not valid"
                );
            }
        } else {
            # If there is no explicit size, we just use the size that the
            # file is currently at. This is the easy branch.
            $traffic_opts{file_size} = $size;
        }
    } else {
        # This will use NACL::STask::VolumeFile for read-only, and will write
        # out a sparse file for read-write:
        $self->_create_traffic_file(
            file  => $file,
            size  => $traffic_opts{file_size},
            is_rw => ($self->_mount_access eq 'rw') ? 1 : 0
        );
    }

    # Make $file an absolute path:
    $file = NACL::MTask::Traffic->get_abs_path($self->_mount, $file);

    $traffic->[$id] = NACL::MTask::Traffic->start(
        command_interface => $self->_client,
        paths             => $file,
        tool              => $tool,
        %traffic_opts
    );

    $Log->exit();
    return $id;
}

=head2 traffic_stop

    # Stop one traffic instance:
    $io_handle->traffic_stop(id => $id);

    # Stop all traffic:
    $io_handle->traffic_stop;

(Must be called as an instance method)

Stop a traffic process that was previously started by B<traffic_start>. If
called with no arguments, stops all traffic processes that are currently
active, that were created by this object.

=over

=item Options

This method recognizes only one parameter:

=over

=item C<< id =E<gt> integer >>

The ID returned by a previous call to B<traffic_start>. Optional, if no ID
is given then all active traffic streams are stopped.

=back

=item Exceptions

The following exceptions may be thrown by B<traffic_stop>:

=over

=item B<NATE::Exceptions::Argument>

If an unknown parameter is passed, or if the value of C<id> is not an
integer, this exception will be thrown.

=item B<NATE::BaseException>

If the C<id> given does not correspond to a traffic stream controlled by
this handle, this exception will be thrown.

=back

In addition, any exception that might be thrown by the B<stop> method of
B<NACL::Data::Traffic> may be thrown here, as well.

=back

=cut

###############################################################################
#
#   Sub Name:       traffic_stop
#
#   Description:    Stop one traffic stream, or all of them, depending on
#                   whether an argument is given.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional arguments
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub traffic_stop {
    $Log->enter();
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate(
        @opts,
        {
            id => { type => SCALAR, regex => qr/^\d+$/, optional => 1 },
        }
    );

    my $traffic = $self->_traffic;
    if ($opts{id}) {
        if (my $traf_handle = $traffic->[$opts{id}]) {
            $traf_handle->stop(if_stopped => 'pass');
            undef $traffic->[$opts{id}];
        } else {
            $Log->exit();
            NATE::BaseException->throw(
                'Invalid "id" argument passed, no such traffic running'
            );
        }
    } else {
        for my $traf_handle (@{$traffic}) {
            next if (! defined $traf_handle);
            $traf_handle->stop(if_stopped => 'pass');
            undef $traf_handle;
        }
    }

    # If there is now no traffic running at all, null the array:
    if (! any { defined($_) } @{$traffic}) {
        @{$traffic} = ();
    }

    $Log->exit();
    return $self;
}

###############################################################################
#
# Private methods
#
###############################################################################

###############################################################################
#
#   Sub Name:       _find_lun
#
#   Description:    For our encapsulated Volume object, find the LUN that
#                   should be used for SAN-based I/O and save it to the _lun
#                   attribute.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional args. These have
#                                                 been pre-validated by
#                                                 create_mount()
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub _find_lun {
    $Log->enter();
    my ($self, %opts) = @_;
    my $options = $opts{options};

    # Just find the first LUN on the volume, unless they specified the LUN
    # name. In that case search for it specifically. This will throw the
    # exception if none is found.
    my $lun = NACL::STask::Lun->find(
        command_interface => $self->_volume->command_interface,
        filter            => {
            volume  => $self->_volume,
            vserver => $self->_volume->vserver,
            ($options->{lun_name}) ? (lun => $options->{lun_name}) : (),
        }
    );

    $self->_lun($lun);

    $Log->exit();
    return;
}

###############################################################################
#
#   Sub Name:       _derive_lun_mountpoint
#
#   Description:    From the given LUN, derive the local-filesystem mountpoint
#                   based on the type arg.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional args
#
#   Returns:        Success:    string
#                   Failure:    throws exception
#
###############################################################################
sub _derive_lun_mountpoint {
    $Log->enter();
    my ($self, %opts) = @_;

    # Save a call to "lun show"... the LUN name is the last element of its
    # full path.
    (my $lunname = $opts{lun}->path) =~ s{.*/}{};

    # Get the data LIFs for the command-interface of this LUN object. We'll
    # use the first one to get the address to use in the mount-point.
    my @ifaces = NACL::C::NetworkInterface->find(
        command_interface => $opts{lun}->command_interface,
        canned_filters    => [ qw(up_data_interfaces) ],
        allow_empty       => 1
    );
    if (! @ifaces) {
        # Why not just let find() throw? I want more control over the message.
        $Log->exit();
        NATE::BaseException->throw(
            'No data LIFs found for the LUN command-interface'
        );
    }
    my $data_ip = $ifaces[0]->get_one_state_attribute('address');

    $Log->exit();
    return "/t/$data_ip/$lunname";
}

###############################################################################
#
#   Sub Name:       _create_nfs_mount
#
#   Description:    Create a mount-point for the client/volume contained in
#                   $self, using NFS as the protocol.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional args. These have
#                                                 been pre-validated by
#                                                 create_mount()
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub _create_nfs_mount {
    $Log->enter();
    my ($self, %opts) = @_;

    # Extra opts to pass to the NACL::STask::Client::Mount->create() call. Used
    # to build the mountopts option, and to conditionally include the chmod
    # option.
    my @opts;

    # No reason to call this accessor so many times:
    my $volume = $self->_volume;

    # We need to provide a path that includes .admin if we are mounting
    # the volume as RW from a LS mirror. But don't do this for SAN!
    if ($opts{is_ls} and ($opts{access} eq 'rw')) {
        if ($self->_san) {
            $Log->warn(
                'SAN and LS mirrors are not compatible. Skipping the ' .
                'application of "/.admin" to the path.'
            );
        } else {
            my $path = $volume->get_one_state_attribute('junction-path');
            push(@opts, path => "/.admin$path");
        }
    }

    # Add things from $opts{options}:
    for my $key (keys %{$opts{options}}) {
        # 'mountopts' is handled specially
        next if ($key eq 'mountopts');
        if ($opts{options}->{$key}) {
            push(@opts, $key => $opts{options}->{$key});
        }
    }
    # For mountopts, we also have to factor in $opts{access} (which means
    # removing the opposite value if it is present).
    my @mountopts = split(/,/ => $opts{options}->{mountopts} || q{});
    if ($opts{access} eq 'rw') {
        @mountopts = grep { lc $_ ne 'ro' } @mountopts;
        push(@mountopts, 'rw');
        push(@opts, chmod => '777');
    } else {
        # Don't need to explicitly test for 'ro', the param validation in
        # create_mount() assures that it is one of 'rw' or 'ro'.
        @mountopts = grep { lc $_ ne 'rw' } @mountopts;
        push(@mountopts, 'ro');
    }
    push(@opts, mountopts => join(q{,} => @mountopts));

    # OK, now we're ready to actually create the mount-point and
    # NACL::STask::Client::Mount object:
    my $mount = NACL::STask::Client::Mount->create(
        command_interface        => $self->_client,
        server_command_interface => $volume->command_interface,
        vserver                  => $volume->vserver,
        volume                   => $volume,
        nacltask_if_exists       => 'purge',
        @opts
    );
    # If that didn't throw an exception, we have a mount object in $mount:
    $self->_mount($mount);

    $Log->exit();
    return $self;
}

###############################################################################
#
#   Sub Name:       _create_cifs_mount
#
#   Description:    Create a mount-point for the client/volume contained in
#                   $self, using CIFS as the protocol.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional args. These have
#                                                 been pre-validated by
#                                                 create_mount()
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub _create_cifs_mount {
    $Log->enter();
    my ($self, %opts) = @_;

    # This will be either an existing share we find, or one we create:
    my $share;
    # The name of the share, for searching and creation:
    my $share_name;
    # These values are used in share creation and share mounting
    my ($properties, $user, $password) =
        @{$opts{options}}{qw(properties user password)};
    # This is just used so many times, it makes sense to only call the
    # accessor once:
    my $volume = $self->_volume;
    # Save this value to avoid calling this method twice:
    my $path = $volume->get_one_state_attribute('junction-path');

    # If this is for an LS mirror, the share will be slightly different based
    # on whether $opts{access} is 'ro' or 'rw'. Otherwise the share is always
    # the same.
    $share_name = $volume->volume;
    if ($opts{is_ls}) {
        $share_name .= "_ls_$opts{access}";
        if ($opts{access} eq 'rw') {
            $path = "/.admin$path";
        }
    }

    # Look for an existing share for this volume:
    $share = NACL::STask::VserverCifsShare->find(
        command_interface => $volume->command_interface,
        filter => {
            vserver      => $volume->vserver,
            'share-name' => $share_name,
        },
        allow_empty => 1
    );
    if (! $share) {
        # If the share doesn't already exist, create it:
        $share = NACL::STask::VserverCifsShare->create(
            command_interface => $volume->command_interface,
            vserver           => $volume->vserver,
            'share-name'      => $share_name,
            path              => $path,
            ($properties) ? ('share-properties' => $properties) : ()
        );
    }

    # OK, now we're ready to actually create the mount-point and
    # NACL::STask::Client::Mount object:
    my $mount = NACL::STask::Client::Mount->create(
        command_interface        => $self->_client,
        server_command_interface => $volume->command_interface,
        vserver                  => $volume->vserver,
        share                    => $share,
        user                     => $user,
        password                 => $password,
        nacltask_if_exists       => 'purge'
    );
    # If that didn't throw an exception, we have a mount object in $mount:
    $self->_mount($mount);

    $Log->exit();
    return $self;
}

###############################################################################
#
#   Sub Name:       _validate_traffic_file
#
#   Description:    Determine if the given file already exists, and if it
#                   does return the file size as given by slist. Return 0
#                   if it doesn't exist. If any other exception occurs, let
#                   it through.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional args. These have
#                                                 been pre-validated.
#
#   Returns:        Success:    file size (0 if file doesn't exist)
#                   Failure:    throws exception
#
###############################################################################
sub _validate_traffic_file {
    my ($self, %opts) = @_;
    my $file = $opts{file};
    my $size;

    try {
        my $volume   = $self->_volume;
        my $pathname = sprintf('/vol/%s/%s', $volume->volume, $file);
        my $result   = NACL::C::Slist->slist(
            command_interface => $volume->command_interface,
            path              => $pathname
        );
        $size = $result->get_parsed_output->[0]->{file_size};
    } catch NACL::APISet::Exceptions::InvalidParamValueException with {
        # This exception means that the path doesn't exist. Any other
        # exception should be allowed to bubble up.
        $size = 0;
    };

    return $size;
}

###############################################################################
#
#   Sub Name:       _create_traffic_file
#
#   Description:    Create a file of a given size for use by the traffic
#                   generation tools.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional args. These have
#                                                 been pre-validated.
#
#   Globals:        %FILE_SIZES
#
#   Returns:        Success:    void
#                   Failure:    throws an exception
#
###############################################################################
sub _create_traffic_file {
    my ($self, %opts) = @_;
    my $file = $opts{file};
    my $size = $opts{size};
    my $rw   = $opts{is_rw};

    if ($rw) {
        # For read-write traffic, use NACL::C::Client::Remote::File to remotely
        # create and write to the file. If we use NACL::STask::VolumeFile the
        # file ends up owned by root, and is thus useless to us.
        require NACL::C::Client::Remote::File;

        my $client   = $self->_client;
        my $fullname =
            NACL::MTask::Traffic->get_abs_path($self->_mount, $file);
        my ($total_units, $one_unit);

        if ($size =~ /^(\d+)([kmg])$/) {
            my ($num, $unit) = ($1, lc $2);
            if ($unit eq 'g') {
                # I don't want to create any 1-gigabyte scalars, so just bump
                # up $total_units by a factor and use a megabyte scalar.
                $total_units = $num * 1024;
                $unit = 'm';
            } else {
                $total_units = $num;
            }
            $one_unit = chr(0) x $FILE_SIZES{$unit};
        } else {
            NATE::BaseException->throw("Value for 'size' ($size) is not valid");
        }

        # Get the remote file handle object:
        my $fh_obj = NACL::C::Client::Remote::File->open(
            command_interface => $client,
            file_name         => $fullname,
            arguments         => {
                access_mode => 'write',
                open_flags  => 'create-new',
            }
        );
        # Write the "$one_unit" at the very end of the file. This will
        # create the rest of the file by extending the file pointer out to
        # the end.
        $fh_obj->write(
            content => $one_unit,
            offset  => ($total_units - 1) * length($one_unit)
        );
        $fh_obj->close;
    } else {
        # For read-only traffic, we need to create the complete file. But
        # the volume is mounted read-only on the client, so we have to go
        # through the filer to do it.

        # Get the volume object, and then get some elements from it:
        my $volume  = $self->_volume;
        my $ci      = $volume->command_interface;
        my $volname = $volume->volume;
        my $vserver = $volume->vserver;
        my $prefix  = "/vol/$volname";

        # OK, we should be able to create the file:
        my $fileobj = NACL::STask::VolumeFile->create(
            command_interface => $ci,
            path              => "$prefix/$file",
            vserver           => $vserver,
            size              => $size,
            nacltask_wait     => 1
        );
    }

    return;
}

###############################################################################
#
#   Sub Name:       DESTROY
#
#   Description:    Class destructor.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub DESTROY {
    my ($self) = @_;

    # We just have to make sure that all traffic streams controlled by this
    # object are stopped, if they haven't already been. This call will do
    # that. If there are no active traffic streams it will do nothing.
    $self->traffic_stop;

    return;
}

1;
