# $Id: //depot/prod/test/Rsierranevada/dp/snapmirror/paloma/rapid/RRS_Volume.pm#3 $
#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

## @summary Rapid Regression Suite volume operations
## @author rjray@netapp.com
## @status shared
## @pod here

package RRS_Volume;

# The use of the "dispatch table" model means some of these appear to be
# unused private subs.
## no critic (ProhibitUnusedPrivateSubroutines)
# The regexp for IPv4 addresses triggers these two:
## no critic (ProhibitComplexRegexes)
## no critic (ProhibitEnumeratedClasses)

use strict;
use warnings;

use Params::Validate qw(
    validate validate_with validation_options
    SCALAR OBJECT BOOLEAN HASHREF CODEREF
);
use Readonly;
use Scalar::Util qw(refaddr);

use Tharn qw($Log);
use NACL::C::Node;
use NACL::C::Client;
use NACL::C::Client::Mkfs;
use NACL::C::NetworkInterface;
use NACL::STask::Client::Mount;
use NACL::STask::Volume;
use NACL::STask::VolumeClone;
use NACL::STask::VolumeSnapshot;
use NACL::STask::VolumeEfficiency;
use NACL::STask::Snapmirror;
use NACL::STask::Lun;
use NACL::STask::LunIgroup;
use NACL::MTask::Data;
use NACL::ComponentUtils qw(_verify_invocation);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;

use RRS_Config;
use RRS_IO;

# This is the "golden snapshot" label, which is used at the end of data
# population to make a final snapshot on a volume.
Readonly my $GOLDEN_LABEL => '__GOLDEN__';

# The default size of volumes, used unless the user specifies an explicit
# value.
Readonly my $DEFAULT_VOLUME_SIZE => '20g';
# The default size of populate data, used unless the user specifies an
# explicit value.
Readonly my $DEFAULT_POPULATE_SIZE => '1g';

# This table maps pedigree types (converted to lower-case and with hyphens
# translated to underscores) to method names. Note that I'm using the names
# here, rather than references, since I'm calling them as methods, not as
# direct subroutine invocations.
Readonly my %CREATOR_TABLE => (
    # Source pedigrees
    rw               => '_create_basic',
    dp_to_rw         => '_create_dp_to_rw',
    t_state_to_rw    => '_create_t_state',
    flexcache        => '_create_with_flexcache',
    flexclone_to_rw  => '_create_flexclone',
    dp_clone_to_rw   => '_create_dp_flexclone',
    restored_fr_tape => '_create_from_tape',
    # Destination pedigrees
    dp               => '_create_basic',
    t_state_to_dp    => '_create_t_state',
);

# This table collects hooks for a given object of this class. It is indexed
# by the refaddr of the object adding a hook or accessing hooks. Each entry
# should be a hash reference keyed by the various "known" hook names.
my %HOOKS_TABLE = ();

# This is a regexp taken from Regexp::Common that both matches and validates
# an IPv4 address. It does no capturing. It is slightly re-written here to
# use the /x option so that it does not trigger the line-length policies of
# PCAST.
Readonly my $IP_ADDR => qr{
    ^
    (?:25[0-5]|2[0-4][0-9]|[0-1]?[0-9]{1,2}) # Octet (0-255)
    [.]                                      # Dot
    (?:25[0-5]|2[0-4][0-9]|[0-1]?[0-9]{1,2}) # Octet (0-255)
    [.]                                      # Dot
    (?:25[0-5]|2[0-4][0-9]|[0-1]?[0-9]{1,2}) # Octet (0-255)
    [.]                                      # Dot
    (?:25[0-5]|2[0-4][0-9]|[0-1]?[0-9]{1,2}) # Octet (0-255)
    $
}x;

# Set the on_fail behavior for Params::Validate to throw an exception (of the
# NATE::Exceptions::Argument class) on errors instead of a plain croak().
validation_options(
    on_fail => sub {
        NATE::Exceptions::Argument->throw(shift);
    }
);

# These are spec-fragments that are used in multiple places in various calls
# to validate() and/or validate_with(). Putting them here cuts down on
# complexity of individual subroutines, and speeds up calls to validation by
# having the specs already computed. Any spec that would appear twice or more
# should be declared here.

# Plain optional scalar, no real validation
Readonly my $OPTIONAL_SCALAR => {
    type     => SCALAR,
    optional => 1,
};
# Optional boolean value
Readonly my $OPTIONAL_BOOLEAN => {
    type     => BOOLEAN,
    optional => 1,
};
# Optional hash-reference value
Readonly my $OPTIONAL_HASHREF => {
    type     => HASHREF,
    optional => 1,
};
# Node objects
Readonly my $NODE_OBJECT => {
    type     => OBJECT,
    optional => 1,
    isa      => 'NACL::C::Node',
};
# Aggregate, by name or by object
Readonly my $AGGR_NAME_OR_OBJECT => {
    type      => SCALAR | OBJECT,
    optional  => 1,
    callbacks => {
        'correct object type' => sub {
            my $obj = shift;
            # We know from type-validation that this is either a
            # plain scalar or a blessed ref:
            return (ref $obj) ? $obj->isa('NACL::C::Aggregate') : 1;
        },
    },
};
# Vserver, by name or by object
Readonly my $VSERVER_NAME_OR_OBJECT => {
    type      => SCALAR | OBJECT,
    optional  => 1,
    callbacks => {
        'correct object type' => sub {
            my $obj = shift;
            # We know from type-validation that this is either a
            # plain scalar or a blessed ref:
            return (ref $obj) ? $obj->isa('NACL::C::Vserver') : 1;
        },
    },
};

# TODO:
#
# After the SAN update for complex volume situations, there are a few things
# that need to be cleaned up in the future:
#
# 1. Make the use of NACL::C::Client object consistent. The connections are
#    actually cached by NATE, so there is no need to jump through hoops to
#    locally cache a NACL::C::Client instance. In particular, the private key
#    "_san_client" is unneeded, it is OK to just create new NACL::C::Client
#    instances as needed.
# 2. And speaking of the private keys for SAN, try to find a better way to
#    carry that information around between methods. While the current method
#    should be adequately cleaning up after itself, it is still confusing to
#    others reading the code.
#
# Anything to do with the SAN hackiness should be marked with a search token,
# "SANHACK".

=head1 NAME

RRS_Volume - Volume creation/allocation for the Rapid Regression Suite

=head1 SYNOPSIS

    use RRS_Volume;

    # Create a handle with all the options:
    $handle = RRS_Volume->new(
        variant             => $variant,
        source_node         => $Node_Src,
        source_aggr_32      => $Aggr_Name_Src_32,
        source_aggr_64      => $Aggr_Name_Src_64
        source_vserver      => $Vserver_Name_Src,
        destination_node    => $Node_Dst,
        destination_aggr_32 => $Aggr_Name_Dst_32,
        destination_aggr_64 => $Aggr_Name_Dst_64,
        destination_vserver => $Vserver_Name_Dst,
        transition_node     => $Transition_Node_Name,
        transition_address  => $Transition_Node_MP_Address,
        transition_volume   => $Transition_Node_Volume_Name,
        volume_size         => $size
    );
    # Create/retrieve volumes, most options are unneeded:
    $src = $handle->get_source_volume;
    $dst = $handle->get_destination_volume;

    # ...or

    # Create a handle with just the variant
    $handle = RRS_Volume->new(variant => $variant);
    # Now a call to get_source_volume() needs the extra info
    $src = $handle->get_source_volume(
        source_node    => $Node_Src,
        source_aggr_64 => $Aggr_Name_Src,
        source_vserver => $Vserver_Name_Src,
    );

=head1 DESCRIPTION

The B<RRS_Volume> class is used to create a handle that stores most/all of the
information needed to create or find volumes for use by Rapid Regression
Suite lifecycle scripts. It takes (at minimum) the name of the variant to get
volumes for. Additionally, it can take and hold on to the other arguments
that are used in the retrieval/creation of volumes. This can be useful for
cases such as fan-out, where there may be multiple calls being made for which
the core options are always the same.

This class does no automatic clean-up of the volumes it returns to the caller.
The caller is responsible for deciding when and if to purge the volumes. The
calls to B<get_source_volume> and B<get_destination_volume> do have a mechanism
for knowing when a volume was created versus being found and reused.

=head1 METHODS

This class provides the following methods. No subroutines are exported by this
package.

=head2 new

    $handle = RRS_Volume->new(
        variant             => $variant,
        source_node         => $Node_Src,
        source_aggr_32      => $Aggr_Name_Src_32,
        source_aggr_64      => $Aggr_Name_Src_64
        source_vserver      => $Vserver_Name_Src,
        destination_node    => $Node_Dst,
        destination_aggr_32 => $Aggr_Name_Dst_32,
        destination_aggr_64 => $Aggr_Name_Dst_64,
        destination_vserver => $Vserver_Name_Dst,
        transition_node     => $Transition_Node_Name,
        transition_address  => $Transition_Node_MP_Address,
        transition_volume   => $Transition_Node_Volume_Name,
        client              => $Client_Name,
        volume_size         => $size
    );

(Must be called as a class method)

Create a new handle object and return it to the caller. All options passed in
are saved on the object for use by later method calls. Note that only the
options listed below are allowed and saved, any unknown options will cause
an exception to be thrown.

All failures in B<new> throw an exception. See C<Exceptions>, below.

=over

=item Options

The following options are recognized by B<new>:

=over

=item C<< variant =E<gt> $variant_name >>

(Required) The name of the variant that volumes will be created for. The name
is used to retrieve a record from the configuration table encapsulated by the
B<RRS_Config> class. This is the only mandatory parameter, and if the value
passed in is not found in the configuration table an exception is thrown.

=item C<< source_node =E<gt> $node_name_or_object >>

(Optional) The node ("command_interface") to be used on the source side, for
creating source volumes. Must be a B<NACL::C::Node> (or sub-type) object.

=item C<< source_aggr_32 =E<gt> $aggr_name_or_object >>

(Optional) The name of the 32-bit aggregate to use when creating source
volumes. May be a B<NACL::C::Aggregate> (or sub-type) object, in which case the
name will be taken from the object. Only used when the specified variant calls
for the source volume to be on a 32-bit aggregate.

=item C<< source_aggr_64 =E<gt> $aggr_name_or_object >>

(Optional) The name of the 64-bit aggregate to use when creating source
volumes. May be a B<NACL::C::Aggregate> (or sub-type) object, in which case the
name will be taken from the object. Only used when the specified variant calls
for the source volume to be on a 64-bit aggregate.

=item C<< source_vserver =E<gt> $vserver_name_or_object >>

(Optional) The name of the vserver to use when creating source volumes. May be
a B<NACL::C::Vserver> (or sub-type) object, in which case the name will be
taken from the object.

=item C<< destination_node =E<gt> $node_name_or_object >>

(Optional) The node ("command_interface") to be used on the destination side,
for creating destination volumes. Must be a B<NACL::C::Node> (or sub-type)
object.

=item C<< destination_aggr_32 =E<gt> $aggr_name_or_object >>

(Optional) The name of the 32-bit aggregate to use when creating destination
volumes. May be a B<NACL::C::Aggregate> (or sub-type) object, in which case the
name will be taken from the object. Only used when the specified variant calls
for the destination volume to be on a 32-bit aggregate.

=item C<< destination_aggr_64 =E<gt> $aggr_name_or_object >>

(Optional) The name of the 64-bit aggregate to use when creating destination
volumes. May be a B<NACL::C::Aggregate> (or sub-type) object, in which case the
name will be taken from the object. Only used when the specified variant calls
for the destination volume to be on a 64-bit aggregate.

=item C<< destination_vserver =E<gt> $vserver_name_or_object >>

(Optional) The name of the vserver to use when creating destination volumes.
May be a B<NACL::C::Vserver> (or sub-type) object, in which case the name will
be taken from the object.

=item C<< transition_node =E<gt> $node_name|$ip_address|$node_obj >>

(Optional) The NATE-hosts name of the node to use when creating "transitional"
(7Mode to CMode) volumes, or the IPv4 address of the node. Alternately, it may
be an object of the class B<NACL::C::Node> or B<NACL::C::SystemNode>. The node
will be used (along with the next parameter, if given) to create a transitional
vserver peering for the sake of creating the transitional volume.

=item C<< transition_address =E<gt> $ip_address|$key_name >>

(Optional) An IPv4 address to use as the multi-path address when creating a
transitional vserver peering between the source-node and the node specified
by C<transition_node>. Alternately, this may be the name of a NATE-hosts file
and/or a key into a hostrec record.

If the parameter value is not an IP address, it is expected to be one of:

    name
    :key
    name:key

The C<name> portion, if present, will be looked up as a NATE-hosts file name,
and an exception thrown if it is not found. If the name is provided and the
file found, then the hostrec record built from it will be used instead of the
hostrec from C<transition_node>.

If the C<key> portion is present, it is used to look up a specific value in
the hostrec record, rather than using the C<default_ip> value. If the key
is not present in the hostrec, or if the value of the key is not a valid IP
address, an exception is thrown. If C<transition_node> is passed as an IP
address, and C<key> is given but C<name> is not, then an error is thrown
(because the key cannot be looked up, as there is no hostrec record at all
to reference).

Note that C<transition_address> is optional even when creating transitional
volumes; it is not required to create the transitional vserver peering.

=item C<< transition_volume =E<gt> $volume_name >>

(Optional) The name of the volume on the 7Mode filer specified by
C<transition_node> that is to be used as the source for creating a new
transitional volume in CMode. This is less useful to specify within B<new>,
unless you plan to create several transitioned volumes from the same 7Mode
source. Most of the time, you'll probably want to pass this to
B<get_source_volume> or B<get_destination_volume>.

=item C<< client =E<gt> $client_name >>

(Optional) The name of the client to use when populating data. If not given,
then C<localhost> will be used. Note that if the volumes you will be creating
are to be used with SAN, this parameter must reflect the client that the rest
of the RRS test-stream will be using. Otherwise later tests will fail to
connect to the SAN LUN that gets created. If this is left as C<localhost> and
SAN is being applied to the volume, a warning will be issued.

=item C<< nfs_options =E<gt> $hashref >>

=item C<< cifs_options =E<gt> $hashref >>

(Optional) If needed, options for the creation/mounting of either NFS volumes
or CIFS shares. These parameters are passed to the B<create_mount> method of
the B<RRS_IO> class unchanged. For their definition and restrictions, see
the B<RRS_IO> documentation.

=item C<< san_options =E<gt> $hashref >>

(Optional) Options used for SAN LUN creation and mounting. Unlike the NFS and
CIFS options, these options are used both during volume creation as well as
data population. The options will be passed to the B<create_mount> method of
B<RRS_IO>, but some will be used by B<get_source_volume>. Those used on
the volume side are:

=over

=item B<lun_name>

The name to assign to the LUN that gets created. If not passed, then the
source volume's name will be used, with C<_lun> appended to it. Generally,
you will not need to use this. It is mostly for internal use when creating
a final volume from an intermediate volume via snapmirror. (In such a case,
the intermediate volume has a random name and the LUN would end up named after
the random name rather than the final volume name.)

=item B<lun_size>

The size to create the LUN as. The default is 50% of the size of the volume
itself. The value may be either a typical size expression (C<20gb>, C<100mb>,
etc.) or a percentage which is applied to the volume size (C<50%>, C<90%>,
etc.). If the size is larger than the volume, an exception is thrown. If the
size is too close to the size of the volume itself, you may encounter errors
later in the lifecycle.

=item B<fstype>

The type of filesystem to create on the LUN. Defaults to C<ext2> for Linux.

=back

Other options for SAN are specified in the B<RRS_IO> documentation.

=item C<< volume_size =E<gt> $string >>

(Optional) The size to create volumes at. The default is C<20g>, which will
create volumes with a size of 20 gigabytes.

=item C<< configuration_file =E<gt> $file_name >>

(Optional) If the B<RRS_Config> class should read the configuration from some
place other than the default, pass that file name here. This value is not saved
on the B<RRS_Volume> object, it is only used to retrieve the B<RRS_Config>
object. This value is subject to the same usage and consideration as the
C<file> option to the the B<new> method in B<RRS_Config>. In fact, it is
passed as the value to that option. This means that if it is not an absolute
path, it will be treated as relative to the the directory that these modules
reside in.

=back

The C<source_*>, C<destination_*> and C<transition_*> options are all marked
optional, as these can also be specified when you call B<get_source_volume> or
B<get_destination_volume>. In general, you will usually need at least the
C<source_*> and C<destination_*> values to be specified at some point, so when
you do is up to you. The C<transition_*> options are only needed when creating
volumes of type C<T-state-to-RW> or C<T-state-to-DP>.

=item Exceptions

There are two possible exceptions that may be thrown by B<new>:

=over

=item B<NATE::Exceptions::Argument>

If an unknown argument is passed, or if an argument is of the wrong type, this
exception will be thrown. This exception will also be thrown if the C<variant>
option's value is not found in the configuration.

=item B<NATE::BaseException>

This exception can be thrown by the B<RRS_Config> constructor. If that occurs,
it is not caught by this routine but rather allowed to return to the caller.

=back

=back

=cut

###############################################################################
#
#   Sub Name:       new
#
#   Description:    Create a RRS_Volume handle object, properly launder the
#                   options, etc.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $class    in      scalar    Class we're blessing into
#                   @opts     in      list      Additional options
#
#   Globals:        $NODE_OBJECT
#                   $AGGR_NAME_OR_OBJECT
#                   $VSERVER_NAME_OR_OBJECT
#                   $OPTIONAL_SCALAR
#
#   Returns:        Success:    object
#                   Failure:    throws exception
#
###############################################################################
sub new {
    $Log->enter();
    my ($class, @opts) = @_;
    # This must be a static method invocation:
    $class->_verify_invocation(style => 'static_only');
    my %opts = validate(
        @opts,
        {
            # Variant is a required arg, so we don't use the generic spec here
            variant             => { type => SCALAR },
            source_node         => $NODE_OBJECT,
            source_aggr_32      => $AGGR_NAME_OR_OBJECT,
            source_aggr_64      => $AGGR_NAME_OR_OBJECT,
            source_vserver      => $VSERVER_NAME_OR_OBJECT,
            destination_node    => $NODE_OBJECT,
            destination_aggr_32 => $AGGR_NAME_OR_OBJECT,
            destination_aggr_64 => $AGGR_NAME_OR_OBJECT,
            destination_vserver => $VSERVER_NAME_OR_OBJECT,
            transition_node     => $OPTIONAL_SCALAR,
            transition_address  => $OPTIONAL_SCALAR,
            transition_volume   => $OPTIONAL_SCALAR,
            client              => $OPTIONAL_SCALAR,
            nfs_options         => $OPTIONAL_HASHREF,
            cifs_options        => $OPTIONAL_HASHREF,
            san_options         => $OPTIONAL_HASHREF,
            volume_size         => $OPTIONAL_SCALAR,
            configuration_file  => $OPTIONAL_SCALAR,
        }
    );
    my $self = { variant => $opts{variant} };

    # Get the RRS_Config object
    my @conf_opts = $opts{configuration_file} ?
        (file => $opts{configuration_file}) : ();
    my $config = RRS_Config->new(@conf_opts);
    # Verify the value of $opts{variant}. $config->get_variant throws an
    # exception if the value is unknown.
    $self->{variant_data} = $config->get_variant(variant => $opts{variant});

    # For the aggregates, for now, we have default names. But if I'd put the
    # defaults in the validation spec fragments above, I'd have had to have
    # separate fragments for 32-bit and 64-bit...
    for (qw(source_aggr_32 destination_aggr_32)) {
        $opts{$_} ||= 'aggr_rrs_32';
    }
    for (qw(source_aggr_64 destination_aggr_64)) {
        $opts{$_} ||= 'aggr_rrs_64';
    }

    # For all these options, validate() has already done any and all validation
    # needed (plus the extra steps above, for aggr names). Just store the
    # values on the same-named keys on $self:
    for my $key (qw(source_node destination_node
                    source_aggr_32 source_aggr_64
                    destination_aggr_32 destination_aggr_64
                    source_vserver destination_vserver
                    transition_node transition_address transition_volume
                    client nfs_options cifs_options san_options volume_size)) {
        $self->{$key} = $opts{$key};
    }

    bless $self, $class;

    $Log->exit();
    return $self;
}

=head2 get_source_volume

    # Using just the defaults given when $handle was created:
    $srcvol = $handle->get_source_volume;

    # Specifying some of the values explicitly: this will use an
    # alternate node to issue the commands and will use a specific
    # name.
    $srcvol = $handle->get_source_volume(
        source_node => $source_node,
        volume_name => $vol_name
    );

(Instance method only)

This method gets the source volume for the variant that C<$handle> was
initialized with. It uses a combination of the data provided when C<$handle>
was created, data from the variants table, and any parameters passed to the
method itself (some of which allow for overriding the values passed to
B<new>).

The return value is a B<NACL::STask::Volume> object. No snapmirror relationship
is set up by this method, only volume creation. If the creation fails, an
exception is thrown.

=over

=item Options

The following options are specifically recognized by this method. In addition,
any options taken by B<NACL::STask::Volume::create> may be passed here as well.
Any options not in the following list will be passed through to the volume
creation unchanged.

=over

=item source_node

=item source_aggr_32

=item source_aggr_64

=item source_vserver

=item destination_node

=item destination_aggr_32

=item destination_aggr_64

=item destination_vserver

=item transition_node

=item transition_address

=item transition_volume

=item client

=item nfs_options

=item cifs_options

=item san_options

=item volume_size

These options are identical to the same-named options to the constructor
B<new>. If passed, any of these will override the value(s) passed when B<new>
was called.

=item C<< volume_name =E<gt> $string >>

Provide an explicit name for the created volume. If this is not passed, a
randomly-generated name is used.

=item C<< populate_size =E<gt> $string >>

The size of the data-set to populate on the source volume. If not passed, then
the default size (C<1g>) is used. Currently, must be in megabytes or gigabytes.
Other measures are not yet supported. See the C<size> option to the
B<populate> method of B<RRS_IO> for details on the format of this parameter.

=item C<< no_populate =E<gt>  0|1 >>

A boolean value that says whether to skip data population. If the value is
C<1> (true), then no data population is done. If the value is C<0> (false)
then data population will be done. The default is C<0>.

=back

Note that in addition to the overridable C<source_*> options, this method
allows you to pass the corresponding C<destination_*> options. These may be
used by some of the variants when creating a source volume, so they are allowed
to be overridden here.

B<NOTE:> While you can pass any B<NACL::STask::Volume::create> options to
this method, know that the option C<nacltask_wait> is explicitly ignored. It
will always be set to C<1>, as volume creation has to wait for completion
before data population can be done.

=item Exceptions

Any exceptions that can be thrown during the creation of a volume may be
seen here. In addition, the following may be thrown explicitly by
B<get_destination_volume>:

=over

=item B<NATE::Exceptions::Argument>

This exception will be thrown if an unknown argument is passed, or if the
value of a passed-in argument fails validation.

=back

=back

=cut

###############################################################################
#
#   Sub Name:       get_source_volume
#
#   Description:    Get/create a source volume, based on the parameters passed
#                   in and already set on $self.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional options
#
#   Globals:        $NODE_OBJECT
#                   $AGGR_NAME_OR_OBJECT
#                   $VSERVER_NAME_OR_OBJECT
#                   $OPTIONAL_SCALAR
#                   %CREATOR_TABLE
#
#   Returns:        Success:    NACL::STask::Volume object
#                   Failure:    throws exception
#
###############################################################################
sub get_source_volume {
    $Log->enter();
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate_with(
        params      => \@opts,
        allow_extra => 1,
        spec        => {
            source_node         => $NODE_OBJECT,
            source_aggr_32      => $AGGR_NAME_OR_OBJECT,
            source_aggr_64      => $AGGR_NAME_OR_OBJECT,
            source_vserver      => $VSERVER_NAME_OR_OBJECT,
            destination_node    => $NODE_OBJECT,
            destination_aggr_32 => $AGGR_NAME_OR_OBJECT,
            destination_aggr_64 => $AGGR_NAME_OR_OBJECT,
            destination_vserver => $VSERVER_NAME_OR_OBJECT,
            transition_node     => $OPTIONAL_SCALAR,
            transition_address  => $OPTIONAL_SCALAR,
            transition_volume   => $OPTIONAL_SCALAR,
            volume_name         => $OPTIONAL_SCALAR,
            volume_size         => $OPTIONAL_SCALAR,
            client              => $OPTIONAL_SCALAR,
            nfs_options         => $OPTIONAL_HASHREF,
            cifs_options        => $OPTIONAL_HASHREF,
            san_options         => $OPTIONAL_HASHREF,
            populate_size       => $OPTIONAL_SCALAR,
            no_populate         => $OPTIONAL_BOOLEAN,
        }
    );
    # The options we pass to the private construction method are based on the
    # options we already have stored on $self, with anything the caller passed
    # in to this call. This allows them to override an existing option if they
    # need to.
    my %call_options = (%{$self}, %opts, type => 'source');

    # What type ("pedigree") of volume are we creating?
    (my $pedigree = lc $self->{variant_data}->{pedigree_src}) =~ s/-/_/g;
    my $create_method = $CREATOR_TABLE{$pedigree};
    if (! $create_method) {
        $Log->exit();
        NATE::BaseException->throw(
            "Source type '$pedigree' is unknown or not yet supported"
        );
    }

    my $volume = $self->$create_method(%call_options);

    # If this was a SAN-using variant, then we created some temporary private
    # keys on $self. Remove them.
    for my $key (grep { /^_san_/ } keys %{$self}) {
        delete $self->{$key};
    }

    $Log->exit();
    return $volume;
}

=head2 get_destination_volume

    # Using just the defaults and what was passed when $handle
    # was created:
    my $dst_vol = $handle->get_destination_volume;

    # Supplying some specific values
    my $dst_vol = $handle->get_destination_volume(
        volume_name         => $dst_name,
        destination_aggr_64 => $new_dest_aggr
    );

(Instance method only)

This method gets the destination volume for the variant that C<$handle> was
initialized with. It uses a combination of the data provided when C<$handle>
was created, data from the variants table, and any parameters passed to the
method itself (some of which allow for overriding the values passed to
B<new>).

The return value is a B<NACL::STask::Volume> object. No snapmirror relationship
is set up by this method, only volume creation. If the creation fails, an
exception is thrown.

=over

=item Options

The following options are specifically recognized by this method. In addition,
any options taken by B<NACL::STask::Volume::create> may be passed here as well.
Any options not in the following list will be passed through to the volume
creation unchanged.

=over

=item source_node

=item source_aggr_32

=item source_aggr_64

=item source_vserver

=item destination_node

=item destination_aggr_32

=item destination_aggr_64

=item destination_vserver

=item transition_node

=item transition_address

=item transition_volume

=item volume_size

These options are identical to the same-named options to the constructor
B<new>. If passed, any of these will override the value(s) passed when B<new>
was called.

=item C<< volume_name =E<gt> $string >>

Provide an explicit name for the created volume. If this is not passed, a
randomly-generated name is used.

=back

Note that in addition to the overridable C<destination_*> options, this method
allows you to pass the corresponding C<source_*> options. These may be used
by some of the variants when creating a destination volume, so they are allowed
to be overridden here.

B<NOTE:> While you can pass any B<NACL::STask::Volume::create> options to
this method, know that the option C<nacltask_wait> is explicitly ignored. It
will always be set to C<1>, as volume creation has to wait for completion
before data population can be done.

=item Exceptions

Any exceptions that can be thrown during the creation of a volume may be
seen here. In addition, the following may be thrown explicitly by
B<get_destination_volume>:

=over

=item B<NATE::Exceptions::Argument>

This exception will be thrown if an unknown argument is passed, or if the
value of a passed-in argument fails validation.

=back

=back

=cut

###############################################################################
#
#   Sub Name:       get_destination_volume
#
#   Description:    Get/create a destination volume, based on the parameters
#                   passed in and already set on $self.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional options
#
#   Globals:        $NODE_OBJECT
#                   $AGGR_NAME_OR_OBJECT
#                   $VSERVER_NAME_OR_OBJECT
#                   $OPTIONAL_SCALAR
#                   %CREATOR_TABLE
#
#   Returns:        Success:    NACL::STask::Volume object
#                   Failure:    throws exception
#
###############################################################################
sub get_destination_volume {
    $Log->enter();
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate_with(
        params      => \@opts,
        allow_extra => 1,
        spec        => {
            source_node         => $NODE_OBJECT,
            source_aggr_32      => $AGGR_NAME_OR_OBJECT,
            source_aggr_64      => $AGGR_NAME_OR_OBJECT,
            source_vserver      => $VSERVER_NAME_OR_OBJECT,
            destination_node    => $NODE_OBJECT,
            destination_aggr_32 => $AGGR_NAME_OR_OBJECT,
            destination_aggr_64 => $AGGR_NAME_OR_OBJECT,
            destination_vserver => $VSERVER_NAME_OR_OBJECT,
            transition_node     => $OPTIONAL_SCALAR,
            transition_address  => $OPTIONAL_SCALAR,
            transition_volume   => $OPTIONAL_SCALAR,
            volume_name         => $OPTIONAL_SCALAR,
            volume_size         => $OPTIONAL_SCALAR,
        }
    );
    # The options we pass to the private construction method are based on the
    # options we already have stored on $self, with anything the caller passed
    # in to this call. This allows them to override an existing option if they
    # need to.
    my %call_options = (%{$self}, %opts, type => 'destination');

    # What type ("pedigree") of volume are we creating?
    (my $pedigree = lc $self->{variant_data}->{pedigree_dst}) =~ s/-/_/g;
    my $create_method = $CREATOR_TABLE{$pedigree};
    if (! $create_method) {
        $Log->exit();
        NATE::BaseException->throw(
            "Destination type '$pedigree' is unknown or not yet supported"
        );
    }

    my $volume = $self->$create_method(%call_options);

    $Log->exit();
    return $volume;
}

=head2 get_golden_snapshot_label

    my $ss_label = $handle->get_golden_snapshot_label;

(Instance or static method)

This method gets the "golden" snapshot label used across the Rapid Regression
Suite. This is just the name given to a final snapshot on created source
volumes. The return value is a string.

=over

=item Options

None.

=item Exceptions

None.

=back

=cut

###############################################################################
#
#   Sub Name:       get_golden_snapshot_label
#
#   Description:    A simple public accessor for the golden snapshot label
#                   we use internally.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      any       Ignored-- this can be a class
#                                                 or instance method
#
#   Globals:        $GOLDEN_LABEL
#
#   Returns:        $GOLDEN_LABEL
#
###############################################################################
sub get_golden_snapshot_label {
    $Log->enter();

    $Log->exit();
    return $GOLDEN_LABEL;
}

=head2 add_hook

    # Add a callback to be invoked later when call_hooks() is
    # called:
    $handle->add_hook(
        hook     => 'postinitialize',
        callback => sub {
            my ($self, %opts) = @_;

            for my (sort keys %opts) {
                print "Key $key -> $opts{$key}\n";
            }

            return;
        }
    );

(Instance method only)

This method adds a hook, or callback, to a specified hook-point that can be
later invoked by the B<call_hooks> method (detailed below). All hooks that
are added to an object are unique to that object; if multiple instances of
this class exist at once, they will not have access to each other's hooks.

The return value of this method is the object reference itself if the
operation was successful, otherwise an exception is thrown.

=over

=item Options

The following options are recognized by this method.

=over

=item C<< hook =E<gt> $string >>

(Required) The name of the hook-point to associate the callback with. This may
be any string, and will be the string that is used to look up the callbacks
when B<call_hooks> is invoked later in your program. May be any string value.

=item C<< callback =E<gt> $coderef >>

(Required) The callback to add, at the hook-point C<hook>. Must be a
code-reference, either a reference to an existing subroutine or a closure. When
the callback is invoked during a call to B<call_hooks>, it will be invoked as
if it were a method of this class. The first argument to the method/closure
will be a reference to the same B<RRS_Volume> object used to add the hook. The
remaining arguments will be the key/value pairs passed as additional arguments
to B<call_hooks>

=back

=item Exceptions

Only one exception may be thrown by this method:

=over

=item B<NATE::Exceptions::Argument>

This exception will be thrown if an unknown argument is passed, if one of the
required options is missing, or if the value of the C<callback> parameter is
something other than a code-reference.

=back

=back

=cut

###############################################################################
#
#   Sub Name:       add_hook
#
#   Description:    Add a hook (callback) for the named hook point, on the
#                   calling object.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional options
#
#   Globals:        %HOOKS_TABLE
#
#   Returns:        Success:    $self
#                   Failure:    throws an exception
#
###############################################################################
sub add_hook {
    $Log->enter();
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate(
        @opts,
        {
            hook     => { type => SCALAR },
            callback => { type => CODEREF },
        }
    );

    # Get the table of hooks for this object. If one doesn't exist, initialize
    # it.
    my $hooks_for_obj = $HOOKS_TABLE{refaddr $self};
    if (! $hooks_for_obj) {
        $hooks_for_obj = {};
        $HOOKS_TABLE{refaddr $self} = $hooks_for_obj;
    }
    # Get the current list of hooks for this specific hook. If such a list
    # doesn't exist, initialize it.
    my $hooks_list = $hooks_for_obj->{$opts{hook}};
    if (! $hooks_list) {
        $hooks_list = [];
        $hooks_for_obj->{$opts{hook}} = $hooks_list;
    }

    # Add the provided callback to the end of the list.
    push(@{$hooks_list}, $opts{callback});

    $Log->exit();
    return $self;
}

=head2 call_hooks

    # Call all hooks for the specified hook-point:
    $handle->call_hooks(hook => 'postinitialize');

(Instance method only)

Call any hooks that have been set up for the specified hook-point on the
calling object. If no hooks have been set up, then it quietly does nothing.
The return value from this is the calling object, if all hooks ran without
error. If any hook throws an exception, the exception is immediately
propagated to the caller of this method, and further hooks are not called.
Additionally, this method may throw an exception if the parameters passed
are not valid.

=over

=item Options

Only the following option is required by this method:

=over

=item C<< hook =E<gt> $string >>

(Required) The name of the hook-point to execute hooks for. May be any string
value.

=back

Any additional options passed to B<call_hooks> will be passed to each hook
that gets called, without any pre-processing. All options are assumed to be
key/value pairs.

=item Exceptions

Any exception thrown by an invoked hook will be immediately sent back to the
caller. Exceptions are not held until the end of hook processing; if an
exception occurs when there are still pending hooks to call, those pending
hooks will not get called.

Only one exception may be thrown directly by this method:

=over

=item B<NATE::Exceptions::Argument>

This exception will be thrown if the C<hook> parameter is not passed.

=back

=back

=cut

###############################################################################
#
#   Sub Name:       call_hooks
#
#   Description:    Call any hook defined for the named hook given in the
#                   parameter "hook", for this object. Remaining arguments are
#                   passed to the hook call.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional options
#
#   Globals:        %HOOKS_TABLE
#
#   Returns:        Success:    $self
#                   Failure:    throws an exception
#
###############################################################################
sub call_hooks {
    $Log->enter();
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate_with(
        params      => \@opts,
        allow_extra => 1,
        spec        => {
            hook => { type => SCALAR },
        }
    );

    my $hookname = delete $opts{hook};

    # First check that there are hooks for this object. Then, check that the
    # hooks include the specific hook that we're calling. Lastly, iterate
    # over the hooks that are there, if the array isn't empty...
    if (my $hooks = $HOOKS_TABLE{refaddr $self}) {
        if (my $hooks_array = $hooks->{$hookname}) {
            for my $hook (@{$hooks_array}) {
                $self->$hook(%opts);
            }
        }
    }

    $Log->exit();
    return $self;
}

=head2 clear_hooks

    # Clear all current hooks/callbacks for the given point:
    $handle->clear_hooks(hook => 'postinitialize');

(Instance method only)

Clears all current hooks in place for the calling object for the specified
hook-point. Does nothing if this object has no hooks, or has no hooks for the
given hook-point.

Returns the calling object on success, throws an exception on failure.

=over

=item Options

Only the following option is recognized by this method:

=over

=item C<< hook =E<gt> $string >>

(Required) The name of the hook-point to clear hooks from. May be any string
value.

=back

=item Exceptions

Only one exception may be thrown by this method:

=over

=item B<NATE::Exceptions::Argument>

This exception will be thrown if an unknown argument is passed, or if the
C<hook> parameter is not passed.

=back

=back

=cut

###############################################################################
#
#   Sub Name:       clear_hooks
#
#   Description:    Clear all the hooks on an object for the specified hook.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional options
#
#   Globals:        %HOOKS_TABLE
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub clear_hooks {
    $Log->enter();
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate(
        @opts,
        {
            hook => { type => SCALAR },
        }
    );

    # Get the table of hooks for this object. We only go forward if one exists.
    if (my $hooks_for_obj = $HOOKS_TABLE{refaddr $self}) {
        # Get the current list of hooks for this specific hook. If such a list
        # doesn't exist, we do nothing.
        if (my $hooks_list = $hooks_for_obj->{$opts{hook}}) {
            @{$hooks_list} = ();
        }
    }

    $Log->exit();
    return $self;
}

###############################################################################
#
# Private methods
#
###############################################################################

###############################################################################
#
#   Sub Name:       _create_volume
#
#   Description:    Do the actual creation of a volume, using the options
#                   passed in and the data in $self->{variant_data} to
#                   derive the parameters for the NACL::STask::Volume::create
#                   call.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      All options for volume creation
#                                                 (see comments below)
#
#   Returns:        Success:    NACL::STask::Volume object
#                   Failure:    throws exception
#
###############################################################################
sub _create_volume {
    $Log->enter();
    my ($self, %opts) = @_;

    # We'll always wait for volume creation, and unless they pass an explicit
    # nacltask_if_exists we always purge an existing volume.
    my %create_opts = (
        nacltask_wait      => 1,
        nacltask_if_exists => 'purge',
    );
    my $vdata = $self->{variant_data};
    # If san_options was passed, we may need some of its parameters for the
    # creation of the LUN:
    my $san_options = delete $opts{san_options};
    # Also, if SAN is in effect, we'll need the client:
    my $client_name = $opts{client} || $self->{client} || 'localhost';

    # Set up command_interface, vserver and type based on the type of volume
    # we're creating. Also choose the correct values from the variant data for
    # space-guarantee and aggregate.
    my $vol_type = delete $opts{type};
    if ($vol_type eq 'source') {
        $create_opts{command_interface} = $opts{source_node};
        $create_opts{aggregate}         =
            ($vdata->{aggr_type_src} eq '32-bit') ?
                $opts{source_aggr_32} : $opts{source_aggr_64};
        $create_opts{vserver}           = $opts{source_vserver};
        $create_opts{type}              = 'RW';
        $create_opts{'space-guarantee'} =
            $vdata->{space_guaranteed_src} ? 'volume' : 'none';
        # Source volumes get this option, so that the built-in default snapshot
        # policy doesn't get in our way. If the caller passes in this option,
        # their value will overwrite this later.
        $create_opts{'snapshot-policy'} = 'none';
        # Source volumes also need this option.
        $create_opts{'unix-permissions'} = '---rwxrwxrwx';
    } elsif ($vol_type eq 'destination') {
        $create_opts{command_interface} = $opts{destination_node};
        $create_opts{aggregate}         =
            ($vdata->{aggr_type_dst} eq '32-bit') ?
                $opts{destination_aggr_32} : $opts{destination_aggr_64};
        $create_opts{vserver}           = $opts{destination_vserver};
        $create_opts{type}              = 'DP';
        $create_opts{'space-guarantee'} =
            $vdata->{space_guaranteed_dst} ? 'volume' : 'none';
    } else {
        $Log->exit();
        NATE::Exceptions::Argument->throw(
            'Missing required argument "type" (or value unrecognized)'
        );
    }
    # Quick sanity-check: Do we have the minimum amount of data to create a
    # volume? Due to the optional nature of some of the parameters, they might
    # get this far without having set all of these:
    if (! ($create_opts{command_interface} && $create_opts{aggregate} &&
           $create_opts{vserver})) {
        $Log->exit();
        NATE::Exceptions::Argument->throw(
            'Missing one or more of the node, aggregate or vserver options'
        );
    }
    # Did they specify an explicit name? If not, we want a random name. But
    # we want the random name to have "rrs_" as a prefix, so we do it manually.
    if ($opts{volume_name}) {
        $create_opts{volume} = $opts{volume_name};
    } else {
        $create_opts{volume} =
            'rrs_' . NACL::STask::Volume->random_name_generator;
    }
    # Volume size defaults to $DEFAULT_VOLUME_SIZE for RRS:
    $create_opts{size} = $opts{volume_size} || $DEFAULT_VOLUME_SIZE;

    # Delete all the "known" passed-in arguments. Anything left after this will
    # be NACL::STask::Volume-specific options passed in by the user.
    delete @opts{qw(source_node destination_node source_aggr_32 source_aggr_64
                    destination_aggr_32 destination_aggr_64 source_vserver
                    destination_vserver volume_name volume_size variant client
                    transition_node transition_address transition_volume
                    variant_data nfs_options cifs_options nacltask_wait)};
    # Now copy anything left over verbatim:
    for (keys %opts) {
        $create_opts{$_} = $opts{$_};
    }

    # Attempt to create the volume:
    my $volume = NACL::STask::Volume->create(%create_opts);

    # Assuming that created without an exception, if it is a source volume we
    # still have to mount it (from the filer perspective) and apply storage
    # efficiency:
    if ($vol_type eq 'source') {
        if (! $create_opts{'junction-path'}) {
            $volume->mount('junction-path' => q{/} . $volume->volume);
        }
        $self->_apply_storage_efficiency(volume => $volume);
        # If this variant is doing SAN, the source volume is where the LUN
        # should be created.
        if ($vdata->{san}) {
            # If they are doing SAN on localhost, issue a warning
            if ($client_name eq 'localhost') {
                $Log->warn(
                    'You are doing SAN set-up on "localhost". This may be ' .
                    'a problem if you later try to connect to this LUN from ' .
                    'a different (more specific) host.'
                );
            }
            # Need a valid client obj for SAN stuff
            my $client = NACL::C::Client->new(name => $client_name);
            # Create the LUN on the filer side
            my ($lun, $igroup) = $self->_create_source_lun(
                variant     => $self->{variant},
                volume      => $volume,
                client      => $client,
                san_options => $san_options
            );
            my $dev_node = $self->_do_client_iscsi_discovery(
                volume      => $volume,
                lun         => $lun,
                client      => $client
            );
            $self->_create_lun_filesystem(
                device      => $dev_node,
                client      => $client,
                san_options => $san_options
            );

            # SANHACK:
            #
            # Save the things that are SAN-related, as we may need them if
            # this is a complex (multi-volume) operation:
            $self->{_san_lun}    = $lun;
            $self->{_san_igroup} = $igroup;
            $self->{_san_client} = $client;
        }
    }

    $Log->exit();
    return $volume;
}

###############################################################################
#
#   Sub Name:       _apply_storage_efficiency
#
#   Description:    Apply storage efficiency (SIS) commands to the given
#                   volume, as specified by the variant data.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional options (just
#                                                 "volume" in this case)
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub _apply_storage_efficiency {
    $Log->enter();
    my ($self, %opts) = @_;
    my $volume = $opts{volume};
    my $vdata = $self->{variant_data};

    # As far as I know, you always have to have dedupe before you can have
    # either type of compression. So I'm predicating this on the true-ness
    # of the 'dedupe' key.
    if ($vdata->{dedupe}) {
        # This enables dedupe and generally turns on efficiency for $volume:
        my $sis = NACL::STask::VolumeEfficiency->on(
            command_interface   => $volume->command_interface,
            vserver             => $volume->vserver,
            volume              => $volume->volume,
            nacltask_if_enabled => 'pass'
        );

        # Now determine which of compression/inline-compression should also
        # be enabled.
        my @opts;
        if ($vdata->{compression}) {
            push(@opts, compression => 'true');
        }
        if ($vdata->{inline_compression}) {
            push(@opts, 'inline-compression' => 'true');
        }
        # Only do the modify op is we're actually changing something:
        if (@opts) {
            $sis->config(@opts);
        }

        # Start up the efficiency engine. This should return almost immediately
        # since we're doing this before data is written:
        $sis->start(
            'scan-old-data' => 'true',
            nacltask_wait   => 1
        );
    }

    $Log->exit();
    return;
}

###############################################################################
#
#   Sub Name:       _populate_and_snapshot
#
#   Description:    Start data population on the volume, then create the
#                   requested number of snapshots while the data is being
#                   generated.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the operation
#                                                 ('volume', etc.)
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub _populate_and_snapshot {
    $Log->enter();
    my ($self, %opts) = @_;

    # Options from the caller, for volume and population-data size:
    my $volume        = $opts{volume};
    my $populate_size = $opts{populate_size} || $DEFAULT_POPULATE_SIZE;
    my $no_populate   = $opts{no_populate}   || 0;
    # What client to use when we populate the data:
    my $client_name   = $opts{client} || $self->{client} || 'localhost';
    my $nfs_options   = $opts{nfs_options};  # Default to undef
    my $cifs_options  = $opts{cifs_options}; # Default to undef
    my $san_options   = $opts{san_options};  # Default to undef
    # Total number of *numbered* snapshots we'll create. We subtract one from
    # the given number to account for our 'golden' snapshot at the end.
    my $total_snapshots = $self->{variant_data}->{snapshots_src} - 1;
    if ($opts{snapshots_used}) {
        $total_snapshots -= $opts{snapshots_used};
    }
    # How wide do numbers need to be (with leading zeros) for SS numbering?
    my $count_width = $total_snapshots ?
        (int(log($total_snapshots) / log(10)) + 1) : 0;
    # This is an sprinf format string for the snapshot names, it'll be given
    # the snapshot number.
    my $ss_pattern = "snapshot_rrs_%0${count_width}d";
    # How long to pause between each snapshot. Scaled over the number of
    # snapshots so that smaller numbers of snapshots aren't all clustered at
    # the start of the data.
    my $snooze = $total_snapshots ? int((229 / $total_snapshots) + 0.5) : 0;
    # If the intended destination is XDP, then the snapshots will need a label.
    my @ss_label = ($self->{variant_data}->{pedigree_dst} =~ /xdp/i) ?
        ('snapmirror-label' => 'hourly') : ();

    # Various params we'll need in controlling the population
    my ($io_handle, $data_handle);

    # If we already have a NACL::C::Client object from the SAN set-up, use it
    # here as well. No sense in having two...
    $io_handle = RRS_IO->new(
        volume  => $volume,
        # SANHACK:
        client  => $self->{_san_client} || $client_name,
        variant => $self->{variant_data}
    );
    if (! $no_populate) {
        # Create the mount-point
        $io_handle->create_mount(
            ($nfs_options)  ? (nfs_options  => $nfs_options)  : (),
            ($cifs_options) ? (cifs_options => $cifs_options) : (),
            ($san_options)  ? (san_options  => $san_options)  : (),
        );
        $data_handle = $io_handle->populate(
            size  => $populate_size,
            path  => 'RRS',
            async => 1
        );
    }

    # Loop to create the snapshots. This does not include the "golden"
    # snapshot-- that'll be created after the data population has completed.
    for my $ssnum (1 .. $total_snapshots) {
        NACL::STask::VolumeSnapshot->create(
            command_interface => $volume->command_interface,
            volume            => $volume->volume,
            vserver           => $volume->vserver,
            snapshot          => sprintf($ss_pattern, $ssnum),
            @ss_label
        );
        # Wait for the designated amount of time:
        Tharn::sleep($snooze);
    }

    if (! $no_populate) {
        # Wait for the data population to complete. If it fails, be sure
        # to delete the mount before re-throwing the exception.
        my $exception;
        try {
            $data_handle->wait_for_completion(timeout => 900);
        } catch NATE::BaseException with {
            $exception = shift;
            $Log->comment(
                'Exception caught from data population: ' .
                $exception->text .
                "\nUnmounting volume before exiting."
            );
        };
        $io_handle->delete_mount;
        if ($exception) {
            $exception->throw;
        }
    }

    # Now create the "golden" snapshot:
    NACL::STask::VolumeSnapshot->create(
        command_interface => $volume->command_interface,
        volume            => $volume->volume,
        vserver           => $volume->vserver,
        snapshot          => $self->get_golden_snapshot_label,
        @ss_label
    );

    $Log->exit();
    return;
}

###############################################################################
#
#   Sub Name:       _create_source_lun
#
#   Description:    Create the LUN for the volume on the filer. This just
#                   does the filer-side steps, it does not do the steps to
#                   mount the LUN on the client or create the initial
#                   filesystem.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the operation
#                                                 ('volume', etc.)
#
#   Returns:        Success:    NACL::STask::Lun instance
#                   Failure:    throws exception
#
###############################################################################
sub _create_source_lun {
    $Log->enter();
    my ($self, %opts) = @_;

    # Options from the caller, for variant, volume, client and SAN options
    my $variant     = $opts{variant};
    my $volume      = $opts{volume};
    my $client      = $opts{client};
    my $san_options = $opts{san_options};

    # We'll need the command-interface from $volume for filer commands:
    my $ci = $volume->command_interface;

    # Get LUN name and size, either from SAN options or from defaults:
    my $lun_name = $san_options->{lun_name} || $volume->volume . '_lun';
    my $lun_size = $san_options->{lun_size} || '50%';

    # Calculate the LUN size if it is in %, otherwise just validate it:
    if ($lun_size =~ /^(\d+)(%|mb?|gb?)$/i) {
        my $num = $1;
        my $unit = lc $2;
        my $vol_size = lc $volume->get_one_state_attribute('size');
        my ($vol_num, $vol_unit) = $vol_size =~ /^(\d+)(.*)$/;

        if ($unit eq q{%}) {
            if ($num > 100) {
                # Over 100%? Seriously?
                NATE::Exceptions::Argument->throw(
                    'SAN option "lun_size" cannot be over 100%'
                );
            }
            $lun_size = ($num / 100) * $vol_num;
            $lun_size .= $vol_unit;
        } else {
            # Check that it isn't larger than the volume itself. I *think*
            # "lun create" would throw the error, but I'd rather avoid the
            # mess entirely since I can test it here.
            my %units = (
                'm' => 1024 ** 2,
                'g' => 1024 ** 3,
                't' => 1024 ** 4,
            );
            my $vsize = $vol_num * $units{substr($vol_unit, 0, 1)};
            my $lsize = $num     * $units{substr($unit, 0, 1)};

            if ($lsize > $vsize) {
                NATE::Exceptions::Argument->throw(
                    "SAN option 'lun_size' ($lun_size) cannot exceed " .
                    "volume size ($vol_size)"
                );
            }
        }
    } else {
        NATE::Exceptions::Argument->throw(
            "Unparsable value for SAN option 'lun_size': $lun_size"
        );
    }

    # Determine OS type of the client. We need this to be a type that
    # "lun create" understands, so for now we're limiting to linux, solaris
    # and windows.
    my $ostype = $client->type;
    if ($ostype =~ /linux/) {
        $ostype = 'linux';
    } elsif ($ostype =~ /solaris/) {
        $ostype = 'solaris';
    } elsif ($ostype =~ /windows/) {
        $ostype = 'windows';
    } else {
        NATE::Exceptions::Argument->throw(
            "Client type '$ostype' is not one we can work with"
        );
    }

    # Get the iSCSI initiator name value from the client. Needed for igroup
    # creation. Yes, this is the only way currently for us to get this value.
    my $initiator;
    my $output = $client->execute(
        command => 'cat /etc/iscsi/initiatorname.iscsi'
    );
    if ($output =~ /^InitiatorName\s*=\s*([-.:\d[:alpha:]]+)/m) {
        $initiator = lc $1; # Node names are not case-sensitive
        if (length($initiator) > 223) {
            NATE::BaseException->throw(
                "InitiatorName '$initiator' is too long (spec says a max " .
                'of 223 bytes)'
            );
        }
    } else {
        NATE::BaseException->throw(
            'Cannot read /etc/iscsi/initiatorname.iscsi on client, or ' .
            "InitiatorName not set. Output was:\n\n$output"
        );
    }

    # Last piece: the igroup name. Just base it on the variant name.
    my $igroup_name = "${variant}_igroup";

    my $lun = NACL::STask::Lun->create(
        command_interface => $ci,
        lun               => $lun_name,
        volume            => $volume,
        vserver           => $volume->vserver,
        size              => $lun_size,
        ostype            => $ostype,
        if_exists         => 'purge',
        nacltask_verify   => 1
    );
    my $igroup = NACL::STask::LunIgroup->create(
        command_interface  => $ci,
        vserver            => $volume->vserver,
        igroup             => $igroup_name,
        protocol           => 'iscsi',
        ostype             => $ostype,
        initiator          => [ $initiator ],
        nacltask_if_exists => 'purge',
        nacltask_verify    => 1
    );
    # N.B. Not big on hard-coding lun-id, but for now we're only creating a
    # single LUN per volume anyway...
    $lun->map(
        igroup                    => $igroup_name,
        'lun-id'                  => 0,
        nacltask_if_mapped_igroup => 'remap',
        nacltask_verify           => 1
    );

    # Called in an array context, return both the LUN and igroup objects.
    # Otherwise, just return the LUN.
    $Log->exit();
    return wantarray ? ($lun, $igroup) : $lun;
}

###############################################################################
#
#   Sub Name:       _do_client_iscsi_discovery
#
#   Description:    Send the client-side commands needed to do the
#                   initialization/discovery for iSCSI.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the operation
#
#   Returns:        Success:    Device node on client this LUN is mapped to
#                   Failure:    throws exception
#
###############################################################################
sub _do_client_iscsi_discovery {
    $Log->enter();
    my ($self, %opts) = @_;

    # Options from the caller, for volume, LUN, client and SAN options
    my $volume      = $opts{volume};
    my $lun         = $opts{lun};
    my $client      = $opts{client};

    # We'll need the command-interface from $volume to determine the iSCSI
    # LIF's IP address:
    my $ci = $volume->command_interface;

    # Get the IP address of the first iscsi-protocol-capable LIF on the
    # node. We're using a canned query here, to get the iscsi data interfaces
    # that are up.
    my @ifaces = NACL::C::NetworkInterface->find(
        command_interface => $ci,
        canned_filters    => [ qw(up_data_iscsi_interfaces) ],
        allow_empty       => 1
    );
    if (! @ifaces) {
        # Why not just let find() throw? I want more control over the message.
        NATE::BaseException->throw(
            'No data LIFs found for the iscsi protocol, cannot do SAN ' .
            'discovery for the client.'
        );
    }
    my $iface_addr = $ifaces[0]->get_one_state_attribute('address');

    # Get the APISet object for $client. The commands we have to run are only
    # available at the APISet level (plus, they need to have root-level
    # privileges set first).
    my $apiset      = $client->apiset;
    my $old_priv    = $apiset->get_privilege;
    my $old_timeout = $apiset->get_timeout;

    # Do "discovery" of the valid targets. This entails sending an iscsiadm
    # command, then bouncing the iscsi service.
    try {
        # When scaled out to 20+ streams, these commands were sometimes hitting
        # the default time-out. Temporarily bump it up to 5 minutes.
        $apiset->set_timeout(timeout => 300);
        $apiset->set_privilege('privilege-level' => 'root');
        $apiset->iscsiadm(
            mode      => 1,
            discovery => 1,
            type      => 'sendtargets',
            'ip-port' => $iface_addr
        );
        $apiset->service_iscsi_restart;
    } finally {
        # Whether we succeed or fail, restore the old privilege level and
        # timeout:
        $apiset->set_privilege('privilege-level' => $old_priv);
        $apiset->set_timeout(timeout => $old_timeout);
    };

    # It appears that there can be a delay between the restart and the LUNs
    # actually showing up to the sanlun command. Give this about 3 tries
    # before we give up.
    my @luns;
    my $vserver = $lun->vserver;
    my $lunpath = $lun->path;
    for my $iter (1 .. 3) {
        @luns = NACL::C::Client::Sanlun->find(
            command_interface => $client,
            allow_empty       => 1,
            filter            => {
                controller   => $vserver,
                lun_pathname => $lunpath,
            }
        );
        last if @luns;
        $Log->comment(
            "Failed to find any LUN after iscsi restart, iteration $iter."
        );
        Tharn::snooze(10);
    }
    if (! @luns) {
        NATE::BaseException->throw(
            'Commands for iSCSI discovery completed successfully, but the ' .
            'LUN did not appear in the sanlun table.'
        );
    }

    # We will probably have seen two device-nodes for this LUN, due to a
    # cluster having two LIFs set up for iSCSI. But we only need one, so
    # (arbitrarily) use the first one.
    $Log->exit();
    return $luns[0]->get_one_state_attribute('device_filename');
}

###############################################################################
#
#   Sub Name:       _create_lun_filesystem
#
#   Description:    Create a filesystem on the LUN through the client
#                   command-interface. Currently only supports what can be
#                   done via NACL::C::Client::Mkfs.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the operation
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub _create_lun_filesystem {
    $Log->enter();
    my ($self, %opts) = @_;

    # Options from the caller, for device, client and SAN options
    my $device      = $opts{device};
    my $client      = $opts{client};
    my $san_options = $opts{san_options};

    my @fstype = $san_options->{fstype} ?
        (fstype => $san_options->{fstype}) : ();

    NACL::C::Client::Mkfs->mkfs(
        command_interface => $client,
        device            => $device,
        @fstype
    );

    $Log->exit();
    return;
}

###############################################################################
#
#   Sub Name:       _move_lun_mapping
#
#   Description:    Unmap the LUN as it currently is mapped, and map the LUN
#                   by the same path that exists on "volume".
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the operation
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub _move_lun_mapping {
    $Log->enter();
    my ($self, %opts) = @_;

    # Options from the caller, just the volume the new LUN is on. The SAN
    # elements are already stored on $self.
    my $volume = $opts{volume};
    # SANHACK:
    my $lun    = $self->{_san_lun};
    my $igroup = $self->{_san_igroup};

    # Start by finding the LUN on $volume with the same path as $lun:
    my $dst_lun = NACL::STask::Lun->find(
        command_interface => $volume->command_interface,
        filter            => {
            volume  => $volume,
            vserver => $volume->vserver,
            lun     => $lun->get_one_state_attribute('lun'),
        }
    );

    # $lun is already mapped, so just unmap it:
    $lun->unmap(igroup => $igroup);

    # Now map the new LUN, found from $volume. Not big on hard-coding lun-id,
    # but for now we're only creating a single LUN per volume anyway...
    $dst_lun->map(
        igroup                    => $igroup,
        'lun-id'                  => 0,
        nacltask_if_mapped_igroup => 'remap',
        nacltask_verify           => 1
    );

    # SANHACK:
    # $self->{_san_lun} needs to point to the "new" LUN, now
    $self->{_san_lun} = $dst_lun;

    $Log->exit();
    return;
}

###############################################################################
#
#   Sub Name:       _create_flexcache
#
#   Description:    Create a Flexcache volume attached to the provided
#                   volume.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the operation
#
#   Returns:        Success:    Flexcache volume object
#                   Failure:    throws exception
#
###############################################################################
sub _create_flexcache {
    $Log->enter();
    my ($self, %opts) = @_;

    my $volume = $opts{volume};
    if (! ($volume && $volume->isa('NACL::C::Volume'))) {
        NATE::Exceptions::Argument->throw(
            'Parameter "volume" not a volume object'
        );
    }

    my $volstate = $volume->state(
        requested_fields => [ qw(vserver aggregate size) ]
    );
    my $cache = NACL::STask::Volume->create(
        command_interface         => $volume->command_interface,
        volume                    => $volume->volume . '_flexcache',
        vserver                   => $volstate->vserver,
        aggregate                 => $volstate->aggregate,
        size                      => $volstate->size,
        'flexcache-origin-volume' => $volume->volume,
        type                      => 'DC',
        nacltask_if_exists        => 'purge'
    );

    $Log->exit();
    return $cache;
}

###############################################################################
#
# These routines are the volume-creators that work for both source-side and
# destination-side.
#
###############################################################################

###############################################################################
#
#   Sub Name:       _create_basic
#
#   Description:    Create and set up a source volume of pedigree "RW" or a
#                   destination volume of pedigree "DP", depending on the
#                   value of $opts{type}.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the creation
#
#   Returns:        Success:    NACL::STask::Volume object
#                   Failure:    throws exception
#
###############################################################################
sub _create_basic {
    $Log->enter();
    my ($self, %opts) = @_;

    # This is the simplest of the dual-side volume creation hooks. We just
    # create a vanilla volume, populate it and create snapshots according to
    # the variant specs (if it is a source volume), and return it.

    # These options should be kept out of the creation:
    my $populate_size = delete $opts{populate_size};
    my $no_populate   = delete $opts{no_populate};
    my $flexcache     = delete $opts{with_flexcache};
    my $nfs_options   = delete $opts{nfs_options};
    my $cifs_options  = delete $opts{cifs_options};
    # san_options should be passed to _create_volume if it is passed here, so
    # leave it in %opts:
    my $san_options   = $opts{san_options};

    # This will both create the volume and take care of storage efficiency:
    my $volume = $self->_create_volume(%opts);

    # Create a flexcache associated with this volume, if requested:
    if ($flexcache) {
        if ($opts{type} eq 'source') {
            # For a source volume, we do it immediately:
            $self->_create_flexcache(volume => $volume);
        } else {
            # But for a destination volume, we have to defer to the
            # 'postinitialize' callback, because we can't add the flexcache
            # until the volume has been initialized.
            $self->add_hook(
                hook     => 'postinitialize',
                callback => sub {
                    my ($self) = @_;

                    $self->_create_flexcache(volume => $volume);
                }
            );
        }
    }

    # If this is a source volume, do data population/snapshots on the volume
    # based on variant settings.
    if ($opts{type} eq 'source') {
        $self->_populate_and_snapshot(
            nfs_options   => $nfs_options,
            cifs_options  => $cifs_options,
            san_options   => $san_options,
            volume        => $volume,
            populate_size => $populate_size,
            no_populate   => $no_populate
        );
    }

    $Log->exit();
    return $volume;
}

###############################################################################
#
#   Sub Name:       _create_with_flexcache
#
#   Description:    Like _create_basic, but with the added step of creating a
#                   flexcache for the volume. Implemented by just adding an
#                   extra param to %opts and calling _create_basic.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the creation
#
#   Returns:        Success:    NACL::STask::Volume object
#                   Failure:    throws exception
#
###############################################################################
sub _create_with_flexcache {
    $Log->enter();
    my ($self, %opts) = @_;

    my $volume = $self->_create_basic(%opts, with_flexcache => 1);

    $Log->exit();
    return $volume;
}

###############################################################################
#
#   Sub Name:       _create_t_state
#
#   Description:    Create a volume as a transition-from-7-mode volume, using
#                   the parameters transition_node, transition_address and
#                   transition_volume to specify the source-point of the
#                   originating volume.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the creation
#
#   Returns:        Success:    NACL::STask::Volume object
#                   Failure:    throws exception
#
###############################################################################
sub _create_t_state {
    $Log->enter();
    my ($self, %opts) = @_;

    # A lot of this is copied from _create_basic, but it is just different-
    # enough that I couldn't directly re-use _create_basic. Apologies.

    # For the variants we cover here, we need these three values. Unless they
    # are all set, we have to throw an exception:
    my ($transition_node, $transition_address, $transition_volume) =
        _process_transition_opts(
            $opts{transition_node}    || q{},
            $opts{transition_address} || q{},
            $opts{transition_volume}  || q{}
        );

    # These need to be withheld from the volume-creation options. Not sure
    # yet if I'll actually need them, though. Might not need to save them.
    my $populate_size = delete $opts{populate_size};
    my $no_populate   = delete $opts{no_populate};
    my $nfs_options   = delete $opts{nfs_options};
    my $cifs_options  = delete $opts{cifs_options};
    my $san_options   = delete $opts{san_options};

    # We'll always wait for volume creation, and unless they pass an explicit
    # nacltask_if_exists we always purge an existing volume. Also, for these
    # cases, "type" (for vol creation) is always DP.
    my %create_opts = (
        nacltask_wait      => 1,
        nacltask_if_exists => 'purge',
        type               => 'DP',
    );
    my $vdata = $self->{variant_data};

    # Set up command_interface, vserver and type based on the type of volume
    # we're creating. Also choose the correct values from the variant data for
    # space-guarantee and aggregate.
    my $vol_type = delete $opts{type};
    if ($vol_type eq 'source') {
        $create_opts{command_interface} = $opts{source_node};
        $create_opts{aggregate}         =
            ($vdata->{aggr_type_src} eq '32-bit') ?
                $opts{source_aggr_32} : $opts{source_aggr_64};
        $create_opts{vserver}           = $opts{source_vserver};
        $create_opts{'space-guarantee'} =
            $vdata->{space_guaranteed_src} ? 'volume' : 'none';
    } elsif ($vol_type eq 'destination') {
        $create_opts{command_interface} = $opts{destination_node};
        $create_opts{aggregate}         =
            ($vdata->{aggr_type_dst} eq '32-bit') ?
                $opts{destination_aggr_32} : $opts{destination_aggr_64};
        $create_opts{vserver}           = $opts{destination_vserver};
        $create_opts{'space-guarantee'} =
            $vdata->{space_guaranteed_dst} ? 'volume' : 'none';
    } else {
        $Log->exit();
        NATE::Exceptions::Argument->throw(
            'Missing required argument "type" (or value unrecognized)'
        );
    }
    # Quick sanity-check: Do we have the minimum amount of data to create a
    # volume? Due to the optional nature of some of the parameters, they might
    # get this far without having set all of these:
    if (! ($create_opts{command_interface} && $create_opts{aggregate} &&
           $create_opts{vserver})) {
        $Log->exit();
        NATE::Exceptions::Argument->throw(
            'Missing one or more of the node, aggregate or vserver options'
        );
    }
    # Did they specify an explicit name? If not, we want a random name. But
    # we want the random name to have "rrs_" as a prefix, so we do it manually.
    if ($opts{volume_name}) {
        $create_opts{volume} = $opts{volume_name};
    } else {
        $create_opts{volume} =
            'rrs_' . NACL::STask::Volume->random_name_generator;
    }
    # Volume size defaults to $DEFAULT_VOLUME_SIZE for RRS:
    $create_opts{size} = $opts{volume_size} || $DEFAULT_VOLUME_SIZE;

    # Delete all the "known" passed-in arguments. Anything left after this will
    # be NACL::STask::Volume-specific options passed in by the user.
    delete @opts{qw(source_node destination_node source_aggr_32 source_aggr_64
                    destination_aggr_32 destination_aggr_64 source_vserver
                    destination_vserver volume_name volume_size variant client
                    transition_node transition_address transition_volume
                    variant_data nfs_options cifs_options nacltask_wait)};
    # Now copy anything left over verbatim:
    for (keys %opts) {
        $create_opts{$_} = $opts{$_};
    }

    # Attempt to create the volume:
    my $volume = NACL::STask::Volume->create(%create_opts);

    # OK, we have the volume now. At this point, we need to make a transition
    # peer relationship between our chosen vserver and the 7-mode side.

    my ($vpt_obj, $snapmirror);
    try {
        # Note that we've already determined the C-Mode end-point earlier, when
        # we chose the values for %create_opts. So I'm re-using those here.
        require NACL::C::VserverPeerTransition;
        $vpt_obj = NACL::C::VserverPeerTransition->create(
            command_interface    => $create_opts{command_interface},
            'local-vserver'      => $create_opts{vserver},
            'src-filer-name'     => $transition_node,
            $transition_address ?
                ('multi-path-address' => $transition_address) : ()
        );

        # Create a snapmirror relationship between the volume we created on
        # this side, and the specified volume on the 7-mode node:
        $snapmirror = NACL::STask::Snapmirror->create(
            command_interface  => $create_opts{command_interface},
            'source-path'      => "$transition_node:$transition_volume",
            'destination-path' => $volume,
            type               => 'TDP',
            throttle           => 'unlimited',
            do_initialize      => 0,
            nacltask_wait      => 1
        );

        # I'm doing the initialize separately from create because I want to
        # know more clearly, from the stacktrace, where any errors occur.
        # Default the time-out to 20min, but if they passed in a timeout use it
        # instead.
        my $timeout = $opts{'method-timeout'} || 1200;
        $snapmirror->initialize('method-timeout' => $timeout);

        # Lastly, if we got this far, do a snapmirror break. This should give
        # us a RW transition volume, as pointed to by $volume.
        $snapmirror->break;

        # The "finally" block will take care of deleting the snapmirror and
        # the vserver transition peering.
    } finally {
        # If we have a problem above, we want to remove the temp snapmirror
        # and the vserver transition peering before we actually throw the
        # exception. This also takes care of those tasks when everything above
        # worked correctly.
        if ($snapmirror) {
            $snapmirror->delete;
        }
        if ($vpt_obj) {
            $vpt_obj->delete;
        }
    };

    # We now have what we want in $volume. But if this is intended to be a
    # source volume, it must also have UNIX permissions set on it and have
    # the junction-point set. It also needs to get the "golden" snapshot
    # created, as well.
    if ($vol_type eq 'source') {
        # Create the "golden" snapshot:
        NACL::STask::VolumeSnapshot->create(
            command_interface => $volume->command_interface,
            volume            => $volume->volume,
            vserver           => $volume->vserver,
            snapshot          => $self->get_golden_snapshot_label
        );

        $volume->modify('unix-permissions' => '---rwxrwxrwx');
        if ($opts{'junction-path'}) {
            $volume->mount('junction-path' => $opts{'junction-path'});
        } else {
            $volume->mount('junction-path' => q{/} . $volume->volume);
        }
    }

    $Log->exit();
    return $volume;
}

###############################################################################
#
#   Sub Name:       _process_transition_opts
#
#   Description:    Process the three transition-related options and return
#                   the "laundered" values. Note that this is a subroutine,
#                   not a method. There is no "$self" in the args list.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $node     in      scalar    Name of the node hostrec, or IP
#                   $address  in      scalar    IP, or key into hostrec hashref
#                   $volume   in      scalar    Name of transition volume
#
#   Globals:        $IP_ADDR
#
#   Returns:        Success:    tuple of (node, address, volume)
#                   Failure:    throws exception
#
###############################################################################
sub _process_transition_opts {
    $Log->enter();
    my ($node, $address, $volume) = @_;
    my ($t_node, $t_address, $hostrec);

    if (! ($node && $volume)) {
        $Log->exit();
        NATE::Exceptions::Argument->throw(
            'Missing at least one of transition_node or transition_volume'
        );
    }

    if (ref $node) {
        # Node is an object, it must be NACL::C::SystemNode or a subclass of
        # that. We get the $hostrec object from it, and the default IP to use
        # as the "real" value for transition_node.
        if ($node->isa('NACL::C::SystemNode')) {
            $hostrec = $node->hostrec;
            $t_node = $hostrec->default_ip;
        } else {
            NATE::Exceptions::Argument->throw(
                'If passed as a reference, transition_node must be a ' .
                'NACL::C::SystemNode or derived subclass.'
            );
        }
    } elsif ($node =~ $IP_ADDR) {
        # Node is already a valid IP address. Use it as the "real" value for
        # transition_node, but we can't derive a hostrec from it.
        $t_node = $node;
    } else {
        # Assume that $node is the name of the hostrec file, and use it to
        # create a NACL::C::Node object. We will then get the hostrec object
        # from that and use it to get the default IP to use as the "real" value.
        $hostrec = NACL::C::Node->new(name => $node)->hostrec;
        $t_node = $hostrec->default_ip;
    }

    # We now have $t_node (or we've thrown an exception). Try to get the
    # address. Note that address is optional, so it's fine for it to just not
    # be set.
    if ($address =~ $IP_ADDR) {
        # If it is a valid IP address, just use it directly.
        $t_address = $address;
    } elsif ($address) {
        # If it isn't an IP address, it should be one (possibly both) of a
        # name and/or hash key. The name (if present) refers to an alternate
        # hostrec file. The hash key (if present) refers to a specific key
        # within either the alternate hostrec or the node hostrec.
        my ($name, $key) = split(/:/, $address, 2);
        if (! $name && ! $key) {
            # If they passed anything, at least one of those two should have
            # been set.
            NATE::Exceptions::Argument->throw(
                "Value '$address' for transition_address is not valid"
            );
        }

        if ($name) {
            # A specific hostrec name was passed in for transition_address.
            # Replace existing value of $hostrec with this one.
            $hostrec = NACL::C::Node->new(name => $name)->hostrec;
            # If there is no $key, this will be our value. If there is a $key,
            # this will be overwritten.
            $t_address = $hostrec->default_ip;
        }

        if ($key) {
            # There was a :KEY passed in, either in addition to or instead of
            # the hostrec name. Either way, we have already set $hostrec to
            # what we want to extract $key from. If $hostrec is not set, then
            # that means that transition_node was an address and there was no
            # explicit name in transition_address, which means we can't use
            # $key.
            if ($hostrec) {
                if (! ($t_address = $hostrec->{$key})) {
                    NATE::Exceptions::Argument->throw(
                        "Key '$key' not found in hostrec or has null value"
                    );
                }
                if ($t_address !~ $IP_ADDR) {
                    NATE::Exceptions::Argument->throw(
                        "Hostrec value '$t_address' (for key $key) is not a " .
                        'valid IP address.'
                    );
                }
            } else {
                NATE::Exceptions::Argument->throw(
                    'Cannot specify transition_address as a key without a ' .
                    'named node for either transition_node or ' .
                    'transition_address.'
                );
            }
        }
    } else {
        # No address was given, so just assign an empty string so we don't
        # get any stray "uninitialized" warnings.
        $t_address = q{};
    }

    $Log->exit();
    return ($t_node, $t_address, $volume);
}

###############################################################################
#
# These routines are the creators specific to source-side volumes.
#
###############################################################################

###############################################################################
#
#   Sub Name:       _create_dp_to_rw
#
#   Description:    Create a source volume by first creating a snapmirror
#                   relationship (on the source node), populating it with
#                   data and snapshots, then breaking the relationship to
#                   convert the DP volume to a RW. Purge the throw-away
#                   source volume and return the new RW volume.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the creation
#
#   Returns:        Success:    NACL::STask::Volume object
#                   Failure:    throws exception
#
###############################################################################
sub _create_dp_to_rw {
    $Log->enter();
    my ($self, %opts) = @_;

    # This is much more involved than the previous. For this one we have to:
    #
    # 1. Create a dummy source, with an anon name, with all the source attrs
    # 2. Create a DP dest using mostly source attrs, with any requested name
    # 3. Set up snapmirror
    # 4. Populate/snapshot
    # 5. Update, to transfer snapshots
    # 6. Do snapmirror break
    # 7. Purge the dummy source
    # 8. Set any storage efficiency options on the now-RW dest
    # 9. Return this volume as if it were the source

    # These options should be kept out of the creation:
    my $populate_size = delete $opts{populate_size};
    my $no_populate   = delete $opts{no_populate};
    my $nfs_options   = delete $opts{nfs_options};
    my $cifs_options  = delete $opts{cifs_options};
    my $san_options   = delete $opts{san_options};
    # Don't want volume_name (if it was passed) passed on to _create_volume:
    my $volume_name   = delete $opts{volume_name};
    # What to do if a volume we're trying to create already exists:
    my $if_exists     = delete $opts{nacltask_if_exists} || 'purge';

    # We'll need to know some details about this variant
    my $vdata  = $self->{variant_data};

    # If they didn't specify a name, create a random one but prepend 'rrs_':
    if (! $volume_name) {
        $volume_name = 'rrs_' . NACL::STask::Volume->random_name_generator;
    }
    # Because the source is not the "real" source, we can't let the SAN LUN
    # be named based on it. If they haven't passed an explicit LUN name,
    # add it to $san_options based on $volume_name.
    if ($vdata->{san} && ! $san_options->{lun_name}) {
        $san_options->{lun_name} = "${volume_name}_lun";
    }

    # 1. Create the (temp) source volume and take care of storage efficiency.
    my $tmp_volume = $self->_create_volume(%opts, san_options => $san_options);

    # 2. Create the destination volume. Cannot use _create_volume for this.
    my $volume = NACL::STask::Volume->create(
        command_interface  => $opts{source_node},
        volume             => $volume_name,
        aggregate          =>
            ($vdata->{aggr_type_src} eq '32-bit') ?
                $opts{source_aggr_32} : $opts{source_aggr_64},
        vserver            => $opts{source_vserver},
        type               => 'DP',
        'space-guarantee'  =>
            $vdata->{space_guaranteed_src} ? 'volume' : 'none',
        nacltask_wait      => 1,
        nacltask_if_exists => $if_exists
    );

    # 3. Set up snapmirror. Don't initialize yet.
    my $snapmirror = NACL::STask::Snapmirror->create(
        command_interface  => $opts{source_node},
        source_ci          => $opts{source_node},
        'source-path'      => $tmp_volume,
        'destination-path' => $volume,
        type               => 'DP',
        do_initialize      => 0,
        nacltask_wait      => 1
    );

    # 4. Populate and create snapshots.
    # Do data population/snapshots on the volume, based on variant settings:
    $self->_populate_and_snapshot(
        nfs_options   => $nfs_options,
        cifs_options  => $cifs_options,
        san_options   => $san_options,
        volume        => $tmp_volume,
        populate_size => $populate_size,
        no_populate   => $no_populate
    );

    # 5. Initialize, to transfer snapshots
    $snapmirror->initialize(
        'source-snapshot' => $self->get_golden_snapshot_label,
        nacltask_wait     => 1
    );

    # 6. Do snapmirror break & purge
    $snapmirror->break;
    $snapmirror->purge;

    # 6 1/2. If this is SAN, move the LUN mapping from $tmp_volume to $volume
    if ($vdata->{san}) {
        $self->_move_lun_mapping(volume => $volume);
        # We have to re-do the iSCSI discovery on the client. Note that the LUN
        # and client objects are taken from the private-key cache values that
        # were set up by the previous parts of SAN set-up.
        $self->_do_client_iscsi_discovery(
            volume      => $volume,
            # SANHACK:
            lun         => $self->{_san_lun},
            client      => $self->{_san_client}
        );
    }

    # 7. Purge the dummy source
    $tmp_volume->purge;

    # 8. Set the UNIX perms, mount the volume (setting the junction-path),
    #    then set any storage efficiency options on the now-RW dest
    $volume->modify('unix-permissions' => '---rwxrwxrwx');
    if ($opts{'junction-path'}) {
        $volume->mount('junction-path' => $opts{'junction-path'});
    } else {
        $volume->mount('junction-path' => q{/} . $volume->volume);
    }
    $self->_apply_storage_efficiency(volume => $volume);

    # 9. Return this volume as if it were the source
    $Log->exit();
    return $volume;
}

###############################################################################
#
#   Sub Name:       _create_flexclone
#
#   Description:    Create a source volume from a RW flexclone volume. Create
#                   a dummy source and a clone from that. Then do the data
#                   population and snapshots on the clone. Return the clone
#                   as the volume, but do not purge the original source.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the creation
#
#   Returns:        Success:    NACL::STask::Volume object
#                   Failure:    throws exception
#
###############################################################################
sub _create_flexclone {
    $Log->enter();
    my ($self, %opts) = @_;

    # For this one we have to:
    #
    # 1. Create a dummy source, with an anon name, with all the source attrs
    # 2. Create a flexclone on the source
    # 3. Set any storage efficiency options on the clone
    # 4. Populate/snapshot on the clone
    # 5. Return this volume as if it were the source

    # These options should be kept out of the creation:
    my $populate_size = delete $opts{populate_size};
    my $no_populate   = delete $opts{no_populate};
    my $nfs_options   = delete $opts{nfs_options};
    my $cifs_options  = delete $opts{cifs_options};
    my $san_options   = delete $opts{san_options};
    # Don't want volume_name or junction-path (if either were passed) passed on
    # to _create_volume:
    my $volume_name   = delete $opts{volume_name};
    my $junction_path = delete $opts{'junction-path'};
    # What to do if the volume we are trying to create already exists:
    my $if_exists     = delete $opts{nacltask_if_exists} || 'purge';

    # We'll need to know some details about this variant
    my $vdata  = $self->{variant_data};

    # If they didn't specify a name, create a random one but prepend 'rrs_':
    if (! $volume_name) {
        $volume_name =
            'rrs_' . NACL::STask::VolumeClone->random_name_generator;
    }
    # Because the source is not the "real" source, we can't let the SAN LUN
    # be named based on it. If they haven't passed an explicit LUN name,
    # add it to $san_options based on $volume_name.
    if ($vdata->{san} && ! $san_options->{lun_name}) {
        $san_options->{lun_name} = "${volume_name}_lun";
    }

    # 1. Create the (temp) source volume.
    my $tmp_volume = $self->_create_volume(%opts, san_options => $san_options);

    # 2. Create the flexclone volume.
    my $volume = NACL::STask::VolumeClone->create(
        command_interface  => $opts{source_node},
        flexclone          => $volume_name,
        'parent-volume'    => $tmp_volume->volume,
        vserver            => $opts{source_vserver},
        type               => 'RW',
        'space-guarantee'  =>
            $vdata->{space_guaranteed_src} ? 'volume' : 'none',
        nacltask_wait      => 1,
        nacltask_if_exists => $if_exists
    );
    # Now that we have it as a clone, "find" it as a regular volume so that
    # we can do population, storage efficiency, etc. on it:
    $volume = NACL::STask::Volume->find(
        command_interface => $opts{source_node},
        filter => {
            volume  => $volume_name,
            vserver => $opts{source_vserver},
        }
    );

    # 2 1/2. If this is SAN, move the LUN mapping from $tmp_volume to $volume
    if ($vdata->{san}) {
        $self->_move_lun_mapping(volume => $volume);
        # We have to re-do the iSCSI discovery on the client. Note that the LUN
        # and client objects are taken from the private-key cache values that
        # were set up by the previous parts of SAN set-up.
        $self->_do_client_iscsi_discovery(
            volume      => $volume,
            # SANHACK:
            lun         => $self->{_san_lun},
            client      => $self->{_san_client}
        );
    }

    # 3. Set any storage efficiency options on the flexclone
    $self->_apply_storage_efficiency(volume => $volume);

    # 4. Populate and create snapshots.
    # First, the clone needs unix-permissions and junction-path set:
    $volume->modify('unix-permissions' => '---rwxrwxrwx');
    if ($junction_path) {
        $volume->mount('junction-path' => $junction_path);
    } else {
        $volume->mount('junction-path' => q{/} . $volume->volume);
    }
    # Do data population/snapshots on the volume, based on variant settings:
    $self->_populate_and_snapshot(
        nfs_options    => $nfs_options,
        cifs_options   => $cifs_options,
        san_options    => $san_options,
        volume         => $volume,
        populate_size  => $populate_size,
        no_populate    => $no_populate,
        snapshots_used => 1
    );

    # 5. Return this volume as if it were the source
    $Log->exit();
    return $volume;
}

###############################################################################
#
#   Sub Name:       _create_dp_flexclone
#
#   Description:    Create a source volume by first creating a snapmirror
#                   relationship (on the source node), initializing it to
#                   make the destination useful, then creating a clone from
#                   the destination.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the creation
#
#   Returns:        Success:    NACL::STask::Volume object
#                   Failure:    throws exception
#
###############################################################################
sub _create_dp_flexclone {
    $Log->enter();
    my ($self, %opts) = @_;

    # For this one we have to:
    #
    # 1. Create a dummy source, with an anon name, with all the source attrs
    # 2. Create a dummy dest using mostly source attrs
    # 3. Set up snapmirror and initialize
    # 4. Create flexclone from the dest volume
    # 5. Set any storage efficiency options on the clone/volume object
    # 6. Populate/snapshot the clone volume
    # 7. Return this volume as the source

    # These options should be kept out of the creation:
    my $populate_size = delete $opts{populate_size};
    my $no_populate   = delete $opts{no_populate};
    my $nfs_options   = delete $opts{nfs_options};
    my $cifs_options  = delete $opts{cifs_options};
    # Don't want volume_name or junction-path (if either were passed) passed on
    # to _create_volume:
    my $volume_name   = delete $opts{volume_name};
    my $junction_path = delete $opts{'junction-path'};
    # san_options should be passed to _create_volume if it is passed here, so
    # leave it in %opts:
    my $san_options   = $opts{san_options};
    # What to do if the volume we're creating already exists:
    my $if_exists     = delete $opts{nacltask_if_exists} || 'purge';

    # We'll need to know some details about this variant
    my $vdata  = $self->{variant_data};

    # If they didn't specify a name, create a random one but prepend 'rrs_':
    if (! $volume_name) {
        $volume_name =
            'rrs_' . NACL::STask::VolumeClone->random_name_generator;
    }

    # 1. Create the (temp) source volume.
    my $tmp_src_volume = $self->_create_volume(%opts);
    # We will need a known snapshot on the dest volume when we create the
    # flexclone, so just go with a known snapshot from the outset.
    NACL::STask::VolumeSnapshot->create(
        command_interface => $opts{source_node},
        vserver           => $opts{source_vserver},
        volume            => $tmp_src_volume->volume,
        snapshot          => '__dp_flexclone__'
    );

    # 2. Create the destination volume. Cannot use _create_volume for this.
    #    Note that 'nacltask_if_exists' is explicitly set to 'purge', since
    #    this is one of the temp volumes.
    my $tmp_dst_name   = 'rrs_' . NACL::STask::Volume->random_name_generator;
    my $tmp_dst_volume = NACL::STask::Volume->create(
        command_interface  => $opts{source_node},
        aggregate          =>
            ($vdata->{aggr_type_src} eq '32-bit') ?
                $opts{source_aggr_32} : $opts{source_aggr_64},
        vserver            => $opts{source_vserver},
        volume             => $tmp_dst_name,
        type               => 'DP',
        'space-guarantee'  =>
            $vdata->{space_guaranteed_src} ? 'volume' : 'none',
        nacltask_wait      => 1,
        nacltask_if_exists => 'purge'
    );

    # 3. Set up snapmirror, don't initialize.
    my $snapmirror = NACL::STask::Snapmirror->create(
        command_interface  => $opts{source_node},
        source_ci          => $opts{source_node},
        'source-path'      => $tmp_src_volume,
        'destination-path' => $tmp_dst_volume,
        type               => 'DP',
        do_initialize      => 0,
        nacltask_wait      => 1
    );
    # Have to do an explicit initialize so that we init from the known
    # snapshot:
    $snapmirror->initialize(
        'source-snapshot' => '__dp_flexclone__',
        nacltask_wait     => 1
    );

    # 4. Create the flexclone volume.
    my $volume = NACL::STask::VolumeClone->create(
        command_interface  => $opts{source_node},
        vserver            => $opts{source_vserver},
        flexclone          => $volume_name,
        'parent-volume'    => $tmp_dst_name,
        'parent-snapshot'  => '__dp_flexclone__',
        type               => 'RW',
        'space-guarantee'  =>
            $vdata->{space_guaranteed_src} ? 'volume' : 'none',
        nacltask_wait      => 1,
        nacltask_if_exists => $if_exists
    );
    # Now that we have it as a clone, "find" it as a regular volume so that
    # we can do population, storage efficiency, etc. on it:
    $volume = NACL::STask::Volume->find(
        command_interface => $opts{source_node},
        filter => {
            volume  => $volume_name,
            vserver => $opts{source_vserver},
        }
    );

    # 5. Set any storage efficiency options on the flexclone
    $self->_apply_storage_efficiency(volume => $volume);

    # 6. Populate and create snapshots.
    # First, the clone needs unix-permissions and junction-path set:
    $volume->modify('unix-permissions' => '---rwxrwxrwx');
    if ($junction_path) {
        $volume->mount('junction-path' => $junction_path);
    } else {
        $volume->mount('junction-path' => q{/} . $volume->volume);
    }
    # Do data population/snapshots on the volume, based on variant settings:
    $self->_populate_and_snapshot(
        nfs_options    => $nfs_options,
        cifs_options   => $cifs_options,
        san_options    => $san_options,
        volume         => $volume,
        populate_size  => $populate_size,
        no_populate    => $no_populate,
        snapshots_used => 1
    );

    # 7. Return this volume as if it were the source
    $Log->exit();
    return $volume;
}

###############################################################################
#
#   Sub Name:       _create_from_tape
#
#   Description:    Create a source volume (populated, snapshot'd) by faking
#                   a tape session with NDMP.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the creation
#
#   Returns:        Success:    NACL::STask::Volume object
#                   Failure:    throws exception
#
###############################################################################
sub _create_from_tape {
    $Log->enter();
    my ($self, %opts) = @_;

    # For this one we have to:
    #
    # 1. Create a dummy source
    # 2. Populate/snapshot dummy source
    # 3. Create dest as intended volume using mostly source attrs, restrict it
    # 4. Activate NDMP on file, execute ndmpcopy from the client, deactivate
    #    NDMP
    # 5. Make dest writable by doing a snapmirror break
    # 6. Return this volume as the source

    # Only need these libraries for this specific variant:
    require NACL::C::SystemServicesNdmp;
    require NACL::C::SystemServicesNdmpLegacy;

    # These options should be kept out of the creation:
    my $populate_size = delete $opts{populate_size};
    my $no_populate   = delete $opts{no_populate};
    my $nfs_options   = delete $opts{nfs_options};
    my $cifs_options  = delete $opts{cifs_options};
    my $san_options   = delete $opts{san_options};
    # Don't want volume_name or junction-path (if either were passed) passed on
    # to _create_volume:
    my $volume_name   = delete $opts{volume_name};
    my $junction_path = delete $opts{'junction-path'};
    # What client to use when we run ndmpcopy:
    my $client_name   = $opts{client} || $self->{client} || 'localhost';
    my $client        = NACL::C::Client->new(name => $client_name);
    # What to do if the volume we're creating already exists:
    my $if_exists     = delete $opts{nacltask_if_exists} || 'purge';

    # We'll need to know some details about this variant
    my $vdata  = $self->{variant_data};

    # If they didn't specify a name, create a random one but prepend 'rrs_':
    if (! $volume_name) {
        $volume_name = 'rrs_' . NACL::STask::Volume->random_name_generator;
    }
    # Because the source is not the "real" source, we can't let the SAN LUN
    # be named based on it. If they haven't passed an explicit LUN name,
    # add it to $san_options based on $volume_name.
    if ($vdata->{san} && ! $san_options->{lun_name}) {
        $san_options->{lun_name} = "${volume_name}_lun";
    }

    # 1. Create the (temp) source volume.
    my $tmp_volume = $self->_create_volume(%opts, san_options => $san_options);

    # 2. Populate/snapshot the temp source
    $self->_populate_and_snapshot(
        nfs_options    => $nfs_options,
        cifs_options   => $cifs_options,
        san_options    => $san_options,
        volume         => $tmp_volume,
        populate_size  => $populate_size,
        no_populate    => $no_populate
    );

    # 3. Create the destination volume. Cannot use _create_volume for this.
    my $volume = NACL::STask::Volume->create(
        command_interface  => $opts{source_node},
        aggregate          =>
            ($vdata->{aggr_type_src} eq '32-bit') ?
                $opts{source_aggr_32} : $opts{source_aggr_64},
        vserver            => $opts{source_vserver},
        volume             => $volume_name,
        type               => 'DP',
        'space-guarantee'  =>
            $vdata->{space_guaranteed_src} ? 'volume' : 'none',
        nacltask_wait      => 1,
        nacltask_if_exists => $if_exists
    );
    # Restrict the dest volume before the ndmpcopy
    $volume->restrict;

    # 4. Activate NDMP on the filer, execute ndmpcopy from the client,
    #    deactivate NDMP
    try {
        NACL::C::SystemServicesNdmp->on(
            command_interface => $opts{source_node},
            node              => q{*}
        );
        NACL::C::SystemServicesNdmpLegacy->on(
            command_interface => $opts{source_node}
        );

        my $iface = NACL::C::NetworkInterface->find(
            command_interface => $opts{source_node},
            filter            => {
                role           => 'node-mgmt',
                'status-oper'  => 'up',
                'status-admin' => 'up',
                'home-node'    => $opts{source_node}->node,
            }
        );
        my $addr = $iface->get_one_state_attribute('address');
        my $source_name = $tmp_volume->volume;
        my $command = '/usr/software/rats/bin/ndmpcopy -t smtape -e ' .
            'SMTAPE_CMODE=Y -aa root:admin -at md5 ' .
            "$addr:/$opts{source_vserver}/$source_name " .
            "$addr:/$opts{source_vserver}/$volume_name";
        my $tries   = 0;
        my $success = 0;

        while ($tries++ < 5) {
            my $output = $client->execute(
                command => $command,
                timeout => 1200
            );
            if ($output =~ /Starting copy.*Transfer successful/is) {
                $success++;
                last;
            } elsif ($output =~ /Issue 'ndmpd on' on the source filer/i) {
                $Log->comment(
                    "NDMP not yet ready on the filer (try $tries of 5)"
                );
                Tharn::snooze(30);
            } else {
                NATE::BaseException->throw(
                    "Unknown error from ndmpcopy:\n\n$output"
                );
            }
        }

        # Sanity check: Did we see certain lines of output?
        if (! $success) {
            NATE::BaseException->throw(
                'When calling ndmpcopy, did not see both "Starting copy" and' .
                ' "Transfer successful" after 5 tries.'
            );
        }
    } finally {
        NACL::C::SystemServicesNdmpLegacy->off(
            command_interface => $opts{source_node}
        );
        NACL::C::SystemServicesNdmp->off(
            command_interface => $opts{source_node},
            node              => q{*}
        );
    };

    # 5. Make the dest RW by doing a snapmirror break. Yes, it works.
    NACL::STask::Snapmirror->break(
        command_interface  => $opts{source_node},
        'destination-path' => $volume
    );
    # If it is SAN, move the mapping from $tmp_volume LUN to $volume LUN
    if ($vdata->{san}) {
        $self->_move_lun_mapping(volume => $volume);
        # We have to re-do the iSCSI discovery on the client. Note that the LUN
        # and client objects are taken from the private-key cache values that
        # were set up by the previous parts of SAN set-up.
        $self->_do_client_iscsi_discovery(
            volume      => $volume,
            # SANHACK:
            lun         => $self->{_san_lun},
            client      => $self->{_san_client}
        );
    }
    # Also need to set permissions and create a junction point now that it is
    # RW.
    $volume->modify('unix-permissions' => '---rwxrwxrwx');
    if ($junction_path) {
        $volume->mount('junction-path' => $junction_path);
    } else {
        $volume->mount('junction-path' => "/$volume_name");
    }
    # AND (almost forgot this), get rid of the temp source volume
    $tmp_volume->purge;

    # 6. Return the dest volume as the final result.
    $Log->exit();
    return $volume;
}

###############################################################################
#
# These routines are the creators specific to destination-side volumes.
#
###############################################################################

1;
