#! /usr/bin/perl

#package dp::snapmirror::paloma::san::DPSan;
package DPSanUtil;

use strict;
use Carp;
use Tharn;
use GNOTE;
use GNOTE::FileSystem;
use Data::Dumper;
use San::San;
use San::SanClient::SanIo;
use Filesys::Tree qw/tree/;

use base qw( Exporter );

# Global variables which will not be directly used.
my $Step = 1;
my %SanSnapMirrorHash;

sub _initialize_params
{
    # All Filer/Vserver specific defaults.
    param('FILER_SRC', '-default', undef);
    param('FILER_DST', '-default', undef);

    # All client specific defaults 
    param('CLIENT', '-default', undef);
    param('RESCAN_TIMEOUT', '-default', 5);
    param('OS_TYPE', '-default', "windows");
    param('CLIENT_IO', '-default', 1);
    param('DEST_CLIENT_IO', '-default', 0);

    my $def_src_ig = $main::VSERVER_SRC."_".$main::OS_TYPE."_ig";
    my $def_dst_ig = $main::VSERVER_DST."_".$main::OS_TYPE."_ig";
    param('IGROUP_SRC', '-default', $def_src_ig);
    param('IGROUP_DST', '-default', $def_dst_ig);

    param('DEV_TYPE', '-default', "raw");
    param('MP_TYPE', '-default', "none");
    param('CONFIG_ISCSI_SUBNET_MATCHING', '-default', "0");         # really not sure if we have
    param('CONFIG_ISCSI_DISCOVERY_PRECLEANUP', '-default', "0");    # to keep these variable
    param('CONFIG_ISCSI_WIN_MPIO', '-default', "0");                # here all the time

    # All Snapmirror specific defaults
    param('SHOW_WORKAROUND', '-default', 0);

    # All Lun specific defaults
    param('INITIAL_LUN_COUNT', '-default', 3);
    param('LUN_SIZE', '-default', "90m");

    return;
}

sub new 
{
    my ($class, $src_vol, $dst_vol) = @_;

    my $self = {};
    bless $self, $class;

    $self->_initialize_params();

    # We need to have 1 igroup per vserver for this testing. 
    # So same vserver should have same igroups
    if ($main::VSERVER_DST eq $main::VSERVER_SRC) {
        $main::IGROUP_DST = $main::IGROUP_SRC;
    }

    ## TODO: Creating a server object to run commands directly.
    ## Consider removing this once lun_show_workaround is removed.
    $self->{'server_obj'} = Server->new(nodes => $main::FILER_DST);

    return ($self);
}

sub dp_snapmirror_life_cycle
{
    my ($self, %input_args) = @_;

    my %args = %input_args;
    $args{'lun_count'}  ||= $main::INITIAL_LUN_COUNT;
    $args{'lun_count'}  = 1;
    $args{'lun_size'}   ||= $main::LUN_SIZE;
    $args{'lun_prefix'} ||= "lun_set1";
    $self->dp_snapmirror_init(%args);

    # Run the snapmirror update with approp. fiji if required.
    %args = %input_args;  # Reset the args.
    $args{'lun_count'} = 1;
    $args{'lun_prefix'} = "lun_set2_001";
    $self->dp_snapmirror_update(%args);

    %args = %input_args;  # Reset the args.
    $args{'lun_count'} = 1;
    $args{'lun_prefix'} = "lun_set3_001";
    $args{'lun_size'}   = "1g";
    $self->dp_snapmirror_abort(%args);

    %args = %input_args;  # Reset the args.
    $self->dp_snapmirror_break(%args);

    # TODO: Resync needs preferably a bigger lun size. 
    # Consider set a min size and increase if user
    # specified < min size.
    %args = %input_args;  # Reset the args.
    $args{'lun_count'} = 1;
    $args{'lun_prefix'} = "lun_set4_001";
    $args{'create_dst_luns'} = 1;
    $self->dp_snapmirror_resync(%args);

    %args = %input_args;  # Reset the args.
    $args{'no_check'} = 1; # This break is just to do flip resync. 
    $self->dp_snapmirror_break(%args);

    $args{'lun_count'} = 1;
    $args{'lun_prefix'} = "lun_set5_001";
    $args{'create_dst_luns'} = 1;
    $self->dp_snapmirror_flip_resync(%args);
}

# Do the test for snapmirror init.
sub dp_snapmirror_init
{
    my ($self, %args) = @_;

    my (@src_luns);
    $args{'lun_prefix'} ||= "lun_set1";
    my $svol      = $args{'src_vol'};
    my $dvol      = $args{'dst_vol'};

    ## ========= Section 1.1.1 ===================
    log_comment("DP_SM_INIT: Creating luns in the source volume\n");

    ## @step Create some luns. 
    my $src_luns = $svol->lun_create(count => $args{'lun_count'},
                                     lun => $args{'lun_prefix'}, 
                                     size => $args{'lun_size'},
                                     ostype => $main::OS_TYPE);

    if (defined($main::CLIENT)) {
        ## Do the client mapping only if user provided the client
        ## parameter.
        log_comment("DP_SM_INIT: Discover the source lun and do IO if required\n");

        lun_map_and_io(vol => $svol, luns => $src_luns,
                       igroup => $main::IGROUP_SRC, read_pct => 0);
    }

    NOTE::logcomment("DP_SM_INIT: Source Lun information:\n", 
                      Dumper($svol->lun_show()));

    # Call the preinit function with preinit args if defined
    if (defined $args{'pre_init_callback'}) {
        $args{'pre_init_callback'}(@{$args{'pre_init_args'}}); 
    }

    ## @step Initialize snapmirror, do initial transfer
    log_comment("DP_SM_INIT: Create and initialize snapmirror relationship\n");
    $self->{sm} = NOTE::GX::SnapMirror->create(src=>$svol, dst=>$dvol);
 
    # If callback is given during transfer, then run init in background
    # and then monitor. Otherwise wait till the init is completed.
    my $bg = defined($args{'init_callback'});
    $self->{sm}->initialize(background => $bg);

    # Monitor the status
    if ($bg) {
        # Wait for it to be start the transfer.
        $self->{sm}->monitor( status => 'Transfer' );

        if (defined $args{'init_callback'}) {
            $args{'init_callback'}(@{$args{'init_args'}}); 
        }

        # Wait for the transfer to complete
        $self->{sm}->monitor( status => 'Transfer' );
    }

    # Call the preinit function with preinit args if defined
    if (defined $args{'post_init_callback'}) {
        $args{'post_init_callback'}(@{$args{'post_init_args'}}); 
    }

    if (defined($main::CLIENT)) {
        log_comment("DP_SM_INIT: Discover new destination LUNs on the client and do read I/O if reqd.\n");

        my $dst_luns = get_undiscovered_dst_luns($self->{sm});
        lun_map_and_io(vol => $dvol, luns => $dst_luns,
                       igroup => $main::IGROUP_DST, read_pct => 100);
    }

    unless ($args{'no_check'}) {
        log_comment("DP_SM_INIT: Collecting all destination lun information\n");
        NOTE::logcomment("DP_SM_INIT: INFO: Destination luns after initial transfer:\n", 
                          Dumper($dvol->lun_show()));

        log_comment("DP_SM_INIT: Verify if transfer/update/snap operation is successful\n");
        $self->verify_sm_operation(%args);
    }

    return (0);
}
 
sub dp_snapmirror_update
{
    my ($self, %args) = @_;
    my @new_luns;

    # TODO: Do all the verifications of the essential args.
    my $svol      = $self->{sm}->src;
    my $dvol      = $self->{sm}->dst;

    ## ========= Section 1.1.2 ===================      
    if ($args{'lun_count'} > 0) {
        log_comment("DP_SM_UPDATE: Create new LUN(s) on the source volume\n");
        my $src_luns = $svol->lun_create(count => $args{'lun_count'},
                                         lun => $args{'lun_prefix'}, 
                                         size => $args{'lun_size'},
                                         ostype => $main::OS_TYPE);

        if (defined($main::CLIENT)) {
            log_comment("DP_SM_UPDATE: Discover the LUN(s) and do IO if required\n");
            lun_map_and_io(vol => $svol, luns => $src_luns,
                           igroup => $main::IGROUP_SRC, read_pct => 0);
        }

        NOTE::logcomment("Source Lun information:\n",
                          Dumper($svol->lun_show()));
    }
 
    # Call the preupdate function with preupdate args if defined
    if (defined $args{'pre_update_callback'}) {
        $args{'pre_update_callback'}(@{$args{'pre_update_args'}}); 
    }

    ## @step update snapmirror
    log_comment("DP_SM_UPDATE: Perform the snapmirror update\n");

    # If callback is given during transfer, then run init in background
    # and then monitor. Otherwise wait till the init is completed.
    my $bg = defined($args{'update_callback'});
    $self->{sm}->update(background => $bg);

    # Monitor the status and call update callback if required.
    if ($bg) {
        # Wait for it to be start the transfer.
        $self->{sm}->monitor(status => 'Transfer');

        if (defined $args{'update_callback'}) {
            $args{'update_callback'}(@{$args{'update_args'}}); 
        }

        # Wait for the transfer to complete
        $self->{sm}->monitor();
    }
 
    if (defined $args{'post_update_callback'}) {
        $args{'post_update_callback'}(@{$args{'post_update_args'}}); 
    }

    if (defined($main::CLIENT)) {
        log_comment("DP_SM_UPDATE: Discover new destination LUNs on the client\n");

        my $dst_luns = get_undiscovered_dst_luns($self->{sm});
        lun_map_and_io(vol => $dvol, luns => $dst_luns,
                       igroup => $main::IGROUP_DST, read_pct => 100);
    }

    unless ($args{'no_check'}) {
       NOTE::logcomment("INFO: Destination luns after sm update:\n", 
                         Dumper($dvol->lun_show()));

        log_comment("DP_SM_UPDATE: Verify if transfer/update is successful\n");
        $self->verify_sm_operation(%args);
    }

    return(0);
}

sub dp_snapmirror_abort
{
    my ($self, %args) = @_;

    my $svol      = $self->{sm}->src;
    my $dvol      = $self->{sm}->dst;

    log_comment("DP_SM_ABORT: Create new LUN(s) on the source volume\n");
    my $src_luns = $svol->lun_create(count  => $args{'lun_count'},
                                     lun    => $args{'lun_prefix'}, 
                                     size   => $args{'lun_size'},
                                     ostype => $main::OS_TYPE);

    if (defined($main::CLIENT)) {
        log_comment("DP_SM_ABORT: Discover new source LUNs on the client\n");
        lun_map_and_io(vol => $svol, luns => $src_luns, run_time => 300,
                       igroup => $main::IGROUP_SRC, read_pct => 0);
    }
    NOTE::logcomment("Source Lun information: \n",
                     Dumper($svol->lun_show()));

    # TODO: We have to do some I/O on the lun to be successful in this
    # testcase. Hence figure out fast on how to do I/O
 
    log_comment("DP_SM_ABORT: Perform the snapmirror update\n");
    $self->{sm}->update(background => 1);

    # Wait for it to be start the transfer.
    $self->{sm}->monitor( status => 'Transfer' );

    # Abort the transfer and verify SM operation.
    log_comment("DP_SM_ABORT: Abort the previous snapmirror update\n");
    $self->{sm}->abort;

    log_comment("DP_SM_ABORT: Collect destination information post abort");
    NOTE::logcomment("Destination Lun information: \n",
                         Dumper($dvol->lun_show()));

    log_comment("DP_SM_ABORT: Verify the expected mismatch between source and destination\n");
    $self->verify_sm_operation(mismatch_expected => 1);

    unless ($args{no_post_abort_update}) {
        # Do another snapmirror update
        log_comment("DP_SM_ABORT: Perform another snapmirror update\n");
        $self->{sm}->update;

        if (defined($main::CLIENT)) {
            log_comment("DP_SM_ABORT: Discover new destination LUNs on the client\n");

            my $dst_luns = get_undiscovered_dst_luns($self->{sm});
            lun_map_and_io(vol => $dvol, luns => $dst_luns,
                           igroup => $main::IGROUP_DST, read_pct => 100);
        }
        NOTE::logcomment("INFO: Destination luns after sm update:\n", 
                         Dumper($dvol->lun_show()));

        log_comment("DP_SM_ABORT: Verify no mismatch between source and destination\n");
        $self->verify_sm_operation(%args);
    }

    return;
}

sub dp_snapmirror_break
{
    my ($self, %args) = @_;
    my $dst_luns;
    my $mismatch = 0;
    my $total_compare = 0;

    my $svol = $self->{sm}->src;
    my $dvol = $self->{sm}->dst;
    
    # Get all the destination luns information.
    unless ($args{'no_check'}) {
        log_comment("DP_SM_BREAK: Collect all the current destination lun information\n");
        $dst_luns = $dvol->lun_show();
        NOTE::logcomment("Destination lun information:\n", Dumper($dst_luns));
    }

    # Call the prebreak function with prebreak args if defined
    if (defined $args{'pre_break_callback'}) {
        $args{'pre_break_callback'}(@{$args{'pre_break_args'}}); 
    }

    # Break the snapmirror relationship
    log_comment("DP_SM_BREAK: Break the snapmirror relationship\n");
    $self->{sm}->break;

    # All of them preceeds this, will be check. So can bypass if asked for.
    return if ($args{'no_check'});
    
    log_comment("DP_SM_BREAK: Verify no mismatch between source and destination\n");
    $args{'no_sm_check'} = 1;
    $self->verify_sm_operation(%args);

    log_comment("DP_SM_BREAK: Collect all post break destination lun information\n");
    my $post_dst_luns = $dvol->lun_show();
    NOTE::logcomment("Destination lun post break:\n", Dumper($post_dst_luns));

    log_comment("DP_SM_BREAK: Verify if serial# is same before and after break\n");
    my %pre_serial = map { $_->{'lun'}, $_->{'serial'} } @$dst_luns;
    my %post_serial = map { $_->{'lun'}, $_->{'serial'} } @$post_dst_luns;

    foreach (keys(%pre_serial)) {
        $total_compare++;
        if ($pre_serial{$_} ne $post_serial{$_}) {
            logresult("FAIL", "BUG: Error Serial number before and after ".
                             "break is different for lun ". $_." ".
                             $pre_serial{$_}.", ".$post_serial{$_}."\n");
            $mismatch++;
        }
    }

    NOTE::logcomment("Total Luns compared = $total_compare, ".
                      "mismatch found = $mismatch\n");

    # TODO: Do we need to unmap the clients as R/O and map them as R/W. 
    # We need to identify that.

    return
}

sub dp_snapmirror_resync
{
    my ($self, %args) = @_;
    my (%pre_serial, %post_serial);
    my $new_dst_luns;
    my $mismatch = 0;
    my $total_compare = 0;
   
    my $svol = $self->{sm}->src;
    my $dvol = $self->{sm}->dst;

    my $fnstr = $args{flip} ? "DP_SM_FLIP_RESYNC": "DP_SM_RESYNC";

    # Create a new LUN preferably bigger size.
    log_comment("$fnstr: Create new LUN(s) on the source volume\n");
    my $src_luns = $svol->lun_create(count  => $args{'lun_count'},
                                     lun    => $args{'lun_prefix'}, 
                                     size   => $args{'lun_size'},
                                     ostype => $main::OS_TYPE);

    if (defined($main::CLIENT)) {
        log_comment("$fnstr: Discover new source LUNs on the client\n");
        lun_map_and_io(vol => $svol, luns => $src_luns,
                       igroup => $main::IGROUP_SRC, read_pct => 0);
    }
    NOTE::logcomment("Luns in source volume: \n", Dumper($svol->lun_show()));

    # Create a new LUN on the destination if specified.
    if ($args{'create_dst_luns'}) {
        log_comment("$fnstr: Create new LUN on the destination volume\n");
        $new_dst_luns = $dvol->lun_create(
                                     count  => $args{'lun_count'},
                                     lun    => $args{'lun_prefix'}."_dst",
                                     size   => $args{'lun_size'},
                                     ostype => $main::OS_TYPE);

        # TODO: Consider doing map and IO here

        # These new lun information should be deleted on a successful resync.
        # After resync, we verify if pre and post resync has same serial
        # number. That time, we will in-correctly flag that there is a 
        # mismatch. Hence creating an entry DONT_COMPARE, so that check
        # below will not compare it.
        %post_serial = map { $_->{'lun'}, "DONT_COMPARE" } @$new_dst_luns;
    }

    # Get all the destination luns information.
    log_comment("$fnstr: Collect all pre resync dst lun information: \n");
    my $dst_luns = $dvol->lun_show();
    NOTE::logcomment("Luns in destination volume before resync: \n",
                     Dumper($dst_luns));

    # Call the preresync function with preresync args if defined
    if (defined $args{'pre_resync_callback'}) {
        $args{'pre_resync_callback'}(@{$args{'pre_resync_args'}}); 
    }

    log_comment("$fnstr: Perform the snapmirror resync\n");
    $self->{sm}->resync();

    log_comment("$fnstr: Collect all post resync destination lun information\n");
    my $post_dst_luns = $dvol->lun_show();
    NOTE::logcomment("Luns in destination volume after resync:\n",
                       Dumper($post_dst_luns));

    log_comment("$fnstr: Verify if serial# is same before and after resync\n");
    %pre_serial = map { $_->{'lun'}, $_->{'serial'} } @$dst_luns;
    map { $post_serial{$_->{'lun'}} = $_->{"serial"} } @$post_dst_luns;

    foreach (keys(%post_serial)) {
        next if ($post_serial{$_} eq "DONT_COMPARE");

        $total_compare++;
        if (exists($pre_serial{$_}) && ($pre_serial{$_} ne $post_serial{$_})) {
            logresult("FAIL", "BUG: Error Serial number before and after ".
                             "resync is different for lun ". $_." ".
                             $pre_serial{$_}.", ".$post_serial{$_}."\n");
            $mismatch++;
        }
    }
    NOTE::logcomment("Total Luns compared = $total_compare, ".
                     "mismatch found = $mismatch\n");

    if (defined($main::CLIENT)) {
        log_comment("$fnstr: Discover new destination LUNs on the client\n");

        my $unmap_dst_luns = get_undiscovered_dst_luns($self->{sm});
        lun_map_and_io(vol => $dvol, luns => $unmap_dst_luns,
                       igroup  => $main::IGROUP_DST, read_pct => 100);
    }

    log_comment("$fnstr: Verify no mismatch between source and destination\n");
    $self->verify_sm_operation(%args);
}
 
sub dp_snapmirror_flip_resync
{
    my ($self, %args) = @_;

    log_comment("DP_SM_FLIP_RESYNC: Testing Flip resync.Delete the existing snapmirror relationship\n".
                "NOTE: From here on src volume = prev dst volume and vice versa\n");

    # Delete the relationship first.
    $self->{sm}->delete;

    # Swap the source and destination and do flip resync
    my $tmp = $self->{sm}->{_src};
    $self->{sm}->{_src} = $self->{sm}->{_dst};
    $self->{sm}->{_dst} = $tmp;

    # Call dp_snapmirror_resync with the same set of params
    $args{flip} = 1;
    $self->dp_snapmirror_resync(%args);
}

sub verify_sm_operation
{
    my ($self, %args) = @_;
 
    my $status = 0;

    $args{mismatch_expected} ||= 0;

    ## @step run lun compare, and fail if no mismatches found prior to 
    ## snapmirror update
    # Compare it from san point of view the 2 volumes. 
    # The compare needs to verify the following for 2 given volumes
    # a) vol type of destination needs to be dp, if sm relationship is on,
    #    otherwise it should be RW)
    # b) list of luns in a volume needs to be same
    # c) compare luns for each lun
    #  c1) lun serial number, lun uuid needs to be different.
    #  c2) lun name should be same. lun path does not matter.
    #  c3) all other lun attributes need to be the same
    #  c4) lun dev id needs to be the same.
    # If these conditions are met, it will return success. Otherwise, it
    # will return failure and will also 
    # array of hashes with offendingproperty that caused this condition.
    if($self->{sm}->lun_compare()) {
        if ($args{mismatch_expected} == 0) {
            logresult("FAIL", "ERROR: BUG: Mismatch between src and ".
                             "destination found\n");
            $status = -1;
        } else {
            NOTE::logcomment("INFO: Lun compare mismatch between src ".
                             "and destination expected and found\n");

        }
    } else {
        if ($args{mismatch_expected} == 0) {
            NOTE::logcomment("INFO: Lun Compare Successful\n");
        } else {
            logresult("FAIL", "ERROR: BUG: Lun compare mismatch between src ".
                             "and destination expected but not found\n");
            $status = -1;
        }
    }

    ## Snapmirror check after transfer
    unless ($args{no_sm_check}) {
        if ($self->{sm}->check()) {
            NOTE::logcomment("ERROR: Snapmirror check has failed:\n");
            $status = -1;
        } else {
            NOTE::logcomment("INFO: Snapmirror check successful\n");
        }
    }
    return ($status);
}

sub log_comment
{
    my ($comment, $level) = @_;

    my $hdr;
    if (!defined($level) || (uc($level) eq "INFO")) {
        $hdr = "INFO: Step $Step, ";
    } else {
        $hdr = "ERROR: Step $Step, ";
    }

    NOTE::logcomment("$hdr $comment\n");
    $Step++;

    return;
}

# This function returns the array of destination side lun object.
sub get_undiscovered_dst_luns
{
    my ($sm) = @_;
    my @new_dst_luns;

    my $dst_luns = $sm->dst->lun_show();
    foreach (@$dst_luns) {
        # Check if the luns are already mapped to the client
        # user had given. If so no need to remap this client.
        if ((!exists($_->{rawName})) || (!length($_->{rawName}))) {
#            ($_->{mapped_to} eq $main::CLIENT)) {
            push(@new_dst_luns, $_);
        }
    }
    return (\@new_dst_luns);
}

sub execute_fiji
{
    my ($self, %args) = @_;

    my $origin_node = $self->{'server_obj'}->set_execution_server(
                                                      Server => $args{node});
    my $cmd = "ngsh -c \"node run -node local fiji filter \'$args{fiji}\'\"";
    my $res = $self->{'server_obj'}->shell(cmd => $cmd);

    if ($res =~ /fiji: filter set/) {
        return (0);
    } else {
        return (-1);
    }
}

sub execute_cmd
{
    my ($self, %args) = @_;
    my ($cmd, $priv);

    # By default run on admin login
    if (!defined($args{mode})) {
        $args{mode} = "admin";
    }

    my $origin_node = $self->{server_obj}->set_execution_server(
                                               -Server => $args{node});
    if ($args{mode} eq "admin") {
        if (length($args{priv})) {
            $priv = "set -confirmations off $args{priv}";
        }
        # Add ngsh -c to run the command in cli (not in bsd)
        $cmd = "ngsh -c \'$priv; $args{cmd}\'";
    } elsif ($args{mode} eq "diag") {
        $cmd = $args{cmd};
    } elsif ($args{mode} eq "7mode") {
        $cmd = "ngsh -c \'node run -node local \"priv set -q $args{priv};".
               " $args{cmd}\"\'";
    }

    my $res = $self->{'server_obj'}->shell(cmd => $cmd);
    return(split(/\n/, $res));
}

sub grep_latest_log
{
    my (%args) = @_;

#    execute_cmd();
}

sub lun_map_and_io
{
    my (%args) = @_;

    # Args for lun_map_to_client
    my $vol      = delete ($args{vol});
    my $igroup   = delete ($args{igroup});
    my $luns     = delete ($args{'luns'});
    my $dev_type = defined($args{dev_type}) ? 
                      delete($args{dev_type}) : $main::DEV_TYPE;
    my $mp_type  = defined($args{mp_type}) ? 
                      delete($args{mp_type}) : $main::MP_TYPE;

    my @lun_names = get_lun_names($luns);
    my @devs = $vol->lun_map_to_client(
                            client => $main::CLIENT,
                            igroup => $igroup,
                            lun    => @lun_names,
                            dev_type => $dev_type,
                            mp_type => $mp_type);

    NOTE::logcomment("Return of lun_map_to_client: \n", Dumper(@devs));
#    my $logstr = "Mapped to device path ";
#    map {$logstr = sprintf("%slun=%s, path=%s\n", $logstr, $_->{lun}, $_->{rawName})} @devs;
#    NOTE::logcomment($logstr);

    NOTE::logcomment("lun show after mapping client:\n",
                     Dumper($vol->lun_show()));

    # Read % is 100 for dest client side and if dest client io is zero
    # avoid running I/O. This is a workaround till things are fixed.
    if ($main::CLIENT_IO && 
          (($args{read_pct} != 100) || $main::DEST_CLIENT_IO)) {
        $vol->SanIo(client => $main::CLIENT, luns => \@lun_names, %args);
    }
}

sub get_lun_names
{
    my ($luns) = @_;
    my @names;

    @names = map {$_->{'lun'}} @$luns;
    return (@names);
}

1;
