# $Id: //depot/prod/test/Rsierranevada/dp/snapmirror/paloma/tools/pcast/lib/NAPC.pm#2 $

package NAPC;

use strict;
use warnings;
use vars qw($VERSION @EXPORT_OK %EXPORT_TAGS);
use subs qw(
    make_fileref
    process_file
    get_physical_stats
    critique_file
    _stats_from_criticstats
    get_file_warnings
    get_line_length_warnings
    get_pod_warnings
    set_data_stats
);
use base 'Exporter';

use Carp qw(carp croak);
use File::Basename qw(basename dirname);
use File::Spec;
use File::chdir;
use Pod::Checker;
use Time::HiRes qw(gettimeofday tv_interval);

use IPC::Run3;
use Perl::Critic;
use Readonly;

Readonly $VERSION => '1.05';

@EXPORT_OK   = qw(
    make_fileref
    process_file
    get_physical_stats
    critique_file
    get_file_warnings
    get_line_length_warnings
    get_pod_warnings
    set_data_stats
);
%EXPORT_TAGS = (all => \@EXPORT_OK);

# When we examine files for warnings, we need the current Perl executable and
# a reliable path to the diagnostics.pm module:

# I hate using hard-coded paths, but this process runs under a different Perl
# than is used for actual development/execution of NetApp code. That means
# there may be libraries under /usr/software that aren't installed here, so we
# have to run the syntax-check using the NetApp perl, not the native one.
my $Perl    = '/usr/software/bin/perl';
my $Diag_pm = '/usr/software/lib/perl5/5.8.8/diagnostics.pm';

# When we look at the length of lines, we flag those that are greater than
# $LINE_LENGTH or $LINE_LENGTH_MAX. The former is the "true" line-length limit
# and the latter is the mandated absolute maximum.
Readonly my $LINE_LENGTH     =>  80;
Readonly my $LINE_LENGTH_MAX => 140;

##############################################################################
#
# Take a file and optionally a base path and create the basic file-ref
# structure that the rest of the routines here (and in NAPC_HTML) expect to
# operate on.
#
##############################################################################
sub make_fileref {
    my ($file, $base) = @_;
    $base ||= $CWD;

    my $fileref = {
        name     => $file,
        basename => basename($file),
        dirname  => dirname($file),
    };

    # dirname returns '.' for leaf files, and we want a null-string instead
    if ($fileref->{dirname} eq q{.}) {
        $fileref->{dirname} = q{};
    }

    # If the filename is already absolute use it for "path", otherwise use
    # $base to make it absolute:
    if (File::Spec->file_name_is_absolute($file)) {
        $fileref->{path} = $file;
    } else {
        $fileref->{path} = File::Spec->catfile($base, $file);
    }

    return $fileref;
}

##############################################################################
#
# Do a full processing of $fileref. Don't get Perl warnings data if
# $args->{skip_warnings} is non-false.
#
##############################################################################
sub process_file {
    my ($fileref, $args) = @_;

    # Make sure $args is defined, and that the "verbose" key is at least 0
    # so that we don't get any warnings:
    $args            ||= {};
    $args->{verbose} ||= 0;

    # Note the time when we start processing the file:
    my $start = [ gettimeofday ];

    # This sets the "revision" key on the hash for $file, gets the inode stats
    # and file length in lines:
    get_physical_stats($fileref, $args);

    # This sets the "critic_*" keys on the hash for $file:
    critique_file($fileref, $args);

    if (! $args->{skip_warnings}) {
        # This augments the critic_* keys with warnings info:
        get_file_warnings($fileref, $args);
    }

    # This creates violation objects on the critic_* keys for lines that
    # are longer than the NetApp coding standards allow:
    get_line_length_warnings($fileref, $args);

    # This runs Pod::Checker on the file and creates level 2 and 3 violations
    # for warnings/errors (respectively) found by it:
    get_pod_warnings($fileref, $args);

    # Once the critiques and warnings are known, calculate data stats:
    set_data_stats($fileref, $args);

    # Now set the "process_time" key on $fileref
    $fileref->{process_time} = tv_interval($start);

    # No need to return a status value-- any of the previous that fails will
    # raise an exception.
    return;
}

##############################################################################
#
# Scan the file for an Id keyword that Perforce should have filled in with
# the path and revision number. Get the number. Get the inode stats, and
# count the number of lines. And anything else to do with the physical file
# that I think of...
#
##############################################################################
sub get_physical_stats {
    my ($fileref, $args) = @_;

    # Get the inode stats for the file, for newness-checking on HTML files
    $fileref->{file_stats} = [ lstat($fileref->{path}) ];

    # Read the whole file, so we can count the number of lines, etc.
    my @lines;
    if (open my $fh, '<', $fileref->{path}) {
        @lines = <$fh>;
        if (! close $fh) {
            # If we fail here, there may be underlying file-sys problems
            croak "Close failed on $fileref->{path}: $!";
        }
    } else {
        croak "Open failed on $fileref->{path}: $!";
    }

    # Get the revision number by scanning against the lines.
    # This is the regexp we match against, looking for the shorter file name
    # followed by "#" and a number:
    #
    # Edit: Perforce was expanding this as a keyword, so changed to this hack
    #       that separates the leading '$' from the rest of the string to
    #       prevent that.
    my $regexp = q{\$} . "Id:\\s+.*?/$fileref->{name}#(\\d+)\\s+\\\$";
    $regexp = qr/$regexp/;
    # Default to "unknown" as the value
    my $revision = 'unknown';
    # We shouldn't have to read too far into the file to find the line,
    # unless the file just doesn't have it...
    foreach my $line (@lines) {
        if ($line =~ /$regexp/) {
            $revision = $1;
            last;
        }
    }
    $fileref->{revision} = $revision;

    # Now check for lines that are longer than 80 or 140 characters. These are
    # the lengths laid down in the NetApp Perl Coding Standards. Later, after
    # the structure for violations has been set up on $fileref this data will
    # be converted to look like regular Perl::Critic complaints. We do it here
    # because we have @lines in memory right now, and I don't want to either
    # re-read it later or store it on $fileref. For sake of cleanliness, the
    # values 80 and 140 are expressed as constants that were declared earlier.
    chomp(@lines);
    # If this is a Windows-encoded file, then there is still a trailing "\r"
    # on each line:
    if ($lines[0] =~ /\r$/) {
        for (@lines) {
            s/\r$//;
        }
    }
    $fileref->{lines_past_length}     = [];
    $fileref->{lines_past_length_max} = [];
    # Lines are indexed here from 0, but indexed from 1 elsewhere. When we add
    # them to the lists, we'll add $idx+1.
    for my $idx (0 .. $#lines) {
        # Lines that contain just a Perforce keyword in a comment are exempt:
        next if $lines[$idx] =~ /\s*#\s*\$(?:Id|Header|File|Revision)/;
        if (length($lines[$idx]) > $LINE_LENGTH_MAX) {
            push(@{$fileref->{lines_past_length_max}}, $idx + 1);
        } elsif (length($lines[$idx]) > $LINE_LENGTH) {
            push(@{$fileref->{lines_past_length}}, $idx + 1);
        }
    }

    return;
}

##############################################################################
#
# Run the file represented by $fileref through the Perl::Critic engine and
# save the critiques (if any) on $fileref.
#
##############################################################################
sub critique_file {
    my ($fileref, $args) = @_;

    my ($perlcriticfile, $critic_severity);

    if (! ($perlcriticfile = $args->{critic_file})) {
        # The default Perl::Critic profile is "perlcritic" in the same dir as
        # this module.
        $perlcriticfile = File::Spec->catfile(dirname(__FILE__), 'perlcritic');
    }
    # The default severity is 1 ("brutal")
    $critic_severity = $args->{critic_severity} || 1;

    # Full path to the file, to pass to Perl::Critic
    my $sourcefile = $fileref->{path};

    $args->{verbose} >= 2 &&
        print "  running Perl::Critic on $fileref->{name}\n";

    # Let's see what the critic has to say about this. Create an instance of
    # the Perl::Critic engine for critiquing the code and run it:
    my $perlcritic = Perl::Critic->new(
        -profile  => $perlcriticfile,
        -severity => $critic_severity
    );
    my @violations = $perlcritic->critique($sourcefile);

    # Get the statistics that Perl::Critic computed during its process:
    my $critic_stats = _stats_from_criticstats($perlcritic->statistics);

    $args->{verbose} >= 2 &&
        print '  critic found ' . scalar(@violations) .  " violation(s)\n";

    # If the user passed "diff_lines" in the args list, it points to a hashref
    # of line numbers that refer to the lines we actually care about. Trim
    # the list of violations to just these lines.
    if (my $lines = $args->{diff_lines}) {
        @violations = grep { $lines->{all} || $lines->{$_->line_number} }
            @violations;
    }

    # The list of violations is sorted by position in the code, which is
    # generally what we'll want to work with anyway. But just in case, let's
    # also create "buckets" where they're sorted by severity. Using the
    # buckets preserves their in-source-code order, as well.
    #
    # Also create buckets for each line. In most cases, there is probably just
    # one violation on a line. But there can be several, and our tool is
    # line-oriented in nature, after all. Also create the matrix of severities
    # by line, which is used in the template to indicate presence of problems
    # on a per-line basis.
    my @all_violations = ();
    my @buckets = ([], [], [], [], [], []); # 0 isn't used as a severity
    my @lines = ();
    my @sev_by_line = ();
    my @sev_obj_by_line = ();
    # This saves the diagnostics that are relevant to this file:
    my %diagnostics = ();

    for my $violation (@violations) {
        my $object = {
            severity      => $violation->severity,
            line_number   => $violation->line_number,
            column_number => $violation->column_number,
            policy        => $violation->policy,
            description   => $violation->description,
            explanation   => $violation->explanation,
        };
        # We use these two a lot, so read them just the once
        my $line_num = $object->{line_number};
        my $severity = $object->{severity};
        # If we haven't saved the diagnostics for this policy yet, do so:
        if (! $diagnostics{$object->{policy}}) {
            (my $diag = $violation->diagnostics) =~ s/^    //mg;
            $diagnostics{$object->{policy}} = $diag;
        }

        if (! defined $sev_by_line[$line_num]) {
            $sev_by_line[$line_num] = [ 0, 0, 0, 0, 0, 0 ];
            $sev_obj_by_line[$line_num] = [ [], [], [], [], [], [] ];
        }

        # Add to the bucket for this severity
        push @{$buckets[$severity]}, $object;
        # Add to the bucket for this line
        push @{$lines[$line_num]}, $object;
        # Add to the sev-by-line buckets
        push @{$sev_obj_by_line[$line_num]->[$severity]}, $object;
        # Increment the sev-by-line count
        $sev_by_line[$line_num]->[$severity]++;
        # Lastly, save this object for the "all" list, so that we don't save
        # the huge objects the Perl::Critic throws around
        push @all_violations, $object;
    }

    # Save the data
    $fileref->{critic_all}             = \@all_violations;
    $fileref->{critic_by_sev}          = \@buckets;
    $fileref->{critic_by_line}         = \@lines;
    $fileref->{critic_sev_by_line}     = \@sev_by_line;
    $fileref->{critic_sev_obj_by_line} = \@sev_obj_by_line;
    # By-bucket totals
    for my $sev (1 .. 5) {
        $fileref->{critic_total_by_sev}->[$sev] = scalar $buckets[$sev];
    }
    # Save the diagnostics for this file
    $fileref->{diagnostics} = \%diagnostics;
    # Save the stats that Perl::Critic came up with
    $fileref->{critic_stats} = $critic_stats;

    return;
}

##############################################################################
#
# Take a Perl::Critic::Statistics object and extract what we want/care about
# into a hash.
#
##############################################################################
sub _stats_from_criticstats {
    my $statsobj = shift;

    my $stats = {
        subs                       => $statsobj->subs(),
        statements                 => $statsobj->statements(),
        lines                      => $statsobj->lines(),
        lines_of_blank             => $statsobj->lines_of_blank(),
        lines_of_comment           => $statsobj->lines_of_comment(),
        lines_of_data              => $statsobj->lines_of_data(),
        lines_of_perl              => $statsobj->lines_of_perl(),
        lines_of_pod               => $statsobj->lines_of_pod(),
        violations_by_severity     => $statsobj->violations_by_severity(),
        violations_by_policy       => $statsobj->violations_by_policy(),
        total_violations           => $statsobj->total_violations(),
        statements_other_than_subs => $statsobj->statements_other_than_subs(),
        average_sub_mccabe         => $statsobj->average_sub_mccabe() || 0,
    };

    # Make a more human-friendly form of this one:
    $stats->{mccabe} = sprintf '%.3f', $stats->{average_sub_mccabe};

    # Calculate some ratios around POD and comments. Watch out for potential
    # div-by-0 problems...
    if ($stats->{lines_of_perl}) {
        $stats->{pod_ratio} = sprintf(
            '%.3f',
            $stats->{lines_of_pod} / $stats->{lines_of_perl}
        );
        $stats->{comment_ratio} = sprintf(
            '%.3f',
            $stats->{lines_of_comment} / $stats->{lines_of_perl}
        );
    }

    return $stats;
}

##############################################################################
#
# Take the file represented in $fileref and run it under "perl -wc". Capture
# the warnings and integrate them into the buckets of data already gathered
# through Perl::Critic.
#
##############################################################################
sub get_file_warnings {
    my ($fileref, $args) = @_;

    # Full path to the file, to pass to "perl -wc"
    my $sourcefile = $fileref->{path};
    # List of all warnings we find
    my @warnings = ();

    $args->{verbose} >= 2 &&
        print "  running 'perl -wc' on $fileref->{name}\n";

    # Get the lines of output. We can't count on PERL5LIB being set up for any
    # extra library paths this would need. It turns out that when this runs
    # under Apache2, PERL5LIB isn't propagated through in the environment. So
    # turn the contents into explicit -I options.
    my @lines;
    my @libs = ();
    if ($args->{libraries}) {
        push(@libs, (map { "-I$_" } @{$args->{libraries}}));
    }
    if ($ENV{PERL5LIB}) {
        push(@libs, (map { "-I$_" } split(/:/ => $ENV{PERL5LIB})));
    }
    # run3() comes from IPC::Run3.
    run3 [ $Perl, @libs, '-wc', $sourcefile ], \undef, \undef, \@lines;
    chomp @lines;

    # As we iterate over the lines we just read, we'll use this regexp to both
    # verify that the line belongs to this file, and also to extract the line
    # number it occurs on. We have to filter like this because there is a TON
    # of bleed-over noise from other Perl modules that I don't have control
    # over...
    my $regexp = qr{
        ^(.*?)\s+
        at\s+$sourcefile\s+
        line\s+(\d+)[.]
    }x;
    my ($description, $line); # Captured from successful regexp matches
    my $policy = 0; # Increments for each new warning (that we use)
    my $wcount = 0; # Total number of warnings for the file
    my $diagnostic; # Used with run3() to capture the diag. text of the warning
    my %seen_this;  # We're seeing duplicate warnings on lines in some cases
    foreach my $warning_line (@lines) {
        if ($warning_line =~ /$regexp/) {
            $description = $1;
            $line        = $2;
        } else {
            next;
        }

        # There are certain warnings we wish to skip, because they aren't
        # relevant to static analysis. Check for those here:
        next if ($description =~ /Can't locate.*in \@INC/ ||
                 $description =~ /BEGIN failed--compilation aborted/);

        # Skip if we've seen this warning for this line:
        next if ($seen_this{"$line:$description"}++);

        # Count the warnings that are actually for this file, and aren't
        # being rejected out-of-hand by the previous tests. This number has
        # to count all of them, even if "diff_lines" causes us to skip most
        # of them.
        $wcount++;

        # If the user passed "diff_lines" in args, it's a hashref of line
        # numbers that we are focusing on. Only retain those warnings that
        # occur on one of those lines.
        if (my $lines = $args->{diff_lines}) {
            next if ((! $lines->{all}) && (! $lines->{$line}));
        }

        my $object = {
            severity      => 0,
            line_number   => $line,
            column_number => 1,
            policy        => "warning$policy",
            description   => $description,
            explanation   => q{},
        };
        push @warnings, $object;

        # Get the diagnostic message for the warning. Note that because this
        # can run under Apache, we have to force output to STDERR rather than
        # reading STDOUT:
        run3 "$Perl $Diag_pm 1>&2", \$warning_line, \undef, \$diagnostic;
        chomp $diagnostic;
        $diagnostic =~ s/^.*[(]#1[)]\n//s;
        $diagnostic =~ s/^    //mg;
        $fileref->{diagnostics}->{"warning$policy"} = $diagnostic;
        $policy++;

        # $line still holds the line number. Use it to increment the sev-0
        # per-line bucket, and add to the other line-based buckets. We may
        # have a warning on a line the Perl::Critic didn't already set up, so
        # check first to see if we need to create "blanks" for the line:
        if (! defined $fileref->{critic_sev_by_line}->[$line]) {
            $fileref->{critic_sev_by_line}->[$line] = [ 0, 0, 0, 0, 0, 0 ];
            $fileref->{critic_sev_obj_by_line}->[$line] =
                [ [], [], [], [], [], [] ];
        }
        $fileref->{critic_sev_by_line}->[$line]->[0]++;
        push @{$fileref->{critic_sev_obj_by_line}->[$line]->[0]}, $object;
        # This one will auto-vivify if needed:
        push @{$fileref->{critic_by_line}->[$line]}, $object;
        # Add to the "all sev-0" bucket
        push @{$fileref->{critic_by_sev}->[0]}, $object;
    }

    $args->{verbose} >= 2 &&
        print '  perl found ' . scalar(@warnings) .  " warnings(s)\n";

    # Add the TOTAL (not just within changed lines) number of warnings to
    # the stats for total violations.
    if ($wcount) {
        $fileref->{critic_stats}->{total_violations} += $wcount;
        $fileref->{critic_stats}->{violations_by_policy}->{warnings} = $wcount;
    }

    if (@warnings) {
        # Merge @warnings into $fileref->{critic_all}
        my @all_issues = (@{$fileref->{critic_all}}, @warnings);
        @all_issues = sort {
            $a->{line_number} <=> $b->{line_number} ||
            $a->{severity}    <=> $b->{severity}
        } @all_issues;
        $fileref->{critic_all} = \@all_issues;
    }

    return;
}

###############################################################################
#
#   Sub Name:       get_line_length_warnings
#
#   Description:    Using line-length information saved in an earlier step,
#                   create pseudo-violations for lines that are longer than
#                   $LINE_LENGTH or $LINE_LENGTH_MAX.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $fileref  in      hashref   Fileref structure that holds
#                                                 pretty much all the data
#                   $args     in      hashref   Arguments passed to all stages
#                                                 of processing
#
#   Returns:        void
#
###############################################################################
sub get_line_length_warnings
{
    my ($fileref, $args) = @_;
    my (
        @line_warnings, $lines, $warn_count, $severity, $column, $description,
        $diagnostic
    );

    # This is the count of ALL the warnings, before we (possibly) prune them
    # down based on the user asking to only check lines with diffs. In fact,
    # we can then use this to skip this whole subroutine if the count is 0.
    $warn_count =
        @{$fileref->{lines_past_length}} + @{$fileref->{lines_past_length_max}};
    if (! $warn_count) {
        # If there are no line-length violations, we're done here.
        return;
    } else {
        # Add the TOTAL (not just within changed lines) number of warnings to
        # the stat for total violations.
        $fileref->{critic_stats}->{total_violations} += $warn_count;
        $fileref->{critic_stats}->{violations_by_policy}->{'coding standards'}
            = $warn_count;
    }

    # If the user passed "diff_lines" in the args list, it means that we're
    # restricting our checks to just those lines specified. If it isn't set,
    # letting $lines stay "undef" works, too.
    $lines = $args->{diff_lines};

    for my $key (qw(lines_past_length lines_past_length_max)) {
        if ($key eq 'lines_past_length') {
            $severity    = 4;
            $column      = $LINE_LENGTH + 1;
            $description = "Line longer than $LINE_LENGTH characters";
            $diagnostic  = << "EOD";
According to NetApp Perl Coding Standards, lines should not be
longer than $LINE_LENGTH characters. If possible, find a logical place
near column $LINE_LENGTH and break the line there.
EOD
        } else {
            $severity    = 5;
            $column      = $LINE_LENGTH_MAX + 1;
            $description = "Line longer than $LINE_LENGTH_MAX characters";
            $diagnostic  = << "EOD";
According to NetApp Perl Coding Standards, lines absolutely must not
exceed $LINE_LENGTH_MAX characters in length, and additionally should not
exceed $LINE_LENGTH characters. Find a logical place, preferably near column
$LINE_LENGTH, and break the line there.
EOD
        }

        for my $line (@{$fileref->{$key}}) {
            # Skip this line if $lines has data and $line isn't in it:
            next if ($lines && (! $lines->{all}) && (! $lines->{$line}));

            # Create a violation object for this line
            my $object = {
                severity      => $severity,
                line_number   => $line,
                column_number => $column,
                policy        => $key,
                description   => $description,
                explanation   => 'See NetApp Perl Coding Standards',
            };
            push(@line_warnings, $object);

            # Set the full diagnostic for this "policy" (for the lightbox)
            $fileref->{diagnostics}->{$key} = $diagnostic;

            # Create "blanks" in the critic_sev_by_line and
            # critic_sev_obj_by_line structures if we haven't seen a complaint
            # on this line yet:
            if (! defined $fileref->{critic_sev_by_line}->[$line]) {
                $fileref->{critic_sev_by_line}->[$line] = [ 0, 0, 0, 0, 0, 0 ];
                $fileref->{critic_sev_obj_by_line}->[$line] =
                    [ [], [], [], [], [], [] ];
            }

            # Now increment the per-line bucket for $severity, and add $object
            # to the object-per-line bucket:
            $fileref->{critic_sev_by_line}->[$line]->[$severity]++;
            push(
                @{$fileref->{critic_sev_obj_by_line}->[$line]->[$severity]},
                $object
            );
            # This one will auto-vivify if $line is new, if needed
            push(@{$fileref->{critic_by_line}->[$line]}, $object);
            # And add to the "all" bucket for $severity
            push(@{$fileref->{critic_by_sev}->[$severity]}, $object);
        }
    }

    $args->{verbose} >= 2 && print '  found ' . scalar(@line_warnings) .
        " line-length violation(s)\n";

    # If there were any warnings (after filtering for diff'd lines) merge them
    # into $fileref->{critic_all}.
    if (@line_warnings) {
        my @all_issues = (@{$fileref->{critic_all}}, @line_warnings);
        @all_issues = sort {
            $a->{line_number} <=> $b->{line_number} ||
            $a->{severity}    <=> $b->{severity}
        } @all_issues;
        $fileref->{critic_all} = \@all_issues;
    }

    return;
}

###############################################################################
#
#   Sub Name:       get_pod_warnings
#
#   Description:    Get the warnings/errors about POD from Pod::Checker
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $fileref  in      hashref   Fileref structure that holds
#                                                 pretty much everything
#                   $args     in      hashref   Arguments passed to all stages
#                                                 of processing
#
#   Returns:        void
#
###############################################################################
sub get_pod_warnings
{
    my ($fileref, $args) = @_;
    my (
        $checker, $buffer, @lines, @pod_warnings, $severity, $line, $text, $fh,
        $regexp, $all_count, $lines, $sev
    );

    # Get a Pod::Checker object to use to get the POD diagnostics from. I
    # prefer the OO approach over the function that Pod::Checker exports.
    # The arguments to new() set the warnings level to the maximum.
    $checker = Pod::Checker->new(-warnings => 2);

    # Open an in-memory filehandle on $buffer, so $checker->parse_from_file
    # can write to it instead of an actual file. Then call the method with
    # the file's full path and the file handle.
    $buffer = q{};
    if (open $fh, '>', \$buffer) {
        $checker->parse_from_file($fileref->{path}, $fh);
        close $fh or croak "Error closing in-memory fh: $!";
    } else {
        croak "Error opening in-memory filehandle: $!";
    }

    # Convert $buffer to lines we can parse:
    @lines = split /\n/ => $buffer;
    # If there are no lines, there is no need to do any of the rest:
    return if (! @lines);
    $all_count = scalar @lines;
    # Add the TOTAL (not just within changed lines) number of warnings to
    # the stat for total violations.
    $fileref->{critic_stats}->{total_violations} += $all_count;
    $fileref->{critic_stats}->{violations_by_policy}->{'documentation'} =
        $all_count;

    # We'll use this to tell if we're checking against changes lines only
    $lines = $args->{diff_lines};

    # This should match each line, leaving the relevant information in the
    # capture buffers:
    $regexp = qr{
        ^[*]{3}\s+              # Leading "*** "
        (WARNING|ERROR):\s+     # $1 - "WARNING" or "ERROR"
        (.*?)\s+                # $2 - Text of the violation
        at\sline\s(\d+|-)       # $3 - Line number or "-"
    }x;

    # Iterate over the lines from Pod::Checker and turn each one into a
    # violation object, like we do with warnings and line-length problems.
    for my $pod_line (@lines) {
        if ($pod_line =~ /$regexp/) {
            $severity = $1;
            $text     = $2;
            $line     = $3;
            # If $line is '-' just move it to line 1 for arguments' sake
            if ($line eq q{-}) {
                $line = 1;
            }
        } else {
            # Should probably warn here, since none of the lines should fail
            # to match the regexp...
            next;
        }

        # If we're checking against diffs and this isn't on a line we're
        # looking at, skip it.
        next if ($lines && (! $lines->{all}) && (! $lines->{$line}));

        # The documentation style of NACL is causing a lot of these warnings.
        # Until/unless I can change their mind on docs-style, disable it:
        next if ($text =~ /multiple occurrence of link target/);

        # Create the pseudo-object for the violation. Note that we are making
        # WARNING-level checker messages severity 2, and ERROR-level messages
        # severity 3.
        $sev = ($severity eq 'WARNING') ? 2 : 3;
        my $object = {
            severity      => $sev,
            line_number   => $line,
            column_number => 1,
            policy        => "pod$severity",
            description   => $text,
            explanation   => 'See the Pod::Checker manual page',
        };
        push @pod_warnings, $object;

        # $line still holds the line number. Use it to increment the right
        # per-line bucket, and add to the other line-based buckets. We may
        # have a warning on a line the Perl::Critic didn't already set up, so
        # check first to see if we need to create "blanks" for the line:
        if (! defined $fileref->{critic_sev_by_line}->[$line]) {
            $fileref->{critic_sev_by_line}->[$line] = [ 0, 0, 0, 0, 0, 0 ];
            $fileref->{critic_sev_obj_by_line}->[$line] =
                [ [], [], [], [], [], [] ];
        }
        $fileref->{critic_sev_by_line}->[$line]->[$sev]++;
        push @{$fileref->{critic_sev_obj_by_line}->[$line]->[$sev]}, $object;
        # This one will auto-vivify if needed:
        push @{$fileref->{critic_by_line}->[$line]}, $object;
        # Add to the "all" bucket for severity $sev:
        push @{$fileref->{critic_by_sev}->[$sev]}, $object;
    }

    # Create the explanation text for the two types of POD-related violations:
    $fileref->{diagnostics}->{podWARNING} = << 'EOD';
This is a warning-level message from Pod::Checker. This may not cause a
problem with the POD formatting, but it is bad style and should probably
be addressed. See the manual page for "Pod::Checker" for more on this
message.
EOD
    $fileref->{diagnostics}->{podERROR} = << 'EOD';
This is a POD error, as found by Pod::Checker. This affects some aspect of
the POD formatting, and should definitely be addressed. See the manual page
for "Pod::Checker" for a more detailed description of this error message.
EOD

    $args->{verbose} >= 2 && print '  found ' . scalar(@pod_warnings) .
        " POD syntax violation(s)\n";

    # If there were any warnings (after filtering for diff'd lines) merge them
    # into $fileref->{critic_all}.
    if (@pod_warnings) {
        my @all_issues = (@{$fileref->{critic_all}}, @pod_warnings);
        @all_issues = sort {
            $a->{line_number} <=> $b->{line_number} ||
            $a->{severity}    <=> $b->{severity}
        } @all_issues;
        $fileref->{critic_all} = \@all_issues;
    }

    return;
}

##############################################################################
#
# Take the data on critiques/warnings and associate some stats for the file
# that aren't already covered by Perl::Critic::Statistics. Also copy a few
# bits from the larger stats for simplicity (keeping everything in one place).
#
##############################################################################
sub set_data_stats {
    my ($fileref, $args) = @_;

    # This is just to save some typing:
    my $cstats = $fileref->{critic_stats};

    # We'll store everything here:
    $fileref->{stats} = {};

    # Copy these directly so that it is easier for index-page per-file-stats
    # access:
    $fileref->{stats}->{average_mccabe} = $cstats->{mccabe};
    $fileref->{stats}->{total_subs} = $cstats->{subs};
    $fileref->{stats}->{total_lines} = $cstats->{lines};

    # These have to be calculated because we're including the warnings in the
    # overall count, which Perl::Critic didn't.
    $fileref->{stats}->{total_violations} = scalar(@{$fileref->{critic_all}});
    $fileref->{stats}->{violations_per_statement} =
        $cstats->{total_violations} / $cstats->{statements_other_than_subs};
    $fileref->{stats}->{violations_per_line} =
        $cstats->{total_violations} / $cstats->{lines};
    # Make human-readable forms (3 decimal places) for use in templates
    $fileref->{critic_stats}->{violations_per_statement} =
        sprintf '%.3f', $fileref->{stats}->{violations_per_statement};
    $fileref->{critic_stats}->{violations_per_line} =
        sprintf '%.3f', $fileref->{stats}->{violations_per_line};

    return;
}

1;
