# $Id: //depot/prod/test/Rsierranevada/dp/snapmirror/paloma/tools/pcast/lib/NAPC_HTML.pm#1 $

package NAPC_HTML;

use strict;
use warnings;
use vars qw($VERSION @EXPORT_OK %EXPORT_TAGS);
use subs qw(
    napc_html_init
    generate_file_html
    generate_index_html
);
use base 'Exporter';

use Carp qw(carp croak);
use Cwd 'cwd';

use JSON;
use PPI;
use PPI::HTML;
use Readonly;
use Template;

Readonly $VERSION => '1.02';

@EXPORT_OK   = qw(napc_html_init generate_file_html generate_index_html);
%EXPORT_TAGS = (all => \@EXPORT_OK);

# Create a JSON handle for encoding data to JSON when we pass it to
# templates. Any of the Perl data that will eventually be manipulated by
# JavaScript has to be encoded this way.
my $Json = JSON->new();

# The handle on the Template::Toolkit engine. Starts out undef, must be
# set up in init_html().
my $Template;

# This is a semi-constant (only computed once, by napc_html_init) string that
# lists the elements of the toolchain for reporting on HTML page footers.
my $Toolchain;

##############################################################################
#
# Initialize the Template::Toolkit object and any other set-up operations that
# are needed by this module.
#
##############################################################################
sub napc_html_init {
    my $args = shift;

    my $root = $args->{template_root} || cwd;

    $Template = Template->new(
        {
            INCLUDE_PATH => $root,
            EVAL_PERL    => 1,
        }
    );

    # Set the value of $Toolchain. If there is no "toolchain" key in the
    # args hashref, try to compute it from $0.
    if ($args->{toolchain}) {
        $Toolchain = $args->{toolchain};
    } else {
        require File::Basename;

        $Toolchain = File::Basename::basename($0);
        if (defined $main::VERSION) {
            $Toolchain .= "/$main::VERSION";
        }
    }
    # Now add elements for NAPC (if loaded) and this module
    if (defined $NAPC::VERSION) {
        $Toolchain .= ", NAPC/$NAPC::VERSION";
    }
    $Toolchain .= ", NAPC_HTML/$VERSION";

    return;
}

##############################################################################
#
# Taking the file information in $fileref, determine if we need to create a
# new HTML syntax-marked file for this entry.
#
##############################################################################
sub generate_file_html {
    my ($fileref, $htmltarget, $args) = @_;

    # Full path to the file, to pass to PPI::HTML
    my $sourcefile = $fileref->{path};

    my $document = PPI::Document->new($sourcefile);
    my $ppi_html = PPI::HTML->new(line_numbers => 1);
    my @lines = split /<br>\n/, $ppi_html->html($document);
    # Now strip the line-numbers PPI provided, and force-close the <span> on
    # each line. This seems wasteful, but it was the only way to get PPI::HTML
    # to markup the code on a line-by-line basis, as opposed to leaving spans
    # open across whole blocks of lines.
    for my $lineno (0 .. $#lines) {
        if ($lines[$lineno] =~ s{^</span>}{}) {
            $lines[$lineno - 1] .= '</span>';
        }
        $lines[$lineno] =~ s{<span\s+class="line_number">\s*\d+:\s+</span>}{}x;
    }

    # Convert the by-line-by-sev data buckets into a simpler structure that
    # can be easily put into JSON for use by the page we generate:
    my @line_notes = ();
    for my $line (@{$fileref->{critic_sev_obj_by_line}}) {
        if (! defined $line) {
            push @line_notes, undef;
            next;
        }

        my $notes = [];
        for my $sev (0 .. 5) {
            $notes->[$sev] =
                [ map { $_->{description} . q{.} } @{$line->[$sev]} ];
        }

        push @line_notes, $notes;
    }

    # The listref of diagnostics is fine, but the contents of the diagnostics
    # might include HTML special characters. So clean them up into a temp
    # array.
    my %sub_table = (
        q{<} => '&lt;',
        q{>} => '&gt;',
        q{"} => '&quot;',
        q{&} => '&amp;',
    );
    my %diagnostics;
    for my $key (keys %{$fileref->{diagnostics}}) {
        my $diag = $fileref->{diagnostics}->{$key};
        $diag =~ s/([<>"&])/$sub_table{$1}/eg;
        $diagnostics{$key} = $diag;
    }

    # Get Template::Toolkit to do the rendering thing it does so well.
    my $result = $Template->process(
        'source_file.tt',
        {
            lines            => \@lines,
            fileref          => $fileref,
            title            => "File: $fileref->{name}",
            diagnostics_json => $Json->encode(\%diagnostics),
            line_notes_json  => $Json->encode(\@line_notes),
            process_time     => $fileref->{process_time},
            toolchain        => $Toolchain,
            $args->{template_args} ? %{$args->{template_args}} : (),
        },
        $htmltarget
    );
    if (! $result) {
        croak q(Error processing template 'source_file.tt': ) .
            $Template->error();
    }

    return;
}

##############################################################################
#
# Generate an index.html file (or whatever we were told to call it by the
# --indexfile parameter) that links to all of the per-file HTML we've
# generated so far.
#
##############################################################################
sub generate_index_html {
    my ($file_data, $indextarget, $args) = @_;

    # Timestamp for the page title
    my $now          = $args->{timestamp} || time;
    my $targets_used = $args->{targets}   || [];

    # Calculate some stats to use on the index page. Could do this in the
    # template itself, but it's cleaner and easier to read here:
    my %stats = (
        total  => 0,
        by_sev => [ 0, 0, 0, 0, 0, 0 ]
    );
    my %file_stats = ();
    for my $file (keys %{$file_data})
    {
        $stats{total} += scalar(@{$file_data->{$file}->{critic_all}});
        for my $sev (0 .. 5)
        {
            $stats{by_sev}->[$sev] +=
                scalar(@{$file_data->{$file}->{critic_by_sev}->[$sev]});
        }

        $file_stats{$file} = {
            revision       => $file_data->{$file}->{revision},
            lines          => $file_data->{$file}->{stats}->{total_lines},
            subs           => $file_data->{$file}->{stats}->{total_subs},
            average_mccabe => $file_data->{$file}->{stats}->{average_mccabe},
            total          => $file_data->{$file}->{stats}->{total_violations},
            per_line       =>
            $file_data->{$file}->{critic_stats}->{violations_per_line},
            per_statement  =>
            $file_data->{$file}->{critic_stats}->{violations_per_statement},
            pod_ratio      => $file_data->{$file}->{critic_stats}->{pod_ratio},
            comment_ratio  =>
            $file_data->{$file}->{critic_stats}->{comment_ratio},
        };
    }

    # For rendering the index, all we need is the sorted list of files, the
    # file data, stats data and list of targets. The template handles the rest.
    my $result = $Template->process(
        'index.tt',
        {
            files      => [ sort keys %{$file_data} ],
            filedata   => $file_data,
            data       => \%stats,
            targets    => $targets_used,
            title      => 'Pcast Run, ' . scalar localtime $now,
            file_stats => $Json->encode(\%file_stats),
            toolchain  => $Toolchain,
        },
        $indextarget
    );
    if (! $result) {
        croak q(Error processing template 'index.tt': ) .
            $Template->error();
    }

    return;
}

1;
