## @summary
##   Perl interface to transfer engine ZAPIs
##
## @type
##   library
##
## @author
##   Keith A. Bare II <bare@netapp.com>
##
## @see-also
##   dp/snapmirror/paloma/unit_test/enginecli.thpl
##
##   dp/snapmirror/paloma/unit_test/zsmcli.pm
##
## @pod here

=pod

=head1 NAME

EngineZapi - Perl interface to transfer engine ZAPIs

=head1 SYNOPSIS

 use Server;
 use dp::snapmirror::paloma::unit_test::EngineZapi;

 my $server = new Server;
 my $engine = new EngineZapi($server);

 my $src = EngineZapi::get_volume_info($server, "src_vol");
 my $dst = EngineZapi::get_volume_info($server, "dst_vol");

 my $result = $engine->destination_prepare($src, $dst);
 my $prepare_id = $result->{"prepare-id"};

 $result = $engine->destination_get_status($dst);
 my ($key, $value);
 while (($key, $value) = each(%$result)) {
     print $key, " => ", $value, "\n";
 }

 my $xfer_snap = EngineZapi::get_snap_uuid_by_name(
         $server, "src_vol", "snap3");
 my $ref_snap = EngineZapi::get_snap_uuid_by_name(
         $server, "src_vol", "snap1");
 $result = $engine->source_start($src, $dst, $prepare_id, $xfer_snap,
                                 "reference-snapshot-id" => $ref_snap);
 my $xfer_id = $result->{"transfer-id"};

 $result = $engine->source_get_status($src, $xfer_id);
 while (($key, $value) = each(%$result)) {
     print $key, " => ", $value, "\n";
 }

 # Assume the transfer has now finished

 $result = $engine->source_conclude($src, $xfer_id);

 $result = $engine->destination_conclude($dst, $prepare_id);

=head1 DESCRIPTION

This library provides an easy-to-use interface for interacting with the
transfer engine from THPL scripts via dblade ZAPIs.  It allows THPL authors to
interact with the engine with minimal concern about how the ZAPI inputs need
to be formatted and how the resulting outputs need to be parsed.

=cut


use strict;
use warnings;

use NATE::Log;
use NATE::Result;
use Server;
use UUID;

use dp::snapmirror::paloma::unit_test::zsmcli;

package EngineZapi;


my $Log = NATE::Log::log_global(1);


#
# These are essentially "static" helper functions.  They'll work if called
# as a class or object method, but need not be.
#

=pod

=head1 STATIC METHODS

This library provides the following utility methods:

=head2 get_volume_info(server, volume, [log])

Arguments:

=over

B<server> - the Server object to use when fetching volume information

B<volume> - the name of the volume of interest

B<log> - I<(optional)> NATE::Log object to log to

=back

Return value:

=over

A reference to a hash containing identifying information about the specified
volume.

=back

=cut

sub get_volume_info
{
    if ($_[0] eq 'EngineZapi' || ref($_[0]) && ref($_[0]) eq 'EngineZapi') {
        shift @_;
    }

    my ($server, $volume, $log) = @_;
    defined($volume) or die 'Invalid arguments!';
    $log = defined($log) ? $log : $Log;

    my $info = {'name' => $volume};

    $log->comment('getting volume info for "' . $volume . '"');


    my @results = $server->volume_show(volume => $volume);
    if (@results != 1) {
        die new NATE::Result::Fatal(
            'unexpected row count for "volume show" output');
    }

    $info->{'dsid'} = $results[0]->{'dsid'};
    $info->{'msid'} = $results[0]->{'msid'};
    $info->{'style'} = $results[0]->{'volume-style'};
    $info->{'type'} = $results[0]->{'type'};

    my $aggr = $results[0]->{'aggregate'};
    my $vserver = $results[0]->{'vserver'};


    @results = $server->cluster_identity_show();
    if (@results != 1) {
        die new NATE::Result::Fatal(
            'unexpected row count for "cluster identity show" output');
    }

    $info->{'cluster-id'} = $results[0]->{'uuid'};


    @results = $server->storage_aggregate_show(aggregate => $aggr);
    if (@results != 1) {
        die new NATE::Result::Fatal(
            'unexpected row count for "storage aggregate show" output');
    }

    $info->{'aggregate-uuid'} = $results[0]->{'uuid'};
    ($info->{'node'},) = $results[0]->{'nodes'} =~ /^(.*?)(,|$)/;


    @results = $server->debug_smdb_table_vserver_id_to_name_show(
        name => $vserver);
    if (@results != 1) {
        die new NATE::Result::Fatal(
            'unexpected row count for "debug smdb table ' .
            'vserver_id_to_name show" output');
    }

    # Not sure what's going on here (Server.pm bug?)
    if (ref($results[0]) eq 'ARRAY' && ref($results[0]->[0]) eq 'HASH') {
        $info->{'vserver-id'} = $results[0]->[0]->{'id'};
    } else {
        $info->{'vserver-id'} = $results[0]->{'id'};
    }

    return $info;
}

=pod

=head2 get_snap_uuid_by_name(server, volume, snapshot, [log])

Arguments:

=over

B<server> - the Server object to use when fetching snapshot information

B<volume> - the name of the volume containing the snapshot of interest

B<snapshot> - the name of the snapshot of interest

B<log> - I<(optional)> NATE::Log object to log to

=back

Return value:

=over

The instance UUID of the specified snapshot, as a string.

=back

=cut

sub get_snap_uuid_by_name
{
    if ($_[0] eq 'EngineZapi' || ref($_[0]) && ref($_[0]) eq 'EngineZapi') {
        shift @_;
    }

    my ($server, $volume, $snap_name, $log) = @_;
    defined($snap_name) or die 'Invalid arguments!';
    $log = defined($log) ? $log : $Log;

    $log->comment('getting uuid for snapshot "' . $snap_name .
                  '" on volume "' .  $volume . '"');

    my @results = $server->volume_snapshot_show(volume => $volume,
                                                snapshot => $snap_name);
    if (@results != 1) {
        die new NATE::Result::Fatal(
            'unexpected row count for "volume snapshot show" output');
    }

    return $results[0]->{'instance-uuid'};
}

sub make_zapi_dsid
{
    if ($_[0] eq 'EngineZapi' || ref($_[0]) && ref($_[0]) eq 'EngineZapi') {
        shift @_;
    }

    my ($dsid_major,) = @_;
    defined($dsid_major) or die 'Invalid arguments!';

    return { 'od-spinnp-dsid' => {
            'od-dsid-major-id' => $dsid_major,
            'od-dsid-minor-id' => 0
        }};
}

sub make_zapi_volume_id
{
    if ($_[0] eq 'EngineZapi' || ref($_[0]) && ref($_[0]) eq 'EngineZapi') {
        shift @_;
    }

    my ($vol_info,) = @_;
    defined($vol_info) or die 'Invalid arguments!';

    return { 'od-volume-identifier' => {
            'volume-msid' => make_zapi_dsid($vol_info->{'msid'}),
            'volume-dsid' => make_zapi_dsid($vol_info->{'type'} eq 'LS'
                                                ? $vol_info->{'dsid'} : 0),
            'cluster-id' => $vol_info->{'cluster-id'},
            'vserver-id' => $vol_info->{'vserver-id'}
        }};
}

sub make_zapi_snapshot_id
{
    if ($_[0] eq 'EngineZapi' || ref($_[0]) && ref($_[0]) eq 'EngineZapi') {
        shift @_;
    }

    my ($snap_uuid,) = @_;
    defined($snap_uuid) or die 'Invalid arguments!';

    return { 'od-snapshot-identifier' => {
            'snapshot-instance-uuid' => $snap_uuid
        }};
}

sub populate_zapi_result_try_one
{
    if ($_[0] eq 'EngineZapi' || ref($_[0]) && ref($_[0]) eq 'EngineZapi') {
        shift @_;
    }

    my ($xml, $result, $name, $xpath, $log) = @_;
    defined($xpath) or die 'Invalid arguments!';
    $log = defined($log) ? $log : $Log;

    my @values = zsmcli::get_values($xml, $xpath);
    if (@values >= 1) {
        @values == 1 or $log->warn('multiple values for ' . $name . ' XPath');
        $result->{$name} = $values[0];
    } else {
        $log->warn('no values for ' . $name . ' XPath');
    }

    return $result;
}

sub populate_zapi_result_require_one
{
    if ($_[0] eq 'EngineZapi' || ref($_[0]) && ref($_[0]) eq 'EngineZapi') {
        shift @_;
    }

    my ($xml, $result, $name, $xpath, $log) = @_;
    defined($xpath) or die 'Invalid arguments!';
    $log = defined($log) ? $log : $Log;

    my @values = zsmcli::get_values($xml, $xpath);
    if (@values >= 1) {
        @values == 1 or $log->warn('multiple values for ' . $name . ' XPath');
        $result->{$name} = $values[0];
    } else {
        die new NATE::Result::Fail('no values for ' . $name . ' XPath');
    }

    return $result;
}

sub populate_zapi_result_common
{
    if ($_[0] eq 'EngineZapi' || ref($_[0]) && ref($_[0]) eq 'EngineZapi') {
        shift @_;
    }

    my ($xml, $result, %options) = @_;
    defined($xml) or die 'Invalid arguments!';
    $result = {} unless (defined($result));
    my $log = defined($options{'log'}) ? $options{'log'} : $Log;

    my @values = zsmcli::get_values($xml, '/results/@status');
    if (@values == 0 || $values[0] ne 'passed') {
        @values = zsmcli::get_values($xml, '/results/@reason');
        if (@values == 0) {
            die new NATE::Result::Fail('unknown ZAPI error');
        } else {
            die new NATE::Result::Fail('ZAPI error, reason: ' . $values[0]);
        }
    }

    populate_zapi_result_require_one(
        $xml, $result, 'failure-reason', '/results/failure-reason', $log);
    if ($result->{'failure-reason'} != 0 &&
        $options{'throw-on-engine-failure'})
    {
        die new NATE::Result::Fail(
            'ZAPI response had a failure-reason of ' . $values[0]);
    }

    # TODO: find some reasonably simple way to include message arguments

    populate_zapi_result_require_one(
        $xml, $result, 'failure-message',
        '/results/failure-message//message-number', $log);
    populate_zapi_result_require_one(
        $xml, $result, 'failure-message-detail',
        '/results/failure-message-detail//message-number', $log);

    return $result;
}

sub populate_zapi_destination_status_common
{
    if ($_[0] eq 'EngineZapi' || ref($_[0]) && ref($_[0]) eq 'EngineZapi') {
        shift @_;
    }

    my ($xml, $result, %options) = @_;
    defined($xml) or die 'Invalid arguments!';
    $result = {} unless (defined($result));
    my $log = defined($options{'log'}) ? $options{'log'} : $Log;

    populate_zapi_result_try_one(
        $xml, $result, 'bytes-transferred',
        '/results/destination-status//bytes-transferred', $log);
    populate_zapi_result_try_one(
        $xml, $result, 'caller-reference-id-1',
        '/results/destination-status//caller-reference-id-1', $log);
    populate_zapi_result_try_one(
        $xml, $result, 'caller-reference-id-2',
        '/results/destination-status//caller-reference-id-2', $log);
    populate_zapi_result_try_one(
        $xml, $result, 'prepare-id',
        '/results/destination-status//prepare-id', $log);
    populate_zapi_result_try_one(
        $xml, $result, 'transfer-id',
        '/results/destination-status//transfer-id', $log);
    populate_zapi_result_try_one(
        $xml, $result, 'transfer-status',
        '/results/destination-status//transfer-status', $log);

    # TODO: also have a source-volume-id, but it's a pain to extract

    return $result;
}


#
# The constructor for EngineZapi objects.
#

=pod

=head1 CONSTRUCTORS

The following methods can be used to a construct an EngineZapi object:

=head2 new(server, [%options])

Arguments:

=over

B<server> - a Server object representing the cluster to contact

=back

Options:

=over

B<log> - NATE::Log object to log to

=back

Return value:

=over

A newly constructed EngineZapi object.

=back

=cut

sub new
{
    my ($class, $server, %options) = @_;
    defined($server) or die 'Invalid arguments!';

    my $self = {};

    $self->{'server'} = $server;
    $self->{'zsmcli'} = new zsmcli($server);

    $self->{'log'} = exists($options{'log'}) ? $options{'log'} : $Log;

    bless ($self, $class);
    return $self;
}


#
# These are instance methods for EngineZapi objects.
#

=pod

=head1 INSTANCE METHODS

The following methods can be called on EngineZapi objects:

=head2 destination_abort(dst_info, xfer_id, [%options])

Arguments:

=over

B<dst_info> - volume information for the destination volume

B<xfer_id> - transfer ID of the transfer to abort

=back

Return value:

=over

A reference to a hash containing the ZAPI outputs.

=back

=cut

sub destination_abort
{
    my ($self, $dst_info, $xfer_id, %options) = @_;
    defined($xfer_id) or die 'Invalid arguments!';
    $options{'log'} = $self->{'log'};

    my $args = {
            'destination-volume-id' => make_zapi_volume_id($dst_info),
            'transfer-id' => $xfer_id
        };

    $self->{'log'}->comment(
        'making d-snapmirror-destination-abort ZAPI for volume "' .
        $dst_info->{'name'} . '"');

    my $xml = $self->{'zsmcli'}->run_command({
            cmd => 'd-snapmirror-destination-abort',
            args => $args
        },
        $dst_info->{'node'});

    my $result = {};
    populate_zapi_result_common($xml, $result, %options);
    return $result;
}

=pod

=head2 destination_abort(dst_info, xfer_id, [%options])

Arguments:

=over

B<dst_info> - volume information for the destination volume

B<prepare_id> - prepare ID; must match the prepare ID in the volume's on-disk
state for the conclude to succeed

=back

Return value:

=over

A reference to a hash containing the ZAPI outputs.

=back

=cut

sub destination_conclude
{
    my ($self, $dst_info, $prepare_id, %options) = @_;
    defined($prepare_id) or die 'Invalid arguments!';
    $options{'log'} = $self->{'log'};

    my $args = {
            'destination-volume-id' => make_zapi_volume_id($dst_info),
            'prepare-id' => $prepare_id
        };
    if ($dst_info->{'style'} eq 'flex') {
        $args->{'destination-flex-volume-dsid'} = { 'od-spinnp-dsid' => {
                'od-dsid-major-id' => $dst_info->{'dsid'},
                'od-dsid-minor-id' => 0
            }};
    }

    $self->{'log'}->comment(
        'making d-snapmirror-destination-conclude ZAPI for volume "' .
        $dst_info->{'name'} . '"');

    my $xml = $self->{'zsmcli'}->run_command({
            cmd => 'd-snapmirror-destination-conclude',
            args => $args
        },
        $dst_info->{'node'});

    my $result = {};
    populate_zapi_result_common($xml, $result, %options);
    return $result;
}

=pod

=head2 destination_get_status(dst_info, [%options])

Arguments:

=over

B<dst_info> - volume information for the destination volume

=back

Return value:

=over

A reference to a hash containing the ZAPI outputs.

=back

=cut

sub destination_get_status
{
    my ($self, $dst_info, %options) = @_;
    defined($dst_info) or die 'Invalid arguments!';
    $options{'log'} = $self->{'log'};

    my $args = {
            'destination-volume-id' => make_zapi_volume_id($dst_info)
        };
    if ($dst_info->{'style'} eq 'flex') {
        $args->{'destination-flex-volume-dsid'} = { 'od-spinnp-dsid' => {
                'od-dsid-major-id' => $dst_info->{'dsid'},
                'od-dsid-minor-id' => 0
            }};
    }

    $self->{'log'}->comment(
        'making d-snapmirror-get-destination-status ZAPI for volume "' .
        $dst_info->{'name'} . '"');

    my $xml = $self->{'zsmcli'}->run_command({
            cmd => 'd-snapmirror-get-destination-status',
            args => $args,
        },
        $dst_info->{'node'});

    my $result = {};
    populate_zapi_result_common($xml, $result, %options);
    populate_zapi_destination_status_common($xml, $result, %options);
    populate_zapi_result_try_one(
        $xml, $result, 'data-engine-type',
        '/results/data-engine-type', $self->{'log'});

    # TODO: report destination stream progress

    return $result;
}

=pod

=head2 destination_prepare(src_info, dst_info, [%options])

Arguments:

=over

B<src_info> - volume information for the source volume

B<dst_info> - volume information for the destination volume

=back

Options:

=over

B<caller-reference-id-1> - explicit caller-reference-id-1 value to use for the
prepare operation

B<caller-reference-id-2> - explicit caller-reference-id-2 value to use for the
prepare operation

=back

Return value:

=over

A reference to a hash containing the ZAPI outputs.

=back

=cut

sub destination_prepare
{
    my ($self, $src_info, $dst_info, %options) = @_;
    defined($dst_info) or die 'Invalid arguments!';
    $options{'log'} = $self->{'log'};

    my ($uuid, $ref1, $ref2);
    if (exists($options{'caller-reference-id-1'})) {
        $ref1 = $options{'caller-reference-id-1'};
    } else {
        UUID::generate($uuid);
        UUID::unparse($uuid, $ref1);
    }
    if (exists($options{'caller-reference-id-2'})) {
        $ref2 = $options{'caller-reference-id-2'};
    } else {
        UUID::generate($uuid);
        UUID::unparse($uuid, $ref2);
    }

    my $args = {
            'caller-reference-id-1' => $ref1,
            'caller-reference-id-2' => $ref2,
            'destination-volume-id' => make_zapi_volume_id($dst_info),
            'source-volume-id' => make_zapi_volume_id($src_info)
        };
    if ($dst_info->{'style'} eq 'flex') {
        $args->{'destination-flex-volume-dsid'} =
            make_zapi_dsid($dst_info->{'dsid'});
    }

    $self->{'log'}->comment(
        'making d-snapmirror-destination-volume-prepare ZAPI for volume "' .
        $dst_info->{'name'} . '"');

    my $xml = $self->{'zsmcli'}->run_command({
            cmd => 'd-snapmirror-destination-volume-prepare',
            args => $args
        },
        $dst_info->{'node'});

    my $result = {};
    populate_zapi_result_common($xml, $result, %options);
    populate_zapi_destination_status_common($xml, $result, %options);

    # TODO: report latest snapshot

    return $result;
}

=pod

=head2 source_abort(src_info, xfer_id, [%options])

Arguments:

=over

B<src_info> - volume information for the source volume

B<xfer_id> - transfer ID of the transfer to abort

=back

Return value:

=over

A reference to a hash containing the ZAPI outputs.

=back

=cut

sub source_abort
{
    my ($self, $src_info, $xfer_id, %options) = @_;
    defined($xfer_id) or die 'Invalid arguments!';
    $options{'log'} = $self->{'log'};

    $self->{'log'}->comment(
        'making d-snapmirror-abort ZAPI for volume "' .
        $src_info->{'name'} . '"');

    my $xml = $self->{'zsmcli'}->run_command({
            cmd => 'd-snapmirror-abort',
            args => { 'transfer-id' => $xfer_id }
        },
        $src_info->{'node'});

    my $result = {};
    populate_zapi_result_common($xml, $result, %options);
    return $result;
}

=pod

=head2 source_conclude(src_info, xfer_id, [%options])

Arguments:

=over

B<src_info> - volume information for the source volume

B<xfer_id> - transfer ID of the transfer to conclude

=back

Return value:

=over

A reference to a hash containing the ZAPI outputs.

=back

=cut

sub source_conclude
{
    my ($self, $src_info, $xfer_id, %options) = @_;
    defined($xfer_id) or die 'Invalid arguments!';
    $options{'log'} = $self->{'log'};

    $self->{'log'}->comment(
        'making d-snapmirror-conclude ZAPI for volume "' .
        $src_info->{'name'} . '"');

    my $xml = $self->{'zsmcli'}->run_command({
            cmd => 'd-snapmirror-conclude',
            args => { 'transfer-id' => $xfer_id }
        },
        $src_info->{'node'});

    my $result = {};
    populate_zapi_result_common($xml, $result, %options);
    return $result;
}

=pod

=head2 source_get_status(src_info, xfer_id, [%options])

Arguments:

=over

B<src_info> - volume information for the source volume

B<xfer_id> - transfer ID of the transfer of interest

=back

Return value:

=over

A reference to a hash containing the ZAPI outputs.

=back

=cut

sub source_get_status
{
    my ($self, $src_info, $xfer_id, %options) = @_;
    defined($xfer_id) or die 'Invalid arguments!';
    $options{'log'} = $self->{'log'};

    $self->{'log'}->comment(
        'making d-snapmirror-get-status ZAPI for volume "' .
        $src_info->{'name'} . '"');

    my $xml = $self->{'zsmcli'}->run_command({
            cmd => 'd-snapmirror-get-status',
            args => { 'transfer-id' => $xfer_id }
        },
        $src_info->{'node'});

    my $result = {};
    populate_zapi_result_common($xml, $result, %options);
    populate_zapi_result_try_one(
        $xml, $result, 'bytes-transferred',
        '/results/bytes-transferred', $self->{'log'});
    populate_zapi_result_try_one(
        $xml, $result, 'checkpoint-bytes-transferred',
        '/results/checkpoint-bytes-transferred', $self->{'log'});
    populate_zapi_result_try_one(
        $xml, $result, 'data-engine-type',
        '/results/data-engine-type', $self->{'log'});
    populate_zapi_result_try_one(
        $xml, $result, 'replication-rate',
        '/results/replication-rate', $self->{'log'});
    populate_zapi_result_try_one(
        $xml, $result, 'time-elapsed',
        '/results/time-elapsed', $self->{'log'});
    populate_zapi_result_try_one(
        $xml, $result, 'transfer-status',
        '/results/transfer-status', $self->{'log'});
    populate_zapi_result_try_one(
        $xml, $result, 'transfer-type',
        '/results/transfer-type', $self->{'log'});

    # TODO: report destination stream progress, destination transfer progress,
    # optional reference snapshot, and transfer snapshot

    return $result;
}

=pod

=head2 source_get_transfer_ids(src_info, dst_info, [%options])

Arguments:

=over

B<src_info> - volume information for the source volume for transfers
of interest

B<dst_info> - volume information for the destination volume for transfers
of interest

=back

Return value:

=over

A reference to a hash containing the ZAPI outputs.

=back

=cut

sub source_get_transfer_ids
{
    my ($self, $src_info, $dst_info, %options) = @_;
    defined($dst_info) or die 'Invalid arguments!';
    $options{'log'} = $self->{'log'};

    my $result = {};
    my $args = {
            'destination-volume-id' => make_zapi_volume_id($dst_info),
            'source-volume-id' => make_zapi_volume_id($src_info)
        };

    $self->{'log'}->comment(
        'making d-snapmirror-get-source-transfer-ids ZAPI for transfers ' .
        'from "' . $src_info->{'name'} . '" to "' . $dst_info->{'name'} . '"');

    my $xml = $self->{'zsmcli'}->run_command({
            cmd => 'd-snapmirror-get-source-transfer-ids',
            args => $args
        },
        $src_info->{'node'});

    populate_zapi_result_common($xml, $result, %options);
    my @transfer_ids = zsmcli::get_values(
        $xml, '/results/transfer-id-list/snapmirror-transfer-id');
    if (@transfer_ids > 0) {
        $result->{'transfer-id-list'} = join(', ', @transfer_ids);
    }

    return $result;
}

=pod

=head2 source_start(src_info, dst_info, prepare_id, xfer_snap, [%options])

Arguments:

=over

B<src_info> - volume information for the source volume

B<dst_info> - volume information for the destination volume

B<prepare_id> - the prepare ID to use in accessing the destination volume;
must match the prepare ID in the destination's on-disk state

B<xfer_snap> - transfer snapshot to use

=back

Options:

=over

B<transfer-id> - explicit transfer-id value to use for the transfer

B<ref-snapshot-id> - reference snapshot to use; if not present, a baseline
transfer is assumed

B<is-quiesced> - "true"/"false" - specifies whether the source volume
is quiesced

B<destination-bandwidth-limit> - specifies bandwidth throttling

B<data-engine-type> - "physical"/"logical" - specifies the type of engine
to use; default is "physical"

B<exo-volume-metadata-to-transfer> - specifies which exo-volume plug-ins
should be run during the transfer

B<preserve-readdir-cookies> - "true"/"false" - specifies whether readdirZ<>()
cookies must be preserved

=back

Return value:

=over

A reference to a hash containing the ZAPI outputs.

=back

=cut

sub source_start
{
    my ($self, $src_info, $dst_info, $prepare_id, $xfer_snap, %options) = @_;
    defined($xfer_snap) or die 'Invalid arguments!';
    $options{'log'} = $self->{'log'};

    my $result = {};
    my $xfer_id;
    if (exists($options{'transfer-id'})) {
        $xfer_id = $options{'transfer-id'};
    } else {
        my $uuid;
        UUID::generate($uuid);
        UUID::unparse($uuid, $xfer_id);
    }
    # Return the transfer ID used, as a convenience
    $result->{'transfer-id'} = $xfer_id;

    my $ref_snap;
    if (exists($options{'ref-snapshot-id'})) {
        $ref_snap = $options{'ref-snapshot-id'};
    }

    my $src_args = {
            'source-aggr-uuid' => $src_info->{'aggregate-uuid'},
            'source-volume-id' => make_zapi_volume_id($src_info),
            'transfer-snapshot-id' => make_zapi_snapshot_id($xfer_snap)
        };
    if (exists($options{'is-quiesced'})) {
        $src_args->{'is-quiesced'} = $options{'is-quiesced'};
    }
    if (defined($ref_snap)) {
        $src_args->{'ref-snapshot-id'} = make_zapi_snapshot_id($ref_snap);
    }
    if ($src_info->{'style'} eq 'flex') {
        $src_args->{'source-flex-volume-dsid'} =
            make_zapi_dsid($src_info->{'dsid'});
    }


    my $args = {
            'd-snapmirror-destinations' => { 'd-snapmirror-destination-info' => {
                'destination-bandwidth-limit' =>
                    (exists($options{'destination-bandwidth-limit'})
                         ? $options{'destination-bandwidth-limit'} : 0),
                'destination-volume-id' => make_zapi_volume_id($dst_info),
                'prepare-id' => $prepare_id
            }},
            'data-engine-type' =>
                (exists($options{'data-engine-type'})
                     ? $options{'data-engine-type'} : 'physical'),
            'is-incremental' => (defined($ref_snap) ? 'true' : 'false'),
            'source-info' => { 'd-snapmirror-source-info' => $src_args },
            'transfer-id' => $xfer_id
        };
    if (exists($options{'exo-volume-metadata-to-transfer'})) {
        $args->{'exo-volume-metadata-to-transfer'} =
            $options{'exo-volume-metadata-to-transfer'};
    }
    if (exists($options{'preserve-readdir-cookies'})) {
        $args->{'preserve-readdir-cookies'} =
            $options{'preserve-readdir-cookies'};
    }

    $self->{'log'}->comment(
        'making d-snapmirror-start ZAPI for volume "' .
        $src_info->{'name'} . '"');

    my $xml = $self->{'zsmcli'}->run_command({
            cmd => 'd-snapmirror-start',
            args => $args
        },
        $src_info->{'node'});

    populate_zapi_result_common($xml, $result, %options);
    return $result;
}

1;
