# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/dp/snapmirror/paloma/unit_test/RR_0/SnapDiff/SIS_clones/common_wrappers.pm#1 $

## @summary common wrappers for SnapDiff SIS-clone unit tests
##
## @description
##    This is a private shared library for the all the unit tests
##    for WAFL_SNAP_DIFF support of constant-time SIS clones in RR.
##
## @author remi@netapp.com
## @status private

package dp::snapmirror::paloma::unit_test::RR_0::SnapDiff::SIS_clones::common_wrappers;
use warnings;
use strict;

use Carp;
use CLI::FileIO qw(wrfile);
use Dense qw(:action dense_toggle dense_set_license);
use FSR::FSRutils;
use NOTE;
use NOTE::Fiji;
use Params::Validate qw(:DEFAULT :types);
use SayWrap qw();
use Tharn;
use Wreck;

# As long as this is set, we will run "stubbed-out" versions of
# these library routines, which should work on any release.
# Change this to zero to attempt to run the "unstubbed" versions
# of the libraries, which require a clone/semainrr build with
# the testing hooks from w-capacity.
our $STUBBED = 0;

# Stubbed-out library calls will create and delete files in the
# root of the volume, prefixed with this string, to indicate
# whether clone splitting is "paused" on a volume or a file.
my $CLONE_SPLIT_DISABLER = "clone_split_disabled";

BEGIN {
    require Exporter;
    our @ISA = qw(Exporter);
    our @EXPORT = qw();
    our @EXPORT_OK = qw(sd_param_init sd_volume_init
			qsm_param_init qsm_qtree_init range_count 
			records_to_ranges compare_record_lists
			coalesce_record_list purge_record_list
			modblock howmany select_random_fbns
			clone_create clone_list clone_split_wait
			clone_splitting_pause_fiji clone_splitting_clear_fiji
			fiji_wait_for_trigger
			filer_stat dumpblock_info wafl_showvbns wafl_touch
			wafl_corrupt_block wafl_corrupt_file wafl_punch_holes
			human_size_to_bytes bytes_to_human_size
			setflag printflag verify_qsm_transfer_size
			reset_counters fetch_counters verify_mirror
		       );
}

# Strips whitespace from a code fragment.
sub strip_whitespace( $ ) {
  $_[0] =~ s/\n\s*/ /g; # eliminate newlines and extra whitespace
  $_[0] =~ s/\A\s*//g;  # strip leading whitespace
  $_[0] =~ s/\s*\z//g;  # strip trailing whitespace
}

sub howmany( $$ ) {
  my ($x, $y) = @_;
  return int(($x + $y - 1) / $y);
}

# Rebuilds a makeshift NATE hostspec from NOTE::Volume object.
sub volume_to_hostspec( $ ) {
  my ($volume) = @_;
  my $hostrec = $volume->Filer->conn->hostp();
  my $hostspec = $hostrec->id();
  $hostspec .= ";hosttype=" . $hostrec->hosttype();
}

# Returns whether the provided ranges are valid:
#    - first FBN must be <= last FBN
#    - ranges must come in sorted order
#    - ranges should not overlap
#    - single block ranges should be just "N", not "N-N"
sub validate_ranges {
  my $rangelist;
  if (scalar(@_) == 1 && ref($_[0]) eq "ARRAY") {
    $rangelist = $_[0];
  } else {
    $rangelist = \@_;
  }
  my $prev = -1;
  foreach my $range (@$rangelist) {
    return undef unless ($range =~ /\A(\d+)(?:\-(\d+))?\z/);
    return undef if (defined($2) && $1 >= $2);
    return undef if ($1 <= $prev);
    $prev = defined $2 ? $2 : $1;
  }
  return 1;
}

## @name ::printflag
## @usage printflag($filer, $flag_name)
## @synopsis
##    Retrieves the value of a flag on a filer, using the 'printflag' command.
sub printflag( $$ ) {
  validate_pos(@_, { isa => 'NOTE::Filer' }, { type => SCALAR });
  my ($filer, $flag) = @_;
  my $output = $filer->say("priv set -q diag ; printflag $flag");
  if ($output =~ /no such flag/) {
    $output = $filer->say("priv set -q test ; printflag $flag");
  }

  unless ($output =~ /\A\Q$flag\E\s+\=\s+(\S+)\s*\z/) {
    croak "printflag failed: $output\n";
  }
  my $value = $1;

  return $value;
}

## @name ::setflag
## @usage setflag($filer, $flag_name, $flag_value)
## @synopsis
##    Sets a flag on a filer, using the 'setflag' command.
sub setflag( $$$ ) {
  validate_pos(@_, { isa => 'NOTE::Filer' },
	           { type => SCALAR }, { type => SCALAR });
  my ($filer, $flag, $value) = @_;
  my $output = $filer->say("priv set -q diag ; setflag $flag $value");
  if ($output =~ /no such flag/) {
    $output = $filer->say("priv set -q test ; setflag $flag $value");
  }

  if ($output) {
    croak "setflag failed: $output\n";
  }

  # Some flags are not settable.  Check the return.
  my $actual = printflag($filer, $flag);
  unless ($actual == $value) {
    croak "setflag \"$flag\" did not retain value\n";
  }

  return 1;
}

## @name ::filer_stat
## @usage filer_stat($volume, $path, %options)
## @synopsis
##    Uses the 'slist' command to retrieve the inode number
##    and size of a file in the given volume.
##    The specified path should be relative to the root of the volume.
##    Returns the inode number in a scalar context.
##    Returns both the inode number and the size otherwise.
## @args snapname Optional: name of snapshot in which to do the lookup
## @args enoent_ok Optional: if set, return undef if the path does not exist
##                           (default behavior is to throw an exception)
sub filer_stat( $$;% ) {
  my @pos = splice(@_, 0, 2);
  validate_pos(@pos, { isa => 'NOTE::Volume' }, { type => SCALAR } );
  my ($volume, $path) = @pos;
  my %opts = validate(@_ => {
			snapname => { type => SCALAR, optional => 1 },
			enoent_ok => { type => SCALAR, optional => 1 },
                     });

  my $fpath;
  if (exists $opts{snapname}) {
    $fpath = sprintf("/vol/%s/.snapshot/%s/%s",
		     $volume->name, $opts{snapname}, $path);
  } else {
    $fpath = sprintf("/vol/%s/%s", $volume->name, $path);
  }
  my $text = $volume->Filer->say("priv set -q diag ; slist $fpath");

  if ($text =~ /\A\Q$fpath\E: Invalid path\s*\z/) {
    return undef if ($opts{enoent_ok});
    die "$text\n";
  }

  unless ($text =~ s/\A\s*Inode\#\s*File size\s*Name\s*\n// &&
	  $text =~ s/\A\-+\s+\-+\s+\-+\s*?\n//) {
    die "error parsing slist header: $text\n";
  }
  my ($fid, $size);
  if ($text =~ s/\A\s*(\d+)\s+(\d+)\s+(\S+)\s*?(?:\n|\z)//) {
    die "path mismatch $3" unless ($3 eq $fpath);
    $fid = $1;
    $size = $2;
  } else {
    die "error matching slist output: $text\n";
  }

  # XXX for the moment, we don't bother parsing
  # the rest of the 'slist' output for all the streams.

  return wantarray ? ($fid, $size) : $fid;
}

## @name ::select_random_fbns
## @usage select_random_fbns( $range, %options )
## @synopsis
##    Randomly chooses a set of blocks to modify, and returns
##    a list of FBN ranges, e.g. ("23", "75-100", "50", "2000").
##    The FBN ranges selected will never overlap or be adjacent.
##    The sole input argument is either a range from which to
##    randomly choose FBNs, or simply the number of blocks in the file.
## @args num_ranges Optional: the number of FBN ranges to return (default=random)
## @args small_ranges Optional: whether to return "small" ranges (default=true)
{
  # Given a maximum count, and a list of potential sub-ranges (each
  # either fixed count or a percentage), randomly choose some sub-count.
  sub random_cnt( $\@ ) {
    my ($max, $range_cnts) = @_;
    my $cnt = $max + 1;

    while ($cnt > $max) {
      my $idx = int(rand(scalar(@$range_cnts)));
      $cnt = $range_cnts->[$idx];
      if ($cnt =~ s/\%\z//) {
	$cnt = sprintf("%.0f", $max * ($cnt / 100.0));
	$cnt = 1 if ($cnt == 0);
      }
    }
    return $cnt;
  }

  my @small_range_cnts = qw(1 20 10% 25% 50%);
  my @large_range_cnts = qw(10% 25% 50%);

  sub select_random_fbns( $;% ) {
    my @pos = splice(@_, 0, 1);
    validate_pos(@pos, { callbacks => {
				'valid range' => sub { validate_ranges($_[0]); }
				      } } );
    my ($range) = @pos;
    my %opts = validate(@_ => {
			num_ranges => {
				type => SCALAR, regex => qr/^\d+$/,
				optional => 1
			},
			small_ranges => {
				type => SCALAR,
				optional => 1, default => 1
			},
                     });

    my ($start, $end);
    die "invalid range \"$range\"\n" unless ($range =~ /\A(\d+)(?:\-(\d+))?\z/);
    if (defined $2) {
      ($start, $end) = ($1, $2);
    } else {
      ($start, $end) = (0, $1 - 1);
    }

    # Choose between 3 and 10 random FBNs (unless given a specific count).
    # For each one, pick a random extent from that starting FBN, either
    # as a fixed constant or a percentage of blocks to the next range.
    my $fbn_count = exists $opts{num_ranges} ?
			$opts{num_ranges} : 3 + int(rand(8));

    # Build the list of random FBNs.  Make sure to avoid duplicates.
    my %fbns_hit = ();
    my %adjacents_hit = (); # to make sure we never pick adjacent FBNs
    my $max = ($end - $start + 1);
    unless (($fbn_count * 3) < $max) {
      die "insufficient FBNs to select ranges\n";
    }
    for (my $i = 0; $i < $fbn_count; $i++) {
      my $fbn;
      do {
	$fbn = int(rand($max)) + $start;
      } while (exists $adjacents_hit{$fbn});
      $fbns_hit{$fbn} = 1;
      $adjacents_hit{$fbn} = 1;
      $adjacents_hit{$fbn + 1} = 1;
      $adjacents_hit{$fbn - 1} = 1;
    }
    my @fbns = sort {$a <=> $b} keys %fbns_hit;

    my @range_cnts = $opts{small_ranges} ?
			@small_range_cnts : @large_range_cnts;

    # For each FBN in the list, pick a count to build the range.
    $fbns[$fbn_count] = $end + 1;
    my @ranges = ();
    for (my $i = 0; $i < $fbn_count; $i++) {
      my $cnt = random_cnt($fbns[$i+1] - $fbns[$i] - 1, @range_cnts);
      if ($cnt == 1) {
	push @ranges, $fbns[$i];
      } else {
	push @ranges, sprintf("%d-%d", $fbns[$i], $fbns[$i] + $cnt - 1);
      }
    }

    return @ranges;
  }
}

## @name ::range_count
## @usage range_count(@ranges)
## @synopsis
##   Calculates the number of blocks being covered by the given ranges.
sub range_count( @ ) {
  my (@ranges) = @_;
  my $block_count = 0;
  foreach my $range (@ranges) {
     die "invalid range $range\n"
       unless ($range =~ /\A(\d+)(?:\-(\d+))?\z/);
     my $count = defined $2 ? $2 - $1 + 1 : 1;
     $block_count += $count;
  }
  return $block_count;
}

## @name ::human_size_to_bytes
## @usage human_size_to_bytes($string)
## @synopsis
##    Given a human-readable size description (e.g. "500g", "64Kb", "32.4 MB"),
##    returns the corresponding number of bytes represented by that string.
##    Rounds the answer to the nearest byte count.
## @bugs
##    It would be nice to use Utils::Units::convert_to_bytes() instead, but
##    that's not very flexible in what it accepts, e.g. it take "12m" but not
##    "12mb" or "12MB", and it won't accept anything with a decimal point
##    like "12.4m".
sub human_size_to_bytes( $ ) {
  my ($human_size) = @_;

  unless ($human_size =~ /\A(\d+(\.\d+)?)\s*([kmgtp])?b?\z/i) {
    die "Could not parse size description \"$human_size\"\n";
  }
  my ($cnt, $decimal, $suffix) = ($1, $2, $3);

  if ($suffix) {
    $suffix =~ tr/A-Z/a-z/; # change all to lower-case
    if ($suffix eq "k") {
      $cnt *= 1024;
    } elsif ($suffix eq "m") {
      $cnt *= 1048576;
    } elsif ($suffix eq "g") {
      $cnt *= 1073741824;
    } elsif ($suffix eq "t") {
      $cnt *= 1099511627776;
    } else { # $suffix eq "p"
      $cnt *= 1125899906842624;
    }
  }

  # round result to nearest byte
  return sprintf("%.0f", $cnt);
}

## @name ::bytes_to_human_size
## @usage bytes_to_human_size($byte_cnt)
## @synopsis
##    Given a byte count, returns a human-readable size description.
##    Rounds the result to no more than two decimal places.
## @bugs
##    It would be nice to have this as part of a standard library like
##    Utils::Units instead of here.  At the moment, the library does not
##    have a routine that will determine the "appropriate" unit for you.
##    The caller has to explicitly request "MB" or "GB", etc.
sub bytes_to_human_size( $ ) {
  my ($size) = @_;
  my $type = "";
  if ($size > 1024) {
    $size /= 1024;
    $type = "KB";
    if ($size > 1024) {
      $size /= 1024;
      $type = "MB";
      if ($size > 1024) {
	$size /= 1024;
	$type = "GB";
	if ($size > 1024) {
	  $size /= 1024;
	  $type = "TB";
	  if ($size > 1024) {
	    $size /= 1024;
	    $type = "PB";
	  }
	}
      }
    }
  }
  $size = sprintf("%.2f", $size);
  $size =~ s/(\.\d)0\z/$1/;
  $size =~ s/\.0\z//;
  return "$size$type";
}

## @name ::records_to_ranges
## @usage records_to_ranges( @record_list )
## @synopsis
##    Converts a list of records (each of which is a hash with at least
##    the fields "fbn", "vbn", and "count") into a list of ranges
##    (i.e. strings of the form "M-N") representing the FBNs described
##    by the records.
sub records_to_ranges( \@ ) {
  my ($recs) = @_;

  my @ranges = ();
  my $cur_fbn = undef;
  my $cur_cnt = 0;

  foreach my $rec (@$recs) {

    # Make sure the record has all the required fields.
    unless (exists $rec->{fbn} && exists $rec->{vbn} && exists $rec->{count}) {
      my $str = "{ ";
      while (my ($key, $value) = each %$rec) {
	$str .= "$key => $value, ";
      }
      $str .= "}";
      die "incomplete record: $str\n";
    }

    # If we haven't started recording a range yet, this one does.
    unless (defined $cur_fbn) {
      $cur_fbn = $rec->{fbn};
      $cur_cnt = $rec->{count};
      next;
    }

    # If this record extends the current range, append it.
    if ($rec->{fbn} == $cur_fbn + $cur_cnt) {
      $cur_cnt += $rec->{count};
      next;
    }

    # Conclude the current range and start a new one.
    if ($cur_cnt == 1) {
      push @ranges, $cur_fbn;
    } else {
      push @ranges, sprintf("%d-%d", $cur_fbn, $cur_fbn + $cur_cnt - 1);
    }
    $cur_fbn = $rec->{fbn};
    $cur_cnt = $rec->{count};

  }

  # Conclude the final range.
  if (defined $cur_fbn) {
    if ($cur_cnt == 1) {
      push @ranges, $cur_fbn;
    } else {
      push @ranges, sprintf("%d-%d", $cur_fbn, $cur_fbn + $cur_cnt - 1);
    }
  }

  return @ranges;
}

## @name ::coalesce_record_list
## @usage coalesce_record_list( @record_list, %options )
## @synopsis
##    Given a list of records (where each record is a hash containing at
##    least the fields "fbn" and "count"), returns a reference to a new
##    record list which matches the first, but for which adjacent records
##    have been coalesced where possible.
## @args ignore_vbns Optional: if set, presume all VBNs are uninitialized
sub coalesce_record_list( \@;% ) {
  my ($recs, %opts) = @_;

  # Count how many records have a vbn.
  my $check_vbns = scalar(grep { exists $_->{vbn} } @$recs);
  $check_vbns = 0 if ($opts{ignore_vbns});

  # Count how many records have a type.
  my $check_type = scalar(grep { exists $_->{type} } @$recs);

  my @records = ();

  my $cur_rec = undef;

  foreach my $rec (@$recs) {
    # If we haven't started a record yet, use this one.
    unless (defined $cur_rec) {
      $cur_rec = { fbn => $rec->{fbn}, count => $rec->{count} };
      $cur_rec->{vbn} = $rec->{vbn} if ($check_vbns);
      $cur_rec->{type} = $rec->{type} if ($check_type);
      next;
    }

    # If this record extends the current record, append it.
    if ($rec->{fbn} == $cur_rec->{fbn} + $cur_rec->{count} &&
	(!$check_vbns || $rec->{vbn} == $cur_rec->{vbn} + $cur_rec->{count}) &&
	(!$check_type || $rec->{type} eq $cur_rec->{type})) {
      $cur_rec->{count} += $rec->{count};
      next;
    }

    # Conclude the current range and start a new one.
    push @records, $cur_rec;
    $cur_rec = { fbn => $rec->{fbn}, count => $rec->{count} };
    $cur_rec->{vbn} = $rec->{vbn} if ($check_vbns);
    $cur_rec->{type} = $rec->{type} if ($check_type);
  }

  # Conclude the final record.
  if (defined $cur_rec) {
    push @records, $cur_rec;
  }

  return \@records;
}

## @name ::purge_record_list
## @usage purge_record_list( @record_list, @ranges )
## @synopsis
##    Given a list of records (where each record is a hash containing
##    at least the fields "fbn" and "count") and ranges, returns a
##    reference to a new list of records which doesn't contain the
##    ranges.
sub purge_record_list( \@\@ ) {
  validate_pos(@_, { type => ARRAYREF },
                   { type => ARRAYREF, callbacks => {
                                'valid range' =>
                                        sub { validate_ranges($_[0]); }
                                }
                   }
              );
  my ($recs, $ranges) = @_;

  my @output = ();

 RECORD: foreach my $caller_rec (@$recs) {

      my %rec = %$caller_rec;
      my $rec = \%rec;

    RANGE: foreach my $cur_range (@$ranges) {
      my ($range_start, $range_end);
      if ($cur_range =~ /\A(\d+)-(\d+)\z/) {
        $range_start = $1;
        $range_end = $2;
      } else {
        $range_start = $range_end = $cur_range;
      }

      # Range ends before this record.
      if ($range_end < $rec->{fbn}) {
        next RANGE;
      }

      # Range completely engulfs this record.
      if ($range_start <= $rec->{fbn} &&
          $range_end >= $rec->{fbn} + $rec->{count} - 1) {
        next RECORD;
      }

      # Range starts after this record.  Since the ranges are sorted,
      # all the remaining ranges will also start after this record.
      if ($range_start >= $rec->{fbn} + $rec->{count}) {
        last RANGE;
      }

      if ($range_start > $rec->{fbn}) {
        # Range starts in the partway through this record.
        # Trim off the first part of the record not covered by the range.
        my $trim_fbns = $range_start - $rec->{fbn};
        my $trim_rec = { fbn => $rec->{fbn},
                         count => $trim_fbns };
        $trim_rec->{vbn} = $rec->{vbn} if (exists $rec->{vbn});
        $trim_rec->{type} = $rec->{type} if (exists $rec->{type});
        push @output, $trim_rec;

        # Range engulfs the remaining part of this record.
        if ($range_end >= $rec->{fbn} + $rec->{count} - 1) {
          next RECORD;
        }
      }

      # Chop FBNs off the start of the record to the end of the range.
      my $chop_fbns = $range_end - $rec->{fbn} + 1;
      $rec->{fbn} += $chop_fbns;
      if (exists $rec->{vbn}) {
        $rec->{vbn} += $chop_fbns;
      }
      $rec->{count} -= $chop_fbns;

    }

    # Done trimming this record according to the ranges.
    push @output, $rec;

  }

  return \@output;
}

## @name ::compare_record_lists
## @usage compare_record_lists( @record_list1, @record_list2, %options )
## @synopsis
##    Given a pair of record lists (where each record is a hash containing
##    at least the fields "fbn" and "count"), iterates through the lists
##    and compares them.  Returns whether the two lists match.
## @args check_vbns Optional: if set, compare the "vbn" field too
## @args check_type Optional: if set, compare the "type" field too
sub compare_record_lists( \@\@;% ) {
  my ($recsA, $recsB, %opts) = @_;

  my $num_recs = scalar(@$recsA);
  if (scalar(@$recsB) != $num_recs) {
    return 0;
  }

  for (my $i = 0; $i < $num_recs; $i++) {
    my ($recA, $recB) = ($recsA->[$i], $recsB->[$i]);

    unless (exists $recA->{fbn} && exists $recA->{count} &&
	    exists $recB->{fbn} && exists $recB->{count}) {
      die "records do not contain both fbn and count\n";
    }
    if ($recA->{fbn} != $recB->{fbn} ||
	$recA->{count} != $recB->{count}) {
      return 0;
    }

    if ($opts{check_vbns}) {
      unless (exists $recA->{vbn} && exists $recB->{vbn}) {
	die "records do not contain vbn\n";
      }
      if ($recA->{vbn} != $recB->{vbn}) {
	return 0;
      }
    }

    if ($opts{check_type}) {
      unless (exists $recA->{type} && exists $recB->{type}) {
	die "records do not contain type\n";
      }
      if ($recA->{type} ne $recB->{type}) {
	return 0;
      }
    }
  }

  return 1;
}

{
  my %sd_default_params = (
	FILER => undef,
	CLIENT => "localhost",
	CLONE_FILE_SIZE => "16MB",
  );

  my %qsm_default_params = (
	FILER_SRC => undef,
	FILER_DST => undef,
	FILER_SRC_AGGR => undef,
	FILER_DST_AGGR => undef,
	CLIENT => "localhost",
	CLONE_FILE_SIZE => "8GB",
  );

  sub param_init( \%% ) {
    my ($default_params, %extra_params) = @_;

    my %params = ();

    while (my ($p, $value) = each %$default_params) {
      next if (exists $extra_params{$p});
      if (defined $value) {
	$params{$p} = param($p, -default => $value);
      } else {
	$params{$p} = param($p, -mesg => $p);
      }
    }

    while (my ($p, $value) = each %extra_params) {
      if (defined $value) {
	$params{$p} = param($p, -default => $value);
      } else {
	$params{$p} = param($p, -mesg => $p);
      }
    }

    my $p = 'CLONE_FILE_SIZE';
    $params{$p} = param($p, -set => human_size_to_bytes($params{$p}));

    return \%params;
  }

## @name ::sd_param_init
## @usage sd_param_init(EXTRA_PARAM1 => $default_value, EXTRA_PARAM2 => undef)
## @synopsis
##    Initializes and returns all the "standard" parameters for
##    SIS-clone tests and returns them as a hash.
##    Additional parameters are specified as arguments, mapping to their
##    default values (or undef if the parameter is required).
  sub sd_param_init(%) {
    param_init(%sd_default_params, @_);
  }

## @name ::qsm_param_init
## @usage qsm_param_init(EXTRA_PARAM1 => $default_value, EXTRA_PARAM2 => undef)
## @synopsis
##    Initializes and returns all the "standard" parameters for
##    SIS-clone QSM tests and returns them as a hash.
##    Additional parameters are specified as arguments, mapping to their
##    default values (or undef if the parameter is required).
  sub qsm_param_init(%) {
    param_init(%qsm_default_params, @_);
  }
}

# Registers FIJI points for preventing clone splitting for all clones
# in a volume, and also for specific FBNs on a given inode.
sub register_fiji () {
  if ($STUBBED) {
    NOTE::Fiji->register(
	quiet => 1,
	tag   => 'delay WAFL messages for a specific inode',
	prog  => '(repeat %d (wafl.wp1m.delay.mdp (&& (eq @vol %s) (== @fileid %d)) (return 1)))',
	param => [ 'repeat', 'volume', 'fileid' ],
    );
  } else {
    NOTE::Fiji->register(
	quiet => 1,
	tag   => 'pause clone splitting on a volume',
	prog  => '(repeat %d (wafl.sis.fixup.skip.vol (== @fsid %d) (return 1)))',
	param => [ 'repeat', 'fsid' ],
    );
    NOTE::Fiji->register(
	quiet => 1,
	tag   => 'pause clone splitting on an inode',
	prog  => '(repeat %d (wafl.sis.fixup.skip.inode (&& (&& (== @fsid %d) (== @inum %d)) (>= @fbn %d)) (return 1)))',
	param => [ 'repeat', 'fsid', 'inum', 'fbn' ],
    );
  }
}

# @name ::volume_init
# @usage volume_init( $filer, $clone_file_size, %options )
# @synopsis
#    This routine calculates the size of the volume needed based on the
#    clone_file_size and num_clones. It creates a flexible volume in the
#    aggregate (if specified, else creates a new aggregate).
#    This requires $filer and $clone_file_size to be specified.
#    Optional params are filer_aggr, num_clones and corruption_test.
#    Returns the NOTE::Volume it creates.
sub volume_init ( $$;% ) {
  my @pos = splice(@_, 0, 2);
  validate_pos(@pos, { type => SCALAR, regex => qr/^.+$/s},
	             { type => SCALAR, regex => qr/^\d+$/ });
  my ($filer, $clone_file_size) = @pos;
  my %opts = validate(@_ => {
			filer_aggr => { type => SCALAR,
					optional => 1 },
			num_clones => {
				type => SCALAR, regex => qr/^\d+$/,
				optional => 1, default => 1 },
			corruption_test => { type => SCALAR, optional => 1 },
                     });

  # For a mixed-mode system, the hostspec should have both
  # "default_ip_7g" and "default_ip_gx" defined.
  # In that situation, ideally we would simply *override* the hosttype to be 7G,
  # and then connect to the filer appropriately.
  # However, there does not appear to be a way to override that field in
  # either a hostspec or hostrec.  Even if we could override it in the hostrec,
  # NOTE::Filer::connect insists on receiving a hostspec, not an object.
  my $hosttype = host($filer)->hosttype();
  unless ($hosttype =~ /\Afiler(\-vsim|\-cfe)?\z/) {
    my $extra = "";
    if ($hosttype =~ /\Afiler-gx/) {
      $extra = "Cannot run against GX -- perhaps just change the hosttype?\n"
    }
    die "${extra}hosttype must be filer or filer-vsim\n";
  }

  my $filer_conn = NOTE::Filer->connect(host => $filer);

  my $aggr;
  if ($opts{filer_aggr}) {
    $aggr = NOTE::Aggregate->attach(Filer => $filer_conn,
				    name => $opts{filer_aggr});
  } else {
    # The only scripts that create aggregates are testing
    # larger-aggregate upgrade, so start with a 32-bit aggregate.
    $aggr = NOTE::Aggregate->create(Filer => $filer_conn, indirect_type => 32,
				    share => 0, recycle => 0);
  }

  # Make the volume big enough to hold an unshared copy of the file
  # and each clone.  Since each clone operation turns the source file
  # into a clone as well, we need enough space for double the number of
  # requested clones.  Then add another 20% for the snap reserve, and
  # then add an extra half of the file size (just for "wiggle room").
  my $size = $clone_file_size;
  $size = $size * $opts{num_clones} * 2.0 * 1.25 + $size / 2;
  $size = 20971520 if ($size < 20971520); # no less than 20MB
  my $human_size = bytes_to_human_size($size);
  $size = sprintf("%.0fk", $size / 1024.0);

  my $volume = NOTE::Volume->create(Filer => $filer_conn, type => "flex",
				    Aggregate => $aggr,	size => $size);
  NOTE::logcomment("Created $human_size volume called " . $volume->name);

  my $host = volume_to_hostspec($volume);
  dense_set_license(host => $host) or die "Could not license SIS\n";
  my $volpath = "/vol/" . $volume->name;
  dense_toggle(host => $host, path_name => $volpath, state => "on");

  if ($opts{corruption_test}) {
    # XXX Currently, this won't be undone by the teardown code.
    setflag($volume->Filer, "wafl_inconsistent_panic_testing", 1);
  }

  unless ($STUBBED) {
    my $bootarg = $volume->Filer->say("priv set -q test ; " .
			"bootargs get enable_constant_time_clone");
    unless ($bootarg eq "enable_constant_time_clone == true") {
      $Log->warn("\"enable_constant_time_clone\" bootarg not enabled");
    }
  }

  return $volume;
}

## @name ::sd_volume_init
## @usage sd_volume_init( %params, %options )
## @synopsis
##    Creates a flexible volume in the aggregate specified by the parameters,
##    and returns the NOTE::Volume object.
##    Sizes the volume according to the clone file size and the number of
##    clones the script intends to create (which defaults to 1).
##    Enables sis/dense on the volume.
## @args num_clones Optional: number of clones to be created (default=1)
## @args corruption_test Optional: prepare volume for corruption-related
##                                 testing (default=0)
sub sd_volume_init( \%;% ) {
  my ($params, %opts) = @_;
  if (exists $params->{FILER_AGGR}) {
    $opts{filer_aggr} = $params->{FILER_AGGR};
  }
  my $volume = volume_init($params->{FILER},
			   $params->{CLONE_FILE_SIZE},
			   %opts);
  unless (exists $params->{FILER_AGGR}) {
    $params->{FILER_AGGR} = $volume->Aggregate->name;
  }

  register_fiji();

  return $volume;
}

## @name ::qsm_qtree_init
## @usage qsm_qtree_init( %params, %options )
## @synopsis
##    Creates the source and destination flexible volumes in the aggregates
##    on the corresponding filers specified by the parameters. 
##    Creates the source qtree on the source volume, and instantiates a
##    qtree object on the destination volume without actually creating it.
##    Returns the source and destination qtree objects (NOTE::Qtree)
##    Sizes the volume according to the clone file size and the number of
##    clones the script intends to create (which defaults to 1).
##    Enables sis/dense on the volume.
## @args num_clones Optional: number of clones to be created (default=1)
## @args corruption_test Optional: prepare volume for corruption-related
##                                 testing (default=0)
sub qsm_qtree_init( \%;% ) {
  my ($params, %opts) = @_;

  # Setup the source volume and qtree.
  my $src_volume = volume_init($params->{FILER_SRC},
				$params->{CLONE_FILE_SIZE},
				filer_aggr => $params->{FILER_SRC_AGGR},
				%opts);
  unless (exists $params->{FILER_SRC_AGGR}) {
    $params->{FILER_SRC_AGGR} = $src_volume->Aggregate->name;
  }
  my $dst_volume = volume_init($params->{FILER_DST},
			       $params->{CLONE_FILE_SIZE},
			       filer_aggr => $params->{FILER_DST_AGGR},
			       %opts);
  unless (exists $params->{FILER_DST_AGGR}) {
    $params->{FILER_DST_AGGR} = $dst_volume->Aggregate->name;
  }

  # Create a qtree on the source volume
  my $src_qtree = NOTE::Qtree->create(Volume => $src_volume);

  # Create a qtree object without actually creating on the destination
  my $dst_qtree = NOTE::Qtree->new(Volume => $dst_volume);

  $src_volume->Filer->license('snapmirror');
  $dst_volume->Filer->license('snapmirror');

  register_fiji();

  return ($src_qtree, $dst_qtree);
}

## @name ::modblock
## @usage modblock( $client, $filepath, @ranges )
## @synopsis
##    Modifies a set of FBN ranges in the named file on the given client.
##    Writes a random block of data to each block in each FBN range.
##    FBN ranges can either be a single FBN, a range (e.g. 0-200), or
##    a range with a stride (e.g. 1000-1500/3).
sub modblock( $$;@ ) {
  my ($CLIENT, $file, @ranges) = @_;
  my $fbns = join(",", @ranges);

  die "No FBN ranges specified.\n" unless (defined $fbns);

  my $code = <<'CODE';
  {
    use Fcntl qw(:DEFAULT :mode :seek);

    my ($file, $fbns) = @_;
    my $BLOCK_SIZE = 4096;

    my $fh;
    sysopen($fh, $file, O_WRONLY|O_CREAT) or die "open($file): $!\n";
    binmode $fh;

    my $value = int(rand(4294967296));
    my $block = pack("L", $value) x 1024;

    foreach my $range (split(/,/, $fbns)) {
      die "invalid fbn range \"$range\"\n"
	unless ($range =~ /\A(\d+)(?:\-(\d+)(?:\/(\d+))?)?\z/);
      my ($start, $end, $stride) = ($1, (defined $2 ? $2 : $1), (defined $3 ? $3 : 1));
      die "negative fbn range \"$range\"\n"
	if ($end && $end < $start);
      die "invalid stride value \"$stride\"\n"
	unless ($stride > 0);

      sysseek($fh, ($BLOCK_SIZE * $start), SEEK_SET) or die "seek: $!\n";
      for (my $fbn = $start; $fbn <= $end; $fbn += $stride) {
	if ($stride > 1) {
	  sysseek($fh, ($BLOCK_SIZE * $fbn), SEEK_SET) or die "seek: $!\n";
	}
	syswrite($fh, $block) or die "write: $!\n";
      }
    }

    close($fh) or die "close: $!\n";

    logresult("PASS");
  }
CODE
  strip_whitespace($code);

  my $obj = subtest($code,
		    '-nobg',
		    '-client' => $CLIENT,
		    '-onexit' => sub {},
		    "--", $file, $fbns);
  my $result = NATE::Result::worst( $obj->results() );
  if (! defined $result || $result->severity() > 0) {
    die "modblock failure: " . join("", $result->messages()) . "\n";
  }

  return 1;
}

## @name ::clone_create
## @usage clone_create( $volume, $source_path, $clone_path, %options )
## @synopsis
##    Creates a SIS clone of a source file.
##    The paths specified should be relative to the root of the volume.
## @args snapname Optional: create a snapshot-backed clone from the
##                          source path in the named snapshot
## @args source_range Optional: reference to a list of ranges to create
##                              sub-file clones from
## @args clone_range Optional: reference to a list of ranges in the clone file
##                             to place sub-file clones at (rather than at the
##                             same ranges as in the source file)
sub clone_create( $$$;% ) {
  my @pos = splice(@_, 0, 3);
  validate_pos(@pos, { isa => 'NOTE::Volume' },
	             { type => SCALAR }, { type => SCALAR });
  my ($volume, $source_path, $clone_path) = @pos;
  my %opts = validate(@_ => {
			snapname => { type => SCALAR, optional => 1 },
			source_range => {
				type => ARRAYREF, optional => 1,
				callbacks => {
					'valid range' =>
						sub { validate_ranges($_[0]); }
					     }
			},
			clone_range => {
				type => ARRAYREF, optional => 1,
				depends => [ 'source_range' ],
				callbacks => {
					'valid range' =>
						sub { validate_ranges($_[0]); }
					     }
			},
                     });

  my $sub_file = (exists $opts{source_range} || exists $opts{clone_range});

  if ($source_path eq $clone_path) {
    die "cannot clone file over itself\n"
      unless (exists $opts{source_range} && exists $opts{clone_range});
  }

  if (exists $opts{source_range} && exists $opts{clone_range}) {
    die "source and clone range lists must have the same number of ranges\n"
      unless (scalar(@{$opts{source_range}}) == scalar(@{$opts{clone_range}}));
    for (my $i = 0; $i < scalar(@{$opts{source_range}}); $i++) {
      die "Params::Validate missed invalid range\n"
	unless ($opts{source_range}->[$i] =~ /\A(\d+)(?:\-(\d+))?\z/);
      my ($source_start, $source_end) = ($1, $2);
      die "Params::Validate missed invalid range\n"
	unless ($opts{clone_range}->[$i] =~ /\A(\d+)(?:\-(\d+))?\z/);
      my ($clone_start, $clone_end) = ($1, $2);
      if (defined $source_end && defined $clone_end) {
	die sprintf("range lengths mismatch: %d-%d vs. %d-%d\n",
		    $source_start, $source_end, $clone_start, $clone_end)
	  unless (($source_end - $source_start) == ($clone_end - $clone_start));
      } else {
	die "singleton range vs. longer range\n"
	  if (defined $source_end || defined $clone_end);
      }
    }
  }

  my $source = sprintf("/vol/%s/%s", $volume->name, $source_path);
  my $clone = sprintf("/vol/%s/%s", $volume->name, $clone_path);

  my $cmd = "clone start $source";
  $cmd .= " $clone" if ($source_path ne $clone_path);
  $cmd .= " -s " . $opts{snapname} if (exists $opts{snapname});

  my $range_cnt = exists $opts{source_range} ?
			scalar(@{$opts{source_range}}) : 0;
  for (my $i = 0; $i < $range_cnt; $i++) {
    die "Params::Validate missed invalid range\n"
      unless ($opts{source_range}->[$i] =~ /\A(\d+)(?:\-(\d+))?\z/);
    my $src_blk = $1;
    my $blk_cnt = defined $2 ? $2 - $1 + 1 : 1;
    my $dst_blk = $src_blk;
    if (exists $opts{clone_range}) {
      die "Params::Validate missed invalid range\n"
	unless ($opts{clone_range}->[$i] =~ /\A(\d+)(?:\-(\d+))?\z/);
      $dst_blk = $1;
      die "mismatched range lengths??\n"
	if ((defined $2 && (($2 - $1 + 1) != $blk_cnt)) ||
	    (! defined $2 && $blk_cnt != 1));
    }
    $cmd .= " -r $src_blk:$dst_blk:$blk_cnt";
  }

  # We can't "$volume->Filer->say($cmd)" here because
  # NOTE::Filer::say() doesn't provide a way to pass a longer timeout.
  my $text = SayWrap::sayEOF( $volume->Filer->conn, $cmd,
			      timeout => 150 ); # 2.5 minutes
  unless ($text =~ /\AClone operation started successfully\. ID: \d+\.\z/) {
    die "clone start failed: $text\n";
  }

  if ($STUBBED) {
    # XXX Note that the clone operation "started", but may not have finished
    # yet.  Really, we ought to poll 'clone status' here until it's done.

    # Create a "parent" file alongside the clone.
    # Create a separate "parent" file for each sub-file clone range.
    my $source_fid = filer_stat($volume, $source_path);
    my $clone_fid = filer_stat($volume, $clone_path);
    unless (exists $opts{source_range}) {
      $opts{source_range} = [ 0 ];
    }
    for (my $i = 0; $i < scalar(@{$opts{source_range}}); $i++) {
      my $srange = $opts{source_range}->[$i];
      my $crange = exists $opts{clone_range} ?
			$opts{clone_range}->[$i] : $srange;
      my $sparent = sprintf("/vol/%s/%s_%s",
			    $volume->name, $source_path, $srange);
      my $cparent = sprintf("/vol/%s/%s_%s",
			    $volume->name, $clone_path, $crange);
      wrfile($volume->Filer->conn, $sparent,
	     "parent for clone fid=$source_fid : $source_path\n")
	or die "error creating $sparent: $@\n";
      if ($source_path ne $clone_path) {
	wrfile($volume->Filer->conn, $cparent,
	       "parent for clone fid=$clone_fid : $clone_path\n")
	  or die "error creating $cparent: $@\n";
      }
    }
  }

  return 1;
}

## @name ::clone_list
## @usage clone_list( $volume, $clone_path, %options )
## @synopsis
##    Looks up a given clone file in the clone metafile.
##    Returns a tuple of the clone inode number and the inode number of
##    the clone's parent inode (in the private space).
##    If the file is no longer a clone, then an empty list is returned.
## @args snapname Optional: snapshot in which to do clone metafile lookup
## @args range Optional: FBN range to lookup (default=0)
sub clone_list( $$;% ) {
  my @pos = splice(@_, 0, 2);
  validate_pos(@pos, { isa => 'NOTE::Volume' }, { type => SCALAR } );
  my ($volume, $path) = @pos;
  my %opts = validate(@_ => {
			snapname => { type => SCALAR, optional => 1 },
			range => {
				type => SCALAR, optional => 1,
				callbacks => {
					'valid range' =>
						sub { validate_ranges($_[0]); }
					     }
			},
                     });

  if ($STUBBED) {
    # Get the inode numbers of the clone and the "parent".
    my %stat_opts = ();
    if (exists $opts{snapname}) {
      $stat_opts{snapname} = $opts{snapname};
    }
    my $clone_fid = filer_stat($volume, $path, %stat_opts);

    $stat_opts{enoent_ok} = 1;
    my $parent_path = $path . "_" . $opts{range};
    my $parent_fid = filer_stat($volume, $parent_path, %stat_opts);
    unless (defined $parent_fid) {
      return (); # no parent found, return as if the split "finished"
    }

    # Now we have the inode numbers for the clone and parent.
    # If the file we created when we "disabled" clone splitting exists,
    # either for the volume or for this clone inode, then report the
    # clone split as still "in progress" -- otherwise, say it's done.
    my $disable_vol_fid = filer_stat($volume,
				     $CLONE_SPLIT_DISABLER,
				     %stat_opts);
    my $disable_ino_fid = filer_stat($volume,
				     $CLONE_SPLIT_DISABLER . "_" . $clone_fid,
				     %stat_opts);
    if (defined $disable_vol_fid || $disable_ino_fid) {
      return ($clone_fid, $parent_fid);
    } else {
      return ();
    }
  }

  my $volpath;
  if (exists $opts{snapname}) {
    $volpath = sprintf("/vol/%s/.snapshot/%s/%s",
		       $volume->name, $opts{snapname}, $path);
  } else {
    $volpath = sprintf("/vol/%s/%s", $volume->name, $path);
  }

  my $range_arg = "";
  my $fbn = undef;
  my $count = undef;
  if (exists $opts{range}) {
    die "Params::Validate missed invalid range\n"
      unless ($opts{range} =~ /\A(\d+)(?:\-(\d+))?\z/);
    $fbn = $1;
    $count = defined $2 ? $2 - $1 + 1 : 1;
    $range_arg = " -r $fbn:$count";
  }

  my $cmd = "priv set -q test ; clone list -p ${volpath}${range_arg}";
  my $text = $volume->Filer->say($cmd);

  if ($text eq 'No entries found!') {
    return ();
  }
  unless ($text =~ /\AClone-inum = (\d+) \((Public|Private)\), start-fbn = (\d+) blk-count = (\d+)\s*?\nParent inum = (\d+) \((Public|Private)\), start-fbn = (\d+)\s*\z/) {
    die "error parsing 'clone list' output: $text\n";
  }
  my ($clone_inum, $clone_priv, $start_fbn, $blk_count) = ($1, $2, $3, $4);
  my ($parent_inum, $parent_priv, $parent_start_fbn) = ($5, $6, $7);

  if (defined $fbn) {
    unless ($start_fbn <= $fbn &&
	    $start_fbn + $blk_count >= $fbn + $count) {
      die "returned clone range does not cover requested range\n";
    }
  }

  unless ($clone_priv =~ /\APublic\z/i && $parent_priv =~ /\APrivate\z/i) {
    die "clone_list() only understands public clones backed by private parents\n";
  }

  # XXX  Here, it could be the case that $start_fbn != $parent_start_fbn,
  # or that the clone is not a public inode, or the parent is not private.
  # However, the interface to this library routine, as presently defined,
  # has no way to return that information.

  return ($clone_inum, $parent_inum);
}

## @name ::clone_split_wait
## @usage clone_split_wait( $volume, $clone_path, %options )
## @synopsis
##    Waits for a SIS clone to be split by polling the output
##    of 'clone list'.
## @args poll_interval Optional: number of seconds between polls (default=5)
## @args timeout Optional: maximum number of seconds to wait (default=3600)
sub clone_split_wait( $$;% ) {
  my @pos = splice(@_, 0, 2);
  validate_pos(@pos, { isa => 'NOTE::Volume' }, { type => SCALAR } );
  my ($volume, $path) = @pos;
  my %opts = validate(@_ => {
			poll_interval => {
				type => SCALAR, regex => qr/^\d+$/,
				optional => 1, default => 5
			},
			timeout => {
				type => SCALAR, regex => qr/^\d+$/,
				optional => 1, default => 3600
			},
                     });

  my $start_time = time;

  while (1) {

    my @fids = clone_list($volume, $path);
    last if (scalar(@fids) == 0);

    my $elapsed = time - $start_time;
    if ($elapsed >= $opts{timeout}) {
      die "gave up waiting for clone split of $path after $elapsed seconds\n";
    }

    snooze($opts{poll_interval});

  }

  NOTE::logcomment("Finished waiting for /vol/" . $volume->name .
		   "/$path to split");
  return 1;
}

# We set fiji points which repeat up to certain (high) count.
my $FIJI_REPEAT = 2147483647;

## @name ::clone_splitting_pause_fiji
## @usage clone_splitting_pause_fiji( $volume, %options )
## @synopsis
##    Sets a FIJI point which will prevent all clone splitting on
##    the given volume from proceeding, or optionally just the
##    clone splitting on a specific file.
##    Returns the NOTE::Fiji object representing the set FIJI.
## @args path Optional: only pause splitting on this specified clone
##                      (relative to the root of the volume)
## @args fbn Optional: only pause splitting on a specific file
##                     past this given FBN (default=0)
sub clone_splitting_pause_fiji( $;% ) {
  my @pos = splice(@_, 0, 1);
  validate_pos(@pos, { isa => 'NOTE::Volume' });
  my ($volume) = @pos;
  my %opts = validate(@_ => {
			path => { type => SCALAR, optional => 1 },
			fbn => {
				type => SCALAR, regex => qr/^\d+$/,
				optional => 1, default => 0,
				depends => [ 'path' ]
			},
                     });

  my $fiji = undef;

  my $fsid = $volume->fsid;

  if ($STUBBED) {
    my $fileid = 64; # root directory
    if (exists $opts{path}) {
      $fileid = filer_stat($volume, $opts{path});
    }
    $fiji = NOTE::Fiji->new(Filer => $volume->Filer,
			    tag => 'delay WAFL messages for a specific inode',
			    repeat => $FIJI_REPEAT,
			    volume => $volume->name,
			    fileid => $fileid);
    $fiji->set;

    # Stat the path to cause the fiji to trigger once.
    filer_stat($volume, (exists $opts{path} ? $opts{path} : "/"));

    # Create a stub file to indicate that "splitting" is disabled.
    my $disabler_file = sprintf("/vol/%s/%s",
				$volume->name, $CLONE_SPLIT_DISABLER);
    $disabler_file .= "_$fileid" if (exists $opts{path});
    wrfile($volume->Filer->conn, $disabler_file,
	   "disable clone splitting on the " .
	   (exists $opts{path} ? "file" : "volume"))
      or die "error creating $disabler_file: $@\n";
    $fiji->{disabler_file} = $disabler_file;

    return $fiji;
  }

  if (exists $opts{path}) {
    my $fileid = filer_stat($volume, $opts{path});
    $fiji = NOTE::Fiji->new(Filer => $volume->Filer,
			    tag => 'pause clone splitting on an inode',
			    repeat => $FIJI_REPEAT,
			    fsid => $fsid,
			    inum => $fileid,
			    fbn => $opts{fbn});
  } else {
    $fiji = NOTE::Fiji->new(Filer => $volume->Filer,
			    tag => 'pause clone splitting on a volume',
			    repeat => $FIJI_REPEAT,
			    fsid => $fsid);
  }

  $fiji->set;
  return $fiji;
}

## @name ::fiji_repeat_count
## @usage fiji_repeat_count( $fiji )
## @synopsis
##    Given a FIJI point (a NOTE::Fiji object) which was set with a
##    repeat count, retrieves the current number of times the fiji
##    was triggered.
##    In a scalar context, returns the number of triggers.
##    In a list context, returns the number of triggers and repeat count.
sub fiji_repeat_count( $ ) {
  validate_pos(@_, { isa => 'NOTE::Fiji' });
  my ($fiji) = @_;

  my $slot = $fiji->slot;
  die "no such fiji set\n" unless (defined $slot);
  my $cmd = "priv set -q diag; fiji dump $slot";

  my ($current_slot, $count, $max_count);
  my $exhausted = 1;

  my $text = $fiji->{_Filer}->say($cmd);
  foreach my $line ( split(/\n/, $text) ) {
    $current_slot = $1 if ( $line =~ /^(\d+):/ );
    next unless (defined $current_slot && $current_slot == $slot);
    if ($line =~ /\A\s+\[\d+\] ITERATOR, id \d+, name repeat, maxCount (\d+), count (\d+)\s*\z/) {
      if (defined $max_count || defined $count) {
	die "multiple iterators in set fiji\n";
      }
      ($max_count, $count) = ($1, $2);
    }
    $exhausted = 0 if ($line =~ /\A\-+\>\s+/);
  }

  unless (defined $current_slot && defined $count && defined $max_count) {
    die "error parsing 'fiji dump' output: no ITERATOR/repeat found\n";
  }

  if ($exhausted) {
    die "fiji exhausted before maxCount $max_count > $count\n"
      if ($count < $max_count);
  } else {
    die "fiji not exhausted after maxCount $max_count <= $count\n"
      if ($count >= $max_count);
  }

  return wantarray ? ($count, $max_count) : $count;
}

## @name ::fiji_wait_for_trigger
## @usage fiji_wait_for_trigger( $fiji )
## @synopsis
##    Waits for the given fiji point to trigger for the first time.
##    This is what NOTE::Fiji::wait should do -- see burt 412519.
##    (This routine will work for fiji points with repeat counts set.)
## @args poll_interval Optional: number of seconds between polls (default=5)
## @args timeout Optional: maximum number of seconds to wait (default=3600)
sub fiji_wait_for_trigger( $;% ) {
  my @pos = splice(@_, 0, 1);
  validate_pos(@pos, { isa => 'NOTE::Fiji' });
  my ($fiji) = @pos;
  my %opts = validate(@_ => {
			poll_interval => {
				type => SCALAR, regex => qr/^\d+$/,
				optional => 1, default => 5
			},
			timeout => {
				type => SCALAR, regex => qr/^\d+$/,
				optional => 1, default => 3600
			},
                     });

  my $start_time = time;

  while (1) {

    my $count = fiji_repeat_count($fiji);
    last if ($count > 0);

    my $elapsed = time - $start_time;
    if ($elapsed >= $opts{timeout}) {
      die "gave up waiting for fiji trigger after $elapsed seconds\n";
    }

    snooze($opts{poll_interval});

  }
}

## @name ::clone_splitting_clear_fiji
## @usage clone_splitting_clear_fiji( $fiji )
## @synopsis
##    Verifies that repeat count on the given FIJI point (a NOTE::Fiji object)
##    has not been exhausted, and then clears the FIJI.
sub clone_splitting_clear_fiji( $ ) {
  validate_pos(@_, { isa => 'NOTE::Fiji' });
  my ($fiji) = @_;

  my ($count, $max_count) = fiji_repeat_count($fiji);
  unless ($count < $max_count) {
    die "fiji repeat count exhausted before clearing\n";
  }

  if ($STUBBED) {
    # Remove the disabler file.
    die "no disabler file name passed via the NOTE::Fiji object\n"
      unless ($fiji->{disabler_file});

    my $cmd = "priv set -q advanced ; rm " . $fiji->{disabler_file};
    my $text = $fiji->{_Filer}->say($cmd);
    die "rm " . $fiji->{disabler_file} . " failed: $text\n"
      if ($text ne "");
  }

  $fiji->clear;
}

## @name ::dumpblock_info
## @usage dumpblock_info( $volume, $inode_number, $block_number, %options )
## @synopsis
##    Retrieves the information about a block in a file.
##    Returns the block information as a hash reference.
##    If the block number is a hole, returns undef.
##    If the block number is past the end of the file,
##    returns the block number of the last block in the file.
##    If the file fits in the inode itself, returns -1.
## @args private Optional: if set, corrupt the inode in the private space
##                         instead of the public space
## @pargs level Optional: level of the block to lookup (default=0)
## @pargs snapid Optional: snapshot ID of the snapshot in which to lookup
##                         the block
sub dumpblock_info( $$$;% ) {
  my @pos = splice(@_, 0, 3);
  validate_pos(@pos, { isa => 'NOTE::Volume' },
	             { type => SCALAR, regex => qr/^\d+$/ },
	             { type => SCALAR, regex => qr/^\d+$/ } );
  my ($volume, $inode, $blknum) = @pos;
  my %opts = validate(@_ => {
			private => { type => SCALAR, optional => 1 },
			level => { type => SCALAR, regex => qr/^\d+$/,
				   optional => 1, default => 0 },
			snapid => { type => SCALAR, regex => qr/^\d+$/,
				    optional => 1, default => 0 },
                     });

  my $cmd = sprintf("priv set -q diag ; dumpblock info %s %d %s%d",
		    $volume->name, $inode,
		    $opts{private} ? "-private " : "", $blknum);
  if ($opts{level} || $opts{snapid}) {
    $cmd .= " " . $opts{level};
    $cmd .= " " . $opts{snapid} if ($opts{snapid});
  }

  my $text = $volume->Filer->say($cmd);
  if ($text eq "block not found -- block is a hole") {
    return undef;
  }
  if ($text =~ /\Ablock not found -- inode has only (\d+) levels\z/) {
    my $max_level = $1;
    if ($max_level == 0) {
      return -1;
    }
    die "inode $inode has only $max_level levels, but requested level " .
      $opts{level} . "\n";
  }
  if ($text =~ /\Ablock not found -- inode has only (\d+) blocks\z/) {
    my $last_block = $1;
    return $last_block;
  }

  my %info = ();

  die "error parsing output from 'dumpblock info': $text\n"
    unless ($text =~ s/\Aok\s*?\n// &&
	    $text =~ s/\Adisk\s+(\S+)\s*?\n//);
  $info{disk} = $1;

  while ($text =~ s/\A([\w\-]+)\s+((0x)[\dA-Fa-f]+|\d+)\s*//) {
    my $field = $1;
    my $value = (defined $3 ? hex($2) : $2);
    $info{$field} = $value;
  }
  die "unexpected extra text from 'dumpblock info': $text\n"
    unless ($text eq "");

  return \%info;
}

## @name ::wafl_showvbns
## @usage wafl_showvbns( $volume, $file_path, %options )
## @synopsis
##    Obtains the list of VVBNs for all the blocks of a file in the
##    given volume.  Returns a reference to a list of change records,
##    each of which is a hash reference containing three keys:
##    fbn, vbn, count.
## @args snapname Optional: snapshot to obtain the block list from
## @args ranges Optional: reference to a list of FBN ranges, requesting only
##                        the specified FBN ranges to be in the returned records
sub wafl_showvbns( $$;% ) {
  my @pos = splice(@_, 0, 2);
  validate_pos(@pos, { isa => 'NOTE::Volume' }, { type => SCALAR });
  my ($volume, $path) = @pos;
  my %opts = validate(@_ => {
			snapname => { type => SCALAR, optional => 1 },
			ranges => {
				type => ARRAYREF, optional => 1,
				callbacks => {
					'valid range' =>
						sub { validate_ranges($_[0]); }
					     }
			},
                     });

  my $volpath = sprintf("/vol/%s/%s", $volume->name, $path);
  if (exists $opts{snapname}) {
    $volpath = sprintf("/vol/%s/.snapshot/%s/%s",
		       $volume->name, $opts{snapname}, $path);
  }

  my $range_check = undef;
  if (exists $opts{ranges}) {
    unless (ref($opts{ranges}) eq "ARRAY") {
      die "invalid ranges parameter\n";
    }
    # Build a closure that checks whether an FBN is in one of the ranges.
    my @ranges = map {
      die "invalid range $_\n" unless ($_ =~ /\A(\d+)(?:\-(\d+))?\z/);
      die "invalid range $_\n" if (defined($2) && $1 >= $2);
      [ $1, defined($2) ? $2 : $1 ]
    } @{$opts{ranges}};
    $range_check = sub {
      foreach my $r (@ranges) {
	if ($r->[0] <= $_[0] && $_[0] <= $r->[1]) {
	  return 1;
	}
      }
      return 0;
    };
  }

  my $cmd = "priv set -q diag ; wafl showvbns $volpath";
  # We can't "$volume->Filer->say($cmd)" here because
  # NOTE::Filer::say() doesn't provide a way to pass a longer timeout,
  # and 'wafl showvbns' can actually take a while.
  my $text = SayWrap::sayEOF( $volume->Filer->conn, $cmd,
			      timeout => 900 ); # 15-minute timeout

  my @records = ();
  my $cur_fbn = undef;
  my $cur_vbn = undef;
  my $cur_cnt = 0;
  while ($text ne "" && $text =~ s/\A(.*?)(?:\n|\z)//) {
    my $line = $1;
    unless ($line =~ /\Afbn (\d+) is vbn (?:\d+), vvbn (\d+)\s*\z/) {
      die "error parsing 'wafl showvbns' output: $line\n";
    }
    my ($fbn, $vbn) = ($1, $2);

    if (defined $range_check) {
      next unless $range_check->($fbn);
    }

    # If we haven't started recording a range yet, this one does.
    unless (defined $cur_fbn) {
      $cur_fbn = $fbn;
      $cur_vbn = $vbn;
      $cur_cnt = 1;
      next;
    }

    # If this record extends the current range, append it.
    if ($fbn == $cur_fbn + $cur_cnt &&
	(($vbn == 0 && $cur_vbn == 0) || ($vbn == $cur_vbn + $cur_cnt))) {
      $cur_cnt++;
      next;
    }

    # Conclude the current range and start a new one.
    my %rec = ( fbn => $cur_fbn, vbn => $cur_vbn, count => $cur_cnt );
    push @records, \%rec;
    $cur_fbn = $fbn;
    $cur_vbn = $vbn;
    $cur_cnt = 1;
  }

  # Conclude the final range
  if (defined $cur_fbn) {
    my %rec = ( fbn => $cur_fbn, vbn => $cur_vbn, count => $cur_cnt );
    push @records, \%rec;
  }

  return \@records;
}

## @name ::wafl_touch
## @usage wafl_touch( $volume, $inode_number, @ranges, $level, %options )
## @synopsis
##    Dirties the buffer indicated by the provided FBN ranges and level
##    in the buffer tree of the given inode in a volume.
## @args private Optional: if set, touch the buftree in the private inode space
sub wafl_touch( $$\@$;% ) {
  my @pos = splice(@_, 0, 4);
  validate_pos(@pos, { isa => 'NOTE::Volume' },
	             { type => SCALAR, regex => qr/^\d+$/ },
	             { type => ARRAYREF, callbacks => {
				'valid range' =>
					sub { validate_ranges($_[0]); }
				}
		     },
	             { type => SCALAR, regex => qr/^\d+$/ } );
  my ($volume, $inode, $ranges, $level) = @pos;
  my %opts = validate(@_ => {
			private => { type => SCALAR, optional => 1 },
                     });

  my $cmd = "priv set -q test ; wafl touch " . $volume->name . " $inode";
  if ($opts{private}) {
    $cmd .= " -private";
  }
  $cmd .= " $level";

  foreach my $range (@$ranges) {

    die "invalid range $range\n" unless ($range =~ /\A(\d+)(?:\-(\d+))?\z/);
    my $start_fbn = $1;
    my $end_fbn = (defined $2 ? $2 : $1);

    my $text = $volume->Filer->say("$cmd $start_fbn $end_fbn");
    if ($text =~ /\Awafl touch succeeded on inode number (\d+) level (\d+) fbn (\d+)\s*\z/) {
      unless ($1 == $inode && $2 == $level && $3 == $start_fbn) {
	die "wafl touch output mismatch\n";
      }
    } elsif ($text =~ /wafl touch failed on inode number \d+ level \d+ fbn \d+ with error (\d+)\s*\z/) {
      die "wafl touch: errno=$1\n";
    } else {
      die "wafl touch error: $text\n";
    }

  }

  return 1;
}

## @name ::wafl_corrupt_block
## @usage wafl_corrupt_block( $volume, $inode_number, $fbn, $level, %options )
## @synopsis
##    Corrupts the block specified by FBN and level in the buffer tree
##    of the given inode, by marking it pseudo-bad.
## @args snapname Optional: snapshot in which to corrupt the block
## @args snapid Optional: snapid of snapshot in which to corrupt the block
##                        (cannot specify both snapname and snapid)
## @args private Optional: if set, corrupt the inode in the private space
##                         instead of the public space
## @args no_online Optional: if set, do not online the volume after corrupting
sub wafl_corrupt_block( $$$$;% ) {
  my @pos = splice(@_, 0, 4);
  validate_pos(@pos, { isa => 'NOTE::Volume' },
	             { type => SCALAR, regex => qr/^\d+$/ },
	             { type => SCALAR, regex => qr/^\d+$/ },
	             { type => SCALAR, regex => qr/^\d+$/ } );
  my ($volume, $inode, $fbn, $level) = @pos;
  my %opts = validate(@_ => {
			snapname => { type => SCALAR, optional => 1 },
			snapid => { type => SCALAR, regex => qr/^\d+$/,
				    optional => 1 },
			private => { type => SCALAR, optional => 1 },
			no_online => { type => SCALAR, optional => 1 },
                     });

  my $snapid = (exists $opts{snapid} ? $opts{snapid} : 0);
  if (exists $opts{snapname}) {
    if (exists $opts{snapid}) {
      die "cannot specifiy both snapname and snapid\n";
    }
    my $snap_status = $volume->snapshot_status();
    unless (exists $snap_status->{$opts{snapname}}) {
      die "no such snapshot \"" . $opts{snapname} . "\"\n";
    }
    $snapid = $snap_status->{$opts{snapname}}->{snapid};
    unless (defined $snapid && $snapid =~ /\A\d+\z/ &&
	    $snapid > 0 && $snapid < 256) {
      die "error obtaining snapid for snapshot \"" . $opts{snapname} . "\"\n";
    }
  }

  my $priv = $opts{private} ? "priv:" : "";
  my $inject_cmd = "pseudo_bad inum=${priv}${inode} level=${level} bno=${fbn} snapid=${snapid}";

  my $rc = wreck($volume->Filer->conn, $volume->name,
		 type => "flex",
		 pseudo_bad => 1,
		 inject_script => $inject_cmd);
  if ($rc != 1 || $@) {
    die "wreck error: $@\n";
  }

  # A successful wreck restricts the volume.  Put it back online.
  unless ($opts{no_online}) {
    $volume->online;
  }

  return 1;
}

## @name ::wafl_corrupt_file
## @usage wafl_corrupt_file( $volume, $inode_number, %options )
## @synopsis
##    Corrupts all the data blocks in a non-holey file,
##    specified by inode number, by marking them all pseudo-bad.
## @args snapname Optional: snapshot in which to corrupt the block
## @args private Optional: if set, corrupt the inode in the private space
##                         instead of the public space
sub wafl_corrupt_file( $$;% ) {
  my @pos = splice(@_, 0, 2);
  validate_pos(@pos, { isa => 'NOTE::Volume' },
	             { type => SCALAR, regex => qr/^\d+$/ } );
  my ($volume, $inode) = @pos;
  my %opts = validate(@_ => {
			snapname => { type => SCALAR, optional => 1 },
			private => { type => SCALAR, optional => 1 },
                     });

  my $snapid = 0;
  if (exists $opts{snapname}) {
    my $snap_status = $volume->snapshot_status();
    unless (exists $snap_status->{$opts{snapname}}) {
      die "no such snapshot \"" . $opts{snapname} . "\"\n";
    }
    $snapid = $snap_status->{$opts{snapname}}->{snapid};
    unless (defined $snapid && $snapid =~ /\A\d+\z/ &&
	    $snapid > 0 && $snapid < 256) {
      die "error obtaining snapid for snapshot \"" . $opts{snapname} . "\"\n";
    }
  }

  my %add_opts = ();
  $add_opts{snapid} = $snapid if ($snapid != 0);
  $add_opts{private} = 1 if ($opts{private});

  my $blknum = 0;
  my $last_block;
  my %holes = ();
  while (1) {
    my $ret = dumpblock_info($volume, $inode, $blknum, %add_opts);
    if (defined $ret) {
      if (ref($ret)) {
	# Block exists.
      } else {
	# $ret is the last block in the file
	die "file changed during operation?\n"
	  unless ($ret == ($blknum - 1));
	$last_block = $ret;
	last;
      }
    } else {
      # Block was a hole.
      $holes{$blknum} = 1;
    }
    $blknum++;
  }

  # Don't keep bringing the volume back online between each block corruption.
  $add_opts{no_online} = 1;

  for ($blknum = 0; $blknum <= $last_block; $blknum++) {
    next if ($holes{$blknum});
    wafl_corrupt_block($volume, $inode, $blknum, 0, %add_opts);
  }

  # A successful wreck restricts the volume.  Put it back online.
  $volume->online;

  return 1;
}

## @name ::wafl_punch_holes
## @usage wafl_punch_holes( $volume, $filepath, @ranges )
## @synopsis
##    Punches holes at specified ranges in the named file of the given volume.
sub wafl_punch_holes( $$\@ ) {
  validate_pos(@_, { isa => 'NOTE::Volume' },
	           { type => SCALAR },
	           { type => ARRAYREF, callbacks => {
				'valid range' =>
					sub { validate_ranges($_[0]); }
				}
		   });
  my ($volume, $filepath, $ranges) = @_;

  my $cmd = "priv set -q test ; file punchhole " .
            "/vol/" . $volume->name .  "/" . $filepath;

  foreach my $range (@$ranges) {
    die "invalid range $range\n" unless ($range =~ /\A(\d+)(?:\-(\d+))?\z/);
    my ($start_fbn, $end_fbn) = ($1, $2);
    $end_fbn = $start_fbn unless (defined $end_fbn);

    my $boffset = $start_fbn * 4096;
    my $nbytes = ( $end_fbn - $start_fbn + 1 ) * 4096;
    my $text = $volume->Filer->say("$cmd $boffset $nbytes");
    unless ($text eq "") {
    	die "file punchhole failed: $text";
    }
  }

  return 1;
}

# @name ::wafl_flush_cache
# @usage wafl_flush_cache($filer)
# @synopsis
#   Issues a 'wafl flush cache' on the FILER thereby flushing the wafl buffers
sub wafl_flush_cache( $ ) {
  validate_pos(@_, { isa => 'NOTE::Filer' });
  my ($filer) = @_;
  my $text = $filer->say('priv set -q diag; wafl flush cache');
  unless ($text eq "") {
    die "Failed to flush cache: $text\n";
  }

  return 1;
}

# @name ::wafl_susp
# @usage wafl_susp($filer, $option)
# @synopsis
#    Issues a wafl_susp $option to the Filer and logs the output.
sub wafl_susp( $$ ) {
  validate_pos(@_, { isa => 'NOTE::Filer' }, { type => SCALAR });
  my ($filer, $option) = @_;
  my $text = $filer->say("priv set -q diag; wafl_susp $option");

  return 1;
}

## @name ::reset_counters
## @usage reset_counters($filer)
## @synopsis
##   Flushes the buffer cache and resets the wafl_susp counters to zero
sub reset_counters ( $ ) {
  validate_pos(@_, { isa => 'NOTE::Filer' });
  my ($filer) = @_;

  # Flush the cache
  wafl_flush_cache($filer);
  # Zero the wafl_susp counters
  wafl_susp($filer, "-z");
}

## @name ::fetch_counters
## @usage fetch_counters($filer)
## @synopsis
##   Dumps the wafl_susp -r output to the log. For now, we will manually
##   look at the log to see how many buffers were loaded by WAFL_SNAP_DIFF
##   for a intialize/update.
##   XXX This needs to be changed to automatically parse the output of
##   'wafl_susp -r' to pick out the values for the numbers of buffers
##   loaded by WAFL_SNAP_DIFF, and return it to the caller.
sub fetch_counters ( $ ) {
  validate_pos(@_, { isa => 'NOTE::Filer' });
  my ($filer) = @_;

  wafl_susp($filer, "-w");
  wafl_susp($filer, "-r");
}

## @name ::verify_qsm_transfer_size
## @usage verify_qsm_transfer_size($transfer_size, $expected_size)
## @synopsis
##   Calculates the overhead multiplier for a QSM transfer. For very small
##   amount of data transfers(<300K), we use a fixed 20K as the overhead. 
##   For transfers greater than 300K we use 5% as the overhead. Uses the
##   overhead and expected transfer size(passed in as KB) to decide if the
##   actual transfer size(passed in as KB) was reasonable, and returns boolean
##   to indicate the same.
sub verify_qsm_transfer_size($$) {
  my ($transfer_size, $expected_size) = @_;
  my $overhead_multiplier;
  
  if ($expected_size <= 300) {
    # For very small transfer sizes, using a percentage of the transfer
    # size (which is a very small amount again) underestimates the overhead
    # of the transfer. We need atleast 3 4K headers(one for the master, one for 
    # the clone, and one for the root dir) in most of our tests. Rounding 
    # this up to 20K.
    $overhead_multiplier = 1 + (20/$expected_size);
  } else {
    $overhead_multiplier = 1.05; #5% overhead for headers
  }

  if (($transfer_size >= $expected_size) &&
	  ($transfer_size < $expected_size * $overhead_multiplier)) {
    return 1;
  } else {
    return 0;
  }
}

## @name ::verify_mirror
## @usage verify_mirror($client, $src_mount_path, $dst_mount_path)
## @synopsis
##   Compares the contents of snapmirror source and destination objects 
##   (volume/qtree) mounted at $src_mount_path and $dst_mount_path on 
##   the $client. If they do not match, logs the details of the discrepancy.
sub verify_mirror ($$$) {
  my ($client, $src_path, $dst_path) = @_;
  my ($in1, $in2, $diff) = DataManip::compare_trees($client,
						    $src_path,
						    $dst_path,
						    ATTRIBUTES => 1);
  if ($@) {
    die "compare_trees failed: $@\n";
  }

  if (scalar(@$in1) || scalar(@$in2) || scalar(@$diff)) {
    if (scalar(@$in1)) {
      NOTE::logcomment("Paths on source but not destination:\n" .
		       join("\n", @$in1) . "\n");
    }
    if (scalar(@$in2)) {
      NOTE::logcomment("Paths on destination but not source:\n" .
		       join("\n", @$in2) . "\n");
    }
    if (scalar(@$diff)) {
      NOTE::logcomment("Paths differing between source and destination:\n" .
		       join("\n", @$diff) . "\n");
    }

    die "source and destination do not match\n";
  }

  return 1;
}

# XXX Still need more routines for:
#    - wafl susp -w/-z + parsing output to get block loads
#    - anything we need for compression...

1;
