=pod

=head1 SKTrace functions

=begin html

<title>SKTrace functions</title>

=end html

This module provides functions for dealing with SKTrace files on a filer.
It can cause a dump, look for dump files, remove dump files, reset the
in-memory traces, and change the trace levels for a given module.

There is currently no way to process the log files and look for specific
messages.  That must be done manually.

=cut

use strict;
use Server;
use Options;
use Tharn qw($Log);

package SKTrace;


=head2 new

Create a new SKTrace object for the given server.

Example:

 my $Server = new Server();
 $Server->set_execution_server(-Server => "node1");
 my $trace = new SKTrace($Server);

=cut

sub new {
  my $class = shift;
  my $server = shift;
  my $self = { SERVER => $server };

  bless ($self, $class);

  return $self;
}


=head2 enable_trace

Change the trace level for a given module, setting it to the given level.

Example:

 $trace->enable_trace("REPL", "6");

=cut

sub enable_trace {
  my $self   = shift;
  my $module = shift;
  my $level  = shift;
  my $server = $self->{SERVER};

  $server->shell(cmd => qq(sudo sysctl sysvar.sktrace.${module}_enable=$level));
}


=head2 clear_membuf

Clear the in-memory buffers of SKTrace events.

Example:

 $trace->clear_membuf();

=cut

sub clear_membuf {
  my $self   = shift;
  my $server = $self->{SERVER};

  $server->shell(cmd => "sudo sysctl sysvar.sktrace.clear=1");
}


=head2 dump

Cause the in-memory SKTrace buffers to be dumped to disk.  Files are created
using the filename base stored in the sysctl sysvar.sktrace.filename.
Returns an array of the full filenames (including the path) that were dumped.

Example:

 my @files = $trace->dump();

=cut

sub dump {
  my $self   = shift;
  my $server = $self->{SERVER};

  my $filebase = $server->shell(cmd => "sudo sysctl -n sysvar.sktrace.filename");

  $server->shell(cmd => "sudo sysctl sysvar.sktrace.dump=1");

  my $dumpfiles = $server->shell(cmd => qq(ls -1 $filebase*));

  if ($dumpfiles =~ /No such file/) {
    $dumpfiles = "";
  }

  my @files = split('\n', $dumpfiles);

  return @files;
}


=head2 get_dump_files

Get the list of all dump files on the server.  This will include any files
dumped by code and any dumped "manually" via the dump() method.  Returns
an array of the full filenames for any dump files found.

Example:

 my @files = $trace->get_dump_files();

=cut

sub get_dump_files {
  my $self   = shift;
  my $server = $self->{SERVER};

  my $filebase = $server->shell(cmd => "sudo sysctl -n sysvar.sktrace.filename");

  my $dumpfiles = $server->shell(cmd => qq(ls -1 $filebase*));

  if ($dumpfiles =~ /No such file/) {
    $dumpfiles = "";
  }

  return split('\n', $dumpfiles);
}


=head2 remove_dump_files

Remove all files matching the pathname and filename prefix where dump files
would be generated.  That is files matching the sysctl sysvar.sktrace.filename
as a prefix.

Example:

 $trace->remove_dump_files();

=cut

sub remove_dump_files {
  my $self   = shift;
  my $server = $self->{SERVER};

  my $filebase = $server->shell(cmd => "sudo sysctl -n sysvar.sktrace.filename");

  $server->shell(cmd => qq(rm $filebase*));
}


=head2 save_dump_files

Copy all dump files to a new location so they won't be destroyed by
subsequent calls to remove_dump_files.  This allows testers to examine
the dump files after several tests have run.  This function returns the
files that were saved in their new location.

Example:

 my @savedfiles = $trace->save_dump_files();

=cut

sub save_dump_files {
  my $self = shift;
  my $server = $self->{SERVER};

  my @files = $self->get_dump_files();

  my $filebase = $server->shell(cmd => "sudo sysctl -n sysvar.sktrace.filename");
  my ($dirbase, undef) = ($filebase =~ m/(.*\/)(.*)/);
  my $timestamp = `date +%Y%m%d%H%M%S`;
  chop $timestamp;

  my $savepath = qq($dirbase$timestamp);

  $server->shell(cmd => qq(mkdir $savepath));
  $server->shell(cmd => qq(cp @files $savepath));

  my $savedfiles = $server->shell(cmd => qq(ls -1 $savepath/*));
  if ($savedfiles =~ /No such file/) {
    $savedfiles = "";
  }

  return split('\n', $savedfiles);
}

sub set_use_sktlogd
{
  my $self   = shift;
  my $enable = shift;
  my $server = $self->{SERVER};

  $enable = $enable ? 1 : 0;

  $server->shell(cmd => "sudo sysctl sysvar.sktrace.mlog=$enable");
}


1;
