=pod

=head1 Wex functions

=begin html

<title>Wex functions</title>

=end html

This module provides functions for interacting with wex (or "volume explore")
on a node.

=cut

use strict;
use Server;
use Options;
use Tharn qw($Log);
use dp::snapmirror::paloma::unit_test::WexUtils;

package Wex;


=head2 new($server)

Create a new Wex object for the given server.

Example:

 my $Server = new Server();
 $Server->set_execution_server(-Server => "node1");
 my $wex = new Wex($Server);

=cut

sub new {
  my $class = shift;
  my $server = shift;
  my $self = { SERVER => $server, CMD => "/usr/sbin/wex" };

  bless ($self, $class);

  return $self;
}


=head2 get_dir_data(volume => "vol", inode => "85p")

Get directory data for the given volume and inode number.

=cut

sub get_dir_data {
  my ($self, %args) = @_;

  my $server = $self->{SERVER};

  my $output = $server->shell(cmd => "$self->{CMD} dir $args{volume}.$args{inode} -dump all");
  if ($output =~ /unable to find/) {
    die("Wex error: $output");
  }

  return split_wex_output($output);
}


=head2 get_aggr_dir_data(aggr => "aggr", voluuid => "0123-000...", metadir => "repl_ckpt")

Get directory data for the given volume directory within the aggregate.

=cut

sub get_aggr_dir_data {
  my ($self, %args) = @_;

  my $server = $self->{SERVER};

  my $output = $server->shell(cmd => "$self->{CMD} dir $args{aggr}./metadir/vvols/$args{voluuid}/$args{metadir} -dump all");
  if ($output =~ /unable to find/) {
    die("Wex error: $output");
  }

  return split_wex_output($output);
}


=head2 split_wex_output(string)

Split the "-dump all" wex output into a map to make accessing values easier.

=cut

sub split_wex_output {
  my ($output) = @_;

  my @arr = split("\n", $output);

  my %values;

  my %current;

  my $record_delim = "";

  foreach my $line (@arr) {
    my ($key, $value) = split("=", $line);

    # If there was no delimiter, just skip the line.
    if ($key eq $line) {
      next;
    }

    # If there's no start token, store the first one we find.
    if ($record_delim eq "") {
      $record_delim = $key;
    }

    # Found the start of a new set of records, so save the current ones,
    # if there are any.
    if (($record_delim eq $key) && (keys(%current))) {
      my %copy = %current;
      $values{$current{name}} = \%copy;
      %current = {};
    }

    $current{$key} = $value;
  }

  # If there are any values in the last set of records, store those.
  if (keys(%current)) {
    my %copy = %current;
    $values{$current{name}} = \%copy;
  }

  return %values;
}

=head2 get_aggr_raw_data(aggr => "aggregate", voluuid => "123-...", metadir => "repl_xfer", start => "0", length => "4k")

Get raw data for the given aggregate's meta-file in the given range.  All
parameters are required.

=cut

sub get_aggr_raw_data {
  my ($self, %args) = @_;

  my $server = $self->{SERVER};

  my $cmd = "$self->{CMD} raw $args{aggr}./metadir/vvols/$args{voluuid}/$args{metadir}\@$args{start}:$args{length}";

  my $output = $server->shell(cmd => $cmd);

  my @arr = split("\n", $output);

  my @bytes = WexUtils::get_raw_data_from_text(@arr);

  return @bytes;
}


=head2 get_raw_data(volume => "vol", inode => "85p", length => 256, [introspection => 0])

Get raw data for the given volume, inode, and length.  If introspection
is given and is true, the inode itself is the source of the raw data.
Otherwise, raw will start at the data for the inode (by appending @0 to
the inode number). When introspection is true, the length is ignored.

=cut

sub get_raw_data {
  my ($self, %args) = @_;

  my $server = $self->{SERVER};

  my $cmd = "$self->{CMD} raw $args{volume}.$args{inode}\@0:$args{length}";

  if (grep { $_ eq "introspection" } keys %args) {
    if ($args{introspection} == 1) {
      $cmd = "$self->{CMD} raw $args{volume}.$args{inode}";
    }
  }

  my $output = $server->shell(cmd => $cmd);

  my @arr = split("\n", $output);

  my @bytes = WexUtils::get_raw_data_from_text(@arr);

  return @bytes;
}


1;
