#
# Utility functions and definitions for Wex parsing.
#
#

use strict;

package WexUtils;

BEGIN {
  our @EXPORT;
  @EXPORT = qw(@repl_on_disk_state_structure_br
	       @repl_on_disk_state_structure_rr
	       @repl_on_disk_checkpoint_structure_br
	       @repl_on_disk_checkpoint_structure_rr
	       %repl_state_field_flags
	       %repl_on_disk_state_enum_br
	       %repl_on_disk_state_enum
	       $null_uuid);
};

our @EXPORT;

my %type_parsers =
  ("uint8_t"  => {func => \&interpret_number,
		  args => {num_bytes => 1}},
   "uint16_t" => {func => \&interpret_number,
		  args => {num_bytes => 2}},
   "uint32_t" => {func => \&interpret_number,
		  args => {num_bytes => 4}},
   "uint64_t" => {func => \&interpret_number,
		  args => {num_bytes => 8}},
   "spinnp_dsid_t" => {func => \&interpret_number,
		       args => {num_bytes => 4}},
   "spinnp_uuid_t" => {func => \&interpret_uuid,
		       args => {}},
   "uuid_t" => {func => \&interpret_uuid,
		args => {}},
   "spinnp_volume_id_t" => {func => \&interpret_volume_id,
			    args => {}},
   "ckpt_volume_id_t" => {func => \&interpret_ckpt_volume_id,
			  args => {}},
   "spinnp_image_id_t" => {func => \&interpret_image_id,
			   args => {}},
   "ckpt_image_id_t" => {func => \&interpret_ckpt_image_id,
			 args => {}});

our $null_uuid   = "00000000-0000-0000-0000-000000000000";


# Get raw byte data from lines of text.
sub get_raw_data_from_text {
  my (@arr) = @_;

  my @bytes;

  foreach my $line (@arr) {

    if ($line =~ /^[^\s]/) {
      next;
    }

    # Nuke the "line number" (the leading 3-digit counter).
    $line =~ s/^ *([0-9A-F]){3}//;

    # Nuke the trailing text version of the bytes.
    # Sometimes those values can look like hex.
    $line =~ s/.{16}$//;

    # Parse the bytes.
    while ($line =~ s/([0-9A-Fa-f]{2} ?)//) {
      push(@bytes, hex($1));
    }
  }

  return @bytes;
}

# Parse raw byte data into a sequence of types with the given names.
sub parse_raw_data {
  my (%args) = @_;

  my %values;
  my $offset = $args{offset};

  foreach my $a (@{$args{structure}}) {
    my ($type, $name) = @{$a};
    my $fn     =   $type_parsers{$type}{func};
    my %fnargs = %{$type_parsers{$type}{args}};

    $fnargs{bytes}  = $args{bytes};
    $fnargs{offset} = $offset;

    ($offset, $values{$name}) = &{$fn}(%fnargs);
  }

  return ($offset, %values);
}

# "Parse" number for the given bit field values.
sub parse_bit_fields {
  my (%args) = @_;

  my @flag_names_found;
  my $bits_found = 0;

  foreach my $bitval (keys %{$args{flags}}) {
    if ($args{bit_field} & $bitval) {
      $bits_found |= $bitval;
      push(@flag_names_found, $args{flags}{$bitval});
    }
  }

  # This isn't working right. :(  It's only applicable when we add new
  # states, which doesn't happen that often.
  my $diff = $bits_found ^ $args{bit_field};

  if ($diff) {
    printf("\tUNKNOWN bits found: 0x%x\n", $diff);
  }

  return @flag_names_found;
}




# Convert some number of bytes into a little-endian number.
sub interpret_number {
  my (%args) = @_;

  my $value = 0;

  for (my $i = 0; $i < $args{num_bytes}; ++$i) {
    $value += (($args{bytes}[$args{offset} + $i]) << $i*8);
  }

  return ($args{offset} + $args{num_bytes}, $value);
}

# Convert some bytes into a spinnp_uuid_t or uuid_t (same structure).
sub interpret_uuid {
  my (%args) = @_;

  my %values;

  ($args{offset}, %values) =
    parse_raw_data(bytes => $args{bytes},
			  structure => [["uint32_t", "time_low"],
					["uint16_t", "time_mid"],
					["uint16_t", "time_hi_and_version"],
					["uint8_t", "clock_seq_hi_and_resv"],
					["uint8_t", "clock_seq_low"],
					["uint8_t", "1"],
					["uint8_t", "2"],
					["uint8_t", "3"],
					["uint8_t", "4"],
					["uint8_t", "5"],
					["uint8_t", "6"]],
			  offset => $args{offset});

  my $str = sprintf("%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
		    $values{time_low}, $values{time_mid},
		    $values{time_hi_and_version},
		    $values{clock_seq_hi_and_resv},
		    $values{clock_seq_low}, $values{1}, $values{2}, $values{3},
		    $values{4}, $values{5}, $values{6});

  return ($args{offset}, $str);
}

# Convert some bytes into a volume ID.
sub interpret_volume_id {
  my (%args) = @_;

  my %values;

  # XXX - only handles a "whole" volume id, not a component.
  ($args{offset}, %values) =
    parse_raw_data(bytes => $args{bytes},
		   structure => [["uint32_t", "volume_id_type"],
				 ["uint32_t", "msid_major"],
				 ["uint32_t", "msid_minor"]],
		   offset => $args{offset});

  my $uuid;
  ($args{offset}, $uuid) = interpret_uuid(bytes => $args{bytes},
						 offset => $args{offset});
  $values{cluster} = $uuid;

  my $vsid;
  ($args{offset}, $vsid) = interpret_number(bytes => $args{bytes},
					    offset => $args{offset},
					    num_bytes => 4);
  $values{vserver_id} = $vsid;

  my %dsid;
  ($args{offset}, %dsid) =
    parse_raw_data(bytes => $args{bytes},
		   structure => [["uint32_t", "dsid_major"],
				 ["uint32_t", "dsid_minor"]],
		   offset => $args{offset});

  $values{dsid_major} = $dsid{dsid_major};
  $values{dsid_minor} = $dsid{dsid_minor};

  my $str = sprintf("{\n\tmsid = {major = %d (0x%x), minor = %d (0x%x)}\n",
		    $values{msid_major}, $values{msid_major},
		    $values{msid_minor}, $values{msid_minor});
  $str .= sprintf("\tcluster = %s\n", $values{cluster});
  $str .= sprintf("\tvserver_id = %s\n", $values{vserver_id});
  $str .= sprintf("\tdsid = {major = %d (0x%x), minor = %d (0x%x)} }",
		  $values{dsid_major}, $values{dsid_major},
		  $values{dsid_minor}, $values{dsid_minor});

  return ($args{offset}, $str);
}

# Convert some bytes into a ckpt_volume_id_t
sub interpret_ckpt_volume_id {
  my (%args) = @_;

  my %values;

  # XXX - only handles a "whole" volume id, not a component.
  ($args{offset}, %values) =
    parse_raw_data(bytes => $args{bytes},
		   structure => [["uint8_t", "volume_id_type"],
				 ["uint32_t", "msid_major"]],
		   offset => $args{offset});

  # Cluster UUID.
  my $uuid;
  ($args{offset}, $uuid) = interpret_uuid(bytes => $args{bytes},
					  offset => $args{offset});
  $values{cluster} = $uuid;

  # Vserver ID.
  my $vsid;
  ($args{offset}, $vsid) = interpret_number(bytes => $args{bytes},
					    offset => $args{offset},
					    num_bytes => 4);
  $values{vserver_id} = $vsid;

  # DSID.
  my %dsid;
  ($args{offset}, %dsid) =
    parse_raw_data(bytes => $args{bytes},
		   structure => [["uint32_t", "dsid_major"]],
		   offset => $args{offset});

  $values{dsid_major} = $dsid{dsid_major};

  # Blade ID.
  my $blade_id;
  ($args{offset}, $blade_id) = interpret_uuid(bytes => $args{bytes},
                                              offset => $args{offset});
  $values{blade_id} = $blade_id;

  # NDMP Session ID.
  my $ndmp_session_id;

  ($args{offset}, $ndmp_session_id) = interpret_uuid(bytes => $args{bytes},
                                                     offset => $args{offset});
  $values{ndmp_session_id} = $ndmp_session_id;

  my $str = sprintf("{\n\ttype = %d\n", $values{volume_id_type});
  $str .= sprintf("\tmsid = %d (0x%x)\n",
		    $values{msid_major}, $values{msid_major});
  $str .= sprintf("\tcluster = %s\n", $values{cluster});
  $str .= sprintf("\tvserver_id = %s\n", $values{vserver_id});
  $str .= sprintf("\tdsid = %d (0x%x)\n",
		  $values{dsid_major}, $values{dsid_major});
  $str .= sprintf("\tblade_id = %s\n", $values{blade_id});
  $str .= sprintf("\tndmp_session_id = %s }\n", $values{ndmp_session_id});

  return ($args{offset}, $str);
}

sub interpret_image_id {
  my (%args) = @_;

  my %values;

  ($args{offset}, %values) =
    parse_raw_data(bytes => $args{bytes},
		   structure => [["uint32_t", "logical_id"],
				 ["uint32_t", "physical_id"]],
		   offset => $args{offset});

  my $str = sprintf("{logical_id = %d (0x%x), physical_id = %d (0x%x)}",
		    $values{logical_id}, $values{logical_id},
		    $values{physical_id}, $values{physical_id});

  return ($args{offset}, $str);
}

sub interpret_ckpt_image_id {
  my (%args) = @_;

  my %values;

  ($args{offset}, %values) =
    parse_raw_data(bytes => $args{bytes},
		   structure => [["uint16_t", "flags"],
				 ["spinnp_uuid_t", "instance_uuid"],
				 ["spinnp_uuid_t", "version_uuid"],
				 ["uint32_t", "logical_id"],
				 ["uint32_t", "physical_id"]],
		   offset => $args{offset});

  my @image_flags =
    WexUtils::parse_bit_fields(flags => \%WexUtils::repl_ckpt_image_flags,
			       bit_field => $values{"flags"});

  my $flags_string = "";
  map { $flags_string .= "\n\t$_" } @image_flags;

  my $str = sprintf("{\n\tflags = 0x%x", $values{"flags"});
  $str .= sprintf($flags_string."\n");
  $str .= sprintf("\tinstance_uuid = %s\n", $values{"instance_uuid"});
  $str .= sprintf("\tversion_uuid = %s\n", $values{"version_uuid"});
  $str .= sprintf("\tlogical_id = %d (0x%x)\n",
		  $values{logical_id}, $values{logical_id});
  $str .= sprintf("\tphysical_id = %d (0x%x) }",
		  $values{physical_id}, $values{physical_id});

  return ($args{offset}, $str);
}


# ------------------------------ REPL specific ------------------------------

# In-volume format.
our @repl_on_disk_state_structure_br =
  (["uint32_t", "version"],
   ["uint32_t", "padding"],
   ["spinnp_uuid_t", "prepare_id"],
   ["uint64_t", "state_fields"],
   ["uint32_t", "on_disk_state"],
   ["spinnp_volume_id_t","target_volume_id"],
   ["spinnp_volume_id_t","source_volume_id"],
   ["spinnp_uuid_t", "reference_id_1"],
   ["spinnp_uuid_t", "reference_id_2"],
   ["uint8_t", "started"],
   ["uint8_t", "raw_dirs"]);

# In-aggregate format.
our @repl_on_disk_state_structure_rr =
  (["uint16_t", "data_size"],
   ["uint32_t", "version"],
   ["uint32_t", "padding_1"],
   ["spinnp_uuid_t", "prepare_id"],
   ["uint64_t", "state_fields"],
   ["uint8_t", "on_disk_state"],
   ["ckpt_volume_id_t","target_volume_id"],
   ["ckpt_volume_id_t","source_volume_id"],
   ["spinnp_uuid_t", "reference_id_1"],
   ["spinnp_uuid_t", "reference_id_2"],
   ["uint8_t", "started"],
   ["uint8_t", "engine_type"],
   ["uint8_t", "logical_mode"],
   ["ckpt_image_id_t", "reference_image_id"],
   ["ckpt_image_id_t", "transfer_image_id"]);

# In-aggregate format with vaults.
our @repl_on_disk_state_structure_sn =
  (["uint16_t", "data_size"],
   ["uint32_t", "version"],
   ["uint32_t", "padding_1"],
   ["spinnp_uuid_t", "prepare_id"],
   ["uint64_t", "state_fields"],
   ["uint8_t", "on_disk_state"],
   ["ckpt_volume_id_t","target_volume_id"],
   ["ckpt_volume_id_t","source_volume_id"],
   ["spinnp_uuid_t", "reference_id_1"],
   ["spinnp_uuid_t", "reference_id_2"],
   ["uint8_t", "started"],
   ["uint8_t", "engine_type"],
   ["uint8_t", "option_raw_dir"],
   ["ckpt_image_id_t", "reference_image_id"],
   ["ckpt_image_id_t", "transfer_image_id"],
   ["uint8_t", "named_data_in_use"],
   ["uint8_t", "named_data_info"]);

# In-volume format.
our @repl_on_disk_checkpoint_structure_br =
  (["uint32_t", "version"],
   ["uint32_t", "padding_1"],
   ["uuid_t", "prepareId"],
   ["uint64_t", "generation"],
   ["uint64_t", "bytes_done"],
   ["uuid_t", "transfer_uniq"],
   ["uuid_t", "sync_transfer_uniq"]);

# In-aggregate format.
our @repl_on_disk_checkpoint_structure_rr =
  (["uint16_t", "data_size"],
   ["uint32_t", "version"],
   ["uint32_t", "padding_1"],
   ["uuid_t", "prepareId"],
   ["uint64_t", "generation"],
   ["uint64_t", "exo_vol_type"],
   ["uint64_t", "bytes_transferred"],
   ["uuid_t", "transfer_uniq"],
   ["uuid_t", "sync_transfer_uniq"]);

our %repl_state_field_flags =
  ((1 << 0) => "ON_DISK_STATE",
   (1 << 1) => "TARGET_VOLUME_ID",
   (1 << 2) => "SOURCE_VOLUME_ID",
   (1 << 3) => "VOLUME_PREPARE_ID_XXX_DONT_USE",
   (1 << 4) => "REFERENCE_ID_1",
   (1 << 5) => "REFERENCE_ID_2",
   (1 << 6) => "STARTED",
   (1 << 7) => "RAW_DIRS_XXX_DONT_USE",
   (1 << 8) => "ENGINE_TYPE",
   (1 << 9) => "OPTION_RAW_DIR",
   (1 << 10) =>"REFERENCE_IMAGE_ID",
   (1 << 11) =>"TRANSFER_IMAGE_ID",
   (1 << 12) =>"NAMED_DATA_IN_USE",
   (1 << 13) =>"NAMED_DATA_INFO");

our %repl_on_disk_state_enum_br =
  (0 => "REPL_CONSTITUENT_VOLUME_PREPARE_STATE_PREPARED",
   1 => "REPL_CONSTITUENT_VOLUME_PREPARE_STATE_PREPARING",
   2 => "REPL_CONSTITUENT_VOLUME_PREPARE_STATE_LOCKED");

our %repl_on_disk_state_enum =
  (0 => "ON_DISK_PREPARING",
   1 => "ON_DISK_PREPARED",
   2 => "ON_DISK_COMPLETE",
   3 => "ON_DISK_LOCKED");

our %repl_ckpt_image_flags =
  ( (1 << 0) => "FLAG_INSTANCE_UUID",
    (1 << 1) => "FLAG_VERSION_UUID",
    (1 << 2) => "FLAG_PHYSICAL_ID",
    (1 << 3) => "FLAG_LOGICAL_ID" );

1;
