# -*- perl -*-

# $Id: $

## @summary XferEngineBase - a base class for Paloma Transfer Engine tests
## @authors beegle@netapp.com
## @pod here

=head1 NAME

XferEngineBase - a base class for Paloma Transfer Engine tests

=head1 SYNOPSIS

    use dp::snapmirror::paloma::unit_test::XferEngineBase;
    
    %testcases = { test_one => "Do test one." };
    &main();
    sub test_one {
        # Run your test steps here.
        return $TCD::PASS;
    }

=head1 DESCRIPTION

This module defines a base for Transfer Engine tests using TCD.  It provides
storage for variables common to most transfer engine tests and functions
for common transfer engine tasks.

This module will also cache the arrays of hashes returned from Server.pm
objects.  For example, the results of "vserver show" on the server are
unlikely to change between test cases.  The first time it is run, this
class will cache the data internally and use those values instead of going
back to the server every time.  You access the cached values by using the
get_xxx_yyy() style functions; to force a call to the server, you may
either use the stored Server object directly, or use the update_xxx_yyy()
style functions.

Data stored by this class is shared across TCD subtests.  Currently, that
data is:

=over

=item B<%testcases>

A hash of testcase name to a description string and an array of
thpl arguments.
For example:

   test1 => { description => "This is test1",
              args => ["THPL_ARG1=1","THPL_ARG2=2"] },
   test2 => { ... },

Use function get_testcases() to get the array of test case names and
descriptions suitable for passing to "new XferEngineBase".

=item B<$Server>

Instance of Server.pm object for communicating with filer nodes.

=item B<@nodes>

The nodes in this cluster as retrieved from Server.pm.

=item B<@vols>

The names of the source and destination volumes to use.  Item 0 is source,
item 1 is destination.

=item B<$mirror>

The NACL Snapmirror object created by any of the create_xxx_volumes(..)
functions.

=back

This class also provides common functions for common transfer engine tasks:

=over

=item B<creating and mirroring volumes>

create_mirror_volumes, create_ls_mirror_volumes, create_cascade_mirror

=item B<initializing, updating, and checking mirrors>

snapmirror_initialize, snapmirror_update, snapmirror_check,
snapmirror_initialize_ls_set, snapmirror_update_ls_set

=item B<getting composite mirror names>

get_error_node_names, get_volume_path, get_volume_path_rr

=item B<getting low-level information>

get_checkpoints, get_pedigree

=back

These functions may be used to verify error messages, if one is expected,
and are extremely simple in their parameters.  If finer grain control is
required, execute the functions via the Server within your test.

=head1 ROUTINES

=cut

# Things to come:
# SKTrace logging and dumping
# turn on ZAPI tracing
# dumping/parsing counters
# scope-based setflag object


use dp::snapmirror::paloma::unit_test::zsmcli;
use dp::snapmirror::paloma::unit_test::Wex;
use NACL::C::Node;
use NACL::C::Client;
use NACL::C::SnapmirrorPolicy;
use NACL::STask::Aggregate;
use NACL::STask::Client::Mount;
use NACL::STask::Volume;
use NACL::STask::VolumeSnapshot;
use NACL::STask::Snapmirror;
use NACL::STask::SystemNodeImage;
use NACL::C::Statistics;
use Options;
use Server; # Use this until we can convert all tests to Command Interface.
use Tharn qw($Log);
use Data::Dumper;

package XferEngineBase;

use base(TCD);

use strict;

# The $requires_debug_build flag is usually set to true because so many
# Paloma unit tests require FIJI points, which are only available in
# debug builds.  If your test does NOT require FIJI points (or any other
# debug-only feature), you should set $requires_debug_build to 0.
BEGIN {
  our @EXPORT;
  @EXPORT = qw(@vols @nodes @nodecis $Server %testcases $srcvol @dstvols @mirrors $requires_debug_build $help_text);
};

our @EXPORT;
our @vols;    # Volume names as strings, one source, one or more destinations.
our @nodes;   # Node names as strings.
our @nodecis; # SystemNode objects for the nodes in the cluster.
our $Server;  # Server.pm object.
our %testcases;
our $srcvol;  # Source volume as object.
our @dstvols; # Destination volumes as objects.
our @mirrors; # Snapmirror relationships as objects.
our $requires_debug_build = 1;
our $help_text = "no help provided";

my $current_nodeci = undef;
my $snapshot_count = 1;
my $aggr_base_name = "NACL_test_aggr_";
my @dstvols_cleanup;
my @mirrors_cleanup;
my @srcvols_cleanup;

# Options to set the volume names.  These should only be used when volumes
# are created during a setup step and the test is run when skipping setup
# (ie, during debugging of a test script).
# This allows volumes to be created with the pseudo-random names in Volume.pm
# and still operated on in later steps.
my $options = new Options;
$options->parameters(definitions =>
[{ name => "XFER_SRC_VOL_NAME",
   description => "Name of the source volume.",
   default => "new_src"
},
 { name => "XFER_DST_VOL_NAME",
   description => "Name of the dest volume.",
   default => "new_dst"
},
 { name => "XFER_DST_VOL_NAME2",
   description => "Name of the second dest volume.",
   default => "new_dst2"
},
 { name => "XFER_HELP",
   description => "Print help on running the test suite and exit.",
   default => "false"
},
 { name => "VOLUME_SPARE",
   description => "Keep the volumes from being purged.",
   default => "0"
}]);

# These variables are populated the first time the information is requested
# from the get_xxx functions like get_volume_show(..).  Subsequent calls will
# use the cached data to make lookups quicker.  To force these values to be
# updated, use the update_xxx functions like update_volume_show(..).
our @all_volumes;
our @all_aggregates;
our @all_vservers;
our @all_mirrors;


=head2 new

Create a new XferEngineBase to work with.  This loads the hash of testcases
into TCD for later execution.

=for html <hr>

=cut

sub new {
  my $class = shift;

  my @tcases;
  foreach my $t (keys (%testcases)) {
    push @tcases, ($t => $testcases{$t}{description});
  }

  my $self = new TCD(-testcases => [@tcases]);

  bless($self, $class);
  return $self;
}


# Version of cleanup for derived tests to call.
sub cleanup {

  foreach my $nci (@nodecis) {
    $nci->run(command => "fiji clear all");
  }

  if ($options->get(option => "VOLUME_SPARE") eq "0") {
    foreach my $m (@mirrors_cleanup) {
      $m->purge();
    }
    foreach my $m (@mirrors) {
      $m->purge();
    }
    foreach my $s (@srcvols_cleanup) {
      $s->purge();
    }
    if ($srcvol != undef) {
      $srcvol->purge();
    }
    foreach my $v (@dstvols_cleanup) {
      $v->purge();
    }
    foreach my $v (@dstvols) {
      $v->purge();
    }
  }

  # Don't let failures to delete the policies fail the test.
  # It's possible they were never created.
  eval {
    NACL::C::SnapmirrorPolicy->delete(command_interface => $nodecis[0],
                                      vserver => "vs0",
                                      policy => "TestDPPolicy");
    NACL::C::SnapmirrorPolicy->delete(command_interface => $nodecis[0],
                                      vserver => "vs0",
                                      policy => "TestXDPPolicy");
  };

}

# Define common methods in 'main' namespace.
# Only init, setup, cleanup, and uninit must be in the 'main' namespace.
# The other methods in this file are in the 'main' namespace for the
# convenience of the test cases using this class.
package main;

use Server;

# ----------------------------------------------------------------------

=head2 run_setup_before_each_testcase

Force the setup/cleanup steps to be run before and after each test case.
This can be handy if your test needs a brand new volume and/or mirror
for the test, but will make the tests run very slowly.

Alternately, you can re-define setup() to NOT create volumes and just
do that work in each of your testcases.

=for html <hr>

=cut

sub run_setup_before_each_testcase {
    Tharn::param("ARGS","-default","-rsc,all");
}

# ----------------------------------------------------------------------

=head2 main

Make a new transfer engine test and run it.

=for html <hr>

=cut

sub main {
    #  Create Test Case Driver object ##
    $Test = new XferEngineBase();
    if ($Test->error){
        $Test->nlog($Test->errmsg());
        return $TCD::FAIL;
    }

    # Performs method callbacks
    $Test->run_test();
    if ($Test->error){
        $Test->nlog($Test->errmsg());
        return $TCD::FAIL;
    }
    exit(0);
}

# ----------------------------------------------------------------------

=head2 init

Initialize objects for the Test to use, such as the $Server.  This function
will also test that the nodes are running a debug build.  You can disable
that test by setting $requires_debug_build to false.

=for html <hr>

=cut

sub init {
  $options->process();

  if (($options->get(option => "XFER_HELP") eq "1") ||
      ($options->get(option => "XFER_HELP") eq "true")) {
    print "Help for this test:\n$help_text\n";
    $Log->log(NATE::Result::NA->new("Help only run."));
    exit;
  }

  @vols = ($options->get(option => "XFER_SRC_VOL_NAME"),
	   $options->get(option => "XFER_DST_VOL_NAME"),
           $options->get(option => "XFER_DST_VOL_NAME2"));

  @nodecis = NACL::C::Node->find();

  foreach my $n (@nodecis) {
    push @nodes, $n->node;
  }

  # For now, keep this around for older code.
  $Server = new Server();

  if (scalar(@nodes) < 2) {
    $Log->warn("Running on only one node; some tests may not pass because of this.");
    push(@nodes, $nodes[0]);
  }

  if ($requires_debug_build) {
    my ($infoarr) = $options->get(option => "NODE_RELEASE_INFO");
    my @info = split(";", @{$infoarr}[0]);

    if ((@info > 4) && !($info[4] =~ /DEBUG/)) {
      $Log->log(NATE::Result::NA->new("Skipping test on non-debug build."));
      exit;
    }
  }

  my ($kenv) = $nodecis[0]->systemshell(command => '"kenv snapmirror.sn_enabled"');
  if (!($kenv =~ /false/)) {
    $Log->log("Running SN-style snapmirror jobs.");
  }
  else {
    $Log->warn("XferEngineBase.pm requires SN-style snapmirror jobs but they are not enabled on the test nodes.  Some tests may not pass because of this.");
  }

  return $TCD::PASS;
}

# ----------------------------------------------------------------------

=head2 setup

Configure the script's setup including state on the nodes.

=for html <hr>

=cut

sub setup {
  foreach my $nci (@nodecis) {
    $nci->run(command => "fiji clear all");
  }

  return $TCD::PASS;
}

# ----------------------------------------------------------------------

=head2 cleanup

Destroy the script's setup, including volumes on the nodes created by
this script.  See the VOLUME_SPARE option for preserving volumes after
the test finishes.

=for html <hr>

=cut

sub cleanup {
  # Do all the work of cleaning up in the class' method so
  # that derived tests can also use it.
  XferEngineBase::cleanup();

  return $TCD::PASS;
}

# ----------------------------------------------------------------------

=head2 uninit

Destroy objects that the Test used.  This implementation does nothing.

=for html <hr>

=cut

sub uninit {
  return $TCD::PASS;
}

# ----------------------------------------------------------------------

=head2 init_mirror_objects

Connect the mirror variable to an existing mirror, if it's not already
connected.  This is needed when tests are run without the setup() step.
When run in the normal case (when setup IS run), this function is a no-op.

=for html <hr>

=cut

sub init_mirror_objects() {
  if (scalar(@mirrors) == 0) {
    my $nodeci = NACL::C::Node->find();

    if ($srcvol eq undef) {
      ($srcvol) = NACL::STask::Volume->find(command_interface => $nodeci,
                                            filter => {volume => $vols[0]});
    }

    if (scalar(@dstvols) == 0) {
      for (my $i=1; $i < scalar(@vols); $i++) {
        $dstvols[$i-1] = NACL::STask::Volume->find(command_interface => $nodeci,
                                                   filter => {volume => $vols[$i]});
      }
    }

    for (my $i=0; $i < scalar(@dstvols); $i++) {
      $mirrors[$i] = NACL::STask::Snapmirror->find(command_interface => $nodeci,
                                                   filter => { "source-path" => $srcvol,
                                                               "destination-path" => $dstvols[$i] });
    }
  }
}



# ----------------------------------------------------------------------
# ---------- Mirrors and Volumes ----------
# ----------------------------------------------------------------------

=head2 create_mirror_volumes

Create, mirror, and manipulate flex volumes in a DP relationship.
Based on the given arguments, this method can populate the source volume
and/or initialize the mirror.

This function will populate the @XferEngineBase::vols with the created
volume names.  Item 0 is always the source and the remaining entries are
the destinations.  This function will also set the $srcvol, @dstvols, and
@mirrors objects.

B<Arguments>

=over

=item B<vol_size> => I<string>

I<Optional> : A size string as given to 'volume create', such as "20M" or "1G".
If not supplied, a default size of "20M" is used.

=item B<dst_vol_size> =>I<string>

I<Optional> : A size string as given to 'volume create', such as "20M" or "1G".
If not supplied, the destination will be the same size as the source volume.

=item B<fill_size> => I<string>

I<Optional> : If 'populate' is 1, fill the source volume to the given size.
This is a size string, as for 'vol_size'.  If not supplied, "13M" is used.

=item B<populate> => I<0|1>

I<Optional> : Whether to populate the source volume with data.
If 'initialize' is also 1, the volume will be populated before the
initialize.  Default: 0.

=item B<initialize> => I<0|1>

I<Optional> : Whether to initialize the snapmirror relationship after
creating it. Default: 1.

=item B<num_dest> => I<int>

I<Optional> : The number of destination volumes to create.  All destinations
will be created on the destination node.  Default: 1.

=item B<type> => I<string>

I<Optional> : One of "LS", "DP", or "XDP".  Default: DP.

=back

B<Examples>

    create_mirror_volumes();

    create_mirror_volumes(populate => 1, initialize => 0);

    create_mirror_volumes(num_dest => 2);

=for html <hr>

=cut

sub create_mirror_volumes {
  my %args = @_;

  my %opts;

  # Defaults.
  $opts{vol_size}  = "20M";
  $opts{dst_vol_size} = undef;
  $opts{fill_size} = "13M";
  $opts{populate}  = 0;
  $opts{initialize} = 1;
  $opts{num_dest} = 1;
  $opts{type} = "DP";

  # Copy any overrides.
  foreach my $k (keys %opts) {
    $opts{$k} = $args{$k} if exists $args{$k};
  }

  if ($opts{dst_vol_size} == undef) {
    $opts{dst_vol_size} = $opts{vol_size};
  }

  # With NACL, it doesn't seem possible to make it create an
  # aggregate for us.  So, just make one and reuse it if it
  # already exists.
  my $aggrname = $aggr_base_name . _cleanup_nodename($nodes[0]);
  my $srcaggr = NACL::STask::Aggregate->create(command_interface => $nodecis[0],
                                               aggregate => $aggrname,
                                               diskcount => 3,
                                               if_exists => 'reuse',
                                               nacltask_wait => 1,
                                               'method-timeout' => 600);

  $aggrname = $aggr_base_name . _cleanup_nodename($nodes[1]);
  my $dstaggr = NACL::STask::Aggregate->create(command_interface => $nodecis[1],
                                               aggregate => $aggrname,
                                               diskcount => 3,
                                               if_exists => 'reuse',
                                               nacltask_wait => 1,
                                               'method-timeout' => 600);

  # Create the test "-tries 1" policies.
  _create_policies();

  # Save any existing source volume for cleanup later.
  if ($srcvol != undef) {
    push @srcvols_cleanup, $srcvol;
  }

  # Create the source volume.
  $srcvol = NACL::STask::Volume->create(command_interface => $nodecis[0],
                                        aggregate => $srcaggr,
                                        type =>"RW",
                                        vserver => "vs0",
                                        size => $opts{vol_size},
                                        "snapshot-policy" => "none",
                                        "nacltask_wait" => 1,
                                        'method-timeout' => 600);
  # Save the source volume's name.
  @vols = ();
  $vols[0] = $srcvol->volume();

  # Save any existing mirrors and destination volumes for cleanup later.
  if (scalar(@mirrors) > 0) {
    push @mirrors_cleanup, @mirrors;
  }

  if (scalar(@dstvols) > 0) {
    push @dstvols_cleanup, @dstvols;
  }

  for (my $i=0; $i < $opts{num_dest}; $i++) {
    # Create the destination volume.
    $dstvols[$i] = NACL::STask::Volume->create(command_interface => $nodecis[0],
                                               aggregate => $dstaggr,
                                               type => "DP", # Volume type is DP for DP and XDP mirrors.
                                               vserver => "vs0",
                                               size => $opts{dst_vol_size},
                                               "nacltask_wait" => 1,
                                               'method-timeout' => 600);
    # Save the destination volume's name.
    push @vols, $dstvols[$i]->volume();

    # SN.0 style snapmirror uses snapmirror policies to manage tries.
    my $policy_name = "TestDPPolicy";
    if ($opts{type} eq "XDP") {
      $policy_name = "TestXDPPolicy";
    }
    $mirrors[$i] = NACL::STask::Snapmirror->create(command_interface => $nodecis[0],
                                                   "source-path" => $srcvol,
                                                   "destination-path" => $dstvols[$i],
                                                   policy => $policy_name,
                                                   nacltask_wait => 1,
                                                   do_initialize => 0,
                                                   type => $opts{type},
                                                   'method-timeout' => 600);
  }

  # Populate the source volume, if necessary.
  if ($opts{populate}) {
    _populate_data(command_interface => $nodecis[0],
                   volume => $srcvol,
                   fill_amount => $opts{fill_size});
  }

  # Initialize the mirror, if necessary.
  if ($opts{initialize}) {
    NACL::STask::VolumeSnapshot->create(command_interface => $nodecis[0],
                                        volume => $srcvol,
                                        vserver => $srcvol->vserver(),
                                        snapshot => "snap".$snapshot_count,
                                        'method-timeout' => 600);

    foreach my $m (@mirrors) {
      $mirrors[0]->initialize(command_interface => $nodecis[0],
                              "source-snapshot" => "snap".$snapshot_count,
                              nacltask_wait => 1,
                              'method-timeout' => 600);
    }

    $snapshot_count++;
  }

  @all_volumes = ();
}

# ----------------------------------------------------------------------

=head2 add_more_data

Add more data to the source volume.

=cut

sub add_more_data {
  my %args = @_;

  my %opts;

  # Defaults.
  $opts{add_size} = "10M";

  # Copy any overrides.
  foreach my $k (keys %opts) {
    $opts{$k} = $args{$k} if exists $args{$k};
  }

  _populate_data(command_interface => $nodecis[0],
                 volume => $srcvol,
                 fill_amount => $opts{fill_size});

  return $TCD::PASS;
}

# ----------------------------------------------------------------------

=head2 snapmirror_initialize

Call snapmirror initialize on the destination volume.  If an error string
is given, expect an error and compare it to the actual error that occurred.

B<Arguments>

=over

=item B<error> => I<string>

I<Optional> : The error string to expect if the initialize fails.
If the initialize should succeed, don't pass an error string or pass 'undef'.

=item B<dst_vol> => I<string>

I<Optional> : Name of the destination volume to initialize.  If none is given,
$XferEngineBase::vols[1] will be used.

=item B<snapshot> => I<string>

I<Optional> : The snapshot name to transfer.  The snapshot must already
exist on the source volume.  If no snapshot name is given, a
new snapshot is created as "snap".$snapshot_count and that is transferred.

=back

B<Examples>

    snapmirror_initialize();  # Should succeed.

    snapmirror_initialize(error => "Raw directory negotiation failed");
    # Should fail with the given error.

    snapmirror_initialize(dst_vol => $XferEngineBase::vols[2]);
    # Run on the given destination and it should succeed.

=for html <hr>

=cut

sub snapmirror_initialize {
  my %args = @_;

  my %opts;

  # Defaults.
  $opts{error} = undef;
  $opts{dst_vol} = $vols[1];
  $opts{snapshot} = undef;

  # Copy any overrides.
  foreach my $k (keys %opts) {
    $opts{$k} = $args{$k} if exists $args{$k};
  }

  my $actual = "";

  if (!$opts{snapshot}) {
    $opts{snapshot} = "snap".$snapshot_count;
    NACL::STask::VolumeSnapshot->create(command_interface => $nodecis[0],
                                        volume => $srcvol,
                                        vserver => $srcvol->vserver(),
                                        snapshot => $opts{snapshot});
    $snapshot_count++;
  }

  # Get the mirror for the destination the caller wants to initialize.
  my $mirror = NACL::STask::Snapmirror->find(command_interface => $nodecis[0],
                                             filter => {'destination-volume' => $opts{dst_vol}});

  if (!$mirror) {
    die("No mirror with name '".$opts{dst_vol}."' found.");
  }

  my @end_states;
  push @end_states, "Snapmirrored";
  push @end_states, "Uninitialized";

  $mirror->initialize("source-snapshot" => $opts{snapshot},
                      nacltask_wait => 0,
                      'method-timeout' => 600);

  $mirror->monitor(nacltask_wait_for_state => \@end_states,
                   nacltask_wait_for_status => "Idle",
                   'method-timeout' => 600);

  my $mirror_state = $mirror->state();

  $actual = $mirror_state->last_transfer_error;

  $Log->comment("snapmirror initialize last transfer error: '".$actual."'");

  if (($actual ne "") and ($actual ne "-")) {
    if (!($opts{error} eq undef)) {
      # Got an error and we expected one.  Compare the strings.
      if ($actual =~ /$opts{error}/) {
	return ($actual =~ $opts{error});
      }
      die("Snapmirror initialize failed, but not with the expected error: expected '$opts{error}', got '$actual'");
    }

    # Got an error, but it wasn't expected.
    die("Snapmirror initialize failed: '$actual'");
  }

  if (!($opts{error} eq undef)) {
    # Didn't get an error, but we expected one.
    die("Snapmirror initialized when it should have failed.");
  }

  return $actual;
}

# ----------------------------------------------------------------------

=head2 snapmirror_initialize_ls_set

Call snapmirror initialize-ls-set on the source volume.  If an error string
is given, expect an error and compare it to the actual error that occurred.

B<Arguments>

=over

=item B<error_string> (string)

I<Optional> : The error string to expect if the initialize fails.
If the initialize should succeed, don't pass an error_string or pass 'undef'.

=back

B<Examples>

    snapmirror_initialize_ls_set();  # Should succeed.

    snapmirror_initialize_ls_set("Raw directory negotiation failed");
    # Should fail with the given error.

=for html <hr>

=cut

sub snapmirror_initialize_ls_set {
  my $errstr  = shift;

  my $actual = "";

  eval {
#    $mirrors[0]->initialize_ls_set(nacltask_wait => 1);

  };
  if ($@) {
    $actual = $@;
    if (!($errstr eq undef)) {
      # Got an error and we expected one.  Compare the strings.
      if ($actual =~ /$errstr/) {
	return ($actual =~ /$errstr/);
      }
      die("Snapmirror initialize-ls-set failed, but not with the expected error: expected '$errstr', got '$actual'");
    }
    # Got an error, but we didn't expect one.
    die("Snapmirror initialize-ls-set failed: '$actual'");
  }

  if (!($errstr eq undef)) {
    # Didn't get an error, but we expected one.
    die("Snapmirror initialized when it should have failed.");
  }

  return $actual;
}

# ----------------------------------------------------------------------

=head2 snapmirror_update

Call snapmirror update on the destination volume.  If an error string
is given, expect an error and compare it to the actual error that occurred.
If a destination volume is not given, $XferEngineBase::vols[1] is used.

B<Arguments>

=over

=item B<error> => I<string>

I<Optional> : The error string to expect if the update fails.
If the update should succeed, don't pass an error string or pass 'undef'.

=item B<dst_vol> => I<string>

I<Optional> : Name of the destination volume to update.  If none is given,
$XferEngineBase::vols[1] will be used.

=item B<snapshot> => I<string>

I<Optional> : The snapshot name to transfer.  The snapshot must already
exist on the source volume.  If no snapshot name is given, a
new snapshot is created as "snap".$snapshot_count and that is transferred.

=back

B<Examples>

    snapmirror_update();  # Should succeed.

    snapmirror_update(dst_vol => "second_dp_mirror"); # Update the given mirror.

    snapmirror_update(error => "Raw directory negotiation failed");
    # Should fail with the given error.

=for html <hr>

=cut

sub snapmirror_update {
  my %args = @_;

  my %opts;

  # Defaults.
  $opts{error} = undef;
  $opts{dst_vol} = $vols[1];
  $opts{snapshot} = undef;

  # Copy any overrides.
  foreach my $k (keys %opts) {
    $opts{$k} = $args{$k} if exists $args{$k};
  }

  my $actual = "";

  if (!$opts{snapshot}) {
    $opts{snapshot} = "snap".$snapshot_count;
    NACL::STask::VolumeSnapshot->create(command_interface => $nodecis[0],
                                        volume => $srcvol,
                                        vserver => $srcvol->vserver(),
                                        snapshot => $opts{snapshot});
    $snapshot_count++;
  }

  # Get the mirror for the destination the caller wants to initialize.
  my $mirror = NACL::STask::Snapmirror->find(command_interface => $nodecis[0],
                                             filter => {'destination-volume' => $opts{dst_vol}});
  if (!$mirror) {
    die("No mirror with name '".$opts{dst_vol}."' found.");
  }

  my @end_states;
  push @end_states, "Snapmirrored";

  $mirror->update("source-snapshot" => $opts{snapshot},
                  nacltask_wait => 0);

  $mirror->monitor(nacltask_wait_for_state => \@end_states,
                   nacltask_wait_for_status => "Idle",
                   'method-timeout' => 600);

  my $mirror_state = $mirror->state();

  $actual = $mirror_state->last_transfer_error;

  $Log->comment("snapmirror update last transfer error: '".$actual."'");

  if (($actual ne "") and ($actual ne "-")) {
    if (!($opts{error} eq undef)) {
      # Got an error and we expected one.  Compare the strings.
      if ($actual =~ /$opts{error}/) {
	return $actual =~ $opts{error};
      }
      die("Snapmirror update failed, but not with the expected error: expected '$opts{error}', got '$actual'");
    }

    # Got an error, but it wasn't expected.
    die("Snapmirror update failed: '$actual'");
  }

  if (!($opts{error} eq undef)) {
    # Didn't get an error, but we expected one.
    die("Snapmirror updated when it should have failed.");
  }

  return $actual;
}

# ----------------------------------------------------------------------

=head2 snapmirror_update_ls_set

Call snapmirror update on the destination volume.  If an error string
is given, expect an error and compare it to the actual error that occurred.
The destination volume updated is always $XferEngineBase::vols[1].

B<Arguments>

=over

=item B<error_string> (string)

I<Optional> : The error string to expect if the initialize fails.
If the update should succeed, don't pass an error_string or pass 'undef'.

=back

B<Examples>

    snapmirror_update_ls_set();  # Should succeed.

    snapmirror_update_ls_set("Raw directory negotiation failed");
    # Should fail with the given error.

=for html <hr>

=cut

sub snapmirror_update_ls_set {
  my $errstr = shift;

  my $actual = "";

  eval {
#    $mirrors[0]->update_ls_set(nacltask_wait => 1);

  };
  if ($@) {
    $actual = $@;
    if (!($errstr eq undef)) {
      # Got an error and we expected one.  Compare the strings.
      if ($actual =~ /$errstr/) {
	return ($actual =~ $errstr);
      }
      die("Snapmirror update failed, but not with the expected error: expected '$errstr', got '$actual'");
    }

    # Got an error, but it wasn't expected.
    die("Snapmirror update failed: '$actual'");
  }

  if (!($errstr eq undef)) {
    # Didn't get an error, but we expected one.
    die("Snapmirror updated when it should have failed.");
  }

  return $actual;
}

# ----------------------------------------------------------------------

=head2 snapmirror_check

Call snapmirror check on the destination volume.  If an error string
is given, expect an error and compare it to the actual error that occurred.
The destination volume checked is always $XferEngineBase::vols[1].

B<Arguments>

=over

=item B<error_string> (string)

I<Optional> : The error string to expect if the initialize fails.
If the update should succeed, don't pass an error_string or pass 'undef'.

=back

B<Examples>

    snapmirror_check();  # Should succeed.

    snapmirror_check("Raw directory negotiation failed");
    # Should fail with the given error.

=for html <hr>

=cut

sub snapmirror_check {
  my $errstr = shift;

  my $actual = "";

  eval {
    $mirrors[0]->check(command_interface => $nodecis[0],
                       nacltask_wait => 1,
                       'method-timeout' => 600);
  };

  if ($@) {
    my $all_failure = $@;
    # If we use all of the failure message, it includes the call stack
    # from perl, which includes the failure text we're expecting and the
    # =~ will match (doh!).  So, just grab the line with the actual
    # CLI command that failed and compare with that.
    ($actual) = grep { $_ =~ "Error: command failed" } split('\n', $all_failure);
    if (!($errstr eq undef)) {
      # Got an error and we expected one.  Compare the strings.
      if ($actual =~ /$errstr/) {
	return $actual =~ $errstr;
      }
      # Make the whole error visible for debugging the tests.
      if ($actual eq "") {
        print $all_failure;
      }
      die("Snapmirror update failed, but not with the expected error: expected '$errstr', got '$actual'");
      return ($actual =~ $errstr);
    }

    # Got an error, but it wasn't expected.
    die("Snapmirror update failed: '$actual'");
  }

  if (!($errstr eq undef)) {
    # Didn't get an error, but we expected one.
    die("Snapmirror checked when it should have failed.");
  }

  return $actual;
}

# ----------------------------------------------------------------------
# ---------- Cached data access ----------
# ----------------------------------------------------------------------

=head2 update_volume_show

Update and return the cached volume information.

B<Examples>

    my @all_vols = update_volume_show();

=for html <hr>

=cut

sub update_volume_show {
  my $output = $XferEngineBase::Server->volume_show();
  @all_volumes = @{ $output };
  return @all_volumes;
}

# ----------------------------------------------------------------------

=head2 get_volume_show

Get the volume information.
If a volume name is given, only volumes matching that name will be returned.

B<Arguments>

=over

=item B<volume_name> (string)

I<Optional> : The volume whose data should be returned.  Names are
compared exactly (not regexp).

=back

B<Examples>

    my @all_vols = get_volume_show();  # Get all vols.

    my ($foo_vol) = get_volume_show("foo");  # Get volume named "foo".

=for html <hr>

=cut

sub get_volume_show {
  my $vol_name = shift;

  if (@all_volumes == 0) {
    update_volume_show();
  }

  if ($vol_name ne undef) {
    my @match_vols = grep { $_->{volume} eq $vol_name } @all_volumes;

    return @match_vols;
  }

  return @all_volumes;
}

# ----------------------------------------------------------------------

=head2 update_aggregate_show

Update and return the cached aggregate information.

B<Examples>

    my @all_aggrs = update_aggregate_show();

=for html <hr>

=cut

sub update_aggregate_show {
  my $output = $XferEngineBase::Server->storage_aggregate_show();
  @all_aggregates = @{ $output };
  return @all_aggregates;
}

# ----------------------------------------------------------------------

=head2 get_aggregate_show

Get the aggregate information.
If an aggregate name is given, only aggregates matching that name will
be returned.

B<Arguments>

=over

=item B<aggr_name> (string)

I<Optional> : The aggregate whose data should be returned.  Names are
compared exactly (not regexp).

=back

B<Examples>

    my @all_aggrs = get_aggregate_show();  # Get all aggregates.

    my ($foo_aggr) = get_aggregate_show("foo");  # Get aggregate named "foo".

=for html <hr>

=cut

sub get_aggregate_show {
  my $aggr_name = shift;

  if (@all_aggregates == 0) {
    update_aggregate_show();
  }

  if ($aggr_name ne undef) {
    my @match_aggrs = grep { $_->{aggregate} eq $aggr_name } @all_aggregates;

    return @match_aggrs;
  }

  return @all_aggregates;
}

# ----------------------------------------------------------------------

=head2 update_vserver_show

Update and return the cached vserver information.

B<Examples>

    my @all_vs = update_vserver_show();

=for html <hr>

=cut

sub update_vserver_show {
  my $output = $XferEngineBase::Server->vserver_show();
  @all_vservers = @{ $output };
  return @all_vservers;
}

# ----------------------------------------------------------------------

=head2 get_vserver_show

Get the vserver information.
If a vserver type is given (as a string) only vservers of that type will
be returned.
If no argument is given, all known vservers will be returned.

B<Arguments>

=over

=item B<vserver_type> (string)

I<Optional> : The type of vserver whose data should be returned.  If there
are more than one vserver of the given type, all matching vserver
information will be returned.

=back

B<Examples>

    my @all_vs = get_vserver_show();  # Get all vservers.

    my ($admin_vs) = get_vserver_show("admin"); # Get "admin" type vserver.

=for html <hr>

=cut

sub get_vserver_show {
  my $vserver_type = shift;

  if (@all_vservers == 0) {
    update_vserver_show();
  }

  if ($vserver_type ne undef) {
    my @type_vservers = grep { $_->{type} eq $vserver_type } @all_vservers;

    return @type_vservers;
  }

  return @all_vservers;
}

# ----------------------------------------------------------------------

=head2 update_snapmirror_show

Update and return the cached snapmirror information.

B<Examples>

    my @all_sm = update_snapmirror_show();

=for html <hr>

=cut

sub update_snapmirror_show {
  my $output = $XferEngineBase::Server->snapmirror_show();
  @all_mirrors = @{ $output };
  return @all_mirrors;
}

# ----------------------------------------------------------------------

=head2 get_snapmirror_show

Get the snapmirror information.
If a destination volume is given (as a string) only mirrors with that as a
destination will be returned.
If no argument is given, all known snapmirrors will be returned.

B<Arguments>

=over

=item B<destination_volume> (string)

I<Optional> : The destination volume of the snapmirror relationship to
get data for.  Volume names are compared exactly (not regexp).  If
there is more than one relationship with the given destination volume,
all those relationships will be returned.

=back

B<Examples>

    my @all_sm = get_snapmirror_show();  # Get all snapmirror relationships.

    my ($foo_sm) = get_snapmirror_show("foo");
    # Get all snapmirror relationships with destination volume "foo".

=for html <hr>

=cut

sub get_snapmirror_show {
  my $dest_vol = shift;

  if (@all_mirrors == 0) {
    update_snapmirror_show();
  }

  if ($dest_vol ne undef) {
    my @dest_mirrors = grep { $_->{'destination-volume'} eq $dest_vol } @all_mirrors;

    return @dest_mirrors;
  }

  return @all_mirrors;
}


# ----------------------------------------------------------------------
# ---------- Other transfer engine functions ----------
# ----------------------------------------------------------------------


# ----------------------------------------------------------------------

=head2 get_error_node_names

Get the node names like they appear in error messages that contain arguments.
That is, return strings of the format "cluster/node".

Returns an array of strings with the same number of items as there are nodes.

B<Examples>

    my @node_err_names = get_error_node_names();
    my $expected_message = sprintf("Error on %s", $node_err_names[0]);

=for html <hr>

=cut

sub get_error_node_names {
  my @all_vservers = get_vserver_show("admin");
  if (@all_vservers != 1) {
    die("Wrong number of admin vservers.");
  }
  my $vserver = $all_vservers[0];

  my @results;

  foreach my $n (@XferEngineBase::nodes) {
    push @results, "node \"$n\" in cluster \"".$vserver->{vserver}."\"";
  }

  return @results;
}

# ----------------------------------------------------------------------

=head2 get_volume_path

Get the path to the given volume in the format "vserver:vol".

B<Arguments>

=over

=item B<vol_name> (string)

I<Required> : The volume name to get the path for.

=back

B<Examples>

    my $vol_path = get_volume_path("foo");
    my $expected_error = sprintf("Error updating %s", $vol_path);

=for html <hr>

=cut

sub get_volume_path {
  my $vol = shift;

  my ($admin_vserver) = NACL::STask::Vserver::find(command_interface => $nodecis[0],
                                                   filter => {type => "admin"});

  if (!$admin_vserver) {
    die("No admin vserver.");
  }

  return $admin_vserver->{vserver}.":".$vol;
}

# ----------------------------------------------------------------------

=head2 get_volume_path_rr

Get the path to the given volume in the format "cluster://vserver/vol".

B<Arguments>

=over

=item B<vol_name> (string)

I<Required> : The volume name to get the path for.

=back

B<Examples>

    my $vol_path = get_volume_path("foo");
    my $expected_error = sprintf("Error updating %s", $vol_path);

=for html <hr>

=cut

sub get_volume_path_rr {
  my $vol = shift;

  my ($admin_vserver) = NACL::STask::Vserver::find(command_interface => $nodecis[0],
                                                   filter => {type => "admin"});

  if (!$admin_vserver) {
    die("No admin vserver.");
  }

  my ($cluster_vserver) = NACL::STask::Vserver::find(command_interface => $nodecis[0],
                                                     filter => {type => "cluster"});

  if (!$cluster_vserver) {
    die("No cluster vserver.");
  }

  return $admin_vserver->{vserver}."://".$cluster_vserver->{vserver}."/".$vol;
}

# ----------------------------------------------------------------------

=head2 get_dest_cluster_vserver

Get the destination volume's cluster name and vserver name.
These are required for most snapmirror commands.

=B<Arguments>

=over

=item B<dst_vol> (string)

I<Required> : The name of the destination volume to get information for.

=back

=for html <hr>

=cut

sub get_dest_cluster_vserver {
  my $dst_vol = shift;

  my @cluster_ids = $XferEngineBase::Server->cluster_identity_show();
  if (@cluster_ids != 1) {
    die("No cluster identification.");
  }

  my $cluster_name = $cluster_ids[0]->{"name"};

  my @dest_vols = get_volume_show($dst_vol);
  if (@dest_vols != 1) {
    die("No destination volume.");
  }
  my $vserver = $dest_vols[0]->{vserver};

  return ($cluster_name, $vserver);
}


# ----------------------------------------------------------------------

=head2 get_volume_hex

Get the hex value of a volume name for some error messages without
a cluster ID.

B<Arguments>

=over

=item B<volume_name> (string)

I<Required> : The name of the volume to get the hex "name" for.

=back

B<Examples>

    my $vol_err_name = get_volume_hex("foo");

=for html <hr>

=cut

sub get_volume_hex {
  my $vol = shift;
  my $msid_only = shift;

  my ($volinfo) = get_volume_show($vol);

  if ($msid_only) {
    return sprintf("0x%0x00000000:0x0", $volinfo->{msid});
  }

  return sprintf("0x%0x00000000:0x%0x00000000",
		 $volinfo->{msid}, $volinfo->{dsid});
}


# ----------------------------------------------------------------------

=head2 get_checkpoints_br

Get all the checkpoint information for the given volume.  This includes
the special all-zero checkpoint entry that contains the on-disk state.
The checkpoints are ordered by how they appear in the repl_ckpt directory.

B<Arguments>

=over

=item B<volume_name> (string)

I<Required> : The name of the volume to get checkpoints for.  Name is
compared exactly (not regexp).

=back

B<Examples>

    my @ckpts = get_checkpoints_br("foo");
    my ($on_disk_values) = grep {$_->{name} eq $WexUtils::null_uuid} @ckpts;
    print "On disk version# = $on_disk_values->{version}\n";

=for html <hr>

=cut

sub get_checkpoints_br {
  my $vol = shift;

  my $wex = new Wex($XferEngineBase::Server);
  my %dirs = $wex->get_dir_data(volume => $vol,
				inode  => "85P/repl_ckpt");

  my @ckpts;

  if (! keys(%dirs)) {
    return @ckpts;
  }

  my @ckpt_order =
    sort { $dirs{$a}{"index"} <=> $dirs{$b}{"index"} } keys %dirs;

  foreach my $ck (@ckpt_order) {
    # Only consider entries with a uuid-like name as checkpoints.
    # (That is, ignore "." and "..".)
    if (! $ck =~ /[a-z0-9\-]+/) {
      next;
    }

    my $inum  = $dirs{$ck}{inum};
    my @bytes = $wex->get_raw_data(volume => $vol,
				   inode  => $inum,
				   length => 256);

    my $structure_ref = \@WexUtils::repl_on_disk_checkpoint_structure;
    if ($ck eq $WexUtils::null_uuid) {
      $structure_ref = \@WexUtils::repl_on_disk_state_structure;
    }

    my ($offset, %values) =
      WexUtils::parse_raw_data(bytes => \@bytes,
			       offset => 0,
			       structure => $structure_ref);

    push @ckpts, \%values;
  }

  return @ckpts;
}


=head2 get_xfer_state

Get the xfer_state information for the given volume.  This does NOT
include any checkpoint information.  Use the get_checkpoints_xx methods
to get that information.

B<Arguments>

=over

=item B<volume_name> (string)

I<Required> : The name of the volume to get xfer_state information for.  
Name is compared exactly (not regexp).

=back

B<Examples>

    my $xfer_state = get_xfer_state("dst_vol");
    if ($xfer_state->{option_named_data} eq "1") {
        ...

=for html <hr>

=cut

sub get_xfer_state {
  my ($vol) = @_;

  my $wex = new Wex($XferEngineBase::Server);
  my ($vol_info) = get_volume_show($vol);

  my $vol_uuid = $vol_info->{uuid};
  my $aggr     = $vol_info->{aggregate};

  # Make sure there's actually a meta-dir for the transfer state.
  my %dirs = $wex->get_aggr_dir_data(aggr => $aggr, voluuid => $vol_uuid);

  if (! grep {"xfer_state"} keys(%dirs)) {
    die("No xfer_state metafile.");
  }

  my @bytes = $wex->get_aggr_raw_data(aggr => $aggr,
				      voluuid => $vol_uuid,
				      metadir => "xfer_state",
				      start => "0",
				      length => "4k");
  my ($offset, %values) = 
    WexUtils::parse_raw_data(bytes => \@bytes,
			     offset => 0,
			     structure => \@WexUtils::repl_on_disk_state_structure_sn);

  return %values;
}


# ----------------------------------------------------------------------

=head2 get_pedigree

Get the pedigree UUID for the given volume name.

B<Arguments>

=over

=item B<volume_name> (string)

I<Required> : The name of the volume to get pedigree information for.  Name is
compared exactly (not regexp).

=item B<pedigree_name> (string)

I<Required> : The name of the pedigree file to read.  Valid values are
"dir_pedigree" for the directory format pedigree and
"dw_pedigree" for the data warehouse format pedigree.

=back

B<Examples>

    my $actual_pedigree = get_pedigree("foo", "dir_pedigree");

=cut

sub get_pedigree {
  my $vol = shift;
  my $pedfile = shift;

  my $wex = new Wex($XferEngineBase::Server);
  # This is here so the wex error will be noticed.  When using the
  # "get_raw_data", it doesn't throw an error when it can't read the values.
  my %dirs = $wex->get_dir_data(volume => $vol,
				inode => "/metadir/repl_pedigree");
  my @bytes = $wex->get_raw_data(volume => $vol,
				 inode =>"/metadir/repl_pedigree/$pedfile",
				 length => 256,
				 introspection => 1);

  my @pedigree_format = (["uuid_t", "pedigree"]);

  # Make it obvious when the value was not read from a metafile.
  if (scalar(@bytes) == 0) {
    die("Failed to read pedigree: $pedfile");
  }

  my $bad_uuid = "ffffffff-ffff-ffff-ffff-ffffffffffff";
  my %values;
  $values{pedigree} = $bad_uuid;

  my $offset = 0;

  ($offset, %values) =
    WexUtils::parse_raw_data(bytes => \@bytes,
			     offset => 88,
			     structure => \@pedigree_format);

  if ($values{pedigree} eq $bad_uuid) {
    die("Failed to read pedigree: $pedfile");
  }

  return $values{pedigree};
}

# ----------------------------------------------------------------------

=head2 get_latest_snapshot

Get the most recent snapshot on the given volume name.

B<Arguments>

=over

=item B<volume_name> (string)

I<Required> : The name of the volume to get the snapshot for.  Name is compared
exactly (not regexp).

=back

B<Examples>

    my $latest_snap = get_latest_snapshot("foo");

=cut

sub get_latest_snapshot {
  my $vol = shift;

  my @snapshots = $Server->volume_snapshot_show(volume => $vol);
  my $latest_snap = $snapshots[scalar(@snapshots)-1];

  return $latest_snap;
}


# ----------------------------------------------------------------------

=head2 get_dw_identity

Get the data warehouse (DW) identity from a volume when given its name.

B<Arguments>

=over

=item B<volume_name> (string)

I<Required> : The name of the volume to get DW identity from.  Name is
compared exactly (not regexp).

=back

B<Examples>

    my $actual_dw_identity = get_dw_identity("foo");

=cut

sub get_dw_identity {
  my $vol = shift;

  my $wex = new Wex($XferEngineBase::Server);
  my %dirs = $wex->get_dir_data(volume => $vol,
				inode => "/metadir/datawarehouse/wholeVolRep");
  my @bytes = $wex->get_raw_data(volume => $vol,
				 inode =>"/metadir/datawarehouse/wholeVolRep/dw_uuid",
				 length => 256,
				 introspection => 1);

  my @dwid_format = (["uuid_t", "dw_identity"]);

  my ($offset, %values) =
    WexUtils::parse_raw_data(bytes => \@bytes,
			     offset => 88,
			     structure => \@dwid_format);

  return $values{dw_identity};
}

# ----------------------------------------------------------------------

=head2 verify_raw_mode_transfer

Verify that the transfer was done in raw mode.  Compare the pedigree of
the given source and destination volumes.

B<Arguments>

=over

=item B<src_volume> (string)

I<Optional> : Name of the source volume for comparing the pedigree.  Default:
$XferEngineBase::vols[0].

=item B<dst_volume> (string)

I<Optional> : Name of the destination volume for comparing the
pedigree.  Default: $XferEngineBase::vols[1].

=back

B<Examples>

    verify_raw_mode_transfer();

    verify_raw_mode_transfer(dst_volume => $XferEngineBase::vols[2]);

=for html <hr>

=cut

sub verify_raw_mode_transfer {
  my %args = @_;

  my %opts;

  $Log->comment("Verifying raw mode transfer was chosen.\n");

  # Defaults.
  $opts{src_volume} = $XferEngineBase::vols[0];
  $opts{dst_volume} = $XferEngineBase::vols[1];

  # Copy any overrides.
  foreach my $k (keys %opts) {
    $opts{$k} = $args{$k} if exists $args{$k};
  }

  # Compare pedigree of dest with that of the source.
  my $dst_pedigree = get_pedigree($opts{dst_volume}, "dir_pedigree");
  my $src_pedigree = "";

  eval {
    # If the source volume has a pedigree meta-file,
    # that's the pedigree we should be using.
    # If there is no pedigree meta-file, use the latest
    # snapshot's instance UUID.
    $src_pedigree = get_pedigree($opts{src_volume}, "dir_pedigree");
  };
  if ($@) {
    my $actual = $@;
    if (($actual =~ "Wex error: ") || ($actual =~ "Failed to read pedigree")) {
      $Log->comment("No pedigree file available.  Using latest snapshot instance UUID.");

      my $src_snap = get_latest_snapshot($opts{src_volume});
      $src_pedigree = $src_snap->{"instance-uuid"};
    }
  }

  $Log->comment("\nSource pedigree: $src_pedigree\n".
		"Dest pedigree:   $dst_pedigree\n");

  if ($src_pedigree ne $dst_pedigree) {
    die("Raw mode was not chosen: pedigree on dest ($dst_pedigree) doesn't match source pedigree ($src_pedigree");
    return $TCD::FAIL;
  }

  return $TCD::PASS;
}

# ----------------------------------------------------------------------

=head2 verify_logical_mode_transfer_rr

Verify that the transfer was done in logical mode.  Compares the pedigree
of the source and destination volumes for inequality.

B<Arguments>

=over

=item B<src_volume> (string)

I<Optional> : Name of the source volume for comparing the pedigree.
Default: $XferEngineBase::vols[0].

=item B<dst_volume> (string)

I<Optional> : Name of the destination volume for comparing the
pedigree.  Default: $XferEngineBase::vols[1].

=back

B<Examples>

    verify_logical_mode_transfer_rr();

    verify_logical_mode_transfer_rr(dst_volume => $XferEngineBase::vols[2]);

=for html <hr>

=cut

sub verify_logical_mode_transfer_rr {
  my %args = @_;

  my %opts;

  $Log->comment("Verifying logical mode transfer was chosen.\n");

  # Defaults.
  $opts{src_volume} = $XferEngineBase::vols[0];
  $opts{dst_volume} = $XferEngineBase::vols[1];

  # Copy any overrides.
  foreach my $k (keys %opts) {
    $opts{$k} = $args{$k} if exists $args{$k};
  }

  eval {
    verify_raw_mode_transfer(%opts);
  };
  if (!$@) {
    die("Transfer was raw mode when it should have been logical.");
    return $TCD::FAIL;
  }

  return $TCD::PASS;
}

# Pass through to the SN version.
sub verify_logical_mode_transfer {
  verify_logical_mode_transfer_sn(@_);
}

# ----------------------------------------------------------------------

=head2 verify_logical_mode_transfer_sn

Verify that the transfer was done in logical mode.  There should not be
a dir-format pedigree when logical mode is chosen.

B<Arguments>

=over

=item B<src_volume> (string)

I<Optional> : Name of the source volume for comparing the pedigree.
Default: $XferEngineBase::vols[0].

=item B<dst_volume> (string)

I<Optional> : Name of the destination volume for comparing the
pedigree.  Default: $XferEngineBase::vols[1].

=back

B<Examples>

    verify_logical_mode_transfer_sn();

    verify_logical_mode_transfer_sn(dst_volume => $XferEngineBase::vols[2]);

=for html <hr>

=cut

sub verify_logical_mode_transfer_sn {
  my %args = @_;

  my %opts;

  $Log->comment("Verifying logical mode transfer was chosen.\n");

  # Defaults.
  $opts{src_volume} = $XferEngineBase::vols[0];
  $opts{dst_volume} = $XferEngineBase::vols[1];

  # Copy any overrides.
  foreach my $k (keys %opts) {
    $opts{$k} = $args{$k} if exists $args{$k};
  }

  # See if there's a dir-format pedigree.
  my $dst_pedigree = "";
  eval {
    $dst_pedigree = get_pedigree($opts{dst_volume}, "dir_pedigree");
  };
  if ($@) {
    my $actual = $@;
    if (($actual =~ "Wex error: ") || ($actual =~ "Failed to read pedigree")) {
      $Log->comment("No pedigree file available.  That's as expected for a logical dir transfer.");

      return $TCD::PASS;
    }
    else {
      die($actual);
    }
  }

  die("Pedigree was '$dst_pedigree', but no dir-format pedigree was expected.");
  return $TCD::FAIL;
}

# ----------------------------------------------------------------------

=head2 verify_lrse_mode_transfer

Verify that the transfer was done in lrse mode.  Compare the DW identity
of the given source and destination volumes.  The source volume must have
the most recent transfer snapshot as its latest snapshot.

B<Arguments>

=over

=item B<src_volume> (string)

I<Optional> : Name of the source volume for comparing the pedigree.  Default:
$XferEngineBase::vols[0].

=item B<dst_volume> (string)

I<Optional> : Name of the destination volume for comparing the
pedigree.  Default: $XferEngineBase::vols[1].

=back

B<Examples>

    verify_lrse_mode_transfer();

    verify_lrse_mode_transfer(dst_volume => $XferEngineBase::vols[2]);

=for html <hr>

=cut

sub verify_lrse_mode_transfer {
  my %args = @_;

  my %opts;

  $Log->comment("Verifying lrse mode transfer was chosen.\n");

  # Defaults.
  $opts{src_volume} = $XferEngineBase::vols[0];
  $opts{dst_volume} = $XferEngineBase::vols[1];

  # Copy any overrides.
  foreach my $k (keys %opts) {
    $opts{$k} = $args{$k} if exists $args{$k};
  }

  # Get the most recent snapshot from the source volume.
  my @snapshots = $XferEngineBase::Server->volume_snapshot_show(volume => $opts{src_volume});
  my $latest_snap = $snapshots[scalar(@snapshots)-1];

  # Compare the instance UUID of the source snapshot to the
  # DW identity on the destination.
  my $expected_dwid = $latest_snap->{'instance-uuid'};
  my $actual_dwid = get_dw_identity($opts{dst_volume});

  $Log->comment("\nSource DW identity: $expected_dwid\n".
		  "Dest DW identity:   $actual_dwid\n");

  if ($expected_dwid ne $actual_dwid) {
    die("LRSE mode was not chosen: DW identity on dest ($actual_dwid) doesn't match source volume's latest snapshot instance UUID ($expected_dwid)");
    return $TCD::FAIL;
  }

  return $TCD::PASS;
}

# ----------------------------------------------------------------------

=head2 verify_on_disk_state

Compare a list of expected on-disk-state values against their actual values.

B<Arguments>

=over

=item B<volume> (string)

I<Required> : The volume name to verify the on-disk-state on.

=item B<expecteds> (hash reference)

I<Required> : The states to check and their expected values.

=back

B<Examples>

    my %exp;
    $exp{version} = 3;
    $exp{started} = 1;
    verify_on_disk_state(volume => "foo", expecteds => \%exp);

=for html <hr>

=cut

sub verify_on_disk_state {
  my %args = @_;
  my $vol = $args{volume};
  my %expecteds = %{$args{expecteds}};

  my %actuals = get_on_disk_state_rr_plus($vol, "sn");

  foreach my $k (keys %expecteds) {
    my $actual;
    $actual = $actuals{$k} if exists $actuals{$k};

    $Log->comment("On disk state: $k = $actual, expected $expecteds{$k}.");

    if ($actual != $expecteds{$k}) {
      $Log->error(die("On disk state: $k = $actual, but expected $expecteds{$k}."));
      return $TCD::FAIL;
    }
  }

  return $TCD::PASS;
}

# ----------------------------------------------------------------------

=head2 get_cm_values

Get a list of counter manager values as a hash with the key as the instance name.

B<Arguments>

=over

=item B<object> (string)

I<Required> : The counter manager object name.

=item B<instance> (string)

I<Required> : The counter manager instance name.

=item B<counter> (string)

I<Required> : The counter manager counter name.

=item B<field> (string)
I<Optional> : The field to get for the counter.  Default: value.

=item B<node> (string)
I<Optional> : The node to limit the counters to.  Default: "*"

=back

B<Examples>

    my %values = get_cm_values(object => "repl_global_ops",
                               instance => "src-data-push_file_update*",
                               counter => "ops_begun");
    if ($values{src-data-push_file_update_anon} > 0) {
        do something;
    }

=for html <hr>

=cut

sub get_cm_values {
  my %args = @_;

  my $field = "value";
  $field = $args{field} if exists $args{field};

  $args{node} = "*" if not exists $args{node};

  # There's a bug in statistics show right now for the handling
  # of -node parameter.  So, don't pass it here, but do the
  # filtering by hand afterwards.
  my @values = $Server->statistics_show(object => $args{object},
                                        instance => $args{instance},
                                        counter => $args{counter});
  my %results;

  foreach my $v (@values) {
    if (($args{node} ne "*") and ($v->{"node"} ne $args{node})) {
      next;
    }
    $results{$v->{"instance-name"}} = $v->{$field};
  }

  return %results;
}

# ----------------------------------------------------------------------

=head2 verify_global_op_counts

Get the repl_global_ops counters and verify the number of ops of the given types
and comparisons.  This function uses the "base-delta" value from the counter
for comparison, so that multiple transfers can be done without having to reset
the counters.

B<Arguments>

=over

=item B<expecteds> (array)

I<Required> : An array of (key, comparator, value) entries.  Each entry is looked up
              by the key in the list of repl_global_ops counters and the value is
              compared against the given value, with the given comparison routine.

=back

B<Examples>

  my @expecteds;
  $expecteds[0] = ['src-data-push_file_update_anon', sub { return int($_[0] <= $_[1]); }, 12];
  $expecteds[1] = ['src-data_push_file_update_by_name', sub { return int($_[0] == $_[1]); }, 0];

  verify_global_op_counts(@expecteds);

=for html <hr>

=cut

sub verify_global_op_counts {
  my @expecteds = @_;

  my %counters = get_cm_values(object => "repl_global_ops",
                               instance => "*",
                               counter => "ops_begun",
                               field => "base-delta");

  foreach my $a (@expecteds) {
    my @values = @{$a};
    my $actual = $counters{$values[0]};
    my $comp = $values[1];
    my $exp = $values[2];
    my $ok = &{$comp}($actual, $exp);
    $Log->comment("$values[0] had $actual delta ops from previous run.");
    if (!$ok) {
      die("Counter value $values[0] not as expected: got $actual, expected $exp.");
    }
  }

  return $TCD::PASS;
}

# ----------------------------------------------------------------------

=head2 get_global_op_counts_before

Get the repl_global_ops counters before the activity of interest.  This
method should be used with the verify_global_op_counts_after to get the
delta.  These methods were added because, inexplicably, the method
verify_global_op_counts stopped returning a valid delta value.

B<Arguments>

=over

=item B<node> (string)

I<Optional> : Limit the statistics to the given node's values, if given.
              If this value is not give, all nodes are queried.

=back

B<Examples>

  my %stats_before = get_global_op_counts_before();

  # Do the thing that generates repl_global_ops.

  my @expecteds;
  $expecteds[0] = ['src-data-push_file_update_anon', sub { return int($_[0] <= $_[1]); }, 12];
  $expecteds[1] = ['src-data_push_file_update_by_name', sub { return int($_[0] == $_[1]); }, 0];

  verify_global_op_counts_after(\%stats_before, @expecteds);

=for html <hr>

=cut

sub get_global_op_counts_before {
  my ($node_arg) = @_;

  if (not $node_arg) {
    $node_arg = "*";
  }

  my %counters = get_cm_values(object => "repl_global_ops",
                               instance => "*",
                               counter => "ops_begun",
                               field => "base-value",
                               node => $node_arg);

  return %counters;
}

# ----------------------------------------------------------------------

=head2 verify_global_op_counts_after

Get the repl_global_ops counters and verify the number of ops of the given types
and comparisons.  This function uses the "base-value" from the counter so it
should be used in conjunction with the get_global_op_counts_before method
which returns the baseline values.

B<Arguments>

=over

=item B<before> (hash ref)

I<Required> : A hash reference to counter "before" values as retrieved from
              a call to get_global_op_counts_before.  These are the values
              used for computing the op delta.

=item B<expecteds> (array ref)

I<Required> : An array ref of (key, comparator, value) entries.  Each entry is looked up
              by the key in the list of repl_global_ops counters and the value is
              compared against the given value, with the given comparison routine.

=item B<node> (string)

I<Optional> : The node to limit the statistics collection to.  This should match
              the value given to get_global_op_counts_before().  If this value is
              not given, all nodes are included.

=back

B<Examples>

  my %stats_before = get_global_op_counts_before();

  # Do the thing that generates repl_global_ops.

  my @expecteds;
  $expecteds[0] = ['src-data-push_file_update_anon', sub { return int($_[0] <= $_[1]); }, 12];
  $expecteds[1] = ['src-data_push_file_update_by_name', sub { return int($_[0] == $_[1]); }, 0];

  verify_global_op_counts_after(before => \%stats_before, expecteds => \@expecteds);

=for html <hr>

=cut

sub verify_global_op_counts_after {
  my %args = @_;

  my $before = $args{before};
  my $expecteds = $args{expecteds};
  my $node = "*";
  $node = $args{node} if exists $args{node};

  my %after = get_cm_values(object => "repl_global_ops",
                            instance => "*",
                            counter => "ops_begun",
                            field => "base-value",
                            node => $node);

  # Compute the deltas.
  # Note that this is not smart enough to sum-up the values
  # from all nodes, so the values are from the last node
  # in the query.  Since we don't know here if the node
  # is important for the comparison, make the caller pass in
  # a node value, if they care.
  my %deltas;
  foreach my $k (keys %after) {
    $deltas{$k} = int($after{$k}) - int($before->{$k});
  }

  foreach my $a (@{$expecteds}) {
    my @values = @{$a};
    my $actual = $deltas{$values[0]};
    my $comp = $values[1];
    my $exp = $values[2];
    my $ok = &{$comp}($actual, $exp);
    $Log->comment("$values[0] had $actual delta ops from previous run.");
    if (!$ok) {
      die("Counter value not as expected: got $actual, expected $exp");
    }
  }

  return $TCD::PASS;
}

# ----------------------------------------------------------------------

=head2 set_src_server

Set the Server object to use nodes[0] as the execution server.

=for html <hr>

=cut

sub set_src_server {
  $XferEngineBase::Server->set_execution_server(-Server => $XferEngineBase::nodes[0]);
}

# ----------------------------------------------------------------------

=head2 set_dst_server

Set the Server object to use nodes[1] as the execution server.

=for html <hr>

=cut

sub set_dst_server {
  $XferEngineBase::Server->set_execution_server(-Server => $XferEngineBase::nodes[1]);
}

# ----------------------------------------------------------------------

=head2 set_force_logical_dirs

Set the paloma_force_logical_dirs flag on the current execution server.

=for html <hr>

=cut

sub set_force_logical_dirs {
  $XferEngineBase::Server->dblade(cmd => "priv set test; setflag paloma_force_logical_dirs 1");
}

# ----------------------------------------------------------------------

=head2 clear_force_logical_dirs

Clear the paloma_force_logical_dirs flag on the current execution server.

=for html <hr>

=cut

sub clear_force_logical_dirs {
  $XferEngineBase::Server->dblade(cmd => "priv set test; setflag paloma_force_logical_dirs 0");
}

# ----------------------------------------------------------------------

=head2 get_newest_source_snapshot

Get the name of the newest snapshot on the source volume.

=for html <hr>

=cut

sub get_newest_source_snapshot {
  my @snaps = NACL::STask::VolumeSnapshot->find(command_interface => $nodecis[0],
                                                filter => {volume => $srcvol->volume});
  if (scalar(@snaps) < 1) {
    die("No snapshots on volume '$srcvol->volume'.");
  }

  return $snaps[0]->snapshot;
}

# ----------------------------------------------------------------------

=head2 format_volume_id_for_zapi

Format the volume with the given name as a string for zsmcli.
Returns a hash reference suitable for formatting with zsmcli.pm.

B<Arguments>

=over

=item B<volname> => I<string>

I<Required> : The volume name to create a zapi-style hash for.

=item B<use_cluster> => I<bool>

I<Optional> : True if the cluster's UUID should be included in the resulting
hash, false if a NULL UUID should be used.  Defaults to true.

=item B<dsid> => I<string>

I<Optional> : The DSID to include in the resulting hash.  If no value is given,
a zero DSID is specified (because most engine interfaces expect that).

=back

B<Examples>

    my $vol = format_volume_id_for_zapi(volname => "foo");
    my $vol = format_volume_id_for_zapi(volname => "foo", dsid => "1234");

=for html <hr>

=cut

sub format_volume_id_for_zapi {
  my %args = @_;
  if (! exists $args{'use_cluster'}) {
    $args{'use_cluster'} = 1;
  }

  if (! exists $args{'dsid'}) {
    $args{'dsid'} = "0";
  }

  if (! exists $args{'volname'}) {
    die("Need volname to format.");
    return;
  }

  my ($info) = get_volume_show($args{'volname'});

  my $dsid_str = $args{'dsid'};
  my $cluster_id_str = $WexUtils::null_uuid;

  if ($args{'use_cluster'}) {
    $XferEngineBase::Server->cli(cmd => "set test");
    my @cluster_id = $XferEngineBase::Server->cluster_identity_show();
    if (@cluster_id != 1) {
      die("No cluster identification.");
    }
    $cluster_id_str = $cluster_id[0]->{"uuid"};
  }

  my $arg = { "od-volume-identifier" =>
	      { "volume-dsid" =>
		{ "od-spinnp-dsid" =>
		  { "od-dsid-major-id" => $dsid_str,
		    "od-dsid-minor-id" => "0" } },
		"volume-msid" =>
		{ "od-spinnp-dsid" =>
 		  { "od-dsid-major-id" => $info->{"msid"},
 		    "od-dsid-minor-id" => "0" } },
		"cluster-id" => $cluster_id_str } };

  return zsmcli::format_odtype($arg);
}


# ----------------------------------------------------------------------


sub turn_on_zapi_tracing {
  my $zapis = shift;

  if ($zapis eq undef) {
    $zapis = "*";
  }

  my $nodestr = join(",", @XferEngineBase::nodes);

  $XferEngineBase::Server->cli(cmd => "set test");
  $XferEngineBase::Server->cli(cmd => "debug zapi trace modify -node $nodestr -zapi $zapis -level all");
}

# ----------------------------------------------------------------------


sub turn_off_zapi_tracing {
  my $zapis = shift;

  if ($zapis eq undef) {
    $zapis = "*";
  }

  my $nodestr = join(",", @XferEngineBase::nodes);

  $XferEngineBase::Server->cli(cmd => "set test");
  $XferEngineBase::Server->cli(cmd => "debug zapi trace modify -node $nodestr -zapi $zapis -level none");
}

# ----------------------------------------------------------------------

=head2 get_on_disk_state

Get the on-disk-state for the given volume.

B<Arguments>

=over

=item B<volume_name> (string)

I<Required> : The name of the volume to get the on-disk-state for.
The name is used exactly as given (not regexp).

=back

B<Examples>

    my $on_disk_values = get_on_disk_state("foo");

=for html <hr>

=cut

sub get_on_disk_state_br {
  my $vol = shift;

  my $wex = new Wex($Server);

  my %dirs = $wex->get_dir_data(volume => $vol, inode => "85P");

  if (! keys(%dirs)) {
    die("Couldn't get meta-file information for volume '$vol'.");
    return $TCD::FAIL;
  }

  %dirs = $wex->get_dir_data(volume => $vol,
			     inode => $dirs{repl_ckpt}{inum});

  if (! keys(%dirs)) {
    die("Couldn't get checkpoint information for volume '$vol'.");
    return $TCD::FAIL;
  }

  my $inum = $dirs{"00000000-0000-0000-0000-000000000000"}{inum};

  my @bytes = $wex->get_raw_data(volume => $vol,
				 inode => $inum,
				 length => 256);

  my ($offset, %values) =
    WexUtils::parse_raw_data(bytes => \@bytes,
		     structure => \@WexUtils::repl_on_disk_state_structure);

  return %values;
}

sub get_on_disk_state_rr_plus {
  my $vol = shift;
  my $rel = shift;

  my $wex = new Wex($Server);
  my ($vol_info) = get_volume_show($vol);

  # Make sure there's actually a meta-dir for the transfer state.
  my %dirs = $wex->get_aggr_dir_data(aggr => $vol_info->{aggregate},
				     voluuid => $vol_info->{uuid});

  if (! grep {"xfer_state"} keys(%dirs)) {
    die("Couldn't get meta-file information for volume '$vol'.");
    return $TCD::FAIL;
  }

  # Get the on-disk-state from the meta-data
  my @bytes = $wex->get_aggr_raw_data(aggr => $vol_info->{aggregate},
				      voluuid => $vol_info->{uuid},
				      metadir => "xfer_state",
				      start => "0",
				      length => "4k");

  my $structure_ref = \@WexUtils::repl_on_disk_state_structure_rr;
  if ($rel eq "sn") {
    $structure_ref = \@WexUtils::repl_on_disk_state_structure_sn;
  }

  my ($offset, %values) = 
    WexUtils::parse_raw_data(bytes => \@bytes,
			     offset => 0,
			     structure => $structure_ref);

  return %values;
}


# ----------------------------------------------------------------------

=head2 verify_no_on_disk_state

Verify there is no transfer on-disk-state for the given volume.

B<Arguments>

=over

=item B<volume_name> (string)

I<Required> : The name of the volume to get checkpoints for.  Name is
compared exactly (not regexp).

=back

B<Examples>

    verify_no_on_disk_state("foo");

=for html <hr>

=cut

sub verify_no_on_disk_state {
  my $vol = shift;
  my $wex = new Wex($Server);

  my %dirs = $wex->get_dir_data(volume => $vol, inode => "85P");

  if (! keys(%dirs)) {
    die("Couldn't get meta-file information for volume '$vol'.");
    return $TCD::FAIL;
  }

  if (grep { $_ eq "repl_ckpt" } keys(%dirs)) {
    die("There is on-disk checkpoint information for the volume '$vol'.");
    return $TCD::FAIL;
  }

  return $TCD::PASS;
}

# ----------------------------------------------------------------------
# ------------------------------ DEPRECATED ----------------------------


# XXX this should be replaced with NOTE functions to create/find a
# different aggregate. - only used by vol move error checker
sub get_other_aggregate {
  my $vol = shift;

  my @match_vols = get_volume_show($vol);

  if (@match_vols != 1) {
    die("No matching volumes to get aggregate for.");
    return undef;
  }

  my $curr_aggr = $match_vols[0]->{aggregate};
  my @all_aggrs = get_aggregate_show();
  my @avail_aggrs = grep {   $_->{aggregate} ne $curr_aggr && \
			   !($_->{aggregate} =~ /aggr0_.*/) && \
			   !($_->{aggregate} =~ /striped.*/) } @all_aggrs;

  if (@avail_aggrs == 0) {
    die("No other aggregates to use.");
    return undef;
  }

  return $avail_aggrs[0]->{aggregate};
}


# ------------------------------ INTERNAL ------------------------------

#
# Arguments, all required:
#
# - command_interface = the node to contact
# - volume = the volume to put data in
# - fill_amount = size to fill, as a string like "5M" or "10K"
#
# A client machine must also be available from the environment.
#
sub _populate_data {
  my %args = @_;

  my %opts;

  # Defaults.
  $opts{fill_amount} = "5M";
  $opts{filename} = "stuff";
  $opts{volume} = $srcvol;
  $opts{command_interface} = $nodecis[0];

  # Copy values and overrides.
  foreach my $k (keys %opts) {
    $opts{$k} = $args{$k} if exists $args{$k};
  }

  $opts{volume}->mount(command_interface => $opts{"command_interface"},
                       "junction-path" => "/".$opts{volume}->volume());

  my ($client) = NACL::C::Client->find();
  my $mount = NACL::STask::Client::Mount->create(command_interface => $client,
                                                 server_command_interface => $opts{"command_interface"},
                                                 vserver => $opts{"volume"}->vserver(),
                                                 path => "/".$opts{"volume"}->volume());

  my $client_path = $mount->state()->mount_point;
  $client->execute(command => "sudo chmod 777 ".$client_path);
#  RichData::populate(path => $client_path, file_avg => 3, client => $client->client(), fill_vol => "50", timeout => 500);
  $client->execute(command => "dd if=/dev/urandom of=".$client_path."/".$opts{filename}." bs=1 count=".$opts{"fill_amount"},
                   timeout => 500);
  $client->execute(command => "mkdir ".$client_path."/subdir");
  $client->execute(command => "dd if=/dev/urandom of=".$client_path."/subdir".$opts{filename}." bs=1 count=100");
}

sub _create_policies {
  # We need a "one retry only" snapmirror policy for SN-style
  # relationships.  Make one, but don't fail if it already exists.
  eval {
    NACL::C::SnapmirrorPolicy->create(command_interface => $nodecis[0],
                                      vserver => "vs0",
                                      policy => "TestXDPPolicy",
                                      tries => 1,
                                      restart => "never");
  };
  if ($@) {
    my $err = $@;
    if (! (($err =~ /duplicate entry/) || ($err =~ /modify/) || ($err =~ /already exists/))) {
      die $err;
    }
  }

  eval {
    NACL::C::SnapmirrorPolicy->create(command_interface => $nodecis[0],
                                      vserver => "vs0",
                                      policy => "TestDPPolicy",
                                      tries => 1,
                                      restart => "never");
  };
  if ($@) {
    my $err = $@;
    if (! (($err =~ /duplicate entry/) || ($err =~ /modify/) || ($err =~ /already exists/))) {
      die $err;
    }
  }

  eval {
    NACL::C::SnapmirrorPolicy->add_rule(command_interface => $nodecis[0],
                                        vserver => "vs0",
                                        policy => "TestXDPPolicy",
                                        "snapmirror-label" => "smtest",
                                        keep => "1");
  };
  if ($@) {
    my $err = $@;
    if (! (($err =~ /duplicate entry/) || ($err =~ /already exists/))) {
      die $err;
    }
  }
}

sub _cleanup_nodename {
  my $nodename = shift;

  my $cleanname = $nodename;
  $cleanname =~ s/\-/_/g;

  return $cleanname;
}

1;
