# $Id: $

## @summary zsmcli - a simple ZSMCLI interface
## @authors beegle@netapp.com
## @pod here

=head1 NAME

zsmcli - a simple ZSMCLI interface

=head1 SYNPOSIS

    use Server;
    use zsmcli;
    
    my $server = new Server();
    my $zsm = new zsmcli($server);
    my $cmd = { cmd => "d-foo", args => "..." };
    
    my $xml = $zsm->run_command($cmd);
    my @values = zsmcli::get_values($xml, "//bar");

=head1 DESCRIPTION

This library simplifies using the zsmcli command and parsing the results.
It is not intended to test the ZAPI workings themselves but to add easy
ZAPI communication to a node.

=head1 ROUTINES

=cut

use strict;
use Server;
use Options;
use Tharn qw($Log);
use XML::Simple;
use XML::XPath;
use XML::XPath::XMLParser;

package zsmcli;

# ----------------------------------------------------------------------

=head2 new

Create a new zsmcli instance for the given server.  The server pointer is
stored, but the current execution server on that object will affect where
the zsmcli command is eventually run.  See run_command(..) for more details.

B<Arguments>

=over

=item server

I<Required> : Server object.

=back

=for html <hr>

=cut

sub new {
  my $class  = shift;
  my $server = shift;
  my $self = { SERVER => $server };

  bless($self, $class);

  return $self;
}

# ----------------------------------------------------------------------

=head2 format_odtype

Format a ZAPI type for zsmcli's consumption.  Generically, this is:
   type-name=[sub-type=1 other-sub-type=2]

B<Arguments>

=over

=item types_hash

I<Required> : Hash reference to type-value pairs, nested for sub-types.

=back

B<Examples>

   { "is-incremental" => "true" } becomes "is-incremental=true"

   { "dest-vol-id" => { "od-volume-id" => { "major-id" => "3",
                                            "minor-id" => "0" } } }
   becomes
   "dest-vol-id=[od-volume-id=[major-id=3 minor-id=0]]"

=for html <hr>

=cut

sub format_odtype {
  my $h = shift;

  # Simple string value, so add no formatting.
  if (ref($h) ne "HASH") {
    return $h;
  }

  my $str = "[";

  foreach my $k (keys %{$h}) {
    $str .= "$k=".format_odtype($h->{$k})." ";
  }

  $str .= "]";
}


# ----------------------------------------------------------------------

=head2 format_command

Format a ZAPI command for zsmcli.  This function takes a hash with two
keys: "cmd" and "args".  "cmd" is the name of the ZAPI you want to
execute (such as d-snapmirror-start).  "args" is a hash of the arguments
to the ZAPI and their values (such as "destination-volume-id" => { ...}).

The command returned from this function is suitable for passing to
zsmcli as a Server's shell command.

B<Arguments>

=over

=item cmd_hash

I<Required> : Hash reference containing the command and arguments to format.

=back

B<Examples>

    my $my_zapi_request = { cmd => "d-snapmirror-start",
                            args => { ... } };
    my $full_cmd = format_command($my_zapi_request);
    $Server->shell(cmd => $full_cmd);

=for html <hr>

=cut

sub format_command {
  my $cmd_hash_ref = shift;

  my $cmd = $cmd_hash_ref->{"cmd"};
  my $argstr = "";

  foreach my $k (keys %{$cmd_hash_ref->{"args"}}) {
    $argstr .= " $k=\"";
    my $h = $cmd_hash_ref->{"args"}->{$k};
    $argstr .= format_odtype($h);
    $argstr .= "\"";
  }

  return "$cmd $argstr";
}

# ----------------------------------------------------------------------

=head2 run_command

Format and execute the given command with the given arguments.  See
format_command for examples of the input format.  The results of the call
are returned if the command executed successfully (parsed command line
and sent the RPC successfully).  If the command could not be executed,
this method will die() with an appropriate error message.

You must still test the output for success or failure of the ZAPI call itself.
The output of this function is the raw XML text returned from zsmcli.  You
may pass it into get_values() in this module to extract data from it or
use your own XML parsing to get at the values (eg. XML::Simple).

If a node name is given as the second argument, then the server's execution
server is set to that node before the command is run.  The execution server
is NOT reverted after the command.

B<Arguments>

=over

=item cmd_hash

I<Required> : Hash reference with the command to format and run.

=item node_name

I<Optional> : Node name to switch to before running the command.

=back

=for html <hr>

=cut

sub run_command {
  my $self = shift;
  my $cmd_hash_ref = shift;
  my $node_str = shift;
  my $server = $self->{SERVER};

  if (defined($node_str)) {
    $main::Log->comment("Switching to node $node_str before running zsmcli command.");
    $server->set_execution_server(-Server => $node_str);
  }

  my $cmd = format_command($cmd_hash_ref);
  my $cmdline = qq(zsmcli $cmd);

  $main::Log->debug("Running command '$cmdline'");

  my $output = $server->shell(cmd => $cmdline);

  my $XML_obj = new XML::Simple;
  my $data = $XML_obj->XMLin($output);

  if ($data->{"status"} eq "failed") {
    $main::Log->error(die("Failed to execute command."));
  }

  return $output;
}


# ----------------------------------------------------------------------

=head2 run_command_everywhere

Run the given command on the stored server object (on all nodes) and
return all the data.  See run_command() for more details.  Returns an
array of XML output, one entry per node where the command was run.

B<Arguments>

=over

=item cmd_hash

I<Required> : Hash reference of the command to format and run.

=back

=for html <hr>

=cut

sub run_command_everywhere {
  my $self = shift;
  my $cmd_hash_ref = shift;
  my $server = $self->{SERVER};
  my @arr;

  foreach my $node ($server->servers) {
    my $data = $self->run_command($node);
    push @arr, $data;
  }

  return \@arr;
}

# ----------------------------------------------------------------------

=head2 get_values

Get values from the given XML text and return the values to the caller.
The desired values are specified using the XPath standard format.  This
function will return an array of results.

Introduction to XPath: http://www.w3schools.com/XPath/xpath_intro.asp

XPath syntax: http://www.w3schools.com/XPath/xpath_syntax.asp

B<Arguments>

=over

=item xml

I<Required> : The XML to search for values in.

=item xpath

I<Required> : The XPath to use when searching the given XML.

=back

B<Examples>

    # Say the following XML is inside $xml:
    <results status="passed">
       <volume-attrs>
          <d-volume-info>
             <creation-time>1250182887</creation-time>
             <dsid>0x000000000000065d</dsid>
             <mirror-types>
                 <d-mirror-types-info>
                        <is-data-protection>false</is-data-protection>
                        <is-load-sharing>true</is-load-sharing>
                        <is-move>false</is-move>
                        <transfer-in-progress>false</transfer-in-progress>
                 </d-mirror-types-info>
             </mirror-types>
          </d-volume-info>
       </volume-attrs>
    </results>

    # To get the dsid value (that is the text between the dsid tags),
    # you can pass any of these XPath values.  "//" is like a wildcard,
    # matching any number of elements before or between the given tags.

    my $dsid = zsmcli::get_values($xml,
                                  "/results/volume-attrs/d-volume-info/dsid");
    my $dsid = zsmcli::get_values($xml, "d-volume-info//dsid");
    my $dsid = zsmcli::get_values($xml, "//dsid");

    # To get the transfer-in-progress flag you can use this:

    my $in_xfer = zsmcli::get_values($xml, "//transfer-in-progress");

    # If you have two tags that appear in different sub-trees, you have
    # to disambiguate them.  For example:
    <foo>
      <bar>
        <thingy>pencil</thingy>
      </bar>
      <baz>
        <thingy>eraser</thingy>
      </baz>
    </foo>

    # This will get you both pencil and eraser.
    my $thingy = zsmcli::get_values($xml, "//thingy");

    # This will get only pencil.
    my $thingy = zsmcli::get_values($xml, "bar//thingy");

    # This will get only eraser.
    my $thingy = zsmcli::get_values($xml, "baz//thingy");


=for html <hr>

=cut

sub get_values {
  my $xml = shift;
  my $path_txt = shift;
  my $path = XML::XPath->new(xml => $xml);

  $main::Log->debug("Looking for nodes in path $path_txt.");
  $main::Log->debug("XML is $xml");

  my $nodeset = $path->find($path_txt);

  my @values;

  $main::Log->debug("Got ". scalar($nodeset->get_nodelist)." result nodes.");

  foreach my $node ($nodeset->get_nodelist) {
    push @values, $node->string_value;
  }

  return @values;
}

# ----------------------------------------------------------------------

=head2 get_transfer_in_progress

Get the values of the transfer-in-progress flags as a hash, indexed by the
member volume name.  This is a convenience method and such information could
be retrieved using get_values above.

B<Arguments>

=over

=item xml

I<Required> : The XML to search for values in.

=back

=for html <hr>

=cut

sub get_transfer_in_progress {
  my $xml = shift;
  my $path = XML::XPath->new(xml => $xml);

  my $nodeset = $path->find("//volume-attrs/d-volume-info");

  my %values;

  foreach my $node ($nodeset->get_nodelist) {
    my $name_nodeset = $path->find("./name", $node);
    my $name = $name_nodeset->string_value;

    my $xfer_nodeset =
      $path->find("./mirror-types/d-mirror-types-info/transfer-in-progress",
		  $node);
    my $xfer = $xfer_nodeset->string_value;

    $values{$name} = $xfer;
  }

  return \%values;
}

sub _slurp_xml {
    my ($output_ref, $lines_ref, $lineno_ref) = @_;

    my $line;
    do {
        $line = $lines_ref->[$$lineno_ref];
        # Remove leading timestamp, etc.
        $line =~ s/^.*?\]\s//;
        push(@$output_ref, $line);
        $$lineno_ref++;
    # Slurp until the outer-most closing XML tag is found
    } while ($line !~ /^<\//);
}

=pod

=head2 extract_zapi_traces

Extracts traced ZAPIs from mgwd.log text.

B<Arguments>

=over

=item log

I<Required> : A fragment of the mgwd.log.

=back

B<Return Value>

=over

=item

A list of hash references, one for each ZAPI extracted from I<log>.  Each hash
reference has two keys: 'request', the XML text of the ZAPI call and 'response'
the XML text of the ZAPI result.

=back

=for html <hr />

=cut

sub extract_zapi_traces {
    my ($log,) = @_;
    my (@request, @response, @result);

    my @lines = split(/\n/, $log);
    my $lineno = 0;
    while ($lineno < scalar(@lines)) {
        my $line = $lines[$lineno];
        if ($line =~ /ZapiInterface:.*Request:/) {
            # Extract the ZAPI request XML
            scalar(@request) == 0 or die 'mismatched ZAPI reqeusts/responses';
            $line =~ s/^.*Request:\s+//;
            push(@request, $line);
            $lineno++;
            _slurp_xml(\@request, \@lines, \$lineno);
        } elsif ($line =~ /ZapiInterface:.*Response:/) {
            # Extract the ZAPI response XML
            scalar(@request) > 0 or die 'mismatched ZAPI reqeusts/responses';
            $line =~ s/^.*Response:\s+//;
            push(@response, $line);
            $lineno++;
            _slurp_xml(\@response, \@lines, \$lineno);

            push(@result, {request  => join("\n", @request),
                           response => join("\n", @response)});
            undef(@request);
            undef(@response);
        } else {
            $lineno++;
        }
    }

    return @result;
}

=pod

=head2 equivalent_numbers

Compares two numbers, treating undef as 0.

B<Arguments>

=over

=item n1

I<Required> : The first number.

=item n2

I<Required> : The second number.

=back

B<Return Value>

=over

=item

True if the numbers are equivalent, false otherwise.

=back

=for html <hr />

=cut

sub equivalent_numbers {
    my ($n1, $n2) = @_;
    if (!defined($n1)) {
        $n1 = 0;
    }
    if (!defined($n2)) {
        $n2 = 0;
    }
    return ($n1 == $n2);
}

=pod

=head2 equivalent_strings

Compares two strings, treating undef as '' (the empty string).

B<Arguments>

=over

=item s1

I<Required> : The first string.

=item s2

I<Required> : The second string.

=back

B<Return Value>

=over

=item 

True if the strings are equivalent, false otherwise.

=back

=for html <hr />

=cut

sub equivalent_strings {
    my ($s1, $s2) = @_;
    if (!defined($s1)) {
        $s1 = '';
    }
    if (!defined($s2)) {
        $s2 = '';
    }
    return ($s1 eq $s2);
}


1;
