# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/dp/snapmirror/qsm/TargetedData.pm#2 $
#
## @summary TargetedData - Targeted data manipulation for QSM tests
## @author	dl-dpgqa-automation@netapp.com
## @status	private
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

TargetedData - Targeted data manipulation for QSM tests

=head1 SYNOPSIS

  use dp::snapmirror::qsm::TargetedData;

  $data_obj = dp::snapmirror::qsm::TargetedData->new(path => 'Z:\temp',
                                                     client => $client);

  # Create 3 directories with 10 files and 4 streams in each directory
  $data->create_dirs(number  => 3,
                     files   => { number=>10 },
                     streams => { number=>4 });

  # Create 6 files with 10 streams in each file
  $data->create_files(number  => 6,
                      streams => { number=>10 });

  # Create 20 streams on current directory with random sizes
  # between 0 and 1000
  $data->create_streams(number => 20, args=>{max_size=>1000});

  # Truncate to zero 50% of all plain files (no streams attached):
  $data_obj->modify(files => { percent=>50, streams=>'no',
                               mode=>'trunc', args=>{size=>0} });

  # Remove 100 streams
  $data_obj->modify(streams => { number=>100, mode=>'remove' });

  # Modify 5 directories with files in them and with or without streams
  # to create 10 more files in each directory.
  # Also modify 20 files without streams to add 2 streams on each file.
  $data_obj->modify(dirs => { number=>5, streams=>'any', files=>'yes',
                              mode=>'create', args=>{files=>{number=>10}} },
                    files => { number=>20, streams=>'no',
                               mode=>'create', args=>{streams=>{number=>2}} });

  # Read files in the populated tree
  $data_obj->read_tree(files => { number=>5, skip=>3, compress=>'yes',
                                  encrypt=>'yes', archive=>'yes',
                                  hidden=>'any', acl=>'no', streams=>'no',
                                  sparse=>'no' });

  # Read stream in the populated tree
  $data_obj->read_tree(streams => { number=>100, compress=>'any',
                                    encrypt=>'any', archive=>'no',
                                    hidden=>'any', acl=>'no' });

  # Compare the contents of the directory tree of the object with the
  # directory tree given in the argument.
  $data_obj->compare(path => $path, contents => 1);

  # Return the hosttype of the client.
  $data_obj->hosttype();

  # Log a warning.
  $data_obj->logWarn($msg);

  # Evaluate code on remote client
  $out = $data_obj->eval(cmd => 'logcomment("Log comment on remote client")');

  # Get the SID for Administrators group on the LAB domain
  $sid = $data_obj->sid(user => 'LAB\Administrators');

  # Get the ACL arguments for an ACL of size 4k using one user
  $args_h = $data_obj->acl_size(size => 4096, users => ['Users']);

  # Set attribute of all elements back to original
  $out = $data_obj->set_original();

=head1 DESCRIPTION

Provides an object interface for targeted data manipulation for QSM tests.
It creates directories, files, streams, links, symbolic links, sparse files
and provides a mechanism for modifying a very specific type of object as well
as how that object is to be modified. On Windows clients, it also creates
files with ACLs as well as setting the compress, encrypt, archive or hidden
attributes on files or directories.

The options size, size_dev, max_size, offset and offset_dev accept unit
modifiers (k, m, g, t). These modifiers can be lower or upper case with
an optional trailing 'b', e.g. '12k', '30 MB'.

=head1 STATUS

  Private

=head1 AUTHOR

  dl-dpgqa-automation@netapp.com
  $Id: //depot/prod/test/Rsierranevada/dp/snapmirror/qsm/TargetedData.pm#2 $
  Last revised $Date: 2013/04/24 $

=head1 COPYRIGHT

  Copyright 2006, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package dp::snapmirror::qsm::TargetedData;

use strict;
use warnings;

use Carp qw(croak);
use Cwd qw(getcwd);
use File::Spec qw();
use File::Compare qw();
use Storable qw(nfreeze thaw);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF);
use Fcntl qw(O_RDONLY O_WRONLY O_BINARY O_CREAT O_TRUNC O_APPEND);
use FileManip qw(write_stream remove_file remove_stream setattr_file stat_stream
                 setattr_stream write_sparse link_file read_file read_stream
                 getattr_file vol_free);

# Check if library is running in NATE environment or standalone
eval { require Tharn };
my $_use_nate = $@ ? 0 : 1;

if ( $_use_nate ) {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
}

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub new($%);
sub _do_call($$%);
sub eval($%);
sub hosttype($;);
sub logWarn($$;);
sub create_dirs($%);
sub _validate_dir_opts($;);
sub _create_dirs($%);
sub create_files($%);
sub _validate_file_opts($;);
sub _create_files($%);
sub create_streams($%);
sub _create_streams($%);
sub set_original($;);
sub _set_original($;$);
sub counters($;%);
sub _counters($;%);
sub sid($%);
sub _sid($%);
sub acl_size($%);
sub modify($%);
sub _modify($%);
sub read_tree($%);
sub _read_tree($%);
sub compare($%);
sub _compare($%);
sub _compare_objects($$$$$;);
sub _compare_object($$$$$$;);
sub _process_objects($$$$$;);
sub _process_object($$$$$$;);
sub _process_size_units($;);
sub _set_on_pct($$;);
sub _get_args($;);
sub _modify_acl($$$;$$);
sub _read_acl($$;);
sub _get_from_list(\@$$;);
sub _get_objects($$$$;$);
sub _get_totals($$;$);
sub _copy_ref($$;);
sub _create_rand_data(\$$;);
sub _create_sparse($$$;);
sub _set_attributes($$$$;);
sub _get_ref($$;);
sub _get_type($;);
sub _get_file($$;);
sub _units($;);
sub _chdir($$;);
sub _chdir_restore($;);
sub _save_db($;);
sub _get_db($;);
sub logVerbose($$;);
sub intRandGauss($$;);
sub randGauss($$;);
sub DESTROY;

#---------------------------------------------------------------------
# Package variables
#---------------------------------------------------------------------
my $_seed;
my %acls;
my @val_char;
my $cifs_sid      = 1;
my $cifs_security = 1;
my $MAX_NUM       = 99999999999;
my $DATA_BLOCK    = 100;
my $MAX_WR_BLOCKS = 1000;
my $MAX_WR_SIZE   = $MAX_WR_BLOCKS*$DATA_BLOCK;
my $DB_FILE       = '.db';
my @ace_masks = qw(
    FILE_READ_DATA FILE_LIST_DIRECTORY FILE_WRITE_DATA FILE_ADD_FILE
    FILE_APPEND_DATA FILE_ADD_SUBDIRECTORY FILE_CREATE_PIPE_INSTANCE
    FILE_READ_EA FILE_WRITE_EA FILE_EXECUTE FILE_TRAVERSE FILE_DELETE_CHILD
    FILE_READ_ATTRIBUTES FILE_WRITE_ATTRIBUTES FILE_ALL_ACCESS
    FILE_GENERIC_READ FILE_GENERIC_WRITE FILE_GENERIC_EXECUTE
);
my @ace_flags = qw(
    CI OI IO NP
);
my @ace_types = qw(
    ACCESS_ALLOWED_ACE_TYPE ACCESS_DENIED_ACE_TYPE
);
my @ace_users = (
    'Users',
    'Guests',
    'Power Users',
    'Administrators',
    'Backup Operators',
);
my %pct_ret = (
    sparse_pct => 0,
    link_pct   => 0,
    slink_pct  => 0,
);
my %units_table = (
    b => 1,
    k => 1024,
    m => 1024*1024,
    g => 1024*1024*1024,
    t => 1024*1024*1024*1024,
);
my @counter_types = qw(create remove rename trunc copy modify sattrs read);
my @counter_otypes = qw(dirs files streams sparse links slinks acls);
my %invalid_filenames = (
    '.'         => 1, # Current directory
    '..'        => 1, # Parent directory
    '.db'       => 1, # Database file for this library
    '.snapshot' => 1, # Snapshot directory on filers
    'nul'       => 1, # The following are invalid file names in windows
    'prn'       => 1,
    'aux'       => 1,
    'con'       => 1,
    'com1'      => 1,
    'com2'      => 1,
    'com3'      => 1,
    'com4'      => 1,
    'com5'      => 1,
    'com6'      => 1,
    'com7'      => 1,
    'com8'      => 1,
    'com9'      => 1,
    'lpt1'      => 1,
    'lpt2'      => 1,
    'lpt3'      => 1,
    'lpt4'      => 1,
    'lpt5'      => 1,
    'lpt6'      => 1,
    'lpt7'      => 1,
    'lpt8'      => 1,
    'lpt9'      => 1,
);

#---------------------------------------------------------------------
# Constants
#---------------------------------------------------------------------
use constant VERBOSE_DIRS    => 0x01;
use constant VERBOSE_FILES   => 0x02;
use constant VERBOSE_STREAMS => 0x04;
use constant VERBOSE_SPARSE  => 0x08;
use constant VERBOSE_ATTRS   => 0x10;
use constant VERBOSE_INFO    => 0x80;
use constant VERBOSE_ALL     => 0xFF;

#---------------------------------------------------------------------

=head2 new

  $data_obj = dp::snapmirror::qsm::TargetedData->new(path => $path
                                                     [, client => $client ]
                                                     [, seed => $seed ]
                                                     [, save => 0|1 ]);

QSM targeted data manipulation object constructor.
The seed for the random number generator can be specified with the
main parameter DATA_SEED or the seed option.

The following named arguments are accepted:

=over

=item Arguments

=over

=item path

(Required) : Objects are created and modified relative to this path.

=item client

(Optional) : Hostspec of client where to perform all operations.
Default is localhost.

=item save

(Optional) : Load the database file of a previously created data into memory.
It saves the data into disk when the object is destroyed.
Default is 0.

=item create

(Optional) : If true, creates the top level directory if it does not exist,
otherwise fails if it does not exist.
Default is 1.

=item seed

(Optional) : Use this seed to initialize the random number generator.
This has the highest precedence and should only be used if the caller
is managing the seed.

=item fillpct

(Optional) : The percentange of space usage that will not be exceeded.
Only supported on Windows. Default is 100.

=item timeout

(Optional) : Timeout for function execution.
Default is 600.

=item dbfile

(Optional) : Unique database file name to allow multiple clients to 
work in the same directory.

=back

=item Returns

TargetedData object.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub new($%) {
    my $class = shift;
    my (%opts) = validate (
        @_,
        {
            path    => { type => SCALAR },
            client  => { type => SCALAR,  default => 'localhost' },
            save    => { type => SCALAR,  default => 0 },
            create  => { type => SCALAR,  default => 1 },
            seed    => { type => SCALAR, optional => 1 },
            fillpct => { type => SCALAR,  default => 100 },
            timeout => { type => SCALAR,  default => 600 },
            dbfile  => { type => SCALAR, default => ""},
        }
    );
    my ($self, $cmd, $fh, %init, $msg, $type, $otype, %free);
    my $path = $opts{path};
    my $save = $opts{save};
    my $create = $opts{create};
    my $fillpct = $opts{fillpct};
    my $dbfile = File::Spec->catfile($opts{path}, $DB_FILE);

    # Save owner of this object
    $opts{_pid} = $$;

    if ( $_use_nate ) {
        # Get hosttype of client
        $fh = Tharn::host($opts{client});
        $opts{_hosttype} = $fh->hosttype();
    } else {
        if ( $^O eq 'MSWin32' ) {
            $opts{_hosttype} = 'windows';
        } else {
            $opts{_hosttype} = "unix-$^O";
        }
    }

    if ( $opts{dbfile} ne "" ) {
        $dbfile = File::Spec->catfile($opts{path},
                                      $opts{dbfile});
    }
    $opts{db_file} = $dbfile;
    
    if ( $opts{client} eq 'localhost' ) {
        # Clean up path (canonical form)
        $opts{path} = File::Spec->canonpath($opts{path});

        $opts{d_used}   = 0;
        $opts{_fillpct} = 0;
        $opts{_modcnt}  = 0;
        $opts{_mcdiff}  = 10;

        if ( $opts{_hosttype} eq 'windows' ) {
            # For Windows, use only lower case letters for file names
            @val_char = ('a'..'z', 0..9, '_');
            # Append a '\' if a drive letter
            $opts{path} .= '\\' if $opts{path} =~ /^.\:$/;

            # Set flag if need to check space
            $opts{_fillpct} = 1 if $opts{fillpct} < 100;

            # Save path for vol_free
            $opts{d_path} = $opts{path};
            if ( $opts{d_path} !~ /^.\:/ ) {
                # Path is relative, need to know which volume this is
                $opts{d_path} = File::Spec->rel2abs($opts{d_path});
            }

            if ( $opts{_fillpct} ) {
                # Get total space
                %free = vol_free(vol => $opts{d_path});

                # Calculate maximum usage space the check for
                $opts{d_total} = $free{total};
                $opts{d_used}  = $free{used};
                $opts{d_max}   = ($opts{fillpct}/100) * $free{total};
            }
        } else {
            @val_char = ('A'..'Z', 'a'..'z', 0..9, '_');
        }

        # Read object from db file
        if ( $save && -e $dbfile ) {
            # Bless object
            $self = bless { %opts }, $class;

            # Read database of directory
            $self->_get_db();
        }

        # Initialize random seed
        if ( defined($opts{seed}) ) {
            $opts{_srand} = $opts{seed};
        } elsif ( defined($_seed) ) {
            $opts{_srand} = $_seed;
        } elsif ( defined($main::DATA_SEED) ) {
            $opts{_srand} = $main::DATA_SEED;
        } else {
            do {
                $opts{_srand} = time() ^ ($$ + ($$ << 15));
            } while (!$opts{_srand});
        }

        # Log the seed to reproduce same results in future runs
        $msg = "DATA_SEED: $opts{_srand}";
        $_use_nate ? Tharn::logcomment($msg) : print "$msg\n";
        srand($opts{_srand});
        $_seed = $opts{_srand} + 1;

        # Return blessed object read from file
        if ( defined($self) ) {
            $self->_counters(clear => 1);
            return $self;
        }

        if ( ! -e $opts{path} ) {
            if ( $create ) {
                # Create new directory
                mkdir $opts{path}
                  or croak "Could not create directory $opts{path}: $!";
            } else {
                croak "Directory $path does not exist";
            }
        }

        # Initialize values
        %init = (
            _mod_id  => 0,
            _mod_id  => 0,
            _remove  => {},
            _objects => { '.' => { -type => 'd', -mod_id => 0 },
                          -type => 'd' },
        );

        # Initialize counters
        foreach $type ( @counter_types ) {
            foreach $otype ( @counter_otypes ) {
                $init{_counters}{$type}{$otype} = 0;
            }
        }
    } else {
        # Change to object's root directory
        # before calling any of the functions
        $opts{_remote_chdir} = 1;

        # Start RPC service on the client
        $cmd = qq{\{
            use dp::snapmirror::qsm::TargetedData;
            eval \{
                our \$obj = dp::snapmirror::qsm::TargetedData->new(
                    path=>q($path),
                    save=>q($save),
                    create=>q($create),
                    fillpct=>q($fillpct),
                );
            \};
            Subtest::ipc->write_message(\$@);
            die \$@ if \$@;
            Subtest::ipc->serve();
            exit(0);
        \}};
        $opts{_sobj} = Tharn::subtest($cmd, '-bg', '-ipc',
                                      '-client', $opts{client},
                                      '-runid', 'TargetedData',
                                      '-onexit', sub{});

        # Check for initialization errors
        $msg = $opts{_sobj}->ipc->read_message();
        if ($msg) {
            $opts{_sobj}->wait_finish();
            croak $msg;
        }
    }

    # Return object
    return bless { %opts, %init }, $class;
}

#---------------------------------------------------------------------
# Function to execute the calling function locally or remotely
#---------------------------------------------------------------------
sub _do_call($$%) {
    my ($self, $function, %opts) = @_;
    my ($func_name, $out, $err, @results);

    # Set alarm handler
    local $SIG{ALRM} = sub { croak "Timeout executing function with timeout=$self->{timeout}" };

    # Set alarm for timeout
    alarm($self->{timeout}) if defined $self->{timeout};

    if ( $self->{client} eq 'localhost' ) {
        # Change to object's root directory
        $self->_chdir('');

        # Call local function
        $out = eval { $self->$function(%opts) };
        $err = $@;

        # Restore previous directory
        eval {
            splice(@{$self->{_prev_cwd}}, 1) if @{$self->{_prev_cwd}} > 1;
            $self->_chdir_restore();
            undef $self->{_cwd};
        };
        alarm(0);
        croak $err if $err;
        croak $@ if $@;
    } else {
        # Get the name of the function to call
        # on the remote client
        $func_name = (caller 1)[3];
        $func_name =~ s/.*::/_/;

        if ( $self->{_remote_chdir} ) {
            # Change to object's root directory
            # Do it only once for the life of the object since
            # it is in a different process, no need to restore
            # to the previous directory when done
            $out = $self->{_sobj}->ipc->eval('$::obj->_chdir("")');

            # Check for any errors
            if ( $@ =~ /socket read failed/ ) {
                @results = $self->{_sobj}->own_results;
                $err = NATE::Result::worst(@results);
                if ( $err && $err->severity > 2 ) {
                    alarm(0);
                    croak join("\n", $err->messages);
                }
            }
            if ( $@ ) {
                alarm(0);
                croak $@;
            }
            $self->{_remote_chdir} = 0;
        }

        # Call proper function on remote client
        $out = $self->{_sobj}->ipc->call("\$::obj->$func_name", %opts);
        alarm(0);
        # Check for any errors
        if ( $@ =~ /socket read failed/ ) {
            @results = $self->{_sobj}->own_results;
            $err = NATE::Result::worst(@results);
            croak join("\n", $err->messages) if $err && $err->severity > 2;
        }
        croak $@ if $@;
    }

    return $out;
}

#---------------------------------------------------------------------

=head2 eval

  $out = $data_obj->eval(cmd => $cmd);

Eval the code on the remote client.

The following named arguments are accepted:

=over

=item Arguments

=over

=item cmd

(Required) : Code to be evaluated on the remote client.

=back

=item Returns

Return value of evaluated code.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub eval($%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            cmd => { type => SCALAR },
        }
    );

    croak "Function can only be called for a remote client"
      if not defined($self->{_sobj});

    return $self->{_sobj}->ipc->eval($opts{cmd});
}

#---------------------------------------------------------------------

=head2 hosttype

  $out = $data_obj->hosttype();

Returns the hosttype of the client.

=over

=item Returns

Return hosttype of client.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub hosttype($;) {
    my $self = shift;

    return $self->{_hosttype};
}

#---------------------------------------------------------------------

=head2 logWarn

  $out = $data_obj->logWarn($msg);

Logs a warning.

=over

=item Returns

None.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub logWarn($$;) {
    my (undef, $msg) = @_;
    $_use_nate ? Tharn::logresult('WARN', $msg) : Carp::carp($msg);
    return;
}

#---------------------------------------------------------------------

=head2 create_dirs

  $out = $data_obj->create_dirs([ dir => $rel_path ]
                                [, number => $ndirs ]
                                [, dev => $ndirs_dev ]
                                [, depth_avg => $depth_avg ]
                                [, depth_dev => $depth_dev ]
                                [, files => \%file_opts ]
                                [, streams => \%stream_opts ]
                                [, args => \%args ]);

Create a number of directories in the specified directory.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dir

(Optional) : Directories are created under this directory.
Path is relative to the object's root path.
Default is the object's current working directory.

=item number

(Optional) : Number of directories to create.
Default is 1.

=item dev

(Optional) : Standard deviation of the number of directories to create.
Default is 0.

=item depth_avg

(Optional) : Average depth to create.
Default is 0.

=item depth_dev

(Optional) : Standard deviation of depth to create.
Default is 0.

=item files

(Optional) : Reference to a hash of options to create files under each
newly created directory. See create_files for options.

=item streams

(Optional) : Reference to a hash of options to create streams on each
newly created directory. See create_streams for options.

=item args

(Optional) : Reference to a hash of options on how to create each directory.

  {
      number   => Maximum number of the specified objects to modify.
      percent  => Modifies this percentage of the specified objects,

      name_size     => Size of directory name.
      name_size_dev => Standard deviation of directory name.
      max_name_size => Size of directory name is a random number between 0
                       and this value (default is 0).
      name_reuse    => Reuse name of previously removed directory
                       (default is 0).
      acl_num       => Maximum number of directories to add an ACL.
      stream_num    => Maximum number of directories to add streams.
      compress_num  => Maximum number of directories to set/clear compress attr.
      encrypt_num   => Maximum number of directories to set/clear encrypt attr.
      archive_num   => Maximum number of directories to set/clear archive attr.
      hidden_num    => Maximum number of directories to set/clear hidden attr.
      acl_pct       => Percentage of directories to add an ACL.
      stream_pct    => Percentage of directories to add streams.
      compress_pct  => Percentage of directories to set/clear compress attr.
      encrypt_pct   => Percentage of directories to set/clear encrypt attr.
      archive_pct   => Percentage of directories to set/clear archive attr.
      hidden_pct    => Percentage of directories to set/clear hidden attr.
      compress      => Set or clear the compress attribute.
      encrypt       => Set or clear the encrypt attribute.
      archive       => Set or clear the archive attribute.
      hidden        => Set or clear the hidden attribute.
      acl_masks     => Use a maximum of this number of access mask flags (0-18)
                       (default is 8; 0 means use maximum available).
      acl_flags     => Use a maximum of this number of flags (0-6)
                       (default is 4; 0 means use maximum available).
      acl_types     => Use a maximum of this number of types (0-3)
                       (default is 1; 0 means use maximum available).
      acl_users     => Use a maximum of this number of users (0-7)
                       (default is 2; 0 means use maximum available).
      users         => List of users/groups to use instead of the defaults.
      acl           => Create directory with an ACL (1 or 2).
                       1: the number of ACEs varies between 1 and the product
                          of the acl_* options
                       2: the number of ACEs is the product of the
                          acl_* options (fixed).
  }

=back

=item Returns

1 if succeeds.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub create_dirs($%) {
    my ($self, %opts) = @_;

    _validate_dir_opts(\%opts);

    return $self->_do_call(\&_create_dirs, %opts);
}

#---------------------------------------------------------------------
# Internal function
# Validate the dir options
#---------------------------------------------------------------------
sub _validate_dir_opts($;) {
    my $opts = shift;

    _validate_file_opts($opts->{files});
    _process_size_units($opts->{streams}{args});

    return;
}

#---------------------------------------------------------------------
# Function to create directories and files with streams
# It runs on the local machine
#---------------------------------------------------------------------
sub _create_dirs($%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            dir       => { type => SCALAR,  default => '' },
            number    => { type => SCALAR,  default => 1 },
            dev       => { type => SCALAR,  default => 0 },
            depth_avg => { type => SCALAR,  default => 0 },
            depth_dev => { type => SCALAR,  default => 0 },
            files     => { type => HASHREF, default => {} },
            streams   => { type => HASHREF, default => {} },
            args      => { type => HASHREF, default => {} },
            # Internal parameter -- current depth
            depth     => { type => SCALAR,  default => 1 },
        }
    );
    my ($ref, $args, $number, $depth, %free);

    # Get database reference of directory where
    # the directories are to be created
    $opts{dir} =~ s/^[\/\\]+//;
    $ref = $self->_get_ref($opts{dir});

    # Get number of directories to create
    $number = intRandGauss($opts{number}, $opts{dev});
    return 0 if $number < 1;

    # Check fill percentage
    if ( $self->{_fillpct} ) {
        # Get used space on volume
        %free = vol_free(vol => $self->{d_path});
        # Set current space usage and return if limit is hit
        $self->{d_used} = $free{used};
        return 0 if $self->{d_used} > $self->{d_max};
    }

    # Change to directory
    $self->_chdir($opts{dir});

    logVerbose(VERBOSE_DIRS, "Creating $number directories on " . getcwd);

    # Set default value for maximum number of streams/attr/acls to create
    $opts{args}{stream_num} = $MAX_NUM unless exists $opts{args}{stream_num};
    $opts{args}{attr_num}   = $MAX_NUM unless exists $opts{args}{attr_num};
    $opts{args}{acl_num}    = $MAX_NUM unless exists $opts{args}{acl_num};

    foreach ( 1..$number ) {
        my $entry;
        # Get arguments on how to create the directories
        $opts{args}{_rmlist} = $self->{_remove}{dirs};
        $opts{args}{_ref}    = $ref;
        $args = _get_args($opts{args});
        $entry = $args->{name};

        if ( $self->{_fillpct} && $self->{d_used} > $self->{d_max} ) {
            # Done populating
            last;
        }

        # Create new directory
        logVerbose(VERBOSE_DIRS, "Create directory $entry");
        mkdir $entry or croak "Could not create directory $entry: $!";
        $self->{_counters}{create}{dirs}++;

        # Add directory to database
        $ref->{$entry} = { -type => 'd', -mod_id => $self->{_mod_id} };
        # Increment the number of directories owned by parent directory
        $ref->{-dirs}++;

        # If parent directory is encrypted or compressed
        # this directory is also encrypted or compressed
        $ref->{$entry}{-encrypt}  = 1 if $ref->{-encrypt};
        $ref->{$entry}{-compress} = 1 if $ref->{-compress};

        # Set attributes on the directory
        if ( %{$args->{sattrs}} ) {
            _set_attributes($entry, $args->{sattrs}, $ref, $opts{args});
        }

        # Create files on this directory
        if ( $opts{files}{number} ) {
            $self->_create_files(dir=>$entry, %{$opts{files}});
        }

        # Create streams on this directory
        if ( $args->{streams} && $opts{streams}{number}
          && ( $opts{args}{stream_num} > 0 )) {
            $self->_create_streams(file=>$entry, %{$opts{streams}});
            $opts{args}{stream_num}--;
        }

        $depth = intRandGauss($opts{depth_avg}, $opts{depth_dev});
        if ( $depth > $opts{depth} ) {
            $self->_create_dirs(%opts, dir=>$entry, depth=>$opts{depth}+1);
        }

        # Set ACL on directory
        if ( $args->{acl} && ( $opts{args}{acl_num} > 0 ) ) {
            logVerbose(VERBOSE_ATTRS, "Set ACL on dir $entry");
            $ref->{$entry}{-acl} = $self->_modify_acl($entry, $args);
            $opts{args}{acl_num}--;
            $self->{_counters}{create}{acls}++;
        }
    }

    # Change back to previous directory
    $self->_chdir_restore();

    return $number;
}

#---------------------------------------------------------------------

=head2 create_files

  $out = $data_obj->create_files([ dir => $rel_path ]
                                 [, number => $nfiles ]
                                 [, dev => $nfiles_dev ]
                                 [, streams => \%stream_opts ]
                                 [, args => \%args ]);

Create a number of files in the specified directory.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dir

(Optional) : Files are created under this directory.
Path is relative to the object's root path.
Default is the object's current working directory.

=item number

(Optional) : Number of files to create.
Default is 1.

=item dev

(Optional) : Standard deviation of the number of files to create.
Default is 0.

=item streams

(Optional) : Reference to a hash of options to create streams on each
newly created file. See create_streams for options.

=item args

(Optional) : Reference to a hash of options on how to create each file.

  {
      size          => Size of file to create.
      size_dev      => Standard deviation of the size of file to create.
      max_size      => Size of file is a random number between 0 and
                       this value (default is 1000).
      name_size     => Size of file name.
      name_size_dev => Standard deviation of file name.
      max_name_size => Size of file name is a random number between 0
                       and this value (default is 0).
      name_reuse    => Reuse name of previously removed file (default is 0).
      acl_num       => Maximum number of files to add an ACL.
      stream_num    => Maximum number of files to add streams.
      sparse_num    => Maximum number of files to be sparse files.
      link_num      => Maximum number of files to be links.
      slink_num     => Maximum number of files to be symbolic links.
      compress_num  => Maximum number of files to set/clear compress attr.
      encrypt_num   => Maximum number of files to set/clear encrypt attr.
      archive_num   => Maximum number of files to set/clear archive attr.
      hidden_num    => Maximum number of files to set/clear hidden attr.
      acl_pct       => Percentage of files to add an ACL.
      stream_pct    => Percentage of files to add streams.
      sparse_pct    => Percentage of files to be sparse files.
      link_pct      => Percentage of files to be links.
      slink_pct     => Percentage of files to be symbolic links.
      compress_pct  => Percentage of files to set/clear compress attr.
      encrypt_pct   => Percentage of files to set/clear encrypt attr.
      archive_pct   => Percentage of files to set/clear archive attr.
      hidden_pct    => Percentage of files to set/clear hidden attr.
      compress      => Set or clear the compress attribute.
      encrypt       => Set or clear the encrypt attribute.
      archive       => Set or clear the archive attribute.
      hidden        => Set or clear the hidden attribute.
      sparse_opts   => Hash reference on how to create each sparse file.
      acl_masks     => Use a maximum of this number of access mask flags (0-18)
                       (default is 8; 0 means use maximum available).
      acl_flags     => Use a maximum of this number of flags (0-6)
                       (default is 4; 0 means use maximum available).
      acl_types     => Use a maximum of this number of types (0-3)
                       (default is 1; 0 means use maximum available).
      acl_users     => Use a maximum of this number of users (0-7)
                       (default is 2; 0 means use maximum available).
      users         => List of users/groups to use instead of the defaults.
      acl           => Create file with an ACL (1 or 2).
                       1: the number of ACEs varies between 1 and the product
                          of the acl_* options
                       2: the number of ACEs is the product of the
                          acl_* options (fixed).
  }

where sparse_opts contain the following:

  {
      ndata      => Number of data regions to create.
      ndata_dev  => Standard deviation of the number of data regions.
      offset     => Offset between data regions.
      offset_dev => Standard deviation of the offset between data regions.
      size       => Size of each data region.
      size_dev   => Standard deviation of the size of each data region.
  }

=back

=item Returns

1 if succeeds.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub create_files($%) {
    my ($self, %opts) = @_;

    _validate_file_opts(\%opts);

    return $self->_do_call(\&_create_files, %opts);
}

#---------------------------------------------------------------------
# Internal function
# Validate the file options
#---------------------------------------------------------------------
sub _validate_file_opts($;) {
    my $opts = shift;

    _process_size_units($opts->{args});
    _process_size_units($opts->{args}{sparse_opts});
    _process_size_units($opts->{streams}{args});

    return;
}

#---------------------------------------------------------------------
# Function to create files with streams
# It runs on the local machine
#---------------------------------------------------------------------
sub _create_files($%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            dir     => { type => SCALAR,  default => '' },
            number  => { type => SCALAR,  default => 1 },
            dev     => { type => SCALAR,  default => 0 },
            streams => { type => HASHREF, default => {} },
            args    => { type => HASHREF, default => {} },
        }
    );
    my ($ref, $args, $cont, $size, $name, $number);

    # Get database reference of directory where
    # the files are to be created
    $opts{dir} =~ s/^[\/\\]+//;
    $ref = $self->_get_ref($opts{dir});

    # Get number of files to create
    $number = intRandGauss($opts{number}, $opts{dev});
    return 0 if $number < 1;

    # Change to directory
    $self->_chdir($opts{dir});

    logVerbose(VERBOSE_DIRS, "Creating $number files on " . getcwd);

    # Set default value for maximum number of streams/sparse/attr/acls to create
    $opts{args}{stream_num} = $MAX_NUM unless exists $opts{args}{stream_num};
    $opts{args}{sparse_num} = $MAX_NUM unless exists $opts{args}{sparse_num};
    $opts{args}{attr_num}   = $MAX_NUM unless exists $opts{args}{attr_num};
    $opts{args}{acl_num}    = $MAX_NUM unless exists $opts{args}{acl_num};

    foreach ( 1..$number ) {
        my $entry;
        # Get arguments on how to create the files
        $opts{args}{_rmlist} = $self->{_remove}{files};
        $opts{args}{_ref}    = $ref;
        $args = _get_args($opts{args});
        $size = $args->{size};
        $entry = $args->{name};
        $self->{d_used} += $size;

        if ( $self->{_fillpct} && $self->{d_used} > $self->{d_max} ) {
            # Done populating
            last;
        }

        # If parent directory is encrypted or compressed
        # this file is also encrypted or compressed
        $ref->{$entry}{-encrypt}  = 1 if $ref->{-encrypt};
        $ref->{$entry}{-compress} = 1 if $ref->{-compress};

        if ( $args->{link} ) {
            logVerbose(VERBOSE_FILES, "Create " .
                       ($args->{link_data}{symbolic} ? 'symbolic ' : '') .
                       "link $entry -> $args->{link_data}{file}");
            link_file(flink=>$entry, %{$args->{link_data}});
            if ( $args->{link_data}{symbolic} ) {
                $ref->{$entry}{-link} = 2;
                $opts{args}{slink_num}--;
                $self->{_counters}{create}{slinks}++;
            } else {
                $ref->{$entry}{-link} = 1;
                $opts{args}{link_num}--;
                $self->{_counters}{create}{links}++;
            }
        } elsif ( $args->{sparse} ) {
            # Need to create sparse file before setting attributes
            logVerbose(VERBOSE_SPARSE, "Create sparse file $entry");
            _create_sparse($entry, $args, $ref);
            $opts{args}{sparse_num}--;
            $self->{_counters}{create}{sparse}++;
        } else {
            # Create new file
            logVerbose(VERBOSE_FILES, "Create file $entry");
            sysopen(FH, $entry, O_WRONLY|O_BINARY|O_CREAT|O_TRUNC)
              or croak "Could not create regular file $entry: $!";
        }

        if ( %{$args->{sattrs}} && ! $args->{link} ) {
            if ( ! $args->{sparse} ) {
                # Close the file to set the attributes
                close(FH);
            }

            # Set attributes on the file before writing to it
            # in order to compress/encrypt the file on the fly
            # instead of creating the file and then do the
            # compression/encryption
            _set_attributes($entry, $args->{sattrs}, $ref, $opts{args});

            if ( ! $args->{sparse} ) {
                # Open the file again
                sysopen(FH, $entry, O_WRONLY|O_BINARY)
                  or croak "Could not create sparse file $entry: $!";
            }
        }

        if ( ! $args->{sparse} && ! $args->{link} ) {
            # Write the file in blocks so any size of file can be created
            while ( $size > 0 ) {
                # Generate data block
                $size -= _create_rand_data($cont, $size);
                # Write current block to file
                print FH $cont or croak "Could not write to file $entry: $!";
            }
            close(FH);
            $self->{_counters}{create}{files}++;
        }

        # Add file type and mod_id
        $ref->{$entry}{-type}   = 'f';
        $ref->{$entry}{-mod_id} = $self->{_mod_id};

        # Increment the number of files owned by parent directory
        $ref->{-files}++;

        # Create streams on this file
        # Do not create streams if it is a link
        if ( ! $args->{link} && $args->{streams} && $opts{streams}{number}
          && ( $opts{args}{stream_num} > 0 ) ) {
            $self->_create_streams(file=>$entry, %{$opts{streams}});
            $opts{args}{stream_num}--;
        }

        # Set ACL on file
        if ( $args->{acl} && ( $opts{args}{acl_num} > 0 ) ) {
            logVerbose(VERBOSE_ATTRS, "Set ACL on file $entry");
            $ref->{$entry}{-acl} = $self->_modify_acl($entry, $args);
            $opts{args}{acl_num}--;
            $self->{_counters}{create}{acls}++;
        }
    }

    # Change back to previous directory
    $self->_chdir_restore();

    return $number;
}

#---------------------------------------------------------------------

=head2 create_streams

  $out = $data_obj->create_streams([ file => $file ]
                                   [, number => $nstreams ]
                                   [, dev => $nstreams_dev ]
                                   [, args => \%args ]);

Create a number of streams in the specified file/directory.

The following named arguments are accepted:

=over

=item Arguments

=over

=item file

(Optional) : The streams are attached to this file/directory.
Default is the object's current working directory.

=item number

(Optional) : Number of streams to create.
Default is 1.

=item dev

(Optional) : Standard deviation of the number of streams to create.
Default is 0.

=item args

(Optional) : Reference to a hash of options on how to create each stream.

  {
      size          => Size of stream to create.
      size_dev      => Standard deviation of the size of stream to create.
      max_size      => Size of stream is a random number between 0 and
                       this value (default is 1000).
      name_size     => Size of stream's name.
      name_size_dev => Standard deviation of stream's name.
      max_name_size => Size of stream's name is a random number between 0
                       and this value (default is 0).
      name_reuse    => Reuse name of previously removed stream (default is 0).
  }

=back

=item Returns

1 if succeeds.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub create_streams($%) {
    my ($self, %opts) = @_;

    _process_size_units($opts{args});

    return $self->_do_call(\&_create_streams, %opts);
}

#---------------------------------------------------------------------
# Function to create streams
# It runs on the local machine
#---------------------------------------------------------------------
sub _create_streams($%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            file   => { type => SCALAR,  default => '.' },
            number => { type => SCALAR,  default => 1 },
            dev    => { type => SCALAR,  default => 0 },
            args   => { type => HASHREF, default => {} },
        }
    );
    my ($ref, $args, $cont, $size, $append, $name, $number);

    # Get database reference of object where
    # the streams are to be created
    $opts{file} =~ s/^[\/\\]+//;
    $ref = $self->_get_ref($opts{file});

    # Get number of streams to create
    $number = intRandGauss($opts{number}, $opts{dev});
    return 0 if $number < 1;
    logVerbose(VERBOSE_FILES, "Creating $number streams on $opts{file}");

    foreach ( 1..$number ) {
        my $entry;
        # Get arguments on how to create the streams
        $opts{args}{_rmlist} = $self->{_remove}{streams};
        $opts{args}{_ref}    = $ref;
        $args = _get_args($opts{args});
        $size = $args->{size};
        $entry = $args->{name};
        $self->{d_used} += $size;

        if ( $self->{_fillpct} && $self->{d_used} > $self->{d_max} ) {
            # Done populating
            last;
        }

        logVerbose(VERBOSE_STREAMS, "Create stream $entry");

        # Create new stream
        $append = 0;
        # Write the stream in blocks so any size of stream can be created
        do {
            # Generate data block
            $size -= _create_rand_data($cont, $size);
            # Write current block to stream
            write_stream(file     => $opts{file},
                         stream   => $entry,
                         contents => $cont,
                         append   => $append);
            # Make sure next iteration the contents are appended
            $append = 1;
        } while ( $size > 0 );
        # Add stream to database
        $ref->{$entry} = { -type => 's', -mod_id => $self->{_mod_id} };
        # Increment the number of streams owned by parent object
        $ref->{-streams}++;
        $self->{_counters}{create}{streams}++;
    }

    return $number;
}

#---------------------------------------------------------------------

=head2 set_original

  $out = $data_obj->set_original();

Set attribute of all elements back to original.
It is used to group modifies into batches so the original flag can
be used to modify elements in a previous batch.

=over

=item Returns

1 if succeeds.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub set_original($;) { shift->_do_call(\&_set_original) }

#---------------------------------------------------------------------
# Set attribute of all elements to original (mod_id = 0)
# It runs on the local machine
#---------------------------------------------------------------------
sub _set_original($;$) {
    my ($self, $ref) = @_;
    $ref = $self->{_objects} if not defined($ref);

    foreach my $key ( keys %$ref ) {
        next if ref($ref->{$key}) ne 'HASH';
        if ( exists($ref->{$key}{-mod_id}) ) {
            $ref->{$key}{-mod_id} = 0;
        }

        # Process this object's children
        $self->_set_original($ref->{$key});
    }
}

#---------------------------------------------------------------------

=head2 counters

  $out = $data_obj->counters();

Get or clear the object counters. The counters are the number of an
specific object type which has been modified or created. The following
is a list of the different counters:

    total   => Totals (counter is not cleared when using the 'clear' option)
    create  => Create counters
    remove  => Remove counters
    rename  => Rename counters
    trunc   => Truncate counters
    copy    => Copy counters
    modify  => Modify counters
    sattrs  => Attributes set counters
    read    => Read counters

Each of the counters have the following specific object counters:

    dirs    => Number of directories
    files   => Number of files
    streams => Number of streams
    sparse  => Number of sparse files
    links   => Number of links
    slinks  => Number of symbolic links
    acls    => Number of ACL's

The following named arguments are accepted:

=over

=item Arguments

=over

=item clear

(Optional) : Clear all the counters except for the 'total' counter.
Defaults to 0.

=back

=item Returns

A hash reference of all the counters.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub counters($;%) { shift->_do_call(\&_counters, @_) }

#---------------------------------------------------------------------
# Get or clear the object counters
#---------------------------------------------------------------------
sub _counters($;%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            clear => { type => SCALAR, default => 0 },
        }
    );
    my (%ret, $ctype, $otype, %total);

    if ( $opts{clear} ) {
        # Clear all counters
        foreach $ctype ( keys %{$self->{_counters}} ) {
            foreach $otype ( keys %{$self->{_counters}{$ctype}} ) {
                $self->{_counters}{$ctype}{$otype} = 0;
            }
        }
        return;
    }

    # Get all the counters and calculate the total counters
    %ret = %{$self->{_counters}};
    $ret{total} = {
        # Default values
        ( map { $_ => 0 } @counter_otypes ),
        # Get totals and overwrite defaults
        %{$self->_get_totals($self->{_objects}{'.'}, \%total)},
    };

    return \%ret;
}

#---------------------------------------------------------------------

=head2 sid

  $sid = $data_obj->sid(user => $user);

Get the SID of the specified user.

The following named arguments are accepted:

=over

=item Arguments

=over

=item user

(Required) : The user/group to get the SID for. Could be just the username
or in the 'domain\username' format.

=back

=item Returns

SID of user.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub sid($%) { shift->_do_call(\&_sid, @_) }

#---------------------------------------------------------------------
# Get the SID for the user
# It runs on the local machine
#---------------------------------------------------------------------
sub _sid($%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            user => { type => SCALAR },
        }
    );

    croak "SID is only supported on Windows"
      if $self->{_hosttype} ne 'windows';

    if ( $cifs_sid ) {
        # Load only once
        require Win32::Security::SID;
        $cifs_sid = 0;
    }

    # Get the SID for the username
    return Win32::Security::SID::ConvertNameToSid($opts{user});
}

#---------------------------------------------------------------------

=head2 acl_size

  $out = $data_obj->acl_size(size => $size, users => \@users);

Calculate the arguments for ACL given the specified size. The size
of the ACL depends on the number of ACEs and the size of each ACE
depends on the SID of the user. A list of users is needed in order
to get the size of each SID (variable length) to calculate the
number of ACEs required for a given size ACL.

For small size ACL the list of users to use must be small while
for large ACLs the list of users to use must be larger to get
accurate results. For an ACL of less than 4k use one user and
for an ACL of 100k use at least 10 users/groups.

The following named arguments are accepted:

=over

=item Arguments

=over

=item size

(Required) : Calculate arguments for this size of ACL.

=item users

(Required) : Reference to an array of users to be used in the creation
of the ACL.

=back

=item Returns

A reference to a hash of arguments for creating or modifying ACLs.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub acl_size($%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            size  => { type => SCALAR },
            users => { type => ARRAYREF },
        }
    );
    my ($naces, $ace_size, $total, $sid, $m, $f, $t, $u);
    my $count = @{$opts{users}};

    croak "List of users is empty" if $count < 1;
    $opts{size} = _units($opts{size});

    # Calculate average size of ACE
    # The size of the ACE is 8 bytes for the header plus the SID
    $ace_size = 0;
    foreach $u ( @{$opts{users}} ) {
        $sid = $self->sid(user => $u);
        croak "Could not get SID for $u" if not defined $sid;
        $ace_size += length($sid);
    }
    # Using the average size of SIDs
    $ace_size = 8 + $ace_size/$count;

    # Calculate the number of ACEs needed for the specified size
    # The size of the ACL is 16 bytes for the header plus the ACEs
    # Get the number of ACEs for a 10% more in order
    # to compensate for the rounding down while calculating
    # the products
    $naces = int(0.5 + (1.1*$opts{size} - 16) / $ace_size);

    # Calculate the products to specified on the arguments to create ACL
    # Use brute force approach by going through all possible combinations
    foreach $t ( 1..@ace_types ) {
        foreach $f ( 1..@ace_flags ) {
            foreach $m ( 1..@ace_masks ) {
                $total = $m * $f * $t * $count;
                if ( $total > 0.9*$naces && $total < 1.1*$naces ) {
                    return { acl       => 2,
                             acl_masks => $m,
                             acl_flags => $f,
                             acl_types => $t,
                             acl_users => $count,
                             users     => $opts{users} };
                }
            }
        }
    }

    return;
}

#---------------------------------------------------------------------

=head2 modify

  $out = $data_obj->modify([ dirs => \%dir_opts ]
                           [, files => \%file_opts ]
                           [, streams => \%stream_opts ]);

Modify the populated tree according to the specified options.
Each of the main options could have any of the following options:

  {
      number   => Maximum number of the specified objects to modify.
      original => Only modify original objects created before a call
                  to this function.
      skip     => Number of the specified objects to skip before modifying.
      skiproot => Do not modify the root directory.
      percent  => Modifies this percentage of the specified objects,
                  it has precedence over the 'number' option.
      compress => Modify compressed files: yes|no|any.
                  yes: only modify compressed files.
                  no:  only modify uncompressed files.
                  any: modify files regardless of the compress attribute.
      encrypt  => Modify encrypted files: yes|no|any.
                  yes: only modify encrypted files.
                  no:  only modify decrypted files.
                  any: modify files regardless of the encrypt attribute.
      archive  => Modify archived files: yes|no|any.
                  yes: only modify archived files.
                  no:  only modify non-archived files.
                  any: modify files regardless of the archive attribute.
      hidden   => Modify hidden files: yes|no|any.
                  yes: only modify hidden files.
                  no:  only modify non-hidden files.
                  any: modify files regardless of the hidden attribute.
      acl      => Modify files with ACLs: yes|no|any.
                  yes: only modify files with ACLs.
                  no:  only modify files without ACLs.
                  any: modify files regardless of ACLs.
      mode     => Modification mode:
                  remove:  remove object.
                  rename:  rename object (dirs and files only).
                  trunc:   truncate object (files and streams only).
                  create:  create under object (dirs and files only).
                  modify:  modify contents of object (files and streams only).
                  sattrs:  set attributes (files and directories only).
                  acl_mod: modify ACEs (files and directories only).
                  acl_del: remove ACLs (files and directories only).
      args     => How to modify the object according to mode
                  (trunc, create, modify, sattrs and acl_mod modes only).
  }

The dirs option could also have:

  {
      streams  => Modify directories with or without streams: yes|no|any.
                  yes: only modify directories with streams.
                  no:  only modify directories without streams.
                  any: modify directories regardless if they contain streams.
      dirs     => Modify directories with or without directories: yes|no|any.
                  yes: only modify directories with directories.
                  no:  only modify directories without directories.
                  any: modify directories regardless if they contain directories
      files    => Modify directories with or without files: yes|no|any.
                  yes: only modify directories with files.
                  no:  only modify directories without files
                  any: modify directories regardless if they contain files.
      depth    => Increase depth of directories, default is 1.
                  If set to 0 it will maintain the current depth structure.
  }

The files option could also have:

  {
      streams  => Modify files with or without streams: yes|no|any.
                  yes: only modify files with streams.
                  no:  only modify files without streams.
                  any: modify files regardless if they contain streams.
      sparse   => Modify files with or without sparse files: yes|no|any.
                  yes: only modify files with sparse files.
                  no:  only modify files without sparse files.
                  any: modify files regardless if they contain sparse files.
  }

Look at create_dirs, create_files or create_streams for specific options
to the args option for the create mode.

The args options for the trunc mode are:

  {
      size     => Truncate to this size.
      size_dev => Standard deviation of the truncate size.
      max_size => Truncate size is a random number between 0 and this value
                  (default is 1000 if size is not given).
  }

The args options for the modify mode are:

  {
      size     => Size of new contents (if append and blocks are not specified).
      size_dev => Standard deviation of the size.
      max_size => Size of new contents is a random number between 0 and this
                  value (default is 1000 if size is not given).
      append   => Append new contents to file or stream.
                  Options size and max_size give the size of data to append.
      blocks   => Number of blocks to modify, offset of each block to modify
                  is selected at random.
                  Option size gives the size of block to modify while max_size
                  gives the maximum offset (do not modify beyond this offset).
      sparse_opts => Hash reference like sparse_opts on create_files.
  }

The args options for the sattrs mode are:

  {
      compress => Set or clear the compress attribute.
      encrypt  => Set or clear the encrypt attribute.
      archive  => Set or clear the archive attribute.
      hidden   => Set or clear the hidden attribute.
  }

The args options for the acl_mod mode are:

  {
      percent   => Percent of ACEs to remove (default is 0).
      acl_masks => Use a maximum of this number of access mask flags (0-18)
                   (default is 8; 0 means use maximum available).
      acl_flags => Use a maximum of this number of flags (0-6)
                   (default is 4; 0 means use maximum available).
      acl_types => Use a maximum of this number of types (0-3)
                   (default is 1; 0 means use maximum available).
      acl_users => Use a maximum of this number of users (0-7)
                   (default is 2; 0 means use maximum available).
      users     => List of users/groups to use instead of the defaults.
  }

Renaming files or directories will not process objects owned by these.
In other words, renaming directories while trying to modify files at the
same time, the files owned by renamed directories will not be processed.
Make two separate calls to modify, one to rename the directories and a
second call to modify the files.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dirs

(Optional) : Reference to a hash of options on how to modify the directories.

=item files

(Optional) : Reference to a hash of options on how to modify the files.

=item streams

(Optional) : Reference to a hash of options on how to modify the streams.

=back

=item Returns

1 if succeeds.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub modify($%) {
    my ($self, %opts) = @_;

    _validate_dir_opts($opts{dirs}{args});
    _validate_file_opts($opts{files}{args});
    _process_size_units($opts{streams}{args});

    return $self->_do_call(\&_modify, %opts);
}

#---------------------------------------------------------------------
# Modify function
# It runs on the local machine
#---------------------------------------------------------------------
sub _modify($%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            dirs    => { type => HASHREF, optional => 1 },
            files   => { type => HASHREF, optional => 1 },
            streams => { type => HASHREF, optional => 1 },
        }
    );
    my ($entry, @args, %free);
    my $rmcount = 0;

    # Default validation hash for the arguments
    my %common = (
        number   => { type => SCALAR,  default => $MAX_NUM },
        original => { type => SCALAR,  default => 0 },
        skip     => { type => SCALAR,  default => 0 },
        skiproot => { type => SCALAR,  default => 0 },
        percent  => { type => SCALAR,  default => 100 },
        compress => { type => SCALAR,  default => 'no',
                     regex => qr/^(no|yes|any)$/ },
        encrypt  => { type => SCALAR,  default => 'no',
                     regex => qr/^(no|yes|any)$/ },
        archive  => { type => SCALAR,  default => 'any',
                     regex => qr/^(no|yes|any)$/ },
        hidden   => { type => SCALAR,  default => 'no',
                     regex => qr/^(no|yes|any)$/ },
        acl      => { type => SCALAR,  default => 'no',
                     regex => qr/^(no|yes|any)$/ },
        mode     => { type => SCALAR,  default => 'remove' },
        args     => { type => HASHREF, default => {} },
    );

    my %def_args = (
        dirs => {
            %common,
            depth   => { type => SCALAR,  default => 1 },
            dirs    => { type => SCALAR,  default => 'no',
                        regex => qr/^(no|yes|any)$/ },
            files   => { type => SCALAR,  default => 'no',
                        regex => qr/^(no|yes|any)$/ },
            streams => { type => SCALAR,  default => 'no',
                        regex => qr/^(no|yes|any)$/ },
        },
        files => {
            %common,
            sparse  => { type => SCALAR,  default => 'no',
                        regex => qr/^(no|yes|any)$/ },
            streams => { type => SCALAR,  default => 'no',
                        regex => qr/^(no|yes|any)$/ },
        },
        streams => \%common,
    );

    # Set defaults for the mode arguments
    foreach $entry ( qw(dirs files streams) ) {
        # Validate arguments for option
        @args = ( %{$opts{$entry}} );
        $opts{$entry} = { validate(@args, $def_args{$entry}) }
          if %{$opts{$entry}};
        if ( !$opts{$entry}{number} || !$opts{$entry}{percent} ) {
            delete $opts{$entry};
            next;
        }

        # Get number of 'remove' modifies
        $rmcount++ if $opts{$entry}{mode} eq 'remove';

        # Set default values for create mode
        if ( $opts{$entry}{mode} eq 'create'
          && exists $opts{$entry}{args}{args} ) {
            $opts{$entry}{args}{args}{stream_num} = $MAX_NUM
              unless defined $opts{$entry}{args}{args}{stream_num};
        }

        # Set default values for modify mode
        if ( $entry eq 'files' && $opts{$entry}{mode} eq 'modify'
          && exists $opts{$entry}{args}{args} ) {
            $opts{$entry}{args}{args}{sparse_num} = $MAX_NUM
              unless defined $opts{$entry}{args}{args}{sparse_num};
        }

        # Set default values for acl_mod and acl_del mode
        if ( $opts{$entry}{mode} =~ /^acl_(mod|del)$/
          && exists $opts{$entry}{args} ) {
            $opts{$entry}{args}{acl_num} = $MAX_NUM
              unless defined $opts{$entry}{args}{acl_num};
        }
    }

    # Return if nothing to modify
    return 1 unless keys %opts;

    # Check fill percentage
    if ( $self->{_fillpct} ) {
        # Get used space on volume
        %free = vol_free(vol => $self->{d_path});
        # Set current space usage and return if limit is hit
        # except if there are any removes
        $self->{d_used} = $free{used};
        $self->{d_remove} = $rmcount;
        return 1 if !$rmcount && $self->{d_used} > $self->{d_max};
    }

    # Set the modification id to keep newly modify/created
    # objects to be modify again in the same call
    $self->{_mod_id}++;

    # Process objects starting at the root
    $self->_process_objects('', $self->{_objects}, \%opts, 0);

    return 1;
}

#---------------------------------------------------------------------

=head2 read_tree

  $out = $data_obj->read_tree([ dirs => \%dir_opts ]
                              [, files => \%file_opts ]);
                              [, streams => \%stream_opts ]);

Read the populated tree according to the specified options.
Each of the main options could have any of the following options:

  {
      number   => Maximum number of the specified objects to read.
      skip     => Number of the specified objects to skip before reading.
      contents => Reads the contents of the file/stream, defaults to 1.
      percent  => Reads this percentage of the specified objects,
                  it has precedence over the 'number' option.
      compress => Read compressed files: yes|no|any.
                  yes: only read compressed files.
                  no:  only read uncompressed files.
                  any: read files regardless of the compress attribute.
      encrypt  => Read encrypted files: yes|no|any.
                  yes: only read encrypted files.
                  no:  only read decrypted files.
                  any: read files regardless of the encrypt attribute.
      archive  => Read archived files: yes|no|any.
                  yes: only read archived files.
                  no:  only read non-archived files.
                  any: read files regardless of the archive attribute.
      hidden   => Read hidden files: yes|no|any.
                  yes: only read hidden files.
                  no:  only read non-hidden files.
                  any: read files regardless of the hidden attribute.
      acl      => Read files with ACLs: yes|no|any.
                  yes: only read files with ACLs.
                  no:  only read files without ACLs.
                  any: read files regardless of ACLs.
  }

The dirs option could also have:

  {
      dirs     => Modify directories with or without directories: yes|no|any.
                  yes: only modify directories with directories.
                  no:  only modify directories without directories.
                  any: modify directories regardless if they contain directories
      files    => Modify directories with or without files: yes|no|any.
                  yes: only modify directories with files.
                  no:  only modify directories without files
                  any: modify directories regardless if they contain files.
      streams  => Modify directories with or without streams: yes|no|any.
                  yes: only modify directories with streams.
                  no:  only modify directories without streams.
                  any: modify directories regardless if they contain streams.
      args     => Options on how to read directories.
  }

The files option could also have:

  {
      sparse   => Read files with or without sparse files: yes|no|any.
                  yes: only read files with sparse files.
                  no:  only read files without sparse files.
                  any: read files regardless if they contain sparse files.
      streams  => Read files with or without streams: yes|no|any.
                  yes: only read files with streams.
                  no:  only read files without streams.
                  any: read files regardless if they contain streams.
      args     => Options on how to read files.
  }

The args option could have:

  {
      acl_num    => Maximum number of files/directories to read ACL's.
      acl_pct    => Percentage of files/directories to read ACL's.
      attr_num   => Maximum number of files/directories to read attributes.
      attr_pct   => Percentage of files/directories to read attributes.
      sparse_num => Maximum number of sparse files to read.
      sparse_pct => Percentage of sparse files to read.
  }

The following named arguments are accepted:

=over

=item Arguments

=over

=item dirs

(Optional) : Reference to a hash of options on how to read the directories.

=item files

(Optional) : Reference to a hash of options on how to read the files.

=item streams

(Optional) : Reference to a hash of options on how to read the streams.

=back

=item Returns

1 if succeeds.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub read_tree($%) { shift->_do_call(\&_read_tree, @_) }

#---------------------------------------------------------------------
# Read function
# It runs on the local machine
#---------------------------------------------------------------------
sub _read_tree($%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            dirs    => { type => HASHREF, optional => 1 },
            files   => { type => HASHREF, optional => 1 },
            streams => { type => HASHREF, optional => 1 },
        }
    );
    my ($entry, @args);

    # Default validation hash for the arguments
    my %common = (
        number   => { type => SCALAR,  default => $MAX_NUM },
        skip     => { type => SCALAR,  default => 0 },
        contents => { type => SCALAR,  default => 1 },
        percent  => { type => SCALAR,  default => 100 },
        compress => { type => SCALAR,  default => 'any',
                     regex => qr/^(no|yes|any)$/ },
        encrypt  => { type => SCALAR,  default => 'any',
                     regex => qr/^(no|yes|any)$/ },
        archive  => { type => SCALAR,  default => 'any',
                     regex => qr/^(no|yes|any)$/ },
        hidden   => { type => SCALAR,  default => 'any',
                     regex => qr/^(no|yes|any)$/ },
        acl      => { type => SCALAR,  default => 'any',
                     regex => qr/^(no|yes|any)$/ },
        args     => { type => HASHREF, default => {} },
    );

    my %def_args = (
        dirs  => {
            %common,
            dirs    => { type => SCALAR,  default => 'any',
                        regex => qr/^(no|yes|any)$/ },
            files   => { type => SCALAR,  default => 'any',
                        regex => qr/^(no|yes|any)$/ },
            streams => { type => SCALAR,  default => 'any',
                        regex => qr/^(no|yes|any)$/ },
        },
        files => {
            %common,
            sparse  => { type => SCALAR,  default => 'any',
                        regex => qr/^(no|yes|any)$/ },
            streams => { type => SCALAR,  default => 'any',
                        regex => qr/^(no|yes|any)$/ },
        },
        streams => \%common,
    );

    # Set defaults for the mode arguments
    foreach $entry ( qw(dirs files streams) ) {
        next if not defined($opts{$entry});

        # Validate arguments for option
        @args = ( %{$opts{$entry}} );
        $opts{$entry} = { validate(@args, $def_args{$entry}) };
        $opts{$entry}{mode} = 'read';
    }

    if ( exists $opts{dirs} ) {
        $opts{dirs}{args}{acl} = 1;
    }
    if ( exists $opts{files} ) {
        $opts{files}{args}{acl} = 1;
        $opts{files}{args}{sparse_opts} = 1;
    }

    # Set the modification id to keep newly modify/created
    # objects to be modify again in the same call
    $self->{_mod_id}++;

    # Process objects starting at the root
    $self->_process_objects('', $self->{_objects}, \%opts, 0);

    return 1;
}

#---------------------------------------------------------------------

=head2 compare

  $out = $data_obj->compare(path => $path
                            [, contents => 0|1 ]
                            [, streams => 0|1 ]
                            [, attributes => 0|1 ]);

Compare the directory tree of the object with the directory tree
given in the argument.

The following named arguments are accepted:

=over

=item Arguments

=over

=item path

(Required) : Path of secondary tree for comparison.

=item contents

(Optional) : If true, compares the contents of the files otherwise
only checks whether files/directories are present.
Default is 1.

=item streams

(Optional) : If true, compares contents of file and directory streams.
Default is 1.

=item attributes

(Optional) : If true, compares file and directory attributes including
ACL, compression, encryption, archive and hidden.
Default is 1.

=item amctime

(Optional) : If true, compares file and directory atime, mtime and ctime.
Default is 0.

=back

=item Returns

The number of differences found.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub compare($%) { shift->_do_call(\&_compare, @_) }

#---------------------------------------------------------------------
# Compare function
# It runs on the local machine
#---------------------------------------------------------------------
sub _compare($%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            path       => { type => SCALAR },
            contents   => { type => SCALAR, default => 1 },
            streams    => { type => SCALAR, default => 1 },
            attributes => { type => SCALAR, default => 1 },
            amctime    => { type => SCALAR, default => 0 },
        }
    );

    # Create secondary data manipulation object
    my $data = dp::snapmirror::qsm::TargetedData->new(
        path   => $opts{path},
        create => 0,
        save   => 1,
    );

    # Process objects starting at the root
    return $self->_compare_objects('', $self->{_objects}, $data->{_objects},
                                   \%opts);
}

#---------------------------------------------------------------------
# Traverse the object's hash and compare each item at a time
# This is a recursive function
#---------------------------------------------------------------------
sub _compare_objects($$$$$;) {
    my $self = shift;
    my ($base, $ref1, $ref2, $mod) = @_;
    my ($key, $item, $type, $level, $file);
    my $ndiff = 0;

    # Check for extra objects in secondary tree
    foreach $item ( grep { ! /^-/o && ! exists $ref1->{$_} } keys %$ref2 ) {
        $file = _get_file($base, $item);
        ($type, $level) = _get_type($ref2->{$item});
        logVerbose($level, "The $type $file does not exists on primary tree");
        $ndiff++;
    }

    # Compare all elements
    foreach $key ( keys %$ref1 ) {
        # Skip internal keys
        next if $key =~ /^-/o;
        $file = _get_file($base, $key);

        # Process the current object
        $ndiff += $self->_compare_object($base, $ref1, $ref2, $mod, $key);

        # Do not traverse beyond a stream
        # Streams do not own any objects
        next if $ref1->{$key}{-type} eq 's';

        # Process objects owned by this object
        $ndiff +=
          $self->_compare_objects($file, $ref1->{$key}, $ref2->{$key}, $mod);
    }

    return $ndiff;
}

#---------------------------------------------------------------------
# Compare each object
#---------------------------------------------------------------------
sub _compare_object($$$$$$;) {
    my $self = shift;
    my ($base1, $ref1, $ref2, $mod, $key) = @_;
    my ($base2, $file2, $type1, $type2, $level, $file1, @attrs1, @attrs2);
    my ($item, $index, $sparse, $val1, $val2, %wattrs1, %wattrs2, @wdiff);
    my $ndiff = 0;
    my $size_diff = 0;

    # Get name and type of object
    ($type1, $level) = _get_type($ref1->{$key});
    $file1 = _get_file($base1, $key);
    if ( $file1 eq '' ) {
        $file1 = $self->{_cwd};
        $file1 =~ s/\\+$//;
    }
    $file1 = "$key on $base1" if $ref1->{$key}{'-type'} eq 's';

    # Check if element exists in secondary tree
    if ( ! exists $ref2->{$key} ) {
        logVerbose($level, "The $type1 $file1 does not exists on " .
                           " secondary tree");
        return ++$ndiff;
    }

    # Get name and type of object on secondary tree
    $type2 = _get_type($ref2->{$key});
    $base2 = _get_file($mod->{path}, $base1);
    $file2 = _get_file($base2, $key);
    $file2 = "$key on $base2" if $ref1->{$key}{'-type'} eq 's';

    # Compare types
    if ( $type1 ne $type2 ) {
        logVerbose($level, "The $type1 $file1 is a $type1 on the primary " .
                           "tree and a $type2 on the secondary tree");
        return ++$ndiff;
    }

    if ( $mod->{attributes} ) {
        # Get attributes
        if ( $ref1->{$key}{'-type'} eq 's' ) {
            if ( $self->{_hosttype} eq 'windows' ) {
                # FileManip does not support stat on streams on windows
                @attrs1 = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
                @attrs2 = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
            } else {
                @attrs1 = stat_stream(file => $base1, stream => $key);
                @attrs2 = stat_stream(file => $base2, stream => $key);
            }
        } else {
            if ( $self->{_hosttype} eq 'windows' ) {
                # Compare ACLs
                $val1 = $self->_read_acl($file1);
                $val2 = $self->_read_acl($file2);
                if ( $val1 ne $val2 ) {
                    logVerbose(VERBOSE_ATTRS,
                               "The ACL is different for the $type1 $file1");
                    $ndiff++;
                }

                # Get windows attributes
                %wattrs1 = getattr_file(file => $file1);
                %wattrs2 = getattr_file(file => $file2);

                @wdiff = grep { $wattrs1{$_} != $wattrs2{$_} } keys %wattrs1;
                logVerbose(VERBOSE_ATTRS, "The following Windows attributes " .
                                          "are different for $type1 $file1: " .
                                          join(', ', @wdiff)) if @wdiff;
                $ndiff++ if @wdiff;
            }

            # Get basic file attributes
            @attrs1 = lstat $file1;
            @attrs2 = lstat $file2;
        }

        if ( $mod->{amctime} ) {
            # Compare atime
            if ( $attrs1[8] != $attrs2[8] ) {
                logVerbose(VERBOSE_ATTRS, "atime mismatch for $type1 $file1: " .
                                          " $attrs1[8]/$attrs2[8]");
                $ndiff++;
            }

            # Compare mtime
            if ( $attrs1[9] != $attrs2[9] ) {
                logVerbose(VERBOSE_ATTRS, "mtime mismatch for $type1 $file1: " .
                                          " $attrs1[9]/$attrs2[9]");
                $ndiff++;
            }

            # Compare ctime
            if ( $attrs1[10] != $attrs2[10] ) {
                logVerbose(VERBOSE_ATTRS, "ctime mismatch for $type1 $file1: " .
                                          " $attrs1[10]/$attrs2[10]");
                $ndiff++;
            }
        }

        # Compare size
        if ( $attrs1[7] != $attrs2[7] ) {
            logVerbose(VERBOSE_ATTRS, "Size mismatch for $type1 $file1: " .
                                      " $attrs1[7]/$attrs2[7]");
            $ndiff++;
            $size_diff = 1;
        }

        # Compare permissions
        if ( $attrs1[2] != $attrs2[2] ) {
            $val1 = sprintf('%0o', $attrs1[2]);
            $val2 = sprintf('%0o', $attrs2[2]);
            logVerbose(VERBOSE_ATTRS, "Permissions mismatch for $type1 " .
                                      "$file1: $val1/$val2");
            $ndiff++;
        }

        # Compare nlinks
        if ( $attrs1[3] != $attrs2[3] ) {
            logVerbose(VERBOSE_ATTRS, "Number of links mismatch for $type1 " .
                                      "$file1: $attrs1[3]/$attrs2[3]");
            $ndiff++;
        }

        # Compare uid
        if ( $attrs1[4] != $attrs2[4] ) {
            logVerbose(VERBOSE_ATTRS, "User id mismatch for $type1 $file1: " .
                                      "$attrs1[4]/$attrs2[4]");
            $ndiff++;
        }

        # Compare gid
        if ( $attrs1[5] != $attrs2[5] ) {
            logVerbose(VERBOSE_ATTRS, "Group id mismatch for $type1 $file1: " .
                                      "$attrs1[5]/$attrs2[5]");
            $ndiff++;
        }
    }

    if ( $mod->{contents} ) {
        # Compare contents
        if ( $ref1->{$key}{'-type'} eq 'f' ) {
            # Compare files
            if ( $ref1->{$key}{'-sparse'} ) {
                # Compare number of data segments on sparse file
                if ( @{$ref1->{$key}{'-sparse'}}
                  != @{$ref2->{$key}{'-sparse'}} ) {
                    logVerbose($level, "Number of data segments mismatch " .
                                       "for $type1 $file1");
                    return ++$ndiff;
                }

                # Compare offsets and sizes of data segments
                $index = 0;
                foreach $item ( @{$ref1->{$key}{'-sparse'}} ) {
                    $sparse = $ref2->{$key}{'-sparse'}[$index++];
                    if ( $item->{offset} != $sparse->{offset} ) {
                        logVerbose($level, "Segment offset mismatch " .
                                           "for $type1 $file1");
                        return ++$ndiff;;
                    }
                    if ( $item->{size} != $sparse->{size} ) {
                        logVerbose($level, "Segment size mismatch " .
                                           "for $type1 $file1");
                        return ++$ndiff;;
                    }
                }
            }

            if ( exists $ref1->{$key}{'-link'}
              && $ref1->{$key}{'-link'} == 2 ) {
                # Compare contents of symbolic link
                $val1 = readlink $file1;
                $val2 = readlink $file2;
                if ( $val1 ne $val2 ) {
                    logVerbose($level, "Symbolic link mismatch " .
                                       "for $file1: $val1 | $val2");
                    return ++$ndiff;;
                }
            } elsif ( ! $size_diff ) {
                # Compare hard links, sparse files and regular files
                # Only compare files if size is the same
                if ( File::Compare::compare($file1, $file2) ) {
                    logVerbose($level, "Content mismatch for $type1 $file1");
                    return ++$ndiff;;
                }
            }
        } elsif ( $mod->{streams} && $ref1->{$key}{'-type'} eq 's' ) {
            # Compare stream contents
            # streams are read in full
            $val1 = read_stream(file => $base1, stream => $key);
            $val2 = read_stream(file => $base2, stream => $key);
            if ( $val1 ne $val2 ) {
                logVerbose($level, "Content mismatch for $type1 $file1");
                return ++$ndiff;;
            }
        }
    }

    return $ndiff;
}

#---------------------------------------------------------------------
# Traverse the object's hash and process each item at a time
# This is a recursive function
#---------------------------------------------------------------------
sub _process_objects($$$$$;) {
    my $self = shift;
    my ($base, $ref, $mod, $depth) = @_;
    my ($more, $key, $ndepth, $ret);
    my @keys = keys %$ref;

    # Process all objects in current object
    # Not using 'keys %$ref' on the foreach to avoid processing
    # the same element twice since adding an entry to the hash
    # results in the keys to be rearranged in an unpredictable way
    foreach $key ( @keys ) {
        # Skip internal keys
        next if $key =~ /^-/o;
        # Return if limit is hit except if there are any removes
        last if $self->{_fillpct} && !$self->{d_remove}
          && $self->{d_used} > $self->{d_max};

        # Process the current object
        $ret = $self->_process_object($base, $ref, $mod, $key, $depth);

        # Change current object name if it was renamed
        $key = $ret if defined $ret;

        # Check if done
        # Calculate if there is any items left to be processed
        # (if any of the counts are not zero)
        $more = 0;
        $more += $mod->{$_}{number} foreach ( keys %$mod );
        last if not $more;
        # Check if current object was not removed/renamed
        next if not exists($ref->{$key});

        # Do not traverse beyond a stream
        # Streams do not own any objects
        next if $ref->{$key}{-type} eq 's';

        $ndepth = $depth;
        if ( $ref->{$key}{-type} eq 'd' ) {
            # Change to directory to process its children
            $self->_chdir($key);
            $ndepth++;
        }

        # Process objects owned by this object
        $self->_process_objects($key, $ref->{$key}, $mod, $ndepth);

        if ( $ref->{$key}{-type} eq 'd' ) {
            # Change back directories
            $self->_chdir_restore();
        }
    }
}

#---------------------------------------------------------------------
# Process each object according to the modify hash of options
#---------------------------------------------------------------------
sub _process_object($$$$$$;) {
    my $self = shift;
    my ($base, $ref, $mod_obj, $obj, $depth) = @_;
    my ($mod, $args, $size, $cont, $name, $cmd, $out, $offset, $ret, $n, @list);
    my ($otype, %free, $du_diff, $du_need, %mopts);
    my @obj_attrs = qw(dirs files streams sparse compress encrypt hidden acl);
    my %obj_types = (s => 'streams', f => 'files', d => 'dirs');
    my $type = $obj_types{$ref->{$obj}{-type}};
    # Use the absolute path for the directory, operations on the
    # current directory fail for windows
    my $stream_base = $ref->{-type} eq 'd' ? $self->{_cwd} : $base;
    $stream_base =~ s/\\*\/*$//;

    # No modify arguments for this type of object
    return if not defined($type);
    return if not defined($mod_obj->{$type});

    if ( $self->{_fillpct} && $self->{_modcnt} > $self->{_mcdiff}) {
        # Get used space on volume
        %free = vol_free(vol => $self->{d_path});
        # Get difference in space on this iteration for this count
        $du_diff = $free{used} - $self->{d_used};

        # Return if limit is hit
        $self->{d_used} = $free{used};
        return if $self->{d_used} > $self->{d_max};

        # Get the space still available to use
        $du_need = $self->{d_max} - $self->{d_used};
        # Adjust how often to poll the system for the used space
        # it depends on the space difference for this iteration
        $self->{_mcdiff} *= ($du_need/10) / $du_diff if $du_diff > 0;
        $self->{_mcdiff} = 300 if $self->{_mcdiff} > 300;
        $self->{_mcdiff} = 1 if $self->{_mcdiff} < 1;
        $self->{_modcnt} = 0;
    }

    # Modify hash for this object
    $mod = $mod_obj->{$type};

    # Do not modify root directory
    return if ($obj eq '.' && $mod->{skiproot});

    # Object has been modified already
    if ( $ref->{$obj}{-mod_id} == $self->{_mod_id} ) {
        return;
    }

    # No more objects of this type to modify
    return if $mod->{number} < 1;

    if ( $type ne 'streams' ) {
        # On Windows the archive bit is set by default on files
        $n = exists($ref->{$obj}{-archive}) ? $ref->{$obj}{-archive} :
             $self->{_hosttype} eq 'windows' ? $type eq 'files' ? 1 : 0 : 0;
        if ( $mod->{archive} ne 'any' && ($mod->{archive} eq 'yes' xor $n) ) {
            # Object does not match required archive attribute
            return;
        }

        foreach $n ( @obj_attrs ) {
            if ( defined($mod->{$n}) && $mod->{$n} ne 'any'
                 && ($mod->{$n} eq 'yes' xor $ref->{$obj}{-$n}) ) {
                # Object does not match required attributes
                return;
            }
        }
    }

    if ( $obj eq '.' && ($mod->{mode} eq 'remove' || $mod->{mode} eq 'sattrs'
      || $mod->{mode} eq 'rename' || $mod->{mode} eq 'copy') ) {
        # Do not remove/rename/copy/sattrs the top level directory
        return;
    }

    # Modify original objects only
    if ( $mod->{original} && $ref->{$obj}{-mod_id} ) {
        return;
    }

    if ( $mod->{skip} > 0 ) {
        # Skip the first objects
        --$mod->{skip};
        return;
    }

    if ( $ref->{$obj}{-link} && ( $mod->{mode} eq 'sattrs'
      || !( $mod->{mode} eq 'remove' || $mod->{mode} eq 'rename' ||
            $mod->{mode} eq 'read' ||
            ( ($mod->{mode} eq 'trunc' || $mod->{mode} eq 'modify')
              && -e $obj && ! -d $obj ) ) ) ) {
        # Remove and rename links unconditionally
        # Modify/truncate symbolic links to files
        # Do not modify/truncate symbolic links to directories
        # Do not set attributes on a link
        return;
    }

    # Check if object is not going to be modified
    # according to percent argument
    return if $mod->{percent} < rand(100);

    # Find file type
    if ( $ref->{$obj}{-sparse} ) {
        $otype = 'sparse';
    } elsif ( $ref->{$obj}{-link} ) {
        $otype = $ref->{$obj}{-link} == 1 ? 'links' : 'slinks';
    } else {
        $otype = $type;
    }

    if ( $mod->{mode} eq 'remove' && $type =~ /^(streams|files|dirs)$/ ) {
        # Remove object
        if ( $type eq 'streams' ) {
            logVerbose(VERBOSE_STREAMS, "Remove stream $obj from $stream_base");
            # Remove stream
            remove_stream(file=>$stream_base, stream=>$obj);
            push @{$self->{_remove}{streams}}, $obj;
        } else {
            # Remove file or directory
            $n = ($type eq 'dirs') ? 1 : 0;
            if ( $n ) {
                logVerbose(VERBOSE_DIRS, 'Remove directory ' . $obj);
            } else {
                logVerbose(VERBOSE_FILES, 'Remove file ' . $obj);
            }
            remove_file(file=>$obj, recursive=>$n);
            if ( $n ) {
                push @{$self->{_remove}{dirs}}, $obj;
            } else {
                push @{$self->{_remove}{files}}, $obj;
            }
        }
        --$ref->{-$type};
        delete $ref->{$obj};
        $self->{_counters}{remove}{$otype}++;
    } elsif ( $mod->{mode} eq 'trunc' && $type =~ /^(streams|files)$/ ) {
        # Truncate object
        $args = _get_args($mod->{args});
        $size = $args->{size};
        $self->{_modcnt}++;
        if ( $type eq 'streams' ) {
            logVerbose(VERBOSE_STREAMS,
                       "Truncate stream $obj on file $stream_base");
            # Truncate stream
            setattr_stream(file=>$stream_base, stream=>$obj, size=>$size);
        } elsif ( $type eq 'files' ) {
            logVerbose(VERBOSE_FILES, "Truncate file $obj");
            # Truncate file
            truncate $obj, $size;
        }
        $ref->{$obj}{-mod_id} = $self->{_mod_id};
        $self->{_counters}{trunc}{$otype}++;
    } elsif ( $mod->{mode} eq 'rename' && $type =~ /^(files|dirs)$/ ) {
        # Rename object
        $args = _get_args($mod->{args});
        $name = $args->{name};
        if ( $type eq 'dirs' ) {
            logVerbose(VERBOSE_DIRS, "Rename directory $obj to $name");
        } else {
            logVerbose(VERBOSE_FILES, "Rename file $obj to $name");
        }
        rename $obj, $name or croak "Could not rename $obj: $!";
        $ref->{$name} = $ref->{$obj};
        delete $ref->{$obj};
        $ref->{$name}{-mod_id} = $self->{_mod_id};
        $ret = $name;
        $self->{_counters}{rename}{$otype}++;
    } elsif ( $mod->{mode} eq 'copy' && $type =~ /^(files|dirs)$/ ) {
        $self->{_modcnt}++;
        # Copy object
        if ( $self->{_hosttype} =~ /^windows/ ) {
            $cmd = $type eq 'files' ? 'copy' : 'xcopy /q/i';
        } elsif ( $self->{_hosttype} =~ /solaris/ ) {
            $cmd = $type eq 'files' ? 'cp -@' : 'cp -r@';
        } else {
            croak "Copy is not supported for $self->{_hosttype}";
        }
        $args = _get_args($mod->{args});
        $name = $args->{name};
        if ( $type eq 'dirs' ) {
            logVerbose(VERBOSE_DIRS, "Copy directory $obj to $name");
        } else {
            logVerbose(VERBOSE_FILES, "Copy file $obj to $name");
        }
        $out = qx($cmd $obj $name 2>&1);
        croak $out if $?;
        $ref->{$name} = $self->_copy_ref($ref->{$obj});
        $ref->{$name}{-mod_id} = $self->{_mod_id};
        $self->{_counters}{copy}{$otype}++;
    } elsif ( $mod->{mode} eq 'create' && $type =~ /^(files|dirs)/ ) {
        # Get optional arguments
        $args = _get_args($mod->{args}{args});

        # Create streams on file/directory
        if ( $args->{streams} && defined($mod->{args}{streams})
          && $mod->{args}{args}{stream_num} ) {
            $self->_create_streams(file=>$obj, %{$mod->{args}{streams}});
            $mod->{args}{args}{stream_num}--;
        }
        if ( $type eq 'dirs' ) {
            # Create files on directory
            if ( defined($mod->{args}{files}) ) {
                $self->_create_files(dir=>$obj, %{$mod->{args}{files}});
            }
            if ( $mod->{depth} || $ref->{$obj}{-dirs} ) {
                $self->_create_dirs(dir=>$obj, %{$mod->{args}});
            }
        }
    } elsif ( $mod->{mode} eq 'modify' && $type =~ /^(streams|files)/ ) {
        # Get size and append flag from arguments
        $args = _get_args($mod->{args});
        $self->{_modcnt}++;
        if ( $ref->{$obj}{-sparse} ) {
            if ( $args->{sparse_opts} ) {
                logVerbose(VERBOSE_SPARSE, "Modify sparse file $obj");
                _create_sparse($obj, $args, $ref);
                $mod->{args}{sparse_num}--;
                $self->{_counters}{modify}{sparse}++;
            }
            return;
        }

        if ( $type eq 'streams' ) {
            logVerbose(VERBOSE_STREAMS,
                       "Modify stream $obj on file $stream_base");
        } else {
            logVerbose(VERBOSE_FILES, "Modify file $obj");
        }

        foreach $out ( @{$args->{offsets}} ) {
            $offset = $out;
            %mopts = $args->{blocks} ? (offset => $offset) : ();
            $n = $args->{append};
            $size = $args->{size};
            if ( $type eq 'streams' ) {
                # Modify stream
                do {
                    $size -= _create_rand_data($cont, $size);
                    write_stream(file     => $stream_base,
                                 stream   => $obj,
                                 contents => $cont,
                                 append   => $n,
                                 %mopts);
                    if ( $args->{blocks} ) {
                        $n = 0;
                        $offset += length($cont);
                    } else {
                        $n = 1;
                    }
                } while ( $size > 0 );
            } else {
                # Modify file
                $n = $args->{blocks} ? 0 : $n ? O_APPEND : O_TRUNC;
                sysopen(FH, $obj, O_WRONLY|O_BINARY|$n)
                  or croak "Could not open file $obj: $!";
                seek(FH, $offset, 0) if $args->{blocks};
                while ( $size > 0 ) {
                    $size -= _create_rand_data($cont, $size);
                    print FH $cont or croak "Could not write to file $obj: $!";
                }
                close(FH);
            }
        }
        $ref->{$obj}{-mod_id} = $self->{_mod_id};
        $self->{_counters}{modify}{$otype}++;
    } elsif ( $mod->{mode} eq 'sattrs' && $type =~ /^(files|dirs)/ ) {
        $args = _get_args($mod->{args});
        if ( %{$args->{sattrs}} ) {
            _set_attributes($obj, $args->{sattrs}, $ref, $mod->{args});
        }
        $ref->{$obj}{-mod_id} = $self->{_mod_id};
        $self->{_counters}{sattrs}{$otype}++;
    } elsif ( $mod->{mode} eq 'acl_mod' && $type =~ /^(files|dirs)/ ) {
        $args = _get_args($mod->{args});
        logVerbose(VERBOSE_ATTRS, "Modify ACL on $type $obj");
        $ref->{$obj}{-acl} = $self->_modify_acl($obj, $args, $ref, 1);
        $ref->{$obj}{-mod_id} = $self->{_mod_id};
        $mod->{args}{acl_num}--;
        $self->{_counters}{modify}{acls}++;
    } elsif ( $mod->{mode} eq 'acl_del' && $type =~ /^(files|dirs)/ ) {
        $args = _get_args($mod->{args});
        logVerbose(VERBOSE_ATTRS, "Delete ACL on $type $obj");
        $self->_modify_acl($obj, $args, $ref, 2);
        delete $ref->{$obj}{-acl};
        $ref->{$obj}{-mod_id} = $self->{_mod_id};
        $mod->{args}{acl_num}--;
        $self->{_counters}{remove}{acls}++;
    } elsif ( $mod->{mode} eq 'read' ) {
        $args = _get_args($mod->{args});
        if ( $type ne 'streams' ) {
            $n = ($type eq 'dirs') ? 'directory' : 'file';
            if ( $args->{rattrs} ) {
                logVerbose(VERBOSE_ATTRS, "Read attributes for $n $obj");
                lstat $obj;
                getattr_file(file => $obj);
                $mod->{args}{attr_num}--;
                $self->{_counters}{read}{attrs}++;
            }
            if ( $args->{acl} ) {
                logVerbose(VERBOSE_ATTRS, "Read ACL for $n $obj");
                $self->_read_acl($obj);
                $mod->{args}{acl_num}--;
                $self->{_counters}{read}{acls}++;
            }
        }

        if ( $type eq 'dirs' ) {
            logVerbose(VERBOSE_FILES, "Read directory $obj");
            if ( opendir DH, $obj ) {
                @list = readdir DH;
                closedir DH;
            }
        } elsif ( $type eq 'files' && $ref->{$obj}{-sparse} ) {
            if ( $args->{sparse} ) {
                logVerbose(VERBOSE_SPARSE, "Read sparse file $obj");
                read_file(file=>$obj);
                $mod->{args}{sparse_num}--;
            }
            # Increment number since this is the counter for regular files
            # and will be decremented later and no regular file will be read
            $mod->{number}++;
        } elsif ( $mod->{contents} ) {
            if ( $type eq 'streams' ) {
                logVerbose(VERBOSE_STREAMS,
                           "Read stream $obj on file $stream_base");
                read_stream(file=>$stream_base, stream=>$obj);
            } elsif ( $otype eq 'links' ) {
                logVerbose(VERBOSE_FILES, "Read link $obj");
                read_file(file=>$obj) if -e $obj;
            } elsif ( $otype eq 'slinks' ) {
                logVerbose(VERBOSE_FILES, "Read symbolic link $obj");
                $out = readlink $obj;
            } elsif ( $type eq 'files' ) {
                logVerbose(VERBOSE_FILES, "Read file $obj");
                read_file(file=>$obj);
            }
            if ( $mod->{number} < 2 ) {
                $n = 0;
                $out = 0;
                # Get remaining number of attributes to be read
                @list = keys %{$mod->{args}};
                map { $n   += $mod->{args}{$_} } grep { /_num$/ } @list;
                map { $out += $mod->{args}{$_} } grep { /_pct$/ } @list;
                if ( $n * $out != 0 ) {
                    $mod->{number}   = $MAX_NUM;
                    $mod->{percent}  = 100;
                    $mod->{contents} = 0;
                }
            }
        }
        $self->{_counters}{read}{$otype}++;
    } else {
        # Should never get here
        croak "Mode <$mod->{mode}> does support object type <$type>";
    }

    # Decrement counter of objects to modify of this type
    --$mod->{number};

    return $ret;
}

#---------------------------------------------------------------------
# Internal function
# Convert all the size arguments to bytes
#---------------------------------------------------------------------
sub _process_size_units($;) {
    my $args = shift;

    return unless defined $args;

    foreach my $item ( qw(size size_dev max_size offset offset_dev) ) {
        if ( exists $args->{$item} ) {
            $args->{$item} = _units($args->{$item});
        }
    }

    return;
}

#---------------------------------------------------------------------
# Internal function
# Returns true according to percentage. If percentage is not defined
# it assumes 100% so returns true as default unless the flag is
# defined on the global hash
#---------------------------------------------------------------------
sub _set_on_pct($$;) {
    my ($mod, $opt) = @_;
    my $pct = $opt . '_pct';
    my $num = $opt . '_num';
    my $ret = exists $pct_ret{$pct} ? $pct_ret{$pct} : 1;

    # Check if this type of object needs to be modified
    if ( defined($mod->{$num}) && ! $mod->{$num} ) {
        return 0;
    }

    # Check if this object will be modified based on percentage
    if ( defined($mod->{$pct}) ) {
        $ret = ($mod->{$pct} > 0 && rand(100) < $mod->{$pct}) ? 1 : 0;
    }

    return $ret;
}

#---------------------------------------------------------------------
# Internal function
# Get the arguments for the modify mode and specify defaults
#---------------------------------------------------------------------
sub _get_args($;) {
    my $mod = shift;
    my ($append, $name_reuse, %sattrs, $percent, $num, $name_size);
    my ($acl, $acl_masks, $acl_flags, $acl_users, $acl_types, $min_name);
    my ($streams, $set_acl, $pct, $sparse_opts, $sdata, $off, $data, $buf);
    my ($link_data, $rattrs);
    my $size   = 0;
    my $sparse = 0;
    my $link   = 0;
    my $slink  = 0;
    my $name = '';
    my $n = @val_char;
    my $blocks = 0;
    my @offsets = (0);
    my @attrs = qw(encrypt compress archive hidden);

    # Process percentage options
    $streams = _set_on_pct($mod, 'stream');
    $set_acl = _set_on_pct($mod, 'acl');
    $rattrs  = _set_on_pct($mod, 'attr');
    # If the sparse percentage is not defined --> set it to 0
    $sparse  = _set_on_pct($mod, 'sparse') if defined($mod->{sparse_opts});
    if ( ! $sparse ) {
        # Sparse has higher precedence over links
        if ( $mod->{_ref}{-dirs} || $mod->{_ref}{-files} ) {
            $link = _set_on_pct($mod, 'link');
            if ( ! $link ) {
                # Hard links have higher precedence over symbolic links
                $slink = _set_on_pct($mod, 'slink');
            }
        }
    }

    if ( $link || $slink ) {
        # Get name of file/directory to link to
        $data = undef;
        foreach ( keys %{$mod->{_ref}} ) {
            if ( ! /^-/ && ($slink || -f) ) {
                # Hard link only to a file
                $data = $_;
                last;
            }
        }
        if ( ! defined $data ) {
            # Could not find appropriate file
            $link = 0;
            $slink = 0;
        }
    }

    if ( $link || $slink ) {
        # Setup link data
        $link_data = { file => $data, symbolic => $slink ? 1 : 0 };
    } elsif ( $sparse ) {
        if ( ref($mod->{sparse_opts}) eq 'ARRAY' ) {
            # Sparse options could be an array ready for write_sparse
            $sparse_opts = $mod->{sparse_opts};
        } elsif ( ref($mod->{sparse_opts}) eq 'HASH' ) {
            # This is a hash so calculate array of data blocks and offsets
            # Calculate number of data segments
            $off = 0;
            $sdata = $mod->{sparse_opts};
            $num = intRandGauss($sdata->{ndata}, $sdata->{ndata_dev});
            $num = 1 if $num < 1;
            foreach ( 1..$num ) {
                # For each data segment calculate the size
                $size = intRandGauss($sdata->{size}, $sdata->{size_dev});
                $off += intRandGauss($sdata->{offset}, $sdata->{offset_dev});
                $size = 1 if $size < 1;
                $data = '';
                while ( $size > 0 ) {
                    # Generate data block
                    $size -= _create_rand_data($buf, $size);
                    $data .= $buf;
                }
                # Save data block and offset on array
                push @$sparse_opts, { offset => $off, data => $data };
                undef $data;
            }
        }
    } elsif ( defined($mod->{size}) && defined($mod->{size_dev}) ) {
        # Calculate size if size_dev is given
        $size = intRandGauss($mod->{size}, $mod->{size_dev});
    } elsif ( defined($mod->{size}) ) {
        # Get the size from the args
        $size = $mod->{size};
    } else {
        # Calculate size from max_size (default 1000)
        $size = defined($mod->{max_size}) ? $mod->{max_size} : 1000;
        $size = int(rand(1.1*$size) - 0.1*$size);
    }

    # Bound size
    $size = 0 if $size < 0;

    # Reuse previously deleted names
    $name_reuse = defined($mod->{name_reuse}) ? $mod->{name_reuse} : 0;

    if ( $name_reuse && @{$mod->{_rmlist}} ) {
        # Reuse a previously removed name
        $name = shift @{$mod->{_rmlist}};
    } else {
        if ( defined($mod->{name_size}) && defined($mod->{name_size_dev}) ) {
            # Both the size and standard deviation are provided,
            # use Gaussian distribution
            $name_size = intRandGauss($mod->{name_size}, $mod->{name_size_dev});
        } elsif ( defined($mod->{name_size}) ) {
            # Get the name_size from the args
            $name_size = $mod->{name_size};
        } elsif ( defined($mod->{max_name_size}) ) {
            # Calculate name_size from max_name_size
            $name_size = int(rand($mod->{max_name_size}));
        } else {
            # Default
            $name_size = intRandGauss(8, 2);
        }

        # Make sure name size is at least the minimum (default is 2)
        $min_name = exists $mod->{min_name_size} ? $mod->{min_name_size} : 2;
        $name_size = $min_name if $name_size < $min_name;

        # Try 100 times for a unique and valid name
        foreach ( 1..100 ) {
            # Create name
            foreach ( 1..$name_size ) {
                $name .= $val_char[int(rand($n))];
            }
            # Check for validity and uniqueness
            if ( ! $invalid_filenames{$name}
              && (! $mod->{_ref} || ! exists $mod->{_ref}{$name}) ) {
                # Found it
                last;
            } else {
                # Reset name and look for another
                $name = '';
            }
        }
        croak "Could not find a unique name" if length($name) == 0;
    }

    # Get the append flag from the args or use default
    $append = defined($mod->{append}) ? $mod->{append} : 0;

    # Get attributes to be set
    %sattrs = map { $_ => $mod->{$_} } grep { defined $mod->{$_} } @attrs;
    foreach ( @attrs ) {
        $pct = $_ . '_pct';
        next if ! defined $mod->{$pct};
        $num = _set_on_pct($mod, $_);
        if ( $num ) {
            $sattrs{$_} = 1 if ! exists $sattrs{$_};
        } else {
            delete $sattrs{$_} if exists $sattrs{$_};
        }
    }

    # Encrypt and compress are mutually exclusive
    if ( $sattrs{encrypt} && $sattrs{compress} ) {
        delete $sattrs{compress};
    }

    if ( $mod->{_ref}{-encrypt} && $sattrs{compress} ) {
         # If parent is encrypted, the file will be encrypted
         delete $sattrs{compress};
    }

    if ( $mod->{_ref}{-compress} && $sattrs{encrypt} ) {
         # If parent is compressed, the file will be compressed
         delete $sattrs{encrypt};
    }

    # Get list of offsets for modification
    if ( defined($mod->{blocks}) ) {
        @offsets = ();
        foreach ( 1..$mod->{blocks} ) {
            push @offsets, int(rand($mod->{max_size}));
        }
        $size = defined($mod->{size}) ? $mod->{size} : 1000;
        $blocks = 1;
        $append = 0;
    }

    # Set/modify ACL on file/directory
    $acl       = defined($mod->{acl}) && $set_acl ? $mod->{acl} : 0;
    $acl_masks = defined($mod->{acl_masks}) ? $mod->{acl_masks} : 8;
    $acl_flags = defined($mod->{acl_flags}) ? $mod->{acl_flags} : 4;
    $acl_types = defined($mod->{acl_types}) ? $mod->{acl_types} : 2;
    $acl_users = defined($mod->{acl_users}) ? $mod->{acl_users} : 2;

    # Percent defaults to 0
    $percent = defined($mod->{percent}) ? $mod->{percent} : 0;

    return { size        => $size,
             streams     => $streams,
             append      => $append,
             rattrs      => $rattrs,
             sattrs      => \%sattrs,
             sparse      => $sparse,
             sparse_opts => $sparse_opts,
             link        => $link || $slink,
             link_data   => $link_data,
             name        => $name,
             blocks      => $blocks,
             offsets     => \@offsets,
             percent     => $percent,
             acl         => $acl,
             acl_masks   => $acl_masks,
             acl_flags   => $acl_flags,
             acl_types   => $acl_types,
             acl_users   => $acl_users,
             users       => $mod->{users},
             name_reuse  => $name_reuse };
}

#---------------------------------------------------------------------
# It creates a unique ACL on a file or modifies the contents of the ACL
#---------------------------------------------------------------------
sub _modify_acl($$$;$$) {
    my ($self, $file, $args, $ref, $modify) = @_;
    my (@mlist, @flist, @ulist, @tlist, $mask, $flag, $user, $type);
    my ($object, @dacl_changes, @sacl_changes, $ace, $ace_curr, $index);
    my ($new_dacl, $new_sacl, @dacl_aces, @sacl_aces, @aces, $entry, $cache);
    my $retry  = 20;
    my $mode   = $args->{acl};
    my @users  = defined($args->{users}) ? @{$args->{users}} : @ace_users;
    my $ntypes = $args->{acl_types} ? $args->{acl_types} : @ace_types;
    my @types  = @ace_types[0..$ntypes-1];
    $modify = 0 if not defined($modify);

    croak "Add/modify ACL is only supported on Windows"
      if $self->{_hosttype} ne 'windows';

    # Get which access, flags, types and users to use for the ACEs
    while ( $retry > 0 && $modify < 2 ) {
        # Get list from ace mask flags
        @mlist = _get_from_list(@ace_masks, $args->{acl_masks}, $mode);
        $entry = join('-', @mlist);

        # Get list from ace flags
        @flist = _get_from_list(@ace_flags, $args->{acl_flags}, $mode);
        $entry .= ':' . join('-', @flist);

        # Get list from ace types
        @tlist = _get_from_list(@types, $args->{acl_types}, $mode);
        $entry .= ':' . join('-', @tlist);

        # If ACE type is DENIED then do not select Users or Administrators
        if ( grep { /DENIED/ } @types[@tlist] ) {
            # Remove the Users and Administrators groups from the list
            @users = grep {!/^(users|administrators)$/i} @users;
            # Add the Guests group to the list if empty
            push @users, 'Guests' if !@users;
        }

        # Get list from ace users
        @ulist = _get_from_list(@users, $args->{acl_users}, $mode);
        $entry .= ':' . join('-', @ulist);

        if ( ! $acls{$entry} ) {
            # Add format of ACL to list of created ACLs
            $acls{$entry} = 1;
            last;
        }
        --$retry;
    }

    croak "Could not create a unique ACL" if ($retry < 1);

    if ( $cifs_security ) {
        # Load modules needed for ACL only once
        require Win32::Security::ACE;
        require Win32::Security::ACL;
        require Win32::Security::NamedObject;
        $cifs_security = 0;
    }

    if ( $modify < 2 ) {
        # Create a list of ACEs
        foreach $user ( @users[@ulist] ) {
            foreach $mask ( @ace_masks[@mlist] ) {
                foreach $flag ( @ace_flags[@flist] ) {
                    foreach $type ( @types[@tlist] ) {
                        logVerbose(VERBOSE_ATTRS, "ACL: user: $user mask: " .
                          "$mask flag: $flag type: $type");
                        # Create ACE
                        push @aces, Win32::Security::ACE->new('FILE',
                                                              $type,
                                                              $flag,
                                                              $mask,
                                                              $user);
                    }
                }
            }
        }
    }

    # Get security for file/directory
    $object = Win32::Security::NamedObject->new('FILE', $file);

    if ( $modify ) {
        # Delete previous entry -- for modify (1) or remove (2)
        delete $acls{$ref->{$file}{-acl}} if exists($ref->{$file}{-acl});

        if ( $modify == 1 ) {
            # Modifying the ACL: remove all inherited ACEs as well
            # as a percentage as requested
            @dacl_aces = grep { ! $_->aceFlags->{INHERITED}
                                && (100-$args->{percent}) > rand(100) }
                                $object->dacl->aces();
            ###@sacl_aces = grep { ! $_->aceFlags->{INHERITED}
            ###                    && (100-$args->{percent}) > rand(100) }
            ###                    $object->sacl->aces();
            @sacl_aces = ();   ### TODO remove this line if uncomment above ###
        }

        # Remove duplicate ACEs
        foreach $ace_curr ( @dacl_aces, @sacl_aces ) {
            $index = 0;
            foreach $ace ( @aces ) {
                if ( $ace->aceType() eq $ace_curr->aceType()
                     && $ace->sid() eq $ace_curr->sid() ) {
                    # This ACE is already in the ACL so remove it
                    splice @aces, $index, 1;
                    last;
                }
                $index++;
            }
        }
    }

    # Separate ACE by type (DACL: Discretionary or SACL: System audit)
    foreach $ace ( @aces ) {
        if ( $ace->aceType() eq 'SYSTEM_AUDIT_ACE_TYPE' ) {
            push @sacl_changes, $ace;
        } else {
            push @dacl_changes, $ace;
        }
    }

    # Create an object for each type with current ACEs less the
    # ones that were removed
    if ( @dacl_changes || @dacl_aces ) {
        $new_dacl = Win32::Security::ACL->new('FILE', @dacl_aces);
        # Add the ACEs to the ACL
        $new_dacl->addAces(@dacl_changes);
        $object->dacl($new_dacl, 'UNPROTECTED_DACL_SECURITY_INFORMATION');
    }
    if ( @sacl_changes || @sacl_aces ) {
        $new_sacl = Win32::Security::ACL->new('FILE', @sacl_aces);
        # Add the ACEs to the ACL
        $new_sacl->addAces(@sacl_changes);
        ### TODO uncomment below if uncomment line 3285
        ###$object->sacl($new_sacl, 'UNPROTECTED_SACL_SECURITY_INFORMATION');
    }

    # Clear the ACL cache on the module
    # The module saves every ACE and ACL created for future use and after
    # creating 10000 ACLs the memory required is about 500MB

    # Clear ACE cache
    $cache = Win32::Security::ACE::SE_FILE_OBJECT->_rawAceCache();
    delete $cache->{$_} foreach keys %$cache;

    # Clear ACL cache
    $cache = Win32::Security::ACL::SE_FILE_OBJECT->_rawAclCache();
    delete $cache->{$_} foreach keys %$cache;

    return $entry;
}

#---------------------------------------------------------------------
# Read the raw ACL of a file/directory
#---------------------------------------------------------------------
sub _read_acl($$;) {
    my ($self, $file) = @_;
    my $acl;

    croak "Read ACL is only supported on Windows"
      if $self->{_hosttype} ne 'windows';

    if ( $cifs_security ) {
        # Load modules needed for ACL only once
        require Win32::Security::ACE;
        require Win32::Security::ACL;
        require Win32::Security::NamedObject;
        $cifs_security = 0;
    }

    # Get security for file/directory
    $file = $self->{_cwd} if $file eq '.';
    eval {
        $acl = Win32::Security::NamedObject->new('FILE', $file)->dacl->rawAcl();
    };
    croak "Error reading ACL for $file: $@" if $@;

    return $acl;
}

#---------------------------------------------------------------------
# It selects N (random) elements at random from the array and returns
# this list keeping the same order as the original list
#---------------------------------------------------------------------
sub _get_from_list(\@$$;) {
    my ($list, $len, $mode) = @_;
    my (%list_h, @list_a, $m);
    my $retry = 10;
    # Number of elements in list
    my $n = @{$list};

    $len = $n if $len > $n;

    # Number of elements to get from list
    if ( $mode == 2 ) {
        $m = $len;
    } else {
        $len = $n if $len == 0;
        $m = 1 + int(rand($len));
    }

    while ( $retry > 0 ) {
        # Weigh each element in the array
        %list_h = map { rand() => $_ } ( 0..$n-1 );
        # Shuffle the list according to the weights
        @list_a = map { $list_h{$_} } sort { $a <=> $b } keys %list_h;
        # Check if the list is complete. On Windows the random
        # number generator produces same numbers back to back
        if ( @list_a == $n ) {
            # Get only the requested number of elements
            # and re-order the elements
            @list_a = sort { $a <=> $b } @list_a[0..$m-1];
            last;
        }
        --$retry;
    }

    croak "Could not get items from acl list" if ($retry < 1);

    return @list_a;
}

#---------------------------------------------------------------------
# Get a list of objects of an specific type
#---------------------------------------------------------------------
sub _get_objects($$$$;$) {
    my $self = shift;
    my ($objects, $base, $ref, $type) = @_;
    my ($object, $key);
    $type = '' if not defined($type);

    foreach $key ( keys %$ref ) {
        # Skip internal keys
        next if $key =~ /^-/o;

        # Absolute path of object
        $object = File::Spec->catfile($base, $key);
        # Add object to list if it is of specified type
        push @$objects, $object if ($ref->{$key}{-type} eq $type);

        # Do not process further this object since
        # it does not have any children
        next if ref($ref->{$key}) ne 'HASH';

        # Process this object's children
        $self->_get_objects($objects, $object, $ref->{$key}, $type);
    }
}

#---------------------------------------------------------------------
# Get total counters by traversing the objects database
#---------------------------------------------------------------------
sub _get_totals($$;$) {
    my ($self, $ref, $total) = @_;

    foreach my $key ( keys %$ref ) {
        next if ref($ref->{$key}) ne 'HASH';

        if ( $ref->{$key}{-sparse} ) {
            $total->{sparse}++;
        } elsif ( $ref->{$key}{-link} ) {
            if ( $ref->{$key}{-link} == 1 ) {
                $total->{links}++;
            } else {
                $total->{slinks}++;
            }
        } elsif ( $ref->{$key}{-type} eq 'd' ) {
            $total->{dirs}++;
        } elsif ( $ref->{$key}{-type} eq 'f' ) {
            $total->{files}++;
        } elsif ( $ref->{$key}{-type} eq 's' ) {
            $total->{streams}++;
        }

        if ( $ref->{$key}{-acl} ) {
            $total->{acls}++;
        }

        $self->_get_totals($ref->{$key}, $total);
    }

    return $total;
}

#---------------------------------------------------------------------
# Recursively copy an item (scalar, array ref or hash ref)
# Used in the 'copy' option for modify to copy the object's tree
#---------------------------------------------------------------------
sub _copy_ref($$;) {
    my ($self, $this) = @_;
    my $new_ref;

    if ( not ref $this ) {
        # Scalar value
        return $this;
    } elsif ( ref $this eq 'ARRAY' ) {
        # Array: return a reference to a local array
        return [ map { $self->_copy_ref($_) } @$this ];
    } elsif ( ref $this eq 'HASH' ) {
        # Hash: return a reference to a local hash
        $new_ref = { map { $_ => $self->_copy_ref($this->{$_}) } keys %$this };
        $new_ref->{-mod_id} = $self->{_mod_id};
        return $new_ref;
    } else {
        # The key item from last iteration is at $_ coming
        # from the map function
        croak "Cannot copy <$_> of unknown type $this";
    }
}

#---------------------------------------------------------------------
# Create a block of random data
#---------------------------------------------------------------------
sub _create_rand_data(\$$;) {
    my ($cont, $size) = @_;
    my ($bsize, $n);
    my $block = '';

    if ( $size < 1 ) {
        $$cont = '';
        return 0;
    }

    # Use the specified size up to the maximum write size
    $bsize = $size < $MAX_WR_SIZE ? $size : $MAX_WR_SIZE;
    # Create data block of a maximum of DATA_BLOCK bytes
    $n = $DATA_BLOCK < $bsize ? $DATA_BLOCK : $bsize;

    # Create the block of random data
    $block .= pack('C', int(rand(255))) foreach (1..$n);

    # Put as many data blocks up to the maximum write size
    $$cont = $block x int($bsize/$n);
    $n = $bsize % $n;
    $$cont .= substr($block, 0, $n) if $n;

    return $bsize;
}

#---------------------------------------------------------------------
# Create a sparse file
#---------------------------------------------------------------------
sub _create_sparse($$$;) {
    my ($entry, $args, $ref) = @_;

    write_sparse(file=>$entry, data=>$args->{sparse_opts});
    $ref->{$entry}{-sparse} = [];
    foreach my $item ( @{$args->{sparse_opts}} ) {
        push @{$ref->{$entry}{-sparse}}, {
            offset => $item->{offset},
            size   => length($item->{data}),
        }
    }

    return;
}

#---------------------------------------------------------------------
# Set windows attributes on file or directory
#---------------------------------------------------------------------
sub _set_attributes($$$$;) {
    my ($object, $sattrs, $ref, $args) = @_;
    my $msg = "Changing attributes on $object:";
    ( $msg .= " $_=$sattrs->{$_}" ) foreach ( keys %$sattrs );
    logVerbose(VERBOSE_ATTRS, $msg);

    # Set attributes
    setattr_file(file=>$object, %$sattrs);

    foreach my $attr ( keys %$sattrs ) {
        # Set type of attributes for objects
        $ref->{$object}{-$attr} = $sattrs->{$attr};

        # Decrement counters
        $args->{$ref . '_num'}--;
    }
}

#---------------------------------------------------------------------
# Gets the reference for the specified file system object
# It must be relative to the path specified in constructor
#---------------------------------------------------------------------
sub _get_ref($$;) {
    my ($self, $obj) = @_;
    my ($object, $dir);
    my $ref = $self->{_objects}{'.'};
    $obj = '' if $obj eq '.';

    # Get object's path relative to object's root
    $object = File::Spec->catfile($self->{_cwd}, $obj);
    $object = File::Spec->abs2rel($object, $self->{path});

    # Remove leading '..\'
    # File::Spec->abs2rel fails when the base is a drive 'Z:'
    $object =~ s/^\.\.\\?//;

    # If abs2rel() told us that we're operating in the current dir, just
    # return $ref unchanged; skip the loop below.
    return $ref if ($object eq q{.});

    # Traverse the object's hash and return the reference
    foreach $dir ( split(/[\/\\]+/, $object) ) {
        croak "Could not find table entry for object <$object> (at elem $dir)"
          if not exists($ref->{$dir});
        $ref = $ref->{$dir};
    }

    return $ref;
}

#---------------------------------------------------------------------
# Get the type string of the object
#---------------------------------------------------------------------
sub _get_type($;) {
    my $object = shift;
    my %obj_types = (
        s => [ 'stream',    VERBOSE_STREAMS ],
        f => [ 'file',      VERBOSE_FILES ],
        d => [ 'directory', VERBOSE_DIRS ],
    );

    my @ret = $object->{'-sparse'} ? ('sparse file', VERBOSE_SPARSE) :
              $object->{'-link'} ?
              ($object->{'-link'} == 1 ? 'link':'symbolic link', VERBOSE_FILES):
              @{$obj_types{$object->{'-type'}}};

    if ( wantarray ) {
        return @ret;
    } else {
        return $ret[0];
    }
}

#---------------------------------------------------------------------
# Get the file name of the object
#---------------------------------------------------------------------
sub _get_file($$;) {
    my ($base, $key) = @_;

    return $base eq '.' ? $key : $key eq '.' ? $base :
           $base eq ''  ? $key : $key eq ''  ? $base :
           File::Spec->catfile($base, $key);
}

#---------------------------------------------------------------------
# Units conversion
#---------------------------------------------------------------------
sub _units($;) {
    my $value = shift;
    my ($val, $m, $n, $b, $mult);

    return unless defined $value;

    ($val, $n, $m, $b) = ($value =~ /(-?\d*\.?\d*)\s*((.?)(.*))/);
    croak "Value must be a number <$value>"
      unless (defined $val && $val ne '' && $val ne '.');
    $mult = $units_table{lc($m)} if $m;
    if ( ($n && !defined $mult) || ($b && lc($b) ne 'b') ) {
        croak "Bad units modifier for <$value>";
    }

    if ( defined $mult && $mult > 1 ) {
        $val *= $mult;
    }

    return $val;
}

#---------------------------------------------------------------------
# Changes to given directory, saving the current working directory
# so it could be restored with _chdir_restore
#---------------------------------------------------------------------
sub _chdir($$;) {
    my ($self, $dir) = @_;
    my $cwd;

    if ( ! defined($self->{_cwd}) ) {
        # Make sure chdir is relative to path
        # so prefix the object's path
        $dir = File::Spec->catfile($self->{path}, $dir);
        $cwd = getcwd() or croak "Could not determine current directory: $!";
    }

    if ( $dir =~ /^\.?$/ ) {
        # Do not chdir if '.' but make an entry
        # on list for _chdir_restore
        push @{$self->{_prev_cwd}}, '';
        return;
    }

    # Save the absolute path for current directory
    push @{$self->{_prev_cwd}}, defined $cwd ? $cwd : $self->{_cwd};

    # Change to directory
    logVerbose(VERBOSE_DIRS, "Changing directory to $dir");
    chdir $dir or croak "Could not change to directory $dir: $!";

    # Set new current working directory
    if ( defined($self->{_cwd}) ) {
        $self->{_cwd} = File::Spec->catfile($self->{_cwd}, $dir) if $dir ne '';
    } else {
        $self->{_cwd} = $dir;
    }

    return;
}

#---------------------------------------------------------------------
# Changes back to previous directory
#---------------------------------------------------------------------
sub _chdir_restore($;) {
    my $self = shift;
    my $prev_dir;

    if ( defined($self->{_prev_cwd}) ) {
        # Get last directory change
        $prev_dir = pop @{$self->{_prev_cwd}};
        return if not $prev_dir;

        # Change back to previous directory
        logVerbose(VERBOSE_DIRS, "Changing directory to $prev_dir --restore--");
        chdir $prev_dir or croak "Could not change to directory $prev_dir: $!";

        # Set new current working directory
        $self->{_cwd} = $prev_dir;
    }

    return;
}

#---------------------------------------------------------------------
# Save database to root directory of tree
#---------------------------------------------------------------------
sub _save_db($;) {
    my $self = shift;
    my $dbfile = $self->{db_file}; #File::Spec->catfile($self->{path}, $DB_FILE);

    sysopen(FH, $dbfile, O_WRONLY|O_BINARY|O_CREAT|O_TRUNC)
      or croak "Could not create db file: $!";
    print FH nfreeze($self) or croak "Could not save db file: $!";
    close(FH);
}

#---------------------------------------------------------------------
# Read database from root directory of tree
#---------------------------------------------------------------------
sub _get_db($;) {
    my $self = shift;
    my $dbfile = $self->{db_file}; # File::Spec->catfile($self->{path}, $DB_FILE);
    my ($buffer, $n);
    my $nread = 0;
    my @ckeys = qw(
        _srand
        _remove
        _mod_id
        _objects
        _counters
    );

    # Open db file
    sysopen(FH, $dbfile, O_RDONLY|O_BINARY)
      or croak "Could not open db file: $!";
    # Read the entire file into buffer
    while ( ($n = sysread(FH, $buffer, 10000, $nread)) ) {
        $nread = length($buffer);
    }
    croak "Could not read db file: $!" if not defined($n);
    close(FH);

    # Thaw the contents of file into a blessed object
    $n = thaw($buffer);

    # Copy all keys from saved object to current object
    ($self->{$_} = $n->{$_}) foreach @ckeys;

    return 1;
}

#---------------------------------------------------------------------
# Internal function
# Logs a message if environment is set to verbose.
#  VERBOSE_DIRS    - show only operations on directories
#  VERBOSE_FILES   - show only operations on files
#  VERBOSE_STREAMS - show only operations on streams
#  VERBOSE_SPARSE  - show only operations on sparse files
#  VERBOSE_ATTRS   - show only operations on attributes
#---------------------------------------------------------------------
sub logVerbose($$;) {
    my ($level, $msg) = @_;
    if ( defined $main::DATA_VERBOSE && $main::DATA_VERBOSE & $level ) {
        $_use_nate ? Tharn::logcomment($msg) : print "$msg\n";
    }
    return;
}

#---------------------------------------------------------------------
# Return the integer of a random number from a Gaussian distribution
# centered about the given mean with the given standard deviation.
# The number is incremented with a uniform probability if the random
# number is not exactly an integer.
#---------------------------------------------------------------------
sub intRandGauss($$;) {
    my ($mean, $stddev) = @_;
    my ($num, $int_num, $fraction);

    $num = randGauss($mean, $stddev);
    $int_num = int($num);
    $fraction = $num - $int_num;
    if ( $fraction > 0 ) {
        $int_num++ if rand() < $fraction;
    }

    return $int_num;
}

#---------------------------------------------------------------------
# Return a random number from a Gaussian distribution centered
# about the given mean with the given standard deviation.
#---------------------------------------------------------------------
{
    my $spare = undef; # static variable across calls to randGauss()

    sub randGauss($$;) {
        my ($mean, $stddev) = @_;
        return $mean unless ($stddev);

        my $r;

        if ($spare) {
            $r = $spare;
            $spare = undef;
        } else {
            # Get two random numbers from the Gaussian distribution
            # centered at zero with a standard deviation of 1.
            my ($x1, $x2, $w, $y1, $y2);
            do {
                $x1 = 2.0 * rand() - 1.0;
                $x2 = 2.0 * rand() - 1.0;
                $w = ($x1 * $x1) + ($x2 * $x2);
            } while ( $w >= 1.0 );
            $w = sqrt( ( -2.0 * log($w) ) / $w );
            $y1 = $x1 * $w;
            $y2 = $x2 * $w;

            $r = $y1;
            $spare = $y2;
        }

        # Map the random number into the desired distribution.
        $r = ($r * $stddev) + $mean;
        return $r;
    }
}

#---------------------------------------------------------------------
# Destructor for object
# It terminates the RPC service on the client if this is the owner
# of such service
#---------------------------------------------------------------------
sub DESTROY {
    my $self = shift;
    my $dbfile;
    my $tobjs = 0;

    # Save current error
    local($@);

    # Make sure only the owner of object runs the destructor
    return if $self->{_pid} != $$;

    # Get number of top level objects
    if ( defined($self->{_objects}) ) {
        $tobjs = grep { ! /^-/o } keys %{$self->{_objects}{'.'}};
    }

    # Save the database file if needed
    if ( $tobjs > 0 && $self->{save} ) {
        $self->_save_db();
        logVerbose(VERBOSE_INFO, "Saving database");
    }

    # Remove database file if there are no objects
    $dbfile = $self->{db_file}; # File::Spec->catfile($self->{path}, $DB_FILE) if ! $tobjs;
    if ( ! $tobjs && -e $dbfile ) {
        unlink $dbfile
          or $self->logWarn("Could not remove database file $dbfile: $!");
    }

    # Terminate daemon
    if ( defined($self->{_sobj}) ) {
        eval {
            $self->{_sobj}->ipc->quit();
            $self->{_sobj}->wait_finish();
        };
        $self->logWarn($@) if $@;
    }

    return;
}

1;

