# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/dp/snapmirror/vsm/restartability/Restart.pm#1 $
#
# Copyright 2005 Network Appliance, Inc.
# All rights reserved.
#
## @summary
##  Common testscript code for testing SnapMirror restartability
##  **ONLY for use by tests in dp/snapmirror/vsm/restartability**
##
## @status private
##
## @author tapl@netapp.com

# NOTE: This package is a subclass of NOTE::SnapMirror
package dp::snapmirror::vsm::restartability::Restart;
our @ISA = (NOTE::SnapMirror);

use strict;
use Tharn;
use LogReader::SnapMirror;
use dp::TestWrapper;
use Carp;
use SayWrap;
use SnapMirror;
use Params::Validate qw(SCALAR ARRAYREF BOOLEAN 
                        CODEREF HASHREF OBJECT SCALARREF);

BEGIN {
    Params::Validate::validation_options(
         on_fail => sub { my @msgs = @_; chomp(@msgs);
			  Carp::confess(Tharn::script_result(@msgs)) }
    );
}

use base qw(Exporter);
our @EXPORT_OK = qw(
    restart_testscript
);

# parameters
param("PHASE",    -default, "all");
param("USE_FIJI", -default, "1");

#################################################################
#
# Regular procedural method
#
#################################################################
sub restart_testscript {
    my (%args) = Params::Validate::validate( @_, 
			       { pretest  => { type => HASHREF,
					       optional => 1 },
				 testcase => { type => HASHREF,
					       optional => 1 },
				 posttest => { type => HASHREF,
					       optional => 1 },
				 teardown => { type => HASHREF,
					       optional => 1 } } );

    my $rsltobj;

    my @phases = _get_phases();

    foreach my $phase (@phases) {
	eval {
	    logcomment("Start phase: $phase");

	    # pass references of functions to these wrappers
	    # add 'phase' to the params
	    if (exists $args{pretest}) {
		@{$args{pretest}{params}} = ($phase);
		dp::TestWrapper::pretest($args{pretest})
	    }
	    if (exists $args{testcase}) {
		@{$args{testcase}{params}} = ($phase);
		dp::TestWrapper::testcase($args{testcase})
	    }
	    if (exists $args{posttest}) {
		@{$args{posttest}{params}} = ($phase);
		dp::TestWrapper::posttest($args{posttest}) 
	    }
	};
	$rsltobj = $@;

	# call the teardown callback function.
	($args{teardown}->{func})->( @{$args{teardown}->{params}} )
	  if (exists $args{teardown});

	last if ($rsltobj);
    }

    logresult($rsltobj) if ($rsltobj);
}

sub _get_phases {
    my @phases;

    my @split = split(/,| /, $::PHASE);
    foreach my $phase (@split) {
	if ($phase =~ /^initialize$/i) {
	    push(@phases, lc($phase));

	} elsif ($phase =~ /^update$/i) {
	    push(@phases, lc($phase));

	} elsif ($phase =~ /^resync$/i) {
	    push(@phases, lc($phase));

	} elsif ($phase =~ /^all$/i) {
	    # user wants everything. greedy...
	    @phases = ();
	    push(@phases, 'update', 'resync', 'initialize');
	    last;

	} else {
	    croak "Unrecognized phase: $phase";
	}
    }

    return @phases;
}


#################################################################
#
# Methods extended from the NOTE::SnapMirror class
#
#################################################################

# make sure to create an 'empty' object first via the new() constructor
sub create_checkpoint {
    my ($self, $phase, %args) = @_;

    my $prepopulate = 1;

    if (exists $args{prepopulate}) {
	$prepopulate = ($args{prepopulate}) ? 1 : $args{prepopulate};
    }

    # both src and dst volumes must be there
    if (!defined($self->src) || !defined($self->dst)) {
	my @caller = caller(0);
	croak "Both source and destination volumes must be created prior ".
	  "to calling " . $caller[1] . ":" . $caller[3];
    }

    $self->{_phase} = $phase;

    # track the snapshots that were created before each transfer.
    # testscripts may need this info to delete the snapshots to
    # test restartability.
    my @created_sshots;

    if ($prepopulate) {
        @created_sshots = $self->populate(%args);
     }

    if ($self->phase eq 'initialize') {
	logcomment("Interrupting $phase and creating a checkpoint");
	$self->set_restart_fiji;

	# timeout after 30mins
	eval { $self->initialize(timeout=>1800) };

    } elsif ($self->phase eq 'update') {
	$self->initialize;
	push(@created_sshots, $self->generate_data_churn(%args)) if ($prepopulate);

        logcomment("Interrupting $phase and creating a checkpoint");
        $self->set_restart_fiji;	      
        eval { $self->update(timeout=>1800) };

    } elsif ($self->phase eq 'resync') {
	logcomment("Creating a VSM relationship and breaking it");
	$self->initialize(timeout=>1800);
	$self->break;
	push(@created_sshots, $self->generate_data_churn(%args)) if ($prepopulate);

	logcomment("Interrupting $phase and creating a checkpoint");
	$self->set_restart_fiji;	      
	eval { $self->resync(timeout=>1800) };

    } else {
	croak "Unknown phase: " . $self->phase;
    }

    my $sm_status = $self->status;

    # verify that checkpoint was created. this will be verified again via
    # /etc/log/snapmirror log in verify_restart() method.
    if ($sm_status->{'status'} ne 'idle with restart checkpoint' &&
	$sm_status->{'current-transfer-error'} 
	ne 'snapmirror transfer failed to complete') {
	croak "Restart checkpoint was not created.";
    }

    return @created_sshots;
}

# overrides the base class initialize method, use for tests when
# special processing is needed after the base snapshot is created
# but before the transfer is aborted
# make sure to create an 'empty' object first via the new() constructor
sub initialize {
    my ($self, %args) = @_;

    # already initialized, must do this check to support previous versions
    # of callers to base initialize
    if ($self->{_init}){
        return;
    }

    # track the snapshots that were created before each transfer.
    # testscripts may need this info to delete the snapshots to
    # test restartability.
    my @created_sshots;

    # if phase is set only need to call base behavior
    if (!$self->{_phase}){
        my $prepopulate = 1;

        if (exists $args{prepopulate}) {
           $prepopulate = ($args{prepopulate}) ? 1 : $args{prepopulate};
        }

        # both src and dst volumes must be there
        if (!defined($self->src) || !defined($self->dst)) {
            my @caller = caller(0);
            croak "Both source and destination volumes must be created prior ".
              "to calling " . $caller[1] . ":" . $caller[3];
        }

        if ($prepopulate) {
            @created_sshots = $self->populate(%args);
        }
    }

    # call the base class initialize
    # timeout after 30mins
    $self->NOTE::SnapMirror::initialize(timeout=>1800);

    $self->{_init} = 1; 
    return @created_sshots;
}

sub restart_xfer {
    my ($self) = @_;

    logcomment("Restarting interrupted '" . $self->phase . "' transfer");

    # find out where the checkpoint was created
    $self->set_chkpt_marker;

    # start LogReader
    $self->start_LogRdr;

    if ($self->phase eq 'initialize') {
	# directly call SnapMirror::snapmirror_initialize() and bypassing
	# all the setup in the inherited initialize() method
	SnapMirror::snapmirror_initialize(host=>$self->dst->Filer->hostrec,
					  dst =>$self->dst_path);

    } elsif ($self->phase eq 'update') {
	$self->update;

    } elsif ($self->phase eq 'resync') {
	$self->resync;

    } else {
	croak "Unknown phase: " . $self->phase;
    }

    # verify that it completed transfer
    if ($self->status('state') ne 'snapmirrored') {
	croak "Aborted '" . $self->phase . "' transfer did not complete";
    }

    # clear the fiji just in case
    $self->clear_restart_fiji;
}

sub start_new_xfer {
    my ($self, $state) = @_;

    logcomment("Starting new transfer via $state");

    # restart the LogReader
    $self->LogRdr->restart;

    $self->$state;

#     if ($state eq 'update') {
# 	$self->update;
#     } elsif ($state eq 'resync') {
# 	$self->resync;
#     } else {
# 	croak "Unknown state: $state";
#     }
}

sub populate {
    my ($self, %args) = @_;
    my @created_sshots;

    # use populate or fill
    if (exists ($args{populate_args})) {

        # prepopulating source volume before snapmirror
        logcomment("Pre-populating with populate " . $self->src->description);

        if ($self->{_modify}) {
            # volume already populated
            $self->src->delete();
        }

        # fill the source with files and create some
        # data churn
        $self->src->populate(%{$args{populate_args}});
        push(@created_sshots, $self->src->snapshot_create);
        for (my $i = 1; $i < 2; ++$i) {

            $self->src->modify(%{$args{modify_args}});
            push(@created_sshots, $self->src->snapshot_create);
        }
        # set the flag to use populate/modify
        $self->{_modify} = 1;

    } else {

        # prepopulating source volume before snapmirror
        logcomment("Pre-populating with fill " . $self->src->description);

        my $done_refill = 0;
        for (my $i = 1; $i < 3; ++$i) {
            # src volume is filled to at least 20% plus current capacity
            # but never more than 35%. need the remaining space to create
            # some data churn.
            my $cap = $self->src->capacity;
            my $fill_percent = int( ((10 * $i) + $cap) % 35 );
            $self->src->fill(percent=>$fill_percent);
            push(@created_sshots, $self->src->snapshot_create);

            # refill only once
            if (!$done_refill) {
                $self->src->refill;
                $done_refill++;
            }
        }
    }

    return @created_sshots;
}

sub verify_restart {
    my ($self) = @_;

    logcomment("Verifying that interrupted '" . $self->phase . "' transfer " .
	       "was restarted");

    # stop the LogReader
    $self->LogRdr->stop;

    my $regex = quotemeta($self->dst_path . " Restart (@ " .
			  $self->chkpt_marker . " KB)");
    my $matches = $self->LogRdr->matches($regex);

    # no match. print snippet of snapmirror log
    if (scalar @{$matches} < 1) {
	die "Phase " . $self->phase . ": " .
	  "Transfer did not continue from checkpoint.\n" . 
	  $self->LogRdr->to_string;
    }
}

sub verify_no_restart {
    my ($self) = @_;

    my $regex;
    my $matches;
    my $restarted = 0;

    logcomment("Verifying that interrupted '" . $self->phase . "' transfer " .
	       "was not restarted");

    # stop the LogReader
    $self->LogRdr->stop;
    
    # first check for 'Abort (Cannot restart)'. if it's there, we know VSM
    # did not restart. if it's not there, then look for 'Restart (@ XXX KB)'.
    # there have been cases whereby both strings were in the log file. The
    # 'Abort (Cannot restart)' string wins.
    $regex = quotemeta($self->dst_path . " Abort (Cannot restart)");
    $matches = $self->LogRdr->matches($regex);

    # nothing, try the other string
    if (scalar @{$matches} < 1) {
	$regex = quotemeta($self->dst_path . " Restart (@ " .
			   $self->chkpt_marker . " KB)");
	$matches = $self->LogRdr->matches($regex);
	$restarted = 1 if (scalar @{$matches} > 1)
    }

    # no match. print snippet of snapmirror log
    if ($restarted) {
	die "Phase " . $self->phase . ": " .
	  "Transfer continued from checkpoint.\n" . 
	  $self->LogRdr->to_string;
    }
}

sub generate_data_churn {
    my ($self, %args) = @_;
    my @sshots;

    if ($self->{_modify} or exists $args{modify_args}) {

        logcomment("Modifying ". $self->src->description);

        for (my $i = 1; $i < 4; ++$i) {
            $self->src->modify(%{$args{modify_args}});
            push(@sshots, $self->src->snapshot_create);
        }
        return @sshots;
    }
             
    logcomment("Filling up " . $self->src->description);

    # check if the fill file is there
    eval { $self->src->fillfile };
    if ($@) {
	# not found. create one.
	$self->src->fill(percent=>15);
    }

    my $refill_once = 1;

    for (my $i = 1; $i < 4; ++$i) {
	# the code tries to simulate data creation in between snapshot
	# creation/deletion overlapping with other snapshots. the goal
	# is to have some of the snapshots holding on the same blocks
	# some snapshots holding on to blocks not held by other snapshots.

	# if we refill too much, we'll run out of space on the vol to
	# create another snapshots
	if ($self->phase eq 'update') {
	    $self->src->refill if ($refill_once-- == 1);
	} else {
	    $self->src->refill;
	}

	push(@sshots, $self->src->snapshot_create);
    }
    return @sshots;
}

sub set_restart_fiji {
    my ($self) = @_;

    # XXX use of parameter... yuck!
    return if (!$::USE_FIJI);

    # abort after 2MB or 500 blocks of data have been written to disk
    # it's 'guaranteed' that there is at least 3MB of data churn to be
    # transferred to the dst volume even if the src volume is only 20MB
    # in size.
    my $out = SayWrap::sayEOF($self->dst->Filer->conn, "fiji filter \"(replica.do.checkpoint(&&(eq \@vol ".$self->dst->name.")(> \@blocks 500))(return 1))\"");

    if ($out !~ /filter set/) {
	croak "Did not set fiji filter: $out";
    }

    $self->{_fiji_set} = 1;
}

sub clear_restart_fiji {
    my ($self) = @_;

    # XXX use of parameter... yuck!
    return if (!$::USE_FIJI);
    return if (!$self->fiji_set);

    my $out = SayWrap::sayEOF($self->dst->Filer->conn, "fiji clear all");
    
    if ($out !~ /filter cleared/) {
	croak "Cannot clear fiji filters: $out";
    }
}

sub set_chkpt_marker {
    my ($self) = @_;
    $self->{_chkpt_marker} = $self->status('transfer-progress');
}

sub chkpt_marker {
    return shift->{_chkpt_marker};
}

sub clear_chkpt_marker {
    my ($self) = @_;
    $self->{_chkpt_marker} = undef;
}

sub LogRdr {
    return shift->{_LogRdr};
}

sub start_LogRdr {
    my ($self) = @_;

    $self->{_LogRdr} = LogReader::SnapMirror->new($self->dst->Filer->conn);

    # start LogReader
    $self->LogRdr->start;
}

sub clear_LogRdr {
    my ($self) = @_;
    $self->{_LogRdr} = undef;
}

sub phase {
    return shift->{_phase};
}

sub fiji_set {
    return shift->{_fiji_set};
}

sub DESTROY {
    my ($self) = @_;

    # clear the fiji if it's still there
    $self->clear_restart_fiji;

    # check for an overriden destructor and call the regular
    # destructor
    $self->SUPER::DESTROY if $self->can("SUPER::DESTROY");
}

1;

