#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/ASUP.pm#1

#***********************************************************************
#
# NAME
#   ASUP.pm - 
#
# SYNOPSIS
#   ASUP Library for the invocation and verfication of 7M and CM SN ASUP.
#
# DESCRIPTION
#   This library is a wrapper for the typical ASUP commands that an end user
#   may desire for the configuration, invocation, and verification of
#   autosupports.  The current library was designed to use Server.pm and NACL,
#   7M or CM SN filer.
#
# AUTHOR
#   dl-cluster-qa
#
# HISTORY
# 7/13/2009 Initial revision
# 8/2/2009 Updates to support minimal content and updates to get_asup_config
# 1/12/2010 1) Updated version check routine
#           2) Updated default time limit for verify_asup_reception routine
#               from 150s to 180s
#           3) invoke_autosupport will display error messages in log
#           4) Added routines for invoke_nht_autosupport,
#               invoke_performance_autosupport, and invoke_weekly_autosupport
#           5) Added get_autosupport_filter and modify_autosupport_filter
#               routines
#           6) Removed UNOWNED_DISKS section from basic check in
#               _verify_sections routine.  We now check for disk ownership
#               before looking for this section.
# 1/31 Adding support for Rolling Rock
#           1) Modify the verify_autosupport_format routine to check for version
#           2) Add RR verify_autosupport_format_RR routines
# 4/28/2011 Making the Library Test Branch Specific for SN
#           1) Remove all checks for BR, 7G, and RR
#           2) Remove the _verify_sections functions
#***********************************************************************

## @summary     ASUP library    

package ASUP;

#use statements
use strict;
use Data::Dumper;
use Server;
my $Server = new Server();

use Tharn qw($Log snooze host connect script_result fatal_result);
use Options;
use Params::Common;
use Params::Validate qw(:all);

use EMS;
use SMTP_Server;
use HTTP_Server;
use HTTPS_Server;
use HTTPS_OnDemand_Server;

use Carp;
use Sys::Hostname;

use File::Path;
use File::Basename;
use XML::LibXML;
use XML::XPath;
use NACL::APISet;
use Clone;
use DataUtils;
use NACL::C::Vserver;
use NACL::C::Node;
use NACL::C::VserverServicesDnsHosts;
use IPC::Cmd qw(can_run);
#use SCON_Utils;
#use Archive::Extract;

#Global variables
our $ontap_version;
our $DEBUG_BURT674766 = Tharn::param("DEBUG_BURT674766");
our $DEBUG_BURT662241 = Tharn::param("DEBUG_BURT662241");

our $GREP_EXEC = can_run("grep");

my @Nodes=$Server->servers();

my $opts = new Options;

$opts->parameters(definitions=>
    [
       {
        name => "DEFAULT_ASUP_HTTP_PUT_URL",
        description => "Should the ASUP PUT url be configured by default. Default is yes.",
        default => 1,
       },
    ]);

my $opt_href = $opts->get_hash();

my $enable_regex = qr/1|on|ON|yes|YES|true|TRUE|0|off|OFF|no|NO|false|FALSE|enable|disable/;
my $content_regex = qr/minimal|complete/i;
my $transport_regex = qr/HTTP|HTTPS|SMTP/i;
my $id_regex = qr/hostname|systemid|true|false/i;


#7M autosupport options
my %args = (
            "enable" => "enable",
            "content" => "content",
            "from" => "address",
            "to" => "address",
            "support_to" => "address",
            "partner_to" => "address",
            "noteto" => "address",
            "cifs_verbose" => "enable",
            "local_nht_data_enable" => "enable", 
            "local_performance_data_enable" => "enable", 
            "nht_data_enable" => "enable", 
            "performance_data_enable" => "enable",
            "support_enable" => "enable",
            "throttle" => "enable",
            "doit" => "message",
            "performance_doit" => "message",
            "retry_count" => "count",
            "retry_interval" => "interval",
            "minimal_subject_id" => "id",
            "mailhost" => "host",
            "support_proxy" => "proxy",
            "support_transport" => "transport",
            "support_url" => "url",
            "change_filext" => "enable",
            "put_url" => "put-url",
            "max_http_size" => "count",
            "max_smtp_size" => "count",
            "payload_format" => "payload_format",
            "validate_digital_certificate" => "enable",
            "periodic_tx_window" => "periodic_tx_window",
            "ondemand_state" => "enable",
            "ondemand_server_url" => "url",
            "ondemand_polling_interval" => "interval",
            "ondemand_remotediag_state" => "enable",
           );

#GX/CM autosupport options
my %gx_args = (
            "enable" => "state",
            "from" => "from",
            "to" => "to",
            "support_to" => "support-address",
            "partner_to" => "partner-address",
            "noteto" => "noteto",
            "local_nht_data_enable" => "local-nht", 
            "local_performance_data_enable" => "local-perf", 
            "nht_data_enable" => "nht", 
            "performance_data_enable" => "perf",
            "support_enable" => "support",
            "throttle" => "throttle",
            "retry_count" => "retry-count",
            "retry_interval" => "retry-interval",
            "minimal_subject_id" => "hostname-subj",
            "mailhost" => "mail-hosts",
            "support_proxy" => "proxy-url",
            "support_transport" => "transport",
            "support_url" => "url", 
            "change_filext" => "change-file-ext",
            "log_file_size" => "log-file-size",
            "reminder" => "reminder",
            "put_url" => "put-url",
            "max_http_size" => "max-http-size",
            "max_smtp_size" => "max-smtp-size",
            "payload_format" => "payload-format",
            "validate_digital_certificate" => "validate-digital-certificate",
            "remove_private_data" => "remove-private-data",
            "periodic_tx_window" => "periodic-tx-window",
            "local_collection" => "local-collection",
            "ondemand_state" => "ondemand-state",
            "ondemand_server_url" => "ondemand-server-url",
            "ondemand_polling_interval" => "ondemand-polling-interval",
            "ondemand_remotediag_state" => "ondemand-remotediag-state",
           );

#Methods

###############################################################################
#Sub: config_asup
#Input Params: hash
#Return values: 1 (success) or the error if a failure
#Summary: Configure autosupport settings on the filer
###############################################################################

sub config_asup {
    my %params = validate(
    @_,
    {
        cifs_verbose => { type  => SCALAR,
                          optional => 1,
                          regex => $enable_regex,
                        },
             content => { type => SCALAR,
                          optional => 1,
                          regex => $content_regex,
                        },
              enable => { type => SCALAR,
                          optional => 1,
                          regex => $enable_regex,
                        },
                from => { type => SCALAR,
                          optional => 1,
                        },
        local_nht_data_enable => { type => SCALAR,
                          optional => 1,
                          regex => $enable_regex,
                        },
        local_performance_data_enable => { type => SCALAR,
                          optional => 1,
                          regex => $enable_regex,
                        },
        minimal_subject_id => { type => SCALAR,
                          optional => 1,
                          regex => $id_regex,
                        },
            mailhost => { type => SCALAR,
                          optional => 1,
                        },
        nht_data_enable => { type => SCALAR,
                          optional => 1,
                          regex => $enable_regex,
                        },
              noteto => { type => SCALAR,
                          optional => 1,
                        },
        performance_data_enable => { type => SCALAR,
                          optional => 1,
                          regex => $enable_regex,
                        },
         retry_count => { type => SCALAR,
                          optional => 1,
                        },
        retry_interval => { type => SCALAR,
                          optional => 1,
                        },
        support_enable => { type => SCALAR,
                          optional => 1,
                          regex => $enable_regex,
                        },
         support_proxy => { type => SCALAR,
                          optional => 1,
                        },
          support_to => { type => SCALAR,
                          optional => 1,
                        },
        support_transport => { type => SCALAR,
                          optional => 1,
                          regex => $transport_regex,
                        },
         support_url => { type => SCALAR,
                          optional => 1,
                        },
            throttle => { type => SCALAR,
                          optional => 1,
                          regex => $enable_regex,
                        },
                  to => { type => SCALAR,
                          optional => 1,
                        },
          partner_to => { type => SCALAR,
                          optional => 1,
                        },
       log_file_size => { type => SCALAR,
                          optional => 1,
                        },
     change_filext => { type => SCALAR,
                          optional => 1,
                        },
            reminder => { type => SCALAR,
                          optional => 1,
                        },
             put_url => { type => SCALAR,
                          optional => 1,
                        },
      payload_format => { type => SCALAR,
                          optional => 1,
                        },
       max_http_size => { type => SCALAR,
                          optional => 1,
                        },
       max_smtp_size => { type => SCALAR,
                          optional => 1,
                        },
       validate_digital_certificate => { type => SCALAR,
                          optional => 1,
                          regex => $enable_regex,
                        },
       periodic_tx_window => { type => SCALAR,
                          optional => 1,
                        },
       remove_private_data => { type => SCALAR,
                          optional => 1,
                          regex => $enable_regex,
                        },
       local_collection => { type => SCALAR,
                          optional => 1,
                          regex => $enable_regex,
                        },       
       ondemand_state => { type => SCALAR,
                           optional => 1,
                           regex => $enable_regex,
                         },
       ondemand_server_url => { type => SCALAR,
                            optional => 1,
                         },
       ondemand_polling_interval => { type => SCALAR,
                         optional => 1,
                         },
       ondemand_remotediag_state => { type => SCALAR,
                                       optional => 1,
                                       regex => $enable_regex,
                                     },
       strict_ondemand_polling_interval_checking => { type => SCALAR,
                         optional => 1,
                         },
    },
    );

    $Log->comment("Performing configuration on node ".$Server->get_execution_server()." : ".Dumper(\%params));

    my $version=_get_version();
    my $errors;

    if ( ( $params{support_url} ) &&
         (!exists $params{put_url}) &&
         $opt_href->{DEFAULT_ASUP_HTTP_PUT_URL} ) {

        $params{put_url} = $params{support_url};
    } 

    my $strict_ondemand_polling_interval_checking = delete $params{strict_ondemand_polling_interval_checking};

    my $cmode_options_to_set;
    my $debug_smdb_table_options_to_set;

    while ( my ($key, $value) = each(%params) ) {
        
        #7M
        if ($version=~/7M/) {
            my $output=undef;
            my @compare;
            my $cdef = "autosupport_".$key;
            if (($key eq "log_file_size") or ($key eq "reminder") or ($key eq "remove_private_data")) {
                $Log->comment("Options log_file_size, reminder, and remove_private_data not supported in 7mode.  These settings will be ignored. \n");
            }
            else {
                eval{
                    $output = $Server->$cdef($args{$key}=>$params{$key});
                    my $modkey=$key;
                    $modkey=~s/_/./g;
                    if ($modkey eq 'put.url') {
                        $modkey = 'support.put_url';
                    }
                    @compare = split(/(autosupport.$modkey\s+)/,$Server->$cdef());
                };
                if ($args{$key} eq "enable") {
                    if (($params{$key} =~ /1|on|ON|yes|YES|true|TRUE|enable/) && ($compare[2] !~ /on/)
                            || ($params{$key} =~ /0|off|OFF|no|NO|false|FALSE|disable/) && ($compare[2] !~ /off/)) {
                        $Log->comment("Verification of autosupport option modification failed for $key. Value passed in did not match value verified...\n");
                        $errors.=$output;
                    }
                }
                elsif ($args{$key} eq "address") {
                    if ($key eq 'partner_to' and ($Server->get_release(format=>"cooked"))=~/(^7\.2)|(^[gG]ord)/) {
                        $Log->comment("Options partner_to not supported in GB.  The setting will be ignored. \n");
                    }
                    else {
                        my @addresses=split(',',$params{$key});
                
                        foreach my $address (@addresses) {
                            $address=~s/\s//g;
                            if ($address ne "" and $address ne '""' and $compare[2] !~ $address) {
                                $Log->comment("Verification of autosupport option modification failed for $key. Value passed in did not match value verified...\n");
                                $errors.=$output;
                            }
                        }
                    }
                }
                elsif ($params{$key} ne "" and $params{$key} ne '""' and  $compare[2] !~ /$params{$key}/i)  {
                    $Log->comment("Failed to verify $key was set correctly \n");
                    $Log->comment("Expected value: $params{$key} and found value: $compare[2] \n");
                    $errors.=$output;
                }
            }
        }
        #GX/CM
        else {
            if ($key eq "content" or $key eq "cifs_verbose") {
                $Log->comment("Content type and CIFS Verbose options are not yet supported in cluster mode.  These settings will be ignored. \n");
            }
            elsif ($key eq "change_filext") {
                $Log->comment("Change FileExt flag no longer supported in this Version.  This settings will be ignored. \n");
            }
            elsif ($key eq "log_file_size" ) {
                $Log->comment("Option log_file_size not supported in this Version.  This setting will be ignored. \n");
            }            
            else {
                if ($key eq "enable" or $key eq "support_enable" or $key eq "ondemand_state" or $key eq "ondemand_remotediag_state") {
                    if ($params{$key} =~ /1|on|ON|yes|YES|true|TRUE/) {
                        $params{$key}="enable";
                    }
                    if ($params{$key} =~ /0|off|OFF|no|NO|false|FALSE/) {
                        $params{$key}="disable";
                    }
                }
                if ($key =~ /minimal_subject_id|local_nht_data_enable|local_performance_data_enable|nht_data_enable|performance_data_enable|reminder|throttle|change_filext|validate_digital_certificate|remove_private_data/) {
                    if ($params{$key} =~ /1|on|ON|yes|YES|TRUE|true|hostname/) {
                        $params{$key}="true";
                    }
                    if ($params{$key} =~ /0|off|OFF|no|NO|FALSE|false|systemid/) {
                        $params{$key}="false";
                    }
                }
                if ($key eq "from") {
                    if ($params{$key} eq "Postmaster") {
                        $params{$key}='Postmaster@localnode';
                    }
                }

                $cmode_options_to_set->{$gx_args{$key}} = $params{$key};

                # For some tests we want to verify that the CLI rejects < 5
                # but typically we would like to be able to run this as quickly
                # as possible
                if ( !$strict_ondemand_polling_interval_checking ) {
                    if ( ($key eq 'ondemand_polling_interval') &&
                         ($params{$key} < 5) ) {
                        delete $cmode_options_to_set->{$gx_args{$key}};
                        $debug_smdb_table_options_to_set->{$gx_args{$key}} = $params{$key}
                    }
                }
            }
        }
    }

    if ( $debug_smdb_table_options_to_set ) {
        eval {
            $Server->debug_smdb_table_nd_autosupport_modify(%$debug_smdb_table_options_to_set);
        };
        my $eval_rc =$@;
        if ($eval_rc) {
            # We will see the parameters printed out in the command above, no
            # point in printing them out in the error message
            $Log->comment("ASUP::config_asup(): Failed to set asup parameters\n");
            return $eval_rc;
        }
    }

    if ( $cmode_options_to_set ) {
        eval {
            $Server->system_node_autosupport_modify(node=>"local", %$cmode_options_to_set);
        };
        my $eval_rc =$@;
        if ($eval_rc) {
            # We will see the parameters printed out in the command above, no
            # point in printing them out in the error message
            $Log->comment("ASUP::config_asup(): Failed to set asup parameters\n");
            return $eval_rc;
        }
    }

    if ( $errors ) {
        return $errors;
    } else {
        return 1;
    }
}

###############################################################################
#Sub: get_asup_config
#Input Params: None
#Return values: option hash
#Summary: Gets the current asup configuration on the filer and returns the hash
###############################################################################

sub get_asup_config {
    my @tmp_val;
    my %return_hash;
    my $output;

    my $version=_get_version();

    #7M
    if ($version=~/7M/) {
        for my $key ( keys %args ) {
            my $cdef = "autosupport_".$key;
            my $option=$cdef;
            $option=~s/\_/\./g;
            eval{
                $output = $Server->$cdef(); 
            };
            if ($key !~ /doit/) {
                if ($output =~ "No such") {
                    $return_hash{$key} = $output;
                }
                else {
                    $option =~ s/autosupport\.put/autosupport\.support\.put/g;
                    @tmp_val = split(/$option\s+/,$output);
                    $tmp_val[1]=~s/\s*\(value might be overwritten in takeover\)//g;
                    $tmp_val[1]=~s/\s//g;
                    if (($tmp_val[1] eq undef) || ($tmp_val[1] eq '\'\'')) {
                        $tmp_val[1]='""';
                    }
                    $return_hash{$key} = $tmp_val[1];
                }
            }
        }
    }
    #GX/CM
    else {
        my $config;
        eval {
            $config=$Server->system_node_autosupport_show(node=>"local");
        };
        for my $key (keys %gx_args){
            if ($config->[0]->{$gx_args{$key}}) {
                if (ref($config->[0]->{$gx_args{$key}}) ne "ARRAY") {
                    $return_hash{$key}=$config->[0]->{$gx_args{$key}};
                }
                else {
                    my $config_value=join(',',@{$config->[0]->{$gx_args{$key}}});
                    $return_hash{$key}=$config_value;
                    }
            }
        }
    }
    
    return \%return_hash;
}

###############################################################################
#Sub: get_asup_destinations
#Input Params: None
#Return values: option hash
#Summary: Gets the current asup destinations for a filer and returns the
#           destination hash
###############################################################################

sub get_asup_destinations {
    my $asup_dest;
    my $version=_get_version();

    if ($version=~/7M/) {
        my @cli_result=split('\n',$Server->cli(cmd=>"autosupport destinations show"));
        my $i=0;
        foreach my $result (@cli_result) {
            if (($result =~ /---/) || ($result =~ /Destinations/) || ($result eq '')) {
                next;
            } 
            $asup_dest->[0]->{'destinations'}->[$i]=$result;
            $i++;
        }
    }
    else { # Cluster mode
        eval {
            $asup_dest=$Server->system_node_autosupport_destinations_show(node=>"local");
        };
    }

    return $asup_dest;
}

###############################################################################
#Sub: invoke_and_verify_asup
#Input Params: Hash
#Return values: 1 (success) or 0 (failure)
#Summary: Combination routine that will:
#   1.  Start up a HTTP, HTTPS, or SMTP server.
#   2.  Invoke the specified autosupport event, or manually invoke one if no
#           event was specified.
#   3.  Verify that the autosupport was received by the server.
#   4.  If no asup save dir was specified, it will return a 1 or a 0 here based
#           upon whether the asup was received or not.
#   5.  If an asup save dir was specified, the routine then attempts to validate
#           the asup format:
#       A) validate the headers for the asup
#       B) validate the sections in the asup
#       C) validate the attachments of the asup
#   6.  Then returns 0 or 1 based on those validations.
#
###############################################################################
sub invoke_and_verify_asup {
    my %params = (
            event=>undef,   #the asup to generate
            asup_save_dir=>undef, #dir to save asups to
            client=>undef, #client to run the server on
            transport=>"HTTPS", #transport type to use
            @_,
    );
    my $transport_server;
    if (!defined $params{client}) {
        $Log->error("A client is required in order to invoke the $params{transport} server \n");
        return 0;
    }
    
    #Configure autosupport settings on the filer and invoke the tansport server
    $Log->comment("Invoke and Verify: Configuring autosupport on the filer \n");
    $Log->comment("Invoke and Verify: Creating $params{transport} Server object \n");
    if ($params{transport} eq "HTTP") {
        config_asup(enable=>"on", support_enable=>"on", support_transport=>$params{transport}, put_url=>$params{client}.":80", support_url=>$params{client}.":80", support_proxy=>'""');
        $transport_server = new HTTP_Server(client => $params{client}, asup_dir=>$params{asup_save_dir});
    }
    elsif ($params{transport} eq "HTTPS") {
        config_asup(enable=>"on", support_enable=>"on", support_transport=>$params{transport}, put_url=>$params{client}.":443", support_url=>$params{client}.":443", support_proxy=>'""', validate_digital_certificate=>"off");
        $transport_server = new HTTPS_Server(client => $params{client}, asup_dir=>$params{asup_save_dir});
    }
    else {
        config_asup(enable=>"on", support_enable=>"on", support_transport=>$params{transport}, mailhost=>$params{client});
        $transport_server = new SMTP_Server(client => $params{client}, asup_dir=>$params{asup_save_dir});
    }
    
    #Invoke the autosupport
    $Log->comment("Invoke and Verify: Invoking an autosupport \n");
    my $invoked_asup=invoke_autosupport(event=>$params{event});
    
    #Verify autosupport was received by the transport server
    my $asup_dirs=verify_autosupport_reception(events=>$invoked_asup, server=>$transport_server);
    
    #Stop the transport server
    $Log->comment("Invoke and Verify: Stopping the $params{transport} server \n");
    $transport_server->stop();

    #if verify asup reception returned 0, then we failed to verify the asup was received
    if ($asup_dirs==0) {
        $Log->comment("Invoke and Verify: Failed to verify that the autosupport was received, $invoked_asup \n");
        return 0;
    }
    #if verify asup reception returned 1, then we validated that the asup was received
    #and no autosupports were saved to disk
    elsif ($asup_dirs==1) {
        return 1;
    }
    #else, the autosupports were saved to disk, so lets validate them
    elsif (defined $params{asup_save_dir}) {
        if (!verify_autosupport_format(asup_dir=>$asup_dirs,validate=>"all")) {
            $Log->comment("Invoke and Verify: Failed to verify autosupport format for invoked autosupport, $invoked_asup \n");
            return 0;
        }
        #the asup checks out okay, return a 1 (success)
        return 1;
    }
    #either an error occurred or an unexpected value was returned, so fail
    else {
        return 0;
    }
}

###############################################################################
#Sub: verify_autosupport_reception
#Input Params: Hash
#Return values: 0 (failure) or 1 (success - no save dirs) or array of save dirs
#Summary: Verify the reception of an autosupport via the specified protocol.
#       Returns a 0 on failure, a 1 on success, or the save dirs if successfully
#       verified and asup_dirs was specifed for the server.
###############################################################################
sub verify_autosupport_reception {
    my $me = (caller(0))[3];
    my %params = (
            events     => undef, #an event or array of events
            server     => undef, #transport server to verify events from
            time_limit => 300,   #amount of time to look for the events (in secs)
            @_,
    );
    #$Log->comment("Events:  ".Dumper($params{events}));
    my $server_type = ref($params{server});
    $Log->comment("$me: Verifying events on server_type: $server_type, event_type: ".ref($params{events})."\n\n");
    
    my $server_messages;
    my $asup_directories=undef;
    my $flag=0; #indicates whether we have found all events or not
    #these are the events we need to look for
    my $generated_events=$params{events};
    my $subject;
    
    my $max_time; #time to exit this routine
    
    #if a single event, HASH
    if (ref($params{events}) ne "ARRAY") {
        #check subject we should look for in the autosupport, if subject is defined for an event then look for that
        if (defined $params{events}->{subject}) {
            $subject=$params{events}->{subject};
        }
        #otherwise, look for the event name
        else {
            $subject=$params{events}->{event};
        }
        #look for the event over a period of 180secs
        $max_time=time()+$params{time_limit};
        my $t1 = time();
OUTER1: while ((time() < $max_time) and ($flag==0)) {

            #check to ensure server is still up
            if (!$params{server}->{subtest}->is_alive()) {
                $Log->error("$me: Server failed to start or has terminated...\n");
                return 0;
            }

            Tharn::sleep(5);

            #get the messages from the transport server subprocess
            $server_messages=$params{server}->get_all();
            if (@{$server_messages} > 0) {
                my $elapsed = time() - $t1;
                $Log->comment("$me: event_type=".ref($params{events})." elapsed($elapsed) MESSAGES SENT TO SERVER: ".Dumper($server_messages)."\n");
            }
            else {
                $Log->debug(sub { $max_time - time()." seconds left until we stop looking for asups!\n"});
            }
            #check each message posted from the transport server against our expected events
            foreach my $message (@$server_messages) {
                #if found, add the save directory to the array and break out
                if ($message =~ /$subject/ and ($message =~ /$params{events}->{node}/ or ($params{events}->{sysid} and $message =~ /$params{events}->{sysid}/)) and $message =~ /Saved to directory\s(.*)/) {
                    my $saved_dir = $1;
                    $Log->comment("$me: Found our autosupport entry!  $message \n");
                    if ($saved_dir ne "") {
                        $Log->comment("$me: ASUP SAVE DIR was $saved_dir \n");
                        push (@$asup_directories,$saved_dir);
                    }

                    if ( $server_type =~ /SMTP/ ) {
                        my $email_file = "$saved_dir/email";
                        if ( $saved_dir && -e $email_file) {
                            my $validated = 0;
                            my $validation_retries = 30;
INNER1:                     while ( !$validated && ($validation_retries > 0) ) {
                                $flag=1;
                                $validation_retries--;
                                $Log->comment("$me: Verifying inclusion of final MIME boundary and BASE64 encoding...");

                                my ($mime_boundary, $base64_encoding) = ASUP::validate_email_file( file => $email_file );

                                if ( !$mime_boundary ) {
                                    $flag=0;
                                    $Log->warn("$me: Failed to find final MIME boundary!");
                                }
                                if ( !$base64_encoding ) {
                                    $flag=0;
                                    $Log->warn("$me: Failed to validate BASE64 encoding!");
                                }

                                if ( !$flag && ($validation_retries > 0) ) {
                                    $Log->warn("$me: Failed to verify MIME contents, retrying in 2 seconds");

                                    my @ls_out;
                                    eval {
                                        @ls_out = ASUP::shell_unix_cmd( cmd => "ls --full-time $saved_dir" );
                                    };
                                    if ( $@ ) {
                                        $Log->comment("$me: ls error = |$@|");
                                    } else {
                                        $Log->comment("$me: ls output: ".join("\n", @ls_out));
                                    }

                                    Tharn::sleep(2);
                                } else {
                                    $validated=1;
                                    last INNER1;
                                }
                            }
                        } elsif ($saved_dir) {
                            $flag=0;
                            $Log->warn("$me: SMTP_Server did not yet create $email_file");
                        } else {
                            $flag=1;
                            $Log->comment("$me: SMTP_Server was not configured to save asup into email file");
                        }
                    } else {
                        $flag = 1;
                    }

                    if ( $flag ) {
                        last OUTER1;
                    } 
                }
            }
            #not yet found, look again  
        }
    }
    #else if an array of events
    else {
        #set the timeout value to look for events number of events at 180s each
        if ($params{time_limit}==180) {
            $max_time=time()+($params{time_limit} * @$generated_events);
        }
        else {
            $max_time=time()+$params{time_limit};   
        }
        $Log->debug(sub {"$me: TIME is ".time()." and max timer is $max_time \n"});
        #while there is time, look for the asups
OUTER2: while ((time() < $max_time) and $flag==0) {
            #check to ensure server is still up
            if (!$params{server}->{subtest}->is_alive()) {
                $Log->error("$me: Server failed to start or has terminated...\n");
                return 0;
            }
 
            Tharn::sleep( 5 );

            #get the messages from the transport server subprocess
            $server_messages=$params{server}->get_all();
            if (@{$server_messages} > 0) {
                $Log->comment("$me: MESSAGES SENT TO SERVER: ".Dumper($server_messages)."\n"); 
            }
            else {
                $Log->debug(sub {$max_time - time()." seconds left until we stop looking for asups!\n"});
            }
            #check each message we got from the tranport server
            foreach my $message (@$server_messages) {
                my $still_looking;
                #$Log->comment("$me: Looking for event in the message $message \n");
                #look a matching event that we were expecting
                my $match=0;
                foreach my $single_event (@$generated_events) {
                    #check subject we should look for in the autosupport, if subject is defined for an event then look for that
                    if (defined $single_event->{subject}) {
                        $subject=$single_event->{subject};
                    }
                    #otherwise, look for the event name
                    else {
                        $subject=$single_event->{event};
                    }
                    $Log->debug(sub {"$me: Looking for $subject from node ".$single_event->{node}." or sysid ".$single_event->{sysid}." \n"});
                    #if found, add the save directory to the array and break out
                    if ($message =~ /$subject/ and ($message =~ /$single_event->{node}/ or ($single_event->{sysid} and $message =~ /$single_event->{sysid}/)) and $match==0 and $message =~ /Saved to directory\s(.*)/) {
                        my $saved_dir = $1;
                        $Log->comment("$me: Found an expected event!  $message \n");
                        if ($saved_dir ne "") {
                            $Log->comment("$me: ASUP SAVE DIR was $saved_dir \n");
                            push (@$asup_directories,$saved_dir);
                        }
                        $match=1;
                        my $warnings = 0;
                        if ( $server_type =~ /SMTP/ ) {
                            my $email_file = "$saved_dir/email";
                            if ( $saved_dir && -e $email_file) {
                                my $validated = 0;
                                my $validation_retries = 30;
INNER2:                         while ( !$validated && ($validation_retries > 0) ) {
                                    $warnings = 0;
                                    $validation_retries--;

                                    my ($mime_boundary, $base64_encoding) = ASUP::validate_email_file( file => $email_file );

                                    if ( !$mime_boundary ) {
                                        $warnings++;
                                        $Log->warn("$me: Failed to find final MIME boundary!");
                                    }
                                    if ( !$base64_encoding ) {
                                        $warnings++;
                                        $Log->warn("$me: Failed to validate BASE64 encoding!");
                                    }

                                    if ( $warnings && ($validation_retries > 0) ) {
                                        $Log->warn("$me: Failed to verify MIME contents, retrying in 2 seconds");
    
                                        my @ls_out;
                                        eval {
                                            @ls_out = ASUP::shell_unix_cmd( cmd => "ls --full-time $saved_dir" );
                                        };
                                        if ( $@ ) {
                                            $Log->comment("$me: ls error = |$@|");
                                        } else {
                                            $Log->comment("$me: ls output: ".join("\n", @ls_out));
                                        }
    
                                        Tharn::sleep(2);
                                    } else {
                                        $validated=1;
                                        last INNER2;
                                    }
                                }
                            } elsif ($saved_dir) {
                                $warnings++;
                                $Log->warn("$me: SMTP_Server did not yet create $email_file");
                            } else {
                                $Log->comment("$me: SMTP_Server was not configured to save asup into email file");
                            }
                        }
                        if ( $warnings ) {
                            push (@$still_looking, $single_event);
                        }
                    }
                    #else, add the message we are currently looking for back to the queue
                    else {
                        #$Log->comment("Event not found yet, Adding message to queue \n");
                        push (@$still_looking, $single_event);
                    }
                }
                #generated_events is now the array of events we are looking for, minus the ones we found
                $generated_events=$still_looking; #pointer to still_looking
                $still_looking=undef;
                #if there are no events left to look for, break out
                if (!defined $generated_events) {
                    $flag=1;
                    last OUTER2;
                }
                #else we still have events to look for, so go back and get more messages!
                $Log->comment("$me: Number of events left to look for: ".@$generated_events."\n");
                $Log->comment("$me: EVENTS left to find - ".Dumper($generated_events)."\n");
            }
        }
    }
    
    #if we have found all of our expected asup events
    if ($flag) {
        $Log->comment("$me: Found all of the events \n");
        #return 1 if no asups were saved to disk
        if (!defined $asup_directories) {
            return 1;
        }
        #else return an array of the directories where the asups are saved
        else {
            return $asup_directories; #returns undef if no directories, otherwise returns a ref to an array of those dirs
        }
    }
    
    #otherwise, we failed to find at least one event so return a failure 0
    
    $Log->comment("$me: Failed to find at least one event \n");
    $Log->comment("$me: Events not found - ".Dumper($generated_events)."\n");
    my $asup_history;

    if ( ref($params{events}) =~ /ARRAY/i ) {
        $asup_history = get_autosupport_history(subject=>"*", node=>$params{events}[0]->{node});
    } else {
        $asup_history = get_autosupport_history(subject=>"*", node=>$params{events}->{node});
    }

    $Log->debug(sub {"$me: Asup history: ".Dumper($asup_history)});
    return 0;
}

###############################################################################
#Sub: verify_autosupport_collection
#Input Params: Hash
#Return values: 0 (failure) or 1 (success)
#Summary: Verify the collection of an autosupport on the filer.
#       Returns a 0 on failure, a 1 on success
###############################################################################
sub verify_autosupport_collection {
    my $me = (caller(0))[3];
    my %params = (
            events => undef, #an event or array of events
            time_limit => 300, #amount of time to look for the events (in secs)
                #we default to 5 mins
            last_update_time=>'>10m',
            destination=>"http",
            @_,
    );
    
        #if a single event
    if (ref($params{events}) ne "ARRAY") {
        my $asup_history;
        my $found=0;
        my $start_time=time();
        my $end_time=$start_time+$params{time_limit};
        $Server->set_execution_server(-Server =>$params{events}->{node});
        my $current_time=time();
        while ($current_time < $end_time and $found == 0) {
            $Log->debug(sub {"Curr TIME = $current_time, END TIME = $end_time \n"});
            if ($params{events}->{event} =~ /^callhome/) {
                $asup_history=ASUP::get_autosupport_history(node=>$params{events}->{node},
                                          trigger=>$params{events}->{event},
                                          subject=>$params{events}->{subject},
                                          destination=>$params{destination},
                                          status=>"transmitting|sent-successful|transmission-failed|re-queued|queued",
                                          "last-update"=>$params{last_update_time});   
            }
            else {
                $asup_history=ASUP::get_autosupport_history(node=>$params{events}->{node},
                                          subject=>'"USER_TRIGGERED (TEST:'.$params{events}->{subject}.')"',
                                          destination=>$params{destination},
                                          status=>"transmitting|sent-successful|transmission-failed|re-queued|queued",
                                          "last-update"=>$params{last_update_time});
            }
            if (defined $asup_history and @{$asup_history} > 0) {
                $found=1;
                $Log->comment("Validated autosupport has been collected via history table results...\n");
            }
            $current_time=time();
            Tharn::sleep(5);
        }
        if (!$found) {
            $Log->error("Autosupport $params{events}->{subject} on node $params{events}->{node} was not reported as collected by the autosupport history show command...\n");
            return 0;
        }
    }
    
    else {
        my $asup_history;
        my $start_time=time();
        my $end_time=$start_time+$params{time_limit};
        foreach my $event (@{$params{events}}) {
            my $found=0;
            $Server->set_execution_server(-Server =>$event->{node});
            my $current_time=time();
            while ($current_time < $end_time and $found == 0) {
                $Log->debug(sub {"Curr TIME = $current_time, END TIME = $end_time \n"});
                if ($event->{event} =~ /^callhome/) {
                    $asup_history=ASUP::get_autosupport_history(node=>$event->{node},
                                          trigger=>$event->{event},
                                          subject=>$event->{subject},
                                          destination=>$params{destination},
                                          status=>"transmitting|sent-successful|transmission-failed|re-queued|queued",
                                          "last-update"=>$params{last_update_time});   
                }
                else {
                    $asup_history=ASUP::get_autosupport_history(node=>$event->{node},
                                          subject=>'"USER_TRIGGERED (TEST:'.$event->{subject}.')"',
                                          destination=>$params{destination},
                                          status=>"transmitting|sent-successful|transmission-failed|re-queued|queued",
                                          "last-update"=>$params{last_update_time});
                }
                if (defined $asup_history and @{$asup_history} > 0) {
                    $found=1;
                    $Log->comment("Validated autosupport has been collected via history table results...\n");
                }
                $current_time=time();
                Tharn::sleep(5);
            }
            if (!$found) {
                $Log->error("Autosupport $event->{subject} on node $event->{node} was not reported as collected by the autosupport history show command...\n");
                return 0;
            }
        }
    }
    
    return 1;
} #end sub verify_autosupport_collection

###############################################################################
#Sub: invoke_autosupport
#Input Params: Hash
#Return values: Hash containing event name/message, the node, and the sysid.
#Summary: Invokes the specified autosupport on a filer.  If no autosupport is
#   specified, it will manually invoke an autosupport.
###############################################################################
sub invoke_autosupport {
    my %params = (
            node => "local", 
            message => 'Manually_Invoked_Autosupport_'.time(),
            subject => undef, #only valid if event is passed in
            event=> undef,
            type=>"test",
            force=> undef,
            uri=> undef,
            @_,
    );
    my $version=_get_version();
    my $event_hash;
    
    while ( my ($key, $value) = each(%params) ) {
        if (!defined($value)) {
            delete $params{$key}; #remove any params that don't have values configured
        }
    }
    
    #if no asup event is specified, manually invoke an autosupport using the
    #default message or the message supplied
    if (!defined $params{event}) {
        #7M
        $Log->comment("Invoke ASUP: Using invoke/doit to invoke asup message $params{message} \n");
        if ($version =~ /7M/ ) {
            eval {
                my $result;
                if ( $version =~ /7M/ && $params{'type'} =~ /performance/ ) {
                    $Log->comment("Invoking the 7M performance asup, autosupport.performance_data.doit");
                    $result=$Server->cli(cmd=>"options autosupport.performance_data.doit $params{message}");
                    #$result=$Server->performance_doit(message=>"'".$params{message}."'");
                    $params{message}="PERFORMANCE SNAPSHOT";
                }
                else{
                    $result=$Server->autosupport_doit(message=>"'".$params{message}."'");
                }
                if ($result) {
                    $Log->comment("DoIt returned an error: $result \n");
                    return 0;
                }
                my $sysconfig=$Server->cli(cmd=>"sysconfig");
                if ($sysconfig =~ /System ID: (\d+)/) {
                    $event_hash->{sysid}=$1;
                }
                else {
                    $event_hash->{sysid}="Sysid not available for node";
                }
            };
        }
        #GX/CM
        else {
            #the performance event doesn't take the message param in rr
            if ( $params{'type'} =~ /performance/ ) {
                eval {
                    $Server->system_node_autosupport_invoke(node=>$params{'node'},type=>$params{'type'});
                };
            }
            else{
                eval {
                    $Server->system_node_autosupport_invoke(%params);
                };
            }
            if ($@) {
                $Log->comment("Invoke returned an error: $@ \n");
                return 0;
            }
        }
        $event_hash->{event}=$params{message};
        $event_hash->{subject}=$params{message};
        $event_hash->{node}=$Server->get_execution_server();

        eval {
            my $res = $Server->system_node_show(node=>$params{'node'});
            $event_hash->{sysid}=$res->[0]->{systemid};
        };

        return $event_hash;
    }
    #else, invoke the asup event specified.  If a subject is specified, use
    # that as the autosupport subject.  Otherwise, use the event name as
    # the autosupport subject.
    else {
        if (defined $params{subject}) {
            $event_hash->{subject}=$params{subject};
        }
        else {
            $event_hash->{subject}=$params{event};
        }
        $Log->comment("Invoke ASUP: Invoking asup via callhome event - $params{event} with subject $event_hash->{subject}\n");
        EMS::invoke_event_without_error_checking(event=>$params{event}, arguments=>'"'.$event_hash->{subject}.'"');
        $event_hash->{event}=$params{event};
        $event_hash->{node}=$Server->get_execution_server();
        return $event_hash;
    }
    #return failure
    return 0;
}

############################################################################
# TestCase: invoke_nht_autosupport
# Parameters: None
# Returns: $invoked_asup
# Purpose: Create files necessary for nht and generate an nht asup
#############################################################################

sub invoke_nht_autosupport {
    my $version=_get_version();
    if ($version=~/7M/) {
        eval {
            $Server->cli(cmd=>"wrfile -a /etc/log/nht_info/sm2.gz");
        };
        eval {
            $Server->cli(cmd=>"wrfile -a /etc/log/nht_info/sm2.gz.ready");
        };
    }
    else {
        eval {
            $Server->shell(cmd=>'sudo echo "nht info gz file" > /mroot/etc/log/nht_info/sm2.gz');
            $Server->shell(cmd=>'sudo touch /mroot/etc/log/nht_info/sm2.gz.ready');
        };
    }
    my $invoked_asup=invoke_autosupport(event=>"callhome.nht.data");
    return $invoked_asup;
} # end sub  invoke_nht_autosupport

############################################################################
# TestCase: invoke_performance_autosupport
# Parameters: None
# Returns: $invoked_asup
# Purpose: Create files necessary for performance and generate a perf asup
#############################################################################

sub invoke_performance_autosupport {
    my $version=_get_version();
    my $invoked_asup;
    if ($version=~/7M/) {
        eval {
            $Server->cli(cmd=>"wrfile -a /etc/log/.cm_stats_hourly_done");    
        };
        $invoked_asup=invoke_autosupport(message=>"callhome.performance.snap",type=>"performance");
    }
    else {
        $invoked_asup=invoke_autosupport(node=>"local",type=>"performance");
        $invoked_asup->{subject}="PERFORMANCE SNAPSHOT";
        $invoked_asup->{node}=$Server->get_execution_server();
    }
    return $invoked_asup;
} # end sub  invoke_performance_autosupport

############################################################################
# TestCase: invoke_weekly_autosupport
# Parameters: None
# Returns: $invoked_asup
# Purpose: Invoke a weekly asup
#############################################################################

sub invoke_weekly_autosupport {
    my $me = (caller(0))[3];

    my %params = (
        event    => 'callhome.weekly',
        subject  => undef,
        @_,
    );

    if ( !$params{subject} ) {
        delete $params{subject};
    }
    my $invoked_asup=invoke_autosupport(%params);
    return $invoked_asup;
} # end sub  invoke_weekly_autosupport

###############################################################################
#Sub: start_transport_server
#Input Params: Hash
#Return values: HTTP,SMTP, or HTTPS server object
#Summary: Creates and initiates a HTTP, SMTP, or HTTPS server.  Then returns a
#   reference to the server object invoked.
###############################################################################
sub start_transport_server {
    my %params = (
            client => undef, #client to run the server on
            asup_save_dir => undef, #directory to save autosupport data to
            transport => "HTTPS", #type of server to create/start
            port=>undef,
            put_supported=>-1,
            post_supported=>-1,
            method_not_supported=>-1,
            ipv6=>undef,
            read_bytes=>"all",
            @_,
    );
    my $transport_server;
    
    if (!defined $params{client}) {
        $Log->error("Start Server: A client is required in order to invoke the $params{transport} server \n");
        return 0;
    }
    #create and start the specified server
    if ($params{transport} eq "HTTPS") {
        $transport_server = new HTTPS_Server(client => $params{client}, 
                                             asup_dir=>$params{asup_save_dir}, 
                                             port=>$params{port},
                                             put_supported=>$params{put_supported},
                                             post_supported=>$params{post_supported},
                                             method_not_supported=>$params{method_not_supported},
                                             ipv6=>$params{ipv6},
                                             read_bytes=>$params{read_bytes});
    }
    elsif ($params{transport} eq "HTTP") {
        $transport_server = new HTTP_Server(client => $params{client},
                                            asup_dir=>$params{asup_save_dir}, 
                                            port=>$params{port},
                                            put_supported=>$params{put_supported},
                                            post_supported=>$params{post_supported},
                                            method_not_supported=>$params{method_not_supported},
                                            ipv6=>$params{ipv6},
                                            read_bytes=>$params{read_bytes});
    }
    elsif ($params{transport} eq "SMTP") {
        $transport_server = new SMTP_Server(client => $params{client}, 
                                            ipv6=>$params{ipv6},
                                            asup_dir=>$params{asup_save_dir});
    }
    else {
        $Log->error("Start Server: Only HTTP, HTTPS, and SMTP are acceptable transport parameters \n");
        return 0;
    }
    #return server object
    return $transport_server;
}

###############################################################################
#Sub: stop_transport_server
#Input Params: Hash
#Return values: 1 (success) or 0 (failure)
#Summary: Terminates the specified transport server.
###############################################################################
sub stop_transport_server {
    my %params = (
            server=>undef, #server object
            @_,
    );
    
    if (!defined $params{server}) {
        $Log->error("Stop Server: A server is required... \n");
        return 0;
    }

    $Log->comment("stop_transport_server: Stopping the $params{server} server \n");
    $params{server}->stop;
    snooze(15);
    
    return 1;
}

###############################################################################
#Sub: stop_transport_servers
#Input Params: servers
#Return values: 1 (success) or 0 (failure)
#Summary: Terminates the specified transport server.
###############################################################################
sub stop_transport_servers {
    my @servers = @_;
    my $rc = 1;

    foreach my $server (@servers) {
        $rc &= ASUP::stop_transport_server( server=> $server );
    }
    
    return $rc;
}


###############################################################################
#Sub: start_ondemand_server
#Input Params: Hash
#Return values: HTTPS server object
#Summary: Creates and initiates a HTTPS Ondemand server. Then returns a
#         reference to the server object invoked.
###############################################################################
sub start_ondemand_server {
    my %params = (
            client=>undef,
            port=>8443,
            delivery_required=>1,
            deliver_uri=>undef,
            retransmit_uri=>undef,
            retransmit_seq_no=>0,
            retransmit_size=>0,
            multiple_retransmits=> 1,
            polling_rate=>-1,
            method_not_supported=>405,
            ondemand_dir=>undef,
            malformed_response=>0,
            malformed_resp_type=>0,
            read_bytes=>"all",
            ipv6=>undef,
            invoke_type=>"none",
            invoke_message=>"AoD Test",
            invoke_uri=>undef,
            multiple_invokes=> 0,
            @_,
    );
    my $ondemand_server;
    
    if (!defined $params{client}) {
        $Log->error("Start Server: A client is required in order to invoke the OnDemand server \n");
        return 0;
    }
   
    $ondemand_server = new HTTPS_OnDemand_Server(client => $params{client}, 
                                             port=>$params{port},
                                             delivery_required=>$params{delivery_required},
                                             deliver_uri=>$params{deliver_uri},
                                             retransmit_uri=>$params{retransmit_uri},
                                             retransmit_seq_no=>$params{retransmit_seq_no},
                                             retransmit_size=>$params{retransmit_size},
                                             multiple_retransmits=>$params{multiple_retransmits},
                                             polling_rate=>$params{polling_rate},
                                             method_not_supported=>$params{method_not_supported},
                                             ondemand_dir=>$params{ondemand_dir},
                                             malformed_response=>$params{malformed_response},
                                             malformed_resp_type=>$params{malformed_resp_type},
                                             ipv6=>$params{ipv6},
                                             read_bytes=>$params{read_bytes},
                                             invoke_type=>$params{invoke_type},
                                             invoke_message=>$params{invoke_message},
                                             invoke_uri=>$params{invoke_uri},
                                             multiple_invokes=>$params{multiple_invokes});

    # Server takes some time to start
    Tharn::sleep(30);

    #return server object
    return $ondemand_server;
}


###############################################################################
#Sub: stop_ondemand_server
#Input Params: Hash
#Return values: 1 (success) or 0 (failure)
#Summary: Terminates the specified ondemand server.
###############################################################################
sub stop_ondemand_server {
    my $result = 0;
    my %params = (
            server=>undef, #server object
            @_,
    );
    
    if (!defined $params{server}) {
        $Log->error("Stop Server: A server is required... \n");
        return 0;
    }

    $Log->comment("stop_transport_server: Stopping the ASUP OnDemand server \n");
    $result = $params{server}->stop;
    snooze(15);
    
    return $result;
}


###############################################################################
#Sub: verify_autosupport_format
#Input Params: Hash
#Return values: 1 (success) or 0 (failure)
#Summary: Validates format of the saved autosupports.  Validation can occur
#       on asup headers, sections, attachments, or all.  Default is all.
###############################################################################
sub verify_autosupport_format {
    my %params = (
            asup_dir=>undef, #dirs where asups were saved
            alt_extension=>0,
            minimal_content=>0,
            validate=>"all", #headers, sections, attachments, manifest, or all
            xfer_protocol=>'PUT|POST', # by default either is ok
            check_status=>1, #by default we should check the status of files in the manifest
            @_,
    );
    
    if (!$params{asup_dir}) {
        $Log->error("Verify ASUP Format: A saved autosupport directory is required in order to validate the format of an asup \n");
        return 0;
    }
    my $content="complete";
    $Log->comment("minimal param(int): $params{minimal_content}");
    if ($params{minimal_content} != 0 ){
        $content="minimal";
    }
    my $version=_get_version();
    
    #verify headers for all versions
    #verify headers format for each asup dir
    if ($params{validate} eq "all" or $params{validate} eq "headers") {
        if (!_verify_headers(asup_dir=>$params{asup_dir},xfer_protocol=>$params{xfer_protocol})) {
            $Log->error("Verify ASUP Format: Failed to verify the headers for the autosupport \n");
            return 0;
        }
    }
    
    if ($params{validate} eq "all" or $params{validate} eq "manifest") {
        if (extract_archive(asup_dir=>$params{asup_dir})) {
            if (verify_manifest(asup_dir=>$params{asup_dir}, check_status=>$params{check_status}, content=>$content) !=1 ) {
                $Log->error("Verify ASUP Format: Failed to verify all attachments from manifest file.  See log for details. \n");
                return 0;
             }
        }
    }
    return 1;       
}

###############################################################################
#Sub: retrieve_archived_asup_dirs
#Input Params: <none>
#Return values: Ref to array of the saved asups on the filer.
#Summary: Retrieves the named files of the current asups that are saved on the
#           filer.
###############################################################################
sub retrieve_archived_asup_dirs {
    my $asupdirs;
    my $version=_get_version();
    
    #get the asup dirs that are saved on the filer
    #7M
    if ($version =~ /7M/) {
        eval {
            $asupdirs = $Server->cli(cmd=>'priv set diag;ls /etc/log/autosupport');
        };
    }
    #GX/CM
    else {
        eval {
            $asupdirs = $Server->shell(cmd=>"ls /mroot/etc/log/autosupport");
        };
    }
    
    my @asuparchive=split(/\t+|\n+/,$asupdirs);
    
    #remove the output from the priv set diag command
    if ($version =~ /7M/) {
        shift(@asuparchive); #first line is output from priv set diag
        shift(@asuparchive); #second line is output from priv set diag
    }
    
    #now sort the array and reverse it so that most recent are first indexed
    @asuparchive=sort(@asuparchive);
    @asuparchive=reverse(@asuparchive);

    my @real_asup_dirs;
    foreach my $asupdir (@asuparchive) {
        if ($asupdir =~ /files/) {
            push (@real_asup_dirs,$asupdir);
        }
    }

    return \@real_asup_dirs;

}

###############################################################################
#Sub: verify_asup_archived
#Input Params: hash
#Return values: asupdir on success, 0 on failure
#Summary: Checks to see if the specified event or message asup has been
#           successfully archived on the filer.
###############################################################################
sub verify_asup_archived {
    my %params = (
            event=>undef,
            @_,
    );
    my $version=_get_version();
    my $archived_asups=retrieve_archived_asup_dirs();
    foreach my $saved_asup (@$archived_asups) {
        my $asupdata=undef;
        $saved_asup=~ s/\s+$//; #remove any trailing whitespace characters
        #7M
        if ($version =~ /7M/) {
            if ($saved_asup =~/files/) {
                $asupdata=$Server->cli(cmd=>"rdfile /etc/log/autosupport/$saved_asup/header.txt");
            }
            elsif ($saved_asup !~ /files/) {
                $asupdata=$Server->cli(cmd=>"rdfile /etc/log/autosupport/$saved_asup");    
            }
        }
        #GX/CM
        else {
            if ($saved_asup =~/files/) {
                $asupdata=$Server->shell(cmd=>'more /mroot/etc/log/autosupport/'.$saved_asup.'/header.txt | grep subject=');
            }
            elsif ($saved_asup !~ /files/) {
                $asupdata=$Server->shell(cmd=>'more /mroot/etc/log/autosupport/'.$saved_asup.' | grep subject=');
                #$asupdata=$Server->shell(cmd=>'ls');
            }
        }
        if ($asupdata =~ $params{event}->{event}) {
            $Log->comment("Verify ASUP Archive: Found our asup in /etc/log/autosupport/$saved_asup \n");
            return $saved_asup;
        }
    }
    $Log->comment("Verify ASUP Archive: Failed to find autosupport $params{event}->{event} in asup archive \n");
    return 0;   

}

###############################################################################
#Sub: modify_autosupport_filter
#Input Params: hash
#Return values: 1 on success, 0 on failure
#Summary: Provides ability to modify whether autosupports are sent locally to
#   addresses in the to field or noteto field.
###############################################################################
sub modify_autosupport_filter {
    my %params = (
            node=>"local",
            'autosupport-message'=>undef,
            to=>"enabled",
            noteto=>"enabled",
            @_,
    );
    my $version=_get_version();
    my $result;
    if ($version=~/7M/) {
        $Server->autosupport_filter_modify('autosupport-message'=>$params{'autosupport-message'}, to=>$params{to}, noteto=>$params{noteto});
        $result=get_autosupport_filter('autosupport-message'=>$params{'autosupport-message'});
        foreach my $res (@$result) {
            if ($res->{to} ne $params{to} or $res->{noteto} ne $params{noteto}) {
                $Log->comment("Autosupport filter modify failed. ".$res->{'autosupport-message'}." does not have settings to:$params{to} and noteto:$params{noteto}\n");
                return 0;
            }
        }
    }
    else {#if 10M
        $Server->system_node_autosupport_filter_modify(node=>$params{node}, 'autosupport-message'=>$params{'autosupport-message'}, to=>$params{to}, noteto=>$params{noteto});
        if ($@) {
           $Log->comment("Autosupport filter modify failed with $@\n");
            return 0; 
        }
    }
    return 1;
}

###############################################################################
#Sub: get_autosupport_filter
#Input Params: hash
#Return values: pointer to array of hash
#Summary: Provides ability to return the current autosupport filter settings
###############################################################################
sub get_autosupport_filter {
    my %params = (
            node=>"local",
            'autosupport-message'=>'*',
            @_,
    );
    my $version=_get_version();
    my $result;
    if ($version=~/7M/) {
        my @show_details=split("\n",($Server->autosupport_filter_show('autosupport-message'=>$params{'autosupport-message'})));
        my $i=0;
        foreach my $show_line (@show_details) {
            if ($show_line =~/(\S*)\s*(enabled|disabled)\s*(enabled|disabled)/) {
                $result->[$i]->{'autosupport-message'}=$1;
                $result->[$i]->{to}=$2;
                $result->[$i]->{noteto}=$3;
            }
            $i++;
        }
    }
    else {#if 10M
        $result=$Server->system_node_autosupport_filter_show(node=>$params{node},'autosupport-message'=>$params{'autosupport-message'});
        
    }
    return $result;
} 

###############################################################################
#Sub: get_autosupport_manifest
#Input Params: hash
#Return values: pointer to array of hash
#Summary: Return the manifest entries for autosupport
###############################################################################
sub get_autosupport_manifest {
    my %params = (
            node=>"local",
            'seq-num'=>undef,
            'prio-num'=>undef,
            subsystem=>undef,
            'cmd-tgt'=>undef,
            'body-file'=>undef,
            cmd=>undef,
            query=>undef,
            'size-collected'=>undef,
            status=>undef,
            error=>undef,
            'content-type'=>undef,
            @_,
    );
    my $version=_get_version();
    my $result;
    while ( my ($key, $value) = each(%params) ) {
        if (!defined($value)) {
            delete $params{$key}; #remove any params that don't have values configured
        }
    }
    if ($version=~/7M/) {
        delete $params{node}; #node param not support in 7M, so remove it
        my @show_details=split("\n",($Server->autosupport_manifest_show(%params)));
        shift(@show_details); #ignore line "Warning: These diagnostic commands are for use by NetApp\r\n"
        shift(@show_details); #ignore line "personnel only.\r\n"
        my $i=0;
        foreach my $show_line (@show_details) {
            if ($show_line =~/\s*AutoSupport Sequence Number:\s*(.*)/) {
                $result->[$i]->{'seq-num'}=$1;
            }
            elsif ($show_line =~/\s*Collection Order:\s*(.*)/) {
                $result->[$i]->{'prio-num'}=$1;
            }
            elsif ($show_line =~/\s*Subsystem:\s*(.*)/) {
                $result->[$i]->{'subsys'}=$1;
            }
            elsif ($show_line =~/\s*Execution domain for command:\s*(.*)/) {
                $result->[$i]->{'cmd-target'}=$1;
            }
            elsif ($show_line =~/\s*The AutoSupport Content filename for this data:\s*(.*)/) {
                $result->[$i]->{'body-file'}=$1;
            }
            elsif ($show_line =~/\s*The actual data being collected: D-blade CLI OR BSD command OR XML table OR filename:\s*(.*)/) {
                $result->[$i]->{'cmd'}=$1;
            }
            elsif ($show_line =~/\s*Query for XML collection:\s*(.*)/) {
                $result->[$i]->{'query'}=$1;
            }
            elsif ($show_line =~/\s*Number of bytes collected:\s*(.*)/) {
                $result->[$i]->{'size-collected'}=$1;
            }
            elsif ($show_line =~/\s*Collection time for this data item \(ms\):\s*(.*)/) {
                $result->[$i]->{'time-collected-ms'}=$1;
            }            
            elsif ($show_line =~/\s*The status of this data item:\s*(.*)/) {
                $result->[$i]->{'status'}=$1;
            }
            elsif ($show_line =~/\s*The textual description of error:\s*(.*)/) {
                $result->[$i]->{'error'}=$1;
            }
            elsif ($show_line =~/\s*The AutoSupport content type for this data:\s*(.*)/) {
                $result->[$i]->{'content_type'}=$1;
                $i++;
            }
        }
    }
    else {#if 10M
        $result=$Server->system_node_autosupport_manifest_show( '-Ignore_errors' => 1, %params);
        
    }
    return $result;
}

###############################################################################
#Sub: get_autosupport_history
#Input Params: hash
#Return values: pointer to array of hash
#Summary: Return the history entries for autosupport
###############################################################################
sub get_autosupport_history {
    my %params = (
            node=>"local",
            'seq-num'=>undef,
            destination=>undef,
            trigger=>undef,
            'last-update'=>undef,
            status=>undef,
            'attempt-count'=>undef,
            subject=>undef,
            uri=>undef,
            error=>undef,
            'generated-on' => undef,
            @_,
    );
    my $version=_get_version();
    my $result=undef;
    while ( my ($key, $value) = each(%params) ) {
        if (!defined($value)) {
            delete $params{$key}; #remove any params that don't have values configured
        }
    }
    if ($version=~/7M/) {
        delete $params{node}; #node parameter not valid for 7M, so remove it here
        my @show_details=split("\n",($Server->autosupport_history_show(%params)));
        shift(@show_details); #ignore line "Warning: These diagnostic commands are for use by NetApp\r\n"
        shift(@show_details); #ignore line "personnel only.\r\n"
        my $i=0;
        foreach my $show_line (@show_details) {
            if ($show_line =~/\s*AutoSupport Sequence Number:\s*(.*)/) {
                $result->[$i]->{'seq-num'}=$1;
            }
            elsif ($show_line =~/\s*Destination for this AutoSupport:\s*(.*)/) {
                $result->[$i]->{'destination'}=$1;
            }
            elsif ($show_line =~/\s*Trigger Event:\s*(\S*)/) {
                $result->[$i]->{'messagename'}=$1;
            }
            elsif ($show_line =~/\s*Time of Last Update:\s*(.*)/) {
                $result->[$i]->{'last-modified'}=$1;
            }
            elsif ($show_line =~/\s*Status of Delivery:\s*(.*)/) {
                $result->[$i]->{'status'}=$1;
            }
            elsif ($show_line =~/\s*Delivery Attempts:\s*(.*)/) {
                $result->[$i]->{'attempt-count'}=$1;
            }
            elsif ($show_line =~/\s*AutoSupport Subject:\s*(.*)/) {
                $result->[$i]->{'subject'}=$1;
            }
            elsif ($show_line =~/\s*Delivery URI:\s*(.*)/) {
                $result->[$i]->{'uri'}=$1;
            }
            elsif ($show_line =~/\s*Last Error:\s*(.*)/) {
                $result->[$i]->{'error'}=$1;
                $i++;
            }
            elsif ($show_line =~/\s*Time of Generation:\s*(.*)/) {
                $i--; # JDS Needed for backward compat until all of SN has generated-on data
                $result->[$i]->{'generated-on'}=$1;
                $i++;
            }
        }
    }
    else {#if 10M
        eval {
            $result=$Server->system_node_autosupport_history_show(%params);
        };
        
    }
    return $result;
}

###############################################################################
#Sub: retransmit_autosupport
#Input Params: hash
#Return values: pointer to array of hash
#Summary: Attempt to retransmit an archived autosupport
###############################################################################
sub retransmit_autosupport {
    my %params = (
            node=>"local",
            'seq-num'=>undef,
            uri=>undef,
            size=>undef,
            @_,
    );
    my $version=_get_version();
    my $result;
    while ( my ($key, $value) = each(%params) ) {
        if (!defined($value)) {
            delete $params{$key}; #remove any params that don't have values configured
        }
    }
    if ($version=~/7M/) {
        delete $params{node}; #node parameter not valid for 7M, so remove it here
        eval {
            $result=$Server->autosupport_history_retransmit(%params);
        };
        if (lc($result) =~/error/) {
            $Log->comment("retransmit_autosupport: Retransmit of seq-num: $params{'seq-num'} failed due to $result \n");
            return $result;
        }
    }
    else {#if 10M
        eval {
            $Server->system_node_autosupport_history_retransmit(%params);  
        };
        my $eval_rc = $@;
        if ($eval_rc) {
            $Log->comment("retransmit_autosupport: Retransmit of seq-num: $params{'seq-num'} failed due to $@ \n");
            return $eval_rc;
        }
    }
    return 1;
}

###############################################################################
#Sub: modify_autosupport_trigger
#Input Params: hash
#Return values: 1 on success, 0 on failure
#Summary: Provides ability to modify autosupport trigger configuration details
###############################################################################
sub modify_autosupport_trigger {
    my %params = (
            node=>"local",
            'autosupport-message'=>undef,
            to=>undef,
            noteto=>undef,
            'basic-additional'=>undef,
            'troubleshooting-additional'=>undef,
            'time-limit'=>undef,
            @_,
    );
    my $version=_get_version();
    my $result;
    while ( my ($key, $value) = each(%params) ) {
        if (!defined($value)) {
            delete $params{$key}; #remove any params that don't have values configured
        }
    }
    if ($version=~/7M/) {
        delete $params{node}; #node param not supported in 7M, so remove it
        $Server->autosupport_trigger_modify(%params);
        $result=get_autosupport_trigger(%params);
        #should we add error checking to ensure the param was set???
        #foreach my $res (@$result) {
        #    while (my ($key,$value)=each(%params)) {
        #        if ($result->{$key} ne $value) {
        #            $Log->comment("Autosupport trigger modify failed.  ".$res->{'autosupport-message'}." has $key set to $result->{$key} when we expected $value\n");
        #            return 0;
        #        }
        #    }
        #}
    }
    else {#if 10M
        $Server->system_node_autosupport_trigger_modify(%params);
        if ($@) {
           $Log->comment("Autosupport trigger modify failed with $@\n");
            return 0; 
        }
    }
    return 1;
}

###############################################################################
#Sub: restore_autosupport_trigger
#Input Params: hash
#Return values: 1 on success, 0 on failure
#Summary: Provides ability to restore previously retrieved autosupport trigger settings
###############################################################################
sub restore_autosupport_trigger {
    my %params = (
            values=>undef,
            @_,
    );

    if ( !$params{values} ) {
        $Log->error( die "Must provide values array of hashes to ASUP::restore_autosupport_trigger()" );
    }

    foreach my $item ( @{$params{values}} ) {
        ASUP::modify_autosupport_trigger( %{$item} );
    }
}

###############################################################################
#Sub: get_autosupport_trigger
#Input Params: hash
#Return values: pointer to array of hash
#Summary: Return the trigger configuration entries for autosupport
###############################################################################
sub get_autosupport_trigger {
    my %params = (
            node=>"local",
            'autosupport-message'=>undef,
            to=>undef,
            noteto=>undef,
            'basic-default'=>undef,
            'troubleshooting-default'=>undef,
            'additional-content'=>undef,
            'basic-additional'=>undef,
            'troubleshooting-additional'=>undef,
            'time-limit'=>undef,
            @_,
    );
    my $version=_get_version();
    my $result;
    while ( my ($key, $value) = each(%params) ) {
        if (!defined($value)) {
            delete $params{$key}; #remove any params that don't have values configured
        }
    }
    if ($version=~/7M/) {
        delete $params{node}; #node param not support in 7M, so remove it
        my @show_details=split("\n",($Server->autosupport_trigger_show(%params)));
        shift(@show_details); #ignore line "Warning: These diagnostic commands are for use by NetApp\r\n"
        shift(@show_details); #ignore line "personnel only.\r\n"
        my $i=-1;
        my $current_section=undef;

        foreach my $show_line (@show_details) {
            if ($show_line =~/\s*EMS Message:\s*(\S*)/) {
                $i++;
                $result->[$i]->{'autosupport-message'}=$1;
            }
            elsif ($show_line =~/\s*Whether to deliver to options autosupport\.to addresses:\s*(\S*)/) {
                $result->[$i]->{'to'}=$1;
            }
            elsif ($show_line =~/\s*Whether to deliver to options autosupport\.noteto addresses:\s*(\S*)/) {
                $result->[$i]->{'noteto'}=$1;
            }
            elsif ($show_line =~/\s*Default subsystems reporting basic info:\s*(.*)/) {
                $current_section='basic-default';
                my $subsystems=$1;
                $subsystems=~s/\s//g; #remove leading spaces
                $result->[$i]->{'basic-default'}=$subsystems;
            }
            elsif ($show_line =~/\s*Default subsystems reporting troubleshooting info:\s*(.*)/) {
                $current_section='troubleshooting-default';
                my $subsystems=$1;
                $subsystems=~s/\s//g; #remove leading spaces
                $result->[$i]->{'troubleshooting-default'}=$subsystems;
            }
            elsif ($show_line =~/\s*Additional content flag:\s*(\S*)/) {
                $result->[$i]->{'additional-content'}=$1;
            }
            elsif ($show_line =~/\s*Additional subsystems reporting basic info:\s*(\S*)/) {
                $current_section='basic-additional';
                my $subsystems=$1;
                $subsystems=~s/\s//g; #remove leading spaces
                $result->[$i]->{'basic-additional'}=$subsystems;
            }
            elsif ($show_line =~/\s*Additional subsystems reporting troubleshooting info:\s*(\S*)/) {
                $current_section='troubleshooting-additional';
                my $subsystems=$1;
                $subsystems=~s/\s//g; #remove leading spaces
                $result->[$i]->{'troubleshooting-additional'}=$subsystems;
            }
            elsif ($show_line =~/\s*The data collection time budget for this trigger.*:\s*(\S*)/) {
                $current_section='time-limit';
                my $time_limit=$1;
                $time_limit=~s/\s//g; #remove leading spaces
                if ( $time_limit =~ /-/ ) {
                    $time_limit = 0;
                }
                $result->[$i]->{$current_section}=$time_limit;
            }
            else {
                if ( $show_line && ($i >= 0) ) {
                    $show_line=~s/\s//g; #remove leading spaces
                    $result->[$i]->{$current_section}=$result->[$i]->{$current_section}.$show_line;
                }
            }
        }
    }
    else {#if 10M
        $result=$Server->system_node_autosupport_trigger_show(%params);
        foreach my $trigger (@$result) {
            if (ref($trigger->{'basic-default'}) eq "ARRAY") {
                my $subsystems=join(",",(@{$trigger->{'basic-default'}}));
                $trigger->{'basic-default'}=$subsystems;
            }
            
            if (ref($trigger->{'troubleshooting-default'}) eq "ARRAY") {
                my $subsystems=join(",",(@{$trigger->{'troubleshooting-default'}}));
                $trigger->{'troubleshooting-default'}=$subsystems;
            }
            
            if (ref($trigger->{'basic-additional'}) eq "ARRAY") {
                my $subsystems=join(",",(@{$trigger->{'basic-additional'}}));
                $trigger->{'basic-additional'}=$subsystems;
            }
            
            if (ref($trigger->{'troubleshooting-additional'}) eq "ARRAY") {
                my $subsystems=join(",",(@{$trigger->{'troubleshooting-additional'}}));
                $trigger->{'troubleshooting-additional'}=$subsystems;
            }

            if ($trigger->{'time-limit'} =~ /-/) {
                $trigger->{'time-limit'}=0;
            }
        }
    }
    return $result;
}

###############################################################################
#Sub: get_subsystems_for_trigger
#Input Params: hash
#Return values: string containing the subsystems separated by a space
#Summary: Return the subsystems defined for a trigger
###############################################################################
sub get_subsystems_for_trigger {
    my %params = (
            node=>"local",
            'autosupport-message'=>undef,
            @_,
    );
    my $trigger=get_autosupport_trigger(%params);
    my $result;
    if ($trigger->[0]->{'basic-default'} ne '-') {
        $result=$trigger->[0]->{'basic-default'};
    }
    if ($trigger->[0]->{'troubleshooting-default'} ne '-') {
        $result=$result.",".$trigger->[0]->{'troubleshooting-default'};
    }
    if ($trigger->[0]->{'basic-additional'} ne '-') {
        $result=$result.",".$trigger->[0]->{'basic-additional'};
    }
    if ($trigger->[0]->{'troubleshooting-additional'} ne '-') {
        $result=$result.",".$trigger->[0]->{'troubleshooting-additional'};
    }
    return $result;
}

###############################################################################
#Sub: verify_manifest
#Input Params: Hash
#Return values: 1 (success) or hash of hashes with failures per asup
#Summary: Validates the saved autosupport attachments match the manifest
#         It is assumed that ASUP::extract_archive() has been called prior
#         to running this routine.
###############################################################################
#enum asup_manifest_status "AutoSupport manifest collection status" {
#        requested                        "ASUP request has been added to the queue and is waiting processing by the collector"
#        working                          "ASUP Collector is actively gathering the needed data"
#        file-not-found                   "ASUP collection failed, a needed file is missing"
#        no-such-table                    "ASUP collection was unable to find the requested SMF table"
#        collection-truncated-size-limit  "ASUP collection truncated due to size limits and partial data is available"
#        collection-skipped-size-limit    "ASUP collection skipped due to size limits and no data is available"
#        collection-truncated-time-limit  "ASUP collection truncated due to time limits and partial data is available"
#        collection-skipped-time-limit    "ASUP collection skipped due to time limits and no data is available"
#        delivery-skipped-size-limit      "ASUP data was skipped at delivery time due to size limits"
#        general-error                    "ASUP collection failed, additional information if any will be in the Error String field"
#        completed                        "ASUP collection is complete, ready for delivery"
#        content-not-collected-mode       "ASUP content was not collected, incompatible operational mode"
#        content-not-collected-precheck   "ASUP content was not collected, pre-check function violation"
#        content-not-collected-privacy    "ASUP content was not collected, the operation is disabled in privacy mode"
#        content-empty                    "ASUP content was collected successfully, but output was empty"
#        collection-aborted               "ASUP dblade collections are aborted, all commands following a command that was aborted by a party other than autosupport running in notifyd


sub verify_manifest {
    my $me = (caller(0))[3];
    my %params = (
            asup_dir=>undef, #dirs of the saved asups, array of dirs
            subsystems=>undef, #subsystems to validate against, passed in via string
            check_status=>1, #should the routine validate file status
            content=>"complete", #check if the minimal flag is on when validating file existence
            @_,
    );
    my $manifest_errors;

    foreach my $dir (@{$params{asup_dir}}) {
        $Log->comment("$me: Checking manifest in asup $dir \n");
        extract_archive(asup_dir=>$dir);  
        
        
        # read the manifest
        my $manifestFile = "$dir/manifest.xml";
        
        # for each entry in the manifest - find it in the directory
        
        # Need to do a directory listing and load all of the files into a dictionary
        # and then when we iterate over the manifest - we can mark the files as being found in 
        # directory.  This way we can determine any extra files in the directory that isn't in the manifest
        # or vice versa
        my %filesInWorkingDirectory = ();
        
        # read in the extracted files of the asup directory, excluding originals
        opendir(DIR,$dir);
        while (my $f = readdir(DIR)) {
        	if ($f ne "." &&
                $f ne ".." &&
                $f ne "manifest.xml" &&
                $f !~ /^msg.*\.txt$/ &&
                $f ne "body.tgz" &&
                $f ne "httppost.log" &&
                $f ne "httpspost.log" &&
                $f ne "httpppre.log" &&
                $f ne "email" &&
                $f ne "body.tar" &&
                $f ne "body.7z") {
                    #$Log->comment("$me: Adding file $f to files in working directoy\n");
                    $filesInWorkingDirectory{ $f } = '';
            }
                if ($f =~ /.xml/) {
                   $Log->trace("$me: Verifying $f is valid XML\n");
                   if ($f =~ /sis_policy_ui/) {
                       $Log->comment("$me: Ignoring XML validity for $f due to BURT 556378");
                   } 
                   elsif ($f =~ /boottimes/) {
                       $Log->comment("$me: Ignoring XML validity for $f due to BURT 489668");
                   }
                   elsif (!validate_xml(xmlfile=>$dir."/".$f )) {
                       $Log->comment("$me: Failed to verify valid XML format for $f \n");
                       closedir(DIR);
                       return 0;
                   }
                }
        }
        closedir(DIR);
        $Log->comment("$me: minimal param value: $params{content}");
        my $xmlPath = XML::XPath->new(filename=>$manifestFile);
        # An XML::XPath nodeset is an object which contains the result of
        # smacking an XML document with an XPath expression; we'll do just
        # this, and then query the nodeset to see what we get.
        my $nodeset = $xmlPath->find('//asup:ROW');
        my @nodelist = $nodeset->get_nodelist;
		my $nodename = $Server->get_execution_server();

        my $Api_Set_Obj;
        if ( $ASUP::DEBUG_BURT662241 ) {
		    #add for BURT662241 - put smftrace --------- deleted if resolved
		    my $host = host("$nodename");
		       $Api_Set_Obj = NACL::APISet->new(
						    hostobj   => $host,
						    category  => "nodename",
						    interface => "CLI",
						    set       => "Systemshell");
		    #end for BURT662241 - put smftrace --------- deleted if resolved
        }

        foreach my $rowNode (@nodelist) {
            my $bodyFile = $rowNode->findvalue('body-file');
            my $cmd = $rowNode->findvalue('cmd');
            my $cmdTgt = $rowNode->findvalue('cmd-tgt');
            my $contentType = $rowNode->findvalue('content-type');
            my $errormessage = $rowNode->findvalue('error');
            my $prioNum = $rowNode->findvalue('prio-num');
            my $query = $rowNode->findvalue('query');
            my $seqNum = $rowNode->findvalue('seq-num');
            my $sizeCollected = $rowNode->findvalue('size-collected');
            my $status = $rowNode->findvalue('status');
            my $subsystem = $rowNode->findvalue('subsys');
            my $timeCollected= $rowNode->findvalue('time-collected-ms');            
            #$Log->comment("$me: Checking for $bodyFile \n");
            #$Log->comment("$me: RowNode Values: $bodyFile,$cmd,$cmdTgt,$contentType,$errormessage,$prioNum,$query,$seqNum,$sizeCollected,$status,$subsystem,$timeCollected\n");
            # Need to verify the size on the disk compared to the manifest
            my $tmpFile = $dir . "/" . $bodyFile;
            my ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks) = stat($tmpFile);
            #if subsystems are passed in, check that commands run are from one of those subsystems	
            if (defined ($params{subsystems})) {
                if ($params{subsystems} !~ /$subsystem/) {
                    $Log->comment("$me: $subsystem subsystem should not have been included in the asup manifest collection.  The allowed subsystems were $params{subsystems} \n");
                    $manifest_errors->{$dir}->{$bodyFile}="$subsystem subsystem should not have been included in this collection";
                }
            }
            #if size is found for a file, check size of attachments for matches
            if ( defined($size) ) {
                #$Log->comment("What's the sizes of the file in/x/eng/rtpeng-testbeds/csiqa/rtpeng_csiqa_scratch_seanf/logs/20100811_075014/asup_10M_verify_management_content/asup_987654-32-0_0067108964_16.7z bytes, $size/$sizeCollected");
                $size=$size; #manifest file shows files size in Bytes, ex. 1123B, so add this to the unzipped file in the dir
            	if ($size ne $sizeCollected) {
            		$Log->comment("$me: File $bodyFile : different size when extracted from body.tgz than reported in the manifest - $size compared to $sizeCollected\n");
                    $manifest_errors->{$dir}->{$bodyFile}="Extracted file size does not match size reported in manifest";
            	} 
            }
            #if no size indicated, then file is not in the directory
            # check if the manifest indicated this
            elsif ($status eq "file-not-found") {
                $Log->trace("$me: File $bodyFile : not included in body.tgz because file-not-found error in manifest\n");
            }
            elsif ($status eq "content-empty" ) {
                $Log->trace("$me: File $bodyFile : not included in body.tgz because This file has not changed since previously sent AutoSupport\n");
            }
            elsif ($status eq "content-not-collected" ) {
                $Log->trace("$me: File $bodyFile : not included in body.tgz because This file has not changed since previously sent AutoSupport\n");
            }            
            elsif ( ($status eq "content-skipped-privacy" || $status eq "content-not-collected-privacy" )&& $params{content} eq "minimal" ){
               $Log->trace("$me: File $bodyFile : not included, because it is private\n");
            }
            elsif ( $status eq "content-not-collected-mode"  ){
               $Log->trace("$me: File $bodyFile : not included because of the filer's mode\n");
            }
            elsif ( $status eq "content-not-collected-precheck"  ){
               $Log->trace("$me: File $bodyFile : not included because This content is only collected for partner AutoSupports\n");
            }
			elsif ( $status eq "collection-skipped-size-limit"  ){
				$Log->trace("$me: File $bodyFile : not included because This file size is over the limit\n");

			}
            elsif ( $status eq "collection-aborted"  ){
               $Log->comment("$me: File $bodyFile : not included because the collection was aborted by a process other than notifyd\n");

            }
            #if the manifest did not indicate the file was missing, then the status of collection must be negative or the file is missing
            else {
                if ($status ne "completed") {
                    $Log->comment("*** $me: Collection problem for file $bodyFile - status was $status due to error: $errormessage\n");

                    if ($bodyFile =~ /leak-data/) {
                        $Log->warn("*** $me: Ignoring collection problem for leak-data.gz - awaiting fix for BURT 570291");
                    }
                    elsif ($params{check_status}) {
                        $manifest_errors->{$dir}->{$bodyFile}="File status was $status because error: $errormessage";
                    }
                    else {
                        $Log->comment("$me: Ignoring status though as check_status was set to 0, so this error will not be reported upstream \n");
                    }
                }
                elsif ( $status eq "completed" && $sizeCollected eq 0 ){
                    $Log->trace("$me: File, $bodyFile size is 0, skipping the check for the files existence\n");
                    next;
                }

                else {
                    $Log->comment("$me: File $bodyFile : Could not find file in body.tgz attachment, though it was indicated in the manifest\n");
                    $manifest_errors->{$dir}->{$bodyFile}="Manifest indicated file, but file was not included in the asup";
                }
            }
            delete $filesInWorkingDirectory{ $bodyFile };

            if ( $ASUP::DEBUG_BURT662241 ) {
			    #add for BURT662241 - put smftrace --------- deleted if resolved
			    if ( $status eq "collection-truncated-time-limit" && $cmdTgt eq "smf_table" ) {
				    my $cmd = "autosupport manifest show -status $status -seq-num $seqNum -field cmd -time-collected $timeCollected";
				    my $smftable =$Server->cli(cmd=>$cmd);
				    my $smftable= (split("\n",$smftable))[-1]; # get last line
				    if ( ASUP::_get_version() =~ /7M/ ) {
					    $cmd= (split(" ",$smftable))[-1]; # get last element
				    } else {
					    $cmd= (split("X\;X",$smftable))[-1]; # get last element
				    }
				    $Log->comment("*** $me: Adding debug smftrace when truncated happen\n");
				    my $raw_output = $Api_Set_Obj->execute_raw_command(command => "ngsh -c \"set diag;debug smdb table $cmd show -smftrace\"");
			    }
			    #end for BURT662241 - put smftrace --------- deleted if resolved
            }

		}
        # Look for files that are in the directory, but not in the manifest file
        my $count_of_files_not_in_manifest=keys(%filesInWorkingDirectory);
        if ($count_of_files_not_in_manifest) {
            $Log->comment("$me: There are $count_of_files_not_in_manifest files that were found in the directory that were not in the manifest.xml\n");
            while ( my ($key, $value) = each(%filesInWorkingDirectory) ) {
                $Log->comment("$me: $key file not indicated in manifest, but was found in the asup directory\n");
                $manifest_errors->{$dir}->{$key}="File not indicated in manifest, but found in asup";
            }
            return 0;
        }  
            
    }
    my $count_of_manifest_errors=keys(%{$manifest_errors});
    if ($count_of_manifest_errors) {
        $Log->comment("$me: Manifest contains errors or does not match content contained in autosupport\n");
        return $manifest_errors;
    }
    else {
        $Log->comment("$me: Found all expected attachments listed in manifest \n");
        return 1;       
    }
}


###############################################################################
#Sub: extract_archive
#Input Params: asup directory
#Return values: 1 (success) or 0 (fail)
#Summary: extracts the asup archives in the directories given
###############################################################################
sub extract_archive {
    my $me = (caller(0))[3];
    my %params = (
            asup_dir=>undef, # asup dir
            timeout=>120, # asup dir
            @_,
    );
    
    my $dir;
    my @dirs;
    my $files;
    
    if ( ref($params{asup_dir}) ) {
      $Log->debug(sub {"$me: asup dir is reference & size, $params{asup_dir}...".keys(%params)});
       @dirs=@{$params{asup_dir}};
      $dir=$dirs[0];
    }
    else {
      $Log->debug(sub {"$me: asup dir is a directory, $params{asup_dir}..."});
      $dir=$params{asup_dir};
    }
    
    # Ideally we should not even be trying to ls this dir
    # until the server has completed writing it out
    $Log->comment("$me: Extracting ASUP Archive in $dir \n");
    my $end_time = time() + $params{timeout};
    my $done      = 0;
    my $rc        = 0;
    while ( !$done && (time() < $end_time) ) {
        eval { $files=`ls $dir`; };
        if ( $@ ) {
            $Log->warn("$me: ls on asup dir resulted in $@");
            $files = "";
        }
        $Log->debug(sub {"$me: Files in the asup dir, $files\n"});
        $rc   = 1;
        $done = 1;
        eval {
            if ($files =~ /body.tgz/) {
                #my $archiveExtract = Archive::Extract->new(archive => "$dir/body.tgz");
                #$archiveExtract->extract(to => $dir) or die $archiveExtract->error;
                ASUP::shell_unix_cmd( cmd => "gzip -d $dir/body.tgz" );
                ASUP::shell_unix_cmd( cmd => "tar -xvf $dir/*.tar -C $dir" );
                ASUP::shell_unix_cmd( cmd => "rm $dir/body.tar" );
            } elsif ($files =~ /body.7z/) {
                $Log->comment("Testing integrity of 7z attachment...");
                my $integrity;
                eval {
                    $integrity = ASUP::shell_unix_cmd( cmd => "/usr/software/bin/7za t $dir/body.7z -r" );
                };
                if ( $@ ) {
                    $integrity = $@;
                }
                $Log->comment($integrity);
                if ($integrity !~ /Everything is Ok/) {
                    $Log->warn("$me: Failed to verify integrity of 7z contents!, saving to $dir/body.7z.save");
                    eval { ASUP::shell_unix_cmd( cmd => "cp -f $dir/body.7z $dir/body.7z.save" ); };
                }
                $Log->comment("$me: Extracting 7z file...");
                ASUP::shell_unix_cmd( cmd => "/usr/software/bin/7za x -o$dir $dir/body.7z" );
                
                $Log->comment("$me: Done extracting.");
                ASUP::shell_unix_cmd( cmd => "rm $dir/body.7z" );
            } elsif ($files =~ /tar.gz/) {
                my $archiveExtract = Archive::Extract->new(archive => "$dir/*.tar.gz");
                $archiveExtract->extract(to => $dir) or die $archiveExtract->error;
                #`gzip -d $dir/*.tar.gz`;
                #`tar -xvf $dir/*.tar -C $dir`;
            } else {
                $rc   = 0;
                $done = 0;
            }
        };

        if ( $@ ) {
            $Log->warn("$me: Extraction resulted in $@");
            $rc   = 0;
            $done = 0;
        }

        if ( $rc == 0 ) {
            $Log->comment("$me: Files in the asup dir, |$files|");
            Tharn::sleep(5);
        }
    }

    if ( $rc == 0 ) {
        $Log->comment("$me: No files to extract: |$files|");
    }

    return $rc;
}

###############################################################################
#Sub: retrieve_manifest_info
#Input Params: asup_dir - array of asup directories
#Return values: hash of hashes containing the manifest information
#Summary: Extracts information from an ASUP manifest file and organizes it like so:
#    $retHash->{$dir}{files}{$subsystem}{$collector}{$prio_num} = 
#                  {
#                      body_file      => collected filename
#                      status         => status of collected file
#                      seq_num        => sequence number
#                      cmd_target     => command target
#                      cmd            => command
#                      size_collected => number of bytes collected for file
#                      time_collected => number of seconds it took to collect the file
#                      error_msg      => errors
#                  }
#  Also, some aggregates are calculated based on the manifest statistics:
#    $retHash->{$dir}{subsystem_aggregates}{$subsystem} = 
#                { size_collected     => Total number of bytes collected for subsystem
#                  time_collected     => Total number of seconds to collect subsystem
#                  num_size_truncated => Number of files that were size-truncated
#                  num_time_truncated => Number of files that were time-truncated
#                  num_delivery_size_truncated => Number of files that were delivery size-truncated
#                  check_size         => Number of bytes collected for subsystem
#                  check_time         => Maximum number of seconds that any one collector took
#                  limit_size         => Approximately 1/2 check_size
#                  limit_time         => Approximately 1/2 check_time
#                };  
#    $retHash->{$dir}{collector_aggregates}{$collector} = 
#                { size_collected     => Total number of bytes collected for subsystem
#                  time_collected     => Total number of seconds to collect subsystem
#                  num_size_truncated => Number of files that were size-truncated
#                  num_time_truncated => Number of files that were time-truncated
#                  num_delivery_size_truncated => Number of files that were delivery size-truncated
#                };  
#
# The limit_* stats are actually to be used for subsequent calls.  Let's say you
# invoke an ASUP with no size or time limits, you can use those limit_* values
# to set a size or time limit, then invoke the same ASUP again.  Those limit_*
# values should guarantee that there will be limiting.  Then you can verify 
# that the check_* values from that subsequent call are equal to the limit_*
# values from the original call.
#         
###############################################################################
sub retrieve_manifest_info {
    my %params = (
            asup_dir=>undef, #dirs of the saved asups, array of dirs
            @_,
    );
    my $manifest_errors;

    my $retHash=undef;
    my @subsystems;
    
    foreach my $dir (@{$params{asup_dir}}) {
        $Log->comment("Checking manifest in asup $dir \n");

        # read the manifest
        my $manifestFile = "$dir/manifest.xml";
         
        my $xmlPath = XML::XPath->new(filename=>$manifestFile);
        # An XML::XPath nodeset is an object which contains the result of
        # smacking an XML document with an XPath expression; we'll do just
        # this, and then query the nodeset to see what we get.
        my $nodeset = $xmlPath->find('//asup:ROW');
        my @nodelist = $nodeset->get_nodelist;
        
        my %size;
        my %time;

        my %size_trunc;
        my %time_trunc;
        my %delivery_trunc;
        my %num_files;
        my %num_slow_files;

        foreach my $rowNode (@nodelist) {
            my $body_file = $rowNode->findvalue('body-file');
            my $cmd = $rowNode->findvalue('cmd');
            my $cmd_target = $rowNode->findvalue('cmd-tgt');
            my $contentType = $rowNode->findvalue('content-type');
            my $error_msg = $rowNode->findvalue('error');
            my $prio_num = int($rowNode->findvalue('prio-num'));
            my $query = $rowNode->findvalue('query');
            my $seq_num = int($rowNode->findvalue('seq-num'));
            my $size_collected = int($rowNode->findvalue('size-collected'));
            my $status = $rowNode->findvalue('status');
            my $subsystem = $rowNode->findvalue('subsys');
            my $time_collected= int($rowNode->findvalue('time-collected-ms'));     

            # determine who collected this file
            my $collector = _determine_collector( target => $cmd_target );

            if ( !exists $num_files{$subsystem} ) {
                $num_files{$subsystem} = 0;
            }
            if ( !exists $size_trunc{$subsystem} ) {
                $size_trunc{$subsystem} = 0;
            }
            if ( !exists $time_trunc{$subsystem} ) {
                $time_trunc{$subsystem} = 0;
            }
            if ( !exists $delivery_trunc{$subsystem} ) {
                $delivery_trunc{$subsystem} = 0;
            }
            if ( !exists $num_files{collector}{$collector} ) {
                $num_files{collector}{$collector} = 0;
            }
            if ( !exists $num_slow_files{sub_collector}{$subsystem}{$collector} ) {
                $num_slow_files{sub_collector}{$subsystem}{$collector} = 0;
            }
            if ( !exists $size_trunc{collector}{$collector} ) {
                $size_trunc{collector}{$collector} = 0;
            }
            if ( !exists $time_trunc{collector}{$collector} ) {
                $time_trunc{collector}{$collector} = 0;
            }
            if ( !exists $delivery_trunc{collector}{$collector} ) {
                $delivery_trunc{collector}{$collector} = 0;
            }

            if ( !$size_collected ) {
                $size_collected = 0;
            }

            if ( !$time_collected ) {
                $time_collected = 0;
            }
            
            # prio_num can be used for relative sequencing within a collector
            $retHash->{$dir}{files}{$subsystem}{$collector}{$prio_num} = 
                                   {   
                                      body_file      => $body_file,
                                      status         => $status,
                                      seq_num        => $seq_num,
                                      cmd_target     => $cmd_target,
                                      cmd            => $cmd,
                                      size_collected => $size_collected,
                                      time_collected => $time_collected,
                                      error_msg      => $error_msg,
                                   };

            # strip off the trailing B
            my $num_size = $size_collected;
               $num_size =~ s/B$//;

            if ( $num_size =~ /K$/ ) {
                $num_size =~ s/K$//;
                $num_size *= 1024;
            } elsif ( $num_size =~ /M$/ ) {
                $num_size =~ s/M$//;
                $num_size *= (1024 * 1024);
            }
 
            # Save sizes/times per collector
            if ( exists $size{$subsystem}{$collector} ) {
                push( @{$size{$subsystem}{$collector}}, $num_size );
            } else {
                $size{$subsystem}{$collector} = [ $num_size ];
            }
   
            # time collected only supported in RR and beyond
            if ( exists $time{$subsystem}{$collector} ) {
                push( @{$time{$subsystem}{$collector}}, $time_collected );
            } else {
                $time{$subsystem}{$collector} = [ $time_collected ];
            }

            $num_files{$subsystem}++;
            $num_files{collector}{$collector}++;

            if ( $time_collected > 100 ) {
                $num_slow_files{sub_collector}{$subsystem}{$collector}++;
            }

            # Since the collectors run in parallel, keep track of the collector
            # that is collecting the largest number of "slow-to-collect" files
            if ( exists $num_slow_files{subsystem}{$subsystem} ) {
                $num_slow_files{subsystem}{$subsystem} = 
                    ($num_slow_files{sub_collector}{$subsystem}{$collector} >
                     $num_slow_files{subsystem}{$subsystem}) ? $num_slow_files{sub_collector}{$subsystem}{$collector} :
                                                               $num_slow_files{subsystem}{$subsystem};
            } else {
                $num_slow_files{subsystem}{$subsystem} = $num_slow_files{sub_collector}{$subsystem}{$collector};
            }

            if ( $status =~ /collection-truncated-size-limit/i or $status =~ /collection-skipped-size-limit/i) {

                # size truncation only supported in RR and beyond
                $size_trunc{$subsystem}++;
                $size_trunc{collector}{$collector}++;

            } elsif ( $status =~ /collection-truncated-time-limit/i or $status =~ /collection-skipped-time-limit/i) {
                
                # time truncation only supported in RR and beyond
                $time_trunc{$subsystem}++;
                $time_trunc{collector}{$collector}++;
            }
            elsif ( $status =~ /delivery-truncated-size-limit/i or $status =~ /delivery-skipped-size-limit/i) {
                
                # delivery truncation only supported in RR and beyond
                $delivery_trunc{$subsystem}++;
                $delivery_trunc{collector}{$collector}++;
            }
        }

        # get the largest size collected by a given subsystem
        my @subsystems = ( keys %size );
        foreach my $subsystem ( keys %size ) {

            # Stats of actual collection values
            my $total_size = 0;
            my $total_time = 0;

            # Stats to use for setting subsequent collection limits
            my $limit_time = 0;
            my $limit_size = 0;

            # Collected stats which should be less than the limits
            my $check_size = 0;
            my $check_time = 0;

            foreach my $collector ( keys %{$size{$subsystem}} ) {

                # add up the array
                my $total_collector_size = 0;
                my $total_collector_time = 0;

                my $largest_size = 0;
                my $largest_time = 0;

                # Sum up the sizes for this collector
                foreach my $sz (@{$size{$subsystem}{$collector}}) {
                    $total_collector_size += $sz;

                    if ( $sz > $largest_size ) {
                        $largest_size = $sz
                    }
                }
    
                # Sum up the times for this collector
                foreach my $tm (@{$time{$subsystem}{$collector}}) {
                    $total_collector_time += $tm;

                    if ( $tm > $largest_time ) {
                        $largest_time = $tm
                    }
                }

                # We want to figure out which collector has the largest files
                # and make our limit be 1/2 of that sum (minus the largest file).
                # my concern is that we might have the following scenario
                # 2,5,20,5,1000 where 1/2 of the total (.5*1032 = 516) because
                # we would only truncate the last file.
                my $half_size = int($total_collector_size-$largest_size/2);
                my $half_time = int($total_collector_time-$largest_time/2);

                $limit_size = ($half_size > $limit_size) ? $half_size : $limit_size;
                $limit_time = ($half_time > $limit_time) ? $half_time : $limit_time;

                # accumulate the total bytes/time for a given subsystem
                # that should be less than the limit
                $check_size += $total_collector_size;
                $check_time = ($total_collector_time > $check_time) ? 
                               $total_collector_time : $check_time;

                # accumulate the total bytes for a given subsystem.
                $total_size += $total_collector_size;

                # Since the collectors are running in parallel, the time will be
                # the max of any one collector time
                $total_time = ($total_collector_time > $total_time) ? 
                               $total_collector_time : $total_time;

                if ( !exists $retHash->{$dir}{collector_aggregates}{$collector} ) {
                    $retHash->{$dir}{collector_aggregates}{$collector} =
                        { size_collected     => $total_collector_size,
                          time_collected     => $total_collector_time,
                          num_files          => $num_files{collector}{$collector},
                          num_size_truncated => $size_trunc{collector}{$collector},
                          num_time_truncated => $time_trunc{collector}{$collector},
                          num_delivery_size_truncated => $delivery_trunc{collector}{$collector},
                        };
                } else {
                    $retHash->{$dir}{collector_aggregates}{$collector}{size_collected} += $total_collector_size;
                    $retHash->{$dir}{collector_aggregates}{$collector}{time_collected} += $total_collector_time;
                }
            }

            # Note, check_size and limit_size will not generally be meaningful in a single call.
            # The intent is that you make 2 calls, one with no limits for a subsystem
            # in which case "limit_size" will be what you will later use when you set the subsystem
            # limits and then invoke another ASUP and the "check_size" will essentially be
            # meaningless.
            #
            # In the subsequent call when subsystem limits ARE imposed, the "check_size" 
            # will be what is checked to make sure that it is less than the subsystem limit
            # we set, and the "limit_size" will be meaningless.

            $retHash->{$dir}{subsystem_aggregates}{$subsystem} = 
                { size_collected     => $total_size,
                  time_collected     => $total_time,
                  num_files          => $num_files{$subsystem},
                  num_slow_files     => $num_slow_files{subsystem}{$subsystem},
                  file_sizes         => $size{$subsystem},
                  file_times         => $time{$subsystem},
                  num_size_truncated => $size_trunc{$subsystem},
                  num_time_truncated => $time_trunc{$subsystem},
                  num_delivery_size_truncated => $delivery_trunc{$subsystem},
                  check_size         => $check_size,
                  check_time         => $check_time,
                  limit_size         => $limit_size,
                  limit_time         => $limit_time, };
        }

    }
    return $retHash;
}

###############################################################################
#Sub: verify_manifest_limits
#Input Params: Hash
#Return values: 1 (success) or 0 (failure)
#Summary: Checks the size or time collected does not exceed the limits
#         passed in.
#         
###############################################################################
sub verify_manifest_limits {
    my %params = (
            asup_dir=>undef, #dirs of the saved asups, array of dirs
            subsystem=>undef, #subsystem to check
            limits=>undef, #limits per subsystem, array of hashes
            check_size=>0, #check size limits?
            check_time=>0, #check time limits?
            @_,
    );

    if ( !$params{limits} ) {
        my $err = "Must supply \"limits\" hash to check against";
        $Log->error($err);
        return 0;
    }

    if ( !($params{check_size} xor $params{check_time}) ) {
        my $err = "Must specify non-zero value for exactly one ".
                   "of the following options \"check_size|check_time\"";
        $Log->error($err);
        return 0;
    }

    my @subsystems;

    if ( ref( $params{subsystem} ) =~ /ARRAY/i ) {
        push( @subsystems, @{$params{subsystem}} );
    } else {
        push( @subsystems, $params{subsystem} );
    }

    my $limits = $params{limits};

    # The logic to determine how many bytes should have been collected
    # is in retrieve_manifest_info()
    my $manifest = retrieve_manifest_info( asup_dir=>$params{asup_dir} );

    # Traverse the limits hash and for each subsystem, check to see if 
    # we are truncating files that exceed the limit.

    # First check that the check_size/time for each subsystem in manifest
    # does not exceeded the limit_size/time
    foreach my $dir (keys %{$manifest}) {
        $Log->comment("Checking manifest size limits in $dir\n");

        foreach my $subsystem ( @subsystems ) {
            if ( $params{check_size} ) {
                if ( $limits->{subsystem_aggregates}{$subsystem}{limit_size} ) {
                    if ( $manifest->{$dir}{subsystem_aggregates}{$subsystem}{check_size} <=
                         $limits->{subsystem_aggregates}{$subsystem}{limit_size} ) {
                        $Log->comment("Collection size for subsystem $subsystem, ".
                                   "EXP: <=|$limits->{subsystem_aggregates}{$subsystem}{limit_size}|, ".
                                   "ACT: |$manifest->{$dir}{subsystem_aggregates}{$subsystem}{check_size}| " );
                    } else {
                        my $err =  "Collection size for subsystem $subsystem, ".
                                   "EXP: <=|$limits->{subsystem_aggregates}{$subsystem}{limit_size}|, ".
                                   "ACT: |$manifest->{$dir}{subsystem_aggregates}{$subsystem}{check_size}| ";
                        $Log->error($err);
                        return 0;
                    }
    
                    # Make sure there is at least one size truncated file entry
                    if ( $manifest->{$dir}{subsystem_aggregates}{$subsystem}{num_size_truncated} == 0 ) {
                        my $err = "Expected to find at least 1 collection-size-truncated file in $subsystem";
                        $Log->error($err);
                        return 0;
                    } else {
                        $Log->comment("Found $manifest->{$dir}{subsystem_aggregates}{$subsystem}{num_size_truncated} ".
                                      "collection-size-truncated file(s) in $subsystem");
                    }
                } else {
                    if ( $manifest->{$dir}{subsystem_aggregates}{$subsystem}{num_size_truncated} > 0 ) {
                        $Log->error("Found $manifest->{$dir}{subsystem_aggregates}{$subsystem}{num_size_truncated} ".
                                      "collection-size-truncated file(s) in $subsystem");
                        return 0;
                    }
                }
            } else {
                if ( $manifest->{$dir}{subsystem_aggregates}{$subsystem}{num_size_truncated} > 0 ) {
                    $Log->error("Found $manifest->{$dir}{subsystem_aggregates}{$subsystem}{num_size_truncated} ".
                                  "collection-size-truncated file(s) in $subsystem");
                    return 0;
                }
            }

            if ( $params{check_time} ) {
                if ( $limits->{subsystem_aggregates}{$subsystem}{limit_time} ) {
                    # Note, for time collection, it is much more likely that
                    # we will have one or two files that take relatively more
                    # time than the others (i.e. many files in the subsecond range
                    # and one multi-second file).  Since they can't tell how long
                    # a file will take to collect until they collect it, we could
                    # be over by one file time.  Since that file might be the long
                    # to collect file, we could be significantly over the time limit.
                    # For example if I set my limit to 1 second, and there are
                    # 5 files that take 50ms each to collect and one 2 second file,
                    # most likely we will collect that 2 second file and go over our
                    # limit by more than 100%.
                    # So we have to just check that there were SOME time truncated files.
    
                    # Make sure there is at least one time truncated file entry
                    if ( $manifest->{$dir}{subsystem_aggregates}{$subsystem}{num_time_truncated} == 0 ) {
                        my $err = "Expected to find at least 1 collection-time-truncated file in $subsystem";
                        $Log->error($err);
                        return 0;
                    } else {
                        $Log->comment("Found $manifest->{$dir}{subsystem_aggregates}{$subsystem}{num_time_truncated} ".
                                      "collection-time-truncated file(s) in $subsystem");
                    }
                } else {
                    # Not limiting this subsystem
                    if ( $manifest->{$dir}{subsystem_aggregates}{$subsystem}{num_time_truncated} > 0 ) {
                        $Log->error("Found $manifest->{$dir}{subsystem_aggregates}{$subsystem}{num_time_truncated} ".
                                      "collection-time-truncated file(s) in $subsystem");
                        return 0;
                    }
                }
            } else {
                # Not limiting times for any subsystem
                if ( $manifest->{$dir}{subsystem_aggregates}{$subsystem}{num_time_truncated} > 0 ) {
                    $Log->error("Found $manifest->{$dir}{subsystem_aggregates}{$subsystem}{num_time_truncated} ".
                                  "collection-time-truncated file(s) in $subsystem");
                    return 0;
                }
            }
        }
    }
    return 1;
}

###############################################################################
#Sub: modify_autosupport_budget
#Input Params: subsystem, size-limit, time-limit
#Return values: 1 on success, 0 on failure
#Summary: Provides ability to modify the size or time budgets for a given subsystem.
###############################################################################
sub modify_autosupport_budget {
    my %params = (
            node=>"local",
            'subsystem'=>'*',
            'size-limit'=>'-1',
            'time-limit'=>'-1',
            @_,
    );
    my $version=_get_version();
    my $result;
    my @args = ();
    if ($version=~/7M/) {
        delete $params{node};
    }

    # can't use undef or zero to signify "not set" because
    # those are valid settings for the limits
    foreach my $item (keys %params) {
        if ( $params{$item} ne "-1" ) {
            push(@args, $item, $params{$item});
        }
    }

    if ($version=~/7M/) {
        $Server->autosupport_budget_modify( @args );
        $result=get_autosupport_budget( @args );
        foreach my $res (@$result) {
            foreach my $item (keys %params) {
                if ( ($item ne "node") && ($params{$item} ne "-1") && ($res->{$item} ne $params{$item}) ) {
                    $Log->comment("Autosupport budget modify failed. Queried ".$res->{$item}." does not equal ".$params{$item});
                    return 0;
                }
            }
        }
    }
    else {#if 10M
        $Server->system_node_autosupport_budget_modify(@args);
        if ($@) {
           $Log->comment("Autosupport filter modify failed with $@\n");
            return 0; 
        }
    }
    return 1;
}

###############################################################################
#Sub: retrieve_saved_autosupport
#Input Params: hash
#Return values: asup data on success, 0 on failure
#Summary: Returns the specified autosupport data
###############################################################################
sub retrieve_saved_autosupport {
    my %params = (
            asup_path=>undef, #path and filename of saved asup to retrieve
            @_,
    );
    my $version=_get_version();
    my $asupdata=undef;
    if ($version =~ /7M/) {
        $asupdata=$Server->cli(cmd=>"rdfile $params{asup_path}");
    }
    #CM
    else {
        $asupdata=$Server->shell(cmd=>"cat $params{asup_path}");
    }
    if (!defined $asupdata) {
        $Log->comment("retrieve_saved_autosupport: Failed to find autosupport in $params{asup_path} \n");
        return 0;
    }
    return $asupdata;   

}

###############################################################################
#Sub: restore_autosupport_budget
#Input Params: hash
#Return values: 1 on success, 0 on failure
#Summary: Provides ability to restore previously retrieved autosupport budget settings
###############################################################################
sub restore_autosupport_budget {
    my %params = (
            values=>undef,
            @_,
    );

    if ( !$params{values} ) {
        $Log->error( "Must provide values array of hashes to ASUP::restore_autosupport_budget()" );
        return;
    }

    foreach my $item ( @{$params{values}} ) {
        next if ($item->{subsystem} eq "mandatory");
        ASUP::modify_autosupport_budget( %{$item} );
    }
}

###############################################################################
#Sub: get_autosupport_budget
#Input Params: hash
#Return values: pointer to array of hash
#Summary: Provides ability to return the current autosupport budget settings
###############################################################################
sub get_autosupport_budget {
    my %params = (
            node=>'local',
            subsystem=>'*',
            'time-limit'=>'*',
            'size-limit'=>'*',
            @_,
    );
    my $version=_get_version();
    my $result;
    my @args = ();
    if ($version=~/7M/) {
        delete $params{node};
    }

    foreach my $item (keys %params) {
        if ( $params{$item} ) {
            push(@args, $item, $params{$item});
        }
    }

    if ($version=~/7M/) {
        my @show_details=split("\n",($Server->autosupport_budget_show(@args)));
        my $i=0;
        foreach my $show_line (@show_details) {
            if ($show_line =~/Subsystem: (.+)/) {
                $result->[$i]->{'subsystem'}=$1;
            }
            elsif ($show_line =~/Time Limit: (.+)/) {
                $result->[$i]->{'time-limit'}=$1;
            }
            elsif ($show_line =~/Size Limit in Octets: (.+)/) {
                $result->[$i]->{'size-limit'}=$1;
                $i++;
            }
        }
    }
    else {#if 10M
        $result=$Server->system_node_autosupport_budget_show(@args);
    }
    return $result;
} 


###############################################################################
#PRIVATE ASUP FUNCTIONS
###############################################################################

###############################################################################
#Sub: _get_licensing
#Input Params: hash
#Return values: Ref to Hash
#Summary: Retrieves the current license information from the node and returns it
###############################################################################
sub _get_licensing {
    my %params = (
            version=>undef,
            @_,
    );
    
    my $license_hash;
    #get the license info
    my $licensing;
    #7M
    if ($params{version}=~/7M/) {
        eval {
            $licensing=$Server->cli(cmd=>"license");
        };
    }
    #GX/CM
    else {
        eval {
            $licensing=$Server->system_node_run(node=>"local",command=>"license"); 
        };
    }
    my @licenses=split("\n",$licensing);
    #populate the license hash {license} = {license info or not licensed}
    foreach my $license (@licenses){
        my @license_values=split(" ",$license);
        $license_hash->{$license_values[0]}=$license_values[1]." ".$license_values[2];
    }
    #return hash of licenses
    return $license_hash;
}

###############################################################################
#Sub: _get_version
#Input Params: <none>
#Return values: Scalar of the version running
#Summary: Retrieves the current Ontap version running on the node
###############################################################################
sub _get_version {

    use NATE::BaseException qw(:try);
    my %params = (
            force=>0, # 1 = do not use cached value of $ontap_version (re-query filer)
            @_,
    );
    
    if ((!$ontap_version) || ($params{force} eq '1')) {
        
        if (!$ontap_version) {
            $Log->comment("ONTAP version is unknown; obtaining from filer...");
        }
        elsif ($params{force}) {
            $Log->comment("Force parameter specified; obtaining version from filer...");
        }

        my $build;
        my $version;
        my $Node = NACL::C::Node->find();
    
        # Automation branch test:main now reflects Sierra Nevada
        # We will therefore assume SN and set 10M or 7M based on result from NACL
        if ($Node->mode() eq 'CMode') {
            $version="SN-10M";
        }
        else {
            $version="SN-7M";
        }
    
        $Log->comment("_get_version: Found node is version $version \n");
        $ontap_version = $version;
    }
    else {
        $Log->comment("_get_version: Using cached ONTAP version $ontap_version \n");
    }
    
    return $ontap_version;
}

###############################################################################
#Sub: _verify_headers
#Input Params: Hash
#Return values: 1 (success) or 0 (failure)
#Summary: Validates the expected headers in each saved autosupport
###############################################################################
sub _verify_headers {
    my $me = (caller(0))[3];
    my %params = (
            asup_dir=>undef, #dirs of the saved asups
            xfer_protocol=>'PUT|POST', # by default either is ok
            @_,
    );
    my @valid_headers;   
    my $version=_get_version();
    
    my $done;

    #check for the required headers in each saved asup
    foreach my $dir (@{$params{asup_dir}}) {
        my $files;

        my $end_time = time() + 30;
        $done = 0;
        while ( !$done && (time() < $end_time ) ) {
            eval { $files=`ls $dir`; };
            if ( $@ || !$files ) {
                $Log->warn("$me: Could not ls $dir");
                Tharn::sleep(5);
            } else {
                $done = 1;
            }
        }

        if ( !$done ) {
            $Log->warn("$me: Failed to find files in $dir");
            return 0;
        }

        #first determine what headers we should expect based on the saved asup type
        my $headers=undef;
        #if ($dir =~ /SMTP/) {
        if ($files =~ /email/) {
            #email_1 contains the raw SMTP asup
            $headers=`head -35 $dir/email`; 
            @valid_headers=("SENDER:","RECEIVER:","MESSAGE:","To:","Date:","Subject:","Mime-Version:","Content-Type:","X-Mailer:","X-Mailer:","X-Netapp-asup-version:",
                    "X-Netapp-asup-content:","X-Netapp-asup-subject:","X-Netapp-asup-system-id:","X-Netapp-asup-os-version:","X-Netapp-asup-hostname:",
                    "X-Netapp-asup-generated-on:","X-Netapp-asup-serial-num:","X-Netapp-asup-model-name:",
                    "X-Netapp-asup-conf-version:","X-Netapp-asup-conf-checksum:","X-Netapp-asup-from:");
            if ($version ne "7G") {
                push (@valid_headers,"X-Netapp-asup-partner-system-id:")
            }
        }
        #elsif ($dir =~ /HTTPS/){
        elsif ($files =~ /httpspost.log/) {
            #httpspost_1.log contains the raw HTTPS asup
            $headers=`head -35 $dir/httpspost.log`;
            @valid_headers=($params{xfer_protocol},"Cache-Control: no-store","Connection: close","Expect: 100-continue","Host:","User-Agent:","Content-Length:","Content-Type:",
                            "X-Netapp-Asup-Conf-Checksum:","X-Netapp-Asup-Conf-Version:","X-Netapp-Asup-Content:","X-Netapp-Asup-From:","X-Netapp-Asup-Generated-On:",
                            "X-Netapp-Asup-Hostname:","X-Netapp-Asup-Model-Name:","X-Netapp-Asup-Os-Version:",
                            "X-Netapp-Asup-Serial-Num:","X-Netapp-Asup-Subject","X-Netapp-Asup-System-Id:","X-Netapp-Asup-Version");
            if ($version ne "7G") {
                push (@valid_headers,"X-Netapp-Asup-Partner-System-Id:")
            }
        }
        else {
            #httppost_1.log contains the raw HTTP asup
            $headers=`head -35 $dir/httppost.log`;
            @valid_headers=($params{xfer_protocol},"Cache-Control: no-store","Connection: close","Expect: 100-continue","Host:","User-Agent:","Content-Length:","Content-Type:",
                            "X-Netapp-Asup-Conf-Checksum:","X-Netapp-Asup-Conf-Version:","X-Netapp-Asup-Content:","X-Netapp-Asup-From:","X-Netapp-Asup-Generated-On:",
                            "X-Netapp-Asup-Hostname:","X-Netapp-Asup-Model-Name:","X-Netapp-Asup-Os-Version:",
                            "X-Netapp-Asup-Serial-Num:","X-Netapp-Asup-Subject","X-Netapp-Asup-System-Id:","X-Netapp-Asup-Version");
            if ($version ne "7G") {
                push (@valid_headers,"X-Netapp-Asup-Partner-System-Id:")
            }
        }
        #now check for the valid headers
        foreach my $basic_header (@valid_headers) {
            if ($headers !~ /$basic_header/) {      
                $Log->warn("$me: Failed to find a basic header, $basic_header, in the saved asup $dir \n");
                return 0;
            }
            $Log->debug(sub {"$me: Validated $basic_header!!!"});
        }
    }
    #found all of the expected headers in the saves asup dirs
    $Log->comment("$me: Found all expected asup headers \n");
    return 1;
}

###############################################################################
#Sub: _determine_collector
#Input Params: command target string
#Return values: string
#Summary: Returns the collector name given a command target name
###############################################################################
sub _determine_collector {
    my %params = (
            target=>undef,
            @_,
    );

    my $retval = "notifyd";

    if ( $params{target} =~ /^dblade/ ) {
        $retval = "dblade";
    } elsif ( $params{target} =~ /^smf/ ) {
        $retval = "mgwd";
    }

    return $retval;
}
###############################################################################
#Sub: rcp_file
#Input Params: host, file to copy, host:<director to copy>
#Return values: string
#Summary: rcp's a file to a specific location
###############################################################################
sub rcp_file {

    my ($host,$file_to_copy, $where_to_copy) = @_;
    my $version=_get_version();
    my $mode = "CMode";
    #7G/7M
    if ($version=~/7M/) {
        $mode="7Mode";
    }
    my $Api_Set_Obj = NACL::APISet->new(hostobj   => $host,
                                     category  => "Node",
                                     interface => "CLI",
                                     set       => "Systemshell",
                                     shell_type => $mode,
                                     context_definition => {username => "diag", password => "netapp1!"}
                                     );
    my $response = $Api_Set_Obj->execute_raw_command(command=>"sudo rcp $file_to_copy $where_to_copy");   # Run
    return $response;
}

###############################################################################
#Sub: _verify_attachments
#Input Params: Hash
#Return values: 1 (success) or 0 (failure)
#Summary: Validates expected attachments in each saved autosupport directory
###############################################################################
sub _verify_attachments {
    my %params = (
            asup_dir=>undef, #dirs where asups were saved
            alt_extension=>0,
            minimal_content=>0,
            @_,
    );
    
    my @valid_attachments;
    
    my $file_ext="gz";
    
    if ($params{alt_extension}) {
        $file_ext="gx";
    }
    
    #check for node version, 7mode
    my $version=_get_version();
    
    #Get the current licensing on the node
    my $licenses=_get_licensing(version=>$version);
    
    #check for minimal content
    if ($params{minimal_content}) {
        if ($version=~/7M/) {
            @valid_attachments=("asup_stats.xml.$file_ext","messages.$file_ext");
        }
        elsif ($version=~/10M/ or $version eq "GX") {
            $Log->error("_verify_attachments: Minimal content type specfied, but not supported in BR \n");
            return 0;
        }
    }
    #else 7M
    elsif ($version=~/7M/) {
    
        #basic attachments that are generated with each complete autosupport
        @valid_attachments=("cm_stats.$file_ext","ems.$file_ext","exports.$file_ext","hosts.$file_ext","nsswitch_conf.$file_ext","rc.$file_ext","resolv_conf.$file_ext","shelf_log");   
    
        #if node is 7M, add some other basic attachments
        if ($version=~/7M/) {
            push(@valid_attachments,("messages.$file_ext","quotas.$file_ext","asup_stats.xml.$file_ext","tar.$file_ext","config_xml.$file_ext"));
        }
    
        #now check for existance of additional files, if present, these should be sent
        #with the complete autosupport
        if ($licenses->{snapmirror} !~ /not licensed/) {
            if ($Server->cli(cmd=>"rdfile /etc/snapmirror.allow") !~ /No such file/) {
                push (@valid_attachments,"sm_allow.$file_ext");
            }
            if ($Server->cli(cmd=>"rdfile /etc/snapmirror.conf") !~ /No such file/) {
                push (@valid_attachments,"sm_conf.$file_ext");
            }
            if ($Server->cli(cmd=>"rdfile /etc/log/snapmirror") !~ /No such file/) {
                push (@valid_attachments,"sm_log.$file_ext");
            }
        }
        if ($Server->cli(cmd=>"rdfile /etc/log/leak_data") !~ /No such file/) {
            push (@valid_attachments,"leak_data.$file_ext");
        }
        if ($Server->cli(cmd=>"rdfile /etc/syslog.conf") !~ /No such file/) {
            push (@valid_attachments,"syslog_conf.$file_ext");
        }
        if ($Server->cli(cmd=>"rdfile /etc/log/membufstats.log") !~ /No such file/) {
            push (@valid_attachments,"memstats.$file_ext");
        }
        if ($Server->cli(cmd=>"rdfile /etc/usermap.cfg") !~ /No such file/) {
            push (@valid_attachments,"usermap_cfg.$file_ext");
        }
        if ($Server->cli(cmd=>"rdfile /etc/log/backup") !~ /No such file/) {
            push (@valid_attachments,"backup.$file_ext");
        }
        if ($Server->cli(cmd=>"rdfile /etc/log/cpdump") !~ /No such file/) {
            push (@valid_attachments,"cpdump.$file_ext");
        }
        if ($Server->cli(cmd=>"rdfile /etc/crash/parity_inconsistencies") !~ /No such file/) {
            push (@valid_attachments,"pi_data.$file_ext");
        }
        if ($Server->cli(cmd=>"rdfile /etc/log/fmm_data") !~ /No such file/) { #Virgo card attachment
            push (@valid_attachments,"fmm_data.$file_ext");
        }
    }
    #else GX/CM
    else {
    
        @valid_attachments=("cm_stats.$file_ext","ems.$file_ext","rc.$file_ext","shelf_log","tar.$file_ext");   
        
        if ($version=~/10M/) {
            push(@valid_attachments,("messages.$file_ext","asup_stats.xml.$file_ext","config_xml.$file_ext"));
        }
        
        if ($Server->shell(cmd=>"ls /mroot/etc/log/leak_data") !~ /No such file/) {
            push (@valid_attachments,"leak_data.$file_ext");
        }
        if ($Server->shell(cmd=>"ls /mroot/etc/log/membufstats.log") !~ /No such file/) {
            push (@valid_attachments,"memstats.$file_ext");
        }
        if ($Server->shell(cmd=>"ls /mroot/etc/log/backup") !~ /No such file/) {
            push (@valid_attachments,"backup.$file_ext");
        }
        if ($Server->shell(cmd=>"ls /mroot/etc/log/cpdump") !~ /No such file/) {
            push (@valid_attachments,"cpdump.$file_ext");
        }
        if ($Server->shell(cmd=>"ls /mroot/etc/crash/parity_inconsistencies") !~ /No such file/) {
            push (@valid_attachments,"pi_data.$file_ext");
        }   
    }
    
    #check each saved asup dir for the required attachments
    foreach my $dir (@{$params{asup_dir}}) { 
        my $files=`ls $dir`;                 
        $Log->comment("_verify_attachments: FILES in asup $dir : \n $files");
        #check for each attachment
        my $event=`cat $dir/msg*.txt | $GREP_EXEC X-Netapp-asup-subject`;
        if ($event =~ /PERFORMANCE|performance/) {
            $Log->comment("_verify_attachments: Performance ASUP \n");
            if ($files !~ /cm_hourly_stats\.$file_ext/ || $files !~ /config_xml\.$file_ext/ || $files !~ /boottimes\.$file_ext/ || $files !~ /asup_stats\.xml\.$file_ext/) {
                $Log->warn("_verify_attachments: Failed to find an expected attachment in the saved PERFORMANCE asup directory $dir \n");
                return 0;
            }
        }
        elsif ($event =~ /MANAGEMENT_LOG|management/) {
            $Log->comment("_verify_attachments: Management/Daily ASUP \n");
            if ($files !~ /tar\.$file_ext/) {
                $Log->warn("_verify_attachments: Failed to find an expected attachment in the saved MANAGEMENT asup directory $dir \n");
                return 0;
            }
        }
        elsif ($event =~ /NHT|nht\.data/) {
            $Log->comment("_verify_attachments: NHT ASUP \n");
            if ($files !~ /sm2\.$file_ext/ || $files !~ /config_xml\.$file_ext/ || $files !~ /asup_stats\.xml\.$file_ext/) {
                $Log->warn("_verify_attachments: Failed to find an expected attachment in the saved NHT asup directory $dir \n");
                return 0;
            }
        }
        else {
            foreach my $basic_attachment (@valid_attachments) { 
                if ($files !~ /$basic_attachment/) {   
                    $Log->warn("_verify_attachments: Failed to find an expected attachment, $basic_attachment, in the saved asup directory $dir \n");
                    return 0;
                }
                $Log->debug(sub {"_verify_attachments: Validated $basic_attachment!!!"});
            }
        }
    }
    #all attachments were found in all asup dirs
    $Log->comment("_verify_attachments: Found all expected attachments \n");
    return 1;
}

###############################################################################
#Sub: verify_files_in_asup_dir
#Input Params: command target string
#Return values: string
#Summary: searches the ASUP directory for the given files
###############################################################################
sub verify_files_in_asup_dir {

    my ($dir, @Attach_array) = @_;

    my @newDir;
    my $array;
    $Log->comment("verify_files_in_asup_dir\n");

    if( ref ($dir) eq 'ARRAY' ){
        @newDir = @{$dir};
    } else {
       $Log->comment("Directory is expected to be an array\n");
        return 0;
   }

    ASUP::extract_archive(asup_dir=>@newDir);
    system("cd @newDir");

    foreach $array (@Attach_array) {
        my $tmp = `ls @newDir/$array`;
        my $tmp1 = "@newDir/$array";

        if ($tmp !~ /$tmp1/) {
           $Log->comment("File not found in asup dir: $array\n");
           return 0;
        }
   }
   return 1;
}

###############################################################################
#Sub: build_asup_content_hash
#Input Params:
#  hash - reference to hash to be filled in by routine
#  dir  - directory which holds all of the xml files for the ASUP
#Return values: none
#Summary: Populates a hash, keyed by filename, of the content of all of the
#         XML files in an asup directory
###############################################################################
sub build_asup_content_hash {

    my $whoami       = (caller(0))[3];
    my %params = (
            hash=>undef,
            dir=>undef,
            @_,
    );

    my $hash = $params{hash};
    my $dir  = $params{dir};

    if ( !$hash || (ref($hash) !~ /HASH/i) ) {
        die script_result("$whoami must be called with hash option and hash must be a hash reference");
    }

    if ( !$dir || !(-d $dir ) ) {
        die script_result("$whoami must be called with dir option, and it must be a directory path");
    }

    my $parser = XML::LibXML->new();

    # JDS cm_stats files are huge and overwhelm the test when trying to parse through it
    #my $files=`ls $dir/*.xml`;
    my $files=`ls $dir/*.xml | $GREP_EXEC -v cm_stats`;
    my @files = split(/\s+/, $files);

    foreach my $file ( @files ) {
        my $filename = fileparse( $file );
        my %h;
        eval {
            my $xmlDoc = $parser->parse_file( $file );
            my @children = $xmlDoc->childNodes();
            DataUtils::build_hash_from_xml( xmlnode => $children[1], hash => \%h );
        };
        if ( $@ ) {

            my $manifest = get_autosupport_manifest();
            my $is_smf_table=0;
            foreach my $item ( @$manifest ) {
                if ( $item->{'body-file'} eq $filename ) {
                    if ( $item->{'cmd-tgt'} eq 'smf_table' ) {
                        $is_smf_table=1;
                    }
                    last;
                }
            }

            if ( $is_smf_table ) {
                $Log->error("failed parsing $file: $@");
            } else {
                $Log->comment("$file: is not formatted as an smf_table, cannot parse");
            }
        } else {
            $hash->{$filename} = Clone::clone(\%h);
        }
    }
}

###############################################################################
#Sub: compare_full_and_compliant_asups
#Input Params:
#  compliant       - hash ref (built by build_asup_content_hash) for a compliant ASUP
#  full            - hash ref (built by build_asup_content_hash) for a full ASUP
#  hash_map_array  - result of "debug smdb table autosupport_hash show"
#                    Note: This must be passed in as the table is destroyed when
#                          ASUP is not configured as "compliant" (i.e. either not
#                          configured, or configured as "full".
#Return values: string
#Summary: Compares a compliant ASUP with a full (complete) ASUP and verifies that
#         data hashed out for compliance reasons matches what is in the full ASUP
###############################################################################
sub compare_full_and_compliant_asups {
    my $whoami       = (caller(0))[3];
    my %params = (
            compliant=>undef,
            full=>undef,
            hash_map_array=>undef,
            @_,
    );

    my $compliant  = $params{compliant};
    my $full       = $params{full};
    my $asup_hash  = $params{hash_map_array};
    my $num_errors = 0;

    if ( !$compliant || (ref($compliant) !~ /HASH/i) ) {
        die script_result("$whoami must be called with compliant option and compliant must be a hash reference");
    }

    if ( !$full || (ref($full) !~ /HASH/i) ) {
        die script_result("$whoami must be called with full option and full must be a hash reference");
    }

    if ( !$asup_hash || (ref($asup_hash) !~ /ARRAY/i) ) {
        die script_result("$whoami must be called with hash_map_array option and full must be a reference to an array of hashes");
    }

    # Get the compliance variable mapping table
    my $asup_remap = get_compliance_xml_mapping();

    # Proc validate compliant ASUP data
    foreach my $filename ( sort keys %{$compliant} ) {
        my @errors;
        my $successes;
        my $hp = $compliant->{$filename};
        foreach my $table ( sort keys %{$hp} ) {
            my $tp = $hp->{$table};
            my $smf_table_name = $tp->{smf_table_name};

            foreach my $asup_def ( sort keys %{$tp->{'asup:TABLE_INFO'}} ) {
                my $remap          = $tp->{'asup:TABLE_INFO'}->{$asup_def}->{'asup:remap'};
                my $xml_field_name = $tp->{'asup:TABLE_INFO'}->{$asup_def}->{'asup:tag'};
                my $smf_field_name = $tp->{'asup:TABLE_INFO'}->{$asup_def}->{'asup:smf_name'};

                # Verify that remap in the XML file matches the remap on the filer
                # find remap from $asup_remap make sure it equals $remap
                my $remap_validated=0;
                foreach my $entry ( @{$asup_remap} ) {
                    if ( ($entry->{'smf-table-name'} eq $smf_table_name) &&
                         ($entry->{'smf-field-name'} eq $smf_field_name) &&
                         ($entry->{'xml-field-name'} eq $xml_field_name) &&
                         ($entry->{'xml-remap'}      eq $remap) ) {
                        $remap_validated=1;
                        last;
                    }
                }

                if ( !$remap_validated ) {
                    push( @errors, "  \nTable $smf_table_name smf_field $smf_field_name ".
                                   "xml_field $xml_field_name remap $remap not found ".
                                   "in asup_xml_fields table on the filer" );
                } elsif (!@errors) {
                    # It is not an error to have a table with no rows.
                    $successes++;
                }


                foreach my $row ( grep(/^asup:ROW/, (sort keys %{$tp} ) ) ) {   
                    my $rp = $tp->{$row};
                    foreach my $asup_var ( sort keys %{$rp} ) {
                        if ( $asup_var eq $xml_field_name ) {
                            my @list_values = {};  
                            # look up hash and replace it with original value
                            my $hash_substituted=0;
                            my $asuphash = $rp->{$asup_var};  
                            if ( $rp->{$asup_var} && ($remap eq 'hash')) {
                                $hash_substituted=1;   
                                $asuphash = $rp->{$asup_var};
                                my $hashRef = $rp->{$asup_var};   
                                if ("$hashRef" =~ m/HASH/) {   
                                     $Log->comment("asup_var = $asup_var, remap = $remap");
                                         foreach my $li_row (sort keys %{$hashRef->{'asup:list'}}) {    
                                                my $li_value= $hashRef->{'asup:list'}->{$li_row};  
                                                foreach my $entry ( @{$asup_hash} ) {  
                                                    if ( $entry->{hash} eq $li_value ) {
                                                        my $testEn = $entry->{hash};
                                                        $Log->comment("testEn = $testEn");
                                                         my $plaintext = $entry->{plaintext};
                                                         my @words = split( /\s+/, $plaintext );
                                                         if ( scalar( @words ) > 1 ) {
                                                            $plaintext =~ s/^\"//g;   
                                                            $plaintext =~ s/\"$//g;   
                                                          }
                                                          $plaintext =~ s/&amp;/&/g;
                                                          $plaintext =~ s/&lt;/</g;
                                                          $plaintext =~ s/&gt;/>/g;
                                                          $plaintext =~ s/&quot;/"/g;
                                                           push(@list_values, $plaintext);
                                                           $hash_substituted=2;
                                                           last;
                                                     }
                                                }
                                         }
                               } else {
                                        foreach my $entry ( @{$asup_hash} ) {   
                                            if ( $entry->{hash} eq $asuphash ) {
                                                # If there are multiple words delimited by spaces,
                                                # there will be "" around them in the table, but not
                                                # in the structure from the ASUP so remove the " so the
                                                # comparison matches
                                                my $plaintext = $entry->{plaintext};
                                                my @words = split( /\s+/, $plaintext );
#                                                if ( scalar( @words ) > 1 ) {
                                                    $plaintext =~ s/^\"//g;   #remove "
                                                    $plaintext =~ s/\"$//g;   #remove "$
#                                                }
        
                                                # LibXML will convert these as we read them from the
                                                # XML files, but in the autosupport_hash table, they are
                                                # still encoded.
                                                $plaintext =~ s/&amp;/&/g;
                                                $plaintext =~ s/&lt;/</g;
                                                $plaintext =~ s/&gt;/>/g;
                                                $plaintext =~ s/&quot;/"/g;
                                                
                                                $rp->{$asup_var} = $plaintext;
                                                $asuphash = $plaintext;
                                                $hash_substituted=2;
                                                last;
                                            }
                                        }
                                 }
     
                                if ( $hash_substituted == 1 ) {
                                    push( @errors, "  \nMAPPING MISSING: Table $smf_table_name smf_field $smf_field_name ".
                                                   "xml_field $xml_field_name remap $remap not found ".
                                                   "in autosupport_hash table on the filer" );
                                }
                            }
                            # some fields contain a newline at the begin; see BURT 673745
                            if ( $full->{$filename}->{$table}->{$row}->{$asup_var} =~ /^\w.+\n\w*/ || $full->{$filename}->{$table}->{$row}->{$asup_var} =~ /\n\w.+\w*/ ){
                                $full->{$filename}->{$table}->{$row}->{$asup_var} =~ s/\n/ /g;
                                $Log->comment("Asup value after trimming of new lines, $full->{$filename}->{$table}->{$row}->{$asup_var}");
                            }
                            # Some fields contain a newline at the between; see BURT 673745
                            if ( $full->{$filename}->{$table}->{$row}->{$asup_var} =~ /\w.+\w*\n\w.+\w*/){
                                $full->{$filename}->{$table}->{$row}->{$asup_var} =~ s/\n/ /g;
                                $Log->comment("Asup value after trimming of new lines, $full->{$filename}->{$table}->{$row}->{$asup_var}");
                            }
                            my $full_asup_var = $full->{$filename}->{$table}->{$row}->{$asup_var};
                            if ("$full_asup_var" =~ m/HASH/ && ($remap eq 'hash')) {
                                 my @full_values = {};
                                  foreach my $li_row (sort keys %{$full_asup_var->{'asup:list'}}) {    
                                       my $li_value= $full_asup_var->{'asup:list'}->{$li_row};  
                                       push(@full_values, $li_value);
                                  }
                                  my $full_size = @full_values;
                                  my $compliant_size = @list_values;
                                  if ($full_size != $compliant_size) {
                                    $Log->error("in $filename $table $row $asup_var, full has $full_size items  while compliant has $compliant_size item,$full_asup_var full_values= $full_values[1] compliant value = $list_values[1]");
                                    $num_errors++;
                                  } else {
                                     for (my $i=1; $i<=$full_size; $i++) {
                                        my $fv = $full_values[$i];
                                        my $lv =  $list_values[$i];
                                        if ("$fv" ne "$lv") {
                                             push( @errors, "  \nCONTENT MISMATCH: Table $smf_table_name smf_field $smf_field_name ".
                                               "xml_field $xml_field_name remap $remap = $lv in compliant HASH ".
                                               "ASUP but should be $fv and i = $i");
                                        } elsif (!@errors) {
                                            $successes++;
                                        }
                                     }
                                  }
                            } else {
                                    if ( ($remap eq 'hash') &&
                                         ("$full_asup_var" ne "$asuphash") ) {
                                        push( @errors, "  \nCONTENT MISMATCH: Table $smf_table_name smf_field $smf_field_name ".
                                                       "xml_field $xml_field_name remap $remap = |$asuphash| in compliant ".
                                                       "ASUP but should be |$full_asup_var|");
                                    } elsif (!@errors) {
                                        $successes++;
                                    }
                            }
                        }
                    }
                }
            }
        }

        if ( @errors ) {
            $Log->error("Errors for File $filename @errors");
            $num_errors++;
        } elsif ( !$successes ) {
            $Log->error("No results for file $filename");
        }
    }
    return $num_errors;
}

###############################################################################
#Sub: _ngsh_7mode
#Input Params:
#  node       - node on which to run the command
#  cmd        - ngsh command to run
#  maxbufsize - maximum buffer size to specify for command with a large amount of
#               output
#Return values: array of hashes
###############################################################################
sub _ngsh_7mode {
    my $whoami       = (caller(0))[3];
    my %params = (
            node       => undef,
            cmd        => undef,
            maxbufsize => undef,
            @_,
    );

    my $node       = $params{node};
    my $cmd        = $params{cmd};
    my $maxbufsize = $params{maxbufsize};

    my $c;
    my $orig_maxbufsize;
    my @aoh;

    if ( !$node  ) {
        $node = $Server->get_execution_server();
    }

    if ( !$cmd ) {
        die script_result("$whoami must be called with cmd option");
    }

    # console connection is created on demand, so make sure it exists
    $Server->console( cmd => "priv set diag" );

    if ( $maxbufsize ) {
        $c = $Server->get_connection(type=>"console_connection");
        $orig_maxbufsize = $c->maxbufsize();
        $c->maxbufsize($maxbufsize);
    }

    eval {
        $Server->console( cmd => "systemshell",
                          noprompt => 1,
                          additional_prompts_timeout => 30,
                          additional_prompts => {
                              'login:'     => "diag",
                              'Password:'  => "netapp1!",
                              "$node%" => "FINAL_PROMPT",
                          },
                        );

        # Running through systemshell from 7mode is really slow...
        my $res =
        $Server->console( cmd => "ngsh -c \'set -showallfields true -showseparator \"X;X\" -privilege diagnostic; $cmd\'",
                          noprompt => 1,
                          additional_prompts_timeout => 300,
                          additional_prompts => {
                              "$node%"     => "FINAL_PROMPT",
                          },
                        );

        $Server->console( cmd => "exit",
                          noprompt => 1,
                          additional_prompts_timeout => 30,
                          additional_prompts => {
                              "$node%"     => "exit",
                              "logout"     => "FINAL_PROMPT",
                              "$node>"     => "FINAL_PROMPT",
                          },
                        );

        # Sometimes we end up with the ngsh command and a blank line included in $res
        # Strip those out if they exist
        my @res = split(/\n/, $res );
        if ( $res[0] =~ /^\s*ngsh -c/ ) {
            shift( @res );
        }

        if ( $res[0] =~ /^\s*$/ ) {
            shift( @res );
        }

        $res = join( "\n", @res );
        
        # This is actually NOT a class method (it doesn't use the $self pointer)
        # so we should be ok running this.
        Node::NG::_cli_output_handler( $Server, output_in=>$res, want_array=>\@aoh );
    };

    if ( $@ ) {
        $Log->error("Attempt to run \"$cmd\" on 7mode filer failed with $@");
    }

    if ( $maxbufsize ) {
        $c->maxbufsize($orig_maxbufsize);
    }

    return \@aoh;
}

###############################################################################
#Sub: get_compliance_hash_table
#Input Params:
#  node       - node on which to run the command
#  maxbufsize - maximum buffer size to specify for command with a large amount of
#               output
#Return values: array of hashes
###############################################################################
sub get_compliance_hash_table {
    my %params = (
            node       => undef,
            maxbufsize => 10000000,
            @_,
    );

    my $node        = $params{node};
    my $maxbufsize  = $params{maxbufsize};

    if ( !$node  ) {
        $node = $Server->get_execution_server();
    }

    my $version=_get_version();
    my $asup_hash;

    if ( $version =~ /7M/ ) {
        $asup_hash = _ngsh_7mode( node => $node,
                                 maxbufsize => $maxbufsize,
                                 cmd => 'debug smdb table autosupport_hash show' );
    } else {
        my $c = $Server->get_connection(type=>"cli_connection");
        my $orig_maxbufsize = $c->maxbufsize();
        $c->maxbufsize( $maxbufsize );

        # So it doesn't match against data in the table
        my @orig_prompts = $c->prompts('^\\S+::>\\s','^\\S+::\\*>\\s');

        $asup_hash = $Server->debug_smdb_table_autosupport_hash_show();  #result of this command line 

        # Restore the prompts
        $c->prompts( @orig_prompts );
        $c->maxbufsize( $orig_maxbufsize );
    }
  
    return $asup_hash;
}

###############################################################################
#Sub: get_compliance_xml_mapping
#Input Params:
#  node       - node on which to run the command
#  maxbufsize - maximum buffer size to specify for command with a large amount of
#               output
#Return values: array of hashes
###############################################################################
sub get_compliance_xml_mapping {
    my %params = (
            node       => undef,
            maxbufsize => 10000000,
            @_,
    );

    my $node        = $params{node};
    my $maxbufsize  = $params{maxbufsize};

    if ( !$node  ) {
        $node = $Server->get_execution_server();
    }

    my $version=_get_version();
    my $asup_hash;

    if ( $version =~ /7M/ ) {
        $asup_hash = _ngsh_7mode( node => $node,
                                 maxbufsize => $maxbufsize,
                                 cmd => 'debug smdb table asup_xml_fields show' );
    } else {
        my $c = $Server->get_connection(type=>"cli_connection");
        my $orig_maxbufsize = $c->maxbufsize();
        $c->maxbufsize( $maxbufsize );

        # So it doesn't match against data in the table
        my @orig_prompts = $c->prompts('^\\S+::>\\s','^\\S+::\\*>\\s');

        $asup_hash = $Server->debug_smdb_table_asup_xml_fields_show();

        # Restore the prompts
        $c->prompts( @orig_prompts );
        $c->maxbufsize( $orig_maxbufsize );
    }

    return $asup_hash;
}

###############################################################################
#Sub: retrieve_client_ipv6_address
#Input Params: Hash
#Return values: address (success) or "" (failure)
#Summary: Returns the global scope IPv6 address of a client
###############################################################################
sub retrieve_client_ipv6_address {
    my %params = (
            client=>undef, #server name/ip address
            @_,
    );

    my $addr = "";
    
    if (!defined $params{client}) {
        die script_result("ASUP::retrieve_client_ipv6_address - A client is required... \n");
    }

    my $ipv6_addr;
    my $host = Tharn::host( $params{client} );

    # First try to extract this from the nate host file
    $ipv6_addr = $host->net_best_address(include=>{family=>"inet6", nettype=>"data"});

    if ( $ipv6_addr ) {
        return $ipv6_addr;
    }

    # Ok, brute-force, try to extract from ifconfig on the client
    my $conn = Tharn::connect('#rsh');

    if ( $host->{hosttype} =~ /win/ ) {
        my $cmd = 'ipconfig';

        my $sentinel = 'DoNe_' . time();
        $cmd = $cmd . "&&echo $sentinel";

        my ($match, $output, $after) =  &Tharn::say($conn, $cmd, $sentinel,
                                                 '-swallow',
                                                 '-timeout', 20,
                                                 '-ignore', "echo $sentinel",
                                                 '-comment', "Attempted '$cmd'");

        # strip output and remove CR/LF
        chomp($output);
        $output =~ s/\r//g;

        # strip sentinel from output
        $output =~ s/\n$sentinel.*//;

        my @output = split(/\n/, $output);

        # are we sure it will always be formatted this way?
        foreach my $line ( @output ) {
            if ( $line =~ /IP Address[\.|\s]+:\s+([a-f|0-9|:]+)/ ) {
                $addr = $1;
                if ( $addr =~ /:/ ) {
                    last;
                }
                $addr = undef;
            }
        }
    } else {
         
        my $cmd = "host $params{client}" ;

        my $sentinel = 'DoNe_' . time();
        $cmd = $cmd . ";echo \"rstatus=\$\?\" ;echo $sentinel";

        my ($match, $output, $after) =  &Tharn::say($conn, $cmd, $sentinel,
                                                 '-swallow',
                                                 '-timeout', 20,
                                                 '-ignore', "echo $sentinel",
                                                 '-comment', "Attempted '$cmd'");

        # strip output and remove CR/LF
        chomp($output);
        $output =~ s/\r//g;

        # strip sentinel and status from output
        $output =~ s/rstatus=(\d+)//g;
        
        if($1 == 0 && $output =~ /IPv6/ ){
       		$output =~ s/\n$sentinel.*//;

		$Log->comment("host output :" . $output);
                if ( $output =~ /IPv6 address\s+([a-f|0-9|:]+)/ ){
			$addr=$1;
                }
        } else {
        	# are we sure it will always be here?
	        my $cmd = '/sbin/ifconfig -a';

        	my $sentinel = 'DoNe_' . time();
	        $cmd = $cmd . ";echo \"rstatus=\$\?\" ;echo $sentinel";

	        my ($match, $output, $after) =  &Tharn::say($conn, $cmd, $sentinel,
        	                                         '-swallow',
                	                	         '-timeout', 20,
                        	                         '-ignore', "echo $sentinel",
                                	                 '-comment', "Attempted '$cmd'");

        	#strip output and remove CR/LF
        	chomp($output);
                $output =~ s/\r//g;

        	# strip sentinel and status from output
        	$output =~ s/"rstatus=\$\?" ;//g;
        	$output =~ s/\n$sentinel.*//;

        	my @output = split(/\n/, $output);

	        # are we sure it will always be formatted this way?
        	foreach my $line ( @output ) {
        	    if ( $line =~ /inet6 addr:\s+([a-f|0-9|:]+)\/\d+\s+Scope:Global/ ) {
	                $addr = $1;
        	        last;
        	    }
        	}
      }
    }

    return $addr;
}

###############################################################################
#Sub: config_ipv6
#Input Params: Hash
#Return values: 1 (success), 0 (failure) or -1 (unsupported)
#Summary: Configure IPv6 on the filer
###############################################################################
sub config_ipv6 {
    my %params = (
            node       => undef,
            enable     => "off",
            @_,
    );

    my $rc = 0;
    my $node        = $params{node};

    if ( !$node  ) {
        $node = $Server->get_execution_server();
    }

    my $version=_get_version();

    my $build=lc($Server->get_release(format=>"cooked"));
    my $state;
    if ( $version =~ /7M/  ) {
        $state = "off";
        if ( $params{enable} =~ /1|on|true|yes|enable/i ) {
            $state = "on";
        }
        eval {
            $Server->cli(cmd => "priv set diag; options ip.v6.enable ".$state);
        };

        if ( !$@ ) {
            $rc = 1;
        }
    } else {
        $state = "false";
        if ( $params{enable} =~ /1|on|true|yes|enable/i ) {
            $state = "true";
        }
        eval {
            #$Server->network_options_ipv6_modify(enabled=>$state);
            $Server->cli(cmd => "network options ipv6 modify -enabled " . $state);
        };

        if ( !$@ ) {
            $rc = 1;
        }
    }

    if ( $rc < 0 ) {
        $Log->comment("Node $node is version $build which does not support IPv6");
    }

    return $rc;
}

###############################################################################
#Sub: validate_xml
#Input Params: Hash
#Return values: 0 (failure) or 1 (success - valid xml file/string)
#Summary: Validate whether or not an xml file is valid (has all the closing
#       tags and such) and return 1 if it's successful.
###############################################################################
sub validate_xml {

    #store the hash as options
    my %options = ( @_, );
    my $result = 1;

    #use the file first, if it exists
    if($options{xmlfile}) {

        my $parser = XML::LibXML->new;
        # validate the file
        eval{ $parser->parse_file($options{xmlfile})};

        # Only fail the function if the xml tags are problems
        if($@) {
            $result = 0;
            $Log->comment("XML Tag Problem: $@");
        }
        else {
            #validate against DTD (if exists)
            $parser->validation(1);
            eval{ $parser->parse_file($options{xmlfile})};

            # Issue warning of DTD related problems
            if($@) {
                $Log->comment("XML DTD Problem: $@");
            }
        }

    }
    #Second use xmlstring if it's passed
    elsif($options{xmlstring}) {
        my $parser = XML::LibXML->new;
        eval{ $parser->parse_string($options{xmlstring})};
        if($@) {
            $result = 0;
            $Log->warn("XML Warning: $@");
        }
    }
    #Fail if nothing is passed
    else {
         $Log->comment("Neither xml string nor file passed.");
         $result = 0;
    }

    return $result;
}

###############################################################################
#Sub: add_hosts_entry
#Input Params: client, ipaddr
#Return values: 0 (failure) or 1 (success)
#Summary: add a static host entry to hosts for 7m, to dns hosts for 10m
###############################################################################
sub add_hosts_entry {
    my %params = (
        client => undef,
        ipaddr => undef,
        @_,
    );

    my $node;
    my $version = _get_version();
    if ( $version =~ /7M/ ) {
        my $entry = "$params{ipaddr} $params{client}";
        $node = $Server->get_execution_server();
        my $host        = host("$node");
        my $version     = _get_version();
        my $Api_Set_Obj = NACL::APISet->new(
            hostobj   => $host,
            category  => "Node",
            interface => "CLI",
            set       => "Systemshell",

        );
        $Log->comment("Adding to an entry to /mroot/etc/hosts file: $entry");
        $Api_Set_Obj->execute_raw_command(
            command => "echo $entry >> /mroot/etc/hosts" );
    } else {    #10M
        $node = NACL::C::Node->find();
        my $vs = NACL::C::Vserver->find(
            command_interface => $node,
            filter            => { type => "admin" }
        );

        $Log->comment(
            "Adding to an entry to dns hosts: vserver:$vs, address:$params{ipaddr}, hostname:$params{client}"
        );
        NACL::C::VserverServicesDnsHosts->create(
            command_interface => $node,
            vserver           => $vs,
            address           => $params{ipaddr},
            hostname          => $params{client}
        );
    }

    return 1;

} ## end sub add_hosts_entry

###############################################################################
#Sub: delete_hosts_entry
#Input Params: ipaddr
#Return values: 0 (failure) or 1 (success)
#Summary: delete a static host entry to hosts for 7m, to dns hosts for 10m
###############################################################################
sub delete_hosts_entry {
    my %params = (
        ipaddr => undef,
        @_,
    );

    my $node;
    my $version = _get_version();
    if ( $version =~ /7M/ ) {
        $node = $Server->get_execution_server();
        my $host        = host("$node");
        my $Api_Set_Obj = NACL::APISet->new(
            hostobj   => $host,
            category  => "Node",
            interface => "CLI",
            set       => "Systemshell",
        );
        my $entry = "/$params{ipaddr}/d";
        $Log->comment(
            "Deleting to an entry from /mroot/etc/hosts file: $entry");
        $Api_Set_Obj->execute_raw_command( command =>
                "sed $entry < /mroot/etc/hosts > /mroot/etc/hostsbak; mv /mroot/etc/hostsbak /mroot/etc/hosts "
        );

    } else {    #10M
        $node = NACL::C::Node->find();
        my $vs = NACL::C::Vserver->find(
            command_interface => $node,
            filter            => { type => "admin" }
        );

        $Log->comment(
            "Deleting an entry from dns hosts: vserver:$vs, address:$params{ipaddr}"
        );
        NACL::C::VserverServicesDnsHosts->delete(
            command_interface => $node,
            vserver           => $vs,
            address           => $params{ipaddr},
        );
    }
    return 1;
} ## end sub delete_hosts_entry

###############################################################################
#Sub: Routine to run a unix command while handling error codes better than backticks
#Return values: ($rc, $cmd_output)
###############################################################################
sub shell_unix_cmd {
    my $me = (caller(0))[3];

    my %opt = validate( @_,
                        { 
                            cmd             => { type     => SCALAR, },
                        },
                      );

    my $cmd             = $opt{cmd};

    # Redirect STDERR to STDOUT
    if ( length($cmd) < 100 ) {
        $Log->comment("$me: $cmd 2>&1");
    } else {
        $Log->debug(sub {"$me: $cmd 2>&1"});
    }
    open( CMD_OUT, "$cmd 2>&1 |" ) || die "$me: during open on $cmd encountered error: $!";
    my @lines = <CMD_OUT>;
    close( CMD_OUT );
    my $rc = $?;

    # Assumption is that all unix commands return 0 for good
    if ( $rc ) {
        die "$me: $cmd returned non-zero return code: $rc";
    }

    chomp( @lines );
    @lines = map { $_ =~ s/\r//g; $_ } @lines;

    return ( wantarray ? @lines : join(" ", @lines) );
}

###############################################################################
#Sub: wait_until_delivered
#Return values: history entry (success), die (failure)
#Summary: Wait until expected ASUP finds its way into the asup history table
###############################################################################
sub wait_until_delivered {
    my $me = (caller(0))[3];

    my %params = (
        status         => 'sent-successful',
        timeout        => 300,
        interval_sleep => 10,
        die_on_error   => 1,
        @_,
    );
    my $asup_history;

    # Make sure you delete any params that ASUP::get_autosupport_history() doesn't understand
    my $timeout       = delete $params{timeout};
    my $sleep         = delete $params{interval_sleep};
    my $die_on_error  = delete $params{die_on_error};

    if ( !defined( $params{destination} ) ) {
        $Log->warn("$me: No destination provided");
    }

    my $t1 = time();
    my $t2 = $t1 + $timeout;

    while ( $t1 < $t2 ) {

        Tharn::sleep($sleep);

        $asup_history=ASUP::get_autosupport_history(%params);

        if (defined($asup_history)) {
            last;
        }

        $t1 = time();
    }

    if ( !defined($asup_history) && $die_on_error ) {
        my %params_no_status = (%params);
        if ( exists $params_no_status{status} ) {
            delete $params_no_status{status};
            $Log->warn("$me: ASUP::get_autosupport_history() without specifying the status");
            ASUP::get_autosupport_history(%params_no_status);
        }
        die "$me: Could not find ASUP::get_autosupport_history() given ".Dumper(\%params);
    }

    return $asup_history;
}

###############################################################################
#Sub: Routine to validate the MIME boundary and BASE64 encoding of an SMTP email file
#Return ($mime_boundary, $base64_encoding);
###############################################################################
sub validate_email_file {
    my $me = (caller(0))[3];

    my %opt = validate( @_,
                        { 
                            file => { type  => SCALAR, },
                        },
                      );

    my $email_file = $opt{file};
    my $mime_boundary;
    my $base64_encoding;
    my @lines;
    if ( !open( FILE, '<', $email_file) ) {
        $Log->warn("$me: failed to open $email_file with $!");
    } else {
        @lines = <FILE>;
        close(FILE);

        @lines = (reverse @lines);
        my $non_blank_line_counter = 0;
        my $mime_boundary_position = 0;
        foreach my $line ( @lines ) {
            chomp($line);
            $line =~ s/\r//g; # Remove trailing carriage returns
            if ( $line ) {
                $non_blank_line_counter++;

                # We expect the MIME boundary to be the last non-blank line and the 
                # BASE64 encoding line to be the line before that
                if ( $line eq '--netapp mime boundary--' ) {
                    $mime_boundary = $line;
                    $mime_boundary_position = $non_blank_line_counter;
                    $Log->comment("$me: Found expected MIME boundary ($mime_boundary)");
                } elsif ( $mime_boundary_position &&
                          ($non_blank_line_counter == $mime_boundary_position + 1) ) {
                    my $length = length($line);
                    if (!$length || (($length%4)!=0)) {
                       $Log->warn("$me: BASE64($line), length($length) is not a multiple of 4!");
                       last;
                    } else {
                       $base64_encoding = $line;
                       $Log->comment("$me: BASE64($line), length($length) is a multiple of 4 as expected");
                       last;
                    }
                } else {
                   $Log->warn("$me: Email file $email_file appears to be misformatted, expected last non-blank line to be --netapp mime boundary-- and the line immediately before that to have a length that is divisible by 4");
                   last;
                }
            }
        }
    }

    return ($mime_boundary, $base64_encoding);
}


=pod

=head1 NAME

ASUP.pm - ASUP configuration, generation, and verification library for 7mode and  Cmode SN.


=head1 SYNOPSIS

    use ASUP;
    ASUP::invoke_and_verify_asup(
            event => "callhome.bad.ram",
            asup_save_dir => $LOGDIR,
            client => $CLIENT,
            transport => "HTTP"
            );


=head1 DESCRIPTION

The methods included in this library were developed in order to assist
engineers in generation and verification of asup events.  The current library
supports 7M and Cmode filers and invokes communication to the filer via the
Server.pm library.  The ASUP.pm library allows the user to configure
autosupport, generate autosupports (both via events and manually), and
verify that the autosupport was received by a SMTP, HTTP, or HTTPS server and
validate the autosupport format (attachments, sections, headers).  If an
asup_save_dir is specified, the autosupport received by the transport server
will be saved off to the indicated directory and the user will be able to
directly access the autosupport data and attachments, for content validation.
  
    
=head1 METHODS

    config_asup();      
        Description:
            This method will allow you to configure the autosupport options.
        Inputs:
            change_filext => <on|off> Change the attachment ext to .gx
            cifs_verbose => <on|off> Turn on cifs verbos.  7M only.
            content => <minimal|complete> ASUP content type.  7M only.
            enable => 7M<on|off> | CM<enable|disable> Enable or disable
                autosupport as a whole.
            from => <mail address> Where the asup should show as being sent from.
            local_nht_data_enable => 7M<on|off> | CM<true|false> Should nhts
                be sent to local addresses.
            local_performance_data_enable => 7M<on|off> | CM<true|false>
                Should performance data asups be sent to local addresses.
            minimal_subject_id => <systemid|hostname> Should the subject line
                contain the system id or hostname. 7M only.
            mailhost => <mailhost name or ip> Hostname of the mail server.
            nht_data_enable => 7M<on|off> | CM<true|false> Enable NHT asups.
            noteto => <mail address> Email addresses to receive summary
                autosupports.
            partner_to => <mail address> Email addresses to be echoed asups sent
                to NetApp support.
            performance_data_enable => 7M<on|off> | CM<true|false> Enable
                performance data asups.
            reminder => 10M<true|false> The asup reminder
            retry_count => <integer> The number of times to attempt to resend an
                asup.
            retry_interval => <integer> The delay interval between asup retries.
            support_enable => 7M<on|off> | CM<enable|disable> Should asups be
                sent to NetApp support.
            support_proxy => <url> Proxy server configuration.
            support_to => <mail address> Email address of NetApp.
            support_transport => <smtp|http|https> Transport to send the asups
                to NetApp.
            support_url => <url> URL of NetApp server.
            throttle => 7M<on|off> | CM<true|false> Should asups be
                throttled/suppressed.
            to => <mail address> Email addresses of local individuals to
                receive asups.
        Outputs:
            1 on successful configuration or error
        Examples:
            see full example below
            
             
    get_asup_config();
        Description:
            This method will get the current asup configuration on the filer.
        Inputs:
            <none>
        Outputs:
            $hash of the asup configuration
        Examples:
            see full example below
            
    get_asup_destinations();
        Description:
            This method will return the current autosupport destinations
                configured on the filer.  CM only.
        Inputs:
            <none>
        Outputs:
            @array of hashes with the current asup destinations.
        Examples:
            see full example below
            
    invoke_and_verify_asup(client=>CLIENT);
        Description:
            This convenience method will configure asup on a filer, invoke an
                autosupport, verify the autosupport is received by the http,
                https, or smtp server, then if a save asup dir was specified
                it will verfiy the autosupport format (headers, sections, and
                attaachments).
        Inputs:
            event => <callhome.*> Specific event to invoke.  If left blank, will
                manually invoke an autosupport.
            asup_save_dir => <local dir> The local directory to save asups to.
                If specified, this will also all cause the verify headers,
                sections, and attachments routines to be run.
            client => <CLIENT> The client to run the http, smtp, or https server
            transport => <HTTPS|SMTP|HTTP> Protocol to use for asup transport
        Outputs:
            0|1 - O on failure of asup reception/or verification, 1 on success
                of asup reception (and asup format verification).
        Examples:
            see full example below
           
    verify_autosupport_reception(events=>EVENT_OR_EVENT_ARRAY,
            server=>TRANSPORT_SERVER);
        Description:
            This method will attempt to look for a received autosupport based
                on the requested events and http|https|smtp server.  It first
                looks for the event subject line, if that is not pupulated,
                then it will try to match based on the event name.
        Inputs:
            events => <hash>|<array of hash> The event or events to look for.
                The hash must consist of the event name and the filer where the
                event was invoked.
            server => <pointer to server object> This is the actual http|https|
                smtp server instance.
            time_limit => <time in sec> Max time to wait for an asup
        Outputs:
            0|1|array_of_dirs Returns 0 if we were unsuccessful in finding all
                of the requested asups.  Returns 1, if we found all of the
                expected asups, but did not save them to disk.  Return an
                array of the directories, if we found all of the expected asups
                and they were all saved to the asup save dir location.
        Examples:
            see full example below
            
    verify_autosupport_collection(events=>EVNENT_OR_EVENT_ARRAY);
        Description:
            This method will attempt to look for a collected autosupport based
                on the requested events.  It will use the autosupport history
                feature to look for the events by trigger name, subject, and
                time-frame.
        Inputs:
            events => <hash>|<array of hash> The event or events to look for.
                The hash must consist of the event name and the filer where the
                event was invoked.
            last_update_time=> <time> Absolute or relative time, how far back
                in the history table should we look for this asup.  Default
                 looks for the asup only within the past 10mins (>10m).
            destination=> <http|smtp|noteto|retransmit> Destination to validate
                collection for.  Default is http.
            time_limit => <time in sec> Max time to wait for an asup.  Default
                is 3 mins.
        Outputs:
            0|1 Returns 0 if we were unsuccessful in finding all
                of the requested asups.  Returns 1, if we found all of the
                expected asups.
        Examples:
            see full example below
        
    invoke_autosupport();
        Description:
            This method will invoke an autosupport on a filer, either an event
                generated asup or a manually invoked asup.
        Inputs:
            message => <text> The message for a manually invoked asup.  This is
                the default if no message/event is specified.  The asup will
                be invoked with the message Manually_Invoked_Autosupport_<time>.
            event => <callhome.*> The callhome event to invoke.
            subject => <text> The subject for the autosupport.  Defaults to the
                event name if not specified.
        Outputs:
            0|event_hash On failure to invoke, it returns a 0.  Else it returns
                the event in hash that includes the event, the  node, and the
                sysid, if available.
        Examples:
            see full example below
            
    invoke_nht_autosupport();
        Description:
            This method will invoke a NHT autosupport on a filer
        Inputs:
            <none>
        Outputs:
            0|event_hash On failure to invoke, it returns a 0.  Else it returns
                the event in hash that includes the event, the  node, and the
                sysid, if available.
        Examples:
            see full example below
            
    invoke_performance_autosupport();
        Description:
            This method will invoke a Performance autosupport on a filer
        Inputs:
            <none>
        Outputs:
            0|event_hash On failure to invoke, it returns a 0.  Else it returns
                the event in hash that includes the event, the  node, and the
                sysid, if available.
        Examples:
            see full example below
         
    invoke_weekly_autosupport();
        Description:
            This method will invoke a Weekly autosupport on a filer
        Inputs:
            <none>
        Outputs:
            0|event_hash On failure to invoke, it returns a 0.  Else it returns
                the event in hash that includes the event, the  node, and the
                sysid, if available.
        Examples:
            see full example below
            
    start_transport_server(client=>CLIENT);
        Description:
            This method will manually start a http|https|smtp server to listen
                for autosupports.
        Inputs:
            client => <CLIENT> The client to invoke the server on.
            asup_dir => <local dir> The local directory to save autosupports
                to.  If not specified, we won't save the received autosupports,
                but we can still verify reception.  We just can't verify the
                autosupport format.
            transport => <HTTP|HTTPS|SMTP> Server type to invoke.  Default is
                HTTPS.
            port => <int> Port to use for server.
        Outputs:
            transport_object Returns a ref to the transport server object.
        Examples:
            see full example below
            
    stop_transport_server(server=>SERVER);
        Description:
            This method will manually stop a http|https|smtp server that is
                listening for autosupports.
        Inputs:
            server => <SERVER> The current https|http|smtp server object to stop
        Outputs:
            0|1 Returns 0 on failure or 1 on success.
        Examples:
            see full example below
            
    verify_autosupport_format(asup_dir => LOCAL_ASUP_DIR);
        Description:
            This method will allow the user to perform format verification on
                a received and saved autosupport.  The verification will be
                automatically done based upon the mode, content type, and
                autosupport message received.  Autosupport format can be
                verified for headers, sections, attachments, or all.  All is
                the default.
        Inputs:
            asup_dir => <local dir> The local directory where the received
                autosupports are saved.
            alt_extension => <0|1> The default is 0, meaning we use .gz for all
                attachments.  If set to 1, we will attempt to validate the
                attachments with .gx extensions.
            minimal_content => <0|1> The default is 0, meaning we should attempt
                to validate the content as complete.  If set to 1, we will only
                validate a minimal content type - available in 7M only.
            validate => <sections|headers|attachments|all> What to validate.
        Outputs:
            0|1 Returns 1 on successful validation and 0 on any failure.  See
                the logs for more details on a failure.
        Examples:
            see full example below
            
    retrieve_archived_asup_dirs();
        Description:
            This method will return a listing of all of the saved asup files
                on a node.
        Inputs:
            <none>
        Outputs:
            @asupfiles Returns a ref to an array of all of the saved asups
                on the filer.
        Examples:
            see full example below
            
    verify_asup_archived(event=>EVENT);
        Description:
            This method will probe the saved asup directories on the filer and
                check to see if the requested event has been saved correctly.
        Inputs:
            event => <hash> The callhome event or asup message to look for.
        Outputs:
            0|asup_save_dir Returns the asup save dir if the asup was
                successfully archived, or 0 if it was not found in the archive.
        Examples:
            see full example below
            
    get_autosupport_filter());
        Description:
            This method will return the current filter configuration for
                autosupports.  Autosupports can be customized to send to the
                to and noteto fields via this option.
        Inputs:
            node=>"local"  The node to filer for, default is current.
            'autosupport-message'=>'*'  The message to filter for, default is
                all.
        Outputs:
            @array of hash Pointer to array of hash with filter settings.
        Examples:
            see full example below
            
    modify_autosupport_filter('autosupport-message'=>MESSAGE);
        Description:
            This method will modify the current filter configuration options
                for an autosupport.
        Inputs:
            node=>"local"  The node on which the message should be modified.
                The default is the current node.
            'autosupport-message'=>undef  The message to modify.
            to=>"enabled"  Whether this asup message should be sent to users in
                the to field.  The default is yes (enabled).
            noteto=>"enabled"  Whether this asup message should be sent to users
                in the noteto field.  The default is yes (enabled).
        Outputs:
            0|1  0 on failure, 1 on success
        Examples:
            see full example below
    
    retrieve_client_ipv6_address( client => CLIENT );
        Description:
            Retrieve the IPv6 address for a given client.
            Note: This is being added prior to having IPv6 addresses
                  coded into the nate hosts file, or having DNS resolution for
                  these addresses.  If the addresses are available in the 
                  nate hosts file, we will get them from there instead of 
                  querying ifconfig.
        Inputs:
            client=>$client  The name of the unix client for which you want to
                             retrieve the IPv6 address.
        Outputs:
            ipv6_address|""  "" on failure, ipv6_address on success
        Examples:
            see full example below

    config_ipv6( node => NODE, enable => <on|off> );
        Description:
            Enable or disable IPv6 on the filer.
        Inputs:
            node=>$node       The node on which to enable|disable IPv6.
            enable=><on|off>  Enable (on) or disable (off)
        Outputs:
            0|1  0 on failure, 1 on success
        Examples:
            see full example below

     validate_xml(xmlfile => FILENAME, xmlstring => XMLSTRING);
        Description:
            Validates an XML file or string.
        Inputs:
            xmlfile=>$filename The path/location with filename
            xmlstring=>$string The actual XML string
        Outputs:
            0|1  0 on failure, 1 on success
        Examples:
            see full example below


=head1 EXAMPLES

Typical example:

    use ASUP;
    use vars qw($LOGDIR $LOGDIR_UNIX $CLIENT);
    my $Server  = Server->new();
    my @Nodes  = $Server->servers();
    $Server->set_execution_server(-Server => $Nodes[0]);
    
    sub verify_http_protocol {
    $Test->description("Check http.");
    
    if (!ASUP::invoke_and_verify_asup(asup_save_dir=>$LOGDIR, client=>$CLIENT,
            transport=>"HTTP")) {
        return $TCD::FAIL;
    }
    return $TCD::PASS;



Ex1: Configure autosupport

    my $result=ASUP::config_asup(enable=>"on", support_enable=>"on",
        support_transport=>$params{transport}, support_url=>$CLIENT,
        mailhost=>"$CLIENT");

Ex2: Get asup configuration

    my $asup_configuration=ASUP::get_asup_config();

Ex3: Invoke an autosupport via "invoke"

    my $invoked_asup=ASUP::invoke_autosupport(message=>"My_Test_ASUP");

Ex4: Invoke an autosupport via an event

    my $invoked_asup=ASUP::invoke_autosupport(event=>"callhome.bad.ram",
        subject=>"Possible Bad RAM");

Ex5: Invoke and verify an asup in one motion.  Will validate format too...

    my $result=ASUP::invoke_and_verify_asup(asup_save_dir=>$LOGDIR,
        client=>$CLIENT, transport=>"SMTP");

Ex6: Manually invoke and verify an asup

    my $asup_server=ASUP::start_transport_server(asup_save_dir=>$LOGDIR,
        client=>$CLIENT, transport=>"SMTP");
    ASUP::config_asup(enable=>"on", support_enable=>"on",
        support_transport=>"SMTP", mailhost=>$CLIENT);
    my $asup=ASUP::invoke_autosupport(message=>"My_Test_ASUP");
    my $asup_dirs=ASUP::verify_autosupport_reception(events=>$asup,
        server=>$asup_server);
    ASUP::stop_transport_server(server=>$asup_server);
    if (!defined $asup_dirs or $asup_dirs==0) {
        return $TCD::FAIL;
    }
    if ($asup_dirs==1) {
        $Test->nlog("No asup dir to validate \n");
        return $TCD::PASS;
    }
    if (ASUP::verify_autosupport_format(asup_dir=>$asup_dirs,validate=>"all");
        return $TCD::PASS;
    }

Ex7: Validate archived asup

    my $asup=ASUP::invoke_autosupport(message=>"My_Test_ASUP");
    my $saved_result=ASUP::verify_asup_archived(event=>$asup);

Ex8: Manually invoke and verify asup collection on the filer only

    ASUP::config_asup(enable=>"on", support_enable=>"on",
        support_transport=>"SMTP", mailhost=>$CLIENT);
    my $asup=ASUP::invoke_autosupport(message=>"My_Test_ASUP");
    ASUP::verify_autosupport_collection(events=>$asup);
 
=cut
