#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/ASUP.pm#1

#***********************************************************************
#
# NAME
#   MimeCompliance.pm -
#
# SYNOPSIS
#   MimeCompliance Library for the invocation and format-validation of 7M and CM SN ASUP.
#
# DESCRIPTION
#	This library is used to enable the end user to invoke an ASUP and verify the email to be
#	MIME compliant. The current library was designed to use Server.pm, ASUP::SMTP_MIME_Server.pm and NACL,  
#   CM SN filer.
#
# AUTHOR
#   c-dl-amittal
#
# HISTORY
# 06/20/2012 Initial Revision
#***********************************************************************
## @summary     MimeCompliance library

package ASUP::MimeCompliance;

#use statements
use strict;
use Data::Dumper;
use Server;
use Tharn qw($Log);
use Options;
use Params::Common;
use Params::Validate qw(:all);
use ASUP;
use ASUP::SMTP_MIME_Server;
use Carp;

use File::Path;
use File::Basename;
use NACL::APISet;
use Clone;
use NACL::C::Node;

#Global variables
our $ontap_version;
my $Server = new Server();
my @Nodes  = $Server->servers();

#Methods

###############################################################################
#Sub: verify_mime_compliance
#Input Params: Hash
#Return values: 1 (success) or 0 (failure)
#Summary: Validates format of the saved autosupports.  Validation can occur
#       on asup headers, sections, attachments, or all.  Default is all.
###############################################################################
sub verify_mime_compliance {
	my $i;
   	my $result=1;
   	my $version = ASUP::_get_version();
   	
	my %params = (
		asup_dir        => undef,    #dirs where asups were saved
		alt_extension   => 0,
		minimal_content => 0,
		validate => "all",    #headers, sections, attachments, manifest, or all
		xfer_protocol => 'PUT|POST',    # by default either is ok
		check_status => 1,    #by default we should check the status of files in the manifest
		@_,
	);
	$Log->enter("Entering verify_mime_compliance");

	if ( !$params{asup_dir} ) {
		$Log->error("Verify MimeCompliance Format: A saved autosupport directory is required in order to validate the format of an asup \n");
		return 0;
	}
	my $content = "complete";
	$Log->comment("minimal param(int): $params{minimal_content}");
	if ( $params{minimal_content} != 0 ) {
		$content = "minimal";
	}
	
	#verify headers for all versions
	#verify headers format for each asup dir
	if ( $params{validate} eq "all" or $params{validate} eq "headers" ) {
		if (!_verify_headers(asup_dir      => $params{asup_dir},
							xfer_protocol => $params{xfer_protocol}))
		{
			$Log->error("Verify MimeCompliance Format: Failed to verify the headers for the autosupport \n");
			return 0;
		}
	}
	
	my %parsed_body = _parse_email( asup_dir => $params{asup_dir});  
	
	$Log->comment("*****************Results:****************");
	$Log->comment("Main content type is Fragment? $parsed_body{isFragment}");
	$Log->comment("Main content type is $parsed_body{contentType}");

	#Usage:
	#$Log->comment($#{$parsed_body{contents}}+1);
	#$Log->comment(%{$parsed_body{contents}[0]}->{isFragment});
   				    
    #Loop through the array of contents(Each array-element is a hash of sub-parts)
	for($i=0;$i<=$#{$parsed_body{contents}};$i++){
		$Log->comment("--Sub part ". ($i+1)." contents--");
		foreach my $key(keys %{$parsed_body{contents}[$i]} ){ #parsed_body{contents} is an array of hashes containing sub-parts of email
			if($key ne "contents"){
				$Log->comment("$key=>$parsed_body{contents}[$i]->{$key}");
			}
			if($parsed_body{contents}[$i]{errorFlag}== 1){
				$result=0;
			}
		} 
	$Log->comment("--Sub Part ".($i+1)." ends--"); 
	}
	return $result;
}

###############################################################################
#PRIVATE MimeCompliance FUNCTIONS
###############################################################################

##############################################################################
#Sub: _parse_email
#Input Params: asup directory
#Return values: Hash (success) or 0 (failure)
#Summary: Parses the email and validates the body-format
###############################################################################
sub _parse_email {
	$Log->enter("Entering _parse_email");
	my %params = (
		asup_dir => undef, #A directory where asups are saved
		@_,
	);

	my $email;
    foreach my $dir ( @{ $params{asup_dir} } ) {
		my $files = `ls $dir`;
		if ( $files =~ /email/ ) 
		{
			$Log->comment("Email found!");
			$email = `cat $dir/email`;
			my $content_type = "Content-Type";
			my %content_type_hash;
			my $container_contents;
			my $content_main_type;
			my %parsed_body;

			#Checking the Mime Version Format
			if ( $email =~ /MIME-Version:(.+)\r\n/im ){
				my $version_result=_verify_version_format( version_number => $1 );
				if ($version_result==1) {
					$Log->comment("Version number qualified!");
				}
				else {
					$Log->error("Version number not qualified!");
				}
			}
			else {
				$Log->error("MIME-Version field not found or is not in correct format");
			}
			
			if($email =~/Date:(.+)\r\n/im){
				$Log->comment("Verifying the date format..");
				my $date_result=_verify_date_format(date=>$1);
				if($date_result==0){
					$Log->error("Improper date format specified!");
				}
				else{
					$Log->comment("Date format passed!");
				}				
			}
			
			#Content-Type marks the beginning of the body
			my $body = index( $email, $content_type );

			#the string "email" now holds the entire email and $body_lines contains the body
			my $body_lines = substr $email, $body;
			my $content_type_with_parameters;

			#Checking for the 'Content Type' of email
			if ( $body_lines =~ /Content-Type:(.+)\r\n/im ) {
				$content_type_with_parameters = $1;
				#If parameters are passed along with the Content-Type field
				if ( $content_type_with_parameters =~ /(.+);(.+)/ ){
					$content_main_type = $1;
					$Log->comment("Content main type:$content_main_type");
				}
				#If there are no parameters passed after Content-Type
				else{
					$content_main_type=$content_type_with_parameters;
				}
                
				if ( $content_main_type =~ /multipart\/(.+)/i ){
					$Log->comment("Multipart and $1 message processing..");
					#Call the function to validate a multipart message
					%parsed_body = _validate_mime_container(
									container_string     => $body_lines,
									primary_content_type => $content_type_with_parameters); 
				}

				else {
					$Log->comment("Non Multipart Message Processing..");
					#Call the function to validate a non-multipart message
					%parsed_body = _validate_mime_fragment(
									fragment_string      => $body_lines,
									primary_content_type => $content_type_with_parameters);
				}
			
			}

		return %parsed_body;
		}
	}
	return 0; 
}

###############################################################################
#Sub: _verify_headers
#Input Params: Hash
#Return values: 1 (success) or 0 (failure)
#Summary: Validates the expected headers in each saved autosupport
###############################################################################
sub _verify_headers {
	$Log->enter("Entering _verify_headers");
	my %params = (
		asup_dir      => undef,         #dirs of the saved asups
		xfer_protocol => 'PUT|POST',    # by default either is ok
		@_,
	);
	my @valid_headers;
	my $version = ASUP::_get_version();

	#check for the required headers in each saved asup
	foreach my $dir ( @{ $params{asup_dir} } ) {
		my $files = `ls $dir`;

	 #first determine what headers we should expect based on the saved asup type
		my $headers = undef;

		if ( $files =~ /email/ ) {

			#email_1 contains the raw SMTP asup
			$headers       = `cat $dir/email`;
			@valid_headers = (
				"SENDER:",                      "RECEIVER:",
				"MESSAGE:",                     "To:",
				"Date:",                        "Subject:",
				"Mime-Version:",                "Content-Type:",
				"From:",
			);
			if ( $version ne "7G" ) {
				push( @valid_headers, "X-Netapp-asup-partner-system-id:" );
			}
		}

		#now check for the valid headers
		foreach my $basic_header (@valid_headers) {
			if ( $headers !~ /$basic_header/ ) {
				$Log->warn("_verify_headers: Failed to find a basic header, $basic_header, in the saved asup $dir \n");
				$Log->exit("Returning 0");
				return 0;
			}
			$Log->debug("_verify_headers: Validated $basic_header!!!");
		}

	}

	#found all of the expected headers in the saves asup dirs
	$Log->comment("_verify_headers: Found all expected asup headers \n");
	$Log->exit("Returning 1");
	return 1;
}

###############################################################################
#Sub: _validate_mime_fragment
#Input Params: Fragment string
#Return values: Hash of the parsed fragment
#Summary: Validates the mime fragment (non-multipart content)
###############################################################################
sub _validate_mime_fragment {
	
	my %params = (
		fragment_string      => undef,    #Body of email between the boundary
		primary_content_type => undef,
		@_,
	);
	$Log->enter("Entering _validate_mime_fragment");
	my $content_type_with_parameters;
	my %result;
	$result{"errorFlag"}=0;
	$result{"contentId"} = "undef";
	$result{"encoding"} = "undef";
	
	#Check if Content-Type is specified for the fragment
	if ( $params{fragment_string} =~ /Content-Type:(.+)\r\n/m ) {
		$content_type_with_parameters= $1;
		$Log->comment("Content type is:$content_type_with_parameters");

		    #If the Content Type is of type text
		if ( $content_type_with_parameters =~ /text\/(.+)/i ) {
			my $text_subtype=$1;
			my $text_parameter;
			
			#Check the sub type
		if ($text_subtype =~ /plain|enriched|richtext/i ) {
			$Log->comment("Valid subtype found for text Content-Type");
			if($text_subtype=~/(.+);(.+)/){ 		#if additional parameters are supplied after a colon
				$text_subtype=$1;
				$text_parameter=$2;
			if($text_parameter=~/charset=(.+)/){   #if charset is specified as parameter
				my $charset_value=$1;
				#check the parameter value of charset
				if($charset_value=~/us-ascii|ISO-8859-[0-10]/i){
					$Log->comment("Valid charset value found:$charset_value");
				}
				else{
					$Log->comment("Unrecognised charset value");
				}
			  }
			else{    #if the parameter specified is not charset
				$Log->error("Un-recognised parameter passed to text/$text_subtype");
				}
			  }
			 
		else{#if parameter is not passed to a text type
			$Log->comment("Parameter absent for a text subtype.Assuming defaults");
			$text_parameter="charset=us-ascii";
			}
			$result{"contentType"} = "text/$text_subtype;$text_parameter";
		   }
		   	 
		   	 
		else{
			$Log->comment("Unrecognised subtype for text Content-Type");
			#If additional parameters are passed after un-recognised text subtype
			if($text_subtype=~/(.+);(.+)/){
			my $text_subtype=$1;
			my $text_parameter=$2;
				if($text_parameter=~/charset=(.+)/){
					if($1!~/us-ascii|ISO-8859-[0-10]/i){
					$Log->comment("Unrecognised charset value and subtype for text");
					$result{"contentType"}="application/octet- stream";
					}
				}
			}
			else{ 	#if no parameter is specified for un-recognised subtype of text
				$Log->comment("Defaults assumed for un-recognised sub-type of text");
				$result{"contentType"}="text/plain;charset=us-ascii";
				}
			}
		}
         
        #Validation for Content-Type:image
		elsif ( $content_type_with_parameters =~ /image\/(.+)/i ) {
			my $image_subtype=$1;
			if ( $image_subtype =~ /jpeg|gif/i ) {
				$Log->comment("Valid subtype found for image Content-Type");
				$result{"contentType"} = "image/$image_subtype";
			}
			else {
				$Log->comment("Unrecognised subtype for image Content-Type");
				$result{"contentType"} = "application/octet-stream";
			}

		}

		#Validation for Content-Type:audio
		elsif ( $content_type_with_parameters =~ /audio\/(.+)/i ) {
			my $audio_subtype=$1;
			if ( $audio_subtype=~ /basic/i ) {
				$Log->comment("Valid subtype found for audio Content-Type");
				$result{"contentType"} = "audio/$audio_subtype";
			}
			else {
				$Log->comment("Unrecognised subtype for audio Content-Type");
				$result{"contentType"} = "application/octet-stream";
			}
		}
	    
	    #Validation for Content-Type:video
		elsif ( $content_type_with_parameters =~ /video\/(.+)/i ) {
			my $video_subtype=$1;
			if ( $video_subtype =~ /mpeg/i ) {
				$Log->comment("Valid subtype found for video Content-Type");
				$result{"contentType"} = "video/$video_subtype";
			}
			else {
				$Log->comment("Unrecognised subtype for video Content-Type");
				$result{"contentType"} = "application/octet-stream";
			}
		}

		#Validation for Content-Type:application
		elsif ( $content_type_with_parameters =~ /application\/(.+)/i ) {
			my $application_subtype=$1;
			if ( $application_subtype =~ /octet-stream|postscript/i ) {
				$Log->comment("Valid subtype found for application Content-Type");
				$result{"contentType"} = "application/$application_subtype";
			}
			else {
				$Log->comment("Unrecognised subtype for application Content-Type");
				$result{"contentType"}="application/$application_subtype";
			}
		}

		#Validation for Content-Type:message
		elsif ( $content_type_with_parameters=~ /message\/(.+)/i ) {
			my $message_subtype=$1;
			#Checking for valid sub-types
			if ( $message_subtype =~ /rfc822|partial|external-body/i ) {
				$Log->comment("Valid subtype found for message Content-Type");
				$result{"contentType"} = "message/$message_subtype";
				
				#If parameters are passed after a colon
				if($content_type_with_parameters=~/message\/external-body;((.|\s)+)/){
				my $message_parameter=$1;
				
				#Content-ID and Encoding must be present for a message/external body message
				if($result{contentID} eq "undef" || $result{encoding} eq "undef"){
					$Log->error("Content-ID or Encoding not specified for a message/external body message");	 	
				}
				
				 if($message_parameter!~/access-type/i){
				 	$Log->error("ACCESS-TYPE parameter absent for a message/external type message");
				 }
				 else{
				 	$Log->comment("Access-type is specified!");
				 	if($message_parameter=~/(.+)=(.+)/){
				 		my $access_type_value=$2;
				 		if($access_type_value=~/FTP|ANON-FTP|TFTP|LOCAL-FILE|MAIL-SERVER/){
				 			$Log->comment("Valid access type found!");
				 		}
				 		else{
				 			$Log->error("Invalid access type for message/external type");
				 		}
				 	 }
				   }
	            }
			}
			else {
				$Log->comment("Unrecognised subtype for application Content-Type");
				# $result{"contentType"}="application/octet-stream";
			}
		}
		
		#If Private values are used for Content-Type
		elsif($content_type_with_parameters=~ /X-.+/i ){
			$Log->comment("Private type used!");
			$result{"contentType"}="$1";
		}
		
		else{
			$Log->error("Un-recognised 'type' used!");
			$result{"contentType"}="undef";
		}
	}
	
	#If Content-Type header is missing
	else {
		$Log->comment("Content-Type is absent.Default values assumed");
		if($params{primary_content_type}=~/multipart\/digest/i){
			$result{"contentType"} = "message/rfc822";
		}
		else{
			$result{"contentType"} = "text/plain;charset=us-ascii";
		}                                                                                                                                                            
	 }
	
	#If Content-Transfer-Encoding is specified
	if ( $params{fragment_string} =~ /Content-Transfer-Encoding:(.+)\r\n/m ) {
		my $encoding_used=$1;
		if($result{contentType}=~/multipart|message/i){
			#Messages of type multipart/message should have encoding only 7bit,8bit or binary
			if($encoding_used=~/7BIT|8BIT|BINARY/i){
				$result{"encoding"} = $encoding_used;
				$result{"encoding"} =~s/\s//g; #removing spaces if any
			}
			else{
				$Log->error("$result{contentType} message cannot have encoding other than 7BIT,8BIT or BINARY");
				$result{"errorFlag"}=1;
				$result{"encoding"} = "invalid";  #Note: It is defined but is invalid
			}
		}
		
		#If the fragment content type is not multipart or message
		else{
			#Check if the encoding is among the accepted ones.
			if($encoding_used=~/7BIT|8BIT|BINARY|quoted-printable|base64|ietf-token|x-token/i){
				 $result{"encoding"} = $encoding_used;
				 $result{"encoding"} =~s/\s//g;  #removing white spaces if any 
			}
			else{
				$Log->error("Encoding is not valid!!");
				$result{"errorFlag"}=1;
				$result{"encoding"}="invalid";     #Encoding is defined but invalid
			}
		}
	}
	else {
		$Log->comment("Encoding is absent.Assume Content-Transfer-Encoding: 7BIT ");
		$result{"encoding"} = "7BIT";
	}

	if ( $params{fragment_string} =~ /Content-ID:((.|\s)+)\r\n/mg ) {
		$result{"contentId"} = $1;
		$Log->comment("Content-ID is: $result{contentId}");
	}

	if ( $params{fragment_string} =~ /(.|\s)*\r\n\r\n([\s\S]+)\r\n/m ) {
			$result{"contents"} = $2;
	}
 
     #verify the encoding
   	if($result{encoding} eq "base64"){
        my $verify_crlf=_verify_encoding(encoded_string=>$result{contents});
         if($verify_crlf==0){
         	$Log->comment("Verify_encoding returned $verify_crlf");
         	$result{"errorFlag"}=1;
         }
    }
    		
	$result{"isFragment"} = "true";
	$Log->exit("Returning result");
	return %result;
}

###############################################################################
#Sub: _validate_mime_container
#Input Params: container string
#Return values: hash(success) or 0 (failure)
#Summary: Validates the mime container (multipart content)
###############################################################################
sub _validate_mime_container {
	$Log->enter("Entering _validate_mime_container");
	my ($cType,$boundary,%result);
	$result{"errorFlag"}=0;
	my @contents;
	my %params = (
		container_string => undef,    # Body of the multi-part message
		primary_content_type =>undef,    # Primary Content Type of the email, with parameters
		@_,
	);

	my $primary_content_type = $params{primary_content_type};
	if ( !$primary_content_type ) {
		$Log->error("Primary Content-Type is required!");
		$result{"errorFlag"}=1;
	}

	#The $params{container_string} holds the entire email body in a container
	if ( $params{container_string} =~ /^Content-Type:((.|\r\n )+)\r\n/m ) {
		$cType = $1;
		#$Log->comment("Content-type in mime-container:$cType");
	}

	#If Content-Type is not specified, assume defaults
	else {
		if ( $primary_content_type =~ /\s*multipart\/digest(.+)/i ) {
			$Log->comment("Assuming defaults for multipart/digest");
			$cType = "message/rfc822";
		}
		else {
			$Log->comment("Assuming default Content-Type for Multipart(non-digest)message");
			$cType = "text/plain;charset=us-ascii";
		}
	}

	if ( $cType =~ /.*boundary="(.+)"/m ) {
		#verify and save the boundary name
		$boundary = $1;
		$Log->comment("Boundary present!=>$boundary");

		#Check if the boundary ends in a white space, and chop it if it does.
		if ( $boundary =~ /(.+)\s$/ ) {
			$boundary = chomp($boundary);
		}
	}
	else {
		$Log->error("Boundary is absent for a multipart message");
		$result{"errorFlag"}=1;
	}

   #Boundary should be defined in order to parse the email
   if($boundary ne ''){
   	$result{"contentType"} = $cType;
	$result{"boundary"}    = $boundary;
	$result{"isFragment"}  = "false";

	my $content_string;
	if ( $params{container_string} =~/--$boundary\r\n((.|\s)*--$boundary)--/img )
	{
        #$1 contains all the email from the first appearance of a boundary till the end
        #All the content after final boundary must be ignored as per rfc 2045
		$content_string = $1;
					
		#Loop if you have mime boundaries within the email
		while ( $content_string =~ /((.|\s)+?)--$boundary/m ) {
			$Log->comment("--Sub-part--");
			my $next_part = $1;
			# +4 is for the two -- and the \r\n
			$content_string = substr $content_string,length($next_part) + length($boundary) + 4;
			if ( $next_part =~ /^Content-Type: multipart\/(.+);/i ) {
				my %refHash =_validate_mime_container(
							container_string     => $next_part,
							primary_content_type => $primary_content_type);
							push(@contents,\%refHash);
			}
			else {
				my %validated =_validate_mime_fragment( 
								fragment_string => $next_part,
								primary_content_type =>$primary_content_type);
				$Log->comment("Pushing the mime fragment!!");
				push( @contents,\%validated );
				}
		}
	}
	else{
		$Log->error("Could'nt parse email:Boundary not terminated properly!");
	}
   }
   
	# If the boundary is ''
	else{
		$Log->warn("Could not parse body due to missing boundary parameter!");
	}

	$result{"contents"} = \@contents;
	$Log->exit("Returning result");
	return %result;
}

###############################################################################
#Sub: _verify_encoding
#Input Params: encoded string
#Return values: 1(success) or 0 (failure)
#Summary: Validates the base-64 encoding of the string
###############################################################################
sub _verify_encoding{
	my $result=1; #Upon error change it to 0 
	$Log->enter("Entering _verify_encoding");
	my %params = (
		encoded_string => undef,    #encoded attatchment string
		@_,
	);
	$Log->comment("Verifying the encoding...");
	
	#Split based on the 'CRLF'
	my @encoded_array=split('\r\n',$params{encoded_string});
	my $last_line=$encoded_array[-1];
	$Log->comment("Last line of the encoding was: \n-->$last_line<--");
	my $length=length($last_line);
	if(($length%4)!=0){
		$Log->error("Length ($length) is not a multiple of 4");
	}
	else{
		$Log->comment("Length ($length) is a multiple of 4");
	}
	
    foreach my $encoded_line(@encoded_array){
    	if($encoded_line=~/(\r|\n)/){
    		$Log->error("Bare CR or LF found in encoding!");
    		$result=0;
    	}
    	my $encoding_length=length($encoded_line);  #This must be <= 76-2 
    	if($encoding_length>74){
    	$Log->warn("Encoding not properly terminated at:$encoded_line");
        $result=0;
    	}
    }
  return $result;
}


###############################################################################
#Sub: _verify_version_format
#Input Params: Version number string
#Return values: 1 (success) or 0 (failure)
#Summary: Validates the Version number
###############################################################################

sub _verify_version_format {
	$Log->enter("Entering _verify_version_format");
	my %params = (
		version_number => undef,    #Version number of MIME 
		@_,
	);

	if ( $params{version_number} =~ /(\d)(\.\d){1}/ ) {
		$Log->exit("Returning 1");
		return 1;
	}
	else {
		$Log->exit(" _verify_version_format returning 0");
		return 0;
	}
}

###############################################################################
#Sub: _verify_date_format
#Input Params: Date string
#Return values: 1 (success) or 0 (failure)
#Summary: Validates the Date field
###############################################################################
sub _verify_date_format{
	$Log->enter('Entering _verify_date_format');
	my %params =(
		date=>undef,
		@_,
	);
	 #rfc 2822 compliance
	if($params{date}=~/['Sun''Mon''Tue''Wed''Thu''Fri''Sat'],\s*(0[1-9]|[12][0-9]|3[01])\s*(Jan|Feb|mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s*(19|20)\d\d\s*([01][0-9]|2[0-3]):([0-5][0-9]):([0-5][0-9]|60)\s*(\+|-)[0-9959]/){
		return 1;
	}
	else{
		$Log->comment("Date::format not supported!");
		$Log->exit("_verify_date_format returning 0");
		return 0;
	}	
}


###############################################################################
#Sub: invoke_and_verify_mime_compliance
#Input Params: Hash
#Return values: 1 (success) or 0 (failure)
#Summary: Combination routine that will:
#   1.  Start up the SMTP server.
#   2.  Invoke the specified autosupport event, or manually invoke one if no
#           event was specified.
#   3.  Verify that the autosupport was received by the server.
#   4.  If no asup save dir was specified, it will return a 1 or a 0 here based
#           upon whether the asup was received or not.
#   5.  If an asup save dir was specified, the routine then attempts to validate
#           the asup mime compliance:
#       A) validate the headers for the asup
#       B) validate the body of the asup
#   6.  Then returns 0 or 1 based on those validations.
###############################################################################
sub invoke_and_verify_mime_compliance {
    my %params = (
            event=>undef,   #the asup to generate
            asup_save_dir=>undef, #dir to save asups to
            client=>undef, #client to run the server on
            transport=>"SMTP", #transport type to use
            num_events=>1,
            @_,
    );
    my $transport_server;
    if (!defined $params{client}) {
        $Log->error("A client is required in order to invoke the $params{transport} server \n");
        return 0;
    }
    
    #Configure autosupport settings on the filer and invoke the tansport server
    $Log->comment("Invoke and Verify: Configuring autosupport on the filer \n");
    $Log->comment("Invoke and Verify: Creating $params{transport} Server object \n");

    ASUP::config_asup(enable=>"on", support_enable=>"on", support_transport=>$params{transport}, mailhost=>$params{client});
    $transport_server = new ASUP::SMTP_MIME_Server(client => $params{client}, asup_dir=>$params{asup_save_dir});

    #Invoke the autosupport
    $Log->comment("Invoke and Verify: Invoking an autosupport \n");
    my $events;
    for (my $i=0; $i<$params{num_events}; $i++) {
        my $message="My_test_autosupport_".time()."_".$i;
        my $event=ASUP::invoke_autosupport(message=>$message);
        push (@$events,$event);
    }
    
    $Log->comment("Looking to see if events were received..");
    #Verify autosupport was received by the transport server
    my $asup_dirs=ASUP::verify_autosupport_reception(events=>$events, server=>$transport_server);
    
    #Stop the transport server
    $Log->comment("Invoke and Verify: Stopping the $params{transport} server \n");
    $transport_server->stop();

    #if verify asup reception returned 0, then we failed to verify the asup was received
    if ($asup_dirs==0) {
        $Log->comment("Invoke and Verify: Failed to verify that the autosupport was received, $events \n");
        return 0;
    }
    #if verify asup reception returned 1, then we validated that the asup was received
    #and no autosupports were saved to disk
    elsif ($asup_dirs==1) {
        return 1;
    }
    #else, the autosupports were saved to disk, so lets validate them
    elsif (defined $params{asup_save_dir}) {
        if(!verify_mime_compliance(asup_dir =>$asup_dirs,validate=>"all",check_status=>0)){
 	 	return 0;
        }
     return 1;
    }
    #either an error occurred or an unexpected value was returned, so fail
    else {
        return 0;
    }
}


=pod

=head1 NAME
MimeCompliance.pm - MIME verification library for asups generated in Cmode SN.


=head1 SYNOPSIS

    use MimeCompliance;
    MimeCompliance::verify_mime_compliance(
	        asup_dir        => $LOGDIR,    #dirs where asups were saved
			alt_extension   => 0,
			minimal_content => 0,
			validate => "all",    #headers, sections, attachments, manifest, or all
			xfer_protocol => 'PUT|POST',    # by default either is ok
			check_status => 1,    #by default we should check the status of files in the manifest
            );


=head1 DESCRIPTION

The methods included in this library were developed in order to assist
engineers in verification of the generated email for mime compliance.  
The current library supports Cmode filers and invokes communication to
the filer via the Server.pm library. The MimeCompliance.pm library allows the 
user to validate the generated autosupport to be MIME compliant.The 
autosupport will be received by an SMTP server and the received autosupport
will be used to validate the format for MIME complaince (the headers and
sections shall be examined).This library makes use of ASUP.pm to invoke and
verify the autosupport reception.
  
    
=head1 METHODS

   verify_mime_compliance(asup_dir => LOCAL_ASUP_DIR);
        Description:
            This method will allow the user to perform format verification on
                a received and saved autosupport.  The verification will be
                automatically done based upon the mode, content type, and
                autosupport message received.  Autosupport format can be
                verified for headers, sections, attachments, or all.  All is
                the default.
        Inputs:
            asup_dir => <local dir> The local directory where the received
                autosupports are saved.
            alt_extension => <0|1> The default is 0, meaning we use .gz for all
                attachments.  If set to 1, we will attempt to validate the
                attachments with .gx extensions.
            minimal_content => <0|1> The default is 0, meaning we should attempt
                to validate the content as complete.  If set to 1, we will only
                validate a minimal content type - available in 7M only.
            validate => <sections|headers|attachments|all> What to validate.
        Outputs:
            0|1 Returns 1 on successful validation and 0 on any failure.  See
                the logs for more details on a failure.
        Examples:
            see full example below
            
	_parse_email(asup_dir=>LOCAL_ASUP_DIR);
		Description:
			This method will parse the email stored in the specified directory 
				and will return the parsed body as a hash value including the 
				various fields present in each sub-part of a multipart email, and 
				their values.
		Inputs:
			asup_dir => <local dir> The local directory where the received
                autosupports are saved.
        Outputs:
        	Hash|0 Returns the HASH of values present in each sub-part of a multipart
        	email or 0 on failure.
        Examples:
            see full example below
            
	_validate_mime_fragment(fragment_string=>$fragment_string, primary_content_type=>$email_type)
		Description:
			This method will validate the individual sub-part of a multipart email
				or the body of a non-multipart email
		Inputs:
			fragment_string =><body_of_fragment> A string containing the body of a fragment  
			primary_content_type =><primary_type_of_email> The primary type of email
		Outputs:
			Hash value consisting of individual fields(headers and their corresponding validated 
				fields).
		Examples:
            see full example below

	_validate_mime_container(container_string=>$container_string, primary_content_type=>$email_type)
		Description:
			This method will validate the multi-part email by calling the _validate_mime_fragment on 
				each fragment(sub part of a multipart email) of a container(the multipart message itself).
		Inputs:
			container_string=><container_body> String containing the body of container
			primary_content_type=><primary_type_of_email> Primary email type
		Outputs:
			Hash of values present in the entire emultipart message, after Validation.
		Examples:
			see full example below
	
	_verify_encoding ($encoded_body)
		Description:
			This method will validate the body of an encoded email to be MIME compliant
		Inputs:
			String containing the body of encoded-message
		Outputs:
			1|0 Returns 1 on success after validation and 0 on error
		Examples:
			See full example below
		
	_verify_version_format($version_number)
		Description:
			This is a helper function that validates the format of the Version number to 
				be MIME compliant
		Inputs:
			A string containing the Version Number
		Outputs:
			1|0 Returns 1 on success and 0 on failure
		Examples:
			See full example below

	_verify_date_format($date_string)
		Description:
			This is a helper function that validates the format of the date string to be MIME Compliant
			as per RFC 2822
		Inputs:
			A string containing the date header of email
		Outputs:
			1|0 Returns 1 if in proper format or a 0 on error		
			
=head1 EXAMPLES

Typical example:

    use MimeCompliance;
    use vars qw($LOGDIR $LOGDIR_UNIX $CLIENT);
    my $Server  = Server->new();
    my @Nodes  = $Server->servers();
    $Server->set_execution_server(-Server => $Nodes[0]);
    
    sub verify_http_protocol {
    $Test->description("Check http.");
    
    if(!MimeCompliance::verify_mime_compliance(asup_dir =>$asup_dirs,validate=>"all",check_status=>0)){
	 $Test->nlog("verify_mime_compliance failed!");
 	 return $TCD::FAIL;
   }
    return $TCD::PASS;

Ex1: Parse email

	my %parsed_body = _parse_email( asup_dir => $params{asup_dir} );

Ex2: Validating a mime container

	my %parsed_body = MimeCompliance::_validate_mime_container(
									container_string     => $body_lines,
									primary_content_type => $content_type_with_parameters
									); 
									
Ex3: Validating a mime fragment
	my %parsed_body = MimeCompliance::_validate_mime_fragment(
									fragment_string      => $body_lines,
									primary_content_type => $content_type_with_parameters
									);

Ex4: Verify encoding of an encoded message

    my $verified_result=MimeCompliance::_verify_encoding(
    								encoded_string => $encoded_body
    								);

Ex5: Verify Version Format
	my $verified_result=MimeCompliance::_verify_version_format(
									version_number => $version_number_string
									);
Ex6: Verify date format
	my $date_verified_result=MimeCompliance::_verify_date_format(date=>$date_string);
	
Ex7: Manually invoke and verify an asup

    my $asup_server=ASUP::start_transport_server(asup_save_dir=>$LOGDIR,
        client=>$CLIENT, transport=>"SMTP");
    ASUP::config_asup(enable=>"on", support_enable=>"on",
        support_transport=>"SMTP", mailhost=>$CLIENT);
    my $asup=ASUP::invoke_autosupport(message=>"My_Test_ASUP");
    my $asup_dirs=ASUP::verify_autosupport_reception(events=>$asup,
        server=>$asup_server);
    ASUP::stop_transport_server(server=>$asup_server);
    if (!defined $asup_dirs or $asup_dirs==0) {
        return $TCD::FAIL;
    }
    if ($asup_dirs==1) {
        $Test->nlog("No asup dir to validate \n");
        return $TCD::PASS;
    }
    if (MimeCompliance::verify_mime_compliance(asup_dir=>$asup_dirs,validate=>"all");
        return $TCD::PASS;
    }
 
=cut

