#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/ASUP_REST.pm#1 $

#***********************************************************************
#
# NAME
#   ASUP_REST.pm - 
#
# SYNOPSIS
#   A NetApp ASUPDW3 REST interface implemented via PERL.
#
# DESCRIPTION
#   This library provides an interface to the NETAPP ASUP REST
#   (Representational State Transfer) interface.  
#   The interface can make requests of the REST server for a given
#   ASUP and get back the following information:
#
# AUTHOR
#   dl-cluster-qa
#
# HISTORY
# 8/9/2010 Initial revision
#***********************************************************************

## @summary     ASUP_REST.pm

package ASUP_REST;

use strict;

use DateTime::Format::Strptime;

my $default_timeout = 30;  # this should be 600

use Options;
use Tharn qw(host snooze subtest $Parent $Log fatal_result);

use NATE::Result;
use Data::Dumper;
use Server;
use Sys::Hostname;
use LWP::UserAgent;
use HTTP::Response;
use HTTP::Status;
use File::Basename;
use DataUtils;
use Clone;
use ASUP;

my $opts = new Options;

my $request_seq=0;

#Global variables
my $Server = new Server();
my @Nodes=$Server->servers();

$opts->parameters(definitions=>
    [
       {
        name => "REST_URL",
        description => "URL for REST server",
        default => "http://support-tools.corp.netapp.com",
       },
       {
        name => "REST_USERID",
        description => "UserId to use when making REST requests",
        default => "?????",
       },
       {
        name => "REST_DATE_FORMAT",
        description => "Date format to be used in REST queries",
        default => '%Y-%m-%dT%T',
       },
    ]);

my $opt_href = $opts->get_hash();

## @name ASUP_REST new method
## @description Creates a new ASUP_REST object
##
## @synopsis
## my $AsupRest = ASUP_REST->new( rest_url    => $rest_url,
##                                rest_userid => $rest_userid );
##
## @arg rest_url    Optional: The URL to use when communicating with the ASUPDW3
##                            REST server.  By default will use the global REST_URL
##                            variable.
## @arg rest_userid Optional: The userid to use when communicating with the ASUPDW3
##                            REST server.  By default will use the global REST_USERID
##                            variable.
##
##
sub new {
    my $this = shift;
    my $class = ref($this) || $this;
    my %options = (
        rest_url    => $opt_href->{REST_URL},
        rest_userid => $opt_href->{REST_USERID},
        @_
    );

    my $self = {
        "rest_url"        => $options{rest_url},
        "rest_userid"     => $options{rest_userid},
        @_,
    };
    bless($self, $class);
    return $self;
}

## @name _get_system_info()
## @description Private method to retrieve (and cache in the object) SYSTEM_ID & SYSTEM_SERIAL 
##              in a mode agnostic way
##
## @synopsis
##   $AsupRest->_get_system_info( node => $node );
##
## @arg node   Required: The node for which we want the SYSTEM_ID & SYSTEM_SERIAL
##                            REST server.  By default will use the global REST_URL
##                            variable.
##
##
sub _get_system_info {
    my $self = shift;
    my %options = (
        node=>undef,
        @_
    );

    if ( !defined $options{node} ) {
        die script_result("ASUP_REST::_get_system_info() must be called with the node option");
    }

    if ( !defined $self->{$options{node}}->{SYSTEM_ID} ) {
        my $version=ASUP::_get_version();
        if ($version eq "7G" or $version=~/7M/) {
            my $sysconfig = Sysconfig->new( $options{node} );
    
            my $system_id = $sysconfig->system_id();
            my $serial_num = $sysconfig->system_serial_number();

            $self->{$options{node}}->{SYSTEM_ID} = $system_id;
            $self->{$options{node}}->{SYSTEM_SERIAL} = $serial_num;
        } else {
            my $sysconfig = $Server->system_node_show( node => $options{node} );
            $self->{$options{node}}->{SYSTEM_ID}     = $sysconfig->[0]->{systemid};
            $self->{$options{node}}->{SYSTEM_SERIAL} = $sysconfig->[0]->{serialnumber};
        }
    }
}

## @name _build_uri()
## @description Private method to construct a URI to be used to query the ASUPDW3 REST interace.
##
## @synopsis
##   $AsupRest->_build_uri( node     => $node,
##                          gen_date => $asup_generation_date,
##                          subject  => $asup_subject_string,
##                          seq_num  => $asup_sequence_number,
##                          list     => $list_content_links,
##                          timeout  => $timeout_for_request );
##
## @arg node     Optional: The node for which the ASUP was generated.
## @arg gen_date Optional: The date and time the ASUP was generated.  Can be built from
##                         the "event log show" for the callhome.xxx event for that ASUP,
##                         or -last-time-sent from "autosupport show" (presuming that the
##                         ASUP was the last one sent out)  Or most definitively from the
##                         header.txt file from the ASUP itself. 
##                         Needs to be in the format yyyy-mm-ddThh:mm:ss
## @arg subject  Optional: The subject text associated with the ASUP.
## @arg seq_num  Optional: The ASUP sequence number.
## @arg asup_id  Optional: Uniquely identifies an ASUP in the ASPUDW3 REST system.
## @arg list     Optional: Should ASUP content links be listed?
##
## @return
##    String in the format (could optionally have subject and/or seq_num in the URI):
## http://support-tools.corp.netapp.com/ASUP_DATA/client_id/scottr/system_id/0101166137/sys_serial_no/210265322/asup_gen_date/2002-05-30T09:00:00
##    or (could optionally have list tacked on the end)
## http://support-tools.corp.netapp.com/ASUP_DATA/client_id/scottr/asup_id/2010032715030069
##
sub _build_uri {
    my $self = shift;
    my %options = (
        node     => undef,
        gen_date => undef,
        subject  => undef,
        seq_num  => undef,
        asup_id  => undef,
        list     => undef,
        @_
    );

    if ( !defined $options{node} ) {
        die script_result("ASUP_REST::_build_uri() must be called with the node option");
    }

    if ( (!defined $options{gen_date}) || (!defined $options{asup_id}) ) {
        die script_result("ASUP_REST::_build_uri() must be called with either the gen_date or asup_id options but not both");
    }

    if ( (defined $options{asup_id}) && ($options{gen_date} || $options{subject} || $options{seq_num} ) ) {
        $Log->warn("ASUP_REST::_build_uri() called with asup_id, gen_date|subject|seq_num options will be ignored");
    }

    if ( (!defined $options{asup_id}) && $options{list} ) {
        $Log->warn("ASUP_REST::_build_uri() called with list, but no asup_id option, list option will be ignored");
    }

    _get_system_info( node => $options{node} );

    my $req_uri;

    if ( defined $options{asup_id} ) {
        $req_uri = $self->{rest_url}.'/ASUP_DATA'.
                   '/client_id/'.$self->{rest_userid}.
                   '/asup_id/'.$options{asup_id};
        if ( $options{list} ) {
            $req_uri = $req_uri.'/list';
        }
    } else {
        $req_uri = $self->{rest_url}.'/ASUP_DATA'.
                   '/client_id/'.$self->{rest_userid}.
                   '/hostname/'.$options{node}.
                   '/system_id/'.$self->{$options{node}}->{SYSTEM_ID}.
                   '/sys_serial_no/'.$self->{$options{node}}->{SYSTEM_SERIAL}.
                   '/asup_gen_date/'.$options{gen_date};

        if ( $options{subject} ) {
            $req_uri = $req_uri.'/asup_subject/'.$options{subject};
        }
    
        if ( $options{seq_num} ) {
            $req_uri = $req_uri.'/asup_seq_no/'.$options{seq_num};
        }
    }

    return $req_uri;
}

# Multiple public wrappers are provided to limit options passed to the swiss army knife routine _request()

## @name query_asup()
## @description Public method to query the REST interface for one or more ASUP Unique
##              identifiers based on information available from the filer.
##
## @synopsis
##   $AsupRest->query_asup( node     => $node,
##                          gen_date => $asup_generation_date,
##                          subject  => $asup_subject_string,
##                          seq_num  => $asup_sequence_number,
##                          timeout  => $timeout_for_request );
##
## @arg node     Required: The node for which the ASUP was generated.
## @arg gen_date Required: The date and time the ASUP was generated.  Can be built from
##                         the "event log show" for the callhome.xxx event for that ASUP,
##                         or -last-time-sent from "autosupport show" (presuming that the
##                         ASUP was the last one sent out)  Or most definitively from the
##                         header.txt file from the ASUP itself. 
##                         Needs to be in the format yyyy-mm-ddThh:mm:ss
## @arg subject  Optional: The subject text associated with the ASUP.
## @arg seq_num  Optional: The ASUP sequence number.
## @arg timeout  Optional: The maximum amount of time (in seconds) to wait until the REST interface
##                         responds.
## @return
##    When called in scalar context, return is $status which is 
##      $status - one of:
##        "REQ_FAILED"      - REST interface failed to respond to request
##        "RES_SUCCESS"     - REST interface responded to request with SUCCESS
##        "RES_IN_PROGRESS" - REST interface responded that the data being requested
##                            is still being processed.
##        "RES_FAILED"      - REST interface responded to request with a FAILURE
##    
##    When called in an array context, return is ($status, $href, $xmldoc) which is 
##      $status - same as above:
##      $href   - xml response from REST interface converted to simplified hash form.
##      $xmldoc - raw XML::LibXML::Document object reference
##    
##
sub query_asup {
    my $self = shift;
    my %options = (
        node     => undef,
        gen_date => undef,
        subject  => undef,
        seq_num  => undef,
        timeout  => $default_timeout,
        @_
    );

    return $self->_request( node     => $options{node}, 
                            gen_date => $options{gen_date},
                            subject  => $options{subject},
                            seq_num  => $options{seq_num},
                            timeout  => $options{timeout} );
}

## @name list_asup_content()
## @description Public method to query the REST interface for a list of content
##              tokens and URLs.  These URLs can then themselves be used to retrieve
##              ASUP content.
##
## @synopsis
##   $AsupRest->list_asup_content( asup_id   => $asup_unique_id,
##                                 recursive => $retrieve_all_content,
##                                 timeout  => $timeout_for_request );
##
## @arg asup_id   Required: The unique id of an ASUP in the ASUPDW3 system.
## @arg recursive Optional: Boolean, true means extract all of the content for the ASUP.
## @arg timeout   Optional: The maximum amount of time (in seconds) to wait until the REST interface
##                          responds.
## @return
##    When called in scalar context, return is $status which is 
##      $status - one of:
##        "REQ_FAILED"      - REST interface failed to respond to request
##        "RES_SUCCESS"     - REST interface responded to request with SUCCESS
##        "RES_IN_PROGRESS" - REST interface responded that the data being requested
##                            is still being processed.
##        "RES_FAILED"      - REST interface responded to request with a FAILURE
##    
##    When called in an array context, return is ($status, $href, $xmldocs) which is 
##      $status - same as above:
##      $href   - xml response from REST interface converted to simplified hash form.
##                Note: if you pass in the "recursive => 1" flag, the list hash results will
##                      be under a top-level LIST key, and the content will be contained under
##                      a top-level CONTENT key.  If recursive is 0, then only the top-level 
##                      LIST key will be present.
##      $xmldocs - array of raw XML::LibXML::Document object references
##
sub list_asup_content {
    my $self = shift;
    my %options = (
        asup_id   => undef,
        recursive => 0,
        timeout   => $default_timeout,
        @_
    );

    my ($status, $h1, $h2, $h3, $xmlDoc);
    my @xmlDocs;

    (undef, $h1, $xmlDoc) = $self->_request( asup_id  => $options{asup_id}, 
                                             list     => 1,
                                             timeout  => $options{timeout} );
    push( @xmlDocs, $xmlDoc );

    my $hp = $h1;

    if ( $options{recursive} ) {
       $h3->{LIST} = Clone::clone($h1);

        my @content_uris;

        DataUtils::extract_field( hash => $h1, array => \@content_uris, field => 'link' );

        foreach my $uri ( @content_uris ) {
           my $name = fileparse( $uri );
           (undef, $h2, $xmlDoc ) = $self->query_asup_content( uri      => $uri,
                                                               timeout  => $options{timeout} );

           push( @xmlDocs, $xmlDoc );
           $h3->{CONTENT}->{$name} = Clone::clone($h2);
        }

        $hp = $h3;
    }

    $status = _aggregate_hash_status($hp);

    return wantarray ? ($status, $hp, \@xmlDocs ) : $status;
}

## @name query_asup_content()
## @description Public method to query the REST interface for the content stored for
##              a given ASUP.
##
## @synopsis
##   ($status, $hash, $xmlDoc ) = $AsupRest->query_asup_content( uri     => $request_uri_for_content,
##                                                               timeout => $timeout_for_request );
##
## @arg uri      Required: The universal resource identifier corresponding to a REST query.
## @arg timeout  Optional: The maximum amount of time (in seconds) to wait until the REST interface
##                         responds.
##
## @return
##    When called in scalar context, return is $status which is 
##      $status - one of:
##        "REQ_FAILED"      - REST interface failed to respond to request
##        "RES_SUCCESS"     - REST interface responded to request with SUCCESS
##        "RES_IN_PROGRESS" - REST interface responded that the data being requested
##                            is still being processed.
##        "RES_FAILED"      - REST interface responded to request with a FAILURE
##    
##    When called in an array context, return is ($status, $href, $xmldoc) which is 
##      $status - same as above:
##      $href   - xml response from REST interface converted to simplified hash form.
##      $xmldoc - raw XML::LibXML::Document object reference
##    
sub query_asup_content {
    my $self = shift;
    my %options = (
        uri       => undef,
        timeout   => $default_timeout,
        @_
    );

    return $self->_request( uri      => $options{uri}, 
                            timeout  => $options{timeout} );
}

## @name _request()
## @description Private method to query the REST interface for ASUP information.
##              End users should use one of the public wrapper functions listed above.
##
## @synopsis
##   $AsupRest->_request( node     => $node,
##                        asup_id  => $asup_unique_id,
##                        list     => $list_content_links,
##                        uri      => $request_uri_for_content,
##                        gen_date => $asup_generation_date,
##                        subject  => $asup_subject_string,
##                        seq_num  => $asup_sequence_number,
##                        timeout  => $timeout_for_request );
##
## NOTE: no parameter checking is done in this private method, all checking is done
##       in wrapper methods that call this method.
## @arg node     Optional: The node for which the ASUP was generated.
## @arg gen_date Optional: The date and time the ASUP was generated.  Can be built from
##                         the "event log show" for the callhome.xxx event for that ASUP,
##                         or -last-time-sent from "autosupport show" (presuming that the
##                         ASUP was the last one sent out)  Or most definitively from the
##                         header.txt file from the ASUP itself. 
## @arg subject  Optional: The subject text associated with the ASUP.
## @arg seq_num  Optional: The ASUP sequence number.
## @arg uri      Optional: The universal resource identifier corresponding to a REST query.
## @arg asup_id  Optional: The unique id of an ASUP in the ASUPDW3 system.
## @arg list     Optional: Should ASUP content links be listed in the response?
## @arg timeout  Optional: The maximum amount of time (in seconds) to wait until the REST interface
##                         responds.
## @return
##    When called in scalar context, return is $status which is 
##      $status - one of:
##        "REQ_FAILED"      - REST interface failed to respond to request
##        "RES_SUCCESS"     - REST interface responded to request with SUCCESS
##        "RES_IN_PROGRESS" - REST interface responded that the data being requested
##                            is still being processed.
##        "RES_FAILED"      - REST interface responded to request with a FAILURE
##    
##    When called in an array context, return is ($status, $href, $xmldoc) which is 
##      $status - same as above:
##      $href   - xml response from REST interface converted to simplified hash form.
##      $xmldoc - raw XML::LibXML::Document object reference
##    
sub _request {
    my $self = shift;
    my %options = (
        node     => undef,
        gen_date => undef,
        subject  => undef,
        seq_num  => undef,
        asup_id  => undef,
        list     => undef,
        uri      => undef,
        timeout  => undef,
        @_
    );

    my $req_uri;
    my %hash;
    my $hp = \%hash;

    # possible values REQ_FAILED, RES_FAILED, RES_SUCCESS, RES_IN_PROGRESS
    my $status = "REQ_FAILED";

    # No parameter checking is going to be done in this private routine.
    # All parameter checking is assumed to be done in the public wrapper
    # routines.

    # Either use the uri the user sent us, or build it from the options hash
    if ( defined $options{uri} ) {
        $req_uri =  $options{uri};
    } else {
        $req_uri =  _build_uri( %options );
    }

    if ( $options{timeout} ) {
        $req_uri = $req_uri.'/timeout/'.$options{timeout};
    }

    # assume we are going to make the request in the foreground
    my $ua = LWP::UserAgent->new;

    # timeout option in will be passed in the REST request, so we need to allow for
    # a little more time on the HTTP request itself.
    $ua->timeout( $options{timeout} + 30 );

    my $res = $ua->get( $req_uri );

    # Check the outcome of the response
    my $xmlDoc;
    if ($res->is_success()) {
$Log->comment("Response = ".Dumper($res->content()) );
        my $parser = XML::LibXML->new();

# FWR will the XML just be a string accessible via content()?
        #$xmlDoc = $parser->parse_string( $res->content() );

        # FWR Hack until REST interface is ready
        #$xmlDoc = $parser->parse_file( '/u/radish/scripts/failedAsup.xml'  );
        #$xmlDoc = $parser->parse_file( '/u/radish/scripts/successfulAsup.xml'  );
        #$xmlDoc = $parser->parse_file( '/u/radish/scripts/inProcessAsup.xml'  );
        #$xmlDoc = $parser->parse_file( '/u/radish/scripts/multipleAsups.xml'  );
        #$xmlDoc = $parser->parse_file( '/u/radish/scripts/listResponse.xml'  );

        my @nodes = $xmlDoc->childNodes();
    
        #_build_hash_from_xml( xmlnode => $nodes[0], hash => $hp );
        DataUtils::build_hash_from_xml( xmlnode => $nodes[0], hash => $hp );

    } else {
        $Log->error("Response = ".Dumper($res->status_line()) );
    }

    $status = _aggregate_hash_status($hp);

    return wantarray ? ($status, $hp, $xmlDoc) : $status;
}

## @name extract_asup_ids()
## @description Public method to extract the "asup_id" tokens from anywhere in the
##              given hash.
##
## @synopsis
##   my @asup_ids = ASUP_REST::extract_asup_ids( hash => \%hash );
##
## @arg hash     Required: A reference to the hash returned by query_asup(), or list_asup_content()
##
sub extract_asup_ids {
    my %options = (
        hash     => undef,
        @_
    );

    if ( !$options{hash} || ((ref $options{hash}) !~ /HASH/i) ) {
        die script_result("ASUP_REST::extract_asup_ids() must be called with hash option and hash must be a hash reference");
    }

    my $hash = $options{hash};
    my @asup_ids;

    DataUtils::extract_field( hash => $hash, array => \@asup_ids, field => 'asup_id' );

    return @asup_ids;
}

## @name _aggregate_hash_status()
## @description Private method to extract and then aggregate all of the 'status_code'
##              entries from a given hash.
##
## @synopsis
##   my @asup_ids = ASUP_REST::aggregate_hash_status( hash => \%hash );
## @arg hash     Required: A reference to the hash returned by query_asup(), or list_asup_content()
##
## @return
##    Scalar $status which is one of:
##        "REQ_FAILED"      - REST interface failed to respond to request
##        "RES_SUCCESS"     - REST interface responded to request with SUCCESS
##        "RES_IN_PROGRESS" - REST interface responded that the data being requested
##                            is still being processed.
##        "RES_FAILED"      - REST interface responded to request with a FAILURE
##
sub _aggregate_hash_status {
    my %options = (
        hash     => undef,
        @_
    );

    if ( !$options{hash} || ((ref $options{hash}) !~ /HASH/i) ) {
        die script_result("ASUP_REST::_aggregate_hash_status() must be called with hash option and hash must be a hash reference");
    }
    my $hash = $options{hash};
    my @states;

    DataUtils::extract_field( hash => $hash, array => \@states, field => 'status_code' );

    # possible values REQ_FAILED, RES_FAILED, RES_SUCCESS, RES_IN_PROGRESS
    # first status is the request status
    my $counter = 0;
    foreach my $status (@states) {
        if ( ($counter == 0) && ($status ne 'SUCCESS') ) {
            return 'REQ_FAILED';
        }

        if ( $counter > 0 ) {
            if ( $status =~ /FAIL/ ) {
                return 'RES_FAILED';
            }

            if ( $status =~ /PROGRESS/ ) {
                return 'RES_IN_PROGRESS';
            }
        }

        $counter++; 
    }

    if ( $counter ) {
        return 'RES_SUCCESS';
    }

    return 'REQ_FAILED';
}

## @name convert_date()
## @description Public method to convert a DateTime string into a format
##              useable in REST requests.
##
## @arg date            Required: Date String
## @arg input_format    Optional: Strptime compatible date formatting string that matches input
##                                format.
## @arg output_format   Optional: Strptime compatible date formatting string that corresponds
##                                to format required by REST queries.
## @arg remove_timezone Optional: Remove timezone from DateTime string (Strptime has problems
##                                with 3 char timezones, they are apparently ambigous)
##
## @synopsis
##
##    convert_date( date => 'Thu Aug 12 00:16:02 EDT 2010' );
##
##
sub convert_date {
    my %options = (
        date            => undef,
        input_format    => '%a %b %d %T %Y',
        output_format   => $opt_href->{REST_DATE_FORMAT},
        remove_timezone => 1,
        @_
    );

    if ( !$options{date} ) {
        die script_result("ASUP_REST::convert_date() must be called with date option and date option must be formatted according to input_format option");
    }

    if ( !$options{input_format} ) {
        die script_result("ASUP_REST::convert_date() must be called with input_format option and input_format option must match format of date option");
    }

    if ( !$options{output_format} ) {
        die script_result("ASUP_REST::convert_date() must be called with output_format option and output_format option must match format expected by REST interface queries");
    }

    my $date          = $options{date};
    my $input_format  = $options{input_format};
    my $output_format = $options{output_format};

    if ( $options{remove_timezone} ) {
        # For default format, assume date string is like : 'Thu Aug 12 00:16:02 EDT 2010'
        my @date = split( /\s+/, $date );

        # Strptime chokes on the timezone (zero-relative 4th field) in date strings
        # like this one: 'Thu Aug 12 00:16:02 EDT 2010'
        # so we must remove it
        @date = (@date[0..3,5]);

        $date = join(" ", @date );
    }

    my $Strp = new DateTime::Format::Strptime( pattern => $input_format );

    my $dt = $Strp->parse_datetime($date);
    $Strp->pattern($output_format);
    my $new_format = $Strp->format_datetime($dt);

#print "Time in old format = |$date|\n";
#print "Time in new format = |$new_format|\n";

    return $new_format;
}


1;

=pod

=head1 NAME

ASUP_REST.pm - Interface to the NetApp ASUPDW3 REST interface


=head1 SYNOPSIS

    use ASUP_REST;
    my $AsupRest = ASUP_REST->new();

    # retrieve $raw_asup_generation_date        using calls outside of this library
    # optionally retrieve $asup_subject_string  using calls outside of this library
    # optionally retrieve $asup_sequence_number using calls outside of this library

    my $formatted_asup_generation_date = ASUP_REST::convert_date( date => $raw_asup_generation_date );
    my ($status, $hash, $xmldoc);
    my $wait_time = xxx; # some positive number of seconds
    my $retries   = yyy; # some positive number of retries
    my $num_fails = zzz; # some positive number of allowable failures

    while ( $retries && ($status != 'RES_SUCCESS') ) {
        ($status, $hash, $xmldoc) = $AsupRest->query_asup( node     => $node,
                                                           gen_date => $formatted_asup_generation_date,
                                                           subject  => $asup_subject_string,
                                                           seq_num  => $asup_sequence_number );

        if ( ($status eq 'RES_IN_PROGRESS') ||
             ($status eq 'RES_FAILED') ) {
            # would we need to retry here too?  I would imagine that soon after an asup was
            # issued, we might get some 'RES_FAILED's before we eventually get some 'RES_IN_PROGRESS's
            # and then eventually a 'RES_SUCCESS'

            $Log->comment( "ASUP_REST request resulted in $status" );

        } elsif ( $status eq 'REQ_FAILED' ) {

            if ( !$num_fails ) {
                $Log->error( die script_result( "ASUP_REST request failed" ) );
            }
            $num_fails--;
        }

        if ( $status != 'RES_SUCCESS') {
            Tharn::snooze( $wait_time );
            $retries--;
        }
    }

    if ( !$num_fails ) {
        $Log->error( die script_result( "ASUP_REST request timed out" ) );
    } else {
       my @asup_ids = ASUP_REST::extract_asup_ids( hash => $hash );
       foreach my $asup_id ( @asup_ids ) {
           my ($status, $content_hash ) = $AsupRest->list_asup_content( asup_id   => $asup_unique_id,
                                                                        recursive => 1 );

           if ( $status eq 'RES_SUCCESS' ) {
               my @content;
               DataUtils::extract_field( hash => $content_hash, array = \@content, field => "CONTENT" );
               foreach my $content (@content) {
                   $Log->comment('Content for block '.$content->{name}.':' );
                   $Log->comment('    '.$content->{data} );
               }
           } else {
               $Log->error("encountered an error");
           }
       }
    }


=head1 DESCRIPTION

This library provides an interface to the NetApp ASUPDW3 REST interface.
The methods provided allow you to:

    Query ASUPDW3 REST interface to get the unique REST ASUP identifier
    Query ASUPDW3 REST interface to get the list of content blocks stored for a specific ASUP.
    Query ASUPDW3 REST interface to get the stored content for a specific ASUP.

It also returns REST data in a simple hash format as well as an XML::LibXML::Document
object and provides utility methods for extracting data from the hash.
    
=head1 METHODS

Note: Instance methods which must be called on an object are listed as 
      $AsupRest->method() while static methods (those that do not need to be
      called with an object) are listed as ASUP_REST::method().

new( rest_url    =>'http://support-tools.corp.netapp.com', 
     rest_userid =>'rest_user');

        Description:
            Creates a new instance of the ASUP_REST interface object.
        Inputs:
            rest_url    => Optional: Base URL for REST interface server
            rest_userid => Optional: Userid to use for REST queries
        Outputs:
            ASUP_REST object instance
        Examples:
            <see synopsis above>
            
$AsupRest->query_asup( node     => $node,
                       gen_date => $asup_generation_date,
                       subject  => $asup_subject_string,
                       seq_num  => $asup_sequence_number,
                       timeout  => $timeout_for_request );

        Description:
            Query REST interface with filer-specific ASUP data to get back unique identifiers (asup_ids)
            to be used in subsequent REST queries.
        Inputs:
            node      => Node ASUP is associated with.
            gen_date  => Generation date/time in format "yyyy-mm-ddThh:mm:ss"
            subject   => Optional: Subject text of ASUP
            seq_num   => Optional: Sequence number of ASUP
            timeout   => Optional: Timeout for request (units: seconds, default 600)
        Outputs:
            When called in scalar context, return is $status which is
              one of:
                "REQ_FAILED"      - REST interface failed to respond to request
                "RES_SUCCESS"     - REST interface responded to request with SUCCESS
                "RES_IN_PROGRESS" - REST interface responded that the data being requested
                                    is still being processed.
                "RES_FAILED"      - REST interface responded to request with a FAILURE
        
            When called in an array context, return is ($status, $href, $xmldoc) which is
              $status - same as above:
              $href   - xml response from REST interface converted to simplified hash form.
              $xmldoc - raw XML::LibXML::Document object reference

        Examples:
            <see synopsis above>
            
$AsupRest->list_asup_content( asup_id   => $asup_unique_id,
                              recursive => $retrieve_all_content,
                              timeout   => $timeout_for_request );

        Description:
            Query REST interface with unique asup_ids to get a list of content blocks
            and optionally the content data itself.
        Inputs:
            asup_id   => REST unique ASUP identifier
            recursive => Optional: Retrieve all content data as well as list them? (default 0 (false))
            timeout   => Optional: Timeout for request (units: seconds, default 600)
        Outputs:
            When called in scalar context, return is $status which is
              one of:
                "REQ_FAILED"      - REST interface failed to respond to request
                "RES_SUCCESS"     - REST interface responded to request with SUCCESS
                "RES_IN_PROGRESS" - REST interface responded that the data being requested
                                    is still being processed.
                "RES_FAILED"      - REST interface responded to request with a FAILURE
        
            When called in an array context, return is ($status, $href, $xmldocs) which is
              $status   - same as above:
              $href     - xml response from REST interface converted to simplified hash form.
              $xmldocs - array of raw XML::LibXML::Document object references
        Examples:
            <see synopsis above>

$AsupRest->query_asup_content( uri     => $request_uri_for_content,
                               timeout => $timeout_for_request );

        Description:
            Method to query the REST interface for the content stored for a given ASUP content block.
        Inputs:
            uri       => URI returned via list_asup_content() that identifies the
                         content block you want data for.
            timeout   => Optional: Timeout for request (units: seconds, default 600)
        Outputs:
            When called in scalar context, return is $status which is
              one of:
                "REQ_FAILED"      - REST interface failed to respond to request
                "RES_SUCCESS"     - REST interface responded to request with SUCCESS
                "RES_IN_PROGRESS" - REST interface responded that the data being requested
                                    is still being processed.
                "RES_FAILED"      - REST interface responded to request with a FAILURE
        
            When called in an array context, return is ($status, $href, $xmldoc) which is
              $status   - same as above:
              $href     - xml response from REST interface converted to simplified hash form.
              $xmldoc   - raw XML::LibXML::Document object reference
        Examples:
            <see synopsis above>
            
ASUP_REST::extract_asup_ids( hash => $hash );
        Description:
            Method to extract data from the passed in hash, no matter how deep,
            that correspond to a "asup_id" key.
        Inputs:
            hash   => Hash reference returned by query_asup(), list_asup_content() or query_asup_content().
        Outputs:
            @asup_ids - array of data corresponding to "asup_id" keys from the input hash.
        Examples:
            <see synopsis above>

ASUP_REST::convert_date( date            => 'Thu Aug 12 00:16:02 EDT 2010',
                         input_format    => '%a %b %d %T %Y',
                         output_format   => '%Y-%m-%dT%T',
                         remove_timezone => 1 );
        Description:
            Public method to convert a DateTime string into a format useable in REST requests.
        Inputs:
            date            Required: Date String
            input_format    Optional: Strptime compatible date formatting string that matches input
                                      format.
            output_format   Optional: Strptime compatible date formatting string that corresponds
                                      to format required by REST queries.
            remove_timezone Optional: Remove timezone from DateTime string (Strptime has problems
                                      with 3 char timezones, they are apparently ambigous)
        Outputs:
            $date - Date formatted according to output_format (default is "yyyy-mm-ddThh:mm:ss") 
        Examples:
            <see synopsis above>

=cut
