# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Admin/License.pm#1 $

## @summary Functions For Licensing Services
##
# Copyright (c) 2005-2010 NetApp, Inc.
# All Rights reserved
#
## @author dl-central-libs
## @status Shared
## @pod here

=head1 NAME

Admin::License - Functions For Licensing Services

=head1 SYNOPSIS

  use Admin::License qw(license);
  my $filer = host($FILER);
  my $conn  = connect('#rsh');

  # license nfs
  license($conn, "nfs") or die $@;

  # check the status of nfs license
  use Admin::License qw(is_licensed :constants);
  my $filer = host($FILER);
  my $conn  = connect('#rsh');
  if (my $rv = is_licensed($conn, 'nfs')) {
     if ($rv & LICENSE_DEMO) {
        print "This is a demo license\n";
     }
  }

=head1 DESCRIPTION

Admin::License provides ways to tell the filer to license/unlicense
services. 

Validation will be done by default. This is done by looking at the
contents of the registry for the particular service. For services which 
need special licenses validation will not be done since they dont have
registry entries.

=head1 NEED ANOTHER LICENSE?

Check out Admin::License::Default

=head1 DEPENDENCIES

This module currently depends on Admin::License::Object for is_licensed()
function.

=head1 FUNCTIONS

=cut

package Admin::License;
use strict;
use warnings;
use Carp qw(croak confess);
use Params::Validate qw(validate_pos SCALAR BOOLEAN OBJECT ARRAYREF);
use Tharn qw(say host connect script_result logcomment);
use gx_licenses qw(map_v1_to_v2 get_licenses);

BEGIN
{
    Params::Validate::validation_options(
        on_fail => sub {my @msgs = @_; chomp(@msgs);
                confess(script_result(@msgs))}
        );
}
use Module::Common qw(validate_mixed);
use Filer::Version qw(are_we_ontapng
		      get_wafl_version
		      WV_IRONCITY_MIN);
BEGIN
{
	Module::Common::setup_export_tags(
		license => [ qw(
			license
			license_raw
			get_license_code
			generate_license
			unlicense
			unlicense_raw
			is_licensed
		) ],
		constants => [ qw(
			LICENSE_OK
			LICENSE_DEMO
			LICENSE_SITE
                        LICENSE_INCLUDED				  
		) ]
	);
}
use constant LICENSE_OK   => 1;
use constant LICENSE_DEMO => 2;
use constant LICENSE_SITE => 4;
use constant LICENSE_INCLUDED => 8;
use SayWrap qw(sayEOF);
use ModelInfo;
use Admin::License::Object;
use Admin::License::Default;

# Validation will be skipped for services that need special licenses
our $Validation = 1;

=head2 license($conn,$service)

Given a Connectrec and a service name, uses the 'license' command
to install a new license.  On success, returns true. Otherwise
returns undef and sets $@.

$service may be a single service name such as 'nfs', or it could be a
list of services.

   license($conn, 'nfs');
   # or
   license($conn, [ qw( nfs cifs http ) ]);

The intent of this function is to guarantee a license is installed
after calling it -- that is, it won't fail just because a license
was already installed, or because the license you specified was
mutually exclusive with another, so you don't have to do extra
work just to get your license installed.

This is actually implemented on top of license_raw() (which fails
if the license is already installed). Users testing the actual
license command should probably use license_raw() instead.

This function does NOT verify what kind of license is installed
already, so for long running scritpts, you may want to make sure
that the license is not a demo license. You may be able to use
the functionality of Admin::License::Object for that.

NOTE: when licensing services that may have mutually exclusive
relationships with other licenses (such as snaplock and snaplock_enterprise),
this function will automatically unlicense the other one.

=cut

sub license
{
	my($conn, $service) = validate_pos(@_,
		{ type => OBJECT, isa => 'Connectrec' },
		{ type => SCALAR | ARRAYREF }
	);

	# no licensing in Tricky yet
    #if (are_we_ontapng($_[0])) {
    #	return 1;
    #}

	return eval {
		my @services = ref($service) ? @$service : ($service);
		foreach my $s (@services) {
			my($license, $special) = get_license_code($conn, $s);
			if (!$license) {
				die "Unknown service '$s'";
			}

			# skip for services with special licenses (e.g. nfs on netcache)
			# since they dont have entries in the registry anyway
			if (!$special) {
				if (is_licensed($conn, $s)) {
					next;
				} elsif ($@) {
					die $@;
				}
			}

			# XXX - handle mutually exclusive licenses
			my $mutex_licenses =
				$Admin::License::Default::MUTUALLY_EXCLUSIVE{ $s };
			if ($mutex_licenses && ref($mutex_licenses) eq 'ARRAY') {
				foreach my $ml (@$mutex_licenses) {
					if ( get_wafl_version($conn->hostp) >=  WV_IRONCITY_MIN) {
						# To allow tests to run on "SnapVault Melding"
						# configuration in IC.0 or later.
						if (($ml eq 'sv_ontap_pri') || ($ml eq 'sv_ontap_sec')) {
							next;
				                }
					}
					if (!unlicense($conn, $ml)) {
						die $@;
					}
				}
			}


			if (ref($license) eq 'ARRAY') {
				#oh, no! its a bunch of licenses that we should try
				foreach my $code (@{$license}) {
					# try each license without caring about success
					if (license_raw($conn, $code, 
							special => $special, service => $s)) {
						last;
					}
				}
			} else {
				#just a normal license
				license_raw($conn, $license, special => $special, service => $s)
					or die $@;
			}
			if ($Validation) {
				if (!$special) {
					if (!is_licensed($conn, $s)) {
						die "Validation failed: license for $s doesn't seem to ".
							"be enabled after adding license\n";
					}
				} else {
					# skip for services with special licenses (e.g. nfs on netcache)
					# since they dont have entries in the registry anyway
					logcomment("Skipping Validation for special licenses");
				}
			}
		}
		return 1;
	};
}

# XXX - this was ripped right out of filer.pm. If there's a better
# way to do it, we should switch to it, rather than relying on the
# output of sysconfig
# model can be 'filer', 'netcache', 'nearstore' , 'gateway' or 'sa_appliance'
sub _guess_model
{
	my ($conn) = validate_pos(@_, { isa => 'Connectrec' });
	return eval {
		# first off, if it's a simulator, we can't rely on sysconfig,
		# so we need to check via the hosttype
		if ($conn->hostp->hosttype() =~ /(filer|cache)-simulator/) {
			if ($1 eq 'cache') {
				return 'netcache';
			} else {
				# check if the filer simulator is promoted or not 
				# one way is to see if test privilege is available or not
				my $out = sayEOF($conn, "priv set -q test");
				if ($out =~ /priv: unrecognized privilege: test/i) {
					return 'promoted_simulator';
				} else {
					return 'filer';
				}
			}
		}

		# if it looks like hardware,
		# get the model from MODEL param using ModelInfo.pm
		# if that does not work then try sysconfig
		my %params = $conn->{hostp}->param_fields();
		my $mi = new ModelInfo($params{MODEL});
		return($mi->{appliance}) if ($mi->{appliance});

                # try sysconfig. If the appliance has been taken over
                # by its cluster partner, try partner sysconfig
                my $text = sayEOF($conn, "sysconfig -v 0");

                my($model) = ($text =~ /^\s+Model Name:\s+(\w+)/sm);
                if (!$model) {
                    if($text !~ /.*?This system has been taken over by .*/i) {
                        croak "Failed to parse result of 'sysconfig -v 0' while guessing ".
                          "the model of appliance";
                    } else {
                        $text = sayEOF($conn, "partner sysconfig -v 0");
                        $model = ($text =~ /^\s+Model Name:\s+(\w+)/sm);
                        if (!$model) {
                            croak "Failed to parse result of 'partner sysconfig -v 0' while guessing ".
                              "the model of taken over appliance in cluster configuration";
                        }
                    }
                }

		# now guess what the product line is
		if ($model =~ /^N/i) {
			# this is an IBM model so we need to look at the Machine Type
			# that will tell us if its a filer(A) or gateway(G) and we 
			# reuse the same licenses
			my($machine_type) = ($text =~ /^\s+Machine Type:\s+IBM-\d\d\d\d-(A|G|C)\d\d/sm);
			if (!$machine_type) {
				croak "Failed to parse result of 'sysconfig -v 0' while ".
						"guessing the model of an IBM appliance";
			}
			return $model =~ /G/i ? "gateway" : "filer";
		} else {
			# this is a NETAPP model
			return
				$model =~ /^R/i ? "nearstore" :
				$model =~ /^G/i ? "gateway" :
				$model =~ /^C/i ? "netcache" :
				$model =~ /^SA/i ? "sa_appliance" :
				"filer";
		}
	}
}

=head2 license_raw($conn,$license[,%opts])

Given a Connectrec and a license code, calls the 'license' command
to install a new license. 

This function uses the "license help" command to check if the kernel
wants the old syntax or the new syntax ("license service=code" or
"license add code"). If the kernel only supports the old syntax,
it will return undef unless you pass the service name via %opts:

  license_raw($conn,$license,service => $service);

If the kernel supports the new syntax, the $service option is ignored.

You may optionally specify the "special" parameter to toggle the
use of the -Special argument to license command, which may be
necessary in some cases (such as enabling NFS on a NetCache)

=cut

sub license_raw
{
	my($conn, $license, %args) = validate_mixed(
		@_,
		[
			{ type => OBJECT, isa => 'Connectrec' },
			{ type => SCALAR, regex => qr(^[a-zA-Z]{7,28}$) }
		],
		{
			special => { type => BOOLEAN, optional => 1 },
			service => { type => SCALAR,  optional => 1 }
		}
	);

	return eval {
		# XXX - support older kernels with difference
		# "license" syntax. (this and other places may need
		# to be tweaked even more -- this is just a starting
		# point...)

		# this is a hack, but a clever hack anyway.
		# by looking at the help, we can tell if this
		# is a new style license, or old style
		# use advanced mode or it does not work for netcache
		my $text = sayEOF($conn, "priv set -q advanced; license help");
		my $old_style = $text !~ /license add/;

		if ($old_style && !$args{service}) {
			die "This kernel requires a service name to install the license, ".
				"but license_raw() did not receive one";
	    }

        my $is_v2 = 0;

        if ( _can_v2_license($conn) ) {
            $args{service} = map_v1_to_v2( $args{service} );
            if ( !$args{service} ) {
                return 1;
            }

            $is_v2 = 1;
        }

		my $sendstr = $old_style ?
			( $args{special} ?
				"priv set -q advanced; license -Special $args{service}=$license" :
				"priv set -q advanced; license $args{service}=$license" ) :
			    ( $args{special} ?
				"priv set -q advanced; license -Special add $license" :
				"priv set -q advanced; license add $license" );

        if ( !$is_v2 ) {
		    # unfortunately some licenses ask for user input (ugh)
		    # we're going to have to deal with these individually
		    if ($args{service} eq 'disk_sanitization') {
                my $error = "Did not get confirmation that license was installed";
                while (1) {
                    my (undef, undef, undef, $prompt) = 
                    say($conn, $sendstr, 
                        '-prompt' => ['license has been installed',
                                      'This Disk Sanitize license must be added from the console.',
                                      'wish to continue' ] );
                    if ($prompt =~ /license has been installed/) {
                        $error = undef;
                    }
                    elsif ($prompt =~ /This Disk Sanitize license must be added from the console./) {
                        $error = "Must use console connection to license disk_sanitization";
                        last;
                    }
                    elsif ($prompt =~ /wish to continue/) {
                        $conn->send("y\r");
                        $sendstr='';
                        next;
                    }
                    else {
                        #must have been a normal filer prompt or EOF
                        last;
                    }
                }
                if ($error) {
                    die $error;
                }

                return 1;
		    } 
        } 

		$text = sayEOF( $conn, $sendstr );
		# XXX - this test may not be good for old syntax. if it breaks,
		# please fix.
		# on success in old style "license enabled" is returned
		# on success in new style "license has been installed" is returned
		# for RR we have new outputs for the $0 license handling
		# those will return 'no license code is necessary'
        # see also http://web.netapp.com/engineering/projects/License/rr-license-model.html 

        # 7 mode v2.0 reports successfull addition as:
        # 'successfully added license key'

        if ($text !~ /(successfully added license key|license has been installed|license enabled|is already installed|no need to install|no license code is necessary|not available in C-mode|not available in partner context)/) {

            die "Did not get confirmation that license was installed\n$text";
		}

		return 1;
	}
}

=head2 is_licensed($conn,$service)

Given a Connectrec and a service name, checks to see if the corresponding
license is installed in the host. On success, it returns a bit-masked
value that you can check against to find out if the license is
a site license, a demo license, or a regular license.

  use Admin::License qw(is_licensed :constants);

  my $rv = is_licensed($conn, 'snaprestore');
  if (!$rv) { # or you can do !($rv & LICENSE_OK)
    print "snaprestore not licensed :(\n";
  } else {
    print "snaprestore is licensed! ";
    if ($rv & LICENSE_DEMO) {
      print "... and it's a demo license!";
    }
    if ($rv & LICENSE_SITE) {
      print "it's a site license!\n";
    }
    print "\n";
  }

If the license isn't installed, or an error is encountered, returns
undef and may set $@.

Dont use this routine to check for services that need a special license.
(e.g. nfs on netcache) It will not work since registry entries are 
not created for such services.

=cut

sub is_licensed
{
	my($conn, $service) = validate_pos(@_,
		{ type => OBJECT, isa => 'Connectrec' },
		{ type => SCALAR }
	);

    # We need to do differently if we are dealing with a v2 license
    # The Object.pm stores a lot of info about each service, but since
    # the v2 registry entries wont be populated, we canoot use Object
    # Class. Once the ONTAP supports v2 entries, the Object.pm can be
    # used directly. Until then the following is a quick hack to get this
    # going!
    if ( _can_v2_license($conn) ){ 
        # Check if a v1 service can be mapped to a v2 package. If not we
        # should consider this to be licensed. We do not have the entire
        # list of all v1 services (possibly get it frm GNOTE?). So we can
        # not determine if a v1 service is a valid string. But we will
        # assume whatever was supplied to $service is a valid v1 service
        # So if we are unable to map, we will set the bits and return the
        # bitmask.
        my $pkg = map_v1_to_v2( $service );
        if ( $pkg ) {
            my $out = sayEOF( $conn, "priv set -q advanced;license show -package $pkg" );
            if ( $out !~ /partner licensing is not available/ ) {
                if ( $out =~ /(No licenses available to show|No licenses are installed)/ ) {
                    return undef;
                }
            }
        }
        my $bm = 0;
        $bm |= LICENSE_SITE;
        $bm |= LICENSE_OK;
        $bm |= LICENSE_INCLUDED; 

        return $bm;
    }

	return eval {
		my $lo = Admin::License::Object->new(
			host => $conn->hostp,
			service => $service
		);
		$lo->refresh_data($conn);

		if (!$lo->licensed()) {
			return undef;
		}
		my $bitmask = 0;

		if ($lo->site()) {
			$bitmask |= LICENSE_SITE;
		}
		if ($lo->demo()) {
			$bitmask |= LICENSE_DEMO;
		}
		if ($lo->license_status =~ /ENABLED/) {
		    $bitmask |= LICENSE_INCLUDED;
		}

		$bitmask |= LICENSE_OK;

		return $bitmask;
	}
}

=head2 get_license_code($conn,$service)

Given a Connectrec and a service name, returns the corresponding license
string. If the corresponding license for the service is not available,
returns undef and sets $@.

Because at the time of writing we do not know of a reliable way to
tell the difference between host appliance types (especially when
things like simulators start to muddle the picture), we do the best
we can to guess what kind of appliance we have, by looking at the 
MODEL param or sometimes by issueing a sysconfig -v command. 

In scalar context this function only returns the license string, but
in array context a boolean flag "special" is also returned, so you
can pass it to license_raw():

  # scalar context
  my $license = get_license_code($host, $service);

  # array context
  my ($license, $special) = get_license_code($host, $service);
  license_raw($conn, $license, special => $special);

v2 licenses are package based and we map a v1 service to a
package. If we can't map (and it is a valid v1 service string),
we will return a 'licensed' string

=cut

sub get_license_code
{
	my($conn, $service) = validate_pos(@_,
		{ type => OBJECT, isa => 'Connectrec' },
		{ type => SCALAR }
	);

	my $model =  _guess_model($conn);
	die $@ if $@;
	# mental note: $model => 'filer' and 'netcache'
	# also includes simulators(filer/cache) and promoted_simulator for filers

	my $t = $Admin::License::Default::DEFAULT_LICENSES{$model};
	if (!$t) {
		# XXX - hmmm. exception or return
		croak("No such model $model");
		return undef;
	}

    my ($code, $special);

	if ($model eq 'netcache') {
		$code = $t->{$service};
		if (!$code) {
			my $t_filer = $Admin::License::Default::DEFAULT_LICENSES{filer};
			if ($code = $t_filer->{$service}) {
				$special = 1;
			}
        }
	} else {
        # Check if we are doing version 2 licensing and return the license
        # code after mapping a v1 service to a v2 package
        if ( _can_v2_license($conn) ){
            my $pkg = map_v1_to_v2( $service );
            # We assume that the $service is a valid v1 service. We do not
            # check it is a valid v1 service as we do not have the list in
            # its entirety. If we can't map, then we will interpret it as
            # already licensed. So we will return a special string 'licensed'
            # that we will be handled by the caller of this function
            if ( $pkg ) {
                $code = get_licenses( version => 2, features => ["$pkg"] );
            } else {
                $code = 'licensed';
            }
        } else {
	 	    $code = $t->{$service};
	    }
    }

	return wantarray ? ($code, $special) : $code;
}

=head2 unlicense($conn,$service)

Given a Connectrec and a service name, uses the 'license' command
to remove the corresponding license.  On success, returns true. Otherwise
returns undef and sets $@.

This is actually implemented on top of ulicense_raw() (which fails
if the license is not already installed) so that it does not issue a
license command if the service is not already installed. This way
it's much more useful for the general community.

Users testing the actual license command should probably use the
unlicense_raw() instead

=cut

sub unlicense
{
	my($conn, $service) = validate_pos(@_,
		{ type => OBJECT, isa => 'Connectrec' },
		{ type => SCALAR }
	);

	return eval {
		# check if the license is special or not 
		my($license, $special) = get_license_code($conn, $service);
		if (!$license) {
			die "Unknown service '$service'";
		}

		# skip checking the registry for services which need special licenses
		# since registry entries are not created for them
		if (!$special) {
			if ($license eq 'licensed' || !is_licensed($conn, $service)) {
				return 1;
			}
			if ($@) {
				die $@;
			}
		}

		if (!unlicense_raw($conn, $service)) {
			die $@;
		}

		if ($Validation) {
			if (!$special) {
			        # only fail if it wasn't auto-enabled since those can't
			        # be deleted
			        my $rv = is_licensed($conn, $service);
				if ($rv && (! ($rv & LICENSE_INCLUDED))) {
					die "Validation failed: license for $service still seem to ".
						"be enabled after removing license\n";
				}
			} else {
				# skip for services with special licenses (e.g. nfs on netcache)
				# since they dont have entries in the registry anyway
				logcomment("Skipping Validation for special licenses");
			}
		}
		return 1;
	};
}

=head2 unlicense_raw($conn,$service)

Given a Connectrec and a service name, calls the 'license' command
to delete a new license. 

=cut

sub unlicense_raw
{
	my($conn, $service) = validate_pos(@_,
		{ type => OBJECT, isa => 'Connectrec' },
		{ type => SCALAR }
	);

	return eval {
		# check if the license is special or not 
		my($license, $special) = get_license_code($conn, $service);
		if (!$license) {
			die "Unknown service '$service'";
		} elsif ( $license eq 'licensed' ) {
            return 1;
        }

		# this is a hack, but a clever hack anyway.
		# by looking at the help, we can tell if this
		# is a new style license, or old style
		# use advanced mode or it does not work for netcache
		my $text = sayEOF($conn, "priv set -q advanced; license help");
		my $old_style = $text !~ /license add/;

        if ( _can_v2_license($conn) ) {
            $service = map_v1_to_v2( $service );
            if ( !$service ) {
                return 1;
            } 
        }
		my $sendstr = $old_style ?
			( $special ?
				"priv set -q advanced; license -Special $service=disable" :
				"priv set -q advanced; license $service=disable" ) :
			( $special ?
				"priv set -q advanced; license -Special delete $service" :
				"priv set -q advanced; license delete $service" );

		# XXX - some licenses cannot be unlicensed e.g. disk_sanitization
		# should we handle those ???

		# for RR many $0 licenses cannot be removed, those are 
		# identified as ENABLED but the command return will say
		# 'cannot be deleted' its ok to ignore these, although
		# it might have been better to test the status return
		# instead of screen scraping the output
                # see also http://web.netapp.com/engineering/projects/License/rr-license-model.html
		$text = sayEOF($conn, $sendstr);
		my $service_re = quotemeta($service);
		if ($text !~ /(unlicensed $service_re|cannot be deleted|successfully deleted)/) {
			die "Failed to unlicense $service: $text\n";
		}
		return 1;
	};
}

=head2 generate_license($service[,%opts])

NOTE: This function creates a new license on the fly, but unless you
explicitly need to test such licenses, B<YOU PROBABLY SHOULD BE USING
license() or get_license() INSTEAD>.

Generates a license for $service, useing the C<secret_code> command.

The command name for secret_code is determined in the following
order: use the C<secret_code> argument, create from the C<build>,
or create from C<$ENV{ONTAPDIR}>.

%opts may contain the following:

=over 4

=item host

The name of the host in which to execute the secret_code. Defaults to
localhost.

=item id

The site id to be passed to secret_code. If not specified, a random
number between 0 and 99999 will be used.

=item days

The number of days until this license expires. This option is
ignored if the nodes argument is also specified

=item nodes

The number of nodes that may be licensed. This option takes
precedend when the days argument is also specified

=item secret_code

The absolute path to the command to be used to generate the license key.
This option takes precedence over the build option and the ONTAPDIR
environment variable

=item build

The absolute path to the root of the ontap build where the secret_code
is. For example, to use the secret_code found under /rlse/mainN,
write it as:

   generate_license('nfs', build => '/rlse/mainN' );

This takes precedenc over ONTAPDIR, but is ignore if the secret_code
option is set.

=back

=cut

# 6/2010 advanced warning, this will always fail going forward
# since the generator will be encrypted in all of the builds
# shortly and only MFG will have the decryption key
# the recommendation is that this functionality be removed from use
# contact sdy@netapp.com or leonard@netapp.com for further information
sub generate_license
{
	my ($service, %args) = validate_mixed(
		@_,
		[ { type => SCALAR } ],
		{
			host    => {
				type     => SCALAR,
				optional => 1,
			},
			id      => {
				type     => SCALAR,
				optional => 1,
				default  => int(rand(100000))
			},
			days    => {
				type     => SCALAR,
				optional => 1,
				default  => 1
			},
			nodes   => {
				type     => SCALAR,
				optional => 1,
				default  => 0
			},
			secret_code => {
				type => SCALAR,
				optional => 1
			},
			build => {
				type => SCALAR,
				optional => 1
			}
		}
	);

	return eval {
		my $secret_code = $args{secret_code};
		if (!$secret_code) {
			$args{build} ||= $ENV{ONTAPDIR};
			$secret_code = File::Spec->rel2abs(
				File::Spec->catfile(
					$args{build},
					'releases',
					'unix',
					'sun4-sunos5',
					'secret_code'
				)
			);
		}

		if (!-x $secret_code) {
			die "$secret_code is not found or is not executable";
		}
	
		my @cmd = ($secret_code);
		# -e and -n can't be specified together...
		if ($args{nodes} > 0) {
			push @cmd, ("-n", $args{nodes});
		} elsif ($args{days} > 0) {
			push @cmd, ("-e", $args{days});
		}
		push @cmd, $args{id};
		push @cmd, $service;

		my $cmd = "@cmd";
		my $output;
		# if we're on a solaris host, we might as well run it locally
		if (!$args{host} && qx|uname| =~ /SunOS/) {
			$output = qx($cmd);
		} else {
			my $host = host($args{host} || "cranford");
			my $conn = connect('#rsh');

			$output = sayEOF($conn, $cmd);
		}
	
		my $service_re = quotemeta($service);
		my $re      = qr{^$service=(.+)$};
		my $license = '';
		my $special = 0;
	
		foreach my $line (split(/(?:\r+)?\n+/, $output)) {
			$line =~ s/^\s+//;
			$line =~ s/\s+$//;
			if ($line =~ /$re/) {
				$license = $1;
				last;
			}
		}
	
		return $license;
	};
}

# Private function that can check if a filer accepts a v2 license
# We use a package scope variable 'LicenseVersion' to store this 
# information
sub _can_v2_license {
    my ($conn) = validate_pos(@_,
        { type => OBJECT, isa => 'Connectrec' }
    );

    my $help = sayEOF($conn, "priv set -q advanced; license help");
    ( $help =~ /license v2 code/ ) ? return 1 : return 0; 
}

1;

=head1 SEE ALSO

Admin::License::Object, Admin::License::Default

=head1 BUGS

    Does not work on Krakus since "count" item in registry did not exist in Krakus.
    OnTap-NG does not have any licenses now so it just returns
    Use of the Zephyr API 'license' is not supported

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = license, sysconfig

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    dl-central-libs
	$Id: //depot/prod/test/Rsierranevada/lib/Admin/License.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005-2010, NetApp Inc., All Rights Reserved.

=cut
