# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Admin/License/Object.pm#2 $

## @summary Object Interface To Single License
## @author dl-central-libs
## @status public
## @pod here

=head1 NAME

Admin::License::Object - License Object

=head1 SYNOPSIS

  use Admin::License::Object;
  my $lo = Admin::License::Object->new(service => 'nfs');
  $lo->refresh_data($conn);

  $lo->service;
  $lo->count;
  $lo->expiration;
  $lo->installation;
  $lo->length;
  $lo->nodes;
  $lo->platform;

  $lo->demo;
  $lo->site;
  $lo->expired;
  $lo->licensed;
  $lo->license_status;

=head1 DESCRIPTION

Admin::License::Object is a simple data storage for information about a
single license. The information that is stored in this object is exactly
what is listed under status.license.$service.* in the registry and 
'license_status' is populated with another registry command  
'registry.walk.options.license', this registry check is made to see if the 
service is licensed in the first place. This extra checking is done since 
the first registry walk is not reliable enough as per burt 174643.

An Admin::License::Object can be created and populated in one of the
following ways:

1. Associate a host to the object

  my $host = host($FILER);
  my $lo = Admin::License::Object->new(host => $host, service => 'nfs' );
  $lo->refresh_data();  # fetch data from $host

2. Pass a connection when needed

  my $host = host($FILER);
  my $conn = connect('#rsh');
  my $lo = Admin::License::Object->new(service => 'snaplock');
  $log->refresh_data($conn); # fetch data from $conn

3. Populate on your own:

  my $lo = Admin::License::Object->new(service => 'http');
  $lo->count(3);
  $lo->demo(1);

=cut

package Admin::License::Object;
use strict;
use Class::MethodMaker
	new_with_init => 'new',
	get_set => [ qw( service count expiration  installation length nodes 
                         platform license_status) ],
	boolean => [ qw(demo site expired) ],
	object  => [
		Hostrec => 'host'
	]
;
use Params::Validate qw(validate_pos);
use Carp qw(croak confess);
use Tharn qw(connect script_result);
BEGIN
{
    Params::Validate::validation_options(
        on_fail => sub {my @msgs = @_; chomp(@msgs);
                confess(script_result(@msgs))}
        );
}
use SayWrap qw( sayEOF );

sub init
{
	my ($self, %args) = @_;

	foreach my $field (qw( host service count demo expiration expired
	  installation length nodes platform site license_status )) {

		if (exists $args{$field}) {
			$self->$field( $args{ $field } );
		}
	}
}

=head2 $lo-E<gt>refresh_data([$conn])

Issues the appropriate command(s) to the appliance (either the one
in $lo-E<gt>host(), or the connection provided via $conn), and populates
the object with the new information.

This looks at the registry values, so you need to be able to issue
the command using 'priv set advanced'.

Throws an exception upon encountering an error.

=cut

sub refresh_data
{
	my $self = shift;

	my ($conn) = validate_pos(@_, { isa => 'Connectrec', optional=>1 });

	if(!$conn) {
        croak("Specify either conn or host") if (!$self->{host});
		$conn = connect($self->host->{id} . '#rsh');
	}

	my $service = $self->service();
	if (!$service) {
		croak( "Can't refresh without service name" );
	}
	my $text = sayEOF(
		   $conn,
	           "priv set -q advanced; registry walk options.license.$service");
        
	if ($text !~ /options.license.$service=(\w+)/) {
		croak( "Failed to parse output of registry walk options.license.$service" );
	}
	$self->license_status($1);
	
        #license is not turned on, so don't collect other information
	
        if ($self->license_status ne 'off') {		
		$text = sayEOF(
			$conn,
			"priv set -q advanced; registry walk status.license.$service");
	
		# XXX - start parsing. after the parse, if we don't have all
		# of the values, then we failed to parse, so die.

		foreach my $field (qw( demo expired site )) {
			if ($text !~ /^status\.license\.$service\.$field=([ \w\d]+)/m) {
				croak( "Failed to parse output of registry walk for $service" );
			}
			$self->$field( $1 eq 'true' )
		}
		foreach my $field (qw( count expiration installation length
	  		nodes platform)) {

			if ($text !~ /^status\.license\.$service\.$field=([ \w\d]+)/m) {
				croak( "Failed to parse output of registry walk for $service" );
			}

			$self->$field( $1 );
		}
	} else {
		return(1);
	}	
}

=head2 $lo-E<gt>licensed()

Returns true if it looks like *some* license was installed, and that it
hasn't expired yet, false otherwise.

This function does not check what kind of license was installed.

=cut

sub licensed
{
	my $self = shift;
	return ($self->license_status()  ne 'off' &&
	  $self->installation() ne 'not installed' &&
	  $self->installation() !~ /disabled/ &&
	  !$self->expired());
}

1;

=head2 $lo-E<gt>service([$service])

Set or get the service name

=head2 $lo-E<gt>count([$count])

Set or get the number of times this license has been installed.

=head2 $lo-E<gt>expiration([$expiration])

Set or get the expiration date

=head2 $lo-E<gt>installation([$installation])

Set or get the installation date. If uninstalled, the string
"not installed" is used (not undef)

=head2 $lo-E<gt>length([$length])

Set or get the length that this license is good for, if this
is a demo license.

=head2 $lo-E<gt>nodes([$nodes])

Set or get the number of nodes this license enables.

=head2 $lo-E<gt>platform([$platform])

Set or get the platform that this license is good for, if applicable.

=head2 $lo-E<gt>demo([$bool])

Set or get if this license is a demo license or not.

=head2 $lo-E<gt>site([$bool])

Set or get if this license is a site license or not.

=head2 $lo-E<gt>expired([$bool])

Set or get if this license is expired or not

=head2 $lo-E<gt>license_status([$bool])

Set or get if the license is on / off from registry walk of options

=head1 SEE ALSO

Admin::License

=head1 STATUS

    Shared

=head1 BUGS

    Does not work on Krakus since "count" item in registry did not exist in Krakus.

=head1 ATTRIBUTES

    ONTAP_COMMANDS = registry

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    dl-central-libs
	$Id: //depot/prod/test/Rsierranevada/lib/Admin/License/Object.pm#2 $
    Last revised $Date: 2013/04/24 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut
