# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/Aggregate.pm#2 $
## @summary Library of aggregate commands
## @author dl-johny@netapp.com
##         (FSS WAFL Quality)
## @cf     false
## @status public
## @pod here

# Submitted into Perforce via burt103435 - Reorganize Test Libraries

package Aggregate::StatusObject;
use Class::MethodMaker
    new_hash_init => 'new',
    get_set       => [ qw(status type state aggr) ],
    hash          => [ qw(options) ];

package Aggregate;
use strict;
use Tharn;
use SPARSE::Util::Logger qw(logTrc logDbg logInfo logWarn logError logStep);
use Utils::Units qw(:constants);
use Data::Dumper;
use Module::Common qw(validate_mixed);
use Aggregate::ImplTemplate;
use Filer::Version qw(get_wafl_version supported_feature UNSUPPORTED_FEATURE_EXCEPTION are_we_ontapng);
use Carp qw(croak);
use misc qw(cli);
# use with aggr_disk_free() command.
use constant DF_SPACE   => 'space';
use constant DF_INODES  => 'inodes';
use constant DF_PRIVATE  => 'private inodes';
use constant DF_RESERVE => 'reserve';
our($Impl, $Validation, $VERSION);
my $Ssd_Pool_Retry = 0;

BEGIN
{
	$Validation = 1;
	$VERSION    = '0.01';

	my @df_space_types  = qw( DF_SPACE DF_INODES DF_PRIVATE DF_RESERVE );
	my @aggr_subs       = qw(
		aggr_add
		aggr_create
		aggr_destroy
		aggr_undestroy
		aggr_disks
		aggr_exists
		aggr_list
		aggr_meta_i2p
		aggr_mirror
		aggr_offline
		aggr_online
		aggr_options
		aggr_purge
		aggr_raidgroup_status
		aggr_disk_free
		aggr_read_fsid
		aggr_rename
		aggr_restrict
		aggr_rewrite_fsid
		aggr_root
		aggr_scrub
                aggr_snap_delta
		aggr_snap_reserve
		aggr_split
		aggr_spare_status
		aggr_status
		aggr_match_plex_raid_info
		aggr_wait_change
                aggr_blocks
		aggr_show_space
		aggr_show_space_hash
                aggr_blockcheck
                aggr_blockcheck_wait
                aggr_blockcheck_read_output
	);

	Module::Common::setup_export_tags(
		df  => [ 'aggr_disk_free', @df_space_types ],
		aggr => [ @aggr_subs ]
	);

}

use Params::Validate
	qw(validate_pos validate BOOLEAN SCALAR ARRAYREF HASHREF);
use SayWrap qw(sayEOF);
use Utils qw(make_conn_and_complain);
use RaidInfo qw(aggregate);
	# XXX - upon popular request, make_conn() shall be used to allow
	#       hostspec/Hostspec/Connectrec as the "filer" argument. But
	#       we want to make sure to complain if it's passed anything
	#       other than a Connectrec, so we will use
	#       make_conn_and_complain(), which will "complain loudly"
	#       if anything other than a Connectrec is passed.

=head1 NAME

Aggregate - Basic Aggregate Functions

=head1 SYNOPSIS

  use Aggregate qw(:all);
  # $conn = connect(...);

  $rv = aggr_create($conn, $aggr, [ $disk1, $disk2 ... ]);
  $rv = aggr_destroy($conn, $aggr,  { destroy_corrupt_volumes => ... });
  $rv = aggr_purge($conn, $aggr,  { destroy_corrupt_volumes => ... });
  $rv = aggr_meta_i2p($conn, $aggr);
  $rv = aggr_online($conn, $aggr);
  $rv = aggr_offline($conn, $aggr);
  $rv = aggr_restrict($conn, $aggr);
  $rv = aggr_rename($conn, $oldaggr, $newaggr);
  $rv = aggr_exists($conn, $aggr);
  $rv = aggr_wait_change($conn, $aggr, want => { status => qr(online) });
  $rv = aggr_scrub($conn, $aggr, $mode);
  # ($completed, $lastscrub) = aggr_scrub($conn, $aggr, 'status');

  # old style of getting info from aggr_status()
  ($state, $status, $vol, %options) = aggr_status($conn, $aggr);

  # new style of getting info from aggr_status()
  $status_obj = aggr_status($conn, $aggr);
  $type   = $status_obj->type();
  $state  = $status_obj->state();
  $status = $status_obj->status();
  $aggr   = $status_obj->aggr();
  %options= $status_obj->options();

  %options = aggr_options($conn, $aggr);

  @disks = aggr_disks($conn, $aggr);
  $vname = aggr_root($conn);

  $hash = aggr_spare_status($conn);
  $hash = aggr_raidgroup_status($conn, $aggr);
  @aggrs = aggr_show_space($conn, $option, $aggr);
  $aggr = aggr_show_space_hash($conn, aggr=>$my_aggr, %args);
  @aggrs = aggr_show_space_hash($conn, %args);

  # load specific parts only (recommended)
  use Aggregate qw(:aggr);  # load commonly used aggr_* functions

  # turn validation off
  {
    local $Aggregate::Validation = 0;
    aggr_online(...);
  }

=head1 DESCRIPTION

This module defines routines related to ONTAP's aggregates.

This library is still in beta status, testers welcome, but please be
warned that the behavior and interface may change.

=head2 $Aggregate::Validation

Certain functions have the ability to issue extra command(s) and double
check if the function actually had the intended effect. On very rare
occasions when you don't want this validation to be done, you can turn it
on/off via a package variable $Aggregate::Validation

Validation is turned on by default. To turn off validation, set the value
C<$Aggregate::Validation> to a false value. If you combine this with the
C<local()> operator, you can limit the effect to a block:

  use Aggregate qw(aggr_online);
  aggr_online(...); # validation effective
  {
    local $Aggregate::Validation = 0;
    aggr_online(...); # no validation
  }
  aggr_online(...); # validation effective

Currently C<aggr_create()>, C<aggr_add()>, C<aggr_online()>, C<aggr_offline()>,
C<aggr_restrict()>, C<aggr_size()>, and C<aggr_rename()> support this feature.

=head1 FUNCTIONS

All functions, unless otherwise stated, return a true value on success,
and false on failure. On failure C<$@> is set to the appropriate error
value via C<die()>.

Function declarations below all take C<$conn>, which is a Connection object
(Connectrec) that is created by calling C<Tharn::connect()>, and a aggregate
name (C<$aggr>)

=head2 aggr_create($conn,$aggr,$disklist[,%opts])

Creates a new aggregate.

The C<$disklist> argument can be either a reference to an array, or a scalar.
If it's scalar, C<aggr_create()> will try to use it as the
E<lt>ndisksE<gt>[@E<lt>disk-sizeE<gt>] notation, while if it's an array ref,
the contents of the array will be treated as disk names and they will be
prepended with the '-d' option.

You may specify the various difference options in C<%opts>
(if any of the following is unspecified, the system default value will be
used):

=over 4

=item mirror

Enable/disable aggregate mirroring

=item force

Force aggregate creation

=item language

Specify the language on the aggregate

=item rgsize

Specify the raidgroup size.

=item raidtype

Specify the raid type. The value must be either 'raid4' or 'raid_dp'

=item randomize_args

Randomizes the command line arguments (except for disk-list)

=item snaplock

Create a SnapLock aggregate (WORM). Requires a snaplock compliance
or snaplock enterprise license.

=item indirect_type

Support for larger aggregate. The value must be either '64' or '32'

=back

=cut

sub aggr_create
{

    param('HYBRID_AGGR', '-default', '0');

    if (my $server = are_we_ontapng($_[0])) {
	    return $server->storage_aggregate_new(aggregate=>$_[1], diskcount=>$_[2], $_[3]=>$_[4]);
	}
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);

    if ( ! $main::HYBRID_AGGR )
    {
        return ($Impl->cmd_create(@_));
    }

    #Hybrid Aggr addition
    my $rc;
    if ( ( $rc = $Impl->cmd_create(@_) ) != 1 )
    {
        #logInfo("rc: $rc");
        logDbg("Error with aggr creation");
        return $rc;
    }

    logInfo("hybrid aggr: $main::HYBRID_AGGR");
    if ( $main::HYBRID_AGGR )
    {
        logInfo("Creating HYBRID aggr");
        $rc = _aggr_hybrid(@_);
    }

    return $rc;

}

#Hybrid Aggregate addition of SSDs and hybrid_enable to on
sub _aggr_hybrid
{

    #aggr_create($fc, $aggr, $aggr_size, %aggr_args)
    my ( $fc, $aggr, $aggr_size, %aggr_args) = @_;

    my %tmp;
    if (  _aggr_hybrid_check( $fc, $aggr, %tmp) )
    {
        logInfo("Aggregate $aggr is already a hybrid aggr");
        return 1;
    }

	#No license currently needed - may need to revisit after decision made about product
    #use Admin::License qw(license is_licensed unlicense);
	##ymichael-vsim2> license add ETHWSMD
	##A snapmanager_vi site license has been installed.
    #sayEOF( $fc, "license add ETHWSMD");

    ##add bootargs to on - not needed now but leaving this commented out in case need to seto to true
    #sayEOF( $fc, "priv set -q test; bootargs set bootarg.dblade.hybrid_aggregates true");

    #add aggr option on
    #sayEOF( $fc, "aggr options $aggr hybrid_enabled on" );
    my $aggropts; 
    for (my $i = 1; $i <= 25; $i++) {
        logInfo("Trying to set hybrid_enabled option to on, attempt $i / 25"); 
        $aggropts = aggr_options($fc, $aggr, hybrid_enabled => "on");
        if ( $aggropts->{hybrid_enabled} ne "on" ) {
            logInfo("hybrid_enabled is not on, trying again in 5");
            sleep 5; 
        } else {
            last;
        }
    }

    if ( $aggropts->{hybrid_enabled} ne "on" ) {
        croak("aggr options $aggr hybrid_enabled is not on.  Set to off still.");
    }

    $Ssd_Pool_Retry = 0;
    my @ssd = _get_ssd_pool(@_);

    my $ssd_numdisks = 3;
    if ( defined $main::HYBRID_SSD_SIZE )
    {
        $ssd_numdisks = $main::HYBRID_SSD_SIZE;
        logDbg("Set ssd number of disks to HYBRID_SSD_SIZE ($main::HYBRID_SSD_SIZE)\n");
    }
    else
    {
        logDbg("Default ssd number of disks to $ssd_numdisks\n");
    }

    if ( scalar(@ssd) < $ssd_numdisks )
    #if ( scalar(@ssd) < 10000 )
    {

        my $hostp = ref($fc->hostp);
        my $hostname = $fc->hostp()->hostname();
        logInfo( "ref of hostp: " . $hostp );
        logInfo( "ref of hostname: " . $hostname );
        if ( $hostp =~ /Vsim/ || $hostp =~ /Simulator/ || $hostname =~ /vsim/)
        #if (1 == 1)
        {
            logInfo( "create vsims SSDs" );
            _aggr_vsim_create_ssd($fc, %aggr_args);
            sayEOF($fc, "disk assign all");
            $Ssd_Pool_Retry = 1;
            @ssd = _get_ssd_pool(@_);

        }
        else
        {
            croak("Number of ssd available is less than desired size: " . scalar(@ssd) . " < " . $ssd_numdisks );
        };
    }

    my $count = 0;
    my $disks = "";
    for( $count=0; $count < $ssd_numdisks; $count++ )
    {
        $disks .= $ssd[$count]->{name} . " ";
    }

    # aggr add hyA -g new -d v6.24 v6.25 v6.26 v6.27
    my $text = sayEOF( $fc, "aggr add $aggr -g new -d $disks");
    croak($text) unless ( $text =~ /has\s+been\s+initiated/ || $text =~ /has\s+completed/ );

    #check if hybrid aggr
    if (  ! _aggr_hybrid_check( $fc, $aggr, %tmp) )
    {
        my $msg = "aggr $aggr is NOT hybrid";
        croak($msg);
    }

    logInfo("$aggr is now a hybrid aggr");
    return 1;

}

#check if the existing aggr is hybrid already
sub _aggr_hybrid_check
{

    my ( $fc, $aggr, %opts ) = @_;

    my $bool = 0;

    my $status_obj = aggr_status($fc, $aggr);
    #logDbg( "aggr status: " . Dumper($status_obj));
    if ( $status_obj->status() =~ /hybrid/ )
    {
        logDbg("hybrid for aggr $aggr");
        $bool = 1;
    }
    else
    {
        my $msg = "aggr $aggr is NOT hybrid";
        logDbg( $msg );
    }

    return $bool;

}


#check the SSD pool and see what is available
sub _get_ssd_pool
{
    my @ssd;
    my $attempt = 1;
    # Retry up to 5 minutes (30 * 10s = 300s)
    # 1st attempt is done right away without sleep
    my $max_attempts = 31;
    my $interval = 10;
    logDbg("SSD pool retry option = $Ssd_Pool_Retry");
    while ( ( @ssd < 3 ) && ( $attempt < 31 ) ) {
        logStep("Getting SSD pool, attempt $attempt");
        @ssd = ();
        logInfo("Add the raid group for SSDs");
        my @spare_pools = aggr_spare_status(@_);
        logDbg( "Spare Pool:" . Dumper(@spare_pools) );
        foreach my $pool (@spare_pools)
        {
            #logDbg( "POOL: " . Dumper($pool));
            #logDbg( "POOL DEFINED: " . ref($pool));
            if ( ref($pool) eq 'ARRAY')
            {
                foreach my $s (@$pool)
                {
                    #logDbg("S: " . Dumper($s) );
                    if ( lc($s->{type}) eq "ssd")
                    {
                        push( @ssd, $s );
                    }
                }
            }
        }
        # If retry option is set, then we're expecting
        # some SSDs to show up. Wait for a minimum of 3 in
        # order for aggr add disks to succeed. Other SSDs can
        # show up later with no consequence.
        if ( !$Ssd_Pool_Retry ) {
            logInfo("Found ". @ssd . " SSDs");
            last;
        } elsif ( @ssd < 3 ) { 
            logStep("Got " . @ssd . " SSDs, trying again til we get >= 3");
            sleep $interval;
            $attempt++;
        } else {
            logStep("Got at least 3 SSDs");
            last;
        }
    }
    return @ssd;

}


#this is to load fake SSDs to the vsim - for smokes and vsim JETI
#see http://wikid.netapp.com/w/User:Ananthan/HyA/HyAonVsim
sub _aggr_vsim_create_ssd
{

    my ( $fc, %aggr_args ) = @_;

    my $adaptor = 2;
    if ( $aggr_args{adaptor} ne "" )
    {
        $adaptor = $aggr_args{adaptor};
    }

    my $ssdnumdisks = 6;
    if ( $aggr_args{ssdnumdisks} ne "" )
    {
        $ssdnumdisks = $aggr_args{ssdnumdisks};
    }


    use boot;
    my $hostspec = $fc->hostp();
    my $hostname = $fc->hostp()->hostname();
	my $passwd   = $fc->hostp()->default_password();

    host($hostname);
    my $con = connect('#console') or croak "failed get console: $@";
    #say($con, "logger $str", -blind);

    my ( $match, $before, $after, $prompt ) = say($con, "priv set -q diag; systemshell", '-noprompt' );
    sleep 2;
    logDbg("prompts |$match|$before|$after|$prompt");

    ( $match, $before, $after, $prompt ) = say($con, "\r", '-wait', ['login:'], '-noprompt' );
    logDbg("prompts2 |$match|$before|$after|$prompt");

    if ( $match =~ /login/ )
    {
        ( $match, $before, $after, $prompt ) = say($con, "diag", '-wait', ['Password'], '-noprompt');
        sleep 2;
        if ($match =~ /Password/)
        {
            #( $match, $before, $after, $prompt ) = say($con, $passwd, '-wait', [$hostname], '-noprompt');
			( $match, $before, $after, $prompt ) = say($con, 'netapp1!', '-wait', [$hostname], '-noprompt');
        }
    }

#make the disks on the vsim
    #sayEOF($con, "setenv PATH /sim/bin:\$PATH");
    sayEOF($con, "cd /sim/dev");
    #sudo /sim/bin/makedisks.main -t 35 -a 2 -n 6
    #-t 35 is SSD
    sayEOF($con, "sudo /sim/bin/makedisks.main -t 35 -a $adaptor -n $ssdnumdisks");
    ( $match, $before, $after, $prompt ) = say($con, "sudo reboot", '-wait', ['SIMLOADER', 'Press Ctrl-C for Boot Menu'], '-noprompt', '-timeout', 300);
    logDbg("prompts3 |$match|$before|$after|$prompt");
    if ( $match =~ /SIMLOADER/ )
    {
        ( $match, $before, $after, $prompt ) = say($con, "boot", '-wait', ['Press Ctrl-C for Boot Menu'], '-noprompt');
    }

    #wait for reboot
    transit_to($hostspec, UP, timeout=>'1200') or die("transit_to: $@");
    #$Server->transit_to(state => $Server->CLI());
}


=head2 aggr_destroy($conn,$aggr)

Destroy a aggregate.

=over 4

=item destroy_corrupt_volumes

If '1', then volume is destroyed even if it is corrupt. If '0'
then corrupt volumes are preserved

=back

=cut

sub aggr_destroy
{
	if (($_[1] eq $main::FILER_AGGR) && ($main::SETUP_DONE == 1) && !$main::IRON_ON_ROOT)
	{
		if ($main::CLEANUP ne "all")
		{
			logError("someone tries to destroy $main::FILER_AGGR, it is forbidden.");
			return undef;
		}
	}
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_destroy(@_));
}

=head2 aggr_undestroy($conn,$aggr)

Undestroy a aggregate.

=cut

sub aggr_undestroy
{
        $Impl = _init_impl(@_);
        $Impl->validation($Validation);
        return ($Impl->cmd_undestroy(@_));
}

=head2 aggr_add($conn,$aggr,$disklist[,%opts])

Adds disks to a aggregate.
For parameter $disklist, if the type is scalar, specify the number of disks
to be added. If the type is an array, aggr_add() will use the '-d' option
to add each disk via its disk name taken from each element of the array.

aggr_add() accepts the following option in %opt :

=over 4

=item force

Force disk addition

=item rg

Specify which raidgroup to add this disk to

=item timeout

Time to wait for validation that the disk has been added

=back

=cut

sub aggr_add
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_add(@_));
}

=head2 aggr_purge($conn,$aggr,[%opts])

A simple convenience function to set an aggregate offline, and then destroy it
along with its hosted flex volumes. aggr_purge() accepts the following options
in %opts:

=over 4

=item vol_only

Destroy only the flex volumes hosted by the aggregate. The aggregate is not
destroyed.

=item destroy_corrupt_volumes

If '1', then volume is destroyed even if it is corrupt. If '0'
then corrupt volumes are preserved

=back

=cut

sub aggr_purge
{
        if (my $server = are_we_ontapng($_[0])) {
	    if ($server->storage_aggregate_offline(aggregate=>$_[1])) {
		return $server->storage_aggregate_delete(aggregate=>$_[1], force=>"", wait=>30);
	    } else {
		return undef;
	    }
	}
	if (($_[1] eq $main::FILER_AGGR) && ($main::SETUP_DONE == 1))
	{
		if ($main::CLEANUP ne "all")
		{
			logError("someone tries to destroy $main::FILER_AGGR, it is forbidden.");
			return undef;
		}
	}
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_purge(@_));
}

=head2 aggr_online($conn,$aggr[,%opts])

Change the state of the aggregate to "online".
Returns success (if the aggregate was changed or is already in the desired
state).
or failure otherwise (if the aggregate couldn't change or didn't exist).
aggr_online() accepts the following options in %opts :

=over 4

=item force

Force onlining of aggregate.

=back

=cut

sub aggr_online
{
        if (my $server = are_we_ontapng($_[0])) {
	    return $server->storage_aggregate_online(aggregate=>$_[1]);
	}
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_online(@_));
}

=head2 aggr_offline($conn,$aggr[,%opts])

Change the state of the aggregate to "offline".
Returns success (if the aggregate was changed or is already in the desired
state).
or failure otherwise (if the aggregate couldn't change or didn't exist).
aggr_offline() accepts the following options in %opts :

=over 4

=item force

Force offlining of aggregate. Will run with 'priv set -q diag' since it
will only work in that mode.

=item cifsdelaytime

The number of seconds to delay before taking the aggregate offline, during
which time CIFS uses are warned of the pending loss of service.

=item all

Offline aggregate with all its volume. This option is not documented.
Do not use it unless it iss *really* necessary. If you want to purge
an aggregate quickly, pls use aggr_purge() instead.

=back

=cut

sub aggr_offline
{
        if (my $server = are_we_ontapng($_[0])) {
	    return $server->storage_aggregate_offline(aggregate=>$_[1]);
	}
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_offline(@_));
}

=head2 aggr_restrict($conn,$aggr[,%opts])

Change the state of the aggregate to "restricted".
Returns success (if the aggregate was changed or is already in the desired
state).
or failure otherwise (if the aggregate couldn't change or didn't exist).
aggr_restrict() accepts the following options in %opts :

=over 4

=item force

Force restricting of aggregate. Will run with 'priv set -q diag' since it
will only work in that mode.

=item cifsdelaytime

The number of seconds to delay before making the aggregate restricted, during
which time CIFS uses are warned of the pending loss of service.

=back

=cut

sub aggr_restrict
{
        if (my $server = are_we_ontapng($_[0])) {
	    return $server->storage_aggregate_restrict(aggregate=>$_[1]);
	}
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_restrict(@_));
}

=head2 aggr_read_fsid($conn,$aggr)

   my $fsid = aggr_read_fsid($conn, $aggr);

Returns the FSID (in hexadecimal) of an aggregate.

=cut

sub aggr_read_fsid
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_read_fsid(@_));
}


=head2 aggr_rewrite_fsid($conn,$aggr,[%opts])

  aggr_rewrite_fsid($conn, $vol, fsid => $aggr_fsid, [force => (0|1)]);

Rewrites the FSID of the desired aggregate. Returns the *original* FSID of
the aggregate prior to the FSID rewrite.

Do not force a rewrite of FSID of an aggregate if you have no idea what
you're doing. Note: You need to restrict the volume first prior to
calling this API.

aggr_rewrite_fsid() accepts the following options in %opts:

=over 4

=item fsid

The desired FSID to be written to the aggregate.

=item force

Optional. Force rewriting of the FSID even though an existing of the
same FSID exists on the system.

=back

=cut

sub aggr_rewrite_fsid
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_rewrite_fsid(@_));
}


=head2 aggr_rename($conn,$oldaggr,$newaggr)

Renames the name of aggregate $oldaggr to $newaggr. Returns success or failure.

=cut

sub aggr_rename
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_rename(@_));
}

=head2 aggr_blocks($conn,$aggr)

Returns a hash of aggregate block size information (block_size blocks fs_size)

=cut

sub aggr_blocks
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_blocks(@_));
}

=head2 aggr_list($conn)

Returns the list of known aggregate names, listed from the 'aggr status' command

  my @vols = aggr_list($conn);

=cut

sub aggr_list
{
        if (my $server = are_we_ontapng($_[0])) {
	    return $server->storage_aggregate_show();
	}
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_list(@_));
}

sub aggr_match_plex_raid_info
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_match_plex_raid_info(@_));
}

=head2 aggr_status($conn,$aggr[,$timeout])

Returns the state, status, host aggregate name, and a hash of aggregate options
of the given aggregate.

  my($state, $status, $vol, %options) = aggr_status($conn, $aggr);

XXX/tapl Return of a list of volumes and options thingy is not working.
         Daisuke needs to look into this for me.

Note that the aggregate options that "aggr status" command returns is not the
same as the extensive one that is returned by aggr_options.

When called in a scalar context, returns an object. Object's accessor methods
have to be called to access aggregate's status info from the object.

  my $status_obj = aggr_status($conn, $vol);
  my $type      = $status_obj->type();
  my $state     = $status_obj->state();
  my $status    = $status_obj->status();
  my %options   = $status_obj->options();
  my $vols_list = $status_obj->volumes(); # returns reference to an array

To access list of volumes in the hosting aggregate...

  # iterate through the list of volumes
  foreach my $vol ( @{$vols_list} ) {
      # do something to $vol
  }

  # offline first volume in the list
  vol_offline($conn, $vols_list->[0]);

IMPORTANT: The options presented by aggr_status() are grabbed from calling
'aggr options', rather than parsing the options column from 'aggr status'.

=cut

sub aggr_status
{
	$Impl = _init_impl(@_);
	my $conn = make_conn_and_complain(shift);
	my($aggr, %args) = validate_mixed(
		@_,
		[ {type => SCALAR} ],
		{
			timeout => { type => SCALAR, optional => 1, default => 300 }
		}
	);

        if (my $server = are_we_ontapng($conn)) {
	    my %aggr;
	    if ($server->storage_aggregate_show(aggregate=>$aggr, want_hash=>\%aggr)) {
		my $status = Aggregate::ImplTemplate::Status->new(cmd=>$Impl->cmd(),
								  validation=>$Impl->validation());
		$status->{state} = $aggr{state};
		if (wantarray()) {
		    die "vol_status against Tricky does not support returning an array yet\n";
		}
		return $status;
	    }
	    return undef;
	}
	$Impl->validation($Validation);
	return $Impl->cmd_status($conn, $aggr, timeout => $args{timeout});
}

# XXX - wait for state/status change
=head2 aggr_wait_change($conn,$aggr[,%opts])

C<aggr_wait_change()> will wait for asynchronous processing on a aggregate that
may change its state/status. For example, to wait for a aggregate to
come up online, you would do:

  use Aggregate qw(aggr_wait_change);
  ...
  my $conn = connect(...);
  my $rv = aggr_wait_change($conn, 'aggr0', state => qr(^online$) );
  my $rv = aggr_wait_change($conn, 'aggr0', status => qr(^normal|raid4$) );

The C<state> or C<status> parameter specifies which field(s) to match.
The function will look at the state and status of an aggregate as return by
'aggr status -r <aggr>' to match against the required state/status.

So in order to match the state as either online or offline, you can do:

  aggr_wait_change($conn, 'aggr0', state => qr{^(?:on|off)line$} );

If you specify multiple keys in either the C<state> or C<status> hash, all
of those criteria must be met before C<aggr_wait_change()> can successfully
return.

You can also specify a C<timeout> and a C<delay> parameter, which specify
the total time to try getting the desired status, and the time to sleep
in between match attempts, respectively.

=cut

sub aggr_wait_change
{
	$Impl = _init_impl(@_);
	my $conn = make_conn_and_complain(shift);
	my($aggr, %args) = validate_mixed(
		@_,
		[ { type => SCALAR } ],
		{
			timeout  => { type => SCALAR, optional => 1, default => 900 },
			state    => { optional => 1, default => qr/online/ },
			status   => { optional => 1},
			delay    => { type => SCALAR, optional => 1, default=> 60 }
		}
	);


    return eval {
		# for efficiency's sake, create a closure which does the
	        # matching.
		my @requirements;
		foreach my $field (qw(state status)) {
			if($args{$field}) {
				if(ref($args{$field}) ne 'Regexp') {
					$args{$field} = qr($args{$field});
				}
				push @requirements, sprintf('$%s =~ /$args{%s}/g', $field, $field);
			}
		}

		my $checker = eval sprintf( <<'EOM', join('&&', @requirements));
			sub {
				my($state, $status) = @_;
				return %s;
			}
EOM
		croak() if $@;

		my $sleepamt = $args{delay};
		my $endtime  = time() + $args{timeout};
		while(1) {
			last if time() > $endtime;

			# XXX/tapl This is currently coded this way because
			# aggr_status() is busted and I'm waiting to migrate
			# this over to the new cmd_status() in ImplTemplate.pm.

			my %rg_status = aggr_raidgroup_status($conn, $aggr)
			    or croak($@);

			# Grab the aggr status from 'aggr status -r'
			# Aggregate vol0 (online, raid4) (block checksums)
			my @status_split = split(/,\s/, $rg_status{'status'});

			# note: the checker() closure will only check the first
			# two elements of the array.
			# $status_split[0] = "online"
			# $status_split[1] = "raid4"

			if($checker->(@status_split)) {
				return 1;
			}
			select(undef,undef,undef,$sleepamt);
		}
		croak("Timed out waiting for change in aggregate '$aggr': " .
			join( ",", map{ "$_ => $args{$_}" } keys %args ));
	};
}

sub aggr_wait_zero
{
	$Impl = _init_impl(@_);
	my($conn, $aggr) = validate_pos( @_, { isa => 'Connectrec' }, 1 );
}

=head2 aggr_options($conn,$aggr[, %opts])

Sets or gets the options of an aggregate. If the function is called with two
arguments, returns a hash of option key/value pairs for the specified aggr.
In scalar context, returns a reference to the hash.

When called with three arguments, will set the optional parameters for the
aggregate depending on what the $opts{key} parameter says.
Returns a hash of option keys/value pairs for the specified aggregate.
In scalar context, returns a reference to the hash

The syntax of the optional parameters is as follows:

    $aggropts = aggr_options($conn, $aggr, fs_size_fixed => 1);
    $aggropts = aggr_options($conn, $aggr, resyncsnaptime => 100);


The options "root, ignore_inconsistent" can be changed in
maintainance mode only.

=cut

sub aggr_options
{
	$Impl = _init_impl(@_);
	my $conn = make_conn_and_complain(shift);
	my($aggr, %opts) = validate_mixed(@_,
		[{ type => SCALAR}],
		{
			root			=> { type => BOOLEAN, optional => 1,
				regex => qr/^\s*$/ },
			ignore_inconsistent => { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },

			nosnap			=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			snapmirrored	=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			fs_size_fixed	=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			snapshot_autodelete => { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			lost_write_protect => { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },

			raidtype		=> { type => SCALAR, optional => 1,
				regex => qr/raid4|raid_dp|raid0/i },

			raidsize		=> { type => SCALAR, optional => 1,
				regex => qr/\d+/ },  # 2-14 : valid value
			resyncsnaptime	=> { type => SCALAR, optional => 1,
				regex => qr/\d+/ },  # > 0 : valid value

            hybrid_enabled => { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },

			# Update cmd_options when a new option which has a valid value of 1/0
		}
	);

	$Impl->validation($Validation);
	return $Impl->cmd_options($conn, $aggr, %opts);

}

=head2 aggr_exists($conn,$aggr)

Returns true if a given aggregate name exists.
(CAVEAT: This function is currently implemented to use aggr_options())

=cut

sub aggr_exists
{
        if (my $server = are_we_ontapng($_[0])) {
	    return $server->storage_aggregate_show(aggregate=>$_[1]);
	}
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_exists(@_));
}

=head2 aggr_disks($conn,$aggr)

Returns an array of disk descriptors for all disks on the aggregate.  Each
disk descriptor is a hash of the form (type name size)

=cut

sub aggr_disks
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return $Impl->cmd_disks(@_);
}

=head2 aggr_root($conn)

Returns the name of the root aggregate on the given filer.

=cut

sub aggr_root
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return $Impl->cmd_root(@_);
}

=head2 aggr_scrub($conn,$aggr,$mode)

Performs the aggr scrub operation, specified by C<$mode>, which can
be one of the following:

  start
  stop
  resume
  suspend
  status

Returns 1 if the operation is successful, *except* for C<status>, which
returns the percentage completed and the last scrub time.

   my($completed, $lastscrub) = aggr_scrub($conn, $aggr, 'status');

C<completed> is set to 0E0 ("0 but true") for both when the percentage
is at 0 or there is no scrubbing being performed on the aggregate.
This is to allow the function to return a "0" while allowing a
construct like:

   my($completed, $lastscrub) = aggr_scrub($conn, $aggr, 'status') or croak();

=cut

sub aggr_scrub
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return $Impl->cmd_scrub(@_);
}

=head2 aggr_snap_delta($conn,$aggr[,%opts])
Returns the output of the snap delta command for the aggregate specified,
in an array of hash references. Keys are taken from header.

The syntax of the optional parameters is as follows:

  $snap_change = aggr_snap_delta($conn, $aggr);
  $snap_change = aggr_snap_delta($conn, $aggr, beginning_snapshot=>snap1);
  $snap_change = aggr_snap_delta($conn, $aggr, beginning_snapshot=>snap1,
                                 ending_snapshot=>snap2);

=item Examples
Volume aggr0
working...

From Snapshot   To                   KB changed  Time         Rate (KB/hour)
--------------- -------------------- ----------- ------------ ---------------
nightly.0       Active File System   83644          0d 23:43  3526.265
nightly.1       nightly.0            88196          1d 00:00  3672.962

Summary...

From Snapshot   To                   KB changed  Time         Rate (KB/hour)
--------------- -------------------- ----------- ------------ ---------------
nightly.1       Active File System   171840         1d 23:43  3600.062

The above output is returned in an arrayref as follows:
[
    {
    'From Snapshot' => 'nightly.0',
    'KB changed' => '83704',
    'Rate (KB/hour)' => '3526.069',
    'Time' => '0d 23:44',
    'To' => 'Active File System'
    },
    {
    'From Snapshot' => 'nightly.1',
    'KB changed' => '88196',
    'Rate (KB/hour)' => '3672.962',
    'Time' => '1d 00:00',
    'To' => 'nightly.0'
    },
    {
    'From Snapshot' => 'nightly.1',
    'KB changed' => '171900',
    'Rate (KB/hour)' => '3599.937',
    'Time' => '1d 23:45',
    'To' => 'Active File System'
    }
]

=cut

sub aggr_snap_delta
{
        $Impl = _init_impl(@_);
        $Impl->validation($Validation);
        return $Impl->cmd_snap_delta(@_);
}

=head2 aggr_snap_reserve($conn,$aggr[,%opt])

Get/Set snap reserve on the aggregate specified by C<$aggr>.

This method accepts the following option in %opt :

    reserve - snap reserve in percentage

This method returns the value of snap reserve of the aggregate if the operation is successful.

Get the value of snap reserve of a aggregate:

   my($reserve) = aggr_snap_reserve($conn, $aggr);

Set the value of snap reserve of a aggregate to 30 percent:

   my($new_reserve) = aggr_snap_reserve($conn, $aggr, reserve => 30);

=cut

sub aggr_snap_reserve
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return $Impl->cmd_snap_reserve(@_);
}

=head2 aggr_split($conn,$aggr,$newaggr)

Split a aggregate

=cut

sub aggr_split
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_split(@_));
}

# XXX - use Aggregate::Copy instead?
sub aggr_copy
{
	croak("TDB");
}

sub aggr_verify
{
	croak("TBD");
}


=head2 aggr_spare_status($conn)

Given a connection to a filer, fetches the output of 'aggr status -s' and
returns a hash containing lists of Aggregate::DiskStatus objects.

The list of disks are keyed by the "pool" that they belong in. If there
are no pools defined, everything will go into slot "0"

  my $conn   = connect(...);
  my $hash = aggr_spare_status($conn);
  foreach my $disk (@{$hash->{0}}) {
    print $disk->name, "\n";
  }


In scalar context, the reference to the resulting hash is returned.
In array context, the hash values are returned directly.

  my $hash = aggr_spare_status($conn);
  my %hash = aggr_spare_status($conn);

=cut

sub aggr_spare_status
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return $Impl->cmd_spare_status(@_);
}

=head2 aggr_raidgroup_status($conn,$aggr)

Given a connection to a filer and a target aggregate, parses the output of
C<aggr status -r $aggr> and returns a hash.

  NOTE: If this function is going to be used in more than one script,
  the returned structure needs to be changed a bit more...

This has is a 4-level nested hash. The top level contains some information
about the aggregate. The second level can be accessed via $hash->{plex},
which is an array of plexes. The third level has the raidgroup information,
and can be accessed via $plex->{raidgroups}. The last level a list of
Aggregate::DiskStatus objects, and can be accessed via $raidgroup->{disks}

  need to work on this doc...

=cut

sub aggr_raidgroup_status
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return $Impl->cmd_raidgroup_status(@_);
}



=head2 aggr_show_space($conn, $option, $aggr)

Given a connection to a filer and a target aggregate, parses the output of
C<aggr show_space [$option] [$aggr]> and returns a array.

This is a 3-level nested array/hash.

The top level array contains information about the aggregate/aggregates.

The second level stores information about space usage of aggregates and they can be accessed via $aggrs[]->{totalSpaceValue} or $aggrs[]->{waflReserveValue} or $aggrs[]->{snapReserveValue} or $aggrs[]->{usableSpaceValue}
or $aggrs[]->{bsrNVLOGValue} or $aggrs[]->{asisValue} or $aggrs[]->{imageBackupValue} etc.

The third level has the volumes within the aggregate information we well as break down allocation of aggregates' total space, Snap reserve and WAFL reserve space. Those information can be accessd via $aggrs[]->{totalSpace}->{allocated} or $aggrs[]->{totalSpace}->{used} or $aggr[]->{totalSpace}->{avail}, same for $aggrs[]->{waflReserve} and $aggrs[]->{snapReserve} values, etc. And for volumes space usage information within a aggregare: $aggrs[]->$vols[]->{allocated} or $aggrs[]->vols[]->{used} or $aggrs[]->vols[]->{guarantee}.

If aggr show_space failed due to invalid $option or $aggr name, the exception will be raised with error messages from aggr show_space.

=cut

sub aggr_show_space
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return $Impl->cmd_show_space(@_);
}


=head2 aggr_show_space_hash($conn, %options)

A variation on the 'aggr_show_space' function (above), this function runs the
'aggregate show_space' command, either for a single aggregate (if specified),
or for all aggregates.

Information returned is in form of a HASHREF (for a single aggregate), or HASH
(for all aggregates). The difference in formats for the two cases is to provide
at least some prevention of misinterpreting 'all aggrs' data as 'single aggr'
data, and vice-versa. An error is thrown if called in a 'scalar' context for the
'all aggrs' case, and also if not in a 'scalar' context for the 'single aggr' case.

The array which is returned by 'aggr_show_space' in the 'vols' field is converted to
a hash. So:

       aggr_show_space:         $aggr->{vols} is an ARRAYREF
       aggr_show_space_hash:    $aggr->{vols} is a HASHREF

This function calls the 'aggr_show_space' function.

=over 4

=item $conn

This is the connection to filer host on which C<aggr show_space> command is to be executed.

=item %options

This is a hash with following optional keys.

report_type
	Units in which the above type of statistics are reported.
	Currently, there are following units: UNIT_KB, UNIT_MB, UNIT_GB
	and UNIT_TB.  The default is UNIT_KB (Utils::Units).

aggr
	This is the name of an aggregate on the filer for which info is to be returned.
        If specified, then a HASHREF will be returned containing the info returned by
        the 'aggr show_space' command. If not specified, then a HASH (not HASHREF) will be
        returned which gives info for all aggregates, with Keys being the names of the aggregates,
        and the Values being HASHREFs which contain the aggregates' information.
        If not given, then info on all aggregates will be given

=back

=cut

sub aggr_show_space_hash
{
        my ($conn) = shift(@_);
	my(%args) = validate_mixed(
		@_,
		[],
		{
			aggr           => { type => SCALAR, optional => 1 },
			report_type    => { type => SCALAR, optional => 1, default => UNIT_KB },
		}
	);

       my $report_type = $args{report_type};
       my $aggr = $args{aggr};
       my %units_hash = ( UNIT_KB => '-k', UNIT_MB => '-m', UNIT_GB => '-g', UNIT_TB => '-t' );
       my $option_flag = $units_hash{report_type};
       my @arglist = ($conn, $option_flag);
       if (defined($aggr)) {
              push(@arglist, $aggr);
       }
       if (wantarray) {
              if (defined($aggr)) {
                      die("aggr_show_space_hash:  AGGR was specified, but was called ".
                          "in ARRAY (hash) context -- s/b SCALAR (hashref)");
              }
       } else {
              if (!defined($aggr)) {
                      die("aggr_show_space_hash:  AGGR was not specified, but was called ".
                          "in SCALAR (hashref) context -- s/b ARRAY (hash)");
              }
       }
       my @aggr_array = aggr_show_space(@arglist);
       my $aggr_hashref = {};
       my $aggr_info;
       foreach my $aggrx (@aggr_array) {
              my $name = $aggrx->{name};
              $aggr_hashref->{$name} = $aggrx;
              my $vol_arrayref = $aggrx->{vols};
              my $vol_hashref = {};
              foreach my $vol (@$vol_arrayref) {
                      $vol_hashref->{$vol->{name}} = $vol;
              }
              $aggrx->{vols} = $vol_hashref;
              my @vol_info_list = values(%{$aggrx->{vols}});
              my @all_hashes = ( $aggrx, @vol_info_list );
              foreach my $hash (@all_hashes) {
                   while (my ($key,$val) = each(%$hash) ) {
                        if (ref($val) =~ /HASH/) {
                              while (my ($key2,$val2) = each(%$val) ) {
                                   if (!ref($val2) && $val2 =~ /^(\d+)KB?/) {
                                       $val->{$key2 . "_kb"} = $1;
                                   }
                              }
                        } elsif (!ref($val) && $val =~ /^(\d+)KB?/) {
                              $hash->{$key . "_kb"} = $1;
                        }
                   }
              }
              if (defined($aggr) && $name eq $aggr) {
                      $aggr_info = $aggrx;
                      last;
              }
       }

       if (wantarray) {
              logDbg("aggr_show_space_hash:  Returning HASH for all aggregates");
              return( %$aggr_hashref );
       } elsif (defined($aggr_info)) {
              logDbg("aggr_show_space_hash:  Returning HASHREF for single aggregate: $aggr");
              return( $aggr_info );
       } else {
              logDbg("aggr_show_space_hash:  No aggregate \"$aggr\" was found -- returning UNDEF");
              return( undef );
       }
}


=head2 aggr_disk_free($conn, %options)

Runs df command on the specified Aggregate. Below is the description of the
parameters of this subroutine.

=over 4

=item $conn

This is the connection to filer host on which C<df> command is to be executed.

=item %options

This is a hash with following optional keys.

path
	This is the name of the aggregate.

space_type
	The types of statistics to be reported. There are three types:
	DF_SPACE, DF_INODES, DF_PRIVATE and DF_RESERVE. Default is DF_SPACE.

report_type
	Units in which the above type of statistics are reported.
	Currently, there are following units: UNIT_KB, UNIT_MB, UNIT_GB
	and UNIT_TB.  The default is UNIT_KB.

snapshot
	If the value of this field is non-zero, the statistics are
	calculated for the snapshot of the requested aggregate.
	The default is 0.

sync
sync_delay
	'sync' option is either 0 or 1. If 0, no delay or sync is performed
	If 1, then sync and delay are done, and either 'sync_delay' option or
	$::DF_SYNC_DELAY global param (see "commonParams.pl") is used to
	determine the length of the delay.

	A 'df' performed immediately after a file operation (e.g. 'truncate')
	might not reflect the file operation in its results. The delay is
	intended to allow all previous file operations to "register" before
	performing the df.

	Default value for 'sync' is now 1. See BURT306218

	Note that there is no way to "guarantee" a delay sufficient to capture
	all previous changes. A greater delay is just more likely to suffice than
	a smaller delay (i.e. delay values are "arbitrary")

no_poll
	Optional. Default is '0'. If '1', then no polling of WAFL counters or other
	values is done before the 'df' operation. However, the delay specified by
	'sync'/'sync_delay' is still done before the 'df' operation.

=back

This subroutine returns a hash reference in scalar context and an array in
array context. For example,

  use Aggregate qw(:df)
  use Utils::Units qw(:constants)

  my ($aggr_df) = aggr_disk_free($conn, path => "aggr0",
			space_type => DF_SPACE, report_type => UNIT_KB)

  my ($total, $used, $avail, $capacity) = aggr_disk_free($conn,
		path => "aggr0", space_type => DF_SPACE,
		report_type => UNIT_KB)

  my ($snap_df) = aggr_disk_free($conn, path => "aggr0", snapshot => 1)

Elements of arrary and values of hash depends on C<space_type> parameter.

=over 4

=item DF_SPACE

Scalar context - The returned hash contains C<total space>, C<used space>,
C<available space> and C<capacity (percent)> keyed by C<total>, C<used>,
C<avail> and C<capacity> respectively.

Array context - The returned array contains C<total space>, C<used space>,
C<available space> and C<capacity (percent)> in the specified order.

=item DF_RESERVE

Scalar context - The returned hash contains C<total space>, C<used space>,
C<available space> and C<reserved space> keyed by C<total>, C<used>,
C<avail> and C<reserved> respectively.

Array context - The returned array contains C<total space>, C<used space>,
C<available space> and C<reserved space> in the specified order.

=item DF_INODES

Scalar context - The returned hash contains C<used nodes>, C<free inodes>
and C<used percentage> keyed by C<iused>, C<ifree> and C<percent_iused>
respectively.

Array context - The returned array contains C<used nodes>, C<free inodes>
and C<used percentage> in the specified order.

With 'dual inode' feature, returns values for 'public' inodes, while
'DF_PRIVATE' returns values for 'private' inodes.

=item DF_PRIVATE

Only valid with 'dual inode' feature. Similar to 'DF_INODES', but returns
values for 'private' inodes, whereas 'DF_INODES' returns values for
'public' inodes.

=back

=cut

sub aggr_disk_free
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	my ($conn, %args) = @_;
	return $Impl->cmd_disk_free($conn, %args, type=>'aggr');
}


=head2 aggr_meta_i2p($conn, $aggr)

Returns the output of the aggregate's meta_i2p command as a hashref,
with fileids as keys, and filenames as values.

  my $i2p = aggr_meta_i2p($conn, $aggr);

=cut

sub aggr_meta_i2p {
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return ($Impl->cmd_meta_i2p(@_));
}



=head2 aggr_mirror($conn, %options);
 Convert the existing aggregate to mirrored aggregate by
 adding disks from same/another pool as specified

=over 4

=item conn
Connection rec for the appliance

=item %options
This is the hash with the following keys

aggr - Name of the aggregate

disks - ARRAYREF of the disks, to be used to create mirror aggr

victim_aggr - Name of the victim aggregate
		(when victim aggregate is specified, disks should not be specified),
		There is no extra logic/enforcement to check this. Scripts are expected to do this
		and by doing this adds extra flexibility to test various invalid combinations

=back

=cut

sub aggr_mirror
{
	$Impl = _init_impl(@_);
	$Impl->validation($Validation);
	return $Impl->cmd_mirror(@_);
}
=head2 aggr_blockcheck(%opt)
This function will start the aggr block_check scanner on the filer aggr
and wait until it finishes scanning


Returns the aggr block_check command output file in case of wait equal to 1.

Returns the scanid of the block_check scanner and ls output of /etc/log/
which has been obtained before executing the aggr block_check command in case of wait equal to 0.

This scanid and ls output can be passed to the aggr_blockcheck_wait function
to get the output file of aggr block_check command later.

=item %opt
This is the hash with the following keys

     filer => hostspec Required
     The host spec to establish the console connection

     aggr  => name of the aggregate Required

     pvbn  => SCALAR|ARRAYREF of pvbn numbers Optional.

     disks => SCALAR|ARRAYREF of disk id's Optional.

     wait => 0|1 Optional
     if 1,Wait until the aggr block_check scanner to complete scanning.
     default is 1.
     if 0,Didn't wait until the block_check scanner to complete.
Usage:
             ($errcode, $path) = aggr_blockcheck(filer=>$FILER,
                                                 aggr => $aggr,
                                                 pvbn=>\@pvbn,
                                                 wait=>1);
             ($scanid, $ls_etc_log_before)=aggr_blockcheck(filer=>$FILER,
                                                           aggr => $aggr,
                                                           pvbn=>\@pvbn,
                                                           wait=>0);

=cut

sub aggr_blockcheck (%)
{
        my %opts=validate(
             @_,
             {       filer => { type=>SCALAR},
                     aggr      => {type => SCALAR},
                     pvbn      => {type => SCALAR|ARRAYREF, optional =>1},
                     disks     => {type => SCALAR|ARRAYREF, optional =>1},
                     wait      => {type => SCALAR, default =>1, optional=>1}
             }

         );

        my $host           = $opts{filer};
        my $aggr           = $opts{aggr};
        my $pvbn           = $opts{pvbn};
        my $disks          = $opts{disks};
        my $wait           = $opts{wait};

        my ($result, $after, $before, $errcode, $match,
            $scanid, $output, $ls_etc_log_before);
        my $str      = "Starting scan on $aggr";
        my $start_msg= "Starting aggr block_check ";
        my $err_msg5 = "usage: ";
        my $err_msg2 = "aggr block_check: ";
        my $err_msg3 ="block_check:";
        my $err_msg4 ="Invalid PVBN";
        my $err_msg1 ="Aggregate $aggr not found";
        my $err_msg6 = "Disk.*not found";
        my $err_msg7 = "Disk.*not part of $aggr";
        my $err_msg8 = "Disk.*not a data disk in $aggr";
        my $string   = qr/$err_msg8.*|$str|$err_msg1|$start_msg|$err_msg2\w+\s\w+\s\w+\s\w+\s.*|$err_msg3\s\w+\s\w+.\w+\s\w+\s\w+.*|$err_msg4.*|$err_msg5.*|$err_msg6.*|$err_msg7.*/;
        my $fh       = host($host);
        ##Initiallizing console connection
        my $fc       = connect('#console');


        if(defined($pvbn)) {

           my $cmd = "priv set diag;aggr block_check $aggr -p @$pvbn";

           ##Getting the ls /etc/log/ before running block_check command
           ##which will be used to get the block_check output file later
           $ls_etc_log_before=cli($host, "priv set -q diag;ls /etc/log/");

           ##Running aggr block_check command
           ($match, $before, $after) =
              say($fc, $cmd, $string,
                  '-swallow',   -eval_error, -noprompt,
                  '-timeout', $::CMDTIMEOUT ,  '-maxidle', $::TIMEOUT);
        }

        if(defined($disks)){

          ##Getting the ls /etc/log/ before running block_check command
          ##which will be used to get the block_check output file later
          $ls_etc_log_before=cli($host, "priv set -q diag;ls /etc/log/");

          my $cmd ="priv set diag;aggr block_check $aggr -d @$disks";
          ##Running aggr block_check command
          ($match, $before, $after) =
              say($fc, $cmd, $string,
                  '-swallow',   -eval_error, -noprompt,
                  '-timeout', $::CMDTIMEOUT ,  '-maxidle', $::TIMEOUT);
        }
        if ($match =~
            /$err_msg1|$err_msg2|$err_msg3|$err_msg4|$err_msg5|$err_msg6|$err_msg7|$err_msg8/)
        {
            $errcode = 1;
            $result  = $match;
            return ($errcode, $result);
        } elsif ($match =~ /$start_msg|$str/) {
            ##waiting for aggr block check to start scanning
            sleep(5);
            $output = cli($host, "priv set -q diag;wafl scan status -A $aggr");

            if ($output =~ /(\d+)\s+aggr\sblock_check\s+Aggregate\s+/) {
                $scanid = $1;
            }
        }

        if (defined($scanid)) {

            if ($wait == 1) {
                $result =
                  aggr_blockcheck_wait(filer             => $host,
                                       scanid            => $scanid,
                                       aggr              => $aggr,
                                       ls_etc_log_before => $ls_etc_log_before);
                $errcode = 0;
            } else {

                return ($scanid, $ls_etc_log_before);

            }
        } else {

            if ($wait == 1) {
                $result =
                  aggr_blockcheck_wait(filer             => $host,
                                       aggr              => $aggr,
                                       ls_etc_log_before => $ls_etc_log_before);
                $errcode = 0;

            } else {
                return (undef, $ls_etc_log_before);
            }
        }

        return ($errcode, $result);
}
=head2 aggr_blockcheck_wait(%opt)
This function will wait for the aggr block_check scanner on the filer aggregate
to complete scanning.

Returns the aggr block_check command output file after completion of scanning.

=item %opt
This is the hash with the following keys

     filer             => hostspec Required

     aggr              => name of the aggregate Required

     scanid            => scanid of the block_check scanner Optional.
                          (Use scanid obtained from aggr_blockcheck function.)

     ls_etc_log_before => ls output of /etc/log/ before obtained before running aggr block_check.
                          (Use ls_etc_log_before obtained from aggr_blockcheck function.)
Usage:
         $blockcheck_path = aggr_blockcheck_wait(filer=>$FILER,
                                                 aggr => $aggr,
                                                 scanid=> $scanid,
                                                 ls_etc_log_before=> $ls_etc_log_before);

=cut


sub aggr_blockcheck_wait(%){

        my %opts=validate(@_,
                          {filer                 => {type => SCALAR},
                           aggr                  => {type => SCALAR},
                           scanid                => {type => SCALAR, optional=>1},
                           ls_etc_log_before     => {type => SCALAR}
                          });

        my $host           = $opts{filer};
        my $aggr           = $opts{aggr};
        my $scanid         = $opts{scanid};
        my $ls_etc_log_before = $opts{ls_etc_log_before};
        my ($output_file, $output);
        my $count =1;
        my $wait = 180;
        my $end_time       = time() + $wait;

        if(defined($scanid)){
           do {
              $output = cli($host, "priv set -q diag;wafl scan status -A $aggr");
              ##Instead of checking Wafl scan status many times which will increase rsh count,
              ##putting sleep to reduce the rsh connection count with the filer
              sleep(5);
              if($output !~ /$scanid/){
                 $count =0;
                }
           } while (($count == 1) && (time() < $end_time));
         }
         my $ls_etc_log_after=cli($host, "priv set -q diag;ls /etc/log/");
         my @ls_etc_log_after = split /\n/, $ls_etc_log_after;
         my @ls_etc_log_before =split /\n/, $ls_etc_log_before;


         my @union =my @inter = my @diff = ();
         my %count = ();

        foreach my $element (@ls_etc_log_after, @ls_etc_log_before) {
                $count{$element}++;
        };
        foreach my $element (keys %count) {
                push @union, $element;
                push @{ $count{$element} > 1 ? \@inter : \@diff }, $element;
        };
        foreach my $diff (@diff){
            if ($diff =~ /(aggr_blockcheck.$aggr.\d+)/){
                my $path ="/etc/log/";
                $diff =$1;
                $output_file = "$path" . "$diff";
            }
       }

       return $output_file;

}
=head2 aggr_blockcheck_read_output(%opt)
This function will read the aggr block_check output file

When defined pvbn
Returns the hash reference with pvbn and the corresponding file path

When defined disks
Returns arrayref of toatl scanned disks,lostblocks vbn,corrupted disks,snapshoot lists
and the number of lost blocks on corruption
Returns arrayref of toatl scanned disks and all other values will be zero.

=item %opt
This is the hash with the following keys

     filer => hostspec Required

     path  => path of the block_check output file Required.

     pvbn  => SCALAR|ARRAYREF of pvbn numbers Optional

     disks => 0|1 defaults to 0 Optional.

Usage:
    $pvbn_file = aggr_blockcheck_read_output(filer=>$FILER,
                                             path => $file_path,
                                             pvbn=>\@pvbn);
    ($scanned_disk,$vbn,$corrupted_disks,$snapshot_list,$lost_block_count)= aggr_blockcheck_read_output(filer=>$FILER,
                                                                                                        path => $file_path,
                                                                                                        disks=>1);

=cut

sub aggr_blockcheck_read_output (%)
{
        my %opts = validate(
            @_,
            {
                    filer => {type => SCALAR },
                    path =>  {type => SCALAR },
                    pvbn =>  {type => SCALAR|ARRAYREF, optional =>1},
                    disks => {type => SCALAR, default=> 0,optional =>1}
            }
        );

        my $fc = $opts{filer};
        my $path = $opts{path};
        my $pvbn_ref = $opts{pvbn};
        my $disks = $opts{disks};
        my $output;
        my $complete_msg = "Completed aggregate block check";
        do{
        $output = cli($fc, "priv set -q diag;rdfile $path");
        sleep(5);
        }until ($output =~ /$complete_msg/s);

        my @lines = split /\n/, $output;

        my $pvbn_file = {};
        my $lost_blocks_msg = "Lost blocks: ";
        my @scanned_disk;
        my (@vvbn_list, @corrupted_disk);
        my $count =0;
        my $snapshot_msg = "The following snapshots contain any of the listed blocks:";
        my $snapshots_corrupt;
        my @snapshot_list ;

        if ($disks == 1){

            foreach (@lines) {
                if($_ =~ /(\w+\.\w+.*)/){
                   my @disk_list = split / /, $1;

                    foreach my $list (@disk_list) {
                               push @scanned_disk, $list;

                    }

                }
                # Store the corrupt blocks and disks in a list
                if($_ =~ /$lost_blocks_msg(.*)/) {
                   my @lost_block = split / /, $1;
                   foreach (@lost_block){
                      if ($_ =~ /(\w+\.\w+)/){
                          if(length(@corrupted_disk) > 0){
                             if ($corrupted_disk[0] !~ $1){
                                 push @corrupted_disk, $1;
                             } else {
                               push @corrupted_disk, $1;
                             }
                          }

                       }
                       if ($_ =~ /^(\d+)/){
                          push @vvbn_list, $1;
                       }
                   }
                }

		# New format in IC.3
		# Lost VVBN: 4150, Disk: 0a.18, Volume: testvol_flex, Inode: 100
                if($_ =~ /Lost\sVVBN:\s(\d+)/) {
                   push @vvbn_list, $1;
		   if($_ =~ m/Disk:\s(.*),\sVolume:/) {
                      push @corrupted_disk, $1;
                   }
                }

                # Determine if there are any corrupt blocks in the snapshots.
                if($_ =~ /$snapshot_msg/) {
                   $snapshots_corrupt = 1;
                }

                # If there are snapshots, then add them to a list.
                if(($_ =~ /\d+:\d+:\d+/) && ($snapshots_corrupt == 1)) {
                     my @line = split / /, $_;
                     push @snapshot_list, $line[0];
                }

              # Store the number of lost blocks
              if($_ =~ /Total\s(\d+)\slost\sblock/) {
                  $count = $1;
              }

          }

          return (\@scanned_disk,\@vvbn_list,\@corrupted_disk,\@snapshot_list,$count);
        }


        if(defined($pvbn_ref)){

            my @pvbn_nos  = @$pvbn_ref;
            foreach my $temp_no (@pvbn_nos) {
                for (my $i = 0; $i < @lines; $i++) {
                    # Str to look for completed block check
                    my $mat_str = "Completed aggregate block check ";

                    if($lines[$i] =~ m/pvbn[ \t]+$temp_no/i) {
                        $lines[$i + 2] =~ m/pathname[ \t]+=[ \t]+(.*)/;
                        $pvbn_file->{$temp_no} = $1;
                    }

                    # New format in IC.3
		    # Lost PVBN: 4150, Disk: 0a.18, Volume: testvol_flex, Inode: 100
                    # Inode 100 has 1 name(s).
                    # Volume: testvol_flex, Inode: 100, Pathname: /vol/testvol_flex/file1

                    if($lines[$i] =~ m/PVBN:\s($temp_no)/i) {
                        $lines[$i + 2] =~ m/Pathname:\s(.*)/;
                        $pvbn_file->{$temp_no} = $1;
                    }
                    if($lines[$i] =~ m/$mat_str/) {
                        last;
                    }

                }
            }

          return $pvbn_file;

      }

}

=head1 SEE ALSO

L<Module::Common>
L<Params::Validate>
L<Utils::Units>

=head1 HISTORY

=head1 TODO

- Fix parsing of list of volumes & others in aggr_status()
- Determine if all the RAID centric APIs, eg. aggr_raidgroup_status(), should
  stay here and not Volume.pm, or be consolidated into this library.
- Add aggr_*() APIs if thery are not in here

TBD after code review

=cut


# internal function that initializes implementor object or dies if
# we are running on release where aggregates do not exist
sub _init_impl {
	my ($conn)  = make_conn_and_complain(shift);
	my $msgref;

	my ($name,$disks,%args) = @_;
	my $feature = 'AGGR';
	if( defined $args{indirect_type} ) { $feature = 'LARGER_AGGR'; }

	if (supported_feature($conn->hostp(), $feature, \$msgref)) {
		$Impl = Aggregate::ImplTemplate->new(
		    cmd        => 'aggr',
		    validation => $Validation
		);
	} else {
		my($filename, $line) = (caller(2))[1,2];
		my $version = get_wafl_version($conn->hostp());
		croak(na_result(UNSUPPORTED_FEATURE_EXCEPTION .". ". $msgref ." : calling aggregate function for release $version from $filename:$line"));
	}
}



1;


