# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/Aggregate/Copy.pm#1 $

## @summary aggr copy library -- action subroutines
##
## @description
##    This package defines routines which perform, start, abort,
##    complete, and get status on aggrcopy transfers.
##
## @dependencies
##    lib/Aggregate.pm
##    lib/SayWrap.pm
##    FSR/FSRutils.pm
##    li/fileSystem.pm
##    lib/FS/Util/snapCmds.pm
##
## @author neha@netapp.com
## @status Work-In-Progress

package Aggregate::Copy;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw( aggrcopy_throttle aggrcopy_start aggrcopy_from_source aggrcopy_cmp_src_dst aggrcopy_cmp_src_dst_ctx aggrcopy_status_list aggrcopy_status aggrcopy_abort aggrcopy_abort_all );

use Aggregate qw(aggr_status aggr_restrict);
use FSR::FSRutils;
use SayWrap;
use Tharn;
use fileSystem;
use FS::Util::snapCmds;
use Volume qw(vol_blocks);
use strict;

## @name :: aggrcopy_throttle
## @usage aggrcopy_throttle($filerConn, $operation, $throttle)
## @synopsis
##    Modifies the speed at which aggrcopy transfer is being done.
##    Returns success or failure, with an error mesage in $@ on failure.
sub aggrcopy_throttle($$$) {
  my ($conn, $operation, $throttle) = @_;

  return eval {
    die "throttle value must be an integer between 1 and 10\n"
      unless (($throttle =~ /^\d+$/) && ($throttle > 0 && $throttle < 11));

    my $text = sayEOF($conn, "aggr copy throttle $operation $throttle");
    die "$@" unless ($text =~ /^copy operation \d+: Throttle adjusted from (10|[1-9])0% to ${throttle}0% of full speed\.$/);
    return 1;
  }
}

# Internal function which constructs and returns the aggr copy command,
# given the source, destination, and an options hash.
sub aggrcopy_cmd_from_opts($$;%) {
  my ($src, $dst, %param) = @_;

  # Setup aggrcopy arguments
  my $cmd = "aggr copy start" .
    ($param{'COPY_ALL'} ? " -S" : "") .
    ($param{'SNAPSHOT'} ? " -s " . $param{'SNAPSHOT'} : "") .
    ($param{'METADATA'} ? " -M" : "") .
    ($param{'CONTEXT'} ? " -C" : "") .
    " $src $dst";

  # Some options only work in diag mode
  if ($param{'METADATA'}) {
    $cmd = "priv set diag ; " . $cmd;
  }

  return $cmd;
}

# The waitstrings for aggr copy depend upon the operation number,
# but it's better to only define the strings in one place.
sub aggrcopy_regx_from_op(;$) {
  my ($op_num) = @_;
  $op_num = "\\d+" unless (defined $op_num);

  # These are a hardcoded regexps which correspond to the ONTAP code.
  # Don't change these unless you know exactly what you're doing.
  my $finish_prog = "^\\s*\\d\\d:\\d\\d:\\d\\d (?:\\w{3}) : " .
    "aggr copy restore $op_num : \\d+ % done\\. " .
    "Estimate \\d+ minutes remaining\\.\\s*(?:\\n|\$)";
  my $finish_wait = "(?m)" .
    "(?:$finish_prog)*" .
    "\\s*\\d\\d:\\d\\d:\\d\\d (?:\\w{3}) : aggr copy restore (\\d+) : 100% done, \\d+ [KMGT]B copied\\.\\s*(?:\\n|\$)";

  return ($finish_prog, $finish_wait);
}

## @name ::aggrcopy_status_list
## @usage aggrcopy_status_list($filerConn)
## @synopsis
##    Returns a reference to a hash of status entries for all aggrcopy
##    operations where the key is the operation number and the value is the
##    progress.
##    Returns a reference to an empty array if no status entries were found.
##    On failure, returns an error message in $@.
sub aggrcopy_status_list($)  {
  my ($conn) = @_;
  my %hash;
  return eval {
    my $text = sayEOF($conn, "aggr copy status");
    if ($text eq "No operations in progress.") {
      return {};
    }

    foreach ( split /\r*\n/, $text ) {
      if ( /aggr copy (?:dump|restore) (\d+) : (?:(\d+)\s*% done|begun)/ ) {
        $hash{$1} = $2 ? $2 : 0;
      }
    }

    if ( %hash ) {
      return \%hash;
    } else {
      die "$text\n";
    }
  }
}

## @name ::aggrcopy_status
## @usage aggrcopy_status($filerConn, $operation)
## @synopsis
##    Returns the progress of the aggrcopy transfer with the given operation
##    number as a percentage.  Returns undef if no aggrcopy operation with
##    the given operation number is in progress.  On failure, returns an error
##    message in $@.
sub aggrcopy_status($$)  {
  my ($conn, $operation) = @_;
  return eval {
    my $text = sayEOF($conn, "aggr copy status $operation");
    if ($text eq "No operations in progress.") {
      return undef;
    } elsif ($text =~ /^\d{2}:\d{2}:\d{2} \w{3} : aggr copy (?:dump|restore) $operation : (\d+)\s*% done/) {
      return $1;
    } elsif ($text =~ /^\d{2}:\d{2}:\d{2} \w{3} : aggr copy (?:dump|restore) $operation : begun, \d+ \w+ to be copied\.$/) {
      return 0;
    } else {
      die "$text\n";
    }
  }
}

## @name ::aggrcopy_abort
## @usage aggrcopy_abort($filerConn, $operation)
## @synopsis
##    Aborts the aggrcopy transfer with the given operation number.
##    Returns success or failure, with an error message in $@ on failure.
sub aggrcopy_abort($$)  {
  my ($conn, $operation) = @_;
  return eval {
    my $text = sayEOF($conn, "aggr copy abort $operation");
    return 1 if ($text =~ /^image operation $operation is now being aborted$/);
    die "unexpected end of output\n" if ($text eq "");
    die "$text\n";
  }
}

## @name ::aggrcopy_abort_all
## @usage aggrcopy_abort($filerConn, $operation)
## @synopsis
##    Aborts the aggrcopy transfer with the given operation number.
##    Returns success or failure, with an error message in $@ on failure.
sub aggrcopy_abort_all($$)  {
  my ($conn, @operations) = @_;
  my($operation);
	
  return eval {
    my $text = sayEOF($conn, "aggr copy abort all");
    die "unexpected end of output\n" if ($text eq "");
	if ($text =~ /is now being aborted$/) {
   	 	return 1;
	}	
	return 0;
  }
}


## @name ::aggrcopy_start
## @usage aggrcopy_start($filerConn, $aggrname, %options)
## @synopsis
##    Starts an aggr copy transfer to the specified aggregate on the given filer.
##    Note, this must be the destination filer.  If the transfer successfully
##    starts, returns a tuple of a subtest object and the operation number.
##    The subtest object is for an active subtest which is waiting for the
##    transfer to complete. On failure, returns undef with error message in $@.
##    If the COMPLETE option is set, then this function returns whether
##    the transfer both started and completed successfully.
##    There is no need to extract results from the returned subtest object,
##    since $@ will contain the error if the transfer failed.
## @param SOURCE Required: the source <aggregate> or <filer>:<aggregate> to copy from
## @param RESTRICT Optional: whether to make sure the destination aggregate is
##                           restricted before starting the transfer
## @param COPY_ALL Optional: whether or not to copy all snapshots
## @param SNAPSHOT Optional: the name of a specific snapshot on the source
##                           aggregate to copy
## @param METADATA Optional: only copy metadata and indirect blocks,
##                           not level-0 blocks
## @param TIMEOUT Optional: timeout period waiting for the command to complete
## @param COMPLETE Optional: wait for transfer to complete before returning if set
## @result
##    The subtest waiting for the transfer to finish logs a FATAL result
##    with key EVAL_ERROR if the transfer fails.  Under normal circumstances,
##    the calling script should wait for the subtest to complete and then call
##    FSR::FSRutils::subtestResults() to determine whether the transfer succeeded.
sub aggrcopy_start($$;%) {
  my ($conn, $dstaggr, %param) = @_;
  return eval {
    die "SOURCE parameter is required\n" unless ($param{'SOURCE'});

    #$COMPLETE = 0 unless ($param{'COMPLETE'});

    # Setup some convenience variables and escaped strings.
    my $src = $param{'SOURCE'};
    my ($srcaggr, $srcfiler) = split_filer_path($param{'SOURCE'});
    if (!$srcfiler) {
	$srcfiler="127.0.0.1";
    }
    my $dstaggr_exp = regx_escape($dstaggr);
    my $srcaggr_exp = regx_escape($srcaggr);
    my $srcfiler_exp = regx_escape($srcfiler);

    # Make sure the destination is restricted (if necessary).
    if ($param{'RESTRICT'}) {
      die $@ unless aggr_restrict($conn, $dstaggr);
    }

    my %opts = ();
    $opts{'TIMEOUT'} = $param{'TIMEOUT'}
      if (defined $param{'TIMEOUT'});

    # Construct the aggr copy command.
    my $cmd = aggrcopy_cmd_from_opts($src, $dstaggr, %param);

    # This is a hardcoded regexp which corresponds to the ONTAP code.  The
    # parentheses also determine how we determine the aggrcopy operation number
    # below. Don't change these unless you know exactly what you're doing.
    my $start_wait = "(?m)" .
      "^This can be a long-running operation. Use Control - C \\(\\^C\\) to interrupt\\.\\s+" .
      "Copy Aggregate: $srcaggr_exp on machine $srcfiler_exp to Aggregate: $dstaggr_exp\\s+" .
      "AGGRCOPY: Starting on aggregate \\d+\\.".
      "\\s*\\d\\d:\\d\\d:\\d\\d (?:\\w{3}) : aggr copy restore (\\d+) : begun, \\d+ [KMGT]B to be copied\\.\\s*(?:\\n|\$)";

    my ($matched, $text) = sayWAIT($conn, $cmd, $start_wait, %opts);
    die $@ if ($@); # sayWAIT failed
    unless ($matched) {
      # waitstring never matched
      die "$text\n" if ($text);
      die "unexpected end of output\n";
    }

    # Re-match the matched string to determine the aggrcopy operation number.
    # Note, this should NEVER fail -- else we would have returned already.
    die "waitstring did not match returned string\n"
      unless ($matched =~ /$start_wait/);
    die "error determining aggrcopy operation number\n" unless (defined $1);
    my $op_num = $1;
    die "invalid operation number\n" unless ($op_num =~ /^\d+$/);

    # Any other text that we got must be an error message.
    die "$text\n" if ($text);

    my ($finish_prog, $finish_wait) = aggrcopy_regx_from_op($op_num);

    # The transfer successfully started,
    # so now we just need to wait for the completion string.
    my $obj = subtest( sub {
			 my ($matched, $text) = sayWAIT($conn, "",
							$finish_wait, %opts);
			 die $@ if ($@); # sayWAIT failed
			 if ($text) {
			   # Strip out any progress lines which may have been
			   # printed between syslogs or other stripped text.
			   $text =~ s/$finish_prog//mg;
			 }

			 unless ($matched) {
			   # waitstring never matched
			   die "$text\n" if ($text);
			   die "unexpected end of output\n";
			 }
			 logresult('FATAL', "EVAL_ERROR=$@") if ($@);
			 logresult('FATAL', "EVAL_ERROR=$text") if ($text);
			 logresult('PASS');
		       },
		       '-bg', '-onexit', sub{}, # don't pass results to parent
		     );

    # If specified, wait for the transfer to complete
    # and check the results once it does.
    if ($param{'COMPLETE'}) {
      Subtest::wait_finish(subtests=>[$obj]);
      die $@ unless subtestResults($obj);
    }

    if (wantarray) {
      my @ret = ($obj, $op_num);
      return @ret;
    } else {
      return $obj;
    }
  }
}

## @name ::aggrcopy_from_source
## @usage aggrcopy_from_source($filerConn, $aggrname, %options)
## @synopsis
##    Starts an aggr copy transfer to the specified aggregate on the given filer.
##    Note, this must be the source filer.  If the transfer successfully
##    starts, returns a tuple of a subtest object and the operation number.
##    The subtest object is for an active subtest which is waiting for the
##    transfer to complete. On failure, returns undef with error message in $@.
##    If the COMPLETE option is set, then this function returns whether
##    the transfer both started and completed successfully.
##    There is no need to extract results from the returned subtest object,
##    since $@ will contain the error if the transfer failed.
## @param DESTINATION Required: the destionation <filer>:<aggregate> to copy to 
## @param COPY_ALL Optional: whether or not to copy all snapshots
## @param SNAPSHOT Optional: the name of a specific snapshot on the source
##                           aggregate to copy
## @param METADATA Optional: only copy metadata and indirect blocks,
##                           not level-0 blocks
## @param TIMEOUT Optional: timeout period waiting for the command to complete
## @param COMPLETE Optional: wait for transfer to complete before returning if set
## @result
##    The subtest waiting for the transfer to finish logs a FATAL result
##    with key EVAL_ERROR if the transfer fails.  Under normal circumstances,
##    the calling script should wait for the subtest to complete and then call
##    FSR::FSRutils::subtestResults() to determine whether the transfer succeeded.
sub aggrcopy_from_source($$;%) {
  my ($conn, $srcaggr, %param) = @_;
  return eval {
    die "DESTINATION parameter is required\n" unless ($param{'DESTINATION'});

    # Setup some convenience variables and escaped strings.
    my ($dstaggr, $dstfiler) = split_filer_path($param{'DESTINATION'});
    die "invalid destination: " . $param{'DESTINATION'} . "\n"
      unless ($dstfiler);

    my $srcaggr_exp = regx_escape($srcaggr);
    my $dstaggr_exp = regx_escape($dstaggr);
    my $dstfiler_exp = regx_escape($dstfiler);

    my %opts = ();
    $opts{'TIMEOUT'} = $param{'TIMEOUT'}
      if (defined $param{'TIMEOUT'});

    # Construct the aggr copy command.
    my $cmd = aggrcopy_cmd_from_opts($srcaggr, $param{'DESTINATION'}, %param);


    # This is a hardcoded regexp which corresponds to the ONTAP code.  The
    # parentheses also determine how we determine the aggrcopy operation number
    # below. Don't change these unless you know exactly what you're doing.
    my $start_wait = "(?m)" .
      "^This can be a long-running operation. Use Control - C \\(\\^C\\) to interrupt\\.\\s+" .
      "Copy Aggregate: $srcaggr_exp on machine (?:\\S+) to Aggregate: $dstaggr_exp\\s+" .
      "AGGRCOPY: Starting on aggregate \\d+\\.".
      "\\s*\\d\\d:\\d\\d:\\d\\d (?:\\w{3}) : aggr copy restore (\\d+) : begun, \\d+ [KMGT]B to be copied\\.\\s*(?:\\n|\$)";

    my ($matched,$text) = sayWAIT($conn, $cmd, $start_wait,%opts);
    die $@ if ($@); # sayWAIT failed
    unless ($matched) {
      # waitstring never matched
      die "$text\n" if ($text);
      die "unexpected end of output\n";
    }

    # Re-match the matched string to determine the aggrcopy operation number.
    # Note, this should NEVER fail -- else we would have returned already.
    die "waitstring did not match returned string\n"
      unless ($matched =~ /$start_wait/);
    die "error determining aggrcopy operation number\n" unless (defined $1);
    my $op_num = $1;
    die "invalid operation number\n" unless ($op_num =~ /^\d+$/);

    # Any other text that we got must be an error message.
    die "$text\n" if ($text);

    my ($finish_prog, $finish_wait) = aggrcopy_regx_from_op($op_num);

    # The transfer successfully started,
    # so now we just need to wait for the completion string.
    my $obj = subtest( sub {
			 my ($matched, $text) = sayWAIT($conn, "",
							$finish_wait, %opts);
			 die $@ if ($@); # sayWAIT failed
			 if ($text) {
			   # Strip out any progress lines which may have been
			   # printed between syslogs or other stripped text.
			   $text =~ s/$finish_prog//mg;
			 }

			 unless ($matched) {
			   # waitstring never matched
			   die "$text\n" if ($text);
			   die "unexpected end of output\n";
			 }
			 logresult('FATAL', "EVAL_ERROR=$@") if ($@);
			 logresult('FATAL', "EVAL_ERROR=$text") if ($text);
			 logresult('PASS');
		       },
		       '-bg', '-onexit', sub{}, # don't pass results to parent
		     );

    # If specified, wait for the transfer to complete
    # and check the results once it does.
    if ($param{'COMPLETE'}) {
      Subtest::wait_finish(subtests=>[$obj]);
      die $@ unless subtestResults($obj);
    }

    if (wantarray) {
      my @ret = ($obj, $op_num);
      return @ret;
    } else {
      return $obj;
    }
   }
}

# This is an internal function and should NOT be called directly
sub get_volumes($$) {
  my ($filer, $aggr) = @_;

  my $obj = aggr_status($filer, $aggr);
  logresult('FATAL', "EVAL_ERROR=$@")
    if ($@);
  my $vols_in_aggr = $obj->volumes();
}

# @name ::aggrcopy_cmp_src_dst
## @usage aggrcopy_cmp_src_dst($filerSrc, $filerDst, $FILER_SRC_AGGR,
##                        $FILER_DST_AGGR, 60)
## @synopsis
##    Creates a snapshot on each volume on source and destination aggregate.
##    Then compares the source and destination volumes using treecompare. 
##    Returns 1 if they match else returns 0. Dies if an error occurs.
##    Repeats the above steps for all the volumes on source and destination
##    aggregates.
sub aggrcopy_cmp_src_dst($$$$$$) {
  my ($filerSrc, $filerDst, $fsrcname, $FILER_SRC_AGGR, $FILER_DST_AGGR, $TIMEOUT) = @_;

  my $vols_in_srcaggr = get_volumes($filerSrc, $FILER_SRC_AGGR);
  my $vols_in_dstaggr = get_volumes($filerDst, $FILER_DST_AGGR);
  logresult('FATAL', "EVAL_ERROR=The source and destination aggregates have different number of volumes, after an aggrcopy")
    unless (scalar(@{$vols_in_srcaggr}) == scalar(@{$vols_in_dstaggr}));

  my  @vols_in_src_aggr = sort @{$vols_in_srcaggr};
  my  @vols_in_dst_aggr = sort @{$vols_in_dstaggr};

  my ($srcvol, $dstvol);
  # Since treecompare can't compare at aggregate level, we compare all the
  # volumes in both source and destination aggregates. 
  while ($srcvol = pop @vols_in_src_aggr) {
    $dstvol = pop @vols_in_dst_aggr;

    logresult('FATAL', "EVAL_ERROR=Volumes on source: $srcvol and destination: $dstvol aggregates are NOT the same")
      unless ($srcvol eq $dstvol);
    # Create a snapshot	on a volume in source and in destination aggregate,
    # for treecompare
    my $src_SnapVol = new snapCmds($filerSrc, 'Vol');
    my $dst_SnapVol = new snapCmds($filerDst, 'Vol');
    my $snapname = 'snapshot_'.time_id();
    $src_SnapVol->snapshot_create($filerSrc, $srcvol, $snapname);
    logresult('FATAL', "EVAL_ERROR=$@") if ($@);
    $dst_SnapVol->snapshot_create($filerDst, $dstvol, $snapname);
    logresult('FATAL', "EVAL_ERROR=$@") if ($@);

    my $src_path = "/vol/$srcvol/.snapshot/$snapname";
    my $dst_path = "/vol/$dstvol/.snapshot/$snapname";
    logresult('FATAL', "EVAL_ERROR=treecompare found a mismatch")
      unless (treecompare($filerDst, $dst_path, $src_path, timeout => $TIMEOUT, filer_dst=>$fsrcname));

    # Cleanup the snapshots that were created for treecompare
    $src_SnapVol->snapshot_delete($filerSrc, $srcvol, $snapname);
    logresult('FATAL', "EVAL_ERROR=$@") if ($@);
    $dst_SnapVol->snapshot_delete($filerDst, $dstvol, $snapname);
    logresult('FATAL', "EVAL_ERROR=$@") if ($@);
  }

}

# @name ::aggrcopy_cmp_src_dst_ctx
## @usage aggrcopy_cmp_src_dst_ctx($filerSrc, $filerDst, $FILER_SRC_AGGR,
##                        $FILER_DST_AGGR, 60)
## @synopsis
##    Creates a snapshot on each volume on source and destination aggregate.
##    Then compares the context for the datablocks both the volumes. 
##    Returns 1 if they match else returns 0. Dies if an error occurs.
##    Repeats the above steps for all the volumes on source and destination
##    aggregates.
sub aggrcopy_cmp_src_dst_ctx($$$$$$) {
  my ($filerSrc, $filerDst, $fsrcname, $FILER_SRC_AGGR, $FILER_DST_AGGR, $TIMEOUT) = @_;

  my $vols_in_srcaggr = get_volumes($filerSrc, $FILER_SRC_AGGR);
  my $vols_in_dstaggr = get_volumes($filerDst, $FILER_DST_AGGR);
  logresult('FATAL', "EVAL_ERROR=The source and destination aggregates have different number of volumes, after an aggrcopy")
    unless (scalar(@{$vols_in_srcaggr}) == scalar(@{$vols_in_dstaggr}));

  my  @vols_in_src_aggr = sort @{$vols_in_srcaggr};
  my  @vols_in_dst_aggr = sort @{$vols_in_dstaggr};

  my ($srcvol, $dstvol);
  # Since treecompare can't compare at aggregate level, we compare all the
  # volumes in both source and destination aggregates. 
  while ($srcvol = pop @vols_in_src_aggr) {
    $dstvol = pop @vols_in_dst_aggr;
	
    logresult('FATAL', "EVAL_ERROR=Volumes on source: $srcvol and destination: $dstvol aggregates are NOT the same")
      unless ($srcvol eq $dstvol);

    my $src_path = "/vol/$srcvol";
    my $dst_path = "/vol/$dstvol";
    logresult('FATAL', "EVAL_ERROR=WAFL I/O context mismatch found")
      unless (compare_vol_ctx($filerDst,$filerSrc,$dst_path, $src_path));

   }

}

# @name ::compare_vol_ctx
## @usage cmp_vol_ctx($filerSrc, $filerDst, $FILER_SRC_AGGR,
##                        $FILER_DST_AGGR)
## @synopsis
## Compares the WAFL I/O context of each and every data block in
## the two volumes
##
sub compare_vol_ctx($$$$) {
	my($filerDst,$filerSrc,$dst_path,$src_path) = @_;
	my($showfh,$source_fh,$dest_fh);
	my($valid,$blockno,$dump_output);
	my($src_vol,$dst_vol,$src_lst,$dst_lst,$file);
	my($src_buftid,$src_FBN,$src_CP);
	my($dst_buftid,$dst_FBN,$dst_CP);

	return eval {
		
		$src_vol = substr($src_path,rindex($src_path,"/") + 1);
		$dst_vol = substr($dst_path,rindex($dst_path,"/") + 1);

		$src_lst = sayEOF($filerSrc,"priv set test;".
			"ls /vol/".$src_vol);

		logcomment("src_lst: $src_lst");

		$dst_lst = sayEOF($filerDst,"priv set test;".
			"ls /vol/".$dst_vol);
		
		logcomment("dst_lst: $dst_lst");

		if (!($src_lst eq $dst_lst)) {
			logresult('FATAL','The source and destination volume do not '.
			'have identical contents');
			return 0;
		}	 		
		my(@file_list) = split("\n","$src_lst");

		foreach $file (@file_list) {

		    if($file =~ '^\.' || $file =~ '^\..') {
			next;
		    }	
		    $source_fh = get_file_handle($filerSrc,"/vol/".$src_vol."/$file");
		    $dest_fh   = get_file_handle($filerDst,"/vol/".$dst_vol."/$file");

		    $blockno  = 0;		
		    $valid = 1;		

		    while ($valid) {

			    ($src_buftid,$src_FBN,$src_CP) = 
				    get_WAFL_IO_context($filerSrc,$source_fh,
					    $src_vol,$blockno);

			    ($dst_buftid,$dst_FBN,$dst_CP) = 
				    get_WAFL_IO_context($filerDst,$dest_fh,
					    $dst_vol,$blockno);
				
			    if (!(($src_buftid eq $dst_buftid) && ($src_FBN eq $dst_FBN) 
				&& ($src_CP eq $dst_CP))) {
					logresult('FATAL','The source and destination' . 
				' context do not match');
			    }

			    if ($src_buftid == -1 && $dst_buftid == -1 ) {
				last;
			    } 		
			    $blockno++;
		    }
		}

		return 1;
	}
}

# @name ::get_WAFL_IO_context
## @usage get_WAFL_IO_context($filer,$file_handle,$vol,$blockno)
## @synopsis
## Returns the WAFL I/O context for a particular block in the file 
##
sub get_WAFL_IO_context ($$$) {
	my($filer,$fh,$vol,$blockno) = @_;
	my($dumpinfo);
	my($buftid,$fbn,$cp);

	$dumpinfo = sayEOF($filer,
		"priv set test;dumpblock info -V $vol $fh $blockno");		

	if ($dumpinfo =~ /ckse-buftreeid ([\d\w]+)\nckse-fbn (\d+)\nckse-cp-count (\d+)/){
	    $buftid = $1;
	    $fbn = $2;
	    $cp = $3;
	}
	elsif ($dumpinfo =~ /block is a hole/){
	    $buftid = 0;
	    $fbn = 0;
	    $cp = 0;
	}
	elsif($dumpinfo =~ /inode has only/){
	    $buftid = -1;
	    $fbn = -1;
	    $cp = -1;
	} 
	elsif($dumpinfo =~ /could not load inode/){
	    $buftid = -1;
	    $fbn = -1;
	    $cp = -1;
	}
	else{
	    logresult("FATAL", "Unable to parse dumpinfo output: $dumpinfo");
	}
 
	return ($buftid, $fbn, $cp);
}



# @name ::get_file_handle
## @usage get_file_handle($filer,$filename)
## @synopsis
## Gets the file handle for a particular file on the filer
##

sub get_file_handle($$) {

	my($filer,$filename) = @_;
	my($showfh,$fh);

	$showfh = sayEOF($filer,"priv set test;showfh $filename");
	die $@ unless ($showfh =~ "fileid=");
	if ($showfh =~ "fileid=([0-9x]+)") {
		$fh = $1;
	}	
	return $fh;
}

1;
