# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Aggregate/ImplTemplate.pm#6 $

package Aggregate::ImplTemplate;
use strict;

# OK, so what does this do you ask?
#
#         new_with_init => 'new'
#     line creates a constructor new() which
#     will call function init() (defined below)
#     right before it returns. For our purposes
#     new() takes a string as an argument.
#
#      new_hash_init => 'hash_init'
#     line creates function hash_init() that
#     accepts hash where each key-value pair is
#       class_var_method => argument_to_that_method
#     so this particular one will let you assign
#     values to 'cmd' and 'validation' since we get nice
#     accessors to them (see below)
#
#      get_set       => [ qw(cmd validation) ]
#    line creates class data vars $cmd and $validation
#    as well as accesor functions for them. So
#    we can do
#        $implTemplateInstance->cmd()
#    to get the value of $cmd
#    and
#        $implTemplateInstance->cmd('stuff')
#    to assign 'stuff' to $cmd
use Class::MethodMaker
        new_with_init => 'new',
        new_hash_init => 'hash_init',
        get_set       => [ qw(cmd validation) ]
;

use Tharn;
use Data::Dumper;
use Module::Common qw(validate_mixed);
use Aggregate::Status;
use Volume::Status;
#use Volume::Clone qw(:all);
use Utils qw(make_host make_conn);

use Filer::Version qw(:releases get_wafl_version  UNSUPPORTED_FEATURE_EXCEPTION are_we_ontapng);

use SPARSE::Util::Logger;

use constant ANCHORSTEAM => '7.0';
use constant SCRIMSHAW   => '6.5';

our($VERSION);

# Construct the common regular expression that is used to extract the RAID
# disk information given by aggr status -(s|r)
my $disk_status_re_str  = join( '\s+',
        (
                       '\s*(spare|dparity|parity|data)',   # raid type
                       '([\w.]+)',                         # Device
                       '([\w]+)',                          # HA
                       '([-\d?]+)',                        # Shelf
                       '([-\d?]+)',                        # Bay
                       '([\w:]+)',                         # channel
                       '([-\w]+)',                         # Pool
                       '(?:(\w+)',                         # Type
                       '(\w+|\w+\/\w+))?',                 # RPM
                       '(\d+)\/(\d+)',                     # Used (MB blks)
                       '(\d+)\/(\d+)'                      # Phys (MB blks)
        )
);

my $disk_status_re = $disk_status_re_str;

use constant RD_ZEROED    => 'ZEROED';
use constant RD_ZEROING  => 'ZEROING';
use constant RD_NOTZEROED => 'NOTZEROED';

## @summary Library of aggregate commands
## @author dl-johny@netapp.com (FSS WAFL Quality)
## @cf     false
## @status private
## @pod here

BEGIN
{
        $VERSION    = '0.02';

        my @df_space_types  = qw( DF_SPACE DF_INODES DF_PRIVATE DF_RESERVE );

        Module::Common::setup_export_tags(
                df  => [ @df_space_types ]
        );

}
use constant VOL_CMD => 'vol';
use constant AGGR_CMD => 'aggr';

use Params::Validate
        qw(validate_pos validate BOOLEAN SCALAR ARRAYREF HASHREF OBJECT UNDEF);
use SayWrap qw(sayEOF);
use Utils qw(make_conn_and_complain);
use RaidInfo qw(aggregate);
        # XXX - upon popular request, make_conn() shall be used to allow
        #       hostspec/Hostspec/Connectrec as the "filer" argument. But
        #       we want to make sure to complain if it's passed anything
        #       other than a Connectrec, so we will use
        #       make_conn_and_complain(), which will "complain loudly"
        #       if anything other than a Connectrec is passed.

sub init
{
        my $self = shift;

        $self->hash_init(@_);

        if ($self->cmd eq VOL_CMD) {
                $self->cmd('vol');
        } elsif ($self->cmd eq AGGR_CMD) {
                $self->cmd('aggr');
        } else {
                Carp::croak("No such type supported for " .
                        __PACKAGE__ . ": " . $self->cmd);
        }

        # create a status object
        $self->{status} =
                Aggregate::ImplTemplate::Status->new(cmd=>$self->cmd(),
                         validation=>$self->validation()
                );

}

my %ConnValidate = (
        conn => {
                type => SCALAR | OBJECT,
                callbacks => {
                        'is valid object' => sub {
                                ref($_[0]) ? (
                                UNIVERSAL::isa($_[0], 'Hostrec') ||
                                UNIVERSAL::isa($_[0], 'Connectrec')) : 1;
                        }
                }
        }
);

my %DiskListValidate = (
        type => SCALAR | ARRAYREF
);

my %TargetValidate = (
      target => {
                type => SCALAR
        }
);

my %BasicValidate = (
        %ConnValidate,
        %TargetValidate
);

# This subroutine is a wrapper around SayWrap::sayEOF() except that
# it checks for config_lock errors which is a transient error on
# volumes. In case of such errors, the command needs to be re-issued
# shortly afterwards.
#
sub sayEOF_wrapper
{
        my $conn = shift;
        my $cmd = shift;
        my @args = @_;

	my $text = undef;
	my $t = 6;

	do {
	    $text = sayEOF($conn, $cmd, @args);
	}
	while( $text =~ /in use by other operations/ && --$t && sleep(10) );

	return $text;
}

sub cmd_create
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);

        # Variable which determines whether the privilege level,
        # should be modified or not, privilege level is modified,
        # based on Burt 631946
        my $change_priv=0;

        my ($target, $disklist, %args) = validate_mixed(
                @_,
                [{type => SCALAR},
                 {type=> SCALAR|ARRAYREF}],
                {
                force    => { type => BOOLEAN, optional => 1 },
                language => { type => SCALAR, optional => 1 },
                mirror   => { type => BOOLEAN, optional => 1 },
                raidtype => {
                        type => SCALAR,
                        optional  => 1,
                        callbacks => {
                                'valid raid type' => sub {
                                        $_[0] eq 'raid4' or $_[0] eq 'raid_dp'
                                }
                        }
                },
                randomize_args => { type => BOOLEAN, optional => 1, default => 1 },
                rgsize  => { type => SCALAR, optional => 1 },
                snaplock => { type => BOOLEAN, optional => 1 },
                snaplock_enterprise => { type => BOOLEAN, optional => 1 },
                timeout => {
                        type => SCALAR,
                        optional => 1,
                        default => 7200,    # 2 hrs for slow ATA filers
                        regexp => qr(^\d+$) # got to look like a number
                },

                #Support for 64-bit aggr naveenm
                indirect_type => {
                        type => SCALAR | UNDEF,
                        optional => 1,
                        callbacks => {
                              'valid value' => sub {
                                       $_[0] eq '32' or $_[0] eq '64' or $_[0] eq undef
                               }
                        }
                },
                purge   => { type=> BOOLEAN, default => 0 }
        } );

        return eval {
                # collect the options that needs to be passed to aggr create
                my %opts;
                if($args{force}) {
                        $opts{ '-f' } = '';
                }
                if($args{mirror}) {
                        $opts{ '-m' } = '';
                }
                if($args{rgsize}) {
                        $opts{ '-r' } = $args{rgsize};
                }
                if($args{raidtype}) {
                        $opts{ '-t' } = $args{raidtype};
                }
                if($args{language}) {
                        $opts{ '-l' } = $args{language};
                }

                #64-bit support naveenm
                if($args{indirect_type}) {
                    $opts{ '-B' } = $args{indirect_type};
                    $change_priv=1;
                } else {
		    # if the indirect type was not provided, then check whether
		    # the LARGE_AGGR parameter has been defined. If so,
		    # set the -B option accordingly.
		    my $large_aggr = $main::LARGE_AGGR;
		    if((defined($large_aggr)) && ($large_aggr == 0)) {
			$opts{ '-B' } = '32';
			$change_priv=1;
		    }
		}

                if ($args{snaplock} || $args{snaplock_enterprise}) {
                        if (get_wafl_version($conn->{hostp}) <= WV_ANCHORSTEAM_MAX) {
                        $opts{ '-L' } = '';
                        # this protects us from SCRIMSHAW confirmation prompts
                        $opts{ '-f' } = '';
                    } else {
                        # snaplock argument is used to create compliance aggregates
                        $opts{ '-L' } = ($args{snaplock} ? "compliance" : "enterprise");
                    }
                }
                
                my $disklist_str;
                if(ref($disklist) eq 'ARRAY') {
			if (ref($disklist->[0]) eq 'ARRAY') {
				$disklist_str = "";
				foreach my $elm (@{$disklist}) {
					$disklist_str .= join(' ',
							('-d', @{$elm}, ' '));
				}
				$opts{ '-m' } = '' if (@{$disklist} == 2);
			} else {
	                        $disklist_str = join(' ', ('-d', @{$disklist}));
			}
                } else {
                        $disklist_str = $disklist;
                }

                # XXX - tapl wanted randomization, so here's some randomization
                # XXX - if this is used across the board, it may be a good idea
                #       to factor it out to a module.
                my @cmd_args;
                if($args{randomize_args}) {
                        my @tmp_opts = keys %opts;
                        while(@tmp_opts) {
                                my $opt = splice(@tmp_opts, rand(@tmp_opts), 1);
                                push @cmd_args, ($opt, $opts{$opt});
                        }
                } else {
                        @cmd_args = %opts;
                }
                # purge if needed
                if ($args{purge}) {
                        # due to burt 117042 we need to destroy
                        # the aggregate with name $target if it exists
                        # when creating volume
                        if ($self->cmd() eq VOL_CMD) {
                                if ($self->cmd_exists($conn, $target)) {
                                        $self->cmd_purge($conn, $target);
                                }
                        }

                        $self->cmd_purge($conn,$target)
                                or die $@;
                }

                # Fix for Burt 651221, code modified to make the script
                # compatible with Burt 631946
                my $priv_set_cmd="";
                if($change_priv == 1){
                         $priv_set_cmd="priv set -q advanced;";
                }

                my $cmd = sprintf(
                    "%s create %s %s %s",
                    $priv_set_cmd . $self->cmd(),
                    $target,
                        join( ' ', @cmd_args ),
                        $disklist_str
                );

                my $conntype = $conn->conntype();

                # ensure an rsh connection - to avoid low disks prompt:
                    # low spare disks warning from 'vol create':
                    #
                    # WARNING! Continuing with vol create will result in having
                    # no spare disk available for one or more raid groups.
                    # Are you sure you want to continue with vol create?
                my $temp_conn = $conn;
                if ($conn->conntype !~ /^rsh/) {
                    Tharn::host($conn->hostp);
                    $temp_conn = Tharn::connect('#rsh');
                }
                logDbg("original conn type = " . $conn->conntype);
                logDbg("temp conn type = " . $temp_conn->conntype);

                my $text = sayEOF_wrapper($temp_conn, $cmd, timeout=>$args{timeout});

                # it's a CONFIG error if there are not enough disks available
                die config_result($text) if ( $text =~ /\d+ disks needed, but only \d+ matching disks? (is|are) available/);

                die $text unless ( $text =~ /has\s+been\s+initiated/ ||
                                   $text =~ /has\s+completed/ );

                # flush the original connection's buffer
                $conn->discardbuff();

                if($self->validation()) {
                        my $sleeptime;
                        my $sleeptime_cap;
                        my $ready = 0;
                        my $endtime = time() + $args{timeout};
                        my $container;

                        if ($self->cmd() eq VOL_CMD) {
                            $container = "volume";
                            $cmd = "priv set -q diag; " .
                                "raid_config info listvol";
                        } else {
                            $container = "aggregate";
                            $cmd = "priv set -q diag; " .
                                "raid_config info listaggr";
                        }

                        # sleep for either 5 or 15 secs for sims or hardware
                        # filers respectively. it's not gonna finish
                        # so quickly. this is to avoid calling raid_config
                        # immediately (to reduce output) since we already know
                        # container creation will take quite a while.
                        #
                        # likewise, cap the sleep time at 2 or 5 mins for
                        # sim or hardware filer respectively
                        if ($conn->hostp()->hosttype() eq 'filer-simulator') {
                                # simulators
                                $sleeptime = 5;
                                $sleeptime_cap = 2*60;
                        } else {
                                # hardware filer
                                $sleeptime = 15;
                                $sleeptime_cap = 5*60;
                        }
                        sleep $sleeptime;

                        while (time() < $endtime) {

                                # avoid calling RaidInfo. this method produces less
                                # output.
                                my $output = sayEOF_wrapper($conn, $cmd);
                                my @info = split(/\n/, $output);
                                foreach my $line (@info) {
					# see prod/common/raidv2/config.c starting @ line 803, function
					# volcmd_util_print_legacy_vol_info(legacy_volume_info_t *legacy_info); and
					my %aggrinfo = split( '[=,]', $line );	# prod/common/wafl/typedefs.h, line 338, typedef enum { ... } } wafl_Mount_state;

					if( $aggrinfo{name} eq $target ) {
					    # reset timer if still creating: typedef wafl_Mount_state: WAFL_VOL_CREATING (prod/common/wafl/typedefs.h)
					    $endtime = time() + $args{timeout} if( $aggrinfo{state} == 4 ); # reset timer if still creating
					    # return if aggregate is ready.
					    return 1 if( $aggrinfo{fsm_state} eq 'NORMAL' && $aggrinfo{state} != 4 && $aggrinfo{state_name} eq 'online');
					}
                                }
                                # exponential backoff of increasing sleep time,
                                # capped at $sleeptime_cap depending on type
                                # of filer
                                $sleeptime *= 2;
                                if ($sleeptime > $sleeptime_cap) {
                                        $sleeptime = $sleeptime_cap;
                                }
                                sleep($sleeptime);
                        }
                        die "Timed out waiting for $container '$target' " .
                                "to complete creation";
                }
                return 1;
        };
}


sub cmd_destroy
{
        my $self     = shift;
        my $conn     = make_conn_and_complain(shift);
        my ($target, %args) = validate_mixed(
                @_,
                [ {type => SCALAR} ],
                {
                   destroy_corrupt_volumes =>
			{ type => BOOLEAN, optional => 1, default => 0 },
		}
	);

        return eval {
                my $target_re = quotemeta($target);
                my $cmd = sprintf(
                        "%s destroy %s -f",
                        $self->cmd(),
                        $target
                );
                # my $text = sayEOF_wrapper($conn, $cmd);
		my $persist_conn;
		my $new_conn_created = 0;
		logDbg("cmd_destroy \"$target\"  destroy_corrupt_volumes: \"$args{destroy_corrupt_volumes}\"");
		if ($args{destroy_corrupt_volumes} &&
		    $conn->conntype() !~ /telnet|cons/) {
			$new_conn_created = 1;
			my $host = $conn->hostp();
			$persist_conn = make_conn($host, "#telnet");
			die("could not create new 'rsh-persist' connection: $@") if ($@);
			$persist_conn->timeout($conn->timeout());
		} else {
			$persist_conn = $conn;
		}
                my (undef, undef, $text) = say($persist_conn, $cmd, '-prompt', '\?');
		if ($args{destroy_corrupt_volumes} &&
		     $text =~ /This\svolume\sis\smarked\scorrupt|This\smessage\sonly\sappears\son\sdevelopment\skernels/) {
			(undef, undef, $text) = say($persist_conn, "y");
		}
		disconnect($persist_conn) if ($new_conn_created);
                if ($text !~ /'$target_re' destroyed/) {
                        die $text;
                }

                if ($self->validation()) {
                        # if target is destroyed, it does not exists anymore
                        !$self->cmd_exists($conn, $target) or die $@;
                }

                return 1;
        };
}

sub cmd_undestroy
{
        my $self     = shift;
        my $conn     = make_conn_and_complain(shift);
        my ($target, %args) = validate_mixed(
                @_,
                [ { type => SCALAR } ],
                {
                }
        );

        return eval {
                my $target_re = quotemeta($target);
                my $cmd = sprintf(
                        "%s undestroy %s",
                        $self->cmd(),
                        $target
                );
                # my $text = sayEOF_wrapper($conn, $cmd);
                my $persist_conn;
                my $new_conn_created = 0;
                logDbg("cmd_undestroy \"$target\"");
                if ($conn->conntype() !~ /telnet|cons/) {
                        $new_conn_created = 1;
                        my $host = $conn->hostp();
                        $persist_conn = make_conn($host, "#telnet");
                        die("could not create new 'rsh-persist' connection: $@") if ($@);
                        $persist_conn->timeout($conn->timeout());
                } else {
                        $persist_conn = $conn;
                }

                my (undef, undef, $text) = say($persist_conn, $cmd, '-prompt', '\?|\>');
                if ($text =~ /undestroy aggregate/) {
                        (undef, undef, $text) = say($persist_conn, "2");
                } else {
			return $text;
		}

                disconnect($persist_conn) if ($new_conn_created);
                if ($text !~ /'$target_re' undestroyed/) {
                        die $text;
                }

                if ($self->validation()) {
                        # if target is undestroyed, it should exists
                        $self->cmd_exists($conn, $target) or die $@;
                }

                return 1;
        };
}

sub cmd_add
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);
        my($target, $disklist, %args) = validate_mixed(
                @_,
                [ { type => SCALAR},
                  { type => SCALAR|ARRAYREF }
                ],
                {
                        force   => { type => BOOLEAN, optional => 1 },
                        rg      => { type => SCALAR, optional => 1 },
                        show    => { type => BOOLEAN, optional => 1 },
                        # number of 4K blocks to skip, to create a hole for fake large aggr. 
                        # Can be specified in decimal or hex (with 0x prefix).
                        skip    => { type => BOOLEAN, optional => 1 },
                        timeout => { type => SCALAR, optional => 1,
                                     default => 45 * 60 },
                        '64bit-upgrade' => { type => SCALAR, optional => 1},
                }
        );

        return eval {
                my %opts = ();
                if ($args{force}) {
                        $opts{ '-f' } = '';
                }
                if ($args{rg}) {
                        $opts{ '-g' } = $args{rg};
                }
                if ($args{show}) {
                        $opts{ '-n' } = '';
                }
                if ($args{skip}) {
                        $opts{ '-s' } =  $args{skip};
                }
                if ($args{timeout} <= 0) {
                        $args{timeout} = 45 * 60;
                }
                if ($args{'64bit-upgrade'}) {
                        $opts{'-64bit-upgrade'} = $args{'64bit-upgrade'};
                }

                my $disklist_str;
                if (ref($disklist) eq 'ARRAY') {
                        $disklist_str = join( ' ', ( '-d', @{$disklist} ) );
                } else {
                        $disklist_str = $disklist;
                }

                my $cmd = sprintf(
                        "%s add %s %s %s",
                        $self->cmd,
                        $target,
                        join( ' ', map{ ($_ => $opts{$_}) } keys %opts ),
                        $disklist_str
                  );

                my $text = sayEOF_wrapper($conn, $cmd);
                if ($text !~ /has\s+been\s+initiated/m && $text !~ /has completed/
		    && $text !~ /requires volume to be quiesced/) {
                        die $text;
                }

                if (!$self->validation) {
                        return 1;
                }

                # Do the validation here

                # wait until state = online and status is NOT growing
                $cmd = sprintf(
                        "%s status -r %s",
                        $self->cmd,
                        $target
                );

                my $success = 0;
                my $endtime = time() + $args{timeout};
                my ($state, $status);
                while(1) {
                        last if time() > $endtime;

                        # Aggregate aggr (online, raid4, growing) (block checksums)
                        #   Plex /vol0/plex0 (online, normal, active)
                        #     RAID group /tap/plex0/rg0 (normal)
                        $text = sayEOF_wrapper($conn, $cmd);
                        if($text =~ /^\s*(Aggregate|Volume)\s+.*\s+\((\w+),\s+(.*)\s+\(/m) {
                                ($state, $status) = ($2, $3);
                                if($state eq 'online' &&
                                   $status !~ /growing/) {
                                        $success = 1;
                                        last;
                                }
                        }
                        select(undef,undef,undef,10);
                }
                if(!$success) {
                        die "Timed out waiting for '$target' to come online. State: $state, Status: $status";
                }
                return 1;
        };
}

sub cmd_purge
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);
        my($target, %args) = validate_mixed(
                @_,
                [ { type => SCALAR} ],
                {
                        # 'vol_only' option only valid in the aggregate
                        # namespace
                        vol_only => {
                                type      => BOOLEAN,
                                optional  => 1,
                                callbacks => {
                                              'Only applies to aggregates' =>
                                              sub {
                                                      $self->cmd() eq AGGR_CMD
                                              }
                                 }
                        },
                       clone_purge => {
                                type => BOOLEAN,
                                default => 1,
                       },
                       busy_wait => {
                                     type => BOOLEAN,
                                     default => 1,
                       },
		       destroy_corrupt_volumes => {
				     type => BOOLEAN,
				     optional => 1,
				     default => 0,
		       },
                }
        );
        return eval {
                if ($args{busy_wait}) {
                        my $out = sayEOF_wrapper($conn, $self->cmd()." status $target");
                        if ($out =~ 'busy') {
                                if ($self->cmd() eq AGGR_CMD) {
                                        Aggregate::aggr_wait_change($conn, $target,
                                                                    want => { state => qr/^(?:on|off)line$/ },
                                                                    timeout=>120
                                                                   ) or die $@;
                                } else {
                                        Volume::vol_wait_change($conn, $target,
                                                                want => { state => qr/^(?:on|off)line$/ },
                                                                timeout=>120 #the default timeout of 900 is too long.
                                                               ) or die $@;
                                }
                        }
                }

                # offline and destroy all flexible volumes in
                # the aggregate first
                if ($self->cmd() eq AGGR_CMD && $self->cmd_exists($conn, $target)) {

                        my $obj = $self->cmd_status($conn, $target) or die $@;

                        # if $target is a trad vol, there will be no list
                        # of volumes we can iterate through.
                        if (defined @{$obj->{volumes}}) {
                                my @vols = @{$obj->{volumes}};

                                foreach my $vol (@vols) {
                                        # cannot recursively call back to self
                                        # as we are in the 'aggr' namespace. to
                                        # purge flex vols, we need to be in the
                                        # 'vol 'namespace. hence to do that,
                                        # directly call Volume::vol_purge()
                                        # to force ourselves back to the 'vol'
                                        # namespace.
                                        Volume::vol_purge($conn, $vol)
                                            or die $@;
                                }
                        }
                }

                if ($self->cmd() eq VOL_CMD
                    && Volume::vol_exists($conn, $target)) {
                        # check for unremovable snaplock volumes under scrimshaw
                        if (get_wafl_version($conn->{hostp}) <= WV_SCRIMSHAW_MAX) {
                            my %options = Volume::vol_options($conn, $target) or die $@;

                            logDbg("cmd_purge: options is " . Dumper(\%options) . "\n");

                            # 'snaplock compliance' licensing does not support
                            # volume destruction prior to the anchorsteam (7.0)
                            # release.
                            #
                            # after AS, such volumes can be destroyed if
                            # all contained files have expired.
                            #
                            # note that 'snaplock enterprise' licensing
                            # does permit snaplock volume destruction
                            if (exists $options{snaplock_compliance}) {
                                logTrc("cmd_purge: target is a snaplock compliance volume: '$target'");
                                die na_result(UNSUPPORTED_FEATURE_EXCEPTION);
                            }
                        }

                        if ($args{clone_purge}) {
                            # destroy all of the volume's clones first

                            # XXX all kinds of evilness occurs in
                            # vvol_sanity.thpl if we use Volume::Clone,
                            # so require it
                            require Volume::Clone;
                            my @clones = Volume::Clone::get_clones(conn=>$conn,
                                                                   vol=>$target);
                            if ($@) {die $@};
                            foreach my $clone (@clones) {
                                    Volume::vol_purge($conn, $clone)
                                              or die $@;
                            }
                    }
                }

                # check if user wants to preserve the aggregate
                if ( (!exists($args{vol_only}) ||
                      !$args{vol_only} ) &&
                    $self->cmd_exists($conn, $target)) {
                        $self->cmd_offline($conn, $target)
                          or die $@;
                        $self->cmd_destroy(  $conn,
					     $target,
                                             destroy_corrupt_volumes => $args{destroy_corrupt_volumes})
                          or die $@;
                }

                return 1;
        };
}


sub cmd_online
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);
        my($target, %args) = validate_mixed(
                @_,
                [ {type => SCALAR} ],
                {
                        force   => { type => BOOLEAN, optional => 1 },
                }
        );

        return eval {
                my %opts = (force => '');

                if(defined $args{force}) {
                        $opts{force} = "-f";
                }

                my $target_re = quotemeta($target);
                my $cmd  = sprintf(
                        "%s online %s %s",
                        $self->cmd,
                        $target,
                        $opts{force}
                );

                my $text = sayEOF_wrapper($conn, $cmd);

                # cmd output is compatible SS(6.5) onwards
                # aggr command is present from AS(7.0) onwards
                my $re = $self->cmd eq AGGR_CMD ?
                        qr<\s*(?:aggr online:\s+)?(?:Aggregate|Volume) ('$target_re'|$target_re) is (?:now|already) online\.> :
                        qr<\s*(?:vol online:\s+)?Volume ('$target_re'|$target_re) is (?:now|already) online\.>;

                if($text !~ /$re/) {
                        die $text;
                }

                if($self->validation()) {
                        $self->cmd_status($conn, $target)
                                or die $@;
                        my $state = $self->{status}->state();

                        if(defined($state) && $state ne 'online') {
                                die "Unable to get $target online (state=$state)";
                        }
                }
                return 1;
        };
}


sub cmd_offline
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);
        my($target, %args) = validate_mixed(
                @_,
                [ {type => SCALAR} ],
                {
                        force         => { type => BOOLEAN, optional => 1 },
                        cifsdelaytime => { type => SCALAR, optional => 1 },
                        # 'all' option only valid in the aggregate namespace
                        all => {
                                type      => BOOLEAN,
                                optional  => 1,
                                callbacks => {
                                              'Only applies to aggregates' =>
                                              sub {
                                                      $self->cmd() eq AGGR_CMD
                                              }
                               }
                        },

                }
        );

        return eval {
                $args{cifsdelaytime} = 0 unless $args{cifsdelaytime};

                # to avoid Perl's warning of using of non-initialized value
                my %opts = (
                        force         => '',
                        cifsdelaytime => '',
                        all           => '',
                        priv          => ''
                );

                if(defined $args{force}) {
                        $opts{force} = "-F";
                        $opts{priv} = "priv set -q diag;";
                }

                 if(defined $args{all}) {
                        $opts{priv}  = "priv set -q diag;";
                        $opts{all}   = "-a";
                        $opts{force} = "-F";
                }

                # -t and -f options are mutually exclusive
                if(defined $args{cifsdelaytime}
                   && !defined($args{force})
                   && !defined($args{all})) {
                        $opts{cifsdelaytime} = "-t $args{cifsdelaytime}";
                }

                my $target_re = quotemeta($target);
                my $cmd  = sprintf(
                        "%s %s offline %s %s %s %s",
                        $opts{priv},
                        $self->cmd(),
                        $target,
                        $opts{all},
                        $opts{cifsdelaytime},
                        $opts{force}
                );

                my $text = sayEOF_wrapper($conn, $cmd);
                # cmd output is compatible SS(6.5) onwards
                # aggr command is present from AS(7.0) onwards
                my $re = $self->cmd eq AGGR_CMD ?
                        qr<\s*(?:aggr offline:\s+)?(?:Aggregate|Volume) ('$target_re'|$target_re) is (?:now|already) offline\.> :
                        qr<\s*(?:vol offline:\s+)?Volume ('$target_re'|$target_re) is (?:now|already) offline\.>;
                if($text !~ /$re/) {
                        die "$text\n";
                }

                if($self->validation) {
                        $self->cmd_status($conn, $target)
                                or die $@;
                        my $state = $self->{status}->state();

                        if(defined($state) &&
                           $state ne 'offline') {
                                die "Unable to get $target offline (state=$state)";
                        }
                }
                return 1;
        };
}


sub cmd_restrict
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);
        my($target, %args) = validate_mixed(
                @_,
                [ {type => SCALAR} ],
                {
                        force         => { type => BOOLEAN, optional => 1 },
                        cifsdelaytime => { type => SCALAR, optional => 1 },
                        # 'all' option only valid in the aggregate namespace
                        all => {
                                type      => BOOLEAN,
                                optional  => 1,
                                callbacks => {
                                              'Only applies to aggregates' =>
                                              sub {
                                                      $self->cmd() eq AGGR_CMD
                                              }
                               }
                        },
                }
        );

        return eval {
                # to avoid Perl's warning of using of non-initialized value
                my %opts = (
                        force         => '',
                        cifsdelaytime => '',
                        all           => '',
                        priv          => ''
                );

                if(defined $args{force}) {
                        $opts{force} = "-F";
                        $opts{priv} = "priv set -q diag;";
                }

                 if(defined $args{all}) {
                        $opts{priv}  = "priv set -q diag;";
                        $opts{all}   = "-a";
                        $opts{force} = "-F";
                }

                if(defined $args{cifsdelaytime}) {
                        $opts{cifsdelaytime} = "-t $args{cifsdelaytime}";
                }

                my $target_re = quotemeta($target);
                my $cmd  = sprintf(
                        "%s %s restrict %s %s %s %s",
                        $opts{priv},
                        $self->cmd(),
                        $target,
                        $opts{all},
                        $opts{cifsdelaytime},
                        $opts{force}
                );

                my $text = sayEOF_wrapper($conn, $cmd, 'timeout' => 240);
                # cmd output is compatible SS(6.5) onwards
                # aggr command is present from AS(7.0) onwards
                my $re = $self->cmd eq AGGR_CMD ?
                        qr<\s*(?:aggr restrict:\s+)?(?:Aggregate|Volume) ('$target_re'|$target_re) is (?:now|already) restricted\.> :
                        qr<\s*(?:vol restrict:\s+)?Volume ('$target_re'|$target_re) is (?:now|already) restricted\.>;

                if($text !~ /$re/) {
                        die $text;
                }

                if($self->validation) {
                        $self->cmd_status($conn, $target)
                                or die $@;
                        my $state = $self->{status}->state();

                        if(defined($state) &&
                           $state ne 'restricted') {
                                die "Unable to get $target restricted (state=$state)";
                        }
                }
                return 1;
        };
}




sub cmd_rename
{
        my $self = shift;
         my $conn = make_conn_and_complain(shift);
        my($from, $to) = validate_pos(@_,
                {type => SCALAR},
                {type => SCALAR}
        );

        return eval {
                my $from_re = quotemeta($from);
                my $to_re   = quotemeta($to);

                my $cmd = sprintf(
                        "%s rename %s %s",
                        $self->cmd,
                        $from,
                        $to
                );

                my $expected;
                if (get_wafl_version($conn->{hostp}) >= WV_ANCHORSTEAM_MIN) {
                        # AS(7.0) and later
                        $expected = "'$from_re' renamed to '$to_re'";
                } else {
                        # SS(6.5)
                        $expected = "$from_re renamed to $to_re";
                }
                # We are not going below SS(6.5)

                my $text = sayEOF_wrapper($conn, $cmd);
                if ($text !~ /$expected/) {
                        die "$text\n";
                }

                if($self->validation) {
                        # old target should not exist anymore
                        !$self->cmd_exists($conn, $from) or die $@;

                        # new target better be there
                        $self->cmd_exists($conn, $to) or die $@;
                }
                return 1;
        };
}


sub cmd_read_fsid
{
        my $self = shift;
         my $conn = make_conn_and_complain(shift);
        my ($target) = validate_pos(@_, {type => SCALAR});

        return eval {
                my $targetexp = quotemeta($target);
                my $cmd = sprintf("priv set -q advanced; %s read_fsid %s",
                                  $self->cmd(),
                                  $target);
                my $fsid = 0;

                my $text = sayEOF_wrapper($conn, $cmd);
                if ($text =~ /\s+'?$targetexp'?\s+has\s+an\s+FSID\s+of\s+(\w+)\./) {
                        $fsid = $1;
                } else {
                        die $text;
                }

                return $fsid;
        };
}

sub cmd_rewrite_fsid
{
        my $self = shift;
         my $conn = make_conn_and_complain(shift);
        my($target, %args) = validate_mixed(
                @_,
                [ {type => SCALAR} ],
                {
                    fsid  => { type => SCALAR },
                    force => { type => SCALAR, optional => 1 },
                }
        );

        return eval {
                my $fsid     = 0;
                my $old_fsid = 0;
                my $cmd;
                my $targetexp = quotemeta($target);

                if(defined $args{fsid}) {
                        $fsid = $args{fsid};
                }

                if (!defined $args{force} ||
                    get_wafl_version($conn->hostp()) <= WV_SCRIMSHAW_MAX) {
                        # no -f option in scrimshaw
                        $cmd = sprintf("priv set -q advanced; " .
                                      "%s rewrite_fsid %s %s",
                                      $self->cmd(),
                                      $target,
                                      $fsid);
                } else {
                    $cmd = sprintf("priv set -q test; " .
                                   "%s rewrite_fsid %s %s -f",
                                   $self->cmd(),
                                   $target,
                                   $fsid);
                }

                my $text = sayEOF_wrapper($conn, $cmd);

                # note: if the regexp string does not contain the requested
                # FSID to be set, it will die.
                if ($text =~ /\s+'?$targetexp'?\s+has\s+had\s+its\s+FSID\s+changed\s+from\s+(\w+)\s+to\s+$fsid\./) {
                        $old_fsid = $1;
                } else {
                        die $text;
                }

                # double check if the container was set to the correct FSID
                if ($self->validation()) {
                        my $set_fsid = $self->cmd_read_fsid($conn, $target);
                        if ($set_fsid ne $fsid) {
                                die "Requested FSID $fsid was not set. Container still have FSID $set_fsid";
                        }
                }

                return $old_fsid;
        };
}

sub cmd_list
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);

        return eval {
                my @names;

                my $cmd  = sprintf( "%s status", $self->cmd );
                my $text = sayEOF_wrapper($conn, $cmd);
                my @lines = split( /\r?\n/, $text );

                my $re = $self->cmd eq AGGR_CMD ?
                        qr<^\s*Aggr\s+State\s+Status\s+Options\s*$> :
                        qr<^\s*Volume\s+State\s+Status\s+Options\s*$>;
                while (my $line = shift @lines) {
                        # find the headers...
                        if ($line =~ /$re/) {
                                last;
                        }
                }

                if (!@lines) {
                        die "Did not get appropriate headers from output of $cmd";
                }

                while ( my $line = shift @lines ) {
                        # we only care about the name of the volume
                        next if($line =~ /vol status:/);
                        if( $line =~ /^\s*(\S+)\s+(\S+)\s+(\S+,?\s+)+\s*(\S+,?)*$/) {
                                my $vol_name=$1;
                                my $state=$2;
                                my $options=$3;
                                if($state =~ /online|offline|restrict|failed/){
				  push(@names, $vol_name) unless($options =~ /snaplock/);
				}
                        }
                }
                return @names;
        }
}


sub cmd_match_plex_raid_info
{
        my $self = shift;
        my ($target_re, $text) = validate_pos(@_,
                        {type => SCALAR},
                        {type => SCALAR}
                        );

        die "no text supplied" unless $text;

        my $plex = '^\s*Plex /'. $target_re .
                        '/(plex\d+): (?:\w+, )*\w+\s*(?:\n|\z)';

        my $raid = '^\s+RAID group /'. $target_re .
                        '/(plex\d+)/(rg\d+): \w+\s*(?:\n|\z)';

        while ($text && ($text =~ s/$plex//s)) {
                my $plex_name = $1;
                while ($text && ($text =~ s/$raid//s)) {
                        die "found RAID group in different plex"
                                unless ($1 eq $plex_name);
                }
        }
        $text =~ s/\A\s+//;
        $text =~ s/\s+\z//;
        return $text;
}


sub cmd_options
{
        # individual params validation is found in Volume.pm and
        # Aggregate.pm
        my $self = shift;
        my $conn = make_conn_and_complain(shift);
        my $target = shift;
        my %opts = @_;

        my $target_type = $self->cmd;
        my $cmd = sprintf( "%s options %s ", $self->cmd(), $target);

        return eval {
                my $targetexp = quotemeta($target);

                if (%opts) {
                        #XXX for now, do not allow more than one option to be set at a time
                        if (scalar (keys(%opts)) > 1) {
                                die ("currently supporting setting of only one option at a time.\n"
                                     ."tried to set ".Dumper(\%opts));
                        }

                        # setting opts
                        for my $key (keys(%opts)) {
                                if ($key eq 'root' &&
                                    get_wafl_version($conn->hostp())
                                    >= WV_TSINGTAO_MIN) {
                                        $cmd .= $key ." ". '-f' ." ";
                                } elsif ($key eq 'nbu_archival_snap') {
                                        $cmd .= $key ." ". $opts{$key} ." ". '-f' ." ";
								} else {
                                        $cmd .= $key ." ". $opts{$key} ." ";
                                }
                        }

                        my $text = sayEOF_wrapper($conn, $cmd);
                        die "unable to set options: $text"
                                if $text =~ /^$target_type options:/;
                        sayEOF_wrapper($conn, "yes")
                                if $text =~ /Are you sure you want to (proceed|continue)/;

                        # no validation? return...
                        if (!$self->validation) {
                                return wantarray ? %opts : \%opts;
                        }

                        # start validation here
                        my $setopts = cmd_options($self,$conn,$target) or die $@;

                        for my $attr (keys(%{$setopts})) {
                                my $expected = $opts{$attr};
                                #XXX !!! if options are set correctly, but setting fails,
                                # the problem is probably that a numeric value gets changed
                                # in conditional below, mistakenly identified as 1|0, on/off switch
                                if ($opts{$attr} &&
                                    $opts{$attr} =~ /1|0|true|false|yes|no\b/i &&
                                    $attr !~ /resyncsnaptime|fractional_reserve|maxdirsize|raidsize|snaplock_(?:minimum|default|maximum)_period|acdirmax|acregmax|acsymmax|actimeo/)  {

                                        $expected = ($opts{$attr} =~ /1|true|yes/i) ? 'on' : 'off';
                                }

                                # when 'vol options <vol> root' is used
                                # look for 'diskroot' and not 'root'.
                                # test if the value is 1 or 0 only.
                                # this is due to Archana's change to
                                # make setting of root be an empty
                                # string.
                                if ($attr eq "root" &&
                                    defined($opts{root}) &&
                                    !$setopts->{diskroot}) {
                                        die "option root isn't set right!";

                                } elsif ($attr eq "guarantee" &&
                                         defined($opts{guarantee}) &&
                                         $setopts->{$attr} !~ /$expected/) {
                                        # this is a mess...
                                        # when you set volume guarantee,
                                        # the value returned can be either
                                        # 'volume' or 'volume(disabled)'
                                        die "option $attr isn't set right!";

                                } elsif ($attr ne "root" &&
                                         $attr ne "guarantee" &&
                                         defined ($opts{$attr}) &&
                                         $setopts->{$attr} ne
                                         $expected) {
                                        die "option $attr isn't set right!";
                                }
                        }
                        return wantarray ? %$setopts : $setopts;

                } else {
                        # getting opts
                        my $text = sayEOF_wrapper($conn, $cmd);
                         die "$text\n" if
                                 $text =~ /No .* named '$targetexp' exists|'$targetexp' is (an aggregate|a volume); use '(aggr|vol) .*'/;

                        die "$text\n" if
                                $text =~ /Invalid (volume|aggregate) name $targetexp/;

                         # prune text of new lines to make option parsing much
                         # easier -- (/s and /m confuses me)
                         $text =~ s/\r?\n/ /g;
                         my %options = map {
                                 /=/ ?
                                         split( /=/, $_ ) :  # key=value option
                                         ($_ => 1)           # boolean option
                         } split( /\s*,\s*/, $text ); # separate each option
                         return wantarray ? %options : \%options;

                        # XXX - commented out call into cmd_status for now..
                        # get both options and status
                        #$self->cmd_status($conn, $target) or die $@;
                        #my %options = $self->{status}->options()
                        #                or die $@;
                        #return wantarray ? %options : \%options;
                }
        };
}



sub cmd_blocks {

        my ($self, $conn, $target) = validate_pos(
                                                  @_,
                                                  {isa=>'Aggregate::ImplTemplate'},
                                                  {type=> SCALAR | HASHREF},
                                                  {type=> SCALAR}
                                                 );

        my $cmd = sprintf("%s status -b %s",
                          $self->cmd(),
                          $target);

        my $output = sayEOF_wrapper($conn, $cmd);
        my $rv;
        # parse the output here
        my @lines = split(/\n/, $output);
        foreach my $line (@lines) {
                if ($line =~ /^$target\s+(\d+)\s+(\d+)\s+(\d+)/) {
                        $rv->{block_size} = $1;
                        $rv->{blocks}     = $2;
                        $rv->{fs_size}    = $3;

                        return $rv;
                }
        }

        die "could not parse ouput: $output";
}


## This function is good to have in Mirror.pm
## rather than defining here, all functions related to mirror should be ideally dealt
## in Mirror.pm, since there is no Mirror.pm leaving it here
sub cmd_mirror
{
	my $self = shift;
	my $conn = make_conn_and_complain(shift);

	my $disklist;
	 my ( $target, %args ) = 
	 		validate_mixed(
				 @_,
				 [{type => SCALAR}], 
				 {
				 disks		=> {type => SCALAR|ARRAYREF, default => "", optional => 1},
				 force    	=> { type => BOOLEAN, default => 1, optional => 1 },
				 timeout 	=> { type => SCALAR, optional => 1,
						 	default => 7200,    # 2 hrs for slow ATA filers
						 	regexp => qr(^\d+$) # got to look like a number
			 			   },
				 victim_aggr 	=> { type => SCALAR, optional => 1, 
					 		regex => qr/^\w+$/ 
						   },


				 #Support for 64-bit aggr naveenm
				 indirect_type 	=> { type => SCALAR, optional => 1,
							 callbacks => {
							       'valid value' => sub {
									$_[0] eq '32' or $_[0] eq '64'
									}
								}
					 	   },
				 purge   	=> { type=> BOOLEAN, default => 0 }
			 	} );

	return eval 
	{

		$disklist = $args{disks};

		 # collect the options that needs to be passed to aggr create
		 my %opts;

		 # force
		 $opts{ '-f' } = ''
			 if( $args{force} ); 


		 # Specify the victim aggregate
		 $opts{ '-v' } = "$args{victim_aggr}"
			if( $args{victim_aggr} );


		 #64-bit support 
		 $opts{ '-B' } = $args{indirect_type}
			 if( $args{indirect_type} ); 


		 # XXX - tapl wanted randomization, so here's some randomization
		 # XXX - if this is used across the board, it may be a good idea
		 #       to factor it out to a module.
		 my @cmd_args;
		 if( $args{randomize_args} ) 
		 {
			 my @tmp_opts = keys %opts;
			 while( @tmp_opts ) 
			 {
				 my $opt = splice( @tmp_opts, rand( @tmp_opts ), 1 );
				 push @cmd_args, ( $opt, $opts{$opt} );
			 }
		 } 
		 else 
		 {
			 @cmd_args = %opts;
		 }


		 my $disklist_str;
		 if( ref($disklist) eq 'ARRAY' ) 
		 {
			if ( ref($disklist->[0]) eq 'ARRAY' ) 
			{
				$disklist_str = "";
				foreach my $elm ( @{$disklist} ) 
				{
					$disklist_str .= join(' ',
							('-d', @{$elm}, ' '));
				}
				$opts{ '-m' } = '' if ( @{$disklist} == 2 );
			} else {
				$disklist_str = join(' ', ('-d', @{$disklist}));
			}
		 } 
		 elsif($disklist ne "")
		 {
			 $disklist_str = $disklist;
		 }


		 #formulate the command
		 my $cmd = sprintf(
		     "%s mirror %s %s %s",
			 $self->cmd(),
		     $target,
			 join( ' ', @cmd_args ),
			 $disklist_str
		 );

		 my $conntype = $conn->conntype();

		 # ensure an rsh connection - to avoid low disks prompt:
		     # low spare disks warning from 'vol create':
		     #
		     # WARNING! Continuing with vol create will result in having
		     # no spare disk available for one or more raid groups.
		     # Are you sure you want to continue with vol create?
		 my $temp_conn = $conn;
		 if ( $conn->conntype !~ /^rsh/ ) 
		 {
		     Tharn::host($conn->hostp);
		     $temp_conn = Tharn::connect('#rsh');
		 }

		 logDbg("original conn type = " . $conn->conntype);
		 logDbg("temp conn type = " . $temp_conn->conntype);


		 my $text = sayEOF_wrapper( $temp_conn, $cmd, timeout=>$args{timeout} );

		 # it's a CONFIG error if there are not enough disks available
		 die config_result( $text ) 
		 	if ($text =~ /\d+ disks needed, but only \d+ matching disks? (is|are) available/ );

		 die $text 
		 	unless ( $text =~ /has\s+been\s+initiated/ ||
				    $text =~ /has\s+completed/ );


		 # flush the original connection's buffer
		 $conn->discardbuff();


		 if($self->validation()) 
		 {
			 my $sleeptime;
			 my $sleeptime_cap;
			 my $ready = 0;
			 my $endtime = time() + $args{timeout};
			 my $container;

			 if ($self->cmd() eq VOL_CMD) 
			 {
			     $container = "volume";
			     $cmd = "priv set -q diag; " .
				 "raid_config info listvol";
			 } 
			 else 
			 {
			     $container = "aggregate";
			     $cmd = "priv set -q diag; " .
				 "raid_config info listaggr";
			 }

			 # sleep for either 5 or 15 secs for sims or hardware
			 # filers respectively. it's not gonna finish
			 # so quickly. this is to avoid calling raid_config
			 # immediately (to reduce output) since we already know
			 # container creation will take quite a while.
			 #
			 # likewise, cap the sleep time at 2 or 5 mins for
			 # sim or hardware filer respectively
			 if ($conn->hostp()->hosttype() eq 'filer-simulator') 
			 {
				 # simumy $impl = Aggregate::ImplTemplate::Status->new(cmd => 'aggr');lators
				 $sleeptime = 5;
				 $sleeptime_cap = 2*60;
			 } 
			 else 
			 {
				 # hardware filer
				 $sleeptime = 15;
				 $sleeptime_cap = 5*60;
			 }

			 sleep $sleeptime;

			 while ( time() < $endtime ) 
			 {
				 # avoid calling RaidInfo. this method produces less
				 # output.
				 my $output = sayEOF_wrapper($conn, $cmd);
				 my @info = split(/\n/, $output);

				 foreach my $line (@info) 
				 {
					# see prod/common/raidv2/config.c starting @ line 803, function
					# volcmd_util_print_legacy_vol_info(legacy_volume_info_t *legacy_info); and
					# prod/common/wafl/typedefs.h, line 338, typedef enum { ... } wafl_Mount_state;
					my %aggrinfo = split( '[=,]', $line );	

					if( $aggrinfo{name} eq $target ) 
					{
					    # reset timer if still creating: typedef wafl_Mount_state: 
					    # WAFL_VOL_CREATING (prod/common/wafl/typedefs.h)
					    $endtime = time() + $args{timeout} 
					    	if( $aggrinfo{state} == 4 ); # reset timer if still creating

					    # return if aggregate is ready.
					    return 1 
					    	if( $aggrinfo{fsm_state} eq 'NORMAL' && 
							$aggrinfo{state} != 4 && 
							$aggrinfo{state_name} eq 'online' && 
							$aggrinfo{mirror_state_name} eq "mirrored" );
					} # if

				 } # foreach

				 # exponential backoff of increasing sleep time,
				 # capped at $sleeptime_cap depending on type
				 # of filer
				 $sleeptime *= 2;
				 if ( $sleeptime > $sleeptime_cap ) 
				 {
					 $sleeptime = $sleeptime_cap;
				 }

				 sleep($sleeptime);

			 } # while

			 die "Timed out waiting for $container '$target' " .
				 "to complete mirror";
		 } # if

		 return 1;

	}; # end of return eval

} # end of cmd_mirror

sub cmd_status
{


        my ($self, $conn, $target, %args)
                = validate_mixed(
                                 @_,
                                 [ {isa=> 'Aggregate::ImplTemplate'},
                                   {type=> SCALAR | HASHREF},
                                   {type=> SCALAR }
                                 ],
                                 { timeout=> {
                                              type => SCALAR,
                                              optional => 1,
                                              default => 300
                                             },
                                 }
                                );
        $conn =  make_conn_and_complain($conn);

        return eval {
                #XXX revisit in Status.pm later. Might not need this
                $self->{status}->cmd($self->cmd())
                        or die $@;

                $self->{status}->fetch_status($conn, $target,
                                              timeout=>$args{timeout})
                        or die $@;

                # XXX need to figure out what aggr status should return
                # now return same thing
                return wantarray ? ($self->{status}->state(),
                                    $self->{status}->status(),
                                    $self->{status}->parent(),
                                    $self->{status}->{options})
                        : $self->{status};
        };
}


sub cmd_wait_change{ die "TBD" }


sub cmd_exists
{
        my $self     = shift;
        my $conn     = make_conn_and_complain(shift);
        my ($target) = validate_pos(@_, {type => SCALAR});

        return eval {
                $self->cmd_options($conn, $target) or die $@;
                return 1;
        };
}



sub cmd_disks
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);
        my ($target) = validate_pos(@_, {type => SCALAR});
        return eval {
                my $cmd  = sprintf(
                        "%s status %s -r",
                        $self->cmd,
                        $target
                );
                my $text = sayEOF_wrapper($conn, $cmd);

                my @disk_list;


                foreach my $line (split(/\n/, $text)) {
                        if ($line =~ /$disk_status_re/) {

                                #my @da = split(/\s+/, $line);
                                my %disk;

                                $disk{'type'} = $1;
                                $disk{'name'} = $2;
                                $disk{'ha'} = $3;
                                $disk{'shelf'} = $4;
                                $disk{'bay'} = $5;
                                $disk{'chan'} = $6;
                                $disk{'used_mb'} = $10;
                                $disk{'user_blks'} = $11;
                                $disk{'phys_mb'} = $12;
                                $disk{'phys_blks'} = $13;
                                $disk{'size'} = "$10/$11";

                                push (@disk_list, {%disk});


                }
          }
                return @disk_list;
        }
}


sub cmd_root
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);

        return eval {
                my @list = $self->cmd_list($conn);
                foreach my $target (@list) {
                        # let cmd_options do the gory details (i.e. figure out
                        # what commands to use, etc
                        # NOTE: Tap says we left out the check for this
                        # cmd_options() so we wouldn't mask any missing root
                        # volume errors.
                        my $opts = $self->cmd_options($conn, $target);
                        if (defined($opts) && $opts->{root}) {
                                return $target;
                        }
                }

                die "No root found";
        };
}


sub cmd_scrub
{
        my $self = shift;

        my $conn = make_conn_and_complain(shift);
        my ($target, $mode) = validate_pos(@_,
                              {type => SCALAR}, {type => SCALAR});

        my $targetre = quotemeta($target);

        my %valid_modes = (start   => 1,
                           stop    => 1,
                           suspend => 1,
                           resume  => 1,
                           status  => 1);

        if (!$valid_modes{$mode}) {
                die "Unsupported " . $self->cmd . " scrub mode: $mode\n";
        }

        if ($self->validation) {
                $self->cmd_exists($conn, $target) or die $@;
        }


        my $cmd = sprintf("%s scrub %s %s",
                          $self->cmd,
                          $mode,
                          $target);

        return eval {

                # do the non-'... scrub status' modes first
                if ($mode ne 'status') {
                        my %mode2waitstr = (start   => 'started',
                                            stop    => 'stopped',
                                            suspend => 'suspended',
                                            resume  => 'resumed');

                        my $waitstr = $mode2waitstr{$mode};
                        my $text = sayEOF_wrapper($conn, $cmd);

                        my $regexp = sprintf("%s scrub: %s scrub of /?%s",
                                             $self->cmd,
                                             $waitstr,
                                             $target);
                        die "$text\n" unless
                          $text =~ m{$regexp};

                        return 1;

                } else {
                        # ... scrub status
                        my $text = sayEOF_wrapper($conn, "$cmd -v");

                        my $regexp = sprintf("%s scrub: status of /?%s",
                                             $self->cmd,
                                             $target);

                        die "$text\n" unless
                          $text =~ m{$regexp};

                        my @lines = split(/\n/, $text);

                        my($completed, $lastscrub);

                        # second line has scrub status
                        # third line has last scrub date
                        if($lines[1] =~
                           /Current scrub is (\d+)% complete|Scrub is not active/) {
                                $completed = $1 || '0E0'; # zero but true
                        } else {
                                die "Could not parse scrub status";
                        }

                        # XXX - I got lazy with the regexp.
                        # it could be improved ;)
                        if($lines[2] =~
                           /Last full scrub completed: ([\w\s\d:]+)|First full scrub not yet completed/) {
                                $lastscrub = $1 || '';
                        } else {
                                die "Could not parse scrub status";
                        }

                        return wantarray ?
                          ($completed, $lastscrub) : $completed;
                }
        };
}


sub cmd_snap_delta 
{
    my $self = shift;
    my $conn = make_conn_and_complain(shift);
    my($target, %args) = validate_mixed(
            @_,
            [ { type => SCALAR} ],
            {
                    beginning_snapshot   => { type => SCALAR, optional => 1 },
                    ending_snapshot      => { type => SCALAR, optional => 1 },
            }
        );

    if ((defined $args{ending_snapshot}) && 
        (!defined $args{beginning_snapshot})) {
        Carp::confess "Param 'beginning_snapshot' is required when " . 
                      "param 'ending_snapshot' is specified.";
    }

    my $targetre = quotemeta($target);
    my $snapcmd;

    if ($self->cmd() eq VOL_CMD) {
            $snapcmd = "snap delta -V $targetre";
    } else {
            $snapcmd = "snap delta -A $targetre";
    }

    if (defined $args{beginning_snapshot}) {
        $snapcmd .= " $args{beginning_snapshot}";
        if (defined $args{ending_snapshot}) {
            $snapcmd .= " $args{ending_snapshot}"
        }
    }

    return eval {
        my $text = sayEOF_wrapper($conn, $snapcmd);

        # Confess if command failed with invalid option or $target.
        if ($text =~ /snap delta:/ || $text =~ /^usage:/) {
             Carp::confess "Invalid options. Command '$snapcmd' returned: '$text'\n";
        }

        my @lines = split(/[\n\r]/, $text);
        chomp(@lines);

        my @hkeys = ();
        my @snap_changes = ();
        foreach my $line (@lines) {
            $line =~ s/^\s+//;

            if($line =~ /^From Snapshot/){
                @hkeys = split(/ \s+/, $line);
                next;
            }

            # workaround spaces in snapshot name 'Active File System'
            $line =~ s/Active File System/Active_File_System/;
            # workaround spaces in 'Time' field (example: 0d 03:43)
            $line =~ s/(\d+d)\s(\d\d:\d\d)/$1_$2/;

            if ($line =~ /^\S+\s+\S+\s+\d+/) {
                my $entry = {};
                my @values = split(/\s+/, $line);
                for (my $i = 0; $i < scalar(@hkeys); $i++) {
                    # restore spaces where required
                    $values[$i] =~ s/_/ /g if ($values[$i] eq 'Active_File_System');
                    $values[$i] =~ s/_/ /g if ($hkeys[$i] eq 'Time');

                    $entry->{$hkeys[$i]} = $values[$i];
                }
                push(@snap_changes, $entry);
            }
        }
        return \@snap_changes;
    };
}


sub cmd_snap_reserve
{
        my $self = shift;

        my $conn = make_conn_and_complain(shift);
        my ($target, %args) = validate_mixed(
                @_,
                [{type => SCALAR}],
                {
                reserve => { type => SCALAR, optional => 1 },
                }
            );

        my $targetre = quotemeta($target);
        my $snapcmd;

        if ($self->cmd() eq VOL_CMD) {
                $snapcmd = "snap reserve -V $targetre";
        } else {
                $snapcmd = "snap reserve -A $targetre";
        }

        return eval {
            if (defined($args{reserve})) {
                # Set the snap reserve
                my $text = sayEOF_wrapper($conn, "$snapcmd $args{reserve}");
                die "$text\n" if ($text);
                return $args{reserve};
            } else {
                    # Determine the snap reserve
                my $re = qr<$target: current snapshot reserve is (\d+)%>;

                my $text = sayEOF_wrapper($conn, $snapcmd);
                die "$text\n" unless ($text =~ m/$re/);

                return $1;
            }
        };
}

sub cmd_split
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);
          my($plex, $newtarget) = validate_pos(
                                   @_,
                                   {type => SCALAR},
                                   {type => SCALAR}
                                );

        return eval {
                my $cmd  = sprintf( "%s split %s %s",
                                    $self->cmd,
                                    $plex,
                                    $newtarget );

                my $targetre = quotemeta($newtarget);

                my $re = $self->cmd eq AGGR_CMD ?
                        qr<\s*new aggregate $targetre created\.\s*$> :
                        qr<\s*new volume $targetre created\.\s*$>;

                my $text = sayEOF_wrapper($conn, $cmd);
                die $text unless
                        $text =~ /$re/;

                # check if the new target is there
                if ($self->validation) {
                        $self->cmd_exists($conn, $newtarget) or die $@;
                }
                return 1;
        };
}

sub cmd_verify
{
        die "TBD";
}

sub cmd_spare_status
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);

        my $text   = sayEOF_wrapper($conn, $self->cmd . " status -s");
        my @lines  = split(/\n/, $text);

        my %pools;
        my $pool = 0;
        my $line;
        foreach my $i (0..$#lines) {
                $line = $lines[$i];
                if($line =~ /Pool(\d+) spare disks/) {
                        $pool = $1;
                } elsif($line =~ /$disk_status_re/) {
                        next if $1 ne 'spare';

                        my $ds = Aggregate::DiskStatus->new(
                                name => $2,
                                ha   => $3,
                                shelf => $4,
                                bay   => $5,
                                chan  => $6,
                                type => $8,
                                used_mb => $10,
                                used_blks => $11,
                                phys_mb   => $12,
                                phys_blks => $13
                        );
                        $ds->state(
                                $line =~ /zeroing/ ? RD_ZEROING:
                                $line =~ /not zeroed/ ? RD_NOTZEROED :
                                RD_ZEROED
                        );
                        $pools{$pool} ||= [];
                        push @{$pools{$pool}}, $ds;
                }
        }
        return wantarray ? %pools : \%pools;
}

sub cmd_raidgroup_status
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);

        my($target) = validate_pos( @_, { type => SCALAR });

        my $target_re = quotemeta($target);
        my $text = sayEOF_wrapper($conn, $self->cmd ." status -r $target");
        my @lines = split( /\n/, $text );

        my($plex_info, $rg_info);
        my %info = (
                status => undef,
                checksum_type => undef,
                plex  => []
        );
        for my $line (@lines) {
                if($line =~ /^Aggregate $target_re \(([^)]+)\) \(([^)]+)\)/) {
                        $info{status}        = $1;
                        $info{checksum_type} = $2;
                } elsif( $line =~ m{^\s+Plex ([/\w\d]+) \(([^)]+)\)} ) {
                        $plex_info = {
                                name   => $1,
                                status => $2,
                                raidgroups => []
                        };
                        push @{$info{plex}}, $plex_info;
                } elsif($line =~ m{^\s+RAID group ([/\w\d]+) \(([^)]+)\)} ) {
                        $rg_info = {
                                name   => $1,
                                status => $2,
                                disks  => []
                        };
                        push @{$plex_info->{raidgroups}}, $rg_info;
                } elsif( $line =~ /^$disk_status_re/ ) {
                        next if $1 eq 'spare';

                        my $ds = Aggregate::DiskStatus->new(
                                name => $2,
                                ha   => $3,
                                shelf => $4,
                                bay   => $5,
                                chan  => $6,
                                used_mb => $10,
                                used_blks => $11,
                                phys_mb   => $12,
                                phys_blks => $13
                        );
                        $ds->state(
                                $line =~ /zeroing/ ? RD_ZEROING :
                                $line =~ /not zeroed/ ? RD_NOTZEROED :
                                RD_ZEROED
                        );

                        push @{$rg_info->{disks}}, $ds;
                }
        }
        return wantarray ? %info : \%info;
}


sub cmd_show_space {
        my($self, $conn, $option, $target) = validate_pos( @_, 1,1,0,0);
	$conn = make_conn_and_complain ($conn);

	my $target_re = quotemeta($target);
      	my $cmd = $self->cmd ." show_space";
	$cmd .= " $option" if defined $option;
	$cmd .= " $target" if defined $target;

        my $text = sayEOF_wrapper($conn, $cmd);

        my @lines = split( /\n/, $text );

        my @aggrs;
	my ( $aggrs, $vols );
	my $volstart; #this is a mark used to parse the output of aggr show_space. 

        $aggrs = {
		  name             => undef,
		  totalSpaceValue  => undef,
		  waflReserveValue => undef,
		  snapReserveValue => undef,
		  usableSpaceValue => undef,
		  bsrNVLOGValue => undef,
		  asisValue => undef,
		  imageBackupValue => undef,
		  vols             => [],
		  totalSpace       => {
				       allocated => undef,
				       used      => undef,
				       avail     => undef
				       },
		  waflReserve      => {
				       allocated => undef,
				       used      => undef,
				       avail     => undef
				      },
		  snapReserve      => {
				       allocated => undef,
				       used      => undef,
				       avail     => undef
				      }
		 };

	$vols = {
		 name      => undef,
		 allocated => undef,
		 used      => undef,
		 guarantee => undef
		};

        for my $line ( @lines ) {

	  next if ( $line =~ /^$/ ); #skip the blank line

	  # if aggr show_space failed with invalid $option or $target,
          # the error messages from aggr show_space will be returned.
	       
	  if ( $line =~ /^aggr show_space:/ ) {
	       die "Invalid options or aggr name: @lines\n";
	  }

	  elsif ( $line =~ /^Aggregate \'(\S+)\'\s*$/ ) {
                        $aggrs->{name} = $1;
		      } 
	        elsif ( $line =~ /^\s+(\d+\w+)\s+(\d+\w+)\s+(\d+\w+)\s+(\d+\w+)\s+(\d+\w+)(\s+(\d+\w+))?(\s+(\d+\w+))?\s*/ ) {
		     $aggrs->{totalSpaceValue}  = $1;
             $aggrs->{waflReserveValue} = $2;      
		     $aggrs->{snapReserveValue} = $3;
		     $aggrs->{usableSpaceValue} = $4;
		     $aggrs->{bsrNVLOGValue} = $5;      
		     $aggrs->{asisValue} = $7;
		     $aggrs->{imageBackupValue} = $9;
		   }
	      elsif ( $line =~ /^Volume\s+Allocated/ ) {
		     $volstart = 1; #turn on $volstart means next lines are about volumes' information
		   }
	      elsif ( $line =~ /^(\S+)\s+(\d+\w+)\s+(\d+\w+)\s+(\w+)\s*$/ and $volstart ) {
		    $vols->{name} = $1;
		    $vols->{allocated} = $2;
		    $vols->{used} = $3;
		    $vols->{guarantee} = $4;
		    push @{$aggrs->{vols}}, $vols;
		    $vols = ();
		  }
	      elsif ( $line =~ /^Aggregate\s+Allocated\s+Used/ ) {
		    $volstart = 0; #turn off $volstart means volumes' information are completed
		  }
	      elsif ( $line =~ /^Total space\s+(\d+\w+)\s+(\d+\w+)\s+(\d+\w+)\s*$/ ) {
		    $aggrs->{totalSpace}->{allocated} = $1;
		    $aggrs->{totalSpace}->{used}      = $2;
		    $aggrs->{totalSpace}->{avail}     = $3;
		  }
	      elsif ( $line =~ /^Snap reserve\s+(\d+\w+)\s+(\d+\w+)\s+(\d+\w+)\s*$/ ) {
		    $aggrs->{snapReserve}->{allocated} = $1;
		    $aggrs->{snapReserve}->{used}      = $2;
		    $aggrs->{snapReserve}->{avail}     = $3;
		  }
	      elsif ( $line =~ /^WAFL reserve\s+(\d+\w+)\s+(\d+\w+)\s+(\d+\w+)\s*$/ ) {
		    $aggrs->{waflReserve}->{allocated} = $1;
		    $aggrs->{waflReserve}->{used}      = $2;
		    $aggrs->{waflReserve}->{avail}     = $3;
		    push @aggrs, $aggrs;
		    $aggrs = ();
		  }
	    }
      
	return wantarray ? @aggrs : \@aggrs;
}	
   

# These constant and mappings are used for the disk_free() function

use constant DF_SPACE   => 'space';
use constant DF_INODES  => 'inodes';
use constant DF_PRIVATE  => 'private inodes';
use constant DF_RESERVE => 'reserve';
use Utils::Units qw(:constants);

my %DF_REPORT_TYPE_OPTS = (
        UNIT_KB() => '-k',
        UNIT_MB() => '-m',
        UNIT_GB() => '-g',
        UNIT_TB() => '-t', # is this supported?
);
#  Note: 'UNIT_B' is available for (but only for) DF_INODES space type, handled in code.

my %DF_SPACE_TYPE_OPTS = (
        DF_SPACE()    => '',
        DF_INODES()   => '-i',
        DF_PRIVATE()  => '-i -P',
        DF_RESERVE()  => '-r'
);

sub _cmd_disk_free_no_poll
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);
	logDbg("Enter _cmd_disk_free_no_poll  Args:\n\n".Dumper({@_}));
        my %args = validate(@_, {
                type        => { type => SCALAR,                         optional=>1 },
                path        => { type => SCALAR,                         optional=>1 },
                vol         => { type => SCALAR,                         optional=>1 },
                aggr        => { type => SCALAR,                         optional=>1 },
                space_type  => { type => SCALAR,   default => DF_SPACE,  optional=>1 },
                report_type => { type => SCALAR,   default => UNIT_KB,   optional=>1 },
		snapshot    => { type => BOOLEAN,  default => 0,         optional=>1 },
		sync        => { type => BOOLEAN,  default => 1,         optional=>1 },
		sync_delay  => { type => SCALAR, default => $::DF_SYNC_DELAY,  optional=>1 },
        } );

        my ($path) = $args{vol} || $args{aggr};
	my $sync = $args{sync};
	my $sync_delay = $args{sync_delay};
	if ($sync && !$sync_delay) {
		$sync_delay = 1;
	}
        my ($dfcmd, $first_header);
        my $pathexp;
        my $cmd_path = $path;

	logDbg("Enter '_cmd_disk_free_no_poll'  Line: ".__LINE__."   Path: $path");
        if ($self->cmd() eq VOL_CMD) {
                if (get_wafl_version($conn->{hostp}) > WV_SCRIMSHAW_MAX) {
                        $dfcmd = "df -V";
                } else {
                        $dfcmd = "df";
                }
                $cmd_path = "/vol/$cmd_path" if ($cmd_path !~ m{^/vol/});
                logcomment("cmd_path = $cmd_path");

        } else {
                $dfcmd = "df -A";
        }

        $pathexp = $cmd_path;

        if ($args{snapshot}) {
                if ($cmd_path =~ m{/$}) {
                        $pathexp = $cmd_path.".snapshot";
                } else {
                        $pathexp = $cmd_path."/.snapshot";
                }
        }

        # If sync is requested, issue 'sync' command and sleep
        # for 30 seconds to force CP and then take df output.
        # - 30 seconds is a arbitary value. No reason, why 30.
        # Currently, there is no reliable way to get correct
        # data from df except this. NA_REVISIT.
	# Default for 'sync' is now 1. See BURT 306218
        if($sync || $sync_delay) {
	    logDbg("_cmd_disk_free_no_poll  Line: ".__LINE__.
	           "\n'sync-ing' before 'df' operation.  sync_delay: \"$sync_delay\" sec.");
	    sayEOF_wrapper($conn, "priv set -q diag; sync; priv set");
	    sleep($sync_delay);
	}

        my $space = $DF_SPACE_TYPE_OPTS{ $args{space_type} };
        if(!defined($space)) {
                die "Unrecognized space type: $args{space_type}\n";
         }

         my $report = $DF_REPORT_TYPE_OPTS{ $args{report_type} };

	 # 'UNIT_B' only availailable for 'DF_INODES' space type.
	 if ($args{space_type} eq DF_INODES  &&  $args{report_type} eq UNIT_B) {
	 	$report = '';
	 } 
         if(!defined($report)) {
                die "Unrecognized report type: $args{report_type}\n";
         }

        # Run the df command and snag the output

        # XXX - we sometimes seem to have a problem with sayEOF when
        # used from console. I don't know exactly what causes it.
        # for now, cross my fingers that nobody uses it from console
	my $text;
	if (my $server = are_we_ontapng($conn)) {
	    $text = sayEOF_wrapper($conn, "dbladecli -c \"$dfcmd $space $report $cmd_path\"");
	} else {
	    $text = sayEOF_wrapper($conn, "priv set -q advanced; $dfcmd $space $report $cmd_path");
	}

        if ($text =~ m%df:\s+(/vol/)?$path:\s(No such file or directory)%) {
                die("cmd_disk_free Error -- $2: $path");
        }
	if ($text =~ /invalid option/i) {
		die("cmd_disk_free Error -- $text");
	}

	logDbg("'_cmd_disk_free_no_poll'  Line: ".__LINE__."   Calling 'df_parser' on text:\n$text");
        return _df_parser($self, $conn, $pathexp, $args{space_type},
                          $args{report_type}, $text);
}


sub cmd_disk_free
{
        my $self = shift;
        my $conn = make_conn_and_complain(shift);
	logDbg("Enter cmd_disk_free  Args:\n\n".Dumper({@_}));
        my %args = validate(@_, {
                path        => { type=>SCALAR,					optional=>1 },
                type        => { type=>SCALAR,					optional=>1 },
                vol         => { type=>SCALAR,					optional=>1 },
                aggr        => { type=>SCALAR,					optional=>1 },
		snapshot    => { type=>BOOLEAN,		default=>0,		optional=>1 },
                space_type  => { type=>SCALAR,		default=>DF_SPACE,	optional=>1 },
                report_type => { type=>SCALAR,		default=>UNIT_KB,	optional=>1 },
		sync        => { type=>BOOLEAN,		default=>1,		optional=>1 },
		sync_delay  => { type=>SCALAR,		default=>$::DF_SYNC_DELAY,	optional=>1 },
                #  'poll' options
		max_time   => { type=>SCALAR,		default=>undef,		optional=>1 },
		no_poll    => { type=>BOOLEAN,		default=>0,		optional=>1 },
		wafl_cntr_thold    => { type=>SCALAR,	default=>undef,		optional=>1 },
        } );

	return eval {
		my $poll_time = 15;
		my $max_time = $args{max_time};
		if ($max_time) {
			logcomment("cmd_disk_free  'zombie blk. processing timeout'\n".
			            "('max_time' option)  is $max_time sec.");
		} else {
			logcomment("cmd_disk_free  'zombie blk. processing timeout'\n".
			            "('max_time' option)  is not set, no timeout will occure\n".
			            "during zombie block processing");
		}
	        my ($vol, $aggr, $cntnr_type, $cntnr_name);
		if ($args{vol} && $args{aggr}) {
			die("cmd_disk_free: Exactly one of 'vol' or 'aggr' options must be supplied");
		} elsif ( !$args{vol} && !$args{aggr} && !($args{path} && $args{type}) ) {
			die("cmd_disk_free: If neither 'vol' or 'aggr' is specified,\n".
			    "then both 'path' and 'type' (container type) must be given");
		} elsif ($args{vol} || ($args{type} =~ /vol/i)) {
			$vol = $args{vol} || $args{path};
			$args{vol} = $vol;
			$cntnr_type = 'volume';
			$cntnr_name = $vol;
		} elsif ($args{aggr} || ($args{type} =~ /aggr/i)) {
			$aggr = $args{aggr} || $args{path};
			$args{aggr} = $aggr;
			$cntnr_type = 'aggregate';
			$cntnr_name = $aggr;
		} elsif ($args{type} =~ /vol/i) {
			die("cmd_disk_free: Exactly one of 'vol' or 'aggr' options must be supplied");
		}
		delete($args{path});

		my $wafl_cntr_name = "wvblk_zombie_blks";
		my $wafl_cntr_spec = join( ":", ($cntnr_type, $cntnr_name, $wafl_cntr_name) );
		my $wafl_stats_cmd = "stats show $wafl_cntr_spec";
	        my %cmd_disk_free_args = %args;
	        my @poll_options = qw( max_time no_poll wafl_cntr_thold );
		my $no_poll = $args{no_poll};
	        for my $x ( @poll_options ) {
			delete($cmd_disk_free_args{$x});
		}
		while (my ($k,$v)=each(%args) ) {
			delete($cmd_disk_free_args{$k}) unless (defined($v));
		}
		if ($no_poll) {
			logDbg("Enter cmd_disk_free  Line: ".__LINE__."   ${cntnr_type}: \"$cntnr_name\"\n".
			       "'no_poll' option was given -- calling '_cmd_disk_free_no_poll' immediately");
			return( _cmd_disk_free_no_poll($self, $conn, %cmd_disk_free_args) );
		} else {
			logDbg("Enter cmd_disk_free  Line: ".__LINE__."   ${cntnr_type}: \"$cntnr_name\"\n".
			       "Begin 'polling' WAFL counter for 'df' availability");
		}
		my $wafl_cntr_thold = $args{wafl_cntr_thold};
		if (!defined($wafl_cntr_thold)) {
			$wafl_cntr_thold = 0;
		}
		my $start_time = time();
		my $end_time = ($max_time) ? $max_time + $start_time : 0;
		my $last_rv;
		my $match_count = 1;
		my $poll_count = 0;
		while (!$end_time || (time() < $end_time)) {
			my $okay_to_df = 0;
			my $num_zombie_blks = undef;
			my $elapsed_time = time() - $start_time;
			my $reply = sayEOF($conn, "priv set -q diag; " . $wafl_stats_cmd);
			logDbg("cmd_disk_free  Line: ".__LINE__."   WAFL 'show stats' reply:\n\"$reply\"");
			(undef, $num_zombie_blks) = ($reply =~ /^($wafl_cntr_spec):(\d+)/);
			if ($num_zombie_blks eq "" || !defined($num_zombie_blks)) {
				logDbg("cmd_disk_free  Line: ".__LINE__."   ${cntnr_type}: \"$cntnr_name\"\n".
				       "Cannot access WAFL counter -- calling '_cmd_disk_free_no_poll' immediately");
				return( _cmd_disk_free_no_poll($self, $conn, %cmd_disk_free_args) );
			}
			logDbg("cmd_disk_free  Line: ".__LINE__.
			       "   Poll iteration: $poll_count   Elapsed: $elapsed_time   ".
			       "Threshold: $wafl_cntr_thold   Zombie Blks: $num_zombie_blks");

if ($num_zombie_blks > 0) { logcomment("Positive 'num_zombie_blks': $num_zombie_blks"); }

			$okay_to_df = ($num_zombie_blks <= $wafl_cntr_thold);
			if ($okay_to_df) {
				logDbg("cmd_disk_free  Line: ".__LINE__.
				       "\n'df' values now available. Elapsed time: $elapsed_time sec.");
				return( _cmd_disk_free_no_poll($self, $conn, %cmd_disk_free_args) );
			}
			sleep($poll_time);
			$poll_count++;
		}
		die("cmd_disk_free: Reliable 'df' values not available within $max_time sec.");
	};
}


sub cmd_meta_i2p {
    my $self     = shift;
    my $conn     = shift;
    my ($target) = validate_pos(@_, { type => SCALAR });

    $conn = make_conn_and_complain($conn);

    return eval {
        my $cmd = sprintf("priv set -q test; %s meta_i2p %s",
                          $self->cmd(), $target);

        my $text = sayEOF_wrapper($conn, $cmd);
        my %content;

        if ($text =~ m/wafl fileid:\s+(\d+)/) {
            $content{$1} = 'wafl';
        }

        if ($text =~ m/raid fileid:\s+(\d+)/) {
            $content{$1} = 'raid';
        }

        while ($text =~ m/^(\d+)\s+(\S+)\s*$/mg) {
            $content{$1} = $2;
        }

        return \%content;
    };
}


sub _df_parser
{
        my ($self, $conn, $pathexp, $space_type, $report_type, $text) = @_;

        my $report_re;
        if ($space_type eq DF_RESERVE) {
                $report_re = '\(?' . '(\d+)' . $report_type .  '\)?';
        } else {
                $report_re = '(\d+)' . $report_type;
        }


        my $version;
        my $ver = get_wafl_version($conn->{hostp});

        if ($ver >= WV_FULLSAIL_MIN && $ver <= WV_FULLSAIL_MAX) {
                $version = '6.4';
        } elsif ($ver >= WV_SCRIMSHAW_MIN && $ver <= WV_SCRIMSHAW_MAX) {
                $version = '6.5';
        } elsif ($ver >= WV_ANCHORSTEAM_MIN && $ver <= WV_ANCHORSTEAM_MAX) {
                $version = '7.0';
        } elsif ($ver >= WV_TSINGTAO_MIN  && $ver <=  WV_TSINGTAO_MAX) {
                $version = '7.1';
        } elsif ($ver >=  WV_GORDONBIERSCH_MIN) {
                # XXX lets pretend we are in main and main is gordonbiersh
                $version = 'latest'; # XXX 7.2  ??
        } else {
                die "unknown version $ver";
        }

        # XXX/ram
        # In Scrimshaw (WV_SCRIMSHAW_MAX) and Anchorsteam,
        # 'df /vol/vol0'    => Filesystem kbytes used avail capacity
        # 'df -k /vol/vol0' => Filesystem total  used avail capacity
        # Looks like WV_SCRIMSHAW_MIN[+1] have different behavior.. sigh

        my $total = defined($report_type) ? "total" : "kbytes"
                if ($ver <= WV_SCRIMSHAW_MAX);

        my $filesystem;
        if ($self->cmd() eq VOL_CMD) {
                $filesystem = "Filesystem";
        } else {
                $filesystem = "Aggregate";
        }

        my %df_report = (
               #XXX assume that for our purposes df output in releases later than anchorsteam
                # will have the same format
             'latest' => {
                  DF_INODES => {
                    header => qr(^\s*$filesystem\s+iused\s+ifree\s+%iused),
                    match  => qr(^\s*$pathexp/?\s+(\d+)\s+(\d+)\s+(\d+)%),
                    },
                  DF_PRIVATE => {
                    header => qr(^\s*$filesystem\s+iused\s+ifree\s+%iused),
                    match  => qr(^\s*$pathexp/?\s+(\d+)\s+(\d+)\s+(\d+)%),
                   },
                  DF_RESERVE => {
                    header => qr(^\s*$filesystem\s+total\s+used\s+avail\s+reserved),
                    match  => qr(\s*$pathexp/?\s+$report_re\s+$report_re\s+$report_re\s+$report_re),
                   },
                  DEFAULT => {
                    header => qr(^\s*$filesystem\s+total\s+used\s+avail\s+capacity),
                    match  => qr(\s*$pathexp/?\s+$report_re\s+$report_re\s+$report_re\s+(\d+|\S+)),
                   },
                },
             '7.1' => {
                  DF_INODES => {
                    header => qr(^\s*$filesystem\s+iused\s+ifree\s+%iused),
                    match  => qr(^\s*$pathexp/?\s+(\d+)\s+(\d+)\s+(\d+)%),
                    },
                  DF_RESERVE => {
                    header => qr(^\s*$filesystem\s+total\s+used\s+avail\s+reserved),
                    match  => qr(^\s*$pathexp/?\s+$report_re\s+$report_re\s+$report_re\s+$report_re),
                   },
                  DEFAULT => {
                    header => qr(^\s*$filesystem\s+total\s+used\s+avail\s+capacity),
                    match  => qr(\s*$pathexp/?\s+$report_re\s+$report_re\s+$report_re\s+(\d+)),
                   },
                },
            '7.0' => {
                  DF_INODES => {
                    header => qr(^\s*$filesystem\s+iused\s+ifree\s+%iused),
                    match  => qr(^\s*$pathexp/?\s+(\d+)\s+(\d+)\s+(\d+)%),
                    },
                  DF_RESERVE => {
                    header => qr(^\s*$filesystem\s+total\s+used\s+avail\s+reserved),
                    match  => qr(^\s*$pathexp/?\s+$report_re\s+$report_re\s+$report_re\s+$report_re),
                   },
                  DEFAULT => {
                    header => qr(^\s*$filesystem\s+total\s+used\s+avail\s+capacity),
                    match  => qr(\s*$pathexp/?\s+$report_re\s+$report_re\s+$report_re\s+(\d+)),
                   },
                },

             '6.5' => {
                  DF_INODES => {
                    header => qr(^\s*$filesystem\s+iused\s+ifree\s+%iused),
                    match  => qr(^\s*$pathexp/?\s+(\d+)\s+(\d+)\s+(\d+)%),
                    },
                  DF_RESERVE => {
                    header => qr(^\s*$filesystem\s+$total\s+used\s+avail\s+reserved),
                    match  => qr(^\s*$pathexp/?\s+$report_re\s+$report_re\s+$report_re\s+$report_re),
                    },
                  DEFAULT => {
                    header => qr(^\s*$filesystem\s+$total\s+used\s+avail\s+capacity),
                    match  => qr(\s*$pathexp/?\s+$report_re\s+$report_re\s+$report_re\s+(\d+)),
                    },
                },

          '6.4' => {
                       DF_INODES => {
                       header => qr(^\s*$filesystem\s+iused\s+ifree\s+%iused),
                    },
                  DF_RESERVE => {
                    header => qr(^\s*$filesystem\s+$total\s+used\s+avail\s+reserved),
                    match  => qr(^\s*$pathexp/?\s+$report_re\s+$report_re\s+$report_re\s+$report_re),
                    },
                  DEFAULT => {
                    header => qr(^\s*$filesystem\s+$total\s+used\s+avail\s+capacity),
                    match  => qr(\s*$pathexp/?\s+$report_re\s+$report_re\s+$report_re\s+(\d+)),
                    },
                },

        );

        if ($space_type eq DF_INODES) {

                # XXX - with -i, the headers are always static (-k, -m, -g,
                # -t are ignored)

                # match header first, and then look for the entry
                if ($text !~ s|$df_report{$version}{DF_INODES}{header}||m) {
                        die "Error parsing df output (new format?)\n" .
                            "Expected $df_report{$version}{DF_INODES}{header}".
                            ", but got $text";
                }

                if ($text =~ m($df_report{$version}{DF_INODES}{match})m) {
                        return wantarray ?
                                ($1, $2, $3) :
                                { iused => $1, ifree => $2,
                                  percent_iused => $3 };
                }
                die "Could not find $pathexp in df output";

       } elsif ($space_type eq DF_PRIVATE) {
                # XXX - with -i -P, the headers are always static (-k, -m, -g,
                # -t are ignored)

                # match header first, and then look for the entry
                if ($text !~ s|$df_report{$version}{DF_PRIVATE}{header}||m) {
                        die "Error parsing df output (new format?)\n" .
                            "Expected $df_report{$version}{DF_PRIVATE}{header}".
                            ", but got $text";
                }

                if ($text =~ m($df_report{$version}{DF_PRIVATE}{match})m) {
                        return wantarray ?
                                ($1, $2, $3) :
                                { iused => $1, ifree => $2,
                                  percent_iused => $3 };
                }
                die "Could not find $pathexp in df output";        

        } elsif ($space_type eq DF_RESERVE) {

                # Modified RE to look for (nnnn) also, since
                # sometimes reserved space is displayed within brackets

                # match header first, and then look for the entry
                if ($text !~ s|$df_report{$version}{DF_RESERVE}{header}||m) {
                        die "Error parsing df output (new format?)\n" .
                            "Expected $df_report{$version}{DF_RESERVE}{header}".
                            ", but got $text";
                }

                if ($text =~ m($df_report{$version}{DF_RESERVE}{match})m) {
                        return wantarray ?
                                ($1, $2, $3, $4) :
                                { total => $1, used => $2,
                                  avail => $3, reserved => $4 };
                }
                die "Could not find $pathexp in df output";

        } else {
                # XXX - output differs between console and rsh?????
                # I've seen that the header output is different between
                # when I manually typ df "blah" and from a script via rsh.
                # I'm currently crossing my fingers that this doesn't cause
                # problems.

                # match header first, and then look for the entry
                logcomment($df_report{$version}{DEFAULT}{header});
                if ($text !~ s|$df_report{$version}{DEFAULT}{header}||m) {
                        die "Error parsing df output (new format?)\n" .
                            "Expected $df_report{$version}{DEFAULT}{header}".
                            ", but got $text";
                }

                if ($text =~ m($df_report{$version}{DEFAULT}{match})m) {
                        logcomment("Parsed df output ($1  $2  $3  $4)");
                        return wantarray ?
                                ($1, $2, $3, $4) :
                                { total => $1, used => $2,
                                  avail => $3, capacity => $4 };
                }
                die "Could not find $pathexp in df output ($1  $2  $3  $4)";
        }
}

package Aggregate::DiskStatus;
use strict;
use Class::MethodMaker
        new_with_init => 'new',
        get_set => [ qw(name ha shelf bay chan used_mb used_blks phys_mb phys_blks state type) ]
;

sub init
{
        my($self, %args) = @_;
        foreach my $field (qw(name ha shelf bay chan used_mb used_blks phys_mb phys_blks state type)) {
                if(exists $args{$field}) {
                        $self->$field($args{$field});
                }
        }
}


=head1 NAME

Aggregate::ImplTemplate - Delegator Object To Process Aggregate/Volume Operations

=head1 SYNOPSIS

  use Aggregate::ImplTemplate;
  # $conn = connect(...);

  my $impl = Aggregate::ImplTemplate->new(
    cmd        => 'aggr', # or 'vol'
    validation => 1,
  );

  $val = $impl->cmd_create(
    conn     => $conn,
    target   => $aggr,
    disklist => [ $disk1, $disk2 ... ]
  );

  $val = $impl->cmd_destroy(
    conn     => $conn,
    target   => $aggr,
    { destroy_corrupt_volumes => ... }
  );

  $val = $impl->cmd_destroy(
    conn     => $conn,
    target   => $aggr,
  );

  # all of these methods have the same signature as cmd_destroy()
  $val = $impl->cmd_purge(...);
  $val = $impl->cmd_exists(...);
  $obj = $impl->cmd_status(...);  # proxy to Aggregate::Status/Volume::Status
  $obj = $impl->cmd_options(...); # proxy to Aggregate::Status/Volume::Status
  @disks = $impl->cmd_disks(...);

  $val = $impl->cmd_restrict(
    conn   => $conn,
    target => $target,
    force  => $bool,
    cifsdelaytime => $time
  );

  $val = $impl->cmd_online(
    conn   => $conn,
    target => $target,
    force  => $bool
  );

  $val = $impl->cmd_offline(
    conn   => $conn,
    target => $target,
    force  => $bool
    cifsdelaytime => $time
  );

  $vname = $impl->cmd_root(conn => $conn);

  $val = $impl->cmd_rename(
    conn => $conn,
    from => $from,
    to   => $to
  );

  # below items are not yet done
  # $val = $impl->cmd_add() (?)
  # $val = $impl->cmd_wait_change($conn, $aggr, want => { status => qr(online) });
  # $val = $impl->cmd_scrub($conn, $aggr, $mode);
  # ($completed, $lastscrub) = cmd_scrub($conn, $aggr, 'status');
  # $hash = cmd_spare_status($conn);
  # $hash = cmd_raidgroup_status($conn, $aggr);

=head1 DESCRIPTION

This module defines routines related to ONTAP's aggregates and volumes.
The reason for the dual nature of this module is that there are a LOT of
code overlap between aggregates and volumes, and therefore it was prudent
to factor out the similarities in such a way that it was still possible
to keep the procedural interface for Aggregate.pm and Volume.pm all while
abstracting out the important bits into a pluggable implementation layer.

Volumes and Aggregates will internally instantiate one of these implementors
and simply proxy most of the calls directly:

  package Aggregate;
  use Aggregate::ImplTemplate;
  my $impl = Aggregate::ImplTemplate->new(cmd => 'aggr');

  sub aggr_create { $impl->cmd_create(@_) }

For those functions whose implementations differ significantly enough or
those that only make sense on either Volume.pm or Aggregate.pm, the
respective modules should write their own implementation.

=head1 METHODS

All functions, unless otherwise stated, return a true value on success,
and false on failure. On failure C<$@> is set to the appropriate error
value via C<die()>.

Function declarations below all take C<$conn>, which is a Connection object
(Connectrec) that is created by calling C<Tharn::connect()>, and a aggregate
name (C<$aggr>)

=head2 cmd_create(%args)

Creates a new aggregate/volume.

  $val = $impl->cmd_create(
    conn     => $conn,
    target   => "aggr1",
    disklist => 3,
    mirror   => $bool,       # optional
    force    => $bool,       # optional
    language => $language,   # optional
    rgsize   => $size,       # optional
    raidtype => $type,       # optional
    randomize_args => $bool  # optional
    snaplock => $bool        # optional
    snaplock_enterprise => $bool        # optional
    timeout  => $number      # optional
    purge    => $bool        # optional
  );

The C<$disklist> argument can be either a reference to an array, or a scalar.
If it's scalar, C<cmd_create()> will try to use it as the
E<lt>ndisksE<gt>[@E<lt>disk-sizeE<gt>] notation, while if it's an array ref,
the contents of the array will be treated as disk names and they will be
prepended with the '-d' option.

You may specify the various difference options in C<%opts>
(if any of the following is unspecified, the system default value will be
used):

=over 4

=item mirror

Enable/disable aggregate/volume mirroring

=item force

Force aggregate/volume creation

=item language

Specify the language on the aggregate/volume

=item rgsize

Specify the raidgroup size.

=item raidtype

Specify the raid type. The value must be either 'raid4' or 'raid_dp'

=item randomize_args

Randomizes the command line arguments (except for disk-list)

=item snaplock

Licenses snaplock and creates a SnapLock Compliant aggregate/volume

=item snaplock_enterprise

Licenses snaplock_enterprise and creates a SnapLock Compliant aggregate/volume

=item timeout

How long to wait for completion

=item purge

Whether to destroy any existing volume of the same name before creating the
new one - due to burt 117042 we need to destroy the aggregate with the given
name if it exists when creating volume

=back

=cut



=head2 cmd_destroy(%args)

Calls the appropriate command to destroy the specified target.

=cut



=head2 cmd_add($conn,$aggr,$disklist[,%opts])

Adds disks to a aggregate/volume.
For parameter $disklist, if the type is scalar, specify the number of disks
to be added. If the type is an array, cmd_add() will use the '-d' option
to add each disk via its disk name taken from each element of the array.

cmd_add() accepts the following option in %opt :

=over 4

=item force

Force disk addition

=item rg

Specify which raidgroup to add this disk to

=item timeout

Time to wait for validation that the disk has been added

=back

=cut




=head2 cmd_purge(%args)

A simple convenience function to set a aggregate/volume offline, and then
destroy it. For aggregates, it also searches all volumes in that aggregate
and calls vol_purge on each volume before purging the aggregate.

=cut



=head2 cmd_online(%args)

  $val = $impl->cmd_online(
    conn   => $conn,
    target => $target,
    force  => $bool
  );

Change the state of the aggregate/volume to "online".
Returns success (if the aggregate/volume was changed or is already in the
desired state), or failure otherwise (state did not change or the target did
not exist).

=over 4

=item force

Force onlining of the target.

=back

=cut




=head2 cmd_offline(%args)

  $val = $impl->cmd_offline(
    conn   => $conn,
    target => $target,
    force  => $bool
    all    => $bool   # only applies to aggregates
  );

Change the state of the aggregate/volume to "offline".
Returns success (if the aggregate/volume was changed or is already in the
desired state), or failure otherwise (state did not change or the target did
not exist).

=over 4

=item force

Force offlining of the target. This will cause the command to run with
'priv set -q diag' since it will only work in that mode.

=item cifsdelaytime

The number of seconds to delay before taking the aggregate offline, during
which time CIFS uses are warned of the pending loss of service.

=back

=cut



=head2 cmd_restrict(%args)

  $val = $impl->cmd_restrict(
    conn   => $conn,
    target => $target,
    force  => $bool,
    cifsdelaytime => $time
  );

Change the state of the aggregate/volume to "restricted".
Returns success (if the target was changed or is already in the desired
state), or failure otherwise (if the aggregate couldn't change or didn't
exist).

cmd_restrict() accepts the following options :

=over 4

=item force

Force restricting of target Will run with 'priv set -q diag' since it
will only work in that mode.

=item cifsdelaytime

The number of seconds to delay before making the target restricted, during
which time CIFS uses are warned of the pending loss of service.

=back

=cut



=head2 cmd_rename(%args)

  $val = $impl->cmd_rename(
    conn => $conn,
    from => $from,
    to   => $to
  );

Renames the name of aggregate or volume $from to $to.
Returns success or failure.

=cut


=head2 cmd_read_fsid($conn,$target)

  $fsid = $impl->cmd_readfsid($con, $target);

Returns the FSID of the container.

=cut

=head2 cmd_rewrite_fsid($conn,$target,%args)

  $impl->cmd_rewrite_fsid($conn, $target, fsid => $target_fsid);

Rewrites the FSID of the desired container. Returns the *original* FSID of
the container prior to the FSID rewrite.

Note: You need to restrict/offline the container first prior to calling
      this API.

=cut


=head2 cmd_list($conn)

Returns the list of known aggregate or volume names, listed from
the appropriate status command

  my @vols = cmd_list(conn => $conn);

=cut


=head2 cmd_options(%args)

This is a direct proxy to Aggregate::Status/Volume::Status' fetch_options()
method. See the documentation in those libraries to find out the exact behavior

=cut


=head2 cmd_status(%args)

This is a direct proxy to Aggregate::Status/Volume::Status' fetch_status()
method. See the documentation in those libraries to find out the exact behavior

=cut


# XXX - wait for state/status change

=head2 cmd_wait_change($conn,$aggr[,%opts])

C<cmd_wait_change()> will wait for asynchronous processing on a aggregate that
may change its state/status. For example, to wait for a aggregate to
come up online, you would do:

  use Aggregate qw(cmd_wait_change);
  ...
  my $conn = connect(...);
  my $val = cmd_wait_change($conn, 'aggr0', state => qr(^online$) );
  my $val = cmd_wait_change($conn, 'aggr0', status => qr(^normal|raid4$) );

The C<state> or C<status> parameter specifies which field(s) to match.
The function will look at the state and status of an aggregate as return by
'aggr status -r <aggr>' to match against the required state/status.

So in order to match the state as either online or offline, you can do:

  cmd_wait_change($conn, 'aggr0', state => qr{^(?:on|off)line$} );

If you specify multiple keys in either the C<state> or C<status> hash, all
of those criteria must be met before C<cmd_wait_change()> can successfully
return.

You can also specify a C<timeout> and a C<delay> parameter, which specify
the total time to try getting the desired status, and the time to sleep
in between match attempts, respectively.

=cut


=head2 cmd_exists(%args)

  $val = $impl->cmd_exists(
    conn   => $conn,
    target => $target
  );

Returns true if a given aggregate or volume name exists.
(CAVEAT: This function is currently implemented to use cmd_options())

=cut


=head2 cmd_disks(%args)

  $val = $impl->cmd_disks(
    conn   => $conn,
    target => $target
  );

Returns an array of disk descriptors for all disks on the aggregate.  Each
disk descriptor is a hash of the form (type name size)

=cut


=head2 cmd_root(%args)

  $vname = $impl->cmd_root(
     conn => $conn
  );

Returns the name of the root aggregate/volume on the given filer.

=cut


=head2 cmd_scrub($conn,$aggr,$mode)

Performs the aggr scrub operation, specified by C<$mode>, which can
be one of the following:

  start
  stop
  resume
  suspend
  status

Returns 1 if the operation is successful, *except* for C<status>, which
returns the percentage completed and the last scrub time.

   my($completed, $lastscrub) = cmd_scrub($conn, $aggr, 'status');

C<completed> is set to 0E0 ("0 but true") for both when the percentage
is at 0 or there is no scrubbing being performed on the aggregate.
This is to allow the function to return a "0" while allowing a
construct like:

   my($completed, $lastscrub) = cmd_scrub($conn, $aggr, 'status') or die;

=cut


=head2 cmd_split($conn,$aggr,$newaggr)

Split a aggregate

=cut


=head2 cmd_spare_status($conn)

Given a connection to a filer, fetches the output of 'aggr status -s' and
returns a hash containing lists of Aggregate::DiskStatus objects.

The list of disks are keyed by the "pool" that they belong in. If there
are no pools defined, everything will go into slot "0"

  my $conn   = connect(...);
  my $hash = cmd_spare_status($conn);
  foreach my $disk (@{$hash->{0}}) {
    print $disk->name, "\n";
  }


In scalar context, the reference to the resulting hash is returned.
In array context, the hash values are returned directly.

  my $hash = cmd_spare_status($conn);
  my %hash = cmd_spare_status($conn);

=cut

=head2 cmd_raidgroup_status($conn,$aggr)

Given a connection to a filer and a target aggregate, parses the output of
C<aggr status -r $aggr> and returns a hash.

  NOTE: If this function is going to be used in more than one script,
  the returned structure needs to be changed a bit more...

This has is a 4-level nested hash. The top level contains some information
about the aggregate. The second level can be accessed via $hash->{plex},
which is an array of plexes. The third level has the raidgroup information,
and can be accessed via $plex->{raidgroups}. The last level a list of
Aggregate::DiskStatus objects, and can be accessed via $raidgroup->{disks}

  need to work on this doc...

=cut

=head2 cmd_show_space($conn, $option, $aggr)
Given a connection to a filer and a target aggregate, parses the output of
C<aggr show_space [$option] [$aggr]> and returns a refence of a array.

This is a 3-level nested array/hash.

The top level array contains information about the aggregate/aggregates. 

The second level stores information about space usage of aggregates and they can be accessed via $aggrs[]->{totalSpaceValue} or $aggrs[]->{waflReserveValue} or $aggrs[]->{snapReserveValue} or $aggrs[]->{usableSpaceValue} etc. 

The third level has the volumes within the aggregate information we well as break down allocation of aggregates' total space, Snap reserve and WAFL reserve space. Those information can be accessd via $aggrs[]->{totalSpace}->{allocated} or $aggrs[]->{totalSpace}->{used} or $aggr[]->{totalSpace}->{avail}, same for $aggrs[]->{waflReserve} and $aggrs[]->{snapReserve} values, etc. And for volumes space usage information within a aggregare: $aggrs[]->$vols[]->{allocated} or $aggrs[]->vols[]->{used} or $aggrs[]->vols[]->{guarantee}.

If cmd_show_space failed due to invalid $option or $aggr name, exception will be raised with error messages from cmd_show_space.

  need to work on this doc...

=cut

=head1 Aggregate::DiskStatus

Aggregate::DiskStatus is a data-storage only object. It has no methods,
and contains the following fields:

=head2 state

=head2 name

=head2 ha

=head2 shelf

=head2 bay

=head2 chan

=head2 used_mb

=head2 used_blks

=head2 phys_mb

=head2 phys_blks

=cut

=head1 SEE ALSO

L<Aggregate::Status>
L<Volume::Status>
L<Params::Validate>
L<Utils::Units>

=cut

=head1 TODO

- tapl: make timeout values consistent to all APIs.
- Fix parsing of list of volumes & others in cmd_status()
- Determine if all the RAID centric APIs, eg. cmd_raidgroup_status(), should
  stay here and not Volume.pm, or be consolidated into this library.
- Add cmd_*() APIs if there're not in here

TBD after code review

=cut


1;


