# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Aggregate/ImplTemplate/Status.pm#1 $
## @summary Implementation Of Status/Options Querying For Aggregate/Volume
## @author dl-johny@netapp.com (FSS WAFL Quality)
## @cf     false
## @status public
## @pod here

package Aggregate::ImplTemplate::Status;
use strict;
use Params::Validate qw(validate SCALAR OBJECT);
use Class::MethodMaker
	new_with_init => 'new',
	new_hash_init => 'hash_init',
	get_set       => [ qw(cmd validation name status state type) ],	#XXX maybe need validation later
	hash          => [ qw(options) ],
;
use Utils qw(make_conn_and_complain);
use SayWrap qw(sayEOF);
use Tharn;
use Data::Dumper;
use Filer::Version qw(get_wafl_version WV_SCRIMSHAW_MAX);

use constant VOL_CMD => 'vol';
use constant AGGR_CMD => 'aggr';

my %BasicValidate = (
	conn => {
		type => SCALAR | OBJECT,
		callbacks => {
			'is valid object' => sub {
				ref($_[0]) ? (
				UNIVERSAL::isa($_[0], 'Hostrec') ||
				UNIVERSAL::isa($_[0], 'Connectrec')) : 1;
			}
		}
	},
	target => {
		type => SCALAR
	},
	timeout => { type => SCALAR, optional => 1, default => 300 }
);

sub init
{
	my $self = shift;

	$self->hash_init(@_);

	if ($self->cmd eq 'vol') {
		$self->cmd('vol');
	} elsif ($self->cmd eq 'aggr') {
		$self->cmd('aggr');
	} else {
		Carp::croak("No such type supported for " .
			__PACKAGE__ . ": " . $self->cmd);
	}
}

sub fetch_status
{
	my $self = shift;
 	my $conn = make_conn_and_complain(shift);
	my $target = shift;
	my %args = @_;
	return eval {
		if (defined($args{timeout}) && $args{timeout} < -1) {
			$args{timeout} = -1;
		}

		my $text = $self->_fetch_raw_status_str($conn, $target, %args)
			or die $@;
		my %info = $self->_parse_raw_status_str($conn, $target, %args,
							text => $text)
			or die $@; # paranoid in case someone adds eval to it
		my %options = $self->fetch_options($conn, $target, %args);
		# "vol options" can legitimately return nothing (when a volume
		# is offline), so it's valid for %options to be empty
		$info{ options } = \%options; 
		
		return $self->_create_structure(%info);
	};
}




sub fetch_options
{
	my $self = shift;
	my $conn = make_conn_and_complain(shift);
	my $target = shift;
	my %args = @_;

	return eval {
		my $target_re = quotemeta($target);

		my $cmd  = sprintf( "%s options %s", $self->cmd(), $target);
		my $text = sayEOF($conn, $cmd);
		
		# note that return string is different from 'status' command
		if ($text =~ /No (aggregate or traditional )?volume named '$target_re' exists/) {
			die $text;
		}

		# prune text of new lines to make option parsing much
		# easier -- (/s and /m confuses me)
		$text =~ s/\r?\n/ /g;

		my %options = map {
			/=/ ?
				split( /=/, $_ ) :  # key=value option
				($_ => 1)           # boolean option
		} split( /\s*,\s*/, $text ); # separate each option

		#update our data structure
		$self->{options} = \%options;
		return wantarray ? %options : \%options;
	};
}

sub _fetch_raw_status_str
{
	# don't do validation in this level -- it's a waste
	my $self = shift;
	my $conn = make_conn_and_complain(shift);
	my $target = shift;
	my %args = @_;

	my $target_re = quotemeta($target);
	my $cmd   = sprintf( "%s status %s", $self->cmd(), $target);
	my $text  = sayEOF($conn, $cmd, TIMEOUT => $args{timeout});

	if ($text =~ /No (?:volume|aggregate) named '$target_re' exists/) {
		die $text;
	}
	return $text;
}

sub _parse_raw_status_str
{
	# don't do validation in this level -- it's a waste
	my $self = shift;
	my $conn = shift;
	my $target = shift;
	my %args = @_;

	my $target_re = quotemeta($target);
	my $text      = $args{text};

	# XXX/daisuke - originally this function had a far more stringent
	# error check, but relaxing it a bit here... we check for the
	# header (because it's important :) but otherwise we just
	# keep on chugging to see if we get all the necessary information.
	# If we can't, oh well. Then we complain. Furthermore, we will no
	# longer parse options from "* status" commands.

	my %info = (name => $target);

	# first, and foremost, deal with 'abnormal statuses' s.a. 'temprorarily busy'
	my $busy = 'temporarily busy';
       	if ($text =~ /$busy/) {
		$info{state} = 'busy';
		# not much we can do now
		return %info;
	}

	my $first_line_re = $self->cmd eq 'aggr' ?
		qr<^\s*Aggr\s+State\s+Status\s+Options\b> :
		qr<^\s*Volume\s+State\s+Status\s+Options\b>;

	if ($text !~ /$first_line_re/sm) {
		die sprintf(
			"Unrecognized format for %s status : '%s'",
			$self->cmd, $text );
	}

	# this bit is the same for both aggregates and volumes.
	# (notice we ignore options here.
	if ($text !~ /^\s*$target_re\s+(\w+)\s+((\w,?\s?)+)\s+(\w+)/m) {
		die "Unable to parse status: $text";
	}

	$info{state} = $1;

	# here is where strangeness can happen. status can wrap
	# around, for example our output can look like this:
	# Volume State      Status            Options
	# snapflex online     raid_dp, flex     snapmirrored=on,
	#                  snapmirrored      fs_size_fixed=on,
        #                                    guarantee=volume(disabled)
	# 
	#
	# ..so we need to grab whatever is in that second column, but
	# ignoring the options that can also span multiple lines.
	# use a little parsing helper for that
	$info{status} = $self->_status_column_helper_parser($target, $text)
		or die $@;

	# parse out the type
	if ($info{status} =~ /aggr\b/) {
		$info{type} = 'aggr';
	} elsif ($info{status} =~ /flex\b/) {
		$info{type} = 'flex';
	} elsif ($info{status} =~ /trad\b/) {
                $info{type} = 'trad';
	} else {		
		if (get_wafl_version($conn->{hostp}) <= WV_SCRIMSHAW_MAX) {
			# For SS and previous releases,
			# guess that the status is ok and assume trad vol
                	$info{type} = 'trad';
		} else {
			die "Unknown container type:" . $info{status};
		}
	}

	# XXX/daisuke - I thought about doing this via inheritance, but
	# something seemed wrong there. gonna live with this big if/else
	# for now
	if ($self->cmd eq 'vol') {
		if ($info{type} eq 'flex') {

			if ($text !~ /^\s*Containing aggregate: '([^']+)'/m) {
				die "Unable to parse status (containing aggregate): $text";
			}
			$info{parent} = $1;
		} elsif ($info{type} eq 'trad') {
			$info{parent} = undef;
			$self->_parse_raid_plex(%args);
		}
	} else {
		if ($info{type} eq 'aggr') {
			# check if the following line is there. if not, bail
			# out saying 'aggr status' cmd output has changed.
	                if ($text !~ /^\s*Volumes:\s+([\w_,<>\(\) ]+\S*)/m) {
                              die "Unable to parse status (volumes): $text";
			}

			my @vollist;

			# grab everything between the 'Volumes:' and 'Plex...'
			# lines
			if ($text =~ /Volumes:(.*?)Plex \//ms) {
				my $list = $1;

				# strip white spaces, newlines, and carriage
				# returns
				$list =~ s/\s*//g;

				foreach (split(/\s*,\s*/, $list)) {
					# ignore '<none>'. if there are
					# no volumes in the aggregate,
					# return an empty array
					if ($_ =~ /<none>/) {
						next;
					}
					push(@vollist, $_);
				}
			}

			# dump the entire list back to the obj's data structure
			$info{volumes} = \@vollist;
		 }
		$self->_parse_raid_plex(%args);
	}
	return %info;
}

sub _parse_raid_plex{}

# this is internal function, so whoever is calling it,
# should know what they are doing
sub _create_structure
{
	my $self = shift;
	my %info = @_;
	foreach my $key (keys(%info)) {
		$self->{$key} = $info{$key};
	}
	return $self;
}


# applicable for volume only
sub parent($;$) {
	my $self   = shift;
	my $parent = shift;
	if ($self->cmd() ne VOL_CMD) {
		return ($self->{name});
	}
	if (defined($parent)) {
	 	$self->{parent} = $parent;	
	}
	return $self->{parent};
}

# applicable for aggregate
# position is 1-based index
sub volumes
{
	my $self= shift;
	my $position = shift;

	return eval {
		if ($self->cmd() ne AGGR_CMD) {
			die "could only call volumes() on aggregate object";
		}
		if (defined($position) && $position <= 0) {
			die "incorrect index $position";
		}
		if (defined($position)) {
			return $self->{volumes}->[$position-1];
		}
		return $self->{volumes};
				
	};
}


sub _status_column_helper_parser 
{
	my $self = shift;
	my $target = shift;
	$target = quotemeta($target);
	my $text = shift;
	
	my $first_line = $self->cmd eq 'aggr' ?
		qr<^\s*Aggr\s+State\s+Status\s+Options\b> :
			qr<^\s*Volume\s+State\s+Status\s+Options\b>;	
	
	
	my $last_line = $self->cmd eq 'aggr' ?
		'Volumes:': $text =~ /flex/ ? 'Containing aggregate' : 'Plex';
	
	$text =~ /$first_line\n(.*)$last_line/sm;
	my $interesting_text = $1;
	if (!$interesting_text) {
		die "changed status string? could not parse $text";
	}

	
	# now grab that first line of the interesting text 
	# and fish out status
	my @lines = split(/\n/, $interesting_text);

	 if (shift(@lines) !~ 
		/^\s*$target\s+(\w+)\s+((\w,?\s?)+)\s+([\s\w,=]+)/m) {
			die "Unable to parse status: $text";
	}
	my $rv = $2;
	
	if (substr($rv, -1,1) eq ' ') {
		chop($rv);
	}
	
	my $status_column;
	# now process remaining lines
	# we will append a comma after each status line
	foreach my $line (@lines) {
		# if we reached a blank line, we are done
		if ($line !~ /\S/ ) {
			last;
		}
		
		my $status_column = (split(/\s\s+/, $line))[1];
		
		# the options can wrap around, and if they do, while
		# status column termninated, we will end up with 
		# options, in $status_column
		# fortunately, options contain '=', so we can spot them that way
		if ($status_column =~ '=') {
			# actualy finished parsing status.
			last;
		}
		$rv .=', ';
		$rv .= $status_column;
	}
	return $rv;
};



1;

__END__

=head1 NAME

Aggregate::ImplTemplate::Status - Underlying Implementation for Aggregate/Volume Status and Option Querying

=head1 SYNOPSIS

  # Don't use this anywhere other than Aggregate::Status and Volume::Status
  use Aggregate::ImplTemplate::Status;
  my $impl = Aggregate::ImplTemplate::Status->new(cmd => 'aggr');
  sub fetch_status {
    ...
    $impl->fetch_status(@_);
    ...
  }

=head1 DESCRIPTION

This is an object to decouple the underlying implementation of Aggregate
and Volume command(s), while maximizing code sharing for both.

There's not much here that a regular user needs to know. Read the docs
for Aggregate::Status and Volume::Status instead.

=cut
