=head1 package Allflavors

OO package to manage a data structure that keeps track of which hardware flavors
have run, passed, and/or failed on a specific build.  

<after a few runs, paste the data structure here to make it easier to understand>

=over 8

=cut

package Allflavors;
use strict;
use Storable; 
use Data::Dumper;

=item new(statdir=>'/x/eng/web-spin/lab')

Creates the object.

=cut

sub new {
	my $class = shift;
	my $self  = {
		analysed=>0,
		passed=>0,
		failed=>0,
		notrun=>0,
		statdir=>"/u/mgarrett/statdir", #this value gets overwritten.  This is a test only value.
		flavors=>[qw/x86 x86_64 x86.debug x86_64.debug x86_64.sim x86_64.debug.sim/],
		@_,
	};

	#some private variables
	$self->{logfile}="$self->{statdir}/lastpass/lastpass.log";
	$self->{buildinfodir}="$self->{statdir}/lastpass/buildinfo";
	$self->{filename}="$self->{buildinfodir}/$self->{changenum}";

	if (!-d $self->{buildinfodir}) { #if the dir doesn't exist
		my $cmd="mkdir -p $self->{buildinfodir}"; #create it
		print "$cmd\n";
		system $cmd;
	}

	#Define, then read in the build information if the stored file already exists.
	$self->{buildinfo}=undef;
	if (-e $self->{filename} ) { $self->{buildinfo}=Storable::lock_retrieve $self->{filename}; }

	bless($self,$class);
	return $self;
}

#update the calculated internal variables
sub _analyse {
	my $self = shift;
	$self->{passed}=0;
	$self->{failed}=0;
	$self->{notrun}=0;
	for my $oneflavor (@{$self->{flavors}}) {
		if (exists $self->{buildinfo}{$oneflavor}) {
			if (exists $self->{buildinfo}{$oneflavor}{passed}) { #if there is at least one pass logged
				$self->{passed}++;
			} else {
				$self->{failed}++;
			}
		} else {
			$self->{notrun}++;
		}
	}
	$self->{analysed}=1;
}

=item all_have_run()

Return shows if all hw flavors have run at least one test?

=cut

sub all_have_run {
	my $self = shift;
	unless ($self->{analysed}) { $self->_analyse }

	if ($self->{notrun}==0) {
		return 1;
	} else {
		return 0;
	}
}

=item status()

Return value shows how many hardware flavors passed/failed/notrun on this build

=cut

sub status {
	my $self = shift;
	unless ($self->{analysed}) { $self->_analyse }

	return "[passed,failed,notrun]=[$self->{passed},$self->{failed},$self->{notrun}]";
}

=item add_result(str_result=>$Test_result, runid=>$ENV{JOBID}, flavor=>$Flavor);

Adds a result to the database for this flavor.

=cut

sub add_result {
	my $self = shift;
	my %parms = (@_);

	if ($parms{str_result} eq "PASS") {
		push(@{$self->{buildinfo}{$parms{flavor}}{passed}},$parms{runid});
	} else {
		push(@{$self->{buildinfo}{$parms{flavor}}{failed}},$parms{runid});
	}

	$self->_analyse;
	$self->_save2disk;
}

#write the info we have collected to disk.
sub _save2disk {
	my $self = shift;
	Storable::lock_nstore $self->{buildinfo}, $self->{filename}; 
}

=item allpass()

Have all flavors passed at least once for this build?

=cut

sub allpass {
	my $self = shift;
	unless ($self->{analysed}) { $self->_analyse }
	if ($self->{passed}==scalar(@{$self->{flavors}})) { #did the number of tests that passed, equal the number of flavors we check?
		return 1;
	} else {
		return 0;
	}
}
	
=item dump()

Dump the contents of $self.  (mostly for debug)

=cut

sub dump {
	my $self = shift;
	print "Contents of object Allflavors: \n";
	print Dumper $self;
}


#returns a formatted date that is used in the logfile.
sub _my_date {
	my ($day, $month, $year) = (localtime)[3,4,5]; #localtime returns a whole bunch of paramaters (like day of week) but we just want element 3,4,5
	$year=$year+1900; #Since the year is 1900 based.
	$month++; #Since the month is 0 based
	return sprintf "$month\-$day\-$year";
}


1;
