#
# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a Armadillo Bridge
## @author kaaron@netapp.com
## @status private
## @pod here

package Armadillo;

# Compiler directives.
use strict;

# Module import.
use Time::HiRes qw(gettimeofday);
use Tharn;
use San::SanLog;
use Utils qw(make_host make_conn);
use Error;
use Data::Dumper;
use San::SanClient::ClientUtil;
use San::SanSwitch::SwitchUtil;
use vars qw($_CONNECTION_COUNT
            @EXPORT
            @ISA);

@ISA = qw(Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Create new Armadillo object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($proto, %args) = @_;

    my ($bridge, $connection, $conn_type, $log, $passwd, $timeout);
    my ($connection_retries, $user, $debug_level);
    my $self = {};
    my $sanlun_obj = {};

    my $class = ref($proto) || $proto;
    bless ($self, $class);

    # init variables
    $debug_level = 0;
    $timeout = 1800;
    $connection_retries = 5;

    # Process arguments.  If an error occurs during the arg processing,
    # The error flag and errmsg values are set and the object is returned.
    # It is up to the caller to check for errors after the new call.
    foreach (keys %args) {
        if (/^-?bridge$/) {
            length($bridge = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"host\":",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/^-?conn_type$/) {
            length($conn_type = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"conn_type\":",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/-?timeout$/) {
            length($timeout = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"timeout\":",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/-?conn_retries$/) {
            length($connection_retries = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"conn_retries\":",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/-?debug_level$/) {
            length($debug_level = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"debug_level\":",
                                    "$args{$_}");
                    return $self;
                };
        }
    }

    # Return unless bridge name is specified
    unless (defined $bridge and length $bridge) {
        $self->_set_err("Mandatory parameter \"bridge\"",
                        "not specified");
        return $self;
    }

    # create a host object for $host
    my $hostrec = Utils::make_host($bridge);

    $self->sanlog(-msg => "Creating a Armadillo Object for bridge: '$bridge'");

    # Set the timeout if one is given.
    $self->timeout($timeout) if (defined $timeout);

    # Set the debug_level if one is given.
    $self->set_debug_level($debug_level) if (defined $debug_level);

    # initialize the number of connect retries on the object
    $self->{connection_retries} = $connection_retries;

    # Determine if a connection type has been provided
    if (!defined($connection)) {
        $connection = $self->_connect($bridge, $timeout, $conn_type);
        if ($self->error) {
            # Error message is already set, return object so caller
            # can get the error.
            return $self;
        }
    }

    # save the connection
    $self->{conn} = $connection;

    # Store connection parameters in object.
    $self->{host} =  $self->{conn}->{hostp}->{hostname};
    $self->{user} =  $self->{conn}->{username};
    $self->{pass} =  $self->{conn}->{password};

    # Initially connected.
    $self->{connected} = 1;

    # save value of conn_type in host object
    $self->{conn_type} = $conn_type;

    return $self;
}

################################################################################
# Subroutine Name: power_cycle ()
# Objective:       Power Cycle the Armadillo
# Details:         Use rpower command from Tharnhost file
################################################################################
sub power_cycle {
    my ($self, %args) = @_;

    my ($machine, $cmd, $status, @output);

    $self->sanlog(-msg => "Power Cyclying ''".$self->get_hostname()."''");

    # Get the rpower command
    $cmd = $self->{'conn'}->{'hostp'}->{'powercmd'};

    # Execute the rpower command through 'nate_server_execute'
    ($status, @output) = &SanUtil::nate_server_execute("-cmd" => $cmd);
    if ($status == 1){
        return $ClientUtil::FALSE;
    }
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: detect_reboot ()
# Objective:       Detect whether host rebooted or not
# Details:         See POD documentation at the end of this file
################################################################################
sub detect_reboot {
    my($self, %args) = @_;

    my ($retry, $alive, $num_tries, $machine, $machine_ip);
    my ($shutdown_timeout, $bootup_timeout, $bootup_sleep);

    $machine = $self->get_hostname();

    # Assigning default timeout values
    $shutdown_timeout = 300;
    $bootup_timeout = 300;
    $bootup_sleep = 60;

    # We will ping the host for a total of ($shutdown_timeout)
    # seconds before we fail out if host didnt stop pinging. This means we are
    # unable to soft reboot the host. Hard reboot is required
    $alive = 1;
    while ($shutdown_timeout > 0) {
        # ping the host to see if it is dead or not
        my $temp_time = time();
        if (&SanUtil::ping_host($machine_ip, $machine) == $ClientUtil::FALSE) {
            $alive = 0;
            $self->sanlog(-msg => "($machine) shutdown successfully");
            last;
        } else {
            $self->sanlog(-msg => "($machine) still alive. sleeping for 5 seconds before trying to ping again");
            &SanUtil::nate_sleep(5);
            $shutdown_timeout = $shutdown_timeout - (time() - $temp_time);
        }
    }

    # check the status of shutdown and report
    if ($alive) {
        $self->_set_err("Not able to soft reboot the Machine ($machine). " .
                        "User should manually try to hard reboot the machine");
        return $ClientUtil::FALSE;
    }

    # sleep just enough to prime the reboot process
    $self->sanlog(-msg => "Sleeping for $bootup_sleep seconds.");
    &SanUtil::nate_sleep($bootup_sleep);

    # We will ping host for a total of ($bootup_timeout + $bootup_sleep)
    # seconds before we fail out if host still didnt ping. This means that host
    # didnt come up after shutdown. Hard reboot is required
    $alive = 0;
    while ($bootup_timeout > 0) {
        # ping host to see if it is up
        my $temp_time = time();
        if (&SanUtil::ping_host($machine_ip, $machine) == $ClientUtil::TRUE) {
            $alive = 1;
            $self->sanlog(-msg => "($machine) rebooted successfully");
            last;
        } else {
            $self->sanlog(-msg => "($machine) not up sleeping 30 seconds");
            &SanUtil::nate_sleep(30);
            $bootup_timeout = $bootup_timeout - (time() - $temp_time);
        }
    }

    # check status of reboot and report
    if ($alive == 0) {
        $self->_set_err("Machine ($machine) never came back up after reboot. " .
                        "User should manually try to hard reboot the machine");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: destroy ()
# Objective:       destroy armadillo object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self) = @_;

    if (defined $self->{conn}) {
        $self->sanlog(-msg => "Disconnecting Armadillo object '" . $self->get_hostname() . "'");
        &Tharn::disconnect($self->{conn});
    }
    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    execute
# Objective: Issue arbitrary command to host and get results.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub execute {
    my ($self, %args) = @_;

    my (@arr, $buffer, $hide, @rows, $cmd_fmt, $hostcmd, $errmsg, $t0, $t1);
	my ($match, $before, $after, $cmd);

    $hide = 0;

    foreach (keys %args) {
        if (/-cmd/) {
            length($cmd = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"cmd\":",
                                    "$args{$_}");
                    return $self;
                };
        }
    }

    # Clear errors.
    $self->_clear_err;

    unless (length $cmd) {
        $self->_set_err("Invalid CLI command");
        return [];
    }

    if (!$hide and $self->{show_cmd}) {
        $self->sanlog(-msg => $cmd, -use_loghdl => 0);
    }

    # Init execution timer
    $t0 = gettimeofday;

    # Execute the command
    my $status;
    my @output;
    my $tmo = $self->timeout();

    # Use 'say' to execute the command
    ($match, $before, $after) = &Tharn::say($self->{conn}, $cmd, '-timeout', $tmo);

    # strip output and remove CR/LF
    chomp($after);
    $after =~ s/\r//g;

    # strip any telnet ctrl characters that get
    # added to the output (this problem seems
    # to only occur with certain platforms.
    $after =~ s/ \cH//g;

    # create an array from the output
    @output = split(/\n/, $after);

    # strip original command out of the output
    if ($output[0] =~ /$cmd/) {
        shift @output;
    }

    # Check output for known error strings to determine status
    if ($after =~ /^ERROR/i || $after =~ /invalid command/i) {
        $self->_set_err("Error: There was an invalid command.");
        return undef;
    }

    # Get the post execution time and save the command time
    $t1 = gettimeofday;
    $self->{command_time} = ($t1 - $t0);

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_connection()
# Objective: Get nate connection object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_connection {
    my ($self) = @_;

    return $self->{conn};
}

###############################################################################
# Method:    get_disk_logins()
# Objective: Get the disk serial numbers that are logged into the bridge.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_logins {
    my ($self) = @_;

    my @disks = ();

    # Execute the command
    my @output = $self->execute(-cmd => "sastargets");
    if ($self->error){
        return undef;
    }

    # @output:
    # 0 NETAPP   X421_HCOBD450A10 Disk          PNG3X34A                    \r\n
    # 1 NETAPP   X421_HCOBD450A10 Disk          PNG4427A                    \r\n
    # 2 NETAPP   X421_HCOBD450A10 Disk          PNG4UWPA                    \r\n
    # Get the serial number from the output and add it to an array:
    foreach my $line (@output) {
        if ($line =~ /\s+([-\S]+)\s+$/) {
            push(@disks, $1);
        }
    }
    print Dumper(@disks);
    return \@disks;
}

###############################################################################
# Method:    get_serial_number()
# Objective: Get the serial number of the Armadillo Bridge
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_serial_number {
    my($self) = @_;

    my(@output, $line, $part_id, $seq_num);

    @output = $self->execute(-cmd => "get serialnumber");
    if ($self->error){
        return undef;
    }

    # Output format:
    # Part Identifier: FB6500N Sequence Number: 100126
    foreach $line (@output){
        if ($line =~ /:\s(.*)\sSequence Number:\s([\S]*)$/) {
            $part_id = $1;
            $seq_num = $2;
        }
    }
    return $part_id.$seq_num;
}

###############################################################################
# Method:    get_fcdatarate()
# Objective: Get the data rate of the Armadillo Bridge
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_fcdatarate {
    my ($self) = @_;

    my (@output, $line, $match_comp, $data_rate);

    @output = $self->execute(-cmd => "get fcdatarate all");
    if ($self->error) {
        $self->_set_err("Failure running command 'get fcdatarate all'");
        return $ClientUtil::FALSE;
    }

    $match_comp = $ClientUtil::FALSE;

    # @output format
    # 4
    # ; Port    Data Rate
    # ;==================
    # 1      auto
    # 2      auto
    foreach $line (@output) {
        if ($line =~ /.*(\d)\s+([\w-]+)$/) {
            $data_rate->{$1}->{'Data Rate'} = $2;
            $match_comp = $ClientUtil::TRUE;
        }
    }
    ## Return #data_rate if we were able to match and get data
    return $data_rate if ($match_comp);
    return $match_comp;
}

###############################################################################
# Method:    get_fcconnmode()
# Objective: Get the connection mode of the Armadillo Bridge
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_fcconnmode {
    my ($self) = @_;

    my (@output, $line, $conn_mode, $match_comp);

    @output = $self->execute(-cmd => "get fcconnmode all");
    if ($self->error) {
        $self->_set_err("Failure running command 'get fcconnmode all'");
        return $ClientUtil::FALSE;
    }

    $match_comp = $ClientUtil::FALSE;

    # @output format
    # 4
    # ; Port    Conn Mode
    # ;==================
    # 1      loop-ptp
    # 2      loop-ptp
    foreach $line (@output) {
        if ($line =~ /.*(\d)\s+([\w-]+)$/) {
            $conn_mode->{$1}->{'Conn Mode'} = $2;
            $match_comp = $ClientUtil::TRUE;
        }
    }
    ## Return #conn_mode if we were able to match and get data
    return $conn_mode if ($match_comp);
    return $match_comp;
}

###############################################################################
# Method:    get_info()
# Objective: Get the info from the Armadillo Bridge
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_info {
    my ($self) = @_;

    my (@output, $line, $info_stats, @info_line, $key, $value);

    @output = $self->execute(-cmd => "info");
    if ($self->error) {
        $self->_set_err("Failure running command 'info'");
        return $ClientUtil::FALSE;
    }

    # @output format:
    # Device Status         = Good
    # Device                = "FibreBridge 6500N"
    # Serial Number         = FB6500N100126
    # Device Version        = A64M
    foreach $line (@output) {
        if ($line =~ /((\S+\s)+)\s+=\s(.+)$/) {
        $info_stats->{$1} = $+;
        }
    }
    return $info_stats;
}

###############################################################################
# Method:    get_fcporterrors()
# Objective: Get the Fiber Channel error stats from the Armadillo Bridge
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_fcporterrors {
    my ($self) = @_;

    my (@output, $line, $error_stats);

    @output = $self->execute(-cmd => "get fcporterrors all");
    if ($self->error) {
        $self->_set_err("Failure running command 'get fcporterrors all'");
        return $ClientUtil::FALSE;
    }

    # @output format:
    # ; Fibre Channel Error Counts
    # ; Port | Link Failures | Sync Loss | Signal Loss | Invalid Tx | Invalid CRC
    # ;==========================================================================
    # 1                 6           1             1          510             0
    # 2                 1           1             1          510             0
    foreach $line (@output) {
        if ($line =~ /.*\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)$/) {
            $error_stats->{$1}->{'Link Failures'} = $2;
            $error_stats->{$1}->{'Sync Loss'} = $3;
            $error_stats->{$1}->{'Signal Loss'} = $4;
            $error_stats->{$1}->{'Invalid Tx'} = $5;
            $error_stats->{$1}->{'Invalid CRC'} = $6;
        }
    }
    return $error_stats;
}

###############################################################################
# Method:    reset_fcerrorstats()
# Objective: Reset the Fiber Channel error stats on the Armadillo
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reset_fcporterrors {
    my ($self) = @_;

    $self->execute(-cmd => "resetfcporterrors all");
    if ($self->error) {
        $self->_set_err("Failure running command 'resetfcporterrors all'");
        return $ClientUtil::FALSE;
    }
    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    reboot()
# Objective: Reboot the Armadillo Bridge
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reboot {
    my ($self) = @_;

    my $count = 10;
    my $status = $ClientUtil::TRUE;

    $self->sanlog(-msg => "Rebooting the bridge '" . $self->get_hostname() . "'");
    &Tharn::say($self->{conn}, "firmwarerestart", '-timeout', $self->timeout(), "-blind");

    while ($status == $ClientUtil::TRUE){
        return $ClientUtil::FALSE if ($count == 0);
        $status = &SanUtil::ping_host($self->get_hostname());
        $count--;
        &SanUtil::nate_sleep(2);
    }
    $self->sanlog(-msg => "Bridge down. Sleeping for 10 seconds.");
    &SanUtil::nate_sleep(10);

    $self->sanlog(-msg => "Reconnecting to the bridge '" . $self->get_hostname() . "'");
    $status = $self->reconnect();
    if ($status == $ClientUtil::FALSE) {
        $self->_set_err("Could not reconnect to bridge");
        return $ClientUtil::FALSE;
    }
    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    hostname()
# Objective: Get client host name from the host record
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_hostname {
    my ($self) = @_;

    return $self->{conn}{hostp}{hostname};
}

###############################################################################
# Method:    hosttype()
# Objective: Get client host type from the host record
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_hosttype {
    my ($self) = @_;

    return $self->{conn}{hostp}{hosttype};
}

###############################################################################
# Method:    reconnect()
# Objective: reconect to host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reconnect {
    my ($self, %args) = @_;

    my ($bridge, $timeout, $conn_type);

    # default action is to disconnect the connect that currently exists
    my $disconnect_existing_conn = exists($args{-disconnect_existing_conn}) ? $args{-disconnect_existing_conn} : 1;
    $self->_clear_err();
    if (defined($self->{conn}) and $disconnect_existing_conn == 1) {
        Tharn::disconnect($self->{conn}, '-ok');
    }

    $bridge = $self->{host};
    $timeout = $self->{timeout};
    $conn_type = $self->{conn_type};

    $self->sanlog(-msg => "Reconnecting to host machine '" . $self->get_hostname() . "'");
    $self->{conn} = $self->_connect($bridge, $timeout, $conn_type);

    if ($self->error) {
        my @error = $self->errmsg();
        my $errmsg  = join('\n', @error);
        $self->_set_err("connect failed: $errmsg");
        return ($ClientUtil::FALSE);
    }

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    timeout
# Objective: set or get telnet timeout
# Details:   See POD documentation at the end of this file.
###############################################################################
sub timeout {
    my ($self, $timeout) = @_;

    if (defined $timeout and length $timeout) {
        $self->{timeout} = $timeout;
    }
    return $self->{timeout};
}

###############################################################################
# Method:    display_info
# Objective: dump client information for diagnostic/audit purposes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub display_info {
    my ($self) = @_;

    my (@output, $string, @disks, %fc_errors, %fc_connmode, %fc_datarate);
    my ($hostname, $comment, $comment_minor, $disk, %info, $key, $serial);

    $comment =       "**************************************************************\n";
    $comment_minor = "--------------------------------------------------------------\n";

    ## Retrieve the necessary data about the Armadillo
    $hostname = $self->get_hostname();
    $serial = $self->get_serial_number();
    @disks = @{$self->get_disk_logins()};
    %info = %{$self->get_info()};
    %fc_errors = %{$self->get_fcporterrors()};
    %fc_connmode = %{$self->get_fcconnmode()};
    %fc_datarate = %{$self->get_fcdatarate()};
    
    push(@output, "\n" . $comment . "Armadillo Summary for $hostname\n" . $comment);

    push(@output, "Serial Number: " . $serial . "\n");

    ## Print the disks in three columns
    push(@output, $comment . "Disks:\n" . $comment_minor);
    my $count = scalar(@disks) / 3;
    while ($count > 0) {
        push(@output, sprintf ("|   %-18s%-18s%-18s   |\n",
                               shift(@disks),
                               shift(@disks),
                               shift(@disks)));
        $count--;
    }
    push (@output, $comment_minor);

    push (@output, $comment);

    ## Skip the serial number since we already printed that
    foreach $key (keys %info) {
        push (@output, sprintf ("%-25s = %-25s\n", $key, $info{$key})) 
             unless $key =~ /^Serial Number/;
    }
    push (@output, $comment);

    ## Print the port errors from the fc ports
    foreach $key (keys %fc_errors) {
        push (@output, "Errors for FC Port #" . $key . ":\n");
        push (@output, sprintf ("%-15s = %-5s\n", "Invalid Tx", $fc_errors{$key}->{'Invalid Tx'}));
        push (@output, sprintf ("%-15s = %-5s\n", "Signal Loss", $fc_errors{$key}->{'Signal Loss'}));
        push (@output, sprintf ("%-15s = %-5s\n", "Sync Loss", $fc_errors{$key}->{'Sync Loss'}));
        push (@output, sprintf ("%-15s = %-5s\n", "Invalid CRC", $fc_errors{$key}->{'Invalid CRC'}));
        push (@output, sprintf ("%-15s = %-5s\n", "Link Failures", $fc_errors{$key}->{'Link Failures'}));
    }
    push (@output, $comment);

    ## Print the connection mode of the fc ports
    foreach $key (keys %fc_connmode){
        push (@output, "Connection Mode for FC Port #" . $key . " = " . $fc_connmode{$key}->{'Conn Mode'});
        push (@output, "\n");
    }
    push (@output, $comment);

    ## Print the data rate of the fc ports
    foreach $key (keys %fc_datarate){
        push (@output, "Data Rate for FC Port #" . $key . " = " . $fc_datarate{$key}->{'Data Rate'});
        push (@output, "\n");
    }

    push (@output, $comment);

    #$self->sanlog(-msg => join('', @output));

    return \@output;
}

###############################################################################
# Method:    ftp_firmware
# Objective: FTP the firmware from a specified server to the bridge.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub ftp_firmware {
    my ($self, %args) = @_;

    my ($conn_str, $connection, $errmsg);
    my ($host, $host_username, $host_pass, $file);
    my ($bridge_name, $bridge_pass, $bridge_user);
    my ($match, $before, $after);

    foreach (keys %args) {
        if (/^-?host$/) {
            length($host = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"host\":",
                                    "$args{$_}");
                    return $ClientUtil::False;
                };
        } elsif (/^-?host_username$/) {
            length($host_username = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"host_username\":",
                                    "$args{$_}");
                    return $ClientUtil::False;
                };
        } elsif (/-?host_pass$/) {
            length($host_pass = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"host_pass\":",
                                    "$args{$_}");
                    return $ClientUtil::False;
                };
        } elsif (/-?file$/) {
            length($file = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"file\":",
                                    "$args{$_}");
                    return $ClientUtil::False;
                };
        }
    }

    $bridge_name = $self->get_hostname();
    $bridge_pass = $self->{pass};
    $bridge_user = $self->{user};

    $conn_str = "" . $host . "#ssh_0;" .
                "conntype=ssh-persist;" .
                "username=" . $host_username . ";" .
                "password=" . $host_pass . ";" .
                "timeout=180;inherit=0;";
    my $max_retries = 5;
    my $retries = 0;

    ## Connect to the host where the file will be ftp'd from
    while (!defined($connection) and $retries < $max_retries) {
        $connection = &Tharn::connect(
            $conn_str,
            -eval=>sub {
                if (@_) {
                    $retries++;
                    $connection = undef;

                    $errmsg = join("", @_);
                    $errmsg =~ s/\n//g;

                    &SanUtil::nate_log('-fatal', "Unable to connect to host.\n$errmsg\n");
                    &Tharn::disconnect($conn_str, '-ok');
                    &SanUtil::nate_sleep(30);
                }
            });
    }

    ## FTP to the bridge to put the firmware file on it
    $self->sanlog(-msg => "Sending FTP Command");
    &Tharn::say($connection, "ftp $bridge_name",
                          '-timeout', "30",
                          '-wait', "$bridge_name:",
                          '-noprompt',
                          '-comment', "Starting FTP");

    ## Send the username to the bridge
    $self->sanlog(-msg => "Sending username");
    &Tharn::say($connection, "$bridge_user",
                          '-timeout', "30",
                          '-prompt', "Password:");

    ## Send the password to the bridge
    $self->sanlog(-msg => "Sending password");
    &Tharn::say($connection, "$bridge_pass",
                          '-timeout', "30",
                          '-prompt', "ftp> ",
                          '-wait', "accepted");

    ## Armadillo requires that binary mode is on
    $self->sanlog(-msg => "Turning on binary mode");
    &Tharn::say($connection, 'bin',
                          '-timeout', "30",
                          '-prompt', "ftp> ",
                          '-wait', "200");

    ## Armadillo requires passive mode turned off
    $self->sanlog(-msg => "Checking passive mode");
    (undef, undef, $after) = &Tharn::say($connection, 'status',
                          '-timeout', "30",
                          '-prompt', "ftp> ");
    if ($after =~ /Passive mode: off/){
        $self->sanlog(-msg => "Passive mode is off.");
    } elsif ($after =~ /Passive mode: off/) {
        $self->sanlog(-msg => "Passive mode is on. Turning it off");
        &Tharn::say($connection, 'passive',
                          '-timeout', "30",
                          '-prompt', "ftp> ",
                          '-wait', "off");
    } else {
        $self->_set_err("Could not determine passive mode status");
        $connection->disconnect();
        return $ClientUtil::False;
    }

    ## Send the firmware file to the bridge
    $self->sanlog(-msg => "Sending file");
    (undef, undef, $after) = &Tharn::say($connection, "put $file",
                          '-timeout', "240",
                          '-prompt', "ftp> ");

    ## Check to make sure the the bridge accepted the firmware
    if ($after =~ /No such file/ or $after =~ /Illegal File Size/ or $after =~ /Cross Flash Protection/) {
        $self->_set_err("Could not FTP file to $bridge_name. " .
                        "This could be a corrupted firmware file or the file doesn't exist.");
        $connection->disconnect();
        return $ClientUtil::False;
    }

    $connection->disconnect();
    return $ClientUtil::True;
}


###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _connect
# Objective: Create telnet connection to host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _connect {
    my ($self, $hostname, $timeout, $conn_type) = @_;

    my ($connect, $conn_str, $errmsg, $host, $id);
    my ($user, $passwd, $hostInfo, $retries, @preferred_conns);

    # Variable init.
    $self->{connected} = 0;

    $id = $Armadillo::_CONNECTION_COUNT++;

    $host = Utils::make_host($hostname);
    my $max_retries = $self->{connection_retries};

    if ($conn_type eq "THARNHOST") {
        $conn_type = "";

        # get list of preferred connection types for this
        # platforms. No error to check in this context.
        @preferred_conns = $self->_get_preferred_conn_types($host);

        # Check to see if any connection types are defined, if not return the default value
        if (!defined($host->{connection_fields}) or
            scalar(keys(%{$host->{connection_fields}})) == 0) {
            # Set the $conn_type to the default of telnet
            $conn_type = "telnet";
        } elsif (scalar(@preferred_conns) > 0) {
            # If we find preferred connection types in the
            # for this platform, try each of those to see if they
            # are defined, if one is, use it.
            foreach my $type (@preferred_conns) {
                if (defined($host->{connection_fields}->{$type})) {
                    $conn_type = $type;
                    last;
                }
            }
        }
        # If conn_type is still empty then get the preferred connection type from the
        # function
        if ($conn_type eq "") {
            $conn_type = $self->_get_preferred_conn_types();
        }
    }

    unless (defined($timeout) and length($timeout)) {
        $self->timeout($host->{connection_fields}{$conn_type}{timeout});
    }

    if (defined($user) and length($user)) {
        $host->{connection_fields}{$conn_type}{username} = $user;
    }

    if (defined($passwd) and length($passwd)) {
        $host->{connection_fields}{$conn_type}{password} = $passwd;
    }

    if ($conn_type eq "ssh" or $conn_type eq "ssh-root") {
        $conn_str = $hostname . "#ssh" . "_" . $id . ";" .
            "conntype=ssh-persist;" .
            "username=" . $host->{connection_fields}{$conn_type}{username} . ";" .
            "password=" . $host->{connection_fields}{$conn_type}{password} . ";" .
            "timeout="  . $self->{timeout} . ";inherit=0;";
    } else {
        $conn_str = $hostname . "#" . $conn_type . ";" .
            "username=" . $host->{connection_fields}{$conn_type}{username} . ";" .
            "password=" . $host->{connection_fields}{$conn_type}{password} . ";" .
            "timeout="  . $self->{timeout} . ";inherit=0;";
    }

    $self->sanlog(-msg_level => 0,
                  -msg => "Connection string: $conn_str");
    while (!defined($connect) and $retries < $max_retries) {
        $connect = &Tharn::connect(
            $conn_str,
            -eval=>sub {
                if (@_) {
                    $retries++;
                    $connect = undef;

                    $errmsg = join("", @_);
                    $errmsg =~ s/\n//g;

                    # Display an error in case user doesn't include error
                    # handler on object instantiation.
                    $self->sanlog(-msg => "WARNING: " .
                                          "unable to connect to Armadillo '" .
                                          $hostname."': $errmsg\n" .
                                          "Attempt $retries of $max_retries");
                    &Tharn::disconnect($conn_str, '-ok');
                    &SanUtil::nate_sleep(30);
                }
            });
    }

    # if retry count has been exhausted, set error and return
    if ($retries == $max_retries) {
        $self->_set_err($errmsg);
        return {};
    }

    # return the connection and the conn type
    return $connect;
}

################################################################################
# Subroutine Name: is_reboot()
# Objective:       reboot host
# Details:         See POD documentation at the end of this file
################################################################################
sub is_reboot {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{reboot} and $self->{reboot} =~ /1|0/) {
        return $self->{reboot};
    } else {
        $self->{reboot} = 0;
    }

    return $self->{reboot};
}

################################################################################
# Subroutine Name: _get_preferred_conn_types
# Objective:    get the preferred type of connection (telnet) for the Armadillo
#               bridge.
################################################################################
sub _get_preferred_conn_types {

    return("telnet");

}

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Armadillo.pm

=head1 Author

Aaron Kirk

=head1 Description

Armadillo.pm provides an API interface to a Aramdillo through the
use of the Telnet library.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm>

=head1 Attributes

=head2 B<connected>

boolean = <1 | 0>

=head2 B<show_cmd>

boolean = <1 | 0> (default: 1 (true) By default log commands executed)

=head2 B<show_data>

boolean = <1 | 0>

=head2 B<protocol>

string = <fcp | iscsi> (default: fcp)

=head2 B<execute>

boolean = <1 | 0> (default: 1)

=head2 B<log>

file handle = File handle used for logging

=head2 B<connect>

connection object = Connection to the host (as returned from Tharn::connect())

=head2 B<conn>

connection object = Connection to the host

=head2 B<host>

string = Host name, connection parameter

=head2 B<user>

string = Username, connection parameter

=head2 B<pass>

string = Password, connection parameter

=head2 B<obj>

hash = Hash of children objects (devmgr, lvm)

=head2 B<command_time>

integer = time in seconds of the last command executed

=head2 B<timeout>

integer = timeout to use for say calls

=head2 B<debug_level>

integer = Current debug level

=head2 B<no_indent>

boolean = <1 | undef> Do not indent log entries when set

=head1 Public Methods

=head2 B<new(%args)>

Creates a new Armadillo object.  Uses the specified IP address or machine name
to create a telnet connection. If the connection is successful, the calling
program will be logged into the CLIENT using either the default user/password
specified in the NATE_HOSTS file or a user and password supplied by the caller.

=over 4

=item Inputs

=item * -host

Machine name or IP address of the remote machine.

=item * I<-user> (default: Tharnhost username entry).

Override default username for armadillo login.

=item * I<-passwd> (default: Tharnhost password entry).

Override default password for armadillo login.

=item * I<-timeout> (default: Tharnhost telnet timeout entry).

Telnet timeout in seconds

=item * I<-debug_level> (default: 0).

Verbosity level for Sanlog::sanlog()

=item * I<-loghdl>

Filehandle to log output too.

Override the Telnet timeout value.

=item * I<-show_cmd> <1 | 0> (default: 1 (true) ).

By default commands executed will be logged   This can be
turned off by passing the -show_cmd with a value of 0.

=item * I<-connection>

If specified, a active/live connection must be passed in, the &connect function
must be called prior to calling the SanClient::new function

=item * I<-conn_type> <THARNHOST | rsh | telnet | ssh | rlogin | console> (NO DEFAULT).

This parameter will determine what connection type will be used for the client.
THIS CAN NOT BE USED IN CONJUNCTION WITH '-connection'. This parameter will NOT
override '-connection'.

=item Outputs

Returns a Armadillo object.  If a successful Telnet connection and login was
achieved, an open Telnet connection is maintained by the object until
the calling program exits, or the B<disconnect> method is explicitly
called.

=item Examples

 # Create a Armadillo object with a connection to a remote machine.

 $bridge = new Armadillo(
        -host => '10.228.213.189',
        -user => 'root',
        -passwd  => 'Password');

  if ($bridge->error) {
     $Test->nlog($bridge->errmsg());
     return $TCase::FAIL;
 }

=back

=cut

1;
