#
# $Id: //depot/prod/test/Rsierranevada/lib/AsupExtractor.pm#1 $
#
# Copyright (c) 2009 Network Appliance, Inc.
#   All rights reserved.

package AsupExtractor;

use Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(extractAsupFile);

use strict;
use MIME::Parser;

=head1 NAME

ASUP - extract information from ASUP files collected from filers

=head1 SYNOPSIS

use AsupExtractor qw(extractAsupFile);

extractAsupFile($input_file, $output_dir);

=head1 DESCRIPTION

This module provides a mechanism to extract ASUP information from files, as
generated by NetApp filers.

=head1 FUNCTIONS

=head2 extractAsupFile(inputFile, outputDir)

This function extracts data from a raw ASUP file, copied from the
/mroot of a NetApp filer.  The contents of the ASUP file are extracted
into seperate files which are placed in the output directory.  The output
directory will be created, if it does not already exist.

Note that this function is intended to allow convenient access to
the raw content of an individual ASUP.  No attempt is made to
perform any form of indexing or data mining of the extracted data
(such as what is done in the NetApp ASUP Data Warehouse).

=head3 ARGUMENTS

=over 4

=item inputFile

The name of the raw ASUP file.

=item outputDir

The name of the output directory to which to extract the files
contained in the ASUP.

=back

=head2 RETURNS

0 on success
non-zero otherwise

=head1 BUGS

This library currently provides very limited functionality:
extract information from a raw ASUP file (a file on disk, without
proper MIME headers).  Ideally, this library would be made to
be more flexible to deal with reading input from arbitrary filehandles,
or input which already has proper MIME headers present.

=cut

sub extractAsupFile {
    my $input_file = shift;
    my $output_dir = shift;

    $input_file || warn "input file required" && return -1;
    $output_dir || warn "output dir required" && return -1;

    # Make sure the output directory exists.
    if (!-d $output_dir) {
	mkdir $output_dir;
	if (!-d $output_dir) {
	    warn "cannot create $output_dir: $!";
	    return -1;
	}
    }

    # The raw ASUP files lack a proper MIME header.  This is normally
    #  tacked on when the ASUP is sent out, because the MIME header
    #  for SMTP transport is different from the MIME header for HTTP
    #  transport.  To compensate for the lack of a header, we fake
    #  up one, here.  Note that the "easy" way to do this would be
    #  to edit the input file, but I'd prefer to treat the input file
    #  as read-only.  Consequently, I create the following command
    #  pipeline to cough up a well-formed MIME header.  And it just so
    #  happens that MIME::Parser is capable of taking a command pipeline
    #  as its input.
    my $boundary = "netapp mime boundary";
    my $cmd = 
      "echo MIME-version: 1.0"
	. "; echo Content-type: multipart/mixed\\; boundary=\\\"$boundary\\\""
	. "; echo"
        . "; echo --$boundary"
	. "; echo Content-type: text/plain"
	. "; echo"
	. "; cat $input_file"
	. " |";

    # Use MIME::Parser to do all the heavy lifting.
    my $parser = new MIME::Parser();
    $parser->output_dir($output_dir);
    $parser->parse_open($cmd);

    # Most (probably all) of the files extracted from the ASUP
    #  are going to be gzipped.  Unzip them.  Note that gunzip will
    #  delete the original by default.
    while (my $gzfile = <$output_dir/*.gz>) {
	`gunzip -f $gzfile`;
	if ($?) { warn "could not gunzip: $?"; }
    }

    # If there are any tar balls, untar them and delete the originals.
    while (my $tarfile = <$output_dir/*_tar_ball $output_dir/*.tar>) {
	`tar -xf $tarfile -C $output_dir`;
	if ($?) {
	    warn "could not untar $tarfile: $?";
	    # if we failed to untar the file, don't delete
	    next;
	}

	unlink $tarfile || warn "could not unlink $tarfile: $!" && return -1;
    }

    return 0;
}

# return value so perl doesn't have a cow.
return 1;
