# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/Backup.pm#1 $

## @summary Backup library -- backup status routines
##
## @description
##    This package defines routines which attempt to get and parse
##    filer output using the "say" command in Tharn.  These routines
##    gather information from the backup status command and return it.
##
##    All the subroutines here expect a filer connection as their
##    first argument.  If the connection is invalid (causing the say
##    command to fail) or if any other errors occur while attempting
##    to get the requested information, these subroutines will throw
##    a fatal signal -- tstlog('-fatal', "Error...") or the like.
##    Don't call these routines from tests unless it is acceptable for
##    the tests to fail if the information isn't available.
##
## @dependencies
## The functions depend on exact output format to parse and extract results.
## Even a minor change in the UI can result in this function not working properly.
##
## @bugs
##
## @author dl-dpgqa-automation@netapp.com
## @cf     false

package Backup;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(getRstBackups);

use SayWrap;
use Tharn;
use strict;

## @name ::getRstBackups
## @synopsis getRstBackups($cfiler, $volume)
## Returns an array of references to hashes. Each hash has following keys:
## id, type, start, level, path
## (Note: Active backups are not returned as part of this list.)
## @description Gets a list of all restartable backups
## @param $cfiler Filer connection object
## @param $volume Volume to get backups for.
## @example To get the id of last (most recent) backup:
##  my @backups = getRstBackups($cfiler, 'myvol');
##  if(@backups) {
##     my $id = (pop @backups)->{id};
##  }else {
##    tstlog(-warn => "No backups found");
##  }

sub getRstBackups($$) {
  my ($filer, $volume) = @_;

  my $text = sayEOF($filer, "backup status");

  # Return if list is empty
  return ()
    if ($text =~ /No backup activity/); 

  # Strip everything up to the beginning of the list
  # (Keep everything after the
  # "--  -----------  ----  ------  ------------  -----  ---------------")
  tstlog('-fatal', "Error parsing output.")
    unless ($text =~ s/.*?\s*\-+\s+\-+\s+\-+\s+\-+\s+\-+\s+\-+\s+\-+\s*(.*)$/$1/s);

  chomp $text;
  my @backups = ();
  my @lines = split(/\n/,$text);
  foreach my $line (@lines) {
    $line =~ s/(\r|\n)$//;
# parse:
#  3  RESTARTABLE  dump          Nov 12 10:13    0    /vol/compat_data/
    if($line =~ 
/^\s*(\d+)\s+\RESTARTABLE\s+(\w+)\s+(\w+\s+\d+\s+\d+:\d+)\s+(\d+)\s+(\/vol\/$volume\/\S*)$/) {
    push (@backups, {id => $1, type => $2, start => $3,
		    level => $4, path => $5});
    }
  }
  return @backups;
}
