#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/Burt.pm#2 $
#
#Copyright (c) 2002 Network Appliance, Inc.
#All rights reserved.
#
## @summary Library of functions to process and track burts 

## @description
## This package has two sets of functions:
## The functions that begin with 'Exit',
## are wrappers for NATE functions that log test results 
## and when appropriate, exit the test script.
##
## The log file output from these functions includes test script call stack information
## so that one can quickly jump from the logged error message 
## into the test script that reported the error.
## The syntax of these log messages are also compatible with the lib/nuffo.thpl
## program that runs as a NATE POSTTEST to collect and store test results.
##
## The set of functions that have a 'burt_' prefix enable test execution
## to correlate test results with known burts.
## When the results do not match the state of the burt, nuffo.thpl
## will add appropriate text to its output, to aid in manually updating
## the relevant burt reprots.
##
## It defines values for test script return codes,
## a set of functions to exit from a test script,
## and a function to announce detected burts.
## @usage
##   use Burt;
##
## @author Charlie Hedstrom (hedstrom@netapp.com)
## @keywords burt return 
## @status public
## @change 2003-09-24 hedstrom: created from earlier disorganized package (config_base.pm)
## @change 2003-11-13 hedstrom: cope with target release = '-' when it is unassigned
## @change 2004-02-18 hedstrom: use burt state during test execution to determine "deferred" status
## @change 2004-02-25 hedstrom: include full call stack with ExitConfigError
## @change 2004-05-11 hedstrom: prepare for NATE module name changes in CallStack, set tab at 4
## @change 2004-08-11 hedstrom: use die for abortive Exit functions (new NATE convention)
## @change 2004-08-11 hedstrom: back out die, CONFIG was reported as FATAL
## @change 2004-08-13 hedstrom: add GetSoftErrorCount
## @change 2004-11-29 hedstrom: add testcase_start
## @change 2004-12-09 hedstrom: add description argument to testcase_start
## @change 2006-08-30 hedstrom: add INFO result for each testcase_start
## @change 2007-08-27 hedstrom: add $deferred status as a return on the ExitSoftError function
## @change 2009-12-22 eharing:  added new path to burt command location 

use strict;

package Burt;
use vars qw(@ISA @EXPORT
    $MAX_SOFTERROR
    @BURTS_UNDER_TEST @BURTS_PASSED @BURTS_FAILED @BURTS_FAILED_REPORTED
    $TESTCASE_HANDLE
    );

require Exporter;
@ISA = ('Exporter');

# Export constants and functions defined by this package
@EXPORT = qw(
    EXIT_PASS EXIT_NOT_APPLICABLE EXIT_ERROR EXIT_SCRIPT_ERROR EXIT_CONFIG_ERROR
    EXIT_CHECKME
    
    burt_begin burt_end
    burt_find
    burt_fixed burt_found burt_get burt_needed
    
    burt_read burt_state
    
    ExitConfigError
    ExitError 
    ExitNA 
    ExitScriptError 
    ExitSoftError 

    GetSoftErrorCount

    testcase_configs
    testcase_start
    testcase_start_description
    testcase_subtest_end
    );

use English qw(-no_match_vars);
use Carp;
use Module::Common qw(validate_mixed);
use Params::Validate qw(ARRAYREF SCALAR UNDEF);
use Tharn;
use SharedTestBed::ConfigTypes;

# Test script completion return codes
# These are compatible with TestMap Extensions for SilkTest
use constant EXIT_PASS => 0;
use constant EXIT_NOT_APPLICABLE => 1;
use constant EXIT_ERROR => 4;
use constant EXIT_SCRIPT_ERROR => 5;
use constant EXIT_CONFIG_ERROR => 6;
use constant EXIT_CHECKME => 101;

# Message text used by lib/nuffo.thpl to track burts
use constant burt_detected_text => "Symptoms detected for Burt ";
use constant burt_not_detected_text => "Symptoms not detected for Burt ";
use constant burt_needed_text => "Possible new problem";
use constant testcase_start_text => "Start Testcase: ";
use constant testcase_start_description => "Description: ";

# Initialize Burt Statistics maintained within this package
@BURTS_UNDER_TEST = ();
@BURTS_PASSED = (); 
@BURTS_FAILED = ();
@BURTS_FAILED_REPORTED = ();
$TESTCASE_HANDLE = undef;

if (defined(&Tharn::testcase_begin)) {
    &NATE::Events::call_on_exit_early_add({
        callback => sub {
            if ($TESTCASE_HANDLE) {
                &Tharn::testcase_end(handle => $TESTCASE_HANDLE);
                $TESTCASE_HANDLE=undef;
            }
        },
        keep_on_fork => 1,
    });
    &NATE::Events::call_on_fork_add({
        callback => sub { $TESTCASE_HANDLE=undef },
        keep_on_fork => 1,
    });
}

# Keep track of failed burt status that is encountered during test execution
my %burt_status;

my $soft_error_count = 0; 


## @name ::burt_begin
## @summary Begin to test for burt symptoms.  
## @description
## Call burt_end when the burt specific test steps are over.
## Any ExitError or ExitSoftError calls made in between
## are considered symptoms of the named burt(s).
## @arg @burts Required: a list of one or more burt numbers 
## that are about to be tested.
sub burt_begin (@) {
    my $aburt;
    foreach $aburt (@_) {
        if (0==grep(/^$aburt$/,@BURTS_UNDER_TEST) ) {
            push (@BURTS_UNDER_TEST,$aburt);
        }
    }
}

## @name ::burt_end
## @summary End testing for specific burt symptoms.
## @arg @burts Required: a list of one or more burt numbers 
sub burt_end (@) {
    my @oldburts = @_;
    my @newburts = ();
    
    # Remove specified burts from @BURTS_UNDER_TEST
    foreach my $aburt (@BURTS_UNDER_TEST) {
        if (0==grep(/^$aburt$/,@oldburts) ) {
            push( @newburts, $aburt);
        }
        # Keep track of passing Burts
        if (0==grep(/^$aburt$/,@BURTS_FAILED) ) {
            if (0==grep(/^$aburt$/,@BURTS_PASSED) ) {
                push( @BURTS_PASSED, $aburt);
            }
            # Issue message so that all passing tests can be found in the log
            # (But stop once an error is seen to reduce log clutter)
            logcomment(burt_not_detected_text."$aburt");
        }
    }
    @BURTS_UNDER_TEST = @newburts;
    
    # Remove specified burts from @BURTS_FAILED_REPORTED
    @newburts = ();
    foreach my $aburt (@BURTS_FAILED_REPORTED) {
        if (0==grep(/^$aburt$/,@oldburts) ) {
            push( @newburts, $aburt);
        }
    }
    @BURTS_FAILED_REPORTED = @newburts;
}

## @name ::burt_fixed
## @summary Report that symptoms of a known burt were not detected.
## @arg $burt Required: the burt number
sub burt_fixed ($) {
    my $burt = shift;
    logcomment(burt_not_detected_text."$burt");
    
    # Keep track of passing Burts
    if (0==grep(/^$burt$/,@BURTS_FAILED) ) {
        if (0==grep(/^$burt$/,@BURTS_PASSED) ) {
            push( @BURTS_PASSED, $burt);
        }
    }
}

## @name ::burt_found
## @summary Report that symptoms of a known burt were detected.
## @arg $burt Required: the burt number
sub burt_found ($) {
    my $burt = shift;
    # Make sure all instances are reported
    # This allows the lack of a 'burt' message to indicate
    # that burt tracking is not supported by a test script
    logcomment(burt_detected_text.$burt);
    if (0==grep(/^$burt$/,@BURTS_FAILED_REPORTED) ) {
        push( @BURTS_FAILED_REPORTED, $burt);
        
        if (0==grep(/^$burt$/,@BURTS_FAILED) ) {
            push( @BURTS_FAILED, $burt);
            
            # Remove it from Passing Burts
            my @newpassingburts = ();
            my $aburt;
            foreach $aburt (@BURTS_PASSED) {
                if ($aburt ne $burt) {
                    push( @newpassingburts, $aburt);
                }
            }
            @BURTS_PASSED = @newpassingburts;
        }
    }
}

## @name ::burt_get
## @summary Return list of burts being tested
sub burt_get () {
    return @BURTS_UNDER_TEST;
}

## @name ::burt_needed
## @summary Report that possible new burt symptoms were detected.
sub burt_needed () {
    logcomment(burt_needed_text);
}

## @name ::burt_find
## @summary Returns a list of burt ids that match the specified query
## @arg $query Required: the burt query text 
## (compatible with 'burt report' command line input)
sub burt_find($) {
    my $query = shift;
    my $burt_fields = shift;

    # Build up the burt report command.
    my $burt_report = &Tharn::param("BURT_REPORT", -default, "/u/burt/bin/burt") . " report ";
    $burt_report .= "-f:'id'";

    # Now find the requested burt
    my $burt_cmd = $burt_report . " -q'$query'";
    my @burts = qx($burt_cmd) 
        or die "Cannot run the burt command: $OS_ERROR\n";

    shift(@burts);  # Get rid of header.

    # return the burt ids.
    return @burts;
}

## @name ::burt_read
## @summary Return a hash of burt data for one burt
## @arg $burt Required: the burt number
## @arg $burt_fields Optional: a blank delimited string of burt field names to be returned.
## The default is 'id state dup target_release open_in_release testme_in_release fixed_in_release untested_in_release'
sub burt_read($;$) {
    my $burt_id = shift;
    my $burt_fields = shift;

    my %burt;

    if (!defined($burt_fields)) {
        $burt_fields = 'id state dup target_release open_in_release testme_in_release fixed_in_release untested_in_release';
    }

    # Build up the burt report command.
    my $burt_report = &Tharn::param("BURT_REPORT", -default, "/u/burt/bin/burt") . " report ";
    $burt_report .= "-f:'$burt_fields'";

    # Now find the requested burt
    my $burt_cmd = $burt_report . " -q'\$id eq $burt_id'";
    my @burts = qx($burt_cmd) 
        or die "Cannot run the burt command: $OS_ERROR\n";

    shift(@burts);  # Get rid of header.
    if (@burts == 1) {
        my @fields = split(" ",$burt_fields);
        my @data = split(" ",$burts[0]);
        my $field_count = 0;
        foreach my $field (@fields) {
            $burt{$field} = $data[$field_count];
            $field_count++;
        }
    } elsif (@burts == 0) {
        print("Burt $burt_id could not be found\n");
    } else {
        print("Multiple Burt $burt_id entries found\n");
    }

    # return the burt fields.
    return %burt;
}

## @name ::burt_state
## @summary Return the state (deferred,old,NEW,OPEN,FIXED,UNTESTED,TESTME) of a burt for the specified target release
## @arg $burt Required: the burt number
## @arg $releases Required: a comma separated string of burt target_release(s) that may own the requested state
## This string is often defined by the NUFFO_TARGET_RELEASE parameter.
## @arg $future_releases Required: a comma separated string of future burt target_release(s) that may own the requested state.
## This string is often defined by the NUFFO_TARGET_RELEASES parameter.
sub burt_state ($;$$) {
    my ($burt,$releases,$future_releases) = @_;
    if (!defined($releases) or $releases =~ m/^\s*$/) {
        $releases = param('NUFFO_TARGET_RELEASE','-default','');
    }
    if (!defined($future_releases) or $future_releases =~ m/^\s*$/) {
        $future_releases = param('NUFFO_TARGET_RELEASES','-default','');
    }
    my $future_rels = ','.$future_releases.',-future-,';
    
    my $burt_state = 'unknown';
    
    # See if we have the burt data
    my %burt = burt_read($burt);
    
    if (defined($burt{state}) and $burt{state} ne '') {
    # An unassigned target release shows as '-'
    if (defined($burt{target_release}) and $burt{target_release} ne '-' and $burt{target_release} ne '') {
        my @burt_releases =split(',',$burt{target_release});
        # The $release parameter may be a comma separated list of current releases (ONTAP, Host Attach Kits, etc)
        my @releases = split(',',$releases);
        # Find the state of this burt with respect to any release listed as current
        if ($burt{state} ne 'MULTISTATE') {
            # Not a MULTISTATE burt - make sure it is current
            foreach my $rel (@burt_releases) {
                if (0 < grep(/^$rel$/,@releases)) {
                    # This burt is listed under one of the current releases
                    $burt_state = $burt{state};
                    last;
                }
            }
        } else {
            # This IS a MULTISTATE burt, find the state with respect to a current release
            foreach my $rel (@releases) {
                if (defined($burt{open_in_release})) {
                    # Surround release list with commas so all entries become fully comma delimited
                    my $open_rels = ','.$burt{open_in_release}.',';
                    my $testme_rels = ','.$burt{testme_in_release}.',';
                    my $fixed_rels = ','.$burt{fixed_in_release}.',';
                    my $untested_rels = ','.$burt{untested_in_release}.',';
                    if ($open_rels =~ m/,$rel,/) {
                        $burt_state = 'OPEN';
                    } elsif ($testme_rels =~ m/,$rel,/) {
                        $burt_state = 'TESTME';
                    } elsif ($fixed_rels =~ m/,$rel,/) {
                        $burt_state = 'FIXED';
                    } elsif ($untested_rels =~ m/,$rel,/) {
                        $burt_state = 'UNTESTED';
                    } else {
                        next;
                    }
                    last;
                }
            } 
        }
        if ($burt_state eq 'unknown') {
            # This burt is not listed in the current releases
            # If this burt is listed in the future, it is deferred
            # else it is an old burt
            foreach my $rel (@burt_releases) {
                if ($future_rels =~ m/,$rel,/) {
                    $burt_state = 'deferred';
                    print("Burt reports that $burt is deferred until $rel\n");
                    last;
                }
            } 
        }
        if ($burt_state eq 'unknown') {
            # This burt is not listed in the current releases
            # and it is NOT deferred, it must be an old burt
            $burt_state = 'old';
            print("Burt reports that $burt was handled in $burt{target_release}\n");
        }
        
        } else {
            # Burt has not been assigned a target release
            # Consider it active, just return its state
            $burt_state = $burt{state};
        }
        if ($burt_state ne 'old' and $burt_state ne 'deferred') {
            print("Burt reports that $burt is in the $burt_state state\n");
        }
    } else {
        print("No burt report found for burt $burt\n");
    }
    
    return $burt_state;
}

## @name ::CallStack
## @summary Adds a call stack to the message list.
## @arg $msg_ref Required: Address of the list of messages
## @arg $depth Required: start depth of the call stack,
## 0 is the calling function's entry
## 1 is that function's caller.
sub CallStack ($$) {
    my $msg_ref = shift;
    my $depth = shift; 
    $depth++;  # Add one level for this subroutine
    
    my $subtest = '';
    while (my ($package,$filename,$line) = caller($depth)) {
        #print("trace: $package,$filename,$line\n");
        # Note: Tharn_Subtest.pm is to be replaced with Subtest.pm
        # Other Tharn prefixed packages to get their own names
        if ($filename =~ /Tharn_Subtest.pm|\/Subtest.pm/) {
            $subtest = 'called via subtest ';
        } elsif ($filename =~ /Tharn_Cmds.pm|Tharn.pm/ and $package eq 'Subtest') {
            $subtest = 'called via subtest ';
        } elsif ($package eq 'TestObj') {
            # The top test invocation
            last;
        } elsif ($filename =~ /Tharn|Connectrec.pm|Hostrec.pm|Param.pm|Rslt.pm/) {
            # Ignore these other NATE layers
        } else {   
            push (@{$msg_ref},$subtest."at $filename($line)");
            $subtest = '';
        }
        $depth++;
    }
}

## @name ::ExitNA
## @summary Exits the test script as NOT APPLICABLE
## @arg $msg Optional: message about why the test script is not applicable.
sub ExitNA (;$) {
    my $msg = shift;
    if (!defined($msg)) {
        $msg = "";
    }
    
    logresult ('NA', "NOT APPLICABLE: $msg");
    #die (na_result($msg));
}

## @name ::ExitError
## @summary Issues a FATAL error message, and reports burts
## @description 
## If no -burt option parameter is passed, the current list of burts defined
## by previous calls to burt_begin are all considered as 'burts under test'.
## If a specific burt value is passed, then only that burt is reported as
## still showing symptoms.
## If no burts are identified in either way, then the error is reported as
## a possible new problem. 
## The text for these burt tracking messages is parsed later by lib/nuffo.thpl.
##
## The message output includes call stack information to quickly edit
## the test script in case the script is out of date, or there is a need to insert a burt number.
## @arg '-burt' $burt Optional: a burt number if this call indicates that burt failed.
## @arg $message Required: A descriptive message about the error
sub ExitError (@) {
    my @burts_under_test = ();
    
    my $option = shift;
    if ($option eq '-burt') {
        # Use the next argument as the burt number, and ignore other BURTS_UNDER_TEST
        my $this_burt = shift;
        if (!defined($this_burt) ) {
            ExitScriptError("A burt must be specified after the '-burt' argument");
        }
        @burts_under_test = ($this_burt);
    } else {
        unshift(@_,$option);
        @burts_under_test = @BURTS_UNDER_TEST;
    }
    
    my @msg = @_;
    
    if (@msg == 0) {
        ExitScriptError("A message parameter must be specified");
    }
    
    my $burted=0;
    my $deferred = 1; # assume the correction to this error has been deferred
    if (0 == @burts_under_test) {
        $deferred = 0; # The correction to this error has NOT been deferred  
    } else {
        foreach my $aburt (@burts_under_test) {
            $burted=1;
            # Make sure all instances are reported
            # This allows the lack of a 'burt' message to indicate
            # that burt tracking is not supported by a test script
            push (@msg, burt_detected_text.$aburt);
            if (0==grep(/^$aburt$/,@BURTS_FAILED_REPORTED) ) {
                push( @BURTS_FAILED_REPORTED, $aburt);
            
                if (0==grep(/^$aburt$/,@BURTS_FAILED) ) {
                    push( @BURTS_FAILED, $aburt);
                }
                if (!defined($burt_status{$aburt})) {
                    $burt_status{$aburt} = burt_state($aburt);
                }
            }
            if ($burt_status{$aburt} ne 'deferred') {
                $deferred = 0;
            }
        }
    }
    
    if (!$burted) {
        push (@msg, burt_needed_text);
    }
    
    # Begin the call stack trace with the call to this function
    CallStack(\@msg,0);
    
    if ($deferred) {
        # Although this error must stop the test, the test has not "FAILED" 
        # since the fix for this error is deferred to a future release
        logresult ('INFO',stacktrace => 0, abort => 1, [@msg]);
    } else {
        logresult ('FATAL',stacktrace => 0,[@msg]);
        #die (fatal_result([@msg]));
    }
}

## @name ::ExitSoftError
## @summary Issues an error message, and reports burts
## @description 
## If no -burt option parameter is passed, the current list of burts defined
## by previous calls to burt_begin are all considered as 'burts under test'.
## If a specific burt value is passed, then only that burt is reported as
## still showing symptoms.
## If no burts are identified in either way, then the error is reported as
## a possible new problem. 
## The text for these burt tracking messages is parsed later by lib/nuffo.thpl.
##
## Each related burt is checked for its current burt status.
## If the burt is deferred to a future release, the error is reported
## with an informational level message, but the FAIL status is ignored.
## The "deferred" status is returned to the called as a true/false 1/0 value.
##
## The message output includes call stack information to quickly edit
## the test script in case the script is out of date, or there is a need to insert a burt number.
## @arg '-burt' $burt Optional: a burt number if this call indicates that burt failed.
## @arg $message Required: A descriptive message about the error
sub ExitSoftError (@) {
    my @burts_under_test = ();
    
    my $option = shift;
    if ($option eq '-burt') {
        # Use the next argument as the burt number, and ignore other BURTS_UNDER_TEST
        my $this_burt = shift;
        if (!defined($this_burt) ) {
            ExitScriptError("A burt must be specified after the '-burt' argument");
        }
        @burts_under_test = ($this_burt);
    } else {
        unshift(@_,$option);
        @burts_under_test = @BURTS_UNDER_TEST;
    }
    
    my @msg = @_;
    
    if (@msg == 0) {
        ExitScriptError("A message parameter must be specified");
    }
    
    
    $soft_error_count++; # Count soft errors so MAX_SOFTERROR can be enforced
    
    my $burted=0;
    my $aburt;
    my $deferred = 1; # assume the correction to this error has been deferred
    if (0 == @burts_under_test) {
        $deferred = 0; # The correction to this error has NOT been deferred  
    } else {
        foreach $aburt (@burts_under_test) {
            $burted=1;
            # Make sure all instances are reported to the log
            # This allows the lack of a 'burt' message to indicate
            # that burt tracking is not supported by a test script
            push (@msg, burt_detected_text.$aburt);
            if (0==grep(/^$aburt$/,@BURTS_FAILED_REPORTED) ) {
                push( @BURTS_FAILED_REPORTED, $aburt);
                
                if (0==grep(/^$aburt$/,@BURTS_FAILED) ) {
                    push( @BURTS_FAILED, $aburt);
                }
                if (!defined($burt_status{$aburt})) {
                    $burt_status{$aburt} = burt_state($aburt);
                }
            }
            if ($burt_status{$aburt} ne 'deferred') {
                $deferred = 0;
            }
        }
    }
    if (!$burted) {
        push (@msg, burt_needed_text);
    }
    
    # Begin the call stack trace with the call to this function
    CallStack(\@msg,0);
    my $abort = 0;
    param('MAX_SOFTERROR','-default','');
    if ($main::MAX_SOFTERROR ne '') {
        if ($main::MAX_SOFTERROR !~ m/^\d+$/) {
            logresult ('WARN',stacktrace => 0, "Invalid MAX_SOFTERROR value: $main::MAX_SOFTERROR, is ignored. It must be a positive integer.");
            param('MAX_SOFTERROR','-set','')
        } else {
            if ($soft_error_count > $main::MAX_SOFTERROR) {
                # We have exceeded the limit of soft errors
                $abort = 1;
                push( @msg, "MAX_SOFTERROR=$main::MAX_SOFTERROR was exceeded, aborting test execution");
            }
        }
    }
    
    if ($deferred) {
        # The test has not "FAILED" 
        # since the fix for this error is deferred to a future release
        logresult ('INFO',stacktrace => 0, abort => $abort, [@msg]);
    } else {
        logresult ('FAIL',stacktrace => 0, abort => $abort, [@msg]);
    }
    return $deferred;
}

## @name ::ExitScriptError
## @summary Exits the test script with a return code indicating a SCRIPT error.
## @description
## This function is called by functions that check their own input parameters
## for expected format and content. The existence of messages from this function
## indicates that modifications are required to the test script. 
## These errors are equivalent to run-time syntax errors.
##
## The message output includes call stack information to quickly edit
## the test script source to correct the problem.
##
## @arg $message Required: A descriptive message about the script error
sub ExitScriptError (@) {
    my @msg = @_;
    
    # Begin the call stack trace with the call to this function
    # Dropping this entry, left people in the dark in some cases
    CallStack(\@msg,0);
    
    logresult ('SCRIPT',stacktrace => 0, [@msg]);
    #die (script_result([@msg]));
}

## @name ::ExitConfigError
## @summary Exits the test script with a return code indicating a Configuration Error
## @arg $message Required: A descriptive message about the configuration error
sub ExitConfigError ($) {
    my @msg = @_;
    
    # Begin the call stack trace with the call to this function
    # Limiting this entry to the caller made it difficult to debug test scripts
    CallStack(\@msg,0);
    
    logresult ('CONFIG',stacktrace => 0, [@msg]);
    #die (config_result([@msg]));
}

## @name ::GetSoftErrorCount
## @summary Returns the currnet number of soft errors that have been seen.
sub GetSoftErrorCount () {
    return $soft_error_count;
}

## @name ::testcase_configs
## @summary Finds the compatible TESTPLAN_CONFIGS entry 
## @description
## If both the TESTPLAN_CONFIGS and TESTBED_CONFIGS parameters have non-empty values,
## its entries are compared against the TESTBED_CONFIGS values.
## If no entries match the test case is skipped (considered as Not Applicable)
sub testcase_configs  (;$$) {
    my $testbed_configs = shift;
    my $testplan_configs = shift;

    if (!defined($testbed_configs)) {
        # TESTBED_CONFIGS holds a list of configuration types
        $testbed_configs = param('TESTBED_CONFIGS','-default','');
    }
    if (!defined($testplan_configs)) {
        $testplan_configs = param('TESTPLAN_CONFIGS','-default','');
    }
    my $test_config = '';
    if ($testplan_configs ne '' && $testbed_configs ne '') {
        if ($testplan_configs =~ m/^ANY$/i) {
            # All tests are run once, anywhere
            $test_config = 'any';
        } else {
            # TESTPLAN_CONFIGS holds a list of configuration type combinations
            my @testplan_combo = split(",",$testplan_configs);

            my @testbed_configs = split(",",$testbed_configs);

            # Pick up configuration types implied from the test bed specification
            @testbed_configs = testbed_config_types(@testbed_configs);

            #print("test bed configs=".join(' ',@testbed_configs)."\n");#DEBUG

            # Look for duplicate TESTPLAN_CONFIGS entries
            # Make sure all specified configuration types are valid
            # Determine that the current test bed has at least one of these combinations
            my %testplan_config_entry;
            foreach my $this_testplan_combo (@testplan_combo) {
                if (defined($testplan_config_entry{"$this_testplan_combo"})) {
                    if ($testplan_config_entry{"$this_testplan_combo"} == 1) {
                        logresult('WARN',"TESTPLAN_CONFIGS has more than one instance of: $this_testplan_combo");
                    } else {
                        $testplan_config_entry{"$this_testplan_combo"}++;
                    }
                    next;
                } else {
                    $testplan_config_entry{"$this_testplan_combo"} = 1;
                }
                # See if every configuration type in this combination is included
                # in the current test bed
                my @testplan_types = split("/|_",$this_testplan_combo);
                my $testplan_type;
                #print("test plan types=".join(' ',@testplan_types)."\n");#DEBUG
                my @invalid_types;
                foreach $testplan_type (@testplan_types) {
                    if (!testbed_config_type_check($testplan_type)) {
                        push(@invalid_types,$testplan_type);
                    }
                }
                if (@invalid_types) {
                    ExitScriptError("An entry in TESTPLAN_CONFIGS has the following invalid configuration types: ".join(',',@invalid_types));
                } else {
                    my $has_all = 1;
                    foreach $testplan_type (@testplan_types) {
                        my $got_testplan_type = 0;
                        foreach my $this_config (@testbed_configs) {
                            if ($this_config eq $testplan_type) {
                                #print("Got match on '$this_config'\n");#DEBUG
                                $got_testplan_type = 1;
                                last;
                            }
                        }
                        if ($got_testplan_type == 0) {
                            #print("Found no test bed configuration match on '$testplan_type'\n");#DEBUG
                            $has_all = 0;
                            last;
                        }
                    }
                    if ($has_all) {
                        $test_config = $this_testplan_combo;
                        last;
                    }
                    if ($test_config ne '') {
                        last;
                    }
                }
            }
            if ($test_config eq '') {
                ExitNA ("Test bed configuration does not include any matching entries from TESTPLAN_CONFIGS=".$testplan_configs);
            } else {
                #print("Test configuration is $test_config\n");#DEBUG
            }
        }
    }
    # Return the matching test plan configuration
    if (wantarray) {
        my @return_configs = split("[_/]",$test_config);
        return @return_configs;
    } else {
        return $test_config;
    }
}

# Start Testcase:

## @name ::testcase_start
## @summary 
## Writes a INFO Result to identify the internal test case that is being started.
## This Result is used by nuffo.thpl to separate test case log records.
## @arg $testcase_name Required: A unigue id or name for the new test case
## Two forms of this parameter are accepted:
##    'preferred_single_word_testcase_name'
##    'overridden/path/testscript.thpl optional_testcase_name'
## The testscript name is detected by the presence of a slash in the text.
## Multiple word test case names are allowed, but will be converted
## by other programs such as nuffo.thpl that replaces whitespace with underscores.
## When no testscript name is passed, this function provides the current
## testscript name in the generated INFO result.
## @arg $testcase_description Optional: A short description of the test case
## @arg hosts => $testcase_host_ref Optional: an array reference to a list of host names
sub testcase_start ($;$%) {
    
    my ($testcase_name,$testcase_description,%opts) = validate_mixed(
        @_,
        [
            { type => SCALAR },
            { type => SCALAR | UNDEF, optional => 1 }
        ],
        {
            hosts => { type => ARRAYREF, optional => 1 }
        }
    );
    my $testcase_hosts_ref = $opts{hosts};
    
    my $testcase_can_start = 1;

    $testcase_name =~ s/^\s+//;
    $testcase_name =~ s/\s+$//;
    my $testcase_full_name;
    if ($testcase_name !~ m/\s/ || $testcase_name !~ m/^\S+\/\S+\s+/) {
        # No test script file name was included 
        # as the first word of the testcase name parameter
        my $testscript_name = testscript_get();
        if ($testscript_name) {
            $testcase_full_name = "$testscript_name $testcase_name";
        } else {
            $testcase_full_name = $testcase_name;
        }
    }
    my $start_message = testcase_start_text."$testcase_full_name";
    if (defined($testcase_description) and $testcase_description !~ m/^\s*$/) {
        $start_message .= " ".testcase_start_description."$testcase_description";
    }
    if (defined($testcase_hosts_ref)) {
        my $testcase_configs = '';
        my $testplan_configs = param('TESTPLAN_CONFIGS',-default,undef);
        if ($testplan_configs) {
            foreach my $hostname (@$testcase_hosts_ref) {
                my $h = host($hostname);
                my %fields = $h->param_fields();
                if ($fields{CONFIGS}) {
                    $testcase_configs .= ','.$fields{CONFIGS};
                }
            }
            eval {
                # The next function call may terminate the test as
                # NA if there is no compatible TESTPLAN_CONFIGS entry
                # SCRIPT ERROR if an undefined configtype was specified
                my $testcase_config = testcase_configs($testcase_configs,$testplan_configs);
                $start_message .= " Overrides: CONFIG_TYPE=$testcase_config";
            }; 
            if ($@) {
                if ($@ =~ m/NOT APPLICABLE: (.*)$/) {
                    my $message = $1;
                    # Return indicator that test case is not applicable
                    logcomment("Not applicable internal test case: $testcase_name on hosts: ".join(",",@$testcase_hosts_ref)." $message");
                    $testcase_can_start = 0;
                } else {
                    die($@);
                }
            }
        }
    }
    if ($testcase_can_start) {
        # Add an INFO result that can be seen without parsing the log files
        logresult('INFO',$start_message);
        if (defined(&Tharn::testcase_begin)) {
            if ($TESTCASE_HANDLE) {
                &Tharn::testcase_end(handle=>$TESTCASE_HANDLE);
            }
            $TESTCASE_HANDLE = Tharn::testcase_begin(name => $testcase_name);
        }
    }
    return $testcase_can_start;
}

## @name ::testcase_subtest_end
## @summary 
## Writes sets the Result for an internal test case subtest that has finished.
## This Result is used by nuffo.thpl to separate test case log records.
## and to record the log files for this subtest
## @arg $testcase_subtest_obj Required: NATE subtest object
## @arg $exit_on_errors Optional: comma separated list of CONFIG,SCRIPT,FAIL,FATAL,PARAM,NA
##      values that define each result type that is supposed to be abortive.
##      Default: "CONFIG,SCRIPT"
sub testcase_subtest_end ($;$) {
    my $testcase_subtest_obj = shift;
    my $exit_on_errors = shift;
    
    if (!defined($exit_on_errors)) {
        $exit_on_errors = 'CONFIG,SCRIPT';
    }
    my @own_results = $testcase_subtest_obj->own_results();
    my $logfile = $testcase_subtest_obj->logfile();
    my $scriptname = $testcase_subtest_obj->scriptname();
    my $test_id = $testcase_subtest_obj->test_id();
    if (ref(\$scriptname) eq 'CODE') {
        $scriptname = "*CODE*";
    } elsif (ref(\$scriptname) ne 'SCALAR') {
        $scriptname = "*".ref(\$scriptname)."*";
    }
    if (@own_results > 0) {
        my $got_a_result = 0;
        foreach my $r (@own_results) {
            my $result_type = $r->type;
            if ($exit_on_errors ne "" && $exit_on_errors =~ /^(.*,)?($result_type)(,.*)?$/) {
                if($r->type eq "PARAM") {
                    logcomment(
                        "exit_on_errors is set to: $exit_on_errors",
                        "Aborting test due to ".$r->type." error(s)"
                    );
                    logresult(
                        $r->type,         abort            => 1,
                        $r->msgs,
                    );
                } else {
                    $r->abortive(1);
                    $r->messages([
                                  "test_id: ".$r->test_id()." logfile: $logfile timestamp: ".$r->timestamp(),
                                  "TEST: $scriptname", "MSGS: ", map( "$_", $r->msgs() ),
                                  "exit_on_errors is set to: $exit_on_errors",
                                  "Aborting test due to ".$r->type." error(s)"
                                 ]
                                );
                    logresult($r);
                    $got_a_result = 1;
                }
            } else {
                if ($r->type eq "PARAM") {
                    logresult(
                        $r->type,         abort            => 0,
                        $r->msgs,
                    );
                } else {
                    $r->abortive(0);
                    $r->messages(["test_id: ".$r->test_id()." logfile: $logfile timestamp: ".$r->timestamp(),
                                  "TEST: $scriptname", "MSGS: ", map( "$_", $r->msgs() )],
                                );
                    logresult($r);
                    $got_a_result = 1;
                }
            }
        }
        if (!$got_a_result) {
            # No errors, defaults to PASS
            logresult( "PASS", abort => 0, 
                [
                 "test_id: $test_id logfile: $logfile timestamp: ".time,
                 "TEST: $scriptname", 
                 "Subtest Passed (no explicit result given)"
                ]);
        }
    } else {
	logresult( "PASS", abort => 0, 
	    [
             "test_id: $test_id logfile: $logfile timestamp: ".time,
             "TEST: $scriptname", 
                   "Subtest Passed (no explicit result given)"
            ]);
    }
}

## @name ::testscript_get
## @summary Returns the name of the active test script.
sub testscript_get () {
    my $msg_ref = shift;
    my $depth = 1; 
    
    my $testscript = '';
    my $previous_script = '';
    while (my ($package,$filename,$line) = caller($depth)) {
        #print("trace: call depth: $depth $package,$filename,$line\n");
        # Note: Tharn_Subtest.pm is to be replaced with Subtest.pm
        # Other Tharn prefixed packages to get their own names
        if ($filename =~ /Tharn_Subtest.pm|\/Subtest.pm/) {
            if ($previous_script) {
                $testscript = $previous_script;
                last;
            }
        } elsif ($filename =~ /Tharn_Cmds.pm|Tharn.pm/ and $package eq 'Subtest') {
        } elsif ($package eq 'TestObj') {
            # The top test invocation (usually ntest)
            last;
        } elsif ($filename =~ /Tharn|Connectrec.pm|Hostrec.pm|Param.pm|Rslt.pm/) {
            # Other NATE layers
        } else { 
            $previous_script = $filename;  
        }
        $depth++
    }
    return $testscript;
}

1;
