# -*- perl -*-
# $Id$

## @summary CConfig 
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

CConfig - CConfig globals

=head1 SYNOPSIS

  use CConfig

=head1 DESCRIPTION

Global name space for CConfig

=cut

# --------------------------------------------------------
# Class: CConfig
#   Rule processing routines
# --------------------------------------------------------

package CConfig;
use strict;
use warnings;

use Options;
use YAML;
use Data::Dumper;

use NACL::STask::Node;

use CConfig::List;
use CConfig::VList;
use CConfig::ListIter;
use CConfig::Attr;
use CConfig::AttrList;
use CConfig::Rule;
use CConfig::Impl;
use CConfig::Job;
use CConfig::ObjSet;
use CConfig::Grammar;
use CConfig::TXMgr;

use Carp;

use Tharn qw($Log snooze param);

our $OPTIONS=Options->new();
our %VARS=();     # var symbol table
our @RO_VARS=();  # list of vars that are "read only" ie set from the environment or via parameters
our $TXMGR;       # global transaction manager object
our $TEST_MODE=0; # flag to indicate that we don't want to call the implementation layer (for unit testing only)
our $SIMMODE=0;   # For future mode where we simulate the filer interactions
our @Nodes;       # NACL Nodes list
our %LOADED;      # hash which prevents 'eval "use $somelib"' from loading same lib multiple times
our %RES=();      # HAsh of all the resources configured keyed by their type (volume/aggregate/etc);


# NOTE: NACL::C is supposed to EVENTUALLY provide an option that would allow extra options to be ignored
#       in which case CDEF_OPTIONS will no longer be needed.
our %CDEF_OPTIONS; # hash which contains valid options for a given CDEF

# -----------------------------------------------
# Function: get
#   Get the specified variable from the symbol table
#
# Usage: 
#     CConfig::get(varname)
#
# Parameters:
#     varname=variable name to retrieve
# -----------------------------------------------
sub get {
    my $varname=shift;
    my $me=(caller(0))[3];

    die "$me: varname arg is required" unless ($varname);

    if (!defined($CConfig::VARS{$varname})) { 
        die "$me: Undefined variable $varname";
    };
    return $CConfig::VARS{$varname}->LIST->DATA;
}

# -----------------------------------------------
# Function: start_tx
#   start a transaction 
#
# Usage: 
#        CConfig::Rule::start_tx(undofile)
#
# Parameters:
#       undofile - undo file name
#
# Return: 
#   1=success, 0=failure 
# -----------------------------------------------
sub start_tx {
    my $undofile=shift;
    my $me=(caller(0))[3];

    my $txmgr=CConfig::TXMgr->new();
    $txmgr->start($undofile);
}

# -----------------------------------------------
# Function: commit 
#   commit a transaction 
#
# Usage: 
#        CConfig::Rule::commit()
#
# Return: 
# -----------------------------------------------
sub commit {
    my $txmgr=CConfig::TXMgr->new();
    $txmgr->commit();
}

# -----------------------------------------------
# Function: undo
#   undo a transaction using commands in the undo file
#
# Usage: 
#     CConfig::Rule::undo($undofile)
#
# Parameters:
#       undofile - undo file name 
# -----------------------------------------------
sub undo {
    my $undofilename=shift;
    my $me=(caller(0))[3];
    my @cmds=();

    # undo does not usually require init to be run
    if ( !@CConfig::Nodes ) {
        @CConfig::Nodes=NACL::STask::Node->find();
    }

    # Force a load of all APISets here so that we have them loaded before we
    # do any subtesting
    foreach my $node (@CConfig::Nodes) {
        $node->apiset(interface => "CLI");
        $node->apiset(interface => "ZAPI");
    }

    eval { CConfig::Subtest::wait_finish() };
    if ( $@ ) {
        $Log->error("$me - CConfig::Subtest::wait_finish() failed with $@");
    }

    open(FH,$undofilename) or croak "ERROR: $me: Cannot open undo file: $undofilename";
    @cmds= reverse <FH>;
    close(FH);

    # run the undo commands
    foreach my $cmd (@cmds) {
        $Log->trace("$me: Executing command: $cmd\n");

        if ( $cmd =~ /\(\"(CConfig\:\:Impl\:\:[a-zA-Z]+)/ ) {
            my $pkg = $1;
            if ( !defined $LOADED{$pkg} ) {
                eval "use $pkg";
                $LOADED{$pkg} = 1;
            }
        }

        if ( $cmd =~ /(NACL\:\:C\:\:[a-zA-Z]+)/ ) {
            my $pkg = $1;
            if ( !defined $LOADED{$pkg} ) {
                eval "use $pkg";
                $LOADED{$pkg} = 1;
            }
        }
            
        eval "$cmd";

        if ($@) {
            $Log->warn("$me: command failed with: $@\n");
        } else {
            $Log->trace("$me: Command Completed Successfully\n");
        }
    }
    CConfig::Subtest::wait_finish();
    return 1;
}

# -----------------------------------------------
# Function: rollback
#   rollback an in flight transaction
#
# Usage: 
#     CConfig::Rule::rollback()
#
# Parameters:
# -----------------------------------------------
sub rollback {
    my $me=(caller(0))[3];

    # wait for background jobs to complete and check results
    eval {CConfig::Subtest::wait_finish()};
    if ( $@ ) {
        $Log->error("$me - CConfig::Subtest::wait_finish() failed with $@");
    }

    $CConfig::TXMGR=CConfig::TXMgr->new();
    $CConfig::TXMGR->rollback();

    # wait for background jobs to complete and check results
    eval {CConfig::Subtest::wait_finish()};
    if ( $@ ) {
        $Log->error("$me - CConfig::Subtest::wait_finish() failed with $@");
    }
    return 1;
}

# -----------------------------------------------
# Function: resinfogen
#   Generate yaml files for each type of resource
#
# Usage: 
#     CConfig::resinfogen($resinfodir)
#
# Parameters:
#       resinfodir - resinfo dir name 
# -----------------------------------------------
sub resinfogen {
    my $resinfodir=shift;
    my $me=(caller(0))[3];
    die "$me: resinfo dir arg is required" unless ($resinfodir);

    foreach my $restype (keys(%CConfig::RES)) {
       my $resfile = "$resinfodir/$restype";
       my $fh = IO::File->new($resfile,">") or croak "ERROR $me: Can't open $resfile: $!";
       my $val=$CConfig::RES{$restype};
       print $fh YAML::Dump($val);
       $fh->close();
    }

    return 1;
}

# -----------------------------------------------
# Function: artifactgen
#   Generate artifact file
#
# Usage: 
#     CConfig::artifactgen(artifactfile)
#
# Parameters:
#       artifactfile - artifact file name 
# -----------------------------------------------
sub artifactgen {
    my $artifactfile=shift;
    my $me=(caller(0))[3];
    die "$me: artifact file name  arg is required" unless ($artifactfile);

    my %varext;
    my $fh = IO::File->new($artifactfile,">") or croak "ERROR $me: Can't open $artifactfile: $!";
    foreach my $var (keys %CConfig::VARS) {
        my $val=$CConfig::VARS{$var};
        if (ref($val) eq "CConfig::ListIter") { $varext{$var}=$val->LIST->DATA; }
    }

    print $fh YAML::Dump(\%varext);
    $fh->close();

    return 1;
}

# -----------------------------------------------
# Function: verify_config
#   Verify a config 
#
# Usage: 
#     CConfig::verify_config(verifyfile)
#
# Parameters:
#       verifyfile - verificaiton file 
# -----------------------------------------------
sub verify_config {
    my $verifyfile=shift;
    my $me=(caller(0))[3];
    die "$me: verify file name arg is required" unless ($verifyfile);

    my $fh = IO::File->new($verifyfile,"<") or croak "ERROR $me: Can't open $verifyfile: $!";
    {
        local $/; 
        my $data = <$fh>;
        print "Loaded: ".Dumper(Load($data));
        #@films_yaml = @{ Load($data) };
    }
    $fh->close();

    return 1;
}

# -----------------------------------------------
# Function: find_parameter
#   Get a variable from a NATE parameter
#
# Usage: 
#     CConfig::find_parameter(parametername, linenumber)
#
# Parameters:
#       parametername - name of the parameter 
#       linenumber - line number in ruleset 
# -----------------------------------------------
sub find_parameter {
    my $paramname=shift;
    my $line=shift;
    my $me=(caller(0))[3];
    my $orig_paramname=$paramname;
    die "$me: parameter name arg is required" unless ($paramname);
    die "$me: line number arg is required" unless ($line);
    
    # Remove parameter name prefix
    $paramname =~ s/^\@//;
    $paramname =~ s/^\$//;
    
    # Grab parameter from NATE
    my $value;

    # Started seeing where Tharn.pm would start complaining that '_' is 
    # reserved and would throw a Script error which eval can't catch
    if ( $paramname ne '_' ) {
        eval {
            $value = param($paramname);
        };
    }
    
    if (defined($value)) {
        if (!grep ( /^$orig_paramname$/, @CConfig::RO_VARS ) ) {
            # If defined in NATE, set/override the parameter value
            $CConfig::VARS{$orig_paramname}=CConfig::ListIter->new(LIST=>CConfig::List->new(split(',',$value)));
            push(@CConfig::RO_VARS,$orig_paramname);
        }
    } else {
        if (! defined($CConfig::VARS{$orig_paramname})) {
            die "Undefined variable $paramname, ruleset line=$line";
        }
    }
    
    return 1;
}

1;
