# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CConfig/AttrList.pm#1 $

## @summary CConfig::AttrList 
## @author kenneth.gates@netapp.com
## @status public
## @pod here

=head1 NAME

AttrList - Reference to a list of CConfig::Attr objects

=head1 SYNOPSIS

  use AttrList 

=head1 DESCRIPTION

This module manages a list of Attr objects

=cut

# --------------------------------------------------------
#    Class: CConfig::AttrList
#      Reference to a list of CConfig::Attr objects
# --------------------------------------------------------

package CConfig::AttrList;
use strict;
use warnings;
use Params::Validate qw(validate validate_pos OBJECT SCALAR BOOLEAN HASHREF ARRAYREF UNDEF);
use Carp;
use Data::Dumper;
use base qw(CConfig::List);

# -----------------------------------------------
# Function: new
#   constructor 
#
# Usage: 
#   CConfig::AttrList->new(data)
#
# Parameters:
#       DATA - reference to an array of CConfig::Attr objects from the WITH clause
#       SA   - reference to an array of item hashes one for each subcommand
#
# Return: 
#   blessed ref
#
# Assumptions:
# -----------------------------------------------
sub new {
    my $invocant = shift;
    my $class    = ref($invocant) || $invocant;
    my $self=$class->SUPER::new();
    $self->{'DATA'}=shift || [];
    $self->{'SA'}=shift || [];
    bless($self, $class);
}

# accessor methods
sub DATA  { my $self = shift; return $self->{DATA}; }
sub SA    { my $self = shift; return $self->{SA}; }

# -----------------------------------------------
# Function: copy
#   deep copy the AttrList
#
# Usage:
#   my $newlist=$attrlist->copy(<attribute value overrides>)
#
# Parameters:
#   attribute value overrides - overrides for attribute values (CConfig::ListIter) copy
#
# Return:
#   New copy of the attribute list.  
# -----------------------------------------------
sub copy {
    my $self=shift;
    my %args = (@_);
    # since the SA list is readonly we don't bother deep copying it
    my $copy=__PACKAGE__->new([],$self->SA);
    # deep copy each attribute (list iterator)
    foreach my $attr (@{$self->DATA()}) { push(@{$copy->{'DATA'}},$attr->copy(%args)); }
    return $copy;
}

# -----------------------------------------------
# Function: add_attr
#   add a new CConfig::Attr to the attr list 
#
# Parameters:
#   attr - attribute to add
#
# Return: none
#
# Assumptions: 
# -----------------------------------------------
sub add_attr {
	my $self=shift;
	my $new_attr=shift;
    push(@{$self->DATA},$new_attr);
}

# -----------------------------------------------
# Function: find_attr
#   find and return the value of the named attribute
#
# Parameters:
#   name - attribute name to find
#
# Return: attribute (or undef)
# -----------------------------------------------
sub find_attr {
	my $self=shift;
	my $name=shift;
    foreach my $attr (@{$self->DATA()}) { 
        next unless(defined($attr->NAME));
        if ($attr->NAME && $attr->NAME eq $name) { return $attr; }
    }
    return undef;
}

# -----------------------------------------------
# Function: remove_attr
#   find and remove the value of the named attribute
#
# Parameters:
#   name - attribute name to find
#
# Return: attribute (or undef)
# -----------------------------------------------
sub remove_attr {
	my $self=shift;
	my $name=shift;
    my $index=-1;
    my $index_to_remove=-1;
    foreach my $attr (@{$self->DATA()}) { 
        $index++;
        next unless(defined($attr->NAME));
        if ($attr->NAME && $attr->NAME eq $name) { 
            $index_to_remove=$index; 
            last;
        }
    }

    if ( $index_to_remove >= 0 ) {
        splice(@{$self->DATA()}, $index_to_remove, 1);
        return 1;
    }
    return 0;
}

# -----------------------------------------------
# Function: advance
#   call advance on each attribute value 
#
# Parameters:
#
# Return: none
#
# Assumptions: 
#   assumes each CConfig::Attr object has a CConfig::ListIter as its value
#   only advances an iterator once, if more that one attribute points to the 
#   same iterator (ie if the list contains references to the same variable 
#   more than once) we only advance the iterator the first time we encounter it
# -----------------------------------------------
sub advance {
	my $self=shift;
    my @iters;

    # for each attribute value pair advance any list iterators that appear in the attribute value
    ATTR: foreach my $attr (@{$self->DATA()}) { 
        # don't advance iterators via pointers, although it won't matter within a single AttrList if pointers
        #   span AttrLists (like when ON attributes point to WITH attributes) double incrementing would be
        #   a problem
        next ATTR if ($attr->REF);
        # check that the attribute points at a ListIter
        if (ref($attr->VALUE) ne "CConfig::ListIter") { 
            my $me=(caller(0))[3];
            croak "ERROR $me: attribute ".$attr->NAME." value must be an instance of CConfig::ListIter not ".ref($attr->VALUE); 
        }
        # see if we advanced this list iterator already
        my $f=0;
        foreach my $i (@iters) { if ($attr->VALUE eq $i) { $f=1; last; } };
        if (!$f) {
            # advance the iterator in the value
            $attr->VALUE->advance(@_);
            push(@iters,$attr->VALUE);
        } 
    }
}

# -----------------------------------------------
# Function: resolve
#   resolve attribute references
#
# Parameters:
#   referent - (optional) AttrList to resolve reference to (defaults to self)
#
# Return: none
#
# Assumptions: 
# -----------------------------------------------
sub resolve {
	my $self=shift;
    my $referent = shift || $self;

    # for each attribute if it is a reference resolve the reference by pointing its value
    #   to the value of its referent
    #   we don't care if the referent is not found since it may be a pointer to a parent
    #   reference which is resolved elsewhere 

    ATTR: foreach my $attr (@{$self->DATA()}) { 
        if ($attr->REF) {
            my $pattr=$referent->find_attr($attr->REF);
            if ($pattr && ($pattr!=$attr)) { 
                $attr->VALUE($pattr->VALUE); 
            }
            next ATTR;
        # This bit is if you have an attribute like MYATTR=VSERVER_{^vserver}_SUFFIX
        } elsif ( $attr->VALUE && $attr->VALUE->LIST && (ref($attr->VALUE->LIST) =~ /CConfig::VList/) ) {
            my $root = $attr->VALUE->LIST->{ROOT};
            my $index = 0;
            foreach my $entry (@$root) {
                # entry could be a string of a Config::Attr
                if ( ref($entry) =~ /CConfig::Attr/ ) {
                    if ( $entry->REF ) {
                        my $pattr=$referent->find_attr($entry->REF);
                        if ($pattr && $pattr!=$entry) {
                            $root->[$index] = $pattr->VALUE();
                        }
                    }
                }
                $index++;
            }
        # This bit is if you have an attribute SELECT WITH filter that references other attributes
        } elsif ( $attr->VALUE && $attr->VALUE->WITH && $attr->VALUE->WITH->DATA ) {
            foreach my $wattr (@{$attr->VALUE->WITH->DATA}) {
                my $wattr_type = ref($wattr);
                if ( $wattr->REF ) {
                    my $pattr=$referent->find_attr($wattr->REF);
                    if ($pattr && $pattr!=$wattr) { 
                        $wattr->VALUE($pattr->VALUE); 
                    }
                } elsif ( $wattr->VALUE()->LIST->{ROOT} ) {
                    my $root = $wattr->VALUE()->LIST->{ROOT};
                    my $index = 0;
                    foreach my $entry (@$root) {
                        if ( ref($entry) =~ /CConfig::Attr/ ) {
                            if ( $entry->REF ) {
                                my $pattr=$referent->find_attr($entry->REF);
                                if ($pattr && $pattr!=$entry) {
                                    $root->[$index] = $pattr->VALUE();
                                }
                            }
                        }
                        $index++;
                    }
                }
            }
        } 
    }
}

# -----------------------------------------------
# Function: to_string
#   convert the list of attributes to a string of -name=value pairs
#
# Parameters:
#
# Return: 
#   string of "-name value, " pairs 
#
# Assumptions: 
#   assumes each CConfig::Attr object has a CConfig::ListIter as its value
# -----------------------------------------------
sub to_string {
    my $self=shift;
    my $str="";
    foreach my $attr (@{$self->DATA()}) { 
        next unless(defined($attr->NAME));
        $str.="-".$attr->NAME()." ";
        if (ref($attr->VALUE()) eq "CConfig::ListIter") { $str.=join(',',@{$attr->VALUE()->get_next()}); }
        $str.=" ";
    }
    return $str; 
}

# -----------------------------------------------
# Function: to_hash
#   convert the list of attributes to a hash 
#
# Usage: 
#   to_hash([FORMAT=>find|create,NUM=><NUM>)
#
# Parameters:
#       FORMAT - find=finder format (prepend prefix), create=create format (no prefix but 
#           if prefix is > or < then modify the value to increment or decrement the
#           attribute value as appropriate)
#       NUM - if FORMAT="find" and NUM="all" then we want to generate an argument list 
#           for the show command to show all objects for all values of the command attributes
#           so for each attribtue we do a get_next(NUM=>ALL) and the resulting values are 
#           formatted into a comma separated list and added to the cli argument list
#           ie find all aggregate with nodes=@NODES becomes storage show aggregate -nodes node1,node2,node3,...
#
# Return: 
#   reference to the hash 
#
# Assumptions: 
#   assumes each CConfig::Attr object has a CConfig::ListIter as its value
# -----------------------------------------------
sub to_hash {
    my $self=shift;
    my (%args) = validate(@_, {
            FORMAT => {type=>SCALAR, optional=>1, default=>"create", regex=>qr/^(create|find)$/},
            NUM => {type=>SCALAR, optional=>1}
            }
    );

    my %hash=();
    foreach my $attr (@{$self->DATA()}) { 
        next unless(defined($attr->NAME));
        next if ( $attr->NAME eq '_default_' );
        if (ref($attr->VALUE()) eq "CConfig::ListIter") { 
            # if FIND flag is set then prepend PREFIX characters (<,>,!) to each attribute value
            if ($args{'FORMAT'}=~/find/i) {
                if ($args{'NUM'} && $args{'NUM'} eq "ALL") {
                    $hash{$attr->NAME()}=join(',',@{$attr->VALUE()->get_next(NUM=>$args{'NUM'})});
                } elsif ($attr->PREFIX()) {
                    my @attrvals;
                    foreach my $attrval (@{$attr->VALUE()->get_next()}) {
                        push(@attrvals,$attr->PREFIX().$attrval);
                    }
                    $hash{$attr->NAME()}=join(',',@attrvals);
                } else {
                    $hash{$attr->NAME()}=join(',',@{$attr->VALUE()->get_next()});
                }

                # If we have an unresolved option on a show, drop it
                if ( $hash{$attr->NAME()} =~ /CConfig::Attr/ ) {
                    delete $hash{$attr->NAME()};
                }
            } elsif ($args{'FORMAT'}=~/create/i && $attr->PREFIX()) {
                my @attrvals;
                foreach my $attrval (@{$attr->VALUE()->get_next()}) {
                    for ($attr->PREFIX()) {
                        my ($num,$suffix)=($attrval=~/(\d+)(.*)/);
                        $_ eq "<" || $_ eq "!" && do { $num--; };
                        $_ eq ">" && do { $num++; };
                        $attrval="${num}${suffix}"
                    }
                    push(@attrvals,$attrval);
                }
                $hash{$attr->NAME()}=join(',',@attrvals);
            } else {
                $hash{$attr->NAME()}=join(',',@{$attr->VALUE()->get_next()});
            }
        }
    }

    # Backwards compatability for Server params
    if ($hash{'-Timeout'}) { 
        $hash{'method-timeout'} = $hash{'-Timeout'}; 
        delete $hash{'-Timeout'}; 
    }
    if (exists $hash{'-subtest_pacing'}) { 
        $hash{'subtest_pacing'} = $hash{'-subtest_pacing'}; 
        delete $hash{'-subtest_pacing'}; 
    }
    if (exists $hash{'-apiset_must'}) { 
        $hash{'apiset_must'} = $hash{'-apiset_must'}; 
        delete $hash{'-apiset_must'}; 
    }
    if (exists $hash{'-apiset_should'}) { 
        $hash{'apiset_should'} = $hash{'-apiset_should'}; 
        delete $hash{'-apiset_should'}; 
    }
    if (exists $hash{'-Verification'}) { 
        delete $hash{'-Verification'}; 
    }

    return %hash; 
}


# -----------------------------------------------
# Function: is_empty
#   See if the attribute list is empty
#
# Parameters:
#
# Return: 
#   0 - list is empty
#   1 - list is not 
#
# Assumptions: 
# -----------------------------------------------
sub is_empty {
    my $self=shift;
    if ($self->DATA && scalar(@{$self->DATA})>0) { return 0; }
    return 1;
}

=head1 FUNCTIONS



=head2 new

    constructor 

=head3 Usage

    CConfig::AttrList->new(data)

=head3 Return

    blessed ref

=head3 Parameters:


=over 4


=item DATA

    reference to an array of CConfig::Attr objects from the WITH clause
    SA   - reference to an array of item hashes one for each subcommand
    

=back


=head2 copy

    deep copy the AttrList

=head3 Usage

    my $newlist=$attrlist->copy(<attribute value overrides>)

=head3 Return

    New copy of the attribute list.  

=head3 Parameters:


=over 4


=item overrides

    overrides for attribute values (CConfig::ListIter) copy
    

=back


=head2 add_attr

    add a new CConfig::Attr to the attr list 

=head3 Return

    none

=head3 Parameters:


=over 4


=item attr

    attribute to add
    

=back


=head2 find_attr

    find and return the value of the named attribute

=head3 Return

    attribute value (or undef)

=head3 Parameters:


=over 4


=item name

    attribute name to find
    

=back


=head2 advance

    call advance on each attribute value 

=head3 Return

    none

=head3 Assumptions

    assumes each CConfig::Attr object has a CConfig::ListIter as its value
    only advances an iterator once, if more that one attribute points to the 
    same iterator (ie if the list contains references to the same variable 
    more than once) we only advance the iterator the first time we encounter it

=head2 to_string

    convert the list of attributes to a string of -name=value pairs

=head3 Return

    string of "-name value, " pairs 

=head3 Assumptions

    assumes each CConfig::Attr object has a CConfig::ListIter as its value

=head2 to_hash

    convert the list of attributes to a hash 

=head3 Usage

    to_hash([FORMAT=>find|create)

=head3 Return

    reference to the hash 

=head3 Assumptions

    assumes each CConfig::Attr object has a CConfig::ListIter as its value

=head3 Parameters:


=over 4


=item FORMAT

    find=finder format (prepend prefix), create=create format (no prefix but 
    if prefix is > or < then modify the value to increment or decrement the
    attribute value as appropriate)
    

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut

1;

