# -*- perl -*-
# $Id$

## @summary CConfig::Calculation 
## @author frank.radish@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Calculation - performs simple calculations on lists

=head1 SYNOPSIS

  use Calculation

=head1 DESCRIPTION

Used by Grammer.pm to provide calculations on lists.

=cut

package CConfig::Calculation;
use Params::Validate qw(:all);
use Data::Dumper;
use CConfig::List;
use Carp;
use strict;
use warnings;

# -----------------------------------------------
# Function: toListIter
#   Convert a scalar into a list and return a ListIter pointing
#   to it.
#
# Usage: 
#   toListIter( operand );
#
# Parameters:
#       operand  - scalar, array, or CCL variable name
#       line     - ruleset line number
#
#  Note: In the case of a scalar, the operand will be converted into 
#        a single entry list.
#
# Return: A ListIter object which points to the array
# -----------------------------------------------
sub toListIter {
    my $operand = shift;
    my $line = shift;
    $line = 'unknown' if (!defined($line));
    
    if ( $operand =~ /^(@|$)/ ) {
        CConfig::find_parameter($operand, $line);
        $operand = $CConfig::VARS{$operand}->copy();
    } elsif ( ref($operand) =~ /array/i ) {
        $operand = CConfig::ListIter->new(LIST=>CConfig::List->new(@$operand))
    } else {
        my @elements = ( $operand );
        $operand = CConfig::ListIter->new(LIST=>CConfig::List->new(@elements))
    }
    return $operand;
}

# -----------------------------------------------
# Function: listCalc
#   Perform the given arithmetic operation on the 2 operands
#
# Usage: 
#   listCalc(op1, op2, $operator);
#
# Parameters:
#       op1      - operand 1, scalar, array, or CCL variable name
#       op2      - operand 2, scalar, array, or CCL variable name
#       operator - operator, one of (+|-|*|/|%)
#       line     - ruleset line number
#
#  Note: In the case of a scalar, the operand will be converted into 
#        a single entry list.
#
# Return: Operation, in list context given the 2 input operands and the
#         operator.
#
# ex: set @arr1 as 1,2,3,4
#     set @arr2 as 1,2,3
#     set @res = @arr1 + @arr2; # equiv to listCalc(@arr1, @arr2, "+");
#     @res will equal (2,4,6,4)
#
#     set @arr1 as 1,2,3,4
#     set @arr2 as 1
#     set @res = @arr1 - @arr2; # equiv to listCalc(@arr1, @arr2, "-");
#     @res will equal (0,2,3,4)
# -----------------------------------------------
sub listCalc {
    my $op1 = shift;
    my $op2 = shift;
    my $op  = shift;
    my $line = shift;
    my @elements;

    $op1 = CConfig::Calculation::toListIter( $op1, $line );
    $op2 = CConfig::Calculation::toListIter( $op2, $line );

    my $op1_len = $op1->LIST->LEN;
    my $op2_len = $op2->LIST->LEN;

    my $max_len = $op1_len > $op2_len ? $op1_len : $op2_len;
    
    my $i = 0;
    my $op1_pos = 0;
    my $op2_pos = 0;

    # make sure the Iterators are at the beginning.
    $op1->POS($op1_pos);
    $op2->POS($op2_pos);

    while ( $i < $max_len ) {
        my ($res,$v1,$v2) = (0, $op1->get(), $op2->get());

        my $do_calc = 1;
        if ( $op1_pos >= $op1_len ) {
            if ( $op eq '-' ) {
                $res = "-$v2";
            } else {
                $res = $v2;
            }
            $do_calc = 0;
        }

        if ( $op2_pos >= $op2_len ) {
            $res = $v1;
            $do_calc = 0;
        }

        # if one array is longer than the other, just copy the entries from
        # the longer array into the result, unchanged, else do operation.
        if ( $do_calc ) {
            eval "\$res = ($v1 $op $v2)";
        }

        push( @elements, $res );

        # advance each iterator, but do not wrap
        $op1->advance(NOWRAP=>1);
        $op2->advance(NOWRAP=>1);

        # If we have 2 arrays of different sizes, we want to have our result be
        # the size of the larger of the 2 arrays, i.e.
        # set @ar1 = 1,2,3;
        # set @ar2 = 1,2;
        # set @ar3 = @ar1+@ar2;  
        # @ar3 should = 2,4,3;

        $op1_pos = $op1->POS;
        $op2_pos = $op2->POS;
        $i = $op1_pos < $op2_pos ? $op1_pos : $op2_pos;
    }
    
    return \@elements;
}

=head1 FUNCTIONS

=head2 toListIter

    Convert a scalar into a list and return a ListIter pointing to it.

=head3 Usage

    toListIter( operand );

=head3 Parameters:

=over 4

=item operand

    required, scalar, array, or CCL variable name

=back

=head3 Return

    A ListIter object which points to the array

=head2 listCalc

    Perform the given arithmetic operation on the 2 operands

=head3 Usage

    listCalc(op1, op2, $operator);

=head3 Parameters:

=over 4

=item op1

    required, first operand, scalar, array, or CCL variable name

=item op2

    required, first operand, scalar, array, or CCL variable name

=item operator

    required, one of (+|-|*|/|%)

=back

=head3 Return

    Result of calculation in array context.
    ex: set @arr1 as 1,2,3,4
        set @arr2 as 1,2,3
        set @res = @arr1 + @arr2; # equiv to listCalc(@arr1, @arr2, "+");
        @res will equal (2,4,6,4)
   
        set @arr1 as 1,2,3,4
        set @arr2 as 1
        set @res = @arr1 - @arr2; # equiv to listCalc(@arr1, @arr2, "-");
        @res will equal (0,2,3,4)

=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut

1;

