# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CConfig/Grammar.pm#1 $

## @summary CConfig::Grammer
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Grammar - grammar definition

=head1 SYNOPSIS

  use Grammar;

=head1 DESCRIPTION

This object represents the CCL grammar

=cut

# --------------------------------------------------------
# Class: CConfig::Grammar
#   CCL Grammar 
# --------------------------------------------------------

package CConfig::Grammar;
use strict;
use warnings;
use List::Util;
use Tharn qw($Log);
use Data::Dumper;
use CConfig::Calculation;
use Params::Validate qw(validate validate_with BOOLEAN SCALAR HASHREF ARRAYREF UNDEF OBJECT);

our $trace_level = 0;

# -----------------------------------------------
# Function: grammarDef
#   return the grammar definition 
#
# Usage: 
#   grammerDef()
#
# Parameters:
#
# Return: grammar definition as a string
# -----------------------------------------------
sub grammarDef() {
    return q{

# Grammer Definition

TRACE_PARSER    : /trace_parser/i '=' INTEGER ';' {
                    $::RD_TRACE = $item{'INTEGER'};
                } 

# -------------------------------------------------------------------
# Literals
# -------------------------------------------------------------------
ATTR     : /[a-z0-9\_\-\.]+/i    # attribute name

# valid chars for Ontap passwords include ( '#', '@', '&', ' ', '.', '_', '!', '$', '%' );
# #, @, $ and % might cause us problems
#ATTRVAL  : /([a-z0-9\_]*(\%|(\%\())[sr]+[:a-z0-9]*(\)[a-z0-9\_]*)*)|([a-z0-9\_\-\/\*\.\!\#\@\&\$\%]+)/i   # attribute value
ATTRVAL  : /([a-z0-9\_]*(\%|(\%\())[sr]+[:a-z0-9]*(\)[a-z0-9\_]*)*)|([a-z0-9\_\-\/\*\.\:\!]+)/i   # attribute value

# VERY IMPORTANT that ATTRVAL is in the order it is given
# The first part of ATTRVAL is to match up with default-autogen-patterns which are in the form
# prefix%SSS | prefix%RR  (Note anywhere you see SSS it could also be some number of Rs)
# prefix%(SSS:start)  where start can be a string or numeric
# prefix%(SSS:start)suffix
# %(SSS:start)suffix

# Note: VARIABLE does not permit '-' because @var1-@var2 which is a subtraction would
#       match @var1- as a VARIABLE.  Attempted to change the regex to only allow '-'
#       when it was not at the end of the variable name, but then ran into errors in
#       the unit tests.
VARIABLE : /[\@\$][a-z0-9\_]+/i {
                    # $var and @var are no treated identically by CCL except for 
                    #   attr=$var 
                    #   which will cause $var to be expanded into a comma separated list
                    if (substr($item[1],0,1) eq '$') { $item[1]=~s/^\$/\@/; }
                    $item[1];
                }
PARAMETER_NAME  : /[a-z0-9\_]+/i
KEYSPEC  : /[a-z0-9\_\-]+/i          # hash indexes
INTEGER  : /[+-]?[0-9]+/             # integers
REAL     : /[+-]?[0-9]+[\.]?[0-9]*/  # Real numbers
NUMBER   : INTEGER | REAL | /NUM/i VARIABLE { if (!exists $CConfig::VARS{$item{'VARIABLE'}} ) {
                                                  # Empty list
                                                  $CConfig::VARS{$item{'VARIABLE'}} = CConfig::ListIter->new();
                                                  CConfig::find_parameter($item{'VARIABLE'}, $line);
                                              }
                                              $CConfig::VARS{$item{'VARIABLE'}}->LIST->LEN; } | /NUMVAL/i VAREXP { $item{'VAREXP'}->get(0); }
# Was code in Rule.pm which assumed that if you were passing in a ListIter, you really wanted the number of items in the list
NUM      : INTEGER | /all/i {"ALL"} | /max/i {"MAX"} | /NUM/i VAREXP { if (!exists $item{'VAREXP'} ) {
                                                                           $item{'VAREXP'} = CConfig::ListIter->new();
                                                                           CConfig::find_parameter($item{'VAREXP'}, $line);
                                                                       }
                                                                       $item{'VAREXP'}->LIST->LEN; } | 
                                                      /NUMVAL/i VAREXP { if (!exists $item{'VAREXP'} ) { 
                                                                             $item{'VAREXP'} = CConfig::ListIter->new();
                                                                             CConfig::find_parameter($item{'VAREXP'}, $line);
                                                                         }
                                                                         $item{'VAREXP'}->get(0); } | VAREXP | {1} 
FE_NUM   : INTEGER | /all/i {"ALL"} | /max/i {"MAX"} | /NUM/i VAREXP { if (!exists $item{'VAREXP'} ) {
                                                                           $item{'VAREXP'} = CConfig::ListIter->new();
                                                                           CConfig::find_parameter($item{'VAREXP'}, $line);
                                                                       }
                                                                       $item{'VAREXP'}->LIST->LEN; } | 
                                                       /NUMVAL/i VAREXP { if (!exists $item{'VAREXP'} ) {
                                                                              $item{'VAREXP'} = CConfig::ListIter->new();
                                                                              CConfig::find_parameter($item{'VAREXP'}, $line);
                                                                          }
                                                                          $item{'VAREXP'}->get(0); } | {'ALL'} 
COMMENT  : /#.*/               # comment
PATH     : /[a-z0-9\_\/]+/i    # directory path
EOF      : /^\Z/
# String Processing
STRING   : /\".*?\"/s 

#STRING   : /\"(.*?)\"/  { $1 } # use this to strip double quotes before sending to server
#STRINGLIT: /\"(.*)\"/  { $1 } 
#STRING   : STRINGLIT  
#         | STRINGLIT '.' STRING { "$item[0].$item[1]"; } 
#         | STRINGLIT '.' VARIABLE  { "$item[0].$item[1]"; } 

REFSTRING   :   '^' ATTRVAL

# complex attribute value - contains at least one variable
#CATTRVAL   : /[a-z0-9\_\-\/\$]+/i 
CATTRVAL    : ATTRVAL <skip:''> '{' VAREXP '}' CATTRVAL_CONT { [$item{'ATTRVAL'},$item{'VAREXP'},@{$item{'CATTRVAL_CONT'}}] }
          | '{' <skip:''> VAREXP '}' CATTRVAL_CONT { [$item{'VAREXP'},@{$item{'CATTRVAL_CONT'}}] }
          | ATTRVAL <skip:''> '{' REFSTRING '}' CATTRVAL_CONT { [$item{'ATTRVAL'},CConfig::Attr->new($item{'ATTRVAL'}.'_'.$item{'REFSTRING'}.'_REF',REF=>$item{'REFSTRING'}),@{$item{'CATTRVAL_CONT'}}] 
          }
          | '{' <skip:''> REFSTRING '}' CATTRVAL_CONT { [CConfig::Attr->new($item{'REFSTRING'}.'_REF',REF=>$item{'REFSTRING'}),@{$item{'CATTRVAL_CONT'}}]}


CATTRVAL_CONT   : ATTRVAL <skip:''> '{' VAREXP '}' CATTRVAL_CONT { [$item{'ATTRVAL'},$item{'VAREXP'},@{$item{'CATTRVAL_CONT'}}] }
              | '{' <skip:''> VAREXP '}' CATTRVAL_CONT { [$item{'VAREXP'},@{$item{'CATTRVAL_CONT'}}] }
              | ATTRVAL <skip:''> '{' REFSTRING '}' CATTRVAL_CONT { [$item{'ATTRVAL'},CConfig::Attr->new($item{'ATTRVAL'}.'_'.$item{'REFSTRING'}.'_REF',REF=>$item{'REFSTRING'}),@{$item{'CATTRVAL_CONT'}}] 
              }
              | '{' <skip:''> REFSTRING '}' CATTRVAL_CONT { [CConfig::Attr->new($item{'REFSTRING'}.'_REF',REF=>$item{'REFSTRING'}),@{$item{'CATTRVAL_CONT'}}]}
              | ATTRVAL { [$item{'ATTRVAL'}] }
              | { [""] }

# -------------------------------------------------------------------
# Keywords
# -------------------------------------------------------------------
LISTOP    : /select(?=\s)/i
ACTION    : /strict_find(?=\s)/i | /find(?=\s)/i | /create(?=\s)/i | /delete(?=\s)/i | /modify(?=\s)/i | /wait(?=\s)/
          | /initialize-ls-set(?=\s)/i | /initialize(?=\s)/i 
          | /efficiency_config(?=\s)/i | /efficiency_on(?=\s)/i | /efficiency_start(?=\s)/i 
          | /quota_modify(?=\s)/i | /quota_on(?=\s)/i | /quota_start(?=\s)/i 
          | /snapshot_autodelete(?=\s)/i | /options(?=\s)/i | /security(?=\s)/i
          | /enable(?=\s)/i | /disable(?=\s)/i 
          | /run(?=\s)/i | /mount(?=\s)/i 
          | /unmap(?=\s)/i | /map(?=\s)/i 
          | /resize(?=\s)/i | /rescan(?=\s)/i | /rename(?=\s)/i | /move(?=\s)/i
          | /accept(?=\s)/i | /reject(?=\s)/i
          | /resume(?=\s)/i | /suspend(?=\s)/i
          | /reserve(?=\s)/i
          | /add-port(?=\s)/i | /remove-port(?=\s)/i 
          | /add(?=\s)/i | /remove(?=\s)/i 
          | /bind(?=\s)/i | /unbind(?=\s)/i 
          | <error: $text *** Invalid action expecting one of strict_find find create delete modify wait 
              initialize-ls-set initialize efficiency_config efficiency_on efficiency_start quota_modify quota_on quota_start snapshot_autodelete options security run mount enable disable map unmap add resize rescan rename move accept reject resume suspend reserve add remove bind unbind add-port remove-port >
SUB_ACTION  : /initialize-ls-set/i | /initialize/i  
            | /efficiency_config/i | /efficiency_on/i | /efficiency_start/i 
            | /quota_modify/i | /quota_on/i | /quota_start/i 
            | /snapshot_autodelete/i | /modify/i | /delete/i | /mount/i | /options/i | /security/i | /enable/i | /disable/i
            | /add-port/i | /remove-port/i 
            | /unmap/i | /map/i  | /resize/i | /rescan/i | /rename/i | /accept/i | /reject/i
            | /resume/i | /suspend/i  | /add/i | /remove/i | /bind/i | /unbind/i | /move/i
            | <error: $text *** Invalid subaction expecting one of 
                initialize-ls-set initialize efficiency_config efficiency_on efficiency_start quota_modify quota_on quota_start snapshot_autodelete modify delete mount options security enable disable map unmap add resize rescan rename move accept reject resume suspend add remove bind unbind add-port remove-port>

RELOP     : '>=' | '<=' | '>' | '<' | '!'  

# -------------------------------------------------------------------
# LIST Iterator keywords
# -------------------------------------------------------------------
START_VAL : /current/i | /first/i | /last/i | /random/i
START     : /start/i '=' INTEGER  | /start/i '=' START_VAL 
#START     : /start/i '=' NUMBER  | /start/i '=' START_VAL 
DIR_VAL   : /right/i | /left/i | /random/i  | /across/i | /down/i
DIRECTION : /direction/i '=' DIR_VAL  
STEP      : /step/i '=' INTEGER   
CHUNK     : /chunk/i '=' INTEGER  
ROOT      : /root/i '='  PATH   
DEPTH     : /depth/i '=' INTEGER  
WIDTH     : /width/i '=' INTEGER  
KEY       : /key/i '=' KEYSPEC  

UNITS     : /K/i | /M/i | /G/i 

LISTLIMIT : INTEGER
            | /NUM/i     <skip:' '> VARIABLE { if ( !exists($CConfig::VARS{$item{'VARIABLE'}}) ) {
                                                   $CConfig::VARS{$item{'VARIABLE'}} = CConfig::ListIter->new();
                                                   CConfig::find_parameter($item{'VARIABLE'}, $line);
                                               }
                                               $CConfig::VARS{$item{'VARIABLE'}}->LIST->LEN; }
            | /NUMVAL/i  <skip:' '> VARIABLE { if ( !exists($CConfig::VARS{$item{'VARIABLE'}}) ) {
                                                   $CConfig::VARS{$item{'VARIABLE'}} = CConfig::ListIter->new();
                                                   CConfig::find_parameter($item{'VARIABLE'}, $line);
                                               }
                                               $CConfig::VARS{$item{'VARIABLE'}}->get(0); }

# LISTSPEC is a list of list values of the form 1,2,3... or v[1..3] or a variable or a list iterator spec
LISTSPEC  : '[' ']' {
                # Empty List
                CConfig::List->new(); 
          }
          | '[' LISTLIMIT '..' LISTLIMIT ']' {
                CConfig::VList->new(STARTI=>$item[2],ENDI=>$item[4]); 
          }
          | '[' INTEGER UNITS '..' INTEGER UNITS ']' {
                die "List start and end units must be the same" if ($item[3] ne $item[6]);
                CConfig::VList->new(STARTI=>$item[2],ENDI=>$item[5],UNITS=>$item[3]); 
          }
          | CATTRVAL <skip:''> '[' LISTLIMIT '..' LISTLIMIT ']' {
                CConfig::VList->new(ROOT=>$item[1],STARTI=>$item[4],ENDI=>$item[6]); 
          }
          | CATTRVAL {
                CConfig::VList->new(ROOT=>$item[1]); 
          }
          | ATTRVAL <skip:''> '[' LISTLIMIT '..' LISTLIMIT ']' {
                CConfig::VList->new(ROOT=>[$item[1]],STARTI=>$item[4],ENDI=>$item[6]); 
          }
          | ATTRVAL(s /,/) {
                CConfig::List->new(@{$item{'ATTRVAL(s)'}}); 
          }
          | STRING(s /,/) {
                CConfig::List->new(@{$item{'STRING(s)'}}); 
          }
          | INTEGER(s /,/) {
                CConfig::List->new(@{$item{'INTEGER(s)'}}); 
          }
       
LISTITER_ARG :  START     { {START=>$item{'START'}} }
             |  DIRECTION { {DIRECTION=>$item{'DIRECTION'}} } 
             |  STEP      { {STEP=>$item{'STEP'}} } 
             |  ROOT      { {ROOT=>$item{'ROOT'}} } 
             |  DEPTH     { {DEPTH=>$item{'DEPTH'}} } 
             |  WIDTH     { {WIDTH=>$item{'WIDTH'}} } 
             |  KEY       { {KEY=>$item{'KEY'}} } 
             |  /with/i   ATTRASSIGN(s) { {WITH=>CConfig::AttrList->new($item{'ATTRASSIGN(s)'})} }
             |  { {} } 
               

LISTITER  : '(' LISTOP NUM LISTSPEC LISTITER_ARG(s) ')' {
                my %args;
                foreach my $larg (@{$item{'LISTITER_ARG(s)'}}) { while (my ($k,$v)=each %$larg) { $args{$k}=$v; } }
                CConfig::ListIter->new(LIST=>$item{'LISTSPEC'},NUM=>$item{'NUM'},%args);
            }
            | '(' LISTOP NUM VECONCAT LISTITER_ARG(s) ')' {
                my %args;
                foreach my $larg (@{$item{'LISTITER_ARG(s)'}}) { while (my ($k,$v)=each %$larg) { $args{$k}=$v; } }
                # if START=CURRENT then load the starting position from the current list position
                if ($args{'START'} =~ /current/i) {
                    $args{'START'}=$item{'VECONCAT'}->LIST->POS;
                }
                $item{'VECONCAT'}->copy(NUM=>$item{'NUM'},%args)
            }
            | '(' LISTOP LISTSPEC LISTITER_ARG(s) ')' {
                my %args;
                foreach my $larg (@{$item{'LISTITER_ARG(s)'}}) { while (my ($k,$v)=each %$larg) { $args{$k}=$v; } }
                CConfig::ListIter->new(LIST=>$item{'LISTSPEC'},NUM=>1,%args);
            }
            | '(' LISTOP VECONCAT LISTITER_ARG(s) ')' {
                my %args;
                foreach my $larg (@{$item{'LISTITER_ARG(s)'}}) { while (my ($k,$v)=each %$larg) { $args{$k}=$v; } }
                # if START=CURRENT then load the starting position from the current list position
                if ($args{'START'} =~ /current/i) {
                    $args{'START'}=$item{'VECONCAT'}->LIST->POS;
                }
                $item{'VECONCAT'}->copy(NUM=>1,%args)
            }


# -------------------------------------------------------------------
# Object functions 
# -------------------------------------------------------------------
            

OBJECT          : /cluster(?=\s)/i | /license(?=\s)/i | /aggregate(?=\s)/i 
                | /vserver(?=\s)/i | /vserver_peer(?=\s)/i
                | /unix_user(?=\s)/i | /unix_group(?=\s)/i | /dns(?=\s)/i | /nfs(?=\s)/i | /nis_domain(?=\s)/i
                | /cifs(?=\s)/i | /cifs_share(?=\s)/i | /cifs_preferred_dc(?=\s)/i
                | /nfs(?=\s)/i | /name_mapping(?=\s)/i
                | /lun_igroup(?=\s)/i | /lun_portset(?=\s)/i | /lun(?=\s)/i
                | /failover_group(?=\s)/i | /vlan(?=\s)/i | /iscsi(?=\s)/i
                | /volume(?=\s)/i | /qtree(?=\s)/i | /quota_policy_rule(?=\s)/i | /quota_policy(?=\s)/i | /sis_policy(?=\s)/i
                | /snapshot_policy(?=\s)/i | /job_schedule(?=\s)/i | /export_policy_rule(?=\s)/i | /export_policy(?=\s)/i
                | /snapmirror(?=\s)/i | /snapshot(?=\s)/i
                | /flex_cache(?=\s)/i | /flex_clone(?=\s)/i
                | /interface(?=\s)/i | /port(?=\s)/i | /port_ifgrp(?=\s)/i | /route(?=\s)/i | /routing_group(?=\s)/i 
                | /node_command(?=\s)/i | /command(?=\s)/i 
                | /test(?=\s)/i | /qos_workload(?=\s)/i | /qos_policy_group(?=\s)/i 
                | /qos_settings_archived_workload(?=\s)/i | /qos_settings_control(?=\s)/i | /qos_settings_read_ahead(?=\s)/i
                | /av_engine(?=\s)/i | /av_access_policy(?=\s)/i 
                | /ccl_debug(?=\s)/i
                | /ip_address(?=\s)/i
                | <error: $text *** Invalid object expecting one of 
                    aggregate av_access_policy av_engine
                    ccl_debug cifs cifs_share cifs_preferred_dc cluster command
                    dns export_policy export_policy_rule failover_group flex_cache flex_clone
                    interface ip_address iscsi job_schedule license lun lun_igroup lun_portset
                    name_mapping nfs nis_domain node_command 
                    port port_ifgrp route routing_group qos_workload qos_policy_group 
                    qos_settings_archived_workload qos_settings_control qos_settings_read_ahead
                    qtree quota_policy quota_policy-rule snapshot snapshot_policy 
                    snapmirror sis_policy
                    vlan volume vserver vserver_peer unix_user unix_group
                    >

ATTRASSIGN      : ATTR '=' LISTITER {
                    CConfig::Attr->new($item{'ATTR'},VALUE=>$item{'LISTITER'});
                } 
                | ATTR '=' '^' ATTRVAL {
                    CConfig::Attr->new($item{'ATTR'},REF=>$item{'ATTRVAL'});
                } 
                | ATTR '=' LISTSPEC {
                    CConfig::Attr->new(
                        $item{'ATTR'},
                        VALUE=>CConfig::ListIter->new(LIST=>$item{'LISTSPEC'})
                    );
                } 
                | ATTR '=' VECONCAT {
                    CConfig::Attr->new( $item{'ATTR'}, VALUE=>$item{'VECONCAT'});
                }
                | ATTR RELOP LISTSPEC {
                    CConfig::Attr->new(
                        $item{'ATTR'},
                        VALUE=>CConfig::ListIter->new(LIST=>$item{'LISTSPEC'}),
                        PREFIX=>$item{'RELOP'}
                    );
                } 
                | ATTR RELOP LISTITER {
                    CConfig::Attr->new($item{'ATTR'},VALUE=>$item{'LISTITER'},PREFIX=>$item{'RELOP'});
                } 

# -------------------------------------------------------------------------
#  VAREXP - variable expression
#  RETURN: CConfig::ListIter
# -------------------------------------------------------------------------
VAREXP          : VARIABLE <skip:''> '{' KEYSPEC '}' {

                    # Check to see if parameter was overridden on the command-line or if it is undefined
                    CConfig::find_parameter($item{'VARIABLE'}, $line);

                    # need to create a new iterator here since KEY is a property of iterator
                    # copy the iterator so the variable list slice will have the same iterator properties
# FWR I added START=>'CURRENT' because I was seeing iterators get reset each time in rules
#     If anyone needs the older behavior, they would just need to add a reset to the iterator between rules
                    $CConfig::VARS{$item{'VARIABLE'}}->copy(KEY=>$item{'KEYSPEC'},START=>'CURRENT');
                }
                | VARIABLE <skip:''> '[' INTEGER '/' INTEGER ']' {
                    # Check to see if parameter was overridden on the command-line or if it is undefined
                    CConfig::find_parameter($item{'VARIABLE'}, $line);
                    # need to create a new list of only the selected elements, and wrap it in a new iterator 
                    # select the N/M'th set of elements
                    #     NODES a b c d e f g h i j k l 
                    #           0 1 2 3 4 5 6 7 8 9 0 1
                    #                               1 1  
                    #     NODES[1/4] = a,b,c (offset= 0,2)
                    #     NODES[3/4] = g,h,i (offset= 6,8)
                    my @elements;
                    my $listref=$CConfig::VARS{$item{'VARIABLE'}}->LIST;
                    my $start  =int(($item[4]-1)*$listref->LEN/$item[6]);
                    my $end    =int($item[4]*$listref->LEN/$item[6]-1);
                    for (my $i=$start; $i<=$end; $i++) { push(@elements,$listref->get(POS=>$i)); }
                    # copy the iterator so the variable list slice will have the same iterator properties
                    $CConfig::VARS{$item{'VARIABLE'}}->copy(LIST=>CConfig::List->new(@elements))
                }
                | VARIABLE <skip:''> '[' INTEGER ']' '{' KEYSPEC '}' {
                    # Check to see if parameter was overridden on the command-line or if it is undefined
                    CConfig::find_parameter($item{'VARIABLE'}, $line);
                    # need to create a new list of only the selected element, and wrap it in a new iterator 
                    my $element=$CConfig::VARS{$item{'VARIABLE'}}->LIST->get(POS=>$item{'INTEGER'});
                    # copy the iterator so the variable list slice will have the same iterator properties
                    $CConfig::VARS{$item{'VARIABLE'}}->copy(LIST=>CConfig::List->new(($element)),KEY=>$item{'KEYSPEC'})
                }
                | VARIABLE <skip:''> '[' INTEGER ']' {
                    # Check to see if parameter was overridden on the command-line or if it is undefined
                    CConfig::find_parameter($item{'VARIABLE'}, $line);
                    # need to create a new list of only the selected element, and wrap it in a new iterator 
                    my $element=$CConfig::VARS{$item{'VARIABLE'}}->LIST->get(POS=>$item{'INTEGER'});
                    # copy the iterator so the variable list slice will have the same iterator properties
                    $CConfig::VARS{$item{'VARIABLE'}}->copy(LIST=>CConfig::List->new(($element)))
                }
                | VARIABLE {
                    # Check to see if parameter was overridden on the command-line or if it is undefined
                    CConfig::find_parameter($item{'VARIABLE'}, $line);
                    $CConfig::VARS{$item{'VARIABLE'}};
                }

# -------------------------------------------------------------------------
#  VECONCAT - variable expression concatenation
#  RETURN: CConfig::ListIter 
# -------------------------------------------------------------------------
VECONCAT        : VAREXP ',' VECONCAT {
                   # concatenate variable contents to list
                   # TODO Hmmm, probably need to do this better, what if VAREXP contains a key
                   my @data=(@{$item{'VAREXP'}->LIST->DATA},@{$item{'VECONCAT'}->LIST->DATA});
                   CConfig::ListIter->new(LIST=>CConfig::List->new(@data));
                }
                | VAREXP  {
                    $item{'VAREXP'};
                } 

# -------------------------------------------------------------------------
#  WITH - with clause
#  RETURN: CConfig::AttrList
# -------------------------------------------------------------------------
WITH            : /with/i <commit> ATTRASSIGNS <commit> SUBACTIONS {
                    # save the attribute assignments and any sub action statements
                    CConfig::AttrList->new($item[3],$item[5])
                }  
                | {
                    CConfig::AttrList->new()
                }

ATTRASSIGNS     : ATTRASSIGN(s) {
                    $item{'ATTRASSIGN(s)'}
                }
                | {
                    []
                }

# -------------------------------------------------------------------------
#  SUBACTIONS - clause of statements to execute on the object after creation
#  RETURN: List of item hash's, one for each sub action
# -------------------------------------------------------------------------
SUBACTIONS      : '{' <commit> SA_STATEMENT(s) '}' {
                    $item[3];
                }  
                | { [] }

# FWR Can I add IF_BLOCK, COMMENT and set into SA_STATEMENT?
#STATEMENT       : IF_BLOCK // | RULE // | COMMENT // | RULE COMMENT // | TRACE_PARSER // | <error: $text *** Invalid statement>

SA_STATEMENT    : SUB_ACTION <commit> WITH <commit> ';' { \%item; }
                | ACTION <commit> ACTION_BODY         {
                        $item{'ACTION_BODY'}->{'ACTION'}=$item{'ACTION'};  
                        $item{'ACTION_BODY'};
                }
                | TRACE_PARSER
                | <error: $text *** expecting valid SUB_ACTION or ACTION list >

# This is special in that default iterators are constructed with NUM=ALL 
FE_LISTITER  : '(' LISTOP FE_NUM LISTSPEC LISTITER_ARG(s) ')' {
                my %args;
                foreach my $larg (@{$item{'LISTITER_ARG(s)'}}) { while (my ($k,$v)=each %$larg) { $args{$k}=$v; } }
                CConfig::ListIter->new(LIST=>$item{'LISTSPEC'},NUM=>$item{'FE_NUM'},%args);
            }
            | '(' LISTOP FE_NUM VECONCAT LISTITER_ARG(s) ')' {
                my %args;
                foreach my $larg (@{$item{'LISTITER_ARG(s)'}}) { while (my ($k,$v)=each %$larg) { $args{$k}=$v; } }
                # if START=CURRENT then load the starting position from the current list position
                if ($args{'START'} =~ /current/i) {
                    $args{'START'}=$item{'VECONCAT'}->LIST->POS;
                }
                $item{'VECONCAT'}->copy(NUM=>$item{'FE_NUM'},%args)
            }
# -------------------------------------------------------------------------
#  This is basically ATTRASSIGN with some special processing so that you 
#  don't have to specify an attribute name, you can just say FOREACH @NODES
#  also we then have to be carefull not to consider the subsequent ON clause 
#  to be a single attribute list with no attribute NAME
# -------------------------------------------------------------------------
FE_ATTRASSIGN   : ATTR '=' FE_LISTITER {
                    CConfig::Attr->new($item{'ATTR'},VALUE=>$item{'FE_LISTITER'});
                } 
                | ATTR '=' '^' ATTRVAL {
                    CConfig::Attr->new($item{'ATTR'},REF=>$item{'ATTRVAL'});
                } 
                | ATTR '=' VECONCAT {
                    CConfig::Attr->new(
                        $item{'ATTR'},
                        VALUE=>$item{'VECONCAT'}->copy(NUM=>"ALL")
                    );
                }
                | ATTR '=' LISTSPEC {
                    CConfig::Attr->new(
                        $item{'ATTR'},
                        VALUE=>CConfig::ListIter->new(LIST=>$item{'LISTSPEC'},NUM=>"ALL")
                    );
                }
                | FE_LISTITER {
                    CConfig::Attr->new(undef,VALUE=>$item{'FE_LISTITER'});
                } 
                | VECONCAT {
                    CConfig::Attr->new(
                        undef,
                        VALUE=>$item{'VECONCAT'}->copy(NUM=>"ALL")
                    );
                }
                | LISTSPEC {
                    CConfig::Attr->new(
                        undef,
                        VALUE=>CConfig::ListIter->new(LIST=>$item{'LISTSPEC'},NUM=>"ALL")
                    );
                }


FOREACH         : /foreach/i FE_ATTRASSIGN ATTRASSIGN(s) {
                    my $attrlist=[$item{'FE_ATTRASSIGN'}];
                    foreach my $attr (@{$item{'ATTRASSIGN(s)'}}) { push(@$attrlist,$attr); }
                    CConfig::AttrList->new($attrlist);
                }  
                | /foreach/i FE_ATTRASSIGN {
                    CConfig::AttrList->new([$item{'FE_ATTRASSIGN'}]);
                }  
                | {
                    CConfig::AttrList->new()
                }

ON              : /on/i <commit> ATTRASSIGN(s) {
                    CConfig::AttrList->new($item{'ATTRASSIGN(s)'})
                }  
                | {
                    CConfig::AttrList->new()
                }

TRACE           : { 
                        #print '-'x80,"\n";
                        use Tharn qw($Log);
                        my $newline_pos = index($_[1],"\n");
                        my $first_non_whitespace = -1;

                        if ( $_[1] =~ /(\S+)/ ) {
                            # Length of pre-match ($`) is the position of match
                            $first_non_whitespace = length($`);
                        }

                        my $first_comment = index($_[1],"\#");
                        if ( $first_non_whitespace > $newline_pos ) {
                            $newline_pos = index($_[1],"\n", $first_non_whitespace );
                        }
                        my $first_line = substr($_[1],$first_non_whitespace,$newline_pos-$first_non_whitespace);

                        # $thisline appears to be relative to the parse, not to the file we are reading
                        my $myline = $thisline;
                        if ( $myline <= $CConfig::Grammar::prev_line{$CConfig::Grammar::trace_level} ) {
                            $myline += $CConfig::Grammar::prev_line{$CConfig::Grammar::trace_level};
                        }
                        $CConfig::Grammar::prev_line{$CConfig::Grammar::trace_level} = $myline;

                        if ( length($first_line) == 0 ) {
                            # Skip null lines
                            # If you don't return anything, TRACE rule will fail
                            1;
                        } elsif ( $first_non_whitespace == $first_comment ) {
                            $Log->comment("RULE TRACE ".$CConfig::Grammar::trace_level.":".$myline."-->".$first_line."<--");
                        } elsif ( $first_line =~ /^\s*if\s+/ ) {
                            $Log->comment("RULE TRACE ".$CConfig::Grammar::trace_level.":".$myline."-->".$first_line."<--");
                        } elsif ( $first_line =~ /^\s*else\s*/ ) {
                            # Do not print out else statements, else grammar will print this out
                            # If you don't return anything, TRACE rule will fail
                            1;
                        } else {
                            # find the end of the statement
                            my $end=0;
                            my $level=0;
                            my $startblock=0;
                            my $endblock=0;
                            my $len = length($_[1]);
                            LOOP: for ($end=0; $end<length($_[1]); $end++) {
                                my $c=substr($_[1],$end,1);
                                my $rest=substr($_[1],$end,$len-$end);
                                # Want to set start/end block only when we first encounter them
                                if ( $c eq '<' ) {
                                    if ( !$startblock && ($rest =~ /^<CODE>/) )   { $startblock = 1; }
                                    if (  $startblock && ($rest =~ /^<\/CODE>/) ) { $endblock = 1; }
                                }
                                # Don't want to check against $level for endblock in case CODE rule has
                                # the wrong number of closed braces, $level could be non-zero at the endblock
                                last LOOP if (($c eq ";") && ($level==0) && ($startblock==0));
                                last LOOP if (($c eq ";")                && ($endblock==1));
                                $level++ if $c eq "{";
                                $level-- if $c eq "}";
                            }
                            my $string = substr($_[1],0,$end+1);
                            my (@out, @lines);

                            if ( $startblock ) {
                                @lines = split("\n", $string);
                                my $linenum=1;
                                my $start=0;
                                foreach my $line (@lines) {
                                    if ( !$start && ($line =~ /<CODE>/) ) { 
                                        $start = 1;
                                    }
                                    if ( $start ) {
                                        # It looks like RecDescent turns "" into " "
                                        if ( ($line !~ /CODE>/) && (length($line) > 1) ) {
                                            push( @out, "NUM: ".$linenum.$line);
                                            $linenum++;
                                        } else {
                                            push( @out, $line);
                                        }
                                    } else {
                                        push( @out, $line);
                                    }
                                }
                                $string = join("\n", @out );
                            } else {
                                @lines = split("\n", $string);
                                foreach my $line (@lines) {
                                    # If line is not just all whitespace
                                    if ( $line !~ /^\s*$/  ) {
                                        push( @out, $line);
                                    }
                                }
                                $string = join("\n", @out );
                            }
                            $Log->comment("RULE TRACE ".$CConfig::Grammar::trace_level.":".$myline."-->".$string."<--");
                        }
                        #print '-'x80,"\n";
                  }


# To allow you to do calculations in a SET rule
GROUPED_OP       : '(' CALCULATION ')' { $item{'CALCULATION'} }
OPERAND          : GROUPED_OP | NUMBER | VARIABLE | /NUM/i VARIABLE { if ( !exists($CConfig::VARS{$item{'VARIABLE'}}) ) {
                                                                          $CConfig::VARS{$item{'VARIABLE'}} = CConfig::ListIter->new();
                                                                          CConfig::find_parameter($item{'VARIABLE'}, $line);
                                                                      } 
                                                                      my $listref = $CConfig::VARS{$item{'VARIABLE'}}->LIST;
                                                                      $listref->LEN; }
CALCULATION      : ADD_CALC | SUB_CALC | MUL_CALC | DIV_CALC | UNARY

UNARY            : OPERAND { $item{'OPERAND'}; }
ADD_CALC         : OPERAND '+' OPERAND { CConfig::Calculation::listCalc($item[1], $item[3], $item[2], $line); }
SUB_CALC         : OPERAND '-' OPERAND { CConfig::Calculation::listCalc($item[1], $item[3], $item[2], $line); }
MUL_CALC         : OPERAND '*' OPERAND { CConfig::Calculation::listCalc($item[1], $item[3], $item[2], $line); }
DIV_CALC         : OPERAND '/' OPERAND { CConfig::Calculation::listCalc($item[1], $item[3], $item[2], $line); }

TRACED_STATEMENT   : TRACE STATEMENT 

STATEMENT       : IF_BLOCK // | RULE // | COMMENT // | RULE COMMENT // | TRACE_PARSER // | <error: $text *** Invalid statement>

# We can't support || because logic is dependent on returning "undef" when a condition fails, causing the
# grammar match to fail, so we can't support intermediate failures that an OR would produce.
# Might be able to figure out a way to handle this, but it doesn't seem like a huge deal since
# there are work-arounds (i.e. use multiple IF blocks to set a variable to 1, then have the
# final IF block just check that variable.

LOGICAL_OPERATOR   : '==' | '>=' | '<=' | '>' | '<' | '!=' | '=~' | '!~' | 'eq' | 'ne' | 'lt' | 'gt' | 'le' | 'ge' | '&&'

REGEX              : /\/[\^\$a-z0-9\-\_]+\/[cgimsox]*/i
                   | /\// VAREXP /\/[cgimsox]*/i { $item[1].$item{VAREXP}->get_next()->[0].$item[3] }

LOGICAL_TERM       : /\(/ LOGICAL_EXPRESSION /\)/ { $item{'LOGICAL_EXPRESSION'} }
                   | VAREXP | REGEX | NUMBER | STRING

LOGICAL_EXPRESSION     : LOGICAL_TERM _LOGICAL_EXPRESSION[ $item{LOGICAL_TERM} ]

_LOGICAL_EXPRESSION     : LOGICAL_OPERATOR LOGICAL_TERM _LOGICAL_EXPRESSION[ $arg[0], $item{LOGICAL_OPERATOR}, $item{LOGICAL_TERM} ]
                        | {
                        use Tharn qw($Log);
                        use Data::Dumper;

                        my $operand1    = $arg[0];
                        my $operator    = $arg[1];
                        my $operand2    = $arg[2];

#local $Data::Dumper::Maxdepth = 3;
#print "operand1 = ".Dumper($operand1)."\n";
#print "operator = ".Dumper($operator)."\n";
#print "operand2 = ".Dumper($operand2)."\n";
                        if ( !defined $operand2 ) {
                            return undef;
                        }

                        my ($opval1, $opval2);

                        if ( ref $operand1 ) {
                            $opval1 = $operand1->get_next()->[0];
                            if ( $operator !~ />|<|==|>=|<=|!=|&&/ ) {
                                $opval1 = '"'.$opval1.'"';
                            }
                        } else {
                            $opval1 = $operand1;
                            if ( ($operator !~ />|<|==|>=|<=|!=|&&/) &&
                                 ($opval1 !~ /"/) ) {
                                $opval1 = '"'.$opval1.'"';
                            }
                        }

                        if ( ref $operand2 ) {
                            $opval2 = $operand2->get_next()->[0];

                            if ( $operator !~ /=~|!~|>|<|==|>=|<=|!=|&&/ ) {
                                $opval2 = '"'.$opval2.'"';
                            }
                        } else {
                            $opval2 = $operand2;
                            if ( ($operator !~ /=~|!~|>|<|==|>=|<=|!=|&&/) &&
                                 ($opval2 !~ /"/) ) {
                                $opval2 = '"'.$opval2.'"';
                            }
                        }
#print "opval1 = ".Dumper($opval1)."\n";
#print "operator = ".Dumper($operator)."\n";
#print "opval2 = ".Dumper($opval2)."\n";

                        my $comp = undef;
                        eval "\$comp = ($opval1 $operator $opval2)";
                        my $return = undef;
                        my $myline = $CConfig::Grammar::prev_line{$CConfig::Grammar::trace_level};
                        if ( $comp ) {
                            $Log->comment("RULE TRACE ".$CConfig::Grammar::trace_level.":".$myline."--> LOGICAL COMPARE: ( $opval1 $operator $opval2 )<--");
                            $return = 1;
                        } else {
                            $Log->comment("RULE TRACE ".$CConfig::Grammar::trace_level.":".$myline."--> LOGICAL MISCOMPARE: ( $opval1 $operator $opval2 )<--");
                        }
                        $return;
            }

IGNORE_IF_LINE    : ...!/^\s?(else|endif)/i /.+/

IGNORE_ELSE_LINE  : ...!/^\s?(endif)/i /.+/

ELSE  : /else/i

ENDIF  : /endif/i

IGNORE_ELSE_CLAUSE  : ELSE IGNORE_ELSE_LINE(s) ENDIF
                    | ENDIF

ELSE_CLAUSE  :      ELSE TRACED_STATEMENT(s) ENDIF
                    | ENDIF


# How this works...  If the LOGICAL_EXPRESSION is true, we are committed to using the first production
#                    but the else lines (if present) will be ignored.
#                    If the first production FAILS, then the second production will ignore the if
#                    block and will execute the else block (if present).
IF_BLOCK     : /if/i '(' LOGICAL_EXPRESSION ')' <commit> TRACED_STATEMENT(s) IGNORE_ELSE_CLAUSE
             | /if/i IGNORE_IF_LINE(s) ELSE_CLAUSE

# print and log are just for literals, if you want something more involved, use "run ccl_debug"
# In the future, they could be expanded in capability

# Should be almost anything...  Anything but ; (for charsets ^ is negate, not start of line)
ANYTHING  : /[^;]+/

RULE            : /print/i <commit> ANYTHING ';'            { 
                    my $print_string = $item{'ANYTHING'};
                    $print_string =~ s/\\\\n/\n/g;
                    $print_string =~ s/^"//g;
                    $print_string =~ s/"$//g;
                    $print_string =~ s/\\\\//g;
                    print $print_string;
                    1; 
                }
                | /logwarn/i <commit> ANYTHING ';'            { 
                    use Tharn qw($Log);
                    my $print_string = $item{'ANYTHING'};
                    $print_string =~ s/\\\\n/\n/g;
                    $print_string =~ s/^"//g;
                    $print_string =~ s/"$//g;
                    $print_string =~ s/\\\\//g;
                    $Log->warn($print_string);
                    1; 
                }
                | /logerror/i <commit> ANYTHING ';'            { 
                    use Tharn qw($Log);
                    my $print_string = $item{'ANYTHING'};
                    $print_string =~ s/\\\\n/\n/g;
                    $print_string =~ s/^"//g;
                    $print_string =~ s/"$//g;
                    $print_string =~ s/\\\\//g;
                    $Log->error($print_string);
                    1; 
                }
                | /logfatal/i <commit> ANYTHING ';'            { 
                    use Tharn qw($Log);
                    my $print_string = $item{'ANYTHING'};
                    $print_string =~ s/\\\\n/\n/g;
                    $print_string =~ s/^"//g;
                    $print_string =~ s/"$//g;
                    $print_string =~ s/\\\\//g;
                    die "CCL: ".$print_string;
                }
                | /log/i <commit> ANYTHING ';'            { 
                    use Tharn qw($Log);
                    my $print_string = $item{'ANYTHING'};
                    $print_string =~ s/\\\\n/\n/g;
                    $print_string =~ s/^"//g;
                    $print_string =~ s/"$//g;
                    $print_string =~ s/\\\\//g;
                    $Log->comment($print_string);
                    1; 
                }
                | /export/i <commit> PARAMETER_NAME '=' EXPORT_ARG ';' { 
                    use Tharn qw($Log logresult);
                    use Data::Dumper;
                    my $data = $item{'EXPORT_ARG'};

                    my ($var, $key);
                    if ( $data =~ /^(\@|\$)(.*)\{(.*)\}/ ) {
                        $var = $1.$2;
                        $key = $3;
                    } else {
                        $var = $data;
                    }
                    if ( !$key ) {
                        $key = "_default_";
                    }

                    if ( exists $CConfig::RO_VARS{$var} )  {
                        $data = $CConfig::RO_VARS{$var}->to_array( KEY=> $key );
                    } elsif ( exists $CConfig::VARS{$var} )  {
                        $data = $CConfig::VARS{$var}->to_array( KEY=> $key );
                    } else {
                        die "CCL: No variable named ".$var." is defined";
                    }

                    # to_array annoyingly returns an array of single item arrays
                    my @data;
                    if ( ref($data) =~ /array/i ) {
                        foreach my $elem ( @$data ) {
                            if ( ref($elem) =~ /array/i ) {
                                foreach my $sub_elem ( @$elem ) {
                                    push( @data, $sub_elem );
                                }
                            } else {
                                push( @data, $elem );
                            }
                        }
                        $data = join(',', @data);
                    }

                    #$Log->comment($item{'PARAMETER_NAME'}."=".Dumper($data));
                    logresult( "PARAM", $item{'PARAMETER_NAME'}."=".$data );
                    1; 
                }
                | /set/i    <commit> SET_BODY               { $item{'SET_BODY'}; }
                | /reset/i  <commit> VARIABLE       ';'     { 
                    $CConfig::VARS{$item{VARIABLE}}->reset();
                } 
                | /routine/i <commit> ROUTINE_DEF            { $item{'ROUTINE_DEF'}; }
                | /include/i <commit> INCLUDE_NAME ';'       { 
                        $CConfig::Grammar::trace_level++;
                        $CConfig::Grammar::prev_line{$CConfig::Grammar::trace_level} = 0;
                        CConfig::Rule::process_file($item{'INCLUDE_NAME'});
                        $CConfig::Grammar::trace_level--;
                }
                | ACTION <commit> ACTION_BODY        { 
                        $item{'ACTION_BODY'}->{'ACTION'}=$item{'ACTION'};  
                        CConfig::Rule::_process_rule($item{'ACTION_BODY'});
                }
                | <error: $text *** expecting set, reset, routine, include, print, log, logwarn, logfatal, logerror, export or valid ACTION >
                

                 # note, if the variable was defined in the environment then it overrides the rule set
SET_BODY         : VARIABLE /as/i LISTSPEC ';' {
                    # Check to see if parameter was overridden on the command-line or if it is undefined
                    eval {CConfig::find_parameter($item{'VARIABLE'}, $line);};
                    # wrap the list in an iterator
                    $CConfig::VARS{$item{VARIABLE}}=CConfig::ListIter->new(LIST=>$item{'LISTSPEC'}) unless grep {$item{VARIABLE} eq $_} @CConfig::RO_VARS;
                    $CConfig::VARS{$item{VARIABLE}};
                 }
                 | VARIABLE /as/i LISTITER ';' {
                    $CConfig::VARS{$item{VARIABLE}}=$item{'LISTITER'} unless grep {$item{VARIABLE} eq $_} @CConfig::RO_VARS; 
                    $CConfig::VARS{$item{VARIABLE}};
                 } 
                 | VARIABLE /as/i VECONCAT ';' {
                    # copy the list iterator so that "set @x as @y" will start @x at the same position as VECONCAT when it was defined
                    $CConfig::VARS{$item{VARIABLE}}=$item{'VECONCAT'}->copy() unless grep {$item{VARIABLE} eq $_} @CConfig::RO_VARS; 
                    $CConfig::VARS{$item{VARIABLE}};
                 } 
                 | VARIABLE /as/i /NUM/i VARIABLE ';' {
                    if ( !grep {$item[1] eq $_} @CConfig::RO_VARS ) {
                        my $listref = $CConfig::VARS{$item[4]}->LIST; 
                        my $len = $listref->LEN; 
                        my @elements = ( $len );
                        $CConfig::VARS{$item[1]}=CConfig::ListIter->new(LIST=>CConfig::List->new(@elements));
                    }
                    $CConfig::VARS{$item[1]};
                 } 
                 | VARIABLE '=' CALCULATION ';' {
                    if ( !grep {$item{VARIABLE} eq $_} @CConfig::RO_VARS ) {
                        my @elements;
                        my $calc_res = $item{CALCULATION};
                        if ( $calc_res =~ /^(@|$)/ ) {
                            @elements = $CConfig::VARS{CALCULATION}->LIST->DATA;
                        } elsif ( ref($calc_res) =~ /array/i ) {
                            @elements = @{$calc_res};
                        } else {
                            @elements = ( $calc_res );
                        }

                        $CConfig::VARS{$item{VARIABLE}}=CConfig::ListIter->new(LIST=>CConfig::List->new(@elements));
                    }
                    $CConfig::VARS{$item{VARIABLE}};
                 } 
                 | VARIABLE /as/i VAREXP ';' {
                    # Allow something like set @nodes as @CLUSTERS[0]{nodes};
                    if ( !grep {$item{VARIABLE} eq $_} @CConfig::RO_VARS ) {
                        my @data=@{$item{'VAREXP'}->LIST->DATA};
                        $CConfig::VARS{$item{VARIABLE}}=CConfig::ListIter->new(LIST=>CConfig::List->new(@data));
                    }
                    $CConfig::VARS{$item{VARIABLE}};
                 } 
                 | VARIABLE /as/i /CALL_ROUTINE/i ROUTINE_NAME /\(/ CALL_ARGS /\)/ /;/ {
                    # Check to see if parameter was overridden on the command-line or if it is undefined
                    eval { CConfig::find_parameter($item{'VARIABLE'}, $line); };
                    my $find_param_rc = $@;
                    # if undefined, no big deal, will be defined after assignment
                    if ( $find_param_rc ) {
                        if ( $find_param_rc !~ /Undefined variable/ ) {
                            die $find_param_rc;
                        }
                    }
                    if ( !grep {$item{VARIABLE} eq $_} @CConfig::RO_VARS ) {
                        my $args         = $item{CALL_ARGS};
                        my $routine_name = $item{ROUTINE_NAME};

                        # when args are on multiple lines, at least get rid of \r and/or \n
                        $args =~ s/\n//g;
                        $args =~ s/\r//g;
                        $args =~ s/=>/,/g;

                        # I tried to do this with split on ',', but ran into trouble when
                        # there were imbedded blanks or commas inside of quotes
                        my $args_length = length($args);
                        my $index = 0;
                        my $stripped_args = "";
                        my $inside_quotes = 0;
                        my $prev_char_code;
                        my @args_list;
                        while ( $index < $args_length ) {
                            my $add = 0;
                            my $char = substr( $args, $index, 1 );
                            # Check to see if we have a non-escaped quote or double quote
                            # 92 is the code for forward slash, using it because forward
                            # slash is proving problematic in this grammar
                            if ( ($char =~ /\"|\'/) && ($prev_char_code != 92 ) ) {
                                $inside_quotes = ($inside_quotes == 0) ? 1 : 0;
                                $add = 1;
                            } else {
                                if ( $inside_quotes ) {
                                    $add = 1;
                                } elsif ( $char !~ /,|\s/ ) {
                                    $add = 1;
                                } elsif ( $char =~ /,/ ) {
                                    push( @args_list, $stripped_args );
                                    $stripped_args = "";
                                }
                            }

                            if ( $add ) {
                                $stripped_args .= $char;
                            }

                            # ord gives us the ascii char code
                            $prev_char_code = ord($char);
                            $index++;
                        }

                        if ( $stripped_args ) {
                            push( @args_list, $stripped_args );
                        }

                        my %args = (@args_list);

                        # Replace %args values ($CConfig::VARS keys) with $CConfig::VARS values
                        foreach my $arg (keys %args) {
                            # Not even sure that anyone would want/need to deal directly with iterators,
                            # but adding it just in case.
                            if ( $args{$arg} =~ /(.*)\:iterator/ ) {
                                my $varname = $1;
                                # Check to see if parameter was overridden on the command-line or if it is undefined
                                CConfig::find_parameter($varname, $line);
                                if ( exists $CConfig::RO_VARS{$varname} )  {
                                    $args{$arg} = $CConfig::RO_VARS{$varname};
                                } elsif ( exists $CConfig::VARS{$varname} )  {
                                    $args{$arg} = $CConfig::VARS{$varname};
                                } else {
                                    die "CCL: No variable named ".$varname." is defined";
                                }
                            } elsif ( $args{$arg} =~ /^(\@|\$)/ ) {

                                my ($var, $key);
                                if ( $args{$arg} =~ /^(\@|\$)(.*)\{(.*)\}/ ) {
                                    $var = $1.$2;
                                    $key = $3;
                                } else {
                                    $var = $args{$arg};
                                }
                                # Check to see if parameter was overridden on the command-line or if it is undefined
                                CConfig::find_parameter($var, $line);
                                if ( $key ) {
                                    if ( exists $CConfig::RO_VARS{$var} )  {
                                        $args{$arg} = $CConfig::RO_VARS{$var}->to_array( KEY=> $key );
                                    } elsif ( exists $CConfig::VARS{$var} )  {
                                        $args{$arg} = $CConfig::VARS{$var}->to_array( KEY=> $key );
                                    } else {
                                        die "CCL: No variable named ".$args{$arg}." is defined";
                                    }
                                # If we call to_array() when no key was specified, we won't pass the 
                                # whole object, but only _default_
                                } else {
                                    if ( exists $CConfig::RO_VARS{$var} )  {
                                       $args{$arg} = $CConfig::RO_VARS{$var}->LIST->DATA;
                                   } elsif ( exists $CConfig::VARS{$var} )  {
                                       $args{$arg} = $CConfig::VARS{$var}->LIST->DATA;
                                   } else {
                                       die "CCL: No variable named ".$args{$arg}." is defined";
                                   }
                                }
                            }
                            # Otherwise, assume this is a literal
                        }

                        if ( ! exists $CConfig::ROUTINES{$routine_name} ) {
                            die "CCL: No ROUTINE named $routine_name was defined";
                        }

                        # Weird, if I include OBJECT from Validate, I get an error, and I also
                        # have to include this here, not just at the top of the file
                        use Params::Validate qw(validate BOOLEAN SCALAR HASHREF ARRAYREF UNDEF);
                        my $routine = $CConfig::ROUTINES{$routine_name};
                        my @return  = $routine->( %args );

                        $CConfig::VARS{$item{VARIABLE}}=CConfig::ListIter->new(LIST=>CConfig::List->new(@return));
                    }
                    $CConfig::VARS{$item{VARIABLE}};
                 } 
                 | <error: $text *** Invalid set command>


ROUTINE_DEF      : ROUTINE_NAME /<CODE>/ PERL_BODY <commit> /<\/CODE>/ /;/ {
                    my $routine_name = $item{ROUTINE_NAME};
                    my $body         = $item{PERL_BODY};
                       $body         = join("\n", @$body);

                    # An attempt to limit what users can do in a routine.
                    if ( $body =~ /(NACL|Server)::/ ) {
                        die "CCL ROUTINE does not support NACL or Server use";
                    }
                    $CConfig::ROUTINES{$routine_name} = sub {  my @rc = eval $body;
                                                               if ( $@ ) {
                                                                   die $@;
                                                               }
                                                               return @rc;
                                                            };
                 } 
                 | <error: $text *** Invalid ROUTINE definition>


# The ':' here is to allow us to pass the actual iterator in rather than just a list
# of data
CALL_ARGS      : /[\@\$a-z0-9\_\{\}\.\:\;\=\-\+\~\>\<\~\!\%\,\'\"\s\^\&\*\#\?\/\\\]*/i
EXPORT_ARG     : /[\@\$a-z0-9\_\{\}]*/i

ROUTINE_ARGS   : /[\@\$a-z0-9\_,]+/i

# Permit : in name in case we want to use the same name at different 'scopes'
ROUTINE_NAME   : /[a-z0-9\_\:]+/i

PERL_BODY      : PERL_LINE(s) { $item{'PERL_LINE(s)'}; }

PERL_LINE      : ...!/^\s?<\/CODE>/ /.+/

INCLUDE_NAME   : /[a-z0-9\-\_\:\.\/]+/i

ACTION_BODY      : NUM OBJECT /as/i VARIABLE WITH FOREACH ON ';' { \%item; }
                 | NUM OBJECT WITH FOREACH ON ';' { \%item; }
                 | <error: $text *** Action body must be of the form [NUM] OBJECT [as VARIABLE] [WITH] [FOREACH]>

startrule       : TRACED_STATEMENT(s) EOF { 1; } 

    };
};

=head1 FUNCTIONS



=head2 grammarDef

    return the grammar definition 

=head3 Usage

    grammerDef()

=head3 Return

    grammar definition as a string

=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut

1;

