# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CConfig/Impl/Aggr.pm#1 $

## @summary CConfig::Impl::Aggr
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Aggr - Aggregate  

=head1 SYNOPSIS

  use Aggr 

=head1 DESCRIPTION

Implements object actions for Aggregates

=cut


# --------------------------------------------------------
# Class: CConfig::Impl::Aggr
#   Aggregate
# --------------------------------------------------------

package CConfig::Impl::Aggr;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log snooze);
use Params::Validate qw(:all);
use String::Random;
use NACL::C::StorageDisk;
use Carp;

use base qw(CConfig::Impl);

# Timeout if user doesn't provide us with one
our $default_timeout = 120;

# -----------------------------------------------
# Function: new
#   create new object
#
# Usage: 
#   new(item)
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Return: 
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self=$class->SUPER::new(
        KEYS=>[qw(aggregate)],
    );

    $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM   => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result=();
    my $strict=0;

    my $subtest = $self->subtest_default( num => $aargs{NUM}, object => $item->{'OBJECT'} );

    SWITCH: for ($item->{'ACTION'}) {
            /strict/i    && do { $strict=1; };
            /find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"find",NUM=>$aargs{'NUM'});

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( nodes disklist mirror-disks ) ], args => \%args );

                eval {
                    @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => \%args} );
                }; 
                if ($@) {
                    @result=();
                    die $@ if ( !$self->noEntries($@) || $strict); 
                }

                my %sargs=$item->{WITH}->to_hash(FORMAT=>"create");
                my $num=0;
                foreach my $r (CConfig::Impl::shuffle(@result)) {
                    # if we haven't picked this obj already then select it
                    if (!$self->RESULT_OBJS->contains($r)) {
                        $num++;
                        $self->RESULT_OBJS->add($r);
                        $Log->debug("$me: found ".$r->{'aggregate'});

                        # generate subactions for this item, passing in primary keys from 'show'
                        CConfig::Rule::gen_subactions($self,$item,%sargs,$self->RESULT_OBJS->getRequired($r));

                        if ( $aargs{'LIMIT'} && ($aargs{'LIMIT'} == $num ) ) {
                            last SWITCH;
                        } elsif ( $aargs{'NUM'} ne "ALL" ) {
                            last SWITCH;
                        }
                    }
                }
                last SWITCH if $aargs{'NUM'} eq "ALL" || $strict;
            };
            /create|find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                if ( $args{node} ) { delete $args{node}; }
                if (!defined($args{'aggregate'})) {

                    my $defPattern = "aggr%SSS";
                    if ($args{'-default-autogen-pattern'}) {
                        $defPattern = $args{'-default-autogen-pattern'};
                        delete $args{'-default-autogen-pattern'};
                    }

                    # if obj name not supplied then fabricate an unused one
                    if (!$self->FILER_OBJS_LOADED) {
                        eval { $self->FILER_OBJS->setRows($self->call_nacl_component_routine( routine => 'fetch')); };
                        $self->FILER_OBJS_LOADED(1);
                    }
                    do {
                        $args{'aggregate'}=$self->gen_string( pattern => $defPattern );
                    } until (!$self->RESULT_OBJS->contains(\%args) && !$self->FILER_OBJS->contains(\%args));

                    # In case someone wants to refer to the auto-generated aggregate
                    my $with = $self->resolveAttrList( key => 'aggregate', value => $args{aggregate}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                } 


                # if the aggregate (or volume for that matter) name is "whatever_NNN" filer will choke on
                # this.  "Whatever_NNN" is reserved for filer use, but "WhateverNNN" or "Whatever_NNN_flex" 
                # is ok.

                # create the object
                $self->check_keys(\%args); # check that all req keys are specified

                CConfig::Rule::gen_subactions($self,$item,%args);

                $self->CREATED_OBJS->add(\%args);
                $self->RESULT_OBJS->add(\%args);

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( nodes disklist mirror-disks ) ], args => \%args );

                # Don't want our method-timeout inherited by children
                if ( !defined($args{'method-timeout'}) ) { 
                    $args{'method-timeout'} = $default_timeout;
                }

                $self->add_undo( cmd => "delete", args => \%args, add_options => { 'force' => 'true', } );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'create', args => \%args );

                last SWITCH;
            };
            /modify/i && do {
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                if ( $args{node} ) { delete $args{node}; }
                # Not calling list2arrayrefs here because none of these are modifiable "nodes disklist mirror-disks"

                $self->add_undo( cmd => "modify", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'modify', args => \%args );
                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}

# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return: 
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;
    my $me=(caller(0))[3];
    my @result=();

    # build result set
    my $resultobjs=$self->RESULT_OBJS;
    # get filer attributes for created objects
    if ($self->CREATED_OBJS->size()) {
        eval {
            @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );
        };
        if ($@) {
            @result=();
            die $@ unless $self->noEntries($@); 
        }
        foreach my $r (@result) { 
            next unless $self->CREATED_OBJS->contains($r); # did we create this object?
            $resultobjs->update($r);
        }
    }
    return $resultobjs->DATA;
}

# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       self - object instance
#
# -----------------------------------------------
sub post {
    my $self=shift;
    my $me=(caller(0))[3];
    my @result=();

    # generate undo actions for created/modified objects
    $self->gen_undo_commands();

    # Make undo wait for background jobs to complete
    $CConfig::TXMGR->push( 'CConfig::Subtest::wait_finish()');

    # wait for background jobs to complete
    CConfig::Subtest::wait_finish();

    my $num_tries       = 0;
    my $num_noentries   = 0;
    my $num_in_cycle    = 0;
    my $print_cycle     = 50;
    my $sleep_per_cycle = 30;
    my $max_cycles      = 2880;
    my $max_sleep       = $max_cycles * $sleep_per_cycle; # 2880*30 = 24 hours
    # wait for all created aggregates to come online
    if ($self->CREATED_OBJS->size()) {
        for (my $done=0; !$done; ) {
            $done=1; # we are done when all created aggregates are online

            if ( $num_in_cycle == 0 ) {
                my $sleep_left = $max_sleep - ($num_tries * $sleep_per_cycle);
                $Log->comment("$me: Will check for aggregates to come online every $sleep_per_cycle seconds for up to ".$sleep_left." seconds");
            }

            eval {
                @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );

            };
            if ($@) {
                @result=();
                die $@ unless $self->noEntries($@); 
                $num_noentries++;
                $done=0;
            }
            foreach my $aggr (@result) {
                if ($aggr->{'state'} ne "online") { $done=0; }
            }

            if ( !$done ) {
                Tharn::sleep($sleep_per_cycle);
            }
            $num_tries++;
            $num_in_cycle++;

            if ( $num_in_cycle >= $print_cycle ) {
                $num_in_cycle = 0;
            }

            # Occasionally run into an infinite loop here.  Might be if the
            # creates are done with BEST_EFFORT and they all fail.
            if ( ($num_tries > $max_cycles) || ($num_noentries > 60) ) {
                die "$me Aggregates failed to come online within $max_sleep seconds";
            }
        }
    }

    return 1;
} 

# -----------------------------------------------
# Function: nacl_component_classname
#   Static method which returns the primary NACL 
#   component classname for this implementation
#
# Usage: 
#   my $nacl_c_classname = nacl_component_classname();
#
# -----------------------------------------------
sub nacl_component_classname {
    return "NACL::C::Aggregate";
}

# -----------------------------------------------
# Function: cdef_root
#   Return cdef_root for this implementation.
#
# Usage: 
#   my $cdef_root = cdef_root();
#
#  Note: This is used for parameter checking until 
#        NACL supports an "ignore_extra_options" flag
#        and possibly for CCL simulation.
# -----------------------------------------------
sub cdef_root {
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( component_name => undef,
                @_ );

    return "storage_aggregate";
}

# -----------------------------------------------
# Function: do_purge
#
#   Static method which overrides the generic CConfig::Impl::do_purge()
#   method which just calls the NACL Component delete.
#   This gives you an opportunity to do more than just a simple
#   delete (in this case, allows retrying the delete.
#
# Usage: 
#   do_purge( "CConfig::Impl::Aggregate", command_interface => $ci,
#             aggregate => "my_aggr" )
#
# Parameters:
#       class - class name, not directly used, but the general mechanism
#               used for this call requires this method signature.
#       command_interface - NACL::C::Node object
#
# -----------------------------------------------
sub do_purge {
    my $class = shift;
    my $me    = (caller(0))[3];
    my %opt   = ( 'command_interface' => undef,
                  @_ );
    my $timeout = 60;

    my %args;
    $args{command_interface} = $opt{command_interface};
    delete $opt{command_interface};

    if ( $opt{'method-timeout'} ) {
        $args{'method-timeout'} = $opt{'method-timeout'};
        $timeout = $args{'method-timeout'};
        delete $opt{'method-timeout'};
    }

    # Handle these cases
    # ERROR: command failed: [Job 301] Job failed: Failed to delete aggregate test_aggr0. 
    # Reason: The specified operation could not be completed as the aggregate is currently busy (dynamic grow)
    #  and
    # Reason: Invalid operation on aggregate 'aggr_base1'. You cannot run operations on an aggregate that is being created. 
    # Try again after the aggregate is online.

    my $sleep_per_cycle = 30;
    my $num_in_cycle    = 0;
    my $print_cycle     = 50;
    my $retries         = 2880;
    my $was_deleted = 0;
    while ($retries > 0) {
        $retries--;
        $num_in_cycle++;
        eval {NACL::C::Aggregate->delete( %args, %opt )};
        my $rc = $@;
        if ( $rc ) {
            if ( $rc =~ /is currently busy|Try again after the aggregate is online/i ) {
                if ( $num_in_cycle == 1 ) {
                    $Log->comment("$me Will try to delete aggregate ".$args{aggregate}." again every $sleep_per_cycle seconds");
                }
                Tharn::sleep($sleep_per_cycle);
            } else {
                die $rc;
            }
        } else {
            $was_deleted = 1;
            last;
        }

        if ( $num_in_cycle >= $print_cycle ) {
            $num_in_cycle = 0;
        }
    }

# FWR
    eval { NACL::C::StorageDisk->zerospares( command_interface => $args{command_interface} ); };

    if ( !$was_deleted ) {
        die "$me failed to delete aggregate ".$args{aggregate};
    }
}

1;
