# -*- perl -*-
# $Id$

## @summary CConfig::Impl::AvEngine
## @author frank.radish@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

AvEngine 

=head1 SYNOPSIS

  use AvEngine; 

=head1 DESCRIPTION

Implements object actions for AvEngine ("antivirus engine") objects

Supported actions: find/modify/enable/disable

=cut

package CConfig::Impl::AvEngine;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use String::Random;
use Carp;

use base qw(CConfig::Impl);

# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = find | modify | enable | disable
#           item->{'OBJECT'} = av_engine
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Return: 
#   blessed object reference
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self=$class->SUPER::new(
        KEYS=>[],
        DEFAULT=>[qw(vendor)]
    );

    $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = find | modify | enable | disable
#           item->{'OBJECT'} = av_engine
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM   => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result=();
    my $strict=0;
    my %args;

    # Can't create an av_engine, so no point in subtesting anything

    # no object to create, so strict_find and find are the same
    SWITCH: for ($item->{'ACTION'}) {
            /find/i    && do { 
                $strict=1;
                # Normally we would pass in args, but engine show tolerates no args passed in
                eval {
                    @result=$self->call_nacl_component_routine( routine => 'fetch' );
                };
                if ($@) {
                    @result=();
                    die $@ if ( !$self->noEntries($@) || $strict); 
                }
                my %sargs;
                foreach my $r (CConfig::Impl::shuffle(@result)) { 
                    # if we haven't picked this obj already then select it
                    if (!$self->RESULT_OBJS->contains($r)) { 
                        $self->RESULT_OBJS->add($r);

                        # generate subactions for this item, passing in primary keys from 'show'
                        CConfig::Rule::gen_subactions($self,$item,%sargs,$self->RESULT_OBJS->getRequired($r));

                        last SWITCH;
                    } 
                }
                last SWITCH;
            };
            /modify/i && do {
                %args=$item->{WITH}->to_hash(FORMAT=>"create");

                my $password;
                # Server supported -password, NACL::C supports "password"
                # CDEFs don't specify password or -password so we have to save
                # password before we call filter_out_invalid()
                if ( $args{'-password'} ) {
                    $password = $args{'-password'};
                } elsif ( $args{password} ) {
                    $password = $args{password};
                }

                $self->filter_out_invalid( component_name => nacl_component_classname(), cmd=> 'modify', args => \%args );

                # add password back in, it might not be valid according to the CDEFs, but it is
                # for the Component
                if ( $password ) {
                    $args{password} = $password;
                }

                # add_undo generally filters out create args which are not valid for delete, however
                # username and password are really NACL::C::VserverCifs options and not cdef options
                # so we have to tell add_undo() not to filter them out.
                $self->add_undo( cmd => "delete", args => \%args, do_not_filter_out => [qw( username password )]  );

                # Were we to let call_nacl_component_routine filter out the invalid arguments, since the 
                # CDEFS don't specify "password" it would be removed from %args
                $self->call_nacl_component_routine( routine => 'modify', filter_out_invalid => 0, args => \%args );
                last SWITCH;
            };
            /enable/i && do {
                # Enable can take a really long time (~45 minutes if AV software download/install is
                # required), user will need to specify a large -Timeout (~2200 seconds?)
                %args=$item->{WITH}->to_hash(FORMAT=>"create");
                $self->add_undo( cmd => "disable", args => \%args );
                $self->call_nacl_component_routine( routine => 'enable', args => \%args );
                last SWITCH;
            };
            /disable/i && do {
                $self->add_undo( cmd => "enable", args => \%args );
                $self->call_nacl_component_routine( routine => 'disable', args => \%args );
                last SWITCH;
            };

            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}

# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;
    my @result=();

    my $resultobjs=$self->RESULT_OBJS;

    return $resultobjs->DATA;
}

# -----------------------------------------------
# Function: post
#   Return found object
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#
# -----------------------------------------------
sub post {
    my $self=shift;
    my $me=(caller(0))[3];
    my @result=();

    # nothing is created for antivirus engine, when we
    # handle undo on modify, we might need to add code here

    return 1;
} 

# -----------------------------------------------
# Function: nacl_component_classname
#   Static method which returns the primary NACL
#   component classname for this implementation
#
# Usage:
#   my $nacl_c_classname = nacl_component_classname();
#
# -----------------------------------------------
sub nacl_component_classname {
    return "NACL::C::AntivirusEngine";
}

# -----------------------------------------------
# Function: cdef_root
#   Return cdef_root for this implementation.
#
# Usage:
#   my $cdef_root = cdef_root();
#
#  Note: This is used for parameter checking until
#        NACL supports an "ignore_extra_options" flag
#        and possibly for CCL simulation.
# -----------------------------------------------
sub cdef_root {
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( component_name => undef,
                @_ );

    return "antivirus_engine";
}



=head1 FUNCTIONS



=head2 new

    create new object

=head3 Return

    blessed object reference

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = find | modify | enable | disable
    item->{'OBJECT'} = av_engine
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = find | modify | enable | disable
    item->{'OBJECT'} = av_engine
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return found object

=head3 Usage

    $obj->post(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut

1;
