# -*- perl -*-
# $Id: $

## @summary CConfig::Impl::CclDebug
## @author radish@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

CclDebug 

=head1 SYNOPSIS

  use CclDebug;

=head1 DESCRIPTION

Implements an object that can be used to aid in ruleset
debugging or when fixing CCL code.  

This object just prints out the parameters for a given 
ruleset action. It supports any action defined
on any Impl object, and doesn't enforce checks on any action, 
so any defined action or subaction can be used.

This makes seeing what lists contain, or how iterators
work much easier.  Also, we have had requests where
people's rulesets are failing, so they aren't getting
the YAML file and they aren't sure what their lists
contain or how their iterators are performing.

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::CclDebug
#   CclDebug
# --------------------------------------------------------

package CConfig::Impl::CclDebug;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log config_result subtest);
use Params::Validate qw(:all);
use String::Random;
use Carp;

use base qw(CConfig::Impl);

# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find | run | modify
#           item->{'OBJECT'} = ccl_debug
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions:
#   object is a singleton, assumes only one rule will be calling this object at a time
# -----------------------------------------------
sub new {
    my $class = shift;
    my $self = $class->SUPER::new(KEYS => [qw(name)],);

    $self->{RESULTS} = [];

    $self;
}

# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage:
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find | run | modify
#           item->{'OBJECT'} = ccl_debug
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self = shift;
    my $item = shift;
    $Log->enter("action $item->{ACTION}");
    my $me = (caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM   => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result;
    my $strict = 0;

    $Data::Dumper::Sortkeys = 1;

    foreach my $action ($item->{'ACTION'}) {
        my %args;
        if ( $action =~ /find/i ) {
           %args  = $item->{WITH}->to_hash(FORMAT => "find", NUM=>$aargs{'NUM'});
        } else {
           %args  = $item->{WITH}->to_hash(FORMAT => "create");
        }

        # Specifically using print here because it formats the Dumper output in the
        # log viewer better
        print "\nCclDebug ".$item->{'ACTION'}." ".Dumper(\%args)."\n";

        $self->{RESULTS} = [\%args];

        if ( $action =~ /run/i ) {
            # generate subactions for this item
            CConfig::Rule::gen_subactions($self,$item,%args);
        }
    }

    $Log->exit();
    return 1;
}

# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;

    return $self->{RESULTS};
}


# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage:
#   $obj->post(<item>)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
# -----------------------------------------------
sub post {
    my $self = shift;
    $Log->enter();

    $Log->exit();
    return 1;
}

=head1 FUNCTIONS



=head2 new

    create new object

=head3 Assumptions

    object is a singleton, assumes only one rule will be calling this object at a time

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find | run | modify
    item->{'OBJECT'} = ccl_debug
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(<item>)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut

1;
