# -*- perl -*-
# $Id:  $

## @summary CConfig::Impl::Vserver
## @author gvg@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Cluster 

=head1 SYNOPSIS

  use Cluster; 

=head1 DESCRIPTION

Implements object actions for Cluster objects

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::Cluster
#   Cluster
# --------------------------------------------------------

package CConfig::Impl::Cluster;
use base qw(CConfig::Impl);
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log config_result);
use Params::Validate qw(:all);
use String::Random;
use Carp;
use NATE::BaseException qw(:try);
use NACL::C::Cluster;

# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions:
#   object is a singleton, assumes only one rule will be calling this object at a time
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self = $class->SUPER::new(KEYS => [qw()],);

    $self;
}

# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage:
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self = shift;
    my $item = shift;
    my $me   = (caller(0))[3];
    $Log->enter();

    #
    # XXX - Hack - Right now, allow only one cluster. All we could do now is,
    #       make sure all nodes given in @NODES are part of one cluster.
    #
    my (%aargs) = validate(@_, {NUM => {type => SCALAR, regex => qr/^1$/,}});

    my @result;
    my $strict = 0;
    my $find   = 0;
    my $create = 0;
    my (@clustered, @unclustered);

  SWITCH: for ($item->{'ACTION'}) {
        /strict/i && do { $strict = 1; };
        /find/i   && do { $find   = 1; };
        /create/i && do { $create = 1; };
        /create|find/i && do {

            my %args =
              ($create)
              ? $item->{WITH}->to_hash(FORMAT => "create")
              : $item->{WITH}->to_hash(FORMAT => "find", NUM => $aargs{'NUM'});
            $Log->debug("Args : " . Dumper(\%args));
            my @nodes = split(",", $args{nodes});
            @nodes = $self->NODES if (!scalar(@nodes));
            $Log->debug(
                   "Nodes picked to create/find cluster: " . join(",", @nodes));

            delete $args{nodes} if ($args{nodes});

            #
            # Go through all nodes and list them as clustered vs unclustered.
            # Based on the state they are in, make sure network interfaces
            # are set to roles expected by nodes own host records.
            #
            foreach my $node (@nodes) {

                my @clus = $self->_clustered($node)
                  if (!grep { /^$node$/ } @clustered);

                if (scalar(@clus)) {
                    push(@clustered, @clus);
                } else {
                    push @unclustered, $node;
                }
            }

            #
            # if strict_find - Report it accordingly.
            #
            if ($strict && $find && scalar(@unclustered)) {
                die config_result("strict_find found unclusterd nodes: "
                      . join(" ", @unclustered)
                      . " It may not be appropriate to proceed further. Exiting"
                      );
            }

            #
            # if 'create', but cluster exists fail as well
            #
            if ($create && scalar(@clustered)) {
                die config_result(
                           "'create' found a already clustered nodes. Exiting");
            }

            #
            # fix a clustername here, if it is not passed.
            #
            $args{clustername} =
              "cluster" . $self->{'RGEN'}->randpattern("ccccc");

            # create the object
            $self->check_keys(\%args);   # check that all req keys are specified

            # generate subactions for this item
            CConfig::Rule::gen_subactions($self, $item, %args);

            #
            # XXX - Hack - cluster delete/cleanup is not supported for now.
            #
            #$self->CREATED_OBJS->add(\%args);
            #$self->RESULT_OBJS->add(\%args);

            $self->_cluster_setup(\@clustered, \@unclustered, \%args);

            last SWITCH;
        };
        croak "$me :ERROR: Unknown action "
          . $item->{'ACTION'}
          . " for object "
          . $item->{'OBJECT'};
    }

    $Log->exit();
    return 1;
}

# -----------------------------------------------
# Function: _clustered
#   Is Clustered already?
#
# Usage:
#   internal only
#
# Parameters:
#       self - object instance
#       nodename - The nodename to be verified
#
# Returns:
#       list of nodes that belong to a cluster or none
# -----------------------------------------------
sub _clustered {
    my ($self, $nodename) = @_;
    $Log->enter("node = $nodename");
    my @clustered;

    my $APISet = CConfig::APISET($nodename);

    try {
        my $resp          = $APISet->cluster_show();
        my $parsed_output = $resp->get_parsed_output();
        foreach my $out (@$parsed_output) {
            if ($out->{node} && $out->{health} eq "true") {
                $Log->comment("$nodename is already clustered");
                push @clustered, $out->{node};
            }
        }

    }
    catch NATE::BaseException with {
        my $ex = shift;
        $Log->comment("error occured for $nodename : " . $ex->text());

        if ($ex->text() =~ /"show" is not a recognized command/) {

            #
            # Node found unclustered. Need to make sure interfaces are
            # set with correct roles as expected by host records.
            #
            $Log->comment("$nodename found unclustered.");

        } else {
            die config_result(
                "Unexpected error occurred during cluster_show(): $ex->text()");
        }
    };

    $Log->exit("@clustered");
    return @clustered;
}

# -----------------------------------------------
# Function: _prep_node
#   Prepare a node to be clustered
#
# Usage:
#   internal only
#
# Parameters:
#       self - object instance
#       nodename - The nodename to be prepared
#
# -----------------------------------------------
sub _prep_node {
    my ($self, $nodename) = @_;
    $Log->enter("node = $nodename");

    foreach my $port (
        qw(DATA_PORT DATA_PORT1 DATA_PORT2 DATA_PORT3 DATA_PORT4 CLUSTER_PORT CLUSTER_PORT1 CLUSTER_PORT2 CLUSTER_PORT3 CLUSTER_PORT4 CLUSTER_MGMT_PORT)
      ) {
        try {
            my $hostrec = CConfig::NODEOBJ($nodename)->hostrec();
            next if (!exists($hostrec->{$port}));
            my $role;
            if ($port =~ /cluster_mgmt/i) {
                $role = "cluster-mgmt";
            } elsif ($port =~ /cluster/i) {
                $role = "cluster";
            } else {
                $role = "data";
            }
            my $APISet = CConfig::APISET($nodename);
            $APISet->network_port_modify(node => $nodename,
                                         port => $hostrec->{$port},
                                         role => $role);
        }
        catch NATE::BaseException with {
            my $ex = shift;
            die config_result(
                          "network_port_modify failed on $nodename for $port : "
                            . $ex->text());

        };
    }

    $Log->exit();
    return 1;
}

# -----------------------------------------------
# Function: _cluster_setup
#   Setup Cluster
#
# Usage:
#   internal only
#
# Parameters:
#       self - object instance
#       clustered - list of nodes that are already part of a cluster
#       unclustered - list of nodes that are not clustered yet
#       args - hashref of arguments to pass to cluster create
#
# -----------------------------------------------
sub _cluster_setup {
    my ($self, $clustered, $unclustered, $args) = @_;
    $Log->enter(  "clustered : "
                . Dumper($clustered)
                . " unclustered : "
                . Dumper($unclustered)
                . " args : "
                . Dumper($args));

    my $clusterip;

    #
    # If none of the nodes are clustered, then we need to create cluster first
    #
    if (!scalar(@$clustered)) {

        #
        # Pick the node that has CLUSTER_MGMT_* configured and run cluster create on that node.
        #
        my ($hrec, $cluster_created, $index);
        my @clusterparams =
          qw(ipaddr1 netmask mtu ipaddr2 mgmt-port license clustername mgmt-ip mgmt-netmask mgmt-gateway);

        for ($index = 0 ; $index < scalar(@$unclustered) ; $index++) {
            $hrec = CConfig::NODEOBJ($unclustered->[$index])->hostrec();
            my @unfound;

            for (@clusterparams) {
                /license/ && do {
                    $args->{license} = $args->{license} || "NWUZZJPJYBFDAA";
                };
                /ipaddr1/ && do {
                    $args->{ipaddr1} = $hrec->{CLUSTER_IP}
                      || $hrec->{CLUSTER_IP1};
                    push @unfound, "CLUSTER_IP1" if (!$args->{ipaddr1});
                };
                /ipaddr2/
                  && ($hrec->{CLUSTER_IP2})
                  && do { $args->{ipaddr2} = $hrec->{CLUSTER_IP2}; };
                /netmask/ && do {
                    $args->{netmask} =
                         $hrec->{CLUSTER_MASK}
                      || $hrec->{CLUSTER_MASK1}
                      || $hrec->{CLUSTER_MASK2};
                    push @unfound, "CLUSTER_MASK" if (!$args->{netmask});
                };
                /mtu/ && do {
                    $args->{mtu} =
                         $hrec->{CLUSTER_MTU}
                      || $hrec->{CLUSTER_MTU1}
                      || $hrec->{CLUSTER_MTU2};
                    push @unfound, "CLUSTER_MTU" if (!$args->{mtu});
                };
                /mgmt-ip/ && do {
                    $args->{'mgmt-ip'} = $hrec->{CLUSTER_MGMT_IP};
                    push @unfound, "CLUSTER_MGMT_IP"
                      if (!$args->{'mgmt-ip'});
                };
                /mgmt-port/ && do {
                    $args->{'mgmt-port'} = $hrec->{CLUSTER_MGMT_PORT};
                    push @unfound, "CLUSTER_MGMT_PORT"
                      if (!$args->{'mgmt-port'});
                };
                /mgmt-netmask/ && do {
                    $args->{'mgmt-netmask'} = $hrec->{CLUSTER_MGMT_MASK};
                    push @unfound, "CLUSTER_MGMT_MASK"
                      if (!$args->{'mgmt-netmask'});
                };
                /mgmt-gateway/ && do {
                    $args->{'mgmt-gateway'} = $hrec->{CLUSTER_MGMT_GATEWAY};
                    push @unfound, "CLUSTER_MGMT_GATEWAY"
                      if (!$args->{'mgmt-gateway'});
                };
            }

            if (!scalar(@unfound)) {
                $Log->comment(
                     "Preparing node $unclustered->[$index] to create a cluster"
                     );
                $self->_prep_node($unclustered->[$index]);

                $Log->comment(  "Creating Cluster on node "
                              . $unclustered->[$index]
                              . " with "
                              . Dumper($args));
                my $cluster;
                try {
                    $cluster =
                      NACL::C::Cluster->create(
                                       'command_interface' =>
                                         CConfig::NODEOBJ($unclustered->[$index]),
                                       'method-timeout' => 300,
                                       %$args);
                }
                catch NATE::BaseException with {
                    my $ex = shift;
                    die config_result("Cluster Create Failed: " . $ex->text());
                };

                # Remove the node that has cluster created just now.
                $cluster_created = 1;
                push @$clustered, $unclustered->[$index];
                delete $unclustered->[$index];
                $clusterip = $args->{ipaddr1};

                $self->CREATED_OBJS->add({%$args, cluster => $cluster});
                $self->RESULT_OBJS->add({%$args, cluster => $cluster});
            } else {
                $Log->comment("$unclustered->[$index] do not have params : "
                    . join(" ", @unfound)
                    . " specified in Hostrec, hence not suitable to run cluster_create"
                    );
            }

        }

        if (!$cluster_created) {
            die config_result("The required params ("
                . join(" ", @clusterparams)
                . ") for cluster_create could not be found from nate_hosts files. Please update nate_hosts files appropriately and rerun the test"
                );
        }

    } else {

        #
        # We found existing cluster - pick up a clusterip so we can join other nodes
        #
        my $hrec = CConfig::NODEOBJ($clustered->[0])->hostrec();
        $clusterip =
             $hrec->{CLUSTER_IP1}
          || $hrec->{CLUSTER_IP2}
          || $hrec->{CLUSTER_IP};
        $Log->comment("Existing cluster found. Using clusterip as $clusterip");
    }

    $Log->debug("Unclustered nodes are :" . join(" ", @$unclustered));
    foreach my $nodename (@$unclustered) {
        $Log->comment("Preparing node $nodename to join the cluster");
        $self->_prep_node($nodename);

        $Log->comment("Joining $nodename to Cluster using $clusterip");
        my $hrec = CConfig::NODEOBJ($nodename)->hostrec();
        my %args = ("clusteripaddr" => $clusterip);
        my @unfound;
        my @cjparams = qw(ipaddr1 ipaddr2 mtu netmask);
        for (@cjparams) {
            /ipaddr1/ && do {
                $args{ipaddr1} = $hrec->{CLUSTER_IP}
                  || $hrec->{CLUSTER_IP1};
                push @unfound, "CLUSTER_IP1" if ($args{ipaddr1});
            };
            /ipaddr2/ && do { $args{ipaddr2} = $hrec->{CLUSTER_IP2} };
            /netmask/ && do {
                $args{netmask} =
                     $hrec->{CLUSTER_MASK}
                  || $hrec->{CLUSTER_MASK1}
                  || $hrec->{CLUSTER_MASK2};
                push @unfound, "CLUSTER_MASK" if ($args{netmask});
            };
            /mtu/ && do {
                $args{mtu} =
                     $hrec->{CLUSTER_MTU}
                  || $hrec->{CLUSTER_MTU1}
                  || $hrec->{CLUSTER_MTU2};
                push @unfound, "CLUSTER_MTU" if ($args{mtu});
            };
        }
        if (scalar(@unfound) < 3) {
            die config_result(
                           "$nodename nate_hosts is missing values for params :"
                             . join(" ", @unfound));
        }

        my $cluster;
        try {
            $cluster =
              NACL::C::Cluster->join(
                                 command_interface => CConfig::NODEOBJ($nodename),
                                 %args);
        }
        catch NATE::BaseException with {
            my $ex = shift;
            die config_result(
                           "Cluster join Failed for $nodename: " . $ex->text());
        };
        $self->CREATED_OBJS->add({%args, cluster => $cluster});
        $self->RESULT_OBJS->add({%args, cluster => $cluster});
    }

    $Log->exit();
    return;
}

# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;

    my $resultobjs=$self->RESULT_OBJS;

    return $resultobjs->DATA;
}


# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage:
#   $obj->post(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#
# -----------------------------------------------
sub post {
    my $self   = shift;
    my $me     = (caller(0))[3];
    $Log->enter();

    # log undo actions
    #
    # XXX - GV: I am not seeing the need to have undo for Cluster yet,
    # since unjoin and destroying cluster is not really fun yet.
    #

    $Log->exit();
    return 1;
}

=head1 FUNCTIONS



=head2 new

    create new object

=head3 Assumptions

    object is a singleton, assumes only one rule will be calling this object at a time

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO
    * Support for Multiple Clusters
    * Once ONTAP is ready, then we have to handle the 'undo' portions of Cluster creation and Join.

=cut

1;
