# -*- perl -*-
# $Id$

## @summary CConfig::Impl::Command
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Command 

=head1 SYNOPSIS

  use Command 

=head1 DESCRIPTION

Implements object actions for Command objects

=cut


# --------------------------------------------------------
# Class: CConfig::Impl::Command
# --------------------------------------------------------

package CConfig::Impl::Command;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use String::Random;
use Carp;
use NACL::C::Node;

use base qw(CConfig::Impl);
my @result=();

# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my $self=$class->SUPER::new();
    return $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM => {type=>SCALAR}
            }
    );

    my %args=$item->{WITH}->to_hash();

    SWITCH: for ($item->{'ACTION'}) {
            /run/i    && do { 
                die "$me: method is a required argument" unless defined($args{'method'}); 
                my $node = $self->CMD_INT;
                my $cli_apiset = $node->apiset(category => "Node", interface => "CLI", set => "CMode");
                my $method=$args{'method'}; delete $args{'method'};

                my ($best_effort, $timeout);
                if ($args{'-best-effort'}) {
                    $best_effort = $args{'-best-effort'};
                    delete $args{'-best-effort'};
                } elsif ($args{'-best_effort'}) {
                    $best_effort = $args{'-best_effort'};
                    delete $args{'-best_effort'};
                }
                if ($args{'method-timeout'}) {
                    $timeout = $args{'method-timeout'};
                    delete $args{'method-timeout'};
                }

                # get additional prompts before we filter them out
                my $prompt_answers = $self->server_compatible_prompts( full_cmd => $method, args => \%args );

                $self->filter_out_invalid( full_cmd=> $method, args => \%args );

                # Add timeout back in because filter_out_invalid() would have removed it
                if ( $timeout ) {
                    $args{'connectrec-timeout'} = $timeout;
                }

                # if command is one which supports additional server style options
                # add them in too.
                if ( $prompt_answers ) {
                    $args{'connectrec-match_table'} = $prompt_answers;
                }

                if ($best_effort) {
                    my $response=eval { $cli_apiset->$method(%args) };
                    if ($@) {
                        $Log->warn("$me: -best_effort flag is set, error will be ignored $@");
                    } else {
                        @result=@{$response->get_parsed_output()};
                    }
                } else {
                    my $response=$cli_apiset->$method(%args);
                    @result=@{$response->get_parsed_output()};
                }
                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}

# -----------------------------------------------
# Function: server_compatible_prompts
#   Return additional prompts required for certain commands.
#   NOTE: Server.pm used to do this for us, but this sort of
#         additional prompt handling is generally done at the
#         Component layer in NACL.  Here we are just providing
#         a minimal compatibility layer for certain Server
#         commands
#
# Usage:
#   $obj->server_compatible_prompts( full_cmd => "vserver_cifs_create", args => \%args );
#
# Parameters:
#       self - object instance
#
# Return:
#         Array of prompt/responses
# -----------------------------------------------
sub server_compatible_prompts {
    my $self=shift;
    my $me=(caller(0))[3];
    my %opt = ( full_cmd => undef,
                args     => undef,
                @_ );

    my $cmd  = $opt{full_cmd};
    my $args = $opt{args};

    if ( !( $cmd && $args ) ) {
        die "$me must be called with full_cmd and args options";
    }

    my $prompt_answers = [];

    # We would really like people to use CCL Impl objects directly, but
    # at a minimum, we want to try to keep support for commands which 
    # were added to CCL rulesets prior to Impl support being developed.
    # i.e. some backwards compatibility
    if ( $cmd =~ /protocols?_cifs_server_new/ ||
         $cmd =~ /vserver_cifs_(new|create|delete)/) {
        if ( !$args->{'-reuse_machine_account'} ) {
            $args->{'-reuse_machine_account'} = 'y';
        }
        my $mach_account = $args->{'-reuse_machine_account'};
        push( @$prompt_answers, "Ok to reuse this account(.+):" => sub {$mach_account} );
    } elsif ($cmd eq "antivirus_engine_enable") {
        if ( $args->{'-eula'} ) {
            my $eula = $args->{'-eula'};
            push( @$prompt_answers, '{yes|no|text}:'         => sub {$eula} );
        }
    }

    if ( $args->{'-password'} ) {
        my $password = $args->{'-password'};
        # Leaving these in mainly so I can see what all of the concrete examples are
        #push( @$prompt_answers, 'Please enter a password for user' => sub {$password} );
        #push( @$prompt_answers, 'Please enter a new password'      => sub {$password} );
        #push( @$prompt_answers, 'Please enter your password:'      => sub {$password} );
        #push( @$prompt_answers, 'Please enter password:'           => sub {$password} );
        #push( @$prompt_answers, 'Enter the password:'              => sub {$password} );
        #push( @$prompt_answers, 'Enter Password:'                  => sub {$password} );
        #push( @$prompt_answers, 'Please enter password again:'     => sub {$password} );
        #push( @$prompt_answers, 'Confirm password:'                => sub {$password} );
        #push( @$prompt_answers, 'Please enter it again:'           => sub {$password} );
        #push( @$prompt_answers, 'Enter it again:'                  => sub {$password} );
        push( @$prompt_answers, qr/(enter|confirm).*password.*:/i   => sub {$password} );
        push( @$prompt_answers, qr/enter it again:/i                => sub {$password} );
    }

    if ( $args->{'-username'} ) {
        my $username = $args->{'-username'};
        # Leaving these in mainly so I can see what all of the concrete examples are
        #push( @$prompt_answers, 'Enter User:'                    => sub {$username} );
        #push( @$prompt_answers, 'Enter the user name:'           => sub {$username} );
        #push( @$prompt_answers, 'Please enter your userid:'      => sub {$username} );

        # Negative look-ahead, i.e. don't match the string 'Please enter a password for user:'
        push( @$prompt_answers, qr/enter(?!.*password).*user.*:/i => sub {$username} );
    }

    return $prompt_answers;
} 
# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;
    return \@result;
} 

# -----------------------------------------------
# Function: post
#   Return command result
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       self - object instance
# -----------------------------------------------
sub post {
    my $self=shift;
    return 1;
} 

=head1 FUNCTIONS



=head2 new

    create new object

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes

=back


=head2 post

    Return command result

=head3 Usage

    $obj->post(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
