# -*- perl -*-
# $Id$

## @summary CConfig::Impl::FlexClone
## @author frank.radish@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

FlexClone 

=head1 SYNOPSIS

  use FlexClone 

=head1 DESCRIPTION

Implements object actions for FlexClone objects

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::FlexClone
#   FlexClone
# --------------------------------------------------------


package CConfig::Impl::FlexClone;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use String::Random;
use CConfig::Job;
use CConfig::Impl::Volume;
use Carp;

use base qw(CConfig::Impl);

# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self=$class->SUPER::new(
        KEYS=>[qw(vserver flexclone )],
        DEFAULT=>[qw(flexclone)]
    );

    $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM   => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result;
    my $strict=0;

    my $subtest = $self->subtest_default( num => $aargs{NUM}, object => $item->{'OBJECT'} );

    SWITCH: for ($item->{'ACTION'}) {
            /strict/i    && do { $strict=1; };
            /find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"find",NUM=>$aargs{'NUM'});
                # remove -forground option, its not valid on show, only on create
                if ($args{'foreground'})  { delete $args{'foreground'}; }
                eval {
                    @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => \%args} );
                };
                if ($@) {
                    @result=();
                    die $@ if ( !$self->noEntries($@) || $strict); 
                }

                my %sargs=$item->{WITH}->to_hash(FORMAT=>"create");
                if ($sargs{'foreground'}) { delete $sargs{'foreground'}; }
                my $num=0;
                foreach my $r (CConfig::Impl::shuffle(@result)) { 
                    # if we haven't picked this obj already then select it
                    if (!$self->RESULT_OBJS->contains($r)) { 
                        $num++;
                        $self->RESULT_OBJS->add($r);
                        $Log->debug(sub {"$me: found ".$r->{'vserver'}.":".$r->{'flexclone'}});

                        # generate subactions for this item, passing in primary keys from 'show'
                        CConfig::Rule::gen_subactions($self,$item,%sargs,$self->RESULT_OBJS->getRequired($r));

                        if ( $aargs{'LIMIT'} && ($aargs{'LIMIT'} == $num ) ) {
                            last SWITCH;
                        } elsif ( $aargs{'NUM'} ne "ALL" ) {
                            last SWITCH;
                        }

                    } 
                }

                last SWITCH if $aargs{'NUM'} eq "ALL" || $strict;
            };
            /create|find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                # compute/default object keys
                if (!defined($args{'vserver'})) { 
                    $args{'vserver'}="vs0"; 
                    # In case someone wants to refer to the auto-generated vserver
                    my $with = $self->resolveAttrList( key => 'vserver', value => $args{vserver}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                }
                if (!defined($args{'parent-volume'})) { 
                    die "FlexClone: create failed, required attribute \"parent-volume\" not found"; 
                }

                if (!defined($args{'parent-snapshot'})) { 
                    my $verMgr = $self->{CMD_INT}->get_version_manager();
                    my $release = $verMgr->get_version_attribute( attribute => "release" );

                    if ( $release =~ /boilermaker/i ) {
                        die "FlexClone: create failed, attribute \"parent-snapshot\" is required in the BoilerMaker release"; 
                    }
                }
                
                if (!defined($args{'flexclone'})) {

                    my $defPattern = $args{'parent-volume'}.'_clone%SSS';
                    if ($args{'-default-autogen-pattern'}) {
                        $defPattern = $args{'-default-autogen-pattern'};
                        delete $args{'-default-autogen-pattern'};
                    }

                    if (!$self->FILER_OBJS_LOADED) {
                        eval { $self->FILER_OBJS->setRows($self->call_nacl_component_routine( routine => 'fetch')); };
                        $self->FILER_OBJS_LOADED(1);
                    }
                    # if obj name not supplied then fabricate an unused one
                    do {
                        $args{'flexclone'}=$self->gen_string( pattern => $defPattern );
                    } until (!$self->RESULT_OBJS->contains(\%args) && !$self->FILER_OBJS->contains(\%args));

                    # In case someone wants to refer to the auto-generated value
                    my $with = $self->resolveAttrList( key => 'flexclone', value => $args{'flexclone'}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                } 

                # create the object
                $self->check_keys(\%args); # check that all req keys are specified

                # generate subactions for this item
                CConfig::Rule::gen_subactions($self,$item,%args);

                # Add object to created objects list record undo actions prior to creating the
                #   object in case the object create times out
                $self->CREATED_OBJS->add(\%args);
                $self->RESULT_OBJS->add(\%args);

                $self->add_undo( component_name => 'NACL::C::Volume', cmd => "delete", args => { vserver => $args{vserver}, volume => $args{flexclone} } );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'create', args => \%args );

                last SWITCH;
            };
            /split_start/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                $self->add_undo( component_name => 'NACL::C::VolumeCloneSplit', cmd => "stop", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, component => 'NACL::C::VolumeCloneSplit', 
                                                    routine => 'start', args => \%args );
                last SWITCH;
            };
            /split_stop/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                $self->add_undo( component_name => 'NACL::C::VolumeCloneSplit', cmd => "start", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, component => 'NACL::C::VolumeCloneSplit', 
                                                    routine => 'stop', args => \%args );
                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}
# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;
    my $me=(caller(0))[3];
    my @result=();

    # update the result objects with attributes from the filer, only need to do this for created objects
    my $resultobjs=$self->RESULT_OBJS;
    if ($self->CREATED_OBJS->size()) {
        eval {
            @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );
        };
        if ($@) {
            @result=();
            die $@ unless $self->noEntries($@);
        }
        foreach my $r (@result) { 
            next unless $self->CREATED_OBJS->contains($r); # did we create this object?
            $resultobjs->update($r);
        }
    }

    return $resultobjs->DATA;
}

# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       self - object instance
# -----------------------------------------------
sub post {
    my $self=shift;
    my $me=(caller(0))[3];


    # generate undo actions for created/modified objects
    $self->gen_undo_commands();

    # Make undo wait for background jobs to complete
    $CConfig::TXMGR->push( 'CConfig::Subtest::wait_finish()');

    # wait for background jobs to complete 
    CConfig::Subtest::wait_finish();

    return 1;
} 

# -----------------------------------------------
# Function: nacl_component_classname
#   Static method which returns the primary NACL
#   component classname for this implementation
#
# Usage:
#   my $nacl_c_classname = nacl_component_classname();
#
# -----------------------------------------------
sub nacl_component_classname {
    return "NACL::C::VolumeClone";
}

# -----------------------------------------------
# Function: cdef_root
#   Return cdef_root for this implementation.
#
# Usage:
#   my $cdef_root = cdef_root();
#
#  Note: This is used for parameter checking until
#        NACL supports an "ignore_extra_options" flag
#        and possibly for CCL simulation.
# -----------------------------------------------
sub cdef_root {
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( component_name => undef,
                @_ );

    if ( $opt{component_name} ) {
        # For delete we use the Volume
        if ( $opt{component_name} =~ /^NACL::C::Volume$/ ) {
            return "volume";
        }
        if ( $opt{component_name} =~ /^NACL::C::VolumeCloneSplit$/ ) {
            return "volume_clone_split";
        }
    }

    return "volume_clone";
}

# -----------------------------------------------
# Function: do_purge
#
#   Static method which overrides the generic CConfig::Impl::do_purge()
#   method which just calls the NACL Component delete.
#
#   Note: This method just calls CConfig::Impl::do_purge() which
#   essentially does what Server volume_purge()
#   did i.e. unmount, offline, delete).
#
# Usage:
#   CConfig::Impl::FlexClone->do_purge( command_interface => $ci,
#             vserver => "vs0", volume => "myvol" )
#
# Parameters:
#       command_interface - NACL::C::Node object
#
# -----------------------------------------------
sub do_purge {
    my $class = shift;
    return CConfig::Impl::Volume->do_purge( @_ );
}


=head1 FUNCTIONS



=head2 new

    create new object

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
