# -*- perl -*-
# $Id$

## @summary CConfig::Impl::IpAddress
## @author frank.radish@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

IpAddress 

=head1 SYNOPSIS

  use CConfig::Impl::IpAddress 

=head1 DESCRIPTION

Implements object actions for IpAddress objects

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::IpAddress
#   IpAddress
# --------------------------------------------------------

package CConfig::Impl::IpAddress;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use NACL::ComponentUtils;
use Carp;
use Network::SCON_DB::IPReservation;
use Network::IPAddress;

use base qw(CConfig::Impl);


# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = reserve
#           item->{'OBJECT'} = ip_address
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self=$class->SUPER::new(
        KEYS=>[qw(address)],
        DEFAULT=>[qw(address)]
    );

    $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = reserve
#           item->{'OBJECT'} = ip_address
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef},
                _INDEX => {type=>SCALAR, default => 0}
            }
    );

    my @result;

    # Action is being called NUM times, but we want to make as few calls to IPManager
    # as necessary, so we pass in _INDEX as a "x of N" indicator and only make the
    # actual call to IPM when x (_INDEX) == N (NUM)

    SWITCH: for ($item->{'ACTION'}) {
            /reserve/i    && do { 

#print "$me: aargs = ".NACL::ComponentUtils::_dump_one(\%aargs)."\n";
                my %args=( version => 'v4',
                           role    => 'data',
                           query   => 'any',
                           $item->{WITH}->to_hash(FORMAT=>"create") );
#print "$me: args = ".NACL::ComponentUtils::_dump_one(\%args)."\n";

                # We are going to squirrel away each request until we reach the last ip_address
                # command, and THEN we will submit the requests to IPManager
                my %req_args = (
                    'role'    => $args{role},
                    'query'   => $args{query},
                    'version' => $args{version},
                    );

                if ( $args{node} ) {
                    $req_args{node}       = $args{node};
                } elsif ( $args{network} ) {
                    # For supporting 1.2.3.4/255.255.255.0 besides just
                    # 3.0.0.0
                    # Typically would encode rule as 
                    #     network={@NODES[1]{DATA_IP2}}/{@NODES[1]{DATA_MASK2}}
                    if ( $args{network} =~ /^(\S+)\/(\S+)$/ ) {
                        my $address = $1;
                        my $mask    = $2;
                        my $NetAddrIP =  NetAddr::IP->new($address, $mask);
                        my $network = $NetAddrIP->network();
                        my ($netid, $network_length) = split('\/', $network);
                        $args{network} = $netid;
                    }
                    $req_args{network_id} = $args{network};
                } elsif ( $args{vlan} ) {
                    $req_args{vlan_id}    = $args{vlan};
                }

                my $hash_key = "";
                foreach my $key ( qw( node network_id vlan_id role query version ) ) {
                    if ( exists $req_args{$key} ) {
                        $hash_key .= $key.'='.$req_args{$key}.',';
                    } else {
                        $hash_key .= $key.'=NULL,';
                    }
                }

                if ( exists $self->{IP_MANAGER_REQUEST}{$hash_key} ) {
                    $self->{IP_MANAGER_REQUEST}{$hash_key}{count}++;
                } else {
                    $self->{IP_MANAGER_REQUEST}{$hash_key}{count} = 1;
                    foreach my $key ( qw( node network_id vlan_id role query version ) ) {
                        if ( exists $req_args{$key} ) {
                            $self->{IP_MANAGER_REQUEST}{$hash_key}{$key} = $req_args{$key};
                        }
                    }
                }
 
                if ( $aargs{NUM} == $aargs{_INDEX} ) {
                    my $variable = $item->{VARIABLE};
                    $variable =~ s/^($|@)//;

                    # Ok, If I want to re-use a block of IPs that I have already reserved (like I would if I were
                    # re-running a redoable ruleset), I would want to make sure I can uniquely identify a reservation.
                    # Might also need to include the CCL rulset file name in there.
                    # $req_args->{job_id} = "static_CCL_var=IPV4_BLOCK1,node=radish-vsim7,version=4,role=data,vlan_id=1041";
                    # Need to assume that people are going to be copying rulesets and not changing variable names, so
                    # need to get the minimum set of attributes to know we are dealing with the same reservation to
                    # build the jobid.  Do NOT want to tack on the time like we used to, and do NOT want to first
                    # delete a reservation of this job_id because if we are re-running a ruleset, we want to re-use the
                    # same IPs.
                    
                    # Might also want to NOT randomize anything in CCL.  This can cause problems with re-running rulesets
                    # like if nodes comes back in a different order
                    my $job_id = 'static_CCL';
                    $job_id .= ':RULESET='.$CConfig::_top_level_ruleset_;

                    # Should tack on NODES= to provide context
                    $job_id .= ':NODES=';
                    my $node_count =0;
                    foreach my $node ( @CConfig::Nodes ) {
                        if ( $node_count == 0 ) {
                            $job_id .= $node->node();
                        } else {
                            $job_id .= ','.$node->node();
                        }
                        $node_count++;
                    }

                    # tack on variable context
                    $job_id .= ':VARIABLE='.$variable;

                    foreach my $key (sort keys %req_args ) {
                        $job_id .= ';'.$key.'='.$req_args{$key};
                    }

                    my $job_id_index = 1;
                    foreach my $key (keys %{$self->{IP_MANAGER_REQUEST}} ) {
                        my $req_args = $self->{IP_MANAGER_REQUEST}{$key};
                        $req_args->{job_id} = $job_id.';count='.$req_args->{count}.';index='.$job_id_index;

                        $job_id_index++;

                        if ( exists $req_args->{node} ) {
                            $req_args->{hostrec} = Hostrec->new( id => $req_args->{node} );

                            # IPReservation does not accept a "node" option
                            delete $req_args->{node};
                        }

                        # To allow a CCL ruleset to be re-usable, we want to re-use a 
                        # reservation if it exists.
                        $req_args->{reuse_reservation} = 1;

                        my ($reservedIp, $error) = Network::SCON_DB::IPReservation->ip_addresses( %{$req_args} );
                        unless( $reservedIp ) {
                           $Log->error("=== ERROR Data IP: $error");
                           foreach my $m ( @{$reservedIp->{'IP'}->{'messages'}} ) { printf("%s\n", $m); }
                           die "$me: Error reserving IP addresses from IPManager";
                        }
                        # IPv6 addresses are expanded by default
                                    $reservedIp->compress_ip_address_list();
                        my @addrs = $reservedIp->ip_address_list();

                        my $vlan_id      = $reservedIp->{IP}->{attrib}{vlan_id};
                        #print "addrs = ".Dumper(\@addrs)."\n";
                        my %ip_info = (
                            gateway      => $reservedIp->ip_address_gateway(),
                            network      => $reservedIp->ip_address_network(),
                            netmask      => $reservedIp->ip_address_netmask(),
                            netmask_bits => $reservedIp->{IP}{attrib}{ip_netmask_bits},
                            version      => $reservedIp->{IP}{attrib}{ip_version},
                        );

                        if ( $vlan_id ) {
                            $ip_info{vlan_id} = $vlan_id;
                        }

                        foreach my $address ( @addrs ) {
                            $ip_info{address}  = $address;
                            $self->RESULT_OBJS->add(\%ip_info);
                        }

                        $CConfig::TXMGR->push( 'CConfig::Impl::IpAddress::release( job_id => "'.$req_args->{job_id}.'" )');
                    }
                } 

                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}

sub release {
    my $me=(caller(0))[3];
    my %args = validate(@_, {
                job_id => {type=>SCALAR},
            }
    );

    my ($DB, $error) = Network::SCON_DB::IPAddrManager->new();

    if ( $DB ) {
        print $me.' calling Network::SCON_DB::IPAddrManager->ip_manager_return( job_id => '.$args{job_id}.")\n";
        my ($ok, $error) = $DB->ip_manager_return( job_id => $args{job_id} );
        if ( !$ok ) {
            die "$me: Error attempting to return IPs associated with jobid=".$args{job_id}." to IPManager: $error";
        }
    } else {
        die "$me: failed connecting to the IPManager DB with $error, could not free job_id=".$args{job_id};
    }
} 

# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;
    my @result=();

    my $resultobjs=$self->RESULT_OBJS;
    return $resultobjs->DATA;
} 


# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       self - object instance
# -----------------------------------------------
sub post {
    my $self=shift;
    
    return 1;
} 

=head1 FUNCTIONS



=head2 new

    create new object

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = reserve 
    item->{'OBJECT'} = ip_address
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = reserve
    item->{'OBJECT'} = ip_address
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
