# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CConfig/Impl/License.pm#1 $

## @summary CConfig::Impl::License
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

License 

=head1 SYNOPSIS

  use License 

=head1 DESCRIPTION

Implements object actions for License objects

=cut


# --------------------------------------------------------
# Class: CConfig::Impl::License
#   License
# --------------------------------------------------------

package CConfig::Impl::License;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use String::Random;
use Carp;

use base qw(CConfig::Impl);

our $V1_KEY_LENGTH = 14;
our $V2_KEY_LENGTH = 28;

# In order
# GX_Base
# GX_Mirror      - older version of SnapMirror_DP
# GX_FlexVol_HPO - striping related
# GX_CIFS
# GX_SnapRestore
# NFS
# SnapMirror_DP
# Striped_Volume - striping
# iSCSI
# FCP
# FlexClone

our %v1_to_v2_converter = (
    VMZIVZGKWVMADF => 'base_1_80_000011',
    TBVMWZGKWVMADF => 'snapmirror',
    FWSBXZGKWVMADF => '',
    RQQQXZGKWVMADF => 'cifs',
    DLOFYZGKWVMADF => 'snaprestore',
    BAKJZZGKWVMADF => 'nfs',
    NUHYZZGKWVMADF => 'snapmirror',
    LJDCBAHKWVMADF => '',
    ZOFNAAHKWVMADF => 'iscsi',
    XDBRBAHKWVMADF => 'fcp',
    PFMUYZGKWVMADF => 'flexclone',

    XBOLYLDPMZSEBG => 'base_1_80_000013',
    VQJPZLDPMZSEBG => 'snapmirror',
    HLHEAMDPMZSEBG => '',
    TFFTAMDPMZSEBG => 'cifs',
    FADIBMDPMZSEBG => 'snaprestore',
    DPYLCMDPMZSEBG => 'nfs',
    PJWADMDPMZSEBG => 'snapmirror',
    NYREEMDPMZSEBG => '',
    BEUPDMDPMZSEBG => 'iscsi',
    ZSPTEMDPMZSEBG => 'fcp',
    RUAXBMDPMZSEBG => 'flexclone',

    ZYSOQSYLTDIFMA => 'base_1_80_000017',
    XNOSRSYLTDIFMA => 'snapmirror',
    JIMHSSYLTDIFMA => '',
    VCKWSSYLTDIFMA => 'cifs',
    HXHLTSYLTDIFMA => 'snaprestore',
    FMDPUSYLTDIFMA => 'nfs',
    RGBEVSYLTDIFMA => 'snapmirror',
    PVWHWSYLTDIFMA => '',
    DBZSVSYLTDIFMA => 'iscsi',
    BQUWWSYLTDIFMA => 'fcp',
    TRFAUSYLTDIFMA => 'flexclone',

    # Translating the AA licenses (i.e. s/n 1-80-123456 to 1-80-000011)
    NWUZZJPJYBFDAA => 'base_1_80_000011',
    LLQDBKPJYBFDAA => 'snapmirror',
    XFOSBKPJYBFDAA => '',
    JAMHCKPJYBFDAA => 'cifs',
    VUJWCKPJYBFDAA => 'snaprestore',
    TJFAEKPJYBFDAA => 'nfs',
    FEDPEKPJYBFDAA => 'snapmirror',
    DTYSFKPJYBFDAA => '',
    RYAEFKPJYBFDAA => 'iscsi',
    PNWHGKPJYBFDAA => 'fcp',
    BYKBBPAMUBDAAA => 'flexclone',
);

our %v2_licenses = (
    base_1_80_000011 => 'ETYYFLXUQUMADFAAAAAAAAAAAAAA',
    base_1_80_000013 => 'GINBJXTZGYSEBGAAAAAAAAAAAAAA',
    base_1_80_000017 => 'IFSEBEPWNCIFMAAAAAAAAAAAAAAA',
    cdmi             => 'WSQFRWDUSAFDAAAAAAAAAAAAAAAA',
    cifs             => 'MJXMPWDUSAFDAAAAAAAAAAAAAAAA',
    fcp              => 'KYSQQWDUSAFDAAAAAAAAAAAAAAAA',
    flexclone        => 'GCKYSWDUSAFDAAAAAAAAAAAAAAAA',
    iscsi            => 'YDVBQWDUSAFDAAAAAAAAAAAAAAAA',
    nfs              => 'APZXOWDUSAFDAAAAAAAAAAAAAAAA',
    snaplock         => 'ERFCUWDUSAFDAAAAAAAAAAAAAAAA',
    snapmanagersuite => 'QLDRUWDUSAFDAAAAAAAAAAAAAAAA',
    snapmirror       => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',
    snapprotectapps  => 'CGBGVWDUSAFDAAAAAAAAAAAAAAAA',
    snaprestore      => 'INOURWDUSAFDAAAAAAAAAAAAAAAA',
    snapvault        => 'SWHNTWDUSAFDAAAAAAAAAAAAAAAA',
);

our %symbolic_to_v2 = reverse %v2_licenses;

our $v2_symbolic_license_keys = join("|", (keys %v2_licenses));

# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my $self=$class->SUPER::new();
    return $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result;
    my %args=$item->{WITH}->to_hash();

    # Server would ignore parameters it didn't understand, but not so with NACL components.
    # A common usage is this (to get the command to execute on each node):
    #    create license with license-code=FEDPEKPJYBFDAA foreach node=@CLUSTERS{nodes} on node=^node;
    #  But "node" isn't actually a valid parameter for 'license add'
    if ( $args{node} ) {
        delete $args{node};
    }

    SWITCH: for ($item->{'ACTION'}) {
            /create/i    && do { 
                # Note, delete is not the mirror of create, you need to provide a feature or package to delete
                # so we are not currently supporting undo on license
                my $node = $self->CMD_INT;
                my $v2_0 = CConfig::Impl::License::is_v2_licensing_supported( node => $self->CMD_INT );
                my $required_package;

                if ( $v2_0 ) {
                    if ( length($args{'license-code'}) < $V2_KEY_LENGTH ) {

                        # Will also convert sybmolic name to a v2 key, 
                        # or a v1 key to a v2 key
                        $args{'license-code'} = $self->convert_license_code( code => $args{'license-code'} );
                    }

                    if ( !$args{'package'} ) {
                       my $symbolic = $symbolic_to_v2{$args{'license-code'}};
                       $Log->comment("NACL::C::SystemLicense::add() requires \"package\" for V2 licenses, using $symbolic for ".
                                     $args{'license-code'}." if incorrect, explicitly specify a package in the license rule");
                       $required_package = $symbolic;
                    } else {
                       $required_package = $args{'package'};
                    }

                    # FWR as of 10/24/11, NACL treats license-code as a scalar, but in newer SN.0
                    # builds with v2.0 licensing, license-code is an array.  For now, keep these
                    # as scalars because that is how NACL views them, but eventually when 
                    # NACL::C::SystemLicense changes, we'll need the following lines

                    ## NACL wants these as ARRAYREFs but CCL wants to see these as strings
                    #$self->list2arrayrefs( fields => [ qw( license-code ) ], args => \%args );
                }

                if ( $args{'license-code'} ) {
                    eval {
                        my $add_args = Storable::dclone(\%args);
                                $self->filter_out_invalid( component_name => "NACL::C::SystemImage", 
                                                           cmd=> "add", args => $add_args );
                        if ( $required_package ) {
                            $add_args->{package} = $required_package;
                        }
                        @result=$self->call_nacl_component_routine( routine => 'add', 
                                                                    args => $add_args,
                                                                    filter_out_invalid => 0 );
                    };

                    if ( $@ ) {
                        if ( $v2_0  && ($@ =~ /Invalid requested field .*feature/)) {
                            # This is just a warning, the license was set properly
                            $Log->warn("NACL::C::SystemImage does not yet fully support v2.0 licensing, but license add succeeded");
                        } else {
                            die $@;
                        }
                    }
                } 
                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}


sub convert_license_code { 
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( code => undef,
                @_ );

    my $new = $opt{code};
    my $orig = $opt{code};
    if ( $orig =~ /$v2_symbolic_license_keys/ ) {
        $new = $v2_licenses{$orig};

        if ( $new ) {
            $Log->comment("$me: Converted symbolic license name: $orig to v2 license key: $new");
        } else {
            $Log->warn("$me: No v2 license key exists corresponding to symbolic name $orig");
        }
    } elsif ( length($orig) == $V1_KEY_LENGTH ) {
        my $feature = $v1_to_v2_converter{$orig};
        if ( $feature ) {
            $new = $v2_licenses{$feature};
        }

        if ( $new ) {
            $Log->comment("$me: Converted v1 license key: $orig to v2: $new, feature: $feature");
        } else {
            $Log->warn("$me: No v2 license key exists corresponding to v1 license key $orig");
        }
    }
    return $new;
}

# -----------------------------------------------
# Function: is_v2_licensing_supported
#   Static method which returns whether or not
#   v2 licensing is supported
#
# Usage:
#   my $is = is_v2_licensing_supported( node => $node );
#
# -----------------------------------------------
sub is_v2_licensing_supported { 
    my $me=(caller(0))[3];
    my %opt = ( node => undef,
                @_ );

    my $node       = $opt{node};
    my $cli_apiset = $node->apiset(category => "Node", interface => "CLI", set => "CMode");
    my $response   = $cli_apiset->system_license_show();
    my $result     = $response->get_parsed_output();
    my $v2_0       = ( exists $result->[0]->{package} ) ? 1 : 0;

    return $v2_0;
}

# -----------------------------------------------
# Function: nacl_component_classname
#   Static method which returns the primary NACL
#   component classname for this implementation
#
# Usage:
#   my $nacl_c_classname = nacl_component_classname();
#
# -----------------------------------------------
sub nacl_component_classname {
    return "NACL::C::License";
}

# -----------------------------------------------
# Function: cdef_root
#   Return cdef_root for this implementation.
#
# Usage:
#   my $cdef_root = cdef_root();
#
#  Note: This is used for parameter checking until
#        NACL supports an "ignore_extra_options" flag
#        and possibly for CCL simulation.
# -----------------------------------------------
sub cdef_root {
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( component_name => undef,
                @_ );

    return "system_license";
}


=head1 FUNCTIONS



=head2 new

    create new object

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
