# -*- perl -*-
# $Id$

## @summary CConfig::Impl::Lun
## @author frank.radish@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Lun 

=head1 SYNOPSIS

  use Lun 

=head1 DESCRIPTION

Implements object actions for Lun objects

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::Lun
#   Lun
# --------------------------------------------------------

package CConfig::Impl::Lun;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use String::Random;
use CConfig::Job;
use Carp;
use NACL::C::Lun;
use NACL::CS::LunGeometry;

use base qw(CConfig::Impl);


# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = lun
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self=$class->SUPER::new(
        KEYS=>[qw(vserver path)],
        DEFAULT=>[qw(path)]
    );

    $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = lun
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result;
    my $strict=0;

    my $subtest = $self->subtest_default( num => $aargs{NUM}, object => $item->{'OBJECT'} );

    SWITCH: for ($item->{'ACTION'}) {
            /strict/i    && do { $strict=1; };
            /find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"find",NUM=>$aargs{'NUM'});

                if ($args{'size'} && ($args{'size'} =~ /B$/))  { $args{'size'} = $self->convertByteString( $args{'size'} ); }

                eval {
                   @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => \%args} );
                };
                if ($@) {
                    @result=();
                    die $@ if ( !$self->noEntries($@) || $strict); 
                }

                my %sargs=$item->{WITH}->to_hash(FORMAT=>"create");
                my $num=0;
                foreach my $r (CConfig::Impl::shuffle(@result)) { 
                    # if we haven't picked this obj already then select it
                    if (!$self->RESULT_OBJS->contains($r)) { 
                        $num++;
                        $self->RESULT_OBJS->add($r);
                        $Log->debug(sub {"$me: found ".$r->{'vserver'}.":".$r->{'path'}});

                        # Forcing this to use CLI because of burt626738
                        # Update with additional data from the geometry command
                        my @geom=$self->call_nacl_component_routine( routine => 'fetch', component_name => 'NACL::CS::LunGeometry', 
                                                                     args => {allow_empty => 1, 
                                                                              apiset_must => { interface => 'CLI' },
                                                                              filter => { vserver => $r->{'vserver'}, 
                                                                                          path    => $r->{'path'} }} );
                        # Look into why geom isn't being filtered
                        foreach my $g (@geom) { 
                            next unless $self->RESULT_OBJS->contains($g);
                            $self->RESULT_OBJS->update($g);
                        }

                        # generate subactions for this item, passing in primary keys from 'show'
                        CConfig::Rule::gen_subactions($self,$item,%sargs,$self->RESULT_OBJS->getRequired($r));

                        if ( $aargs{'LIMIT'} && ($aargs{'LIMIT'} == $num ) ) {
                            last SWITCH;
                        } elsif ( $aargs{'NUM'} ne "ALL" ) {
                            last SWITCH;
                        }

                    } 
                }

                last SWITCH if $aargs{'NUM'} eq "ALL" || $strict;
            };
            /create|find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                # compute/default object keys
                if (!defined($args{'vserver'})) { 
                    $args{'vserver'}="vs0"; 
                    # In case someone wants to refer to the auto-generated vserver
                    my $with = $self->resolveAttrList( key => 'vserver', value => $args{vserver}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                }

                if (!defined($args{'ostype'})) { 
                    die "Lun: create failed, required attribute \"ostype\" not found"; 
                }

                if (defined($args{'path'}) && (defined($args{'volume'}) || defined($args{'lun'}) || defined($args{'qtree'}) ) ) { 
                    die "Lun: create failed, option \"path\" and \"volume,lun,qtree\" are mutually exclusive"; 
                }

                if (!defined($args{'path'})) { 
                    if (!defined($args{'volume'})) { 
                        die "Lun: create failed, if option \"path\" not provided, must provide \"volume\""; 
                    }

                    # path and lun are mutually exclusive
                    if ( !defined($args{'lun'}) ) {
                        my $defPattern = "lun%SSS";
                        if ($args{'-default-autogen-pattern'}) { 
                            $defPattern = $args{'-default-autogen-pattern'}; 
                            delete $args{'-default-autogen-pattern'}; 
                        }
    
                        if (!$self->FILER_OBJS_LOADED) {
                            eval { $self->FILER_OBJS->setRows($self->call_nacl_component_routine( routine => 'fetch')); };
                            $self->FILER_OBJS_LOADED(1);
                        }
                        # if obj name not supplied then fabricate an unused one
                        do {
                            $args{'lun'}=$self->gen_string( pattern => $defPattern );
                            $args{path} = '/vol';
                            $args{path} .= '/' . $args{volume};
                            $args{path} .= '/' . $args{qtree} if (exists $args{qtree});
                            $args{path} .= '/' . $args{lun};
                        } until (!$self->RESULT_OBJS->contains(\%args) && !$self->FILER_OBJS->contains(\%args));

                        # In case someone wants to refer to the auto-generated value
                        my $with = $self->resolveAttrList( key => 'lun',  value => $args{'lun'},  attr_list => $item->{WITH});
                           $with = $self->resolveAttrList( key => 'path', value => $args{'path'}, attr_list => $with);
                        %args=(%args, $with->to_hash(FORMAT=>"create"));
                    } 
       
                    # If the user supplied the lun parameter
                    # path still won't be set
                    if (!defined($args{'path'})) { 
                        $args{path} = '/vol';
                        $args{path} .= '/' . $args{volume};
                        $args{path} .= '/' . $args{qtree} if (exists $args{qtree});
                        $args{path} .= '/' . $args{lun};
                    }

                    # In case someone wants to refer to the auto-generated value
                    my $with = $self->resolveAttrList( key => 'path', value => $args{'path'}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));

                    delete $args{volume};
                    delete $args{qtree} if (exists $args{qtree});
                    delete $args{lun};
                } 

                # create the object
                $self->check_keys(\%args); # check that all req keys are specified

                # generate subactions for this item
                CConfig::Rule::gen_subactions($self,$item,%args);

                # Add object to created objects list record undo actions prior to creating the
                #   object in case the object create times out
                $self->CREATED_OBJS->add(\%args);
                $self->RESULT_OBJS->add(\%args);

                if (!$args{'method-timeout'}) { $args{'method-timeout'} = 120; }

                $self->add_undo( cmd => "delete", args => \%args, add_options => { 'force' => 'true', } );

                $self->call_nacl_component_routine( subtest => $subtest, routine => 'create', args => \%args );

                last SWITCH;
            };
            /modify/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                _remove_keys_other_than_path( \%args );

                $self->add_undo( cmd => "modify", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'modify', args => \%args );
                last SWITCH;
            };
            /move/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                _remove_keys_other_than_path( \%args );
                my %undo_args =  ( vserver => $args{vserver} );
                my %query_args = ( vserver => $args{vserver},
                                  path     => $args{path} );

                # Need to query current for volume/qtree/lun
                my $current=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => \%query_args } );

                my $newpath;
                if ( exists $args{'new-path'} ) {
                    $newpath = $args{'new-path'};
                } else {
                     # of the path components, only qtree and/or lun can change
                     $newpath  = '/vol';
                     $newpath .= '/' . $current->{volume};

                     if ( $args{'new-qtree'} && ( $args{'new-qtree'} !~ /\"\"/) ) {
                         $newpath .= '/' . $args{'new-qtree'};
                     } elsif ( $current->{'new-qtree'} && ( $current->{'new-qtree'} !~ /\"\"/) ) {
                         $newpath .= '/' . $current->{'new-qtree'};
                     }
                     if ( exists $args{'new-lun'}) {
                         $newpath .= '/' . $args{'new-lun'};
                     } else {
                         $newpath .= '/' . $current->{lun};
                     }
                }

                $args{'new-path'}      = $newpath;

                $undo_args{'new-path'} = $args{path};
                $undo_args{path}       = $newpath;

                # Just deal with paths
                delete $args{'new-qtree'} if (exists $args{'new-qtree'});
                delete $args{'new-lun'}   if (exists $args{'new-lun'});

                #lun*> move ?
                #  [ -vserver <vserver name> ]            Vserver Name (default: vs0)
                #   { [-path] <path>                      LUN Path
                #   |  -volume <volume name>              Volume Name
                #    [ -qtree <text (size 0..255)> ]      Qtree Name
                #     [-lun] <text (size 1..255)> }       LUN Name
                #   { [-new-path] <path>                  New LUN Path
                #  | [ -new-qtree <text (size 0..255)> ]  New Qtree Name
                #     [-new-lun] <text (size 0..255)> }   New LUN Name

                $self->add_undo( cmd => "move", args => \%undo_args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'move', args => \%args );

                last SWITCH;
            };
            /rescan/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                # Not "undoable"
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'rescan', args => \%args );
                last SWITCH;
            };
            /resize/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                _remove_keys_other_than_path( \%args );

                my %query_args = %args;
                delete $query_args{size};
                my %undo_args = %query_args;

                # Need to query current size and set that in the add_undo args
                my $current=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => \%query_args } );

                $undo_args{size}  = $current->{size};
                # It is up to the ruleset writer to force reduction resizes in their ruleset, but for 
                # undo, just force unconditionally
                $undo_args{force} = "true";

                $self->add_undo( cmd => "resize", args => \%undo_args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'resize', args => \%args );
                last SWITCH;
            };
            /unmap/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                _remove_keys_other_than_path( \%args );

                $self->add_undo( cmd => "map", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'unmap', args => \%args );
                last SWITCH;
            };
            /map/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                _remove_keys_other_than_path( \%args );

                $self->add_undo( cmd => "unmap", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'map', args => \%args );
                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}

sub _remove_keys_other_than_path {
    my $args = shift;

    if ( !exists $args->{path} ) {
        $args->{path} = '/vol';
        $args->{path} .= '/' . $args->{volume};
        $args->{path} .= '/' . $args->{qtree} if (exists $args->{qtree});
        $args->{path} .= '/' . $args->{lun};
    }

    delete $args->{volume};
    delete $args->{qtree} if (exists $args->{qtree});
    delete $args->{lun};
}

# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;
    my @result=();

    # update the result objects with attributes from the filer, only need to do this for created objects
    my $resultobjs=$self->RESULT_OBJS;
    if ($self->CREATED_OBJS->size()) {
        eval {
           @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );
        };
        if ($@) {
            @result=();
            die $@ unless $self->noEntries($@);
        } else {

            # Update with regular show data
            foreach my $r (@result) { 
                next unless $self->CREATED_OBJS->contains($r); # did we create this object?
                $resultobjs->update($r);
            }

            # Update with additional data from the geometry command
            my @geom=$self->call_nacl_component_routine( routine => 'fetch', component_name => 'NACL::CS::LunGeometry', 
                                                        args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );

            foreach my $r (@geom) { 
                next unless $self->CREATED_OBJS->contains($r); # did we create this object?
                $resultobjs->update($r);
            }
        }
    }
    return $resultobjs->DATA;
} 


# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       self - object instance
# -----------------------------------------------
sub post {
    my $self=shift;
    
    # generate undo actions for created/modified objects
    $self->gen_undo_commands();

    # Make undo wait for background jobs to complete 
    $CConfig::TXMGR->push( 'CConfig::Subtest::wait_finish()'); 

    # wait for background jobs to complete 
    CConfig::Subtest::wait_finish();

    return 1;
} 

# -----------------------------------------------
# Function: nacl_component_classname
#   Static method which returns the primary NACL
#   component classname for this implementation
#
# Usage:
#   my $nacl_c_classname = nacl_component_classname();
#
# -----------------------------------------------
sub nacl_component_classname {
    return "NACL::C::Lun";
}

# -----------------------------------------------
# Function: cdef_root
#   Return cdef_root for this implementation.
#
# Usage:
#   my $cdef_root = cdef_root( component_name => "NACL::C::Lun" );
#
#  Note: This is used for parameter checking until
#        NACL supports an "ignore_extra_options" flag
#        and possibly for CCL simulation.
#        Since Lun has multiple NACL components associated with it,
#        we look to the component_name option to distinguish
# -----------------------------------------------
sub cdef_root {
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( component_name => undef,
                @_ );
    if ( $opt{component_name} ) {
        if ( $opt{component_name} =~ /^LunGeometry$/ ) {
            return "lun_geometry";
        }
    }

    return "lun";
}

=head1 FUNCTIONS



=head2 new

    create new object

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = lun
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = lun
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
