# -*- perl -*-
# $Id$

## @summary CConfig::Impl::LunIgroup
## @author frank.radish@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

LunIgroup 

=head1 SYNOPSIS

  use LunIgroup 

=head1 DESCRIPTION

Implements object actions for LunIgroup objects

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::LunIgroup
#   LunIgroup
# --------------------------------------------------------

package CConfig::Impl::LunIgroup;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use String::Random;
use CConfig::Job;
use Carp;
use NACL::C::LunIgroup;

use base qw(CConfig::Impl);

# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = lun_igroup
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self=$class->SUPER::new(
        KEYS=>[qw(vserver igroup)],
        DEFAULT=>[qw(igroup)]
    );

    $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = lun_igroup
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result;
    my $strict=0;

    my $subtest = $self->subtest_default( num => $aargs{NUM}, object => $item->{'OBJECT'} );

    SWITCH: for ($item->{'ACTION'}) {
            /strict/i    && do { $strict=1; };
            /find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"find",NUM=>$aargs{'NUM'});

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( initiator ) ], args => \%args );

                eval {
                   @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => \%args} );
                };
                if ($@) {
                    @result=();
                    die $@ if ( !$self->noEntries($@) || $strict); 
                }

                my %sargs=$item->{WITH}->to_hash(FORMAT=>"create");
                my $num=0;
                foreach my $r (CConfig::Impl::shuffle(@result)) { 
                    # if we haven't picked this obj already then select it
                    if (!$self->RESULT_OBJS->contains($r)) { 
                        $num++;
                        $self->RESULT_OBJS->add($r);
                        $Log->debug(sub {"$me: found ".$r->{'vserver'}.":".$r->{'igroup'}});

                        # generate subactions for this item, passing in primary keys from 'show'
                        CConfig::Rule::gen_subactions($self,$item,%sargs,$self->RESULT_OBJS->getRequired($r));

                        if ( $aargs{'LIMIT'} && ($aargs{'LIMIT'} == $num ) ) {
                            last SWITCH;
                        } elsif ( $aargs{'NUM'} ne "ALL" ) {
                            last SWITCH;
                        }

                    } 
                }

                last SWITCH if $aargs{'NUM'} eq "ALL" || $strict;
            };
            /create|find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                # Note: port-name is actually a lif of data-protocol iscsi

                #create -vserver vs0 -portset portset0003 -port-name radish-vsim3_iscsi 

                #radish-vsim31312571715::lun portset*> show
                #Vserver   Portset      Protocol Port Names              Igroups
                #--------- ------------ -------- ----------------------- ------------
                #vs0       portset0003  mixed    radish-vsim3_iscsi      -

                # compute/default object keys
                if (!defined($args{'vserver'})) { 
                    $args{'vserver'}="vs0"; 
                    # In case someone wants to refer to the auto-generated vserver
                    my $with = $self->resolveAttrList( key => 'vserver', value => $args{vserver}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                }

                if (!defined($args{'ostype'})) { 
                    die "LunIgroup: create failed, required attribute \"ostype\" not found"; 
                }

                if ( !defined($args{'igroup'}) ) {
                    my $defPattern = "igroup%SSS";
                    if ($args{'-default-autogen-pattern'}) { 
                        $defPattern = $args{'-default-autogen-pattern'}; 
                        delete $args{'-default-autogen-pattern'}; 
                    }

                    if (!$self->FILER_OBJS_LOADED) {
                        eval { $self->FILER_OBJS->setRows($self->call_nacl_component_routine( routine => 'fetch')); };
                        $self->FILER_OBJS_LOADED(1);
                    }
                    # if obj name not supplied then fabricate an unused one
                    do {
                        $args{'igroup'}=$self->gen_string( pattern => $defPattern );
                    } until (!$self->RESULT_OBJS->contains(\%args) && !$self->FILER_OBJS->contains(\%args));

                    # In case someone wants to refer to the auto-generated value
                    my $with = $self->resolveAttrList( key => 'igroup', value => $args{'igroup'}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                } 

                # create the object
                $self->check_keys(\%args); # check that all req keys are specified

                # generate subactions for this item
                CConfig::Rule::gen_subactions($self,$item,%args);

                # Add object to created objects list record undo actions prior to creating the
                #   object in case the object create times out
                $self->CREATED_OBJS->add(\%args);
                $self->RESULT_OBJS->add(\%args);

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( initiator ) ], args => \%args );

                if (!$args{'method-timeout'}) { $args{'method-timeout'} = 120; }

                $self->add_undo( cmd => "delete", args => \%args );

                $self->call_nacl_component_routine( subtest => $subtest, routine => 'create', args => \%args );

                last SWITCH;
            };
            /modify/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                $self->add_undo( cmd => "modify", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'modify', args => \%args );
                last SWITCH;
            };
            /rename/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                my %undo_args = %args;

                # we have to swap the old and new names for undo
                $undo_args{'new-name'} = $args{igroup};
                $undo_args{igroup}     = $args{'new-name'};

                $self->add_undo( cmd => "rename", args => \%undo_args );

                $self->call_nacl_component_routine( subtest => $subtest, routine => 'rename', args => \%args );
                last SWITCH;
            };
            /add/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( initiator ) ], args => \%args );

                my %undo_args = %args;
                $undo_args{force}  = 'true';
                $self->add_undo( cmd => "remove", args => \%undo_args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'add', args => \%args );
                last SWITCH;
            };
            /remove/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( initiator ) ], args => \%args );

                $self->add_undo( cmd => "add", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'remove', args => \%args );
                last SWITCH;
            };
            /bind/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                $self->add_undo( cmd => "unbind", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'bind', args => \%args );
                last SWITCH;
            };
            /unbind/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                $self->add_undo( cmd => "bind", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'unbind', args => \%args );
                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}

# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;
    my @result=();

    # update the result objects with attributes from the filer, only need to do this for created objects
    my $resultobjs=$self->RESULT_OBJS;
    if ($self->CREATED_OBJS->size()) {
        eval {
           @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );
        };
        if ($@) {
            @result=();
            die $@ unless $self->noEntries($@);
        }

        # Update with show data
        foreach my $r (@result) { 
            next unless $self->CREATED_OBJS->contains($r); # did we create this object?
            $resultobjs->update($r);
        }
    }
    return $resultobjs->DATA;
} 


# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       self - object instance
# -----------------------------------------------
sub post {
    my $self=shift;
    
    # generate undo actions for created/modified objects
    $self->gen_undo_commands();

    # Make undo wait for background jobs to complete 
    $CConfig::TXMGR->push( 'CConfig::Subtest::wait_finish()'); 

    # wait for background jobs to complete 
    CConfig::Subtest::wait_finish();

    return 1;
} 

# -----------------------------------------------
# Function: nacl_component_classname
#   Static method which returns the primary NACL
#   component classname for this implementation
#
# Usage:
#   my $nacl_c_classname = nacl_component_classname();
#
# -----------------------------------------------
sub nacl_component_classname {
    return "NACL::C::LunIgroup";
}

# -----------------------------------------------
# Function: cdef_root
#   Return cdef_root for this implementation.
#
# Usage:
#   my $cdef_root = cdef_root( component_name => "NACL::C::LunIgroup" );
#
#  Note: This is used for parameter checking until
#        NACL supports an "ignore_extra_options" flag
#        and possibly for CCL simulation.
#        Since LunIgroup has multiple NACL components associated with it,
#        we look to the component_name option to distinguish
# -----------------------------------------------
sub cdef_root {
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( component_name => undef,
                @_ );

    return "lun_igroup";
}

=head1 FUNCTIONS



=head2 new

    create new object

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = lun_igroup
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = lun_igroup
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
