# -*- perl -*-
# $Id$

## @summary CConfig::Impl::NetPortIfgrp
## @author frank.radish@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

NetPortIfgrp 

=head1 SYNOPSIS

  use NetPortIfgrp 

=head1 DESCRIPTION

Implements object actions for NetPortIfgrp objects
SMF Menu: network port ifgrp

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::NetPortIfgrp
#   NetPortIfgrp object
# --------------------------------------------------------

package CConfig::Impl::NetPortIfgrp;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use String::Random;
use Carp;
use NACL::C::NetworkPort;
use NACL::C::NetworkPortIfgrp;

use base qw(CConfig::Impl);

# -----------------------------------------------
# Function: new
#   create new object
#
# Usage: 
#   new(item)
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = ifgrp
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Return: 
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self=$class->SUPER::new(
        KEYS=>[qw(node ifgrp)],
        DEFAULT=>[qw(ifgrp)]
    );

    $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = ifgrp
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result=();
    my $strict=0;

    my $subtest = $self->subtest_default( num => $aargs{NUM}, object => $item->{'OBJECT'} );

    SWITCH: for ($item->{'ACTION'}) {
            /strict/i    && do { $strict=1; };
            /find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"find",NUM=>$aargs{'NUM'});
                eval {
                    @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => \%args} );
                }; 
                if ($@) {
                    @result=();
                    die $@ if ( !$self->noEntries($@) || $strict); 
                };

                my %sargs=$item->{WITH}->to_hash(FORMAT=>"create");
                my $num=0;
                foreach my $r (CConfig::Impl::shuffle(@result)) { 
                    # if we haven't picked this obj already then select it
                    if (!$self->RESULT_OBJS->contains($r)) {
                        $num++;
                        $self->RESULT_OBJS->add($r);

                        # generate subactions for this item, passing in primary keys from 'show'
                        CConfig::Rule::gen_subactions($self,$item,%sargs,$self->RESULT_OBJS->getRequired($r));

                        if ( $aargs{'LIMIT'} && ($aargs{'LIMIT'} == $num ) ) {
                            last SWITCH;
                        } elsif ( $aargs{'NUM'} ne "ALL" ) {
                            last SWITCH;
                        }

                    } 
                }
                last SWITCH if $aargs{'NUM'} eq "ALL" || $strict;
            };
            /create|find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                if (!defined($args{'node'})) {
                    $args{'node'} = $self->CMD_INT->node();
                }

                if (!defined($args{'ifgrp'})) {
                    my $defPattern = 'a%(s:0)%(s:a)';
                    if ($args{'-default-autogen-pattern'}) {
                        $defPattern = $args{'-default-autogen-pattern'};
                        delete $args{'-default-autogen-pattern'};
                    }

                    if (!$self->FILER_OBJS_LOADED) {
                        eval { $self->FILER_OBJS->setRows($self->call_nacl_component_routine( routine => 'fetch')); };
                        $self->FILER_OBJS_LOADED(1);
                    }
                    # if routing group name not supplied then fabricate an unused one
                    do {
                        $args{'ifgrp'}=$self->gen_string( pattern => $defPattern );
                    } until (!$self->RESULT_OBJS->contains(\%args) && !$self->FILER_OBJS->contains(\%args));

                    # In case someone wants to refer to the auto-generated value
                    my $with = $self->resolveAttrList( key => 'ifgrp', value => $args{'ifgrp'}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                } 

                # create the object
                $self->check_keys(\%args); # check that all req keys are specified

                # generate subactions for this item
                CConfig::Rule::gen_subactions($self,$item,%args);

                $self->CREATED_OBJS->add(\%args);
                $self->RESULT_OBJS->add(\%args);

                $self->add_undo( cmd => "delete", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'create', args => \%args );

                last SWITCH;
            };
            /add-port/i && do {
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                $self->add_undo( cmd => "remove-port", args => \%args );
                # add port will error out if the port is already in the ifgrp, so checking
                # to see if it is already in there before doing the add, so the rule can be
                # "re-doable"
                my $result=$self->call_nacl_component_routine( routine => 'fetch', 
                                                               args => {allow_empty => 1, 
                                                                        filter => { node  => $args{node}, 
                                                                                    ifgrp => $args{ifgrp} }} );
                my $do_add = 1;
                if ( $result ) {
                    if ( grep( /^$args{port}$/, $result->ports() ) ) {
                        $do_add = 0;
                        $Log->comment("$me: port ".$args{port}." is already in ifgrp ".$args{ifgrp}." on node ".$args{node}.", ignoring add-port");
                    }
                }
                
                if ( $do_add ) {
                    $self->call_nacl_component_routine( subtest => $subtest, routine => 'add-port', args => \%args );
                }
                last SWITCH;
            };
            /remove-port/i && do {
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                $self->add_undo( cmd => "add-port", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'remove-port', args => \%args );
                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}
# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;
    my $me=(caller(0))[3];
    my @result=();

    # get filer attributes for created objects
    my $resultobjs=$self->RESULT_OBJS;
    if ($self->CREATED_OBJS->size()) {
        eval {
            @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );
        };
        if ($@) {
            @result=();
            die $@ unless $self->noEntries($@);
        }
        foreach my $r (@result) { 
            next unless $self->CREATED_OBJS->contains($r); # did we create this object?
            $resultobjs->update($r);
        }
    }

    return $resultobjs->DATA;
}

# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       self - object instance
# -----------------------------------------------
sub post {
    my $self=shift;
    my $me=(caller(0))[3];

    # generate undo actions for created/modified objects
    $self->gen_undo_commands();

    # Make undo wait for background jobs to complete
    $CConfig::TXMGR->push( 'CConfig::Subtest::wait_finish()');

    # wait for background jobs to complete
    CConfig::Subtest::wait_finish();

    return 1;
} 

# -----------------------------------------------
# Function: nacl_component_classname
#   Static method which returns the primary NACL
#   component classname for this implementation
#
# Usage:
#   my $nacl_c_classname = nacl_component_classname();
#
# -----------------------------------------------
sub nacl_component_classname {
    return "NACL::C::NetworkPortIfgrp";
}

# -----------------------------------------------
# Function: cdef_root
#   Return cdef_root for this implementation.
#
# Usage:
#   my $cdef_root = cdef_root();
#
#  Note: This is used for parameter checking until
#        NACL supports an "ignore_extra_options" flag
#        and possibly for CCL simulation.
# -----------------------------------------------
sub cdef_root {
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( component_name => undef,
                @_ );

    return "network_port_ifgrp";
}

# -----------------------------------------------
# Function: do_purge
#
#   Static method which overrides the generic CConfig::Impl::do_purge()
#   method which just calls the NACL Component delete.
#
# Usage:
#   CConfig::Impl::NetPortIfgrp->do_purge( command_interface => $ci,
#             node => "mynode", ifgrp => "a0a" )
#
# Parameters:
#       command_interface - NACL::C::Node object
#
# -----------------------------------------------
sub do_purge {
    my $class = shift;
    my %opts = ( command_interface => undef,
                 node => undef,
                 ifgrp => undef,
                 @_ );

           NACL::C::NetworkPort->modify( command_interface => $opts{command_interface},
                                         node => $opts{node},
                                         port => $opts{ifgrp}, 
                                         'up-admin' => 'false' );
    return NACL::C::NetworkPortIfgrp->delete( @_ );
}

# -----------------------------------------------
# Function: do_undo
#   Undo routine which just calls the method on the NACL
#   component.  Impls can override this
#
# Usage:
#   do_undo( 'NACL::C::Volume', '-cmd' => 'modify', %opt );
#
# -----------------------------------------------
sub do_undo {
    my $class = shift;
    my %opt = ( 'command_interface' => undef,
                '-cmd'              => undef,
                @_ );
    my $cmd = delete $opt{'-cmd'};

    if ( !defined $CConfig::LOADED{$class} ) {
        eval "use $class";
        $CConfig::LOADED{$class} = 1;
    }

    if ( $cmd =~ /remove-port/i ) {
        my %port_opt = ( node => undef,
                         ifgrp => undef,
                         @_ );

        # have to admin down the ifgrp before we remove a port from it
        NACL::C::NetworkPort->modify( command_interface => $opt{command_interface},
                                      node              => $port_opt{node},
                                      port              => $port_opt{ifgrp}, 
                                      'up-admin'        => 'false' );
        $class->$cmd( %opt );
    } else {
        $class->$cmd( %opt );
    }
}



=head1 FUNCTIONS



=head2 new

    create new object

=head3 Usage

    new(item)

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = ifgrp
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = ifgrp
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
