# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CConfig/Impl/QuotaPolicyRule.pm#1 $

## @summary CConfig::Impl::QuotaPolicyRule
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here


=head1 NAME

QuotaPolicyRule 

=head1 SYNOPSIS

  use QuotaPolicyRule 

=head1 DESCRIPTION

Implements object actions for QuotaPolicyRule objects

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::QuotaPolicyRule
#   QuotaPolicyRule
# --------------------------------------------------------

package CConfig::Impl::QuotaPolicyRule;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use String::Random;
use Carp;

use base qw(CConfig::Impl);

# -----------------------------------------------
# Function: new
#   create new object
#
# Usage: 
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self=$class->SUPER::new(
        KEYS=>[qw(vserver volume policy-name target type)],
    );

    $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(<item>)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result;
    my $strict=0;

    my $subtest = $self->subtest_default( num => $aargs{NUM}, object => $item->{'OBJECT'} );

    # FWR bug in NACL (BURTS: 494645 & 495942) passes -qtree "" for RR & BR, but it is invalid for BR

    SWITCH: for ($item->{'ACTION'}) {
            /strict/i    && do { $strict=1; };
            /find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"find",NUM=>$aargs{'NUM'});
                eval {
                    @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => \%args} );
                };
                if ($@) {
                    @result=();
                    die $@ if ( !$self->noEntries($@) || $strict); 
                }

                my %sargs=$item->{WITH}->to_hash(FORMAT=>"create");
                my $num=0;
                foreach my $r (CConfig::Impl::shuffle(@result)) { 
                    # if we haven't picked this obj already then select it
                    if (!$self->RESULT_OBJS->contains($r)) {
                        $num++;
                        $self->RESULT_OBJS->add($r);

                        # generate subactions for this item, passing in primary keys from 'show'
                        CConfig::Rule::gen_subactions($self,$item,%sargs,$self->RESULT_OBJS->getRequired($r));

                        if ( $aargs{'LIMIT'} && ($aargs{'LIMIT'} == $num ) ) {
                            last SWITCH;
                        } elsif ( $aargs{'NUM'} ne "ALL" ) {
                            last SWITCH;
                        }

                    } 
                }
                last SWITCH if $aargs{'NUM'} eq "ALL" || $strict;
            };
            /create|find/i    && do { 
                # compute/default object keys
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                if (!defined($args{'vserver'})) { 
                    $args{'vserver'}="vs0"; 
                    # In case someone wants to refer to the auto-generated vserver
                    my $with = $self->resolveAttrList( key => 'vserver', value => $args{vserver}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                }
                if (!defined($args{'policy-name'})) { 
                    die "$me:ERROR: policy-name is a required field for ".$item->{'OBJECT'}." action ".$item->{'ACTION'}."\n"; 
                }
                if (!defined($args{'volume'})) { 
                    die "$me:ERROR: volume is a required field for ".$item->{'OBJECT'}." action ".$item->{'ACTION'}."\n"; 
                }
                if (!defined($args{'type'})) { 
                    $args{'type'}="tree";

                    # In case someone wants to refer to the auto-generated volume
                    my $with = $self->resolveAttrList( key => 'type', value => $args{'type'}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                }

                # create the object
                $self->check_keys(\%args); # check that all req keys are specified

                # generate subactions for this item
                CConfig::Rule::gen_subactions($self,$item,%args);

                # FWR bug in NACL (BURTS: 494645 & 495942) passes -qtree "" for RR & BR, but it is invalid for BR
                if (exists($args{'qtree'})) {
                    my $verMgr = $self->{CMD_INT}->get_version_manager();
                    my $release = $verMgr->get_version_attribute( attribute => "release" );

                    if ( $release =~ /boilermaker/i ) {
                        delete $args{'qtree'};
                    } elsif ( !$args{'qtree'} ) {
                        $args{'qtree'} = '""';
                    }
                }

                if (exists($args{'target'}) && !$args{'target'}) {
                    $args{'target'} = '""';
                }

                $self->CREATED_OBJS->add(\%args);
                $self->RESULT_OBJS->add(\%args);

                $self->add_undo( cmd => "delete", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'create', args => \%args );

                last SWITCH;
            };
            /modify/i && do {
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                # FWR bug in NACL (BURTS: 494645 & 495942) passes -qtree "" for RR & BR, but it is invalid for BR
                if (exists($args{'qtree'})) {
                    my $verMgr = $self->{CMD_INT}->get_version_manager();
                    my $release = $verMgr->get_version_attribute( attribute => "release" );

                    if ( $release =~ /boilermaker/i ) {
                        delete $args{'qtree'};
                    } elsif ( !$args{'qtree'} ) {
                        $args{'qtree'} = '""';
                    }
                }

                if (exists($args{'target'}) && !$args{'target'}) {
                    $args{'target'} = '""';
                }

                $self->add_undo( cmd => "modify", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'modify', args => \%args );
                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}

sub get_results {
    my $self=shift;
    my $me=(caller(0))[3];
    my @result=();

    # get filer attributes for created objects
    my $resultobjs=$self->RESULT_OBJS;
    if ($self->CREATED_OBJS->size()) {
        eval {
            @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );
        };
        if ($@) {
            @result=();
            die $@ unless $self->noEntries($@);
        }
        foreach my $r (@result) { 
            next unless $self->CREATED_OBJS->contains($r); # did we create this object?
            $resultobjs->update($r);
        }
    }

    return $resultobjs->DATA;
}

# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       self - object instance
# -----------------------------------------------
sub post {
    my $self=shift;
    my $me=(caller(0))[3];

    # generate undo actions for created/modified objects
    $self->gen_undo_commands();

    # Make undo wait for background jobs to complete
    $CConfig::TXMGR->push( 'CConfig::Subtest::wait_finish()');

    # wait for background jobs to complete
    CConfig::Subtest::wait_finish();

    return 1;
}

# -----------------------------------------------
# Function: nacl_component_classname
#   Static method which returns the primary NACL
#   component classname for this implementation
#
# Usage:
#   my $nacl_c_classname = nacl_component_classname();
#
# -----------------------------------------------
sub nacl_component_classname {
    return "NACL::C::QuotaPolicyRule";
}

# -----------------------------------------------
# Function: cdef_root
#   Return cdef_root for this implementation.
#
# Usage:
#   my $cdef_root = cdef_root();
#
#  Note: This is used for parameter checking until
#        NACL supports an "ignore_extra_options" flag
#        and possibly for CCL simulation.
# -----------------------------------------------
sub cdef_root {
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( component_name => undef,
                @_ );

    return "volume_quota_policy_rule";
}

=head1 FUNCTIONS



=head2 new

    create new object

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(<item>)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(<item>)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
