# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CConfig/Impl/Snapmirror.pm#1 $

## @summary CConfig::Impl::Snapmirror
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

SnapMirror 

=head1 SYNOPSIS

  use SnapMirror 

=head1 DESCRIPTION

Implements object actions for SnapMirror objects

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::Snapmirror
#   Snapmirror
# --------------------------------------------------------

package CConfig::Impl::Snapmirror;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use String::Random;
use Carp;
use NACL::STask::Snapmirror;
use NACL::GeneralUtils qw(nacl_method_retry);
use NATE::BaseException qw(:try);

use base qw(CConfig::Impl);

# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;

    my $self=$class->SUPER::new(
        KEYS=>[qw(source-cluster source-vserver source-volume destination-cluster destination-vserver destination-volume)],
    );

    $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result;
    my $strict=0;

    my $subtest = $self->subtest_default( num => $aargs{NUM}, object => $item->{'OBJECT'} );

    SWITCH: for ($item->{'ACTION'}) {
            /strict/i    && do { $strict=1; };
            /find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"find",NUM=>$aargs{'NUM'});
                eval {
                    @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => \%args} );
                };
                if ($@) {
                    @result=();
                    die $@ if ( !$self->noEntries($@) || $strict); 
                }

                my %sargs=$item->{WITH}->to_hash(FORMAT=>"create");
                my $num=0;
                foreach my $r (CConfig::Impl::shuffle(@result)) { 
                    # if we haven't picked this obj already then select it
                    if (!$self->RESULT_OBJS->contains($r)) { 
                        $num++;
                        $self->RESULT_OBJS->add($r);

                        # generate subactions for this item, passing in primary keys from 'show'
                        CConfig::Rule::gen_subactions($self,$item,%sargs,$self->RESULT_OBJS->getRequired($r));

                        if ( $aargs{'LIMIT'} && ($aargs{'LIMIT'} == $num ) ) {
                            last SWITCH;
                        } elsif ( $aargs{'NUM'} ne "ALL" ) {
                            last SWITCH;
                        }

                    } 
                }
                last SWITCH if $aargs{'NUM'} eq "ALL" || $strict;
            };
            /create|find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                # default attributes
                if (!defined($args{'source-cluster'})) { 
                    $args{'source-cluster'}=$CConfig::VARS{'@CLUSTERS'}->get(0)->{'_default_'};

                    # In case someone wants to refer to the auto-generated volume
                    my $with = $self->resolveAttrList( key => 'source-cluster', value => $args{'source-cluster'}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                }
                if (!defined($args{'destination-cluster'})) { 
                    $args{'destination-cluster'}=$CConfig::VARS{'@CLUSTERS'}->get(0)->{'_default_'};

                    # In case someone wants to refer to the auto-generated volume
                    my $with = $self->resolveAttrList( key => 'destination-cluster', value => $args{'destination-cluster'}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                }
                # create the object
                $self->check_keys(\%args); # check that all req keys are specified

                # generate subactions for this item
                CConfig::Rule::gen_subactions($self,$item,%args);
                
                # add the item to the object lists
                $self->CREATED_OBJS->add(\%args);
                $self->RESULT_OBJS->add(\%args);

                $self->add_undo( cmd => "delete", args => \%args );

                nacl_method_retry(
                    code        => sub { $self->call_nacl_component_routine( subtest => $subtest, routine => 'create', args => \%args ); },
                    tries_count => 10,
                    sleep_time  => 10,
                    exceptions  => 'NATE::BaseException'
                );

                last SWITCH;
            };
            /initialize-ls-set/i    && do { 
                # initialize the object
                my $result_txt;
                my %args=$item->{WITH}->to_hash();
                # default attributes
                if (!defined($args{'source-cluster'})) { 
                    $args{'source-cluster'}=$CConfig::VARS{'@CLUSTERS'}->get(0)->{'_default_'};
                }
                # no undo for initialize_ls_set
                nacl_method_retry(
                    code        => sub { $self->call_nacl_component_routine( subtest => $subtest, routine => 'initialize_ls_set', args => \%args ); },
                    tries_count => 10,
                    sleep_time  => 10,
                    exceptions  => 'NATE::BaseException'
                );

                last SWITCH;
            };
            /initialize/i    && do { 
                # initialize the object
                my $result_txt;
                my %args=$item->{WITH}->to_hash();
                # default attributes
                if (!defined($args{'source-cluster'})) { 
                    $args{'source-cluster'}=$CConfig::VARS{'@CLUSTERS'}->get(0)->{'_default_'};
                }
                if (!defined($args{'destination-cluster'})) { 
                    $args{'destination-cluster'}=$CConfig::VARS{'@CLUSTERS'}->get(0)->{'_default_'};
                }
                # no undo for initialize
                nacl_method_retry(
                    code        => sub { $self->call_nacl_component_routine( subtest => $subtest, routine => 'initialize', args => \%args ); },
                    tries_count => 10,
                    sleep_time  => 10,
                    exceptions  => 'NATE::BaseException'
                );
                last SWITCH;
            };
            /modify/i && do {
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                $self->add_undo( cmd => "modify", args => \%args );
                nacl_method_retry(
                    code        => sub { $self->call_nacl_component_routine( subtest => $subtest, routine => 'modify', args => \%args ); },
                    tries_count => 10,
                    sleep_time  => 10,
                    exceptions  => 'NATE::BaseException'
                );
                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}

# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;
    my $me=(caller(0))[3];
    my @result=();

    # get filer attributes for created objects
    my $resultobjs=$self->RESULT_OBJS;
    if ($self->CREATED_OBJS->size()) {
        eval {
            # only include required attributes
            @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );
        };
        if ($@) {
            @result=();
            die $@ unless $self->noEntries($@);
        }
        foreach my $r (@result) { 
            next unless $self->CREATED_OBJS->contains($r); # check that we created this object
            $resultobjs->update($r);
        }
    }

    return $resultobjs->DATA;
}

# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       <self> - object instance
# -----------------------------------------------
sub post {
    my $self=shift;

    # generate undo actions for created/modified objects
    $self->gen_undo_commands();

    # Make undo wait for background jobs to complete
    $CConfig::TXMGR->push( 'CConfig::Subtest::wait_finish()');

    # wait for background jobs to complete
    CConfig::Subtest::wait_finish();

    return 1;
}

# -----------------------------------------------
# Function: nacl_component_classname
#   Static method which returns the primary NACL
#   component classname for this implementation
#
# Usage:
#   my $nacl_c_classname = nacl_component_classname();
#
# -----------------------------------------------
sub nacl_component_classname {
    return "NACL::C::Snapmirror";
}

# -----------------------------------------------
# Function: cdef_root
#   Return cdef_root for this implementation.
#
# Usage:
#   my $cdef_root = cdef_root();
#
#  Note: This is used for parameter checking until
#        NACL supports an "ignore_extra_options" flag
#        and possibly for CCL simulation.
# -----------------------------------------------
sub cdef_root {
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( component_name => undef,
                @_ );

    return "snapmirror";
}

# -----------------------------------------------
# Function: do_purge
#
#   Static method which overrides the generic CConfig::Impl::do_purge()
#   method which just calls the NACL Component delete.
#   This gives you an opportunity to do more than just a simple
#   delete (in this case, quiesce, abort and then delete.)
#
# Usage:
#   do_purge( "CConfig::Impl::Snapmirror", command_interface => $ci,
#              source-cluster=>$sc,
#              source-vserver=>$svs,
#              source-volume=>$sv,
#              destination-cluster=>$dc,
#              destination-vserver=>$dvs,
#              destination-volume=>$dv );
#
# Parameters:
#       class - class name, not directly used, but the general mechanism
#               used for this call requires this method signature.
#       command_interface - NACL::C::Node object
#
# -----------------------------------------------
sub do_purge {
    my $class = shift;
    my %opt = ( 'command_interface' => undef,
                @_ );

    # If we are deleting 2 snapmirrors that both use the same source volume, the second delete may fail
    # So handle a limited number of retries.
    my $retries = 3;
RETRY:

    eval {NACL::STask::Snapmirror->purge( %opt, nacltask_if_purged => 'pass' ); };
    my $rc = $@;

    # It appears that if you attempt to delete 2 snapmirrors, which each have the same
    # source volume, the first will succeed and subsequent deletes will need to wait
    # until the first one has completed.
    # or a transfer might be in progress
    if ( $rc ) {
        eval { my $shortrc = (split('\n', $rc))[0]; $Log->warn($shortrc); };
        if ( $rc =~ /With Job ID\s+(\d+)/ ) {
            my @jobs;
            push(@jobs,$1);
            while(CConfig::Job::are_jobs_active(\@jobs)) { sleep 10; }
            CConfig::Job::check_jobs(\@jobs);
        } else {
            Tharn::sleep( 60 );
        }
        $retries--;
        if ( $retries > 0 ) {
            # BURT620753 involves how snapmirror release does not tolerate source or 
            # destination cluster.  Try to call release here without the offending parameters
            # We don't really care if it passes or fails, since it is the $rc from purge()
            # that ultimately determines pass/fail status
            my %release_opts;
            if ( $opt{'source-path'} && $opt{'destination-path'} ) {
                $release_opts{'source-path'}      = $opt{'source-path'};
                $release_opts{'destination-path'} = $opt{'destination-path'};

                if ( $opt{'command_interface'} ) {
                    $release_opts{'command_interface'} = $opt{'command_interface'};
                }
                if ( $opt{'method-timeout'} ) {
                    $release_opts{'method-timeout'} = $opt{'method-timeout'};
                }
            } else {
                if ( $opt{'source-cluster'} ) {
                    $release_opts{'source-path'} = $opt{'source-cluster'}.'://';
                }
                if ( $opt{'source-vserver'} ) {
                    $release_opts{'source-path'} .= $opt{'source-vserver'}.'/';
                }
                if ( $opt{'source-volume'} ) {
                    $release_opts{'source-path'} .= $opt{'source-volume'};
                }

                if ( $opt{'destination-cluster'} ) {
                    $release_opts{'destination-path'} = $opt{'destination-cluster'}.'://';
                }
                if ( $opt{'destination-vserver'} ) {
                    $release_opts{'destination-path'} .= $opt{'destination-vserver'}.'/';
                }
                if ( $opt{'destination-volume'} ) {
                    $release_opts{'destination-path'} .= $opt{'destination-volume'};
                }

                if ( $opt{'command_interface'} ) {
                    $release_opts{'command_interface'} = $opt{'command_interface'};
                }
                if ( $opt{'method-timeout'} ) {
                    $release_opts{'method-timeout'} = $opt{'method-timeout'};
                }
            }

            if ( %release_opts ) {
                $Log->comment("Attempting to work-around BURT620753");
                eval {NACL::STask::Snapmirror->release( %release_opts );};
                if ( $@ ) {
                    $Log->warn("Attempt to work-around BURT620753 failed with $@");
                }
            }
           
            goto RETRY;
        } else {
            die $rc
        }
    } elsif ( $rc ) {
        die $rc
    }
}


=head1 FUNCTIONS



=head2 new

    create new object

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(<item>)

=head3 Parameters:


=over 4


=item <self>

    object instance

=item item

    reference to the parser hash
    

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
