# -*- perl -*-
# $Id: $

## @summary CConfig::Impl::Test
## @author gvg@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Test 

=head1 SYNOPSIS

  use Test;

=head1 DESCRIPTION

Implements object actions for Test objects

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::Test
#   Test
# --------------------------------------------------------

package CConfig::Impl::Test;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log config_result subtest);
use Params::Validate qw(:all);
use String::Random;
use Carp;

use base qw(CConfig::Impl);

# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions:
#   object is a singleton, assumes only one rule will be calling this object at a time
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self = $class->SUPER::new(KEYS => [qw(name)],);

    $self;
}

# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage:
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self = shift;
    my $item = shift;
    $Log->enter("action $item->{ACTION}");
    my $me = (caller(0))[3];
    my (%aargs) = validate(@_, {NUM => {type => SCALAR}});

    my @result;
    my $strict = 0;

  SWITCH: for ($item->{'ACTION'}) {
        /strict/i && do {
            die config_result("test component does not support strict clause");
        };
        /find/i && do {
            die config_result("test component does not support find clause");
        };
        /wait/i && do {
            my $testattr = $item->{WITH}->find_attr("name");
            my %args     = $item->{WITH}->to_hash(FORMAT => "create");
            my @names    = split(",", $args{name});
            $Log->debug("Waiting for the test(s) :" . join(" ", @names));
            foreach my $obj (@{$testattr->LIST->DATA}) {
                if (grep { /^$obj->{name}$/ } @names) {
                    $Log->debug(
                            "Adding test " . $obj->{name} . " for it to wait.");

                    #
                    # Set the wait=true so post() can wait on them since this is wait action.
                    #
                    $obj->{wait} = "true";
                    $self->CREATED_OBJS->add($obj);
                    $self->RESULT_OBJS->add($obj);
                }
            }
            last SWITCH;
        };
        /create/i && do {
            my %args = $item->{WITH}->to_hash(FORMAT => "create");

            # compute/default object keys
            if (!defined($args{'name'})) {

                # if obj name not supplied then fabricate an unused one
                do {
                    $args{'name'} =
                      "test" . $self->{'RGEN'}->randpattern("nnn");
                  } until (   !$self->RESULT_OBJS->contains(\%args)
                           && !$self->FILER_OBJS->contains(\%args));
            }

            #
            # If user has not specififed anything for "wait", set to "true" by default.
            # This will make user aware of it immediately and correct it as necessary.
            # If we set this to 'false' as default, then many subtests can be started in
            # the background, but all will exit when parent CCL exits with out "wait test" rule.
            # We do not want to assume, user has 'wait' rule set subsequently.
            #
            if (!defined($args{'wait'})) { $args{'wait'} = "true"; }

            # create the object
            $self->check_keys(\%args);   # check that all req keys are specified

            # generate subactions for this item
            CConfig::Rule::gen_subactions($self, $item, %args);

            $self->CREATED_OBJS->add(\%args);

            my %params = %args;

            my $timeout =
                ($params{timeout})
                    ? $self->_normalize_duration($params{timeout})
                        && delete $params{timeout}
                    : 0;    # Default: No timeout

            my @subtest = ($params{script}, '-bg');
            delete $params{script};
            push(@subtest, '-client', $params{client})      && delete $params{client} if ($params{client});
            push(@subtest, '-runid', $params{name})         && delete $params{name} if ($params{name});
            push(@subtest, '-username', $params{user})      && delete $params{user} if ($params{user});
            push(@subtest, '-timeout', $timeout) if ($timeout);
            my @nate_params = $self->convert_hash_to_array(%params);

            #
            # TODO - Do we have to support code srtings? unshift(@nate_params, '--');
            #
            push(@subtest, @nate_params);

            $Log->comment("Starting test: " . join(",", @subtest));
            my $subtest;
            eval { $subtest = subtest(@subtest); };
            if ($@) {
                die "creating test failed: $@\n";
            }
            $args{subtest} = $subtest;
            $self->RESULT_OBJS->add(\%args);

            last SWITCH;
        };
        croak "$me :ERROR: Unknown action " . $item->{'ACTION'} . " for object " . $item->{'OBJECT'};
    }

    $Log->exit();
    return 1;
}

# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self = shift;

    # build result set
    my $resultobjs = $self->RESULT_OBJS;

    return $resultobjs->DATA;
}

# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage:
#   $obj->post()
#
# Parameters:
#       self - object instance
# -----------------------------------------------
sub post {
    my $self = shift;
    $Log->enter();
    my @result = ();

    if ($self->CREATED_OBJS->size()) {

        # print "\n\n\n\n CREATED OBJS DATA: ".Dumper($self->CREATED_OBJS->DATA);

        foreach my $obj (@{$self->CREATED_OBJS->DATA}) {
            my $name = $obj->{name};
            next
              unless $self->CREATED_OBJS->contains($obj)
            ;    # did we create this object?

            if ($obj->{wait} ne "true") {
                $Log->comment(
                         "Subtest $name is requested to run in the background");

                #
                # We do not have to do anything here.
                #
            } else {
                my $subtest = $obj->{subtest};
                if ($subtest->is_alive()) {
                    $Log->comment(
                        "Subtest $name is still running. Wait for this job to complete"
                        );
                    $subtest->wait_finish();
                } else {
                    $Log->comment("Subtest $name is already completed");
                }
                my @results = $subtest->own_results();
                $Log->comment("Subtest $name produced the following results: "
                              . join(" ", @results));
            }
        }
    }

    # log undo actions
    #
    # tests are supposed to die, hence no undo required here?
    # NATE repear service automagically kills children promptly and hence I do not think
    # we have to make any special undo operations done in here.
    #
    $Log->exit();
    return 1;
}

sub _normalize_duration {
    my $self     = shift;
    my $duration = shift;
    my $duration_target;

    $Log->enter("duration=$duration");

    # Normalize the duration string into numerical seconds
    if ($duration =~ /^\d+[Hh]$/) {    # if value is hours
        $duration_target = $duration * 3600;
    } elsif ($duration =~ /^\d+[Dd]$/) {    # if value is days
        $duration_target = $duration * 86400;
    } elsif ($duration =~ /^\d+[Mm]$/) {    # if value is minutes
        $duration_target = $duration * 60;
    } elsif ($duration =~ /^\d+[Ss]$/) {    # if value is seconds
        $duration_target = $duration;
    } elsif ($duration =~ /^\d+$/) {
        $duration_target = $duration * 60;    # convert to seconds
    } elsif (defined($duration) && $duration !~ /^\s*$/) {
        die config_result(
            "Could not parse the DURATION parameter. The value was '$duration'."
            );
    }

    $Log->exit("return $duration_target");
    return $duration_target;
}

sub convert_hash_to_array {
    my ($self, %params) = @_;
    $Log->enter();

    my @array = ();
    while (my ($param, $value) = each %params) {
        if (ref($value) eq "ARRAY") {
            $value = join(",", @$value);
        }
        push(@array, "$param=$value");
    }
    $Log->exit("return @array");
    return @array;
}

=head1 FUNCTIONS



=head2 new

    create new object

=head3 Assumptions

    object is a singleton, assumes only one rule will be calling this object at a time

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(<item>)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut

1;
