# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CConfig/Impl/Volume.pm#1 $

## @summary CConfig::Impl::Volume
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Volume 

=head1 SYNOPSIS

  use Volume 

=head1 DESCRIPTION

Implements object actions for Volume objects

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::Volume
#   Volume
# --------------------------------------------------------

package CConfig::Impl::Volume;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use String::Random;
use CConfig::Job;
use Carp;
use NACL::C::Volume;
use NACL::STask::Volume;

use base qw(CConfig::Impl);

# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self=$class->SUPER::new(
        KEYS=>[qw(vserver volume)],
        DEFAULT=>[qw(volume)]
    );

    $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM   => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result;
    my $strict=0;

    my $subtest = $self->subtest_default( num => $aargs{NUM}, object => $item->{'OBJECT'} );

    SWITCH: for ($item->{'ACTION'}) {
            /strict/i    && do { $strict=1; };
            /find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"find",NUM=>$aargs{'NUM'});
                # remove -forground option, its not valid on show, only on create
                if ($args{'foreground'})  { delete $args{'foreground'}; }

                # If we specify size as 20MB, when we attempt to search for it, 
                # 20971520 won't match with 20MB and we will attempt to create the
                # same volume again.
                if ($args{'size'} && ($args{'size'} =~ /B$/))  { $args{'size'} = $self->convertByteString( $args{'size'} ); }

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( disklist mirror-disks ) ], args => \%args );

                # work-around for BURT 605259 where ZAPI fetch dies because volume-options-list-info
                # must be called in vserver context, but we're calling it fetch to get all volumes and
                # not passing in any vserver

                eval {
                   @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, apiset_must => { interface => 'CLI' }, filter => \%args} );
                };
                if ($@) {
                    @result=();
                    die $@ if ( !$self->noEntries($@) || $strict); 
                }

                my %sargs=$item->{WITH}->to_hash(FORMAT=>"create");
                if ($sargs{'foreground'}) { delete $sargs{'foreground'}; }
                my $num=0;
                foreach my $r (CConfig::Impl::shuffle(@result)) { 
                    # if we haven't picked this obj already then select it
                    if (!$self->RESULT_OBJS->contains($r)) { 
                        $num++;
                        $self->RESULT_OBJS->add($r);
                        $Log->debug(sub {"$me: found ".$r->{'vserver'}.":".$r->{'volume'}});

                        # generate subactions for this item, passing in primary keys from 'show'
                        CConfig::Rule::gen_subactions($self,$item,%sargs,$self->RESULT_OBJS->getRequired($r));

                        if ( $aargs{'LIMIT'} && ($aargs{'LIMIT'} == $num ) ) {
                            last SWITCH;
                        } elsif ( $aargs{'NUM'} ne "ALL" ) {
                            last SWITCH;
                        }
                    }
                }

                last SWITCH if $aargs{'NUM'} eq "ALL" || $strict;
            };
            /create|find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                # compute/default object keys
                if (!defined($args{'vserver'})) { 
                    $args{'vserver'}="vs0"; 
                    # In case someone wants to refer to the auto-generated vserver
                    my $with = $self->resolveAttrList( key => 'vserver', value => $args{vserver}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                }

                if (!defined($args{'volume'})) {

                    my $defPattern = "vol%SSS";
                    if ($args{'-default-autogen-pattern'}) { 
                        $defPattern = $args{'-default-autogen-pattern'}; 
                        delete $args{'-default-autogen-pattern'}; 
                    }

                    if (!$self->FILER_OBJS_LOADED) {
                        # work-around for BURT 605259 where ZAPI fetch dies because volume-options-list-info
                        # must be called in vserver context, but we're calling it fetch to get all volumes and
                        # not passing in any vserver
                        eval { $self->FILER_OBJS->setRows($self->call_nacl_component_routine( routine => 'fetch', args => {apiset_must => { interface => 'CLI' }})); };
                        $self->FILER_OBJS_LOADED(1);
                    }
                    # if obj name not supplied then fabricate an unused one
                    do {
                        $args{'volume'}=$self->gen_string( pattern => $defPattern );
                    } until (!$self->RESULT_OBJS->contains(\%args) && !$self->FILER_OBJS->contains(\%args));

                    # In case someone wants to refer to the auto-generated volume
                    my $with = $self->resolveAttrList( key => 'volume', value => $args{volume}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                } 

                # Used to be a check for aggregate, but infinivols don't require one 

                # create the object
                $self->check_keys(\%args); # check that all req keys are specified

                # generate subactions for this item
                CConfig::Rule::gen_subactions($self,$item,%args);

                # Add object to created objects list record undo actions prior to creating the
                #   object in case the object create times out
                $self->CREATED_OBJS->add(\%args);
                $self->RESULT_OBJS->add(\%args);

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( disklist mirror-disks ) ], args => \%args );

                if (!$args{'method-timeout'}) { $args{'method-timeout'} = 120; }

                $self->add_undo( cmd => "delete", args => \%args );

                $self->call_nacl_component_routine( subtest => $subtest, routine => 'create', args => \%args );

                last SWITCH;
            };
            /quota_modify/i    && do { 
                my %args=$item->{WITH}->to_hash();

                $self->add_undo( component_name => 'NACL::C::Quota', cmd => "modify", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, component_name => 'NACL::C::Quota',
                                                    routine => 'modify', args => \%args );
                last SWITCH;
            };
            /quota_on/i    && do { 
                my %args=$item->{WITH}->to_hash();
                $self->add_undo( component_name => 'NACL::C::Quota', cmd => "off", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, component_name => 'NACL::C::Quota',
                                                    routine => 'on', args => \%args );
                last SWITCH;
            };
            /quota_start/i    && do { 
                my %args=$item->{WITH}->to_hash();

                $self->add_undo( component_name => 'NACL::C::Quota', cmd => "stop", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, component_name => 'NACL::C::Quota',
                                                    routine => 'start', args => \%args );
                last SWITCH;
            };
            /modify/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( disklist mirror-disks ) ], args => \%args );

                $self->add_undo( cmd => "modify", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'modify', args => \%args );
                last SWITCH;
            };
            /mount/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                $self->add_undo( cmd => "unmount", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'mount', args => \%args );
                last SWITCH;
            };
            /efficiency_on/i    && do { 
                my %args=$item->{WITH}->to_hash();

                # For some reason, NACL::C::VolumeEfficiency only lists 'path' as a primary key not volume
                # If we don't pass in path, the undo logic will not work properly
                # And of course Ontap doesn't want to see BOTH volume and path, so we have to delete volume
                if ( !exists $args{path} && exists $args{volume} ) {
                    $args{path} = '/vol/'.$args{volume};
                    delete $args{volume};
                }

                $self->add_undo( component_name => 'NACL::C::VolumeEfficiency', cmd => "off", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, component_name => 'NACL::C::VolumeEfficiency',
                                                    routine => 'on', args => \%args );
                last SWITCH;
            };
            /efficiency_config/i    && do { 
                my %args=$item->{WITH}->to_hash();

                my $result=$self->call_nacl_component_routine( routine => 'fetch', 
                                                               component_name => 'NACL::C::VolumeEfficiency',
                                                               args => {allow_empty => 1, 
                                                                        filter => { vserver => $args{vserver},
                                                                                    volume  => $args{volume} }} );
                # At some point they changed from enable to true
                # Comment explaining the undo args cloning is below where we call add_undo()
                my $compression_change_present = 0;
                my $policy_change_present      = 0;
                my $undo_args        = Storable::dclone(\%args);
                my $undo_policy_args = Storable::dclone(\%args);;
                foreach my $key (keys %args) {
                    if ( $key =~ /compression/ ) {
                        $compression_change_present = 1;
                        delete $undo_policy_args->{$key};
                        if ( ($args{$key} =~ /enable/) &&
                             ($result->$key() =~ /true|false/ ) ) {
                            $args{$key} = 'true';
                        } elsif ( ($args{$key} =~ /disable/) &&
                                  ($result->$key() =~ /true|false/ ) ) {
                            $args{$key} = 'false';
                        } elsif ( ($args{$key} =~ /true/) &&
                                  ($result->$key() =~ /able/ ) ) {
                            $args{$key} = 'enable';
                        } elsif ( ($args{$key} =~ /false/) &&
                                  ($result->$key() =~ /able/ ) ) {
                            $args{$key} = 'disable';
                        }
                    }

                    if ( $key =~ /^policy$/ ) {
                        $policy_change_present = 1;
                    }
                }

                # For some reason, NACL::C::VolumeEfficiency only lists 'path' as a primary key not volume
                # If we don't pass in path, the undo logic will not work properly
                # And of course Ontap doesn't want to see BOTH volume and path, so we have to delete volume
                if ( !exists $args{path} && $result->path() ) {
                    $args{path} = $result->path();
                    delete $args{volume};
                }

                # I think that if you do the following:
                #
                #     volume efficiency modify -compression true -policy pol_VS1 -vserver VS1 -path /vol/ksink_src1 -inline-compression true
                #
                # the original policy was "default" and the original compression states were false.
                # Ontap first changes the policy (in this case to one that supports compression), and THEN it changes the 
                # compression states from false to true (which is ok).
                # But when I try to undo the command all in one shot (i.e.
                #
                #     volume efficiency modify -compression false -policy default -vserver VS1 -path /vol/ksink_src1 -inline-compression false
                #
                # Ontap again, first attempts to change the policy to default, but default doesn't support compression, so the undo fails
                # So I have to first undo the compression setting and then the policy name.   Looks backwards here, but that
                # is because undos are executed in reverse order.

                my %undo_subtest = ();
                if ( $policy_change_present && $compression_change_present ) {
                    # If we need to split these up into 2 different commands, we have to do them serially
                    $undo_subtest{subtest} = 0;
                    $self->add_undo( component_name => 'NACL::C::VolumeEfficiency', cmd => "modify", args => $undo_policy_args, %undo_subtest );
                    delete $undo_args->{policy};
                }
                $self->add_undo( component_name => 'NACL::C::VolumeEfficiency', cmd => "modify", args => $undo_args, %undo_subtest );
                $self->call_nacl_component_routine( subtest => $subtest, component_name => 'NACL::C::VolumeEfficiency',
                                                        routine => 'modify', args => \%args );
                last SWITCH;
            };
            /efficiency_start/i    && do { 
                my %args=$item->{WITH}->to_hash();

                # For some reason, NACL::C::VolumeEfficiency only lists 'path' as a primary key not volume
                # If we don't pass in path, the undo logic will not work properly
                # And of course Ontap doesn't want to see BOTH volume and path, so we have to delete volume
                if ( !exists $args{path} && exists $args{volume} ) {
                    $args{path} = '/vol/'.$args{volume};
                    delete $args{volume};
                }

                $self->add_undo( component_name => 'NACL::C::VolumeEfficiency', cmd => "stop", args => \%args );

                $self->call_nacl_component_routine( subtest => $subtest, component_name => 'NACL::C::VolumeEfficiency',
                                                    routine => 'start', args => \%args );
                last SWITCH;
            };
            /snapshot_autodelete/i    && do { 
                my %args=$item->{WITH}->to_hash();
                $self->add_undo( component_name => 'NACL::C::VolumeSnapshotAutodelete', cmd => "modify", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, component_name => 'NACL::C::VolumeSnapshotAutodelete',
                                                    routine => 'modify', args => \%args );
                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}

# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;
    my @result=();

    # update the result objects with attributes from the filer, only need to do this for created objects
    my $resultobjs=$self->RESULT_OBJS;
    if ($self->CREATED_OBJS->size()) {
        eval {
           # work-around for BURT 605259 where ZAPI fetch dies because volume-options-list-info
           # must be called in vserver context, but we're calling it fetch to get all volumes and
           # not passing in any vserver
           @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, apiset_must => { interface => 'CLI' }, filter => { $self->CREATED_OBJS->slicekeys() }} );
        };
        if ($@) {
            @result=();
            die $@ unless $self->noEntries($@);
        }
        foreach my $r (@result) { 
            next unless $self->CREATED_OBJS->contains($r); # did we create this object?
            $resultobjs->update($r);
        }
    }
    return $resultobjs->DATA;
} 


# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       self - object instance
# -----------------------------------------------
sub post {
    my $self=shift;
    
    # generate undo actions for created/modified objects
    $self->gen_undo_commands();

    # Make undo wait for background jobs to complete 
    $CConfig::TXMGR->push( 'CConfig::Subtest::wait_finish()'); 

    # wait for background jobs to complete 
    CConfig::Subtest::wait_finish();

    return 1;
} 

# -----------------------------------------------
# Function: do_purge
#
#   Static method which overrides the generic CConfig::Impl::do_purge()
#   method which just calls the NACL Component delete.
#
#   This gives you an opportunity to do more than just a simple
#   delete (in this case, essentially do what Server volume_purge() 
#   did i.e. unmount, offline, delete).
#
# Usage:
#   do_purge( "CConfig::Impl::Volume", command_interface => $ci,
#             vserver => "vs0", volume => "myvol" )
#
# Parameters:
#       class - class name, not directly used, but the general mechanism
#               used for this call requires this method signature.
#       command_interface - NACL::C::Node object
#
# -----------------------------------------------
sub do_purge {
    my $class = shift;
    my %opt = ( 'command_interface' => undef,
                @_ );

    my %args;
    $args{command_interface} = $opt{command_interface};
    delete $opt{command_interface};
 
    if ( $opt{'method-timeout'} ) {
        $args{'method-timeout'} = $opt{'method-timeout'};
        delete $opt{'method-timeout'};
    }
    CConfig::Impl::Volume->filter_out_invalid( component_name => CConfig::Impl::Volume::nacl_component_classname(), cmd=> 'show', args => \%opt );
    # Note: cdefs for show list foreground as an optional option, but it is not valid
    if ( exists $opt{foreground} ) {
        delete $opt{foreground};
    }

    my $retries=3;
RETRY:

    eval { NACL::STask::Volume->purge( %args, %opt, nacltask_verify => 1, nacltask_if_purged => 'pass' );};

    my $rc = $@;
    if ( $rc ) {
        eval { my $shortrc = (split('\n', $rc))[0]; $Log->warn($shortrc); };
        Tharn::sleep( 60 );
        $retries--;
        if ( $retries > 0 ) {
            goto RETRY;
        } else {
            die $rc;
        }
    } elsif ( $rc ) {
        die $rc
    }
}

# -----------------------------------------------
# Function: do_undo
#   Undo routine which just calls the method on the NACL
#   component.  Impls can override this
#
# Usage:
#   do_undo( 'NACL::C::Volume', '-cmd' => 'modify', %opt );
#
# -----------------------------------------------
sub do_undo {
    my $class = shift;
    my %opt = ( 'command_interface' => undef,
                '-cmd'              => undef,
                @_ );
    my $cmd = delete $opt{'-cmd'};

    if ( !defined $CConfig::LOADED{$class} ) {
        eval "use $class";
        $CConfig::LOADED{$class} = 1;
    }

    if ( ($class =~ /efficiency/i) && ($cmd =~ /stop/i) ) {
        eval {$class->$cmd( %opt );};
        if ( $@ ) {
            my $rc = $@;
            # I could go to the trouble of querying the status before attempting
            # to stop efficiency, but not really worth it...
            if ( $rc !~ /operation is currently idle/i ) {
                die $rc;
            } else {
                $Log->comment("Attempted to stop an Idle volume efficiency operation, error can be ignored");
            }
        }
    } elsif ( ($class =~ /quota/i) && ($cmd =~ /modify|on|off/i) ) {
        eval {$class->$cmd( %opt );};
        if ( $@ ) {
            my $rc = $@;
            if ( $rc !~ /job exists/i ) {
                die $rc;
            } else {
                # This burt was actually filed against snapmirror, but Peter Barry updated the burt
                # with reference to quota on/off in Oct of 2010 and it looks like this might occur with
                # vserver start/stop too...
                $Log->comment("BURT 346761 where Ontap was reporting that a job to modify quota already exists");
                use NACL::CS::Job;
                # This is strictly here for debugging purposes
                my @jobs = NACL::CS::Job->fetch( command_interface => $opt{command_interface}, allow_empty => 1 );
                die $rc;
            }
        }
    } else {
        $class->$cmd( %opt );
    }
}


# -----------------------------------------------
# Function: nacl_component_classname
#   Static method which returns the primary NACL
#   component classname for this implementation
#
# Usage:
#   my $nacl_c_classname = nacl_component_classname();
#
# -----------------------------------------------
sub nacl_component_classname {
    return "NACL::C::Volume";
}

# -----------------------------------------------
# Function: cdef_root
#   Return cdef_root for this implementation.
#
# Usage:
#   my $cdef_root = cdef_root( component_name => "NACL::C::VolumeEfficiency" );
#
#  Note: This is used for parameter checking until
#        NACL supports an "ignore_extra_options" flag
#        and possibly for CCL simulation.
#        Since Volume has multiple NACL components associated with it,
#        we look to the component_name option to distinguish
# -----------------------------------------------
sub cdef_root {
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( component_name => undef,
                @_ );
    if ( $opt{component_name} ) {
        if ( $opt{component_name} =~ /^NACL::C::(VolumeEfficiency|Efficiency)$/ ) {
            return "volume_efficiency";
        }
        if ( $opt{component_name} eq 'NACL::C::VolumeSnapshotAutodelete' ) {
            return "volume_snapshot_autodelete";
        }
        if ( $opt{component_name} =~ /^NACL::C::Quota$/ ) {
            return "volume_quota";
        }
    }

    return "volume";
}

=head1 FUNCTIONS



=head2 new

    create new object

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
