# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CConfig/Impl/Vserver.pm#1 $

## @summary CConfig::Impl::Vserver
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Vserver 

=head1 SYNOPSIS

  use Vserver 

=head1 DESCRIPTION

Implements object actions for Vserver objects

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::Vserver
#   Vserver
# --------------------------------------------------------

package CConfig::Impl::Vserver;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use String::Random;
use Carp;

use NACL::C::Vserver;
use NACL::C::RoutingGroups;

use base qw(CConfig::Impl);

# Timeout if user doesn't provide us with one
our $default_timeout = 120;

# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self=$class->SUPER::new(
        KEYS=>[qw(vserver)],
    );

    $self;
}

# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM   => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result;
    my $strict=0;

    my $subtest = $self->subtest_default( num => $aargs{NUM}, object => $item->{'OBJECT'} );

    SWITCH: for ($item->{'ACTION'}) {
            /strict/i    && do { $strict=1; };
            /find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"find",NUM=>$aargs{'NUM'});

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( ns-switch nm-switch aggrlist aggr-list max-volumes allowed-protocols disallowed-protocols ) ], args => \%args );

                eval {
                    @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => \%args} );
                };
                if ($@) {
                    @result=();
                    die $@ if ( !$self->noEntries($@) || $strict); 
                }

                my %sargs=$item->{WITH}->to_hash(FORMAT=>"create");
                my $num = 0;
                foreach my $r (CConfig::Impl::shuffle(@result)) { 
                    # if we haven't picked this obj already then select it
                    if (!$self->RESULT_OBJS->contains($r)) {
                        $num++;
                        $self->RESULT_OBJS->add($r);

                        # generate subactions for this item, passing in primary keys from 'show'
                        CConfig::Rule::gen_subactions($self,$item,%sargs,$self->RESULT_OBJS->getRequired($r));

                        if ( $aargs{'LIMIT'} && ($aargs{'LIMIT'} == $num ) ) {
                            last SWITCH;
                        } elsif ( $aargs{'NUM'} ne "ALL" ) {
                            last SWITCH;
                        }
                    }
                }
                last SWITCH if $aargs{'NUM'} eq "ALL" || $strict;
            };
            /create|find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                # compute/default object keys
                if (!defined($args{'vserver'})) {

                    my $defPattern  = "vs%SSS";
                    if ($args{'-default-autogen-pattern'}) {
                        $defPattern = $args{'-default-autogen-pattern'};
                        delete $args{'-default-autogen-pattern'};
                    }

                    # if obj name not supplied then fabricate an unused one
                    if (!$self->FILER_OBJS_LOADED) {
                        eval { $self->FILER_OBJS->setRows($self->call_nacl_component_routine( routine => 'fetch')); };
                        $self->FILER_OBJS_LOADED(1);
                    }
                    # if obj name not supplied then fabricate an unused one
                    do {
                        $args{'vserver'}=$self->gen_string( pattern => $defPattern );
                    } until (!$self->RESULT_OBJS->contains(\%args) && !$self->FILER_OBJS->contains(\%args));

                    # In case someone wants to refer to the auto-generated value
                    my $with = $self->resolveAttrList( key => 'vserver', value => $args{'vserver'}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                } 
                # if root volume not defined then fabricate a root volume name
                if (!defined($args{'rootvolume'})) {
                    $args{'rootvolume'}=$args{'vserver'}."_rootvol";

                    # In case someone wants to refer to the auto-generated value
                    my $with = $self->resolveAttrList( key => 'rootvolume', value => $args{'rootvolume'}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                };

                # create the object
                $self->check_keys(\%args); # check that all req keys are specified

                # generate subactions for this item
                CConfig::Rule::gen_subactions($self,$item,%args);

                $self->CREATED_OBJS->add(\%args);
                $self->RESULT_OBJS->add(\%args);

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                # Note: aggrlist is only available on modify, not create
                $self->list2arrayrefs( fields => [ qw( ns-switch nm-switch aggrlist aggr-list max-volumes allowed-protocols disallowed-protocols ) ], args => \%args );

                # having issues where vserver creates will timeout when too many are
                # run in parallel, add late so children won't inherit
                if (!defined($args{'method-timeout'})) {
                    $args{'method-timeout'}=$default_timeout;
                }

                $self->add_undo( cmd => "delete", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'create', args => \%args );

                last SWITCH;
            };
            /modify/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( ns-switch nm-switch aggrlist aggr-list max-volumes allowed-protocols disallowed-protocols ) ], args => \%args );

                $self->add_undo( cmd => "modify", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'modify', args => \%args );
                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}

# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;
    my $me=(caller(0))[3];
    my @result=();

    # get filer attributes for created objects
    my $resultobjs=$self->RESULT_OBJS;
    if (@{$self->CREATED_OBJS->DATA}) {
        eval {
            @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );
        };
        if ($@) {
            @result=();
            die $@ unless $self->noEntries($@);
        }
        foreach my $r (@result) { 
            next unless $self->CREATED_OBJS->contains($r); # did we create this object?
            $resultobjs->update($r);
        }
    }

    return $resultobjs->DATA;
}

# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       self - object instance
# -----------------------------------------------
sub post {
    my $self=shift;
    my $me=(caller(0))[3];
    my @result;

    # generate undo actions for created/modified objects
    $self->gen_undo_commands();

    # Make undo wait for background jobs to complete
    $CConfig::TXMGR->push( 'CConfig::Subtest::wait_finish()');

    # wait for background jobs to complete
    CConfig::Subtest::wait_finish();

    my $num_tries = 0;
    my $num_noentries = 0;
    my $num_in_cycle = 0;
    my $print_cycle = 20;
    # wait for all created vservers to come online
    if ($self->CREATED_OBJS->size()) {
        for (my $done=0; !$done; ) {
            $done=1; # we are done when all created vservers to be running

            if ( $num_in_cycle == 0 ) {
                $Log->comment("$me: Waiting for vservers admin-state to be running");
            }

            eval {
                @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );

            };
            if ($@) {
                @result=();
                die $@ unless $self->noEntries($@); 
                $num_noentries++;
                $done=0;
            }
            foreach my $vserver (@result) {
                if ($vserver->admin_state() ne "running") { $done=0; }
            }

            Tharn::snooze(5) unless $done;
            $num_tries++;
            $num_in_cycle++;

            if ( $num_in_cycle >= $print_cycle ) {
                $num_in_cycle = 0;
            }

            # Occasionally run into an infinite loop here.  Might be if the
            # creates are done with BEST_EFFORT and they all fail.
            if ( ($num_tries > 100) || ($num_noentries > 30) ) {
                die "$me Vservers admin-state failed to change to running within timeout";
            }
        }
    }

    return 1;
}

# -----------------------------------------------
# Function: nacl_component_classname
#   Static method which returns the primary NACL
#   component classname for this implementation
#
# Usage:
#   my $nacl_c_classname = nacl_component_classname();
#
# -----------------------------------------------
sub nacl_component_classname {
    return "NACL::C::Vserver";
}

# -----------------------------------------------
# Function: cdef_root
#   Return cdef_root for this implementation.
#
# Usage:
#   my $cdef_root = cdef_root();
#
#  Note: This is used for parameter checking until
#        NACL supports an "ignore_extra_options" flag
#        and possibly for CCL simulation.
# -----------------------------------------------
sub cdef_root {
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( component_name => undef,
                @_ );

    return "vserver";
}

# -----------------------------------------------
# Function: do_purge
#
#   Static method which overrides the generic CConfig::Impl::do_purge()
#   method which just calls the NACL Component delete.
#   This gives you an opportunity to do more than just a simple
#   delete (in this case, we delete all routing groups and delete
#   the rootvolume).
#
# Usage:
#   do_purge( "CConfig::Impl::Vserver", command_interface => $ci,
#             vserver => "my_vserver" )
#
# Parameters:
#       class - class name, not directly used, but the general mechanism
#               used for this call requires this method signature.
#       command_interface - NACL::C::Node object
#
# -----------------------------------------------
sub do_purge {
    my $class = shift;
    my %opt = ( 'command_interface' => undef,
                @_ );

    # Fetch any/all routing groups that use this vserver (use server for backward compat)
    my @rgs = NACL::C::RoutingGroups->find( command_interface => $opt{command_interface},
                                            allow_empty => 1, filter => {server=> $opt{vserver}} );
    foreach my $rg ( @rgs ) {
        $rg->delete();
    }

    # Fetch the root volume for deletion
    my $state = NACL::CS::Vserver->fetch( command_interface => $opt{command_interface},
                                             filter => {vserver=> $opt{vserver}} );

    if ( $state->{rootvolume} && ( $state->{rootvolume} ne '-' )) {
        CConfig::Impl::Volume::do_purge( 'CConfig::Impl::Volume',  command_interface => $opt{command_interface},
                                         vserver=> $opt{vserver}, volume=> $state->{rootvolume} );
    }

    NACL::C::Vserver->delete( %opt );
}

=head1 FUNCTIONS



=head2 new

    create new object

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
