# -*- perl -*-
# $Id$

## @summary CConfig::Impl::VserverCifs
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

VserverCifs 

=head1 SYNOPSIS

  use VserverCifs 

=head1 DESCRIPTION

Implements object actions for VserverCifs objects
SMF Menu: vserver cifs 

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::VserverCifs
#   VserverCifs object
# --------------------------------------------------------

package CConfig::Impl::VserverCifs;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use String::Random;
use Carp;

use base qw(CConfig::Impl);

# -----------------------------------------------
# Function: new
#   create new object
#
# Usage: 
#   new(item)
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Return: 
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self=$class->SUPER::new(
        KEYS=>[qw(vserver cifs-server domain)],
        DEFAULT=>[qw(cifs-server)]
    );

    $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = volume | aggregate | cluster | vserver
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result=();
    my $strict=0;

    my $subtest = $self->subtest_default( num => $aargs{NUM}, object => $item->{'OBJECT'} );

    SWITCH: for ($item->{'ACTION'}) {
            /strict/i    && do { $strict=1; };
            /find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"find",NUM=>$aargs{'NUM'});

                # Need admin-username to create, but can't use it to find
                my %filter_args = %args;
                if ( exists $filter_args{'admin-username'} ) {
                    delete $filter_args{'admin-username'};
                }
                if ( exists $filter_args{'admin-password'} ) {
                    delete $filter_args{'admin-password'};
                }
                eval {
                    @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => \%filter_args} );
                }; 
                if ($@) {
                    @result=();
                    die $@ if ( !$self->noEntries($@) || $strict); 
                };

                my %sargs=$item->{WITH}->to_hash(FORMAT=>"create");
                my $num=0;
                foreach my $r (CConfig::Impl::shuffle(@result)) { 
                    # if we haven't picked this obj already then select it
                    if (!$self->RESULT_OBJS->contains($r)) {
                        $num++;
                        $self->RESULT_OBJS->add($r);

                        # generate subactions for this item, passing in primary keys from 'show'
                        CConfig::Rule::gen_subactions($self,$item,%sargs,$self->RESULT_OBJS->getRequired($r));

                        if ( $aargs{'LIMIT'} && ($aargs{'LIMIT'} == $num ) ) {
                            last SWITCH;
                        } elsif ( $aargs{'NUM'} ne "ALL" ) {
                            last SWITCH;
                        }

                    } 
                }
                last SWITCH if $aargs{'NUM'} eq "ALL" || $strict;
            };
            /create|find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                if (!defined($args{'vserver'})) { 
                    $args{'vserver'}="vs0"; 
                    # In case someone wants to refer to the auto-generated vserver
                    my $with = $self->resolveAttrList( key => 'vserver', value => $args{vserver}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                }
                if (!defined($args{'method-timeout'})) { $args{'method-timeout'}=120; }

                if (defined($args{'admin-username'})) { $args{'username'}=$args{'admin-username'}; delete $args{'admin-username'}; }
                if (defined($args{'admin-password'})) { $args{'password'}=$args{'admin-password'}; delete $args{'admin-password'}; }

                if (!defined($args{'cifs-server'})) {

                    # Time should result in 10 chars (num seconds since Jan 1, 1970)
                    # Make sure it is only 10 chars long, if necessary trim from the
                    # left (most significant digits won't change for years)
                   
                    my $time = time();
                    my $time_length = length( $time );
                    if ( $time_length > 10 ) {
                        my $extra = $time_length - 10;
                        $time = substr( $time, $extra, 10 );
                    }

                    my $defPattern = $time.'%(RRRR:AAAA)';
                    if ($args{'-default-autogen-pattern'}) {
                        $defPattern = $args{'-default-autogen-pattern'};
                        delete $args{'-default-autogen-pattern'};
                    }

                    if (!$self->FILER_OBJS_LOADED) {
                        eval { $self->FILER_OBJS->setRows($self->call_nacl_component_routine( routine => 'fetch')); };
                        $self->FILER_OBJS_LOADED(1);
                    }

                    # Cifs Server is a little different in that you can only create one per vserver, but
                    # the cifs server name should be unique across all vservers (probably should be unique
                    # across all vservers in a given network).  So we only want to 
                    # search on the "cifs-server" name not on any other parameters
                    my %find_args;
                    # if cifs server name not supplied then fabricate an unused one
                    do {
                        $find_args{'cifs-server'}=$self->gen_string( pattern => $defPattern );
                    } until (!$self->RESULT_OBJS->contains(\%find_args) && !$self->FILER_OBJS->contains(\%find_args));

                    $args{'cifs-server'}=$find_args{'cifs-server'};

                    # In case someone wants to refer to the auto-generated value
                    my $with = $self->resolveAttrList( key => 'cifs-server', value => $args{'cifs-server'}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));

                    # We don't want these getting back in here
                    if (defined($args{'admin-username'})) { delete $args{'admin-username'}; }
                    if (defined($args{'admin-password'})) { delete $args{'admin-password'}; }

                } else {
                    my $cfis_server_len = length( $args{'cifs-server'} );
                    if ( $cfis_server_len > 15 ) {
                        $Log->warn("Impl::VserverCifs: \"cifs-server\" is greater than 15 chars, will be rejected by the filer.");
                        $Log->warn("                   It is better to not specify \"cifs-server\" at all and let CCL default this.");
                    }

                    foreach my $node ( @{$CConfig::VARS{'@NODES'}->LIST->DATA} ) {
                        my $cifs_prefix = $node->{_cifs_prefix_};
                        if ( $args{'cifs-server'} =~ /$cifs_prefix/i ) {
                            $Log->warn("Impl::VserverCifs: Using NODE{_cifs_prefix_} to create \"cifs-server\" is no longer recommended.");
                            $Log->warn("                   It is better to not specify \"cifs-server\" at all and let CCL default this.");
                            last;
                        }
                    }
                }

                # CIFS server must be uppercase
                $args{'cifs-server'} = uc $args{'cifs-server'};

                # create the object
                $self->check_keys(\%args); # check that all req keys are specified

                # Server supported -password, NACL::C supports "password"
                # CDEFs don't allow password so we have to save
                # before we call filter_out_invalid()
                my $username;
                my $password;
                my $reuse_machine_account;
                if ( $args{password} )              { $password              = $args{password}; }
                if ( $args{username} )              { $username              = $args{username}; }
                if ( $args{reuse_machine_account} ) { $reuse_machine_account = $args{reuse_machine_account}; }

                # Eventually NACL will permit extra options to be ignored and this can be removed
                $self->filter_out_invalid( component_name => nacl_component_classname(), cmd=> 'create', args => \%args );

                if ( $password )              { $args{password}              = $password; }
                if ( $username )              { $args{username}              = $username; }
                if ( $reuse_machine_account ) { $args{reuse_machine_account} = $reuse_machine_account; }

                # generate subactions for this item
                CConfig::Rule::gen_subactions($self,$item,%args);

                $self->CREATED_OBJS->add(\%args);
                $self->RESULT_OBJS->add(\%args);

                # add_undo generally filters out create args which are not valid for delete, however
                # username and password are really NACL::C::VserverCifs options and not cdef options
                # so we have to tell add_undo() not to filter them out.
                $self->add_undo( cmd => "delete", args => \%args, do_not_filter_out => [qw( username password )]  );

                # Were we to let call_nacl_component_routine filter, since the CDEFS don't specify "username|password"
                # they would be removed from %args
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'create', filter_out_invalid => 0, args => \%args );
 
                last SWITCH;
            };
            /options/i    && do { 
                my %args=$item->{WITH}->to_hash();

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( wins-servers ) ], args => \%args );

                $self->add_undo( component_name => 'NACL::C::VserverCifsOptions', cmd => "modify", args => \%args );

                $self->call_nacl_component_routine( subtest => $subtest, component_name => 'NACL::C::VserverCifsOptions', 
                                                    routine => 'modify', args => \%args );
                last SWITCH;
            };
            /security/i  && do {
                my %args=$item->{WITH}->to_hash();

                $self->add_undo( component_name => 'NACL::C::VserverCifsSecurity', cmd => "modify", args => \%args );

                $self->call_nacl_component_routine( subtest => $subtest, component_name => 'NACL::C::VserverCifsSecurity',
                                                    routine => 'modify', args => \%args );
                last SWITCH;
            };

            /modify/i && do {
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                $self->add_undo( cmd => "modify", args => \%args );

                $self->call_nacl_component_routine( subtest => $subtest, routine => 'modify', args => \%args );
                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}

# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;
    my $me=(caller(0))[3];
    my @result=();

    # get filer attributes for created objects
    my $resultobjs=$self->RESULT_OBJS;
    if ($self->CREATED_OBJS->size()) {
        eval {
            @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );
        };
        if ($@) {
            @result=();
            die $@ unless $self->noEntries($@);
        }
        foreach my $r (@result) { 
            next unless $self->CREATED_OBJS->contains($r); # did we create this object?
            $resultobjs->update($r);
        }
    }

    return $resultobjs->DATA;
}

# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       self - object instance
# -----------------------------------------------
sub post {
    my $self=shift;
    my $me=(caller(0))[3];

    # generate undo actions for created/modified objects
    $self->gen_undo_commands();

    # Make undo wait for background jobs to complete
    $CConfig::TXMGR->push( 'CConfig::Subtest::wait_finish()');

    # Wait for background jobs to complete
    CConfig::Subtest::wait_finish();

    return 1;
}

# -----------------------------------------------
# Function: nacl_component_classname
#   Static method which returns the primary NACL
#   component classname for this implementation
#
# Usage:
#   my $nacl_c_classname = nacl_component_classname();
#
# -----------------------------------------------
sub nacl_component_classname {
    return "NACL::C::VserverCifs";
}

# -----------------------------------------------
# Function: cdef_root
#   Return cdef_root for this implementation.
#
# Usage:
#   my $cdef_root = cdef_root();
#
#  Note: This is used for parameter checking until
#        NACL supports an "ignore_extra_options" flag
#        and possibly for CCL simulation.
#        Since VserverCifsOptions is also supported by this
#        Impl, we use component_name to distinguish
# -----------------------------------------------
sub cdef_root {
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( component_name => undef,
                @_ );

    if ( $opt{component_name} ) {
        if ( $opt{component_name} eq 'NACL::C::VserverCifsOptions' ) {
            return "vserver_cifs_options";
        } elsif ($opt{component_name} eq 'NACL::C::VserverCifsSecurity' ) {
            return "vserver_cifs_security";
        }
    }

    return "vserver_cifs";
}



=head1 FUNCTIONS



=head2 new

    create new object

=head3 Usage

    new(item)

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = volume | aggregate | cluster | vserver
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
