# -*- perl -*-
# $Id$

## @summary CConfig::Impl::VserverPeer
## @author frank.radish@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

VserverPeer 

=head1 SYNOPSIS

  use VserverPeer 

=head1 DESCRIPTION

Implements object actions for VserverPeer objects

=cut

# --------------------------------------------------------
# Class: CConfig::Impl::VserverPeer
#   VserverPeer
# --------------------------------------------------------

package CConfig::Impl::VserverPeer;
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log);
use Params::Validate qw(:all);
use String::Random;
use CConfig::Job;
use Carp;
use NACL::C::VserverPeer;

use base qw(CConfig::Impl);


# -----------------------------------------------
# Function: new
#   create new object
#
# Parameters:
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = vserver_peer
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# Assumptions: 
# -----------------------------------------------
sub new {
    my $class = shift;
    my @result;
    my $self=$class->SUPER::new(
        KEYS=>[qw(vserver peer-vserver)],
        DEFAULT=>[qw(vserver peer-vserver)]
    );

    $self;
}


# -----------------------------------------------
# Function: action
#   Implement object actions
#
# Usage: 
#   $obj->action(item)
#
# Parameters:
#       self - object instance
#       item - reference to the parser hash
#           item->{'ACTION'} = create | find
#           item->{'OBJECT'} = vserver_peer
#           item->{'WITH'} = CConfig::AttrList of attributes
#
# -----------------------------------------------
sub action {
    my $self=shift;
    my $item=shift;
    my $me=(caller(0))[3];
    my (%aargs) = validate(@_, {
                NUM => {type=>SCALAR},
                LIMIT => {type=>SCALAR,default=>undef}
            }
    );

    my @result;
    my $strict=0;

    my $subtest = $self->subtest_default( num => $aargs{NUM}, object => $item->{'OBJECT'} );

    SWITCH: for ($item->{'ACTION'}) {
            /strict/i    && do { $strict=1; };
            /find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"find",NUM=>$aargs{'NUM'});

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( applications ) ], args => \%args );

                eval {
                   @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => \%args} );
                };
                if ($@) {
                    @result=();
                    die $@ if ( !$self->noEntries($@) || $strict); 
                }

                my %sargs=$item->{WITH}->to_hash(FORMAT=>"create");
                my $num=0;
                foreach my $r (CConfig::Impl::shuffle(@result)) { 
                    # if we haven't picked this obj already then select it
                    if (!$self->RESULT_OBJS->contains($r)) { 
                        $num++;
                        $self->RESULT_OBJS->add($r);
                        $Log->debug(sub {"$me: found ".$r->{'vserver'}.":".$r->{'peer_vserver'}});

                        # Update with additional data from the show-all command
                        # Since we are specifying vserver and peer-vserver, better to let ontap filter this than NACL
                        my @all=$self->call_nacl_component_routine( routine => 'fetch', 
                                                                    component_name => 'NACL::CS::VserverPeerShowAll', 
                                                                    args => {allow_empty => 1, 
                                                                             filter => { vserver             => $r->{'vserver'},
                                                                                         '-let_ontap_filter' => 1,
                                                                                         'peer-vserver'      => $r->{'peer_vserver'} }} );

# FWR still problems in Ontap with "vserver peer show-all".  It doesn't
# show all of the columns when you provide keys in the query.
#local $Data::Dumper::Maxdepth=3;
#print "vserver peer showall = ".Dumper(\@all)."\n";
                        foreach my $a (@all) { 
                            if ( $a->{'peer_vserver'} && !exists $a->{'peer-vserver'} ) {
                                $a->{'peer-vserver'} = $a->{'peer_vserver'};
                            }
                            if ( $a->{'peer-vserver'} && !exists $a->{'peer_vserver'} ) {
                                $a->{'peer_vserver'} = $a->{'peer-vserver'};
                            }
                            $self->RESULT_OBJS->update($a);
                        }

                        # generate subactions for this item, passing in primary keys from 'show'
                        CConfig::Rule::gen_subactions($self,$item,%sargs,$self->RESULT_OBJS->getRequired($r));

                        if ( $aargs{'LIMIT'} && ($aargs{'LIMIT'} == $num ) ) {
                            last SWITCH;
                        } elsif ( $aargs{'NUM'} ne "ALL" ) {
                            last SWITCH;
                        }

                    } 
                }

                last SWITCH if $aargs{'NUM'} eq "ALL" || $strict;
            };
            /create|find/i    && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");
                # compute/default object keys
                if (!defined($args{'vserver'})) { 
                    $args{'vserver'}="vs0"; 
                    # In case someone wants to refer to the auto-generated vserver
                    my $with = $self->resolveAttrList( key => 'vserver', value => $args{vserver}, attr_list => $item->{WITH});
                    %args=(%args, $with->to_hash(FORMAT=>"create"));
                }

                if (!defined($args{'peer-vserver'})) { 
                    die "VserverPeer: create failed, required attribute \"peer-vserver\" not found"; 
                }

                if (!defined($args{'peer-type'}) && !defined($args{'applications'})) {
                     die "VserverPeer: create failed, one of the following required attributes \"peer-type|applications\" not found";
                }

                # create the object
                $self->check_keys(\%args); # check that all req keys are specified

                # generate subactions for this item
                CConfig::Rule::gen_subactions($self,$item,%args);

                # Add object to created objects list record undo actions prior to creating the
                #   object in case the object create times out
                $self->CREATED_OBJS->add(\%args);
                $self->RESULT_OBJS->add(\%args);

                if (!$args{'method-timeout'}) { $args{'method-timeout'} = 120; }

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( applications ) ], args => \%args );

                $self->add_undo( cmd => "delete", args => \%args, add_options => { 'force' => 'true', } );

                $self->call_nacl_component_routine( subtest => $subtest, routine => 'create', args => \%args );

                last SWITCH;
            };
            /modify/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                # NACL wants these as ARRAYREFs but CCL wants to see these as strings
                $self->list2arrayrefs( fields => [ qw( applications ) ], args => \%args );

                $self->add_undo( cmd => "modify", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'modify', args => \%args );
                last SWITCH;
            };
            /accept/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                $self->call_nacl_component_routine( subtest => $subtest, routine => 'accept', args => \%args );
                last SWITCH;
            };
            /reject/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                $self->call_nacl_component_routine( subtest => $subtest, routine => 'reject', args => \%args );
                last SWITCH;
            };
            /resume/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                $self->add_undo( cmd => "suspend", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'resume', args => \%args );
                last SWITCH;
            };
            /suspend/i && do { 
                my %args=$item->{WITH}->to_hash(FORMAT=>"create");

                $self->add_undo( cmd => "resume", args => \%args );
                $self->call_nacl_component_routine( subtest => $subtest, routine => 'suspend', args => \%args );
                last SWITCH;
            };
            croak "$me :ERROR: Unknown action ".$item->{'ACTION'}." for object ".$item->{'OBJECT'};
    };

    return 1;
}

# -----------------------------------------------
# Function: get_results
#   Return created objects
#
# Usage:
#   $obj->get_results()
#
# Parameters:
#       self - object instance
#
# Return:
#         AoH for created/found objects
# -----------------------------------------------
sub get_results {
    my $self=shift;
    my @result=();

    # update the result objects with attributes from the filer, only need to do this for created objects
    my $resultobjs=$self->RESULT_OBJS;
    if ($self->CREATED_OBJS->size()) {
        eval {
           @result=$self->call_nacl_component_routine( routine => 'fetch', args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );
        };
        if ($@) {
            @result=();
            die $@ unless $self->noEntries($@);
        } else {

            # Update with regular show data
            foreach my $r (@result) { 
                next unless $self->CREATED_OBJS->contains($r); # did we create this object?
                $resultobjs->update($r);
            }

            # Update with additional data from the show-all command
            my @all=$self->call_nacl_component_routine( routine => 'fetch', component_name => 'NACL::CS::VserverPeerShowAll', 
                                                        args => {allow_empty => 1, filter => { $self->CREATED_OBJS->slicekeys() }} );

#local $Data::Dumper::Maxdepth=3;
#print "vserver peer showall = ".Dumper(\@all)."\n";
            foreach my $r (@all) { 
                next unless $self->CREATED_OBJS->contains($r); # did we create this object?
#print "Calling update with ".Dumper($r)."\n";
                $resultobjs->update($r);
            }
        }
    }
    return $resultobjs->DATA;
} 


# -----------------------------------------------
# Function: post
#   Return created objects
#
# Usage: 
#   $obj->post()
#
# Parameters:
#       self - object instance
# -----------------------------------------------
sub post {
    my $self=shift;
    
    # generate undo actions for created/modified objects
    $self->gen_undo_commands();

    # Make undo wait for background jobs to complete 
    $CConfig::TXMGR->push( 'CConfig::Subtest::wait_finish()'); 

    # wait for background jobs to complete 
    CConfig::Subtest::wait_finish();

    return 1;
} 

# -----------------------------------------------
# Function: nacl_component_classname
#   Static method which returns the primary NACL
#   component classname for this implementation
#
# Usage:
#   my $nacl_c_classname = nacl_component_classname();
#
# -----------------------------------------------
sub nacl_component_classname {
    return "NACL::C::VserverPeer";
}

# -----------------------------------------------
# Function: cdef_root
#   Return cdef_root for this implementation.
#
# Usage:
#   my $cdef_root = cdef_root( component_name => "NACL::C::VserverPeer" );
#
#  Note: This is used for parameter checking until
#        NACL supports an "ignore_extra_options" flag
#        and possibly for CCL simulation.
# -----------------------------------------------
sub cdef_root {
    my $pkg = shift;
    my $me=(caller(0))[3];
    my %opt = ( component_name => undef,
                @_ );

    return "vserver_peer";
}

=head1 FUNCTIONS



=head2 new

    create new object

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = vserver_peer
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 action

    Implement object actions

=head3 Usage

    $obj->action(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash
    item->{'ACTION'} = create | find
    item->{'OBJECT'} = vserver_peer
    item->{'WITH'} = CConfig::AttrList of attributes
    

=back


=head2 post

    Return created objects

=head3 Usage

    $obj->post(item)

=head3 Parameters:


=over 4


=item self

    object instance

=item item

    reference to the parser hash

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
