# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CConfig/Job.pm#1 $

## @summary Rule 
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Job - Interface to filer JOB command 

=head1 SYNOPSIS

  use Job

=head1 DESCRIPTION

Abstracts interaction with filer JOB command

=cut

# --------------------------------------------------------
# Class: CConfig::Job
#   Utility routines for processing jobs
# --------------------------------------------------------

package CConfig::Job;
use strict;
use warnings;
use Data::Dumper;
use Carp;

use Tharn qw($Log);
use NATE::Log::Call qw(call_trace);
use NACL::CS::Job;
use NACL::CS::JobHistory;

# -----------------------------------------------
# Function: are_jobs_active
#   See if jobs are queued or running
#
# Usage: 
#   are_jobs_running(<joblist>)
#
# Parameters:
#       <joblist> - ref to list of job ids
#
# Return: 
#   1 if any job is still queued or running else 0
#
# Assumptions:
# -----------------------------------------------
sub are_jobs_active {
    my $call = call_trace();

    my $job_ids=shift; # list of job ids to check for
    my $node=$CConfig::Nodes[0];
    my $me=(caller(0))[3];

    my @running=();
 
    @running = NACL::CS::Job->fetch(command_interface => $node, allow_empty => 1, filter => {state=> 'Running'} );
    my $running={};
    map {$running->{$_->id()}=$_} @running;

    my @queued=();
    @queued = NACL::CS::Job->fetch(command_interface => $node, allow_empty => 1, filter => {state=> 'Queued', schedule=>'@now'} );
    my $queued={};
    map {$queued->{$_->id()}=$_} @queued;

    # if no job_ids are specified then return true if anything is waiting or running
    return 1 if !defined($job_ids) && (scalar(@queued) || scalar(@running)); 

    foreach my $job (keys(%$running)) { 
        if (grep($_==$job,@$job_ids)) { return 1; }
    }
    foreach my $job (keys(%$queued)) { 
        if (grep($_==$job,@$job_ids)) { return 1; }
    }

    return 0
}

# -----------------------------------------------
# Function: check_jobs
#   check job results
#
# Usage: 
#   are_jobs_running(<joblist>)
#
# Parameters:
#       <joblist> - list of jobs ids to check
#
# Return: 
#   1 of all jobs passed, else croaks 
#
# Assumptions:
# -----------------------------------------------
sub check_jobs {
    my $call = call_trace();
    my $job_ids=shift; # list of job ids to check for
    my $me=(caller(0))[3];
    my $node=$CConfig::Nodes[0];

    my @result=();
    my %jobs;

    @result = NACL::CS::Job->fetch(command_interface => $node, allow_empty => 1, filter => {id=>join('|',@$job_ids)});
    map {$jobs{$_->id()}=$_} @result;

    my @history = NACL::CS::JobHistory->fetch(command_interface => $node, allow_empty => 1, filter => {id=>join('|',@$job_ids)});
    foreach my $j (@history) {
        push(@result,$j) if $j->completion() ne "-";
    }
    map {$jobs{$_->id()}=$_} @result;

    JOB: foreach my $jid (@$job_ids) {
         croak "ERROR $me: Job: $jid info not found!" unless defined($jobs{$jid});
         # Some of these are Success, Successful, Succeeded, etc...  So long as it
         # looks like "Succe", we're ok...
         if ( !((defined($jobs{$jid}->{'completion'}) &&
                     ($jobs{$jid}->{'completion'} =~ /SnapMirror: done/) &&
                     ($jobs{$jid}->{'completion'} =~ /Succe/))
                 || (defined($jobs{$jid}->{'state'}) &&
                     ($jobs{$jid}->{'state'} =~ /Succe/))
                 || (defined($jobs{$jid}->{'event_type'}) &&
                     ($jobs{$jid}->{'event_type'} =~ /Succe/)))
            ) {
             # The command_interface is huge, printing it out will cause the log file to get
             # enormous
             if ( exists $jobs{$jid}->{command_interface} ) {
                 delete $jobs{$jid}->{command_interface};
             }
             croak "ERROR $me: Job: $jid Failed! \n".Dumper($jobs{$jid})."\n";
         }
     }

    return 1;
}


=head1 FUNCTIONS



=head2 are_jobs_active

    See if jobs are queued or running

=head3 Usage

    are_jobs_running(<joblist>)

=head3 Return

    1 if any job is still queued or running else 0

=head3 Parameters:


=over 4


=item <joblist>

    ref to list of job ids
    

=back


=head2 check_jobs

    check job results

=head3 Usage

    are_jobs_running(<joblist>)

=head3 Return

    1 of all jobs passed, else croaks 

=head3 Parameters:


=over 4


=item <joblist>

    list of jobs ids to check
    

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
