# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CConfig/ListIter.pm#1 $

## @summary List Iterator 
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

ListIter - a list iterator

=head1 SYNOPSIS

  use ListIter

=head1 DESCRIPTION

Represents an iterator that operates over instances of CConfig::List

=cut

# --------------------------------------------------------
# Class: CConfig::ListIter
#   List Iterator
# --------------------------------------------------------

package CConfig::ListIter;

use strict;
use warnings;
use CConfig::List;
use Params::Validate qw(:all);
use Tharn qw($Log snooze);
use Carp qw(carp cluck croak confess);
use Data::Dumper;

# -----------------------------------------------
# Function: new
#   constructor
#
# Usage: 
#   new(
#       LIST=>list,
#       LISTOP=>SELECT,
#       START=>offset|CURRENT,
#       STEP=>step,
#       CHUNK=>chunk size,
#       NUM=>num_objs,
#       DIRECTION=>RIGHT|LEFT,
#       ROOT=>root,
#       DEPTH=>depth,
#       WIDTH=>width,
#       KEY=>key,
#
# Parameters:
#       list      - CConfig::List to build the iterator around 
#       offset    - where to start returning values from (0 based) (CURRENT means start from current list position)
#               note: lists are shared among iterators, each list has a current position that is updated by the last iterator
#                   to move on the list, specifying start=current means a new list iterator will start at the current (last) list
#                   position.
#       step      - how far to move after returning each value
#       chunk     - not yet implemented
#       num_objs  - number of objects to create
#       direction - which way to move after returning each value
#           for path generation these attribtues apply
#       root      - path to use as the root of the generated file paths 
#       depth     - depth of the directory tree to generate
#       width     - width of the directory tree to generate
#       key       - optional hash index to use when returning values from the list.
#                   if the list is a list of hashes then you can specify KEY to indicate
#                   which attributes of each hash you are interested in iterating over
#                   by default you get the raw list value, be that a hash or a scalar
#       with      - optional "WITH" attributes used to filer the selection.
#
# Return: 
#   blessed reference
#
# Assumptions:
# -----------------------------------------------
sub new {
    my $invocant = shift;
    my $class    = ref($invocant) || $invocant;
    my $me=(caller(0))[3];

    my %args = validate(
        @_,
        {
           LIST     => { type  => OBJECT, optional=>1, default=>CConfig::List->new(), isa=>'CConfig::List'},
           LISTOP   => { type  => SCALAR, optional=>1, default=>"SELECT" },
           START    => { type  => SCALAR, optional=>1, default=>0 },
           STEP     => { type  => SCALAR, optional=>1, default=>1 },
           CHUNK    => { type  => SCALAR, optional=>1, default=>1 },
           NUM      => { type  => SCALAR, optional=>1, default=>1 },
           DIRECTION => { type  => SCALAR, optional=>1 },
           ROOT     => { type  => SCALAR, optional=>1  },
           DEPTH    => { type  => SCALAR, optional=>1, default=>1 },
           WIDTH    => { type  => SCALAR, optional=>1, default=>1 },
           KEY      => { type  => SCALAR, optional=>1 },
           POS      => { type  => SCALAR, optional=>1, default=>0 },
           WITH     => { type  => OBJECT, optional=>1, isa=>'CConfig::AttrList' },
           PATH_EXPANDED  => { type  => SCALAR, optional=>1, default=>0 },
        },
    );


    my $self     = {
        LIST    =>$args{'LIST'},
        LISTOP  =>$args{'LISTOP'},
        START   =>$args{'START'},
        STEP    =>$args{'STEP'},
        CHUNK   =>$args{'CHUNK'},
        NUM     =>$args{'NUM'},
        DIRECTION =>$args{'DIRECTION'},
        ROOT    =>$args{'ROOT'},
        PATH_EXPANDED => $args{'PATH_EXPANDED'},
        DEPTH   =>$args{'DEPTH'},
        WIDTH   =>$args{'WIDTH'},
        KEY     =>$args{'KEY'},
        WITH    =>$args{'WITH'},
    };
    bless($self, $class);

    # Set list starting position
    SWITCH: for($self->START()) {
        /current/i && do {
            $self->POS($self->LIST->POS);
            last SWITCH;
        };
        /first/i && do {
            $self->POS(0);
            last SWITCH;
        };
        /last/i && do {
            $self->POS($self->LIST->LASTI);
            last SWITCH;
        };
        /random/i && do {
            $self->POS(int(rand($self->LIST->LEN)));
            last SWITCH;
        };
        $self->POS($self->START);

        # if START is a number and it exceeds the list length then wrap it
        if ($self->START > $self->LIST->LASTI) {
            $self->POS( ($self->LIST->LEN==0) ? 0 : $self->START % $self->LIST->LEN);
        } 
    };    

    # default direction based on step
    $self->DIRECTION(($self->STEP()>=0) ? "RIGHT" : "LEFT") unless $self->DIRECTION(); 

    # if DIRECTION=RANDOM is specified then randomize the STEP sign
    for ($self->DIRECTION()) {
        /LEFT/i && do { $self->STEP(abs($self->STEP())*-1); };
        /RIGHT/i && do { $self->STEP(abs($self->STEP())); };
        /RANDOM/i && do { $self->STEP($self->STEP()*-1) if (rand(2)>1); };
    }

    # if ROOT is specified then we formulate a list of directory paths (make sure we only do this once, in case
    #       we are copied set a flag)
    if ($self->{'ROOT'} && !$self->PATH_EXPANDED) {
        my @paths=$self->_subpath_gen(
            LIST        =>$self->LIST()->DATA(),
            PARENT      =>$self->ROOT(),
            DEPTH       =>$self->DEPTH(),
            WIDTH       =>$self->WIDTH(),
            DIRECTION   =>$self->DIRECTION());
        # replace the list with a new list of the generated paths
        $self->{'LIST'}=CConfig::List->new(@paths);
        $self->PATH_EXPANDED(1);
    }

    return $self;
}

# -----------------------------------------------
# Function:  LIST  - get list we are iterating over
# -----------------------------------------------
sub LIST   { my $self = shift; if(@_) { $self->{'LIST'} = $_[0]; }   return $self->{'LIST'}; }
# -----------------------------------------------
# Function:  LIST  - get iterator LISTOP
# -----------------------------------------------
sub LISTOP { my $self = shift; if(@_) { $self->{'LISTOP'} = $_[0]; } return $self->{'LISTOP'}; }
# -----------------------------------------------
# Function:  START  - get iterator START
# -----------------------------------------------
sub START  { my $self = shift; if(@_) { $self->{'START'} = $_[0]; }  return $self->{'START'}; }
# -----------------------------------------------
# Function:  STEP  - get iterator STEP
# -----------------------------------------------
sub STEP   { my $self = shift; if(@_) { $self->{'STEP'} = $_[0]; }   return $self->{'STEP'}; }
# -----------------------------------------------
# Function:  CHUNK  - get iterator CHUNK
# -----------------------------------------------
sub CHUNK  { my $self = shift; if(@_) { $self->{'CHUNK'} = $_[0]; }  return $self->{'CHUNK'}; }
# -----------------------------------------------
# Function:  NUM  - get iterator NUM
# -----------------------------------------------
sub NUM    { my $self = shift; if(@_) { $self->{'NUM'} = $_[0]; }    return $self->{'NUM'}; }
# -----------------------------------------------
# Function:  ROOT  - get iterator ROOT
# -----------------------------------------------
sub ROOT   { my $self = shift; if(@_) { $self->{'ROOT'} = $_[0]; }   return $self->{'ROOT'}; }
# -----------------------------------------------
# Function:  PATH_EXPANDED  - flag to indicate we already generated the junction path
# -----------------------------------------------
sub PATH_EXPANDED    { my $self = shift; if(@_) { $self->{'PATH_EXPANDED'} = $_[0]; }   return $self->{'PATH_EXPANDED'}; }
# -----------------------------------------------
# Function:  DEPTH  - get iterator DEPTH
# -----------------------------------------------
sub DEPTH  { my $self = shift; if(@_) { $self->{'DEPTH'} = $_[0]; }  return $self->{'DEPTH'}; }
# -----------------------------------------------
# Function:  WIDTH  - get iterator WIDTH
# -----------------------------------------------
sub WIDTH  { my $self = shift; if(@_) { $self->{'WIDTH'} = $_[0]; }  return $self->{'WIDTH'}; }
# -----------------------------------------------
# Function:  DIRECTION  - get iterator DIRECTION
# -----------------------------------------------
sub DIRECTION  { my $self = shift; if(@_) { $self->{'DIRECTION'} = $_[0]; }  return $self->{'DIRECTION'}; }
# -----------------------------------------------
# Function:  KEY  - get iterator KEY
# -----------------------------------------------
sub KEY    { my $self = shift; if(@_) { $self->{'KEY'} = $_[0]; }   return $self->{'KEY'}; }
# -----------------------------------------------
# Function:  STEP  - get iterator WITH
# -----------------------------------------------
sub WITH   { my $self = shift; if(@_) { $self->{'WITH'} = $_[0]; }   return $self->{'WITH'}; }
# -----------------------------------------------
# Function:  POS  - current position in the list
#   There are two list positions, one in the iterator and one in the underlying list, when the iterator moves it modifies its position
#       and also updates the position in the underlying list.
#   When a new iterator is created on a variable with START=CURRENT the new iterator position will be initialized to the underlying list position
#       this allows us to code things like 
#               create ... with ... foreach (SELECT 4 @nodes)               # work on first 4 nodes
#               create ... with ... foreach (SELECT 4 @nodes START=CURRENT) # work on next 4 nodes
#       The foreach clause is special in that it alwys fabricates new iterators so whsereas using nodes=@nodes in a WITH clause will use the 
#           iterator associated with the @nodes variable, specifying nodes=@nodes in a foreach clause will always cause a new
#           iterator to be fabricated so iterator position changes in the foreach don't affect the underlying variable iterator position
#           but what does happen is that when the foreach clause copies the @nodes iterator and initialized it to START=0 that position will
#           be written back to the underlying list, so this has the effect of resetting the underlying list position to 0.  Iterators created
#           prior to this copy will still maintain their current iterator position but new iterators created after the foreach with START=CURRENT will
#           end up starting at 0 instead of the current list iterator position prior to the foreach (confused yet?) 
#           also note that iterators created with START=N will always position their underlying lists to N
# -----------------------------------------------
sub POS    { my $self = shift; if(@_) { $self->{POS}  = $_[0]; $self->LIST->POS($_[0]); } return $self->{POS}; }

# -----------------------------------------------
# Function: _subpath_gen
#   subpath generator
#
# Usage: 
#   _subpath_gen(LIST=><list>,PARENT=><parent>,WIDTH=><width>,DEPTH=><depth>,DIRECTION=><direction>
#
# Parameters:
#       <list> = reference to an array of strings to use to generate directories
#       <parent> = parent directory 
#       <width> = width of generated tree
#       <depth> = depth of generated tree
#       <direction> = direction to generate tree in
#
# Return: 
#   path list
#
# Example:
#   LIST=>vol[1..10] direction=DOWN parent=/ depth=2 width=3
#                                        Parent = /
#                        -----------------------------------------
#       depth=1            /                |                \
#                     vol1                 vol5              vol9
#       depth=2   /     |   \            /   |  \             /       
#                 vol2 vol3 vol4      vol6 vol7 vol8       vol10
#
#           width=1 width=2 width=3
#
#           -----------------------------------------------> Grow Parent 
#   _subpath_gen returns: /vol1 /vol1/vol2 /vol1/vol3 /vol1/vol4 /vol5 /vol5/vol6 /vol5/vol7 /vol5/vol8 /vol9 /vol9/vol10
# -----------------------------------------------
sub _subpath_gen {
    my $self=shift;
    my %args = validate(
        @_,
        {
           LIST => { type  => ARRAYREF, },
           PARENT => { type  => SCALAR, },
           WIDTH => { type  => SCALAR, default=>1, },
           DEPTH => { type  => SCALAR, default=>1, },
           DIRECTION => { type  => SCALAR, default=>"DOWN", },
        },
    );

    if ($args{'DEPTH'}==0) { return (); }

    my @plist;

    if ($args{'DIRECTION'} eq "ACROSS") {
            # width first 
            for (my $w=1; $w <= $args{'WIDTH'}; $w++) {
                my $lelement=shift(@{$args{'LIST'}});
                return @plist unless defined($lelement);
                push(@plist,$args{'PARENT'}.($args{'PARENT'} eq "/"?"":"/").$lelement);
            }
            if ($args{'DEPTH'}>0) {
                for (my $w=1; $w<=$args{'WIDTH'}; $w++) {
                    push(@plist,$self->_subpath_gen(%args,DEPTH=>$args{'DEPTH'}-1,PARENT=>$plist[$w-1]));
                }
            }
    } else {
        # depth first
        for (my $w=1; $w<=$args{'WIDTH'}; $w++) {
                my $lelement=shift(@{$args{'LIST'}});
                return @plist unless defined($lelement);
                my $dir=$args{'PARENT'}.($args{'PARENT'} eq "/"?"":"/").$lelement;
                push(@plist,$dir);
                push(@plist,$self->_subpath_gen(%args,DEPTH=>$args{'DEPTH'}-1,PARENT=>$dir));
        }
    }
    return @plist;
}

# -----------------------------------------------
# Function: get_next
#   pick next series of values from the list that this iter wraps
#   note this only returns scalar values, if the list contains a 
#   hash then KEY will be used to select an attribute.  if KEY
#   is not specified then the _default_ attrubute will be returned
#
# Usage: 
#   $iter->get_next(NOWRAP=>[1|0],NUM=><num>,POS=><pos>,RESTORE_POS=[1|0])
#
# Parameters:
#       <nowrap> = should iterator wrap around to the beginning once it hits the end? (default 1)
#       <num>    = number of object to return (default 1)
#       <pos>    = optional current position of list item to retrieve, by default the iterator
#                       will track its current position but if this is specified it will override
#                       the current iterator position, also the current iterator position will
#                       be set to this new pos
#       <restore_pos> 
#                = should get_next restore the initial position of the iterator? (default 1)
#
# Return: 
#   reference to a list of the next <num> values from the list we are iterating over
#   starting at the current iterator position
# -----------------------------------------------
sub get_next {
    my $self=shift;
    my %args = validate(
        @_,
        {
           NOWRAP => { 
              type  => SCALAR,
              optional => 1,
              default=>$self->{'ROOT'}?1:0,
           },
           NUM => {type=>SCALAR, optional=>1},
           POS => {type=>SCALAR, optional=>1},
           KEY => {optional=>1,  default=>undef},
           FOREACH => {
               type=>SCALAR, 
               optional=>1,
               default=>0
           },
           RESTORE_POS => {
               type=>SCALAR, 
               optional=>1,
               default=>1
           }
        },
    );
    my $me=(caller(0))[3];

    my $lval=[];

    # LASTI < 0 means there is no data in the list
    if ( $self->LIST->LASTI < 0 ) {
        return $lval;
    }

    SWITCH: for ($self->LISTOP()) {
       /SELECT/ && do {
            my $key=$self->KEY ? $self->KEY : '_default_';
            if ( $args{KEY} ) {
                $key = $args{KEY};
            }
#print "LISTOP = ".$self->LISTOP()." ARGS = ".Dumper(\%args)." key = $key, POS = ".$self->POS()." NUM = ".$self->{NUM}." STEP = ".$self->STEP()."\n";

            # args POS overrides POS
            $self->POS($args{'POS'}) if defined($args{'POS'});
            my $ipos=$self->POS(); # save initial list pos so we can restore it at the end
            my $with=$self->WITH(); # any WITH attributes associated with the SELECT

            my $use_current_item=1;
            
            # if $self->FOREACH is true indicates the list is in a foreach clause and therefore the 
            #   key attribute should be ignored since we need the entire hash value from the list in order to properly assign to _@
            #   which has to point to the object so that other attributes of the current foreach object can be accessed via @_ in the rule
            #   foreach attributes also specify NUM=ALL

            # The check for non-zero STEP was added to handle this case: "find ALL volume as @vols with vsroot=false foreach vserver=@vs1;"
            # We ran into the case where this would execute the following 2 commands:
            #     volume show -vserver vs086,vs538 -vsroot false
            #     volume show -vserver vs538 -vsroot false
            # Instead of:
            #     volume show -vserver vs086 -vsroot false
            #     volume show -vserver vs538 -vsroot false
            # How do we handle the case where we might WANT all of the variables concatenated with commas?
            # Also, should AttrList::to_hash() be using commas to separate options, or | ?
            if ($self->{NUM} =~/all/i || ( (defined($args{'NUM'}) && $args{'NUM'} =~ /all/i) && $self->STEP ) ) { 
                while ($self->POS <= $self->LIST->LASTI) {
                    $use_current_item=1;
                    my $v=$self->LIST->get(POS=>$self->POS);
                    # generally for FOREACH clauses we ignore the key and return a list of the underlying objects so that @_ will work
                    #   as described above, BUT if the foreach attribute key happens to point to a non-scalar value (like an array)
                    #   then we will try to return the contents of the non-scalar as the list value, this will mean that @_ will point
                    #   to the objects in the attribute specified by the key

                    if ( $with ) {
                        if ( !$self->matches_with_attributes() ) {
                            # Can't push current item, need to skip the block below and advance to next item
                            $use_current_item=0;
                        }
                    }

                    if ( $use_current_item ) {
                        # if the value is an object(or hash) we need to use the key to retrieve the attribute from the value (unless FOREACH=1)
                        #   if $v->{$key} is an array we attempt to interpolate the array into the result (this is true even if FOREACH=1)
                        my $v_ref  = ref($v);
                        if ($v_ref && (!$args{'FOREACH'} || ref($v->{$key})) ) { 
                            # if the list contains objects then select the key attribute from the object and push it onto the list
                            my $value;
                            my $exists = 0;

                            if ( $v_ref =~ /NACL::CS/ ) {
                                if ( $v->can($key) ) {
                                    $exists = 1;
                                    $value = $v->$key();
                                } elsif ( exists $v->{$key} ) {
                                    $exists = 1;
                                    $value = $v->{$key};
                                }
                            } else {
                                if ( exists $v->{$key} ) {
                                    $exists = 1;
                                    $value = $v->{$key};
                                }
                            }
                            if ( !$exists ) {
                                # Allow checks against non-existent uichange_ and _is_feature_supported strings
                                if ( ($v_ref =~ /Hostrec::Appliance/) && ($key =~ /_uichange_|_is_feature_supported/) ) {
                                    $exists = 1;
                                    $value  = 0;
                                } else {
                                    confess "$me:ERROR: List element:$v_ref has no attribute:$key defined ...\n".Dumper($v)."\n";
                                }
                            }
    
                            if (ref($value) eq "ARRAY") {
                                # if the key points to an arrayref then interpolate the arrayinto the result
                                push (@$lval,@{$value}); 
                            } elsif (ref($value) ne '') {
                                # if its a ref to something else then just punt for now
                                confess "$me:ERROR: Don't know how to handle values of type".ref($value)." for key $key ...\n".Dumper($v)."\n";
                            } else {
                                # else push the scalar
                                push (@$lval,$value); 
                            }
                        } else { 
                            # if the list does not contain objects, or, we are in a foreach then push the value or entire object
                            #   onto the return list, in this case the KEY attribute will be ignored
                            push (@$lval,$v); 
                        }
                    }

                    last if ($self->POS >= $self->LIST->LASTI);
                    # need to watch for infinite loop here, this happens when someone does a "find all ... foreach node=@NODES" since 
                    #   the with attribute for the foreach attribute will typically specify STEP=>0
                    if ($self->STEP==0) {
                        $self->advance_step(NOWRAP=>1,STEP=>1);
                    } else {
                        $self->advance_step(NOWRAP=>1);
                    }
                }
                if ( $args{RESTORE_POS} ) {
                    $self->POS($ipos); # reset list to initial position
                }
                last SWITCH;
            };
       
            # There is a simple case and a more involved case here.
            # Case 1: a "SELECT N @array"
            # Case 2: a "SELECT N @array WITH attribute=value"
            # In Case 1, $i < NUM will generally control terminating the loop
            # In Case 2, Since we might not match based on the WITH attributes, we want to make sure that
            #            we at least terminate on the size of the list.  Even for Case 1 this isn't a bad
            #            check.
            my $i=0;
            my $iterations=0;
            while ( ($iterations < $self->LIST->LEN) && ($i<$self->{NUM}) ) { 

                $iterations++;
                # see if we need to wrap
                if ($self->POS > $self->LIST->LASTI) {
                    if (!$args{'NOWRAP'}) { $self->LIST->LEN>0 ? $self->POS($self->POS % $self->LIST->LEN): $self->POS(0); }
                }

                if ($self->POS <= $self->LIST->LASTI) {
                    my $v=$self->LIST->get(POS=>$self->POS);
                    $use_current_item=1;

                    if ( $with ) {
                        if ( !$self->matches_with_attributes() ) {
                            # Can't push current item, need to skip the block below and advance to next item
                            $use_current_item=0;
                        }
                    }

                    if ( $use_current_item ) {
                        $i++;
                        my $v_ref = ref($v);
                        if ($v_ref ne '' && !$args{'FOREACH'}) {
                            # list value is a reference to a structure then we must have a key to extract an attribute
                            #   note this doesn't have to be a hash, could be an object of some other type
                            my $v_deref=$v->{$key};
                            my $exists = 0;
                            if ( $v_ref =~ /NACL::CS/ ) {
                                if ( $v->can($key) ) {
                                    $exists = 1;
                                    $v_deref = $v->$key();
                                } elsif ( exists $v->{$key} ) {
                                    $exists = 1;
                                    $v_deref = $v->{$key};
                                }
                            } else {
                                if ( exists $v->{$key} ) {
                                    $exists = 1;
                                    $v_deref = $v->{$key};
                                }
                            }

                            if ( !$exists ) {
                                # Allow checks against non-existent uichange_ and _is_feature_supported strings
                                if ( ($v_ref =~ /Hostrec::Appliance/) && ($key =~ /_uichange_|_is_feature_supported/) ) {
                                    $exists  = 1;
                                    $v_deref = 0;
                                } else {
                                    confess "$me:ERROR: List element:$v_ref has no attribute:$key defined ...\n".Dumper($v)."\n";
                                }
                            }

                            if (ref($v_deref) eq 'ARRAY') { 
                                push (@$lval,@$v_deref); 
                            } else {
                                push(@$lval,$v_deref);
                            }
                        } else { 
                            push(@$lval,$v); 
                        }
                    }
                }

                # advance one step
                $self->advance_step(NOWRAP=>$args{'NOWRAP'})
           }
           if ( $args{RESTORE_POS} ) {
               $self->POS($ipos); # reset list to initial position
           }
           last SWITCH;
        };
        confess "Unknown listop: $_";
    }

    return $lval;
}

# -----------------------------------------------
# Function: advance
#   advance iterator NUM steps
#
# Usage:
#   $obj->advance(NOWRAP=>nowrap,STEP=>step,NUM=>num)
#
# Parameters:
#           nowrap    - don't wrap to beginning of list when advancing past end of list
#           step      - amount to advance by for each step
#           num       - multiplier, how many steps to advance, specify this when advancing
#
# Return:
#   New list position
# -----------------------------------------------
sub advance {
    my $self=shift;
    my $me=(caller(0))[3];
    my %args = validate(
        @_,
        {
           NOWRAP => { 
              type  => SCALAR,
              optional => 1,
              default => 0,
           },
           STEP => { 
              type  => SCALAR,
              optional => 1,
              default => $self->STEP(),
           },
           NUM => { 
              type  => SCALAR,
              optional => 1,
              default => $self->NUM(),
           },
        }
    );
    return $self->advance_step(NOWRAP=>$args{'NOWRAP'},STEP=>$args{'STEP'},NUM=>$args{'NUM'});
};

# -----------------------------------------------
# Function: advance_step
#   advance iterator one step
#
# Usage:
#   $obj->advance_step(NOWRAP=>nowrap,STEP=>step,NUM=>num)
#
# Parameters:
#           nowrap    - don't wrap to beginning of list when advancing past end of list
#           step      - amount to advance by for each step
#           num       - multiplier, how many steps to advance, specify this when advancing
#
# Return:
#   New list position
# -----------------------------------------------
sub advance_step {
    my $self=shift;
    my $me=(caller(0))[3];
    my %args = validate(
        @_,
        {
           NOWRAP => { 
              type  => SCALAR,
              optional => 1,
              default => 0,
           },
           STEP => { 
              type  => SCALAR,
              optional => 1,
              default => $self->STEP(),
           },
           NUM => { 
              type  => SCALAR,
              optional => 1,
              default => 1,
           },
        }
    );
    # no list no data, or NUM=ALL
    return if ($self->LIST()->LEN()==0 || $args{'NUM'} =~ /all/i);

    # get list position and length
    my $lpos=$self->POS();
    my $llen=$self->LIST()->LEN();

    # advance the list position
    $lpos+=$args{'STEP'}*$args{'NUM'};
    # see if we ran off the end
    if (($lpos>$self->LIST()->LASTI() || $lpos<0) && !$args{'NOWRAP'}) { $lpos=$lpos%$llen; } 
    # update list position
    $self->POS($lpos);
    return $lpos;
}

# -----------------------------------------------
# Function: append
#   append a new value to the list referenced by this iterator
#
# Parameters:
#           item  - list item to append
#
# Return:
#   appended item
# -----------------------------------------------
sub append {
    my $self=shift;
    return $self->LIST->add(@_); 
}

# -----------------------------------------------
# Function: reset
#   reset the list pointer to the beginning of the list
#
# Parameters: 
#       none
#
# Return: 
#       none
# -----------------------------------------------
sub reset {
    my $self=shift;
    $self->POS(0);
}

# -----------------------------------------------
# Function: copy
#   copy the list iterator (but not the underlying list) so a copied iterator is another iterator
#       into the same list  (unless LIST arg is specified)
#
# Usage:
#   my $newiter=$listiter->copy(<overrides>,[DEEP=><deepflag>])
#
# Parameters:
#   overrides - any of LIST, LISTOP, START, STEP, CHUNK, NUM, DIRECTION, ROOT, DEPTH, WIDTH, KEY
#           that you want to change in the copy
#           specifying START=CURRENT will retain the current iterator POS in the copy
#   deepflag - indicates the underlying list should also be copied
#
# Return:
#   New copy of the list iterator with args overridden
# -----------------------------------------------
sub copy {
    my $self=shift;
    my %args = (@_);
    foreach my $attr (qw(LIST LISTOP START STEP CHUNK NUM DIRECTION ROOT PATH_EXPANDED DEPTH WIDTH WITH KEY)) {
        $args{$attr} = $self->{$attr} if $self->{$attr} && !defined($args{$attr});
    }
    if (defined($args{'DEEP'}) && defined($self->{'LIST'})) {
        $args{'LIST'}=$self->{'LIST'}->copy();
        delete $args{'DEEP'};
    }
    return __PACKAGE__->new(%args); 
}

# -----------------------------------------------
# Function: get
#   get the specified raw data from the list
#
# Usage: 
#   $iter->get(<pos>)
#
# Parameters:
#       <pos>    = optional position of list item to retrieve, by default the current item will be returned
#
# Return: 
# -----------------------------------------------
sub get {
    my $self=shift;
    my $pos=shift || $self->POS;
    my $me=(caller(0))[3];
    return $self->LIST->DATA->[$pos];
}

# -----------------------------------------------
# Function: matches_with_attributes
#   Does the list item at the iterator position match the given "WITH" attributes?
#
# Usage: 
#   $iter->matches_with_attributes(<pos>)
#
# Parameters:
#       <pos>    = optional position of list item to check, by default the current item will be used
#
# Return: 
# -----------------------------------------------
sub matches_with_attributes {
    my $self=shift;
    my $pos=shift || $self->POS;
    my $me=(caller(0))[3];
    my $matches=0;

    if ( $self->{WITH} ) {

        my $obj = $self->get($pos);
        if ( ref($obj) ) {
            my $windex = 0;
            my $num_matches    = 0;
            my $num_attributes = 0;
            
            foreach my $attr (@{$self->{WITH}->DATA}) {
                $num_attributes++;

                my $i = 0;
                # If this is an unresolved reference (i.e. VALUE null), don't bother, can't resolve it here
                # Probably called as a trace from Rule::_process_rule Tracing
                if ( $attr->VALUE ) {
                    my $is_ref=0;
                    if ( $attr->REF ) {
                        $is_ref=1;
                    }

                    my $ipos=$attr->VALUE->POS(); # save initial list pos so we can restore it at the end

                    # Allow wrapping if not a reference since user might decided to pick a random starting point
                    # If it is a reference, we do not want to move the iterator position
                    my $attrVal = $attr->VALUE->get_next(NOWRAP=>$is_ref,NUM=>1,RESTORE_POS=>$is_ref);
                    $i++;

                    # The intent here is to make sure we do not go through the list
                    # more than once even if we wrap.
                    WHILE: while (($attrVal && $attrVal->[0]) && ($i <= $attr->VALUE->LIST->LEN)) {
                        if ( $obj->{$attr->NAME} eq $attrVal->[0] ) {
                            $num_matches++;
                            last WHILE;
                        }

                        # If we are dealing with a ^reference, we do not want to move the iterator
                        # but instead just check the item it is pointing to
                        if ( $is_ref ) {
                            last WHILE;
                        }
                        
                        if ( $i < $attr->VALUE->LIST->LEN ) {
                            $attrVal = $attr->VALUE->get_next(NOWRAP=>$is_ref,NUM=>1,RESTORE_POS=>$is_ref);
                        }
                        $i++;
                    }

                    $attr->VALUE->POS($ipos); # reset attribute list iterator to initial position
                }
            } 

            # If SELECT matches all attributes in the WITH filter, return 1
            if ( $num_matches && ($num_matches == $num_attributes) ) {
                $matches = 1;
            }

        } else {
            confess "Cannot specify SELECT \\w WITH attributes on scalar list";
        }
    } else {
        confess "Cannot call matches_with_attributes as SELECT has no WITH attributes";
    }
    return $matches;
}

# -----------------------------------------------
# Function: to_array
#   Extracts all of the values of the iterator into a simple array.
#
# Usage: 
#   $array_ref = $iter->to_array(KEY=><key>)
#
# Parameters:
#       <key>    = optional key for objects which have keyed data
#
# Return: 
# -----------------------------------------------
sub to_array {
    my $self=shift;

    my %opts = ( KEY => undef,
                 @_  );
    my $me=(caller(0))[3];

    my @elements;
    my $len = $self->LIST->LEN;
    my $i = 0;
    my $pos = 0;
    my $orig_pos = $self->POS();

    $self->POS($pos);

    while ( $i < $len ) {
        push( @elements, $self->get_next( NOWRAP=>1, NUM=>1, RESTORE_POS=>0, KEY=>$opts{KEY} ) );
        $i++;
    }
    $self->POS($orig_pos);

   return \@elements;
}

=head1 FUNCTIONS



=head2 new

    constructor

=head3 Usage

    new(
    LIST=>list,
    LISTOP=>SELECT,
    START=>offset|CURRENT,
    STEP=>step,
    CHUNK=>chunk size,
    NUM=>num_objs,
    DIRECTION=>RIGHT|LEFT,
    ROOT=>root,
    DEPTH=>depth,
    WIDTH=>width,
    KEY=>key,

=head3 Return

    blessed reference

=head3 Parameters:


=over 4


=item direction

    which way to move after returning each value
    for path generation these attribtues apply
    root      - path to use as the root of the generated file paths 
    depth     - depth of the directory tree to generate
    width     - width of the directory tree to generate
    key       - optional hash index to use when returning values form the list.
    if the list is a list of hashes then you can specify KEY to indicate
    which attributes of each hash you are interested in iterating over
    by default you get the raw list value, be that a hash or a scalar
    key       - optional hash index to use when returning values form the list.
    

=back


=head2 LIST  - get list we are iterating over


=head2 LIST  - get iterator LISTOP


=head2 START  - get iterator START


=head2 STEP  - get iterator STEP


=head2 CHUNK  - get iterator CHUNK


=head2 NUM  - get iterator NUM


=head2 ROOT  - get iterator ROOT


=head2 PATH_EXPANDED  - flag to indicate we already generated the junction path


=head2 DEPTH  - get iterator DEPTH


=head2 WIDTH  - get iterator WIDTH


=head2 DIRECTION  - get iterator DIRECTION


=head2 WIDTH  - get iterator WIDTH


=head2 POS  - current position in the list

    The iterator keeps the list updated with the current position so that new iterators can be spawned that specify
    START=CURRENT and they will start at the same position where this iterator left off.

=head2 _subpath_gen

    subpath generator

=head3 Usage

    _subpath_gen(LIST=><list>,PARENT=><parent>,WIDTH=><width>,DEPTH=><depth>,DIRECTION=><direction>

=head3 Return

    path list
    Example:
    LIST=>vol[1..10] direction=DOWN parent=/ depth=2 width=3
    Parent = /
    -----------------------------------------
    depth=1            /                |                \
    vol1                 vol5              vol9
    depth=2   /     |   \            /   |  \             /       
    vol2 vol3 vol4      vol6 vol7 vol8       vol10
    width=1 width=2 width=3
    -----------------------------------------------> Grow Parent 
    _subpath_gen returns: /vol1 /vol1/vol2 /vol1/vol3 /vol1/vol4 /vol5 /vol5/vol6 /vol5/vol7 /vol5/vol8 /vol9 /vol9/vol10

=head2 get_next

    pick next series of values from the list that this iter wraps
    note this only returns scalar values, if the list contains a 
    hash then KEY will be used to select an attribute.  if KEY
    is not specified then the _default_ attribute will be returned

=head3 Return

    reference to a list of the next NUM values from list in LIST member

=head2 advance

    advance iterator NUM steps

=head3 Usage

    $obj->advance(NOWRAP=>nowrap,STEP=>step,NUM=>num)

=head3 Return

    New list position

=head2 advance_step

    advance iterator one step

=head3 Usage

    $obj->advance_step(NOWRAP=>nowrap,STEP=>step,NUM=>num)

=head3 Return

    New list position

=head2 append

    append a new value to the list referenced by this iterator

=head3 Return

    appended item

=head2 reset

    reset the list pointer to the beginning of the list

=head3 Return

    none

=head2 copy

    copy the list iterator (but not the underlying list) so a copied iterator is another iterator
    into the same list  (unless LIST arg is specified)

=head3 Usage

    my $newiter=$listiter->copy(<overrides>,[DEEP=><deepflag>])

=head3 Return

    New copy of the list iterator with args overridden

=head3 Parameters:


=over 4


=item overrides

    any of LIST, LISTOP, START, STEP, CHUNK, NUM, DIRECTION, ROOT, DEPTH, WIDTH, KEY
    that you want to change in the copy
    specifying START=CURRENT will retain the current iterator POS in the copy

=item deepflag

    indicates the underlying list should also be copied
    

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut

1;
