# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CConfig/ObjSet.pm#1 $

## @summary Rule 
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

ObjSet - an object set

=head1 SYNOPSIS

  use ObjSet

=head1 DESCRIPTION

A set of objects.  Each object is assumed to have a key that uniquely identifies it.  
Duplciates are not allowed

=cut

# --------------------------------------------------------
# Class: CConfig::ObjSet
#   A set of objects identified by keys
# --------------------------------------------------------

package CConfig::ObjSet;
use strict;
use warnings;
use Params::Validate qw(:all);
use Data::Dumper;
use Tharn qw($Log);
use Carp;
use Storable;

# -----------------------------------------------
# Function: new
#   create new set of objects
#
# Usage: 
#   ObjSet->new(KEYS=>keylist,REQKEYS=>reqkeys,COMPAT=>compatkeys,DEFAULT=>default_attr)  
#
# Parameters:
#       keylist - list of object keys, hash attributes to use to uniquely identify an object
#       reqkeys - list of required keys, these keys are always required to be defined in the
#           object in order for the object to be added to the set.
#           by default these will be the same as keylist 
#       compatkeys - hash of compatible keys
#       default_attr - the default attributes, when an object is added to the object set 
#           the _default_ attriute will be set to the value of this attribute
#
# Return: 
# Assumptions:
# -----------------------------------------------
sub new {
    my $class = shift;
    my (%args) = validate(@_, {
            KEYS    => {type=>ARRAYREF, optional=>0},
            REQKEYS => {type=>ARRAYREF, optional=>1},
            DEFAULT => {type=>ARRAYREF, optional=>1},
            COMPAT  => {type=>HASHREF,  optional=>1}
            }
    );
    bless {
        KEYS=>$args{'KEYS'} || [],
        REQKEYS=>$args{'REQKEYS'} || $args{'KEYS'},
        DEFAULT=>$args{'DEFAULT'},
        COMPAT=>$args{'COMPAT'} || {},
        DATA=>[],
    }, $class;
}

# selectors
sub DATA    { my $self = shift; if(@_) { $self->{DATA}    = $_[0]; } return $self->{DATA}; }
sub KEYS    { my $self = shift; if(@_) { $self->{KEYS}    = $_[0]; } return $self->{KEYS}; }
sub REQKEYS { my $self = shift; if(@_) { $self->{REQKEYS} = $_[0]; } return $self->{REQKEYS}; }
sub COMPAT  { my $self = shift; if(@_) { $self->{COMPAT}  = $_[0]; } return $self->{COMPAT}; }
sub DEFAULT { my $self = shift; if(@_) { $self->{DEFAULT} = $_[0]; } return $self->{DEFAULT}; }

# -----------------------------------------------
# Function: add
#   add a copy of the object to set 
#
# Usage: 
#   ObjSet->add(<object_in>*)
#
# Parameters:
#       <object_in> - object (hash ref) to add 
# -----------------------------------------------
sub add {
    my $self = shift;
    my $object_in = shift;
    my $me=(caller(0))[3];

    # check object has all required keys specified
    foreach my $k (@{$self->REQKEYS}) {
        if ( !exists($object_in->{$k}) ) {
            my $found = 0;
            if ( ref($object_in) =~ /NACL::CS/ ) {
                if ( $object_in->can($k) ) {
                    $found = 1;
                }
            }
            if ( !$found ) {
                croak "ERROR $me: missing required attribute: $k when adding object to set";
            }
        }
    }

    my $obj;
    my $cmd_int;
    # Storable::dclone will choke on the CommandInterface object
    if ( exists $object_in->{command_interface} ) {
        $cmd_int = delete $object_in->{command_interface};
    }

    $obj = Storable::dclone($object_in);

    if ( $cmd_int ) {
        $object_in->{command_interface} = $cmd_int;
    }

    # assign the default attribute
    my @default_fields;
    if ($self->DEFAULT) {
        @default_fields = @{$self->DEFAULT};
    } else {
        @default_fields = @{$self->KEYS};
    }
    if ( ref($obj) =~ /NACL::CS/ ) {
        my @fields;
        foreach my $item ( @default_fields ) {
            my $value;
            # Ok, the reason for all of this is that NACL::CS
            # objects are supposed to go through the MethodMaker
            # interface to access fields.  This has advantages in 
            # that it makes it easier to catch misspellings and it
            # handles the -|. to _ translations in the attribute names.
            #
            # The disadvantage is that we (CCL) add in fields 
            # which are NOT accessible via the MethodMaker interface
            # So depending on which attribute we are asking for, we have
            # to get it a different way.
            if ( $obj->can( $item ) ) {
                $value = $obj->$item();
                push( @fields, $value );
            } elsif ( exists $obj->{$item} ) {
                $value = $obj->{$item};
                push( @fields, $value );
            }
        }
        $obj->{'_default_'}=join('.',@fields); 
    } else {
        $obj->{'_default_'}=join('.',map {$obj->{$_}} @default_fields); 
    }

    # push unique objects only, gee maybe we should die on duplicates
    push(@{$self->{'DATA'}},$obj) unless $self->contains($obj); 
    return $self;
}

# -----------------------------------------------
# Function: setRows
#   set row from server result array
#
# Usage: 
#   ObjSet->setRows(<array>)
#
# Parameters:
#       <array> - ref to array of hashes to add
# -----------------------------------------------
sub setRows {
    my $self = shift;
    my $me=(caller(0))[3];

    $self->DATA([]);
    foreach my $row (@_) { 
        push(@{$self->DATA},$row); 
    }
    return $self;
}

# -----------------------------------------------
# Function: contains
#   tests if set contains an object
#
# Usage: 
#   ObjSet->contains(obj*)
#
# Parameters:
#       obj - object (hash) ref 
#
# Return: 
#   if object contained in set then returns the object ref
#   otherwise undef 
# -----------------------------------------------
sub contains {
    my $self = shift;
    my $target=shift;
    my $targetfound=0;

         foreach my $obj (@{$self->DATA}) {
             $targetfound=1;
             foreach my $k (@{$self->KEYS}) {
                 my $objval;
                 my $tarval;
                 if ( ref($obj) =~ /NACL::CS/ ) {
                     if ( $obj->can($k) ) {
                         $objval = $obj->$k();
                     }
                 } elsif ( $obj->{$k} ) {
                     $objval = $obj->{$k};
                 }

                 if ( ref($target) =~ /NACL::CS/ ) {
                     if ( $target->can($k) ) {
                         $tarval = $target->$k();
                     }
                 } elsif ( $target->{$k} ) {
                     $tarval = $target->{$k};
                 }

                 if ( $objval && $tarval ) {
                     if ( $objval ne $tarval ) {
                         $targetfound = 0;
                     }
                 }
             }
             return $obj if $targetfound;
         }

    return undef;
}

# -----------------------------------------------
# Function: update
#   update the set entry from the supplied hash
#
# Usage: 
#   ObjSet->update(obj*)
#
# Parameters:
#       obj - object (hash) ref to use to update the set with
#
# Return: 
#       true
# -----------------------------------------------
sub update {
    my $self = shift;
    my $obj = shift;
    my $me=(caller(0))[3];
    my $target=$self->contains($obj);
    croak "ERROR $me: object not found in set:".Dumper($obj) unless defined($target);

    foreach my $k (keys %$obj) { 
        # If the object is a NACL::CS object, if we ever attempt
        # to save the object with Storable::dclone(), it will fail
        # on the connection records in the command_interface object.
        if ( $k ne 'command_interface' ) {
            $target->{$k}=$obj->$k();
        }
    }
    return 1; 
}

# -----------------------------------------------
# Function: size
#   return size of object list 
#
# Usage: 
#
# Parameters:
#
# Return: 
#   size of object list
# -----------------------------------------------
sub size {
    return scalar(@{shift->DATA});
}

# -----------------------------------------------
# Function: slice
#   return a list of object key values
#
# Usage: 
#   $objset->slice(key)
#
# Parameters:
#   key - attribute to return values from
#
# Return: 
#   list of object values for the specified attribute
#       attribute->val1,val2,val3
# -----------------------------------------------
sub slice {
    my $self=shift;
    my $key=shift || "_default_";
    my @attlist;
    foreach my $attr (@{$self->DATA}) {
        if ( ref($attr) =~ /NACL::CS/ ) {
            push( @attlist, $attr->$key() );
        } else {
            push( @attlist, $attr->{$key} );
        }
    }
    return @attlist;
}

# -----------------------------------------------
# Function: slicekeys
#   return a list of object key and key value list pairs for each required attribute
#
# Usage: 
#   $objset->slicekeys()
#
# Parameters:
#
# Return: 
#   list of key and value list paris for each required attribute
#       reqattr1->val1|val2|val3
#       reqattr2->val1|val2|val3, etc
# -----------------------------------------------
sub slicekeys {
    my $self=shift;
    my @result;
    my $hcompat = $self->COMPAT;
    foreach my $k (@{$self->REQKEYS}) {
        my @data = $self->slice($k);

        my $compatkey;
        my $valstring;
        if ( $hcompat && (defined($hcompat->{$k}))) {
            $compatkey = $hcompat->{$k};
        }
        if ( !@data && $compatkey) {
            @data = $self->slice($compatkey);
        }
        $valstring = join('|',@data);
        push (@result,$k); push(@result,$valstring);
        if ( $compatkey ) {
            push (@result,$compatkey); push(@result,$valstring);
        }
    }
    return @result;
}

# -----------------------------------------------
# Function: getRequired()
#   Given a hash of fields for a command, will return a hash subset which 
#   only contains entries for the required keys.
#
# Usage: 
#   $objset->getRequired( \%show_hash )
#
# Parameters:
#
#  Hash Ref, keyed by command field names, typically from a show command.
#
# Return: 
#   Subset of the input hash corresponding to the required keys
# -----------------------------------------------
sub getRequired {
    my $self=shift;
    my $hr=shift;
    my %result;

    my $hcompat = $self->COMPAT;
    foreach my $k (@{$self->REQKEYS}) {
        my $val;
        my $compatkey;
        if ( defined $hr->{$k} ) {
            $val = $hr->{$k};
        } elsif ($hcompat && (defined($hcompat->{$k}))) {
            $compatkey = $hcompat->{$k};
            $val = $hr->{$compatkey};
        }
        $result{$k} = $val;
        if ( $compatkey ) {
            $result{$compatkey} = $val;
        }
    }
    return %result;
}


=head1 FUNCTIONS



=head2 new

    create new set of objects

=head3 Usage

    ObjSet->new(KEYS=>keylist,REQKEYS=>reqkeys,DEFAULT=>default_attr)  

=head3 Parameters:


=over 4


=item keylist

    list of object keys, hash attributes to use to uniquely identify an object

=item reqkeys

    list of required keys, these keys are always required to be defined in the
    object in order for the object to be added to the set.
    by default these will be the same as keylist 

=item default_attr

    the default attributes, when an object is added to the object set 
    the _default_ attriute will be set to the value of this attribute
    

=back


=head2 add

    add object to set 

=head3 Usage

    ObjSet->add(<obj>*)

=head3 Parameters:


=over 4


=item <obj>

    object (hash ref) to add 

=back


=head2 setRows

    set row from sever result array

=head3 Usage

    ObjSet->setRows(<array>)

=head3 Parameters:


=over 4


=item <array>

    ref to array of hashes to add

=back


=head2 contains

    tests if set contains an object

=head3 Usage

    ObjSet->contains(obj*)

=head3 Return

    if object contained in set then returns the object ref
    otherwise undef 

=head3 Parameters:


=over 4


=item obj

    object (hash) ref 
    

=back


=head2 update

    update the set entry from the supplied hash

=head3 Usage

    ObjSet->update(obj*)

=head3 Return

    true

=head3 Parameters:


=over 4


=item obj

    object (hash) ref to use to update the set with
    

=back


=head2 size

    return size of object list 

=head3 Return

    size of object list

=head2 slice

    return a list of object key values

=head3 Usage

    $objset->slice(key)

=head3 Return

    list of object values for the specified attribute
    attribute->val1,val2,val3

=head3 Parameters:


=over 4


=item key

    attribute to return values from
    

=back


=head2 slicekeys

    return a list of object key and key value list pairs for each required attribute

=head3 Usage

    $objset->slicekeys()

=head3 Return

    list of key and value list paris for each required attribute
    reqattr1->val1,val2,val3
    reqattr2->val1,val2,val3, etc

=head2 getRequired

    Given a hash of fields for a command, will return a hash subset which 
    only contains entries for the required keys.

=head3 Usage

    ObjSet->getRequired(<hashref>)

=head3 Return

    Subset of the input hash corresponding to the required keys

=head3 Parameters:


=over 4


=item <hashref>

    Hash Ref, keyed by command field names, typically from a show command.

=back

=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut

1;
