# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CConfig/Rule.pm#1 $

## @summary Rule 
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Rule - Rule Processing 

=head1 SYNOPSIS

  use Rule

=head1 DESCRIPTION

Rule processing utilities

=cut

# --------------------------------------------------------
# Class: CConfig::Rule
#   Rule processing routines
# --------------------------------------------------------

package CConfig::Rule;
use strict;
use warnings;

use CConfig; # so that legacy code that doesn't do a use CConfig will still work
use Parse::RecDescent;
use Storable qw(freeze thaw);
use Data::Dumper;
use NACL::ComponentUtils;

use Utils qw(make_host);
use Tharn;
use NATE::Inc qw(find_file);

### tool tracking code ###
our $CCL_VERSION       = "5.4";
use Sys::Hostname      qw(hostname);
use Tharn_UsageService qw(log_usage);
use NATE::OS           qw(login_name);
### tool tracking code ###

use CConfig::Impl;
use CConfig::Subtest;
use NACL::STask::Node;
use NACL::C::ClusterIdentity;
use CConfig::Impl::License;
use File::Basename;

use IO::Handle;
#use File::Data;
 
our $PARSER;      # the recdescent parser object
our $isSlave=0;   # is this a slave process

$::RD_ERRORS = 1; # Make sure the parser dies when it encounters an error
$::RD_WARN   = 1; # Enable warnings. This will warn on unused rules &c.
$::RD_HINT   = 1; # Give out hints to help fix problems.

# -----------------------------------------------
# Function: init 
#   Initialize the parser from a grammar, and clear variables
#
# Usage: 
#        CConfig::Rule::init()
#
# Parameters:
#
# Return: 
#   1=success, 0=failure 
# -----------------------------------------------
sub init {
    my $me=(caller(0))[3];
    my %args=(@_);
    $CConfig::OPTIONS->parameters(definitions=>
    [
    {
        name => "BEST_EFFORT",
        default=>0,
        description => "best effort flag - if set create errors will be ignored",
    },
    {
        name => "STRICT_FIND",
        default=>0,
        description => "strict find flag - finds don't automatically try to do a create if a resource is not found",
    },
    {
        name => "SUBTEST_PACING",
        default=>"",
        description => "Subtest Pacing.".
                       "Example 1: 2:10, only subtest when creating 2 or more objects, and subtest 10 at a time.".
                       "Example 2: 0:<anything> disables subtesting globally.".
                       "Example 3: 1:10,volume:2:12,interface:0 all implementations will subtest when creating 1 or more objects and subtest 10 at a time, except volume which will only subtest when creating 2 or more objects and will subtest 12 at a time and interface which will never subtest.",
    },
    {
        name => "DISALLOW_ZAPI",
        default=>"",
        description => "Force apiset_must to be CLI for these objects",
    },
    {
        name => "SHUFFLE",
        default=>1,
        description => "Shuffle (randomize) filer show command output",
    },
    ]);

# DISALLOW_API as of 5/2/12
#        "cifs_preferred_dc,dns,export_policy,export_policy_rule,interface,lun,lun_igroup,nfs,nis_domain,quota_policy_rule,route,routing_group,snapshot_policy,unix_user,unix_group,volume",

# Reasons for ZAPI failures
# BURT 594824: interface fails in "lun_test" & "netintf_test" & "cifs_test" because of 'status-admin', 'auto-revert', 'dns-zone', 'allow-lb-migrate'
# BURT 594824: volume fails in "vserver2_test" because of 'group' name isn't converted to integer for ZAPI
# BURT 595717: snapshot_policy modify fails in "snapshot_test" because ZAPI modify requires "enabled" ?

    $CConfig::OPTIONS->process( sfopairs=>1 );

    my $opt_hash = $CConfig::OPTIONS->get_hash();

    # Still leave default 1:10 because ZAPI is not supported for all components and
    # CLI needs to be limited to 10, and there is no way for me to know ahead of time, if ZAPI support exists
    my $cli_default_pacing   = "1:10,aggregate:1:4,interface:0,port:0,route:0,routing_group:0,snapmirror:0,vserver:0";
    my $zapi_default_pacing  = "1:50,aggregate:1:4,vserver:0,snapmirror:0";
    #my $zapi_default_pacing  = "1:10,aggregate:1:4,interface:1:50,port:0,qtree:1:50,route:1:50,routing_group:1:50,snapmirror:1:50,snapshot:1:50,volume:1:50,vserver:1:2";

    # Since this is just a default, maybe just go with existence of FILER_APISET_MUST|SHOULD_INTERFACE
    # Not going to support different pacing on different filers, if you want that, use the
    # -subtest_pacing parameter on individual rulesets

    # What if I just allowed per ruleset overrides of pacing parameters?
    if ( $opt_hash->{SUBTEST_PACING} eq "" ) {
        my $should_apiset = Tharn::param( "FILER_APISET_SHOULD_INTERFACE" );
        my $must_apiset   = Tharn::param( "FILER_APISET_MUST_INTERFACE" );
        if ( ($must_apiset  && ($must_apiset   =~ /ZAPI/i)) ||
            ($should_apiset && ($should_apiset =~ /ZAPI/i)) ) {
            $opt_hash->{SUBTEST_PACING} = $zapi_default_pacing;
            CConfig::Impl->PREFER_ZAPI( 1 );
        } else {
            $opt_hash->{SUBTEST_PACING} = $cli_default_pacing;
        }
    }

    if ( $opt_hash->{SHUFFLE} ) {
        CConfig::Impl->SHUFFLE( 1 );
    } else {
        CConfig::Impl->SHUFFLE( 0 );
    }

    my $report_args = "";
    foreach my $var ( qw( UNDO BEST_EFFORT RULESET STRICT_FIND SITE RANDOM_NODE SUBTEST_PACING DISALLOW_ZAPI ) ) {
        if ( $opt_hash->{$var} ) {
            $report_args = $report_args.$var."=".$opt_hash->{$var}." ";
        }
    }

    # Tool Tracking.  Lets us see who is calling CCL and how many times.
    log_usage( params=> { tool      => 'CCL',
                          tool_args => $report_args,
                          version   => $CCL_VERSION,
                          hostname  => hostname(),
                          username  => login_name(),
                        } );

    CConfig::Subtest::SUBTEST_PACING( $opt_hash->{SUBTEST_PACING} );
    CConfig::Impl::DISALLOW_ZAPI(     $opt_hash->{DISALLOW_ZAPI} );

    # initiaze vars
    %CConfig::VARS=();
    $CConfig::VARS{'@_'}=CConfig::ListIter->new(LIST=>CConfig::List->new());

    # initialize variables
    @CConfig::Nodes  = NACL::STask::Node->find();

    # add args to symbol table
    foreach my $var (keys %args) {
        if (defined($args{$var})) {
            if (ref($args{$var}) eq "ARRAY")  {
                $CConfig::VARS{'@'.$var}=CConfig::ListIter->new(LIST=>CConfig::List->new(@{$args{$var}}));
            } else {
                $CConfig::VARS{'@'.$var}=CConfig::ListIter->new(LIST=>CConfig::List->new($args{$var}));
            }
            push(@CConfig::RO_VARS,'@'.$var);
        }
    }

    # set @NODES = list of hostrec objects
    my $nlist=CConfig::List->new();
    my $i=0;
    my $params = $CConfig::OPTIONS->get_hash();

    # The node names coming in on the command-line can be node1;nodeattribute=value,node2
    foreach my $nodespec (@{$args{NODES}}) {
        my $nodehost=Utils::make_host($nodespec);
        my $nodename = $nodehost->{hostname};

        my $nodeid = $nodename;
        my $cifs_prefix;
        $nodeid =~ s/-/_/g;

        $cifs_prefix = $nodeid;
        $cifs_prefix =~ s/_//g;
        $cifs_prefix =~ tr/a-z/A-Z/;

        $nodehost->{'_default_'}=$nodename;
        $nodehost->{'node'}=$nodename;
        $nodehost->{'_index_'}=$i++;
        $nodehost->{'_node_id_'}=$nodeid;
        $nodehost->{'_cifs_prefix_'}=$cifs_prefix;

        my $cmd_int = CConfig::Impl::find_node_component( name => $nodename );
        my $ver_mgr = $cmd_int->get_version_manager();
 
        $nodehost->{'_v2_licensing_supported_'} = CConfig::Impl::License::is_v2_licensing_supported( node => $cmd_int );

        # Retrieving all of this takes time, so unless someone really needs it
        # I'm not going to retrieve it all

        $nodehost->{'_ontap_release_'}       = $ver_mgr->get_version_attribute( attribute => "release" );
        $nodehost->{'_ontap_change_number_'} = $ver_mgr->get_version_attribute( attribute => "change_number" );

        # takes less than a second to load the uichange strings
        my @uichanges                        = @{$ver_mgr->get_version_attribute( attribute => "uichange_strings" )};

        foreach my $uichange ( @uichanges ) {
            $nodehost->{'_uichange_'.$uichange} = 1;
        }

        my $start = time();
        my @features                        = keys( %NACL::C::_Version::Features_To_Check );

        my $last = $start;
        foreach my $feature ( @features ) {
            my $time = time();
            my $elapsed = $time - $last; $last = $time;
            $Log->debug(sub {"feature $feature took $elapsed seconds"});
            my $supp = 0;
            # If there is an error during is_feature_supported() report error, and
            # default support to 0 (i.e. not supported)
            try {
                $supp = $cmd_int->is_feature_supported( feature => $feature );
            } catch NATE::BaseException with {
                my $ex = shift;
                $Log->warn("$me: Error while executing is_feature_supported( feature => $feature ): " . $ex->text());

            };
            $nodehost->{'_is_supported_'.$feature} = $supp;
        }

        my $elapsed = time() - $start;
        # Just so I can tell if loading is_supported starts taking a significant amount of time
        # Currently takes between 1-3 seconds per node
        if ( $elapsed > 10 ) {
            print "Loading is_supported took $elapsed seconds\n";
        }

        #$nodehost->{'_ontap_branch_'}        = $ver_mgr->get_version_attribute( attribute => "branch" );
        #$nodehost->{'_ontap_build_options_'} = $ver_mgr->get_version_attribute( attribute => "build_options" );
        #$nodehost->{'_ontap_build_path_'}    = $ver_mgr->get_version_attribute( attribute => "build_path" );
        #$nodehost->{'_ontap_wafl_version_'}  = $ver_mgr->get_version_attribute( attribute => "wafl_version" );
        #$nodehost->{'_ontap_node_version_'}  = $ver_mgr->get_version_attribute( attribute => "node_version" );

        my $lifinfo;
        eval {

            # Yank these once STask::Node change goes in
            package Hostrec::Appliance;
            our %Nettype_Converter = (
                MGMT         => 'management',
                DATA         => 'data',
                CLUSTER      => 'cluster',
                CLUSTER_MGMT => 'cluster_management',
                INTERCLUSTER => 'intercluster',
                RLM          => 'rlm',
            );

            our %Field_Converter = (
                IP           => 'address',
                ADDRESS      => 'address',
                PORT         => 'interface',
                MASK         => 'netmask',
                NETMASK      => 'netmask',
                GATEWAY      => 'gateway',
                NAME         => 'name',
                MTU          => 'mtu',
            );

            package CConfig::Rule;

            $lifinfo = $cmd_int->get_hostrec_lif_info( source => 'net_records' );
        };

        if ( $@ ) {
            $Log->warn( "$me: NACL::STask::Node->get_hostrec_lif_info() failed with $@");
        } else {
            foreach my $lifkey (keys %{$lifinfo}) {
                my $hkeyprefix = '_'.lc $lifkey;
                my ($hkey, $data);
                
                my @lif_nodes;
                foreach my $entry ( @{$lifinfo->{$lifkey}} ) {
                    my $node_hkey = $hkeyprefix.'_node_';
                    foreach my $field ( qw( IP MASK GATEWAY PORT MTU NAME ) ) {
                        # Need to handle case where we don't have all of the keys?
                        # i.e. what if we are missing MTUs or MASKs?
                        if ( exists $entry->{$field} ) {
                            $hkey = $hkeyprefix.'_'.lc $field.'_';
                            $data = $entry->{$field};

                            # Hack until I fix NACL::STask::Node, which I broke :) FWR
                            if ( ($field  =~ /NAME/i) &&
                                 ($lifkey =~ /DATA/i) &&
                                 ($data   !~ /^$nodename/) ) {
                                $data = $nodename."_".$data;
                            }

                            my @elements;
                            if ( exists $nodehost->{$hkey} ) {
                                @elements = @{$nodehost->{$hkey}};
                            }
                            push( @elements, $data );
                            if ( $field =~ /IP/i ) {
                                push( @lif_nodes, $nodename );
                            }
                            $nodehost->{$hkey} = [ @elements ];
                        } else {
                            $Log->warn( "$me: Do not seem to have a definition for $lifkey $field");
                        }
                        $nodehost->{$node_hkey} = [ @lif_nodes ];
                    }
                }
            }
        }

        foreach my $pair ( @{$params->{SFOPAIRS}} ) {
            if ( $nodename eq $pair->[0] ) {
                $nodehost->{'_sfo_node_'}=0;
                last;
            } elsif ( $nodename eq $pair->[1] ) {
                $nodehost->{'_sfo_node_'}=1;
                last;
            }
        }

        $nlist->add($nodehost);
    }

    $CConfig::VARS{'@NODES'}=CConfig::ListIter->new(LIST=>$nlist);

    # initialize @clusters
    my %clusters;
    my %cluster_node_index;
    my @clusters;
    $i=0;
    # add nodes to clusters (if clusters exist)
    eval {
        foreach my $nodespec (@{$args{NODES}}) {
            my $nodehost=Utils::make_host($nodespec);
            my $nodename = $nodehost->{hostname};
            my $node = CConfig::Impl::find_node_component( name => $nodename );
            my $clusterid=NACL::CS::ClusterIdentity->fetch( command_interface => $node );
            my $clustername=$clusterid->{name};

            if (!defined($clusters{$clustername}))  {
                # create new cluster
                $cluster_node_index{$clustername}=0;
                $nodehost->{'_index_'}=$cluster_node_index{$clustername}++;
                $clusters{$clustername}={_default_=>$clustername,_index_=>$i++,nodes=>[$nodehost]};
                push(@clusters,$clusters{$clustername});
            } else {
                # add node to existing cluster
                $nodehost->{'_index_'}=$cluster_node_index{$clustername}++;
                push(@{$clusters{$clustername}->{'nodes'}},$nodehost);
            }
        }
    };

    # TODO: need to make this more specific, differentiate between cluster not found and other errors
    if ($@) {
    }

    $CConfig::VARS{'@CLUSTERS'}=CConfig::ListIter->new(LIST=>CConfig::List->new(@clusters));
    push(@CConfig::RO_VARS,'@CLUSTERS');

    # initialize transaction object
    $CConfig::TXMGR=CConfig::TXMgr->new();

    $PARSER=Parse::RecDescent->new(CConfig::Grammar::grammarDef());
}


# -----------------------------------------------
# Function: process_file 
#   Parse and execute a ruleset file
#
# Usage: 
#        CConfig::Rule::process_file(ruleset_file)
#
# Parameters:
#       ruleset_file - ruleset file name
#
# Return: 
#   instance of Parse::RecDescent
#   1=success, 0=failure 
# -----------------------------------------------
sub process_file {
    my $ruleset=shift;
    my $me=(caller(0))[3];
    my $filename;

    # Burt 473470 - Search for Rulesets in NATELIB path if not found
    #               as an absolute or relative-from-pwd path
    if ( -f $ruleset ) {
        $filename=$ruleset;
    } else {
        eval { $filename = find_file( $ruleset ); };
        if ( $@ ) {
            die "ERROR $me: Ruleset file: $filename not found";
        }
    }

    # read ruleset
    open(FILE,$filename) or die "ERROR $me: Ruleset file: $filename not found or unreadable";

    # Want to keep the name of the top level ruleset
    if ( ! $CConfig::_top_level_ruleset_ ) {
        $CConfig::_top_level_ruleset_ = basename($filename,(".rs"));
    }

    my @input=<FILE>;
    close(FILE);

    defined $PARSER->startrule("@input") || die "ERROR $me: Parser Error: $@";
}

# -----------------------------------------------
# Function: process_ruleset 
#   Parse and execute a ruleset 
#
# Usage: 
#        CConfig::Rule::process_ruleset(ruleset)
#
# Parameters:
#       ruleset - ruleset to process 
#
# Return: 
#   1=success, 0=failure 
# -----------------------------------------------
sub process_ruleset {
    my $ruleset=shift;
    my $me=(caller(0))[3];
    $PARSER->startrule($ruleset) || die "ERROR $me: Parser Error: $@";
}

# -----------------------------------------------
# Function: process_rule 
#   Parse and execute a single rule
#
# Usage: 
#        CConfig::Rule::process_rule(rule)
#
# Parameters:
#       rule - rule to process 
#
# Return: 
#   1=success, 0=failure 
# -----------------------------------------------
sub process_rule {
    my $rule=shift;
    my $me=(caller(0))[3];
    $PARSER->startrule($rule) || die "ERROR $me: Parser Error: $@";
}


# -----------------------------------------------
# Function: _process_rule
#   process a parsed rule (internal)
#
# Usage: 
#   _process_rule(item)
#
# Parameters:
#       item - reference to the parser hash
#
# Return: 
#   1=success, 0=failure 
# -----------------------------------------------
sub _process_rule {
    my $item=shift;
    my $me=(caller(0))[3];
    my $calledby=(caller(1))[3];
    my $ts_start=time;
    my @item_trace=(); # cache the %ITEM array for TEST_MODE
    @CConfig::BG_JOBS=();

    $Log->trace("$me: Rule start @ $ts_start");
    $Log->trace("$me: ACTION: ".$item->{'ACTION'}." OBJECT: ".$item->{'OBJECT'}." WITH: ".$item->{'WITH'}->to_string());

    # Tell Subtest to start a new subtest directory for this top-level rule
    if ( $calledby =~ /RecDescent/ ) {
        CConfig::Subtest::NEW_RULE(1);
    }

    $DB::single=1;

    my $impl=CConfig::Impl::create_impl($item);

    # TODO: if ($item->{'NUM'} eq "MAX") { lookup max objects } 
    # TODO: implement create all functionality
    if ($item->{'NUM'} && $item->{'NUM'} =~ /all/i && $item->{'ACTION'} =~ /create/i) {
        die "ERROR $me: 'create all' is not implemented";
    }

    #   If NUM=ref to list iterator we will use the length of the list as the NUM value
    my $NUM = $item->{'NUM'};
    if (ref($item->{'NUM'}) eq "CConfig::ListIter") { $NUM=$item->{'NUM'}->get_next(); $NUM=$NUM->[0]; } 

    if ($item->{'FOREACH'} && !$item->{'FOREACH'}->is_empty()) {
        my $f_attrs=$item->{'FOREACH'}->DATA; # list of foreach attribute references

        # Handling of foreach attributes:
        #   The foreach clause looks at its attributes like this
        #       foreach <primary attr> [<secondary attr>*]
        #
        #   To generate the loop driver the primary attribute get_next() is called ONCE to get a list of values to use to drive the foreach loop
        #       A single get_next is used to generate the list driver so that you can say foreach (SELECT 4 @abc ...) and have the foreach loop
        #       iterated over only the next 4 values from @abc
        #       By default if no iterator is specified we wrap it in (SELECT ALL @abc)
        #
        #   The primary attribute is parsed with FE_ATTRASSIGN wich ensures that NUM=>ALL by default 
        #       the prinary attribute is used to generate a WITH attribute (if it has an attribute name) with STEP=>0 so it won't increment
        #       inside the foreach. 
        #       The primary attribute itself should not increment between foreach attributes either since it has already been queried with get_next to 
        #           retrieve all the values it needs to have during the rule execution, however the with attribute generated form the primary attribute
        #           (which represents an iterator over the driver attributes) should increment between each foreach iterations
        #       It primary attribute itself should only increment after the entire rule has executed.
        #
        #   The secondary attributes are parsed with ATTRASSIGN so don't get any special default treatment
        #       they are also used to generate WITH attributes with STEP=>0 to make them constant within a foreach iteration
        #       they get incremented between foreach iterations, the STEP to use to increment the WITH iterators should be taken from the foreach iterators
        #           and both the WITH and secondary foreach iterators should be incremented after each foreach iteration
        #       at the end of the rule these attributes will already be incremented by the foreach iterations so don't need incremented again
        #
        #   FOREACH=>1 tells get_next to return whole objects rather than single attributes, we need this so that
        #   we can assign to the @_ variable
        #   We build a new iteartor out of the results of the get_next and add the foreach attribute to the WITH clause
        #   (if its named), that wey (SELECT 4 @abc) ends up becoming a shorted list of 4 elements within the WITH clause

        # primary attribute
        my $f_p_attr=$f_attrs->[0];
        my $f_p_name=$f_attrs->[0]->NAME;
        my $f_p_vals=$f_attrs->[0]->VALUE->get_next(FOREACH=>1); # this will drive the iteration
        my $f_p_w_attr;

        # if its a named attribute add it to the list of WITH attributes
        if (defined($f_p_name)) {
            $f_p_w_attr=CConfig::Attr->new(
                    $f_p_name,
                    VALUE=>CConfig::ListIter->new(LIST=>CConfig::List->new(@$f_p_vals),STEP=>0)
            );
            $item->{'WITH'}->add($f_p_w_attr);
        }

        # list of references to secondary attributes
        my @f_s_attrs=@$f_attrs[1..$#$f_attrs];

        # subsequent attribute values get added to the WITH attributes also, but here we just add them directly
        my %f_s_w_attr=();
        foreach my $f_attr (@f_s_attrs) {
            my $f_attr_name=$f_attr->NAME;
            my $f_attr_val=$f_attr->VALUE;

            # convert FOREACH attributes to regular attributes in the rule
            #   if the attribute is of the form name=value then add name=value to the WITH attributes
            #    step is 0 since foreach attributes don't iterate inside the foreach loop
            die "ERROR: $me: Secondary foreach attributes must be named" unless (defined($f_attr_name));
            my $f_w_attr=CConfig::Attr->new(
                $f_attr_name,
                VALUE=>$f_attr_val->copy(NUM=>1,STEP=>0)
            );
            $item->{'WITH'}->add($f_w_attr);

            # index the secondary WITH attributes by attribute name
            $f_s_w_attr{$f_attr_name}=$f_w_attr;
        }

        # Resolve ^ references after foreach so we can point to the foreach attribute(s)
        $item->{'WITH'}->resolve();
        $item->{'ON'}->resolve($item->{'WITH'});


        foreach my $f_val (@$f_p_vals) { 

            # set the list data of the @_ variable list iterator to contain only the current f_val.
            #   All variable values are list iterators but when a variable is used in a WITH construct
            #   the grammar fabricates a new iterator around the variable so that only the variables
            #   list data will be shared between iterators.  The exact attributes returned from a hash
            #   valued list will be determined by the KEY in the list iterator.  Thus by setting the @_
            #   variable list data to the $f_val we can allow different parts of the hash to be 
            #   used for different WITH attributes (each of which may be using different iterators
            #   to access different parts of the same list via KEYS) in the same grammar rule.
            #   Unfortunately this means that if you want to hold a reference to the @_ variable
            #   value outside the foreach loop you need to deep copy the iterator.

            $CConfig::VARS{'@_'}->LIST->DATA([$f_val]); 

            # now create NUM objects, advancing the WITH iterators after each, note the foreach
            #   attribute value does not advance due to the STEP=>0 in its iterator
            #   If NUM=ALL we will just do one find, calling ALL on each iterator and passing it all 
            #       down to the CLI, save all objects that are returned
            if ($CConfig::TEST_MODE) { 
                if ($NUM =~ /all/i) {
                        push(@item_trace,{ACTION=>$item->{'ACTION'},NUM=>$NUM,OBJECT=>$item->{'OBJECT'},WITH=>$item->{'WITH'}->to_string(),ON=>$item->{'ON'}->to_string});
                } else {
                    for (my $i=0; $i<$NUM; $i++) {
                        push(@item_trace,{ACTION=>$item->{'ACTION'},NUM=>$NUM,OBJECT=>$item->{'OBJECT'},WITH=>$item->{'WITH'}->to_string(),ON=>$item->{'ON'}->to_string});
                        $item->{'WITH'}->advance();
                        $item->{'ON'}->advance();
                        $item->{'SA'}->{'WITH'}->advance() if defined($item->{'SA'});
                    }
                    if (ref($item->{'NUM'}) eq "CConfig::ListIter") { $item->{'NUM'}->advance(); $NUM=$item->{'NUM'}->get_next(); $NUM=$NUM->[0]; } 
                }
            } else {
                if ($NUM =~ /all/i) {
                        $impl->invoke_action($item,NUM=>"ALL");  
                } else {
                    my %extra_args = ();
                    if ( $item->{OBJECT} eq 'ip_address' ) {
                        $extra_args{_INDEX} = 1;
                    }
                    # The problem we're trying to solve with -single_command is when there might be 1000 
                    # objects, but the user wants to do a strict_find to get 100 of them, so they will 
                    # actually do 100 shows of 1000 volumes.
                    my $single_command = 0;
                    if ( $item->{'ACTION'} =~ /^strict_/ ) {
                        my $sc=$item->{'WITH'}->find_attr('-single_command');
                        if ($sc) {
                            $single_command=$sc->VALUE->get_next()->[0];
                        }
                    }

                    if ( $single_command ) {
                        $impl->invoke_action($item,NUM=>"ALL",LIMIT=>$NUM);  
                    } else {
                        for (my $i=0; $i<$NUM; $i++) {
                            # Just so we don't have to do string compares over and over
                            if ( $extra_args{_INDEX} ) {
                                $extra_args{_INDEX} = $i + 1;
                            }
    
                            $impl->invoke_action($item,NUM=>$NUM,%extra_args); 
                            $item->{'WITH'}->advance();
                            $item->{'ON'}->advance();
    
                            # Advancing each subactions WITH iterators
                            foreach my $sub_action ( @{$item->{'WITH'}->{'SA'}} ) {
                                $sub_action->{'WITH'}->advance();
                            }
                        }
                    }
                    if (ref($item->{'NUM'}) eq "CConfig::ListIter") { $item->{'NUM'}->advance(); $NUM=$item->{'NUM'}->get_next(); $NUM=$NUM->[0]; } 
                }
            }
            # Advance the primary foreach with attribute (this is the fabricated with attribute that cooresponds to the 
            #   primary with attribute, its populated with the driver attribute values)
            $f_p_w_attr->VALUE()->advance(STEP=>1) if defined($f_p_w_attr);
            # Need to advance the secondary foreach attributes, then set their cooresponding WITH attributes positions to match
            foreach my $f_attr (@f_s_attrs) { 
                # this is a little risky if the same iterator appears more than once in the FOREACH list we will accidentally
                #   increment it twice, need to port (or abstract) code from AttrList::advance
                $f_attr->VALUE()->advance(); 
                # set the cooresponding WITH attribute position directly from the secondary foreach attribute
                $f_s_w_attr{$f_attr->NAME}->VALUE->POS($f_attr->VALUE->POS);
            }
        }
        # advance the primary foreach iterator after the rule execution
        $f_p_attr->VALUE->advance();

        # reset the @_ to null
        $CConfig::VARS{'@_'}->LIST->DATA([]); 
    } else {
        # resolve any reference among the WITH attributes
        $item->{'WITH'}->resolve();

        if ($CConfig::TEST_MODE) { 
            if ($NUM =~ /all/i) {
                    push(@item_trace,{ACTION=>$item->{'ACTION'},NUM=>$NUM,OBJECT=>$item->{'OBJECT'},WITH=>$item->{'WITH'}->to_string(),ON=>$item->{'ON'}->to_string});
            } else {
                for (my $i=0; $i<$NUM; $i++) {
                    push(@item_trace,{ACTION=>$item->{'ACTION'},NUM=>$NUM,OBJECT=>$item->{'OBJECT'},WITH=>$item->{'WITH'}->to_string(),ON=>$item->{'ON'}->to_string});
                    $item->{'WITH'}->advance();
                    $item->{'ON'}->advance();
                    $item->{'SA'}->{'WITH'}->advance() if defined($item->{'SA'});
                }
            }
        } else {
            if ($NUM =~ /all/i) {
                $impl->invoke_action($item,NUM=>"ALL");
            } else { 
                my %extra_args = ();
                if ( $item->{OBJECT} eq 'ip_address' ) {
                    $extra_args{_INDEX} = 1;
                }
                my $single_command = 0;
                if ( $item->{'ACTION'} =~ /^strict_/ ) {
                    my $sc=$item->{'WITH'}->find_attr('-single_command');
                    if ($sc) {
                        $single_command=$sc->VALUE->get_next()->[0];
                    }
                }

                if ( $single_command ) {
                    $impl->invoke_action($item,NUM=>"ALL",LIMIT=>$NUM);  
                } else {
                    for (my $i=0; $i<$NUM; $i++) {
                        # Just so we don't have to do string compares over and over
                        if ( $extra_args{_INDEX} ) {
                            $extra_args{_INDEX} = $i + 1;
                        }
                        $impl->invoke_action($item,NUM=>$NUM,%extra_args); 
                        $item->{'WITH'}->advance();
                        $item->{'ON'}->advance();
     
                        # Advancing each subactions WITH iterators
                        foreach my $sub_action ( @{$item->{'WITH'}->{'SA'}} ) {
                            $sub_action->{'WITH'}->advance();
                        }
                    }
                }
            }
        }
    }

    $DB::single=1; 

    # assign variables
    if ($CConfig::TEST_MODE) {
        $CConfig::VARS{$item->{'VARIABLE'}}=CConfig::ListIter->new(LIST=>CConfig::List->new(@item_trace)) if defined($item->{'VARIABLE'});
        $Log->debug(sub {"VARIABLE ".$item->{'VARIABLE'}." = ".NACL::ComponentUtils::Dumper($CConfig::VARS{$item->{'VARIABLE'}})});
    } else {
        # retrieve results from action and execute subactions
        $impl->invoke_post();

        my $results;
        ### If the OBJECT_YAML_DIR input param is defined, the user is requesting
        ### to create YAML file for each type of object specified in the ruleset, 
        ### with the object info for every object of that type that was created/discovered.
        if (defined $CConfig::OBJECT_YAML_DIR) {
            $results=$impl->get_results();
            if (defined ($CConfig::RES{$item->{'OBJECT'}} )) {
               my $val=$CConfig::RES{$item->{'OBJECT'}};
               push (@$val, @$results);
               $CConfig::RES{$item->{'OBJECT'}} = $val;
            }
            else {
               $CConfig::RES{$item->{'OBJECT'}}   = $results;
            }

        }
        if (defined($item->{'VARIABLE'})) {
            if ( !$results ) {
              $results=$impl->get_results();  
            }
            $CConfig::VARS{$item->{'VARIABLE'}}=CConfig::ListIter->new(LIST=>CConfig::List->new(@$results));
        }
    } 

    my $ts_elapsed=time-$ts_start;
    $Log->trace("$me: Rule end: time elapsed=$ts_elapsed");
    return 1;
}

# -----------------------------------------------
# Function: gen_subactions
#       Generate and queue subactions for the current item
#       By default generated subactions are queued to the items SA_LIST property and executed
#           after the call to Impl->post, this is so that undo processing works correctly
#           main rule action must be completed and post processed before any sub-actions begin
#           else sub-action undo's will follow main rule undo
#
# Usage: 
#       gen_subactions(impl,item,%args)

# Parameters:
#       impl - reference to the implementation object for this item
#       item - reference to the item to extract subactions from
#       %args - hash of name/value pairs to be added to the subaction WITH clause
# -----------------------------------------------
sub gen_subactions {
    my $impl=shift;
    my $item=shift;
    my %args=(@_);
    my $me=(caller(0))[3];
    my @subitems=();
    my $deferr=1; # by default sub-actions are deferred until after main rule post processing

    my %subimpls; # when dealing with subactions, if they are for a different Impl object,
                  # we will create one so we can get it's keys.  These are needed so we can
                  # limit what we copy from the parent.

    # make a copy of each subitem for the current item, inherit the WITH attributes from the parent item 
    # and push the subitem onto the SA_LIST stack in the implementation obj for later execution
    foreach my $subitem (@{$item->{'WITH'}->SA}) {
        my %copy=%{$subitem};
        # either its a sub-action or a regular action
        if (defined($subitem->{'SUB_ACTION'})) {
            # subactions work on their parent object and inherit their parents target 
            $copy{'ACTION'}=$subitem->{'SUB_ACTION'};
            $copy{'OBJECT'}=$item->{'OBJECT'};
            $copy{'NUM'}=1;
            $copy{'ON'}=$item->{'ON'};
        }

        if ( !exists $subimpls{$copy{'OBJECT'}} ) {
            $subimpls{$copy{'OBJECT'}}=CConfig::Impl::create_impl(\%copy);
        } 

        # copy the WITH attributes list itself from the subitem, START=>CURRENT so list iterator positions are preserved
        $copy{'WITH'}=$subitem->{'WITH'}->copy(START=>'CURRENT'); 

        # Deep copy the parent's ON so we execute on the same node (important for
        # multi-cluster
        $copy{'ON'}=$item->{'ON'}->copy(START=>'CURRENT');

        # resolve parent attribute references in subcommands 
        $copy{'WITH'}->resolve($item->{'WITH'},COPY=>1);

        # retrieve sub-impl's keys
        my @attribs_subaction_needs;
        my $do_filter;
        if ( (exists $subimpls{$copy{'OBJECT'}}) && ($copy{'ACTION'} eq "modify") ) {
            # ACTION name doesn't always match up to command name 
            # (like volume "efficiency_on" matches up to NACL::C::Efficiency "on") but
            # the main scenario where we need to be careful in copying parent attributes
            # to a sub action is when you create an object, then call modify, generally
            # you don't want to copy all of the create arguments into modify
            @attribs_subaction_needs = $subimpls{$copy{'OBJECT'}}->get_required_keys( action => $copy{'ACTION'} );
            $do_filter = 1;
        }

        # When you have the case where the subaction has a variable which is a
        # reference to a parent variable, which in turn is a reference to a
        # foreach variable, we need to freeze that parent reference
        # in the subaction
        # example is destination-vserver which refers to create's vserver which
        #    refers to @_ (foreach loop variable):
        # create NUM @CLUSTERS[0]{nodes} volume as @v_ls_c0 with
        #     volume={$_{volume}}_lsv[1..100]
        #     vserver=@_{vserver}
        #     size=50MB
        #     type=DP
        #     aggregate=@aggrs_c0
        #     space-guarantee=none {
        #         create snapmirror with
        #             type=LS
        #             source-cluster=$CLUSTERS[0]
        #             source-vserver=^vserver
        #             destination-cluster=$CLUSTERS[0]
        #             destination-vserver=^vserver
        #             destination-volume=^volume;
        #     } foreach source-volume=@v_src_c0 on cluster=@CLUSTERS[0];

        foreach my $attr (@{$copy{'WITH'}->DATA}) {
            my $ref = $attr->REF();
            if ( $ref && !(grep(/^$ref$/, @attribs_subaction_needs)) ) {
                push( @attribs_subaction_needs, $ref );
            }
        }

        # copy parent attributes into subitem
        foreach my $attr (@{$item->{'WITH'}->DATA}) {
            # add an attribute to the WITH clause for parent attributes
            #   note: parent attribute values (iterators) are "frozen" in the child (step is 0, start is current)
            #       subactions are queued and implemented post create, this is to accomodate background creation of resources
            #       so we have to copy the current parent WITH attribute values to the child
            #       We bypass the necessity of the deep copy by creating a new list and installing the parent value in it

            # Since we are "freezing" the list for the copies attribute, we must specify NUM=ALL
            # If not, it will default to 1, and we will only get the first attribute

            if ( !$copy{'WITH'}->find_attr($attr->NAME) ) {
                # Only copy parent attribute for 'modify' if it is a required option of sub impl
                if ( ($do_filter && grep( /^$attr->NAME()$/, @attribs_subaction_needs )) ||
                     !$do_filter ) {
                    $copy{'WITH'}->add_attr(
                            CConfig::Attr->new(
                                $attr->NAME,
                                VALUE=>CConfig::ListIter->new(NUM=>'ALL', 
                                        LIST=>CConfig::List->new(@{$attr->VALUE->get_next()})
                                    )
                            )
                    );
                } 
            }
        }

        # add an attribute to the with clause for each arg to this routine (this is to accomodate fabricated attributes)
        foreach my $attr (keys(%args)) {
            if ( !$copy{'WITH'}->find_attr($attr) ) {
                # Only copy parent attribute for 'modify' if it is a required option of sub impl
                if ( ($do_filter && grep( /^$attr$/, @attribs_subaction_needs )) ||
                     !$do_filter ) {
                    $copy{'WITH'}->add_attr(
                        CConfig::Attr->new(
                            $attr,
                            VALUE=>CConfig::ListIter->new(LIST=>
                                    CConfig::List->new($args{$attr})
                                )
                        )
                    );
                }
            }
        }

        # if action is not immediate then push it onto the SA_LIST stack
        #   for execution during post processing
        if ($deferr) { push(@{$impl->{'SA_LIST'}},\%copy);
        } else { CConfig::Rule::_process_rule(\%copy); }
    }
    return;
}

# -----------------------------------------------
# These functions were moved to CConfig.pm but 
#   left these here to prevent breaking existing code
# -----------------------------------------------
sub get         { return CConfig::get(@_); }
sub start_tx    { return CConfig::start_tx(@_); }
sub commit      { return CConfig::commit(@_); }
sub rollback    { return CConfig::rollback(@_); }
sub undo        { return CConfig::undo(@_); }
sub artifactgen { return CConfig::artifactgen(@_); }


=head1 FUNCTIONS



=head2 init 

    Initialize the parser from a grammar, and clear variables

=head3 Usage

    CConfig::Rule::init()

=head3 Return

    1=success, 0=failure 

=head2 start_tx

    start a transaction 

=head3 Usage

    CConfig::Rule::start_tx(undofile)

=head3 Return

    1=success, 0=failure 

=head3 Parameters:


=over 4


=item undofile

    undo file name
    

=back


=head2 commit 

    commit a transaction 

=head3 Usage

    CConfig::Rule::commit()

=head2 process_file 

    Parse and execute a ruleset file

=head3 Usage

    CConfig::Rule::process_file(ruleset_file)

=head3 Return

    instance of Parse::RecDescent
    1=success, 0=failure 

=head3 Parameters:


=over 4


=item ruleset_file

    ruleset file name
    

=back


=head2 process_ruleset 

    Parse and execute a ruleset 

=head3 Usage

    CConfig::Rule::process_ruleset(ruleset)

=head3 Return

    1=success, 0=failure 

=head3 Parameters:


=over 4


=item ruleset

    ruleset to process 
    

=back


=head2 process_rule 

    Parse and execute a single rule

=head3 Usage

    CConfig::Rule::process_rule(rule)

=head3 Return

    1=success, 0=failure 

=head3 Parameters:


=over 4


=item rule

    rule to process 
    

=back


=head2 _process_rule

    process a parsed rule (internal)

=head3 Usage

    _process_rule(item)

=head3 Return

    1=success, 0=failure 

=head3 Parameters:


=over 4


=item item

    reference to the parser hash
    

=back


=head2 gen_subactions

    Generate a and queue subactions for the current item
    By default generated subactions are queued to the items SA_LIST property and executed
    after the call to Impl->post, this is so that undo processing works correctly
    main rule action must be completed and post processed before any sub-actions begin
    else sub-action undo's will follow main rule undo

=head3 Usage

    gen_subactions(impl,item,%args)

=head3 Parameters:


=over 4


=item impl

    reference to the implementation object for this item

=item item

    reference to the item to extract subactions from

=item %args

    hash of name/value pairs to be added to the subaction WITH clause

=back


=head2 undo

    undo a transaction using commands in the undo file

=head3 Usage

    CConfig::Rule::undo($undofile)

=head3 Parameters:


=over 4


=item undofile

    undo file name 

=back


=head2 rollback

    rollback an in flight transaction

=head3 Usage

    CConfig::Rule::rollback()

=head2 artifactgen

    Generate artifact file

=head3 Usage

    CConfig::Rule::artifactgen(artifactfile)

=head3 Parameters:


=over 4


=item artifactfile

    artifact file name 

=back


=head2 get

    Get the specified variable

=head3 Usage

    CConfig::Rule::get(varname)


=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut

1;

