# -*- perl -*-
# $Id$

## @summary Subtest 
## @author dl-coati-core
## @cf     false
## @status public
## @pod here

=head1 NAME

Subtest - Interface to Subtest

=head1 SYNOPSIS

  use CConfig::Subtest;

=head1 DESCRIPTION

Handles pacing subtests so that we do not go over filer limits and also
handles retrieving subtest results.

=cut

# --------------------------------------------------------
# Class: CConfig::Subtest
#   Utility routines for handling subtests
# --------------------------------------------------------

package CConfig::Subtest;
use strict;
use warnings;
use Data::Dumper;
use Carp;
use NACL::ComponentUtils;

use Tharn qw($Log);
use NATE::Log::Call qw(call_trace);

# Subtests are prefaced by strings which contain a sequence number in order that
# they are displayed in order in the log viewers
our $subtest_seq           = 0;
our %preface_seq           = ();

our $conn_per_second_limit = 10;
our $new_rule              = 0;
our %subtest_pacing;
    $subtest_pacing{GLOBAL}{MIN}       = 1;
    $subtest_pacing{GLOBAL}{AT_A_TIME} = 10;

our @PENDING_SUBTESTS      = ();
our $num_pending_subtests  = 0;

sub DEFAULT_AT_A_TIME {if(@_) { $subtest_pacing{GLOBAL}{AT_A_TIME} = $_[0]; } return $subtest_pacing{GLOBAL}{AT_A_TIME}; }
sub DEFAULT_MIN       {if(@_) { $subtest_pacing{GLOBAL}{MIN}       = $_[0]; } return $subtest_pacing{GLOBAL}{MIN}; }
sub CONN_PER_SECOND   {if(@_) { $conn_per_second_limit             = $_[0]; } return $conn_per_second_limit; }
sub NEW_RULE          {if(@_) { $new_rule                          = $_[0]; } return $new_rule; }

# Ontap limits us to 10 connections/second to avoid DOS (denial of service) attacks
# So we will 1/10 * 1.25 or 125ms between each of the subtests
# This is independent of how many subtests will be running at one time.
sub CONN_LIMIT_SLEEP  { return ((1.0/CONN_PER_SECOND()) * 1.25); }

# -----------------------------------------------
# Function: SUBTEST_PACING
#   Parse the SUBTEST_PACING command-line option.
#
# Usage: 
#   CConfig::Subtest::SUBTEST_PACING( $opt_hash->{SUBTEST_PACING} );
#
# Format:
#   Example 1: '2:10' - only subtest when creating 2 or more objects, and subtest 10 at a time.
#   Example 2: '0'    - disabled subtesting globally.
#   Example 3: '1:10,volume:2:12,interface:0' 
#                     - all implementations will subtest when creating 1 or more objects and subtest 10 at a time, 
#                       except 'volume' which will only subtest when creating 2 or more objects and will subtest 12 at a time 
#                          and 'interface' which will never subtest.
#                       NOTE:  Global pacing must always be specified first
#
# -----------------------------------------------
sub SUBTEST_PACING   { 
    if (@_) { 
        # first is always the global setting
        my @items = split( ',', $_[0] );
        my @global = split( ':', $items[0] );
        if ( scalar(@global) ) {
           $subtest_pacing{GLOBAL}{MIN} = $global[0]; 
        }
        if ( scalar(@global) > 1 ) {
           $subtest_pacing{GLOBAL}{AT_A_TIME} = $global[1]; 
        }

        # shift off the 1st (global) arg
        shift @items;
        foreach my $item (@items) {
            my @spec = split( ':', $item );
            if ( scalar(@spec) < 2 ) {
                die "SUBTEST_PACING must provide at a minimum an implementation_name:number";
            } else {
               my $impl_name = $spec[0];
               $subtest_pacing{$impl_name}{MIN} = $spec[1];
               if ( scalar(@spec) > 2 ) {
                   $subtest_pacing{$impl_name}{AT_A_TIME} = $spec[2];
               }
            }
        }
    } 

#print "subtest_pacing = |".Dumper(\%subtest_pacing)."|\n";

    return \%subtest_pacing; 
}

# -----------------------------------------------
# Function: SUBTEST_MIN
#   Returns the minimum number of objects to subtest.
#   In other words, if 2 is returned, then this means that you
#   should subtest if you are going to create 2 or more objects.
#   If 0 is returned, do not subtest at all.
#
# Usage: 
#   my $intstmin  = CConfig::Subtest::SUBTEST_MIN( object => 'interface' );
#   my $globstmin = CConfig::Subtest::SUBTEST_MIN( );
#
# -----------------------------------------------
sub SUBTEST_MIN   { 
    my %opt = ( object => "GLOBAL", @_ );
    if ( exists $subtest_pacing{$opt{object}}{MIN} ) {
        return $subtest_pacing{$opt{object}}{MIN};
    }
    return $subtest_pacing{GLOBAL}{MIN};
}

# -----------------------------------------------
# Function: SUBTEST_AT_A_TIME
#   Returns the max number of subtests to run at a time
#   In other words, if 10 is returned, then this means that you
#   should wait for subtests to finish if you have already started 10.
#
# Usage: 
#   my $intstmin  = CConfig::Subtest::SUBTEST_AT_A_TIME( object => 'interface' );
#   my $globstmin = CConfig::Subtest::SUBTEST_AT_A_TIME( );
#
# -----------------------------------------------
sub SUBTEST_AT_A_TIME   { 
    my %opt = ( object => "GLOBAL", @_ );
    if ( exists $subtest_pacing{$opt{object}}{AT_A_TIME} ) {
        return $subtest_pacing{$opt{object}}{AT_A_TIME};
    }
    return $subtest_pacing{GLOBAL}{AT_A_TIME};
}

# -----------------------------------------------
# Function: _method2label
#
#   Takes a fully qualified name like NACL::C::Volume::create or
#   CConfig::Impl::Volume::do_purge and will
#
#     1) Strip off the command NACL::C:: and/or CConfig::Impl::
#     2) Convert :: to _
#
# Return:
#
#   my $label = _method2label( method => $method );
#
# -----------------------------------------------
sub _method2label {
    my %opt = ( method => "default", @_ );

    my $method = $opt{method};
    $method =~ s/NACL::C:://;
    $method =~ s/CConfig::Impl:://;
    $method =~ s/::/_/g;
    return $method;
}

# -----------------------------------------------
# Function: _get_subtest_directory
#
#   When subtesting out actions, we need to group the subtest output
#   into directories, and preface these directories with sequence
#   numbers so that they are displayed in by the log viewers in
#   order.  Currently the toplevel sequence number is 4 digits,
#   zero padded on the left.
#
#   The outer directory (governed by _get_subtest_directory()) will be like:
#
#       0003_Volume_create
#
#   The inner directory (governed by _get_subtest_seq_label()) will be like:
#
#       000000001_Volume_create
#
#   and you end up with:
#
#    0003_Volume_create/000000001_Volume_create
#    0003_Volume_create/000000002_Volume_create
#    0003_Volume_create/000000003_Volume_create
#
# Return:
#
#   my $label = _get_subtest_directory( method => $method );
#
# -----------------------------------------------
sub _get_subtest_directory {
    my %opt = ( method => "default", @_ );

    my $method = _method2label( method => $opt{method});
    my $seq = 0;
    if ( !exists $preface_seq{$method} ) {
        # If a new top-level rule, but never encountered this implementation/method before, unset NEW_RULE
        # so we don't end up creating multiple directories
        # In other words if you have 2 top level "create volume" rules, clear
        # NEW_RULE for the first one.
        if ( NEW_RULE() ) {
            NEW_RULE(0);
        }
        $subtest_seq++;
        $preface_seq{$method} = $subtest_seq;
    } elsif ( NEW_RULE() ) {
        $subtest_seq++;
        $preface_seq{$method} = $subtest_seq;
        NEW_RULE(0);
    }

    my $subtest_dir = sprintf( "%04d_%s/", $preface_seq{$method}, $method );

    return $subtest_dir;
}

# -----------------------------------------------
# Function: _get_subtest_seq_label
#
#   When subtesting out actions, we need to group the subtest output
#   into directories, and preface these directories with sequence
#   numbers so that they are displayed in by the log viewers in
#   order.
#
#   The outer directory (governed by _get_subtest_directory()) will be like:
#
#       0003_Volume_create
#
#   The inner directory (governed by _get_subtest_seq_label()) will be like:
#
#       000000001_Volume_create
#
#   and you end up with:
#
#    0003_Volume_create/000000001_Volume_create
#    0003_Volume_create/000000002_Volume_create
#    0003_Volume_create/000000003_Volume_create
#
# Return:
#
#   my $label = _get_subtest_seq_label( method => $method );
#
# -----------------------------------------------
sub _get_subtest_seq_label {
    my %opt = ( method => "default", @_ );

    # Here $method is just used to select a unique serial number
    my $method = _method2label( method => $opt{method}).$opt{method};
    if ( exists $preface_seq{$method} ) {
        $preface_seq{$method}++;
    } else {
        $preface_seq{$method} = 1;
    }
    return sprintf( "%09d", $preface_seq{$method} );
}

# -----------------------------------------------
# Function: wait_finish
#   Wait until pending subtests are finished
#
# Usage: 
#   wait_finish()
#
# Assumptions:
# -----------------------------------------------
sub wait_finish {
    my $call = call_trace();
    my $me=(caller(0))[3];
    my $calledby=(caller(1))[3];
    my $calledby2=(caller(2))[3];
    my @calledby = ( $calledby, $calledby2 );

    my $fail = 0;
    my $bged = 0;
    if ( @PENDING_SUBTESTS ) {
        # Wait for this batch of parallel tests to finish
        Subtest::wait_finish( subtests => \@PENDING_SUBTESTS );

        foreach my $st ( @PENDING_SUBTESTS ) {
            my @results=$st->own_results();
            foreach my $result (@results) {
                if ( ref($result) =~ /^NATE::Result::Pass/ ) {
                    if ( $result->misc() && ($result->misc() =~ /(\d+)/) ) {
                        push(@CConfig::BG_JOBS,$1);
                        $bged++;
                    }
                } elsif ( ref($result) =~ /^NATE::Result::Warn/ ) {
                    my $misc = $result->misc();
                    if ( $misc && ($misc =~ /Job\s+(\d+)\s+(is running in the background|has been queued)/) ) {
                        # Examples (Ontap reports these as errors, but generally are not):
                        # ERROR: Job 316 is running in the background
                        # ERROR: Job 437 has been queued
                        push(@CConfig::BG_JOBS,$1);
                        $bged++;
                    }
                } elsif ( ref($result) =~ /^NATE::Result::(Fatal|Fail|Script|Config)$/ ) {
                    $fail++;
                }
            }
        }

        if ( $bged ) {
            # wait for background jobs to complete
            if (@CConfig::BG_JOBS) {
                $Log->trace("Waiting for jobs ".join(',',@CConfig::BG_JOBS));
                while(CConfig::Job::are_jobs_active(\@CConfig::BG_JOBS)) { sleep 2; }
            }

            # check bg jobs
            if (@CConfig::BG_JOBS) { CConfig::Job::check_jobs(\@CConfig::BG_JOBS); }
        }

        $num_pending_subtests = 0;
        @PENDING_SUBTESTS = ();
        @CConfig::BG_JOBS=();

        if ( $fail ) {
            die "CConfig::Subtest::wait_finish() - $fail subtests failed";
        }
    }

    return 1;
}

# -----------------------------------------------
# Function: subtest
#   Subtest a CConfig::Impl action
#
# Usage: 
#   subtest( class => "NACL::C::Volume", method => "create", args => \%args );
#   subtest( class => "CConfig::Impl::Volume", method => "do_purge", args => \%args );
#
# Parameters:
#       class     - Class name of object which will be passed as the first argument to a
#                   "static" method.
#       method    - name of 'static' routine to call
#       args      - Arguments to pass to method
#       at_a_time - Number of subtests to do at a time (i.e. if you have 100 subtests, and
#                   you want to do them 10 at a time, then wait for completion, then another
#                   10 and so on...)  At a time needs to be set because:
#
#                   1) Filer connection limits (around 125)
#                   2) As you do more "near-simultaneous' actions, each action tends to 
#                      take longer.  Eventually, if not limited by at_a_time, would take
#                      so long that commands would timeout.
#
# Return: 
#   1
#
# -----------------------------------------------
sub subtest {
    my $call = call_trace();
    my $me=(caller(0))[3];

    my %opt = ( class     => undef,
                method    => undef,
                args      => undef,
                at_a_time => undef,
                @_
              );

  
    if ( ! $opt{at_a_time} ) {
        if ( defined $opt{args}{'-subtest_at_a_time'} ) {
            $opt{at_a_time} = delete $opt{args}{'-subtest_at_a_time'};
        } else {
            $opt{at_a_time} = DEFAULT_AT_A_TIME();
        }
    }

    my $conn_limit_sleep = 1;
    if ( $opt{args} ) {
        if ( $opt{args}{apiset_should} && ($opt{args}{apiset_should}{interface} eq 'ZAPI') ) {
            $conn_limit_sleep = 0;
        }
        if ( $opt{args}{apiset_must} && ($opt{args}{apiset_must}{interface} eq 'ZAPI') ) {
            $conn_limit_sleep = 0;
        }
    } 

    # In order not to hit the connections/second limit, sleep a
    # fraction of a second if there are currently pending subtests
    if ( $conn_limit_sleep && @PENDING_SUBTESTS ) {
        Tharn::sleep( CONN_LIMIT_SLEEP() );
    }

    my $generic_runid = $opt{method};

    # Some purge routines are on the individual Impl (i.e. Volume) objects and
    # some can just use the generic Impl.pm
    # If routine name is the generic CConfig::Impl::purge, tack on the
    # NACL class name
    if ( $generic_runid =~ /^CConfig\:\:Impl\:\:[a-zA-Z_]+$/ ) {
        if ( $opt{class} =~ /\:\:([a-zA-Z]+)$/ ) {
            $generic_runid = $1.'_'.$generic_runid;
        }
    }

    # Give subtests more descriptive names
    # including their NACL class names
    if ( ($generic_runid =~ /do_undo/) &&
         (exists $opt{args}{-cmd}) ) {
         my $undo_name = $opt{args}{-cmd};
         # If we pass in class NACL::C::VolumeEfficiency
         #               -cmd off
         # We'd rather end up with rollback_VolumeEfficiency_off instead of
         #                         rollback_Volume_off
         if ( (exists $opt{class}) &&
              ($opt{class} =~ /\:\:([a-zA-Z]+)$/) ) {
            $generic_runid = 'rollback_'.$1.'_'.$undo_name;
         } else {
             $generic_runid =~ s/do_undo/$undo_name/;
             $generic_runid = 'rollback_'.$generic_runid;
         }
    } elsif ( $generic_runid =~ /do_purge/ ) {
         $generic_runid =~ s/do_purge/purge/;
         $generic_runid = 'rollback_'.$generic_runid;
    }

    # Generate the directory/label
    my $preface = _get_subtest_directory( method => $generic_runid );
    my $subseq  = _get_subtest_seq_label( method => $generic_runid );

    # Create the runid string
    my $runid_string = _method2label( method => $preface.$subseq . "_".$generic_runid );
    my $st =  Tharn::subtest(
                \&_do_subtest,
                "-bg",
                "-runid",
                $runid_string,
                "-onexit",
                sub { },
                "-show-results",
                '--',
                'class',
                $opt{class},
                'method',
                $opt{method},
                'args',
                $opt{args},
          );

    push( @PENDING_SUBTESTS, $st );
    $num_pending_subtests++;

    if ( $num_pending_subtests >= $opt{at_a_time} ) {
        wait_finish();
    }

    return 1;
}

# -----------------------------------------------
# Function: _do_subtest
#   Subtest callback routine, called by the NATE subtest mechanism.
#
# Usage: 
#   _do_subtest( class => "NACL::C::Volume", method => "create", args => \%args );
#
# Parameters:
#       class     - Class name of object which will be passed as the first argument to a
#                   "static" method (generally NACL::C::* or CConfig::Impl::* but could be
#                   anything).
#       method    - name of static routine to call
#       args      - Arguments to pass to method
#
# Return: 
#   1
#
# -----------------------------------------------
sub _do_subtest {
    my $me=(caller(0))[3];
    my %opt = ( class     => undef,
                method    => undef,
                args      => undef,
                @_
              );

    my $node = $opt{args}{command_interface};

    my $job;
    if ( exists $opt{args}{foreground} ) {
        if ( $opt{args}{foreground} =~ /false/ ) {
            $opt{args}{job_component} = \$job;
        }
    }

    # Make sure class is loaded
    if ( !defined $CConfig::LOADED{$opt{class}} ) {
        eval "use $opt{class}";
        $CConfig::LOADED{$opt{class}} = 1;
    }

    # Call the static (class) method
    my $class  = $opt{class};
    my $method = $opt{method};

#use Data::Dumper;
#local $Data::Dumper::Maxdepth=3;
#print "FWR $me ".Dumper($opt{args})."\n";

    eval {$class->$method( %{$opt{args}} );};

    my $rc = $@;
    # Ontap declares ERRORs for some things which aren't really errors
    if ( $rc =~ /Job\s+(\d+)\s+(is running in the background|has been queued)/ ) {
        my $result = NATE::Result::Warn->new( $rc );
        $result->misc($rc);
        $result->log();
    } elsif ( $rc ) {
        my $result = NATE::Result::Fail->new( $rc );
        $result->misc($rc);
        $result->log();
    } else {
        my $result = NATE::Result::Pass->new( "PASS" );
        if ( $job ) {
            $result->misc($job->id());
        }
        $result->log();
    }
}

=head1 FUNCTIONS



=head2 wait_finish

    Wait until all pending subtests are finished.

=head3 Usage

    wait_finish()

=head3 Return

    1

=head3 Parameters:


=head2 subtest

    Subtest a static method call.

=head3 Usage

    subtest( class => "NACL::C::Volume",       method => "create", args => \%args );
    subtest( class => "CConfig::Impl::Volume", method => "do_purge", args => \%args, at_a_time => 10 );

=head3 Return

    1

=head3 Parameters:


=over 4


=item <class>

    class name to be passed as first argument to static method
    
=item <method>

    method name of routine to be called

=item <args>

    hashref of argument hash to be passed to "method"

=item <at_a_time>

    number of subtests to run at a time.  (i.e. maybe you have 100 volume creates, if
        at_a_time=10, you will create 10 volumes, wait until they finish, then do 10 more
        and so on...

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut


1;
