# -*- perl -*-
# $Id $

## @summary Rule 
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

TXMgr - an object set

=head1 SYNOPSIS

  use TXMgr

=head1 DESCRIPTION

Manager a set of possible nested transactions and their associated undo actions

=cut

# --------------------------------------------------------
# Class: CConfig::TXMgr
#   A transaction manager
# --------------------------------------------------------

package CConfig::TXMgr;
use strict;
use warnings;
use Params::Validate qw(:all);
use Data::Dumper;
use Carp;
use Tharn qw($Log snooze);

my $singleton;

# -----------------------------------------------
# Function: new
#   create new transaction manager
#
# Usage: 
#   TXMgr->new()  
#
# Parameters:
# Return: 
# Assumptions:
# -----------------------------------------------
sub new {
    my $class = shift;
    return $singleton if (defined($singleton));
    $singleton=bless {
        STACK=>[],
    }, $class;
}

sub STACK   { my $self = shift; if(@_) { $self->{STACK} = $_[0]; } return $self->{STACK}; }
sub CMDS    { my $self = shift; return $self->{STACK}->[-1]->{'CMDS'}; }
sub UNDOFILE    { my $self = shift; return $self->{STACK}->[-1]->{'UNDOFILE'}; }
sub UNDOFH      { my $self = shift; return $self->{STACK}->[-1]->{'UNDOFH'}; }

# -----------------------------------------------
# Function: start
#   start a new transaction
#
# Usage: 
#   TXMgr->start(<undofile>)
#
# Parameters:
#       <undofile> - optional undo file to log actions to
# -----------------------------------------------
sub start {
    my $self = shift;
    my $undofile = shift;
    my $undofh   = undef;
    my $me=(caller(0))[3];

    if (defined($undofile)) {
        open($undofh,">$undofile") or croak "ERROR $me: Cant open undofile file: $undofile";
        $undofh->autoflush(1);
    };

    push(@{$self->STACK},{
            UNDOFILE=>$undofile,
            UNDOFH=>$undofh,
            CMDS=>[]
        });

    return;
}

# -----------------------------------------------
# Function: top
#   return the transaction at the top of the stack
#
# Parameters:
# -----------------------------------------------
sub top {
    my $self = shift;
    return $self->STACK->[-1];
}

# -----------------------------------------------
# Function: push
#   push command onto stack for current transaction
#
# Usage:
#   push(<cmd>);
#
# Parameters:
#   <cmd> - command to push
# -----------------------------------------------
sub push {
    my $self = shift;
    my $cmd = shift;
    my $me=(caller(0))[3];

    $Log->debug(sub {"UNDO PUSH: $cmd"});
    # check that we are in a transaction
    if (scalar(@{$self->STACK})==0) {
        die "$me: Not in a transaction, call TXMgr::start to start a transaction before calling push";
    }
    # push command onto stack
    push(@{$self->CMDS},$cmd);
    # if undofile is specified then write command to file
    my $fh=$self->UNDOFH;
    print $fh $cmd."\n" if defined($fh);
}

# -----------------------------------------------
# Function: commit
#   commit a transaction - if a undofile was specified then 
#       the internal commands are written to the undo file
#
# Parameters:
# -----------------------------------------------
sub commit {
    my $self = shift;
    my $me=(caller(0))[3];
    my $fh=$self->UNDOFH;
    close($fh) if defined($fh);
    pop(@{$self->STACK});
};

# -----------------------------------------------
# Function: rollback
#   rollback a transaction
#
#   Usage:
#       rollback();
#
# Parameters:
# -----------------------------------------------
sub rollback {
    my $self = shift;
    my $me=(caller(0))[3];
    my @cmds=();

    if (scalar(@{$self->STACK}>0)) {
        my $fh=$self->UNDOFH;
        close($fh) if defined($fh);

        @cmds=reverse @{$self->CMDS};

        # run the undo commands
        foreach my $cmd (@cmds) {
            $Log->trace("$me: Executing command: $cmd\n");

            if ( $cmd =~ /\(\"(CConfig\:\:Impl\:\:[a-zA-Z]+)/ ) {
                my $pkg = $1;
                if ( !defined $CConfig::LOADED{$pkg} ) {
                    eval "use $pkg";
                    $CConfig::LOADED{$pkg} = 1;
                }
            }

            if ( $cmd =~ /(NACL\:\:C\:\:[a-zA-Z]+)/ ) {
                my $pkg = $1;
                if ( !defined $CConfig::LOADED{$pkg} ) {
                    eval "use $pkg";
                    $CConfig::LOADED{$pkg} = 1;
                }
            }

            $Log->debug(sub {"Executing command = |$cmd|"});

            eval "$cmd";

            if ($@) {
                $Log->warn("$me: command failed with: $@\n");
            } else {
                $Log->trace("$me: Command Completed Successfully\n");
            }
        }

        pop(@{$self->STACK}) if scalar(@{$self->STACK});
        return 1;
    }
    return 0;
}

# -----------------------------------------------
# Function: rollback_last
#   rollback the last command in a transaction
#
#   Usage:
#       my $cmd = rollback_last();
#
# Parameters:
#     none
# Return:
#   command string or "undef" if no commands are left in the queue
# -----------------------------------------------
sub rollback_last {
    my $self = shift;
    my $me=(caller(0))[3];
    my @cmds=();
    my $cmd;

    if (scalar(@{$self->STACK}>0)) {
        my $fh=$self->UNDOFH;
        close($fh) if defined($fh);

        @cmds=@{$self->{STACK}->[-1]->{'CMDS'}};
        $cmd = pop( @cmds );
        $self->{STACK}->[-1]->{'CMDS'} = \@cmds;

        # run the undo command
        $Log->trace("$me: Executing command: $cmd\n");

        if ( $cmd =~ /\(\"(CConfig\:\:Impl\:\:[a-zA-Z]+)/ ) {
            my $pkg = $1;
            if ( !defined $CConfig::LOADED{$pkg} ) {
                eval "use $pkg";
                $CConfig::LOADED{$pkg} = 1;
            }
        }

        if ( $cmd =~ /(NACL\:\:C\:\:[a-zA-Z]+)/ ) {
            my $pkg = $1;
            if ( !defined $CConfig::LOADED{$pkg} ) {
                eval "use $pkg";
                $CConfig::LOADED{$pkg} = 1;
            }
        }

        $Log->debug(sub {"Executing command = |$cmd|"});

        eval "$cmd";

        if ($@) {
            $Log->warn("$me: command failed with: $@\n");
        } else {
            $Log->trace("$me: Command Completed Successfully\n");
        }
    }

    # If no more entries exist
    if ( !scalar( @cmds ) ) {
        pop(@{$self->STACK}) if scalar(@{$self->STACK});
    }
    return $cmd;
}

# -----------------------------------------------
# Function: undo
#   undo a transaction from an undo file
#
#   Usage:
#       undo([<undofile>]);
#
# Parameters:
#   <undofile> - rollback commands from undo file 
# -----------------------------------------------
sub undo {
    my $self = shift;
    my $undofile = shift;
    my $me=(caller(0))[3];
    my @cmds=();

    open(FH,$undofile) or croak "ERROR: $me: Cant open undo file: $undofile";
    @cmds= reverse <FH>;
    close(FH);

    # run the undo commands
    foreach my $cmd (@cmds) {
        $Log->trace("$me: Executing command: $cmd\n");
        if ( $cmd =~ /\(\"(CConfig\:\:Impl\:\:[a-zA-Z]+)/ ) {
            my $pkg = $1;
            if ( !defined $CConfig::LOADED{$pkg} ) {
                eval "use $pkg";
                $CConfig::LOADED{$pkg} = 1;
            }
        }

        if ( $cmd =~ /(NACL\:\:C\:\:[a-zA-Z]+)/ ) {
            my $pkg = $1;
            if ( !defined $CConfig::LOADED{$pkg} ) {
                eval "use $pkg";
                $CConfig::LOADED{$pkg} = 1;
            }
        }

        eval "$cmd";

        if ($@) {
            $Log->warn("$me: command failed with: $@\n");
        } else {
            $Log->trace("$me: Command Completed Successfully\n");
        }
    }
    return 1;
}


=head1 FUNCTIONS



=head2 new

    create new transaction manager

=head3 Usage

    TXMgr->new()  

=head2 start

    start a new transaction

=head3 Usage

    TXMgr->start(<undofile>)

=head3 Parameters:


=over 4


=item <undofile>

    optional undo file to log actions to

=back


=head2 top

    return the transaction at the top of the stack

=head2 push

    push command onto stack for current transaction

=head3 Usage

    push(<cmd>);

=head3 Parameters:


=over 4


=item <cmd>

    command to push

=back


=head2 commit

    commit a transaction - if a undofile was specified then 
    the internal commands are written to the undo file

=head2 rollback

    rollback a transaction

=head3 Usage

    rollback();

=head2 undo

    undo a transaction from an undo file

=head3 Usage

    undo([<undofile>]);

=head3 Parameters:


=over 4


=item <undofile>

    rollback commands from undo file 

=back



=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut

1;
