# -*- perl -*-
# $Id$

## @summary CConfig::VList 
## @author kenneth.gates@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

VList - a list 

=head1 SYNOPSIS

  use VList

=head1 DESCRIPTION

An anstraction of a list of values, includes a current list position

=cut

# --------------------------------------------------------
# Class: CConfig::VList
#   Class to abstract a list of values and include the concept of a current position
# --------------------------------------------------------

package CConfig::VList;
use base qw(CConfig::List);
use Params::Validate qw(:all);
use Tharn qw($Log snooze);
use Data::Dumper;
use Carp;
use strict;
use warnings;

# -----------------------------------------------
# Function: new
#   constructor
#
# Usage: 
#   new(list)
#
# Parameters:
#       list      - reference to an array 
#
# Return: blessed reference
# -----------------------------------------------
sub new {
    my $invocant = shift;
    my %args=(@_);
    my $class    = ref($invocant) || $invocant;
    my $self=$class->SUPER::new();
    foreach my $arg (qw(ROOT STARTI ENDI UNITS)) { $self->{$arg}=$args{$arg} if defined($args{$arg}); }
    if (defined($args{'ROOT'})) {
        my $me=(caller(0))[3];
        die "$me: ROOT is not an array reference its a: ".ref($args{'ROOT'}) unless ref($args{'ROOT'}) eq "ARRAY";
    };
    $self->{'REIFIED'}=0; # track if we have been converted to a real list 
    bless($self, $class);
}

# -----------------------------------------------
# Function:  DATA - list data
#   if we have list data then we have been reified so return the list pointer as usual
# -----------------------------------------------
sub DATA   { 
    my $self = shift; 
    $self->reify();
    if(@_) { $self->{DATA} = $_[0]; } 
    return $self->{DATA}; 
}
# -----------------------------------------------
# Function:  ROOT- return the root of the identifier
#   Note that the VLIST wraps iterators, the position of the list
#       is maintained by the iterator that this list is wrapped by, lists themselves
#       don't maintain a current position so the parent iterator position is passed down
#       to the list value iterators via the POS argument
# -----------------------------------------------
sub ROOT    { 
    my $self = shift; 

    my %args = validate(
        @_,
        {
           POS => { 
              type  => SCALAR,
              default => $self->POS,
           },
        }
    );

    return undef if (!defined($self->{'ROOT'})); 

    my $id="";
    # assemble the root components
    foreach my $comp (@{$self->{'ROOT'}}) { 
        if (ref($comp) eq "CConfig::ListIter") {
            my @elements;
            foreach my $val (@{$comp->get_next(POS=>$args{'POS'})}) { 
                push( @elements, $val );
            }
            $id .= join( ",", @elements );
        } else {
            $id.=$comp; 
        }
    }
    return $id;
}

# -----------------------------------------------
# Function:  STARTI
# -----------------------------------------------
sub STARTI    { my $self = shift; if(@_) { $self->{'STARTI'}  = $_[0]; } return $self->{'STARTI'}; }
# -----------------------------------------------
# Function:  ENDI
# -----------------------------------------------
sub ENDI    { my $self = shift; if(@_) { $self->{'ENDI'}  = $_[0]; } return $self->{'ENDI'}; }
# -----------------------------------------------
# Function:  UNITS
# -----------------------------------------------
sub UNITS    { my $self = shift; if(@_) { $self->{'UNITS'}  = $_[0]; } return $self->{'UNITS'}; }
# -----------------------------------------------
# Function:  LEN  - length of data array
#   The LEN of the embedded list will be the greater of the index range or the length of the longest embedded list
# -----------------------------------------------
sub LEN    { 
    my $self = shift; 
    if ($self->{'REIFIED'}) { return $self->SUPER::LEN(@_); }
    my $len=0;
    if (defined($self->STARTI) && defined($self->ENDI)) { $len=$self->ENDI-$self->STARTI+1; }
    if (defined($self->{'ROOT'})) { 
        # get length of all lists, pick max
        foreach my $comp (@{$self->{'ROOT'}}) { 
            if (ref($comp) eq "CConfig::ListIter") {
                $len=($len<$comp->LIST->LEN)? $comp->LIST->LEN : $len;
            } else {
                $len=($len<1)? 1 : $len;
            }
        }
    };
    return $len;
}
# -----------------------------------------------
# Function:  LASTI- last index of data in list
# -----------------------------------------------
sub LASTI  { 
    my $self = shift; 
    if ($self->{'REIFIED'}) { return $self->SUPER::LASTI(@_); }
    return $self->LEN-1;
}

# -----------------------------------------------
# Function: reify
#   Make the list "real", compute the list values and add them to the DATA arrary
#
# Usage: 
#   reify()
#
# Parameters:
#
# Return: 
#
# Assumptions:
# -----------------------------------------------
sub reify   { 
    my $self = shift; 
    if (!$self->{'REIFIED'}) {
        for (my $i=0; $i<$self->LEN; $i++) {
            push(@{$self->{DATA}},$self->get(POS=>$i));
        }
        $self->{'REIFIED'}=1;
    }
    return;
}

# -----------------------------------------------
# Function: get
#   return value of a list element.  
#
# Usage: 
#   get([POS=>pos])
#
# Parameters:
#   pos - optional, list index to return value for, by default
#           will use the current list position
#
# Return: 
#   data element at the current cursor pos
#
# Assumptions:
# -----------------------------------------------
sub get {
    my $self=shift;
    if ($self->{'REIFIED'}) { return $self->SUPER::get(@_); }

    my %args = validate(
        @_,
        {
           POS => { 
              type  => SCALAR,
              default => $self->POS,
           },
        }
    );
    my $velement="";
    if ( defined($self->{'ROOT'}) ) {
        $velement.=$self->ROOT(POS=>$args{'POS'});
    }

    if (defined($self->STARTI)) { 
        if ($args{'POS'}>$self->LASTI) { $args{'POS'}=$args{'POS'} % $self->LEN; } 
        $velement.=$self->STARTI+$args{'POS'};
    }
    if (defined($self->UNITS))  { $velement.=$self->UNITS; }
    return $velement;
}


# -----------------------------------------------
# Function: copy
#   copy a list 
#
# Usage: 
#   copy()
#
# Parameters:
#
# Return: 
#   ref to copy of self
# -----------------------------------------------
sub copy {
    my $self=shift;
    # can't just call superclass copy since accessing the DATA element will cause us to reify which
    #   will freeze the value of any variables in the root at the point that copy is called
    my $copy=__PACKAGE__->new(); 
    foreach my $attr (qw(ROOT STARTI ENDI UNITS DATA POS REIFIED)) { $copy->{$attr}=$self->{$attr} if defined($self->{$attr}); }
    return $copy;
}

=head1 FUNCTIONS



=head2 new

    constructor

=head3 Usage

    new(list)

=head3 Return

    blessed reference

=head2 DATA - list data

    if we have list data then we have been reified so return the list pointer as usual

=head2 ROOT- return the root of the identifier

    Note that the VLIST wraps iterators, the position of the list
    is maintained by the iterator that this list is wrapped by, lists themselves
    don't maintain a current position so the parent iterator position is passed down
    to the list value iterators via the POS argument

=head2 STARTI


=head2 ENDI


=head2 UNITS


=head2 LEN  - length of data array

    The LEN of the embedded list will be the greater of the index range or the length of the longest embedded list

=head2 LASTI- last index of data in list


=head2 reify

    Make the list "real", compute the list values and add them to the DATA arrary

=head3 Usage

    reify()

=head2 get

    return value of a list element.  

=head3 Usage

    get([POS=>pos])

=head3 Return

    data element at the current cursor pos

=head3 Parameters:


=over 4


=item pos

    optional, list index to return value for, by default
    will use the current list position
    

=back


=head2 copy

    copy a list 

=head3 Usage

    copy()

=head3 Return

    ref to copy of self


=head1 SEE ALSO

=head1 HISTORY

=head1 TODO

=cut

1;

