# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/CIFS.pm#1 $


## @summary Library of CIFS routines
## @author  dl-maryv (FSS Protocols QA)
## @status  public
## @pod     here

=head1 NAME

CIFS - Package of CIFS routines

=head1 DESCRIPTION

This library is just a placeholder for all CIFS routines,
a collection of CIFS routines from all libraries.

=head1 SYNPOSIS

    use CIFS qw(:all);

    my $h = host($FILER);
    my $conn = connect("#rsh");

    $ret = share_exists($conn, $share);
    
=head1 SEE-ALSO

    lib/setup::cifs_setup
    lib/ExportFS::share_exists
    lib/ExportFS::share_add
    lib/ExportFS::share_delete
    lib/MapShare::map_cifs_share
    lib/MapShare::unmap_cifs_share

=head1 FUNCTIONS

Refer the specific lib for description of each routine.

=for html 
<hr> </hr>

=cut

package CIFS;

use strict;

BEGIN {
  use Module::Common;
  Module::Common::setup_export_tags(
	setup =>  [ qw(cifs_setup) ],
	share =>  [ qw(share_exists share_add share_delete) ],
	map   =>  [ qw(map_cifs_share unmap_cifs_share) ],
	file  =>  [ qw(set_win_file_size) ],
	limits => [ qw(get_cifs_limits) ]
  );
}

use setup qw(cifs_setup);
use ExportFS qw(:cifs);
use MapShare qw(:all);
use Utils qw(make_conn);
use Utils::Units qw(:constants convert_to_bytes convert_units);
use Params::Validate qw(validate_pos validate BOOLEAN SCALAR OBJECT ARRAYREF HASHREF UNDEF);
use Tharn;
use Carp qw(croak);
use CLI::FileIO qw(rdfile wrfile);
use POSIX qw(floor);
use Options;
use Params::Common;
use Module::Common qw(validate_mixed);
use Data::Dumper;

my $options = new Options;
$options->parameters(
    definitions => [
        {  name        => 'P4_DIR',
           spec        => "s",
           default     => 'x:\agnostic',
           description => 'Defined where the P4 trees in the minwin env',
        },
    ]);

=head2 set_win_file_size($spec, $path, $size)

Sets the file size on the windows box

=over

=item Arguments

=over

=item spec

(Required) windows client spec

=item path

(Required) path to file. Needs to have double back slashes
example, "\\\\f:\\file"

=item size

(Required) desired file size in bytes. Ex 8000

=back

=item Examples: 

set_win_file_size('ncl-w2k452.lab;hosttype=windows',
                  '\\\\F:\\FILE\',8000);

=back

=for html 
<hr> </hr>

=cut

sub set_win_file_size
{
	my ($spec, $path, $size) = 
		validate_pos(@_,
			     {type => SCALAR|HASHREF },
			     {type => SCALAR, regex=>qr/\\/ }, # regex could be more strict
			     {type => SCALAR, regex=>qr/^\d+$/ }
			    );
			
	my $conn = make_conn($spec);
	my $cmd = "cmd.exe /c x:\\bin\\seteof $path $size";
	my $wait = "file size set to $size";
	
	say($conn, "$cmd", '-wait', "$wait");
	return 1;
	

}

=head2 get_cifs_limits(filer => $conn, avg_groups => $avg_groups, client => $cconn, p4_dir => "x:\agnostic")

Returns the cifs limits of a filer. 

=over

=item Description

Based on the algorithm published by cifs
this routine gets cifs limits for a particular 
filer based on it's model and memory and 
optionally the average group credentails of 
all connected users

http://wikid.netapp.com/w/CIFS_Resource_Limits#Algorithm

=item Arguments

=over

=item filer

(Required) filer hostid or connection object

=item avg_groups

(Optional) The weighted average of the number of memberships 
for users connecting over cifs.  This is used to take in to 
consideration the memory taken by SIDs in the cred associated 
with the connections. It will vary based on the number of groups 
each user opening cifs session is member of and the number of i
sessions opened by the user.

For example, if the load is run with following config: 

User   Groups  Sessions 
--------------------------------
User1  5       1
User2  10      2 
User3  15      1

avg_groups will be the weighted average, which will be
(5 * 1 + 10 * 2 + 15 * 1)/(1 + 2 + 1)

If you are going to run the entire load with single user, you can simply
set avg_groups to the number of groups the user is member of and avoid
the above calculation.

Please see burt 268164 for more information

If avg_groups = 0

The filer and Active Directory will be queried and avg_groups will be 
calculated based on the currently connected users and their local and 
AD group credentials. 

This mode requires a minwin client preferably the domain controller.

If avg_groups = 1 (default)

This will assume a non AD envrionment and base calculations on a single 
user with a single group membership.

If avg_groups > 1

Assumes the avg_groups calculation has been performed prior to calling 
the function and plugs the value directly into the remaining calulations.

=item client

(Optional) client hostid or connection object

Required if avg_groups = 0

=item p4_dir

(Optional) The path to the p4 tree.

Required if avg_groups = 0

Defaults to x:\agnostic

=back

=item Examples: 

    my %cifs_limits =  get_cifs_limits(filer => $filer);
    my $max_shares = $cifs_limits{maxshares}; 

=back

=for html 
<hr> </hr>

=cut

sub get_cifs_limits {
    $options->process();
    my $Options = $options->get_hash();
    my (%opts) =
      validate_mixed(@_,
                     [],
                     {filer  => {type => SCALAR | OBJECT, optional => 0},
                      client => {type => SCALAR | OBJECT, optional => 1},
                      avg_groups =>
                        {type => SCALAR | UNDEF, optional => 1, default => ''},
                      p4_dir => {type     => SCALAR,
                                 optional => 1,
                                 default  => $Options->{'P4_DIR'}
                        },
                      mem => {type => SCALAR | UNDEF, optional => 1},
                     }
                     );

    my $Host = host($opts{'filer'});
    my ($output,               $ontap_version, $filer,
        $Server,               $memory_size,   $memory_units,
        $adjusted_memory_size, $model_name,    $version,
        $nvmem_size,           $nvmem_units
        );
    if ($Host->hosttype =~ /ng/) {
        $Server = Server->new();
        $Server->system_node_run(-Timeout => 300,
                                 output   => \$output,
                                 'node'   => $opts{'filer'},
                                 command  => "sysconfig -a"
                                 );
    } else {
        $filer = make_conn($opts{'filer'});
        (undef, undef, $output) =
          say($filer, "sysconfig -a", -timeout, 300, -noverbose);
    }

    my @sysconfig_output = split /\n/, $output;
    if ($output =~ /.*Model\s+Name:\s+(\w+\d+)/) {
        $model_name = ($1);
    } else {
        croak("Could not parse sysconfig output for model name");
    }
    if ($output =~ /.*Memory\s+Size:\s+(\d+)\s+(\w+).*/) {
        $memory_size  = $1;
        $memory_units = $2;
        $memory_units =~ s/MB/M/;

    } else {
        croak("Could not parse sysconfig output for memory size");
    }

    if ($output =~ /.*NVMEM\s+Size:\s+(\d+)\s+(\w+).*/) {
        $nvmem_size  = $1;
        $nvmem_units = $2;
        $nvmem_units =~ s/MB/M/;
    }

    # NVMEM Size:         512 MB of Main Memory Used
    if ($output =~ /NetApp Release (.+)/) {
        $version = $1;
    } else {
        croak("Could not parse sysconfig output for ONTAP version");
    }

    if ($opts{'mem'}) {
        $memory_size  = $opts{'mem'};
        $memory_units = 'M';
    }

    #cifs limits differ as per the model number and memory size.
    #from the above algorithm
    $adjusted_memory_size = $memory_size;
    if ($model_name =~ /FAS250/) {

        #64mb is reserved for nvram memory
        $adjusted_memory_size -= 64;
    }
    if ($model_name =~ /FAS270/) {

        #128mb is reserved for nvram memory
        $adjusted_memory_size -= 128;
    }
    if ($model_name =~ /FAS2040/) {
        if ($version !~ /^(R8|8).0/ || $version =~ /7-Mode/) {

            # Set usable memory. See burt 365988
            $memory_size = 2816;

            #256MB reserved for nvram memory in IC.X
            $adjusted_memory_size = $memory_size - 256;

        } else {

            #512MB reserved for nvram memory in IC.X
            $adjusted_memory_size -= 512;
        }
    } ## end if ($model_name =~ /FAS2040/)

    my $msg;
    if ($opts{'avg_groups'} !~ /\d+/ && $Host->hosttype !~ /ng/) {
        $Log->comment(  "avg_groups not defined, calculating avg_groups"
                      . " based on connected users");

        #---------------------------------------------------------------------
        # Per Raghavendra Chowdary get the the largest number of domain group
        # memberships and if any exceed the default of one group membershipts
        # use that in the avg_group calculations
        #---------------------------------------------------------------------
        $Log->comment("Getting domain group membership for connected users");
        my $client = make_conn($opts{'client'});

        my (undef, undef, $output) = say($filer, "cifs sessions",
                             -timeout, 3600, -maxbufsize, 25000000, -noverbose);
        my @lines = split /\n/, $output;
        my %sessions = ();

        # 172.16.12.135(KC28B9-E0) (NAS\administrator - pcuser)
        foreach my $line (@lines) {
            if ($line =~ /\(\w+\\(\S+)\s+/) {
                my $user = $1;
                if (!grep(/$user/, @{$sessions{'users'}})) {
                    push(@{$sessions{'users'}}, $user);
                    $sessions{$user}{'sessions'} = 1;
                }
                $sessions{$user}{'sessions'}++;
            }
        } ## end foreach my $line (@lines)
        my $totalSessions     = 0;
        my $totalCredentials  = 0;
        my $calulateAvgGroups = 0;
        foreach my $user (@{$sessions{'users'}}) {
            $sessions{$user}{'groups'} = 0;
            for (my $i = 1 ; $i <= 3 ; $i++) {
                my $cmd =
                  "cmd /c cscript $opts{'p4_dir'}\\tools\\windows\\vbs\\listGroupMemberships.vbs $user";
                my (undef, undef, $output) = say($client, $cmd, -timeout, 300);
                if ($output =~ /Group Memberships: (\d+)/) {
                    $sessions{$user}{'groups'} = $1;
                    last;
                }
                sleep 10;
            } ## end for (my $i = 1 ; $i <= ...

            $totalSessions += $sessions{$user}{'sessions'};
            $totalCredentials +=
              ($sessions{$user}{'sessions'} * $sessions{$user}{'groups'});
            if ($sessions{$user}{'groups'} > 1) { $calulateAvgGroups = 1; }
        } ## end foreach my $user (@{$sessions...

	#<<<
        $msg  = "Domain Information\n";
        $msg .= "  Total Sessions  . . : $totalSessions\n";
        $msg .= "  Users Found . . . . : " . @{$sessions{'users'}} . "\n";
        $msg .= "  Total Credentials . : $totalCredentials\n";
        #>>>
        $Log->comment($msg);

        if ($calulateAvgGroups) {
            $opts{'avg_groups'} = floor($totalCredentials / $totalSessions);
        } else {
            $Log->comment( "No users connected or all users have the "
                         . "default group membership. Setting avg_groups to 0");
            $opts{'avg_groups'} = 0;
        }

    } elsif ($opts{'avg_groups'} == 0 && $Host->hosttype !~ /ng/) {
        $Log->comment(  "Using user specified value of "
                      . "$opts{'avg_groups'} for avg_groups");
    }

    #Create a hash of cifs limits.
    my %cifs_limits;

    $cifs_limits{'num_lcl_groups'}      = 0;
    $cifs_limits{'num_lcl_memberships'} = 0;
    if ($Host->hosttype !~ /ng/) {

        $Log->comment("Getting local group info");
        my $data = rdfile($filer, "/etc/lclgroups.cfg");
        my @lines = split(/\n/, $data);
        for (my $i = 0 ; $i < @lines ; $i++) {
            my $line = $lines[$i];
            if ($line =~ /\[\s+"/) {
                $cifs_limits{'num_lcl_groups'}++;
            } elsif ($line =~ /^S-/) {
                $cifs_limits{'num_lcl_memberships'}++;
            }
        }

        $Log->comment("Getting local user info");
        $cifs_limits{'num_lcl_users'} = 0;
        my $data = rdfile($filer, "/etc/passwd");
        my @lines = split(/\n/, $data);
        for (my $i = 0 ; $i < @lines ; $i++) {
            my $line = $lines[$i];
            if ($line =~ /^\w+/) {
                $cifs_limits{'num_lcl_users'}++;
            }
        }

    } ## end if ($Host->hosttype !~...

    $cifs_limits{'num_mem_bytes'} =
         convert_to_bytes($adjusted_memory_size . $memory_units)
      or croak("$@");

    if (   $cifs_limits{'num_lcl_groups'} > 7
        && $cifs_limits{'num_lcl_memberships'} > 2
        && $Host->hosttype !~ /ng/) {

        $Log->comment(  "Non standard local group configuration detected. "
                      . "Calculating local group usage");
        $cifs_limits{'lcl_user_group_mem'} =
          (100 * $cifs_limits{'num_lcl_groups'} +
            130 * $cifs_limits{'num_lcl_memberships'} +
            1000 * $cifs_limits{'num_lcl_users'});
    } else {
        $cifs_limits{'lcl_user_group_mem'} = 0;
    }

    $cifs_limits{'num_64mb_mem_blocks'} =
      floor(($cifs_limits{'num_mem_bytes'} - $cifs_limits{'lcl_user_group_mem'})
            >> 26);

    if ($cifs_limits{'num_64mb_mem_blocks'} > 96) {
        $cifs_limits{'num_64mb_mem_blocks'} = 96;
    }

    if ($model_name =~ /F7|FAS2|FAS6/) {
        $cifs_limits{'max_users_algorithm'} =
          "((64 MB Blocks - 2) * 1000) + 1200";
        $cifs_limits{'maxusers'} =
          (($cifs_limits{'num_64mb_mem_blocks'} - 2) * 1000) + 1200;
    } elsif ($model_name =~ /R1|F8|FAS9|FAS3/) {
        $cifs_limits{'max_users_algorithm'} =
          "((64 MB Blocks - 3) * 1000) + 3000";
        $cifs_limits{'maxusers'} =
          (($cifs_limits{'num_64mb_mem_blocks'} - 3) * 1000) + 3000;
    }

    $cifs_limits{'avg_groups_algorithm'} =
      'UserN Groups * UserN Sessions + User(N+1) Groups * User(N+1) Sessions...) / Total Sessions';
    $cifs_limits{'local_user_and_group_memory_algorithm'} =
      '100 * Local Groups + 130 * Local Group Memberships + 1000 * Local Users';
    $cifs_limits{'num_64mb_mem_blocks_algorithm'} =
      '(Bytes - Local User And Group Memory) >> 26';

    if ($version !~ /^(R8|8).0/ || $version =~ /7-Mode/) {
        $cifs_limits{'max_connections_algorithm'} =
          '(Max Users * 1700) / (1400 + Average Groups * 100)';
        $cifs_limits{'maxlockedfiles_algorithm'} =
          'Max Open Files + (64 MB Blocks * 2048)';
        $cifs_limits{'maxshareconn_algorithm'} = '4 * Max Users';
        $cifs_limits{'maxopenfiles_algorithm'} = '20 * Max Users';
        $cifs_limits{'maxlocks_algorithm'}     = 'Max Locked Files * 2';
        $cifs_limits{'maxshares_algorithm'}    = '2 * Max Users';

        if ($opts{'avg_groups'} == 0) {
            $cifs_limits{'maxconnections'} = $cifs_limits{'maxusers'};
        } else {
            $cifs_limits{'maxconnections'} =
              floor(($cifs_limits{'maxusers'} * 1700) /
                    (1400 + ($opts{'avg_groups'} * 100)));
        }

        $cifs_limits{'maxshares'}    = 2 * $cifs_limits{'maxusers'};
        $cifs_limits{'maxshareconn'} = 4 * $cifs_limits{'maxusers'};
        $cifs_limits{'maxopenfiles'} = 20 * $cifs_limits{'maxusers'};
        $cifs_limits{'maxlockedfiles'} =
          $cifs_limits{'maxopenfiles'} +
          ($cifs_limits{'num_64mb_mem_blocks'} * 2048);
        $cifs_limits{'maxlocks'} = $cifs_limits{'maxlockedfiles'} * 2;
    } elsif ($version =~ /^(R8|8).0/ && $version =~ /Cluster-Mode/) {
        $cifs_limits{'max_connections_algorithm'} = 'Hard coded';
        $cifs_limits{'maxlockedfiles_algorithm'}  = 'Hard coded';
        $cifs_limits{'maxshareconn_algorithm'}    = 'Hard coded';
        $cifs_limits{'maxopenfiles_algorithm'}    = 'Hard coded';
        $cifs_limits{'maxlocks_algorithm'}        = 'Hard coded';
        $cifs_limits{'maxshares_algorithm'}       = 'Hard coded';
        $cifs_limits{'max_users_algorithm'}       = 'Hard coded';
        $cifs_limits{'maxshares'}                 = 32000;
        $cifs_limits{'maxshareconn'}              = 131070;
        $cifs_limits{'maxopenfiles'}              = 131070;

        if (convert_units($memory_size, UNIT_MB, UNIT_GB) < 8) {
            $cifs_limits{'maxconnections'} = 3753;
            $cifs_limits{'maxlockedfiles'} = 1411072;
            $cifs_limits{'maxlocks'}       = 2822144;
        } else {
            $cifs_limits{'maxconnections'} = 10000;
            $cifs_limits{'maxlockedfiles'} = 2116608;
            $cifs_limits{'maxlocks'}       = 4233216;
        }
    } ## end elsif ($version =~ /^(R8|8).0/...
    #<<<
    my $algorithm;
    my $msg = "CIFS Limit Information\n";
    $msg .= "  Version . . . . . . . . : $version\n";
    $msg .= "  Model . . . . . . . . . : $model_name\n";
    $msg .= "  Memory/Bytes  . . . . . : " . $memory_size . $memory_units . "/" . convert_to_bytes($memory_size . $memory_units) . "\n";

    if ($nvmem_size) {
        $msg .= "  NVMEM/Bytes . . . . . . : " . $nvmem_size . $nvmem_units . "/" . convert_to_bytes($nvmem_size . $nvmem_units) . "\n";
        $algorithm = "  Memory - NVRAM"; 
	$msg .= "  Adjusted Memory/Bytes . : " . $adjusted_memory_size . $memory_units . "/$cifs_limits{'num_mem_bytes'} ($algorithm)\n";
    } ## end if ($nvmem_size)
    $msg .= "  Average Groups  . . . . : $opts{'avg_groups'} ($cifs_limits{'avg_groups_algorithm'})\n";
    $msg .= "  Local Users . . . . . . : $cifs_limits{'num_lcl_users'}\n";
    $msg .= "  Local Groups  . . . . . : $cifs_limits{'num_lcl_groups'}\n";
    $msg .= "  Local Group Memberships : $cifs_limits{'num_lcl_memberships'}\n";
    $msg .= "  Local User/Group Memory : $cifs_limits{'lcl_user_group_mem'} = ($cifs_limits{'local_user_and_group_memory_algorithm'})\n";
    $msg .= "  64 MB Blocks  . . . . . : $cifs_limits{'num_64mb_mem_blocks'} = ($cifs_limits{'num_64mb_mem_blocks_algorithm'})\n";
    $msg .= "  Max Users . . . . . . . : $cifs_limits{'maxusers'} = ($cifs_limits{'max_users_algorithm'})\n";

    if ($version !~ /8.0/ || $version =~ /7-Mode/) {
        $msg .= "  Max Connections . . . . : $cifs_limits{'maxconnections'} = ($cifs_limits{'max_connections_algorithm'})\n";
        $msg .= "  Max Shares  . . . . . . : $cifs_limits{'maxshares'} = ($cifs_limits{'maxshares_algorithm'})\n";
        $msg .= "  Max Share Connections . : $cifs_limits{'maxshareconn'} = ($cifs_limits{'maxshareconn_algorithm'})\n";
        $msg .= "  Max Open Files  . . . . : $cifs_limits{'maxopenfiles'} = ($cifs_limits{'maxopenfiles_algorithm'})\n";
        $msg .= "  Max Locked Files  . . . : $cifs_limits{'maxlockedfiles'} = ($cifs_limits{'maxlockedfiles_algorithm'})\n";
        $msg .= "  Max Locks . . . . . . . : $cifs_limits{'maxlocks'} = ($cifs_limits{'maxlocks_algorithm'})";
    } elsif ($version =~ /8.0/ && $version =~ /Cluster-Mode/) {
        $msg .= "  Max Connections . . . . : $cifs_limits{'maxconnections'} = ($cifs_limits{'max_connections_algorithm'})\n";
        $msg .= "  Max Shares  . . . . . . : $cifs_limits{'maxshares'} = ($cifs_limits{'maxshares_algorithm'})\n";
        $msg .= "  Max Share Connections . : $cifs_limits{'maxshareconn'} = ($cifs_limits{'maxshareconn_algorithm'})\n";
        $msg .= "  Max Open Files  . . . . : $cifs_limits{'maxopenfiles'} = ($cifs_limits{'maxopenfiles_algorithm'})\n";
        $msg .= "  Max Locked Files  . . . : $cifs_limits{'maxlockedfiles'} = ($cifs_limits{'maxlockedfiles_algorithm'})\n";
        $msg .= "  Max Locks . . . . . . . : $cifs_limits{'maxlocks'} = ($cifs_limits{'maxlocks_algorithm'})\n";
    } ## end elsif ($version =~ /8.0/ ...
    #>>>
    $Log->comment($msg);
    return (%cifs_limits);

} ## end sub get_cifs_limits
    
    
1;

