#Hey emacs, this is basically -*- perl -*-
# $Id$
#
## @summary abe - Library for the abe tests
## @author	dl-maryv (FSS Protocols QA)
## @status	public
## @notes	Created by Vladimir Yakubov
## @pod here

=head1 NAME

ABE - Library for the Access Based Enumeration

=head1 SYNOPSIS

  use CIFS::Abe;

  abe_setup(qtree_security => 'unix', windows_version => '5.1',
                       setup_h => \%setup_h);

  abe_cleanup(cifs_drive => 'T:');
  
  set_permissions (target_drive => $tdrive, target_directory => $tdir,
  					target_file	=> $tfile, perm => $perm);

  read_dir_cmd (target_drive => $tdrive, target_directory => $tdir,
  					target_file	=> $tfile, client => $CLIENT);

=head1 DESCRIPTION

Provides the common routines for the ABE tests.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id$
  Last revised $Date$

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package CIFS::Abe;

use strict;
use warnings;
use base qw(Exporter);

use Tharn qw(host connect logresult say logcomment disconnect script_result matchstrings);
use Carp qw(croak);
use FS::Util::Setup qw(create_vol common_cleanup);
use MapShare qw(map_cifs_share unmap_cifs_share);
use ExportFS qw(share_add share_delete);
use Users qw(addPasswdEntry addGroupEntry delPasswdEntry delGroupEntry);
use SPARSE::Util::Logger;
use Network::Utils qw(run_getXXbyYY);
use filer qw(filer_options_set filer_qtree_security_set filer_vol_snap_delete);
use Params::Validate qw(validate ARRAYREF HASHREF SCALAR UNDEF);
use POSIX qw(uname);
use IO::Seekable qw(SEEK_SET);
use File::Basename qw(basename);
use CIFS::Ntfs qw(cacls_get cacls_compare);
use CIFS::Setup qw(cifs_setup);
use CIFS::CommonParams qw(:all);
use CIFS::widelink qw(deleteFile);
use SayWrap qw(sayEOF);
use Admin::License qw(license);
use File::Find qw(find);
use TestEnv::MinWin qw(minwin_compatible execute_cmd);
use Utils qw(make_host make_conn);
use Volume qw(:all);
use mount qw(mount umount mounts);
#use Win32::FileSecurity qw(Get EnumerateRights);


BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our $VERSION = '1.0';

# Test file and directory names
#my $test_file_name = 'vss-clt-001.txt';
#my $test_dir_name = 'tdir';
#my $test_file_msg = 'this file contains the string vss-clt-001';

# Variables used by the setup and cleanup routine
my @filer_vol_path;
my @create_vol_called;
my @fc_users_created;
my @fc_groups_created;
my @cifs_share_completed;
my @cifs_map_completed;
my @qtree_security_completed;
my %test_files;

# For CIFS shares to be accessable the $CIFS_USER account
# must exist from the Unix side of the filer. We just add
# the $CIFS_USER account to /etc/passwd and the root
# group to /etc/group.
my $admin_group = 'root';

# Functions to be exported
our @EXPORT_OK = qw(
    abe_setup
    abe_cleanup
	turn_on_abe_on_share
	set_permissions
	read_dir_cmd
	clean_share
	create_dataset
	change_perms
	setup_homedirs
	change_dir
	open_file
	rename_file_dir
	delete_file_dir
	create_file_dir
	type_file
	read_dir_gui
	change_dir_gui
	open_file_gui
	make_file_gui
	create_widelink
	change_qtree_security
);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub abe_setup(%);
sub abe_cleanup(%);
sub turn_on_abe_on_share(%);
sub set_permissions(%);
sub read_dir_cmd(%);
sub create_dataset(%);
sub change_perms(%);
sub setup_homedirs(%);
sub change_dir(%);
sub open_file(%);
sub rename_file_dir(%);
sub delete_file_dir(%);
sub create_file_dir(%);
sub type_file(%);
sub read_dir_gui(%);
sub change_dir_gui(%);
sub open_file_gui(%);
sub make_file_gui(%);
sub create_widelink(%);
sub change_qtree_security(%);

#---------------------------------------------------------------------

=head2 abe_setup

abe_setup(setup_h => \%setup_h
          [, qtree_security => $security ]
          [, windows_version => $win_version ]);

The following named arguments are accepted:

=over

=item Arguments

=over

=item setup_h

(Required) : Reference to setup hold hash for cleanup.

=item qtree_security

(Optional) : Security setting for a qtree. Defaults to unix.

=item windows_version

(Optional) : MS Windows version number

=back

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub abe_setup(%) {
    my %opts = validate(
        @_,
        {
            setup_h         => { type => HASHREF },
            qtree_security  => { type => SCALAR, default => 'mixed',
                                 regex => qr/^(mixed|ntfs|unix)$/ },
            windows_version =>  { type => SCALAR, default => '5',
                                 regex => qr/^(5\.0|5\.1|5\.2|5)$/ }
        }
    );
    my ($cifs_drive, $exti, $fc, $host, $i, $op, @osrel, @triples, $volume);

    logStep("BEGIN: abe_setup: setup the environment");

    $host = host('localhost');
    logStep("abe_setup: Verify client has correct Windows version");
    @osrel = uname();
    if ($osrel[2] !~ /$opts{windows_version}/) {
        croak("client running $osrel[2] expected $opts{windows_version} " .
              "windows release");
    }

    # The default filer, vol, and share to use are set first. If additional
    # triples were passed in then process them in this loop also.
    @triples = ({ filer => $main::FILER, vol => $main::FILER_VOL,
                  share => $main::CIFS_SHARE });
    if (defined($opts{setup_h}->{ext_setup})) {
        push @triples, @{$opts{setup_h}->{ext_setup}};
    }
    $opts{setup_h}->{ext_setup} = \@triples;

    $i = 0;
    foreach $exti (@triples) {
        logStep("abe_setup: Create volume $exti->{vol}");
        $create_vol_called[$i] = 1;
        if ($i == 0) {
            $op = $main::EXTERN_SETUP;
        } else {
            $op = 0;
        }
        ($fc, $volume) = create_vol(filer => $exti->{filer},
                                    filer_vol => $exti->{vol},
                                    extern_setup => $op,
                                    guarantee => 'none') or croak($@);
        $exti->{fc} = $fc;
        $exti->{vol} = $volume;
        $filer_vol_path[$i] = "/vol/$exti->{vol}";
		
		logStep("abe_setup: change qtree security to mixed for the testvol_flex ");
		say($fc, "cifs terminate -t 0");
		say($fc, "qtree security /vol/testvol_flex  ntfs");

		logStep("abe_setup: Setup CIFS on the filer");
		if ($main::MODE eq "7G") {
			cifs_setup(host   => $exti->{filer}, 
			    'auth-type'   => 'ad',
		  		cifs_pass     => $main::CIFS_PASSWD,
				domain        => $main::DOMAIN,
				domain_user   => $main::CIFS_USER, 
   			   	domain_pass   => $main::CIFS_PASSWD );
		} else {
        	
        	cifs_setup(  host => $exti->{filer}, 
			   'server-name' => $exti->{filer}, 
			   'auth-type'   => 'ad',
  			   cifs_pass     => $main::CIFS_PASSWD,
			   domain        => $main::DOMAIN,
			   domain_user   => $main::CIFS_USER, 
   			   domain_pass   => $main::CIFS_PASSWD );
		}
	  sleep 50;

        share_add($fc, $exti->{share}, $filer_vol_path[$i], verify => 1,
                  purge_existing => 1, umask => 0, widelink => 1,
                  nosymlink_strict_security => 1)
            or croak("failed to export share $exti->{share}");
        $cifs_share_completed[$i] = 1;

        # Set the qtree security mode on the filer volume
        if (filer_qtree_security_set($fc, $filer_vol_path[$i],
            $opts{qtree_security}) == 0) {
            croak("QTREE set failed on $filer_vol_path[$i] for $opts{qtree_security}");
        }
        $qtree_security_completed[$i] = 1;

        #$cifs_drive = map_cifs_share(
        #    server   => $fc->connhost(),
        #    share    => $exti->{share},
        #    client   => 'localhost',
        #    drive    => '*',
        #    username => $main::CIFS_USER,
        #    password => $main::CIFS_PASSWD
        #    )
        #    or croak($@);
        #logDbg("mapped drive $cifs_drive");
        #if (!$cifs_drive) {
        #    croak("could not map the drive");
        #}
#
        #$cifs_drive .= ":" unless $cifs_drive =~ /:$/;
        #logDbg("Share $exti->{share} of $host mounted as $cifs_drive");
        #$exti->{cifs_drive} = $cifs_drive;
        #$cifs_map_completed[$i] = 1;

        # Put the default fc, vol, and cifs_drive into the setup_h also
        if ($i == 0) {
            $opts{setup_h}->{fc} = $fc;
            $opts{setup_h}->{vol} = $volume;
            $opts{setup_h}->{cifs_drive} = $cifs_drive;
        }

        $i++;
    }
	
    logStep("END: abe_setup:");
    return;
}

#---------------------------------------------------------------------

=head2 abe_cleanup

abe_cleanup([dirs        => \@dirs]
            [,files      => \@files]
            [,cifs_drive => $cifs_drive]
            [,options    => \%options]
            [,snapshots  => \@snapshots]
            [,ext_setup  => \@ext_setup);

Cleanup after a test run if necessary. Use this cleanup
routine if you use abe_setup.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dirs

(Optional) : List of directories to cleanup on exit.

=item files

(Optional) : List of files to cleanup on exit.

=item cifs_drive

(Optional) : Drive letter where share is mapped to.

=item options

(Optional) : Hash of options and values to restore

=item snapshots

(Optional) : List of snapshots to cleanup on exit.

=item ext_setup

(Optional) : List of {cifs_drive}, {snapshots}, and {symfile} for
extended setup of {fc}, {vol}, and {share}.

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub abe_cleanup(%) {
    my %opts = validate(
        @_,
        {
            dirs       => { type => ARRAYREF | UNDEF, optional => 1 },
            files      => { type => ARRAYREF | UNDEF, optional => 1 },
            cifs_drive => { type => SCALAR | UNDEF, optional => 1 },
            options    => { type => HASHREF | UNDEF, optional => 1 },
            snapshots  => { type => ARRAYREF | UNDEF, optional => 1 },
            vol        => { type => SCALAR | UNDEF, optional => 1 },
            fc         => { type => HASHREF | UNDEF, optional => 1 },
            ext_setup  => { type => ARRAYREF | UNDEF, optional => 1 },
			clean_homedir  => { type => SCALAR , default => 0 }
        }
    );
    my (@arr, $cleanup, $dir, $exti, $fc, $fh, $filer_vol, $i, $snapshot);
	my $homedir_vol = $main::FILER_VOL;
	
    logStep("CLEANUP == $main::CLEANUP");

    eval {


        $i = 0;
        foreach $exti (@{$opts{ext_setup}}) {
            logStep("abe_cleanup: Reconnect filer $exti->{filer}");
            eval {
                $fh = host($exti->{filer});
                $fc = connect($main::FILER_CONN, -timeout, $main::TIMEOUT);
                $fc->timeout($main::CMDTIMEOUT);
            };
            if ($@) {
                undef $fc;
                logWarn("abe_cleanup: $@");
            }

            if ($main::CLEANUP ne 'none') {
                # cleanup files only for the first filer on the list
                if ($i == 0 && $cifs_map_completed[$i]) {
                    if (defined($opts{files}) && @{$opts{files}} > 0) {
                        # The unlink would fail if somebody accidently stuck
                        # "/" as a file because it is a directory so there is
                        # no need to look for it on the list.
                        logStep("abe_cleanup: removing files");
                        unlink @{$opts{files}} or
                            logDbg("abe_cleanup: unlink failed: $!");
                        @{$opts{files}} = ();
                    }

                    if (defined($opts{dirs}) && @{$opts{dirs}} > 0) {
                        logStep("abe_cleanup: removing directories");
                        # Traverse array in reverse order so that the directories
                        # are removed newest to oldest.
                        while ($dir = pop @{$opts{dirs}}) {
                            # Remove the directory
                            rmdir $dir or
                                logDbg("abe_cleanup: rmdir $dir failed: $!");
                        }
                        @{$opts{dirs}} = ();
                    }
                }
				
				if ($opts{clean_homedir} == 1) {
					logStep("abe_cleanup: turn off ABE on cifs.homedir share");
					turn_on_abe_on_share(filer => $main::FILER, share => 'cifs.homedir', abe_on_off => 0);
				}
				logStep("abe_cleanup: removing files/dirs on the testshare");
				clean_share(client => $main::CLIENT, clean_homedir_share => $opts{clean_homedir});

                if ($cifs_map_completed[$i] && $exti->{cifs_drive}) {
                    logStep("abe_cleanup: unmapping CIFS drive: $exti->{cifs_drive}");
                    eval {
                        unmap_cifs_share(mappath => $exti->{cifs_drive}, force => 1);
                    };
                    logWarn("abe_cleanup: $@") if $@;
                }

                if ($fc && $cifs_share_completed[$i]) {
					if ($opts{clean_homedir} == 1) {
						$exti->{share} = $main::HOMEDIR_SHARE;
					} 
					logStep("abe_cleanup: delete CIFS share: $exti->{share}");
					eval {
						share_delete($fc, $exti->{share});
					};
                    logWarn("abe_cleanup: $@") if $@;
                }

                if ($fc && $fc_users_created[$i]) {
                    logStep("abe_cleanup: Remove users added to filer");
                    delPasswdEntry($fc, $fc_users_created[$i]);
                    logWarn("abe_cleanup: $@") if ($@);
                }

                if ($fc && $fc_groups_created[$i]) {
                    logStep("abe_cleanup: Remove groups added to filer");
                        delGroupEntry($fc, $fc_groups_created[$i]) or
                            logWarn("abe_cleanup: $@");
                }

                # restore options only for the first filer on the list
                @arr = keys(%{$opts{options}});
                if ($i == 0 && $fc && defined($opts{options}) && $#arr >= 0) {
                    logStep("abe_cleanup: restoring options");
                    eval {
                        foreach $i (@arr) {
                           filer_options_set($fc, $i, ${$opts{options}}{$i});
                        }
                    };
                    logWarn("abe_cleanup: $@") if $@;
                }
            }

            # Delete /etc/symlink.translations file and symlink file
            if ($exti->{symfile}) {
                eval {
                    deleteFile($fc->{hostp}->{id}, $fc->connhost(),
                               '/vol/$homedir_vol/etc/symlink.translations');
                };
                logWarn("abe_cleanup: $@") if $@;
                eval {
                    deleteFile($fc->{hostp}->{id}, $fc->connhost(),
                               $exti->{symfile});
                };
                logWarn("abe_cleanup: $@") if $@;
            }

            # Call common cleanup if create_vol was called
            if ($fc && $create_vol_called[$i]) {

                logStep("abe_cleanup: Run common cleanup on vol $exti->{vol}");
                if ($i == 0 || $main::CLEANUP =~ /none|volume/) {
                    $cleanup = $main::CLEANUP;
                } else {
                    # If not the default volume then (test|all) map to volume
                    # for secondary volume cleanup.
                    $cleanup = 'volume';
                }
                # common_cleanup has purge of FILER_VOL hardcoded in so undef
                # it here
                $filer_vol = $main::FILER_VOL;
                undef $main::FILER_VOL;
                common_cleanup($fc, cleanup => $cleanup, vol => $exti->{vol});
                $main::FILER_VOL = $filer_vol;
                logWarn("abe_cleanup: $@") if $@;
            }
            $i++;
        }
    };

    logWarn("abe_cleanup: $@") if $@;
    logStep("END: abe_cleanup:");
    return;
}

#---------------------------------------------------------------------

=head2 test_dir_and_file_create

%results = test_dir_and_file_create(target_drive => $tdrive
                                   [, huge       => $huge ]
                                   [, many       => $many ]);

%results = (
           dirs  => reference to array of dirs that were created,
           files => reference to array of files that were created,
           dir_tmstmp => timestamp on the directory,
           file_tmstmp => timestamp on the file
)

Create a test directory and file on the FILER volume.

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive for directory and file creation

=item test_dir_name

(Optional) : name of the directory to create

=item test_file_name

(Optional) : name of the file to create

=item test_message

(Optional) : message to write to the file

=item huge

(Optional) : whether a huge file should be created in the directory

=item many

(Optional) : whether many files and directories should be created in the
directory

=back

=item Returns

Results hash containing information about the test files and directories

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub test_dir_and_file_create(%) {
    my %opts = validate(
        @_,
        {
            target_drive => { type => SCALAR, regex => qr/^[A-Za-z]\:$/ },
			test_dir_name => { type => SCALAR, default => ""},
			test_file_name => { type => SCALAR, default => ""},
			test_message => { type => SCALAR, default => ""},
        }
    );

    my (@dirs, @files, @stat_arr, $dir_tmstmp, $file_tmstmp, $i, $j, $tpath);

    # don't catch error on mkdir because directory may already exist
	logStep("test_dir_and_file_create: Creating directory $opts{target_drive}/$opts{test_dir_name}");
    mkdir("$opts{target_drive}/$opts{test_dir_name}");
    sleep 5;

	#my file_name = $opts{test_file_name}
	if ($opts{test_file_name} ne "") {
		open(OUTFILE, ">$opts{target_drive}/$opts{test_dir_name}/$opts{test_file_name}")
			or croak("Couldn't create file: $!");

		print OUTFILE "$opts{test_message}\n"
			or croak("Couldn't write to file: $!");

		if ($opts{huge}) {
			seek(OUTFILE, 10000000000, SEEK_SET);
			print OUTFILE "end of file.\n";
		}

    	close(OUTFILE);
	}

    return;
}

#---------------------------------------------------------------------

=head2 turn_on_abe_on_share

$results = turn_on_abe_on_share(filer => $filer,
								share => $share_name, 
								abe_on_off => $abe_on_off);

%results = (0|1 - success or failure of operation)

Turn on ABE on a give share.

The following named arguments are accepted:

=over

=item Arguments

=over

=item filer

(Required) : filer where the share lives

=item share

(Required) : name of the share on which to turn on ABE

=item abe_on_off

(Required) : boolean with 1 to turn on and 0 to turn off ABE

=back

=item Returns

Results 1 or 0 depending on operation success

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub turn_on_abe_on_share(%) {
    my %opts = validate(
        @_,
        {
            filer => { type => SCALAR},
			share => { type => SCALAR},
			abe_on_off => { type => SCALAR, regex => qr/(0|1)/},
        }
    );

    my $result = 0;

    my $rsh_conn = make_conn($opts{filer});
	
	if($opts{abe_on_off}) {
		logStep("turn_on_abe_on_share: Turning on ABE on $opts{share}");
    	my ($out1, $out2, $out3) = say($rsh_conn, "cifs shares -change $opts{share} -accessbasedenum");
		if($out3 =~ /have been successfully modified/) {
			$result = 1;
		} else {
			$result = 0;
		}
	} else {
		logStep("turn_on_abe_on_share: Turning off ABE on $opts{share}");
		my ($out1, $out2, $out3) = say($rsh_conn, "cifs shares -change $opts{share} -noaccessbasedenum");
		if($out3 =~ /have been successfully modified/) {
			$result = 1;
		} else {
			$result = 0;
		}
	}
	return $result;
}

#---------------------------------------------------------------------

=head2 set_permissions

set_permissions(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				perm    => $perm,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=item perm

(Required) : permission bit which to set.

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub set_permissions(%) {
    my %opts = validate(
        @_,
        {
            target_drive  	=> { type => SCALAR, regex => qr/^[A-Za-z]\:$/ },
			target_directory => { type => SCALAR },
			target_file   	=> { type => SCALAR, default => "" },
			perm    	  	=> { type => SCALAR }
        }
    );
	
	my ($win, %res1, %res2, $res);

    require CIFS::Abe_gui;
    import CIFS::Abe_gui qw(find_target_my_computer verify_previous_version_tab
                            close_my_computer explorer_open_with_target set_security);
	
	
	# Gets the rights for all files listed on the command line.
    
    
	my $test_dir_name = $opts{target_directory};
	my $test_file_name = $opts{target_file};
	
	logStep("set_permissions: Changing permissions on $opts{target_drive}/$opts{target_directory} to $opts{perm}");

    $win = explorer_open_with_target(target_text => $opts{target_drive});

    $win = find_target_my_computer(target_text => $test_dir_name, window => $win);
	
	set_security(target_text => $test_dir_name, window => $win);
   
    return;
}
#---------------------------------------------------------------------

=head2 read_dir_cmd

read_dir_cmd(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub read_dir_cmd(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR },
			cifs_passwd => { type => SCALAR },
			file => { type => SCALAR },
        }
    );
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype);

	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{share};
	my $fh = host($main::FILER);
	my $filer_ip = $fh->default_ip();

	logStep("read_dir_cmd: mapping CIFS drive");
	$cifs_drive = map_cifs_share(
		server   => $filer_ip,
		share    => $share,
		client   => 'localhost',
		drive    => '*',
		username => $opts{cifs_user},
		password => $opts{cifs_passwd}
		)
		or croak($@);
	logDbg("mapped drive $cifs_drive");
	logStep("read_dir_cmd: successfully mapped filer $filer_ip\\$share to CIFS drive $cifs_drive");
	if (!$cifs_drive) {
		croak("could not map the drive");
	}
	foreach my $file (split(" ", $opts{file})){
		my ($path, $found) = split(",", $file);
	
		if ($found == 1) {
			logStep("read_dir_cmd: Checking the existence of the file $cifs_drive\\$path, we are expecting to find it");
		} else {
			logStep("read_dir_cmd: Checking the existence of the file $cifs_drive\\$path, we are not expecting to find it");
		}		
		sleep 5;

		my ($out1, $out2, $out3) = say($rsh_conn, "dir $cifs_drive\\$path");
		if ($out3 =~ /File Not Found/) {
			logStep("read_dir_cmd: File $cifs_drive\\$path was not found");
			if ($found == 1) {
				logresult('FAIL', "We were expecting to find the directory/file $cifs_drive\\$path and but not found it");
				$result = 0;
			} else {
				logresult('PASS', "We were not expecting to find the directory/file $cifs_drive\\$path and did not see it");
				$result = 1;
			}
		} else {
			logStep("read_dir_cmd: File $cifs_drive\\$path was found");
			if ($found == 1) {
				logresult('PASS', "We were expecting to find the directory/file $cifs_drive\\$path and found it");
				$result = 1;
			} else {
				logresult('FAIL', "We were not expecting to find the directory/file $cifs_drive\\$path and but found it");
				$result = 0;
			}
		}	
	}
	
	logStep("read_dir_cmd: unmapping CIFS drive $cifs_drive");
	eval {
		unmap_cifs_share(mappath => $cifs_drive, force => 1);
	};
	logWarn("read_dir_cmd: $@") if $@;
   
    return $result;
}
#---------------------------------------------------------------------

=head2 clean_share

clean_share(target_drive    => $tdrive,
			client => $CLIENT,
			);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub clean_share(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			clean_homedir_share => { type => SCALAR }
        }
    );
	
	my $home_dir = $main::HOMEDIR;

	my ($rsh_conn, @dir, $line, $directory, $drive_letter, $share, $username, $password, $result);
	$rsh_conn = make_conn($opts{client});
	
	
	if ($opts{clean_homedir_share} == 1) {
		$share    = $home_dir;			
	} else {
		$share    = $main::CIFS_SHARE;
		
	}
	my $fh = host($main::FILER);
	my $filer_ip = $fh->default_ip();
	my $filer_conn = make_conn($main::FILER);
	
	$username = $main::CIFS_USER;
	$password = $main::CIFS_PASSWD;
	
	$drive_letter = map_cifs_share(
		server   => $filer_ip,
		share    => $share,
		client   => 'localhost',
		drive    => '*',
		username => $username,
		password => $password
		)
		or croak($@);
	logDbg("clean_share: mapped drive $drive_letter");
	logStep("clean_share: successfully mapped share $main::FILER\\$share to CIFS drive $drive_letter");
	if (!$drive_letter) {
		croak("could not map the share");
	}
	
	my $drive_letter_saved = $drive_letter;
	
	if ($opts{clean_homedir_share} == 1) {
		$drive_letter    = join ("", $drive_letter, '\\', $main::HOMEDIR_SHARE);			
		say($filer_conn, "priv set diag; rm /etc/cifs_homedir.cfg");
		say($filer_conn, "cifs homedir load");
	}

	logStep("clean_share: Deleting files and directories on drive $drive_letter");
	my ($out1, $out2, $out3) = say($rsh_conn, "del $drive_letter\\* /F /Q");
	
		
	($out1, $out2, $out3) = say($rsh_conn, "dir $drive_letter");
	@dir = split("\r", $out3);
	foreach $line (@dir) {
		if ($line =~ /<DIR>/){
			my @tmp_line = split (" ", $line);
			$directory = $tmp_line[4];
			($out1, $out2, $out3) = say($rsh_conn, "rmdir /S /Q $drive_letter\\$directory");
			if ($out3 =~ /The directory is not empty/) {
				change_perms(client => $main::CLIENT, path => $directory, user => $main::TEST_CIFS_USER, perm => 'F', drive_letter => $drive_letter);
				logStep("clean_share: Second try at deleting directory $drive_letter\\$directory");
				($out1, $out2, $out3) = say($rsh_conn, "rmdir /S /Q $drive_letter\\$directory");
				if ($out3 =~ /The directory is not empty/) {
					logStep("clean_share: Unable to delete directory $drive_letter\\$directory");
				}
			}
		}
	}

	($out1, $out2, $out3) = say($rsh_conn, "dir $drive_letter");
	if ($out3 =~ /File Not Found/) {
		$result = 1;
	} else {
		$result = 0;
	}
	
	logStep("clean_share: unmapping CIFS share $drive_letter_saved");
	eval {
		unmap_cifs_share(mappath => $drive_letter_saved, force => 1);
	};
	logWarn("clean_share: $@") if $@; 

   
    return $result;
}
#--------------------------------------------------------------------
=head2 create_dataset

clean_share(target_drive    => $tdrive,
			client => $CLIENT,
			dir_list => $dir_list,
			file_list => $file_list
			);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub create_dataset(%) {
    my %opts = validate(
        @_,
        {
			dir_list => { type => SCALAR },
			file_list => { type => SCALAR },
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::CIFS_PASSWD},
        }
    );
	
	#my $cifs_drive;

	my $share = $opts{cifs_share};
	my ($filer, $hosttype) = split(";", $main::FILER);
	
	logStep("create_dataset: mapping CIFS drive");
	
	my $fh = host($filer);
	my $filer_ip = $fh->default_ip();
		
	
	my $cifs_drive = map_cifs_share( 
		server   => $filer_ip,
		share    => $share,
		client   => 'localhost',
		drive    => '*',
		username => $opts{cifs_user},
		password => $opts{cifs_passwd}
		)
		or croak($@);
	logDbg("mapped drive $cifs_drive");
	logStep("create_dataset: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
	if (!$cifs_drive) {
		croak("could not map the drive");
	}	
	
	logStep("create_dataset: create directories");
	
	foreach my $dir (split(" ", $opts{dir_list})) {
		test_dir_and_file_create(target_drive => $cifs_drive, test_dir_name => $dir);
		foreach my $file (split(" ", $opts{file_list})) {
			test_dir_and_file_create(target_drive => $cifs_drive, test_dir_name => $dir, test_file_name => $file);
		}
	}
	
	logStep("create_dataset: unmapping CIFS drive $cifs_drive");
	eval {
		unmap_cifs_share(mappath => $cifs_drive, force => 1);
	};
	logWarn("create_dataset: $@") if $@;
	
   
    return;
}
#--------------------------------------------------------------------

=head2 change_perms

read_dir_cmd(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub change_perms(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			path => { type => SCALAR },
			user => { type => SCALAR },
			perm => { type => SCALAR },
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::CIFS_PASSWD},
			drive_letter => { type => SCALAR, default => ""}
        }
    );
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype);

	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{cifs_share};
	($filer, $hosttype) = split(";", $main::FILER);

	my $fh = host($filer);
	my $filer_ip = $fh->default_ip();

	if($opts{drive_letter} eq "") {
		logStep("change_perms: mapping CIFS drive");
		$cifs_drive = map_cifs_share(
			server   => $filer_ip,,
			share    => $share,
			client   => 'localhost',
			drive    => '*',
			username => $opts{cifs_user},
			password => $opts{cifs_passwd}
			)
			or croak($@);
		logDbg("change_perms: mapped drive $cifs_drive");
		logStep("change_perms: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
		if (!$cifs_drive) {
			croak("could not map the drive");
		}
	} else {
		$cifs_drive = $opts{drive_letter};
		logStep("change_perms: CIFS drive is already mapped to $cifs_drive");
	}
	
	logStep("change_perms: Changing permissions on the file $cifs_drive\\$opts{path} to $opts{perm} for the user $opts{user}");
		
	my ($out1, $out2, $out3) = say($rsh_conn, "dir $cifs_drive");
	($out1, $out2, $out3) = say($rsh_conn, "cacls $cifs_drive\\$opts{path} /E /P $opts{user}\:$opts{perm}");
	if ($out3 =~ /processed dir:/ || $out3 =~ /processed file:/) {
		logStep("change_perms: Change successful");
		$result = 1;
		
	} else {
		logStep("change_perms: Change failed");
		logresult('FATAL', "change_perms: Change failed");
		$result = 0;
	}		
	
	sleep 5;
	
	if($opts{drive_letter} eq "") {
		logStep("change_perms: unmapping CIFS drive $cifs_drive");
		say($rsh_conn, "net use $cifs_drive /delete /y");
		logWarn("change_perms: $@") if $@;  
	}
    
   
    return $result;
}
#---------------------------------------------------------------------

=head2 setup_homedirs

$results = setup_homedirs(filer => $filer,
								share => $share_name, 
								abe_on_off => $abe_on_off);

%results = (0|1 - success or failure of operation)

Turn on ABE on a give share.

The following named arguments are accepted:

=over

=item Arguments

=over

=item filer

(Required) : filer where the share lives

=item share

(Required) : name of the share on which to turn on ABE

=item abe_on_off

(Required) : boolean with 1 to turn on and 0 to turn off ABE

=back

=item Returns

Results 1 or 0 depending on operation success

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub setup_homedirs(%) {
    my %opts = validate(
        @_,
        {
        }
    );

    my $result = 0;
	
    my ($filer, $hosttype) = split(";", $main::FILER);
    my $rsh_conn = make_conn($filer);
	my $fh = host($filer);
	my $filer_ip = $fh->default_ip();
	my $homedir_vol = $main::FILER_VOL;
	my $home_dir = $main::HOMEDIR;
	my $user_share = $main::HOMEDIR_SHARE;
	
	my $drive = map_cifs_share(
			server   => $filer_ip,
			share    => $main::CIFS_SHARE,
			client   => 'localhost',
			drive    => '*',
			username => $main::CIFS_USER,
			password => $main::CIFS_PASSWD
			)
		or croak($@);
	test_dir_and_file_create(target_drive => $drive, test_dir_name => $home_dir);
	eval {
			unmap_cifs_share(mappath => $drive, force => 1);
	};
    say($rsh_conn, "cifs shares -add $home_dir /vol/$homedir_vol/$home_dir");
	
	logStep("setup_homedirs: setup /etc/cifs_homedir.cfg file");
	my ($out1, $out2, $out3) = say($rsh_conn, "priv set diag; rm /etc/cifs_homedir.cfg");
	($out1, $out2, $out3) = say($rsh_conn, "cifs homedir load");
	($out1, $out2, $out3) = say($rsh_conn, "wrfile -a /etc/cifs_homedir.cfg /vol/$homedir_vol/$home_dir");
	($out1, $out2, $out3) = say($rsh_conn, "rdfile /etc/cifs_homedir.cfg");
	($out1, $out2, $out3) = say($rsh_conn, "cifs homedir load");
	($out1, $out2, $out3) = say($rsh_conn, "cifs homedir");
	($out1, $out2, $out3) = say($rsh_conn, "options cifs.home_dir_namestyle ntname");
	
	logStep("setup_homedirs: change qtree security to mixed");
	($out1, $out2, $out3) = say($rsh_conn, "cifs terminate -t 0");
	($out1, $out2, $out3) = say($rsh_conn, "qtree security /vol/$homedir_vol ntfs");
	($out1, $out2, $out3) = say($rsh_conn, "cifs restart");
	
	sleep 10;
	
	
	logStep("setup_homedirs: mapping CIFS drive");
	my $cifs_drive = map_cifs_share(
		server   => $filer_ip,
		share    => $home_dir,
		client   => 'localhost',
		drive    => '*',
		username => $main::CIFS_USER,
		password => $main::CIFS_PASSWD
		)
		or croak($@);
	logDbg("change_perms: mapped drive $cifs_drive");
	logStep("change_perms: successfully mapped filer $filer\\home to CIFS drive $cifs_drive");
	if (!$cifs_drive) {
		croak("could not map the drive");
	}
	
	logStep("setup_homedirs: create directory for the test user");
	test_dir_and_file_create(target_drive => $cifs_drive, test_dir_name => $user_share);
	
	logStep("setup_homedirs: unmapping CIFS drive $cifs_drive");
	eval {
		unmap_cifs_share(mappath => $cifs_drive, force => 1);
	};
	logWarn("setup_homedirs: $@") if $@;   
	
	logStep("setup_homedirs: verifying the creation of the homedirectory");
	my $homedir = map_cifs_share(
		server   => $filer,
		share    => $user_share,
		client   => 'localhost',
		drive    => '*',
		username => $main::TEST_CIFS_USER,
		password => $main::TEST_CIFS_PASSWD
		)
		or croak($@);
	logDbg("change_perms: mapped drive $homedir");
	logStep("setup_homedirs: successfully mapped homedir $filer\\$main::TEST_CIFS_USER to CIFS drive $homedir");
	if (!$homedir) {
		croak("could not map the homedir");
	}
	logStep("setup_homedirs: unmapping CIFS homedir $homedir");
	eval {
		unmap_cifs_share(mappath => $homedir, force => 1);
	};
	logWarn("setup_homedirs: $@") if $@; 
	
	return $result;
}

#---------------------------------------------------------------------

=head2 change_perms

read_dir_cmd(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub change_dir(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			path => { type => SCALAR },
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::CIFS_PASSWD},
			drive_letter => { type => SCALAR, default => ""}
        }
    );
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype);

	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{cifs_share};
	($filer, $hosttype) = split(";", $main::FILER);

	if($opts{drive_letter} eq "") {
		logStep("change_dir: mapping CIFS drive");
		$cifs_drive = map_cifs_share(
			server   => $filer,
			share    => $share,
			client   => 'localhost',
			drive    => '*',
			username => $opts{cifs_user},
			password => $opts{cifs_passwd}
			)
			or croak($@);
		logDbg("change_dir: mapped drive $cifs_drive");
		logStep("change_dir: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
		if (!$cifs_drive) {
			croak("could not map the drive");
		}
	} else {
		$cifs_drive = $opts{drive_letter};
		logStep("change_dir: CIFS drive is already mapped to $cifs_drive");
	}
	
	logStep("change_dir: Changing directory to $cifs_drive\\$opts{path}");
		
	my ($out1, $out2, $out3) = say($rsh_conn, "cd $cifs_drive\\$opts{path}");
	if ($out3 =~ /The system cannot find the path specified/) {
		logresult('FAIL', "We were not able to cd into a directory");
	} else {
		logresult('PASS', "We were able to successfully cd into a directory");
	}
	
	if($opts{drive_letter} eq "") {
		logStep("change_dir: unmapping CIFS drive $cifs_drive");
		eval {
			unmap_cifs_share(mappath => $cifs_drive, force => 1);
		};
		logWarn("change_dir: $@") if $@;  
	}
    
   
    return $result;
}
#---------------------------------------------------------------------

=head2 open_file

read_dir_cmd(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub open_file(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			path => { type => SCALAR },
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::TEST_CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::TEST_CIFS_PASSWD},
			drive_letter => { type => SCALAR, default => ""}
        }
    );
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype);

	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{cifs_share};
	($filer, $hosttype) = split(";", $main::FILER);

	if($opts{drive_letter} eq "") {
		logStep("open_file: mapping CIFS drive");
		$cifs_drive = map_cifs_share(
			server   => $filer,
			share    => $share,
			client   => 'localhost',
			drive    => '*',
			username => $opts{cifs_user},
			password => $opts{cifs_passwd}
			)
			or croak($@);
		logDbg("open_file: mapped drive $cifs_drive");
		logStep("open_file: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
		if (!$cifs_drive) {
			croak("could not map the drive");
		}
	} else {
		$cifs_drive = $opts{drive_letter};
		logStep("open_file: CIFS drive is already mapped to $cifs_drive");
	}
	
	logStep("open_file: Trying to open file $cifs_drive\\$opts{path}, we are not supposed to be able to do so");
		
	my ($out1, $out2, $out3) = say($rsh_conn, "$cifs_drive\\$opts{path}");
	if ($out3 =~ /is not recognized as an internal or external command/) {
		logresult('PASS', "We were unable to open the file $cifs_drive\\$opts{path}");
	} else {
		logresult('FAIL', "We were not supposed to ba able to open the file $cifs_drive\\$opts{path}, but were able to");
	}
	
	if($opts{drive_letter} eq "") {
		logStep("open_file: unmapping CIFS drive $cifs_drive");
		eval {
			unmap_cifs_share(mappath => $cifs_drive, force => 1);
		};
		logWarn("open_file: $@") if $@;  
	}
    
   
    return $result;
}
#---------------------------------------------------------------------

=head2 rename_file_dir

read_dir_cmd(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub rename_file_dir(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			path => { type => SCALAR },
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::TEST_CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::TEST_CIFS_PASSWD},
			drive_letter => { type => SCALAR, default => ""}
        }
    );
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype);

	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{cifs_share};
	($filer, $hosttype) = split(";", $main::FILER);

	if($opts{drive_letter} eq "") {
		logStep("rename_file_dir: mapping CIFS drive");
		$cifs_drive = map_cifs_share(
			server   => $filer,
			share    => $share,
			client   => 'localhost',
			drive    => '*',
			username => $opts{cifs_user},
			password => $opts{cifs_passwd}
			)
			or croak($@);
		logDbg("rename_file_dir: mapped drive $cifs_drive");
		logStep("rename_file_dir: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
		if (!$cifs_drive) {
			croak("could not map the drive");
		}
	} else {
		$cifs_drive = $opts{drive_letter};
		logStep("rename_file_dir: CIFS drive is already mapped to $cifs_drive");
	}
	
	logStep("rename_file_dir: Trying to rename $cifs_drive\\$opts{path}");
		
	my ($out1, $out2, $out3) = say($rsh_conn, "ren $cifs_drive\\$opts{path} renamed1");
	#print "out3 = $out3\n";
	if ($out3 =~ /Access is denied./ || $out3 =~ /The system cannot find the file specified./) {
		logresult('PASS', "We were unable to rename the file/directory $cifs_drive\\$opts{path}");
	} else {
		logresult('FAIL', "We were not supposed to ba able to rename the file/directory  $cifs_drive\\$opts{path}, but were able to");
	}
	
	if($opts{drive_letter} eq "") {
		logStep("rename_file_dir: unmapping CIFS drive $cifs_drive");
		eval {
			unmap_cifs_share(mappath => $cifs_drive, force => 1);
		};
		logWarn("rename_file_dir: $@") if $@;  
	}
    
   
    return $result;
}
#---------------------------------------------------------------------

=head2 delete_file_dir

read_dir_cmd(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub delete_file_dir(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			path => { type => SCALAR },
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::TEST_CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::TEST_CIFS_PASSWD},
			drive_letter => { type => SCALAR, default => ""},
			dir => { type => SCALAR, default => 0}
        }
    );
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype);

	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{cifs_share};
	($filer, $hosttype) = split(";", $main::FILER);

	if($opts{drive_letter} eq "") {
		logStep("delete_file_dir: mapping CIFS drive");
		$cifs_drive = map_cifs_share(
			server   => $filer,
			share    => $share,
			client   => 'localhost',
			drive    => '*',
			username => $opts{cifs_user},
			password => $opts{cifs_passwd}
			)
			or croak($@);
		logDbg("delete_file_dir: mapped drive $cifs_drive");
		logStep("delete_file_dir: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
		if (!$cifs_drive) {
			croak("could not map the drive");
		}
	} else {
		$cifs_drive = $opts{drive_letter};
		logStep("delete_file_dir: CIFS drive is already mapped to $cifs_drive");
	}
	
	logStep("delete_file_dir: Trying to delete $cifs_drive\\$opts{path}");
	
	if ($opts{dir} == 1) {
		my ($out1, $out2, $out3) = say($rsh_conn, "rmdir /S /Q $cifs_drive\\$opts{path}");
		#print "out3 = $out3\n";
		if ($out3 =~ /Access is denied./ || $out3 =~ /The system cannot find the file specified./ || $out3 =~ /Could Not Find/) {
			logresult('PASS', "We were unable to delete the directory $cifs_drive\\$opts{path}");
		} else {
			logresult('FAIL', "We were not supposed to ba able to delete the directory $cifs_drive\\$opts{path}, but were able to");
		}
	} else {
		my ($out1, $out2, $out3) = say($rsh_conn, "del $cifs_drive\\$opts{path} /F /Q");
		#print "out3 = $out3\n";
		if ($out3 =~ /Access is denied./ || $out3 =~ /The system cannot find the file specified./ || $out3 =~ /Could Not Find/) {
			logresult('PASS', "We were unable to delete the file $cifs_drive\\$opts{path}");
		} else {
			logresult('FAIL', "We were not supposed to be able to delete the file $cifs_drive\\$opts{path}, but were able to");
		}
	}
	
	if($opts{drive_letter} eq "") {
		logStep("delete_file_dir: unmapping CIFS drive $cifs_drive");
		eval {
			unmap_cifs_share(mappath => $cifs_drive, force => 1);
		};
		logWarn("delete_file_dir: $@") if $@;  
	}
    
   
    return $result;
}
#---------------------------------------------------------------------

=head2 create_file_dir

read_dir_cmd(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub create_file_dir(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			path => { type => SCALAR },
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::TEST_CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::TEST_CIFS_PASSWD},
			drive_letter => { type => SCALAR, default => ""},
			dir => { type => SCALAR, default => 0}
        }
    );
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype);

	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{cifs_share};
	($filer, $hosttype) = split(";", $main::FILER);

	if($opts{drive_letter} eq "") {
		logStep("create_file_dir: mapping CIFS drive");
		$cifs_drive = map_cifs_share(
			server   => $filer,
			share    => $share,
			client   => 'localhost',
			drive    => '*',
			username => $opts{cifs_user},
			password => $opts{cifs_passwd}
			)
			or croak($@);
		logDbg("create_file_dir: mapped drive $cifs_drive");
		logStep("create_file_dir: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
		if (!$cifs_drive) {
			croak("could not map the drive");
		}
	} else {
		$cifs_drive = $opts{drive_letter};
		logStep("create_file_dir: CIFS drive is already mapped to $cifs_drive");
	}
	
	logStep("create_file_dir: Trying to create file/dir $cifs_drive\\$opts{path}");
	
	if ($opts{dir} == 1) {
		my ($out1, $out2, $out3) = say($rsh_conn, "mkdir $cifs_drive\\$opts{path}");
		#print "out3 = $out3\n";
		if ($out3 =~ /Access is denied./ || $out3 =~ /already exists./) {
			logresult('PASS', "We were not able to create directory with the same name as existing file $cifs_drive\\$opts{path}");
		} else {
			logresult('FAIL', "We were able to create directory with the same name as existing file $cifs_drive\\$opts{path}, but were supposed to");
		}
	} else {
		my ($out1, $out2, $out3) = say($rsh_conn, "echo this is a test > $cifs_drive\\$opts{path}");
		#print "out3 = $out3\n";
		if ($out3 =~ /Access is denied./ || $out3 =~ /already exists./) {
			logresult('PASS', "We were not able to create file with the same name as existing file $cifs_drive\\$opts{path}");
		} else {
			logresult('FAIL', "We were able to create file with the same name as existing file $cifs_drive\\$opts{path}, but were supposed to");
		}
	}
	
	if($opts{drive_letter} eq "") {
		logStep("create_file_dir: unmapping CIFS drive $cifs_drive");
		eval {
			unmap_cifs_share(mappath => $cifs_drive, force => 1);
		};
		logWarn("create_file_dir: $@") if $@;  
	}
    
   
    return $result;
}
#---------------------------------------------------------------------
=head2 type_file

read_dir_cmd(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub type_file(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			path => { type => SCALAR },
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::TEST_CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::TEST_CIFS_PASSWD},
			drive_letter => { type => SCALAR, default => ""},
        }
    );
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype);

	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{cifs_share};
	($filer, $hosttype) = split(";", $main::FILER);

	if($opts{drive_letter} eq "") {
		logStep("type_file: mapping CIFS drive");
		$cifs_drive = map_cifs_share(
			server   => $filer,
			share    => $share,
			client   => 'localhost',
			drive    => '*',
			username => $opts{cifs_user},
			password => $opts{cifs_passwd}
			)
			or croak($@);
		logDbg("type_file: mapped drive $cifs_drive");
		logStep("type_file: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
		if (!$cifs_drive) {
			croak("could not map the drive");
		}
	} else {
		$cifs_drive = $opts{drive_letter};
		logStep("type_file: CIFS drive is already mapped to $cifs_drive");
	}
	
	logStep("type_file: Trying to print filer $cifs_drive\\$opts{path}");
	
	my ($out1, $out2, $out3) = say($rsh_conn, "type $cifs_drive\\$opts{path}");
	#print "out3 = $out3\n";
	if ($out3 =~ /Access is denied./ || $out3 =~ /already exists./) {
		logresult('PASS', "We were not able to create file with the same name as existing file $cifs_drive\\$opts{path}");
	} else {
		logresult('FAIL', "We were able to create file with the same name as existing file $cifs_drive\\$opts{path}, but were supposed to");
	}
	
	if($opts{drive_letter} eq "") {
		logStep("type_file: unmapping CIFS drive $cifs_drive");
		eval {
			unmap_cifs_share(mappath => $cifs_drive, force => 1);
		};
		logWarn("type_file: $@") if $@;  
	}
    
   
    return $result;
}
#---------------------------------------------------------------------
=head2 read_dir_gui

read_dir_gui(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub read_dir_gui(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::TEST_CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::TEST_CIFS_PASSWD},
			drive_letter => { type => SCALAR, default => ""},
			path => { type => SCALAR },
        }
    );
	
	require Win32::GuiTest;
	import Win32::GuiTest qw(FindWindowLike GetWindowText SendKeys WaitWindowLike
	                      GetListViewContents GetChildWindows SetFocus
	                      GetForegroundWindow  WaitWindow PushChildButton
                      SendMouse GetTabItems SelTabItemText GetClassName GetComboText 
					  GetTreeViewSelPath GetComboContents GetListContents GetListText WMGetText
					  GetWindowID);
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype, $found);
	my (@windows, $ttext, $window, $window2, $proptext, $prevproptext, $out1, $out2, $out3, 
		$test_dir, $test_file, $title, $current_test_dir);
	
	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{cifs_share};
	($filer, $hosttype) = split(";", $main::FILER);

	if($opts{drive_letter} eq "") {
		logStep("read_dir_gui: mapping CIFS drive");
		$cifs_drive = map_cifs_share(
			server   => $filer,
			share    => $share,
			client   => 'localhost',
			drive    => '*',
			username => $opts{cifs_user},
			password => $opts{cifs_passwd}
			)
			or croak($@);
		logDbg("read_dir_gui: mapped drive $cifs_drive");
		logStep("read_dir_gui: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
		if (!$cifs_drive) {
			croak("could not map the drive");
		}
	} else {
		$cifs_drive = $opts{drive_letter};
		logStep("read_dir_gui: CIFS drive is already mapped to $cifs_drive");
	}
	
	open_explorer(drive_letter => $cifs_drive, window_title => $share, rsh_conn => $rsh_conn, path => "");
	

	my $window_found = WaitWindowLike(0, $share, "", "", 10, 1);
	@windows = FindWindowLike(0, $share, "");
	
	foreach my $file (split(" ", $opts{path})){
		my ($ttext, $found) = split(",", $file);
		print "0 - ttext = $ttext\n";
		$window = $windows[0];
		$result = 0;
		$proptext = "";		
		$current_test_dir = "";
		my $counter = 0;
				
		if ($found == 1) {
			logStep("read_dir_gui: Checking the existence of the file $cifs_drive\\$ttext, we are expecting to find it");
		} else {
			logStep("read_dir_gui: Checking the existence of the file $cifs_drive\\$ttext, we are not expecting to find it");
		}	
		
		if ($ttext =~ /\\/) {
			($test_dir, $test_file) = split(/\\/, $ttext);
			
			if ($test_dir ne $current_test_dir) {
				SendKeys("%{F4}");
				sleep(2);
				open_explorer(drive_letter => $cifs_drive, window_title => $share, rsh_conn => $rsh_conn, path => $test_dir);
			}
			$title = $test_file;
			$current_test_dir = $test_dir;
		} else {
			$title = $ttext;
		}
		#print "test_dir = $test_dir, test_file = $test_file; title = $title\n";

		SendKeys("{HOME}");
		sleep(2);

		# Names of explorer items will be queried by properties window text
		# (bug in Gui_Test does not allow windows 2003 to read lists)
		SendKeys("%f");
		sleep(2);
		SendKeys("r");

		sleep(10);
		WaitWindow("Properties", 10);

		$window2 = GetForegroundWindow();

		sleep(2);

		$proptext = GetWindowText($window2);
		$prevproptext = "";
		print "0 - proptext = $proptext; title = $title; prevproptext = $prevproptext; window2 = $window2\n";

		while ($proptext !~ /$title/ && $prevproptext ne $proptext)  {
			print "1 - proptext = $proptext; title = $title; prevproptext = $prevproptext; window2 = $window2\n";
			SetFocus($window2);

			PushChildButton($window2, "Cancel");

			sleep(2);

			SendKeys("{DOWN}");

			sleep(1);

			SendKeys("%f");

			sleep(2);

			SendKeys("r");
			sleep(10);

			WaitWindow("Properties", 10);

			$window2 = GetForegroundWindow();

			sleep(2);

			$prevproptext = $proptext;

			$proptext = GetWindowText($window2);
			print "2 - proptext = $proptext; title = $title; prevproptext = $prevproptext; window2 = $window2\n";
			
			print "11 - counter = $counter\n";
			$counter++;
		}

		SetFocus($window2);
		PushChildButton($window2, "Cancel");
	
		if ($proptext =~ /$title/) {
			logStep("read_dir_gui: File $cifs_drive\\$ttext was found; found = $found");
			if ($found == 1) {
				logresult('PASS', "1 - We were expecting to find the directory/file $cifs_drive\\$ttext and we found it");					
				$result = 1;
			} else {
				logresult('FAIL', "2 - We were not expecting to find the directory/file $cifs_drive\\$ttext but found it");
				$result = 0;
			}
		} 
		if ($result == 0) {
			logStep("read_dir_gui: File $cifs_drive\\$ttext was not found; found = $found");
			if ($found == 1) {
				logresult('FAIL', "3 - We were expecting to find the directory/file $cifs_drive\\$ttext but not found it");
				$result = 0;
			} else {
				logresult('PASS', "4 - We were not expecting to find the directory/file $cifs_drive\\$ttext and not found it");
				$result = 1;
			}
		}	
	}	
		
    
	if($opts{drive_letter} eq "") {
		logStep("read_dir_gui: unmapping CIFS drive $cifs_drive");
		($out1, $out2, $out3) = say($rsh_conn, "net use $cifs_drive /delete /y");
	}
   
    return $result;
}
#---------------------------------------------------------------------
=head2 open_explorer

read_dir_cmd(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub open_explorer(%) {
    my %opts = validate(
        @_,
        {
			path => { type => SCALAR },
			window_title => { type => SCALAR },
			rsh_conn => { type => HASHREF },
			drive_letter => { type => SCALAR, default => ""}
        }
    );
	
	my ($window2, $result);

	logStep("open_explorer: Opening explorer to location $opts{drive_letter}\\$opts{path}");

	SendKeys("^{ESCAPE}");
	sleep(2);
	SendKeys("r");
	sleep(2);
	SendKeys("explorer.exe $opts{drive_letter}\\$opts{path}");
	sleep(2);
	SendKeys("{ENTER}");
	sleep(5);


	WaitWindow($opts{window_title}, 10);
	$window2 = GetForegroundWindow();
	print "13 - window2 = $window2\n";
	if ($window2 == 0) {
		logStep("open_explorer: unmapping CIFS drive $opts{drive_letter}");
		say($opts{rsh_conn}, "net use $opts{drive_letter} /delete /y");
		logresult('FATAL', "Explorer window didn't open");					
	}

	SendKeys("%v");
	sleep(2);
	SendKeys("l");
	sleep(2);
    
   
    return $window2;
}
#---------------------------------------------------------------------
=head2 change_dir_gui

change_dir_gui(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub change_dir_gui(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::TEST_CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::TEST_CIFS_PASSWD},
			drive_letter => { type => SCALAR, default => ""},
			path => { type => SCALAR },
        }
    );
	
	require Win32::GuiTest;
	import Win32::GuiTest qw(FindWindowLike GetWindowText SendKeys WaitWindowLike
	                      GetListViewContents GetChildWindows SetFocus
	                      GetForegroundWindow  WaitWindow PushChildButton
                      SendMouse GetTabItems SelTabItemText GetClassName GetComboText 
					  GetTreeViewSelPath GetComboContents GetListContents GetListText WMGetText
					  GetWindowID);
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype, $found);
	my (@windows, $ttext, $window, $window2, $proptext, $prevproptext, $out1, $out2, $out3, 
		$test_dir, $test_file, $title, $current_test_dir);
	
	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{cifs_share};
	($filer, $hosttype) = split(";", $main::FILER);

	if($opts{drive_letter} eq "") {
		logStep("read_dir_gui: mapping CIFS drive");
		$cifs_drive = map_cifs_share(
			server   => $filer,
			share    => $share,
			client   => 'localhost',
			drive    => '*',
			username => $opts{cifs_user},
			password => $opts{cifs_passwd}
			)
			or croak($@);
		logDbg("change_dir_gui: mapped drive $cifs_drive");
		logStep("change_dir_gui: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
		if (!$cifs_drive) {
			croak("could not map the drive");
		}
	} else {
		$cifs_drive = $opts{drive_letter};
		logStep("change_dir_gui: CIFS drive is already mapped to $cifs_drive");
	}
	
	my $window_found = open_explorer(drive_letter => $cifs_drive, window_title => $share, rsh_conn => $rsh_conn, path => $opts{path});
	
	my $text = GetWindowText($window_found);
	my $class = GetClassName($window_found);
	print "text = $text; class = $class\n";
	sleep(10);
	SendKeys("{ENTER}");
	
	if($text =~ /Windows Explorer/ && $class =~ /32770/) {
		logresult('PASS', "We were unable to open a directory $cifs_drive\\$opts{path}");
		$result = 1;
	} else {
		logresult('FAIL', "We were able to open a directory $cifs_drive\\$opts{path}, but we were not supposed to");
		$result = 0;
	}
	
		
    
	if($opts{drive_letter} eq "") {
		logStep("change_dir_gui: unmapping CIFS drive $cifs_drive");
		($out1, $out2, $out3) = say($rsh_conn, "net use $cifs_drive /delete /y");
	}
   
    return $result;
}
#---------------------------------------------------------------------
=head2 open_file_gui

change_dir_gui(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub open_file_gui(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::TEST_CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::TEST_CIFS_PASSWD},
			drive_letter => { type => SCALAR, default => ""},
			path => { type => SCALAR },
        }
    );
	
	require Win32::GuiTest;
	import Win32::GuiTest qw(FindWindowLike GetWindowText SendKeys WaitWindowLike
	                      GetListViewContents GetChildWindows SetFocus
	                      GetForegroundWindow  WaitWindow PushChildButton
                      SendMouse GetTabItems SelTabItemText GetClassName GetComboText 
					  GetTreeViewSelPath GetComboContents GetListContents GetListText WMGetText
					  GetWindowID);
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype, $found);
	my (@windows, $ttext, $window, $window2, $proptext, $prevproptext, $out1, $out2, $out3, 
		$test_dir, $test_file, $title, $current_test_dir);
	
	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{cifs_share};
	($filer, $hosttype) = split(";", $main::FILER);

	if($opts{drive_letter} eq "") {
		logStep("open_file_gui: mapping CIFS drive");
		$cifs_drive = map_cifs_share(
			server   => $filer,
			share    => $share,
			client   => 'localhost',
			drive    => '*',
			username => $opts{cifs_user},
			password => $opts{cifs_passwd}
			)
			or croak($@);
		logDbg("open_file_gui: mapped drive $cifs_drive");
		logStep("open_file_gui: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
		if (!$cifs_drive) {
			croak("could not map the drive");
		}
	} else {
		$cifs_drive = $opts{drive_letter};
		logStep("open_file_gui: CIFS drive is already mapped to $cifs_drive");
	}
	
	logStep("open_file_gui: Opening file $opts{drive_letter}\\$opts{path}");

	SendKeys("^{ESCAPE}");
	sleep(2);
	SendKeys("r");
	sleep(2);
	SendKeys("explorer.exe $opts{drive_letter}\\$opts{path}");
	sleep(2);
	SendKeys("{ENTER}");
	sleep(5);


	#WaitWindow($opts{window_title}, 120);
	$window2 = GetForegroundWindow();
	print "13 - window2 = $window2\n";
	if ($window2 == 0) {
		logStep("open_file_gui: unmapping CIFS drive $opts{drive_letter}");
		say($opts{rsh_conn}, "net use $opts{drive_letter} /delete /y");
		logresult('FATAL', "Explorer window didn't open");					
	}

	my $text = GetWindowText($window2);
	my $class = GetClassName($window2);
	print "text = $text; class = $class\n";
	#WaitWindowLike(0, $share, "", "", 10, 1);
	#@windows = FindWindowLike(0, $share, "");
	sleep(10);
	SendKeys("{ENTER}");
	
	if($text =~ /Windows Explorer/ && $class =~ /32770/) {
		logresult('PASS', "We were unable to open a file $cifs_drive\\$opts{path}");
		$result = 1;
	} else {
		logresult('FAIL', "We were able to open a file $cifs_drive\\$opts{path}, but we were not supposed to");
		$result = 0;
	}
	
		
    
	if($opts{drive_letter} eq "") {
		logStep("open_file_gui: unmapping CIFS drive $cifs_drive");
		($out1, $out2, $out3) = say($rsh_conn, "net use $cifs_drive /delete /y");
	}
   
    return $result;
}
#---------------------------------------------------------------------
=head2 make_file_gui

make_file_gui(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub make_file_gui(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::TEST_CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::TEST_CIFS_PASSWD},
			drive_letter => { type => SCALAR, default => ""},
			dir => { type => SCALAR },
			file => { type => SCALAR, default => "" },
        }
    );
	
	require Win32::GuiTest;
	import Win32::GuiTest qw(FindWindowLike GetWindowText SendKeys WaitWindowLike
	                      GetListViewContents GetChildWindows SetFocus
	                      GetForegroundWindow  WaitWindow PushChildButton
                      SendMouse GetTabItems SelTabItemText GetClassName GetComboText 
					  GetTreeViewSelPath GetComboContents GetListContents GetListText WMGetText
					  GetWindowID);
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype, $found);
	my (@windows, $ttext, $window, $window2, $proptext, $prevproptext, $out1, $out2, $out3, 
		$test_dir, $test_file, $title, $current_test_dir);
	
	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{cifs_share};
	($filer, $hosttype) = split(";", $main::FILER);

	if($opts{drive_letter} eq "") {
		logStep("make_file_gui: mapping CIFS drive");
		$cifs_drive = map_cifs_share(
			server   => $filer,
			share    => $share,
			client   => 'localhost',
			drive    => '*',
			username => $opts{cifs_user},
			password => $opts{cifs_passwd}
			)
			or croak($@);
		logDbg("make_file_gui: mapped drive $cifs_drive");
		logStep("make_file_gui: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
		if (!$cifs_drive) {
			croak("could not map the drive");
		}
	} else {
		$cifs_drive = $opts{drive_letter};
		logStep("make_file_gui: CIFS drive is already mapped to $cifs_drive");
	}
	
	if ($opts{file} eq "") {
		my $window_found = open_explorer(drive_letter => $cifs_drive, window_title => $share, rsh_conn => $rsh_conn, path => '');
	} else {
		my $window_found = open_explorer(drive_letter => $cifs_drive, window_title => $share, rsh_conn => $rsh_conn, path => $opts{dir});
	}	
	
	sleep(2);
	SendKeys("%f");
	sleep(2);
	SendKeys("{RIGHT}");
	sleep(2);
	if ($opts{file} eq "") {
		SendKeys("f");
		sleep(2);
		SendKeys("$opts{dir}");
	} else {
		SendKeys("t");
		sleep(2);
		SendKeys("$opts{file}");
	}
	#sleep(2);
	#SendKeys("$opts{file}");
	sleep(2);
	SendKeys("{ENTER}");
	sleep(5);


	#WaitWindow($opts{window_title}, 10);
	$window2 = GetForegroundWindow();
	print "13 - window2 = $window2\n";
	if ($window2 == 0) {
		logStep("make_file_gui: unmapping CIFS drive $opts{drive_letter}");
		say($opts{rsh_conn}, "net use $opts{drive_letter} /delete /y");
		logresult('FATAL', "Explorer window didn't open");					
	}

	my $text = GetWindowText($window2);
	my $class = GetClassName($window2);
	print "1 - text = $text; class = $class\n";
	if($text =~ /Rename/) {
		SendKeys("{ENTER}");
		sleep(5);
		
		$window2 = GetForegroundWindow();
		$text = GetWindowText($window2);
		$class = GetClassName($window2);
		print "2 - text = $text; class = $class\n";
	}	
		
	#WaitWindowLike(0, $share, "", "", 10, 1);
	#@windows = FindWindowLike(0, $share, "");
	sleep(10);
	SendKeys("{ENTER}");
	
	print "text = $text\n";
	
	if($text =~ /Error Renaming File or Folder/) {
		logresult('PASS', "We were unable to create file $cifs_drive\\$opts{dir}\\$opts{file}");
		$result = 1;
	} else {
		logresult('FAIL', "We were able to create file $cifs_drive\\$opts{dir}\\$opts{file}, but we were not supposed to");
		$result = 0;
	}
	
		
    
	if($opts{drive_letter} eq "") {
		logStep("make_file_gui: unmapping CIFS drive $cifs_drive");
		($out1, $out2, $out3) = say($rsh_conn, "net use $cifs_drive /delete /y");
	}
   
    return $result;
}
#--------------------------------------------------------------------
=head2 create_widelink

create_widelink(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub create_widelink(%) {
    my %opts = validate(
        @_,
        {
			unix_client => { type => SCALAR },
			vol_source => { type => SCALAR },
			vol_dest => { type => SCALAR },
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::TEST_CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::TEST_CIFS_PASSWD},
			drive_letter => { type => SCALAR, default => ""},
        }
    );
	
	my ($result, $rsh_conn);

	$rsh_conn = make_conn($main::FILER);
	my $fh = host($main::FILER);
	my $filer_ip = $fh->default_ip();
	my $unix_conn = make_conn($opts{unix_client});
	
	#Create volume on a filer
	$result = vol_create($rsh_conn, $opts{vol_dest}, 2);
    print "result = $result\n";
	
	#mount volume on the unix client
	#say ($unix_conn, "mkdir -p /t/$main::FILER/vol_$opts{vol_source}-tcp-3");
	say ($unix_conn, "mkdir -p /t/$main::FILER/$opts{vol_source}");
	say ($unix_conn, "mount -o rw,bg,hard,intr,noquota,vers=3,proto=tcp $main::FILER:/vol/$opts{vol_source} /t/$main::FILER/$opts{vol_source}");
	#my $mountpoint = mount($main::FILER,"/vol/$opts{vol_source}",client=>$opts{unix_client}, proto=>"tcp", vers=>3, mountopts=>"rw,bg,hard,intr,noquota")
    #    or die $@;
	#print "mountpoint = $mountpoint\n";
	
	#create a symlink
	say ($unix_conn, "r ln -s /t/$main::FILER/$opts{vol_dest}/* widelink");
	
	#make sure that widelinks are enabled on a share
	say ($rsh_conn, "cifs shares -change $opts{cifs_share} -widelink");
	
	#create a symlink.translations file on a filer
	say ($rsh_conn, "rm /etc/symlink.translations");
	say ($rsh_conn, "wrfile -a /etc/symlink.translations widelink /t/$main::FILER/$opts{vol_source}/* \\$main::FILER\$opts{cifs_share}\$opts{vol_dest}\*");
	
	sleep 30;
   
    return $result;
}
#--------------------------------------------------------------------

=head2 change_qtree_security

$results = change_qtree_security(filer => $filer,
								share => $share_name, 
								abe_on_off => $abe_on_off);

%results = (0|1 - success or failure of operation)

Turn on ABE on a give share.

The following named arguments are accepted:

=over

=item Arguments

=over

=item filer

(Required) : filer where the share lives

=item share

(Required) : name of the share on which to turn on ABE

=item abe_on_off

(Required) : boolean with 1 to turn on and 0 to turn off ABE

=back

=item Returns

Results 1 or 0 depending on operation success

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub change_qtree_security(%) {
    my %opts = validate(
        @_,
        {
			qtree => { type => SCALAR},
			security_type => { type => SCALAR},
        }
    );

    my $result = 0;

    my $rsh_conn = make_conn($main::FILER);
	
	logStep("change_qtree_security: Changing qtree security on $opts{qtree}");
	my ($out1, $out2, $out3) = say($rsh_conn, "cifs terminate -t 0");
	($out1, $out2, $out3) = say($rsh_conn, "qtree security /vol/$opts{qtree} $opts{security_type}");
	sleep 10;
	($out1, $out2, $out3) = say($rsh_conn, "cifs restart");
	
	say($rsh_conn, "qtree");
	
	return $result;
}

#---------------------------------------------------------------------
