#!/usr/local/bin/perl
#
# Copyright (c) 2005 Network Appliance, Inc.
# All Rights reserved
#
## @summary
##  Library for ACL testing
##
## @description
##  Provides functions to test ACL
#
## @dependencies	none
#
## @status	        public
#
## @author	        vladimir@netapp.com,zina@netapp.com,olgav@netapp.com

=head1 NAME

Acl - Library for testing the ACL

=head1 SYNOPSIS

  use CIFS::Acl;

  change_shares($connobj,$CIFS_SHARE,$FILER_VOL,$create_delete,$security_style);
  
  Parameter $connobj        Required:	Provides the connection to node
  Parameter $CIFS_SHARE     Required:	Provides the name for share
  Parameter $FILER_VOL      Required:	Provides the volume name for share
  Parameter $create_delete  Required:	"0" for create; "1" for delete
  Parameter $security_style Optional:	"unix", "ntfs", or "mixed". If not specified then 
                                        security style will not changed
  
  Ex.:
  change_shares($fc,$CIFS_SHARE,$FILER_VOL,0);        # Created share on node
  change_shares($fc,$CIFS_SHARE,$FILER_VOL,1);        # Deleted share on node
  change_shares($fc,$CIFS_SHARE,$FILER_VOL,0,"ntfs"); # Created share on node 
                                                      # and changed security style 

=head1 DESCRIPTION

Create/delete a share on 7G/7MODE/GX node for the Acl tests.

=head1 STATUS

  Shared

=head1 AUTHOR

  olgav@netapp.com(Protocols QA)
  $Id$
  Last revised $Date$

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut


package CIFS::Acl;

require Exporter;

use strict;
use warnings;
use Tharn;
use CIFS::CommonParams;
use filer qw(filer_qtree_security_set);
use Server;

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our $VERSION = '1.0';

# Functions to be exported
our @ISA = qw(Exporter);
our @EXPORT =
  qw(change_shares);

#---------------------------------------------------------------------
# Global variables for this module
#---------------------------------------------------------------------
my $server;

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub change_shares(%);

#---------------------------------------------------------------------

sub change_shares(%) {

	my $fc = $_[0];
	my $share = $_[1];
	my $vol = $_[2];
	my $delete = $_[3];
	my $security_style = $_[4];
	my $reply;
 
	if ($fc->hostp()->hosttype() =~ m/filer-ng/){
	    # GX branch
          # create a new Server object first thing
  	    $server = new Server();

	    if($delete) {
		   eval {
		     $server->cli(cmd => "vserver cifs share show -vserver $main::VSERVER -share-name $share");
		   };
		   if (!$@){
	          eval {
		  	     $server->cli(cmd => "vserver cifs share delete -vserver $main::VSERVER -share-name $share");
		      };
		      if ($@){
			     logresult("FAIL", "Share $share was not deleted correctly");
		      }
		  }else {
		      logresult("INFO", "Share $share does not exist");
		  }
	    } else {
               if ($security_style){
                   $server->cli(cmd => "volume modify -vserver $main::VSERVER -volume $vol -security-style ntfs");
               }
		   eval {
		       $server->cli(cmd => "vserver cifs share create -vserver $main::VSERVER -share-name $share -path /$vol");
		   };
		   if ($@){
		       die("Share $share was not added");
		   }	   
		   
	   }

	} else {
	   # 7G/7MODE branch
	   if($delete) {
		say($fc, "cifs shares -delete $share");
		(undef, undef, my $reply) = say($fc, "cifs shares");
		foreach my $line (split ("\n",$reply)){
			if($line =~ /$share/){
				logresult("FAIL", "Share $share was not deleted correctly");
			}
		}
	   } else {
            if ($security_style){
                   filer_qtree_security_set($fc, "/vol/$vol/", "ntfs");
            }
		say($fc, "cifs shares -add $share /vol/$vol");
		my $share_found = 0;
		(undef, undef, my $reply) = say($fc, "cifs shares $share");
		foreach my $line (split ("\n",$reply)){
			if($line =~ /$share/){
				$share_found = 1;
			}
		}
		if(!$share_found) {
			die("Share $share was not added");
		}
	   }
	}
}


1;
