#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CIFS/Ad.pm#1 $
#
## @summary	Ad - Library for the active directory tests
## @author	dl-maryv (FSS Protocols QA)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Ad - Library for the Active Directory tests

=head1 SYNOPSIS

  use CIFS::Ad;

  ad_setup(ad_server => $AD_CHILD_SERVER, dns_server => $AD_CHILD_SERVER,
           domain => $cdomain, domain_user => $CIFS_DOMAIN_ADMIN,
           domain_pass => $CIFS_DOMAIN_ADMIN_PASSWD,
           setup_h => \%setup_h);

  ad_cleanup(%setup_h);

  $ret =  ad_val_filer(ad_server => $AD_CHILD_SERVER, filer => $FILER,
                       domain => $cdomain, username => $CIFS_DOMAIN_ADMIN,
                       password => $CIFS_DOMAIN_ADMIN_PASSWD);

  ad_val_domain(child_server => $AD_CHILD_SERVER,
                child_domain => $cdomain, dns_mode => 'ddns',
                domain => $AD_DOMAIN, mode => 'mixed',
                root_server => $AD_ROOT_SERVER,
                username => $CIFS_USER, password => $CIFS_PASSWD,
                windows_version => '5.2');

  @ret = adsi_query(query => $query);

  static_ip_set(ip => '192.168.2.42', mask => '255.255.255.0',
                gate => '192.168.2.1', dns => ('127.0.0.1'));

  $ret = ad_add_delete_domain_user(cifs_client => $client, user => $user,
                                   pass => $pass, flag => 'add');

  @groups = ad_list_groups(fc => $fc);

  testme();

=head1 DESCRIPTION

Provides the common routines for the Active Directory tests.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: //depot/prod/test/Rsierranevada/lib/CIFS/Ad.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package CIFS::Ad;

use strict;
use warnings;
use base qw(Exporter);

use Tharn qw(connect host logresult);
use Carp qw(croak);
use SPARSE::Util::Logger;
use Params::Validate qw(validate ARRAYREF HASHREF OBJECT SCALAR);
use POSIX qw(uname);
use SayWrap qw(sayEOF);
use CIFS::Ntfs qw(ntfs_cifs_setup);
use Network::Dns qw(dns_get_options dns_configure dns_restore_options);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our $VERSION = '1.0';

# Hash reference of DNS info to be restored for the filer
my $saved_dns_info;

# Functions to be exported
our @EXPORT_OK = qw(
    ad_setup
    ad_cleanup
    ad_val_filer
    ad_val_domain
    adsi_query
    static_ip_set
    ad_add_delete_domain_user
    ad_list_groups
);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub _LDAP_ADsPath($$$);
sub ad_setup(%);
sub ad_cleanup(%);
sub ad_val_filer(%);
sub ad_val_domain(%);
sub adsi_query(%);
sub static_ip_set(%);
sub ad_add_delete_domain_user(%);
sub ad_list_groups(%);
sub testme();

# Local function to build an LDAP ADsPath
sub _LDAP_ADsPath($$$) {
    my ($server, $domain, $stuff) = @_;
    my (@arr, $h, $query, $srv_name);

    $h = host($server);
    $srv_name = $h->hostname();
    $query = "LDAP://$srv_name/";
    if ($stuff) {
        $query .= "$stuff,";
    }

    @arr = split(/\./, $domain);
    foreach (@arr) {
        $query .= "dc=" . $_ . ",";
    }
    # Remove last comma
    $query =~ s/,$//;

    return $query;
}

#---------------------------------------------------------------------

=head2 ad_setup

 ad_setup(setup_h => \%setup_h
          [, windows_version => $win_version ]);

The following named arguments are accepted:

=over

=item Arguments

=over

=item ad_server

(Required) : hostspec where active directory domain server is setup.

=item domain

(Required) : Domain name setup on ad_server.

=item domain_user

(Required) : Domain administrator name.

=item domain_pass

(Required) : Domain administrator password.

=item dns_server

(Required) : hostspec of DNS server for the domain.

=item setup_h

(Required) : Reference to setup hold hash for cleanup.

=item windows_version

(Optional) : MS Windows version number. Defaults to 5.

=back

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub ad_setup(%) {
    my %opts = validate(
        @_,
        {
            ad_server       => { type => SCALAR },
            domain          => { type => SCALAR },
            domain_user     => { type => SCALAR },
            domain_pass     => { type => SCALAR },
            dns_server      => { type => SCALAR },
            setup_h         => { type => HASHREF },
            windows_version => { type => SCALAR, default => '5',
                                 regex => qr/^(5\.0|5\.1|5\.2|5)$/ }
        }
    );
    my ($fc, $fh, @osrel);

    logStep("BEGIN: ad_setup: setup the environment");

    logStep("ad_setup: Verify client has correct Windows version");
    @osrel = uname();
    if ($osrel[2] !~ /$opts{windows_version}/) {
        croak("client running $osrel[2] expected $opts{windows_version} " .
              "windows release");
    }

    # Connect to filer
    $fh = host($main::FILER);
    $fc = connect($main::FILER_CONN, -timeout, $main::TIMEOUT);
    $fc->timeout($main::CMDTIMEOUT);

    logStep("ad_setup: Save DNS state on the filer");
    $saved_dns_info = dns_get_options(conn => $fc);

    logStep("ad_setup: Setup DNS on the filer");
    $fh = host($opts{dns_server});
    dns_configure(conn => $fc,
                  enable => 'on',
                  domain => $opts{domain},
                  srv_ip_addr => $fh->default_ip());

    logStep("ad_setup: Setup CIFS on the filer");
    $opts{setup_h}->{ad_server} = $opts{ad_server};
    $opts{setup_h}->{domain} = $opts{domain};
    ntfs_cifs_setup(filer => $main::FILER,
                    ad_setup => 'yes',
                    dns_name => $opts{domain},
                    dns_server => $fh->default_ip(),
                    domain => $opts{domain},
                    domain_user => $opts{domain_user},
                    domain_pass => $opts{domain_pass},
                    mode => 'ntfs');

    logStep("END: ad_setup:");
    return;
}

#----------------------------------------------------------------------

=head2 ad_cleanup

ad_cleanup(setup_h => $setup_h);

The following named arguments are accepted:

=over

=item Arguments

=over

=item ad_server

(Optional) : hostspec where active directory domain server is setup.

=item domain

(Optional) : Domain name setup on ad_server.

=back

=back

=for html <hr />

=cut
#--------------------------------------------------------------------

sub ad_cleanup(%) {
    my %opts = validate(
        @_,
        {
            ad_server => { type => SCALAR, optional => 1 },
            domain    => { type => SCALAR, optional => 1 },
        }
    );

    my ($err, $fc, $fh, $filer_name, $objNameSpace, $objSec, $query);

    logStep("CLEANUP == $main::CLEANUP");

    require Win32::OLE;
    import Win32::OLE;

    eval {
        logStep("ad_cleanup: Remove filer from AD computers container");
        $fh = host($main::FILER);
        if (defined ($opts{domain}) && defined($opts{ad_server})) {
            # Build the ADsPath for quering the computers container
            $query = _LDAP_ADsPath($opts{ad_server}, $opts{domain},
                                  'cn=computers');

            # Bind to the Active Directory server and run the query
            $objNameSpace = Win32::OLE->GetObject("LDAP:");
            $err = Win32::OLE->LastError();
            if ($err == 0) {
                $objSec = $objNameSpace->OpenDSObject("$query",
                                                      $main::CIFS_USER,
                                                      $main::CIFS_PASSWD, 1);
                $err = Win32::OLE->LastError();
                if ($err == 0) {
                    # Get the filer name
                    $filer_name = $fh->hostname();
                    $filer_name =~ tr/a-z/A-Z/;

                    $objSec->Delete('computer','cn='.$filer_name);
                    $err = Win32::OLE->LastError();
                    # Ignore errors if it is not there to delete as cifs_setup
                    # could have died.
                    if ($err != 0 && $err !~ /no such object/) {
                        logWarn("ad_cleanup: Delete OLE error: $err");
                    }
                    $objSec->Close;
                } else {
                    logWarn("ad_cleanup: OpenDSObject OLE error: $err");
                }
                $objNameSpace->Close;
            }
        }
    };
    logWarn("ad_cleanup: $@") if ($@);

    eval {
        if (defined($saved_dns_info)) {
            logStep("ad_cleanup: Restore filer DNS information");
            # Connect to filer
            $fh = host($main::FILER);
            $fc = connect($main::FILER_CONN, -timeout, $main::TIMEOUT);
            $fc->timeout($main::CMDTIMEOUT);

            dns_restore_options(conn => $fc, options => $saved_dns_info);
        }
    };
    logWarn("ad_cleanup: $@") if ($@);

    logStep("END: ad_cleanup");
    return;
}

#---------------------------------------------------------------------

=head2 adsi_query

 @res = adsi_query(query => $query);

The following named arguments are accepted:

=over

=item Arguments

=over

=item password

(Required) : Password for the username.

=item query

(Required) : LDAP query string to process.

=item username

(Required) : User name to use on the LDAP query.

=back

=item Returns

An array containing the results of the query.

=back

=for html <hr />

=cut

#-------------------------------------------------------------------

sub adsi_query(%) {
    my %opts = validate(
        @_,
        {
            password => { type => SCALAR },
            query    => { type => SCALAR },
            username => { type => SCALAR },
        }
    );
    my ($err, $i, @list, $objNameSpace, $objSec);

    require Win32::OLE;
    import Win32::OLE qw(in);

    # Bind to the Active Directory server and run the query
    $objNameSpace = Win32::OLE->GetObject("LDAP:") or
        croak("adsi_query: Cannot create LDAP object\n: $!");
    $objSec = $objNameSpace->OpenDSObject("$opts{query}", $opts{username},
                                          $opts{password}, 1);
    $err = Win32::OLE->LastError();

    if ($err == 0) {
        # Build an array of results
        foreach $i (in $objSec) {
            push @list, $i->{AdsPath};
        }
        $objSec->Close;
    } else {
        croak("adsi_query: OLE error: $err\n");
    }

    $objNameSpace->Close;

    return(@list);
}

#---------------------------------------------------------------------

=head2 ad_val_filer

 $ret =  ad_val_filer(ad_server => $AD_ROOT_SERVER, filer => $FILER,
                      domain => $AD_DOMAIN, username => $CIFS_USER,
                      password => $CIFS_PASSWD);

Verify filer is in active directory domain.
The following named arguments are accepted:

=over

=item Arguments

=over

=item ad_server

(Required) : hostspec where active directory domain server is setup.

=item domain

(Required) : Active directory domain

=item filer

(Required) : Hostspec of a filer

=item password

(Required) : Password for username.

=item username

(Required) : User name to do the validation as.

=back

=item Returns

1 if filer is in active domain, 0 otherwise

=back

=for html <hr />

=cut

#-------------------------------------------------------------------

sub ad_val_filer(%) {
    my %opts = validate(
        @_,
        {
            ad_server => { type => SCALAR },
            domain    => { type => SCALAR },
            filer     => { type => SCALAR },
            password  => { type => SCALAR },
            username  => { type => SCALAR }
        }
    );
    my ($filer_name, $h, $query, @query_res, $rc);

    # Build the ADsPath for the domain
    $query = _LDAP_ADsPath($opts{ad_server}, $opts{domain}, 'cn=computers');

    # Send the query to the Active Directory Server
    @query_res = adsi_query(query => $query, username => $opts{username},
                            password => $opts{password});

    # Get the filer name
    $h = host($opts{filer});
    $filer_name = $h->hostname();
    $filer_name =~ tr/a-z/A-Z/;

    # See if the filer name was returned
    $rc = 0;
    foreach (@query_res) {
        if ($_ =~ /$filer_name/) {
            $rc = 1;
        }
    }

    return $rc;
}

#---------------------------------------------------------------------

=head2 ad_val_domain

 ad_val_domain(dns_mode => 'standard', domain => $AD_DOMAIN,
               mode => 'mixed', root_server => $AD_ROOT_SERVER,
               username => $CIFS_USER, password => $CIFS_PASSWD,
               windows_version => '5.2');

Validate that the active directory domain has been setup according to
the passed in parameters.

The following named arguments are accepted:

=over

=item Arguments

=over

=item child_domain

(Required) : hostspec where active directory domain server is setup.

=item child_server

(Required) : hostspec of child domain active directory server.

=item dns_mode

(Required) : Type of DNS mode. Valid values are standard, ddns, or unix.

=item domain

(Required) : Active directory domain

=item mode

(Required) : Active directory mode. Valid values are mixed or native.

=item password

(Required) : Password for username.

=item root_server

(Required) : hostspec of root domain active directory server.

=item username

(Required) : Username to validate with.

=item windows_version

(Optional) : MS Windows version number. Defaults to 5.

=back

=item Returns

1 if filer is in active domain, 0 otherwise

=back

=for html <hr />

=cut

#-------------------------------------------------------------------

sub ad_val_domain(%) {
    my %opts = validate(
        @_,
        {
            child_domain    => { type => SCALAR, optional => 1 },
            child_server    => { type => SCALAR, optional => 1 },
            dns_mode        => { type => SCALAR,
                                 regex => qr/(standard|ddns|unix)/ },
            domain          => { type => SCALAR },
            mode            => { type => SCALAR,
                                 regex => qr/(mixed|native)/ },
            password        => { type => SCALAR },
            root_server     => { type => SCALAR },
            username        => { type => SCALAR },
            windows_version => { type => SCALAR, default => '5',
                                 regex => qr/^(5\.0|5\.1|5\.2|5)$/ }
        }
    );
    my (@arr, $err, $h, $msg, $objNameSpace, $objSec, $query, $srv_name, $value);

    require Win32::OLE;
    import Win32::OLE qw(in);

    # Build the ADsPath for the domain
    $query = _LDAP_ADsPath($opts{root_server}, $opts{domain}, '');

    # Bind to the Active Directory server and run the query
    $objNameSpace = Win32::OLE->GetObject("LDAP:") or
        croak("ad_val_domain: Cannot create LDAP object\n: $!");
    $objSec = $objNameSpace->OpenDSObject("$query", $opts{username},
                                          $opts{password}, 1);
    $err = Win32::OLE->LastError();
    if ($err != 0) {
        croak("ad_val_domain: OLE error: $err\n");
    }

    # Make sure the Active Directory Server is running
    # Returns 0 => disabled  1 => enabled
    $value = $objSec->Get("serverState");
    $err = Win32::OLE->LastError();
    if ($err == 0) {
        if ($value == 0) {
            $msg += "Active Directory Server is disabled\n";
        }
    } else {
        croak("ad_val_domain: serverState: OLE error: $err\n");
    }

    # Check the mode of the Active Directory Server
    # Returns 0 => native  1 => mixed
    $value = $objSec->Get("nTMixedDomain");
    $err = Win32::OLE->LastError();
    if ($err == 0) {
        if (($opts{mode} eq "native" && $value != 0) ||
            ($opts{mode} eq "mixed" && $value != 1)) {
            $msg += "Active Directory: expected mode: $opts{mode}\n";
        }
    } else {
        croak("ad_val_domain: nTMixedDomain: OLE error: $err\n");
    }

    # XXX
    # Need to validate if child/root but not seeing a good
    # way right now.
    #
    # Need to validate if DDNS is set but not seeing a good
    # way right now.
    #
    # Need to validate if DDNS is set but not seeing a good
    # way right now. (operatingSystemVersion didn't work)

    $objSec->Close;
    $objNameSpace->Close;

    croak ($msg) if ($msg);
    return;
}

#-----------------------------------------------------------------------

=head2 static_ip_set

static_ip_set(ip => $ip, mask => $mask, gate => $gate, dns => \@dns);

Configure TCP/IP for a static ip address, gateway, and dns.

The following named arguments are accepted:

=over

=item Arguments

=over

=item ip

(Required) : Target IP to configure TCP/IP.

=item mask

(Required) : Target Netmask to configure TCP/IP.

=item gate

(Required) : Target Gateway IP to configure TCP/IP.

=item dns

(Required) : Target DNS address'es to configure TCP/IP.

=back

=back

=for html <hr />

=cut
#-----------------------------------------------------------------------

sub static_ip_set(%) {
    my %opts = validate(
        @_,
        {
            ip   => { type => SCALAR,
                    regex => qr/^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/ },
            mask   => { type => SCALAR,
                    regex => qr/^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/ },
            gate => { type => SCALAR,
                    regex => qr/^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/ },
            dns  => { type => ARRAYREF }
        }
    );

    my ($res, $i, $res2, $j, @ip_address, @netmask, @gateway, @gate_metric);

    require Win32::OLE;
    import Win32::OLE qw(in);

    @ip_address = ($opts{ip});
    @netmask = ($opts{mask});
    @gateway = ($opts{gate});
    @gate_metric = (1);

    $res = Win32::OLE->GetObject("winmgmts:\\\\.\\root\\CIMV2") or croak("static_ip_set (GetObject): $!");

        $res2 = $res->ExecQuery("Select * from Win32_NetworkAdapterConfiguration where IPEnabled=TRUE");
        foreach $j (in $res2) {
            $j->EnableStatic(\@ip_address, \@netmask);
            if (Win32::OLE->LastError()) {
                croak("static_ip_set (EnableStatic): ".Win32::OLE->LastError()."\n");
            }
            $j->SetGateways(\@gateway, \@gate_metric);
            if (Win32::OLE->LastError()) {
                croak("static_ip_set (SetGateways): ".Win32::OLE->LastError()."\n");
            }
            $j->SetDNSServerSearchOrder($opts{dns});
            if (Win32::OLE->LastError()) {
                croak("static_ip_set (SetDNSServerSearchOrder): ".Win32::OLE->LastError()."\n");
            }
        }
    return;
}

#-------------------------------------------------------------------------

=head2 ad_add_delete_domain_user

$ret = ad_add_delete_domain_user(user         => $user,
                                 pass         => $pass,
                                 flag         => 'add');

CIFS client command line interface to Add/delete a domain user.

The following name arguments are accepted:

=over

=item Arguments

=over

=item user

(Required) : The domain user to add or delete. It should be 20 characters
or less (limitation of Windows net command).

=item pass

(Required) : The domain user's password.

=item flag

(Optional) : The add/delete flag. Defaults to add.

=back

=item Returns

Will croak on failure.

=back

=for html <hr />

=cut
#--------------------------------------------------------------------------

sub ad_add_delete_domain_user(%) {
    my %opts = validate(
        @_,
        {
            user        => { type => SCALAR },
            pass        => { type => SCALAR },
            flag        => { type => SCALAR , default => 'add',
                             regex => qr/^(add|delete)$/}
        }
    );

    my($cmd, $out, $pass, $user);
    $user = $opts{user};
    $pass = $opts{pass};

    $out = length($user);
    if ($out > 20) {
        croak("ad_add_delete_domain_user: user $user name length $out " .
              "> 20 character maximum");
    }

    if ($opts{flag} eq 'add') {
        $cmd = "net user $user $pass /add /domain";
    }
    else {
        $cmd = "net user $user /delete /domain";
    }

    $out = qx($cmd);
    if ($out !~ /completed\s+successfully/) {
        croak("ad_add_delete_domain_user: cmd: $cmd output: $out\n");
    }
    return;
}

#-------------------------------------------------------------------------

=head2 ad_list_groups

@groups = ad_list_groups(fc => $fc);

Retrieve the filer local groups.

The following name arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : The filer's connection object.

=back

=item Returns

The following are returned in an array of hashes:
each hash contains:

$ret->{Name} = group name
$ret->{Info} = info
$ret->{Rid}  = Rid number
$ret->{Role} = The groups role

=back

=for html <hr />

=cut
#--------------------------------------------------------------------------

sub ad_list_groups(%) {
    my %opts = validate(
        @_,
        {
            fc => { type => OBJECT },
        }
    );

    my ($fc) = $opts{fc};

    my $group_text = sayEOF($fc, "useradmin group list");
    my (@groups) = ();
    my $group = {};

    for (split(/\n/, $group_text)) {
        $_ =~ s/^\s*//;
        $_ =~ s/\s*$//;

        if (!/\S+/) {
            next;
        }

        if (/Name:\s*(.*)/) {
            $group = {};
            $$group{Name} = $1;
            push(@groups, $group);
        } elsif (/(\w+):\s*(.*)/) {
            $$group{$1} = $2;
        }
    }

    return(@groups);
}

#---------------------------------------------------------------------

=head2 testme

Self test routine for this library.

Use test_lib.thpl to invoke this routine.

Example
    ntest -sim PKGNAME="CIFS::Ad" SUBNAME=testme CLIENT='minwin_clnt'
    FILER=myfiler util/test_pkg.thpl

=over

=item Arguments

=over

=item CLIENT

(Required) : Minwin compatible client on which to run testme.

=item FILER

(Required) : Filer to test.

=back

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub testme() {
    use CIFS::CommonParams;
    use Utils qw(subtest_remote);
    use TestEnv::MinWin qw(minwin_compatible);

    my $host = host($main::CLIENT);
    if ($host->hosttype() !~ /windows/ || !minwin_compatible($main::CLIENT)) {
        logresult('CONFIG', 'The host ' . $host->hostname() .
                  ' is not MinWin compatible, exiting.');
        return;
    }

    if ($main::CLIENT ne 'localhost') {
        eval {
            subtest_remote(test   => 'util/test_pkg.thpl',
                           client => $main::CLIENT);
        };

        logresult('CONFIG', $@) if ($@);
        exit(0);
    }

    my ($fh, $fc);

    $fh = host($main::FILER);
    $fc = connect($main::FILER_CONN, -timeout, $main::TIMEOUT);
    $fc->timeout($main::CMDTIMEOUT);

    eval {
        my @groups = ad_list_groups(fc => $fc);
    };

    if ($@) {
        logresult('FAIL', $@);
    } else {
        logresult('PASS', 'testme(): Ntfs.pm');
    }
    return;
}

1;
