package CIFS::AuditLog::ParseXML;

use strict;
use warnings;
use CIFS::XML::Rules;
use XML::LibXML;
use Data::Dumper;
use Data::Compare ();
use File::Basename qw(fileparse);

use Exporter ();
our @ISA       = qw( Exporter );
our @EXPORT_OK = qw( geteventsfor get_default_xsd );

my $ThisModule = "CIFS/AuditLog/ParseXML.pm";

#my $XSDFilename = "Ntap_audit_log.xsd";
my $XSDFilename = "netappMain.xsd";

#my $TopTag = "ntap-audit-log";
my $TopTag       = "Events";
my $CloseEventId = 5002;

our (%Filters, %Rules, %Events);

sub new {
    my $class  = shift;
    my %config = @_;
    my $self   = {};
    my $data;

    eval {
        require 'CIFS/AuditLog/filters';
        require 'CIFS/AuditLog/rules';
        require 'CIFS/AuditLog/getevents';
    };
    if ($@) {
        die "Error while loading files: $@";
    }
    if (not defined $config{source}) {
        die "Need to specify a source (string or file)\n";
    }
    my $parser = new CIFS::XML::Rules(
        rules => [
            _default => sub { warn "Unhandled Tag: $_[0]\n"; exit },
            %Rules,
        ],
        normalisespaces => 1,
        stripspaces     => 3 | 4,
    );
    if ($config{source} eq "string") {
        if (not defined $config{xml}) {
            die "No XML content given\n";
        }
        my $data;
        eval { $data = $parser->parsestring($config{xml}); };
        if ($@) {
            warn "Error parsing string:" . $@ . "END\n";
            return undef;
        }
        $self = {Events => $data,};
        bless $self;
        return $self;
    }
    unless ($config{'no-validation'}) {
        my $ex = schema_validation(%config);
        if ($ex) {
            warn "Schema validation FAILED: $ex\n\n";
            return undef;
        } else {
            print "Schema validation PASSED\n\n";
        }
    } else {
        print "Schema validation SKIPPED\n\n";
    }
    if ($config{source} eq "file") {
        eval { $data = $parser->parsefile($config{xml}); };
        if ($@) {
            warn "Error parsing file: $config{xml} $@\n";
            return undef;
        }
    } elsif ($config{source} eq "xmlstring") {
        eval { $data = $parser->parsestring($config{xmlstring}); };
        if ($@) {
            warn "Error parsing file: $config{xmlstring} $@\n";
            return undef;
        }
    }
    my $n = scalar @{$data->{$TopTag}{Event}};
    $self = {
        Events       => delete $data->{$TopTag}{Event},
        attributes   => $data->{$TopTag},
        no_of_events => $n,
    };
    bless $self;
    return $self;
} ## end sub new

sub schema_validation {
    my %config = @_;
    my $schemafile
        = defined $config{xsdfile} ? $config{xsdfile} : get_default_xsd();
    my $schema = XML::LibXML::Schema->new(location => $schemafile);
    my $parser = XML::LibXML->new;
    my $doc;
    if ($config{source} eq "xmlstring") {
        $doc = $parser->parse_string($config{xmlstring});
    } elsif ($config{source} eq "file") {
        $doc = $parser->parse_file($config{xml});
    }
    eval { $schema->validate($doc) };
    return $@;
} ## end sub schema_validation

sub get_default_xsd {
    return (fileparse($INC{$ThisModule}))[1] . $XSDFilename;
}

sub _eliminate_close_events {
    my $self  = shift;
    my $first = $self->{no_of_events};
    @{$self->{Events}}
        = grep { $_->{id} ne "$CloseEventId" } @{$self->{Events}};
    $self->{no_of_events} = @{$self->{Events}};
    my $ret = $first - $self->{no_of_events};
    print "$ret close Events deleted\n";
    return $ret;
} ## end sub _eliminate_close_events

sub dump {
    my $self   = shift;
    my %config = @_;
    print Dumper [$self->getevents(filter => "minimal", %config)];
}

sub compareevents {
    my $self     = shift;
    my %config   = @_;
    my %defaults = %{$config{defaults}};
    my @expected;
    my $deleteclose = $config{deleteclose};
    my @fields      = @{$config{fields}};

    foreach my $expected (@{$config{expected}}) {
        push @expected, {%defaults, %$expected};
    }
    @expected = map { _getfiltered($_, [@fields]) } @expected;
##     @expected = @{$config{expected}};

    if ($deleteclose) {
        $self->_eliminate_close_events();
    }

    my @actualevents = $self->getevents(filter => [@fields]);

    my $no_expected = @expected;
    my $no_actual   = @actualevents;

    if ($no_expected != $no_actual) {
        warn
            "Number of Events do not match: Expected: $no_expected, Actual: $no_actual. Skipping comparison\n";
        print Dumper [@actualevents];
        return 0;
    }

    my $result = Data::Compare::Compare([@actualevents], [@expected]);
    if ($result) {
        printf "%d Events match expected!\n", scalar @expected;
    } else {
        print "Events do not match!\n";
        print "Expected: ";
        print Dumper [@expected];
        print "Actual: ";
        print Dumper [@actualevents];
        return 0;
    }
    return 1;
} ## end sub compareevents

sub checktime {
    my $self     = shift;
    my $prevtime = 0;
    my $n        = @{$self->{Events}};
    my $xml      = $self->{filename};
    for (my $i = 0; $i < @{$self->{Events}}; $i++) {
        my $event = ${$self->{Events}}[$i];
        (my $time = $event->{datetime}) =~ s/[^\d\.]//g;
        if ($prevtime > $time) {
            printf
                "$xml: Chronological error in Event #%d - %s (%s) on path [%s].\n%4d) $prevtime\n%4d) $time\n",
                $i + 1, $event->{EventID}, $event->{EventName},
                $event->{ObjectName}, $i, $i + 1;
            return 0;
        }
        $prevtime = $time;
    }
    print "All OK.\n";
    return 1;
} ## end sub checktime

sub _get_close_events {
    my $self = shift;
    my @closeevents = grep { $_->{id} eq $CloseEventId } @{$self->{Events}};
}

sub _inarray {
    my ($element, $arrayref) = @_;
    return grep { $_ eq $element } @$arrayref;
}

sub _getfiltered {
    my ($event, $filter) = @_;
    my %event = map { $_ => $event->{$_} }
        grep { _inarray($_, $filter) } keys %$event;
    return \%event;
} ## end sub _getfiltered

sub getevents {
    my $self   = shift;
    my %config = @_;
    my %where  = defined $config{where} ? %{$config{where}} : ();
    my $filter = defined $config{filter} ? $config{filter} : undef;
    my @return = @{$self->{Events}};
    foreach my $given_tag (keys %where) {
        my $searchvalue = $where{$given_tag};
        @return = grep {
            defined $_->{$given_tag}
                and $_->{$given_tag} eq $searchvalue
        } @return;
    }

    $filter = "all" unless defined $filter;

    $filter = $Filters{$filter} unless ref $filter eq ref [];

    if (defined $filter) {
        @return = map { _getfiltered($_, $filter) } @return;
##         foreach my $event (@return) {
##             foreach my $key (%$event) {
##                 print "Deleting $key\n" and delete $event->{$key} unless _inarray($key, $filter);
##             }
##         }
    }
    return @return;
} ## end sub getevents

sub pygetevents {
    my $self = shift;
    $Data::Dumper::Pair = ": ";
    my $str = Dumper [$self->getevents()];
    $str =~ s/^\$\w+\s*=\s*|;$//g;
    $str =~ s/undef,/None,/g;
    return $str;
} ## end sub pygetevents

sub geteventsfor {
    my $event = shift;
    die "Unknown event: $event\n" unless defined $Events{$event};
    return @{$Events{$event}};
}

1;

