# -*- perl -*-
# $Id$

## @summary Routines for invoking character mapping command
## @author dl-maryv (FSS Protocols QA)
## @status Shared
## @pod here

=head1 NAME

Charmap - Routines for invoking character mapping command

=head1 SYNOPSIS

    use CIFS::Charmap;

    my $fh = host($FILER);

    charmap_is_supported ($fh); 
    charmap_is_enabled ($fh); 
    charmap_command($fh,<vol_name>,<nfs_hex:unicode_hex,...>); 
    charmap_command_cleanup($fh,<vol_name>); 
    .....

=head1 DESCRIPTION

These functions are used primarily by Character map tests.
It handles charmap command set for a volume such as set, display and cleanup

=head1 ROUTINES

=cut

package CIFS::Charmap;

require Exporter;
use strict;
use Tharn;
use Data::Dumper;
use Carp;
use SayWrap qw(sayEOF);
use vars qw (@ISA @EXPORT_OK);

@ISA = qw(Exporter);
@EXPORT_OK = qw (
		charmap_is_supported
		charmap_is_enabled
		charmap_command
		charmap_command_cleanup
		 		 );

use Filer::Version qw(get_wafl_version WV_GORDONBIERSCH_MIN
		      UNSUPPORTED_FEATURE_EXCEPTION);


=head2 charmap_is_supported

This function checks if charmap feature is supported in current 
ontap release 

=over

=item fh

(Required) :  filer host object

=back

=cut

sub charmap_is_supported {

    my $fh = shift;

    my $version = get_wafl_version($fh);
    if ($version < WV_GORDONBIERSCH_MIN) {
		croak(UNSUPPORTED_FEATURE_EXCEPTION . 
        ": character mapping unsupported in release prior to GORDONBIERSCH");
		return 0;
    }
    return 1;
}


=head2 charmap_is_enabled

This function checks if character mapping is only enabled in
advanced mode

=over

=item fh

(Required): Filer host object

=back

=cut

sub charmap_is_enabled{

    my $fh = shift;
    my $result = sayEOF($fh,"priv set admin; charmap");

    if ($result =~ /^charmap not found/i) {
		return 0;
    }
    return 1;
}


=head2 charmap_command

This function issues character mapping commands for set or display
and returns command output

=over

=item conn

(Required): connectspec of device

=item vol

(Required): volume to set/display character mapping

=item value

(Optional): pair of arguments representing nfs character value 
and cifs unicode value. 

=back

=cut


sub charmap_command($$$) {

    my ($fh,$vol,$val_pair) = @_;
    my $result = "";
      
    my $cmd ="priv set advanced; charmap";
    $result = sayEOF($fh,"$cmd $vol $val_pair");

    return($result);
}


=head2 charmap_command_cleanup

This function deletes an existing charmap for a given volume

Returns output of specified command

=over

=item conn

(Required): connectspec of device

=item vol

(Required): volume to delete the character mapping

=back

=cut

sub charmap_command_cleanup($$) {

    my ($fh,$vol) = @_;
    my $result = "";
   
    my $cmd ="priv set advanced; charmap $vol " . '""';
    $result = sayEOF($fh,$cmd );

    return($result);
}

=head1 STATUS

Shared

=head1 ATTRIBUTES

ONTAP_COMMANDS = charmap

ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    FSS Protocols QA (dl-maryv)
    $Id$
    Last revised $Date: 2005/07/22 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

1;
