# $Id:$
# 
# Copyright (c) 2007 Network Appliance, Inc.
# All rights reserved.
#
## @summary Library of functions to assist the configuration of cifs.
## @author David Low (davidlow@netapp.com)
## @status public
## @pod here


package CifsSetupHelper;

use strict;
use Exporter;
use CIFS::Setup qw(cifs_setup);
use Tharn qw (say connect disconnect host $Log);
use Data::Dumper;
use Carp;


=head1 NAME

CifsSetupHelper.pm - Helper module for setting up cifs on a filer or vfiler
using CIFS:Setup.pm.

=head1 SYNOPSIS

    use CIFS::CifsSetupHelper;

    my $conf_helper;

    # to setup a vfiler:
    # $conf_helper = CifsSetupHelper->new($FILER, $CLIENT, $vfiler_name);

    # to setup the main filer:
    $conf_helper = CifsSetupHelper->new($FILER, $CLIENT);

    if (defined $conf_helper) {
        $Log->comment("CifsSetupHelper objected created fine.\n");
    } else {
        $Log->log(
            config_result("Failed to create CifsSetupHelper object.\n")
        );
    }

    # rerun cifs setup if cifs is already set up:
    $conf_helper->force_rerun_setup(1);
    
    my $ok;
    if (define $AD_DOMAIN && defined $AD_USER && defined $AD_USER_PASSWD) {
        $ok = $conf_helper->run_cifs_setup($AD_DOMAIN,$AD_USER,$AD_USER_PASSWD);
    } else {
        $ok = $conf_helper->run_cifs_setup();
    }
    
    if ($ok){
        $Log->comment("Cifs set up fine.\n");

        # Cifs is set up to run cifs related tests.

    } else {
        $Log->log(config_result("Failed to run cifs setup.\n"));

        # failed cifs setup, can not run cifs related tests.

    }
 
=cut

=head1 DESCRIPTION

This module simplifies the process of setting up cifs on a main filer or on a 
vfiler.  It checks for the most common causes of cifs setup failures, such
as cifs is not licensed, dns is not set up, and filer time is too far out of
sync with domain server.  To simplify cifs setup, this module uses most of the 
default setup parameters of CIFS::Setup.pm.  The non-default parameter 
settings are using only active directory domain authentication and not using 
WINS.  The user has the option to specify the Active Directory domain and 
credentials to use for cifs setup.  If they are not specified, this module 
tries to use the default Active Directory domain for the DNS domain the filer 
is using.

=cut

our $AUTOLOAD; 
my %autoload_methods = (
    # To set or get the log level setting:
    log_level   => 1,           # 0 = no logging from this module
                                # 1 = default, normal module logging.
                                # 2 = most verbose module logging.

    # To set or get the force_rerun_setup settting:
    force_rerun_setup => 0,     # 0 = do not run cifs setup if it is 
                                #     already set up.
                                # 1 = rerun cifs setup if it is set up.

    # To set or get the cifs license code:
    cifs_code => "GTLYRVH",     # default cifs site license code.

    # To allow cifs setup to run on a vfiler:
    vfiler => undef,
);


=head1 CONSTRUCTOR

=cut

=head2 new($filer, $client, $vfiler)

Create a cifs setup helper object.

This is the constructor method. It takes a filer hostspec object, a client 
hostspec object, and an optional vfiler name.  The CLIENT must be a UNIX 
machine.  If the cifs setup is for a vfiler, the name of the vfiler is 
specified with the $vfiler parameter.

B<Usage:> 

    $conf_helper = CifsSetupHelper->new($FILER, $CLIENT);
    $conf_helper = CifsSetupHelper->new($FILER, $CLIENT, $vfiler);

B<Returns:> C<$conf_helper> a cifs configuration helper object

=cut

sub new {
    my($this, $filer, $client, $vfiler)= (@_);
    if ((!defined $filer) || (!defined $client)) { 
        # missing required args.
        #
        $Log->log(fail_result("Missing required arguments: filer hostspec and "
            . "client hostspec."));
        return undef;
    }

    my($class)= ref($this) || $this;
    my($self)= { %autoload_methods };
    bless ($self, $class);

    if ($vfiler) {
        $self->vfiler($vfiler);
    } else {
        $self->vfiler(undef);
        $self->{"VFCP"} = "";  # no command prefix for the main filer.
    }

    my $h_filer = host($filer);
    my $filer_conn = connect("#rsh");
    if (!defined $filer_conn) {
        $Log->log(fail_result("Failed to create rsh connection to filer."));
        return undef;
    }
    my $h_client  = host($client);
    my $client_conn = connect("#telnet");
    if (!defined $client_conn) {
        $Log->log(fail_result("Failed to create telnet connection to client."));
        return undef;
    }
    $self->{"FILER_HOST"} = $h_filer; 
    $self->{"CLIENT_HOST"} = $h_client; 
    $self->{"FILER_CONN"} = $filer_conn;
    $self->{"CLIENT_CONN"} = $client_conn;

    # The following are the default site dependent AD domain to use if
    # the user does not specified an AD domain to use for cifs setup.
    # To add a new default AD domain, add a hash containing the following
    # "NAME" => AD domain name
    # "USER" => admin user of this domain
    # "PASSWD" => admin user password
    #
    $self->{"KNOWN_DOMAINS"} = [
        {  # original SVL cifs domain:
           "NAME" => "cifs.lab.netapp.com",
           "USER" => "administrator",
           "PASSWD" => "BigCheese" 
        },
        {  # new SVL cifs domain.
           "NAME" => "lab.netapp.com",
           "USER" => "labclient",
           "PASSWD" => "sundance" 
        },
        {  # RTP cifs domain.
           "NAME" => "rtp.netapp.com",
           "USER" => "mgtuser",
           "PASSWD" => "Passw0rd" 
        },
        {  # RTP susqa cifs domain.
           "NAME" => "susqa.rtp.netapp.com",
           "USER" => "susqa",
           "PASSWD" => "susqa" 
        },
        {  # NB BTC cifs domain.
           "NAME" => "lab.eng.btc.netapp.in",
           "USER" => "smdev1",
           "PASSWD" => "smdev1" 
        },
    ];

    return $self;
}

=head1 METHODS

=cut

=head2 add_cifs_license()

This methods adds the cifs license if it is not already licensed.

B<Usage:> $conf_helper-E<gt>add_cifs_license($license_code);

B<Access:> public

B<Input:> $license_code = the cifs license code.

B<Return value:> 1 if cifs is licensed. 0 if cifs is not licensed, and the
attemp to add it fails.

=cut

sub add_cifs_license {
    my ($self, $license_code) = @_;
    $self->_logmsg("Entering add_cifs_license().", 2);
    my $filer_conn = $self->{"FILER_CONN"};

    my $result = $self->is_cifs_licensed();
    if (!$result) {
        # cifs not licensed, add it.
        _send_cmd($filer_conn, "license add $license_code");
        $result = $self->is_cifs_licensed();
        if (!$result) {
            $self->_logmsg("Failed to add cifs license:$license_code.");
            return 0;  # failed to add the license.
        }
    } 
    $self->_logmsg("Exiting add_cifs_license().", 2);
    return 1;  # license is OK.
}


=head2 check_cifs_setup()

This method checks to see if "cifs setup" should be run.
If the returned status is "RUNNING", the given cifs domain
is already running, and no need to run cifs setup.  If the
retruned status is "CHECK-OK", "cifs setup"
may be run next.  If the returned status is "CHECK-FAILED", 
the user should look into the MESSAGE to see what failed, and 
decide how to correct the failure before running cifs setup.

B<Usage:> $conf_helper-E<gt>check_cifs_setup($cifs_domain,
    $cifs_license_code);

B<Access:> public

B<Input:> 

          $cifs_domain = name of cifs domain to check.
          $cifs_license_code = optional, cifs license code.  If not
                       given, and if cifs is not licensed, it will use the
                       default license code obtained by $self->cifs_code().

B<Return value:> A hash containing the following "STATUS" and "MESSAGE" pair.

          "STATUS" => "RUNNING", cifs is already running with the
                       passed in cifs domain, $cifs_domain.
          "MESSAGE" => a string containing the current cifs domain
                       name.
          ---------
          "STATUS" => "CHECK-OK", ready to run cifs setup (all checks 
                       passed), but the given cifs domain is not running.
          "MESSAGE" => a string containing the checks that are done. 
                       Possible values are "CIFS-DOMAIN", "ADD-LICENSE",
                       "SYNC-TIME", "VFILER-STATUS", and "CIFS-RUNNING".
          ---------
          "STATUS" => "CHECK-FAILED", filer is not ready to run cifs 
                       setup, some checks failed.
          "MESSAGE" => a string containing the checks that failed. 
                       Possible values are "CIFS-DOMAIN", "DNS-SETUP", 
                       "ADD-LICENSE", "SYNC-TIME", "VFILER-STATUS", and 
                       "UNKNOWN-FAILURE".  If there are multiple failures, 
                       they are separated using ",".  

=cut

sub check_cifs_setup {
    my ($self, $cifs_domain, $cifs_license_code) = @_;
    $self->_logmsg("Entering check_cifs_setup().", 2);
    if (!defined $cifs_license_code) {
        $cifs_license_code = $self->cifs_code();
    }
    my $fail_str;
    my $pass_str;
    my %result = (
        "STATUS" => "CHECK-FAILED",
        "MESSAGE" => "UNKNOWN-FAILURE"
    );

    # check vfiler status if vfiler is defined.
    #
    my $status = "CHECK-OK";
    if ($self->vfiler()) {
        $self->_logmsg("checking vfiler status next.");
        my $vfiler_running = $self->is_vfiler_running();
        if (!$vfiler_running) {
            $result{"STATUS"} = "CHECK-FAILED";
            $result{"MESSAGE"} = "VFILER-STATUS";
            return %result; # no need to check further if vfiler is not running.
        } else {
            $pass_str = "VFILER-STATUS";
        }
    }

    my %domain_info;
    my $is_running = $self->is_cifs_setup(host => $self->{"FILER_HOST"});
    if (!$is_running) {
        $self->_logmsg("Trying to restart cifs if possible.");
        $is_running = $self->cifs_restart();
    }

    if ($is_running) {
        %domain_info = $self->get_cifs_domain_info();
        if ($domain_info{"DOMAIN_NAME"} =~ /^\s*$cifs_domain\s*$/i) {
            
			# do sync_time so that force setup can run without time problem.
            if ($self->force_rerun_setup()) {
			    $self->_logmsg("Sync time next.");
			    if (!$self->sync_time()) {
                    $result{"STATUS"} = "CHECK-FAILED";
                    $result{"MESSAGE"} = "SYNC-TIME";
                    return %result;
			    }
            }

            $result{"STATUS"} = "RUNNING";
            $result{"MESSAGE"} = $domain_info{"DOMAIN_NAME"};
            $self->_logmsg("cifs is running, returning status: " 
                . $result{"STATUS"});
            return %result;
        }
        # cifs running, but with a different domain.
        $pass_str = $pass_str . "CIFS-RUNNING";
    }

    # check dns setup on filer.
    #
    $self->_logmsg("checking filer dns setup next.");
    my $dns_domain = $self->get_default_dns_domain();
    if (!defined $dns_domain) {
        $status = "CHECK-FAILED";
        $fail_str = "DNS-SETUP";
    } else {
        $pass_str = $pass_str . ",DNS-SETUP";
    }

    # do sync_time
    #
    $self->_logmsg("Sync time next.");
    if (!$self->sync_time()) {
        $status = "CHECK-FAILED";
        $fail_str = $fail_str . ",SYNC-TIME";
    } else {
        $pass_str = $pass_str . ",SYNC-TIME";
    }

    # Check and add cifs license if needed.
    #
    $self->_logmsg("Check cifs license next.");
    my $cifs_lic = $self->add_cifs_license($cifs_license_code);
    if (!$cifs_lic) {
        $status = "CHECK-FAILED";
        $fail_str = $fail_str . ",ADD-LICENSE";
    } else {
        $pass_str = $pass_str . ",ADD-LICENSE";
    }

    $fail_str =~ s/^\s*,+//;  # remove leading "," if there is any.
    $pass_str =~ s/^\s*,+//;
    $result{"STATUS"} = $status;
    if ($status eq "CHECK-OK") {
        $result{"MESSAGE"} = $pass_str;
    } else {
        $result{"MESSAGE"} = $fail_str;
    }
    $self->_logmsg("Exiting check_cifs_setup().", 2);
    return %result;
}


=head2 cifs_restart()

This method tries to restart cifs using "cifs restart".

B<Usage:> $conf_helper-E<gt>cifs_restart();

B<Access:> public

B<Input:> none

B<Return value:> 1 if cifs restarts successfully. 0 if cifs does not restart.

=cut

sub cifs_restart {
    my $self = shift;
    $self->_logmsg("Entering cifs_restart().", 2);
    my $filer_conn = $self->{"FILER_CONN"};
    my @outputs = _send_cmd($filer_conn, $self->{"VFCP"} . "cifs restart");
    if (grep (/CIFS local server is running/i, @outputs)) {
        $self->_logmsg("Cifs restarted successfully.");
        return 1;  # cifs restarted.
    } 
    $self->_logmsg("Exiting cifs_restart().", 2);
    return 0;  #  cifs did not restart.
}


=head2 cleanup()

If "CIFS::cifs_setup()" fails, a "cifs setup" may be pending.  This method
tries to terminates a pending "cifs setup" by sending a ^C to the filer using
a telnet connection.

B<Usage:> $conf_helper-E<gt>cleanup($host_obj);

B<Access:> public

B<Input:> $host_obj = filer host object 

B<Return value:> 1=^C was sent.  0=^C was not sent.

=cut

sub cleanup {
    my $self = shift;
    my $host_obj = shift;
    $self->_logmsg("Entering cleanup().", 2);

    my $tc = connect($host_obj->{id} . "#telnet");
    if (!defined $tc) {
        $self->_logmsg("Failed to open telnet connection to filer.");
        return 0;
    }
    $self->_logmsg("Sending a ctrl+C to terminate a pending filer command.");
    my $ctrl_C = sprintf( "%c", 003 );    # [Ctrl]+C char;
    say( $tc, $ctrl_C, '-nocr', '-blind' );
    sleep 2;
    disconnect($tc);
    $self->_logmsg("Exiting cleanup().", 2);
    return 1;
}


=head2 get_cifs_domain_info()

This method runs the "cifs domaininfo" command on the filer
and returns the current cifs domain info in a hash.

B<Usage:> $conf_helper-E<gt>get_cifs_domain_info();

B<Access:> public

B<Input:> none

B<Return value:> A hash containing the following info corresponding to the
output of "cifs domaininfo".  If error occurred, undef is returned.

   "NETBIOS_DOMAIN" => Net Bios Domain.
   "DOMAIN_NAME" => domain name.
   "TYPE" => domain type.
   "AD_SITE" => filer AD site.
   "CONNECTED_DCS" => currently connected DCs.
   "TOTAL_DCS_FOUND" => total DS addresses found.
   "PREFERRED_DCS" => preferred addresses (DC).
   "FAVORED_DCS" => favored addresses (DC).
   "OTHER_DCS" => other addresses (DC).
   "CONNECTED_AD_SERVER" => connected AD LDAP server.
   "PREFERRED_ADS" => preferred addresses (AD).
   "FAVORED_ADS" => favored addresses (AD).
   "OTHER_ADS" => other addresses (AD).

=cut

sub get_cifs_domain_info {
    my $self = shift;
    $self->_logmsg("Enterning get_cifs_domain_info().", 2);
    my $fc = $self->{"FILER_CONN"};
    my $cifs_domain_info;
    ( undef, undef, $cifs_domain_info ) = say(
        $fc, $self->{"VFCP"} . "cifs domaininfo"
    );
    if ($cifs_domain_info !~ /Domain Name:\s+/) {
        $self->_logmsg("Failed to query CIFS domain info.");
        return undef;
    }    
    $self->_logmsg("Successfully queried CIFS domain info:\n" 
       . "$cifs_domain_info.");
    
    # parse the cifs domain info.
    #
    my @lines = split (/\n/,$cifs_domain_info);
    my %info;
    my $i = 0;
    while ($i<=$#lines) {
        if ($lines[$i] =~ /^\s*NetBios Domain:\s+(.+)/ ) {
            $info{"NETBIOS_DOMAIN"} = $1;
        } elsif ($lines[$i] =~ /Domain Name:\s+(.+)/ ) {
            $info{"DOMAIN_NAME"} = $1;
        } elsif ($lines[$i] =~ /^\s*Type:\s+(.+)/ ) {
            $info{"TYPE"} = $1;
        } elsif ($lines[$i] =~ /^\s*Filer AD Site:\s+(.+)/ ) {
            $info{"AD_SITE"} = $1;
        } elsif ($lines[$i] =~ /^\s*Current Connected DCs:\s+(.+)/ ) {
            $info{"CONNECTED_DCS"} = $1;
            $info{"CONNECTED_DCS"} =~ s/^\\+//;  # remove the leading "\"
        } elsif ($lines[$i] =~ /^\s*Total DC addresses found:\s+(.+)/ ) {
            $info{"TOTAL_DCS_FOUND"} = $1;
        } elsif ($lines[$i] =~ /^\s*Preferred Addresses:/ ) {
            if ( (defined $info{"TOTAL_DCS_FOUND"}) 
                && (!defined $info{"CONNECTED_AD_SERVER"}) ) {
                $info{"PREFERRED_DCS"} = _read_next_addrs(\@lines,$i);

            } else {
                $info{"PREFERRED_AD_SERVERS"} = _read_next_addrs(\@lines,$i);
            }
            $i++;
            next;
        } elsif ($lines[$i] =~ /^\s*Favored Addresses:/ ) {
            if ( (defined $info{"TOTAL_DCS_FOUND"}) 
                && (!defined $info{"CONNECTED_AD_SERVER"}) ) {
                $info{"FAVORED_DCS"} = _read_next_addrs(\@lines,$i);

            } else {
                $info{"FAVORED_AD_SERVERS"} = _read_next_addrs(\@lines,$i);
            }
            $i++;
            next;
        } elsif ($lines[$i] =~ /^\s*Other Addresses:/ ) {
            if ( (defined $info{"TOTAL_DCS_FOUND"}) 
                && (!defined $info{"CONNECTED_AD_SERVER"}) ) {
                $info{"OTHER_DCS"} = _read_next_addrs(\@lines,$i);

            } else {
                $info{"OTHER_AD_SERVERS"} = _read_next_addrs(\@lines,$i);
            }
            $i++;
            next;
        } elsif ($lines[$i] =~ /^\s*Connected AD LDAP Server:\s+(.+)/ ) {
            $info{"CONNECTED_AD_SERVER"} = $1;
            $info{"CONNECTED_AD_SERVER"} =~ s/^\\+//;  # remove leading "\".
        }
        $i ++;
    }
    $self->_logmsg("Exiting get_cifs_domain_info().", 2);
    return %info; 
}


=head2 get_default_dns_domain()

    Get the default DNS domain name.

B<Access:> public

B<Usage:> $conf_helper-E<gt>get_default_dns_domain();

B<Input:> none

B<Return value:> the default DNS domain name from "dns info".  If there
 is no default DNS domain found, undef is returned.

=cut

sub get_default_dns_domain {
    my ($self) = shift;
    $self->_logmsg("Entering get_default_dns_domain().", 2);
    my $fc = $self->{"FILER_CONN"};
    my ( undef, undef, $out, undef ) = say( $fc, $self->{"VFCP"} . "dns info");
    my ($dns_domain) = ($out=~ /default domain:\s+(\S+)\s*/i); 
    $self->_logmsg("Exiting get_default_dns_domain().", 2);
    return $dns_domain;
}


=head2 get_known_domain()

This method looks up an array of hard coded AD domains.  To add a new 
known domain, see the comments for $self->{"KNOWN_DOMAINS"} in the new()
constructor.

B<Usage:> $conf_helper-E<gt>get_known_domain($domain);

B<Access:> public

B<Input:> $domain = name of AD domain.

B<Return value:> If the given domain is known, this method returns a ref to a
hash containing the following.  If the given domain is unknown, this method 
returns undef.

    "NAME" => name of cifs domain.
    "USER" => admin user of domain.
    "PASSWD" => admin user password.

=cut

sub get_known_domain {
    my ($self, $domain) = @_;
    $self->_logmsg("Entering get_known_domain().", 2);
    foreach my $d (@{$self->{"KNOWN_DOMAINS"}}) {
        if ($d->{"NAME"} =~ /^\s*$domain\s*$/i) {
            if (defined $d->{"USER"} && defined $d->{"PASSWD"}) {
                return $d; # known domain hash.
            } 
        }
    }
    $self->_logmsg("Domain \"$domain\" is NOT known.");
    $self->_logmsg("Exiting get_known_domain().", 2);
    return undef;  # unknown domain
}


=head2 is_cifs_licensed()

This method queries the cifs license of the filer.

B<Usage:> $conf_helper-E<gt>is_cifs_licensed();

B<Access:> public

B<Input:> none

B<Return value:> 1 if cifs is licensed, and 0 if not.

=cut

sub is_cifs_licensed {
    my $self = shift;
    $self->_logmsg("Entering is_cifs_licensed().", 2);
    my $filer_conn = $self->{"FILER_CONN"};

    my @outputs = _send_cmd($filer_conn, "license");
    if (grep (/cifs\s+not\s+licensed/, @outputs)) {
        $self->_logmsg("CIFS is not licensed.");
        return 0;  # not licensed.
    } 
    $self->_logmsg("CIFS is licensed.");
    $self->_logmsg("Exiting is_cifs_licensed().", 2);
    return 1;  # license is there.
}


=head2 is_cifs_setup()

Check to see if cifs is already set up and running.  Since the methed
CIFS::Setup::is_cifs_setup() does not work on vfilers, we will implement 
the check here.

B<Usage:> $conf_helper-E<gt>is_cifs_setup();

B<Access:> public

B<Input:>  none

B<Return value:> 1= cifs is set up.  0= cifs is not set up.

=cut

sub is_cifs_setup {
    my $self = shift;

    my $fc = $self->{"FILER_CONN"};
    my $cmd = "cifs sessions";
    my (undef, undef, $out) = say($fc, $self->{"VFCP"} . $cmd);
    if (($out =~ /not licensed/) || ($out =~ /not running/)) {
        return 0;
    } 
    return 1;
}


=head2 is_vfiler_running()

Check to see if the vfiler is running using vfiler status. 

B<Usage:> $conf_helper-E<gt>is_vfiler_running();

B<Access:> public

B<Input:>  none

B<Return value:> 1= vfiler is running.  0= vfiler is not running.

=cut

sub is_vfiler_running {
    my $self = shift;

    $self->_logmsg("Entering is_vfiler_running().", 2);
    my $vfiler_name = $self->vfiler();
    if (!defined $vfiler_name) {
        $self->_logmsg("Vfiler is not specified.");
        return 0; # no such vfiler
    }

    my $fc = $self->{"FILER_CONN"};
    my $cmd = "vfiler status";
    my (undef, undef, $out) = say($fc, $cmd);
    if ($out =~ /^\s*$vfiler_name\s+running\s*$/m) { 
        $self->_logmsg("Vfiler \"$vfiler_name\" is running.");
        return 1;
    } 
    $self->_logmsg("Vfiler \"$vfiler_name\" is not running.");
    $self->_logmsg("Exiting is_vfiler_running().", 2);
    return 0;
}


=head2 run_cifs_setup()

    Run cifs_setup of CIFS::Setup.pm

B<Access:> public

B<Input:>

    $domain = optional, name of AD domain.
    $user   = optional, name of domain user.
    $passwd = optional, password of domain user.

B<Return value:> 1 if cifs setup is successful, 0 otherwise.

=cut

sub run_cifs_setup {
    my ($self, $domain, $user, $passwd) = @_;
    if (!defined $domain | !defined $user |!defined $passwd) {
        # get the domain from the known domains.
        $domain = $self->get_default_dns_domain();
        if (!defined $domain) {
            $self->_logmsg("Do not have AD domain info to run "
                . "run_cifs_setup().", 2);
            return 0;
        } 
        my $h = $self->get_known_domain($domain);
        if (!defined $h) {
            $self->_logmsg("Domain $domain is not known.", 2);
            return 0;
        } 
        $domain = $h->{"NAME"};
        $user = $h->{"USER"};
        $passwd = $h->{"PASSWD"};
    }

    my %check_result = $self->check_cifs_setup($domain);
    if ($check_result{"STATUS"} eq "RUNNING") {
        if (!$self->force_rerun_setup()) {
            return 1;  # the given cifs domain is already set up and running.
        }
    } elsif  ($check_result{"STATUS"} eq "CHECK-FAILED") {
        $self->_logmsg("check_cifs_setup($domain) failed: "
            . $check_result{"MESSAGE"} , 2);
        return 0;  # check failed, can not run cifs_setup.
    } 

    # now run cifs setup using minumum parameters.
    my $ret_cifs_running = 0;
    my %args;
    my $host = $self->{"FILER_HOST"};
    $args{'host'}        = $host;
    $args{'server-name'} = substr ($host->hostname(),0,15); # 15 ch or less.
    $args{'domain'}      = $domain;
    $args{'domain_user'} = $user;
    $args{'domain_pass'} = $passwd;
    $args{'auth-type'}   = 'ad'; # Active Directory Domain Authentication only.
    $args{'wins_option'} = 3;    # Do not use WINS.
    $args{'cifs_pass'} = $self->{"FILER_HOST"}{"default_password"}; 
    if ($self->vfiler()) {
        $args{'vfiler'} = $self->vfiler();
        $args{'server-name'} = substr ($self->vfiler(),0,15); 
    }

    $self->_logmsg("running cifs setup with the following args hash.\n"
        . Dumper(%args), 2);
    eval { cifs_setup(%args); };
    if ($@) {
        $self->_logmsg("cifs setup failed:" . $@, 2);
        $self->cleanup($self->{"FILER_HOST"});
        return $ret_cifs_running;
    }
    else {
        # check cifs again after configuration.
        #
        my $ret = $self->is_cifs_setup(host => $host);  # return 1 if set up.
        if ( $ret == 1 ) {
            $self->_logmsg("CIFS is configured successfully and running.", 2);
            $ret_cifs_running = 1;
        }
        else {
            $self->_logmsg("CIFS is not running after configuration.", 2);
            $ret_cifs_running = 0;
        }
    }
    return $ret_cifs_running;
}



=head2 sync_time()

This method syncs the filer time to an NetApp rdate server.  
If "options timed.enable on" and "options timed.servers" is not blank,
this method assumes the time is sync and skips the use of rdate.

B<Usage:> $conf_helper-E<gt>sync_time();

B<Access:> public

B<Input:> none

B<Return value:> 1 if time is synced to an rdate server, and 0 if not.

=cut

sub sync_time {
    my $self = shift;
    my $filer_conn = $self->{"FILER_CONN"};
    $self->_logmsg("Entering sync_time().", 2);
       
    # the following is a list of rdate time servers that NatApp IT supports.
    #
    my @rdate_servers = (  
        "timehost0.eng.netapp.com",
        "timehost1.eng.netapp.com", 
        "timehost2.eng.netapp.com", 
        "ntp0.eng.netapp.com", 
        "ntp1.eng.netapp.com", 
        "ntp2.eng.netapp.com", 
        "ntp3.eng.netapp.com", 
        "ntp4.eng.netapp.com",
        ); 

    # Check for active timed setup.  If it is already set, assume no
    # need to sync time and return successful.
    #
    my $timed_active = 1;
    my $time_synced = 0;
    my @cmd_outs = _send_cmd($filer_conn, "options timed");
    for (my $i = 0; $i <= $#cmd_outs; $i++) {
        my $tmp = $cmd_outs[$i];
        $tmp =~ s/\(.*\)//; # remove anything that is inside "()".
        if ($tmp =~ /timed.enable\s+off/) {  # timed is not active.
            $timed_active = 0;
            last;
        } elsif ( $tmp =~ /timed.servers\s+$/ ) {
            # Timed.enable is on, but no timed.servers sepecified.
            # We will set timed.enable off and try to sync time using rdate.
            #
            _send_cmd($filer_conn, "options timed.enable off");
            $timed_active = -1; # remember to set timed.enable to on later.
        }
    }
    if ($timed_active > 0) {
        $self->_logmsg("Found timed.enable on.  Assume time is sync.");
        $time_synced = 1;
        goto END_SYNC_TIME;
    }

    # 
    # run rdate with each server to see if any of them works.
    #
    foreach my $server (@rdate_servers) {
        $self->_logmsg("Trying to sync filer time to time server $server.");
        my $cmd = "rdate $server"; 
        @cmd_outs = _send_cmd($filer_conn, $cmd);
        if ($cmd_outs[0] !~ /^rdate:\s+/) {
            # command successful.  Time sync to time server done.
            $self->_logmsg("Successfully synchronized filer time to "
                . "$server using rdate.");
            $time_synced = 1;
            last;
        }
    }

END_SYNC_TIME:;
    if ( $timed_active < 0 ) { # need to set timed.enable back to on.
        _send_cmd($filer_conn, "options timed.enable on");
    }

    $self->_logmsg("Exiting sync_time().", 2);
    if (!$time_synced) {
        $self->_logmsg("Failed to synchronize filer time to a NatApp time "
            . "server.");
        return 0;
    }
    return 1;
}


# AUTOLOAD methods:
#

=head2 cifs_code()

This method sets or queries the cifs license code.  If the $settings parameter
is given, it is used to udpate the cifs license code.  If this parameter is
missing, this method returns the current cifs license code.

B<Usage:> $conf_helper-E<gt>cifs_code($setting);

B<Access:> public

B<Input:> $setting = optional, set the cifs license code to be used.

B<Return value:> currect cifs license code.

=cut


=head2 force_rerun_setup()

This method sets or queries the force_rerun_setup flag.  If no input parameter 
is given, the current setting is returned.  If the $setting parameter is 
given, the flag is set to $setting.  If the force_rerun_setup flag is set to
1, cifs setup is run even if cifs is already set up.  If this flag is set to
0, cifs setup is not run if it is already set up.

B<Usage:> $conf_helper-E<gt>force_rerun_setup($setting);

B<Access:> public

B<Input:> $setting = optional, 0=do not run cifs setup if cifs is set up 
already.  1=rerun cifs setup if cifs is already set up.

B<Return value:> currect setting of force_rerun_setup.

=cut


=head2 log_level()

This method sets or queries the log level setting.  If the optional log level
 parameter is missing, this method queries and returns the current log level. 
 If the optional log level parameter is given, the log level is set to that 
value.

B<Usage:> $conf_helper-E<gt>log_level($log_level);

B<Access:> public

B<Input:> $log_level = optional, the log level value to set.  0=do not log 
messages of this module.  1=log normal messages.  2=log all of the messages.  

B<Return value:> current log level.

=cut


=head2 vfiler()

This method sets or queries the vfiler name that will be running cifs setup. 
 If the $vfiler_name parameter is given, it is used to udpate the vfiler() 
setting, and the next call to run_cifs_setup() will be running "cifs setup" on 
this vfiler.  If this parameter is missing, this method returns the current 
vfiler name, which may be "undef" if the vfiler name is never initialized by 
specifying the $vfiler parameter when calling constructor 
CifsSetupHelper->new() or by calling vfiler() with the vfiler name. 
To run cifs setup on multiple vfilers of the same physical filer, for each
vfiler use vfiler($vfiler_name) to specify the vfiler and then call 
run_cifs_setup().

B<Usage:> $conf_helper-E<gt>vfiler($vfiler_name);

B<Access:> public

B<Input:> $vfiler_name = optional, name of vfiler to set up cifs.

B<Return value:> currect vfiler name.

=cut


# private methods:
#

=head2 _extract_servers_from_list()

This method extracts the servers from the input string.

B<Access:> private

B<Input:> $s_str = a string containing a "," separated list of servers.

B<Return value:> An array containing servers read from the input string.
Undef, if no server found.

=cut

sub _extract_servers_from_list {
    my $s_str = shift; 
    my @ads;
    my @servers = split (",",$s_str);
    for (my $i=0; $i<=$#servers; $i++) {
        my $s = $servers[$i];
        $s =~ s/^\s*//;    # remove the leading white spaces.
        $s =~ /^(\S+)\s*/; # only want the first word.
        $s = $1;
        if ($s !~ /^\s*\$/) { # none blank
            push (@ads, $s);
        }
    }
    if ($#ads >=0) {
        return @ads;
    } else {
        return undef;
    }
}


=head2 _get_ad_servers()

This method extracts the AD servers from the input hash.
These servers correspond to the AD servers from the output
of "cifs domaininfo".

B<Access:> private

B<Input:> $ref_domainion = a reference to a hash that contains the cifs domain
info as returned by get_cifs_domain_info().

B<Return value:> An array containing the AD servers read from the input 
hash.  Undef if fails.

=cut

sub _get_ad_servers {
    my $ref_domaininfo = shift;
    my @ads;
    if (defined $ref_domaininfo->{"CONNECTED_AD_SERVER"}) {
        push (@ads, $ref_domaininfo->{"CONNECTED_AD_SERVER"});
    }
    if (defined $ref_domaininfo->{"PREFERRED_AD_SERVERS"}) {
        # the ADs may be in a "," separated list.
        #
        my @s = _extract_servers_from_list(
            $ref_domaininfo->{"PREFERRED_AD_SERVERS"}
        );
        if ($s[0]) { 
            push (@ads, @s);
        }
    }
    if (defined $ref_domaininfo->{"FAVORED_AD_SERVERS"}) {
        my @s = _extract_servers_from_list(
            $ref_domaininfo->{"FAVORED_AD_SERVERS"}
        );
        if ($s[0]) { 
            push (@ads, @s);
        }
    }
    if (defined $ref_domaininfo->{"OTHER_AD_SERVERS"}) {
        my @s = _extract_servers_from_list(
            $ref_domaininfo->{"OTHER_AD_SERVERS"}
        );
        if ($s[0]) { 
            push (@ads, @s);
        }
    }
    if ($#ads<0) {
        return undef;
    } else {
        return @ads;
    }
}


=head2 _logmsg()

This method is used by other methods to write messages to the test log.
It uses logcommnet() to write the messages, depending on the setting of 
$self->log_level().  If $self->log_level() == 0, nothing is logged.
If $self->log_level() >= 1, the given message is logged if $log_lvl is
less than or equal to $self->log_level().  The message is logged 
with a header string to indicate that this message is from this module. 

B<Access:> private

B<Input:>

    $msg_str = The message to log.

    $log_lvl = optional, default = 1.  The value controls whether or not
               a message is logged.  A message is logged if $log_lvl is
               less than or equal to  $self->log_level().

B<Return value:> none

=cut

sub _logmsg {
    my $self = shift;
    my $msg_str = shift;
    my $log_lvl = shift;
    $log_lvl = 1 if (!defined $log_lvl);
    return if ($self->log_level() ==0);  # no module logging.
    if ($log_lvl <= $self->log_level()) {
        my $header_msg = "CifsSetupHelper.pm-log: ";
        $Log->comment($header_msg . $msg_str);
    }  
}


=head2 _read_next_addrs()

This method reads the lines of an array pointed to by
$ref_lines to look for all the addresses starting at line
number $i.  It stops reading further when a line wihtout
":" is found.

B<Access:> private

B<Input:>
 
          $ref_lines = a reference to an array containing the lines to read.
          $i = the index to the array where to start reading.

B<Return value:> A string containing a "," separated list of addresses.

=cut

sub _read_next_addrs {
    my ($ref_lines, $i) = @_;
    my $ret_str;
    while ($i<$#$ref_lines) {
        my $next_line = $$ref_lines[$i+1];
        if ($next_line =~ /:/) {
            last;  # no more address to read.
        }
        $next_line =~ s/^\s*//;   # remove leading blanks.
        $next_line =~ s/\s*$//;   # remove trailing blanks.
        if ($next_line !~ /^$/) { # skip blank lines
            $ret_str = $ret_str . ",$next_line";  # a "," separated list.
        }
        $i++;
    }
    $ret_str =~ s/^,//; # remove the leading ",".
    return $ret_str;
}



=head2 _send_cmd()

This method sends the given command to the filer connection 
and returns the output in an array.

B<Access:> private

B<Input:> 

          $conn = connection to filer.  
          $cmd = command to send.
          $echo = 1: return all the lines including the command itself.
                  0: do not return the command by removing it from the output.

B<Return value:> An array containing the output of the command.

=cut

sub _send_cmd {
    my ($conn, $cmd, $echo) =@_;
    $echo = 0 if (undef $echo);
    my $out;
    (undef, undef, $out) = say ($conn, $cmd);
    my @outs = split (/\n/, $out);
    for (my $i = 0; $i <= $#outs; $i++) {
        $outs[$i] =~ s/\r//g; # remove \r characters.
    }

    if ($echo ==0) {
        # remove the command from the first line if it is there.
        my $true_cmd = $cmd;
        $true_cmd =~ s/^\s*vfiler\s+run\s+-q\s+\S+\s+//i;
        my $first_line = $outs[0];
        chomp $first_line;
        if ($first_line =~ /\s*$true_cmd\s*$/) {
           # this is the echo of the command, remove it.
           shift (@outs);
        }
    }
    return @outs;
}

sub AUTOLOAD {
	my $self = shift;
	my $type = ref($self)
		    or croak "$self is not an object";

	my $name = $AUTOLOAD;
    return if $name =~ /::DESTROY$/;
	$name =~ s/.*://;   # strip fully-qualified portion

	unless (exists $self->{$name} ) {
	    croak "Can't access undefined method '$name' in class $type";
	}

	if (@_) {
	    $self->{$name} = shift;
        if ($name =~ /^vfiler$/) {
            # need to set the command prefix for vfiler:
            $self->{"VFCP"} = "vfiler run -q " . $self->{$name} . " ";
        }
	} 
    return $self->{$name};
} 


=head1 SEE ALSO

=head1 AUTHOR

David Low, davidlow@netapp.com

=head1 HISTORY

Initial version.

=head1 TODO

=cut

=head1 COPYRIGHT

Copyright (c) 2008 Network Appliance, Inc.

=cut

1;
