# $Id: //depot/prod/test/Rsierranevada/lib/CIFS/Client.pm#4 $
#
## @summary     Common library to handle Cifs client operations via rsh
##
## @description
##   This package defines routines which attempt to get and parse
##   filer output using the "say" command in Tharn.
##
##   All the subroutines here expect a connection object as their
##   argument.  If the connection is invalid (causing the say
##   command to fail) or if any other errors occur while attempting
##   to get the requested information, these subroutines will return 0.
##
## @dependencies
##   Some functions in this lib depend the exisitance of certain tools
##   located in the P4 tree. The required tools are:
##
##   tools/windows/pstools:
##   tools/windows/vbs:
##
## @bugs
## @author      bellizzi@netapp.com
## @modified    $Date: 2013/07/30 $

=head1 NAME
  
CIFS::Client - Library of client related functions
  
=head1 SYNOPSIS

 use Cifs::Client;
    
=head1 ROUTINES
    
=cut

package CIFS::Client;

use strict;
use Tharn;
$Log->comment("Starting " . __FILE__);
use POSIX qw(strftime);
use config_base qw(host_is_alive);
use Data::Dumper;
use Options;
use Params::Validate qw(
    validate_pos
    validate
    BOOLEAN
    SCALAR
    OBJECT
    ARRAYREF
    HASHREF
    UNDEF
);
use MapShare qw(
    list_maps
    map_cifs_share
    unmap_cifs_share
);
use Net::Ping::External qw(ping);
use Utils::DNS;
use NACL::MTask::PacketCapture;
use NACL::C::Client;
use SystemTest::Sleep;
use vars qw (@ISA @EXPORT $VERSION $DEBUG);

$VERSION = 1.00;
@ISA     = qw (Exporter);
@EXPORT  = qw (
    check_connectivity
    create_client_logpath
    get_ip_config
    get_drive_letters
    get_win32_pids
    get_win32_process_list
    ipv6_address_to_ipv6_literal
    ipv6_share_to_ipv6_literal
    ipv6_literal_to_ipv6_address
    kill_win32_pid
    kill_all_win32_pids
    mapDrive
    map_share
    unmap_share
    monitor_autoit_lock_file
    remove_mapped_drives
    prep_for_windows_subtest
    test_share
    client_diagnostics
    process_cifs_clients
    capture_start
    capture_stop
);

my $Options = new Options;
$Options->parameters(
    definitions => [
        {   name    => 'PSTOOLS_DIR',
            spec    => "s",
            default => 'v2.3',
            description =>
                'The pstools directory under tools/windows/pstools',
        },
        {   name         => 'PROCESS_TOOL',
            spec         => "s",
            default      => 'microsoft',
            valid_values => ['', 'pstools', 'vbs', 'microsoft'],
            description  => 'The tool to use when managing processes',
        },
        {   name        => 'LOGDIR_SHARE',
            spec        => "s",
            description => 'The CIFS share for win32 clients '
                . 'to reach the nate LOGDIR',
        },
        {   name        => 'LOGDIR_USER',
            spec        => "s",
            description => 'The username used when mapping to LOGDIR_SHARE',
        },
        {   name        => 'LOGDIR_PASSWD',
            spec        => "s",
            description => 'The passwd used when mapping to LOGDIR_SHARE',
        },
        {   name => 'LOGDIR_DRIVE',
            spec => "s",
            description =>
                'The drive letter used when mapping to LOGDIR_SHARE',
        },
        {   name        => 'TOP_LOGDIR_WINDOWS',
            spec        => "s",
            description => 'Import for NATE params',
        },
        {   name        => 'HOME_SHARE',
            spec        => "s",
            description => 'The CIFS share for win32 clients '
                . 'to reach the users home share',
        },
        {   name        => 'HOME_SHARE_USER',
            spec        => "s",
            description => 'The username used when mapping to HOME_SHARE',
        },
        {   name        => 'HOME_SHARE_PASSWD',
            spec        => "s",
            description => 'The passwd used when mapping to HOME_SHARE',
        },
    ]
);

=head2 get_win32_pids

Returns a hashref of pids

 @pids = get_win32_pids(conn => $conObj, process => cmd);

=over

=item conn 

(Required) Conn object of the client required to be passed.

=item p4_dir 

(Optional) The windows path to the P4 tree

=item process 

(Required) The process name (Can be regex)

=item pstools_dir 

(Optional) The pstools version dir in p4 (default: v2.3)

=item process_tool

(Optional) Defines which tool to use pstools or vbs or microsoft (default: microsoft)

=back

=cut

sub get_win32_pids {

    $Log->enter();

    $Options->process();
    my $options = $Options->get_hash();
    my (%opts) = validate(
        @_,
        {   conn    => {type => OBJECT, optional => 0},
            process => {type => SCALAR, optional => 0},
            p4_dir  => {
                type     => SCALAR,
                optional => 1,
                default  => $options->{'P4_DIR'}
            },
            pstools_dir => {
                type     => SCALAR,
                optional => 1,
                default  => $options->{'PSTOOLS_DIR'}
            },
            process_tool => {
                type     => SCALAR,
                optional => 1,
                default  => $options->{'PROCESS_TOOL'}
            },
            timeout => {
                type     => SCALAR,
                optional => 1,
                default  => 300
            },
            verbose => {
                type     => SCALAR,
                optional => 1,
                default  => 1
            }
        }
    );

    my $hostName = $opts{'conn'}->hostp()->hostname();
    my ($line, $output, $cmd, @pids, $error);

    #---------------------------------------
    # Create the command we need to execute
    #---------------------------------------
    if ($opts{'process_tool'} eq 'pstools') {
        $cmd
            = $opts{'p4_dir'}
            . '\tools\windows\pstools\\'
            . $opts{'pstools_dir'}
            . '\pslist';

       #----------------------------------------------------------------------
       # Remove extension in process name since pstools does not display them
       #----------------------------------------------------------------------
        $opts{'process'} =~ s/^(.+)\.(com|exe|dll)/$1/i;

    } elsif ($opts{'process_tool'} eq 'vbs') {
        $cmd
            = 'cscript /nologo '
            . $opts{'p4_dir'}
            . '\tools\windows\vbs\ps.vbs';

        #-------------------------------------------
        # Add exe extension since vbs displays them
        #-------------------------------------------
        if ($opts{'process'} !~ /\.(com|exe|dll)/) {
            $opts{'process'} .= '\.exe';
        }
    } elsif ($opts{'process_tool'} eq 'microsoft') {
        $cmd = 'tasklist';

        #-------------------------------------------
        # Add exe extension since vbs displays them
        #-------------------------------------------
        if ($opts{'process'} !~ /\.(com|exe|dll)/) {
            $opts{'process'} .= '\.exe';
        }
    }

    for (my $i = 1; $i <= 3; $i++) {

        $error = 0;
        $Log->comment("Getting the process ID(s) for $opts{'process'} "
                . "on $hostName - attempt #"
                . $i);

        #----------------------
        # Get the process list
        #----------------------
        eval {
            (undef, undef, $output)
                = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
                -noverbose);
        };

        my $error = $@;
        if ($error) {
            $Log->comment("$cmd\n $output\n $error");
            die "Error getting the process ID(s) for $opts{'process'}";
        } else {
            $Log->debug("$cmd\n $output");

        }

        #-----------------------
        # Check for tool errors
        #-----------------------
        if ($opts{'process_tool'} eq 'pstools'
            && (   $output !~ /PsList/
                || $output
                =~ /object not found|A device attached to the system is not functioning/
            )
            ) {
            $Log->comment("$cmd\n $output");
            $Log->comment("Error getting process list");
            $error = 1;
            snooze 5;
            next;
        } elsif (
            $opts{'process_tool'} eq 'vbs'
            && (   $output !~ /Name\s+PID/
                || $output =~ /CScript Error: Loading script/i)
            ) {
            $Log->comment("$cmd\n $output");
            $Log->comment("Error getting process list");
            $error = 1;
            snooze 5;
            next;
        } elsif ($opts{'process_tool'} eq 'microsoft'
            && $output !~ /Image Name/i) {
            $Log->comment("$cmd\n $output");
            $Log->comment("Error getting process list");
            $error = 1;
            snooze 5;
            next;
        }

        #----------------------------
        # Search the commands output
        #----------------------------
        @pids = ();
        foreach $line (split(/\n/, $output)) {

            chomp($line);
            $line =~ s/\s+$//g;

            #-----------------------------------------------------
            # Check the current line for the process name and pid
            #-----------------------------------------------------
            if ($line =~ /^$opts{'process'}\s+(\d+)\s+/) {
                push(@pids, $1);
            }

        }

        #------------------------------------------------------
        # We made it here without errors break out of the loop
        #------------------------------------------------------
        if (!$error) { last; }
    }

    $Log->exit();
    if ($error) {
        die "Error getting the process ID(s) for $opts{'process'}";
    } else {
        if ($opts{'verbose'} && @pids == 0) {
            $Log->comment($output);
        }
        return @pids;
    }
} ## end sub get_win32_pids

=head2 get_win32_process_list

Returns a hashref of process names and pids

 $pdata = get_win32_process_list(conn => $conObj);

=over

=item conn 

(Optional) Conn object of the client required to be passed.

=item p4_dir 

(Optional) The windows path to the P4 tree

=item pstools_dir 

(Optional) The pstools version dir in p4 (default: v2.3)

=item process_tool

(Optional) Defines which tool to use pstools or vbs or microsoft (default: microsoft)

=back

=cut

sub get_win32_process_list {

    $Log->enter();

    $Options->process();
    my $options = $Options->get_hash();
    my (%opts) = validate(
        @_,
        {   conn   => {type => OBJECT, optional => 0},
            p4_dir => {
                type     => SCALAR,
                optional => 1,
                default  => $options->{'P4_DIR'}
            },
            pstools_dir => {
                type     => SCALAR,
                optional => 1,
                default  => $options->{'PSTOOLS_DIR'}
            },
            process_tool => {
                type     => SCALAR,
                optional => 1,
                default  => $options->{'PROCESS_TOOL'}
            },
            timeout => {
                type     => SCALAR,
                optional => 1,
                default  => 300
            }
        }
    );

    my $hostName = $opts{'conn'}->hostp()->hostname();
    my ($line, $output, $cmd, %pdata);

    if ($opts{'process_tool'} eq 'pstools') {
        $cmd
            = $opts{'p4_dir'}
            . '\tools\windows\pstools\\'
            . $opts{'pstools_dir'}
            . '\pslist';

       #----------------------------------------------------------------------
       # Remove extension in process name since pstools does not display them
       #----------------------------------------------------------------------
        $opts{'process'} =~ s/^(.+)\.(com|exe|dll)/$1/i;

    } elsif ($opts{'process_tool'} eq 'vbs') {
        $cmd
            = 'cscript /nologo '
            . $opts{'p4_dir'}
            . '\tools\windows\vbs\ps.vbs';

        #-------------------------------------------
        # Add exe extension since vbs displays them
        #-------------------------------------------
        if ($opts{'process'} !~ /\.(com|exe|dll)/) {
            $opts{'process'} .= '\.exe';
        }
    } elsif ($opts{'process_tool'} eq 'microsoft') {
        $cmd = 'tasklist';

        #-------------------------------------------
        # Add exe extension since vbs displays them
        #-------------------------------------------
        if ($opts{'process'} !~ /\.(com|exe|dll)/) {
            $opts{'process'} .= '\.exe';
        }
    }

    $Log->comment("Getting the process list on $hostName");

    # Get the process list
    eval {
        (undef, undef, $output)
            = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
            -noverbose);
    };

    my $error = $@;
    if ($error) {
        $Log->comment("$cmd\n $output\n $error");
        die "Error getting the process list on $hostName";
    } else {
        $Log->debug("$cmd\n $output");
    }

    #------------------------------------------------------------------------
    # Check for expected tool output. If these strings are missing we had an
    # error
    #------------------------------------------------------------------------
    if ($opts{'process_tool'} eq 'pstools'
        && (   $output !~ /PsList/
            || $output
            =~ /object not found|A device attached to the system is not functioning/
        )
        ) {
        $Log->comment("$cmd\n $output");
        die "Error getting process list";
    } elsif (
        $opts{'process_tool'} eq 'vbs'
        && (   $output !~ /Name\s+PID/
            || $output =~ /CScript Error: Loading script/i)
        ) {
        $Log->comment("$cmd\n $output");
        die "Error getting process list";
    } elsif ($opts{'process_tool'} eq 'microsoft' && $output !~ /Image Name/i)
    {
        $Log->comment("$cmd\n $output");
        die "Error getting process list";
    }

    #----------------------------
    # Search the commands output
    #----------------------------
    foreach $line (split(/\n/, $output)) {

        chomp($line);
        $line =~ s/\s+$//g;

        #--------------------
        # Ignore header info
        #--------------------
        if ($opts{'process_tool'} eq 'pstools') {
            if ($line
                =~ /(^PsList|^Copyright|^Sysinternals|^Process\s+information|Name\s+Pid|Image Name|^=)/
                ) {
                next;
            }
        }

        #-----------------------------------------------------
        # Check the current line for the process name and pid
        #-----------------------------------------------------
        if ($line =~ /^(\S+)\s+(\d+)\s+/) {
            my $name = $1;
            my $pid  = $2;

            $pdata{'pids'}{$pid}   = $name;
            $pdata{'names'}{$name} = $pid;
            push(@{$pdata{'list'}}, $line);
            push(@{$pdata{$name}},  $pid);
        }

    }

    $Log->exit();
    return \%pdata;
} ## end sub get_win32_process_list

=head2 kill_win32_pid

Kills a process id

 kill_win32_pid(conn => $conObj, pid => 12345);

=over

=item conn

(Required) Conn object of the client required to be passed.

=item p4_dir

(Optional) The windows path to the P4 tree (default x:\\agnostic)

=item pid

(Required) The process id you want to kill

=item pstools_dir 

(Optional) The pstools version dir in p4 (default: v2.3)

=item process_tool

(Optional) Defines which tool to use pstools or vbs or microsoft (default: microsoft)

=back

=cut

sub kill_win32_pid {

    $Log->enter();

    $Options->process();
    my $options = $Options->get_hash();
    my (%opts) = validate(
        @_,
        {   conn   => {type => OBJECT, optional => 0},
            pid    => {type => SCALAR, optional => 0},
            p4_dir => {
                type     => SCALAR,
                optional => 1,
                default  => $options->{'P4_DIR'}
            },
            pstools_dir => {
                type     => SCALAR,
                optional => 1,
                default  => $options->{'PSTOOLS_DIR'}
            },
            process_tool => {
                type     => SCALAR,
                optional => 1,
                default  => $options->{'PROCESS_TOOL'}
            },
            timeout => {
                type     => SCALAR,
                optional => 1,
                default  => 300
            }
        }
    );

    my $hostName = $opts{'conn'}->hostp()->hostname();
    my ($cmd, $ret, $err, $output);

    $Log->comment("Killing process $opts{'pid'} on $hostName");

    if ($opts{'process_tool'} eq 'pstools') {
        $cmd
            = $opts{'p4_dir'}
            . '\tools\windows\pstools\\'
            . $opts{'pstools_dir'}
            . '\pskill -t '
            . $opts{'pid'};
    } elsif ($opts{'process_tool'} eq 'vbs') {
        $cmd
            = 'cscript /nologo '
            . $opts{'p4_dir'}
            . '\tools\windows\vbs\kill.vbs localhost '
            . $opts{'pid'};
    } elsif ($opts{'process_tool'} eq 'microsoft') {
        $cmd = "tskill $opts{'pid'} /a";
    }

    eval {
        (undef, undef, $output)
            = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
            -noverbose);
    };

    my $error = $@;
    if ($error) {
        $Log->comment("$cmd\n $output\n $error");
        die "Error killing process";
    } else {
        $Log->debug("$cmd\n $output");
    }

    $Log->exit();

    snooze 5;
} ## end sub kill_win32_pid

=head2 kill_all_win32_pids

Kills the specified pids dead

 kill_all_win32_pids(conn          => $conObj, 
                     process       => cmd, 
		     max_attempts  => 25, 
		     wait          => 60)

=over

=item conn

(Required) Conn object of the client required to be passed.

=item p4_dir

(Optional) The windows path to the P4 tree

=item pid

(Required) The process name you want to kill

=item max_attempts

(Required) The max number of attempts to try and kill

=item wait

(Required) The time to wait before attempting to kill hte process again

=item pstools_dir 

(Optional) The pstools version dir in p4 (default: v2.3)

=item process_tool

(Optional) Defines which tool to use pstools or vbs or microsoft (default: microsoft)

=back

=cut

sub kill_all_win32_pids {

    $Log->enter();

    $Options->process();
    my $options = $Options->get_hash();
    my (%opts) = validate(
        @_,
        {   conn         => {type => OBJECT, optional => 0},
            process      => {type => SCALAR, optional => 0},
            max_attempts => {type => SCALAR, optional => 0},
            p4_dir       => {
                type     => SCALAR,
                optional => 1,
                default  => $options->{'P4_DIR'}
            },
            pstools_dir => {
                type     => SCALAR,
                optional => 1,
                default  => $options->{'PSTOOLS_DIR'}
            },
            process_tool => {
                type     => SCALAR,
                optional => 1,
                default  => $options->{'PROCESS_TOOL'}
            },
            'wait'  => {type => SCALAR, optional => 1, default => 1},
            timeout => {
                type     => SCALAR,
                optional => 1,
                default  => 300
            }
        }
    );

    my $hostName = $opts{'conn'}->hostp()->hostname();

    my $attempts = 0;
    $Log->comment(
        "Killing all processes matching $opts{'process'} on $hostName");
    my @pids = get_win32_pids(
        'conn'    => $opts{'conn'},
        'timeout' => $opts{'timeout'},
        'process' => $opts{'process'},
        'verbose' => 0
    );

    $Log->debug("Found pids: " . Dumper(\@pids));
    my %allDead;
    my @p;
    foreach my $pid (@pids) {
        $allDead{$pid} = 0;
        if ($pid =~ /\d+/ && $pid == 0) {
            next;
        } elsif ($pid =~ /\d+/ && $pid != 0) {
            kill_win32_pid(
                'conn'    => $opts{'conn'},
                'timeout' => $opts{'timeout'},
                'pid'     => $pid
            );
            while ($allDead{$pid} != 1 && $attempts < $opts{'max_attempts'}) {
                @p = get_win32_pids(
                    'conn'    => $opts{'conn'},
                    'timeout' => $opts{'timeout'},
                    'process' => $opts{'process'},
                    'verbose' => 0
                );
                if (grep(/$pid/, @p)) {
                    kill_win32_pid(
                        'conn'    => $opts{'conn'},
                        'timeout' => $opts{'timeout'},
                        'pid'     => $pid
                    );
                } else {
                    $allDead{$pid} = 1;
                }
                $attempts++;
                if ($opts{'wait'}) {
                    snooze $opts{'wait'};
                } else {
                    snooze 5;
                }
            }
        } else {
            $allDead{$pid} = 1;
        }
    }

    if (grep(/0/, keys(%allDead)) == 0
        && $attempts == $opts{'max_attempts'}) {
        $Log->comment(Dumper(\@p));
        die "Could not kill all $opts{'process'} processes after "
            . $opts{'max_attempts'}
            . " attemtps\n";
    }

    $Log->exit();

} ## end sub kill_all_win32_pids

=head2 check_connectivity

Verifies network connectivity and rsh functionality to a host

 check_connectivity(conn => $conObj);

=over

=item conn

(Required) Conn object of the client required to be passed.

=back

=cut

sub check_connectivity {

    $Log->enter();

    my (%opts) = validate(
        @_,
        {   conn    => {type => OBJECT, optional => 0},
            timeout => {
                type     => SCALAR,
                optional => 1,
                default  => 60
            }
        }
    );

    my $hostName = $opts{'conn'}->hostp()->hostname();
    my $hostType = $opts{'conn'}->hostp()->hosttype();
    my $hostIp   = $opts{'conn'}->hostp()->default_ip();
    my ($pingStatus, $rshStatus, $output);

    for (my $i = 1; $i <= 3; $i++) {
        $pingStatus = 0;
        $rshStatus  = 0;
        $Log->comment(
            "Testing connectivity to $hostName($hostIp) attempt #" . $i);

        #---------------
        # Ping the host
        #---------------
        $Log->debug("Verifying IP access to $hostName($hostIp) using ICMP");

        if (!ping('host' => $hostIp)) {
            $Log->comment("Error pinging $hostName($hostIp)");
            snooze 10;
            next;
        } else {
            $pingStatus = 1;
        }

        #------------------------------
        # Check if rsh/staf is working
        #------------------------------
        $Log->debug("Verifying OS on $hostName($hostIp) is $hostType");
        my $cmd = 'cmd.exe /c ver';

        if ($hostType =~ /windows/i) {
            eval {
                (undef, undef, $output)
                    = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
                    -noverbose);
            };

            my $error = $@;
            if ($error || $output !~ /Microsoft Windows/) {
                $Log->comment("$cmd\n $output\n $error");
                $Log->comment("Error accessing $hostName($hostIp) via say");
                next;
            } else {
                $Log->debug("$cmd\n $output");
                $rshStatus = 1;
            }
        } elsif ($hostType =~ /linux|sun|solaris/i) {
            eval {
                (undef, undef, $output)
                    = say($opts{'conn'}, 'uname', -timeout, $opts{'timeout'},
                    -noverbose);
            };

            my $error = $@;
            if ($error || $output !~ /linux|sun|solaris/i) {
                $Log->comment("$cmd\n $output\n $error");
                $Log->comment("Error accessing $hostName($hostIp) via say");
                next;
            } else {
                $Log->debug("$cmd\n $output");
                $rshStatus = 1;
            }
        }

        if ($rshStatus) { last; }

    }

    $Log->exit();

    if (!$pingStatus || !$rshStatus) {
        snooze 10;
        die "Error verifying connectivity to $hostName($hostIp)";
    }
} ## end sub check_connectivity

=head2 get_drive_letters

Returns a array of drive letters thats are currently mapped

 @drives = get_drive_letters(conn => $conObj);

=over

=item conn

(Required) Conn object of the client required to be passed.

=back

=cut

sub get_drive_letters {

    $Log->enter();

    my (%opts) = validate(
        @_,
        {   conn    => {type => OBJECT, optional => 0},
            timeout => {type => OBJECT, optional => 1, default => 300}
        }
    );

    my $hostName = $opts{'conn'}->hostp()->hostname();
    my $hostType = $opts{'conn'}->hostp()->hosttype();
    my $hostIp   = $opts{'conn'}->hostp()->default_ip();

    $Log->comment("Getting mapped drives on $hostName($hostIp)");
    my $cmd = 'cmd.exe /c net use';
    my (undef, undef, $output)
        = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'});

    if ($output !~ /The\s+command\s+completed\s+successfully/) {
        $Log->comment("$cmd\n $output");
        die "Error getting drive listing\n";
    } else {
        $Log->debug("$cmd\n $output");
    }

    my @drives;
    foreach my $line (split(/\n/, $output)) {
        chomp;
        if ($line =~ /\s+(\w):\s+/) { push(@drives, $1); }
    }
    $Log->exit();

    return @drives;
} ## end sub get_drive_letters

=head2 remove_mapped_drives

Removes one or more drive mappings

 remove_mapped_drives(conn   => $conObj, 
                      remove => *, 
		      ignore => d);

=over

=item conn

(Required) Conn object of the client required to be passed.

=item remove

(Required) The drive letters you want to remove

=item ignore

(Optional) The drive letters you do not remove when remove = *

=back

=cut

sub remove_mapped_drives {

    $Log->enter();

    my (%opts) = validate(
        @_,
        {   conn    => {type => OBJECT, optional => 0},
            remove  => {type => SCALAR, optional => 0},
            ignore  => {type => SCALAR, optional => 1},
            timeout => {type => OBJECT, optional => 1, default => 300}
        }
    );

    my $hostName = $opts{'conn'}->hostp()->hostname();
    my $hostType = $opts{'conn'}->hostp()->hosttype();
    my $hostIp   = $opts{'conn'}->hostp()->default_ip();

    my $msg = "Removing mapped drive(s) $opts{'remove'} ";
    if ($opts{'ignore'}) { $msg .= "(except $opts{'ignore'}) "; }
    $msg .= "on $hostName($hostIp)";
    $Log->comment($msg);

    my @driveLetters = get_drive_letters('conn' => $opts{'conn'});
    foreach my $drive (@driveLetters) {
        if ($opts{'remove'} ne '*' && $opts{'remove'} !~ /$drive/i) {
            $Log->debug(
                "Skipping $drive because $opts{'remove'} !~ /$drive/i");
            next;
        }
        if ($opts{'ignore'} ne '' && $opts{'ignore'} =~ /$drive/i) {
            $Log->debug(
                "Skipping $drive because $opts{'ignore'} =~ /$drive/i");
            next;
        }
        $Log->comment("Removing mapped drive $drive on $hostName($hostIp)");
        my $output;
        my $cmd = "cmd.exe /c net use $drive: /d /y";
        eval {
            (undef, undef, $output)
                = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'}, -noverbose);
        };

        my $error = $@;
        if ($error || $output !~ /$drive:\s+was\s+deleted\s+successfully/i) {
            $Log->comment("$cmd\n $output\n $error");
            die "Unable to remove $drive";
        } else {
            $Log->debug("$cmd\n $output");
        }
    }

    $Log->exit();

} ## end sub remove_mapped_drives

=head2 monitor_autoit_lock_file

Monitors a lock file created by GUI automation scripts on a windows client 

 monitor_autoit_lock_file(conn   => $conObj
                          checks => 25,
			  wait   => 5);

=over

=item conn

(Required) Conn object of the client required to be passed.

=item checks

(Required) The number of times to check for the existanc of the lock file

=item wait

(Required) The time to wait between checks

=back

=cut

sub monitor_autoit_lock_file {

    $Log->enter();

    my (%opts) = validate(
        @_,
        {   conn    => {type => OBJECT, optional => 0},
            checks  => {type => SCALAR, optional => 0},
            wait    => {type => SCALAR, optional => 0},
            timeout => {type => OBJECT, optional => 1, default => 300}
        }
    );

    $Log->comment("Waiting for AutoIt lock file");
    my $c     = 0;
    my $found = 0;
    my $cmd   = 'cmd.exe /c dir c:\autoit.lck';
    my $output;

    while ($found == 0 && $c <= 300) {
        eval {
            (undef, undef, $output)
                = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
                -noverbose);
        };

        $Log->debug("$cmd\n $output");
        foreach my $line (split(/\n/, $output)) {
            if ($line =~ /dir/) {
                next;
            } elsif ($line =~ /autoit\.lck/) {
                $found = 1;
            }
        }
        if ($found == 0) {
            snooze 1;
            $c++;
        }
    }

    if ($found == 0) {
        $Log->comment("$cmd\n $output");
        die "The autoit.lck file was not created "
            . "during the wait time of 300 seconds\n";
    } else {
        $Log->debug("$cmd\n $output");
    }

    $Log->comment("Monitoring AutoIt lock file");
    for ($c = 1; $c <= $opts{'checks'}; $c++) {
        eval {
            (undef, undef, $output)
                = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
                -noverbose);
        };

        $Log->debug("$cmd\n $output");
        foreach my $line (split(/\n/, $output)) {
            if ($line =~ /dir/) {
                next;
            } elsif ($line =~ /File\s+Not\s+Found/) {
                $Log->comment("AutoIt lock file cleared");
                return;
            } elsif ($line =~ /autoit\.lck/) {
                snooze $opts{'wait'};
            }
        }
    }
    $Log->exit();

    die "The autoit.lck file was not cleared during "
        . "the wait time of "
        . $opts{'wait'} * $opts{'checks'}
        . " seconds";
} ## end sub monitor_autoit_lock_file

=head2 mapDrive

Map a drive letter to a unc path and return the drive letter

This function is deprecated please use Cifs::Client::map_share

=cut

sub mapDrive {

    my %opts = @_;
    my ($line, $drive, $output, $cmd, $hostObj, $hostIp, $hostType,
        $hostName);

    $hostObj  = $opts{'conn'}->hostp();
    $hostName = $hostObj->hostname();
    $hostType = $hostObj->hosttype();
    $hostIp   = $hostObj->default_ip();

    logresult("WARN",
        "This function is deprecated. Consider using Cifs::Client::map_share"
    );
    $Log->comment(
        "Mapping $opts{'drive'} to \\\\$opts{'server'}\\$opts{'share'} on $hostName($hostIp)"
    );

    $cmd
        = "net use $opts{'drive'} \\\\$opts{'server'}\\$opts{'share'} /u:$opts{'cifs_user'} $opts{'cifs_passwd'}";

    (undef, undef, $output)
        = say($opts{'conn'}, $cmd, -timeout, 300, -eval_error, -noverbose);
    $Log->debug("$cmd\n $output");
    if ($output !~ /command\s+completed\s+successfully/) {
        return (0, undef);
    }

    foreach my $line (split(/\n/, $output)) {
        chomp;

        # Drive W: is now connected to \\SS-3\kc33b13-e0.
        if (   $opts{'drive'} =~ /\*/
            && $line =~ /Drive\s+(\w):\s+is\s+now\s+connected/) {
            return (1, $1);
        } elsif ($opts{'drive'} !~ /\*/
            && $line =~ /command\s+completed\s+successfully/) {
            return (1, $opts{'drive'});
        }
    }
    return (0, undef);
} ## end sub mapDrive

=head2 get_ip_config

Gets the ipconfig information for a windoze box and retruns a hashref in the form of 

 ipInfo: $VAR1 = {
          Local Area Connection => { 
                                       dns_servers => 172.16.0.6,
                                       gateway => ,
                                       ip => 172.16.12.131,
                                       netbios => ,
                                       primary_wins => 172.16.0.21,
                                       state => ,
                                       dns_suffix => nas.ssqa.rtp.netapp.com,
                                       mac => 00:14:5E:3C:99:14,
                                       mask => 255.254.0.0,
                                       secondary_wins => 
                                     },
          Local Area Connection 2 => { 
                                         dns_servers => 172.16.0.6,
                                         gateway => 10.60.89.1,
                                         ip => 10.60.89.160,
                                         netbios => Disabled,
                                         primary_wins => ,
                                         state => ,
                                         dns_suffix => rtp.netapp.com,
                                         mac => 00:14:5E:3C:99:15,
                                         mask => 255.255.255.0,
                                         secondary_wins => 
                                       }
        };


 $ipinfo = get_ip_config(conn => $conObj);

=over

=item conn

(Required) Conn object of the client required to be passed.

=back

=cut

sub get_ip_config {

    $Log->enter();

    my (%opts) = validate(
        @_,
        {   conn    => {type => OBJECT, optional => 0},
            timeout => {type => OBJECT, optional => 1, default => 300}
        }
    );

    my $hostName = $opts{'conn'}->hostp()->hostname();

    $Log->comment("Getting IP configuration for $hostName");

    my ($output, %ipInfo, $interface);
    my $cmd = 'cmd.exe /c ipconfig /all';
    eval {
        (undef, undef, $output)
            = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
            -noverbose);
    };

    $Log->debug("$cmd\n $output");
    $ipInfo{'hostname'}           = '';
    $ipInfo{'primary_dns_suffix'} = '';

    foreach my $line (split(/\n/, $output)) {
        chomp($line);
        $line =~ s/\s+$//g;
        if ($line =~ /.+Host Name.+:\s+(.*)$/) {
            $ipInfo{'hostname'} = $1;
        } elsif ($line =~ /.+Primary Dns Suffix.+:\s+(.*)$/) {
            $ipInfo{'primary_dns_suffix'} = $1;
        } elsif ($line =~ /\s+adapter\s+(.+):/) {
            $interface                            = $1;
            $ipInfo{$interface}{'state'}          = '';
            $ipInfo{$interface}{'mask'}           = '';
            $ipInfo{$interface}{'gateway'}        = '';
            $ipInfo{$interface}{'dns_suffix'}     = '';
            $ipInfo{$interface}{'secondary_wins'} = '';
            $ipInfo{$interface}{'primary_wins'}   = '';
            $ipInfo{$interface}{'dns_servers'}    = '';
            $ipInfo{$interface}{'netbios'}        = '';
            push(@{$ipInfo{'adapters'}}, $interface);
        }

        if ($interface) {
            if ($line =~ /.+Media State.+:\s+(Media disconnected)/) {
                $ipInfo{$interface}{'state'} = $1;
            } elsif ($line =~ /.+Physical Address.+:\s+(.*)$/) {
                my $mac = $1;
                $mac =~ s/-/:/g;
                $ipInfo{$interface}{'mac'} = $mac;
            } elsif ($line =~ /.+Address.+:\s+(.*)$/) {

 # Vista/2k8
 # Link-local IPv6 Address . . . . . : fe80::3095:394d:d251:6ca1%12(Preferred)
 # IPv4 Address. . . . . . . . . . . : 10.60.76.96(Preferred)
 # IPv6 Address. . . . . . . . . . . : fd20:81be:b255:8100:0:10:0:2(Preferred)
 # Link-local IPv6 Address . . . . . : fe80::3181:f361:3d14:85bd%11(Preferred)
 # IPv4 Address. . . . . . . . . . . : 172.16.13.252(Preferred)
 # 2003
 # IP Address. . . . . . . . . . . . : 10.60.92.59
 # IP Address. . . . . . . . . . . . : fe80::214:5eff:fe3c:b8db%4
 # IP Address. . . . . . . . . . . . : 172.16.14.194
 # IP Address. . . . . . . . . . . . : fd20:81be:b255:8100:0:10:0:29
 # IP Address. . . . . . . . . . . . : fe80::214:5eff:fe3c:b8da%5
 # IP Address. . . . . . . . . . . . : fe80::ffff:ffff:fffd%6
 # IP Address. . . . . . . . . . . . : fe80::5efe:172.16.14.194%2
 # IP Address. . . . . . . . . . . . : fe80::5efe:10.60.92.59%2
 # XP
 # IP Address. . . . . . . . . . . . : 10.60.89.160
 # IP Address. . . . . . . . . . . . : fe80::214:5eff:fe3c:9915%4
 # IP Address. . . . . . . . . . . . : 172.16.12.131
 # IP Address. . . . . . . . . . . . : fd20:81be:b255:8100:0:10:0:34
 # IP Address. . . . . . . . . . . . : fe80::214:5eff:fe3c:9914%5
 # IP Address. . . . . . . . . . . . : fe80::ffff:ffff:fffd%6
 # IP Address. . . . . . . . . . . . : fe80::5efe:172.16.12.131%2
 # IP Address. . . . . . . . . . . . : fe80::5efe:10.60.89.160%2
 #

                my $addr = $1;
                if ($addr =~ /(\d+\.\d+\.\d+\.\d+)/) {
                    push(@{$ipInfo{$interface}{'ipv4'}}, $1);
                } elsif ($addr
                    =~ /([0-9a-f]+:[0-9a-f]+:[0-9a-f]+:[0-9a-f]+:[0-9a-f]+:[0-9a-f]+:[0-9a-f]+:[0-9a-f]+)/i
                    ) {
                    push(@{$ipInfo{$interface}{'ipv6'}}, $1);
                } elsif ($addr
                    =~ /([0-9a-f]+::[0-9a-f]+:[0-9a-f]+:[0-9a-f]+:[0-9a-f]+%[0-9a-f]+)/
                    ) {
                    $ipInfo{$interface}{'link_local_ipv6'} = $1;
                }
            } elsif ($line =~ /.+Subnet Mask.+:\s+(.*)$/) {
                $ipInfo{$interface}{'mask'} = $1;
            } elsif ($line =~ /.+Default Gateway.+:\s+(.*)$/) {
                $ipInfo{$interface}{'gateway'} = $1;
            } elsif ($line =~ /.+DNS Servers.+:\s+(.*)$/) {
                $ipInfo{$interface}{'dns_servers'} = $1;
            } elsif ($line =~ /.+Primary\s+WINS\s+Server.+:\s+(.*)$/) {
                $ipInfo{$interface}{'primary_wins'} = $1;
            } elsif ($line =~ /.+Secondary\s+WINS\s+Server.+:\s+(.*)$/) {
                $ipInfo{$interface}{'secondary_wins'} = $1;
            } elsif (
                $line =~ /.+Connection-specific\s+DNS\s+Suffix.+:\s+(.*)$/) {
                $ipInfo{$interface}{'dns_suffix'} = $1;
            } elsif ($line =~ /.+NetBIOS\s+over\s+Tcpip.+:\s+(.*)$/) {
                $ipInfo{$interface}{'netbios'} = $1;
            }
        }
    }
    $Log->exit();

    return \%ipInfo;
} ## end sub get_ip_config

=head2 map_share

Map a unc path and return the drive letter if one was specified. If the mapping fails 
perform client diagnostics and retry.

 $drive = map_share(conn        => $conObj, 
                    share       => share, 
		    cifs_user   => user, 
		    cifs_passwd => password, 
		    drive       => w|*|)

=over

=item conn

(Required) Conn object of the client required to be passed.

=item share

(Required) The UNC path you want to map to

=item drive

(Required) The drive letter you want to use. Can be blank or *.

=item cifs_user

(Required) The user credentials

=item cifs_user

(Required) The password for the user credentials

=item timeout

(Required) The timeout for used commands

=item reuse

(Optional) Reuse existing drive mappings

=item diags

(Optional) Perform client side diags if the mapping fails

=item tests

(Optional) The share tests to be performed r|w|rw

=back

=cut

sub map_share {

    $Log->enter();

    my (%opts) = validate(
        @_,
        {   conn     => {type => OBJECT, optional => 0},
            share    => {type => SCALAR, optional => 0},
            username => {type => SCALAR, optional => 1},
            password => {type => SCALAR, optional => 1},
            drive    => {type => SCALAR, optional => 1, default => ''},
            timeout  => {type => SCALAR, optional => 0, default => '300'},
            reuse    => {type => SCALAR, optional => 1, default => 1},
            diags    => {type => SCALAR, optional => 1, default => 1},
            tests    => {type => SCALAR, optional => 1, default => 'rw'},
            delay    => {type => SCALAR, optional => 1, default => 60},
            verbose  => {type => SCALAR, optional => 1, default => 0},
            attempts => {type => SCALAR, optional => 0, default => 3},
            trace => {type => SCALAR | UNDEF, optional => 1, default => 1}
        }
    );

    my $hostname = $opts{'conn'}->hostp()->hostname();

    $Log->comment("Parsing share path $opts{'share'}");
    $opts{'share'} =~ /^[\\]+([^\\]+)[\\]+([^\\]+)/;
    my $server = $1;
    my $share  = $2;
    $Log->comment("Server:$server Share:$share");
    my $address;
    if ($server) {
        $address = ipv6_literal_to_ipv6_address('name' => $server);
        if ($address !~ /(^\d+\.\d+\.\d+\.\d+$|:)/) {
            $Log->comment("Resolving $address");
            my $DNS = get_ip_by_name('name' => $address);

            if ($DNS->{'ERROR'} && $opts{'trace'}) {

                die "Unable to get IP address from server name $address";
            }
            if ($DNS->{'INET4'}) {
                $address = $DNS->{'INET4'};
            } elsif ($DNS->{'INET6'}) {
                $address = $DNS->{'INET6'};
            }
        }
    } else {
        die "Unable to determine server from share path: $opts{'share'}";
    }

    #-----------------------------------------------------
    # Translate IPv6 addresses to ipv6-literal.net format
    #-----------------------------------------------------
    $opts{'share'} = ipv6_share_to_ipv6_literal('share' => $opts{'share'});

    my $escapedPath = $opts{'share'};
    $escapedPath =~ s/(\\|\.)/\\$1/g;

    #----------------------------------
    # Make sure we can ping the server
    #----------------------------------
    my ($cmd, $output, $msg);
    if ($address) {
        for (my $i = 1; $i <= $opts{'attempts'}; $i++) {
            my $failed = 0;
            $Log->comment("Pinging $address using ICMP attempt #" . $i);
            $cmd = "ping $address";
            eval {
                (undef, undef, $output)
                    = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
                    -noverbose);
            };

            my $error = $@;
            if ($error || $output !~ /Reply from/) {
                $Log->comment("$cmd\n $output\n $error");
                $msg    = "Ping failed to $address\n";
                $failed = 1;
                $Log->warn($msg);
            } else {
                $Log->debug("$cmd\n $output");
            }
            if ($i == $opts{'attempts'} && $failed) {
                client_diagnostics('conn' => $opts{'conn'});
                die $msg;
            } elsif (!$failed) {
                last;
            }
        }
    }

    #--------------------------------------------------------------------
    # List the current mappings. Check if the existing mapping is in the
    # unavailable state or has the incorrect drive letter assigned. If
    # so remove the mapping. If everything checks out and reuse is set
    # reuse the drive mapping.
    #--------------------------------------------------------------------
    $msg = "Searching existing mappings for ";
    if ($opts{'drive'} =~ /\w+/) {
        $msg .= "$opts{'drive'} mapped to $opts{'share'}";
    } elsif ($opts{'drive'} =~ /\*/) {
        $msg .= "any drive mapped to $opts{'share'}";
    } else {
        $msg .= "any mapping to $opts{'share'}";
    }
    $Log->comment($msg);

    my @maps = list_maps(
        'client'  => $opts{'conn'},
        'timeout' => $opts{'timeout'},
        'verbose' => $opts{'verbose'}
    );

    foreach my $map (@maps) {
        $Log->debug("\n"
                . "Share  . . . . : ${$map}{'path'}\n "
                . "Requested Share:$opts{'share'} \n"
                . "Drive  . . . . : ${$map}{'drive'}\n "
                . "Requested Drive:$opts{'drive'}");
        if (   $opts{'drive'} =~ /\w/
            && ${$map}{'drive'} =~ /$opts{'drive'}/i
            && ${$map}{'path'} !~ /^$escapedPath$/i) {
            $Log->comment("Existing mapping of ${$map}{'path'} is using the "
                    . "specified drive letter ($opts{'drive'}), removing");
            unmap_cifs_share(
                'client'  => $opts{'conn'},
                'drive'   => ${$map}{'drive'},
                'force'   => 1,
                'verbose' => $opts{'verbose'}
            );
        } elsif (${$map}{'path'} =~ /^$escapedPath$/i) {
            if (${$map}{'status'} =~ /Unavailable/) {
                $Log->comment(
                    "Existing mapping of $opts{'path'} is unavailable, removing"
                );
                unmap_cifs_share(
                    'client'  => $opts{'conn'},
                    'mappath' => ${$map}{'path'},
                    'force'   => 1,
                    'verbose' => $opts{'verbose'}
                );
            } elsif ($opts{'drive'} =~ /\w/
                && ${$map}{'drive'} !~ /$opts{'drive'}/i) {
                $Log->comment(
                    "Existing mapping of $opts{'path'} does not have the "
                        . "correct drive letter ($opts{'drive'}), removing");
                unmap_cifs_share(
                    'client'  => $opts{'conn'},
                    'drive'   => ${$map}{'drive'},
                    'force'   => 1,
                    'verbose' => $opts{'verbose'}
                );
            } elsif ($opts{'reuse'}
                && $opts{'drive'}   =~ /\*/
                && ${$map}{'drive'} =~ /\w/i) {
                $Log->comment(
                    "Reusing existing mapping ${$map}{'drive'} ${$map}{'path'}"
                );

                test_share(
                    'conn'  => $opts{'conn'},
                    'share' => $opts{'share'},
                    'tests' => $opts{'tests'},
                    'trace' => $opts{'trace'}
                );
                $Log->exit();
                return ${$map}{'drive'};
            } elsif ($opts{'reuse'} && $opts{'drive'} !~ /\*/) {
                $Log->comment(
                    "Reusing existing driveless mapping ${$map}{'path'}");

                test_share(
                    'conn'  => $opts{'conn'},
                    'share' => $opts{'share'},
                    'tests' => $opts{'tests'},
                    'trace' => $opts{'trace'}
                );
                $Log->exit();
                return ${$map}{'drive'};
            } elsif ($opts{'reuse'}
                && $opts{'drive'} =~ /\*/
                && !${$map}{'drive'}) {
                $Log->comment(
                    "Forcing creation of new drive mapping to ${$map}{'path'}"
                );

                last;
            } elsif (!$opts{'reuse'}
                && ${$map}{'drive'}
                && $opts{'drive'} =~ /\w/) {
                $Log->comment(
                    "Removing existing mapping ${$map}{'drive'} ${$map}{'path'}"
                );
                unmap_cifs_share(
                    'client'  => $opts{'conn'},
                    'drive'   => ${$map}{'drive'},
                    'force'   => 1,
                    'verbose' => $opts{'verbose'}
                );
            } elsif (!$opts{'reuse'} && !${$map}{'drive'}) {
                $Log->comment(
                    "Removing existing driveless mapping ${$map}{'path'}");
                unmap_cifs_share(
                    'client'  => $opts{'conn'},
                    'mappath' => ${$map}{'path'},
                    'force'   => 1,
                    'verbose' => $opts{'verbose'}
                );
            }
        }
    }

    #---------------------------------------------------------------------
    # If no mapping was found or one was found but it was removed map the
    # share
    #---------------------------------------------------------------------
    my $Capture;
    for (my $i = 1; $i <= $opts{'attempts'}; $i++) {
        my $msg = "Mapping to $opts{'share'} ";
        if ($opts{'username'}) {
            $msg .= "as $opts{'username'} ";
        }
        $msg .= "attempt #" . $i;
        $Log->comment($msg);
        if ($i == 2) {
            $opts{'verbose'} = 1;
            if ($opts{'trace'}) {
                $Capture = capture_start($opts{'conn'}, $address);
            }
        }

        my %args = (
            'client'  => $opts{'conn'},
            'share'   => $opts{'share'},
            'drive'   => $opts{'drive'},
            'timeout' => $opts{'timeout'},
            'verbose' => $opts{'verbose'}
        );
        if ($opts{'username'}) {
            $args{'username'} = $opts{'username'};
            $args{'password'} = $opts{'password'};
        }
        $Log->debug("map_cifs_share args:\n" . Dumper(\%args));
        my $returnValue = map_cifs_share(%args);
        
	$Log->debug("map_cifs_share returned: \"$returnValue\"");

        if (!$returnValue) {

            $Log->comment("Unable to create mapping to $opts{'share'}");

            if ($i == $opts{'attempts'}) {

                if ($opts{'trace'} && $Capture) {
                    capture_stop($Capture, $opts{'conn'});
                }

               #--------------------------------------------------------------
               # If the drive mappimng failed for any reason, perform client
               # side diagnostics
               #--------------------------------------------------------------
                if ($opts{'diags'}) {
                    $Log->comment("Performing client side diagnostics");
                    client_diagnostics(
                        'conn'  => $opts{'conn'},
                        'share' => $opts{'share'}
                    );
                }
                die "Unable to create mapping to $opts{'share'}\n";
            } else {
                snooze $opts{'delay'};
            }

        } else {

            #-------------------------------------------------------------
            # List the current mappings. If the mapping exists and is not
            # unavailable use the drive
            #-------------------------------------------------------------
            $Log->comment("Verifying mapping to $opts{'share'}");
            my @maps = list_maps(
                'client'  => $opts{'conn'},
                'timeout' => $opts{'timeout'},
                'verbose' => $opts{'verbose'}
            );

            my ($drive, $share, @found);
            foreach my $map (@maps) {
                if (   ${$map}{'path'} eq $opts{'share'}
                    && ${$map}{'status'} !~ /Unavailable/) {
                    my $hash = {
                        'drive' => ${$map}{'drive'},
                        'path'  => ${$map}{'path'},
                    };
                    push(@found, $hash);
                }
            }

            if (@found) {
                $Log->debug(
                    @found . ' potential maps found for verification');
            } else {
                $msg
                    = "No drive mappings found matching $opts{'share'} "
                    . "whose status is not Unavailable "
                    . Dumper(\@maps);
                if ($i == $opts{'attempts'}) {
                    if ($opts{'trace'} && $Capture) {
                        capture_stop($Capture, $opts{'conn'});
                    }
                    die $msg;
                } else {
                    $Log->comment($msg);
                    snooze $opts{'delay'};
                    next;
                }
            }

            if ($opts{'drive'} =~ /\w/) {

                $Log->debug(
                    'Checking for mappings using the requested drive letter');
                foreach my $map (@found) {
                    $Log->debug('Checking mapping; ' . Dumper($map));
                    if ($opts{'drive'} eq ${$map}{'drive'}) {
                        $drive = ${$map}{'drive'};
                        $share = ${$map}{'share'};
                    }
                }

                if (!$drive) {
                    $msg
                        = "Unable to verify mapping to share $opts{'share'}: "
                        . "a drive was specified as $opts{'drive'} but no "
                        . "mapped drive letter was found: "
                        . Dumper(\@found);
                    if ($i == $opts{'attempts'}) {
                        if ($opts{'trace'} && $Capture) {
                            capture_stop($Capture, $opts{'conn'});
                        }
                        die $msg;
                    } else {
                        $Log->comment($msg);
                        snooze $opts{'delay'};
                        next;
                    }
                }
            } elsif ($opts{'drive'} eq '*') {
                $Log->debug('Checking for mappings using any drive letter');
                foreach my $map (@found) {
                    $Log->debug('Checking mapping; ' . Dumper($map));
                    if (${$map}{'drive'} eq $returnValue) {
                        $drive = ${$map}{'drive'};
                        $share = ${$map}{'share'};
                    }
                }
                if (!$drive) {
                    $msg
                        = "Unable to verify mapping to share $opts{'share'}: "
                        . "any drive was specified as $opts{'drive'} but no "
                        . "mapped drive letter was found: "
                        . Dumper(\@found);
                    if ($i == $opts{'attempts'}) {
                        if ($opts{'trace'} && $Capture) {
                            capture_stop($Capture, $opts{'conn'});
                        }
                        die $msg;
                    } else {
                        $Log->comment($msg);
                        snooze $opts{'delay'};
                        next;
                    }
                }

            } else {
                $Log->debug('Checking for mappings using no drive letter');
                foreach my $map (@found) {
                    $Log->debug('Checking mapping; ' . Dumper($map));
                    if (${$map}{'path'} eq $opts{'share'}
                        && !${$map}{'drive'}) {
                        $share = ${$map}{'path'};
                    }
                }

                if (!$share) {
                    $msg = '"No driveless mapping was found '
                        . Dumper(\@found);
                    if ($i == $opts{'attempts'}) {
                        if ($opts{'trace'} && $Capture) {
                            capture_stop($Capture, $opts{'conn'});
                        }
                        die $msg;
                    } else {
                        $Log->comment($msg);
                        snooze $opts{'delay'};
                        next;
                    }
                }
            }

            if ($opts{'trace'} && $Capture) {
                capture_stop($Capture, $opts{'conn'});
            }

            test_share(
                'conn'     => $opts{'conn'},
                'share'    => $opts{'share'},
                'tests'    => $opts{'tests'},
                'attempts' => $opts{'attempts'},
                'trace'    => $opts{'trace'}
            );

            $Log->exit();

            return $drive;

        }
    }
} ## end sub map_share

=head2 unmap_share

Map a unc path and return the drive letter if one was specified. If the mapping fails 
perform client diagnostics and retry.

$drive = unmap_share(conn  => $conObj,
                     share => share);

=over

=item conn

(Required) Conn object of the client required to be passed.

=item share

(Required) The UNC path you want to map to

=item timeout

(Required) The timeout for used commands

=back

=cut

sub unmap_share {

    $Log->enter();

    my (%opts) = validate(
        @_,
        {   conn  => {type => OBJECT, optional => 0},
            delay => {type => SCALAR, optional => 1, default => '5*2'},
            drive => {type => SCALAR | UNDEF,   optional => 1},
            share => {type => SCALAR | UNDEF,   optional => 1},
            maps  => {type => ARRAYREF | UNDEF, optional => 1},
            timeout  => {type => SCALAR, optional => 0, default => '300'},
            verbose  => {type => SCALAR, optional => 1, default => 1},
            attempts => {type => SCALAR, optional => 1, default => 5}
        }
    );

    # Translate IPv6 addresses to ipv6-literal.net format
    $opts{'share'} = ipv6_share_to_ipv6_literal('share' => $opts{'share'});

    # Prepare the regex
    my $path = $opts{'share'};
    $path =~ s/(\\|\.)/\\$1/g;
    my $drive;
    if ($opts{'drive'}) {
        $drive = $opts{'drive'};
        $drive =~ s/://g;
    }

    my $max_attempts = 5;
    my $sleep = SystemTest::Sleep->new(spec => $opts{delay});
    for (my $attempt = 1; $attempt <= $max_attempts; $attempt++) {

        my $msg = "Removing ";
        if ($opts{'drive'}) { $msg .= "$opts{'drive'} "; }
        $msg .= "$opts{'share'} attempt #" . $attempt;
        $Log->comment($msg);

        if ($attempt > 1) { $opts{'verbose'} = 1; }

        # Get the existing mappings
        my @maps = list_maps(
            'client'  => $opts{'conn'},
            'timeout' => $opts{'timeout'},
            'verbose' => $opts{'verbose'}
        );

        $Log->debug("maps: " . Dumper(\@maps));

        # Walk through each mapping and remove accordingly
        $msg = "Checking existing mappings for ";
        if ($opts{'drive'}) { $msg .= "$opts{'drive'} "; }
        $msg .= $opts{'share'};
        $Log->comment($msg);

        foreach my $map (@maps) {
            $Log->debug("drive:${$map}{'drive'} path:${$map}{'path'}");
            if (   $drive
                && ${$map}{'drive'} =~ /$drive/i
                && ${$map}{'path'}  =~ /^$path$/i) {
                $Log->comment("Removing ${$map}{'drive'} ${$map}{'path'}");
                unmap_cifs_share(
                    'client'  => $opts{'conn'},
                    'drive'   => ${$map}{'drive'},
                    'force'   => 1,
                    'verbose' => $opts{'verbose'}
                );
            } elsif (!$drive
                && ${$map}{'drive'} =~ /[a-z]/i
                && ${$map}{'path'}  =~ /^$path$/i) {
                $Log->comment("Removing ${$map}{'drive'} ${$map}{'path'}");
                unmap_cifs_share(
                    'client'  => $opts{'conn'},
                    'drive'   => ${$map}{'drive'},
                    'force'   => 1,
                    'verbose' => $opts{'verbose'}
                );
            } elsif (!$drive && ${$map}{'path'} =~ /^$path$/i) {
                $Log->comment("Removing ${$map}{'path'}");
                unmap_cifs_share(
                    'client'  => $opts{'conn'},
                    'mappath' => ${$map}{'path'},
                    'force'   => 1,
                    'verbose' => $opts{'verbose'}
                );
            }
        }

        $Log->comment("Waiting for the client to remove the mapping");
        $sleep->delay(increment => 1);

        @maps = list_maps(
            'client'  => $opts{'conn'},
            'timeout' => $opts{'timeout'},
            'verbose' => $opts{'verbose'}
        );

        $msg = "Verifying ";
        if ($opts{'drive'}) { $msg .= "$opts{'drive'} "; }
        $msg .= "$opts{'share'} has been removed";
        $Log->comment($msg);
        my $removed = 1;
        foreach my $map (@maps) {
            $Log->debug("drive:${$map}{'drive'} path:${$map}{'path'}");
            if (   $drive
                && ${$map}{'drive'} =~ /$drive/i
                && ${$map}{'path'}  =~ /^$path$/i) {
                $Log->comment(
                    "${$map}{'drive'} ${$map}{'path'} still exists");
                $removed = 0;
            } elsif (!$drive
                && ${$map}{'drive'} =~ /[a-z]/i
                && ${$map}{'path'}  =~ /^$path$/i) {
                $Log->comment(
                    "${$map}{'drive'} ${$map}{'path'} still exists");
            } elsif (!$drive && ${$map}{'path'} =~ /^$path$/i) {
                $Log->comment("${$map}{'path'} still exists");
            }
        }
        if ($removed) {
            last;
        } elsif ($attempt == $max_attempts) {
            my $msg = "Unable to remove ";
            if ($opts{'drive'}) { $msg .= "$opts{'drive'} "; }
            $msg .= "$opts{'share'}";
            die $msg;
        }
    }
    $Log->exit();

} ## end sub unmap_share

=head2 create_client_logpath {

Creates the directory TOP_LOGDIR/$client_dir/clientname

 $path = create_client_logpath(conn => $conn, logdir => $logdir);

=over

=item conn

(Required) Conn object of the client required to be passed.

=item top_logdir_windows

(Required) The NATE param LOGDIR_WINDOWS

=item timeout

(Required) The timeout for used commands

=back

=cut

sub create_client_logpath {

    $Log->enter();

    my (%opts) = validate(
        @_,
        {   conn               => {type => OBJECT},
            top_logdir_windows => {type => SCALAR},
            client_dir         => {type => SCALAR, default => 'clients'},
            timeout            => {type => SCALAR, default => '300'}
        }
    );

    my $hostname           = $opts{'conn'}->hostp()->hostname();
    my $client_dir         = $opts{'client_dir'};
    my $top_logdir_windows = $opts{'top_logdir_windows'};
    my ($jobId, $output);

    #----------------------------------------------
    # Get a directory listing of the log directory
    #----------------------------------------------
    $Log->comment("Searching for  $top_logdir_windows\\$client_dir");
    my $cmd = "cmd.exe /c dir $top_logdir_windows";
    eval {
        (undef, undef, $output)
            = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
            -noverbose);
    };

    my $error = $@;
    if ($error) {
        $Log->comment("$cmd\n $output\n $error");
        die "Error searching for the $client_dir directory";
    } elsif ($output =~ /cannot find the drive specified/) {
        $Log->comment("$cmd\n $output");
        die "Error $client_dir directory not found";
    } else {
        $Log->debug("$cmd\n $output");
    }

    #-------------------------------------------------------
    # If the $client_dir directory does nto exist, create it
    #-------------------------------------------------------
    if ($output !~ /<DIR>\s+$client_dir/) {
        $Log->comment("Creating $top_logdir_windows\\$client_dir");
        $cmd = "cmd.exe /c mkdir $top_logdir_windows\\$client_dir";
        eval {
            (undef, undef, $output)
                = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
                -noverbose);
        };

        my $error = $@;
        if ($error) {
            $Log->comment("$cmd\n $output\n $error");
            die "Error creating the $client_dir directory";
        } else {
            $Log->debug("$cmd\n $output");
        }
    }

    #-----------------------------------------------------
    # Get a directory listing of the $client_dir directory
    #-----------------------------------------------------
    $cmd = "cmd.exe /c dir $top_logdir_windows\\$client_dir";
    eval {
        (undef, undef, $output)
            = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
            -noverbose);
    };

    $error = $@;
    if ($@) {
        $Log->comment("$cmd\n $output\n $error");
        die "Error listing the $client_dir directory";
    } elsif ($output =~ /cannot find the drive specified/) {
        $Log->comment("$cmd\n $output");
        die "Error $client_dir directory was not created";
    } else {
        $Log->debug("$cmd\n $output");
    }

    #------------------------------------------------------
    # If the client's directory does not exist, create it
    #------------------------------------------------------
    if ($output !~ /<DIR>\s+$hostname$/) {
        $Log->comment("Creating $top_logdir_windows\\$client_dir\\$hostname");
        $cmd
            = "cmd.exe /c if not exist $top_logdir_windows\\$client_dir\\$hostname "
            . "mkdir $top_logdir_windows\\$client_dir\\$hostname";
        eval {
            (undef, undef, $output)
                = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
                -noverbose);
        };

        my $error = $@;
        if ($error) {
            $Log->comment("$cmd\n $output\n $error");
            die "Error creating the $client_dir\\$hostname directory";
        } else {
            $Log->debug("$cmd\n $output");
        }
    }

    #--------------------------------------------------------------
    # Get a directory listing of the $client_dir\hostname directory
    #--------------------------------------------------------------
    $cmd = "cmd.exe /c dir $top_logdir_windows\\$client_dir\\$hostname";
    eval {
        (undef, undef, $output)
            = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'},
            -noverbose);
    };

    $error = $@;
    if ($error) {
        $Log->comment("$cmd\n $output\n $error");
        die "Error listing the $client_dir\\$hostname directory";
    } elsif ($output =~ /cannot find the drive specified/) {
        $Log->comment("$cmd\n $output");
        die "Error $client_dir\\$hostname directory was not created";
    } else {
        $Log->debug("$cmd\n $output");
    }
    $Log->exit();

    return "$top_logdir_windows\\$client_dir\\$hostname";
} ## end sub create_client_logpath

=head2 ipv6_share_to_ipv6_literal

Returns the ipv6-literal.net form of a UNC path containing an IPv6 address.

 $share = ipv6_share_to_ipv6_literal(share => \\\\fd20:81be:b255:8100:0:1:0:1\\kc55b7);

=over

=item share 

(Required) The UNC path of the share

=back

=cut

sub ipv6_share_to_ipv6_literal {

    $Log->enter();

    my (%opts) = validate(@_, {share => {type => SCALAR}});
    my $original_share = $opts{'share'};

    #\\\\fd20:81be:b255:8100:0:1:0:1\\kc55b7
    if ($opts{'share'} =~ /^([\\]+)([^\\]+)([\\]+)(.*)/) {
        my $sharePart1 = $1;
        my $sharePart2 = $2;
        my $sharePart3 = $3;
        my $sharePart4 = $4;
        $sharePart2 = ipv6_address_to_ipv6_literal('address' => $sharePart2);
        $opts{'share'}
            = $sharePart1 . $sharePart2 . $sharePart3 . $sharePart4;

        if ($opts{'share'} ne $original_share) {
            $Log->debug("Share is now $opts{'share'}");
        }
    }

    $Log->exit();
    return $opts{'share'};

} ## end sub ipv6_share_to_ipv6_literal

=head2 ipv6_address_to_ipv6_literal
       
Converts an IPv6 address to Microsofts .ip6-literal.net format
              
 $ipv6LiteralName = ipv6_address_to_ipv6_literal(address => fd20:81be:b255:8100:0:1:0:1);

=over

=item address
    
(Required) IPv6 Address
        
=back
    
=cut    

sub ipv6_address_to_ipv6_literal {

    $Log->enter();

    my (%opts) = validate(@_, {address => {type => SCALAR}});
    my $address = $opts{'address'};

    if ($address =~ /:/) {
        $Log->comment("Converting $opts{'address'} to ipv6-literal format");
        $address =~ s/:/-/g;
        $address .= '.ipv6-literal.net';
        $Log->comment("$opts{'address'} is now $address");
    }

    $Log->exit();

    return $address;

} ## end sub ipv6_address_to_ipv6_literal

=head2 ipv6_literal_to_ipv6_address
       
Converts an IPv6 address to Microsofts .ip6-literal.net format
              
 $ipv6Address = ipv6_literal_to_ipv6_address(name => fd20-81be-b255-8100-0-1-0-1.ipv6-literal.net);

=over

=item name
    
(Required) ipv6-literal.net name
        
=back
    
=cut    

sub ipv6_literal_to_ipv6_address {

    $Log->enter();

    my (%opts) = validate(@_, {name => {type => SCALAR}});
    my $address = $opts{'name'};

    if ($address =~ /ipv6-literal\.net/i) {
        $Log->comment("Converting $opts{'name'} to IPv6 address");
        $address =~ s/\.ipv6-literal\.net//i;
        $address =~ s/-/:/g;
        $Log->debug("$opts{'name'} is now $address");
    }

    $Log->exit();

    return $address;

} ## end sub ipv6_literal_to_ipv6_address

=head2 prep_for_windows_subtest
       
Do whatever is needed on the client to get it ready for subtests
	      
 $share = prep_for_windows_subtest(conn => $conn);

=over

=item conn
    
(Required) The client connection object
	
=item timeout
    
(Optional) The say timeout to use
	
=back
    
=cut    

sub prep_for_windows_subtest {

    $Log->enter();
    $Log->comment("Preparing for subtesting on windows");

    $Options->process();
    my $options = $Options->get_hash();

    my (%opts) = validate(
        @_,
        {   conn    => {type => OBJECT, optional => 0},
            timeout => {
                type     => SCALAR,
                optional => 1,
                default  => 300
            }
        }
    );

    #------------------------------------
    # Map a drive to the nate log server
    #------------------------------------
    if ($options->{'LOGDIR_SHARE'}) {
        $Log->comment("Mapping to LOGDIR_SHARE");
        map_share(
            'conn'     => $opts{'conn'},
            'share'    => $options->{'LOGDIR_SHARE'},
            'drive'    => '',
            'username' => $options->{'LOGDIR_USER'},
            'password' => $options->{'LOGDIR_PASSWD'},
            'tests'    => 'r',
            'timeout'  => $opts{'timeout'},
	    'trace'    => 0
        );
    }

    #----------------------------------------------------------------
    # If this is not here staf will say it can't write to the LOGDIR
    #----------------------------------------------------------------
    $Log->comment("Testing TOP_LOGDIR_WINDOWS");
    test_share(
        'conn'  => $opts{'conn'},
        'share' => $options->{'TOP_LOGDIR_WINDOWS'},
        'trace' => 1
    );

    #------------------------------------
    # Map a drive to the nate log server
    #------------------------------------
    if ($options->{'HOME_SHARE'}) {
        $Log->comment("Mapping to HOME_SHARE");
        map_share(
            'conn'     => $opts{'conn'},
            'share'    => $options->{'HOME_SHARE'},
            'drive'    => '',
            'username' => $options->{'HOME_SHARE_USER'},
            'password' => $options->{'HOME_SHARE_PASSWD'},
            'tests'    => 'r',
            'timeout'  => $opts{'timeout'}
        );
    }

    $Log->exit();

} ## end sub prep_for_windows_subtest

=head2 test_share

Tests read and write access to a share

 test_share(conn => $conObj, share => share);

=over

=item conn

(Required) Conn object of the client required to be passed.

=item share

(Required) The UNC path you want to map to

=item tests

(Optional) The tests to perform r=dir listing, w=create/delete file, rw=both

=item diags

(Optional) Perform client side diags if the test fails

=back

=cut

sub test_share {

    $Log->enter();

    my (%opts) = validate(
        @_,
        {   conn     => {type => OBJECT, optional => 0},
            share    => {type => SCALAR, optional => 0},
            diags    => {type => SCALAR, optional => 1, default => 1},
            tests    => {type => SCALAR, optional => 1, default => 'rw'},
            attempts => {type => SCALAR, optional => 1, default => 3},
            trace => {type => SCALAR | UNDEF, optional => 1, default => 1},
            delay => {type => SCALAR,         optional => 1, default => 60},
        }
    );

    $Log->comment("Parsing share path $opts{'share'}");
    $opts{'share'} =~ /^[\\]+([^\\]+)[\\]+([^\\]+)/;
    my $server = $1;
    my $share  = $2;
    $Log->comment("Server:$server Share:$share");
    my $address;
    if ($server) {
        $address = ipv6_literal_to_ipv6_address('name' => $server);
    } else {
        die script_result(
            "Unable to determine server from share path: $opts{'share'}");
    }

    my ($cmd, $output, $Capture);
    my $hostname = $opts{'conn'}->hostp()->hostname();
    if ($opts{'tests'} =~ /r/i) {
        for (my $i = 1; $i <= $opts{'attempts'}; $i++) {
            if ($i > 1) { snooze $opts{'delay'}; }
            if ($i == 2) {
                if ($opts{'trace'}) {
                    $Capture = capture_start($opts{'conn'}, $address);
                }
            }
            $Log->comment(
                "Testing read access to $opts{'share'} attempt #" . $i);

            $cmd = "cmd.exe /c dir $opts{'share'}";

            eval {
                (undef, undef, $output)
                    = say($opts{'conn'}, $cmd, -timeout, 60, -noverbose);
            };

            my $error = $@;
            if ($error || $output !~ /<DIR>|bytes\s+free|Directory\s+of/) {
                if ($i == $opts{'attempts'}) {
                    if ($opts{'trace'} && $Capture) {
                        capture_stop($Capture, $opts{'conn'});
                    }
                    if ($opts{'diags'}) {
                        client_diagnostics(
                            'conn'  => $opts{'conn'},
                            'share' => $opts{'share'}
                        );
                    }
                    die
                        "Error listing $opts{'share'}:$cmd\n $output\n $error";
                }
            } else {
                $Log->debug("$cmd\n $output");
                if ($opts{'trace'} && $Capture) {
                    capture_stop($Capture, $opts{'conn'});
                }
                last;
            }
        }
    }

    if ($opts{'tests'} =~ /w/i) {
        for (my $i = 1; $i <= $opts{'attempts'}; $i++) {

            if ($i > 1) { snooze $opts{'delay'}; }
            if ($i == 2) {
                if ($opts{'trace'}) {
                    $Capture = capture_start($opts{'conn'}, $address);
                }
            }

            my $testFile
                = 'test_share_'
                . $opts{'conn'}->hostp->hostname() . '_'
                . strftime("%m%d%y%H%M%S", localtime());
            $Log->comment("Testing write access to "
                    . "$opts{'share'}\\$testFile attempt #"
                    . $i);
            $cmd = "cmd /c echo \"Created by CIFS::Client::test_share\" > $opts{'share'}\\$testFile";

            eval {
                (undef, undef, $output)
                    = say($opts{'conn'}, $cmd, -timeout, 300, -noverbose);
            };

            my $error = $@;

            if ($output && $output =~ /No space left on device|There is not enough space|invalid|incorrect/) {
                $Log->comment("Skipping write testing due to lack of space");
                last;
            }

            if ($error || $output) {
                if ($i == $opts{'attempts'}) {
                    my $msg
                        = "Error writing test file "
                        . "$opts{'share'}\\$testFile: "
                        . "$cmd\n $output\n $error";
                    $Log->comment($msg);
                    if ($opts{'trace'} && $Capture) {
                        capture_stop($Capture, $opts{'conn'});
                    }
                    client_diagnostics(
                        'conn'  => $opts{'conn'},
                        'share' => $opts{'share'}
                    );
                    die $msg;
                } else {
                    next;
                }
            } else {
                $Log->debug("$cmd\n $output");
            }

            snooze 1;

            $cmd = "cmd.exe /c dir $opts{'share'}\\$testFile";

            eval {
                (undef, undef, $output)
                    = say($opts{'conn'}, $cmd, -timeout, 60, -noverbose);
            };

            $error = $@;
            if ($error || $output !~ /$testFile/) {
                if ($i == $opts{'attempts'}) {
                    my $msg
                        = "Error verifying write access to: "
                        . "$opts{'share'}\\$testFile: "
                        . "$cmd\n $output\n $error";
                    $Log->comment($msg);
                    if ($opts{'trace'} && $Capture) {
                        capture_stop($Capture, $opts{'conn'});
                    }
                    client_diagnostics(
                        'conn'  => $opts{'conn'},
                        'share' => $opts{'share'}
                    );
                    die $msg;
                } else {
                    next;
                }
            } else {
                $Log->debug("$cmd\n $output");
            }

            snooze 1;

            $cmd = "cmd.exe /c del /q $opts{'share'}\\$testFile";
            eval {
                (undef, undef, $output)
                    = say($opts{'conn'}, $cmd, -timeout, 60, -noverbose);
            };

            $error = $@;
            if ($error || $output) {
                if ($i == $opts{'attempts'}) {
                    my $msg
                        = "Error removing "
                        . "$opts{'share'}\\$testFile: "
                        . "$cmd\n $output\n $error";
                    $Log->comment($msg);
                    if ($opts{'trace'} && $Capture) {
                        capture_stop($Capture, $opts{'conn'});
                    }
                    client_diagnostics(
                        'conn'  => $opts{'conn'},
                        'share' => $opts{'share'}
                    );
                    die $msg;
                } else {
                    next;
                }
            } else {
                $Log->debug("$cmd\n $output");
                last;
            }

        }

        if ($opts{'trace'} && $Capture) {
            capture_stop($Capture, $opts{'conn'});
        }

        snooze 1;

        $Log->exit();
    }
} ## end sub test_share

=head2 client_diagnostics

Performs basic client diagnostic to capture client and network share 
status at the time of failures

 client_diagnostics(conn => $conObj, share => share);

=over

=item conn

(Required) Conn object of the client required to be passed.

=item share

(Required) The UNC path you want to test

=back

=cut

sub client_diagnostics {

    $Log->enter();

    my (%opts) = validate(
        @_,
        {   conn    => {type => OBJECT, optional => 0},
            share   => {type => SCALAR, optional => 1},
            timeout => {type => SCALAR, default  => '300'}
        }
    );

    check_connectivity('conn' => $opts{'conn'});

    if ($opts{'share'}) {

        $Log->comment("Testing $opts{'share'}");
        $opts{'share'} =~ /.+\\([^\\]+)\\([^\\]+)/;
        my $server = $1;
        my $share  = $2;

        $Log->comment("Querying DNS for $server");
        my $cmd = "nslookup $server";
        eval { say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'}); };

        my $error = $@;
        if ($error) {
            $Log->comment("Error excuting $cmd: $error");
        }

        my $output;
        my $address = ipv6_literal_to_ipv6_address('name' => $server);

        $Log->comment("Pinging $address using ICMP");
        $cmd = "ping $address";
        eval {
            (undef, undef, $output)
                = say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'});
        };

        $error = $@;
        if ($error) {
            $Log->comment("Error excuting $cmd: $error");
        }

        if ($output =~ /Reply from/) {
            if ($output =~ /$share\s+Disk/) {
                $Log->comment("Listing $opts{'share'}");

                $cmd = "cmd.exe /c dir $opts{'share'}";
                eval { say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'}); };

                $error = $@;
                if ($error) {
                    $Log->comment("Error excuting $cmd: $error");
                }

            }
        }
    }

    $Log->comment("Capturing network state and configuration info");

    my $cmd = 'net use';
    eval { say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'}); };

    my $error = $@;
    if ($error) {
        $Log->comment("Error excuting $cmd: $error");
    }

    $cmd = 'ipconfig /all';
    eval { say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'}); };
    $error = $@;
    if ($error) {
        $Log->comment("Error excuting $cmd: $error");
    }

    $cmd = 'netstat';
    eval { say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'}); };
    $error = $@;
    if ($error) {
        $Log->comment("Error excuting $cmd: $error");
    }

    $cmd = 'route print';
    eval { say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'}); };
    $error = $@;
    if ($error) {
        $Log->comment("Error excuting $cmd: $error");
    }

    $cmd = 'arp -a';
    eval { say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'}); };
    $error = $@;
    if ($error) {
        $Log->comment("Error excuting $cmd: $error");
    }

    $Log->comment("Capturing the process list");

    $cmd = 'tasklist';
    eval { say($opts{'conn'}, $cmd, -timeout, $opts{'timeout'}); };

    $error = $@;
    if ($error) {
        $Log->comment("Error excuting $cmd: $error");
    }

    $Log->exit();
} ## end sub client_diagnostics

=head2 process_cifs_clients

Common wrapper for client processing

 process_cifs_clients(script => $script, runid => $runid, clients => \@clients);

=over

=item options

(Required) deprecated

=item script

(Required) The name of the calling script

=item runid

(Required) base runid string

=item clients

(Required) Arrayref of clients

=back

=cut

sub process_cifs_clients {

    $Log->enter();

    #-------------------------------------------------
    # 04/15/10 bellzzi@netapp.com
    # This line conflicts with TestManager lite tests
    #-------------------------------------------------
    #use Utils::SSQA qw(strip_vars get_max_clients check_subtest_status);
    $Options->process();
    my $options = $Options->get_hash();

    my (%opts) = validate(
        @_,
        {   options => {type => HASHREF | UNDEF, optional => 1},
            script  => {type => SCALAR},
            runid   => {type => SCALAR},
            clients => {type => ARRAYREF}
        }
    );

    my $volumeIndex          = 0;
    my $subtestId            = 1;
    my $cifsServerAliasIndex = 0;
    my $dataInterfaceIndex   = 0;
    my $maxClients           = Utils::SSQA::get_max_clients();
    my $clientOptions        = Utils::SSQA::strip_vars();
    my $Remote               = new RemoteExecution;
    my ($passed, $failed, $running, $complete, @processed, $clientConnType);

    if ($options->{'USE_STAF'}) {
        $clientConnType = '#staf';
    } else {
        $clientConnType = '#rsh';
    }

    foreach my $key (keys(%{$options})) {
        if (!grep { $_ eq $key } @Tharn::Alias_params) {
            $clientOptions->{$key} = $options->{$key};
        }
    }

    foreach my $client (@{$opts{'clients'}}) {

        #--------------------------------------------
        # This is a CIFS load so use windows clients
        #--------------------------------------------
        if (host($client)->hosttype() !~ /win/i) { next; }

        #-------------------
        # Connect to client
        #-------------------
        my $clientConnObj
            = connect($clientConnType, '-nologin', -timeout, 300);

        #-----------------------------------
        # Verify connectivity to the client
        #-----------------------------------
        eval { check_connectivity('conn' => $clientConnObj); };
        if ($@) {

          #-------------------------------------------------------------------
          # Reset the $cifsServerAliasIndex if we are at the end of the array
          #-------------------------------------------------------------------
            if ($cifsServerAliasIndex
                >= @{$options->{'CIFS_SERVER_ALIASES'}} - 1) {
                $cifsServerAliasIndex = 0;
            } else {
                $cifsServerAliasIndex++;
            }

            #-----------------------------------------------------------------
            # Reset the $dataInterfaceIndex if we are at the end of the array
            #-----------------------------------------------------------------
            if ($dataInterfaceIndex >= @{$options->{'DATA_INTERFACES'}} - 1) {
                $dataInterfaceIndex = 0;
            } else {
                $dataInterfaceIndex++;
            }

            #----------------------------------------------------------
            # Reset the $volumeIndex if we are at the end of the array
            #----------------------------------------------------------
            if ($volumeIndex >= @{$options->{'VOLUME'}} - 1) {
                $volumeIndex = 0;
            } else {
                $volumeIndex++;
            }
            $Log->warn($@);
            next;
        }

        my $subtestTarget = 'localhost';
        if ($options->{'SUBTEST_TYPE'} eq 'remote') {
            $subtestTarget = $client;
            prep_for_windows_subtest('conn' => $clientConnObj);
        }

      # ----------------------------------------------------------------------
      # If CIFS_SERVER_ACCESS_MODE is data_interfaces use the IP address
      # in DATA_INTERFACES referenced by $dataInterfaceIndex. See burt 363274
      # for more info If CIFS_SERVER_ACCESS_MODE is cifs_server_aliases
      # use the name in CIFS_SERVER_ALIASES referenced by
      # $cifsServerAliasIndex
      # --------------------------------------------------------------------
        if (   $options->{'DATA_INTERFACES'}
            && $options->{'CIFS_SERVER_ACCESS_MODE'} eq 'data_interfaces') {

            $clientOptions->{'CIFS_SERVER'}
                = ${$options->{'DATA_INTERFACES'}}[$dataInterfaceIndex];

        } elsif ($options->{'CIFS_SERVER_ALIASES'}
            && $options->{'CIFS_SERVER_ACCESS_MODE'} eq 'cifs_server_aliases')
        {

            $clientOptions->{'CIFS_SERVER'}
                = ${$options->{'CIFS_SERVER_ALIASES'}}[$cifsServerAliasIndex];

        }

        $Log->comment("Client $client will use "
                . "$clientOptions->{'CIFS_SERVER'} for CIFS access");

       #----------------------------------------------------------------------
       # If VOLUME > 1 and VOLUME_USAGE eq 'multiple' assign volumes in round
       # robin format
       #----------------------------------------------------------------------
        if (@{$options->{'VOLUME'}} > 1
            && $options->{'VOLUME_USAGE'} eq 'multiple') {

            $clientOptions->{'VOLUME_INDEX'} = $volumeIndex;

        }

        $Log->comment("Client $client will use target volume "
                . "${$clientOptions->{'VOLUME'}}[$clientOptions->{'VOLUME_INDEX'}]"
        );

        $clientOptions->{'CLIENT_ID'} = $client;
        $Remote->execute(
            'host'  => "$subtestTarget;prefer_family=inet4",
            'cmd'   => $opts{'script'},
            'runid' => "${client}-$opts{'runid'}-$options->{'TIMESTAMP'}",
            'nate_params' => $clientOptions,
            'handler'     => sub { },
        );

        #---------------------------------------------------------
        # Limit the amount of subtests if MAX_SUBTESTS is defined
        #---------------------------------------------------------
        $subtestId++;
        if (   $subtestId != 1
            && ($subtestId % $options->{'MAX_SUBTESTS'}) == 1
            && $options->{'SUBTEST_TYPE'} eq 'local') {

       #----------------------------------------------------------------------
       # Instead of using $Remote->wait() keep parsing the subtests to find
       # failures so we can exit immeadiately since this could potentially be
       # a long running test
       #----------------------------------------------------------------------
            $Log->comment(
                "Waiting for $options->{'MAX_SUBTESTS'} processes to finish");
            ($passed, $failed, $running, $complete) = (0, 0, 0, 0);
            while ($complete < @processed) {
                ($passed, $failed, $running, $complete)
                    = Utils::SSQA::check_subtest_status(
                    'obj'   => $Remote,
                    'regex' => $options->{'TIMESTAMP'}
                    );
                if ($failed) {
                    $Log->exit();
                    die 'Error cleaning up client(s)';
                }
                snooze 15;
            }
        }

        push(@processed, $client);

       #----------------------------------------------------------------------
       # If the configured clients is equal to MAX_CLIENTS, stop configuring
       # clients
       #----------------------------------------------------------------------
        if (@processed == $maxClients) {
            $Log->comment(
                "MAX_CLIENTS threshold of $maxClients has been reached");
            last;
        }

        #----------------------------------------------------------
        # Check the subtest status and die if there are any errors
        #----------------------------------------------------------
        ($passed, $failed, $running, $complete)
            = Utils::SSQA::check_subtest_status(
            'obj'   => $Remote,
            'regex' => $options->{'TIMESTAMP'}
            );
        if ($failed) {
            $Log->exit();
            die 'Error cleaning up client(s)';
        }

        #-------------------------------------------------------------------
        # Reset the $cifsServerAliasIndex if we are at the end of the array
        #-------------------------------------------------------------------
        if ($cifsServerAliasIndex >= @{$options->{'CIFS_SERVER_ALIASES'}} - 1)
        {
            $cifsServerAliasIndex = 0;
        } else {
            $cifsServerAliasIndex++;
        }

        #-----------------------------------------------------------------
        # Reset the $dataInterfaceIndex if we are at the end of the array
        #-----------------------------------------------------------------
        if ($dataInterfaceIndex >= @{$options->{'DATA_INTERFACES'}} - 1) {
            $dataInterfaceIndex = 0;
        } else {
            $dataInterfaceIndex++;
        }

        #----------------------------------------------------------
        # Reset the $volumeIndex if we are at the end of the array
        #----------------------------------------------------------
        if ($volumeIndex >= @{$options->{'VOLUME'}} - 1) {
            $volumeIndex = 0;
        } else {
            $volumeIndex++;
        }

    }

    #----------------------------------------------------------------------
    # Instead of using $Remote->wait() keep parsing the subtests to find
    # failures so we can exit immeadiately since this could potentially be
    # a long running test
    #----------------------------------------------------------------------
    ($passed, $failed, $running, $complete) = (0, 0, 0, 0);
    while ($complete < @processed) {
        ($passed, $failed, $running, $complete)
            = Utils::SSQA::check_subtest_status(
            'obj'   => $Remote,
            'regex' => $options->{'TIMESTAMP'}
            );
        if ($failed) {
            $Log->exit();
            die 'Error cleaning up client(s)';
        }
        snooze 15;
    }

    $Log->exit();

} ## end sub process_cifs_clients

sub capture_start {
    my ($conn, $server) = @_;
    my $hostname = $conn->hostp()->hostname();
    my $Capture;
    my $client = NACL::C::Client->new(name => $hostname);
    my $details = [{target_machine => $client}];
    if ($server) {
        if (!ref($server)) {
            my $address = ipv6_literal_to_ipv6_address('name' => $server);
            if ($address !~ /(^\d+\.\d+\.\d+\.\d+$|:)/) {
                $Log->comment("Resolving $address");
                my $DNS = get_ip_by_name('name' => $address);

                if ($DNS->{'ERROR'}) {
                    die "Unable to get IP address from server name $address";
                }
                if ($DNS->{'INET4'}) {
                    $address = $DNS->{'INET4'};
                } elsif ($DNS->{'INET6'}) {
                    $address = $DNS->{'INET6'};
                }
            }
            $details->[0]->{capture_filter} = "host $address";
        } else {
            my $filter;
            foreach my $s (@{$server}) {
                my $address = ipv6_literal_to_ipv6_address('name' => $s);
                if ($address !~ /(^\d+\.\d+\.\d+\.\d+$|:)/) {
                    $Log->comment("Resolving $address");
                    my $DNS = get_ip_by_name('name' => $address);

                    if ($DNS->{'ERROR'}) {
                        die
                            "Unable to get IP address from server name $address";
                    }
                    if ($DNS->{'INET4'}) {
                        $address = $DNS->{'INET4'};
                    } elsif ($DNS->{'INET6'}) {
                        $address = $DNS->{'INET6'};
                    }
                }
                $filter .= "host $address or ";
            }
            $filter =~ s/or $//;
            $details->[0]->{capture_filter} = $filter;
        }
    }
    $Log->comment("Enabling packet capture on $hostname with the following parameters\n". Dumper($details));
    $Capture
        = NACL::MTask::PacketCapture->new(packet_capture_details => $details);
    $Capture->start();
    sleep 2;
    return $Capture;
} ## end sub capture_start

sub capture_stop {

    my ($Capture, $conn) = @_;

    $Options->process();
    my $options  = $Options->get_hash();
    my $hostname = $conn->hostp()->hostname();

    $Log->comment("Stopping packet capture on $hostname");
    sleep 2;
    if ($Capture) {
        eval { $Capture->stop(); };
        if ($@) { $Log->comment($@); }
        eval { $Capture->show(); };
        if ($@) { $Log->comment($@); }
        sleep 10;
        my $top_logdir_windows = $options->{'TOP_LOGDIR_WINDOWS'};
        if ($top_logdir_windows) {
            my $log_path = create_client_logpath(
                'conn'               => $conn,
                'top_logdir_windows' => $top_logdir_windows,
                'timeout'            => '300'
            );
            sleep 2;
            my $params = $Capture->{'packet_capture_params'};
            foreach my $target (keys(%{$params})) {
                my $files = $params->{$target}->{'capture_file'};
                foreach my $id (keys(%{$files})) {
                    my $file = $params->{$target}->{'capture_file'}->{$id};
                    my $cmd  = "cmd.exe /c copy $file $log_path";
                    my $output;
                    $Log->comment("Copying $file $log_path");
                    eval {
                        (undef, undef, $output)
                            = say($conn, $cmd, -timeout, 300, -noverbose);
                    };

                    my $error = $@;
                    if ($error) {
                        $Log->comment("$cmd\n $output\n $error");
                        die "Error copying capture file to $log_path";
                    } else {
                        $Log->debug("$cmd\n $output");

                    }
                }
            }
            sleep 2;
        }
        eval { $Capture->cleanup(); };
        if ($@) { $Log->comment($@); }
        $Capture = undef;
    }

} ## end sub capture_stop

1;
