# $Id$

## @summary Library for Windows Event Log operations from NATE on Windows
## @author mueller@netapp.com
## @pod here

=head1 NAME

cifsFileCreate

=head1 SYNOPSIS

    use CIFS::EventLog qw(:all);

=head1 ROUTINES

=cut

package CIFS::EventLog;
use strict;

BEGIN {
  use Module::Common;
  Module::Common::setup_export_tags(
    ops => [ qw(cifsEventLogOpen
                cifsEventLogClose
		cifsEventLogSearch
                cifsEventLogRead
                cifsEventLogWrite
		cifsEventLogSave
                cifsEventLogClear
		cifsEventLogOldestIndex
                cifsEventLogNumEntries) ],
    constant_read => [ qw(EVENTLOG_SEEK_READ
                          EVENTLOG_SEQUENTIAL_READ
			  EVENTLOG_FORWARDS_READ
			  EVENTLOG_BACKWARDS_READ) ],
    constant_write => [ qw(EVENTLOG_ERROR_TYPE
                           EVENTLOG_WARNING_TYPE
			   EVENTLOG_INFORMATION_TYPE
			   EVENTLOG_AUDIT_SUCCESS
			   EVENTLOG_AUDIT_FAILURE) ]
  );
}

use Params::Validate v0.70 qw(validate SCALAR SCALARREF HASHREF UNDEF);
use Carp qw(croak);
use base qw(Exporter);
use Win32::EventLog;
use Win32;
use CIFS::Misc qw(:all);

sub cifsEventLogOpen {
    my (%args) = validate (@_, {
	sSourceName	=> { type => SCALAR },
	sServerName	=> { type => SCALAR, optional => 1, default => $ENV{ComputerName} }
    });
    my $handle;

    if ($handle = Win32::EventLog->new($args{sSourceName}, $args{sServerName})) {
      # Currently, this never fails locally.  If sSourceName is not an existing
      # log, it opens 'Application'.  Weird.
      if ($handle->{handle} != 0) {
        return $handle;
      }
    }
    
    return 0;
}

sub cifsEventLogClose {
    my (%args) = validate (@_, {
	hEventLog	=> { type => HASHREF }
    });

    $args{hEventLog}->Close();
    return 1;
}

sub cifsEventLogNumEntries {
    my (%args) = validate (@_, {
	hEventLog	=> { type => HASHREF }
    });
    my $uEntries;

    $args{hEventLog}->GetNumber($uEntries);
    return $uEntries;
}

sub cifsEventLogClear {
    my (%args) = validate (@_, {
        hEventLog	=> { type => HASHREF },
        sBackupFilename	=> { type => SCALAR, optional => 1, default => () }
    });

    $args{hEventLog}->Clear($args{sBackupFilename});
    return 1;
}

sub cifsEventLogSave {
    my (%args) = validate (@_, {
        hEventLog	=> { type => HASHREF },
	sBackupFilename	=> { type => SCALAR }
    });

    $args{hEventLog}->Backup($args{sBackupFilename});
    return 1;
}

sub cifsEventLogOldestIndex {
    my (%args) = validate (@_, {
        hEventLog	=> { type => HASHREF }
    });
    my $id;

    $args{hEventLog}->GetOldest($id);
    return $id;
}

sub cifsEventLogRead {
    my (%args) = validate (@_, {
        hEventLog	=> { type => HASHREF },
	uEventID	=> { type => SCALAR }
    });
    my $hashRef;

    if ($args{hEventLog}->Read(EVENTLOG_FORWARDS_READ|EVENTLOG_SEEK_READ,
                           $args{uEventID},
			   $hashRef)) {
      Win32::EventLog::GetMessageText($hashRef);
      return %{$hashRef};
    }

    return 0;
}

sub cifsEventLogWrite {
    my (%args) = validate (@_, {
        hEventLog	=> { type => HASHREF },
	sMessage	=> { type => SCALAR },
	sSource		=> { type => SCALAR, optional => 1, default => "NATE-CIFS" },
	uEventType	=> { type => SCALAR, optional => 1, default => EVENTLOG_INFORMATION_TYPE },
	uCategory	=> { type => SCALAR, optional => 1, default => 0 },
	uEventID	=> { type => SCALAR, optional => 1, default => 1 },
	sData		=> { type => SCALAR, optional => 1, default => undef } 
    });
    my %reportHash = (
    	Source => $args{sSource},
	EventType => $args{uEventType},
	Category => $args{uCategory},
	EventID => $args{uEventID},
	Data => $args{sData},
	Strings => $args{sMessage}
    );

    if ($args{hEventLog}->Report(\%reportHash)) {
      return 1;
    }

    return 0;
}

sub cifsEventLogSearch {
    my (%args) = validate (@_, {
        hEventLog	=> { type => HASHREF },
	sMessage	=> { type => SCALAR, optional => 1 },
	sSource		=> { type => SCALAR, optional => 1 },
	sEventType	=> { type => SCALAR, optional => 1 },
	sCategory	=> { type => SCALAR, optional => 1 },
	uEventID	=> { type => SCALAR, optional => 1 },
	sData		=> { type => SCALAR, optional => 1 }
    });

    if (keys %args < 2) {
      $! = 1; # Incorrect function
      return 0;
    }

    my $numEntries = cifsEventLogNumEntries(hEventLog=>$args{hEventLog});
    my $oldestIndex = cifsEventLogOldestIndex(hEventLog=>$args{hEventLog});
    my $hashRef;
    my %hashCopy;
    my $currId;
    my @results;

    for ($currId = $oldestIndex; $currId <= $numEntries; $currId++) {
      if ($args{hEventLog}->Read(EVENTLOG_FORWARDS_READ|EVENTLOG_SEEK_READ,
                                 $currId,
	  		         $hashRef)) {
        Win32::EventLog::GetMessageText($hashRef);
        # We have an entry, does it match?
	if ($args{sMessage}) { ($hashRef->{Message} !~ /$args{sMessage}/mi) ? next : 1; }
	if ($args{sSource}) { ($hashRef->{Source} !~ /$args{sSource}/mi) ? next : 1; }
	if ($args{uEventType}) { ($hashRef->{EventType} !~ /$args{sEventType}/mi) ? next : 1; }
	if ($args{uCategory}) { ($hashRef->{Category} !~ /$args{sCategory}/mi) ? next : 1; }
	if ($args{uEventID}) { ($hashRef->{EventID} !~ /$args{uEventID}/mi) ? next : 1; }
	if ($args{sData}) { ($hashRef->{Data} !~ /$args{sData}/mi) ? next : 1; }
	%hashCopy = %$hashRef;
	push (@results, \%hashCopy);
      }
    }

    (@results > 0) ? return @results : return ();
}

1;
