# $Id$

## @summary Library for CIFS file operations from NATE on Windows
## @author mueller@netapp.com
## @pod here

=head1 NAME

cifsFileCreate

=head1 SYNOPSIS

    use CIFS::FileOp qw(:all);

=head1 ROUTINES

=cut

package CIFS::FileOp;
use strict;

# Sorry about your namespace, but it's about to be overloaded. :-(
# Most of these are constants defined in Win32API::File that will
# need to be used for some of the advanced functionality in the
# exported functions.

BEGIN {
  use Module::Common;
  Module::Common::setup_export_tags(
    ops => [ qw(cifsFileCreate
                cifsFileClose
                cifsFileCopy
                cifsFileDelete
                cifsFileRename
		cifsFileRead
		cifsFileWrite
		cifsFileGetSize
		cifsFileGetType
		cifsFileGetAttributes
		cifsFileSetAttributes
		cifsFileSetSparse
		cifsFileSetFilePointer
		cifsFileCreateDirectory
		cifsFileDeleteDirectory
		cifsFileGetFreeSpace) ],
    constant_file => [ qw(FILE_READ_DATA
                          FILE_WRITE_DATA
			  FILE_APPEND_DATA
			  FILE_READ_EA
			  FILE_WRITE_EA
			  FILE_READ_ATTRIBUTES
			  FILE_WRITE_ATTRIBUTES
			  FILE_GENERIC_READ
			  FILE_GENERIC_WRITE
			  FILE_GENERIC_EXECUTE
			  FILE_CREATE_PIPE_INSTANCE
			  FILE_TRAVERSE
			  FILE_LIST_DIRECTORY
			  FILE_ADD_FILE
			  FILE_ADD_SUBDIRECTORY
			  FILE_EXECUTE
			  FILE_DELETE_CHILD
			  FILE_ALL_ACCESS) ],
    constant_file_attribute => [ qw(FILE_ATTRIBUTE_ARCHIVE
                                    FILE_ATTRIBUTE_COMPRESSED
				    FILE_ATTRIBUTE_HIDDEN
				    FILE_ATTRIBUTE_NORMAL
				    FILE_ATTRIBUTE_OFFLINE
				    FILE_ATTRIBUTE_READONLY
				    FILE_ATTRIBUTE_SYSTEM
					FILE_ATTRIBUTE_SPARSE_FILE
				    FILE_ATTRIBUTE_TEMPORARY) ],
    constant_file_flag => [ qw(FILE_FLAG_BACKUP_SEMANTICS
                               FILE_FLAG_DELETE_ON_CLOSE
			       FILE_FLAG_NO_BUFFERING
			       FILE_FLAG_OVERLAPPED
			       FILE_FLAG_POSIX_SEMANTICS
			       FILE_FLAG_RANDOM_ACCESS
			       FILE_FLAG_SEQUENTIAL_SCAN
			       FILE_FLAG_WRITE_THROUGH) ],
    constant_file_share => [ qw(FILE_SHARE_READ
                                FILE_SHARE_WRITE
				FILE_SHARE_DELETE) ],
    constant_file_seek => [ qw(FILE_BEGIN
	                       FILE_CURR
			       FILE_END) ],
    constant_file_create => [ qw(CREATE_ALWAYS
                                 CREATE_NEW
			         OPEN_ALWAYS
			         OPEN_EXISTING
			         TRUNCATE_EXISTING) ]
  );
}

use Params::Validate v0.70 qw(validate SCALAR SCALARREF HASHREF UNDEF);
use Carp qw(croak);
use base qw(Exporter);
use Win32::File qw(GetAttributes SetAttributes ARCHIVE COMPRESSED DIRECTORY HIDDEN NORMAL OFFLINE READONLY SYSTEM TEMPORARY);
use Win32API::File 0.08 qw( :ALL );
use CIFS::Misc qw(:all);

use constant ENCRYPTED => 0x4000;
use constant NOINDEX => 0x2000;
use constant REPARSE => 0x400;
use constant SPARSE => 0x200;
use constant FSCTL_SET_SPARSE => 0x900c4;

=head2 cifsFileCreate(%opts)

=over

=item Arguments

=over

=item filename

(Required) The name of the file (including the path) that you want to open or create.

=item bitsCreateMode

(Optional) A flag that determines the manner in which to create the file.
Default = ke (OPEN_EXISTING)
Possible Values = kc (OPEN_ALWAYS), ke (OPEN_EXISTING), tc (TRUNCATE_EXISTING),
te (CREATE_ALWAYS), nc (CREATE_NEW)

=item bitsFileAccess

(Optional) Requested access level(s) for creating this file.
Default = r (GENERIC_READ)

=item bitsSharedAccess

(Optional) A flag that determines how this file is shared with other processes.
Default = rw (FILE_SHARE_READ and FILE_SHARE_WRITE)

=item bitsAttributes

(Optional) A flag that determines which file attributes to use when creating this file.
Default = a (FILE_ATTRIBUTE_ARCHIVE)

=item bitsFlags

(Optional) Special options for creating this file.
Default = 0

=item secAttributes

(Optional) A security attribute data structure packed into a string.
Default = NULL

=item hModel

(Optional) An open handle to an existing file from which to mimic access when
performing this operation.
Default = 0

=back

=item Examples

<insert a usage example here>
         
=item Dependencies

A MinWin-compatible Windows client

=item Notes

This command must be run from NATE running on Windows, it cannot be run from
a UNIX host running NATE. 

=back

=for html 
<hr> </hr>

=cut

sub cifsFileCreate {
    my (%args) = validate (@_, {
	sFilename	=> { type => SCALAR },
	sCreateMode	=> { type => SCALAR, optional => 1, default => "ke" },
	sFileAccess	=> { type => SCALAR, optional => 1, default => "r" },
	sSharedAccess	=> { type => SCALAR, optional => 1, default => "rw" },
	sAttributes	=> { type => SCALAR, optional => 1, default => "a" },
	uFlags	=> { type => SCALAR, optional => 1, default => 0 },
	pSecAttributes	=> { type => SCALAR, optional => 1, default => [] },
	hModelFile	=> { type => SCALAR, optional => 1, default => 0 }
       });
    my %params = (
        Flags => $args{uFlags},
	Attributes => $args{sAttributes},
	Security => $args{pSecAttributes},
	Model => $args{hModelFile},
	Access => $args{sFileAccess},
    # TODO: A sCreateMode of 'te' (CREATE_ALWAYS) currently returns a
    # 'parameter incorrect' error.  Need to determine the cause of this.
    # Meanwhile, we'll throw in the CREATE_ALWAYS() param in this case.
	Create => ($args{sCreateMode} eq "te") ? CREATE_ALWAYS() : $args{sCreateMode},
	Share => $args{sSharedAccess}
	);
    my $handle;

    # createFile is a built-in wrapper around CreateFile that supports the
    # letter-to-bits attributes that we (and hopefully you) like.
    if ($handle = createFile($args{sFilename}, \%params)) {
      return $handle;
    } else {
      return undef;
    }
}

sub cifsFileClose {
    my (%args) = validate (@_, {
        hFile	=> { type => SCALAR }
    });

    if (CloseHandle($args{hFile})) {
      return 1;
    } else {
      return 0;
    }
}

sub cifsFileCopy {
    my (%args) = validate (@_, {
        sFileSource	=> { type => SCALAR },
	sFileDest	=> { type => SCALAR },
	bForce	=> { type => SCALAR, optional => 1, default => 0 }
    });

    if (CopyFile($args{sFileSource}, $args{sFileDest}, !$args{bForce})) {
      return 1;
    } else {
      return 0;
    }
}

sub cifsFileDelete {
    my (%args) = validate (@_, {
        sFilename	=> { type => SCALAR }
    });

    if (DeleteFile($args{sFilename})) {
      return 1;
    } else {
      return 0;
    }
}

sub cifsFileRename {
    my (%args) = validate (@_, {
        sFileSource	=> { type => SCALAR },
	sFileDest	=> { type => SCALAR },
	bForce	=> { type => SCALAR, optional => 1, default => 0 }
    });

    if (MoveFileEx($args{sFileSource}, $args{sFileDest}, $args{bForce} ? MOVEFILE_REPLACE_EXISTING() : 0)) {
      return 1;
    } else {
      return 0;
    }
}

sub cifsFileRead {
    my (%args) = validate (@_, {
        sFilename	=> { type => SCALAR, optional => 1, default => () },
	hFile		=> { type => SCALAR, optional => 1, default => 0 },
	uNumBytes	=> { type => SCALAR },
	pOverlapped	=> { type => SCALAR, optional => 1, default => [] }
    });
    my $bytes;
    my $hFile = $args{hFile};
    my $bOpenedFile = 0;

    if ((!$hFile && !$args{sFilename}) || ($hFile && $args{sFilename})) {
      $! = 1; # Incorrect function
      return -1;
    }

    if (!$hFile) {
      $hFile = createFile($args{sFilename}, 'r');
      $bOpenedFile = 1;
    }

    if ($hFile) {
      if (ReadFile($hFile, $bytes, $args{uNumBytes}, [], $args{pOverlapped})) {
        if ($bOpenedFile) { cifsFileClose(hFile => $hFile); }
        return $bytes;
      }
    }
    if ($bOpenedFile) { cifsFileClose(hFile => $hFile); }
    return -1;
}

sub cifsFileWrite {
    my (%args) = validate (@_, {
        sFilename	=> { type => SCALAR, optional => 1, default => () },
	hFile		=> { type => SCALAR, optional => 1, default => 0 },
	sBytes		=> { type => SCALAR },
	pOverlapped	=> { type => SCALAR, optional => 1, default => [] }
    });
    my $bytesWritten;
    my $hFile = $args{hFile};
    my $bOpenedFile = 0;

    if ((!$hFile && !$args{sFilename}) || ($hFile && $args{sFilename})) {
      $^E = 1; # Incorrect function
      return -1;
    }

    if (!$hFile) {
      $hFile = createFile($args{sFilename}, 'rwe');
      if ($hFile) { $bOpenedFile = 1; }
    }

    if ($hFile) {
      if (WriteFile($hFile, $args{sBytes}, 0, $bytesWritten, $args{pOverlapped})) {
        if ($bOpenedFile) { cifsFileClose(hFile => $hFile); }
        return $bytesWritten;
      }
    }
    if ($bOpenedFile) { cifsFileClose(hFile => $hFile); }
    return -1;
}

sub cifsFileGetSize {
    my (%args) = validate (@_, {
        sFilename	=> { type => SCALAR }
    });
    my $filesize;

    if (! -e $args{sFilename}) {
      $^E = 2; # The system cannot find the file specified
      return -1;
    }

    if ((undef, undef, undef, undef,
	 undef, undef, undef, $filesize,
	 undef, undef, undef, undef, undef) = stat($args{sFilename})) {
      return $filesize;
    }

    return -1;
}

sub cifsFileGetType {
    my (%args) = validate (@_, {
        sFilename	=> { type => SCALAR }
    });
    my $hFile;
    my $filetype;

    if ($hFile = cifsFileCreate(sFilename => $args{sFilename}, sCreateMode => OPEN_EXISTING)) {
      if ($filetype = GetFileType($hFile)) {
        cifsFileClose(hFile => $hFile);
	if ($filetype eq FILE_TYPE_DISK) {
          return "DISK";
	} elsif ($filetype eq FILE_TYPE_CHAR) {
          return "CHAR";
	} elsif ($filetype eq FILE_TYPE_PIPE) {
          return "PIPE";
	} else {
          return "UNKNOWN";
	}
      }
      cifsFileClose(hFile => $hFile);
    }

    return 0;
}

sub cifsFileGetAttributes {
    my (%args) = validate (@_, {
        sFilename	=> { type => SCALAR }
    });
    my $attr;
    my $attrLetters;

    if (! -e $args{sFilename}) {
      $^E = 2; # The system cannot find the file specified
      return 0;
    }

    if (GetAttributes($args{sFilename}, $attr)) {
      if ($attr & ARCHIVE) { $attrLetters .= "a"; }
      if ($attr & COMPRESSED) { $attrLetters .= "c"; }
      if ($attr & DIRECTORY) { $attrLetters .= "d"; }
      if ($attr & HIDDEN) { $attrLetters .= "h"; }
      if ($attr & NORMAL) { $attrLetters .= "n"; }
      if ($attr & OFFLINE) { $attrLetters .= "o"; }
      if ($attr & READONLY) { $attrLetters .= "r"; }
      if ($attr & SYSTEM) { $attrLetters .= "m"; }
      if ($attr & TEMPORARY) { $attrLetters .= "t"; }
      if ($attr & ENCRYPTED) { $attrLetters .= "e"; }
      if ($attr & NOINDEX) { $attrLetters .= "i"; }
      if ($attr & REPARSE) { $attrLetters .= "p"; }
      if ($attr & SPARSE) { $attrLetters .= "s"; }
      return $attrLetters;
    }
    return 0;
}

sub cifsFileSetAttributes {
    my (%args) = validate (@_, {
      sFilename		=> { type => SCALAR },
      sAttributes	=> { type => SCALAR }
    });
    my $attr;

    if (! -e $args{sFilename}) {
      $! = 2; # The system cannot find the file specified
      return 0;
    }

    if ($args{sAttributes}) {
      if ($args{sAttributes} =~ /a/) { $attr |= ARCHIVE; }
      if ($args{sAttributes} =~ /c/) { $attr |= COMPRESSED; }
      if ($args{sAttributes} =~ /d/) { $attr |= DIRECTORY; }
      if ($args{sAttributes} =~ /h/) { $attr |= HIDDEN; }
      if ($args{sAttributes} =~ /n/) { $attr |= NORMAL; }
      if ($args{sAttributes} =~ /o/) { $attr |= OFFLINE; }
      if ($args{sAttributes} =~ /r/) { $attr |= READONLY; }
      if ($args{sAttributes} =~ /m/) { $attr |= SYSTEM; }
      if ($args{sAttributes} =~ /t/) { $attr |= TEMPORARY; }
      if ($args{sAttributes} =~ /e/) { $attr |= ENCRYPTED; }
      if ($args{sAttributes} =~ /i/) { $attr |= NOINDEX; }
      if ($args{sAttributes} =~ /p/) { $attr |= REPARSE; }
      if ($args{sAttributes} =~ /s/) { $attr |= SPARSE; }
      if (SetAttributes($args{sFilename}, $attr)) {
        return 1;
      }
    }
    return 0;
}

sub cifsFileSetSparse {
    my (%args) = validate (@_, {
        hFile		=> { type => SCALAR },
	});
    my $bytes;
    my $hFile = $args{hFile};
    
    if (DeviceIoControl($hFile, FSCTL_SET_SPARSE(), [], 0, [], 0, $bytes, [])) {
      return 1;
    }

    return 0;
}

sub cifsFileSetFilePointer {
    my (%args) = validate (@_, {
	hFile		=> { type => SCALAR },
	uOffset		=> { type => SCALAR },
	sFromWhere	=> { type => SCALAR },
	});
    
    if (SetFilePointer($args{hFile}, $args{uOffset}, [], $args{sFromWhere})) {
        return 1;
    }	    

    return 0;
}

sub cifsFileCreateDirectory {
    my (%args) = validate (@_, {
      sPathname		=> { type => SCALAR }
    });
    
    mkdir $args{sPathname} || return 0;
    return 1;
}

sub cifsFileDeleteDirectory {
    my (%args) = validate (@_, {
      sPathname		=> { type => SCALAR }
    });
    
    rmdir $args{sPathname} || return 0;
    return 1;
}

sub cifsFileGetFreeSpace {
    my (%args) = validate (@_, {
      cDriveLetter	=> { type => SCALAR }
    });

    my $path = $args{cDriveLetter} . ":/";
    if (! -e $path) {
      $! = 2; # The system cannot find the file specified
      return -1;
    }
    
    my ($bytes) = (`cmd /c dir $args{cDriveLetter}:` =~ /([\d,]+) bytes free/);
    $bytes =~ s/,//g;
    return $bytes;
}

1;
