# -*- perl -*-
# $Id$
#
# Copyright (c) 2004 Network Appliance Inc.
# All right reserved.
#
## @summary Routines for setting up ldap options, sending the queries and wcc
## command to the ldap server
##
## @description Some functions are used to setup ldap options for different
## kinds of ldap servers like OpenLDAP, Novell e-Directory and windows ldap
## servers. Some functions are used to send the queries to the ldap server,
## after getting back the results from the ldap server, it will extract them and
## returns to caller functions. Some functions are used to send the wcc command
## with different kinds of options to the filer. After getting back the results,
## they would work on, extract the results and then return to caller functions
##
## @dependencies
##
## use Tharn;    
##
##
## @author dl-maryv (FSS Protocols QA)
##
## @change 2005-06-30 huynhle: add the function addomain_conf
##
##

package CIFS::LdapUtils;
require Exporter;
use strict;
use Tharn;
use SayWrap qw(sayEOF);
use filer qw(filer_options_set);
use vars qw (@ISA @EXPORT_OK);
@ISA = qw(Exporter);
@EXPORT_OK = qw(openldap_conf
                edirldap_conf
                addomain_conf
                getpw_byname
                getpw_byuid
                getgr_byname
                getgr_bygid
                getgr_list
                wcc_u_uname
                wcc_s_dm_ntname
                wcc_s_ntname);
                
## @name ::openldap_conf
##
## @summary this function configures all the ldap options so the filer can
## connect to OpenLDAP server. It gets ldap_base, ldap_name, ldap_passwd and
## ldap_servers as arguments. Other ldap options are set to default values as
## below.
##
## @arg $fc, %options
## $fc - required. filer connection object
## %options - required. hash of arguments
##            ldap_base ==> string - name of the base
##            ldap_name ==> string - name of ldap server's administrator
##            ldap_passwd ==> string - passwd of ldap server's administrator
##            ldap_servers ==> IP address - IP address of the ldap server
## example: openldap_conf($fc, ldap_base => "dc=netapp,dc=com", 
##                             ldap_name => "cn=admin,dc=netapp,dc=com",
##		               ldap_passwd => "hello", 
##                             ldap_servers => "172.23.100.14");
##

sub openldap_conf
{
    my($fc, %options) = @_;

    my %ldap_options = ();

    # Initialize the options strings for all the LDAP options
    my $ldap_addomain = "ldap.ADdomain";
    my $ldap_base = "ldap.base";
    my $ldap_base_group = "ldap.base.group";
    my $ldap_base_netgroup = "ldap.base.netgroup";
    my $ldap_base_passwd = "ldap.base.passwd";
    my $ldap_enable = "ldap.enable";
    my $ldap_name = "ldap.name";
    my $ldap_attribute_gecos = "ldap.nssmap.attribute.gecos";
    my $ldap_attribute_gidnumber = "ldap.nssmap.attribute.gidNumber";
    my $ldap_attribute_groupname = "ldap.nssmap.attribute.groupname";
    my $ldap_attribute_homeDirectory = "ldap.nssmap.attribute.homeDirectory";
    my $ldap_attribute_loginShell = "ldap.nssmap.attribute.loginShell";
    my $ldap_attribute_memberNisNetgroup = "ldap.nssmap.attribute" .
                                           ".memberNisNetgroup";
    my $ldap_attribute_memberUid = "ldap.nssmap.attribute.memberUid";
    my $ldap_attribute_netgroupname = "ldap.nssmap.attribute.netgroupname";
    my $ldap_attribute_nisNetgroupTriple = "ldap.nssmap.attribute" .
                                           ".nisNetgroupTriple";
    my $ldap_attribute_uid = "ldap.nssmap.attribute.uid";
    my $ldap_attribute_uidNumber = "ldap.nssmap.attribute.uidNumber";
    my $ldap_attribute_userPassword = "ldap.nssmap.attribute.userPassword";
    my $ldap_objectClass_nisNetgroup = "ldap.nssmap.objectClass.nisNetgroup";
    my $ldap_objectClass_posixAccount = "ldap.nssmap.objectClass.posixAccount";
    my $ldap_objectClass_posixGroup = "ldap.nssmap.objectClass.posixGroup";
    my $ldap_passwd = "ldap.passwd";
    my $ldap_port = "ldap.port";
    my $ldap_servers = "ldap.servers";
    my $ldap_servers_preferred = "ldap.servers.preferred";
    my $ldap_ssl_enable = "ldap.ssl.enable";
    my $ldap_usermap_enable = "ldap.usermap.enable";

    # Set all the default values for some of LDAP options
    $ldap_options{$ldap_addomain} = "\"\"";
    $ldap_options{$ldap_base_group} = "\"\"";
    $ldap_options{$ldap_base_netgroup} = "\"\"";
    $ldap_options{$ldap_base_passwd} = "\"\"";
    $ldap_options{$ldap_enable} = 'on';
    $ldap_options{$ldap_attribute_gecos} = 'gecos';
    $ldap_options{$ldap_attribute_gidnumber} = 'gidNumber';
    $ldap_options{$ldap_attribute_groupname} = 'cn';
    $ldap_options{$ldap_attribute_homeDirectory} = 'homeDirectory';
    $ldap_options{$ldap_attribute_loginShell} = 'loginShell';
    $ldap_options{$ldap_attribute_memberNisNetgroup} = 'memberNisNetgroup';
    $ldap_options{$ldap_attribute_memberUid} = 'memberUid';
    $ldap_options{$ldap_attribute_netgroupname} = 'cn';
    $ldap_options{$ldap_attribute_nisNetgroupTriple} = 'nisNetgroupTriple';
    $ldap_options{$ldap_attribute_uid} = 'uid';
    $ldap_options{$ldap_attribute_uidNumber} = 'uidNumber';
    $ldap_options{$ldap_attribute_userPassword} = 'userPassword';
    $ldap_options{$ldap_objectClass_nisNetgroup} = 'nisNetgroup';
    $ldap_options{$ldap_objectClass_posixAccount} = 'posixAccount';
    $ldap_options{$ldap_objectClass_posixGroup} = 'posixGroup';
    $ldap_options{$ldap_port} = 389;
    $ldap_options{$ldap_servers_preferred} = "\"\"";
    $ldap_options{$ldap_ssl_enable} = 'off';
    $ldap_options{$ldap_usermap_enable} = 'off';
    
    if(defined($options{ldap_base})){
        $ldap_options{$ldap_base} = $options{ldap_base};
    }

    if(defined($options{ldap_name})){
        $ldap_options{$ldap_name} = $options{ldap_name};
    }
    
    if(defined($options{ldap_passwd})){
        $ldap_options{$ldap_passwd} = $options{ldap_passwd};
    }
    
    if(defined($options{ldap_servers})){
        $ldap_options{$ldap_servers} = $options{ldap_servers};
    }

    foreach my $option_key (keys %ldap_options){
        my $option_value = $ldap_options{$option_key};
        sayEOF($fc, "options $option_key $option_value");
        #filer_options_set($fc, $option_key, $option_value);
    }
    
    return 1;
    
}



## @name ::edirldap_conf
##
## @summary this function configures all the ldap options so the filer can
## connect to edir ldap novell server. It gets ldap_base, ldap_name, ldap_passwd
## and ldap_servers as arguments. Other ldap options are set to default values
## as below.
##
## @arg $fc, %options
## $fc - required. filer connection object
## %options - required. hash of arguments
##            ldap_base ==> string - name of the base
##            ldap_name ==> string - name of ldap server's administrator
##            ldap_passwd ==> string - passwd of ldap server's administrator
##            ldap_servers ==> IP address - IP address of the ldap server
## example: edirldap_conf($fc, ldap_base => "o=org", 
##                             ldap_name => "cn=admin,o=org",
##		               ldap_passwd => "hello", 
##                             ldap_servers => "172.26.2.125");
##

sub edirldap_conf
{
    my($fc, %options) = @_;

    my %ldap_options = ();

    # Initialize the options strings for all the LDAP options
    my $ldap_addomain = "ldap.ADdomain";
    my $ldap_base = "ldap.base";
    my $ldap_base_group = "ldap.base.group";
    my $ldap_base_netgroup = "ldap.base.netgroup";
    my $ldap_base_passwd = "ldap.base.passwd";
    my $ldap_enable = "ldap.enable";
    my $ldap_name = "ldap.name";
    my $ldap_attribute_gecos = "ldap.nssmap.attribute.gecos";
    my $ldap_attribute_gidnumber = "ldap.nssmap.attribute.gidNumber";
    my $ldap_attribute_groupname = "ldap.nssmap.attribute.groupname";
    my $ldap_attribute_homeDirectory = "ldap.nssmap.attribute.homeDirectory";
    my $ldap_attribute_loginShell = "ldap.nssmap.attribute.loginShell";
    my $ldap_attribute_memberNisNetgroup = "ldap.nssmap.attribute" .
                                           ".memberNisNetgroup";
    my $ldap_attribute_memberUid = "ldap.nssmap.attribute.memberUid";
    my $ldap_attribute_netgroupname = "ldap.nssmap.attribute.netgroupname";
    my $ldap_attribute_nisNetgroupTriple = "ldap.nssmap.attribute" .
                                           ".nisNetgroupTriple";
    my $ldap_attribute_uid = "ldap.nssmap.attribute.uid";
    my $ldap_attribute_uidNumber = "ldap.nssmap.attribute.uidNumber";
    my $ldap_attribute_userPassword = "ldap.nssmap.attribute.userPassword";
    my $ldap_objectClass_nisNetgroup = "ldap.nssmap.objectClass.nisNetgroup";
    my $ldap_objectClass_posixAccount = "ldap.nssmap.objectClass.posixAccount";
    my $ldap_objectClass_posixGroup = "ldap.nssmap.objectClass.posixGroup";
    my $ldap_passwd = "ldap.passwd";
    my $ldap_port = "ldap.port";
    my $ldap_servers = "ldap.servers";
    my $ldap_servers_preferred = "ldap.servers.preferred";
    my $ldap_ssl_enable = "ldap.ssl.enable";
    my $ldap_usermap_enable = "ldap.usermap.enable";

    # Set all the default values for some of LDAP options
    $ldap_options{$ldap_addomain} = "\"\"";
    $ldap_options{$ldap_base_group} = "\"\"";
    $ldap_options{$ldap_base_netgroup} = "\"\"";
    $ldap_options{$ldap_base_passwd} = "\"\"";
    $ldap_options{$ldap_enable} = 'on';
    $ldap_options{$ldap_attribute_gecos} = 'gecos';
    $ldap_options{$ldap_attribute_gidnumber} = 'gidNumber';
    $ldap_options{$ldap_attribute_groupname} = 'cn';
    $ldap_options{$ldap_attribute_homeDirectory} = 'homeDirectory';
    $ldap_options{$ldap_attribute_loginShell} = 'loginShell';
    $ldap_options{$ldap_attribute_memberNisNetgroup} = 'memberNisNetgroup';
    $ldap_options{$ldap_attribute_memberUid} = 'memberUid';
    $ldap_options{$ldap_attribute_netgroupname} = 'cn';
    $ldap_options{$ldap_attribute_nisNetgroupTriple} = 'nisNetgroupTriple';
    $ldap_options{$ldap_attribute_uid} = 'uid';
    $ldap_options{$ldap_attribute_uidNumber} = 'uidNumber';
    $ldap_options{$ldap_attribute_userPassword} = 'userPassword';
    $ldap_options{$ldap_objectClass_nisNetgroup} = 'nisNetgroup';
    $ldap_options{$ldap_objectClass_posixAccount} = 'posixAccount';
    $ldap_options{$ldap_objectClass_posixGroup} = 'posixGroup';
    $ldap_options{$ldap_port} = 389;
    $ldap_options{$ldap_servers_preferred} = "\"\"";
    $ldap_options{$ldap_ssl_enable} = 'off';
    $ldap_options{$ldap_usermap_enable} = 'off';
    
    if(defined($options{ldap_base})){
        $ldap_options{$ldap_base} = $options{ldap_base};
    }

    if(defined($options{ldap_name})){
        $ldap_options{$ldap_name} = $options{ldap_name};
    }
    
    if(defined($options{ldap_passwd})){
        $ldap_options{$ldap_passwd} = $options{ldap_passwd};
    }
    
    if(defined($options{ldap_servers})){
        $ldap_options{$ldap_servers} = $options{ldap_servers};
    }

    foreach my $option_key (keys %ldap_options){
        my $option_value = $ldap_options{$option_key};
        sayEOF($fc, "options $option_key $option_value");
        #filer_options_set($fc, $option_key, $option_value);
    }
    
    return 1;
    
}

## @name ::addomain_conf
##
## @summary this function configures all the ldap options so the filer can
## connect to Windows AD server which holds Unix users information. It gets
## ldap_addomain, ldap_base, ldap_name, ldap_passwd and ldap_usermap_base as
## arguments. Other ldap options are set to default values as below.
##
## @arg $fc, %options
## $fc - required. filer connection object
## %options - required. hash of arguments
##            ldap_addomain ==> string - name of Windows AD domain
##            ldap_base ==> string - name of the base
##            ldap_name ==> string - name of ldap server's administrator
##            ldap_passwd ==> string - passwd of ldap server's administrator
##            ldap_usermap_base ==> string - name of the base to start mapping
##                                           the users
## example: addomain_conf($fc,
##  ldap_addomain ==> "linux.sa.shell.com"
##  ldap_base => "dc=linux,dc=sa,dc=shell,dc=com", 
##  ldap_name => "cn=administrator,cn=users,dc=linux,dc=sa,dc=shell,dc=com",
##  ldap_passwd => "hello",
##  ldap_usermap_base ==> "ou=automationusers,dc=linux,dc=sa,dc=shell,dc=com);
##

sub addomain_conf
{
    my($fc, %options) = @_;

    my %ldap_options = ();

    # Initialize the options strings for all the LDAP options
    my $ldap_addomain = "ldap.ADdomain";
    my $ldap_base = "ldap.base";
    my $ldap_base_group = "ldap.base.group";
    my $ldap_base_netgroup = "ldap.base.netgroup";
    my $ldap_base_passwd = "ldap.base.passwd";
    my $ldap_enable = "ldap.enable";
    my $ldap_name = "ldap.name";
    my $ldap_attribute_gecos = "ldap.nssmap.attribute.gecos";
    my $ldap_attribute_gidnumber = "ldap.nssmap.attribute.gidNumber";
    my $ldap_attribute_groupname = "ldap.nssmap.attribute.groupname";
    my $ldap_attribute_homeDirectory = "ldap.nssmap.attribute.homeDirectory";
    my $ldap_attribute_loginShell = "ldap.nssmap.attribute.loginShell";
    my $ldap_attribute_memberNisNetgroup = "ldap.nssmap.attribute" .
                                           ".memberNisNetgroup";
    my $ldap_attribute_memberUid = "ldap.nssmap.attribute.memberUid";
    my $ldap_attribute_netgroupname = "ldap.nssmap.attribute.netgroupname";
    my $ldap_attribute_nisNetgroupTriple = "ldap.nssmap.attribute" .
                                           ".nisNetgroupTriple";
    my $ldap_attribute_uid = "ldap.nssmap.attribute.uid";
    my $ldap_attribute_uidNumber = "ldap.nssmap.attribute.uidNumber";
    my $ldap_attribute_userPassword = "ldap.nssmap.attribute.userPassword";
    my $ldap_objectClass_nisNetgroup = "ldap.nssmap.objectClass.nisNetgroup";
    my $ldap_objectClass_posixAccount = "ldap.nssmap.objectClass.posixAccount";
    my $ldap_objectClass_posixGroup = "ldap.nssmap.objectClass.posixGroup";
    my $ldap_passwd = "ldap.passwd";
    my $ldap_port = "ldap.port";
    my $ldap_servers = "ldap.servers";
    my $ldap_servers_preferred = "ldap.servers.preferred";
    my $ldap_ssl_enable = "ldap.ssl.enable";
    my $ldap_usermap_att_unixaccount = "ldap.usermap.attribute.unixaccount";
    my $ldap_usermap_att_windowsaccount = "ldap.usermap.attribute" . 
                                          ".windowsaccount";
    my $ldap_usermap_base = "ldap.usermap.base";
    my $ldap_usermap_enable = "ldap.usermap.enable";

    # Set all the default values for some of LDAP options
    $ldap_options{$ldap_base_group} = "\"\"";
    $ldap_options{$ldap_base_netgroup} = "\"\"";
    $ldap_options{$ldap_base_passwd} = "\"\"";
    $ldap_options{$ldap_enable} = 'on';
    $ldap_options{$ldap_attribute_gecos} = 'gecos';
    $ldap_options{$ldap_attribute_gidnumber} = 'gidNumber';
    $ldap_options{$ldap_attribute_groupname} = 'cn';
    $ldap_options{$ldap_attribute_homeDirectory} = 'homeDirectory';
    $ldap_options{$ldap_attribute_loginShell} = 'loginShell';
    $ldap_options{$ldap_attribute_memberNisNetgroup} = 'memberNisNetgroup';
    $ldap_options{$ldap_attribute_memberUid} = 'memberUid';
    $ldap_options{$ldap_attribute_netgroupname} = 'cn';
    $ldap_options{$ldap_attribute_nisNetgroupTriple} = 'nisNetgroupTriple';
    $ldap_options{$ldap_attribute_uid} = 'sAMAccountName';
    $ldap_options{$ldap_attribute_uidNumber} = 'uidNumber';
    $ldap_options{$ldap_attribute_userPassword} = 'userPassword';
    $ldap_options{$ldap_objectClass_nisNetgroup} = 'nisNetgroup';
    $ldap_options{$ldap_objectClass_posixAccount} = 'User';
    $ldap_options{$ldap_objectClass_posixGroup} = 'Group';
    $ldap_options{$ldap_port} = 389;
    $ldap_options{$ldap_servers} = "\"\"";
    $ldap_options{$ldap_servers_preferred} = "\"\"";
    $ldap_options{$ldap_ssl_enable} = 'off';
    $ldap_options{$ldap_usermap_att_unixaccount} = 'sAMAccountName';
    $ldap_options{$ldap_usermap_att_windowsaccount} =
                                                'shellLNXWindowsAccountName';
    $ldap_options{$ldap_usermap_enable} = 'on';
    
    if(defined($options{ldap_addomain})){
        $ldap_options{$ldap_addomain} = $options{ldap_addomain};
    }
    if(defined($options{ldap_base})){
        $ldap_options{$ldap_base} = $options{ldap_base};
    }

    if(defined($options{ldap_name})){
        $ldap_options{$ldap_name} = $options{ldap_name};
    }
    
    if(defined($options{ldap_passwd})){
        $ldap_options{$ldap_passwd} = $options{ldap_passwd};
    }
    
    if(defined($options{ldap_usermap_base})){
        $ldap_options{$ldap_usermap_base} = $options{ldap_usermap_base};
    }

    foreach my $option_key (keys %ldap_options){
        my $option_value = $ldap_options{$option_key};
        sayEOF($fc, "options $option_key $option_value");
        #filer_options_set($fc, $option_key, $option_value);
    }
    
    return 1;    
}

## @name ::getpw_byname
##
## @summary this function runs the "getXXbyYY getpwbyname_r <user_name>"
## command. After gets back the result from the ldap server, it will extract all
## different values username, password, uid, gid, gecos, home directory and
## shell from the output and return them in a list
##
## @arg $fc, $username
## $fc - required. Filer connection object
## $username - required. The user name
##
## example: getpw_byname($fc, <user_name>)
##
##

sub getpw_byname
{
    my($fc, $username) = @_;
    my $result = sayEOF($fc,
                        "priv set advanced; getXXbyYY getpwbyname_r $username");
    
    my @list_rt;
    my $temp;

    if($result =~ /Could not get/){
        die "Can't get the query result from ldap server";
    }else{
        # extract the user name and push it in the list
        $result =~ /pw_name\s=\s([ \w]+)/;
        $temp = $1;
        push(@list_rt, $temp);

        # extract the password and push it in the list
        $result =~ /pw_passwd\s=\s(\S+)/;
        $temp = $1;
        push(@list_rt, $temp);
    
        # extract the uid and push it in the list
        $result =~ /pw_uid\s=\s(\d+)/;
        $temp = $1;
        push(@list_rt, $temp);
    
        # extract the gid and push it in the list
        $result =~ /pw_gid\s=\s(\d+)/;
        $temp = $1;
        push(@list_rt, $temp);
    
        # extract the gecos and push it in the list
        $result =~ /pw_gecos\s=\s([ \w]+)/;
        $temp = $1;
        push(@list_rt, $temp);
    
        # extract the home directory and push it in the list
        $result =~ /pw_dir\s=\s([\/\w]+)/;
        $temp = $1;
        push(@list_rt, $temp);
    
        # extract the shell and push it in the list
        $result =~ /pw_shell\s=\s([\/\w]+)/;
        $temp = $1;
        push(@list_rt, $temp);
    }
   
    return @list_rt;  
}

## @name ::getpw_byuid
##
## @summary this function runs the "getXXbyYY getpwbyuid_r <uid_number>" 
## command. After gets back the result from the ldap server, it will extract all
## different values username, password, uid, gid, gecos, home directory and
## shell from the output and return them in a list
##
## @arg $fc, $uid
## $fc - required. Filer connection object
## $uid - required. The uid number
##
## example: getpw_byname($fc, <uid_number>)
##
##

sub getpw_byuid
{
    my($fc, $uid) = @_;
    my $result = sayEOF($fc,
                        "priv set advanced; getXXbyYY getpwbyuid_r $uid");

    my @list_rt;
    my $temp;

    if($result =~ /Could not get/){
        die "Can't get the query result from ldap server";
    }else{
        # extract the user name and push it in the list
        $result =~ /pw_name\s=\s([ \w]+)/;
        $temp = $1;
        push(@list_rt, $temp);
    
        # extract the password and push it in the list
        $result =~ /pw_passwd\s=\s(\S+)/;
        $temp = $1;
        push(@list_rt, $temp);
    
        # extract the uid and push it in the list
        $result =~ /pw_uid\s=\s(.*),/;
        $temp = $1;
        push(@list_rt, $temp);
    
        # extract the gid and push it in the list  
        $result =~ /pw_gid\s=\s(\d+)/;
        $temp = $1;
        push(@list_rt, $temp);
    
        # extract the gecos and push it in the list
        $result =~ /pw_gecos\s=\s([ \w]+)/;
        $temp = $1;
        push(@list_rt, $temp);
   
        # extract the home directory and push it in the list
        $result =~ /pw_dir\s=\s([\/\w]+)/;
        $temp = $1;
        push(@list_rt, $temp);
    
        # extract the shell and push it in the list
        $result =~ /pw_shell\s=\s([\/\w]+)/;
        $temp = $1;
        push(@list_rt, $temp);
    }
    
    return @list_rt;
}

## @name ::getgr_byname
##
## @summary this function runs the "getXXbyYY getgrbyname <group_name>" command.
## After gets back the result from the ldap server, it will extract the group
## name and gid from the output and return them in a list
##
## @arg $fc, $grname
## $fc - required. Filer connection object
## $grname - required. the group name
##
## example: getgr_byname($fc, <group_name>)
##
##

sub getgr_byname
{
    my($fc, $grname) = @_;
    my $result = sayEOF($fc,
                        "priv set advanced; getXXbyYY getgrbyname $grname");
    
    my @list_rt;
    my $temp;
    
    if($result =~ /Could not get/){
        die "Can't get the query result from ldap server";
    }else{
        # extract the group name and push it in the list
        $result =~ /name\s=\s(\w+)/;
        $temp = $1;
        push(@list_rt, $temp);
    
        # extract the gid and push it in the list
        $result =~ /gid\s=\s(\d+)/;
        $temp = $1;
        push(@list_rt, $temp);
    }
    
    return @list_rt;
}

## @name ::getgr_bygid
##
## @summary this function runs the "getXXbyYY getgrbygid <gid_number>" command.
## After gets back the result from the ldap server, it will extract the group
## name and gid from the output and return them in a list
##
## @arg $fc, $gid
## $fc - required. Filer connection object
## $gid - required. The gid number
##
## example: getgr_bygid($fc, <gid_number>)
##
##

sub getgr_bygid
{
    my($fc, $gid) = @_;
    my $result = sayEOF($fc,
                     "priv set advanced; getXXbyYY getgrbygid $gid");
    
    my @list_rt;
    my $temp;
    
    if($result =~ /Could not get/){
        die "Can't get the query result from ldap server";
    }else{
        # extract the group name and push it in the list
        $result =~ /name\s=\s(\w+)/;
        $temp = $1;
        push(@list_rt, $temp);
    
        # extract the gid and push it in the list
        $result =~ /gid\s=\s(\d+)/;
        $temp = $1;
        push(@list_rt, $temp);
    }
    
    return @list_rt;
}

## @name ::getgr_list
##
## @summary this function runs the "getXXbyYY getgrlist <user_name>" command.
## After gets back the result from the ldap server, it will extract the user
## name, the groups names and the number of group that the user belongs to from
## the output and return them.
##
## @arg $fc, $username
## $fc - required. Filer connection object
## $gid - required. The user name
##
## example: getgr_list($fc, <user_name>)
##
##

sub getgr_list
{
    my($fc, $username) = @_;
    my $result = sayEOF($fc,
                        "priv set advanced; getXXbyYY getgrlist $username");
    
    my $u_name;
    my @groups;
    my $num_groups;
    
    if($result =~ /Could not get/){
        die "Can't get the query result from ldap server";
    }else{
        # extract the user name
        $result =~ /pw_name\s=\s(\w+)/;
        $u_name = $1;
        
        # extract all the group names and push them in the list
        $result =~ /Groups:\s([ \d]+)/;
        @groups = split /\s/, $1;
        
        # Figure out the number of all groups that the user belongs to
        $num_groups = $#groups + 1;
    }
    
    # returns user_name, number of all groups and the groups names
    return ($u_name, $num_groups, @groups)
    
}

## @name ::wcc_u_uname
##
## @summary this function runs the "wcc -u <unix_user_name>" command. After it
## gets back the result, it will extract all different values
## windows_domain_name, windows_user_name and unix_user_name from the output and
## return them.
##
## @arg $fc, $username
## $fc - required. Filer connection object
## $username - required. The unix user name
##
## example: wcc_u_uname($fc, <unix_user_name>)
##
##

sub wcc_u_uname{
    my($fc, $u_user_name) = @_;
    my $result = sayEOF($fc, "wcc -u $u_user_name");
    my $domain_name;
    my $nt_user_name;
    my $unix_user_name;
    
    if($result=~/\(NT - UNIX\) account name\(s\):\s+\((.*)\\(.*)\s+-\s+(.*)\)/){
       $domain_name = $1;
       $nt_user_name = $2;
       $unix_user_name = $3;
    }else{
       die "CIFS is not running, licensed or the unix user's name is invalid";
    }
    
    return ($domain_name, $nt_user_name, $unix_user_name);
}


## @name ::wcc_s_dm_ntname
##
## @summary this function runs the "wcc -s <windows_domain_name>\
## <windows_user_name>" command. After it gets back the result, it will extract
## all different values windows_domain_name, windows_user_name and
## unix_user_name from the output and return them.
##
## @arg $fc, $username
## $fc - required. Filer connection object
## $username - required. The windows user name in the form
## <windows_domain_name>\<windows_user_name>
##
## example: wcc_s_dm_ntname($fc, <windows_domain_name>\<windows_user_name>)
##
##

sub wcc_s_dm_ntname{
    my($fc, $domain_username) = @_;
    my $result = sayEOF($fc, "wcc -s $domain_username");
    my $domain_name;
    my $nt_user_name;
    my $unix_user_name;
    
    if($result=~/\(NT - UNIX\) account name\(s\):\s+\((.*)\\(.*)\s+-\s+(.*)\)/){
       $domain_name = $1;
       $nt_user_name = $2;
       $unix_user_name = $3;
    }else{
    die "CIFS is not running, licensed or the windows user's name is invalid";
    }
    
    return ($domain_name, $nt_user_name, $unix_user_name);
}


## @name ::wcc_s_ntname
##
## @summary this function runs the "wcc -s <windows_user_name>" command. After
## it gets back the result, it will extract all different values
## windows_domain_name, windows_user_name and unix_user_name from the output and
## return them.
##
## @arg $fc, $username
## $fc - required. Filer connection object
## $username - required. The windows user name
##
## example: wcc_s_ntname($fc, <windows_user_name>)
##
##

sub wcc_s_ntname{
    my($fc, $nt_user_name) = @_;
    my $result = sayEOF($fc, "wcc -s $nt_user_name");
    my $domain_name;
    my $nt_user_name;
    my $unix_user_name;
    
    if($result=~/\(NT - UNIX\) account name\(s\):\s+\((.*)\\(.*)\s+-\s+(.*)\)/){
       $domain_name = $1;
       $nt_user_name = $2;
       $unix_user_name = $3;
    }else{
    die "CIFS is not running, licensed or the windows user's name is invalid";
    }
    
    return ($domain_name, $nt_user_name, $unix_user_name);
}


