#Hey emacs, this is basically -*- perl -*-
# $Id: $
#
## @summary Ntfs - Library for the NTFS-only tests
## @author	dl-maryv (FSS Protocols QA)
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Ntfs - Library for the NTFS-only tests

=head1 SYNOPSIS

  use CIFS::Ntfs qw(ntfs_setup, ntfs_cleanup);

  %setup_h = ntfs_setup(mode => 'multi');

  ntfs_cleanup(vol => \@volumes_created);

  $text = ntfs_qtree_security_set(file => $filename);

  ntfs_cifs_setup(mode => $mode);

  @acls = cacls_get(name => 'Z:\testfile');

  $out = cacls_compare(acl_a => \@acls1, acl_b => \@acls2);

  $out = cacls_print(cacls_name => $cacls_name, acl_a => \@acls1);

  $out = cacls_set(file => $filename,  user => $user_name);

  %groups = lclgroups_get(fc => $fc);

=head1 DESCRIPTION

Provides common routines for the NTFS-only tests.

ntfs_setup gets the initial filer wafl.default_security_style
so it can reset it on cleanup. It also calls cifs setup.

ntfs_cleanup purge any volumes created during the test run then
resets the wafl.default_security_style to the pre-test setting.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: $
  Last revised $Date: $

=head1 COPYRIGHT

  Copyright 2006, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package CIFS::Ntfs;

use strict;
use warnings;
use base qw(Exporter);

use Admin::License qw(license is_licensed unlicense);
use Carp qw(croak);
use ExportFS qw(share_add share_delete);
use FileManip qw(remove_file);
use filer qw(filer_options_set filer_cifs_access filer_qtree_get
             filer_qtree_security_set);
use FS::Util::Setup qw(create_vol common_cleanup);
use GenericMount qw(umount);
use MapShare qw(map_cifs_share unmap_cifs_share);
use nfs::GenericExport qw(unexport_path);
use nfs::RemoteFileIO qw(client_rmdir);
use Params::Validate qw(validate ARRAYREF SCALAR OBJECT UNDEF HASHREF);
use SayWrap qw(sayEOF);
use CIFS::Setup  qw(cifs_setup);
use SPARSE::Util::Logger qw(logStep logWarn);
use SPARSE::Util::FilerPerfWatch;
use SPARSE::Resource::Filer qw(sparse_filer);
use Tharn qw(connect host logresult);
use Users qw(addCIFSUser removeCIFSUser);
use Volume qw(vol_purge);
use CLI::FileIO qw(rdfile wrfile);
use Network::Dns qw(dns_get_options dns_configure dns_restore_options);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our $VERSION = '1.0';

# Functions to be exported
our @EXPORT_OK = qw(
    ntfs_setup
    ntfs_cleanup
    ntfs_qtree_security_set
    ntfs_cifs_setup
    cacls_get
    cacls_compare
    cacls_print
    cacls_set
    lclgroups_get
    testme
);

#---------------------------------------------------------------------
# Global variables for this module
#---------------------------------------------------------------------
my $cifs_license = 'cifs';
my $nfs_license = 'nfs';
my $cacls = 'cacls';
my $usermap = '/etc/usermap.cfg';
my %shares_h;
my $cleanup_msg_displayed = 0;
my $filer_obj;
my $trace;

#---------------------------------------------------------------------
# State variables used by the setup and cleanup routine
# Typical access is via the filer name ex: $state{$filer}{share_name}
# This allows N filers to be setup at one time via ntfs_setup and each
# filer is guranteed to only be setup once.
#---------------------------------------------------------------------
my %state;

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub ntfs_setup(%);
sub ntfs_cleanup(%);
sub ntfs_qtree_security_set(%);
sub ntfs_cifs_setup(%);
sub cacls_get(%;);
sub cacls_compare(%;);
sub cacls_print(%);
sub cacls_set(%;);
sub _list_cmp($$;);
sub _cacls_cmp();
sub lclgroups_get(%;);
sub testme();

#---------------------------------------------------------------------

=head2 ntfs_setup

    ntfs_setup([filer          => $FILER],
               [mode           => 'multi' | 'ntfs'],
               [cifs_license   => 'yes' | 'no'],
               [nfs_license    => 'yes' | 'no'],
               [create_vol     => 'yes' | 'no'],
               [setup_cifs     => 'yes' | 'no'],
               [write_usermap  => 'yes' | 'no' | 'remove'],
               [cifs_users     => \%test_users ],
               [default_share  => 'yes' | 'no'],
               [share_name     => $CIFS_SHARE],
               [nfs_enable     => 'yes' | 'no'],
               [guitest        => 'yes' | 'no'],
                ret_h          => \%setup_h);

Store the initial wafl.default_security_style to reset on cleanup.
Runs the common cifs setup. Sets the filer
to multi-protocol or ntfs-only mode.

The following named arguments are accepted:

=over

=item Arguments

=over

=item filer

(Optional) : Filer hostspec to setup. Defaults to $main::FILER.

=item mode

(Optional) : Protocol setting for the filer can be 'ntfs' or 'multi'. Defaults to 'multi'.

=item cifs_license

(Optional) : If defined will either license or unlicense CIFS on the filer. If cifs_license => 'yes' will license CIFS. If cifs_license => 'no' will unlicense CIFS.

=item nfs_license

(Optional) : If defined will either license or unlicense NFS on the filer. If nfs_license => 'yes' will license NFS. If nfs_license => 'no' will unlicense NFS.

=item create_vol

(Optional) : If create_vol => 'yes' create default volume. If create_vol => 'no' do not create default volume.

=item setup_cifs

(Optional) : If setup_cifs => 'yes' will setup CIFS with the specified mode. If setup_cifs => 'no' will not setup CIFS. Defaults to 'yes'.

=item write_usermap

(Optional) : If write_usermap => 'yes' will write \\hostname\administrator == root into usermap.cfg. If set to 'remove', then the file is removed from the filer.
Defaults to 'yes'.

=item cifs_users

(Optional) : Hashref of user accounts to create on filer.

=item default_share

(Optional) : If default_share => 'yes' create and map default share. If default_share => 'no' do not create and map default share.

=item share_name

(Optional) : share_name => 'share_name' name of the default share. Defaults to $main::CIFS_SHARE.

=item nfs_enable

(Optional) : Enable NFS on create_vol

=item guitest

(Optional) : Checks to see if windows client has screen saver. Can't run gui tests on windows client if screen saver is on.

=item ad_setup

(Optional) : If defined and set to 'yes' will add filer to an active directory domain.

=item domain

(Optional) : Domain name to add filer to, used in cifs_setup. Defaults to
$main::CIFS_DOMAIN.

=item domain_user

(Optional) : Domain user that administers filer in domain,used in cifs_setup.
Defaults to $main::CIFS_DOMAIN_ADMIN.

=item domain_pass

(Optional) : Domain user password,used in cifs_setup. Defaults to
$main::CIFS_DOMAIN_ADMIN_PASSWD.

=item dns_server

(Optional) : DNS server IP address.

=item dns_name

(Optional) : DNS domain name used in call to cifs_setup when
adding filer to an active directory domain. Defaults to domain.

=item ret_h

(Required) : Hash ref passed by caller. To be used by caller and ntfs_cleanup.

=back

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub ntfs_setup(%) {
    my %opts = validate(
        @_,
        {
            filer           => { type => SCALAR | UNDEF,
                                 default => $main::FILER },
            mode            => { type => SCALAR, default => 'multi',
                                 regex => qr/^(ntfs|multi)$/ },
            cifs_license    => { type => SCALAR, optional => 1,
                                 regex => qr/^(yes|no)$/ },
            nfs_license     => { type => SCALAR, optional => 1,
                                 regex => qr/^(yes|no)$/ },
            create_vol      => { type => SCALAR, optional => 1,
                                 regex => qr/^(yes|no)$/ },
            setup_cifs      => { type => SCALAR, default => 'yes',
                                 regex => qr/^(yes|no)$/ },
            write_usermap   => { type => SCALAR, default => 'yes',
                                 regex => qr/^(yes|no|remove)$/ },
            cifs_users      => { type => HASHREF, optional => 1 },
            default_share   => { type => SCALAR, optional => 1,
                                 regex => qr/^(yes|no)$/ },
            share_name      => { type => SCALAR,
                                 default => $main::CIFS_SHARE },
            nfs_enable      => { type => SCALAR, default => 'no',
                                 regex => qr/^(yes|no)$/ },
            guitest         => { type => SCALAR, default => 'no',
                                 regex => qr/^(yes|no)$/ },
            ad_setup        => { type => SCALAR, default => 'no',
                                 regex => qr/^(yes|no)$/ },
            domain          => { type => SCALAR,
                                 default => $main::CIFS_DOMAIN },
            domain_user     => { type => SCALAR,
                                 default => $main::CIFS_DOMAIN_ADMIN },
            domain_pass     => { type => SCALAR,
                                 default => $main::CIFS_DOMAIN_ADMIN_PASSWD },
            dns_name        => { type => SCALAR, optional => 1 },
            dns_server      => { type => SCALAR, optional => 1 },
            ret_h           => { type => HASHREF }
          }
    );

    my (%args, @arr, $cifs_drive, $cifs_user, $cmd, $fc, $fh, $i,
        $fname, $mode, $new_text, $purge, %qtree_out, $r, %test_users, $volume);

    logStep("BEGIN: ntfs_setup");

    # Get connection object to filer
    eval {
        $fh = host($opts{filer});
        $fname = $fh->hostname();
        $fc = connect($main::FILER_CONN, -timeout, $main::TIMEOUT);
        $fc->timeout($main::CMDTIMEOUT);
    };
    croak("ntfs_setup: Get connection object to filer $@") if ($@);

    # Save return values
    ${$opts{ret_h}}{filer} = $opts{filer};
    ${$opts{ret_h}}{fc} = $fc;
    ${$opts{ret_h}}{shares} = \%shares_h;

    if ($opts{guitest} eq 'yes') {
        logStep("ntfs_setup: gui screen_lock test");
        # Disable the screen lock on the local client and just leave it off
        # Croak if the screen lock was on.
        logStep("ntfs_setup: Verify client has screen saver turned off");
        require CIFS::Vss_gui;
        import CIFS::Vss_gui qw(screen_lock);
        screen_lock(setting => 'off', croak_if_on => 'yes');
    }

    # Reference to state for this filer
    $r = \%{$state{$fname}};

    if (defined($opts{cifs_license})) {
        logStep("ntfs_setup: CIFS license check and set");
        $r->{cifs_is_licensed} = is_licensed($fc, $cifs_license);
        croak("ntfs_setup: check cifs license failed $@") if $@;
        if (($opts{cifs_license} eq 'yes') && !$r->{cifs_is_licensed}) {
            license($fc, $cifs_license);
            croak("ntfs_setup: cifs license failed $@") if $@;
            $r->{cifs_license_called} = 1;
        } elsif (($opts{cifs_license} eq 'no') && $r->{cifs_is_licensed}) {
            unlicense($fc, $cifs_license);
            croak("ntfs_setup: cifs unlicense failed $@") if $@;
            $r->{cifs_unlicense_called} = 1;
        }
    }

    if (defined($opts{nfs_license})) {
        logStep("ntfs_setup: NFS license check and set");
        $r->{nfs_is_licensed} = is_licensed($fc, $nfs_license);
        croak("ntfs_setup: nfs license check failed $@") if $@;
        if (($opts{nfs_license} eq 'yes') && !$r->{nfs_is_licensed}) {
            license($fc, $nfs_license);
            croak("ntfs_setup: nfs license failed $@") if $@;
            $r->{nfs_license_called} = 1;
        } elsif (($opts{nfs_license} eq 'no') && $r->{nfs_is_licensed}) {
            unlicense($fc, $nfs_license);
            croak("ntfs_setup: nfs unlicense failed $@") if $@;
            $r->{nfs_unlicense_called} = 1;
        }
    }

    $r->{filer_vol_path} = "/vol/$main::FILER_VOL";
    if (defined($opts{create_vol}) && $opts{create_vol} eq 'yes') {
        logStep("ntfs_setup: Create volume");
        $r->{create_vol_called} = 1;
        # Using $main::FILER_VOL as volume name
        ($fc, $volume) = create_vol(filer => $opts{filer},
                                    nfs_enable => $opts{nfs_enable})
            or croak("ntfs_setup: Create vol failed $@");

        # At this point the test has asked us to create the vol and the
        # vol already existed or did not exist and was created. Make sure
        # the qtree security style for FILER_VOL matches the protocol mode
        # requested for the CIFS setup.
        %qtree_out = filer_qtree_get($fc, vols => 1);

        # If mode is multi, map name to mixed for comparison
        $mode = $opts{mode};
        $mode = 'mixed' if ($mode =~ /multi/);

        if ($qtree_out{$main::FILER_VOL}{''}{Style} ne $mode) {
            logStep("Set the qtree security mode on $main::FILER_VOL to $mode");
            if (filer_qtree_security_set($fc, $r->{filer_vol_path}, $mode)
                == 0) {
                croak("ntfs_setup: qtree set mode $mode failed on $r->{filer_vol_path}");
            }
        }
    }

    # Create username to use for cifs based on whether AD is setup or not
    # Default is workgroup
    $fname =~ tr/a-z/A-Z/;

    $cifs_user = "$fname\\$main::CIFS_USER";
    # Reassign $main::CIFS_USER and so the fully qualified user name
    # gets propagated to subtests correctly
    $main::CIFS_USER = $cifs_user;
    $new_text = "\n$cifs_user == root\n";

    if (defined($opts{domain})) {
        $opts{domain} =~ s/\s+//g;
        if ($opts{domain} ne '') {
            # Using Domains
            $new_text .= "$main::CIFS_DOMAIN_ADMIN == root\n";
        }
    }
    ${$opts{ret_h}}{cifs_user} = $cifs_user;
    ${$opts{ret_h}}{cifs_passwd} = $main::CIFS_PASSWD;

    if ($opts{write_usermap} eq 'yes') {
        logStep("ntfs_setup: Setup $usermap file on $fname");
        $r->{filer_usermap_text} = rdfile($fc, $usermap);
        wrfile($fc, $usermap, "$r->{filer_usermap_text}$new_text")
            or croak("ntfs_setup: can't write $usermap: $@");
    } elsif ($opts{write_usermap} eq 'remove') {
        $r->{filer_usermap_text} = rdfile($fc, $usermap);
        $cmd = "priv set advanced; rm $usermap";
        sayEOF($fc, $cmd);
        $new_text = rdfile($fc, $usermap);
        croak("ntfs_setup: Couldn't remove $usermap on filer") if (!$@);
    }

    @arr = keys(%{$opts{cifs_users}});
    if (defined($opts{cifs_users}) && @arr > 0) {
        @{$r->{cifs_users_created}} = ();
        logStep("ntfs_setup: Add CIFS user to $fname");
        %test_users = %{$opts{cifs_users}};
        foreach $i (keys %test_users) {
            # First remove CIFS user, ignore any error returned as the
            # user may not exist
            removeCIFSUser($fc, $i);
            addCIFSUser($fc, $i,
                        $test_users{$i}{passwd},
                        $test_users{$i}{group});
            croak("ntfs_setup: can't add CIFS user $i: $@") if $@;
            push @{$r->{cifs_users_created}}, $i;
        }
    }

    if (defined($opts{setup_cifs}) && $opts{setup_cifs} eq 'yes') {
        logStep("ntfs_setup: Setup filer $fname as $opts{mode}");
        if ($opts{ad_setup} eq 'yes') {
            if (!defined($opts{dns_name})) {
                $opts{dns_name} = $opts{domain};
            }
            %args = (ad_setup    => $opts{ad_setup},
                     dns_name    => $opts{dns_name});

            if (defined($opts{dns_server}) && $opts{dns_server} ne '') {
                logStep("ntfs_setup: Save the DNS state of the filer");
                $r->{saved_dns_info} = dns_get_options(conn => $fc);

                logStep("ntfs_setup: Setup DNS on the filer");
                $fh = host($opts{dns_server});
                dns_configure(conn => $fc,
                              enable => 'on',
                              domain => $opts{domain},
                              srv_ip_addr => $fh->default_ip());
                $args{dns_server} = $fh->default_ip();
            }
        } else {
            %args = ();
        }
        ntfs_cifs_setup(domain      => $opts{domain},
                        domain_user => $opts{domain_user},
                        domain_pass => $opts{domain_pass},
                        filer => $opts{filer},
                        mode  => $opts{mode},
                        %args);

        if (defined($opts{default_share}) && $opts{default_share} eq 'yes') {
            logStep("ntfs_setup: share add $opts{share_name}");
            $purge = ($opts{share_name} =~ /^(etc\$|c\$|home)$/i) ? 0 : 1;
            # if share name is ETC$, C$, or HOME don't add them,
            # they already exist, they just need to be mapped to a cifs drive
            if ($purge) {
                share_add($fc, $opts{share_name}, $r->{filer_vol_path},
                          verify => 1, purge_existing => $purge, umask => 0)
                    or croak("ntfs_setup: failed to share $opts{share_name}");
            }
            $shares_h{$opts{share_name}} = $purge;

            logStep("ntfs_setup: map CIFS share");
            $cifs_drive = map_cifs_share(
                        server   => $fc->connhost(),
                        share    => $opts{share_name},
                        client   => 'localhost',
                        drive    => $main::CIFS_DRIVE,
                        username => $cifs_user,
                        password => $main::CIFS_PASSWD
                        )
                    or croak("ntfs_setup: could not map the drive $@");

            if (!$cifs_drive) {
                croak("ntfs_setup: could not map the drive");
            }

            $cifs_drive .= ":" unless $cifs_drive =~ /:$/;
            push @{$opts{ret_h}{drives}}, $cifs_drive;
        }
    }

    if ($main::TRACE) {
        # Initialize pktt utility to collect packet traces
        logStep('Start pktt trace');
        $filer_obj = sparse_filer($main::FILER);
        $trace = perf_watch($filer_obj, "pktt");
        $trace->watchStart();
    }

    logStep("END: ntfs_setup");
    return;
}

#---------------------------------------------------------------------

=head2 ntfs_cleanup

    ntfs_cleanup([filer         => $FILER],
                 [cifs_user     => 'Administrator'],
                 [cifs_passwd   => 'whatever'],
                 [options       => \%options],
                 [vol           => $volume],
                 [drives        => \@drives],
                 [shares        => \%shares],
                 [vols          => \@volumes],
                 [dirs          => \@directories],
                 [files         => \@files],
                 [share_access  => \%share_access],
                 [share_acls    => \%share_acls],
                 [users_created => \@users],
                 [cifs_users    => \@cifs_users],
                 [unexport      => \@exported_paths],
                 [cc            => $client],
                 [unmount_vols  => \@mounted_vols],
                 [unmount_dirs  => \@nfs_dirs],
                 [unmount_files => \@nfs_files],
                 [ad_setup      => 'yes'],
                 [ad_server     => $ad_server],
                 [domain        => $ad_domain]);

Cleanup after a test run. Unmap CIFS drive. Restore filer options if changed
during test run. Purge volumes created after test run. Purge default volume
if created by ntfs_setup. Unmap any drives if mapped by test. Delete any
shares added by test. Purge volumes, dirs and files created by test. Restore
share access and acls changed by test. Delete users created by test. Delete
CIFS users added by test. Unlicenses CIFS if ntfs_setup licensed CIFS.
Unlicenses NFS if ntfs_setup licensed NFS. Resets filer
wafl.default_security_style to pre-test setting. Use this cleanup
routine if you use ntfs_setup.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Optional) : Filer connection object.

=item filer

(Optional) : Filer hostspec to setup. Defaults to $main::FILER.

=item cifs_user

(Optional) : CIFS username that was used in the filer usermap.cfg file.

=item cifs_passwd

(Optional) : CIFS password to go with cifs_user.

=item options

(Optional) : reset any options that were changed in setup.

=item vol

(Optional) : Default volume to purge after test run.

=item drives

(Optional) : drives to unmap after test run from localhost.

=item shares

(Optional) : shares to delete after test run.

=item vols

(Optional) : List of volumes to purge after test run.

=item dirs

(Optional) : List of directories to remove after test run.

=item files

(Optional) : List of files to remove after test run.

=item share_access

(Optional) : Hash of hashes used to set shares access to pretest state.

=item share_acls

(Optional) : Hash of hashes used to delete share level acls added by test.

=item cifs_users

(Optional) : Hashref of cifs users added by test.

=item unexport

(Optional) : Array of vol paths to unexport that were exported by test.

=item cc

(Optional) : Nfs Client connection object. If defined used to cleanup any mounted vols, dirs and files used by client.

=item unmount_vols

(Optional) : Array of vols mounted by Nfs Client during test.

=item unmount_dirs

(Optional) : Array of dirs created by Nfs Client during test.

=item unmount_files

(Optional) : Array of files created by Nfs Client during test.

=item ad_setup

(Optional) : If defined and set to 'yes' filer was added to an active directory domain by test, so call ad_cleanup.

=item ad_server

(Optional) : Active directory server to pass to ad_cleanup.

=item domain

(Optional) : Domain name to pass to ad_cleanup.

=back

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub ntfs_cleanup(%) {
    my %opts = validate(
        @_,
        {   fc             => { type => OBJECT   | UNDEF, optional => 1 },
            filer          => { type => SCALAR, default => $main::FILER },
            options        => { type => HASHREF  | UNDEF, optional => 1 },
            drives         => { type => ARRAYREF | UNDEF, optional => 1 },
            shares         => { type => HASHREF  | UNDEF, optional => 1 },
            vols           => { type => ARRAYREF, optional => 1 },
            dirs           => { type => ARRAYREF, optional => 1 },
            files          => { type => ARRAYREF, optional => 1 },
            share_access   => { type => HASHREF,  optional => 1 },
            share_acls     => { type => HASHREF,  optional => 1 },
            cifs_users     => { type => HASHREF  | UNDEF, optional => 1 },
            unexport       => { type => ARRAYREF, optional => 1 },
            cc             => { type => OBJECT   | UNDEF, optional => 1 },
            unmount        => { type => ARRAYREF, optional => 1 },
            unmount_vols   => { type => ARRAYREF, optional => 1 },
            unmount_dirs   => { type => ARRAYREF, optional => 1 },
            unmount_files  => { type => ARRAYREF, optional => 1 },
            ad_setup       => { type => SCALAR   | UNDEF, optional => 1 },
            ad_server      => { type => SCALAR   | UNDEF, optional => 1 },
            domain         => { type => SCALAR   | UNDEF, optional => 1 },
            cifs_user      => { type => SCALAR   | UNDEF, optional => 1 },
            cifs_passwd    => { type => SCALAR   | UNDEF, optional => 1 },
        }
    );

    my ($acl_user, @arr, $fc, $fh, $i, $r, $ret_val, $text, $user, $vol);

    if (!$cleanup_msg_displayed) {
        logStep("CLEANUP == $main::CLEANUP");
        $cleanup_msg_displayed = 1;
    }

    if ($main::TRACE && defined $trace) {
        # Shut down pktt tracing and generate reports
        logStep('Shut down pktt trace');
        $trace->watchStop();
        $trace->watchReport();
    }

    eval {
        # Reconnect to filer
        $fh = host($opts{filer});
        logStep('ntfs_cleanup: cleanup related to filer ' . $fh->hostname());
        $fc = connect($main::FILER_CONN, -timeout, $main::TIMEOUT);
        $fc->timeout($main::CMDTIMEOUT);

        # Reference to state for this filer
        $r = $state{$fh->hostname()};

        if ($main::CLEANUP ne 'none') {
            if (defined($opts{files}) && @{$opts{files}} > 0) {
                logStep("ntfs_cleanup: unlinking files");
                unlink @{$opts{files}} or
                    logWarn("ntfs_cleanup: unlink failed: $!");
                @{$opts{files}} = ();
            }

            if (defined($opts{dirs}) && @{$opts{dirs}} > 0) {
                logStep("ntfs_cleanup: removing directories");
                # Traverse array in reverse order so that the directories
                # are removed newest to oldest.
                foreach my $dir (@{$opts{dirs}}) {
                    # Change the mode to make the directory removable,
                    # but more importantly to clear out the ACL.
                    # Remove the directory
                    chmod 0777, $dir;
                    rmdir $dir or
                        logWarn("ntfs_cleanup: rmdir $dir failed: $!");
                }

                @{$opts{dirs}} = ();
            }

            # restore options
            @arr = keys(%{$opts{options}});
            if ($fc && defined($opts{options}) && @arr > 0) {
                logStep("ntfs_cleanup: restoring options");
                foreach $i (@arr) {
                    eval {
                        filer_options_set($fc, $i, ${$opts{options}}{$i});
                    };

                    logWarn("ntfs_cleanup: $@") if $@;
                }
            }

            @arr = keys(%{$opts{share_acls}});
            if ($fc && defined($opts{share_acls}) && @arr > 0) {
                logStep("ntfs_cleanup: deleting share level acls");
                foreach $i (@arr) {
                    eval {
                       $text = filer_cifs_access(
                              $fc,
                              share       => $i,
                              user        => $opts{share_acls}{$i},
                              delete_flag => 'yes');
                    };

                    logWarn("ntfs_cleanup: $@") if $@;
                }
            }

            @arr = keys(%{$opts{share_access}});
            if ($fc && defined($opts{share_access}) && @arr > 0) {
                logStep("ntfs_cleanup: restoring share permissions");
                foreach $i (@arr) {
                    foreach $acl_user (keys %{$opts{share_access}{$i}{$i}{acl}}) {
                        eval {
                           $text = filer_cifs_access(
                                  $fc,
                                  share  => $i,
                                  user   => $acl_user,
                                  rights =>
                                  $opts{share_access}{$i}{$i}{acl}{$acl_user});
                        };

                        logWarn("ntfs_cleanup: $@") if $@;
                    }
                }
            }

            @arr = keys(%{$opts{shares}});
            if ($fc && defined($opts{shares}) && @arr > 0) {
                logStep("ntfs_cleanup: deleting shares");
                foreach $i (@arr) {
                    if (${$opts{shares}}{$i} ==  1) {
                        eval {
                            share_delete($fc, $i);
                        };
                        logWarn("ntfs_cleanup: $@") if $@;
                    }
                }
            }

            # Unmapping shares is done after the shares are removed
            # which is backwards but is done this way because sometimes
            # windows thinks files are still open even if they have been
            # closed. The force option on unmap_cifs_share doesn't work
            # so this is the only thing that seems to work.
            if (defined($opts{drives}) && @{$opts{drives}} > 0) {
                logStep("ntfs_cleanup: unmapping drives");
                my $drive;
                while ($drive = pop @{$opts{drives}}) {
                    unmap_cifs_share(mappath => $drive);
                    logWarn("ntfs_cleanup: $@") if $@;
                }
            }

            if ($fc && $r->{filer_usermap_text}) {
                logStep("ntfs_cleanup: Restore $usermap file on filer");
                wrfile($fc, $usermap, $r->{filer_usermap_text})
                    or logWarn("ntfs_cleanup: can't write $usermap: $@");
            }

            if ($fc && defined($r->{cifs_users_created})) {
                logStep("ntfs_cleanup: Remove cifs users added to the filer");
                foreach my $user (@{$r->{cifs_users_created}}) {
                    removeCIFSUser($fc, $user);
                    logWarn("ntfs_cleanup: $@") if $@;
                }
            }

            if ($fc && defined($opts{vols}) && @{$opts{vols}} > 0) {
                # Remove all volumes created by test
                foreach $vol (@{$opts{vols}}) {
                    eval {
                        vol_purge($fc, $vol);
                    };

                    logWarn("ntfs_cleanup: $@") if $@;
                }
            }

            # If we have a client connection object we have mounted a file
            # system so lets unmount the file system and remove any files
            # or directories created during test
            if (defined($opts{cc})) {
                if (defined($opts{unmount_files}) &&
                    @{$opts{unmount_files}} > 0) {
                    ## @step Removing files
                    logStep('ntfs_cleanup: Removing files');
                    foreach my $file (@{$opts{unmount_files}}) {
                        eval {
                            $ret_val = remove_file(file => $file,
                                                   user => 'root');
                            if (!$ret_val) {
                                logWarn("ntfs_cleanup: $@") if $@;
                            }
                        };

                        logWarn("ntfs_cleanup: $@") if $@;
                    }
                }

                if (defined($opts{unmount_dirs}) &&
                    @{$opts{unmount_dirs}} > 0) {
                    ## @step Removing directories
                    logStep('ntfs_cleanup: Removing directories');
                    foreach my $dir (@{$opts{unmount_dirs}}) {
                        eval {
                            $ret_val = client_rmdir($opts{cc}, $dir);
                            if (!$ret_val) {
                                logWarn("ntfs_cleanup: $@") if $@;
                            }
                        };

                        logWarn("ntfs_cleanup: $@") if $@;
                    }
                }

                if (defined($opts{unmount}) && @{$opts{unmount}} > 0) {
                    ## @step Unmounting
                    logStep('ntfs_cleanup: Unmounting');
                    foreach $vol (@{$opts{unmount}}) {
                        eval {
                            $ret_val = umount($vol, client=>$opts{cc},
                                              force => '1');
                            if (!$ret_val) {
                                logWarn("ntfs_cleanup: $@") if $@;
                            }
                        };

                        logWarn("ntfs_cleanup: $@") if $@;
                    }
                }

                if ($fc && defined($opts{unmount_vols}) &&
                    @{$opts{unmount_vols}} > 0) {
                    ## @step Removing mounted volumes created by test
                    logStep('ntfs_cleanup: Removing mounted volumes ' .
                            'created by test');
                    foreach $vol (@{$opts{unmount_vols}}) {
                        eval {
                            vol_purge($fc, $vol);
                        };

                        logWarn("ntfs_cleanup: $@") if $@;
                    }
                }
            }

            if ($fc && defined($opts{unexport}) && @{$opts{unexport}} > 0) {
                ## @step Unexporting path
                logStep('ntfs_cleanup: Unexporting path');
                foreach $vol (@{$opts{unexport}}) {
                    $ret_val = unexport_path($fc, $vol);
                    if (!$ret_val) {
                        logWarn("ntfs_cleanup: $@") if $@;
                    }
                }
            }

            if ((!$r->{cifs_is_licensed}) && $r->{cifs_license_called}) {
                logStep("ntfs_cleanup: unlicense CIFS");
                unlicense($fc, $cifs_license);
                logWarn("ntfs_cleanup: $@") if $@;
            } elsif ($r->{cifs_is_licensed} && $r->{cifs_unlicense_called}) {
                logStep("ntfs_cleanup: license CIFS");
                license($fc, $cifs_license);
                logWarn("ntfs_cleanup: $@") if $@;
            }

            if ((!$r->{nfs_is_licensed}) && $r->{nfs_license_called}) {
                logStep("ntfs_cleanup: unlicense NFS");
                unlicense($fc, $nfs_license);
                logWarn("ntfs_cleanup: $@") if $@;
            } elsif ($r->{nfs_is_licensed} && $r->{nfs_unlicense_called}) {
                logStep("ntfs_cleanup: license NFS");
                license($fc, $nfs_license);
                logWarn("ntfs_cleanup: $@") if $@;
            }
        }

        eval {
            if ($fc && defined($r->{saved_dns_info})) {
                logStep("ntfs_cleanup: Restore filer DNS information");
                dns_restore_options(conn => $fc,
                                    options => $r->{saved_dns_info});
            }
        };
        logWarn("ntfs_cleanup: $@") if $@;

        # Call common cleanup if create_vol was called
        if ($fc && $r->{create_vol_called}) {
            common_cleanup($fc);
            logWarn("ntfs_cleanup: common_cleanup: $@") if $@;
        }
    };

    logWarn("ntfs_cleanup: $@") if $@;
    logStep("END: ntfs_cleanup");

    return;
}

#---------------------------------------------------------------------

=head2 ntfs_qtree_security_set

    my $text = ntfs_qtree_security_set(file => $filename);

This function changes security for qtree.

This function is used for special test case where a warning message is
displayed on stderr on the filer when setting qtree security mode to unix
on an NTFS-only filer without an NFS license.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Filer connection object.

=item qtree_path

(Required) : path on filer to qtree to change security on.

=item qtree_mode

(Required) : security mode for qtree. Either mixed | unix | ntfs.

=back

=item Returns

Returns text string with 'qtree security' output

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub ntfs_qtree_security_set(%) {
    my %opts = validate(
        @_,
        {
            fc         => { type => OBJECT },
            qtree_path => { type => SCALAR },
            qtree_mode => { type => SCALAR,
                            regex => qr/^(mixed|ntfs|unix)$/ },
        }
    );

    my($fc) = $opts{fc};
    my ($qtreepath) = $opts{qtree_path};
    my ($mode) = $opts{qtree_mode};

    my $cmd = "qtree security $qtreepath $mode 2>&1";
    my($text) = sayEOF($fc,$cmd);

    return $text;
}

#---------------------------------------------------------------------

=head2 ntfs_cifs_setup

    ntfs_cifs_setup(mode =>$mode);

This function is a wrapper function that calls cifs_setup to setup CIFS on the
filer with the specified mode. This wrapper is used to isolate the details of
setting up in a domain environment or on WORKGROUP.

Will setup cifs in a domain environment or under WORKGROUP. If domain
is an empty string will setup under WORKGROUP otherwise will attempt to
setup under a domain using passed in or default values. Croaks if their
is an error in call to cifs_setup.

The following named arguments are accepted:

=over

=item Arguments

=over

=item filer

(Required) : Name of filer to connect to.

=item mode

(Optional) : Protocol mode to setup cifs. Either 'Multi' protocol or 'Ntfs'
only. Defaults to 'Multi' protocol.

=item ad_setup

(Optional) : If defined and set to 'yes' will add filer to an active directory
domain.

=item domain

(Optional) : Domain name to add filer to, used in cifs_setup. Allows a
domain name with just spaces for testing purposes. Defaults to
$main::CIFS_DOMAIN.

=item domain_user

(Optional) : Domain user that administers filer in domain,used in cifs_setup.
Defaults to $main::CIFS_DOMAIN_ADMIN.

=item domain_pass

(Optional) : Domain user password,used in cifs_setup. Defaults to
$main::CIFS_DOMAIN_ADMIN_PASSWD.

=item dns_server

(Optional) : DNS server IP address.

=item dns_name

(Optional) : DNS domain name used in call to cifs_setup when
adding filer to an active directory domain.

=back

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub ntfs_cifs_setup(%) {
    my %opts = validate(
        @_,
        {
            filer       => { type => SCALAR },
            mode        => { type => SCALAR, default => 'multi',
                             regex => qr/^(ntfs|multi)$/ },
            ad_setup    => { type => SCALAR, default => 'no',
                             regex => qr/^(yes|no)$/ },
            domain      => { type => SCALAR,
                             default => $main::CIFS_DOMAIN },
            domain_user => { type => SCALAR,
                             default => $main::CIFS_DOMAIN_ADMIN },
            domain_pass => { type => SCALAR,
                             default => $main::CIFS_DOMAIN_ADMIN_PASSWD },
            dns_server  => { type => SCALAR, optional => 1,
                             regex=>qr/(?:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/ },
            dns_name    => { type => SCALAR, optional => 1 },
        }
    );

    my (%args, $fc, $fh, $err_msg, @proto);

    # Determine whether to use the domain or the workgroup
    $opts{domain} =~ s/\s+//g;
    if ($opts{domain} ne '') {
        # Special hook to allow passing in an empty domain name
        if ($opts{domain} eq 'XX_empty_XX') {
            $opts{domain} = $opts{dns_name} = '    ';
        }

        %args = (domain      => $opts{domain},
                 domain_user => $opts{domain_user},
                 domain_pass => $opts{domain_pass});
    } else {
        %args = (workgroup => 'WORKGROUP');
    }

    if ($opts{ad_setup} eq 'yes') {
        if (!($opts{dns_server} && $opts{dns_name})) {
            croak('ntfs_cifs_setup: AD setup requested but dns_server and/or ' .
                  'dns_name parameters not passed in');
        }
    }

    # ad_setup is used to set up the filer under a different active
    # directory domain then the one it was initially set up with.
    # The extra params dns_domainname and dns_server could
    # be different then the initial dns_domainname and dns_server so
    # they're passed in to cifs setup. If DNS was initially setup up and
    # it's not going ad_setup will be set to 'no' or won't be passed in and
    # defaults to no. ad_setup is ONLY USED if DNS resolver was NOT initially
    # setup or if its going to CHANGE.
    if ($opts{ad_setup} eq 'yes') {
        $args{dns_enable} = 1;
        $args{dns_domainname} = $opts{dns_name};
        $args{dns_server_list} = $opts{dns_server};
    }

    # If the protocol requested is ntfs then cifs_setup has to be run
    # twice, the first time to multi then the second time to ntfs. This
    # gurantees that the filer is in NTFS-only mode.
    if ($opts{mode} eq 'ntfs') {
        @proto = ( 'multiprotocol', 'ntfs' );
    } elsif ($opts{mode} eq 'multi') {
        @proto = ( 'multiprotocol' ); 
	} else {
        @proto = ( $opts{mode} );
    }

	
    foreach my $p (@proto) {
 	if ($main::MODE eq "7G") {
		cifs_setup(host          => $opts{filer},
			'auth-type'      => 'ad',
			cifs_pass        => $main::CIFS_PASSWD,
			passwd_group     => 'yes',
			'security-style' => $p,
       	        	%args);
	 } else {      	
	      cifs_setup(host		 => $opts{filer},
			'server-name'	 => $opts{filer},
			'auth-type'  	 => 'ad',
		  	cifs_pass	 => $main::CIFS_PASSWD,
			passwd_group	 => 'yes',
			'security-style' => $p,
                   	%args);
	}

        if ($@) {
            # save the cifs_setup error
            $err_msg = $@;

            # cifs_setup should cleanup after itself but it can leave
            # the command 'cifs setup' running and prompting for input.
            # When 'cifs setup' gets converted to ZAPI this problem should
            # go away. For now grab the telnet connection to send a ^C to
            # the filer.
            eval {
                $fh = host($opts{filer});
                $fc = connect("#telnet", -timeout, $main::TIMEOUT);
                $fc->timeout($main::CMDTIMEOUT);
                $fc->send_intr();
            };
            croak("ntfs_cifs_setup: cifs setup failed $err_msg");
        }
    }

    return;
}

#---------------------------------------------------------------------

=head2 cacls_get

    @acls = cacls_get(file => $filename);

This function gets the acls for the specified file.

The following named arguments are accepted:

=over

=item Arguments

=over

=item file

(Required) : Absolute file name to get acls for.

=back

=item Returns01_verify_new_user.stub.log

Returns an array of acls for the file. Each entry
in the array is a hash containing 'who' and 'access'
as keys.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cacls_get(%;) {
    my %opts = validate(
        @_,
        {
            file => { type => SCALAR },
        }
    );
    my (@ret, $who, $access, $text);
    my $file = $opts{file};

    # Add '\' when file is the mapped drive
    $file =~ s/(.:)$/$1\\/;

    # Execute cacls command
    $text = qx($cacls $file 2>&1);
    croak "cacls_get: file: $file: $text" if $?;

    # Remove filename from output
    my $qfile = quotemeta($file);
    $text =~ s/^.*$qfile //;
    # Change unknown account to a question mark
    $text =~ s/<Account Domain not found>/\?:/g;

    # Process the output one line at a time
    foreach ( split /\r?\n/, $text ) {
        # Remove leading and trailing spaces
        s/^\s+//;
        s/\s+$//;
        next if /^$/;
        if ( /([^:]+):(.*)/ ) {
            # Set the current who and access
            $who    = $1;
            $access = $2;
            push @ret, { who => $who };
            if ( $access !~ /special access:/ ) {
                # No special access so just save it
                $ret[-1]{access} = $access;
            } else {
                # Special access, initialize array reference
                $ret[-1]{access} = [];
            }
        } else {
           # Save access for current who
           push @{$ret[-1]{access}}, $_;
        }
    }

    return @ret;
}

#---------------------------------------------------------------------

=head2 cacls_compare

    $out = cacls_compare(acl_a => \@acls1, acl_b => \@acls2);

This function compares two acls.

The following named arguments are accepted:

=over

=item Arguments

=over

=item acls_a

(Required) : Array reference of first acl to compare.

=item acls_b

(Required) : Array reference of second acl to compare.

=back

=item Returns

Returns 1 if both acls are the same, otherwise returns 0.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cacls_compare(%;) {
    my %opts = validate(
        @_,
        {
            acl_a => { type => ARRAYREF },
            acl_b => { type => ARRAYREF },
        }
    );
    # Check length of arrays
    return 0 if ( @{$opts{acl_a}} != @{$opts{acl_b}} );

    # Sort both lists of acls
    my @sort_acla = sort _cacls_cmp @{$opts{acl_a}};
    my @sort_aclb = sort _cacls_cmp @{$opts{acl_b}};

    # Compare lists of acls
    my $n = 0;
    foreach my $acl_a ( @sort_acla ) {
        return 0 if ( $acl_a->{who} ne $sort_aclb[$n]{who} or
                      _list_cmp($acl_a->{access}, $sort_aclb[$n++]{access}) );
    }

    return 1;
}

#---------------------------------------------------------------------

=head2 cacls_print

    $out = cacls_print(cacls_name => $cacls_name, acl_a => \@acls1);

This function prints an acl.

The following named arguments are accepted:

=over

=item Arguments

=over

=item acls_a

(Required) : Array reference of acl to print.

=back

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cacls_print(%) {
    my %opts = validate(
        @_,
        {
            cacls_name => { type => SCALAR },
            acl_a      => { type => ARRAYREF },
        }
    );

    print ("\n", $opts{cacls_name}, "          \n");

    if (ref($opts{acl_a}) eq 'ARRAY') {
        foreach my $acl (@{$opts{acl_a}}) {
            if (ref($acl) eq 'HASH') {
                if (defined($acl->{who}) && defined($acl->{access})) {
                    print ("Who - ", $acl->{who}, "          \n");
                    if (ref($acl->{access}) eq 'ARRAY') {
                        foreach my $access (@{$acl->{access}}) {
                            print (" ", $access, "          \n");
                        }
                    } else {
                        print (" ", $acl->{access}, "          \n");
                    }

                    print ("\n");
                }
            }
        }
    }
    return;
}

#---------------------------------------------------------------------

=head2 cacls_set

    $out = cacls_set(file => $filename,  user => $user_name);

This function sets or removes an acl for the specified file.

The following named arguments are accepted:

=over

=item Arguments

=over

=item file

(Required) : Absolute file name to get acls for.

=item user

(Required) : User to remove access for.

=back

=item Returns

Returns 1 on success 0 on failure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub cacls_set(%;) {
    my %opts = validate(
        @_,
        {
            file => { type => SCALAR },
            user => { type => SCALAR },
        }
    );

    my ($text);
    my $file = $opts{file};
    my $user = $opts{user};

    # Add '\' when file is the mapped drive
    $file =~ s/(.:)$/$1\\/;

    # Execute cacls command
    $text = qx($cacls $file \/e \/r $user 2>&1);
    croak "cacls_set: file: $file: $text" if $?;
    return 1;
}

#---------------------------------------------------------------------
# Internal function.
# Compares two elements and returns -1, 0 or 1 if first element is
# less, equal or greater than respectively.
# The elements could be either a scalar or array reference.
# Scalar is defined to be less than an array reference.
#---------------------------------------------------------------------
sub _list_cmp($$;) {
    my ($a, $b) = @_;
    if ( ! ref($a) and ! ref($b) ) {
        # Both are scalars
        return($a cmp $b);
    } elsif ( ref($a) xor ref($b) ) {
        # Scalar and reference
        return ( ref($a) ? 1 : -1 )
    } else {
        # Both are references
        my $out;
        my $n = 0;
        # Sort both arrays
        my @sorta = sort @{$a};
        my @sortb = sort @{$b};
        # Compare each value in the arrays
        foreach my $aval ( @sorta ) {
            if ( $out = $aval cmp $sortb[$n++] ) {
                return $out;
            }
            # In case array A is longer
            last if ($n == @sortb);
        }
        # If it gets this far, all elements have been
        # the same so return difference in length
        return scalar(@sorta - @sortb);
    }
}

#---------------------------------------------------------------------
# Internal function.
# Sorting function for cacls.
#---------------------------------------------------------------------
sub _cacls_cmp() {
    return $a->{who} cmp $b->{who} || _list_cmp($a->{access}, $b->{access});
}

#---------------------------------------------------------------------

=head2 lclgroups_get

    %groups = lclgroups_get(fc => $fc);

Reads the /etc/lclgroups.cfg file and returns a hash with the
Administrators, Users, Guests, Replicators, Backup Operators, and
Power Users SIDs. Each type of group has an array of SIDs.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connection to the filer.

=back

=item Returns

Result hash containing the Administrators, Users, Guests, Replicators,
Backup Operators, and Power Users SIDs (ex: $group{Users}[1]).

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub lclgroups_get(%;) {
    my %opts = validate(
        @_,
        {
            fc => { type => OBJECT },
        }
    );
    my ($data, @lines, $mode, %ret);

    # Read lclgroups.cfg
    $mode = 'Unknown Group';
    $data = rdfile($opts{fc}, "/etc/lclgroups.cfg");
    croak ("lclgroups_get: Cannot read /etc/lclgroups.cfg: $@") if ($@);
    @lines = split('\n', $data);
    foreach (@lines) {
        if (/^\[ \"Administrators\"/) {
            $mode = 'Administrators';
            next;
        }
        if (/^\[ \"Users\"/) {
            $mode = 'Users';
            next;
        }
        if (/^\[ \"Guests\"/) {
            $mode = 'Guests';
            next;
        }
        if (/^\[ \"Replicators\"/) {
            $mode = 'Replicators';
            next;
        }
        if (/^\[ \"Backup Operators\"/) {
            $mode = 'Backup Operators';
            next;
        }
        if (/^\[ \"Power Users\"/) {
            $mode = 'Power Users';
            next;
        }
        if (/^S-/) {
            push @{$ret{$mode}}, "$_";
        }
    }

    return %ret;
}

#---------------------------------------------------------------------

=head2 testme

Self test routine for this library.

Use test_lib.thpl to invoke this routine.

Example
    ntest -sim PKGNAME="CIFS::Ntfs" SUBNAME=testme CLIENT='minwin_clnt'
    CIFS_DOMAIN='cifs.lab.domain' CIFS_DOMAIN_ADMIN='admin_name'
    CIFS_DOMAIN_ADMIN_PASSWD='password' FILER=myfiler util/test_pkg.thpl

=over

=item Arguments

=over

=item CLIENT

(Required) : Minwin compatible client on which to run testme.

=item FILER

(Required) : Filer to test.

=back

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub testme() {
    use CIFS::CommonParams;
    use Utils qw(subtest_remote);
    use TestEnv::MinWin qw(minwin_compatible);

    my $host = host($main::CLIENT);
    if ($host->hosttype() !~ /windows/ || !minwin_compatible($main::CLIENT)) {
        logresult('CONFIG', 'The host ' . $host->hostname() .
                  ' is not MinWin compatible, exiting.');
        return;
    }

    if ($main::CLIENT ne 'localhost') {
        eval {
            subtest_remote(test   => 'util/test_pkg.thpl',
                           client => $main::CLIENT);
        };

        logresult('CONFIG', $@) if ($@);
        exit(0);
    }

    my (%setup_h);

    # Setup environment
    eval {
        ntfs_setup(mode  => 'ntfs',
                   ret_h => \%setup_h);
    };

    if ($@) {
        my $err = $@;
        # Clean environment
        ntfs_cleanup(%setup_h);
        logresult('FAIL', $err);
        return;
    }

    # Clean environment
    eval {
        ntfs_cleanup(%setup_h);
    };

    if ($@) {
        logresult('FAIL', $@);
    } else {
        logresult('PASS', 'testme(): Ntfs.pm');
    }
    return;
}

1;
