# this is basically -*- perl -*-
# $Id: /depot/prod/test/main/lib/CIFS/SMB3_Gap_Setup.pm
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
## @author olgav@netapp.com
## @status	design
# 

package CIFS::SMB3_Gap_Setup;


# Compiler directives.
#
use strict;
use warnings;
#
# Module import.
#
use File::Temp;
use Tharn qw($Log);
use Options;
use Params::Common;
use CIFS::Params_CIFSGX;
use base qw(Exporter);
use TestEnv::MinWin qw(execute_cmd);

use NACL::STask::VserverCifsShare;

use NACL::MTask::CIFS;

use NACL::C::NetworkInterface;
use NACL::C::VserverServicesDns;
use NACL::C::VserverCifsOptions;
use NACL::C::VserverExportPolicyRule;

use NACL::CS::VserverCifsSession;
use NACL::CS::VserverCifsSessionFile;
use NACL::CS::VserverLocks;
use NACL::CS::Volume;

use NATE::BaseException qw(:try);
use NACL::Exceptions::NoElementsFound;
use NACL::APISet::Exceptions::TimeoutException;

use NACL::GeneralUtils qw(arrange_by_create_time);

use Params::Validate qw(validate ARRAYREF SCALAR OBJECT UNDEF HASHREF);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our @EXPORT_OK = qw(
    cifs_server_create
    check_cifs_session_file_state
    check_oplock_sharelevel_lock
    check_cifs_session_state
    get_clients_ipv6
    delete_all_vol
    cleanup_client
);
#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub cifs_server_create(%);
sub check_cifs_session_file_state(%);
sub check_oplock_sharelevel_lock(%);
sub check_cifs_session_state(%);
sub get_clients_ipv6(%);
sub delete_all_vol(%);
sub cleanup_client(%);

#
# Global Variables.
#

my $Options = new Options();
$Options->set_default("CIFS_SERVER" => "cifsvs");

$Options->process();
my $Params=$Options->get_hash();


###############################################################################
# Subroutine  : check_cifs_session_file_state
# Input       : cifs session file object and options to check
# Return      : true/fails
# Description : This subroutine checks specific state for cifs session file command.
###############################################################################
sub check_cifs_session_file_state(%) {

    my %params = validate(
       @_,
       {
            cifs_session_file_state_obj => { isa => 'NACL::CS::VserverCifsSessionFile'},
            file_id             => { optional => 1, default => undef},
            connection_id       => { optional => 1, default => undef},
            session_id          => { optional => 1, default => undef},
            file_type           => { optional => 1, default => undef},
            open_mode           => { optional => 1, default => undef},
            hosting_aggregate   => { optional => 1, default => undef},
            hosting_volume      => { optional => 1, default => undef},
            share               => { optional => 1, default => undef},
            path                => { optional => 1, default => undef},
            share_mode          => { optional => 1, default => undef},
            range_locks         => { optional => 1, default => undef},
            continuously_available => { optional => 1, default => undef},
            reconnected         => { optional => 1, default => undef},

       }
    );

    my @compare_fields;
    my $differences = 0;
    my $CifsSessionFile_state = $params{cifs_session_file_state_obj};

    if ($params{file_id}) { push @compare_fields, 'file_id'; }
    if ($params{connection_id}) { push @compare_fields, 'connection_id'; }
    if ($params{session_id}) { push @compare_fields, 'session_id'; }
    if ($params{file_type}) { push @compare_fields, 'file_type'; }
    if ($params{open_mode}) { push @compare_fields, 'open_mode'; }
    if ($params{hosting_aggregate}) { push @compare_fields, 'hosting_aggregate'; }
    if ($params{hosting_volume}) { push @compare_fields, 'hosting_volume'; }
    if ($params{share}) { push @compare_fields, 'share'; }
    if ($params{path}) { push @compare_fields, 'path'; }
    if ($params{share_mode}) { push @compare_fields, 'share_mode'; }
    if ($params{range_locks}) { push @compare_fields, 'range_locks'; }
    if ($params{continuously_available}) { push @compare_fields, 'continuously_available'; }
    if ($params{reconnected}) { push @compare_fields, 'reconnected'; }

    $Log->comment("\n ----> Comparing using fields: " . join( ',', @compare_fields ));

    foreach my $field (@compare_fields) {
        if ($params{$field} ne $CifsSessionFile_state->$field) {
            NATE::BaseException->throw(
	                    "$field for cifs session file command is $CifsSessionFile_state->$field compare to  $params{$field}, unable to compare" );
            $differences = 1;
        }
    }

    if ($differences) {
        NATE::BaseException->throw("Compare Failed. \n");
    }else {
        $Log->comment("\n ---------------- CIFS session file state is as expected!");
    }
}

##############################################################################
# Subroutine  : check_cifs_session_state
# Input       : cifs session object and options to check
# Return      : true/fails
# Description : This subroutine checks specific state for cifs session file command.
###############################################################################
sub check_cifs_session_state(%) {

    my %params = validate(
       @_,
       {
            cifs_session_state_obj => { isa => 'NACL::CS::VserverCifsSession'},
            protocol_version            => { optional => 1, default => undef},
            files                       => { optional => 1, default => undef},
            session_id                  => { optional => 1, default => undef},
            connection_id               => { optional => 1, default => undef},
            hostname                    => { optional => 1, default => undef},
            connected_time              => { optional => 1, default => undef},
            idle_time                   => { optional => 1, default => undef},
            address                     => { optional => 1, default => undef},
            shares                      => { optional => 1, default => undef},
            cifs_user                   => { optional => 1, default => undef},
            auth_mechanism              => { optional => 1, default => undef},
            unix_user                   => { optional => 1, default => undef},
            continuously_available      => { optional => 1, default => undef},
            ca_protection               => { optional => 1, default => undef},
            lif_address                 => { optional => 1, default => undef},
            windows_user                => { optional => 1, default => undef},

       }
    );

    my @compare_fields;
    my $differences = 0;
    my $CifsSession_state = $params{cifs_session_state_obj};

    if ($params{protocol_version}) { push @compare_fields, 'protocol_version'; }
    if ($params{connection_id}) { push @compare_fields, 'connection_id'; }
    if ($params{session_id}) { push @compare_fields, 'session_id'; }
    if ($params{files}) { push @compare_fields, 'files'; }
    if ($params{hostname}) { push @compare_fields, 'hostname'; }
    if ($params{connected_time}) { push @compare_fields, 'connected_time'; }
    if ($params{idle_time}) { push @compare_fields, 'idle_time'; }
    if ($params{shares}) { push @compare_fields, 'shares'; }
    if ($params{address}) { push @compare_fields, 'address'; }
    if ($params{cifs_user}) { push @compare_fields, 'cifs_user'; }
    if ($params{auth_mechanism}) { push @compare_fields, 'auth_mechanism'; }
    if ($params{continuously_available}) { push @compare_fields, 'continuously_available'; }
    if ($params{unix_user}) { push @compare_fields, 'unix_user'; }
    if ($params{ca_protection}) { push @compare_fields, 'ca_protection'; }
    if ($params{lif_address}) { push @compare_fields, 'lif_address'; }
    if ($params{windows_user}) { push @compare_fields, 'windows_user'; }

    $Log->comment("\n ----> Comparing using fields: " . join( ',', @compare_fields ));

    foreach my $field (@compare_fields) {
        if ($params{$field} ne $CifsSession_state->$field) {
            NATE::BaseException->throw(
	                    "$field for cifs session file command is $CifsSession_state->$field compare to  $params{$field}, unable to compare" );
            $differences = 1;
        }
    }

    if ($differences) {
        NATE::BaseException->throw("Compare Failed. \n");
    }else {
        $Log->comment("\n ---------------- CIFS session state is as expected!");
    }
}

##############################################################################
# Subroutine  : check_oplock_sharelevel_lock
# Input       : 
# Return      : 1/0
# Description : This subroutine executes NACL::CS::VserverLocks->fetch and 
#               checks output.
###############################################################################
 sub check_oplock_sharelevel_lock(%) {

    my %params = validate(
       @_,
       {
            node                        => { optional => 1, default => undef},
            vserver                     => { optional => 1, default => undef},
            client_ip                   => { optional => 1, default => undef},
            path                        => { optional => 1, default => undef},
            oplock                      => { optional => 1, default => undef},
            smb_open_type               => { optional => 1, default => undef},
            smb_connect_state           => { optional => 1, default => undef},

       }
    );
    my $VserverLock_obj;
    my $return_code = 1;
    my $exception_object;
    
    try {
        $VserverLock_obj  = NACL::CS::VserverLocks->fetch(command_interface       => $params{node},
                                                      filter => {vserver          => $params{vserver}, 
                                                                 'client-address' => $params{client_ip},
                                                                 path             => $params{path},
                                                                 type             => "op-lock",
                                                                }
                                                      );

        if (($VserverLock_obj->oplock_level() eq $params{oplock}) && ($VserverLock_obj->smb_connect_state() eq "connected")) {
            $Log->comment("----> Vserver locks shows $params{oplock} oplock for files created by the user. Oplock level: " . $VserverLock_obj->oplock_level() . " ; SMB connect state: " . $VserverLock_obj->smb_connect_state());

        }else {
            $Log->comment("***** ----> Vserver oplocks show should be $params{oplock}. SMB connect state should be connected");
            $return_code = 0;
        }

        $VserverLock_obj  = NACL::CS::VserverLocks->fetch(command_interface       => $params{node},
                                                      filter => {vserver          => $params{vserver},
                                                                 'client-address' => $params{client_ip},
                                                                 path             => $params{path},
                                                                 type             => "share-level",
                                                                }
                                                      );

        if (($VserverLock_obj->smb_open_type() eq $params{smb_open_type}) && ($VserverLock_obj->smb_connect_state() eq $params{smb_connect_state}) ) {
            $Log->comment("----> Vserver share level lock shows: smb open type - " . $VserverLock_obj->smb_open_type() . " ; smb connect state - " . $VserverLock_obj->smb_connect_state());

        }else {
            $Log->comment("***** ----> Vserver locks show should display $params{smb_open_type} share-locks for files created by the user. SMB connect state should be $params{smb_connect_state}");
            $return_code = 0;
        }

    }catch NACL::Exceptions::NoElementsFound with {
        $exception_object = shift;
	$Log->comment( "\n ***** Exception raised!: error" . $exception_object->text() );
        $return_code = 0;   
    };


    return $return_code;
 
 }

##############################################################################
# Subroutine  : cifs_server_create
# Input       : 
# Return      : 1/0
# Description : This subroutine create cifs server for vserver on node.
###############################################################################
 sub cifs_server_create(%) {

    my %params = validate(
       @_,
       {
            node           => { optional => 1, default => undef},
            vserver        => { optional => 1, default => $Params->{VSERVER} },
            client         => { optional => 1, default => undef},
            cifs_server    => { optional => 1, type => SCALAR, default => $Params->{CIFS_SERVER} },
            cifs_domain_fullname     => { optional => 1, type => SCALAR, default => $Params->{CIFS_DOMAIN_LONG} },
            cifs_domain_shortname    => { optional => 1, type => SCALAR, default => $Params->{CIFS_DOMAIN} },
            ad_server      => { optional => 1, type => ARRAYREF, default => $Params->{AD_SERVER} },
            ad_server_ip   => { optional => 1, type => ARRAYREF, default => $Params->{AD_SERVER_IP} },
            vserver        => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
            ad_server_user => { optional => 1, type => SCALAR, default => $Params->{AD_SERVER_USER} },
            ad_server_password       => { optional => 1, type => SCALAR, default => $Params->{AD_SERVER_PASSWORD} },
            rootshare      => { optional => 1, type => SCALAR, default => "ROOTSHARE" },
            position       => { optional => 1, type => SCALAR, default => $Params->{POSITION} },
            nis_server     => { optional => 1, type => SCALAR, default => $Params->{NIS_SERVER} },
            nis_domain     => { optional => 1, type => SCALAR, default => $Params->{NIS_DOMAIN} },
            dns_server_ip  => { optional => 1, type => ARRAYREF, default => $Params->{DNS_SERVER_IP} },
            dns_domain     => { optional => 1, type => ARRAYREF, default => $Params->{DNS_DOMAIN} },
            timezone       => { optional => 1, type => SCALAR, default => $Params->{TIMEZONE} },
            nsswitch       => { optional => 1, type => SCALAR, default => $Params->{NS_SWITCH} },
            ou             => { optional => 1, type => SCALAR, default => $Params->{OU} },
            export_policy  => { optional => 1, default => $Params->{EXPORT_POLICY}},
            client_match   => { optional => 1, default => "0.0.0.0/0"},
            if_exists      => { optional => 1, default => "reuse"},
            check_two_lifs => { optional => 1, default => "false"},
       }
    );

    my $return_code = 1;
    my ($exception_object, $output, $VserverServicesDns, $ip_obj, $ip, @data_lif_obj, $data_lif, $lif, $node, );    
    my $map_patt_user = "$params{cifs_domain_shortname}\\\\$params{ad_server_user}";
    my @dns_server_IP = @{$params{dns_server_ip}};
    my @dns_domain = map(lc, @{$params{dns_domain}});
    my $client_obj = $params{client};

    try {

        $VserverServicesDns  = NACL::C::VserverServicesDns->find(command_interface => $params{node},
                                                      filter => {
                                                                 vserver    => $params{vserver},
                                                                }
                                          );
        
        $VserverServicesDns->modify(command_interface => $params{node},
                                             vserver           => $params{vserver},
                                             state             => "enabled",
                                            );
    }catch NACL::Exceptions::NoElementsFound with {

        NACL::C::VserverServicesDns->create(command_interface => $params{node},
                                            vserver           => $params{vserver},
                                            domains           => [@dns_domain],
                                            state             => "enabled",
                                            'name-servers'    => [@dns_server_IP],
                                           );
    };

    $node = $params{node}->get_one_state_attribute('node');

    try {
        NACL::C::VserverExportPolicyRule->find(command_interface =>  $params{node},
                                               filter => {vserver =>  $params{vserver},}
                                               );
    }
    catch NACL::Exceptions::NoElementsFound with {

        NACL::C::VserverExportPolicyRule->create(command_interface =>  $params{node},
                                                 vserver           =>  $params{vserver},
                                                 policyname        =>  $params{export_policy},
                                                 clientmatch       =>  $params{client_match},
                                                 rorule=> ['any'],
                                                 rwrule=> ['any'],
                                                 );
    };

    if ($params{check_two_lifs} eq "true") {
        try {
            @data_lif_obj = NACL::C::NetworkInterface->find(command_interface => $params{node},
                                                                filter => {
                                                                           vserver  => $params{vserver},
                                                                           role     => "data",
                                                                           }
                                                                );
            $lif = 0;
            foreach $data_lif (@data_lif_obj) {
                $output = $data_lif->get_one_state_attribute('home-node');
                if ($output eq $node) {
                    $output = $data_lif->get_one_state_attribute('status-admin');
                    if ($output ne "up") { 
                        $data_lif->modify('status-admin' => 'up');
                    }
                    $output = $data_lif->get_one_state_attribute('is-home');
                    if ($output eq "false") {
                        $data_lif->revert();
                    }
                    $lif++;              
                }     
                if ($lif > 1 ) {last;}     
            }

        }catch NATE::BaseException with {
 
            my $exception_object = shift;
            $Log->comment("\n ***** Exception raised!: " . $exception_object->text(). "\n" );

        };
    }

    eval {
        NACL::MTask::CIFS->setup(command_interface => $params{node},
                                 vserver           => $params{vserver},
                                 if_exists         => $params{if_exists},
                                 'cifs-server'     => uc $params{cifs_server},
                                 'cifs_domain'     => lc $params{cifs_domain_fullname},
                                 'dns_domain'      => $dns_domain[0],
                                 username          => $params{ad_server_user},
                                 password          => $params{ad_server_password},
                                 'unix_groups'     => [
                                                       {
                                                        'name' => "root",
                                                        'id' => "0",
                                                       },
                                                       {
                                                        'name' => "nobody",
                                                        'id' => "65535",
                                                       },
                                                       {
                                                        'name' => "pcuser",
                                                        'id' => "65534",
                                                       },
                                                      ],
                                 'unix_users'      => [
                                                       {
                                                        'user' => "root",
                                                        'id' => "0",
                                                        'primary-gid' => "0",
                                                       },
                                                       {
                                                        'user' => "nobody",
                                                        'id' => "65535",
                                                        'primary-gid' => "65535",
                                                       },
                                                       {
                                                        'user' => "pcuser",
                                                        'id' => "65534",
                                                        'primary-gid' => "65534",
                                                       },
                                                      ],
                                 'vserver_name_mapping' => 
                                                      [ 
                                                       {
                                                        'pattern'     => $map_patt_user,
                                                        'direction'   => "win-unix",
                                                        'replacement' => "root",
                                                        'position'    => $params{position},
                                                       },
                                                      ],
                                 'vserver_name_mapping' =>
                                                      [ 
                                                       {
                                                        'pattern'     => "root",
                                                        'direction'   => "unix-win",
                                                        'replacement' => $map_patt_user,
                                                        'position'    => $params{position},
                                                       },
                                                      ],
                                );
    };
    if ($@) {

        $return_code = 0;
        $Log->warn("\"***** ---------------- Failed to create $params{cifs_server} CIFS server on $params{vserver} vserver." . $@);

    }else {

        $Log->comment("\n ---------------- $params{cifs_server} CIFS server created successfully on $params{vserver} vserver.\n");

        $Log->comment("\n ---------------- Modify CIFS server options for exportpolicy.\n");
        NACL::C::VserverCifsOptions->modify(command_interface   => $params{node},
                                            vserver             => $params{vserver},
                                            'is-exportpolicy-enabled' => "true");
        eval {

            NACL::STask::VserverCifsShare->delete(command_interface => $params{node},
                                                  vserver           => $params{vserver},
                                                  'share-name'      => '!*$',
                                                 );
        };

        NACL::STask::VserverCifsShare->create(command_interface => $params{node},
                                              vserver           => $params{vserver},
                                              'share-name'      => $params{rootshare},
                                              path              => "/",
                                              nacltask_if_exists => "purge",
                                             );
        sleep 30;

        $Log->comment("\n ---------------- Net View $params{cifs_server} cifs server from client.\n");
                
        $ip_obj = NACL::C::NetworkInterface->find(command_interface => $params{node},
                                                     filter => { vserver => $params{vserver}, role => "data"}
                                                    );

        $ip = $ip_obj->get_one_state_attribute('address');
 
        $output = $client_obj->execute(command => "net view $ip");
                
        if ($output !~ /$params{rootshare}/i) {

            $Log->warn("***** ---------------- Cannot net view $params{cifs_server} CIFS server by IP.");
            $return_code = 0;

        }else {

            $output = $client_obj->execute(command => "net view $params{cifs_server}");
                
            if ($output !~ /$params{rootshare}/i) {
                $Log->warn("---------------- Create DNS host for $params{cifs_server} CIFS server.\n ***** ---------------- Cannot net view $params{cifs_server} CIFS server by name.");
            }                
        }
   }

    return $return_code;
 
 }

##############################################################################
# Subroutine  : get_clients_ipv6
# Input       : 
# Return      : client ipv6
# Description : This subroutine retuns windows client ipv6.
###############################################################################
 sub get_clients_ipv6(%)
 {

    my %params = validate(
       @_,
       {
            client       => { optional => 1, default => undef},
       }
    );
    my (@client_config, @client_ipv6, $output, @output); 

    my $Client_obj = $params{client};
    
    $output = $Client_obj->execute(command => "ipconfig");

    if ($output =~ /Temporary IPv6 Address/i) {
        $Log->comment(" ");
        $Log->comment("************************************ WARNING ***********************************");
        $Log->comment("Disable Temp IPv6 address on client to make sure only real IPv6 address is used.\n (Run 'netsh interface ipv6 set....' commands.)\n");
        $Log->comment("********************************************************************************");
        $Log->comment(" ");
    }
    
    @output = split(/\n/, $output);

    foreach my $line (@output) {
    
        if ($line =~ /^\s+(IPv6 Address|Temporary IPv6 Address)/i) {
#        if ($line =~ /^\s+IPv6 Address/i) {
            @client_config = split( /:\s+/, $line, 2);
            chomp($client_config[1]);
            @client_ipv6 = split( /[(]/, $client_config[1], 2);
#            last;
        }	
    }
     
    return $client_ipv6[0];
 
 }

##############################################################################
# Subroutine  : delete_all_vol
# Input       : aggr
# Return      : 
# Description : Delete all but root volumes on aggregate.
##############################################################################
sub delete_all_vol(%) {
    my %params = validate(
       @_,
       {
            node    => { optional => 1, default => undef},
            vserver => { optional => 1, default => undef},
            aggr    => { optional => 1, default => undef},
       }
    );

    my ($exception_object, %common_opts, @cs_objects, @volumes );

    try {
        @cs_objects = NACL::CS::Volume->fetch(
                                             command_interface=>$params{node},
                                             filter => { aggregate => $params{aggr},
                                                         vserver   => $params{vserver},
                                                         vsroot    => 'false'
                                                       },
                                             requested_fields => [qw(create-time)]
                                            );
    }
    catch NACL::Exceptions::NoElementsFound with {
        $Log->comment( "---------------- NO volumes on $params{aggr}");
        return 1;
    };

    my @arranged_cs_objects = arrange_by_create_time(volume_cs_objects => \@cs_objects );

    foreach my $cs (@arranged_cs_objects) {
         push @volumes, $cs->get_task_instance();
    }

    foreach my $vol (@volumes) {
        my $vol_name = $vol->get_one_state_attribute('volume');
        my $state = $vol->get_one_state_attribute('state');
        $Log->comment( "----------------Deleting $vol_name volume with $state state");
        if ($state eq "restricted") {
            $vol->modify(state => "online");    
        }
        try{
            $vol->purge(%common_opts);
        }
        catch NACL::APISet::Exceptions::TimeoutException with {
            $exception_object = shift;
            $Log->comment( "---------------- " . $exception_object->text());
            return 0;
        };
    }

    return 1;
}
 
################################################################################
################################################################################
## @name :: cleanup_client                                                     #
## @summary This subroutine does cleanup mapped shares on client               #
## @return  Nothing                                                            #
################################################################################

sub cleanup_client(%) {
    my %params = validate(
       @_,
       {
            cifs_server => { optional => 1, default => undef},
            data_lif    => { optional => 1, default => undef},
            client      => { optional => 1, default => undef},
       }
    );

    my $Client_obj = $params{client};
    my $cifs_server = $params{cifs_server};
    my $Cifs_server_ip = $params{data_lif};

    $Log->comment("------- DATA network interface is $Cifs_server_ip and cifs server is $cifs_server.");
    
        ## DELETING PRE-MAPPED FILER SHARES IF ANY BEFORE STARTING THE TESTS.
        $Log->comment("Deleting the Filer shares if any mapped earlier before starting the tests.");
        my $output = $Client_obj->execute(command => "net use");

        while ($output =~ /.*([A-Z]:).*($Cifs_server_ip|$cifs_server)/i) {
            $output = $Client_obj->execute(command => "net use $1 /DELETE /Y");
            if ($output !~ /$1 was deleted successfully/i) {
                    die("Failed to delete mapped drive $1 from the Windows client.\n\nDelete all the mapped shares of $cifs_server Filer from the Windows client and rerun the script."); 
            }
            $output = $Client_obj->execute(command => "net use");
        }
        
        $Client_obj->execute(command => "ipconfig /flushdns");
}

1;
