# $Id$

## @summary Library to test CIFS acl inheritanc.
## @author witte@netapp.com

package CIFS::Sanity::sdtest;

use strict;
use CIFS::Security qw(:all);
use CIFS::FileOp qw(:all);
use Tharn;

BEGIN {
  use Module::Common;
  Module::Common::setup_export_tags(
    ops => [ qw(sd_inheritance_test) ]
    );
}

#-------------------------------------------------------------
#  Subroutine to test security.
#
#  Params:
#    fq test share
#    fq reference share
#    test cases
#    verbose

my $filer;
my $reference_host;
my $test_cases;
my $verbose;
my $global_pass = 1;
my $reference_drive;
my $test_drive;

my $test_dir;
my $test_dir_ref;

my $test_idir;
my $test_idir_ref;

my $test_ifile;
my $test_ifile_ref;

my @CommandMasks;
my @CommandFlags;

sub sd_inheritance_test
  {
    $filer = shift;
    $reference_host = shift;
    $test_cases = shift;
    $verbose = shift;


    #--------------------------------------------------------------
    # Map the test drive and the reference drive.  Fail out if we
    # cannot map either of them.
    #-------------------------------------------------------------
    print "mapping: \\\\$reference_host\\sanity-ntfs\n";
    $reference_drive =  map_drive("\\\\$reference_host\\sanity-ntfs",
				    "cifs-sanity\\admintestuser", "admintestuser");

    print "mapping: \\\\$filer\\sanity-ntfs\n";
    $test_drive      =  map_drive("\\\\$filer\\sanity-ntfs",
				     "cifs-sanity\\admintestuser", "admintestuser");


    if ($test_drive eq undef)
      {
	logresult('fail', "Could not map \\\\$filer\\sanity-ntfs");
	$global_pass = 0;
      }

    if ($reference_drive eq undef)
      {
	logresult('fail', "Could not map \\\\$reference_drive\\sanity-ntfs");
	$global_pass = 0;
      }

    if (!$global_pass)
      {
	&Cleanup;
	return 0;
      }


    #------------------------------------------------------------
    # Define all of the paths we will use during our tests.
    #------------------------------------------------------------
    $test_dir     = "$test_drive/InheritanceTestDir";
    $test_dir_ref = "$reference_drive/InheritanceTestDir";

    $test_idir     = "$test_dir/NewDir";
    $test_idir_ref = "$test_dir_ref/NewDir";

    $test_ifile     = "$test_dir/NewFile.txt";
    $test_ifile_ref = "$test_dir_ref/NewFile.txt";


    &CreateTestDirectories;
    if (!$global_pass)
      {
	&Cleanup;
	return 0;
      }


    &LoadCommands;
    if (!$global_pass)
      {
	&Cleanup;
	return 0;
      }

    &DoTests;
    &Cleanup;

    return $global_pass;

}



#-------------------------------------------------------------
#-------------------------------------------------------------
#-------------------------------------------------------------


sub CreateTestDirectories
  {
    if (!cifsFileCreateDirectory(sPathname => "$test_dir"))
      {
	&Cleanup;
	
	die("Could not create top level test directory.");
      }

    if (!cifsFileCreateDirectory(sPathname => "$test_dir_ref"))
      {
	&Cleanup;
	die("Could not create top level reference directory.");
      }
}

sub DoTests
  {
    my $i;
    my $pass;
    my $logresult;

    $_ = $test_cases;

    #
    # First try all combinations of inheritance and mask mix.
    #
    if (/all/i || /enumerate/i)
      {
	$logresult = 'pass';
	print("[Starting ENUMERATE inheritance test]\n");

	for ($i=0; $i <= $#CommandMasks; $i++)
	  {
	    $pass = &SetTestSecurity($i);
	    &verbose_print("Setting security to $CommandFlags[$i] $CommandMasks[$i]",$pass);

	    if ($pass)
	      {
		$pass = &TryInheritance(1);
		&verbose_print("Attempting inheritance of $CommandFlags[$i] $CommandMasks[$i]",
			       $pass);
	      }

	    $pass or last;
	  }

	$pass or $logresult = 'fail';
	logresult($logresult, "Inheritance ENUMERATION test");
      }


    if (/all/i || /Default SD/i)
      {
	$logresult = 'pass';
	print("[Starting DEFAULT SD inheritance test]\n");
	#
	# Now test the case where there are no inheritable aces.  This test makes sure that
	# We set the same default as NTFS.
	#
	$pass = &SetDefaultSDTestSecurity;
	&verbose_print("Setting up default SD inheritance scenario", $pass);
	
	if ($pass)
	  {
	    $pass = &TryInheritance(1);
	    &verbose_print("Trying default SD inheritance", $pass);
	  }
	$pass or $logresult = 'fail';
	logresult($logresult, "DEFAULT SD test");
      }

    if (/all/i || /Large SD 1/i)
      {
	$logresult = 'pass';
	print("[Starting LARGE SD 1 inheritance test]\n");
	#
	# Try a really large SD that we CAN inherit. The 2030 value
	# as arrived at through trial an error with Windows CP clients.
	#
	$pass = &SetLargeSD(2030, "OI|CI|NP");
	&verbose_print("Setting up a large but inheritable SD", $pass);
	
	if ($pass)
	  {
	    $pass = &TryInheritance(0);
	    &verbose_print("Trying large but inheritable inheritance", $pass);
	  }

	$pass or $logresult = 'fail';
	logresult($logresult, "LARGE SD test number 1");
      }

    if (/all/i || /Large SD 2/i)
      {
	$logresult = 'pass';
	print("[Starting LARGE SD 2 inheritance test]\n");
	#
	# Try a really large SD that expands when inherited.
	# as arrived at through trial an error with Windows CP clients.
	#
	$pass = &SetLargeSD(2030, "OI|CI|IO");
	&verbose_print("Setting up a large but inheritable SD", $pass);
	
	if ($pass)
	  {
	    $pass = &TryInheritance(0);
	    &verbose_print("Trying large but inheritable inheritance", $pass);
	  }

	$pass or $logresult = 'fail';
	logresult($logresult, "LARGE SD test number 2");

      }

    $global_pass = $pass;
}


sub Cleanup
  {
    &CleanDir($test_dir);
    &CleanDir($test_dir_ref);
    &unmap_drive($test_drive);
    &unmap_drive($reference_drive);
  }


sub TryInheritance
  {
    my $do_cacls = shift;

    my $pass = 1;
    my $match_pass;

    my $test_dir_made = 0;
    my $test_dir_ref_made = 0;
    my $test_file_made = 0;
    my $test_file_ref_made = 0;

    my $test_file_fh;
    my $test_file_ref_fh;

    # Create a file and a directory in each of the test directory
    # and the test directory. Each should succeed in both, or fail
    # in both, anything else is an error.
    $test_dir_made = cifsFileCreateDirectory(sPathname => "$test_idir");
    &verbose_print("Creating $test_idir", $test_dir_made);

    $test_dir_ref_made = cifsFileCreateDirectory(sPathname => "$test_idir_ref");
    &verbose_print("Creating $test_idir_ref", $test_dir_ref_made);


    $test_file_fh = cifsFileCreate(sFilename => "$test_ifile",
				   sCreateMode => "te");
    $test_file_made = ($test_file_fh ne undef);
    &verbose_print("Creating $test_ifile", $test_file_made);
    if ($test_file_made)
      {
	cifsFileClose(hFile => $test_file_fh);
      }


    $test_file_ref_fh = cifsFileCreate(sFilename => "$test_ifile_ref",
				       sCreateMode => "te");
    $test_file_ref_made = ($test_file_ref_fh ne undef);
    &verbose_print("Creating $test_ifile_ref", $test_file_ref_made);
    if ($test_file_ref_made)
      {
	cifsFileClose(hFile => $test_file_ref_fh);
      }


    if (!$test_dir_made && $test_dir_ref_made)
      {
	print("Could not create child dir on test drive, but could on reference drive.\n");
	$pass = 0;
	goto cleanup;
      }
    if ($test_dir_made && !$test_dir_ref_made)
      {
	print("Could not create child dir on reference drive, but could on test drive.\n");
	$pass = 0;
	goto cleanup;
      }
    if (!$test_file_made && $test_file_ref_made)
      {
	print("Could not create child file on test drive, but could on reference drive.\n");
	$pass = 0;
	goto cleanup;
      }
    if ($test_file_made && !$test_file_ref_made)
      {
	print("Could not create child file on reference drive, but could on test drive.\n");
	$pass = 0;
	goto cleanup;
      }

    # Now compare the SDs on the files, if the files were created.
    if ($test_dir_made)
      {
	$match_pass = cifsSecurityMatch(sSource => "$test_idir",
					sTarget => "$test_idir_ref",
				        uType => DACL|OWNER);
	&verbose_print("Trying to match directories", $match_pass);
	if (!$match_pass)
	  {
	    $pass = 0;
	    if ($do_cacls)
	      {
		&output_cacls("Directory", $test_idir, $test_idir_ref);
	      }
	  }
      }

    if ($test_file_made)
      {
	$match_pass = cifsSecurityMatch(sSource => "$test_ifile",
					sTarget => "$test_ifile_ref",
				        uType => DACL|OWNER);
	&verbose_print("Trying to match files", $match_pass);
	if (!$match_pass)
	  {
	    $pass = 0;
	    if ($do_cacls)
	      {
		&output_cacls("File", $test_ifile, $test_ifile_ref);
	      }
	  }
      }

#     $pass or die;
#     die "he he\n";

  cleanup:

    if ($test_dir_made)             { &CleanDir($test_idir);        }
    if ($test_dir_ref_made)         { &CleanDir($test_idir_ref);    }
    if ($test_file_made)            { &CleanFile($test_ifile);      }
    if ($test_file_ref_made)        { &CleanFile($test_ifile_ref);  }

    $global_pass = $pass;
    return $pass;
  }


#---------------------------------------------------------------
# Subroutine to get rid of a directory that may have
# weird ACLs on it.  Set ACL first, then delete directory.
#---------------------------------------------------------------
sub CleanDir
  {
    my $dir_path = $_[0];

    my $ace;
    my @aceList;

    $ace = cifsSecurityBuildACE(sUser => "S-1-1-0",
				sType => "ALLOW",
				sAccessMask => "F");
    push(my @aceList, $ace);
    cifsSecuritySet(sPathname => "$dir_path",
		    uAceList => \@aceList,
		    bInherit => 0);

    cifsFileDeleteDirectory(sPathname => "$dir_path");
  }

#---------------------------------------------------------------
# Subroutine to get rid of a file that may have
# weird ACLs on it.  Set ACL first, then delete directory.
#---------------------------------------------------------------
sub CleanFile
  {
    my $file_path = $_[0];

    my $ace;
    my @aceList;

    $ace = cifsSecurityBuildACE(sUser => "S-1-1-0",
				sType => "ALLOW",
				sAccessMask => "F");
    push(my @aceList, $ace);
    cifsSecuritySet(sPathname => "$file_path",
		    uAceList => \@aceList,
		    bInherit => 0);

    cifsFileDelete(sFilename => "$file_path");
  }


#-------------------------------------------------------------
# Subroutine to set SD on the test directory on both
# that test and reference machine.  We use the "S-1-5-..." form
# for SIDS so that we don't have to look anything up.  We also
# use a hard coded SID for our main test ACE, since we don't
# want to collide with the test user.
#--------------------------------------------------------------
sub SetTestSecurity
{
  my $test_index = shift;

  my $test_dir_set = 0;
  my $test_dir_ref_set = 0;
  my $pass = 1;

  my $ace;
  my @aceList;

  my $access_mask  =  $CommandMasks[$test_index];
  my $access_flags =  $CommandFlags[$test_index];

  # Always put DENY ACES first, since this is what Windows clients do.
  # Interleave to make the server collapse them.
  $ace = cifsSecurityBuildACE(sUser => "S-1-5-21-11111-22222-44444",
			      sType => "DENY",
			      sFlags => "$access_flags",
			      sAccessMask => "$access_mask");
  push(@aceList, $ace);

  # Creator owner
  $ace = cifsSecurityBuildACE(sUser=> "S-1-3-0",
			      sType => "DENY",
			      sFlags => "$access_flags",
			      sAccessMask => "FILE_WRITE_EA");
  push(@aceList, $ace);

  $ace = cifsSecurityBuildACE(sUser => "S-1-5-21-11111-22222-44444",
			      sType => "DENY",
			      sFlags => "$access_flags",
			      sAccessMask => "FILE_WRITE_EA");
  push(@aceList, $ace);

  # Creator group
  $ace = cifsSecurityBuildACE(sUser=> "S-1-3-1",
			      sType => "DENY",
			      sFlags => "$access_flags",
			      sAccessMask => "FILE_WRITE_EA");
  push(@aceList, $ace);


  # Always give everyone full control, so we can clean up.
  $ace = cifsSecurityBuildACE(sUser => "S-1-1-0",
			      sType => "ALLOW",
			      sFlags => "FULL_INHERIT",
			      sAccessMask => "F");
  push(@aceList, $ace);


  # Now add the other allow aces.
  $ace = cifsSecurityBuildACE(sUser => "S-1-5-21-11111-22222-33333",
			      sType => "ALLOW",
			      sFlags => "$access_flags",
			      sAccessMask => "$access_mask");
  push(@aceList, $ace);


  $ace = cifsSecurityBuildACE(sUser=> "S-1-3-0",
			      sType => "ALLOW",
			      sFlags => "$access_flags",
			      sAccessMask => "$access_mask");
  push(@aceList, $ace);

  $ace = cifsSecurityBuildACE(sUser => "S-1-5-21-11111-22222-33333",
			      sType => "ALLOW",
			      sFlags => "$access_flags",
			      sAccessMask => "FILE_WRITE_EA");
  push(@aceList, $ace);

  $ace = cifsSecurityBuildACE(sUser=> "S-1-3-1",
			      sType => "ALLOW",
			      sFlags => "$access_flags",
			      sAccessMask => "$access_mask");
  push(@aceList, $ace);


  # Set the SD on the directory.  Either we should be able to set on both or
  # neither.
  $test_dir_set = cifsSecuritySet(sPathname => "$test_dir",
				  uAceList => \@aceList,
				  bInherit => 0);

  $test_dir_ref_set = cifsSecuritySet(sPathname => "$test_dir_ref",
				      uAceList => \@aceList,
				      bInherit => 0);

  if ($test_dir_set && !$test_dir_ref_set)
    {
      &verbose_print("Able to set test directory, but not reference directory", 0);
      $pass = 0;
    }
  if (!$test_dir_set && $test_dir_ref_set)
    {
      &verbose_print("Able to set reference directory, but not test directory", 0);
      $pass = 0;
    }

#  $pass or die;

  $global_pass = $pass;
  return $pass;
}

#-------------------------------------------------------------
# Subroutine to set SD on the test directory on both
# that test and reference machine.  We use the "S-1-5-..." form
# for SIDS so that we don't have to look anything up.  We also
# use a hard coded SID for our main test ACE, since we don't
# want to collide with the test user.
#--------------------------------------------------------------
sub SetDefaultSDTestSecurity
{
  my $test_dir_set = 0;
  my $test_dir_ref_set = 0;
  my $pass = 1;

  my $ace;
  my @aceList;

  &verbose_print("Setting default test security: no inheritable ACEs");

  # Always give everyone full control
  $ace = cifsSecurityBuildACE(sUser => "S-1-1-0",
			      sType => "ALLOW",
			      sFlags => "",
			      sAccessMask => "F");
  push(@aceList, $ace);


  $test_dir_set = cifsSecuritySet(sPathname => "$test_dir",
				  uAceList => \@aceList,
				  bInherit => 0);

  $test_dir_ref_set = cifsSecuritySet(sPathname => "$test_dir_ref",
				      uAceList => \@aceList,
				      bInherit => 0);

  if ($test_dir_set && !$test_dir_ref_set)
    {
      &verbose_print("Able to set test directory, but not reference directory", 0);
      $pass = 0;
    }
  if (!$test_dir_set && $test_dir_ref_set)
    {
      &verbose_print("Able to set reference directory, but not test directory.", 0);
      $pass = 0;
    }

  $global_pass = $pass;
  return $pass;
}


#-------------------------------------------------------------
# Subroutine to setup an arbitrarily large SD
#--------------------------------------------------------------
sub SetLargeSD
{
  my $num_aces = shift;
  my $inheritance_flags = shift;

  my $test_dir_set = 0;
  my $test_dir_ref_set = 0;
  my $pass = 1;
  my $i;

  my $ace;
  my @aceList;

  $ace = cifsSecurityBuildACE(sUser => "S-1-1-0",
			      sType => "ALLOW",
			      sFlags => "CI|OI|NP",
			      sAccessMask => "F");
  push(@aceList, $ace);

  # We want to add a bunch of allow ACEs, each of which only
  # is inherited one level.
  for ($i=0; $i < $num_aces; $i++)
    {
      my $sid = "S-1-5-21-11111-22222-$i";

      $ace = cifsSecurityBuildACE(sUser => $sid,
				  sType => "ALLOW",
				  sFlags => $inheritance_flags,
				  sAccessMask => "FILE_GENERIC_READ");
      push(@aceList, $ace);
    }


  $test_dir_set = cifsSecuritySet(sPathname => "$test_dir",
				  uAceList => \@aceList,
				  bInherit => 0);

  $test_dir_ref_set = cifsSecuritySet(sPathname => "$test_dir_ref",
				      uAceList => \@aceList,
				      bInherit => 0);

  if ($test_dir_set && !$test_dir_ref_set)
    {
      &verbose_print("Able to set test directory, but not reference directory", 0);
      $pass = 0;
    }
  if (!$test_dir_set && $test_dir_ref_set)
    {
      &verbose_print("Able to set reference directory, but not test directory.", 0);
      $pass = 0;
    }

  $global_pass = $pass && $test_dir_set;
  return $global_pass;
}



#--------------------------------------------------------
# Subroutine to load the set of test cases we are going
# to run during inheritance testing.
#--------------------------------------------------------
sub LoadCommands
  {

    # Don't need to try all access masks, just the ones that
    # are special. To these we add FILE_WRITE_EA just for 
    # fun.
    my @AceAccessMasks = ("FILE_GENERIC_READ",
			  "FILE_GENERIC_WRITE",
			  "FILE_GENERIC_EXECUTE");

    my @AceFlags = ("CI",
		    "OI",
		    "IO",
		    "NP");
    my $mask;
    my $flags;

    foreach $mask (@AceAccessMasks)
      {
	for (my $i = 1; $i <= 16; $i++)
	  {
	    my $dash = "";
	    my $flags = "";

	    if ($i & 1)   { $flags .= "$AceFlags[0]";      $dash = "|"; }
	    if ($i & 2)   { $flags .= "$dash$AceFlags[1]"; $dash = "|"; }
	    if ($i & 4)   { $flags .= "$dash$AceFlags[2]"; $dash = "|"; }
	    if ($i & 8)   { $flags .= "$dash$AceFlags[3]"; $dash = "|"; }

	    push @CommandMasks, $mask;
	    push @CommandFlags, $flags
	  }
      }
  }


sub map_drive
{
  my $path = shift;
  my $user = shift;
  my $passwd = shift;

  my $output = `net use * $path /user:$user $passwd 2>&1`;
  if ($output =~ m/error/gi) {
    return undef;
  }
  return $1 if ($output =~ m/.*Drive\s+(\w:).*/g);
}

sub unmap_drive
{
  my $drive = shift;

  system("net use /d $drive");
}

sub verbose_print
  {
    if ($verbose == 0) { return; }

    my $text = shift;
    my $pass = shift;


    if ($pass eq undef)
      {
	if ($verbose >= 2)
	  {
	    print "[$text]\n";
	  }
      }
    elsif ($pass && ($verbose >= 2))
      {
	print "[$text: ok]\n";
      }
    elsif (!$pass)
      {
	print "[$text: ***FAIL***]\n";
      }
}

# Routine to dump calcs comparison between the OK and the failure case.
# We only do this in verbose mode >= 2.  The strange "s ///" command is 
# because NATE dies on printed CACLS output for some unknown reason
# unless we massage the output.
sub output_cacls
  {
    my $type = shift;
    my $test = shift;
    my $reference = shift;

    my @output;
    my $line;


    if ($verbose < 1) { return; }

    print "---- Test $type -----------------------------------\n";
    @output = `cacls $test`;
    foreach $line (@output)
      {
	$line =~ s/\s\s\s\s/\ /g;               #Collapse some whitespace
	$line =~ s/[^ \d\w\(\)\[\]\:\-\\]//g;   # Need to do this, or NATE croaks?
	print "$line\n";
      }

    print "---- Reference $type ------------------------------\n";
    @output = `cacls $reference`;
    foreach $line (@output)
      {
	$line =~ s/\s\s\s\s/\ /g;               #Collapse some whitespace
	$line =~ s/[^ \d\w\(\)\[\]\:\-\\]//g;   # Need to do this, or NATE croaks?
	print "$line\n";
      }
  }

1;
