# $Id$

## @summary Library for CIFS file security operations from NATE on Windows
## @author mueller@netapp.com
## @pod here

=head1 NAME

=head1 SYNOPSIS

    use CIFS::Security qw(:all);

=head1 ROUTINES

=cut

package CIFS::Security;
use strict;

BEGIN {
  use Module::Common;
  Module::Common::setup_export_tags(
    constants => [ qw(DACL
		      SACL
		      OWNER
		      CONTROL) ],
    ops => [ qw(cifsSecurityDisplay
	        cifsSecurityBuildACE
	        cifsSecurityChange
		cifsSecuritySet
		cifsSecurityCheck
		cifsSecurityOwner
		cifsSecurityMatch
		cifsSecuritySID) ]
  );
}

use Params::Validate v0.70 qw(validate SCALAR SCALARREF HASHREF UNDEF ARRAYREF);
use Carp;
use base qw(Exporter);
use Data::Dumper;
use Win32::Security::ACL;
use Win32::Security::ACE;
use Win32::Security::NamedObject;
use Win32::Security::SID;
use CIFS::Misc qw(:all);

use constant DACL => 1;
use constant SACL => 2;
use constant OWNER => 4;
use constant CONTROL => 8;

sub aceTypeToString {
  my ($aceType, $aceFlags) = @_;
  
  if ($aceType =~ /ACCESS_ALLOWED/) {
    return "ALLOW";
  } elsif ($aceType =~ /ACCESS_DENIED/) {
    return "DENY";
  } elsif ($aceType =~ /SYSTEM_AUDIT/) {
    if ($aceFlags->{'SUCCESSFUL_ACCESS_ACE_FLAG'} && $aceFlags->{'FAILED_ACCESS_ACE_FLAG'}) {
      return "ALL";
    } elsif ($aceFlags->{'SUCCESSFUL_ACCESS_ACE_FLAG'}) {
      return "SUCCESS";
    } else {
      return "FAIL";
    }
  } else {
    return "?";
  }
}

sub aceAccessMaskToString {
  my $aceMask = shift @_;
  my $access;
  
  if ($aceMask->{FULL}) {
    return "FULL";
  }
  if ($aceMask->{MODIFY}) {
    return "MODIFY";
  }
  if ($aceMask->{READ}) {
    return "READ";
  }
  
  if ($aceMask->{FILE_READ_DATA}) {
    $access .= "FILE_READ_DATA|";
  }
  if ($aceMask->{FILE_LIST_DIRECTORY}) {
    $access .= "FILE_LIST_DIRECTORY|";
  }
  if ($aceMask->{FILE_WRITE_DATA}) {
    $access .= "FILE_WRITE_DATA|";
  }
  if ($aceMask->{FILE_ADD_FILE}) {
    $access .= "FILE_ADD_FILE|";
  }
  if ($aceMask->{FILE_APPEND_DATA}) {
    $access .= "FILE_APPEND_DATA|";
  }
  if ($aceMask->{FILE_ADD_SUBDIRECTORY}) {
    $access .= "FILE_ADD_SUBDIRECTORY|";
  }
  if ($aceMask->{FILE_CREATE_PIPE_INSTANCE}) {
    $access .= "FILE_CREATE_PIPE_INSTANCE|";
  }
  if ($aceMask->{FILE_READ_EA}) {
    $access .= "FILE_READ_EA|";
  }
  if ($aceMask->{FILE_WRITE_EA}) {
    $access .= "FILE_WRITE_EA|";
  }
  if ($aceMask->{FILE_EXECUTE}) {
    $access .= "FILE_EXECUTE|";
  }
  if ($aceMask->{FILE_TRAVERSE}) {
    $access .= "FILE_TRAVERSE|";
  }
  if ($aceMask->{FILE_DELETE_CHILD}) {
    $access .= "FILE_DELETE_CHILD|";
  }
  if ($aceMask->{FILE_READ_ATTRIBUTES}) {
    $access .= "FILE_READ_ATTRIBUTES|";
  }
  if ($aceMask->{FILE_WRITE_ATTRIBUTES}) {
    $access .= "FILE_WRITE_ATTRIBUTES|";
  }
  if ($aceMask->{FILE_ALL_ACCESS}) {
    $access .= "FILE_ALL_ACCESS|";
  }
  if ($aceMask->{FILE_GENERIC_READ}) {
    $access .= "FILE_GENERIC_READ|";
  }
  if ($aceMask->{FILE_GENERIC_WRITE}) {
    $access .= "FILE_GENERIC_WRITE|";
  }
  if ($aceMask->{FILE_GENERIC_EXECUTE}) {
    $access .= "FILE_GENERIC_EXECUTE|";
  }
  if ($aceMask->{DELETE}) {
    $access .= "DELETE|";
  }
  if ($aceMask->{READ_CONTROL}) {
    $access .= "READ_CONTROL|";
  }
  if ($aceMask->{WRITE_DAC}) {
    $access .= "WRITE_DAC|";
  }
  if ($aceMask->{WRITE_OWNER}) {
    $access .= "WRITE_OWNER|";
  }
  if ($aceMask->{GENERIC_READ}) {
    $access .= "GENERIC_READ|";
  }
  if ($aceMask->{GENERIC_WRITE}) {
    $access .= "GENERIC_WRITE|";
  }
  if ($aceMask->{GENERIC_EXECUTE}) {
    $access .= "GENERIC_EXECUTE|";
  }
  if ($aceMask->{GENERIC_ALL}) {
    $access .= "GENERIC_ALL|";
  }
  $access =~ s/\|$//;
  return $access;
}

sub aceFlagsToString {
  my $aceFlags = shift @_;
  my $flags;
  
  if ($aceFlags->{CONTAINER_INHERIT_ACE}) {
    $flags .= 'CI,';
  }
  if ($aceFlags->{OBJECT_INHERIT_ACE}) {
    $flags .= 'OI,';
  }
  if ($aceFlags->{INHERIT_ONLY_ACE}) {
    $flags .= 'IO,';
  }
  if ($aceFlags->{NO_PROPAGATE_INHERIT_ACE}) {
    $flags .= 'NP,';
  }
  $flags =~ s/,$//;
  if ($aceFlags->{INHERITED_ACE}) {
    $flags .= ':INHERITED';
  } else {
    $flags .= ':EXPLICIT';
  }
  return $flags;
}

sub cifsSecurityDisplay {
  my (%args) = validate (@_, {
    sPathname	=> { type => SCALAR },
    sAccounts	=> { type => SCALAR, optional => 1, default => undef },
    uType	=> { type => SCALAR, optional => 1, default => DACL|SACL }
  });

  if (! -e $args{sPathname}) {
    $! = 2; # The system cannot find the file specified
    return 0;
  }
  
  my %sec_hash;
  my $object = Win32::Security::NamedObject->new('FILE',
                                                 $args{sPathname});
  my @types = (DACL, SACL);
  # XXX: Why isn't this eval block catching the croak with a bad filename?
  #      Probably need to make a workaround.
  # XXX: Is there a better (and clean) way to print this access mask out?
  #      Most people (incl. me) can't do bitwise comparisons in their head. :-P
  # XXX: SACL support not available yet.
  eval {
    foreach my $type (@types) {
    if ($args{uType} & $type) {
      my $acl;
      my $sType;
      if ($type == DACL) {
        $acl = $object->dacl();
        $sType = 'DACL';
      } else {
        $acl = $object->sacl();
        $sType = 'SACL';
      }
      my @aces = $acl->aces();
      my $count = 0;
      print "$sType (" . @aces . " entries):\n";
      foreach my $ace (@aces) {
	my $skip = 0;
	my $name = Win32::Security::SID::ConvertSidToName($ace->sid());
	if ($args{sAccounts}) { # if sUsers were specified, we'll only print the ones on the list
	  $skip = 1;
	  my @users = split(/,/, $args{sAccounts});
	  foreach my $user (@users) {
	    if ($user =~ /^S(?:-\d+)+$/) {
	      $user = Win32::Security::SID::ConvertSidToName(Win32::Security::SID::ConvertStringSidToSid($user));
	    } else {
	      $user =~ s/\//\\/;
	    }
	    if (lc($user) eq lc($name)) {
	      $skip = 0;
	    }
	  }
	}
	if ($skip) { next; }
	printf("\t%s:%s:0x%08X:%s\n", $name, aceTypeToString($ace->aceType, $ace->explainAceFlags()), $ace->rawAccessMask(),
	  aceFlagsToString($ace->aceFlags()));
	$sec_hash{$count}{Name} = $name;
	$sec_hash{$count}{aceType} = aceTypeToString($ace->aceType, $ace->explainAceFlags());
	$sec_hash{$count}{aceMask} = $ace->rawAccessMask();
	$sec_hash{$count}{aceFlags} = aceFlagsToString($ace->aceFlags());
	($sec_hash{$count}{aceFlags}, $sec_hash{$count}{aceInheritance}) = (aceFlagsToString($ace->aceFlags()) =~ /^(.*):(.*)$/);
	$count++;
      }
    }
    }
  };
  if ($@) {
    return 0;
  }
  return %sec_hash;
}

sub cifsSecurityOwner {
  my (%args) = validate (@_, {
    sPathname	=> { type => SCALAR },
    sOwner	=> { type => SCALAR, optional => 1, default => undef }
  });
  
  if (! -e $args{sPathname}) {
    $! = 2; # The system cannot find the file specified
    return 0;
  }

  my $object = Win32::Security::NamedObject->new('FILE',
					         $args{sPathname});
  my $trustee;
  my $sid;
  
  if ($args{sOwner}) {
    if ($args{sOwner} =~ /^S(?:-\d+)+$/) {
      $sid = Win32::Security::SID::ConvertStringSidToSid($args{sOwner});
    } else {
      $sid = Win32::Security::SID::ConvertNameToSid($args{sOwner});
    }
    if (!$sid) {
      $^E = Win32::GetLastError();
      return 0;
    } else {
      eval {
        $object->ownerSid($sid);
        $trustee = $object->ownerTrustee();
      };
    }
  } else {
    eval { $trustee = $object->ownerTrustee(); };
  }
  
  if (!$@) { return $trustee; }
  return 0;
}

sub cifsSecuritySID {
    my (%args) = validate (@_, {
      sAccount	=> { type => SCALAR }
    });
    my $retVal;
   
    if ($args{sAccount} =~ /^S(?:-\d+)+$/) {
      # this is a string SID, return the account name
      $retVal = Win32::Security::SID::ConvertSidToName(Win32::Security::SID::ConvertStringSidToSid($args{sAccount}));
      if ($retVal eq $args{sAccount}) {
        return 0;
      }
      return $retVal;
    } else {
      # this is an account name, return the string SID
      $retVal = Win32::Security::SID::ConvertSidToStringSid(Win32::Security::SID::ConvertNameToSid($args{sAccount}));
      if ($retVal !~ /^S(?:-\d+)+$/) {
        return 0;
      }
      return $retVal;
    }
}

sub cifsSecurityCheck {
  my (%args) = validate (@_, {
    sPathname	=> { type => SCALAR },
    sAccount	=> { type => SCALAR },
    sAccessMask	=> { type => SCALAR, optional => 1, default => undef }
  });
  
  if (! -e $args{sPathname}) {
    $! = 2; # The system cannot find the file specified
    return 0;
  }
  
  # Go through each ACE for 'sPathname' and determine if 'sUser' has
  # (at least) 'sAccessMask' access rights.
  my $object = Win32::Security::NamedObject->new('FILE',
					         $args{sPathname});
  my $sid;
  my $dacl;
  
  if ($args{sAccount} =~ /^S(?:-\d+)+$/) {
    $sid = Win32::Security::SID::ConvertStringSidToSid($args{sAccount});
  } else {
    $sid = Win32::Security::SID::ConvertNameToSid($args{sAccount});
  }
  if ($sid) {
    $dacl = $object->dacl();
    my @aces = $dacl->aces();
    foreach my $ace (@aces) {
      if ($ace->sid() eq $sid) {
	my $uAccessMask = Win32::Security::ACE::SE_FILE_OBJECT->dbmAccessMask()->build_mask($args{sAccessMask});
	if ((($uAccessMask & $ace->rawAccessMask()) == $uAccessMask) || !$args{sAccessMask}) {
	  if ($ace->aceType() eq 'ACCESS_DENIED_ACE_TYPE') {
	    return 0;
	  } else {
	    return 1;
	  }
	}
      }
    }
  }
  return 0;
}

sub cifsSecurityBuildACE {
  my (%args) = validate (@_, {
    sUser	=> { type => SCALAR },
    sType	=> { type => SCALAR, optional => 1, default => 'ALLOW' },
    sAccessMask	=> { type => SCALAR, optional => 1, default => 'F' },
    sFlags	=> { type => SCALAR, optional => 1, default => '' }
  });
  my $ace;
  
  eval {
    my @aceParams = ($args{sFlags}, $args{sAccessMask}, $args{sUser});
    $ace = Win32::Security::ACE->new('FILE', $args{sType}, @aceParams);
  };
  if ($@) {
    return 0;
  }
  return $ace;
}

sub cifsSecuritySet {
  
  my (%args) = validate (@_, {
    sPathname	=> { type => SCALAR },
    uAceList	=> { type => ARRAYREF, optional => 1, default => '' },
    bInherit	=> { type => SCALAR, optional => 1, default => 1 }
  });

  if (! -e $args{sPathname}) {
    $! = 2; # The system cannot find the file specified
    return 0;
  }

  my @dacl;
  my @sacl;

  my $use_dacl = 0;
  my $use_sacl = 0;

  foreach my $ace ( @{ $args{uAceList} } ) {
    if ($ace->aceType() eq 'SYSTEM_AUDIT_ACE_TYPE') {
      $use_sacl = 1;
      push(@sacl, $ace);
    } else {
      $use_dacl = 1;
      push(@dacl, $ace);
    }
  }
  
  eval {
    my $dacl;
    my $sacl;

    if ($use_dacl) {
      $dacl = Win32::Security::ACL->new('FILE', @dacl);
    }
    if ($use_sacl) {
      $sacl = Win32::Security::ACL->new('FILE', @sacl);
    }
    my $object = Win32::Security::NamedObject->new('FILE', $args{sPathname});
    if ($use_dacl) {
      $object->dacl($dacl, $args{bInherit} ? 'UNPROTECTED_DACL_SECURITY_INFORMATION' : 'PROTECTED_DACL_SECURITY_INFORMATION');
    }
    if ($use_sacl) {
      $object->sacl($sacl, $args{bInherit} ? 'UNPROTECTED_SACL_SECURITY_INFORMATION' : 'PROTECTED_SACL_SECURITY_INFORMATION');
    }
  };
  if ($@) {
    return 0;
  }
  return 1;
  
}

sub cifsSecurityMatch {

  my (%args) = validate (@_, {
    sSource	=> { type => SCALAR },
    sTarget	=> { type => SCALAR },
    uType	=> { type => SCALAR, optional => 1, default => DACL|SACL|OWNER }
  });

  if (!(-e $args{sSource} && -e $args{sTarget})) {
    $! = 2; # The system cannot find the file specified
    return 0;
  }

  my $source_object = Win32::Security::NamedObject->new('FILE', $args{sSource});
  my $target_object = Win32::Security::NamedObject->new('FILE', $args{sTarget});

  my @acl_types = (DACL, SACL);
  
  foreach my $type (@acl_types) {
  if ($args{uType} & $type) {
    my @source_aces;
    my @target_aces;
    if ($type == DACL) {
      @source_aces = $source_object->dacl()->aces();
      @target_aces = $target_object->dacl()->aces();
    } else {
      @source_aces = $source_object->sacl()->aces();
      @target_aces = $target_object->sacl()->aces();
    }
    
    # The premise here is to test that, for every group of common ace types in
    # the acl, each ace is represented in both. The order makes no difference
    # in a particular group.
    my $source_group = 0;
    my $target_group = 0;
    my $last_source_ace_type = @source_aces[0]->rawAceType();
    my $last_target_ace_type = @target_aces[0]->rawAceType();
    my $match;

    my @merged_source_aces = _mergeAceList(uAceList => \@source_aces);
    my @merged_target_aces = _mergeAceList(uAceList => \@target_aces);

    foreach my $source_ace (@merged_source_aces) {
      $match = 0;
      $target_group = 0;
      if ($source_ace->rawAceType() != $last_source_ace_type) {
        $source_group++;
      }
      foreach my $target_ace (@merged_target_aces) {
        if ($target_group != $source_group) {
          if ($target_ace->rawAceType() != $last_target_ace_type) {
            $target_group++;
          }
        }
	#print $source_group . " | " . $target_group . "\n";
	#print "Trustee: " . $source_ace->trustee() . " == " . $target_ace->trustee() . "\n";
	#printf ("Type: 0x%x == 0x%x\n", $source_ace->rawAceType(), $target_ace->rawAceType());
	#printf ("Type: 0x%x == 0x%x\n", $source_ace->rawAccessMask(), $target_ace->rawAccessMask());
	#printf ("Flag: 0x%x == 0x%x\n", $source_ace->rawAceFlags(), $target_ace->rawAceFlags());
        if ($source_ace->trustee() eq $target_ace->trustee() &&
            $source_ace->rawAceType() == $target_ace->rawAceType() &&
            $source_ace->rawAccessMask() == $target_ace->rawAccessMask() &&
            $source_ace->rawAceFlags() == $target_ace->rawAceFlags()) {
            #print "Matched!\n\n";
            $match = 1;
            last;
        }
	$last_target_ace_type = $target_ace->rawAceType();
      }
      $last_source_ace_type = $source_ace->rawAceType();
      if (!$match) {
        return 0;
      }
    }
    
  }
  }
  
  if ($args{uType} & OWNER) {
    if ($source_object->ownerTrustee() ne $target_object->ownerTrustee()) {
      return 0;
    }
  }
  
  # This checks for auto-inheritance (SE_DACL_AUTO_INHERIT) as well, should we?
  # This check is currently off by default.
  if ($args{uType} & CONTROL) {
    my $source_control = $source_object->control();
    my $target_control = $target_object->control();
    
    foreach my $source_key (keys %$source_control) {
      if ($target_control->{$source_key} != 1) {
	      return 0;
      }
    }
  }
  
  return 1;
}

sub _mergeAceList {
  my (%args) = validate (@_, {
    uAceList    => { type => ARRAYREF }
  });

  my @base_aces = @{ $args{uAceList} };
  my @merged_aces;
  my $found;

  foreach my $base_ace (@base_aces) {
    $found = 0;
    foreach my $merged_ace (@merged_aces) {
      if ($base_ace->trustee eq $merged_ace->trustee &&
               $base_ace->rawAceType() == $merged_ace->rawAceType()) {
        if ($base_ace->rawAccessMask() == $merged_ace->rawAccessMask()) {
	  $merged_ace->rawAceFlags($merged_ace->rawAceFlags() | $base_ace->rawAceFlags());
	} elsif ($base_ace->rawAceFlags() == $merged_ace->rawAceFlags()) {
          $merged_ace->rawAccessMask($merged_ace->rawAccessMask() | $base_ace->rawAccessMask());
	}
	$found = 1;
      }
    }
    if (!$found) {
      push @merged_aces, $base_ace;
    }
  }

  return @merged_aces;
}

sub cifsSecurityChange {
  my (%args) = validate (@_, {
    sPathname	=> { type => SCALAR },
    uAceList	=> { type => ARRAYREF }
  });
  
  if (!(-e $args{sPathname})) {
    $! = 2; # The system cannot find the file specified
    return 0;
  }
  
  my $object = Win32::Security::NamedObject->new('FILE', $args{sPathname});
  
  eval {
    my $dacl = $object->dacl();
    my $sacl = $object->sacl();
    my @dacl_aces = $dacl->aces();
    my @sacl_aces = $sacl->aces();
    my @change_aces = @{ $args{uAceList} };
    my $index = 0;

    my @dacl_changes;
    my @sacl_changes;

    foreach my $ace ( @{ $args{uAceList} } ) {
      if ($ace->aceType() eq 'SYSTEM_AUDIT_ACE_TYPE') {
        push(@sacl_changes, $ace);
      } else {
        push(@dacl_changes, $ace);
      }
    }
    
    foreach my $dacl_change (@dacl_changes) {
      foreach my $dacl_ace (@dacl_aces) {
        $index = 0;
        if (!$dacl_ace->aceFlags()->{INHERITED_ACE}) {
          if ($dacl_ace->sid() eq $dacl_change->sid()) {
            splice(@dacl_aces, $index, 1);
          }
        }
        $index++;
      }
    }
    
    foreach my $sacl_change (@sacl_changes) {
      foreach my $sacl_ace (@sacl_aces) {
        $index = 0;
        if (!$sacl_ace->aceFlags()->{INHERITED_ACE}) {
          if ($sacl_ace->sid() eq $sacl_change->sid()) {
            splice(@sacl_aces, $index, 1);
          }
        }
        $index++;
      }
    }

    my $new_dacl = Win32::Security::ACL->new('FILE', @dacl_aces);
    my $new_sacl = Win32::Security::ACL->new('FILE', @sacl_aces);
    $new_dacl->addAces(@dacl_changes);
    $new_sacl->addAces(@sacl_changes);
    $object->dacl($new_dacl);
    $object->sacl($new_sacl);
  };
  if ($@) {
    return 0;
  }
  
  return 1;
}

1;
