# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CIFS/Setup.pm#1 $

## @summary CIFS Setup routines
## @author dl-test-infra@netapp.com
## @status Shared
## @pod here


=head1 NAME

CIFS::Setup - Filer setup routines for CIFS.

=head1 SYNOPSIS

    use CIFS::Setup qw(is_cifs_setup cifs_setup);

    my $host = host($FILER);

    my $ret = is_cifs_setup(host=>$host);

    eval {
        cifs_setup(host=>.....);
    };
    logresult('fatal', $@) if ($@);



=head1 DESCRIPTION

    This library contains CIFS setup routines.

    Use test/cifs/setup.thpl to test this lib

=head1 SEE-ALSO

    Network::Nis::nis_configure

=head1 ROUTINES

=cut

package CIFS::Setup;

use strict;
use warnings;

BEGIN
{
    # doing the export in the BEGIN block to avoid Exporter problems
    # during run-time due to circular use statements
    use base qw(Exporter);
    our @EXPORT_OK = qw(is_cifs_setup cifs_setup);

    Params::Validate::validation_options(
        on_fail => sub {my @msgs = @_; chomp(@msgs);
                confess(script_result(@msgs))}
        );
}

use Params::Validate v0.70 qw(validate SCALAR OBJECT HASHREF);
use Carp qw(croak confess);
use Tharn qw(say logcomment logresult disconnect script_result matchstrings $Log);
use Utils qw(make_host make_conn);
use Admin::License qw(license);
use TracePkg;



=head2 is_cifs_setup

    is_cifs_setup( host => $host )

Check if CIFS is already setup on filer

The routine returns 1 if CIFS is setup already and 0 otherwise.

It accepts the following named arguments

=over

=item host

(Required) : hostspec or host object of filer

=back

=for html <hr>

=cut

sub is_cifs_setup {

    my (%args) = validate (@_, {
                         host => { type=>SCALAR|HASHREF },
                        } );

    my $conn = make_conn($args{host});
    die($@) if ($@); # remove when the return eval is removed

    my (undef, undef, $out) = say($conn, 'cifs sessions');
    return(($out =~ /(not licensed|not running)/s) ? 0 : 1);

}

=head2 cifs_setup

    cifs_setup( %opts )

A routine to setup CIFS on the filer.

This routine is by no means *complete*. Has to be updated to support
user auth style 2 (NT domain user authentication)

To turn on debugs for this function add the following to you ntest environment
or command line:

NTAP_QA_TRACE='CIFS::Setup=all=enabled'

This test uses, then disconnects, the "#telnet" connection. (non-interactive
"cifs setup" exists, but it is way too hidden and limited and not expected to
be used by customers, so it cannot be trusted)

Some of the things to be done for this routine

add support for nt4

migrate to using the zapi style. Basically

    * change domain to domain-name
    * change domain_user to login-user
    * change domain_pass to login-password
    * change root_pass to default-root-password
    * implement pdc-ip-address
    * implement server-name
    * in ZAPI workgroup name is not specified - how do they do that ???
      - The domain-name field serves as the workgroup name. ~~ Garrett

    * Actually, you should just use the ZAPIs themselves, eh? ~~ Garrett

The routine takes the following named arguments

=over

=item host

(Required) : The filer DNS name, Hostrec or Connectrec.

=item auth-type

(Optional) : The authentication style that determines the
method by which clients will be authenticated when connecting
to this CIFS server.
Possible values: 'ad'(Active Directory), 'workgroup'(Workgroup),
'passwd'(/etc/passwd).  Default: 'workgroup'.
This routine does not support nt4 as yet. When it does 'nt4' will be
a possible value.

Note: The domain, domain_user and domain_pass need to be specified
if auth-type is ad.

=item workgroup

(Optional) : The workgroup of which the filer would be part of.
Defaults to "WORKGROUP" if not specified and auth-type is 'workgroup'.

=item domain

(Optional) : The domain name of which the filer would be a part of.
If the domain name is specified then the domain_user and domain_pass
must also be specified.

=item domain_user

(Optional) : The username on the domain with administrative
privileges to add a filer to it.

=item domain_pass

(Optional) : The password of the "domain_user".

=item root_pass

(Optional) : A scalar specifying the root password of the filer.
Defaults to the hostspec "default_password".

=item cifs_pass

(Required) : The password for the cifs administrative user.
Deafults cifs password used is cifspass123

=item dns_enable

(Optional) : If defined will configure DNS when prompted by cifs setup.
If defined must provided 'dns_domainname' and 'dns_server_list' also.

=item dns_domainname

(Optional) : Filer's domain name. Must be defined if 'dns_enable' is defined.

=item dns_server_list

(Optional) : Comma separted list of DNS server IP(s).
Must be defined if dns_enable is defined.

=item timed_enable

(Optional) : If defined will configure time services when prompted by cifs setup.
If defined must provided 'timed_server_list' also.

=item timed_server_list

(Optional) : Comma separted list of time server names or IP(s).
Must be defined if timed_enable is defined.

=item passwd_group

(Optional) : If set then create the default /etc/passwd and/or /etc/group files.
Default is 'no'.

=item vfiler

(Optional) : If variable vfiler is defined, then cifs  setup is done on the vfiler

=item server-name

(Optional) : CIFS Server name.

=item security-style

(Optional) : The security style determines wheter or not the CIFS service will
support multiprotocol access. These styles are subject to change in a future
release. Possible values : 'ntfs','multiprotocol'. Default: 'multiprotocol'

=item create_admin

(Optional) : If set to 1 will create the CIFS administrator account if prompted.
Possible values : 'yes' / 'no'.  Default : 'yes'

=item ou-name

(Optional) : The distinguished name of the organizational unit that the CIFS
service will become a member of. If cifs.new_setup.enable = on, the filer
will join the 'CN=Computers' OU if ou-name is not found in the list of
available OUs. Note: This information is only used when the auth-type is 'ad'.

=item select_site_name

(Optional) : If set to 'yes' will enable selection of the AD site name if prompted.
Possible values : 'yes' / 'no'. Default : 'no'.

=item site-name

(Optional) : The name of the site to choose when select_site_name is yes. If
select_site_name = yes and site-name is not defined the Default-First-Site-Name
will be selected.

=item wins_option

(Optional) : Not supported when cifs.new_setup.enable = on. If set to 1 WINS
will be enabled using the existing WINS servers if they exist otherwise the
wins servers will be set to wins_server_list. If set to 2, WINS will be
enabled and any existing WINS servers will be overwritten with
wins_server_list. If set to 3 WINS will be disabled.
Possible values : 1,2,3. Default : 3.";

=item wins_servers

(Optional) : Not supported when cifs.new_setup.enable = on. A comma seperated
list of WINS servers.

=item local_admins

(Optional) : User or group ID to add to the BUILTIN\adminisrators group

=item timeout

(Optional) : The timeout for waiting for prompts. Default: 60.

=back

=cut


sub cifs_setup {

    my $Trace = new TracePkg;
    # while adding args to cifs_setup make sure to add them to cifs_setup_old
    # or remove the args from the list before calling cifs_setup_old
    my (%args) = validate (@_, {
                    timeout => { type=>SCALAR, default=>60 },
                    host => { type=>SCALAR|HASHREF },
                    local_admins => { type=>SCALAR, optional=>1 },
                    wins_option => { type=>SCALAR, optional=>1, default=>3 },
                    wins_server_list => { type=>SCALAR,
		                          optional=>1,
                                          regex=>qr/(?:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/
					  },
                    vfiler=> { type => SCALAR, optional =>1 },
                    'auth-type' => { type=>SCALAR, optional=>1,  # used ZAPI style
                                    regex=>qr/^(ad|workgroup|passwd)$/,
                                    #regex=>qr/^(ad|nt4|workgroup|passwd)$/,
                                    default=>'workgroup' },
                    'server-name' => { type=>SCALAR, optional=>1,
                                    regex=>qr/\w{1,15}/ },
                    workgroup => { type=>SCALAR, optional=>1 ,default=>'workgroup'},
                    domain => { type=>SCALAR, optional=>1, regex=>qr/\w+/,
                                depends=>['domain_user','domain_pass'] },
                    domain_user => { type=>SCALAR, regex=>qr/\w+/, optional=>1 },
                    domain_pass => { type=>SCALAR, optional=>1 },
		    'ou-name' => { type=>SCALAR, optional=>1, default=>'CN=computers' },
                    root_pass => { type=>SCALAR, optional=>1 },
                    cifs_pass => { type=>SCALAR, default=>'cifspass123' },
                    'security-style' => { type=>SCALAR, optional=>1, # used ZAPI style
                                default=>'multiprotocol',
                                regex=>qr/^ntfs|multiprotocol$/ },
                    create_admin => { type=>SCALAR, optional=>1,
                                regex=>qr/^(yes|no)$/, default=>'yes' },
                    dns_enable => { type=>SCALAR, optional=>1,
                                depends=>['dns_server_list','dns_domainname'] },
                    dns_domainname => { type=>SCALAR, optional=>1 },
                    dns_server_list => { type=>SCALAR, optional=>1, # comma separated list of DNS server IP's
                                regex=>qr/(?:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/ },
                    timed_enable => { type=>SCALAR, optional=>1,
                                depends=>['timed_server_list'] },
                    timed_server_list => { type=>SCALAR, optional=>1 }, # comma separated list of time server name or IP's
                    passwd_group => {type=>SCALAR, regex =>qr/^(yes|no)$/, default => 'yes' },
                    select_site_name => {type=>SCALAR, regex =>qr/^(yes|no)$/, default => 'yes' },
                    'site-name' => {type=>SCALAR, optional=>1, default => ''},
                    });

    # Set the user authentication selction based on auth-type
    my $auth_type = _compute_auth_type(%args);

    # check whether to execute the old or new style cifs routine
    my $rsh_conn = make_conn($args{host});
    die($@) if ($@); # remove when the return eval is removed
    logcomment("Remove administrative user");
    #my $ret = removeCIFSUser($rsh_conn, "administrator")
    #    or logError($@);
    _say_vfiler($args{vfiler},$rsh_conn,"useradmin user delete administrator");
    my (undef, undef, $out) = _say_vfiler($args{vfiler},$rsh_conn,"options cifs.new_setup.enable");
    if (($out =~ /No such option cifs.new_setup.enable/) ||
        ($out =~ /cifs.new_setup.enable\s+off/)) {
        # execute old style of cifs setup
        logcomment("Using OLD style of cifs setup");
        my %old_args = %args;
        # delete the args that are not supported by cifs_setup_old
        # if needed these args need to be coded in cifs_setup_old and tested
        delete $old_args{'server-name'};
        delete $old_args{dns_enable};
        delete $old_args{dns_domainname};
        delete $old_args{dns_server_list};
        delete $old_args{timed_enable};
        delete $old_args{timed_server_list};
        delete $old_args{'security-style'};
        delete $old_args{passwd_group};
        delete $old_args{select_site_name};
        delete $old_args{'site-name'};
        delete $old_args{wins_option};
        delete $old_args{wins_server_list};
        delete $old_args{local_admins};

        my $rv;
        eval {
            $rv = cifs_setup_old(%old_args);
        };
        die $@ if ($@);
        return $rv;
    }
    logcomment("Using NEW style of cifs setup");
    #Checking the Filer security option field   values
    my (undef, undef, $out_store) = _say_vfiler($args{vfiler},$rsh_conn,"options security.passwd.rules.history");

    $out_store =~ /history\s+(\d+)/;
    $out_store = "$1";
    #Temporarily Resetting security.passwd.rules.history to 0
    _say_vfiler($args{vfiler},$rsh_conn, "options security.passwd.rules.history 0");

    # Create the telnet connection for the filer and fill up the
    # root_pass in our args if we don't know it yet.
    my $h = make_host($args{host});
    my $telnet_conn = make_conn($args{host},"#telnet");
    die($@) if ($@); # remove when the return eval is removed
    if (defined($main::CMDTIMEOUT)) {
        $telnet_conn->timeout($main::CMDTIMEOUT);
    }

    $args{"root_pass"} =  $telnet_conn->hostp()->default_password();

    # terminate and license CIFS
    _say_vfiler($args{vfiler},$telnet_conn,"cifs terminate -t 0");
    # XXX what errors to look out for ???

    license($telnet_conn, "cifs") or die $@;

    # Create a replytable.
    my %replytable = (
        "Do you want to make the system visible via WINS\\? .*"          => $args{wins_option} =~ /[1-2]/ ? "yes" : "no",
        "\\(1\\)(?:.|\n)Keep the current WINS(?:.|\n)*Selection \\(1-3\\)\?" => $args{wins_option},
        "IPv4 address\\(es\\) of your WINS name server\\(s\\).*"         => $args{wins_server_list},
        "Would you like to specify additional WINS name servers"         => "no",
        "Enter the password for the root user"                           => $args{root_pass},
        "Enter the password for root"                                    => $args{root_pass},
        "Retype the password:"                                           => "",
        "(?s:The default name of this filer will be .*".
        "Do you want to modify this name\\?)"                            => "no",
        "(?s:This filer will be a member of the Windows Workgroup .*".
        "Do you want to modify this name\\?)"                            => exists($args{workgroup}) ? "yes" : "no",
        "Workgroup name for the filer\\?.*:"                             => $args{workgroup},
        "Do you want to modify this list\\?"                             => "no",
        "Do you want to delete the existing filer account information\\?" => "yes",
        "create the .*\\\\administrator account\\?"                      => $args{create_admin},
        "Please enter the Windows.*user.*:"                              => $args{domain_user},
        "Enter the name of the Windows user"                             => $args{domain_user},
        "Password for.*:"                                                => $args{domain_pass},
        "Do you wish to continue and join.*in NT4 mode\\?"               => "",
        "Active Directory container for filer account\\?"                => $args{'ou-name'},
        "Do you want to continue and change the current filer account information\\?" => 'yes',
        "Would you like to reconfigure this filer to be an NTFS-only filer\\?" => (($args{'security-style'} eq 'ntfs') ? 'yes' : 'no'),
        "Do you want to configure this filer as a NTFS-only filer"       => (($args{'security-style'} eq 'ntfs') ? 'yes' : 'no'),
        "Data ONTAP CIFS services support four styles of user authentication(?:.|\n)*Selection \\(1-4\\)\?" => $auth_type,
        "\\(1\\) Multiprotocol filer[\r|\n]*\\(2\\) NTFS-only filer[\r|\n]*Selection \\(1-2\\)\?"
                                                                         => (($args{'security-style'} eq 'multiprotocol') ? 1 : 2),
        "\\(1\\) NTFS-only filer[\r|\n]*\\(2\\) Multiprotocol filer[\r|\n]*Selection \\(1-2\\)\?"
                                                                         => (($args{'security-style'} eq 'multiprotocol') ? 2 : 1),
        "\\(1\\) Ignore Active Directory Site information when choosing domain controllers[\r|\n]*\\(2\\) Choose a site name from the list of available sites[\r|\n]*Selection \\(1-2\\)\?"
                                                                         => (($args{select_site_name} eq 'no') ? 1 : 2),
        # site-name
        "Choose a Site name from the following list of Sites(?:.|\n)*Selection \\(1-\\d+\\)\?"
	                                                                 => '',  # default : Default-first-site-name
        # ou-name
        "The user that you specified has permission to create(?:.|\n)*\\(1\\) CN=computers(?:.|\n)*Selection \\(1-\\d+\\)"            => $args{'ou-name'},
        "\\(3\\) Choose from a subset of containers by specifying a search filter(?:.|\n)*Selection \\(1-3\\)"
	                                                                 => exists($args{'ou-name'}) ? 3 : 1,
        "Please enter some or all of a container specification to limit the(?:.|\n)*Enter the search string"
	                                                                 => $args{'ou-name'},
        "\\(1\\) $args{'ou-name'}(?:.|\n)*Selection \\(1-2\\)"          => 1,
        "Would you like to change this name\\?"                          => exists($args{'server-name'}) ? 'yes' : 'no',
	"Would you like to specify a user or group that can administer CIFS" =>  exists($args{'local_admins'}) ? 'yes' : 'no',
	"Enter the name of a user or group that will administer CIFS on the filer" => $args{'local_admins'},
        "Enter the CIFS server name for the filer"                       => $args{'server-name'},
        "Active Directory Domain name\\?"                                => $args{domain},
        "What is the name of the Active Directory domain"                => $args{domain},
        "Do you want to re-use this AD account\\?"                       => "", # default : yes
        "Do you want to re-use this machine account"                     => "", # default : yes
        "Enter the new password for .*\\\\administrator"                 => $args{cifs_pass},
        "Would you like to specify a user or group that can administer CIFS\\?" => 'no',
        "What is the name of the Workgroup\\?"                           => "",
        "Should CIFS create a default /etc/passwd file"                  => $args{passwd_group},
        "Should CIFS create a default /etc/group file"                   => $args{passwd_group},
        "Should CIFS create default /etc/passwd and /etc/group files"    => $args{passwd_group},
        "Would you like to enable NIS group caching"                     => "", # default : yes
        "Enter the hour\\(s\\) when NIS should update the group cache"   => "", # default : 24hrs
        "Would you like to specify additional hours"                     => "", # default : no
        "Do you want to configure the filer\\'s DNS resolver service"     => exists($args{dns_enable}) ? 'yes' : 'no',
        "What is the filer\\'s DNS domain name"                           => $args{dns_domainname},
        "What are the IPv4 address\\(es\\) of your authoritative DNS name server\\(s\\)" => $args{dns_server_list},
        "What are the IPv4/IPv6 address\\(es\\) of your authoritative DNS name server\\(s\\)" => $args{dns_server_list},
        "Would you like to specify additional DNS name servers"          => "no",
        "Would you like to configure time services\\?"                   => exists($args{timed_enable}) ? 'yes' : 'no',
        "Enter the time server host\\(s\\) and\\/or address\\(es\\)"     => $args{timed_server_list},
        "Would you like to specify additional time servers\\?"           => 'no',
        "Would you like to reconfigure this filer to be a multiprotocol filer\\?" => (($args{'security-style'} eq 'multiprotocol') ? 'yes' : 'no'),
        "Do you want to create the filer's machine account in"           => "", #default : yes
    );

    # wait for "cifs setup" prompts in %replytable, and act on
    # them, until the prompt coming back is one of the built-in
    # say prompts (i.e. the filer's normal prompt, showing that
    # cifs setup is done)

    _say_vfiler($args{vfiler},$telnet_conn,"cifs setup",'-blind');
    my ($tries, $prompt, $reply, $last_prompt, $after);
    for ($tries=100;$tries;$tries--) {
        eval {
            (undef, undef, $after, $prompt) =
              say($telnet_conn, "", '-timeout', $args{timeout}, '-prompt', [keys %replytable]);
        };
        my $error = $@;
        if ($error) {
            $telnet_conn->send_intr();
            die $error;
        } elsif ($after =~ /vfiler cannot be run/) {
            sleep 10;
            _say_vfiler($args{vfiler}, $telnet_conn, "cifs setup", '-blind');
            next;
        }

        $Trace->debug(msg=>"Prompt: $prompt");
        last if (matchstrings($prompt,[$telnet_conn->prompts()]));
	# if we don't have a match, just # take the default
        if (!matchstrings($prompt, [keys %replytable])) {
            eval { say($telnet_conn, "\r", '-nocr', '-blind'); };
            my $error = $@;
            if ($error) {
                $telnet_conn->send_intr();
                die $error;
            }
        }
        foreach my $regex (keys %replytable) {
            next unless(matchstrings($prompt,$regex));
                $Trace->debug(msg=>"Regex: $regex");
                if ($prompt =~ /Retype the password/) {
                    if ($last_prompt =~ /Enter the password for the root user/) {
                        $reply = $args{root_pass};
                    } else {
                        $reply = $args{cifs_pass};
                    }
		# Get the number for the desired container from the container
		# selection list
                } elsif ($prompt =~ /The user that you specified has permission to create(?:.|\n)*\(1\) CN=computers(?:.|\n)*Selection \(1-\d+\)/i) {
		    my $msg = "\nSearching for OU $args{'ou-name'}\n";
		    my $ou_found = 0;
                    foreach my $line (split(/\r\n/, $prompt)) {
                        chomp($line);
                        if ($line =~ /\((\d+)\)\s+$args{'ou-name'}/) {
		            $msg .= "Found OU $args{'ou-name'} as selection #$1";
			    $ou_found = 1;
                            $reply = $1;
			}
                    }
                    $Trace->debug(msg=>"$msg");
		    if (!$ou_found) {
                        logresult('warn', "$args{'ou-name'} not found defaulting to CN=computers");
		    }
		# Get the number for the desired site from the site
		# selection list
                } elsif ($prompt =~ /Choose a Site name from the following list of Sites(?:.|\n)*Default-First-Site-Name/i && $args{'site-name'}) {
		    my $msg = "\nSearching for site $args{'site-name'}\n";
		    my $site_found = 0;
                    foreach my $line (split(/\r\n/, $prompt)) {
                        chomp($line);
                        if ($line =~ /\((\d+)\)\s+$args{'site-name'}/) {
		            $msg .= "Found site $args{'ou-name'} as selection #$1";
			    $site_found = 1;
                            $reply = $1;
			}
                    }
                    $Trace->debug(msg=>"$msg");
		    if (!$site_found) {
                        logresult('warn', "$args{'site-name'} not found defaulting to Default-First-Site-Name");
		    }
                } else {
                    $reply = $replytable{$regex};
                }
                $reply .= "\r";
                $Trace->debug(msg=>"Reply: $reply");
                eval { say($telnet_conn, $reply, '-nocr', '-blind'); };
                my $error = $@;
                if ($error) {
                    $telnet_conn->send_intr();
                    die $error;
                }
        }
        $last_prompt = $prompt; # This is the last prompt we saw.
    }

    # If we are out of tries, something is wrong.
    if (!$tries) {
        $telnet_conn->send_intr();
        die "setup::cifs(): too many prompts during \"cifs setup\"\n";
    }

    # Check whether the cifs setup succeeded.
    my (undef, undef, $output) = _say_vfiler($args{vfiler},$telnet_conn, "cifs");
    $Trace->debug(msg=>"cifs command output: $output");

    #Reverting back the security.passwd.rules.history value to actual value
    _say_vfiler($args{vfiler},$telnet_conn, "options security.passwd.rules.history $out_store");
    disconnect($telnet_conn);
    unless($output =~ /The following commands are available/) {
        croak "setup::cifs():cifs setup unsuccessful : \"$output\"\n";
    }
    return 1;

}

# old style of setting up cifs
sub cifs_setup_old() {
    my (%args) = validate (@_, {
                    host => { type=>SCALAR|HASHREF },
                    'auth-type' => { type=>SCALAR, optional=>1,  # used ZAPI style
                                    regex=>qr/^(ad|workgroup|passwd)$/,
                                    #regex=>qr/^(ad|nt4|workgroup|passwd)$/,
                                    default=>'workgroup' },
                    workgroup => { type=>SCALAR, optional=>1 },
                    domain => { type=>SCALAR, optional=>1, regex=>qr/\w+/,
                                depends=>['domain_user','domain_pass']},
                    domain_user => { type=>SCALAR, regex=>qr/\w+/, optional=>1 },
                    domain_pass => { type=>SCALAR, optional=>1 },
                    'ou-name' => { type=>SCALAR, optional=>1 },
                    root_pass => { type=>SCALAR, optional=>1 },
                    cifs_pass => { type=>SCALAR,default=>'cifspass123' },
                    create_admin => { type=>SCALAR, optional=>1, default=>'yes' },
                    vfiler=> { type => SCALAR, optional =>1 }
                });

    my $cmd_prefix = ($args{vfiler}) ? "vfiler run $args{vfiler}" : "";


    # root_pass in our args if we don't know it yet.
    # Create rsh connection for the filer
    my $rsh_conn = make_conn($args{host});
    die($@) if ($@); # remove when the return eval is removed
    logcomment("Remove administrative user");
    say($rsh_conn,"$cmd_prefix useradmin user delete administrator");

    #Checking the Filer security option field security.passwd.rules.history  values and storing it
    my (undef, undef, $out_store) = say($rsh_conn, "$cmd_prefix options security.passwd.rules.history");
    $out_store =~ /(\d+)/;
    $out_store = "$1";

    # Disabling the NIS options for old setup to work
    say($rsh_conn, "$cmd_prefix options nis.enable off");

    # Disabling the LDAP options for old setup to work
    say($rsh_conn, "$cmd_prefix options ldap.enable off");

    #Temporarily Resetting security.passwd.rules.history to 0
    say($rsh_conn, "$cmd_prefix options security.passwd.rules.history 0");

    # Create the telnet connection for the filer and fill up the
    my $h = make_host($args{host});
    my $c = make_conn($args{host},"#telnet");
    die($@) if ($@); # remove when the return eval is removed

    $args{"root_pass"} =  $c->hostp()->default_password();

    # Set the user authentication selction based on auth-type
    my $auth_type = _compute_auth_type(%args);

    # terminate and license CIFS
    say($c,"$cmd_prefix cifs terminate -t 0");
    license($c, "cifs") or die $@;

    # Create a replytable.
    my %replytable = (
        "Do you want to make the system visible via WINS\\?"             => "no",
        "Do you wish to make the system visible via WINS\\?"             => "no",
        "Do you want to configure this filer as a NTFS-only filer\\?"    => "no",
        "Enter the password for root"                                    => $args{root_pass},
        "Set the password for .*\\\\administrator"                       => $args{cifs_pass},
        "Retype the password:"                                           => "",
        "(?s:The default name of this filer will be .*".
        "Do you want to modify this name\\?)"                            => "no",
        "(?s:This filer will be a member of the Windows Workgroup .*".
        "Do you want to modify this name\\?)"                            => exists($args{workgroup}) ? "yes" : "no",
        "(?s:This filer will be a member of the Workgroup .*".
        "Do you want to modify this name\\?)"                            => exists($args{workgroup}) ? "yes" : "no",
        "Enter the Workgroup for the filer.*:"                           => $args{workgroup},
        "Do you want to modify this list\\?"                             => "no",
        "What type of authentication will this filer use\\?"             => $auth_type,
        "Do you want to delete the existing filer account information\\?" => "yes",
        "Create the .*\\\\administrator account\\?"                      => $args{create_admin},
        "Enter the Windows Domain for the filer.*:"                      => $args{domain},
        "Please enter the Windows.*user.*:"                              => $args{domain_user},
        "Password for.*:"                                                => $args{domain_pass},
        "Do you wish to continue and join.*in NT4 mode\\?"               => "",
        "Active Directory container for filer account\\?"                => $args{'ou-name'},
        "Would you like to reconfigure this filer to be a multiprotocol filer\\?"   => 'no',
        );

    # wait for "cifs setup" prompts in %replytable, and act on
    # them, until the prompt coming back is one of the built-in
    # say prompts (i.e. the filer's normal prompt, showing that
    # cifs setup is done)
    say($c,"$cmd_prefix cifs setup",'-blind');
    my ($tries, $prompt, $reply, $last_prompt, $after);
    for ($tries=100;$tries;$tries--) {
        (undef, undef, $after, $prompt) =
            say($c,"",'-prompt',[keys %replytable]);
        last if (matchstrings($prompt,[$c->prompts()]));
        if (!matchstrings($prompt,[keys %replytable])) {
            die "Unrecognized prompt \"$prompt\" encountered\n";
        }
        foreach my $regex (keys %replytable) {
            next unless(matchstrings($prompt,$regex));
            if ($prompt =~ /Retype the password/) {
                if ($last_prompt =~ /Enter the password for root/) {
                    $reply = $args{"root_pass"};
                } else {
                    $reply = $args{"cifs_pass"};
                }
            } else {
                $reply = $replytable{$regex};
            }
            $reply .= "\r";
            say($c,$reply,'-nocr','-blind');
        }
        $last_prompt = $prompt; # This is the last prompt we saw.
    }

    # If we are out of tries, something is wrong.
    if (!$tries) {
        die "setup::cifs(): too many prompts during \"cifs setup\"\n";
    }

    # Check whether the cifs setup succeeded.
    my (undef, undef, $output) = say($c, "$cmd_prefix cifs");
    unless($output =~ /The following commands are available/) {
        die "setup::cifs():cifs setup unsuccessful : \"$output\"\n";
    }

    #Reverting back the security.passwd.rules.history value to actual value
    logcomment("the resetting of original values of security");
    say($rsh_conn, "$cmd_prefix options security.passwd.rules.history $out_store");

    disconnect($c);
    return 1;
}


sub _compute_auth_type(%) {

    my (%args) = @_;
    my $flag = 0;
    my $auth_type;

    my $cmd_prefix = ($args{vfiler}) ? "vfiler run $args{vfiler}" : "";
    my $fh = make_conn($args{host});
    my (undef, undef, $out_compute) = say($fh, "$cmd_prefix options cifs.new_setup.enable");
    if (($out_compute =~ /No such option cifs.new_setup.enable/) ||
        ($out_compute =~ /cifs.new_setup.enable\s+off/)) {
        $flag=1;
    }
    if ($args{'auth-type'} eq 'passwd') {
        # XXX - dont know of anything that depends on this auth style
        $auth_type = ($flag) ? 3 : 4;

    } elsif (($args{'auth-type'} eq 'workgroup') || (!$args{'workgroup'})) {
        # check if workgroup is specified - else default it to WORKGROUP
        #if (!$args{'workgroup'}) {
            logcomment("Setting workgroup to 'WORKGROUP'");
            # XXX - get rid of defaulting to "WORKGROUP"
            $args{workgroup} = 'WORKGROUP';
       # }
         $auth_type = ($flag) ?  2 : 3;
    } elsif ($args{'auth-type'} eq 'nt4') {
        # XXX - this is not supported as yet
        logresult('warn', "This user auth style is not supported by this routine");
         $auth_type = ($flag) ? "" : 2;
    } else {
        # check if domain name etc are also specified
        croak("'ad' user auth style has been chosen, and ".
                "domain, domain_user, domain_pass also need to be specified")
        if (!$args{domain} || !$args{domain_user} || !$args{domain_pass});
         $auth_type = 1 ;
    }

    return($auth_type);
}

sub _say_vfiler {
    my ($vfiler, $conn, $sendstr, @opts) = @_;
    if ($vfiler) {
        $sendstr = "vfiler run $vfiler $sendstr";
    }
  AGAIN: {
        my @say_results;
        eval { (@say_results) = say($conn, $sendstr, @opts); };
        my $error = $@;
        if ($error) {
            $conn->send_intr();
            die $error;
        } elsif (defined($say_results[2])
                 && $say_results[2] =~ /vfiler cannot be run/) {
            redo AGAIN;
        }
        return wantarray ? @say_results : $say_results[0];
    }
}

1;

=head1 BUGS / ENHANCEMENTS

    Update the routines to use ZAPI.
    Implement all the ZAPI cifs API concerned with cifs setup.

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = options cifs, vfiler run xxx options cifs, cifs, vfiler run xxx cifs

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    dl-test-infra@netapp.com
    $Id: //depot/prod/test/Rsierranevada/lib/CIFS/Setup.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

