# this is basically -*- perl -*-
# $Id: /depot/prod/test/main/lib/CIFS/Setup_GX.pm
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
## @author olgav@netapp.com
## @status	design
# 
=head1 NAME

Setup_GX - Library for the 10mode/GX VSERVER and CIFS setup

=head1 SYNOPSIS

  use CIFS::Setup_GX;

  vserver_setup(nodes => 'hifive', vserver => 'hifive_vs',
                interface_type => 'DATA');

  cifs_setup( nodes => 'hifive');

  cifs_setup( nodes => 'hifive',
              cifs_domain_fullname => 'cifsqa.lab.netapp.com',
              cifs_domain_shortname => 'cifsqa',
              cifs_server => 'hifive_cifsvs',
              vserver => 'hifive_vs');

  create_volume(node => $filer, 
                filer_vol => $volumename, 
                volsecurity_style => $sec_style,
                unix_permissions => $unix_perm,
                vserver     => $vserver,
                'aggr_name' => $aggr,   
                );   
  
  check_nblade();

  get_release();

  create_homedir_share(vserver => $vserver,
                       share_name  => "%w",
                       share_pattern => "%w" );
 
=head1 DESCRIPTION

Provides the common routines for the CIFS tests.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id$
  Last revised $Date$

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package CIFS::Setup_GX;

use strict;
#use warnings;
use Options;
use Server;
use Tharn;
use Data::Dumper;
use TracePkg;
use English;
use Params::Common;
use Params::Volume;
use CIFS::Params_CIFSGX;
use Params::Validate qw(validate ARRAYREF SCALAR OBJECT UNDEF HASHREF);
use base qw(Exporter);
use TestEnv::MinWin qw(execute_cmd);


BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our @EXPORT_OK = qw(
    cifs_setup
    vserver_setup
    cifs_cleanup
    vserver_cleanup
    create_volume
    check_nblade
    get_release
    create_homedir_share
    ldap_client_setup
    vserver_create_general
    ipv6_win
    ipv6_linux

);
#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub cifs_setup(%);
sub vserver_setup(%);
sub cifs_cleanup(%);
sub vserver_cleanup(%);
sub create_volume(%);
sub check_nblade(%);
sub get_release(%);
sub create_homedir_share(%);
sub ldap_client_setup(%);
sub vserver_create_general(%);
sub ipv6_win(%);
sub ipv6_linux(%);

#---------------------------------------------------------------------

my $Trace = new TracePkg;
my $Options = new Options();

# create a new Server object first thing
my ($Server, @Nodes);

$Options->set_default("CIFS_SERVER" => "cifsvs");

$Options->process();
my $Params=$Options->get_hash();

# Client for cifs_cleanup()
my @client = @{$Params->{CLIENT}};
my $client = $client[0];

my $nsswitch = 'file';

sub _init_server {
    unless ( $Server ) {
       $Server = new Server();
       @Nodes = $Server->servers();
    }
}

#
# CIFS setup
#
sub cifs_setup(%) {
    $Trace->entry();
    my %params = validate( 
          @_,
          {
            node           => { optional => 1, type => SCALAR},
            cifs_server    => { optional => 1, type => SCALAR, default => $Params->{CIFS_SERVER} },
            cifs_domain_fullname    => { optional => 1, type => SCALAR, default => $Params->{CIFS_DOMAIN_LONG} },
            cifs_domain_shortname    => { optional => 1, type => SCALAR, default => $Params->{CIFS_DOMAIN} },
            ad_server      => { optional => 1, type => ARRAYREF, default => $Params->{AD_SERVER} },
            ad_server_ip   => { optional => 1, type => ARRAYREF, default => $Params->{AD_SERVER_IP} },
            vserver        => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
            ad_server_user => { optional => 1, type => SCALAR, default => $Params->{AD_SERVER_USER} },
            ad_server_password  => { optional => 1, type => SCALAR, default => $Params->{AD_SERVER_PASSWORD} },
            rootshare      => { optional => 1, type => SCALAR, default => "ROOTSHARE" },
            position       => { optional => 1, type => SCALAR, default => $Params->{POSITION} },
            nis_server    => { optional => 1, type => SCALAR, default => $Params->{NIS_SERVER} },
            nis_domain    => { optional => 1, type => SCALAR, default => $Params->{NIS_DOMAIN} },
            dns_server_ip => { optional => 1, type => ARRAYREF, default => $Params->{DNS_SERVER_IP} },
            dns_domain      => { optional => 1, type => ARRAYREF, default => $Params->{DNS_DOMAIN} },
            timezone      => { optional => 1, type => SCALAR, default => $Params->{TIMEZONE} },
            nsswitch      => { optional => 1, type => SCALAR, default => $Params->{NS_SWITCH} },
            ou            => { optional => 1, type => SCALAR, default => $Params->{OU} },

          }
    );
    
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));

    _init_server();

    my $map_patt_admin = "$params{cifs_domain_shortname}\\\\$params{ad_server_user}";
    my ($reply, @ns_switch, @vsreply, $check, $vsreply);
    my %server_add_args;

    my $filer = $params{node};
    $Trace->comment(msg=> "node = $filer");
	
    my @adserver = map(uc, @{$params{ad_server}});
    my @adserver_ip = @{$params{ad_server_ip}};
    my @dnsdomain = map(uc, @{$params{dns_domain}});
    my @dnsserver_ip = @{$params{dns_server_ip}};

    $Trace->comment(msg=> "AD_SERVER= @adserver AD_SERVER_IP= @adserver_ip");
    $Trace->comment(msg=> "DNS - @dnsdomain @dnsserver_ip");
  
    if ($#adserver > 4) {
          $Trace->warn(msg=> "There are too many Domain controlers specifed in AD_SERVER!  Only 5 can be configured.");
    }

    if ($#adserver != $#adserver_ip) {
          $Trace->warn(msg=> "AD_SERVER and AD_SERVER_IP must have the same number of elements!  Only 5 can be configured. $#adserver $#adserver_ip");
    }
    eval {
          $Server->vserver_cifs_show(
                  domain        => uc $params{cifs_domain_fullname},
                  'cifs-server' => uc $params{cifs_server},
                  );
    };
    if ( $@ ) {
        eval {	 
            $Server->vserver_show(
                     vserver => $params{vserver},
                     );
        };
        if ($@) { 
            $Trace->comment(msg => "VSERVER $params{vserver} does not exist on node.");
            return 0;
        }
        
        eval {
            @vsreply = $Server->network_interfaces_show(
                          server => $params{vserver},
                          fields => "address",
                          );
       };              
       if ($@) {
            $Trace->comment(msg => "Network data interface not setted.");
            return 0;
       }else {
            my $connection= 0;
            # ping data interface on the node
            foreach my $row (@vsreply){
                my %row = %$row;
                my @ip_addr=split(/:/, $row{address});
                if ($#ip_addr < 1) {
                    $reply = $Server->network_ping(
                                  node => $filer,
                                  destination   => $row{address},
                                   ); 
                    if ($reply =~ /is alive/i) {
                        $connection=0; 
                        last;
                    }else {
                        $connection++; 
                    }
                }else {
                    my @ipv6_mgmt = $Server->network_interfaces_show(
                                               server => $params{node},
                                               fields => "address",
                                               role   => "*mgmt*"
                                               );                           
                    $reply = $Server->cli(node => "$params{node}",
                                           cmd=> "run $params{node} ping6 -S $row{address} $ipv6_mgmt[0]->{address}",
                                           );       
                    if ($reply =~ /no answer from|address is not configured|usage: ping6/i) {
                        $connection++; 
                        
                    }else {
                        $connection=0; 
                        last;
                    }
                }
            }
            if ($connection>0) {
                $Trace->comment(msg => "Can not ping network data interface.");
                return 0;
            }
       }

    }else {
          $Trace->comment(msg => "CIFS SERVER $params{cifs_server} exists on node.");
          return 1;
    }

    $Trace->comment(msg => "CIFS configuration starts at this point.");

    eval {
        $vsreply = $Server->system_service_ntp_config_modify( enabled => "false" );
    };
    if (!$@){
        $Trace->comment(msg => "Setup time and timezone."); 
        # Setup time and timezone
        my ($year, $dst) = (localtime)[5,8];
        $year += 1900;
        # Need to set the timezone first for now
        eval {
            $Server->system_node_date_modify(
		      node => $filer,
		      timezone => $params{timezone},
                      );
        };
        if ($@){
            $Trace->warn(msg => "Error setting timezone to $params{timezone} on $filer");
        }
        if (open(IF, "/usr/sbin/ntpdate -q $params{ad_server_ip}[0]|")){
            my %h = (Jan => 1, Feb => 2, Mar => 3, Apr => 4, May => 5, Jun => 6,
                 Jul => 7, Aug => 8, Sep => 9, Oct => 10, Nov => 11, Dec => 12);
       
            while (<IF>){
               # Output looks like:
               # bryans@rylos<~/p4/test/os/boot>==> /usr/sbin/ntpdate -q 10.98.203.74
               # server 10.98.203.74, stratum 2, offset 0.002678, delay 0.03366
               # 23 Aug 21:44:22 ntpdate[11517]: adjust time server 10.98.203.74 offset 0.002678 sec
               # bryans@rylos<~/p4/test/os/boot>==>
               if (/^\s*(\d+)\s+(\S+)\s+(\d+):(\d+):(\d+).*\s+([\-\d\.]+) sec$/){
                   $Trace->comment(msg => "Time message from server $params{ad_server_ip}[0] => $_");
                   # We matched the time line
                   # The bummer is that the time doesn't show DST, so need to adjust
                   my ($day, $mon, $hour, $min, $sec, $offset) = ($1, $h{$2}, $3, $4, $5, $6);
                   if (abs($offset) > 600){
                       my $hostname = `/bin/hostname`;
                       chomp($hostname);
                       $Trace->warn(msg => "Time on host $hostname appears to be *way* off, offset = $offset");
                   }
                   eval "use Time::Local";
                   if ($@){
                       $Trace->comment(msg => "Could not include Time::Local to handle time adjustment");
                       return 0;
                   } else {
                       $Trace->comment(msg => "Handling time using timelocal $mon/$day/$year $hour:$min:$sec + $offset");
                       # $sec and $min won't change, everything else is in theory
                       ($sec, $min, $hour, $day, $mon, $year) = localtime(timelocal($sec, $min, $hour, $day, $mon - 1, $year - 1900) + $offset);
  
                       # Adjust the month and year
                       $mon++;
                       $year += 1900;
                   }
                   my $date = sprintf("\"%d/%d/%d %d:%02d:%02d\"", $mon, $day, $year, $hour, $min, $sec);
                   $Trace->comment(msg => "Setting date to $date, I think time is " . (scalar localtime));
                   # Use close enough timezone for now.  Once we get the time set properly, we'll fix up the timezone
                   eval {
                       $Server->system_node_date_modify(node => "*", 
                                        date => $date,
                                        );
                   };
                   if ($@){
                        $Trace->comment(msg => $@);
                        return 0;
                   }
                   last;
               }
            }
            close(IF);

        }
        $Server->system_service_ntp_config_modify( enabled => "true" );
    }

    my $output;
    # Just for kicks, verify the time for the log
    $Server->shell(cmd => "date", output => \$output);
    $Trace->comment(msg => "Date on $filer is $output");


    # DNS

##########
#
#    if ( $Params->{RELEASE} =~ /Boilermaker/i ) { 
#
#        eval {
#            $reply = $Server->system_services_dns_show(node    => $filer);
#        };
#        if (!$@) {
#            eval {
#                $Server->system_services_dns_delete(node    => $filer);
#            };
#            if ( $@ ) {
#                $Trace->warn(msg => "Cannot delete DNS on $filer - $@");
#                return 0;
#            }        
#        }        
#    
#        eval {
#            $Server->system_services_dns_create(
#	                         node    => $filer,
#	                         domains => $params{dns_domain},
#	                         state   => "enabled",
#	                         'name-servers' => $params{dns_server_ip},
#	                         );
#        };
#        if ( $@ ) {
#            $Trace->warn(msg => "Cannot create DNS on $filer - $@");
#            return 0;
#        }
#			
#        eval {
#            $Server->system_services_dns_hosts_create(
#                                                 address    => $dnsserver_ip[0],
#                                                 hostname   => $dnsdomain[0],
#                                                 );
#        }; 
#        if ( $@ ) {
#            $Trace->comment(msg => "DNS - $@");
#            return 0;
#        }
#    
#    }else {
#    
#        eval {
#            $reply = $Server->vserver_services_dns_show(vserver => $params{vserver});
#        };
#        if (!$@) {
#            eval {
#                $Server->vserver_services_dns_delete(vserver => $params{vserver});
#            };
#            if ( $@ ) {
#                $Trace->warn(msg => "Cannot delete DNS on $params{vserver} - $@");
#                return 0;
#            }        
#        }        
#        
#        eval {
#            $Server->vserver_services_dns_create(
#	                         vserver => $params{vserver},
#	                         domains => $params{dns_domain},
#	                         state   => "enabled",
#	                         'name-servers' => $params{dns_server_ip},
#	                         );
#        };
#        if ( $@ ) {
#            $Trace->warn(msg => "Cannot create DNS on $filer - $@");
#            return 0;
#        }
#    }    
        
#old

    #$reply = $Server->system_services_dns_show(node    => $filer);
    #$Trace->comment(msg => "ns1 - $reply->[0]->{ns1}  ns2 - $reply->[0]->{ns2}  ns3 - $reply->[0]->{ns3}");
	
    #$check = 0;
    #foreach my $adsv ( @adserver_ip ){
    #    if (($adsv eq $reply->[0]->{ns1})||($adsv eq $reply->[0]->{ns2})||($adsv eq $reply->[0]->{ns3})) {
    #        $check = 1;
    #        last;
    #    }
    #}
    #if ( $check == 1) {
    #    eval {
    #        $Server->system_services_dns_modify(
    #                     node    => $filer,
    #                     state   => "enabled",
    #                     );
    #    };
    #    if ( $@ ) {
    #         $Trace->warn(msg => "DNS - $@");
    #    }	
    #}else {
    #    $Trace->warn(msg => "Check the DNS setup.");
    #    return 0;
    #}
			
    #eval {
    #    $Server->system_services_dns_hosts_create(
    #              address    => $adserver_ip[0],
    #              hostname   => $adserver[0],
    #              );
    #};
    #if ( $@ ) {
    #    $Trace->comment(msg => "DNS - $@");
    #}
    
    # NIS 
    
    @ns_switch = @{$params{nsswitch}};
    foreach my $ns (@ns_switch){
					  
        if ($ns =~ /file|all/i) {
            eval {
                @vsreply = $Server->vserver_services_unix_user_show(
                             vserver     => $params{vserver},
                             user          => "root",
                             );
            };
            if ($@) {
                eval {
                    $Server->vserver_services_unix_user_create(
                          vserver => $params{vserver},
                          user    => "root",
                          id      => "0",
                          'primary-gid' => "0",  
                          ); 
                };
                if ($@) {
                    $Trace->warn(msg => "$@");
                }
            }

            eval {
                @vsreply = $Server->vserver_services_unix_group_show(
                               vserver     => $params{vserver},
                               name        => "root",
                               );
            };
            if ($@) {
                eval {
                    $Server->vserver_services_unix_group_create(
                          vserver => $params{vserver},
                          name    => "root",
                          id      => "0",
                          );            
                };
                if ($@) {
                    $Trace->warn(msg => "$@");
                }
            }
            eval {
                @vsreply = $Server->vserver_services_unix_user_show(
                             vserver     => $params{vserver},
                             user          => "nobody",
                             );
            };
            if ($@) {
                eval {
                    $Server->vserver_services_unix_user_create(
                          vserver => $params{vserver},
                          user    => "nobody",
                          id      => "65534",
                          'primary-gid' => "65534",  
                          ); 
                };
                if ($@) {
                    $Trace->warn(msg => "$@");
                }
            }

            eval {
                @vsreply = $Server->vserver_services_unix_group_show(
                               vserver     => $params{vserver},
                               name        => "nobody",
                               );
            };
            if ($@) {
                eval {
                    $Server->vserver_services_unix_group_create(
                          vserver => $params{vserver},
                          name    => "nobody",
                          id      => "65534",
                          );            
                };
                if ($@) {
                    $Trace->warn(msg => "$@");
                }
            }
        }

        if ($ns =~ /nis|all/i) {
 
            eval {    
                @vsreply = $Server->vserver_services_nis_domain_show(
                               vserver     => $params{vserver},
                               domain      => $params{nis_domain},
                               );
            };
            if ($@ ) {
                eval {
                    $Server->vserver_services_nis_domain_create(
                          vserver => $params{vserver},
                          domain  => $params{nis_domain},
                          servers => $params{nis_server},
                          active  => "true",  
                          ); 
                };
                if ( $@ ) {
                    $Trace->warn(msg => "$@");
                }
            }else {
                eval {
                    $Server->vserver_services_nis_domain_modify(
                          vserver => $params{vserver},
                          domain  => $params{nis_domain},
                          servers => $params{nis_server},
                          active  => "true",  
                          ); 
	        };
                if ( $@ ) {
                    $Trace->warn(msg => "$@");
                }
            }
        }	
    }

    # CIFS setup
    eval {
        $Server->vserver_name_mapping_show(
                  vserver     => $params{vserver},
                  direction   => "win-unix",
                  pattern     => "$map_patt_admin",
                  replacement => "root",
                  );
    };
    if ( $@ ) {
        eval {
            $vsreply = $Server->vserver_name_mapping_show(
                      vserver     => $params{vserver},
                      direction   => "win-unix",
                      pattern     => "$map_patt_admin",
                      );
            $Server->vserver_name_mapping_delete(
                      vserver     => $params{vserver},
                      direction   => "win-unix",
                      position    =>  $vsreply->[0]{position},
                      );
        };

        eval {
            $Server->vserver_name_mapping_show(
                      vserver     => $params{vserver},
                      direction   => "win-unix",
                      position    => $params{position},
                      );
            $Server->vserver_name_mapping_delete(
                      vserver     => $params{vserver},
                      direction   => "win-unix",
                      position    => $params{position},
                      );
                      
        };
        
        eval {
            $Server->vserver_name_mapping_create(
                      vserver     => $params{vserver},
                      direction   => "win-unix",
                      position    => $params{position},
                      pattern     => "$map_patt_admin",
                      replacement => "root",
                      );
        };
        if ( $@ ) {
              $Trace->warn(msg => "Can not map $map_patt_admin user - $@");
        }  
    }  			  
    
    eval {
        $Server->vserver_name_mapping_show(
                  vserver     => $params{vserver},
                  direction   => "unix-win",
                  pattern     => "root",
                  replacement => "$map_patt_admin",
                  );
    };				  
    if ( $@ ) {
       eval {
            $Server->vserver_name_mapping_show(
	              vserver     => $params{vserver},
	              direction   => "unix-win",
	              position    =>  $params{position},
                      );
            $Server->vserver_name_mapping_delete(
                      vserver     => $params{vserver},
                      direction   => "unix-win",
                      position    =>  $params{position},
                      );
        };

        eval {
            $vsreply = $Server->vserver_name_mapping_show(
                      vserver     => $params{vserver},
                      direction   => "unix-win",
                      pattern     => "root",
                      );
            $Server->vserver_name_mapping_delete(
                      vserver     => $params{vserver},
                      direction   => "unix-win",
                      position    =>  $vsreply->[0]{position},
                      );
                      
        };

        eval {
           $Server->vserver_name_mapping_create(
                      vserver     => $params{vserver},
                      direction   => "unix-win",
                      position    => $params{position},
                      pattern     => "root",
                      replacement => "$map_patt_admin",
                      );
        };
        if ( $@ ) {
              $Trace->warn(msg => "Can not map root user - $@");
        }  
    } 

    eval {

        if ( defined $params{ou} ) {
            $server_add_args{ou} = "OU=".$params{ou};
            $server_add_args{"-Verification"} = undef;
        }
        $Server->vserver_cifs_create(
                  vserver     => lc $params{vserver},
                  'cifs-server' => uc $params{cifs_server},
                  domain      => uc $params{cifs_domain_fullname},
                  -userid     => $params{ad_server_user},
                  -password   => $params{ad_server_password},
                  -reuse_machine_account => "y",
                  %server_add_args,
                  );
        check_nblade();
    };
    if ( $@ ) {
        $Trace->warn(msg => "$@");
        return 0;
    }
    
    eval {
        $Server->vserver_export_policy_rule_show(
                  vserver     => $params{vserver},
                  policyname  => "default",
                  );
    };
    if ($@){
        eval {
            $Server->vserver_export_policy_rule_create(
                      vserver     => $params{vserver},
                      policyname  => "default",
                      clientmatch => "0.0.0.0/0",
                      rorule	=> "any",
                      rwrule	=> "any",
                      superuser	=> "any",
                      'allow-suid' => "true",
                      'allow-dev'  => "true",
                      );
        };
        if ($@){
            $Trace->warn(msg => $@);
        }
    }

    eval {	
        $Server->vserver_cifs_share_create(
                  vserver       => $params{vserver},
                  'share-name'  => $params{rootshare},
                  path          => "/",
                  );
        check_nblade();
    };
    if ( $@ ) {
        $Trace->warn(msg => "$@");
    }
  
    eval {
        $Server->vserver_cifs_options_modify(
                  vserver       => $params{vserver},
                  'default-unix-user' => "root",
                  );
        check_nblade();
    };
    if ( $@ ) {
        $Trace->comment(msg => "$@");
    } 


    $Trace->exit();
    return 1;
}



#
# VServer setup
#
sub vserver_setup(%) {
    $Trace->entry();
    my %params = validate( 
       @_,
       {
            node            => { optional => 1, type => SCALAR },
            vserver         => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
            initvol_files   => { optional => 1, type => SCALAR, default => $Params->{INITIAL_VOLUME_FILES} },
            initvol_size    => { optional => 1, type => SCALAR, default => $Params->{INITIAL_VOLUME_SIZE} },
            volsecurity_style    => { optional => 1, type => SCALAR, default => $Params->{VOL_SECURITY_STYLE} },
            aggr_sleep      => { optional => 1, type => SCALAR, default => $Params->{AGGRSLEEP} },
            aggr_loop       => { optional => 1, type => SCALAR, default => $Params->{AGGRLOOP} },
            aggr_type       => { optional => 1, type => SCALAR, default => $Params->{AGGRTYPE} },
            aggr_size       => { optional => 1, type => SCALAR, default => $Params->{AGGR_SIZE} },
            aggr_name       => { optional => 1, type => SCALAR, default => $Params->{AGGR_NAME}},                
            interface_type  => { optional => 1, type => SCALAR | UNDEF, default => ["CLUSTER", "DATA", "MGMT"]},
            nis_server      => { optional => 1, type => SCALAR, default => $Params->{NIS_SERVER} },
            nis_domain      => { optional => 1, type => SCALAR, default => $Params->{NIS_DOMAIN} },
            nsswitch        => { optional => 1, type => SCALAR, default => $Params->{NS_SWITCH} },
       }
    );
	  
    _init_server();

    my %server_add_args;
    my (%data, %row, $reply, @vsreply, $res);
    my $interface_number;
    my $filer = $params{node};
	
    $Trace->comment(msg=> "Setup vserver for node = $filer");
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));

    eval {	 
          $Server->vserver_show(
               vserver => $params{vserver},
               );
    };
    if ($@) {
          # VSERVER does not exist on node
          $Trace->comment(msg => "VSERVER $params{vserver} does not exist on node");
          if ($@ = ~ /There are no entries matching your query/){
               $res = vserver_create(
                           node    => $filer,      
                           vserver => $params{vserver}, 
                           interface_type  => $params{interface_type},
                           initvol_files   => $params{initvol_files},
                           initvol_size    => $params{initvol_size},
                           volsecurity_style => $params{volsecurity_style},
                           aggr_sleep      => $params{aggr_sleep},
                           aggr_loop       => $params{aggr_loop},
                           aggr_type       => $params{aggr_type},
                           aggr_size       => $params{aggr_size},
                           aggr_name       => $params{aggr_name},  
                           nis_server      => $params{nis_server},
                           nis_domain      => $params{nis_domain},			
                           );
               if (!$res) {
                    return 0;
               }

          }else {
               $Trace->comment(msg => "$@");
               return 0;
          }
    }else {
          # VSERVER exists on node
          $Trace->comment(msg => "VSERVER $params{vserver} exists on node");
          eval {
               $reply = $Server->network_interfaces_show(
                        server => $params{vserver},
                        fields => "address",
                        );
          };
          if ($@){
               $Trace->comment(msg => "There are no network interface for vserver. Create new vserver.");
               eval {	 
                      @vsreply = $Server->volume_show(
                                   vserver => $params{vserver},
                                   );
               };
               if (!$@){
                   foreach my $row (@vsreply){
                       %row = %$row;
                       eval {
                           $Server->volume_purge(
                                vserver => $params{vserver},
                                volume  => $row{volume},
                                );
                       };
                       if ($@) {
                           $Trace->comment(msg => "$@");
                       }
                   }

                   eval {
                      $Server->volume_purge(
                           vserver => $params{vserver},
                           volume  => "*",
                           );
                   };
                   if ($@) {
                      $Trace->comment(msg => "Can not delete volume - $@");
                   }
               }
               eval {
                   $Server->network_routing_groups_route_delete( 
                          server  => $params{vserver},
                          'routing-group' => "*",
                          );
               };
               if ($@){
                   $Trace->comment(msg => "$@");
               }
               eval {
                   $Server->network_routing_groups_delete( 
                          server  => $params{vserver},
                          'routing-group' => "*",
                          );
               };
               if ($@){
                   $Trace->comment(msg => "$@");
               }
               eval {	 
                    $Server->vserver_delete(
                              vserver => $params{vserver},
                              );
               };
               if ($@) {
                    $Trace->comment(msg => "Can not delete vserver $params{vserver} - $@");
                    return 0;
               }
               # Create vserver
               $res = vserver_create(
                              node    => $filer,
                              vserver => $params{vserver}, 
                              interface_type  => $params{interface_type},
                              initvol_files   => $params{initvol_files},
                              initvol_size    => $params{initvol_size},
                              volsecurity_style => $params{volsecurity_style},
                              aggr_sleep      => $params{aggr_sleep},
                              aggr_loop       => $params{aggr_loop},
                              aggr_type       => $params{aggr_type},
                              aggr_size       => $params{aggr_size},
                              aggr_name       => $params{aggr_name},  
                              nis_server      => $params{nis_server},
                              nis_domain      => $params{nis_domain},			  
                              );
               if (!$res) {
                        $Trace->warn(msg => "Cannot create new vserver.");
                        return 0;
               }
		   
         }else {
               my $IP_vsaddress = $reply->[0]{'address'};
               $Trace->comment(msg => "IP address for vserver $params{vserver} on node is $IP_vsaddress");
               $interface_number = 0;
   	
               # Get all data interfaces from host file
               my @vif_info = $Options->get_interface_table(nodes => $filer, interface_type=>$params{interface_type});
               $Trace->comment(msg => "Output = " . Data::Dumper::Dumper(\@vif_info)); 				  
               foreach my $ref (@vif_info){
                    %data = %$ref;
                    if ($IP_vsaddress eq $data{address}) {
                          # vserver has a correct IP address
                          last;
                    }else {
                          $interface_number++;
                    }
               }			   
               if ( $interface_number > $#vif_info) {
                    # vserver has a different IP address then we have in host file
                    $Trace->comment(msg => "VSERVER IP address $IP_vsaddress does not exist in host file.");
                    return 0;
               }else {
                    eval {	 
                             @vsreply = $Server->volume_show(
                                       vserver => $params{vserver},
                                       );
                    };
                    if ($@) {
                            $Trace->comment(msg => "Check root volume for vserver $params{vserver}on node - $@");
                            return 0;
                    }
                    $Trace->comment(msg => "VSERVER $params{vserver} with IP = $IP_vsaddress exists on node $filer");	   
               }
          }
    }
    $Trace->exit();
    return 1;

}


sub vserver_create(%) {
    $Trace->entry();
    my %params = validate( 
         @_,
         {
            node          => { optional => 1, type => SCALAR },
            vserver       => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
            initvol_files => { optional => 1, type => SCALAR, default => $Params->{INITIAL_VOLUME_FILES} },
            initvol_size  => { optional => 1, type => SCALAR, default => $Params->{INITIAL_VOLUME_SIZE} },
            volsecurity_style => { optional => 1, type => SCALAR, default => $Params->{VOL_SECURITY_STYLE} },
            aggr_sleep    => { optional => 1, type => SCALAR, default => $Params->{AGGRSLEEP} },
            aggr_loop     => { optional => 1, type => SCALAR, default => $Params->{AGGRLOOP} },
            aggr_type     => { optional => 1, type => SCALAR, default => $Params->{AGGRTYPE} },
            aggr_size     => { optional => 1, type => SCALAR, default => $Params->{AGGR_SIZE} },
            aggr_name     => { optional => 1, type => SCALAR, default => $Params->{AGGR_NAME}},            
            interface_type  => { optional => 1, type => SCALAR | UNDEF, default => ["CLUSTER", "DATA", "MGMT"]},
            nis_server      => { optional => 1, type => SCALAR, default => $Params->{NIS_SERVER} },
            nis_domain      => { optional => 1, type => SCALAR, default => $Params->{NIS_DOMAIN} },
            nsswitch        => { optional => 1, type => SCALAR, default => $Params->{NS_SWITCH} },
         }
    );
	
    _init_server ();

    my (%data, %row, $reply);
    my %server_add_args;

    my $filer = $params{node};
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
    $Trace->comment(msg=> "vserver for node = $filer");
	
    my @aggrtypes = split(/,/, $params{aggr_type});
    my $aggrtype = $aggrtypes[0];

    my $use_failover_group = $Server->has_change(change => "failover-groups") &&
                                (lc $params{interface_type}) eq "data" ?
               	                        "use-failover-group" : undef; 
    my $use_failover_group_val = $Server->has_change(change => "failover-groups") &&
                                (lc $params{interface_type}) eq "data" ?
          	                        "disabled" : undef; 

    # Get all data interfaces from host file
    my @vif_info = $Options->get_interface_table(nodes => $filer, interface_type=>"DATA");
    $Trace->comment(msg => "Output = " . Data::Dumper::Dumper(\@vif_info)); 	
   
    # check IP on the host file
    my $interface_number = 0;
    my $ip_number;
    my @interface;

    foreach my $ref (@vif_info){
       %data = %$ref;
       eval {
            @interface = $Server->network_interfaces_show(
                          role  => "data",
                          fields => "address",
                          );
       };              
       if (!$@) {
            $ip_number = 0;
            # check IP on the node
            foreach my $row (@interface){
                     %row = %$row;
                     if ($row{address} eq $data{address}) {
                        last;
                     }else {
                        $ip_number++;
                     }
            }
       }
		             
       # find out a first free IP address in host file
       if ( $ip_number > $#interface) {
            # Create vserver
            $Trace->comment(msg => "First free IP address is $data{address}.");
            $Trace->comment(msg => "Create VSERVER $params{vserver}.");
   
            eval {
                $reply = $Server->storage_aggregate_show(
                                aggregate => $params{aggr_name},
                                nodes => $filer,
                                );
            };
            if ( $@ ) {
                $Trace->comment(msg => "Create new aggregate $params{aggr_name}.");
                eval {
                    $reply = $Server->storage_aggregate_create(
                                   aggregate => $params{aggr_name},
                                   diskcount => $params{aggr_size},
                                   raidtype  => $aggrtype,
                                   nodes     => $filer,
                                   );
                };
                # The aggregate may not be created quickly enough and this command may fail.
                # We should hang out for a little
                # while and then check on the aggregate creation
                sleep $params{aggr_sleep};
                my $loopcount = $params{aggr_loop};
                while ($loopcount--){
                    $reply = $Server->storage_aggregate_show(
                                      aggregate => $params{aggr_name},
                                      nodes     => $filer,
                                      );
                    if ($reply->[0]->{state} =~ /creating/i){
                        sleep $params{aggr_sleep};
                        next;
                    }else {
                        last;
                    }
                }
            }
            eval {
                $reply = $Server->storage_aggregate_show(
                                aggregate => $params{aggr_name},
                                nodes 	=> $filer,
                                );
            };
            if ( $@ ) {
                $Trace->comment(msg => "Cannot create aggregate $params{aggr_name} - $@"); 
                return 0;
            } 

            # Create a virtual server
            eval {
                $Server->vserver_create(
                         vserver	=> $params{vserver},
                         rootvolume	=> "root_".$params{vserver},
                         aggregate	=> $params{aggr_name},
                         'ns-switch' => "$nsswitch",
                         language	=> "en_US",
                         comment	=> $params{vserver},
                         'rootvolume-security-style' => $params{volsecurity_style},
                         );
            };
            if ($@){
                $Trace->comment(msg => $@);
                return 0;
            }
	
            $Server->volume_modify(
                        vserver	=> $params{vserver},
                        volume	=> "root_".$params{vserver},
                        size	=> $params{initvol_size},
                        files	=> $params{initvol_files},
                        unix_permissions => "777",
                        );

            $Server->network_routing_groups_create(
                       server  => $params{vserver},
                       'routing-group' => "d172.17.152.0/22",
                       subnet => "172.17.152.0/22",
                       role   => "data",
                       );


            $Server->network_routing_groups_route_create(
                       server  => $params{vserver},
                       'routing-group' => "d172.17.152.0/22",
                       gateway => $data{gateway},
                       );
	 
            $Server->network_interfaces_create(
                     server   => $params{vserver},
                     lif      => $data{name},
                     role     => "data",
                     'home-node' => $filer,
                     'home-port' => $data{port},
                     'status-admin' => "up",
                     'failover-policy' => "nextavail",
                     address  => $data{address},
                     netmask  => $data{netmask},
                     'routing-group' => "d172.17.152.0/22",
                     $use_failover_group => $use_failover_group_val,
                     );

            $reply = $Server->network_ping(
                          node => $filer,
                          destination   => $data{address},
                          );
            if ($reply !~ /is alive/i) {
                return 0;
            }					  
 
            eval {
                $Server->vserver_services_nis_domain_create(
                          vserver => $params{vserver},
                          domain  => $params{nis_domain},
                          servers => $params{nis_server},
                          active  => "true",  
                          ); 
            };
            if ( $@ ) {
                $Trace->warn(msg => "$@");
            }
            eval {
                $Server->vserver_nfs_create(
                          vserver => $params{vserver},
                          access  => "true",
 #                         udp     => "enabled",  
 #                         tcp     => "enabled",  
 #                         v2      => "enabled",  
 #                         v3      => "enabled",  
 # for RRT                         v4      => "enabled",
 #                         spinauth => "enabled",
                          ); 
            };
            if ( $@ ) {
                $Trace->warn(msg => "$@");
            }
            eval {
                $Server->vserver_modify(
                          vserver => $params{vserver},
                          'ns-switch' => $params{nsswitch},
                          ); 
            };
            if ( $@ ) {
                $Trace->warn(msg => "$@");
            }

            eval {
                $Server->vserver_services_unix_user_create(
		                vserver => $params{vserver},
		                user 	=> "root",
		                id		=> 0,
		                'primary-gid' => 0,
		                );
            };
            if ( $@ ) {
	           $Trace->warn(msg => "$@");
            } 
            eval {
		        $Server->vserver_services_unix_group_create(
		                vserver => $params{vserver},
		                name 	=> "root",
		                id		=> 0,
		                );
	        };
            if ( $@ ) {
	           $Trace->warn(msg => "$@");
            } 
            last;
       }else {
            $interface_number++;
       }
    }
    if ( $interface_number > $#vif_info) {
          $Trace->comment(msg => "Do not have a free IP address for the new vserver in a host file.");
          return 0;	
    }			
    $Trace->exit();
    return 1;
	
}

#
# CIFS cleanup
#
sub cifs_cleanup(%) {
    $Trace->entry();
    my %params = validate( 
          @_,
          {
            node           => { optional => 1, type => SCALAR},
            cifs_server    => { optional => 1, type => SCALAR, default => $Params->{CIFS_SERVER} },
            cifs_domain_fullname    => { optional => 1, type => SCALAR, default => $Params->{CIFS_DOMAIN_LONG} },
            ad_server      => { optional => 1, type => ARRAYREF, default => $Params->{AD_SERVER} },
            ad_server_ip   => { optional => 1, type => ARRAYREF, default => $Params->{AD_SERVER_IP} },
            vserver        => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
            ad_server_user => { optional => 1, type => SCALAR, default => $Params->{AD_SERVER_USER} },
            ad_server_password  => { optional => 1, type => SCALAR, default => $Params->{AD_SERVER_PASSWORD} },
            position       => { optional => 1, type => SCALAR, default => $Params->{POSITION} },
            rootshare      => { optional => 1, type => SCALAR, default => "ROOTSHARE" },
          }
    );
	
    _init_server ();

    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
    
    my %server_add_args;

    my $filer = $params{node};
    $Trace->comment(msg=> "node = $filer   client = $client");
		
    my @adserver = map(uc, @{$params{ad_server}});
    my @adserver_ip = @{$params{ad_server_ip}};
    
    $Trace->comment(msg=> "AD_SERVER= @adserver");
    $Trace->comment(msg=> "AD_SERVER_IP= @adserver_ip");
  
    eval {
          $Server->vserver_cifs_show(
                  vserver       => lc $params{vserver},
                  'cifs-server' => uc $params{cifs_server},
                  );
    };
    if ( $@ ) {
          $Trace->comment(msg => "CIFS SERVER $params{cifs_server} does not exist on node - $@");
          return 0;
    }else {
          $Trace->comment(msg => "CIFS SERVER $params{cifs_server} exists on node");
          
      eval {
           $Server->vserver_cifs_delete(
                  vserver       => lc $params{vserver},
                  -userid     => $params{ad_server_user},
                  -password   => $params{ad_server_password},
                  );
           check_nblade();
      };
      if ( $@ ) {
           $Trace->comment(msg => "Cannot delete cifs server - $@");
           return 0;
      }
    }

#    eval {
#        $Server->vserver_cifs_domain_preferred_dc_remove(
#                  vserver  => $params{vserver},
#                  domain   => $params{cifs_domain_fullname},
#                  lif      => '*',
#                  );
#    };
#    if ( $@ ) {
#        $Trace->comment(msg => "cifs domain preferred dc - $@");
#    }

    $Trace->exit();
    return 1;
}

#
# VServer cleanup
#
sub vserver_cleanup(%) {
    $Trace->entry();
    my %params = validate( 
       @_,
       {
         node            => { optional => 1, type => SCALAR },
         vserver         => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
         interface_type  => { optional => 1, type => SCALAR | UNDEF, default => ["CLUSTER", "DATA", "MGMT"]},
       }
    );
    my (@row, $reply);  
    my %server_add_args;
    my $filer = $params{node};
	
    _init_server ();

    $Trace->comment(msg=> "Setup vserver for node = $filer");
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
        
    eval {	 
          $Server->vserver_show(
               vserver => $params{vserver},
               );
    };
    if ($@) {
          # VSERVER does not exist on node
          $Trace->comment(msg => "VSERVER $params{vserver} does not exist on node");
          return 0;
    }else {
      if ($params{interface_type} eq "DATA") {
          # VSERVER exists on node
          $Trace->comment(msg => "VSERVER $params{vserver} exists on node");
          eval {
               @row = $Server->network_interfaces_show(
                                        server => $params{vserver},
                                        role   => "data",
                                        );
          };
          if ($@) {
               #DATA network interface does not exist on node
               $Trace->comment(msg => "DATA network interface does not exist on node - $@");
          }else {	
   	           eval {
                    $reply = $Server->network_interfaces_delete(
                                        server => $params{vserver},
                                        lif => "*",
                                        );
               };
               if ($@){
                    $Trace->comment(msg => "Cannot delete network interface - $@");
               }
          }

          eval {
               $Server->network_routing_groups_route_delete( 
                          server  => $params{vserver},
                          'routing-group' => "*",
                          );
          };
          if ($@){
               $Trace->comment(msg => "$@");
          }
          eval {
               $Server->network_routing_groups_delete( 
                          server  => $params{vserver},
                          'routing-group' => "*",
                          );
          };
          if ($@){
               $Trace->comment(msg => "$@");
          }

          eval {
               $Server->volume_purge(
                           vserver => $params{vserver},
                           volume  => "*",
                           );
          };
          if ($@){
              eval {
                    @row = $Server->volume_show(
                           vserver => $params{vserver},
                           );
              };
              if (!$@) {   
                 foreach my $row (@row) {  
                    eval {   
                           $Server->volume_purge(
                                    vserver => $params{vserver},
                                    volume  => $row->{volume},
                                    );
                    };
                    if ($@) {
                           $Trace->comment(msg => "$@");
                    }
                 };
                 eval {
                    $Server->volume_purge(
                          vserver => $params{vserver},
                          volume  => "*",
                          );
                 };
                 if ($@){
                    $Trace->comment(msg => "Cannot delete volume - $@");
                 }
             }
          }

          eval {	 
              $Server->vserver_delete(
                              vserver => $params{vserver},
                              );
          };
          if ($@) {
              $Trace->comment(msg => "Can not delete vserver $params{vserver} - $@");
              return 0;
          }
       }
    }
	   
    $Trace->exit();
    return 1;

}

#
# Create volume
#
sub create_volume(%)
{
    $Trace->entry();
    my %params = validate( 
       @_,
       {
        node          => { optional => 1, type => SCALAR},
        vserver       => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
        filer_vol     => { optional => 1, type => SCALAR, default => $Params->{FILER_VOL} },
        aggr_name     => { optional => 1, type => SCALAR, default => $Params->{AGGR_NAME}},    
        volsecurity_style    => { optional => 1, type => SCALAR, default => $Params->{VOL_SECURITY_STYLE} },
        vol_striped_format    => { optional => 1, type => SCALAR, default => $Params->{VOL_STRIPE_FORMAT} },
        vol_type        => { optional => 1, type => SCALAR, default => $Params->{VOLUME_TYPE} },
        unix_permissions => { optional => 1, type => SCALAR, default => $Params->{UNIX_PERMISSIONS} },
        initvol_size    => { optional => 1, type => SCALAR, default => $Params->{INITIAL_VOLUME_SIZE} },
        av_policy        => { optional => 1, type => SCALAR, default => $Params->{ANTIVIRUS_ON_ACCESS_POLICY} },
        policy        => { optional => 1, type => SCALAR, default => $Params->{EXPORT_POLICY} }, 
        junction_path => { optional => 1, type => SCALAR, default => $Params->{JUNCTION_PATH} },
        }
    );
    
    _init_server ();

    if (!defined $params{junction_path}) {
        $params{junction_path} = "/$params{filer_vol}";
    }

    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
    my $filer = $params{node};

    my $vsreply;
    
    eval {
        $vsreply = $Server->storage_aggregate_show(
                    aggregate => $params{aggr_name},
                    ); 
    };   
    
    if (($vsreply->[0]{'volume-style'} eq "striped")&& ($params{vol_striped_format} eq "data")) {   
	
        if ($params{volsecurity_style} eq "ntfs") {

            eval {
                $Trace->comment(msg => "Create coral volume with ntfs security-style.");		
                $Server->volume_create(
                            vserver => $params{vserver},
                            volume => $params{filer_vol},
                            aggregate => $params{aggr_name},
                            state => "online",
                            type => $params{vol_type},
                            'junction-path' => $params{junction_path}, 
                            size => $params{initvol_size},
                            'space-guarantee' => "volume",
                            'snapshot-policy' => "default",
                            policy => $params{policy},
                            'security-style'  => $params{volsecurity_style},
                            'stripe-format' => $params{vol_striped_format},
                            'antivirus-on-access-policy' => $params{av_policy},
                            comment => $params{filer_vol} . "_ntfs_coral_volume_created_script",
                            foreground => "true",
                            -Timeout=>600,
                            );

            };
            if ($@) {
                $Trace->comment(msg => "Cannot create volume - '$@'");
                return 0;
            }

        }else   {

            $Trace->comment(msg => "Create coral volume with $params{volsecurity_style} security-style.");		
            eval {
                 $Server->volume_create(
                             vserver => $params{vserver},
                             volume => $params{filer_vol},
                             aggregate => $params{aggr_name},
                             state => "online",
                             type => $params{vol_type},
                             'junction-path' => $params{junction_path},
                             size => $params{initvol_size},
                             'space-guarantee' => "volume",
                             'snapshot-policy' => "default",
                             policy => $params{policy},
                             'stripe-format' => $params{vol_striped_format},
                             'security-style'  => $params{volsecurity_style},
                             'unix-permissions' => $params{unix_permissions},
                             'antivirus-on-access-policy' => $params{av_policy},
                             comment => $params{filer_vol} . "_mixed_or_unix_coral_volume_created_by_script",
                             foreground => "true",
                             -Timeout=>600,
                             );
            };
            if ($@) {
                $Trace->comment(msg => "Cannot create volume - '$@'");
                return 0;
            }
        }
		 
    } else {
	 
        if ($params{volsecurity_style} eq "ntfs") {
            $Trace->comment(msg => "Create volume with ntfs security-style.");		 
            eval {
                $Server->volume_create(
                            vserver => $params{vserver},
                            volume => $params{filer_vol},
                            aggregate => $params{aggr_name},
                            state => "online",
                            type => $params{vol_type},
                            'junction-path' => $params{junction_path},
                            size => $params{initvol_size},
                            'space-guarantee' => "volume",
                            'snapshot-policy' => "default",
                            policy => $params{policy},
                            'security-style'  => $params{volsecurity_style},
                            'antivirus-on-access-policy' => $params{av_policy},
                            comment => $params{filer_vol} . "_ntfs_volume_created_script",
                            foreground => "true",
                            -Timeout=>600, 
                            );

            };
            if ($@) {
                $Trace->comment(msg => "Cannot create volume - '$@'");
                return 0;
            }

        }else   {
            $Trace->comment(msg => "Create volume with $params{volsecurity_style} security-style.");		

            eval {
                $Server->volume_create(
                            vserver => $params{vserver},
                            volume => $params{filer_vol},
                            aggregate => $params{aggr_name},
                            state => "online",
                            type => $params{vol_type},
                            'junction-path' => $params{junction_path},
                            size => $params{initvol_size},
                            'space-guarantee' => "volume",
                            'snapshot-policy' => "default",
                            policy => $params{policy},
                            'security-style'  => $params{volsecurity_style},
                            'unix-permissions' => $params{unix_permissions},
                            'antivirus-on-access-policy' => $params{av_policy},
                            comment => $params{filer_vol} . "_mixed_or_unix_volume_created_by_script",
                            foreground => "true",
                            -Timeout=>600,
                            );
            };
            if ($@) {
                $Trace->comment(msg => "Cannot create volume - '$@'");
                return 0;
            }
         }	  
    } 

    return 1;   

}
#---------------------------------------------------------
sub check_nblade(%)
{
    my ($vsreply, @vsreply);
    my $wait_limit = 0;
    
    _init_server();

    if ( $Params->{NEW_FUNCTIONALITIES} ) {
            
        while (1) {
        
            eval {
               $vsreply = $Server->diag_nblade_replay_show();
            };
        
            if ($vsreply->[0]->{result} eq 'succeeded') {
                $Trace->comment(msg => "------ $vsreply->[0]->{result} ");
                return 1;
            }elsif ($vsreply->[0]->{result} eq 'failed') {
                $Trace->comment(msg => "------ $vsreply->[0]->{result} ");
                return 0;
            }elsif ($wait_limit > 5 ){
                $Trace->comment(msg => "------ Waiting more than 2 minutes for result.");
                return 0;
            }
            
            sleep 5;
            $wait_limit++;
        }
    }    
# 25
    return 1;
}

#---------------------------------------------------------
sub get_release(%)
{

    _init_server ();
    $Options->set("RELEASE" => $Server->get_release(output => "cooked"));    
    $Trace->comment(msg => "RELEASE ===================== $Params->{RELEASE}");  
    
    $Options->set("NEW_FUNCTIONALITIES" => "1");
 
    # New Functionality for login - 'secd-login-passwd-prompt'

    # New Functionality for preferred-dc - 'vserver-cifs-domain-preferred-dc-v2' 
    if ( $Server->has_change(change => 'vserver-cifs-domain-preferred-dc-v2') ) {
        $Options->set("NEW_PREFERRED_DC" => "1");
    }else {
        $Options->set("NEW_PREFERRED_DC" => "0" );
    }  
    
    # New Functionality for share path - 'path-validation' 
    if ( $Server->has_change(change => 'path-validation') ) {
        $Options->set("NEW_PATH_VALIDATION" => "1");
    }else {
        $Options->set("NEW_PATH_VALIDATION" => "0" );
    }    

    # New Functionality for SecD - 'secd-diag-node-mandatory-argument'  
    if ( $Server->has_change(change => 'secd-diag-node-mandatory-argument') ) {
        $Options->set("NEW_SECD_FUNC" => "1");
    }else {
        $Options->set("NEW_SECD_FUNC" => "0" );
    }    
 
    # New functionality for SYMLINK  - 'symlink-properties-default'
    if ( $Server->has_change(change => 'symlink-properties-default') ) {
        $Options->set("SYMLINK_DEFAULT_ENABLED" => "1");
    }else {
        $Options->set("SYMLINK_DEFAULT_ENABLED" => "0" );
    }    

    $Trace->comment(msg => "============= New Functionality is $Params->{NEW_FUNCTIONALITIES} ;  New SecD is $Params->{NEW_SECD_FUNC}");  

    return 1;
}

#---------------------------------------------------------
sub create_homedir_share(%)

{
    $Trace->entry();
    my %params = validate( 
       @_,
       {
        vserver       => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
        share_name     => { optional => 1, type => SCALAR, },    
        share_pattern => { optional => 1, type => SCALAR, },
        }
    );
    
    my $share_properties = "oplocks,browsable,changenotify,homedirectory";

    _init_server ();
    
    if ( $Params->{NEW_FUNCTIONALITIES} ) {
        eval {
            $Server->vserver_cifs_share_create(
                        vserver       => $params{vserver},
                        'share-name'  => $params{share_name},
                        path          => $params{share_pattern},
                        'share-properties' => $share_properties,
                        );
            check_nblade();
        };
        if ($@) {
            if ( $@ !~ /duplicate entry/) {
                $Log->comment("***** Cannot create cifs homedir share - '$@'.");
                $Log->comment(" ---------------------------------------------");
                return 0;
            }
        }
    }else {
        eval {
            $Server->vserver_cifs_home_directory_share_pattern_create(
                            vserver       => $params{vserver},
                            share         => $params{share_name},
                            pattern       => $params{share_pattern},
                            );
        };
        if ($@) {
            if ( $@ !~ /duplicate entry/ ) {
                $Trace->comment(msg => "***** Can not create cifs homedir share pattern  - '$@'");
                $Log->comment(" ---------------------------------------------");
                return 0;
            }
        }	
    }

    return 1;

}

#---------------------------------------------------------
sub ldap_client_setup(%)

{
    $Trace->entry();
    my %params = validate( 
       @_,
       {
        node        => { optional => 1, type => SCALAR},
        vserver     => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
        ldap_client => { optional => 1, type => SCALAR, default => $Params->{LDAP_CLIENT_CONFIG} },    
        ldap_server => { optional => 1, type => SCALAR, default => $Params->{LDAP_SERVER} },
        ldap_schema => { optional => 1, type => SCALAR, default => $Params->{LDAP_SCHEMA} },
        base_dn     => { optional => 1, type => SCALAR, default => $Params->{LDAP_BASE_DN} },
        bind_dn     => { optional => 1, type => SCALAR, default => $Params->{LDAP_BIND_DN} },
        bind_password => { optional => 1, type => SCALAR, default => $Params->{LDAP_BIND_PASSWORD} },
       }
    );
    
    _init_server ();
 
    eval {
         $Server->vserver_services_ldap_client_show(
                         'client-config'  => $params{ldap_client},
                         schema  => $params{ldap_schema},
                         servers => $params{ldap_server},
                         );
    };
    if ($@) {        
        eval {
            $Server->cli(node => "$params{node}",
                  cmd=> "vserver services ldap client create  -client-config $params{ldap_client} -schema $params{ldap_schema} -servers $params{ldap_server} -base-dn $params{base_dn} -bind-dn $params{bind_dn}", 
                 'additional_prompts' => {
                     'Please enter password:'=> "$params{bind_password}",
                     'Confirm password:'=> "$params{bind_password}",
                 },
                 'additional_prompts_timeout' => 100,
                 "^$params{node}" => "*::.*>",
                 );
       };
    }
    
    eval {
         $Server->vserver_services_ldap_show(
                        vserver       => $params{vserver},
                        'client-config'  => $params{ldap_client},
                         );
    };
    if ($@) {        
        eval {
            $Server->vserver_services_ldap_create(
                        vserver       => $params{vserver},
                        'client-config'  => $params{ldap_client},
                        'client-enabled' => "true",
                        );
        };
        if ($@) {
            if ( $@ !~ /duplicate entry/) {
                $Log->comment("***** Cannot create ldap - '$@'.");
                $Log->comment(" ---------------------------------------------");
                return 0;
            }
        }
    }

    eval {
        $Server->vserver_services_ldap_modify(
                        vserver       => $params{vserver},
                        'client-config'  => $params{ldap_client},
                        'client-enabled' => "true",
                        );
    };
 
    return 1;

}

#---------------------------------------------------------

sub vserver_create_general(%) {
    $Trace->entry();
    my %params = validate( 
         @_,
         {
            node          => { optional => 1, type => SCALAR },
            nodename      => { optional => 1, type => SCALAR },
            vserver       => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
            initvol_files => { optional => 1, type => SCALAR, default => $Params->{INITIAL_VOLUME_FILES} },
            initvol_size  => { optional => 1, type => SCALAR, default => $Params->{INITIAL_VOLUME_SIZE} },
            volsecurity_style => { optional => 1, type => SCALAR, default => $Params->{VOL_SECURITY_STYLE} },
            aggr_sleep    => { optional => 1, type => SCALAR, default => $Params->{AGGRSLEEP} },
            aggr_loop     => { optional => 1, type => SCALAR, default => $Params->{AGGRLOOP} },
            aggr_type     => { optional => 1, type => SCALAR, default => $Params->{AGGRTYPE} },
            aggr_size     => { optional => 1, type => SCALAR, default => $Params->{AGGR_SIZE} },
            aggr_name     => { optional => 1, type => SCALAR, default => $Params->{AGGR_NAME}}, 
            lif           => { optional => 1, type => SCALAR }, 
            home_port     => { optional => 1, type => SCALAR},
            address       => { optional => 1, type => SCALAR},
            netmask       => { optional => 1, type => SCALAR},
            gateway       => { optional => 1, type => SCALAR},
            routing_group   => { optional => 1, type => SCALAR},            
            interface_type  => { optional => 1, type => SCALAR | UNDEF, default => ["CLUSTER", "DATA", "MGMT"]},
            nis_server      => { optional => 1, type => SCALAR, default => $Params->{NIS_SERVER} },
            nis_domain      => { optional => 1, type => SCALAR, default => $Params->{NIS_DOMAIN} },
            nsswitch        => { optional => 1, type => SCALAR, default => $Params->{NS_SWITCH} },
            
         }
    );
	
    _init_server ();

    my $reply;

    my @aggrtypes = split(/,/, $params{aggr_type});
    my $aggrtype = $aggrtypes[0];

    my $filer = $params{node};
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
    $Trace->comment(msg=> "vserver for node = $params{nodename}");
	   
    eval {
        $reply = $Server->storage_aggregate_show(
                                aggregate => $params{aggr_name},
                                nodes => $params{nodename},
                                );
    };
    if ( $@ ) {
        $Trace->comment(msg => "Create new aggregate $params{aggr_name}.");
        eval {
            $reply = $Server->storage_aggregate_create(
                                   aggregate => $params{aggr_name},
                                   diskcount => $params{aggr_size},
                                   raidtype  => $aggrtype,
                                   nodes     => $params{nodename},
                                   );
        };
        # The aggregate may not be created quickly enough and this command may fail.
        # We should hang out for a little
        # while and then check on the aggregate creation
        sleep $params{aggr_sleep};
        my $loopcount = $params{aggr_loop};
        while ($loopcount--){
            $reply = $Server->storage_aggregate_show(
                                      aggregate => $params{aggr_name},
                                      nodes     => $params{nodename},
                                      );
            if ($reply->[0]->{state} =~ /creating/i){
                sleep $params{aggr_sleep};
                next;
            }else {
                last;
            }
        }
    }
    eval {
        $reply = $Server->storage_aggregate_show(
                                aggregate => $params{aggr_name},
                                nodes 	=> $params{nodename},
                                );
    };
    if ( $@ ) {
        $Trace->comment(msg => "Cannot create aggregate $params{aggr_name} - $@"); 
        return 0;
    } 

    # Create a virtual server
    eval {
        $Server->vserver_create(
                         vserver	=> $params{vserver},
                         rootvolume	=> "root_".$params{vserver},
                         aggregate	=> $params{aggr_name},
                         'ns-switch'    => $params{nsswitch},
                         language	=> "en_US",
                         comment	=> $params{vserver},
                         'rootvolume-security-style' => $params{volsecurity_style},
                         );
    };
    if ($@){
        $Trace->comment(msg => $@);
        return 0;
    }

    $Server->volume_modify(
                        vserver	=> $params{vserver},
                        volume	=> "root_".$params{vserver},
                        size	=> $params{initvol_size},
                        files	=> $params{initvol_files},
                        );
	 
    $Server->network_interfaces_create(
                     server   => $params{vserver},
                     lif      => $params{lif},
                     role     => $params{interface_type},
                     'home-node' => $params{nodename},
                     'home-port' => $params{home_port},
                     'status-admin' => "up",
                     'failover-policy' => "nextavail",
                     address  => $params{address},
                     netmask  => $params{netmask},                     
                     );

    $reply = $Server->network_ping(
                          node => $params{nodename},
                          destination   => $params{address},
                          );
    if ($reply !~ /is alive/i) {
        return 0;
    }					  

    $Server->network_routing_groups_route_create(
                       server  => $params{vserver},
                       'routing-group' => $params{routing_group},
                       gateway => $params{gateway},
                       ); 
    eval {
        $Server->vserver_services_nis_domain_create(
                          vserver => $params{vserver},
                          domain  => $params{nis_domain},
                          servers => $params{nis_server},
                          active  => "true",  
                          ); 
    };
    if ( $@ ) {
        $Trace->warn(msg => "$@");
    }
    eval {
        $Server->vserver_nfs_create(
                          vserver => $params{vserver},
                          access  => "true",
#                          udp     => "enabled",  
#                          tcp     => "enabled",  
#                          v2      => "enabled",  
#                          v3      => "enabled",  
#                          v4      => "enabled",
#                          spinauth => "enabled",
                          ); 
    };
    if ( $@ ) {
        $Trace->warn(msg => "$@");
    }

    eval {
        $Server->vserver_services_unix_user_create(
		                vserver => $params{vserver},
		                user 	=> "root",
		                id		=> 0,
		                'primary-gid' => 0,
		                );
    };
    if ( $@ ) {
        $Trace->warn(msg => "$@");
    } 

    eval {
        $Server->vserver_services_unix_group_create(
		                vserver => $params{vserver},
		                name 	=> "root",
		                id		=> 0,
		                );
    };
    if ( $@ ) {
        $Trace->warn(msg => "$@");
    } 
		
    $Trace->exit();
    
    return 1;
	
}

#---------------------------------------------------------

sub ipv6_win (%) {
    $Trace->entry();
	$Trace->comment(msg => "#############IPV6_WIN##############################");
    my %params = validate( 
         @_,
         {
            address          => { optional => 1, type => SCALAR }
		 }
		 );
    my $ip_addr=$params{address};
    my @ipv6_addr=split(/:/, $ip_addr);
    print "$#ipv6_addr \n";
    
    if ($#ipv6_addr < 1)
    {
        return $ip_addr;
    }
        
    my ( $add_zero, @ipv6_win, $i);
    $add_zero=8-$#ipv6_addr;

    if ($add_zero>0)
    {
        foreach ( @ipv6_addr)
        {
            if ( $_)
            {
                push (@ipv6_win, $_);
            }
            else
            {
                for($i=0;$i<$add_zero;$i++)
                {
                    push (@ipv6_win, '0');
                }
            }
        }
    }
    $ip_addr = join('-',@ipv6_win);
    $ip_addr= $ip_addr.".ipv6-literal.net";
	$Trace->exit();
    return $ip_addr;
}

#---------------------------------------------------------
sub ipv6_linux (%) {
    $Trace->entry();
	$Trace->comment(msg => "#############IPV6_LINUX##############################");
    my %params = validate( 
         @_,
         {
            address          => { optional => 1, type => SCALAR }
		 }
		 );
    my $ip_addr=$params{address};
    my @ipv6_addr=split(/:/, $ip_addr);
    print "$#ipv6_addr \n";
    
    if ($#ipv6_addr < 1)
    {
        return $ip_addr;
    }
    
    $ip_addr= "'[".$ip_addr."]'";
	$Trace->exit();
    return $ip_addr;
}

1;


#---------------------------------------------------------

