#Hey emacs, this is basically -*- perl -*-
# $Id: $
#
## @summary Vss - Library for the Vss tests
## @author	dl-maryv (FSS Protocols QA)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Vss - Library for the Volume Shadow Services tests

=head1 SYNOPSIS

  use CIFS::Vss qw(vss_setup, vss_cleanup);

  vss_setup(qtree_security => 'ntfs', windows_version => '5.1',
                       setup_h => \%setup_h);

  vss_cleanup(cifs_drive => 'T:');

  vss_pver1_install();

  vss_pver2_install();

  %results = test_dir_and_file_create(target_drive => 'T:');

  test_file_append(target_drive => 'T:');

  test_file_check(target_drive => 'T:', lines => 4);

  prev_ver_tab_check(target_drive => 'T:', min_snapshots => 5);

  prev_ver_content_check(target_drive => 'T:');

  prev_ver_restore_check(target_drive => 'T:');

  vss_snapshot_restore(snapshot_name => 'vsstest001', fc => $fc);

  testme();

=head1 DESCRIPTION

Provides the common routines for the VSS tests.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: $
  Last revised $Date: $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package CIFS::Vss;

use strict;
use warnings;
use base qw(Exporter);

use File::stat;
use Tharn qw(host connect logresult);
use Carp qw(croak);
use FS::Util::Setup qw(create_vol common_cleanup);
use MapShare qw(map_cifs_share unmap_cifs_share);
use ExportFS qw(share_add share_delete);
use SPARSE::Util::Logger;
use CIFS::Setup  qw(cifs_setup);
use filer qw(filer_options_set filer_qtree_security_set filer_vol_snap_delete
             filer_vol_option_set);
use CLI::FileIO qw(rdfile wrfile);
use Params::Validate qw(validate ARRAYREF HASHREF SCALAR UNDEF);
use POSIX qw(uname);
use CIFS::Ntfs qw(cacls_get cacls_compare cacls_print);
use SayWrap qw(sayEOF);
use Admin::License qw(license is_licensed);
use File::Find qw(find);
use TestEnv::MinWin qw(minwin_compatible);
use FileManip qw(remove_file);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our $VERSION = '1.0';

# Test file and directory names
my $test_file_name = 'vsstest001.txt';
my $test_dir_name = 'vsstestdirectory';
my $test_file_msg = "this file contains the string vsstest001\n";
my $test_file_append_msg = "This is text that has been appended to the file\n";

# Previous version install images
my $pver1_file = 'X:\vss\twcli32.msi';
my $pver2_file = 'X:\vss\ShadowCopyClient.msi';

# Variables used by the setup and cleanup routine
my @filer_vol_path;
my @create_vol_called;
my %cifs_setup_completed;
my @cifs_share_completed;
my @cifs_map_completed;
my @qtree_security_completed;
my %test_files;
my $usermap = '/etc/usermap.cfg';
my $filer_usermap_text;

# Functions to be exported
our @EXPORT_OK = qw(
    vss_setup
    vss_cleanup
    vss_pver1_install
    vss_pver2_install
    test_dir_and_file_create
    test_file_append
    test_file_check
    prev_ver_tab_check
    prev_ver_content_check
    prev_ver_restore_check
    vss_snapshot_restore
    testme
);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub vss_setup(%);
sub vss_cleanup(%);
sub vss_pver1_install();
sub vss_pver2_install();
sub test_dir_and_file_create(%);
sub test_file_append(%);
sub test_file_check(%);
sub prev_ver_tab_check(%);
sub prev_ver_content_check(%);
sub prev_ver_restore_check(%);
sub vss_snapshot_restore(%);
sub testme();

#---------------------------------------------------------------------

=head2 vss_setup

vss_setup(setup_h => \%setup_h
          [, qtree_security => $security ]
          [, windows_version => $win_version ]);

Performs the setup for the VSS tests including checking
the for the screen saver on the client.

The following named arguments are accepted:

=over

=item Arguments

=over

=item setup_h

(Required) : Reference to setup hold hash for cleanup.

=item qtree_security

(Optional) : Security setting for a qtree. Defaults to ntfs.

=item windows_version

(Optional) : MS Windows version number

=back

=back

=for html <hr />

=cut
#---------------------------------------------------------------------

sub vss_setup(%) {
    my %opts = validate(
        @_,
        {
            auto_snapshots  => { type => SCALAR, default => 'on',
                                 regex => qr/^(on|off)$/ },
            setup_h         => { type => HASHREF },
            qtree_security  => { type => SCALAR, default => 'ntfs',
                                 regex => qr/^(unix|mixed|ntfs)$/ },
            windows_version => { type => SCALAR, default => '5',
                                 regex => qr/^(5\.0|5\.1|5\.2|5)$/ },
        }
    );
    my (%args, $cifs_drive, $cifs_user, $err_msg, $exti, $fc,
        $fh, $host, $hostname, $i, $new_text, $nfs_enable, $op, @osrel,
        @triples, $volume);

    logStep("BEGIN: vss_setup: setup the environment");

    logStep("vss_setup: Verify client has correct Windows version");
    @osrel = uname();
    if ($osrel[2] !~ /$opts{windows_version}/) {
        croak("vss_setup: client running $osrel[2] expected " .
              "$opts{windows_version} windows release");
    }

    # Disable the screen lock on the client and just leave it off
    # Croak if the screen lock was on.
    logStep("vss_setup: Verify client has screen saver turned off");
    require CIFS::Utils_gui;
    import CIFS::Utils_gui qw(screen_lock);
    screen_lock(setting => 'off', croak_if_on => 'yes');

    # Set nfs_enable for the create_vol call based on the QTREE security style
    if ($opts{qtree_security} eq 'ntfs') {
        $nfs_enable = 'no';
    } else {
        $nfs_enable = 'yes';
    }

    # The default filer, vol, and share to use are set first. If additional
    # triples were passed in then process them in this loop also.
    @triples = ({ filer => $main::FILER, vol => $main::FILER_VOL,
                  share => $main::CIFS_SHARE });
    if (defined($opts{setup_h}->{ext_setup})) {
        push @triples, @{$opts{setup_h}->{ext_setup}};
    }
    $opts{setup_h}->{ext_setup} = \@triples;

    $i = 0;
    foreach $exti (@triples) {
        logStep("vss_setup: Create volume $exti->{vol}");
        $create_vol_called[$i] = 1;
        if ($i == 0) {
            $op = $main::EXTERN_SETUP;
        } else {
            $op = 0;
        }
        ($fc, $volume) = create_vol(filer => $exti->{filer},
                                    filer_vol => $exti->{vol},
                                    extern_setup => $op,
                                    guarantee => 'none',
                                    nfs_enable => $nfs_enable) or
            croak("vss_setup: $@");
        $exti->{fc} = $fc;
        $exti->{vol} = $volume;
        $filer_vol_path[$i] = "/vol/$exti->{vol}";

        if ($nfs_enable eq 'yes') {
            logStep("vss_setup: Verify NFS is licensed");
            is_licensed($fc, 'nfs');
            croak("vss_setup: check nfs license failed $@") if $@;
        }

        $host = host($exti->{filer});
        $hostname = $host->hostname();
        if (!defined($cifs_setup_completed{$hostname})) {
            # Determine whether to use the domain or the workgroup
            $main::CIFS_DOMAIN =~ s/\s+//g;
            if ($main::CIFS_DOMAIN ne '') {
                %args = (
                    domain => $main::CIFS_DOMAIN,
                    domain_user => "$main::CIFS_DOMAIN_ADMIN",
                    domain_pass => $main::CIFS_DOMAIN_ADMIN_PASSWD,
                    cifs_pass => $main::CIFS_DOMAIN_ADMIN_PASSWD
                    );
            } else {
                %args = (
                    workgroup => 'WORKGROUP'
                    );
            }
            logStep("vss_setup: Setup CIFS on the filer");
	    if ($main::MODE eq "7G") {
		cifs_setup(host  	=> $exti->{filer}, 
			  'auth-type'   => 'ad',
			   passwd_group => 'yes', 
			   %args);
 	    } else {
               cifs_setup(host		=> $exti->{filer},
			 'server-name'  => $exti->{filer}, 
			  passwd_group  => 'yes',
			 'auth-type'    => 'ad',
			 %args);
	   }
           if ($@) {
                # save the cifs_setup error
                $err_msg = $@;

                # cifs_setup should cleanup after itself but it can leave
                # the command 'cifs setup' running and prompting for input.
                # When 'cifs setup' gets converted to ZAPI this problem should
                # go away. For now grab the telnet connection to send a ^C to
                # the filer.
                eval {
                    $fh = host($exti->{filer});
                    $fc = connect("#telnet", -timeout, $main::TIMEOUT);
                    $fc->timeout($main::CMDTIMEOUT);
                    $fc->send_intr();
                };
                croak("vss_setup: $err_msg");
            }
            $cifs_setup_completed{$hostname} = 1;
        }

        # Set the qtree security mode on the filer volume
        if (filer_qtree_security_set($fc, $filer_vol_path[$i],
            $opts{qtree_security}) == 0) {
            croak("vss_setup: QTREE set failed on $filer_vol_path[$i] for " .
                  "$opts{qtree_security}");
        }
        $qtree_security_completed[$i] = 1;

        share_add($fc, $exti->{share}, $filer_vol_path[$i], verify => 1,
                  purge_existing => 1, umask => 0, widelink => 1,
                  nosymlink_strict_security => 1, caching => 'no')
            or croak("vss_setup: failed to share $exti->{share}");
        $cifs_share_completed[$i] = 1;

        # Create username to use for cifs based on whether AD is setup or not
        # Default is workgroup
        $hostname =~ tr/a-z/A-Z/;

        $cifs_user = "$hostname\\$main::CIFS_USER";
        # Reassign $main::CIFS_USER and so the fully qualified user name
        # gets propagated to subtests correctly
        $main::CIFS_USER = $cifs_user;
        $new_text = "\n$cifs_user == root\n";

        if (defined($main::CIFS_DOMAIN)) {
            $main::CIFS_DOMAIN =~ s/\s+//g;
            if ($main::CIFS_DOMAIN ne '') {
                # Using Domains
                $new_text .= "$main::CIFS_DOMAIN_ADMIN == root\n";
            }
        }

        logStep("vss_setup: Setup $usermap file on filer");
        $filer_usermap_text = rdfile($fc, $usermap);
        wrfile($fc, $usermap, $filer_usermap_text . $new_text)
            or croak("vss_cleanup: can't write $usermap: $@");

        $cifs_drive = map_cifs_share(
            server   => $fc->connhost(),
            share    => $exti->{share},
            client   => 'localhost',
            drive    => '*',
            username => $cifs_user,
            password => $main::CIFS_PASSWD
            )
            or croak("vss_setup: $@");
        logDbg("mapped drive $cifs_drive");
        if (!$cifs_drive) {
            croak("vss_setup: could not map the drive");
        }

        $cifs_drive .= ":" unless $cifs_drive =~ /:$/;
        logDbg("Share $exti->{share} of $host mounted as $cifs_drive");
        $exti->{cifs_drive} = $cifs_drive;
        $cifs_map_completed[$i] = 1;

        # Put the default fc, vol, and cifs_drive into the setup_h also
        if ($i == 0) {
            $opts{setup_h}->{fc} = $fc;
            $opts{setup_h}->{vol} = $volume;
            $opts{setup_h}->{cifs_drive} = $cifs_drive;
        }

        if ($opts{auto_snapshots} eq 'off') {
            logStep("vss_setup: Turn off automatic snapshots");
            $exti->{auto_snapshots} = 1;
            filer_vol_option_set($fc, $volume, 'nosnap', 'on');
        }

        $i++;
    }

    logStep("END: vss_setup:");
    return;
}

#---------------------------------------------------------------------

=head2 vss_cleanup

vss_cleanup([dirs        => \@dirs]
            [,files      => \@files]
            [,cifs_drive => $cifs_drive]
            [,options    => \%options]
            [,snapshots  => \@snapshots]
            [,ext_setup  => \@ext_setup);

Cleanup after a test run if necessary. Use this cleanup
routine if you use vss_setup.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dirs

(Optional) : List of directories to cleanup on exit.

=item files

(Optional) : List of files to cleanup on exit.

=item cifs_drive

(Optional) : Drive letter where share is mapped to.

=item options

(Optional) : Hash of options and values to restore

=item snapshots

(Optional) : List of snapshots to cleanup on exit.

=item ext_setup

(Optional) : List of {cifs_drive}, and {snapshots} for
extended setup of {fc}, {vol}, and {share}.

=back

=back

=for html <hr />

=cut
#--------------------------------------------------------------------

sub vss_cleanup(%) {
    my %opts = validate(
        @_,
        {
            dirs       => { type => ARRAYREF | UNDEF, optional => 1 },
            files      => { type => ARRAYREF | UNDEF, optional => 1 },
            cifs_drive => { type => SCALAR | UNDEF, optional => 1 },
            options    => { type => HASHREF | UNDEF, optional => 1 },
            snapshots  => { type => ARRAYREF | UNDEF, optional => 1 },
            vol        => { type => SCALAR | UNDEF, optional => 1 },
            fc         => { type => HASHREF | UNDEF, optional => 1 },
            ext_setup  => { type => ARRAYREF | UNDEF, optional => 1 }
        }
    );
    my (@arr, $cleanup, $dir, $exti, $fc, $fh, $filer_vol, $hostname, $i,
        $snapshot);

    logStep("CLEANUP == $main::CLEANUP");

    eval {
        # The snapshots for the default filer come in on this option. Add
        # them into the array of snapshots to cleanup for all filers.
        if (defined($opts{snapshots}) && @{$opts{snapshots}} > 0) {
            $opts{ext_setup}->[0]->{snapshots} = $opts{snapshots};
        }

        $i = 0;
        foreach $exti (@{$opts{ext_setup}}) {
            logStep("vss_cleanup: Reconnect filer $exti->{filer}");
            eval {
                $fh = host($exti->{filer});
                $fc = connect($main::FILER_CONN, -timeout, $main::TIMEOUT);
                $fc->timeout($main::CMDTIMEOUT);
            };
            if ($@) {
                undef $fc;
                logWarn("vss_cleanup: $@");
            }

            if ($main::CLEANUP ne 'none') {
                # restore options only for the first filer on the list
                @arr = keys(%{$opts{options}});
                if ($i == 0 && $fc && defined($opts{options}) && $#arr >= 0) {
                    logStep("vss_cleanup: restoring filer options");
                        foreach $i (@arr) {
                            eval {
                                filer_options_set($fc, $i,
                                                  ${$opts{options}}{$i});
                            };
                            logWarn("vss_cleanup: $@") if $@;
                        }
                }

                # cleanup files only for the first filer on the list
                if ($i == 0 && $cifs_map_completed[$i]) {
                    if (defined($opts{files}) && @{$opts{files}} > 0) {
                        # The unlink would fail if somebody accidently stuck
                        # "/" as a file because it is a directory so there is
                        # no need to look for it on the list.
                        logStep("vss_cleanup: removing files");
                        unlink @{$opts{files}} or
                            logDbg("vss_cleanup: unlink failed: $!");
                        @{$opts{files}} = ();
                    }

                    if (defined($opts{dirs}) && @{$opts{dirs}} > 0) {
                        logStep("vss_cleanup: removing directories");
                        # Traverse array in reverse order so that the
                        # directories are removed newest to oldest.
                        while ($dir = pop @{$opts{dirs}}) {
                            eval {
                                # Remove the directory and all files in it
                                remove_file(file => $dir, recursive => 1,
                                            client => 'localhost');
                            };
                            logWarn("vss_cleanup: $@") if $@;
                        }
                        @{$opts{dirs}} = ();
                    }

                    # Make sure the test directory is deleted
                    if (-d "$exti->{cifs_drive}/$test_dir_name") {
                        eval {
                            remove_file(file =>
                                        "$exti->{cifs_drive}/$test_dir_name",
                                        recursive => 1, client => 'localhost');
                        };
                        logWarn("vss_cleanup: $@") if $@;
                    }
                }

                # delete snapshots
                if (defined($exti->{snapshots}) && @{$exti->{snapshots}} > 0) {
                    logStep("vss_cleanup: removing snapshots");
                    eval {
                        while ($snapshot = pop @{$exti->{snapshots}}) {
                            # Remove the snapshot
                            if (ref $snapshot) {
                                filer_vol_snap_delete($snapshot->{fc},
                                                      $snapshot->{vol},
                                                      $snapshot->{snapshot});
                            } else {
                                if ($fc) {
                                    filer_vol_snap_delete($fc, $exti->{vol},
                                                          $snapshot);
                                }
                            }
                        }
                    };
                    logWarn("vss_cleanup: $@") if $@;
                }

                if ($fc && $cifs_share_completed[$i]) {
                    logStep("vss_cleanup: delete CIFS share: $exti->{share}");
                    eval {
                        share_delete($fc, $exti->{share});
                    };
                    logWarn("vss_cleanup: $@") if $@;
                }

                # Unmapping shares is done after the shares are removed
                # which is backwards but is done this way because sometimes
                # windows thinks files are still open even if they have been
                # closed. The force option on unmap_cifs_share doesn't work
                # so this is the only thing that seems to work.
                if ($cifs_map_completed[$i] && $exti->{cifs_drive}) {
                    logStep("vss_cleanup: unmapping CIFS drive: $exti->{cifs_drive}");
                    unmap_cifs_share(mappath => $exti->{cifs_drive},
                                     force => 1);
                    logWarn("vss_cleanup: $@") if $@;
                }

                if ($fc && $filer_usermap_text) {
                    logStep("vss_setup: Restore $usermap file on filer");
                    wrfile($fc, $usermap, $filer_usermap_text)
                        or logWarn("vss_cleanup: can't write $usermap: $@");
                }
                if (defined($exti->{auto_snapshots})) {
                    logStep("vss_setup: Turn back on automatic snapshots");
                    filer_vol_option_set($fc, $exti->{vol}, 'nosnap', 'off');
                }
            }

            # Call common cleanup if create_vol was called
            if ($fc && $create_vol_called[$i]) {
                # Set the qtree security mode on the filer volume back to ntfs
                if ($qtree_security_completed[$i]) {
                    logStep("vss_cleanup: Set qtree mode back to ntfs");
                    eval {
                        filer_qtree_security_set($fc, $filer_vol_path[$i],
                                                 'ntfs');
                    };
                    logWarn("vss_cleanup: $@") if $@;
                }

                logStep("vss_cleanup: Run common cleanup on vol $exti->{vol}");
                if ($i == 0 || $main::CLEANUP =~ /none|volume/) {
                    $cleanup = $main::CLEANUP;
                } else {
                    # If not the default volume then (test|all) map to volume
                    # for secondary volume cleanup.
                    $cleanup = 'volume';
                }
                # common_cleanup has purge of FILER_VOL hardcoded in so undef
                # it here
                $filer_vol = $main::FILER_VOL;
                undef $main::FILER_VOL;
                common_cleanup($fc, cleanup => $cleanup, vol => $exti->{vol});
                $main::FILER_VOL = $filer_vol;
                logWarn("vss_cleanup: $@") if $@;
            }
            $i++;
        }
    };
    logWarn("vss_cleanup: $@") if $@;
    logStep("END: vss_cleanup:");
    return;
}

#---------------------------------------------------------------------

=head2 vss_pver1_install

vss_pver1_install();

Install Previous Version Client # 1. Use this for the previous version
client that comes with win2k3 server.

=for html <hr />

=cut
#---------------------------------------------------------------------

sub vss_pver1_install() {
    my $msiexec = "msiexec /i $pver1_file /q";

    if (! -e $pver1_file) {
        croak("vss_pver1_install: $pver1_file not found");
    }

    system($msiexec);

    return;
}

#---------------------------------------------------------------------

=head2 vss_pver2_install

vss_pver2_install();

Install Previous Version Client # 2. Use this for the previous version
client that is downloadable from the MS site.

=for html <hr />

=cut
#---------------------------------------------------------------------

sub vss_pver2_install() {
    my $msiexec = "msiexec /i $pver2_file /q";

    if (! -e $pver2_file) {
        croak("vss_pver1_install: $pver2_file not found");
    }

    system($msiexec);

    return;
}

#---------------------------------------------------------------------

=head2 test_dir_and_file_create

%results = test_dir_and_file_create(target_drive => $tdrive
                                   [, file_size  => $file_size ]
                                   [, num_files  => $num_files ]);

%results = (
           dirs  => reference to array of dirs that were created,
           files => reference to array of files that were created,
           dir_tmstmp => timestamp on the directory,
           file_tmstmp => timestamp on the file
)

Create a test directory and file on the FILER volume.

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive for directory and file creation

=item file_size

(Optional) : size of file in MBs to create in the directory. Default
is to just create it with the $test_file_msg

=item num_files

(Optional) : indicates the number of files that should be created in the
directory in addition to the test file and directory. Default is 0.

=back

=item Returns

Results hash containing information about the test files and directories

=back

=for html <hr />

=cut
#--------------------------------------------------------------------

sub test_dir_and_file_create(%) {
    my %opts = validate(
        @_,
        {
            target_drive => { type => SCALAR, regex => qr/^[A-Za-z]\:$/ },
            file_size    => { type => SCALAR, regex => qr/^\d+$/,
                              default => 0 },
            num_files    => { type => SCALAR, regex => qr/^\d+$/,
                              default => 0 }
        }
    );

    my (@dirs, @files, $name, $dir_tmstmp, $file_tmstmp, $i, $j, $out, $sb,
        $tpath);

    $tpath = "$opts{target_drive}/$test_dir_name";

    # Remove the directory if it already exits
    if (-d $tpath) {
        remove_file(file => $tpath, recursive => 1, client => 'localhost');
    }
    mkdir("$tpath") or
        croak("test_dir_and_file_create: Couldn't mkdir $tpath: $!");

    # If file size is set then use dt to create the file with a known
    # pattern. The file is created and verified. Otherwise
    # create the default text file
    if ($opts{file_size} > 0) {
        $out = `X:\\bin\\dt.exe of=$opts{target_drive}\\$test_dir_name\\$test_file_name pattern=iot passes=1 limit=$opts{file_size}m dispose=keep 2>&1`;

        if ($out !~ /Total errors detected: 0.1/i) {
            croak("test_dir_and_file_create: dt failed: $out");
        }

        $sb = stat("$tpath/$test_file_name")
            or croak("test_dir_and_file_create: Couldn't stat file: $!");

        # Convert file size from bytes to MBs
        $i = $sb->size / (1024 * 1024);
        if ($i != $opts{file_size}) {
            croak("test_dir_and_file_create: Incorrect file size. exp: " .
                  "$opts{file_size} MB got: $i MB");
        }
    } else {
        open(OUTFILE, ">$tpath/$test_file_name")
            or croak("test_dir_and_file_create: Couldn't create file: $!");
        print OUTFILE "$test_file_msg"
            or croak("test_dir_and_file_create: Couldn't write to file: $!");
        close(OUTFILE);
    }

    $sb = stat("$tpath/$test_file_name")
        or croak("test_dir_and_file_create: Couldn't stat file: $!");
    $file_tmstmp = $sb->mtime;

    $sb = stat("$tpath")
        or croak("test_dir_and_file_create: Couldn't stat directory: $!");
    $dir_tmstmp = $sb->mtime;

    push @dirs, "$tpath";
    push @files, "$tpath/$test_file_name";

    if ($opts{num_files}) {
        # Add the test dir and file to the list of files created.
        $test_files{"$tpath"} = 1;
        $test_files{"$tpath/$test_file_name"} = 1;

        # Note that we don't track the files @files or @dirs since cleanup
        # will remove the whole directory and everything underneath it
        for ($i = 1; $i<=10; $i++) {
            $name = "$tpath/d$i";
            mkdir("$name")
                or croak("test_dir_and_file_create: $name: $!");
            $test_files{"$name"} = 1;
            $tpath = $name;
            for ($j = 1; $j<=(int ($opts{num_files} / 10)); $j++) {
                $name = "$tpath/f$j.txt";
                open(OUTFILE, ">$name")
                    or croak("test_dir_and_file_create: $name: $!");
                print OUTFILE "$test_file_msg"
                    or croak("test_dir_and_file_create: Couldn't write to $name: $!");
                close(OUTFILE);
                $test_files{"$name"} = 1;
            }
        }
    }

    return { dirs  => \@dirs,
             files => \@files,
             dir_tmstmp => $dir_tmstmp,
             file_tmstmp => $file_tmstmp};
}

#---------------------------------------------------------------------

=head2 test_file_append

test_file_append(target_drive => $tdrive);

Append to the test file on the FILER volume.

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive for test file append

=back

=back

=for html <hr />

=cut
#--------------------------------------------------------------------

sub test_file_append(%) {
    my %opts = validate(
        @_,
        {
            target_drive => { type => SCALAR, regex => qr/^[A-Za-z]\:$/ }
        }
    );
    my $fname = "$opts{target_drive}\\$test_dir_name\\$test_file_name";

    open(OUTFILE, ">>$fname")
        or croak("test_file_append: Couldn't open $fname: $!");

    print OUTFILE $test_file_append_msg
        or croak("test_file_append: Couldn't write to $fname: $!");

    close(OUTFILE);

    return;
}

#---------------------------------------------------------------------

=head2 test_file_check

test_file_check(target_drive => $tdrive, lines => 2);

Verify the contents of the test file. The number of lines
indicates what is expected to be in the file. The first line
is expected to contain $test_file_msg and all subsequent
lines are expected to contain $test_file_append_msg.

The following named arguments are accepted:

=over

=item Arguments

=over

=item contents

(Optional) : the contents of the file to check. This is a
reference to an array.

=item lines

(Required) : number of lines in the file.

=item target_drive

(Required) : target drive for test file append

=back

=item Returns

Error string, empty if there was no error.

=back

=for html <hr />

=cut
#--------------------------------------------------------------------

sub test_file_check(%) {
    my %opts = validate(
        @_,
        {
            contents     => { type => ARRAYREF, optional => 1 },
            lines        => { type => SCALAR, regex => qr/^\d+$/ },
            target_drive => { type => SCALAR, regex => qr/^[A-Za-z]\:$/ }
        }
    );
    my (@contents, $err, $expected, $i, $j);

    if (defined($opts{contents})) {
        # Copy the file contents into a local array
        @contents = @{$opts{contents}};
    } else {
        # Read the contents of the file into an array
        open(INFILE, "<$opts{target_drive}\\$test_dir_name\\$test_file_name")
            or croak("test_file_append: Couldn't open file: $!");
        @contents = <INFILE>;
        close(INFILE);
    }

    $err = '';
    if ($#contents < 0) {
        if ($opts{lines} > 0) {
            $err .= "$test_file_name empty should have $opts{lines} lines\n";
        }
        return $err;
    }

    # Check the contents of each line of the file.
    # The first line is expected to contain $test_file_msg and all
    # subsequent lines are expected to contain $test_file_append_msg.
    for ($i = 0; $i < $opts{lines}; $i++) {
        if ($i == 0) {
            $expected = $test_file_msg;
        } else {
            $expected = $test_file_append_msg;
        }
        if ($contents[$i] ne $expected) {
            $err .= "$test_file_name error in line $i\n";
            $err .= "expected: $expected\n";
            if (defined($contents[$i])) {
                $err .= "got: $contents[$i]\n";
            } else {
                $err .= "got: (NOTHING)\n";
            }
        }
    }

    # This is the case where there are more lines in the file than expected.
    if ($#contents > $i) {
        $j = $#contents + 1;
        $err .= "$test_file_name expected $opts{lines} got $j lines\n";
        for (; $i < $j; $i++) {
            $err .= "Extra Line: $contents[$i]\n";
        }
    }

    return $err;
}

#---------------------------------------------------------------------

=head2 prev_ver_tab_check

prev_ver_tab_check(target_drive    => $tdrive
                  [, min_snapshots => $minsnap ]);

Check previous version tab for test file and directory. There is
no good way to correlate the item in the list view with the actual
snapshot. The min_snapshots argument is used to see if there is
at least N snapshots in the list. Do not change the content of
the error messages as the tests that use this will parse the message.

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive for previous version tab check

=item min_snapshots

(Optional) : minimum number of snapshots in previous version list view

=back

=back

=for html <hr />

=cut
#--------------------------------------------------------------------

sub prev_ver_tab_check(%) {
    my %opts = validate(
        @_,
        {
            target_drive  => { type => SCALAR, regex => qr/^[A-Za-z]\:$/ },
            min_snapshots => { type => SCALAR, regex => qr/^\d+$/,
                               default => 1 }
        }
    );

    my (%res1, %res2, $res, $win);

    require CIFS::Utils_gui;
    import CIFS::Utils_gui qw(find_target_explorer
                              verify_previous_version_tab
                              close_window explorer_open_with_target);

    # Sleep to give windows a chance to see the snapshots that were just
    # created on the filer.
    sleep(10);

    $win = explorer_open_with_target(target_text => $opts{target_drive});
    croak("prev_ver_tab_check: window open failed for $opts{target_drive}")
        if (!defined($win));

    $res = find_target_explorer(target_text => $test_dir_name, window => $win);
    if (!defined($res)) {
        %res1 = verify_previous_version_tab(window => $win,
                                            target_text => $test_dir_name);
    }
    close_window(window => $win);
    croak($res) if (defined($res));

    $win = explorer_open_with_target(target_text =>
                                     $opts{target_drive}."\\$test_dir_name");
    croak("prev_ver_tab_check: window open failed for " .
        $opts{target_drive}."\\$test_dir_name") if (!defined($win));

    $res = find_target_explorer(target_text => $test_file_name, window => $win);
    if (!defined($res)) {
        %res2 = verify_previous_version_tab(window => $win,
                                            target_text => $test_file_name);
    }
    close_window(window => $win);
    croak($res) if (defined($res));

    $res = '';
    if ($res1{"present"} == 0) {
        $res .= "Previous Version Tab Not Present on target $test_dir_name\n";
    }
    if ($res1{"listed"} == 0) {
        $res .= "Previous Version Tab does not list snapshots on target $test_dir_name\n";
    } else {
        if ($res1{"listed"} < $opts{min_snapshots}) {
            $res .= "Previous Version Tab does not list minimum number " .
                    "$opts{min_snapshots} of snapshots on target " .
                    "$test_dir_name\n";
        }
    }
    if ($res2{"present"} == 0) {
        $res .= "Previous Version Tab Not Present on target $test_file_name\n";
    }
    if ($res2{"listed"} == 0) {
        $res .= "Previous Version Tab does not list snapshots on target $test_file_name\n";
    } else {
        if ($res2{"listed"} < $opts{min_snapshots}) {
            $res .= "Previous Version Tab does not list minimum number " .
                    "$opts{min_snapshots} of snapshots on target " .
                    "$test_file_name\n";
        }
    }

    if ($res ne '') {
        croak($res);
    }

    return;
}

#---------------------------------------------------------------------

=head2 prev_ver_content_check

prev_ver_content_check(lines => 2, target_drive => $tdrive, version => 3);

Check for previous version View text matches original text written to test file.
It is important to turn off automatic snapshots if your test uses this
library routine.

The following named arguments are accepted:

=over

=item Arguments

=over

=item lines

(Required) : number of lines in the file. The first line is
expected to contain $test_file_msg and all subsequent
lines are expected to contain $test_file_append_msg.

=item target_drive

(Required) : target drive for previous version tab check

=item version

(Required): The number of the previous version to get, where 0 is from the
most recent snapshot taken.

=item use_copy

(Optional): Use the copy button on the previous version window to get
the text. Default is no.

=back

=back

=for html <hr />

=cut
#----------------------------------------------------------------------

sub prev_ver_content_check(%) {
    my %opts = validate(
        @_,
        {
            lines        => { type => SCALAR, regex => qr/^\d+$/ },
            target_drive => { type => SCALAR, regex => qr/^[A-Za-z]\:$/ },
            use_copy     => { type => SCALAR, regex => qr/^(yes|no)$/,
                              default => 'no' },
            version      => { type => SCALAR, regex => qr/^\d+$/ },
        }
    );

    my ($err, %res1, $res, @res2, $win);

    require CIFS::Utils_gui;
    import CIFS::Utils_gui qw(find_target_explorer verify_previous_version_tab
                            close_window close_properties_window
                            explorer_open_with_target previous_version_copy_get
                            previous_version_text_get);

    # Sleep to give windows a chance to see the snapshots that were just
    # created on the filer.
    sleep(10);

    $win = explorer_open_with_target(target_text =>
                                     $opts{target_drive}."\\$test_dir_name");
    croak("prev_ver_content_check: window opened failed for " .
        $opts{target_drive}."\\$test_dir_name") if (!defined($win));

    $res = find_target_explorer(target_text => $test_file_name, window => $win);
    if (!defined($res)) {
        %res1 = verify_previous_version_tab(closewin => 'no', window => $win,
                                            target_text => $test_file_name);
    } else {
        close_window(window => $win);
        croak($res);
    }

    $res = "";
    if ($res1{"present"} == 0) {
        $res .= "Previous Version Tab Not Present on target $test_file_name\n";
    }

    if ($res1{"listed"} == 0) {
        $res .= "Previous Version Tab does not list snapshots on target $test_file_name\n";
    }

    if ($res =~ /Previous/) {
        close_properties_window(window => $res1{window});
        close_window(window => $win);
        croak("prev_ver_content_check: $res");
    }

    if ($opts{use_copy} eq 'no') {
        @res2 = previous_version_text_get(version => $opts{version});
    } else {
        @res2 = previous_version_copy_get(test_file_name => $test_file_name,
                                          version => $opts{version});
    }

    close_properties_window(window => $res1{window});
    close_window(window => $win);

    $err = test_file_check(contents => \@res2, lines => $opts{lines},
                           target_drive => $opts{target_drive});
    if ($err ne '') {
        $res .= "Previous Version version $opts{version}\n$err";
    }

    if ($res =~ /Previous/) {
        croak("prev_ver_content_check: $res");
    }

    return;
}

#---------------------------------------------------------------------

=head2 vss_snapshot_restore

vss_snapshot_restore(snapshot_name => $sname,
                     fc            => $fc);

Use the "snap restore" filer command to restore the test file from the given
snapshot.

The following named arguments are accepted:

=over

=item Arguments

=over

=item snapshot_name

(Required) : Name of snapshot to restore from.

=item fc

(Required) : Filer connection object.

=back

=back

=for html <hr />

=cut
#----------------------------------------------------------------------

sub vss_snapshot_restore(%) {
    my %opts = validate(
        @_,
        {
            snapshot_name => { type => SCALAR },
                       fc => { type => HASHREF }
        }
    );

    my ($cmd, $ttext);

    license($opts{fc}, 'snaprestore') or croak("vss_snapshot_restore: $@");

    $cmd = "snap restore -f -t file -s $opts{snapshot_name} $filer_vol_path[0]/$test_dir_name/$test_file_name";

    $ttext = sayEOF($opts{fc}, $cmd);

    croak("vss_snapshot_restore: $ttext") if ($ttext);

    return;
}

#---------------------------------------------------------------------

=head2 prev_ver_restore_check

prev_ver_restore_check(target_drive   => $tdrive, lines => 1, version => 0,
                      [, dir_check    => $dir_check ]
                      [, acl_check    => $acl_check ]
                      [, attr_check   => $attr_check ]
                      [, time_check   => $time_check ]
                      [, file_size    => $file_size ]
                      [, num_files    => $num_files]);

Check for previous version Restore text matches original text written to test
file. It is important to turn off automatic snapshots if your test uses this
library routine.

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive for previous version tab check

=item lines

(Optional) : number of lines in the file. The first line is
expected to contain $test_file_msg and all subsequent
lines are expected to contain $test_file_append_msg. Default is 0.

=item version

(Required): The number of the previous version to get, where 0 is from the
most recent snapshot taken.

=item dir_check

(Optional) : whether to restore and check directory (0 is default).
Otherwise the test file will be used.

=item acl_check

(Optional) : whether to check for preserved ACL (0 is default).

=item attr_check

(Optional) : whether to check for preserved attributes (0 is default).

=item time_check

(Optional) : the timestamp of the original version of the file or directory
(0 is default).

=item file_size

(Optional) : size of file in MBs created in the directory. If set then
dt was used to create the file.

=item num_files

(Optional) : indicates the number of files to check for being restored in
addition to the test file and test directory. Default is 0.

=back

=back

=for html <hr />

=cut
#----------------------------------------------------------------------

sub prev_ver_restore_check(%) {
    my %opts = validate(
        @_,
        {
            target_drive => { type => SCALAR, regex => qr/^[A-Za-z]\:$/ },
            dir_check    => { type => SCALAR, regex => qr/(0|1)/,
                              default => 0 },
            acl_check    => { type => SCALAR, regex => qr/(0|1)/,
                              default => 0 },
            attr_check   => { type => SCALAR, regex => qr/(0|1)/,
                              default => 0 },
            time_check   => { type => SCALAR, default => 0 },
            file_size    => { type => SCALAR, regex => qr/^\d+$/,
                              default => 0 },
            num_files    => { type => SCALAR, regex => qr/(0|1)/,
                              default => 0 },
            lines        => { type => SCALAR, regex => qr/^\d+$/,
                              default => 0 },
            version      => { type => SCALAR, regex => qr/^\d+$/ },
        }
    );

    my ($err, $win, %res1, $res, @res, @res2, $orig_attr, $new_attr,
        @orig_acl, @new_acl, $pass, $mod_attr, %manyfiles, $fpath, $sb,
        $fdospath);

    require CIFS::Utils_gui;
    import CIFS::Utils_gui qw(find_target_explorer verify_previous_version_tab
                            close_window close_properties_window
                            explorer_open_with_target
                            previous_version_restore_get);
    require Win32::File;
    import Win32::File qw(GetAttributes SetAttributes HIDDEN);

    # Sleep to give windows a chance to see the snapshots that were just
    # created on the filer.
    sleep(10);

    if ($opts{dir_check}) {
        $win = explorer_open_with_target(target_text => $opts{target_drive});
        croak("prev_ver_restore_check: window open failed for " .
            "$opts{target_drive}") if (!defined($win));
        $res = find_target_explorer(target_text => $test_dir_name, window => $win);
        %res1 = verify_previous_version_tab(closewin => 'no', window => $win,
                                            target_text => $test_dir_name);
        $fpath = $opts{target_drive}."/".$test_dir_name;
        $fdospath = $opts{target_drive}."\\".$test_dir_name;
    } else {
        $win = explorer_open_with_target(target_text => $opts{target_drive}."\\$test_dir_name");
        croak("prev_ver_restore_check: window open failed for " .
            $opts{target_drive}."\\$test_dir_name") if (!defined($win));
        $res = find_target_explorer(target_text => $test_file_name, window => $win);
        %res1 = verify_previous_version_tab(closewin => 'no', window => $win,
                                            target_text => $test_file_name);
        $fpath = $opts{target_drive}."/".$test_dir_name."/".$test_file_name;
        $fdospath = $opts{target_drive}."\\".$test_dir_name."\\".$test_file_name;
    }

    $res = '';
    if ($res1{"present"} == 0) {
        $res .= "Previous Version Tab Not Present on target $fpath\n";
    }

    if ($res1{"listed"} == 0) {
        $res .= "Previous Version Tab does not list snapshots on target $fpath\n";
    }

    if ($res =~ /Previous/) {
        close_properties_window(window => $res1{window});
        close_window(window => $win);
        croak("prev_ver_restore_check: $res");
    }

    # Get the file attributes before the restore
    if ($opts{attr_check}) {
        GetAttributes($fpath, $orig_attr) or
            $res .= "prev_ver_restore_check: GetAttributes failed: $!";
        # Since doing GUI testing then the HIDDEN attribute was not
        # set or the file wouldn't have worked. Just bang the bit.
        $mod_attr = $orig_attr | HIDDEN();
        SetAttributes($fpath, $mod_attr) or
            $res .= "prev_ver_restore_check: SetAttributes failed: $!";
    }

    # Get the ACLs before the restore and then change the ACL so that the
    # current file has an ACL that is different.
    if ($opts{acl_check}) {
        @orig_acl = cacls_get(file => $fpath);

        @res = `X:\\bin\\SetACL.exe -on "$fdospath" -ot file -actn ace -ace "n:Everyone;s:n;p:read,delete,read_dacl,write_dacl;m:set;w:dacl"`;
        $pass = 0;
        foreach (@res) {$pass = 1 if $_ =~ /successful/;}
        if (!$pass) {
            $res .= "prev_ver_restore_check: SetACL failed $fdospath";
        }
    }

    # Restore the file or directory
    $err = previous_version_restore_get(version => $opts{version});

    close_properties_window(window => $res1{window});
    close_window(window => $win);

    if ($err != 0) {
        croak("Error restoring $fpath\nMost likely due to insufficient " .
              "space on the volume");
    }

    if ($opts{file_size}) {
        # If the file size is set then this is a huge file so use dt.exe
        # to verify the contents. Ignore dir_check because we want to look
        # at the restored file (even if the whole directory was restored).
        my $out = `X:\\bin\\dt.exe if=$opts{target_drive}\\$test_dir_name\\$test_file_name pattern=iot passes=1 limit=$opts{file_size}m dispose=keep 2>&1`;
        if ($out !~ /Total errors detected: 0.1/i) {
            croak("prev_ver_restore_check: dt verification failed: $out");
        }
    } else {
        # Only read the contents of the file if the fpath is a file
        if (-f $fpath) {
            if ($opts{file_size}) {
            } else {
                open(INFILE, "<$fpath");
                @res2 = <INFILE>;
                close(INFILE);
            }
        }

        # Only one file to check
        $err = test_file_check(contents => \@res2, lines => $opts{lines},
                               target_drive => $opts{target_drive});
        if ($err ne '') {
            $res .= "Previous Version version $opts{version}\n$err";
        }

        if ($res =~ /Previous/) {
            croak("prev_ver_restore_check: $res");
        }
    }

    if ($opts{time_check}) {
        $sb = stat($fpath) or
            croak("prev_ver_restore_check: Couldn't stat $fpath: $!");
        if ($opts{time_check} != $sb->mtime) {
            croak("prev_ver_restore_check: Timestamps do not match from " .
                  "original to restored version. " .
                  "exp: $opts{time_check} got: $sb->mtime");
        }
    }

    if ($opts{attr_check}) {
        GetAttributes($fpath, $new_attr) or
            croak("prev_ver_restore_check: GetAttributes failed $fpath: $!");
        if ($new_attr != $orig_attr) {
            croak("prev_ver_restore_check: Attributes do not match from " .
                  "original to restored version." .
                  "exp: $orig_attr got: $new_attr");
        }
    }

    if ($opts{acl_check}) {
        @new_acl = cacls_get(file => $fpath);
        if (!cacls_compare(acl_a => \@orig_acl, acl_b => \@new_acl)) {
            cacls_print(cacls_name => 'Original ACL', acl_a => \@orig_acl);
            cacls_print(cacls_name => 'New ACL', acl_a => \@new_acl);
            croak("prev_ver_restore_check: ACL's do not match before and " .
                  "after restore $fpath");
        }
    }

    if ($opts{num_files}) {
        find sub { $manyfiles{"$File::Find::name"} = 1 }, "$opts{target_drive}/$test_dir_name";
        foreach (keys %manyfiles) {
            if ($test_files{$_} != 1) {
                croak("prev_ver_restore_check: Restored files do not " .
                      "match original list: $_");
            }
        }
        foreach (keys %test_files) {
            if ($manyfiles{$_} != 1) {
                croak("prev_ver_restore_check: File $_ not restored");
            }
        }
    }

    return;
}

#---------------------------------------------------------------------

=head2 testme

Self test routine for this library.

Use test_lib.thpl to invoke this routine.

Example
    ntest -sim PKGNAME="CIFS::Vss" SUBNAME=testme CLIENT=minwin_clnt
    CIFS_DOMAIN='cifs.lab.domain' CIFS_DOMAIN_ADMIN='admin_name'
    CIFS_DOMAIN_ADMIN_PASSWD='password' FILER=myfiler util/test_pkg.thpl

=over

=item Arguments

=over

=item CLIENT

(Required) : Minwin compatible client on which to run testme.

=item FILER

(Required) : Filer to test.

=back

=back

=for html <hr />

=cut
#----------------------------------------------------------------------

sub testme() {
    use CIFS::CommonParams;
    use Utils qw(subtest_remote);

    my ($err, %ret_values, $tmp_var, $host);

    $host = host($main::CLIENT);
    if ($host->hosttype() !~ /windows/ || !minwin_compatible($main::CLIENT)) {
        logresult('CONFIG', 'The host ' . $host->hostname() .
                  ' is not MinWin compatible, exiting.');
        return;
    }

    if ($main::CLIENT ne 'localhost') {
        eval {
            subtest_remote(test   => 'util/test_pkg.thpl',
                           client => $main::CLIENT);
        };

        logresult('CONFIG', $@) if ($@);
        exit(0);
    }

    eval {
        vss_setup(setup_h => \%ret_values, windows_version => '5');

        vss_pver1_install();

        vss_pver2_install();

        $tmp_var = test_dir_and_file_create(target_drive => $ret_values{cifs_drive});

        $ret_values{dirs} = $tmp_var->{dirs};
        $ret_values{files} = $tmp_var->{files};
    };

    if ($@) {
        $err = $@;
        vss_cleanup(%ret_values);
        logresult('FAIL', $err);
        return;
    }

    eval {
        test_file_append(target_drive => $ret_values{cifs_drive});
    };

    if ($@) {
        $err = $@;
        vss_cleanup(%ret_values);
        logresult('FAIL', $err);
        return;
    }

    eval {
        prev_ver_tab_check(target_drive => $ret_values{cifs_drive});
        prev_ver_tab_check_neg(target_drive => $ret_values{cifs_drive});
        prev_ver_content_check(target_drive => $ret_values{cifs_drive});
        prev_ver_restore_check(target_drive => $ret_values{cifs_drive});
    };

    if ($@) {
        $err = $@;
        vss_cleanup(%ret_values);
        logresult('FAIL', $err) if ($err !~ /Previous/);
        return;
    }

    eval {
        vss_cleanup(%ret_values);
    };

    if ($@) {
        logresult('FAIL', $@);
    } else {
        logresult('PASS', 'testme(): Vss.pm');
    }

    return;
}

1;
