# this is basically -*- perl -*-
# $Id: /depot/prod/test/main/lib/CIFS/Widelink_GX.pm
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
## @author olgav@netapp.com
## @status	design

=head1 NAME

Widelink_GX - Library for the Cmode/GX scripts in widelink area

=head1 SYNOPSIS

  use CIFS::Widelink_GX;

  create_symlink_on_mount_point(vserver => 'hifive_vs',
            cifs_server    => 'cifs_hifive,
            vol_name       => 'vol_to_test',
            unix_path      => '/www/test/',
            link_name      => 'my_link',
            nfs_client     => 'lincq1');

  create_domain_user( user_name      => 'user1',
			user_passwd    => 'PASSword',			
			wind_client    => 'xp-sp3-vm4',

  delete_volume_share(cifs_server => 'hifive_cifsvs',
            vol_name       => 'vol_to_test',
            cifs_share     => '*');
   
  
=head1 DESCRIPTION

Provides the common routines for the CIFS tests.

=head1 STATUS

  Shared

=head1 AUTHOR

  olgav@netapp.com  
  $Id$
  Last revised $Date$

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package CIFS::Widelink_GX;

use strict;
use warnings;
use Options;
use Server;
use Tharn;
use Data::Dumper;
use TracePkg;
use English;
use Params::Common;
use Params::Volume;
use Params::Validate qw(validate ARRAYREF SCALAR OBJECT UNDEF HASHREF);
use base qw(Exporter);
use CIFS::Params_CIFSGX;
use TestEnv::MinWin qw(execute_cmd);
use MapShare qw(map_cifs_share unmap_cifs_share);
use CIFS::Setup_GX qw(check_nblade);
use CIFS::Setup_GX qw(ipv6_win ipv6_linux);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
                         on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our @EXPORT_OK = qw(
    create_symlink_on_mount_point
    delete_symlink_on_mount_point
    check_widelink_from_client
    cleanup_homedir
    check_nfs_nis
    delete_volume_share
    create_domain_user
    delete_domain_user
    check_from_client
    check_snapmirror_from_client
);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub create_symlink_on_mount_point(%);
sub delete_symlink_on_mount_point(%);
sub check_widelink_from_client(%);
sub cleanup_homedir(%);
sub check_nfs_nis(%);
sub delete_volume_share(%);
sub create_domain_user(%);
sub delete_domain_user(%);
sub check_from_client(%);
sub check_snapmirror_from_client(%);

#---------------------------------------------------------------------

my $Trace = new TracePkg;
my $Options = new Options();

$Options->set_default("CIFS_SERVER" => "cifsvs");
$Options->process();
my $Params=$Options->get_hash();

# create a new Server object first thing
my $Server = new Server();
my @Nodes = $Server->servers();

my $admin = "Administrator";
my (@vsreply, $vsreply, %vsreply, $output, $row);

#---------------------------------------------------------
#
# Create symlink on mount point
#

 sub create_symlink_on_mount_point(%)
 {
     $Trace->entry();
     my %params = validate( 
          @_,
          {
            vserver        => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
            cifs_server    => { optional => 1, type => SCALAR, default => $Params->{CIFS_SERVER} },
            vol_name       => { optional => 1, type => SCALAR, default => $Params->{FILER_VOL} },
            unix_path      => { optional => 1, type => SCALAR },
            cifs_path      => { optional => 1, type => SCALAR, default => "" },
            link_name      => { optional => 1, type => SCALAR },
            nfs_client     => { optional => 1, type => SCALAR },
            cifs_server_IP => { optional => 1, type => SCALAR },

          }
    );
    
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
    my $host = host($params{nfs_client});
    my $nfscc = connect("#rsh;username=root"); 

    my $mountpath = "/t/$params{cifs_server}/$params{vol_name}";
    my $mountdir = "/t/$params{cifs_server}";
    my $output;
    my $Cifs_server_IP;
    $Trace->comment(msg => "DATA network interface is $params{cifs_server_IP}");
 
    $Trace->comment( msg => "Un-mounting the Volume if mounted earlier on the NFS Client.");
    say($nfscc, "umount $mountpath", '-timeout',120);
 
    $Trace->comment( msg => "Mounting the Volume on the NFS Client.");
    say($nfscc, "mkdir -p $mountpath", '-timeout',120);
    say($nfscc, "chmod -Rf 777 $mountdir", '-timeout',120);
    $Cifs_server_IP=$params{cifs_server_IP};
    $Cifs_server_IP=ipv6_linux(address=>$Cifs_server_IP);
    say($nfscc, "mount $Cifs_server_IP:/$params{vol_name} $mountpath", '-timeout',120);
    say($nfscc, "cd $mountpath/$params{cifs_path};ln -s $params{unix_path} $params{link_name}", '-timeout',120);

    (undef, undef, $output) = say($nfscc, "ls -la $mountpath/$params{cifs_path}", '-timeout',120);
    if ($output !~ /$params{link_name}/i) {
        $Trace->comment(msg => "==================================================");	
        $Trace->comment( msg => "Cannot create symlink.");
        $Trace->comment(msg => "==================================================");
        return 0;
    }
say($nfscc, "pwd", '-timeout',120); 

#    say($nfscc, "umount $mountpath", '-timeout',120);
    return 1;
 }

#--------------------------------------------------------
#
# Delete symlink on mount point
#

 sub delete_symlink_on_mount_point(%)
 {
     $Trace->entry();
     my %params = validate( 
          @_,
          {
            cifs_server    => { optional => 1, type => SCALAR, default => $Params->{CIFS_SERVER} },
            vol_name       => { optional => 1, type => SCALAR, default => $Params->{FILER_VOL} },
            link_name      => { optional => 1, type => SCALAR },
            nfs_client     => { optional => 1, type => SCALAR },
          }
    );
    
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
    my $host = host($params{nfs_client});
    my $nfscc = connect("#rsh;username=root"); 

    my $mountpath = "/t/$params{cifs_server}/$params{vol_name}";
    my $output;
    my $return_code = 1;

say($nfscc, "pwd", '-timeout',120); 
say($nfscc, "cd", '-timeout',120);
    $Trace->comment( msg => "Remove symlink and umount the volume on the NFS client.");
    say($nfscc, "cd $mountpath;rm $params{link_name}", '-timeout',120);
    sleep 30;
    (undef, undef, $output) = say($nfscc, "ls -la $mountpath", '-timeout',120);
    if ($output =~ /$params{link_name}/i) {	
        $Trace->comment(msg => "==================================================");
        $Trace->comment( msg => "Can not delete symlink.");
        $Trace->comment(msg => "==================================================");
        $return_code=0;
    }
say($nfscc, "pwd", '-timeout',120); 
say($nfscc, "cd", '-timeout',120);
 
    say($nfscc, "umount $mountpath", '-timeout',120);
    say($nfscc, "rm -rf $mountpath", '-timeout',120);

    return $return_code;
	
 } #end of _delete_symlink_on_mount_point

#---------------------------------------------------------
#
# Check widelink from client
#

 sub check_widelink_from_client(%)
 {
     $Trace->entry();
     my %params = validate( 
          @_,
          {
            cifs_server    => { optional => 1, type => SCALAR, default => $Params->{CIFS_SERVER} },
            cifs_drive     => { optional => 1, type => SCALAR, default => $Params->{CIFS_DRIVE} },
            cifs_domain    => { optional => 1, type => SCALAR, default => $Params->{CIFS_DOMAIN} },
            cifs_share     => { optional => 1, type => SCALAR, default => $Params->{CIFS_SHARE} },
            file_name      => { optional => 1, type => SCALAR },
            text_to_test   => { optional => 1, type => SCALAR },
            link_name      => { optional => 1, type => SCALAR },
            wind_client    => { optional => 1, type => SCALAR },
            user_name      => { optional => 1, type => SCALAR },
            user_passwd    => { optional => 1, type => SCALAR },			
          }
    );
    
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
    my $client = $params{wind_client};
    my $return_code = 1;

    my $drive = (defined $params{cifs_drive}) ? $params{cifs_drive} : "*";
    sleep 30;
    if ($params{user_name} !~ /$admin/i) {
        $drive = map_cifs_share(server   =>$params{cifs_server},
                        share    =>$params{cifs_share},
                        client   =>$client,
                        domain   =>$params{cifs_domain},
                        username =>$params{user_name},
                        password =>$params{user_passwd},
                        drive    =>$drive,
                        );
    }else {
        $drive = map_cifs_share(server   =>$params{cifs_server},
                        share    =>$params{cifs_share},
                        client   =>$client,
                        domain   =>$params{cifs_domain},
                        drive    =>$drive,
                        );
    }
    if ( $drive ) { 
        my $path = "$drive\\";

        $output = execute_cmd($client, "dir $path");	
        if ($output !~ /$params{link_name}/i) {	
            $Trace->comment(msg => "==================================================");
            $Trace->comment( msg => "Cannot see symlink.");
            $Trace->comment(msg => "=================================================="); 
            $return_code=0;
        }
        for ( my $i = 0; $i <= 6; $i++ ){
            $output = execute_cmd($client, "dir $path$params{link_name}\\");
            if ($output !~ /File Not Found/i) { last;}
        }    
        if ($output !~ /$params{user_name}|has no label|$params{cifs_share}/i) {	
            $Trace->comment(msg => "==================================================");
            $Trace->comment( msg => "Can not open $params{link_name} link.");
            $Trace->comment(msg => "==================================================");
            $return_code=0;
        }
        $output = execute_cmd($client, "more $path$params{link_name}\\$params{file_name}");	
        if ($output !~ /$params{text_to_test}/) {	
            $Trace->comment(msg => "==================================================");
            $Trace->comment( msg => "Can not read file or symlink is to wrong share.");
            $Trace->comment(msg => "==================================================");
            $return_code=0;
        }
        sleep 10;
        #unmap_cifs_share(drive => $drive, client => $client,);
        $output = execute_cmd($client, "net use $drive /delete /yes");
        if ($output !~ /deleted successfully/i) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment(msg => "Failed to unmap drive.");
            $Trace->comment(msg => "==================================================");
            $return_code=0;
        }

    }else {
        $Trace->comment(msg => "==================================================");
        $Trace->comment(msg => "Failed to map drive.");
        $Trace->comment(msg => "==================================================");
        $return_code=0;
    }
 
    return $return_code;
 }

#--------------------------------------------------------
#
# Check NFS and NIS on node
#

sub check_nfs_nis(%) {
    $Trace->entry();
    my %params = validate( 
          @_,
          {
            vserver    => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
          }
    );
    
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
    my $return_code = 1;
    my $count;
	my $vserver = $params{vserver};
	
    ### NFS check
    eval {
        @vsreply = $Server->vserver_nfs_show(
                   vserver       => $vserver,
                   );
    };
    if ($@){
        $Trace->comment(msg => "==================================================");
        $Trace->comment(msg => "Check NFS for $vserver - '$@'");
        $Trace->comment(msg => "==================================================");
        $return_code = 0;	
    }else {
        $count = $#vsreply; 
        foreach my $row (@vsreply){
            my %row = %$row;
            if ($row{access} eq "true") {
                last;
            }
            $count--;
        }
        if ($count < 0) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment(msg => "Check NFS for $vserver.");
            $Trace->comment(msg => "==================================================");
            $return_code = 0;	
        }
    }

    ### NIS check
    eval {
        @vsreply = $Server->vserver_services_nis_domain_show(
                   vserver       => $vserver,
                   );
    };
    if ($@) {
        $Trace->comment(msg => "==================================================");
        $Trace->comment(msg => "Check NIS for $vserver - '$@'");
        $Trace->comment(msg => "==================================================");
        $return_code = 0;	
    }else {
        $count = $#vsreply; 
        foreach my $row (@vsreply){
            my %row = %$row;
            if ($row{active} eq "true") {
                last;
            }
            $count--;
        }
        if ($count < 0) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment(msg => "Check NIS for $vserver.");
            $Trace->comment(msg => "==================================================");
            $return_code = 0;	
        }
    }

    return $return_code;

} #end of _check_nfs_nis()

#---------------------------------------------------------
#
# Delete volume and share on node
#

 sub delete_volume_share(%)
 {
    $Trace->entry();
    my %params = validate( 
          @_,
          {
            vserver        => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
            cifs_share     => { optional => 1, type => SCALAR, default => $Params->{CIFS_SHARE} },
            vol_name       => { optional => 1, type => SCALAR, default => $Params->{FILER_VOL} },
          }
    );
    
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));

    my $vserver = $params{vserver};
    my $volume_name = $params{vol_name};
    my $share_name = $params{cifs_share};
    my $return_code = 1;

    eval {
        $Server->vserver_cifs_share_show(
                       vserver => $vserver,
                       'share-name'  => $share_name, );
    };
    if (!$@) {
        eval {
            $Server->vserver_cifs_share_delete(
                       vserver       => $vserver,
                       'share-name'  => $share_name,
                       );
            check_nblade();           
        };
        if ( $@ ) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment( msg => "Share not deleted! - '$@'");
            $Trace->comment(msg => "==================================================");
            $return_code = 0;
        }
    }
    
    eval {
        $Server->volume_show(
                       vserver => $vserver,
                       volume  => $volume_name, );
    };
    if (!$@) {
        eval {
            $Server->volume_purge(
                           vserver => $vserver,
                           volume => $volume_name,
                           -Timeout=>600,
                           );
#                           foreground => "true",
        };
        if ( $@ ) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment( msg => "Volume not deleted! - '$@'");
            $Trace->comment(msg => "==================================================");
            $return_code = 0;
        }
    }
 
    return $return_code;
 }


#---------------------------------------------------------
#
# Create DOMAIN user
#

 sub create_domain_user(%)
 {
    $Trace->entry();
    my %params = validate( 
        @_,
        {
            user_name      => { optional => 1, type => SCALAR },
            user_passwd    => { optional => 1, type => SCALAR },			
            wind_client    => { optional => 1, type => SCALAR },
        }
    );
    
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
    my $return_code = 1;

   # my $command = "net user $params{user_name} /delete /domain";
   # execute_cmd($params{wind_client}, $command);

    my $command = "net user $params{user_name} $params{user_passwd} /add /expires:never /active:yes /domain";
    $output = execute_cmd($params{wind_client}, $command);
    if (!$output) {$return_code = 0;}

    return $return_code;
 }
#---------------------------------------------------------
#
# Delete usr from DOMAIN
#

 sub delete_domain_user(%)
 {
    $Trace->entry();
    my %params = validate( 
        @_,
        {
            user_name      => { optional => 1, type => SCALAR },
            wind_client    => { optional => 1, type => SCALAR },
        }
    );
    
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
    my $return_code = 1;

    my $command = "net user $params{user_name} /delete /domain";
    $output = execute_cmd($params{wind_client}, $command);
    if (!$output) {$return_code = 0;}    

    return $return_code;
 }

#---------------------------------------------------------
#
# Cleanup home directory
#

 sub cleanup_homedir(%)
 {
    $Trace->entry();
    my %params = validate( 
           @_,
           {
             vserver        => { optional => 1, type => SCALAR, default => $Params->{VSERVER} },
             hd_share       => { optional => 1, type => SCALAR },
             hd_search_path => { optional => 1, type => SCALAR },
             new_homedir_func   => { optional => 1, type => SCALAR, default => $Params->{NEW_FUNCTIONALITIES} },
           }
    );
     
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
    my $return_code = 1;
    my $vserver = $params{vserver};
	
    if ( $params{new_homedir_func} ) {
        eval {
            $Server->vserver_cifs_share_delete(
                    vserver       => $vserver,
                    'share-name'  => $params{hd_share},
                    );
            check_nblade();        
        };
        if ( $@ ) {
            if ( $@ !~ /There are no entries matching your query/i ) {
                $Trace->comment(msg => "==================================================");
                $Trace->comment(msg => "Can not delete cifs homedir share  - '$@'");
                $Trace->comment(msg => "==================================================");
                $return_code=0;
            }
        }	

    }else {
    
        eval {
            $Server->vserver_cifs_home_directory_modify(
                       vserver     => $vserver,
                       enabled     => "disabled",
                       );
        };
        if ( $@ ) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment(msg => "Can not modify cifs homedir - '$@'");
            $Trace->comment(msg => "==================================================");
            $return_code=0;
        }
    
        eval {
            $Server->vserver_cifs_home_directory_share_pattern_delete(
                    vserver       => $vserver,
                    share         => $params{hd_share},
                    );
        };
        if ( $@ ) {
            if ( $@ !~ /There are no entries matching your query/i ) {
                $Trace->comment(msg => "==================================================");
                $Trace->comment(msg => "Can not delete cifs homedir share pattern  - '$@'");
                $Trace->comment(msg => "==================================================");
                $return_code=0;
            }
        }	
    }

    eval {
        $Server->vserver_cifs_home_directory_search_path_remove(
                        vserver       => $vserver,
                        path          => $params{hd_search_path},
                        );
        check_nblade();                
    };
    if ( $@ ) {
        if ( $@ !~ /There are no entries matching your query/i ) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment(msg => "Can not delete cifs homedir share pattern  - '$@'");
            $Trace->comment(msg => "==================================================");
            $return_code=0;
        }
    }
	
    return $return_code;
 }

#---------------------------------------------------------
sub check_from_client(%)
{
    $Trace->entry();
    my %params = validate( 
           @_,
           {
            cifs_server    => { optional => 1, type => SCALAR, default => $Params->{CIFS_SERVER} },
            cifs_drive     => { optional => 1, type => SCALAR, default => $Params->{CIFS_DRIVE} },
            cifs_domain    => { optional => 1, type => SCALAR, default => $Params->{CIFS_DOMAIN} },
            cifs_share     => { optional => 1, type => SCALAR, default => $Params->{CIFS_SHARE} },
            file_name      => { optional => 1, type => SCALAR },
            text_to_test   => { optional => 1, type => SCALAR },
            wind_client    => { optional => 1, type => SCALAR },
            user_name      => { optional => 1, type => SCALAR },
            user_passwd    => { optional => 1, type => SCALAR },			

           }
    );
     
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
    my $return_code = 1;
    my $client = $params{wind_client};

    my $drive = (defined $params{cifs_drive}) ? $params{cifs_drive} : "*";
    sleep 10;

    if ($params{user_name} !~ /$admin/i) {
        $drive = map_cifs_share(server   =>$params{cifs_server},
                        share    =>$params{cifs_share},
                        client   =>$client,
                        domain   =>$params{cifs_domain},
                        username =>$params{user_name},
                        password =>$params{user_passwd},
                        drive    =>$drive,
                        );
    }else {
        $drive = map_cifs_share(server   =>$params{cifs_server},
                        share    =>$params{cifs_share},
                        client   =>$client,
                        domain   =>$params{cifs_domain},
                        drive    =>$drive,
                        );
    }
    if ( $drive ) { 
        my $path = "$drive\\";

        $output = execute_cmd($client, "dir $path");
        if (($output =~/Access denied/) || ($output =~/network path was not found/)) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment( msg => "Can not dir $path.");
            $Trace->comment(msg => "==================================================");  
            $return_code=0;
        }
        sleep 10;
        $output = execute_cmd($client, "more $path$params{file_name}");	
        if ($output !~ /$params{text_to_test}/) {
            $Trace->comment(msg => "==================================================");	
            $Trace->comment( msg => "Can not read file or maped to wrong share.");
            $Trace->comment(msg => "==================================================");
            $return_code=0;
        }
        $path = "$drive\\MYTESTDIR";
        $output = execute_cmd($client, "mkdir $path");
        if (($output =~/Access denied/) || ($output =~/network path was not found/)) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment( msg => "Can not mkdir $path.");
            $Trace->comment(msg => "==================================================");
            $return_code=0;
        }
        $output = execute_cmd($client, "echo this is my file > $path\\myfile.txt");
        if (($output =~/Access denied/) || ($output =~/network path was not found/)) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment( msg => "Can not create $path\\myfile.txt.");
            $Trace->comment(msg => "==================================================");
            $return_code=0;
        }
        $output = execute_cmd($client, "dir $path");
        if (($output =~/Access denied/) || ($output =~/network path was not found/)) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment( msg => "Can not dir $path."); 
            $Trace->comment(msg => "==================================================");
            $return_code=0;
        }
        $output = execute_cmd($client, "rmdir $path /s /q");
        $output = execute_cmd($client, "net use $drive /delete /yes");
        if ($output !~ /deleted successfully/i) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment(msg => "Failed to unmap drive.");
            $Trace->comment(msg => "==================================================");
            $return_code=0;
        }

    }else {
        $Trace->comment(msg => "==================================================");
        $Trace->comment(msg => "Failed to map drive.");
        $Trace->comment(msg => "==================================================");
        $return_code=0;
    }
    sleep 10;
    return $return_code;

}

#---------------------------------------------------------
sub check_snapmirror_from_client(%)
{
    $Trace->entry();
    my %params = validate( 
           @_,
           {
            cifs_server    => { optional => 1, type => SCALAR, default => $Params->{CIFS_SERVER} },
            cifs_drive     => { optional => 1, type => SCALAR, default => $Params->{CIFS_DRIVE} },
            cifs_domain    => { optional => 1, type => SCALAR, default => $Params->{CIFS_DOMAIN} },
            cifs_share     => { optional => 1, type => SCALAR, default => $Params->{CIFS_SHARE} },
            file_name      => { optional => 1, type => SCALAR },
            text_to_test   => { optional => 1, type => SCALAR },
            wind_client    => { optional => 1, type => SCALAR },
           }
    );
     
    $Trace->comment(msg => "Input params = " . Data::Dumper::Dumper(\%params));
    my $return_code = 1;
    my $client = $params{wind_client};

    my $drive = (defined $params{cifs_drive}) ? $params{cifs_drive} : "*";

    $drive = map_cifs_share(server   =>$params{cifs_server},
                        share    =>$params{cifs_share},
                        client   =>$client,
                        domain   =>$params{cifs_domain},
                        drive    =>$drive,
                        );
    if ( $drive ) { 
        my $path = "$drive\\";

        $output = execute_cmd($client, "dir $path");
        if (($output =~/Access denied/) || ($output =~/network path was not found/)) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment( msg => "Can not dir $path.");
            $Trace->comment(msg => "==================================================");
            $return_code=0;
        }
        $output = execute_cmd($client, "more $path\\$params{file_name}");	
        if ($output !~ /$params{text_to_test}/) {	
            $Trace->comment(msg => "==================================================");
            $Trace->comment( msg => "Can not read file or maped to wrong share.");
            $Trace->comment(msg => "==================================================");
            $return_code=0;
        }

        $output = execute_cmd($client, "net use $drive /delete /yes");
        if ($output !~ /deleted successfully/i) {
            $Trace->comment(msg => "==================================================");
            $Trace->comment(msg => "Failed to unmap drive.");
            $Trace->comment(msg => "==================================================");
            $return_code=0;
        }

    }else {
        $Trace->comment(msg => "==================================================");
        $Trace->comment(msg => "Failed to map drive.");
        $Trace->comment(msg => "==================================================");
        $return_code=0;
    }
    sleep 10;
    return $return_code;

}

#---------------------------------------------------------



1;




