# $Id$

## @summary Library for using the ReadDirectoryChangesW API on Windows.
##          This shouldn't be used directly, use the CIFS::ChangeNotify
##          APIs instead.
## @author mueller@netapp.com
## @pod here

=head1 NAME

ReadDirectoryChanges API

=head1 SYNOPSIS

    use Win32::ReadDirectoryChanges;

=head1 ROUTINES

=cut

# This package was found in it's early stages on CPAN, but had to be completely
# re-worked in order to be actually useful for NetApp testing purposes. We will
# most likely want to use this module over that which is included with PERL.

package Win32::ReadDirectoryChanges;

use 5.006001;
use strict;

require Exporter;
our @ISA = qw(Exporter);

our @EXPORT = qw(
	FILE_NOTIFY_CHANGE_FILE_NAME
	FILE_NOTIFY_CHANGE_DIR_NAME
	FILE_NOTIFY_CHANGE_NAME
	FILE_NOTIFY_CHANGE_ATTRIBUTES
	FILE_NOTIFY_CHANGE_SIZE
	FILE_NOTIFY_CHANGE_LAST_WRITE
	FILE_NOTIFY_CHANGE_LAST_ACCESS
	FILE_NOTIFY_CHANGE_CREATION
	FILE_NOTIFY_CHANGE_EA
	FILE_NOTIFY_CHANGE_SECURITY
	FILE_NOTIFY_CHANGE_STREAM_NAME
	FILE_NOTIFY_CHANGE_STREAM_SIZE
	FILE_NOTIFY_CHANGE_STREAM_WRITE
	
	FILE_ACTION_UNSPECIFIED
	FILE_ACTION_ADDED
	FILE_ACTION_REMOVED
	FILE_ACTION_MODIFIED
	FILE_ACTION_RENAMED_OLD_NAME
	FILE_ACTION_RENAMED_NEW_NAME
	FILE_ACTION_ADDED_STREAM
	FILE_ACTION_REMOVED_STREAM
	FILE_ACTION_MODIFIED_STREAM
	FILE_ACTION_REMOVED_BY_DELETE
	FILE_ACTION_ID_NOT_TUNNELLED
	FILE_ACTION_TUNNELLED_ID_COLLISION
);

our $VERSION = '0.01';

use Win32::API;
use threads;
use threads::shared;

# Exported stuff
sub FILE_NOTIFY_CHANGE_FILE_NAME    { 0x00000001 }
sub FILE_NOTIFY_CHANGE_DIR_NAME     { 0x00000002 }
sub FILE_NOTIFY_CHANGE_NAME         { 0x00000003 }
sub FILE_NOTIFY_CHANGE_ATTRIBUTES   { 0x00000004 }
sub FILE_NOTIFY_CHANGE_SIZE         { 0x00000008 }
sub FILE_NOTIFY_CHANGE_LAST_WRITE   { 0x00000010 }
sub FILE_NOTIFY_CHANGE_LAST_ACCESS  { 0x00000020 }
sub FILE_NOTIFY_CHANGE_CREATION     { 0x00000040 }
sub FILE_NOTIFY_CHANGE_EA           { 0x00000080 }
sub FILE_NOTIFY_CHANGE_SECURITY     { 0x00000100 }
sub FILE_NOTIFY_CHANGE_STREAM_NAME  { 0x00000200 }
sub FILE_NOTIFY_CHANGE_STREAM_SIZE  { 0x00000400 }
sub FILE_NOTIFY_CHANGE_STREAM_WRITE { 0x00000800 }

sub FILE_ACTION_UNSPECIFIED            { 0x00000000 }
sub FILE_ACTION_ADDED                  { 0x00000001 }
sub FILE_ACTION_REMOVED                { 0x00000002 }
sub FILE_ACTION_MODIFIED               { 0x00000003 }
sub FILE_ACTION_RENAMED_OLD_NAME       { 0x00000004 }
sub FILE_ACTION_RENAMED_NEW_NAME       { 0x00000005 }
sub FILE_ACTION_ADDED_STREAM           { 0x00000006 }
sub FILE_ACTION_REMOVED_STREAM         { 0x00000007 }
sub FILE_ACTION_MODIFIED_STREAM        { 0x00000008 }
sub FILE_ACTION_REMOVED_BY_DELETE      { 0x00000009 }
sub FILE_ACTION_ID_NOT_TUNNELLED       { 0x0000000A }
sub FILE_ACTION_TUNNELLED_ID_COLLISION { 0x0000000B }

# Internal stuff
my ($CloseHandle,
    $GetCurrentProcess,
    $OpenProcessToken,
    $LookupPrivilegeValue,
    $AdjustTokenPrivileges,
    $CreateFile,
    $ReadDirectoryChanges,
    $CreateIoCompletionPort,
    $GetQueuedCompletionStatus);

sub FILE_LIST_DIRECTORY { 0x00000001 }

sub FILE_SHARE_READ   { 0x00000001 }
sub FILE_SHARE_WRITE  { 0x00000002 }
sub FILE_SHARE_DELETE { 0x00000004 }

sub OPEN_EXISTING { 3 }

sub FILE_FLAG_BACKUP_SEMANTICS { 0x02000000 }
sub FILE_FLAG_OVERLAPPED       { 0x40000000 }

sub TOKEN_ADJUST_PRIVILEGES { 0x0020 }
sub TOKEN_QUERY             { 0x0008 }

sub SE_BACKUP_NAME        { 'SeBackupPrivilege' }
sub SE_RESTORE_NAME       { 'SeRestorePrivilege' }
sub SE_CHANGE_NOTIFY_NAME { 'SeChangeNotifyPrivilege' }

sub SE_PRIVILEGE_ENABLED  { 2 }

sub DWORD_SIZE { 4 }

sub _InitializeAPI {
	my $kernel32 = 'kernel32.dll';
	my $advapi32 = 'advapi32.dll';

	$CloseHandle = new Win32::API($kernel32,
	                              'CloseHandle',
	                              'N', 'I') || die;
	$GetCurrentProcess     = new Win32::API($kernel32,
	                                        'GetCurrentProcess',
	                                        '', 'N') || die;
	$OpenProcessToken      = new Win32::API($advapi32,
	                                        'OpenProcessToken',
	                                        'NNP', 'I' ) || die;
	$LookupPrivilegeValue  = new Win32::API($advapi32,
	                                        'LookupPrivilegeValue',
	                                        'PPP', 'I') || die;
	$AdjustTokenPrivileges = new Win32::API($advapi32,
	                                        'AdjustTokenPrivileges',
	                                        'NIPNPP', 'I') || die;
	$CreateFile = new Win32::API($kernel32,
	                             'CreateFileA',
	                             'PNNPNNN', 'N') || die;
	$ReadDirectoryChanges = new Win32::API($kernel32,
	                                       'ReadDirectoryChangesW',
	                                       'NPNINPPP', 'I') || die;
	$CreateIoCompletionPort = new Win32::API($kernel32,
	                                         'CreateIoCompletionPort',
						 'NNPN', 'N') || die;
	$GetQueuedCompletionStatus = new Win32::API($kernel32,
	                                            'GetQueuedCompletionStatus',
						    'NPPPN', 'I') || die;
}

sub _EnablePrivileges {
	my $phToken = pack("L", 0);

	if($OpenProcessToken->Call($GetCurrentProcess->Call(),
	                           TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY,
	                           $phToken)) {
		my $hToken = unpack("L", $phToken);

		_SetPrivilege($hToken, SE_BACKUP_NAME, 1);
		_SetPrivilege($hToken, SE_RESTORE_NAME, 1);
		_SetPrivilege($hToken, SE_CHANGE_NOTIFY_NAME, 1);

		$CloseHandle->Call($hToken);
	}
}

sub _SetPrivilege {
	my ($hToken, $pszPriv, $bSetFlag) = @_;

	my $iResult;
	my $pLuid = pack("Ll", 0, 0);

	if($LookupPrivilegeValue->Call("\x00\x00", $pszPriv, $pLuid))
	{
		my $pPrivStruct = pack("LLlL",
		                       1,
		                       unpack("Ll", $pLuid),
		                       (($bSetFlag)? SE_PRIVILEGE_ENABLED : 0));

		$iResult = (0 != $AdjustTokenPrivileges->Call($hToken,
		                                              0,
		                                              $pPrivStruct,
		                                              length($pPrivStruct),
		                                              0, 0));
	}
	return ($iResult);
}

# Methods
sub new {
	my $class = shift;
	$class = ref($class) || $class;
	my $self = {
		# the defaults
		path    => undef,
		subtree => 0,
		filter  => FILE_NOTIFY_CHANGE_NAME,
		# overridden
		@_,
	};
	bless $self, $class;
	# Put a delay in here so that we aren't capturing the notifications from previous operations
	select(undef, undef, undef, 1);
	return $self if $self->initialize(@_);
	return undef;
}

sub initialize {
	my $self = shift;

	$CloseHandle->Call($self->{handle}) if (exists($self->{handle}) &&
	                                        $self->{handle} > 0);
	$self->{handle} = $CreateFile->Call($self->{path},
	                                    FILE_LIST_DIRECTORY,
	                                    FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
	                                    0,
	                                    OPEN_EXISTING,
	                                    FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OVERLAPPED,
	                                    0);

	$self->{port} = $CreateIoCompletionPort->Call($self->{handle},
	                                              0,
						      $self,
						      0);

	share(@{$self->{results}});

	$self->{listening} = 0;
	share($self->{listening});

	# Create a thread to capture directory changes
	$self->{thread} = threads->new("handle_changes", $self);

	while (!$self->{listening}) { select(undef, undef, undef, 0.25); };

	return ($self->{handle} > 0) ? 1 : undef;
}

sub read_changes {
	my $self = shift;

	return undef unless (exists($self->{handle}) && $self->{handle} > 0);

	my $nBufferLength = 64*1024;

	$self->{buffer} = "\x00" x $nBufferLength;

	my $pBytesReturned = pack("L", 0);
	my $pOverlapped = pack("PPNNN", 0, 0, 0, 0, 0);

	my $iResult = $ReadDirectoryChanges->Call($self->{handle},
						  $self->{buffer},
	                                          $nBufferLength,
	                                          $self->{subtree},
	                                          $self->{filter},
	                                          $pBytesReturned,
						  $pOverlapped,
						  0);
}

sub handle_changes {
	my $self = shift;

	my $pUnused = pack("L", 0);
	my $pBytesReturned = pack("L", 0);
	my $pOverlapped = pack("PPNNN", 0, 0, 0, 0, 0);

	while (1) {
		# Send the notification request, and then asynchronously wait for a response.
		&read_changes($self);

		$self->{listening} = 1;
		# Wait until either a notification is received or the completion port is closed.
		my $iReturn = $GetQueuedCompletionStatus->Call($self->{port},
			                                       $pBytesReturned,
							       $pUnused,
							       $pOverlapped,
							       0xFFFFFFFF); # Infinite wait

		if ($iReturn && $self->{listening}) {
			my $bytesReturned = unpack("L", $pBytesReturned);

			my $pBuffer = $self->{buffer};

			if ($bytesReturned > 0) {
				my ($NextEntryOffset, $Action, $FileNameLength, $FileName);

				while (1) {
					($NextEntryOffset, $Action, $FileNameLength) = unpack("LLL", $pBuffer);
					(undef, undef, undef, $FileName) = unpack("LLLa$FileNameLength", $pBuffer);
					$FileName = pack "C*", unpack "S*", $FileName;
					
					lock @{$self->{results}};
					push @{$self->{results}}, $Action, $FileName;
					
					last if($NextEntryOffset <= 0);
					$pBuffer = substr($pBuffer, $NextEntryOffset);
				}
			} else {
				# This can happen if there are many changes at once or if there's some other
				# weirdness that can't be defined. It's basically change notify's way of
				# telling the client *something* happened, but if you want to know what, go
				# figure it out on your own.
				lock @{$self->{results}};
				push @{$self->{results}}, 0, "* Unspecified *";
			}
		} else {
			last;
		}
	}
}

DESTROY {
	my $self = shift;
	$self->{listening} = 0;
	$CloseHandle->Call($self->{handle}) if (exists($self->{handle}) &&
	                                        $self->{handle} > 0);
	$CloseHandle->Call($self->{port}) if (exists($self->{port}) &&
	                                      $self->{port} > 0);
	select(undef, undef, undef, 0.25);
	return 1;
}

sub details {
	my $self = shift;
	my @result_list;

	# This sleep is in place so that ONTAP has enough time to post the notification message, as it
	# posts changes on a scheduled basis, not right away.
	select(undef, undef, undef, 1);
	
	lock @{$self->{results}};
	my @results = @{$self->{results}};
	@{$self->{results}} = ();

	while (@results) {
		push @result_list, [ ( splice @results, 0, 2 ) ];
	}

	return @result_list;
}

_InitializeAPI;
_EnablePrivileges;

1;
