#Hey emacs, this is basically -*- perl -*-
# $Id$
#
## @summary qdir - Library for the abe query directory tests
## @author	nsahu (CIFS Protocols QA)
## @status	public
## @notes	Created by Nilesh Sahu
## @pod here

=head1 NAME

ABE - Library for the query directory

=head1 SYNOPSIS

  use CIFS::qdir;

  qdir_setup(qtree_security => 'unix', windows_version => '5.1',
                       setup_h => \%setup_h);

  qdir_cleanup(cifs_drive => 'T:');
  

=head1 DESCRIPTION

Provides the common routines for the query directory tests.

=head1 STATUS

  Shared

=head1 AUTHOR

  nsahu (CIFS Protocols QA)
  $Id$
  Last revised $Date$

=head1 COPYRIGHT

  Copyright 2009, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package CIFS::qdir;

use strict;
use warnings;
use base qw(Exporter);

use Tharn qw(host connect logresult say logcomment disconnect script_result matchstrings);
use Carp qw(croak);
use FS::Util::Setup qw(create_vol common_cleanup);
use MapShare qw(map_cifs_share unmap_cifs_share);
use ExportFS qw(share_add share_delete);
use Users qw(addPasswdEntry addGroupEntry delPasswdEntry delGroupEntry);
use SPARSE::Util::Logger;
use Network::Utils qw(run_getXXbyYY);
use filer qw(filer_options_set filer_qtree_security_set filer_vol_snap_delete);
use Params::Validate qw(validate ARRAYREF HASHREF SCALAR UNDEF);
use POSIX qw(uname);
use IO::Seekable qw(SEEK_SET);
use File::Basename qw(basename);
use CIFS::Ntfs qw(cacls_get cacls_compare);
use CIFS::Setup qw(cifs_setup);
use CIFS::CommonParams qw(:all);
use CIFS::widelink qw(deleteFile);
use SayWrap qw(sayEOF);
use Admin::License qw(license);
use File::Find qw(find);
use TestEnv::MinWin qw(minwin_compatible execute_cmd);
use Utils qw(make_host make_conn);
use Volume qw(:all);
use mount qw(mount umount mounts);
#use Win32::FileSecurity qw(Get EnumerateRights);


BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our $VERSION = '1.0';

# Test file and directory names
#my $test_file_name = 'vss-clt-001.txt';
#my $test_dir_name = 'tdir';
#my $test_file_msg = 'this file contains the string vss-clt-001';

# Variables used by the setup and cleanup routine
my @filer_vol_path;
my @create_vol_called;
my @fc_users_created;
my @fc_groups_created;
my @cifs_share_completed;
my @cifs_map_completed;
my @qtree_security_completed;
my %test_files;

# For CIFS shares to be accessable the $CIFS_USER account
# must exist from the Unix side of the filer. We just add
# the $CIFS_USER account to /etc/passwd and the root
# group to /etc/group.
my $admin_group = 'root';

# Functions to be exported
our @EXPORT_OK = qw(
    qdir_setup
    qdir_cleanup
    verify_dir_attrs
    create_dataset
    change_attrs
    change_qtree_security
    verify_sorted_output
    verify_output
    verify_time
);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub qdir_setup(%);
sub qdir_cleanup(%);
sub verify_dir_attrs(%);
sub create_dataset(%);
sub change_attrs(%);
sub verify_sorted_output(%);
sub verify_output(%);
sub verify_time(%);
#---------------------------------------------------------------------

=head2 qdir_setup

qdir_setup(setup_h => \%setup_h
          [, qtree_security => $security ]
          [, windows_version => $win_version ]);

The following named arguments are accepted:

=over

=item Arguments

=over

=item setup_h

(Required) : Reference to setup hold hash for cleanup.

=item qtree_security

(Optional) : Security setting for a qtree. Defaults to unix.

=item windows_version

(Optional) : MS Windows version number

=back

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub qdir_setup(%) {
    my %opts = validate(
        @_,
        {
            setup_h         => { type => HASHREF },
            qtree_security  => { type => SCALAR, default => 'mixed',
                                 regex => qr/^(mixed|ntfs|unix)$/ },
            windows_version =>  { type => SCALAR, default => '6',
                                 regex => qr/^(6\.0)$/ }

        }
    );
    my ($cifs_drive, $exti, $fc, $host, $i, $op, @osrel, @triples, $volume);

    logStep("BEGIN: qdir_setup: setup the environment");

    $host = host('localhost');
    logStep("qdir_setup: Verify client has correct Windows version");
    @osrel = uname();
    if ($osrel[2] !~ /$opts{windows_version}/) {
        croak("client running $osrel[2] expected $opts{windows_version} " .
              "windows release");
    }

    # The default filer, vol, and share to use are set first. If additional
    # triples were passed in then process them in this loop also.
    @triples = ({ filer => $main::FILER, vol => $main::FILER_VOL,
                  share => $main::CIFS_SHARE });
    if (defined($opts{setup_h}->{ext_setup})) {
        push @triples, @{$opts{setup_h}->{ext_setup}};
    }
    $opts{setup_h}->{ext_setup} = \@triples;

    $i = 0;
    my $filer_conn = make_conn($main::FILER);
    say($filer_conn, "options cifs.smb2.client.enable on");
    say($filer_conn, "options cifs.smb2.enable on");

    foreach $exti (@triples) {
        logStep("qdir_setup: Create volume $exti->{vol}");
        $create_vol_called[$i] = 1;
        if ($i == 0) {
            $op = $main::EXTERN_SETUP;
        } else {
            $op = 0;
        }
        ($fc, $volume) = create_vol(filer => $exti->{filer},
                                    filer_vol => $exti->{vol},
                                    extern_setup => $op,
                                    guarantee => 'none') or croak($@);
        $exti->{fc} = $fc;
        $exti->{vol} = $volume;
        $filer_vol_path[$i] = "/vol/$exti->{vol}";
		
		logStep("qdir_setup: change qtree security to mixed for the testvol_flex ");
		say($fc, "cifs terminate -t 0");
		say($fc, "qtree security /vol/testvol_flex  ntfs");

		logStep("qdir_setup: Setup CIFS on the filer");
		if ($main::MODE eq "7G") {
			cifs_setup(host   => $exti->{filer}, 
			    'auth-type'   => 'ad',
		  		cifs_pass     => $main::CIFS_PASSWD,
				domain        => $main::DOMAIN,
				domain_user   => $main::CIFS_USER, 
   			   	domain_pass   => $main::CIFS_PASSWD );
		} else {
        	
        	cifs_setup(  host => $exti->{filer}, 
			   'server-name' => $exti->{filer}, 
			   'auth-type'   => 'ad',
  			   cifs_pass     => $main::CIFS_PASSWD,
			   domain        => $main::DOMAIN,
			   domain_user   => $main::CIFS_USER, 
   			   domain_pass   => $main::CIFS_PASSWD );
		}
	  sleep 50;

        share_add($fc, $exti->{share}, $filer_vol_path[$i], verify => 1,
                  purge_existing => 1, umask => 0, widelink => 1,
                  nosymlink_strict_security => 1)
            or croak("failed to export share $exti->{share}");
        $cifs_share_completed[$i] = 1;

        # Set the qtree security mode on the filer volume
        if (filer_qtree_security_set($fc, $filer_vol_path[$i],
            $opts{qtree_security}) == 0) {
            croak("QTREE set failed on $filer_vol_path[$i] for $opts{qtree_security}");
        }
        $qtree_security_completed[$i] = 1;

        # Put the default fc, vol, and cifs_drive into the setup_h also
        if ($i == 0) {
            $opts{setup_h}->{fc} = $fc;
            $opts{setup_h}->{vol} = $volume;
            $opts{setup_h}->{cifs_drive} = $cifs_drive;
        }

        $i++;
    }
	
    logStep("END: qdir_setup:");
    return;
}

#---------------------------------------------------------------------

=head2 abe_cleanup

abe_cleanup([dirs        => \@dirs]
            [,files      => \@files]
            [,cifs_drive => $cifs_drive]
            [,options    => \%options]
            [,snapshots  => \@snapshots]
            [,ext_setup  => \@ext_setup);

Cleanup after a test run if necessary. Use this cleanup
routine if you use abe_setup.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dirs

(Optional) : List of directories to cleanup on exit.

=item files

(Optional) : List of files to cleanup on exit.

=item cifs_drive

(Optional) : Drive letter where share is mapped to.

=item options

(Optional) : Hash of options and values to restore

=item snapshots

(Optional) : List of snapshots to cleanup on exit.

=item ext_setup

(Optional) : List of {cifs_drive}, {snapshots}, and {symfile} for
extended setup of {fc}, {vol}, and {share}.

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub qdir_cleanup(%) {
    my %opts = validate(
        @_,
        {
            dirs       => { type => ARRAYREF | UNDEF, optional => 1 },
            files      => { type => ARRAYREF | UNDEF, optional => 1 },
            cifs_drive => { type => SCALAR | UNDEF, optional => 1 },
            options    => { type => HASHREF | UNDEF, optional => 1 },
            snapshots  => { type => ARRAYREF | UNDEF, optional => 1 },
            vol        => { type => SCALAR | UNDEF, optional => 1 },
            fc         => { type => HASHREF | UNDEF, optional => 1 },
            ext_setup  => { type => ARRAYREF | UNDEF, optional => 1 },
			clean_homedir  => { type => SCALAR , default => 0 }
        }
    );
    my (@arr, $cleanup, $dir, $exti, $fc, $fh, $filer_vol, $i, $snapshot);
	my $homedir_vol = $main::FILER_VOL;
	
    logStep("CLEANUP == $main::CLEANUP");

    eval {


        $i = 0;
        foreach $exti (@{$opts{ext_setup}}) {
            logStep("qdir_cleanup: Reconnect filer $exti->{filer}");
            eval {
                $fh = host($exti->{filer});
                $fc = connect($main::FILER_CONN, -timeout, $main::TIMEOUT);
                $fc->timeout($main::CMDTIMEOUT);
            };
            if ($@) {
                undef $fc;
                logWarn("qdir_cleanup: $@");
            }

            if ($main::CLEANUP ne 'none') {
                # cleanup files only for the first filer on the list
                if ($i == 0 && $cifs_map_completed[$i]) {
                    if (defined($opts{files}) && @{$opts{files}} > 0) {
                        # The unlink would fail if somebody accidently stuck
                        # "/" as a file because it is a directory so there is
                        # no need to look for it on the list.
                        logStep("qdir_cleanup: removing files");
                        unlink @{$opts{files}} or
                            logDbg("qdir_cleanup: unlink failed: $!");
                        @{$opts{files}} = ();
                    }

                    if (defined($opts{dirs}) && @{$opts{dirs}} > 0) {
                        logStep("qdir_cleanup: removing directories");
                        # Traverse array in reverse order so that the directories
                        # are removed newest to oldest.
                        while ($dir = pop @{$opts{dirs}}) {
                            # Remove the directory
                            rmdir $dir or
                                logDbg("qdir_cleanup: rmdir $dir failed: $!");
                        }
                        @{$opts{dirs}} = ();
                    }
                }
				
				if ($opts{clean_homedir} == 1) {
					logStep("qdir_cleanup: turn off ABE on cifs.homedir share");
					turn_on_abe_on_share(filer => $main::FILER, share => 'cifs.homedir', abe_on_off => 0);
				}
				logStep("qdir_cleanup: removing files/dirs on the testshare");
				clean_share(client => $main::CLIENT, clean_homedir_share => $opts{clean_homedir});

                if ($cifs_map_completed[$i] && $exti->{cifs_drive}) {
                    logStep("qdir_cleanup: unmapping CIFS drive: $exti->{cifs_drive}");
                    eval {
                        unmap_cifs_share(mappath => $exti->{cifs_drive}, force => 1);
                    };
                    logWarn("qdir_cleanup: $@") if $@;
                }

                if ($fc && $cifs_share_completed[$i]) {
					if ($opts{clean_homedir} == 1) {
						$exti->{share} = $main::HOMEDIR_SHARE;
					} 
					logStep("qdir_cleanup: delete CIFS share: $exti->{share}");
					eval {
						share_delete($fc, $exti->{share});
					};
                    logWarn("qdir_cleanup: $@") if $@;
                }

                if ($fc && $fc_users_created[$i]) {
                    logStep("qdir_cleanup: Remove users added to filer");
                    delPasswdEntry($fc, $fc_users_created[$i]);
                    logWarn("qdir_cleanup: $@") if ($@);
                }

                if ($fc && $fc_groups_created[$i]) {
                    logStep("qdir_cleanup: Remove groups added to filer");
                        delGroupEntry($fc, $fc_groups_created[$i]) or
                            logWarn("qdir_cleanup: $@");
                }

                # restore options only for the first filer on the list
                @arr = keys(%{$opts{options}});
                if ($i == 0 && $fc && defined($opts{options}) && $#arr >= 0) {
                    logStep("qdir_cleanup: restoring options");
                    eval {
                        foreach $i (@arr) {
                           filer_options_set($fc, $i, ${$opts{options}}{$i});
                        }
                    };
                    logWarn("qdir_cleanup: $@") if $@;
                }
            }

            # Delete /etc/symlink.translations file and symlink file
            if ($exti->{symfile}) {
                eval {
                    deleteFile($fc->{hostp}->{id}, $fc->connhost(),
                               '/vol/$homedir_vol/etc/symlink.translations');
                };
                logWarn("qdir_cleanup: $@") if $@;
                eval {
                    deleteFile($fc->{hostp}->{id}, $fc->connhost(),
                               $exti->{symfile});
                };
                logWarn("qdir_cleanup: $@") if $@;
            }

            # Call common cleanup if create_vol was called
            if ($fc && $create_vol_called[$i]) {

                logStep("qdir_cleanup: Run common cleanup on vol $exti->{vol}");
                if ($i == 0 || $main::CLEANUP =~ /none|volume/) {
                    $cleanup = $main::CLEANUP;
                } else {
                    # If not the default volume then (test|all) map to volume
                    # for secondary volume cleanup.
                    $cleanup = 'volume';
                }
                # common_cleanup has purge of FILER_VOL hardcoded in so undef
                # it here
                $filer_vol = $main::FILER_VOL;
                undef $main::FILER_VOL;
                common_cleanup($fc, cleanup => $cleanup, vol => $exti->{vol});
                $main::FILER_VOL = $filer_vol;
                logWarn("qdir_cleanup: $@") if $@;
            }
            $i++;
        }
    };

    logWarn("qdir_cleanup: $@") if $@;
    logStep("END: qdir_cleanup:");
    return;
}

#---------------------------------------------------------------------

=head2 test_dir_and_file_create

%results = test_dir_and_file_create(target_drive => $tdrive
                                   [, huge       => $huge ]
                                   [, many       => $many ]);

%results = (
           dirs  => reference to array of dirs that were created,
           files => reference to array of files that were created,
           dir_tmstmp => timestamp on the directory,
           file_tmstmp => timestamp on the file
)

Create a test directory and file on the FILER volume.

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive for directory and file creation

=item test_dir_name

(Optional) : name of the directory to create

=item test_file_name

(Optional) : name of the file to create

=item test_message

(Optional) : message to write to the file

=item huge

(Optional) : whether a huge file should be created in the directory

=item many

(Optional) : whether many files and directories should be created in the
directory

=back

=item Returns

Results hash containing information about the test files and directories

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub test_dir_and_file_create(%) {
    my %opts = validate(
        @_,
        {
            target_drive => { type => SCALAR, regex => qr/^[A-Za-z]\:$/ },
			test_dir_name => { type => SCALAR, default => ""},
			test_file_name => { type => SCALAR, default => ""},
			test_message => { type => SCALAR, default => ""},
                        client => { type => SCALAR, default => ""},
                        timeflag => { type => SCALAR, default => '0'},
                        flag => { type => SCALAR, default => '0'},
        }
    );

    my (@dirs, @files, @stat_arr, $dir_tmstmp, $file_tmstmp, $i, $j, $tpath, $rsh_conn, $out1, $out2, $out3);

    # don't catch error on mkdir because directory may already exist
    logStep("test_dir_and_file_create: Creating directory $opts{target_drive}/$opts{test_dir_name}");
    mkdir("$opts{target_drive}/$opts{test_dir_name}");
    sleep 5;
    if ($opts{client} ne "") {
        logStep("NOW HERE INSIDE test_dir_and_file_create");
        $rsh_conn = make_conn($main::FILER);
    }

        my $op;
        my $h;
        my @temp;
	#my file_name = $opts{test_file_name}
	if ($opts{test_file_name} ne "") {
                if ($opts{timeflag} == 1) {
                    ($out1, $out2, $out3) = say($rsh_conn, "date");
		    open(OUTFILE, ">$opts{target_drive}/$opts{test_dir_name}/$opts{test_file_name}")
			or croak("Couldn't create file: $!");
                    if ($out3 =~ /([0-9]+:[0-9]+):[0-9]+/) {
                        print "NOW CHECKING $&";
                        print $&;
                        @temp = split(":", $&);
                        print $temp[0];
                        print $temp[1];
                        print $temp[2];
                      if (($temp[0]) > 12) {
                         $h = (($temp[0]) - 12);
                         $op = "0".$h.":".$temp[1]." "."PM";
                      } else {
                          $op = $temp[0].":".$temp[1]." "."AM";
                      }
                   }
                } else {
                    open(OUTFILE, ">$opts{target_drive}/$opts{test_dir_name}/$opts{test_file_name}")
                        or croak("Couldn't create file: $!");
                }

                if ($opts{flag} == 1) {
                   ($out1, $out2, $out3) = say($rsh_conn, "date");
                   print OUTFILE "$opts{test_message}\n"
                        or croak("Couldn't write to file: $!");
                   if ($out3 =~ /([0-9]+:[0-9]+):[0-9]+/) {
                      print "NOW CHECKING $&";
                      print $&;
                      @temp = split(":", $&);
                      print $temp[0];
                      print $temp[1];
                      print $temp[2];
                      if (($temp[0]) > 12) {
                         $h = (($temp[0]) - 12);
                         $op = "0".$h.":".$temp[1]." "."PM";
                      } else {
                         $op = $temp[0].":".$temp[1]." "."AM";
                      } 
                   }
                } else {
                   print OUTFILE "$opts{test_message}\n"
			or croak("Couldn't write to file: $!");
                }

		if ($opts{huge}) {
			seek(OUTFILE, 10000000000, SEEK_SET);
			print OUTFILE "end of file.\n";
		}

    	close(OUTFILE);
	}

    #print "NOW OUTSIDE LOOP";
    #print $op;
    if (($opts{flag} == 1) || ($opts{timeflag} == 1)) {
        return $op;
    } else {
        return;
    }
}

#---------------------------------------------------------------------

=head2 read_dir_cmd

read_dir_cmd(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub verify_dir_attrs(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			share => { type => SCALAR, default => $main::CIFS_SHARE},
                        path => { type => SCALAR },
			cifs_user => { type => SCALAR },
			cifs_passwd => { type => SCALAR },
			file => { type => SCALAR , default => undef},
                        attr => { type => SCALAR },
                        flag => { type => SCALAR},
        }
    );
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype, $i, $j);

	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{share};
        $i = 0;
        $j = 0;
	my $fh = host($main::FILER);
	my $filer_ip = $fh->default_ip();

	logStep("verify_dir_attrs: mapping CIFS drive");
	$cifs_drive = map_cifs_share(
		server   => $filer_ip,
		share    => $share,
		client   => 'localhost',
		drive    => '*',
		username => $opts{cifs_user},
		password => $opts{cifs_passwd}
		)
		or croak($@);
	logDbg("mapped drive $cifs_drive");
	logStep("verify_dir_attrs: successfully mapped filer $filer_ip\\$share to CIFS drive $cifs_drive");
	if (!$cifs_drive) {
		croak("could not map the drive");
	}
 
        
        if (defined ($opts{file})) {
	    foreach my $file (split(" ", $opts{file})){
            $i++;
     	    my ($out1, $out2, $out3) = say($rsh_conn, "dir /a$opts{attr} $cifs_drive\\$opts{path}\\$file");
		if ($out3 =~ /File Not Found/) {
                        if ($opts{flag} == 1) {
			    logStep("verify_dir_attrs: File $cifs_drive\\$opts{path}\\$file was not found");
                        } else {
                            logStep("verify_dir_attrs: File $cifs_drive\\$opts{path}\\$file was not found");
                            $j++;
                        }
		} else {
                        if ($opts{flag} == 1) {
			    logStep("verify_dir_attrs: File $cifs_drive\\$opts{path} was found");
                            $j++;
                        } else {
                            logStep("verify_dir_attrs: File $cifs_drive\\$opts{path} was found");
                        }
		}	
	    }
            if ($opts{flag} == 0) {
                if ($i == $j) {
                    $result = 1;
                } else {
                    $result = 0;
                }
            } elsif ($opts{flag} == 1) {
                if ($i == $j) {
                    $result = 1;
                } else {
                    $result = 0;
                }            
            }
        } else {
            my ($out1, $out2, $out3) = say($rsh_conn, "dir /a$opts{attr} $cifs_drive\\$opts{path}");	
            if ($out3 =~ /File Not Found/) {
                logStep("verify_dir_attrs: File $cifs_drive\\$opts{path} was not found");
                $result = 0;
            } else {
                logStep("verify_dir_attrs: File $cifs_drive\\$opts{path} was found");
                $result = 1;
            }
        } 
	logStep("verify_dir_attrs: unmapping CIFS drive $cifs_drive");
	eval {
		unmap_cifs_share(mappath => $cifs_drive, force => 1);
	};
	logWarn("read_dir_cmd: $@") if $@;
   
    return $result;
}
#---------------------------------------------------------------------

=head2 clean_share

clean_share(target_drive    => $tdrive,
			client => $CLIENT,
			);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub clean_share(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			clean_homedir_share => { type => SCALAR }
        }
    );
	
	my $home_dir = $main::HOMEDIR;

	my ($rsh_conn, @dir, $line, $directory, $drive_letter, $share, $username, $password, $result);
	$rsh_conn = make_conn($opts{client});
	
	
	if ($opts{clean_homedir_share} == 1) {
		$share    = $home_dir;			
	} else {
		$share    = $main::CIFS_SHARE;
		
	}
	my $fh = host($main::FILER);
	my $filer_ip = $fh->default_ip();
	my $filer_conn = make_conn($main::FILER);
	
	$username = $main::CIFS_USER;
	$password = $main::CIFS_PASSWD;
	
	$drive_letter = map_cifs_share(
		server   => $filer_ip,
		share    => $share,
		client   => 'localhost',
		drive    => '*',
		username => $username,
		password => $password
		)
		or croak($@);
	logDbg("clean_share: mapped drive $drive_letter");
	logStep("clean_share: successfully mapped share $main::FILER\\$share to CIFS drive $drive_letter");
	if (!$drive_letter) {
		croak("could not map the share");
	}
	
	my $drive_letter_saved = $drive_letter;
	
	if ($opts{clean_homedir_share} == 1) {
		$drive_letter    = join ("", $drive_letter, '\\', $main::HOMEDIR_SHARE);			
		say($filer_conn, "priv set diag; rm /etc/cifs_homedir.cfg");
		say($filer_conn, "cifs homedir load");
	}

	logStep("clean_share: Deleting files and directories on drive $drive_letter");
	my ($out1, $out2, $out3) = say($rsh_conn, "del $drive_letter\\* /F /Q");
	
		
	($out1, $out2, $out3) = say($rsh_conn, "dir $drive_letter");
	@dir = split("\r", $out3);
	foreach $line (@dir) {
		if ($line =~ /<DIR>/){
			my @tmp_line = split (" ", $line);
			$directory = $tmp_line[4];
			($out1, $out2, $out3) = say($rsh_conn, "rmdir /S /Q $drive_letter\\$directory");
			if ($out3 =~ /The directory is not empty/) {
				change_perms(client => $main::CLIENT, path => $directory, user => $main::TEST_CIFS_USER, perm => 'F', drive_letter => $drive_letter);
				logStep("clean_share: Second try at deleting directory $drive_letter\\$directory");
				($out1, $out2, $out3) = say($rsh_conn, "rmdir /S /Q $drive_letter\\$directory");
				if ($out3 =~ /The directory is not empty/) {
					logStep("clean_share: Unable to delete directory $drive_letter\\$directory");
				}
			}
		}
	}

	($out1, $out2, $out3) = say($rsh_conn, "dir $drive_letter");
	if ($out3 =~ /File Not Found/) {
		$result = 1;
	} else {
		$result = 0;
	}
	
	logStep("clean_share: unmapping CIFS share $drive_letter_saved");
	eval {
		unmap_cifs_share(mappath => $drive_letter_saved, force => 1);
	};
	logWarn("clean_share: $@") if $@; 

   
    return $result;
}
#--------------------------------------------------------------------
=head2 create_dataset

clean_share(target_drive    => $tdrive,
			client => $CLIENT,
			dir_list => $dir_list,
			file_list => $file_list
			);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub create_dataset(%) {
    my %opts = validate(
        @_,
        {
			dir_list => { type => SCALAR },
			file_list => { type => SCALAR, default => undef},
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::CIFS_PASSWD},
                        sizeflag => { type => SCALAR, default => '0'},
                        dateflag => { type => SCALAR, default => '0'},
                        timeflag => { type => SCALAR, default => '0'},
                        client => { type => SCALAR, default => "" }
        }
    );
	
	#my $cifs_drive;

	my $share = $opts{cifs_share};
	my ($filer, $hosttype) = split(";", $main::FILER);
	
	logStep("create_dataset: mapping CIFS drive");
	
	my $fh = host($filer);
	my $filer_ip = $fh->default_ip();
		
	
	my $cifs_drive = map_cifs_share( 
		server   => $filer_ip,
		share    => $share,
		client   => 'localhost',
		drive    => '*',
		username => $opts{cifs_user},
		password => $opts{cifs_passwd}
		)
		or croak($@);
	logDbg("mapped drive $cifs_drive");
	logStep("create_dataset: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
	if (!$cifs_drive) {
		croak("could not map the drive");
	}	
	
	logStep("create_dataset: create directories");
        my @timearray;
	
	foreach my $dir (split(" ", $opts{dir_list})) {
		test_dir_and_file_create(target_drive => $cifs_drive, test_dir_name => $dir);
                if (defined ($opts{file_list})) { 
		    foreach my $file (split(" ", $opts{file_list})) {
                        if ($opts{sizeflag} == 1) {
                            if ($opts{timeflag} == 1) {
                                push(@timearray, test_dir_and_file_create(target_drive => $cifs_drive, test_dir_name => $dir, test_file_name => $file, test_message => "This is $file\.", client => $opts{client}, flag => '1'));
                            } else { 
			        test_dir_and_file_create(target_drive => $cifs_drive, test_dir_name => $dir, test_file_name => $file, test_message => "This is $file\.", flag => '0');
                            }
                        } else {
                            if ($opts{dateflag} == 1) {
                               sleep 60;
                            } 
                            if ($opts{timeflag} == 1) {
                               sleep 60;
                               push(@timearray, test_dir_and_file_create(target_drive => $cifs_drive, test_dir_name => $dir, test_file_name => $file, client => $opts{client}, timeflag => '1'));
                            } else {
                               test_dir_and_file_create(target_drive => $cifs_drive, test_dir_name => $dir, test_file_name => $file);
                            }
                        } 
		    }
                }
	}
	
	logStep("create_dataset: unmapping CIFS drive $cifs_drive");
	eval {
		unmap_cifs_share(mappath => $cifs_drive, force => 1);
	};
	logWarn("create_dataset: $@") if $@;
	
   
    if ($opts{timeflag} == 1) {
       return @timearray;
    } else {
       return;
    }
}
#--------------------------------------------------------------------

=head2 change_perms

read_dir_cmd(target_drive    => $tdrive,
                target_directory    => $tdir,
				[target_file    => $tfile],
				user		=> $user,);

Check previous version tab for test file and directory in FILER vol

The following named arguments are accepted:

=over

=item Arguments

=over

=item target_drive

(Required) : target drive on which the file or directory lives

=item target_dir

(Required) : directory on which permissions need to be changed

=item target_file

(Optional) : file on which permissions need to be changed

=back

=back

=for html <hr />

=cut

#--------------------------------------------------------------------

sub change_attrs(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			path => { type => SCALAR },
			user => { type => SCALAR },
                        file_list => { type => SCALAR , default => undef},
			attr => { type => SCALAR},
			cifs_share => { type => SCALAR, default => $main::CIFS_SHARE},
			cifs_user => { type => SCALAR, default => $main::CIFS_USER},
			cifs_passwd => { type => SCALAR, default => $main::CIFS_PASSWD},
			drive_letter => { type => SCALAR, default => ""},
                        flag => { type => SCALAR, default => ""},
        }
    );
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype);

	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{cifs_share};
	($filer, $hosttype) = split(";", $main::FILER);

	my $fh = host($filer);
	my $filer_ip = $fh->default_ip();

	if($opts{drive_letter} eq "") {
		logStep("change_perms: mapping CIFS drive");
		$cifs_drive = map_cifs_share(
			server   => $filer_ip,,
			share    => $share,
			client   => 'localhost',
			drive    => '*',
			username => $opts{cifs_user},
			password => $opts{cifs_passwd}
			)
			or croak($@);
		logDbg("change_perms: mapped drive $cifs_drive");
		logStep("change_perms: successfully mapped filer $filer\\$share to CIFS drive $cifs_drive");
		if (!$cifs_drive) {
			croak("could not map the drive");
		}
	} else {
		$cifs_drive = $opts{drive_letter};
		logStep("change_perms: CIFS drive is already mapped to $cifs_drive");
	}
	
	logStep("change_attrs: Changing attributes on the files $opts{file_list} to $opts{attr} for the user $opts{user}");
		
	my ($out1, $out2, $out3) = say($rsh_conn, "dir $cifs_drive");
        ($out1, $out2, $out3) = say($rsh_conn, "attrib +$opts{attr} $cifs_drive\\$opts{path}");
        if (defined ($opts{file_list})) {
            foreach my $file (split(" ", $opts{file_list})) {
                if ($opts{flag} == 1) {
	            ($out1, $out2, $out3) = say($rsh_conn, "attrib +$opts{attr} $cifs_drive\\$opts{path}\\$file");
                } else {
                    ($out1, $out2, $out3) = say($rsh_conn, "attrib -$opts{attr} $cifs_drive\\$opts{path}\\$file");
                }
            }	
        } else {
            ($out1, $out2, $out3) = say($rsh_conn, "attrib +$opts{attr} $cifs_drive\\$opts{path}");
        }
	sleep 5;
	
	if($opts{drive_letter} eq "") {
		logStep("change_perms: unmapping CIFS drive $cifs_drive");
		say($rsh_conn, "net use $cifs_drive /delete /y");
		logWarn("change_perms: $@") if $@;  
	}
    
    return $result;
}

sub verify_sorted_output(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			share => { type => SCALAR, default => $main::CIFS_SHARE},
                        path => { type => SCALAR },
			cifs_user => { type => SCALAR },
			cifs_passwd => { type => SCALAR },
			file => { type => SCALAR },
                        attr => { type => SCALAR },
        }
    );
	
	my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype, $i, $j);

	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{share};
        $i = 0;
        $j = 0;
	my $fh = host($main::FILER);
	my $filer_ip = $fh->default_ip();

	logStep("verify_sorted_output: mapping CIFS drive");
	$cifs_drive = map_cifs_share(
		server   => $filer_ip,
		share    => $share,
		client   => 'localhost',
		drive    => '*',
		username => $opts{cifs_user},
		password => $opts{cifs_passwd}
		)
		or croak($@);
	logDbg("mapped drive $cifs_drive");
	logStep("verify_sorted_output: successfully mapped filer $filer_ip\\$share to CIFS drive $cifs_drive");
	if (!$cifs_drive) {
		croak("could not map the drive");
	}
 
        
     	my ($out1, $out2, $out3) = say($rsh_conn, "dir /o$opts{attr} $cifs_drive\\$opts{path}");
        print $out3;
        my $res = verify_output($out3,$opts{file}); 
        if ($res == 0) {
           logStep("verify_sorted_output: Output of diretory command is not sorted as expected\n");
           $result = 0;
        } else {
           logStep("verify_sorted_output: Output of diretory command is sorted as expected\n");
           $result = 1;
        }
	logStep("verify_sorted_output: unmapping CIFS drive $cifs_drive");
	eval {
		unmap_cifs_share(mappath => $cifs_drive, force => 1);
	};
	logWarn("verify_sorted_output: $@") if $@;
   
    return $result;
}

sub verify_output {
    my($op, $list) = @_;
    my @op1 = split(/\n/, $op);
    my @con = split(/ /, $list);
    my $con,
    my $op1,
    my $op2,
    my $i = 0;
    my $j = 0;
    my $k = 0;
    foreach $con (@con) {
        $k++;
    }
    foreach $op1 (@op1) {
        if ($op1 =~ /[0-3][0-9]\/[0-3][0-9]/ && $op1 !~ /DIR/) {
            my @op2 = split(/\s+/, $op1);
            if ($op2[4] eq $con[$i]) {
                $i++;
                $j++;
            }
        }
    }
    if ($j == $k) {
        return 1
    } else {
        return 0
    }
}

sub verify_time(%) {
    my %opts = validate(
        @_,
        {
			client => { type => SCALAR },
			share => { type => SCALAR, default => $main::CIFS_SHARE},
                        path => { type => SCALAR },
			cifs_user => { type => SCALAR },
			cifs_passwd => { type => SCALAR },
			file_list => { type => SCALAR },
                        attr => { type => SCALAR },
                        time_list => { type => ARRAYREF },
        }
    );
    
    my ($result, $exti, $cifs_drive, $test_file_name, $test_dir_name, $rsh_conn, $filer, $share, $hosttype, @filelist, @splitop, $i, $j, $k, $out1, $out2, $out3, $count, @time_list);

        logStep("NOW INSIDE VERIFY_TIME");
        my $var = $opts{time_list};
        @time_list = @$var;

        foreach my $temp1 (@time_list){
           print $temp1;
        }

	$rsh_conn = make_conn($opts{client});
    
	$share = $opts{share};
	my $fh = host($main::FILER);
	my $filer_ip = $fh->default_ip();

	logStep("verify_time: mapping CIFS drive");
	$cifs_drive = map_cifs_share(
		server   => $filer_ip,
		share    => $share,
		client   => 'localhost',
		drive    => '*',
		username => $opts{cifs_user},
		password => $opts{cifs_passwd}
		)
		or croak($@);
	logDbg("mapped drive $cifs_drive");
	logStep("verify_time: successfully mapped filer $filer_ip\\$share to CIFS drive $cifs_drive");
	if (!$cifs_drive) {
		croak("could not map the drive");
	}
 
        $i=0;
        $j=0;
        $k=0;
        $count=0;
     	($out1, $out2, $out3) = say($rsh_conn, "dir /t$opts{attr} $cifs_drive\\$opts{path}");
        print $out3;
        @filelist = split(" ", $opts{file_list});

        foreach my $temp (@filelist) {
           $count++;
        }

        @splitop = split("\r", $out3);
        foreach my $op1 (@splitop) {
            print "NOW INSIDE MAIN VERIFY LOOP";
            print $count;
            print $op1;
            if ($op1 =~ /[0-2][0-9]\:[0-5][0-9]/ && $op1 !~ /DIR/) {
               print "NOW INSIDE SECOND LOOP";       
               print $filelist[$i];
               if ($op1 =~ /$filelist[$i]/) {
                  print "NOW INSIDE THIRD LOOP";
                  $i++;
                  print $time_list[$j];
                  if ($op1 =~ /$time_list[$j]/) {
                     print "NOW INSIDE FOURTH LOOP";
                     logStep($time_list[$j]);
                     logStep($op1);
                     $k++;
                     $j++;
                 }
              }
           }
        } 
        if ($k == $count) {
           logStep("verify_time: Output of diretory command is  as expected\n");
           $result = 1;
        } else {
           logStep("verify_time: Output of diretory command is not as expected\n: Please confirm that filer and windows client are in same timezone, if not then put them in same timezone");
           $result  = 0;
        }
	logStep("verify_time: unmapping CIFS drive $cifs_drive");
	eval {
		unmap_cifs_share(mappath => $cifs_drive, force => 1);
	};
	logWarn("verify_time: $@") if $@;
   
    return $result;
}


#---------------------------------------------------------------------
