#  $Id: //depot/prod/test/Rsierranevada/lib/CLI/FileIO.pm#1 $

package CLI::FileIO;

## @summary Functions to read/write files from the appliance command line
##
## @description
##   This library provides wrappers for "rdfile" and "wrfile"
##   commands, which allow reading and writing of simple text files
##   over rsh.
##
##   One might use these methods instead of a real file protocol (such
##   as NFS or CIFS) in order to modify /etc files (which may only be
##   written by root or administrator) without having to arrange for
##   test code to run as that user.
##
##   "simple text files" means no control characters other than
##   newlines, lines no more than 200 characters long, etc.
##
## @example
##    # remove "DNS" settings from /etc/rc on a filer
##
##    use CLI::FileIO qw(rdfile wrfile);
##
##    param('FILER','-mesg',"hostspec for a filer");
##    my $h = host($FILER);
##    my $c = connect("#rsh");
##
##    my $data = rdfile($c,"/etc/rc")
##        or logresult("FATAL","Could not read /etc/rc: $@");
##    $data =~ s/options dns.*\n//;
##    wrfile($c,"/etc/rc",$data)
##        or logresult("FATAL","Could not write /etc/rc: $@");
##
## @author dl-central-libs


# Placing exports in BEGIN block to prevent circular include problem.
BEGIN {
    require Exporter;
    @ISA = qw(Exporter);
    @EXPORT = qw(rdfile wrfile);
}

use Utils qw(make_conn);
use VFILER::Util qw(get_vfiler_root_path);


use strict;

use Tharn;

## @name ::rdfile
## @summary Read a file using hidden filer command 'rdfile'
## @description
##   read a file using filer's hidden "rdfile" command.
##
##   files are assumed to be text files, and carriage returns
##   preceeding newlines are dropped.
##
## @synopsis
##     rdfile($conn, $filename);
##
## @arg $conn Required: Filer connection to use.  (rsh connections,
## particuarly rsh-socket, should work reliably, telnet/console should
## work if debug printfs and syslog messages do not get in the way and
## if we can detect echoed back input properly)
##
## @arg $filename Required: Name of file to read (the argument to
## "rdfile")
##
## @returns
##    The data read.  On failure, returns undef, with the perl special
##    variable "$@" containing a relevant error message.
##
## @author lincoln@netapp.com

sub rdfile
{
    return eval {
	my ($conn, $filename) = @_;
	($conn,$filename) = check_vfiler($conn,$filename);
	my (undef, undef, $text) = say($conn, "rdfile $filename",'-eval_error');
	die $@ if $@;
	$text =~ s/\r+\n/\n/g;
	my $filename_re = quotemeta($filename);
	if ($conn->is_persay()) {
	    $text =~ s/^rdfile $filename_re\n//;
	}
	if ($text =~ /^$filename_re\: /) {
	    die "error from rdfile reading $filename.  rdfile says:\n" .
		"$text\n";
	}
	return $text;
    };
}


## @name ::wrfile
## @summary Write a file using hidden filer command 'wrfile'
## @description
##   Write a file using filer's hidden "wrfile" command with added options ( -private | -private -d | -private -i inum) 
##
## @synopsis
##     wrfile($conn, $filename, $data[,%opts]);
##
## @arg $conn Required: Filer connection to use.  (rsh connections,
## particuarly rsh-socket, should work reliably, telnet/console should
## work if debug printfs and syslog messages do not get in the way)
##
## @arg $filename Required: Name of file to write (the argument to "wrfile")
##
## @arg $data Required: Contents of the file to write (the input to "wrfile")
##
## @arg %opts Optional: options can be set:
##
##    private=>1 or 0, default is 0 
##             It will create a private file
##
##    private=>1 or 0 followed by pri_dir=>1 or 0 
##             It will create a private directory
##
##    private=>1 or 0 followed by -i <inum>
##             It will create a private file with the inode number specified 
##
##    
##
## @returns
##    A true value on success.  On failure, returns undef, with the
##    perl special variable "$@" containing a relevant error message.
##
## @example my $stat = wrfile($Conn, $path, private=>1, inum=><inode-no>)
## @example my $stat = wrfile($Conn, $path, private=>1)
## @example my $stat = wrfile($Conn, $path, private=>1, pri_dir=>1)
##
## @author lincoln@netapp.com , cbose@netapp.com

sub wrfile ( $$$;% ) {
    my ($conn,$filename,$data,%opts) = @_;
    

    return eval {
	($conn,$filename) = check_vfiler($conn,$filename);
	# invoke wrfile
         my $private  = ($opts{private}) ? "-private" : "";
         my $inum     = ($opts{inum}) ? "-i $opts{inum}" : "";
	 my $pri_dir  = ($opts{pri_dir}) ? "-d" : ""; 
       
	 if(defined($opts{private})){
                if($inum && !$pri_dir){
	         say($conn,"priv set -q test;wrfile $filename $private $inum",'-blind','-eval_error');
	   	} elsif($pri_dir && !$inum){
         	say($conn,"priv set -q test;wrfile $filename $private $pri_dir",'-blind','-eval_error');
                } elsif($pri_dir && $inum){
		die("Invalid Option. Correct Usage is :\n wrfile <filename> [...]\n | wrfile <filename> -private [...]\n | wrfile <filename> -private -d [...]\n | wrfile <filename> -private -i <inum> [...]");
		} else{
	     	say($conn,"priv set -q test;wrfile $filename $private",'-blind','-eval_error');
                }
         } else{
	   	say($conn,"wrfile $filename",'-blind','-eval_error');
	     }

	die $@ if $@;

	
	# send file contents as input to wrfile

	if ($conn->conntype =~ /rsh-socket/) {
	    $conn->send($data);
	} else {
	    # For non-rsh-socket connections, one cannot send input
	    # too quickly.  Also, non-rsh-socket connections echo back
	    # the input, which we should gather anyway, so we can
	    # match on the echo back of each line before we send the
	    # next line.
	    my (@lines) = split(/\n/,$data);
	    foreach my $line (@lines) {
		my $linere = "(?:" . Tharn::string2regex($line) . ")";
		$linere =~ s/\t/\\s+/g;
		my $endtime = Tharn::timeout2time($conn->timeout());
		$conn->send($line . "\n");
		if (!defined($conn->expect(match=>[$linere],timeout=>$endtime))) {
		    die("Error waiting for echo back of line \n".
			"  \"$line\"\n".
			"(regular expression:\n".
			"  \"$linere\"\n".
			"): $@");
		}
	    }
	}

    
	# Finish wrfile command and wait for following prompt

	if ($conn->conntype =~ /rsh/) {
	    # over rsh, wrfile input is finished in a normal way by
	    # closing input, and we know the command is finished if we
	    # see output close (EOF).
	    $conn->send_EOF();
	    $conn->expect();
	    if ($@ !~ /EOF/) {
		die "Error waiting for end of file after end of input to wrfile: $@";
	    }
	} else {
	    # over telnet or console there is no way to send end of
	    # input, so wrfile must be finished by killing the running
	    # wrfile command, which should result in a normal filer
	    # prompt (which we are reponsible for consuming)
            $conn->send_intr();
	    $conn->expect(match=>[$conn->prompts()])
		or die "Error waiting for prompt after end of input to wrfile: $@";
	}
	return 1;
    };
}

sub check_vfiler($$)
{
	my($conn,$filename) = @_;
	return eval {
	    my $hostrec=$conn->hostp();
	    my $pfiler = $hostrec->{PFILER};
	    if($pfiler) {
	    	my $hostid=$hostrec->id();
	    	$conn = make_conn($pfiler);
	    	# check if the filename is in rootpath
	    	# else use the given path
	    	if ($filename !~ m(^/vol/))  {
	    		my $rootpath = get_vfiler_root_path($conn,$hostid);
	    		$filename = $rootpath . $filename;
	    	}
	    }
	    return ($conn,$filename);
	};
}

1;
	   
