#
# $Id
#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Content Repository Class
## @author ngok@netapp.com
## @status shared
## @pod here

package CR;

use strict;
use warnings;

use Data::Dumper;
use Params::Validate qw(:all); # for SCALAR
use Tharn qw($Log);
use NACL::APISet; # for validate_params()
use NACL::C::Node;
use NACL::C::Vserver;
use NACL::C::Volume;
use NACL::C::Aggregate;
use NACL::C::Repository;
use CR::UniqueName;
use CR::Exceptions::ResidueVolumes qw(:try); # for "otherwise"
use CR::Setup;

=head1 NAME

ContentRepository - The class to encapsulate methods and data of Content Repository

=head1 SYNOPSIS

=begin text

 use CR;

 my $cr1 = CR->create(project => 'beacon');
 print $cr->vserver(), " is ready\n" if ($cr->is_ready());
 $cr1->delete();

 my $cr2 = CR->new();
 $cr2->delete();

=end text

=head1 AUTHOR

Sidney Ho (ngok@netapp.com)

=head1 DESCRIPTION

This is a NACL AppBuilder to support test program that uses Content Repository.
This Perl module contains only the constructors and destructor.
Extra functionalities are provided by Perl modules under the CR directory.

=head1 ROUTINES

=cut

############################################################
#     constructors
############################################################

=head2 new

Constructor to connect to a Vserver intended for Repository operations.
It requires the Vserver name
and a node name of the cluster for making the connection.

=over 4

=item *
project

The kind of project is intended for.  Default is 'beacon'.

=item *
vserver

The name of the Vserver of the Repository. (mandatory)

=item *
node

This is a cluster node name (mandatory).
This is usually the FILER of a HostRec which is part of the cluster
corresponding to the Vserver.
A connection will be opened using this node to issue ngsh commands to the node.

=back

Post-conditions:

=over 4

=item 1.
Instantiated CR object.

=item 2.
Instantiated object of NACL Node component as a field in the CR object.

=item 3.
Instantiated object of NACL Vserver component as a field in the CR object.

=item 4.
Root volume of Vserver is online.

=back

=cut

#
# new - instantiate object
#
# Fields defined in the object
#   1. node              - node name
#   2. node_obj          - object of 'node'
#   3. vserver           - vserver name
#   4. vserver_obj       - object of 'vserver'
#   5. repository_obj    - repository object of 'vserver'
#   6. ngsh              - APISet object for issuing the ngsh commands
# NOTE: Should have the same fields as in create() method except
#   7. rootvolume        - root volume of vserver (initiated on use)
#   8. rootvolume_obj    - object of 'rootvolume'(initiated on use)
#
sub new {
  $Log->enter();

  my ($proto, %args) = @_;
  my $class = ref($proto) || $proto;

  my $self = NACL::APISet::validate_params($class,
                                           params => \%args,
                                           spec => {
                                             node => { type => SCALAR, optional => 0 },
                                             vserver => { type => SCALAR, optional => 0 },
                                             interface => { type => SCALAR, optional => 0 },
                                             project => { type => SCALAR, optional => 1 },
                                            },
                                           );
  # project should be 'beacon' but is not being checked now

  $self->{node_obj} = NACL::C::Node->new(node => $args{node},
                                         hostrec => Tharn::host($args{node})
                                        );

  $self->{vserver_obj} = NACL::C::Vserver->new(command_interface => $self->{node_obj},
                                               vserver => $self->{vserver}
                                              );

#  $self->{repository_obj} = NACL::C::Repository->find(command_interface => $self->{node_obj},
#                                                      filter => {vserver => $args{vserver}}
#                                                     );

  $self->{ngsh} = $self->{node_obj}->apiset(category => "Node",
                                            interface => $args{interface},
                                            set => "CMode"
                                           );

  if (defined($args{connection})) {
    $self->{linux} = NACL::APISet->new(connection => $args{connection},
                                       category   => "Host",
                                       interface  => "CLI",
                                       set        => "Linux",
                                       version_info => { "cdefs_from_build" => 0 },
                                      );
  }

  bless $self, $class;

  # Make sure root volume is online
  $self->_get_rootvolume_obj()->online();

  $Log->exit();
  return $self;
}

=head2 create

Constructor to create an instance of Content Repository.
It needs a Vserver name to be associated with the Content Repository.
It returns a CR object
so that setup functions can be performed on it.

=over 4

=item *
project

The kind of project is intended for (optinoal).  Default is 'beacon'.

=item *
vserver

The name of the Vserver to contain the Repository (mandatory).
If the vserver already exists, it will be reused.

=item *
node

This is a cluster node name (mandatory).
This is usually the FILER of a HostRec which is part of the cluster
corresponding to the Vserver.
A connection will be opened using this node to issue ngsh commands to the node.

=item *
aggreggates

A comma-seperated list of aggregates to a contain the object volume(s) (optional).
Default is to use an existing aggregate with most available space.

=item *
size

The total size of all the object volume(s) (optional).

=item *
junction-path

This is the Junction Path to be used by client to mount the namespace volume (optional).
Default is '/NS'.

=item *
repository-role

For regular repository, use 'RW'.
For repository used for snapmirroring, use 'DP'.
This is optional and the default is 'RW'.

=back

Pre-conditions:

=over 4

=item 1. A functional cluster with nodes, disks and physical network set up.

=back

Post-conditions:

=over 4

=item 1.
Instantiated CR object.

=item 2.
Instantiated object of NACL Node component as a field in the CR object.

=item 3.
Instantiated object of NACL Vserver component as a field in the CR object.

=item 4.
Added Repository license.

=item 5.
1 or more non-root aggregate.
If aggregates are created, there will be 1 per node.

=item 6.
Created a CR Vserver if not already exists.

=item 7.
Root volume of Vserver is online

=item 8.
Created or cleared to an empty Repository.

=back

=cut

sub create {
  $Log->enter();
  my ($proto, %args) = @_;
  my $class = ref($proto) || $proto;

  my $self = NACL::APISet::validate_params($class,
                                           params => \%args,
                                           spec => {
                                             project => { type => SCALAR, optional => 1 },
                                             node => { type => SCALAR, optional => 0 },
                                             vserver => { type => SCALAR, optional => 0 },
                                             aggregates => { type => SCALAR, optional => 1 },
                                             size => { type => SCALAR, optional => 1 },
                                             'junction-path' => { type => SCALAR, optional => 1 },
                                            'repository-role' => { type => SCALAR, optional => 1 },
                                           },
                                          );
  # project should be 'beacon' but is not being checked now
  my $project = delete $args{project};
  my $node = delete $args{node};

  $self->{node_obj} =	NACL::C::Node->new(node => $node,
                                           hostrec => Tharn::host($node));

  $self->{ngsh} = $self->{node_obj}->apiset(category => "Node",
                                            interface => "CLI",
                                            set => "CMode");
  # Add Repository license
  # $ngsh->system_license_add('license-code' => 'BIUWGKPJYBFDAA');

  # Check if Vserver already exist
  $self->{vserver_obj} = NACL::C::Vserver->find(command_interface => $self->{node_obj},
                                             filter => { vserver => $args{vserver}},
                                             allow_empty => 1);
  bless $self, $class;

  if (defined($self->{vserver_obj})) {
    # Reuse Vserver
    $self->{repository_obj} =
    NACL::C::Repository->find(command_interface => $self->{node_obj},
                              filter => {vserver => $args{vserver} },
                              allow_empty => 1);
    if (defined($self->{repository_obj})) {
       # delete repository (but not vserver)
       $self->delete();
    } else {
      # not recorded in repository as CR Vserver
      $self->_delete_vserver(); # delete first
      $self->_create_vserver(); # recreate
    }

    # Check that no volumes remain that belong to this Vserver
    if (my @residue = $self->_residue_volumes()) {
      CR::Exceptions::ResidueVolumes->throw("Previous volumes ".join(",",@residue)." remained in repository for vserver ".$args{vserver});
    }

    # Make sure rootvolume is online (create if necessary)
    my $rootvolume_obj = $self->_get_rootvolume_obj(allow_empty => 1);
    if (defined($rootvolume_obj)) {
      $rootvolume_obj->online();
    } else {
       # workaround for partially deleted VServer
       $self->_create_rootvolume(); # recreate rootvolume for now
    };
  } else {
    # New Vserver
    $self->{vserver_obj} = $self->_create_vserver(instantiating => 1);
  }

  # Workaround: snapshot may hang CR volume deletion
  $self->{vserver_obj}->modify('snapshot-policy' => 'none');

  # Create repository
  my $role = delete $args{'repository-role'};
  if (defined($role)) {
    $self->{repository_obj} = NACL::C::Repository->create(command_interface => $self->{node_obj},
                                                          vserver => $args{vserver},
                                                          'repository-role' => $role);
  } else {
    $self->{repository_obj} = NACL::C::Repository->create(
    command_interface => $self->{node_obj},
    vserver => $args{vserver});
  }

  # default junction-path is /NS
  $args{'junction-path'} = '/NS' if !defined($args{'junction-path'});

  # make namespace volume name the same as junction path (w/o leading '/')
  $args{volume} = $args{'junction-path'};
  $args{volume} =~ s|^/||;

  # define aggregates
  if (!defined($args{aggregates})) {
    $args{aggregates} = $self->_get_aggregate(%args);
  }

  # Create repository volumes
  #TODO: should use NACL::C::RepositoryVolume->create(...) when it's ready
  $self->{ngsh}->repository_volume_create(%args);

  $Log->exit();
  return $self;
}

############################################################
#     internal supporting routines
############################################################

sub _create_vserver {
  my ($self, %args) = @_;
  my $aggr = $self->_get_aggregate(%args);

  # create vserver in the first aggregate
  # rootvolume will be created automatically
  $self->{rootvolume} = 'root_' . $self->{vserver};
  $self->{vserver_obj} = NACL::C::Vserver->create(command_interface => $self->{node_obj},
                                                  vserver => $self->{vserver},
                                                  rootvolume => $self->{rootvolume},
                                                  aggregate => $aggr,
                                                  'ns-switch' => [ 'file' ],
                                                  'rootvolume-security-style' => 'mixed'
                                                 );

  # make rootvolume mode 0777
  $self->{rootvolume_obj} = NACL::C::Volume->new(command_interface => $self->{node_obj},
                                                 volume => $self->{rootvolume},
                                                 vserver => $self->{vserver});
                                                 $self->{rootvolume_obj}->modify('unix-permissions' => '---rwxrwxrwx'
                                                );
    return $self->{vserver_obj};
  }

sub _create_rootvolume {
  my ($self, %args) = @_;
  my $aggr = $self->_get_aggregate(%args);

  if (!defined($self->{rootvolume})) {
    $self->{rootvolume} = 'root_' . $self->{vserver};
  }
  return $self->{rootvolume_obj} = NACL::C::Volume->create(command_interface => $self->{node_obj},
                                                           volume => $self->{rootvolume},
                                                           aggregate => $aggr,
                                                           vsroot => 'true',
                                                           'security-style' => 'mixed',
                                                           'unix-permissions' => '---rwxrwxrwx',
                                                           vserver => $self->{vserver});
}

use constant DELAY => 10; # 10 seconds
use constant RETRIES => 6; # totally 1 minute
sub _residue_volumes {
  my $self = shift;
  my @volumes;
  for my $i (1..RETRIES) {
    last unless (@volumes = NACL::C::Volume->find(
    command_interface => $self->{node_obj},
    filter => {vserver => $self->{vserver}}));
    sleep DELAY;
  }

  # get only the names
  @volumes = map { $_->state()->volume() } @volumes;

  if (!defined($self->{rootvolume})) {
    $self->{rootvolume} = $self->{vserver_obj}->state()->rootvolume();
  }

  # return only non-root volumes
  return grep { $_ ne $self->{rootvolume} } @volumes;
}

#
# _get_rootvolume_obj - get the root volume object of the vserver
#
# object fields set as side-effect:
# 1. rootvolume
# 2. rootvolume_obj
#
sub _get_rootvolume_obj {
  my ($self, %args) = @_;

  if (!defined($self->{rootvolume})) {
    $self->{rootvolume} = $self->{vserver_obj}->state()->rootvolume();
  }

  $args{allow_empty} = 0 if !defined($args{allow_empty});
  if (!defined($self->{rootvolume_obj})) {
    $self->{rootvolume_obj} = NACL::C::Volume->find(command_interface => $self->{node_obj},
                                                    filter => { volume => $self->{rootvolume},
                                                    vserver => $self->{vserver}},
                                                    allow_empty => $args{allow_empty});
  }
  return $self->{rootvolume_obj};
}

#
# _get_aggregate - get the first aggregate seen in node
#
use constant MIN_AGGR_SPACE => 1000000000; # min 1GB
sub _get_aggregate {
  my ($self, %args) = @_;

  if (defined($args{aggregates})) {
    my @aggr = split ',', $args{aggregates};
    return $aggr[0];
  }

  # find non-root aggr with an appropriate size
  my @aggrs = NACL::C::Aggregate->find(command_interface => $self->{node_obj});
  @aggrs = sort {$b->state()->availsize() <=> $a->state()->availsize()} @aggrs;
  return $aggrs[0]->aggregate() if $aggrs[0]->state()->availsize() > MIN_AGGR_SPACE;

  # create one
  my $aggr = unique_name('aggr');
  return NACL::C::Aggregate->create(command_interface => $self->{node_obj},
                                    aggregate => $aggr,
                                    diskcount => 3)->aggregate();
}

############################################################
#     setters and getters
############################################################

=head2 set_flag

Set an internal flag that changes the behavior of the Repository.

=over 4

=item *
new_dataports

During dataports creation, any existing dataports with the same IP address will be refreshed with the new specification.
If this flag is not set, setup_dataport() will preserve existing dataports with the same IP addresses.
Default is 0 (i.e. keeping existing dataports.)

=item *
lif_auto_revert

Indicate whether LIF created will have the auto-revert property or not.
Default is 0 (false).
Setting this flag to 1 (true) if you want all the data LIFs to
automatically revert back to their original node after recovery from
a LIF failover.

=back

=cut

sub set_flag {
  $Log->enter();
  my ($self, %args) =	@_;
  %args = NACL::APISet::validate_params($self,
                                        params => \%args,
                                        spec => {
                                          new_dataports => { type => SCALAR, optional => 1 },
                                          lif_auto_revert => { type => SCALAR, optional => 1 },
                                        },
                                       );
  $self->{$_} = $args{$_} foreach (keys(%args));
  $Log->debug("set_flag args: %args\n");
  $Log->exit();
}

=head2 vserver

Return the name of the vserver of this Content Repository object.

=cut

sub vserver {
  $Log->enter();
  my $self = shift;
  $Log->exit();
  return $self->{vserver};
}

=head2 is_ready

Return a boolean whether setup has finished and the Content Repository is ready to be used.

=cut

sub is_ready {
  $Log->enter();
  my $self = shift;
  my $response = $self->{repository_obj}->state()->status();
  $Log->exit();
  if ($response =~ /ready/) {
    return 1;
  } else {
    return 0;
  }
}

############################################################
#     deletions
############################################################

=head2 delete

Release all volumes used and delete the Repository.
Vserver may be deleted as well dependent on the boolean parameter passed in.

=over 4

=item *
delete_vserver

This is an boolean to specify whether both the Repository and the Vserver
are to be deleted at the same time.
If it is 0 (false) or is not passed as argument,
only the Repository will be deleted
and the Vserver will be kept around.

=back


=cut

use constant DEL_TIMEOUT => 300; # 5 min
sub delete {
  $Log->enter();
  my ($self, %args) =	@_;

  # $self->{repository_obj}->delete('method-timeout' => DEL_TIMEOUT);
  $self->{ngsh}->repository_delete(vserver => $self->{vserver},
                                   'connectrec-timeout' => DEL_TIMEOUT);
  if ($args{delete_vserver}) {
    $self->_delete_vserver();
  }
  $Log->exit();
}

sub _delete_vserver {
  my $self = shift;
  my $ngsh = $self->{ngsh};

  $self->_clear_nis_domain();
  try {
  $ngsh->network_interface_delete(vserver => $self->vserver,
                                  lif => '*');
  }
  otherwise {}; # ignore any exception

  try {
    $ngsh->network_routing_groups_route_delete(vserver => $self->vserver,
                                               'routing-group' => '*');
  }
  otherwise {}; # ignore any exception

  try {
    $ngsh->network_routing_groups_delete(vserver => $self->vserver,
                                         'routing-group' => '*');
  }
  otherwise {}; # ignore any exception
  try {
    $ngsh->vserver_export_policy_rule_delete(vserver => $self->vserver,
                                             policyname => '*');
  }
  otherwise {}; # ignore any exception

  my $rootvolume_obj = $self->_get_rootvolume_obj();
  $rootvolume_obj->offline();
  $rootvolume_obj->delete();

  $ngsh->vserver_delete(vserver => $self->{vserver});
}

# delete nis-domain if exist
sub _clear_nis_domain {
  my $self = shift;
  try {
    $self->{ngsh}->vserver_services_nis_domain_delete(vserver => $self->{vserver},
                                                      domain => '*');
  }
  otherwise {}; # ignore any exception
}

=head1 EXCEPTIONS

=over 4

=item *
CR::Exception

This is the base class of all the exceptions pertaining to Content Repository.

=item *
CR::Exceptions::ArgumentError

The argument is of the wrong ref type as expected.
The ref type causing the exception will be in the message.

=item *
CR::Exceptions::NoAggregate

There are 2 situations (with different messages) when this exception is thrown:

=over 4

=item 1.
An existing aggregate is required for an operation and it is not found.

=item 2.
Not enough disks (i.e. less than 3) in the node to create an aggregate which is requied by the operation.

=back


=item *
CR::Exceptions::NoVServer

This exception will be thrown if an operation expecting an existing Vserver but none is found.

=item *
CR::Exceptions::NoRootVolume

This exception will be thrown if it is detected that no root volume is associated with a CR Vserver.

=item *
CR::Exceptions::VolumeNotFound

If a volume is expected but not found,
this exception will be thrown and the volume name will be attached to the exception.

=item *
CR::Exceptions::ResidueVolumes

When recycling a Vserver with an existing name, the previous instance of the Vserver will be delete first.
However, if the deletion cannot be completed with all its volumes deleted within a certain time frame,
the creation will be aborted by throwing this exception.

=back

Note that there are other exceptions that may be thrown by the underlying modules used.
They are implementation dependent and cannot be all listed here.

=cut

1;
