#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#$Id$
## @summary ALM Connection Module
## @author aldrin@netapp.com
## @status shared
## @pod here

package CR::ALMConn;

use warnings;
use strict;

use NATE::Log qw(log_global);

use LWP;
use LWP::UserAgent;
use HTTP::Response;
use HTTP::Status qw(status_message is_success);
use MIME::Base64;
use XML::DOM;

my $LOG = NATE::Log->new();

$CR::ALMConn::PROD_URL = 'http://rtpalmas.rtp.netapp.com';
$CR::ALMConn::TEST_URL = 'http://rtpalmstageas.rtp.netapp.com';
$CR::ALMConn::REST = '/qcbin/rest';
%CR::ALMConn::RES = (
    domain          => sub { return "/domains/$_[0]"; },
    project         => sub { return "/projects/$_[0]"; },

    is_auth         => '/is-authenticated',
    auth            => '/qcbin/authentication-point/authenticate',
    session         => '/site-session',
    'de-auth'       => '/qcbin/authentication-point/logout',

    test_set        => '/test-sets',
    test_set_folder => '/test-set-folders',
    test_inst       => '/test-instances',
    test            => '/tests',
);

my $XML_PARSER = XML::DOM::Parser->new();
my $DEFAULT_PAGE_SIZE = 200;

################################################################################
#
#   Create a new ALMConn object. This object is used for authorizing,
#   de-authorizing, and querying ALM for information. The ALMConn object may be
#   used for accessing subroutines in this module, but also contains necessary
#   constants that are used in class-level subroutines (requires an ALMConn
#   object to have been initialized).
#
#   Input:
#       A reference to a hash containing named arguments. Currently recognized
#           arguments are as follows:
#           DOMAIN  -> The outermost 'namespace' or 'partition' in which ALM
#                      data is separated. Default Value: ONTAP
#           PROJECT -> The namespace or partition below DOMAIN. A DOMAIN
#                      in ALM may have multiple projects. Default Value: ONTAP
#           url     -> The server that the desired ALM instance is being hosted
#                      on. The default value is the development/staging ALM
#                      server ($CR::ALMConn::TEST_URL). A constant is provided
#                      for convenient access to the production ALM server
#                      ($CR::ALMConn::PROD_URL).
#                      Default Value: http://rtpalmstageas.rtp.netapp.com
#
#   Output:
#       A reference to an ALMConn object. This object may be used for accessing
#           subroutines in this module and is generally designed to make
#           interfacing with ALM more intuitive.
#
################################################################################
sub new {
    my ($class, $args) = @_;

    my $self = bless({}, $class);

    _init_self($self, $args);

    return $self;
}

################################################################################
#
#   Initialize variables used in other subroutines for accessing the desired
#   ALM server, the desired ALM DOMAIN and PROJECT namespaces, the desired
#   response format from the ALM server, and initializes the LWP object and
#   'cookie jar' (storage of cookies for this LWP object).
#
#   Input:
#       The ALMConn object whose variables needs to be initialized.
#       A reference to a hash of values to set object variables to. This hash
#           is explained in ALMConn->new().
#
#   Output:
#       The ALMConn object whose variables have now been initialized to usable
#           values.
#
################################################################################
sub _init_self {
    my ($self, $args) = @_;

    $self->{browser} = LWP::UserAgent->new();
    $self->{browser}->cookie_jar({});

    $self->{domain} = $CR::ALMConn::RES{domain}('ONTAP');
    if ($args->{DOMAIN}) {
        $self->{domain} = $CR::ALMConn::RES{domain}($args->{DOMAIN});
    }

    $self->{project} = $CR::ALMConn::RES{project}('ONTAP');
    if ($args->{PROJECT}) {
        $self->{project} = $CR::ALMConn::RES{project}($args->{PROJECT});
    }

    $self->{base_url} = $args->{url} ? $args->{url} : $CR::ALMConn::TEST_URL;
    $self->{proj_url} = $self->{base_url}.$CR::ALMConn::REST.
                        $self->{domain}.$self->{project};
    $self->{hdr} = {'Accept' => 'application/xml'};

    $LOG->comment("initialized ALMConn object attributes:");
    foreach my $self_key (keys %{$self}) {
        $LOG->comment("\t$self_key => $self->{$self_key}");
    }

    return;
}

################################################################################
#
#   Checks if the current ALMConn object is authorized with the ALM server.
#
#   Input:
#       An initialized ALMConn object.
#
#   Output:
#       The HTTP response from the ALM server to a HTTP post to the 'is_auth'
#           ALM REST resource.
#
################################################################################
sub _is_auth {
    my $self = $_[0];

    return $self->{browser}->post($self->{base_url}.$CR::ALMConn::RES{is_auth});
}

################################################################################
#
#   Authorizes the ALMConn object with the ALM server.
#
#   Input:
#       An initialized ALMConn object.
#       A reference to a hash of arguments. Currently recognized arguments are:
#           USER -> the username to use for authorization with the ALM server
#           PASS -> the password corresponding to the ALM username
#
#   Output:
#       The HTTP response from the ALM server to an HTTP post to the 'auth'
#           ALM REST resource.
#
################################################################################
sub authorize {
    my ($self, $args) = @_;

    if (!$args || !scalar keys %{$args}) {
        $LOG->comment("Checking if authorized with ALM");
        return _is_auth($self);
    }

    if (!$args->{USER} || !$args->{PASS}) {
        $LOG->warn("Received invalid user/password for ALM authentication");
        return HTTP::Response->new(400, 'Invalid username/password');
    }

    my $encoded_str = encode_base64("$args->{USER}:$args->{PASS}");
    my $form_data = [];

    $LOG->comment("authenticating...");

    return $self->{browser}->post($self->{base_url}.$CR::ALMConn::RES{auth},
                                  $form_data,
                                  'Authorization' => "Basic $encoded_str");
}

################################################################################
#
#   De-authorizes the ALMConn object with the ALM server.
#
#   Input:
#       An initialized ALMConn object.
#       A reference to a hash of arguments. This hash is currently unused.
#
#   Output:
#       1 if successful, 0 if false
#
################################################################################
sub de_authorize {
    my ($self, $args) = @_;

    my $form_data = [];
    $LOG->comment("deauthorizing with ALM");
    my $res = $self->{browser}->get($self->{base_url}.
                                    $CR::ALMConn::RES{'de-auth'}, $form_data);

    if (defined $res && response_is_good($res)) {
        $LOG->comment("deauthorized!");
        return 1;
    }

    $LOG->warn("failed to deauthorize with ALM");
    return 0;
}

################################################################################
#
#   Queries ALM for a particular 'test' entity. This is a subset of the
#   meta-data associated with a test-instance in a test-set in ALM.
#
#   Input:
#       An initialized ALMConn object.
#       A reference to a hash of arguments. Currently recognized arguments are:
#           test_set    -> A test-set entity retrieved from ALM. This entity is
#                          represented by a hash with property names as keys
#                          and property values as key-values.
#           hdr_options -> Any HTTP header options recognized and accepted by
#                          ALM. For example, the default header option,
#                          'Accept' => 'application/xml', requests ALM data
#                          in XML format. Another option is 'Accept' =>
#                          'application/ant'. This option requests ALM data in
#                          the ANT format. Other acceptable header options may
#                          be found in the ALM REST documentation:
#                          http://rtpalmas.rtp.netapp.com/qcbin/Help/
#                          doc_library/api_refs/REST/webframe.html
#
#   Output:
#       A reference to a list of ALM test entities. Each entity is represented
#           by a hash, mapping entity properties (e.g. ID) to property values.
#
################################################################################
sub test_get {
    my ($self, $args) = @_;

    my @tests = ();
    my $test_set_entity = $args->{test_set};
    my $entities = $self->test_instance_get({
        test_set_id => $test_set_entity->{id},
    });

    my $num_entities = 0;

    foreach my $entity (@{$entities}) {
        my $query_args = {
            hdr_options => $args->{hdr_options} || q{},
            res_type => 'test',
            query_attrs => "id[$entity->{'test-id'}]",
            query_fields => q{},
        };

        my $response = $self->rest_query($query_args);

        if ($response) {
            $LOG->comment("parsing response to test entity query");

            my $test_entities = $self->xml_entities_parse($response->content);

            foreach my $test_entity (@{$test_entities}) {
                push(@tests, $test_entity);
            }
        }

        $num_entities++;
    }

    $LOG->comment("Retrieved $num_entities test entities from ALM");

    return \@tests;
}

################################################################################
#
#   Queries ALM for a particular 'test-instance' entity. This is the primary
#   entity that ALM uses to represents tests within a test-set.
#
#   Input:
#       An initialized ALMConn object.
#       A reference to a hash of arguments. Currently recognized arguments are:
#           test_set_id -> The ID of a test-set entity in ALM. To query
#                          test-instances from ALM you must query for
#                          cycle-id = test-set ID where the test-set ID is the
#                          ID of the test-set containing the desired
#                          test-instances.
#           hdr_options -> Any HTTP header options recognized and accepted by
#                          ALM. For example, the default header option,
#                          'Accept' => 'application/xml', requests ALM data
#                          in XML format. Another option is 'Accept' =>
#                          'application/ant'. This option requests ALM data in
#                          the ANT format. Other acceptable header options may
#                          be found in the ALM REST documentation:
#                          http://rtpalmas.rtp.netapp.com/qcbin/Help/
#                          doc_library/api_refs/REST/webframe.html
#
#   Output:
#       A reference to a list of ALM test-instance entities. Each entity is
#           represented by a hash, mapping entity properties (e.g. ID) to
#           property values.
#
################################################################################
sub test_instance_get {
    my ($self, $args) = @_;
    my ($page_size, $need_bigger_page) = ($DEFAULT_PAGE_SIZE, 0);
    my ($count, $response, $entities) = (-1, undef, undef);

    if (!$args->{test_set_id}) {
        print({*STDERR} "Invalid test-set-id supplied\n");
        return;
    }

    my $query_args = {
        hdr_options => $args->{hdr_options} || q{},
        res_type => 'test_inst',
        query_attrs => "cycle-id[$args->{test_set_id}]",
        query_fields => 'test-id',
    };

    do {
        $query_args->{page_size} = $page_size;

        $response = $self->rest_query($query_args);

        if ($response){
            ($count, $entities) = $self->xml_entities_parse($response->content);

            if ($page_size < $count) {
                $LOG->comment("data page size ($page_size) smaller than ".
                              "retrieved number ($count) of test-instance ".
                              "entities");
                $page_size = $count;
                $need_bigger_page = 1;
            }
        }
        else {
            last;
        }

    } while($need_bigger_page);

    return $entities;
}

################################################################################
#
#   Queries ALM for a particular 'test-set' entity.
#
#   Input:
#       An initialized ALMConn object.
#       A reference to a hash of arguments. Currently recognized arguments are:
#           test_set_path -> The full path to the test set in ALM.
#           hdr_options   -> Any HTTP header options recognized and accepted by
#                            ALM. For example, the default header option,
#                            'Accept' => 'application/xml', requests ALM data
#                            in XML format. Another option is 'Accept' =>
#                            'application/ant'. This option requests ALM data in
#                            the ANT format. Other acceptable header options may
#                            be found in the ALM REST documentation:
#                            http://rtpalmas.rtp.netapp.com/qcbin/Help/
#                            doc_library/api_refs/REST/webframe.html
#
#   Output:
#       A reference to a list of ALM test-instance entities. Each entity is
#           represented by a hash, mapping entity properties (e.g. ID) to
#           property values.
#
################################################################################
sub test_set_get {
    my ($self, $args) = @_;
    my @test_sets = ();

    my $query_args = {
        hdr_options => $args->{hdr_options} || q{},
        res_type => 'test_set',
        query_fields => 'id',
    };

    if ($args->{test_set_path}) {
        $LOG->comment("Retrieving single test set...");
        $LOG->comment("Setting test_set_path appropriately...");

        my @test_set_folders = split(/\//, $args->{test_set_path});
        my $test_set_name = pop(@test_set_folders);

        $query_args->{query_attrs} = "name[$test_set_name];";
        $args->{test_set_folder_path} = join(q{/}, @test_set_folders);
    }

    $LOG->comment("getting test set folder...");
    my $test_set_folder = $self->test_set_folder_get($args);

    if (!$test_set_folder->{id}) {
        $LOG->warn("could not find '$args->{test_set_folder_path}'");
        return [];
    }

    $query_args->{query_attrs} .= "parent-id[$test_set_folder->{id}]";

    my $response = $self->rest_query($query_args);

    if ($response) {
        my $entities = $self->xml_entities_parse($response->content);

        foreach my $entity (@{$entities}) { push(@test_sets, $entity); }
    }

    return \@test_sets;
}

################################################################################
#
#   Queries ALM for a particular 'test-set-folder' entity. A test-set-folder
#   is any parent directory (regardless of distance) of the desired test-set.
#   For example, given a test-set of '/ONTAP/ONTAP/SierraNevada/sierranevada.0/
#   sierranevada.0/NOS/Searchlight/Regression/DR/Automated-Regression' there
#   are 8 test-set-folder entities above the test-set 'Automated-Regression'.
#   These test-set-folders are (in order): ROOT, SierraNevada, sierranevada.0,
#   sierranevada.0, NOS, Searchlight, Regression, and DR. This subroutine is
#   useful for recursively traversing these test-set-folders in order to query
#   test-set information from ALM.
#
#   Input:
#       An initialized ALMConn object.
#       A reference to a hash of arguments. Currently recognized arguments are:
#           test_set_path -> The full path to the test set in ALM.
#           hdr_options   -> Any HTTP header options recognized and accepted by
#                            ALM. For example, the default header option,
#                            'Accept' => 'application/xml', requests ALM data
#                            in XML format. Another option is 'Accept' =>
#                            'application/ant'. This option requests ALM data in
#                            the ANT format. Other acceptable header options may
#                            be found in the ALM REST documentation:
#                            http://rtpalmas.rtp.netapp.com/qcbin/Help/
#                            doc_library/api_refs/REST/webframe.html
#
#   Output:
#       A reference to a list of ALM test-instance entities. Each entity is
#           represented by a hash, mapping entity properties (e.g. ID) to
#           property values.
#
################################################################################
sub test_set_folder_get {
    my ($self, $args) = @_;

    if (!$args->{test_set_folder_path}) {
        print({*STDERR} "Invalid test_set_folder_path value supplied\n");
        return {};
    }

    my @test_set_folders = split(/\//, $args->{test_set_folder_path});
    my $query_args = {
        hdr_options => $args->{hdr_options},
        res_type => 'test_set_folder',
        query_fields => 'id',
    };

    my ($test_set_folder, $id) = ({}, 0);

    foreach my $folder (@test_set_folders) {
        $query_args->{query_attrs} = "parent-id[$id];name[$folder]";

        my $response = $self->rest_query($query_args);
        if ($response) {
            my $entities = $self->xml_entities_parse($response->content);

            if (scalar @{$entities} == 1) {
                $test_set_folder = $entities->[0];
                $id = $entities->[0]->{id};
            }
            elsif (scalar @{$entities} > 1) {
                $LOG->warn("found too many test-set-folder matches");
            }
            elsif (scalar @{$entities} < 1) {
                $LOG->warn("no test-set-folders found\n");
                last;
            }
        }
        else {
            last;
        }
    }

    return $test_set_folder;
}

################################################################################
#
#   Convenience subroutine for sending RESTful queries to the ALM REST
#   interface. This subroutine takes arguments and serves as a consistent method
#   of making HTTP GET requests.
#
#   Input:
#       An initialized ALMConn object. Some object properties are used:
#           proj_url    -> The URL and REST resource of the ALM project to be
#                          queried. This is then combined with key-value pairs
#                          in a reference to a hash of arguments (mentioned
#                          below) to construct the full URL and REST resource
#                          for the ALM entity (e.g. test-set, test-instance,
#                          etc.) to be queried. For more information on what
#                          key-value pairs can/should be included in the
#                          arguments to this subroutine, see 'build_url'.
#       A reference to a hash of arguments. Currently recognized arguments are:
#           hdr_options -> Any HTTP header options recognized and accepted by
#                          ALM. For example, the default header option,
#                          'Accept' => 'application/xml', requests ALM data
#                          in XML format. Another option is 'Accept' =>
#                          'application/ant'. This option requests ALM data in
#                          the ANT format. Other acceptable header options may
#                          be found in the ALM REST documentation:
#                          http://rtpalmas.rtp.netapp.com/qcbin/Help/
#                          doc_library/api_refs/REST/webframe.html
#
#   Output:
#       The HTTP Response from the ALM server if the response is a success,
#           undefined otherwise.
#
################################################################################
sub rest_query {
    my ($self, $args) = @_;

    my $url = build_url($self->{proj_url}, $args);
    my $form_data = [];
    my %hdr_options = $args->{hdr_options} ? %{$args->{hdr_options}} :
                                             %{$self->{hdr}};

    my $response = $self->{browser}->get($url, $form_data, %hdr_options);

    if (response_is_good($response)) {
        return $response;
    }
    else {
        $LOG->error("Received failure response from ALM query");
        $LOG->comment("queried url '$url' with options:");

        foreach my $hdr_key (keys %hdr_options) {
            $LOG->comment("\t$hdr_key => $hdr_options{$hdr_key}");
        }

        $LOG->comment("received response: ".($response->decoded_content));
    }
}

################################################################################
#
#   Convenience subroutine for parsing XML returned from the ALM server.
#   Currently implemented via DOM Parser instead of SAX Parser. For queries
#   with large responses, SAX should be more efficient and performant. DOM was
#   used because it was simpler to implement.
#
#   Input:
#       An initialized ALMConn object.
#       The XML data to be parsed.
#       Additional subroutine options. Currently only debug is a recognized
#           option and, if true, triggers extra debug print statements.
#
#   Output:
#       The total number of found entities in ALM for the the previous HTTP
#           query.
#       A reference to the list of entities found by ALM. Each item in the list
#           is a hash containing detail information for the entity.
#
################################################################################
sub xml_entities_parse {
    my ($self, $xml_data, $options) = @_;
    my @entities = ();

    my $xml_doc = $XML_PARSER->parse($xml_data);
    my ($entity_count, $xml_entities) = xml_entities_get($xml_doc, {});

    if ($options && $options->{debug}) {
        $LOG->comment('processing '.$xml_entities->getLength.' entities');
    }

    for my $entity_ndx (0..($xml_entities->getLength - 1)) {
        push(@entities, xml_entity_parse($xml_entities->item($entity_ndx)));
    }

    return ($entity_count, \@entities);
}

################################################################################
#
#   Convenience subroutine for parsing XML returned from the ALM server. This
#   subroutine takes the initial XML format and strips the outermost container.
#   This extracts information about all entities returned by ALM which may span
#   more than the number of results returned in a single HTTP response from ALM.
#
#   Input:
#       The XML document to be parsed.
#       Additional subroutine options. Currently only debug is a recognized
#           option and, if true, triggers extra debug print statements.
#
#   Output:
#       The total number of found entities in ALM for the the previous HTTP
#           query.
#       A reference to the nodelist of entities in the XML data returned by
#           ALM. This is a list of XML nodes directly below the outermost XML
#           node for the ALM XML format.
#
################################################################################
sub xml_entities_get {
    my ($xml_doc, $options) = @_;

    my $root_node = $xml_doc->getFirstChild;

    my $num_entities = $root_node->getAttribute('TotalResults');
    if ($options->{debug}) {
        $LOG->comment("num entities received: $num_entities");
    }

    my $entities = $root_node->getElementsByTagName('Entity');
    if ($options->{debug}) {
        $LOG->comment($entities->getLength." entities gotten");
    }

    return ($num_entities, $entities);
}

################################################################################
#
#   Convenience subroutine for parsing XML returned from the ALM server. This
#   subroutine takes XML and converts it into a hash representing the entity
#   information.
#
#   Input:
#       The XML node to be parsed that represents an ALM entity.
#
#   Output:
#       The entity represented as a hash instead of in XML.
#
################################################################################
sub xml_entity_parse {
    my $xml_entity = $_[0];
    my $entity_data = {};

    foreach my $field_node ($xml_entity->getElementsByTagName('Field')) {
        my $field_name = $field_node->getAttribute('Name');

        my $field_vals = $field_node->getElementsByTagName('Value');

        if ($field_vals->getLength > 0) {
            my $field_val_node = $field_vals->item(0);

            if ($field_val_node->hasChildNodes) {
                $field_val_node = $field_val_node->getFirstChild;

                if ($field_val_node) {
                    my $field_val = $field_val_node->getData;
                    $entity_data->{$field_name} = $field_val;
                }
            }
        }

    }

    return $entity_data;
}
#TODO add debug stuffs to the above subroutine

################################################################################
#
#   Convenience subroutine for building a URL with arbitrary parameters.
#   Currently the ALM entity properties, property-values, and desired properties
#   are passed in as a pre-constructed string. If necessary in the future this
#   method should be extended to accomodate more flexible URL parameters.
#
#   Input:
#       The URL to append URL parameters to.
#       A reference to a hash of arguments denoting what URL parameters to
#           append to the given URL. Currently, recognized arguments are:
#           query_attrs  -> Attributes for an ALM query. An ALM query works as
#                           a filter so that ALM is not overloaded by trying to
#                           retrieve more entities than are necessary. For
#                           example if the desired entity has name
#                           'Automated-Regression' then the query to ALM should
#                           include 'name[Automated-Regression]' as a URL
#                           parameter. This way only entities whose name is
#                           'Automated-Regression' will be returned for further
#                           action. Multiple query attributes may be specified
#                           and are semi-colon (;) delimited.
#           query_fields -> List of entity properties that will be used for
#                           further action. This prevents ALM from returning
#                           more data than is necessary. If the entity name is
#                           the only useful information, then ALM need not
#                           return extra entity information aside from the name
#                           of the entity. This would be done in the form
#                           'fields=name'. Multiple fields may be specified and
#                           are comma-delimited.
#           page_size    -> The number of entities whose information should be
#                           sent in a single page. For example, assume there are
#                           200 entities that ALM has found based on a given
#                           query. If the page_size is only 100, then ALM will
#                           return information on the first 100 entities, and
#                           will need to be queried again with special
#                           parameters that specify the second 100 entities are
#                           desired.
#
#   Output:
#       The new URL with additional parameters appended as necessary.
#
################################################################################
sub build_url {
    my ($url, $args) = @_;

    my $url_param = $CR::ALMConn::RES{$args->{res_type}}.
                    "?query={$args->{query_attrs}}".
                    "&fields=$args->{query_fields}";

    if ($args->{page_size}) {
        $url_param .= "&page-size=$args->{page_size}";
    }

    return $url.$url_param;
}

################################################################################
#
#   Convenience subroutine for checking the type and result of an HTTP response
#   from the ALM server. If the previous HTTP request was unsuccessful for
#   reasons relating to authorization, formatting, etc. then the corresponding
#   HTTP response will describe an error that occurred when the HTTP request
#   was sent.
#
#   Input:
#       The HTTP response received from the ALM server.
#
#   Output:
#       0 (false) if the received HTTP response is undefined. 1 (true) if the
#           HTTP response is non-zero and non-empty, the 'content-type' of the
#           is an expected type ('application/xml'), and the op code is an HTTP
#           success code.
#
################################################################################
sub response_is_good {
    my $res = $_[0];

    if (!defined $res) {
        return 0;
    }

    return $res && $res->header('content-type') eq 'application/xml'
        && is_success($res->code);
}

1;

=pod

=head1 NAME
    ALMConn
=head1 DESCRIPTION
    C<ALMConn> is an object-oriented class (more or less) that aims to provide
    a simple method of accessing ALM via the REST API. It is currently a work
    in progress.
    This means that an HTTP post command is sent to the ALM server's REST
    interface, and the HTTP response sent back to this client is returned.
=head1 GLOBALS
    C<ALMConn> has several globals that are important and used throughout the
    module.

    =over
    =item *
        ALMConn::PROD_URL provides the default (that will likely stay the
        default for many years to come) URL for accessing the production
        ALM server.
    =item *
        ALMConn::TEST_URL provides the default (that will likely stay the
        default for many years to come) URL for accessing the
        staging/development ALM server.
    =item *
        ALMConn::REST is the base resource for the ALM REST API.
    = item *
        ALMConn::RES is a hash containing various resources that this module
        currently uses to a fair extent.
    = item *
        XML_PARSER is a private global object that serves as the XML Parser
        for this module.
    =back
=cut
