## Copyright (c) 2012 NetApp Inc.
## All rights reserved.
##
## @name CRCapabilityObject.pm
##
## @summary Content repository capability object representation
##
## @description
##   Represents the capability object for content repository
##
## @keywords cr, http, capabilityobject
##
## @author Asif Hoda
##
## @example
##
## @burt
##    533470
##
## @status private

=pod

=head1 Package Name

CR::CRCapabilityObject.PM

=head1 Author/Maintainer

Asif Hoda

=head1 Description
Represents the capability object for content repository

=cut

package CR::CRCapabilityObject;

# Compiler directives.
use strict;
use warnings;

use JSON;
use HTTP::Message;
use Hash::Flatten qw(:all);
use NATE::Log qw(log_global);my $Log = log_global();

use CR::CRHttpRequest;
use CR::CRCdmiUtils;
use CR::CRCdmiObject;

our @ISA = qw( CR::CRCdmiObject);

################################################################################
# Default values for capability object read
################################################################################
my (
    $_def_capability_read_method,  $_def_capability_read_uri,
    $_def_capability_read_headers, $_def_capability_read_body
);

$_def_capability_read_method  = "GET";
$_def_capability_read_uri     = "http://localhost:8081/cdmi_capabilities/";
$_def_capability_read_headers = {
    "X-CDMI-Specification-Version" => "1.0.1",
    "Accept"                       => "application/cdmi-capability"
};

$_def_capability_read_body = "";

################################################################################
# Default values for capability object update
################################################################################
my (
    $_def_capability_update_method,  $_def_capability_update_uri,
    $_def_capability_update_headers, $_def_capability_update_body
);

$_def_capability_update_method  = "PUT";
$_def_capability_update_uri     = "http://localhost:8081/cdmi_capabilities/";
$_def_capability_update_headers = {
    "X-CDMI-Specification-Version" => "1.0.1",
    "Content-Type"                 => "application/cdmi-capability"
};

$_def_capability_update_body = "";
################################################################################

################################################################################
# Default values for capability object create
################################################################################
my (
    $_def_capability_create_method,  $_def_capability_create_uri,
    $_def_capability_create_headers, $_def_capability_create_body
);

$_def_capability_create_method  = "PUT";
$_def_capability_create_uri     = "http://localhost:8081/cdmi_capabilities/";
$_def_capability_create_headers = {
    "Accept"                       => "application/cdmi-capability",
    "X-CDMI-Specification-Version" => "1.0.1",
    "Content-Type"                 => "application/cdmi-capability"
};

$_def_capability_create_body = "";
################################################################################

################################################################################
# Default values for capability object delete
################################################################################
my (
    $_def_capability_delete_method,  $_def_capability_delete_uri,
    $_def_capability_delete_headers, $_def_capability_delete_body
);

$_def_capability_delete_method  = "DELETE";
$_def_capability_delete_uri     = "http://localhost:8081/cdmi_capabilities/";
$_def_capability_delete_headers = { "X-CDMI-Specification-Version" => "1.0.1" };

$_def_capability_delete_body = "";
################################################################################

=head2 new

Creates a new Capability Object.

Owns a CR:CRHttpRequest, HTTP::Response and LWP::UserAgent and blesses the hash

=over

=item Arguments

B<lwp> = I<LWP::UserAgent object>

I<Optional> User specified LWP::UserAgent object. The user supplied
LWP::UserAgent is used to send data to CDMI server. If not specified then the
default LWP::UserAgent is used.

=back

=cut

sub new {

    my ( $class, $lwp ) = @_;

    # If user has paseed us an LWP::UserAgent then use that else default
    if ( !$lwp ) {
        $lwp = new LWP::UserAgent;
    }

    my $self = {
        _http_request    => undef,
        _server_response => undef,
        _browser         => $lwp,
    };

    bless $self, $class;
}

sub destroy {
    my ($self) = shift;
    $self->{_server_response} = undef;
    $self->{_http_request}    = undef;
    $self->{_browser}         = undef;
}

sub create {
    $Log->enter();
    my ( $self, $http_request ) = @_;
    $Log->exit();
    return $self->_issue_request($http_request);
}

sub update {
    $Log->enter();
    my ( $self, $http_request ) = @_;
    $Log->exit();
    return $self->_issue_request($http_request);
}

sub read {
    $Log->enter();
    my ( $self, $http_request ) = @_;
    $Log->exit();
    return $self->_issue_request($http_request);
}

sub delete {
    $Log->enter();
    my ( $self, $http_request ) = @_;
    $Log->exit();
    return $self->_issue_request($http_request);
}

sub _issue_request {
    my ( $self, $http_request ) = @_;
    my ( $_def_method, $_def_uri, $_def_headers, $_def_body );

    # Set the defaults based on who called us
    my ($caller) = ( caller(1) )[3];
    if ( $caller eq "CR::CRCapabilityObject::read" ) {
        $_def_method  = $_def_capability_read_method;
        $_def_uri     = $_def_capability_read_uri;
        $_def_headers = $_def_capability_read_headers;
        $_def_body    = $_def_capability_read_body;
    }
    elsif ( $caller eq "CR::CRCapabilityObject::update" ) {
        $_def_method  = $_def_capability_update_method;
        $_def_uri     = $_def_capability_update_uri;
        $_def_headers = $_def_capability_update_headers;
        $_def_body    = $_def_capability_update_body;
    }
    elsif ( $caller eq "CR::CRCapabilityObject::create" ) {
        $_def_method  = $_def_capability_create_method;
        $_def_uri     = $_def_capability_create_uri;
        $_def_headers = $_def_capability_create_headers;
        $_def_body    = $_def_capability_create_body;
    }
    elsif ( $caller eq "CR::CRCapabilityObject::delete" ) {
        $_def_method  = $_def_capability_delete_method;
        $_def_uri     = $_def_capability_delete_uri;
        $_def_headers = $_def_capability_delete_headers;
        $_def_body    = $_def_capability_delete_body;
    }

    if ( "CR::CRHttpRequest" eq ref($http_request) ) {

        # user wants to override the defaults
        if ( "UNINIT" eq $http_request->get_method() ) {
            $http_request->set_method($_def_method);
        }
        if ( "UNINIT" eq $http_request->get_uri() ) {
            $http_request->set_uri($_def_uri);
        }
        if ( "UNINIT" eq $http_request->get_header() ) {
            $http_request->set_header($_def_headers);
        }
        if ( "UNINIT" eq $http_request->get_body() ) {
            $http_request->set_body($_def_body);
        }
    }
    else {
        $Log->trace(
            "Using the default parameters for Capabilities Object operations");
        $http_request = new CR::CRHttpRequest();
        $http_request->set_method($_def_method);
        $http_request->set_uri($_def_uri);
        $http_request->set_header($_def_headers);
        $http_request->set_body($_def_body);
    }

    if ( (defined $http_request->header("ResetHeaders")) &&
         ("True" eq $http_request->header("ResetHeaders")) ) {

        # We are checking if the user has set {} as the headers. In case user
        # sets {} as the headers then user intends to send no headers to server.
        # CR::CRHttpRequest::set_header when sees {} as the headers to be set,
        # inserts a dummy header by name "ResetHeaders" and value as "True". We
        # check if the header is present, and in case it is, then clear all the
        # headers.
        $http_request->clear;
    }

    $self->{_http_request} = $http_request;

    $self->_send();
    return $self->{_server_response};
}

1;
################################################################################
# Lib Ends
################################################################################
