## Copyright (c) 2012 NetApp Inc.
## All rights reserved.
##
## @name CRCDMIObject.pm
##
## @summary Content repository abstract object representation
##
## @description
##   Represents a generic object for content repository
##
## @keywords cr
##
## @author Asif Hoda
##
##
## @status private
##
##

=pod

=head1 Package Name

CR::CRCdmiObject.PM

=head1 Author/Maintainer

Asif Hoda

=head1 Description
Represents an abstract object for content repository

=cut

# Compiler directives.
package CR::CRCdmiObject;

use strict;
use warnings;
use NATE::Log qw(log_global);
my $Log = log_global();
use Hash::Flatten qw(:all);
use IO::Socket::INET;

=head2 get_response_header_value {

Get the value of the header whose key is specified in argument.

=over

=item Arguments

B<server_response> = HTTP::Response

I<Required> Server response from which the object ID needs to be extracted.

=back

=cut

sub get_response_header_value {
    $Log->enter();
    my ( $self, $key ) = @_;

    my $server_response = $self->{_server_response};

    # Check if the server response is of type HTTP::Response.
    if ( "HTTP::Response" eq ref($server_response) ) {
        return $server_response->header("$key");
    }

    $Log->exit();
    return undef;
} ## end sub get_response_header_value

=head2 get_object_id

Get the object ID of the object created from the CDMI server response.

=over

=item Arguments

B<server_response> = HTTP::Response

I<Required> Server response from which the object ID needs to be extracted.

=back

=cut

sub get_object_id {
    $Log->enter();
    my ($self) = @_;
    my $server_response = $self->{_server_response};

    # Check if the server response is of type HTTP::Response.
    if ( "HTTP::Response" eq ref($server_response) ) {

        # Obtaining the OID for a server response for DELETE is meaningless
        # So will check if the request had anything but PUT/POST/GET.
        if (   ( $server_response->request->method eq "POST" )
            || ( $server_response->request->method eq "PUT" )
            || ( $server_response->request->method eq "GET" ) ) {

            # Check if "X-CDMI-Specification-Version" header is present to
            # denote a CDMI response. If CDMI response then return "objectID"
            # from HTTP body
            my ( $response_body, $flat_response_body );
            my $cdmi_spec_hdr
                = $server_response->header("X-CDMI-Specification-Version");

            if ( !defined($cdmi_spec_hdr) || $cdmi_spec_hdr eq "" ) {

               # non cdmi response
               # would return undef if location header is not found as that is
               # the only way to determnine an OID for non CDMI dataobject
                my $location_hdr = $server_response->header("Location");
                if ( defined($location_hdr) && $location_hdr ne "" ) {
                    return
                        substr( $location_hdr,
                        rindex( $location_hdr, "/" ) + 1,
                        length($location_hdr) );
                } else {
                    return undef;
                }
            } else {

                # cdmi response
                if ( $server_response->content ) {

                    # Check if the response is a valid JSON
                    my $json = JSON->new->utf8;
                    $response_body
                        = $json->allow_nonref->utf8->decode(
                        $server_response->content )
                        or return undef;
                    $flat_response_body = flatten($response_body);
                    if ( exists $flat_response_body->{"objectID"} ) {
                        return $flat_response_body->{"objectID"};
                    }
                }
            }
        }
    }
    $Log->exit();
    return undef;
} ## end sub get_object_id

=head2 get_first_level_field_value

Get the value of a first level JSON body field of a CDMI content type response.

=over

=item Arguments

B<server_response> = HTTP::Response

I<Required> Server response from which the first level JSON body field is to be returned.

B<first_level_key> = String

I<Required> Key of the first level field whose value is requested.

=back

=cut

sub get_first_level_field_value {

    $Log->enter();
    my ( $self, $first_level_key ) = @_;

    if ( !defined $first_level_key ) {
        return undef;
    }

    if ( $first_level_key eq "metadata" ) {
        return undef;
    }

    my $server_response = $self->{_server_response};

    # Check if the server response is of type HTTP::Response.
    if ( "HTTP::Response" eq ref($server_response) ) {

        if (   ( $server_response->request->method eq "POST" )
            || ( $server_response->request->method eq "PUT" )
            || ( $server_response->request->method eq "GET" ) ) {

            # Check if "X-CDMI-Specification-Version" header is present to
            # denote a CDMI response.
            my $cdmi_spec_version
                = $server_response->header("X-CDMI-Specification-Version");
            if ( !defined($cdmi_spec_version) || $cdmi_spec_version eq "" ) {

                # non cdmi response
                return undef;
            }

            # cdmi response
            if ( $server_response->content ) {

                # Check if the response is a valid JSON
                my $json = JSON->new->utf8;
                my $response_body
                    = $json->allow_nonref->utf8->decode(
                    $server_response->content )
                    or return undef;

                return $response_body->{$first_level_key};
            }
        }
    }

    $Log->exit();
    return undef;
} ## end sub get_first_level_field_value

=head2 get_current_version_oid

Get the object ID of the current data object version created from the CDMI server response.

=over

=item Arguments

B<server_response> = HTTP::Response

I<Required> Server response from which the object ID needs to be extracted.

=back

=cut

sub get_current_version_oid {
    $Log->enter();
    my ($self) = @_;
    my $server_response = $self->{_server_response};

    # Check if the server response is of type HTTP::Response.
    if ( "HTTP::Response" eq ref($server_response) ) {

        # Obtaining the OID for a server response for DELETE is meaningless
        # So will check if the request had anything but PUT/POST/GET.
        if (   ( $server_response->request->method eq "POST" )
            || ( $server_response->request->method eq "PUT" )
            || ( $server_response->request->method eq "GET" ) ) {

            # Check if "X-CDMI-Specification-Version" header is present to
            # denote a CDMI response. If CDMI response then return "objectID"
            # from HTTP body
            my ( $response_body, $flat_response_body );
            if ( $server_response->header("X-CDMI-Specification-Version") eq
                "" ) {

               # non cdmi response
               # would return undef if location header is not found as that is
               # the only way to determnine an OID for non CDMI dataobject
                if ( $server_response->header("Location") ne "" ) {
                    my $loc_hdr = $server_response->header("Location");
                    return substr( $loc_hdr, rindex( $loc_hdr, "/" ) + 1,
                        length($loc_hdr) );
                }
            } else {

                # cdmi response
                if ( $server_response->content ) {

                    # Check if the response is a valid JSON
                    my $json = JSON->new->utf8;
                    $response_body
                        = $json->allow_nonref->utf8->decode(
                        $server_response->content )
                        or return undef;
                    if (exists $response_body->{"metadata"}
                        ->{"cdmi_version_current"} ) {
                        return $response_body->{"metadata"}
                            ->{"cdmi_version_current"};
                    }
                }
            }
        }
    }
    $Log->exit();
    return undef;
} ## end sub get_current_version_oid

=head2 get_object_name

Get the object name of the object created from the CDMI server response.

=over

=item Arguments

B<server_response> = HTTP::Response

I<Required> Server response from which the object name needs to be extracted.

=back

=cut

sub get_object_name {
    $Log->enter();
    my ($self) = @_;
    my $server_response = $self->{_server_response};

    # Check if the server response is of type HTTP::Response.
    if ( "HTTP::Response" eq ref($server_response) ) {

        # Obtaining the name for a server response for POST/DELETE
        # is meaningless, as POST is not supported for named objects
        # and DELETE has no reponse body.
        if (   ( $server_response->request->method eq "PUT" )
            || ( $server_response->request->method eq "GET" ) ) {

            # Check if "X-CDMI-Specification-Version" header is present to
            # denote a CDMI response. If CDMI response then return "objectID"
            # from HTTP body
            my ( $response_body, $flat_response_body );
            if ( $server_response->header("X-CDMI-Specification-Version") ne
                "" ) {

                # non cdmi response for named object has no name
                # would return undef
            } else {

                # cdmi response
                if ( $server_response->content ) {

                    # Check if the response is a valid JSON
                    my $json = JSON->new->utf8;
                    $response_body
                        = $json->allow_nonref->utf8->decode(
                        $server_response->content )
                        or return undef;
                    $flat_response_body = flatten($response_body);
                    if ( exists $flat_response_body->{"objectName"} ) {
                        return $flat_response_body->{"objectName"};
                    }
                }
            }
        }
    }
    $Log->exit();
    return undef;
} ## end sub get_object_name

=head2 verify_headers

Verify the HTTP headers from the CDMI server against user provided hash.

=over

=item Arguments

B<headers_hash_ref> = HashRef

I<Required> Hash ref with the headers and theie values thereof that need to be
verified against the server response

B<server_response> = HTTP::Response

I<Required> Server response from against which the user passed headers hash will
be matched

=back

=cut

sub verify_headers {
    $Log->enter();
    $Log->trace(
        "#########################################################\n
        Verifying HTTP Headers
        \n#########################################################"
    );

    my ( $self, $headers_hash_ref ) = @_;
    my $server_response   = $self->{_server_response};
    my $unmatched_headers = 0;

    # Verify if $server_response is a HTTP::Response
    if ( "HTTP::Response" ne ref($server_response) ) {
        $Log->trace(
            "Invalid server response passed for HTTP header verification");
        $unmatched_headers += 1;
        return $unmatched_headers;
    }

    # verify of $headers_hash_ref is a HASH
    if ( ref($headers_hash_ref) ne "HASH" ) {
        $Log->trace(
            "Invalid user specified HTTP headers passed to be verified");
        $unmatched_headers += 1;
        return $unmatched_headers;
    }

    # Compare
    while ( my ( $key_user_header, $value_user_header )
        = each(%$headers_hash_ref) ) {

        # verify if the header is present
        if ( $server_response->header($key_user_header) ) {
            if ( $value_user_header eq "" ) {

                # Users just wants to verify for the presence of the header
                $Log->trace( "[ MATCH ] Header ["
                        . $key_user_header
                        . "] present in HTTP server response \n[ SKIP ] It's value for comparison"
                );
            } else {

                # User wants the header and the value to match
                if ( $server_response->header($key_user_header) eq
                    $value_user_header ) {
                    $Log->trace( "[ MATCH ] User passed ["
                            . $key_user_header
                            . "] => ["
                            . $value_user_header
                            . "] matches ["
                            . $key_user_header
                            . "] => ["
                            . $server_response->header($key_user_header)
                            . "]" );
                } else {
                    $unmatched_headers += 1;
                    $Log->trace( "[ MISMATCH ] User passed ["
                            . $key_user_header
                            . "] => ["
                            . $value_user_header
                            . "] doesn't match ["
                            . $key_user_header
                            . "] => ["
                            . $server_response->header($key_user_header)
                            . "]" );
                }
            }
        } else {
            $unmatched_headers += 1;
            $Log->trace( "[ MISMATCH ] Header ["
                    . $key_user_header
                    . "] absent in HTTP server response" );
        }
    }

    $Log->exit();
    return $unmatched_headers;
} ## end sub verify_headers

=head2 verify_body

Verify the HTTP body from the CDMI server against user value.

=over

=item Arguments

B<user_body_ref> = HashRef

I<Required> Hash ref with the HTTP body that needs to be verified against the
CDMI server response

B<server_response> = HTTP::Response

I<Required> Server response from against which the user passed http body hashref
will be matched

=back

=cut

sub verify_body {
    $Log->enter();
    $Log->trace(
        "#########################################################\n
        Verifying HTTP Body
        \n#########################################################"
    );

    my ( $self, $user_body_ref ) = @_;
    my $server_response = $self->{_server_response};
    my $unmatched_body  = 0;

    # Determine if the HTTP reqponse body is CDMI or non CDMI . Best way is to
    # peek into the client request and see if client sent
    # 'X-CDMI-Specification-Version' header. Presence of that tells us that we
    # a CDMI response to be verified which should be in JSON format
    if ( !defined( $server_response->header("X-CDMI-Specification-Version") )
        || $server_response->header("X-CDMI-Specification-Version") eq "" ) {

        # non cdmi response
        if ( $user_body_ref eq $server_response->content ) {
            $Log->trace("[MATCH] http body matched with the given response");
        } else {
            $unmatched_body += 1;
            $Log->trace(
                "[MISMATCH] http body did not match with the given response");
        }
    } else {

        # cdmi response
        # check if server_response is a HTTP::Response
        if ( "HTTP::Response" ne ref($server_response) ) {
            $Log->trace(
                "Invalid server response passed for HTTP body verification");
            $unmatched_body += 1;
            return $unmatched_body;
        }

        # check if the server response has a body
        if ( !( $server_response->content ) ) {
            $Log->trace("Server response has no content");
            $unmatched_body += 1;
            return $unmatched_body;
        }

        # check if the response body is valid JSON or not
        # if yes then hashify the HTTP body
        my $json_response_body = 0;
        my $json               = new JSON;
        if ( $json->decode( $server_response->content ) ) {
            $json_response_body
                = $json->allow_nonref->utf8(0)
                ->decode( $server_response->content )
                or return -1;
        }

        # flatten the HTTP body
        my $flat_json_response_body = flatten($json_response_body);

        # check if the $user_body_ref is a HASH
        if ( ref($user_body_ref) ne "HASH" ) {
            $Log->trace("Invalid user specified body passed to be verified");
            $unmatched_body += 1;
            return $unmatched_body;
        }

        # flatten the user body reference
        my $flat_user_body_ref = flatten($user_body_ref);

        ########################################################################
        ## Hash::Flatten when sees an empty hash as the value for some key, just
        ## ignores the key in while flattening the hash. Hence the resultant
        ## flattened hash no key with the {} as value.
        ##
        ## User supplied "metadata" key is allowed to have {} as value as this
        ## indicates server has flushed all metadata as user asked it to in form
        ## of supplying {} for "metadata" key in http request.
        ##
        ## With the above discussion currently the metadata with {} value check
        ## can never be done.
        ##
        ## Hence the folllowing fix
        ########################################################################
        if ( exists( $json_response_body->{"metadata"} )
            && !( keys %{ $json_response_body->{"metadata"} } ) ) {
            $flat_json_response_body->{"metadata"} = "{}";
        }
        if ( exists( $user_body_ref->{"metadata"} )
            && !( keys %{ $user_body_ref->{"metadata"} } ) ) {
            $flat_user_body_ref->{"metadata"} = "{}";
        }
        ########################################################################

        # start comparing
        while ( my ( $key_user_body, $value_user_body )
            = each(%$flat_user_body_ref) ) {
            if ( exists( $flat_json_response_body->{$key_user_body} ) ) {
                if ( $value_user_body eq "" ) {

            # case when value is "", meaning user is interested in verifying
            # just the presence, not the value per say. useful when values are
            # unpredictable at the http request issue time
                    $Log->trace( "[ MATCH ]  User passed http item ["
                            . $key_user_body
                            . "] present in HTTP body from server response \n[ SKIP ] It's value for comparison"
                    );
                } else {

                # user wants to compare both the presence of key and the value
                # thereof
                    if ( $value_user_body eq
                        $flat_json_response_body->{$key_user_body} ) {

                        # Both user supplied key value pair matched
                        $Log->trace( "[ MATCH ] User passed ["
                                . $key_user_body . "]=>["
                                . $value_user_body
                                . "] matches server response ["
                                . $key_user_body . "]=>["
                                . $flat_json_response_body->{$key_user_body}
                                . "]" );
                    } else {

                        # key present however values don't match
                        $unmatched_body += 1;
                        $Log->trace( "[ MISMATCH ] User passed ["
                                . $key_user_body . "]=>["
                                . $value_user_body
                                . "] doesn't match server response ["
                                . $key_user_body . "]=>["
                                . $flat_json_response_body->{$key_user_body}
                                . "]" );
                    }
                }
            } else {
                $unmatched_body += 1;
                $Log->trace( "[ MISMATCH ]  User passed http item ["
                        . $key_user_body
                        . "] absent in HTTP body" );
            }
        }
    }
    $Log->exit();
    return $unmatched_body;
} ## end sub verify_body

=head2 _send

Sends the owning HTTP::Request object to the server using the owing
user set LWP::UserAgent. In case user doesn't specify LWP::UserAgent then
default LWP::UerAgent is used

=over

=back

=cut

sub _send {
    my ($self) = @_;
    $Log->trace(
        "#########################################################\n
        Client Request
        \n#########################################################"
    );

    $Log->trace( $self->{_http_request}->as_string );
    $Log->trace("##########################################################");

    # _read_response_http_body if present indicates that user wants us to dump
    # the HTTP response from the server in the file specified by the value of
    # _read_response_http_body
    #
    # if absent this would indicate that user wants us to retain the HTTP body
    # from server in memory. this will not work for large HTTP file downloads
    #
    # in case of large file downloads users must compulsorily specify this
    # variable which enables us to do a post download analysis
    if ( defined $self->{_read_response_http_body} ) {

        # user wants us to dump the HTTP body from server response in the file
        # specified by _read_response_http_body
        $self->{_server_response}
            = $self->{_browser}->request( $self->{_http_request},
            $self->{_read_response_http_body} );
    } else {

        # user wants us to keep the HTTP body from server in memory
        $self->{_server_response}
            = $self->{_browser}->request( $self->{_http_request} );
    }

    $Log->trace(
        "#########################################################\n
        Server Response
        \n#########################################################"
    );

    $Log->trace( $self->{_server_response}->as_string );
    $Log->trace("##########################################################");
} ## end sub _send

sub _write_to_socket {

    ############################################################################
    ## Get all the args
    my ( $self, %args ) = @_;
    my $method  = $self->{_http_request}->method();
    my $uri     = $self->{_http_request}->uri();
    my $headers = $self->{_http_request}->headers()->as_string();
    my $body    = $self->{_http_request}->content();

    ############################################################################
    ## Extract the IP address, port
    my $socket_info    = $self->{_socket}         || undef;
    my $socket_timeout = $self->{_socket_timeout} || 10;
    my $packet         = $self->{_packet}         || undef;
    my ( $socket_ip_addess, $socket_port ) = split( /\|/, $socket_info );
    my $content_length = length($body);

    ############################################################################
    ## If user did not pass the packet construct the packet to send over socket
    if ( !defined($packet) ) {
        $packet = $method . ' ' . $uri . ' HTTP/1.1' . "\r\n";

        if ( $headers !~ /Host\s*:\s*/i ) {
            $packet
                .= 'Host: ' . $socket_ip_addess . ':' . $socket_port . "\r\n";
        }

        ## If the request does not contain the content-length header and
        ## http method is either POST or PUT then set the content-length header
        if ( $headers !~ /Content-Length\s*:\s*/i ) {
            $packet .= 'Content-Length: ' . $content_length . "\r\n"
                if ( $method eq 'POST' or $method eq 'PUT' );
        }

        ## Remove the content-length header from request if test does not
        ## require the content-length header
        if ( $headers =~ /Content-Length\s*:\s*remove-content-length-header/ )
        {
            $headers =~ s/Content-Length\s*:.*\n//g;
        }

        $packet .= $headers;
        $packet .= "Pragma: no-cache\r\n" if ( $method ne 'DELETE' );
        $packet .= "\r\n" . $body;
    }

    ############################################################################
    $Log->trace(
        "\n#########################################################\n
        Client Request
        \n#########################################################"
    );

    $Log->trace($packet);
    $Log->trace("##########################################################");

    ############################################################################
    ## Create a socket and connect it to the remote peer

    my $socket = IO::Socket::INET->new(
        PeerAddr => $socket_ip_addess,
        PeerPort => $socket_port,
        Proto    => 'tcp',
        Timeout  => 15,                # Empirical. Not sure if this works !!!
    ) or die "Couldn't connect to Server : $!\n";

    # Set socket options
    # Set autoflush
    $socket->autoflush(1);

    # Set linger options
    $socket->setsockopt( SOL_SOCKET, SO_LINGER, pack( "II", 1, 10 ) )
        or die "Can't set SO_LINGER on the socket : $!";

    ############################################################################
    ## send the created HTTP packet over the socket and handle error if any

    if ( !defined( $socket->send($packet) ) ) {
        $Log->trace("socket send failed");

        # Bad happened. Sent TCP RST to CDMI server and close
        $socket->shutdown(0);
        return undef;
    }

    ############################################################################
    ## recv the cdmi server response over the socket
    my $recv_length = 4086;

    # Add the created socket to the read set
    my $read_set = IO::Select->new($socket) or die "IO::Select $!";
    my ( $server_response, $recv_data, $sockets_ready_for_read,
        $socket_ready_for_read );

    my $whileFlag = 1;

    # The $read_socket_inactivity_time is an empirical value.
    # What this solves:
    # The CDMI server after responding to a request keeps the TCP connection
    # open hoping the CDMI client [test] to reuse the TCP connection to send
    # further requests [persistent connection]. The CDMI client should
    # proactively close the connection if it has nothing to send to server.
    my $read_socket_inactivity_time = $socket_timeout;    # in secs

    while ($whileFlag) {

        # We will only wait for reads to be watched
        ($sockets_ready_for_read)
            = IO::Select->select( $read_set, undef, undef,
            $read_socket_inactivity_time );

        if ( defined $sockets_ready_for_read ) {

            foreach $socket_ready_for_read (@$sockets_ready_for_read) {

                my $no_bytes_read
                    = sysread( $socket_ready_for_read, $recv_data,
                    $recv_length );

                if ( !defined($no_bytes_read) ) {

                    # Error happened. No point in continuing further.
                    $Log->trace("sysread error: $!");
                    $whileFlag = 0;
                }
                if ( !$no_bytes_read ) {

                    # Remote peer gracefully closed so shall we.
                    $Log->trace("Info: Remote peer gracefully closed");
                    $whileFlag = 0;
                }

                if ($no_bytes_read) {

                    # Data will keep coming in next iteration of while loop
                    $server_response .= $recv_data;
                }
            }
        } else {

            # No more data to read
            $whileFlag = 0;
        }
    }

    $read_set->remove($socket_ready_for_read);

    # Shut down gracefully.
    $socket->shutdown(2);

    ## If CDMI server does not reply within 5 seconds (defined in the
    ## $read_socket_inactivity_time variable) then library shuts down the
    ## socket gracefully.
    ## Due to this the $server_response variable remains undef.
    ## To resolve this issue we are creating a dummy HTTP response string
    ## which later will be parsed to a HTTP::Response object
    if ( !defined($server_response) ) {
        $server_response
            = "HTTP/1.1 499 Client Closed Request\r\nContent-Length: 0";
    }

    ############################################################################
    ## Let's feed in the cdmi server response buffer into the CRCdmiObject based
    ## object
    ## HTTP::Response->parse is dumb not to indicate any error if occurs while
    ## parsing a string to make it a HTTP::Response. Hence we are unable to
    ## handle any parsing error condition. Hope, aal izz well ;).
    $self->{_server_response} = HTTP::Response->parse($server_response);

    # Peek into the the received HTTP headers to see if the CDMI server sent
    # chunked Transfer-Encoding. If it did then _dechunkify the HTTP body
    # in place.
    if (defined(
            $self->{_server_response}->headers->header('Transfer-Encoding')
        )
        && ($self->{_server_response}->headers->header('Transfer-Encoding') eq
            "chunked" )
        ) {
        $self->{_server_response}
            ->content( _dechunkify( $self->{_server_response}->content ) );
    }

    ############################################################################
    ## Log the server response
    $Log->trace(
        "\n#########################################################\n
        Server Response
        \n#########################################################"
    );
    $Log->trace( $self->{_server_response}->as_string );
    $Log->trace("##########################################################");

    ############################################################################
    ## We need fill in the HTTP::Response object's _http_request part as this
    ## class contains an interface to obtain the HTTP::Request that resulted in
    ## this HTTP::Response.
    $self->{_server_response}->request( $self->{_http_request} );

} ## end sub _write_to_socket

sub _dechunkify {
    my $chunked_body = shift;

    my $while                        = 1;
    my $crlf                         = "\r\n";
    my $chunk_body                   = undef;
    my $http_body_final              = undef;
    my $curr_chunk_size_start_offset = 0;
    my $curr_chunk_size_end_offset   = 0;
    my $curr_chunk_start_offset      = 0;
    my $curr_chunk_end_offset        = 0;
    my $curr_chunk_size              = 0;

    while ($while) {

        # find the offset where the crlf starts and assign it
        # to $curr_chunk_size_end_offset
        $curr_chunk_size_end_offset
            = index( $chunked_body, $crlf, $curr_chunk_size_start_offset );

        # read from $curr_chunk_size_start_offset for length
        # [$curr_chunk_size_end_offset - $curr_chunk_size_start_offset]
        $curr_chunk_size
            = substr( $chunked_body, $curr_chunk_size_start_offset,
            ( $curr_chunk_size_end_offset - $curr_chunk_size_start_offset ) );

        # hex convert the chunk size and assign it to $curr_chunk_size
        $curr_chunk_size = hex($curr_chunk_size);

        # add 2 to $curr_chunk_size_end_offset get the offset where
        # the chunk starts and assign it to $curr_chunk_start_offset
        $curr_chunk_start_offset = $curr_chunk_size_end_offset + 2;

        # read from offset $curr_chunk_start_offset for length
        # of $curr_chunk_size
        $chunk_body = substr( $chunked_body, $curr_chunk_start_offset,
            $curr_chunk_size );

        # set the $curr_chunk_end_offset to
        # $curr_chunk_size + $curr_chunk_start_offset
        $curr_chunk_end_offset = $curr_chunk_start_offset + $curr_chunk_size;

        # move the $curr_chunk_size_start_offset to 2 + $curr_chunk_end_offset
        $curr_chunk_size_start_offset = $curr_chunk_end_offset + 2;

        $while++;
        if ( !$curr_chunk_size ) {
            $while = 0;
        }

        $http_body_final .= $chunk_body;
    }

    return $http_body_final;
} ## end sub _dechunkify

1;
################################################################################
# Lib Ends
################################################################################
