## Copyright (c) 2012 NetApp Inc.
## All rights reserved.
##
## @name CRCdmiResourceTracker.pm
##
## @summary CDMI resources tracker
##
## @description
##   Represents the resources that tests want to track while CDMI tests are
##   ongoing.
##
## @keywords cr, http
##
## @author Asif Hoda
##
## @burt
##    664805
##
## @status private
##
##

=pod

=head1 Package Name

CR::CRCdmiResourceTracker.PM

=head1 Author/Maintainer

Asif Hoda

=head1 Description
Represents the resources that tests want to track while CDMI tests are ongoing.

=cut

package CR::CRCdmiResourceTracker;

use strict;
use warnings;
use NATE::Log qw(log_global);
use Hash::Flatten qw(:all);
use Options;
use NACL::CS::NetworkConnectionsActive;
use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::APISet;
my $Log = log_global();

sub new {

    my ( $class, %args ) = @_;
    my $resources_to_be_tracked = $args{resources_to_be_tracked} || undef;

    my $options = new Options;
    $options->parameters(
        definitions => [
            {
                name        => "MBUF_LEAK_TOLERANCE",
                default     => "4",
                description => "Limit of Mbuf leak tolerance",
            },
            {
                name        => "CONN_ACTIVE_LEAK_TOLERANCE",
                default     => "5",
                description => "Limit of active network connection tolerance",
            },
            {
                name        => "EXEC_LEAK_TOLERANCE",
                default     => "0",
                description => "Limit of exec leak tolerance",
            },
            {
                name        => "CDMI_ACLC_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.aclc leak tolerance",
            },
            {
                name        => "CDMI_ACLI_TOLERANCE",
                default     => "4",
                description => "Limit of cdmi.acli leak tolerance",
            },
            {
                name        => "CDMI_ACLO_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.aclo leak tolerance",
            },
            {
                name        => "CDMI_CONN_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.conn leak tolerance",
            },
            {
                name        => "CDMI_DATI_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.dati leak tolerance",
            },
            {
                name        => "CDMI_HTPO_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.htpo leak tolerance",
            },
            {
                name        => "CDMI_HTTP_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.http leak tolerance",
            },
            {
                name        => "CDMI_LKPO_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.lkpo leak tolerance",
            },
            {
                name        => "CDMI_MDI_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.mdi leak tolerance",
            },
            {
                name        => "CDMI_MDO_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.mdo leak tolerance",
            },
            {
                name        => "CDMI_MSHO_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.msho leak tolerance",
            },
            {
                name        => "CDMI_PLYO_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.plyo leak tolerance",
            },
            {
                name        => "CDMI_RESO_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.reso leak tolerance",
            },
            {
                name        => "CDMI_RPCO_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.rpco leak tolerance",
            },
            {
                name        => "CDMI_VERI_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.veri leak tolerance",
            },
            {
                name        => "CDMI_VERO_TOLERANCE",
                default     => "0",
                description => "Limit of cdmi.vero leak tolerance",
            },
            {
                name        => "NBLD_NBEC_TOLERANCE",
                default     => "2000",
                description => "Limit of nbld.nbec leak tolerance",
            },
        ]
    );

    # Get the user passed values
    $options->process();
    my $mbuf_leak_tolerance = $options->get( option => 'MBUF_LEAK_TOLERANCE' );
    my $conn_active_leak_tolerance =
      $options->get( option => 'CONN_ACTIVE_LEAK_TOLERANCE' );
    my $exec_leak_tolerance = $options->get( option => 'EXEC_LEAK_TOLERANCE' );
    my $cdmi_aclc           = $options->get( option => 'CDMI_ACLC_TOLERANCE' );
    my $cdmi_acli           = $options->get( option => 'CDMI_ACLI_TOLERANCE' );
    my $cdmi_aclo           = $options->get( option => 'CDMI_ACLO_TOLERANCE' );
    my $cdmi_conn           = $options->get( option => 'CDMI_CONN_TOLERANCE' );
    my $cdmi_dati           = $options->get( option => 'CDMI_DATI_TOLERANCE' );
    my $cdmi_htpo           = $options->get( option => 'CDMI_HTPO_TOLERANCE' );
    my $cdmi_http           = $options->get( option => 'CDMI_HTTP_TOLERANCE' );
    my $cdmi_lkpo           = $options->get( option => 'CDMI_LKPO_TOLERANCE' );
    my $cdmi_mdi            = $options->get( option => 'CDMI_MDI_TOLERANCE' );
    my $cdmi_mdo            = $options->get( option => 'CDMI_MDO_TOLERANCE' );
    my $cdmi_msho           = $options->get( option => 'CDMI_MSHO_TOLERANCE' );
    my $cdmi_plyo           = $options->get( option => 'CDMI_PLYO_TOLERANCE' );
    my $cdmi_reso           = $options->get( option => 'CDMI_RESO_TOLERANCE' );
    my $cdmi_rpco           = $options->get( option => 'CDMI_RPCO_TOLERANCE' );
    my $cdmi_veri           = $options->get( option => 'CDMI_VERI_TOLERANCE' );
    my $cdmi_vero           = $options->get( option => 'CDMI_VERO_TOLERANCE' );
    my $nbld_nbec           = $options->get( option => 'NBLD_NBEC_TOLERANCE' );

    # Defaulting the resources to be tracked if user has not provided. User,
    # if chooses, to provide the resources to be tracked hash, has to confer
    #  with the following hash interface.
    if ( !defined($resources_to_be_tracked) ) {
        $resources_to_be_tracked = {
            rtags => {
                'rtags_modules' => [ 'cdmi', 'ClCdmi', 'nbld' ],
                'tolerance'     => {
                    'default_overall_tolerance' => $mbuf_leak_tolerance,
                    'cdmi.aclc'                 => $cdmi_aclc,
                    'cdmi.acli'                 => $cdmi_acli,
                    'cdmi.aclo'                 => $cdmi_aclo,
                    'cdmi.conn'                 => $cdmi_conn,
                    'cdmi.dati'                 => $cdmi_dati,
                    'cdmi.htpo'                 => $cdmi_htpo,
                    'cdmi.http'                 => $cdmi_http,
                    'cdmi.lkpo'                 => $cdmi_lkpo,
                    'cdmi.mdi'                  => $cdmi_mdi,
                    'cdmi.mdo'                  => $cdmi_mdo,
                    'cdmi.msho'                 => $cdmi_msho,
                    'cdmi.plyo'                 => $cdmi_plyo,
                    'cdmi.reso'                 => $cdmi_reso,
                    'cdmi.rpco'                 => $cdmi_rpco,
                    'cdmi.veri'                 => $cdmi_veri,
                    'cdmi.vero'                 => $cdmi_vero,
                    'nbld.nbec'                 => $nbld_nbec,
                },
            },
            connections => {
                'service' => {
                    'protocol'  => ['cdmi'],
                    'tolerance' => $conn_active_leak_tolerance,
                },
            },
            exec => {

                # Asif: We don't use this array anywhere currently.
                # We have hard coded value in the plugin.
                'exec_sysvar' => ['sysvar.nblade.debug.core.exec_free'],
                'tolerance'   => $exec_leak_tolerance,
            },
        };
    }

    # Resource map representation [Update this when new pluggin are added]
    #    _pre_resource_map--|--rtag--|--node1--|--rtag1--value1
    #                       |        |         |--rtag2--value2
    #                       |        |         |-- ...
    #                       |        |
    #                       |        |--node2--|--rtag1--value1
    #                       |                  |--rtag2--value2
    #                       |                  |-- ...
    #                       |
    #                       |--network.connections--value
    #                       |
    #                       |--sysvar.nblade.debug.core.exec_free--value
    #                       |
    #   _post_resource_map--|--rtag--|--node1--|--rtag1--value1
    #                       |        |         |--rtag2--value2
    #                       |        |         |-- ...
    #                       |        |
    #                       |        |--node2--|--rtag1--value1
    #                       |                  |--rtag2--value2
    #                       |                  |-- ...
    #                       |
    #                       |--network.connections--value
    #                       |
    #                       |--sysvar.nblade.debug.core.exec_free--value

    my $self = {
        _resources_to_be_tracked => $resources_to_be_tracked,
        _pre_resource_map        => undef,
        _post_resource_map       => undef,
        _num_leaked_resources    => 0,
    };

    bless $self, $class;

    # Validate if the $resources_to_be_tracked keys are valid
    # [Update this when new pluggin are added]
    my $valid_tracked_resoure_types = [ 'rtags', 'connections', 'exec' ];
    my $resource_types_to_be_tracked =
      [ keys %{ $self->{_resources_to_be_tracked} } ];

    # Call the plugins for all the resource types
    foreach my $resource_type_to_be_tracked (@$resource_types_to_be_tracked) {
        if ( grep { $_ eq $resource_type_to_be_tracked }
            @$valid_tracked_resoure_types )
        {
            $self->_dispatch_to_fetch_plugin( '_pre_resource_map',
                $resource_type_to_be_tracked );
        }
        else {
            $Log->trace( "Error: Invalid resource ["
                  . $resource_type_to_be_tracked
                  . "] passed to be tracked. Aborting." );
            return undef;
        }
    }

    return $self;
} ## end sub new

sub conclude_resource_watch {
    my ($self) = @_;
    $Log->trace("conclude_resource_watch");

    # Fill the _post_resource_map part of the $self
    my $resource_types_to_be_tracked =
      [ keys %{ $self->{_resources_to_be_tracked} } ];
    foreach my $resource_type_to_be_tracked (@$resource_types_to_be_tracked) {

        # Generate the _post_resource_map section on the map
        $self->_dispatch_to_fetch_plugin( '_post_resource_map',
            $resource_type_to_be_tracked );

        # Compare and conclude the pre and post runs
        $self->_dispatch_to_compare_plugin($resource_type_to_be_tracked);
    }
    return $self->{_num_leaked_resources};
}

sub _dispatch_to_fetch_plugin {
    $Log->trace("_dispatch_to_fetch_plugin");
    my ( $self, $pre_or_post, $resource_type_to_be_tracked ) = @_;

    # Plugin dispatch table [Update this when new pluggin are added]
    my %actions = (
        'rtags'       => \&plugin_rtag_fetch,
        'connections' => \&plugin_connection_fetch,
        'exec'        => \&plugin_exec_fetch,
    );

    $actions{$resource_type_to_be_tracked}->( $self, $pre_or_post );
}

sub _dispatch_to_compare_plugin {
    $Log->trace("_dispatch_to_compare_plugin");
    my ( $self, $resource_type_to_compare ) = @_;

    # Plugin dispatch table [Update this when new pluggin are added]
    my %actions = (
        'rtags'       => \&plugin_rtag_compare,
        'connections' => \&plugin_connection_compare,
        'exec'        => \&plugin_exec_compare,
    );

    $actions{$resource_type_to_compare}->($self);
}

sub plugin_exec_fetch {

    # Every fetch plugin will do these 2 things:
    # 1. Fetch the relevent data from Data OnTap.
    # 2. Fill in the _pre_resource_map/_post_resource_map [$pre_or_post]

    $Log->trace("plugin_exec_fetch");
    my ( $self, $pre_or_post ) = @_;

    # Get all the active network connection filered by the protocol.
    my $node = NACL::C::Node->find();

    # Discovering all the NODES that consitute this cluster
    my @cluster_nodes = NACL::C::Node->find( command_interface => $node );

    # For each node let's get the exec count
    foreach my $cluster_node (@cluster_nodes) {
        my $cluster_node_apiset = NACL::APISet->new(
            name      => $cluster_node->name(),
            category  => "Node",
            interface => "CLI",
            set       => "Systemshell"
        );
        $self->{$pre_or_post}->{'sysvar.nblade.debug.core.exec_free'} +=
          $cluster_node_apiset->sysctl(
            'name' => 'sysvar.nblade.debug.core.exec_free' )
          ->get_parsed_output()->[0]->{'sysvar.nblade.debug.core.exec_free'};
    }    # foreach my $cluster_node (@cluster_nodes)

}    # end plugin_exec_fetch

sub plugin_connection_fetch {

    # Every fetch plugin will do these 2 things:
    # 1. Fetch the relevent data from Data OnTap.
    # 2. Fill in the _pre_resource_map/_post_resource_map [$pre_or_post]

    $Log->trace("plugin_connection_fetch");
    my ( $self, $pre_or_post ) = @_;

    # Get all the active network connection filered by the protocol.
    my $node = NACL::C::Node->find();

    # Fill the network.connections
    # Get all the active network connection filtered by vserver and protocol
    # service
    my $options = new Options;
    $options->process();
    my @NetworkConnectionsActive = ('');
    try {
        @NetworkConnectionsActive = NACL::CS::NetworkConnectionsActive->fetch(
            command_interface => $node,
            filter            => {
                service => join(
                    ',',
                    @{
                        $self->{_resources_to_be_tracked}->{connections}
                          ->{service}->{protocol}
                      }
                ),
                vserver => $options->get( option => 'VSERVER' ),
            }
        );
    }
    catch NACL::Exceptions::NoElementsFound with {
        my $exception = shift;
        $Log->trace("No active network connections found for CDMI protocol.");
        $self->{_pre_resource_map}->{'network.connections'} = 0;
    };

    if ( scalar(@NetworkConnectionsActive) ) {
        $self->{$pre_or_post}->{'network.connections'} =
          scalar(@NetworkConnectionsActive);
    }
}    # end plugin_connection_fetch

sub plugin_rtag_fetch {

    # Every fetch plugin will do these 2 things:
    # 1. Fetch the relevent data from Data OnTap.
    # 2. Fill in the _pre_resource_map/_post_resource_map [$pre_or_post]

    $Log->trace("plugin_rtag_fetch");
    my ( $self, $pre_or_post ) = @_;

    my $node = NACL::C::Node->find();

    # Discovering all the NODES that consitute this cluster
    my @cluster_nodes = NACL::C::Node->find( command_interface => $node );

    my $rtag_modules_to_track =
      $self->{_resources_to_be_tracked}->{rtags}->{rtags_modules};

    # For each node let's get the all rtag counts
    foreach my $cluster_node (@cluster_nodes) {
        my $cluster_node_apiset = $cluster_node->get_7m_or_nodescope_apiset();

        # Fetch rtag belonging to each module
        foreach my $rtag_module_to_track (@$rtag_modules_to_track) {

            # Initialize the rtags map
            my @rtags_to_be_tracked =
              @{ $cluster_node_apiset->rtag( module => $rtag_module_to_track )
                  ->get_parsed_output() };

            foreach my $rtag_to_be_tracked (@rtags_to_be_tracked) {
                $self->{$pre_or_post}->{rtags}->{ $cluster_node->name() }
                  ->{"$rtag_to_be_tracked->{'module_id'}"} =
                  $rtag_to_be_tracked->{'mbufs-in-use'};

            }    # foreach my $rtag_to_be_tracked (@rtags_to_be_tracked)
        }    # foreach my $rtags_modules ($self->{_resources_to_be_tracked} ...
    }    # foreach my $cluster_node (@cluster_nodes)
} ## end sub plugin_rtag_fetch

sub plugin_rtag_compare {

    # Every compare plugin will do these 3 things:
    # 1. Be the resource map layout aware.
    # 2. Do a compare based on the cluster/vserver topology.
    # 3. Increment the _num_leaked_resources if tolerance in breached.

    $Log->trace("plugin_rtag_compare");
    my ($self) = @_;

    # 1. For each node get all the compare all the RTAGS in pre and post
    # Flatten RTAG preresource map
    my $flattened_pre_rtag_resource_map =
      flatten( $self->{_pre_resource_map}->{rtags}, { HashDelimiter => '_' } );

    # Flatten RTAG postresource map
    my $flattened_post_rtag_resource_map =
      flatten( $self->{_post_resource_map}->{rtags}, { HashDelimiter => '_' } );

    # Compare hashes
    for ( keys %$flattened_post_rtag_resource_map ) {
        my $resource_tolerance =
          $self->{_resources_to_be_tracked}->{rtags}->{tolerance}
          ->{default_overall_tolerance};

        # Split the flattened key to get the RTAG name
        my @resource_tolerance_split = split( '_', $_ );

        if (
            defined $self->{_resources_to_be_tracked}->{rtags}->{tolerance}
            ->{ $resource_tolerance_split[1] } )
        {
            $resource_tolerance =
              $self->{_resources_to_be_tracked}->{rtags}->{tolerance}
              ->{ $resource_tolerance_split[1] };
        }

        unless ( exists $flattened_pre_rtag_resource_map->{$_} ) {
            $Log->trace("[Error] $_: not found in pre resource map. Ouch!!!");
            next;
        }
        my $no_leaked_mbufs =
          $flattened_post_rtag_resource_map->{$_} -
          $flattened_pre_rtag_resource_map->{$_};

        if ( $no_leaked_mbufs > $resource_tolerance ) {
            $Log->trace( "[Error]" 
                  . $_
                  . " value pre-calibration ["
                  . $flattened_pre_rtag_resource_map->{$_}
                  . "] vs ["
                  . $flattened_post_rtag_resource_map->{$_}
                  . "] post-calibration. Tolerance ["
                  . $resource_tolerance
                  . "] breached. ["
                  . $no_leaked_mbufs
                  . "] mbufs leaked." );

            # Increment the RTAG leaks if above tolerance levels.
            $self->{_num_leaked_resources} +=
              ( $flattened_post_rtag_resource_map->{$_} -
                  $flattened_pre_rtag_resource_map->{$_} );
        }
        elsif ($no_leaked_mbufs) {
            $Log->trace( "[Error]" 
                  . $_
                  . " value pre-calibration ["
                  . $flattened_pre_rtag_resource_map->{$_}
                  . "] vs ["
                  . $flattened_post_rtag_resource_map->{$_}
                  . "] post-calibration. ["
                  . $no_leaked_mbufs
                  . "] mbufs leaked." );
        }
    }

} ## end sub plugin_rtag_compare

sub plugin_connection_compare {

    # Every compare plugin will do these 3 things:
    # 1. Be the resource map layout aware.
    # 2. Do a compare based on the cluster/vserver topology.
    # 3. Increment the _num_leaked_resources if tolerance in breached.

    $Log->trace("plugin_connection_compare");
    my ($self) = @_;

    my $no_leaked_connections =
      $self->{_post_resource_map}->{'network.connections'} -
      $self->{_pre_resource_map}->{'network.connections'};

    if ($no_leaked_connections) {
        $Log->trace( "[Error]"
              . "Cluster wide connection leak value pre-calibration ["
              . $self->{_pre_resource_map}->{'network.connections'}
              . "] vs ["
              . $self->{_post_resource_map}->{'network.connections'}
              . "] post-calibration. Tolerance ["
              . $self->{_resources_to_be_tracked}->{connections}->{service}
              ->{tolerance} 
              . "]" );
    }

    if ( $no_leaked_connections >
        $self->{_resources_to_be_tracked}->{connections}->{service}
        ->{tolerance} )
    {
        $Log->trace( "[Error]"
              . "Cluster wide connection leak value pre-calibration ["
              . $self->{_pre_resource_map}->{'network.connections'}
              . "] vs ["
              . $self->{_post_resource_map}->{'network.connections'}
              . "] post-calibration. Tolerance ["
              . $self->{_resources_to_be_tracked}->{connections}->{service}
              ->{tolerance}
              . "] breached. ["
              . $no_leaked_connections
              . "] conenctions leaked." );

        # Increment the leaks if above tolerance levels.
        $self->{_num_leaked_resources} += $no_leaked_connections;
    }

} ## end sub plugin_connection_compare

sub plugin_exec_compare {

    # Every compare plugin will do these 3 things:
    # 1. Be the resource map layout aware.
    # 2. Do a compare based on the cluster/vserver topology.
    # 3. Increment the _num_leaked_resources if tolerance in breached.

    $Log->trace("plugin_exec_compare");
    my ($self) = @_;

    my $no_leaked_execs =
      $self->{_post_resource_map}->{'sysvar.nblade.debug.core.exec_free'} -
      $self->{_pre_resource_map}->{'sysvar.nblade.debug.core.exec_free'};

    if ($no_leaked_execs) {
        $Log->trace( "[Error]"
              . "Cluster wide exec context leak value pre-calibration ["
              . $self->{_pre_resource_map}
              ->{'sysvar.nblade.debug.core.exec_free'} 
              . "] vs ["
              . $self->{_post_resource_map}
              ->{'sysvar.nblade.debug.core.exec_free'}
              . "] post-calibration. Tolerance ["
              . $self->{_resources_to_be_tracked}->{'exec'}->{tolerance}
              . "]" );
    }

    if ( $no_leaked_execs >
        $self->{_resources_to_be_tracked}->{'exec'}->{tolerance} )
    {
        $Log->trace( "[Error]"
              . "Cluster wide exec context leak value pre-calibration ["
              . $self->{_pre_resource_map}->{'sysvar.nblade.debug.core.exec_free'}
              . "] vs ["
              . $self->{_post_resource_map}->{'sysvar.nblade.debug.core.exec_free'}
              . "] post-calibration. Tolerance ["
              . $self->{_resources_to_be_tracked}->{'exec'}->{tolerance}
              . "] breached. ["
              . $no_leaked_execs
              . "] exec context(s) leaked." );

        # Increment the leaks if above tolerance levels.
        $self->{_num_leaked_resources} += $no_leaked_execs;
    }

} ## end sub plugin_exec_compare

sub destroy {
    my ($self) = shift;
    $self->{_pre_resource_map}  = undef;
    $self->{_post_resource_map} = undef;
} ## end sub destroy

1;
################################################################################
# Lib Ends
################################################################################
