#
# $Id: //depot/prod/test/main/lib/CR/CRCdmiUtils.pm#1 $
#
## @summary  CDMI utils module
## @author   dl-cr-obj-store@netapp.com
## @status   Private
## @pod      here

=head1 NAME

CR::CRCdmiUtils

=head1 DESCRIPTION

This is CR class for CDMI utils

=head1 METHODS

=cut

package CR::CRCdmiUtils;

# Compiler directives.
use strict;
use warnings;

use Server;
use NACL::APISet;
use NACL::C::Cluster;
use NACL::CS::Cluster;
use NACL::CS::Vserver;
use NACL::CS::NetworkInterface;
use NACL::C::Node;
use NACL::C::Vserver;
use Tharn qw($Log);
use File::Compare;
use Scalar::Util qw(reftype);
use Hash::Flatten qw(:all);
use Data::Validate::IP;
use URI::Split qw(uri_split);
use NACL::STask::Fiji;
use NATE::BaseException qw(:try);


#
# Import exceptions
#
use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::Exceptions::UnexpectedState;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException qw(:try);
use NACL::APISet::Exceptions::NoMatchingEntriesException qw(:try);

################################################################################
# METHODS
################################################################################
sub prerun {
    my %args           = @_;
    my $node           = $args{node};
    my $vserver        = $args{vserver};
    my $cdmiport       = $args{cdmiport};
    my $urischeme      = $args{urischeme};
    my $skipvalidation = $args{skipvalidation};

    my $cluster_state = undef;
    my $vserver_state = undef;
    my $volume_state  = undef;
    my $interface_state = undef;

    # Will throw NACL::Exceptions::NoElementsFound if there is no infinite
    # volume hosted on vsever
    $volume_state = NACL::CS::Volume->fetch(
        command_interface => $node,
        filter            => {
            vserver        => $vserver,
            'volume-style' => "infinitevol"
        }
    );

    # Will throw NACL::Exceptions::NoElementsFound if no data lif under the
    # vserver has cdmi protocol enabled
    $interface_state = NACL::CS::NetworkInterface->fetch(
        command_interface => $node,
        randomize_order   => 1,
        filter            => {
            vserver         => $vserver,
            'role'          => 'data',
            'data-protocol' => ['cdmi'],
            'status-oper'   => 'up',
            'status-admin'  => 'up',
        }
    );

    if ( !$skipvalidation ) {

        # Will throw NACL::Exceptions::NoElementsFound if the NODE is "invalid"
        $cluster_state = NACL::CS::Cluster->fetch(
            command_interface => $node,
            filter            => {
                health      => "true",
                eligibility => "true"
            }
        );

        # Will throw NACL::Exceptions::NoElementsFound if the VSERVER doesn't
        # exist
        $vserver_state = NACL::CS::Vserver->fetch(
            command_interface => $node,
            filter            => { vserver => $vserver, }
        );

        $Log->comment( "
        \*********************************************************************\n
        Following will be asserted:\n
        - Assert if the NODE part of cluster and is healthy.\n
        - Assert if VSERVER exists in cluster of which the NODE is a part of.\n
        - Assert if VSERVER is a repository vserver.\n
        - Assert if an infinite volume is hosted on VSERVER.\n
        - Assert if the infinite volume is online.\n
        - Assert if the infinite volume has a junction path.\n
        - Assert if the infinite volume has object store enabled.\n
        - Assert if VSERVER has at least one CDMI enabled data lif.\n
        - Assert if CDMI port is valid cdmi port.\n
        - Assert if uri scheme is valid.\n
        ************************************************************************
        " );

        $Log->comment("******************************************************");
        $Log->comment(" Assert if the NODE part of cluster and is healthy");
        $Log->comment("******************************************************");

        $Log->comment( "[" . $node->node . "] is healthy and eligible" );

        $Log->comment("******************************************************");
        $Log->comment(" Assert if VSERVER exists in cluster");
        $Log->comment("******************************************************");

        $Log->comment(
            "[" . $vserver_state->vserver . "]exists in the cluster" );

        $Log->comment("******************************************************");
        $Log->comment(" Assert if VSERVER is a repository vserver");
        $Log->comment("******************************************************");
        if ( $vserver_state->is_repository ne 'true' ) {
            NACL::Exceptions::UnexpectedState->throw( "vserver "
                  . $vserver_state->vserver
                  . " is not a repository vserver" );
        }
        $Log->comment(
            "[" . $vserver_state->vserver . "] is a repository vserver" );

        $Log->comment("******************************************************");
        $Log->comment( " Assert if VSERVER hosts at least one data lif with "
              . "CDMI protocol enabled" );
        $Log->comment("******************************************************");

        $Log->comment( "Randomly chosen lif ["
              . $interface_state->address
              . "] has CDMI protocol enabled and has operational and admin "
              . " statues UP and data role" );

        $Log->comment("******************************************************");
        $Log->comment(" Assert if an infinite volume is hosted on VSERVER");
        $Log->comment("******************************************************");

        $Log->comment( "volume ["
              . $volume_state->volume
              . "] hosted under ["
              . $vserver_state->vserver
              . "] is an infinite volume" );

        $Log->comment("******************************************************");
        $Log->comment(" Assert if the infinite volume is online");
        $Log->comment("******************************************************");
        if ( $volume_state->state ne 'online' ) {
            NACL::Exceptions::UnexpectedState->throw(
                "infinite volume " . $volume_state->volume . " is not online" );
        }
        $Log->comment( "volume [" . $volume_state->volume . "] is online" );

        $Log->comment("******************************************************");
        $Log->comment(" Assert if the infinite volume has a junction path.");
        $Log->comment("******************************************************");
        if ( $volume_state->junction_path eq '' ) {
            NACL::Exceptions::UnexpectedState->throw( "infinite volume "
                  . $volume_state->volume
                  . " has no junction path" );
        }
        $Log->comment( "volume ["
              . $volume_state->volume
              . "] has junction path "
              . $volume_state->junction_path );

        $Log->comment("******************************************************");
        $Log->comment(" Assert if infinite volume has object store enabled.");
        $Log->comment("******************************************************");
        if ( $volume_state->enable_object_store ne 'true' ) {
            NACL::Exceptions::UnexpectedState->throw( "infinite volume ["
                  . $volume_state->volume
                  . "] has no object store" );
        }
        $Log->comment(
            "volume [" . $volume_state->volume . "] has object store enabled" );

        $Log->comment("******************************************************");
        $Log->comment(" Assert if CDMI port is valid cdmi port");
        $Log->comment("******************************************************");
        #                        HTTP   HTTPS
        my @valid_cdmi_ports = ( 8081 , 8443 );

        if ( !scalar( grep( /\b$cdmiport\b/i, @valid_cdmi_ports ) ) ) {
            NACL::Exceptions::UnexpectedState->throw(
                "CDMI port [" . $cdmiport . " is not standard CDMI port" );
        }

        $Log->comment("CDMI port validated to be valid CDMI port");

        $Log->comment("******************************************************");
        $Log->comment(" Assert if uri scheme is valid");
        $Log->comment("******************************************************");
        my @valid_uri_scheme = ( 'http', 'https' );

        if ( !scalar( grep( /\b$urischeme\b/i, @valid_uri_scheme ) ) ) {
            NACL::Exceptions::UnexpectedState->throw(
                "[" . $urischeme . " is not a valid CDMI uri scheme" );
        }
        $Log->comment( "[" . $urischeme . "] is a valid CDMI uri scheme" );
    }

    # TODO: Optimize, sending a hashref is a bad idea
    my $ret_hash = {
        random_data_lif => $interface_state->address,
        urischeme       => $urischeme,
        cdmiport        => $cdmiport,
        junction_path   => $volume_state->junction_path,
        uri             => $urischeme . '://'
          . $interface_state->address . ':'
          . $cdmiport . '/',
    };

    return $ret_hash;
} ## end sub validate_prerun

=head2 validate_cluster_has_atleast_two_nodes

Validates if the cluster has atleast two nodes.

=over

=item Arguments

B<node_obj> = I<NACL::C::Node object> | <filer spec>

I<Required> Filer on which to validate the nodes.

B<NodeInfo> = Reference

I<Required> PerlRef which will contain the HASHREF with the Node info on
return. HashRef is owned by the caller.

=back

=cut

sub validate_cluster_has_atleast_two_nodes {
    my ( $node_obj, $node_info ) = @_;

    my @node_arr = NACL::C::Cluster->find( command_interface => $node_obj );
    my $num_nodes = @node_arr;
    if ( $num_nodes < 2 ) {
        $Log->comment( "ERROR: A minimum of 2 nodes is needed to run these "
              . "test cases. Your cluster has ["
              . $num_nodes
              . "]\n" );
        return -1;
    } ## end if ( $num_nodes < 2 )

    $Log->comment(
        "There are [" . $num_nodes . "] nodes present in the cluster\n" );

    # Add each node's name as a key into the NodeInfo hash.
    foreach my $n (@node_arr) {
        $node_info->{ $n->node() }->{exists} = 1;
    }

    return ( 0, $node_info );
} ## end sub validate_cluster_has_atleast_two_nodes

=head2 populate_node_conn_info

Populates the node connection info only for the node specified in the FILER
parameter. Doesn't populate for all the nodes in the cluster.

=over

=item Arguments

B<node_name> = String

I<Required> Name of the node for which information needs to be populated.

B<NodeInfo> = Reference

I<Required> PerlRef containing the information about the node. Typically
returned by validate_cluster_has_atleast_two_nodes

=back

=cut

sub populate_node_conn_info {

    # Presently we populate the node connection info only for the
    # node specified in the FILER parameter. No need to do this
    # for every node in the cluster as we don't talk to any other
    # nodes in this test.
    my ( $node_name, $node_info ) = @_;

    if (defined($node_name)){
        $Log->comment(
                      "Caching connection entries for Node: [" . $node_name . "]\n" );
    }
    my $hostObj = Tharn::host($node_name);

    $Log->comment(
        "Adding ngsh connection info for Node: [" . $node_name . "]\n" );

    my $version_manager = NACL::APISet::Version->new(
        hostobj    => $hostObj,
        category   => "Node",
        shell_type => "CMode"
    );
    $node_info->{$node_name}->{ngshCLI} = NACL::APISet->new(
        hostobj      => $hostObj,
        category     => "Node",
        interface    => "CLI",
        set          => "CMode",
        version_info => {
            compatibility    => 0,    # Disable compatibility
            # cdefs_from_build => 1,    # Load CDEFs directly from build root.
            version_manager =>
              $version_manager, # Reusing already created version manager object
          }

    );

    $Log->comment(
        "Adding FreeBSD connection info for Node: [" . $node_name . "]\n" );
    $node_info->{$node_name}->{freeBSDCLI} = NACL::APISet->new(
        hostobj    => $hostObj,
        category   => "Node",
        interface  => "CLI",
        set        => "Systemshell",
        shell_type => 'CMode',
    );

    $Log->comment( "Adding Server.pm connection info for Node: ["
          . $node_name . "].\n"
          . "This one is a temporary requirement until BURT 451150 "
          . "brings\nin support for 'debug smdb table' commands into "
          . "NACL APISet\n" );
    my $server_obj = Server->new;
    $server_obj->set_execution_server( -Server => "$node_name" );
    $node_info->{$node_name}->{Server} = $server_obj;

    return $node_info;
} ## end sub populate_node_conn_info

=head2 validate_vserver_exists

Validates if the specified VServer exists in the specified Node.

=over

=item Arguments

B<vserver_name> = String

I<Required> Name of the VServer who's presence needs to be validated on the Node

B<node_obj> = NACL::C::Node

I<Required> Node object on which the VServer presence needs to be verified

=back

=cut

sub validate_vserver_exists {
    my ( $vserver_name, $node_obj ) = @_;
    my $repos_vserver_obj = undef;
    my $caught_exception  = 0;
    try {
        $repos_vserver_obj = NACL::C::Vserver->find(
            command_interface => $node_obj,
            filter            => { vserver => $vserver_name },
        );
    } ## end try
    catch NACL::Exceptions::NoElementsFound with {
        $Log->comment( "ERROR: Vserver ["
              . $vserver_name
              . "] does not exist "
              . "in cluster\n" );
        $caught_exception = 1;
    };
    return -1 if ($caught_exception);

    $Log->comment( "Vserver [" . $vserver_name . "] exists in the cluster\n" );

    return 0;
} ## end sub validate_vserver_exists

=head2 validate_vserver_is_repository

Validates if the VServer has a repository set up on it.

=over

=item Arguments

B<vserver_name> = String

I<Required> VServer which should have a repository created on it.

B<node_obj> = NACL::C::Node

I<Required> Node object on which the VServer is present.

=back

=cut

sub validate_vserver_is_repository {
    my ( $vserver_name, $node_name, $node_info ) = @_;

    # As there is no NACL component available yet for the repository vserver
    # we fall back on the APISet to perform this set of commands.
    my $ngshCLI = $node_info->{$node_name}->{ngshCLI};

    # Determine the vserver is a repository vserver
    my $repos_resp;
    my $caught_exception = 0;
    try {
        $repos_resp =
          $ngshCLI->repository_volume_show( vserver => $vserver_name );
    }
    catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        my $exception_object = shift;
        my $response_object  = $exception_object->get_response_object();
        $Log->comment(
            "ERROR: Command invoked: " . $response_object->get_command() );
        $Log->comment( "ERROR: Response raw output at time of exception:\n"
              . $response_object->get_raw_output() );
        $Log->comment( "ERROR: Vserver ["
              . $vserver_name
              . "] is not a repository "
              . "vserver\n" );
        $caught_exception = 1;
    } ## end with
    otherwise {
        $Log->comment("ERROR: Unexpected exception occurred\n");
        $caught_exception = 1;
    };
    return -1 if ($caught_exception);

    my @repos_out = @{ $repos_resp->get_parsed_output() };

    # For debug
    # $Log->comment("Repository info: ".Dumper(@repos_out)."\n");

    if ( $repos_out[0]->{'status'} ne "normal" ) {
        $Log->comment( "ERROR: Vserver ["
              . $vserver_name
              . "] is a repository "
              . "vserver. But it is not 'normal'. Present repository "
              . "status is ["
              . $repos_out[0]->{'status'}
              . "]\n" );
        return -1;
    } ## end if ( $repos_out[0]->{'status'...

    $Log->comment( "Vserver ["
          . $vserver_name
          . "] is a repository vserver "
          . "and is in 'normal' (aka usable) state\n" );

    return ( 0, $node_info );
} ## end sub validate_vserver_is_repository

=head2 determine_repository_data_lifs

Determine the data lifs on which the repository set is done.

=over

=item Arguments

B<vserver_name> = String

I<Required> VServer which should have a repository created on it.

B<node_obj> = NACL::C::Node

I<Required> Node object on which the VServer is present.

B<NodeInfo> = HashRef

I<Required> Perl HashRef containing the information about the node. Typically
returned by validate_cluster_has_atleast_two_nodes. This will add a key called
"reposDataLIF" under the node, which will have the IP address of the data LIF.

=back

=cut

sub determine_repository_data_lifs {
    my ( $vserver_name, $node_name, $node_info ) = @_;

    # As there is no NACL component available yet for the vserver data LIFs
    # we fall back on the APISet to perform this set of commands.
    my $ngshCLI = $node_info->{$node_name}->{ngshCLI};

    # Determine the data LIFs of the repository vserver
    my $repos_resp;
    my $caught_exception = 0;
    try {
        $repos_resp = $ngshCLI->network_interface_show(
            vserver         => $vserver_name,
            'data-protocol' => 'cdmi'
        );
    } ## end try
    catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        my $exception_object = shift;
        my $response_object  = $exception_object->get_response_object();
        $Log->comment(
            "ERROR: Command invoked: " . $response_object->get_command() );
        $Log->comment( "ERROR: Response raw output at time of exception:\n"
              . $response_object->get_raw_output() );
        $Log->comment( "ERROR: Vserver ["
              . $vserver_name
              . "] does not have any data "
              . "LIFs configured\n" );
        $caught_exception = 1;
    } ## end with
    otherwise {
        $Log->comment("ERROR: Unexpected exception occurred\n");
        $caught_exception = 1;
    };
    return -1 if ($caught_exception);

    my @repos_out = @{ $repos_resp->get_parsed_output() };

    # For debug
    # $Log->comment("Data LIFs info: ".Dumper(@repos_out)."\n");

    # Locate a data LIF that's hosted on the node specified in the FILER
    # parameter. We will send all CDMI commands to this data LIF
    foreach my $lif_obj (@repos_out) {
        if (   ( $lif_obj->{'curr-node'} eq $node_name )
            && ( $lif_obj->{'status-oper'} eq "up" ) )
        {

            # We pick the first data LIF that matches the above condition and
            # store it in the NodeInfo hash
            $node_info->{$node_name}->{reposDataLIF} = $lif_obj;
            last;
        } ## end if ( ( $lif_obj->{'curr-node'...
    } ## end foreach my $lif_obj (@repos_out)

    if ( exists $node_info->{$node_name}->{reposDataLIF} ) {
        $Log->comment( "The Data LIF IP address for node ["
              . $node_name
              . "] that we will use\nin this test is ["
              . $node_info->{$node_name}->{reposDataLIF}->{'address'}
              . "]\n" );
    }
    else {
        $Log->comment(
                "ERROR: There is no repository vserver data LIF hosted on"
              . "the FILER ["
              . $node_name
              . "]. For this test it is "
              . "necessary a repository data LIF be present.\n" );
        return -1;
    } ## end else [ if ( exists $NodeInfo->...

    return ( 0, $node_info );
} ## end sub determine_repository_data_lifs

=head2 extract_capability_oids_from_rdb

Extracts the OIDs of the capability objects.

=over

=item Arguments

B<vserver_name> = String

I<Required> VServer which should have a repository created on it.

B<node_obj> = NACL::C::Node

I<Required> Node object on which the VServer is present.

B<NodeInfo> = HashRef

I<Required> Perl HashRef containing the information about the node. Typically
returned by validate_cluster_has_atleast_two_nodes. This will add a key called
"reposDataLIF" under the node, which will have the IP address of the data LIF.

=back

=cut

sub extract_capability_oids_from_rdb {
    my %args    = @_;
    my $node    = $args{node};
    my $vserver = $args{vserver};

    # Extract the underlying APISet
    my $node_apiset = $node->apiset();

    # No APISets are present hence the last respite. File a burt to provide
    # the CDEFs onto the build.
    my $raw_output = $node_apiset->execute_command( command =>
          "debug smdb table cdmi_capabilities show -vserver $vserver" );

    my $capability_oid          = {};
    my $standard_nacl_delimiter = 'X;X';
    my @arr_raw_output          = split /\n/, $raw_output;
    foreach my $row_raw_output (@arr_raw_output) {
        my ( $vserver, $object_type, $oid ) =
          split( $standard_nacl_delimiter, $row_raw_output );

        # Just capture the OIDs ignoring unwanted columns and rows
        if ( !( $object_type =~ m/^(O|o)bject/ ) ) {
            $capability_oid->{$object_type} = $oid;
        }
    }
    return $capability_oid;
} ## end sub extract_capability_oids_from_rdb

sub repository_oid_info {
    my ( $vserver_name, $node_name, $node_info, $oid_dataobject ) = @_;

    # Run the 'repository oid show' command
    # As there is no NACL component available yet for the repository
    # vserver we fall back on the APISet to perform this set of
    # commands.
    my $ngshCLI = $node_info->{$node_name}->{ngshCLI};

    my $repos_oid_resp;
    my $caught_exception = 0;
    try {
        $repos_oid_resp = $ngshCLI->repository_oid_show(
            vserver => $vserver_name,
            oid     => $oid_dataobject
        );
    }
    catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        my $exception_object = shift;
        my $response_object  = $exception_object->get_response_object();
        $Log->comment(
            "ERROR: Command invoked: " . $response_object->get_command() );
        $Log->comment( "ERROR: Response raw output at time of exception:\n"
              . $response_object->get_raw_output() );
        $Log->comment( "ERROR: Repository Vserver ["
              . $vserver_name
              . "] OID show command failed\n" );
        $caught_exception = 1;
    }
    otherwise {
        $Log->comment("ERROR: Unexpected exception occurred\n");
        $caught_exception = 1;
    };
    return -1 if ($caught_exception);

    my @repos_oid_out = @{ $repos_oid_resp->get_parsed_output() };

    # For debug
    $Log->comment( "Repository OID info: " . Dumper(@repos_oid_out) . "\n" );
}

=head2 compare_files

Compare files

=over

=item Arguments

B<left_file> = File Path

I<Required> 1st file to be compared.

B<right_file> = File Path

I<Required> 2nd file to be compared.

=back

=cut

sub compare_files {
    my ( $left_file, $right_file ) = @_;
    return compare( $left_file, $right_file );

}

=head2 fiji_filter

Enables a FIJI

=over

=item Arguments

B<node_obj> = NACL::C::Node

I<Optional> Node object on which the VServer is present.

B<filter> = String

I<Required> Fiji filter.

=back

=cut

sub set_fiji_cluster {
    my %args = @_;
    my $cmd_interface = $args{node_obj};
    my $filter = $args{filter};
    my $ret = 0;

    $cmd_interface->refresh_command_interface(no_version_change => 1);
    my @node_list = NACL::C::Node->find(command_interface => $cmd_interface,);
    foreach my $node (@node_list) {
        try {
            NACL::STask::Fiji->set(
                                        command_interface => $node,
                                        tag               => 'set_fiji',
                                        program           => $filter,
                                        nacltask_param => ['set_fiji'],
                    );
        }
        catch NATE::BaseException with {
            my $exception_object = shift;
            $Log->comment( $exception_object->stacktrace() );
            $ret =1;
        }
    }
    return $ret;
} ## end sub fiji_filter

=head2 find_random_data_lif

find random lif

=over

=item Arguments

B<vserver_name> = String

I<Required> VServer which should have a repository created on it.

B<node_obj> = NACL::C::Node

I<Required> Node object on which the VServer is present.

B<role> = String

I<Required> role eaither data/mgmt...

B<data_protocol> = string

I<Required> protocol cdmi/nfs....

=back

=cut

sub find_random_lif {
    my %args = @_;
    my $node = $args{node};
    my $vserver = $args{vserver};
    my $role = $args{role} || 'data';
    my $data_proto = $args{data_proto} || 'cdmi';
    my $oper_status = $args{oper_status} || 'up';
    my $admin_status = $args{admin_status} || 'up';

    # Will throw NACL::Exceptions::NoElementsFound if no data lif under the
    # vserver has cdmi protocol enabled
    my $random_lif = NACL::CS::NetworkInterface->fetch(
        randomize_order   => 1,
        command_interface => $node,
        requested_fields => ['address'],
        filter            => {
            vserver        => $vserver,
            'role'         => $role,
            'status-oper'  => $oper_status,
            'status-admin' => $admin_status,
            'data-protocol' => [$data_proto],
        }
    );
    #returning data lif address
    return $random_lif->address;
} ## end sub find_random_data_lif

=head2 fiji_cdmi_enable_spinnp_request_infra_error

Enables FIJI to inject infra error in spinnp request path.

=over

=item Arguments

B<node_obj> = NACL::C::Node

I<Required> Node object on which the VServer is present.

B<handler_state> = Integer

I<Required> Handler state for which the FIJI is to be enabled.

B<error> = Integer

I<Required> Infra error to inject.

=back

=cut

sub fiji_cdmi_enable_spinnp_request_infra_error {
        my %args = @_;
    my $node_obj = $args{node_obj};
    my $handler_state = $args{state_no};
    my $infra_error = $args{error_no};

    # Enable FIJI
    my $fiji_command = "(cdmi.spinnp.request.infra.error(== \@handlerState "
                       . $handler_state . ")(return $infra_error))";

    my $enable_ret = CR::CRCdmiUtils::set_fiji_cluster(
                            'node_obj' => $node_obj,
                            'filter' => $fiji_command,
                            );

    if(0 != $enable_ret) {
        $Log->comment("Failed to enable FIJI for handler state: " .
                      $handler_state . ", infra error: " . $infra_error);
    }

    return $enable_ret;
} # end fiji_cdmi_enable_spinnp_request_infra_error()

=head2 fiji_cdmi_enable_spinnp_response_spinnp_error

Enables FIJI to inject spinnp error in spinnp response path.

=over

=item Arguments

B<node_obj> = NACL::C::Node

I<Required> Node object on which the VServer is present.

B<handler_state> = Integer

I<Required> Handler state for which the FIJI is to be enabled.

B<error> = Integer

I<Required> Spinnp error to inject.

=back

=cut

sub fiji_cdmi_enable_spinnp_response_spinnp_error {
    my %args = @_;
    my $node_obj = $args{node_obj};
    my $handler_state = $args{state_no};
    my $spinnp_error = $args{error_no};

    # Enable FIJI
    my $fiji_command = "(cdmi.spinnp.response.spin.error(== \@handlerState "
                       . $handler_state . ")(return $spinnp_error))";

    my $enable_ret = CR::CRCdmiUtils::set_fiji_cluster(
                            'node_obj' => $node_obj,
                            'filter' => $fiji_command
                            );

    if(0 != $enable_ret) {
        $Log->comment("Failed to enable FIJI for handler state: " .
                      $handler_state . ", infra error: " . $spinnp_error);
    }

    return $enable_ret;
} # end fiji_cdmi_enable_spinnp_response_spinnp_error()


=head2 check_locks

Checks locks on a file / directory.

=over

=item Arguments

B<node_obj> = NACL::C::Node

I<Required> Node object on which the VServer is present.

B<file_name> = String

I<Required> Name of the file on which lock is to be checked.

=back

=cut

sub check_locks {
    my ( $node_info, $node_name, $file_name) = @_;

    $Log->comment("Checking locks on $file_name.");

    my $ngsh_cli = $node_info->{$node_name}->{ngshCLI};

    my $locks_leaked = 0;

    try {
        my $locks_resp = $ngsh_cli->vserver_locks_show('path' => $file_name);

        if($locks_resp->get_processed_output() =~ /no entries matching/) {
            $Log->comment("No locks leaked for object $file_name.");
        } else {
            $Log->comment("Locks leaked for object : $file_name.");
            $locks_leaked = 1;
        }
    }
    catch NACL::APISet::Exceptions::NoMatchingEntriesException with {

        $Log->comment("No locks leaked for object $file_name.");
    };

    if(0 != $locks_leaked) {
        return -1;
    }

    return 0;
} # end check_locks()

=head2 memleak_before_test

Checks the count of various rtaggged Gbuffers.

=over

=item Arguments

B<node_obj> = NACL::C::Node

I<Required> Node object on which the VServer is present.

B<Mem_Leak_Before> = Hash

I<Required> Stores the memleak output of command 'node run local rtag -t mbuf' in hash

=back

=cut

sub memleak_before_test {
    my ($node_info, $node_name) = @_;

    $Log->comment("Memleak Before TestCase");

    my $ngsh_cli = $node_info->{$node_name}->{ngshCLI};
    my $Mem_Leak_Before = $_[2];
    my $raw_output = $ngsh_cli->execute_raw_command(command => "node run local rtag -t mbuf");
    my $line;
    my $MemLeakKey;
    my $MemLeakValue;
    for (keys %$Mem_Leak_Before) {
        delete $Mem_Leak_Before->{$_};
    }
    foreach $line (split('\n',$raw_output)) {
        chomp($line);
        ($MemLeakKey, $MemLeakValue) = split(' ',$line);
        if( (!defined($MemLeakKey)) or (!defined($MemLeakValue)) ) {
            next;
        }
        if (($MemLeakKey eq "") or ($MemLeakKey eq "Module:") or ($MemLeakKey eq "Total:")) {
            next;
        }
        $Mem_Leak_Before->{$MemLeakKey} = $MemLeakValue;
    }
    return 0;
} # end memleak_before_test()

=head2 memleak_after_test

Checks the count of various rtaggged Gbuffers and evaluates diff with passed hash

=over

=item Arguments

B<node_obj> = NACL::C::Node

I<Required> Node object on which the VServer is present.

B<Mem_Leak_Before> = Hash

I<Required> Stores the memleak output of command 'node run local rtag -t mbuf' in hash

=back

=cut

sub memleak_after_test {
    my ($node_info, $node_name) = @_;

    $Log->comment("Memleak After TestCase");

    my $ngsh_cli = $node_info->{$node_name}->{ngshCLI};
    my %Mem_Leak_Before = %{$_[2]};
    my $raw_output = $ngsh_cli->execute_raw_command(command => "node run local rtag -t mbuf");
    my $line;
    my $MemLeakKey;
    my $MemLeakValue;
    my $leak;
    my %Mem_Leak_After;

    for (keys %Mem_Leak_After) {
        delete $Mem_Leak_After{$_};
    }
    %Mem_Leak_After = ();

    foreach $line (split('\n',$raw_output)) {
        chomp($line);
        ($MemLeakKey, $MemLeakValue) = split(' ',$line);
        if( (!defined($MemLeakKey)) or (!defined($MemLeakValue)) ) {
            next;
        }
        if (($MemLeakKey eq "") or ($MemLeakKey eq "Module:") or ($MemLeakKey eq "Total:")) {
            next;
        }
        $Mem_Leak_After{$MemLeakKey} = $MemLeakValue;
    }
    for my $key ( keys %Mem_Leak_After ) {
        if (($key eq "Module:") or ($key eq "Total:")) {
            next;
        }
        my $valueBefore = $Mem_Leak_Before{$key};
        my $valueAfter = $Mem_Leak_After{$key};
        if( !defined($valueBefore)) {
            $valueBefore = 0;
        }
        if( !defined($valueAfter)) {
            $valueAfter = 0;
        }
        $leak = $valueAfter - $valueBefore;
        if ( $leak > 0 ) {
            $Log->comment("Error !! $key module leaked = $leak bufs");
        }
    }
    return 0;
} # memleak_after_test

=head2 verify_location_header

Verify the format of Location header in CDMI/NonCDMI create response

=over

=item Arguments

B<response> = HTTP::Response

    (Required) This should be an object of type HTTP::Response. Automatically returned
               by CR::CdmiObject module after receiving response from CDMI server.

B<expected_uri_scheme> = Scalar

    (Required) Expected URI scheme in the localtion header. Allowed values: http|https

B<expected_cdmi_port> = Scalar

    (Required) Expected CDMI port in the localtion header.

B<expected_dir> = Scalar

    (Optional) Expected OID directory. Default value 'cdmi_objectid'

=back

=cut

sub verify_location_header {
    my $args                = shift;
    my $response            = $args->{response};
    my $expected_uri_scheme = $args->{expected_uri_scheme};
    my $expected_cdmi_port  = $args->{expected_cdmi_port};
    my $expected_dir        = $args->{expected_dir} || 'cdmi_objectid';

    ## Extract location header from response object
    my $location = $response->{'_headers'}->{'location'};
    $Log->comment('Info: The Location is ' . $location);

    ## Split the Location URI
    my ($scheme, $auth, $path, $query, $frag) = uri_split($location);

    ## Validate URI scheme
    if ($scheme ne $expected_uri_scheme) {
        $Log->comment('Error: URI scheme ['
            . $scheme . '] does not match with expected scheme ['
            . $expected_uri_scheme . ']');
        return 0;
    }
    $Log->comment('Info: URI scheme ['
            . $scheme . '] matches with expected scheme ['
            . $expected_uri_scheme . ']');

    ## Extract IP and Port
    my ($ip, $port) = split(/:/, $auth);
    ## Validate the IP address
    my $validate_ip_add = Data::Validate::IP->new();
    if (!$validate_ip_add->is_ipv4($ip)) {
        $Log->comment('Error: IP in the localtion header ['
            . $ip . '] is NOT a valid IPv4 address');
        return 0;
    }
    $Log->comment('Info: IP in the localtion header ['
            . $ip . '] is a valid IPv4 address');

    ## Validate CDMI port
    if ($port ne $expected_cdmi_port) {
        $Log->comment('Error: CDMI port ['
        . $port .'] does NOT match with expected port ['
        . $expected_cdmi_port . ']');
        return 0;
    }
    $Log->comment('Info: CDMI port ['
        . $port .'] matches with expected port ['
        . $expected_cdmi_port . ']');

    my ($empty, $dir, $oid) = split(/\//, $path);
    ## Validate OID directory
    if ($dir ne $expected_dir) {
        $Log->comment('Error: OID directory ['
        . $dir . '] does not match with expected directory ['
        . $expected_dir . ']');
        return 0;
    }
    $Log->comment('Info: OID directory ['
        . $dir . '] matches with expected directory ['
        . $expected_dir . ']');

    ## Validate OID
    if ($oid !~ /[a-zA-Z0-9]/) {
        $Log->comment('Error: OID [' . $oid . '] is not valid');
        return 0;
    }
    $Log->comment('Info: OID [' . $oid . '] is a valid one');

    $Log->comment('Info: Location header is correct');
    return 1;
}

1;
################################################################################
# Lib Ends
################################################################################
