## Copyright (c) 2012 NetApp Inc.
## All rights reserved.
##
## @name CRContainerObject.pm
##
## @summary Content repository container object representation
##
## @description
##   Represents the container object for content repository
##
## @keywords cr, http, containerobject
##
## @author Asif Hoda
##
## @example
##
## my $method = "PUT";
## my $uri = "http://uri_of_cdmi_server/NS/TestContainer";
## my $my_headers = {
##    "X-CDMI-Specification-Version"=>"1.0.1",
##    "Host"=>"localhost",
##    "Content-Type"=>"application/cdmi-object",
##    "Accept"=>"application/cdmi-object"
## };
## my $httpbody =   {
##    "mimetype" => "text/plains",
##    "value" => "Hello CDMI World Updated!"
##
## my $my_http_request = new CRHttpRequest();
##
## $my_http_request->set_uri($uri);
## $my_http_request->set_method($method);
## $my_http_request->set_header($my_headers);
## $my_http_request->set_body($httpbody);
##
## $my_http_request->get_uri();
## $my_http_request->get_method();
## $my_http_request->get_header();
## $my_http_request->get_body();
##
## @burt
##    513215
##
## @status private
##
## @duration 1-2 seconds
##

=pod

=head1 Package Name

CR::CRContainerObject.PM

=head1 Author/Maintainer

Asif Hoda,Padmanabh Joshi

=head1 Description
Represents the container object for content repository

=cut

package CR::CRContainerObject;

# Compiler directives.
use strict;
use warnings;

use JSON;
use HTTP::Message;
use Hash::Flatten qw(:all);
use LWP::UserAgent;

use NATE::Log qw(log_global);my $Log = log_global();

use CR::CRHttpRequest;
use CR::CRCdmiUtils;
use CR::CRCdmiObject;

our @ISA = qw( CR::CRCdmiObject);

################################################################################
# Default values for object creation
################################################################################
my ($_def_create_method,  $_def_create_uri,
    $_def_create_headers, $_def_create_body
);

$_def_create_method  = "PUT";
$_def_create_uri     = "http://localhost:8081/NS/TestContainer";
$_def_create_headers = {
    "X-CDMI-Specification-Version" => "1.0.1",
    "Content-Type"                 => "application/cdmi-container",
    "Accept"                       => "application/cdmi-container"
};
$_def_create_body = {};
################################################################################
################################################################################
# Default values for object updation
################################################################################
my ($_def_update_method,  $_def_update_uri,
    $_def_update_headers, $_def_update_body
);

$_def_update_method  = "PUT";
$_def_update_uri     = $_def_create_uri;
$_def_update_headers = {
    "X-CDMI-Specification-Version" => "1.0.1",
    "Content-Type"                 => "application/cdmi-container"
};
$_def_update_body = {};
################################################################################
################################################################################
# Default values for object read
################################################################################
my ( $_def_read_method, $_def_read_uri, $_def_read_headers, $_def_read_body );

$_def_read_method  = "GET";
$_def_read_uri     = $_def_create_uri;
$_def_read_headers = {
    "X-CDMI-Specification-Version" => "1.0.1",
    "Accept"                       => "application/cdmi-container"
};
$_def_read_body = "";
################################################################################
################################################################################
# Default values for object delete
################################################################################
my ($_def_delete_method,  $_def_delete_uri,
    $_def_delete_headers, $_def_delete_body
);

$_def_delete_method  = "DELETE";
$_def_delete_uri     = $_def_create_uri;
$_def_delete_headers = { "X-CDMI-Specification-Version" => "1.0.1" };
$_def_delete_body    = "";
################################################################################

=head2 new

Creates a new container object.

Owns a CR:CRHttpRequest, HTTP::Response and LWP::UserAgent and blesses the hash

=over

=item Arguments

B<lwp> = I<LWP::UserAgent object>

I<Optional> User specified LWP::UserAgent object. The user supplied
LWP::UserAgent is used to send data to CDMI server. If not specified then the
default LWP::UserAgent is used.

=back



=cut

sub new {
    my ( $class, %args ) = @_;
    my $lwp         = $args{browser}     || undef;
    my $socket_info = $args{socket_info} || undef;

    # If user has paseed us an LWP::UserAgent then use that else default
    if ( !defined($lwp) ) {
        $lwp = new LWP::UserAgent;
    }

    my $self = {
        _socket                  => $socket_info,
        _http_request            => undef,
        _server_response         => undef,
        _browser                 => $lwp,
        _read_response_http_body => undef,
    };

    bless $self, $class;
} ## end sub new

sub destroy {
    my ($self) = shift;
    $self->{_socket}                  = undef;
    $self->{_server_response}         = undef;
    $self->{_http_request}            = undef;
    $self->{_browser}                 = undef;
    $self->{_read_response_http_body} = undef;
} ## end sub destroy

=head2 create

Creates a CDMI or a non CDMI container object.

=over

=item Arguments

B<http_request> = I<HTTP::Request object>

I<Optional> User specified HTTP::Request object. The values from this object
will override the default container object creation parameters i.e http headers,
http body, http method and uri

=back

=cut

sub create {
    $Log->enter();
    my ( $self, $http_request ) = @_;
    $Log->exit();
    return $self->_issue_request($http_request);
} ## end sub create

=head2 update

Updates a CDMI or a non CDMI container object.

=over

=item Arguments

B<http_request> = I<HTTP::Request object>

I<Optional> User specified HTTP::Request object. The values from this object
will override the default container object updation parameters i.e http headers,
http body, http method and uri

=back

=cut

sub update {
    $Log->enter();
    my ( $self, $http_request ) = @_;
    $Log->exit();
    return $self->_issue_request($http_request);
} ## end sub update

=head2 read

Reads a CDMI or a non CDMI container object.

=over

=item Arguments

B<http_request> = I<HTTP::Request object>

I<Optional> User specified HTTP::Request object. The values from this object
will override the default container object read parameters i.e http headers,
http body, http method and uri

=back

=cut

sub read {
    $Log->enter();
    my ( $self, $http_request, $read_response_http_body ) = @_;

    if ($read_response_http_body) {
        $self->{_read_response_http_body} = $read_response_http_body;
    }

    $Log->exit();
    return $self->_issue_request($http_request);
} ## end sub read

=head2 delete

Delets a CDMI or a non CDMI container object.

=over

=item Arguments

B<http_request> = I<HTTP::Request object>

I<Optional> User specified HTTP::Request object. The values from this object
will override the default container object delete parameters

=back

=cut

sub delete {
    $Log->enter();
    my ( $self, $http_request ) = @_;
    $Log->exit();
    return $self->_issue_request($http_request);
} ## end sub delete

sub _issue_request {
    my ( $self, $http_request ) = @_;
    my ( $_def_method, $_def_uri, $_def_headers, $_def_body );

    # Set the defaults based on who called us
    my ($caller) = ( caller(1) )[3];
    if ( $caller eq "CR::CRContainerObject::create" ) {
        $_def_method  = $_def_create_method;
        $_def_uri     = $_def_create_uri;
        $_def_headers = $_def_create_headers;
        $_def_body    = $_def_create_body;
    } elsif ( $caller eq "CR::CRContainerObject::update" ) {
        $_def_method  = $_def_update_method;
        $_def_uri     = $_def_update_uri;
        $_def_headers = $_def_update_headers;
        $_def_body    = $_def_update_body;
    } elsif ( $caller eq "CR::CRContainerObject::read" ) {
        $_def_method  = $_def_read_method;
        $_def_uri     = $_def_read_uri;
        $_def_headers = $_def_read_headers;
        $_def_body    = $_def_read_body;
    } elsif ( $caller eq "CR::CRContainerObject::delete" ) {
        $_def_method  = $_def_delete_method;
        $_def_uri     = $_def_delete_uri;
        $_def_headers = $_def_delete_headers;
        $_def_body    = $_def_delete_body;
    } ## end elsif ( $caller eq "CR::CRContainerObject::delete")

    if ( "CR::CRHttpRequest" eq ref($http_request) ) {

        # user wants to override the defaults
        if ( "UNINIT" eq $http_request->get_method() ) {
            $http_request->set_method($_def_method);

        }
        if ( "UNINIT" eq $http_request->get_uri() ) {
            $http_request->set_uri($_def_uri);

        }

        if ( "UNINIT" eq $http_request->get_header() ) {
            $http_request->set_header($_def_headers);

        } else {
            for my $key ( $http_request->header_field_names() ) {
                print $key, " : ", $http_request->header($key);
                if (   ( $key eq "Authorization" )
                    && ( ( my $method = $http_request->method() ) ne "PUT" ) )
                {
                    $http_request->set_header($_def_headers);
                    last;
                } ## end if ( ( $key eq "Authorization"...
            } ## end for my $key ( $http_request...
        } ## end else [ if ( "UNINIT" eq $http_request...

        if ( "UNINIT" eq $http_request->get_body() ) {
            $http_request->set_body($_def_body);
        }
    } else {
        $Log->trace(
            "Using the default parameters for container object operations");
        $http_request = new CR::CRHttpRequest();
        $http_request->set_method($_def_method);
        $http_request->set_uri($_def_uri);
        $http_request->set_header($_def_headers);
        $http_request->set_body($_def_body);
    } ## end else [ if ( "CR::CRHttpRequest"...

    if ( defined( $http_request->header("ResetHeaders") ) ) {

      # We are checking if the user has set {} as the headers. In case user
      # sets {} as the headers then user intends to send no headers to server.
      # CR::CRHttpRequest::set_header when sees {} as the headers to be set,
      # inserts a dummy header by name "ResetHeaders" and value as "True". We
      # check if the header is present, and in case it is, then clear all the
      # headers.
        $http_request->clear;
    } ## end if ( defined( $http_request...

    $self->{_http_request} = $http_request;

    if ( defined $self->{_socket} ) {
        $self->_write_to_socket();
    } else {
        $self->_send();
    }

    return $self->{_server_response};
} ## end sub _issue_request

1;
################################################################################
# Lib Ends
################################################################################
