## Copyright (c) 2012 NetApp Inc.
## All rights reserved.
##
## @name CRHttpRequest.pm
##
## @summary HTTP request class for content repository objects
##
## @description
##   Represents the HTTP request that is sent to the CDMI server under test.
##
## @keywords cr, http
##
## @author Asif Hoda
##
## @example
##
## my $browser = new LWP::UserAgent;
## my $data_object = new CRDataObject();
## $data_object->create($browser);
## my $verify_create_httpheader = {
##     "X-CDMI-Specification-Version" => "1.0",
##     "Content-Type"=> "application/cdmi-object"
## };
## $data_object->verify_headers($verify_create_httpheader);
##
## my $verify_create_httpbody = {
##     "completionStatus"=> "Complete",
##     "mimetype"=> "text/plain",
##     "objectID"=> "",
##     "objectName"=> "",
##     "capabilitiesURI"=> "",
##     "domainURI"=> "",
##     "objectURI"=> "/TestDataObject.txt",
##     "parentURI"=> "/"
## };
##
## $data_object->verify_body($verify_create_httpbody);
## $data_object->update($browser,undef,$json_file);
## $data_object->read($browser);
## $data_object->delete($browser);
##
## @burt
##    513215
##
## @status private
##
## @duration 1-2 seconds
##

=pod

=head1 Package Name

CR::CRHttpRequest.PM

=head1 Author/Maintainer

Asif Hoda,Padmanabh Joshi

=head1 Description
Represents the HTTP request that is sent to the CDMI server under test.

=head2 Base Class

HTTP::Request

=cut

package CR::CRHttpRequest;

# Compiler directives.
use strict;
use warnings;
use Readonly;

use HTTP::Request;
use HTTP::Headers;
use JSON;
use Params::Validate qw[:all];
use NATE::Log qw(log_global);my $Log = log_global();
use Fcntl qw(SEEK_SET SEEK_END);
use MIME::Base64;
use URI::Escape;

our @ISA = qw( HTTP::Request Exporter);

our @EXPORT = qw($expected_missing_authentication_code
  $expected_object_created_code
  $expected_invalid_parameter_code
  $expected_object_updated_code
  $expected_object_read_code
  $expected_object_delete_code
  $expected_bad_request_code
  $expected_forbidden_code
  $expected_not_found_code
  $expected_method_not_allowed
  $expected_request_timed_out_code
  $expected_not_acceptable_code
  $expected_entity_too_large_code
  $expected_length_required
);

Readonly::Scalar our $expected_object_read_code            => "200";
Readonly::Scalar our $expected_object_created_code         => "201";
Readonly::Scalar our $expected_object_updated_code         => "204";
Readonly::Scalar our $expected_object_delete_code          => "204";
Readonly::Scalar our $expected_invalid_parameter_code      => "400";
Readonly::Scalar our $expected_bad_request_code            => "400";
Readonly::Scalar our $expected_missing_authentication_code => "401";
Readonly::Scalar our $expected_forbidden_code              => "403";
Readonly::Scalar our $expected_not_found_code              => "404";
Readonly::Scalar our $expected_method_not_allowed          => "405";
Readonly::Scalar our $expected_not_acceptable_code         => "406";
Readonly::Scalar our $expected_request_timed_out_code      => "408";
Readonly::Scalar our $expected_entity_too_large_code       => "413";
Readonly::Scalar our $expected_length_required             => "411";

sub new {
    my ( $class, $http_method, $http_uri, $http_header, $http_content ) = @_;
    my $self = $class->SUPER::new();
    bless $self, $class;

    # Post blessing processing for c'tor overload
    if ($http_method) {
        $self->set_method($http_method);
    }

    if ($http_uri) {
        $self->set_uri($http_uri);
    }

    if ($http_header) {
        $self->set_header($http_header);
    }

    if ($http_content) {
        $self->set_body($http_content);
    }

    return $self;
} ## end sub new

sub destroy {

    # For future use
}

sub set_header {
    my ( $self, $input_headers ) = @_;
    if ( ref($input_headers) ne "HASH" ) {
        $Log->error("http headers must be a hash ref");
        return -1;
    }

    if ( !( scalar keys %$input_headers ) ) {

        # Empty hash, user doesn't want to set any headers
        #
        # Following is a dummy key value inserted to the hash so that get_header
        # may check for this dummy header and return appropriately to indicate
        # that user doesn't want to send the default headers, rather user
        # doesn't want to send the default headers at all
        $self->header( "ResetHeaders" => "True" );
        return 1;
    } ## end if ( !( scalar keys %$input_headers...

    # Add headers to http request
    while ( ( my $header, my $value ) = each %$input_headers ) {
        $self->push_header( $header => $value );
        $Log->trace( "Setting header : " . $header . " to value : " . $value );
    }
    return 1;
} ## end sub set_header

sub get_header {
    my ($self) = @_;
    if ( $self->headers->as_string ne "" ) {
        return $self->headers;

    }
    return "UNINIT";
} ## end sub get_header

sub set_body {

    #Sets $_http_body
    my ( $self, $http_body, $chunk_sizes, $cdmi_http_body_order ) = @_;

    # $chunk_sizes will bring in an arrayref with various user supplied chunked
    # sizes. Among them a randomly chosen chunk size will be used. The
    # randomized choice of chunk size is made while creating every HTTP body
    # chunks. In effect, if the $chunk_sizes arrayref has more than one element,
    # then HTTP body being sent to CDMI server might have chunks with varying
    # chunk sizes.
    # Check if the user passed REF type is ARRAY else error

    if ( ref($chunk_sizes) ne "ARRAY" && defined $chunk_sizes ) {
        $Log->trace( "Chunk size expected to be type ARRAYREF, passed REF is of"
              . " type "
              . ref($chunk_sizes) );
        return -1;
    }

    if ( !$chunk_sizes ) {

        # user has not passed the chunked size
        #
        # chunk_size is required is any of the following cases
        # a. user has set a file name in HTTP body for non CDMI request
        # b. user has set a file name in "value" field of JSON body in CDMI
        #    request
        #
        # in the above cases if user has failed to set the chunk size then we
        # set the same to 2048, which is a arbit value.
        #
        # if user has sent a $chunk_sizes we respect that
        $chunk_sizes = [2048];
    }

    # $cdmi_http_body_order will bring in an arrayref with an ordered list of
    # request message body field names. We will abide the order and will assure
    # that $http_body, if a hash, will be stringified and send over the wire
    # with request message body items in order as specified in
    # $cdmi_http_body_order.
    if ( !( defined $cdmi_http_body_order ) ) {
        $cdmi_http_body_order = [
            'valuetransferencoding', 'move',
            'copy',                  'mimetype',
            'metadata',              'valuerange',
            'value'
        ];
    }

    if ( ref($cdmi_http_body_order) ne "ARRAY" ) {
        $Log->trace( "CDMI http body order expected to be type ARRAYREF, passed"
              . " REF is of type "
              . ref($cdmi_http_body_order) );
        return -1;
    }

    if ($http_body) {

        # Find the type
        if ( ref($http_body) eq "HASH" ) {

            # Peek into the hash to see if the "value" field is a file path
            if ( exists $http_body->{value} ) {

                # Check if the "value" is a openable file
                if ( -e $http_body->{value} ) {

                # Test if the file is a bin file
                #
                # The tests -T and -B determine if a file is text or binary.
                # But people who know a lot about filesystems know there’s no
                # bit (at least in Unix-like operating systems) to indicate
                # that a file is a binary or text file, so how can Perl tell?
                # The answer is that Perl cheats: it opens the file, looks at
                # the first few thousand bytes, and makes an educated guess.
                # If it sees a lot of null bytes, unusual control characters,
                # and bytes with the high bit set, then that looks like a
                # binary file. If there’s not much weird stuff, then it looks
                # like text. It sometimes guesses wrong. If a text file has a
                # lot of Swedish or French words (which may have characters
                # represented with the high bit set, as some
                # ISO-8859-something variant, or perhaps even a Unicode
                # version), it may fool Perl into declaring it binary. So it’s
                # not perfect, but if you need to separate your source code
                # from compiled files, or HTML files from PNGs, these tests
                # should do the trick

                    my $_created_http_body_file = 0;
                    if ( -B $http_body->{value} ) {

                        # It's a bin file
                        $_created_http_body_file =
                          $self->_create_http_body_file( $http_body, "binary" );
                    }
                    else {

                        # It's a text file
                        $_created_http_body_file =
                          $self->_create_http_body_file( $http_body, "text" );
                    }
                    $self->content(
                        $self->_send_chunked(
                            $_created_http_body_file, $chunk_sizes
                        )
                    );
                }
                else {
                    $self->content(
                        $self->_order_http_body(
                            $http_body, $cdmi_http_body_order
                        )
                    );
                }
            }
            else {

                # otherwise it's just a plain hash
                $http_body = to_json($http_body);
                $self->content($http_body);
            }
        }
        elsif ( -e $http_body ) {

            # $http_body is a file path
            $self->content( $self->_send_chunked( $http_body, $chunk_sizes ) );
        }
        else {

            # $http_body is a perl string most probably for non CDMI request
            $self->content($http_body);
        }
    } ## end if ($http_body)

    return 1;
} ## end sub set_body

sub get_body {
    my ($self) = @_;
    if ( $self->content eq "" ) {
        return "UNINIT";
    }
    return $self->content;
} ## end sub get_body

sub set_method {
    my ( $self, $http_method ) = @_;

    $self->method($http_method);
    return 1;
} ## end sub set_method

sub get_method {
    my ($self) = @_;
    if ( !defined( $self->method ) ) {
        return "UNINIT";
    }
    return $self->method;
} ## end sub get_method

sub set_uri {
    my ( $self, $http_uri ) = @_;

    $self->uri($http_uri);
    return 1;
} ## end sub set_uri

sub get_uri {
    my ($self) = @_;
    if ( !defined( $self->uri ) ) {
        return "UNINIT";
    }
    return $self->uri;
} ## end sub get_uri

##This function sets the basic authentication for each HTTP request when called .
sub set_credentials {
    my $self              = shift;
    my %input_credentials = validate(
        @_,
        {
            'username' => {
                type     => SCALAR,
                optional => 0,
            },
            'password' => {
                type     => SCALAR,
                optional => 0,
            }
        }
    );

    #Add credentials to HTTP request object
    my $_uname = $input_credentials{username};
    my $_pass  = $input_credentials{password};
    $self->authorization_basic( $_uname, $_pass );
    $Log->trace( "Setting credentials :" . $_uname . "::" . $_pass );
    return 1;
} ## end sub set_credentials

################################################################################
##                               PRIVATE METHODS                              ##
################################################################################
our $offset = 0;

sub _send_chunked {

    my ( $self, $file_name, $chunk_sizes ) = @_;

    return sub {
        open( FILE, $file_name ) or die "can't open $file_name: $!";
        binmode(FILE);

        seek( FILE, $offset, SEEK_SET );
        my $buff;
        my $read = read( FILE, $buff, $chunk_sizes->[ rand @$chunk_sizes ] );

        $offset = $offset + $read;
        close(FILE);

        if ( !$buff ) {
            $offset = 0;
        }
        return $buff;
      }
} ## end sub _send_chunked

sub _create_http_body_file {
    my ( $self, $http_body, $encoding ) = @_;

    # CDMI spec mandates
    # "The request and response body JSON fields may be specified or returned in
    # any order, with the exception that, if present, for data objects, the
    # valuerange and value fields shall appear last and in that order."
    #
    # Hence the following:
    #
    # 1. Keep the items to be ordered in the final JSON body in this array
    # 2. Delete the kept item from HTTP body
    # 3. Write the HTTP body to tmp file
    # 4. Insert the _last_elem in the temp file

    # 1. Keep the items to be ordered in the final JSON body in this array
    my $_last_elem = [];

    if ( exists $http_body->{value} ) {
        push( @$_last_elem, ( "value", $http_body->{value} ) );

        # 2. Delete the kept item from HTTP body
        delete $http_body->{value};
    }

    # 3. Write the HTTP body to tmp file
    # Now $http_body has the value key and value removed
    # will write this to the file
    $http_body = to_json($http_body);
    chop($http_body);
    $http_body = $http_body . ",\"" . $_last_elem->[0] . "\"\:\"";

    my $_file = "./testdata/tmp.upload." . rand() . ".txt";
    $Log->trace( "Temp HTTP request file name [" . $_file . "]" );

    # Writing the temp file
    open FILE, ">$_file" or die "can't open $_file: $!";
    print FILE $http_body;
    seek( FILE, 0, SEEK_END );

    # 4. Insert the _last_elem in the temp file
    open( IPFILE, $_last_elem->[1] ) or die "can't open $_last_elem->[0]: $!";
    binmode(IPFILE);
    my $buff;
    while ( read( IPFILE, $buff, 1024 ) ) {
        if ( $encoding eq "binary" ) {
            $buff = encode_base64($buff);
        }
        print FILE $buff;
        seek( FILE, 0, SEEK_END );
    }

    # Complete the JSON
    print FILE "\"\}";

    # Close the files
    close IPFILE;
    close FILE;

    # Return the file name of the temp file that we just wrote
    # This file content will be chunked in most likelihood
    return $_file;

} ## end sub _create_http_body_file

sub _order_http_body {

    # --------------------------------------------------------------------------
    # CDMI specification mandates, when present, the following fields position
    # in JSON formatted HTTP body will appear as per the mentioned rules.
    # --------------------------------------------------------------------------
    # Field names:
    # --------------------------------------------------------------------------
    # 1) valuetransferencoding
    # 2) move
    # 3) copy
    # 4) mimetype
    # 5) metadata
    # 6) valuerange
    # 7) value
    # --------------------------------------------------------------------------
    # Rules:
    # --------------------------------------------------------------------------
    # Rule 1:
    # (1), (2), (3) & (4) MAY come in any order but MUST come before (5).
    # --------------------------------------------------------------------------
    # Rule 2:
    # (6) & (7) should be penultimate & last fields respectively.
    # --------------------------------------------------------------------------

    # Ensuring the mandate

    # Get the agrs
    my ( $self, $hash, $order ) = @_;
    my $staging_hash = {};

    # Delete the keys present in $order from $hash
    foreach my $order_key (@$order) {
        if ( exists $hash->{$order_key} ) {
            $staging_hash->{$order_key} = $hash->{$order_key};
            delete $hash->{$order_key};
        }
    }

    my $hash_string = encode_json($hash);

    # Strip the last }
    my $last_brace_index = rindex( $hash_string, '}' );
    substr( $hash_string, $last_brace_index, 1, '' );

    # If the hash has been emptied then don't put the leading ,
    if ( $hash_string ne '{' ) {
        $hash_string .= ',';
    }

    # Insert the ordered key-value pair
    foreach my $order_key (@$order) {
        if ( exists $staging_hash->{$order_key} ) {
            if ( ref( $staging_hash->{$order_key} ) ) {
                $hash_string .= '"'
                  . $order_key . '":'
                  . encode_json( $staging_hash->{$order_key} ) . ',';
            }
            else {
                $hash_string .=
                  '"' . $order_key . '":"' . $staging_hash->{$order_key} . '",';
            }
        }
    }

    # Add the last } back
    my $last_comma_index = rindex( $hash_string, ',' );
    substr( $hash_string, $last_comma_index, 1, '}' );

    return $hash_string;
} ## end sub _order_http_body

1;
################################################################################
# Lib Ends
################################################################################
