## Copyright (c) 2011 NetApp Inc.
## All rights reserved.
##
## @name CRLargeDataObject.pm
##
## @summary Content repository Large data object CRUD methods
##
## @description
##   Provides all Create,Update,Read and Delete methods for
##       Large data object
##
## @keywords Largedataobject
##
## @author   Padmanabh Joshi
##
## @example
##
## my $result = $data_object->create(
##          'CDMI_CONTENT'    => $Create_Datafile,
##          'CDMI_DATALIF_IP' => $DATALIF,
##          'CDMIMODE' => 1
##         );
##
##
##
##
## @status private
##
## @duration
##

=pod

=head1 Package Name

  CRLargeDataObject . PM

=head1 Author / Maintainer

  pjoshi @netapp . com, dl-cr-obj-store @netapp . com

=head1 Description Represents of the large data object

=cut

package CR::CRLargeDataObject;

# Compiler directives.
use strict;
use warnings;
use Tharn qw($Log);

use LWP;
use HTTP::Request;
use Params::Validate qw[:all];
use JSON;
use CR::CRCdmiObject;
use MIME::Base64;

#use HTTP::Headers;
our @ISA = qw( CR::CRCdmiObject);

my $fh;
################################################################################
##Creating object using  defualt constructor
sub new {
    my ($class) = @_;
    my $self = {
        _http_request    => undef,
        _server_response => undef
    };
    bless $self, $class;
} ## end sub new

sub destroy {
    my ($self) = shift;
    $self->{_server_response} = undef;
    $self->{_http_request}    = undef;
}

##Create Data object Method
sub create {
    my $self          = shift;
    my %object_params = validate(
        @_,
        {   'CDMIMODE' => {
                type     => SCALAR,
                optional => 0
            },
            'CDMI_PORT' => {
                type     => SCALAR,
                optional => 1,
                default  => '8081'
            },
            'Accept' => {
                type     => SCALAR,
                optional => 1,
                default  => 'application/cdmi-object'
            },
            'Content-Type' => {
                type     => SCALAR,
                optional => 1,
                default  => 'application/cdmi-object'
            },
            'X-CDMI-Specification-Version' => {
                type     => SCALAR,
                optional => 1,
                default  => '1.0.1'
            },
            'CDMI_TRANSPORT' => {
                type     => SCALAR,
                optional => 1,
                default  => 'http'
            },
            'CDMI_METHOD' => {
                type     => SCALAR,
                optional => 1,
                default  => 'POST'
            },
            'CDMI_DATALIF_IP' => {
                type     => SCALAR,
                optional => 0
            },
            'PARENT_URI' => {
                type     => SCALAR,
                optional => 1,
                default  => 'cdmi_objectid'
            },
            'Content-Length' => {
                type     => SCALAR,
                optional => 1,
                default  => '0'
            },
            'CDMI_CONTENT' => {
                type     => SCALAR,
                optional => 1,
                default =>
                    '{"mimetype" : "text/plain", "metadata" : {}, "value" :  HI THIS IS DEFAULT DATA OF CDMI DATA OBJECT}'
            },
            'ADD_HEADERS' => {
                type     => HASHREF,
                optional => 1
                }

        }
    );
    my $cdmicontent;
    my $filename;
    my $http_req;
    my $readfunc;
    my $hdr = HTTP::Headers->new();

    #Forming the URI Path
    my $transport     = $object_params{'CDMI_TRANSPORT'};
    my $method        = $object_params{'CDMI_METHOD'};
    my $datalifip     = $object_params{'CDMI_DATALIF_IP'};
    my $cdmiport      = $object_params{'CDMI_PORT'};
    my $parenturi     = $object_params{'PARENT_URI'};
    my $contentlength = $object_params{'Content-Length'};
    my $addheader     = $object_params{'ADD_HEADERS'};
    my $cdmimode      = $object_params{'CDMIMODE'};

    my $uri
        = $transport . "\:" . "//"
        . $datalifip . "\:"
        . $cdmiport . "/"
        . $parenturi . "/";

    #Forming the Header section
    my $accept      = $object_params{'Accept'};
    my $contenttype = $object_params{'Content-Type'};
    my $cdmispec    = $object_params{'X-CDMI-Specification-Version'};

    ##Forming Header Object.  'x', indicates that we should exclude the header
    ## to test -ve scenario.
    if ( $datalifip ne 'X' ) {
        $hdr->header( Host => $datalifip );
    }
    if ( ( $accept ne 'X' ) && ($cdmimode) ) {
        $hdr->header( Accept => $accept );
    }
    if ( $contenttype ne 'X' ) {

        if ($cdmimode) {
            $hdr->header( Content_Type => $contenttype );
        } else {
            $contenttype = "text/plain";
            $hdr->header( Content_Type => $contenttype );
        }
    } ## end if ( $contenttype ne 'X')

    if ( ( $cdmispec ne 'X' ) && ($cdmimode) ) {
        $hdr->header( X_CDMI_Specification_Version => $cdmispec );
    }

    #Putting additional header if any for negative test cases.
    if ( defined(%$addheader) ) {
        my %temphash = %$addheader;
        foreach my $field ( keys %temphash ) {
            $hdr->push_header( $field => $temphash{$field} );

        }
    } ## end if ( defined(%$addheader...

    #Getting Content of Request
    if ( !( -e $object_params{'CDMI_CONTENT'} ) ) {
        $Log->comment( $object_params{'CDMI_CONTENT'}
                . " : Input file not found, Hence exiting the test\n" );
        return $TCD::CONFIG;

    } else {

        #Reading the input File
        $filename = $object_params{'CDMI_CONTENT'};
        open( $fh, '<', $filename )
            or die "Couldn't open file: $!";
        $readfunc = sub {
            read( $fh, my $buf, 1024 );
            return $buf;
        };

        $http_req = HTTP::Request->new( $method, $uri, $hdr, $readfunc );

    } ## end else [ if ( !( -e $object_params...

    my $browser = new LWP::UserAgent;

    #$Log->comment( "HTTP Request\n" . $http_req->as_string());
    $self->{_server_response} = $browser->request($http_req);

    $Log->comment(
        " #######################PRINTING: Request  START####################\n"
    );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment( $http_req->as_string() );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment(
        " #######################PRINTING: Request  END#######################\n"
    );

    $Log->comment(
        " HTTP Status :=========>" . $self->{_server_response}->code() );
    $Log->comment(
        "###############------------PRINTING: Response AS STRING START######################\n"
    );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment( $self->{_server_response}->as_string() );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment(
        "######################PRINTING: Response  AS STRING END########################\n"
    );

    return ( $self->{_server_response} );
} ## end sub create

##Update Data object Method
sub update {
    my $self          = shift;
    my %object_params = validate(
        @_,
        {   'CDMI_PORT' => {
                type     => SCALAR,
                optional => 1,
                default  => '8081'
            },
            'CDMIMODE' => {
                type     => SCALAR,
                optional => 0,
            },
            'Accept' => {
                type     => SCALAR,
                optional => 1,
                default  => 'application/cdmi-object'
            },
            'Content-Type' => {
                type     => SCALAR,
                optional => 1,
                default  => 'application/cdmi-object'
            },
            'X-CDMI-Specification-Version' => {
                type     => SCALAR,
                optional => 1,
                default  => '1.0.1'
            },
            'Content-Length' => {
                type     => SCALAR,
                optional => 1,
                default  => '0'
            },
            'CDMI_METHOD' => {
                type     => SCALAR,
                optional => 1,
                default  => 'PUT'
            },
            'CDMI_DATALIF_IP' => {
                type     => SCALAR,
                optional => 0,
            },
            'PARENT_URI' => {
                type     => SCALAR,
                optional => 1,
                default  => 'cdmi_objectid'
            },
            'CDMI_OBJECTID' => {
                type     => SCALAR,
                optional => 0,
            },
            'CDMI_CONTENT' => {
                type     => SCALAR,
                optional => 0,
                default =>
                    '{"mimetype" : "text/plain", "metadata" : {}, "value" : HI THIS IS DEFAULT DATA OF CDMI DATA UPDATE OBJECT}'
            },
            'ADD_HEADERS' => {
                type     => HASHREF,
                optional => 1,
            },
            'CDMI_TRANSPORT' => {
                type     => SCALAR,
                optional => 1,
                default  => 'http'
                }

        }
    );
    my $cdmicontent;
    my $filename;
    my $http_req;
    my $readfunc;
    my $hdr = HTTP::Headers->new();

    #Forming the uri Path
    my $transport     = $object_params{'CDMI_TRANSPORT'};
    my $method        = $object_params{'CDMI_METHOD'};
    my $datalifip     = $object_params{'CDMI_DATALIF_IP'};
    my $cdmiport      = $object_params{'CDMI_PORT'};
    my $parenturi     = $object_params{'PARENT_URI'};
    my $contentlength = $object_params{'Content-Length'};
    my $cdmiobjectid  = $object_params{'CDMI_OBJECTID'};
    my $addheader     = $object_params{'ADD_HEADERS'};
    my $cdmimode      = $object_params{'CDMIMODE'};

    my $uri
        = $transport . "\:" . "//"
        . $datalifip . "\:"
        . $cdmiport . "/"
        . $parenturi . "/"
        . $cdmiobjectid;
    $Log->comment( "uri ====>:" . $uri );

    #Forming the Header section
    my $accept      = $object_params{'Accept'};
    my $contenttype = $object_params{'Content-Type'};
    my $cdmispec    = $object_params{'X-CDMI-Specification-Version'};

    ##Forming Header Object.  'x', indicates that we should exclude the header to test -ve scenario.
    if ( $datalifip ne 'X' ) {
        $hdr->header( Host => $datalifip );
    }
    if ( ( $accept ne 'X' ) && ($cdmimode) ) {
        $hdr->header( Accept => $accept );

    }
    if ( $contenttype ne 'X' ) {
        if ($cdmimode) {
            $hdr->header( Content_Type => $contenttype );
        } else {
            $contenttype = "text/plain";
            $hdr->header( Content_Type => $contenttype );
        }
    } ## end if ( $contenttype ne 'X')

    if ( ( $cdmispec ne 'X' ) && ($cdmimode) ) {
        $hdr->header( X_CDMI_Specification_Version => $cdmispec );
    }

    #Putting additional header if any for negative test cases.
    if ( defined(%$addheader) ) {
        my %temphash = %$addheader;
        foreach my $field ( keys %temphash ) {
            $hdr->push_header( $field => $temphash{$field} );
        }
    } ## end if ( defined(%$addheader...

    #Getting Content of Request
    if ( !( -e $object_params{'CDMI_CONTENT'} ) ) {
        $Log->comment( $object_params{'CDMI_CONTENT'}
                . " : Input file not found, Hence exiting the test\n" );
        return $TCD::CONFIG;

    } else {

        #Reading the input File
        $filename = $object_params{'CDMI_CONTENT'};
        open( $fh, '<', $filename )
            or die "Couldn't open file: $!";
        $readfunc = sub {
            read( $fh, my $buf, 65536 );
            return $buf;
        };
        $contentlength = -s ($filename);
        if ( $cdmispec ne 'X' ) {
            $hdr->header( 'Content-Length' => $contentlength );
        }
        $http_req = HTTP::Request->new( $method, $uri, $hdr, $readfunc );

    } ## end else [ if ( !( -e $object_params...

    #Sending  the HTTP request through User Agent browser
    my $browser = new LWP::UserAgent;
    $self->{_server_response} = $browser->request($http_req);

    $Log->comment(
        " #######################PRINTING: Request  START####################\n"
    );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment( $http_req->as_string() );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment(
        " #######################PRINTING: Request  END#######################\n"
    );

    $Log->comment(
        " HTTP Status :=========>" . $self->{_server_response}->code() );
    $Log->comment(
        "###############------------PRINTING: Response AS STRING START######################\n"
    );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment( $self->{_server_response}->as_string() );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment(
        "######################PRINTING: Response  AS STRING END########################\n"
    );

    return ( $self->{_server_response} );
} ## end sub update

##Read Data object Method
sub read {
    my $self          = shift;
    my %object_params = validate(
        @_,
        {   'CDMI_PORT' => {
                type     => SCALAR,
                optional => 1,
                default  => '8081'
            },
            'Accept' => {
                type     => SCALAR,
                optional => 1,
                default  => 'application/cdmi-object'
            },
            'Content-Type' => {
                type     => SCALAR,
                optional => 1,
                default  => 'application/cdmi-object'
            },
            'X-CDMI-Specification-Version' => {
                type     => SCALAR,
                optional => 1,
                default  => '1.0.1'
            },
            'Content-Length' => {
                type     => SCALAR,
                optional => 1,
                default  => '0'
            },
            'CDMI_METHOD' => {
                type     => SCALAR,
                optional => 1,
                default  => 'GET'
            },
            'CDMI_DATALIF_IP' => {
                type     => SCALAR,
                optional => 0,
            },
            'PARENT_URI' => {
                type     => SCALAR,
                optional => 1,
                default  => 'cdmi_objectid'
            },
            'CDMI_OBJECTID' => {
                type     => SCALAR,
                optional => 0,
                default  => '0'
            },
            'ADD_HEADERS' => {
                type     => HASHREF,
                optional => 1,

            },
            'CDMI_TRANSPORT' => {
                type     => SCALAR,
                optional => 1,
                default  => 'http'
            },
            'RESTOREFILE' => {
                type     => SCALAR,
                optional => 0,

            },
            'CDMIMODE' => {
                type     => SCALAR,
                optional => 0,
            },
            'ENCODETYPE' => {
                type     => SCALAR,
                optional => 1,
                default  => 'utf-8'
                }

        }
    );
    my $cdmicontent;
    my $filename;
    my $http_req;
    my $readfunc;
    my $hdr = HTTP::Headers->new();

    #Forming the uri Path
    my $transport     = $object_params{'CDMI_TRANSPORT'};
    my $method        = $object_params{'CDMI_METHOD'};
    my $datalifip     = $object_params{'CDMI_DATALIF_IP'};
    my $cdmiport      = $object_params{'CDMI_PORT'};
    my $parenturi     = $object_params{'PARENT_URI'};
    my $contentlength = $object_params{'Content-Length'};
    my $cdmiobjectid  = $object_params{'CDMI_OBJECTID'};
    my $addheader     = $object_params{'ADD_HEADERS'};
    my $cdmimode      = $object_params{'CDMIMODE'};
    my $encodetype    = $object_params{'ENCODETYPE'};

    my $uri
        = $transport . "\:" . "//"
        . $datalifip . "\:"
        . $cdmiport . "/"
        . $parenturi . "/"
        . $cdmiobjectid;
    $Log->comment( "uri ====>:" . $uri );

    #Forming the Header section
    my $accept      = $object_params{'Accept'};
    my $contenttype = $object_params{'Content-Type'};
    my $cdmispec    = $object_params{'X-CDMI-Specification-Version'};

    ##Forming Header Object.  'x', indicates that we should exclude the header to test -ve scenario.
    if ( $datalifip ne 'X' ) {
        $hdr->header( Host => $datalifip );
    }
    if ( ( $accept ne 'X' ) && ($cdmimode) ) {
        $hdr->header( Accept => $accept );

    }
    if ( ( $contenttype ne 'X' ) && ($cdmimode) ) {
        $hdr->header( Content_Type => $contenttype );
    }

    if ( ( $cdmispec ne 'X' ) && ($cdmimode) ) {
        $hdr->header( X_CDMI_Specification_Version => $cdmispec );
    }

    if ( ( $encodetype ne 'X' ) && ($cdmimode) ) {
        $hdr->header( valuetransferencoding => $encodetype );
    }

    #Putting additional header if any for negative test cases.
    if ( defined(%$addheader) ) {
        my %temphash = %$addheader;
        foreach my $field ( keys %temphash ) {
            $hdr->push_header( $field => $temphash{$field} );
        }
    } ## end if ( defined(%$addheader...

    #Reading the input File
    $filename = $object_params{'RESTOREFILE'};

    $http_req = HTTP::Request->new( $method, $uri, $hdr );

    #Sending  the HTTP request through User Agent browser
    my $browser = new LWP::UserAgent;
    $self->{_server_response} = $browser->request( $http_req, $filename );

    $Log->comment(
        " #######################PRINTING: Request  START####################\n"
    );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment( $http_req->as_string() );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment(
        " #######################PRINTING: Request  END#######################\n"
    );

    $Log->comment(
        " HTTP Status :=========>" . $self->{_server_response}->code() );
    $Log->comment(
        "###############------------PRINTING: Response AS STRING START######################\n"
    );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment( $self->{_server_response}->as_string() );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment(
        "######################PRINTING: Response  AS STRING END########################\n"
    );

    return ( $self->{_server_response} );
} ## end sub read

##Delete Data object Method
sub delete {
    my $self          = shift;
    my %object_params = validate(
        @_,
        {   'CDMI_PORT' => {
                type     => SCALAR,
                optional => 1,
                default  => '8081'
            },
            'CDMI_METHOD' => {
                type     => SCALAR,
                optional => 1,
                default  => 'DELETE'
            },
            'CDMI_DATALIF_IP' => {
                type     => SCALAR,
                optional => 0,
            },
            'PARENT_URI' => {
                type     => SCALAR,
                optional => 1,
                default  => 'cdmi_objectid'
            },
            'CDMI_OBJECTID' => {
                type     => SCALAR,
                optional => 0,
                default  => '0'
            },
            'ADD_HEADERS' => {
                type     => HASHREF,
                optional => 1,

            },
            'CDMI_TRANSPORT' => {
                type     => SCALAR,
                optional => 1,
                default  => 'http'
                }

        }
    );

    my $http_req;
    my $hdr = HTTP::Headers->new();

    #Forming the uri Path
    my $transport    = $object_params{'CDMI_TRANSPORT'};
    my $method       = $object_params{'CDMI_METHOD'};
    my $datalifip    = $object_params{'CDMI_DATALIF_IP'};
    my $cdmiport     = $object_params{'CDMI_PORT'};
    my $parenturi    = $object_params{'PARENT_URI'};
    my $cdmiobjectid = $object_params{'CDMI_OBJECTID'};
    my $addheader    = $object_params{'ADD_HEADERS'};

    my $uri
        = $transport . "\:" . "//"
        . $datalifip . "\:"
        . $cdmiport . "/"
        . $parenturi . "/"
        . $cdmiobjectid;
    $Log->comment( "uri ====>:" . $uri );

    #Putting additional header if any for negative test cases.
    if ( defined(%$addheader) ) {
        my %temphash = %$addheader;
        foreach my $field ( keys %temphash ) {
            $hdr->push_header( $field => $temphash{$field} );
        }
    } ## end if ( defined(%$addheader...

    $http_req = HTTP::Request->new( $method, $uri, $hdr );

    #Sending  the HTTP request through User Agent browser
    my $browser = new LWP::UserAgent;
    $self->{_server_response} = $browser->request($http_req);

    $Log->comment(
        " #######################PRINTING: Request  START####################\n"
    );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment( $http_req->as_string() );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment(
        " #######################PRINTING: Request  END#######################\n"
    );

    $Log->comment(
        " HTTP Status :=========>" . $self->{_server_response}->code() );
    $Log->comment(
        "###############------------PRINTING: Response AS STRING START######################\n"
    );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment( $self->{_server_response}->as_string() );
    $Log->comment(" \n");
    $Log->comment(" \n");
    $Log->comment(
        "######################PRINTING: Response  AS STRING END########################\n"
    );

    return ( $self->{_server_response} );
} ## end sub delete
##Delete Data object Method

##Base64 Encoding.
sub encode_to_base64 {
    my $self   = shift;
    my %params = validate(
        @_,
        {   'BASE64FILE' => {
                type     => SCALAR,
                optional => 0
            },
            'DATAFILE' => {
                type     => SCALAR,
                optional => 0
            },
            'DECODEFILE' => {
                type     => SCALAR,
                optional => 0
            },
            'CDMIMODE' => {
                type     => SCALAR,
                optional => 0
            }
        }
    );
    my $base64file = $params{'BASE64FILE'};
    my $datafile   = $params{'DATAFILE'};
    my $decodefile = $params{'DECODEFILE'};
    my $cdmimode   = $params{'CDMIMODE'};

    my ( $buf, $data, $data1, $n );

    ##Core logic of base64 encoding
    open( READFILE, '<', $datafile )
        or die "Couldn't open INPUT DATA FILE: $datafile $!";
    open( WRITEFILE, '>', $base64file )
        or die "Couldn't CREATE Base64 INPUT FILE: $base64file $!";
    open( WRITEFILE1, '>', $decodefile )
        or die "Couldn't CREATE DecodeBase64 INPUT FILE: $decodefile $!";
    my $prefix_string
        = "{\"mimetype\" \: \"text\/plain\",\"valuetransferencoding\"\: \"base64\",\"value\"\ :\"";
    my $post_string = "\"\}";
    binmode READFILE;

    if ($cdmimode) {
        print WRITEFILE $prefix_string;
    }
    while ( ( $n = read( READFILE, $buf, 60 * 57 ) ) != 0 ) {
        $data = encode_base64( $buf, "" );
        print WRITEFILE $data;
        $data1 = decode_base64($data);
        print WRITEFILE1 $data1;
    } ## end while ( ( $n = read( READFILE...
    if ($cdmimode) {
        print WRITEFILE $post_string;
    }
    close(WRITEFILE);
    close(WRITEFILE1);
    close(READFILE);

    $Log->comment(
        "###################### BASE64 ENCODING DONE########################\n"
    );
    return 1;
} ## end sub encode_to_base64

1;

