# Copyright (c) 2011 Network Appliance, Inc.
# All rights reserved.

## @summary unit test library
## @author  dl-crqa-auto@netapp.com

package CR::CRUtils;

use strict;
use warnings;
use Tharn;
use Options;
use CR::CommonParams;

use NACL::C::Volume;
use NACL::C::Vserver;
use NACL::STask::Aggregate;
use NACL::STask::Volume;
use NACL::STask::Snapmirror;
use NACL::C::Interface;
use NATE::BaseException;
use NACL::C::JobHistory;
use NACL::C::ClusterIdentity;
use NACL::C::ClusterPeer;
use NACL::C::ClusterPeerHealth;
use NACL::C::Snapmirror;
use NACL::C::Snapshot;
use Network::Trace;
use NACL::C::Node;
use NACL::C::Job;
use Connectrec qw();
use Hostrec qw();
use TestUtils;
use Data::Dumper;
use Data::Random qw( rand_words );
use Time::Local;
use NACL::APISet;
use NACL::C::VolumeCheck;
use NACL::C::DebugVreport;
use NACL::C::RepositoryVolumeConstituent;
use NACL::C::UnitNormalization;

use vars qw( $Options %SiteData $Site $Client_Trace_Obj $Api_Set_Obj $Api_Set_Obj_Client $Api_Set_Obj_Client_C $Tshark_pid $Log_directory);
our %_wordHits;

=head2 ::new

Creates a CR::CRUtils object and returns it.

=over

=item Arguments

=item C<<"connection"=>$Node >>

(Mandatory) A connection to one of the nodes. This will be
stored in the object while executing other methods if invoked
as instance methods.

=item Examples

=back

    my $util = CR::CRUtils->new( connection => $Node, )

=cut

sub new {
    my ($class, %args) = @_;
    my $self = {};
    bless $self, $class;
    $Options = new Options;
    $Options->process();
    $self->{_connection} = $args{connection};
    $self->{_nodes} = $self->get_all_nodes;
    %SiteData = TestUtils::get_site_info();        # From TestUtils
    return $self;
}

=head2 ::delete_repo

Delete a repository volume after unmounting and offlining it.

=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as instance method, otherwise necessary.

=item C<<"vserver"=>$vserver >>

(Optional) Name of the vserver for which the repository belongs to.
Defaults to "vs0" if not passed.

=item Examples

=back

Instance method:
    $CRUtils->delete_repo();

Class method:
    CR::CRUtils->delete_repo(
       command_interface => $Node,
       vserver => $vserver_name,
    );

=cut

sub delete_repo {
    my ($self, %args) = @_;
    my (%opts, @repos) ;
    my $connection = $args{connection} || $self->{_connection};
    my $vserver_name = $args{vserver} || $Options->get(option=>'VSERVER');
    @repos = NACL::C::Volume->find(
              command_interface => $connection,
              filter => { vserver => $vserver_name,
                          'volume-style' => 'infinitevol',
                        },
              allow_empty => 1,
            );
    foreach my $repo (@repos){
        #Delete the repository
        my $repo_type = $repo->get_one_state_attribute('type');
        #Unmount
        try{
            $repo->unmount(
            "method-timeout" => $Options->get(option=>"TIMEOUT")
            );
        }catch NACL::APISet::Exceptions::CommandFailedException with{
            my $exception_obj = shift;
            $Log->comment("$exception_obj->text()") unless ($repo_type eq 'DP');
        }catch NACL::APISet::Exceptions::ResponseException with{
            my $exception_obj = shift;
            $Log->comment("$exception_obj->text()") unless ($repo_type eq 'DP');
        };
        #Offline
        try{
            $repo->offline(
                "method-timeout" => $Options->get(option=>"TIMEOUT"),
            );
        }catch NACL::APISet::Exceptions::CommandFailedException with{
            my $exception_obj = shift;
            $Log->comment("$exception_obj->text()") unless ($repo_type eq 'DP');
        }catch NACL::APISet::Exceptions::ResponseException with{
            my $exception_obj = shift;
            $Log->comment("$exception_obj->text()") unless ($repo_type eq 'DP');
        };
        #Delete the repo
        $repo->delete("method-timeout" => $Options->get(option=>"TIMEOUT"));
    }
}

=head2 ::delete_aggrs

Delete all aggregates in a cluster that have zero volumes
on them. Immediately after deleting the aggregates, if the
user calls this method, all the aggregates that have zero
volumes will be deleted.

=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as instance method, otherwise necessary.

=item Examples

=back

Instance method:
    $CRUtils->delete_aggrs();

Class method:
    CR::CRUtils->delete_aggrs(
       command_interface => $Node,
    );

=cut

sub delete_aggrs {
    my ($self, %args) = @_;
    my $connection = $args{connection} || $self->{_connection};

    my @aggrs = NACL::STask::Aggregate->find(
                  command_interface => $connection,
                  filter =>{ volcount => 0 }
                );

    foreach my $aggr (@aggrs) {
        $aggr->purge();
    }
}

=head2 ::repository_create

Delete all aggregates in a cluster that have zero volumes
on them. Immediately after deleting the aggregates, if the
user calls this method, all the aggregates that have zero
volumes will be deleted.

=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as instance method, otherwise necessary.

=item Examples

=back

Instance method:
    $CRUtils->delete_aggrs();

Class method:
    CR::CRUtils->delete_aggrs(
       command_interface => $Node,
    );

=item Returns
=over
Returns a reference to a NACL Volume object.

=cut

sub repository_create {
    my ( $self, %args ) = @_;
    my $repository;

    my (%opts, @aggr_list);
    my $connection = $args{connection} || $self->{_connection};
    $opts{vserver} = $args{vserver} ||
                                 $Options->get(option=>'VSERVER');
    $opts{'junction-path'} = $args{'junction-path'} ||
                                 $Options->get(option=>'CVOL_JUNCTION_PATH');
    my $aggr_count = $args{aggr_count} ||
                                 $Options->get(option=>'AGGR_COUNT');
    $opts{'volume'} = $args{'volume'} ||
                                 $Options->get(option=>'CVOL_NAME');
    # Shorten the repository creation time
    $opts{'directory-bitwidth'} = $args{'directory-bitwidth'} || 2;
    $opts{'unix-permissions'} = $args{'unix-permissions'} ||
                                                    '---rwxrwxrwx';
    $opts{'type'} = $args{'type'} || 'RW';
    $opts{'space-guarantee'} = $args{'space-guarantee'} || 'volume';
    $opts{'size'} = $args{'size'} if $args{'size'};
    my $reuse = delete $args{'reuse'} || 0;
    my @node_names = $self->get_all_nodes(
                        connection => $connection
                     );
    my @assigned_aggr_names;
    my @ols_aggr_list;
    my $ols_enable =
           ( $Options->get( option => 'ENABLE_OBJECT_STORE' ) == 1 &&
             $Options->defined( option => 'OLS_AGGR_LIST' ) ) ? 1 : 0;
    if ($ols_enable) {
        @ols_aggr_list
                = @{ $Options->get( option => 'OLS_AGGR_LIST' ) };
    }

    # Find source infinite volumes and then create aggregates on destination
    # to match source infinitevols for RR style of snapmirror
    if ($opts{'type'} eq 'DP') {
        if (!$Options->defined( option => 'SRC_NODE' ) ||
            !$Options->defined( option => 'SRC_VSERVER' )) {
            $Log->comment('Test failed: SRC_NODE and SRC_VSERVER are mandatory parameters');
            return 0;
        }
        my $src_node = NACL::C::Node->new(
                        node=>$Options->get(option=>'SRC_NODE')
                                     );
        my @obj_vol = NACL::C::Volume->find(
                command_interface => $src_node,
                filter            => {
                    'vserver'          => $Options->get( option => 'SRC_VSERVER' ),
                    'constituent-role' => '!ns_mirror',
                    'is-constituent'   => 'true',
                },
            );
        my $new_aggr_count = scalar @obj_vol;
        $Log->comment("Requested repository with $aggr_count aggregates compares with \n".
                      "required $new_aggr_count aggregates to match source infinitevols");
        $aggr_count = $new_aggr_count if ($aggr_count < $new_aggr_count);
    }

    ##########################################################
    # Find if there is a repository present under this vserver
    ##########################################################
    $repository =  NACL::C::Volume->find(
                        command_interface   => $connection,
                        allow_empty         => 1,
                        filter              => {'volume-style' => 'infinitevol',
                                                'vserver'      => $opts{vserver}},
                      );
    ##################################################################
    # Find the vserver.
    ##################################################################
    my $vserver_obj = NACL::C::Vserver->find(
                         command_interface => $connection,
                         filter => { 'vserver' => $opts{vserver}},
                      );
    ##################################################################
    # If there is a repository, if it matches the aggr count and
    # the user wants to reuse it, return the repository object.
    # Else delete the repo.
    ##################################################################
    if ( $repository ) {
        $Log->comment('A repository is present');
        my $present_aggr_count =
                $self->get_obj_vols(connection => $connection,
                                    'vserver' => $opts{vserver}) + 1;
        if ( $present_aggr_count == $aggr_count && $reuse == 1 ) {
            return $repository;
        }

        $self->delete_repo(
           connection => $connection,
           vserver => $opts{vserver}
        );
    }
    ########################################################
    # Get all aggregates with no volumes
    ########################################################
    my @available_aggrs = NACL::C::Aggregate->find(
                             command_interface => $connection,
                             filter => { volcount => 0 },
                             allow_empty => 1
                          );
    my $available_aggr_count = scalar( @available_aggrs );
    ############################################################
    # Handle the inequality cases. If the aggr counts match, we
    # do not need to do anything.
    ############################################################
    if ( !$ols_enable ) {
        if ( $aggr_count < $available_aggr_count ) {
            my $to_be_removed = $available_aggr_count - $aggr_count;
            splice(@available_aggrs,0,$to_be_removed);
        } elsif ( $aggr_count > $available_aggr_count ) {
            my $to_be_created = $aggr_count - $available_aggr_count;
            my @new_aggrs = $self->create_aggrs(
                            nodes => \@node_names,
                            connection => $connection,
                            aggr_count => $to_be_created
                         );
            @available_aggrs = (@available_aggrs, @new_aggrs);
        }
    } else {
        my %tmp_hash;
        my $count = 0;
        foreach my $item (@ols_aggr_list) {
            $tmp_hash{$item} = 1;
            push @assigned_aggr_names, $item;
            #TODO: check to see if ols_aggregates are avilable
            #create otherwise
        }
        my $aggr_item;
        #available aggregates will not necessarirly have the ols list
        #This calcuation will prove wrong
        foreach my $item (@available_aggrs) {
            $aggr_item = $item->state()->aggregate();
            if (!defined $tmp_hash{$aggr_item}) {
                $tmp_hash{$aggr_item} = 1;
                $count++;
                push @assigned_aggr_names, $aggr_item;
            }
            last if ($count >= $aggr_count);
        }
        if ($count < $aggr_count) {
            my $to_be_created = $aggr_count - $count;
            my @new_aggrs = $self->create_aggrs(
                            nodes => \@node_names,
                            connection => $connection,
                            aggr_count => $to_be_created
                         );
            foreach my $aggr (@new_aggrs) {
                push @assigned_aggr_names, $aggr->state()->aggregate();
            }
        }
        $Log->comment("Assigning @assigned_aggr_names with OLS enabled");
    }
    ###########################################################
    # Create the aggr-list for assignment
    ###########################################################
    if ( !$ols_enable ||
         ($ols_enable && !@assigned_aggr_names) ) {
        foreach my $aggr (@available_aggrs) {
            push @assigned_aggr_names, $aggr->state()->aggregate();
        }
    }
    ##########################################################
    # Add the vserver root aggregate to the assigned aggr list
    ##########################################################
    push @assigned_aggr_names, $vserver_obj->state()->aggregate();

    $Log->comment("Assigning @assigned_aggr_names to vserver");
    $vserver_obj->modify('aggr-list' => \@assigned_aggr_names );

    #################################################################
    # Create the repository.
    #################################################################
    delete $opts{'junction-path'} if ($opts{'type'} eq 'DP');
    delete $opts{'unix-permissions'} if ($opts{'type'} eq 'DP');
    if ( !$ols_enable ) {
        $repository = NACL::C::Volume->create(
                         command_interface => $connection,
                          foreground => 'true',
                         'method-timeout' =>
                                     $Options->get(option=>"TIMEOUT"),
                         %opts
                      );
    } else {
        # this option is automaticaly mirror to DP volume on SN
        delete $opts{'directory-bitwidth'} if ($opts{'type'} eq 'DP');
        # this option does not work util the burt556167 is fixed
        delete $opts{'size'} if ($opts{'type'} eq 'DP');

        $Log->comment("Create infinitevol $opts{'volume'} with OLS enabled");
        $repository = NACL::C::Volume->create(
                          command_interface => $connection,
                          'ols-aggr-list' => \@ols_aggr_list,
                          'enable-object-store' => 'true',
                          foreground => 'true',
                          'method-timeout' =>
                                     $Options->get(option=>"TIMEOUT"),
                          %opts
                      );
    }

    return $repository;
}

=head2 ::create_aggrs

Creates a specified number of aggregates.

=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as instance method, otherwise necessary.

=item C<<"nodes"=>$nodenames >>

Reference an array of node names.

=item C<<"aggr_count"=>$count >>

(Optional) The number of aggregates to be created.
Default will be taken from CR::CommonParams which is 3.

=item C<<"aggr_prefix"=>$aggr_prefix >>

(Optional) The prefix for the aggrs to be created. The
postfix will be an _ and a number.


=item Examples

=back

Instance method:
    $CRUtils->create_aggrs(
      nodes => \@node_names,
    );

=item Returns

A reference to an array of NACL::STask::Aggregate objects.

=over

=cut


sub create_aggrs {
    my ( $self, %args ) = @_;
    my ($aggr_count, $aggr_prefix, @nodes, @aggr_list);
    my $connection = $args{connection} || $self->{_connection};
    my $node_names = $args{nodes};
    my $node_count = scalar(@$node_names);
    my $count = 0;
    $aggr_count = $args{aggr_count} || $Options->get(option=>'AGGR_COUNT');
    $aggr_prefix = $args{aggr_prefix} || $Options->get(option=>'AGGR_PREFIX');
    my $timer = $Options->get('option'=>'AGGR_CREATE_WAIT');
    my $sleep_time = $Options->get(option=>'AGGR_CHECK_INTERVAL');

    $Log->comment("Creating $aggr_count aggregates with $aggr_prefix prefix");

    for my $index ( 1 .. $aggr_count ) {
	my $aggr_name = newName($aggr_prefix);
        my $node = $node_names->[$count++ % $node_count];
        my $aggr = $self->create_aggr(
                      aggregate => $aggr_name,
                      nodes => [$node],
                      connection => $connection
                   );
        push @aggr_list, $aggr;
    }

    # Wait until they are created
    foreach my $aggr ( @aggr_list ) {
        $aggr->wait_for_creation( polling_interval => $sleep_time,
                                  'method-timeout' => $timer );
    }

    return @aggr_list;
}

=head2 ::get_obj_vols

Obtains either number of object volumes or a list of object volumes
depending on the context.

=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as instance method, otherwise necessary.

=item C<<"volume"=>$repo_vol_name >>

(Optional) Repository volume name.
Defaults to "repo_vol".

=item C<<"vserver"=>$vserver_name >>

(Optional) Name of the vserver.
Default will be taken from CR::CommonParams which is "vs0".

=item Examples

=back

Instance method:
    $CRUtil-get_obj_vols(
      nodes => \@node_names,
    );

=item Returns

In scalar context, returns number of object volumes.
In list context, returns an array of references to
NACL::C::Volume objects.

=over

=cut

sub get_obj_vols {
    my ( $self, %args ) = @_;
    my $repository = $args{repository} || undef;
    my $connection = $args{connection} || $self->{_connection};
    my $volname = $args{volume} || $Options->get(option=>'CVOL_NAME');
    my $vserver = $args{vserver} || $Options->get(option=>'VSERVER');

    my @obj_vols = NACL::C::Volume->find(
                     command_interface   => $connection,
                     filter              => {
                            'vserver'           => $vserver,
                            'is-constituent'    => 'true',
                            'constituent-role'  => 'data',
                     },
                   );


    return wantarray ? @obj_vols : scalar(@obj_vols);
}

#Copies elements of one hash to another. Useful if arguments differnet from
# options that need to be passed to lower level functions.

sub hashOverlay {
    my %args      = @_;

    my $source    = $args{source};
    my $target    = $args{target} || {};
    my $overwrite = defined $args{overwrite} ? $args{overwrite} : 1;

    if ( $source && ref $target eq 'HASH' ) {
        for my $key ( keys %{$source} ) {
            if ( ref $source->{$key} eq 'HASH' ) {
                $target->{$key} = _hashOverlay(
                    target    => $target->{$key},
                    source    => $source->{$key},
                    overwrite => $overwrite,
                );
            }
            else {
                if ( !defined( $target->{$key} ) || $overwrite ) {
                    $target->{$key} = $source->{$key};
                }
            }
        }
    }
    return $target;
}

=head2 ::create_aggr

Creates an aggregate.

=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"aggregate"=>$aggr_name >>

(Optional) Name of the new aggregate. If not passed, the new
aggregate name will start with AG prefix and a random name
will be selected.

=item C<<"background"=>[true|false] >>

(Optional) If background is true, the method will return immediately
after the aggregate create command is issued. Otherwise it will wait
till the aggregate becomes online. Default is true.

=item diskcount

(Optional) Numebr of disks to be used while creating the aggregate.
Default will be taken from CR::CommonParams.

==item

=item Examples

=back

Instance method:
    $CRUtils->create_aggr(
      nodes => \@node_names,
    );

=item Returns

In scalar context, returns number of object volumes.
In list context, returns an array of references to
NACL::C::Volume objects.

=over

=cut

sub create_aggr {
    my ($self, %args) = @_;

    my %opts;
    my (@existing_obj, $aggr_obj, $timer, $sleep_time);
    my $connection = delete $args{connection} || $self->{_connection};
    my $background = delete $args{background} || 'true';

    $opts{'diskcount'} = $args{diskcount} || $Options->get(option=>'AGGR_DISK_COUNT');
    $opts{'nodes'} = [$connection->get_one_state_attribute('node')];
    $opts{aggregate} = delete $args{aggregate} || newName( "AG" );
    $timer = $Options->get('option'=>'AGGR_CREATE_WAIT');
    $sleep_time = $Options->get(option=>'AGGR_CHECK_INTERVAL');

    hashOverlay( source => \%args,
                 target => \%opts
               );

    @existing_obj = NACL::STask::Aggregate->find( command_interface => $connection,
                         filter =>{ aggregate => $opts{aggregate}},
                         allow_empty => 1,
                       );
    if (@existing_obj) {
        $Log->comment("Warning: $opts{aggregate} already exists");
        $aggr_obj = $existing_obj[0];
    } else {
        $aggr_obj = NACL::STask::Aggregate->create(
                      command_interface => $connection,
                      nacltask_wait     => ($background !~ /true/i),
                      %opts
                    );
    }

    return $aggr_obj;
}

# Generates a new word from dictionary excluding netapp prohibited
# characters
sub _newWord {
    my @r;
    my $len = shift || 8;

    my $retry = 0;
    while (1) {
        @r = rand_words( size => 32 );
        for (@r) {
            next unless ( length($_) == $len );
            next if ( $_ =~ /(-|')/ );
            next if ( $_ =~ /shared/i );    # Reserved word for GenericMount
            next if ( $_wordHits{$_} );
            $_wordHits{$_} = 1;
            return $_;
        }
        if ( ++$retry > 29 ) {
#            NOTE::logcomment("Resetting newName after $retry failed attempts");
            %_wordHits = undef;
        }
    }
}

#Creates a new name starting with user specified prefix.
sub newName {
    my ( $descriptor ) = @_;

    my $name;
    $name = sprintf( "%05d_%s_%s", $$, _newWord(4), _newWord(8) );
    $name = $descriptor . $name if ($descriptor);
    return $name;
}

=head2 ::get_all_nodes

Obtains all the node names in a cluster

=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item Examples

=back

Instance method:
    $CRUtils->get_all_nodes();

=item Returns

In scalar context, this returns number of nodes in a cluster. In a list
context, this returns an array of node names that belong to a cluster.

=over

=cut

sub get_all_nodes {
    my ($self, %args) =  @_;
    my @node_names;

    my $ci_node = $args{connection} || $self->{_connection};
    my @nodes = NACL::C::Node->find(
                  command_interface => $ci_node
                );
    foreach my $node (@nodes) {
        my $node_name = $node->get_one_state_attribute('node');
        push @node_names, $node_name;
    }

    return wantarray ? @node_names : scalar(@node_names);
}

=head2 ::get_data_ip

Obtains a data interface ip address of a vserver.

=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item Examples

=back

Instance method:
    $server_addr = $CRUtils->get_data_ip();

=item Returns

Returns an ip address of given vserver. If not found throws a BaseException.

=over

=cut

sub get_data_ip {
    my ( $self, %args ) = @_;
    my $ip_addr;
    my $vserver = $args{vserver} || $Options->get(option=>'VSERVER');
    my $node = $args{connection} || $self->{_connection};
    my $node_name = $args{node} || $node->get_one_state_attribute('node');
    my @iface_objs = NACL::C::Interface->find(
                      command_interface => $node,
                      filter => {
                                'vserver' => $vserver,
#                                'home-node' => $node_name,
                                'role' => 'data',
                                'status-admin' => 'up',
                      },
                    );

    if ( @iface_objs ) {
        # Return a random one if there are multiple ones
        my $obj = $iface_objs[int(rand(@iface_objs))];
        $ip_addr = $obj->get_one_state_attribute('address');
    } else {
        NATE::BaseExceptin->throw("Server ip address is not found");
    }
    return $ip_addr;
}

=head2 ::snapmirror_create

Will create snapmirror
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"source_connection"=> $src_connection >>

(Mandatory) Source cluster connection


=item Examples

=back

Class method:
    CR::CRUtils->snapmirror_create(
        connection => $Dst_Node,
        source_connection => $Src_Node,
        foreground => 'false', # Put the job in background
        job_component => \$Job_obj,
    );

=item Returns


=over

=cut

sub snapmirror_create{
    my ($self, %args ) = @_;
    $args{method}='create';
    return _snapmirror_action(%args);
}


=head2 ::snapmirror_abort

Will abort snapmirror relationship
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"source_connection"=> $src_connection >>

(Mandatory) Source cluster connection


=item Examples

=back

Class method:
    CR::CRUtils->snapmirror_abort(
        connection => $Dst_Node,
        source_connection => $Src_Node,
        foreground => 'false', # Put the job in background
        job_component => \$Job_obj,
    );

=item Returns


=over

=cut
sub snapmirror_abort{
    my ($self, %args ) = @_;
    $args{method}='abort';
    return _snapmirror_action(%args);
}


=head2 ::snapmirror_promote

Will promote snapmirror relationship
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"source_connection"=> $src_connection >>

(Mandatory) Source cluster connection


=item Examples

=back

Class method:
    CR::CRUtils->snapmirror_promote(
        connection => $Dst_Node,
        source_connection => $Src_Node,
        foreground => 'false', # Put the job in background
        job_component => \$Job_obj,
    );

=item Returns


=over

=cut
sub snapmirror_promote{
    my ($self, %args ) = @_;
    $args{method}='promote';
    return _snapmirror_action(%args);
}


=head2 ::snapmirror_modify

Will modify snapmirror relationship
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"source_connection"=> $src_connection >>

(Mandatory) Source cluster connection


=item Examples

=back

Class method:
    CR::CRUtils->snapmirror_modify(
        connection => $Dst_Node,
        source_connection => $Src_Node,
        foreground => 'false', # Put the job in background
        job_component => \$Job_obj,
    );

=item Returns


=over

=cut
sub snapmirror_modify{
    my ($self, %args ) = @_;
    $args{method}='modify';
    return _snapmirror_action(%args);
}


=head2 ::snapmirror_resume

Will resume snapmirror relationship
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"source_connection"=> $src_connection >>

(Mandatory) Source cluster connection


=item Examples

=back

Class method:
    CR::CRUtils->snapmirror_resume(
        connection => $Dst_Node,
        source_connection => $Src_Node,
        foreground => 'false', # Put the job in background
        job_component => \$Job_obj,
    );

=item Returns


=over

=cut
sub snapmirror_resume{
    my ($self, %args ) = @_;
    $args{method}='resume';
    return _snapmirror_action(%args);
}


=head2 ::snapmirror_update

Will update snapmirror relationship
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"source_connection"=> $src_connection >>

(Mandatory) Source cluster connection


=item Examples

=back

Class method:
    CR::CRUtils->snapmirror_update(
        connection => $Dst_Node,
        source_connection => $Src_Node,
        foreground => 'false', # Put the job in background
        job_component => \$Job_obj,
    );

=item Returns


=over

=cut
sub snapmirror_update{
    my ($self, %args ) = @_;
    $args{method}='update';
    return _snapmirror_action(%args);
}

=head2 ::snapmirror_delete

Will delete snapmirror relationship
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"source_connection"=> $src_connection >>

(Mandatory) Source cluster connection


=item Examples

=back

Class method:
    CR::CRUtils->snapmirror_delete(
        connection => $Dst_Node,
        source_connection => $Src_Node,
        foreground => 'false', # Put the job in background
        job_component => \$Job_obj,
    );

=item Returns


=over

=cut

sub snapmirror_delete{
    my ($self, %args ) = @_;
    $args{method}='delete';
    return _snapmirror_action(%args);
}

=head2 ::snapmirror_initialize

Will initialize snapmirror relationship
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"source_connection"=> $src_connection >>

(Mandatory) Source cluster connection


=item Examples

=back

Class method:
    CR::CRUtils->snapmirror_initialize(
        connection => $Dst_Node,
        source_connection => $Src_Node,
        foreground => 'false', # Put the job in background
        job_component => \$Job_obj,
    );

=item Returns


=over

=cut
sub snapmirror_initialize{
    my ($self, %args ) = @_;
    $args{method}='initialize';
    return _snapmirror_action(%args);
}

=head2 ::snapmirror_break

Will break snapmirror relationship
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"source_connection"=> $src_connection >>

(Mandatory) Source cluster connection


=item Examples

=back

Class method:
    CR::CRUtils->snapmirror_break(
        connection => $Dst_Node,
        source_connection => $Src_Node,
        foreground => 'false', # Put the job in background
        job_component => \$Job_obj,
    );

=item Returns


=over

=cut

sub snapmirror_break{
    my ($self, %args ) = @_;
    $args{method}='break';
    return _snapmirror_action(%args);
}

=head2 ::snapmirror_resync

Will resync snapmirror
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"source_connection"=> $src_connection >>

(Mandatory) Source cluster connection


=item Examples

=back

Class method:
    CR::CRUtils->snapmirror_resync(
        connection => $Dst_Node,
        source_connection => $Src_Node,
        foreground => 'false', # Put the job in background
        job_component => \$Job_obj,
    );

=item Returns


=over

=cut

sub snapmirror_resync{
    my ($self, %args ) = @_;
    $args{method}='resync';
    return _snapmirror_action(%args);
}

=head2 ::snapmirror_action

will do snapmirror actions like create delete, initialize, resync, break
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"source_connection"=> $src_connection >>

(Mandatory) Source cluster connection

=item C<<"method"=> $initialize>>

(Mandatory) SourceMirror action

=item Examples

=back

Class method:
    CR::CRUtils->snapmirror_action(
        connection => $Dst_Node,
        source_connection => $Src_Node,
        method => 'resync',
        foreground => 'false', # Put the job in background
        job_component => \$Job_obj,
    );

=item Returns


=over

=cut

sub snapmirror_action{
    my ($self, %args ) = @_;
    return _snapmirror_action(%args);
}

sub _snapmirror_action{
    my (%args) = @_;
    my (%opts);
    my $Dst_Node = $args{connection};
    my $Dst_Vserver = $args{'destination_vserver'};
    my $Src_Node = $args{'source_connection'};
    my $Src_Vserver = $args{'source_vserver'};
    my $Job_obj_ref = $args{'job_component'};

    my $method = $args{method};
    #my $foreground = $args{foreground} || 'true';
    #optional arguments - needed for modify
    $opts{foreground} = $args{foreground} if $args{foreground};
    $opts{throttle} = $args{throttle} if $args{throttle};
    $opts{tries} = $args{tries} if $args{tries};
    $opts{schedule} = $args{schedule} if $args{schedule};

     if(!$Dst_Vserver){
        $Dst_Vserver = NACL::C::Vserver->find(
            command_interface=>$Dst_Node,
            filter=> {'is-repository' => 'true'}) ;
     }

    my $Dst_Vol_obj = NACL::C::Volume->find(
                        command_interface=>$Dst_Node,
                        filter => {'volume-style' => 'infinitevol',
                        vserver => $Dst_Vserver},
                        );

    my $Dst_Volume = $Dst_Vol_obj->get_one_state_attribute('volume');

    if(!$Src_Vserver){
        my $src_vs = NACL::C::Vserver->find(
            command_interface=>$Src_Node,
            filter=> {'is-repository' => 'true'}
            );
     }

    my $Src_Vol_obj = NACL::C::Volume->find(
                        command_interface => $Src_Node,
                        filter => {'volume-style' => 'infinitevol',
                                    vserver => $Src_Vserver},
                        );
    my $Src_Volume = $Src_Vol_obj->get_one_state_attribute('volume');
    my $Dst_Clus_obj = NACL::CS::ClusterIdentity->fetch(
                                             command_interface => $Dst_Node
                                  );
    my $Dst_Cluster = $Dst_Clus_obj->name();

    my $Src_Clus_obj = NACL::CS::ClusterIdentity->fetch(
                                             command_interface => $Src_Node
                                  );
    my $Src_Cluster = $Src_Clus_obj->name();
    my $action_obj ;

    #snapmirror create does not have a foreground option
    #and does not create a job. Hence we will always make a component call
    #for it
    #In case of create, $action_obj will be the newly created snapmirror object
    my $is_component_call;
    if ($method =~ /create/) {
        $is_component_call = 1;
        delete $opts{foreground};
    }

    #If foreground exists and is true or if its a component call
    if ($args{foreground} eq 'true' || $is_component_call) {
        $action_obj =
            NACL::C::Snapmirror->$method(command_interface     =>$Dst_Node,
                                         'destination-volume'  => $Dst_Volume,
                                         'source-volume'       => $Src_Volume,
                                         'destination-vserver' => $Dst_Vserver,
                                         'source-vserver'      => $Src_Vserver,
                                         'destination-cluster' => $Dst_Cluster,
                                         'source-cluster'      => $Src_Cluster,
                                         'method-timeout'      => 600,
                                         %opts,
                                         );
    }
    else{
        #Burt Id : 518304
        #Do a Apiset call as thpl users can get the job_id
        my $NACL_Api_Set_Obj =
            NACL::APISet->new(hostobj   => $Dst_Node->hostrec(),
                              category  => "Node",
                              interface => "CLI",
                              set       => "CMode",
                             );
        my $apiset_method = "snapmirror_${method}";
        $action_obj = $NACL_Api_Set_Obj->$apiset_method(
                    'destination-volume' => $Dst_Volume,
                    'source-volume' => $Src_Volume,
                    'destination-vserver' => $Dst_Vserver,
                    'source-vserver' => $Src_Vserver,
                    'destination-cluster' => $Dst_Cluster,
                    'source-cluster' => $Src_Cluster,
                    %opts,
                    );
        my $parsed_output = $action_obj->{parsed_output};
        my ($Job_hash) = @$parsed_output;
        $action_obj->{job_id}=$Job_hash->{job_id};
    }
    return $action_obj;
}

=head2 :: snapshot_restore

will do snapmirror actions like create delete, initialize, resync, break
=over

=item Arguments

=item C<<"connection"=>$Node >>
(Mandatory) connection

=item C<<"snapshot"=> $snapshot_name>>

(Mandatory) Snapshot name to restore to .

=item Examples

=back

Class method:
    CR::CRUtils->snapmirror_action(
        connection => $Node,
        snapshot => $snapshot_name,
    );

=item Returns


=over

=cut

sub snapshot_restore{
    my ($self, %args ) = @_;
    my (%opts);
    my $Node = $args{connection} || $self->{_connection};
    my $Vserver = $args{vserver} || $Options->get(option=>'VSERVER');
    my $Snapshot = $args{snapshot};
    my $flag_ignore_mount =  $args{'ignore-mount'};
    my $volume_object = NACL::C::Volume->find(
                        command_interface=>$Node,
                        filter => {vserver => $Vserver,},
        );
    my $Volume = $volume_object->get_one_state_attribute('volume');
    my $Junction_path = $volume_object->get_one_state_attribute('junction-path');
    #unmount volume
    $volume_object->unmount();
    #Restore snapshot
    NACL::C::Snapshot->restore(
                    command_interface => $Node,
                    snapshot => $Snapshot,
                    vserver => $Vserver,
                    volume => $Volume,
                    #foreground => 'true',
                    );

    # the last step of restore is resync betw ns and ns_mirror (bkgd proc)
    # if ns_mirror is not in idle state, wait till it completes
    my $snapmirror_obj = NACL::STask::Snapmirror->find(
                             command_interface => $Node,
                             'filter'          => {
                                 'vserver' => $Vserver,
                                 'relationship-capability' => '8.2*',
                             },
                             'allow_empty' => 'true',
                         );
    if ($snapmirror_obj) {
        $snapmirror_obj->monitor('method-timeout'  =>  60,
                                 'polling_interval'=>  5,
                                 'nacltask_wait_for_status' => 'Idle',
                                );
    }

    $volume_object->mount('junction-path' => $Junction_path) if(!$flag_ignore_mount);
}


=head2 ::wait_till_job_succeeds

Wait till a particular job (in background) finishes

=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"name"=> 'job name'>>

(Mandatory) job name

=item Examples

=back

Class method:
    CR::CRUtils->wait_till_job_succeeds(
            connection => $Dst_Node,
            name => 'snapmirror transfer',
    );

=item Returns


=over

=cut

sub wait_till_job_succeeds{
    my ($self, %args ) = @_;
    my $Node = $args{connection} || $self->{_connection};
    my $id = $args{id};
    my $Job_obj = NACL::C::Job->find(
                        command_interface => $Node,
                        filter => { id => $id },
                        allow_empty => 1,
                        );
    if ($Job_obj){
        $Job_obj->watch_progress('method-timeout' => 500);
    }else{
        $Log->comment("Looks like the job : $id is not found ");
    }
}

=head2 ::check_snapmirror_fields

check Snapmirror state and status

=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"source_connection"=> $src_connection >>

(Mandatory) Source cluster connection

=item C<<"expected_state"=> 'Snapmirrored'>>

(Mandatory) SourceMirror state

=item C<<"expected_status"=> 'Transferring'>>

(Mandatory) SourceMirror status

=item Examples

=back

Class method:
    CR::CRUtils->check_snapmirror_fields(
        connection => $Dst_Node,
        source_connection => $Src_Node,
        expected_state => 'Snapmirrored',
        expected_status => 'Transferring'
    );

=item Returns


=over

=cut
sub check_snapmirror_fields {
    my ($self, %args ) = @_;
    my $Dst_Node = $args{connection} || $self->{_connection};
    my $Src_Node = $args{'source_connection'};
    my $Src_Vserver = $args{'source_vserver'};
    my $Dst_Vserver = $args{'destination_vserver'};
    my $expected_state = $args{'expected_state'};
    my $expected_status = $args{'expected_status'};
    # wait for 10 seconds, as it takes some time to show the status properly
    sleep(10);
    my @failures=();

    if(!$Dst_Vserver){
        $Dst_Vserver = NACL::C::Vserver->find(
                        command_interface=>$Dst_Node,
                        filter=> {'is-repository' => 'true'}) ;
    }

    my @Dst_snapm_objs = NACL::C::Snapmirror->find(
                            command_interface => $Dst_Node,
                            filter => {'vserver' => $Dst_Vserver,
                                       'is-constituent' => "false"},
                            allow_empty => 1,
                            );
    push (@failures,'Destination : snapmirror show fails') unless (@Dst_snapm_objs);
    $Log->comment("Destination Snamirror objects : @Dst_snapm_objs");
    foreach my $Dst_snapm_obj(@Dst_snapm_objs){
        if($Dst_snapm_obj->get_one_state_attribute('state') ne $expected_state){
            $Log->comment("Destination cluster is not in $expected_state state");
            $Log->comment('Instead it is :'. $Dst_snapm_obj->get_one_state_attribute('state'));
            push (@failures,"Destination cluster is not in $expected_state state".
                        'Instead it is :'.
                         $Dst_snapm_obj->get_one_state_attribute('state')
                        );
        }
        if($Dst_snapm_obj->get_one_state_attribute('status') ne $expected_status){
            $Log->comment("Destination cluster is not in $expected_status status");
            push (@failures,"Destination cluster is not in $expected_status status");
        }
    }

    if(!$Src_Vserver){
            $Src_Vserver = NACL::C::Vserver->find(
                            command_interface=>$Src_Node,
                            filter=> {'is-repository' => 'true'}) ;
    }

    my $dst_clus_obj = NACL::CS::ClusterIdentity->fetch(
                                             command_interface => $Dst_Node
                                  );

    #filter snapmirror by destination clustername
    my @Src_snapm_objs =  NACL::C::Snapmirror->find(
                            command_interface => $Src_Node,
                            filter => {'vserver' => $Src_Vserver,
                                        'is-constituent' => "false",
                                        'destination-cluster' => $dst_clus_obj->name(),},
                            allow_empty => 1,
                            );
    push (@failures,'Source : snapmirror show fails') unless (@Src_snapm_objs);
    $Log->comment("Source Snamirror objects : @Src_snapm_objs");
    foreach my $Src_snapm_obj(@Src_snapm_objs){
        if($Src_snapm_obj->get_one_state_attribute('state') ne $expected_state){
            $Log->comment("Source cluster is not in $expected_state state");
            push (@failures,"Source cluster is not in $expected_state state");
        }
        if($Src_snapm_obj->get_one_state_attribute('status') ne $expected_status){
            $Log->comment("Source cluster is not in $expected_status status");
            push (@failures,"Source cluster is not in $expected_status status");
        }
    }
    return @failures;
}

=head2 ::check_cluster_peer

check cluster peer ping and health

=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) If used as an instance method, otherwise necessary.

=item C<<"source_connection"=> $src_connection >>

(Optional) If used as an instance method, otherwise necessary.

=item Examples

=back

Class method:
    CR::CRUtils->check_cluster_peer(
        connection => $Dst_Node,
        source_connection => $Src_Node,
    );

=item Returns

=over

=cut
sub check_cluster_peer {
    my ($self, %args ) = @_;
    my $dst_node = $args{connection} || $self->{_connection};
    my $src_node = $args{'source_connection'};

    #src cluster find and ping
    my $dst_clus = NACL::CS::ClusterIdentity->fetch(
                                             command_interface => $dst_node
                                  );
    my $dst_cluster_name = $dst_clus->name();

    my $peer = NACL::C::ClusterPeer->find(
                 command_interface =>  $src_node,
                 filter => { cluster => $dst_cluster_name },
              );

    $peer->state();

    my @cluster_peer_ping_states;
    try {
        @cluster_peer_ping_states = NACL::C::ClusterPeer->ping(
                                      command_interface => $src_node,
                                   );

        foreach my $element (@cluster_peer_ping_states) {
            foreach my $key (keys %{$element}) {
                $Log->comment("src ping dst $key: $element->{$key}");
            }
        }
    } catch NACL::Exceptions::NoElementsFound with {
        $Log->comment("NoElementsFound in src ping");
    };
    $Log->comment("End of src ping");

    #dst cluster find and ping
    my $src_clus = NACL::CS::ClusterIdentity->fetch(
                                             command_interface => $src_node
                                  );
    my $src_cluster_name = $src_clus->name();

    $peer = NACL::C::ClusterPeer->find(
                 command_interface =>  $dst_node,
                 filter => { cluster => $src_cluster_name },
              );

    $peer->state();

    try {
        @cluster_peer_ping_states = NACL::C::ClusterPeer->ping(
                                      command_interface => $dst_node,
                                   );

        foreach my $element (@cluster_peer_ping_states) {
            foreach my $key (keys %{$element}) {
                $Log->comment("dst ping src $key: $element->{$key}");
            }
        }
    } catch NACL::Exceptions::NoElementsFound with {
        $Log->comment("NoElementsFound in dst ping");
    };
    $Log->comment("End of dst ping");

    #src cluster peer health check
    my $peer_health;
    try {
        $peer_health = NACL::C::ClusterPeerHealth->find(
                         command_interface =>  $src_node,
                      );
        $peer_health->state();
        $Log->comment("src health ElementsFound");
    } catch NACL::Exceptions::NoElementsFound with {
        $Log->comment("src health noElementsFound");
    };

    #dst cluster peer health check
    try {
       $peer_health = NACL::C::ClusterPeerHealth->find(
                         command_interface =>  $dst_node,
                      );
            $peer_health->state();
        $Log->comment("dst health ElementsFound");
    } catch NACL::Exceptions::NoElementsFound with {
        $Log->comment("dst health noElementsFound");
    };

}

=head2 ::convert_size_into_KB
Converts all sizes : KB MB GB TB PB  into bytes
=over

=item Examples

=back
Class method:
    @sizes_in_bytes = CR::CRUtils->convert_size_into_KB(
                        1KB,
                        50 PB
                       );

=cut

sub convert_size_into_KB{
    my ($self, @sizes ) = @_;
    my @sizes_inKB;
    foreach my $size(@sizes){
        print "$size \n";
        chomp($size);
        my ($size_unprocessed,$byte_type,$size_in_KB);
        $size =~ m /(.*)([KMGTP]B)/g;
        $size_unprocessed=$1;
        $byte_type = $2 || 'KB';
        $size_in_KB=0;
        if($byte_type eq 'PB'){
            $size_in_KB = $size_unprocessed * 1024 * 1024 * 1024 * 1024;
        }elsif($byte_type eq 'TB'){
            $size_in_KB = $size_unprocessed * 1024 * 1024 * 1024;
        }elsif($byte_type eq 'GB'){
            $size_in_KB = $size_unprocessed * 1024 * 1024;
        }elsif($byte_type eq 'MB'){
            $size_in_KB = $size_unprocessed * 1024;
        }else{
            $size_in_KB = $size_unprocessed;
        }
        push (@sizes_inKB, $size_in_KB.'KB');
    }
    return @sizes_inKB;
}

=head2 ::check_object_state
Check an object state and status
=over

=item Examples
given a job object, and check job state is 'Running' or not
=back
Class method:
    $object_status = CR::CRUtils->check_object_state(
                            object => $object,
                            check_state => 'state'
                            expected_status => 'Running'
                            timeout => 100,
                            );

=cut
sub check_object_state {
    my ($self, %args) = @_;

    my ($object, $expected_status, $state_attribute, $timeout);
    $object = $args{object} if $args{object};
    $expected_status = $args{expected_status} if $args{expected_status};
    $state_attribute = $args{check_state} if $args{check_state};
    $timeout = $args{timeout} if $args{timeout};

    my $ret_flag = 0;

    if (!defined $timeout) {
        $timeout = 100;
    }
    if (!defined $state_attribute) {
        $state_attribute = 'state';
    }
    if (!defined $expected_status) {
        $expected_status = 'Success';
    }

    if (defined $object) {
        my $i;
        my $obj_status;
        sleep(2);

        # A timeout wait loop to get an expected status
        # Checking status of an object will be failure if time out
        for ($i = 0; $i < $timeout; $i++) {
            $obj_status = $object->get_one_state_attribute($state_attribute);
            last if ($obj_status eq $expected_status);
            sleep(1);
        }
        $ret_flag = $obj_status if ($i < $timeout);
    } else {
        $Log->comment('This object is not defined');
    }

    return $ret_flag;
}

=head2 ::time_to_epoch_seconds
Converts job start/end time to epoch seconds:
currently, handle the time format of job starttime or endtime:
"dd/mm hours:minutes:seconds"
=over

=item Examples

=back
Class method:
    $seconds = CR::CRUtils->time_to_epoch_seconds(
                        "07/19 12:38:20"
                       );

=cut
sub time_to_epoch_seconds {
    my ($self, $time_str ) = @_;

    my $date_time = `date`;
    my @times = split(' ', $date_time);
    my $year = $times[@times - 1];

    my ($date, $time, $mm, $dd, $hr, $min, $sec);
    my $epoc_sec = 0;

    try {
        ($date, $time) = split(' ', $time_str);
        ($mm, $dd) = split('/', $date);
        ($hr, $min, $sec) = split(':', $time);
        $epoc_sec = timelocal($sec, $min, $hr, $dd, $mm - 1, $year);

    } catch NATE::BaseException with {
        my $ex = shift;
        $Log->comment('Exception thrown from time_to_epoch_seconds() '.
                    'in CRUtils.pm');
        $ex->throw();
    };

    return $epoc_sec;
}


=head2 ::check_volume_space_guarantee
Volume space guarantee
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) Defaults

=item Examples

=back
Class method:
    CR::CRUtils->check_volume_space_guaranteee(
                    connection => $Node,
                    vserver => $Vserver,
                    );

=cut

sub check_volume_space_guarantee{
    my ($self, %args ) = @_;
    my $Node = $args{connection} || $self->{_connection};
    my $repo_vs  = $args{vserver};
    #check the value for space-guarantee
    if(!$repo_vs){
        $repo_vs = NACL::C::Vserver->find(
            command_interface=>$Node,
            filter=> {'is-repository' => 'true'}
            );
    }

    my $RepoVol_Obj = NACL::C::Volume->find(
                        command_interface => $Node,
                        filter => {'volume-style' => 'infinitevol',
                                    vserver => $repo_vs,
                                  },
                        );
    my @failure_text=();
    if ($RepoVol_Obj->get_one_state_attribute('space-guarantee') ne 'volume'){
        $Log->comment('Upon DP create, value of space-guarantee is not volume');
        push (@failure_text,'Upon DP create, value of space-guarantee is not volume');
    }
    return @failure_text;
}

=head2 ::start_client_packet_trace
Start Client side tracing
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) Defaults

=item C<<"vserver"=> $Vserver >>

(Optional) Name of the vserver for which the repository belongs to.
Defaults to "vs0" if not passed.

=item Examples

=back
Class method:
    CR::CRUtils->start_client_packet_trace(
                    connection => $Src_Node,
                    vserver => $Vserver,
                    );

=cut

sub start_client_packet_trace{
    my ($self, %args ) = @_;
    my $Node = $args{connection} || $self->{_connection};
    my $Vserver = $args{vserver} || $Options->get(option=>'VSERVER');
    my $Client = $args{client} || $Options->get(option=>'CLIENT');
    my ($Client_Name) = @$Client;
    my $client_hostrec = Hostrec->new(id => "$Client_Name");
    my $Node_name = $Node->name();
    my $capture_id = $args{capture_id};
    #Get Data IP
    my $data_lif = NACL::C::NetworkInterface->find(command_interface=>$Node,
                                        filter => {vserver => $Vserver,
                                                    role => 'data',
                                                },
                                        );
    my $Data_ip = $data_lif->get_one_state_attribute('address');
    $Log_directory = $Options->get(option => 'LOGDIR');
    #Find the client Interface
    my $ifconfig;
    my ($set, %opts);

    #Start packet trace on client
    if($client_hostrec->{OS} eq 'SunOS'){
        $set = 'Solaris';
        $Api_Set_Obj_Client_C = NACL::APISet->new(
                                 name => $Client_Name ,
                                 category => "Host",
                                 interface => "CLI",
                                 set => $set,
                                 );
        $ifconfig = $Api_Set_Obj_Client_C->execute_raw_command(
                        command => "/usr/sbin/ifconfig -a",
                        );
        $ifconfig =~ s/:/ /g;
    }else{
        $set = 'Linux';
        $Api_Set_Obj_Client_C = NACL::APISet->new(
                                 name => $Client_Name ,
                                 category => "Host",
                                 interface => "CLI",
                                 set => $set,
                                 );
        $ifconfig = $Api_Set_Obj_Client_C->execute_raw_command(
            command => "/sbin/ifconfig",
            );
    }
    my @lines = split(/\n/,$ifconfig);
    my ($interface) = grep (/^e/,@lines);
    ($interface) = split(/ /, $interface);

    my $tshark_cmd = '/usr/software/bin/sudo /usr/software/bin/tshark -i '.$interface.' -f "host '.$Data_ip.'" -w '.$Log_directory.'/'.$Client_Name.'_'.$interface.'_'.$capture_id.'.trc';
    $Log->comment("Tshark Command : $tshark_cmd");
    $Api_Set_Obj_Client_C->execute_raw_command(
        command => "$tshark_cmd &",
        );
    my $tshark_ps_output = $Api_Set_Obj_Client_C->execute_raw_command(
        command => "/usr/software/bin/sudo ps -aef | grep tshark",
        );
    $tshark_ps_output =~ s/^\s*//;
    my @tshark_ps_tmp = split (/ /,$tshark_ps_output);
    $Tshark_pid = $tshark_ps_tmp[1];
    $Log->comment("$tshark_ps_output \n processs id : $Tshark_pid ");
    $Log->comment('Started Client Side tracing .. ');
}

=head2 ::stop_and_copy_client_packet_trace
Stop Client side tracing and copy files
=over

=item Examples

=back
Class method:
    my $client_trace_file = CR::CRUtils->
            stop_and_copy_client_packet_trace();

=item Returns
Return client trace file name

=over

=cut
sub stop_and_copy_client_packet_trace{
    my ($self, %args ) = @_;
    my @client_traces_files;
    my $client_trace_file;
    my $Log_directory = $Options->get(option => 'LOGDIR');
    my $Client = $args{client} || $Options->get(option=>'CLIENT');
    my $capture_id = $args{capture_id};
    my ($Client_Name) = @$Client;
    my $client_hostrec = Hostrec->new(id => "$Client_Name");

    $Api_Set_Obj_Client_C->execute_raw_command(
            command => "/usr/software/bin/sudo kill 2 $Tshark_pid",
        );
    my $tshark_ps_output = $Api_Set_Obj_Client_C->execute_raw_command(
            command => "/usr/software/bin/sudo ps -aef | grep tshark",
        );
     $Log->comment("Output after killing tshark on client : $tshark_ps_output");
     @client_traces_files = <${Log_directory}/*${capture_id}.trc>;
     $Log->comment('Stoping Client Side tracing .. ');
    ($client_trace_file) = @client_traces_files;
    $Log->comment("Client side trace files : $client_trace_file");
    system("/usr/software/bin/sudo chmod 777 $client_trace_file");
    return wantarray ? @client_traces_files : $client_trace_file;
}

=head2 ::start_server_packet_trace

Start the Server side tracing
=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) Defaults

=item C<<"vserver"=> $Vserver >>

(Optional) Name of the vserver for which the repository belongs to.
Defaults to "vs0" if not passed.

=item C<<"client"=> $Client >>

(Optional) Defaults to $Client

=item Examples

=back
Class method:
    CR::CRUtils->start_server_packet_trace(
                    connection => $Src_Node,
                    vserver => $Vserver,
                    client => $Client,
                    );
=cut

sub start_server_packet_trace{
    my ($self, %args ) = @_;
    my $Node = $args{connection} || $self->{_connection};
    my $Vserver = $args{vserver} || $Options->get(option=>'VSERVER');
    my $Client = $args{client} || $Options->get(option=>'CLIENT');
    my ($Client_Name) = @$Client;
    my $client_hostrec =  Hostrec->new(id => "$Client_Name");
    my $Node_name = $Node->name();
    my $Host_obj = $Node->hostrec();
    #Start Packet trace on the node's LIF
    #Verify and Get the interface
    #Get Data IP and port
    my $data_lif = NACL::C::NetworkInterface->find(command_interface=>$Node,
                                        filter => { vserver => $Vserver,
                                                    role => 'data',
                                                },
                                        );
    my $Data_ip = $data_lif->get_one_state_attribute('address');
    my $data_lif_port = $data_lif->get_one_state_attribute('curr-port');
    my $data_lif_node = $data_lif->get_one_state_attribute('curr-node');

    #Get Node IP
    my $node_mgmt_lif = NACL::C::NetworkInterface->find(
                                        command_interface=>$Node,
                                        filter => {role => 'node-mgmt',
                                                'home-node' => $data_lif_node,
                                        },
    );
    my $Node_ip = $node_mgmt_lif->get_one_state_attribute('address');

    my ($set, %opts);
    if($client_hostrec->{OS} eq 'SunOS'){
        $set = 'Solaris';
    }else{
        $set = 'Linux';
    }
    #Client Side Settings
    $Api_Set_Obj_Client = NACL::APISet->new(
                                 name => $Client_Name ,
                                 category => "Host",
                                 interface => "CLI",
                                 set => $set,
                                 );
    #using /tmp - hoping not to disturb user's current rsa settings
    #Remove previous rsa entries and generate new(#TODO : may be use if exists)
    my @commands_on_client_machine = (
    "ssh-keygen -R $Node_name",
    "ssh-keygen -R $Node_ip",
    "rm -rf /tmp/ssh_key_${Client_Name} ; rm -rf /tmp/ssh_key_${Client_Name}.pub; ssh-keygen -t rsa -f /tmp/ssh_key_${Client_Name} -N ''",
    "cd ; pwd; ssh-keyscan -t rsa $Node_ip >> .ssh/known_hosts",
    "cd ;pwd;  ssh-keyscan -t rsa $Node_name >> .ssh/known_hosts",
    "pwd; cat .ssh/known_hosts",
    );
    foreach my $command (@commands_on_client_machine){
        my $command_output = $Api_Set_Obj_Client->execute_raw_command(
            command => "$command",
        );
        $Log->comment("$command : $command_output");
    }
    my $public_key_contents = $Api_Set_Obj_Client->execute_raw_command(
            command => "cat /tmp/ssh_key_${Client_Name}.pub",
            );
    chomp($public_key_contents);
    my $Api_Set_Obj_diag = NACL::APISet->new(hostobj => $Host_obj,
                                 category => "Node",
                                 interface => "CLI",
                                 set => "Systemshell",
                                );
    my @commands_on_server_in_diag_mode = (
    "cd; rm -rf .ssh; mkdir -p .ssh; touch .ssh/authorized_keys; echo \"$public_key_contents\" > .ssh/authorized_keys",
    "cat .ssh/authorized_keys",
    "mkdir -p /mroot/trace_file_backups",
    "mv /mroot/*.trc /mroot/trace_file_backups",
    );
    foreach my $command (@commands_on_server_in_diag_mode){
        my $command_output = $Api_Set_Obj_diag->execute_raw_command(
            command => "$command",
        );
        $Log->comment("$command : $command_output");
    }
    $Api_Set_Obj = NACL::APISet->new(hostobj => $Host_obj,
                                 category => "Node",
                                 interface => "CLI",
                                 set => "CMode",
                                 );
    my @commands_on_server = (
    "run local pktt delete $data_lif_port",
    "run local pktt start $data_lif_port",
    );
    foreach my $command (@commands_on_server){
        my $command_output = $Api_Set_Obj->execute_raw_command(
            command => "$command",
        );
        $Log->comment("$command : $command_output");
    }
 }

=head2 ::stop_and_copy_server_packet_trace

Stops the Server side tracing, dumps the trc file into LOGDIR

=over

=item Arguments

=item C<<"connection"=>$Node >>

(Optional) Defaults

=item C<<"vserver"=> $Vserver >>

(Optional) Name of the vserver for which the repository belongs to.
Defaults to "vs0" if not passed.

=item C<<"client"=> $Client >>

(Optional) Defaults to $Client

=item Examples

=back
Class method:
    my $server_side_trace_file = CR::CRUtils->stop_and_copy_server_packet_trace(
                    connection => $Src_Node,
                    vserver => $Vserver,
                    client => $Client,
                    );
=item Returns
Return client trace file name

=over
=cut

sub stop_and_copy_server_packet_trace{
    my ($self, %args ) = @_;
    my $Node = $args{connection} || $self->{_connection};
    my $Vserver = $args{vserver} || $Options->get(option=>'VSERVER');
    my $Client = $args{client} || $Options->get(option=>'CLIENT');
    my ($Client_Name) = @$Client;
    my $Node_name = $Node->node();
    #Need a better way to get/construct the LOGDIR
    my $Log_directory = $Options->get(option => 'LOGDIR');
    my $user_log_home_dir = "$ENV{'LOGDIR'}"; # Issue : solaris runs apiset as root
    my $dir_to_cp_trc_files = $Log_directory || "${user_log_home_dir}/latest";
    my $capture_id = $args{capture_id};
    #stop Packet trace on the node's LIF
    #Get the interface
    #Get Data IP and port
    my $data_lif = NACL::C::NetworkInterface->find(command_interface=>$Node,
                                        filter => {vserver => $Vserver,
                                                    role => 'data',
                                                },
                                        );
    my $Data_ip = $data_lif->get_one_state_attribute('address');
    my $data_lif_port = $data_lif->get_one_state_attribute('curr-port');
    my $data_lif_node = $data_lif->get_one_state_attribute('curr-node');
    #Get Node IP
    my $node_mgmt_lif = NACL::C::NetworkInterface->find(command_interface=>$Node,
                                        filter => {
                                                role => 'node-mgmt',
                                                'home-node' => $data_lif_node,
                                        },
    );
    my $Node_ip = $node_mgmt_lif->get_one_state_attribute('address');

    my @commands_on_server = (
    "run local pktt status $data_lif_port ",
    "run local pktt dump $data_lif_port ",
    "run local pktt stop $data_lif_port ",
    "run local pktt list",
    );
    foreach my $command (@commands_on_server){
        my $command_output = $Api_Set_Obj->execute_raw_command(
            command => "$command",
        );
        $Log->comment("$command : $command_output");
    }
    #List Pkt Trace on Server
    my $pktt_list = $Api_Set_Obj->execute_raw_command(
        command => "run local pktt list",
        );
    $Log->comment("pktt list shows : $pktt_list ");
    #Identify which file has to be fetched from Node
    my @pkttt_list_array = split(/\n/,$pktt_list);
    my @pktt_trcs_for_interfaces = sort(grep (/^$data_lif_port/,@pkttt_list_array));
    my $latest_trc_file = pop(@pktt_trcs_for_interfaces);
    $Log->comment("Trace file to be fetched : $latest_trc_file");
    #Fetch trc files using Client Object
    my $fetch_trc_files = $Api_Set_Obj_Client->execute_raw_command(
        command => "cd ; scp  -i /tmp/ssh_key_${Client_Name} diag\@$Node_ip:/mroot/$latest_trc_file $dir_to_cp_trc_files/${Data_ip}_${data_lif_port}_${capture_id}.trc",
        );
    $Log->comment("SCP/ FETCHing trc files shows : $fetch_trc_files");
    #workaround for Issue : abcdefgh
    system("/usr/software/bin/sudo chmod 777 $dir_to_cp_trc_files/*.trc");
    #Packet analysis on the fetched trc file
    my $file = "$dir_to_cp_trc_files/${Data_ip}_${data_lif_port}_${capture_id}.trc";
    return $file;
}

=head2 ::get_timestemp

It calculates the current time, formats it, concatenates it and returns the final value

=over

=item Arguments

None

=item Examples

=back

Instance method:
    $time_stamp = $CRUtil->get_timestamp();

=item Returns

Time-stamp

=over

=cut

sub get_timestamp {
    my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst )
        = localtime(time);
    if ( $mon < 10 )  { $mon  = "0$mon"; }
    if ( $hour < 10 ) { $hour = "0$hour"; }
    if ( $min < 10 )  { $min  = "0$min"; }
    if ( $sec < 10 )  { $sec  = "0$sec"; }
    $year = $year + 1900;
    $mon = $mon + 1;

    return $year . '_' . $mon . '_' . $mday . '_' . $hour . '_' . $min . '_'
        . $sec;
}


=head2 get_min_max_volume_size

Gets the minimum and the maximum infinitevol size for the volume

=over

=item Arguments

=item C<$args)>
(Optional) Reference to a hash which can have keys -
connection - the node object one which volume size is to be determined
vserver    - the name of the vserver on which the volume size is to be determined

=item Returns

=item C<$size)>
Reference to a hash containing keys - max_size and min_size.
Values against the keys would be the respective size in bytes.

=item Usage
    my $size = $Cr_Utils_Obj->get_min_max_volume_size();
    $min_size = $size->{'min_size'};
    $max_size = $size->{'max_size'};

=back

=cut

sub get_min_max_volume_size {
    my ($self, %args) = @_;

    my $size = {};
    my $node_obj = $args{connection} || $self->{_connection};
    my $vserver  = $args{vserver} || $Options->get(option=>'VSERVER');
    my $host_obj = $node_obj->hostrec();
    my $zapi_apiset_obj = NACL::APISet->new(hostobj   => $host_obj,
                                            category  => 'Node',
                                            interface => 'ZAPI',
                                            set       => 'CMode',
                                           );

    $zapi_apiset_obj->set_vserver('vserver-name' => $vserver);

    my $response_obj =
        $zapi_apiset_obj->volume_get_limits();
    my $output = $response_obj->get_parsed_output();
    $size->{'max_size'} = $output->[0]{'max-infinitevol-size'};
    $size->{'min_size'} = $output->[0]{'min-infinitevol-size'};

    return $size;
}

=head2 :: get_job_id

For a given Job object this function will return a job id if the job is running

If the job is Queued or not in running state it will wait for few seconds to
see if the job is in running state.

=over

=item Arguments

None

=item Examples

=back

Instance method:
    $job_id = $CRUtil->get_job_id( job_object => $job_obj);

=item Returns

$job_id if the job is running

=over

=cut
sub get_job_id{
    my ($self, %args) = @_;
    my $job_obj = $args{'job_object'};
    my $job_status;
    my $wait_count;
    until (($job_status eq 'Running') || ($wait_count == 20)){
        $job_status = $job_obj->get_one_state_attribute('state');
        $wait_count++;
        sleep(1);
    }
    my $job_id = $job_obj->get_one_state_attribute('id');
    if ($job_status ne 'Running'){
            $Log->comment("Job is not running but :  $job_status");
            return -1;
    }
    return $job_id;
}

=head2 get_destination_volume_size

Gets the minimum required size of destination volume

=over

=item Arguments

=item C<$args)>
Reference to a hash which should have keys -
vserver    (Required) - the name of the vserver on which the volume size is to
                        be determined
connection (Optional) - the node object one which volume size is to be
                        determined
src_vol_obj (Required) - the source volume object

=item Returns

=item C<$size)>
Minimum size of the destination volume

=item Usage
    $Dst_Volume_Size =
    $Cr_Utils_Obj->get_destination_volume_size(vserver       => $Dst_Vserver
                                               'src_vol_obj' => $Src_Repo_Obj);

=back

=cut

sub get_destination_volume_size {
    my ($self, %args) = @_;

    my $size;
    my $node_obj = $args{connection} || $self->{_connection};
    my $vserver  = $args{vserver} || $Options->get(option=>'VSERVER');
    my $src_volume_obj = $args{src_vol_obj};

    my $volume_size =
        $self->get_min_max_volume_size('vserver' => $vserver);
    my $src_volume_size = $src_volume_obj->get_one_state_attribute('size');
    $size = $src_volume_size > $volume_size->{'min_size'} ?
            $src_volume_size : $volume_size->{'min_size'};

    return $size;
}

=head2 ::start_packet_trace

Starts the packet trace using the utility Network::Trace

=over

=item Arguments

Reference to Testcase name, Data IP and Time Stamp

=item Examples

=back

Instance method:
    CR::CRUtils->start_packet_trace(\$test_name, $Data_IP, $Time_stamp);

=item Returns

Packet trace object

=over

=cut

sub start_packet_trace {

    my $self = shift;
    my $test = shift;
    my $data_ip = shift;
    my $time_stamp = shift;
    ${$test} = new Network::Trace (
                              address => $data_ip,
                              sudo    => '/usr/software/bin/sudo',
                              id      => ${$test} . '_' . $time_stamp,
                        );
        sleep 8; # Waiting for packet trace to start

}

=head2 :: disable_fiji

Disable whatever fault injection that is enabled.

=over

=item Arguments

connection: required for command interface

=item Examples

=back

Class method:
    CR::CRUtils->disable_fiji(connection => $node);

Instance method:
    $CRUtil->disable_fiji(connection => $node);

=item Returns

none

=over

=cut

sub disable_fiji {
    my ($self, %args) = @_;
    my $node_obj = $args{connection} || $self->{_connection};
    $Log->comment("Disable fault injections");
    my @fiji_objs =  NACL::C::DebugFaultInjection->find
                         (command_interface => $node_obj,
                          filter            => {enabled => 'true'},
                          allow_empty       => 1);

    foreach my $fiji_obj (@fiji_objs) {
        $fiji_obj->disable_fault();
    }
}


=head2 :: online_offlined_volume

Check if the infiniteVol is in mixed/offline state, then run volume online
and mount.

=over

=item Arguments

connection: required for command interface

vserver: required

junction_path: default is CVOL_JUNCTION_PATH from Options, "-" will not mount.

=item Examples

=back

Class method:
    CR::CRUtils->online_off_volume(connection => $node, 'junction_path' => "-");

Instance method:
    $CRUtil->online_off_volume(connection => $node, 'junction_path' => "/DST");

=item Returns

none

=over

=cut

sub online_offlined_volumes {
    my ($self, %args) = @_;
    my $node_obj = $args{connection} || $self->{_connection};

    $Log->comment("online volumes that are offlined");
    #burt 593848 preventis mixed state volume being returned.
    my @offlined_objs = NACL::C::Volume->find
                                (command_interface => $node_obj,
                                 filter  => {state => 'mixed|offline',
                                             'is-constituent'   => 'true' },
                                 allow_empty => 1);

    if (!@offlined_objs) {
        return;
    }

    foreach my $offlined_obj (@offlined_objs) {
        # check if volume type is DP, if so then default is no mounting
        my $def_junction_path;
        my $vserver = $offlined_obj->get_one_state_attribute('vserver');
        if ($offlined_obj->get_one_state_attribute('type') eq 'DP') {
             $def_junction_path = "-";
        } else {
             $def_junction_path = $Options->get(option=>'CVOL_JUNCTION_PATH');
        }
        my $junction_path_ref = $args{'junction_path_ref'};
        my %junction_path_hash = %$junction_path_ref;
        my $junction_path = $junction_path_hash{$vserver} || $def_junction_path;

        # online and mount volume
        if ($offlined_obj) {
            $offlined_obj->online();
            unless ($junction_path =~ /^-$/) {
                $offlined_obj->mount('junction-path' => $junction_path);
            }
        }

        # display the state of all constituent volumes
        my @constituent_vol_objs = NACL::C::Volume->find
                                      (command_interface => $node_obj,
                                       filter  => {'is-constituent' => 'true',
                                                    vserver         => $vserver,
                                       },
                                       allow_empty => 1);
        foreach my $vol_obj (@constituent_vol_objs) {
            if ($vol_obj->get_one_state_attribute('state') ne 'online') {
                $Log->comment("not all volumes are online");
            }
        }
    }
}

=head2 :: remove_snapmirror_schedule

change the schedule of the snapmirror to "".

=over

=item Arguments

connection: required for command interface

=item Examples

=back

Class method:
    CR::CRUtils->remove_snapmirror_schedule(connection => $node);

Instance method:
    $CRUtil->remove_snapmirror_schedule(connection => $node);

=item Returns

none

=over

=cut

sub remove_snapmirror_schedule {

    my ($self, %args) = @_;
    my $node_obj = $args{connection} || $self->{_connection};

    $Log->comment("remove snapmirror schedule");
    my @snapmirror_objs = NACL::C::Snapmirror->find(
                                  command_interface => $node_obj,
                                  filter    => {'is-constituent'   => 'false'},
                                  allow_empty       => 1,
                                 );
    if (@snapmirror_objs) {
        foreach my $snapmirror_obj (@snapmirror_objs) {
            my $schedule_str = $snapmirror_obj->get_one_state_attribute('schedule');
            if ($schedule_str eq '-') {
                $Log->comment("node is snapmirror source");
                next;
            } elsif ($schedule_str eq '""') {
                $Log->comment("snapmirror schedule not set");
                next;
            }
            # give it 10 minutes to wait for the snapmirror Xfer to complete
            my $retry = 20;
            my $sm_queued_job_obj;
            while ($retry > 0 && !$sm_queued_job_obj) {
                $sm_queued_job_obj = NACL::C::Job->find(
                                       command_interface => $node_obj,
                                       filter       => {state => 'Queued',
                                                        jobtype  => 'SnapMirror*'},
                                       allow_empty  => 1,
                                      );
                if (!$sm_queued_job_obj) {
                    sleep 30;
                    $retry--;
                }
            }
            try {
                $snapmirror_obj->modify(schedule => "\"\"",
                                        foreground => 'true');
            } catch NACL::APISet::Exceptions::CommandFailedException with {
                #burt 545376
                $Log->comment("snapmirror modify schedule failed, retry");
                sleep 300;
                $snapmirror_obj->modify(schedule => "\"\"",
                                        foreground => 'true');
            };
            $snapmirror_obj->get_one_state_attribute('schedule');
        }
    }
}

=head2 :: remove_snapshot_schedule

change the schedule of the snapshot of specified infiniteVol to none.

=over

=item Arguments

connection: required for command interface

=item Examples

=back

Class method:
    CR::CRUtils->remove_snapshot_schedule(connection => $node);

Instance method:
    $CRUtil->remove_snapshot_schedule(connection => $node);

=item Returns

none

=over

=cut

sub remove_snapshot_schedule {
    my ($self, %args) = @_;
    my $node_obj = $args{connection} || $self->{_connection};

    # set snapshot policy on infinitevol to be none
    $Log->comment("set snapshot policy on inifiniteVol to be none");
    my @repo_objs = NACL::C::Volume->find(command_interface => $node_obj,
                                         filter            =>
                                             {'volume-style' => 'infinitevol'},
                                         allow_empty       => 1,
                   );

    if (@repo_objs) {
        foreach my $repo_obj (@repo_objs) {
            if ($repo_obj->get_one_state_attribute('snapshot-policy') eq 'none') {
                next;
            }
            # give it 6 minute to wait for snapshot to complete
            my $retry = 24;
            my $ss_queued_job_obj;
            while ($retry > 0 && !$ss_queued_job_obj) {
                $ss_queued_job_obj = NACL::C::Job->find(
                                              command_interface => $node_obj,
                                              filter            =>
                                                  {state   => 'Queued',
                                                   jobtype => 'Snapshot*'},
                                              allow_empty       => 1,
                                     );

                if (!$ss_queued_job_obj) {
                    sleep 15;
                    $retry--;
                }
            }
            try {
                $repo_obj->modify('snapshot-policy' => 'none',
                                  foreground        => 'true');
            } catch NACL::APISet::Exceptions::CommandFailedException with {
                #burt 545376
                $Log->comment("snapshot policy nodify failed, retry");
                sleep 300;
                $repo_obj->modify('snapshot-policy' => 'none',
                                  foreground        => 'true');
            };
            $repo_obj->get_one_state_attribute('snapshot-policy');
        }
    }
}

=head2 :: run_debug_info

run the command on the client to gather debug information.
Global parameters DEBUG_CMD has to be defined.
DEBUG_CMD=command to run the client, implemented with "system" function

=over

=item Arguments

none

=item Examples

ntest sm_cleanup_test.thpl SRC_NODE=crfas3270x35a DST_NODE=crfas3270x36a
SRC_VSERVER=vs0 DST_VSERVER=vs0 DEBUG_CMD='~/p4/RR/CR/util/cr_info.pl --logdir
/x/eng/natelogs/crqanightly/dyutemp --node crfas3270x35a --config_only'`
=back

Class method:
    CR::CRUtils->run_debug_info();

Instance method:
    $CRUtil->run_debug_info();

=item Returns

none

=over

=cut

sub run_debug_info {
    my ($self, %args) = @_;
    if ($Options->get(option=>'DEBUG_CMD')) {
        my $debug_cmd = $Options->get(option=>'DEBUG_CMD');

        $Log->comment("run $debug_cmd to get debug info");
        system("/usr/software/bin/sudo $debug_cmd");
    } else {
        return;
    }
}


=head2 :: cr_cleanup

cr_cleanup calls the following functions to restore the filer.
    disable_fiji
    remove_snapshot_schedule
    remove_snapmirror_schedule
    online_offlined_volume

=over

=item Arguments

connection: required for command interface
vserver: required

=item Examples

=back

Class method:
    CR::CRUtils->cr_cleanup();

Instance method:
    $CRUtil->cr_cleanup();

=item Returns

none

=over

=cut

sub cr_cleanup {
    my ($self, %args) = @_;
    my $node_obj = $args{connection};
    if ($node_obj) {
        $Log->comment("run generic cleanup function before exit");
        CR::CRUtils->disable_fiji(connection => $node_obj);
        CR::CRUtils->remove_snapshot_schedule(connection => $node_obj );
        CR::CRUtils->remove_snapmirror_schedule(connection => $node_obj);
        CR::CRUtils->online_offlined_volumes(connection => $node_obj,
                                 junction_path_ref => $args{'junction_path_ref'}
                                            );
    }
}


=head2 :: promote_ns_mirror

 Promotes a ns mirror on $args{vserver} to become the new namespace
 constituent

=over

=item Arguments

connection: required for command interface

vserver: required

=item Examples

=back

Class method:
    CR::CRUtils->promote_ns_mirror();

Instance method:
    $CRUtil->promote_ns_mirror();

=item Returns

none

=over

=cut

sub promote_ns_mirror {

    my ($self, %args) = @_;
    my $node_obj = $args{connection} || $self->{_connection};
    my $vserver = $args{vserver};

    my $ns_obj = NACL::C::Volume->find(
        'command_interface' => $node_obj,
        'filter'            => {
            'constituent-role' => 'namespace',
            'vserver'          => $vserver,
        }
    );
    my $ns_name = $ns_obj->volume;
    my $ns_jp   = $ns_obj->get_one_state_attribute('junction-path');
    my $ns_msid = $ns_obj->get_one_state_attribute('msid');

    my $ns_mirror_obj = NACL::C::Volume->find(
        'command_interface' => $node_obj,
        filter              => {
            'constituent-role' => 'ns_mirror',
            'vserver'          => $vserver,
        }
    );
    my $ns_mirror_name = $ns_mirror_obj->volume();

    $Log->comment('Removing intracluster NS mirror SM relationship');
    my $sm_obj = NACL::STask::Snapmirror->find(
        'command_interface' => $node_obj,
        'filter'            => {
            'vserver'            => $vserver,
            'destination-volume' => $ns_mirror_obj,
        },
    );
    $sm_obj->wait_on_attribute(
                        attribute_to_check => 'status',
                        till_value         => 'Idle',
                        polling_interval   => 2
                        );

    # Do a snapmirror update
    $sm_obj->update();
    $sm_obj->break( nacltask_verify => 1 );
    $sm_obj->purge();

    $Log->comment('Modifying ns mirror filesys-size-fixed to false');
    $ns_mirror_obj->modify( 'filesys-size-fixed' => 'false' );

    $Log->comment('Offlining and deleting ns constituent');
    $ns_obj->delete( 'disable-offline-check' => 'true', force => 'true' );

    my @vreports = NACL::C::DebugVreport->find(
        'command_interface' => $node_obj,
        'allow_empty'       => 1,
    );

    if ( scalar(@vreports) > 0 ) {
        foreach my $vr (@vreports) {
            $vr->fix();
        }
    }

    $ns_mirror_obj->rename( newname => $ns_name, foreground => 'true' );

    $Log->comment( 'Unassigning ns mirror and reassigning with constituent role'
            . ' set to namespace' );
    my $repo_vol_const_obj = NACL::C::RepositoryVolumeConstituent->new(
                                   command_interface => $node_obj,
                                   'volume' => $ns_name,
                                   'vserver' => $vserver,
    );

    $repo_vol_const_obj->unassign();
    $repo_vol_const_obj->assign( 'role' => 'namespace',
                                'origMsid' => $ns_msid
    );

    $ns_obj = NACL::C::Volume->find(
        'command_interface' => $node_obj,
        'filter'            => {
            'vserver'          => $vserver,
            'constituent-role' => 'namespace',
        },
    );

    # workaround for burt598819 - change ns export policy if needed
    my $expected_ns_policy = 'repos_namespace_export_policy';
    if ( $ns_obj->get_one_state_attribute('policy') ne $expected_ns_policy ) {
        $ns_obj->modify( policy => $expected_ns_policy );
    }

    $Log->comment('Mounting the new namepsace volume');
    $ns_obj->mount( 'junction-path' => $ns_jp );

    $Log->comment('Removing any dangling redirector files');
    my $vol_check = NACL::C::VolumeCheck->find(
        'command_interface' => $node_obj,
        'allow_empty'       => 1,
        'filter'            => {
            'vserver' => $vserver,
            'volume'  => $ns_obj->volume(),
        },
    );

    $Log->comment('Repair the volume and do a Volume Check');
    my $volumecheck_obj = NACL::C::VolumeCheck->start(
                                command_interface => $node_obj,
                                'vserver' => $vserver,
                                'volume' => $ns_name,
                                'repair' => 'true'
                                );

    NACL::C::VolumeCheck->wait_on_attribute(
                        command_interface => $node_obj,
                        'vserver' => $vserver,
                        'method-timeout'   => 1500,
                        attribute_to_check => 'state',
                        till_value         => 'success',
                        polling_interval   => 2,
    );

    my $iv_vserver_obj = NACL::C::Vserver->find(
        command_interface => $node_obj,
        filter            => {
            'vserver'      => $vserver
        }
    );

    $Log->comment('Creating a new NS mirror by enabling and disabling snapdiff');
    my $iv_volume_obj = NACL::C::Volume->find(
        command_interface => $node_obj,
        filter            => {
            'volume-style' => 'infinitevol',
            'vserver'      => $vserver
        }
    );
    $iv_volume_obj->modify('enable-snapdiff' => 'true', foreground => 'true' );
    $iv_volume_obj->wait_on_attribute(
                        attribute_to_check => 'state',
                        till_value         => 'online',
                        polling_interval   => 2,
    );

    $iv_volume_obj->modify('enable-snapdiff' => 'false', foreground => 'true' );

    return;
}


=head2 :: get_target_dc

 This method is used to find a target DC volume that satisfies
 test requirements. If it's not present, it can shrink or restore
 DC size by tearing down current IV and recreating IV, and get
 the first DC volume of the newly-created IV as target DC.

=over

=item Arguments

connection: required for command interface

type: a scalar with value 'SHRINK' or 'RESTORE', means to shrink
      or restore DC size

para-ref: a reference to a hash. Because in smoke, there is always
          an IV exists. So 'infinitevol', 'vserver', and original
          IV states 'iv-attrs' are required. 'shrink-flag' is also
          a required paramter, it should be 1 at the beginning.
          Others are optional. If not specified,
          'threshold-min-dc-size' and 'max-data-constituent-size'
          will be '256MB' by default.
          {
          'existing-min-dc-size' => ...,
          'threshold-min-dc-size' => ...,
          'max-data-constituent-size' => ...,
          'shrink-flag' => ...,
          'vserver' => ...,
          'infinitevol' => ...,
          'iv-attrs' => ...
          }

=item Examples

=back

Class method:
    CR::CRUtils->get_target_dc(
        connection => ...,
        type => ...,
        all => ...,
        'para-ref' => ...
    );

Instance method:
    $CRUtil->get_target_dc(
        connection => ...,
        type => ...,
        all => ...,
        'para-ref' => ...
    );

=item Returns

One/All of the current DCs that satisfies test requirement or the
first/all DC(s) of the newly-created IV after shrinking/restoring
DC size

=over

=cut

sub get_target_dc {

    my ($self, %args) = @_;
    my $node_obj      = $args{connection} || $self->{_connection};
    my $type          = $args{type};
    my $all_flag      = $args{all};
    my $ref           = $args{'para-ref'};
    my $iv_obj        = $ref->{infinitevol};
    my %opts = %{$ref->{'iv-attrs'}};

    my $threshold_min_dc_size = $ref->{'threshold-min-dc-size'} || '256MB';
    my $max_dc_size           = $ref->{'max-data-constituent-size'} || '256MB';
    my $api_set_obj           = $node_obj->apiset();

    my @dc_vol_objs;
    my $volume = 'repo_vol';
    my $satisfy_flag_for_all_dcs = 1;

    # If infinite volume exists before test starts, get the IV attributes
    $volume = $iv_obj->state()->volume() if ($iv_obj);

    if ($type eq 'SHRINK') {

        $Log->comment('Get all DC volumes of the current IV');

        # Get all DC volumes of the current IV
        @dc_vol_objs = NACL::C::Volume->find(
            command_interface => $node_obj,
            filter            => {
                'constituent-role' => 'data',
                vserver            => $ref->{'vserver'},
            }
        );

        $Log->comment(
            'Loop through existing DC to find whehter any DC satisfies test '
            . 'requirement'
        );

        # Loop through exsiting DC to find whether any DC satisfies test
        # requirement
        foreach (@dc_vol_objs) {
            my $size = $_->get_one_state_attribute('size');
            my $converted_size = NACL::C::UnitNormalization::to_megabytes(
                $size
            );
            my $converted_thres_size = NACL::C::UnitNormalization::to_megabytes(
                $threshold_min_dc_size
            );
            $converted_size =~ s{(\d+)MB}{$1}e;
            $converted_thres_size =~ s{(\d+)MB}{$1}e;

            my $used_percent = $_->get_one_state_attribute('percent-used');
            $used_percent =~ s{(\d+)[%]}{($1*0.01)}e;

            # If current DC has a size smaller or equal to
            # max-data-constituent-size and used-percent below 20%, return it as
            # target DC volume
            if($converted_size <= $converted_thres_size && $used_percent < 0.2){
                my $name = $_->volume();

                $Log->comment(
                    "No need to shrink DC size, DC volume $name satisfies test "
                    . "requirement\n"
                );

                # Set the shrink flag to be 0, thus no restoring DC action
                # is needed to be done
                $ref->{'shrink-flag'} = 0;

                if (!$all_flag) {
                    return $_;
                }
            } else {

                # If there is any DC that doesn't satisfy test requirement,
                # set this flag to 0
                $satisfy_flag_for_all_dcs = 0;
            }
        }

        # If all DC satisfy test requirement, put them into @target_dcs
        if ($satisfy_flag_for_all_dcs) {
            if ($all_flag) {
                return @dc_vol_objs;
            }
        }

        $Log->comment("Shrink DC size to satisfy test requirement...\n");

        # If no current DC is in EMPTY state and equal to/smaller than
        # max-data-constituent-size, do the following

        # Get the original state of repos_dc_min_size table
        my $result = CR::C::DebugSmdbTableReposDcMinSize->find(
            command_interface => $node_obj,
            allow_empty => 1
        );

        # If there is an entry for repos_dc_min_size table, record the
        # existing value for min-data-constituent-size field
        if ($result) {
            $ref->{'existing-min-dc-size'} =
                $result->get_one_state_attribute('min-data-constituent-size');
            my $size_in_bytes = $ref->{'existing-min-dc-size'};
            $ref->{'existing-min-dc-size'} =
                NACL::C::UnitNormalization::to_megabytes($size_in_bytes);
        }

        if ( $ref->{'existing-min-dc-size'} ne $threshold_min_dc_size) {

            $Log->comment(
                'If entry exists for repos_dc_min_size table, delete it'
            );

            # If entry exists for repos_dc_min_size table, delete it
            if ( $ref->{'existing-min-dc-size'} ) {
                $api_set_obj->execute_raw_command(
                    'command' => 'debug smdb table repos_dc_min_size delete'
                    . ' *'
                );
            }

            $Log->comment(
                'Create a new entry in repos_dc_min_size table, with '
                . "min-data-constituent-size to be $threshold_min_dc_size"
            );

            # No matter entry exists or not, create a new entry with
            # min-data-constituent-size to be threshold-min-dc-size
            $api_set_obj->execute_raw_command(
                'command' => 'debug smdb table repos_dc_min_size create '
                . "-min-data-constituent-size $threshold_min_dc_size"
            );
        }

        $Log->comment('Purge the current IV...');

        # Delete the IV
        NACL::STask::Volume->purge(
            command_interface => $node_obj,
            volume            => $volume,
            vserver           => $ref->{'vserver'},
            'method-timeout'  => 1000,
        ) if ($iv_obj);

        $Log->comment(
            'Create a new IV after manipulating repos_dc_min_size table if '
            . 'test is running on filer, with max-data-constituent-size '
            . 'specified when creation'
        );

        # Create a new IV, with max-data-consitituent-size specified
        $iv_obj = NACL::C::Volume->create(
            command_interface  => $node_obj,
            volume             => $volume,
            vserver            => $ref->{'vserver'},
            'max-data-constituent-size' => $max_dc_size,
            'unix-permissions' => '---rwxrwxrwx',
            foreground         => 'true',
            'method-timeout' => 1000,
        );
    } # end if($type eq 'SHRINK')

    if ($type eq 'RESTORE' && $ref->{'shrink-flag'} == 0) {
        $Log->comment(
            'No need to restore DC size, because Shrinking DC size was not '
            . "performed\n"
        );
    }

    if ($type eq 'RESTORE' && $ref->{'shrink-flag'} != 0) {

        $Log->comment("Restore DC size to original...\n");

        # Deleting IV should be performed before setting repos_dc_min_size table
        # For example, if threshold-min-dc-size & max-data-constituent-size are
        # 256MB. The original repos-dc-min-size is 1TB. After shrinking DC size,
        # each DC is 256MB. 'debug smdb table repos_dc_min_size create
        # -min-data-constituent-size 1TB' won't work for the current filer
        # configuration during restoring

        $Log->comment('Purge the current IV...');

        # Delete the IV
        NACL::STask::Volume->purge(
            command_interface => $node_obj,
            volume            => $volume,
            vserver           => $ref->{'vserver'},
            'method-timeout'  => 1000,
        ) if ($iv_obj);

        $Log->comment('Restore repos_dc_min_size table to its original state');

        if ( $ref->{'existing-min-dc-size'} ne $threshold_min_dc_size ) {

            # Delete the current entry in repos_dc_min_size table
            $api_set_obj->execute_raw_command(
                'command' => 'debug smdb table repos_dc_min_size delete *'
            );

            # If there is an entry exists before test cass starts, restore
            # it to its original
            if ( $ref->{'existing-min-dc-size'} ) {
                my $min_dc_size = $ref->{'existing-min-dc-size'};
                $api_set_obj->execute_raw_command(
                    'command' => 'debug smdb table repos_dc_min_size create'
                    ." -min-data-constituent-size $min_dc_size"
                );
            }
        }

        $Log->comment('Create a new IV after restoring repos_dc_min_size table,'
        . ' without specifying max-data-constituent-size when creation'
        );

        # Create a new IV, without max-data-constituent-size specified
        $iv_obj = NACL::C::Volume->create(
            command_interface  => $node_obj,
            volume             => $volume,
            vserver            => $ref->{'vserver'},
            'unix-permissions' => '---rwxrwxrwx',
            foreground         => 'true',
            'method-timeout' => 1000,
            %opts
        );

    } #end if ($type eq 'RESTORE')

    # Get all DC volumes of the newly-created IV
    @dc_vol_objs = NACL::C::Volume->find(
         command_interface => $node_obj,
         filter            => {
             'constituent-role' => 'data',
              vserver            => $ref->{'vserver'},
         }
    );

    if ($all_flag == 1) {
        return @dc_vol_objs;
    } else {
        if ($type eq 'SHRINK') {
            my $chosen_dc_name = $dc_vol_objs[0]->volume();

            $Log->comment(
                "After shrinking DC size, DC volume $chosen_dc_name was chosen "
                . "as target DC volume\n"
            );

        }

        return $dc_vol_objs[0];
    }
}
1;
