## Copyright (c) 2012 NetApp Inc.
## All rights reserved.
##
## @name CdmiDataObject.pm
##
## @summary CDMI Data object requests
##
## @description
##   Implements CDMI Data object requests
##
## @keywords cr, cdmi, http, dataobject
##
## @author Pavan Verma
##
## @example
##
## @status private

=pod

=head1 Package Name

CR::CdmiDataObject

=head1 Author/Maintainer

Pavan Verma

=head1 Description
Implements CDMI Data object requests

=cut

package CR::CdmiDataObject;

# Compiler directives
use strict;
use warnings;

use NATE::Log qw(log_global);
use WWW::Curl;
use WWW::Curl::Easy;
use HTTP::Response qw();
use File::stat;

my $Log = log_global();

################################################################################
# Read a single object
################################################################################
sub Read_Object
{
    my $host = shift(@_);
    my $port = shift(@_);
    my $objname = shift(@_);
    my $curl = shift(@_);

    #my $url = 'http://' . $host . ':' . $port . '/cdmi_objectid/' . $objname;
    my $url = 'http://' . $host . ':' . $port . $objname;
    $curl->setopt(CURLOPT_URL, $url);

    my $response_body;
    $curl->setopt(CURLOPT_WRITEDATA,\$response_body);

    # Starts the actual request
    my $retcode = $curl->perform;

    # Looking at the results...
    if ($retcode == 0)
    {
        my $response_code = $curl->getinfo(CURLINFO_HTTP_CODE);
        if ($response_code != 200)
        {
            $Log->error("Read object request failed with error $response_code");
            return 1;
        }
        # judge result and next action based on $response_code
    }
    else
    {
        # Error code, type of error, error message
        my $errstr = $curl->strerror($retcode);
        my $errbuf = $curl->errbuf;
        $Log->error("An error happened: $retcode, $errstr, $errbuf");
        return 1;
    }

    return 0;
}

################################################################################
# Update a single object
################################################################################
sub Update_Object
{
    my $host = shift(@_);
    my $port = shift(@_);
    my $objname = shift(@_);
    my $datafile = shift(@_);
    my $curl = shift(@_);

    #my $url = 'http://' . $host . ':' . $port . '/cdmi_objectid/' . $objname;
    my $url = 'http://' . $host . ':' . $port . $objname;
    $curl->setopt(CURLOPT_URL, $url);

    open (my $infile, $datafile) or $Log->error('$!');
    $curl->setopt(CURLOPT_READDATA, $infile);
    my $sb = stat($datafile);
    $curl->setopt(CURLOPT_INFILESIZE, $sb->size);

    my $response_body;
    $curl->setopt(CURLOPT_WRITEDATA,\$response_body);

    # Starts the actual request
    my $retcode = $curl->perform;

    # Looking at the results...
    if ($retcode == 0)
    {
        my $response_code = $curl->getinfo(CURLINFO_HTTP_CODE);
        if ($response_code != 204)
        {
            $Log->error("Update object request failed with error $response_code");
            return 1;
        }
        # judge result and next action based on $response_code
    }
    else
    {
        # Error code, type of error, error message
        my $errstr = $curl->strerror($retcode);
        my $errbuf = $curl->errbuf;
        $Log->error("An error happened: $retcode, $errstr, $errbuf");
        return 1;
    }

    return 0;
}

################################################################################
# Delete a single object
################################################################################
sub Delete_Object
{
    my $host = shift(@_);
    my $port = shift(@_);
    my $objname = shift(@_);
    my $curl = shift(@_);

    #my $url = 'http://' . $host . ':' . $port . '/cdmi_objectid/' . $objname;
    my $url = 'http://' . $host . ':' . $port . $objname;
    $curl->setopt(CURLOPT_URL, $url);

    my $response_body;
    $curl->setopt(CURLOPT_WRITEDATA,\$response_body);

    # Starts the actual request
    my $retcode = $curl->perform;

    # Looking at the results...
    if ($retcode == 0)
    {
        my $response_code = $curl->getinfo(CURLINFO_HTTP_CODE);
        if ($response_code != 204)
        {
            $Log->error("Delete object request failed with error $response_code");
            return 1;
        }
        # judge result and next action based on $response_code
    }
    else
    {
        # Error code, type of error, error message
        my $errstr = $curl->strerror($retcode);
        my $errbuf = $curl->errbuf;
        $Log->error("An error happened: $retcode, $errstr, $errbuf");
        return 1;
    }

    return 0;
}

################################################################################
# Read the objects specified in input list
################################################################################
sub Read_Many
{
    my $host = shift(@_);
    my $port = shift(@_);
    my $objsArrRef = shift(@_);

    my $count = scalar(@{$objsArrRef});
    my $curl = WWW::Curl::Easy->new;

    $curl->setopt(CURLOPT_HEADER, 1);
    $curl->setopt(CURLOPT_CONNECTTIMEOUT, 15);
    $curl->setopt(CURLOPT_TIMEOUT, 30);

    for (my $i = 0; $i < $count; $i++)
    {
        my $result = Read_Object($host, $port, ${$objsArrRef}[$i], $curl);
        if ($result != 0)
        {
            $Log->error("Exiting because of error $result");
            return $i;
        }
    }

    return $count;
}

################################################################################
# Update the objects specified in input list
################################################################################
sub Update_Many
{
    my $host = shift(@_);
    my $port = shift(@_);
    my $datafile = shift(@_);
    my $objsArrRef = shift(@_);

    my $count = scalar(@{$objsArrRef});
    my $curl = WWW::Curl::Easy->new;

    $curl->setopt(CURLOPT_HEADER, 1);
    $curl->setopt(CURLOPT_CONNECTTIMEOUT, 15);
    $curl->setopt(CURLOPT_TIMEOUT, 30);
    $curl->setopt(CURLOPT_PUT, 1);

    my @headers;
    $headers[0] = "Expect:";
    $headers[1] = "Content-Type: text/plain";
    $curl->setopt(CURLOPT_HTTPHEADER, \@headers);

    for (my $i = 0; $i < $count; $i++)
    {
        my $result = Update_Object($host, $port, ${$objsArrRef}[$i], $datafile, $curl);
        if ($result != 0)
        {
            $Log->error("Exiting because of error $result");
            return $i;
        }
   }

    return $count;
}

################################################################################
# Delete the objects specified in input list
################################################################################
sub Delete_Many
{
    my $host = shift(@_);
    my $port = shift(@_);
    my $objsArrRef = shift(@_);

    my $count = scalar(@{$objsArrRef});
    my $curl = WWW::Curl::Easy->new;

    $curl->setopt(CURLOPT_HEADER, 1);
    $curl->setopt(CURLOPT_CONNECTTIMEOUT, 15);
    $curl->setopt(CURLOPT_TIMEOUT, 30);
    $curl->setopt(CURLOPT_CUSTOMREQUEST, "DELETE");

    for (my $i = 0; $i < $count; $i++)
    {
        my $result = Delete_Object($host, $port, ${$objsArrRef}[$i], $curl);
        if ($result != 0)
        {
            $Log->error("Exiting because of error $result");
            return $i;
        }
   }

    return $count;
}

1;
