## Copyright (c) 2012 NetApp Inc.
## All rights reserved.
##
## @name CdmiDataObjectNamed.pm
##
## @summary Named CDMI Data object requests
##
## @description
##   Implements Named CDMI Data object requests
##
## @keywords cr, cdmi, http, named dataobject
##
## @author Pavan Verma
##
## @example
##
## @status private

=pod

=head1 Package Name

CR::CdmiDataObjectNamed

=head1 Author/Maintainer

Pavan Verma

=head1 Description
Implements Named CDMI Data object requests

=cut

package CR::CdmiDataObjectNamed;

# Compiler directives
use strict;
use warnings;

use NATE::Log qw(log_global);
use WWW::Curl;
use WWW::Curl::Easy;
use HTTP::Response qw();
use File::stat;

my $Log = log_global();

################################################################################
# Create a single named object
################################################################################
sub Create_Object
{
    my $host = shift(@_);
    my $port = shift(@_);
    my $datafile = shift(@_);
    my $objname = shift(@_);

    my $curl = WWW::Curl::Easy->new;

    $curl->setopt(CURLOPT_HEADER, 1);
    $curl->setopt(CURLOPT_PUT, 1);
    $curl->setopt(CURLOPT_CONNECTTIMEOUT, 15);
    $curl->setopt(CURLOPT_TIMEOUT, 30);

    my @headers;
    $headers[0] = "Expect:";
    $headers[1] = "Content-Type: text/plain";
    $curl->setopt(CURLOPT_HTTPHEADER, \@headers);

    (substr($objname, 0, 1) eq "/") or $Log->error("Object name must start with slash");
    my $url = 'http://' . $host . ':' . $port . $objname;

    $curl->setopt(CURLOPT_URL, $url);

    open (my $infile, $datafile) or $Log->error('$!');

    $curl->setopt(CURLOPT_READDATA, $infile);

    my $response_body;
    $curl->setopt(CURLOPT_WRITEDATA,\$response_body);

    # Starts the actual request
    my $retcode = $curl->perform;

    close $infile;

    # Looking at the results...
    if ($retcode == 0)
    {
        my $response_code = $curl->getinfo(CURLINFO_HTTP_CODE);
        if ($response_code != 201)
        {
            $Log->error("Create named object request failed with error $response_code");
            return '';
        }
        return $objname;
    }
    else
    {
        # Error code, type of error, error message
        my $errstr = $curl->strerror($retcode);
        my $errbuf = $curl->errbuf;
        $Log->error("An error happened: $retcode, $errstr, $errbuf");
        return '';
    }
}

################################################################################
# Create a bunch of named objects
################################################################################
sub Create_Many
{
    my $host = shift(@_);
    my $port = shift(@_);
    my $datafile = shift(@_);
    my $count = shift(@_);
    my $container = shift(@_);
    my $objname_prefix = shift(@_);

    my @objects;
    my $objectname;
    my $retval;

    # Add trailing slash to container name if required
    if (substr($container, length($container) - 1, 1) ne "/")
    {
        $container .= "/";
    }

    for (my $i = 0; $i < $count; $i++)
    {
        $objectname = $container . $objname_prefix . 'obj_' . $i;
        $retval = Create_Object($host, $port, $datafile, $objectname);
        if ($retval ne '')
        {
            push(@objects, $objectname);
        }
        else
        {
            $Log->error("Exiting because of error");
            return $i;
        }
    }

    return ($count, @objects);
}

1;
