# Hey emacs, this is basically -*- perl -*-
#
# Copyright (c) 2010 Network Appliance, Inc.
# All rights reserved.
#
# Content Repository/CheckerUtils.pm

package CR::CheckerUtils;

# Compiler directives.
use strict;
use warnings;
use Options;
use Tharn;

##------------------------------------------------------------
# Module Imports
##-------------------------------------------------------------

use base qw(TCD);
use TCD;
use Tharn qw($Log);
use Data::Dumper;
use NACL::APISet;
use NACL::C::Node;
use NACL::C::Volume;
use NACL::Transit;
use NACL::C::Cluster;
use CR::PosixUtils;
use NATE::BaseException qw(:try);
use Params::Validate qw(:all);
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::STask::Client::Mount;
use NACL::C::Client;
use NACL::C::CommandInterface;
use NACL::C::Component;
use File::Spec;
use NACL::C::SystemNode;
use NACL::C::Aggregate;
use NACL::C::VolumeCheck;
use NACL::STask::Vserver;
use MIME::Base64;
use Carp qw(croak);
use Config;
use NACL::CS::NetworkInterface;

# The external interface into this library
our @EXPORT = qw(
  volumeOffline
  volumeRestrict
  inode_num
  change_node_state
  disable_fijis
  checkerutils_setup
  checkerutils_cleanup
  new_file
  start_scan
  get_data_file
  new
);

# Global variables
use vars qw(
  $Test
  $options
  $Client
  @Cleanup
  $vserver
  $client
  $mount
  $ns_msid
);

########Get the inode number for a directory#########
sub inode_num {
    my ( $node, $volume, $subdir ) = @_;
    my $inode_no;

    $Log->comment("Routine inode_num called with @_");
    eval {

        my $Api_Set_Obj = NACL::APISet->new(
            name      => $node,
            category  => "Node",
            interface => "CLI",
            set       => "Nodescope",
        );
        my $dir_path = "/vol/" . $volume . "/" . $subdir;
        my @parsed_output =
          @{ $Api_Set_Obj->slist( path => $dir_path )->get_parsed_output() };

        $inode_no = $parsed_output[0]->{inode};
        $Log->comment("inode_number : $inode_no");

    };
    if ($@) {
        $Log->comment("inode_num() failed: $@");
    }
    return ($inode_no);
}    #end of inode_num

#############Change node state###############
sub change_node_state {
    my ( $node, $state ) = @_;

    $Log->comment("Routine change_node_state called with @_");

    # STATES : CLI, FIRMWARE, USERNAME

    # Open a transit connection
    eval {
        sleep(15);
        my $transit_obj = NACL::Transit->new( name => $node, );

        my $current_state = $transit_obj->get_state;
        $Log->comment("current state :$current_state");

        $current_state = $transit_obj->change_state( to => $state );
        $Log->comment("current state :$current_state");
    };
    if ($@) {
        die "change_node_state failed: $@";
    }

}

###### Disable all the fijis on all CR nodes   ####
# Usage: disable_fijis("fiji1", "fiji2", ..."fijiN");

sub disable_fijis {
    $Log->comment("disable_fijis() called ");
    my $self = shift;
    my @fijis = @_;
    eval {
        foreach my $fiji (@fijis)
        {

            #disable all fijis on cluster node
            NACL::C::DebugFaultInjection->disable_fault(
                command_interface => $self->{node},
                'fault-id'        => $fiji,
                'node'            => '*'
            );
        }
    };
    if ($@) {
        die "disable_fijis() failed: $@";
    }
}

##########Vol unmount ############
# Usage: volumeUnmount($volume_name)
######################################
sub volumeUnmount {
    $Log->comment("Routine volumeUnmount called with @_");
    my $self = shift;
    my %opts = validate(
        @_,
        {   volume            => { type => SCALAR,  optional => 0 },
            vserver           => { type => SCALAR,  optional => 0 },
        }
    );

    $Log->comment("Unmounting the volume $opts{volume}...");
    # Find the volume we want to unmount
    my $unmount_volume = NACL::C::Volume->find(
        command_interface => $self->{node},
        filter            => {
            'vserver' => $opts{vserver},
            'volume'  => $opts{volume},
        },
    );
    $unmount_volume->unmount();
    $unmount_volume->wait_on_attribute( 'method-timeout'=> '60',attribute_to_check => 'junction-active', till_value => '-' );
    $Log->comment("Status : Volume $opts{volume} is unmounted");
    # Below commands are executed as a workaround for CLUSFS issue
    $self->{node_shell}->execute_raw_command(command => 'cd /clus/'.$opts{vserver});
    $self->{node_shell}->execute_raw_command(command => 'ls');
}

sub volumeMount {
    $Log->comment("Routine volumeMount called with @_");
    my $self = shift;
    my %opts = validate(
        @_,
        {   volume     => { type => SCALAR,  optional => 0 },
            vserver    => { type => SCALAR,  optional => 0 },
            junc_path  => { type => SCALAR,  optional => 1, default => 1 },
        }
    );
    my $junc_path   = $opts{junc_path};

    $Log->comment("Mounting the volume $opts{volume}...");
    my $mount_volume = NACL::C::Volume->find(
        command_interface => $self->{node},
        filter            => {
            'vserver' => $opts{vserver},
            'volume'  => $opts{volume},
        },
    );
    if ($opts{junc_path} eq '1'){
        #Check volume type.
        my $role = $mount_volume->state(requested_fields => [ qw(constituent-role) ]);
        if ($role->constituent_role() =~ /data/){
            #Junction path for Object volume
            if ($opts{volume} =~ /repo_vol_(.*)/){
                $junc_path = '/.system/constituents/'.$1;
            }else {
                $junc_path = '/.system/constituents/'.$opts{volume};
            }
        } else {
            #Junction path for Namespace volume
            $junc_path = "/NS";
        }
    }
    $mount_volume->mount('junction-path' => $junc_path);
    $mount_volume->wait_on_attribute( 'method-timeout'=> '60',attribute_to_check => 'junction-path', till_value => $junc_path );
    sleep(5);
    $Log->comment("Status : Volume $opts{'volume'} is mounted");
    # Below commands are executed as a workaround for CLUSFS issue
    $self->{node_shell}->execute_raw_command(command => 'cd /clus/'.$opts{vserver});
    $self->{node_shell}->execute_raw_command(command => 'ls');
    # ZAPI takes min 60 sec on filer to work properly after the mount though show command displays junction-path is mounted
    #sleep(90);
}

##########Vol offline ############
# Usage: volumeOffline($volume_name)
######################################
sub volumeOffline {
    $Log->comment("Routine volumeOffline called with @_");
    my %opts = validate(
        @_,
        {   volume            => { type => SCALAR,  optional => 0 },
            vserver           => { type => SCALAR,  optional => 0 },
            command_interface => { type => HASHREF, optional => 0 },
        }
    );

    $Log->comment("Modifying the status of the volume $opts{volume} to offline...");
    # Find the volume we want to turn offline
    my $offline_volume = NACL::C::Volume->find(
        command_interface => $opts{command_interface},
        filter            => {
            'vserver' => $opts{vserver},
            'volume'  => $opts{volume},
        },
    );
    sleep (5);
    # Unmount the volume if it's not unmounted
    if ($offline_volume->get_one_state_attribute('junction-active') eq 'true') {
        $offline_volume->unmount();
        $offline_volume->wait_on_attribute( 'method-timeout'=> '60',attribute_to_check => 'junction-active', till_value => '-' );
    }
    $offline_volume->offline();
    $offline_volume->wait_on_attribute( 'method-timeout'=> '60',attribute_to_check => 'state', till_value => 'offline' );
    $Log->comment("Status : Volume $opts{volume} is Offline");
}

##########Vol restrict ############
# Usage: volumeRestrict($volume_name)
######################################
sub volumeRestrict {
    my %opts = validate(
        @_,
        {   volume            => { type => SCALAR,  optional => 0 },
            vserver           => { type => SCALAR,  optional => 0 },
            command_interface => { type => HASHREF, optional => 0 },
        }
    );
    $Log->comment("Modifying the status of the volume $opts{volume} to restricted...");
    # Find the volume we want to turn restrict
    my $restrict_volume = NACL::C::Volume->find(
        command_interface => $opts{command_interface},
        filter            => {
            'vserver' => $opts{vserver},
            'volume'  => $opts{volume}
        },
    );
    # Unmount the volume if it's not unmounted
    if ($restrict_volume->get_one_state_attribute('junction-active') eq 'true') {
        $restrict_volume->unmount();
        $restrict_volume->wait_on_attribute( 'method-timeout'=> '60',attribute_to_check => 'junction-active', till_value => '-' );
    }
    $restrict_volume->restrict();
    $restrict_volume->wait_on_attribute( 'method-timeout'=> '60',attribute_to_check => 'state', till_value => 'restricted' );
    $Log->comment("Status : Volume $opts{volume} is Restricted");
}

##########Vol Online ############
# Usage: volumeOnline($volume_name)
######################################
sub volumeOnline {
    my $self = shift;
    my %opts = validate(
        @_,
        {   volume  => { type => SCALAR, optional => 0 },
            vserver => { type => SCALAR, optional => 0 },
            command_interface =>
                { type => HASHREF, default => $self->{node} },
        }
    );
    $Log->comment("Modifying the status of the volume $opts{volume} to online...");
    # Find the volume we want to turn online
    my $online_volume = NACL::C::Volume->find(
        command_interface => $opts{command_interface},
        filter            => {
            'vserver' => $opts{vserver},
            'volume'  => $opts{volume}
        },
    );
    # if volume is not mounted first make the volume online and then mount the volume
    if ($online_volume->get_one_state_attribute('junction-active') eq 'true') {
        $Log->comment("Volume : $opts{volume} is already online");
    } else {
        $online_volume->online();
        $online_volume->wait_on_attribute( 'method-timeout'=> '60',attribute_to_check => 'state', till_value => 'online' );
        $Log->comment("Status : Volume $opts{volume} is Online");
        $self->volumeMount(
            volume            => $opts{volume},
            vserver           => $opts{vserver},
        );
    }
}

########## Get Node For an Aggregate ############
# Usage: get_node_name_for_aggr($aggr_name)
######################################
sub get_node_name_for_aggr {
    my $self      = shift;
    my $aggr_name = shift;

    my @aggr = NACL::C::Aggregate->find(
        command_interface => $self->{node},
        filter            => { aggregate => $aggr_name }
    );
    my $node_name = $aggr[0]->state()->nodes();
    return $$node_name[0];
}

##########################################################################################################
##########################################################################################################

sub new {
    $Log->comment("new called");
    my $class = shift;
    my $self  = {};

    $self->{node} = NACL::C::Node->find();

    my $vserver = NACL::STask::Vserver->find(
        command_interface => $self->{node},
        filter            => { 'is-repository' => 'true', }
    );
    $self->{cifs_share} = $main::CIFS_SHARE || "cifs_share";

    my $output = NACL::CS::NetworkInterface->fetch(
        command_interface => $self->{node},
        filter            => {
            'vserver'      => $vserver->vserver,
            'role'         => 'data',
            'status-oper'  => 'up',
            'status-admin' => 'up'
        },
        requested_fields => ['address'],
    );

    $self->{data_ip} = $output->address();

    # Find the clients here
    $Log->comment("Find the Windows client...");
    $client = NACL::C::Client->new( name => $main::WINCLIENT );
    if ( !$client->is_windows() ) {
        $Log->error("It is not a Windows client");
        return $TCD::CONFIG;
    }

    $Log->comment("Creating objects to communicate with filer");
    $self->{node_cli} = NACL::APISet->new(
        hostobj      => $self->{node}->hostrec(),
        category     => 'Node',
        interface    => 'CLI',
        set          => 'CMode',
    );

    $self->{node_shell} = NACL::APISet->new(
        hostobj      => $self->{node}->hostrec(),
        category     => 'Node',
        interface    => 'CLI',
        set          => 'Systemshell',
        shell_type   => 'CMode',
        version_info => { "cdefs_from_build" => 0 },
    );

    $self->{client}     = $client;
    $self->{vserver}    = $vserver->vserver;
    $self->{ns_vol}     = undef;
    $self->{ns_msid}    = undef;
    $self->{drive}      = undef;
    $self->{mount}      = undef;
    bless $self, $class;

    return $self;
}

sub checkerutils_setup {
    $Log->comment("checkerutils_setup called");
    my $self   = shift;
    my %opts = validate(
        @_,
        {   'cifs_domain' => { type => SCALAR, default => 'CNC' },
            'cifs_user'   => { type => SCALAR, default => 'adadmin' },
            'cifs_passwd' => { type => SCALAR, default => 'adadmin' },
        }
    );

    $self->{mount} = NACL::STask::Client::Mount->create(
        command_interface        => $self->{client},
        server_command_interface => $self->{node},
        share                    => $self->{cifs_share},
        vserver                  => $self->{vserver},
        user                     => $opts{cifs_domain} . '\\' . $opts{cifs_user},
        password                 => $opts{cifs_passwd},
        persistent               => 'no',
    );
    $self->{drive} = $self->{mount}->get_one_state_attribute('mount_point');

    my $ns = NACL::C::Volume->find(
        command_interface => $self->{node},
        filter            => {
            vserver            => $self->{vserver},
            'constituent-role' => 'namespace',
        },
    );
    $self->{ns_vol}  = $ns->volume;
    $self->{ns_msid} = $ns->get_one_state_attribute('msid');

    return 1;
}

sub checkerutils_cleanup {
    $Log->comment("checkerutils_cleanup called");
    my $self   = shift;
    my $result = 0;

    if ( $self->{mount} ) {
        NACL::C::Client::Mount->delete(
            command_interface => $self->{client},
            mount_point =>
              $self->{mount}->get_one_state_attribute('mount_point')
        );
    }
    else {
        $Log->comment("No mount point found.");
    }
    $result = 1;

    return $result;
}

sub start_scan {
    $Log->comment("start_scan() called with @_");
    my $self = shift;
    my %opts  = validate(
        @_,
        {
            'volume'    => { type => SCALAR, },
            'file-path' => { type => SCALAR, optional => 1 },
            'repair'    => { type => SCALAR, },
            'error'     => { type => SCALAR, optional => 1 },
            'message'   => { type => SCALAR, optional => 1 },
        }
    );
    my $response;
    my $volume    = $opts{'volume'};
    my $file_path = $opts{'file-path'};
    my $repair    = $opts{'repair'};
    my $error     = $opts{'error'};
    my $message   = $opts{'message'};
    try {
        $response = $self->{node_cli}->volume_check_start(
            vserver     => $self->{vserver},
            volume      => $volume,
            'file-path' => $file_path,
            'repair'    => $repair,
        );

    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        $response = $exception_object->get_response_object();
        $Log->comment( "Caught a " . ref($exception_object) . "!" );
        $Log->comment( "error is: " . $exception_object->text() );
        if ( index( $exception_object->text(), $response ) < 0 ) {
            $Log->comment("Different error received");
        }
    };
    $Log->comment( "response OUTPUT :\n " . Dumper $response);
    return $response;
}

sub file_exist {
    $Log->comment("corrupt_file() called with @_");
    my $self = shift;

    my %opts = validate(
        @_,
        {
            filepath => { type => SCALAR, },
            volume   => { type => SCALAR, },
        }
    );
    my $filepath = $opts{filepath};
    my $volume   = $opts{volume};
    my $status   = 0;
    try {
        my $fullpath =
          File::Spec->catfile( $self->get_junc_path($volume), $filepath );
        my $output = $self->{node_cli}->volume_file_show_fhmap(
            'vserver' => $self->{vserver},
            'path'    => $fullpath,
        );
        $status = 1;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        my $resp             = $exception_object->get_response_object();
        $Log->comment( "Caught a " . ref($exception_object) . "!" );
        $Log->comment( "error is: " . $exception_object->text() );
        if (
            index( $exception_object->text(), 'No such file or directory' ) <
            0 )
        {
            $Log->comment("Different error received");
        }
    };

    return $status;
}

sub corrupt_file {
    $Log->comment("corrupt_file() called with @_");
    my $self = shift;

    my %opts = validate(
        @_,
        {
            filepath    => { type => SCALAR, },
            type        => { type => SCALAR, },
            corruptpath => { type => SCALAR, optional => 1 },
        }
    );
    my $filepath    = $opts{filepath};
    my $type        = $opts{type};
    my $corruptpath = $opts{corruptpath};
    my $cmd;
    if ( $type =~ /corrupt_redirector/ ) {
        $cmd = "sudo echo corrupting > $filepath";
    }
    elsif ( $type =~ /corrupt_data_file/ ) {
        $cmd = "sudo mv $filepath $corruptpath";
    }
    elsif ($type =~ /dangling_redirector/
        || $type =~ /dangling_data_file/ )
    {
        $cmd = "sudo rm -rf $filepath";
    }
    my $output = $self->{node_shell}->execute_raw_command( command => $cmd );
    $Log->comment( "SHELL COMMAND OUTPUT :\n " . Dumper $output);
    if ( $output ne '' ) {
        $Log->comment("Corruption on $filepath failed");
        return 0;
    }
    $Log->comment("Corruption on $filepath is successful");
    return 1;
}

sub dangling_base_redirector {
    $Log->comment("dangling_base_redirector() called with @_");
    my $self = shift;

    my %opts = validate(
        @_,
        {
            files => { type => HASHREF, default => '' },
            filename =>
              { type => SCALAR, optional => 1, dependencies => [qw(volume )], },
            volume => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(filename )]
            },
        }
    );
    my ( $filename, $volume );
    my $files = $opts{files};
    if ($files) {
        $filename = $files->{'base'}->{'data'}->{'name'};
        $volume   = $files->{'base'}->{'data'}->{'volume'};
    }
    else {
        $filename = $opts{filename};
        $volume   = $opts{volume};
    }
    my $filepath =
      File::Spec->catfile( '/clus', $self->{vserver}, $self->get_junc_path($volume),
        $filename );
    return (
        $self->corrupt_file(
            filepath => $filepath,
            type     => 'dangling_redirector'
        )
    );
}

sub dangling_base_data_file {
    $Log->comment("dangling_base_data_file() called with @_");
    my $self = shift;

    my %opts = validate(
        @_,
        {
            files => { type => HASHREF, default => '' },
            filename =>
              { type => SCALAR, optional => 1, dependencies => [qw(volume )], },
            volume => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(filename )]
            },
        }
    );
    my ( $filename, $volume );
    my $files = $opts{files};
    if ($files) {
        $filename = $files->{'base'}->{'redirector'}->{'name'};
        $volume   = $files->{'base'}->{'redirector'}->{'volume'};
    }
    else {
        $filename = $opts{filename};
        $volume   = $opts{volume};
    }
    my $filepath =
      File::Spec->catfile( '/clus', $self->{vserver}, $self->get_junc_path($volume),
        $filename );
    return (
        $self->corrupt_file(
            filepath => $filepath,
            type     => 'dangling_data_file'
        )
    );
}

sub dangling_stream_redirector {
    $Log->comment("dangling_stream_redirector() called with @_");
    my $self = shift;

    my %opts = validate(
        @_,
        {
            files    => { type => HASHREF, default => '' },
            filename => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(volume stream)],
            },
            volume => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(filename stream)]
            },
            stream => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(volume filename)]
            },
        }
    );
    my ( $filename, $volume, $stream );
    my $files = $opts{files};
    if ($files) {
        $filename = $files->{'base'}->{'data'}->{'name'};
        $volume   = $files->{'base'}->{'data'}->{'volume'};
        $stream   = $files->{'stream'}[0]->{'data'}->{'name'};
    }
    else {
        $filename = $opts{filename};
        $volume   = $opts{volume};
        $stream   = $opts{stream};
    }
    my $filepath =
      File::Spec->catfile( '/clus', $self->{vserver}, $self->get_junc_path($volume),
        $filename, 'nmatr%', $stream );
    return (
        $self->corrupt_file(
            filepath => $filepath,
            type     => 'dangling_redirector'
        )
    );
}

sub dangling_stream_data_file {
    $Log->comment("dangling_stream_data_file() called with @_");
    my $self = shift;

    my %opts = validate(
        @_,
        {
            files    => { type => HASHREF, default => '' },
            filename => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(volume stream)],
            },
            volume => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(filename stream)]
            },
            stream => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(volume filename)]
            },
        }
    );
    my ( $filename, $volume, $stream );
    my $files = $opts{files};
    if ($files) {
        $filename = $files->{'base'}->{'redirector'}->{'name'};
        $volume   = $files->{'base'}->{'redirector'}->{'volume'};
        $stream   = $files->{'stream'}[0]->{'redirector'}->{'name'};
    }
    else {
        $filename = $opts{filename};
        $volume   = $opts{volume};
        $stream   = $opts{stream};
    }
    my $filepath =
      File::Spec->catfile( '/clus', $self->{vserver}, $self->get_junc_path($volume),
        $filename, 'nmatr%', $stream );
    return (
        $self->corrupt_file(
            filepath => $filepath,
            type     => 'dangling_data_file'
        )
    );
}

sub corrupt_base_redirector {
    $Log->comment("corrupt_base_redirector() called with @_");
    my $self = shift;

    my %opts = validate(
        @_,
        {
            files => { type => HASHREF, default => '' },
            filename =>
              { type => SCALAR, optional => 1, dependencies => [qw(volume )], },
            volume => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(filename )]
            },
        }
    );
    my ( $filename, $volume );
    my $files = $opts{files};
    if ($files) {
        $filename = $files->{'base'}->{'redirector'}->{'name'};
        $volume   = $files->{'base'}->{'redirector'}->{'volume'};
    }
    else {
        $filename = $opts{filename};
        $volume   = $opts{volume};
    }
    my $filepath =
      File::Spec->catfile( '/clus', $self->{vserver}, $self->get_junc_path($volume),
        $filename );
    return (
        $self->corrupt_file(
            filepath => $filepath,
            type     => 'corrupt_redirector'
        )
    );
}

sub corrupt_base_data_file {
    $Log->comment("corrupt_base_data_file() called with @_");
    my $self = shift;

    my %opts = validate(
        @_,
        {
            files => { type => HASHREF, default => '' },
            filename =>
              { type => SCALAR, optional => 1, dependencies => [qw(volume )], },
            volume => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(filename )]
            },
        }
    );
    my ( $filename, $volume );
    my $files = $opts{files};
    if ($files) {
        $filename = $files->{'base'}->{'data'}->{'name'};
        $volume   = $files->{'base'}->{'data'}->{'volume'};
    }
    else {
        $filename = $opts{filename};
        $volume   = $opts{volume};
    }
    my $junc_path = $self->get_junc_path($volume);
    my $filepath =
      File::Spec->catfile( '/clus', $self->{vserver}, $junc_path, $filename );
    my @dirs = split( '/', $filename );
    $dirs[$#dirs] = get_timestamp() . '_corrupted';
    my $corruptfile = join( '/', @dirs );
    my $corruptpath =
      File::Spec->catfile( '/clus', $self->{vserver}, $junc_path, $corruptfile );
    return (
        $self->corrupt_file(
            filepath    => $filepath,
            type        => 'corrupt_data_file',
            corruptpath => $corruptpath
        )
    );
}

sub corrupt_stream_redirector {
    $Log->comment("corrupt_stream_redirector() called with @_");
    my $self = shift;

    my %opts = validate(
        @_,
        {
            files    => { type => HASHREF, default => '' },
            filename => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(volume stream)],
            },
            volume => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(filename stream)]
            },
            stream => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(volume filename)]
            },
        }
    );
    my ( $filename, $volume, $stream );
    my $files = $opts{files};
    if ($files) {
        $filename = $files->{'base'}->{'redirector'}->{'name'};
        $volume   = $files->{'base'}->{'redirector'}->{'volume'};
        $stream   = $files->{'stream'}[0]->{'redirector'}->{'name'};
    }
    else {
        $filename = $opts{filename};
        $volume   = $opts{volume};
        $stream   = $opts{stream};
    }
    my $filepath =
      File::Spec->catfile( '/clus', $self->{vserver}, $self->get_junc_path($volume),
        $filename, 'nmatr%', $stream );
    return (
        $self->corrupt_file(
            filepath => $filepath,
            type     => 'corrupt_redirector'
        )
    );
}

sub corrupt_stream_data_file {
    $Log->comment("corrupt_stream_data_file() called with @_");
    my $self = shift;

    my %opts = validate(
        @_,
        {
            files    => { type => HASHREF, default => '' },
            filename => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(volume stream)],
            },
            volume => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(filename stream)]
            },
            stream => {
                type         => SCALAR,
                optional     => 1,
                dependencies => [qw(volume filename)]
            },
        }
    );
    my ( $filename, $volume, $stream );
    my $files = $opts{files};
    if ($files) {
        $filename = $files->{'base'}->{'data'}->{'name'};
        $volume   = $files->{'base'}->{'data'}->{'volume'};
        $stream   = $files->{'stream'}[0]->{'data'}->{'name'};
    }
    else {
        $filename = $opts{filename};
        $volume   = $opts{volume};
        $stream   = $opts{stream};
    }
    my $junc_path = $self->get_junc_path($volume);
    my $filepath =
      File::Spec->catfile( '/clus', $self->{vserver}, $junc_path, $filename, 'nmatr%',
        $stream );
    my $corruptpath =
      File::Spec->catfile( '/clus', $self->{vserver}, $junc_path, $filename, 'nmatr%',
        get_timestamp() . '_corrupted' );
    return (
        $self->corrupt_file(
            filepath    => $filepath,
            type        => 'corrupt_data_file',
            corruptpath => $corruptpath
        )
    );
}
########## Multiple Inconsistencies ############
# Usage: multiple_inconsistencies(
#           files => $file_details,
#           types => \@corrupt_types);
#
# Types of inconsistencies :
#   corrupt base data file
#   dangling base data file
#   corrupt base redirector file
#   dangling base redirector file
#   corrupt stream redirector
#   dangling stream data file
#   corrupt stream data file
#   dangling stream redirector
#
#TODO: Enhancement to function adding one more param to get number
#	of times each type of inconistency should be created.
################################################
sub multiple_inconsistencies {
    $Log->comment("multiple_inconsistency() called with @_");
    my $self = shift;

    my %opts = validate(
        @_,
        {
            files => { type => HASHREF,  default => '' },
            types => { type => ARRAYREF, default => '' },
        }
    );

    my $files = $opts{files};
    my $types = $opts{types};

    my $count     = -1;
    my $iteration = 0;
    my $return_val;
    my $base_volume    = $files->{'base'}->{'data'}->{'volume'};
    my $base_filename  = $files->{'base'}->{'data'}->{'name'};
    my $redir_volume   = $files->{'base'}->{'redirector'}->{'volume'};
    my $redir_filename = $files->{'base'}->{'redirector'}->{'name'};

    foreach my $corrupt_type (@$types) {
        $count++;
        my $stream_redir = $files->{'stream'}[$count]->{'redirector'}->{'name'};
        my $stream_data  = $files->{'stream'}[$count]->{'data'}->{'name'};
        if ( $corrupt_type =~ /dangling\s+stream\s+redirector/i ) {
            $return_val = $self->dangling_stream_redirector(
                stream   => $stream_redir,
                volume   => $redir_volume,
                filename => $redir_filename,
            );
        }
        elsif ( $corrupt_type =~ /corrupt\s+stream\s+redirector/i ) {
            $return_val = $self->corrupt_stream_redirector(
                stream   => $stream_redir,
                volume   => $redir_volume,
                filename => $redir_filename,
            );
        }
        elsif ( $corrupt_type =~ /dangling\s+stream\s+data\s+file/i ) {
            $return_val = $self->dangling_stream_data_file(
                stream   => $stream_data,
                volume   => $base_volume,
                filename => $base_filename,
            );
        }
        elsif ( $corrupt_type =~ /corrupt\s+stream\s+data\s+file/i ) {
            $return_val = $self->corrupt_stream_data_file(
                stream   => $stream_data,
                volume   => $base_volume,
                filename => $base_filename,
            );
        }
        if ( $return_val == 0 ) {
            $Log->comment(
                "Error occured while executing $corrupt_type inconsistency");
            return 0;
        }
    }
    foreach my $type (@$types) {
        if ( $type =~ /dangling base redirector/ ) {
            $return_val = $self->dangling_base_redirector( files => $files );
        }
        elsif ( $type =~ /dangling base data file/ ) {
            $return_val = $self->dangling_base_data_file( files => $files );
        }
        elsif ( $type =~ /corrupt base redirector/ ) {
            $return_val = $self->corrupt_base_redirector( files => $files );
        }
        elsif ( $type =~ /corrupt base data file/ ) {
            $return_val = $self->corrupt_base_data_file( files => $files );
        }
        if ( $return_val == 0 ) {
            $Log->comment("Error occured while executing $type inconsistency");
            return 0;
        }
    }

    return 1;
}
####  Get Junction path  #####
# $junc-path = get_junc_path($vol_name);
#############################
sub get_junc_path {
    my ( $self, $volume_name ) = @_;
    $Log->comment("Get the Junction path for the volume $volume_name ");
    my $volume_param = NACL::C::Volume->find(
        command_interface => $self->{node},
        filter            => { volume => $volume_name }
    );
    return ( $volume_param->get_one_state_attribute('junction-path') );
}

##--------------------------------------------------------
# create_files
#
# Helper to create base and stream files over cifs client
# Inputs:
#   1. Total number of streams to be created
#   2. Mounted drive
#   3. Prefix for the file
#   4. Directory path where the file to be created
#
# Output:
#   1. Redirector and data file details of Base and Stream files
##----------------------------------------------------------
sub create_files () {
    $Log->comment("create_files() called with @_");
    my $self = shift;

    my %opts = validate(
        @_,
        {
            total_streams => { type => SCALAR, default => 0 },
            drive         => { type => SCALAR, default => $self->{drive} },
            file_prefix   => { type => SCALAR, default => 'cr_iron_' },
            dir_name      => { type => SCALAR, default => '\\' },
        }
    );
    my $total_streams = $opts{total_streams};
    my $drive         = $opts{drive};
    my $file_prefix   = $opts{file_prefix};
    my $dir_name      = $opts{dir_name};
    my $timestamp     = get_timestamp();
    my $base_file     = $file_prefix . 'base_' . $timestamp;
    $Log->comment(
            "Creating a base file by running following command : " . 'echo '
          . $base_file . ' > '
          . $drive
          . $dir_name
          . $base_file );
    my $output =
      $client->execute( command => 'echo '
          . $base_file . ' > '
          . $drive
          . $dir_name
          . $base_file );

    for ( my $i = 1 ; $i <= $total_streams ; $i++ ) {
        my $stream_file = $file_prefix . 'stream_' . $timestamp . '.' . $i;
        $Log->comment(
            "Creating a stream file by running following command : " . 'echo '
              . $stream_file . ' > '
              . $drive
              . $dir_name
              . $base_file . ':'
              . $stream_file );
        $output =
          $client->execute( command => 'echo '
              . $stream_file . ' > '
              . $drive
              . $dir_name
              . $base_file . ':'
              . $stream_file );

    }
        #sleep(5);
    my $file_details = { $self->get_data_file($base_file) };
    $Log->comment( "Base and Stream details of Redirector and data files : "
          . Dumper($file_details) );
    return $file_details;
}

###############################################################################
# get_data_file
#
# Helper function to get the Redirector and data file details of all base and
#   stream files.
#
# Input:
#   1. Redirector file path.
#
# Output:
#   1. Redirector and data file details of Base and Stream files
#
#   This will return the complete structure of the files:
#   my %file_details = ( 'base'=>
#                             {'redirector'=>{'inode'=> '' , 'volume' =>'', 'name'=>''},
#                              'data'=>{'inode'=> '' , 'volume' =>'', 'name'=>''}
#                             },
#                         'stream'=> [
#                                 {'redirector'=>{'inode'=> '' , 'volume' =>'', 'name'=>''},
#                                  'data'=>{'inode'=> '' , 'volume' =>'', 'name'=>''}
#                                 }]
#                                );
###############################################################################
sub get_data_file() {
    $Log->comment("get_data_file() called with @_");
    my ( $self, $filename ) = @_;
    my %file_details;

    $file_details{'base'}->{'redirector'}->{'volume'} = $self->{ns_vol};
    my $node_name
        = $self->get_node_name_for_aggr( $self->get_aggregate( $self->{ns_vol} ) );
    my $APISet_ns = NACL::APISet->new(
        name      => $node_name,
        category  => "Node",
        interface => "CLI",
        set       => "Nodescope",
    );

    # Get the slist of inodes and filename of base and stream redirector file
    my $dir_path = File::Spec->catfile( '/vol/' . $self->{ns_vol}, $filename );
    my @parsed_output =
      @{ $APISet_ns->slist( path => "'" . $dir_path . "'" )
          ->get_parsed_output() };

    # Adding Sleep - BURT617913
    sleep(10);

    # First element is the base redirector file
    $file_details{'base'}->{'redirector'}->{'inode'} =
      $parsed_output[0]->{inode};
    my %redirector_file = $self->get_inode_param(
        $file_details{'base'}->{'redirector'}->{'volume'},
        $file_details{'base'}->{'redirector'}->{'inode'}
    );
    $file_details{'base'}->{'redirector'}->{'name'} =
      $redirector_file{'name'}[0];

    # matching data file
    my $redir_fh = $self->get_file_handle(
        $self->{ns_msid}, '00',
        $file_details{'base'}->{'redirector'}->{'inode'},
        $redirector_file{'generation'}
    );

    # Get the Data Volume name and File Inode from the forward pointer
    (
        $file_details{'base'}->{'data'}->{'volume'},
        $file_details{'base'}->{'data'}->{'inode'}
    ) = $self->_hex_info( $self->{ns_vol}, $parsed_output[0]->{inode} );
    if (   $file_details{'base'}->{'data'}->{'volume'} !~ /unknown/
        && $file_details{'base'}->{'data'}->{'inode'} != 0 )
    {

        # Get the Data File path
        my @streams;

        my %data_file = $self->get_inode_param(
            $file_details{'base'}->{'data'}->{'volume'},
            $file_details{'base'}->{'data'}->{'inode'}
        );

        $dir_path = File::Spec->catfile(
            '/vol/' . $file_details{'base'}->{'data'}->{'volume'},
            $data_file{'name'}[0] );
        $file_details{'base'}->{'data'}->{'name'} = $data_file{'name'}[0];
        my $obj_file_name =
          $self->_construct_data_file_name( "01", "00", $redir_fh );
        if (
            index( $file_details{'base'}->{'data'}->{'name'}, $obj_file_name ) <
            0 )
        {
            $Log->warn(
                "Constructed data file name does not match the actual file name"
            );
        }

        # Get the slist of inodes and filename of base and stream data file
        $node_name = $self->get_node_name_for_aggr(
            $self->get_aggregate(
                $file_details{'base'}->{'data'}->{'volume'}
            )
        );

        # Initialize the APISet object to Node containing data file
        my $APISet_data = NACL::APISet->new(
            name      => $node_name,
            category  => "Node",
            interface => "CLI",
            set       => "Nodescope",
        );
        my @slist_stream =
          @{ $APISet_data->slist( path => "'" . $dir_path . "'" )
              ->get_parsed_output() };

        # Adding Sleep - BURT617913
        sleep(10);

        # Cleanup the APISet object
        $APISet_data->cleanup_APISet();

        foreach my $file ( @{ $parsed_output[0]->{'extended-inode-info'} } ) {
            my %stream_file;
            $stream_file{'redirector'}->{'inode'}  = $file->{inode};
            $stream_file{'redirector'}->{'name'}   = $file->{name};
            $stream_file{'redirector'}->{'volume'} = $self->{ns_vol};
            (
                $stream_file{'data'}->{'volume'},
                $stream_file{'data'}->{'inode'}
            ) = $self->_hex_info( $self->{ns_vol}, $file->{inode} );
            my %redirector_stream = $self->get_inode_param(
                $stream_file{'redirector'}->{'volume'},
                $stream_file{'redirector'}->{'inode'}
            );
            my $redir_fh = $self->get_file_handle(
                $self->{ns_msid}, '00',
                $stream_file{'redirector'}->{'inode'},
                $redirector_stream{'generation'}
            );
            foreach
              my $stream ( @{ $slist_stream[0]->{'extended-inode-info'} } )
            {

                if ( $stream->{inode} == $stream_file{'data'}->{'inode'} ) {
                    $stream_file{'data'}->{'name'} = $stream->{name};
                    my $obj_file_name =
                      $self->_construct_data_file_name( "01", "00", $redir_fh );
                    if (
                        index( $stream_file{'data'}->{'name'}, $obj_file_name )
                        < 0 )
                    {
                        $Log->warn(
"Constructed data file name does not match the actual file name"
                        );
                    }
                }
            }
            push( @streams, {%stream_file} );
        }
        if ( $#{ $parsed_output[0]->{'extended-inode-info'} } !=
            $#{ $slist_stream[0]->{'extended-inode-info'} } )
        {
            $Log->comment(
"Number of streams in redirector and data files are not matching"
            );
            foreach
              my $stream ( @{ $slist_stream[0]->{'extended-inode-info'} } )
            {
                my %stream_file;
                if ( $#streams >= 0 ) {
                    foreach my $file (@streams) {
                        if ( $stream->{inode} != $file->{'data'}->{'inode'} ) {
                            $stream_file{'data'}->{'name'}  = $stream->{name};
                            $stream_file{'data'}->{'inode'} = $stream->{inode};
                        }
                    }
                }
                else {
                    $stream_file{'data'}->{'name'}  = $stream->{name};
                    $stream_file{'data'}->{'inode'} = $stream->{inode};
                }
                push( @streams, {%stream_file} );
            }
        }
        $file_details{'stream'} = [@streams];
    }
    return %file_details;
}

##--------------------------------------------------------
# _hex_info
#
# Helper to get the data file's volume and inode number
# Inputs:
#   1. Namespace volume name
#   2. Redirector file inode
#
# Output:
#   1. Data file's volume name
#   2. Data file's inode number
##----------------------------------------------------------
sub _hex_info {
    $Log->comment("hex_info() called with @_");
    my ( $self, $volume, $inode ) = @_;
    my ( $data_vol, $data_inode );
    my $cmd    = 'wex raw ' . $volume . '.' . $inode;
    my $output = $self->{node_shell}->execute_raw_command( command => $cmd );
    my @temp   = $output =~ /\s+\w+\s+([\w+\s]{23})/g;
    if ( $temp[12] =~ /[\w+\s]{12}(\w+)\s+(\w+)\s+(\w+)\s+(\w+)/i ) {
        try {
            my $data_vol_id = hex( $4 . $3 . $2 . $1 );
            $Log->comment("Data volume msid value is $data_vol_id");
            my @data_volume = @{
                $self->{node_cli}->volume_show(
                    vserver => $self->{vserver},
                    'msid'  => $data_vol_id
                  )->get_parsed_output()
              };
            $data_vol = $data_volume[0]->{volume};
            $Log->comment("Data volume msid value is $data_vol");
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $exception_object = shift;
            my $resp             = $exception_object->get_response_object();
            $Log->comment( "Caught a " . ref($exception_object) . "!" );
            $Log->comment( "error is: " . $exception_object->text() );
            if (
                index( $exception_object->text(),
                    'There are no entries matching your query' ) < 0
              )
            {
                $Log->comment("Different error received");
            }
            $data_vol = 'unknown';
            $Log->comment("Data volume msid value is $data_vol");
        };
    }
    if ( $temp[14] =~ /(\w+)\s+(\w+)\s+(\w+)/i ) {
        $data_inode = hex( $3 . $2 . $1 );
        $Log->comment("Data volume inode value is $data_inode");
    }

    return ( $data_vol, $data_inode );
}

##--------------------------------------------------------
# get_inode_param
#
# Helper to get the file's details like inode,nlink...
# Inputs:
#   1. Volume name of the file
#   2. Path of the file
#
# Output:
#   1. Details in hash format
##----------------------------------------------------------
sub get_inode_param {
    $Log->comment("get_inode_param() called with @_");
    my ( $self, $vol_name, $file_path ) = @_;
    my ( @name, %inode_param );
    my $cmd = 'wex inode ' . $vol_name . '.' . $file_path . ' -dump all';
    my $output = $self->{node_shell}->execute_raw_command( command => $cmd );

    %inode_param = $output =~ /(.*)=(.*)\n/g;
    @name        = $output =~ /name=(.*)\n/g;
    $inode_param{'name'} = \@name;

    return %inode_param;
}

##--------------------------------------------------------
# get_file_handle
#
# Helper to get the file handle
# Inputs (in decimals):
#   1. MSID value of the file
#   2. Flags value of the file
#   3. FileID value of the file
#   4. Generation value of the file
# Output:
#   Filehandle of the file
##----------------------------------------------------------
sub get_file_handle {
    $Log->comment("get_file_handle called with @_");

    my ( $self, $msid, $flags, $fileid, $gen ) = @_;
    my $fh = undef;

    my $i_msid = pack( "Q", $msid );
    my $str_msid = unpack( 'H*', $i_msid );

    my $flags_ino_igen = pack( "III", $flags, $fileid, $gen );
    my $str_flags_ino_igen = unpack( 'H*', $flags_ino_igen );

    $fh = $str_msid . $str_flags_ino_igen;
    $Log->comment("Returning fh: $fh");

    return $fh;
}

#---------------------------------------------------------------------
# _construct_data_file_name
#
# Helper to construct the data file name.
# Inputs:
#   1. Hex string representing 1 byte version.
#   2. Hex string representing 1 byte backpointer flags.
#   3. Hex string representing the redir_fh.
# Output:
#   base64 encoded backpointer name.
#---------------------------------------------------------------------
sub _construct_data_file_name {
    $Log->comment("_construct_data_file_name called with @_");
    my ( $self, $redir_version, $bp_flags, $redir_fh ) = @_;

    croak 'Invalid redir_version length: ' . $redir_version
      if ( length($redir_version) != 2 );
    croak 'Invalid bp_flags length: ' . $bp_flags if ( length($bp_flags) != 2 );
    croak 'Invalid redir_fh length: ' . $redir_fh
      if ( length($redir_fh) != 40 );

    my $bin_buf_chksum = pack "H2H2H*", $redir_version, $bp_flags, $redir_fh;
    my $hex_buf_chksum = unpack "H*", $bin_buf_chksum;
    my $sum = $self->_compute_hbuf_checksum($hex_buf_chksum);
    my $bin_buf_sum_little_endian = pack "v",
      $sum;    # v: An unsigned short (16-bit) in "VAX" (little-endian) order.
    my $hex_buf_sum = unpack "H*", $bin_buf_sum_little_endian;
    my $bin_object_file_name = pack "H4H2H2H*", $hex_buf_sum, $redir_version,
      $bp_flags, $redir_fh;
    my $obj_file_name =
      encode_base64( $bin_object_file_name, "" );    # Don't put '\n' at the end
    $obj_file_name =~ tr/\//_/;    # Translate '/' characters to '_'.
    my $len = length($obj_file_name);
    $Log->comment("constructed object file name ($len bytes): $obj_file_name");
    return $obj_file_name;
}

#---------------------------------------------------------------------
# _compute_hbuf_checksum
#
# Helper to compute checksum on a hex string.
# Input:
#   Hex string that looks like:
#   "f5200000c72ed12e010000000604008000000000000000006000000076e39419"
#   The length of this string must be divisible evenly by 4.
#   This string contains 16-bit unsigned short's in little endian format.
# Output:
#   Returns 16-bit checksum.
#---------------------------------------------------------------------
sub _compute_hbuf_checksum {
    $Log->comment("_compute_hbuf_checksum called with @_");
    my ( $self, $hbuf ) = @_;

    croak 'Invalid hex buffer length for checksum computation: ' . $hbuf
      if ( ( length($hbuf) % 2 ) != 0 );

    my $bin_H = pack "H*", $hbuf;
    my @bin_C = unpack "C*", $bin_H;
    my $sum1  = 0xff;
    my $sum2  = 0xff;
    foreach my $val (@bin_C) {
        $sum1 += $val;
        $sum2 += $sum1;
    }
    $sum1 = $self->_reduce8($sum1);
    $sum2 = $self->_reduce8($sum2);
    $sum2 <<= 8;
    $sum1 += $sum2;

    my $ret = ( ~$sum1 ) & 0xffff;             # Invert the bits
    my $sum_str = sprintf( "0x%04x", $ret );
    $Log->comment("returning checksum: $sum_str for $hbuf");
    return $ret;
}

#---------------------------------------------------------------------
# _reduce8
#
# Given a 64-bit value,
#
#   1. Add the two 32-bit words, add back the carry if any. This results
#      in a single 32-bit value.
#   2. Add the two 16-bit words, add back the carry if any. This results
#      in a single 16-bit value.
#   3. Add the two bytes, add back the carry if any. This results
#      in a single byte value.
#
# This logic assumes little-endian platform.
#---------------------------------------------------------------------
sub _reduce8 {
    my ( $self, $val ) = @_;

    my $sum32 = $val;
    my $max32 = 0xffffffff;
    my $max16 = 0xffff;
    my $max8  = 0xff;

    if ( $Config{longsize} == 8 ) {
        my $v32_0 = $val & $max32;
        my $v32_1 = ( $val >> 32 ) & $max32;
        $sum32 = $v32_0 + $v32_1;
        if ( $sum32 > $max32 ) {
            $sum32 -= $max32;
        }
    }

    my $v16_0 = $sum32 & $max16;
    my $v16_1 = ( $sum32 >> 16 ) & $max16;
    my $sum16 = $v16_0 + $v16_1;
    if ( $sum16 > $max16 ) {
        $sum16 -= $max16;
    }

    my $v8_0 = $sum16 & $max8;
    my $v8_1 = ( $sum16 >> 8 ) & $max8;
    my $sum8 = $v8_0 + $v8_1;
    if ( $sum8 > $max8 ) {
        $sum8 -= $max8;
    }

    return $sum8;
}

### Get the aggr of a volume in a vserver ####
sub get_aggregate {
    $Log->comment("get_aggregate() called with @_");
    my ( $self, $vol_name ) = @_;
    my @volume = NACL::C::Volume->find(
        command_interface => $self->{node},
        filter            => { volume => $vol_name }
    );
    my $aggr_name = $volume[0]->state()->aggregate();
    return $aggr_name;
}

sub get_timestamp {
    my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst ) =
      localtime(time);
    if ( $mon < 10 )  { $mon  = "0$mon"; }
    if ( $hour < 10 ) { $hour = "0$hour"; }
    if ( $min < 10 )  { $min  = "0$min"; }
    if ( $sec < 10 )  { $sec  = "0$sec"; }
    $year = $year + 1900;
    $mon  = $mon + 1;

    return $year . '_' . $mon . '_' . $mday . '_' . $hour . '_' . $min . '_'
      . $sec;
}
