# Copyright (c) 2011 Network Appliance, Inc.
# All rights reserved.

## @summary Cleanup utilities for CR tests
## @author jpeters@netapp.com
## @status shared
## @pod here

package CR::Cleanup;

use strict;
use warnings;

use TCD;

use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::C::DebugFaultInjection;
use NACL::C::Volume;
use NACL::C::SystemServicesNtpConfig;
use NACL::C::Snapshot;
use NACL::STask::Snapmirror;

=head1 NAME

CR::Cleanup - A class to handle cleanup for CR tests.

=head1 DESCRIPTION

This class handles onlining volumes that are offlined as part of a test, deleting snapshots
that are created as part of a test, and so on.

=head1 SYNOPSIS

=begin text

=end text

=head1 AUTHOR

Jessica Peters (jpeters@netapp.com)

=cut


sub new
{
    my $class = shift;
    my $test = shift;
    my $node = shift;
    my $vserver = shift;

    my $self = {};
    $self->{TEST} = $test;
    $self->{NODE} = $node;
    $self->{VSERVER} = $vserver;
    $self->{REENABLE_NTP} = 0;
    $self->{FAULTS_TO_DISABLE} = {};
    $self->{VOLUMES_TO_MOUNT} = {};
    $self->{VOLUMES_TO_ONLINE} = {};

    bless $self, $class;
    return $self;
}

sub add_snapshot_to_delete
{
    my ($self, $name) = @_;

    push(@{$self->{SNAPS_TO_DELETE}}, $name);
}

sub reenable_ntp
{
    my ($self) = @_;
    $self->{REENABLE_NTP} = 1;
}

sub enable_fault_injection
{
    my $self = shift;
    my %params = @_;
    my @nodes = NACL::C::Node->find();

    $self->{FAULTS_TO_DISABLE}->{$params{'fault-id'}} = 1;

    foreach my $node (@nodes) {
        if(!exists($params{'fault-mode'})) {
            $params{'fault-mode'} = 'GX_FAULT_MODE_ALWAYS';
        }

        $params{'command_interface'} = $self->{NODE};
        $params{'node'} = $node->node;

        NACL::C::DebugFaultInjection->set_fault_mode(%params);
        NACL::C::DebugFaultInjection->enable_fault(
            command_interface => $self->{NODE},
            'fault-id' => $params{'fault-id'},
            'node' => $params{'node'}
        );
    }
}

sub disable_fault_injection
{
    my $self = shift;
    my %params = @_;
    my @nodes = NACL::C::Node->find();

    foreach my $node (@nodes) {
        NACL::C::DebugFaultInjection->disable_fault(
            command_interface => $self->{NODE},
            'fault-id' => $params{'fault-id'},
            'node' => $node->node
        );
    }

    delete($self->{FAULTS_TO_DISABLE}->{$params{'fault-id'}});
}

sub unmount_and_offline_volume
{
    my $self = shift;
    my $volume = shift;

    # Find the volume we want to turn offline 
    my $vol = NACL::C::Volume->find(
        command_interface => $self->{NODE},
        filter => {
            vserver           => $self->{VSERVER},
            volume            => $volume
        },
    );

    # Unmount the volume if it's not unmounted
    if ($vol->get_one_state_attribute('junction-active') eq 'true') {
        my $jp = $vol->get_one_state_attribute('junction-path');
        $self->{VOLUMES_TO_MOUNT}->{$volume} = $jp;
        $vol->unmount();
    }

    $self->{VOLUMES_TO_ONLINE}->{$volume} = 1;
    try {
        $vol->offline();
    } catch NACL::APISet::Exceptions::CommandFailedException with {
        # burt 592431 work around after encounters volume busy error
        my $msg = shift;
        if ( $msg->text() =~ /volume busy/i ) {
            $self->{TEST}->nlog("WARN: volume offline command failed, retry " .
                $msg->text());
            sleep 20;
            $vol->offline();
        } else {
            NACL::APISet::Exceptions::CommandFailedException->throw(
                'text' => $msg->text() );
        }
    };

    return $vol;
}

sub online_volume
{
    my $self = shift;
    my $volume = shift;

    NACL::C::Volume->online(
        command_interface => $self->{NODE},
        vserver           => $self->{VSERVER},
        volume            => $volume
    );

    delete($self->{VOLUMES_TO_ONLINE}->{$volume});
}

sub cleanup
{
    my $self = shift;
    my $success = 1;

    # clean up fault injection
    $success = 0 if (!$self->cleanup_fault_injection());

    # clean up offline volumes
    $success = 0 if (!$self->cleanup_offline_volumes());

    # clean up snapshots
    $success = 0 if (!$self->cleanup_snapshots());

    # reenable NTP
    if ($self->{REENABLE_NTP}) {
        try {
            NACL::C::SystemServicesNtpConfig->modify(
                command_interface => $self->{NODE},
                enabled => 'true');
        } catch NATE::BaseException with {
            my $msg = shift;
            $self->{TEST}->nlog("WARN: NTP re-enable failed with: " . 
                $msg->text());
            $success = 0;
        };
    }

    return $success;
}

sub cleanup_snapshots {
    my $self    = shift;
    my $success = 1;

    # clean up snapshots
    for my $snapshot ( @{ $self->{SNAPS_TO_DELETE} } ) {
        try {
            my @ns_mirrors = NACL::C::Volume->find(
                command_interface => $self->{NODE},
                filter            => {
                    'vserver'          => $self->{VSERVER},
                    'is-constituent'   => 'true',
                    'constituent-role' => 'ns_mirror'
                },
                allow_empty => 1,
            );
            my $volume_filter = '';
            if (@ns_mirrors) {
                foreach my $ns_mirror (@ns_mirrors) {
                    $ns_mirror = '!' . $ns_mirror->volume;
                }
                $volume_filter = join ',', @ns_mirrors;
            }
            else {
                $volume_filter = '*';
            }

            # add retry in case ns_mirror update is running
            my $retry_ss_delete = 0;
            try {
                NACL::C::Snapshot->delete(
                    command_interface => $self->{NODE},
                    snapshot          => $snapshot . '*',
                    vserver           => $self->{VSERVER},
                    volume            => $volume_filter,
                    force             => 'true',
                    'ignore-owners'   => 'true',
                    foreground        => 'true'
                );
            } catch NATE::BaseException with {
                my $delete_msg = shift;
                if ($delete_msg->text() =~ /Snapshot is in use/i ) {
                    $self->{TEST}->nlog("retry snapshot delete");
                    $retry_ss_delete = 1;
                } else {
                    NATE::BaseException->throw( $delete_msg->text() );
                }
            };

            if ($retry_ss_delete) {
                my $n_poll=6;
                while ($retry_ss_delete && $n_poll >0) {
                    my $snapmirror = NACL::STask::Snapmirror->find(
                                         command_interface => $self->{NODE},
                                         'filter'          => {
                                             'vserver' => $self->{VSERVER},
                                             'status' => '!Idle',
                                         },
                                         'allow_empty' => 'true',
                                     );
                    last if (!$snapmirror);
                    $n_poll--;
                    sleep 10;
                }
                NACL::C::Snapshot->delete(
                    command_interface => $self->{NODE},
                    snapshot          => $snapshot . '*',
                    vserver           => $self->{VSERVER},
                    volume            => $volume_filter,
                    force             => 'true',
                    'ignore-owners'   => 'true',
                    foreground        => 'true'
                )
            }
        }
        catch NATE::BaseException with {
        my $msg = shift;
        if ( $msg->text() !~ /no entries matching your query/ ) {
            $self->{TEST}->nlog( "WARN: snapshot cleanup "
                  . "failed with: "
                  . $msg->text() );
            $success = 0;
        }
    };
}
return $success;
}

sub cleanup_offline_volumes
{
my $self = shift;
my $success = 1;

    for my $vol (keys(%{$self->{VOLUMES_TO_ONLINE}})) {
        try {
            NACL::C::Volume->online(
                command_interface => $self->{NODE},
                vserver           => $self->{VSERVER},
                volume            => $vol,
            );
        } catch NATE::BaseException with {
            $success = 0;
        };
    }

    for my $vol (keys(%{$self->{VOLUMES_TO_MOUNT}})) {
        try {
            NACL::C::Volume->mount(
                command_interface => $self->{NODE},
                vserver           => $self->{VSERVER},
                volume            => $vol,
                'junction-path'   => $self->{VOLUMES_TO_MOUNT}->{$vol}
            );
        } catch NATE::BaseException with {
            $success = 0;
        };
    }

    return $success;
}

sub cleanup_fault_injection
{
    my $self = shift;
    my $success = 1;

    my @nodes = NACL::C::Node->find();

    for my $fault (keys(%{$self->{FAULTS_TO_DISABLE}})) {
        foreach my $node (@nodes) {
            try {
                NACL::C::DebugFaultInjection->disable_fault(
                    command_interface => $self->{NODE},
                    'fault-id' => $fault,
                    'node' => $node->node
                );
            } catch NATE::BaseException with {
                $success = 0;
            };
        }
    }

    return $success;
}

1;
