#===============================================================================
#     REVISION: $Id$
#         FILE: CR::ClientUtils
#
#  DESCRIPTION: This Library contains the helper functions for all common tasks
#               specific to Client operations.
#
#         BUGS: ---
#        NOTES: Not yet
#       AUTHOR: dl-svl-nosauto
#      COMPANY: NetApp, Inc
#      VERSION: 1.0
#      CREATED: Thursday 18,October,2012 09:16:39  EDT
# LAST CHANGED: $DateTime$
# CHANGED BY  : $Author$
#===============================================================================

package CR::ClientUtils;

use strict;
use warnings;
use NACL::C::Client::Remote::File;
use NACL::STask::VserverCifsShare;
use NACL::C::Client;
use NACL::STask::Client::Mount;
use NATE::PerlRPC::Worker::Process;
use Tharn qw ($Log);

use Exporter 'import';
our @EXPORT_OK = qw(
    create_cifs_files
    get_windows_mount_obj
    get_windows_os_type
);

use Params::Validate qw(SCALAR OBJECT ARRAYREF HASHREF validate);

#===  FUNCTION  ================================================================
#         NAME: create_cifs_files
#      PURPOSE: To create files through cifs
#   PARAMETERS: win_command_interface (Required):
#                   Object obtained through command_interface method of Windows
#                   mount object.
#               file_prefix (Optional):
#                   Prefix to be used in name of the created files.
#                   Default prefix: file
#               count (Optional):
#                   Number of files to be created. Default file count is 1
#               arguments (Required)
#                   Arguments required by NACL::C::Client::File like:
#                   access_mode, open_flags, permissions, share_mode etc.
#      RETURNS: 1
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#        USAGE:
#          my $txt_file_objs = create_cifs_files(
#                  win_command_interface => $Win_Mount_Obj->command_interface(),
#                  file_prefix => $Win_Mount_Obj->mount_point() . q{/} . '.txt',
#                  count => $cifs_count,
#                  arguments => {
#                      access_mode => 'write',
#                      open_flags => 'create-new',
#                      share_mode => 'none',
#                  }
#          );
#
#===============================================================================
sub create_cifs_files{

    my %opts = validate(@_,
        { win_command_interface => {type => OBJECT, optional => 0},
          file_prefix => {type => SCALAR, optional => 1,
              default => 'file' },
          count => {type => SCALAR, optional => 1,
              default => 1, regex => qr/^\d+$/ },
          arguments => {type => HASHREF, optional => 0,},
          close_files => {type => SCALAR, optional => 1,
              default => 1, },
        }
    );

    $Log->comment('Creating NATE RPC worker object');
    my $nate_rpc_worker_obj = NATE::PerlRPC::Worker::Process->new(
        hostrec => $opts{win_command_interface}->hostrec(),
    );

    my @created_file_objs;

    for my $file_count ( 1 .. $opts{count} ) {

        my $file_path = $opts{file_prefix} . ".$file_count";

        $opts{arguments}->{file_name} = $file_path;

        $Log->comment("Creating file $file_path");
        my $fh_obj = $nate_rpc_worker_obj->call(
             package => 'NACL::C::Client::File',
             method => 'open',
             args => [%{$opts{arguments}}],
             wrap_result => 1,
        );
        push @created_file_objs, $fh_obj;
        $fh_obj->close() if ( $opts{close_files} );
    }

    return \@created_file_objs;
}

#===  FUNCTION  ================================================================
#         NAME: get_windows_mount_obj
#      PURPOSE: To mount and get a windows mount object
#   PARAMETERS: node_obj - Command interface of any one of the node in cluster.
#               win_client_obj - Windows client object created/found through
#                   NACL::C::Client
#               vserver_name - Name of the vserver for which cifs share needs
#                   to be mapped (mounted).
#               juntion_path - Junction path of the cifs share
#
#      RETURNS: Windows mount object
#  DESCRIPTION: It first finds the CIFS share meeting the criteria, then
#               maps it and finally returns the mapped object.
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub get_windows_mount_obj{

    my %opts = validate(
        @_,
        {   node_obj       => { type => OBJECT, optional => 0 },
            win_client_obj => { type => OBJECT, optional => 0 },
            vserver_name   => { type => SCALAR, optional => 0, },
            junction_path  => { type => SCALAR, optional => 0, },
            data_ip        => { type => SCALAR, optional => 0, },
            user           => {
                type     => SCALAR,
                optional => 1,
            },
            password => {
                type     => SCALAR,
                optional => 1,
            },
        }
    );

    # If user did not provide user name and password for cifs
    if (!$opts{user} or !$opts{password}){
        # and param CIFS_DOMAIN CIFS_USER and CIFS_PASSWD is set
        if ($ENV{CIFS_DOMAIN} and $ENV{CIFS_USER} and $ENV{CIFS_PASSWD}){
            # We set user name and password using the environment parameters
            $opts{user} = "$ENV{CIFS_DOMAIN}\\$ENV{CIFS_USER}";
            $opts{password} = "$ENV{CIFS_PASSWD}";
            $Log->warn("default CIFS username to $opts{user} and password to $opts{password}");
        }
    }

    my ( $node_obj, $win_client_obj, $vserver_name, $junction_path, $data_ip )
        = delete(
        @opts{qw(node_obj win_client_obj vserver_name junction_path data_ip)}
        );

    # If user name or password not provided, remove the parameters
    unless (exists($opts{user}) && exists($opts{password})){
        delete(@opts{qw(user password)});
    }

    my $cifs_share_obj = NACL::STask::VserverCifsShare->find(
        command_interface => $node_obj,
        filter            => {
            vserver => $vserver_name,
            path    => $junction_path
        },
    );

    my $cifs_share_name
        = $cifs_share_obj->get_one_state_attribute('share-name');

    # Only if data ip is specified
    if ($data_ip) {
        $opts{server} = $data_ip;
    }

    my $win_mount_obj = NACL::STask::Client::Mount->create(
        command_interface        => $win_client_obj,
        share                    => $cifs_share_name,
        server_command_interface => $node_obj,
        vserver                  => $vserver_name,
        %opts,
    );

    return $win_mount_obj;

}

#===  FUNCTION  ================================================================
#         NAME: get_windows_os_type
#      PURPOSE: To get windows OS name
#   PARAMETERS: client_obj  (Required)
#      RETURNS: A scalar denotes windows OS type, e.g Windows XP, Windows 7,
#               Windows 8, Server 2003, Server 2008
#       THROWS: none
#     COMMENTS: none
#     SEE ALSO: n/a
#        USAGE: get_windows_os_type( client_obj => $win_client_obj );
#===============================================================================
sub get_windows_os_type {
    my %opts = validate(@_,{client_obj => {type => OBJECT, optional => 0},});

    # Validate a windows client is provided
    my $win_client_obj = $opts{client_obj};
    if (!$win_client_obj->is_windows()) {
        $Log->comment('Provided client is not windows client!');
        return;
    }

    # Available OS types information
    my %os_types = (
        win7   => 'Windows 7',
        win8   => 'Windows 8',
        winxp  => 'Windows XP',
        win2k3 => 'Server 2003',
        win2k8 => 'Server 2008',
    );

    # Get Windows OS type
    my $os_type;
    my $output = $win_client_obj->execute(
        command => 'systeminfo | findstr /B /C:' . q{"} . 'OS Name' . q{"}
    );

    if ($output =~ /$os_types{win7}/i) {
        $os_type = $os_types{win7};
    } elsif ($output =~ /$os_types{win8}/i) {
        $os_type = $os_types{win8};
    } elsif ($output =~ /$os_types{winxp}/i) {
        $os_type = $os_types{winxp};
    } elsif ($output =~ /$os_types{win2k3}/i) {
        $os_type = $os_types{win2k3};
    } elsif ($output =~ /$os_types{win2k8}/i) {
        $os_type = $os_types{win2k8};
    }
    return $os_type;
}
1;
