# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary
##  Common CR setup functions call to be share by beacon and searchlight
##  setup scripts
##
## @see-also CR/beacon_setup.thpl CR/searchlight_setup.thpl
##

package CR::CommonSetup;

package main;
use CR::CommonParams;
use strict;
use warnings;
no warnings 'redefine';

# Module import.
use CR::Setup::Constants;
use TCD;
use Params::Validate qw(validate_with UNDEF HASHREF SCALAR HASHREF);
use CR::::C::DebugSmdbTableReposDcMinSize;
use NACL::STask::SystemLicense;
use NACL::STask::Snapmirror;
use NACL::C::Node;
use NACL::C::Vserver;
use NACL::C::VserverServicesUnixUser;
use NACL::C::VserverServicesUnixGroup;
use NACL::C::VserverServicesNisDomain;
use NACL::C::VserverNameMapping;
use NACL::C::VserverCifsOptions;
use NACL::STask::Volume;
use NACL::C::VolumeEfficiency;
use NACL::C::Aggregate;
use NACL::C::Interface;
use NACL::C::ExportPolicyRule;
use NACL::C::VserverCdmi;
use NACL::C::VserverCdmiDomain;
use NACL::C::VserverNfs;
use NACL::C::VserverCifs;
use NACL::C::VserverCifsShare;
use NACL::C::VserverCifsSuperuser;
use NACL::C::VserverCifsDomainPreferredDc;
use NACL::C::VserverServicesDns;
use NACL::C::Disk;
use NACL::C::JobLog;
use NACL::C::Job;
use NACL::C::SecurityLogin;
use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::TimeoutException;
use NACL::APISet::Exceptions::ResponseException;
use NACL::Exceptions::HostUnreachable;
use NACL::Exceptions::NoElementsFound;
use NACL::C::Snapmirror;
use NACL::C::Network;
use NACL::STask::NetworkRoutingGroupsRoute;
use NACL::C::SystemServicesNtpServer;
use Digest::SHA1 qw(sha1_hex);
use NATE::Error;

use TestUtils;

# Global Variables.
use vars qw($Options $Node $Apiset %SiteData $Site $NA $Vsadmim_name
    @Vserver_Name_List);

sub UseCommonSetup {
    $Options = new Options;

    $Options->parameters(
        definitions => [
            {   name => "ALLOW_AGGR_REUSE",
                description =>
                    "Do not delete aggregates if they are listed in AGGR_NAME_LIST or OLS_AGGR_NAME",
                spec    => "boolean",
                default => 1,
            },
            {   name        => "PREFERRED_DISK_TYPE",
                description => "List of disk type prefer by aggregate create",
                spec        => "s@",
                default     => [ 'ATA', 'SATA' ],
            },
            {   name        => "AGGR_RAID_TYPE",
                description => "Aggregate raid type, can be raid4 or raid_dp",
                spec        => "s",
                default     => "raid4",
                valid_values => [ "raid_dp", "raid4" ],
            },
            {   name        => "REPLACE_NA_WITH_PASS",
                description => "Force NA test to be PASS test",
                spec        => "boolean",
                default     => 0,
            },
            {   name => "DIRECTORY_BITWIDTH",
                description =>
                    "value for -directory-bitwidth when create InfiniteVol, default value is 13",
                spec          => "s",
                less          => 21,
                greater_equal => 0,
            },
            {   name => "SPACE_GUARANTEE",
                description =>
                    "Set the -space-guarantee flag when create the InfiniteVol",
                spec         => "s",
                valid_values => [ "none", "volume" ],
            },
            {   name        => "EXPORT_OBJECT_VOLUME",
                description => "Set the object constituents",
                spec        => "boolean",
                default     => 0,
            },
            {   name        => "DEBUG_LOGGING",
                description => "Enable changes to the logging level",
                spec        => "boolean",
                default     => 1,
            },
            {   name         => "LOGGING_LEVEL",
                description  => "The loging level for mgwd log and job log",
                spec         => "s",
                default      => "debug",
                valid_values => [
                    "emerg",   "alert",  "crit", "err",
                    "warning", "notice", "info", "debug"
                ],
            },
            {   name        => "VSADMIN_ON_DATA_LIF",
                description => "Enable Vserver Admin on data lif",
                spec        => "boolean",
                default     => 0,
            },
            {   name        => "VSADMIN_PASSWORD",
                description => "vsadmin password, default is netapp1!",
                spec        => "s",
                default     => "netapp1!",
            },
            {   name        => 'SRC_NODE',
                description => 'Node name of a source cluster',
            },
            {   name        => 'DST_NODE',
                description => 'Node name of a destination cluster',
            },
            {   name        => 'SRC_VSERVER',
                description => 'Source vserver name',
                default     => 'vs0',
            },
            {   name => 'NIS_DOMAIN',
                description =>
                    'NIS Domain name, if not provide, will try to get it from site info files',
                spec    => 's',
                default => 'netapp.com',
            },
            {   name => 'NIS_SERVERS',
                description =>
                    'NIS Domain Servers IP, if not provide, will try to get it from site info files',
                spec    => 's@',
                default => ['10.60.132.113'],
            },
            {   name        => "NTP_SERVER_IP",
                description => "NTP Server IP address",
                spec        => "s@",
                default     => [qw(10.98.203.74 10.98.18.43 10.98.13.43)],
            },
            {   name        => 'PURGE_INTERCLUSTER',
                description => 'Purges existing intercluster interfaces (Default: 0)',
                spec        => "boolean",
                default     => 0,
            },
            {   name        => 'MAKE_INTERCLUSTER',
                description => 'Creates intercluster interfaces if they do not already exist (Default: 1)',
                spec        => "boolean",
                default     => 1,
            },
        ]
    );
    $Options->process();
    $Node   = NACL::C::Node->find();
    $Apiset = $Node->apiset(
        category  => "Node",
        interface => "CLI",
        set       => "CMode",
    );
    $NA
        = $Options->get( option => 'REPLACE_NA_WITH_PASS' )
        ? $TCD::PASS
        : $TCD::NA;
    %SiteData = get_site_info();

# Find out the site, default it to svl if it's not define and can't be set from tharnhost
    if ( $Options->defined( option => 'SITE' ) ) {
        $Site = $Options->get( option => 'SITE' );
    }
    else {
        $Site = $Node->hostrec()->SITE() ? $Node->hostrec()->SITE() : 'svl';
    }
    $Vsadmim_name = 'vsadmin';

    $Test->step("Make sure all nodes are healthy");
    my @nodes = NACL::C::Node->find( command_interface => $Node );
    foreach my $node (@nodes) {
        my $health = $node->get_one_state_attribute('health');
        if ( $health ne 'true' ) {
            throw NATE::Error::Simple(
                'Not all the nodes are healthy, failing');
        }
    }

    if ( $Node->hostrec()->hosttype() =~ /vsim/ ) {
        $Test->step("Turning off vsim raid scrubbing");
        $Apiset->execute_raw_command(
            command => 'node run * options raid.media_scrub.enable off' );
    }

    my $num_repos_vservers = $Options->get( option => 'NUM_REPOS_VSERVER' );
    my $num_flex_vservers  = $Options->get( option => 'NUM_FLEX_VSERVER' );
    my @user_vserver_names = @{ $Options->get( option => 'VSERVER_NAMES' ) };
    my $num_total_vservers = $num_repos_vservers + $num_flex_vservers;

    @user_vserver_names = _unique_list(@user_vserver_names);

    if ( scalar(@user_vserver_names) > 0 ) {

        # use user specified list if it was provided and ensure that
        # the number of provided vserver names is equal to the amount
        # the user wants to create
        if ( $num_total_vservers != scalar(@user_vserver_names) ) {
            $Test->nlog( 'VSERVER_NAMES and the sum of NUM_REPOS_VSERVER + '
                    . 'NUM_FLEX_VSERVER were not equal. Please provide the same number'
                    . ' of vserver names' );
            return $TCD::FAIL;
        }

        @Vserver_Name_List = @user_vserver_names;
    }
    else {
        if ( $num_total_vservers == 1 ) {

            # use single user specified vserver option if only 1 vserver is
            # being created
            push( @Vserver_Name_List, $Options->get( option => 'VSERVER' ) );
        }
        else {

            # no vserver names were specified so create them
            for ( my $i = 0; $i < $num_total_vservers; $i++ ) {
                my $vs_name = 'vs' . $i;
                push( @Vserver_Name_List, $vs_name );
            }
        }
    }
}

sub debug_logging {
    $Test->description();
    my $result = $TCD::PASS;

    if ( $Options->get( option => 'DEBUG_LOGGING' ) eq 1 ) {
        my $level = $Options->get( option => 'LOGGING_LEVEL' );

        # execute the log command on each node
        my @nodes = NACL::C::Node->find( command_interface => $Node );
        my $node_cnt = scalar(@nodes);
        foreach my $node (@nodes) {
            my $node_name = $node->get_one_state_attribute('node');
            $Test->nlog("enable debug logging on node $node_name");
            $Test->step(
                "Changing the 'job log' logging level for CR* modules to $level"
            );
            my $apiset = $node->apiset(
                category  => "Node",
                interface => "CLI",
                set       => "CMode"
            );
            NACL::C::JobLog->modify(
                command_interface => $node,
                'module'          => 'CR*',
                'level'           => $level
            );

            $Test->step(
                "Changing the 'logger mgwd log' logging level for CR* modules to $level"
            );
            my $res
                = $apiset->execute_raw_command(
                command => "logger mgwd log modify -module CR* -level "
                    . $Options->get( option => 'LOGGING_LEVEL' ) );
            if ( $res =~ /Error:/ ) {
                $Log->warn(
                    "Failed to change logging level. Are you running a debug build?\n"
                        . "To surpress this warning add DEBUG_LOGGING=0 to the ntest parameters or smoke plan"
                );
            }

            $Test->step(
                "Changing the 'logger mgwd log' logging level for tables::repos* modules to $level"
            );
            $res
                = $apiset->execute_raw_command( command =>
                    "logger mgwd log modify -module tables::repos* -level "
                    . $Options->get( option => 'LOGGING_LEVEL' ) );
            if ( $res =~ /Error:/ ) {
                $Log->warn(
                    "Failed to change logging level. Are you running a debug build?\n"
                        . "To surpress this warning add DEBUG_LOGGING=0 to the ntest parameters or smoke plan"
                );
            }

            $Test->step(
                "Changing the 'logger mgwd log' logging level for repos module to $level"
            );
            $res
                = $apiset->execute_raw_command(
                command => "logger mgwd log modify -module repos -level "
                    . $Options->get( option => 'LOGGING_LEVEL' ) );
            if ( $res =~ /Error:/ ) {
                $Log->warn(
                    "Failed to change logging level. Are you running a debug build?\n"
                        . "To surpress this warning add DEBUG_LOGGING=0 to the ntest parameters or smoke plan"
                );
            }

            $Test->step(
                'Disabling DNS reverse lookup for ssh.'
            );
            $node->apiset( set => 'Systemshell' )->execute_raw_command(
                command => 'sudo sed -i .bak "s/#VerifyReverseMapping no/VerifyReverseMapping no/g" /etc/ssh/sshd_config ; sudo /etc/rc.d/sshd restart'
            );
        }
    }
    else {
        $result = $NA;
    }

    return $result;
}

# This test will try to remove everything in a cluster.
# This only trigger when RESETCLUSTER=1 is set.
sub clean_cluster {
    $Test->description(
        "Delete all existing repositories, vservers and volumes.",
        "This test is enable by setting RESETCLUSTER=1",
    );
    if ( $Options->get( option => "RESETCLUSTER" ) == 1 ) {
        $Test->step("Make sure all aggregates are online");
        my @aggrs = NACL::C::Aggregate->find( 'command_interface' => $Node, );
        foreach (@aggrs) {
            $_->online()
                if (
                $_->get_one_state_attribute('state') =~ m/offline|restrict/ );
        }

        $Test->step('Make sure all the volumes are online');
        my @volumes = NACL::STask::Volume->find( 'command_interface' => $Node, );
        foreach my $vol (@volumes) {
            $vol->online()
                if ( $vol->get_one_state_attribute('state') !~ /online/ );
        }

        my @vservers = NACL::C::Vserver->find(
            command_interface => $Node,
            filter            => { type => 'data' },
            allow_empty       => 1
        );
        foreach my $vserver (@vservers) {
            delete_vserver($vserver);
        }
        return $TCD::PASS;
    }
    return $NA;
}

sub delete_aggr {
    $Test->description(
        "Delete all aggregates",
        "This test is enable by DELETE_AGGR=1 and ALLOW_AGGR_REUSE",
        "If ALLOW_AGGR_REUSE=1, this test will keep the aggregates",
        "listed in AGGR_NAME_LIST and OLS_AGGR_LIST",
        "If ALLOW_AGGR_REUSE=0, this test will delete all aggregates.",
    );
    my $zero_spare = 0;
    my %keep_aggr;
    my @aggrList = _get_aggr_list();
    @keep_aggr{@aggrList} = (1) x @aggrList
        if ( $Options->get( option => 'ALLOW_AGGR_REUSE' ) );
    if ( $Options->get( option => 'DELETE_AGGR' ) == 1 ) {
        my @nacl_aggrs = NACL::C::Aggregate->find(
            command_interface => $Node,
            filter            => { volcount => 0 },
            allow_empty       => 1
        );
        foreach my $nacl_aggr (@nacl_aggrs) {
            my $aggr_name = $nacl_aggr->get_one_state_attribute('aggregate');
            if ( !defined $keep_aggr{$aggr_name} ) {
                $Test->step( "Deleting aggregate: " . $aggr_name );
                $nacl_aggr->delete(
                    "method-timeout" => $Options->get( option => "TIMEOUT" )
                );
                $zero_spare = 1;
            }
            else {
                $Test->step( "ALLOW_AGGR_REUSE is set and $aggr_name exist in"
                        . " AGGR_NAME_LIST or OLS_AGGR_LIST" );
            }
        }

        # We will do disk zerospares on real hardware only.
        if ( $zero_spare == 1 && $Node->hostrec()->hosttype() !~ /vsim/ ) {
            $Test->step( "Host type is "
                    . $Node->hostrec()->hosttype()
                    . ", initial disk zerospares" );
            NACL::C::Disk->zerospares( command_interface => $Node );
        }
        return $TCD::PASS;
    }
    return $NA;
}

# This add license
sub add_license {
    $Test->description("Add License to the cluster");
    if ( $Test->get_param('APISET_CDEFS_FROM_BUILD') eq '1' ) {
        $Test->nlog(
            'May fail to add license with APISET_CDEFS_FROM_BUILD=1 (burt524588)'
        );
        $Test->nlog(
            'Please use APISET_CLI_CMODE_CDEFS_FROM_BUILD=1 if add_license fail.'
        );
    }
    my @license = @{ $Options->get( option => "FEATURES" ) };
    foreach my $feature (@license) {
        my $was_added;
        NACL::STask::SystemLicense->add(
            command_interface => $Node,
            feature           => $feature,
            _was_added        => \$was_added,
        );
        if ($was_added) {
            $Test->step("Feature $feature added");
        }
        else {
            $Test->step(
                "Feature $feature already exists, don't need to do anything."
            );
        }
    }
    return $TCD::PASS;
}

# This create aggregate if they are not already created.
# If the script don't need to create aggregate, it will return NA
# aggregate are created in round robin on the cluster
sub create_aggr {
    my @nodes         = NACL::C::Node->find( command_interface => $Node );
    my $node_cnt      = scalar(@nodes);
    my $status        = $NA;
    my $cnt           = 0;
    my $aggr_disk     = $Options->get( option => 'AGGR_DISK_COUNT' );
    my $timeout_timer = $Options->get( option => 'AGGR_CREATE_WAIT' );
    my $sleep_time    = $Options->get( option => 'AGGR_CHECK_INTERVAL' );
    my $aggr_type     = $Options->get( option => 'AGGR_RAID_TYPE' );
    my $vs_disk_count = $Options->get( option => 'VSERVER_AGGR_DISK_COUNT');
    my @aggrList      = _get_aggr_list();
    $Test->description();
    if ($aggr_type =~ /raid4/){
        my $msata_aggr = NACL::C::Disk->find(
            'command_interface' => $Node,
            filter              => { 'type' => 'MSATA', },
            'allow_empty'       => 1,
        );
        if ($msata_aggr){
            $Log->warn('Cannot use raid4 with MSATA disk, falling back to raid_dp');
            $aggr_type = 'raid_dp';
            if ($aggr_disk < 4){
                $Log->warn("Aggregate disk count too low for MSATA, change form $aggr_disk to 4");
                $aggr_disk = 4;
            }
            if ($vs_disk_count < 4) {
                $Log->warn("Vserver aggregate disk count too low for MSATA, change form $vs_disk_count to 4");
                $vs_disk_count = 4;
            }
        }
    }

    # Create aggregate here
    foreach my $aggr (@aggrList) {

        # Get the node name in round robin.
        my $node
            = $nodes[ $cnt++ % $node_cnt ]->get_one_state_attribute('node');
        my $tmp = NACL::C::Aggregate->find(
            command_interface => $Node,
            filter            => { aggregate => $aggr },
            allow_empty       => 1,
        );

        if ( $tmp && $tmp->get_one_state_attribute('nodes')->[0] =~ /$node/ )
        {
            $Test->step("Aggregate $aggr already exist.");
        }
        else {
            if ($tmp) {
                if ( $tmp->get_one_state_attribute('volcount') ne '0' ) {
                    $Test->nlog(
                        "Can't delete aggregate $aggr because there are volumes in it"
                    );
                }
                else {
                    $Test->step(
                        "Found aggregate $aggr but it's not on node $node, deleting it"
                    );
                    $tmp->delete( "method-timeout" =>
                            $Options->get( option => "TIMEOUT" ) );
                }
            }
            my $disk_type = prefer_disk_type_get($node);
            $Test->step(
                "Creating aggregate " . $aggr . " on node " . $node );
            my @params = (
                aggregate => $aggr,
                diskcount =>
                    ( $aggr eq $Options->get( option => 'VSERVER_AGGR' ) )
                ? $vs_disk_count
                : $aggr_disk,
                "method-timeout" => $Options->get( option => "TIMEOUT" ),
                "nodes"          => [$node],
                "raidtype"       => $aggr_type,
            );

         # Will use the preferred disk type if we find the preferred disk type
         # in the filer.
            push( @params, 'disktype' => $disk_type ) if ($disk_type);
            NACL::C::Aggregate->create(
                command_interface => $Node,
                @params,
            );
            $status = $TCD::PASS;
        }
    }

    # Wait for aggregate to go online
    $Test->step('Waiting all required aggregates to go online');
    foreach my $aggr (@aggrList) {
        my $tmp = NACL::C::Aggregate->find(
            command_interface => $Node,
            filter            => { aggregate => $aggr },
        );
        while ($tmp->get_one_state_attribute('state') ne 'online'
            && $timeout_timer > 0 )
        {
            if ( $tmp->get_one_state_attribute('state') eq '-' ) {
                $Test->nlog(
                    'cannot get the state of the aggr, node may be down');
                return $TCD::FATAL;
            }
            sleep $sleep_time;
            $timeout_timer -= $sleep_time;
        }
        if ( $tmp->get_one_state_attribute('state') ne 'online' ) {
            $Test->step(
                'Aggregate create time out, may be change AGGR_CREATE_WAIT and try again'
            );
            NACL::APISet::Exceptions::TimeoutException->throw(
                "Aggregate create timeout");
        }
    }

# Extra check to make sure all aggregate are online, if not, print a warning message
    my @aggrs = NACL::C::Aggregate->find( command_interface => $Node );
    foreach my $aggr (@aggrs) {
        if ( $aggr->get_one_state_attribute('state') ne 'online' ) {
            $Log->warn( 'Aggregate '
                    . $aggr->get_one_state_attribute('aggregate')
                    . ' is not online.' );
        }
    }
    $Test->step("All required aggregate are online");
    return $status;
}

# Create the Vserver
sub create_vserver {
    $Test->description("Creating vserver(s)");
    my @aggrList           = _get_aggr_list();
    my $num_repos_vservers = $Options->get( option => 'NUM_REPOS_VSERVER' );
    my $num_flex_vservers  = $Options->get( option => 'NUM_FLEX_VSERVER' );

    my @existing_vservers = NACL::C::Vserver->find(
        'command_interface' => $Node,
        'allow_empty'       => 1,
        'filter'            => { type => 'data', },
    );

    if ( !$Options->get( option => 'ALLOW_MULTIPLE_VSERVER' )
        && scalar(@existing_vservers) > 1 )
    {
        $Test->nlog( 'If ALLOW_MULTIPLE_VSERVER=0 only a single vserver'
                . ' is supported. Re-run the script with RESETCLUSTER=1 to'
                . ' remove any existing vservers' );
        return $TCD::FAIL;
    }

    # remove vservers if they already exist
    foreach my $vs_name (@Vserver_Name_List) {
        my $vserver = NACL::C::Vserver->find(
            'command_interface' => $Node,
            'allow_empty'       => 1,
            'filter'            => { vserver => $vs_name, },
        );

        if ($vserver) {
            delete_vserver($vserver);
        }
    }

    my @unused_vserver_names = @Vserver_Name_List;

    for ( my $i = 0; $i < $num_repos_vservers; $i++ ) {
        my $vs_name          = shift @unused_vserver_names;
        my $vs_root_vol_name = $vs_name . '_root';

        $Test->step("Creating repository vserver $vs_name");

        my $vs = NACL::C::Vserver->create(
            'command_interface' => $Node,
            vserver             => $vs_name,
            rootvolume          => $vs_root_vol_name,
            aggregate           => $Options->get( option => 'VSERVER_AGGR' ),
            'ns-switch'         => $Options->get( option => 'NS_SWITCH' ),
            'rootvolume-security-style' =>
                $Options->get( option => 'SECURITY_STYLE' ),
            'is-repository'     => 'true',
            'method-timeout'    => $Options->get( option => 'TIMEOUT' ),
        );

        $Test->step("Modifying $vs_name aggr-list");
        $vs->modify( 'aggr-list' => \@aggrList );

        $Test->step("Modifying vserver root volume permissions");
        NACL::STask::Volume->modify(
            'command_interface' => $Node,
            'unix-permissions' =>
                $Options->get( option => 'UNIX_PERMISSIONS' ),
            'volume'  => $vs_root_vol_name,
            'vserver' => $vs_name,
        );
    }

    for ( my $i = 0; $i < $num_flex_vservers; $i++ ) {
        my $vs_name          = shift @unused_vserver_names;
        my $vs_root_vol_name = $vs_name . '_root';

        $Test->step("Creating non-repository vserver $vs_name");

        my $vs = NACL::C::Vserver->create(
            'command_interface' => $Node,
            vserver             => $vs_name,
            rootvolume          => $vs_root_vol_name,
            aggregate           => $Options->get( option => 'VSERVER_AGGR' ),
            'ns-switch'         => $Options->get( option => 'NS_SWITCH' ),
            'rootvolume-security-style' =>
                $Options->get( option => 'SECURITY_STYLE' ),
            'is-repository'     => 'false',
            'method-timeout'    => $Options->get( option => 'TIMEOUT' ),
        );

        $Test->step("Modifying $vs_name aggr-list");
        $vs->modify( 'aggr-list' => \@aggrList );

        $Test->step("Modifying vserver root volume permissions");
        NACL::STask::Volume->modify(
            'command_interface' => $Node,
            'unix-permissions' =>
                $Options->get( option => 'UNIX_PERMISSIONS' ),
            'volume'  => $vs_root_vol_name,
            'vserver' => $vs_name,
        );
    }

    return $TCD::PASS;
}

# Create the export rule
sub create_export_rule {
    $Test->description("Create default export rule");
    foreach my $vserver (@Vserver_Name_List) {
        NACL::C::ExportPolicyRule->create(
            command_interface => $Node,
            'policyname'      => 'default',
            'clientmatch'     => '0.0.0.0/0',
            'rorule'          => ['any'],
            'rwrule'          => ['any'],
            'protocol'        => ['any'],
            'vserver'         => $vserver,
            'anon'            => '0',
            'superuser'       => ['any'],
            'ruleindex'       => '1',
            'allow-suid'      => 'true',
            'allow-dev'       => 'true',
        );
    }
    return $TCD::PASS;
}

sub add_ntp_servers {
    $Test->description('Adding NTP Servers');

    my @nodes = NACL::C::Node->find( command_interface => $Node, );

    foreach my $n (@nodes) {

        # Site config doesn't have ntp servers
        foreach my $s ( @{ $Options->get( option => 'NTP_SERVER_IP' ) } ) {
            my $ntp_server = NACL::C::SystemServicesNtpServer->find(
                command_interface => $Node,
                filter            => {
                    node   => $n->node,
                    server => $s,
                },
                allow_empty => 1
            );

            if ( !$ntp_server ) {
                NACL::C::SystemServicesNtpServer->create(
                    command_interface => $Node,
                    node              => $n->node,
                    server            => $s,
                );
            }
        }
    }

    return $TCD::PASS;
}

# Join the nis domain
sub add_nis_domain {
    $Test->description("Adding NIS Domain to vserver");
    my $nis_domain;
    my @nis_servers;
    if ( $Options->defined( 'option' => 'NIS_DOMAIN' ) ) {
        $nis_domain = $Options->get( 'option' => 'NIS_DOMAIN' );
    }
    else {
        $Test->step("Getting NIS Domain from Site Data file");
        $nis_domain = $SiteData{$Site}->{nis_domain};
    }
    if ( $Options->defined( 'option' => 'NIS_SERVERS' ) ) {
        @nis_servers = @{ $Options->get( 'option' => 'NIS_SERVERS' ) };
    }
    else {
        $Test->step("Getting NIS Servers IP from Site Data file");
        @nis_servers = split( /\s+/, $SiteData{$Site}->{nis_servers} );
    }

    my @localusers
        = split( /,/, $Options->get( 'option' => 'LOCAL_UNIX_USERS' ) );

    foreach my $vserver (@Vserver_Name_List) {
        foreach my $userdata (@localusers) {
            my ( $name, $uid, $gid, $fullname ) = split( /:/, $userdata );
            $Test->step(
                "Add user: $name, UID: $uid, GID $gid, Full name: $fullname if user does not exist"
            );
            NACL::C::VserverServicesUnixUser->find_or_create(
                'command_interface'     => $Node,
                'user'                  => $name,
                'vserver'               => $vserver,
                create_specific_options => {
                    'primary-gid' => $gid,
                    'id'          => $uid,
                    'full-name'   => $fullname,
                }
            );
        }
        my @localgroups
            = split( /,/, $Options->get( 'option' => 'LOCAL_UNIX_GROUPS' ) );
        foreach my $groupdata (@localgroups) {
            my ( $name, $id ) = split( /:/, $groupdata );
            $Test->step(
                "Create group: $name, ID: $id if group does not exist");
            NACL::C::VserverServicesUnixGroup->find_or_create(
                'command_interface'     => $Node,
                'name'                  => $name,
                'vserver'               => $vserver,
                create_specific_options => { 'id' => $id, }
            );
        }
        $Test->step("Join NIS Domain: $nis_domain");
        NACL::C::VserverServicesNisDomain->create(
            command_interface => $Node,
            'domain'          => $nis_domain,
            'servers'         => \@nis_servers,
            'vserver'         => $vserver,
            'active'          => "true",
        );
    }
    my $ping_success = 0;
    foreach (@nis_servers) {
        $Test->step("Verify if $_ is online");
        try {
            NACL::C::Network->ping(
                command_interface => $Node,
                'node'            => $Node,
                'destination'     => $_,
            );
            $ping_success = 1;
        }
        catch NACL::Exceptions::HostUnreachable with {
            $Test->nlog("Unable to reach NIS Server $_");
        };
    }
    if ( 0 == $ping_success ) {
        $Test->nlog("Fail to reach any NIS Domain, failling");
        return $TCD::FAIL;
    }
    return $TCD::PASS;
}

# Join the dns domain
sub add_dns {
    $Test->description("Adding $Site DNS to vserver");
    my @dns_servers;
    my @dns_domains;
    if (   $Options->defined( 'option' => 'DNS_SERVER_IP' )
        && $Options->defined( 'option' => 'DNS_DOMAIN' ) )
    {
        $Test->nlog(
            'Overriding DNS Servers IP and DNS Domain Name provided by site data file'
        );
        @dns_servers = @{ $Options->get( 'option' => 'DNS_SERVER_IP' ) };
        @dns_domains = @{ $Options->get( 'option' => 'DNS_DOMAIN' ) };
    }
    else {
        @dns_servers = $SiteData{$Site}->{mgmt_dns1};
        if ( $SiteData{$Site}->{mgmt_dns2} ) {
            push( @dns_servers, $SiteData{$Site}->{mgmt_dns2} );
        }
        @dns_domains = split( /\s+/, $SiteData{$Site}->{dns_domains} );
    }
    foreach my $vserver (@Vserver_Name_List) {
        NACL::C::VserverServicesDns->create(
            command_interface => $Node,
            'domains'         => \@dns_domains,
            'name-servers'    => \@dns_servers,
            'vserver'         => $vserver,
        );
    }
    return $TCD::PASS;
}

# Create nfs services
sub create_nfs {
    $Test->description("Create NFS Services");
    my $default_protocols = {
        'v3'        => 'disabled',
        'v4.0'      => 'disabled',
        'v4.1'      => 'disabled',
        'v4.1-pnfs' => 'disabled',
    };

    my $default_options = ();

    my @required_protocols = @{ $Options->get( option => "NFS_PROTOCOL" ) };

    $default_options->{'v4.0-acl'} = $Options->get( option => "V4_ACL" )
        if ( $Options->get( option => "V4_ACL" ) );
    $default_options->{'v4.0-read-delegation'}
        = $Options->get( option => "V4_READ_DELEGATION" )
        if ( $Options->get( option => "V4_READ_DELEGATION" ) );
    $default_options->{'v4.0-write-delegation'}
        = $Options->get( option => "V4_WRITE_DELEGATION" )
        if ( $Options->get( option => "V4_WRITE_DELEGATION" ) );
    $default_options->{'v4-id-domain'}
        = $Options->get( option => "V4_ID_DOMAIN" )
        if ( $Options->get( option => "V4_ID_DOMAIN" ) );
    $default_options->{'v4.1-acl'} = $Options->get( option => "V4_1_ACL" )
        if ( $Options->get( option => "V4_1_ACL" ) );

    foreach my $protocol (@required_protocols) {
        $default_protocols->{$protocol} = 'enabled';
    }

    foreach my $vserver (@Vserver_Name_List) {
        my @nfs = NACL::C::VserverNfs->find(
            command_interface => $Node,
            filter            => { 'vserver' => $vserver },
            allow_empty       => 1,
        );
        if ( scalar(@nfs) > 0 ) {
            $Test->nlog("Deleting old NFS settings");
            NACL::C::VserverNfs->delete(
                command_interface => $Node,
                'vserver'         => $vserver,
            );
        }

        NACL::C::VserverNfs->create(
            command_interface => $Node,
            'access'          => 'true',
            'vserver'         => $vserver,
            %$default_protocols,
            %$default_options,
        );

        _vserver_nfs_modify(
            vserver          => $vserver,
            'mount-rootonly' => 'disabled'
        );
    }
    return $TCD::PASS;
}

# Create CIFS server

sub create_cifs {
    $Test->description("CIFS Server");

    return $NA unless ( $Options->get( option => 'CREATE_CIFS' ) == 1 );

    foreach my $vserver (@Vserver_Name_List) {
        $Test->step( 'Create CIFS server for ' . $vserver );

        NACL::C::VserverCifsDomainPreferredDc->add(
            command_interface => $Node,
            vserver           => $vserver,
            domain => $Options->get( 'option' => 'CIFS_DOMAIN_LONG' ),
            "preferred-dc" => $Options->get( 'option' => 'AD_SERVER_IP' ),
        );

        my $cifs_server_name
            = _generate_cifs_server_name( vserver_name => $vserver );

        NACL::C::VserverCifs->create(
            command_interface => $Node,
            vserver           => $vserver,
            'cifs-server'     => $cifs_server_name,
            domain => $Options->get( 'option' => 'CIFS_DOMAIN_LONG' ),
            username => $Options->get( 'option' => 'AD_SERVER_USER' ),
            password => $Options->get( 'option' => 'AD_SERVER_PASSWORD' ),
        );

        NACL::C::VserverCifsOptions->modify(
            'command_interface' => $Node,
            'default-unix-user' => 'nobody',
            'vserver'           => $vserver,
            'wins-servers' => $Options->get( 'option' => 'AD_SERVER_IP' ),
        );

        my @superusers = @{ $Options->get( 'option' => 'CIFS_SUPERUSERS' ) };
        sleep(10);    # Sleep 10 second to work around BURT602514
        foreach my $superuser (@superusers) {
            $Test->step("Adding superuser $superuser");
            NACL::C::VserverCifsSuperuser->create(
                'command_interface' => $Node,
                'vserver'           => $vserver,
                'accountname'       => $superuser,
                'domain' => $Options->get( 'option' => 'CIFS_DOMAIN' ),
            );
        }

        if ( $Options->get( option => 'SKIP_NAME_MAPPING_SETUP' ) ) {
            $Test->step( 'We have been told to not configure the name-mapping '
                         . 'rules. Skipping it' );
        } else {
            $Test->step( 'Configuring the name-mapping rules' );

            NACL::C::VserverNameMapping->create(
                'command_interface' => $Node,
                'vserver'           => $vserver,
                'position'          => 2,
                'direction'         => 'win-unix',
                'pattern'           => $Options->get( option => 'CIFS_DOMAIN' )
                    . '\\\\(.+)',
                'replacement' => '\\1',
            );
            NACL::C::VserverNameMapping->create(
                'command_interface' => $Node,
                'vserver'           => $vserver,
                'position'          => 2,
                'direction'         => 'unix-win',
                'pattern'           => '(.+)',
                'replacement'       => $Options->get( option => 'CIFS_DOMAIN' )
                    . '\\\\\\1',
            );
        }

        $Test->step('Create CIFS share on the Volume');

        my $vol = NACL::CS::Volume->fetch(
            'command_interface' => $Node,
            'filter'            => {
                'vserver'         => $vserver,
                'vsroot'          => 'false',
                'is-constituent'  => 'false',
                'junction-active' => 'true',
            },
            'allow_empty'       => 1,
            'requested_fields'  => [ 'junction-path' ],
        );
        if( $vol ) {
            NACL::C::VserverCifsShare->create(
                'command_interface' => $Node,
                'share-name'        => $Options->get( 'option' => 'CIFS_SHARE' ),
                'path'    => $vol->junction_path,
                'vserver' => $vserver,
            );
        } else {
            $Log->comment("No mounted volume for Vserver $vserver found, skipping");
        }
    }
    return $TCD::PASS;
}

# Create CDMI server

sub create_cdmi {
    $Test->description("CDMI Server");

    if ( $Options->get( option => 'CREATE_CDMI' ) != 1 ) {
        $Log->comment("Skipping CDMI configuration as CREATE_CDMI is not 1");
        return $NA;
    }

    # CDMI server cannot be configured unless CIFS server is configured
    if ( $Options->get( option => 'CREATE_CIFS' ) != 1 ) {
        $Log->comment( "Cannot configure CDMI unless CIFS is configured. "
                . "Rerun the script with CREATE_CIFS=1" );
        return $TCD::FAIL;
    }

    foreach my $vserver (@Vserver_Name_List) {
        my $vserverObj = NACL::C::Vserver->find(
            'command_interface' => $Node,
            'filter'            => { vserver => $vserver, },
        );

        if ($vserverObj->get_one_state_attribute('is-repository') =~ /false/)
        {
            $Log->comment( "Vserver "
                    . $vserver
                    . " is not a repository vserver. "
                    . "Skipping CDMI configuration for it" );
            next;
        }

        $Test->step( 'Configuring CDMI server for ' . $vserver );
        try {
            NACL::C::VserverCdmi->create(
                command_interface => $Node,
                vserver           => $vserver,
            );
        }
        catch NACL::APISet::Exceptions::CommandNotFoundException with {
            $Log->comment( "You are attempting to configure CDMI on a build "
                    . "that does not support it. Rerun the script "
                    . "with CREATE_CDMI=0" );
            return $TCD::FAIL;
        };

        $Test->step("Fetch NetBIOS Domain/Workgroup Name");
        my $cifsServer = NACL::C::VserverCifs->find(
            'command_interface' => $Node,
            'filter'            => { 'vserver' => $vserver }
        );

        my $netbiosDomainOrWorkGroupName
            = $cifsServer->get_one_state_attribute('domain-workgroup');

        if ( !$netbiosDomainOrWorkGroupName ) {
            $Test->nlog("CIFS is not setup for vserver");
            return $TCD::FAIL;
        }

        $Test->step("Set the delegated domain for the CDMI Root Domain");
        NACL::C::VserverCdmiDomain->modify(
            'command_interface' => $Node,
            'vserver'           => $vserver,
            'domain'            => '/cdmi_domains/',
            'delegated-domain'  => $netbiosDomainOrWorkGroupName,
        );

        # Turn on Volume Efficiency on all constituents on RW volume
        # This is needed for CDMI Named Objects which do SIS Clone operations.
        # SIS clone engine uses Dedup engine which is controlled by the
        # volume efficiency command

        my $repo_vol = NACL::STask::Volume->find(
            'command_interface' => $Node,
            'filter'            => {
                'vserver'      => $vserver,
                'volume-style' => 'infinitevol',
            },
        );

        #Enabling volume efficiency on RW volume
        my $type = $repo_vol->get_one_state_attribute('type');

        if ($type eq 'RW') {
            $Test->step(
            'Turn on volume efficiency for the InfiniteVol constituents');
            NACL::C::VolumeEfficiency->on(
                'command_interface' => $Node,
                'volume'  => $repo_vol->get_one_state_attribute('volume'),
                'vserver' => $vserver,
            );
        }
    }
    return $TCD::PASS;
}

# Create a data interface based on the information provided by the tharn host
sub create_data_interface {
    $Test->description(
        "Create data interface",
        "This test will fail if:",
        "1. tharnhost of both nodes are missing one of the following:",
        "  a. DATA_IP or DATA_IP1",
        "  b. DATA_PORT or DATA_PORT1",
        "  c. DATA_MASK or DATA_MASK1",
        "  d. DATA_GATEWAY or DATA_GATEWAY1",
        "2. The data lif already exist",
    );
    my $data_lif_num = uc( $Options->get( option => "DATA_LIF_NUM" ) );
    return $NA if ( $data_lif_num eq 0 );

    my @lifs                  = _find_lif_with_free_data_ip();
    my @vservers_without_lifs = @Vserver_Name_List;
    my $vserver;
    my $index   = 0;    #index for iterating through vserver name list
    my $numlifs = 0;    #index to keep track of the number of lifs
    if ( scalar(@Vserver_Name_List) == 0 ) {
        return $TCD::NA;
    }
    if ( @lifs == 0 ) {
        $Test->nlog('No data IP available');
        return $TCD::FAIL;
    }
    else {
        foreach my $lif (@lifs)
        {    #Iterate through every lif, and assign it to each vserver
            $vserver
                = $Vserver_Name_List[ $index % scalar(@Vserver_Name_List) ];
            _assign_data_interface(
                node          => $lif->{'node'},
                lif_num       => $lif->{'lif_num'},
                interface_num => $lif->{'interface_num'},
                vserver       => $vserver,
            );
            $index++;
            if ( $numlifs == 0 ) {
                shift(@vservers_without_lifs);
            }
            if ( $index % scalar(@Vserver_Name_List) == 0 ) {
                $numlifs++;
            }
            if ( $numlifs == $data_lif_num && $data_lif_num ne 'AUTO' ) {
                last;
            }
        }
    }
    if ( scalar(@vservers_without_lifs) > 0 && $numlifs == 0 ) {
        $Test->nlog( "WARNING: There were not enough free data IPs to "
                . "assign data LIFs to all of the vservers.\n"
                . "Vservers without data LIFS:" );
        foreach (@vservers_without_lifs) {
            $Test->nlog("$_");
        }
    }
    return $TCD::PASS;
}

# Creates an intercluster interface based on the information provided by the
# tharn host on each node in the cluster
sub create_intercluster_interface {
    $Test->description('Create intercluster interface');
    my @nodes = NACL::C::Node->find( command_interface => $Node );
    my $purge_intercluster = $Options->get( option => "PURGE_INTERCLUSTER");
    my $make_intercluster = $Options->get( option => "MAKE_INTERCLUSTER" );
    my $ipspaces_on = 0;
    my $vserver_obj;
    my $admin_vserver_name;

    # check if ipspaces feature is in the build
    # if ipspaces is in, get admin vserver name - burt 676956
    # Intercluster LIF can only reside under admin vserver and is not allowed
    # to be under node vserver in ipspaces.
    if ($Node->has_uichange(uichange => "ipspaces-ms4")) {
        $ipspaces_on = 1;
        my $vserver_obj = NACL::C::Vserver->find(
           'command_interface' => $Node,
           'filter'            => {
               'type'       => 'admin',
           },
        );
        $admin_vserver_name = $vserver_obj->vserver();
    }

    foreach my $node (@nodes) {

        my $host    = $node->hostrec();
        my $vserver_name = ($ipspaces_on) ? $admin_vserver_name : $host->id();
        my %params  = $host->param_fields();

        $Test->step('Create intercluster lif on ' . $host->id() . " for vserver $vserver_name");
        my $lif;
        my $tmp_lif = NACL::C::Interface->find(
            command_interface => $node,
            filter            => {
                role    => 'intercluster',
                'home-node'    => $host->id(),
            },
            allow_empty => 1
        );
        if ($tmp_lif && $purge_intercluster == 0) {
            $Test->nlog('Using already created lif');
            $lif = $tmp_lif;
        }
        elsif ($tmp_lif && $purge_intercluster == 1 )
        {
            $Test->nlog('Purging existing intercluster lif');
            NACL::C::Interface->delete(
                command_interface => $node,
                lif => $tmp_lif->lif,
                vserver => $vserver_name,
            );
            $tmp_lif = undef;
        }
        elsif (!exists( $params{'INTERCLUSTER_IP'} )
            && !exists( $params{'INTERCLUSTER_IP1'} )
            && !exists( $params{'DATA_IP2'} ) )
        {
            $Log->warn('INTERCLUSTER_IP missing in hostrec. ');
            $Log->warn('Please add to tharnhost so that DR can be enabled.');
            return $NA;
        }
        if($make_intercluster == 1 && !$tmp_lif)
        {
            $Test->nlog('Creating intercluster lif');
            my $ip
                = $params{'INTERCLUSTER_IP'}
                || $params{'INTERCLUSTER_IP1'}
                || $params{'DATA_IP2'};
            my $mask
                = $params{'INTERCLUSTER_MASK'}
                || $params{'INTERCLUSTER_MASK1'}
                || $params{'DATA_MASK2'};
            my $port
                = $params{'INTERCLUSTER_PORT'}
                || $params{'INTERCLUSTER_PORT1'}
                || $params{'DATA_PORT2'};

            $lif = NACL::C::Interface->create(
                command_interface => $node,
                'vserver'         => $vserver_name,
                'lif'             => 'intercluster_' . $host->id(),
                'role'            => 'intercluster',
                'home-node'       => $host->id(),
                'home-port'       => $port,
                'address'         => $ip,
                'netmask'         => $mask,
            );
        }

        if (!$ipspaces_on && defined($lif))
        {
            my $routing_group_name = $lif->get_one_state_attribute('routing-group');
            my $route = NACL::STask::NetworkRoutingGroupsRoute->find(
                command_interface => $node,
                filter => {
                    'routing-group' => $routing_group_name,
                    'vserver'       => $vserver_name,
                },
                allow_empty => 1,
            );
            if (!$route)
            {
                $Test->step('Create network route for routing group ' . $routing_group_name);
                my $gateway
                    = $params{'INTERCLUSTER_GATEWAY'}
                    || $params{'INTERCLUSTER_GATEWAY1'}
                    || $params{'DATA_GATEWAY2'};

                NACL::STask::NetworkRoutingGroupsRoute->create(
                    command_interface => $node,
                    'vserver'         => $vserver_name,
                    'routing-group'   => $routing_group_name,
                    'destination'     => '0.0.0.0/0',
                    'gateway'         => $gateway,
                );
            }
        }
    }
    return $TCD::PASS;
}

# Create the Infinitevol
sub create_repository {
    $Test->description("Create Infinitevol");

    my $need_expand = 0;
    my $job;
    my @aggrList = _get_aggr_list();

    my @params = (
        command_interface => $Node,
        'type'            => $Options->get( option => "CVOL_TYPE" ),
    );

    # We should not set any other flag for volume type other then RW
    if ( $Options->get( option => 'CVOL_TYPE' ) =~ m/rw/i ) {
        push(
            @params,
            'junction-path' =>
                $Options->get( option => "CVOL_JUNCTION_PATH" ),
            'unix-permissions' =>
                $Options->get( option => "UNIX_PERMISSIONS" ),
        );
        if ( $Options->defined( option => 'DIRECTORY_BITWIDTH' ) ) {
            push( @params,
                'directory-bitwidth' =>
                    $Options->get( option => 'DIRECTORY_BITWIDTH' ) );
        }
        if ( $Options->defined( option => 'SNAPSHOT_POLICY' ) ) {
            push( @params,
                'snapshot-policy' =>
                    $Options->get( option => 'SNAPSHOT_POLICY' ) );
        }
    }

    # SPACE_GUARANTEE and volume size two options, should support
    # to create "RW" and "DP" two types of volume
    if ( $Options->defined( option => 'CVOL_SIZE' ) ) {
        my $requested_size = $Options->get( option => "CVOL_SIZE" );

        if ( $requested_size =~ /^\+/ ) {
            $need_expand = 1;
        }
        else {
            push( @params, 'size' => $Options->get( option => "CVOL_SIZE" ) );
        }
    }
    if ( $Options->defined( option => 'SPACE_GUARANTEE' ) ) {
        push( @params,
            'space-guarantee' => $Options->get( option => 'SPACE_GUARANTEE' )
        );
    }

    if (   $Options->get( option => "ENABLE_OBJECT_STORE" ) == 1
        && $Options->defined( option => 'OLS_AGGR_LIST' ) )
    {
        my @ols_aggr_list = @{ $Options->get( option => 'OLS_AGGR_LIST' ) };

        push( @params, 'ols-aggr-list'       => \@ols_aggr_list );
        push( @params, 'enable-object-store' => 'true' );
    }

    if ( $Options->get( option => "ENABLE_SNAPDIFF" ) == 1 ) {
        push @params, 'enable-snapdiff' => 'true';
    }

    # Always create as background task
    push( @params, 'foreground' => 'false' );

    my @repos_vservers = NACL::C::Vserver->find(
        'command_interface' => $Node,
        'allow_empty' => 1,    # we might only be creating flexvol vservers
        filter => { 'is-repository' => 'true', }
    );

    my $cvol_name = $Options->get( option => 'CVOL_NAME' );
    my $iv_name = $cvol_name;
    my $iv_count = 0;
    foreach my $vs (@repos_vservers) {

        push( @params, 'job_component' => \$job );
        push( @params, 'vserver'       => $vs );

        if ( $Options->get( option => 'CREATE_IV_OF_INCR_NAMES' ) == 1 ){
            push @params, 'volume' => $iv_name;
            $Test->step( "Create infinitevol $iv_name"
                    . ' using background job on vserver '
                    . $vs->vserver );
            $iv_count++;
            $iv_name = $cvol_name . $iv_count;

        } elsif ( $Options->get( option => 'CREATE_IV_OF_RAND_NAMES' ) != 1 ) {
            push @params, 'volume' => $cvol_name;
            $Test->step( "Create infinitevol $cvol_name"
                    . ' using background job on vserver '
                    . $vs->vserver );
        } elsif ( $Options->get( option => 'CREATE_IV_OF_RAND_NAMES' ) == 1 ) {
            $Test->step('Create IV with random name on vserver ' . $vs->vserver);
        }

        my $repo_vol;
        # Putting these checks outside the try catch to minimize confusion in
        # the log in the case where the create fails
        my @large_aggrs = NACL::C::Aggregate->find(
            command_interface => $Node,
            filter            => {
                root => 'false',
                size => '>1.0TB',
            },
            allow_empty       => 1,
        );

        my $repos_min_size_table;
        try {
            $repos_min_size_table = CR::C::DebugSmdbTableReposDcMinSize->find(
                command_interface => $Node,
                allow_empty       => 1,
            );
        } otherwise  {
            $Log->warn('Unable to retrieve repos_dc_min_size table.');
        };

        try {
            $repo_vol = NACL::STask::Volume->create(@params,);
        } catch NACL::Exceptions::JobFailed with {
            my $ex = shift;

            if (!@large_aggrs &&
                !$repos_min_size_table &&
                $Node->hostrec()->hosttype() !~ /vsim/
            ) {
                $Log->warn(
                    'Creating an Infinitevol on a real Filer requires at least'
                    . ' 1.33 TB of usable space. If you'
                    . ' are doing functional testing, please run'
                    . ' CR/reduce_dc_size_check.thpl on a DEBUG build to'
                    . ' reduce the minimum DC size'
                );
            }
            $ex->throw();
        };

        if ($need_expand) {
            my $requested_size = $Options->get( option => "CVOL_SIZE" );

            $Test->step( "Modify the InfiniteVol to " . $requested_size );
            $repo_vol->modify( size => $requested_size );
        }

        if ( $Options->get( option => 'EXPORT_OBJECT_VOLUME' ) == 1 ) {
            $Test->step( 'EXPORT_OBJECT_VOLUME=1 is set, '
                    . "changing object volume to 'default' export policy" );
            my @obj_vol = NACL::STask::Volume->find(
                command_interface => $Node,
                filter            => {
                    'vserver'          => $vs,
                    'is-constituent'   => 'true',
                    'constituent-role' => 'data',
                },
            );
            foreach my $volume (@obj_vol) {
                $volume->modify( 'policy' => 'default' );
            }
        }
        if (   $Options->get( 'option' => 'CVOL_TYPE' ) =~ m/rw/i
            && $repo_vol->get_one_state_attribute('junction-path') ne
            $Options->get( 'option' => 'CVOL_JUNCTION_PATH' ) )
        {
            $Log->warn('Junction path not set, may be you hit BURT627006');
            $repo_vol->unmount();
            $repo_vol->mount( 'junction-path' =>
                    $Options->get( 'option' => 'CVOL_JUNCTION_PATH' ) );
        }

    }
    return $TCD::PASS;
}

sub create_flexvol {
    $Test->description();

    my @flex_vservers = NACL::C::Vserver->find(
        'command_interface' => $Node,
        'allow_empty'       => 1,
        'filter'            => {
            'is-repository' => 'false',
            'type' => 'data',
        },
    );

    foreach my $flex_vs (@flex_vservers) {
        my @aggr_list = $flex_vs->get_one_state_attribute('aggr-list');
        my $aggr_with_most_avail_space = undef;
        my $most_avail_space           = 0;

        # extract all aggr names if vserver aggr list is '-'
        if ( scalar(@aggr_list) == 1 && $aggr_list[0] eq '-' ) {
            my @aggrs
                = NACL::C::Aggregate->find( 'command_interface' => $Node );
            @aggr_list = map( $_->aggregate, @aggrs );
        }

        $Test->step(
                  "Finding an aggregate with most available size for flexvol"
                . " creation" );
        foreach my $aggr_name (@aggr_list) {
            my $aggr = NACL::C::Aggregate->find(
                'command_interface' => $Node,
                filter              => { aggregate => $aggr_name }
            );

            if ( $aggr->get_one_state_attribute('availsize')
                > $most_avail_space )
            {
                $aggr_with_most_avail_space = $aggr_name;
                $most_avail_space
                    = $aggr->get_one_state_attribute('availsize');
            }
        }

        $Test->step("Creating flexvol on $aggr_with_most_avail_space");
        my $fv = NACL::STask::Volume->create(
            'command_interface' => $Node,
            'volume'            => $Options->get( option => 'FLEXVOL_NAME' ),
            'vserver'           => $flex_vs->vserver,
            'aggregate'         => $aggr_with_most_avail_space,
        );

        $fv->mount( 'junction-path' => '/'
                . $Options->get( option => 'FLEXVOL_NAME' ) );
    }

    return $TCD::PASS;
}

# Helper function to delete vserver
sub delete_vserver {
    my $vserver = shift();
    my @cifs;
    my @volumes;
    my $root_vol;
    my $vs_name = $vserver->get_one_state_attribute('vserver');

    # Clean up CIFS server configure
    @cifs = NACL::C::VserverCifs->find(
        command_interface => $Node,
        'filter'          => { 'vserver' => $vs_name, },
        allow_empty       => 1,
    );
    $Test->step("Delete CIFS setting");
    foreach my $cifs (@cifs) {
        $cifs->delete(
            'username' => $Options->get( 'option' => 'AD_SERVER_USER' ),
            'password' => $Options->get( 'option' => 'AD_SERVER_PASSWORD' ),
        );
    }

    # Delete Infinitevol for this vserver, if it exists.
    my @repos = NACL::STask::Volume->find(
        command_interface => $Node,
        allow_empty       => 1,
        filter            => {
            'volume-style' => 'infinitevol',
            'vserver'      => $vs_name
        }
    );
    foreach my $repo (@repos) {
        if ( $repo->get_one_state_attribute('junction-active') =~ m/true/i ) {
            $Test->step("Unmount infinivol");
            $repo->unmount();
        }
        unless ( $repo->get_one_state_attribute('state') =~ m/(-|offline)/i )
        {
            my $job;
            $Test->step("Offline infinivol using async command");
            $repo->offline(
                'foreground'    => 'false',
                'job_component' => \$job
            );
            _bg_job_helper(
                'job'      => $job,
                'wait_msg' => 'Waiting for infinivol to go offline',
                'erro_msg' => 'Volume offline job fail',
                'warn_msg' => "Offline doesn't seem to be a background job",
            ) if $job;
        }

        my $job;
        $Test->step("Delete infinivol using async command");
        $repo->delete( 'job_component' => \$job, 'foreground' => 'false' );
        _bg_job_helper(
            'job'      => $job,
            'wait_msg' => 'Waiting for infinivol to finish delete job',
            'erro_msg' => 'Volume delete job fail',
            'warn_msg' => "Delete doesn't seem to be a background job",
        ) if $job;
    }

    try {

        # Clean up CDMI configuration if CDMI is present in this build
        $Test->step("Delete CDMI configuration");

        my @cdmi = NACL::C::VserverCdmi->find(
            command_interface => $Node,
            'filter'          => { 'vserver' => $vs_name, },
            allow_empty       => 1,
        );

        foreach my $cdmi (@cdmi) {
            $cdmi->delete( 'vserver' => $vs_name );
        }
    }
    catch NACL::APISet::Exceptions::CommandNotFoundException with {
        $Log->comment(
            "Skipping CDMI cleanup as CDMI code is not present in this build"
        );
    }
    catch NACL::APISet::Exceptions::MethodNotFoundException with {
        $Log->comment(
            "Skipping CDMI cleanup as CDMI code is not present in this build"
        );
    };

    @volumes = NACL::STask::Volume->find(
        command_interface => $Node,
        filter            => { vserver => $vs_name },
        allow_empty       => 1
    );

    $Test->step( "Stopping vserver " . $vs_name . " if it's not stopped" );
    $vserver->stop()
        if ( $vserver->get_one_state_attribute('admin-state') ne 'stopped' );

    $root_vol = undef;
    foreach my $volume (@volumes) {
        if ( $vserver->get_one_state_attribute('rootvolume') ne
            $volume->get_one_state_attribute('volume') )
        {
            $Test->step( "Unmount, offline and delete volume: "
                    . $volume->get_one_state_attribute('volume') );
            $volume->unmount()
                if ( $volume->get_one_state_attribute('junction-active')
                =~ m/true/i );
            $volume->offline()
                if (
                $volume->get_one_state_attribute('state') !~ m/offline/i );
            $volume->delete(
                "method-timeout" => $Options->get( option => "TIMEOUT" ) );
        }
        else {
            $root_vol = $volume;
        }
    }
    if ( defined $root_vol ) {
        $Test->step( "Deleting root volume: "
                . $root_vol->get_one_state_attribute('volume') );
        $root_vol->offline()
            if ( $root_vol->get_one_state_attribute('state') !~ m/offline/i );
        $root_vol->delete(
            "method-timeout" => $Options->get( option => "TIMEOUT" ) );
    }
    $Test->step( "Deleting vserver: " . $vs_name );
    {    # Code to work around issue in burt 483819
        try {
            $vserver->delete(
                "method-timeout" => $Options->get( option => "TIMEOUT" ) );
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            $vserver->start();
            $vserver->delete(
                "method-timeout" => $Options->get( option => "TIMEOUT" ) );
            $Log->warn( "Apply work around for BURT 483819 to delete vserver "
                    . $vs_name );
        }
    };
}

sub xget_bytes_from_string {
    my ($string_size) = @_;
    return _get_bytes_from_string($string_size);
}

# Return the size in bytes from a string
sub _get_bytes_from_string() {
    my $string_size = lc(shift);
    my $size;

    # Translate from KB, MB, GB, TB, or PB into bytes
    if ( $string_size =~ /k/ ) {
        $string_size =~ s/^\+([0-9]+).*/$1/;
        $size = $string_size * 1024;
    }
    elsif ( $string_size =~ /m/ ) {
        $string_size =~ s/^\+([0-9]+).*/$1/;
        $size = $string_size * ( 1024**2 );
    }
    elsif ( $string_size =~ /g/ ) {
        $string_size =~ s/^\+([0-9]+).*/$1/;
        $size = $string_size * ( 1024**3 );
    }
    elsif ( $string_size =~ /t/ ) {
        $string_size =~ s/^\+([0-9]+).*/$1/;
        $size = $string_size * ( 1024**4 );
    }
    elsif ( $string_size =~ /p/ ) {
        $string_size =~ s/^\+([0-9]+).*/$1/;
        $size = $string_size * ( 1024**5 );
    }
    elsif ( $string_size =~ /b/ ) {
        $string_size =~ s/^\+([0-9]+).*/$1/;
        $size = $string_size;
    }
    else {

        # This is an error case.  We could not parse
        # the string as either KB, MB, GB, TB, or PB.
        #
        # Disregard the string by setting $size to 0.
        $Test->nlog( "Could not parse CVOL_SIZE option: " . $string_size );
        $size = 0;
    }
    return $size;
}

# Return the size of the largest aggregate in the supplied array
sub _largest_aggr_size() {

    my @aggrList     = @_;
    my $largest_size = 0;

    foreach my $aggr_name (@aggrList) {

        my $aggr = NACL::C::Aggregate->find(
            command_interface => $Node,
            filter            => { aggregate => $aggr_name },
            allow_empty       => 0,
        );

        my $size = $aggr->get_one_state_attribute('size');

        if ( $size > $largest_size ) {
            $largest_size = $size;
        }
    }

    return $largest_size;
}

# Return the size of the OLS aggregates
sub _ols_aggr_size() {

    my $ols_size = 0;

    if ( $Options->defined( option => 'OLS_AGGR_LIST' ) ) {
        my @aggrList = @{ $Options->get( option => 'OLS_AGGR_LIST' ) };

        foreach my $aggr_name (@aggrList) {
            my $aggr = NACL::C::Aggregate->find(
                command_interface => $Node,
                filter            => { aggregate => $aggr_name },
                allow_empty       => 0,
            );
            $ols_size += $aggr->get_one_state_attribute('size') if ($aggr);
        }
    }

    return $ols_size;
}

# Generate an array containing unique item based on the input array
# The item will use the same order of the input
sub _unique_list(@) {
    my @raw_list = @_;
    my @result;
    my %tmp_hash;
    foreach my $item (@raw_list) {
        push( @result, $item ) unless defined $tmp_hash{$item};
        $tmp_hash{$item} = 1;
    }
    return @result;
}

# This function get the prefer disk type from the system based on the PREFERRED_DISK_TYPE
# It will also look for the AGGR_DISK_COUNT parameter to make sure there are enough
# disk before returning the preferred disk type.
# If non of the preferred disk type found, it return nothing.
sub prefer_disk_type_get($) {
    my $node = $_[0];
    foreach my $type ( @{ $Options->get( option => 'PREFERRED_DISK_TYPE' ) } )
    {
        my @disks = NACL::C::Disk->find(
            command_interface => $Node,
            filter            => {
                'type'      => $type,
                'aggregate' => '-',
                'owner'     => $node,
            },
            allow_empty => 1,
        );
        return $type
            if (
            scalar(@disks) >= $Options->get( option => 'AGGR_DISK_COUNT' ) );
    }
}

# This function will wait for the background job to complete
# Input:
#   job      - The NACL::C::Job object
#   wait_msg - The message we want to display as steps while we are waiting.
#   warn_msg - The warning message we want to display if the job complete alerady.
sub _bg_job_helper {
    my %opt = validate_with(
        params => \@_,
        spec   => {
            'job'      => { type => HASHREF },
            'wait_msg' => { type => SCALAR | UNDEF, optional => 1 },
            'warn_msg' => { type => SCALAR | UNDEF, optional => 1 },
            'erro_msg' => { type => SCALAR | UNDEF, optional => 1 },
        },
    );
    my $job      = $opt{'job'};
    my $wait_msg = $opt{'wait_msg'};
    my $warn_msg = $opt{'warn_msg'};
    my $erro_msg = $opt{'erro_msg'};
    my $job_state;
    $Test->step($wait_msg) if ($wait_msg);
    {
        try {
            $job_state = $job->get_one_state_attribute('state');
        }
        catch NACL::Exceptions::NoElementsFound with {
            $Log->warn(
                "Job disappear from job list, may be it's changing state, retry again"
            );
            $Test->nlog(
                "Job disappear from job list, may be it's changing state, retry again"
            );
            $job_state = $job->get_one_state_attribute('state');
        };
        sleep(5) if ( $job_state =~ m/queued/i );
    };
    if ( $job_state !~ m/(running|queued)/i && $warn_msg ) {
        $Log->warn($warn_msg);
        $Test->nlog($warn_msg);
    }
    $job->watch_progress(
        'method-timeout' => $Options->get( option => "TIMEOUT" ) );
    if (   $job->get_one_state_attribute('state') !~ m/success/i
        && $erro_msg )
    {
        NACL::APISet::Exceptions::ResponseException->throw(
            text            => $erro_msg,
            response_object => $job
        );
    }
}

sub _get_aggr_list {
    return _unique_list(
        @{ $Options->get( option => 'AGGR_NAME_LIST' ) },
        (          $Options->defined( option => 'OLS_AGGR_LIST' )
                && $Options->get( 'option' => 'ENABLE_OBJECT_STORE' ) == 1
            )
        ? @{ $Options->get( option => 'OLS_AGGR_LIST' ) }
        : (),
        $Options->get( option => 'VSERVER_AGGR' ),
    );
}

sub _generate_cifs_server_name {
    my %params = @_;

    my $cifsvsname;

    if ( !$Options->get( 'option' => 'CIFS_SERVER' ) ) {
        $cifsvsname = $Node->get_one_state_attribute('node');
        $cifsvsname .= $params{'vserver_name'};
        $cifsvsname =~ s/-|_//g;
        my $i = length($cifsvsname);
        if ( $i > 15 ) {
            $cifsvsname = substr( sha1_hex("$cifsvsname"), 0, 15 );
        }
    }
    else {
        $cifsvsname = $Options->get( 'option' => 'CIFS_SERVER' );
    }

    $cifsvsname = uc $cifsvsname;
}

sub _find_lif_with_free_data_ip {
    my @lifs = ();
    my @all_nodes = NACL::C::Node->find( command_interface => $Node );
    foreach my $node (@all_nodes) {
        my $host   = $node->hostrec();
        my %params = $host->param_fields();

        #mapped list of lif names to avoid if we have doubles
        my @interfaces = map( $_->lif,
            NACL::C::Interface->find( 'command_interface' => $node));
        my $lif_num = 1;
        #list of DATA_IP's from tharn host file
        my @addresses = grep( /DATA_IP/, keys %params );
        foreach my $interface_num (@addresses) {

            #strip interface_num from the full name eg. DATA_IP1 will now be 1
            $interface_num =~ s/DATA_IP//;
            my $lif = NACL::C::Interface->find(
                command_interface => $node,
                filter            => {
                    'address' => $params{ 'DATA_IP' . $interface_num },
                    'role'    => '*',
                },
                allow_empty => 1,
            );
            if ( !$lif ) {
               #keep incrementing lif_num until we find a data lif without naming conflicts
                my $hostname = $host->id;
                while (grep(/^data$lif_num.$hostname$/,@interfaces) > 0 )
                {   $lif_num++;}
                #push lif object into return array
                my $return_lif = {
                    node          => $node,
                    interface_num => $interface_num,
                    lif_num       => $lif_num,
                };
                push( @lifs, $return_lif );
                $lif_num++;
            }
        }
    }
    return @lifs;
}

# Create a data interface based on the information provided by the tharn host
sub _assign_data_interface {
    my %args          = @_;
    my $node          = $args{'node'};
    my $vserver       = $args{'vserver'};
    my $lif_num       = $args{'lif_num'};
    my $interface_num = $args{'interface_num'};
    my $host          = $node->hostrec();
    my %params        = $host->param_fields();

    $Test->step(
        "Create data lif on " . $host->id() . " for vserver $vserver." );
    my $lif = NACL::C::Interface->create(
        command_interface => $Node,
        'vserver'         => $vserver,
        'lif'             => "data$lif_num" . "_" . $host->id(),
        'role'            => 'data',
        'data-protocol' =>
            \@{ $Options->get( option => 'DATA_LIF_PROTOCOL' ) },
        'home-node' => $host->id(),
        'home-port' => $params{ 'DATA_PORT' . $interface_num },
        'address'   => $params{ 'DATA_IP' . $interface_num },
        'netmask'   => $params{ 'DATA_MASK' . $interface_num },
    );

    if ( $Options->get( option => 'VSADMIN_ON_DATA_LIF' ) ) {
        $Test->step(
            'VSADMIN_ON_DATA_LIF = 1, changing data lif firewall policy to mgmt'
        );
        $lif->modify( 'firewall-policy' => 'mgmt' );
    }
    my $routing_group = $lif->get_one_state_attribute('routing-group');
    my @route         = NACL::STask::NetworkRoutingGroupsRoute->find(
        command_interface => $Node,
        filter            => {
            'vserver'       => $vserver,
            'routing-group' => $routing_group,
        },
        'allow_empty' => 1,
    );

    # Setup the gateway if needed
    if ( !@route ) {
        $Test->step(
            "Create network route for routing group " . $routing_group );
        NACL::STask::NetworkRoutingGroupsRoute->create(
            command_interface => $Node,
            'vserver'         => $vserver,
            'routing-group'   => $routing_group,
            'destination'     => '0.0.0.0/0',
            'gateway'         => $params{ 'DATA_GATEWAY' . $interface_num },
        );
    }

    if ( $Options->get( option => 'VSADMIN_ON_DATA_LIF' ) ) {
        try {
            $Test->step("Setting $Vsadmim_name password");
            NACL::C::SecurityLogin->password(
                command_interface => $node,
                username          => $Vsadmim_name,
                vserver           => $vserver,
                newpwd => $Options->get( option => 'VSADMIN_PASSWORD' ),
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $ex = shift;
            $ex->throw
                unless ( $ex->text
                =~ /New password must be different than old password/ );
        };
        $Test->step("Unlock $Vsadmim_name account");
        NACL::C::SecurityLogin->unlock(
            command_interface => $node,
            username          => $Vsadmim_name,
            vserver           => $vserver,
        );
    }
    return $TCD::PASS;
}

sub _vserver_nfs_modify {
    my %parameters = @_;

    $parameters{'vserver'} = $Options->get( option => 'VSERVER' )
        unless ( exists $parameters{'vserver'} );

    {
        try {
            $Apiset->vserver_nfs_modify(%parameters);
        }
        catch NATE::BaseException with {
            my $err_msg = shift()->text();
            return $TCD::FAIL
                if $err_msg !~ m/invalid argument "-mount-rootonly"/;
            $Log->warn('"-mount-rootonly" is not a valid NFS server option');
            return $TCD::PASS;
        }
    }

    return $TCD::PASS;
}

1;
