#===============================================================================
#
#         FILE: $Id$
#
#  DESCRIPTION: This Library contains the helper functions for all common tasks
#               which are neither specific to DR (snapmirror/snapshot), Posix,
#               Client or Setup operations.
#
#         BUGS: ---
#        NOTES: Not yet
#       AUTHOR: dl-svl-nosauto
#      COMPANY: NetApp, Inc
#      VERSION: 1.0
#      CREATED: Thursday 18,October,2012 09:16:39  EDT
#     MODIFIED: $DateTime$
# LAST CHANGED: $Author$
#===============================================================================

package CR::CommonUtils;

use strict;
use warnings;
use String::Random;
use List::MoreUtils qw( any );
use NACL::C::SystemNode;
use NACL::CS::SystemNodeDate;
use Tharn qw( $Log );
use NACL::C::RepositoryPlacementFileCreate;
use NACL::C::JobLog;
use Data::Compare;
use CR::CRHttpRequest;
use CR::CRContainerObject;
use CR::CRDataObject;
use NATE::BaseException qw(:try);
use Params::Validate qw(SCALAR OBJECT ARRAYREF HASHREF validate);
use NACL::C::DiagNbladeRepositoryFhmapCache;
use Data::Dumper;
use Data::Compare;
use File::Basename qw(dirname);
use Exporter 'import';

our @EXPORT_OK = qw(
                    enable_autoboot_on_node
                    get_counted_random_string
                    set_fiji
                    fiji_clear_all
                    verify_avail_files_count_on_dcs
                    recalculate_bucket_status
                    verify_files_in_fhmap
                    verify_storage_service_via_fhmap
                    verify_lif_migration_on_node
                    set_force_create_on_dc
                    verify_fhmap_data_for_file
                    get_fhmap_output_for_file
                    set_debug_logging
                    get_current_min
                    get_current_sec
                    create_container
                    delete_container
                    create_named_object
                    delete_named_object
                    read_named_object
                    create_nameless_object
                    delete_nameless_object
                    read_nameless_object
                   );
our %EXPORT_TAGS = (cdmi => [qw(create_container
                                delete_container
                                create_named_object
                                delete_named_object
                                read_named_object
                                create_nameless_object
                                delete_nameless_object
                                read_nameless_object)],
                   );


#===  FUNCTION  ================================================================
#         NAME: enable_autoboot_on_node
#      PURPOSE: To enable autoboot on specified node.
#   PARAMETERS: node_obj (required): The object (command_interface) of the node
#               on which autoboot needs to be enabled.
#      RETURNS: 1
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#        USAGE: enable_autoboot_on_node (node_obj => $command_interface);
#===============================================================================
sub enable_autoboot_on_node{

    my %opts = validate(@_, { node_obj => {type => OBJECT, optional => 0} });

    my $node_api = $opts{node_obj}->apiset(
        interface => 'CLI',
        category  => 'Node',
        set       => 'Nodescope',
    );

    $node_api->bootargs_set(
        arg   => 'AUTOBOOT',
        value => 'true'
    );

    return 1;

}

#===  FUNCTION  ================================================================
#         NAME: get_counted_random_string
#      PURPOSE: To get a random string of length specified by count.
#   PARAMETERS: 1. count (Required): Lenght of the desired random string.
#      RETURNS: Random string having length of specified count.
#  DESCRIPTION: String will be comprised of all Alphanumeric characters plus "_"
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub get_counted_random_string{

    my %opts = validate(@_,{
        count => {type => SCALAR,
                  regex => qr/^\d+$/,}}
    );

    my $str = String::Random->new();

    my $random_string = $str->randregex("\\w{$opts{count}}");

    return $random_string;
}

#===  FUNCTION  ================================================================
#         NAME: set_fiji
#      PURPOSE: sets the specified fiji on all the node objects specified
#   PARAMETERS: 1. fiji
#               2. node_objs : Reference of the array of NACL::C::Node Objects
#      RETURNS: true
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================

sub set_fiji {

    my %opts = validate(@_, { node_objs => {type => ARRAYREF, optional => 0},
                              fiji => {type => SCALAR, optional => 0},
                            });

    my $fiji = $opts{'fiji'};
    my @node_objs = @{$opts{'node_objs'}};

    foreach my $node_obj (@node_objs) {
        NACL::C::SystemNode->run(
                command_interface => $node_obj,
                node    => $node_obj->node(),
                command => $opts{'fiji'}
        );
    }

    return 1;
}

#===  FUNCTION  ================================================================
#         NAME: fiji_clear_all
#      PURPOSE: clear all the fiji on all the node objects specified
#   PARAMETERS: 1. node_objs : Reference of the array of NACL::C::Node Objects
#      RETURNS: true
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================

sub fiji_clear_all {

    my %opts = validate(@_, { node_objs => {type => ARRAYREF, optional => 0} });

    my @node_objs = @{$opts{'node_objs'}};

    foreach my $node_obj (@node_objs) {
        NACL::C::SystemNode->run(
                command_interface => $node_obj,
                node    => $node_obj->node(),
                command => 'fiji clear all'
        );
    }

    return 1;
}

#===  FUNCTION  ================================================================
#         NAME: verify_avail_files_count_on_dcs
#      PURPOSE: To verify if the available count to create files is not less
#               than specified value.
#   PARAMETERS: dc_objs (Required): Array ref containing DC objects whose
#                   available file count (inodes) is to be compared.
#               file_count (Required): The file count which is to be compared.
#      RETURNS: False if the available file count is less than specified value
#               else True
#  DESCRIPTION: It first calculates the value of files for the DC and then the
#               value of files-used. It then checks if the subtraction of files
#               and files-used is not less than specified count. If it is then
#               a false is return else it moves to next DC. This process is
#               performed for all the specified DCs.
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub verify_avail_files_count_on_dcs{

    my %opts = validate(@_, {
        dc_objs => {type => ARRAYREF, optional => 0},
        file_count => {type => SCALAR, optional => 1}});

    my ( %initial_files_used );

    foreach my $dc_obj ( @{$opts{dc_objs}} ) {

        my $files_count = $dc_obj->get_one_state_attribute('files');
        my $files_used_count = $dc_obj->get_one_state_attribute('files-used');

        my $dc_name = $dc_obj->volume();

        my $expected_file_count = $opts{file_count};
        my $actual_file_count = $files_count - $files_used_count;

        if ( $actual_file_count < $expected_file_count ){
            $Log->comment("Volume $dc_name doesn't have enough space."
                        . ' Please ensure that each DC has space to create'
                        . ' at least 200 files');
            return 0;
        }

        $initial_files_used{$dc_obj->get_one_state_attribute('msid')} =
            $files_used_count;

    }

    return \%initial_files_used;

}

#===  FUNCTION  ================================================================
#         NAME: recalculate_bucket_status
#      PURPOSE: To forcefully recalculate the bucket status ahead of schedule.
#   PARAMETERS: volume_obj (Required): NACL Component object of IV
#               command_interface (Optional): Node obj which is available.
#      RETURNS: 1
#       THROWS: no exceptions
#     COMMENTS: It only works for one DCG as of now. See TODO.
#         TODO: Will be nice to parameterize the DCG so that user can provide
#               it as an input. Additional points for having an option if user
#               wants to recalculate for all the DCGs.
#     SEE ALSO: n/a
#===============================================================================
sub recalculate_bucket_status{

    my %opts = validate(@_, {
        volume_obj => {type => OBJECT, optional => 0},
        command_interface => {type => OBJECT, optional => 1}
    });

    my $vserver_name = $opts{volume_obj}->vserver();
    my $volume_name = $opts{volume_obj}->volume();
    my $node_obj = $opts{command_interface}
                   || $opts{volume_obj}->command_interface();

    # find dc group id
    my $repos_plac_obj = NACL::C::RepositoryPlacementFileCreate->find(
        command_interface => $node_obj,
        filter            => {
            vserver => $vserver_name,
            volume  => $volume_name,
        },
    );

    my $dc_group_id
        = $repos_plac_obj->get_one_state_attribute('constituent-group-id');

    $Log->comment('Restoring DC buckets to actual state');
    NACL::C::RepositoryPlacementFileCreate->recalculate(
        command_interface      => $node_obj,
        'vserver'              => $vserver_name,
        'constituent-group-id' => $dc_group_id,
        'volume'               => $volume_name,
    );

    return 1;

}

#===  FUNCTION  ================================================================
#         NAME: verify_files_in_fhmap
#      PURPOSE: To verify that given files are present in FHMAP.
#   PARAMETERS: vserver_obj (Required): NACL Component object of InfiniteVol
#               path (Required): Directory path including junction-path where
#                   the files are present.
#               files (Required): Array ref containing files present in path.
#               node_name (Required): Name of the node on which FHMAP is to be
#                   checked.
#      RETURNS: 1 if all the specified files are present in FHMAP else 0.
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub verify_files_in_fhmap{

    my %opts = validate(@_, {
        vserver_obj => {type => OBJECT, optional => 0},
        path => {type => SCALAR, optional => 0},
        files => {type => ARRAYREF, optional => 0},
        node_name => {type => SCALAR, optional => 0},
        command_interface => {type => OBJECT, optional => 1},
    });

    my $command_interface = $opts{command_interface}
                            || $opts{vserver_obj}->command_interface();

    my $return_value = 1;

    my @files = @{$opts{files}};

    foreach my $file ( @files ) {

        eval {
            my $fhmap_obj = NACL::C::DiagNbladeRepositoryFhmapCache->find(
               command_interface => $command_interface,
               filter => {
                   'namespace-path' => "$opts{path}/$file",
                   vserver => $opts{vserver_obj}->vserver(),
                   node => $opts{node_name},
               },
            );

            $Log->comment("File $opts{path}/$file is present in FHMAP");
         };

         if ( $@ ) {

            $Log->comment("File $opts{path}/$file is not present in FHMAP."
                          . " Reason: $@");
            $return_value = 0;
        }
    }

    return $return_value;

}

#===  FUNCTION  ================================================================
#         NAME: verify_lif_migration_on_node
#      PURPOSE: To verify that data lif has migrated to node other than home
#               node and is up.
#   PARAMETERS: 1. remote_node_obj
#                  Remote node object i.e node which is not home node (Required)
#               2. lif_name
#                  Data lif to be checked (Required)
#      RETURNS: Network interface object
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub verify_lif_migration_on_node {

    my %opts = validate(@_, {
        remote_node_obj => {type => OBJECT, optional => 0},
        lif_filter => {type => HASHREF, optional => 0},
    });

    my $net_int_obj;

    # Trying in loop for lif migration to complete as lif migration can take
    # few seconds before it is successfully performed.
    for ( 1 .. 5 ){
        $net_int_obj = NACL::C::Interface->find(
            command_interface => $opts{remote_node_obj},
            filter            => {
                role           => 'data',
                'curr-node'    => $opts{remote_node_obj}->node(),
                'is-home'      => 'false',
                'status-admin' => 'up',
                %{$opts{lif_filter}},
            },
            allow_empty => 1,
        );

        if ($net_int_obj){
            $Log->comment('LIF migration successfully completed to node: '
                          . $opts{remote_node_obj}->node());
            last;
        } else {
            $Log->comment('Waiting for LIF migration to complete on node: '
                          . $opts{remote_node_obj}->node());
            sleep 5;
        }

    }

    return $net_int_obj;

} ## end sub verify_lif_migration_on_node


#===  FUNCTION  ================================================================
#         NAME: verify_fhmap_data_for_file
#      PURPOSE: To verify the actual o/p of fhmap with the expected data.
#   PARAMETERS: vserver_obj (Required): NACL vserver obj to which IV belongs.
#               file_path (Required): Complete path with junction-path to be
#                   looked up in FHMAP.
#               node_name (Required): Node on which is FHMAP is to be expected.
#               attr_ref (Required): Referrence to hash containing fields as
#                   keys with actual values to be compared. Do no include
#                   NS and Data FH and als entry-flags.
#               command_interface (Required): Node on which FHMAP command is to
#                   be executed.
#      RETURNS: 1 in case of success else 0 in case of failure.
#  DESCRIPTION: If command interface is not provided then it picks one from
#               vserver object. Currently namespace-filehandle, data-filehandle
#               and entry-flags are excluded from comparison.
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub verify_fhmap_data_for_file{

    my %opts = validate(@_, {
        vserver_obj => {type => OBJECT, optional => 0},
        file_path => {type => SCALAR, optional => 0},
        node_name => {type => SCALAR, optional => 0},
        attr_ref => {type => HASHREF, optional => 0},
        command_interface => {type => OBJECT, optional => 1},
    });

    my $command_interface = $opts{command_interface}
                            || $opts{vserver_obj}->command_interface();


    my $api_set_obj = $command_interface->apiset();

    my $fh_response = $api_set_obj->diag_nblade_repository_fhmap_cache_show(
       'namespace-path' => $opts{file_path},
       node => $opts{node_name},
       vserver => $opts{vserver_obj}->vserver(),
    );

    my @fh_parsed_output = $fh_response->get_parsed_output;

    my %fhmap_fields = %{$fh_parsed_output[0]->[0]};

    # Removing following fields since currently we are not comparing it
    delete $fhmap_fields{'entry-flags'};

    $opts{attr_ref}->{filer} = $opts{node_name};
    $opts{attr_ref}->{node} = $opts{node_name};

    if ( ! Compare (\%fhmap_fields, $opts{attr_ref}) ){
        $Log->error('Mismatch found in FHMAP output.'
                    . ' Expected: ' . Dumper($opts{attr_ref})
                    . ' Actual: ' . Dumper(\%fhmap_fields));
        return 0;

    }

    return 1;
}

#===  FUNCTION  ================================================================
#         NAME: set_force_create_on_dc
#      PURPOSE: To select a DC to send all the creates to it forcefully.
#   PARAMETERS: iv_obj (Required): NACL Component IV object to which the DC
#               belongs.
#               dc_name (Required): Name of the DC on which all the creates
#               would go.
#               vserver
#      RETURNS: Referrence to hash containing attribute values to be reset.
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub set_force_create_on_dc{

    my %opts = validate(@_, {
        iv_obj => {type => OBJECT, optional => 0},
        dc_name => {type => SCALAR, optional => 0},
    });

    $Log->comment(
        'Modify repository placment file-create policy, set create-force-volume'
        . " to be the target DC $opts{dc_name}"
    );

    my @keys = qw (
        create-force-volume
        constituent-group-id
        create-empty-ratio
        create-regular-ratio
        volume
        vserver
    );

    my $iv_name = $opts{iv_obj}->volume();
    my $vserver_name = $opts{iv_obj}->get_one_state_attribute('vserver');

    my $repos_plac_obj = NACL::C::RepositoryPlacementFileCreate->find(
        command_interface => $opts{iv_obj}->command_interface(),
        filter            => {
            vserver => $vserver_name,
            volume  => $iv_name,
        },
    );

    my %params;
    foreach my $key (@keys) {
        $params{$key} = $repos_plac_obj->get_one_state_attribute($key);
    }

    $repos_plac_obj->modify(
        %params,
        'create-force-volume'  => $opts{dc_name}
    );

   return [$repos_plac_obj, \%params];
}

#===FUNCTION====================================================================
#         NAME: verify_storage_service_via_fhmap
#      PURPOSE: To verify file's storage service type by validating DC attribute
#               got from fhmap cache is one of the DCs that lies in expected DCG
#   PARAMETERS: files (Required): A reference to an array of file names
#               junction_path (Optional): IV junction path
#               dcg_info (Required): A reference to a hash that contains storage
#                                    service information of DCs
#               storage_service_type (Required): Expected storage service type
#               node_obj (Required): NACL Component object of IV
#               vserver_name (Required): Name of vserver
#      RETURNS: 1
#       THROWS: no exceptions
#     COMMENTS: It only works for one DCG as of now. See TODO.
#         TODO: Will be nice to parameterize the DCG so that user can provide
#               it as an input. Additional points for having an option if user
#               wants to recalculate for all the DCGs.
#     SEE ALSO: n/a
#===============================================================================
sub verify_storage_service_via_fhmap {

    my %opts = validate(@_,
        {files         => {type => ARRAYREF, optional => 0},
         junction_path => {type => SCALAR, optional => 1, default => '/NS'},
         dcg_info      => {type => HASHREF, optional => 0},
         storage_service_type => {type => SCALAR, optional => 0},
         vserver_obj   => {type => OBJECT, optional => 0},}
    );
    my $ss_type       = $opts{storage_service_type};
    my $dcg_info      = $opts{dcg_info};
    my $junction_path = $opts{junction_path};
    my $vserver_obj   = $opts{vserver_obj};
    my $node_obj      = $vserver_obj->command_interface();
    my $node_api_obj  = $node_obj->apiset();

    foreach my $file_name (@{$opts{files}}) {

        # Get the namespace path of the file
        my $file_ns_path = $junction_path . q{/} . $file_name;

        my $err_msg = 'show failed: Failed to get file handle. '
            . 'Error: No such file or directory';

        # Do a fhmap cache show
        my $response = $node_api_obj->diag_nblade_repository_fhmap_cache_show(
            'namespace-path' => $file_ns_path,
            node =>  $node_obj->node(),
            vserver => $vserver_obj->vserver()
        );

        if ($response =~ /$err_msg/) {
            $Log->error(
                "Unexpected! Cannot find file $file_name through fhmap cache"
                . "Error message: $response\n"
            );
            return 0;
        }

        $Log->comment("File $file_name is found through fhmap cache show");

        my @parsed_output = $response->get_parsed_output();
        my $data_const_name = $parsed_output[0]->[0]->{'data-constituent'};
        if ( any {$_->volume() eq $data_const_name} @{$dcg_info->{$ss_type}}) {
            $Log->comment(
                "File $file_name is created on DC $data_const_name, which "
                . "belongs to $ss_type DCG"
            );
        } else {
            $Log->error(
                "Unexpected! File $file_name is created on DC $data_const_name,"
                . " which does not belong to $ss_type DCG"
            );
            return 0;
        }
    }
    return 1;
}

#===  FUNCTION  ================================================================
#         NAME: get_fhmap_output_for_file
#      PURPOSE: To gather fhmap fields and it's values for a given file
#               belonging to given vserver.
#   PARAMETERS: vserver_obj (Required): NACL vserver obj to which IV belongs.
#               file_path (Required): Complete path with junction-path to be
#                   looked up in FHMAP.
#               command_interface (Required): Node on which FHMAP command is to
#                   be executed.
#      RETURNS: Hash containing fhmap fields with corresponding values.
#  DESCRIPTION: It executes volume_file_repository_fhmap_show command for a
#               given file, gathers the fhmap o/p data and returns the hash.
#               Filehandles are currently excluded.
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub get_fhmap_output_for_file{

    my %opts = validate(@_, {
        vserver_obj => {type => OBJECT, optional => 0},
        file_path => {type => SCALAR, optional => 0},
        command_interface => {type => OBJECT, optional => 1},
    });

    my $command_interface = $opts{command_interface} ||
                            $opts{vserver_obj}->command_interface();

    my $response = $command_interface
                   ->apiset()->volume_file_repository_fhmap_show(
        vserver => $opts{vserver_obj}->vserver(),
        path    => $opts{file_path},
    );

    my @parsed_output = $response->get_parsed_output();
    my %fh_fields     = %{ $parsed_output[0]->[0] };

    # Deleting following fields as they are not present in
    # diag nblade repository fhmap o/p
    delete $fh_fields{path};
    delete $fh_fields{filehandle};

    return %fh_fields;

}

#===  FUNCTION  ================================================================
#         NAME: get_current_min
#      PURPOSE: Get the minute from the current time
#   PARAMETERS: node_obj (required): The object (command_interface) of the node
#      RETURNS: Minute value
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub get_current_min {
    my %opts = validate(@_, { node_obj => {type => OBJECT, optional => 0} });

    my $node_obj = $opts{node_obj};
    my $date_state_obj =
       NACL::CS::SystemNodeDate->fetch(command_interface => $node_obj,
                                       filter => {node => $node_obj->node()});
    my $current_time = $date_state_obj->date();
    my $min;
    if ($current_time =~ /(\d+)\/(\d+)\/(\d+)\s+(\d+):(\d+):(\d+)/) {
        $min = $5;
    }

    return $min;
}

#===  FUNCTION  ================================================================
#         NAME: get_current_sec
#      PURPOSE: Get the second from the current time
#   PARAMETERS: node_obj (required): The object (command_interface) of the node
#      RETURNS: Second value
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub get_current_sec {
    my %opts = validate(@_, { node_obj => {type => OBJECT, optional => 0} });

    my $node_obj = $opts{node_obj};
    my $date_state_obj =
       NACL::CS::SystemNodeDate->fetch(command_interface => $node_obj,
                                       filter => {node => $node_obj->node()});
    my $current_time = $date_state_obj->date();
    my $sec;
    if ($current_time =~ /(\d+)\/(\d+)\/(\d+)\s+(\d+):(\d+):(\d+)/) {
        $sec = $6;
    }

    return $sec;
}

#===  FUNCTION  ================================================================
#         NAME: set_debug_logging
#      PURPOSE: Set debug logging on the given node
#   PARAMETERS: node objects (required): A reference to an array of node objects
#      RETURNS: 1 in case of success else 0 in case of failure
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#        USAGE: set_debug_logging(ref_node_objs => \@node_objs,
#                                 level => 'info');
#===============================================================================
sub set_debug_logging {
    my %opts = validate(@_, {
             ref_node_objs => {type => ARRAYREF, optional => 0},
             level => {type => SCALAR, optional => 1, default => 'debug'},
        });

    my @node_objs = @{$opts{ref_node_objs}};
    my $level = $opts{level};

    my @valid_level = qw(emerg alert crit err warning notice info debug);

    my $found_level = grep {$_ eq $level} @valid_level;
    if (!$found_level) {
        $Log->warn("Failed to change the logging level to $level");
        return 0;
    }

    foreach my $node_obj (@node_objs) {
        # execute the log command on the specific node
        my $node_name = $node_obj->node();
        $Log->comment("Enable debug logging on node $node_name");

        $Log->comment(
            "Changing the 'job log' logging level for CR* modules to $level");
        my $apiset = $node_obj->apiset(
                category  => 'Node',
                interface => 'CLI',
                set       => 'CMode'
            );
        NACL::C::JobLog->modify(
            command_interface => $node_obj,
            'module'          => 'CR*',
            'level'           => $level
        );

        $Log->comment(
            'Changing the \'logger mgwd log\' logging level for CR* modules to '
            .$level);
        my $res = $apiset->execute_raw_command(
            command => "logger mgwd log modify -module CR* -level $level");

        if ( $res =~ /Error:/ ) {
            $Log->warn(
                'Failed to change logging level: you may not use a debug build'
            );
        }

        $Log->comment(
            'Changing the \'logger mgwd log\' logging level for tables::repos* '
            ."modules to $level");
        $res = $apiset->execute_raw_command( command =>
            "logger mgwd log modify -module tables::repos* -level $level");

        if ( $res =~ /Error:/ ) {
            $Log->warn(
                'Failed to change logging level: you may not use a debug build'
            );
        }

        $Log->comment(
            'Changing the \'logger mgwd log\' logging level for repos module '
            ."to $level");
        $res = $apiset->execute_raw_command(
            command => "logger mgwd log modify -module repos -level $level");

        if ( $res =~ /Error:/ ) {
            $Log->warn(
                'Failed to change logging level: you may not use a debug build'
            );
        }
    }

    return 1;

}


###############################################################################
# Name        : create_container
# Description : Creates a container given a volume object
# Input       : Hash with following keys
#               vol_obj         Volume object on which the container is to be
#                               created (Required)
#               name            Container name, defaults to cdmi
#               uri_scheme      Uri scheme, defaults to http
#               cdmi_port       CDMI port to use, defaults to 8081
#               body            Metadata (Optional)
#                               Example:
#                       $c_create_body = {
#                          metadata => {
#                                       usmd1 => 'value1',
#                                       usmd2 => 'value2',
#                                       usmd3 => {
#                                                 usmd4 => 'value4',
#                                                 usmd5 => 'value5',
#                                                 usmd6 => {
#                                                           usmd7 => 'value7',
#                                                           usmd8 => 'value8',
#                                                           usmd9 => 'value9',
#                                                          },
#                                                },
#                                      },
#                       };
# Returns     : On success, hash ref with keys container_obj and uri
#               undef on failure
###############################################################################
sub create_container {
    my (%params) = @_;

    my $vol_obj = $params{vol_obj};
    my $name  = $params{name} || 'cdmi';
    my $uri_scheme = $params{uri_scheme} || 'http';
    my $cdmi_port  = $params{cdmi_port} || 8081;
    my $body = $params{body};

    my $node_obj = $vol_obj->command_interface();
    my $junc_path = $vol_obj->get_one_state_attribute('junction-path');

    my $interface_cs_obj = NACL::CS::NetworkInterface->fetch(
                                command_interface => $node_obj,
                                filter            => {
                                    vserver         => $vol_obj->vserver(),
                                    'role'          => 'data',
                                    'data-protocol' => ['cdmi'],
                                    'status-oper'   => 'up',
                                    'status-admin'  => 'up',}
                           );

    #Forming uri - e.g. http://10.53.19.131:8081/NS/cdmi/
    my $uri = $uri_scheme.q{://}.$interface_cs_obj->address().q{:}.
              $cdmi_port.$junc_path.q{/}.$name.q{/};

    $Log->comment('Creating container: '.$uri);
    my $http_req_obj = CR::CRHttpRequest->new();
    $http_req_obj->set_uri($uri);
    if ($body) {
        $http_req_obj->set_body($body);
    }
    my $container_obj = CR::CRContainerObject->new();
    my $response      = $container_obj->create($http_req_obj);

    #Check if creation succeeded
    if ($response->is_success()) {
        $Log->comment('Container created with response: '.
                      $response->decoded_content());
    } else {
        $Log->comment('**FAIL** Something went wrong: '.
                      $response->status_line());
        return;
    }

    return {
            container_obj => $container_obj,
            uri           => $response->base()->as_iri()
           };
}


###############################################################################
# Name        : delete_container
# Description : deletes given container
# Input       : Ref Hash with following keys (output of create_container)
#               container_obj Container object (returned by create_container)
#               uri           Uri of the container object
# Returns     : True  on success
#               False on failure
###############################################################################
sub delete_container {
    my ($params) = @_;

    my $c_obj = $params->{container_obj};
    my $uri = $params->{uri};

    $Log->comment('Deleting container: '.$uri);
    my $http_req_obj = CR::CRHttpRequest->new();
    $http_req_obj->set_uri($uri);
    my $response = $c_obj->delete($http_req_obj);

    #Check if deletion succeeded
    if ($response->is_success()) {
        $Log->comment('Container deleted with response: '.
                      $response->status_line());
        return 1;
    } else {
        $Log->comment('**FAIL** Something went wrong: '.
                      $response->status_line());
        return;
    }
}


###############################################################################
# Name        : create_named_object
# Description : Creates named object
# Input       : Hash with following keys
#               uri     Uri of the container (Required)
#               name    Name of the named object (Defaults to named<timestamp>)
# Returns     : Object of type CR::DataObject and uri on success
#               False on failure
###############################################################################
sub create_named_object {
    my (%params) = @_;
    my $uri = $params{uri};
    my $name = $params{name} || 'named'.time();

    $Log->comment('Creating named data object: '.$name);
    my $http_req_obj = CR::CRHttpRequest->new();
    $http_req_obj->set_uri($uri.$name);
    $http_req_obj->set_method('PUT');

    my $data_obj = CR::CRDataObject->new();
    my $response = $data_obj->create($http_req_obj);

    #Check if creation succeeded
    if ($response->is_success()) {
        $Log->comment('Named object created with response: '.
                      $response->decoded_content());
        return {
                data_obj => $data_obj,
                uri      => $response->base()->as_iri(),
               };
    } else {
        $Log->comment('**FAIL** Something went wrong: '.
                      $response->status_line());
        return;
    }
}


###############################################################################
# Name        : read_named_object
# Description : Reads a named object
# Input       : Ref Hash with following keys
#               uri             Uri of the container
#               read_over_nfs   Volume mount object
#                               To be provided when object is to be read over
#                               NFS (Currently not supported)
#               read_over_cdmi  True if named object is to be read over CDMI
#                               Currently defaults to 1
# Returns     : True    if read was successful
#               False   if read failed
###############################################################################
sub read_named_object {
    my ($params) = @_;

    #my $read_over_nfs = $params{read_over_nfs};
    my $read_over_cdmi = $params->{read_over_cdmi} || 1;
    my $uri = $params->{uri};
    my $read_status;

    if ($read_over_cdmi) {
        $Log->comment('Reading named data object over CDMI: '.$uri);
        $read_status = read_data_object(data_obj => $params->{data_obj},
                                        uri      => $params->{uri}) ;
    }

    #venkatc (CDMI Dev):
    #I will be disabling support for Named Object reads over NFS and CIFS in
    #the current sprint. Only read over CDMI will be supported. Support for
    #reading over NFS and CIFS for Named Objects will be added when the
    #File-Object duality implementation work eventually starts.
    #This requires work in the FOD engine, the Nblade CDMI server and the
    #policy engine. Only when all these three things fall into place will you
    #be able to properly read Named Objects over NFS/CIFS.

    #Following code can be enabled once the support to read over NFS is enabled
    #if ($read_over_nfs) {
    #    $Log->comment('Reading named data object over NFS');
    #    my $src_mount_obj = $read_over_nfs;
    #    my $client_obj = $src_mount_obj->command_interface();
    #    my $mount_point = $src_mount_obj->mount_point();
    #    my $junc_path = $1 if (($mount_point =~ /\/(\w+)$/));
    #    my $file_path = $1 if ($uri =~ /$junc_path(.*)/);
    #    my $response_obj;

    #    try {
    #        $response_obj =
    #            $client_obj->apiset()->ls(path => $mount_point.$file_path);
    #        $Log->comment('Named object was read with response: '.
    #                      $response_obj->get_processed_output());
    #        $read_status = 1;
    #    } catch NACL::APISet::Exceptions::CommandFailedException with {
    #        my $exception_obj = shift;
    #        $response_obj = $exception_obj->get_response_object();
    #        my $output = $response_obj->get_processed_output();
    #        $Log->comment('**FAIL** Could not read file over nfs: '.$output);
    #        $read_status = 0;
    #    };
    #}

    return $read_status;
}


###############################################################################
# Name        : delete_named_object
# Description : Deletes named object
# Input       : Ref Hash with following keys
#               data_obj   Named object to be deleted
#               uri        Uri of the container
# Returns     : See delete_data_object()
###############################################################################
sub delete_named_object {
    my ($params) = @_;

    $Log->comment('Deleting named data object: '.$params->{uri});
    return delete_data_object(data_obj => $params->{data_obj},
                              uri      => $params->{uri});
}


###############################################################################
# Name        : create_nameless_object
# Description : Creates nameless object
# Input       : Hash with following keys
#               vol_obj Volume object on which the data object is to be created
#               uri_scheme     Uri scheme (defaults to http)
#               cdmi_port      cdmi port (defaults to 8081)
# Returns     : Object of type CR::DataObject and uri on success
#               False on failure
###############################################################################
sub create_nameless_object {
    my (%params) = @_;

    my $vol_obj = $params{vol_obj};
    my $uri_scheme = $params{uri_scheme} || 'http';
    my $cdmi_port  = $params{cdmi_port} || 8081;
    my $node_obj = $vol_obj->command_interface();

    my $interface_cs_obj = NACL::CS::NetworkInterface->fetch(
                                command_interface => $node_obj,
                                filter            => {
                                    vserver         => $vol_obj->vserver(),
                                    'role'          => 'data',
                                    'data-protocol' => ['cdmi'],
                                    'status-oper'   => 'up',
                                    'status-admin'  => 'up',}
                           );

    my $uri = $uri_scheme.q{://}.$interface_cs_obj->address().q{:}.
              $cdmi_port.q{/cdmi_objectid/};

    $Log->comment('Creating nameless data object at '.$uri);
    my $http_req_obj = CR::CRHttpRequest->new();
    $http_req_obj->set_uri($uri);

    my $nameless_obj = CR::CRDataObject->new();
    my $response     = $nameless_obj->create($http_req_obj);

    #Check if creation succeeded
    if ($response->is_success()) {
        $Log->comment('Nameless object created with response: '.
                      $response->decoded_content());
        return {
                data_obj => $nameless_obj,
                uri      => $response->base()->as_iri(),
               };
    } else {
        $Log->comment('**FAIL** Something went wrong: '.
                      $response->status_line());
        return;
    }
}


###############################################################################
# Name        : read_nameless_object
# Description : Reads a named object
# Input       : Output of create_nameless_object
#               Ref to hash with following keys
#               uri             Uri of the container
#               data_obj        Data object to be read
# Returns     : True    if read was successful
#               False   if read failed
###############################################################################
sub read_nameless_object {
    my ($params) = @_;

    my $oid = $params->{data_obj}->get_object_id();

    $Log->comment('Reading nameless data object: '.$params->{uri}.$oid);
    return read_data_object(data_obj => $params->{data_obj},
                            uri      => $params->{uri}.$oid) ;
}


###############################################################################
# Name        : delete_nameless_object
# Description : Delete nameless object
# Input       : Output of create_nameless_object
#               Ref to hash with following keys
#               uri             Uri of the container
#               data_obj        Data object to be deleted
# Returns     : See delete_data_object()
###############################################################################
sub delete_nameless_object {
    my ($params) = @_;

    my $oid = $params->{data_obj}->get_object_id();

    $Log->comment('Deleting nameless data object: '.$params->{uri}.$oid);
    return delete_data_object(data_obj => $params->{data_obj},
                              uri      => $params->{uri}.$oid);
}


###############################################################################
# Name        : read_data_object
# Description : Reads a named or nameless object
#               Called by read_named_object and read_nameless_object
# Input       : Hash with following keys
#               uri             Uri of the container
#               data_obj        Data object to be read
# Returns     : True    if read was successful
#               False   if read failed
###############################################################################
sub read_data_object {
    my (%params) = @_;

    my $read_status;

    my $http_req_obj = CR::CRHttpRequest->new();
    $http_req_obj->set_uri($params{uri});
    my $response = $params{data_obj}->read($http_req_obj);

    #Check if read was successful
    if ($response->is_success()) {
        $Log->comment('Data object was read with response: '.
                      $response->decoded_content());
        $read_status = 1;
    } else {
        $Log->comment('**FAIL** Something went wrong: '.
                      $response->status_line());
    }

    return $read_status;
}


###############################################################################
# Name        : delete_data_object
# Description : Deletes named or nameless data object
#               Used by delete_named_object and delete_nameless_object
# Input       : Hash with following keys
#               data_obj   Data object to be deleted
#               uri        Uri of the data object
# Returns     : True  on success
#               False on failure
###############################################################################
sub delete_data_object {
    my (%params) = @_;
    my $uri = $params{uri};
    my $data_obj = $params{data_obj};

    my $http_req_obj = CR::CRHttpRequest->new();
    $http_req_obj->set_uri($uri);
    my $response = $data_obj->delete($http_req_obj);

    #Check if the data object was deleted
    if ($response->is_success()) {
        $Log->comment('Deleted data object with response: '.
                      $response->status_line());
        return 1;
    } else {
        $Log->comment('**FAIL** Something went wrong: '.
                      $response->status_line());
        return;
    }
}


1;
