#
# $Id: //depot/prod/test/Rsierranevada/lib/CR/ContentRepository.pm#1 $
#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Content Repository Class
## @author ngok@netapp.com
## @status shared
## @pod here

package CR::ContentRepository;

use strict;
use warnings;

use Tharn qw($Log);
use NACL::APISet::Exceptions::ResponseException qw(:try);
use CR::Exceptions::ArgumentError qw(:try);
use CR::Exceptions::NoAggregate qw(:try);
use CR::Exceptions::NoVServer qw(:try);
use CR::Exceptions::NoRootVolume qw(:try);
use CR::Exceptions::VolumeNotFound qw(:try);
use CR::Exceptions::ResidueVolumes qw(:try);
use CR::Exceptions::InvalidNetmask qw(:try);
use Params::Validate qw(:all);
use NACL::APISet; # for valid_params()
use CR::UniqueName;
use Data::Dumper;

use constant PARTITION_BITS => 4;  # default used by setup_bootarg()
use constant TREE_WIDTH_BITS => 4; # default used by setup_bootarg()
use constant DEFAULT_AGGR_DISKCOUNT => 3; # used by _create_aggregate()
my $extended_timeout = 240; # in seconds

=head1 NAME

ContentRepository - The class to encapsulate methods and data of Content Repository

=head1 SYNOPSIS

=begin text

 use ContentRepository;

 my $cr = ContentRepository->new(vserver => 'cr_vserver', 
                                 command_interface => 'node1');
 $cr->load_image();
 $cr->set_flag(create_volumes => 1);
 $cr->setup_bootarg(partitionbits => 6, 
                    tree_width_bits => 6, 
                    ols_enabled => 1,
                    reboot => 0); # need to reboot to take effect

 $cr->ols_volumes(volumes => 8);
 $cr->mds_volumes(volumes => [ 'mds1', 'mds2', 'mds3', 'mds4' ]);
 $cr->obj_volumes(volumes => ['objvol1', 'objvol2'], 
        'object-volume-group' => 'archive_ovg');
 $cr->ns_volumes(volumes => 'ns1');

 $cr->setup_dataport('home-node' => 'node1',
		    'home-port' => 'e0d',
		    address => '172.17.121.16');

 $cr->setup_MDS();
 $cr->setup_NFS( domain => 'netapp.com',
                servers => [ '10.56.32.32', '10.56.32.33' ] );

 my $bool = $cr->is_ready();

 $cr->delete();

=end text

=head1 AUTHOR

Sidney Ho (ngok@netapp.com)

=head1 DESCRIPTION

This is a NACL AppBuilder to support test program that uses Content Repository.

This initial release only implements setup of the Content Repository.
Furture releases will support operations on the Content Repository from NAS or REST clients.

Parameters such as volume can take a single argument or multiple arguments.
If it is multi-valued, it must be an arrayref of several arguments.

=head1 ROUTINES

=cut

############################################################
#     constructors
############################################################

=head2 create

Constructor to create a fully functional instance of Content Repository.
It sets the create_volumes flag to 1.
It expects a functional cluster with its nodes.
Then it goes through these steps:

=over 4

=item 1. Create a CR-enabled Vserver using largest aggregate to hold the root volume.

=item 2. Create a Repository and associate the Vserver to it.

=item 3. Create on each node 2 OLS volumes, 2 MDS volumes and 1 OBJ volume.

=item 4. Create 1 NS volume on the largest aggregate.

=back

Valid parameters are

=over 4

=item *
command_interface

This is one of the node names of the cluster that contains the Repository (mandatory).
This is usually the FILER of a HostRec which is part of the cluster..
A connection will be opened using this node to communicate..

=item *
vserver

The name of the Vserver to contain the Repository. (optional)
If vserver name is not given, a name will be generated automatically.
Use the getter method 'vserver()' to obtain the name.

=back 

=cut

sub create {
    $Log->enter();
    my ($proto, %args) = @_;
    my $class = ref($proto) || $proto;

    %args = NACL::APISet::validate_params($class,
		     params => \%args,
		     spec => {
		     command_interface => { type => SCALAR, optional => 0 },
		     vserver => { type => SCALAR, optional => 1 },
		 },
	     );
    if (!defined($args{vserver})) {
	$args{vserver} = unique_name('vs');
    }

    my $self = $proto->new(%args);
    my $admin_cli = $self->{admin_cli};
    $self->set_flag(create_volumes => 1, new_dataports => 1);

    $self->mds_volumes(volumes => 2*@{$self->{nodes}});
    $self->ols_volumes(volumes => 2*@{$self->{nodes}});
    $self->obj_volumes(volumes => scalar(@{$self->{nodes}}), 
		     'object-volume-group' => unique_name('ovg'));
    $self->ns_volumes(volumes => 1);

    # Setup any data ports from nate_host info
    $self->setup_dataports_from_nate_host();

    # MDS specific setup
    $self->setup_MDS();

    #todo: find from where to obtain the domain and servers
    $self->setup_NFS(domain => 'netapp.com',
		     servers => ['10.56.32.32', '10.56.32.33']);

    $Log->exit();
    return $self;
}

=head2 new

Constructor to create an uninitialized instance of Content Repository.
It needs a Vserver name to be associated with the Content Repository.
It returns an unintialized ContentRepository object 
so that setup functions can be performed on it.

=over 4

=item *
vserver

The name of the Vserver to contain the Repository.
If the vserver already exists, it will be reused.
If this parameter is not specified, a unique Vserver name will be generated.

=item *
command_interface

This is one of the node name of the Vserver (mandatory).
This is usually the FILER of a HostRec which is part of the cluster
corresponding to the Vserver.
A connection will be opened using this node to communicate with the node.

=back

Pre-conditions:

=over 4

=item 1. A functional cluster with nodes, disks and physical network set up.

=back

Post-conditions:

=over 4

=item 1.
An instantiated object.

=item 2.
Got all nodes in cluster.

=item 3.
Got all system shell to each node.

=item 4.
Added Repository license.

=item 5.
1 or more non-root aggregate.
If aggregates are created, there will be 1 per node.

=item 6.
Created a CR Vserver if not already exists.

=item 7.
Root volume of Vserver is online

=item 8.
Created or cleared to an empty Repository.

=back

=cut

sub new {
    $Log->enter();
    my ($proto, %args) = @_;
    my $class = ref($proto) || $proto;

    %args = NACL::APISet::validate_params($class,
		     params => \%args,
		     spec => {
		     command_interface => { type => SCALAR, optional => 0 },
		     vserver => { type => SCALAR, optional => 1 },
		     'repository-role' => { type => SCALAR, optional => 1 },
		 },
	     );

    my $self = $class->_connect(%args);
    my $admin_cli = $self->{admin_cli};

    # Add Repository license
#    $admin_cli->system_license_add('license-code' => 'BIUWGKPJYBFDAA');

    # Add NFS license
    $admin_cli->system_license_add('license-code' => 'TJFAEKPJYBFDAA');

    # Check if Vserver already exist
    my @parsed_output = @{$admin_cli->vserver_show()->get_parsed_output()};
    my $vserver_exists = grep $args{vserver} eq $_->{vserver}, @parsed_output;

    # Check and create Vserver
    if ($vserver_exists) {
	# reuse Vserver (IMPORTANT: it may not be a CR Vserver)
	@parsed_output = @{$admin_cli->repository_volume_show()->get_parsed_output()};
	if (grep $args{vserver} eq $_->{vserver}, @parsed_output) {
	    # delete repository and Vserver
	    $self->delete();
	}
	else { # not recorded in repository as CR Vserver
	    $self->_delete_vserver();
	    $self->_create_vserver();
	}

	if (my @residue = $self->_residue_volumes()) {
	    CR::Exceptions::ResidueVolumes->throw("Previous volumes ".join(",",@residue)." remained in repository for vserver ".$self->{vserver});
	}

	try {
	    # Get rootvolume must perform after residue volume check
	    # otherwise the shown volume may be in the process of being deleted
	    $self->{rootvolume} = $self->_get_rootvolume(); 
	    $admin_cli->volume_online(volume => $self->{rootvolume}, vserver => $self->{vserver});
	}
	catch CR::Exceptions::NoRootVolume with {
	    # workaround for partially deleted VServer
	    $self->_create_rootvolume(); # recreate rootvolume for now
	};
    }
    else {
	$self->_create_vserver(instantiating => 1);
    }

    # Create repository
    if (defined($args{'repository-role'})) {
	$admin_cli->repository_volume_create(vserver => $self->{vserver},
			       'repository-role' => $args{'repository-role'});
    }
    else {
	$admin_cli->repository_volume_create(vserver => $self->{vserver});
    }

    # Workaround: snapshot may hang CR volume deletion
    $admin_cli->vserver_modify(vserver => $self->{vserver},
			       'snapshot-policy' => 'none');

    $Log->exit();
    return $self;
}

sub _create_vserver {
    my ($self, %args) = @_;
    my $admin_cli = $self->{admin_cli};

    # Check Aggregate
    my @aggregates = $self->_get_aggregates(%args);
    
    # create vserver in the largest aggregate
    $self->{rootvolume} = 'root_' . $self->{vserver};
    $admin_cli->vserver_create(vserver => $self->{vserver},
			       aggregate => $aggregates[0],
			       rootvolume => $self->{rootvolume},
			       'ns-switch' => 'file',
			       'rootvolume-security-style' => 'mixed');
    # make rootvolume mode 0777
    $admin_cli->volume_modify(volume => $self->{rootvolume},
			      vserver => $self->{vserver},
			      'unix-permissions' => '---rwxrwxrwx');
}

=head2 connect

Constructor to connect to a Vserver intended for Repository operations.
It requires the Vserver name
and a node name of the cluster for making the connection.

=over 4

=item *
vserver

The name of the Vserver of the Repository. (mandatory)

=item *
command_interface

This is one of the node name of the Vserver (mandatory).
This is usually the FILER of a HostRec which is part of the cluster
corresponding to the Vserver.
A connection will be opened using this node to communicate with the node.

=back

Post-conditions:

=over 4

=item 1.
Instantiated object of an existing Content Repository Vserver.

=item 2.
Got all nodes in cluster.

=item 3.
Got all system shell to each node.

=item 4.
Root volume of Vserver is online

=back

=cut

sub connect {
    $Log->enter();

    my ($proto, %args) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->_connect(%args);
    my $admin_cli = $self->{admin_cli};

    # Make sure root volume is online
    $self->{rootvolume} = $self->_get_rootvolume();
    $admin_cli->volume_online(volume => $self->{rootvolume}, vserver => $self->{vserver});

    # Do we need to check license? seems irrelevant

    $Log->exit();
    return $self;
}

#
# _connect - instantiate object
#
# Fields defined in the object
#   1. vserver           - vserver name
#   2. command_interface - connection node name
#   3. admin_cli         - admin cli connection to command_interface above
#   4. nodes             - nodes in cluster
#   5. system_shells     - shell connections for each node in cluster
#   6. hosts             - hostRec of each node
#
sub _connect {
    $Log->enter();

    my ($proto, %args) = @_;
    my $class = ref($proto) || $proto;

    my $self = NACL::APISet::validate_params($class,
		     params => \%args,
		     spec => {
		     vserver => { type => SCALAR, optional => 0 },
		     command_interface => { type => SCALAR, optional => 0 },
		 },
	     );
    my $host = Tharn::host($self->{command_interface});

    # Open a CLI connection
    my $admin_cli = NACL::APISet->new(hostobj => $host,
				     category => 'Node',
			  	    interface => 'CLI',
				          set => 'CMode');

    $self->{admin_cli} = $admin_cli;
    bless $self, $class;

    # Find all the nodes of cluster
    my @parsed_output = @{$admin_cli->system_node_show()->get_parsed_output()};
    my @nodes = map $_->{node}, @parsed_output;
    $self->{nodes} = \@nodes;

    # Open SystemShell connections
    my @system_shells;
    my @hosts;
    foreach my $node (@{$self->{nodes}}) {
	my $host = Tharn::host($node);
	my $shell = 
	    NACL::APISet->new(hostobj => $host,
			      category => 'Node',
			      interface => 'CLI',
			      set => 'Systemshell',
                  version_info => { "cdefs_from_build" => 0 },  
			      shell_type => 'CMode',
			      context_definition => {username => 'diag', 
						     password => 'netapp1!'},
			      );
	push(@hosts, $host);
	push(@system_shells, $shell);
    }
    $self->{system_shells} = \@system_shells;
    $self->{hosts} = \@hosts;

    $Log->exit();
    return $self;
}

############################################################
#     volume and assignment
############################################################

=head2 volumes

Create one or more volumes before assigning any role.
(NOTE: This is not implemented)

Valid parameters are

=over 4

=item *
volumes

The names of the volumes. It accepts an ARRAY or a SCALAR.

=back

=cut

sub volumes {
}

=head2 ols_volumes

Add object_locator volumes.
This function can only be called once during initialization of the Repository.
It will locate the volumes in one of the Vserver nodes.
If it is not found, it will throw the VolumeNotFound exception
unless create_volumes is enabled.

Valid parameters are

=over 4

=item *
volumes

The names of the volumes

=item *
aggreggate

The names of the aggregates each corresponding to a name in the volume list. 
This parameter can be an ARRAY or a SCALAR. 
If there are more volumes than aggregates,
the aggregate list will be reused from the beginning repeatingly
until all volumes are associated with aggregates.

=item *
size

The size of the corresponding volume in the volume list. 
It can be an ARRAY or a SCALAR. 
If there are more volumes than sizes,
the size list will be resued repeatingly.

=back

=cut

sub ols_volumes {
    $Log->enter();
    my @volumes = _assign_volumes(@_, role => 'object_locator');
    $Log->exit();
    return @volumes;
}

=head2 mds_volumes

Add one or more metadata volumes.
This function can only be called once during initialization of repository.
It will locate the volumes in one of the Vserver nodes.
If it is not found, it will throw the VolumeNotFound exception
unless create_volumes is enabled.

Valid parameters are

=over 4

=item *
volumes

The names of the volumes.

=item *
aggreggate

The aggregates of the corresponding volume in the volume list.
This parameter can be an ARRAY or a SCALAR. 
If there are more volumes than aggregates,
the aggregate list will be reused repeatingly.

=item *
size

The size of the corresponding volume in the volume list. 
It can be an ARRAY or a SCALAR. 
If there are more volumes than sizes,
the size list will be resued repeatingly.

=back

=cut

sub mds_volumes {
    $Log->enter();
    my ($self, %args) =	@_;

    if ($self->{no_mls_volumes}) {
	# special flag for mls negative test case
	$Log->exit();
	return undef;
    }

   my @volumes = _assign_volumes(@_, role => 'metadata');
    $Log->exit();
    return @volumes;
}

=head2 obj_volumes

Add one or more object volumes.
This function may be called several times with one or more volume names.
It will locate the volumes in one of the Vserver nodes.
If it is not found, it will throw the VolumeNotFound exception
unless create_volumes is enabled.

Valid parameters are

=over 4

=item *
volumes

The names of the volumes.

=item *
object-volume-group

The name of the OVG to which these volumes belong.
This is a mandatory field.
This parameter must be a SCALAR.
An exception will be thrown if it is not given.
(NOT IMPLEMENTED: If default-object-volume-group is defined,
the default-object-volume-group will be used and no exeception will be thrown.)

=item *
aggreggate

The aggregate of the corresponding volume in the volume list.
This parameter can be an ARRAY or a SCALAR. 
If there are more volumes than aggregates,
the aggregate list will be reused repeatingly.

=item *
size

The size of the corresponding volume in the volume list.
It can be an ARRAY or a SCALAR. 
If there are more volumes than sizes,
the size list will be resued repeatingly.

=back

=cut

sub obj_volumes {
    $Log->enter();
    my ($self, %args) =	@_;
    %args = NACL::APISet::validate_params($self,
		 params => \%args,
		 spec => {
		     volumes => { type => SCALAR|ARRAYREF, optional => 0 },
		     'object-volume-group' => { type => SCALAR, optional => 0 },
		     aggregate => { type => SCALAR|ARRAYREF, optional => 1 },
		     size => { type => SCALAR|ARRAYREF, optional => 1 },
		 },
	     );

    my @volumes = $self->_assign_volumes(%args, role => 'object');
    $Log->exit();
    return @volumes;
}

=head2 ns_volumes

Add a namespace volume.
It will locate the volume in one of the Vserver nodes.
If it is not found, it will throw the VolumeNotFound exception
unless create_volumes is enabled.

Valid parameters are

=over 4

=item *
volume

The name of the volume.
Only 1 namespace volume is currently supported by CR.
But this method does not enforce this restriction.

=back

=cut

sub ns_volumes {
    $Log->enter();
    my @volumes = _assign_volumes(@_, role => 'namespace');
    $Log->exit();
    return @volumes;
}

# _assign_volumes - common routine for adding and assigning role to volumes
my %abbr = (
	    metadata => 'mds',
	    namespace => 'ns',
	    object => 'obj',
	    object_locator => 'ols',
	    );
sub _assign_volumes {
    my ($self, %args) =	@_;
    %args = NACL::APISet::validate_params($self,
		 params => \%args,
		 spec => {
		     volumes => { type => SCALAR|ARRAYREF, optional => 0 },
		     role => { type => SCALAR, optional => 0 },
		     'type' => { type => SCALAR, optional => 1 },
		     aggregate => { type => SCALAR|ARRAYREF, optional => 1 },
		     size => { type => SCALAR|ARRAYREF, optional => 1 },
		     'object-volume-group' => { type => SCALAR, optional => 1 },
		     'junction-path' => { type => SCALAR, optional => 1 },
		 },
	     );

    # Put volume names into an array
    my $admin_cli = $self->{admin_cli};
    my @volumes;
    if ('ARRAY' eq ref($args{volumes})) {
	@volumes = @{$args{volumes}};
    }
    elsif (!ref($args{volumes})) {
	if ($args{volumes} =~ /^\d+$/) {
	    # volumes is digit - generate that many volumes automatically
	    for (my $i = 0; $i < $args{volumes}; $i++) {
		$volumes[$i] = unique_name($abbr{$args{role}});
	    }
	}
	else {
	    # there is only 1 volume
	    @volumes = ($args{volumes});
	}
    }
    else {
	# illegal ref, throw an exception
	CR::Exceptions::ArgumentError->throw('volumes arg: Unsupported ref type '.ref($args{volumes}));
    }

    # Get aggregates to create volumes
    my @aggregates; 
    if (!defined($args{aggregate})) {
	@aggregates = $self->_get_aggregates(volumes => \@volumes);
    }
    elsif (!ref($args{aggregate})) {
	@aggregates = ($args{aggregate});
    }
    elsif ('ARRAY' eq ref($args{aggregate})) {
	@aggregates = @{$args{aggregate}};
    }
    else {
	# illegal ref, throw an exception
	CR::Exceptions::ArgumentError->throw('aggregate arg: Unsupported ref type '.ref($args{aggregate}));
    }

    # Get sizes for volumes
    my @sizes;
    if (!defined($args{size})) {
	@sizes = ('80MB'); # changed from 20MB then 40MB
    }
    elsif (!ref($args{size})) {
	@sizes = ($args{size});
    }
    elsif ('ARRAY' eq ref($args{size})) {
	@sizes = @{$args{size}};
    }
    else {
	# illegal ref, throw an exception
	CR::Exceptions::ArgumentError->throw('size arg: Unsupported ref type '.ref($args{size}));
    }

    # Get existing volumes
    my @exist_vols = $self->_get_volumes();

    # Assign role to volume; creating volume if necessary
    my $aggregate_index = 0;
    my $size_index = 0;
    foreach my $volume (@volumes) {

	# unmount or create volume before assigning the role
	if (grep $_ eq $volume, @exist_vols) {
	    if ('namespace' eq $args{role}) {
		#bug: should check to see if junction-path is there
	    }
	    # perform an unmount if volume exists
	    $admin_cli->volume_unmount(volume => $volume,
				  vserver => $self->{vserver});
	}
	elsif (!$self->{create_volumes}) {
	    CR::Exceptions::VolumeNotFound->throw('Volume '.$volume.' does not exist');
	}
	else {
	    # prepare parameters before creating the volume
	    my %vol_args = (volume => $volume, 
			    size => $sizes[$size_index],
			    aggregate => $aggregates[$aggregate_index],
			    vserver => $self->{vserver});
	    if (defined($args{type}) && 'DP' eq $args{type}) {
		# no unix-permission, no namespace for DP
		$vol_args{type} = $args{type};
	    }
	    else {
		if (exists($args{type}) && 'RW' ne $args{type}) {
		    CR::Exceptions::ArgumentError->throw("Unsupported volume type " . $args{type});
		}
		$vol_args{'unix-permissions'} = '---rwxrwxrwx';
		if ('namespace' eq $args{role}) {
		    if (exists($args{'junction-path'})) {
			$vol_args{'junction-path'} = $args{'junction-path'};
		    }
		    else {
			$vol_args{'junction-path'} = "/$volume"; # use the volume name if not given
		    }
		}
	    }

	    $admin_cli->volume_create(%vol_args); # create the volume

	    # allow max number of inodes for object volumes
	    if ('object' eq $args{role}) {
		my $size = $sizes[$size_index];
		my $files;
		use constant FACTOR => 0.95;
		if ($size =~ /^([\d\.]+)GB$/) {
		    $files = int($1*250000*FACTOR); # 1 inode per 4K
		}
		elsif ($size =~ /^([\d\.]+)MB$/) {
		    $files = int($1*250*FACTOR); # 1 inode per 4K
		}
		elsif ($size =~ /^([\d\.]+)KB$/) {
		    $files = int($1*.25*FACTOR); # 1 inode per 4K
		}
		
		if ($files) {
		    $admin_cli->volume_modify(volume => $volume,
					      files => $files,
					      vserver => $self->{vserver});
		}
	    }

	    $aggregate_index = ++$aggregate_index % @aggregates;
	    $size_index = ++$size_index % @sizes;
	}
    }
	
    # do assignment
    if ('object' eq $args{role}) {
	$admin_cli->repository_volume_assign(role => $args{role},
					volumes => join(',', @volumes),
					'obj-volume-group' => $args{'object-volume-group'},
					vserver => $self->{vserver});
    }
    else {
	$admin_cli->repository_volume_assign(role => $args{role},
					     volumes => join(',', @volumes),
					     'connectrec-timeout' => $extended_timeout,
					     vserver => $self->{vserver});
    }
    return @volumes;
}

# create the root volume of the vserver and use the name "root_$vserver"
sub _create_rootvolume {
    my $self = shift;
    my $admin_cli = $self->{admin_cli};
    my @aggregates = $self->_get_aggregates();

    if (!defined($self->{rootvolume})) {
	$self->{rootvolume} = 'root_' . $self->{vserver};
    }
    $admin_cli->volume_create(volume => $self->{rootvolume},
			      aggregate => $aggregates[0],
			      vsroot => 'true',
			      'security-style' => 'mixed',
			      'unix-permissions' => '---rwxrwxrwx',
			      vserver => $self->{vserver});
}

# get the root volume of the vserver
sub _get_rootvolume {
    my $self = shift;
    my $admin_cli = $self->{admin_cli};
    my $rootvol;

    try {
	$rootvol = ${$admin_cli->vserver_show(vserver => $self->{vserver})->get_parsed_output()}[0]->{rootvolume};
    }
    otherwise {
	CR::Exceptions::NoVServer->throw("Can't find VServer " . $self->{vserver});
      };

    if ('-' eq $rootvol) {
	CR::Exceptions::NoRootVolume->throw("No root volume found for vserver " . $self->{vserver});;
      }
    return $rootvol;
}


# get non-root volumes
sub _get_volumes {
    my $self = shift;
    my $admin_cli = $self->{admin_cli};
    my @parsed_output = @{$admin_cli->repository_volume_show_roles()->get_parsed_output()};
    @parsed_output = grep $_->{vserver} eq $self->{vserver}, @parsed_output;
    return map $_->{volume}, @parsed_output;
}

#
# _get_aggregates - return list of aggregates in cluster ordered by available space
#
sub _get_aggregates {
    my ($self, %args) = @_;
    my $admin_cli = $self->{admin_cli};
    my @parsed_output = @{$admin_cli->storage_aggregate_show()->get_parsed_output()};    
    if (@parsed_output < 2 * @{$self->{nodes}}) {
#	if (!$self->{create_volumes} && !$args{instantiating}) {
#	    CR::Exceptions::NoAggregate->throw('No aggregates in cluster');
#	}
	return $self->_create_aggregates();
    }

    #todo: filter out root aggregate of each node
    @parsed_output = sort {$b->{availsize} <=> $a->{availsize}} @parsed_output;

    if (defined($args{volumes})) {
	my @node_aggregates;
	my $i = 0;
	# get aggregate from each node
	foreach my $node (@{$self->{nodes}}) {
	    ($node_aggregates[$i++]) = grep $_->{nodes} =~ /$node/, @parsed_output;
	}
	# elongate the list
	while (@{$args{volumes}} > @node_aggregates) {
	    push(@node_aggregates, @node_aggregates);
	}
	return map $_->{aggregate}, @node_aggregates;
    }
    else {
	return map $_->{aggregate}, @parsed_output;
    }

}

# _create_aggregates - create 1 non-root aggregate for each node
#
#    assumed there is 1 root aggregate already exists for each node
sub _create_aggregates {
    my $self = shift;
    my $admin_cli = $self->{admin_cli};
    my @aggregates;

    foreach my $node (@{$self->{nodes}}) {
	if(1 >= @{$admin_cli->storage_aggregate_show(nodes => $node)->get_parsed_output()}) {
	    my @parsed_output = @{$admin_cli->storage_disk_show(owner => $node)->get_parsed_output()};
	    my @unused = grep $_->{state} eq 'spare', @parsed_output;
	    
	    # check diskcount
	    if (DEFAULT_AGGR_DISKCOUNT > @unused) {
		CR::Exceptions::NoAggregate->throw('Insufficient disks for creating aggregate in node ', $node);
	      }
	    
	    # create aggregate
	    my $aggregate = unique_name('aggr');
	    push(@aggregates, $aggregate);
	    $admin_cli->storage_aggregate_create(aggregate => $aggregate,
						 diskcount => DEFAULT_AGGR_DISKCOUNT, 
						 nodes => $node);
	}
    }
    return @aggregates;
}

############################################################
#     setters and getters
############################################################

=head2 set_flag

Set an internal flag that changes the behavior of the Repository.

=over 4

=item *
create_volumes

Automatically create volumes that do not exist.
Default is 0 (false) and an exception will be thrown on non-existing volumes.
Set this flag to 1 to create the volumes without throwing the exception.

=item *
new_dataports

During dataports creation, any existing dataports with the same IP address will be refreshed with the new specification.
If this flag is not set, setup_dataport() will preserve existing dataports with the same IP addresses.
Default is 0 (i.e. keeping existing dataports.)

=item *
lif_auto_revert

Indicate whether LIF created will have the auto-revert property or not.
Default is 0 (false).
Setting this flag to 1 (true) if you want all the data LIFs to
automatically revert back to their original node after recovery from 
a LIF failover.

=back

=cut

sub set_flag {
    $Log->enter();
    my ($self, %args) =	@_;
    %args = NACL::APISet::validate_params($self,
		     params => \%args,
		     spec => {
		     create_volumes => { type => SCALAR, optional => 1 },
		     new_dataports => { type => SCALAR, optional => 1 },
		     lif_auto_revert => { type => SCALAR, optional => 1 },
		 },
	     );
    $self->{$_} = $args{$_} foreach (keys(%args));
    $Log->debug("set_flag args: %args\n");
    $Log->exit();
}

=head2 vserver

Return the name of the vserver of this Content Repository object.

=cut

sub vserver {
    $Log->enter();
    my $self = shift;
    $Log->exit();
    return $self->{vserver};
}

=head2 is_ready

Return a boolean whether setup has finished and the Content Repository is ready to be used.

=cut

sub is_ready {
    $Log->enter();
    my $self = shift;
    my $admin_cli = $self->{admin_cli};
    my $response = $admin_cli->repository_volume_show();
    $Log->exit();
    if ($response =~ /ready/) {
	return 1;
    }
    else {
	return 0;
    }
}

############################################################
#     dataport and network
############################################################

=head2 setup_dataports_from_nate_host

Setup the dataports from information in the nate_host file.

=cut

sub setup_dataports_from_nate_host {
    $Log->enter();
    my $self = shift;

    for (my $i = 0; $i <= $#{$self->{nodes}}; $i++) {
	my $node = ${$self->{nodes}}[$i];
	my $host = ${$self->{hosts}}[$i];
	my %args;

	# a data port
	if (defined($host->DATA_PORT) && defined($host->DATA_IP)) {
	    %args = ('home-node' => $node,
		     'home-port' => $host->DATA_PORT,
		     address => $host->DATA_IP);
	    $args{netmask} = $host->DATA_MASK if $host->DATA_MASK;
	    $args{gateway} = $host->DATA_GATEWAY if $host->DATA_GATEWAY;
	    $self->setup_dataport(%args);
	}

	# data port 1
	if (defined($host->DATA_PORT1) && defined($host->DATA_IP1)) {
	    %args = ('home-node' => $node,
		     'home-port' => $host->DATA_PORT1,
		     address => $host->DATA_IP1);
	    $args{netmask} = $host->DATA_MASK1 if $host->DATA_MASK1;
	    $args{gateway} = $host->DATA_GATEWAY1 if $host->DATA_GATEWAY1;
	    $self->setup_dataport(%args);
	}

	# data port 2
	if (defined($host->DATA_PORT2) && defined($host->DATA_IP2)) {
	    %args = ('home-node' => $node,
		     'home-port' => $host->DATA_PORT2,
		     address => $host->DATA_IP2);
	    $args{netmask} = $host->DATA_MASK2 if $host->DATA_MASK2;
	    $args{gateway} = $host->DATA_GATEWAY2 if $host->DATA_GATEWAY2;
	    $self->setup_dataport(%args);
	}
    }
    $Log->exit();
}

=head2 setup_dataport

Set up the data port on the specified node with its IP address,
netmask, and gateway.

=over 4

=item *
home-node

The node on which the data port resides. (mandatory)
This is the name of the HostRec.

=item *
home-port

The port id (mandatory)
(such as 'e0d')

=item *
address

The IP address of the port. (mandatory)

=item *
netmask

The netmask of the port. (optional)
If it is not given, 
it is assumed 255.255.255.0.

=item *
gateway

The gateway of the subnet to which the IP address belongs. (optional)
If it is not given,
it is assumed to be host id 1 of the subnet.
For example, IP address 172.17.121.16 netmask 255.255.255.0
will yield gateway address 172.17.121.1.

=back

=cut

sub setup_dataport {
    $Log->enter();
    my ($self, %args) =	@_;
    my $admin_cli = $self->{admin_cli};
    %args = NACL::APISet::validate_params($self,
		 params => \%args,
		 spec => {
		     lif => { type => SCALAR, optional => 1 },
		     'home-node' => { type => SCALAR, optional => 0 },
		     'home-port' => { type => SCALAR, optional => 0 },
		     address => { type => SCALAR, optional => 0 },
		     netmask => { type => SCALAR, optional => 1 },
		     gateway => { type => SCALAR, optional => 1 },
		 },
	     );

    # check if address is already used in a dataport
    try {
	my $port = shift @{$admin_cli->network_interface_show(address => $args{address})->get_parsed_output()};
	if ($self->{new_dataports}) {
	    # delete existing dataport having the identical address
	    $admin_cli->network_interface_delete(lif => $port->{lif}, 
						 vserver => $port->{vserver});
	    # proceed to create dataport
	}
	else {
	    return; # keep existing dataport
	}
    }
    catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
	$Log->comment("NACL::APISet::Exceptions::NoMatchingEntriesException caught and ignored.");
	# proceed to create dataport
    };

    if (!defined($args{netmask})) {
	$args{netmask} = '255.255.255.0';
    }
    if (!defined($args{lif})) {
	$args{lif} = unique_name('lif');
    }

    my $gateway;
    if (!defined($args{gateway})) {
	$gateway = _gateway($args{address}, $args{netmask});
    }
    else {
	$gateway = $args{gateway};
        # gateway is not a valid argument for network_interface_create
        # remove it before passing %args as arguments
	delete $args{gateway}; 
    }

    if (defined($self->{lif_auto_revert})) {
	$args{'auto-revert'} = $self->{lif_auto_revert} ? 'true' : 'false';
    }

    $admin_cli->network_interface_create(%args, role => 'data', 
		      server => $self->{vserver}); # No exception if duplicate

    # setup routing group route
    my $routing_group = _data_routing_group($args{address}, $args{netmask});
    try {
	$admin_cli->network_routing_groups_route_create(server => $self->{vserver},
							'routing-group' => $routing_group,
							gateway => $gateway);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
	my $ex = shift;
	if ($ex->text() !~ /duplicate entry/) {
	    $ex->throw();
	}
	$Log->comment("NACL::APISet::Exceptions::ResponseException '".
		      $ex->text()."' caught and ignored.");
    };
    $Log->exit();
}

my @netmask_table = ("0.0.0.0",
		     "128.0.0.0",
		     "192.0.0.0",
		     "224.0.0.0",
		     "240.0.0.0",
		     "248.0.0.0",
		     "252.0.0.0",
		     "254.0.0.0",
		     "255.0.0.0",
		     "255.128.0.0",
		     "255.192.0.0",
		     "255.224.0.0",
		     "255.240.0.0",
		     "255.248.0.0",
		     "255.252.0.0",
		     "255.254.0.0",
		     "255.255.0.0",
		     "255.255.128.0",
		     "255.255.192.0",
		     "255.255.224.0",
		     "255.255.240.0",
		     "255.255.248.0",
		     "255.255.252.0",
		     "255.255.254.0",
		     "255.255.255.0",
		     "255.255.255.128",
		     "255.255.255.192",
		     "255.255.255.224",
		     "255.255.255.240",
		     "255.255.255.248",
		     "255.255.255.252",
		     "255.255.255.254",
		     "255.255.255.255");
sub _netmask_to_bitmask {
    my $netmask = shift;
    for (my $i = 0; $i < $#netmask_table; $i++) {
	if ($netmask eq $netmask_table[$i]) {
	    return $i;
	}
    }
    CR::Exceptions::InvalidNetmask->throw("Netmask $netmask is invalid");
}

sub _network_id {
    my ($address, $netmask) = @_;
    my @address = split '\.', $address;
    my @netmask = split '\.', $netmask;
    my @result;
    for (my $i = 0; $i < 4; $i++) {
	$result[$i] = ($address[$i]+0) & ($netmask[$i]+0); # Perl bug workaround
    }
    return @result;
}

sub _data_routing_group {
    my ($address, $netmask) = @_;
    my @network = _network_id($address, $netmask);
    my $bitmask = _netmask_to_bitmask($netmask);
    return "d".join(".", @network)."/$bitmask";
}

sub _gateway {
    my ($address, $netmask) = @_;
    my @network = _network_id($address, $netmask);
    $network[3] |= 1;
    return join(".", @network);
}

=head2 setup_MDS

Setup firewall for pre-defined mdsd port.
(Per burt441153)

=cut

sub setup_MDS {
    $Log->enter();
    my ($self, %args) =	@_;
    my $admin_cli = $self->{admin_cli};
    try {
	$admin_cli->system_services_firewall_policy_service_create(service => "mdsd", protocol => "tcp/udp", port => 962);
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
	my $ex = shift;
	if ($ex->text() !~ /duplicate entry/) {
	    $ex->throw();
	}
	$Log->comment("NACL::APISet::Exceptions::CommandFailedException '".
		      $ex->text()."' caught and ignored.");
    };
    $Log->exit();
}

=head2 setup_NFS

Set the user changeable NFS related parameters
and set up NIS on it.

=over 4

=item *
domain

NIS domain name to where the CR Vserver belongs.

=item *
servers

NIS servers of the domain.

=back


=cut

sub setup_NFS {
    $Log->enter();
    my ($self, %args) =	@_;
    my $admin_cli = $self->{admin_cli};
    %args = NACL::APISet::validate_params($self,
		 params => \%args,
		 spec => {
		     servers => { type => SCALAR|ARRAYREF, optional => 0 },
		     domain => { type => SCALAR, optional => 0 },
		 },
	     );

    $self->_clear_nis_domain();

    if ('ARRAY' eq ref($args{servers})) {
	$args{servers} = join(',', @{$args{servers}});
    }

    $admin_cli->vserver_services_nis_domain_create(%args, 
					      vserver => $self->{vserver}, 
					      active => 'true');
    $admin_cli->vserver_modify(vserver => $self->{vserver}, 
			       'ns-switch' => 'nis');

    try {
	$admin_cli->vserver_nfs_create(vserver => $self->{vserver}, 
				       v2 => 'disabled',
				       v3 => 'enabled',
				       "v4.0" => 'disabled',
				       "v4.1" => 'disabled',
				       udp => 'disabled');
    }
    catch NACL::APISet::Exceptions::ResponseException with {
	my $ex = shift;
	if ($ex->text() !~ /duplicate entry/) {
	    $ex->throw();
	}
	$Log->comment("NACL::APISet::Exceptions::ResponseException '".
		      $ex->text()."' caught and ignored.");
    };

    try {
	my @parsed_output = @{$admin_cli->vserver_export_policy_rule_show(vserver => $self->{vserver},
						policyname => 'default')->get_parsed_output()};

	if (@parsed_output) {
	    $admin_cli->vserver_export_policy_rule_modify(vserver => $self->{vserver},
						      policyname => 'default',
						      ruleindex => 1,
						      superuser => 'any',
						      anon => 0);
	}
    }
    catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
	$admin_cli->vserver_export_policy_rule_create(vserver => $self->{vserver},
						      policyname => 'default',
						      clientmatch => '0.0.0.0/0',
						      rorule => 'any',
						      rwrule => 'any',
						      superuser => 'any',
						      anon => 0);
    };
    $Log->exit();
}

############################################################
#     deletions
############################################################

=head2 delete

Release all volumes used and delete the Repository.
Vserver may be deleted as well dependent on the boolean parameter passed in.

=over 4

=item *
delete_vserver

This is an boolean to specify whether both the Repository and the Vserver 
are to be deleted at the same time.
If it is 0 (false) or is not passed as argument, 
only the Repository will be deleted
and the Vserver will be kept around.

=back


=cut

sub delete {
    $Log->enter();
    my ($self, %args) =	@_;
    $self->_delete_repository();

    if ($args{delete_vserver}) {
	$self->_delete_vserver();
    }
    $Log->exit();
}

# delete repository
sub _delete_repository {
    my $self = shift;
    my $admin_cli = $self->{admin_cli};
    my $prompts_answers = ['.*you sure you wish to proceed? {y|n}:.*'=> 'y'];
    $admin_cli->repository_volume_delete(vserver => $self->{vserver},
				  'connectrec-timeout' => $extended_timeout,
				  'connectrec-match_table' => $prompts_answers);
}

my $delay = 10; # 10 seconds
my $retries = 6; # totally 1 minute
sub _residue_volumes {
    my $self = shift;
    my @volumes;
    for(my $i = 0; (@volumes = $self->_get_volumes()) && $i < $retries; $i++) {
	sleep $delay;
    }
    return @volumes;
}

sub _delete_volume {
    my ($self, %args) = @_;
    my $admin_cli = $self->{admin_cli};
    my $prompts_answers = ['.*Do you want to continue? {y|n}:.*'=> 'y'];
    $admin_cli->volume_offline(volume => $args{volume},
			 vserver => $self->{vserver},
			 'connectrec-match_table' => $prompts_answers);

    $prompts_answers = ['.*? {y|n}:.*'=> 'y'];
    $admin_cli->volume_delete(volume => $args{volume}, 
			 vserver => $self->{vserver},
			 foreground => 'true',
			 'connectrec-match_table' => $prompts_answers);
}

# delete nis-domain if exist
sub _clear_nis_domain {
    my $self = shift;
    my $admin_cli = $self->{admin_cli};
    try {
	$admin_cli->vserver_services_nis_domain_delete(vserver => $self->{vserver},
						       domain => '*');
    }
    otherwise {}; # ignore any exception
}

sub _delete_vserver {
    my $self = shift;
    my $admin_cli = $self->{admin_cli};
    $self->_clear_nis_domain();
    try {
	$admin_cli->network_interface_delete(vserver => $self->vserver,
					     lif => '*');
    }
    otherwise {}; # ignore any exception
    try {
	$admin_cli->network_routing_groups_route_delete(vserver => $self->vserver,
							'routing-group' => '*');
    }
    otherwise {}; # ignore any exception
    try {
	$admin_cli->network_routing_groups_delete(vserver => $self->vserver,
						  'routing-group' => '*');
    }
    otherwise {}; # ignore any exception
    try {
	$admin_cli->vserver_export_policy_rule_delete(vserver => $self->vserver,
						       policyname => '*');
    }
    otherwise {}; # ignore any exception

    try {
	if (!defined($self->{rootvolume})) {
	    $self->{rootvolume} = $self->_get_rootvolume();
	}
	$self->_delete_volume(volume => $self->{rootvolume});
    }
    otherwise {}; # ignore any exception

    $admin_cli->vserver_delete(vserver => $self->{vserver});
}

############################################################
#     boot and image
############################################################

=head2 setup_bootarg

Set the user changable bootarg used by Content Repository.
Valid parameters are 

=over 4

=item *
partitionbits

Positive non-zero value (default 4)

=item *
tree_width_bits

Positive non-zero value (default 4)

=item *
ols_enabled

Either 1 (true) or 0 (false). 
Default is 1.

=item *
reboot

Reboot each node after its bootargs are updated.
Default is 0 (false).  
Set this flag to 1 to cause a reboot.

=back

After calling this method, the nodes of the Vserver need to be rebooted
in order for the argument to take effect.

=cut

sub setup_bootarg {
    $Log->enter();
    my ($self, %args) = @_;
    my %boot_spec = (
	     partitionbits => { type => SCALAR, optional => 1 },
	     tree_width_bits => { type => SCALAR, optional => 1 },
	     ols_enabled => { type => SCALAR, optional => 1 },
	     reboot => {type => SCALAR, optional => 1 },
	     );

    %args = NACL::APISet::validate_params($self,
				    params => \%args,
				    spec => \%boot_spec);

    if (!defined($args{partitionbits})) {
	$args{partitionbits} = PARTITION_BITS;
    }
    if (!defined($args{tree_width_bits})) {
	$args{tree_width_bits} = TREE_WIDTH_BITS;
    }
    if (!$args{ols_enabled}) {
	$args{ols_enabled} = 1;
    }

    # sysctl to all nodes in cluster
    foreach my $shell (@{$self->{system_shells}}) {
        # Sending all the bootargs parameters
	$self->sysctl($shell, "kern.bootargs=bootarg.nblade.cr_posix_enabled=$args{ols_enabled}");
	$self->sysctl($shell, "kern.bootargs=bootarg.mgwd.ols.tree_width_bits=$args{tree_width_bits}");
	$self->sysctl($shell, "kern.bootargs=bootarg.mgwd.ols.partitionbits=$args{partitionbits}");
	$self->sysctl($shell, "kern.bootargs=bootarg.nblade.web.policy_enabled=1");
	$self->sysctl($shell, "kern.bootargs=bootarg.nblade.no_local_fast_path=0");
	$self->sysctl($shell, "kern.bootargs=bootarg.nblade.localbypass_in=0");
	$self->sysctl($shell, "kern.bootargs=bootarg.nblade.localbypass_out=0");
    }
    $Log->exit();
}

sub sysctl {
    $Log->enter();
    my $self = shift;
    my $system_shell = shift;
    my $command = shift;
    $system_shell->execute_raw_command(command => "sudo sysctl $command");
    $Log->exit();
}

=head2 reboot

Reboot a node associated with this Content Repository.

=cut

sub reboot {
    $Log->enter();
    my ($self, %args) =	@_;
    my $admin_cli = $self->{admin_cli};
    %args = NACL::APISet::validate_params($self,
		     params => \%args,
		     spec => {
		     node => {type => SCALAR, optional => 0 },
		 },
	     );

    try {
	$admin_cli->system_reboot(node => $args{node},
				  'connectrec-timeout' => 1);
	CR::Exceptions::RebootFailed->throw("Failed to reboot $args{node}");
    }
    otherwise {}; # ignore any exception
    $Log->exit();
}

=head2 load_image

Load the system image into node given by the FILER parameter.

=over 4

=item *
package

URL of system image to be loaded.
If the package argument is not given,
the latest image from 
http://10.57.83.10/engineering/rlse/DOT/crdevN/final/bedrock/export/x86_64.debug/tarball/image.tgz 
will be loaded.

=item *
reboot

Reboot each the node after the system image is updated.
Default is 0 (false).  
Set this flag to 1 to cause a reboot.

=back

=cut

sub load_image {
    $Log->enter();
    my ($self, %args) =	@_;
    my $admin_cli = $self->{admin_cli};
    %args = NACL::APISet::validate_params($self,
		     params => \%args,
		     spec => {
		     package => { type => SCALAR, optional => 1 },
		     reboot => {type => SCALAR, optional => 1 },
		 },
	     );

    if (!defined($args{package})) {
	#todo: Should use the correct image for 32-bit and 64-bit, etc.
        #      A way to automatically detected them is preferred.
	$args{package} = "http://10.57.83.10/engineering/rlse/DOT/crdevN/final/bedrock/export/x86_64.debug/tarball/image.tgz";
    }
    
    my $reboot = $args{reboot};
    delete $args{reboot};

    my $prompts_answers = ['.*Enter User:.*'=> "",
			   '.*Enter Password:.*'=> ""];
    my $node = $self->{command_interface}; # i.e. FILER parameter
    $admin_cli->system_node_image_update(%args, 
					 node => $node, 
				   setdefault => 'true',
			    'replace-package' => 'true',
			 'connectrec-timeout' => 900,
		     'connectrec-match_table' => $prompts_answers);

    if ($reboot) {
	$self->reboot(node => $node);
    }
    $Log->exit();
}

=head1 EXCEPTIONS

=over 4

=item *
CR::Exception

This is the base class of all the exceptions pertaining to Content Repository.

=item *
CR::Exceptions::ArgumentError

The argument is of the wrong ref type as expected.
The ref type causing the exception will be in the message.

=item *
CR::Exceptions::NoAggregate

There are 2 situations (with different messages) when this exception is thrown:

=over 4

=item 1.
An existing aggregate is required for an operation and it is not found.

=item 2.
Not enough disks (i.e. less than 3) in the node to create an aggregate which is requied by the operation.

=back


=item *
CR::Exceptions::NoVServer

This exception will be thrown if an operation expecting an existing Vserver but none is found.

=item *
CR::Exceptions::NoRootVolume

This exception will be thrown if it is detected that no root volume is associated with a CR Vserver.

=item *
CR::Exceptions::VolumeNotFound

If a volume is expected but not found,
this exception will be thrown and the volume name will be attached to the exception.

=item *
CR::Exceptions::ResidueVolumes

When recycling a Vserver with an existing name, the previous instance of the Vserver will be delete first.
However, if the deletion cannot be completed with all its volumes deleted within a certain time frame,
the creation will be aborted by throwing this exception.

=back

Note that there are other exceptions that may be thrown by the underlying modules used.
They are implementation dependent and cannot be all listed here.

=cut

1;
