#===============================================================================
#     REVISION: $Id$
#         FILE: DRUtilities.pm
#
#  DESCRIPTION: This library contains common subroutines for tasks specific
#               InfiniteVol Snapmirror Relationship.
#
#         BUGS: Not known
#        NOTES: ---
#       AUTHOR: dl-svl-nosauto
#      COMPANY: NetApp, Inc
#      VERSION: 1.0
# LAST CHANGED: $DateTime$
# CHANGED BY  : $Author$
#===============================================================================

use strict;
use warnings;

use NACL::C::Node;
use NACL::STask::Snapmirror;
use NACL::STask::Volume;
use NACL::C::Client;
use NACL::STask::Client::Mount;

use CR::CommonParams;
use Tharn qw($Log);
use Params::Validate qw(SCALAR OBJECT ARRAYREF HASHREF validate);
use Options;

our @EXPORT_OK = qw();


##############################################################################
# Name          : create_uninit_sm()
# Description   : For a snapmirror relationship, does the following
#                 Deletes snapmirror relationship
#                 Release sm relationship from source
#                 Deletes destination volume
#                 Creates destination volume
#                 Creates snapmirror relationship
# Input         : Reference to Snapmirror Object and Dst Volume object
# Returns       : Nothing
#                 $TCD::FAIL if NACL calls fail
##############################################################################
sub create_uninit_sm {
    my (%params) = @_;

    my $sm_obj = $params{snapmirror_obj};
    my $dst_iv_obj = $params{dst_iv_obj};
    my $src_node_obj = $params{src_node_obj};
    my $create_sm = $params{create_sm};
    my $init_sm = $params{init_sm};
    my $dst_node_obj = $dst_iv_obj->command_interface();

    my %dst_vol_params;
    my $dst_state_obj = $dst_iv_obj->state();
    foreach my $attribute (qw(volume vserver type size)) {
        $dst_vol_params{$attribute} = $dst_state_obj->$attribute();
    }

    my %snapmirror_params =
        ('destination-path' => $sm_obj->destination_path(),
         'source-path'      => $sm_obj->source_path());

    $sm_obj->purge(nacltask_release => 1,
                   source_ci        => $src_node_obj);
    $dst_iv_obj->purge();

    $dst_iv_obj =
        NACL::STask::Volume->create(command_interface => $dst_node_obj,
                                    %dst_vol_params,
                                    nacltask_wait => 1);

    if ($create_sm || $init_sm) {
        $sm_obj =
            NACL::STask::Snapmirror->create(command_interface => $dst_node_obj,
                                            %snapmirror_params,
                                            'do_initialize' => $init_sm);
    }

    return;
}


###############################################################################
# Name          : get_init_objs()
# Description   : Can be used by DR scripts to find/create NACL objects
# Input         : Nothing
# Returns       : Reference to hash with keys
#                 SRC_NODE - Source node object
#                 DST_NODE - Destination node object
#                 ALL_SRC_NODES - Reference to array.
#                                 Has all source node objects in source cluster
#                 ALL_DST_NODES - Reference to array.
#                                 Has all source node objects in dst cluster
#                 CLIENT        - Client object
#                 IV_SNAPMIRROR - Contains all IV snapmirrors.
#                                 Array of hashes with keys:
#                                 SM_OBJ - snapmirror object
#                                 SRC_VOL_OBJ - source volume object
#                                 DST_VOL_OBJ - destination volume object
#                 FLEX_SNAPMIRROR - Contains all flexvol snapmirrors
#                                   Array of hashes with keys:
#                                   SM_OBJ - snapmirror object
#                                   SRC_VOL_OBJ - source volume object
#                                   DST_VOL_OBJ - destination volume object
#                 TIMEOUT    - Timeout specified (or default from CommonParams)
# Usage Example : $Obj_Holder = get_init_objs();
#                 $SM_Obj = $Obj_Holder->{IV_SNAPMIRROR}[0]{SM_OBJ};
#                 This will give the first IV snapmirror found
###############################################################################
sub get_init_objs {
    my $obj_holder;

    my $options = Options->new();
    $options->process();

    #Get the source and destination node names and client
    my $src_node_name = $options->get(option => 'SRC_NODE');
    my $dst_node_name = $options->get(option => 'DST_NODE');
    $obj_holder->{CLIENT} = NACL::C::Client->find();
    $obj_holder->{TIMEOUT} = $options->get(option => 'TIMEOUT');

    if (!($src_node_name and $dst_node_name)) {
        $Log->comment('SRC_NODE, DST_NODE and CLIENT are mandatory parameters');
        return;
    }

    #Get the node/filer/vsim objects
    $Log->comment('Creating required objects');
    $obj_holder->{SRC_NODE} = NACL::C::Node->new(name => $src_node_name);
    $obj_holder->{DST_NODE} = NACL::C::Node->new(name => $dst_node_name);
    @{$obj_holder->{ALL_SRC_NODES}} =
             NACL::C::Node->find(command_interface => $obj_holder->{SRC_NODE});
    @{$obj_holder->{ALL_DST_NODES}} =
             NACL::C::Node->find(command_interface => $obj_holder->{DST_NODE});

    #Get all the non-constituent snapmirror objects on the dst node
    my @snapmirror_objs = NACL::STask::Snapmirror->find(
                                  command_interface => $obj_holder->{DST_NODE},
                                  filter => {'is-constituent' => 'false'},
                                  allow_empty => 1,
                          );

    #Check which of these is an InfiniteVolume relationship.
    foreach my $sm_obj (@snapmirror_objs) {
        my $dst_obj = NACL::STask::Volume->find(
                          command_interface => $obj_holder->{DST_NODE},
                          filter => {
                                     vserver => $sm_obj->destination_vserver(),
                                     volume  => $sm_obj->destination_volume(),
                                    }
                         );
        my ($src_vserver_name);
        if ($sm_obj->source_path() =~ /^(\S+):(\S+)$/i) {
            $src_vserver_name = $1;
        }
        my $src_obj = NACL::STask::Volume->find(
                                  command_interface => $obj_holder->{SRC_NODE},
                                  filter => {
                                             volume => $sm_obj->source_volume(),
                                             vserver => $src_vserver_name
                                            }
                      );
        my $sm_details = {
                          SM_OBJ => $sm_obj,
                          SRC_VOL_OBJ => $src_obj,
                          DST_VOL_OBJ => $dst_obj,
                         };

        if ($dst_obj->get_one_state_attribute('volume-style') =~
                                                               /infinitevol/) {
            push @{$obj_holder->{IV_SNAPMIRROR}}, $sm_details;
        } else {
            push @{$obj_holder->{FLEX_SNAPMIRROR}}, $sm_details;
        }
    }

    return $obj_holder;
}


##############################################################################
# Name        : volume_restore()
# Description : Common subroutine that does the following:
#               Deletes the snapmirror relationship
#               Deletes the source and destination volume
#               Creates the source and destination volume back with size given
#               Creates Snapmirror
#               Initializes Snapmirror
#
#               This subroutine is mostly used when you need to restore
#               size of source and destination volumes, where testcases
#               have modified it.
#
# Input       : Hash with following keys
#               src_vol_size - Size of the source volume to be created
#               dst_vol_size - Size of the destination volume to be created
#               src_vol_obj  - Source Volume Object
#               dst_vol_obj  - Destination Volume Object
#               sm_obj       - Snapmirror Object
# Returns     : Nothing
##############################################################################
sub volume_restore {
    my (%params) = @_;

    my $src_vol_size = $params{src_vol_size};
    my $dst_vol_size = $params{dst_vol_size};
    my $src_vol_obj  =  $params{src_vol_obj};
    my $dst_vol_obj  =  $params{dst_vol_obj};
    my $sm_obj  =  $params{sm_obj};

    my (%src_vol_params, %dst_vol_params);
    my $src_state_obj = $src_vol_obj->state();
    my $src_node_obj = $src_vol_obj->command_interface();
    my $dst_node_obj = $dst_vol_obj->command_interface();

    foreach my $attribute (qw(volume vserver)) {
        $dst_vol_params{$attribute} = $dst_vol_obj->$attribute();
    }

    foreach my $attribute (qw(volume vserver junction-path snapshot-policy
                              unix-permissions)) {
        $src_vol_params{$attribute} = $src_state_obj->$attribute();
    }

    my %sm_params = ('destination-path' => $sm_obj->destination_path(),
                     'source-path'      => $sm_obj->source_path());

    #We cannot take the volume to its initial size, so we'll delete and create
    #it back again
    $sm_obj->purge(nacltask_release => 1,
                   source_ci        => $src_node_obj);
    $src_vol_obj->purge();
    $dst_vol_obj->purge();

    $src_vol_obj =
        NACL::STask::Volume->create(command_interface => $src_node_obj,
                                    %src_vol_params,
                                    size => $src_vol_size,
                                    type => 'RW',
                                    foreground => 'true');

    $src_vol_size = $src_vol_obj->state()->size();
    if ($dst_vol_size < $src_vol_size) {
        $dst_vol_size = $src_vol_size;
    }

    $dst_vol_obj =
        NACL::STask::Volume->create(command_interface => $dst_node_obj,
                                    %dst_vol_params,
                                    size       => $dst_vol_size,
                                    type => 'DP',
                                    foreground => 'true');

    $sm_obj =
        NACL::STask::Snapmirror->create(command_interface => $dst_node_obj,
                                        %sm_params);

    return;
}


1;
