#
# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary CR Iron Job Class
## @author sundeep@netapp.com
## @status shared
## @pod here

package CR::IronJob;

use strict;
use warnings;

use Data::Dumper;
use NACL::APISet;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use Params::Validate qw(:all);
use SPARSE::Util::Logger;
use Tharn qw($Log);

# The external interface into this library
our @EXPORT = qw(
    start_scan_on_volume
    stop_scan_on_volume
    verify_job_status_with_fiji
);

=head1 NAME

IronJob - The class to allow creation and manipulation of CR Iron jobs.

=head1 SYNOPSIS

=begin text

 use IronJob;

 %job_details = start_scan_on_volume(cli => $admin_cli,
                                vserver => $vserver
                                volume => $volume);

 stop_scan_on_volume(cli => $admin_cli,
                     vserver => $vserver
                     volume => $volume);

=end text

=head1 AUTHOR

Sundeep Singh (sundeep@netapp.com)

=head1 DESCRIPTION

This is a NACL AppBuilder to support the CR Iron jobs for Content Repository.

=head1 ROUTINES

=cut

logLevel('info');


=head2 start_scan_on_volume

The purpose of this subroutine is to start the CR Iron jobs
on specified volumes belong to specified vserver.

There are no optional parameters.

=head2 stop_scan_on_volume

The purpose of this subroutine is to stop the CR Iron jobs
on specified volumes belong to specified vserver.

There are no optional parameters.

=cut

##---------------------------------------------------
# get_job_details
#
# Inputs:
#  - cli: ngsh connection
#  - vserver: repository vserver
#  - volume: volume on which to start scan
#
# Output:
#  Returns details of the jobs spawned for the scan
#  The structure being returned today:
#
#  job_details => {
#          'master-job' => {
#                    'job-id' => Master job ID
#                    'node' => The node on which this job is running
#           }
#          'child-job' => {
#                    'job-id' => Child job ID
#                    'node' => The node on which this job is running
#           }
#  }
##-------------------------------------------------------

sub get_job_details ( % )
{
    $Log->comment("get_job_details() called with @_");

    my %opts = validate(
        @_,
        {
            cli => { type => HASHREF },
            vserver => {type => SCALAR},
            volume => {type => SCALAR},
        }
    );


    # Attempt to start a scan on the volume specified
    my $admin_cli = $opts{cli};
    my $vserver = $opts{vserver};
    my $volume = $opts{volume};

    # We need to figure out the role of the volume
    my $cons_role;
    my @parsed_output = @{$admin_cli->volume_show(
                            vserver => $vserver,
                            volume => $volume,
                            'is-constituent' => 'true')->get_parsed_output()};
    $cons_role = $parsed_output[0]->{'constituent-role'};

    $Log->comment("Got constituent role: $cons_role");

    # Check the status of the command to see if the jobs have been created.
    my %child_job;
    my %master_job;
    my $snapshot = "repos_iron_" . $volume;
    my $tries = 0;
    while (1)
    {
        #sleep(5); # Sleep for 5 secs to ensure job has been created
        my @parsed_output;
        try {
            @parsed_output = @{$admin_cli->job_show(name => "\"$cons_role scan on $volume\"",
                                     state => "Running")->get_parsed_output()};

            if (scalar(@parsed_output) > 0) {
                # Found child job
                my $child_job_id = $parsed_output[0]->{id};
                my $master_job_id;

                $Log->comment("Got child-job-id: $child_job_id and node $parsed_output[0]->{node}");

                %child_job = ('job-id' => $child_job_id,
                              'node' => $parsed_output[0]->{node});

                # Now get the master job id as well.
                my $description = $parsed_output[0]->{description};
                if ($description =~ /mj: (.*)\"/) {
                    $Log->comment("Got master-job-id: $1");
                    $master_job_id = $1;
                }
                @parsed_output = @{$admin_cli->job_show(id => $master_job_id)->get_parsed_output()};

                $Log->comment("Got master job node: $parsed_output[0]->{node}");
                %master_job = ('job-id' => $master_job_id,
                               'node' => $parsed_output[0]->{node});

                #$admin_cli->volume_snapshot_show(vserver => $vserver, snapshot => $snapshot);
                last;
            }
        } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
            # Expected if child job not yet spawned
            $tries ++;
            if ($tries > 60) {
                # Waited for over a minute
                die "Timeout waiting for job to start";
            }
        };
        sleep(1); # Sleep for 1 sec to ensure job has been created
    }

    my %result = (
                    'master-job' => \%master_job,
                    'child-job' => \%child_job
                 );

    my $master_job_id = $master_job{'job-id'};
    my $master_job_node = $master_job{'node'};
    my $child_job_id = $child_job{'job-id'};
    my $child_job_node = $child_job{'node'};

    $Log->comment("returning: $master_job_id $master_job_node and $child_job_id $child_job_node");
    return %result;
}

##---------------------------------------------------
# start_scan_on_volume
#
# Inputs:
#  - cli: ngsh connection
#  - vserver: repository vserver
#  - volume: volume on which to start scan
#
# Output:
#  Returns details of the jobs spawned for the scan
#  The structure being returned today:
#
#  job_details => {
#          'master-job' => {
#                    'job-id' => Master job ID
#                    'node' => The node on which this job is running
#           }
#          'child-job' => {
#                    'job-id' => Child job ID
#                    'node' => The node on which this job is running
#           }
#  }
##-------------------------------------------------------

sub start_scan_on_volume ( % )
{
    $Log->comment("start_scan_on_volume() called with @_");

    my %opts = validate(
        @_,
        {
            cli => { type => HASHREF },
            vserver => {type => SCALAR},
            volume => {type => SCALAR},
        }
    );


    # Attempt to start a scan on the volume specified
    my $admin_cli = $opts{cli};
    my $vserver = $opts{vserver};
    my $volume = $opts{volume};

    $admin_cli->volume_check_start(vserver => $vserver, volume => $volume);

    return get_job_details(cli => $admin_cli,
                           vserver => $vserver,
                           volume => $volume);

}

##---------------------------------------------------
# stop_scan_on_volume
#
# Inputs:
#  - cli: ngsh connection
#  - vserver: repository vserver
#  - volume: volume on which to start scan
#
##---------------------------------------------------

sub stop_scan_on_volume ( % )
{
    $Log->comment("stop_scan_on_volume() called with @_");

    my %opts = validate(
        @_,
        {
            cli => { type => HASHREF },
            vserver => {type => SCALAR},
            volume => {type => SCALAR},
        }
    );


    # Attempt to stop a scan on the volume specified
    my $admin_cli = $opts{cli};
    my $vserver = $opts{vserver};
    my $volume = $opts{volume};


    my (%result) = get_job_details(cli => $admin_cli,
                                 vserver => $vserver,
                                 volume => $volume);

    my (%master_job) = %{$result{"master-job"}};
    my $master_job_id = %master_job->{"job-id"};
    my (%child_job) = %{$result{'child-job'}};
    my $child_job_id = %child_job->{'job-id'};

   print Dumper(\%result);
   print Dumper(\%master_job);
   print Dumper(\%child_job);

    $admin_cli->volume_check_stop(vserver=> $vserver, volume => $volume);

    my $tries = 0;
    while (1)
    {
        # Sleep for 5 seconds for the jobs to be stopped.
        sleep(5);
        try {
            $admin_cli->job_show(id => $master_job_id,
                             state => "Failure");
            $admin_cli->job_show(id => $child_job_id,
                             state => "Failure");
            last;
        } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
           $tries++;
           if ($tries > 12) {
               die "Timeout waiting for jobs to be stopped";
           }
        }
    }
}

### Note the below code can be removed if not needed later
sub verify_job_status_with_fiji ( % )
{
    $Log->comment("verify_job_status_with_fiji() called with @_");

    my %opts = validate(
        @_,
        {
            cli => { type => HASHREF },
            masterjob => {type => SCALAR},
            childjob => {type => SCALAR},
            vserver => {type => SCALAR},
            volume => {type => SCALAR},
            state => {type => SCALAR},

        }
    );


    # Attempt to stop a scan on the volume specified
    my $admin_cli = $opts{cli};
    my $master_job_id = $opts{masterjob};
    my $child_job_id = $opts{childjob};
    my $vserver = $opts{vserver};
    my $volume = $opts{volume};
    my $state = $opts{state};


    #my (%result) = get_job_details(cli => $admin_cli,
    #                            vserver => $vserver,
    #                            volume => $volume);

    #my (%master_job) = %{$result{"master-job"}};
    #my $master_job_id = %master_job->{"job-id"};
    #my (%child_job) = %{$result{'child-job'}};
    #my $child_job_id = %child_job->{'job-id'};

    #print Dumper(\%result);
    #print Dumper(\%master_job);
    #print Dumper(\%child_job);

    #$Log->comment("Show master job: $master_job_id");

    my $tries = 0;
    while (1)
    {
        # Sleep for 12 seconds for the jobs to be stopped.
        sleep(12);
        try {
            $admin_cli->job_show(id => $master_job_id,
                             state => $state);
            $admin_cli->job_show(id => $child_job_id,
                             state => $state);
            #$admin_cli->volume_check_show(volume => $volume,
            #                            state => $state);
            
            last;
        } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
           $tries++;
           if ($tries > 12) {
               die "Timeout waiting for job status";
           }
        }
    }
}