#
# Copyright (c) 2010 Network Appliance, Inc.
# All rights reserved.
#
# Content Repository/PosixUtils.pm

package CR::PosixUtils;

our @EXPORT_OK = qw($CPX_DONGLE_PREFIX);
# cpx dongle file prefix
# must be in sync with CPX_CREATE_PREFIX in
# ng_shared/src/include/cr_shared_redirector.h
our $CPX_DONGLE_PREFIX = ".~DOT_CR1992";


# Define common methods in 'main' namespace
package main;

##------------------------------------------------------------
# Imports
##-------------------------------------------------------------
use CR::nfs::CommonParams;
use CR::nfs::NfsOps qw(set_env
  clean_env
  file_make_with_client
  file_make
  scalar_validate
  file_rename_with_client
  dir_make
  dir_make_with_client
  symlink_make_with_client
  file_create_with_client
  object_remove_with_client
  file_setattr_with_client
  file_getattr_with_client
  file_read_verify_with_client
  file_write_verify_with_client);

use nfs::NfsDef qw(FATTR4_FILES_FREE FATTR4_SIZE UNCHECKED);
use Params::Validate qw(SCALAR OBJECT ARRAYREF HASHREF BOOLEAN validate);
use File::Spec;
use ZapiTest;
use NACL::APISet;
use NACL::C::ClusterDate;
use NACL::C::EventConfig;
use NACL::C::Node;
use NACL::C::NetworkInterface;
use NATE::BaseException qw(:try);
use NACL::Exceptions::NoElementsFound;
use NACL::APISet::Exceptions::NoMatchingEntriesException qw(:try);
use File::Basename qw(dirname basename);
use MIME::Base64;
use Carp qw(croak);
use nfs::NfsDef qw (nfs_error);
use CR::C::DebugSmdbTableReposConstituents;
use Data::Dumper;
use CR::C::DebugSmdbTableVserverIdToName;
use Time::Local;
use POSIX qw(strftime);

# The library requries that the calling thpl define
# the following variables
use vars qw(
  $Test
  $VERSION
  $VSERVER
  $PROTOCOL
  $TESTDIR
  $COUNT
  $NAMESPACE
  $NAMESPACE_VOL
  $DOCLEANUP
  $NFS4MINORVERSION
  $NFS_CLIENT_TIMEOUT
  $NFS_LEASE_PERIOD
  $RETRY
  );

# The following objects are used to talk to the filer
my ($Node, $Api_set_nodes);

# The external interface into this library
our @EXPORT_OK = qw(posix_init
  posix_setup
  posix_uninit
  posix_cleanup
  fiji_filter
  fiji_clear
  create_files
  create_directory
  create_recursive_directory_path
  create_file
  rename_file
  read_verify_files
  write_verify_files
  truncate_files
  find_obj_files
  rename_files
  create_placement_rule
  create_object_volume
  get_default_ovg_name
  verify_object_in_ovg
  verify_object_in_volume
  corrupt_data_file_names
  corrupt_redirector
  corrupt_redirectors
  make_redirectors_dangling
  make_data_files_dangling
  create_temporary_files
  validate_files
  get_iv_on_repos_vserver_with_default_dcg
  reconnect_nfs_client_connections
  wait_for_grace_period
  );

##------------------------------------------------------------
# Local Variables
##-------------------------------------------------------------
my %env_h;        # The local handle used by the library
my @all_files;    # The files that need to be cleaned up
my @dirs;         # The directories that need to be cleaned up

##-------------------------------------------------------------
# posix_options
#
# posix_add_options is called by the thpl
# to add options posix utils references to a hash
##--------------------------------------------------------------
sub posix_options() {
    return (
        {
         name         => 'PROTOCOL',
         description  => 'The underlying transport protocol to use',
         valid_values => [qw(tcp)],
         default      => 'tcp',
        },
        {
         name         => 'NFSVERSION',
         description  => 'The NFS protocol version to use',
         valid_values => [3, 4],
         default      => '3',
        },
        {
         name => 'NAMESPACE',
         description => 'The junction path of the namespace volume against which to run these tests',
         default => '/NS',
        },
        {
         name        => 'TESTDIR',
         description => 'Path used under the mount point',
         spec        => 's',
         default     => '/',
        },
        {
         name => 'DOCLEANUP',
         description  => 'Set to 0 to leave files and directories after the test run',
         spec         => 's',
         default      => 1,
         valid_values => [0, 1],
        },
        {
         name        => 'COUNT',
         description => 'Number of Files',
         default     => 1,
        },
        {
         name         => 'NFS4MINORVERSION',
         description  => 'The NFS4 Minor version to use',
         valid_values => [0, 1],
         default      => '0',
        },
        {
         name        => 'NFS_CLIENT_TIMEOUT',
         description => 'Timeout value for NFS clients to use',
         default     => '30',
        },
        {
         name        => 'RETRY',
         description => "No of NFS op retries",
         default     => '0',
        }
    );
} ## end sub posix_options()

##-------------------------------------------------------------
# posix_read_options
#
# posix_read_options is called by the thpl
# to read options posix utils references
##--------------------------------------------------------------
sub posix_read_options($) {
    my $Options = shift;
    $PROTOCOL           = $Options->get(option => "PROTOCOL");
    $VERSION            = $Options->get(option => "NFSVERSION");
    $NAMESPACE          = $Options->get(option => "NAMESPACE");
    $TESTDIR            = $Options->get(option => "TESTDIR");
    $DOCLEANUP          = $Options->get(option => "DOCLEANUP");
    $COUNT              = $Options->get(option => "COUNT");
    $NFS4MINORVERSION   = $Options->get(option => "NFS4MINORVERSION");
    $NFS_CLIENT_TIMEOUT = $Options->get(option => "NFS_CLIENT_TIMEOUT");
    $NFS_LEASE_PERIOD   = $Options->get(option => "NFS_LEASE_PERIOD");
    $RETRY              = $Options->get(option => "RETRY");

} ## end sub posix_read_options($)
##-------------------------------------------------------------
# posix_init
#
# posix_init is called by the thpl before executing any tests to
# allow for creation of objects that would be needed by the tests.
# Returns a hash which contains the $node object created here. A
# hash allows us to extend this to return more stuff in the future
# without having to worry about breaking existing scripts.
##--------------------------------------------------------------
sub posix_init() {
    $Log->comment("posix_init called");

    $Node->{host} = Hostrec->new(id => $main::FILER);

    $Test->step("Creating objects to communicate with filer");

    my $command_interface = NACL::C::Node->find();

    # Fetching all the nodes in the cluster
    my @nodes = NACL::C::Node->find(command_interface => $command_interface);

    $Node->{nodes} = \@nodes;
    my ($node) = grep { $_->node() eq $main::FILER } @nodes;

    $Node->{cli} = $node->apiset();

    $Node->{shell} = $node->apiset(set          => 'Systemshell',
                                   version_info => {"cdefs_from_build" => 0},);

    $Node->{vscli} = $node->apiset(set => 'Vserver');

    # vol offline/online is slow
    $Node->{vscli}->set_timeout('connectrec-timeout' => 300);
    $Node->{vscli}
      ->set_vserver('vserver-context-name' => $main::VSERVER_GX_NAME);
    $Node->{vscli}->set_privilege("privilege-level" => "test");

    $Node->{zapi} = $node->apiset(interface    => 'ZAPI',
                                  version_info => {"cdefs_from_build" => 0},);
    $Node->{zapi}->set_vserver("vserver-name" => $main::VSERVER_GX_NAME);
    $Log->comment(
           "Created: $Node->{cli} $Node->{shell} $Node->{vscli} $Node->{zapi}");
    $NAMESPACE_VOL = ${
        $Node->{vscli}->volume_show('junction-path'  => $NAMESPACE,
                                    'is-constituent' => 'true'
          )->get_parsed_output()}[0]->{volume};

    # Exporting the data structures to be used for cleanup
    $Node->{cleanup_dir}   = \@dirs;
    $Node->{cleanup_files} = \@all_files;

    return {node => $Node};
} ## end sub posix_init()

##-------------------------------------------------------------
# posix_setup()
#
# Called by the thpl file as a part of init() to provide for
# configuring the script's setup. Eg: create volumes, mount, etc.
# Input Parameters: User can provide following input parementers:
#     NFS Version (user_nfs_version) - NFS client objects for
#         the specified nfs versions would be created
#     NFS Minor Version(user_minor_version) - NFS client objects
#         for the specified minor versions would also be created
##--------------------------------------------------------------
sub posix_setup {
    $Log->comment("posix_setup called");

    # Input parameters
    my (%opts) =
      validate(@_,
               {user_nfs_version => {type  => SCALAR,
                                     regex => qr/^([234]|all|not2|not3|not4)$/,
                                     optional => 1,},
                user_minor_version => {type     => SCALAR,
                                       regex    => qr/^([01]|both)$/,
                                       optional => 1,},});

    # Use user_nfs_version if user has specified it else use $VERSION
    my $user_nfs_ver = ($opts{user_nfs_version}) || $VERSION;

    # Use user_minor_version if user has specified it else use $NFS4MINOVERSION
    my $user_minor_ver = $opts{user_minor_version} || $NFS4MINORVERSION;

    # Setup environment
    eval {
        $Log->comment("Create NFS Client objects");
        set_env(setup_h   => \%env_h,
                version   => $user_nfs_ver,
                protocol  => $PROTOCOL,
                minor_ver => $user_minor_ver,
                timeout   => $NFS_CLIENT_TIMEOUT,);

        # TODO verify vserver/data lif/vol is for repository
        # XXX: why?

        my @nodes =
          map { $_->{node} }
          @{$Node->{cli}->system_node_show()->get_parsed_output()};

        #        $Log->step("Clear all previously set FIJI points");
        fiji_clear(cli => $Node->{cli}, nodes => \@nodes);

        # Similarly, take care of DOCLEANUP as well.
        # XXX: Maybe there is a better way to deal with this.
        if (!defined($DOCLEANUP)) {
            $DOCLEANUP = 1;
        }

        #TESTDIR might not have been defined by the thpl as an option
        if (!defined($TESTDIR) || $TESTDIR eq '/') {
            $TESTDIR = "/";
        } else {
            $Test->step("Setup the test directories");

            # For any sub-directories we need to create them
            eval { _lookup_file_info($main::NAMESPACE_VOL, $TESTDIR); };

            if ($@) {

                # Unable to find directory, so create it
                dir_make(
                    hostname  => $main::FILER,
                    volume    => $main::NAMESPACE,   #NOTE: misnamed (mount pnt)
                    path      => "/",
                    name      => $TESTDIR,
                    version   => $VERSION,
                    protocol  => $PROTOCOL,
                    minor_ver => $NFS4MINORVERSION);

                # Add all directories to be cleanup up later
                push(@dirs, $TESTDIR);
            } ## end if ($@)
        } ## end else [ if (!defined($TESTDIR)...
    };

    if ($@) {
        $Log->comment("posix_setup: $@");
        die $@;
    }

    return $TCD::PASS;
} ## end sub posix_setup

##-------------------------------------------------------------
# posix_cleanup()
#
# Called by the thpl to provide for any unconfiguration
# after the script has been run. Eg: unmount, etc.
##--------------------------------------------------------------
sub posix_cleanup() {
    $Log->comment("posix_cleanup called");

    if ($DOCLEANUP == 0) {
        return $TCD::PASS;
    }

    # Set up the files and directories to be cleaned-up
    $env_h{files} = \@all_files;
    $env_h{dirs}  = \@dirs;
    $Log->comment("clean_env called with '@dirs'");
    clean_env(%env_h);

    # TODO: Should we care about cleanup faling?
    return $TCD::PASS;
} ## end sub posix_cleanup()

##-------------------------------------------------------------
# posix_uninit()
#
# Called by thpl at the end to uninitialize objects
##--------------------------------------------------------------
sub posix_uninit() {
    $Log->comment("posix_uninit called");
    return $TCD::PASS;
}

##-------------------------------------------------------------
# Other common functions to share by scripts
#
# The TCD coding convention says that we need to prefix an '_'
# for internal routines. Don't know if these qualify as
# internal though.
##--------------------------------------------------------------

#---------------------------------------------------------------------
# get_default_ovg_name
#
# Returns the name of the default ovg
#---------------------------------------------------------------------
sub get_default_ovg_name() {
    my $response = $Node->{vscli}->repository_rule_show(rule => 'default');
    my @vols = @{$response->get_parsed_output()};
    return $vols[0]->{'default-data-constituent-group'};
}

#---------------------------------------------------------------------
# verify_object_in_ovg
#
# Helper to verify that an object is stored in a particular obj vol group
#
# Params:
# - fh: The file handle of the object
# - ovg: The ovg in which to verify
#---------------------------------------------------------------------
sub verify_object_in_ovg ( $$ ) {
    $Log->comment("verify_object_in_ovg called with @_");
    $Log->comment("Using Admin CLI: $Node->{vscli}");

    my ($fh, $ovg) = @_;
    my $volume   = undef;
    my $response = undef;
    my @ovgvols  = undef;

    $Log->comment("msid: $msid");

    # The following command does not return the msids if
    # I use the switch "-fields".
    $response =
      $Node->{vscli}
      ->repository_volume_show('default-data-constituent-group' => $ovg);
    @ovgvols = @{$response->get_parsed_output()};

    # There could be multiple volumes in the ovg
    foreach my $vol (@ovgvols) {

        eval { verify_object_in_volume($fh, $vol->{volume}); };

        if ($@) {

            # Could fail for all but 1 volume in the ovg
            if ($@ !~ /File does not exist in volume/) {

                # Something else went wrong.
                die $@;
            }
        } else {
            $volume = $vol->{volume};
            last;
        }
    } ## end foreach my $vol (@ovgvols)

    if (!defined($volume)) {
        die "File does not exist in ovg: $ovg";
    }

    $Log->comment("Found $fh in $ovg:$volume");
} ## end sub verify_object_in_ovg ( $$ )

#---------------------------------------------------------------------
# verify_object_in_volume
#
# Helper to verify that an object is stored in a particular volume
#
# Params:
# - fh: The file handle of the object
# - volume: The volume in which to verify
#---------------------------------------------------------------------
sub verify_object_in_volume ( $$ ) {
    $Log->comment("verify_object_in_volume called with @_");

    my ($fh, $volume) = @_;
    my $msid = unpack('II', pack('H*', $fh));

    $Log->comment("msid: $msid");

    $Log->comment("Using Admin CLI: $Node->{vscli}");

    my $response = $Node->{vscli}->volume_show(msid => $msid);

    my @vols = @{$response->get_parsed_output()};

    # The above call should have returned just one volume
    my $vol = @vols[0]->{'volume'};

    $Log->comment("Got volume $vol and msid is $msid");
    if ($vol ne $volume) {
        die "File does not exist in volume";
    }
} ## end sub verify_object_in_volume ( $$ )

##--------------------------------------------------------
# fiji_filter
#
# Sets up a fiji filter on one or more nodes (typically
# all nodes since requests might go to either one).
#
# Arguments:
# - cli: NACL NG shell object
# - nodes: Array of nodes on which to set the filter
# - filter: fiji filter to define
##----------------------------------------------------------
sub fiji_filter ( % ) {
    my %opts = validate(@_,
                        {cli    => {type => HASHREF},
                         nodes  => {type => ARRAYREF},
                         filter => {type => SCALAR},});

    foreach my $node (@{$opts{nodes}}) {
        my $result =
          $opts{cli}->system_node_run(node    => $node,
                                      command => "fiji filter '$opts{filter}'");
        if ($result->get_raw_output() !~ /fiji: filter set/) {
            die $result->get_raw_output();
        }
    }
} ## end sub fiji_filter ( % )

##--------------------------------------------------------
# fiji_clear
#
# Clears all fiji filters on a set of nodes.
#
# Arguments:
# - cli: The NACL NG shell object
# - nodes: Array of nodes to clear filters
##----------------------------------------------------------
sub fiji_clear ( % ) {
    my %opts = validate(@_,
                        {cli   => {type => HASHREF},
                         nodes => {type => ARRAYREF},});

    foreach my $node (@{$opts{nodes}}) {
        my $result =
          $opts{cli}->system_node_run(node    => $node,
                                      command => 'fiji clear all');
        if ($result->get_raw_output() !~ /fiji: filter cleared/) {
            die $result->get_raw_output();
        }
    }
} ## end sub fiji_clear ( % )

##--------------------------------------------------------
# create_files
#
# Arguments:
# - path: The full directory path in the repository
# - count: The number of files that need to be created
# - size: The size of the file
# - prefix: The prefix to use for the file names
# - fill: The data to fill file with
# - cleanup: The flag to decide whether to delete the file
#            in posix_cleanup
#
# Returns:
# An array of hashes with information about the created
# files in the form:
#   {
#     client_obj => ...,
#     ns_file    => {
#                       volume => (volume name),
#                       junction-path => (volume junction path)
#                       path => (ns file path, relative to ns root),
#                       name => (ns file name)
#                       nfs_fh => (nfs fh to get to the file)
#                   },
#     obj_file   => {
#                       volume => (volume name),
#                       junction-path => (obj vol junction path),
#                       path => (obj file path, relative to obj vol root),
#                       name => (obj file name/ns file handle),
#                       fh => (obj file handle)
#                   }
#   }
##----------------------------------------------------------
sub create_files ( % ) {
    $Log->comment("create_files called with: @_");

    my %opts = validate(@_,
                        {path   => {type => SCALAR, default => $TESTDIR},
                         count  => {type => SCALAR, default => $COUNT},
                         size   => {type => SCALAR, default => 8 * 1024 + 511},
                         fill   => {type => SCALAR, default => 'A'},
                         cleanup => { type => BOOLEAN, default => 1 },
                         prefix => {type => SCALAR, default => ''}});

    $Log->comment("Creating files for each client object");
    my @results;

    foreach my $client_obj (@{$env_h{nfs_client_objs}}) {

        my $ver       = $client_obj->version();
        my $minor_ver = $client_obj->minor_version();

        # The reason we need to generate unique prefix for each NFS object is
        # because if $COUNT is more than 1 then user would like to create
        # $COUNT unique files for each NFS object. So if $COUNT = 2 because of
        # the below logic we would create 2 unique files for each NFS object
        # and hence we would end up creating 6 files in total as expected.
        # Previously this was not unique and it was overwriting
        # existing file because of UNCHECKED create and it was a bug.
        my $file_prefix = $opts{prefix} . '_' . "NFSv$ver.$minor_ver" . '_';

        eval {
            push(@results,
                 _create_files(client_obj => $client_obj,
                               path       => $opts{path},
                               count      => $opts{count},
                               size       => $opts{size},
                               fill       => $opts{fill},
                               cleanup => $opts{cleanup},
                               prefix     => $file_prefix));
        };

        if (ref($@) || $@) {
            die $@;
        }
    } ## end foreach my $client_obj (@{$env_h...

    return @results;
} ## end sub create_files ( % )

##--------------------------------------------------------
# create_file
#
# Arguments:
# - path : The full directory path in the repository
# - name : The name of the file to be created
# - cr_mode : create mode, one of UNCHECKED, GUARDED, EXCLUSIVE
# - attr : attributes of the file to be created
# - client_uid, client_gid : uid and gid to be used to create the object
##----------------------------------------------------------
sub create_file ( % ) {
    $Log->comment("create_file called with: @_");

    my %opts = validate(@_,
                        {path => {type    => SCALAR,
                                  default => $TESTDIR},
                         name => {type    => SCALAR,
                                  default => 'file'},
                         cr_mode => {type    => SCALAR,
                                     default => UNCHECKED,
                                     regex   => qr/^\d+$/},
                         attr => {type    => HASHREF,
                                  default => {}},
                         client_uid => {type    => SCALAR,
                                        default => 0,
                                        regex   => qr/^[1-9]\d*$/},
                         client_gid => {type    => SCALAR,
                                        default => 0,
                                        regex   => qr/^[1-9]\d*$/},});

    my @results;
    $Log->comment("Creating file for each client object");
    foreach my $client_obj (@{$env_h{nfs_client_objs}}) {
        eval {
            my @curr_files;
            if ($opts{client_uid} != 0) {
                $client_obj->rpc()->auth_sys_set(uid => $opts{client_uid},
                                                 gid => $opts{client_uid});
            }

            @curr_files =
              file_create_with_client(client_obj => $client_obj,
                                      path       => $opts{path},
                                      name       => $opts{name},
                                      cr_mode    => $opts{cr_mode},
                                      attr       => $opts{attr},);

            if ($opts{client_uid} != 0) {
                $client_obj->rpc()->auth_sys_set();
            }

            push(@files, @curr_files);

            # XXX: should return @curr_files and the validation should be
            # initiated from the test itself?

            $Log->comment("Validate ns/obj split");
            push(@results,
                 _validate_files(client_obj => $client_obj,
                                 path       => $opts{path},
                                 ns_files   => \@curr_files,
                                 op         => '(CREATE): new file'));
        };
        if (ref($@) || $@) {
            die $@;
        }
    } ## end foreach my $client_obj (@{$env_h...

    return @results;
} ## end sub create_file ( % )

##--------------------------------------------------------
# validate_files
#
# This is a wrapper on _validate_files
##----------------------------------------------------------
sub validate_files ( % ) {
    $Log->comment("validate_files called with: @_");

    my %opts = validate(@_,
                        {path => {type    => SCALAR,
                                  default => $TESTDIR},
                         files => {type => ARRAYREF},});

    eval {
        foreach my $file (@{$opts{files}}) {
            my $client_obj = $file->{client_obj};
            $Log->comment("Got client_obj: $client_obj");
            $Log->comment("Got ns_file: $file->{ns_file}");

            _validate_files(client_obj => $client_obj,
                            path       => $opts{path},
                            ns_files   => [$file->{ns_file}->{name}],
                            op         => '(VALIDATE): existing file');
        } ## end foreach my $file (@{$opts{files...
    };

    if (ref($@) || $@) {
        die "validate_files() failed: $@";
    }
} ## end sub validate_files ( % )

##--------------------------------------------------------
# create_directory
#
# Arguments:
# - path     : The full directory path in the repository
#              where the directories are to be created
# - name     : The prefix to use for the directory names
# - count    : The number of directories that need to be created
# - attr     : A hash of attributes to be applied to the directories
#              when each is created
##----------------------------------------------------------
sub create_directory ( % ) {
    $Log->comment("create_directory called with: @_");

    my %opts = validate(@_,
                        {path  => {type => SCALAR, default => $TESTDIR},
                         name  => {type => SCALAR, default => 'dir'},
                         count => {type    => SCALAR,
                                   default => $COUNT,
                                   regex   => qr/^[1-9]\d*$/},
                         attr => {type => HASHREF, default => {}},});

    #TODO: Make this work for multiple client objects
    # A client object is created each for combination of
    # transport protocol and nfs version etc..

    $Log->comment("Creating directories");
    my @client_objects = @{$env_h{nfs_client_objs}};
    my $client_obj     = $client_objects[0];
    my @directories;
    eval {
        @directories =
          dir_make_with_client(client_obj => $client_obj,
                               path       => $opts{path},
                               name       => $opts{name},
                               count      => $opts{count},);

        $Log->comment("Created directories");
    };
    if (ref($@) || $@) {
        die $@;
    }

    return @directories;
} ## end sub create_directory ( % )

##--------------------------------------------------------
# create_recursive_directory_path
#
# Arguments:
# - path     : The full directory path in the repository
# - attr     : A hash of attributes to be applied to the directories
#              when each is created
# - name     : Base name to be used while creating recursive directory path
# - count    : Number of recursive directories to be created
# Notes      : Both path and name cannot be specified
##----------------------------------------------------------
sub create_recursive_directory_path ( % ) {
    $Log->comment("create_recursive_directory_path called with: @_");

    my %opts = validate(@_,
                        {path =>
                           {type => SCALAR, default => undef, optional => 1},
                         attr => {type => HASHREF, default => {}},
                         name =>
                           {type => SCALAR, default => undef, optional => 1},
                         count => {type     => SCALAR,
                                   default  => $COUNT,
                                   optional => 1,
                                   regex    => qr/^[1-9]\d*$/},});

    #TODO: Make this work for multiple client objects
    # A client object is created each for combination of
    # transport protocol and nfs version etc..

    # Checking if both path and name are specified
    if ($opts{path} && $opts{name}) {
        die
          "Both 'path' and 'name' cannot be feeded to - create_recursive_directory_path\n";
    }

    my @client_objects = @{$env_h{nfs_client_objs}};
    my $client_obj     = $client_objects[0];
    my $current_path   = $TESTDIR;
    my ($base_dir, @directories);

    # If user has specified the complete directory path to be created
    if ($opts{path}) {

        $opts{path} =~ s{/+}{/}g;
        $opts{path} =~ s{^/}{}g;

        $Log->comment("Creating recursive directory path - '$opts{path}'");
        @directories = split m{/}, $opts{path};

        # If user has just specified the base directory name
        # and count of directories to be created
    } elsif ($opts{name}) {

        @directories = map { $opts{name} . $_ } (1 .. $opts{count});

    } else {
        die
          "Either 'path' or 'name' should be provided to - create_recursive_directory_path\n";
    }

    # Storing the base dir, to be used while doing recursive cleanup
    $base_dir = $directories[0];

    foreach my $dir (@directories) {
        $Log->comment("Creating the directory - '$dir'");
        eval {
            dir_make_with_client(client_obj => $client_obj,
                                 path       => $current_path,
                                 name       => $dir,
                                 count      => 1,
                                 attr       => $opts{attr},);

        };

        if (ref($@) || $@) {
            die $@;
        }

        # Changing the current directory to change the location for new directory
        $current_path = $current_path . "/$dir";

    } ## end foreach my $dir (@directories)

    # Collecting the directory to be cleanued
    push @dirs, $TESTDIR . '/' . $base_dir;

    $Log->comment("Created the recursive directory path");

    # Returning the created directory
    $Log->comment("Returning the directory path - '$current_path'");
    return $current_path;
} ## end sub create_recursive_directory_path ( % )

##--------------------------------------------------------
# create_symlink
#
# Arguments:
# - path     : The full directory path in the repository
#              where the symlinks are to be created
# - name     : The prefix to use for the symlink names
# - target   : The target of the symbolic link
# - count    : The number of symlinks that need to be created
##----------------------------------------------------------
sub create_symlink ( % ) {
    $Log->comment("create_symlink called with: @_");

    my %opts = validate(@_,
                        {target => {type => SCALAR},
                         path   => {type => SCALAR, default => $TESTDIR},
                         name   => {type => SCALAR, default => 'symlink'},
                         count => {type    => SCALAR,
                                   default => $COUNT,
                                   regex   => qr/^[1-9]\d*$/},});

    #TODO: Make this work for multiple client objects
    # A client object is created each for combination of
    # transport protocol and nfs version etc..

    $Log->comment("Creating symlinks");
    my @client_objects = @{$env_h{nfs_client_objs}};
    my $client_obj     = $client_objects[0];
    my @symlinks;
    $Log->comment("Got client_obj: $client_obj");
    eval {
        @symlinks =
          symlink_make_with_client(client_obj => $client_obj,
                                   path       => $opts{path},
                                   target     => $opts{target},
                                   name       => $opts{name},
                                   count      => $opts{count},);

        $Log->comment("Created symlinks");
    };
    if (ref($@) || $@) {
        die $@;
    }

    return @symlinks;
} ## end sub create_symlink ( % )

##--------------------------------------------------------
# rename_files
#
# Arguments:
# - files : A array of hashes with the following format
#   {
#     client_obj => ...,
#     ns_file    => {
#                       volume => (volume name),
#                       junction-path => (volume junction path)
#                       path => (ns file path, relative to ns root),
#                       name => (ns file name)
#                       nfs_fh => (nfs fh to get to the file)
#                   },
#     obj_file   => {
#                       volume => (volume name),
#                       junction-path => (obj vol junction path),
#                       path => (obj file path, relative to obj vol root),
#                       name => (obj file name/ns file handle),
#                       fh => (obj file handle)
#                   }
#   }
# - todir : The new directory where renamed file goes
# - prefix : the prefix to add to every file name
##----------------------------------------------------------
sub rename_files ( % ) {
    $Log->comment("rename_file called with: @_");

    my %opts = validate(@_,
                        {files  => {type => ARRAYREF},
                         todir  => {type => SCALAR, default => ''},
                         prefix => {type => SCALAR, default => ''},
                         suffix => {type => SCALAR, default => ''}});

    if (prefix eq '' && suffix eq '') {
        die "Have to specify at least one of prefix or suffix";
    }

    eval {
        foreach my $file (@{$opts{files}}) {
            my $client_obj = $file->{client_obj};
            $Log->comment("Got client_obj: $client_obj");

            my $fromdir  = dirname($file->{ns_file}->{path});
            my $fromname = basename($file->{ns_file}->{path});

            $Log->comment("Got fromdir: $fromdir and fromname: $fromname");

            # todir defaults to fromdir
            if ($opts{todir} eq '') {
                $opts{todir} = $fromdir;
            }

            my $toname = $opts{prefix} . $fromname . $opts{suffix};

            my $renamed_file =
              file_rename_with_client(client_obj => $client_obj,
                                      fromdir    => $fromdir,
                                      fromname   => $fromname,
                                      todir      => $opts{todir},
                                      toname     => $toname);

            $Log->comment("Renamed file : $renamed_file");

            # If file name is already in cleanup_files, means the cleanup flag
            # of the original file is 1. Keep the renamed file the same
            foreach (@{ $Node->{cleanup_files} }) {
                if ( $_ eq $file->{ns_file}->{name} ) {
                    _remove_file_from_cleanup($file->{ns_file}->{path});
                    _add_files_to_cleanup($opts{todir}, [$toname]);
                }
            }
        } ## end foreach my $file (@{$opts{files...
    };

    if (ref($@) || $@) {
        die $@;
    }

    return @files;
} ## end sub rename_files ( % )

##--------------------------------------------------------
# truncate_files
#
# Arguments:
# - files: An array that contains the following structure
#   {
#     client_obj => ...,
#     ns_file    => {
#                       volume => (volume name),
#                       junction-path => (volume junction path)
#                       path => (ns file path, relative to ns root),
#                       name => (ns file name)
#                       nfs_fh => (nfs fh to get to the file)
#                   },
#     obj_file   => {
#                       volume => (volume name),
#                       junction-path => (obj vol junction path),
#                       path => (obj file path, relative to obj vol root),
#                       name => (obj file name/ns file handle),
#                       fh => (obj file handle)
#                   }
#   }
#  The above array is obtained directly from a call
#  to create_files
##----------------------------------------------------------
sub truncate_files ( @ ) {
    $Log->comment("truncate_files called with @_");

    my (@files) = @_;

    foreach my $fileobj (@files) {
        my $client_obj = $fileobj->{client_obj};
        my $ns_file    = $fileobj->{ns_file};

        my $attr =
          $client_obj->version() != 4 ? {size => 0} : {FATTR4_SIZE() => 0};

        $Log->comment("Path: $ns_file->{path}");
        file_setattr_with_client(client_obj => $client_obj,
                                 path       => $ns_file->{path},
                                 attr       => $attr);

        $attr = file_getattr_with_client(client_obj => $client_obj,
                                         path       => $ns_file->{path});

        if ($attr->{size} != 0) {
            die "getattr returned a non-zero size";
        }
    } ## end foreach my $fileobj (@files)
} ## end sub truncate_files ( @ )

##--------------------------------------------------------
# unlink_files
#
# Arguments:
# - files: An array that contains the following structure
#   {
#     client_obj => ...,
#     ns_file    => {
#                       volume => (volume name),
#                       junction-path => (volume junction path)
#                       path => (ns file path, relative to ns root),
#                       name => (ns file name)
#                       nfs_fh => (nfs fh to get to the file)
#                   },
#     obj_file   => {
#                       volume => (volume name),
#                       junction-path => (obj vol junction path),
#                       path => (obj file path, relative to obj vol root),
#                       name => (obj file name/ns file handle),
#                       fh => (obj file handle)
#                   }
#   }
#  The above array is obtained directly from a call
#  to create_files
##----------------------------------------------------------
sub unlink_files ( @ ) {
    $Log->comment("unlink_files called with @_");

    my (@files) = @_;

    foreach my $fileobj (@files) {
        my $client_obj = $fileobj->{client_obj};
        my $ns_file    = $fileobj->{ns_file};

        $Log->comment("Path: $ns_file->{path}");
        object_remove_with_client(client_obj => $client_obj,
                                  path       => $ns_file->{path});
        _remove_file_from_cleanup($ns_file->{path});
    }
} ## end sub unlink_files ( @ )

##--------------------------------------------------------
# find_object_files
#
# Arguments:
# - path: The full directory path in the repository
# - files: Ann array of file names for which to lookup obj files.
#
# Returns:
# An array of hashes with information about the found
# files in the form:
#   {
#     client_obj => ...,
#     ns_file    => {
#                       volume => (volume name),
#                       junction-path => (volume junction path)
#                       path => (ns file path, relative to ns root),
#                       name => (ns file name)
#                       nfs_fh => (nfs fh to get to the file)
#                   },
#     obj_file   => {
#                       volume => (volume name),
#                       junction-path => (obj vol junction path),
#                       path => (obj file path, relative to obj vol root),
#                       name => (obj file name/ns file handle),
#                       fh => (obj file handle)
#                   }
#   }
##----------------------------------------------------------
sub find_object_files ( % ) {
    $Log->comment("find_object_files called with: @_");

    my %opts = validate(@_,
                        {path  => {type => SCALAR},
                         files => {type => ARRAYREF}});

    # Use the first client object avaialable
    my $filer_conn = $env_h{fc};
    my $client_obj = @{$env_h{nfs_client_objs}}[0];

    $Log->comment("Get the object files");

    #    my @obj_files = _find_obj_files( filer_conn => $filer_conn,
    #                                    volume => $client_obj->mnt_path(),
    #                                    path => $opts{path},
    #                                    ns_files => $opts{files} );

## TODO: once the nfs burt got fixed, the below call should be removed and above
## call should be uncommented.

    my @obj_files = _find_obj_files(filer_conn => $filer_conn,
                                    volume     => $main::JUNCTION_PATH,
                                    path       => $opts{path},
                                    ns_files   => $opts{files});

    return @obj_files;
} ## end sub find_object_files ( % )

##--------------------------------------------------------
# create_object_volume
#
# Arguments:
# - name: The name of the volume to create.
# - space: The size of the volume to e created.
# - ovg: The name of the ovg to
##---------------------------------------------------------
sub create_object_volume ( % ) {
    $Log->comment("create_volume called with @_");
    $Log->comment("Using Admin CLI: $Node->{vscli}");

    my %opts = validate(@_,
                        {name  => {type => SCALAR},
                         ovg   => {type => SCALAR},
                         space => {type => SCALAR, default => "21MB"}});

    my $aggrName;

    eval { $aggrName = _get_aggregate_with_free_space($opts{space}); };

    if ($@) {
        die "create_object_volume: $@";
    }

    try {

        # Create Volume on aggregate
        $Test->step("Making a volume on aggregate $aggrName");

        $Node->{vscli}->volume_create(volume             => $opts{name},
                                      aggregate          => $aggrName,
                                      'unix-permissions' => '---rwxrwxrwx');

        $Log->comment("Created a volume $opts{name} on aggregate $aggrName");

        # Assign object role to the volume and add it to
        # a test ovg
        $Test->step("Assigning object role to the volume");

        $Node->{vscli}->repository_volume_assign(
                                                role    => 'object',
                                                volumes => $opts{name},
                                                'obj-volume-group' => $opts{ovg}
        );

    } ## end try
    catch NATE::BaseException with {
        my $ex = shift;
        die $ex->text();
    };
} ## end sub create_object_volume ( % )

##--------------------------------------------------------
# create_placement_rule
#
# Arguments:
# - name: The name of the rule to create.
# - expr: The rule expression.
# - type: The type parameter
# - ovg: The object volume group
##---------------------------------------------------------
sub create_placement_rule ( % ) {
    $Log->comment("create_placement_rule called with @_");
    $Log->comment("Using Admin CLI: $Node->{vscli}");

    my %opts = validate(@_,
                        {name => {type => SCALAR},
                         expr => {type => SCALAR},
                         type => {type => SCALAR},
                         ovg  => {type => SCALAR}});

    try {
        $Node->{vscli}->repository_rule_create(
                                  rule        => $opts{name},
                                  expression  => $opts{expr},
                                  'rule-type' => $opts{type},
                                  'default-data-constituent-group' => $opts{ovg}
        );

        $Log->comment("Adding placement rule $test_rule to ruleset for CR");

        my $repository_ruleset_add_response = undef;

        $repository_ruleset_add_response =
          $Node->{vscli}->repository_ruleset_add_rule(rule      => $opts{name},
                                                      ruleindex => '1');
    } ## end try
    catch NATE::BaseException with {
        my $ex = shift;
        die $ex->text();
    };

    # XXX: If snoozing is not done at this point
    # As the panic does not occur in the create test,
    # it could be because volumes and repository rules are
    # being created here
    $Log->comment("Snoozing for 15 seconds before creating file\n");
    snooze(15);
} ## end sub create_placement_rule ( % )

##--------------------------------------------------------
# lookup_file_handle
#
# The routine uses showfh to lookup a file. The advantage of using this
# over ZAPI is that this contains the flags as well.
#
# Arguments:
# - volume: The volume which has our file.
# - file: The file path to lookup, excluding the volume name.
# - xsid: To use 'dsid' or 'msid'
##----------------------------------------------------------
sub lookup_file_handle( $$$ ) {
    $Log->comment("lookup_file_handle called with @_");

    my ($volume, $file, $xsid) = @_;
    my $fh = undef;

    # Get the node corresponding to a volume
    my $response =
      $Node->{vscli}->volume_show(volume => $volume,
                                  field  => "node");

    my @vols = @{$response->get_parsed_output()};

    my $node = @vols[0]->{'node'};

    my $ordinal_vol_name = _get_ordinal_volume_name($volume);

    #node run -node username-vsim1 -command showfh /vol/ns1/onekbfile
    my $showfhcmd =
        "node run -node "
      . $node
      . " -command showfh " . "'" . "/vol/"
      . $ordinal_vol_name . "/"
      . $file . "'";

    my $output = $Node->{cli}->execute_raw_command(command => $showfhcmd);

    # This was my
    #   $output = substr($output, index($output, $file) + length($file));
    # Changed it because the output does not contain filename anymore
    # Will it come back?

    $Log->comment("$output");
    if ($output !~ /.*dsid.*/) {
        die "showfh failed!";
    }

    my @values = split(/[= ]/, $output);

    my $xsid_flag = ($xsid eq 'dsid') ? ('dsid') : ('msid');

    my %hash_values = (@values);
    my $msid        = $hash_values{$xsid_flag};
    my $flags       = $hash_values{"flags"};
    my $fileid      = $hash_values{"fileid"};
    my $gen         = $hash_values{"gen"};
    my $snap_id     = $hash_values{'snapid'};

    if ($xsid eq 'origmsid'){

        if ($snap_id != 0){
            croak 'Looks like FH is being fetched using origmsid for a SS file'
                  . ' which is not implemented by this function';
        }

        my $repos_cons_table_obj = CR::C::DebugSmdbTableReposConstituents->find(
            command_interface => ${$Node->{nodes}}[0],
            filter => {vserver => $VSERVER,
                       msid => hex($msid)}
        );

        my $orig_msid
            = $repos_cons_table_obj->get_one_state_attribute('origMsid');
        $msid = sprintf('%#014x', $orig_msid);
    }

    my $i_msid = pack("Q", hex($msid));
    my $str_msid = unpack('H*', $i_msid);

    my $flags_ino_igen = pack("III", hex($flags), hex($fileid), hex($gen));
    my $str_flags_ino_igen = unpack('H*', $flags_ino_igen);

    $fh = $str_msid . $str_flags_ino_igen;
    $Log->comment("Returning fh: $fh");

    return $fh;
} ## end sub lookup_file_handle( $$$ )

##--------------------------------------------------------
# read_verify_files
#
# Arguments:
# - files: An array that contains the following structure
#          {
#             client_obj => ...,
#             ns_file => ...,
#             obj_file => ...  (this component is not used)
#          }
# - fill: character occupying the file, default 'A'
#  The above array is obtained directly from a call
#  to create_files
##----------------------------------------------------------
sub read_verify_files ( @ ) {
    $Log->comment("read_verify_files called with @_");

    my %opts = validate(@_,
                        {files => {type => ARRAYREF},
                         fill  => {type => SCALAR, default => 'A'}});

    foreach my $fileobj (@{$opts{files}}) {
        my $client_obj = $fileobj->{client_obj};
        my $ns_file    = $fileobj->{ns_file};
        my $path       = '';

        if (!$ns_file->{nfs_fh}) {
            $path = $ns_file->{path};
        } else {
            $client_obj->curr_fh($ns_file->{nfs_fh});
        }

        $Log->comment("Path: $ns_file->{path}");
        file_read_verify_with_client(client_obj => $client_obj,
                                     path       => $path,
                                     fill       => $opts{fill});
        if (!$ns_file->{nfs_fh}) {
            $ns_file->{nfs_fh} = $client_obj->curr_fh();
        }
    } ## end foreach my $fileobj (@{$opts...
} ## end sub read_verify_files ( @ )

##--------------------------------------------------------
# write_verify_files
#
# Arguments:
# - files: An array that contains the following structure
#          {
#             client_obj => ...,
#             ns_file => ...,
#             obj_file => ...  (this component is not used)
#          }
# - fill: character occupying the file, default 'A'
#  The above array is obtained directly from a call
#  to create_files
##----------------------------------------------------------
sub write_verify_files ( @ ) {
    $Log->comment("write_verify_files called with @_");

    my %opts = validate(@_,
                        {files => {type => ARRAYREF},
                         fill  => {type => SCALAR, default => 'A'}});

    foreach my $fileobj (@{$opts{files}}) {
        my $client_obj = $fileobj->{client_obj};
        my $ns_file    = $fileobj->{ns_file};
        my $path       = '';

        if (!$ns_file->{nfs_fh}) {
            $path = $ns_file->{path};
        } else {
            $client_obj->curr_fh($ns_file->{nfs_fh});
        }

        $Log->comment("Path: $ns_file->{path}");
        file_write_verify_with_client(client_obj => $client_obj,
                                      path       => $path,
                                      fill       => $opts{fill});
        if (!$ns_file->{nfs_fh}) {
            $ns_file->{nfs_fh} = $client_obj->curr_fh();
        }
    } ## end foreach my $fileobj (@{$opts...
} ## end sub write_verify_files ( @ )

##-----------------------------------------------------
# The following routines are used to create inconsistencies
# in the repository file system
##------------------------------------------------------

##-----------------------------------------------------
## make_redirectors_dangling
## Helper to Create dangling redirector files
#
# Arguments:
# An array of hashes with information about the created
# files in the form:
#   {
#     client_obj => ...,
#     ns_file    => {
#                       volume => (volume name),
#                       junction-path => (volume junction path)
#                       path => (ns file path, relative to ns root),
#                       name => (ns file name)
#                       nfs_fh => (nfs fh to get to the file)
#                   },
#     obj_file   => {
#                       volume => (volume name),
#                       junction-path => (obj vol junction path),
#                       path => (obj file path, relative to obj vol root),
#                       name => (obj file name/ns file handle),
#                       fh => (obj file handle)
#                   }
#   }
##----------------------------------------------------

sub make_redirectors_dangling ( % ) {
    $Log->comment("make_redirectors_dangling() called with @_");

    my %opts = validate(@_, {files => {type => ARRAYREF},});

    eval {
        foreach my $file (@{$opts{files}}) {

            my $obj_file = $file->{obj_file};
            my $file_to_delete =
              File::Spec->catfile("/clus", $main::VSERVER_GX_NAME,
                                  $obj_file->{'junction-path'},
                                  $obj_file->{path});

            $Log->comment(
                "Log into system shell and delete the obj file : $file_to_delete "
            );

            my $Free_BSD_Cli = $Node->{shell};

            # Send a command over the CLI
            my $raw_response = $Free_BSD_Cli->execute_raw_command(
                                               command => "rm $file_to_delete");

            # execute_raw_command() always returns the raw response as a string
            $Log->comment(
                     "The node returned the following raw text: $raw_response");

            if ($raw_response =~ /No such file/) {
                $Log->comment("The file $file_to_delete does not exist");
                die "The deletion for file failed: $raw_response";
            } else {
                eval {

                    # validate after delete
                    $Log->comment(
                               "Validate that deletion of file was successful");

                    my $path =
                      File::Spec->catfile('/vol', $obj_file->{'volume'},
                                          $obj_file->{path});

                    my $result =
                      _invoke_zapi('file-read-file', 'path', $path, 'offset', 0,
                                   'length', 40);

                    my $data = $result->child_get_string('data');
                    $Log->comment("Data read from data file: $data");
                };
                if ($@ =~ /No such file or directory/) {

                    # file-read-file should fail
                    $Log->comment("data file deleted, $@");
                } else {
                    $Log->comment("Data file still exists: $@");
                    die "data file still exists: $@";
                }

                eval {
                    $Log->comment(
                                "Ensure that the redirector file still exists");

                    my $ns_file = $file->{ns_file};
                    my $path = File::Spec->catfile('/vol', $ns_file->{'volume'},
                                                   $ns_file->{path});

                    my $result =
                      _invoke_zapi('file-read-file', 'path', $path, 'offset', 0,
                                   'length', 40);

                    my $data = $result->child_get_string('data');
                    $Log->comment("Data read from redirector file: $data");
                };
                if ($@) {
                    $Log->comment("Redirector file also got deleted: $@");
                    die "Redirector file also got deleted: $@";
                }
            } ## end else [ if ($raw_response =~ /No such file/)
        } ## end foreach my $file (@{$opts{files...
    };

    if ($@) {
        $Log->comment("make_redirectors_dangling() failed: $@");
        die "make_redirectors_dangling() failed: $@";
    }
} ## end sub make_redirectors_dangling ( % )

##-----------------------------------------------------
## make_data_files_dangling
## Helper to make existing data files dangling
#
# Arguments:
# An array of hashes with information about the created
# files in the form:
#   {
#     client_obj => ...,
#     ns_file    => {
#                       volume => (volume name),
#                       junction-path => (volume junction path)
#                       path => (ns file path, relative to ns root),
#                       name => (ns file name)
#                       nfs_fh => (nfs fh to get to the file)
#                   },
#     obj_file   => {
#                       volume => (volume name),
#                       junction-path => (obj vol junction path),
#                       path => (obj file path, relative to obj vol root),
#                       name => (obj file name/ns file handle),
#                       fh => (obj file handle)
#                   }
#   }
##----------------------------------------------------
sub make_data_files_dangling ( % ) {
    $Log->comment("make_data_files_dangling() called with @_");

    my %opts = validate(@_, {files => {type => ARRAYREF},});

    eval {
        foreach my $file (@{$opts{files}}) {

            my $ns_file = $file->{ns_file};
            my $file_to_delete =
              File::Spec->catfile("/clus", $main::VSERVER_GX_NAME,
                                  $ns_file->{'junction-path'},
                                  $ns_file->{path});

            $Log->comment("Delete namespace file: rm $file_to_delete");

            # Send a command over the CLI
            my $raw_response =
              $Node->{shell}
              ->execute_raw_command(command => "rm $file_to_delete");

            # execute_raw_command() always returns the raw response as a string
            $Log->comment("Response for delete op: $raw_response");

            if ($raw_response != '') {
                die "The deletion for file failed";
            } else {
                eval {

                    # validate after delete
                    $Log->comment(
                               "Validate that deletion of file was successful");

                    my $path = File::Spec->catfile('/vol', $ns_file->{'volume'},
                                                   $ns_file->{path});

                    my $result =
                      _invoke_zapi('file-read-file', 'path', $path, 'offset', 0,
                                   'length', 40);

                    my $data = $result->child_get_string('data');
                    $Log->comment("Data read from redirector file: $data");
                };
                if ($@ =~ /No such file or directory/) {

                    # file-read-file should fail
                    $Log->comment("Redirector deleted, $@");
                } else {
                    $Log->comment("Redirector still exists: $@");
                    die "Redirector still exists: $@";
                }

                eval {
                    $Log->comment("Ensure that the data file still exists");

                    my $obj_file = $file->{obj_file};
                    my $path =
                      File::Spec->catfile('/vol', $obj_file->{'volume'},
                                          $obj_file->{path});

                    my $result =
                      _invoke_zapi('file-read-file', 'path', $path, 'offset', 0,
                                   'length', 40);

                    my $data = $result->child_get_string('data');
                    $Log->comment("Data read from data file: $data");
                };
                if ($@) {
                    $Log->comment("Data file also got deleted: $@");
                    die "Data file also got deleted: $@";
                }
            } ## end else [ if ($raw_response != '')
        } ## end foreach my $file (@{$opts{files...
    };

    if ($@) {
        $Log->comment("make_data_files_dangling() failed: $@");
        die "make_data_files_dangling() failed: $@";
    }

}    # make_data_files_dangling

##-----------------------------------------------------
## corrupt_data_file_names
## Helper to corrupt existing data file names
#
# Arguments:
# An array of hashes with information about the created
# files in the form:
#   {
#     client_obj => ...,
#     ns_file    => {
#                       volume => (volume name),
#                       junction-path => (volume junction path)
#                       path => (ns file path, relative to ns root),
#                       name => (ns file name)
#                       nfs_fh => (nfs fh to get to the file)
#                   },
#     obj_file   => {
#                       volume => (volume name),
#                       junction-path => (obj vol junction path),
#                       path => (obj file path, relative to obj vol root),
#                       name => (obj file name/ns file handle),
#                       fh => (obj file handle)
#                   }
#   }
##----------------------------------------------------
sub corrupt_data_file_names ( % ) {
    $Log->comment("corrupt_data_file_names() called with @_");

    my %opts = validate(@_, {files => {type => ARRAYREF},
                             name => {type => SCALAR, optional => 1, 
                             default => '0000000000000000000000000000000000000000'},
    });

    my $corrupt_value = $opts{name};
    eval {
        foreach my $file (@{$opts{files}}) {

            my $obj_file = $file->{obj_file};
            my ($a, $subdir) = split('/', $obj_file->{path});

            my $file_to_corrupt =
              File::Spec->catfile("/clus",
                                  $main::VSERVER_GX_NAME,
                                  $obj_file->{'junction-path'},
                                  $subdir,
                                  $obj_file->{name});

            my $file_corrupted =
              File::Spec->catfile("/clus", $main::VSERVER_GX_NAME,
                                  $obj_file->{'junction-path'},
                                  $subdir, $corrupt_value);

            $Log->comment(
                        "Rename obj file: mv $file_to_corrupt $file_corrupted");

            # Send a command over the CLI
            my $raw_response =
              $Node->{shell}->execute_raw_command(
                              command => "mv $file_to_corrupt $file_corrupted");

            # execute_raw_command() always returns the raw response as a string
            $Log->comment("Response for rename: $raw_response");

            if ($raw_response != '') {
                die "The rename for file failed";
            } else {
                $Log->comment("Validate that rename of file was successful");

                # validate after rename
                eval {
                    my $obj_file = $file->{obj_file};
                    my $path =
                      File::Spec->catfile('/vol', $obj_file->{'volume'},
                                          $obj_file->{path});

                    my $result =
                      _invoke_zapi('file-read-file', 'path', $path, 'offset', 0,
                                   'length', 40);

                    my $data = $result->child_get_string('data');
                    $Log->comment("Data read from data file: $data");
                };
                if ($@ =~ /No such file or directory/) {

                    # file-read-file should fail
                    $Log->comment("File renamed: $@");
                } else {
                    $Log->comment("Maybe data file name still exists: $@");
                    die "Maybe data file name still exists: $@";
                }
            } ## end else [ if ($raw_response != '')
        } ## end foreach my $file (@{$opts{files...
    };

    if ($@) {
        $Log->comment("corrupt_data_file_names() failed: $@");
        die "corrupt_data_file_names() failed: $@";
    }

}    # corrupt_data_file_names

#---------------------------------------------------------------------
# corrupt_redirector
#
# Helper to modify redirector file
#---------------------------------------------------------------------
sub corrupt_redirector {
    $Log->comment("corrupt_redirector called with @_");
    my ($path, $offset, $data) = @_;

    $offset ||= '0';
    $data   ||= '00000000';

    my $len_bytes_written =
      _invoke_zapi('file-write-file', 'path', $path, 'offset', $offset, 'data',
                   $data);

} ## end sub corrupt_redirector

##---------------------------------------------------------------------
## corrupt_redirectors
## Helper to corrupt existing redirector files.
##
# Arguments:
# An array of hashes with information about the created
# files in the form:
#   {
#     client_obj => ...,
#     ns_file    => {
#                       volume => (volume name),
#                       junction-path => (volume junction path)
#                       path => (ns file path, relative to ns root),
#                       name => (ns file name)
#                       nfs_fh => (nfs fh to get to the file)
#                   },
#     obj_file   => {
#                       volume => (volume name),
#                       junction-path => (obj vol junction path),
#                       path => (obj file path, relative to obj vol root),
#                       name => (obj file name/ns file handle),
#                       fh => (obj file handle)
#                   }
#   }
##----------------------------------------------------

sub corrupt_redirectors( % ) {
    $Log->comment("corrupt_redirectors() called with @_");

    my %opts = validate(@_, {files => {type => ARRAYREF},});

    eval {
        foreach my $file (@{$opts{files}}) {
            my $client   = $file->{client_obj};
            my $path     = $file->{ns_file}->{path};
            my $vol_path = $file->{ns_file}->{volume};

            my $redir_path = '/vol' . "/" . $vol_path . "/" . $path;
            my $offset =
              20;    # TODO : Sync with change expected in redirector format
            my $data   = 11111111;
            my $length = 4;

            $Log->comment("Make the redirector corrupt at $redir_path");

            #te_temporary_file Call the routine for making the redirector corrupt
            corrupt_redirector($redir_path, $offset, $data);

            # ensure that the data has changed
            my $read_response =
              $Node->{zapi}->file_read_file(path   => $redir_path,
                                            offset => $offset,
                                            length => $length);

            my @out_array = @{$read_response->get_parsed_output()};
            $Log->comment("The data from file: $out_array[0]->{data}");

            if ($data != $out_array[0]->{data}) {
                $Log->comment("The data does not match");
                die "Corrupt data does not match";
            }
        } ## end foreach my $file (@{$opts{files...
    };

    if ($@) {
        die "corrupt_redirectors() failed: $@";
    }
} ## end sub corrupt_redirectors( % )

##-------------------------------------------------------------------
## create_temporary_files
#
# Helper to create temporary files in the NS volume, which will
# have a corresponding data file as well.
#
# Arguments:
# - path: The directory to use for creating these temproary files
# - prefix: The prefix to use for the file names
# - count: The number of temporary files to create
##--------------------------------------------------------------------
sub create_temporary_files ( % ) {
    $Log->comment("create_temporary_files called with @_");

    my %opts = validate(@_,
                        {path => {type    => SCALAR,
                                  default => $TESTDIR},
                         prefix => {type => SCALAR, default => ''},
                         count  => {type => SCALAR, default => $COUNT},});
    my (@nodes, $dongle_file);

    eval {
        @nodes =
          map { $_->{node} }
          @{$Node->{cli}->system_node_show()->get_parsed_output()};

        $Log->comment("Clear all previously set FIJI points");
        fiji_clear(cli => $Node->{cli}, nodes => \@nodes);

        $Log->comment("Set the FIJI point to make sure that create fails.");
        fiji_filter(
                 cli   => $Node->{cli},
                 nodes => \@nodes,
                 filter =>
                   'repeat 10000 cpx.create.subop.rename.failure(nil)(return 1)'
        );
        sleep(10);

        $Log->comment("Create the file in path: $TESTDIR");
        eval {
            my @files = create_files(path   => $opts{path},
                                     count  => $opts{count},
                                     prefix => $opts{prefix});
        };

        if ($@ =~ /NFS3ERR_IO/) {

            # Expected to fail
            $Log->comment("Create file failed with error: $@");

            eval {
                my $index      = 1;
                my $client_obj = ${$env_h{nfs_client_objs}}[0];
                my $ver        = $client_obj->version();
                my $minor_ver  = $client_obj->minor_version();

                # The temporary filename after the creation failed.
                $dongle_file =
                    $CR::PosixUtils::CPX_DONGLE_PREFIX
                  . $opts{prefix} . '_'
                  . "NFSv$ver.$minor_ver" . '_';
                for ($index = 1 ; $index <= $opts{count} ; $index++) {

                    # Validate that the dongle file does exist
                    my $path =
                      File::Spec->catfile("/vol", $NAMESPACE_VOL, $opts{path},
                                          $dongle_file);

                    if ($opts{count} != 1) {
                        $path = $path . "." . $index;
                    }

                    $Log->comment(
                              "Validate that the dongle file \'$path\' exists");

                    my $response =
                      _invoke_zapi('file-read-file', 'path', $path, 'offset', 0,
                                   'length', 40);

                    my $data = $response->child_get_string('data');
                    $Log->comment("Data read from redirector file: $data");
                } ## end for ($index = 1 ; $index...
            };
            if (!$@) {

                # file-read-file should succeed
            } else {
                $Log->comment("Temporary file does not exist: $@");
                die "Temporary file does not exist: $@";
            }
        } else {

            # Expecting I/O error.
            die "Create file should have failed with NFS3ERR_IO: $@";
        }
    };

    # cleanup FIJI points
    $Log->comment("Clean-up all set FIJI points");
    fiji_clear(cli => $Node->{cli}, nodes => \@nodes);

    if ($@) {
        die "create_temporary_files() failed: $@";
    }

    return $dongle_file;
} ## end sub create_temporary_files ( % )

##-----------------------------------------------------
## corrupt_redirector_link_count
## Helper to increase the link count of a redirector file
#
# Arguments:
# An array of hashes with information about the created
# files in the form:
#   {
#     client_obj => ...,
#     ns_file    => {
#                       volume => (volume name),
#                       junction-path => (volume junction path)
#                       path => (ns file path, relative to ns root),
#                       name => (ns file name)
#                       nfs_fh => (nfs fh to get to the file)
#                   },
#     obj_file   => {
#                       volume => (volume name),
#                       junction-path => (obj vol junction path),
#                       path => (obj file path, relative to obj vol root),
#                       name => (obj file name/ns file handle),
#                       fh => (obj file handle)
#                   }
#   }
##----------------------------------------------------
sub corrupt_redirector_link_count ( % ) {
    $Log->comment("corrupt_redirector_link_count() called with @_");

    my %opts = validate(@_, {files => {type => ARRAYREF},
                             count => {type => SCALAR , optional => 1, default => 1},
    });

    eval {
        foreach my $file (@{$opts{files}}) {
            for (my $i=0; $i < $opts{count}; $i++) {
            my $ns_file = $file->{ns_file};
            my $file_to_link =
              File::Spec->catfile("/clus", $main::VSERVER_GX_NAME,
                                  $ns_file->{'junction-path'},
                                  $ns_file->{path});
            my $rand_val = int(rand(1000000));
            my $file_link =
              File::Spec->catfile("/clus",
                                  $main::VSERVER_GX_NAME,
                                  $ns_file->{'junction-path'},
                                  "criron_link_file_$rand_val");
            $Log->comment(
                 "create hardlink namespace file: ln $file_to_link $file_link");

            # Send a command over the CLI
            my $raw_response =
              $Node->{shell}
              ->execute_raw_command(command => "ln $file_to_link $file_link");

            # execute_raw_command() always returns the raw response as a string
            $Log->comment("Response for link op: $raw_response");
            if ($raw_response != '') {
                die "The hardlink creation for file failed";
            } else {
                eval {

                    # validate after delete
                    $Log->comment(
                               "Validate that hardlink of file was successful");

                    my $path = File::Spec->catfile('/vol', $ns_file->{'volume'},
                                                  "criron_link_file_$rand_val");

                    my $result =
                      _invoke_zapi('file-read-file', 'path', $path, 'offset', 0,
                                   'length', 40);

                    my $data = $result->child_get_string('data');
                    $Log->comment("Data read from redirector file: $data");
                };
                if ($@ =~ /No such file or directory/) {

                    # file-read-file should fail
                    $Log->comment("File not created, Hardlink failed, $@");
                    die "Hardlink failed : $@";
                }
            } ## end else [ if ($raw_response != '')
            } ## end for (my $i=0; $i < $opt...
        } ## end foreach my $file (@{$opts{files...
    };

    if ($@) {
        $Log->comment("corrupt_redirector_link_count() failed: $@");
        die "corrupt_redirector_link_count() failed: $@";
    }

}    # corrupt_redirector_link_count
##-----------------------------------------------------
## corrupt_data_link_count
## Helper to increase the link count of a data file
#
# Arguments:
# An array of hashes with information about the created
# files in the form:
#   {
#     client_obj => ...,
#     ns_file    => {
#                       volume => (volume name),
#                       junction-path => (volume junction path)
#                       path => (ns file path, relative to ns root),
#                       name => (ns file name)
#                       nfs_fh => (nfs fh to get to the file)
#                   },
#     obj_file   => {
#                       volume => (volume name),
#                       junction-path => (obj vol junction path),
#                       path => (obj file path, relative to obj vol root),
#                       name => (obj file name/ns file handle),
#                       fh => (obj file handle)
#                   }
#   }
##----------------------------------------------------
sub corrupt_data_link_count ( % ) {
    $Log->comment("corrupt_data_link_count() called with @_");

    my %opts = validate(@_, {files => {type => ARRAYREF},
                             count => {type => SCALAR , optional => 1, default => 1},});

    eval {
        foreach my $file (@{$opts{files}}) {

            for (my $i=0; $i < $opts{count}; $i++) {
            my $obj_file = $file->{obj_file};
            my $file_to_link =
              File::Spec->catfile("/clus", $main::VSERVER_GX_NAME,
                                  $obj_file->{'junction-path'},
                                  $obj_file->{path});
            my $rand_val  = int(rand(1000000));
            my $file_link = $file_to_link . '.hl.' . $rand_val;

            #my $file_link = File::Spec->catfile("/clus",
            #                                         $main::VSERVER_GX_NAME,
            #                                         $obj_file->{'junction-path'},
            #                                         "criron_link_file_$rand_val");
            $Log->comment(
                      "create hardlink data file: ln $file_to_link $file_link");

            # Send a command over the CLI
            my $raw_response =
              $Node->{shell}
              ->execute_raw_command(command => "ln $file_to_link $file_link");

            # execute_raw_command() always returns the raw response as a string
            $Log->comment("Response for link op: $raw_response");
            if ($raw_response != '') {
                die "The hardlink creation for file failed";
            } else {
                eval {

                    # validate after delete
                    $Log->comment(
                               "Validate that hardlink of file was successful");

                    my $path =
                      File::Spec->catfile('/vol', $obj_file->{'volume'},
                                        $obj_file->{path} . '.hl.' . $rand_val);

                    #"criron_link_file_$rand_val");

                    my $result =
                      _invoke_zapi('file-read-file', 'path', $path, 'offset', 0,
                                   'length', 40);

                    my $data = $result->child_get_string('data');
                    $Log->comment("Data read from redirector file: $data");
                };
                if ($@ =~ /No such file or directory/) {

                    # file-read-file should fail
                    $Log->comment("File not created, Hardlink failed, $@");
                    die "Hardlink failed : $@";
                }
            } ## end else [ if ($raw_response != '')
            } ## end for (my $i=0; $i < $opt...
        } ## end foreach my $file (@{$opts{files...
    };

    if ($@) {
        $Log->comment("corrupt_data_link_count() failed: $@");
        die "corrupt_data_link_count() failed: $@";
    }
}    # corrupt_data_link_count

##--------------------------------------------------------
# Internal routines follow. These are not exposed outside
# the library
##---------------------------------------------------------

##---------------------------------------------------------
# _get_proto_string
#
# Get a string from the version and the transport protocol
# fields of the passed in NfsClient obj
##----------------------------------------------------------
sub _get_proto_string ( % ) {
    $Log->comment("_get_proto_string called with @_");

    my ($client_obj_ref) = @_;

    $Log->comment("$client_obj_ref");

    my $client_ver  = $client_obj_ref->version();
    my $client_prot = $client_obj_ref->protocol();
    my $prot_str    = $client_ver . '_' . $client_prot;

    return $prot_str;
} ## end sub _get_proto_string ( % )

##---------------------------------------------------------
# _create_files
#
# The internal routine that actually does the file create
##----------------------------------------------------------
sub _create_files ( % ) {
    $Log->comment("_create_files called with: @_");

    my %opts = validate(@_,
                        {client_obj => {type => HASHREF},
                         path       => {type => SCALAR},
                         count      => {type => SCALAR},
                         size       => {type => SCALAR},
                         fill       => {type => SCALAR},
                         cleanup    => {type => BOOLEAN},
                         prefix     => {type => SCALAR, optional => ''}});

    my $filer_conn     = $env_h{fc};
    my $client_obj_ref = $opts{client_obj};
    my $client_obj     = %$client_obj_ref;

    $Log->comment(
                 "client_obj: $client_obj and client_obj_ref: $client_obj_ref");

    my @curr_files;
    my $file_name;

    if ($opts{prefix} eq '') {
        $file_name = "cpx_" . _get_proto_string($client_obj_ref);
    } else {
        $file_name = $opts{prefix};
    }

    # something odd, just to check partial blocks (mainly test code)
    #my $attr = { size => $opts{size}, FATTR4_SIZE() => $opts{size} };
    my $version = $client_obj_ref->version();

    my $minor_ver =
      (defined $client_obj_ref->minor_version())
      ? $client_obj_ref->minor_version()
      : '0';

    my $protocol = $client_obj_ref->protocol();
    my $attr =
      $version == 4 ? {FATTR4_SIZE() => $opts{size}} : {size => $opts{size}};

    $Log->comment("Creating $opts{count} files with filename $file_name ");
    $Log->comment(
                "\t for client_obj ver:$version.$minor_ver protocol:$protocol");

    # TODO: What happens when the caller does not specify all the params
    @curr_files = file_make_with_client(client_obj => $client_obj_ref,
                                        path       => $opts{path},
                                        name       => $file_name,
                                        count      => $opts{count},
                                        attr       => $attr,
                                        fill       => $opts{fill},);

    # If cleanup flag is 1, add the file to an array of files that needs to be
    # cleaned up by posix_cleanup
    if ($opts{cleanup} == 1) {
        _add_files_to_cleanup($opts{path}, \@curr_files);
    }

    $Log->comment("Files created: @curr_files");

    $Log->comment("Validate ns/obj split");
    return
      _validate_files(client_obj => $client_obj_ref,
                      path       => $opts{path},
                      ns_files   => \@curr_files,
                      op         => '(CREATE): new files');
} ## end sub _create_files ( % )

#---------------------------------------------------------------------
# _validate_files
#
# Validate that a set of files on a namespace volume maps to a valid
# set of object files, and vice versa.
#---------------------------------------------------------------------
sub _validate_files( % ) {

    $Log->comment("_validate_files called with @_");

    my %opts = validate(@_,
                        {client_obj => {type => HASHREF},
                         path       => {type => SCALAR, optional => ''},
                         ns_files   => {type => ARRAYREF},
                         op         => {type => SCALAR}});

    my $filer_conn = $env_h{fc};

    $Log->comment(
                "Get the object files corresponding to the files just created");

##NFSv4 synthetic client is having mnt_path as NS and not /NS as in NFSv3 client.

    #    my @obj_files = _find_obj_files( filer_conn => $filer_conn,
    #                                    volume => $opts{client_obj}->mnt_path(),
    #                                    path => $opts{path},
    #                                    ns_files => $opts{ns_files} );

    my @obj_files = _find_obj_files(filer_conn => $filer_conn,
                                    volume     => $main::JUNCTION_PATH,
                                    path       => $opts{path},
                                    ns_files   => $opts{ns_files});

    $Log->comment("Make sure the count matches");
    my $validation = scalar_validate(op       => $opts{op},
                                     version  => $opts{client_obj}->version(),
                                     protocol => $opts{client_obj}->protocol(),
                                     expected => scalar(@{$opts{ns_files}}),
                                     returned => scalar(@obj_files));

    if ($validation !~ /validation succeeded/) {
        die fail_result($validation);
    }

    my (@ns_vol_obj, $ns_vol, @results);

    $Log->comment("Using Admin CLI: $Node->{vscli}");
    try {
        my $response =
          $Node->{vscli}->volume_show('constituent-role' => 'namespace');

        @ns_vol_obj = @{$response->get_parsed_output()};

    }
    catch NATE::BaseException with {
        my $ex = shift;
        die $ex->text();
    };

    $ns_vol = $ns_vol_obj[0]->{'volume'};

    for my $i (0 .. $#{$opts{ns_files}}) {

        #        push(@results, {
        #            client_obj => $opts{client_obj},
        #            ns_file => {
        #                           volume => $main::NAMESPACE_VOL,
        #                           'junction-path' => $opts{client_obj}->mnt_path(),
        #                           path => "$opts{path}" . "/${$opts{ns_files}}[$i]",
        #                           name => ${$opts{ns_files}}[$i]
        #                       },
        #            obj_file => $obj_files[$i],
        #        });

## TODO: Remove the below code once NFS mnt_path issue is fixed and uncomment the
## previous call.

        push(@results,
             {client_obj => $opts{client_obj},
              ns_file    => {
                   volume          => $main::NAMESPACE_VOL,
                   'junction-path' => $main::JUNCTION_PATH,
                   path =>
                     File::Spec->catfile($opts{path}, ${$opts{ns_files}}[$i]),
                   name => ${$opts{ns_files}}[$i],
                   fh =>
                     lookup_file_handle(
                       $ns_vol,
                       File::Spec->catfile($opts{path}, ${$opts{ns_files}}[$i]),
                       'msid'),},
              obj_file => $obj_files[$i],});

    } ## end for my $i (0 .. $#{$opts...

    return @results;
} ## end sub _validate_files( % )

#---------------------------------------------------------------------
# _find_obj_files
#
# Returns an array of object files corresponding to a set of namespace
# files
#---------------------------------------------------------------------
sub _find_obj_files( % ) {

    $Log->comment("_find_obj_files called with @_");

    my %opts = validate(@_,
                        {filer_conn => {type => HASHREF},
                         volume     => {type => SCALAR},
                         path       => {type => SCALAR, optional => ''},
                         ns_files   => {type => ARRAYREF}});

    # Get a list of all volumes
    my @vols;

    $Log->comment("Using Admin CLI: $Node->{vscli}");

    try {
        my $response = $Node->{vscli}->volume_show('is-constituent' => 'true');

        @vols = @{$response->get_parsed_output()};
    }
    catch NATE::BaseException with {
        my $ex = shift;
        die $ex->text();
    };

    # Find the namespace volume
    my $ns_vol = _volume_from_mnt_path($opts{volume}, @vols);

    return map {
        _find_obj_file($ns_vol, File::Spec->catfile($opts{path}, $_), @vols)
    } @{$opts{ns_files}};
} ## end sub _find_obj_files( % )

#---------------------------------------------------------------------
# _construct_data_file_name
#
# Helper to construct the data file name.
# Inputs:
#   1. Hex string representing 1 byte version.
#   2. Hex string representing 1 byte backpointer flags.
#   3. Hex string representing the redir_fh.
# Output:
#   base64 encoded backpointer name.
#---------------------------------------------------------------------
sub _construct_data_file_name( $$$ ) {
    $Log->comment("_construct_data_file_name called with @_");
    my ($redir_version, $bp_flags, $redir_fh) = @_;

    croak 'Invalid redir_version length: ' . $redir_version
      if (length($redir_version) != 2);
    croak 'Invalid bp_flags length: ' . $bp_flags if (length($bp_flags) != 2);
    croak 'Invalid redir_fh length: ' . $redir_fh if (length($redir_fh) != 40);

    my $bin_buf_chksum = pack "H2H2H*", $redir_version, $bp_flags, $redir_fh;
    my $hex_buf_chksum = unpack "H*", $bin_buf_chksum;
    my $sum                       = _compute_hbuf_checksum($hex_buf_chksum);
    my $bin_buf_sum_little_endian = pack "v",
      $sum;    # v: An unsigned short (16-bit) in "VAX" (little-endian) order.
    my $hex_buf_sum = unpack "H*", $bin_buf_sum_little_endian;
    my $bin_object_file_name = pack "H4H2H2H*", $hex_buf_sum, $redir_version,
      $bp_flags, $redir_fh;
    my $obj_file_name =
      encode_base64($bin_object_file_name, "");    # Don't put '\n' at the end
    $obj_file_name =~ tr/\//_/;    # Translate '/' characters to '_'.
    my $len = length($obj_file_name);
    $Log->comment("constructed object file name ($len bytes): $obj_file_name");
    return $obj_file_name;
} ## end sub _construct_data_file_name( $$$ )

#---------------------------------------------------------------------
# _find_obj_file
#
# Helper to find the object file for a single namespace file
#---------------------------------------------------------------------
sub _find_obj_file( $$@ ) {
    $Log->comment("_find_obj_file called with @_");

    my ($ns_vol, $ns_file, @vols) = @_;

    my $redir_fh = lookup_file_handle($ns_vol, $ns_file, 'origmsid');
    my ($version_hbuf, $obj_fh, $inode) = _read_redir_file($ns_vol, $ns_file);
    my $obj_file_name =
      _construct_data_file_name($version_hbuf, "00", $redir_fh);

    #my $obj_msid = unpack('Q', pack('H*', $obj_fh));
    # We only need the MSID
    # TODO: endian-ness
    my $obj_msid = unpack('II', pack('H*', $obj_fh));

    foreach my $vol (@vols) {
        $Log->comment(
                "Find the correct object volume by msid, check $vol->{volume}");
        $Log->comment("Object msid: $obj_msid and Volume msid: $vol->{msid}");
        if ($vol->{msid} eq $obj_msid) {
            my $sub_dir_path = _get_subdir_path($vol->{volume}, $inode);
            $Log->comment("Comparing object file name with path");
            _validate_obj_file_name($obj_file_name, $sub_dir_path);
            my $fh = lookup_file_handle($vol->{volume}, $sub_dir_path, 'msid');
            $Log->comment("Comparing object fh $obj_fh with $fh");
            if ($fh eq $obj_fh) {
                $Log->comment("Found match for $ns_file");
                return {volume          => $vol->{'volume'},
                        'junction-path' => $vol->{'junction-path'},
                        path            => "$sub_dir_path",
                        name            => $obj_file_name,
                        fh              => $obj_fh};
            }
        } ## end if ($vol->{msid} eq $obj_msid)
    } ## end foreach my $vol (@vols)

    $Log->comment("Did not find matching obj file for $ns_file");
    return;
} ## end sub _find_obj_file( $$@ )

#---------------------------------------------------------------------
# _reduce8
#
# Given a 64-bit value,
#
#   1. Add the two 32-bit words, add back the carry if any. This results
#      in a single 32-bit value.
#   2. Add the two 16-bit words, add back the carry if any. This results
#      in a single 16-bit value.
#   3. Add the two bytes, add back the carry if any. This results
#      in a single byte value.
#
# This logic assumes little-endian platform.
#---------------------------------------------------------------------
sub _reduce8( $ ) {
    my ($val) = @_;

    my $sum32 = $val;
    my $max32 = 0xffffffff;
    my $max16 = 0xffff;
    my $max8  = 0xff;

    if ($Config{longsize} == 8) {
        my $v32_0 = $val & $max32;
        my $v32_1 = ($val >> 32) & $max32;
        $sum32 = $v32_0 + $v32_1;
        if ($sum32 > $max32) {
            $sum32 -= $max32;
        }
    }

    my $v16_0 = $sum32 & $max16;
    my $v16_1 = ($sum32 >> 16) & $max16;
    my $sum16 = $v16_0 + $v16_1;
    if ($sum16 > $max16) {
        $sum16 -= $max16;
    }

    my $v8_0 = $sum16 & $max8;
    my $v8_1 = ($sum16 >> 8) & $max8;
    my $sum8 = $v8_0 + $v8_1;
    if ($sum8 > $max8) {
        $sum8 -= $max8;
    }

    return $sum8;
} ## end sub _reduce8( $ )

#---------------------------------------------------------------------
# _compute_hbuf_checksum
#
# Helper to compute checksum on a hex string.
# Input:
#   Hex string that looks like:
#   "f5200000c72ed12e010000000604008000000000000000006000000076e39419"
#   The length of this string must be divisible evenly by 4.
#   This string contains 16-bit unsigned short's in little endian format.
# Output:
#   Returns 16-bit checksum.
#---------------------------------------------------------------------
sub _compute_hbuf_checksum( $ ) {
    $Log->comment("_compute_hbuf_checksum called with @_");
    my ($hbuf) = @_;

    croak 'Invalid hex buffer length for checksum computation: ' . $hbuf
      if ((length($hbuf) % 2) != 0);

    my $bin_H = pack "H*", $hbuf;
    my @bin_C = unpack "C*", $bin_H;
    my $sum1  = 0xff;
    my $sum2  = 0xff;
    foreach $val (@bin_C) {
        $sum1 += $val;
        $sum2 += $sum1;
    }
    $sum1 = _reduce8($sum1);
    $sum2 = _reduce8($sum2);
    $sum2 <<= 8;
    $sum1 += $sum2;

    my $ret = (~$sum1) & 0xffff;             # Invert the bits
    my $sum_str = sprintf("0x%04x", $ret);
    $Log->comment("returning checksum: $sum_str for $hbuf");
    return $ret;
} ## end sub _compute_hbuf_checksum( $ )

sub _validate_hbuf( $ ) {
    my ($hbuf) = @_;

    $Log->comment("_validate_hbuf called with @_");

    croak 'Invalid hbuf ' . $hbuf
      if ($hbuf !~ /^([0-9a-fA-F]{4})([0-9a-fA-F]+)$/);
}

#-----------------------------------------------------------------------
#_validate_obj_file_name
#
#Helper to validate constructed object file name with sub directory path
#obtained from inode
#-----------------------------------------------------------------------
sub _validate_obj_file_name( $$ ) {
    my ($obj_file_name, $sub_dir_path) = @_;

    $Log->comment("_validate_obj_file_name called with @_");
    $index = index $sub_dir_path, $obj_file_name;
    if ($index > 0) {
        $Log->comment("Object Filename validation Passed: $obj_file_name");
    } else {
        croak 'Invalid Object Filename ' . $obj_file_name;
    }
} ## end sub _validate_obj_file_name( $$ )

#---------------------------------------------------------------------
# _validate_hbuf_checksum
#
# Helper to validate checksum on a hex string that includes the checksum.
# Input:
#   Hex string that looks like:
#   "f5200000c72ed12e010000000604008000000000000000006000000076e39419"
#   The length of this string must be divisible evenly by 4.
#   This string contains 16-bit unsigned short's in little endian format.
# Output:
#   None.
#   Fails if checksum validation fails.
#---------------------------------------------------------------------
sub _validate_hbuf_checksum( $ ) {
    my ($hbuf) = @_;

    _validate_hbuf($hbuf);

    my ($embedded_sum_hbuf, $the_rest_hbuf) = ($1, $2);

    croak 'Invalid hex buffer length for checksum validation: ' . $hbuf
      if ((length($hbuf) % 2) != 0);

    my $bin_embedded_sum = pack "H4", $embedded_sum_hbuf;
    my $embedded_sum = unpack "v", $bin_embedded_sum
      ;    # v: An unsigned short (16-bit) in "VAX" (little-endian) order.

    my $ret = _compute_hbuf_checksum($the_rest_hbuf);

    my $sum_info =
      sprintf("embedded: 0x%04x calculated:0x%04x", $embedded_sum, $ret);

    croak "Checksum validation failed: ($sum_info) for " . $hbuf
      if ($ret != $embedded_sum
          || ((~$ret + $embedded_sum) & 0xffff) != 0xffff);

    $Log->comment("Checksum validation PASSED ($sum_info) for $hbuf");
    return;
} ## end sub _validate_hbuf_checksum( $ )

#---------------------------------------------------------------------
# Pull the inode number out of the redir file content. First stuff the
# 64 byte readable string back into the 32 byte binary. Then get the
# inode number out of those 32 bytes. It is in bytes 24-27.
#---------------------------------------------------------------------
sub _get_inode_num( $ ) {
    my ($hbuf) = @_;

    $Log->comment("_get_inode_num called with @_");

    _validate_hbuf($hbuf);

    my $bin_hbuf = pack("H*", $hbuf);
    my ($before, $value, $after) = unpack("H48lH8", $bin_hbuf);
    my $inode = sprintf("%08x", $value);

    $Log->comment("_get_inode_num got inode from redir: $inode");
    return ($inode);
} ## end sub _get_inode_num( $ )

#---------------------------------------------------------------------
# Go get the path from the d-blade given the inode number. Once the path
# is returned strip the volume name off of it and return just subdir/file
#---------------------------------------------------------------------
sub _get_subdir_path( $$ ) {
    my ($volume, $inode) = @_;

    $Log->comment(
              "_get_subdir_path called with volume: $volume and inode: $inode");

    # Get the node corresponding to a volume
    my $response =
      $Node->{vscli}->volume_show(volume => $volume,
                                  field  => "node");
    my @vols = @{$response->get_parsed_output()};
    my $node = @vols[0]->{'node'};

    my $getpathfrominode =
      "node run -node " . $node . " -command inodepath -v $volume 0x$inode";
    my $output =
      $Node->{cli}->execute_raw_command(command => $getpathfrominode);

    $Log->comment("$output");

    if ($output !~ /.*Primary*/) {
        die "get_subdir_path failed!";
    }

    # getting the file path after "Primary pathname = <fullpath>"
    $output =~ m{.*Primary pathname = (/.*)\s};
    my $fullpath = $1;

    # Strip out /vol/<constituent>
    $fullpath =~ m{/vol/\w+(.*)};
    my $pathname = $1;
    $Log->comment("Full file path: $fullpath");
    $Log->comment("Returning path: $pathname");
    return ($pathname);
} ## end sub _get_subdir_path( $$ )

#---------------------------------------------------------------------
# _read_redir_file
#
# Helper to read the redirector file contents to get the object filehandle.
# Returns an array containing a hex string representing 1 byte version,
#                  followed by the hex string representing the obj_fh.
#---------------------------------------------------------------------
sub _read_redir_file( $$ ) {
    $Log->comment("_read_redir_file called with @_");
    my ($vol, $file) = @_;

    my $path = File::Spec->catfile('/vol', $vol, $file);
    my $result;

    # Adding the logic to retry the zapi if the volume is busy for some reason
    # We have seen this happening almost frequently on Smokes especially
    # after vol move or if the volume is close to saturation

    # Explanation: The file-read-file ZAPI is implemented using ClusFS/SpinVFS,
    # and SpinVFS apparently returns EJUKEBOX (which the zapi layer translates
    # to EVOLUMEBUSY) on the first access to a file after a volume move.
    # Refer burt548444

    my $emsg = 'Volume is busy';
    for (1 .. 5) {
        eval {
            $result = _invoke_zapi('file-read-file', 'path', $path, 'offset', 0,
                                   'length', 32);
            last;
        };

        if ($@) {
            $Log->comment("$_: Unable to invoke zapi: $@");
            die $@ if $@ !~ /$emsg/;
        }

    } ## end for (1 .. 5)
    my $redir = $result->child_get_string('data');

    # Validate the length (32 hex bytes, 64 characters).
    if (length($redir) != 64) {
        die fail_result('Redirector content length is not 64 bytes: ' . $redir);
    }
    _validate_hbuf_checksum($redir);
    my $inode = _get_inode_num($redir);

    # Redir format is checksum-magic-version-flags-redir
    # Version:1
    if ($redir !~ /^(.{4})0000c72ed12e(.{2})000000(.{40})$/) {
        die fail_result('Invalid redirector content ' . $redir);
    }
    my ($redir_version, $obj_fh) = ($2, $3);

    $Log->comment(
        "returning version $redir_version and obj file handle: $obj_fh and inode: $inode"
    );
    return ($redir_version, $obj_fh, $inode);
} ## end sub _read_redir_file( $$ )

#---------------------------------------------------------------------
# lookup_file_handle
#
# Helper to determine the file handle for a specified file.
# The $dsid parameter specifies whether to use the 'dsid' or 'msid'
#
# use lookup_file_handle, not lookup_file_handle_
# the zapi does not return flags
#---------------------------------------------------------------------
sub lookup_file_handle_ ( $$$ ) {

    $Log->comment("lookup_file_handle called with @_");
    my ($vol, $file, $dsid) = @_;

    my $info = _lookup_file_info($vol, $file);
    $Log->comment("_lookup_file_info returned: $info");

    # TODO object file name seems to use native
    # endian; what if client endianness is different?
    my $dsid_string =
      unpack('H*', pack('II', ($info->child_get_string($dsid))));

    $Log->comment("$dsid: $dsid_string");

    my $fh_string = unpack(
        'H*',
        pack(
            'III',
            0,                                       # TODO flags ever non-zero?
            $info->child_get_string('inode-number'),
            $info->child_get_string('inode-gen-number')));

    $Log->comment("Returning fh: $dsid_string $fh_string");
    return "${dsid_string}${fh_string}";
} ## end sub lookup_file_handle_ ( $$$ )

#---------------------------------------------------------------------
# _lookup_file_info
#
# Helper to lookup a file's attributes using ZAPI.
#---------------------------------------------------------------------
sub _lookup_file_info( $$ ) {
    $Log->comment("_lookup_file_info called with @_");
    my ($vol, $file) = @_;

    my $path = '/vol/' . $vol . $file;

    my $result = _invoke_zapi('file-get-file-info', 'path', $path);
    return $result->child_get('file-info');
}

#---------------------------------------------------------------------
# _volume_from_mnt_path
#
# Helper to lookup the name of a volume from its mount point path.
#---------------------------------------------------------------------
sub _volume_from_mnt_path( $@ ) {
    $Log->comment("_volume_from_mnt_path called with @_");
    my ($mp, @vols) = @_;

    foreach my $vol (@vols) {
        if ($vol->{'junction-path'} eq $mp) {
            return $vol->{volume};
        }
    }
    return;
} ## end sub _volume_from_mnt_path( $@ )

#---------------------------------------------------------------------
# _invoke_zapi
#
# Helper to invoke a ZAPI call
#---------------------------------------------------------------------
sub _invoke_zapi {

    #XXX: Use network interface show to get the mgmt IP
    my $hostr = host($main::FILER);
    my $zapi = new ZapiTest($hostr->{default_ip}, 'admin', 'netapp1!');
    $zapi->set_legacy_http_conn(0);
    $zapi->set_vfiler_server($main::VSERVER_GX_NAME);

    # Check the result as well
    my ($result) = $zapi->invoke(@_);
    $Log->comment("ZAPI Result : " . $result);

    if ($result->attr_get('status') ne 'passed') {
        die fail_result($result->attr_get('reason'));
    }
    return $result;
} ## end sub _invoke_zapi

#---------------------------------------------------------------------
# _get_aggregate_with_free_space
#
# Params:
# - free_space: This is the free space that should be available
#               in the aggr.
#
# This function can throw many kinds of failures. In case of failure
# the returned aggregate name would be undef
#---------------------------------------------------------------------
sub _get_aggregate_with_free_space {
    my ($free_space) = @_;
    my $aggrName = undef;
    my $r;

    my $avl_aggr_show;

    $Log->comment("_get_aggregate_with_free_space called with @_");
    $Log->comment("Using admin CLI object: $Node->{cli}");

    try {

        $aggrName = ${
            $Node->{cli}->storage_aggregate_show(aggregate => 'ag_test_*',
                                                 availsize => $free_space,
                                                 state     => 'online'
              )->get_parsed_output()}[0]->{aggregate};

    }
    catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        my $rand_val = int(rand(1000000));
        my $aggrNode;

        # If we found no aggregate that we can use, create a new one
        my @parsed_output =
          @{$Node->{cli}->system_node_show()->get_parsed_output()};
        my @nodes = map $_->{node}, @parsed_output;

        foreach my $node (@nodes) {
            $Log->comment("Attempting to find disks from node $node");
            $r =
              $Node->{cli}->storage_disk_show("state" => "spare",
                                              home    => $node);

            $Log->comment(  "Node $node has "
                          . scalar(@{$r->get_parsed_output()})
                          . " spare disks");

            if (scalar(@{$r->get_parsed_output()}) >= 2) {
                $aggrNode = $node;
            }
        } ## end foreach my $node (@nodes)

        if (!$aggrNode) {
            die "Must have at least one node with 2 spare disks";
        }

        # Create aggregate
        $aggrName = "ag_test_$rand_val";
        $Test->step("Making an aggregate with 2 disks");
        $Node->{cli}->storage_aggregate_create(aggregate => $aggrName,
                                               nodes     => $aggrNode,
                                               diskcount => 2,
                                               raidtype  => "raid4");

        # Aggregate create takes some time to complete
        for (my $i = 0 ; $i < 60 ; $i++) {
            $r = $Node->{cli}->storage_aggregate_show(aggregate => $aggrName);
            if (${$r->get_parsed_output()}[0]->{state} eq "online") {
                last;
            }
            snooze(5);
        }

        if (${$r->get_parsed_output()}[0]->{state} ne "online") {
            die "Failed to create $aggrName in time";
        }
    };

    return $aggrName;
} ## end sub _get_aggregate_with_free_space

#---------------------------------------------------------------------
# _remove_file_from_cleanup
#
# Params:
# - file name: file name to remove from cleanup.
#
# May not know if file exists, so no error if not present
#---------------------------------------------------------------------
sub _remove_file_from_cleanup($) {
    my $rm_file = shift;
    $rm_file =~ s/\A(\/*)//;

    #remove the file from the list of files
    my $i = 0;
    foreach my $file_name (@all_files) {
        if ($rm_file == $file_name) {
            splice(@all_files, $i, 1);
            last;
        }
        ++$i;
    }
} ## end sub _remove_file_from_cleanup($)

sub _add_files_to_cleanup($@) {
    $path  = shift;
    $array = shift;

    $path .= '/';
    $path =~ s/\A\/*//;

    for $file_name (@$array) {
        push(@all_files, $path . $file_name);
    }
} ## end sub _add_files_to_cleanup($@)

#---------------------------------------------------------------------
# flush_fhmap
#
# Helper to flush the FHMAP on the Node whose Data IP is provided
#
# Arguments:
# data_ip (Required) - Data IP of the node from which FHMAP is to be flushed
#
# Usage:
# flush_fhmap(10.53.20.215)
#---------------------------------------------------------------------
sub flush_fhmap {

    # Declaring variables specific to the subroutine
    my ($data_ip, $prompts_answers, $api_set_obj, $data_ip_node);

    ($data_ip) = @_;

    # Getting the DataIP node
    $data_ip_node = get_data_ip_node($data_ip);

    # Checking if we already have APISet connection object to the required node
    $api_set_obj = get_node_api_cli_obj($data_ip_node);

    # Flushing the FHMAP
    $Test->nlog("Flushing the FHMAP from the node: '$data_ip_node'");
    $prompts_answers =
      ['.*This command will remove the entire fhmap cache on node*' => 'y'];
    $api_set_obj->diag_nblade_repository_fhmap_cache_clear_all(
                                    'node'                   => $data_ip_node,
                                    'connectrec-match_table' => $prompts_answers
    );

} ## end sub flush_fhmap

#---------------------------------------------------------------------
# get_node_api_cli_obj
#
# Helper to check if we already have a defined APISet CLI object to a node.
# If not then create it and set it to the specified node
#
# Arguments:
# Node Name (Required) - Node name whose APISet CLI object is to be established
#
# Usage:
# get_node_api_cli_obj(foobar-vsim1)
#
# Returns:
# APISet CLI object to the specified node
#---------------------------------------------------------------------
sub get_node_api_cli_obj {
    my ($data_ip_node) = @_;
    if (!$data_ip_node) {
        croak "Node name is required in order get/set APISet CLI object";
    }

    # Checking if we already have APISet connection object to the required node
    if (exists($Api_set_nodes->{$data_ip_node . '_cli'})) {
        $Test->step(
            "Using the already defined APISet connection object to the node - '$data_ip_node'"
        );
    } elsif ((($Node->{cli}->{hostobj}->{hostname}) ne $data_ip_node)) {
        $Test->step("Creating a API set CLI object to the Node: $data_ip_node");
        $Api_set_nodes->{$data_ip_node . '_cli'} =
          NACL::APISet->new(hostobj   => host($data_ip_node),
                            category  => "Node",
                            interface => "CLI",
                            set       => "CMode",);
    } else {
        $Test->step(
            "Using the existing APISet connection object to the node - '$data_ip_node' created during setup"
        );
        $Api_set_nodes->{$data_ip_node . '_cli'} = $Node->{cli};
    }

    return $Api_set_nodes->{$data_ip_node . '_cli'};

} ## end sub get_node_api_cli_obj

#---------------------------------------------------------------------
# get_data_ip_node
#
# Helper to get the node-name for the data ip that is provided as input
#
# Arguments:
# DataIP (Required) - DataIP using which the Node-name is to be calculated
#
# Usage:
# get_data_ip_node(192.168.0.2)
#
# Returns:
# The node-name having the input DataIP
#---------------------------------------------------------------------
sub get_data_ip_node {
    my ($data_ip) = @_;

    # This will get assigned for only the node which matches the Data IP
    # Or this will get initialized to 0 if no network interface found
    my $data_lif_obj =
      NACL::C::NetworkInterface->find(command_interface => $Node->{nodes}->[0],
                                      filter => {address => $data_ip},);
    return $data_lif_obj->get_one_state_attribute('curr-node');
} ## end sub get_data_ip_node

#------------------------------------------------------------------------------
# get_iv_on_repos_vserver_with_default_dcg
#
# Helper to get the Infinite Vol object and corresponding repository
# vserver object having default DCG
#
# Arguments:
# Node Object (Required) - Command Interface node object to connect to the
# cluster
#
# Usage:
# get_iv_on_repos_vserver_with_default_dcg($Node_Obj)
#
# Returns:
# Reference to array containing IV obj with default DCG & repository vserver obj
# If no Infinite Vol was found, it will return 0
#-------------------------------------------------------------------------------
sub get_iv_on_repos_vserver_with_default_dcg {

    my $node_obj = shift;

    my @repo_vserver_objs =
      NACL::C::Vserver->find(command_interface => $node_obj,
                             filter            => {'is-repository' => 'true',});

    my $iv_obj = 0;
    foreach my $vserver_obj (@repo_vserver_objs) {
        $iv_obj = NACL::C::Volume->find(command_interface => $node_obj,
                                        filter            => {
                                             vserver => $vserver_obj->vserver(),
                                             'volume-style' => 'infinitevol',
                                             'is-managed-by-service' => 'false',
                                        },
                                        allow_empty => 1,);

        if ($iv_obj) {
            return [$iv_obj, $vserver_obj];
        }
    } ## end foreach my $vserver_obj (@repo_vserver_objs)

    $Log->error('IV on a repository vserver having default DCG not found.');
    return 0;
} ## end sub get_iv_on_repos_vserver_with_default_dcg

#------------------------------------------------------------------------------
# reconnect_nfs_client_connections
#
# Helper to reconnect NFS client connections when TCP connection is reset in
# scenario of LIF is brought down or the LIF failover happens or the LIF is
# intentionally migrated to other node.
#
# Arguments:
# None
#
# Usage:
# reconnect_nfs_client_connections();

# Returns:
# 1
#-------------------------------------------------------------------------------
sub reconnect_nfs_client_connections {
    foreach my $client_obj (@{$env_h{nfs_client_objs}}) {
        $client_obj->{object}->{rpc}->reconnect();
        $client_obj->mount(path => $NAMESPACE);
    }

    return 1;
}

#------------------------------------------------------------------------------
# wait_for_grace_period()
#
# Helper to wait for GRACE period to end. It tries to create a file through
# NFSv4 OPEN. If it gets NFS4ERR_GRACE then it retries after sleeping for
# 30 seconds. It tries this process until it succeeds but not more than 5 times.
# If it is able to create a file then it closes the file and pushes it for
# cleanup.
#
# Arguments:
# nfsv4_client_obj (Optional): NFSv4 client object to be used for creating the
# test file. Default is NFSv4 client object created by posix_setup
#
# dir_fh (Optional): Directory FH where the test file needs to be created.
# Default is mount FH fetched from NFSv4 client object created by posix_setup
#
# Usage:
# wait_for_grace_period();
# or
# wait_for_grace_period(
#     nfsv4_client_obj => $client_obj,
#     dir_fh => directory_filehandle,
# );
#
# Returns:
# 1 if the file was created and GRACE period was ended.
# -1 if the file could not be created because GRACE period was still going on.
# 0 if the file creation failed for some unknown reason.
#-------------------------------------------------------------------------------
sub wait_for_grace_period {

    my %opts = @_;

    $Log->comment('Reconnecting NFS client connections');
    reconnect_nfs_client_connections();

    my ($client_obj) = $opts{nfsv4_client_obj}
      || grep { $_->version() == 4 } @{$env_h{nfs_client_objs}};
    my $dhandle = $opts{dir_fh} || $client_obj->{object}->{mnt_fh};
    my $file_name = 'check_grace_period_' . time();

    my $open_op_status;
    for (1 .. 5) {

        try {
            my $open_op_response =
              $client_obj->open(name    => $file_name,
                                cr_mode => 2,
                                dhandle => $dhandle,
                                attr    => {FATTR4_MODE() => 777});

            $open_op_status = 1;
            $client_obj->close(stateid => $open_op_response->{stateid});
            push @{$Node->{cleanup_files}}, $file_name;

        } ## end try
        catch NATE::BaseException with {

            my $err_msg = shift;
            if ($err_msg =~ /NFS4ERR_GRACE/) {
                $Log->comment(
                            'NFS server is in GRACE period and hence retrying');
                sleep 30;
                $open_op_status = -1;
            } else {
                $Log->warn("OPEN failed for unknown reason: $err_msg");
                $open_op_status = 0;
            }

        };

        # Exiting from the loop after catching the expected error code
        last if ($open_op_status != -1);
    } ## end for (1 .. 5)

    # Checking the status of open op
    if ($open_op_status == -1) {
        $Log->warn('Server is still in GRACE period');
    }

    return $open_op_status;

} ## end sub wait_for_grace_period

##--------------------------------------------------------
# _get_ordinal_volume_name
#
# This helper function finds the ordinal-value for the
# supplied volume. If the value is numeric then it
# builds the ordinal name and returns it.
# If the value is not numeric i.e either it is 'base' or '-'
# then it doesn't do anything and returns the existing value
# w/o changing anything.
#
# Arguments:
# - volume name     : Flex vol name
# Notes:
# This function is only applicable for FlexVols
##----------------------------------------------------------
sub _get_ordinal_volume_name {

    my ($vol_name) = @_;

    my $vol_obj = NACL::C::Volume->find(
                                      command_interface => ${$Node->{nodes}}[0],
                                      filter            => {
                                                 vserver => $VSERVER,
                                                 volume  => $vol_name,},);

    my $vol_orginal_name = $vol_obj->get_one_state_attribute('name-ordinal');

    # 'name-ordinal' can have following valid values and we are interested only
    # when it's numeric:
    # 1. base
    # 2. -
    # 3. Numeric
    if ($vol_orginal_name =~ /\d+/) {
        $vol_name = $vol_name . "($vol_orginal_name)";
    }

    return $vol_name;
} ## end sub _get_ordinal_volume_name

#------------------------------------------------------------------------------
# refresh_nfs_client
#
# Helper to refresh the client whenever the client has expired, so that it
# renews the expired object.
#
# Arguments:
# Client obj to be refreshed
#
# Usage:
# reconnect_nfs_client_connections( $obj );
# Returns:
#-------------------------------------------------------------------------------
sub refresh_nfs_client {
    my $client_obj = shift;
    my $exp_code;

    $Log->comment('Refreshing the nfs client object');
    if ($client_obj->version() == 4 && $client_obj->minor_version() == 1) {
        eval { $client_obj->sequence(); };

        $exp_code = 'NFS4ERR_BADSESSION|NFS4ERR_STALE_CLIENTID';

        if ($@ =~ /$exp_code/) {
            eval {
                $client_obj->exchange_id();
                $client_obj->create_session();
                $client_obj->sequence();
            };
            $Test->nlog("Fail to create session for nfsv4.1 obj : $@") if ($@);
        }
    } else {
        eval { $client_obj->renew(); };
        $exp_code = 'NFS4ERR_STALE_CLIENTID';

        if ($@ =~ /$exp_code/) {
            eval {
                $client_obj->setclientid();
                $client_obj->setclientid_confirm();
            };
            $Test->nlog("Fail to setclientid for nfsv4 obj: $@") if ($@);
        }
    } ## end else [ if ($client_obj->version...

}    ## refresh_nfs_client

#------------------------------------------------------------------------------
# verify_storage_service_for_files
#
# Helper to verify given set of files belong to expected DCG.
#
# Arguments:
# files - Array ref containing files to be verified.
# path - Path including the junction path where the files are present.
# service - Expected storage service the files should belong to.
# volume_name - The InfiniteVol name (IV) containing the files.
# vserver_name - The vserver name to which the IV belongs.
#
# Usage:
# verify_storage_service_for_files(
#     files => \@thpl_files,
#     path => $iv_jp,
#     service => 'gold',
#     volume_name => $Infinite_Vol_Obj->volume(),
#     vserver_name => $Infinite_Vol_Obj->vserver(),
# );
# Returns:
#-------------------------------------------------------------------------------
sub verify_storage_service_for_files{

    my %opts = validate(@_,
        {files   => {type => ARRAYREF, optional => 0},
         path   => {type => SCALAR, optional => 0,},
         service  => {type => SCALAR, optional => 1,
                       default => 'default'},
         volume_name   => {type => SCALAR, optional => 0},
         vserver_name  => {type => SCALAR, optional => 1,
                       default => $VSERVER}},
    );

    my @test_files = @{$opts{files}};
    foreach my $file ( @test_files ) {

        $file =  ( $opts{path} ) ? ( $opts{path} . q{/} . $file )
                                 : ( $file );

        my $resp = $Node->{cli}->volume_storage_service_locate_file(
            'volume'   => $opts{volume_name},
            'vserver'  => $opts{vserver_name},
            'filename' => $file,
        );

        my $raw_output = $resp->get_raw_output();
        if ( $raw_output !~ /is in storage service "$opts{service}"/ ) {
            $Log->error("File $file is not present in storage service"
                          . $opts{service});
            return 0;
        }
    }

    return 1;

}

##-------------------------------------------------------------
# set_systemnodedate_utcdateandtime
# Modify the system node time
##--------------------------------------------------------------
sub set_systemnodedate_utcdateandtime {
    my (%params) = validate(@_,
                           {cluster_date_obj   => {type => HASHREF,
                                                   optional => 0},
                            node_obj           => {type => OBJECT,
                                                   optional => 1},
                            increase_hour_by   => {type => SCALAR,
                                                   optional => 1},});

    my $epoch_time = get_system_utcdateandtime_as_epoch_time(
                         cluster_date_obj => $params{cluster_date_obj});
    $Test->nlog("Epoch time : $epoch_time");
    if ($params{increase_hour_by}){
        $epoch_time += int ($params{increase_hour_by} * 60 * 60);
    }
    $Test->nlog("Modified Epoch time : $epoch_time");
    my $utc_date_and_time = POSIX::strftime('%Y%m%d%H%M.%S',
                                            localtime($epoch_time));

    $Test->nlog("System time : UTC Date and Time : $utc_date_and_time");

    NACL::C::ClusterDate->modify(command_interface => $params{node_obj},
                                 utcdateandtime => $utc_date_and_time,
                                 );
    return 1;
}

##-------------------------------------------------------------
# set_systemnodedate_utcdateandtime
# Return the system node time
##--------------------------------------------------------------
sub get_system_utcdateandtime_as_epoch_time {
    my (%params) = validate(@_,
                            {cluster_date_obj   => {type => HASHREF,
                                                    optional => 0},
                             command_interface  => {type => HASHREF,
                                                    optional => 1},
                             node               => {type => SCALAR,
                                                    optional => 1},});

    my $sys_utc_time = $params{cluster_date_obj}->state()->utc_date();
    $Test->nlog("System UTC time : $sys_utc_time");
    $sys_utc_time =~ s/\"//g;
    my ($mon,$mday,$year,$hour,$min,$sec) = split(/\D/,$sys_utc_time);
    $Test->nlog("Mon : $mon Day: $mday Year : $year Hour : $hour Min: $min"
              . " Sec: $sec");
    return timelocal($sec,$min,$hour,$mday,int ($mon - 1),$year);
}

##--------------------------------------------------------------
# Helper function to extract the vserver id given vserver name
##--------------------------------------------------------------
sub get_vserver_id_from_name
{
    my %opts = validate(@_,
                       {vserver_name   => {type    => SCALAR,
                                           default => $VSERVER},
                        node_obj       => {type    => OBJECT}});
    my $smdb_vserver_obj = CR::C::DebugSmdbTableVserverIdToName->find(
                               command_interface => $opts{node_obj},
                               filter         => {name => $opts{vserver_name}});
    return $smdb_vserver_obj->get_one_state_attribute('id');
}

##-------------------------------------------------------------
# Verify that a Repository.NS.Dangling.Lock EMS message was
# generated.
# Arguments:
#    - seqnum: Sequence number of the EMS message
#    - node_obj: Node object to use as command interface for
#                fetching EMS event
# Returns:
#    TCD::PASS if the event with given sequence number
#    is a Repository.NS.Dangling.Lock event, else return
#    TCD::FAIL.
##--------------------------------------------------------------
sub verify_NS_dangling_lock_event
{
    my %opts = validate(@_,
                       {seqnum   => {type    => SCALAR,},
                        node_obj => {type    => OBJECT}});

    my $event = NACL::CS::EventLog->fetch(
                    command_interface => $opts{node_obj},
                    filter            => { seqnum => $opts{seqnum} } );

    my $smdb_vserver_id = get_vserver_id_from_name(node_obj => $opts{node_obj});
    my $event_msg = $event->event();

    my $ns_dangling_lock = 'Repository.NS.Dangling.Lock';
    my $begin_msg = "$ns_dangling_lock: Failures prevented the cleanup of a"
                  . ' transient lock with UUID';
    my $lock_uuid = '[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}';
    my $end_msg = "on the namespace constituent in Vserver $smdb_vserver_id.";

    $Test->nlog("Expected : $begin_msg $lock_uuid $end_msg");
    $Test->nlog("Actual : $event_msg");

    if ($event_msg !~ m/$begin_msg $lock_uuid $end_msg/){
        $Log->warn->nlog("Event failed to match expected: $begin_msg $lock_uuid $end_msg"
                . ", found $event_msg"
        );
    }

    if($event_msg !~ m/$ns_dangling_lock/) {
        $Test->nlog("Event failed to match expected $ns_dangling_lock"
. ", found $event_msg");
        return $TCD::FAIL;
    }
    return $TCD::PASS;
}

##-------------------------------------------------------------
# Verify that a Repository.DC.Dangling.Lock EMS message was
# generated.
# Arguments:
#    - seqnum: Sequence number of the EMS message
#    - node_obj: Node object to use as command interface for
#                fetching EMS event
# Returns:
#    TCD::PASS if the event with given sequence number
#    is a Repository.DC.Dangling.Lock event, else return
#    TCD::FAIL.
##--------------------------------------------------------------
sub verify_DC_dangling_lock_event
{
    my %opts = validate(@_,
                        {seqnum   => {type   => SCALAR,},
                         node_obj => {type   => OBJECT}});

    my $event = NACL::CS::EventLog->fetch(
                    command_interface => $opts{node_obj},
                    filter            => { seqnum => $opts{seqnum} } );
    my $smdb_vserver_id = get_vserver_id_from_name(node_obj => $opts{node_obj});
    my $event_msg = $event->event();
    my $dc_dangling_lock = 'Repository.DC.Dangling.Lock';
    my $begin_msg = "$dc_dangling_lock: Failures prevented the cleanup of a"
                  . ' lock with UUID';
    my $lock_uuid = '[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}';
    my $end_msg = "on the data constituent in Vserver $smdb_vserver_id.";

    $Test->nlog("Expected : $begin_msg $lock_uuid $end_msg");
    $Test->nlog("Actual : $event_msg");

    if ($event_msg !~ m/$begin_msg $lock_uuid $end_msg/){
        $Log->warn->nlog("Event failed to match expected: $begin_msg $lock_uuid $end_msg"
                . ", found $event_msg"
        );
    }

    if($event_msg !~ m/$dc_dangling_lock/) {
        $Test->nlog("Event failed to match expected $dc_dangling_lock"
. ", found $event_msg");
        return $TCD::FAIL;
    }
    return $TCD::PASS;
}

return $TCD::TRUE;

#
# Local Variables:
# indent-tabs-mode:nil
# tab-width:4
# End:
#
