#
# $Id
#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Content Repository Class
## @author ngok@netapp.com
## @status shared
## @pod here

package CR::Setup;

use strict;
use warnings;

use Data::Dumper;
use Params::Validate qw(:all); # for SCALAR
use Tharn qw($Log);
use NACL::APISet;
use NACL::C::Interface;
use NACL::C::Route;
use NACL::C::VserverServicesNisDomain;
use NACL::C::VserverNfs;
use NACL::C::ExportPolicyRule;
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use CR::UniqueName;

BEGIN {
    use Exporter ();
    our (@ISA, @EXPORT);
    @ISA = qw(Exporter CR);
    @EXPORT = qw(setup_dataports_from_nate_host setup_dataport setup_vserver);
}

############################################################
#     setup dataport
############################################################

=head2 setup_dataports_from_nate_host

Setup the dataports from information in the nate_host file.

=cut

sub setup_dataports_from_nate_host {
    $Log->enter();
    my $self = shift;
    my @nodes = NACL::C::Node->find(command_interface => $self->{node_obj});

    foreach my $node_obj (@nodes) {
	my $node = $node_obj->state()->node();
	my $host = Tharn::host($node);
	my %args;

	# a data port
	if (defined($host->DATA_PORT) && defined($host->DATA_IP)) {
	    %args = ('home-node' => $node,
		     'home-port' => $host->DATA_PORT,
		     address => $host->DATA_IP);
	    $args{netmask} = $host->DATA_MASK if $host->DATA_MASK;
	    $args{gateway} = $host->DATA_GATEWAY if $host->DATA_GATEWAY;
	    $self->setup_dataport(%args);
	}

	# data port 1
	if (defined($host->DATA_PORT1) && defined($host->DATA_IP1)) {
	    %args = ('home-node' => $node,
		     'home-port' => $host->DATA_PORT1,
		     address => $host->DATA_IP1);
	    $args{netmask} = $host->DATA_MASK1 if $host->DATA_MASK1;
	    $args{gateway} = $host->DATA_GATEWAY1 if $host->DATA_GATEWAY1;
	    $self->setup_dataport(%args);
	}

	# data port 2
	if (defined($host->DATA_PORT2) && defined($host->DATA_IP2)) {
	    %args = ('home-node' => $node,
		     'home-port' => $host->DATA_PORT2,
		     address => $host->DATA_IP2);
	    $args{netmask} = $host->DATA_MASK2 if $host->DATA_MASK2;
	    $args{gateway} = $host->DATA_GATEWAY2 if $host->DATA_GATEWAY2;
	    $self->setup_dataport(%args);
	}
    }
    $Log->exit();
}

=head2 setup_dataport

Set up the data port on the specified node with its IP address,
netmask, and gateway.

=over 4

=item *
home-node

The node on which the data port resides. (mandatory)
This is the name of the HostRec.

=item *
home-port

The port id (mandatory)
(such as 'e0d')

=item *
address

The IP address of the port. (mandatory)

=item *
netmask

The netmask of the port. (optional)
If it is not given,
it is assumed 255.255.255.0.

=item *
gateway

The gateway of the subnet to which the IP address belongs. (optional)
If it is not given,
it is assumed to be host id 1 of the subnet.
For example, IP address 172.17.121.16 netmask 255.255.255.0
will yield gateway address 172.17.121.1.

=back

=cut

sub setup_dataport {
    $Log->enter();
    my ($self, %args) =	@_;

    %args = NACL::APISet::validate_params($self,
		 params => \%args,
		 spec => {
		     lif => { type => SCALAR, optional => 1 },
		     'home-node' => { type => SCALAR, optional => 0 },
		     'home-port' => { type => SCALAR, optional => 0 },
		     address => { type => SCALAR, optional => 0 },
		     netmask => { type => SCALAR, optional => 1 },
		     gateway => { type => SCALAR, optional => 1 },
		 },
	     );

    # check if address is already used in a dataport
    my $port = NACL::C::Interface->find(command_interface => $self->{node_obj},
					filter => {vserver => $self->{vserver},
						   address => $args{address}},
					allow_empty => 1);
    if (defined($port)) {
	if ($self->{new_dataports}) {
	    # delete existing dataport having the identical address
	    $port->delete();
	    # proceed to create dataport
	}
	else {
	    return; # keep existing dataport
	}
    }

    if (!defined($args{netmask})) {
	$args{netmask} = '255.255.255.0';
    }
    if (!defined($args{lif})) {
	$args{lif} = unique_name('lif');
    }

    my $gateway;
    if (!defined($args{gateway})) {
	$gateway = _gateway($args{address}, $args{netmask});
    }
    else {
	$gateway = delete $args{gateway};
    }

    if (defined($self->{lif_auto_revert})) {
	$args{'auto-revert'} = $self->{lif_auto_revert} ? 'true' : 'false';
    }

    NACL::C::Interface->create(command_interface => $self->{node_obj},
			       server => $self->{vserver},
			       role => 'data',
			       %args,
			       ); # No exception if duplicate

    #IMPORTANT: Assumed routing groups were created during LIFs creation
    # Setup routing group route
    my $routing_group = _data_routing_group($args{address}, $args{netmask});
    try {
	NACL::C::Route->create(command_interface => $self->{node_obj},
			       server => $self->{vserver},
			       'routing-group' => $routing_group,
			       gateway => $gateway,
			       destination => "0.0.0.0/0");
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
	my $ex = shift;
	if ($ex->text() !~ /duplicate entry/) {
	    $ex->throw();
	}
	$Log->comment("NACL::APISet::Exceptions::CommandFailedException '".
		      $ex->text()."' caught and ignored.");
    };
    $Log->exit();
}

my @netmask_table = ("0.0.0.0",
		     "128.0.0.0",
		     "192.0.0.0",
		     "224.0.0.0",
		     "240.0.0.0",
		     "248.0.0.0",
		     "252.0.0.0",
		     "254.0.0.0",
		     "255.0.0.0",
		     "255.128.0.0",
		     "255.192.0.0",
		     "255.224.0.0",
		     "255.240.0.0",
		     "255.248.0.0",
		     "255.252.0.0",
		     "255.254.0.0",
		     "255.255.0.0",
		     "255.255.128.0",
		     "255.255.192.0",
		     "255.255.224.0",
		     "255.255.240.0",
		     "255.255.248.0",
		     "255.255.252.0",
		     "255.255.254.0",
		     "255.255.255.0",
		     "255.255.255.128",
		     "255.255.255.192",
		     "255.255.255.224",
		     "255.255.255.240",
		     "255.255.255.248",
		     "255.255.255.252",
		     "255.255.255.254",
		     "255.255.255.255");
sub _netmask_to_bitmask {
    my $netmask = shift;
    for my $i (0..$#netmask_table) {
	if ($netmask eq $netmask_table[$i]) {
	    return $i;
	}
    }
    CR::Exceptions::InvalidNetmask->throw("Netmask $netmask is invalid");
}

sub _network_id {
    my ($address, $netmask) = @_;
    my @address = split '\.', $address;
    my @netmask = split '\.', $netmask;
    my @result;
    for my $i (0..3) {
	$result[$i] = ($address[$i]+0) & ($netmask[$i]+0); # Perl bug workaround
    }
    return @result;
}

sub _data_routing_group {
    my ($address, $netmask) = @_;
    my @network = _network_id($address, $netmask);
    my $bitmask = _netmask_to_bitmask($netmask);
    return "d".join(".", @network)."/$bitmask";
}

sub _gateway {
    my ($address, $netmask) = @_;
    my @network = _network_id($address, $netmask);
    $network[3] |= 1;
    return join(".", @network);
}

=head2 setup_vserver

Set the user changeable vserver related parameters
and set up NIS on it.

=over 4

=item *
domain

NIS domain name to where the CR Vserver belongs.

=item *
servers

NIS servers of the domain.

=back

=cut

############################################################
#     setup Vserver
############################################################

sub setup_vserver {
    $Log->enter();
    my ($self, %args) =	@_;
    %args = NACL::APISet::validate_params($self,
		 params => \%args,
		 spec => {
		     servers => { type => ARRAYREF, optional => 0 },
		     domain => { type => SCALAR, optional => 0 },
		 },
	     );

    $self->_clear_nis_domain();

    NACL::C::VserverServicesNisDomain->create(command_interface => $self->{node_obj},
					      %args,
					      vserver => $self->{vserver},
					      active => 'true');
    $self->{vserver_obj}->modify('ns-switch' => [ 'nis' ]);

    my $nfs = NACL::C::VserverNfs->find(command_interface => $self->{node_obj},
					allow_empty => 1);

    if (!defined($nfs)) {
	NACL::C::VserverNfs->create(command_interface => $self->{node_obj},
				  vserver => $self->{vserver},
				  v2 => 'disabled',
				  v3 => 'enabled',
				  "v4.0" => 'disabled',
				  "v4.1" => 'disabled',
				  udp => 'disabled');
    }

    my $rule = NACL::C::ExportPolicyRule->find(command_interface => $self->{node_obj},
						filter =>
						{ vserver => $self->{vserver},
						  policyname => 'default'},
						allow_empty => 1);;

    if (defined($rule)) {
      NACL::C::ExportPolicyRule->modify(command_interface => $self->{node_obj},
					  vserver => $self->{vserver},
					  policyname => 'default',
					  ruleindex => $rule->ruleindex(),
					  superuser => 'any',
					  anon => 0);
    }
    else {
      NACL::C::ExportPolicyRule->create(command_interface => $self->{node_obj},
					  vserver => $self->{vserver},
					  policyname => 'default',
					  clientmatch => '0.0.0.0/0',
					  rorule => 'any',
					  rwrule => 'any',
					  superuser => 'any',
					  anon => 0);
    }
    $Log->exit();
}

############################################################
#     setup NFS & NIS & DNS (part of setup_Vserver)
############################################################

# burt436901 - RFE: Feature enhancements for cr_setup.thpl: NIS, DNS and DP Repository support
# (to be implemented by restructuring setup_vserver)
