# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary BasicSetup Library - Base
## @author  galane@netapp.com
## @pod     here

=head1 NAME

CR::Setup::BasicSetup::Base

=head1 DESCRIPTION

Base basic setup module to configure vservers and infinitevols.

=cut

package CR::Setup::BasicSetup::Base;
use CR::Setup::Space;
use Digest::SHA1 qw(sha1_hex);
use Moose;
use MooseX::Params::Validate;
use namespace::autoclean;
use NACL::C::JobLog;
use NACL::C::LoggerMgwdLog;
use NACL::CS::Aggregate;
use NACL::STask::NetworkInterface;
use NACL::STask::NetworkRoutingGroupsRoute;
use NACL::STask::VserverServicesDns;
use NACL::STask::VserverServicesNisDomain;
use NACL::STask::VserverNfs;
use NACL::STask::SystemLicense;
use NACL::STask::Vserver;
use NATE::Log qw(log_global);
use TestUtils;

has infinitevols => (
    isa     => 'ArrayRef[NACL::STask::Volume]',
    is      => 'rw',
    traits  => ['Array'],
    handles => {
        add_infinitevol     => 'push',
        get_infinitevol     => 'get',
        clear_infinitevols  => 'clear',
        filter_infinitevols => 'grep',
    },
    predicate => 'has_infinitevols',
);

has infinitevol_configs => (
    isa     => 'ArrayRef[CR::Setup::BasicSetup::Base::InfinitevolConfig]',
    is      => 'rw',
    builder => '_build_infinitevol_config',
    traits  => ['Array'],
    handles => {
        add_infinitevol_config     => 'push',
        get_infinitevol_config     => 'get',
        clear_infinitevol_configs  => 'clear',
        filter_infinitevol_configs => 'grep',
    },
    predicate => 'has_infinitevol_configs',
);

has lif_manager => (
    isa     => 'CR::Setup::BasicSetup::Base::LIFManager',
    is      => 'rw',
    lazy    => 1,
    builder => '_build_lif_manager',
);

has node => (
    isa       => 'NACL::C::Node',
    is        => 'rw',
    predicate => 'has_node',
    clearer   => 'clear_node',
    trigger   => \&_node_was_set,
);

has version => (
    isa     => 'Str',
    is      => 'ro',
    default => 'latest',
);

has vservers => (
    isa     => 'ArrayRef[NACL::STask::Vserver]',
    is      => 'rw',
    traits  => ['Array'],
    handles => {
        add_vserver     => 'push',
        clear_vservers  => 'clear',
        filter_vservers => 'grep',
        get_vserver     => 'get',
    },
    predicate => 'has_vservers',
);

has vserver_configs => (
    isa     => 'ArrayRef[CR::Setup::BasicSetup::Base::VserverConfig]',
    is      => 'rw',
    builder => '_build_vserver_config',
    traits  => ['Array'],
    handles => {
        add_vserver_config    => 'push',
        get_vserver_config    => 'get',
        clear_vserver_configs => 'clear',
    },
    predicate => 'has_vserver_configs',
);

has _log => (
    isa     => 'NATE::Log',
    is      => 'ro',
    default => sub { log_global() }
);

has _site_info => (
    isa     => 'HashRef[Any]',
    is      => 'ro',
    builder => '_build_site_info',
);

##------------------------------------------------------------
# Public Interfaces
##------------------------------------------------------------
sub clear_all_configs {
    my $self = shift;

    $self->clear_vserver_configs();
    $self->clear_infinitevol_configs();
}

sub create {
    my $self = shift;

    $self->add_system_licenses();
    $self->enable_logging();
    $self->lif_manager->purge_lifs(role => 'intercluster');
    $self->lif_manager->purge_lifs(role => 'data');
    $self->vserver_setup();
    $self->lif_manager->create_intercluster_lifs();
    $self->infinitevol_setup();

    foreach my $vs_cfg (@{$self->vserver_configs}) {
        $self->create_cifs_for_vserver( vs_config => $vs_cfg );

        # only create CIFS share if vserver has a RW infinitevol
        my $create_cifs_share = undef;
        $create_cifs_share = $self->filter_infinitevol_configs(
            sub { $_->type eq 'RW' && $_->vserver eq $vs_cfg->vserver }
        );
        $self->create_cifs_share( vs_config => $vs_cfg ) if $create_cifs_share;
    }
}

sub infinitevol_setup {
    my $self = shift;

    foreach my $cfg ( @{ $self->infinitevol_configs } ) {
        $self->create_infinitevol( iv_config => $cfg );
        $self->create_storage_services( iv_config => $cfg );
    }
}

sub vserver_setup {
    my $self = shift;

    foreach my $cfg ( @{ $self->vserver_configs } ) {
        $self->create_vserver( vs_config => $cfg );
        $self->add_dns_domain( vs_config => $cfg );
        $self->add_nis_domain( vs_config => $cfg );
        $self->create_nfs_for_vserver( vs_config => $cfg );
        $self->lif_manager->create_data_lifs( vs_config => $cfg );
    }
}

##------------------------------------------------------------
# Builders
##------------------------------------------------------------

sub _build_infinitevol_config {
    my $self = shift;

    my @configs   = ();
    my $iv_config = CR::Setup::BasicSetup::Base::InfinitevolConfig->new();
    push( @configs, $iv_config );

    return \@configs;
}

sub _build_lif_manager {
    my $self = shift;
    return CR::Setup::BasicSetup::Base::LIFManager->new(
        node => $self->node );
}

sub _build_site_info {
    my $self = shift;

    my %site_info = get_site_info();
    my $site = $self->node->hostrec()->SITE();

    $site = 'svl' unless (defined $site);

    return $site_info{$site};
}

sub _build_vserver_config {
    my $self = shift;

    my @configs   = ();
    my $vs_config = CR::Setup::BasicSetup::Base::VserverConfig->new();
    push( @configs, $vs_config );

    return \@configs;
}

sub _node_was_set {
    my $self = shift;
    $self->lif_manager->node($self->node);
}

##------------------------------------------------------------
# Private methods
##------------------------------------------------------------
sub create_cifs_for_vserver {
    my ( $self, $vs_config ) = validated_list( \@_,
        vs_config => { isa => 'CR::Setup::BasicSetup::Base::VserverConfig' },
    );

    NACL::C::VserverCifsDomainPreferredDc->add(
        'command_interface' => $self->node,
        'vserver'           => $vs_config->vserver,
        'domain'            => $vs_config->cifs_config->cifs_domain_long,
        'preferred-dc'      => $vs_config->cifs_config->ad_server_ip,
    );

    sleep(10); # Preferred DC changes can take several seconds to propagate

    my $cifs_server_name = $self->node;
    $cifs_server_name .= $vs_config->vserver;
    $cifs_server_name =~ s/-|_//g;
    my $i = length($cifs_server_name);
    if ( $i > 15 ) {
        $cifs_server_name = substr( sha1_hex($cifs_server_name), 0, 15 );
    }

    NACL::C::VserverCifs->create(
        'command_interface' => $self->node,
        'vserver'           => $vs_config->vserver,
        'cifs-server'       => $cifs_server_name,
        'domain'            => $vs_config->cifs_config->cifs_domain_long,
        'username'          => $vs_config->cifs_config->ad_server_user,
        'password'          => $vs_config->cifs_config->ad_server_password,
    );
    NACL::C::VserverCifsOptions->modify(
        'command_interface' => $self->node,
        'default-unix-user' => 'root',
        'wins-servers'      => $vs_config->cifs_config->wins_server_ip,
        'vserver'           => $vs_config->vserver,
    );

    sleep(10);    # Sleep 10 second to work around BURT602514
    foreach my $superuser (@{ $vs_config->cifs_config->cifs_superusers }) {
        NACL::C::VserverCifsSuperuser->create(
            'command_interface' => $self->node,
            'vserver'           => $vs_config->vserver,
            'accountname'       => $superuser,
            'domain'            => $vs_config->cifs_config->cifs_domain,
        );
    }

    NACL::C::VserverNameMapping->create(
        'command_interface' => $self->node,
        'vserver'           => $vs_config->vserver,
        'position'          => 2,
        'direction'         => 'win-unix',
        'pattern' => $vs_config->cifs_config->cifs_domain . '\\\\(.+)',
        'replacement' => '\\1',
    );
    NACL::C::VserverNameMapping->create(
        'command_interface' => $self->node,
        'vserver'           => $vs_config->vserver,
        'position'          => 2,
        'direction'         => 'unix-win',
        'pattern'           => '(.+)',
        'replacement' => $vs_config->cifs_config->cifs_domain . '\\\\\\1',
    );
}

sub create_cifs_share {
    my ( $self, $vs_config ) = validated_list( \@_,
        vs_config => { isa => 'CR::Setup::BasicSetup::Base::VserverConfig' },
    );
    NACL::STask::VserverCifsShare->create(
        'command_interface'  => $self->node,
        'vserver'            => $vs_config->vserver,
        'share-name'         => $vs_config->cifs_config->cifs_share_name,
        'share-properties'   => $vs_config->cifs_config->share_properties,
        'path'               => $vs_config->cifs_config->cifs_share_path,
        'nacltask_if_exists' => $vs_config->cifs_config->exists_action,
    );
}

sub purge_cifs_for_vserver {
    my ( $self, $vs_config ) = validated_list( \@_,
        vs_config => { isa => 'CR::Setup::BasicSetup::Base::VserverConfig' },
    );

    my @vserver_cifs = NACL::C::VserverCifs->find(
        command_interface => $self->node,
        filter => {
            vserver => $vs_config->vserver,
        },
        allow_empty => 1,
    );
    if ( @vserver_cifs) {
        foreach my $cifs (@vserver_cifs) {
            $cifs->delete(
                'admin-username'    => $vs_config->cifs_config->ad_server_user,
                'admin-password'    => $vs_config->cifs_config->ad_server_password,
            );
        }
    }

    my @preferred_dcs = NACL::C::VserverCifsDomainPreferredDc->find(
        command_interface => $self->node,
        filter => {
            vserver => $vs_config->vserver,
        },
        allow_empty => 1,
    );
    if ( @preferred_dcs ) {
        foreach my $dc (@preferred_dcs) {
            $dc->NACL::C::VserverCifsDomainPreferredDc::remove();
        }
    }

    my @name_mappings = NACL::C::VserverNameMapping->find(
        command_interface => $self->node,
        filter => {
            vserver => $vs_config->vserver,
        },
        allow_empty => 1,
    );
    if ( @name_mappings ) {
        foreach my $map (@name_mappings) {
            $map->delete();
        }
    }

}

sub create_infinitevol {
    my ( $self, $iv_config ) = validated_list( \@_,
        iv_config => { isa => 'CR::Setup::BasicSetup::Base::InfinitevolConfig' },
    );

    my %create_params = (
        'command_interface'     => $self->node,
        'nacltask_if_exists'    => $iv_config->exists_action,
        'nacltask_is_infinivol' => 1,
        'snapshot-policy'       => $iv_config->snapshot_policy,
        'volume'                => $iv_config->volume,
        'vserver'               => $iv_config->vserver,
        'type'                  => $iv_config->type,
    );

    if ( $iv_config->type eq 'RW' ) {
        $create_params{'junction-path'}    = $iv_config->junction_path;
        $create_params{'unix-permissions'} = $iv_config->unix_permissions;
    }

    my $iv = NACL::STask::Volume->create( %create_params );
    $self->add_infinitevol( $iv );

    return $iv;
}

sub create_nfs_for_vserver {
    my ( $self, $vs_config ) = validated_list( \@_,
        vs_config => { isa => 'CR::Setup::BasicSetup::Base::VserverConfig' },
    );

    NACL::STask::VserverNfs->create(
        'command_interface'  => $self->node,
        'vserver'            => $vs_config->vserver,
        'nacltask_if_exists' => $vs_config->exists_action,
        'access'             => 'true',
        'v3'                 => 'enabled',
        'v4.0'               => 'disabled',
        'v4.1'               => 'enabled',
    );
}

sub create_storage_services {
    my ( $self, $iv_config ) = validated_list( \@_,
        iv_config => { isa => 'CR::Setup::BasicSetup::Base::InfinitevolConfig' },
    );

    return unless ( defined $iv_config->storage_services );

    my $space = CR::Setup::Space->new(version => $self->version);

    # FIXME remove once nacl volume create is supported for constituents
    my $cli = $self->node->apiset(
        category  => 'Node',
        interface => 'CLI',
        set       => 'CMode',
    );

    my @ss_configs = @{ $iv_config->storage_services };

    my $iv = NACL::STask::Volume->find(
        'command_interface' => $self->node,
        'filter'            => {
            'vserver' => $iv_config->vserver,
            'volume'  => $iv_config->volume,
        }
    );

    if ( $iv->get_one_state_attribute('is-managed-by-service') ne 'true' ) {
        $iv->modify( 'is-managed-by-service' => 'true' );
    }

    my $aggr = $space->aggregate_with_most_usable_space(
        vserver => $iv_config->vserver
    );

    foreach my $ss_cfg (@ss_configs) {
        foreach ( 1 .. $ss_cfg->num_constituents ) {
            $cli->volume_create(
                'vserver'          => $iv_config->vserver,
                'constituent-role' => 'data',
                'aggregate'        => $aggr->aggregate,
                'storage-service'  => $ss_cfg->storage_service,
            );
        }
    }
}

sub create_vserver {
    my ( $self, $vs_config ) = validated_list( \@_,
        vs_config => { isa => 'CR::Setup::BasicSetup::Base::VserverConfig' },
    );

    my $vs = NACL::STask::Vserver->create(
        'command_interface'         => $self->node,
        'nacltask_if_exists'        => $vs_config->exists_action,
        'aggregate'                 => $vs_config->aggregate,
        'is-repository'             => $vs_config->is_repository,
        'ns-switch'                 => $vs_config->ns_switch,
        'rootvolume'                => $vs_config->rootvolume,
        'rootvolume-security-style' => $vs_config->rootvolume_security_style,
        'vserver'                   => $vs_config->vserver,
        'nacltask_cifs_username'    => $vs_config->cifs_config->ad_server_user,
        'nacltask_cifs_password'    => $vs_config->cifs_config->ad_server_password,
    );

    $self->add_vserver( $vs );

    return $vs;
}

sub add_dns_domain {
    my ( $self, $vs_config ) = validated_list( \@_,
        vs_config => { isa => 'CR::Setup::BasicSetup::Base::VserverConfig' },
    );

    my @dns_servers = ( $self->_site_info->{'mgmt_dns1'} );

    if ( defined $self->_site_info->{'mgmt_dns2'} ) {
        push( @dns_servers, $self->_site_info->{'mgmt_dns2'} );
    }

    my @dns_domains = split( /\s+/, $self->_site_info->{'dns_domains'} );

    my $dns = NACL::STask::VserverServicesDns->create(
        'command_interface'  => $self->node,
        'nacltask_if_exists' => $vs_config->exists_action,
        'domains'            => \@dns_domains,
        'name-servers'       => \@dns_servers,
        'vserver'            => $vs_config->vserver,
    );

    return $dns;
}

sub add_nis_domain {
    my ( $self, $vs_config ) = validated_list( \@_,
        vs_config => { isa => 'CR::Setup::BasicSetup::Base::VserverConfig' },
    );

    my $nis = NACL::STask::VserverServicesNisDomain->create(
        'command_interface'  => $self->node,
        'nacltask_if_exists' => $vs_config->exists_action,
        'domain'             => $self->_site_info->{'nis_domain'},
        'servers'            => [ $self->_site_info->{'nis_servers'} ],
        'vserver'            => $vs_config->vserver,
        'active'             => 'true',
    );

    return $nis;
}

sub add_system_licenses {
    my ( $self, $licenses ) = validated_list(
        \@_,
        licenses => {
            isa      => 'ArrayRef[Str]',
            optional => 1,
            default  => [ "NFS", "CIFS", "Snapmirror_DP", "SnapRestore" ]
        },
    );

    foreach my $feature ( @{$licenses} ) {
        NACL::STask::SystemLicense->add(
            'command_interface' => $self->node,
            'feature'           => $feature,
            'nacltask_verify'   => 1
        );
    }
}

sub enable_logging {
    my ( $self, %params ) = validated_hash(
        \@_,
        level => {
            isa      => 'Str',
            optional => 1,
            default  => 'debug',
        },
        job_modules => {
            isa      => 'ArrayRef[Str]',
            optional => 1,
            default  => [ 'CR*' ],
        },
        mgwd_modules => {
            isa      => 'ArrayRef[Str]',
            optional => 1,
            default  => [ 'CR*', 'repos', 'tables::repos*' ],
        },
    );

    my @all_nodes = NACL::C::Node->find( command_interface => $self->node );

    foreach my $node (@all_nodes) {
        foreach my $job_mod ( @{ $params{job_modules} } ) {
            NACL::C::JobLog->modify(
                command_interface => $node,
                module            => $job_mod,
                level             => $params{level},
            );
        }
        foreach my $mgwd_mod ( @{ $params{mgwd_modules} } ) {
            NACL::C::LoggerMgwdLog->modify(
                command_interface => $node,
                module            => $mgwd_mod,
                level             => $params{level},
            );
        }
    }
}

##------------------------------------------------------------
# Internal Helpers
##------------------------------------------------------------

# Used to create modifiable config objects from the CR::Setup::Defaults
# NOTE: This should not be called directly, only extended by subclasses
package CR::Setup::BasicSetup::Base::Configs;
use Moose;
use namespace::autoclean;

has _defaults => (
    isa     => 'CR::Setup::Defaults::Base',
    is      => 'ro',
    default => sub { CR::Setup::Defaults::Base->new() }
);

sub BUILD {
    my $self = shift;
    my $component = $self->_default_component;

    foreach my $key ( keys %{ $self->_defaults->$component } ) {
        my $type
            = $self->_defaults->$component->meta->get_attribute($key)->type_constraint;

        # replace all the '-' in the hashes with '_' so we can use
        # CR::Setup::Defaults reader methods to grab the default values
        $key =~ s/-/_/g;

        $self->meta->add_attribute(
            $key => (
                is      => 'rw',
                isa     => $type,
                default => sub { $self->_defaults->$component->$key },
                lazy    => 1,
            )
        );
    }

}

package CR::Setup::BasicSetup::Base::ExistsAction;
use Moose::Role;

has exists_action => (
    isa     => 'Str',
    is      => 'rw',
    default => 'purge',
);

package CR::Setup::BasicSetup::Base::VserverConfig;
use Moose;
use namespace::autoclean;
extends 'CR::Setup::BasicSetup::Base::Configs';

with 'CR::Setup::BasicSetup::Base::ExistsAction';

has _default_component => (
    isa     => 'Str',
    is      => 'ro',
    default => 'Vserver',
);

has cifs_config => (
    isa     => 'CR::Setup::BasicSetup::Base::CIFSConfig',
    is      => 'rw',
    clearer => 'clear_cifs_config',
    default => sub { CR::Setup::BasicSetup::Base::CIFSConfig->new(); },
);

has num_data_lifs => (
    isa       => 'Int',
    is        => 'rw',
    default   => 1,
    predicate => 'has_num_data_lifs',
    clearer   => 'clear_num_data_lifs',
);

package CR::Setup::BasicSetup::Base::InfinitevolConfig;
use Moose;
use namespace::autoclean;
extends 'CR::Setup::BasicSetup::Base::Configs';

with 'CR::Setup::BasicSetup::Base::ExistsAction';

has _default_component => (
    isa     => 'Str',
    is      => 'ro',
    default => 'Infinitevol',
);

has storage_services => (
    isa     => 'ArrayRef[CR::Setup::BasicSetup::Base::StorageServiceConfig]',
    is      => 'rw',
    traits  => ['Array'],
    handles => {
        add_storage_service_config    => 'push',
        get_storage_service_config    => 'get',
        clear_storage_service_configs => 'clear',
    },
    predicate => 'has_storage_services',
);

has type => (
    isa       => 'Str',
    is        => 'rw',
    default   => 'RW',
);

has vserver => (
    isa       => 'Str',
    is        => 'rw',
    default   => 'vs0',
    predicate => 'has_vserver',
    clearer   => 'clear_vserver',
);

package CR::Setup::BasicSetup::Base::StorageServiceConfig;
use Moose;
use namespace::autoclean;

has num_constituents => (
    isa       => 'Int',
    is        => 'rw',
    default   => 0,
    predicate => 'has_num_constituents',
    clearer   => 'clear_num_constituents',
);

has storage_service => (
    isa       => 'Str',
    is        => 'rw',
    predicate => 'has_storage_service',
    clearer   => 'clear_storage_service',
);

package CR::Setup::BasicSetup::Base::CIFSConfig;
use Moose;
use namespace::autoclean;

with 'CR::Setup::BasicSetup::Base::ExistsAction';

has ad_server => (
    isa     => 'ArrayRef[Str]',
    is      => 'rw',
    default => sub { [qw(cnc-ad1 cnc-ad2)] },
);

has ad_server_ip => (
    isa     => 'ArrayRef[Str]',
    is      => 'rw',
    default => sub { [qw(10.98.203.74 10.98.202.208)] },
);

has ad_server_user => (
    isa     => 'Str',
    is      => 'rw',
    default => 'adadmin',
);

has ad_server_password => (
    isa     => 'Str',
    is      => 'rw',
    default => 'adadmin',
);

has kerberos_realm => (
    isa     => 'Str',
    is      => 'rw',
    default => 'CNC.SPIN.ENG.NETAPP.COM',
);

has cifs_domain => (
    isa     => 'Str',
    is      => 'rw',
    default => 'CNC',
);

has cifs_domain_long => (
    isa     => 'Str',
    is      => 'rw',
    default => 'CNC.SPIN.ENG.NETAPP.COM',
);

has cifs_share_name => (
    isa     => 'Str',
    is      => 'rw',
    default => 'cifs_share',
);

has cifs_share_path => (
    isa     => 'Str',
    is      => 'rw',
    default => '/NS',
);

has cifs_superusers => (
    isa     => 'ArrayRef[Str]',
    is      => 'rw',
    default => sub { [qw( Administrator adadmin )] },
);

has share_properties => (
    isa     => 'ArrayRef[Str]',
    is      => 'rw',
    default => sub { [qw(browsable)] },
);

has wins_server_ip => (
    isa     => 'ArrayRef[Str]',
    is      => 'rw',
    default => sub { [qw(10.98.203.74)] },
);

package CR::Setup::BasicSetup::Base::LIFManager;
use Moose;
use MooseX::Params::Validate;
use namespace::autoclean;

has net_host_params => (
    isa => 'ArrayRef[CR::Setup::BasicSetup::Base::LIFManager::NetHostParam]',
    is => 'rw',
    predicate => 'has_net_host_params',
    traits  => ['Array'],
    handles => {
        add_net_host_param    => 'push',
        clear_net_host_params => 'clear',
    }
);

has node => (
    isa       => 'NACL::C::Node',
    is        => 'rw',
    predicate => 'has_node',
    clearer   => 'clear_node',
);

has num_intercluster_lifs => (
    isa     => 'Int',
    is      => 'rw',
    default => '1',
);

has unused_net_host_params => (
    isa => 'ArrayRef[CR::Setup::BasicSetup::Base::LIFManager::NetHostParam]',
    is  => 'rw',
    predicate => 'has_unused_net_host_params',
    traits    => ['Array'],
    handles   => {
        add_unused_net_host_param      => 'push',
        clear_unused_net_host_params   => 'clear',
        num_unused_net_host_params     => 'count',
        delete_unused_net_host_param   => 'delete',
        filter_unused_net_host_params  => 'grep',
        index_of_unused_net_host_param => 'first_index',
    },
);

has vserver_configs => (
    isa     => 'ArrayRef[CR::Setup::BasicSetup::Base::VserverConfig]',
    is      => 'rw',
    predicate => 'has_vserver_configs',
    traits    => ['Array'],
    handles   => {
        add_vserver_config      => 'push',
        clear_vserver_configs   => 'clear',
    },
);

sub create_data_lifs {
    my ( $self, %params ) = validated_hash( \@_,
        vs_config        => { isa => 'CR::Setup::BasicSetup::Base::VserverConfig', optional => 1 },
    );
    my @all_nodes = NACL::C::Node->find( command_interface => $self->node );

    # if desired the user can pass in a vs config similar to other components in basic setup.
    if(defined $params{vs_config}) {
        $self->clear_vserver_configs();
        $self->add_vserver_config($params{vs_config});
    }

    foreach my $vs (@{ $self->vserver_configs }) {
        my $num_lifs_desired = $vs->num_data_lifs;
        my $lifs_per_node    = $num_lifs_desired / scalar(@all_nodes);
        if ( $lifs_per_node < 1 ) { $lifs_per_node = 1; } # place 1 per node until we reach desired # lifs

        $self->find_unused_net_host_params( nodes => \@all_nodes );

        foreach my $node ( @all_nodes ) {
            my @available_params_for_node = $self->filter_unused_net_host_params(
                sub { $_->node eq $node->node }
            );

            if ( scalar(@available_params_for_node) < 1 ) {
                CR::Setup::Exception->throw(
                    'Unable to create data LIFs: No available DATA_IPs on '
                        . $node->node . ' for vserver ' . $vs->vserver
                );
            }

            my $lifs_to_create = $lifs_per_node;

            while($lifs_to_create > 0 && $num_lifs_desired  > 0) {
                my $unused_param = shift @available_params_for_node;
                if ( !$unused_param ) {
                    CR::Setup::Exception->throw(
                        'Unable to create the desired number of data LIFs on '
                            . $node->node . ' for vserver '
                            . $vs->vserver . '. All available IPs consumed.'
                    );
                }

                $self->delete_unused_net_host_param(
                    $self->index_of_unused_net_host_param(sub{ $unused_param })
                );
                $self->create_network_interface(
                    net_host_param => $unused_param,
                    vserver        => $vs->vserver
                );
                $lifs_to_create = $lifs_to_create - 1;
                $num_lifs_desired = $num_lifs_desired - 1;
            }
            last if $num_lifs_desired == 0;
        }
    }
}

sub create_intercluster_lifs {
    my $self = shift;
    my $num_lifs_created = 0;

    my @all_nodes = NACL::C::Node->find( command_interface => $self->node );
    $self->find_unused_net_host_params( nodes => \@all_nodes );

    # just create intercluster lifs in a round robin fashion
    while($num_lifs_created < $self->num_intercluster_lifs) {
        foreach my $node (@all_nodes) {
            my @available_params_for_node = $self->filter_unused_net_host_params(
                sub { $_->node eq $node->node }
            );

            if ( scalar(@available_params_for_node) >= 1 ) {
                my $unused_param = shift @available_params_for_node;
                $self->delete_unused_net_host_param(
                    $self->index_of_unused_net_host_param(
                        sub { $unused_param }
                    )
                );
                $self->create_network_interface(
                    net_host_param => $unused_param,
                    vserver        => $node->node,
                    role           => 'intercluster',
                );
                $num_lifs_created = $num_lifs_created + 1;
            }
            last if($num_lifs_created == $self->num_intercluster_lifs);
        }

        if (   $num_lifs_created < $self->num_intercluster_lifs
            && $self->num_unused_net_host_params == 0 )
        {
            CR::Setup::Exception->throw(
                'Unable to create desired number of intercluster LIFs: Not '
                . 'enough available DATA_IPs in tharnhost file.' );
        }
    }
}

sub create_network_interface {
    my ( $self, %params ) = validated_hash( \@_,
        net_host_param => { isa => 'CR::Setup::BasicSetup::Base::LIFManager::NetHostParam' },
        role           => { isa => 'Str', default => 'data', optional => 1 },
        vserver        => { isa => 'Str' },
    );
    my $net_host_param = $params{'net_host_param'};

    my $lif_name
        = $params{role}
        . $net_host_param->index
        . '_' . $net_host_param->node;

    my %create_params = (
        command_interface  => $self->node,
        vserver            => $params{vserver},
        lif                => $lif_name,
        role               => $params{role},
        address            => $net_host_param->data_ip,
        'home-node'        => $net_host_param->node,
        'home-port'        => $net_host_param->data_port,
        netmask            => $net_host_param->data_mask,
        nacltask_if_exists => 'purge',
    );

    if ( $params{role} eq 'data' ) {
        my $lif = NACL::STask::NetworkInterface->create(
            'data-protocol' => [ 'nfs', 'cifs' ],
            %create_params,
        );
        my $routing_group = $lif->get_one_state_attribute('routing-group');
        NACL::STask::NetworkRoutingGroupsRoute->create(
            'command_interface'  => $self->node,
            'vserver'            => $params{vserver},
            'routing-group'      => $routing_group,
            'destination'        => '0.0.0.0/0',
            'gateway'            => $net_host_param->data_gateway,
            'nacltask_if_exists' => 'reuse',
        );
    }
    else {
        NACL::STask::NetworkInterface->create(%create_params);
    }
}

sub find_unused_net_host_params {
    my ( $self, %params ) = validated_hash( \@_,
         nodes => { isa => 'ArrayRef[NACL::C::Node]' },
    );

    $self->clear_net_host_params() if $self->has_net_host_params();
    $self->clear_unused_net_host_params() if $self->has_unused_net_host_params();

    foreach my $node ( @{ $params{nodes} } ) {
        $self->_find_net_host_params(node => $node);
    }
    foreach my $nhp ( @{ $self->net_host_params } ) {
        my $lif = NACL::STask::NetworkInterface->find(
            command_interface => $self->node,
            allow_empty       => 1,
            filter            => { address => $nhp->data_ip },
        );
        if ( !defined $lif ) {
            $self->add_unused_net_host_param( $nhp );
        }
    }
}

sub purge_lifs {
    my ( $self, %params ) = validated_hash( \@_,
         role => { isa => 'Str', default => 'data' },
    );

    my @lifs = NACL::STask::NetworkInterface->find(
        command_interface => $self->node,
        allow_empty       => 1,
        filter            => { role => $params{role} },
    );
    foreach (@lifs) { $_->purge if ( defined $_ ); }
}

sub _find_net_host_params {
    my ( $self, %params ) = validated_hash( \@_,
        node => { isa => 'NACL::C::Node' },
    );

    my $host           = $params{node}->hostrec();
    my %host_params    = $host->param_fields();

    my @ips   = map { $host_params{$_} } grep { /DATA_IP\d*/ } keys %host_params;
    my @ports = map { $host_params{$_} } grep { /DATA_PORT\d*/ } keys %host_params;
    my @masks = map { $host_params{$_} } grep { /DATA_MASK\d*/ } keys %host_params;
    my @gateways = map { $host_params{$_} } grep { /DATA_GATEWAY\d*/ } keys %host_params;

    # this assumes all DATA_IP entries have correstponding port,mask,and gateway entries
    my $count = 0;
    foreach ( @ips ) {
        my $nhp = CR::Setup::BasicSetup::Base::LIFManager::NetHostParam->new(
            data_ip      => $ips[$count],
            data_port    => $ports[$count],
            data_mask    => $masks[$count],
            data_gateway => $gateways[$count],
            'index'      => $count,
            node         => $host->id,
        );
        $self->add_net_host_param($nhp);
        $count = $count + 1;
    }
}

package CR::Setup::BasicSetup::Base::LIFManager::NetHostParam;
use Moose;
use namespace::autoclean;

has data_ip      => ( isa => 'Str', is => 'ro' );
has data_port    => ( isa => 'Str', is => 'ro' );
has data_mask    => ( isa => 'Str', is => 'ro' );
has data_gateway => ( isa => 'Str', is => 'ro' );
has 'index'      => ( isa => 'Str', is => 'ro' );
has node         => ( isa => 'Str', is => 'ro' );

package CR::Setup::BasicSetup::Base::MultiCluster;
use CR::Setup::Expectation;
use Moose;
use NACL::C::ClusterPeer;
use NACL::C::ClusterPeerHealth;
use NACL::C::Vserver;
use NACL::CS::ClusterIdentity;
use namespace::autoclean;

has dst_cluster_id => (
    isa     => 'Str',
    is      => 'rw',
    builder => '_build_dst_cluster_id',
    lazy    => 1,
);

has src_cluster_id => (
    isa     => 'Str',
    is      => 'rw',
    builder => '_build_src_cluster_id',
    lazy    => 1,
);

has [ 'dst_node_name', 'src_node_name' ] => (
    isa => 'Str',
    is  => 'rw',
);

has dst_node => (
    isa     => 'NACL::C::Node',
    is      => 'rw',
    builder => '_build_dst_node',
    lazy    => 1,
);

has src_node => (
    isa     => 'NACL::C::Node',
    is      => 'rw',
    builder => '_build_src_node',
    lazy    => 1,
);

has dst_intercluster_lif => (
    isa     => 'NACL::CS::NetworkInterface',
    is      => 'rw',
    builder => '_build_dst_intercluster_lif',
    lazy    => 1,
);

has src_intercluster_lif => (
    isa     => 'NACL::CS::NetworkInterface',
    is      => 'rw',
    builder => '_build_src_intercluster_lif',
    lazy    => 1,
);

has sm_config => (
    isa      => 'CR::Setup::BasicSetup::Base::MultiCluster::SnapmirrorConfig',
    is       => 'rw',
    required => 1,
);

sub _build_dst_cluster_id {
    my $self = shift;
    my $dst_cluster_id = NACL::CS::ClusterIdentity->fetch(
        command_interface => $self->dst_node, );
    return $dst_cluster_id->name;
}

sub _build_src_cluster_id {
    my $self = shift;
    my $src_cluster_id = NACL::CS::ClusterIdentity->fetch(
        command_interface => $self->src_node, );
    return $src_cluster_id->name;
}

sub _build_dst_node {
    my $self = shift;

    if ( !defined $self->dst_node_name ) {
        if ( defined $ENV{DST_NODE} ) {
            $self->dst_node_name( $ENV{DST_NODE} );
        }
        else {
            CR::Setup::Exception->throw(
                'Unable to create destination node component. Provide DST_NODE during'
                    . ' script invocation or dst_node_name during component creation.'
            );
        }
    }

    return NACL::C::Node->new( node => $self->dst_node_name );
}

sub _build_dst_intercluster_lif {
    my $self = shift;
    NACL::CS::NetworkInterface->fetch(
        command_interface => $self->dst_node,
        filter            => { role => 'intercluster' },
    );
}

sub _build_src_node {
    my $self = shift;

    if ( !defined $self->src_node_name ) {
        if ( defined $ENV{SRC_NODE} ) {
            $self->src_node_name( $ENV{SRC_NODE} );
        }
        else {
            CR::Setup::Exception->throw(
                'Unable to create source node component. Provide SRC_NODE during'
                    . ' script invocation or src_node_name during component creation.'
            );
        }
    }

    return NACL::C::Node->new( node => $self->src_node_name );
}

sub _build_src_intercluster_lif {
    my $self = shift;
    NACL::CS::NetworkInterface->fetch(
        command_interface => $self->src_node,
        filter            => { role => 'intercluster' },
    );
}

sub create_cluster_peering {
    my $self = shift;

    NACL::C::ClusterPeer->create(
        command_interface => $self->src_node,
        'peer-addrs'      => [ $self->dst_intercluster_lif->address ],
    );
    NACL::C::ClusterPeer->create(
        command_interface => $self->dst_node,
        'peer-addrs'      => [ $self->src_intercluster_lif->address ],
    );

}

sub purge_cluster_peering {
    my $self = shift;

    my $src_peering = NACL::C::ClusterPeer->find(
        command_interface => $self->src_node,
        allow_empty       => 1,
        filter            => { cluster => $self->dst_cluster_id, },
    );
    $src_peering->delete() if defined $src_peering;

    my $dst_peering = NACL::C::ClusterPeer->find(
        command_interface => $self->dst_node,
        allow_empty       => 1,
        filter            => { cluster => $self->src_cluster_id, },
    );
    $dst_peering->delete() if defined $dst_peering;
}

sub verify_cluster_peer_health {
    my $self = shift;

    my %cluster_peer_nodes = (
        $self->src_cluster_id => $self->dst_node,
        $self->dst_cluster_id => $self->src_node
    );

    foreach my $cluster_id ( keys %cluster_peer_nodes ) {
        my @cluster_peer_healths = ();
        this_block {
            @cluster_peer_healths = NACL::C::ClusterPeerHealth->find(
                command_interface => $cluster_peer_nodes{$cluster_id},
                filter            => { 'destination-cluster' => $cluster_id },
            );
        } should( eventually( pass() )->every(5)->within(120)->seconds() );
        foreach my $cph (@cluster_peer_healths) {
            this_block {
                $cph->get_one_state_attribute('cluster-health');
            } should( eventually( Be->lexically_equal_to( 'true' ) )->every(5)->within(60)->seconds() );
        }
    }
}

sub create_vserver_peering {
    my $self = shift;
    NACL::STask::VserverPeer->create(
        command_interface  => $self->src_node,
        vserver            => $self->sm_config->src_vserver->vserver,
        'peer-vserver'     => $self->sm_config->dst_vserver->vserver,
        'peer-cluster'     => $self->dst_cluster_id,
        applications       => ['snapmirror'],
        nacltask_if_exists => 'reuse',
    );
    NACL::STask::VserverPeer->accept(
        command_interface  => $self->dst_node,
        vserver            => $self->sm_config->dst_vserver->vserver,
        'peer-vserver'     => $self->sm_config->src_vserver->vserver,
    );
}

sub purge_vserver_peering {
    my $self = shift;
    NACL::STask::VserverPeer->purge(
        command_interface  => $self->src_node,
        vserver            => $self->sm_config->src_vserver->vserver,
        'peer-vserver'     => $self->sm_config->dst_vserver->vserver,
        nacltask_if_purged => 'pass',
    );
    NACL::STask::VserverPeer->purge(
        command_interface  => $self->dst_node,
        vserver            => $self->sm_config->dst_vserver->vserver,
        'peer-vserver'     => $self->sm_config->src_vserver->vserver,
        nacltask_if_purged => 'pass',
    );
}

sub create_snapmirror_relationship {
    my $self = shift;
    NACL::STask::Snapmirror->create(
        command_interface     => $self->dst_node,
        'source-cluster'      => $self->src_cluster_id,
        'destination-cluster' => $self->dst_cluster_id,
        'source-vserver'      => $self->sm_config->src_vserver->vserver,
        'destination-vserver' => $self->sm_config->dst_vserver->vserver,
        'source-volume'       => $self->sm_config->src_infinitevol->volume,
        'destination-volume'  => $self->sm_config->dst_infinitevol->volume,
        'nacltask_if_exists'  => 'reuse',
        'nacltask_wait'       => 1,
    );
}

sub purge_snapmirror_relationship {
    my $self = shift;
    NACL::STask::Snapmirror->purge(
        command_interface     => $self->dst_node,
        source_ci             => $self->src_node,
        'source-cluster'      => $self->src_cluster_id,
        'destination-cluster' => $self->dst_cluster_id,
        'source-vserver'      => $self->sm_config->src_vserver->vserver,
        'destination-vserver' => $self->sm_config->dst_vserver->vserver,
        'source-volume'       => $self->sm_config->src_infinitevol->volume,
        'destination-volume'  => $self->sm_config->dst_infinitevol->volume,
        nacltask_if_purged    => 'pass',
        nacltask_verify       => 1,
    );
}

package CR::Setup::BasicSetup::Base::MultiCluster::SnapmirrorConfig;
use Moose;
use namespace::autoclean;

has [ 'dst_vserver', 'src_vserver' ] => (
    isa      => 'CR::Setup::BasicSetup::Base::VserverConfig',
    is       => 'rw',
    required => 1,
);

has [ 'dst_infinitevol', 'src_infinitevol' ] => (
    isa      => 'CR::Setup::BasicSetup::Base::InfinitevolConfig',
    is       => 'rw',
    required => 1,
);

1;
