# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary Setup Cifs Info Class
## @author  Adam.Ciapponi@netapp.com
## @pod     here

=head1 NAME

CR::Setup::CifsInfo

=head1 DESCRIPTION

This module implements an object to encapsulate the cifs info.

=head1 METHODS

=cut

package CR::Setup::CifsInfo;

use strict;
use warnings;

use Data::Dumper::Concise;
use Digest::SHA1 qw(sha1_hex);
use Hostrec;
use NACL::STask::Volume;
use NACL::C::VserverCifs;
use NACL::C::VserverCifsDomainPreferredDc;
use NACL::C::VserverCifsSuperuser;
use NACL::C::VserverCifsOptions;
use NACL::C::VserverNameMapping;
use NACL::STask::VserverCifsShare;
use NACL::STask::SystemLicense;
use NATE::BaseException qw(:try);
use Tharn qw($Log);

my %defaults = (
    AD_SERVER => [qw(cnc-ad1 cnc-ad2)],
    AD_SERVER_IP => [qw(10.98.203.74 10.98.202.208)],
    AD_SERVER_USER => 'adadmin',
    AD_SERVER_PASSWORD => 'adadmin',
    KERBEROS_REALM => 'CNC.SPIN.ENG.NETAPP.COM',
    CIFS_DOMAIN => 'CNC',
    CIFS_DOMAIN_LONG => 'CNC.SPIN.ENG.NETAPP.COM',
    CIFS_SUPERUSERS => [qw(Administrator adadmin)],
    WINS_SERVER_IP => [qw(10.98.203.74)],
);

sub new {
    my $class = shift;
    my %params = @_;

    my $self = {};

    my %envs_hash = ();
    foreach my $param (keys %defaults) {
        if (defined $ENV{$param}) {
            if (ref($defaults{$param}) eq 'ARRAY') {
                my @a = split(',', $ENV{$param});
                $envs_hash{$param} = \@a;
            } else {
                $envs_hash{$param} = $ENV{$param};
            }
        }
    }

    my $hostrec;
    if (defined $envs_hash{'AD_SERVER'} ) {
        # Only the first Active Directory server specified will be used for hostrec
        my @ad_servers = @{ $envs_hash{'AD_SERVER'} };
        try {
            $hostrec = Hostrec->new( id => $ad_servers[0] );
        } otherwise {
            $Log->log('Unable to retrieve hostrec for:' . $ad_servers[0]);
        };
    }

    my %hostrec_hash = ();
    if ($hostrec) {
        foreach my $param (keys %defaults) {
            if ( $hostrec->$param() ) {
                if (ref($defaults{$param}) eq 'ARRAY') {
                    my @a = split(',', $hostrec->$param());
                    $hostrec_hash{$param} = \@a;
                } else {
                    $hostrec_hash{$param} = $hostrec->$param();
                }
            }
        }
    }

    foreach my $param (keys %defaults) {
        my $lowercase = lc $param;
        if ( defined $envs_hash{$param}  ) {
            $self->{$lowercase} = $envs_hash{$param};
        } else {
            if ( defined $hostrec_hash{$param} ) {
                $self->{$lowercase} = $hostrec_hash{$param};
            } else {
                $self->{$lowercase} = $defaults{$param};
            }
        }
    }
    $Log->log("CIFS Information:\n" . Dumper($self));
    $self->{'helpers'} = $params{'helpers'};
    bless $self, $class;
    return $self;
}

sub setup_cifs_for_vserver {
    my $self = shift;
    my %params = @_;

    my $node = $params{'node'};
    my $vserver = $params{'vserver'};
    $vserver = $self->{'helpers'}->vserver_stask_for( vserver => $vserver );

    my $cifs_share_name = $params{'cifs_share_name'};
    if ( ! defined $cifs_share_name ) {
        $cifs_share_name = 'cifs_share';
    }
    my $cifs_share_path = $params{'cifs_share_path'};
    if ( ! defined $cifs_share_path ) {
        my $iv = NACL::CS::Volume->fetch(
           command_interface => $node,
           filter => {
                vserver => $vserver->vserver,
                'volume-style' => 'infinitevol',
           },
           requested_fields => [qw(junction-path)],
        );
        $cifs_share_path = $iv->junction_path;
    }

    my $cifs_server_name = $node->node;
    $cifs_server_name .= $vserver->vserver;
    $cifs_server_name =~ s/-|_//g;
    my $i = length($cifs_server_name);
    if ( $i > 15 ) {
        $cifs_server_name = substr( sha1_hex($cifs_server_name), 0, 15 );
    }

    my $share_properties = $params{'share-properties'};
    if ( ! defined $share_properties ) {
        $share_properties = [qw(browsable)];
    }

    NACL::STask::SystemLicense->add(
        command_interface => $node,
        feature           => 'CIFS',
        nacltask_verify   => 1,
    );

    NACL::C::VserverCifsDomainPreferredDc->add(
        'command_interface' => $node,
        'vserver'           => $vserver->vserver,
        'domain'            => $self->{'cifs_domain_long'},
        'preferred-dc'      => $self->{'ad_server_ip'},
    );

    sleep 10; # Preferred DC changes can take several seconds to propagate

    NACL::C::VserverCifs->create(
        'command_interface' => $node,
        'vserver'           => $vserver->vserver,
        'cifs-server'       => $cifs_server_name,
        'domain'            => $self->{'cifs_domain_long'},
        'username'          => $self->{'ad_server_user'},
        'password'          => $self->{'ad_server_password'},
    );

    NACL::C::VserverCifsOptions->modify(
        'command_interface' => $node,
        'default-unix-user' => 'root',
        'wins-servers'      => $self->{'wins_server_ip'},
        'vserver'           => $vserver->vserver,
    );

    my @superusers = @{ $self->{'cifs_superusers'} };
    sleep(10);    # Sleep 10 second to work around BURT602514
    foreach my $superuser (@superusers) {
        NACL::C::VserverCifsSuperuser->create(
            'command_interface' => $node,
            'vserver'           => $vserver->vserver,
            'accountname'       => $superuser,
            'domain'            => $self->{'cifs_domain'},
        );
    }

    NACL::C::VserverNameMapping->create(
        'command_interface' => $node,
        'vserver'           => $vserver->vserver,
        'position'          => 2,
        'direction'         => 'win-unix',
        'pattern' => $self->{'cifs_domain'} . '\\\\(.+)',
        'replacement' => '\\1',
    );
    NACL::C::VserverNameMapping->create(
        'command_interface' => $node,
        'vserver'           => $vserver->vserver,
        'position'          => 2,
        'direction'         => 'unix-win',
        'pattern'           => '(.+)',
        'replacement' => $self->{'cifs_domain'} . '\\\\\\1',
    );

    my $cifs_share = NACL::STask::VserverCifsShare->create(
        'command_interface'  => $node,
        'vserver'            => $vserver->vserver,
        'share-name'         => $cifs_share_name,
        'share-properties'   => $share_properties,
        'path'               => $cifs_share_path,
        'nacltask_if_exists' => "purge",
    );

    return $cifs_share;
}

sub purge_cifs_for_vserver {
    my $self = shift;
    my %params = @_;

    my $node = $params{'node'};
    my $vserver = $params{'vserver'};
    $vserver = $self->{'helpers'}->vserver_stask_for( vserver => $vserver );

    my @vserver_cifs = NACL::C::VserverCifs->find(
        command_interface => $node,
        filter => {
            vserver => $vserver->vserver,
        },
        allow_empty => 1,
    );
    if ( @vserver_cifs) {
        foreach my $cifs (@vserver_cifs) {
            $cifs->delete(
                'admin-username'    => $self->{'ad_server_user'},
                'admin-password'    => $self->{'ad_server_password'},
            );
        }
    }


    my @preferred_dcs = NACL::C::VserverCifsDomainPreferredDc->find(
        command_interface => $node,
        filter => {
            vserver => $vserver->vserver,
        },
        allow_empty => 1,
    );
    if ( @preferred_dcs ) {
        foreach my $dc (@preferred_dcs) {
            $dc->remove();
        }
    }

    my @name_mappings = NACL::C::VserverNameMapping->find(
        command_interface => $node,
        filter => {
            vserver => $vserver->vserver,
        },
        allow_empty => 1,
    );
    if ( @name_mappings ) {
        foreach my $map (@name_mappings) {
            $map->delete();
        }
    }

    return;
}

1;
