################################################################################
# Copyright (c) 2012 NetApp, Inc.
# All rights reserved.
################################################################################

package CR::Setup::DataPolicy;

use strict;
use warnings;

use CR::Setup::Constants;
use JSON::XS qw();
use Params::Validate qw(:all);
use Scalar::Util qw(blessed);
use Storable qw(dclone);

sub new {
    my $class  = shift;
    my %params = validate_with(
        params => \@_,
        spec   => {
            data_policy => {
                type      => HASHREF | OBJECT,
                optional  => $TRUE,
                callbacks => {
                    'should be a CR::Setup::DataPolicy if it is an OBJECT' =>
                        sub {
                        if ( blessed $_[0] ) {
                            if ( $_[0]->isa('CR::Setup::DataPolicy') ) {
                                return $TRUE;
                            }
                            else {
                                return $FALSE;
                            }
                        }
                        else {
                            return $TRUE;
                        }
                        }
                }
            },
            data_policy_json       => { type => SCALAR,   optional => $TRUE },
            rules                  => { type => ARRAYREF, default  => [] },
            ruleset_format_version => { type => SCALAR,   default  => '1.0' },
        },
        allow_extra => $FALSE,
    );
    my $self;
    if ( defined $params{'data_policy'} ) {
        my @rules = @{ $params{'data_policy'}->{'rules'} };
        @rules
            = map { CR::Setup::DataPolicy::Rule->new( rule => $_ ) } @rules;
        $params{'data_policy'}->{'rules'} = \@rules;
        $self = $params{'data_policy'};
    }
    elsif ( defined $params{'data_policy_json'} ) {
        my $decoded_data_policy
            = JSON::XS->new->decode( $params{'data_policy_json'} );
        my @rules = @{ $decoded_data_policy->{'rules'} };
        @rules
            = map { CR::Setup::DataPolicy::Rule->new( rule => $_ ) } @rules;
        $decoded_data_policy->{'rules'} = \@rules;
        $self = $decoded_data_policy;
    }
    else {
        $self = {
            ruleset_format_version => $params{'ruleset_format_version'},
            rules                  => $params{'rules'},
        };
    }
    bless $self, $class if !blessed $self;
    return $self;
}

sub add_rule {
    my $self   = shift;
    my %params = validate_with(
        params => \@_,
        spec   => {
            rule_label => {
                type    => SCALAR,
                default => 'Rule ' . scalar( @{ $self->{'rules'} } )
            },
        },
        allow_extra => $TRUE,
    );
    unshift(
        @{ $self->{'rules'} },
        CR::Setup::DataPolicy::Rule->new(%params)
    );
    return $self;
}

sub add_default_rule {
    my $self   = shift;
    my %params = validate_with(
        params => \@_,
        spec   => {
            rule_label => { type => SCALAR,   default => 'default' },
            rule_scope => { type => ARRAYREF, default => [] },
        },
        allow_extra => $TRUE,
    );
    unshift(
        @{ $self->{'rules'} },
        CR::Setup::DataPolicy::Rule->new(%params)
    );
    return $self;
}

sub to_hash {
    my $self   = shift;
    my %params = validate_with(
        params      => \@_,
        spec        => {},
        allow_extra => $FALSE,
    );
    my $hash = dclone($self);
    $hash = { %{$hash} };
    my @rules = @{ $hash->{'rules'} };
    @rules = map { $_->to_hash } @rules;
    $hash->{'rules'} = \@rules;
    return $hash;
}

sub to_json {
    my $self   = shift;
    my %params = validate_with(
        params      => \@_,
        spec        => {},
        allow_extra => $FALSE,
    );
    return JSON::XS->new->encode( $self->to_hash );
}

package CR::Setup::DataPolicy::Rule;

use CR::Setup::Constants;
use JSON::XS qw();
use Params::Validate qw(:all);
use UUID::Tiny;
use Scalar::Util qw(blessed);
use Storable qw(dclone);

sub new {
    my $class  = shift;
    my %params = validate_with(
        params => \@_,
        spec   => {
            rule => {
                type      => HASHREF | OBJECT,
                callbacks => {
                    'should be a CR::Setup::DataPolicy::Rule if it is an OBJECT'
                        => sub {
                        if ( blessed $_[0] ) {
                            if ( $_[0]->isa('CR::Setup::DataPolicy::Rule') ) {
                                return $TRUE;
                            }
                            else {
                                return $FALSE;
                            }
                        }
                        else {
                            return $TRUE;
                        }
                        }
                },
                optional => $TRUE
            },
            rule_json => { type => SCALAR, optional => $TRUE },
            rule_label =>
                { type => SCALAR, default => 'randomlabel' . rand() },
            rule_id => { type => SCALAR, default => create_UUID_as_string() },
            rule_scope => { type => ARRAYREF, default => [] },
            rule_epoch => {
                type    => HASHREF,
                default => { epoch_reference => 'ctime' }
            },
            rule_epochs    => { type => HASHREF, optional => $TRUE },
            storageservice => { type => SCALAR,  default  => '-' },
        },
        allow_extra => $FALSE,
    );

    my $self;
    if ( defined $params{'rule'} ) {
        $self = $params{'rule'};
    }
    elsif ( defined $params{'rule_json'} ) {
        $self = JSON::XS->new->decode( $params{'rule_json'} );
    }
    else {
        my $rule_epochs;
        if ( defined $params{'rule_epochs'} ) {
            $rule_epochs = $params{'rule_epochs'};
        }
        else {
            $rule_epochs = {
                0 => {
                    'local' => {
                        'metadata' => {
                            'storageservice' => $params{'storageservice'},
                        }
                    }
                }
            };
        }
        $self = {
            rule_label  => $params{'rule_label'},
            rule_id     => $params{'rule_id'},
            rule_scope  => $params{'rule_scope'},
            rule_epoch  => $params{'rule_epoch'},
            rule_epochs => $rule_epochs,
        };
    }

    bless $self, $class if !blessed $self;
    return $self;
}

sub to_hash {
    my $self   = shift;
    my %params = validate_with(
        params      => \@_,
        spec        => {},
        allow_extra => $FALSE,
    );
    my $hash = dclone($self);
    $hash = { %{$hash} };
    return $hash;
}

sub to_json {
    my $self   = shift;
    my %params = validate_with(
        params      => \@_,
        spec        => {},
        allow_extra => $FALSE,
    );
    return JSON::XS->new->encode( $self->to_hash );
}
1;
