# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary Setup Helpers Library - Base Version
## @author  Adam.Ciapponi@netapp.com
## @pod     here

=head1 NAME

CR::Setup::Helpers::Base

=head1 DESCRIPTION

This module implements version independent helpers and is a base class for the version specific helpers.

=head1 METHODS

=cut

package CR::Setup::Helpers::Base;
use strict;
use warnings;
use CR::Setup::Constants;
use CR::Setup::Expectation;
use CR::Setup::Exception;
use CR::C::DebugSmdbTableReposDcMinSize;
use Data::Dumper::Concise;
use NACL::STask::Snapmirror;
use NACL::STask::Volume;
use NACL::STask::Vserver;
use NATE::BaseException qw(:try);
use Params::Validate qw(:all);
use Scalar::Util qw(blessed);
use Tharn qw($Log);

our @ISA = qw(CR::Setup::Helpers);

=head2 new

This does initialization of the helpers object.

=cut

sub new {
    my $class = shift;
    my %params = @_;

    my $self = {
        defaults => $params{'defaults'},
        cdefs_from_build => $FALSE,
    };
    bless $self, $class;
    return $self;
}

=head2 set_default_node

Sets the default node for the helpers instance. If no node is specified in commands requiring it,
the default will be used.

=cut

sub set_default_node {
    my $self = shift;
    my %params = @_;

    $self->{'node'} = $params{'node'};
    return;
}

=head2 get_node

Gets a node object for use.

=cut

sub get_node {
    my $node = NACL::C::Node->find();
    return $node;
}

=head2 get_apiset_for_node

Gets an apiset object for a given node.

=cut

sub get_apiset_for_node {
    my $self = shift;
    my %params = @_;

    my $node = delete $params{'node'};
    if ( !defined $node ) {
        $node = $self->{'node'};
    }
    my $privilege = delete $params{'privilege'};
    my $set_timeout = delete $params{'set_timeout'};
    my $cdefs_from_build = delete $params{'cdefs_from_build'};
    if ( ! defined $cdefs_from_build ) {
        $cdefs_from_build = $self->{'cdefs_from_build'};
    }

    my $apiset = NACL::APISet->new(
        hostobj   => $node->hostrec(),
        category  => 'Node',
        version_info => { cdefs_from_build => $cdefs_from_build },
        %params
    );

    if ($set_timeout) {
        $apiset->set_timeout(
            'connectrec-timeout' => $self->{'defaults'}->Common->timeout );
    }
    if ($privilege) {
        $apiset->set_privilege( 'privilege-level' => $privilege );
    }

    return $apiset;
}

=head2 get_admin_cli

Gets an admin privilege cmode cli object.

=cut

sub get_admin_cli {
    my $self = shift;
    my %params = @_;

    $params{'privilege'} = 'admin';
    $params{'interface'} = 'CLI';
    $params{'set'} = 'CMode';
    $params{'set_timeout'} = $TRUE;

    return $self->get_apiset_for_node(%params);
}

=head2 get_advanced_cli

Gets an advanced privilege cmode cli object.

=cut

sub get_advanced_cli {
    my $self = shift;
    my %params = @_;

    $params{'privilege'} = 'advanced';
    $params{'interface'} = 'CLI';
    $params{'set'} = 'CMode';
    $params{'set_timeout'} = $TRUE;

    return $self->get_apiset_for_node(%params);
}

=head2 get_diagnostic_cli

Gets a diagnostic privilege cmode cli object.

=cut

sub get_diagnostic_cli {
    my $self = shift;
    my %params = @_;

    $params{'privilege'} = 'diagnostic';
    $params{'interface'} = 'CLI';
    $params{'set'} = 'CMode';
    $params{'set_timeout'} = $TRUE;

    return $self->get_apiset_for_node(%params);
}

=head2 get_test_cli

Gets a test privilege cmode cli object.

=cut

sub get_test_cli {
    my $self = shift;
    my %params = @_;

    $params{'privilege'} = 'test';
    $params{'interface'} = 'CLI';
    $params{'set'} = 'CMode';
    $params{'set_timeout'} = $TRUE;

    return $self->get_apiset_for_node(%params);
}

=head2 get_zapi

Gets a zapi apiset object.

=cut

sub get_zapi {
    my $self = shift;
    my %params = @_;

    $params{'interface'} = 'ZAPI';
    $params{'set'} = 'CMode';
    $params{'set_timeout'} = $FALSE;
    $params{'cdefs_from_build'} = $FALSE;

    return $self->get_apiset_for_node(%params);
}

=head2 get_systemshell

Gets a systemshell apiset object.

=cut

sub get_systemshell {
    my $self = shift;
    my %params = @_;

    $params{'interface'} = 'CLI';
    $params{'set'} = 'Systemshell';
    $params{'shell_type'} = 'CMode';
    $params{'set_timeout'} = $FALSE;
    $params{'cdefs_from_build'} = $FALSE;

    return $self->get_apiset_for_node(%params);
}

=head2 find_from_defaults

Finds stasks from defaults.

=cut

sub find_from_defaults {
    my $self = shift;
    my %params = @_;

    my $node = $params{'node'};
    if ( !defined $node ) {
        $node = $self->{'node'};
    }

    if ( $params{'type'} eq 'infinitevol') {
        return $self->volume_stask_for(
            node         => $node,
            volume       => $self->{'defaults'}->Infinitevol->volume,
            vserver      => $self->{'defaults'}->Vserver->vserver,
            allow_empty  => 1,
        );
    } elsif ( $params{'type'} eq 'vserver' ) {
        return $self->vserver_stask_for(
            node        => $node,
            vserver     => $self->{'defaults'}->Vserver->vserver,
            allow_empty => 1,
        );
    } else {
        CR::Setup::Exception->throw(
            'find_from_defaults - Invalid usage for field type: ' . $params{'type'}
        );
    }
}

=head2 delete_volume_if_exists

Purges the specified volume if it is defined.

=cut

sub delete_volume_if_exists {
    my $self = shift;
    my %params = @_;

    $Log->log('Deleting volume if it exists.');
    if (defined $params{'volume'}) {
        my $volume_stask = $self->volume_stask_for(%params);
        $Log->log('Volume: ' . $volume_stask->volume . ' is being deleted.');
        $volume_stask->purge( 'method-timeout' => $self->{'defaults'}->Common->timeout );
    } else {
        $Log->log('Volume was not present, no need to delete.');
    }

    return;
}

=head2 minimum_data_constituent_size

Gets the minimum data constituent size for the given node.

=cut

sub minimum_data_constituent_size {
    my $self = shift;
    my %params = @_;

    my $node = $params{'node'};
    if ( !defined $node ) {
        $node = $self->{'node'};
    }

    my $filer_default = 1 * $TB;
    my $vsim_default = 256 * $MB;
    my $override = undef;

    try {
        my @results = CR::C::DebugSmdbTableReposDcMinSize->find(
            command_interface => $node,
            allow_empty => 1
        );
        if ( scalar(@results) > 0 ) {
            $override = $results[0]->get_one_state_attribute('min-data-constituent-size');
        }
    } otherwise {
        $Log->log('Smf table repos_dc_min_size does not exist.');
        $filer_default = 256 * $MB;
    };

    if ( $override ) {
        $Log->log('Minimum DC Size was overridden, returning that size (bytes): ' . $override);
        return $override;
    } else {
        if ( $node->hostrec()->hosttype() =~ /vsim/ ) {
            $Log->log('Minimum DC Size is the vsim default (bytes): ' . $vsim_default);
            return $vsim_default;
        } else {
            $Log->log('Minimum DC Size is the filer default (bytes): ' . $filer_default);
            return $filer_default;
        }
    }
}

=head2 disable_background_operations

Disables all background operations. Especially those that can result in size changes.

=cut

sub disable_background_operations {
    my $self = shift;
    my %params = @_;

    my $node = $params{'node'};
    if ( !defined $node ) {
        $node = $self->{'node'};
    }

    # Disable vserver root volumes snapshot
    my @rootvolumes = NACL::STask::Volume->find(
        command_interface => $node,
        filter => {
            vsroot            => 'true',
            'snapshot-policy' => '!none',
        },
        allow_empty => 1
    );
    foreach my $rv (@rootvolumes) {
        $rv->modify( 'snapshot-policy' => 'none');
    }

    # Quiesce all mirrors
    my @mirrors = NACL::STask::Volume->find(
        command_interface => $node,
        filter => {
            'constituent-role' => 'ns_mirror',
        },
        allow_empty => 1
    );
    @mirrors = map { $_->vserver . ':' . $_->volume } @mirrors;
    @mirrors = map {
        NACL::STask::Snapmirror->find(
            command_interface => $node,
            filter => {
                'destination-path' => $_
            }
        );
    } @mirrors;
    foreach my $mirror (@mirrors) {
        try {
            $mirror->quiesce(
                'nacltask_verify' => 1,
                'method-timeout'  => $self->{'defaults'}->Common->timeout
            );
        }
        catch NACL::Exceptions::VerifyFailure with {
            my $exception = shift;
            $Log->log('Unable to quiesce Snapmirror: ' . $exception->text());
        };
    }

    return;
}

=head2 enable_background_operations

Reenables useful background operations that were disabled by disable_background_operations().

=cut

sub enable_background_operations {
    my $self = shift;
    my %params = @_;

    my $node = $params{'node'};
    if ( !defined $node ) {
        $node = $self->{'node'};
    }

    # We don't actually want the vserver root set back to default for snapshot-policy

    # Resume all mirrors
    my @mirrors = NACL::STask::Volume->find(
        command_interface => $node,
        filter => {
            'constituent-role' => 'ns_mirror',
        },
        allow_empty => 1
    );
    @mirrors = map { $_->vserver . ':' . $_->volume } @mirrors;
    @mirrors = map {
        NACL::STask::Snapmirror->find(
            command_interface => $node,
            filter => {
                'destination-path' => $_
            }
        );
    } @mirrors;
    foreach my $mirror (@mirrors) {
        $mirror->resume(
            'method-timeout' => $self->{'defaults'}->Common->timeout
        );
    }

    return;
}

sub nacl_for {
    my $self = shift;
    my %params = @_;

    my $node = delete $params{'node'};
    if ( !defined $node ) {
        $node = $self->{'node'};
    }

    my $nacl_type = delete $params{'nacl_type'};
    my $package = delete $params{'package'};
    my $primary_key = delete $params{'primary_key'};
    my $target = delete $params{'target'};
    if ( !defined $target ) {
        $target = delete $params{$primary_key};
    }
    my $allow_empty = delete $params{'allow_empty'};
    my @keys = @{ delete $params{'keys'} };
    if ( !defined $allow_empty ) {
        $allow_empty = 0;
    }

    if ( $nacl_type eq 'stask' ) {
        my $stask_package = 'NACL::STask::' . $package;

        my $stask;
        if ( blessed $target ) {
            if ( $target->isa($stask_package) ) {
                $stask = $target;
            } elsif ( $target->isa('NACL::C::' . $package ) || $target->isa('NACL::CS::' . $package) ) {
                my %filter = map { $target->$_ } @keys;
                %filter = (%filter, %params);
                $stask = $stask_package->find(
                    command_interface => $node,
                    filter => \%filter,
                    allow_empty => $allow_empty
                );
            } else {
                CR::Setup::Exception->throw(
                    "Invalid target usage for helper stask_for(). $target is not an appropriate object."
                );
            }
        } else {
            if ( !ref($target) ) {
                my %filter = (
                    $primary_key => $target
                );
                %filter = (%filter, %params);
                $stask = $stask_package->find(
                    command_interface => $node,
                    filter => \%filter,
                    allow_empty => $allow_empty
                );
            } else {
                CR::Setup::Exception->throw(
                    "Invalid target usage for helper stask_for(). $target is not a scalar."
                );
            }
        }

        return $stask;
    } elsif ( $nacl_type eq 'state' ) {
        my $state_package = 'NACL::CS::' . $package;

        my $state;
        if ( blessed $target ) {
            if ( $target->isa($state_package) ) {
                $state = $target;
            } elsif ( $target->isa('NACL::C::' . $package ) || $target->isa('NACL::STask::' . $package) ) {
                $state = $target->state;
            } else {
                CR::Setup::Exception->throw(
                    "Invalid target usage for helper stask_for(). $target is not an appropriate object."
                );
            }
        } else {
            if ( !ref($target) ) {
                my %filter = (
                    $primary_key => $target
                );
                %filter = (%filter, %params);
                $state = $state_package->fetch(
                    command_interface => $node,
                    filter => \%filter,
                    allow_empty => $allow_empty
                );
            } else {
                CR::Setup::Exception->throw(
                    "Invalid target usage for helper stask_for(). $target is not a scalar."
                );
            }
        }

        return $state;
    } else {

    }
}

sub nacl_for_list {
    my $self = shift;
    my %params = @_;

    my $list = delete $params{'list'};
    my $function = delete $params{'function'};

    my @list_array;
    if ( ref($list) eq 'ARRAY' ) {
        @list_array = @{ $list };
    } elsif ( !ref($list) ) {
        @list_array = split(',', $list);
    } else {

    }

    my @formatted_array = ();
    @formatted_array = map { $self->$function( target => $_, %params) } @list_array;

    return \@formatted_array;
}

sub stask_for {
    my $self = shift;
    my %params = @_;

    $params{'nacl_type'} = 'stask';

    return $self->nacl_for(%params);
}

sub stasks_for_list {
    my $self = shift;
    my %params = @_;

    $params{'function'} = 'stask_for';

    return $self->nacl_for_list(%params);
}

sub state_for {
    my $self = shift;
    my %params = @_;

    $params{'nacl_type'} = 'state';

    return $self->nacl_for(%params);
}

sub states_for_list {
    my $self = shift;
    my %params = @_;

    $params{'function'} = 'state_for';

    return $self->nacl_for_list(%params);
}

sub aggregate_for {
    my $self = shift;
    my %params = @_;

    $params{'package'} = 'Aggregate';
    $params{'keys'} = [qw(aggregate)];
    $params{'primary_key'} = 'aggregate';

    return $self->nacl_for(%params);
}

sub aggregates_for_list {
    my $self = shift;
    my %params = @_;

    $params{'function'} = 'aggregate_for';

    return $self->nacl_for_list(%params);
}

sub aggregates_for_vserver {
    my $self = shift;
    my %params = @_;

    my $node = $params{'node'};
    if ( !defined $node ) {
        $node = $self->{'node'};
    }

    my $context = $params{'context'};
    if ( !defined $context ) {
        $context = 'cluster';
    }

    my $nacl_type = $params{'nacl_type'};
    my $package;
    my $find_function;

    if ($nacl_type eq 'stask') {
        $package = 'NACL::STask::Aggregate';
        $find_function = 'find';
    } elsif ($nacl_type eq 'state') {
        $package = 'NACL::CS::Aggregate';
        $find_function = 'fetch';
    } else {
        CR::Setup::Exception->throw(
            "Invalid nacl type provided for helper aggregates_for_vserver(): $nacl_type"
        );
    }

    my $vserver = $params{'vserver'};
    $vserver = $self->vserver_state_for( vserver => $vserver, command_interface => $node );
    my $aggr_list = $vserver->aggr_list;

    my @aggrs;
    if ( $aggr_list eq '-' ) {
        if ( $context eq 'cluster' ) {
            # - means all mroot aggregates
            @aggrs = $package->$find_function(
                command_interface => $node,
                filter => {
                   'has-mroot' => 'false'
                },
                allow_empty => 1
            );
        } elsif ( $context eq 'vserver' ) {
            # - means no aggregates
            @aggrs = ();
        } else {
            CR::Setup::Exception->throw(
                "Invalid context usage for helper aggregates_for_vserver(): $context"
            );
        }
    } else {
        @aggrs = @{
            $self->aggregates_for_list(
                node      => $node,
                list      => $aggr_list,
                nacl_type => $nacl_type
            )
        };
    }

    return \@aggrs;
}

sub aggregate_stask_for {
    my $self = shift;
    my %params = @_;

    $params{'nacl_type'} = 'stask';

    return $self->aggregate_for(%params);
}

sub aggregate_stasks_for_list {
    my $self = shift;
    my %params = @_;

    $params{'function'} = 'aggregate_stask_for';

    return $self->nacl_for_list(%params);
}

sub aggregate_stasks_for_vserver {
    my $self = shift;
    my %params = @_;

    $params{'nacl_type'} = 'stask';

    return $self->aggregates_for_vserver(%params);
}

sub aggregate_state_for {
    my $self = shift;
    my %params = @_;

    $params{'nacl_type'} = 'state';

    return $self->aggregate_for(%params);
}

sub aggregate_states_for_list {
    my $self = shift;
    my %params = @_;

    $params{'function'} = 'aggregate_state_for';

    return $self->nacl_for_list(%params);
}

sub aggregate_states_for_vserver {
    my $self = shift;
    my %params = @_;

    $params{'nacl_type'} = 'state';

    return $self->aggregates_for_vserver(%params);
}

sub vserver_for {
    my $self = shift;
    my %params = @_;

    $params{'package'} = 'Vserver';
    $params{'keys'} = [qw(vserver)];
    $params{'primary_key'} = 'vserver';

    return $self->nacl_for(%params);
}

sub vservers_for_list {
    my $self = shift;
    my %params = @_;

    $params{'function'} = 'vserver_for';

    return $self->nacl_for_list(%params);
}

sub vserver_stask_for {
    my $self = shift;
    my %params = @_;

    $params{'nacl_type'} = 'stask';

    return $self->vserver_for(%params);
}

sub vserver_stasks_for_list {
    my $self = shift;
    my %params = @_;

    $params{'function'} = 'vserver_stask_for';

    return $self->nacl_for_list(%params);
}

sub vserver_state_for {
    my $self = shift;
    my %params = @_;

    $params{'nacl_type'} = 'state';

    return $self->vserver_for(%params);
}

sub vserver_states_for_list {
    my $self = shift;
    my %params = @_;

    $params{'function'} = 'vserver_state_for';

    return $self->nacl_for_list(%params);
}

sub volume_for {
    my $self = shift;
    my %params = @_;

    $params{'package'} = 'Volume';
    $params{'keys'} = [qw(volume vserver)];
    $params{'primary_key'} = 'volume';

    return $self->nacl_for(%params);
}

sub volume_for_list {
    my $self = shift;
    my %params = @_;

    $params{'function'} = 'volume_for';

    return $self->nacl_for_list(%params);
}

sub volume_stask_for {
    my $self = shift;
    my %params = @_;

    $params{'nacl_type'} = 'stask';

    return $self->volume_for(%params);
}

sub volume_stasks_for_list {
    my $self = shift;
    my %params = @_;

    $params{'function'} = 'volume_stask_for';

    return $self->nacl_for_list(%params);
}

sub volume_state_for {
    my $self = shift;
    my %params = @_;

    $params{'nacl_type'} = 'state';

    return $self->volume_for(%params);
}

sub volume_states_for_list {
    my $self = shift;
    my %params = @_;

    $params{'function'} = 'volume_state_for';

    return $self->nacl_for_list(%params);
}

=head2 shell_exec

Executes a command on the specified client and returns the result and exit code.

=cut

sub shell_exec {
    my $self = shift;
    my %params = validate_with(
        params => \@_,
        spec => {
            client => { isa => 'NACL::C::Client' },
            command => { type => SCALAR },
            code_should => { isa => 'CR::Setup::Matcher', optional => $TRUE },
            result_should => { isa => 'CR::Setup::Matcher', optional => $TRUE },
        },
        allow_extra => $TRUE,
    );

    my $client = delete $params{'client'};
    my $command = delete $params{'command'};
    my $code_matcher = delete $params{'code_should'};
    my $result_matcher = delete $params{'result_should'};
    my $return_val = {};

    $return_val->{'result'} = $client->execute(
        command => $command,
        strip_cr => $FALSE,
        %params
    );
    # The execute expands \n to \r\n
    $return_val->{'result'} =~ s/\r\n/\n/g;
    my %code_check = ();
    $code_check{'log_level'} = $params{'log_level'} if defined $params{'log_level'};
    if ( $client->is_windows ) {
        # Windows adds an extra newline to every call
        if ( $return_val->{'result'} eq "\n" ) {
            # Empty results just give a newline
            $return_val->{'result'} = '';
        }
        else {
            # Nonempty results give \r\r\n which turns into \r\n with the replace so remove last two
            $return_val->{'result'} = substr($return_val->{'result'}, 0, -2);
        }
        $code_check{'command'} = 'echo %ERRORLEVEL%';
    } else {
        $code_check{'command'} = 'echo $?';
    }
    $return_val->{'code'} = int($client->execute(%code_check));

    if ( defined $code_matcher ) {
        this($return_val->{'code'})->should($code_matcher);
    }
    if ( defined $result_matcher ) {
        this($return_val->{'result'})->should($result_matcher);
    }

    return $return_val;
}

1;
