
# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary Setup Matcher Library
## @author  Adam.Ciapponi@netapp.com
## @pod     here

=head1 NAME

CR::Setup::Matcher

=head1 DESCRIPTION

This module is a base class for the implementation of RSpec like matchers in perl.  They provide a simple and straightforward interface to
perform comparisons of complex data structures with a minimum of additional code by allowing a developer to
write statements that are more similar to plain english than function calls.

It is not necessary to implement a new Matcher class for each type of comparison you need to make.  Most
common uses such as return value comparison and exception handling are already defined within existing classes.
For information on the basic set of available comparison operators, see L<CR::Setup::Expectation|lib-CR-Setup-Expectation-pm/Expectation>

NOTE: Matcher is a base class and should never be used to instantiate Matcher objects.  See the following child classes
for more information on available Matcher object types to handle specific types of complex comparisons:

=over

=item L<CR::Setup::Matcher::DeeplyEqualMatcher|lib-CR-Setup-Matcher-DeeplyEqualMatcher-pm/DeeplyEqualMatcher>

=item L<CR::Setup::Matcher::EvaluationMatcher|lib-CR-Setup-Matcher-EvaluationMatcher-pm/EvaluationMatcher>

=item L<CR::Setup::Matcher::ExceptionMatcher|lib-CR-Setup-Matcher-ExceptionMatcher-pm/ExceptionMatcher>

=item L<CR::Setup::Matcher::MockObjectMatcher|lib-CR-Setup-Matcher-MockObjectMatcher-pm/MockObjectMatcher>

=item L<CR::Setup::Matcher::PeriodicMatcher|lib-CR-Setup-Matcher-PeriodicMatcher-pm/PeriodicMatcher>

=back

=cut

package CR::Setup::Matcher;

# Compiler directives.
use strict;
use warnings;

# Module import.
use CR::Setup::Unbless;
use Data::Dumper::Concise;
use Tharn qw($Log);

=head1 METHODS

=head2 new

This is the base constructor for the Matcher class. It is initialized with the expected value and the
custom failure messages for exceptions thrown by the object during comparison operations.

Note that the two failure message methods do not need to be overridden, though it is recommended to
do so if there is any need for specialized error messages.

=cut

sub new {
    my ( $class, $expected, $failure_message, $negative_failure_message )
        = @_;
    unless ( defined $failure_message ) {
        $failure_message = "Expected matcher to match but it did not.";
    }
    unless ( defined $negative_failure_message ) {
        $negative_failure_message
            = "Expected matcher not to match but it did.";
    }
    my $self = {
        'expected'                 => $expected,
        'failure_message'          => $failure_message,
        'negative_failure_message' => $negative_failure_message
    };
    bless $self, $class;
    return $self;
}

=head2 matches

This method returns a boolean based on whether or not the target of the matcher matches the expected value.
It is typically overridden in a child class to account for the complex data structures the Matcher object
is used on.

=head3 usage

For methods dealing with primitives, you must use CR::Setup::Unbless::unbless_this before using the value
passed into the method in order to have type safety and nicer failure messages.

See the child classes mentioned above for examples of overridden versions of this method

=cut

sub matches {
    my ( $self, $target ) = @_;
    $self->{'actual'} = unbless_this($target);
    return 1;
}

=head2 failure_message

This method handles the error message returned by the exception thrown when a should statement does not match.

=head3 usage

When overridden, this method is often similar in structure to this method, where the only true difference is in
the output and its formatting.

=cut

sub failure_message {
    my ($self) = @_;
    return
          $self->{'failure_message'}
        . "\nExpected: "
        . Dumper( $self->{'expected'} )
        . "\nActual: "
        . Dumper( $self->{'actual'} ) . "\n";
}

=head2 negative_failure_message

This method handles the error message returned by the exception thrown when a should_not statement does not match.

=head3 usage

As with the failure_message method above, an overridden version of this method is often very similar except for
the output and its formatting.

=cut

sub negative_failure_message {
    my ($self) = @_;
    return
          $self->{'negative_failure_message'}
        . "\nExpected: "
        . Dumper( $self->{'expected'} )
        . "\nActual: "
        . Dumper( $self->{'actual'} ) . "\n";
}

=head2 with_property

This function is used for building up the behaviour of the matcher by adding a property to the instance.
It can then be used in a variety of ways, such as passing a Node interface to a Matcher object, or
to perform comparisons against these defined object properties.

=head3 usage

When redefined in a child class, it is typically used as the return value for the new function.  For
example, to create a Node property:

=begin text

    sub with_node {
        my ( $self, $node ) = @_;
        return $self->with_property( 'node', $node );
    }

=end text

=cut

sub with_property {
    my ( $self, $property, $value ) = @_;
    $self->{$property} = $value;
    return $self;
}
1;
