# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary Setup Evaluation Matcher
## @author  Adam.Ciapponi@netapp.com
## @pod     here

=head1 NAME

CR::Setup::Matchers::EvaluationMatcher

=head1 DESCRIPTION

This module implements a matcher that determines whether it matches by calling eval

=head1 METHODS

=cut

package CR::Setup::Matchers::EvaluationMatcher;

# Compiler directives.
use strict;
use warnings;
use CR::Setup::Constants;
use CR::Setup::Exception;
use CR::Setup::Unbless;
use base qw(CR::Setup::Matcher);

=head2 matches

Builds up a string to evaluate comparisons/expressions and return a boolean value.

=cut

sub matches {
    my ( $self, $target ) = @_;
    $self->{'actual'} = unbless_this($target);

    # Code blocks are always dereferenced into their return value
    if ( ref( $self->{'actual'} ) eq 'CODE' ) {
        $self->{'actual'} = &{ $self->{'actual'} };
    }
    my $str = $self->{'prefix'};
    if ( $self->{'actual_dereferenced'} && ref( $self->{'actual'} ) ) {
        $str .= $self->get_dereference_char('actual');
        $str .= '{$self->{\'actual\'}}';
    }
    else {
        $str .= '$self->{\'actual\'}';
    }
    if ( $self->{'expected_in_expression'} ) {
        $str .= $self->{'middle'};
        if ( $self->{'expected_dereferenced'} && ref( $self->{'expected'} ) )
        {
            $str .= $self->get_dereference_char('expected');
            $str .= '{$self->{\'expected\'}}';
        }
        else {
            $str .= '$self->{\'expected\'}';
        }
    }
    $str .= $self->{'suffix'};
    if ( eval($str) ) {
        return $TRUE;
    }
    else {
        return $FALSE;
    }
}

=head2 with_prefix

Sets the prefix for the eval string before actual.

=cut

sub with_prefix {
    my ( $self, $prefix ) = @_;
    return $self->with_property( 'prefix', $prefix );
}

=head2 with_middle

Sets the middle part for the eval string between actual and expected. Only used if expected is in the expression.

=cut

sub with_middle {
    my ( $self, $middle ) = @_;
    return $self->with_property( 'middle', $middle );
}

=head2 with_suffix

Sets the suffix for the eval string after actual or after expected if it is in the expression.

=cut

sub with_suffix {
    my ( $self, $suffix ) = @_;
    return $self->with_property( 'suffix', $suffix );
}

=head2 with_expected_in_expression

Sets a flag to use the expected value in the eval string.

=cut

sub with_expected_in_expression {
    my ($self) = @_;
    return $self->with_property( 'expected_in_expression', $TRUE );
}

=head2 with_actual_dereferenced

Sets a flag to dereference the actual value if it is a reference.

=cut

sub with_actual_dereferenced {
    my ($self) = @_;
    return $self->with_property( 'actual_dereferenced', $TRUE );
}

=head2 with_expected_dereferenced

Sets a flag to dereference the expected value if it is a reference.

=cut

sub with_expected_dereferenced {
    my ($self) = @_;
    return $self->with_property( 'expected_dereferenced', $TRUE );
}

=head2 get_dereference_char

Gets the character needed to dereference a property of the matcher (actual/expected).

=cut

sub get_dereference_char {
    my ( $self, $property ) = @_;
    my $char;
    if ( ref( $self->{$property} ) eq 'SCALAR' ) {
        $char = '$';
    }
    elsif ( ref( $self->{$property} ) eq 'ARRAY' ) {
        $char = '@';
    }
    elsif ( ref( $self->{$property} ) eq 'HASH' ) {
        $char = '%';
    }
    elsif ( ref( $self->{$property} ) eq 'CODE' ) {
        $char = '&';
    } else {
        CR::Setup::Exception->throw('Cannot dereference type: ' . ref($self->{$property}));
    }
    return $char;
}
1;
