
# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary Setup Exception Matcher
## @author  Adam.Ciapponi@netapp.com
## @pod     here

=head1 NAME

CR::Setup::Matchers::ExceptionMatcher

=head1 DESCRIPTION

This module implements a matcher that determines whether it matches by looking at whether or not an exception was thrown.

=head1 METHODS

=cut

package CR::Setup::Matchers::ExceptionMatcher;

# Compiler directives.
use strict;
use warnings;
use CR::Setup::Constants;
use CR::Setup::Exception;
use Data::Dumper::Concise;
use NATE::BaseException qw(:try);
use Tharn qw($Log);

use base qw(CR::Setup::Matcher);

=head2 matches

Determines matching against whether or not an exception was thrown and if it was thrown was it the correct exception returning a boolean value.

=cut

sub matches {
    my ( $self, $target ) = @_;

    $Log->log(
        'Executing code block and expecting exception to be thrown.')
        if ( $self->{'expected'}->{'status'} eq 'failed' );
    $self->{'actual'} = { 'status' => 'passed' };
    $self->{'expected'}->{'exception'} = $self->{'exception'} if $self->{'exception'};
    $self->{'expected'}->{'exact_error_message'} = $self->{'exact_error_message'} if $self->{'exact_error_message'};
    $self->{'expected'}->{'error_message_like'} = $self->{'error_message_like'} if $self->{'error_message_like'};
    try {
        &$target;
    }
    otherwise {
        my $exception = $@;
        $Log->log('Exception was thrown.');
        $Log->log('Exception reference type: ' . ref($exception));
        $Log->log("Exception message:\n" . Dumper($exception->{'-text'}));
        $Log->debug("Exception:\n" . Dumper($exception) );
        $self->{'actual'}->{'status'}    = 'failed';
        $self->{'actual'}->{'exception'} = $exception;
        $self->{'actual'}->{'message'}   = $exception->{'-text'};
        my $exception_return_ref = $self->{'and_return_exception'};
        ${$exception_return_ref} = $exception if exists $self->{'and_return_exception'};
        my $error_message_return_ref = $self->{'and_return_error_message'};
        ${$error_message_return_ref} = $exception->{'-text'} if exists $self->{'and_return_error_message'};
        &{ $self->{'otherwise'} } if defined $self->{'otherwise'};
    }
    finally {
        &{ $self->{'finally'} } if defined $self->{'finally'};
    };
    if ( $self->{'actual'}->{'status'} ne $self->{'expected'}->{'status'} ) {
        $Log->debug('Does not match because it does not match the expected result status.');
        return $FALSE;
    }
    if ( defined $self->{'exception'} ) {
        my $expected_exception = $self->{'exception'};
        if ( !ref($expected_exception) ) {
            unless (
                $self->{'actual'}->{'exception'}->isa($expected_exception) )
            {
                $Log->debug('Does not match because it is not the expected exception.');
                return $FALSE;
            }
        }
        elsif ( ref($expected_exception) eq "ARRAY" ) {
            my $matches_exception = $FALSE;
            foreach ( @{ $self->{'exception'} } ) {
                if ( $self->{'actual'}->{'exception'}->isa($_) ) {
                    $matches_exception = $TRUE;
                    last;
                }
            }
            unless ($matches_exception) {
                $Log->debug('Does not match because it is not in the list of expected exceptions.');
                return $FALSE;
            }
        } else {
           CR::Setup::Exception->throw("The expected exception ($expected_exception) was not a SCALAR or ARRAYREF.");
        }
    }
    if ( defined $self->{'error_message_like'}
        && $self->{'actual'}->{'message'} !~ $self->{'error_message_like'} )
    {
        $Log->debug('Does not match because it does not match the expected error message regex.');
        return $FALSE;
    }
    if ( defined $self->{'exact_error_message'}
        && $self->{'actual'}->{'message'} ne $self->{'exact_error_message'} )
    {
        $Log->debug('Does not match because it does not match the expected exact error message.');
        return $FALSE;
    }

    $Log->debug('Matches all criteria.');
    return $TRUE;
}

=head2 with_success

Sets appropriate state for a matcher that does not expect an exception to be raised.

=cut

sub with_success {
    my ($self) = @_;
    $self->{'failure_message'}
        = "Expected the code block to not throw an exception matching all of the criteria specified but it did.";
    $self->{'negative_failure_message'}
        = "Expected the code block to throw an exception matching all of the criteria specified but it did not.";
    return $self->with_property( 'expected', { 'status' => 'passed' } );
}

=head2 with_failure

Sets appropriate state for a matcher that does expects an exception to be raised.

=cut

sub with_failure {
    my ($self) = @_;
    $self->{'failure_message'}
        = "Expected the code block to throw an exception matching all of the criteria specified but it did not.";
    $self->{'negative_failure_message'}
        = "Expected the code block to not throw an exception matching all of the criteria specified but it did.";
    return $self->with_property( 'expected', { 'status' => 'failed' } );
}

=head2 with_otherwise

Adds a handler for the otherwise block in the try statement.

=cut

sub with_otherwise {
    my ( $self, $otherwise ) = @_;
    return $self->with_property( 'otherwise', $otherwise );
}

=head2 with_finallly

Adds a handler for the finally block in the try statement.

=cut

sub with_finally {
    my ( $self, $finally ) = @_;
    return $self->with_property( 'finally', $finally );
}

=head2 with_exception

Adds criteria for matching whether or not the exception is the correct one.

=cut

sub with_exception {
    my ( $self, $exception ) = @_;
    return $self->with_property( 'exception', $exception );
}

=head2 with_error_message_like

Adds criteria for matching the message of the exception thrown against a regex. Use qr//.

=cut

sub with_error_message_like {
    my ( $self, $error_message_like ) = @_;
    return $self->with_property( 'error_message_like', $error_message_like );
}

=head2 with_exact_error_message

Adds criteria for matching the message of the exception thrown against an exact string.

=cut

sub with_exact_error_message {
    my ( $self, $exact_error_message ) = @_;
    return $self->with_property( 'exact_error_message',
        $exact_error_message );
}

=head2 and_return_exception

Adds hook for returning the exception

=cut

sub and_return_exception {
    my ( $self, $return_reference ) = @_;
    return $self->with_property( 'and_return_exception', $return_reference );
}

=head2 and_return_error_message

Adds hook for returning the exception error message.

=cut

sub and_return_error_message {
    my ( $self, $return_reference ) = @_;
    return $self->with_property( 'and_return_error_message', $return_reference );
}
1;
