
# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary Setup Mock Object Matcher
## @author  Adam.Ciapponi@netapp.com
## @pod     here

=head1 NAME

CR::Setup::Matchers::MockObjectMatcher

=head1 DESCRIPTION

This module implements a matcher that determines whether it matches by looking at whether or not the specified calls were made to the Test::MockObject.

=head1 METHODS

=cut

package CR::Setup::Matchers::MockObjectMatcher;

# Compiler directives.
use strict;
use warnings;
use CR::Setup::Constants;
use CR::Setup::Matcher;
use Data::Dumper::Concise;
use Params::Validate qw(:all);
use Test::Deep::NoTest;
use Tharn qw($Log);

use base qw(CR::Setup::Matcher);

=head2 matches

Determines matching against whether or not the mock object was called correctly and returns a boolean value.

=cut

sub matches {
    my $self = shift;
    my ($target) = validate_pos( @_, { can => '_calls' } );
    $self->{'actual'} = {};

    $self->{'qualifier'} = 'exactly' if !defined $self->{'qualifier'};
    my $method = $self->{'expected'};
    my $qualifier = $self->{'qualifier'};
    my $number = defined $self->{'number'} ? $self->{number} : 1;
    my $args = $self->{'args'};

    $self->{'expected'} = {
        method => $method,
        qualifier => $qualifier,
        number => $number,
        args => $args,
    };

    if ($qualifier eq 'at position') {
        my $call = $target->_calls->[$number];
        if (!defined $call) {
            $self->{'actual'}->{'call at position'} = 'No call at position ' . $number;
            return $FALSE;
        } else {
            $self->{'actual'}->{'call at position'} = $call;
            my $actual_method = $call->[0];
            my @actual_args = @{$call->[1]};
            shift @actual_args; #self
            $self->{'actual'}->{'number'} = $number;
            $self->{'actual'}->{'method'} = $actual_method;
            if ($method eq $actual_method) {
                if ( defined $args ) {
                    if ( ref($args) eq 'ARRAY' ) {
                        $self->{'actual'}->{'args'} = \@actual_args;
                        if (eq_deeply($args, \@actual_args)) {
                            return $TRUE;
                        } else {
                            return $FALSE;
                        }
                    } elsif ( ref($args) eq 'HASH' ) {
                        my %actual_args_hash = @actual_args;
                        $self->{'actual'}->{'args'} = \%actual_args_hash;
                        if (eq_deeply($args, \%actual_args_hash)) {
                            return $TRUE;
                        } else {
                            return $FALSE;
                        }
                    } else {
                        # Invalid usage
                    }
                } else {
                    return $TRUE;
                }
            } else {
                return $FALSE;
            }
        }
    } else {
        my @calls = @{ $target->_calls };
        $self->{'actual'}->{'call history'} = \@calls;
        my $num_times = 0;
        foreach my $call (@calls) {
            my $actual_method = $call->[0];
            my @actual_args = @{$call->[1]};
            shift @actual_args; #self
            if ($method eq $actual_method) {
                if ( defined $args ) {
                    if ( ref($args) eq 'ARRAY' ) {
                        if (eq_deeply($args, \@actual_args)) {
                            $num_times++;
                        } else {
                            # Doesn't Match
                            next;
                        }
                    } elsif ( ref($args) eq 'HASH' ) {
                        my %actual_args_hash = @actual_args;
                        if (eq_deeply($args, \%actual_args_hash)) {
                            $num_times++;
                        } else {
                            # Doesn't Match
                            next;
                        }
                    } else {
                        # Invalid usage
                    }
                } else {
                    $num_times++;
                }
            } else {
                # Doesn't Match
                next;
            }
        }
        $self->{'actual'}->{'number'} = $num_times;
        if ($qualifier eq 'exactly') {
            if ($num_times == $number) {
                return $TRUE;
            } else {
                return $FALSE;
            }
        } elsif ($qualifier eq 'at most') {
            if ($num_times <= $number) {
                return $TRUE;
            } else {
                return $FALSE;
            }
        } elsif ($qualifier eq 'at least') {
            if ($num_times >= $number) {
                return $TRUE;
            } else {
                return $FALSE;
            }
        }
    }
}

=head2 at_most

Sets the qualifier for the number of matching calls to be at most the specified value.

=cut

sub at_most {
    my $self = shift;
    my ($number) = validate_pos( @_, { type => SCALAR } );
    $self->with_property( 'number', $number );
    return $self->with_property( 'qualifier', 'at most' );
}

=head2 at_least

Sets the qualifier for the number of matching calls to be at least the specified value.

=cut

sub at_least {
    my $self = shift;
    my ($number) = validate_pos( @_, { type => SCALAR } );
    $self->with_property( 'number', $number );
    return $self->with_property( 'qualifier', 'at least' );
}

=head2 at_least

Sets the qualifier for the number of matching calls to be exactly the specified value.

=cut

sub exactly {
    my $self = shift;
    my ($number) = validate_pos( @_, { type => SCALAR } );
    $self->with_property( 'number', $number );
    return $self->with_property( 'qualifier', 'exactly' );
}

=head2 zero

For convenience, sets the number of matching calls or the position of the call to 0.

=cut

sub zero {
    my $self = shift;
    validate_pos(@_);
    return $self->with_property( 'number', 0 );
}

=head2 one

For convenience, sets the number of matching calls or the position of the call to 1.

=cut

sub one {
    my $self = shift;
    validate_pos(@_);
    return $self->with_property( 'number', 1 );
}

=head2 two

For convenience, sets the number of matching calls or the position of the call to 2.

=cut

sub two {
    my $self = shift;
    validate_pos(@_);
    return $self->with_property( 'number', 2 );
}

=head2 three

For convenience, sets the number of matching calls or the position of the call to 3.

=cut

sub three {
    my $self = shift;
    validate_pos(@_);
    return $self->with_property( 'number', 3 );
}

=head2 time

Does nothing but is here so that the matcher can be used in a manner that is a proper english statement.

=cut

sub time {
    my $self = shift;
    validate_pos(@_);
    return $self;
}

=head2 times

Does nothing but is here so that the matcher can be used in a manner that is a proper english statement.

=cut

sub times {
    my $self = shift;
    validate_pos(@_);
    return $self;
}

=head2 with_args

Sets the args that need to be matched for the specified call.

=cut

sub with_args {
    my $self = shift;
    my ($args) = validate_pos( @_, { type => ARRAYREF|HASHREF } );
    return $self->with_property( 'args', $args );
}

=head2 at_position

Sets the qualifier for the call to be at the specified position. This results in only single call comparison.

=cut

sub at_position {
    my $self = shift;
    my ($number) = validate_pos( @_, { type => SCALAR } );
    $self->with_property( 'number', $number );
    return $self->with_property( 'qualifier', 'at position');
}
1;
