# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary Setup Library Composite Mock
## @author  Adam.Ciapponi@netapp.com
## @pod     here

=head1 NAME

CR::Setup::Mocks::CompositeMock

=head1 DESCRIPTION

This module implements a mock that is a composite of Test::MockModule and Test::MockObject to get
the benefits of being able to mock an already loaded module combined with the call history and
additional mocking functionality present in a mock object.

=head1 METHODS

=cut

package CR::Setup::Mocks::CompositeMock;
# Compiler directives.
use strict;
use warnings;

use Data::Dumper::Concise;
use Test::MockModule;
use Test::MockObject;
use Tharn qw($Log);

=head2 new

This initializes both components of the composite mock, a Test::MockObject and a Test::MockModule.

=cut

sub new {
    my $class = shift;
    my %params = @_;

    my $module = $params{'module'};
    if ( defined $params{'no_auto'} ) {
        $module = Test::MockModule->new( $module, no_auto => $params{'no_auto'} );
    }
    else {
        $module = Test::MockModule->new($module);
    }

    my $object = $params{'object'};
    $object = Test::MockObject->new($object);

    my $self = {
        module => $module,
        object => $object,
    };
    bless $self, $class;

    return $self;
}

=head2 bind_method

This creates a mock method on the Test::MockModule that calls the corresponding function on
the Test::MockObject. If method is a reserved method of Test::MockObject (ie new), an alias
can be provided and that will be called instead on the Test::MockObject.

=cut

sub bind_method {
    my $self = shift;
    my %params = @_;

    my $module = $self->module;
    my $object = $self->object;
    my $method = $params{'method'};
    my $alias  = $params{'alias'};

    if (defined $method) {
        if (defined $alias) {
            $module->mock(
                $method,
                sub {
                    my $class = shift;
                    return $object->$alias(@_);
                }
            );
        } else {
            $module->mock(
                $method,
                sub {
                    my $class = shift;
                    return $object->$method(@_);
                }
            );
        }
    } else {
        # Throw Exception
    }

    return $self;
}

=head2 module

Getter for the Test::MockModule.

=cut

sub module {
    my $self = shift;
    return $self->{'module'};
}

=head2 object

Getter for the Test::MockObject.

=cut

sub object {
    my $self = shift;
    return $self->{'object'};
}

=head2 DESTROY

The destructor explicitly unmocks methods on the MockModule as those are the most dangerous
to leave around.

=cut

sub DESTROY {
    my $self = shift;
    $self->module->unmock_all() if defined $self->module;
    undef $self;
}

1;
