# Copyright (c) 2013 Network Appliance, Inc.
# All rights reserved.

## @summary Sfmod Transfer Library
## @author  dl-crqa-auto@netapp.com
## @pod     here

=head1 NAME

CR::Setup::SfmodTransfer

=head1 DESCRIPTION

Module to encapsulate SFMoD operations.

=head1 METHODS

=cut

package CR::Setup::SfmodTransfer;
use CR::Setup::Constants;
use CR::Setup::Expectation;
use List::Util qw(first);
use NACL::C::VolumeFileMove;
use NACL::C::VolumeFileRepositoryFhmap;
use NACL::STask::Volume;
use Params::Validate qw(:all);
use Tharn qw($Log);

sub new {
    my $class  = shift;
    my %params = @_;

    if ( !defined $params{'fhmap'} ) {
        $params{'fhmap'} = NACL::CS::VolumeFileRepositoryFhmap->fetch(
            command_interface => $params{'nodes'}->[0],
            filter            => {
                vserver => $params{'vserver'}->vserver,
                path    => $params{'file_path'}->{'on_filer'},
            },
        );
    }
    if ( !defined $params{'infinitevol'} ) {
        $params{'infinitevol'} = NACL::STask::Volume->find(
            command_interface => $params{'nodes'}->[0],
            filter            => {
                vserver        => $params{'vserver'}->vserver,
                'volume-style' => 'infinitevol'
            },
        );
    }
    if ( !defined $params{'infinitevol_jp'} ) {
        $params{'infinitevol_jp'} = $params{'infinitevol'}
            ->get_one_state_attribute('junction-path');
    }

    if ( !defined $params{'src_dc'} ) {
        $params{'src_dc'} = NACL::STask::Volume->find(
            command_interface => $params{'nodes'}->[0],
            filter            => {
                vserver => $params{'vserver'}->vserver,
                volume  => $params{'fhmap'}->data_constituent
            },
        );
    }
    if ( !defined $params{'src_dc_jp'} ) {
        $params{'src_dc_jp'}
            = $params{'src_dc'}->get_one_state_attribute('junction-path');
    }

    if ( !defined $params{'dst_dc'} ) {
        $params{'dst_dc'} = NACL::STask::Volume->find(
            command_interface => $params{'nodes'}->[0],
            filter            => {
                vserver => $params{'vserver'}->vserver,
                volume  => '!' . $params{'fhmap'}->data_constituent,
                'constituent-role' => 'data'
            },
        );
    }
    if ( !defined $params{'dst_dc_jp'} ) {
        $params{'dst_dc_jp'}
            = $params{'dst_dc'}->get_one_state_attribute('junction-path');
    }

    $params{'dc_file_path'} = $params{'fhmap'}->data_path;
    $params{'dc_file_path'} =~ s/\Q$params{'src_dc_jp'}\E//;
    my $basename
        = ( File::Spec::Unix->splitpath( $params{'dc_file_path'} ) )[-1];
    $params{'dc_directory'} = $params{'dc_file_path'};
    $params{'dc_directory'} =~ s/\Q\/$basename\E$//;

    $params{'dc_file'}       = $basename;
    $params{'dc_file_query'} = '*/' . $basename;

    my $self = \%params;
    bless $self, $class;
    return $self;
}

sub move {
    my $self   = shift;
    my %params = validate_with(
        params => \@_,
        spec   => {
            max_throughput => {
                type    => SCALAR,
                default => 0,
            },
            report_time => {
                type    => SCALAR,
                default => 10,
            },
            scanner_paused => {
                type    => SCALAR,
                default => 'false',
            },
            transfer_method => {
                type    => SCALAR,
                regex   => qr/^ksmf|policy_check$/,
                default => 'policy_check',
            },
            validate_success => {
                type    => SCALAR,
                default => $FALSE,
            },
        },
    );

    if (   $self->{'dst_dc'}->get_one_state_attribute('junction-path') ne '-'
        && $params{'transfer_method'} eq 'ksmf' )
    {
        $self->{'systemshell'}->execute_raw_command(
            command => 'sudo mkdir -p -m 777 '
                . File::Spec::Unix->join(
                '/clus',
                $self->{'vserver'}->vserver,
                substr( $self->{'dst_dc_jp'},    1 ),
                substr( $self->{'dc_directory'}, 1 )
                )
        );
    }

    my $dst_node = first {
        $_->node eq $self->{'dst_dc'}->get_one_state_attribute('node');
    }
    @{ $self->{'nodes'} };

    my $response;
    my $success_message;
    if ( $params{'transfer_method'} eq 'ksmf' ) {
        $success_message = qr/^\s*$/m;
        $response
            = $dst_node->apiset->execute_raw_command(
                  command => 'debug smdb table sfod_operation_node create'
                . ' -source-vserver-id '
                . $self->{'vserver'}->get_one_state_attribute('id')
                . ' -source-msid '
                . $self->{'src_dc'}->get_one_state_attribute('msid')
                . ' -source-dsid '
                . $self->{'src_dc'}->get_one_state_attribute('dsid')
                . ' -source-file-path '
                . $self->{'dc_file_path'}
                . ' -destination-dsid '
                . $self->{'dst_dc'}->get_one_state_attribute('dsid')
                . ' -destination-file-path '
                . $self->{'dc_file_path'}
                . ' -scanner-paused '
                . $params{'scanner_paused'}
                . ' -report-time '
                . $params{'report_time'}
                . ' -max-throughput '
                . $params{'max_throughput'}
                . ' -plugin-interfaces 8'
                . ' -hold-quiescence false'
                . ' -operation move' );
    }
    elsif ( $params{'transfer_method'} eq 'policy_check' ) {
        $success_message = qr/Started file move with job UUID/;
        $response
            = $dst_node->apiset->execute_raw_command(
                  command => 'repository policy check'
                . ' -vserver '
                . $self->{'vserver'}->vserver
                . ' -path '
                . $self->{'file_path'}->{'on_filer'}
                . ' -msid '
                . $self->{'dst_dc'}->get_one_state_attribute('msid')
                . ' -scanner-paused '
                . $params{'scanner_paused'}
                . ' -max-throughput '
                . $params{'max_throughput'}
                . ' -report-time '
                . $params{'report_time'} );
    }

    if ( $params{'validate_success'} ) {
        this($response)->should( Match->regex($success_message) );
    }

    return $response;
}

sub wait_for_job {
    my $self   = shift;
    my %params = validate_with(
        params => \@_,
        spec   => {
            state    => { type => SCALAR, },
            fhmap_dc => {
                callbacks => {
                    'isa nacl volume' => sub {
                        $_[0]->isa('NACL::C::Volume')
                            || $_[0]->isa('NACL::CS::Volume')
                            || $_[0]->isa('NACL::STask::Volume');
                    },
                },
                optional => $TRUE,
            },
        },
    );

    this_block {
        my $job = NACL::CS::VolumeFileMove->fetch(
            command_interface => $self->{'nodes'}->[0],
            filter =>
                { 'destination-file-path' => [ $self->{'dc_file_query'} ], },
            allow_empty => 1,
        );
        $job ? $job->scanner_status : '';
    }
    should( eventually( Be->lexically_equal_to( $params{'state'} ) )->every(1)
            ->within(2)->minutes() );

    if ( defined $params{'fhmap_dc'} ) {

        # ls -l will usually flush stale clusFS info
        $self->{'systemshell'}->execute_raw_command(
            command => 'ls -l '
                . File::Spec::Unix->join(
                '/clus',
                $self->{'vserver'}->vserver,
                $self->{'file_path'}->{'on_filer'}
                )
        );

        # fhmap show uses clusFS which may not instantly update across the
        # cluster
        this_block {
            my $fh_after = NACL::CS::VolumeFileRepositoryFhmap->fetch(
                command_interface => $self->{'nodes'}->[0],
                filter            => {
                    vserver => $self->{'vserver'}->vserver,
                    path    => $self->{'file_path'}->{'on_filer'},
                },
            );
            $fh_after->data_constituent;
        }
        should(
            eventually(
                Be->lexically_equal_to( $params{'fhmap_dc'}->volume )
                )->every(1)->within(30)->seconds
        );
    }

    return;
}

sub move_back {
    my $self   = shift;
    my %params = validate_with(
        params => \@_,
        spec   => {
            max_throughput => {
                type    => SCALAR,
                default => 0,
            },
            report_time => {
                type    => SCALAR,
                default => 10,
            },
            scanner_paused => {
                type    => SCALAR,
                default => 'false',
            },
            transfer_method => {
                type    => SCALAR,
                regex   => qr/^ksmf|policy_check$/,
                default => 'policy_check',
            },
            validate_success => {
                type    => SCALAR,
                default => $FALSE,
            },
        },
    );

    my $src_node = first {
        $_->node eq $self->{'src_dc'}->get_one_state_attribute('node');
    }
    @{ $self->{'nodes'} };

    my $response;
    my $success_message;
    if ( $params{'transfer_method'} eq 'ksmf' ) {
        $success_message = qr/^\s*$/m;
        $response
            = $src_node->apiset->execute_raw_command(
                  command => 'debug smdb table sfod_operation_node create'
                . ' -source-vserver-id '
                . $self->{'vserver'}->get_one_state_attribute('id')
                . ' -source-msid '
                . $self->{'dst_dc'}->get_one_state_attribute('msid')
                . ' -source-dsid '
                . $self->{'dst_dc'}->get_one_state_attribute('dsid')
                . ' -source-file-path '
                . $self->{'dc_file_path'}
                . ' -destination-dsid '
                . $self->{'src_dc'}->get_one_state_attribute('dsid')
                . ' -destination-file-path '
                . $self->{'dc_file_path'}
                . ' -scanner-paused '
                . $params{'scanner_paused'}
                . ' -max-throughput '
                . $params{'max_throughput'}
                . ' -report-time '
                . $params{'report_time'}
                . ' -plugin-interfaces 8'
                . ' -hold-quiescence false'
                . ' -operation move' );
    }
    elsif ( $params{'transfer_method'} eq 'policy_check' ) {
        $success_message = qr/Started file move with job UUID/;
        $response
            = $src_node->apiset->execute_raw_command(
                  command => 'repository policy check'
                . ' -vserver '
                . $self->{'vserver'}->vserver
                . ' -path '
                . $self->{'file_path'}->{'on_filer'}
                . ' -msid '
                . $self->{'src_dc'}->get_one_state_attribute('msid')
                . ' -scanner-paused '
                . $params{'scanner_paused'}
                . ' -max-throughput '
                . $params{'max_throughput'}
                . ' -report-time '
                . $params{'report_time'} );
    }

    if ( $params{'validate_success'} ) {
        this($response)->should( Match->regex($success_message) );
    }

    return $response;
}

sub get_job {
    my $self = shift;

    my $job = NACL::C::VolumeFileMove->find(
        command_interface => $self->{'nodes'}->[0],
        filter =>
            { 'destination-file-path' => [ $self->{'dc_file_query'} ], },
    );

    return $job;
}

1;
