# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary Setup Space Library - 8.2 SierraNevada Release
## @author  galane@netapp.com
## @pod     here

=head1 NAME

CR::Setup::Space::SierraNevada

=head1 DESCRIPTION

Space module for the SierraNevada release.

=head1 METHODS

=cut

package CR::Setup::Space::SierraNevada;

use strict;
use warnings;
use CR::Setup::Constants::SierraNevada;
use NACL::CS::VolumeShowFootprint;
use Params::Validate qw(:all);

use base qw(CR::Setup::Space::Base);

##------------------------------------------------------------
# Public Interfaces
##------------------------------------------------------------

=head2 aggregate_with_least_usable_space

Finds and returns the aggeregate in a given aggregate list that has the least usable
space

=cut
sub aggregate_with_least_usable_space {
    my $self = shift;
    my %params = validate_with(
        params => \@_,
        spec   => {
            aggregate_list => { type => ARRAYREF, optional => 1 },
            vserver        => { type => SCALAR,   optional => 1 },
            space_required => { type => SCALAR,   optional => 1 },
        }
    );

    my @aggrs;

    if( defined $params{vserver} ) {
        @aggrs = @{
            $self->{helpers}->aggregate_states_for_vserver( vserver => $params{vserver} )
        };
    } elsif( defined $params{aggregate_list} ) {
        @aggrs = @{
            $self->{helpers}->aggregate_states_for_list(
                list => $params{aggregate_list}
            )
        };
    } else {
        CR::Setup::Exception->throw(
            'Must provide vserver or aggregate_list parameter to aggregate_with_least_usable_space()'
        );
    }

    if ( scalar(@aggrs) > 0 ) {
        @aggrs = sort {
            $self->usable_space_for_aggregate(aggregate_state => $a) <=>
            $self->usable_space_for_aggregate(aggregate_state => $b)
        } @aggrs;
    } else {
        CR::Setup::Exception->throw(
            'No aggregates specified or empty aggregate list for aggregate_with_least_usable_space()'
        );
    }

    if ( defined $params{space_required} ) {
        if ( $self->usable_space_for_aggregate( aggregate_state => $aggrs[0] )
            < $params{space_required} )
        {
            CR::Setup::Exception->throw(
                      'No aggregates able to accomodate the specified size '
                    . $params{space_required}
                    . ' bytes for aggregate_with_least_usable_space()' );
        }
    }

    return $aggrs[0];
}

=head2 aggregate_with_most_usable_space

Finds and returns the aggeregate in a given aggregate list that has the highest usable
space

=cut
sub aggregate_with_most_usable_space {
    my $self = shift;
    my %params = validate_with(
        params => \@_,
        spec   => {
            aggregate_list => { type => ARRAYREF, optional => 1 },
            space_required => { type => SCALAR,   optional => 1 },
            vserver        => { type => SCALAR,   optional => 1 },
        }
    );

    my @aggrs;

    if( defined $params{vserver} ) {
        @aggrs = @{
            $self->{helpers}->aggregate_states_for_vserver( vserver => $params{vserver} )
        };
    } elsif( defined $params{aggregate_list} ) {
        @aggrs = @{
            $self->{helpers}->aggregate_states_for_list(
                list => $params{aggregate_list}
            )
        };
    } else {
        CR::Setup::Exception->throw(
            'Must provide vserver or aggregate_list parameter to aggregate_with_most_usable_space()'
        );
    }

    if ( scalar(@aggrs) > 0 ) {
        @aggrs = reverse sort {
            $self->usable_space_for_aggregate(aggregate_state => $a) <=>
            $self->usable_space_for_aggregate(aggregate_state => $b)
        } @aggrs;
    } else {
        CR::Setup::Exception->throw(
            'No aggregates specified or empty aggregate list for aggregate_with_most_usable_space()'
        );
    }

    if ( defined $params{space_required} ) {
        if ( $self->usable_space_for_aggregate( aggregate_state => $aggrs[0] )
            < $params{space_required} )
        {
            CR::Setup::Exception->throw(
                      'No aggregates able to accomodate the specified size '
                    . $params{space_required}
                    . ' bytes for aggregate_with_most_usable_space()' );
        }
    }

    return $aggrs[0];
}

=head2 aggregates_with_more_than_usable_space

Finds and returns aggregates with more than the specified space required

=cut
sub aggregates_with_more_than_usable_space {
    my $self = shift;
    my %params = validate_with(
        params => \@_,
        spec   => {
            aggregate_list => { type => ARRAYREF, optional => 1 },
            vserver        => { type => SCALAR,   optional => 1 },
            space_required => { type => SCALAR },
        }
    );

    my @aggrs;
    my @aggrs_with_more_space;

    if( defined $params{vserver} ) {
        @aggrs = @{
            $self->{helpers}->aggregate_states_for_vserver( vserver => $params{vserver} )
        };
    } elsif( defined $params{aggregate_list} ) {
        @aggrs = @{
            $self->{helpers}->aggregate_states_for_list(
                list => $params{aggregate_list}
            )
        };
    } else {
        CR::Setup::Exception->throw(
            'Must provide vserver or aggregate_list parameter to aggregate_with_more_than_usable_space()'
        );
    }

    if ( scalar(@aggrs) > 0 ) {
        @aggrs_with_more_space = grep {
            $self->usable_space_for_aggregate( aggregate_state => $_ )
                > $params{space_required}
        } @aggrs;
    } else {
        CR::Setup::Exception->throw(
            'No aggregates specified or empty aggregate list for aggregate_with_more_than_usable_space()'
        );
    }

    if ( !(scalar(@aggrs_with_more_space) > 0) ) {
        CR::Setup::Exception->throw(
                'No aggregates with more usable space than '
                . $params{space_required}
                . ' bytes in aggregate_with_more_than_usable_space()' );
    }

    return \@aggrs_with_more_space;
}

=head2 usable_space_for_aggregate

Determines the usable space for an aggregate and returns that number.

=cut
sub usable_space_for_aggregate {
    my $self = shift;
    my %params = validate_with(
        params => \@_,
        spec   => {
            aggregate => { type => SCALAR, optional => 1 },
            aggregate_state => { isa => 'NACL::CS::Aggregate', optional => 1 },
            volume_footprints => { type => ARRAYREF, optional => 1 },
        }
    );

    my $aggregate_state;
    my $usable_space;
    my @volume_footprints;

    if ( !defined $params{volume_footprints} ) {
        @volume_footprints = NACL::CS::VolumeShowFootprint->fetch(
            command_interface => $self->{node},
            filter            => {
                aggregate => $params{'aggregate'},
            },
            allow_empty => 1
        );
    } else {
        @volume_footprints = @{ $params{volume_footprints} };
    }

    if( defined $params{aggregate_state} ) {
        $aggregate_state = $params{aggregate_state};
    } else {
        $aggregate_state = $self->{helpers}->aggregate_state_for( aggregate => $params{aggregate} );
    }

    if( scalar(@volume_footprints) > 0 ) {
        $usable_space = $aggregate_state->availsize;

        foreach my $vol_footprint ( @volume_footprints ) {
            $usable_space = $usable_space - $self->calculate_wafl_reserve(aggregate_state => $aggregate_state);

            $usable_space = $usable_space - $self->calculate_dedupe_reserve(
                aggregate_state  => $aggregate_state,
                volume_footprint => $vol_footprint,
            );

            $usable_space = $usable_space - $self->calculate_volume_reserve(
                volume_footprint => $vol_footprint,
            );
        }
    } else {
        $usable_space = $aggregate_state->availsize - 0.09 * $aggregate_state->size;
    }

    return $usable_space;
}

##------------------------------------------------------------
# Private methods
##------------------------------------------------------------
sub calculate_dedupe_reserve {
    my $self = shift;
    my %params = validate_with(
        params => \@_,
        spec   => {
            aggregate_state  => { isa => 'NACL::CS::Aggregate' },
            volume_footprint => { isa => 'NACL::CS::VolumeShowFootprint' },
        }
    );

    my $metafile_footprint
        = $params{volume_footprint}->dedupe_metafiles_footprint =~ /^\d+$/
        ? $params{volume_footprint}->dedupe_metafiles_footprint
        : 0;
    my $metafile_temp_footprint
        = $params{volume_footprint}->dedupe_metafiles_temporary_footprint =~ /^\d+$/
        ? $params{volume_footprint}->dedupe_metafiles_temporary_footprint
        : 0;

    my $dedupe_reserve = $params{aggregate_state}->size * 0.06 - ( $metafile_footprint + $metafile_temp_footprint );

    return $dedupe_reserve;
}

sub calculate_volume_reserve {
    my $self = shift;
    my %params = validate_with(
        params => \@_,
        spec   => { volume_footprint => { isa => 'NACL::CS::VolumeShowFootprint' }, }
    );

    my $vol_footprint = $params{volume_footprint};

    my $vol_state = $self->{helpers}->volume_state_for(
        volume  => $vol_footprint->volume,
        vserver => $vol_footprint->vserver,
    );

    my $vol_reserve = 0;

    if( $vol_state->space_guarantee eq 'volume' ) {
        my $reserve = ( $vol_state->size / $THICK_PROVISIONING_RATIO ) - $vol_footprint->total_footprint;
        $vol_reserve = $reserve if ( $reserve > 0 );
    } elsif( $vol_state->space_guarantee eq 'none' ) {
        $vol_reserve = $vol_state->size / $THIN_PROVISIONING_RATIO;
    }

    return $vol_reserve;
}

sub calculate_wafl_reserve {
    my $self = shift;
    my %params = validate_with(
        params => \@_,
        spec   => { aggregate_state => { isa => 'NACL::CS::Aggregate' }, }
    );

    my $wafl_reserve = $params{aggregate_state}->size * 0.03;

    return $wafl_reserve;
}

1;
