=head1

This package is purely to fix issues in TCD that make it undesirable with the new driver.

Issues fixed:
    With the repeated setup and cleanup, there is no need to abort file execution on setup failure.

=cut
package CR::Setup::TCD;

#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/main/lib/TCD.pm#30 $

## @summary Provides ability to control what tests are executed
## in a script containing multiple tests
##
## @author knopp@netapp.com, pressley@netapp.com
## @status public
## @pod here

package TCD;

use vars qw(
    @StdMethods
    %HshTestRepeat
);

our @EXPORT=qw(
   new
   description
   display
   get_dir
   get_loghdl
   get_name
   get_param
   get_testcase
   get_next_testcase
   get_previous_testcase
   get_previous_result
   get_worst_result
   worst_subtest_result
   log
   ndisplay
   nlog
   read_config
   run_tests
   step
   get_summary
   testcases
);

#
# Export Global variables
#
$TCD::SCRIPT = -8;
$TCD::CONFIG = -7;
$TCD::NA     = -6;
$TCD::FATAL  = -5;
$TCD::UNDEF  = -4;
$TCD::DISABLE= -1;
$TCD::PASS   =  1;
$TCD::FAIL   =  0;
$TCD::TRUE   =  1;
$TCD::FALSE  =  0;
$TCD::POSITIVE = 1;
$TCD::NEGATIVE = 0;

#
# Compiler directives
#
use strict;

#
# Module import
#
use POSIX qw(strftime);
use Scalar::Util qw(blessed);
use Cwd qw(abs_path);
use Time::Local;
use IO::Tee;
use IO::File;
use Tie::IxHash;
use File::Compare;
use File::Basename;
use File::stat;
use Error;
use CR::Setup::TCDTimer;
use Burt qw(testcase_start);
use Exporter;
use Data::Dumper;
use Tharn qw($Log);

#
# List of files we inherit from
#
our @ISA = qw(Exporter Error Timer);

sub testcases
{
    my (@TESTS)=@_;
    $Log->enter("");

    package main;
    use vars qw($Test);

    $Test = new TCD(-testcases=>[@TESTS]);
    if ($Test->error){
        $Test->log($Test->errmsg);
        $Log->exit("");
        return $TCD::FAIL;
    }

    # Performs method callback
    $Test->run_test();
    if ($Test->error){
        $Test->log($Test->errmsg);
        $Log->exit("");
        return $TCD::FAIL;
    }
    $Log->exit("");
    exit(0);
}

###############################################################################
# Method:    new
# Objective: Create a new TCD object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new
{
    my ($class,%args) = @_;
    $Log->enter("");
    my (
        $client,
        $config,
        $fullpath,
        $use_testcases,
        $self,
        @argv,
        @tee,
        @methodlist,
        %testcases,
       );

    $self = {};
    bless $self;

    no strict 'refs';

    $self->{stdout}=(defined(fileno(\*ntest::NTEST_STDOUT)))
                    ? \*ntest::NTEST_STDOUT
                    : new IO::Tee();

    $self->{stderr}=(defined(fileno(\*ntest::NTEST_STDERR)))
                    ? \*ntest::NTEST_STDERR
                    : new IO::Tee();

    $self->{log} = $self->{stdout};
    select((select($self->{log}), $| = 1)[0]);

    use strict 'refs';

    # Retrieve Test Case arguments
    # from command line.
    @argv =($self->get_param("TESTCASES") or $self->get_param("ARGS"))
        ? split('\s+',($self->get_param("TESTCASES") or $self->get_param("ARGS")))
        : "all";

    if (defined($args{-options})) {
        my $optstr = $args{-options};
        $optstr=~ s/\s+//g; # strip any whitespace
        my @optionlist=split(/\,/,$optstr);
        push @argv,(@optionlist);
    }

    $self->{argv}=\@argv;

    unless (scalar(@argv)) {
        $self->_set_err("="x80,
                        "Error: No test parameters of options were specified for the\n".
                        "       thpl parameter 'TESTCASES' on the command line.",
                        "="x80);

        Tharn::logresult(NATE::Result::Config->new("Invalid command line arguments",abortive=>0));
        $Log->exit("");
        return $self;
    }

    # Check to see that we were given either a 'tests' or a 'testcases'
    # parameter.

    no strict;
    tie %testcases,Tie::IxHash;
    use strict;

    my $cnt=0;
    push my @testcases,@{$args{-testcases}};
    while ($cnt < scalar(@testcases)){
        $testcases{$testcases[$cnt]}=$testcases[$cnt+1];
        push @methodlist, $testcases[$cnt];
        $cnt+=2;
    }

    $self->{testcases}=\%testcases;
    $self->{methodlist}=\@methodlist;

    # Make sure we were given at least one test
    unless (scalar(@{$self->{methodlist}})>0) {
        $self->_set_err("Error: TCD::new either parameter \"tests\" ",
                        "or \"testcases\" must be specified");
        $Log->exit("");
        return $self;
    }

    # Validate input argument list to thpl file.
    $self->_args_list($self->{methodlist});
    if ($self->error) {
        $Log->exit("");
        return $self
    };

    # Test Name
    $fullpath=$self->get_param("UTESTNAME");
    if ($self->error) {
        $Log->exit("");
        return $self
    }

    # Project, testcase path, and testcase name are located
    # after the 'test' directory name.
    # '.../test/[project]/[tcpath]/[tesTCD]'
    my ($tcpath,$testname)=$fullpath=~/test\/(.*)\/(.*).thpl$/;
    $self->{tcpath}=$tcpath;
    $self->{suite}=$tcpath;
    $self->{suite}=~s/\//_/g;

    # Testcase information
    my ($base,$dir,$ext)=fileparse($ENV{TESTNAME},'\..*');
    $self->{testscript_dir}=$dir;
    $self->{testscript}=$base;

    if ($self->_is_quiet()){
        $self->{stdout} = IO::Tee->new();
        $self->{log}    = IO::Tee->new();
        select((select($self->{log}), $| = 1)[0]);
    }

    # Display testcase name being executed
    $self->display("","Test Script: ".$self->{testscript}.".thpl");

    # Store standard methods in global array.
    push @StdMethods,"init","setup","cleanup","uninit";

    $self->{testbed}=`/bin/pwd`;
    chomp($self->{testbed});
    $self->nlog("TESTBED: ".$self->{testbed});

    # Retrieve variables from config file
    $config=$self->get_param("CONFIG");

    $self->_clear_err() if (defined $config);

    if (defined $config){
            $self->read_config($config);
            return $self if ($self->error);
    }

    $self->{next}     = "";
    $self->{previous} = "";
    $Log->exit("");
    return $self;
}

###############################################################################
# Method:    description
# Objective: Display test case header
# Details:   See POD documentation at the end of this file.
###############################################################################
sub description
{
    my ($self, @name)=@_;
    $Log->enter("");
    my(
       @header,
       $bFirst,
       $newline,
       $type,
       );

    # initialize
    $bFirst=1;

    # hold a copy of description in tcase to use for logging.
    $self->{description}="@name";

    unless (@name) {
        if (ref($self->_get_nuffo_description())eq "ARRAY"){
             push @name,@{$self->_get_nuffo_description()};
        }
        else {
            push @name,$self->_get_nuffo_description();
        }
    }

    $type = ($self->{testname} =~ /^init$|^setup$|^cleanup$|^uninit$/)
        ?  "TEST CONFIG"
        :  "TEST CASE  ";

    push @header,
    "\n",
    "#"x 80,
    "##  TEST SCRIPT: ".$self->{testscript}.".thpl",
    "##  $type: ".$self->{testname}."()";


    foreach my $line (@name[0..scalar(@name)-1]){

        $newline=($bFirst==1)
            ? "##  DESCRIPTION: ".$line
            : "##               ".$line;

        push @header,$newline;

        $bFirst=0;
    }

    push @header,"#"x 80;

    $Log->comment(join ("\n", @header));

    $self->log(@header);

    $Log->exit("");
}

###############################################################################
# Method:    display
# Objective: print output to stdout
# Details:   See POD documentation at the end of this file.
###############################################################################
sub display
{
    my ($self, @line) = @_;
    $Log->enter("");

    print {$self->{stdout}} join("\n",@line)."\n";
    $self->{stdout}->flush();
    $Log->exit("");
}

###############################################################################
# Method:    get_dir
# Objective: get testcase directory path
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_dir
{
    my ($self) = @_;
    $Log->enter("");

    $self->_clear_err();


    if (defined $self->{testscript_dir}){
        $Log->exit("");
        return $self->{testscript_dir};
    }
    else {
        $self->_set_err("Error: TCD::get_dir() TCD object is missing testcase directory");
        $Log->exit("");
        return "";
    }
}

###############################################################################
# Method:    get_loghdl
# Objective: Return a reference to the log handle used to log output
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_loghdl
{
    my ($self) = @_;
    $Log->enter("");
    $Log->exit("");
    return \$self->{log};
}

################################################################################
# Subroutine Name: get_name
# Objective:       Returns script name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name
{
    my ($self) = @_;
    $Log->enter("");

    $self->_clear_err();

    if (defined $self->{testscript}){
        $Log->exit("");
        return $self->{testscript};
    }
    else {
        $self->_set_err("Error: TCD::get_name() Test case name is not set");
        $Log->exit("");
        return "";
    }
}

################################################################################
# Subroutine Name: get_testcase
#                  get_next_testcase
#                  get_previous_testcase
# Objective:       Return appropriate test case name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_testcase
{
    my ($self) = @_;
    $Log->enter("");

    $self->_clear_err();

    if (defined $self->{testname}){
        $Log->exit("");
        return $self->{testname};
    }
    else {
        $self->_set_err("Error: TCD::get_testcase() Test case name is not set");
        $Log->exit("");
        return "";
    }
}

sub get_next_testcase
{
    my ($self) = @_;
    $Log->enter("");

    $self->_clear_err();

    if (defined $self->{next}){
        $Log->exit("");
        return $self->{next};
    }
    else {
        $self->_set_err("Error: TCD::get_next_testcase() Next test case name is not set");
        $Log->exit("");
        return "";
    }
}

sub get_previous_testcase
{
    my ($self) = @_;
    $Log->enter("");

    $self->_clear_err();

    if (defined $self->{previous}){
        $Log->exit("");
        return $self->{previous};
    }
    else {
        $self->_set_err("Error: TCD::get_previous_testcase() Next test case name is not set");
        $Log->exit("");
        return "";
    }
}

################################################################################
# Subroutine Name: get_previous_result()
# Objective:       Return appropriate test case name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_previous_result
{
    my ($self) = @_;
    $Log->enter("");

    $self->_clear_err();

    my $result =  $self->{result};
    if ($result){
        $Log->exit("");
        return $result->[-1]->{status}; # note array[-1] is last element
    }
    else{
        $Log->exit("");
        $self->_set_err("Error: TCD::get_previous_result() No results found");
        return;
    }
}

################################################################################
# Subroutine Name: get_worst_result()
# Objective:       Return the worst result of all the testcases
# Details:         See POD documentation at the end of this file
################################################################################
sub get_worst_result
{
    my ($self) = @_;
    $Log->enter("");

    $self->_clear_err();

    my $worst_result = $TCD::PASS;

    my $results =  $self->{result};
    foreach my $result (@{$results}) {

        # if this result is worse than the worst result
        # update the worst result
        if ($result->{status} < $worst_result) {
            $worst_result = $result->{status};
        }
    }

    # set an error if there is no worst result
    if (not defined($worst_result)) {
        $self->_set_err("Could not determine worst result");
        $Log->exit("");
        return "";
    }

    $Log->exit("");
    return $worst_result;
}

###############################################################################
# Method:    get_param
# Objective: get parameter value or use default if not set
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_param
{
    my ($self, $param,$default) = @_;
    $Log->enter("");
    my (
        $value,
        @value,
        );

    $self->_clear_err();
    Tharn::param($param,"-default", undef);
    $value=$Tharn::Params{$param}->{value};
    $value=$default unless (defined $value);

    # If called in list context, split values and return array.
    if (wantarray) {
        # Some parameters (FILER, etc.) allow multiple values to
        # be separated by either ',' or ':'.
        # If the value ends in '.cfg', the caller is requesting the
        # values to be read from a config file, rather than specifying
        # them all on the command line.  This makes specifying a large
        # number of clients much easier.
        if ($value =~ /\.cfg$/) {
            open(CFG, $value) or
            do {
                $self->_set_err("Can't open config file $value: $!");
                return;
            };

            while (my $line=<CFG>) {
                chomp $line;

                # Skip blank or commented out lines.
                next if ($line =~ /^\#/ or $line =~ /^\s*$/);
                push @value, $line;
            }
        }
        else {
            @value = split(/[:|,]/, $value);
        }
        $Log->exit("");
        return @value;
    }
    else {
        $Log->exit("");
        return $value;
    }
}

###############################################################################
# Method:    log
# Objective: Log test output to stdout and file
# Details:   See POD documentation at the end of this file.
###############################################################################
sub log
{
    my ($self, @line) = @_;
    $Log->enter("");

    print {$self->{log}} join("\n", @line) . "\n";
    $self->{log}->flush();
    $Log->exit("");
}

###############################################################################
# Method:    ndisplay
# Objective: print output to stdout and nate logs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub ndisplay
{
    my ($self, @line) = @_;
    $Log->enter("");

    if (@line) {
        $Log->comment(join ("\n", @line));
        $self->display(@line);
    }
    $Log->exit("");
}

###############################################################################
# Method:    nlog
# Objective: Log test output to stdout, tcase, and nate logs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nlog
{
    my ($self, @line) = @_;
    $Log->enter("");

    if (@line) {
        $Log->comment(join("\n", @line));
        $self->log(@line);
    }
    $Log->exit("");
}

##########################################################################4#####
# Method:    read_config ()
# Objective: read config file for environment variables
# Details:   See POD documentation at the end of this file.
###############################################################################
sub read_config
{
    my ($self,$config)=@_;

    $Log->enter("");

    # Clear the error buffer
    $self->_clear_err();

    # Validate that a config file path was passed in.
    unless (defined $config and length $config){
        $self->_set_err("'config' is a required argument");
        $Log->exit("");
        return undef;
    }

    # Append '.cfg' file extension if not provided
    $config .= ".cfg" if ($config !~ /\.cfg$/);

    # Determine path of config file.  Unless config path starts
    # with '.', '~', or '/' the path is relative to the directory
    # location of the test thpl being executed.
    $config = ($config =~ /^[.~\/]/)
        ? abs_path($config)
        : $self->get_dir().$config;

    # Display location of config file
    $self->nlog("CONFIG: ".$config);

    # Return error if the file can't be opened
    unless (open(CONFIG,"<".$config))  {
        $self->_set_err("Unable to open config file '".$config."'",$!);
        return $TCD::FALSE;
    }

    # Process the config file parameter value pairs
    while(my $line=<CONFIG>){

        chomp($line);

        # Ignore lines starting with comment tag '#' or
        # containing  the '=' character.
        next if ($line=~/^#/);
        next if ($line!~/\=/);
        my ($param,$value)=$line=~/(\S+)=['"]?((\S+\s?)[^'"\s#]+)/;

        if (not exists $ENV{$param}){
            $ENV{$param}=$value;
        }

    }

    # Close the config file
    close(CONFIG);
    $Log->exit("");
}

##########################################################################4#####
# Method:    run_test
# Objective: run test methods from program
# Details:   See POD documentation at the end of this file.
###############################################################################
sub run_test
{
    my ($self) = @_;
    $Log->enter("");
    my (
        $bFound,
        $bNoSetup,
        $bNoCleanup,
        $bNoSetClean,
        $bRepeatSetup,
        $bCleanup,
        $bSetup,
        $bQuiet,
        $bNoPre,
        $bNoPost,
        $errmsg,
        $idx,
        $test,
        $testname,
        @result,
        @missing,
        );

    $self->_clear_err();

    $self->{result}=\@result;

    $bFound=0;

    # Display list of available tests and return
    # if the list (-l) option is specified
    if (scalar(grep{/^-l$/} @{$self->{options}})) {
        $Log->trace("Returning Test List");
        $self->_list_tests($self->{methodlist});
        $Log->exit("");
        return $TCD::FALSE;
    }
    $Log->trace("Not returning Test List");

    # if the help (-h) option is specified
    if (scalar(grep{/^-h/} @{$self->{options}})) {
        $Log->trace("Returning help");
        $self->_help();
        $Log->exit("");
        return $TCD::FALSE;
    }
    $Log->trace("Not Returning help");

    # Make sure if -st is explicitly called that
    # no other test case arguments accompany it.
    if (scalar(grep{/^-st$/} @{$self->{options}}) and
        (scalar (@{$self->{tests}}) or scalar (@{$self->{options}})>1)) {
        $Log->trace("Invalid command line arguments found with -st");
        $self->_set_err("="x80,
                        "Error: TCD::run_test() When specifiying the setup ",
                        "switch (-st) explicitly, no other command options or ",
                        "tests can accompany it.  Only TESTCASES=-st is allowed",
                        "="x80,
                       );
        Tharn::logresult(NATE::Result::Config->new("Invalid command line arguments",abortive=>0));
        $Log->exit("");
        return $TCD::FALSE;
    }
    $Log->trace("No invalid command line arguments found with -st");
    unless (scalar (@{$self->{tests}})) {
        $self->{NOLOG}=1;
    }

    # Determine if setup/cleanup should be run once
    # per Test Case or once per Test Case.  The
    # default is to run setup/cleanup once per
    # Test Case.

    $bRepeatSetup=scalar(grep{/^-rsc$/} @{$self->{options}});
    $bSetup=scalar(grep{/^-st$/} @{$self->{options}});
    $bCleanup=scalar(grep{/^-cl$/} @{$self->{options}});
    $bNoCleanup=scalar(grep{/^-nc$/} @{$self->{options}});
    $bNoSetup=scalar(grep{/^-ns$/} @{$self->{options}});
    $bNoSetClean=scalar(grep{/^-nsc$/} @{$self->{options}});
    $bQuiet=scalar(grep{/^-q$/} @{$self->{options}});
    $bNoPre=scalar(grep{/^-ndpre$/} @{$self->{options}});
    $bNoPost=scalar(grep{/^-ndpos$/} @{$self->{options}});

    if ($bCleanup eq $TCD::TRUE) {
        $Log->trace("Cleanup Requested");
        # If cleanup requested execute at the beginning of
        # test.  Only perform again if 1 or more tests are requested
        # to be executed.
        my @tmp=(defined @{$self->{tests}} and scalar(@{$self->{tests}}))
            ? ("init","cleanup","setup",@{$self->{tests}},"cleanup","uninit")
            : ("init","cleanup","uninit");

        $self->{tests}=\@tmp;
    }
    elsif ($bSetup eq $TCD::TRUE) {
        # If setup requested don't execute tests
        # just setup code.  This is because setup is
        # automatically executed whenever a test is
        # requested to be run,
        $Log->trace("Setup and don't execute tests requested");
        my @tmp=("init","setup","uninit");
        $self->{tests}=\@tmp;
    }
    elsif ($bNoSetup eq $TCD::TRUE) {
        $Log->trace("Bypass Setup Requested");
        # Determine if user wants to bypass setup
        my @tmp=("init",@{$self->{tests}},"cleanup","uninit");
        $self->{tests}=\@tmp;
    }
    elsif ($bNoCleanup eq $TCD::TRUE) {
        # Determine if user wants to bypass cleanup
        $Log->trace("Bypass Cleanup Requested");
        my @tmp=("init","setup",@{$self->{tests}},"uninit");
        $self->{tests}=\@tmp;
    }
    elsif ($bNoSetClean eq $TCD::TRUE) {
        # Determine if user wants to bypass setup and cleanup
        $Log->trace("Bypass Setup and Cleanup Requested");
        my @tmp=("init",@{$self->{tests}},"uninit");
        $self->{tests}=\@tmp;
    }
    elsif ($bRepeatSetup eq $TCD::TRUE){
        # Run setup and cleanup for each requested test
        # to be executed.
        $Log->trace("Setup and Cleanup will be run between each test");
        my @tmp;
        foreach my $test (@{$self->{tests}}){
            # Not running setup and cleanup for disabled test cases
            if ($test =~ /^\-/) {
                push @tmp,$test;
            } else {
                push @tmp,"setup",$test,"cleanup";
            }
        }
        @tmp=("init",@tmp,"uninit");
        $self->{tests}=\@tmp;
    }
    else{
        # Run setup once before running all selected
        # tests followed by running cleanup at the end.
        $Log->trace("Setup and cleanup will be run once");
        my @tmp=("init","setup",@{$self->{tests}},"cleanup","uninit");
        $self->{tests}=\@tmp;
    }

    $self->{repeat_setup}=$bRepeatSetup;
    $self->{no_setup}=$bNoSetup;

    # Check that at least one valid test was supplied for
    # options that require it.

    if (scalar(grep{/^-ns$|^-nc$|^-nsc$|^-q$/} @{$self->{options}})){
        $Log->trace("Found arguments that apply to tests");
        foreach $test (@{$self->{tests}}){
            ($testname)=grep{/^(\-|)$test$/i} @{$self->{methodlist}};
            next if (not defined($testname));
            $bFound=1;
        }

        if (not $bFound) {
            $Log->trace("No valid test names supplied with selected option");
            $self->_set_err("="x80,
                       "Error: TCD::run_test(): "," No valid test names supplied with selected option",
                       "="x80);
            Tharn::logresult(NATE::Result::Config->new("Invalid command line arguments",abortive=>0));
            $Log->exit("");
            return $TCD::FALSE;

        }
    } else {
        $Log->trace("No argments were found that must be applied to tests");
    }

    foreach $test (@{$self->{tests}}){
         if (not grep{/^(\-|)$test$/i} ("init","setup",
                  @{$self->{methodlist}},"cleanup","uninit")) {

            $self->_set_err("="x80,"Error: TCD::run_test(): ".
                            "No valid test named '".$test."' found!!!",
                       "="x80);

           # Log result and add it to the results list.
            Tharn::logresult(NATE::Result::Config->new("Invalid test name",abortive=>0));
            $Log->exit("");
            return $TCD::FALSE;
         }
    }

    #
    # If LISTTESTCASES has been set, all we are going to do is print
    # testcases and return.
    #
    my $just_list=$self->get_param("LISTTESTCASES");
    if ($just_list =~ /1|^y|^t/i) {
        $Log->trace("LISTTESTCASES is set");
        foreach $test (@{$self->{tests}}){
            # skip disabled and standard test cases
            next if ($test=~/^\-/);
            next if ($test=~ /^init$|^setup$|^cleanup$|^uninit$/);

            my $description = ${$self->{testcases}}{$test};
            $self->display("testcase : $test : $description");
        }
        $Log->exit("");
        return;
    }
    $Log->trace("LISTTESTCASES is not set");

    # Execute requested Test Cases
    $Log->debug("Entering testcase loop");
    $Log->debug("Testcases :".Dumper(\@{$self->{tests}}));
    while ($idx <= $#{$self->{tests}}){

        $test=${$self->{tests}}[$idx];
        $Log->debug("Processing test: $test");
        if ($idx>0) {
            $self->{previous} = ${$self->{tests}}[$idx - 1];
        } else {
            $self->{previous} = "";
        }
        if ($idx < scalar(@{$self->{tests}})) {
            $self->{next} = ${$self->{tests}}[$idx + 1];
        } else {
            $self->{next} = "";
        }
        $idx++;

        $self->{step_nbr}=0;

        $self->{method}=$test;

        # Create a hash to store each tests results
        my $hRes={};
        $self->{test_nbr}++;

        # Specified testnames are case independent (test1==TeSt1).
        # Check if this is the case.

        ($testname)=grep{/^(\-|)$test$/i} ("init","setup",
                                           @{$self->{methodlist}},"cleanup","uninit");


        if (not defined($testname)) {
            $Log->trace("Test $test is not defined - skipping");
            next;
        };

        # If testname is valid execute it and store its results in
        # a hash, otherwise inform the user it doesn't exist.
        package main;

        # Check to see if test has been disabled.
        if ($testname=~/^\-/) {
            $Log->trace("Test $testname is disabled - skipping");
            package TCD;
            $self->{description}="**Unable to capture description due to test being disabled**";
            $hRes={
                test=>substr($testname,1),
                type=>'tst',
                status=>$TCD::DISABLE,
                start_dt=>strftime("%Y-%m-%d",localtime()),
                start_tm=>strftime("%H:%M:%S",localtime()),
                time=>0
                };

            push(@result,$hRes);
        }
        # Check to see if test has been defined
        elsif (defined(&{*main::.$testname})) {
            $Log->trace("Test $testname exists - executing");

            if ( !$bNoPre &&
                 ($testname !~ /^(init|setup|cleanup|uninit)$/) ) {
                my $diag_pre_routine;
                if ( defined( &{*main::.'diag_pre_testcase'} )) {
                    $diag_pre_routine = 'diag_pre_testcase';
                }

                if ( $diag_pre_routine ) {
                    package main;
                    eval {
                        $Log->comment("----------------------------------------");
                        $Log->comment("INFO: Calling Pre-Testcase main::$diag_pre_routine");
                        $Log->comment("----------------------------------------");
                        no strict 'refs';
                        &{*main::.$diag_pre_routine}($testname);
                    };
                    if ( $@ ) {
                        $Log->warn("Pre-Routine $diag_pre_routine failed with $@");
                    }
                }
            }

            package TCD;
            $hRes=$self->_run_method($testname);
            # Store each test result hash into an array.
            push(@result,$hRes);

            if ( !$bNoPost &&
                 ($testname !~ /^(init|setup|cleanup|uninit)$/) ) {
                my $diag_post_routine;
                if ( defined(  &{*main::.'diag_post_testcase'})) {
                    $diag_post_routine = 'diag_post_testcase';
                }

                if ( $diag_post_routine ) {
                    package main;
                    eval {
                        $Log->comment("----------------------------------------");
                        $Log->comment("INFO: Calling Post-Testcase main::$diag_post_routine");
                        $Log->comment("----------------------------------------");
                        no strict 'refs';
                        &{*main::.$diag_post_routine}($testname,$hRes->{status});
                    };
                    if ( $@ ) {
                        $Log->warn("Post-Routine $diag_post_routine failed with $@");
                    }
                }
            }

            #return if ($self->error);

            # Store each test result hash into an array.
            ##push(@result,$hRes);

            # End all testing if fatal error occurs.
            if ($hRes->{status} eq $TCD::FATAL) {
                Tharn::tstlog(-comment=>"FATAL ERROR on $testname\(\)");
                $self->{msg} = "ERROR FATAL: Testing aborted due to fatal ".
                "return from $testname\(\)";

                # Set the index to the next to last element.  Check to see
                # if it's cleanup.  If it is, then leave it and execute
                # cleanup and uninit.  If it's not, set the index to the last
                # element (uninit).
                if ((($testname ne 'cleanup') and ($testname ne 'uninit'))) {
                    $idx = scalar(@{$self->{tests}}) - 2;
                    if (($self->{tests}->[$idx] ne 'cleanup') or ($testname eq 'init')) {
                        # Go to last test in list, which should be uninit
                        $idx++;
                    }
                }
                next;
            }
        }
        # Inform user that no method stub exists for
        # the test case
        else {
            package TCD;

            if (not scalar(grep{/^init$|^setup$|^cleanup$|^uninit$/} @StdMethods)){
                $self->display("Test '$test' not found\n");
                push @missing, $test;
            }
            $Log->trace("Test $test could not be found - skipping");
            next;
        }

        # If Running Single setup() or init() and either method Fails its not
        # worth executing remaining test cases.
        # CR SETUP FIX
        if (
            (
                $hRes->{test}   =~ /^init$/i and
                $hRes->{status} =~ /$TCD::FAIL|$TCD::FATAL|$TCD::SCRIPT|$TCD::CONFIG|$TCD::NA/
            )
            or
            (
                $hRes->{test}   =~ /^setup$/i and
                $hRes->{status} =~ /$TCD::FAIL|$TCD::FATAL|$TCD::SCRIPT|$TCD::CONFIG|$TCD::NA/ and
                $bRepeatSetup eq $TCD::FALSE
            )
            ) {
            $self->{msg}="ERROR: Testing aborted due to error from $testname\(\)";

            # Set the index to the next to last element.  Check to see
            # if it's cleanup.  If it is, then leave it and execute
            # cleanup and uninit.  If it's not, set the index to the last
            # element (uninit).
            if ((($testname ne 'cleanup') and ($testname ne 'uninit'))) {
                $idx = scalar(@{$self->{tests}}) - 2;
                if (($self->{tests}->[$idx] ne 'cleanup') or ($testname eq 'init')) {
                    # Go to last test in list, which should be uninit
                    $idx++;
                }
            }
            next;
        }
    }

    # Display overall summary of test results
    $self->_run_summary(\@result) if (scalar(@result)>0);

    # Delete summary files that have minimal content.
    if (-e $self->{summary_file}) {
        if (stat($self->{summary_file})->size < 10) {
            unlink ($self->{summary_file}) or
                do{
                    $self->_set_err("Error: TCD::run_test(): Unable to delete empty",
                                    "file ".$self->{summary_file}. " : $!");
                    return $self;
                }
        }
    }
=comment
    if ($self->{msg}=~/FATAL|CONFIG/i) {

        # Create a new Result object, log it, and add it to the results list.
        my $nate_result_obj =
                    Tharn::logresult(
                          type=>"CONFIG",
                          abort=>1,
                          stacktrace=>0,
                          msg=>"Testcase aborted" );

        # Log an existing Result object and add it to the results list.
        Tharn::logresult($nate_result_obj);

    }
=cut
    $Log->exit("");
}

###############################################################################
# Method:    step
# Objective: Display formatted test step
# Details:   See POD documentation at the end of this file.
###############################################################################
sub step
{
    my ($self,@step)=@_;
    $Log->enter("");
    my(
       $bFirst,
       $header,
       $len,
       $newline,
       @output,
       );

    if (not defined $step[0]) {
        $Log->exit("");
        return;
    }

    # initialize
    $bFirst = 1;
    $self->{step_nbr} += 1;

    # Add the appropriate string to the testcase header
    # If we aren't processing the firt step, show how much time the
    # previous step took
    if ($self->{step_nbr} > 1) {
        my $current_step_time = time - $self->{step_time};
        my $time_string = $self->timer_format(sprintf("%12.2f", $current_step_time),
                                              1);
        $header = "==  STEP " . $self->{step_nbr} . " (+$time_string):";
    } else {
        $header = "==  STEP " . $self->{step_nbr} . ":";
    }

    # Save the current time so we know how long the next step takes
    $self->{step_time} = time;

    # size of step number + semicolon
    $len = length($self->{step_nbr}) + 1;

    push @output,
        "\n",
        "="x80;

    foreach my $line (@step[0..scalar(@step)-1]) {

        $newline = ($bFirst==1)
            ? $header." ".$line
            : "==        "." "x($len).$line;

        push @output, $newline;
        $bFirst=0;
    }

    push @output, "="x 80;

    $Log->exit("");
    $self->nlog(@output);
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _args_list
# Objective: validate thpl files TESTCASES value
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _args_list
{
    my ($self,$tests) = @_;
    $Log->enter("");
    my (
        $arg,
        $argstr,
        $bOneSetup,
        $errmsg,
        $test,
        @args,
        @options,
        @tests,
        $fh,
        @tmp,
        $name,
        @methodlist,
        );

    $self->_clear_err();

    # initialize test counters
    $self->{test_nbr}=0;
    $self->{test_cnt}=0;
    @methodlist=@$tests;

    # Separate the tests from the
    # the command line options into
    # their own arrays, @tests and @options
    # respectively
    foreach $argstr (@{$self->{argv}}){
        @args=split(/\,/,$argstr);

        foreach $arg (@args){
            if ($arg=~/^-/){  # Command line option
                push @options,$arg;
            }
            elsif ($arg=~/\.\./){  # Index List specified
                my ($start,$end)=split(/\.\./,$arg);
                if ($start<=$end){
                    while($start<=$end){
                        push @tests,$methodlist[$start - 1]
                            if (defined($methodlist[$start - 1]));
                        $start++;
                    }
                }
                else {
                    while($start>=$end){
                        push @tests,$methodlist[$start -1]
                            if (defined($methodlist[$start - 1]));
                        $start--;
                    }
                }
            }
            elsif ($arg=~/^[0-9]{1,3}$/){ # Index specified
                push @tests,$methodlist[$arg -1]
                    if (defined($methodlist[$arg-1]));
            }
            elsif($arg=~/^\d+\*\d+$/){  # Index specified with repeat
                my ($id,$repeat)= $arg=~/(\d+)\*(\d+)$/;
                foreach (my $x=0; $x < $repeat;$x++){
                    push @tests,$methodlist[$id -1]
                        if (defined($methodlist[$id-1]));
                }
            }
            elsif ($arg=~/\%/){ # Wildcard specified
                my @tmp=split(/\%/," $arg ");
                my $arg=join('.*',@tmp);

                # Remove leading/trailing spaces
                $arg =~ s/^\s+//;
                $arg =~ s/\s+$//;

                my @arrs = grep{/^$arg$/} @methodlist;
                foreach my $arr (@arrs){
                    push @tests,$arr;
                }
            }
            elsif($arg=~/^all\*\d+$/i or $arg=~/^\%\*\d+$/){ # repeat all tests
                my (undef,$repeat)=split(/\*/,$arg);
                foreach (my $x; $x < $repeat;$x++){
                    foreach my $name (@methodlist) {
                        push @tests,$name;
                    }
                }
            }
            elsif($arg=~/^all$/i or $arg=~/^\%$/){
                push @tests,(@methodlist);  # all tests
            }
            elsif($arg=~/\S+\*\d+/){  # exact match with repeat
                my ($name,$repeat)= $arg=~/(\S+)\*(\d+)$/;
                foreach (my $x; $x < $repeat;$x++){
                    push @tests,$name;
                }
            }
            else {  # exact match
                push @tests,$arg;
            }
        }
    }

    # Check if valid test options were supplied
    if (scalar(@options) > 0){
        foreach (@options){
            unless (/^-q$|^-rsc$|^-ns$|^-nc$|^-nsc$|^-cl$|^-e$|^-st$|^-l$|^-h$|^-ndpre$|^-ndpos$/){
                $self->_set_err("="x80,"Error: Invalid option in TESTCASES list '$_'","="x80);
                Tharn::logresult(NATE::Result::Config->new("Invalid TESTCASES option '$_'",abortive=>0));
                $Log->exit("");
                return;
            }
        }
    }

    # Store reference to options array
    $self->{options}=\@options;

    # Check if at least one valid test name was supplied and
    # TESTCASES request was not list(-l), setup(-st), or cleanup(-cl) operation.
    if (scalar(@tests) eq 0 and not scalar(grep{/^-l$/} @{$self->{options}}) and
        not scalar(grep{/^-cl$|-st$|-h$/} @{$self->{options}})){
        $self->_set_err("="x80,
                        "Error: No test names were entered on the command line",
                        "       for test execution.",
                        "="x80);
        Tharn::logresult(NATE::Result::Config->new("No test names entered",abortive=>0));
        $Log->exit("");
        return ;
    }

    # Store reference to tests array
    $self->{tests}=\@tests;

    # Check that user hasn't mistakingly placed one of
    # of the standard methods in the @TESTS array
    if (scalar(grep{/^init$|^setup$|^cleanup$|^uninit$/} @{$self->{methodlist}})){
        $self->_set_err("="x80,
                        "Error: The following methods should not be included ".
                        "in the testcase\n       ",
                        "\@TESTS array; init, setup, cleanup, uninit",
                        "="x80,);
        $Log->exit("");
        return ;
    }

    # Store number of TCDs to execute
    $self->{test_cnt}= scalar(@tests);
    $Log->exit("");
    return;
}

###############################################################################
# Method:    _is_quiet
# Objective: check to see if logging should be displayed to stdout
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _is_quiet
{
    my ($self)=@_;
    $Log->enter("");
    $Log->exit("");
    return (scalar(grep{/^-q/} @{$self->{options}}));
}

###############################################################################
# Method:    _list_tests
# Objective: List available test names
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _list_tests
{
    my ($self,$tests)=@_;
    $Log->enter("");
    my (
        $cnt,
        @err,
        $rtn,
        $state,
        );

    $self->_clear_err();

    # Print test summary to the display only.
    $self->display(
                   "#"x 50,
                   "##  Available Test Cases",
                   "#"x 50);

    foreach my $tst (@$tests){
        $cnt++;
        if ($tst=~/^\-/){
            $tst=substr($tst,1);
            $state="(DISABLED)";
        }
        else {
            $state="";
        }

        $self->display( "$cnt) ".$tst." $state");
    }

    $self->display("\n");
    $Log->exit("");
}

###############################################################################
# Method:    _run_method
# Objective: Run method
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _run_method
{
    my ($self,$test)=@_;
    $Log->enter("");
    my (
        $base,
        $bHarn,
        $bNewExp,
        $logfilename,
        $tcpath,
        $output_file,
        $hRes,
        $state,
        $eval_msg,
        );

    $self->_clear_err();

    # Place a marker in NATE LOG to show where next
    # test starts
    my $description = "";
    if (defined($self->{testcases}{$test})) {
        $description = $self->{testcases}{$test};
    }
    Burt::testcase_start($test, $description) if ($test!~/^init$|^setup$|^cleanup$|^uninit$/i);

    $self->{testname}=$test;

    # Unique log file name for test case
    ($base,undef,undef)=fileparse($ENV{LOGDIR},'\..*');
    $self->{nlog_date}=substr($base,2);
    $tcpath=$self->{tcpath};
    $tcpath=~s/\//_/g;
    $logfilename=sprintf("%s+%s+%s.%s",
                     $tcpath,
                     $self->{testscript},
                     $self->{testname},
                     substr($base,2));

    # Append test case cycle id to logfilename if test
    # is being run repeatedly
    $HshTestRepeat{$test}++;
    if ($HshTestRepeat{$test} > 1) {
        $logfilename.="_".$HshTestRepeat{$test};
    }

    no strict 'refs';
    $hRes={};
    $hRes->{type}=($test=~/^init$|^setup$|^cleanup$|^uninit$/i) ? "cfg" : "tst";
    $hRes->{start_dt}=strftime("%Y-%m-%d",localtime());;
    $hRes->{start_tm}=strftime("%H:%M:%S",localtime());;


    # Start a timer
    $self->timer_start("tmA");

    # test name
    $hRes->{test}=$test;

    #
    # Context switch to the "main" context and
    # execute the test in that context
    #
    package main;
    eval {
        $hRes->{status}=&{*main::.$test}($self);
    };

    #
    # Save the possible error string to a variable in the TCD context
    #
    $eval_msg = $@;

    #
    # Clear the variable returned by eval so we get accurate
    # return value on the next call
    #
    $@=undef;

    #
    # Context Switch back to the TCD context
    #
    package TCD;

    #
    # If the testcase died, check known strings for a script error.
    # If it's not a script error than we can call it a fatal error.
    #
    if ($eval_msg) {
        my $current_status = $self->_perlmsg_lookup($eval_msg);
        my $ref_msg = ref($eval_msg);
        if (length($ref_msg)) {
            $self->nlog("TCD::_run_method():" .
                        "testcase exited with reference $ref_msg: $eval_msg");
        } else {
            $self->nlog("TCD::_run_method(): testcase exited: $eval_msg");
        }
        $hRes->{status} = $current_status;
    }

    $hRes->{status} = $TCD::UNDEF if (not defined $hRes->{status});

    $self->{log}->flush();

    # Stop timer
    $hRes->{time} = $self->timer_stop("tmA");

    my $result_obj;
    if ($hRes->{status} == $TCD::CONFIG) {
        $self->log("-- FAIL (CONFIG)");
        $result_obj=NATE::Result::Config->new("Testcase Complete: Test configuration failed", abortive=>0);
    }
    elsif ($hRes->{status} == $TCD::SCRIPT){
        $result_obj=NATE::Result::Script->new("Testcase Complete: Script Error", stacktrace=>1, abortive=>0);
        $self->log("-- FAIL (SCRIPT)");
    }
    elsif($hRes->{status} == $TCD::PASS){
        $self->log("-- PASS");
        $result_obj=NATE::Result::Pass->new("Testcase Complete: $test\(\) passed");
    }
    elsif($hRes->{status} =~ /$TCD::FAIL|$TCD::FATAL/){
        $self->log("-- FAIL");
        $result_obj=NATE::Result::Fail->new("Testcase Complete: $test\(\) failed");
    }
    elsif($hRes->{status} == $TCD::NA){
        $self->nlog("Test case does not apply for this configuration");
        $self->log("-- NA");
        $result_obj=NATE::Result::NA->new("Testcase Complete: $test\(\) is not applicable",abortive=>0);
    }
    else {
        $hRes->{status} = $TCD::SCRIPT;
        $self->nlog("Test case did not return a recognized status such as "  .
                    "\$TCD::PASS, \$TCD::FAIL, \$TCD::FATAL, \$TCD::CONFIG " .
                    "\$TCD::NA, \$TCD::SCRIPT.");
        $result_obj=NATE::Result::Script->new("Testcase Complete: Script Error",stacktrace=>1,abortive=>0);
        $self->log("-- FAIL (SCRIPT)");
    }

    # do not log 'standard method' results if TCD_STRICT_RESULTS is set
    my $not_strict_results = ! $self->get_param("TCD_STRICT_RESULTS", 0);

    # is the current test a 'standard method'?
    my $not_a_std_method = !scalar(grep $test eq $_, @StdMethods);

    # is this a setup exception? (this check was copied from line run_test();
    # without it, no NATE result will be logged for these errors
    # and then ntest will interpret that as a PASS (not what we want).
    my $setup_exception = ($hRes->{test} =~ /^init$|^setup$/i and
                           $hRes->{status} =~ /$TCD::FAIL|$TCD::FATAL|$TCD::SCRIPT|$TCD::CONFIG|$TCD::NA/);

    # evaluate the conditions to determine if we should log a result
    if ($not_strict_results || $not_a_std_method || $setup_exception) {
        Tharn::logresult( $result_obj );
    }

    use strict 'refs';
    $Log->exit("");
    return $hRes;
}

###############################################################################
# Method:    _run_summary
# Objective: Display test summary when run standalone (outside harness)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _run_summary
{
    my ($self,$result)=@_;
    $Log->enter("");
    my (
        $buffer,
        $cnt,
        $logsdir,
        $testdir,
        $line,
        $status,
        $rel_location,
        $tot_pass,
        $tot_fail,
        $tot_disable,
        $tot_time,
        $tot_undef,
        );

    # Initialize counters
    $tot_pass=0;
    $tot_fail=0;
    $tot_disable=0;
    $tot_undef=0;

    $rel_location="No Nate Log File \(LOGDIR=-\)";

    # Print test summary to the display only.
    if (defined $ENV{LOGDIR} and length($ENV{LOGDIR})){

        ($logsdir,$testdir)=$ENV{LOGDIR}=~/\/(\S+)\/(\S+)$/;

         $rel_location="/".$logsdir."/".$testdir;
    }

    $self->{stdout}=defined(fileno(\*ntest::NTEST_STDOUT))
                    ? \*ntest::NTEST_STDOUT
                    : new IO::Tee();


    $buffer = sprintf("\n\n%-80s\n%s %s\n%s %s\n%s %s\n%s %s\n%-80s\n",
                      "#"x80,
                      "##  TESTBED: ",$self->{testbed},
                      "##  TESTPATH:",$self->{tcpath},
                      "##  TESTSCRIPT:",$self->{testscript}.".thpl",
                      "##  TESTLOG: ", $rel_location,
                      "#"x80);

    $self->ndisplay($self->_summary($buffer));

    if (defined $self->{msg}) {
        $self->ndisplay($self->{msg}."\n\n");
    }
    else {
        $self->ndisplay("\n\n");
    }
    $Log->exit("");

}

sub get_summary
{
    $Log->exit("");
    my ($self,$buffer) = @_;

    $buffer = sprintf("\n\n%-80s\n%s %s\n%-80s\n",
                      "#"x80,
                      "##  TEST CASE:", $self->{testscript},
                      "#"x80);
    $Log->exit("");
    return $self->tryme($buffer);
}

sub _summary
{
    my ($self, $buffer)=@_;
    $Log->enter("");

    my $status;

    # Initialize counters
    my $cnt         = 0;
    my $tot_time    = 0;
    my $tot_tests   = 0;
    my $tot_pass    = 0;
    my $tot_fail    = 0;
    my $tot_disable = 0;
    my $tot_config  = 0;
    my $tot_script  = 0;
    my $tot_na      = 0;
    my $tot_undef   = 0;
    my $tot_notrun  = 0;

    # The tests array get preloaded with the cfg cases whether they
    # exist or not.  If they don't exist, don't count them.
    $tot_tests = scalar(@{$self->{tests}}) if defined @{$self->{tests}};
    $tot_tests-- unless(defined(&{*main::."init"}));
    $tot_tests-- unless(defined(&{*main::."setup"}));
    $tot_tests-- unless(defined(&{*main::."cleanup"}));
    $tot_tests-- unless(defined(&{*main::."uninit"}));

    $buffer = "" if (not defined $buffer);

    $buffer .=sprintf("%-4.4s %-4.4s %-8.8s %-12.12s %-52s\n".
                      "%4.4s %4.4s %-8.8s %-12.12s %-52s\n",
                      "Nbr","Type","Result","Time(s)","Test Case",
                      "="x4,"="x4,"="x8,"="x12,"="x48);

    no strict 'refs';
    foreach my $hsh (@{$self->{result}}){
        $cnt++;

        $status = $self->_get_status($hsh->{status});

        # Don't count the standard methods in the total pass/fail counts
        if ($hsh->{type} and $hsh->{type} =~ "tst") {
            $tot_pass    += 1 if($hsh->{status}==$TCD::PASS);
            $tot_fail    += 1 if($hsh->{status}=~/$TCD::FAIL|$TCD::FATAL/);
            $tot_config  += 1 if($hsh->{status}==$TCD::CONFIG);
            $tot_script  += 1 if($hsh->{status}==$TCD::SCRIPT);
            $tot_disable += 1 if($hsh->{status}==$TCD::DISABLE);
            $tot_na      += 1 if($hsh->{status}==$TCD::NA);
            #$tot_undef  += 1 if($hsh->{status}==$TCD::UNDEF);
        } else {
            $tot_tests--;
        }
        $tot_time+=$hsh->{time};

        $buffer.=sprintf("%4d %-5.5s%-8.8s %12.2f %-52s\n",
                         $cnt,
                         (defined $hsh->{type} and $hsh->{type} =~/tst/i)? "tst" : "cfg",
                         (defined $status) ? $status : "undef",
                         $hsh->{time},
                         $hsh->{test});
    }

    $tot_notrun = $tot_tests - $tot_pass - $tot_fail -
                  $tot_config - $tot_script - $tot_disable - $tot_na;


    $buffer .= sprintf("%-80s\n%s%s\n%s%s\n%s%s\n%s%s\n%s%s\n%s%s\n%s%s\n%s%s\n",
                       "="x80,
                       "               Time:     ", $self->timer_format(sprintf("%12.2f",$tot_time)),
                       "               Pass:     ", $tot_pass,
                       "               Fail:     ", $tot_fail,
                       "               Disable:  ", $tot_disable,
                       "               Config:   ", $tot_config,
                       "               Script:   ", $tot_script,
                       "               Non App:  ", $tot_na,
                       "               Not Run:  ", $tot_notrun);

    #$Log->comment($buffer));

    use strict 'refs';
    $Log->exit("");
    return $buffer;
}

###############################################################################
# Method:    _get_nuffo_description
# Objective: description to use in NUFFO
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_nuffo_description
{
    my ($self) = @_;
    $Log->enter("");
    my $description;

    $description=$self->{testcases}{$self->get_testcase()};
    $Log->exit("");
    return (defined $description) ? $description : "";
}

###############################################################################
# Method:    _get_status
# Objective: Return string version of status id
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_status
{
    my ($self, $status_id)=@_;
    $Log->enter("");
    my $status;

    $status="Passed"    if ($status_id eq $TCD::PASS);
    $status="Failed**"  if ($status_id eq $TCD::FAIL or $status_id eq $TCD::FATAL);
    $status="Config"    if ($status_id eq $TCD::CONFIG);
    $status="Disable-"  if ($status_id eq $TCD::DISABLE);
    $status="Script"    if ($status_id eq $TCD::SCRIPT);
    $status="Undef"     if ($status_id eq $TCD::UNDEF);
    $status="N/A"       if ($status_id eq $TCD::NA);

    $Log->exit("");
    return $status;
}

sub _help
{
   my ($self)=@_;
   $Log->enter("");
   my (

       @buf,
       @help,
       $switch,
       $name,
       $description,
       );

       $switch = <<end_switch;
       Command Line Switches

           Switch  Name                  Description
           ======  ====================  ==============================================
            -h     Help                  This message

            -l     List                  List available test cases to be run

            -st    Setup                 Execute setup only

            -cl    Cleanup               Execute cleanup before executing test cases

            -q     Quiet                 Execute test cases w/o displaying output to stdout

            -ns    No Setup              Execute test cases w/o performing setup

            -nc    No Cleanup            Execute test cases w/o performing cleanup

            -nsc   No Setup/Cleanup      Execute test cases w/o performing setup or cleanup

            -rsc   Repeat Setup/Cleanup  Execute setup and cleanup for each test case

            -ndpre No Diag Pre-Testcase  Execute test cases w/o running any diagnostic pre-testcase routines

            -ndpos No Diag Post-Testcase Execute test cases w/o running any diagnostic post-testcase routines

       Other Options

            o A test case can be repeatedly executed using any one of the following solutions.

                TESTCASES=mytest,mytest,mytest    # execute test case 'mytest' 3 times
                TESTCASES=mytest*100              # execute test case 'mytest' 100 times
                TESTCASES=2*50                    # execute test case at index position 2 50 times.

end_switch

   $Log->exit("");
   $self->log($switch);
}

###############################################################################
# Method:    _perlmsg_lookup
# Objective: Determine whether a given argument matches a Perl compilation
#            or runtime error message, and return a corresponding Result type.
# Details:   This is called when a testcase dies and gets caught by an
#            eval and we are trying to figure out if it's a SCRIPT error
#            or a CONFIG error based on the string returned.  If the
#            item that was returned is not a string we don't handle
#            that here.
#
#            This method was copied from Subtest.pm.
###############################################################################
sub _perlmsg_lookup
{
    my ($self, $msg) = @_;
    $Log->enter("");

    my @Script_msgs = (
        qr/compilation (?:errors|aborted)/o,
        qr/syntax error/o,
        qr/^Global symbol/o,
        qr/found where operator expected/o,
        qr/Missing (?:comma|command|operator|name|semicolon|.* bracket)/o,
        qr/bareword /oi,
        qr/argument (?:is not|.+ isn't) /o,
        qr/undefined (?:value|subroutine|variable|format)/oi,
        qr/^Can't (?:call method|locate|use|modify|bless|return|coerce)\b/o,
        qr/^Odd number of/o,
        qr/^Prototype mismatch/o,
        qr/^Assertion failed/o,
        qr/^Bad name after/o,
    );
    my @Config_msgs = (
        qr/^Can\'t exec/o,
    );

    $Log->exit("");
    my $msg_ref = ref($msg);

    # Handle Exceptions
    # Get the type of result from the method result_type and convert it to
    # a TCD result
    if (blessed($msg_ref) && ($msg_ref->isa("NATE::BaseException"))) {
        if (UNIVERSAL::can($msg_ref,"result_type")) {
            my $result_type = $msg_ref->result_type();
            $Log->comment("Got an exception object. Result Type is $result_type");
            if ($result_type eq 'FAIL') {
                return $TCD::FAIL;
            } elsif ($result_type eq 'FATAL') {
                return $TCD::FATAL;
            } elsif ($result_type eq 'SCRIPT') {
                return $TCD::SCRIPT;
            } elsif ($result_type eq 'CONFIG') {
                return $TCD::CONFIG;
            } else {
                # could not understand the result
                $Log->debug("TCD cannot understand result '$result_type'");
                return $TCD::FAIL;
            }
        } else {
            # object does not have result_type method
            $Log->debug("Object does not have method 'result_type'");
            return $TCD::FAIL;
        }
    }
    #
    # If we got a NATE Fatal, back return fatal, this will be the last
    # testcase executed.  If we got any other non-string back, just
    # return FAIL.  If we got a string, check for script and config
    # errors against the known defined messages above.  Otherwise
    # just return FAIL.
    #
    return $TCD::FATAL  if ($msg_ref eq "NATE::Result::Fatal");
    return $TCD::FAIL   if ($msg_ref);
    return $TCD::SCRIPT if grep($msg =~ $_, @Script_msgs);
    return $TCD::CONFIG if grep($msg =~ $_, @Config_msgs);
    return $TCD::FAIL;
}

############################################################################
# Subroutine: _map_nate_result_to_TCD_status
# Parameters: NATE severity code of the failure, abortive flag
# Returns: TCD status code
# Purpose: Convert from a NATE severity code to a TCD status code
#
# Note: Not all NATE results have a direct TCD mapping.  Most of the missing
#       severities are reasonably clear what they should be mapped to, but Inspect
#       is a little questionable.  The NATE documentation says that the test was not
#       able to determine a result and the log files must be examined manually to
#       determine pass/fail.  It would seem safest to count this as a failure.
#
#    NATE::Result::Script::severity  = 6 ==> $TCD::SCRIPT = -8
#    NATE::Result::Config::severity  = 5 ==> $TCD::CONFIG = -7
#    NATE::Result::Fatal::severity   = 4 ==> $TCD::FATAL  = -5
#    NATE::Result::Fail::severity    = 4 ==> $TCD::FAIL   =  0
#    NATE::Result::Inspect::severity = 3 ==> $TCD::FAIL   =  0 (no exact TCD code match)
#    NATE::Result::NA::severity      = 2 ==> $TCD::NA     = -6
#    NATE::Result::Warn::severity    = 1 ==> $TCD::PASS   =  1 (no exact TCD code match)
#    NATE::Result::Info::severity    = 0 ==> $TCD::PASS   =  1 (no exact TCD code match)
#    NATE::Result::Param::severity   = 0 ==> $TCD::PASS   =  1 (no exact TCD code match)
#    NATE::Result::Pass::severity    = 0 ==> $TCD::PASS   =  1
#############################################################################

sub _map_nate_result_to_TCD_status {
    my $self     = shift;
    my $severity = shift;
    my $abortive = shift;

    # Currently Fatal and Fail are the same severity, they only differ
    # in abortiveness.  In case this ever changes, I am coding this
    # as if they are not the same severity and interpretting an
    # abortive Fail as Fatal, and a non-abortive Fatal as Fail.
    if ( $severity      == NATE::Result::Script::severity() ) {
        return $TCD::SCRIPT;
    } elsif ( $severity == NATE::Result::Config::severity() ) {
        return $TCD::CONFIG;
    } elsif ( $severity == NATE::Result::Fatal::severity() ) {
        if ( $abortive ) {
            return $TCD::FATAL;
        } else {
            return $TCD::FAIL;
        }
    } elsif ( $severity == NATE::Result::Fail::severity() ) {
        if ( $abortive ) {
            return $TCD::FATAL;
        } else {
            return $TCD::FAIL;
        }
    } elsif ( $severity == NATE::Result::Inspect::severity() ) {
        return $TCD::FAIL;
    } elsif ( $severity == NATE::Result::NA::severity() ) {
        return $TCD::NA;
    } elsif ( ($severity == NATE::Result::Info::severity()) ||
              ($severity == NATE::Result::Param::severity()) ||
              ($severity == NATE::Result::Pass::severity()) ) {
        return $TCD::PASS;
    } else {
        $Log->warn("Do not understand severity $severity");
        return $TCD::FAIL;
    }
}

############################################################################
# Subroutine: worst_subtest_result
# Parameters: array of subtest object references
# Returns: TCD Execution Result
# Purpose: Find the most severe subtest NATE result and map it to a TCD result
#############################################################################
sub worst_subtest_result {
    my $self = shift;
    my @st = @_;
    my @res;
    my $worst_of_all_subtests = NATE::Result::Pass::severity();
    my $abortive = 0;

    for my $st (@st) {
        @res = $st->results();
        if (@res) {
            my $worst = NATE::Result::worst(@res);

            # keep track of the worst of the worst
            if ( $worst->severity() > $worst_of_all_subtests ) {
                $worst_of_all_subtests = $worst->severity();
                $abortive = $worst->abortive();
            } elsif ( $worst->severity() == $worst_of_all_subtests ) {
                if ( $worst->abortive() > $abortive ) {
                    $abortive = $worst->abortive();
                }
            }

        } else {
            $Log->debug("No result  for subtest ".$st->runid()."- skipping");
        }
    }
    $Log->debug( "Worst result of all subtests is ".$worst_of_all_subtests." abortiveness = ".$abortive );

    my $tcd_rc =  $self->_map_nate_result_to_TCD_status($worst_of_all_subtests,$abortive);

    $Log->debug("Mapped to TCD equate = $tcd_rc");
    return $tcd_rc;
}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

TCD.pm

=head1 Author

Gary Knopp

=head1 Description

TCD (Multi Test Driver) extends the ntest command line interface by providing
the ability to execute specific testcase methods in a single testscript.

A Command line variable 'TESTCASES' is used to provide list of test names
(methods) to execute in the test script.  If not specified, all tests in
the testscript are executed once.

If TCD_STRICT_RESULTS is set to a non-zero value, then TCD will NOT log NATE
results for the init/setup/cleanup/uninit methods - if they PASS. Any FAIL
result which terminates the run will still generate a NATE result, regardless
of which method failed.

Also included are additional method that utilize IO:Tee module that enables
output to be mutliplexed to nate logs and STDOUT in a single call.

=head1 Synopsis

Here are some examples for executing a test script using various TESTCASES value combinations.

=over 4

    ntest -noconsole Syntax/storage_show.thpl TESTCASES=[]

=item * TESTCASES=mytest1

Enter a single test name.

=item * TESTCASES=mytest1,mytest2

Enter comma seperated list of test names. Space seperated values are
allowed also if the entire TESTCASES string value is quoted.

=item * TESTCASES=1,3..5,2..1,mytest1

Enter the index position or index range (base 1).

=item * TESTCASES=mytest%

Enter test names using the wildcard character '%'.

=item * TESTCASES=all

Enter keyword 'all' (or %) to execute all tests.

=item * TESTCASES=3*100

    Execute test case at position 3 in the @TESTS array 100 times.


       Command Line Switches to TESTCASES parameter

           Switch   Name                  Description
           ======   ====================  ==============================================
            -h      Help                  This message

            -l      List                  List available test cases to be run

            -st     Setup                 Execute setup only

            -cl     Cleanup               Execute cleanup before executing test cases

            -q      Quiet                 Execute test cases w/o displaying output to stdout

            -ns     No Setup              Execute test cases w/o performing setup

            -nc     No Cleanup            Execute test cases w/o performing cleanup

            -nsc    No Setup/Cleanup      Execute test cases w/o performing setup or cleanup

            -rsc    Repeat Setup/Cleanup  Execute setup and cleanup for each test case

            -ndpre  No Diag Pre-Testcase  Execute test cases w/o running any diagnostic pre-testcase routines

            -ndpos  No Diag Post-Testcase Execute test cases w/o running any diagnostic post-testcase routines

       Other Options

            o A test case can be repeatedly executed using any one of the following solutions.

                TESTCASES=mytest,mytest,mytest    # execute test case 'mytest' 3 times
                TESTCASES=mytest*100              # execute test case 'mytest' 100 times
                TESTCASES=2*50                    # execute test case at index position 2 50 times.

=back

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

Error.pm

TCD::Timer.pm


=head1 Public Methods

=head2 B<new(%args)>

=over 4

=item Description

Creates a new Mult Test Driver (TCD) object.

=item Inputs

=item * -testcases

Array reference of test method names that can be executed

=item * I<-options>

Comma separated list of TCD options.  Examples: -q -rsc -ns -nc -nsc -cl -e -st -l -h

=item Outputs

Returns a TCD object.

=item Examples

 # Test case method names to control
 my @TESTS=qw( test1 test2 test3);

 # Create a Test Case Object
 $Test = new TCD(-testcases=>\@TESTS);
 if ($Test->error){
     $Test->nlog($Test->errmsg);
     return $TCD::FAIL;
 }

=back

=head2 B<description(@descr)>

This method creates a standard format header for the testcase method being executed.
The header contains the name of the testcase script, test method, along with the
user defined description. Alternatively, calling this method with no arguments
will use the current testcase's description from the @Testcases array in the
header. This method should only be called once at the beginning of each public
method and standard methods if used (init(), setup(), cleanup(),
and uninit()).

=over 4

=item Inputs

=item * $descr|@arr|undef

Detailed description of the test

=item Outputs

None

=item Examples

 sub mytest1()
 {

    # print testcase header
    $Test->description(
                    "This test verifies that a single",
                    "volume group containing 1 volume",
                    "can be snapped"
                );
    ...
    ...
    ...
    return $TCD::PASS;
 }

Alternatively:

sub mytest1()
 {

    # print testcase header
    $Test->description();   # Uses description from @Testcases for this testcase.
    ...
    ...
    ...
    return $TCD::PASS;
 }

Note: Support has been added for diagnostic pre and post routines.  These routines (if
      defined in your script) will be run before and after your testcase is run, but will
      not contribute to the pass/fail status of that testcase.  These routines provide a
      convenient place to run filer commands that may aide in triaging failures.
      If you do not define routines that start with diag_pre_/diag_post_, this feature
      will not be activated.

      If you define a "diag_pre_testcase()" routine, that routine will be executed prior to
      running each testcase in your script.  Likewise, if you define a "diag_post_testcase()"
      routine, it will be executed after each testcase is run. The diag_pre_testcase() routine
      will be passed the name of the testcase that will be called after it.  The diag_post_testcase()
      routine will be passed that name as well as the status of the testcase run before it.

      A diagnostic pre-routine might be needed if you need to clear out filer tables prior to
      running a testcase.  In the diagnostic post-routine, you may want to only run diagnostic
      commands in the event of a testcase failure, but the routine will be called regardless,
      so you are not limited in that regard.

      Ex:

      sub diag_post_testcase {
          my $tc_name = shift;
          my $tc_result = shift;

          if ( $tc_result != $TCD::PASS ) {
              foreach my $server ($Server->servers()) {

                  $Server->set_execution_server(-Server =>$server);

                  eval {$Server->network_interface_show();};
                  eval {$Server->network_port_show();};
              }
          }
      }

=back

=head2 B<display(@msg)>

This method displays information to STDOUT.  This information is not
captured in any log files.  A carriage return is automatically appended
to the input buffer.

=over 4

=item Inputs

=item * @msg

An array or scalar of information to be logged to a file for the test being run.

=item Outputs

None

=item Examples

 # Determine if error occurred...
 $Test->display("This text will only be written to STDOUT");


=back

=head2 B<get_dir()>

Directory location of testcase being executed

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the directory name of the testcase.  If an error
occurs the error flag is set, errmsg is set to the failure reason, and "" is returned.


=item Examples

    $Test->get_dir("Testcase is located in directory'".$Test->get_dir()."'");

=back

=head2 B<get_loghdl()>

This method retrieves the log handle that is used by the the TCD object
to multiplex log results.  This handle can be passed to other modules that
are used by a test for logging additional information.

=over 4

=item Inputs

none

=item output

Returns the file handle from TCD object.

=item Examples

    $loghdl=$Test->get_loghdl();

=back

=head2 B<get_name()>

Name of testcase

=over 4

=item Inputs

none

=item Outputs

Returns the name of the test case.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.


=item Examples

    $Test->nlog("Test case name is '".$Test->get_name()."'");

=back

=head2 B<get_param($param,I<$default>)>

This method retrieves any environment variables that have been set.

=over 4

=item Inputs

=item * $param

The name of the parameter to retrieve

=item * I<$default>

value to use if parameter is not set.

=item output

Returns a value if successful.  If an error occurred, the error flag is set,
errmsg is set to the failure reason.

=item Examples

    $volume=$Test->get_param("VOLUME","vol1");
    if ($Test->error) {
        $Test->log($Test->errmsg);
        return $TCD::FAIL;
    }


=back

=head2 B<get_testcase()>

Name of test case that is being executed

=over 4

=item Inputs

none

=item Outputs

Returns the name of the test case.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.


=item Examples

    $Test->nlog("Test case name is '".$Test->get_testcase()."'");

=back


=head2 B<get_next_testcase()>

Name of the next test case that will be executed

=over 4

=item Inputs

none

=item Outputs

Returns the name of the next test case.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.


=item Examples

    $Test->nlog("Next test case name is '".$Test->get_next_testcase()."'");

=back

=head2 B<get_previous_testcase()>

Name of the previous test case that was executed

=over 4

=item Inputs

none

=item Outputs

Returns the name of the previous test case.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.

=item Examples

    $Test->nlog("Previous Test case name is '".$Test->get_previous_testcase()."'");

=back

=head2 B<get_previous_result()>

Result of the previous test case that was executed

=over 4

=item Inputs

none

=item Outputs

Returns the result of the previous test case.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.

=item Examples

    $result = $Test->get_previous_result();

=back

=head2 B<get_worst_result()>

Return the worst result value of all testcases that ran

=over 4

=item Inputs

none

=item Outputs

A TCD Result value ($TCD::SCRIPT, $TCD::CONFIG, etc...)

=item Examples

    $worst_result = $Test->get_worst_result();
    if ($worst_result <= $TCD::FAIL) {
       # at least one test fataled
    } elsif ($worst_result == $TCD::PASS) {
       # all tests passed
    }

=back

=head2 B<worst_subtest_result(@subtests)>

This method return the worst result value of all the subtests that are
passed in, and converts that NATE result into a TCD result.

=over 4

=item Inputs

=item * @subtests

An array of Subtest objects.

=item Outputs

The subtests passed in will be traversed and the worst NATE result
of all subtests will be mapped to a TCD Result value ($TCD::SCRIPT,
$TCD::CONFIG, etc...) based on the following mapping:

    NATE::Result::Script::severity  = 6 ==> $TCD::SCRIPT = -8
    NATE::Result::Config::severity  = 5 ==> $TCD::CONFIG = -7
    NATE::Result::Fatal::severity   = 4 ==> $TCD::FATAL  = -5
    NATE::Result::Fail::severity    = 4 ==> $TCD::FAIL   =  0
    NATE::Result::Inspect::severity = 3 ==> $TCD::FAIL   =  0 (no exact TCD code exists)
    NATE::Result::NA::severity      = 2 ==> $TCD::NA     = -6
    NATE::Result::Warn::severity    = 1 ==> $TCD::PASS   =  1 (no exact TCD code exists)
    NATE::Result::Info::severity    = 0 ==> $TCD::PASS   =  1 (no exact TCD code exists)
    NATE::Result::Param::severity   = 0 ==> $TCD::PASS   =  1 (no exact TCD code exists)
    NATE::Result::Pass::severity    = 0 ==> $TCD::PASS   =  1

=item Examples

    $worst_result = $Test->worst_subtest_result(@subtests);
    if ($worst_result <= $TCD::FAIL) {
       # at least one subtest failed
    } elsif ($worst_result == $TCD::PASS) {
       # all subtests passed
    }

=back

=head2 B<log(@msg)>

This method logs information to both STDOUT and a test logfile opened by the
TCD object.

=over 4

=item Inputs

=item * @msg

An array or scalar of information to be logged to a file for the test being run.

=item Outputs

None

=item Examples

 # Determine if error occurred...
 $TCD->log(
            "This text will be written to a log file as well",
            "as STDOUT"
            );

=back

=head2 B<ndisplay(@output)>

This method will log information to both STDOUT and the nate log.

=over 4

=item Inputs

=item * @output

An array or scalar of information to be logged to a file for the test being run.

=item Outputs

None

=item Examples

# Determine if error occurred...
$TCD->ndisplay("This text will only be written to STDOUT and NATE log file");

=back

=head2 B<nlog(@buf)>

This method will log information to both nate log, testcase log, and STDOUT.

=over 4

=item Inputs

=item * @buf

An array or scalar of information to be logged for the test being run.

=item Outputs

None

=item Examples

 # Determine if error occurred...
 $TCD->nlog(
            "This text will be written to a NATE, testcase log file, as well",
            "as STDOUT"
            );

=back

=head2 B<read_config(@buf)>

This method will read a config file of parameter value pairs and
store them into the environment

=over 4

=item Inputs

=item * $config

A scalar value to a config file (*.cfg) containing param value pairs
to be stored in the environment for the test to use.
If only a file name is provied the default location of the config
file is relative to the location of the test thpl being executed.

This function is automaticaly called when a CONFIG is passed in on
the command line.

=item Outputs

None

=item Examples

 ...  if the test was being run from /x/eng/testbeds/mylab/mytestbed

 # location of config file would be  /u/knopp/p4/test/TCD/mytest.cfg
 ntest CONFIG=mytest.cfg TCD/template.thpl

 # location of config file would be
 # /x/eng/testbeds/mylab/mytestbed/mytest.cfg
 ntest CONFIG=./mytest.cfg TCD/template.thpl

 # location of config file would be /u/knopp/mytest.cfg
 ntest CONFIG=~/mytest.cfg TCD/template.thpl

 The following param value formats in the config file are acceptable.
 All strings after a comment character '#' are ignored.

 myparam=myvalue
 myparam=my multiword value
 myparam='my multiword value'
 myparam="my multiword value"
 myparam=my multiword value   ### my parameter comment


=back

=head2 B<run_tests()>

This method executes the test names in the @TESTS array as identified on the command line.
The following standard methods will automatically be called if defined in the testscript;
init() and setup() before running the tests followed by cleanup() and uninit().

Standard methods init() and uninit() should be used for object creation and deletion.  While
setup() and cleanup() should be used for configuration and teardown.

Each testmethod callback should return a 1 ($TCD::PASS) or 0 ($TCD::FAIL) to indicate
the result of the test.  If either of the configuration methods init() or setup() fail
then the whole testcase is considered a failure and none of the test cases are executed. If any
test method returns a fatal error ($TCD::FATAL) test execution is aborted and no cleanup()
is performed.

=over 4

=item Inputs

none

=item Outputs

None

=item Examples

  # Tests to run
  # Use '-' character to disable test from running
  my @TESTS=qw(mytest1 mytest2 -mytest3);

  # TCD
  $Test = new TCD(-testcases=>[@TESTS]);
  if ($Test->error){
      $Test->nlog($Test->errmsg);
      return $TCD::FAIL;
  }

  $Test->run_test();
  if ($Test->error){
      $Test->error($Test->errmsg);
      return $TCD::FAIL;
  }

  sub mytest1{
      $Test->description("test1 descr");
      $Test->nlog("test1 passes");
      return $TCD::PASS;

  }

  sub mytest2{
      $Test->description("test2 descr");
      $Test->nlog("test2 fails");
      return $TCD::FAIL;
  }

  sub mytest3{
      $Test->description("test3 descr");
      $Test->nlog("test3 disabled");
      return $TCD::DISABLE;
  }


=back

=head2 B<step(@descr)>

This method creates a standard format header for each testcase step in a testcase method.
Sequential number of steps is maintained by TCD.

=over 4

=item Inputs

=item * $descr|@arr

Description of the test step

=item Outputs

None

=item Examples

    $Test->step("Mount the NFS created filesystems");

=back

=head1 TODO

=over 4

=item * Add support for automatic result entry in Nuffo.

=back

=cut

