# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary Setup Test Library
## @author  Adam.Ciapponi@netapp.com
## @pod     here

=head1 NAME

CR::Setup::Test

=head1 DESCRIPTION

This module contains the base for all setup tests.

=head1 METHODS

=cut

package CR::Setup::Test;

# Compiler directives.
use strict;
use warnings;

# Module import
use CR::Setup::CifsInfo;
use CR::Setup::Cleanup;
use CR::Setup::Constants;
use CR::Setup::Defaults;
use CR::Setup::Helpers;
use CR::Setup::TestParams;
use CR::Setup::Validation;
use Data::Dumper::Concise;
use NACL::APISet;
use NACL::C::Client;
use NATE::BaseException qw(:try);
use Options;
use Tharn qw($Log);
use TestUtils qw(get_site_info);

# This should be in order up top but Validation uses UnitTest which includes TCD
# undoing the path
use CR::Setup::TCD;

use base qw(Exporter);

use vars qw(
    $Cifs_Info
    $Cleanup
    $Defaults
    $Driver_Options
    $Helpers
    @Linux_Clients
    $Site_Info
    @Solaris_Clients
    $Validation
    $Version
    @Windows_Clients
);

our @EXPORT = qw(
    configure_testcases
    $Cifs_Info
    $Cleanup
    $Defaults
    $Helpers
    @Linux_Clients
    $Site_Info
    @Solaris_Clients
    $Validation
    $Version
    @Windows_Clients
);

# UNIVERSAL::isa has a bug in its verbose warnings and will
# flag most NACL components as incorrectly using isa even when they
# are using it correctly. Perlcritic considers NACL's usage as correct.
# Unfortunately there is no way to turn off verbose warnings once it is
# enabled, thus requiring turning the report_warning function to a NOP.
use UNIVERSAL::isa;
package UNIVERSAL::isa;
no warnings 'redefine';
sub report_warning {
    return;
}
package CR::Setup::Test;

# Always include some libraries
package main;
# Basic constants ie $TRUE/$FALSE
use CR::Setup::Constants;
# Basic matchers and should statements
use CR::Setup::Expectation;
# Composite mock
use CR::Setup::Mocks::CompositeMock;
# Make Dumper available by default
use Data::Dumper::Concise;
# Make array/list operations available by default
use List::Util qw(first max maxstr min minstr reduce shuffle sum);
use List::MoreUtils qw(:all);
# Rounding functions like round to nearest
use Math::Round qw(:all);
# Try/catch blocks available by default
use NATE::BaseException qw(:try);
# Mocking libraries
use Test::MockModule;
use Test::MockObject;
use Test::MockObject::Extends;
package CR::Setup::Test;

########## Exported Functions  ###################################################

sub configure_testcases {
    my @testcases = @_;

    # Determine configuration for testcases
    my $configuration = {};
    my $first_testcase = shift(@testcases);
    if ( $first_testcase eq 'configuration') {
        $configuration = shift(@testcases);
    } else {
        # First testcase wasn't configuration, put it back onto the array
        unshift(@testcases, $first_testcase);
    }

    # Get test driver options
    $Driver_Options = new Options;
    $Driver_Options->process();

    # Get versioning information
    my $force_version = $Driver_Options->get( option => 'FORCE_VERSION' );
    $Version = 'latest';
    if ( $force_version ne 'auto' ) {
        $Version = $force_version;
        $Site_Info = _get_site_info();
    } elsif ( !$configuration->{'disable_version_check'} && $force_version eq 'auto' ) {
        # The behavior of this section has not been tested on mixed mode clusters.
        # Since it only grabs the version of the first node in the NODES list the
        # tests would have to have that nodes build version specified in order to
        # be executed.
        my @nodes = @{ $Driver_Options->get( option => 'NODES' ) };
        my $hostrec = Hostrec->new( id => $nodes[0] );
        $Version = _get_version_info($hostrec);
        $Site_Info = _get_site_info($hostrec);
    }

    _handle_clients();
    _handle_constants($Version);

    $Cleanup    = CR::Setup::Cleanup->new();
    $Defaults   = CR::Setup::Defaults->new($Version);
    $Helpers    = CR::Setup::Helpers->new( version => $Version, defaults => $Defaults );
    $Validation = CR::Setup::Validation->new( version => $Version, defaults => $Defaults, helpers => $Helpers );
    $Cifs_Info = CR::Setup::CifsInfo->new( helpers => $Helpers );

    $configuration->{'testsets'} = $Driver_Options->get( option => 'TESTSETS' );

    # Process testcases
    my @processed_testcases = ();
    my @not_implemented_testcases = ();
    while (@testcases) {
        my $test_name = shift(@testcases);
        my $test_info = shift(@testcases);

        if ( ref($test_info) eq 'HASH' ) {
            if ( ! main->can($test_name) ) {
                # Don't attempt to run not implemented or disabled testcases
                push(@not_implemented_testcases, $test_name, $test_info->{'desc'});
            } else {
                # Only run if configuration is correct
                my $testcase_should_run = _testcase_matches_config(
                    'test_config'    => $test_info,
                    'testset_config' => $configuration,
                    'version'        => $Version
                );
                if ($testcase_should_run) {
                    push (@processed_testcases, $test_name, $test_info->{'desc'});
                }
            }
        } else {
            if ( ! main->can($test_name) ) {
                # Don't attempt to run not implemented or disabled testcases
                push(@not_implemented_testcases, $test_name, $test_info);
            } else {
                # Run test case
                push (@processed_testcases, $test_name, $test_info);
            }
        }
    }

    while (@not_implemented_testcases) {
        my $test_name = shift(@not_implemented_testcases);
        my $test_desc = shift(@not_implemented_testcases);
        if ( $test_name =~ /^-/ ) {
            $Log->warn('Testcase: ' . $test_name . '() => "' . $test_desc . '" is disabled.');
        } else {
            $Log->warn('Testcase: ' . $test_name . '() => "' . $test_desc . '" is not yet implemented.');
        }
    }

    if ( scalar(@processed_testcases) > 0 ) {
        return _execute_testcases(@processed_testcases);
    } else {
        $Log->warn('No testcases are applicable to run.');
        return _exit_execution();
    }
}

########## Local Functions  ###################################################

sub _execute_testcases {
    my @testcases = @_;
    $Log->enter('');

    $Log->log("Applicable testcases:\n" . Dumper(\@testcases));

    package main;
    use vars qw($Test);

    $Test = new TCD(
        -options   => '-rsc',
        -testcases => [@testcases]
    );
    if ($Test->error){
        $Test->log($Test->errmsg);
        $Log->exit('');
        return $TCD::FAIL;
    }

    # Performs method callback
    $Test->run_test();
    if ($Test->error){
        $Test->log($Test->errmsg);
        $Log->exit('');
        return $TCD::FAIL;
    }
    $Log->exit('');
    exit(0);
}

sub _get_version_info {
    my ($hostrec) = @_;
    # The behavior of this section has not been tested on mixed mode clusters.
    # Since it only grabs the version of the first node in the NODES list the
    # tests would have to have that nodes build version specified in order to
    # be executed.

    $Log->log("Determining version info");

    my $version_mgr = NACL::APISet::Version->new(
        hostobj    => $hostrec,
        category   => "Node",
        shell_type => "CMode",
    );

    my $branch = $version_mgr->get_version_attribute( attribute => "branch" );
    my $version = $version_mgr->get_version_attribute( attribute => "release" );
    my $build
        = $version_mgr->get_version_attribute( attribute => "node_version" );

    $Log->log(
        "Version: $version Build: $build Branch: $branch");

    $version =~ s/(.*)_//g;

    return $version;
}

sub _get_site_info {
    my ($hostrec) = @_;

    # Specified SITE is highest priority
    # Next highest priority is SITE from hostrec
    # If not in either default to svl
    my %sites = get_site_info;
    my $site;
    if ( $Driver_Options->defined( option => 'SITE' ) ) {
        $site = $Driver_Options->get( option => 'SITE' );
    }
    elsif (defined $hostrec) {
        $site = $hostrec->SITE() ? $hostrec->SITE() : 'svl';
    }
    else {
        $site = 'svl';
    }
    $Log->log('Using site: ' . $site);
    my $site_info = $sites{$site};

    return $site_info;
}

sub _testcase_matches_config {
    my %params = @_;

    my $test_config = $params{'test_config'};
    my $testset_config = $params{'testset_config'};
    my $version = $params{'version'};

    if ( ! defined $test_config->{'versions'}) {
        # Currently configuration is solely based off of versioning
        return $TRUE;
    } else {
        my $versioning = $test_config->{'versions'};
        if (defined $versioning->{'all'}) {
            # Just check if testset matches
            # Version matches this criteria
            my @possible_sets = @{$testset_config->{'testsets'}};
            my @testsets = @{$versioning->{'all'}};
            if (grep(/^all$/, @possible_sets)) {
                return $TRUE;
            } else {
                # Match specific
                foreach my $set (@possible_sets) {
                    if (grep(/^$set$/, @testsets)) {
                        return $TRUE;
                    }
                }
            }
        } else {
            my @node_version = split(/\./, $version);
            my $nv = sprintf("%02d",$node_version[0]) . sprintf("%02d", $node_version[1]);
            foreach my $test_version (keys %{$versioning}) {
                if ($test_version =~ /(.*?)([0-9\.]+)/) {
                    my @tv_dot = split(/\./, $2);
                    my $tv = sprintf("%02d",$tv_dot[0]) . sprintf("%02d", $tv_dot[1]);
                    my $operator = $1;
                    unless ($operator) {
                        $operator = '==';
                    }
                    if (eval ('$nv' . $operator . '$tv' )) {
                        # Version matches this criteria
                        my @possible_sets = @{$testset_config->{'testsets'}};
                        my @testsets = @{$versioning->{$test_version}};
                        if (grep(/^all$/, @possible_sets)) {
                            return $TRUE;
                        } else {
                            # Match specific
                            foreach my $set (@possible_sets) {
                                if (grep(/^$set$/, @testsets)) {
                                    return $TRUE;
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    return $FALSE;
}

sub _handle_clients {
    my @client_names = ();
    if ($Driver_Options->defined( option => 'CLIENT' )) {
        @client_names = @{ $Driver_Options->get( option => 'CLIENT' ) };
    }

    my @windows_client_names = ();
    if ($Driver_Options->defined( option => 'WIN_CLIENTS' )) {
        @windows_client_names = @{ $Driver_Options->get( option => 'WIN_CLIENTS' ) };
    }

    # Ensure client names does not include names in WIN_CLIENTS
    if ( @windows_client_names ) {
        @client_names = grep {
            my $client = $_;
            grep {
                $_ ne $client
            } @windows_client_names
        } @client_names;
    }

    @Linux_Clients = ();
    foreach my $client_name (@client_names) {
        try {
            my $client = NACL::C::Client->new( name => $client_name );
            push(@Linux_Clients, $client);
        } otherwise {
            $Log->log('Unable to initialize client: ' . $client_name);
        };
    }

    # Separate out solaris clients
    @Solaris_Clients = grep { $_->type() =~ /^unix-solaris$/ } @Linux_Clients;
    @Linux_Clients = grep { $_->type() !~ /^unix-solaris$/ } @Linux_Clients;

    # Separate out windows clients
    @Windows_Clients = grep { $_->is_windows() } @Linux_Clients;
    @Linux_Clients = grep { ! $_->is_windows() } @Linux_Clients;

    # Whatever is put into WIN_CLIENTS is assumed to be windows
    foreach my $client_name (@windows_client_names) {
        try {
            my $client = NACL::C::Client->new( name => $client_name );
            push(@Windows_Clients, $client);
        } otherwise {
            $Log->log('Unable to initialize windows client: ' . $client_name);
        };
    }

    return;
}

sub _handle_constants {
    my ($version) = @_;

    # Use version specific constants

    if ($version eq 'latest') {
        package main;
        use CR::Setup::Constants::SierraNevada;
        package CR::Setup::Test;
    } elsif ($version =~ m/^8\.3\./) {
        package main;
        use CR::Setup::Constants::FullSteam;
        package CR::Setup::Test;
    } elsif ($version =~ m/^8\.2\./) {
        package main;
        use CR::Setup::Constants::SierraNevada;
        package CR::Setup::Test;
    } else {
        # Invalid version
    }
}

sub _exit_execution {
    package main;
    exit(0);
}
1;
