# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary Balanced creation / expansion Library - 8.3 FullSteam Release
## @pod     here

=head1 NAME

CR::Setup::Validation::Balanced_FS

=head1 DESCRIPTION

This module implements balanced creation / expansion library for the FullSteam release.

=head1 METHODS

=cut

package CR::Setup::Validation::Balanced::FullSteam;

use strict;
use warnings;

use base qw(CR::Setup::Validation::Balanced::SierraNevada);

# Module import

use CR::Setup::Constants;
use CR::Setup::Constants::SierraNevada;
use List::Util qw(first max maxstr min minstr reduce shuffle sum);
use List::MoreUtils qw(:all);
use Math::Round qw(:all);
use NACL::C::Node;
use Tharn qw($Log);
use Test::Deep::NoTest;
use Data::Dumper::Concise;
use Params::Validate qw(:all);

=head2 validate_infinitevol

This function validates the correctness of Infinitevol expansion and creation operations

=cut

sub validate_infinitevol {
    my $self                = shift;
    my %parameters          = @_;
    my @validate_these_only = ();
    @validate_these_only = @{ $parameters{'validate_these_only'} }
        if defined $parameters{'validate_these_only'};
    my %validations = (
        create => [
            qw(
                ols_placement_on_creation_is_valid
                ols_size_on_creation_is_valid
                )
        ],
        expand => [
            qw(
                ols_size_after_expansion_is_valid
                )
        ]
    );

    my $ok = $self->SUPER::validate_infinitevol(%parameters);

    my $result         = $ok;
    my @validate_these = @{ $validations{ $self->{'operation'} } }
        if defined $validations{ $self->{'operation'} };

    foreach my $method (@validate_these) {
        if (@validate_these_only) {
            $ok = $self->$method()
                if ( grep( /^$method$/, @validate_these_only ) );
        }
        else {
            $ok = $self->$method();
        }
        $result = $ok && $result;
    }

    return $result;
}

=head2 ols_placement_on_creation_is_valid

This function returns boolean as to whether or not the OLS constituent was placed correctly.

=cut

sub ols_placement_on_creation_is_valid {
    my $self       = shift;
    my %parameters = @_;

    $Log->log('Validation Starting: OLS Constituents Placement on Creation');
    if ( $self->{'final_state'}->{'enable_object_store'} eq 'false' ) {
        $Log->log(
            'Validation Ending: OLS Constituents Placement on Creation not applicable'
        );
        return $TRUE;
    }

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    my @nodes = @{ $self->{'final_state'}->{'list_of_nodes'} };
    @nodes = NACL::CS::Node->fetch( command_interface => $node )
        unless @nodes;
    my %availsizes   = %{ $self->_get_available_size() };
    my $num_of_nodes = $self->_get_number_of_nodes(
        'node'       => $node,
        'availsizes' => \%availsizes
    );

    my @ols = $self->_get_constituent_by_role( 'role' => 'ols' );
    if ( scalar @ols == $num_of_nodes ) {
        $Log->log('Validation Success:  Correct number of OLS created');
    }
    else {
        $Log->log('Validation Failure: '
                . @ols
                . ' OLS created, while '
                . @nodes
                . ' are expected' );
        return $FALSE;
    }

    my @ols_nodes      = map { $_->node } @ols;
    my @ols_aggregates = map { $_->aggregate } @ols;

    foreach my $n ( map { $_->node } @nodes ) {
        if ( !grep( /^$n$/, @ols_nodes ) ) {
            $Log->log( 'Validation Failure: No OLS created in node: ' . $n );
            return $FALSE;
        }
    }
    $Log->log('Validation Success: OLS exist in all nodes');

    my $ols_aggr_list = $self->{'final_state'}->{'ols_aggr_list'};
    if (@$ols_aggr_list) {
        foreach my $ols_aggr (@ols_aggregates) {
            if ( !grep( /^$ols_aggr$/, @{$ols_aggr_list} ) ) {
                $Log->log(
                    'Validation Failure: OLS aggregate was not placed on an aggregate in the ols-aggr-list'
                );
                return $FALSE;
            }
        }
        $Log->log(
            'Validation Success: OLS aggregates were placed on aggregates in the ols--aggr-list'
        );
    }
    else {
        foreach my $ols_aggr (@ols_aggregates) {
            if ( !grep( /^$ols_aggr$/, keys %availsizes ) ) {
                $Log->log(
                    'Validation Failure: OLS aggregate was not placed on an aggregate in the vserver aggr-list'
                );
                return $FALSE;
            }
        }
        $Log->log(
            'Validation Success: OLS aggregates were placed on aggregates in the vserver aggr-list'
        );
    }

    $Log->log(
        'Validation Ending: OLS Constituents Placement on Creation was Successful'
    );

    return $TRUE;
}

=head2 ols_size_on_creation_is_valid

This function returns boolean as to whether or not the OLS constituent was sized correctly.

=cut

sub ols_size_on_creation_is_valid {
    my $self       = shift;
    my %parameters = @_;

    $Log->log('Validation Starting: OLS Constituent Sizing on Creation');
    if ( $self->{'final_state'}->{'enable_object_store'} eq 'false' ) {
        $Log->log(
            'Validation Ending: OLS Constituent Sizing on Creation not applicable'
        );
        return $TRUE;
    }

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    my %availsizes   = %{ $self->_get_available_size() };
    my $num_of_nodes = $self->_get_number_of_nodes(
        'node'       => $node,
        'availsizes' => \%availsizes
    );

    my $data_aggr_list = $self->{'final_state'}->{'data_aggr_list'};
    my $num_dc_nodes   = $self->{'final_state'}->{'num_dc_nodes'};
    unless ( defined $num_dc_nodes ) {
        $num_dc_nodes = $self->get_number_of_dc_nodes(
            'node'           => $node,
            'availsizes'     => \%availsizes,
            'data-aggr-list' => $data_aggr_list
        );
    }

    my @ols = $self->_get_constituent_by_role( 'role' => 'ols' );
    if ( scalar @ols == $num_of_nodes ) {
        $Log->log('Validation Success:  Correct number of OLS created');
    }
    else {
        $Log->log('Validation Failure: '
                . @ols
                . ' OLS created, while '
                . $num_of_nodes
                . ' are expected' );
        return $FALSE;
    }

    my @ols_sizes = uniq( map( $_->size, @ols ) );
    if ( scalar @ols_sizes == 1 ) {
        $Log->log(
            'Validation Success:  All OLS constituents have the same size');
    }
    else {
        $Log->log(
            'Validation Failure:  Not all OLS constituents have the same size'
        );
        return $FALSE;
    }

    my $max_ols_size = $DEFAULT_OLS_MAX_SIZE;
    $Log->log('Validation Info: Default max OLS Constituent Size is '
            . $max_ols_size
            . ' bytes' );

    # Hash of results from different sizing limites
    my %sizings = ();

    # Min usable space among the OLS aggregates
    my %ols_aggrs        = ();
    my @ols_usable_space = ();
    foreach my $ols (@ols) {
        my $ols_aggr_name = $ols->aggregate;
        my $usable_space  = $self->max_usable_space(
            'node'      => $node,
            'aggregate' => $ols_aggr_name,
            'volume'    => $ols,
            'size'      => $self->{'initial_state'}->{'aggrs_sizes'}->{'size'}
                ->{$ols_aggr_name},
            'availsize'       => $availsizes{$ols_aggr_name},
            'space-guarantee' => 'volume'
        ) unless defined $ols_aggrs{$ols_aggr_name};
        $ols_aggrs{$ols_aggr_name} = 1;
        $Log->log( 'Validation Info: usable space is ' . $usable_space );
        push( @ols_usable_space, $usable_space );
    }
    @ols_usable_space = sort { $a <=> $b } @ols_usable_space;
    $sizings{'Minimum Usable Space of the OLS aggregates'}
        = $ols_usable_space[0];

    my $ns_size
        = $self->_get_constituent_by_role( 'role' => 'namespace' )->size;
    my @ns_mirror = $self->_get_constituent_by_role( 'role' => 'ns_mirror' );
    my $ns_mirror_size = 0;
    foreach my $nm (@ns_mirror) {
        $ns_mirror_size += $nm->size;
    }
    my $infinitevol_size = $self->{'final_state'}->{'infinitevol'}->size;
    $sizings{'Max OLS Aggregate Size for Vserver'} = nhimult( $WAFL_BLOCK,
        ( $infinitevol_size - $ns_mirror_size - $ns_size ) / $num_of_nodes );

    # Percentage of Infinitevol
    my $specified_infinitevol_size
        = $self->{'final_state'}->{'size_specified'};
    if ( defined $specified_infinitevol_size ) {
        $Log->log('Validation Info: Specified InfiniteVol Size is '
                . $specified_infinitevol_size
                . ' bytes' );
        $Log->log(
            'Validation Info: Using Specified InfiniteVol Size for Percentage Calculation'
        );
        $infinitevol_size = $specified_infinitevol_size;
    }
    else {
        my $default_infinitevol_size
            = (   $num_dc_nodes
                * $self->{'helpers'}
                ->minimum_data_constituent_size( node => $node ) ) / 0.5;
        $Log->log('Validation Info: Default InfiniteVol Size is '
                . $default_infinitevol_size
                . ' bytes' );
        $Log->log(
            'Validation Info: Using Default InfiniteVol Size for Percentage Calculation'
        );
        $infinitevol_size = $default_infinitevol_size;
    }

    $sizings{'25% of the InfiniteVol size'}
        = nhimult( $WAFL_BLOCK, 0.25 * $infinitevol_size / $num_of_nodes );

    # Max namespace constituent size
    $sizings{'Default Maximum OLS Constituent Size'} = $max_ols_size;

    # Choose smallest of the methods
    my $method_chosen
        = ( sort { $sizings{$a} <=> $sizings{$b} } keys %sizings )[0];

    $Log->log( 'Validation Info: Smallest size method is ' . $method_chosen );
    $Log->log('Validation Info: Size from that method is '
            . $sizings{$method_chosen}
            . ' bytes' );

    if ( $sizings{$method_chosen} < $MIN_OLS_SIZE ) {
        $Log->log(
            'Validation Info: OLS limit chosen to be Minimum FlexVol Size');
        $Log->log('Validation Info: OLS size should be '
                . $MIN_NS_SIZE
                . ' bytes' );
        $Log->log(
            'Validation Info: OLS size was ' . $ols_sizes[0] . ' bytes' );

        my $difference = abs( $MIN_OLS_SIZE - $ols_sizes[0] );
        $Log->log(
            'Validation Info: Difference between actual and expected is '
                . $difference
                . ' bytes' );

        if ( $difference > 0 ) {
            $Log->log(
                'Validation Failure: OLS did not match Minimum FlexVol Size'
            );
            return $FALSE;
        }
        else {
            $Log->log('Validation Success: OLS matched Minimum FlexVol Size');
        }
    }
    else {
        $Log->log(
            'Validation Info: OLS limit chosen to be ' . $method_chosen );
        $Log->log('Validation Info: OLS size should be '
                . $sizings{$method_chosen}
                . ' bytes' );
        $Log->log(
            'Validation Info: OLS size was ' . $ols_sizes[0] . ' bytes' );

        my $difference = abs( $sizings{$method_chosen} - $ols_sizes[0] );
        $Log->log(
            'Validation Info: Difference between actual and expected is '
                . $difference
                . ' bytes' );

        if ( $difference > $DEFAULT_TOLERANCE ) {
            $Log->log(
                'Validation Failure: OLS did not match ' . $method_chosen );
            return $FALSE;
        }
        else {
            $Log->log( 'Validation Success: OLS matched ' . $method_chosen );
        }
    }

    $Log->log(
        'Validation Ending: OLS Constituent Sizing on Creation was Successful'
    );
    return $TRUE;
}

=head2 ols_size_after_expansion_is_valid

This function returns boolean as to whether or not the ols constituents were sized correctly.

=cut

sub ols_size_after_expansion_is_valid {
    my $self       = shift;
    my %parameters = @_;

    $Log->log('Validation Starting: OLS Constituents Sizing on Expansion');
    if ( $self->{'final_state'}->{'enable_object_store'} eq 'false' ) {
        $Log->log(
            'Validation Ending: OLS Constituents Sizing on Expansion not applicable'
        );
        return $TRUE;
    }

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    my %availsizes   = %{ $self->_get_available_size() };
    my $num_of_nodes = $self->_get_number_of_nodes(
        'node'       => $node,
        'availsizes' => \%availsizes
    );

    # OLS after expansion
    my @ols = $self->_get_constituent_by_role( 'role' => 'ols' );
    if ( scalar @ols == $num_of_nodes ) {
        $Log->log('Validation Success:  Correct number of OLS created');
    }
    else {
        $Log->log('Validation Failure: '
                . @ols
                . ' OLS created, while '
                . $num_of_nodes
                . ' are expected' );
        return $FALSE;
    }

    my @ols_sizes = uniq( map( $_->size, @ols ) );
    if ( scalar @ols_sizes == 1 ) {
        $Log->log(
            'Validation Success:  All OLS constituents have the same size');
    }
    else {
        $Log->log(
            'Validation Failure:  Not all OLS constituents have the same size'
        );
        return $FALSE;
    }

    # OLS before expansion
    my @original_ols = $self->_get_constituent_by_role(
        'role'  => 'ols',
        'state' => 'initial'
    );
    my @original_ols_sizes = map( $_->size, @ols );

    # Hash of results from different sizing limites
    my %sizings = ();

    # Usable aggregate space for OLS
    my %ols_aggrs        = ();
    my @ols_usable_space = ();
    foreach my $ols (@ols) {
        my $ols_aggr_name = $ols->aggregate;
        my $usable_space  = $self->max_usable_space(
            'node'      => $node,
            'aggregate' => $ols_aggr_name,
            'volume'    => $ols,
            'size'      => $self->{'initial_state'}->{'aggrs_sizes'}->{'size'}
                ->{$ols_aggr_name},
            'availsize'       => $availsizes{$ols_aggr_name},
            'space-guarantee' => 'volume'
        ) unless defined $ols_aggrs{$ols_aggr_name};
        $ols_aggrs{$ols_aggr_name} = 1;
        $Log->log( 'Validation Info: usable space is ' . $usable_space );
        push( @ols_usable_space, $usable_space );
    }
    @ols_usable_space = sort { $a <=> $b } @ols_usable_space;
    $sizings{'Usable Aggregate Space for OLS'} = $ols_usable_space[0];

    # Max OLS Aggregate Size for Vserver
    my $ns_size
        = $self->_get_constituent_by_role( 'role' => 'namespace' )->size;
    my @ns_mirror = $self->_get_constituent_by_role( 'role' => 'ns_mirror' );
    my $ns_mirror_size = 0;
    foreach my $nm (@ns_mirror) {
        $ns_mirror_size += $nm->size;
    }
    my $infinitevol_size = $self->{'final_state'}->{'infinitevol'}->size;
    $sizings{'Max OLS Aggregate Size for Vserver'} = nhimult( $WAFL_BLOCK,
        ( $infinitevol_size - $ns_mirror_size - $ns_size ) / $num_of_nodes );

    # Percentage of Infinitevol
    my $specified_infinitevol_size
        = $self->{'final_state'}->{'size_specified'};
    $Log->log('Validation Info: Specified InfiniteVol Size is '
            . $specified_infinitevol_size
            . ' bytes' );

    $sizings{'25% of the InfiniteVol size'} = nhimult( $WAFL_BLOCK,
        0.25 * $specified_infinitevol_size / $num_of_nodes );

    # Max namespace constituent size
    my $max_ols_size = $DEFAULT_OLS_MAX_SIZE;
    $Log->log('Validation Info: Max Namespace Constituent Size is '
            . $max_ols_size
            . ' bytes' );
    $sizings{'Maximum OLS Constituent Size'} = $max_ols_size;

    # Choose smallest of the methods
    my $method_chosen
        = ( sort { $sizings{$a} <=> $sizings{$b} } keys %sizings )[0];

    $Log->log( 'Validation Info: Smallest size method is ' . $method_chosen );
    $Log->log('Validation Info: Size from that method is '
            . $sizings{$method_chosen}
            . ' bytes' );

    if ( $sizings{$method_chosen} < $ols_sizes[0] ) {
        $Log->log(
            'Validation Info: Desired size for expansion would shrink the OLS, validating that the size remained constant'
        );
        if ( $original_ols_sizes[0] == $ols_sizes[0] ) {
            $Log->log('Validation Success: OLS size was not changed');
        }
        else {
            $Log->log('Validation Failure: OLS was changed');
            return $FALSE;
        }
    }
    else {
        $Log->log(
            'Validation Info: Desired size for expansion would not shrink the OLS, validating the new size'
        );
        if ( $sizings{$method_chosen} < $MIN_OLS_SIZE ) {
            $Log->log(
                'Validation Info: OLS limit chosen to be Minimum FlexVol Size'
            );
            $Log->log('Validation Info: OLS size should be '
                    . $MIN_OLS_SIZE
                    . ' bytes' );

            my $difference = abs( $MIN_OLS_SIZE - $ols_sizes[0] );
            $Log->log(
                'Validation Info: Difference between actual and expected is '
                    . $difference
                    . ' bytes' );

            if ( $difference > 0 ) {
                $Log->log(
                    'Validation Failure: OLS did not match Minimum FlexVol Size'
                );
                return $FALSE;
            }
            else {
                $Log->log(
                    'Validation Success: OLS matched Minimum FlexVol Size');
            }
        }
        else {
            $Log->log(
                'Validation Info: OLS limit chosen to be ' . $method_chosen );
            $Log->log('Validation Info: OLS size should be '
                    . $sizings{$method_chosen}
                    . ' bytes' );

            my $difference = abs( $sizings{$method_chosen} - $ols_sizes[0] );
            $Log->log(
                'Validation Info: Difference between actual and expected is '
                    . $difference
                    . ' bytes' );

            if ( $difference > $DEFAULT_TOLERANCE ) {
                $Log->log( 'Validation Failure: OLS did not match '
                        . $method_chosen );
                return $FALSE;
            }
            else {
                $Log->log(
                    'Validation Success: OLS matched ' . $method_chosen );
            }
        }
    }
    $Log->log(
        'Validation Ending: OLS Constituents Sizing on Expansion was Successful'
    );

    return $TRUE;
}

1;
