# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary Balanced Infinite Volume Library
## @author  Adam.Ciapponi@netapp.com
## @pod     here

=head1 NAME

CR::Setup::Validation::Balanced

=head1 DESCRIPTION

This module contains methods for testing Balanced infinite volumes

=head1 METHODS

=cut

package CR::Setup::Validation::Balanced::SierraNevada;

# Compiler directives.
use strict;
use warnings;

# Module import
use CR::Setup::TestParams;
use CR::Setup::Constants;
use CR::Setup::Constants::SierraNevada;
use List::Util qw(first max maxstr min minstr reduce shuffle sum);
use Params::Validate qw(:all);
use Math::Round qw(:all);
use List::MoreUtils qw(:all);
use NACL::C::Node;
use Tharn qw($Log);
use Test::Deep::NoTest qw(eq_deeply);
use Data::Dumper::Concise;

use CR::Setup::Expectation;
use CR::UnitTest;    #FIXME remove this
use NATE::BaseException qw(:try);
use NACL::APISet;
use NACL::C::Snapmirror;
use NACL::C::Volume;
use NACL::CS::VolumeShowFootprint;
use NACL::C::Vserver;
use Storable qw(dclone);

our @ISA = qw(CR::Setup::Validation::Balanced);

=head2 new

Basic initialization for the validation library.

=cut

sub new {
    my $class   = shift;
    my %params  = @_;
    my $options = new Options;

    $options->parameters(definitions=>[
        {   name        => 'AGGR_STABLE_WAIT_TIME',
            description => 'Number of seconds to wait for aggregates size to become stabilized',
            spec        => 's',
            default     => '60',
        }
    ]);

    $options->process();
    my $wait_time = $options->get( option => 'AGGR_STABLE_WAIT_TIME' );

    my $self = {
        node                => $params{'node'},
        defaults            => $params{'defaults'},
        helpers             => $params{'helpers'},
        operation           => $params{'operation'},
        initial_state       => undef,
        final_state         => undef,
        aggregate_wait_time => $wait_time,
    };

    bless $self, $class;
    return $self;
}

=head2 get_initial_state

This function records information on all volumes, and size and available size information on all aggregates before InfiniteVol
creation or expansion

=cut

sub get_initial_state {
    my $self       = shift;
    my %parameters = @_;

    unless ( defined $parameters{'node'} ) {
        $parameters{'node'} = $self->{'node'};
    }

    $self->{'helpers'}
        ->disable_background_operations( node => $parameters{'node'} );

    $self->_stabilize_aggregates(%parameters)
        if defined $parameters{'vserver'};

    $self->{'initial_state'}->{'volume_info_by_aggr'}
        = $self->_get_volume_info_by_aggr(%parameters);
    $self->{'initial_state'}->{'volume_footprint_info_by_aggr'}
        = $self->_get_volume_footprint_info_by_aggr(%parameters);
    $self->{'initial_state'}->{'aggrs_sizes'}
        = $self->_get_aggrs_sizes(%parameters);

    $self->{'helpers'}
        ->enable_background_operations( node => $parameters{'node'} );

    return $self->{'initial_state'};
}

=head2 get_final_state

This function records information on all volumes, and size and available size information on all aggregates after InfiniteVol
creation or expansion

=cut

sub get_final_state {
    my $self       = shift;
    my %parameters = @_;

    unless ( defined $parameters{'node'} ) {
        $parameters{'node'} = $self->{'node'};
    }

    $self->{'helpers'}
        ->disable_background_operations( node => $parameters{'node'} );

    $self->_stabilize_aggregates(%parameters)
        if defined $parameters{'vserver'};

    $self->{'final_state'}->{'volume_info_by_aggr'}
        = $self->_get_volume_info_by_aggr(%parameters);
    $self->{'final_state'}->{'volume_footprint_info_by_aggr'}
        = $self->_get_volume_footprint_info_by_aggr(%parameters);
    $self->{'final_state'}->{'aggrs_sizes'}
        = $self->_get_aggrs_sizes(%parameters);

    $self->{'helpers'}
        ->enable_background_operations( node => $parameters{'node'} );
    return $self->{'final_state'};
}

=head2 validate_infinitevol

This function validates the correctness of Infinitevol expansion and creation operations

=cut

sub validate_infinitevol {
    my $self                = shift;
    my %parameters          = @_;
    my @validate_these_only = ();
    @validate_these_only    = @{ $parameters{'validate_these_only'} }
        if defined $parameters{'validate_these_only'};
    my %validations = (
        create => [
            qw(
                namespace_placement_on_creation_is_valid
                ns_mirror_placement_on_creation_is_valid
                data_placement_on_creation_is_valid
                namespace_size_on_creation_is_valid
                ns_mirror_size_on_creation_is_valid
                dcs_size_on_creation_is_valid
                infinitevol_size_on_creation_is_valid
                )
        ],
        expand => [
            qw(
                namespace_size_after_expansion_is_valid
                infinitevol_size_after_expansion_is_valid
                )
        ]
    );

    my $result         = $TRUE;
    my @validate_these = @{ $validations{ $self->{'operation'} } }
        if defined $validations{ $self->{'operation'} };

    unless (@validate_these) {
        return $FALSE;
    }

    foreach my $method (@validate_these) {
        my $ok = $TRUE;
        if ( @validate_these_only ) {
            $ok = $self->$method()
                if ( grep( /^$method$/, @validate_these_only ) );
        }
        else {
            $ok = $self->$method();
        }
        $result = $ok && $result;
    }
    return $result;
}

=head2 balanced_create

This function does balanced creation

=cut

sub balanced_create {
    my $self       = shift;
    my %parameters = @_;

    # Node setup
    my $node = $parameters{'node'};

    # Vserver setup
    my $vs = $parameters{'vserver'};
    $vs = NACL::CS::Vserver->fetch(
        command_interface => $node,
        filter            => { vserver => $vs->vserver }
    );
    my $vs_name      = $vs->vserver;
    my @vs_aggr_list = $vs->aggr_list;
    $self->{'helpers'}->disable_background_operations( node => $node );

    # Parameters for the create command
    my $cli  = $parameters{'cli'};
    my $zapi = $parameters{'zapi'};

    my $use_zapi_for_create = $parameters{'use_zapi_for_create'};
    unless ( defined $use_zapi_for_create ) {
        $use_zapi_for_create = $FALSE;
    }

    my %create_args = %{ $parameters{'create_args'} };
    unless ( defined $create_args{'vserver'} ) {
        $create_args{'vserver'} = $vs_name;
    }
    unless ( defined $create_args{'snapshot-policy'} ) {
        $create_args{'snapshot-policy'} = 'none';
    }

    my $create_to_max = $parameters{'create_to_max'};
    unless ( defined $create_to_max ) {
        $create_to_max = $FALSE;
    }

    my $use_zapi_for_max = $parameters{'use_zapi_for_max'};
    unless ( defined $use_zapi_for_max ) {
        $use_zapi_for_max = $FALSE;
    }

    my $max_factor = $parameters{'max_factor'};
    unless ( defined $max_factor ) {
        $max_factor = 1;
    }

    $self->_stabilize_aggregates(
        'node'           => $node,
        'vserver'        => $vs,
        'data-aggr-list' => $create_args{'data-aggr-list'}
    );

    my %dereferenced_create_args = %create_args;
    foreach my $array_element
        qw(ols-aggr-list data-aggr-list ns-mirror-aggr-list) {
        if ( defined $create_args{$array_element} ) {
            $dereferenced_create_args{$array_element} = join ',',
                @{ $create_args{$array_element} };
        }
    }

    if ($create_to_max) {
        my $max_was_found = $FALSE;
        if ($use_zapi_for_max) {
            my $max_size
                = $max_factor * $self->_get_infinitevol_max_size(%parameters);
            $create_args{'size'}              = $max_size;
            $dereferenced_create_args{'size'} = $max_size;
            $max_was_found                    = $TRUE;
        }
        else {
            $dereferenced_create_args{'size'} = '20PB';
            my %max_args = %dereferenced_create_args;
            %max_args
                = _zapi_to_cli_mapping_for_create_async(
                %{ dclone( \%dereferenced_create_args ) } )
                if $use_zapi_for_create;
            my $response;
            this_block {
                $response = $cli->volume_create(
                    'foreground' => 'true',
                    %max_args
                );
            }
            should( fail()->and_return_error_message( \$response ) );
            if ( $response =~ m/maximum size use "([0-9]+)B"/ ) {
                my $max_size = nlowmult( $WAFL_BLOCK, $max_factor * $1 );
                $max_size
                    = $create_args{'space-guarantee'} eq 'none'
                    ? $max_size - $THIN_PROVISIONING_MAX_SIZE_BUFFER
                    : $max_size - $THICK_PROVISIONING_MAX_SIZE_BUFFER;
                $create_args{'size'}              = $max_size;
                $dereferenced_create_args{'size'} = $max_size;
                $max_was_found                    = $TRUE;
            }
            else {
                $Log->log('Validation Failure: Could not find max size');
            }
        }
        this($max_was_found)->should( Be->true() );
    }

    if ( !$use_zapi_for_create ) {

 # Create should succeed, if it should fail, just run the command individually
        this_block {
            $cli->volume_create(
                'foreground' => 'true',
                %dereferenced_create_args
            );
        }
        should( pass() );
    }
    else {

        # zapi apiset already contains vserver info and will cause an error
        # if we try to pass in a vserver, so remove it
        delete $create_args{'vserver'};
        my $response;
        this_block {
            $response = $zapi->volume_create_async(%create_args);
        }
        should( pass() );

        # wait for create job to complete
        this(
            validate_zapi_job_id(
                response_obj        => $response,
                job_tag             => 'result-jobid',
                job_description     => 'Volume Create',
                wait_complete       => 'true',
                'command_interface' => $node
                ) == $TCD::FAIL
        )->should( Be->false() );

        %create_args = _zapi_to_cli_mapping_for_create_async(%create_args);
    }

    $self->{'helpers'}->enable_background_operations( node => $node );

    my $iv = NACL::CS::Volume->fetch(
        command_interface => $node,
        filter            => {
            'volume-style' => 'infinitevol',
            'vserver'      => $vs_name
        },
    );

    $create_args{'vserver'}        = $vs;
    $create_args{'infinitevol'}    = $iv;
    $create_args{'vs_aggr_list'}   = \@vs_aggr_list;
    $create_args{'size_specified'} = $create_args{'size'};
    $self->_set_final_state_info(%create_args);

    return $TRUE;
}

=head2 balanced_expand_and_validate

This function does balanced expansion.

=cut

sub balanced_expand {
    my $self       = shift;
    my %parameters = @_;

    # Node setup
    my $node = $parameters{'node'};

    # Vserver setup
    my $vs = $parameters{'vserver'};
    $vs = NACL::CS::Vserver->fetch(
        command_interface => $node,
        filter            => { vserver => $vs->vserver }
    );
    my $vs_name      = $vs->vserver;
    my @vs_aggr_list = $vs->aggr_list;

    $self->{'helpers'}->disable_background_operations( node => $node );

    # Parameters for the expansion command
    my $cli  = $parameters{'cli'};
    my $zapi = $parameters{'zapi'};

    my %expand_args = %{ $parameters{'expand_args'} };
    unless ( defined $expand_args{'vserver'} ) {
        $expand_args{'vserver'} = $vs_name;
    }

    my %validation_args = ();
    if ( defined $parameters{'validation_args'} ) {
        %validation_args = %{ $parameters{'validation_args'} };
    }

    my $use_zapi_for_expand = $parameters{'use_zapi_for_expand'};
    unless ( defined $use_zapi_for_expand ) {
        $use_zapi_for_expand = $FALSE;
    }

    my $expand_command = $parameters{'expand_command'};
    unless ( defined $expand_command ) {
        $expand_command = 'volume_modify';
    }

    my $expand_to_max = $parameters{'expand_to_max'};
    unless ( defined $expand_to_max ) {
        $expand_to_max = $FALSE;
    }

    my $use_zapi_for_max = $parameters{'use_zapi_for_max'};
    unless ( defined $use_zapi_for_max ) {
        $use_zapi_for_max = $FALSE;
    }

    my $max_factor = $parameters{'max_factor'};
    unless ( defined $max_factor ) {
        $max_factor = 1;
    }

    if (   $expand_command eq 'volume_modify'
        && $use_zapi_for_expand == $FALSE )
    {
        $expand_args{'foreground'} = 'true';
    }

    my $iv = NACL::CS::Volume->fetch(
        command_interface => $node,
        filter            => {
            'volume-style' => 'infinitevol',
            'vserver'      => $vs_name
        },
        requested_fields => ['space-guarantee']
    );
    $parameters{'space-guarantee'} = $iv->space_guarantee;

    my $ns = NACL::CS::Volume->fetch(
        command_interface => $node,
        filter            => {
            'is-constituent'   => 'true',
            'constituent-role' => 'namespace',
            'vserver'          => $vs_name
        }
    );
    my $original_ns_size = $ns->size;

    my $ns_mirror_before = NACL::CS::Volume->fetch(
        command_interface => $node,
        filter            => {
            'is-constituent'   => 'true',
            'constituent-role' => 'ns_mirror',
            'vserver'          => $vs_name
        },
        allow_empty => 1,
    );

    $self->_stabilize_aggregates(
        'node'           => $node,
        'vserver'        => $vs,
        'data-aggr-list' => $expand_args{'data-aggr-list'}
    );

    if ($expand_to_max) {
        my $max_was_found = $FALSE;
        if ($use_zapi_for_max) {
            my $max_size
                = $max_factor * $self->_get_infinitevol_max_size(%parameters);
            $expand_args{'size'}               = $max_size;
            $validation_args{'size_specified'} = $max_size;
            $max_was_found                     = $TRUE;
        }
        else {
            my $response;
            this_block {
                $response = $cli->volume_modify(
                    'foreground' => 'true',
                    'size'       => '20PB',
                    'vserver'    => $expand_args{'vserver'},
                    'volume'     => $expand_args{'volume'},
                );
            }
            should( fail()->and_return_error_message( \$response ) );
            if ( $response =~ m/maximum size use "([0-9]+)B"/ ) {
                my $max_size = nlowmult( $WAFL_BLOCK, $max_factor * $1 );
                $max_size
                    = $iv->space_guarantee eq 'none'
                    ? $max_size - $THIN_PROVISIONING_MAX_SIZE_BUFFER
                    : $max_size - $THICK_PROVISIONING_MAX_SIZE_BUFFER;
                $expand_args{'size'}               = $max_size;
                $validation_args{'size_specified'} = $max_size;
                $max_was_found                     = $TRUE;
            }
            else {
                $Log->log('Validation Failure: Could not find max size');
            }
        }
        this($max_was_found)->should( Be->true() );
    }

# Expansion should succeed, if it should fail, just run the command individually
    if ( $use_zapi_for_expand == $FALSE ) {
        this_block {
            $cli->$expand_command(%expand_args);
        }
        should( pass() );
    }
    else {
        my $response;
        this_block {
            $response = $zapi->volume_modify_iter_async(
                query => {
                    'volume-id-attributes' =>
                        { 'name' => $expand_args{'volume'} }
                },
                attributes => {
                    'volume-space-attributes' =>
                        { 'size' => $expand_args{'size'} }
                },
            );
        }
        should( pass() );

        # wait for expansion job to complete
        this(
            validate_zapi_job_id(
                response_obj        => $response,
                job_tag             => 'jobid',
                job_description     => 'Volume Modify',
                wait_complete       => 'true',
                'command_interface' => $node
                ) == $TCD::FAIL
        )->should( Be->false() );
    }

    $self->{'helpers'}->enable_background_operations( node => $node );

    my $iv_expanded = NACL::CS::Volume->fetch(
        command_interface => $node,
        filter            => {
            'volume-style' => 'infinitevol',
            'vserver'      => $vs_name
        },
    );

    $expand_args{'vserver'}        = $vs;
    $expand_args{'infinitevol'}    = $iv_expanded;
    $expand_args{'vs_aggr_list'}   = \@vs_aggr_list;
    $expand_args{'size_specified'} = $validation_args{'size_specified'};
    $expand_args{'max-namespace-constituent-size'}
        = $validation_args{'max-namespace-constituent-size'};

    $self->_set_final_state_info(%expand_args);

    return $TRUE;
}

=head2 max_usable_space

This function returns the maximum usable space based upon the space guarantee, available space, and the size of the aggregate

=cut

sub max_usable_space {
    my $self                      = shift;
    my %parameters                = @_;
    my $dedupe_metadata_footprint = 0;
    my $total_volume_footprint    = 0;
    my $volume_reserve            = 0;

    $parameters{'expansion'}
        = $self->{'operation'} eq 'expand' ? $TRUE : $FALSE;

    my $space_info = $self->_get_volume_space_info_for_aggregate(
        aggregate => $parameters{'aggregate'},
        node      => $parameters{'node'},
        volume    => $parameters{'volume'},
        expansion => $parameters{'expansion'}
    );

    $dedupe_metadata_footprint = $space_info->{'dedupe_metadata_footprint'};
    $Log->log(
        "Validation Info: dedupe metadata footprint is $dedupe_metadata_footprint"
    );

    $total_volume_footprint = $space_info->{'total_volume_footprint'};
    $Log->log(
        "Validation Info: total volume footprint is $total_volume_footprint"
    );

    $volume_reserve = $space_info->{'volume_reserve'};
    $Log->log("Validation Info: volume reserver is $volume_reserve");

    my $wafl_reserve = 0.03 * $parameters{'size'};
    $Log->log("Validation Info: wafl reserve is $wafl_reserve");

    my $dedupe_reserve
        = 0.06 * $parameters{'size'} - $dedupe_metadata_footprint;
    $Log->log("Validation Info: dedupe reserve is $dedupe_reserve");

    my $actual_available
        = $parameters{'availsize'}
        - $wafl_reserve
        - $dedupe_reserve
        - $volume_reserve;

    if ( $parameters{'expansion'} ) {
        $actual_available -= $CONSTITUENT_EXPANSION_TAX;
    }
    else {
        $actual_available -= $CONSTITUENT_CREATION_TAX;
    }

    if ( $actual_available < 0 ) {
        $actual_available = 0;
    }

    $actual_available += $total_volume_footprint
        if $parameters{'expansion'};

    if ( $parameters{'space-guarantee'} eq 'volume' ) {
        return nlowmult( $WAFL_BLOCK,
            $THICK_PROVISIONING_RATIO * $actual_available );
    }
    elsif ( $parameters{'space-guarantee'} eq 'none' ) {
        return nlowmult( $WAFL_BLOCK,
            $THIN_PROVISIONING_RATIO * $actual_available );
    }
    else {
        return;
    }
}

=head2 find_largest_aggregates

This function returns the largest aggregates. These are aggregates that can be chosen for namespace creation.

Once tiebreaker logic is available for multiple with the same size, this will always return a single aggregate.

=cut

sub find_largest_aggregates {
    my $self       = shift;
    my %parameters = @_;

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    # Vserver setup
    my %availsizes   = %{ $self->_get_available_size() };
    my @vs_aggr_list = @{ $self->{'final_state'}->{'vs_aggr_list'} };
    my %usablesizes  = ();

    foreach my $aggr_name ( keys %availsizes ) {
        my $aggr = NACL::CS::Aggregate->fetch(
            command_interface => $node,
            filter            => { 'aggregate' => $aggr_name }
        );

        $usablesizes{$aggr_name} = $self->max_usable_space(
            'node'            => $node,
            'aggregate'       => $aggr->aggregate,
            'size'            => $aggr->size,
            'availsize'       => $availsizes{$aggr_name},
            'volume'          => $parameters{'volume'},
            'space-guarantee' => 'volume'
        );
    }

    my @largest_aggregates = ();

    for ( my $i = 0; $i < ( keys %usablesizes ); $i++ ) {
        my $cur_max = (
            reverse sort { $usablesizes{$a} <=> $usablesizes{$b} }
                keys %usablesizes
        )[$i];
        if (   @largest_aggregates > 0
            && $usablesizes{ $largest_aggregates[-1] }
            > $usablesizes{$cur_max} )
        {
            last;
        }
        else {
            $Log->log( 'Validation Info: Possible largest aggregate is '
                    . $cur_max );
            push( @largest_aggregates, $cur_max );
        }
    }

    # Tiebreaker, first in vserver aggregate list
    foreach my $vs_aggr (@vs_aggr_list) {
        if ( grep( /^$vs_aggr$/, @largest_aggregates ) ) {
            $Log->log(
                'Validation Info: Largest aggregate chosen is ' . $vs_aggr );
            @largest_aggregates = ($vs_aggr);
            last;
        }
    }

    return @largest_aggregates;
}

=head2 namespace_placement_on_creation_is_valid

This function returns boolean as to whether or not the namespace constituent was placed correctly.

=cut

sub namespace_placement_on_creation_is_valid() {
    my $self       = shift;
    my %parameters = @_;

    $Log->log(
        'Validation Starting: Namespace Constituent Placement on Creation');

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    # Namespace setup
    my $ns = $self->_get_constituent_by_role( 'role' => 'namespace' );
    my $ns_aggr_name = $ns->aggregate;
    $parameters{'namespace'} = $ns;

    $Log->log('Validation Success: Namespace was created');

    my $ns_aggr_specified = $self->{'final_state'}->{'namespace_aggregate'};
    my %availsizes        = %{ $self->_get_available_size() };

    if ( !grep( /^$ns_aggr_name$/, keys %availsizes ) ) {
        $Log->log(
            'Validation Failure: Namespace was not on an aggregate in the vserver aggregate list'
        );
        return $FALSE;
    }
    else {
        $Log->log(
            'Validation Success: Namespace was on an aggregate in the vserver aggregate list'
        );
        if ($ns_aggr_specified) {
            if ( $ns_aggr_name ne $ns_aggr_specified ) {
                $Log->log(
                    'Validation Failure: Namespace was not on the specified aggregate. Expected: '
                        . $ns_aggr_specified
                        . ' Actual: '
                        . $ns_aggr_name );
                return $FALSE;
            }
            else {
                $Log->log(
                    'Validation Success: Namespace was on the specified aggregate. Expected: '
                        . $ns_aggr_specified
                        . ' Actual: '
                        . $ns_aggr_name );
            }
        }
        else {
            my @possible_aggrs = $self->find_largest_aggregates(
                'volume' => $parameters{'namespace'} );
            if ( !grep( /^$ns_aggr_name$/, @possible_aggrs ) ) {
                $Log->log(
                    'Validation Failure: Namespace was not placed on the largest aggregate in the vserver aggregate list'
                );
                return $FALSE;
            }
            else {
                $Log->log(
                    'Validation Success: Namespace was placed on the largest aggregate in the vserver aggregate list'
                );
            }
        }
    }

    $Log->log(
        'Validation Ending: Namespace Constituent Placement on Creation was Successful'
    );

    return $TRUE;
}

=head2 ns_mirror_placement_on_creation_is_valid

This function returns boolean as to whether or not the ns_mirror constituent was placed correctly.
Currently this does not support multiple namespace mirrors (SnapDiff enabled).

=cut

sub ns_mirror_placement_on_creation_is_valid {
    my $self       = shift;
    my %parameters = @_;

    $Log->log(
        'Validation Starting: Namespace Mirror Constituent Placement on Creation'
    );

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    # Vserver setup
    my %availsizes     = %{ $self->_get_available_size() };
    my $data_aggr_list = $self->{'final_state'}->{'data_aggr_list'};
    my $num_dc_nodes   = $self->{'final_state'}->{'num_dc_nodes'};
    unless ( defined $num_dc_nodes ) {
        $num_dc_nodes = $self->get_number_of_dc_nodes(
            'node'           => $node,
            'availsizes'     => \%availsizes,
            'data-aggr-list' => $data_aggr_list
        );
    }

    # NS Mirror setup
    my $ns_mirror = $self->_get_constituent_by_role( 'role' => 'ns_mirror' );
    my $ns_mirror_aggr_name;
    my $ns_mirror_aggr;
    my $ns_mirror_aggr_node;

    if ( defined $ns_mirror ) {
        $ns_mirror_aggr_name
            = $self->_get_constituent_by_role( 'role' => 'ns_mirror' )
            ->aggregate;
        $ns_mirror_aggr = NACL::CS::Aggregate->fetch(
            command_interface => $node,
            filter            => { 'aggregate' => $ns_mirror_aggr_name }
        );
        $ns_mirror_aggr_node = @{ $ns_mirror_aggr->nodes }[0];
    }

    my $ns_mirror_aggr_list = $self->{'final_state'}->{'ns_mirror_aggr_list'};

    # Namespace setup
    my $ns_aggr_name
        = $self->_get_constituent_by_role( 'role' => 'namespace' )->aggregate;
    my $ns_aggr = NACL::CS::Aggregate->fetch(
        command_interface => $node,
        filter            => { 'aggregate' => $ns_aggr_name }
    );
    my $ns_aggr_node = @{ $ns_aggr->nodes }[0];

    if ( $num_dc_nodes > 1 ) {

        # Namespace mirror should be created
        if ( !$ns_mirror_aggr_name ) {
            $Log->log('Validation Failure: Namespace mirror was not created');
            return $FALSE;
        }
        else {
            $Log->log('Validation Success: Namespace mirror was created');
            if (@$ns_mirror_aggr_list) {
                if (!grep( /^$ns_mirror_aggr_name$/, @{$ns_mirror_aggr_list} )
                    )
                {
                    $Log->log(
                        'Validation Failure: Namespace mirror was not placed on an aggregate in the ns-mirror-aggr-list'
                    );
                    return $FALSE;
                }
                else {
                    $Log->log(
                        'Validation Success: Namespace mirror was placed on an aggregate in the ns-mirror-aggr-list'
                    );
                }
            }
            else {
                if ( !grep( /^$ns_mirror_aggr_name$/, keys %availsizes ) ) {
                    $Log->log(
                        'Validation Failure: Namespace mirror was not placed on an aggregate in the vserver aggr-list'
                    );
                    return $FALSE;
                }
                else {
                    $Log->log(
                        'Validation Success: Namespace mirror was placed on an aggregate in the vserver aggr-list'
                    );
                }
            }
            if ( $ns_aggr_node eq $ns_mirror_aggr_node ) {
                $Log->log(
                    'Validation Failure: Namespace mirror was placed on the same node as the namespace'
                );
                return $FALSE;
            }
            else {
                $Log->log(
                    'Validation Success: Namespace mirror was not placed on the same node as the namespace'
                );
            }
        }
    }
    else {

        # Namespace mirror shouldn't be created
        if ($ns_mirror_aggr_name) {
            $Log->log('Validation Failure: Namespace mirror was created');
            return $FALSE;
        }
        else {
            $Log->log('Validation Success: Namespace mirror was not created');
        }
    }

    $Log->log(
        'Validation Ending: Namespace Mirror Constituent Placement on Creation was Successful'
    );
    return $TRUE;
}

=head2 data_placement_on_creation_is_valid

This function returns boolean as to whether or not all data constituents were placed correctly.

=cut

sub data_placement_on_creation_is_valid {
    my $self       = shift;
    my %parameters = @_;

    $Log->log('Validation Starting: Data Constituent Placement on Creation');

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    # Vserver setup
    my %availsizes = %{ $self->_get_available_size() };

    # Data constituent setup
    my %data_aggrs     = ();
    my $data_aggr_list = $self->{'final_state'}->{'data_aggr_list'};
    my $num_dc_nodes   = $self->{'final_state'}->{'num_dc_nodes'};
    unless ( defined $num_dc_nodes ) {
        $num_dc_nodes = $self->get_number_of_dc_nodes(
            'node'           => $node,
            'availsizes'     => \%availsizes,
            'data-aggr-list' => $data_aggr_list
        );
    }
    my @dcs       = $self->_get_constituent_by_role( 'role' => 'data' );
    my @dcs_found = ();
    my %dc_nodes  = ();

    foreach my $dc (@dcs) {
        my $dc_aggr_name = $dc->aggregate;
        my $dc_aggr      = NACL::CS::Aggregate->fetch(
            command_interface => $node,
            filter            => { 'aggregate' => $dc_aggr_name }
        );

        if (@$data_aggr_list) {
            if ( !grep( /^$dc_aggr_name$/, @{$data_aggr_list} ) ) {
                $Log->log('Validation Failure: Data constituent ('
                        . $dc->volume
                        . ') was not placed on an aggregate in the data-aggr-list'
                );
                return $FALSE;
            }
            else {
                $Log->log('Validation Success: Data constituent ('
                        . $dc->volume
                        . ') was placed on an aggregate in the data-aggr-list'
                );
            }
        }
        else {
            if ( !grep( /^$dc_aggr_name$/, keys %availsizes ) ) {
                $Log->log('Validation Failure: Data constituent ('
                        . $dc->volume
                        . ') was not placed on an aggregate in the vserver aggr-list'
                );
                return $FALSE;
            }
            else {
                $Log->log('Validation Success: Data constituent ('
                        . $dc->volume
                        . ') was placed on an aggregate in the vserver aggr-list'
                );
            }
        }

        $Log->log( 'Validation Info: data constituent is ' . $dc->volume );
        my $aggr_node = @{ $dc_aggr->nodes }[0];
        $Log->log(
            'Validation Info: data constituent is on node ' . $aggr_node );
        $dc_nodes{$aggr_node} = 1;
        push( @dcs_found, $dc );
    }

    if ( $num_dc_nodes != keys %dc_nodes ) {
        $Log->log(
            'Validation Failure: data constituents do not span all required nodes'
        );
        return $FALSE;
    }
    else {
        $Log->log(
            'Validation Success: data constituents span all required nodes');
    }

    $Log->log(
        'Validation Ending: Data Constituent Placement on Creation was Successful'
    );
    return $TRUE;
}

=head2 get_initial_aggregate_state

This function provides a hash for available size lookup for initial conditions.

=cut

sub get_initial_aggregate_state {
    my $self       = shift;
    my %parameters = @_;

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    # Vserver setup
    my $vs         = $parameters{'vserver'};
    my @aggr_list  = @{ $vs->aggr_list };
    my %availsizes = ();
    $Log->log(
        'Grabbing inital aggregate state for vserver: ' . $vs->vserver );
    foreach my $aggr_name (@aggr_list) {
        my $aggr = NACL::CS::Aggregate->fetch(
            command_interface => $node,
            filter            => { 'aggregate' => $aggr_name },
        );
        $availsizes{$aggr_name} = $aggr->availsize;
    }
    return %availsizes;
}

=head2 namespace_size_on_creation_is_valid

This function returns boolean as to whether or not the namespace constituent was sized correctly.
NOTE: This will not handle snapdiff enabled.

=cut

sub namespace_size_on_creation_is_valid {
    my $self       = shift;
    my %parameters = @_;

    $Log->log(
        'Validation Starting: Namespace Constituent Sizing on Creation');

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    # Vserver setup
    my %availsizes     = %{ $self->_get_available_size() };
    my $data_aggr_list = $self->{'final_state'}->{'data_aggr_list'};
    my $num_dc_nodes   = $self->{'final_state'}->{'num_dc_nodes'};
    unless ( defined $num_dc_nodes ) {
        $num_dc_nodes = $self->get_number_of_dc_nodes(
            'node'           => $node,
            'availsizes'     => \%availsizes,
            'data-aggr-list' => $data_aggr_list
        );
    }

    # Namespace setup
    my $ns_aggr_name;
    my $ns_size;
    my $ns_aggr;
    my $ns = $self->_get_constituent_by_role( 'role' => 'namespace' );
    if ( defined $ns ) {
        $Log->log('Validation Success: Namespace was created');
        $ns_aggr_name = $ns->aggregate;
        $ns_aggr      = NACL::CS::Aggregate->fetch(
            command_interface => $node,
            filter            => { 'aggregate' => $ns_aggr_name }
        );
        $ns_size = $ns->size;
    }
    else {
        $Log->log('Validation Failure: Namespace was not created');
        return $FALSE;
    }

    my $max_ns_size
        = $self->{'final_state'}->{'max_namespace_constituent_size'};
    unless ( defined $max_ns_size ) {
        $max_ns_size = $DEFAULT_NS_MAX_SIZE;
    }
    $Log->log('Validation Info: Max Namespace Constituent Size is '
            . $max_ns_size
            . ' bytes' );

    # Hash of results from different sizing limites
    my %sizings = ();

    # Max usable space in the namespace aggregate,
    my $ns_usable_space = $self->max_usable_space(
        'node'            => $node,
        'aggregate'       => $ns_aggr->aggregate,
        'size'            => $ns_aggr->size,
        'volume'          => $ns,
        'availsize'       => $availsizes{$ns_aggr_name},
        'space-guarantee' => 'volume'
    );
    $sizings{'Maximum Usable Space of the Namespace Aggregate'}
        = $ns_usable_space;

    my $ns_mirror = $self->_get_constituent_by_role( 'role' => 'ns_mirror' );
    if ( defined $ns_mirror ) {

        # Max usable space in the namespace mirror aggregate
        my $ns_mirror_aggr = NACL::CS::Aggregate->fetch(
            command_interface => $node,
            filter            => { 'aggregate' => $ns_mirror->aggregate }
        );
        my $ns_mirror_usable_space = $self->max_usable_space(
            'node'            => $node,
            'aggregate'       => $ns_mirror_aggr->aggregate,
            'size'            => $ns_mirror_aggr->size,
            'volume'          => $ns_mirror,
            'availsize'       => $availsizes{ $ns_mirror_aggr->aggregate },
            'space-guarantee' => 'volume'
        );
        $sizings{'Maximum Usable Space of the Namespace Mirror Aggregate'}
            = $ns_mirror_usable_space;
    }

    # Percentage of Infinitevol
    my $specified_infinitevol_size
        = $self->{'final_state'}->{'size_specified'};
    my $infinitevol_size;
    my $size_factor = $self->{'final_state'}->{'enable_object_store'} eq 'false' ? 0.75 : 0.5;
    if ( defined $specified_infinitevol_size ) {
        $Log->log('Validation Info: Specified InfiniteVol Size is '
                . $specified_infinitevol_size
                . ' bytes' );
        $Log->log(
            'Validation Info: Using Specified InfiniteVol Size for Percentage Calculation'
        );
        $infinitevol_size = $specified_infinitevol_size;
    }
    else {
        my $default_infinitevol_size
            = (   $num_dc_nodes
                * $self->{'helpers'}
                ->minimum_data_constituent_size( node => $node ) ) / $size_factor;
        $Log->log('Validation Info: Default InfiniteVol Size is '
                . $default_infinitevol_size
                . ' bytes' );
        $Log->log(
            'Validation Info: Using Default InfiniteVol Size for Percentage Calculation'
        );
        $infinitevol_size = $default_infinitevol_size;
    }

    if ( $num_dc_nodes > 1 ) {
        $sizings{'12.5% of the InfiniteVol size'}
            = nhimult( $WAFL_BLOCK, 0.125 * $infinitevol_size );
    }
    else {
        $sizings{'25% of the InfiniteVol size'}
            = nhimult( $WAFL_BLOCK, 0.25 * $infinitevol_size );
    }

    # Max namespace constituent size
    $sizings{'Maximum Namespace Constituent Size'} = $max_ns_size;

    # Choose smallest of the methods
    my $method_chosen
        = ( sort { $sizings{$a} <=> $sizings{$b} } keys %sizings )[0];

    $Log->log( 'Validation Info: Smallest size method is ' . $method_chosen );
    $Log->log('Validation Info: Size from that method is '
            . $sizings{$method_chosen}
            . ' bytes' );

    if ( $sizings{$method_chosen} < $MIN_NS_SIZE ) {
        $Log->log(
            'Validation Info: Namespace limit chosen to be Minimum FlexVol Size'
        );
        $Log->log('Validation Info: Namespace size should be '
                . $MIN_NS_SIZE
                . ' bytes' );
        $Log->log(
            'Validation Info: Namespace size was ' . $ns_size . ' bytes' );

        my $difference = abs( $MIN_NS_SIZE - $ns_size );
        $Log->log(
            'Validation Info: Difference between actual and expected is '
                . $difference
                . ' bytes' );

        if ( $difference > 0 ) {
            $Log->log(
                'Validation Failure: Namespace did not match Minimum FlexVol Size'
            );
            return $FALSE;
        }
        else {
            $Log->log(
                'Validation Success: Namespace matched Minimum FlexVol Size'
            );
        }
    }
    else {
        $Log->log( 'Validation Info: Namespace limit chosen to be '
                . $method_chosen );
        $Log->log('Validation Info: Namespace size should be '
                . $sizings{$method_chosen}
                . ' bytes' );
        $Log->log(
            'Validation Info: Namespace size was ' . $ns_size . ' bytes' );

        my $difference = abs( $sizings{$method_chosen} - $ns_size );
        $Log->log(
            'Validation Info: Difference between actual and expected is '
                . $difference
                . ' bytes' );

        if ( $difference > $DEFAULT_TOLERANCE ) {
            $Log->log( 'Validation Failure: Namespace did not match '
                    . $method_chosen );
            return $FALSE;
        }
        else {
            $Log->log(
                'Validation Success: Namespace matched ' . $method_chosen );
        }
    }

    $Log->log(
        'Validation Ending: Namespace Constituent Sizing on Creation was Successful'
    );
    return $TRUE;
}

=head2 ns_mirror_size_on_creation_is_valid

This function returns boolean as to whether or not the namespace mirror constituent was sized correctly.
Only run if a namespace mirror should be created.

=cut

sub ns_mirror_size_on_creation_is_valid {
    my $self       = shift;
    my %parameters = @_;

    $Log->log(
        'Validation Starting: NS Mirror Constituent Sizing on Creation');

    # Vserver setup
    my %availsizes     = %{ $self->_get_available_size() };
    my $data_aggr_list = $self->{'final_state'}->{'data_aggr_list'};
    my $num_dc_nodes   = $self->{'final_state'}->{'num_dc_nodes'};
    unless ( defined $num_dc_nodes ) {
        $num_dc_nodes = $self->get_number_of_dc_nodes(
            'node'           => $parameters{'node'},
            'availsizes'     => \%availsizes,
            'data-aggr-list' => $data_aggr_list
        );
    }
    if ( $num_dc_nodes == 1 ) {
        $Log->log(
            'Validation Info: NS Mirror Constituent is not created on single node'
        );
        $Log->log(
            'Validation Ending: NS Mirror Constituent Sizing on Creation was Successful'
        );
        return $TRUE;
    }

    # NS and mirror setup
    my $ns = $self->_get_constituent_by_role( 'role' => 'namespace' );
    unless ( defined $ns ) {
        $Log->log('Validation Failure: Namespace was not created');
        return $FALSE;
    }

    my $ns_mirror = $self->_get_constituent_by_role( 'role' => 'ns_mirror' );
    unless ( defined $ns_mirror ) {
        $Log->log('Validation Failure: NS Mirror was not created');
        return $FALSE;
    }

    my $ns_mirror_size = $ns_mirror->size;
    my $ns_size        = $ns->size;

    $Log->log( 'Validation Info: Namespace size is ' . $ns_size . ' bytes' );
    $Log->log(
        'Validation Info: NS mirror size is ' . $ns_mirror_size . ' bytes' );

    if ( $ns_size != $ns_mirror_size ) {
        $Log->log(
            'Validation Failure: NS Mirror was not the same size as the Namespace'
        );
        return $FALSE;
    }
    else {
        $Log->log(
            'Validation Success: NS Mirror was the same size as the Namespace'
        );
    }

    $Log->log(
        'Validation Ending: NS Mirror Constituent Sizing on Creation was Successful'
    );
    return $TRUE;
}

=head2 dcs_size_on_creation_is_valid

This function returns boolean as to whether or not the  dcs were sized correctly.

=cut

sub dcs_size_on_creation_is_valid {
    my $self       = shift;
    my %parameters = @_;

    $Log->log('Validation Starting: Data Constituents Sizing on Creation');

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    # Vserver setup
    my %availsizes = %{ $self->_get_available_size() };

    # Data constituent setup
    my @dcs = $self->_get_constituent_by_role( 'role' => 'data' );
    my $max_dc_size = $self->{'final_state'}->{'max_data_constituent_size'};
    my $cluster_max_vol_size = $self->_get_cluster_max_vol_size(
        node       => $node,
        availsizes => \%availsizes
    );
    unless ( defined $max_dc_size
        && $max_dc_size < $cluster_max_vol_size )
    {
        $max_dc_size = $cluster_max_vol_size;
    }
    $Log->log('Validation Info: Max data constituent size is '
            . $max_dc_size
            . ' bytes' );

    # Do node sums and validate against max size for each constituent
    my %nodes = ();

    foreach my $dc (@dcs) {
        my $dc_aggr_name = $dc->aggregate;
        my $dc_aggr      = NACL::CS::Aggregate->fetch(
            command_interface => $node,
            filter            => { 'aggregate' => $dc_aggr_name }
        );
        my $dc_aggr_node = @{ $dc_aggr->nodes }[0];

        my $dc_size = $dc->size;
        if ( $dc_size > $max_dc_size ) {
            $Log->log('Validation Info: Data constituent '
                    . $dc->volume
                    . ' has total size '
                    . $dc_size
                    . ' bytes' );
            $Log->log('Validation Failure: Data constituent '
                    . $dc->volume
                    . ' was larger than maximum data constituent size '
                    . $max_dc_size
                    . ' bytes' );
            return $FALSE;
        }

        if ( $dc_size < $self->{'helpers'}
            ->minimum_data_constituent_size( node => $node ) )
        {
            $Log->log('Validation Info: Data constituent '
                    . $dc->volume
                    . ' has total size '
                    . $dc_size
                    . ' bytes' );
            $Log->log('Validation Failure: Data constituent '
                    . $dc->volume
                    . ' was smaller than minimum data constituent size '
                    . $self->{'helpers'}
                    ->minimum_data_constituent_size( node => $node )
                    . ' bytes' );
            return $FALSE;
        }

        if ( exists $nodes{$dc_aggr_node} ) {
            $nodes{$dc_aggr_node} += $dc_size;
        }
        else {
            $nodes{$dc_aggr_node} = $dc_size;
        }
    }

# Difference between largest node and smallest node total sizes should be within tolerance
    my $largest_node
        = ( reverse sort { $nodes{$a} <=> $nodes{$b} } keys %nodes )[0];
    my $smallest_node = ( sort { $nodes{$a} <=> $nodes{$b} } keys %nodes )[0];
    my $tolerance
        = $self->{'helpers'}->minimum_data_constituent_size( node => $node );
    my $difference = abs( $nodes{$largest_node} - $nodes{$smallest_node} );

    $Log->log('Validation Info: Largest node '
            . $largest_node
            . ' has total size '
            . $nodes{$largest_node}
            . ' bytes' );
    $Log->log('Validation Info: Smallest node '
            . $smallest_node
            . ' has total size '
            . $nodes{$smallest_node}
            . ' bytes' );
    $Log->log('Validation Info: Difference between largest and smallest is '
            . $difference
            . ' bytes' );
    $Log->log( 'Validation Info: Tolerance is +/- ' . $tolerance . ' bytes' );

    if ( $difference > $tolerance ) {
        $Log->log(
            'Validation Failure: Data Constituents sizes were not balanced across nodes'
        );
        return $FALSE;
    }
    else {
        $Log->log(
            'Validation Success: Data Constituents sizes were balanced across nodes'
        );
    }

    $Log->log(
        'Validation Ending: Data Constituents Sizing on Creation was Successful'
    );
    return $TRUE;
}

=head2 infinitevol_size_on_creation_is_valid

This function returns boolean as to whether or not the infinitevol was sized correctly.

=cut

sub infinitevol_size_on_creation_is_valid {
    my $self       = shift;
    my %parameters = @_;

    $Log->log('Validation Starting: InfiniteVol Sizing on Creation');

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    # Vserver Setup
    my %availsizes     = %{ $self->_get_available_size() };
    my $data_aggr_list = $self->{'final_state'}->{'data_aggr_list'};
    unless ( defined $data_aggr_list ) {
        $data_aggr_list = [];
    }
    my $num_dc_nodes = $self->{'final_state'}->{'num_dc_nodes'};
    unless ( defined $num_dc_nodes ) {
        $num_dc_nodes = $self->get_number_of_dc_nodes(
            'node'           => $node,
            'availsizes'     => \%availsizes,
            'data-aggr-list' => $data_aggr_list
        );
    }

    # InfiniteVol setup
    my $infinitevol = $self->{'final_state'}->{'infinitevol'};
    unless ( defined $infinitevol ) {
        $Log->log('Validation Failure: InfiniteVol was not created');
        return $FALSE;
    }

    my $infinitevol_size = $infinitevol->size;
    $Log->log('Validation Info: InfiniteVol size is '
            . $infinitevol_size
            . ' bytes' );

    my @constituents = $self->_get_constituent_by_role( 'role' => 'all' );

   # Use of anything other than raw values is not supported for size_specified
    my $size_specified = $self->{'final_state'}->{'size_specified'};

    # Size validation
    if ( defined $size_specified ) {
        $Log->log('Validation Info: InfiniteVol size specified is '
                . $size_specified
                . ' bytes' );
        if ( $infinitevol_size < $size_specified ) {
            my $difference = abs( $infinitevol_size - $size_specified );
            $Log->log(
                'Validation Info: Resulting size was less than the specified size by '
                    . $difference
                    . ' bytes' );
            $Log->log(
                'Validation Failure: InfiniteVol size was less than the specified size'
            );
            return $FALSE;
        }
        else {
            $Log->log(
                'Validation Success: InfiniteVol size was not less than the specified size'
            );
            my $tolerance
                = $num_dc_nodes
                * $self->{'helpers'}
                ->minimum_data_constituent_size( node => $node );
            $Log->log('Validation Info: InfiniteVol size tolerance is '
                    . $tolerance
                    . ' bytes' );

            my $difference = abs( $infinitevol_size - $size_specified );
            $Log->log(
                'Validation Info: Difference between actual and expected is '
                    . $difference
                    . ' bytes' );

            if ( $difference > $tolerance ) {
                $Log->log(
                    'Validation Failure: InfiniteVol size was not within tolerance'
                );
                return $FALSE;
            }
            else {
                $Log->log(
                    'Validation Success: InfiniteVol size was within tolerance'
                );
            }
        }
    }
    else {

# InfiniteVol is default size
# This is very complicated to do generally for what scenarios we create in test
# For now this will not have any validation
        $Log->log('Validation Info: InfiniteVol size is not specified.');
    }

    my $sum_ = $self->_sum_of_constituents(@constituents);
    $Log->log(
        'Validation Info: Sum of  constituents is ' . $sum_ . ' bytes' );

    if ( $infinitevol_size != $sum_ ) {
        $Log->log(
            'Validation Failure: InfiniteVol size was not the sum of all  constituents'
        );
        return $FALSE;
    }
    else {
        $Log->log(
            'Validation Success: InfiniteVol size was the sum of all  constituents'
        );
    }

    $Log->log(
        'Validation Ending: InfiniteVol Sizing on Creation was Successful');
    return $TRUE;
}

=head2 get_number_of_dc_nodes

This function takes in the availsizes hash and the data-aggr-list and gives the number of nodes the dcs span.

=cut

sub get_number_of_dc_nodes {
    my $self       = shift;
    my %parameters = @_;

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    # Vserver setup
    my %availsizes = %{ $parameters{'availsizes'} };

    # Volume setup
    my @data_aggr_list = ();
    if ( defined $parameters{'data-aggr-list'} ) {
        $Log->log('Validation Info: data-aggr-list was specified');
        @data_aggr_list = @{ $parameters{'data-aggr-list'} };
        $Log->log(@data_aggr_list);
    }

    if ( @data_aggr_list > 0 ) {
        $Log->log(
            'Validation Info: data-aggr-list is being used for num_dc_nodes'
        );
        my %data_aggr_list_hash = ();
        foreach my $aggr (@data_aggr_list) {
            $Log->log(
                'Validation Info: data-aggr-list aggregate is ' . $aggr );
            $data_aggr_list_hash{$aggr} = 1;
        }
        return $self->_get_number_of_nodes(
            'node'       => $node,
            'availsizes' => \%data_aggr_list_hash
        );
    }
    else {
        $Log->log(
            'Validation Info: vserver aggr-list is being used for num_dc_nodes'
        );
        return $self->_get_number_of_nodes(
            'node'       => $node,
            'availsizes' => \%availsizes
        );
    }
}

=head2 namespace_size_after_expansion_is_valid

This function returns boolean as to whether or not the namespace constituent was sized correctly.

=cut

sub namespace_size_after_expansion_is_valid {
    my $self       = shift;
    my %parameters = @_;

    $Log->log(
        'Validation Starting: Namespace Constituent Sizing on Expansion');

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    my %availsizes = %{ $self->_get_available_size() };

    # Namespace setup
    my $ns_aggr_name;
    my $ns_size;
    my $ns_aggr;
    my $ns = $self->_get_constituent_by_role( 'role' => 'namespace' );
    if ( defined $ns ) {
        $Log->log('Validation Success: Namespace was present');
        $ns_aggr_name = $ns->aggregate;
        $ns_aggr      = NACL::CS::Aggregate->fetch(
            command_interface => $node,
            filter            => { 'aggregate' => $ns_aggr_name }
        );
        $ns_size = $ns->size;
        $Log->log('Validation Info: Current Namespace size is '
                . $ns_size
                . ' bytes' );
    }
    else {
        $Log->log('Validation Failure: Namespace was not present');
        return $FALSE;
    }
    my $original_ns = $self->_get_constituent_by_role(
        'role'  => 'namespace',
        'state' => 'initial'
    );
    my $original_ns_size;

    if ( defined $original_ns ) {
        $original_ns_size = $original_ns->size;
        $Log->log('Validation Info: Original Namespace size was '
                . $original_ns_size
                . ' bytes' );
    }

    my $max_ns_size
        = $self->{'final_state'}->{'max_namespace_constituent_size'};
    unless ( defined $max_ns_size ) {
        $max_ns_size = $DEFAULT_NS_MAX_SIZE;
    }
    $Log->log('Validation Info: Max Namespace Constituent Size is '
            . $max_ns_size
            . ' bytes' );

    # Hash of results from different sizing limites
    my %sizings = ();

    # NS Mirror setup
    my $ns_mirror = $self->_get_constituent_by_role( 'role' => 'ns_mirror' );
    my $ns_mirror_size;
    my $ns_mirror_aggr;
    if ( defined $ns_mirror ) {
        $ns_mirror_size = $ns_mirror->size;
        $Log->log('Validation Info: Current Namespace Mirror size is '
                . $ns_mirror_size
                . ' bytes' );

        # Max usable space in the namespace mirror aggregate
        $ns_mirror_aggr = NACL::CS::Aggregate->fetch(
            command_interface => $node,
            filter            => { 'aggregate' => $ns_mirror->aggregate }
        );
        my $ns_mirror_usable_space = $self->max_usable_space(
            'node'            => $node,
            'aggregate'       => $ns_mirror_aggr->aggregate,
            'size'            => $ns_mirror_aggr->size,
            'volume'          => $ns_mirror,
            'availsize'       => $availsizes{ $ns_mirror_aggr->aggregate },
            'space-guarantee' => 'volume',
            'expansion'       => $TRUE
        );
        $sizings{'Maximum Usable Space of the Namespace Mirror Aggregate'}
            = $ns_mirror_usable_space;
        $Log->log(
            'Validation Info: Maximum Usable Space of the Namespace Mirror Aggregate is '
                . $ns_mirror_usable_space
                . ' bytes' );
    }

    # Percentage of Infinitevol
    my $specified_infinitevol_size
        = $self->{'final_state'}->{'size_specified'};
    $Log->log('Validation Info: Specified InfiniteVol Size is '
            . $specified_infinitevol_size
            . ' bytes' );

    if ( defined $ns_mirror ) {
        $sizings{'12.5% of the InfiniteVol size'}
            = nhimult( $WAFL_BLOCK, 0.125 * $specified_infinitevol_size );
    }
    else {
        $sizings{'25% of the InfiniteVol size'}
            = nhimult( $WAFL_BLOCK, 0.25 * $specified_infinitevol_size );
    }

    # Max namespace constituent size
    $sizings{'Maximum Namespace Constituent Size'} = $max_ns_size;

    # Max usable aggregate size
    my $ns_usable_space = $self->max_usable_space(
        'node'            => $node,
        'aggregate'       => $ns_aggr->aggregate,
        'size'            => $ns_aggr->size,
        'volume'          => $ns,
        'availsize'       => $availsizes{ $ns_aggr->aggregate },
        'space-guarantee' => 'volume',
        'expansion'       => $TRUE
    );
    $sizings{'Maximum Usable Space of the Namespace Aggregate'}
        = $ns_usable_space;
    $Log->log(
        'Validation Info: Maximum Usable Space of the Namespace Aggregate is '
            . $ns_usable_space
            . ' bytes' );

    # Choose smallest of the methods
    my $method_chosen
        = ( sort { $sizings{$a} <=> $sizings{$b} } keys %sizings )[0];

    $Log->log( 'Validation Info: Smallest size method is ' . $method_chosen );
    $Log->log('Validation Info: Size from that method is '
            . $sizings{$method_chosen}
            . ' bytes' );

# Max usable space is rather complicated, this will err on the side of caution to prevent sporadic failures
# If the current size of the ns is less than the smallest size method and is greater than minimum size
# And the aggregate is full as best as it can be determined then the validation will assume that is the
# max usable space of the aggregate, this is not a guarantee that it did the right thing
    if ((   $self->aggregate_is_full(
                'aggregate' => $ns_aggr,
                node        => $node
            )
            && $ns_size < ( $sizings{$method_chosen} - $DEFAULT_TOLERANCE )
            && $ns_size >= $MIN_NS_SIZE
        )
        || (defined $ns_mirror
            && $self->aggregate_is_full(
                'aggregate' => $ns_mirror_aggr,
                node        => $node
            )
            && $ns_mirror_size
            < ( $sizings{$method_chosen} - $DEFAULT_TOLERANCE )
            && $ns_mirror_size >= $MIN_NS_SIZE
        )
        )
    {
        $Log->log(
            'Validation Info: Namespace limit chosen to be Max Usable Space of the Namespace Aggregate'
        );
        $Log->log(
            'Validation Info: Sizing is not validated in this case as it cannot be determined reliably'
        );

        # Exit as successful before other validation
        $Log->log(
            'Validation Ending: Namespace Constituent Sizing on Expansion was Successful'
        );
        return $TRUE;
    }

    if ( $sizings{$method_chosen} < $original_ns_size ) {
        $Log->log(
            'Validation Info: Desired size for expansion would shrink the namespace, validating that the size remained constant'
        );
        if ( $ns_size == $original_ns_size ) {
            $Log->log('Validation Success: Namespace size was not changed');
        }
        else {
            $Log->log('Validation Failure: Namespace was changed');
            return $FALSE;
        }
    }
    else {
        $Log->log(
            'Validation Info: Desired size for expansion would not shrink the namespace, validating the new size'
        );
        if ( $sizings{$method_chosen} < $MIN_NS_SIZE ) {
            $Log->log(
                'Validation Info: Namespace limit chosen to be Minimum FlexVol Size'
            );
            $Log->log('Validation Info: Namespace size should be '
                    . $MIN_NS_SIZE
                    . ' bytes' );

            my $difference = abs( $MIN_NS_SIZE - $ns_size );
            $Log->log(
                'Validation Info: Difference between actual and expected is '
                    . $difference
                    . ' bytes' );

            if ( $difference > 0 ) {
                $Log->log(
                    'Validation Failure: Namespace did not match Minimum FlexVol Size'
                );
                return $FALSE;
            }
            else {
                $Log->log(
                    'Validation Success: Namespace matched Minimum FlexVol Size'
                );
            }
        }
        else {
            $Log->log( 'Validation Info: Namespace limit chosen to be '
                    . $method_chosen );
            $Log->log('Validation Info: Namespace size should be '
                    . $sizings{$method_chosen}
                    . ' bytes' );

            my $difference = abs( $sizings{$method_chosen} - $ns_size );
            $Log->log(
                'Validation Info: Difference between actual and expected is '
                    . $difference
                    . ' bytes' );

            if ( $difference > $DEFAULT_TOLERANCE ) {
                $Log->log( 'Validation Failure: Namespace did not match '
                        . $method_chosen );
                return $FALSE;
            }
            else {
                $Log->log( 'Validation Success: Namespace matched '
                        . $method_chosen );
            }
        }
    }
    $Log->log(
        'Validation Ending: Namespace Constituent Sizing on Expansion was Successful'
    );
    return $TRUE;
}

=head2 infinitevol_size_after_expansion_is_valid

This function returns boolean as to whether or not the infinitevol was sized correctly.

=cut

sub infinitevol_size_after_expansion_is_valid {
    my $self       = shift;
    my %parameters = @_;

    $Log->log('Validation Starting: InfiniteVol Sizing on Expansion');

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    # InfiniteVol setup
    my $infinitevol_size = $self->{'final_state'}->{'infinitevol'}->size;
    $Log->log('Validation Info: InfiniteVol size is '
            . $infinitevol_size
            . ' bytes' );

    my @constituents = $self->_get_constituent_by_role( 'role' => 'all' );
    my @dcs          = $self->_get_constituent_by_role( 'role' => 'data' );
    my %dc_nodes     = ();

    foreach my $dc (@dcs) {
        my $dc_aggr_name = $dc->aggregate;
        my $dc_aggr      = NACL::CS::Aggregate->fetch(
            command_interface => $node,
            filter            => { 'aggregate' => $dc_aggr_name }
        );

        $Log->log( 'Validation Info: data constituent is ' . $dc->volume );
        my $aggr_node = @{ $dc_aggr->nodes }[0];
        $Log->log(
            'Validation Info: data constituent is on node ' . $aggr_node );
        $dc_nodes{$aggr_node} = 1;
    }

    my $num_dc_nodes = scalar keys %dc_nodes;

   # Use of anything other than raw values is not supported for size_specified
    my $size_specified = $self->{'final_state'}->{'size_specified'};

    $Log->log('Validation Info: InfiniteVol size specified is '
            . $size_specified
            . ' bytes' );
    if ( $infinitevol_size < $size_specified ) {
        my $difference = abs( $infinitevol_size - $size_specified );
        $Log->log(
            'Validation Info: Resulting size was less than the specified size by '
                . $difference
                . ' bytes' );
        $Log->log(
            'Validation Failure: InfiniteVol size was less than the specified size'
        );
        return $FALSE;
    }
    else {
        $Log->log(
            'Validation Success: InfiniteVol size was not less than the specified size'
        );
        my $tolerance
            = $num_dc_nodes
            * $self->{'helpers'}
            ->minimum_data_constituent_size( node => $node );
        $Log->log('Validation Info: InfiniteVol size tolerance is '
                . $tolerance
                . ' bytes' );

        my $difference = abs( $infinitevol_size - $size_specified );
        $Log->log(
            'Validation Info: Difference between actual and expected is '
                . $difference
                . ' bytes' );

        if ( $difference > $tolerance ) {
            $Log->log(
                'Validation Failure: InfiniteVol size was not within tolerance'
            );
            return $FALSE;
        }
        else {
            $Log->log(
                'Validation Success: InfiniteVol size was within tolerance');
        }
    }

    # InfiniteVol size excludes the reserve constituent
    my $sum_of_constituents = $self->_sum_of_constituents(@constituents);
    $Log->log('Validation Info: Sum of  constituents is '
            . $sum_of_constituents
            . ' bytes' );

    if ( $infinitevol_size != $sum_of_constituents ) {
        $Log->log(
            'Validation Failure: InfiniteVol size was not the sum of all constituents'
        );
        return $FALSE;
    }
    else {
        $Log->log(
            'Validation Success: InfiniteVol size was the sum of all constituents'
        );
    }

    $Log->log(
        'Validation Ending: InfiniteVol Sizing on Expansion was Successful');
    return $TRUE;
}

=head2 aggregate_is_full


NOTE: The tolerance is sufficiently large on this that it may say its full when there is still some space to go on really large aggregates

=cut

# TODO: Somehow handle thin provisioning (reserve), this should work for namespace
sub aggregate_is_full {
    my $self       = shift;
    my %parameters = @_;

    my $aggr                = $parameters{'aggregate'};
    my $availsize           = $aggr->availsize;
    my $size                = $aggr->size;
    my $wafl_dedupe_reserve = nhimult( $WAFL_BLOCK, 0.09 * $size );

    my @vols = NACL::CS::Volume->fetch(
        command_interface => $parameters{'node'},
        filter            => { aggregate => $aggr->aggregate },
        requested_fields  => [ 'space-guarantee', 'size' ],
        allow_empty       => 1
    );

    my $footprint = 0;
    foreach my $vol (@vols) {
        if ( $vol->space_guarantee eq 'none' ) {
            $footprint += $vol->size / $THIN_PROVISIONING_RATIO;
        }
    }

    $Log->log(
        "Validation Info: Available size: $availsize, Wafl and Dedupe Reserve: $wafl_dedupe_reserve, Volume footprint: $footprint"
    );

    if ( abs( $availsize - $wafl_dedupe_reserve - $footprint ) > 0.01 * $size
        && $availsize > $wafl_dedupe_reserve )
    {
        return $FALSE;
    }
    else {
        return $TRUE;
    }
}

###############################################

=head2 _get_aggrs_sizes

Helper function to get aggregates sizes and available sizes

=cut

sub _get_aggrs_sizes {
    my $self           = shift;
    my %parameters     = @_;
    my %aggr_size      = ();
    my %aggr_availsize = ();

    my @aggrs = NACL::CS::Aggregate->fetch(
        command_interface => $parameters{'node'} );

    foreach my $aggr (@aggrs) {
        $aggr_size{ $aggr->aggregate }      = $aggr->size;
        $aggr_availsize{ $aggr->aggregate } = $aggr->availsize;
    }
    return { size => \%aggr_size, 'availsize' => \%aggr_availsize };
}

=head2 _get_volume_info_by_aggr

Helper function to get informaton about volumes by aggregates

=cut

sub _get_volume_info_by_aggr {
    my $self           = shift;
    my %parameters     = @_;
    my %volume_by_aggr = ();

    my @volumes = NACL::CS::Volume->fetch(
        command_interface => $parameters{'node'},
        filter            => { 'volume-style' => '!infinitevol' }
    );

    my @aggrs = NACL::CS::Aggregate->fetch(
        command_interface => $parameters{'node'} );

    foreach my $aggr (@aggrs) {
        $volume_by_aggr{ $aggr->aggregate } = []
            unless defined $volume_by_aggr{ $aggr->aggregate };
    }

    foreach my $vol (@volumes) {
        $Log->log(
            'Validation Info: _get_volume_info_by_aggr aggregate, volume name '
                . $vol->aggregate . ', '
                . $vol->volume );
        push( @{ $volume_by_aggr{ $vol->aggregate } }, $vol );
    }

    return \%volume_by_aggr;
}

=head2 _get_volume_footprint_info_by_aggr

Helper function to get informaton about volumes footprint by aggregates

=cut

sub _get_volume_footprint_info_by_aggr {
    my $self                     = shift;
    my %parameters               = @_;
    my %volume_footprint_by_aggr = ();

    my @volumes = NACL::CS::Volume->fetch(
        command_interface => $parameters{'node'},
        filter            => { 'volume-style' => '!infinitevol' }
    );

    my @aggrs = NACL::CS::Aggregate->fetch(
        command_interface => $parameters{'node'} );

    foreach my $aggr (@aggrs) {
        $volume_footprint_by_aggr{ $aggr->aggregate } = []
            unless defined $volume_footprint_by_aggr{ $aggr->aggregate };
    }

    foreach my $vol (@volumes) {
        my $footprint = NACL::CS::VolumeShowFootprint->fetch(
            command_interface => $parameters{'node'},
            filter            => { 'volume' => $vol->volume },
        );
        push( @{ $volume_footprint_by_aggr{ $vol->aggregate } }, $footprint );
    }

    return \%volume_footprint_by_aggr;
}

=head2 _get_volume_space_info_for_aggregate

This function finds the dedupe metadata footprint and volume footprint information on the aggregate provided

=cut

sub _get_volume_space_info_for_aggregate {
    my $self                      = shift;
    my %parameters                = @_;
    my $dedupe_metadata_footprint = 0;
    my $total_volume_footprint    = 0;
    my $total_volume_reserve      = 0;

    my @vols = @{ $self->{'final_state'}->{'volume_info_by_aggr'}
            ->{ $parameters{'aggregate'} } };

    my @footprints
        = @{ $self->{'final_state'}->{'volume_footprint_info_by_aggr'}
            ->{ $parameters{'aggregate'} } };

    foreach my $vol (@vols) {
        my @fps = grep { $_->volume eq $vol->volume } @footprints;
        my $cs = $fps[0];

        my $metafile_footprint
            = $cs->dedupe_metafiles_footprint() =~ /^\d+$/
            ? $cs->dedupe_metafiles_footprint()
            : 0;
        my $metafile_temp_footprint
            = $cs->dedupe_metafiles_temporary_footprint() =~ /^\d+$/
            ? $cs->dedupe_metafiles_temporary_footprint()
            : 0;
        my $volume_footprint = 0;
        if ( $cs->total_footprint() =~ /^\d+$/ ) {
            $volume_footprint += $cs->total_footprint();
            $volume_footprint
                = $volume_footprint
                - $metafile_footprint
                - $metafile_temp_footprint;
        }

        if ((   (      $vol->vserver ne $parameters{'volume'}->vserver
                    or $vol->is_constituent eq 'false'
                )
                and !$parameters{'expansion'}
            )
            or $parameters{'expansion'}
            )
        {
            $dedupe_metadata_footprint
                = $dedupe_metadata_footprint
                + $metafile_footprint
                + $metafile_temp_footprint;

            my $volume_reserve = 0;
            my $volume_size    = $vol->size;
            if ( $vol->space_guarantee eq 'volume' ) {
                $volume_reserve
                    = $volume_size / $THICK_PROVISIONING_RATIO
                    > $volume_footprint
                    ? $volume_size / $THICK_PROVISIONING_RATIO
                    - $volume_footprint
                    : 0;
            }
            elsif ( $vol->space_guarantee eq 'none' ) {
                $volume_reserve = $volume_size / $THIN_PROVISIONING_RATIO;
            }
            $total_volume_reserve += $volume_reserve;
        }

        if (    $parameters{'expansion'}
            and $vol->volume eq $parameters{'volume'}->volume )
        {
            my $new_volume_footprint = 0;
            my $volume_size          = $vol->size;
            if ( $vol->space_guarantee eq 'volume' ) {
                $new_volume_footprint
                    = $volume_size / $THICK_PROVISIONING_RATIO
                    > $volume_footprint
                    ? $volume_size / $THICK_PROVISIONING_RATIO
                    : $volume_footprint;
            }
            elsif ( $vol->space_guarantee eq 'one' ) {
                $new_volume_footprint
                    = $volume_size / $THIN_PROVISIONING_RATIO;
            }
            $total_volume_footprint += $new_volume_footprint;
        }
    }

    my %info = ();
    $info{'dedupe_metadata_footprint'} = $dedupe_metadata_footprint;
    $info{'total_volume_footprint'}    = $total_volume_footprint;
    $info{'volume_reserve'}            = $total_volume_reserve;

    return \%info;
}

=head2 _get_cluster_max_vol_size

This function finds the max vol size for a cluster based on the availsizes hash.

=cut

sub _get_cluster_max_vol_size {
    my $self       = shift;
    my %parameters = @_;

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    # Vserver setup
    my %availsizes = %{ $parameters{'availsizes'} };
    my %h          = ();

    foreach my $aggr_name ( keys %availsizes ) {
        my $aggr = NACL::CS::Aggregate->fetch(
            command_interface => $node,
            filter            => { 'aggregate' => $aggr_name }
        );
        my $aggr_node = @{ $aggr->nodes }[0];
        $h{$aggr_node} = 1;
    }
    my @nodes = keys %h;
    my $max_vol_size = min map { $_->max_vol_size } map {
        NACL::CS::Node->fetch(
            command_interface => $node,
            filter            => { node => $_ },
            requested_fields  => [qw(max-vol-size)],
            )
    } @nodes;

    return $max_vol_size;
}

=head2 _get_number_of_nodes

This function takes in the availsizes hash and gives the number of node the aggregates span.

=cut

sub _get_number_of_nodes {
    my $self       = shift;
    my %parameters = @_;

    # Node setup
    my $node = $parameters{'node'};
    unless ( defined $node ) {
        $node = $self->{'node'};
    }

    # Vserver setup
    my %availsizes = %{ $parameters{'availsizes'} };
    my %h          = ();

    foreach my $aggr_name ( keys %availsizes ) {
        my $aggr = NACL::CS::Aggregate->fetch(
            command_interface => $node,
            filter            => { 'aggregate' => $aggr_name }
        );
        my $aggr_node = @{ $aggr->nodes }[0];
        $h{$aggr_node} = 1;
    }
    return scalar( keys %h );
}

=head2 _sum_of_constituents

This function gives the sum of the sizes from an array of constituents.

=cut

sub _sum_of_constituents {
    my $self         = shift;
    my @constituents = @_;
    my $total        = 0;

    foreach my $constituent (@constituents) {
        $total += $constituent->size;
        $Log->log('Constituent volume '
                . $constituent->volume . ' is '
                . $constituent->size
                . ' bytes' );
    }

    return $total;
}

=head2 _get_infinitevol_max_size

This function gets the max InfiniteVol size

=cut

sub _get_infinitevol_max_size {
    my %params = @_;
    my %limit_options;

    my %create_args = $params{'create_args'};

    $limit_options{'space-guarantee'} = $create_args{'space-guarantee'}
        if ( defined $create_args{'space-guarantee'} );
    $limit_options{'namespace-aggregate'}
        = $create_args{'namespace-aggregate'}
        if ( defined $create_args{'namespace-aggregate'} );
    $limit_options{'max-namespace-constituent-size'}
        = $create_args{'max-namespace-constituent-size'}
        if ( defined $create_args{'max-namespace-constituent-size'} );
    $limit_options{'data-aggr-list'} = $create_args{'data-aggr-list'}
        if ( defined $create_args{'data-aggr-list'} );
    $limit_options{'enable-snapdiff'} = $create_args{'enable-snapdiff'}
        if ( defined $create_args{'enable-snapdiff'} );
    $limit_options{'ns-mirror-aggr-list'}
        = $create_args{'ns-mirror-aggr-list'}
        if ( defined $create_args{'ns-mirror-aggr-list'} );

    my $zapi = $params{'zapi'};
    my $response;
    this_block {
        $response = $zapi->volume_get_limits(%limit_options);
    }
    should( pass() );

    my @output = @{ $response->get_parsed_output };
    $Log->log('Validation Info: volume-get-limits returned a max size of '
            . $output[0]->{'max-infinitevol-size'}
            . ' bytes' );
    return $limit_options{'space-guarantee'} eq 'none'
        ? $output[0]->{'max-infinitevol-size'}
        - $THIN_PROVISIONING_MAX_SIZE_BUFFER
        : $output[0]->{'max-infinitevol-size'}
        - $THICK_PROVISIONING_MAX_SIZE_BUFFER;
}

=head2 _get_infinitevol_max_size

This function translates the zapi keywords into their cli counterparts for volume-create-async zapi

=cut

sub _zapi_to_cli_mapping_for_create_async {
    my %params = @_;

    $params{'volume'}         = delete $params{'volume-name'};
    $params{'policy'}         = delete $params{'export-policy'};
    $params{'group'}          = delete $params{'group-id'};
    $params{'user'}           = delete $params{'user-id'};
    $params{'nvfail'}         = delete $params{'is-nvfail-enabled'};
    $params{'maxdir-size'}    = delete $params{'max-dir-size'};
    $params{'security-style'} = delete $params{'volume-security-style'};
    $params{'state'}          = delete $params{'volume-state'};
    $params{'type'}           = delete $params{'volume-type'};

    return %params;
}

=head2 _stabilize_aggregates

This function wait for aggregate to become stable

=cut

sub _stabilize_aggregates {
    my $self       = shift;
    my %parameters = @_;

# It make take a few seconds after deletion for wafl to finish freeing up blocks
    $Log->log(
        'Validation Starting: Waiting for the available sizes on the aggregates to settle'
    );
    my %availsizes = $self->get_initial_aggregate_state(
        'node'    => $parameters{'node'},
        'vserver' => $parameters{'vserver'}
    );
    my $num_dc_nodes = $self->get_number_of_dc_nodes(
        'node'           => $parameters{'node'},
        'availsizes'     => \%availsizes,
        'data-aggr-list' => $parameters{'data-aggr-list'}
    );
    my $is_stable = $FALSE;

    for ( my $i = 0; $i < $self->{'aggregate_wait_time'}; $i += 10 ) {
        sleep 10;
        my %new_availsizes = $self->get_initial_aggregate_state(
            'node'    => $parameters{'node'},
            'vserver' => $parameters{'vserver'}
        );
        if ( eq_deeply( \%availsizes, \%new_availsizes ) ) {
            $Log->log(
                'Validation Info: Available sizes was constant for 10 seconds'
            );
            $is_stable = $TRUE;
            last;
        }
        else {
            $Log->log(
                'Validation Info: Available sizes was not constant for 10 seconds'
            );
            %availsizes = %new_availsizes;
        }
    }

    if ( !$is_stable ) {
        $Log->log('Validation Failure: Available Sizes never stabilized');
    }
    else {
        $Log->log('Validation Success: Available Sizes stabilized');
        $Log->log(
            'Validation Ending: Waiting for the available sizes on the aggregates to settle'
        );
    }
    this($is_stable)->should( Be->true() );
}

=head2 _get_constituent_by_role

This function returns constituenta for specified roles

=cut

sub _get_constituent_by_role {
    my $self       = shift;
    my %parameters = @_;
    my %volume_info_by_aggr;
    my @vols = ();

    $parameters{'state'} = 'final' unless defined $parameters{'state'};

    if ( lc( $parameters{'state'} ) eq 'final' ) {
        %volume_info_by_aggr
            = %{ $self->{'final_state'}->{'volume_info_by_aggr'} };
    }
    else {
        %volume_info_by_aggr
            = %{ $self->{'initial_state'}->{'volume_info_by_aggr'} };
    }

    foreach ( ( keys %volume_info_by_aggr ) ) {
        push( @vols, @{ $volume_info_by_aggr{$_} } );
    }

    my @constituents = grep {
        (   $_->constituent_role eq $parameters{'role'}
                || ( lc( $parameters{'role'} ) eq "all"
                && $_->is_constituent eq 'true' )
            )
            && ( $_->vserver eq $self->{'final_state'}->{'vserver'}->vserver )
    } @vols;

    return wantarray ? @constituents : $constituents[0];
}

=head2 _get_available_size

This function returns the available size of the aggregates in vserver aggregate list

=cut

sub _get_available_size {
    my $self  = shift;
    my @aggrs = ();

    my %availsizes   = ();
    my @vs_aggr_list = @{ $self->{'final_state'}->{'vs_aggr_list'} };

    if ( $vs_aggr_list[0] eq '-' ) {
        %availsizes
            = %{ $self->{'initial_state'}->{'aggrs_sizes'}->{'availsize'} };
    }
    else {
        foreach (@vs_aggr_list) {
            $availsizes{$_} = $self->{'initial_state'}->{'aggrs_sizes'}
                ->{'availsize'}{$_};
        }
    }

    return \%availsizes;
}

=head2 _set_final_state_info

This function sets Vserver and Infintevol information upon creation or expansion

=cut

sub _set_final_state_info {
    my $self   = shift;
    my %params = validate_with(
        params => \@_,
        spec   => {
            vserver        => { type => HASHREF | UNDEF,  default => undef },
            infinitevol    => { type => HASHREF | UNDEF,  default => undef },
            vs_aggr_list   => { type => ARRAYREF | UNDEF, default => '-' },
            size_specified => { type => SCALAR | UNDEF,   default => undef },
            enable_object_store => { type => SCALAR, default => 'false' },
            max_namespace_constituent_size =>
                { type => SCALAR | UNDEF, default => undef },
            max_data_constituent_size =>
                { type => SCALAR | UNDEF, default => undef },
            namespace_aggregate =>
                { type => SCALAR | UNDEF, default => undef },
            data_aggr_list => { type => ARRAYREF,       default => [] },
            num_dc_nodes   => { type => SCALAR | UNDEF, default => undef },
            ns_mirror_aggr_list => { type => ARRAYREF, default => [] },
            ols_aggr_list       => { type => ARRAYREF, default => [] },
            list_of_nodes       => { type => ARRAYREF, default => [] }
        },
        allow_extra => $TRUE,
    );

    foreach my $field ( keys %params ) {
        my $alias_field = $field;
        $alias_field =~ s/_/-/g;
        $self->{'final_state'}->{$field} = $params{$field};
        if ( defined $params{$alias_field} ) {
            $self->{'final_state'}->{$field} = $params{$alias_field};
        }
    }
}

1;
