#===============================================================================
#     REVISION: $Id$
#         FILE: SetupUtils.pm
#
#  DESCRIPTION: This library contains common subroutines for tasks specific
#               InfiniteVol setup.
#
#         BUGS: Not known
#        NOTES: ---
#       AUTHOR: dl-svl-nosauto
#      COMPANY: NetApp, Inc
#      VERSION: 1.0
#      CREATED: Friday 26,October,2012 02:07:38  EDT
# LAST CHANGED: $DateTime$
# CHANGED BY  : $Author$
#===============================================================================

use strict;
use warnings;

use NACL::STask::Volume;
use NACL::CS::Disk;

use Tharn qw($Log);
use Params::Validate qw(SCALAR OBJECT ARRAYREF HASHREF validate);
use NACL::C::VserverDataPolicy;

our @EXPORT_OK = qw (
    create_dc_on_infinitevol
    delete_created_dc_on_infinitevol
    import_data_policy
    calculate_avg_files_per_constituent
    convert_dc_bucket_type
    find_aggr_most_usable_space
    configure_multiple_dcgs
    disable_multiple_dcgs
);

my %Clean_Up;

#===  FUNCTION  ================================================================
#         NAME: create_dc_on_infinitevol
#  DESCRIPTION: Creates a data constituent on the infinitevolume.
#               In order to do so, it needs to find or create an aggregate that
#               can accommodate the constituent.
#               It also creates a closure which cleans up everything it created
#               and keeps it in a global hash.
#               Need to call delete_created_dc_on_infinitevol to execute the
#               closure.
#        INPUT: vserver_obj (Required)
#               vserver object (on which the infinitevol is created on)
#
#               node_to_aggr_create_on (Optional)
#               Node on which aggr is to be created.
#
#               dc_volume_params (Optional)
#               Additional volume parameters to be used while creating DCs
#
#      RETURNS: DC volume object
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#        USAGE: $dc_obj = create_dc_on_infinitevol(
#                             vserver_obj => $vserver_obj,
#                             node_to_aggr_create_on => your-vm-name
#                             dc_volume_params
#                                 => {'is-managed-by-service' => 'true',
#                                             'storage-service' => 'silver'}
#                         );
#===============================================================================
sub create_dc_on_infinitevol {
    my %params = validate(@_, { vserver_obj => {
                                                type => OBJECT,
                                                optional => 0,
                                               },
                                node_to_aggr_create_on => {
                                                           type => SCALAR,
                                                           optional => 1,
                                                          },
                                dc_volume_params => {
                                                   type => HASHREF,
                                                   optional => 1,
                                               }
                              }
                 );

    my $vserver_obj = $params{vserver_obj};
    my %dc_volume_params;
    if  ( $params{dc_volume_params} ) {
        %dc_volume_params = %{$params{dc_volume_params}};
    }

    my $node_obj = $vserver_obj->command_interface();
    my @aggr_list = $vserver_obj->get_one_state_attribute('aggr-list');

    my ($new_aggr_flag, $needed_avail_size, $target_aggr_obj);
    my $aggr_name_filter = join(q{,}, @aggr_list);

    #If aggr-list is - , check for all the aggrs in the cluster
    if ($aggr_name_filter =~ /-/) {
        $aggr_name_filter = q{*};
    }

    my @aggr_objs = NACL::CS::Aggregate->fetch(
                       command_interface => $node_obj,
                       filter            => {'aggregate' => $aggr_name_filter},
                    );

    #Refer http://wikid.netapp.com/w/ContentRepository/Setup/
    #Aggregate_Modelling_for_InfiniteVol
    @aggr_objs =
        reverse sort {($a->availsize() - 0.09 * $a->size()) <=>
                      ($b->availsize() - 0.09 * $b->size())
                     } @aggr_objs;

    #For a vsim, 256MB constituent is created wheres for filer it is 1TB.
    #Sizes below are in bytes
    if ($node_obj->hostrec->MODEL() eq 'VSIM') {
        $needed_avail_size = 268435456;
    } else {
        $needed_avail_size = 1024 ** 4;
    }

    #The first aggregate after the sort - which has teh most available space
    my $most_avail_aggr_size =
        ($aggr_objs[0]->availsize()) - (0.09 * $aggr_objs[0]->size());

    if ($most_avail_aggr_size >= $needed_avail_size) {
        $Log->comment('Found an aggr to create new data constituent');
        $target_aggr_obj = $aggr_objs[0];
    } else {
        $Log->comment('Did not find an aggr to create new data constituent, '.
                    'will create one');
        $new_aggr_flag = 1;
    }

    #No aggr was found - create one.
    if ($new_aggr_flag) {

        #Find the node on which to create one
        my $node_to_use = $params{node_to_aggr_create_on};
        if ($node_to_use) {
            my @disks_objs =
                  NACL::CS::Disk->fetch(command_interface => $node_obj,
                                        filter => {'container-type' => 'spare',
                                                   owner  => $node_to_use},
                  );
            if (scalar(@disks_objs) < 3) {
                $Log->warn('Cannot create aggr - specified node has '.
                           scalar(@disks_objs).'spare disks');
                return;
            }
        } else {
            my $avail_disks_count = 0;
            my @all_node_objs =
                NACL::C::Node->find(command_interface => $node_obj);
            foreach my $node_obj (@all_node_objs) {
                my @disks_objs =
                 NACL::CS::Disk->fetch(command_interface => $node_obj,
                                       filter => {'container-type' => 'spare',
                                                  owner  => $node_obj->node()},
                 );
                if (scalar(@disks_objs) > $avail_disks_count) {
                    $avail_disks_count = scalar(@disks_objs);
                    $node_to_use = $node_obj->node();
                }
            }
            if ($avail_disks_count < 3) {
                $Log->warn('No nodes found with atleast 3 spare disks');
                return;
            }
        }

        my $aggr_name = 'aggr_for_add_dc_'.time();
        $target_aggr_obj =
         NACL::STask::Aggregate->create(command_interface => $node_obj,
                                        aggregate         => $aggr_name,
                                        nodes             => [$node_to_use],
                                        diskcount         => 3,
         );
        if ($aggr_name_filter ne q{*}) {
            $target_aggr_obj->add_to_aggr_list_of_vserver(
                                           vserver => $vserver_obj->vserver());
        }
    }

    #Create the data constituent
    my $response_obj =
        $node_obj->apiset()->volume_create(
            vserver => $vserver_obj->vserver(),
            aggregate => $target_aggr_obj->aggregate(),
            'constituent-role' => 'data',
            %dc_volume_params,
        );

    my $dc_obj;
    if ($response_obj->get_processed_output() =~
                                    /Created constituent (\S+) successfully/i) {
        $Log->comment("Constituent $1 created successfully");
        $dc_obj =
            NACL::STask::Volume->find(command_interface => $node_obj,
                                  filter => {
                                             volume => $1,
                                             vserver => $vserver_obj->vserver()
                                            }
            );
    } else {
        $Log->warn('*** FAIL *** Something went wrong');
    }

    $Clean_Up{$dc_obj->vserver()}{$dc_obj->volume()} =
        sub {
             if ($dc_obj) {
                 $Log->comment('Deleting constituent '.$dc_obj->volume());
                 $dc_obj->purge();
             }

             if ($new_aggr_flag) {
                 $target_aggr_obj->remove_from_aggr_list_of_vserver(
                                                vserver => $dc_obj->vserver());
                 $Log->comment('Deleting aggr created '.
                             $target_aggr_obj->aggregate());
                 $target_aggr_obj->purge();
             }
            };

    return $dc_obj;
}


#===  FUNCTION  ================================================================
#         NAME: delete_created_dc_on_infinitevol
#  DESCRIPTION: Deletes data constituent created by create_dc_on_infinitevol
#               If an aggr was created in create_dc_on_infinitevol, deletes
#               that too.
#        INPUT: None
#      RETURNS: None
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: create_dc_on_infinitevol
#        USAGE: delete_created_dc_on_infinitevol($dc_obj)
#===============================================================================
sub delete_created_dc_on_infinitevol {
    my ($dc_obj) = @_;

    return $Clean_Up{$dc_obj->vserver()}{$dc_obj->volume()}->();
}

#===  FUNCTION  ================================================================
#         NAME: import_data_policy
#      PURPOSE: This function exports the data policy and imports it.
#   PARAMETERS: None
#      RETURNS: 1
#  DESCRIPTION: This is needed when DCs are unassigned & need to be resassigned.
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub import_data_policy {

    my %opts = validate(@_,
        {node_obj => {type => OBJECT, optional => 0},
         vserver_name => {type => SCALAR, optional => 0},
         data_policy => {type => SCALAR, optional => 0},}
    );

    # Importing the policy
    NACL::C::VserverDataPolicy->do_import(
        command_interface => $opts{node_obj},
        vserver           => $opts{vserver_name},
        file_data_policy  => $opts{data_policy},
    );

    return 1;
} ## end sub import_data_policy

#===  FUNCTION  ================================================================
#         NAME: calculate_avg_files_per_constituent
#   PARAMETERS: node_obj - Command interface to any node in the object
#               vserver_name - The vserver to operate on
#               @bucket_rows - the rows in the ksmf_repos_ov table to sum
#               %initial_files_used - a hash containing the number of files
#                   used for each DC. Hash keys are the DC msid
#      RETURNS: Incremented bucket file count.
#  DESCRIPTION: Sums the new number of files used for each DC in a given
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub calculate_avg_files_per_constituent {

    my %opts = validate (@_,
        {node_obj => {type => OBJECT, optional => 0},
         vserver_name => {type => SCALAR, optional => 0},
         bucket_rows => {type => ARRAYREF, optional => 0},
         initial_files_used => {type => HASHREF, optional => 0},}
    );

    my @bucket_rows = @{ $opts{'bucket_rows'} };
    my %initial_files_used = %{ $opts{'initial_files_used'} };
    my $bucket_file_count = 0;

    foreach my $row (@bucket_rows) {
        my $msid = $row->get_one_state_attribute('volMsid');
        my $dc = NACL::C::Volume->find(
            'command_interface' => $opts{node_obj},
            'filter'            => {
                'vserver' => $opts{vserver_name},
                'msid'    => $msid,
            },
        );
        $bucket_file_count += ( $dc->get_one_state_attribute('files-used')
                - $initial_files_used{$msid} );
    }

    $bucket_file_count /= scalar(@bucket_rows);

    return $bucket_file_count;
} ## end sub calculate_avg_files_per_constituent

#===  FUNCTION  ================================================================
#         NAME: convert_dc_bucket_type
#      PURPOSE: To convert bucket type of each of the specified rows of table
#               ksmf_repos_ov to any of the specified bucket type.
#   PARAMETERS: node_obj (Required): Command interface of one of the node in
#                   the cluster.
#               bucket_rows (Required): Reference to array containing rows from
#                   table ksmf_repos_ov whose value needs to be modified.
#               bucket_types (Required): Reference to array containing
#                   the bucket values to be changed.
#                   Valide values in the array ref:
#                   REPOS_VOL_REGULAR_BUCKET
#                   REPOS_VOL_RESERVED_BUCKET
#                   REPOS_VOL_EMPTY_BUCKET
#                   REPOS_VOL_FULL_BUCKET
#      RETURNS: 1
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#        USAGE:
#               convert_dc_bucket_type(
#                   node_obj => $Node_Obj,
#                   bucket_rows => \@empty_bucket_rows,
#                   bucket_types => \@required_buckets,
#               );
#===============================================================================
sub convert_dc_bucket_type{

    my %opts = validate (@_,
        {node_obj => {type => OBJECT, optional => 0},
         bucket_rows => {type => ARRAYREF, optional => 0},
         bucket_types => {type => ARRAYREF, optional => 0},}
    );


    my $cli = $opts{node_obj}->apiset(
        category  => 'Node',
        interface => 'CLI',
        set       => 'CMode',
    );

    $cli->set_privilege( 'privilege-level' => 'test' );

    foreach my $bucket_row ( @{$opts{bucket_rows}} ){

        my $required_bucket = shift @{$opts{bucket_types}};

        my $orig_msid = $bucket_row->get_one_state_attribute('origMsid');
        my $vs_id     = $bucket_row->get_one_state_attribute('vserver');
        my $msid      = $bucket_row->get_one_state_attribute('volMsid');
        my $ovg_id    = $bucket_row->get_one_state_attribute('ovg-id');
        my $dir_bw = $bucket_row->get_one_state_attribute('directoryBitWidth');

        # modifying the table needs to be done at diag privilege so grab a CLI
        # apiset and execute the raw command
        my $cmd = "debug smdb table ksmf_repos_ov modify -vserver $vs_id "
                . "-volMsid $msid -ovg-id $ovg_id -directoryBitWidth "
                . "$dir_bw -origMsid $orig_msid -createBucket "
                . $required_bucket;

        $Log->comment('Modifying ksmf_repos_ov table to place DC having'
                    . " origMsid $orig_msid into $required_bucket bucket.");
        $cli->execute_raw_command( command => $cmd );
    }

    return 1;

}

#===  FUNCTION  ================================================================
#         NAME: find_aggr_most_usable_space
#  DESCRIPTION: On a given node, finds aggr with most usable space
#        INPUT: Node object
#      RETURNS: aggr state object
#       THROWS: no exceptions
#     COMMENTS: none
#===============================================================================
sub find_aggr_most_usable_space {
    my ($node_obj) = @_;

    my @aggr_objs = NACL::CS::Aggregate->fetch(command_interface => $node_obj);

    #Refer http://wikid.netapp.com/w/ContentRepository/Setup/
    #Aggregate_Modelling_for_InfiniteVol
    @aggr_objs =
        reverse sort {($a->availsize() - 0.09 * $a->size()) <=>
                      ($b->availsize() - 0.09 * $b->size())
                     } @aggr_objs;

    return $aggr_objs[0];
}


#===  FUNCTION  ================================================================
#         NAME: configure_multiple_dcgs
#   PARAMETERS: infinitevol_obj - Infinite volume object
#               node_names - Names of the nodes on which new DC will be created
#               vserver_obj - Vserver object
#               storage_service_for_current_dcs - Storage service type for
#                                                 existing DCs on IV
#               storage_service_for_created_dcs - Storage service type for
#                                                 created DCs on IV
#               data_policy - Storage service data policy that needs to be
#                             imported (optional)
#      RETURNS: Return a hash reference contains a reference to created DC
#               objects, original data policy, and DCG info.
#  DESCRIPTION: This function enables storage service on IV, configure the
#               storage service type for existing DCs, create new DCs on IV,
#               and configure storage service type for the new DCs
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub configure_multiple_dcgs {

    my %opts = validate (@_,
        {infinitevol_obj => {type => OBJECT, optional => 0},
         node_names => {type => ARRAYREF, optional => 0},
         vserver_obj => {type => OBJECT, optional => 0},
         storage_service_for_current_dcs => {type => SCALAR, optional => 0},
         storage_service_for_created_dcs => {type => SCALAR, optional => 0},
         data_policy => {type => SCALAR, optional => 1},}
    );
    my (
        $data_policy, $iv_obj, $vserver_obj, @created_dc_objs, @current_dc_objs,
        $ss_info
    );
    $iv_obj = $opts{infinitevol_obj};
    $data_policy = $opts{data_policy} if ($opts{data_policy});
    $vserver_obj = $opts{vserver_obj};

    my $vserver_name = $vserver_obj->vserver();
    my $node_obj = $vserver_obj->command_interface();
    my $node_api_obj = $node_obj->apiset();

    # Default policy to be imported
    my $default_policy = '{
        "ruleset_format_version" : "1.0",
        "rules" : [
            {
                "rule_label" : "Thpl",
                "rule_id" : "00fc0534-10dd-7812-ec8c-a9356fa8ce97",
                "rule_scope" : [{"objectName" : "starts thpl."}],
                "rule_epoch" : { "epoch_reference" : "ctime" },
                "rule_epochs" : { "0" : { "local" : { "metadata" : {
                    "storageservice" : "gold" } } } }
            },
            {
                "rule_label" : "Txt",
                "rule_id" : "00fc0534-10dd-7812-ec8c-a9356fa8cd99",
                "rule_scope" : [{"objectName" : "starts txt."}],
                "rule_epoch" : { "epoch_reference" : "ctime" },
                "rule_epochs" : { "0" : { "local" : { "metadata" : {
                    "storageservice" : "gold" } } } }
            },
            {
                "rule_label" : "Image",
                "rule_id" : "00fc0534-10dd-7812-ec8c-a9356fa8cf01",
                "rule_scope" : [{"objectName" : "starts jpg."}],
                "rule_epoch" : { "epoch_reference" : "ctime" },
                "rule_epochs" : { "0" : { "local" : { "metadata" : {
                    "storageservice" : "gold" } } } }
            },
            {
                "rule_label" : "default",
                "rule_id" : "00fc0534-10dd-7812-ec8c-a9356fa8ca00",
                "rule_scope" : [],
                "rule_epoch" : {"epoch_reference" : "ctime" },
                "rule_epochs" : { "0" : { "local" : { "metadata" : {
                    "storageservice" : "silver" } } } }
            }
        ]
    }';

    $Log->comment('Store the original data-policy on vserver');

    # Use API set object because burt620870 hasn't been fixed
    my $original_data_policy = $node_api_obj->execute_raw_command(
        command => "vserver data-policy export -vserver $vserver_name"
    );

    $Log->comment(
        'Modifying IV to enable storage-service and set current DCs storage'
        . " service type to $opts{storage_service_for_current_dcs}"
    );

    $iv_obj->modify(
        'is-managed-by-service' => 'true',
        'storage-service' => $opts{storage_service_for_current_dcs},
    );

    @current_dc_objs = NACL::C::Volume->find(
        command_interface => $node_obj,
        filter            => {
            vserver => $vserver_name,
            'constituent-role' => 'data'
        }
    );

    $Log->comment('Store storage service information for current DCs');
    $ss_info->{$opts{storage_service_for_current_dcs}} = \@current_dc_objs;

    my $ss_for_created_dcs = $opts{storage_service_for_created_dcs};
    $Log->comment("Creating new DCs with $ss_for_created_dcs storage service");

    foreach my $node_name ( @{$opts{node_names}} ){
        push @created_dc_objs, create_dc_on_infinitevol(
            vserver_obj => $opts{vserver_obj},
            node_to_aggr_create_on => $node_name,
            dc_volume_params => {
                'constituent-role' => 'data',
                'storage-service' => $opts{storage_service_for_created_dcs},
            }
        );
    }

    $Log->comment('Store storage service information for created DCs');
    $ss_info->{$opts{storage_service_for_created_dcs}} = \@created_dc_objs;

    # import data policy
    import_data_policy(
        node_obj => $node_obj,
        vserver_name => $vserver_obj->vserver(),
        data_policy => $data_policy | $default_policy,
    );

    return {dc_objs =>\@created_dc_objs,
            dcg_info => $ss_info,
            original_data_policy => $original_data_policy};
}

#===  FUNCTION  ================================================================
#         NAME: disable_multiple_dcgs
#   PARAMETERS: created_dc_objs - DC objects created through
#                                 configure_multiple_dcgs()
#               infinitevol - Infinite volume object
#      RETURNS: Return 1 when succeeds, else 0
#  DESCRIPTION: Delete DCs created through configure_multiple_dcgs(), disable
#               storage service on IV
#       THROWS: no exceptions
#     COMMENTS: none
#     SEE ALSO: n/a
#===============================================================================
sub disable_multiple_dcgs {
    my %opts = validate (@_,
        {infinitevol_obj => {type => OBJECT, optional => 0},
         vserver_obj => {type => OBJECT, optional => 0},
         created_dc_objs => {type => ARRAYREF, optional => 0},
         original_data_policy => {type => SCALAR, optional => 0},}
    );

    my $iv_obj = $opts{infinitevol_obj};

    $Log->comment('Deleting the created DCs');

    foreach my $dc_obj ( @{$opts{created_dc_objs}} ) {
       $Log->comment('Deleting the dc: ' . $dc_obj->volume());
       delete_created_dc_on_infinitevol($dc_obj);
    }

    $Log->comment('Modifying the IV to disable storage-service');

    $iv_obj->modify('is-managed-by-service' => 'false');

    $Log->comment('Importing the original data policy');

    NACL::C::VserverDataPolicy->do_import(
        command_interface => $opts{vserver_obj}->command_interface(),
        vserver => $opts{vserver_obj}->vserver(),
        file_data_policy => $opts{original_data_policy},
    );

    return 1;
}
1;
