#
# $Id: //depot/prod/test/Rsierranevada/lib/CR/Snapmirror.pm#1 $
#
# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Snapmirror Class
## @author davidnew@netapp.com
## @status shared
## @pod here

package CR::Snapmirror;

use strict;
use warnings;

use Data::Dumper;
use NACL::APISet;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use Math::Round;
use Params::Validate qw(:all);
use SPARSE::Util::Logger;

use constant DEL_TIMEOUT => 1800; # 30 min

=head1 NAME

Snapmirror - The class to encapsulate methods and data of "repository volume snapmirror" commands set

=head1 SYNOPSIS

=begin text

 use Snapmirror;

 $cr->repository_volume_snapmirror_break('destination-cluster' => $dest_cluster,
                                         'destination-vserver' => $dest_vserver,
                                         'source-cluster'      => $source_cluster,
                                         'source-vserver'      => $source_vserver,
                                         foreground            => $value);

 $cr->repository_volume_snapmirror_create('destination-cluster' => $dest_cluster,
                                          'destination-vserver' => $dest_vserver,
                                          'source-cluster'      => $source_cluster,
                                          'source-vserver'      => $source_vserver,
                                          foreground            => $value);

 $cr->repository_volume_snapmirror_delete('destination-cluster' => $dest_cluster,
                                          'destination-vserver' => $dest_vserver,
                                          'source-cluster'      => $source_cluster,
                                          'source-vserver'      => $source_vserver,
                                          foreground            => $value);

 $cr->repository_volume_snapmirror_modify('destination-cluster' => $dest_cluster,
                                          'destination-vserver' => $dest_vserver,
                                          'source-cluster'      => $source_cluster,
                                          'source-vserver'      => $source_vserver,
                                          'schedule'            => $schedule,
                                          foreground            => $value);

 $cr->repository_volume_snapmirror_resync('destination-cluster' => $dest_cluster,
                                         'destination-vserver' => $dest_vserver,
                                         'source-cluster'      => $source_cluster,
                                         'source-vserver'      => $source_vserver,
                                         foreground            => $value);

 $cr->repository_volume_snapmirror_show(instance => "",
                                      vserver  => $vserver,
                                      schedule => $cron_job);

 $cr->repository_volume_snapmirror_update('destination-cluster' => $dest_cluster,
                                          'destination-vserver' => $dest_vserver,
                                          'source-cluster'      => $source_cluster,
                                          'source-vserver'      => $source_vserver,
                                          foreground            => $value);
=end text

=head1 AUTHOR

David New (davidnew@netapp.com)

=cut

logLevel('info');

######################################
# Repository Volume Snapmirror Break
######################################

=head2 repository_volume_snapmirror_break

Purpose of this subroutine is to break repository snapmirror relationship to source.

Required parameters are

=over 4

=item *
destination_cluster

Snapmirror destination cluster node

=item *
destination_vserver

Snapmirror destination cluster vserver

=item *
source_cluster

Snapmirror source cluster cluster

=item *
source_vserver

Snapmirror source cluster vserver

=back

Optional parameters are

=over 4

=item *
foreground

Run the job in foreground (true/false)

=back

=cut

sub repository_volume_snapmirror_break {
  logStep("Calling repository volume snapmirror break");

  my ($self, %args) =@_;
  my $admin_cli = $self->{ngsh};

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           'destination-cluster' => {type => SCALAR, optional => 0},
                                           'destination-vserver' => {type => SCALAR, optional => 0},
                                           'source-cluster'      => {type => SCALAR, optional => 0},
                                           'source-vserver'      => {type => SCALAR, optional => 0},
                                           foreground            => {type => SCALAR, optional => 1}
                                         },
                                       );

  $admin_cli->repository_volume_snapmirror_break(%args,
                                                  'connectrec-timeout'=>DEL_TIMEOUT);
  logInfo("repository volume snapmirror break executed successfully\n");
  return 1;
}

######################################
# Repository Volume Snapmirror Create
######################################

=head2 repository_volume_snapmirror_create

Purpose of this subroutine is to create repository snapmirror relationship to source.

Required parameters are

=over 4

=item *
destination_cluster

Snapmirror destination cluster node

=item *
destination_vserver

Snapmirror destination cluster vserver

=item *
source_cluster

Snapmirror source cluster cluster

=item *
source_vserver

Snapmirror source cluster vserver

=back

Optional parameters are

=over 4

=item *
foreground

Run the job in foreground (true/false)

=back

=cut

sub repository_volume_snapmirror_create {
  logStep("Calling repository volume snapmirror create");

  my ($self, %args) =@_;
  my $admin_cli = $self->{ngsh};

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           'destination-cluster' => {type => SCALAR, optional => 0},
                                           'destination-vserver' => {type => SCALAR, optional => 0},
                                           'source-cluster'      => {type => SCALAR, optional => 0},
                                           'source-vserver'      => {type => SCALAR, optional => 0},
                                           foreground            => {type => SCALAR, optional => 1}
                                         },
                                       );

  $admin_cli->repository_volume_snapmirror_create(%args,
                                                  'connectrec-timeout'=>DEL_TIMEOUT);
  logInfo("repository volume snapmirror create executed successfully\n");
  return 1;
}

######################################
# Repository Volume Snapmirror Delete
######################################

=head2 repository_volume_snapmirror_delete

Purpose of this subroutine is to delete repository snapmirror relationship from source.

Required parameters are

=over 4

=item *
destination_cluster

Snapmirror destination cluster node

=item *
destination_vserver

Snapmirror destination cluster vserver

=item *
source_cluster

Snapmirror source cluster cluster

=item *
source_vserver

Snapmirror source cluster vserver

=back

Optional parameters are

=over 4

=item *
foreground

Run the job in foreground (true/false)

=back

=cut

sub repository_volume_snapmirror_delete {
  logStep("Calling repository volume snapmirror delete");

  my ($self, %args) =@_;
  my $admin_cli = $self->{ngsh};

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           'destination-cluster' => {type => SCALAR, optional => 0},
                                           'destination-vserver' => {type => SCALAR, optional => 0},
                                           'source-cluster'      => {type => SCALAR, optional => 0},
                                           'source-vserver'      => {type => SCALAR, optional => 0},
                                           foreground            => {type => SCALAR, optional => 1}
                                         },
                                       );

  $admin_cli->repository_volume_snapmirror_delete(%args,
                                                  'connectrec-timeout'=>DEL_TIMEOUT);
  logInfo("repository volume snapmirror delete executed successfully\n");
  return 1;
}

######################################
# Repository Volume Snapmirror Modify
######################################

=head2 repository_volume_snapmirror_modify

Purpose of this subroutine is to modify repository volume snapmirror options

Required parameters are

=over 4

=item *
destination_cluster

Snapmirror destination cluster node

=item *
destination_vserver

Snapmirror destination cluster vserver

=item *
source_cluster

Snapmirror source cluster cluster

=item *
source_vserver

Snapmirror source cluster vserver

=item *
schedule

Snapmirror schedule from "cron job show"

=back

Optional parameters are

=over 4

=item *
foreground

Run the job in foreground (true/false)

=back

=cut

sub repository_volume_snapmirror_modify{
  logStep("Calling repository volume snapmirror modify");

  my ($self, %args) =@_;
  my $admin_cli = $self->{ngsh};

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           'destination-cluster' => {type => SCALAR, optional => 0},
                                           'destination-vserver' => {type => SCALAR, optional => 0},
                                           'source-cluster'      => {type => SCALAR, optional => 0},
                                           'source-vserver'      => {type => SCALAR, optional => 0},
                                           schedule              => {type => SCALAR, optional => 0},
                                           foreground            => {type => SCALAR, optional => 1}
                                         },
                                       );

  ##  It's better to clear the snapmirror schedule with "" first then only assigned the wanted schedule.
  ##  If assigning the same schedule name to the snapmirror, it might not work as expected.
  if ($args{schedule} ne "\"\"") {
    $admin_cli->repository_volume_snapmirror_modify('destination-cluster' => $args{'destination-cluster'},
                                                    'destination-vserver' => $args{'destination-vserver'},
                                                    'source-cluster'      => $args{'source-cluster'},
                                                    'source-vserver'      => $args{'source-vserver'},
                                                    schedule              => "\"\"",
                                                    foreground            => $args{foreground});
  }
  $admin_cli->repository_volume_snapmirror_modify(%args);
  logInfo("repository volume snapmirror modify with schedule $args{schedule} executed successfully\n");
  return 1;
}

######################################
# Repository Volume Snapmirror Resync
######################################

=head2 repository_volume_snapmirror_resync

Purpose of this subroutine is to resync repository snapmirror relationship to source.

Required parameters are

=over 4

=item *
destination_cluster

Snapmirror destination cluster node

=item *
destination_vserver

Snapmirror destination cluster vserver

=item *
source_cluster

Snapmirror source cluster cluster

=item *
source_vserver

Snapmirror source cluster vserver

=back

Optional parameters are

=over 4

=item *
foreground

Run the job in foreground (true/false)

=back

=cut

sub repository_volume_snapmirror_resync {
  logStep("Calling repository volume snapmirror resync");

  my ($self, %args) =@_;
  my $admin_cli = $self->{ngsh};

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           'destination-cluster' => {type => SCALAR, optional => 0},
                                           'destination-vserver' => {type => SCALAR, optional => 0},
                                           'source-cluster'      => {type => SCALAR, optional => 0},
                                           'source-vserver'      => {type => SCALAR, optional => 0},
                                           foreground            => {type => SCALAR, optional => 1}
                                         },
                                       );

  $admin_cli->repository_volume_snapmirror_resync(%args,
                                                  'connectrec-timeout'=>DEL_TIMEOUT);
  logInfo("repository volume snapmirror resync executed successfully\n");
  return 1;
}

######################################
# Repository Volume Snapmirror Show
######################################

=head2 repository_volume_snapmirror_show

Purpose of the subroutine is to display the snapmirror details.

Required parameters are

none

Optional parameters are

=over 4

=item *
instance

Selected fields to display

=item *
vserver

Name of the repository vserver

=item *
schedule

name of the snapmirror cron job

=back

=cut

sub repository_volume_snapmirror_show{
  logStep("Calling repository volume snapmirror show");

  my ($self,%args)=@_;
  my $admin_cli = $self->{ngsh};
  my @parsed_output;

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           instance => {type => SCALAR, optional => 1},
                                           vserver  => {type => SCALAR, optional => 1},
                                           schedule => {type => SCALAR, optional => 1}
                                         },
                                        );
  @parsed_output=@{$admin_cli->repository_volume_snapmirror_show(%args)->get_parsed_output()};
  logInfo("repository volume snapmirror show executed successfully\n");
  return @parsed_output;
}

######################################
# Repository Volume Snapmirror Update
######################################

=head2 repository_volume_snapmirror_update

Purpose of this subroutine is to update repository snapmirror from source (pulling).

Required parameters are

=over 4

=item *
destination_cluster

Snapmirror destination cluster node

=item *
destination_vserver

Snapmirror destination cluster vserver

=item *
source_cluster

Snapmirror source cluster cluster

=item *
source_vserver

Snapmirror source cluster vserver

=back

Optional parameters are

=over 4

=item *
foreground

Run the job in foreground (true/false)

=back

=cut

sub repository_volume_snapmirror_update {
  logStep("Calling repository volume snapmirror update");

  my ($self, %args) =@_;
  my $admin_cli = $self->{ngsh};

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           'destination-cluster' => {type => SCALAR, optional => 0},
                                           'destination-vserver' => {type => SCALAR, optional => 0},
                                           'source-cluster'      => {type => SCALAR, optional => 0},
                                           'source-vserver'      => {type => SCALAR, optional => 0},
                                           foreground            => {type => SCALAR, optional => 1}
                                         },
                                       );

  $admin_cli->repository_volume_snapmirror_update(%args,
                                                  'connectrec-timeout'=>DEL_TIMEOUT);
  logInfo("repository volume snapmirror update executed successfully\n");
  return 1;
}
