#
# $Id: //depot/prod/test/Rsierranevada/lib/CR/Snapshot2.pm#1 $
#
# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Snapshot2 Class
## @author davidnew@netapp.com
## @status shared
## @pod here

package CR::Snapshot2;

use strict;
use warnings;

use CR::Job;
use Data::Dumper;
use NACL::APISet;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use Math::Round;
use Params::Validate qw(:all);
use SPARSE::Util::Logger;

use constant DEL_TIMEOUT => 300; # 5 min

=head1 NAME

Snapshot2 - The class to encapsulate methods and data of "volume snapshot" commands set

=head1 SYNOPSIS

=begin text

 use Snapshot2;

 $cr->volume_snapshot_create(snapshot   => $snapshot_name,
                             vserver    => $vserver,
                             volume     => $volume,
                             comment    => $comment,
                             foreground => $value);

 $cr->volume_snapshot_delete(vserver    => $vserver,
                             volume     => $volume,
                             snapshot   => $snapshot_name,
                             foreground => $value);

 $cr->volume_snapshot_exists(vserver  => $vserver,
                             volume   => $volume,
                             snapshot => $snapshot_name);

 $cr->volume_snapshot_promote(vserver  => $vserver,
                              volume   => $volume,
                              snapshot => $snapshot_name);

 $cr->volume_snapshot_show(instance    => "",
                           vserver     => $vserver,
                           volume      => $volume,
                           snapshot    => $snapshot_name,
                           comment     => $snapshot_comment,
                           state       => $state,
                           create-time => $createTime,
                           size        => $size,
                           blocks      => $blocks,
                           usedblocks  => $usedblocks);

 $cr->volume_snapshot_show_options(vserver  => $vserver,
                                   volume   => $volume,
                                   snapshot => $snapshot_name);

 $cr->volume_snapshot_verify(vserver  => $vserver,
                             volume   => $volume,
                             snapshot => $snapshot_name);
 
 #$cr->volume_snapshot_delete(vserver  => $vserver,
 #                            snapshot => $snapshot_name);;

 #$cr->volume_snapshot_show(vserver  => $vserver,
 #                          snapshot => $snapshot_name);;

=end text

=head1 AUTHOR

David New (davidnew@netapp.com)

=cut

logLevel('info');

##################################################################
# Volume Snapshot Create
##################################################################

=head2 volume_snapshot_create

Purpose of this subroutine is to create infinivol volume snapshots

Required parameters are

=over 4

=item *
$vserver

The infinivol vserver

=item *
volume

name of the infinivol volume

=item *
$snapshot

The name of the snapshot

=back

Optional parameters are

=over 4

=item *
comment

Text comment for snapshot

=item *
foreground

Run the job in foreground (true/false)

=back

=cut

sub volume_snapshot_create{
  logStep("Calling volume snapshot create");

  my ($self, %args) =@_;
  my $admin_cli = $self->{ngsh};

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           vserver    => {type => SCALAR, optional => 0},
                                           volume     => {type => SCALAR, optional => 0},
                                           snapshot   => {type => SCALAR, optional => 0},
                                           comment    => {type => SCALAR, optional => 1},
                                           foreground => {type => SCALAR, optional => 1}
                                         },
                                       );
  $admin_cli->volume_snapshot_create(%args);
  logInfo("CLI 'volume snapshot created' executed successfully.\n");
  return 1;
}

##################################################################
# Volume Snapshot Delete
##################################################################

=head2 volume_snapshot_delete

The purpose of this subroutine is to delete $snapshot from infinivol vserver $vserver

Optional parameters are

=over 4

=item *
vserver

Name of the infinivol vserver

=item *
volume

name of the infinivol volume

=item *
snapshot

Name of the snapshot to be deleted

=item *
foreground

Run the job in foreground (true/false)

=back

=cut

sub volume_snapshot_delete {
  logStep("Calling volume snapshot delete");
  
  my($self,%args)=@_;
  my $admin_cli = $self->{ngsh};
  my (@parsed_output);
  my $success;

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           vserver    => {type => SCALAR, optional => 1},
                                           volume     => {type => SCALAR, optional => 1},
                                           snapshot   => {type => SCALAR, optional => 1},
                                           foreground => {type => SCALAR, optional => 1}
                                         },
                                        );

  try {
    $success=$admin_cli->volume_snapshot_delete(%args,
                                                'connectrec-timeout'=>DEL_TIMEOUT);
    $success=1;
  }
  catch NACL::APISet::Exceptions::ResponseException with {
    my $ex = shift;

    if ($ex =~ m/no entries matching/g) {
      $success=1;
    } elsif ($ex) {
      $success=0;
      $ex->throw();
    }
  };

  if ($success==1) {
    ## Removing element foreground as it's not supported by volume snapshot show
    delete($args{foreground});
    try {
      @parsed_output =$self->CR::Snapshot2::volume_snapshot_show(%args);
      $success=0;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
      my $ex = shift;

      if (($ex =~ m/no entries matching/g) || ($ex =~ m/currently empty/g)) {
        $success=1;
      } elsif ($ex) {
        $success=0;
        $ex->throw();
      }
    };

    if ($success==1) {
      logInfo("CLI 'volume snapshot delete' executed successfully.\n");
      return 1;
    } else {
      logInfo("CLI 'volume snapshot delete' failed to execute.\n");
      return 0;
    }
  }
}

##################################################################
# Volume Snapshot Exist checking
##################################################################

=head2 volume_snapshot_exists

Purpose of the subroutine is to find out whether snapshot ($snapshot) exists
in the infinivol ($vserver)

Required parameters are

=over 4

=item *
vserver
The infinivol vserver

=item *
volume

name of the infinivol volume

=item *
snapshot

The name of the snapshot

=back

=cut

sub volume_snapshot_exists{
  logStep("Checking to see if snapshot exists");
  
  my($self,%args)=@_;
  my $admin_cli = $self->{ngsh};
  my @parsed_output;

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           vserver  => {type => SCALAR, optional => 0},
                                           volume   => {type => SCALAR, optional => 0},
                                           snapshot => {type => SCALAR, optional => 0}
                                         },
                                        );

  @parsed_output =$self->CR::Snapshot2::volume_snapshot_show(%args);
  if( grep $args{snapshot} eq $_->{snapshot},@parsed_output){
    logInfo("Snapshot $args{snapshot} exist.\n");
    return 1;
  } else {
    logInfo("Snapshot $args{snapshot} does not exist.\n");
    return 0;
  }
}

##################################################################
# Volume Snapshot Promote
##################################################################

=head2 volume_snapshot_promote

The purpose of this subroutine is to promote the Active file system to $snapshot.
The subroutine waits 180 seconds for the promote functionality to Complete.

Required parameters are

=over 4

=item *
volume

name of the infinivol volume

=item *
snapshot	

Name of the snapshot

=item *
vserver

Name of the infinivol vserver

=back

=cut

sub volume_snapshot_promote{
  logStep("Calling volume snapshot promote");
  
  my($self,%args)=@_;
  my $admin_cli = $self->{ngsh};
  my ($exception_obj,$ret_value);

  my $job_name="\"volume snapshot promote\"";

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           vserver  => {type => SCALAR, optional => 0},
                                           volume   => {type => SCALAR, optional => 0},
                                           snapshot => {type => SCALAR, optional => 0}
                                         },
                                        );
  $admin_cli->volume_snapshot_promote(%args);

  #Upon failure, check if the promote job is still running, if yes, sleep for 180 secs.
  while($self->CR::Job::job_status(name =>$job_name, state => "Running")) {
    logInfo("Sleeping for 180 secs for the Promote job to complete");
    sleep(180);
  }
  $ret_value =$self->CR::Job::job_status(name =>$job_name, state => "Failure");
  if ($ret_value){
    logInfo("CLI 'volume snapshot promote' executed successfully.\n");
    return 0;
  } else {
    logInfo("CLI 'volume snapshot promote' failed to execute.\n");
    return 1;
  }
}

##################################################################
# Volume Snapshot Show
##################################################################

=head2 volume_snapshot_show

Purpose of the subroutine is to display the snapshot details.

Required parameters are

none

Optional parameters are

=over 4

=item *
instance

Selected fields to display

=item *
vserver

Name of the infinivol vserver

=item *
volume

name of the infinivol volume

=item *
snapshot

name of the snapshot

=item *
comment

comment of the snapshot

=item *
state

State of the snapshot. i.e.:valid

=item *
create-time

Creation Time of the snapshot

=item *
size

size of the snapshot

=item *
blocks

total block % of the snapshot

=item *
useblocks

use block % of the snapshot

=back

=cut

sub volume_snapshot_show{
  logStep("Calling volume snapshot show");

  my ($self,%args)=@_;
  my $admin_cli = $self->{ngsh};
  my @parsed_output;

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           instance      => {type => SCALAR, optional => 1},
                                           vserver       => {type => SCALAR, optional => 1},
                                           volume        => {type => SCALAR, optional => 1},
                                           snapshot      => {type => SCALAR, optional => 1},
                                           comment       => {type => SCALAR, optional => 1},
                                           state         => {type => SCALAR, optional => 1},
                                           'create-time' => {type => SCALAR, optional => 1},
                                           size          => {type => SCALAR, optional => 1},
                                           blocks        => {type => SCALAR, optional => 1},
                                           usedblocks    => {type => SCALAR, optional => 1}
                                         },
                                        );
  if(%args){
    @parsed_output=@{$admin_cli->volume_snapshot_show(%args)->get_parsed_output()};
    logInfo("List of infinivol snapshot on vserver $self->{vserver}:\n");
    return @parsed_output;
  } else{
    @parsed_output=@{$admin_cli->volume_snapshot_show(vserver  => $args{vserver},)->get_parsed_output()};
    logInfo("List of infinivol snapshots on vserver $self->{vserver}:\n");
    return @parsed_output;
  }
}

##################################################################
# Volume Snapshot Show Options
##################################################################

=head2 volume_snapshot_show_options

Verify "volume snapshot show -options" works as expected.

Required parameters are

=over 4

=item *
vserver

Name of the vserver.

=item *
volume

Name of the infinivol vserver.

=item *
snapshot

Snapshot name

=back

=cut

sub volume_snapshot_show_options {
  logStep("Calling volume snapshot show with options");

  my ($self, %args) =@_;
  my $admin_cli = $self->{ngsh};
  my %hash;
  my (@parsed_output,@instance);
  my ($snapshot,$Vserver,$state,$create_time,$size,$blocks,$usedblocks,$comment);
  my ($count,$option,$data,$match,$key,$value);

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           vserver  => {type => SCALAR, optional => 0},
                                           volume   => {type => SCALAR, optional => 0},
                                           snapshot => {type => SCALAR, optional => 0},
                                         },
                                        );

  try {
    @parsed_output =$self->CR::Snapshot2::volume_snapshot_show(%args);
  }
  catch NACL::APISet::Exceptions::ResponseException with {
    my $ex = shift;

    if ($ex) {
      logInfo("NACL::APISet::Exceptions::ResponseException : ".$ex->text()."");
    }
  };

  foreach (@parsed_output) {
    logInfo("volume snapshot show -vserver $args{vserver} -snapshot $args{snapshot}:\n".
            "Snapshot name  : $_->{snapshot}\n".
            "Volume         : $_->{volume}\n".
            "Vserver        : $_->{vserver}\n".
            "Snapshot state : $_->{state}\n".
            "Create time    : $_->{'create-time'}\n".
            "Size           : $_->{size}\n".
            "Total Blocks % : $_->{blocks}\n".
            "Used Blocks %  : $_->{usedblocks}\n".
            "Comment        : $_->{comment}\n");

    $snapshot   = $_->{snapshot};
    $Vserver    = $_->{vserver};
    $state  = $_->{state};
    $create_time  = $_->{'create-time'};
    $size       = $_->{size};
    $blocks     = $_->{blocks};
    $usedblocks = $_->{usedblocks};
    $comment    = $_->{comment};
  }

  for ($count=1; $count<=9; $count++) {
    $match = 0;

    if($count==1) { $option="instance";     $data="-vserver $args{vserver} -snapshot $args{snapshot}" }
    if($count==2) { $option="vserver";      $data="$args{vserver}" }
    if($count==3) { $option="snapshot";     $data="$args{snapshot}" }
    if($count==4) { $option="comment";      $data="$comment" }
    if($count==5) { $option="state";        $data="$state" }
    if($count==6) { $option="size";         $data="$size" }
    if($count==7) { $option="blocks";       $data="$blocks" }
    if($count==8) { $option="usedblocks"  ; $data="$usedblocks" }
    if($count==9) { $option="create-time";  $data="$create_time" }

    try {
      @parsed_output =$self->CR::Snapshot2::volume_snapshot_show($option => $data);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
      my $ex = shift;

      if ($ex) {
        logInfo("NACL::APISet::Exceptions::ResponseException : ".$ex->text()."");
      }
    };
  
    if($count==1) {
      %hash = ();
        
      $hash{'vserver'}  = $args{vserver};
      $hash{'snapshot'} = $args{snapshot};
      foreach (@parsed_output) {
        while (($key, $value)=each(%$_)) {
          $value =~ s/^\s+|\s+$//g;
          @instance = split(/: /, $value);

          if($instance[0] eq "Snapshot Size")              { $instance[0]="size" }
          if($instance[0] eq "Snapshot state")             { $instance[0]="state" }
          if($instance[0] eq "Percentage of Total Blocks") { $instance[0]="blocks" }
          if($instance[0] eq "Comment")                    { $instance[0]="comment" }
          if($instance[0] eq "Percentage of Used Blocks")  { $instance[0]="usedblocks" }
          if($instance[0] eq "Create time of snapshot")    { $instance[0]="create-time";
                                                             $instance[1]="\"".$instance[1]."\"";
                                                           }
        }
        $hash{$instance[0]} = $instance[1];
      }
      @parsed_output = ({%hash});
    }
 
    foreach (@parsed_output) {
      logInfo("volume snapshot show -$option $data:\n".
              "Snapshot name  : $_->{snapshot}\n".
              "Vserver        : $_->{vserver}\n".
              "Snapshot state : $_->{state}\n".
              "Create time    : $_->{'create-time'}\n".
              "Size           : $_->{size}\n".
              "Total Blocks % : $_->{blocks}\n".
              "Used Blocks %  : $_->{usedblocks}\n".
              "Comment        : $_->{comment}\n");

      if ($_->{snapshot}   eq $snapshot &&
          $_->{vserver}    eq $Vserver &&
          $_->{state}  eq $state &&
          $_->{'create-time'}  eq $create_time &&
          $_->{size}       eq $size &&
          $_->{blocks}     eq $blocks &&
          $_->{usedblocks} eq $usedblocks &&
          $_->{comment}    eq $comment) {
        $match=1;
        logInfo("'volume snapshot show -$option $data' show correct snapshot");
        last;
      }
    }
    if($match==0) {
      logInfo("'volume snapshot show -$option $data' show incorrect snapshot");
      return 0;
    }
  }
  return 1;
}

##################################################################
# Volume Snapshot Verification
##################################################################

=head2 volume_snapshot_verify

Verify "volume snapshot show for infinivol" consistent with "volume snapshot show".
Checking snapshot size, used block % and total block %.
If it does not consistent, throws an error and reports failure.

Required parameters are

=over 4

=item *
vserver

Name of the vserver.

=item *
volume

Name of the infinivol volume.

=item *
snapshot

Snapshot name

=back

=cut

sub volume_snapshot_verify{
  logStep("Verify infinivol and object volume snapshot info/calculation is correct");

  my ($self, %args) =@_;
  my $admin_cli = $self->{ngsh};
  my ($size1,$total1,$used1);
  my ($size2,$total2,$used2);
  my ($size,$total,$used);
  my ($rep_volname,$volname,$vol,$error);
  my (@parsed_output,@parsed_output2);

  %args = NACL::APISet::validate_params ($self,
                                         params => \%args,
                                         spec   => {
                                           vserver  => {type => SCALAR, optional => 0},
                                           volume   => {type => SCALAR, optional => 0},
                                           snapshot => {type => SCALAR, optional => 0},
                                         },
                                        );

  @parsed_output=$self->CR::Snapshot2::volume_snapshot_show(%args);

  foreach (@parsed_output) {
    $rep_volname = $_->{volume};
    $size1       = $_->{size};
    $total1      = $_->{blocks};
    $used1       = $_->{usedblocks};
    logInfo("Infinivol volume snapshot : \n".
            "Volume        : $rep_volname\n".
            "Size          : $size1\n".
            "Total Block % : $total1\n".
            "Used Block %  : $used1\n");
  }

$total1 =~ s/%//g;
  $used1  =~ s/%//g;

  delete($args{volume});
  @parsed_output=$self->CR::Snapshot2::volume_snapshot_show(%args);

  $size2 = 0;
  $total2 = 0;
  $used2 = 0;

  foreach $vol (@parsed_output) {
    if ($vol->{volume} eq "$rep_volname") { next; }
    $size    = $vol->{size};
    $used    = $vol->{usedblocks};
    $volname = $vol->{volume};

    $used  =~ s/%//g;

    #print "\nVolume Name : $volname";
    #print "\nVolume Size : $size";
    #print "\nUsed Block% : $used\n";
    
    $size2 = $size2 + $size;

    @parsed_output2=$self->CR::Volume::volume_show(vserver=>$args{vserver},
                                                     volume=>$volname);
    foreach (@parsed_output2) {
      if($total1==0) {
        $total2   = $total2 + $_->{size};
      } else {
        $total2 = $total2 + $size*100/$total;
      }
 
      if($used==0) {
        $used2 = $used2 + $_->{"used-afs"};
      } else {
        $used2 = $used2 + ($size*100/$used - $size);
      }
    }
  }

  if($used2==0) {
    logInfo("Error, snapshot block used size cannot be 0 for all constituent volumes");
    return 0;
  }

  $total2 = $size2/($total2+$size2)*100;
  $used2 = $size2/($used2+$size2)*100;

  logInfo("\nInfinivol volume snapshot show\n".
          "Size           : $size1\n".
          "Total Blocks % : $total1\n".
          "Used Blocks  % : $used1\n\n".
          "Sum of volume snapshot show\n".
          "Size           : $size2\n".
          "Total Blocks % : $total2\n".
          "Used Blocks  % : $used2");

  $total2 = round($total2);
  $used2 = round($used2);

  if($size1==$size2 && $total1==$total2 && $used1==$used2) {
    logInfo("Verify infinivol and constituent volumes snapshot $args{snapshot} info/calculation - PASSED.\n");
    return 1;
  } else {
    if($size1  != $size2)  { logInfo("Infinivol Snapshot Size and Object Volumes Snapshot Size compare failed.\n") }
    if($total1 != $total2) { logInfo("Infinivol Total Blocks % and Object Volumes Total Blocks % compare failed.\n") }
    if($used1  != $used2)  { logInfo("Infinivol Used Blocks % and Object Volumes Used Blocks % compare failed.\n") }
    logInfo("Verify infinivol and constituent volumes snapshot $args{snapshot} info/calculation - FAILED.\n");
    return 0;
  }
}

##################################################################
# Volume Snapshot Delete
##################################################################
#
#=head2 volume_snapshot_delete
#
#The purpose of this subroutine is to delete $snapshot from volume
#
#Optional parameters are
#
#=over 4
#
#=item *
#snapshot
#Name of the snapshot to be deleted
#
#=item *
#vserver
#
#Name of the infinivol vserver
#
#=back
#
#=cut
#
#sub volume_snapshot_delete {
#  logStep("Calling volume snapshot delete");
#
#  my($self,%args)=@_;
#  my $admin_cli = $self->{ngsh};
#  my (@parsed_output);
#  my $success;
#
#  %args = NACL::APISet::validate_params ($self,
#                                         params => \%args,
#                                         spec   => {
#                                           vserver     => {type => SCALAR, optional => 1},
#                                           snapshot    => {type => SCALAR, optional => 1}
#                                         },
#                                        );
#
#  try {
#    $success=$admin_cli->volume_snapshot_delete(%args,
#                                                'connectrec-timeout'=>DEL_TIMEOUT);
#    $success=1;
#  }
#  catch NACL::APISet::Exceptions::ResponseException with {
#    my $ex = shift;
#
#    if ($ex =~ m/no entries matching/g) {
#      $success=1;
#    } elsif ($ex) {
#      $success=0;
#      $ex->throw();
#    }
#  };
#
#  if ($success==1) {
#    ## Removing element foreground as it's not supported by volume snapshot show
#    delete($args{foreground});
#    
#    try {
#      @parsed_output =$self->CR::Snapshot2::volume_snapshot_show(%args);
#      $success=0;
#    }
#    catch NACL::APISet::Exceptions::ResponseException with {
#      my $ex = shift;
#
#      if (($ex =~ m/no entries matching/g) || ($ex =~ m/currently empty/g)) {
#        $success=1;
#      } elsif ($ex) {
#        $success=0;
#        $ex->throw();
#      }
#    };
#
#    if ($success==1) {
#      logInfo("CLI 'volume snapshot delete' executed successfully.\n");
#      return 1;
#    } else {
#      logInfo("CLI 'volume snapshot delete' failed to execute.\n");
#      return 0;
#    }
#  }
#}

##################################################################
# Volume Snapshot Show
##################################################################
#
#=head2 volume_snapshot_show
#
#Purpose of the subroutine is to display the snapshot details.
#
#Optional parameters are
#
#=over 4
#
#=item *
#vserver
#
#Name of the infinivol vserver
#
#=item *
#snapshot
#
#name of the snapshot
#
#=back
#
#=cut
#
#sub volume_snapshot_show{
#  logStep("Calling volume snapshot show");
#
#  my ($self,%args)=@_;
#  my $admin_cli = $self->{ngsh};
#  my @parsed_output;
#
#  %args = NACL::APISet::validate_params ($self,
#                                         params => \%args,
#                                         spec   => {
#                                           vserver => {type => SCALAR, optional => 1},
#                                           snapshot => {type => SCALAR, optional => 1},
#                                         },
#                                        );
#
#  delete($args{volume});
#
#  if($args{snapshot}){
#    @parsed_output=@{$admin_cli->volume_snapshot_show(%args)->get_parsed_output()};
#    logInfo("List of volume snapshot on vserver $self->{vserver}:\n");
#    return @parsed_output;
#  } else{
#    @parsed_output=@{$admin_cli->volume_snapshot_show(vserver=>$args{vserver})->get_parsed_output()};
#    logInfo("List of volumes snapshot on vserver $self->{vserver}:\n");
#    return @parsed_output;
#  }
#}

