# Copyright (c) 2011 Network Appliance, Inc.
# All rights reserved.

## @summary unit test library
## @author  richard.chow@netapp.com
## @pod     here

=head1 NAME

CR::UnitTest

=head1 DESCRIPTION

This module contains methods for beacon CLI unit tests.

=head1 METHODS

=cut

package CR::UnitTest;

# Compiler directives.
use strict;
use warnings;

# Module import.
use base qw(TCD);
use NATE::BaseException qw(:try);

use TCD;
use Params::Validate qw(validate_with SCALAR HASHREF ARRAYREF);
use NACL::APISet;
use NACL::C::Node;
use NACL::C::Aggregate;
use NACL::C::VolumeSnapshot;
use NACL::C::NetworkInterface;
use NACL::C::Interface;
use NACL::C::Route;
use NACL::C::Job;
use NACL::C::JobHistory;
use Options;
use CR::CommonParams;
use NACL::STask::Aggregate;
use NACL::STask::Vserver;
use NACL::STask::Volume;
use NACL::C::Disk;
use NACL::C::ExportPolicyRule;
use CR::Setup::Constants;

use List::Util qw(max);

our @EXPORT = qw(
    getCli
    getCliForNode
    getAdminCli
    getAdvancedCli
    getTestCli
    getZapi
    getShell
    getNode
    getRowByField
    getRowsByField
    process_subtest
    verify
    testcases
    reset_rw_repo
    validate_zapi_job_id
    is_searchlight
    is_beacon
    is_object_store
    create_aggregates
    delete_temp_aggregates
    create_data_lif
    find_or_create_vserver
    find_or_create_infinitevol
    create_export_rule
    find_create_empty_aggregates_for_nodes
);

# Global Variables
my $Options;
my @Temp_Aggr_List;
my %Aggr_Owner_Map;
my $Cached_Node;

########## Exported Functions  ###################################################

=head2 getCli

This function creates a CMode CLI with default privilege (diag)  and returns it.

=cut

sub getCli {
    my $node = shift;
    unless ( defined $node && $node->isa("NACL::C::Node") ) {
        $node = NACL::C::Node->find();
    }

    my $cli = getCliForNode($node);
    $cli->set_privilege( "privilege-level" => "diag" );

    return $cli;
}

=head2 getCliForNode

This function creates an APISet with CLI interface, default privilege (diag), and returns it.

=cut

sub getCliForNode {
    my $node = shift;

    my $cli = NACL::APISet->new(
        hostobj      => $node->hostrec(),
        category     => 'Node',
        interface    => 'CLI',
        set          => 'CMode',
    );

    unless ( defined($Options) ) {
        $Options = new Options;
        $Options->process();
    }

    $cli->set_timeout(
        'connectrec-timeout' => $Options->get( option => "TIMEOUT" ) );
    return $cli;
}

=head2 getAdminCli

This function creates an APISet with CLI interface , set the privilege to admin and returns it.

=cut

sub getAdminCli {
    my $node = shift;
    unless ( defined $node && $node->isa("NACL::C::Node") ) {
        $node = NACL::C::Node->find();
    }

    my $cli = getCliForNode($node);
    $cli->set_privilege( "privilege-level" => "admin" );

    return $cli;
}

=head2 getAdvancedCli

This function creates an APISet with CLI interface , set the privilege to advanced and returns it.

=cut

sub getAdvancedCli {
    my $node = shift;
    unless ( defined $node && $node->isa("NACL::C::Node") ) {
        $node = NACL::C::Node->find();
    }

    my $cli = getCliForNode($node);
    $cli->set_privilege( "privilege-level" => "advanced" );

    return $cli;
}

=head2 getTestCli

This function creates an APISet with CLI interface , set the privilege to test and returns it.

=cut

sub getTestCli {
    my $node = shift;
    unless ( defined $node && $node->isa("NACL::C::Node") ) {
        $node = NACL::C::Node->find();
    }

    my $cli = getCliForNode($node);
    $cli->set_privilege( "privilege-level" => "test" );

    return $cli;
}

=head2 getZapi

This function creates an APISet with ZAPI interface and returns it.

=cut

sub getZapi {
    my $node = shift;
    unless ( defined $node && $node->isa("NACL::C::Node") ) {
        $node = NACL::C::Node->find();
    }

    $TCD::Log->warn('APISET_CDEFS_FROM_BUILD=1 is not supported')
        if $ENV{APISET_CDEFS_FROM_BUILD};

    return NACL::APISet->new(
        hostobj      => $node->hostrec(),
        category     => 'Node',
        interface    => 'ZAPI',
        set          => 'CMode',
        version_info => { "cdefs_from_build" => 0 }
    );
}

=head2 getShell

This function create an APISet with SystemShell interface and returns it.

=cut

sub getShell {
    my $node = shift;
    unless ( defined $node && $node->isa("NACL::C::Node") ) {
        $node = NACL::C::Node->find();
    }

    return NACL::APISet->new(
        hostobj      => $node->hostrec(),
        category     => 'Node',
        interface    => 'CLI',
        set          => 'Systemshell',
        shell_type   => 'CMode',
        version_info => { "cdefs_from_build" => 0 },
    );
}

=head2 getNode

This function creates a node object and returns it.

=cut

sub getNode {
    my @nodes = NACL::C::Node->find();
    return wantarray ? @nodes : $nodes[0];
}

=head2 getRowByField

This function returns the first row of the parsed output of a response object
fileterd by field specified.

=cut

sub getRowByField {
    my $value;
    my $field;

    my @parsed_output = _getOutputByField( \$field, @_ );

    eval { $value = $parsed_output[0]->{$field}; };

    $value = "" if ($@);
    return $value;
}

=head2 getRowsByField

This function returns all rows of the parsed output of a response object
fileterd by field specified.

=cut

sub getRowsByField {
    my $value;
    my $field;

    my @parsed_output = _getOutputByField( \$field, @_ );
    eval {
        for ( my $i = 0; $i < @parsed_output; $i++ )
        {
            $value = $value . "," if $i != 0;
            $value = $value . $parsed_output[$i]->{$field};
        }
    };

    $value = "" if ($@);
    return $value;
}

=head2 process_subtest

This function overwrites Tharn::subtest().

=cut

sub process_subtest(@) {
    my @params = @_;
    push( @params, -onexit => sub { } );
    my $subtest = Tharn::subtest(@params);
    my $status  = $TCD::PASS;
    my @results = $subtest->rslts();
    foreach my $result (@results) {
        if ( $result->type() =~ /FAIL|SCRIPT|CONFIG/i ) {
            $status = $TCD::FAIL;
        }
    }
    return $status;
}

=head2 verify

This function performs verification testing on a test case.

=cut

sub verify {
    my $cli    = shift;
    my $cmd    = shift;
    my %params = @_;

    my $should_pass = delete( $params{'-should_pass'} );
    my $should_fail = delete( $params{'-should_fail'} );
    my $should_wait = delete( $params{'-should_wait'} );

    if ( $should_pass && $should_fail ) {
        NATE::BaseException->throw(
                  "Only one requirement can be specified at a same time: "
                . "'-should_pass', '-should_fail', or 'should_wait' " );
    }

    if (   defined( Tharn::param('USE_OLD_ALGORITHM') )
        && Tharn::param('USE_OLD_ALGORITHM') =~ /1/
        && $cmd                              =~ m/volume_create/
        && $cli->{'interface'}               =~ m/cli/i )
    {
        $cmd = 'repository_volume_create';
        $params{'foreground'} = 'true'
            unless $params{'foreground'} =~ /false/;
        $params{'privilege-level'}   = 'test';
        $params{'use-new-algorithm'} = 'false';
    }

    return _positiveTest( $cli, $cmd, %params ) if $should_pass;
    return _negativeTest( $cli, $cmd, %params ) if $should_fail;
    _waitfor( $cli, $cmd, %params ) if $should_wait;

    if ( !$should_pass && !$should_fail && !$should_wait ) {
        NATE::BaseException->throw(
                  "A requirement, '-should_pass', '-should_fail', "
                . "or '-should_wait', shall be specified" );
    }
}

=head2 create_aggregates

This function create list of aggregates, aggregate create by
this function will be delete by functions delete_temp_aggregates

Input:
    command_interface
    number_of_aggregates => 1,
    aggregate_suffix     => "tmp_aggr"
    nodes                => [ 'node1' ],
    nacltask_min_size    => '1GB',
    purge_volumes        => 0,

=cut

sub create_aggregates {
    my %params = validate_with(
        params => \@_,
        spec   => {
            command_interface      => { type => HASHREF },
            create_specify_options => {
                type     => HASHREF,
                optional => 1,
            },
            number_of_aggregates   => {
                type    => SCALAR,
                default => 1,            # Default to 1 aggregate
                regex   => qr/^\d+$/,    # This must be a numeric value
            },
            nodes             => { type => ARRAYREF, optional => 1 },
            nacltask_min_size => { type => SCALAR, optional => 1 },
            purge_volumes     => { type => SCALAR, default => 0 },
            aggregate_suffix  => { type => SCALAR, default => 'tmp_aggr' },
        }
    );

    my @result_list;
    my $test_name = $main::Test->get_testcase();

    if( defined($params{create_specify_options}) ) {
        $TCD::Log->warn(
            "The option 'create_specify_options' has been deprecated."
            . "\nUse the option 'nodes' if you want to specify which nodes to"
            . " create aggregates on."
            . "\nUse the option 'nacltask_min_size' if you need an aggregate of"
            . " certain size."
        );
    }

    # The minimum number of disks for an aggregate changes based on the build type
    my $cli = getTestCli();
    my $response
        = $cli->system_node_image_show( fields => 'build-type');
    my $build_type = $response->{parsed_output}[0]->{'build-type'};

    my $raid_dp_disks = ($build_type =~ /DEBUG/i) ? 3 : 5;
    my $raid4_disks = ($build_type =~ /DEBUG/i) ? 2 : 3;

    my %create_options;
    $create_options{nodes} = $params{nodes};
    $create_options{nodes} = $params{create_specify_options}->{nodes}
        unless defined $create_options{nodes};
    my $auto_detect_node = !defined($create_options{nodes});

    # MSATA disks are not allow in raid4 aggregates so modify the defaults based
    # on whether they are present.
    my $msata_disks = NACL::C::Disk->find(
        command_interface => $params{command_interface},
        filter            => {
            type => 'msata',
        },
        allow_empty       => 1,
    );
    if ($msata_disks) {
        # Override with raid_dp and increase the diskcount by 1
        $create_options{raidtype} = 'raid_dp';
        #Ensure the disk count is at least 3
        $create_options{diskcount} = max(
            $params{create_specify_options}->{diskcount} + 1,
            $raid_dp_disks
        );
    } else {
        # Default to raid4 with 2 disks
        if ( defined( $params{create_specify_options}->{raidtype} ) ) {
            $create_options{raidtype} = $params{create_specify_options}->{raidtype};
        } else {
            $create_options{raidtype} = 'raid4';
        }
        if ( defined( $params{create_specify_options}->{diskcount} ) ) {
            $create_options{diskcount} = $params{create_specify_options}->{diskcount};
        } else {
            $create_options{diskcount} = $raid4_disks;
        }
    }
    $create_options{'method-timeout'} = $Options->get( 'option' => 'TIMEOUT' );

    my $aggr_suffix = 0;
    for ( my $i = 1; $params{number_of_aggregates} >= $i; $i++ ) {
        my $created;
        $aggr_suffix++;
        my $aggr_name = $params{aggregate_suffix} . $aggr_suffix;
        while ( $Aggr_Owner_Map{$aggr_name} =~ m/^$test_name$/ ) {
            $aggr_suffix++;
            $aggr_name = $params{aggregate_suffix} . $aggr_suffix;
        }
        if ( $auto_detect_node ) {
            my $create_node = _find_node_with_most_spare_disk(
                command_interface => $params{command_interface} );
            $create_options{nodes} = [$create_node];
        }

        # STask by default will wait for aggregate create successfully.
        my $aggr = NACL::STask::Aggregate->find_or_create(
            command_interface       => $params{'command_interface'},
            create_specific_options => \%create_options,
            aggregate               => $aggr_name,
            _was_created            => \$created,
        );
        $aggr->grow_size( nacltask_min_size => $params{nacltask_min_size} )
            if defined( $params{nacltask_min_size} );
        $TCD::Log->log('Fetching aggregate information for debug purposes');
        $aggr->state( requested_fields => [ 'size','diskcount','raidtype','nodes' ] );
        push( @result_list, $aggr );
        if ( $created ) {
            push( @Temp_Aggr_List, $aggr_name );
        }
        $Aggr_Owner_Map{$aggr_name} = $test_name;
        if ( $params{purge_volumes} ) {
            _purge_constituent_volumes(
                command_interface => $params{command_interface},
                aggregate         => $aggr,
            );
        }
    }
    return wantarray ? @result_list : $result_list[0];
}

=head2 _find_node_with_most_spare_disk

This function find node contain most spare disk

Input:
    command_interface

=cut

sub _find_node_with_most_spare_disk {
    my %params = validate_with(
        params => \@_,
        spec   => { command_interface => { type => HASHREF }, }
    );

    my @nodes = NACL::C::Node->find(
        'command_interface' => $params{'command_interface'}, );
    my $empty_disk_count = 0;
    my $node;
    foreach (@nodes) {
        my @disks = NACL::C::Disk->find(
            'command_interface' => $params{'command_interface'},
            'filter'            => {
                'owner'          => $_->node,
                'container-type' => 'spare',
            },
            'allow_empty'       => 1,
        );
        if ( scalar(@disks) > $empty_disk_count ) {
            $empty_disk_count = scalar(@disks);
            $node             = $_->node;
        }
    }
    NATE::BaseException->throw( 'Could not find a node that has spare disks.' )
        if ( $empty_disk_count == 0 );

    return $node;
}

=head2 delete_temp_aggregates

This functions delete all tempory created aggregates
This in general should only be called during uninit

Input:
    command_interface

=cut

sub delete_temp_aggregates {
    my %params = validate_with(
        params => \@_,
        spec   => { command_interface => { type => HASHREF }, },
    );
    @Temp_Aggr_List = sort(@Temp_Aggr_List);
    my $prev_aggr = "";
    while ( scalar(@Temp_Aggr_List) > 0 ) {
        my $aggr = pop(@Temp_Aggr_List);
        unless ( $aggr eq $prev_aggr ) {
            NACL::STask::Aggregate->purge(
                command_interface => $params{'command_interface'},
                'aggregate'          => $aggr,
                'nacltask_if_purged' => 'pass',
                'method-timeout'     => $Options->get( 'option' => 'TIMEOUT' ),

            );
            $prev_aggr = $aggr;
        }
    }
    if ( $params{'command_interface'}->hostrec()->hosttype() !~ m/vsim/i ) {
        NACL::C::Disk->zerospares(
            command_interface => $params{'command_interface'} );
    }
    %Aggr_Owner_Map = ();
}

=head2 find_or_create_vserver

This function checks whether a Vserver matching the specified criteria is present
and if so returns it. If there is no Vserver found matching the specified criteria,
then a new Vserver of that type is created and returned to the user.
Input:
    %options_common_to_find_and_create,
=cut

sub find_or_create_vserver {
    my %parameters = @_;
    my $node = delete $parameters{'node'};
    unless ( defined $node && $node->isa("NACL::C::Node") ) {
        $node = get_cached_node();
    }

    my $timeout = delete $parameters{'method-timeout'};
    $timeout = $Options->get( 'option' => 'TIMEOUT' )
        unless $timeout;

    $parameters{'vserver'} = $Options->get( option => 'VSERVER' )
        unless ( exists $parameters{'vserver'} );

    $parameters{'aggregate'} = $Options->get( option => 'VSERVER_AGGR' )
        unless ( exists $parameters{'aggregate'} );

    $parameters{'ns-switch'} = $Options->get( option => 'NS_SWITCH' )
        unless ( exists $parameters{'ns-switch'} );

    $parameters{'is-repository'} = 'true'
        unless ( exists $parameters{'is-repository'} );

    $parameters{'rootvolume-security-style'}
        = $Options->get( option => 'SECURITY_STYLE' )
        unless ( exists $parameters{'rootvolume-security-style'} );

    $parameters{'rootvolume'} = $parameters{'vserver'} . '_root'
        unless ( exists $parameters{'rootvolume'} );

    my $vs = NACL::STask::Vserver->find(
        command_interface => $node,
        filter            => { 'vserver' => $parameters{'vserver'}, },
        allow_empty       => 1
    );

    if ($vs) {
        foreach my $k ( keys %parameters ) {
            my $value   = $parameters{$k};
            my $setting = $vs->get_one_state_attribute($k);

            $value   = join( ',', @$value )   if ref($value)   eq "ARRAY";
            $setting = join( ',', @$setting ) if ref($setting) eq "ARRAY";

            if ( $setting ne $value ) {
                $main::Log->comment( $setting . "not equal to " . $value );
                $vs->purge( 'method-timeout'  => $timeout );
                undef $vs;
                last;
            }
        }
    }

    $vs = NACL::STask::Vserver->create(
        command_interface => $node,
        'method-timeout'  => $timeout,
        %parameters
    ) unless $vs;

    return $vs;
}

=head2 find_or_create_infinitevol

This function checks whether an InfiniteVol matching the specified criteria is present
and if so returns it. If there is none found matching the specified criteria,
then a new InfinteVol of that type is created and returned to the user.
Input:
    %options_common_to_find_and_create,
=cut

sub find_or_create_infinitevol {
    my %parameters = @_;

    my $node = delete $parameters{'node'};
    unless ( defined $node && $node->isa("NACL::C::Node") ) {
        $node = get_cached_node();
    }

    my $timeout = delete $parameters{'method-timeout'};
    $timeout = $Options->get( 'option' => 'TIMEOUT' )
        unless $timeout;

    $parameters{'vserver'} = $Options->get( 'option' => 'VSERVER' )
        unless ( exists $parameters{'vserver'} );

    $parameters{'volume'} = $Options->get( 'option' => 'CVOL_NAME' )
        unless ( exists $parameters{'volume'} );

    $parameters{'snapshot-policy'}
        = $Options->get( 'option' => 'SNAPSHOT_POLICY' )
        unless ( exists $parameters{'snapshot-policy'} );

    $parameters{'junction-path'}
        = $Options->get( 'option' => 'CVOL_JUNCTION_PATH' )
        unless ( exists $parameters{'junction-path'} or
                 $parameters{'type'} =~ /DP/i );

    my $iv = NACL::STask::Volume->find(
        command_interface => $node,
        filter            => {
            'vserver'      => $parameters{'vserver'},
            'volume-style' => 'infinitevol'
        },
        allow_empty => 1
    );

    if ($iv) {
        foreach my $k ( keys %parameters ) {
            if ( $iv->get_one_state_attribute($k) ne $parameters{$k} ) {
                $iv->purge( 'method-timeout'  => $timeout );
                undef $iv;
                last;
            }
        }
    }

    $iv = NACL::STask::Volume->create(
        command_interface => $node,
        'method-timeout'  => $timeout,
        %parameters
    ) unless $iv;

    return $iv;
}

=head2 create_export_rule
This function creates a default export policy for testing
=cut

sub create_export_rule {
    my $policy_name = 'default';
    my $node        = getNode();

    NACL::C::ExportPolicyRule->create(
        command_interface => $node,
        'policyname'      => $policy_name,
        'clientmatch'     => '0.0.0.0/0',
        'rorule'          => ['any'],
        'rwrule'          => ['any'],
        'protocol'        => ['any'],
        'vserver'         => $Options->get( option => 'VSERVER' ),
        'anon'            => '0',
        'superuser'       => ['any'],
        'ruleindex'       => '1',
        'allow-suid'      => 'true',
        'allow-dev'       => 'true',
    );
    return $policy_name;
}

# add support for DISABLE_TESTCASES and versioning
sub testcases {
    $Options = new Options;
    $Options->parameters(
        definitions => [
            {   name        => "DISABLE_TESTCASES",
                description => "Array of test cases we want to disable",
                spec        => "s@",
            },
        ]
    );
    $Options->process();
    my @test_cases = @_;
    my @disabled_test = @{ $Options->get( option => "DISABLE_TESTCASES" ) };
    my @new_tests;
    my %d_hash;
    @d_hash{@disabled_test} = (1) x @disabled_test;

    my $ver_info = _get_version_info();

    while (@test_cases) {
        my $test_name = shift(@test_cases);
        my $test_info = shift(@test_cases);
        $test_name = '-' . $test_name if ( $d_hash{$test_name} );

        # handle disabled tests
        if ( $d_hash{$test_name} ) {
            $test_name = '-' . $test_name;
            push( @new_tests, $test_name, $test_info );
        }
        elsif ( ref($test_info) eq "HASH" ) {

            # check if test is compatible with the node's build
            my $run_status = _determine_run_status(
                builds_for_test => $test_info->{build},
                build_on_node   => $ver_info->{build},
            );

            if ( $run_status == $TRUE ) {
                push( @new_tests, $test_name, $test_info->{desc} );
            }
        }
        else {

            # test does not apply to specific build so run it
            push( @new_tests, $test_name, $test_info );
        }
    }

    TCD::testcases(@new_tests);
}

# reset RW repository to original state / settings
sub reset_rw_repo {
    my %parameters = validate_with(
        params => \@_,
        spec   => {
            command_interface => { type => HASHREF },
            vserver           => { type => SCALAR }
        }
    );

    my $node    = $parameters{'command_interface'};
    my $repo_vs = $parameters{'vserver'};
    my $repo;
    my $repo_ok;
    my $min_repo_vols = 3;

    my $default_vol        = $Options->get( option => 'CVOL_NAME' );
    my $default_jp         = $Options->get( option => 'CVOL_JUNCTION_PATH' );
    my $default_permission = $Options->get( option => 'UNIX_PERMISSIONS' );

    # get repository
    $repo = NACL::STask::Volume->find(
        command_interface => $node,
        filter            => {
            'vserver'      => $repo_vs,
            'volume-style' => 'infinitevol'
        },
        allow_empty => 1,
    );
    if($repo)
    {
        # ensure all constituents are online
        my @repo_vols = NACL::STask::Volume->find(
            command_interface => $node,
            filter            => { 'vserver' => $repo_vs }
        );

        if ( @repo_vols >= $min_repo_vols ) {
            for ( my $i = 0; $i < @repo_vols; $i++ ) {
                $repo_vols[$i]->online()
                    if $repo_vols[$i]->get_one_state_attribute('state') ne
                        'online';
            }
            $repo_ok = 1;
        }

        # delete all snapshots
        if ($repo_ok) {
            my @repo_snapshots = NACL::C::VolumeSnapshot->find(
                command_interface => $node,
                filter            => {
                    'vserver' => $repo_vs,
                    'volume'  => $repo->volume,
                },
                allow_empty => 1
            );

            foreach my $snapshot (@repo_snapshots) {
                $snapshot->delete( 'foreground' => 'true' );
            }
        }

        if ($repo_ok) {

            # mount to original junction path
            if ( $repo->get_one_state_attribute('state') ne 'online' ) {
                $repo->online(
                    'method-timeout' => $Options->get( option => "TIMEOUT" ),
                );
            }
            $repo->unmount(
                'method-timeout' => $Options->get( option => "TIMEOUT" ),
            );

            $repo->mount(
                'junction-path'  => $default_jp,
                'method-timeout' => $Options->get( option => "TIMEOUT" ),
            );

            # set unix permission to original setting
            $repo->modify(
                'unix-permissions' => $default_permission,
                'method-timeout'   => $Options->get( option => "TIMEOUT" ),
            );

            # set volume name to original value
            my $actual_vol = $repo->volume;
            if ( $actual_vol ne $default_vol ) {
                $repo->rename(
                    'newname' => $default_vol,
                    'method-timeout'   => $Options->get( option => "TIMEOUT" ),
                );

                $repo = NACL::STask::Volume->find(
                    command_interface => $node,
                    filter            => {
                        'vserver'      => $repo_vs,
                        'volume-style' => 'infinitevol'
                    }
                );
            }
        }

        if (  !$repo_ok
            || $repo->get_one_state_attribute('state') ne 'online' )
        {
            NATE::BaseException->throw('Attempts to revert the repository ' .
                'to a normal state failed, please investigate.');
        }
    }
    else
    {
        $repo = NACL::STask::Volume->create(
            command_interface  => $node,
            'vserver'          => $repo_vs,
            'volume'           => $default_vol,
            'junction-path'    => $default_jp,
            'unix-permissions' => $default_permission,
            'method-timeout'   => $Options->get( option => 'TIMEOUT' ),
        );
    }
    return $TCD::PASS;
}

#Validate the job id returned to a ZAPI call
sub validate_zapi_job_id {
    my %parameters      = @_;
    my $response        = $parameters{'response_obj'};
    my $job_description = $parameters{'job_description'};
    my $node            = $parameters{'command_interface'};
    my $wait_complete   = $parameters{'wait_complete'};
    $wait_complete = 'false' unless $wait_complete;

    my $job_done = 0;

    my $expected_state = $parameters{'expected_state'};
    $expected_state = 'Success|Queued|Running' unless $expected_state;

    my $job_tag = $parameters{'job_tag'};
    $job_tag = 'jobid' unless $job_tag;

    my $job_id;
    my $output = $response->get_raw_output();

    if ( $output =~ m/<$job_tag>(.*?)<\/$job_tag>/s ) {
        $job_id = $1;
    }

    $TCD::Log->warn("No job id found") unless $job_id;

    my $job = NACL::C::Job->find(
        command_interface => $node,
        filter            => { id => $job_id },
        allow_empty       => 1
    );

    # Try to find the job in the hisotry or poll for it if needed
    if ( !$job ) {
        $job = NACL::C::JobHistory->find(
            command_interface => $node,
            filter            => { id => $job_id },
            allow_empty       => 1
        );

        if ( !$job ) {

            # poll for the job
            my $timeout = 60;

            while ( !$job || $timeout < 60 ) {
                $job = NACL::C::Job->find(
                    'command_interface' => $node,
                    'allow_empty'       => 1,
                    'filter'            => { id => $job_id },
                );

                $timeout += 5;
                sleep(5);
            }
        }
        else {
            $job_done = 1;
        }
    }

    if ($job) {
        if ( $job->get_one_state_attribute('description')
            =~ m/$job_description/i )
        {
            return $TCD::PASS if ($job_done);
            $job->watch_progress(
                'method-timeout' => $Options->get( option => "TIMEOUT" ) )
                if ( $wait_complete eq 'true' );
            return $job->get_one_state_attribute('state')
                =~ m/$expected_state/i
                ? $TCD::PASS
                : $TCD::FAIL;
        }
        else {
            return $TCD::FAIL;
        }
    }
    else {
        return $TCD::FAIL;
    }
}

#Check if this is searchlight build
sub is_searchlight {
    return $Options->get( option => 'CR_VERSION' ) =~ m/^searchlight$/i;
}

#Check if this is beacon build
sub is_beacon {
    return $Options->get( option => 'CR_VERSION' ) =~ m/^beacon$/i;
}

sub is_object_store {
    return 0;
}

=head2 create_network_interface

This function creates a data lif on the vserver and return the lif it creates.
    It will also set up the route group. Gateway info will be fetched from tharn host
    file.

Input:

    command_interface
    vserver
    num                         => 1

=cut

sub create_data_lif {
    my %parameters = validate_with(
        params => \@_,
        spec   => {
            command_interface => { type => HASHREF },
            vserver           => {
                type    => SCALAR,
                default => $Options->get( option => 'VSERVER' ),
            },
            num => { type => SCALAR, default => 1 },
        }
    );
    my @nodes        = ();
    my $Node         = $parameters{'command_interface'};
    my @all_nodes    = NACL::C::Node->find( command_interface => $Node );
    my $data_lif_num = $parameters{'num'};
    my $vserver      = $parameters{'vserver'};

    foreach my $node (@all_nodes) {
        my $host   = $node->hostrec();
        my %params = $host->param_fields();
        my $address
            = exists( $params{'DATA_IP'} )
            ? $params{'DATA_IP'}
            : $params{'DATA_IP1'};

        my $lif = NACL::C::Interface->find(
            command_interface => $Node,
            filter            => {
                'role'    => 'data',
                'address' => $address
            },
            allow_empty => 1,
        );
        push( @nodes, $node ) if ( !$lif );
    }
    if ( @nodes == 0 ) {
        $TCD::Log->error('No data IP available');
        return $TCD::FAIL;
    }
    elsif ( $data_lif_num eq 'AUTO' ) {
        foreach my $node (@nodes) {
            _assign_data_interface(
                'node'              => $node,
                'command_interface' => $Node,
                'vserver'           => $vserver,
            );
        }
    }
    elsif ( $data_lif_num eq 1 ) {
        return _assign_data_interface(
            'node'              => $nodes[0],
            'command_interface' => $Node,
            'vserver'           => $vserver,
        );
    }
}

########## Local Functions  ###################################################

# run this test if the requirement is to wait for an expected result
sub _waitfor {
    my $cli    = shift;
    my $cmd    = shift;
    my %params = @_;

    my $warning_msg_exact = delete( $params{'-exact_warning_message'} );
    $warning_msg_exact = delete( $params{'-contained_in_output'} )
        unless $warning_msg_exact;
    my $warning_msg_like  = delete( $params{'-like_warning_message'} );
    my $negative_matching = delete( $params{'-negative_matching'} );
    my $interval          = delete( $params{'-interval'} );
    my $timeout           = delete( $params{'-timeout'} );
    my $suppress          = 'true';
    my $response;
    my $output;
    my $ok;

    $interval = 30  if !$interval;
    $timeout  = 120 if !$timeout;

    if ( !$warning_msg_exact && !$warning_msg_like ) {
        NATE::BaseException->throw(
            "A message for '-contain_in_output' " . "shall be specified" );
    }

    while ( $timeout > 0 && !$ok ) {
        $response = $cli->$cmd(%params);
        $output   = $response->{'processed_output'};               #CLI
        $output   = $response->get_raw_output() unless $output;    #ZAPI
        $ok = _processErrMsg( $warning_msg_exact, $warning_msg_like, $output,
            $suppress, $negative_matching );
        sleep $interval;
        $timeout -= $interval;
        last if $ok;
    }

    if ( !$ok ) {
        my $args;

        foreach my $k ( keys %params ) {
            $args = $args . $k . " " . $params{$k};
        }
        NATE::BaseException->throw(
            "Time out waiting for '$cmd $args' to return " . "
            $warning_msg_like $warning_msg_exact"
        );
    }
}

# run this test if the requirement is for the test to pass
sub _positiveTest {
    my $cli    = shift;
    my $cmd    = shift;
    my %params = @_;
    my $response;

    my $warning_msg_exact = delete( $params{'-exact_warning_message'} );
    $warning_msg_exact = delete( $params{'-contained_in_output'} )
        unless $warning_msg_exact;
    my $warning_msg_like = delete( $params{'-like_warning_message'} );

    $response = $cli->$cmd(%params);

    if ( $warning_msg_exact || $warning_msg_like ) {
        my $output = $response->{'processed_output'};    #CLI
        $output = $response->get_raw_output() unless $output;    #ZAPI
        _processErrMsg( $warning_msg_exact, $warning_msg_like, $output );
    }

    return $response;
}

# run this test if the requirement is for the test to fail
sub _negativeTest {
    my $cli           = shift;
    my $cmd           = shift;
    my %params        = @_;
    my $err_msg_exact = delete( $params{'-exact_error_message'} );
    my $err_msg_like  = delete( $params{'-like_error_message'} );
    my $err_msg_rxed;
    my $response;
    my $output;
    my $processed = 0;

    $main::Log->comment('********* Expect the following to fail *********');
    try {
        $response = $cli->$cmd(%params);
        $output   = $response->{'processed_output'};               #CLI
        $output   = $response->get_raw_output() unless $output;    #ZAPI
    }
    catch NATE::BaseException with {
        $err_msg_rxed = shift()->text();
        $processed    = 1;
        _processErrMsg( $err_msg_exact, $err_msg_like, $err_msg_rxed );

       # we only care about returning error message if we want to extract some
       # info from it (_processErrMsg will still throw an exception if the
       # user is validating the msg for correctness)
        return $err_msg_rxed;
    }
    finally {
        if ( $processed == 0 && ( $err_msg_exact || $err_msg_like ) ) {
            $output = $err_msg_rxed if $err_msg_rxed;
            _processErrMsg( $err_msg_exact, $err_msg_like, $output );
        }
        elsif ( $processed == 0 ) {
            NATE::BaseException->throw("This test shall fail, but it passed");
        }
    };
}

# process error message
sub _processErrMsg {
    my $err_msg_exact  = shift;
    my $err_msg_like   = shift;
    my $err_msg_rxed   = shift;
    my $suppress       = shift;
    my $negative_match = shift;

# remove new line characters, as regular expression only do pattern match on one line
    $err_msg_rxed =~ s/\n//g if $err_msg_rxed;

# remove leading 'X' on each word, as each word is enclosed with leading and trailing
# "X" when returned from API response
    $err_msg_rxed =~ s/X\b//g if $err_msg_rxed;
    $err_msg_rxed =~ s/\bX//g if $err_msg_rxed;
    $err_msg_exact = _trimStr($err_msg_exact);
    $err_msg_like  = _trimStr($err_msg_like);
    $err_msg_like =~ s/,/|/g if $err_msg_like;
    $err_msg_like = '\b' . '(' . $err_msg_like . ')' . '\b'
        if $err_msg_like;

    #'^(?:(?!\b(Active)\b).)*$\r?\n?'

    if ($negative_match) {
        $err_msg_exact = '^(?:(?!' . $err_msg_exact . ').)*$\r?\n?'
            if $err_msg_exact;
        $err_msg_like = '^(?:(?!' . $err_msg_like . ').)*$\r?\n?'
            if $err_msg_like;
    }

    if ($err_msg_exact) {
        return _runRegEx(
            $err_msg_rxed,
            $err_msg_exact,
            "Expecting error message: $err_msg_exact, \n"
                . "but received error message: $err_msg_rxed",
            $suppress
        );
    }
    elsif ($err_msg_like) {
        my @err_msgs_rxed = split( /\n/, $err_msg_rxed );
        return _runRegEx(
            $err_msgs_rxed[0],
            $err_msg_like,
            "Expecting error message key words: $err_msg_like, \n "
                . "not contained in the received error message: $err_msgs_rxed[0]",
            $suppress
        );
    }
}

# remove leading and trailing spaces from a string
sub _trimStr {
    my $str = shift;
    $str =~ s/^\s+// if $str;
    $str =~ s/\s+$// if $str;
    return $str;
}

# throw an exception if a pattern is not matched in the message
sub _runRegEx {
    my $message  = shift;
    my $pattern  = shift;
    my $throwMsg = shift;
    my $suppress = shift;
    my $ok;

    $ok = ( $message =~ m/$pattern/ );
    if ( !$suppress && !$ok ) {
        NATE::BaseException->throw("$throwMsg");
    }

    return $ok;
}

# get parsed output of a response object
sub _getOutputByField {
    my $field    = shift;
    my $response = shift;
    my $column   = shift;
    my %filter   = @_;

    ${$field} = $column;
    my @parsed_output = @{ $response->get_parsed_output() };

    foreach my $key ( keys %filter ) {
        @parsed_output = grep( $_->{$key} eq $filter{$key}, @parsed_output );
    }

    return @parsed_output;
}

# Helper functions to purge volumes inside a aggregate
sub _purge_constituent_volumes {
    my %params = validate_with(
        params => \@_,
        spec   => {
            command_interface => { type => HASHREF },
            aggregate         => { type => HASHREF | SCALAR },
        }
    );
    my @vols = NACL::STask::Volume->find(
        'command_interface' => $params{'command_interface'},
        'filter'            => { aggregate => $params{'aggregate'}, },
        'allow_empty'       => 1,
    );
    foreach (@vols) {
        $_->purge();
    }
}

# Create a data interface based on the information provided by the tharn host
# , and only DATA_IP or DATA_IP1 are used
sub _assign_data_interface {
    my %args   = @_;
    my $node   = $args{'node'};
    my $ci     = $args{'command_interface'};
    my $host   = $node->hostrec();
    my %params = $host->param_fields();

    $args{'vserver'} = $Options->get( option => 'VSERVER' )
        unless defined $args{'vserver'};

    my $vserver = $args{'vserver'};

    my $lif = NACL::C::Interface->create(
        command_interface => $ci,
        'vserver'         => $vserver,
        'lif'             => 'data1_' . $host->id(),
        'role'            => 'data',
        'data-protocol' =>
            \@{ $Options->get( option => 'DATA_LIF_PROTOCOL' ) },
        'home-node' => $host->id(),
        'home-port' => exists( $params{'DATA_PORT'} ) ? $params{'DATA_PORT'}
        : $params{'DATA_PORT1'},
        'address' => exists( $params{'DATA_IP'} ) ? $params{'DATA_IP'}
        : $params{'DATA_IP1'},
        'netmask' => exists( $params{'DATA_MASK'} ) ? $params{'DATA_MASK'}
        : $params{'DATA_MASK1'},
    );

    if ( $Options->get( option => 'VSADMIN_ON_DATA_LIF' ) ) {
        $lif->modify( 'firewall-policy' => 'mgmt' );
    }
    my $routing_group = $lif->get_one_state_attribute('routing-group');
    my @route         = NACL::C::Route->find(
        command_interface => $ci,
        filter            => {
            'vserver'       => $vserver,
            'routing-group' => $routing_group,
        },
        'allow_empty' => 1,
    );

    # Setup the gateway if needed
    if ( !@route ) {
        NACL::C::Route->create(
            command_interface => $ci,
            'vserver'         => $vserver,
            'routing-group'   => $routing_group,
            'destination'     => '0.0.0.0/0',
            'gateway'         => exists( $params{'DATA_GATEWAY'} )
            ? $params{'DATA_GATEWAY'}
            : $params{'DATA_GATEWAY1'},
        );
    }
    return $TCD::PASS;
}

sub _get_version_info {

    # The behavior of this section has not been tested on mixed mode clusters.
    # Since it only grabs the version of the first node in the NODES list the
    # tests would have to have that nodes build version specified in order to
    # be executed.
    my $ver_info;
    my @nodes = @{ $Options->get( option => 'NODES' ) };

    $TCD::Log->comment("Determining version info for $nodes[0]");
    my $host = Hostrec->new( id => $nodes[0] );

    my $version_mgr = NACL::APISet::Version->new(
        hostobj    => $host,
        category   => "Node",
        shell_type => "CMode",
    );

    my $branch = $version_mgr->get_version_attribute( attribute => "branch" );
    my $build = $version_mgr->get_version_attribute( attribute => "release" );
    my $node_ver
        = $version_mgr->get_version_attribute( attribute => "node_version" );

    $TCD::Log->comment(
        "Node Version: $node_ver Build: $build Branch: $branch");

    $ver_info = {
        version => $node_ver,
        build   => $build,
        branch  => $branch,
    };

    return $ver_info;
}

sub _determine_run_status {

    # Currently this method only accepts build versions in the format
    # of X.X.X

    my %params          = @_;
    my $run_status      = $FALSE;
    my $disabled_flag   = $FALSE;
    my @builds_for_test = @{ $params{builds_for_test} };
    my $node_build      = $params{build_on_node};

    # we only care about build version so remove non-digit characters
    $node_build =~ s/\D//g;

    foreach my $test_build (@builds_for_test) {

        # remove dots from test build version
        $test_build =~ s/\.//g;

        if ( $test_build =~ s/^<=//g && $node_build <= $test_build ) {
            $TCD::Log->comment(
                "Node build $params{node_build} is <= to test build version");
            $run_status = $TRUE;
        }
        elsif ( $test_build =~ s/^>=//g && $node_build >= $test_build ) {
            $TCD::Log->comment(
                "Node build $params{node_build} is >= to test build version");
            $run_status = $TRUE;
        }
        elsif ( $test_build =~ s/^-//g && $node_build == $test_build ) {
            $TCD::Log->comment( "This test is disabled for builds equal to "
                    . "$params{'build_on_node'}" );
            $run_status    = $FALSE;
            $disabled_flag = $TRUE;
            last;    #  specifying '-' should override open ranges
        }
        elsif ( $test_build == $node_build ) {
            $TCD::Log->comment(
                "Node build $params{node_build} is equal to test build version"
            );
            $run_status = $TRUE;
        }
    }

    if ( $run_status == $FALSE && $disabled_flag == $FALSE ) {
        $TCD::Log->warn(
            "The node build ($params{node_build}) did not match a possible"
                . " build for this test (@builds_for_test)" );
    }

    return $run_status;
}

=head2 allocate_aggregates_to_nodes

This function finds or creates empty aggregates, and ensures they are spread evenly
amoung the nodes provided.

Input:

    command_interface
    nodes
    num_of_aggrs

=cut
sub find_create_empty_aggregates_for_nodes {
    my %param        = @_;
    my $node = delete $param{'command_interface'};
    my $nodes = delete $param{'nodes'};
    my $num_of_aggrs = delete $param{'num_of_aggrs'};
    my @aggr_available;
    my $added_aggr;
    my %aggr_added;

    while ( $num_of_aggrs > 0 ) {
        foreach (@$nodes) {
            my $node_name = $_->node;
            $added_aggr = 0;
            my @empty_aggr = NACL::C::Aggregate->find(
                command_interface => $node,
                filter            => {
                    volcount => 0,
                    nodes    => [$node_name],
                    'diskcount' => '2',
                    'raidtype'  => 'raid4',
                },
                allow_empty => 1,
            );
            if (@empty_aggr) {
                foreach (@empty_aggr) {
                    unless ( $aggr_added{ $_->aggregate } ) {
                        push( @aggr_available, $_ );
                        $aggr_added{ $_->aggregate } = 1;
                        $added_aggr = 1;
                        last;
                    }
                }
            }

            unless ($added_aggr) {
                my @new_aggr = create_aggregates(
                    'command_interface'      => $node,
                    'create_specify_options' => {
                        'nodes'     => [$node_name],
                        'method-timeout' =>
                            $Options->get( option => 'AGGR_CREATE_WAIT' ),
                    },
                    number_of_aggregates => 1,
                    'purge_volumes'      => 1,
                );
                push( @aggr_available, @new_aggr );
                $aggr_added{ $new_aggr[0]->aggregate } = 1;
            }

            $num_of_aggrs -= 1;
            last if $num_of_aggrs <= 0;
        }
    }
    return @aggr_available;
}

=head2 get_cached_node
function to get a cached copy of node object.
=cut
sub get_cached_node {
    $Cached_Node ||= getNode();
    return $Cached_Node;
}

1;
