=head1 NAME

Utils.pm - Utility Functions for Content Repository InfiniteVol.

=head1 SYNOPSIS

Note: C<$CLIENT> B<MUST> be a UNIX machine.

use CR::Utils;

=head1 DESCRIPTION

Note: C<$CLIENT> B<MUST> be a UNIX machine.

=cut

package CR::Utils;

use misc;
use Tharn;
use Data::Dumper;
require Exporter;

use NACL::C::Volume;
use NACL::C::Aggregate;
use NACL::C::Client;
use NACL::C::Client::Mount;
use NATE::BaseException;
use NACL::C::NetworkInterface;

use vars qw(@ISA @EXPORT @EXPORT_OK);
@ISA = ('Exporter');

our @EXPORT = qw(
    &get_ns_local_dcs
    &get_ns_remote_dcs
    &do_mount
    &do_unmount
    &get_nfs_data_lif
    &make_cons_offline
    &make_cons_online
    &get_constituents
    &get_constituent_objects
    &generate_cons_jpath_hash
    );


=head2 get_ns_local_dcs()

This function will get the list of Data Constituents which coexists with 
Namespace Constituents on a Node.

=over

=item Inputs

=over

=item C << $command_interface>>

(Required )Node Component Object.

=item C<< $vserver >>

(Optional)Repository Vserver name

=cut

sub get_ns_local_dcs {
    my ( $node, $vserver) = @_;
    my $res;
    my @ns_local_dcs;

    try{
        $res = NACL::C::Volume->find(command_interface       => $node,
                                        filter  => { 
                                            'vserver'           => $vserver,
                                            'constituent-role'  => 'namespace',
                                        }
                                       );

        my $ns_aggr = $res->get_one_state_attribute('aggregate');

        $res = NACL::C::Aggregate->find(command_interface => $node,
                                        filter => {'aggregate' => $ns_aggr},
                                       );

        my $ns_node = $res->get_one_state_attribute('nodes'); 

        my @res = NACL::C::Aggregate->find(command_interface => $node,
                                        filter => {'nodes' => $ns_node },
                                        );

        my @ns_local_aggrs = map {$_->get_one_state_attribute('aggregate')} @res ;
        my $aggr_list = join(',', @ns_local_aggrs);

        @res = NACL::C::Volume->find(command_interface      => $node,
                                     filter => { 
                                        'vserver'             => $vserver,
                                        'aggregate'           => $aggr_list,
                                        'constituent-role'  => 'data',
                                     },
                                     );

        @ns_local_dcs = map {$_->get_one_state_attribute('volume')} @res;

    } catch NATE::BaseException with {
        my $exception_object = shift;
        $Log->comment("Error: : ". Dumper($exception_object));
        return undef;
    };

    return @ns_local_dcs;
}

=head2 get_ns_remote_dcs()

This function will get the list of Data Constituents which does not coexists with 
Namespace Constituents on a Node.

=over

=item Inputs

=over

=item C << $command_interface>>

(Required )Node Component Object.

=item C<< $vserver >>

(Optional)Repository Vserver name

=cut

sub get_ns_remote_dcs {
    my ($node, $vserver) = @_;
    my $res;
    my @ns_remote_dcs;

    try{
        $res = NACL::C::Volume->find(command_interface       => $node,
                                        filter  => { 
                                            'vserver'           => $vserver,
                                            'constituent-role'  => 'namespace',
                                        }
                                       );

        my $ns_aggr = $res->get_one_state_attribute('aggregate');

        $res = NACL::C::Aggregate->find(command_interface => $node,
                                        filter => { 'aggregate' => $ns_aggr},
                                       );

        my $ns_node = $res->get_one_state_attribute('nodes');

        my $ns_nodes = join(',', @$ns_node);

        $Log->comment("ns_nodes = $ns_nodes");

        my @nodes = NACL::C::Node->find(command_interface => $node);

        my @node_list = map{$_->get_one_state_attribute('node')} @nodes;
        $Log->comment("nodes = @node_list ");

        my @aggr_node_list=();

        foreach my $n (@node_list) {
              push (@aggr_node_list, $n) if ( $n !~ m/$ns_nodes/ );
        }

        $Log->comment("aggr_node_list = @aggr_node_list");

        my @res = NACL::C::Aggregate->find(command_interface => $node,
                                        filter => {'nodes' => \@aggr_node_list },
                                        );

        my @ns_remote_aggrs = map {$_->get_one_state_attribute('aggregate')} @res ;
        my $aggr_list = join(',', @ns_remote_aggrs);

        @res = NACL::C::Volume->find(command_interface      => $node,
                                     filter => { 
                                        'vserver'             => $vserver,
                                        'aggregate'           => $aggr_list,
                                        'constituent-role'  => 'data',
                                     },
                                     );

        @ns_remote_dcs = map {$_->get_one_state_attribute('volume')} @res;
    } catch NATE::BaseException with {
        my $exception_object = shift;
        $Log->comment("Error: : ". Dumper($exception_object));
        return undef;
    };

    return @ns_remote_dcs;
}

=head2 do_mount()

This function will use NACL::C::Client::Mount to mount InfinteVol on 
CLIENT .

=over

=item Inputs

=over

=item C << $server>>

(Required )Data LIF IP of the Repository Vserver to be used for mount.

=item C<< $vers >>

(Required)NFS Version to be used to mount.

=item C<< $junction_path >>

(Required)Junction Path of the Infinitevol to be used for mount.

=item C<< $mountopts >>

(Optional)Mount Options [-o] options to be used for mount.

=cut

sub do_mount {
    my ($server, $vers, $junction_path, $mountopts) = @_;
    my $client = NACL::C::Client->find();
    my $mount;
    my $mount_path;

    my @mount_params = (command_interface   => $client,
                        'server'            => $server,
                        'path'              => $junction_path,
                        'vers'              => $vers,
                        'retries'           => 1,
                        );
    push (@mount_params, 'mountopts' => $mountopts) if ( defined $mountopts);
 
    try {
        $mount = NACL::C::Client::Mount->create(@mount_params);

        $mount_path = $mount->get_one_state_attribute('mount_point');
        $Log->comment("Mount_Path: $mount_path created" );
        return $mount_path;

    } catch NATE::BaseException with {
        my $exception_object = shift;
        $Log->comment("Error: : ". Dumper($exception_object));
        return undef;
    };

}

=head2 do_unmount()

This function will use NACL::C::Client::Mount to delete the mountpoint on 
CLIENT .

=over

=item Inputs

=over

=item C << $mount_path>>

(Required )Mount path to be deleted.

=cut

sub do_unmount {
    my ($mount_path) = @_;
    $Log->comment("Arg passed : $mount_path");
    try{
        my $client = NACL::C::Client->find();
        NACL::C::Client::Mount->delete(command_interface    => $client,
                                       'mount_point'        => $mount_path,
                                       'force'              => 'true',
                                       );

    } catch NATE::BaseException with {
        my $exception_object = shift;
        $Log->comment("Error: : ". Dumper($exception_object));
        return undef;
    };
    return 1;
}

=head2 get_nfs_data_lif()

This function will get the data lif having nfs as data protocol for the given 
vserver.
=over

=item Inputs

=over

=item C << $command_interface>>

(Required )Node Component Object.

=item C<< $vserver >>

(Required)Repository Vserver name

=cut

sub get_nfs_data_lif {
    my ($node, $vserver) = @_;
    my @lifs;

    try {
        my @res = NACL::C::NetworkInterface->find(command_interface         =>$node,
                                                      filter => {
                                                          'vserver'         => $vserver,
                                                          'data-protocol'   => ['nfs'],
                                                          'status-admin'    => 'up',
                                                      },
                                                  );
        @lifs = map { $_->get_one_state_attribute('address') } @res;

    } catch NATE::BaseException with {
        my $exception_object = shift;
        $Log->comment("Error: : ". Dumper($exception_object));
        return undef;
    };
    return @lifs;
}

=head2 make_cons_offline()

This function will offline the consituents provided.
=over

=item Inputs

=over

=item C << $command_interface>>

(Required )Node Component Object.

=item C<< $vserver >>

(Required)Repository Vserver name

=item C<< @cons >>

(Required)Constituent name list to offline.

=cut

sub make_cons_offline {
    my ($node, $vserver, @cons_list) = @_;
    my $fail =0;

    foreach my $cons (@cons_list) {

        try {

            NACL::C::Volume->unmount(command_interface  => $node,
                                     'volume'           => $cons,
                                     'vserver'          => $vserver,
                                    );

            NACL::C::Volume->offline(command_interface  => $node,
                                     'volume'           => $cons,
                                     'vserver'          => $vserver,
                                     'foreground'       => 'true',
                                    );

        } catch NATE::BaseException with {
            my $exception_object = shift;

            $Log->comment("Error: : ". Dumper($exception_object));
            return undef;
        };
    }
    return 1;
}

=head2 make_cons_online()

This function will online the consituents provided.
=over

=item Inputs

=over

=item C << $command_interface>>

(Required )Node Component Object.

=item C<< $vserver >>

(Required)Repository Vserver name

=item C<< $cons_jpath_hashref >>

(Required)Consituent name and its junction path hash.

=cut

sub make_cons_online {
    my ($node, $vserver, $cons_jpath_hashref) = @_;
    my $fail =0;

    foreach my $cons (keys %$cons_jpath_hashref) {
        try {
            NACL::C::Volume->online(command_interface   => $node,
                                    'volume'            => $cons,
                                    'vserver'           => $vserver,
                                    'foreground'        => 'true',
                                    );

            NACL::C::Volume->mount(command_interface    => $node,
                                    'volume'            => $cons,
                                    'vserver'           => $vserver,
                                    'junction-path'     => $cons_jpath_hashref->{$cons},
                                    );

        } catch NATE::BaseException with {
            my $exception_object = shift;
            $Log->comment("Error: : ". Dumper($exception_object));
            return undef;
        };

    }
return 1;
}

=head2 get_constituents()

This function will get the Constituents of the given repository vserver 
and the consituent role if provided.
=over

=item Inputs

=over

=item C << $command_interface>>

(Required )Node Component Object.

=item C<< $vserver >>

(Required)Repository Vserver name

=item C<< $role >>

(Optinal)Consituent role.

=cut

sub get_constituents {
    my ($node, $vserver, $role) = @_;
    my @cons_list;

    try {
        if ( defined $role ) {
            @cons_list = NACL::C::Volume->find(command_interface            => $node,
                                                 filter => {
                                                        'vserver'           => $vserver,
                                                        'is-constituent'    => 'true',
                                                        'constituent-role'  => $role
                                                 },
                                               );
        } else {
            @cons_list = NACL::C::Volume->find(command_interface            => $node,
                                                 filter => {
                                                        'vserver'           => $vserver,
                                                        'is-constituent'    => 'true',
                                                 },
                                               );
        }

    } catch NATE::BaseException with {
        my $exception_object = shift;
        $Log->comment("Error: : ". Dumper($exception_object));
        return undef;
    };

    my @cons = map {$_->get_one_state_attribute('volume')} @cons_list ;
    return @cons;
}

=head2 get_constituent_objects()

This function will get the Volume objects for the Constituents of the given 
repository vserver and the consituent role if provided.
=over

=item Inputs

=over

=item C << $command_interface>>

(Required )Node Component Object.

=item C<< $vserver >>

(Required)Repository Vserver name

=item C<< $role >>

(Optional)Consituent role.

=cut

sub get_constituent_objects {
    my ($node, $vserver, $role) = @_;
    my @cons_list;

    try {
        if ( defined $role ) {
            @cons_list = NACL::C::Volume->find(command_interface            => $node,
                                                 filter => {
                                                        'vserver'           => $vserver,
                                                        'is-constituent'    => 'true',
                                                        'constituent-role'  => $role
                                                 },
                                               );
        } else {
            @cons_list = NACL::C::Volume->find(command_interface            => $node,
                                                 filter => {
                                                        'vserver'           => $vserver,
                                                        'is-constituent'    => 'true',
                                                 },
                                               );
        }

    } catch NATE::BaseException with {
        my $exception_object = shift;
        $Log->comment("Error: : ". Dumper($exception_object));
        return undef;
    };

    return @cons_list;
}

=head2 generate_cons_jpath_hash()

This function will genarate a hash of consituent name and its junction path 
=over

=item Inputs

=over

=item C << $command_interface>>

(Required )Node Component Object.

=item C<< $vserver >>

(Required)Repository Vserver name

=cut

sub generate_cons_jpath_hash {
    my ( $node, $vserver) = @_;
    my @cons = get_constituents($node, $vserver);
    my %jpath_map;

    foreach my $cons ( @cons ) {

        try {
            my $res = NACL::C::Volume->find(command_interface       => $node,
                                                filter => {
                                                    'vserver'       => $vserver,
                                                    'volume'        => $cons,
                                                },
                                            );
            $jpath_map{$cons} = $res->get_one_state_attribute('junction-path');

        } catch NATE::BaseException with {
            my $exception_object = shift;
            $Log->comment("Error: : ". Dumper($exception_object));
            return undef;
        };
    }
    return \%jpath_map;
}

1;