# $Id: $
#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary VolumeMove - Volume move class
## @author natha@netapp.com
## @status shared
## @pod here

=head1 NAME

 VolumeMove - Volume move class

=head1 DESCRIPTION

 C<CR::VolumeMove> provides methods for volume move tests.

=head1 METHODS

=cut

package CR::VolumeMove;

use nfs::NfsClient;
use nfs::NfsDef qw( UNCHECKED FATTR4_MODE
    OPEN4_SHARE_ACCESS_WRITE OPEN4_SHARE_ACCESS_READ
    OPEN4_SHARE_ACCESS_BOTH  OPEN4_SHARE_DENY_BOTH
    OPEN4_SHARE_DENY_NONE    OPEN4_SHARE_DENY_READ );
use TestEnv::MinWin qw(execute_cmd);
use NACL::C::VolumeMove;
use NACL::C::VserverLocks;
use NACL::CS::NetworkInterface;
use NACL::CS::Volume;
use NACL::CS::VolumeMoveTargetAggr;
use NACL::CS::Vserver;
use NACL::Exceptions::NoElementsFound qw( :try );
use Tharn qw( $Log );
use WaflironLib qw( populate_percent_data );

my $mark = '-' x 80;

=head2 nfs_create_file

  my $open = CR::VolumeMove::nfs_create_file(
      nfs_client  => $nfs_client,
      file_name   => $file_name,
      [ access_lock => $access_lock, ]
      [ deny_lock   => $deny_lock, ]
  );

 Creates a file with name passed in the variable $file_name under
 the mount point of the NFS synthetic client passed in $nfs_client.

 The following parameters are accepted:

=over

=item Arguments

=over

=item nfs_client

 (Required) : An object type of nfs::NfsClient

=item file_name

 (Required) : Name of the file to be created

=item access_lock

 (Optional) : Access lock type: OPEN4_SHARE_ACCESS_WRITE, OPEN4_SHARE_ACCESS_READ
             (default is OPEN4_SHARE_ACCESS_WRITE)

=item deny_lock

 (Optional) : Deny lock type: OPEN4_SHARE_DENY_NONE, OPEN4_SHARE_DENY_BOTH
              (default is OPEN4_SHARE_DENY_NONE)

=back

=item Return:

 Reference to a hash containing the status and output of procedure.

=back

=cut

sub nfs_create_file {
    my %args        = @_;
    my $nfs_client  = $args{nfs_client};
    my $file_name   = $args{file_name};
    my $access_lock = $args{access_lock} || OPEN4_SHARE_ACCESS_WRITE;
    my $deny_lock   = $args{deny_lock} || OPEN4_SHARE_DENY_NONE;
    my $result;

    $access_lock = OPEN4_SHARE_ACCESS_READ
        if ( $access_lock eq 'OPEN4_SHARE_ACCESS_READ' );
    $access_lock = OPEN4_SHARE_ACCESS_WRITE
        if ( $access_lock eq 'OPEN4_SHARE_ACCESS_WRITE' );
    $access_lock = OPEN4_SHARE_ACCESS_BOTH
        if ( $access_lock eq 'OPEN4_SHARE_ACCESS_BOTH' );
    $deny_lock = OPEN4_SHARE_DENY_NONE
        if ( $deny_lock eq 'OPEN4_SHARE_DENY_NONE' );
    $deny_lock = OPEN4_SHARE_DENY_READ
        if ( $deny_lock eq 'OPEN4_SHARE_DENY_READ' );

    eval {
        $result = $nfs_client->open(
            name    => $file_name,
            dhandle => $nfs_client->mnt_fh,
            cr_mode => UNCHECKED,
            access  => $access_lock,
            deny    => $deny_lock,
            attr    => { FATTR4_MODE() => 0777 }
        );
    };

    if ($@) {
        $Log->comment(
            "\n$mark\nCould not create file $file_name: $@\n$mark\n");
        return undef;
    }

    return $result;
} ## end sub nfs_create_file

=head2 get_win_oplock

  CR::VolumeMove::get_win_oplock(
      $win_client,
      $filepath,
      [ $oplock ]
  );

 Creates a file with name passed in the variable $filepath under
 the mount point of the CIFS client passed in $win_client.

 The following parameters are accepted:

=over

=item Arguments

=over

=item win_client

 (Required) : An object of Hostrec::Windows

=item filepath

 (Required) : Full path name of the file to be created

=item opts

 (Optional) : A HASH ref of the options.
              eg, $opts = {
                      open_mode   => $open_mode,   # dafault: rw
                      shared_mode => $shared_mode, # dafault: sr
                      create_mode => $create_mode, # dafault: cn
                  };

=item expect_timeout

 (Optional) : Timeout value for expect. Default is 60 seconds.

=back

=back

=cut

sub get_win_oplock {
    my $win_client     = shift;
    my $filepath       = shift;
    my $opts           = shift;
    my $expect_timeout = shift || '15';
    my $open_mode      = $opts->{open_mode} || 'rw';
    my $shared_mode    = $opts->{shared_mode} || 'sr';
    my $create_mode    = $opts->{create_mode} || 'cn';

    require Expect;
    my $rsh = new Expect;

    # Spawn a rsh connection to the minwin client and get the cmd prompt
    $rsh->spawn("rsh $win_client cmd");
    $rsh->expect( $expect_timeout, -re, ">" );

    # Open the $filepath with open32.exe util
    $rsh->send( "X:\\bin\\open32.exe $filepath $open_mode "
            . "$shared_mode $create_mode\r" );
    $rsh->expect( $expect_timeout, -re, "exit:" );

    # Get a oplock on the $filepath
    $rsh->send("l\r");
    $rsh->expect( $expect_timeout, -re, "exit:" );
    sleep 300;
    $rsh->send("c\r");
    $rsh->expect( $expect_timeout, -re, "exit:" );
    $rsh->send("q\r");
    $rsh->expect( $expect_timeout, -re, ">" );
    $rsh->close();
} ## end sub get_win_oplock

sub cifs_close_file {
    my $win_client     = shift;
    my $expect_timeout = '60';
    my $pid            = undef;

    require Expect;
    my $rsh = new Expect;

    # Spawn a rsh connection to the minwin client and get the cmd prompt
    $rsh->spawn("rsh $win_client cmd");
    $rsh->expect( $expect_timeout, '>' );

    # Get PID of open32.exe application
    $rsh->send( 'tasklist /fi "IMAGENAME eq open32.exe"' . "\r" );
    $rsh->expect( $expect_timeout, 'Windows>' );
    my $before = $rsh->before();
    while ( $before =~ /open32.exe\s+(\d+)\s+/g ) {
        $pid = $1;
        $rsh->send("taskkill /F /PID $pid\r");
        $rsh->expect( $expect_timeout, '>' );
    }
    sleep 30;
    $rsh->close();
} ## end sub cifs_close_file

=head2 cifs_delete_file

  CR::VolumeMove::cifs_delete_file(
      $win_client,
      $filepath,
  );

 Creates a file with name passed in the variable $filepath under
 the mount point of the CIFS client passed in $win_client.

 The following parameters are accepted:

=over

=item Arguments

=over

=item win_client

 (Required) : An object of Hostrec::Windows

=item filepath

 (Required) : Full path name of the file to be created

=back

=back

=cut

sub cifs_delete_file {
    my $win_client = shift;
    my $filepath   = shift;

    $Log->comment("\n$mark\nAttempting to delete file $filepath\n$mark\n");
    my $output = execute_cmd( $win_client, "del $filepath" );

    if ( $output =~ m/\w+/ ) {
        $Log->comment( "\n$mark\nFailed to delete the file $filepath"
                . "\nError: $output\n$mark\n" );
        return undef;
    }

    $Log->comment(
        "\n$mark\nSuccessfully deleted the file $filepath\n$mark\n");
    return 1;
} ## end sub cifs_delete_file

=head2 data_lifs

  CR::VolumeMove::data_lifs(
      command_interface => $command_interface,
      vserver           => $vserver,
  );

  Uses C<NACL::CS::NetworkInterface> to find data lifs associated with
  a vserver. This method finds data lifs in a randomized order.

 The following parameters are accepted:

=over

=item Arguments

=over

=item command_interface

 (Required) : Object of NACL::C::Node. A component object that represents the
              host on which commands can be executed.

=item vserver

 (Required) : The name of the vserver that contains the lif

=back

=item Return:

 Array of data lifs IP adresses or IP address of a single data lif.

=back

=cut

sub data_lifs {
    my %args              = @_;
    my $command_interface = $args{command_interface};
    my $vserver           = $args{vserver};
    my @data_lifs;

    my @data_lifs_obj = NACL::CS::NetworkInterface->fetch(
        command_interface => $command_interface,
        randomize_order   => 1,
        filter            => {
            vserver => $vserver,
            role    => 'data',
        }
    );

    foreach ( 0 .. $#data_lifs_obj ) {
        push @data_lifs, $data_lifs_obj[$_]->address();
    }

    return wantarray ? @data_lifs : $data_lifs[0];
} ## end sub data_lifs

=head2 junction_path

  CR::VolumeMove::junction_path(
      command_interface => $command_interface,
      vserver           => $vserver,
  );

 Uses C<NACL::CS::Volume> to find the junction path of a volume.

 The following parameters are accepted:

=over

=item Arguments

=over

=item command_interface

 (Required) : Object of NACL::C::Node. A component object that represents the
              host on which commands can be executed.

=item vserver

 (Required) : The name of the vserver associated with the volume.

=item volume

 (Required) : The name of the volume.

=back

=item Return:

 Junction path of the volume.

=back

=cut

sub junction_path {
    my %args              = @_;
    my $command_interface = $args{command_interface};
    my $vserver           = $args{vserver};
    my $infinite_volume   = $args{infinite_volume};

    my $volume_state_obj = NACL::CS::Volume->fetch(
        command_interface => $command_interface,
        requested_fields  => ['junction-path'],
        filter            => {
            vserver => $vserver,
            volume  => $infinite_volume,
        }
    );
    my $junction_path = $volume_state_obj->junction_path();

    return $junction_path;
} ## end sub junction_path

=head2 nfs_client_setup

  CR::VolumeMove::nfs_client_setup(
      data_lif        => $data_lif,
      'junction-path' => junction_path,
      proto           => $proto,
  );

 Creates an object of nfs::NfsClient and mounts the junction
 path of the infinite volume from the synthetic client.

 The following parameters are accepted:

=over

=item Arguments

=over

=item data_lif

 (Required) : IP address of a data lif associated with the vserver which
 hosts the infinite volume.

=item junction-path

 (Required) : The junction path of the infinite volume.

=item proto

 (Required) : Protocol to be used for the NFS mount operation.

=back

=item Return:

 nfs::NfsClient object.

=back

=cut

sub nfs_client_setup {
    my %args          = @_;
    my $data_lif      = $args{data_lif};
    my $junction_path = $args{'junction-path'};
    my $proto         = $args{proto};
    my $version       = $args{version} || '4';

    my $nfs_client_obj = nfs::NfsClient->new(
        hostname => $data_lif,
        proto    => $proto,
        version  => $version,
    );
    if ( defined $nfs_client_obj ) {
        $Log->comment( "\n$mark\nMount the volume on the synthetic"
                . " client\n$mark\n" );
        $nfs_client_obj->mount( path => $junction_path );
        $Log->comment( "\n$mark\nMount Path: "
                . $nfs_client_obj->mnt_path
                . "\n$mark\n" );
    } else {
        $Log->comment("Failed to create synthetic client object!\n");
        return undef;
    }

    return $nfs_client_obj;
} ## end sub nfs_client_setup

=head2 vserver_locks_show

  CR::VolumeMove::vserver_locks_show(
      command_interface => $command_interface,
      vserver           => $vserver,
      volume            => $volume,
      want_share_lock   => $want_share_lock,
      %other_options
  );

 Uses C<NACL::C::VserverLocks> or API calls to find the
 locks of a file on a vserver.

 The following parameters are accepted:

=over

=item Arguments

=over

=item command_interface

 (Required) : Object of NACL::C::Node. A component object that represents the
              host on which commands can be executed.

=item vserver

 (Required) : The name of the vserver associated with the volume.

=item volume

 (Required) : The name of the data constituent volume.

=item want_share_lock

 (Optional) : Agrument to specify whether sharelock mode is needed
              (default is 'undef')

=back

=item Return:

 Object of NACL::C::VserverLocks or the processed output of the
 'vserver locks show' command

=back

=cut

sub vserver_locks_show {
    my %args              = @_;
    my $command_interface = $args{command_interface};
    my $vserver           = $args{vserver};
    my $volume            = $args{volume};
    my $want_share_lock   = $args{want_share_lock} || undef;

    # When sharelock mode is required for verification
    if ( defined($want_share_lock) && $want_share_lock eq 'true' ) {
        my $apiset = $command_interface->apiset(
            category  => 'Node',
            interface => 'CLI',
            set       => 'CMode',
        );
        $apiset->set( showallfields => 'false' );
        my $output = $apiset->vserver_locks_show( volume => $volume )
            ->get_processed_output();
        $apiset->set( showallfields => 'true' );
        return $output;
    } else {
        my $vs_locks_obj = NACL::C::VserverLocks->find(
            command_interface => $command_interface,
            filter            => {
                vserver => $vserver,
                volume  => $volume,
            }
        );
        my ( $key, $value );
        $Log->comment("\n\nVserver Locks:\n$mark\n");
        while ( ( $key, $value ) = each %$vs_locks_obj ) {
            $key =~ s/_/ /g;
            $Log->comment( ucfirst($key) . ': ' . $value );
        }
        $Log->comment("\n$mark\n");
        return $vs_locks_obj;
    }
} ## end sub vserver_locks_show

=head2 vserver_locks_break

  CR::VolumeMove::vserver_locks_break(
      command_interface => $command_interface,
      vserver           => $vserver,
  );

 Uses C<NACL::C::VserverLocks> to break all the locks on a vserver

 The following parameters are accepted:

=over

=item Arguments

=over

=item command_interface

 (Required) : Object of NACL::C::Node. A component object that represents the
              host on which commands can be executed.

=item vserver

 (Required) : The name of the vserver associated with the volume.

=back

=back

=cut

sub vserver_locks_break {
    my %args              = @_;
    my $command_interface = $args{command_interface};
    my $vserver           = $args{vserver};

    NACL::C::VserverLocks->break(
        command_interface => $command_interface,
        vserver           => $vserver,
        lif               => '*'
    );
} ## end sub vserver_locks_break

=head2 verify_vserver_locks

  CR::VolumeMove::verify_vserver_locks(
      $initail_locks, $final_locks
  );

 Compares between two NACL::C::VserverLocks objects

 The following parameters are accepted:

=over

=item Arguments

=over

=item initail_locks

 (Required) : Object of NACL::C::VserverLocks

=item final_locks

 (Required) : Object of NACL::C::VserverLocks

=back

=back

=cut

sub verify_vserver_locks {
    my $initial_locks = shift;
    my $final_locks   = shift;
    my $result        = 1;

    foreach ( keys %$initial_locks ) {
        if ( $initial_locks->{$_} eq $final_locks->{$_} ) {
            $Log->comment( "[MATCH] Lock [$_] matches ["
                    . $initial_locks->{$_}
                    . "]\n" );
        } else {
            $Log->comment("\n[MISMATCH] Lock [$_]\n");
            $result = 0;
        }
    }

    return $result;
} ## end sub verify_vserver_locks

=head2 verify_vserver_locks_change

  CR::VolumeMove::verify_vserver_locks_change(
      sharelock_mode => $sharelock_mode,
      locks          => $final_locks,
      volume         => $volume,
      path           => $path
  );

 Verifies the raw output of the 'vserver locks show' command.

 The following parameters are accepted:

=over

=item Arguments

=over

=item sharelock_mode

 (Required) : Sharelock mode (e.g. 'read-deny_none')

=item oplock_level

 (Optional) : Oplock Level (e.g. 'batch')

=item locks

 (Required) : Processed output of API call for command 'vserver locks show'

=item volume

 (Required) : Name of the data constituent volume on which file is created.

=item path

 (Required) : NS path of the file.

=back

=item Return:

 1 - Pass, 0 - Fail

=back

=cut

sub verify_vserver_locks_change {
    my %args           = @_;
    my $sharelock_mode = $args{sharelock_mode};
    my $oplock_level   = $args{oplock_level};
    my $locks          = $args{locks};
    my $volume         = $args{volume};
    my $path           = $args{path};
    my $result         = 1;

    if ( defined($oplock_level) ) {
        if ( $locks
            =~ /$volume.*$path.*cifs.*share-level.*Sharelock Mode:\s+$sharelock_mode/s
            && $locks =~ /op-lock.*Oplock Level:\s+$oplock_level/s ) {
            $Log->comment( "\n$mark\n[WinOplock] Actual Locks MATCHES with "
                    . "the Expected Locks\n$mark\n" );
        } else {
            $Log->comment( "\n$mark\nERROR: [WinOplock] Actual Locks does "
                    . "NOT MATCH with the Expected Locks\n$mark\n" );
            $result = 0;
        }
    } else {
        if ( $locks
            =~ /$volume.*$path.*nfsv\d(\.\d)?.*share-level.*Sharelock Mode:\s+$sharelock_mode/s
            ) {
            $Log->comment( "\n$mark\nActual Locks MATCHES with "
                    . "the Expected Locks\n$mark\n" );
        } else {
            $Log->comment( "\n$mark\nERROR: Actual Locks does "
                    . "NOT MATCH with the Expected Locks\n$mark\n" );
            $result = 0;
        }
    }

    return $result;
} ## end sub verify_vserver_locks_change

=head2 data_constituent_volume

  CR::VolumeMove::data_constituent_volume(
      command_interface => $command_interface,
      vserver           => $vserve,
      path              => $path
  );

 Find the data constituent volume on which a file is created.

 The following parameters are accepted:

=over

=item Arguments

=over

=item command_interface

 (Required) : Object of NACL::C::Node. A component object that represents the
              host on which commands can be executed.

=item vserver

 (Required) : The name of the vserver associated with the infinite volume.

=item path

 (Required) : NS path of the file.

=back

=item Return:

 Data Constituent volume name on which the file is created.

=back

=cut

sub data_constituent_volume {
    my %args              = @_;
    my $command_interface = $args{command_interface};
    my $vserver           = $args{vserver};
    my $path              = $args{path};
    my $dc_volume;

    ### Burt: 575743
    ### The method NACL::C::VolumeFile::show_fhmap is not returning an object
    ### So we cant call the 'state()' method to get data constituent volume.
    # NEED TO MODIFY AFTER the burt is FIXED
    #my $vs_lock_obj = NACL::C::VolumeFile->show_fhmap(
    #    command_interface => $command_interface,
    #    path              => $path,
    #);
    #$Log->comment('************** '.$vs_lock_obj.' **************');
    #foreach (keys %$vs_lock_obj) {
    #    $Log->comment( $_ . ' => ' . $obj->{$_} );
    #}

    ## Work around for the above issue
    my $apiset = $command_interface->apiset(
        category  => 'Node',
        interface => 'CLI',
        set       => 'CMode',
    );
    $apiset->set( showallfields => 'false' );
    $apiset->set_timeout( 'connectrec-timeout' => 300 );
    my $output = $apiset->volume_file_repository_fhmap_show(
        vserver => $vserver,
        path    => $path
    )->get_processed_output();
    $apiset->set( showallfields => 'true' );
    if ( $output =~ /Data Constituent:\s+(\S+_data\d{4})/s ) {
        $dc_volume = $1;
        $Log->comment( "\n$mark\nVolume on which file $path is"
                . " stored: [$dc_volume]\n$mark\n" );
    } else {
        $Log->warn("\nCould not find DC volume");
        return undef;
    }

    return $dc_volume;
} ## end sub data_constituent_volume

=head2 volume_move

  CR::VolumeMove::volume_move(
      command_interface  => $command_interface,
      'constituent-role' => $constituent_role,
      constituent_volume => $constituent_volume,
      vserver            => $vserve,

  );

 Moves a constituent volume (data/namespace/ols/ns_mirror) to a new aggregate.

 The following parameters are accepted:

=over

=item Arguments

=over

=item command_interface

 (Required) : Object of NACL::C::Node. A component object that represents the
              host on which commands can be executed.

=item constituent-role

 (Required) : Role of the constituent volume.

=item vserver

 (Required) : The name of the vserver associated with the infinite volume.

=item constituent_volume

 (Optional) : Name of the constituent volume to be moved.

=item aggregate

 (Optional) : Name of the destination aggregate.

=item cutover-window

 (Optional) : Specifiy Cutover Time Window.

=item cutover-action

 (Optional) : Specify Action For Cutover.

=back

=item Return:

 Object of NACL::C::Job.

=back

=cut

sub volume_move {
    my %args               = @_;
    my $command_interface  = $args{command_interface};
    my $constituent_role   = $args{'constituent-role'};
    my $constituent_volume = $args{constituent_volume} || undef;
    my $destination_aggr   = $args{aggregate} || undef;
    my $vserver            = $args{vserver};
    my $cutover_window     = $args{'cutover-window'} || 45;
    my $cutover_action     = $args{'cutover-action'} || 'defer_on_failure';
    my $cutover_attempts   = $args{'cutover-attempts'} || 3;
    my $method_timeout     = $args{'method-timeout'} || 600;

    # Find constituent volume to be moved
    if ( !defined($constituent_volume) ) {
        $constituent_volume = _find_constituent_volumes(
            command_interface  => $command_interface,
            'constituent-role' => $constituent_role,
            'is-moving'        => 'false',
            state              => 'online',
            vserver            => $vserver,
        );
        $constituent_volume = $constituent_volume->volume();
    }
    $Log->comment(
        "\n$mark\nVolume to be moved: $constituent_volume\n$mark\n");

    # Find destination aggregate
    if ( !defined($destination_aggr) ) {
        $destination_aggr = find_available_aggregate(
            command_interface => $command_interface,
            volume            => $constituent_volume,
            vserver           => $vserver
        );
    }
    $Log->comment(
        "\n$mark\nDestination Aggregate: $destination_aggr\n$mark\n");

    # Start volume move job
    my $job;
    if ( defined($constituent_volume) && defined($destination_aggr) ) {
        NACL::C::VolumeMove->start(
            command_interface       => $command_interface,
            vserver                 => $vserver,
            volume                  => $constituent_volume,
            'destination-aggregate' => $destination_aggr,
            'cutover-window'        => $cutover_window,
            'cutover-action'        => $cutover_action,
            'cutover-attempts'      => $cutover_attempts,
            'method-timeout'        => $method_timeout,
            job_component           => \$job,
        );
    } else {
        $Log->comment("\n$mark\nCould NOT start volume move\n$mark\n");
        return 0;
    }

    return $job;
} ## end sub volume_move

=head2 vol_move_completed

  CR::VolumeMove::vol_move_completed( $vol_move_job );

 Waits until a job is completed.

=over

=item Arguments

=over

=item vol_move_job

 (Required) : Object of a NACL::C::Job;

=back

=item Return:

 1 - Pass, 0 -Fail

=back

=cut

sub vol_move_completed {
    my $vol_move_job = shift;

    ## Return ZERO is $vol_move_job is NOT of type NACL::C::Job
    return 0 if ( ref($vol_move_job) ne 'NACL::C::Job' );

    # Watch the volume move progress
    if ($vol_move_job) {
        $vol_move_job->watch_progress( 'method-timeout' => 300 );
    }

    # Verify completion status of the volume move job
    if ( $vol_move_job->state()->completion() =~ /Success/i ) {
        $Log->comment("\n$mark\nVolume move successful.\n$mark\n");
        undef $vol_move_job;
        return 1;
    } else {
        $Log->comment("\n$mark\nVolume move failed\n$mark\n");
        undef $vol_move_job;
        return 0;
    }
} ## end sub vol_move_completed

=head2 vol_move_is_running

  CR::VolumeMove::vol_move_is_running( $vol_move_job );

  Checks if job is in running state or not.

=over

=item Arguments

=over

=item vol_move_job

 (Required) : Object of a NACL::C::Job;

=back

=item Return:

 1 - Pass, 0 -Fail

=back

=cut

sub vol_move_is_running {
    my $vol_move_job = shift;

    ## Return ZERO is $vol_move_job is NOT of type NACL::C::Job
    return 0 if ( ref($vol_move_job) ne 'NACL::C::Job' );

    my %status = (
        queued  => 'Queued',
        running => 'Running',
        success => 'Success',
        failure => 'Failure',
        error   => 'Error'
    );
    my $job_status = $vol_move_job->state()->state();
    my $result     = undef;
    my $retries    = 30;
    while ( $retries > 0 ) {
        if ( $job_status eq 'Running' ) {
            $result = 1;
            last;
        } else {
            $result = 0;
            $retries--;
            sleep 2;
        }
        $job_status = $vol_move_job->state()->state();
    }

    return $result;
} ## end sub vol_move_is_running

=head2 populate_infinite_vol

  CR::VolumeMove::populate_infinite_vol(
      command_interface => $command_interface,
      connection_type   => $connection_type,
      client            => $client,
      file_count        => 100,
      percentage        => '10%',
      volume            => $volume,
      'junction-path'   => $junction_path,
      vserver           => $vserver
  );

  Populates the infinite volume by filling up its data constituent
  till the total used percentage of the infinite volume becomes same
  as the value passed in the 'percentage' agrument.

=over

=item Arguments

=over

=item command_interface

 (Required) : Object of NACL::C::Node. A component object that represents
              the host on which commands can be executed.

=item connection_type

  (Optional) : Type connection that will established with the client.
               Default is 'ssh'

=item client

  (Required) : Client machine name (eg, galadriel)

=item file_count

  (Optional) : Number of total files to be created. Default is 50.

=item volume

  (Optional) : Name of the infinite volume. Default is 'repo_vol'.

=item junction-path

  (Optional) : Name of the junction path of the infinite volume.
               Default is '/NS'.

=item percentage

  (Required) : Percentage till which the infinte volume needs to be filled.

=item vserver

  (Optional) : Name of the vserver associated with the infinite volume.
               Default is 'vs0'.

=back

=item Return:

  1 - Pass, 0 - Fail

=back

=cut

sub populate_infinite_vol {
    my %args              = @_;
    my $command_interface = $args{command_interface};
    my $connection_type   = $args{connection_type} || 'ssh';
    my $client            = $args{client};
    my $file_count        = $args{file_count} || '50';
    my $infinite_vol      = $args{volume} || 'repo_vol';
    my $junction_path     = $args{'junction-path'} || '/NS';
    my $percentage        = $args{percentage};
    my $vserver           = $args{vserver} || 'vs0';

    my $volume_params = {
        volume             => $infinite_vol,
        vserver            => $vserver,
        'junction-path'    => $junction_path,
        'junction-active'  => 'true',
        'stripe-format'    => 'data',
        'user'             => 'root',
        'unix-permissions' => 777,
        'wait_time'        => 1800,
    };

    $Log->comment(
        "\n$mark\nPopulating DC vols of the infinite volume: $client\n$mark\n"
    );
    populate_percent_data(
        command_interface => $command_interface,
        volume_params     => $volume_params,
        connection_type   => $connection_type,
        client            => $client,
        file_count        => $file_count,
        percentage        => $percentage,
    );

    my $vol_state_obj = NACL::CS::Volume->fetch(
        command_interface => $command_interface,
        requested_fields  => ['percent-used'],
        filter            => {
            vserver => $vserver,
            volume  => $infinite_vol,
        }
    );
    my $percent_used = $vol_state_obj->percent_used();
    $Log->comment("\n$mark\nUsed Percentage: $percent_used\n$mark\n");

    ( $percent_used eq $percentage ) ? return 1 : return 0;
} ## end sub populate_infinite_vol

=head2 find_available_aggregate

  CR::VolumeMove::find_available_aggregate(
      command_interface => $command_interface,
      volume            => $volume,
      vserver           => $vserver
  );

  Finds the target aggregates suitable for a volume to be moved.

=over

=item Arguments

=over

=item command_interface

 (Required) : Object of NACL::C::Node. A component object that represents
              the host on which commands can be executed.

=item volume

  (Required) : Name of the volume to be moved.

=item vserver

  (Required) : Name of the vserver associated with the infinite volume.
               This option is used to find the root aggregate of the vserver.

=back

=item Return:

  Array of aggregate names or name of a single aggregate.

=back

=cut

sub find_available_aggregate {
    my %args              = @_;
    my $command_interface = $args{command_interface};
    my $volume            = $args{volume};
    my $vserver           = $args{vserver};
    my @available_aggr;
    my @target_aggr;

    # Get target aggregates for volume move
    try {
        @available_aggr = NACL::CS::VolumeMoveTargetAggr->fetch(
            command_interface => $command_interface,
            randomize_order   => 1,
            filter            => {
                vserver => $vserver,
                volume  => $volume,
            }
        );
    }
    catch NACL::Exceptions::NoElementsFound with {
        my $exception = shift;
        if ( $exception =~ /No matching volume move target aggr\(s\) found/ )
        {
            $Log->comment( "\n$mark\nNo matching volume move target"
                    . "aggr(s) found\n$mark\n" );
        }
    };
    foreach (@available_aggr) { push @target_aggr, $_->aggregate() }

    return wantarray ? @target_aggr : $target_aggr[0];
} ## end sub find_available_aggregate

#-------------------------------------------------------------------------------
# Subroutine Name: _find_constituent_volumes
# Description    : Finds the constituent volumes of an infinite volume
# Arguments      : command_interface, constituent_role, volume_state, vserver
# Return Type    : Object of NACL::CS::Volume
#-------------------------------------------------------------------------------
sub _find_constituent_volumes {
    my %args              = @_;
    my $command_interface = $args{command_interface};
    my $constituent_role  = $args{'constituent-role'};
    my $is_moving         = $args{'is-moving'};
    my $volume_state      = $args{state};
    my $vserver           = $args{vserver};

    ## Find volume with the given filters
    my $constituent_volumes = NACL::CS::Volume->fetch(
        command_interface => $command_interface,
        randomize_order   => 1,
        requested_fields  => [ 'size', 'aggregate' ],
        filter            => {
            'constituent-role' => $constituent_role,
            'is-moving'        => $is_moving,
            state              => $volume_state,
            vserver            => $vserver,
        }
    );

    return $constituent_volumes;
} ## end sub _find_constituent_volumes

1;
