#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#$Id$
## @summary
## A suite of useful wiki-related subroutines
## @description
## A consolidation of wiki-related subroutines for easier wiki-related scripting
## @author aldrin@netapp.com
## @status shared
## @pod here

package CR::Wiki;

use strict;
use warnings;
use CR::WikiTable;

use POSIX qw(:sys_wait_h);
use NATE::Log qw(log_global);
use IPC::Open3;
use Readonly;
use English qw(-no_match_vars);

################################################################################
#
# CONSTANTS
#
################################################################################

Readonly my $LOG => NATE::Log->new();
%CR::Wiki::PAGE_CONSTANTS = (
    'TRIAGE'            => "== Triage Summary ==\n",
    'AUTOMATION'        => "== Auto Update ==\n",
    'ARCHIVES'          => "== Archives ==\n",
    'SUMMARY'           => "==== Nightly Smokes ====\n",
);

################################################################################
#
# GLOBALS
#
################################################################################

my %ORIG_CONTENT = ();
my $wsvn_path = "/x/eng/localtest/noarch/esis/wsvnwiki/src/wikitools/wsvn";

################################################################################
#
# SUBROUTINES
#
################################################################################

################################################################################
#
#   Add the rows of the given WikiTable table to the wiki page.
#
#   Input:
#       Wiki text to modify.
#       WikiTable table to add to the wiki page.
#
#   Output:
#       Modified wiki text.
#
################################################################################
sub table_update {
    my ($content, $table) = @_;

    my $begin = "$table->{$CR::WikiTable::BEGIN_TAG}";
    my $body = "$table->{$CR::WikiTable::BODY_TAG}";

    my $table_header = $table->header_string();
    if ($content =~ s/($begin\n).*?($body\n)/$1$table_header$2/s) {
        $LOG->comment("successfully updated $begin");
    }
    else {
        $LOG->error("Could not find '$begin\n'");
    }

    my $table_text = $table->body_string();
    if ($content =~ s/($body\n)/$1$table_text/) {
        $LOG->comment("successfully updated $body");
    }
    else {
        $LOG->error("Could not find '$body'");
    }

    return $content;
}

################################################################################
#
#   Check if the given WikiTable table is present in the wiki page.
#
#   Input:
#       Wiki text to modify.
#       WikiTable table to search for.
#
#   Output:
#       True if the WikiTable table was found.
#       False otherwise.
#
################################################################################
sub table_find {
    my ($content, $table) = @_;

    return $content =~ m/$table->{$CR::WikiTable::BODY_TAG}/;
}

################################################################################
#
#   Retrieve all table rows on the wiki page that fall within the given table
#   markup tags.
#
#   Input:
#       Wiki text to modify.
#       Start tag of the WikiTable to parse.
#       Body tag of the WikiTable to parse.
#       End tag of the WikiTable to parse.
#
#   Output:
#       Reference to the list of table rows. Each row is a hash representing the
#           attributes of the row.
#
################################################################################
sub rows_get {
    my ($content, $start_tag, $body_tag, $end_tag) = @_;
    my @table_rows = ();

    #print("using tags: $start_tag, $body_tag, $end_tag\n");

    #All of the text between the given header and body, body and end tags
    #need to be extracted for every table matching the appropriate tags.
    while ($content =~ m/.*?$start_tag(.*?)$body_tag(.*?)$end_tag/sg) {
        my ($table_hdr, $table_bdy) = ($1, $2);
        my @table_headers = ();

        #This is a regex to grab any lines that denote a wikitable header
        #(column name). These are of the format '!...\n' and can sometimes
        #start with '<'
        my $regex_str = qr{[^<]!(.*)};

        #parse the column names of the table into a list
        while ($table_hdr =~ m/$regex_str/g) { push (@table_headers, $1); }

        #change the regex to search for lines denoting a wikitable column
        #value. For example, a wikitable row cell might be in the format
        #'|...\n'. Since '|' is a metacharacter, we escape it with '\'
        $regex_str = qr{\|(.*)};
        my ($row_data, $ndx) = ({}, 0);

        while ($table_bdy =~ m/$regex_str/g) {
            my $row = $1;

            #In this case (the row has more than just '|-' or '|}') we
            #have found a line with some useful information
            if ($row ne q{-} && $row ne '}') {
                $row_data->{$table_headers[$ndx++]} = $row;
            }
            #In this case, we have found a line that is either the separator
            #between the header and body of the table, or the end of the body
            #of the table. Either way, we push a reference to a hash of the row
            #into a list and reset the hash for the next table row.
            elsif (($row eq q{-} || $row eq '}') && (keys %{$row_data})) {
                push(@table_rows, $row_data);
                ($row_data, $ndx) = ({}, 0); #reset these for next tuple
            }
        }
    }

    return \@table_rows;
}

################################################################################
#
#   Add a new table to the wiki page.
#
#   Input:
#       Wiki text to modify.
#       WikiTable table to add to the wiki page.
#       Markup tag for section to add the WikiTable table to.
#       Hash of other options to add. Currently only 'date' and 'build' are
#           used. If 'date' is specified then the given date will be added in
#           the appropriate section. If 'build' is specified then the given
#           build will be added below the date, followed by the content of the
#           WikiTable table.
#
#   Output:
#       Modified wiki text.
#
################################################################################
sub table_add {
    my ($content, $table, $section_tag, $options) = @_;
    my ($table_text, $date, $build) = ($table->to_string(), q{}, q{});

    if ($options->{date}) { $date = $table->{date}; }
    if ($options->{build}) { $build = $table->{build}; }

    #if date was specified as an argument then make sure to add a header
    #for the provided date if it doesn't already exist
    if ($date) {
        if ($content !~ m/$section_tag\n$date/) {
            $content =~ s/($section_tag\n)/$1$date/;
        }

        #add this table with the specified build. Even if build was not
        #specified as an argument, it should be considered an empty
        #string and so will still work.
        $content =~ s/($date)/$1$build$table_text/;
    }
    #If the date was not provided as an argument, then only try to
    #find the header for the build, if build was specified. Otherwise
    #the table should be added directly under the name of the section.
    else { $content =~ s/($section_tag\n)/$1$build$table_text/; }

    return $content;
}

################################################################################
#
#   Replace table contents on wiki with given table by using the markup tags
#   associated with the Wikitable table.
#
#   Input:
#       Wiki text to modify.
#       WikiTable table to set on the wiki page.
#
#   Output:
#       Modified wiki text.
#
################################################################################
sub table_set {
    my ($content, $table) = @_;

    my $begin_tag = $table->{$CR::WikiTable::BEGIN_TAG};
    my $body_tag = $table->{$CR::WikiTable::BODY_TAG};

    my $table_header = $table->header_string();
    if ($content =~ s{($begin_tag\n).*?($body_tag\n)}{$1$table_header$2}s) {
        $LOG->comment("successfully set new table header\n");
    }
    else {
        $LOG->warn("could not find tags '$begin_tag\n' and '$body_tag\n'");
    }

    my $table_text = $table->body_string();
    #Adding /x seems to break this regex
    #$content =~ s{($body_tag\n).*?(\Q$CR::WikiTable::TABLE_END\E)}
    #             {$1$table_text$2}sx;
    if ($content =~ s{($body_tag\n).*?(\Q$CR::WikiTable::TABLE_END\E)}
                     {$1$table_text$2}s) {
        $LOG->comment("successfully set new table body\n");
    }
    else {
        $LOG->warn("could not find tags '$body_tag\n' and ".
                   "'$CR::WikiTable::TABLE_END'");
    }

    return $content;
}

################################################################################
#
#   Initializes a temporary directory for wiki pages to be checked out to,
#   modified and then committed from. Once this script has finished updating
#   the wiki page, it will remove the temporary directory to make room for
#   future instances of this script (a uniquely named directory must be made
#   for each concurrent instance of this script running otherwise there are
#   file modification conflicts).
#
#   Input:
#       None
#
#   Output:
#       A closure that will remove the appropriate temporary working directory.
#
################################################################################
sub wiki_wrkdir_init {
    my $wrkdir_id = 0;

    while (-e "/tmp/wiki_wrkdir_$wrkdir_id") {
        $wrkdir_id++;
        $LOG->comment("Someone else is updating the wiki...\n".
                      "Checking the $wrkdir_id-th wiki workspace...");
    }

    my $wrkdir = "/tmp/wiki_wrkdir_$wrkdir_id";
    system("mkdir -p $wrkdir");
    chdir($wrkdir);
    system('echo "[DEFAULT]" > .wsvn');
    system('echo "username = smokes1" >> .wsvn');
    system('echo "password = smokes123" >> .wsvn');

    #Return a closure for removing the working directory previously created.
    return sub {
        chdir(q{..});
        system("rm -rf $wrkdir");
    };
}

################################################################################
#
#   Pulls and reads a wiki page from the netapp wiki server, then maintains the
#   contents of the wiki page in association with the name of the wiki page.
#   The original contents of the wiki page is then used later when checking for
#   conflicts during a commit.
#
#   Input:
#       The name of the desired wiki page to be read.
#
#   Output:
#       A string that is the contents of the wiki page.
#
################################################################################
sub wiki_page_read {
    my $page_name = $_[0];

    system("$wsvn_path checkout $page_name");

    my $wikid;
    #enable "slurp" mode. This means that when the contents of the file is read
    #everything will be read until the end of file and stored in a single
    #scalar variable (string)
    local ($INPUT_RECORD_SEPARATOR, $ARG) = (undef, undef);

    open($wikid, '+<', $page_name) or croak("Could not read from wiki file\n");
    my $contents = <$wikid>;
    close($wikid) or croak("Unable to close wiki file\n");

    #Store what has been read from the file now so that we can use it to
    #robustly check for conflicts when we need to commit our changes
    $ORIG_CONTENT{$page_name} = $contents;

    return $contents;
}

################################################################################
#
#   Writes to the wiki page, and commits the updated content. First this
#   subroutine executes another update to be sure the wiki page has not been
#   changed recently. If another update is seen, that update is then used as
#   the base and will be modified using a passed in callback subroutine
#   (closure).
#
#   Input:
#       The name of the wiki page to be written and committed.
#       The content to overwrite the wiki page with. This should be the wiki
#           page's original contents with changes to reflect updates.
#       A reference to a hash of extra arguments:
#           * Number of retries if conflicts are encountered. Default is 4.
#           * A callback subroutine for applying appropriate updates to
#               the wiki page in case the original wiki page content has
#               been modified.
#           * A boolean 'should_clear' determining whether old entries
#               on the page should be removed.
#
#   Output:
#       A closure to remove the temporary working directory when finished.
#
################################################################################
sub wiki_page_write {
    my ($page_name, $content, $args) = @_;

    my $commit_cmd = "$wsvn_path commit $page_name";
    my $update_cmd = "$wsvn_path update $page_name";
    my ($wikid, $in_fd, $out_fd, $err_fd);
    my ($should_retry, $num_retries) = (0, 4);
    if (defined $args && $args->{retries}) {
        $num_retries = $args->{retries};
    }

    #enable "slurp" mode. This means that when you do a read from the file
    #it will read until end of file and store it all as a single string
    local ($INPUT_RECORD_SEPARATOR, $ARG) = (undef, undef);

    do {
        $LOG->comment("Trying to commit updated page '$page_name'...");

        #Update the local version of the desired wiki pages
        my $update_pid = open3($in_fd, $out_fd, $err_fd, $update_cmd);
        waitpid($update_pid, 0);

        #read wiki page in case someone else updated while we were
        #parsing test logs and querying ALM, etc.
        my $wiki_content;
        if (open($wikid, '<', $page_name)) {
            $wiki_content = <$wikid>;
            close($wikid) or croak("Could not close updated wiki file\n");
        }
        else { croak("Could not read updated wiki file\n"); }

        #if the content now on the wiki page doesn't match the content
        #when we started, then we're going to have to re-apply our updates
        #to the new wiki page content
        if ($wiki_content ne $ORIG_CONTENT{$page_name}) {
            $LOG->comment('Conflict occurred when writing to wiki page');

            #add new test results to newly seen updates
            if (defined $args && defined $args->{callback}) {
                $content = $args->{callback}->($wiki_content,
                                               $args->{should_clear});
            }
        }

        #write either content passed in, or updated content if there
        #has been an update after wiki_page_read was originally called
        if (open($wikid, '>', $page_name))  {
            print({$wikid} $content);
            close($wikid) or croak("Unable to close wiki file\n");
        }
        else { croak("Could not write to wiki file\n"); }

        $LOG->comment("committing...");
        my $commit_pid = open3($in_fd, $out_fd, $err_fd, $commit_cmd);

        #If the wsvn tool outputs 'changed on server' or the
        #first letter indicates a conflict (C) then our attempt to
        #commit failed.
        $should_retry = grep {
            $_ =~ /page changed on server|^C/i
        } <$out_fd>;

        waitpid($commit_pid, 0);

    #should_retry becomes false when there were no errors from wsvn
    #when we committed our modified wiki page content, and we want
    #to limit the number of retries so we're not in an infinite loop
    #for any reason
    } while ($should_retry && ($num_retries--) > 0);

    $LOG->comment("Finished committing. Remaining retries: $num_retries");

    return;
}

################################################################################
#
#   Initialize wiki page contents with the appropriate section markers.
#   Unfortunately, I'm not sure how to adapt this to be general for anyone in
#   a way that is simple. So this is based on the SVL searchlight wiki page:
#   http://wikid.netapp.com/w/Searchlight_Nightly_SVL_Automations_Run_Result
#
#   Input:
#       Type of page: 'smoke_holder' or 'summary'. This is because the
#           automation archive and summary archive pages do not have all
#           sections that the summary page has.
#
#   Output:
#       Modified wiki text.
#
################################################################################
sub init_page {
    my ($page_type, $content, $page_markers) = @_;

    #Both the main summary page and the summary holder have the TRIAGE summary
    #section
    if ($page_type ne 'smoke_holder'
     && $content !~ m/$CR::Wiki::PAGE_CONSTANTS{'TRIAGE'}/s) {
        $content .= "$CR::Wiki::PAGE_CONSTANTS{'TRIAGE'}\n".
                    "$CR::Wiki::PAGE_CONSTANTS{'SUMMARY'}\n";

        #Only the main summary page shows BURTs
        if ($page_type eq 'summary') {
            $content .= "==== Product Burt(s) ====\n".
                        "==== Script Burt(s) ====\n";
        }
    }

    #both the main summary page and the smoke holder page have the automation
    #section
    if (($page_type eq 'summary' || $page_type eq 'smoke_holder')
      && $content !~ m/$CR::Wiki::PAGE_CONSTANTS{'AUTOMATION'}/s) {
        $content .= "$CR::Wiki::PAGE_CONSTANTS{'AUTOMATION'}\n";
    }

    #All pages should have the archive link section
    if ($content !~ m/$CR::Wiki::PAGE_CONSTANTS{'ARCHIVES'}/s) {
        $content .= $CR::Wiki::PAGE_CONSTANTS{'ARCHIVES'};
#                    "$page_markers->{ARCHIVE_LINKS}\n";
    }

    return $content;
}

################################################################################
#
#   Clear any tables in the given text that are older than a specified date.
#
#   Input:
#       Text to clear old entries off of (presumably the content of the wiki
#           page)
#       Window of entries from today to keep. This is currently only in days,
#           although can support months and years. Must be a positive number
#           to be accepted.
#
#   Output:
#       Modified wiki text.
#
################################################################################
sub old_entries_clear {
    my ($content, $days_to_live) = @_;

    #for each date header do the following
    while ($content =~ m/(== (\w{3}) (\d\d), \d\d(\d\d) ==)/g) {
        #the date has been captured in the above regex
        my $date_str = $1;
        my ($month, $day, $year) = ($2, $3, $4);

        #convert the month string (e.g. jan, feb, etc.) to its ID (e.g. 0, 1,
        #etc.) then break out of the loop so to be slightly more efficient.
        foreach my $month_ndx (0..$#CR::WikiTable::MONTH_LIST) {
            if ($month =~ m/$CR::WikiTable::MONTH_LIST[$month_ndx]/i) {
                $month = $month_ndx;
                last;
            }
        }

        #given the numeric representation of the year, month, and day of this
        #table, save it as 'old date'. This will be compared to the current
        #date to determine if this table contains old entries.
        my $old_date = sprintf('%02d%02d%02d', $year, $month + 1, $day);

        #determine 2 days before today's date
        my $curr_date = get_date_offset(0, 0, 2);

        #although the default is 2 days, this will change the date if
        #a different time offset is provided
        if ($days_to_live > 0) {
            $curr_date = get_date_offset(0, 0, $days_to_live);
        }

        #if this table does contain old content, then replace everything
        #from this table all the way to the bottom of the page (excluding the
        #archive links at the bottom) from the wiki page.
        if ($old_date lt $curr_date) {
            $content =~ s/$date_str.*== Archives ==\n/== Archives ==\n/s;
            last;
        }
    }

    return $content;
}

################################################################################
#
#   Get the date corresponding to year_off years before this year, month_off
#   months before this month, and day_off days before today. A positive offset
#   retrieves a prior date, whereas a negative offset retrieves a following
#   date.
#
#   Input:
#       Year offset.
#       Month offset.
#       Day offset.
#
#   Output:
#       Year, Month, Day in the format 'yymmdd'
#
################################################################################
sub get_date_offset {
    my ($year_off, $month_off, $day_off) = @_;

    my @pst_time = gmtime(time);
    my ($day, $month, $year) = @pst_time[ 3, 4, 5 ];

    #return a numerical representation of the year, month, and day given
    #the appropriate offset numbers (positive offset means that long before
    #today)
    return sprintf('%02d%02d%02d', ($year % 100) - $year_off,
                   ($month + 1) - $month_off, $day - $day_off);
}

1; #Module was successfully loaded

=pod

=head1 NAME
    Wiki
=head1 DESCRIPTION
    C<Wiki> is module that provides an interface for modifying a wiki page.
    Ideally this module houses code that is general for any wiki page.
=head1 GLOBALS
    C<Wiki> has one global that is important and used throughout the module.
    Wiki::PAGE_CONSTANTS is a hash of markers for sections of the wiki page.
    The markers currently maintained are somewhat particular to the wiki page
    'http://wikid.netapp.com/w/Searchlight_Nightly_SVL_Automations_Run_Result'
    simply because it is difficult to imagine what sections a generic wiki
    may have.
=cut
