#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#$Id$
## @summary
## A suite of useful subroutines for managing a wiki table.
## @author aldrin@netapp.com
## @status shared
## @pod here

package CR::WikiTable;

use strict;
use warnings;

$CR::WikiTable::DEBUG = 0;

$CR::WikiTable::ATTR_KEY = 'attributes';
$CR::WikiTable::TABLE_KEY = 'table';
$CR::WikiTable::STYLE_KEY = 'style';

$CR::WikiTable::BEGIN_TAG = 'begin';
$CR::WikiTable::BODY_TAG  = 'body';
$CR::WikiTable::END_TAG   = 'end';
$CR::WikiTable::TABLE_END = "|}\n";

@CR::WikiTable::MONTH_LIST = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);

#the year field is 2 digits... this YEAR_OFFSET value should be correct for
#another 87 years
my $YEAR_OFFSET = 2000;

################################################################################
#
#   Creates a WikiTable object that will contain all necessary information
#   including:
#       * Table header
#       * Table row data
#       * Table section markup (flags to regex for in the text)
#
#   Input:
#       The name of the module (to be called like 'WikiTable->new(...)')
#       A reference to a hash of table tags. There should be three tags with
#           '$CR::WikiTable::BEGIN_TAG', '$CR::WikiTable::BODY_TAG', and
#           '$CR::WikiTable::END_TAG' as keys and the values for each of these
#           keys will then be used to find desired table data in the wiki text.
#       A reference to a list of attributes (column names) for the header of
#           this wikitable.
#       A string containing wiki table options for this wiki table. For example,
#           'width="100%"' will make the table span the full width of the
#           containing div.
#           'class="wikitable sortable"' makes this wikitable have all
#           css properties of a wikitable (not sure what that entails) as well
#           as making the attributes (columns) of this wikitable sortable (not
#           sure if lexicographically or numerically, or both).
#
#   Output:
#       A reference to a WikiTable object.
#
################################################################################
sub new {
    my ($class, $tags, $attributes, $options) = @_;
    my $self = {};

    $self->{$CR::WikiTable::TABLE_KEY} = ();

    #These tags will allow for parts of the table, and the entire table itself,
    #to be more easily manipulated via regexes
    if (ref $tags eq 'HASH') {
        $self->{$CR::WikiTable::BEGIN_TAG} = $tags->{$CR::WikiTable::BEGIN_TAG};
        $self->{$CR::WikiTable::BODY_TAG}  = $tags->{$CR::WikiTable::BODY_TAG};
        $self->{$CR::WikiTable::END_TAG}   = $tags->{$CR::WikiTable::END_TAG};
    }
    else { print("Table tags must be in the form of a hash\n"); }

    if (ref $attributes eq 'ARRAY') {
        #attributes are column names
        $self->{$CR::WikiTable::ATTR_KEY} = \@{$attributes};
    }
    else { print("Table attributes must be in the form of an array\n"); }

    if (!$options) {
        #these options are css options for wiki table formatting
        $options = 'width="100%" class="wikitable sortable"';
    }

    $self->{$CR::WikiTable::STYLE_KEY} = $options;

    return bless($self, $class);
}

################################################################################
#
#   Extracts a formatted date string from a given build string.
#
#   Input:
#       The build to extract the date from.
#
#   Output:
#       If build has proper month, year, day format then the date of the build
#           is returned in the format 'mm dd, yyyy'.
#           undefined otherwise.
#
################################################################################
sub _build_date_get {
    my ($build) = @_;
    my ($year, $month, $day);

    #build date is in the format 'yymmdd_hhMM' where:
    #   y - year digit
    #   m - month digit
    #   d - day digit
    #   h - hour digit
    #   M - minute digit
    if ($build =~ m/(\d\d)(\d\d)(\d\d)_\d\d\d\d/) {
        ($month, $year, $day) = ($CR::WikiTable::MONTH_LIST[$2 - 1],
                                 $YEAR_OFFSET + $1, $3,);
    }

    #This is to be sure we correctly parsed the build date
    if ($year && $month && $day) {
        return "$month $day, $year";
    }

    return;
}

################################################################################
#
#   Adds new row data to this wikitable's row data. The new row data should be
#   given as a hash (reference) where each attribute (column) of the row is
#   a key-value pair mapping the name of the attribute (column) to its value.
#   Also, if 'build' is an attribute in this WikiTable and this table does not
#   yet have a specified date, then the date and build are set appropriately
#   based on the added (presumably first) row data. This presumes a different
#   table for each date (which corresponds to a build). This may need to be
#   better abstracted in the future.
#
#   Input:
#       A WikiTable object reference.
#       A reference to a hash of arguments. The provided arguments are key-value
#           pairs mapping attribute (column) names to attribute (column) values.
#
#   Output:
#       None
#
################################################################################
sub row_add {
    my ($self, $args) = @_;

    #print(STDERR "adding row...\n");

    my $row_data = {};
    foreach my $attr_name (@{$self->{$CR::WikiTable::ATTR_KEY}}) {
        if (!$self->{date} && $attr_name =~ m/build/i) {
            #The date header is an <h2>
            $self->{date} = '== '._build_date_get($args->{$attr_name})." ==\n";

            #The build header is an <h4>
            $self->{build} = '==== '.$args->{$attr_name}." ====\n";
        }

        $row_data->{$attr_name} = $args->{$attr_name};
    }

    push (@{$self->{$CR::WikiTable::TABLE_KEY}}, $row_data);
    return;
}

################################################################################
#
#   Retrieves all of the row data from this WikiTable.
#
#   Input:
#       A WikiTable object reference.
#
#   Output:
#       A reference to a list where each item in the list is a reference to a
#           hash (%{$reference->[index]})
#
################################################################################
sub rows_get {
    my ($self, $args) = @_;

    return $self->{$CR::WikiTable::TABLE_KEY};
}

################################################################################
#
#   Converts this WikiTable object into a string.
#
#   Input:
#       A WikiTable object reference.
#
#   Output:
#       This WikiTable object as a string in wikitable format.
#
################################################################################
sub to_string {
    my $self = $_[0];

    my $header = $self->{_special_header} ? $self->{_special_header}
                                          : header_string($self);

    #The format for self_string (the string representation of this WikiTable)
    #is:
    #   <BEGIN_TAG> (markup)
    #   <table header> (column name formatting)
    #   <BODY_TAG> (markup)
    #   <row data> (this is a hash where the keys correspond to column names)
    #   <TABLE_END> (the '|}' denoting the end of table content)
    #   <END_TAG> (markup)
    my $self_string = $self->{$CR::WikiTable::BEGIN_TAG}."\n".$header;
    $self_string .= $self->{$CR::WikiTable::BODY_TAG}."\n".body_string($self);
    $self_string .= $CR::WikiTable::TABLE_END."\n".
                    $self->{$CR::WikiTable::END_TAG}."\n";

    return $self_string;
}

################################################################################
#
#   Converts the header of this WikiTable object into a string.
#
#   Input:
#       A WikiTable object reference.
#
#   Output:
#       The columns of this WikiTable object as a string in wikitable format.
#
################################################################################
sub header_string {
    my $self = $_[0];
    my $header = "{| $self->{$CR::WikiTable::STYLE_KEY}\n";

    #$header is the string for the table column headers. Since the names of the
    #columns are in the $CR::WikiTable::ATTR_KEY key of the $self object, we
    #just iterate over these and proper formatting for column names is to start
    #the line with '!' and end the line with a '\n'
    foreach my $attr (@{$self->{$CR::WikiTable::ATTR_KEY}}) {
        $header .= "!$attr\n";
    }

    #Chomp the header so that there isn't unnecessary space inside
    #the table text
    chomp($header);

    return $header;
}

################################################################################
#
#   Converts the content of this WikiTable object into a string.
#
#   Input:
#       A WikiTable object reference.
#
#   Output:
#       The content of this WikiTable object as a string in wikitable format.
#
################################################################################
sub body_string {
    my $self = $_[0];

    my ($row_delim, $attr_delim, $entry_str) = (q{|-}, q{|}, q{});

    foreach my $entry (@{$self->{$CR::WikiTable::TABLE_KEY}}) {
        $entry_str .= "$row_delim\n";

        #Iterate over the column names (keyed in $self by
        #$CR::WikiTable::ATTR_KEY) to be able to access the columns of the
        #table body data
        foreach my $attr_name (@{$self->{$CR::WikiTable::ATTR_KEY}}) {
            if ($entry->{$attr_name} && $entry->{$attr_name} ne q{}) {
                $entry_str .= "$attr_delim$entry->{$attr_name}\n";
            }
            else { $entry_str .= "$attr_delim\n"; }
        }
    }

    return $entry_str;
}

1;

=pod

=head1 NAME
    WikiTable
=head1 DESCRIPTION
    C<WikiTable> is an object-oriented class (more or less) that takes the
    responsibility of maintaining information for a wiki table and correctly
    converting this information to a human-readable format (that is compatible
    with how the wiki displays tables).
=head1 GLOBALS
    C<WikiTable> has several globals that are important and used throughout the
    module.

    =over
    =item *
        ATTR_KEY, TABLE_KEY, STYLE_KEY are all constants for accessing
        specific values from the WikiTable object.
    =item *
        BEGIN_TAG, BODY_TAG, and END_TAG are all constants for consistent,
        standardized access to appropriate tag types for a given wiki table.
        Additionally, TABLE_END is made as a constant so that should any
        extra formatting for tables be desired it need only be changed for this
        variable.
    =item *
        WikiTable::MONTH_LIST is a list of the months for consistent conversion
        from builds/dates to a human-readable month
    =item *
        YEAR_OFFSET is a global for WikiTable only that is used to convert from
        a 2 digit year (e.g. 00 or 01 to 2000 or 2001). This seems silly, but
        when the 4-digit year is incorrect (100 years off) simply changing this
        variable will be all that is needed to make things right again.
=cut
