#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/CR/nfs/NfsOps.pm#3 $
#
## @summary    nfsops_v2_v3_v4 overlap test routines
## @author    dl-cr-dev
## @pod here

=head1 NAME

NfsOps - nfsops_v2_v3_v4 overlap test routines

=head1 SYNOPSIS

 use CR::nfs::NfsOps;

 $out = set_env( setup_h  => \%setup_h,
                 allowed  => [ 3, 4 ],
                 version  => 4,
                 protocol => 'tcp',
                 minor_ver => 1);

 clean_env( %setup_h );

 ($fc, $volume) = nfsops_vol_create( vol_name => 'temp_volume',
                                     vol_size => '20m' );

 nfsops_vol_purge( fc     => $fc,
                   volume => $volume );

 object_remove( hostname => $fc->hostp()->hostname(),
                volume   => $volume,
                path     => $files[0] );

 object_remove_with_client( client_obj => $client_obj,
                            path       => $files[0] );


 @files = file_make( hostname => $fc->hostp()->hostname(),
                     volume   => $volume,
                     name     => $file_name,
                     count    => 1,
                     fill     => 'A' );

 @files = file_make_with_client( client_obj => $client_obj,
                                 path    => $path,
                                 name    => $name,
                                 count   => $count,
                                 attr    => \%attr,
                                 fill    => 'A' );

 @files = file_rename_with_client( client_obj => $client_obj,
                                 fromdir      => $fromdir
                                 fromname     => $fromname
                                 todir        => $todir
                                 toname       => $toname);

 @dirs = dir_make( hostname => $fc->hostp()->hostname(),
                   volume   => $volume,
                   name     => $dir_name,
                   count    => 4 );

 @dirs = dir_make_with_client( client_obj => $client_obj,
                   path     => $path,
                   name     => $dir_name,
                   count    => 4 );

 @links = link_make( hostname => $fc->hostp()->hostname(),
                     volume   => $volume,
                     name     => $link_name,
                     target   => $target_name,
                     count    => 1 );

 @symlinks = symlink_make( hostname => $fc->hostp()->hostname(),
                           volume   => $volume,
                           name     => $symlink_name,
                           target   => $target_name,
                           count    => 1 );

 @symlinks = symlink_make_with_client( client_obj => $client_obj,
                                        name     => $symlink_name,
                                        target   => $target_name,
                                        count    => 1 );

 @devices = device_make( hostname => $fc->hostp()->hostname(),
                         volume   => $volume,
                         name     => $device_name,
                         type     => $dev,
                         count    => 1 ) );

 file_write( hostname => $fc->hostp()->hostname(),
             volume   => $volume,
             path     => $files[0],
             offset   => 0,
             data     => $data,
             stable   => FILE_SYNC,
             stateid  => $stateid );

 file_getattr_with_client( client_obj => $client_obj,
                           path       => $files[0] );

 file_setattr( hostname => $fc->hostp()->hostname(),
               volume   => $volume,
               path     => $files[0],
               attr     => { mode => 0777 } );

 file_setattr_with_client( client_obj => $client_obj,
                           path       => $files[0],
                           attr       => { mode => 0777 } );

 $valid = scalar_validate( op       => 'LOOKUP',
                           version  => $ver,
                           protocol => $proto,
                           expected => $expected,
                           returned => $returned );

 $valid = hash_validate( op       => $op,
                         version  => $ver,
                         protocol => $proto,
                         expected => \%expected,
                         returned => \%returned );

 $valid = hash_validate_change( op       => 'REMOVE',
                                version  => $ver,
                                protocol => $proto,
                                initial  => \%initial,
                                current  => \%current );

 nfsops_lookup_path( client => $client_obj,
                     path   => $files[0] );

 nfsops_lookup_snapshot_path( client => $client_obj,
                              file_obj => $file_object,
                              snapshot => $snapshot_name);
 nfsops_lookup_dongle_file( client => $client_obj,
                            path => $dongle_file_path,);

 %attrs = nfsops_common_attrs( version  => $ver,
                               returned => \%returned );

 %previous = nfsops_save_common_attrs( attr_h => \%attrs );

 my $exp = nfsops_access_bits( bitmask => $expected );

 %stats = nfsops_fsstats( fc     => $fc,
                          volume => $volume );

 %devices = nfsops_set_devices ( device => $device );

 @ids = nfsops_get_new_id( fc   => $fc,
                           type => 'uid' );

 file_read_verify_with_client(client_obj => $client_obj,
                              path       => $path,
                              fill       => 'A');

 file_write_verify_with_client(client_obj => $client_obj,
                               path       => $path,
                               fill       => 'A');

=head1 DESCRIPTION

Provides generic functionality for use in nfsops_v2_v3_v4 overlap testing.

A setup routine is provided which establishes the workspace on the filer
and creates the NFS client objects which will be used during the test. A
corresponding cleanup routine restores the filer workspace and umounts the
NFS client objects.

A set of routines is provided for creating filesystem objects that may be
needed during the test. Validation routines compare results of an NFS call
with a set of expected results and report any discrepancies. Utility
routines provide common operations that may be shared among several tests,
such as looking up a path, converting results to a more readable common
format, or gathering information directly from the filer.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: //depot/prod/test/main/lib/nfs/NfsOps.pm#3 $
  Last revised $Date: 2006/12/08 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package CR::nfs::NfsOps;

use strict;
use warnings;
use base qw(Exporter);

use Tharn;
use Carp qw(croak);
use SPARSE::Util::Logger;
use Cwd qw(getcwd);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF BOOLEAN);
use filer qw(filer_options_get filer_options_set);
use FS::Util::Setup qw(create_vol common_cleanup);
use Volume qw(vol_purge vol_destroy);
use Network::Utils qw(run_getXXbyYY);
use File::Basename qw(basename dirname);
use FileManip qw(remove_file);
use SayWrap qw(sayEOF);
use GenericMount qw(mount umount);
use nfs::GenericExport qw(unexport_path);
use nfs::Utils qw(validate_nfs_version);
use nfs::NfsClient;
use nfs::NfsDef qw(FATTR4_FILEID FATTR4_FSID FATTR4_MODE FATTR4_NUMLINKS
    FATTR4_OWNER FATTR4_OWNER_GROUP FATTR4_RAWDEV FATTR4_SIZE
    FATTR4_TIME_ACCESS FATTR4_TIME_MODIFY FATTR4_TYPE FATTR4_TIME_METADATA
    FILE_SYNC NFBLK NFCHR NFDIR NFLNK UNCHECKED ACCESS_READ ACCESS_LOOKUP
    ACCESS_MODIFY ACCESS_EXTEND ACCESS_DELETE ACCESS_EXECUTE STATEID0
    OPEN4_SHARE_ACCESS_WRITE);
use nfs::NfsXdr qw(xdr_access xdr_bitmap4 xdr_cb_compound xdr_cb_exit
  xdr_cb_info xdr_cb_nlm_granted xdr_cb_nsm_notify xdr_cb_pid xdr_cb_port
  xdr_cb_register xdr_commit xdr_compound xdr_compound_res xdr_cookie
  xdr_cookieverf xdr_create xdr_daemon_info xdr_dump xdr_export xdr_fattr4
  xdr_fhandle xdr_fsinfo xdr_fsstat xdr_getattr xdr_link xdr_lookup xdr_mkdir
  xdr_mknod xdr_mount xdr_nfstime xdr_nlm_lock xdr_nlm_stat xdr_nlm_test
  xdr_opaque xdr_pathconf xdr_read xdr_readdir xdr_readdirplus xdr_readlink
  xdr_remove xdr_rename xdr_sattr xdr_setattr xdr_specdata xdr_stateid4
  xdr_symlink xdr_uint xdr_uint64 xdr_write xdr_name );

use nfs::CompoundRequests qw(validate_minor_ver);
use NACL::C::Node;
use NACL::C::Interface;

my $Node_Obj = NACL::C::Node->find();

my $data_lif_obj = NACL::C::Interface->find(
            command_interface => $Node_Obj,
            filter => {vserver=>$main::VSERVER}
        );
my $Data_IP = $data_lif_obj->get_one_state_attribute('address');

my $ng_check = 0;
my $f = host ($main::FILER);
my $dongle_file_prefix = '.~DOT_CR1992';

if ($f->hosttype() =~ /filer-ng/) {
  use nfs_setup_gx ;
  $ng_check = 1;
}


BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess( Tharn::script_result( @msgs ) ) }
    );
};

our @EXPORT_OK = qw( set_env
                     clean_env
                     nfsops_vol_create
                     nfsops_vol_purge
                     object_remove
                     object_remove_with_client
                     file_make
                     file_make_with_client
                     file_create_with_client
                     file_rename_with_client
                     file_delete_with_client
                     dir_make
                     dir_make_with_client
                     link_make
                     symlink_make
                     symlink_make_with_client
                     device_make
                     file_write
                     file_getattr_with_client
                     file_setattr
                     file_setattr_with_client
                     scalar_validate
                     hash_validate
                     hash_validate_change
                     nfsops_lookup_path
                     nfsops_lookup_snapshot_path
                     nfsops_lookup_dongle_file
                     nfsops_common_attrs
                     nfsops_save_common_attrs
                     nfsops_access_bits
                     nfsops_fsstats
                     nfsops_set_devices
                     nfsops_get_new_id
                     get_gx_v4_id_domain
                     testme
                     file_read_verify_with_client
                     file_write_verify_with_client);

#------------------------------------------------------------------------
# Version for package
#------------------------------------------------------------------------
our $VERSION = '0.01';

#------------------------------------------------------------------------
# Package global variables
#------------------------------------------------------------------------
my $nfs_v4_enable;
my $nfs_v4_acl;
my $nfs_v4_rd;
my $nfs_v4_wd;
my $nfs_mount_rootonly;
my $create_vol_called = 0;

#------------------------------------------------------------------------
# Subroutine prototypes
#------------------------------------------------------------------------
sub set_env(%);
sub clean_env(%);
sub nfsops_vol_create(%);
sub nfsops_vol_purge(%);
sub object_remove(%);
sub object_remove_with_client(%);
sub file_make(%);
sub dir_make(%);
sub dir_make_with_client(%);
sub link_make(%);
sub symlink_make(%);
sub symlink_make_with_client(%);
sub device_make(%);
sub file_write(%);
sub file_getattr_with_client(%);
sub file_setattr(%);
sub file_setattr_with_client(%);
sub scalar_validate(%);
sub hash_validate(%);
sub hash_validate_change(%);
sub nfsops_lookup_path(%);
sub nfsops_lookup_snapshot_path(%);
sub nfsops_common_attrs(%);
sub nfsops_save_common_attrs(%);
sub nfsops_access_bits(%);
sub nfsops_fsstats(%);
sub nfsops_set_devices(%);
sub nfsops_get_new_id(%);
sub gx_get_v4_id_domain();
sub testme();
#------------------------------------------------------------------------

=head2 host

  $hostobj = host( hostspec [,complainspec] [,paramdef]... );

Since "DATA_IP" is not always defined for tharn hosts, and its value is
essential the same as the value for  "DATA_IP1".  Hence, a function is
required in to override Tharn::host, so that when "DATA_IP" is not
defined, its value will be set to the the value for "DATA_IP1".
=cut

#------------------------------------------------------------------------

sub host(@){
    my $hostobj = Tharn::host(@_);

    if ($hostobj->{"DATA_IP1"} && !$hostobj->{"DATA_IP"}){
        $hostobj->{"DATA_IP"} = $hostobj->{"DATA_IP1"};
    }
    return $hostobj;
}

#------------------------------------------------------------------------

=head2 set_env

  $out = set_env( setup_h     => \%setup_h
                  [, allowed  => \@allowed ]
                  [, version  => $version ]
                  [, protocol => 'udp'|'tcp'|'both' ]
                  [, timeout  => $timeout ]
                  [, minor_ver => $minor_ver ]);

NfsOps overlap test setup routine. This routine establishes the environment
for the nfsops tests. It creates and exports the test volume if indicated.
An array of NFS client objects is created, one for each of the tested
version/protocol pairs, and the objects are mounted from the filer.

All setup state is passed back to the test on the 'setup_h' named parameter.
This hash is populated with the following information:
    volume          => the test volume that will be used
    fc              => the connection object to the filer
    nfs_client_objs => an array of client objects that are created
                       ( one for each version/protocol combination )

The following named arguments are accepted:

=over

=item Arguments

=over

=item setup_h

(Required) : Hash reference where setup state is passed back to test.

=item allowed

(Optional) : An array reference indicating what NFS version options are
allowed for this test. An error occurs if the version parameter does not
match one of these values. Default is 'all' versions.

=item version

(Optional) : NFS version(s) that will be tested. Acceptable values are
2, 3, 4, 'not2', 'not3', 'not4', and 'all'. Default is 'all' which will
test versions 2, 3, and 4.

=item protocol

(Optional) : The NFS connection protocol to be tested: 'udp', 'tcp', or
'both'. Default is 'both'.

=item timeout

(Optional) : Timeout value for connection (default is 30).

=item minor_ver

(Optional) : NFSv4 minor version.default is 0

=back

=item Returns

1 if success, otherwise croaks with an error.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub set_env ( % ) {
    my %opts = validate(
        @_,
        {
            setup_h   => { type => HASHREF },
            allowed   => { type => ARRAYREF, default => [ 2, 3, 4 ] },
            version   => { type => SCALAR, default => 'all' },
            protocol  => { type => SCALAR, default => 'both',
                                           regex   => qr/^(both|tcp|udp)$/ },
            timeout   => { type => SCALAR, default => 30 },
            minor_ver => { type => SCALAR, default => 0 },
        }
    );

    my $set_h = $opts{setup_h};
    my ( $vol_name, $hostname, $ver, @nfs_client_objs, @vers, @proto, @minor_ver, %mode );

    logStep('BEGIN: set_env: setup environment for test');
    logStep('set_env: Validate that host is unix');
    my $h = host( 'localhost' );

    croak "Client must be a unix host" if ( $h->hosttype() !~ /^unix/ );

    logStep('set_env: Validate allowed NFS versions');
    @vers = validate_nfs_version( version => $opts{version},
                                  allowed => $opts{allowed} );

    @proto = ($opts{protocol} eq 'both') ? ('tcp', 'udp') : ($opts{protocol});

    @minor_ver= validate_minor_ver (minor_ver => $opts{minor_ver});

    logStep('set_env: Validate NFS version/protocol compatibility');

    croak "Invalid NFS version/protocol combination: " .
            "v$opts{version}, $opts{protocol}"
            if ( $opts{version} eq 4 && $opts{protocol} eq 'udp' );

    if ($ng_check == 0) {
        # Common setup -- connects to the filer, then creates and exports
        # a volume depending on the $main::SETUP_DONE parameter
        logStep('set_env: 7G/7M specific setup');
        logStep('set_env: Create test volume');
        $create_vol_called = 1;
        ($set_h->{fc}, $vol_name) = create_vol() or croak $@;
        $hostname = $set_h->{fc}->hostp()->hostname();

        # Full path for volume
        $set_h->{volume} = "/vol/$vol_name";

        # Save reserved port state
        logStep('set_env: Save root only mount state');
        $nfs_mount_rootonly = filer_options_get( $set_h->{fc},
                                                 'nfs.mount_rootonly' );

        logStep('set_env: Enable mounting from non-reserved ports');
        filer_options_set( $set_h->{fc}, 'nfs.mount_rootonly', 'off' )
                if ( $nfs_mount_rootonly eq 'on' );
    } else {
       logStep ('set_env: C-Mode specific setup');
       my %set_h;
       my $ref1 = nfs_setup_gx::set_env_gx( setup_h  => \%set_h,
                  allowed  => [ 3, 4 ],
                  version  => $opts{version},
                  protocol => $opts{protocol} );

        # Assigning the data ip
        $hostname = $Data_IP;

        logStep("set_env: Using hostname: $hostname");
        $set_h->{fc} = connect($main::FILER_CONN);
        # Full path for volume
        $set_h->{volume} = $ref1->{volume};
    }

    # Create array of NFS clients
    logStep('set_env: Create and mount NFS clients');

    foreach my $proto ( @proto ) {
        foreach my $vers ( @vers ) {
            next if ( $vers == 4 && $proto eq 'udp' );

            if ( $vers != 4 ) {
                push @nfs_client_objs,
                        nfs::NfsClient->new( hostname => $hostname,
                                             version  => $vers,
                                             proto    => $proto,
                                             timeout  => $opts{timeout},
                                             );
            } elsif ( $vers == 4 ) {

                foreach my $minor_ver (@minor_ver) {

                    push @nfs_client_objs,
                                nfs::NfsClient->new( hostname => $hostname,
                                                     version  => $vers,
                                                     proto    => $proto,
                                                     minor_version  => $minor_ver,
                                                     timeout  => $opts{timeout},
                                                     retry => (defined $main::RETRY) ? $main::RETRY : 0,
                                                    )
                }
            }
        }
    }
    # Do not continue if list of NFS client objects is empty
    croak 'NFS client(s) not created' if ( $#nfs_client_objs == -1 );

    $set_h->{nfs_client_objs} = \@nfs_client_objs;

    # Mount the NFS objects
    foreach my $nfs_obj (@nfs_client_objs) {
        $nfs_obj->mount( path => $set_h->{volume} );
    }

    if ( $nfs_client_objs[0]->minor_version() == 1 ) {
        logDbg('Register for callback register service');
        eval {
            $nfs_client_objs[0]->callback_register();
        };
        croak "Could not register for callback service : $@" if ( $@ );
    }

    # Ensure volume is writable (for object creation)
    logStep('set_env: Setattr mode=0777 on test volume');

    if ( $nfs_client_objs[0]->version() == 4 ) {
        $mode{FATTR4_MODE()} = 0777 ;
    } else {
        $mode{mode} = 0777;
    }

    $nfs_client_objs[0]->rpc()->auth_sys_set( uid => 0, gid => 0 );

    if ($ng_check == 0 ) {
        $nfs_client_objs[0]->setattr( fhandle => $nfs_client_objs[0]->mnt_fh(),
                                      attr    => \%mode );
    }
    $nfs_client_objs[0]->rpc()->auth_sys_set();

    logStep('END: set_env');
    return 1;
}

#------------------------------------------------------------------------

=head2 clean_env

  clean_env( [ volume          => $volume ],
             [ fc              => $fc ],
             [ nfs_client_objs => \@nfs_client_objs ],
             [ files           => \@files ],
             [ dirs            => \@dirs ] );

nfsops overlap test cleanup routine. This routine restores the state of
the filer to the pre-test condition. Files and directories created by the test
are removed. NFS client objects that were mounted are unmounted.

The following named arguments are accepted:

=over

=item Arguments

=over

=item volume

(Optional) : The volume on the filer where the nfsops tests are performed.

=item fc

(Optional) : The filer connection object.

=item nfs_client_objs

(Optional) : An array of NFS client objects that have been mounted.

=item files

(Optional) : An array of files that have been created in the volume
directory. These files will be removed.

=item dirs

(Optional) : An array of directories that have been created in the volume
directory. These directories and any subdirectories/files will be removed.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub clean_env ( % ) {
    my %opts = validate(
                        @_,
                        {
                        volume          => { type => SCALAR | UNDEF,   optional => 1 },
                        fc              => { type => OBJECT | UNDEF,   optional => 1 },
                        nfs_client_objs => { type => ARRAYREF | UNDEF, optional => 1 },
                        files           => { type => ARRAYREF | UNDEF, optional => 1 },
                        dirs            => { type => ARRAYREF | UNDEF, optional => 1 },
                        }
                       );
    my ( $h, $curr_dir, $mountpoint );

    logStep("CLEANUP == $main::CLEANUP");
    eval {

        eval {
            # Don't depend on fc parameter received
            logStep('clean_env: Re-connect to the filer');
            $h = host( $main::FILER );
            $opts{fc} = connect( $main::FILER_CONN, -timeout, $main::TIMEOUT );
            $opts{fc}->timeout( $main::CMDTIMEOUT );
        };

        if ($@) {
            logWarn("clean_env: $@");
            undef $opts{fc};
        }

        if ( $main::CLEANUP ne 'none' ) {
            logStep('clean_env: Umount the NFS client objects');

            if ( defined( $opts{nfs_client_objs} ) ) {

                foreach my $nfs_obj ( @{$opts{nfs_client_objs}} ) {
                    eval { $nfs_obj->umount( path => $opts{volume} ) };
                }
            }

            if ( defined( $opts{fc} ) && defined( $opts{volume} ) &&
                ( ( defined( $opts{files} ) && $#{$opts{files}} > -1 ) ||
                ( defined( $opts{dirs} ) && $#{$opts{dirs}} > -1 ) ) ) {

                logStep('clean_env: Mounting Volume');

                if ($h->hosttype !~ /filer-ng/) {
                    #logStep("connection host = $opts{fc}->connhost()");
                    $mountpoint = mount( $opts{fc},
                                         $opts{volume} ) or
                        logWarn("clean_env_first: mount failed: $@");
                } else {
                    my $nfs_server = $Data_IP;
                    eval {
                            $mountpoint =  mount ($nfs_server,
                                                  $main::JUNCTION_PATH,
                                                  mntopts=>'vers=3'
                                                 )
                    };
                    if($@){
                        logWarn("clean_env: mount of $nfs_server failed: $@");
                    }
                }
            }

            if ( defined( $mountpoint ) ) {
                $curr_dir = getcwd();

                if ( chdir $mountpoint == 0 ) {
                    logWarn("clean_env: chdir failed: $!");
                    umount( $mountpoint ) or
                        logWarn("clean_env: umount failed: $@");
                    undef $mountpoint;
                }
            }

            # Remove files that were created
            if ( defined( $mountpoint ) && defined( $opts{files} ) &&
                    $#{$opts{files}} > -1 ) {

                # The unlink would fail if somebody accidentally stuck
                # "/" as a file because it is a directory so there is
                # no need to look for it on the list.
                logStep('clean_env: Removing files');

                unlink @{$opts{files}} or logWarn("clean_env: unlink failed: $!");
                @{$opts{files}} = ();
            }

            # Remove directories that were created
            if ( defined( $mountpoint ) && defined( $opts{dirs} ) && $#{$opts{dirs}} > -1 ) {
                logStep('clean_env: Removing directories');
                # Traverse array in reverse order so that the directories
                # are removed newest to oldest.
                my $dir;

                while ( $dir = pop @{$opts{dirs}} ) {
                    # Remove entire directory tree
                    chmod 0777, $dir;
                    remove_file( file      => "$mountpoint/$dir",
                                 recursive => 1,
                                 client    => 'localhost' );
                }
                @{$opts{dirs}} = ();
            }

            if ( defined( $mountpoint ) ) {
                chdir $curr_dir or logWarn("clean_env: chdir $curr_dir failed: $!");
                umount( $mountpoint ) or logWarn("clean_env: umount failed: $@");
            }

            my $h = host($main::FILER);

            if ($h->hosttype !~ /filer-ng/) {
                # Call common cleanup if create_vol was called

                if ( defined( $opts{fc} ) && $create_vol_called != 0 ) {
                    logStep("Run common cleanup");
                    common_cleanup( $opts{fc} );
                    logWarn("clean_env: $@") if $@;
                }

                if ( $main::CLEANUP ne 'none' && $opts{volume} ) {
                    # Unexport the volume
                    logStep("Unexport volume");
                    unexport_path( $opts{fc}, $opts{volume} ) or croak $@;
                }
            }
        }
    };
    logWarn("clean_env: $@") if $@;

    logStep('END: clean_env');
    return;
}
#------------------------------------------------------------------------

=head2 nfsops_vol_create

  ($fc, $volume) = nfsops_vol_create( vol_name => $vol_name,
                                      vol_size => $vol_size );

Creates a volume with the name and size indicated by the input parameters.
All other volume creation parameters are the default values for the
create_vol function.

Permission on the volume is set to 0777 to allow all operations on the
volume.

The following named arguments are accepted:

=over

=item Arguments

=over

=item vol_name

(Required) : The name of the volume to be created.

=item vol_size

(Required) : The size of the volume to be created. Must conform to
'vol create' limits on the filer.

=item version

(Optional) : The NFS version to use. Defaults to 3

=item protocol

(Optional) : The protocol to use (tcp or udp). Defaults to tcp.

=item minor_ver

(Optional) : Minor version for NFSv4 Default is 0.

=back

=item Returns

An array containing the filer connection object and the volume created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_vol_create ( % ) {
    my %opts = validate(
        @_,
        {
            vol_name  => { type => SCALAR },
            vol_size  => { type => SCALAR },
            version   => { type => SCALAR, default => 3 },
            protocol  => { type => SCALAR, default => 'tcp' },
            minor_ver => { type => SCALAR, default => 0 },
        }
    );
    my ( $fc, $volume, $nfs_obj, $fh_mnt, $mode );

    logDbg('Create the temporary volume');
    my $f=host($main::FILER);

    if ($f->hosttype()!~/filer-ng/) {
        ($fc, $volume) = create_vol( filer_vol    => $opts{vol_name},
                                     vol_size     => $opts{vol_size},
                                     extern_setup => 0 ) or croak $@;

        $volume = "/vol/$volume";
    } else {
        my $server_obj=new Server(node=>$main::FILER);

        eval {
            $server_obj->volume_show (vserver=>$main::VSERVER_GX_NAME,
                                      volume=>$opts{vol_name});
        };

        if ($@){
            $server_obj->volume_create(vserver=>$main::VSERVER_GX_NAME,
                                       volume=>$opts{vol_name},
                                       aggregate=>$main::FILER_GX_AGGR,
                                       size=> $opts{vol_size},
                                       policy=>$main::EXPORT_GX_POL,
                                       'junction-path'=> "/$opts{vol_name}");
        }
        $volume="/$opts{vol_name}";
        host($main::FILER);
        $fc=connect('#ssh');
    }

    logDbg('Set the permission on the temporary volume to allow writing');

    $nfs_obj = ( $opts{version} == 4 ) ?
        nfs::NfsClient->new( hostname => $opts{hostname},
                             version  => $opts{version},
                             proto    => $opts{protocol},
                             minor_version  => $opts{minor_ver},
                             timeout  => (defined $main::NFS_CLIENT_TIMEOUT) ? $main::NFS_CLIENT_TIMEOUT : 30 ,
                             retry => (defined $main::RETRY) ? $main::RETRY : 0
                            )

        : nfs::NfsClient->new( hostname => $opts{hostname},
                               version  => $opts{version},
                               proto    => $opts{protocol},
                              );

    $fh_mnt = $nfs_obj->mount( path => $volume );
    $nfs_obj->rpc()->auth_sys_set( uid => 0,
                                   gid => 0 );

    $mode = ($opts{version} == 4) ? FATTR4_MODE() : 'mode';

    $nfs_obj->setattr( fhandle => $nfs_obj->mnt_fh(),
                       attr    => { $mode => 0777 } );
    $nfs_obj->rpc()->auth_sys_set();

    return ($fc, $volume);
}

#------------------------------------------------------------------------

=head2 nfsops_vol_purge

  nfsops_vol_purge( fc     => $conn,
                    volume => $volume_name );

A simple wrapper to the Volume::vol_purge routine.

For some reason, the vol_purge was not destroying the volume in tests,
so Volume::vol_destroy is called as well.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : The filer connection object.

=item volume

(Required) : The volume name that is to be destroyed.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_vol_purge ( % ) {
    my %opts = validate(
        @_,
        {
            fc       => { type => OBJECT },
            volume   => { type => SCALAR },
            vol_name => { type => SCALAR, optional => 1 },
        }
    );

    my $f=host($main::FILER);

    if ($f->hosttype() =~ /filer-ng/) {

        if (defined($opts{vol_name})) {
            logStep("Purge temporary gx volume: $opts{vol_name}");
            my $server_obj = new Server(node=>$main::FILER);
            $server_obj->volume_unmount(vserver=>$main::VSERVER_GX_NAME, volume=>$opts{vol_name});
            $server_obj->volume_offline(vserver=>$main::VSERVER_GX_NAME, volume=>$opts{vol_name});
            $server_obj->volume_delete(vserver=>$main::VSERVER_GX_NAME, volume=>$opts{vol_name});
        }
    } else {
        vol_purge( $opts{fc}, $opts{volume} );

        # For some reason, vol_purge does not destroy the volume. Do it now.
        vol_destroy( $opts{fc}, $opts{volume} );
    }
    return;
}

#------------------------------------------------------------------------

=head2 object_remove

  object_remove( hostname    => $hostname,
                 volume      => $volume,
                 path        => $path,
                 isDirectory => $isDirectory,
                 version     => $version,
                 protocol    => $protocol,
                 minor_ver   => $minor_ver);

Removes an object from the specified host on the specified volume,
Establishes a new client object mounts the volume, looks up the
directory path.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer from which the object
will be removed.

=item volume

(Required) : The volume on the filer where the object resides.

=item path

(Required) : The path including the object name where the object
resides on the filer, relative to the mount point.

=item isDirectory

(Optional) : Indicates whether the object being removed is a directory.
0 indicates false, all other numbers indicate true. Defaults to false.

=item version

(Optional) : The NFS version to use. Defaults to 3

=item protocol

(Optional) : The protocol to use (tcp or udp). Defaults to tcp.

=item minor_ver

(Optional) : Minor Version for NFSv4. Defaults to 0.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub object_remove ( % ) {
    my %opts = validate(
        @_,
        {
            hostname    => { type => SCALAR },
            volume      => { type => SCALAR },
            path        => { type => SCALAR },
            isDirectory => { type => SCALAR, default => 0 },
            version     => { type => SCALAR, default => 3 },
            protocol    => { type => SCALAR, default => 'tcp' },
            minor_ver   => { type => SCALAR, default => 0 },
        }
    );
    my $f = host("$main::FILER");

    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = $Data_IP;
    }
    my ( $nfs_obj, $fh_mnt, $rmv_failed );

    logDbg('BEGIN: object_remove');
    logDbg('object_remove: Create and mount v3 nfs_obj for object removal');

    $nfs_obj = ( $opts{version} == 4 ) ?
        nfs::NfsClient->new( hostname => $opts{hostname},
                             version  => $opts{version},
                             proto    => $opts{protocol},
                             minor_version  => $opts{minor_ver},
                             timeout  => (defined $main::NFS_CLIENT_TIMEOUT) ? $main::NFS_CLIENT_TIMEOUT : 30,
                             retry => (defined $main::RETRY) ? $main::RETRY : 0
                            )

        : nfs::NfsClient->new( hostname => $opts{hostname},
                               version  => $opts{version},
                               proto    => $opts{protocol}
                              );

    $fh_mnt = $nfs_obj->mount( path => $opts{volume} );

    eval {
        object_remove_with_client( client_obj   => $nfs_obj,
                                   path         => $opts{path},
                                   isDirectory  => $opts{isDirectory}
                                  );
    };
    $rmv_failed = $@ if $@;

    logDbg('object_remove: Umount the nfs_obj');
    $fh_mnt = $nfs_obj->umount( path => $opts{volume} );

    logDbg('END: object_remove');
    croak $rmv_failed if $rmv_failed;
    return;
}

#------------------------------------------------------------------------

=head2 object_remove_with_client

  object_remove( client_obj   => $client_obj,
                 path         => $path,
                 isDirectory  => $isDirectory);

Removes an object from the specified client.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client_obj

(Required) : The client object with which the object
will be removed.

=item path

(Required) : The path including the object name where the object
resides on the filer, relative to the mount point.

=item isDirectory

(Optional) : Indicates whether the object being removed is a directory.
0 indicates false, all other numbers indicate true. Defaults to false.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub object_remove_with_client ( % ) {
    my %opts = validate(
        @_,
        {
            client_obj  => { type => HASHREF },
            path        => { type => SCALAR },
            isDirectory => { type => SCALAR, default => 0 },
        }
    );

    my ( $nfs_obj, $object, $rmv_failed );
    logDbg('BEGIN: object_remove_with_client');
    $nfs_obj = $opts{client_obj};
    $object = basename( $opts{path} );
    logDbg('object_remove_with_client: Get the filehandle of the parent directory');

    nfsops_lookup_path( client => $nfs_obj,
                        path   => dirname( $opts{path} ) );

    eval {
        logDbg('object_remove_with_client: Remove the object');

        if ($opts{isDirectory} && $nfs_obj->version() != 4) {
            $nfs_obj->rmdir( name => $object );
        } else {
            $nfs_obj->remove( name => $object );
        }
    };

    $rmv_failed = $@ if $@;

    logDbg('END: object_remove_with_client');
    croak "object_remove_with_client: $rmv_failed" if $rmv_failed;
    return;
}

#------------------------------------------------------------------------

=head2 file_make

  @files = file_make( hostname    => $hostname,
                      volume      => $volume
                      [, path     => $path ]
                      [, name     => $name ]
                      [, count    => $count ]
                      [, attr     => \%attr ]
                      [, fill     => 'A' ]
                      [, version  => $version]
                      [, protocol => $protocol]
                      [, minor_ver => $minor_ver]);

Creates one or more files on the specified host and volume.
Establishes and mounts a new NFS client object, looks up the filehandle of the
directory in which the new files will be created.

Files will be created with mode 0777 by default, and changed to
the attributes specified in the attr parameter if present.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer on which the objects will be created.

=item volume

(Required) : The volume on the filer where the objects will be created..

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number beginning with 1. Defaults to 'file'.

=item count

(Optional) : The number of objects to be created. Default is 1.

=item attr

(Optional) : A hash of attributes to be applied to the objects when
each is created.

=item fill

(Optional) : A single character to be used to fill the created file(s).
A GETATTR call will return the size of the file, and the character will
be written repetitively to fill the file to the size specified in the GETATTR.

=item version

(Optional) : The NFS version to use. Defaults to 3

=item protocol

(Optional) : The protocol to use (tcp or udp). Defaults to tcp.

=item minor_ver

(Optional) : Minor Version for NFSv4. Defaults to 0.

=back

=item Returns

An array containing the file names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub file_make ( % ) {
    my %opts = validate(
        @_,
        {
            hostname  => { type => SCALAR },
            volume    => { type => SCALAR },
            path      => { type => SCALAR, default => '' },
            name      => { type => SCALAR, default => 'file' },
            count     => { type => SCALAR, default => 1, regex => qr/^[1-9]\d*$/ },
            attr      => { type => HASHREF,default => {} },
            fill      => { type => SCALAR, default => 'A', regex => qr/^.$/ },
            version   => { type => SCALAR, default => 3 },
            protocol  => { type => SCALAR, default => 'tcp' },
            minor_ver => { type => SCALAR, default => 0 },
            }
    );
    my $f = host("$main::FILER");

    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = $Data_IP;
    }
    my ( $nfs_obj, $timeout, $fh_mnt, $curr_fh, @files, $mode );
    logDbg('BEGIN: file_make');
    # Allow a longer timeout when using client object to fill a file
    $timeout = ( $opts{fill} ) ? 300 : 30;
    logDbg('file_make: Create and mount v4 nfs_obj for file creation');

    $nfs_obj = ( $opts{version} == 4 ) ?
        nfs::NfsClient->new( hostname => $opts{hostname},
                             version  => $opts{version},
                             proto    => $opts{protocol},
                             minor_version  => $opts{minor_ver},
                             timeout  => (defined $main::NFS_CLIENT_TIMEOUT) ? $main::NFS_CLIENT_TIMEOUT : 30,
                             retry => (defined $main::RETRY) ? $main::RETRY : 0
                            )

        : nfs::NfsClient->new( hostname => $opts{hostname},
                               version  => $opts{version},
                               proto    => $opts{protocol},
                              );
    $fh_mnt = $nfs_obj->mount( path => $opts{volume} );

    eval {
        @files = file_make_with_client( client_obj   => $nfs_obj,
                                           path         => $opts{path},
                                           name         => $opts{name},
                                           count        => $opts{count},
                                           attr         => $opts{attr},
                                           fill         => $opts{fill},
                                         );
    };

    my $fm_failed = $@ if $@;
    logDbg('END: file_make_with_client');
    croak "file_make_with_client: $fm_failed" if $fm_failed;

    logDbg('file_make: Umount the nfs_obj');
    $fh_mnt = $nfs_obj->umount( path => $opts{volume} );

    logDbg('END: file_make');
    return @files;
}

#------------------------------------------------------------------------

=head2 file_make_with_client

  @files = file_make_with_client( client_obj => $client_obj,
                                  [, path     => $path ]
                                  [, name     => $name ]
                                  [, count    => $count ]
                                  [, attr     => \%attr ]
                                  [, fill     => 'A' ]
                                 );

Creates one or more files using the specified client object.
The client object must have a mounted volume prior to calling
this subroutine. Looks up the filehandle of the directory in
which the new files will be created.

Files will be created with mode 0777 by default, and changed to
the attributes specified in the attr parameter if present.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client_obj

(Required) : The client object to use when performing NFSv3 ops. This
subroutine assumes that there is already a volume mounted using NFSv3 on
this client.

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number beginning with 1. Defaults to 'file'.

=item count

(Optional) : The number of objects to be created. Default is 1.

=item attr

(Optional) : A hash of attributes to be applied to the objects when
each is created.

=item fill

(Optional) : A single character to be used to fill the created file(s).
A GETATTR call will return the size of the file, and the character will
be written repetitively to fill the file to the size specified in the GETATTR.

=back

=item Returns

An array containing the file names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub file_make_with_client ( % ) {
    my %opts = validate(
        @_,
        {
            client_obj => { type => OBJECT},
            path       => { type => SCALAR, default => ''},
            name       => { type => SCALAR, default => 'file' },
            count      => { type => SCALAR, default => 1, regex => qr/^[1-9]\d*$/ },
            attr       => { type => HASHREF,default => {} },
            fill       => { type => SCALAR, default => 'A' },
        }
    );

    my ( $curr_fh, @files, $mode,$create,$open );
    logDbg('BEGIN: file_make');

    if ( $opts{path} ) {
        logDbg('file_make: Get file fh from path');
        nfsops_lookup_path( client => $opts{client_obj},
                            path   => $opts{path} );
    }
    $curr_fh = $opts{client_obj}->curr_fh();

    logDbg("file_make: Create $opts{count} files and set the attributes");
    # Call SETATTR separately to allow setting atime, mtime at creation

    foreach my $iter ( 1 .. $opts{count} ) {
        my $file_name;
        # Add a dongle only if there are more than one with
        # same name
        if ($opts{count} != 1) {
            $file_name = "$opts{name}\.$iter";
        } else {
            $file_name = $opts{name};
        }

        if ( $opts{client_obj}->version() == 4) {
            $open = $opts{client_obj}->open( dhandle => $curr_fh,
                                             name    => $file_name,
                                             access  => OPEN4_SHARE_ACCESS_WRITE ,
                                             cr_mode => UNCHECKED,
                                             attr    => { FATTR4_MODE() => 0777 },
                                           );

        } else {
            $create = $opts{client_obj}->create( dhandle => $curr_fh,
                                                 name    => $file_name,
                                                 cr_mode => UNCHECKED,
                                                 attr    => { mode => 0777 });
        }

        push @files, $file_name;
        $opts{client_obj}->setattr( attr => $opts{attr} ) if ( $opts{attr} );

        if ( $opts{fill} ) {
            my $getattr = $opts{client_obj}->getattr();
            my $size = ($opts{client_obj}->version() == 4) ?
                $getattr->{fattr}{FATTR4_SIZE()} : $getattr->{fattr}{'size'};

            my $offset = 0;

            while ( $offset < $size - 8192 ) {
## TODO: optimise the code by using stateid => defined($open) ? $open->{stateid} : 0

                my $written = ($opts{client_obj}->version() == 4) ?
                        $opts{client_obj}->write( offset  => $offset,
                                                  data    => $opts{fill}x8192,
                                                  stable  => FILE_SYNC ,
                                                  stateid => $open->{stateid} )
                      : $opts{client_obj}->write( offset => $offset,
                                                  data   => $opts{fill}x8192,
                                                  stable => FILE_SYNC );

                $offset += ($opts{client_obj}->version() == 4) ?
                    $written->{results}[-1]{result}{count} : $written->{count};
            }
            ($opts{client_obj}->version() == 4) ?
                $opts{client_obj}->write( offset  => $offset,
                                          data    => $opts{fill}x( $size-$offset ),
                                          stable  => FILE_SYNC ,
                                          stateid => $open->{stateid} ) :
                $opts{client_obj}->write( offset => $offset,
                                          data   => $opts{fill}x( $size-$offset ),
                                          stable => FILE_SYNC );
        }

        $opts{client_obj}->close( stateid => $open->{stateid} )
                if ( $opts{client_obj}->version() == 4 );

    }
    logDbg('END: file_make');
    return @files;
}

#------------------------------------------------------------------------

=head2 file_create_with_client

  @files = file_create_with_client( client_obj => $client_obj,
                                  [, path     => $path ]
                                  [, name     => $name ]
                                  [, count    => $count ]
                                  [, cr_mode  => $cr_mode ]
                                  [, attr     => \%attr ]
                                 );

Creates one or more files using the specified client object.
The client object must have a mounted volume prior to calling
this subroutine. Looks up the filehandle of the directory in
which the new files will be created.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client_obj

(Required) : The client object to use when performing NFSv3 ops. This
subroutine assumes that there is already a volume mounted using NFSv3 on
this client.

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number beginning with 1. Defaults to 'file'.

=item count

(Optional) : The number of objects to be created. Default is 1.

=item attr

(Optional) : A hash of attributes to be applied to the objects when
each is created.

=item cr_mode

(Optional) : Create mode: UNCHECKED (default), GUARDED, EXCLUSIVE.
Used only in NFSv3.

=back

=item Returns

An array containing the file names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub file_create_with_client ( % ) {
    my %opts = validate(
        @_,
        {
            client_obj => { type => OBJECT },
            path       => { type => SCALAR, default => '' },
            name       => { type => SCALAR, default => 'file' },
            count      => { type => SCALAR, default => 1, regex => qr/^[1-9]\d*$/ },
            cr_mode    => { type => SCALAR, default => UNCHECKED, regex => qr/^\d+$/},
            attr       => { type => HASHREF,default => {} },
        }
    );

    my ( $curr_fh, @files, $mode, $open );
    logDbg('BEGIN: file_create_with_client');

    if ( $opts{path} ) {
        logDbg('file_create_with_client: Get file fh from path');
        nfsops_lookup_path( client => $opts{client_obj},
                            path   => $opts{path} );
    }
    $curr_fh = $opts{client_obj}->curr_fh();
    logDbg("file_create_with_client: Create $opts{count} files and set the attributes");

    foreach my $iter ( 1 .. $opts{count} ) {
        my $file_name = "$opts{name}\.$iter";
        # Add a dongle only if there are more than one with
        # same name
        if ($opts{count} == 1) {
            $file_name = $opts{name};
        }

        my $create = ($opts{client_obj}->version() ==4) ?
                $open = $opts{client_obj}->open( dhandle => $curr_fh,
                                         name    => $file_name,
                                         attr    => $opts{attr},
                                         cr_mode => $opts{cr_mode},
                                       ) :

                $opts{client_obj}->create( dhandle => $curr_fh,
                                           name    => $file_name,
                                           cr_mode => $opts{cr_mode},
                                           attr    => $opts{attr});

        push @files, $file_name;

        $opts{client_obj}->close( stateid => $open->{stateid} )
                if ( $opts{client_obj}->version() == 4 );
    }

    logDbg('END: file_create_with_client');
    return @files;
}

#------------------------------------------------------------------------

=head2 file_rename_with_client

  @files = file_rename_with_client( client_obj => $client_obj,
                                  [, fromdir   => $fromdir ]
                                  [, fromname  => $fromname ]
                                  [, todir     => $todir ]
                                  [, toname    => $toname ]);

Renames a file using the specified client object.
The client object must have a mounted volume prior to calling
this subroutine. Looks up the filehandle of the directory in
which the original file exists and where the renamed file will
be created.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client_obj

(Required) : The client object to use when performing NFS ops. This
subroutine assumes that there is already a volume mounted using NFS on
this client.

=item fromdir

(Required) : The path on the filer where the file to be renamed exists,
relative to the mount point.

=item fromname

(Required) : The base name to be used for the object/file which is
to be renamed.

=item todir

(Optional) : The path on the filer where the renamed file will exist,
relative to the mount point. Defaults to fromdir.

=item toname

(Required) : The base name of the file after rename

=back

=item Returns

An array containing the renamed file

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub file_rename_with_client ( % ) {
    my %opts = validate(
        @_,
        {
            client_obj => { type => OBJECT },
            fromdir    => { type => SCALAR },
            fromname   => { type => SCALAR },
            todir      => { type => SCALAR, optional => 0 },
            toname     => { type => SCALAR },
        }
    );

    my ( $from_fh, $to_fh );
    logDbg('BEGIN: file_rename_with_client');
    logDbg('file_rename_with_client: Get file fh from fromdir');

    nfsops_lookup_path( client => $opts{client_obj},
                        path   => $opts{fromdir} );
    $from_fh = $opts{client_obj}->curr_fh();
    $to_fh = $from_fh;

    if ( $opts{todir} ) {
        logDbg('file_rename_with_client: Get file fh from todir');
        nfsops_lookup_path( client => $opts{client_obj},
                            path   => $opts{todir} );
        $to_fh = $opts{client_obj}->curr_fh();
    }

    my $create = $opts{client_obj}->rename( fromdir => $from_fh,
                                            fromname => $opts{fromname},
                                            todir => $to_fh,
                                            toname => $opts{toname});

    logDbg('END: file_rename_with_client');
    return $opts{toname};
}

#------------------------------------------------------------------------

=head2 dir_make_with_client

  @dirs = dir_make_with_client( client_obj => $client_obj,
                                [, path     => $path ]
                                [, name     => $name ]
                                [, count    => $count ]
                                [, attr     => \%attr ]
                               );

Creates one or more directories on the specified host and volume.
Establishes and mounts a new NFS client object, looks up the filehandle
of the directory in which the new directories will be created.

Directories will be created with mode 0777 by default, and changed to
the attributes specified in the attr parameter if present.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client_obj

(Required) : The client object to use when performing NFS ops. This
subroutine assumes that there is already a volume mounted using NFS on
this client.

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number. Defaults to 'dir'.

=item count

(Optional) : The number of objects to be created. Defaults to 1.

=item attr

(Optional) : A hash of attributes to be applied to the objects when
each is created.

=back

=item Returns

An array containing the directory names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub dir_make_with_client ( % ) {
    my %opts = validate(
        @_,
        {
            client_obj => { type => OBJECT },
            path       => { type => SCALAR, default => '' },
            name       => { type => SCALAR, default => 'dir' },
            count      => { type => SCALAR, default => 1, regex => qr/^[1-9]\d*$/ },
            attr       => { type => HASHREF, default => {} },
        }
    );
    my ( $curr_fh, @dirs, $mode );
    logStep('BEGIN: dir_make_with_client');

    if ( $opts{path} ) {
        logStep('dir_make_with_client: Get directory fh from path');
        nfsops_lookup_path( client => $opts{client_obj},
                            path   => $opts{path} );
    }
    $curr_fh = $opts{client_obj}->curr_fh();
    logStep("dir_make_with_client: Create $opts{count} directories and set the attributes");

    foreach my $iter ( 1 .. $opts{count} ) {
        my $dir_name = "$opts{name}\.$iter";

        # Add a dongle only if there are more than one with
        # same name
        if ($opts{count} == 1) {
            $dir_name = $opts{name};
        }

        $opts{client_obj}->version() == 4 ?
            $opts{client_obj}->create( dhandle => $curr_fh,
                                       name => $dir_name,
                                       type => NFDIR ,
                                       attr => { FATTR4_MODE() => 0777 } )

            : $opts{client_obj}->mkdir( dhandle => $curr_fh,
                                        name => $dir_name,
                                        attr => { 'mode' => 0777 } );

        push @dirs, $dir_name;

        $opts{client_obj}->setattr( attr => $opts{attr} ) if ( $opts{attr} );
    }

    logStep('END: dir_make_with_client');
    return @dirs;
}

#------------------------------------------------------------------------

=head2 dir_make

  @dirs = dir_make( hostname => $hostname,
                    volume      => $volume
                    [, path     => $path ]
                    [, name     => $name ]
                    [, count    => $count ]
                    [, attr     => \%attr ]
                    [, version  => $version]
                    [, protocol => $protocol]
                    [, minor_ver => $minor_ver]);

Creates one or more directories on the specified host and volume.
Establishes and mounts a new NFS client object, looks up the filehandle
of the directory in which the new directories will be created.

Directories will be created with mode 0777 by default, and changed to
the attributes specified in the attr parameter if present.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer on which the objects will be created.

=item volume

(Required) : The volume on the filer where the objects will be created.

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number. Defaults to 'dir'.

=item count

(Optional) : The number of objects to be created. Defaults to 1.

=item attr

(Optional) : A hash of attributes to be applied to the objects when
each is created.

=item version

(Optional) : The NFS version to use. Defaults to 3

=item protocol

(Optional) : The protocol to use (tcp or udp). Defaults to tcp.

=item minor_ver

(Optional) : Minor Version for NFSv4. Defaults to 0.

=back

=item Returns

An array containing the directory names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub dir_make ( % ) {
    my %opts = validate(
        @_,
        {
            hostname  => { type => SCALAR },
            volume    => { type => SCALAR },
            path      => { type => SCALAR, default => ''},
            name      => { type => SCALAR, default => 'dir' },
            count     => { type => SCALAR, default => 1, regex => qr/^[1-9]\d*$/ },
            attr      => { type => HASHREF,default => {} },
            version   => { type => SCALAR,default => 3 },
            protocol  => { type => SCALAR, default => 'tcp' },
            minor_ver => { type => SCALAR, default => 0 },
        }
    );
    my $f = host("$main::FILER");

    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = $Data_IP;
    }

    my ( $nfs_obj, $fh_mnt, $curr_fh, @dirs, $mode );
    logStep('BEGIN: dir_make');
    logStep('dir_make: Create and mount nfs_obj for directory creation');

    $nfs_obj = ( $opts{version} == 4 ) ?
        nfs::NfsClient->new( hostname => $opts{hostname},
                             version  => $opts{version},
                             proto    => $opts{protocol},
                             minor_version  => $opts{minor_ver},
                             timeout  => (defined $main::NFS_CLIENT_TIMEOUT) ? $main::NFS_CLIENT_TIMEOUT : 30,
                             retry => (defined $main::RETRY) ? $main::RETRY : 0
                            )

        : nfs::NfsClient->new( hostname => $opts{hostname},
                               version  => $opts{version},
                               proto    => $opts{protocol},
                              );
    $fh_mnt = $nfs_obj->mount( path => $main::JUNCTION_PATH );

    eval {
         @dirs = dir_make_with_client( client_obj   => $nfs_obj,
                                       path         => $opts{path},
                                       name         => $opts{name},
                                       count        => $opts{count},
                                       attr         => $opts{attr},
                                     );
    };

    my $dm_failed = $@ if $@;
    logStep('END: dir_make_with_client');
    croak "dir_make_with_client: dm_failed with $@" if $dm_failed;

    logStep('dir_make: Umount the nfs_obj');
    $fh_mnt = $nfs_obj->umount( path => $opts{volume} );

    logStep('END: dir_make');
    return @dirs;
}

#------------------------------------------------------------------------

=head2 link_make

  @links = link_make( hostname    => $hostname,
                      volume      => $volume,
                      target      => $target
                      [, path     => $path ]
                      [, name     => $name ]
                      [, count    => $count ]
                      [, version  => $version]
                      [, protocol => $protocol]
                      [, minor_ver => $minor_ver]);

Creates one or more hard links on the specified host and volume.
Establishes and mounts a new NFS client object, looks up the
filehandle of the directory in which the new links will be created.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer on which the objects
will be created.

=item volume

(Required) : The volume on the filer where the objects will be created..

=item target

(Required) : The path/name of the target for the links. All created
objects will point to the same target object

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number. Defaults to 'link'.

=item count

(Optional) : The number of objects to be created. Defaults to 1.

=item version

(Optional) : The NFS version to use. Defaults to 3

=item protocol

(Optional) : The protocol to use (tcp or udp). Defaults to tcp.

=item minor_ver

(Optional) : Minor Version for NFSv4. Defaults to 0.

=back

=item Returns

An array containing the link names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub link_make ( % ) {
    my %opts = validate(
        @_,
        {
            hostname  => { type => SCALAR },
            volume    => { type => SCALAR },
            target    => { type => SCALAR },
            path      => { type => SCALAR, default => '' },
            name      => { type => SCALAR, default => 'link' },
            count     => { type => SCALAR, default => 1, regex => qr/^[1-9]\d*$/ },
            version   => { type => SCALAR, default => 3 },
            protocol  => { type => SCALAR, default => 'tcp' },
            minor_ver => { type => SCALAR, default => 0 },
        }
    );
    my $f = host("$main::FILER");

    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = $Data_IP;
    }
    my ( $nfs_obj, $fh_mnt, $dir_fh, @links );
    logDbg('BEGIN: link_make');
    logDbg('link_make: Create and mount nfs_obj for link creation');

    $nfs_obj = ( $opts{version} == 4 ) ?
        nfs::NfsClient->new( hostname => $opts{hostname},
                             version  => $opts{version},
                             proto    => $opts{protocol},
                             minor_version  => $opts{minor_ver},
                             timeout  => (defined $main::NFS_CLIENT_TIMEOUT) ? $main::NFS_CLIENT_TIMEOUT : 30,
                             retry => (defined $main::RETRY) ? $main::RETRY : 0,
                            )

        : nfs::NfsClient->new( hostname => $opts{hostname},
                               version  => $opts{version},
                               proto    => $opts{protocol},
                              );

    $fh_mnt = $nfs_obj->mount( path => $opts{volume} );

    if ( $opts{path} ) {
        logDbg('link_make: Get fh of target directory from path');
        nfsops_lookup_path( client => $nfs_obj,
                            path   => $opts{path} );
    }
    $dir_fh = $nfs_obj->curr_fh();

    logDbg('link_make: Lookup the target filehandle');
    nfsops_lookup_path( client => $nfs_obj,
                        path   => $opts{target} );
    my $fhandle = $nfs_obj->curr_fh();

    logDbg("link_make: Create $opts{count} links");

    foreach my $iter ( 1 .. $opts{count} ) {
        my $link_name = "$opts{name}\.$iter";
        $nfs_obj->link( fhandle => $fhandle,
                        todir   => $dir_fh,
                        name    => $link_name );
        push @links, $link_name;

        $nfs_obj->lookup( dhandle => $dir_fh,
                          name    => $link_name );
        $nfs_obj->setattr( attr => $opts{attr} ) if $opts{attr};
    }

    logDbg('link_make: Umount the nfs_obj');
    $fh_mnt = $nfs_obj->umount( path => $opts{volume} );

    logDbg('END: link_make');
    return @links;
}

#------------------------------------------------------------------------

=head2 symlink_make

  @symlinks = symlink_make( hostname => $hostname,
                            volume   => $volume,
                            target   => $target
                            [, path  => $path ]
                            [, name  => $name ]
                            [, attr  => \%attr ]
                            [, count => $count ]
                            [, version => $version]
                            [,protocol => $protocol]
                            [, minor_ver => $minor_ver]);

Creates one or more symbolic links on the specified host and volume.
Establishes and mounts a new NFS client object, looks up the filehandle
of the directory in which the new symlinks will be created.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer on which the objects
will be created.

=item volume

(Required) : The volume on the filer where the objects will be created..

=item target

(Required) : The path/name of the target for the symlinks. All created
objects will point to the same target object

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number. Defaults to 'symlink'.

=item attr

(Optional) : A hash of attributes to be applied to the objects when
each is created.

=item count

(Optional) : The number of objects to be created. Defaults to 1.

=item version

(Optional) : The NFS version to use. Defaults to 3

=item protocol

(Optional) : The protocol to use (tcp or udp). Defaults to tcp.

=item minor_ver

(Optional) : Minor Version for NFSv4. Defaults to 0.

=back

=item Returns

An array containing the symlink names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub symlink_make ( % ) {
    my %opts = validate(
        @_,
        {
            hostname => { type => SCALAR },
            volume   => { type => SCALAR },
            target   => { type => SCALAR },
            path     => { type => SCALAR, default => '' },
            name     => { type => SCALAR,  default  => 'symlink' },
            attr     => { type => HASHREF, default  => {} },
            count    => { type => SCALAR,  default  => 1,
                                           regex    => qr/^[1-9]\d*$/ },
            version     => { type => SCALAR, default => 3 },
            protocol    => { type => SCALAR, default => 'tcp' },
            minor_ver   => { type => SCALAR, default => 0 },
        }
    );
    my $f = host("$main::FILER");

    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = $Data_IP;
    }
    my ( $nfs_obj, $fh_mnt, $curr_fh, @links );
    logDbg('BEGIN: symlink_make');
    logDbg('symlink_make: Create and mount v4 nfs_obj for symlink creation');

    $nfs_obj = ( $opts{version} == 4 ) ?
        nfs::NfsClient->new( hostname => $opts{hostname},
                             version  => $opts{version},
                             proto    => $opts{protocol},
                             minor_version  => $opts{minor_ver},
                             timeout  => (defined $main::NFS_CLIENT_TIMEOUT) ? $main::NFS_CLIENT_TIMEOUT : 30,
                             retry => (defined $main::RETRY) ? $main::RETRY : 0
                            )

        : nfs::NfsClient->new( hostname => $opts{hostname},
                               version  => $opts{version},
                               proto    => $opts{protocol},
                              );
    $fh_mnt = $nfs_obj->mount( path => $opts{volume} );

    eval {
         @links = symlink_make_with_client(client_obj   => $nfs_obj,
                                           target       => $opts{target},
                                           path         => $opts{path},
                                           name         => $opts{name},
                                           attr         => $opts{attr},
                                           count        => $opts{count},
                                          );
    };

    my $symlinkmake_failed = $@ if $@;
    logDbg('END: symlink_make_with_client');

    croak "symlink_make_with_client: dm_failed" if $symlinkmake_failed;

    logDbg('symlink_make: Umount the nfs_obj');
    $fh_mnt = $nfs_obj->umount( path => $opts{volume} );

    logDbg('END: symlink_make');
    return @links;
}

#------------------------------------------------------------------------

=head2 symlink_make_with_client

  @symlinks = symlink_make_with_client( client_obj => $nfs_obj,
                                        target   => $target
                                        [, path  => $path ]
                                        [, name  => $name ]
                                        [, attr  => \%attr ]
                                        [, count => $count ] );

Creates one or more symbolic links on the specified host and volume.
Establishes and mounts a new NFS client object, looks up the filehandle
of the directory in which the new symlinks will be created.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client_obj

(Required) : The client object to use when performing NFS ops. This
subroutine assumes that there is already a volume mounted using NFS on
this client.

=item target

(Required) : The path/name of the target for the symlinks. All created
objects will point to the same target object

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number. Defaults to 'symlink'.

=item attr

(Optional) : A hash of attributes to be applied to the objects when
each is created.

=item count

(Optional) : The number of objects to be created. Defaults to 1.

=back

=item Returns

An array containing the symlink names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub symlink_make_with_client ( % ) {
    my %opts = validate(
        @_,
        {
            client_obj => { type => OBJECT },
            target     => { type => SCALAR },
            path       => { type => SCALAR, default => '' },
            name       => { type => SCALAR, default  => 'symlink' },
            attr       => { type => HASHREF,default  => {} },
            count      => { type => SCALAR, default  => 1, regex => qr/^[1-9]\d*$/ },
        }
    );

    my ( $curr_fh, @links );
    logDbg('BEGIN: symlink_make_with_client');

    if ( $opts{path} ) {
        logDbg('symlink_make_with_client: Get file fh from path');
        nfsops_lookup_path( client => $opts{client_obj},
                            path   => $opts{path} );
    }
    $curr_fh = $opts{client_obj}->curr_fh();
    logDbg("symlink_make_with_client: Create $opts{count} symlinks");

    foreach my $iter ( 1 .. $opts{count} ) {
        my $symlink_name = "$opts{name}\.$iter";

        # Add a dongle only if there are more than one with
        # same name
        if ($opts{count} == 1) {
            $symlink_name = $opts{name};
        }

        ($opts{client_obj}->version() == 4) ?
            $opts{client_obj}->create(dhandle => $curr_fh,
                                      name    => $symlink_name,
                                      data    => $opts{target},
                                      type    => NFLNK )

            : $opts{client_obj}->symlink( dhandle => $curr_fh,
                                          name    => $symlink_name,
                                          data    => $opts{target} );
        push @links, $symlink_name;

        $opts{client_obj}->lookup( dhandle => $curr_fh,
                                   name    => $symlink_name );
        $opts{client_obj}->setattr( attr => $opts{attr} ) if $opts{attr};
    }

    logDbg('END: symlink_make_with_client');
    return @links;
}

#------------------------------------------------------------------------

=head2 device_make

  @devs = device_make( hostname    => $hostname,
                       volume      => $volume,
                       type        => $type
                       [, path     => $path ]
                       [, name     => $name ]
                       [, attr     => \%attr ]
                       [, count    => $count ]
                       [, version  => $version]
                       [, protocol => $protocol]
                       [, minor_ver => $minor_ver]);

Creates one or more device files on the specified host and volume.
Establishes and mounts a new NFS client object, looks up the filehandle
of the directory in which the new devices will be created.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer on which the objects will be created.

=item volume

(Required) : The volume on the filer where the objects will be created..

=item type

(Required) : The type of device to be created. Can be one of NFCHR,
NFBLK, NFSOCK, or NFFIFO.

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number. Defaults to 'dev'.

=item attr

(Optional) : A hash of attributes to be applied to the objects when
each is created.

=item count

(Optional) : The number of objects to be created. Defaults to 1.

=item version

(Optional) : The NFS version to use. Defaults to 3

=item protocol

(Optional) : The protocol to use (tcp or udp). Defaults to tcp.

=item minor_ver

(Optional) : Minor Version for NFSv4. Defaults to 0.

=back

=item Returns

An array containing the device names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub device_make ( % ) {
    my %opts = validate(
        @_,
        {
            hostname  => { type    => SCALAR },
            volume    => { type    => SCALAR },
            type      => { type    => SCALAR },
            path      => { type    => SCALAR, default => '' },
            name      => { type    => SCALAR,  default => 'dev' },
            attr      => { type    => HASHREF,
                           default => { FATTR4_MODE() => 0777 } },
            count     => { type    => SCALAR, default => 1, regex => qr/^[1-9]\d*$/ },
            version   => { type => SCALAR, default => 3 },
            protocol  => { type => SCALAR, default => 'tcp' },
            minor_ver => { type => SCALAR, default => 0 },
        }
    );
    my $f = host("$main::FILER");

    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = $Data_IP;
    }
    my ( $fh_mnt, $nfs_obj, $h, $rdev, $curr_fh, @devices, $mode );
    my ( $major, $minor ) = ( 0,0 );

    logDbg('device_make: Create and mount v4 nfs_obj for device creation');

    $nfs_obj = ( $opts{version} == 4 ) ?
        nfs::NfsClient->new( hostname => $opts{hostname},
                             version  => $opts{version},
                             proto    => $opts{protocol},
                             minor_version  => $opts{minor_ver},
                             timeout  => (defined $main::NFS_CLIENT_TIMEOUT) ? $main::NFS_CLIENT_TIMEOUT : 30,
                             retry => (defined $main::RETRY) ? $main::RETRY : 0
                            )

        : nfs::NfsClient->new( hostname => $opts{hostname},
                               version  => $opts{version},
                               proto    => $opts{protocol},
                              );
    $fh_mnt = $nfs_obj->mount( path => $opts{volume} );

    if ( $opts{path} ) {
        logDbg('device_make: Get directory fh from path');
        nfsops_lookup_path( client => $nfs_obj,
                            path   => $opts{path} );
    }
    $curr_fh = $nfs_obj->curr_fh();

    logDbg("device_make: Create $opts{count} devices");
    # Device major, minor numbers are taken from the localhost

    if ( $opts{type} == NFCHR || $opts{type} == NFBLK ) {
        $h = host( 'localhost' );
        my $out = `df -k /`;
        my @dfrslt = split /\n/, $out;
        my @dfinfo = split / +/, $dfrslt[1];
        $rdev = ( stat( "/dev/zero" ) )[6];
        $rdev = ( stat( $dfinfo[0] ) )[6] if $opts{type} == NFBLK;

        if ( $h->hosttype() =~ /solaris/ ) {
            # Format of solaris rdev result is funky: handle it here
            $minor = $rdev & 0xFFFF;
            $major = ( $rdev>>18 ) & 0xFFFF;
        } else {
            $minor = $rdev & 0xFF;
            $major = ( $rdev>>8 ) & 0xFF;
        }
    }
    $nfs_obj->rpc()->auth_sys_set( uid => 0, gid => 0 );

    foreach my $iter ( 1 .. $opts{count} ) {
        my $dev_name = "$opts{name}\.$iter";
        my $result;
        # NFSv3 create only supports regular files, use mknod instead

        my $op = ($opts{version} == 3) ? 'mknod' : 'create';

        $mode = ($opts{version} == 4) ? FATTR4_MODE() : 'mode';

        $result = $nfs_obj->$op( dhandle => $curr_fh,
                                 name    => $dev_name,
                                 type    => $opts{type},
                                 major   => $major,
                                 minor   => $minor,
                                 attr    => { $mode => 0777 } );
        push @devices, $dev_name;

        $nfs_obj->lookup( dhandle => $curr_fh,
                          name    => $dev_name );

        $result = $nfs_obj->setattr( attr => $opts{attr} ) if $opts{attr};
    }
    $nfs_obj->rpc()->auth_sys_set();

    logDbg('device_make: Umount the nfs_obj');
    $fh_mnt = $nfs_obj->umount( path => $opts{volume} );

    return @devices
}

#------------------------------------------------------------------------

=head2 file_write

  file_write( hostname  => $hostname,
              volume    => $volume,
              [ path      => $path, ]
              data      => $data,
              stable    => $stable,
              stateid   => $stateid
              [, offset => $offset ]
              [, version  => $version]
              [, protocol => $protocol]
              [, minor_ver => $minor_ver]);

Writes data to a file on the hostname specified.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer on which the file to be written
is residing.

=item volume

(Required) : The volume on the filer where file to be written is located..

=item path

(Optional) : The volume relative path of the file to be written.
Otherwise, uses current fh in client_obj

=item data

(Required) : The data to be written to the file.

=item stable

(Required) : A specification of how the NFS v4 write is to be performed.
Can be FILE_SYNC, DATA_SYNC, or UNSTABLE.

=item stateid

(Required) : The NFS v4 stateid of the file to be written, if one is
known. Defaults to 0.

=item offset

(Optional) : The starting byte offset in the file for the data to be
written. Defaults to 0.

=item version

(Optional) : The NFS version to use. Defaults to 3

=item protocol

(Optional) : The protocol to use (tcp or udp). Defaults to tcp.

=item minor_ver

(Optional) : Minor Version for NFSv4. Defaults to 0.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub file_write ( % ) {
    my %opts = validate(
        @_,
        {
            hostname  => { type => SCALAR },
            volume    => { type => SCALAR },
            data      => { type => SCALAR },
            stable    => { type => SCALAR },
            stateid   => { type => SCALAR },
            path      => { type => SCALAR, default => ''},
            offset    => { type => SCALAR, default => 0 },
            version   => { type => SCALAR, default => 3 },
            protocol  => { type => SCALAR, default => 'tcp' },
            minor_ver => { type => SCALAR, default => 0 },
        }
    );
    my $f = host("$main::FILER");

    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = $Data_IP;
    }
    my ($nfs_obj);
    logDbg('BEGIN: file_write');
    logDbg('file_write: Create and mount v3 nfs_obj for file writing');

    $nfs_obj = ( $opts{version} == 4 ) ?
        nfs::NfsClient->new( hostname => $opts{hostname},
                             version  => $opts{version},
                             proto    => $opts{protocol},
                             minor_version  => $opts{minor_ver},
                             timeout  => (defined $main::NFS_CLIENT_TIMEOUT) ? $main::NFS_CLIENT_TIMEOUT : 30,
                             retry => (defined $main::RETRY) ? $main::RETRY : 0
                             )

        : nfs::NfsClient->new( hostname => $opts{hostname},
                               version  => $opts{version},
                               proto    => $opts{protocol},
                              );
    $nfs_obj->mount( path => $opts{volume} );

    if ( $opts{path} ) {
        logDbg('file_write: Get file fh from path');
        nfsops_lookup_path( client => $nfs_obj,
                            path   => $opts{path} );
    }

    logDbg('file_write: Write data to file');
    $nfs_obj->write( offset  => $opts{offset},
                     data    => $opts{data},
                     stable  => $opts{stable},
                     stateid => $opts{stateid} );

    logDbg('END: file_write');
    return;
}

#------------------------------------------------------------------------

=head2 file_getattr_with_client

  file_getattr_with_client( client_obj => $client_obj,
                            [path       => $path] );

Gets attributes from a file using the specified client.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client_obj

(Required) : The client object with which the file attributes should
be retrieved.

=item path

(Optional) : The volume relative path of the file to be accessed.
Otherwise, uses current fh in client_obj

=back

=item Returns

A hash of the file attributes.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub file_getattr_with_client ( % ) {
    my %opts = validate(
        @_,
        {
            client_obj => { type => HASHREF },
            path       => { type => SCALAR, default => '' },
        }
    );
    my $nfs_obj = $opts{client_obj};
    logDbg('BEGIN: file_getattr_with_client');

    if ( $opts{path} ) {
        logDbg('file_getattr: Get file fh from path');
        nfsops_lookup_path( client => $nfs_obj,
                            path   => $opts{path} );
    }

    logDbg('file_getattr: Get file attributes');
    my $result = $nfs_obj->getattr();
    logDbg('END: file_getattr_with_client');
    return $result->{fattr};
}

#------------------------------------------------------------------------

=head2 file_setattr

  file_setattr( hostname    => $hostname,
                volume      => $volume,
                [ path        => $path, ]
                attr        => $attr,
                [, version  => $version]
                [, protocol => $protocol]
                [, minor_ver => $minor_ver]);

Sets attributes on a file on the hostname specified.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer on which the file to be modified
is residing.

=item volume

(Required) : The volume on the filer where file to be modified is located..

=item path

(Optional) : The volume relative path of the file to be modified.
Otherwise, uses current fh in client_obj

=item attr

(Required) : The attributes to set on the file.

=item version

(Optional) : The NFS version to use. Defaults to 3

=item protocol

(Optional) : The protocol to use (tcp or udp). Defaults to tcp.

=item minor_ver

(Optional) : Minor Version for NFSv4. Defaults to 0.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub file_setattr ( % ) {
    my %opts = validate(
        @_,
        {
            hostname => { type => SCALAR },
            volume   => { type => SCALAR },
            path     => { type => SCALAR, default => '' },
            attr     => { type => HASHREF },
            version  => { type => SCALAR, default => 3 },
            protocol => { type => SCALAR, default => 'tcp' },
            minor_ver   => { type => SCALAR, default => 0 },
        }
    );
    my $f = host("$main::FILER");

    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = $Data_IP;
    }

    my ($nfs_obj);
    logDbg('BEGIN: file_setattr');
    logDbg('file_setattr: Create and mount nfs_obj');

    $nfs_obj = ( $opts{version} == 4 ) ?
        nfs::NfsClient->new( hostname => $opts{hostname},
                             version  => $opts{version},
                             proto    => $opts{protocol},
                             minor_version  => $opts{minor_ver},
                             timeout  => (defined $main::NFS_CLIENT_TIMEOUT) ? $main::NFS_CLIENT_TIMEOUT : 30,
                             retry => (defined $main::RETRY) ? $main::RETRY : 0
                            )

        : nfs::NfsClient->new( hostname => $opts{hostname},
                               version  => $opts{version},
                               proto    => $opts{protocol},
                              );

    $nfs_obj->mount( path => $opts{volume} );

    eval {
        file_setattr_with_client( client_obj => $nfs_obj,
                                  path       => $opts{path},
                                  attr       => $opts{attr} );
    };

    my $setattr_failed = $@ if $@;

    $nfs_obj->umount( path => $opts{volume} );
    logDbg('END: file_setattr');

    croak $setattr_failed if $setattr_failed;
    return;
}

#------------------------------------------------------------------------

=head2 file_setattr_with_client

  file_setattr_with_client( client_obj => $client_obj
                            [ path       => $path, ]
                            attr       => $attr);

Sets attributes on a file using the specified client.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client_obj

(Required) : The client object with which the file attributes should
be set.

=item path

(Optional) : The volume relative path of the file to be modified.
Otherwise, uses current fh in client_obj

=item attr

(Required) : The attributes to set on the file.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub file_setattr_with_client ( % ) {
    my %opts = validate(
        @_,
        {
            client_obj => { type => HASHREF },
            path       => { type => SCALAR, default => '' },
            attr       => { type => HASHREF },
        }
    );
    my $nfs_obj = $opts{client_obj};
    logDbg('BEGIN: file_setattr_with_client');

    if ( $opts{path} ) {
        logDbg('file_setattr: Get file fh from path');
        nfsops_lookup_path( client => $nfs_obj,
                            path   => $opts{path} );
    }

    logDbg('file_setattr: Set file attributes');
    $nfs_obj->setattr( attr => $opts{attr} );
    logDbg('END: file_setattr_with_client');
    return;
}
#------------------------------------------------------------------------

=head2 scalar_validate

  $validation = scalar_validate( op          => $op,
                                 version     => $version,
                                 protocol    => $protocol,
                                 expected    => $expected,
                                 returned    => $returned
                                 [, operator => 'eq'|'ne'|'lt'|'gt' ] );

Validates a scalar result from an NFS operation against an expected
result. Compares the expected and returned values passed in using the
operator given and reports an error if the comparison fails. The error
message is formatted with op, version, and protocol information.

The following named arguments are accepted:

=over

=item Arguments

=over

=item op

(Required) : A string describing the operation/results being validated.
Used in error messages to make them more readable.

=item version

(Required) : The NFS version used for the results being validated.
Used in error messages to make them more readable.

=item protocol

(Required) : The NFS communication protocol used for the results being
validated. Used in error messages to make them more readable.

=item expected

(Optional) : The expected value from the NFS operation.

=item returned

(Required) : The returned value from the NFS operation.

=item operator

(Optional) : The comparison operator to be used between the two values.
May be one of 'eq' to validate the two values are equal, 'ne' to validate
that the two values are not equal, 'gt' to validate that the expected
value is greater than the returned value, or 'lt' to validate that the
expected value is less than the returned value. Defaults to 'eq'.

=back

=item Returns

A string indicating successful validation, or an error.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub scalar_validate ( % ) {
    my %opts = validate(
        @_,
        {
            op       => { type => SCALAR },
            version  => { type => SCALAR },
            protocol => { type => SCALAR },
            expected => { type => SCALAR },
            returned => { type => SCALAR },
            expected_min => { type => SCALAR, default => 0 },
            operator => { type => SCALAR, default => 'eq',
                          regex => qr/eq|ne|lt|gt|between/ },
        }
    );
    my $op    = $opts{op};
    my $ver   = $opts{version};
    my $proto = $opts{protocol};

    logDbg('BEGIN: scalar_validate');
    logDbg('scalar_validate: Check that reply exists');

    return "scalar_validate: $op result undefined"
            if ( !defined( $opts{returned} ) );

    if ( $opts{operator} eq 'eq' &&
            !( $opts{expected} eq $opts{returned} ) ) {
        return "scalar_validate: $op results do not match: " .
                "\'$opts{expected}\' expected, \'$opts{returned}\' returned: " .
                "NFS v$ver, $proto";
    }

    if ( $opts{operator} eq 'ne' &&
            !( $opts{expected} ne $opts{returned} ) ) {
        return "scalar_validate: $op results match: " .
                "\'$opts{expected}\' given, \'$opts{returned}\' returned: " .
                "NFS v$ver, $proto";
    }

    if ( $opts{operator} eq 'gt' &&
            !( $opts{expected} gt $opts{returned} ) ) {
        return "scalar_validate: $op results do not validate: " .
                "\'$opts{expected}\' is not greater than \'$opts{returned}\': " .
                "NFS v$ver, $proto";
    }

    if ( $opts{operator} eq 'lt' &&
            !( $opts{expected} lt $opts{returned} ) ) {
        return "scalar_validate: $op results do not match: " .
                "\'$opts{expected}\' is not less than \'$opts{returned}\': " .
                "NFS v$ver, $proto";
    }

    if ( $opts{operator} eq 'between' &&
            !( $opts{expected_min} <=  $opts{returned} ) &&
                 !( $opts{expected} >= $opts{returned} ) ) {
        return "scalar_validate: $op results do not match: " .
                "\'$opts{returned}\' is not in range \'$opts{expected_min}\' and \'$opts{expected}\' " .
                "NFS v$ver, $proto";
    }

    logDbg('END: scalar_validate');
    return 'scalar_validate: validation succeeded';
}

#------------------------------------------------------------------------

=head2 hash_validate

  $validation = hash_validate( op       => $op,
                               version  => $version,
                               protocol => $protocol,
                               expected => \%expected,
                               returned => \%returned
                               [ key    => $key ] );

Validates a hash result from an NFS operation against an expected
result. Compares the expected and returned values passed in and reports
an error if they differ. Only those keys present in the expected hash
are validated, additional keys in the returned hash are ignored.
Nested hashes are compared through recursive calls to hash_validate.
The error message is formatted with op, version, and protocol information.

The following named arguments are accepted:

=over

=item Arguments

=over

=item op

(Required) : A string describing the operation/results being validated.
Used in error messages to make them more readable.

=item version

(Required) : The NFS version used for the results being validated.
Used in error messages to make them more readable.

=item protocol

(Required) : The NFS communication protocol used for the results being
validated. Used in error messages to make them more readable.

=item expected

(Required) : The hash of expected values from the NFS operation.

=item returned

(Required) : The hash of returned values from the NFS operation.

=item key

(Optional) : A string describing the parent key for reporting purposes
when recursive calls are made to hash_validate. Defaults to ''.

=back

=item Returns

A string indicating successful validation, or an error.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub hash_validate ( % ) {
    my %opts = validate(
        @_,
        {
            op       => { type => SCALAR },
            version  => { type => SCALAR },
            protocol => { type => SCALAR },
            expected => { type => HASHREF },
            returned => { type => HASHREF },
            key      => { type => SCALAR, default => '' },
        }
    );
    my $pnt_key;
    my $ver      = $opts{version};
    my $proto    = $opts{protocol};
    my %expected = %{$opts{expected}};
    my %returned = %{$opts{returned}};

    if ( $opts{key} ) {
        $pnt_key = "$opts{key}:";
    } else {
        $pnt_key = '';
    }

    logDbg('BEGIN: hash_validate');
    logDbg('hash_validate: Check that result exists');

    return "hash_validate: $opts{op} result undefined"
            if ( !defined( $opts{returned} ) );

    foreach my $key ( keys %expected ) {

        if ( ref( $expected{$key} ) eq 'HASH' ) {
            my $validation = hash_validate( op       => $opts{op},
                                            version  => $ver,
                                            protocol => $proto,
                                            expected => $expected{$key},
                                            returned => $returned{$key},
                                            key      => "$pnt_key$key" );

            if ( $validation !~ /validation succeeded/ ) {
                return $validation;
            }
        } elsif ( ref( $expected{$key} ) eq 'ARRAY' &&
                  ref( $returned{$key} ) eq 'ARRAY' ) {
            # perform deep array comparison (as hashes)
            my %expected_array = map{ $_ => $expected{$key}->[$_] } 0 .. $#{$expected{$key}};
            my %returned_array = map{ $_ => $returned{$key}->[$_] } 0 .. $#{$returned{$key}};
            my $validation = hash_validate( op       => $opts{op},
                                            version  => $ver,
                                            protocol => $proto,
                                            expected => \%expected_array,
                                            returned => \%returned_array,
                                            key      => "$pnt_key$key" );
            if ( $validation !~ /validation succeeded/ ) {
                return $validation;
            }

        } else {

            if ( $expected{$key} ne $returned{$key} ) {
                return "hash_validate: $pnt_key$key mismatch - " .
                           "\'$expected{$key}\' expected, " .
                           "\'$returned{$key}\' returned: NFS v$ver, $proto";
            }
        }
    }

    logDbg('END: hash_validate');
    return 'hash_validate: validation succeeded';
}

#------------------------------------------------------------------------

=head2 hash_validate_change

  $validation = hash_validate_change( op       => $op,
                                      version  => $version,
                                      protocol => $protocol,
                                      initial  => \%initial,
                                      current  => \%current
                                      [ key    => $key ] );

Validates that current hash values have changed from initial hash values.
Compares the initial and current values passed in and reports an error
if they do not differ. Only those keys present in the initial hash
are validated, additional keys in the current hash are ignored.
Nested hashes are compared through recursive calls to hash_validate_change.
The error message is formatted with op, version, and protocol information.

The following named arguments are accepted:

=over

=item Arguments

=over

=item op

(Required) : A string describing the operation/results being validated.
Used in error messages to make them more readable.

=item version

(Required) : The NFS version used for the results being validated.
Used in error messages to make them more readable.

=item protocol

(Required) : The NFS communication protocol used for the results being
validated. Used in error messages to make them more readable.

=item initial

(Required) : The hash of initial values to be validated.

=item current

(Required) : The hash of current values to be validated.

=item key

(Optional) : A string describing the parent key for reporting purposes
when recursive calls are made to hash_validate_change. Defaults to ''.

=back

=item Returns

A string indicating successful validation, or an error.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub hash_validate_change ( % ) {
    my %opts = validate(
        @_,
        {
            op       => { type => SCALAR },
            version  => { type => SCALAR },
            protocol => { type => SCALAR },
            initial  => { type => HASHREF },
            current  => { type => HASHREF },
            key      => { type => SCALAR, default => '' },
        }
    );
    my $pnt_key;
    my $ver = $opts{version};
    my $proto = $opts{protocol};
    my %initial = %{$opts{initial}};
    my %current = %{$opts{current}};

    if ( $opts{key} ) {
        $pnt_key = "$opts{key}:";
    } else {
        $pnt_key = '';
    }

    logDbg('BEGIN: hash_validate_change');
    logDbg('hash_validate_change: Check that result exists');

    return "hash_validate_change: $opts{op} result undefined"
            if ( !defined( $opts{current} ) );

    foreach my $key ( keys %initial ) {

        if ( ref( $initial{$key} ) eq 'HASH' ) {
            my $validation = hash_validate_change( op       => $opts{op},
                                                   version  => $ver,
                                                   protocol => $proto,
                                                   initial  => $initial{$key},
                                                   current  => $current{$key},
                                                   key      => "$pnt_key$key" );

            if ( $validation !~ /validation succeeded/ ) {
                return $validation;
            }

        } else {

            if ( $initial{$key} eq $current{$key} ) {
                return "hash_validate_change: $pnt_key$key did not change - " .
                        "\'$initial{$key}\' initial, " .
                        "\'$current{$key}\' current: NFS v$ver, $proto";
            }
        }
    }

    logDbg('END: hash_validate_change');
    return 'hash_validate_change: validation succeeded';
}

#------------------------------------------------------------------------

=head2 nfsops_lookup_path

  nfsops_lookup_path( client => $client,
                      path   => $path );

Calls a series of LOOKUP operations on a path (relative to
mountpoint) and sets the current filehandle of the NFS object
to the last element of the path.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client

(Required) : An NFS object that has previously called the mount function.

=item path

(Required) : The path, relative to the mount point, to be looked up.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_lookup_path ( % ) {
    my %opts = validate(
        @_,
        {
            client => { type => OBJECT },
            path   => { type => SCALAR },
        }
    );
    my ($name, @path);
    logDbg('BEGIN: nfsops_lookup_path');
    logDbg('nfsops_lookup_path: Verify mount and reset current filehandle');

    # Verify that client has been mounted
    croak 'NFS client object has no mount filehandle (not mounted)'
            if ( !$opts{client}->mnt_fh() );

    # Set curr_fh to mnt_fh
    $opts{client}->curr_fh($opts{client}->mnt_fh());
    logDbg('nfsops_lookup_path: Split path and call lookups');

    # Get rid of contiguous '/'s and split path into segments
    $opts{path} =~ s{/+}{/}g;
    $opts{path} =~ s{^/}{};
    @path = split /\//, $opts{path};

    foreach $name (@path) {
        eval { $opts{client}->lookup( name => $name ) };
        croak "LOOKUP error on $name in path '$opts{path}'\n$@" if $@;
    }

    logDbg('END: nfsops_lookup_path');
    return;
}

#------------------------------------------------------------------------

=head2 nfsops_lookup_snapshot_path

  nfsops_lookup_snapshot_path(  client => $client,
                                file_obj  => $file,
                                snapshot   => $snapshot_name );

Creates a snapshot file by fetching junction-path and file-name from the file_obj
Then calls a series of LOOKUP operations on a path (relative to
mountpoint) and sets the current filehandle of the NFS object
to the last element of the path.

The following named arguments are accepted:

=over

=item Arguments

=over

=item  client

(Required) : An NFS object that has previously called the mount function.

=item file_obj

(Required) : The file object of the file to be LOOKED up in the snapshot.

=item snapshot

(Required) : The snapshot to be used for looking the file.

=back

=item Returns

The snapshot file path.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_lookup_snapshot_path ( % ) {
    my %opts = validate(
        @_,
        {
            client => { type => OBJECT },
            file_obj => { type => HASHREF },
            snapshot   => { type => SCALAR },
        }
    );
    my ($name, @path, $snapshot_file);

    logDbg('BEGIN: nfsops_lookup_snapshot_path');

    logDbg('nfsops_lookup_snapshot_path: Verify mount and reset current filehandle');

    # Verify that client has been mounted
    croak 'NFS client object has no mount filehandle (not mounted)'
            if ( !$opts{client}->mnt_fh() );

    # Set curr_fh to mnt_fh
    $opts{client}->curr_fh($opts{client}->mnt_fh());

    logDbg('nfsops_lookup_snapshot_path: Split path and call lookups');

    # Get rid of contiguous '/'s i.e path that contains at least 2 continuous
    # forward slash ('//') and split path into segments
    $snapshot_file = $opts{file_obj}->{ns_file}->{'junction-path'} .
                      "/.snapshot/$opts{snapshot}/" .
                      $opts{file_obj}->{ns_file}->{path};

    my $relative_path = ".snapshot/$opts{snapshot}/" .
                      $opts{file_obj}->{ns_file}->{path};

    nfsops_lookup_path( client => $opts{client},
                        path   => $relative_path );

    logDbg('END: nfsops_lookup_snapshot_path');
    return $snapshot_file;
}

#------------------------------------------------------------------------

=head2 nfsops_lookup_dongle_file

  nfsops_lookup_dongle_file(  client => $client,
                              path   => $dongle_file_path,)

The following named arguments are accepted:

=over

=item Arguments

=over

=item  client

(Required) : A NFS client object pointing to the DATA IP

=item path

(Required) : The file path whose dongle file is to be lookedup
EG: If the dongle file to be looked up is /foo/bar/.~DOT_CR1992fred
    then just provide the parent file path /foo/bar/fred w/o dongle name

=back

=item should_be_present

(Optional) : Specify this option to true (non-zero), if the test expects
the dongle file to be present. Default is 0 (False) i.e it doesn't expects the
dongle file to be present and will return 0 if the dongle file is present

=item Returns

True (1) if the specified dongle file was not found when expected not to be
found or it was found when expected to be found else False (0) if the file is
present when not expected to be found or if the file was not found when expected
to be found

=back

=for html <hr />

=cut

#------------------------------------------------------------------------

sub nfsops_lookup_dongle_file {

    # Taking the input parameters
    my %opts = validate(
        @_,
        {   client => { type => OBJECT },
            path   => { type => SCALAR },
            should_be_present => { type => BOOLEAN,
                                   default => 0 },
        }
    );

    # Getting the NFS version
    my $client_obj = $opts{client};
    my $nfs_ver    = $client_obj->version();

    $Log->comment("Processing the path - $opts{path} for dongle file");
    # Constructing the dongle file path

    my $dir_path    = dirname( $opts{path} );
    $Log->comment("Directory path of dongle file is: $dir_path");

    my $file_name   = basename( $opts{path} );
    $Log->comment("File name is: $file_name");

    # This is because if the file is directly present under the junction-path,
    # dirname would return '.' & dongle file path becomes ./.~DOT_CR1992file
    # As a result of which nfsops_lookup_path fails but only for NFSv4 and with
    # NFS4ERR_BADNAME. Not sure if it is expected behavior.
    # I see no reason why NFSv4 lookup of '.' should fail with NFS4ERR_BADNAME
    my $dongle_file
        = ($dir_path eq q{.})
        ? ($dongle_file_prefix . $file_name)
        : join( q{/}, $dir_path, $dongle_file_prefix . $file_name );
    $Log->comment("Dongle file path is: $dongle_file");

    # Generating the expected error codes
    my $test_result = 1;
    my $exp_err;
    if ( $nfs_ver == 3 ) {
        $exp_err = 'NFS3ERR_NOENT';
    } elsif ( $nfs_ver == 4.0 ) {
        $exp_err = 'NFS4ERR_NOENT';
    } elsif ( $nfs_ver == 4.1 ) {
        $exp_err = 'NFS4ERR_NOENT';
    }

    # Trying to lookup the dongle file.
    try {
        nfsops_lookup_path( client => $client_obj, path => $dongle_file );

        # If the dongle file is expected to be found then set the return value
        # to zero if it is not found
        if($opts{should_be_present}){
            $Log->comment( "Dongle file $dongle_file found expectedly"
                    . " for NFSv$nfs_ver" );
        } else {
            $Log->error( "FAIL: Dongle file $dongle_file found unexpectedly"
                    . " for NFSv$nfs_ver" );
            $test_result = 0;
        }
    }
    catch NATE::BaseException with {
        my $lookup_err = shift;

        # If the dongle file is not expected to be found and NFS vserver didn't
        # return expected error code then set the return value to zero
        if(!$opts{should_be_present}){
            if ( $lookup_err !~ /$exp_err/ ) {
                $Log->error( "NFSv$nfs_ver Lookup of $dongle_file failed with"
                        . " an unexpected error. Expected: $exp_err."
                        . " Actual: $lookup_err" );
                $test_result = 0;
            } else {
                $Log->comment("NFSv$nfs_ver returned expected error code: "
                              . $exp_err);
            }
        } else {
            # If the dongle file is expected to be found then set the return
            # value to zero if it is not found
            $Log->error( "FAIL: Dongle file $dongle_file not found unexpectedly"
                    . " for NFSv$nfs_ver. Reason: $lookup_err" );
            $test_result = 0;
        }
    };

    return $test_result;
} ## end sub nfsops_lookup_dongle_file

#------------------------------------------------------------------------

=head2 nfsops_common_attrs

  %attrs = nfs_common_attrs( version  => $version,
                             returned => \%returned );

Converts a hash of v2, v3, v4 attributes to a common format for comparison
between different NFS versions. Certain hash keys are removed, and NFS v4
operation codes are converted to hash keys matching those of v2 and v3.

The following named arguments are accepted:

=over

=item Arguments

=over

=item version

(Required) : The NFS version which generated the attribute keys and values
in the current format.

=item returned

(Required) : A hash of attribute keys matching the format returned by
the version indicated.

=back

=item Returns

A hash containing the attributes with common format key names.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_common_attrs ( % ) {
    my %opts = validate(
        @_,
        {
            version  => { type => SCALAR },
            returned => { type => HASHREF },
        }
    );
    my %common = %{$opts{returned}};
    logDbg('BEGIN: nfsops_common_attrs');

    if ( $opts{version} == 2 ) {
        $common{acc_seconds}   = $common{atime}{seconds};
        $common{acc_usec_nsec} = $common{atime}{useconds};
        delete $common{atime};
        $common{mod_seconds}   = $common{mtime}{seconds};
        $common{mod_usec_nsec} = $common{mtime}{useconds};
        delete $common{mtime};
        my $rdev = $common{rdev};
        delete $common{rdev};
        $common{rdev}{specdata2} = $rdev & 0xFF;
        $common{rdev}{specdata1} = ($rdev>>8) & 0xFF;

    } elsif ( $opts{version} == 3 ) {
        $common{acc_seconds}   = $common{atime}{seconds};
        $common{acc_usec_nsec} = $common{atime}{nseconds};
        delete $common{atime};
        $common{mod_seconds}   = $common{mtime}{seconds};
        $common{mod_usec_nsec} = $common{mtime}{nseconds};
        delete $common{mtime};

    } else {
        $common{nlink}  = $common{FATTR4_NUMLINKS()};
        delete $common{FATTR4_NUMLINKS()};
        $common{fsid}   = $common{FATTR4_FSID()}{major};
        delete $common{FATTR4_FSID()};
        $common{fileid} = $common{FATTR4_FILEID()};
        delete $common{FATTR4_FILEID()};
        $common{size}   = $common{FATTR4_SIZE()};
        delete $common{FATTR4_SIZE()};
        $common{ctime}  = $common{FATTR4_TIME_METADATA()};
        delete $common{FATTR4_TIME_METADATA()};
        $common{mode}   = $common{FATTR4_MODE()};
        delete $common{FATTR4_MODE()};
        $common{type}   = $common{FATTR4_TYPE()};
        delete $common{FATTR4_TYPE()};
        $common{acc_seconds}   = $common{FATTR4_TIME_ACCESS()}{seconds};
        $common{acc_usec_nsec} = $common{FATTR4_TIME_ACCESS()}{nseconds};
        delete $common{FATTR4_TIME_ACCESS()};
        $common{mod_seconds}   = $common{FATTR4_TIME_MODIFY()}{seconds};
        $common{mod_usec_nsec} = $common{FATTR4_TIME_MODIFY()}{nseconds};
        delete $common{FATTR4_TIME_MODIFY()};
        $common{rdev} = $common{FATTR4_RAWDEV()};
        delete $common{FATTR4_RAWDEV()};
        $common{uid} = ( split '@', $common{FATTR4_OWNER()} )[0];
        $common{uid} = getpwnam( $common{uid} );
        delete $common{FATTR4_OWNER()};
        $common{gid} = ( split '@', $common{FATTR4_OWNER_GROUP()} )[0];
        $common{gid} = getgrnam( $common{gid} );
        delete $common{FATTR4_OWNER_GROUP()};
    }

    logDbg('END: nfsops_common_attrs');
    return %common;
}

#------------------------------------------------------------------------

=head2 nfsops_save_common_attrs

  %saved_attrs = nfsops_save_common_attrs( attr_h => \%attr_h );

Generates a hash consisting of a subset of attributes that are common
to all three NFS versions. The subset does not keep attributes that
are not common to all three versions or that can be determined by
other means from the filer. The main purpose of this subroutine is to
select and keep a set of attributes for comparison between several
NFS versions and calls.

The following named arguments are accepted:

=over

=item Arguments

=over

=item attr_h

(Required) : A hash of attributes in nfsops_common_attrs format.

=back

=item Returns

The hash of common attribute keys and values.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_save_common_attrs ( % ) {
    my %opts = validate(
        @_,
        {
            attr_h => { type => HASHREF },
        }
    );
    my %previous;
    logDbg('BEGIN: nfsops_save_common_attrs');

    $previous{nlink}  = $opts{attr_h}{nlink};
    $previous{fsid}   = $opts{attr_h}{fsid};
    $previous{fileid} = $opts{attr_h}{fileid};
    $previous{size}   = $opts{attr_h}{size};
    $previous{ctime}  = $opts{attr_h}{ctime};
    delete $previous{ctime}{nseconds};
    delete $previous{ctime}{useconds};

    logDbg('END: nfsops_save_common_attrs');
    return %previous;
}

#------------------------------------------------------------------------

=head2 nfsops_access_bits

  $access_bits = nfsops_access_bits( bitmask => $bitmask );

Generates a string of access permissions that are returned as a
bitmask from the NFS v3 and v4 ACCESS operation. Indicates which
access bits are set in the given bitmask.

The following named arguments are accepted:

=over

=item Arguments

=over

=item bitmask

(Required) : A bitmask of access rights as returned from an NFS v3 or
v4 ACCESS operation.

=back

=item Returns

A string indicating the set access types, or 'none' if the bitmask is 0.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_access_bits ( % ) {
    my %opts = validate(
        @_,
        {
            bitmask => { type => SCALAR },
        }
    );
    my $bitmask = $opts{bitmask};
    my $acc_bits;
    logDbg('BEGIN: nfsops_access_bits');

    my %access = ( ACCESS_READ    => ACCESS_READ,
                   ACCESS_LOOKUP  => ACCESS_LOOKUP,
                   ACCESS_MODIFY  => ACCESS_MODIFY,
                   ACCESS_EXTEND  => ACCESS_EXTEND,
                   ACCESS_DELETE  => ACCESS_DELETE,
                   ACCESS_EXECUTE => ACCESS_EXECUTE );

    foreach my $key ( sort { $access{$b} <=> $access{$a} } keys %access ) {

        if ( $bitmask >= $access{$key} ) {
            $bitmask  -= $access{$key};
            $acc_bits .= "$key | ";
        }
    }

    logDbg('END: nfsops_access_bits');

    if ( $acc_bits ) {
        return substr( $acc_bits, 0, -3 );
    } else {
        return 'none';
    }
}

#------------------------------------------------------------------------

=head2 nfsops_fsstats

  %fsstats = nfsops_fsstats( fc     => $fc,
                             volume => $volume );

Gathers fsstat information directly from the filer using the df command.
Parses the output for space information (total bytes, free bytes, available
bytes) and file information (total inodes, free inodes, available
inodes).

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : The filer connection object.

=item volume

(Required) : The volume on the filer where the nfsops tests are performed.

=back

=item Returns

A hash of the fsstat information.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_fsstats ( % ) {
    my %opts = validate(
        @_,
        {
            fc     => { type => OBJECT },
            volume => { type => SCALAR },
        }
    );
    my ( $cmd, $results, @dfrslt, @volrslt, $ufiles, %fsstat );
    logDbg('BEGIN: nfsops_fsstats');

    my $h=host($main::FILER);

    if ($h->hosttype() =~ /filer-ng/){
        my @results_out;

        eval{
            my $server_obj = new Server(node=>$main::FILER);
            $opts{volume}=~s/\///;
            @results_out = $server_obj -> volume_show (vserver=>$main::VSERVER_GX_NAME, volume=>$opts{volume});
        };
        $fsstat{fbytes}=$results_out[0]->{'available'};
        $fsstat{abytes}=$results_out[0]->{'available'};
        $fsstat{tbytes}=$results_out[0]->{'total'};
        $fsstat{tfiles}=$results_out[0]->{'files'};
        my $ufiles=$results_out[0]->{'files-used'};
        $fsstat{afiles}=$fsstat{tfiles}-$ufiles;
        $fsstat{ffiles}=$fsstat{tfiles}-$ufiles;
        return %fsstat;
    }

    logDbg('nfsops_fsstats: Call df on volume');
    $cmd = "df $opts{volume}";
    $results = sayEOF( $opts{fc}, $cmd );

    # Get space information
    @dfrslt  = split /\n/, $results;
    @volrslt = split / +/, $dfrslt[1];
    $fsstat{tbytes} = int( $volrslt[1] ) * 1024;
    $fsstat{fbytes} = int( $volrslt[3] ) * 1024;
    $fsstat{abytes} = $fsstat{fbytes};

    logDbg('nfsops_fsstats: Call df -i on volume');
    $cmd = "df -i $opts{volume}";
    $results = sayEOF( $opts{fc}, $cmd );

    # Get inode information
    @dfrslt  = split /\n/, $results;
    @volrslt = split / +/, $dfrslt[1];
    $ufiles  = int( $volrslt[1] );
    $fsstat{ffiles} = int( $volrslt[2] );
    $fsstat{afiles} = $fsstat{ffiles};
    $fsstat{tfiles} = $fsstat{ffiles} + $ufiles;

    logDbg('END: nfsops_fsstats');
    return %fsstat;
}

#------------------------------------------------------------------------

=head2 nfsops_set_devices

  %devs = nfsops_set_devices( device => 'char'|'block'|'both' );

Creates a hash of device types and ids based on the device value passed in.
The hash indicates a character device type, a block device type, or both.
The returned hash is useful in tests involving devices for stepping through
the types of devices to be created and tested.

The following named arguments are accepted:

=over

=item Arguments

=over

=item device

(Required) : The type(s) of device(s) to be included. May be 'char',
'block', or 'both'.

=back

=item Returns

A hash, in which the keys are the type of device (block and/or
char) and the values are the NFS numeric value of that type.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_set_devices ( % ) {
    my %opts = validate(
        @_,
        {
            device => { type => SCALAR, regex => qr/char|block|both/ },
        }
    );
    my ( $dev, %devices );
    $dev = $opts{device};

    if ( $dev eq 'char' ) {
        %devices = ( char  => NFCHR )
    } elsif ( $dev eq 'block' ) {
        %devices = ( block => NFBLK )
    } elsif ( $dev eq 'both' ) {
        %devices = ( char  => NFCHR, block => NFBLK )
    } else {
        croak "Invalid device specification: $dev";
    }

    return %devices;
}

#------------------------------------------------------------------------

=head2 nfsops_get_new_id

  @ids = nfsops_get_new_id( fc   => $fc,
                            type => 'uid'|'gid' );

Selects a new uid or gid from 'testu1', 'testu2', 'testg1', or 'testg2'.
Type parameter determines if a uid or gid is selected. Tries to assign a new
uid or gid that is different from the current user and that exists on the
filer. If both uids or gids are unusable, an error is returned.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : A filer connection object to the filer to be used.

=item type

(Required) : A string that identifies the type of id to be selected. Must
be either 'uid' or 'gid'.

=back

=item Returns

An array containing the user or group name and the user or group
id number, or an error if a new id cannot be assigned.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_get_new_id ( % ) {
    my %opts = validate(
        @_,
        {
            fc   => { type => HASHREF },
            type => { type => SCALAR, regex => qr/uid|gid/ },
        }
    );

    if ( $opts{type} eq 'uid' ) {
        # Assign 'testu1' if it exists and is not current user
        my $h = host($main::FILER);
        my $testu1;

        if ($h->hosttype()=~/filer-ng/){
            $testu1=run_getXXbyYY_gx();
        } else{
            $testu1 = run_getXXbyYY( $opts{fc}, 'getpwbyname_r', 'testu1' );
        }

        ($testu1)  = ( $testu1 =~ /pw_uid = (\d+)/ );
        return ( 'testu1', $testu1 ) if ( $testu1 && $testu1 != $> );

        # Assign 'testu2' if it exists and is not current user
        my $testu2;

        if ($h->hosttype()=~/filer-ng/){
            $testu2=run_getXXbyYY_gx();
        } else{
            $testu2 = run_getXXbyYY( $opts{fc}, 'getpwbyname_r', 'testu2' );
        }

        ($testu2)  = ( $testu2 =~ /pw_uid = (\d+)/ );
        return ( 'testu2', $testu2 ) if ( $testu2 && $testu2 != $> );

        # Croak because a new uid could not be assigned
        croak 'Could not assign testu1 or testu2: does not exist on ' .
                'filer or is the current user';

    } else {
        # Assign 'testg1' if it exists and is not current user's primary group
        my $h = host($main::FILER);
        my $testg1;

        if ($h->hosttype()=~/filer-ng/){
            $testg1=run_getXXbyYY_gx();
        } else{
            $testg1 = run_getXXbyYY( $opts{fc}, 'getgrbyname', 'testg1' );
        }

        ($testg1)  = ( $testg1 =~ /gid = (\d+)/ );

        return ( 'testg1', $testg1 ) if ( $testg1 != (split " ", $))[0] );

        # Assign 'testg2' if it exists and is not current user's primary group
        my $testg2;

        if ($h->hosttype()=~/filer-ng/){
            $testg2=run_getXXbyYY_gx();
        } else{
            $testg2 = run_getXXbyYY( $opts{fc}, 'getgrbyname', 'testg2' );
        }

        ($testg2)  = ( $testg2 =~ /gid = (\d+)/ );
        return ( 'testg2', $testg2 ) if ( $testg2 != (split " ", $))[0] );

        # Croak because a new gid could not be assigned
        croak "Could not assign testg1 or testg2: does not exist on " .
                "filer or is the current user's group";
    }
}

sub testme () {
    logcomment("Performing NfsOps self test");

    require nfs::CommonParams;

    my ( %set_h, %hash1, %hash2 );
    my $ver = 3;
    my $proto = 'tcp';

    # Test setup routine
    eval {
        set_env( setup_h  => \%set_h,
                 allowed  => [ 2, 3, 4 ],
                 version  => $ver,
                 protocol => $proto );
        };
    logresult('FAIL', "set_env failed: $@") if $@;

    # Test cleanup routine
    eval {
        clean_env( %set_h );
    };
    logresult('FAIL', "clean_env failed: $@") if $@;

    # Test scalar validation for equality
    my $out = scalar_validate(
                          op       => 'TESTME: scalar validation, equality',
                          version  => $ver,
                          protocol => $proto,
                          expected => $proto,
                          operator => 'eq',
                          returned => $proto );

    logresult('FAIL', "scalar_validate equality failed: $out")
            if $out !~ /validation succeeded/;

    # Test scalar validation for inequality
    $out = scalar_validate(
                          op       => 'TESTME: scalar validation, inequality',
                          version  => $ver,
                          protocol => $proto,
                          expected => $proto,
                          operator => 'ne',
                          returned => $ver );

    logresult('FAIL', "scalar_validate inequality failed: $out")
            if $out !~ /validation succeeded/;

    # Test hash validation
    for my $index (1..10) {
        $hash1{$index}      = $index;
        $hash1{sub}{$index} = $index;
        $hash2{$index}      = $index;
        $hash2{sub}{$index} = $index;
    }
    $out = hash_validate( op       => 'TESTME: hash validation',
                          version  => $ver,
                          protocol => $proto,
                          expected => \%hash1,
                          returned => \%hash2 );
    logresult('FAIL', "hash_validate failed: $out")
            if $out !~ /validation succeeded/;

    # Test hash change validation
    for my $index (1..10) {
        $hash1{$index} = $index;
        $hash1{sub}{$index} = $index;
        $hash2{$index} = $index + 1;
        $hash2{sub}{$index} = $index + 1;
    }
    $out = hash_validate_change(
                          op       => 'TESTME: hash change validation',
                          version  => $ver,
                          protocol => $proto,
                          initial  => \%hash1,
                          current  => \%hash2 );

    logresult('FAIL', "hash_validate_change failed: $out")
            if $out !~ /validation succeeded/;

    logresult('PASS', "SELF TEST of NfsOps.pm");
}

sub run_getXXbyYY_gx () {
    my $server_obj=Server->new(node => $main::FILER);
    my $output;

    eval {
        $server_obj->vserver_services_unix_user_show(
                        vserver => $::VSERVER_GX_NAME,
                        user          => "testu1");
    };

    if ($@) {

        eval {
            $server_obj->vserver_services_unix_user_create(
                          vserver => $::VSERVER_GX_NAME,
                          user    => "testu1",
                          id      => "3196",
                          'primary-gid' => "3196",
                         );
       };

       if ($@) {
        logWarn("$@");
       } else {
        $output='pw_uid = 3196';
       }

    } else {
        $output='pw_uid = 3196';
    }

    eval {
          $server_obj->vserver_services_unix_group_show(
                           vserver => $::VSERVER_GX_NAME,
                           user          => "testg1");
    };

    if ($@) {

       eval {
            $server_obj->vserver_services_unix_group_create(
                           vserver => $::VSERVER_GX_NAME,
                           name => "testg1",
                           id      => "3196"
                         );
       };
       if ($@) {
               logWarn("$@");
       } else {
            $output.=' gid = 3196';
       }

    } else {
         $output.=' gid = 3196';
    }

    return $output;

}

sub get_gx_v4_id_domain () {

    my $server_obj = new Server (node=>$main::FILER);
    my @results_out=$server_obj -> vserver_nfs_show(vserver=>$main::VSERVER_GX_NAME);

    if ($results_out[0]->{v4}){
        return $results_out[0]->{'v4-id-domain'};
    } else {
       return 0;
    }

}


=head2 file_read_verify_with_client

  file_read_verify_with_client( client_obj => $client_obj,
                               [, path     => $path ]
                               [, fill     => 'A' ]);

reads and verifies the contents of the file using the specified client.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client_obj

(Required) : The client object with which the file contents should
be verified.

=item path

(Optional) : The volume relative path of the file to be accessed.
Otherwise, uses current fh in client_obj

=item fill

(Optional) : A single character expected to fill the file(s).
A GETATTR call will return the size of the file, and the character will
be read repetitively to fill the file to the size specified in the GETATTR.

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub file_read_verify_with_client( % ) {
    my %opts = validate(
        @_,
        {
            client_obj => { type => HASHREF },
            path       => { type => SCALAR,   default => '' },
            fill       => { type => SCALAR,   regex    => qr/^.$/ }
        }
    );
    my $client_obj = $opts{client_obj};

    logDbg('BEGIN: file_read_verify_with_client');

    if ( $opts{path} ) {
        logDbg('file_read_verify_with_client: Get file fh from path');

        nfsops_lookup_path( client => $client_obj,
                            path   => $opts{path} );
    }

    my $getattr = $client_obj->getattr();
    my $size = $opts{client_obj}->version() == 4 ? $getattr->{fattr}{FATTR4_SIZE()} : $getattr->{fattr}{size};
    logDbg("read file size=>$size");
    my $offset  = 0;

    while ( $offset < $size ) {
        my $read_rsp = $client_obj->read( offset  => $offset,
                                          count   => $size - $offset,
                                          stateid => STATEID0 );
        my $rsp_count = $client_obj->version() == 4
                        ? length($read_rsp->{data})
                        : $read_rsp->{count};
        $offset += $rsp_count;
        my $validation = scalar_validate( op       => 'file_read_verify_with_client at offset ' . $offset,
                                          version  => $client_obj->version(),
                                          protocol => $client_obj->protocol(),#TODO, better way?
                                          expected => $opts{fill} x $rsp_count,
                                          returned => $read_rsp->{data});

        if ($validation !~ /validation succeeded/) {
            die fail_result( $validation );
        }
    }
    logDbg('END: file_read_verify_with_client');
}

#------------------------------------------------------------------------

=head2 file_write_verify_with_client

  file_write_verify_with_client( client_obj => $client_obj,
                                [, path     => $path ],
                                [, fill     => 'A' ]);

writes, reads, and verifies the contents of the file using the specified client.

The following named arguments are accepted:

=over

=item Arguments

=over


=item client_obj

(Required) : The client object with which the file contents should
be writen and verified.

=item path

(Optional) : The volume relative path of the file to be accessed.
Otherwise, uses current fh in client_obj

=item fill

(Optional) : A single character to be used to fill the written file(s).
A GETATTR call will return the size of the file, and the character will
be written repetitively to fill the file to the size specified in the GETATTR.

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub file_write_verify_with_client( % ) {
    my %opts = validate(
        @_,
        {
            client_obj => { type => HASHREF },
            path       => { type => SCALAR,   default => '' },
            fill       => { type => SCALAR,   regex    => qr/^.$/ }
        }
    );
    my $client_obj = $opts{client_obj};

    logDbg('BEGIN: file_write_verify_with_client');

    if ( $opts{path} ) {
        logDbg('file_write_verify_with_client: Get file fh from path');
        nfsops_lookup_path( client => $client_obj,
                            path   => $opts{path} );
    }

    my $getattr = $client_obj->getattr();

    my $size = $client_obj->version() == 4 ? $getattr->{fattr}{FATTR4_SIZE()} : $getattr->{fattr}{size};
    my $offset  = 0;

    while ( $offset < $size ) {
        #my $curr_size = 8192 < $size - $offset ? $size - $offset : 8192;
        my $curr_size = $size - $offset;
        my $write_rsp = $client_obj->write( offset  => $offset,
                                            data    => $opts{fill} x $curr_size,
                                            stateid => STATEID0 );
        $offset += $write_rsp->{count};
    }

    file_read_verify_with_client(\%opts);

    logDbg('END: file_write_verify_with_client');
}

#------------------------------------------------------------------------

=head2 file_delete_with_client

  file_delete_with_client( client_obj => $client_obj,
                           file_names     => \@file_names,
                           [, path     => $path ]);

  Delete an array of files using the specified client.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client_obj

(Required) : The client object used to delete the files

=item file_names

(Required) : An array reference to the names of files that needs
to be deleted

=item path

(Optional) : The volume relative path of the file to be accessed.
Default is '/'

=item Returns 1 when succeeds

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub file_delete_with_client {
    my %params = validate(@_, {
        client_obj => {type => OBJECT, optional => 0},
        path => {type => SCALAR, optional => 1, default => q{/}},
        file_names => {type => ARRAYREF, optional => 0}});

    my @file_names = @{$params{file_names}};
    foreach my $file_name (@file_names) {
        object_remove_with_client(
            client_obj => $params{client_obj},
            path => $params{path} . $file_name
        );
        $Log->comment("Deleting file $file_name\n");
    }
    return 1;
}

1;


#/*
# * Local Variables:
# * indent-tabs-mode:nil
# * tab-width:4
# * End:
# */
