#
## @summary     filtering NFS packets in a .trc file
## @author      karthik7@netapp.com
## @pod here

=head1 NAME

Tshark - nfs v2 v3 v4 filtering NFS packets in a .trc file

=head1 SYNOPSIS

 use CR::nfs::Tshark;


=head1 DESCRIPTION

Filters NFS packets in a .trc file

=head1 STATUS

  Shared

=head1 AUTHOR

  karthik7 (CR QA Automation)

=head1 COPYRIGHT

  Copyright 2010, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package CR::nfs::Tshark;

use strict;
use warnings;

use Params::Validate qw(:all);
use NATE::BaseException;
use Tharn qw($Log);
use Data::Dumper;
use Options;
use CR::CommonParams;

use TracePkg;
use Params::NFS;
use Options;

#------------------------------------------------------------------------
#use constant DEFAULT_DELIM => "X;X";
#------------------------------------------------------------------------
# Version for package
#------------------------------------------------------------------------
our $VERSION = '0.01';

#------------------------------------------------------------------------
# Package global variables
use vars qw($Options $class %args $self %Ops_codes %Readdirplus @Packets @Layers);

=head2 ::new

Creates a CR::nfs::Tshark object and returns it.

=over

=item Arguments

=item C<<"file"=>$server_side_trace_file >>

(Mandatory) A trc file

=item C<<"protocol"=>nfs >>

(Optional) Defaults to nfs

=item C<<"version"=>v3 >>

(Optional) Defaults to v3

=item C<<"procedure" => READIRPLUS >>

(Mandatory) NFS operation

=item Examples

=back

    my $readdirplus_obj = CR::nfs::Tshark->new(
                    file => $server_side_trace_file,
                    protocol => 'nfs',
                    version => 'v3', #default
                    procedure => 'READIRPLUS',
                   );
=cut
sub new {
    ($class, %args) = @_;
    $Options = new Options;
    $Options->process();
    $self = {};
    %Ops_codes = ('nfs' => {
                        'v3'  => {
                             'NULL' => 0,
                             'GETATTR' => 1,
                             'SETATTR' => 2,
                             'LOOKUP' => 3,
                             'ACCESS' => 4,
                             'READLINK' => 5,
                             'READ' => 6,
                             'WRITE' => 7,
                             'CREATE' => 8,
                             'MKDIR' => 9,
                             'SYMLINK' => 10,
                             'MKNOD' => 11,
                             'REMOVE' => 12,
                             'RMDIR' => 13,
                             'RENAME' => 14,
                             'LINK' => 15,
                             'READDIR' => 16,
                             'READIRPLUS' => 17,
                             'FSSTAT' => 18,
                             'FSINFO' => 19,
                             'PATHCONF' => 20,
                             'COMMIT' => 21,
                             },
                      },
                    );
    bless $self, $class;
    $self->{_tshark} = '/usr/software/bin/tshark';
    $self->{_file} = $args{file};
    $self->{_protocol} = $args{protocol} || 'nfs';
    $self->{_version} = $args{version} || 'v3';
    $self->{_procedure} = $args{procedure};
    my $protocol_value = $Ops_codes{$self->{_protocol}}->
                                          {$self->{_version}}->
                                               {$self->{_procedure}};
    #filter example : nfs.procedure_v3 == 17
    my $filter = '\'' . $self->{_protocol}.
                  '.procedure_'.$self->{_version}.
                  ' == ' .
                  $protocol_value.
                  '\'';
    my $non_verbose_cmd = $self->{_tshark} .' -r '. $self->{_file} . ' -R '. $filter ;
    $Log->comment("command : $non_verbose_cmd");
    my $non_verbose_output = `$non_verbose_cmd 2>/dev/null`;
    my $verbose_cmd = $non_verbose_cmd . ' -V ';
    my $verbose_output = `$verbose_cmd 2>/dev/null`;
    $self->{_tshark_check_help_mesage} = `$self->{_tshark} --help`;
    if ($!){
        $Log->warn("Error while running tshark $!");
    }
    $Log->comment("Executed Command to parse the tshark file : $verbose_cmd");
    $self->{_raw_output} = $non_verbose_output;
    $self->{_raw_output_verbose} = $verbose_output;
    @Packets = split (/\n\nFrame /,$verbose_output);
    $self->{_packets} = \@Packets;
    @Layers = _extract('Network','0',@Packets);
    $self->{_layers} = \@Layers;
    return $self;
}


=head2 ::get_readirplus_hash

Get a redirplus hash contanning the file stat info

=over

=item Arguments

=item Examples

=back
    my $readdirplus_obj = CR::nfs::Tshark->get_readirplus_hash();
=cut

sub get_readirplus_hash{
    my ($self) = @_;
    my $layers = $self->{_layers};
    my @layers = @$layers;
    my @entries = _extract('Entry','1',@layers);
    foreach my $entry_name (@entries){
        $entry_name =~ /Entry: name (.*)/;
        _construct_readirplus_hash($entry_name);
    }
    return %Readdirplus;
}

=head2 ::_extract

private subroutine to parse tshark packets
=over

=cut

sub _extract{
    my ($filter_condition,$intendation_level,@elements) = @_;
    my @sub_elements;
    my @sub_element_lines;
    my $space_count = ($intendation_level * 4);
    foreach my $element (@elements){
        my @element_lines = split (/\n/,$element);
        my $total_lines_in_element = @element_lines;
        my $line_count=1;#flag
        foreach my $sub_element_line(@element_lines){
            $sub_element_line =~ s/^\s{$space_count}// if ($space_count);
            #Strip the bracket but preserve the indentation
            $sub_element_line =~ s/^(\s*)\[(.*)\]$/$1$2/g;
            if(($sub_element_line =~ /^\S/) || ($line_count == $total_lines_in_element)){
                my $sub_element = join ("\n", @sub_element_lines);
                push (@sub_elements, $sub_element);
                @sub_element_lines=(); # empty this temp varibale for next iteration
            }
                push (@sub_element_lines, $sub_element_line);
                $line_count++;
        }
    }
    #Extract
    my @extracted_sub_elements = grep(/^$filter_condition/, @sub_elements);
    return wantarray ? @extracted_sub_elements : 1;
}

=head2 ::_construct_readirplus_hash

private subroutine to construct readir hash
=over

=cut
sub _construct_readirplus_hash{
 my ($element) = @_;
 my @element_lines = split (/\n/,$element);
 my ($entry_as_key, $flag_atime, $flag_mtime, $flag_ctime);
 foreach my $element_line(@element_lines){
    if($element_line =~ /^Entry: name (.*)/){
        $entry_as_key = $1;
        next if (($entry_as_key eq '.') || ($entry_as_key eq '..'));
    }elsif($element_line =~ /\sattributes (.*) mode:(.*) uid:/){
        $Readdirplus{$entry_as_key}{file_type}= $1;
        $Readdirplus{$entry_as_key}{mode}= $2;
    }elsif($element_line =~ /mode: (.*)/){
        $Readdirplus{$entry_as_key}{mode}= $1;
    }elsif($element_line =~ /nlink: (.*)/){
        $Readdirplus{$entry_as_key}{nlink}= $1;
    }elsif($element_line =~ /size: (.*)/){
        $Readdirplus{$entry_as_key}{size}= $1;
    }elsif($element_line =~ /atime: (.*)/){
        $Readdirplus{$entry_as_key}{atime}= $1;
        $flag_atime = 1;
    }elsif($element_line =~ /mtime: (.*)/){
        $Readdirplus{$entry_as_key}{mtime}= $1;
        $flag_mtime = 1;
    }elsif($element_line =~ /ctime: (.*)/){
        $Readdirplus{$entry_as_key}{ctime}= $1;
        $flag_ctime = 1;
    }else{
        if ($flag_atime && ($element_line =~ /\s{4}seconds: (.*)/)){
            $Readdirplus{$entry_as_key}{atime_seconds}= $1;
        }elsif($flag_atime && ($element_line =~ /\snano seconds: (.*)/)){
            $Readdirplus{$entry_as_key}{atime_nano_seconds}= $1;
            $flag_atime = 0;
        }elsif ($flag_mtime && ($element_line =~ /\s{4}seconds: (.*)/)){
            $Readdirplus{$entry_as_key}{mtime_seconds}= $1;
        }elsif($flag_mtime && ($element_line =~ /\snano seconds: (.*)/)){
            $Readdirplus{$entry_as_key}{mtime_nano_seconds}= $1;
            $flag_mtime = 0;
        }elsif ($flag_ctime && ($element_line =~ /\s{4}seconds: (.*)/)){
            $Readdirplus{$entry_as_key}{ctime_seconds}= $1;
        }elsif($flag_ctime && ($element_line =~ /\snano seconds: (.*)/)){
            $Readdirplus{$entry_as_key}{ctime_nano_seconds}= $1;
            $flag_ctime = 0;
        }
        else{
            1;
        }
    }
  }
}
1;
