# Hey emacs, this is basically -*- perl -*-
#
# Copyright (c) 2010 Network Appliance, Inc.
# All rights reserved.
#
# lib/CR/wafl_corruption.pm
#
## @summary Routines for using 'wreck' corruption command to
##    introduce WAFL inconsistencies before running WAFL Iron
##
## @description
##    The 'wreck' command used to introduce WAFL inconsistencies
##    in the repository for testing purposes, like:
##   - 'inode_block_count_corruption': Corrupt block count in a inode of a file
##   - 'dir_generation_corruption': Corrupt a directory generation of a directory
##   - 'dir_entry_file_name_hash_corruption': Corrupt filename hash of the directory entry
##   - 'dir_entry_pointing_to_another_inode': Corrupt the directory entry to point to another inode
##   - 'data_blocks_lost': Corrupt indirect blocks in a buftree
##   - 'dir_entry_dir_name_corruption': Corrupt directory name in a directory entry
##   - 'dir_entry_file_name_corruption': Corrupt file name in a directory entry
##   - 'file_generation_corruption': Corrupt inode generation of a file
## @synopsis
##    use CR::wafl_corruption;
##    cr_wreck (  files => $filesref,
##                vol_type => $vol_type,
##                inject_string => $inject_string,
##                wafliron => $wafliron,
##                check_messages => $check_messages
##    );
##
## @examples
##    use CR::wafl_corruption;
##    wafl_corruption_init();
##    wafl_corruption_setup();
##    eval{
##	      cr_wreck (  files => $filesref,
##                    vol_type => $vol_type,
##                    inject_string => $inject_string,
##                    wafliron => $wafliron,
##                    check_messages => $check_messages
##        );
##    };
##    if ($@) {
##        $Log->comment("cr_wreck() failed: $@");
##        return $TCD::FAIL;
##    }
##
##
## @usage
##
##    cr_wreck (  files => $filesref,
##                vol_type => $vol_type,
##                inject_string => $inject_string,
##                wafliron => $wafliron,
##                check_messages => $check_messages
##    );
##
##    $fileref        - Reference of an array of hashes with information
##                      about the created files in the form:
##                      {
##                          client_obj => ...,
##                          ns_file    => {
##                              volume => (volume name),
##                              junction-path => (volume junction path)
##                              path => (ns file path, relative to ns root),
##                              name => (ns file name)
##                              nfs_fh => (nfs fh to get to the file)
##                          },
##                          obj_file   => {
##                              volume => (volume name),
##                              junction-path => (obj vol junction path),
##                              path => (obj file path, relative to obj vol root),
##                              name => (obj file name/ns file handle),
##                              fh => (obj file handle)
##                          }
##                      }
##
##    $vol_type       - 'ns', or 'obj'. Volume type Namespace constituent or
##                       Data Constituent
##
##    $inject_string  - 'inode_block_count_corruption','dir_generation_corruption','dir_entry_file_name_hash_corruption','dir_entry_file_name_corruption',
##                        'dir_entry_pointing_to_another_inode', 'dir_entry_dir_name_corruption',
##                        'file_generation_corruption' or 'data_blocks_lost'
##                        Type of WAFL corruption
##
##    $wafliron       - 0 or 1,  default is 0.
##                       0 - Don't run wafliron after corruption.
##                       1 - Run wafliron after corruption.
##
##    $check_messages  - 0 or 1, default is 0.
##                       0 - Don't check wafliron mlog messages after corruption.
##                       1 - Check wafliron mlog messages after corruption.
##    Note: The check_messages option should be used only with wafliron
##
## @burt : <481796>
## @author Jayakumar C

package wafl_corruption;

# Define common methods in 'main' namespace
package main;
# The external interface into this library
our @EXPORT_OK = qw( wafl_corruption_init
                     wafl_corruption_setup
                     wafl_corruption_cleanup
                     wafl_corruption_uninit
                     cr_wreck
                     get_file_path
                     start_wafliron
                     inject_corruption
                     get_snap_id
					);
use strict;
use warnings;

# Module import.
use TCD;
use Options;
use NACL::APISet;
use NATE::BaseException;
use Data::Dumper;
use Server;
use Node::NG::Transit;
use LogReader::MlogMessages;
use Params::Validate qw(SCALAR OBJECT ARRAYREF HASHREF validate);

# The library requries that the calling thpl define
# the following variables

# Global variables
use vars qw(
    $Test
    $Options
    $FILER
    $Log
    $VSERVER_GX_NAME
    $TESTDIR
    $Server
    $Host
    @Nodes
    $Wipoc_flag
    $Wipt_flag
    );

# The following objects are used to talk to the filer
my $Node;

##-------------------------------------------------------------
# wafl_corruption_init
#
# Called by the thpl file as a part of init() to provide for
# creation of objects that would be needed by the tests.
##--------------------------------------------------------------
sub wafl_corruption_init ()
{
    $Log->comment("wafl_corruption_init() is called");
    eval {
        $Log->comment("======Creating Host And Connection Objects");
        $Host = host($FILER)         or die($@);
        $Log->comment("======Creating a server object");
        $Server = new Server;
    };
    if ($@) {
        $Log->comment("wafl_corruption_init() failed : $@");
    }
} # End of wafl_corruption_init

##-------------------------------------------------------------
# wafl_corruption_setup()
#
# Called by the thpl file as a part of setup() to provide for
# configuring the script's setup.
##--------------------------------------------------------------
sub wafl_corruption_setup()
{
    $Log->comment("wafl_corruption_setup() is called");
    eval {
        @Nodes         = $Server->servers;
        $Node          = $Nodes[0];
        ### Resetting values  ######
        $Log->comment("======Resetting the flags to default values =======");
        reset_wafliron_flags();
    };
    if ($@) {
        $Log->comment("wafl_corruption_setup() failed: $@");
    }
}## End of wafl_corruption_setup()

##-------------------------------------------------------------
# wafl_corruption_cleanup()
#
# Called by the thpl to provide for any unconfiguration
# after the script has been run.
##--------------------------------------------------------------
sub wafl_corruption_cleanup()
{
    $Log->comment("wafl_corruption_cleanup() is called");
    eval {
        $Log->comment("======Resetting the flags to default values =======");
        reset_wafliron_flags();
    };
    if ($@) {
        $Log->comment("wafl_corruption_cleanup() failed: $@");
    }
}## End of wafl_corruption_cleanup()
##-------------------------------------------------------------
# wafl_corruption_uninit()
#
# Called by thpl at the end to uninitialize objects
##--------------------------------------------------------------
sub wafl_corruption_uninit()
{
    $Log->comment("wafl_corruption_uninit() is called");

}## End of wafl_corruption_uninit()

#--------------------------------------------------------------
# This function will return snapid of snapshot
# in given volume and '0' on failure.
#
#------------------------------------------------------------
sub get_snap_id (%)
{
 my %opts = validate(
        @_,
        {
            snapshot   => { type => SCALAR },
            volume => { type => SCALAR }
        }
      );
my $Snap_id;
 eval { 
    my $res =
          $Server->volume_snapshot_show(vserver  => $VSERVER_GX_NAME,
                                        volume   => $opts{volume},
                                        snapshot => $opts{snapshot}
                                       ) or die "snapshot creation failed";
    $Log->comment("====== Snapshot show output is:\n" . Dumper(\$res));

        $Log->comment("====== Getting the snapid of snapshot : $opts{snapshot}");
        $Snap_id = ${$$res[0]}{'physical-snap-id'};
};
if($@) {
  die("get_snap_id() failed: $@");
} else {
    return $Snap_id;
}
}

##----------------------------------------------------
# This function will return us file path in object 
# or name space volume
# it will take list of files @file and volume_type as
# parameter 
##----------------------------------------------------
sub get_file_path ( % ) 
{
     my %opts = validate(
        @_,
        {
            files    => { type => HASHREF },
            vol_type => { type => SCALAR, regexp => qr/ns|obj/i }
        }
      );

    my $files = $opts{files};
    my $vol_type = $opts{vol_type};
    my %ret_hash;
    my ($created_file, $volume, $input_file, $filename, $input_dir); 
        if ($vol_type =~ /ns/i){
            $created_file = $files->{ns_file};
            $volume = $created_file->{'volume'};
            $input_file = "$created_file->{'path'}";
            $filename = $created_file->{'name'};
            if ($created_file->{'path'} =~ /(.*)$filename/){
                $input_dir = $1;
            }
        } elsif ($vol_type =~ /obj/i){
            $created_file = $files->{obj_file};
            $volume = $created_file->{'volume'};
            $input_file = "$created_file->{'path'}";
            $filename = $created_file->{'name'};
            if ($input_file =~ /(.*)\/[$filename]/){$input_dir = $1;}
        } else {
              die "======Wrong vol_type : $vol_type";
        }
        my $aggr = get_aggr(vserver => $VSERVER_GX_NAME, volume => $volume);
        $Node = get_node ( aggr => $aggr );

        $ret_hash{'volume'}   = $volume;
        $ret_hash{'path'}     = $input_file;
        $ret_hash{'filename'} = $filename;
        $ret_hash{'dirpath'}  = $input_dir;
        $ret_hash {'aggr'} = $aggr;
        $ret_hash {'Node'} = $Node;

    return \%ret_hash ;
}

#---------------------------------------------------------------------------
# This wreck can handle multiple corruptions as well as it will inject 
# single corruption, We are writing this version of wreck function 
# in order to get wafliron call out of cr_wreck() to increase flexibility.
#
#-----------------------------------------------------------------------------
sub inject_corruption(%) {
    $Log->comment("inject_corruption() called with @_");
     my %opts = validate(
        @_,
        {
            container_type => {  type => SCALAR, default => "vol", regexp => qr/vol|aggr/i }, 
            vol_type => {  type => SCALAR, optional => 0, regexp => qr/ns|obj/i },
            inject_string => {type => ARRAYREF },
            volume => { type => SCALAR} 
      }
    );#end of validate function
    my $cont_type = $opts{container_type};
    my $vol_type = $opts{vol_type};
    my @inject_strings = @{$opts{inject_string}};
    my $volume = $opts{volume};
    my $inject_file = '/etc/'.$volume;
    my ($aggr,$inject_string,$file_status,$wreck_status);
    $Test->step("Get the Node and Aggregate of $volume vol in Vserver: $VSERVER_GX_NAME");
    $aggr = get_aggr( vserver => $VSERVER_GX_NAME,
                               volume  => $volume);
    $Node = get_node ( aggr => $aggr );
    $Server->set_execution_server(-Server => $Node);

    $file_status = &_create_inject_file ( Node => $Node,
                                          server => $Server,
                                          inject_strings => [@inject_strings],
                                          inject_file => $inject_file
                                         );
     die "=========== create_inject_file failed" unless ($file_status);
     $inject_string = $inject_file;
     eval {
         $Test->step("Run wreck on the volume : $volume");
         $wreck_status = &_wreck ( type          => $cont_type,
                                   server        => $Server,
                                   inject_string => $inject_string,
                                   volname       => $volume,
                                   aggrname      => $aggr
                                  );
     };
    if ($@) {
        die("======_wreck FAILED with error:$@");
    }
    return 1;
}

##------------------------------------------------------------------
# Start Wafliron on aggregate to fix or make sure that corruption
# either got fixed or does not exist
#
##------------------------------------------------------------------
sub start_wafliron(%) {
    my %opts = validate ( @_ , {
         aggr => { type => SCALAR  },
         wafl_iron_flag_set => { type => SCALAR, default => 0 },
         check_messages => { type => SCALAR, default => 0 },
         wi_messages =>  {   type => ARRAYREF, optional => 1 },
         wi_count => {   type => ARRAYREF, optional => 1 }, }
    );
    
    my $aggr = $opts{aggr};
    my $wafl_iron_flag_set = $opts{wafl_iron_flag_set};
    my $check_messages = $opts {check_messages};
    my  @wi_messages;
    my  @wi_count;
    my ($Logh,$messages,  $wafl_status);
    if( $check_messages == 1){
        @wi_messages = @{$opts{wi_messages}};
        @wi_count = @{$opts{wi_count}};
        if( @wi_messages == undef || @wi_count == undef ){
                    die("wi_messages and wi_count are undefined");
        }
    }
    # Set or Reset wafliron flags to fix or verify corruption
    if( $wafl_iron_flag_set == 1 ) {
        $Test->step("Set the flags to run wafliron");
        set_wafliron_flags();
    } else {
        $Test->step ("reset wafliron flags to verify no corruption");
        reset_wafliron_flags ("1","1");
    }

    $Node = get_node ( aggr => $aggr );
    $Server->set_execution_server(-Server => $Node);

        
    $Test->step("Run wafliron on the aggr : $aggr");
    #### Start logging  ######
    $Logh = &_start_log(server => $Server);
    $wafl_status = &_wafliron ( server => $Server,
                                filer_aggr => $aggr
                               ); 
    die "=======_wafliron failed" unless($wafl_status);
    #### Stop logging   ######
    $messages = &_stop_log(handle => $Logh, server => $Server);

    # This is to verify wafliron fixed the corruption in expected way
    $Log->comment("====== MESSAGES are: $messages");
    if ($check_messages == 1 ){
        $Test->step("Verify the Mlog messages for the requied logs");
        $Log->step("Step------ Validate expected corruption found message and data location.",
                    " Expected Result:expected messages should be found in logs");
        $Log->comment(  "======Looking for error messages in "
                         . "WALFIRON LOG MESSAGES => $messages");

        #@wi_messages = @{$opts{wi_messages}};
        #@wi_count = @{$opts{wi_count}};
        my $error_check = &_verify_message( server => $Server,
                                            messages => $messages,
                                            wi_messages => \@wi_messages,
                                            wi_count => \@wi_count
                                           );
    die "======_error_check FAILED" unless($error_check);
    } else {
        $Log->comment("======  Skipping message verification  =====");
    }

}
##-----------------------------------------------------------------
## Called by the thpl file to introduce WAFL inconsistencies
## in the repository
##-----------------------------------------------------------------
sub cr_wreck ( % )
{
	#my %opts = @_;
    $Log->comment("cr_wreck() called with @_");
    my %opts = validate(
        @_,
        {
            files  =>  {   type => ARRAYREF, optional => 0 },
            container_type => {  type => SCALAR, default => "vol", regexp => qr/vol|aggr/i },
            vol_type => {  type => SCALAR, optional => 0, regexp => qr/ns|obj/i },
            inject_string => {type => SCALAR,optional => 0, regexp => qr/inode_block_count_corruption|dir_generation_corruption|dir_entry_file_name_hash_corruption|dir_entry_pointing_to_another_inode|data_blocks_lost|file_generation_corruption|dir_entry_file_name_corruption|dir_entry_dir_name_corruption/i  }, 
            wafliron => { type => SCALAR, optional => 1, default => 0 },
            check_messages => { type => SCALAR, optional => 1, default => 0 },
            wi_messages =>  {   type => ARRAYREF, optional => 1 },
            wi_count => {   type => ARRAYREF, optional => 1 },
        }
    );
    my ($input_dir,$created_obj_file, $created_ns_file, $aggr, $volume, $num,$input_file, $messages, $Logh, $created_obj_file2, $file,
            $file_id, $file_id2, $res, $wafl_status,@wi_messages,@wi_count,$wreck_status, $filename, @dirs, $parent_dir, $input_file2,$fullpath);
    my @files = @{$opts{files}};
    my $cont_type = $opts{container_type};
    my $vol_type = $opts{vol_type};
    my $inject_string = $opts{inject_string};
    my $wafliron = $opts{wafliron};
    my $check_messages = $opts{check_messages};
    #@wi_messages = $opts(wi_messages);
    #@wi_count = $opts(wi_count);
    eval {
        if ($vol_type =~ /ns/i){
            $created_ns_file = $files[0]->{ns_file};
            $volume = $created_ns_file->{'volume'};
            $input_file = "$created_ns_file->{'path'}";
            $filename = $created_ns_file->{'name'};
            if ($created_ns_file->{'path'} =~ /(.*)$filename/){
                $input_dir = $1;
            }
        } elsif ($vol_type =~ /obj/i){
            $created_obj_file = $files[0]->{obj_file};
            $volume = $created_obj_file->{'volume'};
            $input_file = "$created_obj_file->{'path'}";
            $filename = $created_obj_file->{'name'};
            #### Modifying path with directory name mentioned in Junction path
            #$fullpath = $created_obj_file->{'path'};
            #$fullpath = "$created_obj_file->{'junction-path'}";
            #@dirs = split( '/', $fullpath );
            #if ( $fullpath =~ /\w+$dirs[0](.*)/ ) {$input_file = $1.$input_file;}
            if ($input_file =~ /(.*)\/[$filename]/){$input_dir = $1;}
        } else {
              die "======Wrong vol_type : $vol_type";
        }
        $Test->step("Get the Node and Aggregate of the volume in a Vserver : $VSERVER_GX_NAME");
        $aggr = get_aggr( vserver => $VSERVER_GX_NAME,
                               volume  => $volume);
        $Node = get_node ( aggr => $aggr );
        $Server->set_execution_server(-Server => $Node);
        $Test->step("Reset the flags to default values");
        reset_wafliron_flags();
        $Test->step("Creating an injection file ");
        $Log->comment("====== Creating an injection file\n");
        if ($inject_string =~ /inode_block_count_corruption/){
            #$num = int(rand(255));
            # Assigning count=250 instead of any random number
            $inject_string = "inode_blk_cnt inum=path:$input_file count=250";
            @wi_messages = ("Decreasing block count","Block counts corrected");
            @wi_count = ("1","1");
        } elsif ($inject_string =~ /dir_generation_corruption/){
            @dirs = split('/',$input_dir);
            if ($input_dir =~ /(.*)$dirs[$#dirs]/){$parent_dir = $1;};
            $inject_string = "dir_gen inum=path:$parent_dir dirent=$dirs[$#dirs] gen=12345678";
            @wi_messages = ("Inodes link counts corrected:","Directory entries pointing to bad files:","lost files or directories relinked into lost\+.\\w+");
            @wi_count = ("1","1","1");
        } elsif($inject_string =~ /file_generation_corruption/){
            $inject_string = "inode_gen inum=path:$input_file gen=0xfff";
            @wi_messages = ( "Bad generation", "1 lost files or directories relinked into lost\+.found");
            @wi_count    = ("1", "1");
        } elsif ($inject_string =~ /dir_entry_file_name_corruption/){
            #The start and end are staring and ending character in the string which should be filled with the ascii character as mentioned by fill field
            #First two characters are filled with "a"
            $inject_string ="dir_name_pattern inum=path:$input_dir dirent=$filename start=1 end=2 fill=97";
            # TODO: Need clarification for wafliron Mlog messages 
            @wi_messages = ("Clearing RAID pseudo bad flag");
            @wi_count = ("1");
        } elsif ($inject_string =~ /dir_entry_dir_name_corruption/){
            #The start and end are staring and ending character in the string which should be filled with the ascii character as mentioned by fill field
            #First two characters are filled with "a"
            @dirs = split('/',$input_dir);
            if ($input_dir =~ /(.*)$dirs[$#dirs]/){$parent_dir = $1;};
            $inject_string ="dir_name_pattern inum=path:$parent_dir dirent=$dirs[$#dirs] start=1 end=2 fill=97";
            # TODO: Need clarification for wafliron Mlog messages
            @wi_messages = ("Clearing RAID pseudo bad flag");
            @wi_count = ("1");
        } elsif ($inject_string =~/dir_entry_file_name_hash_corruption/){
            $inject_string = "dir_hash inum=path:$input_dir dirent=number:1 hash=0xabcdabcd";
            @wi_messages = ("has incorrect name hash, correcting","Directory hashes fixed");
            @wi_count = ("1","1");
        } elsif ($inject_string =~ /dir_entry_pointing_to_another_inode/){
            my ($i,@inject_strings, $inject_file, $file_status, $gen,$count);
            $inject_file = '/etc/'.$volume;
            $count = 0;
            for ($i= 1; $i<= ($#files); $i++) {
                if (($vol_type =~ /obj/)and ($files[$i]->{obj_file}->{'volume'} =~ $volume)){
                    #$fullpath = "$files[$i]->{obj_file}->{'junction-path'}";
                    #@dirs = split( '/', $fullpath );
                    #if ( $fullpath =~ /\w+$dirs[0](.*)/ ) {$input_file2 = $1."$files[$i]->{obj_file}->{'path'}";}
                    $input_file2 = $files[$i]->{obj_file}->{'path'};
                    $count++;
                }elsif (($vol_type =~ /ns/)and ($files[$i]->{ns_file}->{'volume'} =~ $volume)){
                    $input_file2 = $files[$i]->{ns_file}->{'path'};
                    $count++;
                };
            }
            if ($count == 0){
                $Log->comment("Need atleast two files in the same volume : $volume");
                die "Need atleast two files in the same volume: $volume";
            }
            # Get the generation number of the second file
            $gen = &_get_file_gen ( Node => $Node,
                                    server => $Server,
                                    input_file => $input_file2,
                                    volume => $volume
                                   ) or 0;
            # Multiple corrupts in the directory entry 1) Corrupting generation of the inode 2) corrupting the inode info
            @inject_strings = ("dir_gen inum=path:$input_dir dirent=$filename gen=$gen","dir_fileid inum=path:$input_dir dirent=$filename inum2=path:$input_file2");
            $file_status = &_create_inject_file ( Node => $Node,
                                                  server => $Server,
                                                  inject_strings => [@inject_strings],
                                                  inject_file => $inject_file
                                                 ) or 0;
            die "=========== create_inject_file failed" unless ($file_status);
            $inject_string = $inject_file;
            @wi_messages = ("nlink too small","Inodes link counts corrected","lost files or directories relinked into lost\+.\\w+");
            @wi_count = ("1","1","1");
        } elsif ($inject_string =~ /data_blocks_lost/){
		    $inject_string = "ind_vol_bno inum=path:$input_file level=1 bno=10 new_bno=0xabcdabcd index=1";
            @wi_messages = ("Clearing out of range vbn","Block counts corrected","lost blocks collected into 1 files in lost\+.\\w+");
            @wi_count = ("1","1","1");

        } else {
              die "====== Wrong inject_string : $inject_string";
        }

        $Test->step("Run wreck on the volume : $volume");
        $wreck_status = &_wreck ( type          => "vol",
                                  server        => $Server,
                                  inject_string => $inject_string,
                                  volname       => $volume,
                                  aggrname      => $aggr
                                ) or 0;
        die "======_wreck FAILED" unless ($wreck_status);

        $Test->step("Set the flags to run wafliron");
        set_wafliron_flags();
        if ($wafliron == 1){
            $Test->step("Run wafliron on the aggr : $aggr");
            #### Start logging  ######
            $Logh = &_start_log(server => $Server);
            $wafl_status = &_wafliron ( server => $Server,
                                        filer_aggr => $aggr
                                      ) or 0;
            die "=======_wafliron failed" unless($wafl_status);
            #### Stop logging   ######
            $messages = &_stop_log(handle => $Logh, server => $Server);
            $Log->comment("====== MESSAGES are: $messages");
            if ($check_messages ==1){
                $Test->step("Verify the Mlog messages for the requied logs");
                $Log->step("Step------ Validate expected corruption found message and data location.",
                    " Expected Result:expected messages should be found in logs");
                $Log->comment(  "======Looking for error messages in "
                     . "WALFIRON LOG MESSAGES => $messages");
                my $error_check = &_verify_message( server => $Server,
                                                    messages => $messages,
                                                    wi_messages => \@wi_messages,
                                                    wi_count => \@wi_count
                                                  ) or 0;
                die "======_error_check FAILED" unless($error_check);
            } else {
                $Log->comment("======  Skipping message verification  =====");
            }
        } else {
              $Log->comment("======  Skipping wafliron  =====");
        }
    };
    if ($@) {
        $Log->comment("cr_wreck() failed: $@");
        die "cr_wreck() failed: $@";
    }
}# End of cr_wreck()

####  Set the flags to wafliron safe values   ####
sub set_wafliron_flags ()
{
    $Log->comment("======Setting the flags to WAFLIRON safe values =======");
    my @flags = ("wafl_iron_panic_on_corruption","wafl_inconsistent_panic_testing");
    my @flag_values = ("0","1");
    my $set_flags = &_setflags ( server => $Server,
                                 flags => \@flags,
                                 flag_values => \@flag_values
                               )or 0;
    die "======_setflags FAILED" unless($set_flags);
    ## Flags set done
}

####  Resetting the flags to defaults   ####
sub reset_wafliron_flags ()
{
    my @flag_values = @_;
    $Log->comment("======Resetting the flags to default values =======");
    my	@flags = ("wafl_iron_panic_on_corruption","wafl_inconsistent_panic_testing");
    #setting wafl_iron_panic_on_corruption to "0" every time
    $flag_values[0] = "0";
    #In wafliron cases sometime corruption gets in snapshot too, wafliron will
    #panic if we dont keep wafl_inconsistent_panic_testing to 1
    $flag_values[1] = defined($flag_values[1]) ? $flag_values[1] : 0;
    my	$set_flags = &_setflags ( server => $Server,
                                  flags => \@flags,
                                  flag_values => \@flag_values
                                )or 0;
    die "======_setflags FAILED" unless($set_flags);
    ## Flags set done
}

### Get the node and aggr of a volume in a vserver ####
sub get_aggr ()
{
    $Log->comment("get_aggr() called with @_");
    my %opts = @_;
    my $vserver = $opts{vserver};
    my $volume = $opts{volume};
    # Get all the volumes belonging to the Vserver.
    my @vs0_vols = $Server->volume_show ( vserver => $vserver,
                                          'volume' => $volume);
    my $aggr_vs =  $vs0_vols[0]->{aggregate};
    return $aggr_vs;
}

### Get the node of a aggregate ###########
sub get_node ()
{
    $Log->comment("get_node() called with @_");
    my %opts = @_;
    my $aggr_vs = $opts{aggr};
    #Run aggr show to get all aggrs and nodes
    my $output = $Server->storage_aggregate_show(aggregate => $aggr_vs);
    my $Node = ${$$output[0]}{nodes};
    return $Node;
}

# Run the wreck tool
############ _wreck ############################################################
#                                                                              #
# wreck tool is ran in node shell with test privileges.                        #
#                                                                              #
################################################################################
sub _wreck()
{
	$Log->comment("_wreck() called with @_");
	my %opts          = @_;
    my $type          = $opts{type};
    my $inject_string = $opts{inject_string};
    my $volname = $opts{volname};
    my $aggr = $opts{aggrname};
    my ($Logh, $messages, $read_data, $res, $i, $count, $cmd, $ret);
    my $err_flag = 0;
    if ($type eq "aggr") {
        $cmd = qq( "priv set test;wreck -A $aggr '$inject_string'" );
        $ret = $Server->system_node_run( node    => $Node,
                                         command => $cmd,
                                         '-Timeout' => 300);
    } elsif ($type eq "vol") {
        $cmd = qq( "priv set test;wreck $volname '$inject_string'" );
        $ret = $Server->system_node_run( node    => $Node,
                                         command => $cmd,
                                         '-Timeout' => 300);
    } else {
        die "check the options passed to _wreck";
    }
    $Log->comment("====== wreck return output is:\n ". Dumper $ret);
    die "wrecking failed" unless ($ret =~ /Finished/i || $ret =~ /Wrote vbn .*/i );
	return 1;
}

# Set the flags
############ _setflags #########################################################
#                                                                              #
# set flags is ran in node shell with test privileges.                         #
#                                                                              #
################################################################################
sub _setflags()
{
    $Log->comment("_setflags() called with @_");
    my %opts = @_;
    my @flags = @{$opts{flags}};
    my @flag_values    = @{$opts{flag_values}};
	$Log->step(  "Step------Set flags $flags[0] $flag_values[0]"
               . "$flags[1] $flag_values[1] and verify"
               . "using printflag command. Expected Result: All flags are set");

    $Log->comment(  "======Setting the value of the flag \""
                  . "$flags[0]\".");
    $Server->system_node_run( node    => $Node,
                              command => "setflag $flags[0] $flag_values[0]");
    $Server->system_node_run( node    => $Node,
                              command => "printflag $flags[0]");
    $Log->comment(  "======Setting the value of the flag \""
                  . "$flags[1]\".");
    $Server->system_node_run( node    => $Node,
                              command => "setflag $flags[1] $flag_values[1]");
    $Server->system_node_run( node    => $Node,
                              command => "printflag $flags[1]");
    return 1;
}

# Start logging the messages before running the wafliron
########## _log_start ##########################################################
#                                                                              #
# Start to monitor the messages logged by current Wafliron session.            #
#                                                                              #
################################################################################
sub _start_log()
{
    $Log->comment("_start_log() called with @_");
    my %opts = @_;
    $Log->comment("======Start logging..");
    my $logrdr = LogReader::MlogMessages->new($opts{server});
    $logrdr->start;
    # Sleep for 2 seconds delay to capture the logs
    sleep 2;
    return $logrdr;
}

# Stop logging the messages after wafliron is stopped
############ _log_stop #########################################################
#                                                                              #
# stop  monitoring  the messages logged by current Wafliron session and return #
# the set of wafl messages logged between start and stop.                      #
#                                                                              #
################################################################################
sub _stop_log()
{
    $Log->comment("_stop_log() called with @_");
    my %opts = @_;
    my ($lines, $report, $logrdr);
    $logrdr = $opts{handle};
    # Sleep for 5 seconds delay to capture the logs
    sleep 15;
    $logrdr->stop;
    $Log->comment("======Finished logging..");
    $lines = $logrdr->matches('.');
    if (@$lines) {
        $report = join "\n", @$lines;
    }
    return $report;
}

# Verify the log messages contains the required wafliron logs
sub _verify_message ()
{
    $Log->comment("_verify_message() called with @_");
    my %opts = @_;
    my $messages = $opts{messages};
    my @wi_messages = @{$opts{wi_messages}};
    my @wi_count    = @{$opts{wi_count}};
    my ($count, $i, $err_flag);
    $err_flag =1;
    $Log->step(
        "Step------ Validate expected corruption found message and data location.",
        " Expected Result:expected messages should be found in logs");

    $Log->comment(  "======Looking for error messages in "
                  . "WALFIRON LOG MESSAGES => $messages");

    $Log->comment("====== Validating for message\n");
    $count = 0;
    $i     = 0;
        foreach (@wi_messages) {
            $count = 0;
            while ($messages =~ /$_/g) { $count++; }

            $Log->comment("====== Message found $count times");
            if ( $count != $wi_count[$i]) {
                $Log->error(
                    "======FAILED: Message FOUND with wrong count => $_ \"$count\", should be \"$wi_count[$i]\" times\n");
                $err_flag = 0;
            }
            last if ($err_flag == 0);
            $Log->comment(
                       "======PASSED: Message FOUND with right count => $_ \"$count\" times\n");
            $i++;
        } ## end foreach (@wi_messages)
    return $err_flag;
}



# Run wafliron on the required volume/aggregate
sub _wafliron()
{
    $Log->comment("_wafliron() called with @_");
    my %opts = @_;
    my $filer_aggr = $opts{filer_aggr};
    my ($res, $err_flag);
    $err_flag = 1;
    $Log->step("Step------set privilage to test and Start wafliron on aggregate"
               . "Expected result: wafliron should start successful");
    $Server->set(privilege => "test");
    $Server->storage_aggregate_wafliron_start(aggregate => $filer_aggr);
    $Log->step(
            "Step------validate that wafliron is running by using wafliron show"
              . "Expected Result:iron-status should be in scanning state");
    &_cmd_wait_or_check_state(server  => $Server,
                              cmd     => "wafliron_show",
                              options => \{aggregate => $filer_aggr},
                              field   => 'iron-status',
                              value   => 'scanning');
    $Log->comment("======Polling for wafliron to complete");
    &_cmd_wait_or_check_state(server  => $Server,
                              cmd     => "wafliron_show",
                              options => \{aggregate => $filer_aggr},
                              field   => 'iron-status',
                              value   => 'not_running',
                              wait    => 1);
    eval {
        $res = "";
        $res = $Server->storage_aggregate_wafliron_show(aggregate => $filer_aggr);
    };
    $Log->comment("====== Wafliron show output is:\n" . Dumper(\$res));
    if ($@) {
        $Log->comment("wafliron failed: $@");
        $err_flag = 0;
    }
   return $err_flag;
}


#################################################################################

############### _cmd_wait_or_check_state ########################################
# Executes a particular type of show cmd and polls or check for the given 'field'
#       to reach the given 'value' depending on value of parameter 'wait' .
#
# Required options:
#               server => Server object
#               cmd    => command to execute
#               cmd implemented so far:
#                       aggregate_show => storage_aggregate_show
#                       volume_show    => volume_show
#                       wafliron_show => storage_aggregate_wafliron_show
#               options=> reference of arguments to be passed to the show call.
#               field  => attribute of show for which checking has to be done.
#               value  => state of field to wait for..
#               wait   => 1 if polling required
#                         0 if only check required.(default)
#################################################################################
sub _cmd_wait_or_check_state () {
	$Log->comment("_cmd_wait_or_check_state() called with @_");
    my %opts    = @_;
    my $server  = $opts{server};
    my $field   = $opts{field};
    my $value   = $opts{value};
    my $cmd     = $opts{cmd};
    my $options = $opts{options};
    my ($t_val, $output, $flag) = (0, "", 1);
    my ($timeout);
    my @option = %$$options;

    if ($opts{wait}) {
        $timeout = $opts{timeout} ? $opts{timeout} : 300;
    } else {
        $timeout = 0;
    }
    if ($cmd =~ /aggregate_show/) {
        eval {
            $output = "";
            $output = $server->storage_aggregate_show(@option);
        };
        if ($@) {
            $Log->comment("Aggregate show failed: $@");
            return 0;
        }
        $Log->comment("====== Aggregate show output is:\n" . Dumper(\$output));
    } elsif ($cmd =~ /volume_show/) {
        eval {
            $output = "";
            $output = $server->volume_show(@option);
        };
        if ($@) {
            $Log->comment("Volume show failed: $@");
            return 0;
        }
        $Log->comment("====== Volume show output is:\n" . Dumper(\$output));
    } elsif ($cmd =~ /wafliron_show/) {
        eval {
            $output = "";
            $output = $server->storage_aggregate_wafliron_show(@option);
        };
        if ($@) {
            $Log->comment("Wafliron show failed: $@");
            return 0;
        }
        $Log->comment("====== Wafliron show output is:\n" . Dumper(\$output));
    }
    $t_val       = time;
    $flag        = 0 if (${$$output[0]}{$field} =~ /$value/);
    $timeout += $t_val;
    while ($flag and $t_val < $timeout) {
        if ($cmd =~ /aggregate_show/) {
            eval {
                $output = "";
                $output = $server->storage_aggregate_show(@option);
            };
            if ($@) {
                $Log->comment("Aggregate show failed: $@");
                return 0;
            }
            $Log->comment(
                       "====== Aggregate show output is:\n" . Dumper(\$output));
        } elsif ($cmd =~ /volume_show/) {
            eval {
                $output = "";
                $output = $server->volume_show(@option);
            };
            if ($@) {
                $Log->comment("Volume show failed: $@");
                return 0;
            }
            $Log->comment("====== Volume show output is:\n" . Dumper(\$output));
        } elsif ($cmd =~ /wafliron_show/) {
            eval {
                sleep(15);
                $output = "";
                $output = $server->storage_aggregate_wafliron_show(@option);
            };
            if ($@) {
                $Log->comment("Wafliron show failed: $@");
                return 0;
            }
            $Log->comment("====== Wafliron show output is:\n" . Dumper(\$output));
        } ## end elsif ($cmd =~ /wafliron_show/)
        eval { $flag = 0 if (${$$output[0]}{$field} =~ /$value/); };
        if ($@) {
                $Log->comment("Checking required value failed: $@");
                return 0;
            }
        $t_val = time;
    } ## end while ($flag and $t_val <...
    if ($flag and $t_val >= $timeout) {
        $Log->comment("======Waiting for $field to reach $value Timed Out");
        return 0;
    }
    return 1;
} ## end sub _cmd_wait_or_check_state ()

# Create inject file to inject multiple corruptions
sub _create_inject_file()
{
    $Log->comment("_create_inject_file() called with @_");
    my %opts = @_;
    my $Node = $opts{Node};
    my $inject_file = $opts{inject_file};
    my @inject_strings = @{$opts{inject_strings}};
    $Log->step("Step------set privilage to test and Start wafliron on aggregate"
               . "Expected result: wafliron should start successful");
    $Log->comment("====== Removing an injection file\n");
    # Remove the file if already exist
    $Server->system_node_run( node    => $Node,
                              command => "rm $inject_file",
                              '-Timeout' => 300);
    $Log->comment("====== Creating an injection file\n");
    eval {
        foreach (@inject_strings) {
            my $cmd = qq("wrfile -a $inject_file \'$_\'");
            $Server->system_node_run( node    => $Node,
                                      command => $cmd,
                                      '-Timeout' => 300);
        } ## End of foreach @inject_strings
    };
    if ($@) {
        $Log->comment("_create_inject_file() failed: $@");
        return 0;
    }
    return 1;
} ## end sub _create_inject_file()

#### Get the existing Hash value of the file######
sub _get_file_gen()
{
    $Log->comment("_get_file_gen() called with @_");
    my %opts = @_;
    my $Node = $opts{Node};
    my $input_file = $opts{input_file};
    my $volume = $opts{volume};
    my ($res,$gen);
    ($gen, $res) = ("", "");
    $Server->system_node_run( node    => $Node,
                              command => "showfh /vol/$volume/$input_file",
                              output  => \$res);
    ($gen) = $res =~ /gen=(\w*)/;
    return $gen;
}  ### end sub _get_file_gen()
1;
