#!/usr/local/bin/perl
# $Id: //depot/prod/test/Rsierranevada/lib/CSM.pm#1 $
## @summary     Provides the api to monitor performance, isolate issues, gather error counters, and surmise CSM health.
## @author      cwilcox@netapp.com,dl-cluster-qa@netapp.com
## @pod here

=head1 NAME

CSM.pm

=head1 SYNOPSIS
 SYNOPSIS:
    use Server;
    use CSM;
    $obj = new Server;
    $obj = new CSM(server => $Server);

    $obj->initialize(server => $Server);

    $obj->csm_health_summary(server => $Server,
                             level  => low|medium|high,
                            );

    $obj->get_session_error_counters(server => $Server);

C<CSM> Cluster Session Manager, or CSM provides a managed abstraction layer or tunnel for SpinNP sessions between the blades of each node in the cluster.
       Local node session communication between Nblade, and Dblade is handled by NullTX transport, while RC (Reliable Connection) transports 
       internodal session messaging between the cluster blades. CSM uses UUID endpoints to map paths between blades. CSM connects blades... not
       nodes through mapped Logical Interfaces (LIFs). Interaction exists with VifMgr, VLDB, and MGWD. 
       
       Internodal Blade Session Communication:
       SpinNP <-> Blade UUID <-> RC <-> UDP <-> IP <-> Ethernet <-> IP <-> UDP <-> RC <-> Blade UUID <-> SpinNP
          
       Local Blade Session Communication:
       SpinNP <-> Blade UUID <-> NullTX <-> Blade UUID <-> SpinNP

       http://wikid.netapp.com/w/CSM 

=head1 DESCRIPTION

C<CSM.pm> CSM basically creates a tunnel by managing network resources and state in real time for applications. Because the transport
          layer is constantly building and tearing down connections in response to application requests/responses, enforcing flow-control,
          and dynamically adjusting to network conditions, a good litmus test is validating the performance of the applications that are
          being transported across the session.

          The CSM.pm library provides API for:
          1. Searching for errored, and/or troubled sessions
          2. Verifying network cluster lif connectivity using 'cluster-ping'.
          3. Verifying the CSM tunnel linking striped volumes across the cluster to all msid(s) using 'csmping'.
          4. Verifying network cluster port health.
          5. Gathering cluster-wide error counters using variances on the following commands:
             a. statistics show -category {striped|csm}
             b. debug csm session show
             c. debug csm counters show -category {nulltx|session|rc|csm|cache}
          6. Provides a low, medium, and high level impact CSM Health Summary which increases the granularity, and scope of potential CSM issues.

=head1 FUNCTIONS

=cut

package CSM;

#use statements
use strict;
use Data::Dumper;
use Server;
use Tharn qw($Log);
use Options;
use Params::Common;
use Params::Validate qw(:all);
use Carp;

#Global VARS

=pod

=head2 new

C<new>

B<Description>
Creates, and initializes the CSM object which collects data for the primary litmus test
in determining CSM health using the 'debug csm session show' command. During "low", and
"medium" levels for CSM Health Summary, the state of these sessions will determine how 
deep the inspection will drill for errors. 

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Blessed hash structure as a CSM object

=cut

sub new {
    my $class = shift;
    my @sessions;
    my %params = validate(
    @_,
    {
       server => { type  => HASHREF },
    }
    );

    my $Server = $params{server};
    my $self = {_sessions => undef};
    bless $self,$class;

    $self->{_sessions} = $self->initialize(server=>$Server);

    return $self;
}

=pod

=head2 initialize

C<initialize>

B<Description>
Used to initialize the CSM object which collects data for the primary litmus test
in determining CSM health using the 'debug csm session show' command. The scope of
the statistics is filtered via the @SESSION_FIELDS array.

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Cluster CSM sessions stored within an array-of-hash.

=cut


sub initialize {
my @SESSION_FIELDS = qw(node uniquifier state type
                        initiator-string target-string
                        initiator target req-rcvd resp-sent
                        resp-rcvd req-sent req-pending
                        rsp-pending requests timedout-resp
                        timeout
                       );
    my $self = shift;
    my @session_aoh;
    my @output;
    my %params = validate(    
    @_,
    {
       server => { type => HASHREF },
    }
    );
    my $Server = $params{server};
    my @Nodes = $Server->servers();

    eval {        
        @output = $Server->debug_csm_session_show();
    };
    
    if ($@) {
        $Log->error("The 'debug csm session show' command failed due to an unknown error: $@");
    }
    foreach my $rec (@output) {
        my %session_fields;
        foreach my $field (@SESSION_FIELDS) {
            $session_fields{$field} = $rec->{$field};
        }
        push @session_aoh, {%session_fields};
    }
    return \@session_aoh;
}

=pod

=head2 csm_health_summary

C<csm_health_summary>

B<Description>
Method used to gather and aggregate cluster health records into a summarized 
report for determining the overall health of CSM. The intensity, and impact
can be controlled by a user param-defined level of low, medium, or high. 
Each error instance found will generate a THARN WARN result.

B<Input>

=over 4

=item * B<server> => $Server_obj

=item * B<level>

=over 4

=item * B<low> (default)
        If any CSM sessions are not "ESTABLISHED", cluster connectivity checks
        are performed, and errored session counters are returned.

=item * B<medium>
        If any CSM sessions are not "ESTABLISHED", OR troubled connections exist,
        cluster connectivity checks are performed, and errored counters are returned
        for session, csm, and lif categories.

=item * B<high>
        Checks for CSM sessions not in "ESTABLISHED" state.
        Checks for troubled connections.
        Performs cluster connectivity checks.
        Performs striped volume connectivity checks.
        Returns errored counters for: 
            1. csm
            2. session
            3. striped
            4. nulltx
            5. cache
            6. lif
            7. rc
        
=back

=back

B<Output>
Returns all cluster-wide error status, and counters in a hash of array.

Example output:
CSM Health Summary: $VAR1 = {
'rc_error_counters' => [
  'WARNING: RC Errors from command \'debug csm counters show -category rc -node cqa-6070-13 -name BindSentOpAbort\' = 315',
  'WARNING: RC Errors from command \'debug csm counters show -category rc -node cqa-perf-1 -name BindSentOpAbort\' = 354',
  'WARNING: RC Errors from command \'debug csm counters show -category rc -node cqa-perf-1 -name LoSndAbort\' = 48',
  'WARNING: RC Errors from command \'debug csm counters show -category rc -node cqa-perf-2 -name BindSentOpAbort\' = 72',
  'WARNING: RC Errors from command \'debug csm counters show -category rc -node cqa-perf-2 -name LoPassiveOpenNoMem\' = 2',
  'WARNING: RC Errors from command \'debug csm counters show -category rc -node cqa-perf-2 -name LoSndAbort\' = 84'
 ],
'cache_error_counters' => [
  'WARNING: Cache Errors from command \'debug csm counters show -category cache -node cqa-6070-13 -name dsToBladeInvalidates\' = 2',
  'WARNING: Cache Errors from command \'debug csm counters show -category cache -node cqa-6070-13 -name vifToBladeInvalidates\' = 2',
  'WARNING: Cache Errors from command \'debug csm counters show -category cache -node cqa-perf-1 -name dsToBladeInvalidates\' = 2',
  'WARNING: Cache Errors from command \'debug csm counters show -category cache -node cqa-perf-1 -name vifToBladeInvalidates\' = 2',
  'WARNING: Cache Errors from command \'debug csm counters show -category cache -node cqa-perf-2 -name dsToBladeInvalidates\' = 1',
  'WARNING: Cache Errors from command \'debug csm counters show -category cache -node cqa-perf-2 -name vifToBladeInvalidates\' = 1'
 ],
'striped_error_counters' => [
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-6070-13 -counter client-csm-heartbeat-connect-failure\' = 266',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-6070-13 -counter client-csm-req-send-retried\' = 10',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-6070-13 -counter client-csm-req-send-failure\' = 21',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-6070-13 -counter client-csm-resp-failure\' = 75',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-6070-13 -counter server-heartbeat-stalled-wafl-status\' = 2431981',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-6070-14 -counter server-heartbeat-stalled-wafl-status\' = 1157151',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-1 -counter client-csm-heartbeat-connect-failure\' = 437',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-1 -counter client-csm-req-send-retried\' = 10',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-1 -counter client-csm-req-send-failure\' = 41',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-1 -counter client-csm-resp-failure\' = 100',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-1 -counter server-heartbeat-stalled-wafl-status\' = 2325788',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-2 -counter client-csm-heartbeat-connect-failure\' = 460',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-2 -counter client-csm-req-send-retried\' = 10',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-2 -counter client-csm-req-send-failure\' = 39',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-2 -counter client-csm-resp-failure\' = 81',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-2 -counter server-heartbeat-stalled-wafl-status\' = 2389427'
 ],
'cluster_errored_ports' => 0,
'cluster_connectivity_errors' => 0,
'session_error_counters' => [
   'WARNING: Session Errors from command \'debug csm counters show -category session -node cqa-6070-13 -name closefailed\' = 7',
   'WARNING: Session Errors from command \'debug csm counters show -category session -node cqa-perf-1 -name closefailed\' = 6',
   'WARNING: Session Errors from command \'debug csm counters show -category session -node cqa-perf-2 -name closefailed\' = 7'
 ],
'nulltx_error_counters' => [],
'csmping_striped_failures' => 0,
'lif_error_counters' => [],
'csm_error_counters' => [
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-6070-13 -name activeConnFailures\' = 72',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-6070-13 -name activeConnStartFailures\' = 361',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-6070-13 -name getSessionForDataSetFailRsp\' = 115',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-1 -name activeConnFailures\' = 72',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-1 -name activeConnStartFailures\' = 383',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-1 -name getSessionForDataSetFailRsp\' = 157',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-2 -name activeConnFailures\' = 72',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-2 -name activeConnStartFailures\' = 114',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-2 -name connReject_NoBlade\' = 84',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-2 -name getSessionForDataSetFailRsp\' = 127'
 ]
};

=cut

sub csm_health_summary {
    my $self = shift;
    my %summary;
    my %params = validate(
    @_,
    {
      server => { type => HASHREF },
       level => { type => SCALAR,
                  default => "low",
                },
    }
    );
    my $Server = $params{server};

    $Log->comment("We are performing a $params{level} impact CSM health summary.\n");

    my $badness_check = $self->get_bad_sessions();
    
    if ((defined(@$badness_check)) || ($params{level} eq "high") || ($params{level} eq "medium")) {
        $Log->comment("We will report CSM sessions that are not ESTABLISHED, and"
                      ."cluster network connectivity checks will be executed.\n"
                     );
        $summary{bad_sessions} = @$badness_check;
        $summary{cluster_connectivity_errors} = $self->get_cluster_connectivity_errors(server=>$Server);
        $summary{cluster_errored_ports} = $self->get_cluster_errored_ports(server=>$Server);
        $Log->comment("Gathering Session Errors:\n");
        $summary{session_error_counters} = $self->get_session_error_counters(server=>$Server);

        if ($params{level} ne "low") {
            $summary{troubled_sessions} = $self->get_troubled_sessions();
            $Log->comment("Gathering CSM Errors:\n");    
            $summary{csm_error_counters} = $self->get_csm_error_counters(server=>$Server);
            $Log->comment("Gathering LIF Errors:\n");
            $summary{lif_error_counters} = $self->get_lif_error_counters(server=>$Server);
        }
        
        if ($params{level} eq "high") {
            $Log->comment("Csmping Striped Volumes:\n");
            $summary{csmping_striped_failures} = $self->csmping_striped_vols(server=>$Server);
            $Log->comment("Gathering Striped Errors:\n");
            $summary{striped_error_counters} = $self->get_striped_error_counters(server=>$Server);
            $Log->comment("Gathering RC Errors:\n");
            $summary{rc_error_counters} = $self->get_rc_error_counters(server=>$Server);
            $Log->comment("Gathering Cache Errors:\n");
            $summary{cache_error_counters} = $self->get_cache_error_counters(server=>$Server);
            $Log->comment("Gathering Nulltx Errors:\n");
            $summary{nulltx_error_counters} = $self->get_nulltx_error_counters(server=>$Server);
        }
    } else {
        $Log->comment("Passed $params{level} CSM Health Summary!");
        $summary{passed} = "Passed";   
    }
    return \%summary;
}
    
=pod

=head2 csmping_striped_vols

C<csmping_striped_vols>

B<Description>
Used to verify CSM connectivity to striped volumes across the cluster using the  
'csmping' utility for acro style tests.
Each error instance found will generate a THARN WARN result.

coraltest csmping -a -m 2147484747
Will check connectivity between localhost and all constituent volumes of the coral set 2147484747
contacting localhost...
sending csmping from localhost to owner of volume 1625...: Reachable
sending csmping from localhost to owner of volume 1626...: Reachable
sending csmping from localhost to owner of volume 1627...: Reachable
sending csmping from localhost to owner of volume 1628...: Reachable
sending csmping from localhost to owner of volume 1629...: Reachable
sending csmping from localhost to owner of volume 1630...: Reachable
sending csmping from localhost to owner of volume 1631...: Reachable
sending csmping from localhost to owner of volume 1632...: Reachable
8 Volumes Reachable, 0 Volume unreachable in this coralset from localhost

B<Input>

=over 4

=item * B<server> => $Server_obj

=item * B<ping_all_vols> => 0|1
        If bool turned on, the csmping binary will ping all striped volumes from each
        node. If bool turned off, it will choose a random volume to send a csmping.
        The default is off. 
=back

B<Output>
Array of striped volumes found to be unreachable.

=cut

sub csmping_striped_vols {
    my $self = shift;
    my @errored_stripe_vols;
    my %params = validate(
    @_,
    {
       server => { type => HASHREF },
       ping_all_vols => { type => BOOLEAN,
                          default => undef,
                        }
    }
    );
    my $Server = $params{server};
    my @Nodes = $Server->servers();
    my @striped_vols;

    eval{
        @striped_vols = $Server->volume_show('volume-style' => "striped",
                                             'state' => "online",
                                             'vsroot' => "false"
                                            );
    };

    if ($@) {
        if ($@ =~ /There are no entries matching your query/) {
            $Log->error("There were no striped volumes found upon which to execute the csmping test");
        } else {
            $Log->error("The 'volume show' command for striped volumes failed due to an unknown error: $@");
        }
        return;
    }

    if (!$params{ping_all_vols}) {
        my $sum_vol = int(@striped_vols);
        my $index = int(rand($sum_vol));
        my $volume = $striped_vols[$index];
        undef(@striped_vols); 
        push @striped_vols, {%$volume};
    }
  
    foreach my $node (@Nodes) {
        $Server->set_execution_server(-Server => $node);
        foreach my $vol (@striped_vols) {
            my $csm_ping_result = $Server->shell(cmd=>"coraltest csmping -a -m $vol->{msid}");
            $csm_ping_result =~ /(\d+)\sVolume\sunreachable/;
            if ($1 > 0) {
                push @errored_stripe_vols, "WARNING: $vol->{vserver}:$vol->{volume} failed coraltest csmping on $node: $&";
            }
        }
    }
    if (@errored_stripe_vols) {
        foreach (@errored_stripe_vols) {
            $Log->warn($_);
        }
        return \@errored_stripe_vols;
    }
}

=pod

=head2 get_cluster_connectivity_errors

C<get_cluster_connectivity_errors>

B<Description>
Used to verify whether cluster connectivity issues could be accountable for CSM problems by using the 'cluster ping-cluster'
command. 
Each error instance found will generate a THARN WARN result.

Example command output:
Host is cqa-6070-13
Getting information from ngsh
Local = 192.168.20.140 192.168.21.140 192.168.20.141 192.168.21.141
Remote = 192.168.20.176 192.168.21.176 192.168.20.178 192.168.22.178 192.168.20.142 192.168.21.142 192.168.20.143 192.168.21.143
Ping status:
32 paths up, 0 paths down at 1500 size
32 paths up, 0 paths down at 4500 size
32 paths up, 0 paths down at 9000 size
RPC status:
8 paths up, 0 paths down (tcp check)
8 paths up, 0 paths down (udp check)

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Array of path failures from 'cluster ping-cluster' command.

=cut

sub get_cluster_connectivity_errors {
    my $self = shift;
    my @cluster_connectivity_errors;
    my %params = validate(
    @_,
    {
       server => { type => HASHREF },
    }
    );
    my $Server = $params{server};
    my @Nodes = $Server->servers();

    my $Pings;
    foreach my $ping_from_node (@Nodes) {
        eval{
            $Pings = $Server->cluster_ping_cluster(node=>$ping_from_node,
                                                   'site-list' => "true",
                                                   );
        };
        if ($@) {
            $Log->error("The 'cluster ping cluster' command failed due to an unknown error: $@");
        }
        # Validate Ping Results for paths down for cluster ports 
        while ($Pings =~ /\s(\d+)\spaths\sdown\sat\s(\d+)\ssize/g) {
            if ($1 > 0) {
                push @cluster_connectivity_errors, "WARNING: There were $& packets while performing a cluster ping from $ping_from_node.";
            }
        }
        # Validate Ping Results for paths down tcp|udp check
        while ($Pings =~ /\s(\d+)\spaths\sdown\s\({tcp|udp}\scheck\)/g) {
            if ($1 > 0) {
                push @cluster_connectivity_errors, "WARNING: $& while performing a cluster ping from $ping_from_node.";
            }
        }
    }
    if (@cluster_connectivity_errors) {
        foreach (@cluster_connectivity_errors) {
            $Log->warn($_);
        }
        return \@cluster_connectivity_errors;
    }
}

=pod

=head2 get_cluster_errored_ports

C<get_cluster_errored_ports>

B<Description>
Used to verify network cluster ports are "up", and configured the same throughout
the cluster. 
These can cause an environment that affects the performance, and health of CSM.
Each error instance found will generate a THARN WARN result.

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Array of warning messages pinpointing the port attribute(s) in error, or misconfigured.

=cut


sub get_cluster_errored_ports {
    my $self = shift;
    my @cluster_errored_ports;
    my %params = validate(
    @_,
    {
       server => { type => HASHREF },
    }
    );
    my $Server = $params{server};
    my @Nodes = $Server->servers();
    my @Ports;

    eval{
        @Ports = $Server->network_port_show(role => "cluster");
    };
    if ($@) {
        $Log->error("The 'network port show' command failed due to an unknown error: $@");
    }

    my $compare_a = shift(@Ports);
    foreach my $compare_b (@Ports) {
        if ($compare_a->{mtu} != $compare_b->{mtu}) {
            push @cluster_errored_ports, "ERROR: The MTU setting on cluster port $compare_b->{port}, on $compare_b->{node}"
                                         ."does not match $compare_a->{mtu} which was derived from $compare_a->{port} on"
                                         ."$compare_a->{node}.\n"
        }
    }

    my @ports_down = grep($_->{link} != "up", @Ports);
    foreach (@ports_down) {
        push @cluster_errored_ports, "ERROR: Port $_->{port} on $_->{node} link is in a $_->{link} state.\n";
    }

    if (@cluster_errored_ports) {
        foreach (@cluster_errored_ports) {
            $Log->error($_);
        }
        return \@cluster_errored_ports;
    }
}

=pod

=head2 get_striped_error_counters

C<get_striped_error_counters>

B<Description>
Used to derive errors from specific striped counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Array of warning messages pinpointing the non-zero striped error counter.

=cut

sub get_striped_error_counters {
    my @STRIPED_ERROR_COUNTERS = qw(server-csm-req-awakened server-csm-req-stalled server-csm-req-discarded server-csm-req-unmarshal-error
                                    client-csm-resp-failure client-csm-resp-unmarshal-error client-csm-req-send-failure client-csm-req-send-retried
                                    client-csm-heartbeat-connect-gaveup client-csm-heartbeat-connect-failure client-csm-multicast-connect-failure
                                    client-csm-multicast-connect-gaveup server-heartbeat-stalled-wafl-status server-heartbeat-resumed-wafl-status
                                    server-heartbeat-stalled-adminlock server-heartbeat-resumed-adminlock server-heartbeat-resumed-offline
                                    lockstate-cache-resets lockstate-cache-failures lockstate-invalidated-pulls f-to-m-invalidated-in-flight
                                    f-to-m-need-ticket f-to-m-inode-invalid f-to-m-inode-missing d-to-f-invalidated-in-flight d-to-f-need-ticket
                                    d-to-f-inode-invalid d-to-f-inode-missing xstripe-write-avoid-tkt-enospc xstripe-write-fail-use-reserve
                                    xstripe-write-use-reserve extending-write-avoid-tkt-enospc extending-write-fail-use-reserve
                                    extending-write-use-reserve kinetic-token-revoke kinetic-token-discard xwrite-hit-zombie xwrite-failed
                                    xwrite-suspended csm-resp-fail-gbuf csm-send-fail-gbuf
                                   );
    my $self = shift;
    my @errored_counters;
    my %params = validate(
    @_,
    {
       server => { type => HASHREF },
    }
    );
    my $Server = $params{server};
    my @Nodes = $Server->servers();
    my @striped_stats_error;
    my $counter_string = join(",",@STRIPED_ERROR_COUNTERS);
    foreach my $node (@Nodes) {
            eval{
                @striped_stats_error = $Server->statistics_show(
                                           category  => "striped",
                                           object  => "striped-statistics",
                                           node => $node,
                                           counter => $counter_string,
                                          );
            };
            if ($@) {
                if ($@ =~ /There are no entries matching your query/) {
                    $Log->error("Requested striped statistics do not exist. High probability: No striped volumes provisioned in cluster.");
                } else {
                    $Log->error("The 'statistics show' command failed for an unknown error: $@");
                }
                return;
            } else {
                grep {if ($_->{value} > 0) {
                         $Log->warn("WARNING: Striped Errors from command 'statistics show -category striped -object striped-statistics -node $_->{node} -counter $_->{counter}' = $_->{value}");
                         push @errored_counters, "WARNING: Striped Errors from command 'statistics show -category striped -object striped-statistics -node $_->{node} -counter $_->{counter}' = $_->{value}";
                     }
                } @striped_stats_error;
            }
    }
    return \@errored_counters;
}

=pod

=head2 get_lif_error_counters

C<get_lif_error_counters>

B<Description>
Used to derive errors from specific lif counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Array of warning messages pinpointing the non-zero lif error counter.

=cut

sub get_lif_error_counters {
    my @LIF_ERROR_COUNTERS = qw(recv-errors sent-errors);
    my $self = shift;
    my @errored_counters;
    my %params = validate(
    @_,
    {
       server => { type => HASHREF },
    }
    );
    my $Server = $params{server};
    my @Nodes = $Server->servers();
    my @lifs;
    my @lif_stats_error;
    my $counter_string = join(",",@LIF_ERROR_COUNTERS);
    foreach my $node (@Nodes) {
        @lifs = $Server->network_interface_show(
                    server => $node,
                    role => "cluster",
                   );
        foreach my $lif (@lifs) {
            eval{
                @lif_stats_error = $Server->statistics_show(
                                           category => "lif",
                                           object  => $lif->{server}.":".$lif->{lif},
                                           counter => $counter_string,
                                          );
            };
            if ($@) {
                $Log->error("The 'statistics show' command failed due to an unknown error: $@");
            } else {
                grep {if ($_->{value} > 0) {
                         $Log->warn("WARNING: LIF Errors from command 'statistics show -category lif -object $_->{object} -counter $_->{counter}' = $_->{value}");
                         push @errored_counters, "WARNING: LIF Errors from command 'statistics show -category lif -object $_->{object} -counter $_->{counter}' = $_->{value}";
                         $Log->comment("Dumping netstat to logs:\n");
                         $Server->dblade(cmd=>'priv set -q diag;netstat -p udp');
                         $Server->dblade(cmd=>'priv set -q diag;netstat -i');
                         $Server->dblade(cmd=>'priv set -q diag;netstat -s');
                     }
                } @lif_stats_error;
            }
        }
    }
    return \@errored_counters;
}

=pod

=head2 get_csm_error_counters

C<get_csm_error_counters>

B<Description>
Used to derive errors from specific csm counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Array of warning messages pinpointing the non-zero csm error counter.

=cut

sub get_csm_error_counters {
    my @CSM_ERROR_COUNTERS = qw(activeConnFailures activeConnStartFailures connReject_BindToSession connReject_BrokenSession
                                connReject_Mallocation connReject_NoBlade connReject_ObsoleteConn connReject_SessionInit
                                connReject_SharedSessionExists connReject_UniquifierClash getSessionForDataSetFailRsp 
                                getSessionForVifFailRsp registerRcListenAddressFailed
                               );
    my $self = shift;
    my @errored_counters;
    my %params = validate(
    @_,
    {
       server => { type => HASHREF },
    }
    );
    my $Server = $params{server};
    my @Nodes = $Server->servers();
    my @csm_counter_error;
    my $counter_string = join(",",@CSM_ERROR_COUNTERS);
    foreach my $node (@Nodes) {
            eval{
                @csm_counter_error = $Server->debug_csm_counters_show(
                                           category => "csm",
                                           node => $node,
                                           name => $counter_string,
                                          );
            };
            if ($@) {
                $Log->error("The 'debug csm counters show' command failed for an unknown error: $@");
            } else {
                grep {if ($_->{value} > 0) {
                         $Log->warn("WARNING: CSM Errors from command 'debug csm counters show -category csm -node $_->{node} -name $_->{name}' = $_->{value}");
                         push @errored_counters, "WARNING: CSM Errors from command 'debug csm counters show -category csm -node $_->{node} -name $_->{name}' = $_->{value}";
                     }
                } @csm_counter_error;
            }
    }
    return \@errored_counters;
}

=pod

=head2 get_rc_error_counters

C<get_rc_error_counters>

B<Description>
Used to derive errors from specific RC counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Array of warning messages pinpointing the non-zero RC error counter.

=cut

sub get_rc_error_counters {
    my @RC_ERROR_COUNTERS = qw(AcceptWaitLoAborted AcceptWaitLoAborted2 BadBuf BadDataVerifierAck BadHdrVerifier BadLen BadMagic
                               BadVersion BindRcvdAckBad BindRcvdOpAbort BindRcvdOpAbort2 BindRcvdSeqBad BindSentAckBad BindSentAckBadAbort
                               BindSentInvalid BindSentOpAbort BuildPktNoBuf ConnUserWaitDrop ConnUserWaitLoAborted ConnUserWaitLoAborted2
                               FastTimerClosedBusyBad ListenOpBad LoActiveOpenRspNoMem LoPassiveOpenNoMem LoRcvRcInputNull LoRemoteCloseAbort
                               LoRemoteCloseAbort2 LoSndAbort LocalCloseAbort LocalCloseBadState SyncInvalid SyncOpAbort SyncOpAbort2
                               SyncRcvBadDataVerifier SyncRcvBadDataVerifierBind SyncRcvBadDataVerifierBrsp SyncRcvGbufTypeMismatch
                               SyncRcvStrangeBuf Unknown
                              );
    my $self = shift;
    my @errored_counters;
    my %params = validate(
    @_,
    {
       server => { type => HASHREF },
    }
    );
    my $Server = $params{server};
    my @Nodes = $Server->servers();
    my @rc_counter_error;
    my $counter_string = join(",",@RC_ERROR_COUNTERS);
    foreach my $node (@Nodes) {
            eval{
                @rc_counter_error = $Server->debug_csm_counters_show(
                                           category => "rc",
                                           node => $node,
                                           name => $counter_string,
                                          );
            };
            if ($@) {
                $Log->error("The 'debug csm counters show' command failed for an unknown error: $@");
            } else {
                grep {if ($_->{value} > 0) {
                         $Log->warn("WARNING: RC Errors from command 'debug csm counters show -category rc -node $_->{node} -name $_->{name}' = $_->{value}");
                         push @errored_counters, "WARNING: RC Errors from command 'debug csm counters show -category rc -node $_->{node} -name $_->{name}' = $_->{value}";
                     }
                } @rc_counter_error;
            }
    }
    return \@errored_counters;
}

=pod

=head2 get_session_error_counters

C<get_session_error_counters>

B<Description>
Used to derive errors from specific session counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Array of warning messages pinpointing the non-zero session error counter.

=cut

sub get_session_error_counters {
    my @SESSION_ERROR_COUNTERS = qw(badchan badchannel badinitop badinterface badprototag badseq badstateforinit badstaterequest
                                    bindfailed closefailed createfailed createrespfailed marshalfailure negoifacefailed negosessfailed
                                    noreqforresp spinnp_abort_operation_error spinnp_bind_connection_error spinnp_close_session_error
                                    spinnp_create_session_error spinnp_get_session_info_error spinnp_gss_auth_begin_error
                                    spinnp_gss_auth_complete_error spinnp_gss_continue_init_error spinnp_gss_destroy_error
                                    spinnp_gss_init_error spinnp_negotiate_interface_error spinnp_negotiate_resp_store_error
                                    spinnp_negotiate_session_error spinnp_null_proc_error spinnp_recovery_complete_error
                                    spinnp_replay_response_error spinnp_response_ids_error spinnp_set_seq_window_error
                                    spinnp_unbind_connection_error unbindfailed unmarshalfailure
                                   );

    my $self = shift;
    my @errored_counters;
    my %params = validate(
    @_,
    {
       server => { type => HASHREF },
    }
    );
    my $Server = $params{server};
    my @Nodes = $Server->servers();
    my @session_counter_error;
    my $counter_string = join(",",@SESSION_ERROR_COUNTERS);
    foreach my $node (@Nodes) {
            eval{
                @session_counter_error = $Server->debug_csm_counters_show(
                                           category => "session",
                                           node => $node,
                                           name => $counter_string,
                                          );
            };
            if ($@) {
                $Log->error("The 'debug csm counters show' command failed for an unknown error: $@");
            } else {
                grep {if ($_->{value} > 0) {
                         $Log->warn("WARNING: Session Errors from command 'debug csm counters show -category session -node $_->{node} -name $_->{name}' = $_->{value}");
                         push @errored_counters, "WARNING: Session Errors from command 'debug csm counters show -category session -node $_->{node} -name $_->{name}' = $_->{value}";
                     }
                } @session_counter_error;
            }
    }
    return \@errored_counters;
}

=pod

=head2 get_nulltx_error_counters

C<get_nulltx_error_counters>

B<Description>
Used to derive errors from nulltx striped counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Array of warning messages pinpointing the non-zero nulltx error counter.

=cut

sub get_nulltx_error_counters {
    my @NULLTX_ERROR_COUNTERS = qw(noMemActiveConn noMemPassiveConn);
    my $self = shift;
    my @errored_counters;
    my %params = validate(
    @_,
    {
       server => { type => HASHREF },
    }
    );
    my $Server = $params{server};
    my @Nodes = $Server->servers();
    my @nulltx_counter_error;
    my $counter_string = join(",",@NULLTX_ERROR_COUNTERS);
    foreach my $node (@Nodes) {
            eval{
                @nulltx_counter_error = $Server->debug_csm_counters_show(
                                           category => "nulltx",
                                           node => $node,
                                           name => $counter_string,
                                          );
            };
            if ($@) {
                $Log->error("The 'debug csm counters show' command failed for an unknown error: $@");
            } else {
                grep {if ($_->{value} > 0) {
                         $Log->warn("WARNING: Nulltx Errors from command 'debug csm counters show -category nulltx -node $_->{node} -name $_->{name}' = $_->{value}");
                         push @errored_counters, "WARNING: Nulltx Errors from command 'debug csm counters show -category nulltx -node $_->{node} -name $_->{name}' = $_->{value}";
                     }
                } @nulltx_counter_error;
            }
    }
    return \@errored_counters;
}

=pod

=head2 get_cache_error_counters

C<get_cache_error_counters>

B<Description>
Used to derive errors from specific cache counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Array of warning messages pinpointing the non-zero cache error counter.

=cut

sub get_cache_error_counters {
    my @CACHE_ERROR_COUNTERS = qw(dsToBladeInvalidates vifToBladeInvalidates);
    my $self = shift;
    my @errored_counters;
    my %params = validate(
    @_,
    {
       server => { type => HASHREF },
    }
    );
    my $Server = $params{server};
    my @Nodes = $Server->servers();
    my @cache_counter_error;
    my $counter_string = join(",",@CACHE_ERROR_COUNTERS);
    foreach my $node (@Nodes) {
            eval{
                @cache_counter_error = $Server->debug_csm_counters_show(
                                           category => "cache",
                                           node => $node,
                                           name => $counter_string,
                                          );
            };
            if ($@) {
                $Log->error("The 'debug csm counters show' command failed for an unknown error: $@");
            } else {
                grep {if ($_->{value} > 0) {
                         $Log->warn("WARNING: Cache Errors from command 'debug csm counters show -category cache -node $_->{node} -name $_->{name}' = $_->{value}");
                         push @errored_counters, "WARNING: Cache Errors from command 'debug csm counters show -category cache -node $_->{node} -name $_->{name}' = $_->{value}";
                     }
                } @cache_counter_error; 
            }
    }
    return \@errored_counters;
}

=pod

=head2 get_bad_sessions

C<get_bad_sessions>

B<Description>
Used to find any CSM session that is NOT in an "ESTABLISHED" state.
Each error instance found will generate a THARN WARN result.
B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Array of warning messages reporting any CSM sessions NOT in an "ESTABLISHED" state. Also details
the blade interface endpoints which make up that specific CSM session.

=cut

sub get_bad_sessions {
    my $self = shift;
    my @bad_sessions;
    my @not_established = grep($_->{state} != "ESTABLISHED",@{$self->{_sessions}});
    if (@not_established) {
        foreach my $errored (@not_established) {
            my $initiator = $errored->{'initiator-string'};
            my $initiator_uuid = $errored->{'initiator'};
            my $target_uuid = $errored->{'target'};
            my $target = $errored->{'target-string'};
            my $state = $errored->{state};
            push @bad_sessions, "WARNING: Session from $initiator:$initiator_uuid to $target:$target_uuid is in a $state state!";
        }
        if (@bad_sessions) {
            foreach (@bad_sessions) {
                $Log->warn($_);
            }
            return \@bad_sessions;
        }
    } else { 
        return undef; 
    }
}

=pod

=head2 get_troubled_sessions

C<get_troubled_sessions>

B<Description>
Used to find CSM related stats which might reveal intermittent, or underlying issues in the cluster.
Each error instance found will generate a THARN WARN result.

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Array of warning messages reporting CSM sessions performance dispairity. Also details
the blade interface endpoints which make up that specific CSM session.

=cut

sub get_troubled_sessions {
    my $self = shift;
    my @troubled_sessions;
    my @outstanding_resp = grep(($_->{'req-rcvd'} != $_->{'resp-sent'}),@{$self->{_sessions}});
    if (@outstanding_resp) {
        foreach my $troubled (@outstanding_resp) {
            my $req_rcvd = $troubled->{'req-rcvd'};
            my $resp_sent = $troubled->{'resp-sent'};
            my $target = $troubled->{'target-string'};
            my $target_uuid = $troubled->{'target'};
            my $initiator = $troubled->{'initiator-string'};
            my $initiator_uuid = $troubled->{'initiator'};
            push @troubled_sessions, "WARNING: Requests received: $req_rcvd do not equal Responses sent: $resp_sent.\n"
                                      ."$initiator:$initiator_uuid is having problems communicating with $target:$target_uuid.\n";
        }
    }

    my @outstanding_req = grep(($_->{'req-sent'} != $_->{'resp-rcvd'}),@{$self->{_sessions}});
    if (@outstanding_req) {

        foreach my $troubled (@outstanding_req) {
            my $resp_rcvd = $troubled->{'resp-rcvd'};
            my $req_sent = $troubled->{'req-sent'};
            my $initiator = $troubled->{'initiator-string'};
            my $initiator_uuid = $troubled->{'initiator'};
            my $target = $troubled->{'target-string'};
            my $target_uuid = $troubled->{'target'};
            push @troubled_sessions, "WARNING: Responses received: $resp_rcvd does not equal Requests sent: $req_sent.\n"
                                     ."$initiator:$initiator_uuid is having problems communicating with $target:$target_uuid.\n";
        }
    }
    if (@troubled_sessions) {
        foreach (@troubled_sessions) {
            $Log->warn($_);
        }
        return \@troubled_sessions;
    } else {
        return undef;
    }
}

1;

=pod

=head1 LIMITATIONS
The initial offering of the CSM.pm library utilizes "current" values, and is meant mainly as a post-processing tool. Future versions
will attempt to take samples and report delta values in-paralell.

=cut

