# $Id: //depot/prod/test/Rsierranevada/lib/CSetup/Engine.pm#1 $

## @summary The engine which does setup and cleanup on 
##hosts(Filers and Clients)
## @author dl-setupcleanup-libs@netapp.com
## @pod here

package CSetup::Engine;
use strict;
use warnings;
use Tharn qw(script_result logresult config_result logcomment);
use Params::Validate qw(validate SCALAR HASHREF UNDEF OBJECT);
BEGIN
{
    Params::Validate::validation_options(
                                         on_fail => sub {
                                             my @msgs = @_; chomp(@msgs);
                                             die script_result([@msgs]);
                                         }
                                        );
}


require Exporter;
our @ISA=qw(Exporter);
our @EXPORT_OK=qw(host_setup);

=head1 NAME

Engine - The engine which does setup and cleanup on hosts(Filers and Clients).

=head1 SYNOPSIS

    use CSetup::Engine qw(host_setup)
    use CSetup::SnapShot;
    my $snap_impl = SnapShot->new(system => tada-u1);
    #The states hash defines all the states ,involved in a setup process,
    #the hash  has  Base,Application and Plan as the states 
    #involved in the setup process
    my $states = {
                  Base =>        {
                                 parent => undef,
                                 install_action => $base_callback,
                                 snapshot_action => create,
                                 },
                  Application => {
                                 parent => Base,
                                 install_action => $application_callback,
                                 snapshot_action => create,
                                 },
                  Plan =>        {
                                 parent => Application,
                                 install_action => $plan_callback,
                                 snapshot_action => create,
                                 }
                  };
    host_setup(
               states=>$states,
               revert_to=>Base,
               to_state=>Plan,
               snap_object=>$snap_impl
              );

=head1 DESCRIPTION

The calling script has to define the various states involved in the setup process for a 
host in the States hash.States hash has state-name as the key and an anonymous hash as the value.
The anonymous hash contains parent state for the current state ,install actions(code reference)
to reach to the current state and  snapshot action to be performed on reaching the state.
The install action can also be set to undef for a state,in that case parent 
and current state are the same

The snapshot action can take any one of the three values:

=over

create: create a new snapshot of this state

delete: delete any already existing snapshot for this state

readonly: do not create a new snapshot and also do not delete any existing snapshot for this state
 
=back

For Example,
$states = {
           Base  => {
               parent => undef,
               install_action => $init_from_scratch,
               snapshot_action => create,
           },
           Foo  => {
               parent => Base,
               install_action => $init_foo_from_base,
               snapshot_action => "readonly",
           },
           Bar  => {
               parent => Foo,
               install_action => $init_bar_from_Foo,
               snapshot_action => delete,
           }
       };

The user, calling host_setup will pass the $states defined,$revert_to ,$to_state and $snap_object

host_setup(
           state=>$states
           revert_to=>"Base",
           to_state=>"Bar",
           snap_object=>$snap_impl_class
          );

The host_setup subroutine will first revert the host to the revert_to state,loop through the states 
hash defined, executes the code references defined on the hash and sets the host to the to_state


=head1 host_setup


host_setup(
          states=>$states
          revert_to=>"Base",
          to_state=>"Bar",
          snap_object=>$snap_impl_class
          );

=over

=item states

(Required) : Hash reference which contains all the states defined

=item to_state

(Required) : The name of the state to reach

=item revert_to

(optional) : Defaults to to_state, revert to this state before trying to reach to_state,revert_to can 
also be set to "reinit" means no revert at all, start from the scratch,which is the highest state defined
 on the hash.

=item snap_object

(Required) : An object instance of CSetup::Snapshot, which will provide the methods to check what snapshots
 are available, create a snapshot, or revert to a snapshot

=back

=cut

sub host_setup {
    my (%opts) = validate(@_, {
                               states => { type => HASHREF },
                               revert_to => { type => SCALAR|UNDEF, default=>undef },
                               to_state => { type => SCALAR },
                               snap_object => { type => OBJECT },
                               exe_client=>{ type => SCALAR}
                              }
                          );
    my $states=$opts{'states'};
    my $revert_to=$opts{'revert_to'};
    my $to_state=$opts{'to_state'};
    my $snap_impl=$opts{'snap_object'};
    my $exe_client=$opts{'exe_client'};
    if (!defined($revert_to)) {
        $revert_to = $to_state;
    } elsif ($revert_to =~ /^reinit$/i) {
            $revert_to = undef;
    } else {
        die "Unknown $revert_to state" if (!$states->{$revert_to});
    }
       
    # Walk up the tree, attempting to revert to an earlier state
    # and building a list of actions we will have to take once we
    # do revert to there.
       
    my $attempt_revert=0;
    my @active_states=();
    my $host = $snap_impl->{host};
    my $hostname= $host->hostname();
    for (my $state = $to_state; defined($state); $state=$states->{$state}->{parent} ) {
        if (defined($revert_to) && $state eq $revert_to) {
            $attempt_revert=1;
            if($hostname eq $exe_client){
                $attempt_revert=0;
                $states->{$state}->{snapshot_action}="readonly";
             }   
        }
        if ($attempt_revert) {
        eval {
             if ($snap_impl->is_available(state => $state)) {
                $snap_impl->revert_to(state => $state);
                last;
                }
             };
             if ($@) {
                 # Bail out in case of config error
                 if (ref($@) && $@->can("type") && $@->type() =~ /CONFIG/){
                     die config_result($@);       
                 }
                logcomment($@);
                $attempt_revert=0;
             }
       }
       unshift @active_states, $state;
    }
       
    foreach my $state (@active_states) {
    my $install_action = $states->{$state}->{install_action};
    $install_action->() if $install_action;
    my $snapshot_action = $states->{$state}->{snapshot_action};
    if ($snapshot_action eq "create") {
       $snap_impl->create(state => $state);
    } elsif ($snapshot_action eq "delete") {
      $snap_impl->delete(state => $state);
    } elsif ($snapshot_action eq "readonly") {
      # do nothing
    } else {
      die config_result("No such snapshot_action '$snapshot_action'");
    }
    }
    }

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = none

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    dl-setupcleanup-libs@netapp.com
    $Id: //depot/prod/test/Rsierranevada/lib/CSetup/Engine.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005,2009 Network Appliance Inc., All Rights Reserved.

=cut

1;
