# $Id: //depot/prod/test/Rsierranevada/lib/CSetup/Snapshot/Aggregate.pm#1 $

## @summary Aggregate snap create and aggregate snap restore, used for 7g/7-mode
## @author dl-setupcleanup-libs
## @status Shared
## @pod here

package CSetup::Snapshot::Aggregate;

use strict;
use warnings;

use Tharn qw($Log script_result host connect say param);
use Params::Validate qw(validate SCALAR);
use SayWrap qw(sayEOF);

BEGIN
{
    Params::Validate::validation_options(
                                         on_fail => sub {
                                             my @msgs = @_; chomp(@msgs);
                                             die script_result([@msgs]);
                                         },
                                        );
}

use CLI::FileIO qw(rdfile wrfile);
use Aggregate qw(aggr_root aggr_list aggr_exists aggr_disks aggr_purge);
use SnapCmds qw(snapshot_exists snapshot_delete snapshot_create snaprestore_volume);
use Hammer qw(ontap_hammer_abort);
use Admin::License qw(license);
use boot qw(boot);
use RaidInfo;
use CSetup::Utils qw(read_conf);

sub new {

    my ($proto) = shift @_;
    my (%args) = validate (@_, {
                 'host' => { isa => 'Hostrec' },
                }
            );
    my $class = ref($proto) || $proto;

    my $self = {
	host => $args{host},
    };

    #Boot the filer ,if its not booted
    boot($args{host},powercycle => 1);

    if(!$@) {
      #Disable the cluster 
      $Log->comment("Disabling cluster");
      my $conn = connect("#console");
      my $out = sayEOF($conn, "cf disable;");
      $Log->comment("$out");
    } else {
      $Log->comment("Unable to boot the filer");
    }
    bless($self,$class);
    return $self;
}

#
# Check if specified state has available snapshot(s) to revert to
#
sub is_available {

    my ($self) = shift @_;
    my (%args) = validate (@_, {
                 'state' => { type=>SCALAR },
                }
            );

    # connect to the host (filer in this case)
    my $host = $self->{host};
    my $conn = connect("#rsh");

    my $snapname = _snapname_for_state($args{state});

    # check for the existance of the data file
    # this file lives on the host(filer)
    # if file does not exist return false
    # read the data file /etc/csetup/state-$state
    my $datafile = _datfile_for_state($args{state});
    $Log->comment("Reading data file '$datafile'");
    my $data = rdfile($conn,$datafile);
    if ($@) {
        # return 0 if file is not found - die for other failures
        if ($@ =~ /No such file or directory/i) {
            $Log->comment("Restoration of state '$args{state}' ".
                        "not available because datafile '$datafile' ".
                        "is not present : $@");
            return 0;
        } else {
            die("Restoration of state '$args{state}' ".
                "not available because : $@");
        }
    }
    $Log->trace("Data is $data");
    # Return 0 if there is no data in file
    if (!$data) {
        $Log->comment("Restoration of state '$args{state}' ".
                      "not available because datafile '$datafile' ".
                      "is empty");
        return 0;
    }

    if (!&_check_desc($args{state})) {
        $Log->comment("Restoration of state '$args{state}' ".
                      "not available because Snapshot description ".
                      "didnt match");
        return 0;
    }

    # format of the data file double colon separated list of items
    # /etc/csetup-state-$state
    # <aggrname>::<aggr-attributes>
    # Aggr attributes will look like colon separated key=value pairs
    # eg
    # aggr1::disks=4a.d 4a.d:

    # for each aggregate in the file 
    foreach my $line (split(/\n/,$data)) {
        next if ($line !~ /::/); # skip the desc 
        my ($aggr, $attributes) = split (/::/,$line);
        $Log->debug("Aggr is $aggr, attributes are $attributes");

        # check if the aggregate exists 
        my $ret = aggr_exists($conn, $aggr);
        if (!$ret) {
            $Log->comment("Restoration of state '$args{state}' not available ".
                            "because aggregate '$aggr' is missing");
            return 0;
        }

        # check if a snapshot exists for it 
        $ret = snapshot_exists($conn,'-A '.$aggr,$snapname);
        if (!$ret) {
            $Log->comment("Restoration of state '$args{state}' ".
                        "not available because snapshot '$snapname'".
                        " for aggregate $aggr is missing");
            return 0;
        }

        # check the attributes of the aggregate
        foreach my $each_attr (split/:/,$attributes) {
            $Log->debug("Aggregate attribute is $each_attr");
            my ($attrkey, $value) = ($each_attr =~ /^(\S+)=(.*)$/);

            $Log->comment("Attr $attrkey Value $value");
            if ($attrkey =~ /disks/) {
                $Log->debug("Checking the disks attributes $value");
                # get the current list of disks from the system
                my @current_disks = aggr_disks($conn,$aggr) 
                    or die("Restoration of state '$args{state}' ".
                                "not available because getting ".
                                "list of aggr '$aggr' disks failed :$@");
                my @current_disk_list;
                for my $href ( @current_disks ) {
                    push @current_disk_list, $href->{name};
                }
                # check the disks attributes to see if they match
                my @disk_list = split(/\s/,$value);
                $Log->debug("Compare aggr $aggr disks @current_disk_list ".
                        "with disks from data file @disk_list"); 
                if (_compare_arrays(\@current_disk_list,\@disk_list)) {
                    $Log->debug("Current disks for aggr $aggr @current_disk_list".
                        "match the list of disks in datafile @disk_list");
                } else {
                    $Log->comment("Restoration of state '$args{state}' not available ".
                                "because aggregate '$aggr' has since moved to ".
                                "a different set of disks. (from ".
                                "'@disk_list' to '@current_disk_list')");
                    return 0;
                }
            }
            # XXX add checking for more attr here - what is missing ?
        }
    }

    return 1;
}


#
# create snapshots of all aggreates and create a new data file
#
sub create {

    my ($self) = shift @_;
    my (%args) = validate (@_, { 
                 'state' => { type=>SCALAR },
                }
            );

    # connect to the host (filer in this case)
    my $host = $self->{host};
    my $conn = connect("#rsh");

    my $datafile = _datfile_for_state($args{state});
    my $snapname = _snapname_for_state($args{state});
    my $data; # data to be written to the datafile
    my @aggr_list = aggr_list($conn); # get list of aggregates
    $data = &_desc_for_state($args{state});
    $data .= "\n";
    # get the attributes of all the aggrs that we are going to snapshot
    foreach my $aggr (@aggr_list) {
        my $line = $aggr. '::';
        # get the aggr attributes
        my @aggr_disks = aggr_disks($conn,$aggr) 
            or die("Creating state '$args{state}' failed since ".
                "getting the list of aggr '$aggr' disks failed : $@");
        my @disk_list;
        for my $href ( @aggr_disks ) {
            push @disk_list, $href->{name};
        }
        $line .= "disks=@disk_list";
        $data .= $line;
        $data .= "\n";
    }

    # Write a new copy of the data file
    $Log->trace("Writing data $data to file $datafile");
    wrfile($conn, $datafile, $data)
        or die("Creating state '$args{state}' failed : ".
                "Could not write $datafile : $@");

    # Create a snapshot on every aggregate (with the snapshot name that $state maps to).
    foreach my $aggr (@aggr_list) {
        $Log->trace("Creating snapshot $snapname for aggr $aggr");
        eval {
            snapshot_create($conn, "-A ".$aggr, $snapname);
        
        };
        if ($@) {
            # get rid of the data file since its not valid
            say($conn, "priv set -q advanced; rm $datafile"); # does not return anything to verify  
            my $data = rdfile($conn,$datafile);
            die("Creating state '$args{state}' failed since snapshot creation ".
                "failed. $@. Tried to remove invalid data file '$datafile' and ".
                "that failed too") if ($data);
            die("Creating state '$args{state}' failed since snapshot creation ".
                "failed : $@");
        }
    }

    return 1;
}

#
# Delete all the aggregate snapshots for specified state
#
sub delete {

    my ($self) = shift @_;
    my (%args) = validate (@_, { 
                 'state' => { type=>SCALAR },
                }
            );

    # connect to the host (filer in this case)
    my $host = $self->{host};
    my $conn = connect("#rsh");

    my $snapname = _snapname_for_state($args{state});

    # get list of aggregates
    my @aggr_list = aggr_list($conn);
    foreach my $aggr (@aggr_list) {
        # purge the snapshot if it exists
        my $ret = snapshot_exists($conn,'-A '.$aggr,$snapname);
        if (!$ret) {
            # no need to fail if it does not exist
            $Log->trace("Snapshot $snapname for aggregate $aggr does not exist");
        } else {
            $Log->trace("Deleting snapshot $snapname for aggr $aggr");
            snapshot_delete($conn, '-A '.$aggr, $snapname)
                or die("Deleting state '$args{state}' failed since delete ".
                        "of snapshot '$snapname' of aggr '$aggr' failed : $@");
        }
    }

    # remove the data file /etc/csetup-state-<state>
    my $datafile = _datfile_for_state($args{state});
    say($conn, "priv set -q advanced; rm $datafile"); # does not return anything to verify  
    my $data = rdfile($conn,$datafile);
    $Log->warn("Was unable to purge the datafile '$datafile' ".
                "while deleting state '$args{state}'") if ($data);

    return 1;
}


#
# revert the filer to the specified state
#
sub revert_to {

    my ($self) = shift @_;
    my (%args) = validate (@_, { 
                 'state' => { type=>SCALAR },
                }
            );


    # connect to the host (filer in this case)
    my $host = $self->{host};
    my $conn = connect("#rsh");

    my $snapname = _snapname_for_state($args{state});

    # restore the root aggregate first 
    # so we can catch any failures sooner
    my $root_aggr = aggr_root($conn) 
        or die("Reverting to state '$args{state}' failed since ".
                "discovering root volume failed : $@");
    $Log->comment("Restoring the root aggregate $root_aggr to snapshot $snapname");
    # snaprestore of root volume requires will reboot filer and uses a console connection
    # license snaprestore to restore root vol
    license($conn, 'snaprestore') 
        or die("Reverting to state '$args{state}' failed since licensing ".
                "of snaprestore failed :$@"); 
    _snaprestore_root_aggregate(filer=>$host, 
                                aggregate=>$root_aggr,
                                snapname => $snapname,
                                ) or die("Reverting to state '$args{state}' failed ".
                                        "since snaprestore of roor aggr '$root_aggr' ".
                                        "failed : $@");
    # license again since root vol got restored
    license($conn, 'snaprestore') 
        or die("Reverting to state '$args{state}' failed since licensing ".
                "of snaprestore failed :$@"); 

    # For each non-root aggregate, if it has a snapshot for this state, 
    # revert it to that state. If it does not, then destroy it.
    my @aggr_list = aggr_list($conn);
    foreach my $aggr (@aggr_list) {
        next if ($aggr eq $root_aggr); # skip root aggr since its done already

        # check if snapshot exists - do we do that or fail if it does not
        my $ret = snapshot_exists($conn,'-A '.$aggr,$snapname);
        if (!$ret) {
            # purge the aggregate if snapshot does not exist
            $Log->comment("Snapshot '$snapname' for aggregate '$aggr' ".
                        "does not exist. Purging aggr '$aggr'");
            aggr_purge($conn,$aggr) 
                or die("Reverting to state '$args{state}' failed ".
                        "since purging of aggr '$aggr' failed : $@");
        } else {
            $Log->comment("Restoring aggr '$aggr' to snapshot '$snapname'");
            snaprestore_volume($conn, '-A '.$aggr, $snapname) 
                or die("Reverting to state '$args{state} failed since ".
                        "snaprestore of aggr '$aggr' to snapshot ".
                        "'$snapname' failed : $@");
            # the snapshot used to restore the aggr still remains after the restore
        }
    }

    # Perform a number of cleanup steps which one would have expected to be 
    # cleaned up as part of "revert a whole filer to state", 
    # but which are not part of aggregate state:

    # Unfail failed disks if any
    # Get raid info from filer
    my $ri = new RaidInfo($conn) or
        die("Reverting to state '$args{state}' failed : ".
            " Unable to get raid info from filer");
    my %failed_disks =  $ri->failed_disks();
    my @failed = keys(%failed_disks);
    foreach  my $disk (@failed) {
        $Log->comment("Unfailing disk '$disk'");
        my (undef, undef, $out) = say($conn, "priv set -q advanced; disk unfail $disk");
        die("Reverting to state '$args{state}' failed since ".
                "unfailing failed disk '$disk' failed : $out")
            if ($out !~ /Unfailing disk |should now be spare/i);
        # dont wait for the unfail to finish
    }

    # Zero the spares if any (this command does not return anything)
    $Log->comment("Zeroing spares on filer");
    say($conn, "disk zero spares");
    # dont wait for the disks to get zeroed

    # Stop hammer (since hammer stays running across reboots)
    $Log->comment("Aborting all running hammers on filer");
    ontap_hammer_abort(fc=>$conn); # this does not return any success or failure

    return 1;
}

#
# local routine to create snapname given state
#
sub _datfile_for_state {
    my ($state) = @_;

    my $datafile = "/etc/csetup-state-$state";
    return $datafile;
}

#
# local routine to create snapname given state
#
sub _snapname_for_state {
    my ($state) = @_;
    
    my $snapname = '__csetup__'.$state;
    return $snapname;
}

#
# local routine to compare arrays
#
sub _compare_arrays {
    my ($first, $second) = @_;
    return 0 unless @$first == @$second;
    for (my $i = 0; $i < @$first; $i++) {
        return 0 if $first->[$i] ne $second->[$i];
    }
    return 1;
}

# 
# local routine to snap restore a root aggr
#
sub _snaprestore_root_aggregate {
    my (%args) = validate(@_, {
                               filer => { isa => 'Hostrec' },
                               aggregate => { type => SCALAR },
                               snapname => { type => SCALAR},
                               timeout => { regex => qr(^\d+$), default => 120 },
                               });

    my $conn = connect($args{filer}->id . "#console");
    
    # BURT-  407249: Workaround for product bug - 39955.
    # Correct way to check if snapshot was restored is to check 
    # for appropriate message. Since some of the builds dont display standard
    # message for snap restore. Code checks for:
    # a) Message - "System Rebooting ..."
    # b) Creates a file on filer 'file_407249' before snap restore and 
    #    checks if file exists after snap restore.
    # This workaround will be removed once product bug is fixed.
    
    $Log->comment("WorkAround for BURT 407249 - Restore the snapshot". 
                  " and validate if  snapshot got restored");
    my $text = "Workaround for BURT 407249";
    wrfile($conn, "/etc/file_407249", $text)
           or die "can't write to /etc/file_407249: $@";
    
    my $reboot_prompt = "System reboot";
    my (undef, undef, $output) =
        say($conn,"snap restore -f -s $args{snapname} -A $args{aggregate}",
            '-noprompt','-prompt',$reboot_prompt);
    boot($conn) or die $@;
    
    my $file_text = rdfile($conn, "/etc/file_407249") or die $@;
    if($file_text =~ /No such file or directory/) {
        $Log->comment("Snapshot $args{snapname} restored successfully"); 
        return 1;
    } else {
        $Log->error("Snapshot restore failed.");
        return 0;
    }
}


#
# local routine which returns description for a given state
#
sub _desc_for_state {
    my ($state) = @_;
    my @param_names;
    if ($state =~ /^Base$/i) {
    @param_names=qw(
                    FILER_CSETUP_BASE FILER_REAL_ZERO
                    FILER_DISK_TYPE FILER_IMAGE
                    FILER_BUILD_TYPE FILER_ONTAPDIR 
                    FILER_INSECURE_SERVICES NTP_SERVER
                    FILER_MINI_DISK_SIZE FILER_NEW_FS
                  );
    } elsif ($state =~ /^Application$/i) {
        my $filer_fas = param('FILER_FAS_LABEL', '-default', "");
	if ($filer_fas) {
	    read_conf($filer_fas);
	}
	@param_names=qw(
			   FILER_CSETUP_APPLICATION FILER_SERVICES
			   FILER_VOL_LIST FILER_ADMIN_USERS
			   FILER_ETC_PASSWD_USERS FILER_COMMAND
                           FILER_ROOT_AGGR_RENAME FILER_ROOT_VOL_RENAME
		       );
    } elsif ($state =~ /^Plan$/i) {
	@param_names=qw(
			   FILER_CSETUP_PLAN
		       );
    } else {
	die "Unknown state '$state'\n";
    }
    my @arr;
    foreach my $key (@param_names) {
    my $value = param($key);
    if ($key =~ /^FILER_IMAGE$/) {
       if (-l $value) {
          eval {
               $value = readlink($value);
          };
          if ($@) {
                   die "$@";
          }
       }
    }
    if (defined($value)) {
        push @arr, "$key=$value";
    }
    }
    return join(" ",@arr);
}

#
# local routine which checks the current FBS  description and the stored
# description are the same.
#
sub _check_fbsdesc {
    my $csetup_ignore = param('CSETUP_IGNORE', '-default', undef);
    my $file = _datfile_for_state("Base");
    my $conn = connect("#rsh");
    my $data = rdfile($conn,$file);
    my @lines = split(/\n/,$data);
    my $basedesc = &_desc_for_state("Base");
    my $desc = $lines[0];
    if (defined($csetup_ignore)) {
       $desc = _ignore_csetup($desc,$csetup_ignore); 
       $basedesc = _ignore_csetup($basedesc,$csetup_ignore); 
    }
    if ("$desc" eq "$basedesc") {
       return 1;
    } 
    return 0;
}

#
# local routine which checks the current FAS  description and the stored
# description are the same.
#
sub _check_fasdesc {
    my $csetup_ignore = param('CSETUP_IGNORE', '-default', undef);
    if (!&_check_fbsdesc()) {
        return 0;
    }
    my $file = _datfile_for_state("Application");
    my $conn = connect("#rsh");
    my $data = rdfile($conn,$file);
    my @lines = split(/\n/,$data);
    my $fasdesc = &_desc_for_state("Application");
    my $desc = $lines[0];
    if (defined($csetup_ignore)) {
       $desc = _ignore_csetup($desc,$csetup_ignore); 
       $fasdesc = _ignore_csetup($fasdesc,$csetup_ignore); 
    }
    if ("$desc" eq "$fasdesc") {
       return 1;
    } 
    return 0;
}

#
# local routine which checks the current FPS description and the stored description
# are the same.
#
sub _check_fpsdesc {
    my $csetup_ignore = param('CSETUP_IGNORE', '-default', undef);
    if (!&_check_fbsdesc()) {
        return 0;
    }
    if (!&_check_fasdesc()) {
        return 0;
    }
    my $file = _datfile_for_state("Plan");
    my $conn = connect("#rsh");
    my $data = rdfile($conn,$file);
    my @lines= split(/\n/,$data);
    my $desc = $lines[0];
    my $fpsdesc = &_desc_for_state("Plan");
    if (defined($csetup_ignore)) {
       $desc = _ignore_csetup($desc,$csetup_ignore); 
       $fpsdesc = _ignore_csetup($fpsdesc,$csetup_ignore); 
    }
    if ("$desc" eq "$fpsdesc") {
       return 1;
    } 
    return 0;
}

#
# local routine which checks the description for any state  with the stored
# description of that state.
#
sub _check_desc {
    my ($state) = @_;
    if ($state =~ /^Base$/i) {
       return &_check_fbsdesc;
    }
    if ($state =~ /^Application$/i) {
       return &_check_fasdesc;
    }
    if ($state =~ /^Plan$/i) {
       return &_check_fpsdesc;
    }
    return 0;
}

#
# local routine which removes the setup attributes from the descriptions 
#
sub _ignore_csetup {
    my ($str,$csetup_ignore) = @_;
    my ($regex,$reg);
    my @regs = split(/\s/,$csetup_ignore);
    foreach my $reg (@regs) {
      if ($reg =~ /\*$/) {
          $reg.="?";
      }
      $str =~ s/(\s|^)$reg(\s|$)/ /;
    }
    return $str;
}
         

=head1 NAME

CSetup::Snapshot::Aggregate - Aggregate snap create and 
aggregate snap restore, used for 7g/7-mode

=head1 SYNOPSIS

    use CSetup::Snapshot::Aggregate;

    my $host = host($FILER);
    my $obj = CSetup::Snapshot::Aggregate->new(host=>$host);

    $obj->is_available(state=>$state);

    $obj->create(state=>$state);

    $obj->revert_to(state=>$state);

    $obj->delete(state=>$state);


=head1 DESCRIPTION

This library contains methods that manipulate the aggregate snapshots.

The snapshot names have a mapping to the state name. __csetup__<state>
is the name given to snapshots created for state <state>.
Other snapshots that are not created by this library are not used.

There is a data file created in the root volume namely 
/etc/csetup-state-<state>. This file is created by the create method
and read by the is_available method. It contains the list of aggregates
that have been snapshot for that state and the attributes of aggregate
when the snapshot was taken. Currently it holds the list of disks 
that the aggregate is made up of.

=head1 METHODS

=head2 is_available

This method takes as input the name of the state.

Presence of snapshots are checked for all the aggregates 
listed in the state specific data file.
Checking is also done to verify that the aggregate attribues 
have not changed from what is present in the data file.

If the aggregate is missing or the snapshot is missing or 
attributes have changed then it returns false.

=head2 create

This method takes as input the name of the state.

It creates a snapshot __csetup__<state> for every aggregate and 
write out a new copy of the data file /etc/csetup-state-<state>

=head2 delete

This method takes as input the name of the state.

It removes all aggregate snapshots that the specified state maps to.

=head2 revert_to

This method takes as input the name of the state.

It reverts the system to the specified state. 

This is done by doing a snaprestore of all the aggregates on the system.
The root aggregate is first reverted. This action will trigger a reboot
of the system.

Also performs the following cleanup steps 

* failed disks if any are unfailed

* spares disks if any are zeroed

* all ontap hammers are aborted

=head1 FOR_TESTING

Use Example/testCSetupSnapshotAggregate.thpl for testing this lib

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = rm, disk unfail, disk zero spares, snap restore

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    dl-setupcleanup-libs@netapp.com
    $Id: //depot/prod/test/Rsierranevada/lib/CSetup/Snapshot/Aggregate.pm#1 $ 
    Last revised $Date: 2013/01/22 $ 

=head1 COPYRIGHT

    Copyright 2005,2009, Network Appliance Inc., All Rights Reserved.

=cut

1;
