# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CSetup/Snapshot/VI.pm#1 $

## @summary Library to take/restore snapshots of guests running in VMWare Virtual Infrastructure.
## @author bkavita@netapp.com, dl-setupcleanup-libs
## @status Shared
## @pod here

package CSetup::Snapshot::VI;

use strict;
use warnings;
use Tharn qw($Log script_result snooze config_result host);
use Params::Validate qw(validate SCALAR);
BEGIN
{
    Params::Validate::validation_options(
       on_fail => sub {
                      my @msgs = @_; 
                      chomp(@msgs);
                      die script_result([@msgs]);
       }, );
};
require Exporter;
our @ISA=qw(Exporter);
our @EXPORT_OK=qw(is_available revert_to create delete);

use VMware::VIRuntime;

=head1 NAME

CSetup::Snapshot::VI - Snapshot methods for guests running in VMWare Virtual Infrastructure.

=head1 SYNOPSIS

use CSetup::Snapshot;

my $host = host($client);

my $obj = CSetup::Snapshot->new(host=>$host);

$obj->is_available(state=>$state);

$obj->create(state=>$state);

$obj->revert_to(state=>$state);

$obj->delete(state=>$state);

=head1 DESCRIPTION

This library contains methods to create or restore snapshots of guests that are
running in VMWare Virtual Infrastructure (ESX, VirtualCenter, etc.)

=cut

sub new {
    my ($system) = shift @_;
    my (%opts) = validate (@_, {
                 'host' => { isa => 'Hostrec' }, 
                 } );
    my $class = ref($system) || $system;
    my $host = $opts{host};
    my $session = _login($host);
    my $self = {
                 host=> $host, 
                 session => $session,};
    bless($self,$class);
    return $self;
 
}

=head1 METHODS

=head2 is_available

 This method takes as input the name of the state.

 Checks if snapshot is available. If the snapshot is available then it returns 
 true else it return false.

=cut

sub is_available {
    my $self = shift;
    my (%args) = validate (@_, {'state' => { type=>SCALAR }, } );
     
    my $host = $self->{host};
    my $session = $self->{session};
    my $vm_name = $host->id();
    my $vm = _find_vm($session, $vm_name);
    my $snapshot = _get_snapshot($vm, state => $args{state});

    return ($snapshot) ? 1 : 0;
}


=head2 revert_to

 This method takes as input the name of the state.

 Checks if snapshot is available. If the snapshot is available then it will revert 
 the VM client to state specific snapshot else will throw an error.

=cut

sub revert_to {
    my $self = shift;
    my (%args) = validate (@_, {'state' => { type=>SCALAR }, } );
     
    my $host = $self->{host};
    my $session = $self->{session};
    my $vm_name = $host->id();
    my $vm = _find_vm($session, $vm_name);
    
    my $snapshot = _get_snapshot($vm, state => $args{state});
    my $snap_name = _generate_snapname($args{state});
    if ($snapshot) {
        eval{Vim::get_view(mo_ref =>$snapshot->{snapshot})->RevertToSnapshot();};
        die("Failed to revert to snapshot $snap_name - $@") if($@);
        $Log->comment("Restored snapshot - $snap_name successfully");
        _boot($vm, $vm_name);
    }
}


=head2 create

 This method takes as input the name of the state.

 Checks if the snapshot with exists. If the snapshot exists it first deletes
 the snapshot and then creates a snapshot.
 It creates a snapshot __csetup__<state> for the VM client.

=cut

sub create {
    my $self = shift;
    my (%args) = validate (@_, {'state' => { type=>SCALAR }, } );
     
    my $host = $self->{host};
    my $session = $self->{session};
    my $vm_name = $host->id();
    my $vm = _find_vm($session, $vm_name);
    
    _shutdown($vm, $vm_name);
    my $old_snapshot = _get_snapshot($vm, state => $args{state});
    my $snap_name = _generate_snapname($args{state});
   
    if ($old_snapshot) {
        eval{Vim::get_view(
                           mo_ref =>$old_snapshot ->{snapshot})->RemoveSnapshot( 
                                    removeChildren => 1 ); };
        die("Unable to delete the snapshot - $@") if($@);
        $Log->comment("Removed the snapshot - $snap_name.");
    } 
    eval{$vm->CreateSnapshot(name => $snap_name, 
                             description=>"Snapshot for $args{state} state", 
                             memory=>0, 
                             quiesce=>1);};

    die("Failed to create a snapshot - $@") if $@;
    $Log->comment("Created snapshot - $snap_name successfully");
    _boot($vm, $vm_name);
}


=head2 delete

 This method takes as input the name of the state.

 Checks if snapshot is available. If the snapshot is available then it will delete 
 the snapshot that maps to the specified state else will throw an error.

=cut

sub delete {
    my $self = shift;
    my (%args) = validate (@_, {'state' => { type=>SCALAR }, } );
     
    my $session = $self->{session};
    my $host = $self->{host};
    my $vm_name = $host->id();
    my $vm = _find_vm($session, $vm_name);
    my $old_snapshot = _get_snapshot($vm, state => $args{state});
    my $snap_name = _generate_snapname($args{state});
    
    if ($old_snapshot) {
        eval{Vim::get_view(
                        mo_ref => $old_snapshot->{snapshot} )->RemoveSnapshot( 
                                  removeChildren => 1 );};
        die("Unable to delete the snapshot - $@") if($@);
        $Log->comment("Removed the snapshot - $snap_name successfully.");
    } else {
        $Log->comment("Snapshot $snap_name does not exits.");
    }
}


# Name : _login()
# Description : Local method to create a session of virtual infrastructure
# Input Parameters : None
# Return Values : $session - Session object of VMWare Virtual Infrastructure

sub _login {
    my $host_obj = shift;
    my $session;

    my $esx_hostid = $host_obj->{VM_HOST};
    if (!$esx_hostid) {
        die config_result("Field 'VM_HOST' missing from host record "
                          .$host_obj->id().
                          ". Perhaps it is not a virtual machine?");
    }
    
    my $esx_host_obj = host($esx_hostid);
    my $service_url = "https://".$esx_host_obj->default_ip()."/sdk/vimService";
    my $user = $host_obj->{VI_USERNAME}
               or die config_result("Field 'VI_USERNAME' missing from host record "
                                    .$host_obj->id().
                                    ". Perhaps it is not an ESX VI server?");
    my $pass = $host_obj->{VI_PASSWORD}
               or die config_result("Field 'VI_PASSWORD' missing from host record "
                                    .$host_obj->id().
                                    ". Perhaps it is not an ESX VI server?");
       
    $Log->comment("Login to the virtual infrastructure server ....");
    eval {$session = Vim::login(service_url => $service_url, 
                                user_name => $user,
                                password => $pass); };
    die config_result("Failed to login - $@") if($@);
    return $session;
}


# Name : _find_vm()
# Description : Local method to create object of virtual machine.
# Input Parameters : $session - Session object of VMWare Virtual Infrastructure
#                             $vm_name - Name of the virtual machine
# Return Values : $vm_obj - Object of the given virtual machince

sub _find_vm {
    my $session = shift;
    my $vm_name=shift;
      
    my $vm_obj;
    $Log->comment("Searching  for $vm_name on the".
                  " virtual infrastructure server ....");
    eval {$vm_obj = $session->find_entity_view(
                              view_type=>'VirtualMachine',
                              filter=>{'name'=>$vm_name}); };
    die config_result("Failed to find the virtual machine on ESX Server. \n $@") 
              if($@ || !$vm_obj); 
    return $vm_obj;
}


# Name : _get_snapshot()
# Description : Local method to check if snapshot exists in 
#                     VirtualMachineSnapshotTree
# Input Parameters : $snap_name - Name of the snapshot
# Return Values : $snap_ref - Reference to the VirtualMachineSnapshot

sub _get_snapshot {
    my $virtual_machine = shift;
    my (%args) = validate (@_, {'state' => { type=>SCALAR }, } );

    my $snap_name = _generate_snapname($args{state});
    
    my @keys = keys(%$virtual_machine);
    if (exists($virtual_machine->{snapshot})) {
        my $snap_ref;
        eval {$snap_ref = $virtual_machine->snapshot->rootSnapshotList;};
        die("Failed to retrieve the snapshot list - $@") if ($@);
        return _get_snapshot_from_snaplist($snap_name, $snap_ref);
    } 
}


# Name : _get_snapshot_from_snaplist()
# Description : Local method to get snapshot from snapshot list.
# Input Parameters : $snap_name - Name of snapshot
#                    $snapshot_list - Reference to root snapshot list 
# Return Values : $snapshot - Reference to the VirtualMachineSnapshot

sub _get_snapshot_from_snaplist {
    my ($snap_name,$snapshot_list) = @_;
    
    return undef if (!$snapshot_list); 
    foreach my $snapshot (@$snapshot_list) {
        return $snapshot if ($snapshot->name eq $snap_name); 
    }
    foreach my $snapshot (@$snapshot_list) {
        my $match = _get_snapshot_from_snaplist(
                         $snap_name, $snapshot->childSnapshotList);
        return $match if ($match);
    }
    return undef;
}


# Name : _generate_snapname()
# Description : Local method to create snapshot name for given state
# Input Parameters : $state - Name of state
# Return Values : $snap_name - Name of snapshot

sub _generate_snapname {
    my $state = shift;
    my $snap_name = '__csetup__'.$state;
     
    return $snap_name;
}


# Name : _shutdown
# Description : Local method to shutdown vm client.
# Input Parameters : $vm - VM Client Object
#                    $vm_name - Name of the virtual machine
# Return Values : None

sub _shutdown {
    my $vm_obj = shift;
    my $vm_name = shift;
    
    $vm_obj->update_view_data();
    if ($vm_obj->runtime->powerState->val eq 'poweredOn') {
        $Log->comment("Shutting down the VM Client - $vm_name");    
        $vm_obj->ShutdownGuest();
        $Log->comment("Sleeping till the client is successfully powered off..");
        until ($vm_obj->runtime->powerState->val eq 'poweredOff') {
            $vm_obj->update_view_data();
            snooze(5); 
        }
        $Log->comment("VM Client - $vm_name successfully powered off.");
    }
}


# Name : _boot
# Description : Local method to power on vm client.
# Input Parameters : $vm - VM Client Object
#                    $vm_name - Name of the virtual machine
# Return Values : None

sub _boot {
    my $vm_obj = shift;
    my $vm_name = shift;
    
    $Log->comment("Turning On the VM Client - $vm_name");    
    $vm_obj->PowerOnVM();
    $Log->comment("Sleeping till the client is successfully powered on..");
    until ($vm_obj->runtime->powerState->val eq 'poweredOn') {
        $vm_obj->update_view_data();
        snooze(5); 
    }
    snooze(120);
    $Log->comment("VM Client - $vm_name successfully powered on.");
}

=head1 FOR_TESTING
 
 Use Example/testCSetupSnapshotVI.thpl for testing this lib.

=head1 STATUS

 Shared

=cut

1;

=head1 AUTHOR

  bkavita@netapp.com, dl-setupcleanup-libs
  $Id: //depot/prod/test/Rsierranevada/lib/CSetup/Snapshot/VI.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

 Copyright 2009, Network Appliance Inc., All Rights Reserved.

=cut
