# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CSetup/Utils.pm#1 $


## @summary Miscellaneous routines for Common
## Setup/Cleanup
## @author dl-setupcleanup-libs@netapp.com
## @pod here

=head1 NAME

CSetup::Utils

=head1 DESCRIPTION

Miscellaneous routines for Common Setup / Cleanup

=head1 ROUTINES

=cut

package CSetup::Utils;
use strict;
use warnings;
use Tharn qw($Log script_result config_result subtest host param say);
use Params::Validate qw(validate ARRAYREF);
use NATE::OS qw(login_name);
use NATE::Inc qw(get_file);
require Exporter;
our @ISA=qw(Exporter);
BEGIN {
    our @EXPORT_OK=qw(make_subtest_hook hold_jig read_conf set_option get_option);
    Params::Validate::validation_options(
                                         on_fail => sub {
                                             my @msgs = @_; chomp(@msgs);
                                             die script_result([@msgs]);
                                         },
                                        );
};


# Parse a string into arguments as a shell does (parse a string into
# words, where quotes and escapes can include whitespace into a
# "word")

sub _parse_args {
    my ($str) = @_;
    my @words;
    my $word=undef;
    my $in_quotes=0;
    while (1) {
        if (!$in_quotes && $str =~ /\G\s+/cg) {
            # whitespace ends a word (but only if not in quotes)
            if (defined($word)) {
                push @words, $word;
                $word=undef;
            }
        } elsif ($str =~ /\G\\(.)/scg) {
            # backslash makes next character literal (inside or outside any quotes)
            $word = "" if !defined($word);
            $word .= $1;
        } elsif ($str =~ /\G\"/cg) {
            # double quote
            # starts a new word (even if it will be an empty string)
            $word = "" if !defined($word);
            if (!$in_quotes) {
                # begin quote
                $in_quotes='"';
            } elsif ($in_quotes eq '"') {
                # end quote
                $in_quotes=0;
            } elsif ($in_quotes eq "'") {
                # double quote is literal inside single quoted string
                $word .= '"';
            }
        } elsif ($str =~ /\G\'/cg) {
            # single quote (exactly complementary to double quote; no
            # additional behavior)
            # starts a new word (even if it will be an empty string)
            $word = "" if !defined($word);
            if (!$in_quotes) {
                # begin quote
                $in_quotes="'";
            } elsif ($in_quotes eq "'") {
                # end quote
                $in_quotes=0;
            } elsif ($in_quotes eq '"') {
                # single quote is literal inside double quoted string
                $word .= "'";
            }
        } elsif ($str =~ /\G(.)/scg) {
            # any other single character
            $word = "" if !defined($word);
            $word .= $1;
        } else {
            # does not even match ".", so must be no more input
            last;
        }
    }
    if ($in_quotes) {
        die config_result("Unterminated [$in_quotes] quote in string $str");
    }
    if (defined($word)) {
        push @words, $word;
    }
    return \@words;
}

# parse a string into a list of subtests, similar to how ntest parses
# its command line arguments and its auxiliary test strings (PREJOB,
# POSTTEST, etc.), where there are global parameters and options,
# followed by a scriptspec, followed by one or more parameter
# assignments or options for just that scriptspec, followed by a
# second scriptspec, and so on.  returns an array ref of subtest
# specs, where each subtest spec is an array ref whose first element
# is the scriptspec and further elements are its parameter assignments
# and options.

# Example: "foo.thpl -client localhost '{logcomment(q(hello world))}' x=1 -timeout 3600"
# returns [["foo.thpl","-client","localhost"],["{logcomment(q(hello world))}","x=1","-timeout","3600"]]

sub _parse_subtests {
    my ($str) = @_;
    my @subtests;
    my $current_subtest;
    my @current_subtest_params;
    my @global_params;
    my $args = _parse_args($str);
    while (@$args) {
        my $arg = shift @$args;
        if ($arg =~ /^(?:\w+)\*?\=/) {
            # is a parameter assignment
            if ($current_subtest) {
                push @current_subtest_params, $arg;
            } else {
                push @global_params, $arg;
            }
        } elsif ($arg =~ /^-/) {
            # is a subtest option
            my $arg_arg = shift @$args;
            if (!defined($arg_arg)) {
                die config_result("Missing argument for option '$arg' in subtest list '$str'");
            }
            if ($current_subtest) {
                push @current_subtest_params, ($arg, $arg_arg);
            } else {
                push @global_params, ($arg, $arg_arg);
            }
        } else {
            # is a subtest
            if (defined($current_subtest)) {
                push @subtests, [$current_subtest, @global_params, @current_subtest_params];
            }
            $current_subtest=$arg;
            @current_subtest_params=();
        }
    }
    if (defined($current_subtest)) {
        push @subtests, [$current_subtest, @global_params, @current_subtest_params];
    }
    return \@subtests;
}

=head2 make_subtest_hook

A function that takes a list of subtests and their arguments, and returns a callback which
will call them in series (or undef if the list of subtests is empty)

=cut


sub make_subtest_hook {
    my ($subtest_str) = @_;
    my $subtest_list = _parse_subtests($subtest_str);
    my $callback;
    if (!@$subtest_list) {
        # empty 
        $callback=undef;
    } elsif (grep { $_->[0] =~ /^invalid$/i } @$subtest_list) {
        # "invalid"
        $callback = sub { die config_result("Subtest list contained magic test script named 'invalid',".
                         "telling us not even to attempt to run this list of tests"); };
    } else {
        $callback = sub {
            foreach my $aref (@$subtest_list) {
                subtest(@$aref);
            }
        };
    }
    return $callback;
}

=head2  hold_jig

A library routine which holds the current set of machines for manual inspection
It takes as an argument the list of machines whose state needs to be preserved
and sleeps forever.

=cut

sub hold_jig {
    my ($client, $filer, $tmp, @client, @filer, $user);
    my (@hosts) = ();
    my (%args) = validate(@_, {
                 'hosts' => { type => ARRAYREF },
                }
            );

    # Check whether MAIL_TO parameter is defined by the user.
    if (! defined $ENV{MAIL_TO}) {
    	$Log->error("Mail recepient not specified for hold jig (MAIL_TO)");
	return;
    }

    # Get the list of hosts and check whether hosts list is not empty !!
    @hosts = @{$args{'hosts'}};
    if (! @hosts) {
    	$Log->warn("No hosts specified to hold the jig for.");
	return;
    }
    $Log->comment("Hosts: @hosts");

    ## Identify which are filers and which are not, and 
    ## form a comma-separated variable of each type.
    ## Eg: $filer=FILER1,FILER2; $client=CLIENT1,CLIENT2
    foreach $tmp (@hosts) {
	my $obj = host($tmp);
	if (($obj->hosttype) =~ /filer/) {
	    push @filer, $tmp;
	} else {
	    push @client, $tmp;
	}
    }
    $filer = join ',', @filer;
    $client = join ',', @client;
    $Log->comment("Filers: $filer");
    $Log->comment("Clients: $client");

    ## Identify the username if the user has not explicitly passed it.
    $user = login_name(); 
    $Log->comment("Username = $user");

    ## Call hold_jig.thpl with the FILER and CLIENT information.
    subtest("tools/hold_jig.thpl", 
	    "RESERVE_FILERS=$filer",
	    "RESERVE_CLIENTS=$client",
	    "username=$user",
	    "HOLD_JIG=3",
	    "CTE=1");

    ## Wait till the subtest finishes.
    Subtest::wait_finish();
}

=head2 read_conf

A library routine to read and export set of parameters provided in conf file 
present in csetup/conf directory which is relative to the depot path
Thus making parameters available in NATE environtment.

=cut

sub read_conf {
    my ($label) = @_;
    my $file ="csetup/conf/$label\.conf";
    my $config = get_file($file) || die config_result("Not able to open the file $file: $!");
    foreach(split(/\n/, $config)) {
           if (m/^=/) {
              die config_result("Variables in $file are not defined in required PARAM_NAME=PARAM_VALUE format");
           }
           if (m/(.+?)[=](.+)/) {
              my $key = $1; my $value = $2;
              $key =~ s/^\s+//; $key =~ s/\s+$//;
              $value =~ s/^\s+//; $value =~ s/\s+$//;
              param($key,'-set',$value);
           }
    }
    close(FH);
}

=head2 set_option

This is a routine which will set the option on a filer. The parameters passed
are the connection handle, the option-name and the option-value. First, the 
current value of the option is checked by called get_option. If the option is
already set to the desired value, nothing is done. Else, the corresponding
option is set to the desired value and get_option is called to verify whether
the setting of the option was succesful.

=cut

sub set_option {
    my ($conn, $optname, $optval) = @_;
    my $cur_val; 
    
    # Get the current val. If same, then simply return.
    $cur_val = get_option($conn, $optname); # Get current value.
    if ($cur_val eq $optval) {
        $Log->comment("Already set: $optname = $optval");
        return 1;
    }
    
    # Else, set the option and do error-checking.
    say($conn, "options $optname $optval");
    $cur_val = get_option($conn, $optname);
    if ($cur_val eq $optval) {
        $Log->comment("Succesfully set: $optname = $optval");
        return 1;
    } else {
        $Log->error("Could not set $optname to $optval");
        return 0;
    }
}

=head2 get_option

This routine is used to get the value of an option. 

=cut

sub get_option {
    my ($conn, $option) = @_;
    my ($output, $value);

    #say($conn, "priv set diag");
    (undef, undef, $output) = say($conn, "options $option");
    $output =~ /^\s*(\S+)\s+(\S+)\s*.*$/;
    $value = $2;
    $Log->comment("Value of $option is $value");
    
    return $value;
}

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = none

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    dl-setupcleanup-libs@netapp.com
    $Id: //depot/prod/test/Rsierranevada/lib/CSetup/Utils.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005,2009, Network Appliance Inc., All Rights Reserved.

=cut

1;
