#Hey emacs, this is basically -*- perl -*-
# $Id$

#***********************************************************************
#
# NAME
#   ChangeModel
#
# SYNOPSIS
#   Rebrand a NetApp filer to be IBM, or vice versa;
#
# DESCRIPTION
#
# AUTHOR
#   dl-coati-core
#***********************************************************************

package ChangeModel;

use strict;
use Tharn qw($Staf subtest $Log host script_result);
use Data::Dumper;

use ChangeModel::SysDiag;
use ChangeModel::SlDiag;

# Hash of netapp to ibm mappings
# Important to keep this table updated, as there appears to be no way (currently) to 
# logically determine this information.  Also, though I have a spreadsheet of
# what SHOULD be allowable for each model, I have found discrepancies, like the 3040, 3070
# and 6070.  These SHOULD be AN0 or AN1, but I am only seeing "A" as a suffix.
# Also, for Shuar and the sldiag models (32xx and 62xx), they don't display the allowable
# model strings at all.
#
# Also, be very careful of the spreadsheet.  The models aren't all listed in increasing
# NetApp filer model numbers, OR are they listed in increasing IBM model numbers, so it is
# very easy to pick the wrong number when transcribing into this file.
#
# http://projects.eng.netapp.com/IBMRebranding_Pubs/MTM/IBM_Nseries_MTMs.pdf
#
# The SYS_MODEL from LOADER is the same format we need when changing the model 
our %netapp_2_ibm = qw(
    FAS2020   IBM-2859-AN0
    FAS2040   IBM-2859-AN1
    FAS2050   IBM-2862-AN0
    FAS3020   IBM-2864-AN0
    FAS3040   IBM-2869-A
    FAS3050   IBM-2865-AN0
    FAS3070   IBM-2868-A
    FAS3140   IBM-2858-AN0
    FAS3160   IBM-2858-AN2
    FAS3170   IBM-2858-AN1
    FAS3210   IBM-2858-CN0
    FAS3220   IBM-2858-EN5
    FAS3240   IBM-2858-EN1
    FAS3250   IBM-2858-EN6
    FAS3270   IBM-2858-EN2
    FAS6030   IBM-2866-A
    FAS6040   IBM-2866-AN1
    FAS6070   IBM-2867-A
    FAS6080   IBM-2867-AN1
    FAS6210   IBM-2867-CN0
    FAS6220   IBM-2867-CN0
    FAS6240   IBM-2867-EN1
    FAS6280   IBM-2867-EN2
    FAS6290   IBM-2867-EN2
    FAS2240-2 IBM-2857-AN2
    FAS2240-4 IBM-2857-AN4
    FAS2220   IBM-2857-AN5
    FAS8020   IBM-2868-CN0
    FAS8040   IBM-2868-CN1
    FAS8060   IBM-2868-CN2
    FAS8080   IBM-2868-EN4
);


# The short name is what Ontap returns as the model name for IBM
# The LOADER returns the long IBM names
# Ontap returns the same FAS names for NetApp filers
our %ibm_short_2_ibm_long = qw(
    N3300  IBM-2859-AN0
    N3400  IBM-2859-AN1
    N3600  IBM-2862-AN0
    N5200  IBM-2864-AN0
    N5500  IBM-2865-AN0
    N5300  IBM-2869-A
    N5600  IBM-2868-A
    N6040  IBM-2858-AN0
    N6060  IBM-2858-AN2
    N6070  IBM-2858-AN1
    N6210  IBM-2858-CN0
    N6220  IBM-2858-EN5
    N6240  IBM-2858-EN1
    N6250  IBM-2858-EN6
    N6270  IBM-2858-EN2
    N7600  IBM-2866-A
    N7800  IBM-2867-A
    N7700  IBM-2866-AN1
    N7900  IBM-2867-AN1
    N7550T IBM-2867-CN0
    N7750T IBM-2867-EN1
    N7950T IBM-2867-EN2
    N3220  IBM-2857-AN2
    N3240  IBM-2857-AN4
    N3150  IBM-2857-AN5
    NS8020     IBM-2868-CN0
    NS8040     IBM-2868-CN1
    NS8060     IBM-2868-CN2
    NS8080EX   IBM-2868-EN4
);

our %ibm_2_netapp = reverse %netapp_2_ibm;

# Sometimes the IBM model numbers can be -Cxx or -Exx depending
# on whether or not there is an ioxm (?) module?
# So allow for these variants as well.

our %alternate_netapp_2_ibm = qw(
    FAS3220   IBM-2858-CN5
    FAS3240   IBM-2858-CN1
    FAS3250   IBM-2858-CN6
    FAS3270   IBM-2858-CN2
);

our %alternate_ibm_short_2_ibm_long = qw(
    N6220  IBM-2858-CN5
    N6240  IBM-2858-CN1
    N6250  IBM-2858-CN6
    N6260  IBM-2858-CN2
);

our %alternate_ibm_2_netapp = reverse %alternate_netapp_2_ibm;

our $maint_mode_prompt = '\*>';

our $vespa_models         = qr/FAS224|N3240|IBM-2857-AN2|IBM-2857-AN4/;
our $simple_toggle_models = qr/FAS224|FAS222|N3240|N3150|IBM-2857-AN2|IBM-2857-AN4|IBM-2857-AN5/;
our $sldiag_models        = qr/FAS224|FAS32|FAS62|FAS222|N62|N755|N775|N795|N3220|N3240|N3150|NS8020|NS8040|NS8060|NS8080EX|FAS8020|FAS8040|FAS8060|FAS8080|N7550T|N7950T|FAS6220|FAS6290/;

our @Absolut_M_part_numbers = qw( 111-00616 111-00639 );
our @Reyka_M_part_numbers = qw( 111-01462 111-01465 );

## @get_subtest_seq()
## @description Generate a unique 3 digit number-string for subtest tagging.
##
## @synopsis
##
##    $seq = ChangeModel::get_subtest_seq( );
##
## @return       :  3 digit number with, leading zeros, ex: "001"
##------------------------------------------------------------------------------
# Subtest sequence number
my $subtest_seq = 0;

sub get_subtest_seq {
    $subtest_seq++;
    return sprintf( "%03d", $subtest_seq );
}


## @is_sldiag_model()
## @description Is this filer model one which uses sldiag to change model number?
##
## @arg model
## @arg host        Required: NACL Host object
## @arg conn        Required: NACL ConnectRec object
##
## @synopsis
##
##    ChangeModel::is_sldiag_model( "FAS6210" );
##
## @return       :  0|1
##------------------------------------------------------------------------------
sub is_sldiag_model {
    my $model = shift;
    if ( $model =~ $sldiag_models ) {
        return 1;
    }
    return 0;
}

## @get_current_model()
## @description Routine which attempts to get the current filer model by the closest
##              (transit-wise) source
##
## @arg transit_obj Required: NACL Transit object
## @arg host        Required: NACL Host object
## @arg conn        Required: NACL ConnectRec object
##
## @synopsis
##
##    $model = ChangeModel::get_current_model( transit_obj => $trans,
##                                             host => $host, conn => $conn );
##
## @return       :  "model string"
##------------------------------------------------------------------------------
sub get_current_model {
    my $whoami = (caller(0))[3];
    my %opt = (
                  transit_obj => undef,
                  host        => undef,
                  conn        => undef,
                  @_
              );

    my $current_model;

    if ( !$opt{transit_obj} || !$opt{conn} || !$opt{host} ) {
        die "$whoami - Must provide transit_obj, host and conn options";
    }

    my $model_name_regex = qr(Model Name:\s+([0-9a-zA-Z\-]+));
    my $trans = $opt{transit_obj};
    my $host  = $opt{host};
    my $conn  = $opt{conn};
    my $hostname = $host->{hostname};

    my $state = $trans->get_state( timeout => 900, maxidle => 900 );
    $Log->comment("$whoami - $hostname: Initial state: $state");
    if ( $state =~ /USERNAME|PASSWORD/ ) {
        $trans->change_state( to               => 'CLI',
                              iftakenover      => 'takeback', 
                              inhibit_takeover => 1,
                              timeout          => 900,
                              maxidle          => 900 );
        $state = $trans->get_state( timeout => 900, maxidle => 900 );
        $Log->comment("$whoami - $hostname: Needed to login, new state: $state");
    }

    if ( $state !~ /^(CLI|MAINT|FIRMWARE)$/ ) {
        return $current_model;
    }

    $state = $trans->get_state( timeout => 900, maxidle => 900 );

    if ( $state eq 'CLI' ) {
        my $get_model = 'run -node local -command sysconfig -h';
        if ( $host->{hosttype} !~ /ng/ ) {
            $get_model = 'sysconfig -h';
        }
        $Log->comment("$whoami - $hostname: state is: $state, executing $get_model");
        my $output = $conn->execute( command => $get_model, timeout => 900 );
        if ( $output =~ $model_name_regex ) {
            $current_model = $1;
        }
    } elsif ( $state eq 'MAINT' ) {
        my $get_model = 'sysconfig -v';
        $Log->comment("$whoami - $hostname: state is: $state, executing $get_model");
        my $output = $conn->execute( command => $get_model,
                                     timeout => 900,
                                     prompts => [$maint_mode_prompt] );
        if ( $output =~ $model_name_regex ) {
            $current_model = $1;
        }
    } elsif ( $state eq 'FIRMWARE' ) {
        $Log->comment("$whoami - $hostname: state is: $state, calling ChangeModel::get_loader_sys_model()");
        $current_model = ChangeModel::get_loader_sys_model( host => $host, conn => $conn );
    } else {
        $Log->warn("$whoami - $hostname: unexpected state: $state, should be CLI|MAINT|FIRMWARE");
    }

    # Ok, the firmware model is the same format we need when we set the model, but
    # ontap and maint_mode return the short IBM version string
    if ( $current_model ) {
        if ( defined $ibm_short_2_ibm_long{$current_model} ) {
            $current_model = $ibm_short_2_ibm_long{$current_model};
        }
    } else {
        $Log->warn("$whoami - $hostname: Could not determine filer model string");
    }

    return $current_model;
}

## @get_loader_sys_model()
## @description Routine which queries SYS_MODEL from LOADER
##
## @arg host Required: NATE host object
## @arg conn Required: NATE Connection object
##
## @synopsis
##
##    $model = ChangeModel::get_loader_sys_model( host => $host, conn => $conn );
##
## @return       :  "model string"
##------------------------------------------------------------------------------
sub get_loader_sys_model {
    my %opt = (
                  host => undef,
                  conn => undef,
                  @_
              );
    my $printenv_cmd          = 'printenv SYS_MODEL';
    my $printenv_regex        = qr/SYS_MODEL\s+(\S+)/;

    my $sysmodel;
    if ( $opt{host} && $opt{conn} ) {
        $opt{conn}->discardbuff();
        my $output = $opt{conn}->execute( command => $printenv_cmd,
                                          prompts => $opt{host}->prompt_fw() );

        if ( $output =~ $printenv_regex ) {
            $sysmodel = $1;
        }
    }

    return $sysmodel;
}

## @transit_to_final()
## @description Routine which transits to a given state
##
## @arg transit_obj Required: NATE host object
## @arg state       Required: FIRMWARE|CLI|FIRMWARE_THEN_CLI
##
## @synopsis
##
##    $model = ChangeModel::transit_to_final( transit_obj => $trans, state => $state );
##
##    Note: FIRMWARE_THEN_CLI is not a valid Transit.pm state, but just one that we
##          are using in this routine.
##
## @return       :  nothing
##------------------------------------------------------------------------------
sub transit_to_final {
    my $whoami = (caller(0))[3];
    my %opt = (
                  transit_obj => undef,
                  conn_obj    => undef,
                  state       => undef,
                  timeout     => 900,
                  maxidle     => 900,
                  @_
              );

    my $curr;
    if ( $opt{transit_obj} && $opt{state} && ($opt{state} =~ /^(FIRMWARE|CLI|FIRMWARE_THEN_CLI)$/) ) {
        $curr = $opt{transit_obj}->get_state( timeout => $opt{timeout}, maxidle => $opt{maxidle} );
        $Log->comment("$whoami - current_state = $curr, transit_to = ".$opt{state});
        # I would have thought that NACL::Transit would handle this, but at least in the BR automation
        # branch it doesn't and figured I would just add this check in main and propagate to the other
        # branches
        if ( $curr ne $opt{state} ) {
            if ( $opt{state} =~ /^FIRMWARE/ ) {
                $opt{transit_obj}->change_state( to               => 'FIRMWARE', 
                                                 iftakenover      => 'takeback',
                                                 inhibit_takeover => 1,
                                                 powercycle       => 1,
                                                 timeout          => $opt{timeout},
                                                 maxidle          => $opt{maxidle} );
                $curr = $opt{transit_obj}->get_state( timeout => $opt{timeout}, maxidle => $opt{maxidle} );
            }
    
            # the powercycle option is that if the transit fails, powercycle and try one more time
            if ( $opt{state} =~ /CLI$/) {
                $Log->comment("$whoami - current_state = $curr, transit_to = ".$opt{state});
                $opt{transit_obj}->change_state( to               => 'CLI', 
                                                 iftakenover      => 'takeback', 
                                                 inhibit_takeover => 1,
                                                 powercycle       => 1,
                                                 timeout          => $opt{timeout},
                                                 maxidle          => $opt{maxidle} );
    
                # Often the first node that comes up a second or two faster than the other will 
                # takeover and it might take quite a while before giveback occurs (have run into
                # many cases where a veto might exist for over 15 minutes
                eval {
                    if ( $opt{conn_obj} ) {
                        $opt{conn_obj}->execute( command => "storage failover giveback -fromnode local -override-vetoes true",
                                                 match_table => [
                                                              '{y|n}:' => 'y',
                                                          ],
                                                 timeout => 300, );
                    }
                };
            }
            $curr = $opt{transit_obj}->get_state( timeout => $opt{timeout}, maxidle => $opt{maxidle} );
            $Log->comment("$whoami - final_state = $curr");
        }
    } else {
        die "$whoami - Must provide transit_obj and state options, and state must equal FIRMWARE|CLI|FIRMWARE_THEN_CLI";
    }
}

## @target_model()
## @description Routine which determines a specific model name from a
##              generic model_type (i.e. netapp to FAS6070, or toggle to IBM-2859-A)
##
## @arg current_model        Required: Specific model name (i.e. FAS3070, IBM-2859-A, etc...)
## @arg requested_model_type Optional: Generic model_type (netapp|ibm|toggle|same), default:netapp
##
## @synopsis
##
##    $model = ChangeModel::target_model( current_model => 'FAS2040', requested_model_type => 'toggle' );
##
##  Note: The "same" requested_model_type is purely for testing purposes in that it will set the
##        model to what it currently is set to.
##
## @return       :  "model string"
##------------------------------------------------------------------------------
sub target_model {
    my $whoami = (caller(0))[3];
    my %opt = (
                  current_model         => undef,
                  requested_model_type  => 'netapp',
                  @_
              );

    if ( !$opt{current_model} ) {
       die "$whoami - Must pass in option current_model";
    }

    my $current_model   = $opt{current_model};
    my $req_model_type  = $opt{requested_model_type};

    my $target_model = "";

    my $to_ibm    = 0;
    my $to_netapp = 0;

    # This first one is strictly a testing state.  Means set it to what it
    # currently is set to.  For testing out the "set" path on new hardware where
    # you don't want to take the chance of "bricking" a brand-new filer... :)
    if ( $req_model_type eq "same" ) {
        $target_model = $current_model;
    } elsif ( $req_model_type eq "toggle" ) {
        if ( $current_model =~ /^ibm/i ) {
            $to_netapp = 1;
        } elsif ( $current_model =~ /^fas/i ) {
            $to_ibm = 1;
        }
    } elsif ( ($req_model_type eq "netapp") &&
              ($current_model =~ /^ibm/i) ) {
        $to_netapp  = 1;
    } elsif ( ($req_model_type eq "ibm") &&
              ($current_model =~ /^fas/i) ) {
        $to_ibm = 1;
    }

    if ( $to_netapp ) {
        if ( defined $ibm_2_netapp{$current_model} ) {
            $target_model = $ibm_2_netapp{$current_model};
        } elsif ( defined $alternate_ibm_2_netapp{$current_model} ) {
            $target_model = $alternate_ibm_2_netapp{$current_model};
        } else {
            die "$whoami - No NetApp equivalent tabled for $current_model";
        }

        # Since Absolut-M and Reyka-M have the same IBM model number,
        # need to determine the corresponding unique NetApp model number

        if ( $target_model =~ /FAS6280|FAS6290/i ) {
            use NACL::C::Node;
            use Hostrec;
            use Connectrec;
            use NACL::Transit;
            use NATE::BaseException qw(:try);

            my $node = NACL::C::Node->find();

            my $transit_obj;
            my $host_obj;
            my $conn_obj;
            my $node_name = $node->node();

            try {
                $host_obj = Hostrec->new( id => $node_name );
                $conn_obj
                    = Connectrec->new( host => $host_obj, id => "console" );
                $transit_obj = NACL::Transit->new( name => $conn_obj );
            } catch NATE::BaseException with {
                my $ex = shift;
                $Log->error("Could not create a Transit object");
                $ex->throw();
            };

	    my $saved_state = $transit_obj->get_state();

            if ( $saved_state ne "MAINT" ) {
                try {
                    $transit_obj->change_state( to => 'MAINT' );
                }
                catch NACL::Transit::Exceptions::TransitException with {
                    my $ex = shift;
                    $Log->error("Could not Transit to Maintenance Mode");
                    $ex->throw();
                };
            }

            my $sysconfig_prompts = [
                "^$node_name>", "^$node_name\\*>",
                "^dblade>",     "^dblade\\*>",
                "^sysconfig>",  "^sysconfig\\*>",
                "^\\>",         "^\\*>",
                "\\*>",
            ];

            my $output;
            try {
                $output = $conn_obj->execute(
                    command => "sysconfig -v",
                    timeout => 900,
                    prompts => $sysconfig_prompts,
                );
            } catch NATE::BaseException with {
                my $ex = shift;
                $Log->error('Error executing sysconfig -v :');
                $ex->throw();
            };

            if ( $output
                =~ /Part Number:\s+($Absolut_M_part_numbers[0]|$Absolut_M_part_numbers[1])/i
                )
            {
                $Log->comment("Absolut-M part number detected: $1");
                $target_model = "FAS6280";
            } elsif ( $output
                =~ /Part Number:\s+($Reyka_M_part_numbers[0]|$Reyka_M_part_numbers[1])/i
                )
            {
                $Log->comment("Reyka-M part number detected: $1");
                $target_model = "FAS6290";
	    }

            if ($transit_obj->get_state() ne "$saved_state"  ) {
                try {
                    $transit_obj->change_state( to => "$saved_state" );
                }
                catch NACL::Transit::Exceptions::TransitException with {
                    my $ex = shift;
                    $Log->error("Could not Transit to $saved_state");
                    $ex->throw();
                };
            }
        }
    } elsif ( $to_ibm ) {
        if ( defined $netapp_2_ibm{$current_model} ) {
            $target_model = $netapp_2_ibm{$current_model};
        } elsif ( defined $alternate_netapp_2_ibm{$current_model} ) {
            $target_model = $alternate_netapp_2_ibm{$current_model};
        } else {
            die "$whoami - No IBM equivalent tabled for $current_model";
        }
    }

    return $target_model;
}

## @get_subtest_results()
## @description Routine which waits for subtests to complete, then 
##              retrieves the results
##
## @arg subtests Required: array ref to subtest array
## @arg errors   Optional: array ref to error array
##
## @synopsis
##
##    $model = ChangeModel::get_subtest_results( subtests => \@subtests, errors => \@errors );
##
## @return       :  0|1
##------------------------------------------------------------------------------
sub get_subtest_results {
    my $whoami = (caller(0))[3];
    my %opt = ( subtests => undef,
                errors   => undef,
                @_ );

    if ( ref($opt{subtests}) !~ /ARRAY/ ) {
       die "$whoami - Must pass in option subtests as an arrayref";
    }

    if ( $opt{errors} && (ref($opt{errors}) !~ /ARRAY/) ) {
       die "$whoami - The errors option, if passed in, must be an arrayref";
    }

    my $script_sev = NATE::Result::Script::severity();
    my $fail_sev   = NATE::Result::Fail::severity();
    my $rc         = $script_sev;
    my @subtests   = ();

    if ( $opt{subtests} ) {
        @subtests = @{$opt{subtests}};

        Subtest::wait_finish( subtests => \@subtests );

        $rc = NATE::Result::Pass::severity();

        for my $st (@subtests) {
            my @res = $st->own_results();
            if (@res) {
                my $worst = NATE::Result::worst(@res);

                if ( $worst->severity() >= $rc ) {
                    $rc = $worst->severity();
                }

                # Add result strings into errors array
                if ( $opt{errors} && ($worst->severity() >= $fail_sev) ) {
                    push( @{$opt{errors}}, 'Subtest: '.$worst->runid().' failed with '.join( ',', $worst->messages()) );
                }
            }
        }
    }

    return $rc;
}


sub cold_shutdown {
    my $whoami = (caller(0))[3];
    my %opt = ( nodes => undef,
                @_ );

    if ( ref($opt{nodes}) !~ /ARRAY/ ) {
       die "$whoami - Must pass in option nodes as an arrayref";
    }

    my $nodes = $opt{nodes};

    my $gx_shutdown_cold_retries = 3;
    my $shutdown_timeout = 200;

RETRY_SHUTDOWN_COLD:

    $gx_shutdown_cold_retries--;

    # Timeout will be 300, 400, 500, 600 (max) seconds
    $shutdown_timeout += 100;

    my $st;
    my @results;

    $Log->comment("Cold Shutdown of nodes");
    # Doing this separately in case node is 7mode
    my @errors = ();
    my @subtests;
    my $failure;
    eval {
        my $index = 0;
        foreach my $node (@{$nodes}) {
            my $seq = get_subtest_seq( );
            $st = subtest( 'os/boot/gx_boot.thpl',
                           'gx_boot_TASKS=shutdown_cold',
                           'gx_boot_NODES='.$node,
                           '-bg',
                           '-timeout', $shutdown_timeout,
                           '-runid', $seq."_gx_boot_shutdown_cold_".$node,
                           "-onexit", sub { } );
            push( @subtests, $st );
            $index++;
        }
        Subtest::wait_finish( subtests => \@subtests );
    };

    if ( $@ ) {
        $failure = $@;
        push( @errors, $failure );
    }
    
    ChangeModel::get_subtest_results( subtests => \@subtests,  errors => \@errors );
    @subtests=();
    
    if ( @errors ) {
        $Log->warn("$whoami: gx_boot.thpl shutdown_cold failed with |".join("\n", @errors)."|");
        if ( $gx_shutdown_cold_retries < 0 ) {
            die "$whoami: Exhausted all gx_boot.thpl shutdown_cold retries";
        } else {
    
            # only retry on nodes that failed
            my $new_nodes = [];
            foreach my $node (@$nodes) {
                my $check_runid = '_gx_boot_shutdown_cold_'.$node;
                foreach my $error (@errors) {
                    if ( $error =~ /$check_runid/ ) {
                        push( @$new_nodes, $node );
                        last;
                    }
                }
            }
            $nodes = $new_nodes;
            $Log->comment("$whoami: retry gx_boot.thpl shutdown_cold in 10 seconds");
            Tharn::sleep(10);
            goto RETRY_SHUTDOWN_COLD;
        }
    }
}

1;
