#Hey emacs, this is basically -*- perl -*-
# $Id$

#***********************************************************************
#
# NAME
#   ChangeModel::SlDiag
#
# SYNOPSIS
#   Rebrand a NetApp filer to be IBM, or vice versa, via sldiag.
#
# DESCRIPTION
#
# AUTHOR
#   dl-coati-core
#***********************************************************************

package ChangeModel::SlDiag;

use strict;
use Tharn qw($Staf subtest $Log host fail_result script_result);
use Archive::Tar;
use Archive::Tar::File;
use Data::Dumper;
use File::Temp;
use NATE::BaseException qw(:try);
use NACL::APISet::Version;

my $boot_diags_cmd          = 'boot_diags';

## @is_simple_toggle_model()
## @description Is this filer model one which uses IBM/FAS as generic models
##              when rebranding?
##
## @arg model   Required: IBM or NetApp Model string
##
## @synopsis
##
##    ChangeModel::SlDiag::is_simple_toggle_model( "FAS2240-2" );
##
## @return       :  0|1
##------------------------------------------------------------------------------
sub is_simple_toggle_model {
    my $model = shift;
    my $toggle_models = $ChangeModel::simple_toggle_models;
    if ( $model =~ $toggle_models ) {
        return 1;
    }
    return 0;
}

## @requires_sp_model_refresh()
## @description Is this filer model one which requires the SP to refresh the SYS_MODEL?
##
## @arg model   Required: IBM or NetApp Model string
##
## @synopsis
##
##    ChangeModel::SlDiag::requires_sp_model_refresh( "FAS2240-2" );
##
## @return       :  0|1
##------------------------------------------------------------------------------
sub requires_sp_model_refresh {
    my $model = shift;
    my $refresh_models = $ChangeModel::simple_toggle_models;
    if ( $model =~ $refresh_models ) {
        return 1;
    }
    return 0;
}

## @need_to_set_model_on_midplane()
## @description Does this filer require you to change the model on midplanes?
##
## @arg model   Required: IBM or NetApp Model string
##
## @synopsis
##
##    ChangeModel::SlDiag::need_to_set_model_on_midplane( "FAS2240-2" );
##
## @return       :  0|1
##------------------------------------------------------------------------------
sub need_to_set_model_on_midplane {
    my $model = shift;

    # Currently only Vespa and Ruckus are sldiag with no need to set model on 
    # the midplane, may change in the future
    my $no_midplane_models = $ChangeModel::simple_toggle_models;
    if ( $model =~ $no_midplane_models ) {
        return 0;
    }
    return 1;
}

## @target_model()
## @description Routine which determines a specific model name from a
##              generic model_type (i.e. netapp to FAS6070, or toggle to IBM-2859-A)
##
## @arg current_model        Required: Specific model name (i.e. FAS3070, IBM-2859-A, etc...)
## @arg requested_model_type Optional: Generic model_type (netapp|ibm|toggle|same), default:netapp
##
## @synopsis
##
##    $model = ChangeModel::SlDiag::target_model( current_model => 'FAS2040', requested_model_type => 'toggle' );
##
##  Note: The "same" requested_model_type is purely for testing purposes in that it will set the
##        model to what it currently is set to.
##
## @return       :  "model string" or "IBM|FAS" for Vespa
##------------------------------------------------------------------------------
sub target_model {
    my $whoami = (caller(0))[3];
    my %opt = (
                  current_model         => undef,
                  requested_model_type  => 'netapp',
                  @_
              );

    if ( !$opt{current_model} ) {
       die "$whoami - Must pass in option current_model";
    }

    my $current_model   = $opt{current_model};
    my $req_model_type  = $opt{requested_model_type};
    my $target          = ChangeModel::target_model( %opt );

    # If filer that uses simple FAS|IBM as target string (like Vespa)
    if ( is_simple_toggle_model( $target ) ) {
        if ( $target =~ /^FAS/ ) {
            $target = 'FAS';
        } elsif ( $target =~ /^IBM/ ) {
            $target = 'IBM';
        }
    }

    return $target;
}

## @change_model()
## @description Routine which will attempt to change the Model of a NetApp
##              filer to be that of it's IBM equivalent, or vice versa.
##              For 32xx and 62xx filers, model changing is accomplished
##              via sldiag, and must be done on each node controller and
##              midplanes at the same time.
##
## @arg node_pair            Required: ARRAY ref to SFO pair of nodes
## @arg requested_model_type Required: netapp|ibm|toggle
## @arg final_transit_state  Optional: Any NACL Transit.pm state 
##                                         (i.e. "FIRMWARE", "CLI", etc...)
##
## @synopsis
##
##    ChangeModel::change_model( node_pair=>\@pair, requested_model_type=>"toggle", 
##                               final_transit_state => "FIRMWARE" );
##
## @return       :  1|die
##------------------------------------------------------------------------------
sub change_model {
    my $whoami = (caller(0))[3];
    my @errors;
    my $_errors = 0;

    sub _declare_error {
        $_errors++;
        return undef;
    }

    # Prompts can be "SP> ", "SP*> ", "SP name> ", or "SP name*> "
    # Though I have also seen the space after the ">" cut off, so
    # this regex should handle them all.
    my $sp_prompts = ['SP\s*\S*>\s?'];
    my %opt = (
                  node_pair            => undef,
                  requested_model_type => undef,
                  final_transit_state  => undef,
                  @_
              );

    my $pair          = $opt{node_pair};
    my $model_type    = $opt{requested_model_type};
    my $final_state   = $opt{final_transit_state};

    if ( !$pair || !$model_type ) {
        die script_result("$whoami - Must specify node_pair and requested_model_type options");
    }

    # We're going to at least allow change_model to CHECK to see if a single node is
    # branded as we expect it, but will not allow changes if only given a single node
    # Or actually any number of nodes other than 2.
    my $num_nodes = scalar(@$pair);

    my (%slobjs, $conn, $nodehost, $fw_prompts, $trans);
    my ($current_model, $target_model, $output);
    foreach my $node (@$pair) {

        $conn     = Tharn::connect($node.'#console');
        $nodehost = Tharn::host( $node );

        $fw_prompts = $nodehost->prompt_fw();

        $trans = NACL::Transit->new(name => $conn);

        # save state for later use
        $slobjs{CONN}{$node}        = $conn;
        $slobjs{HOST}{$node}        = $nodehost;
        $slobjs{PROMPTS}{$node}     = $fw_prompts;
        $slobjs{TRANSIT}{$node}     = $trans;
        $slobjs{FINAL_STATE}{$node} = $final_state;
    }

    # Running into problems where Transit doesn't seem to handle filers that are in
    # weird states very well.  Use gx_boot to cold_shutdown the nodes
    ChangeModel::cold_shutdown( nodes => $pair );

    # Getting in a state to get version manager
    # Transit to CLI if we haven't
    $trans->change_state(
        to               => 'CLI',
        inhibit_takeover => 1,
        iftakenover      => 'takeback',
        timeout          => 900,
        maxidle          => 900
    );
    $Log->comment( "Successfully transit to state: CLI" );

    # Create version manager object
    my $version_manager;
    my $promoted;
    try {
        $version_manager = NACL::APISet::Version->new(
            hostobj    => $nodehost,
            category   => "Node",
            shell_type => "CMode"
        );

        $promoted = $version_manager->is_promoted();
        $Log->debug( "Result of checking build promotion: $promoted" );
    } catch NATE::BaseException with {
        my $ex = shift;
        $Log->error( "Failed to get version manager" . $ex->text() . $ex->stacktrace() );
    }

    # New sldiag model menu only supported in release 8.2.1 and greater
    my $sldiag_system_model_enabled;
    if( $version_manager->is_version_ge( release => '8.2.1' ) ) {
        $sldiag_system_model_enabled = 1;
    }

    my $initial_transit_retries = 3;
INITIAL_TRANSIT:

    $initial_transit_retries--;

    if( $promoted ) {
        foreach my $node (@$pair) {
            $conn = $slobjs{CONN}{$node};

            # Ensure at loader prompt
            $trans->change_state(
                to               => 'FIRMWARE',
                inhibit_takeover => 1,
                iftakenover      => 'takeback',
                timeout          => 900,
                maxidle          => 900 );

            eval {
                $conn->execute(
                    command => 'setenv sldiag-oem-enabled? true',
                    timeout => 20,
                );
            };
            if ( $@ ) {
                $Log->error( "Could not set sldiag oem: $@");
            }
        }
    }

    # This can take a little while...  Subtest to parallelize.
    my @subtests;
    foreach my $node (@$pair) {
        $trans         = $slobjs{TRANSIT}{$node};
        $conn          = $slobjs{CONN}{$node};
        $nodehost      = $slobjs{HOST}{$node};

        # Get the current model, if possible, by doing trivial transits
        $current_model = ChangeModel::get_current_model( transit_obj => $trans, conn => $conn, host => $nodehost );

        # If we are unable to get the current model from CLI|MAINT|FIRMWARE, we have to transit to firmware
        if ( !$current_model ) {
            my $seq = ChangeModel::get_subtest_seq();
            my $st = subtest( sub { $trans->change_state( to => 'FIRMWARE', 
                              inhibit_takeover => 1,
                              iftakenover      => 'takeback',
                              timeout          => 900,
                              maxidle          => 900 ); },
                               "-runid", $seq."_".$node."_transit_to_firmware",
                               "-bg",
                               "-onexit", sub { },
                      );
            push( @subtests, $st );
        } else {
            $slobjs{CURRENT_MODEL}{$node} = $current_model;
        }
    }

    my @transit_errors;
    if ( @subtests ) {
        ChangeModel::get_subtest_results( subtests => \@subtests, errors => \@transit_errors );

        if ( @transit_errors ) {
            if ( $initial_transit_retries > 0 ) {
                $Log->warn("Error transiting to FIRMWARE ".Dumper(\@transit_errors));
                Tharn::sleep(120);
                goto INITIAL_TRANSIT;
            } else {
                $Log->error("Error transiting to FIRMWARE ".Dumper(\@transit_errors));
                push( @errors, @transit_errors );
                goto FINAL_TRANSIT;
            }
        }
    }

    foreach my $node (@$pair) {

        # If we could not get the current model before a transit to FIRMWARE, then try again
        if ( !$slobjs{CURRENT_MODEL}{$node} ) {
            $trans         = $slobjs{TRANSIT}{$node};
            $conn          = $slobjs{CONN}{$node};
            $nodehost      = $slobjs{HOST}{$node};
    
            # Get the current model
            $current_model = ChangeModel::get_current_model( transit_obj => $trans, conn => $conn, host => $nodehost );
    
            if ( !$current_model ) {
                # We've already transited to firmware and still can't get the model, now just give up
                push(@errors, "Could not find current model for $node");
    
            } else {
                $slobjs{CURRENT_MODEL}{$node} = $current_model;
            }
        }

    }

    if ( @errors ) {
        goto FINAL_TRANSIT;
    }

    my $nop = 0;
    foreach my $node (@$pair) {
        $slobjs{TARGET_MODEL}{$node}    = $slobjs{CURRENT_MODEL}{$node};
        $slobjs{NOP}{$node}    = 0;

        # Target Model of "" means we already ARE this model
        eval {
            $target_model = ChangeModel::SlDiag::target_model( current_model => $slobjs{CURRENT_MODEL}{$node},
                                                               requested_model_type => $model_type );
        };

        if ( !$@ ) {
            if ( !$target_model ) {
                $Log->comment("Node: $node, Current model: $slobjs{CURRENT_MODEL}{$node}, Target: $model_type is a NOP");
                $slobjs{NOP}{$node}    = 1;
                $nop++;
            } else {
                $Log->comment("Node: $node, Current model: $slobjs{CURRENT_MODEL}{$node}, Target: $target_model");
                $slobjs{TARGET_MODEL}{$node}    = $target_model;
                if ( $num_nodes != 2 ) {
                    push(@errors, "sldiag node $node needs to be changed to target_model $target_model, but num_nodes $num_nodes != 2");
                    die fail_result("$whoami - ".join("\n", @errors));
                }
            }
        } else {
            push(@errors, "sldiag node: $node, requested_model_type: $model_type, but failed with $@");
            die fail_result("$whoami - ".join("\n", @errors));
        }
    }

    # If nothing to do on either node we are done.  If only one node needs it's model changed, the other
    # will just be set again to it's current model.
    if ( $nop == $num_nodes ) {
        goto FINAL_TRANSIT;
    }

    # This can take a little while...  Subtest to parallelize.
    @subtests=();
    foreach my $node (@$pair) {
        my $trans = $slobjs{TRANSIT}{$node};

        # If user wants to transit to CLI, and we are going to convert this node
        # then make sure we reboot before we go to the CLI
        if ( ($slobjs{FINAL_STATE}{$node} =~ /CLI/) && (!$slobjs{NOP}{$node}) ) {
            $slobjs{FINAL_STATE}{$node} = "FIRMWARE_THEN_CLI";
        }

#print "\n\nIn TRANSIT_TO_FIRMWARE block node = $node, state = ".$trans->get_state()."\n\n";
        if ( $trans->get_state( timeout => 900, maxidle => 900 ) !~ /^(FIRMWARE|MAINT)$/ ) {
            my $seq = ChangeModel::get_subtest_seq();
            my $st = subtest( sub { $trans->change_state( to => 'FIRMWARE',
                              inhibit_takeover => 1,
                              iftakenover      => 'takeback',
                              timeout          => 900,
                              maxidle          => 900 ); },
                               "-runid", $seq."_".$node."_transit_to_firmware",
                               "-bg",
                               "-onexit", sub { },
                      );
            push( @subtests, $st );
        }
    }

    ChangeModel::get_subtest_results( subtests => \@subtests, errors => \@errors );

    if ( @errors ) {
        goto FINAL_TRANSIT;
    }

    # depending on release, we may need to login as admin/netapp1! or naroot/<no password>
    my $sp_user_index = 0;
    my @spusers = qw( admin naroot );
    sub get_sp_user {
        if ( $sp_user_index >= scalar(@spusers) ) {
            $sp_user_index = 0;
        }
        my $rc = $spusers[$sp_user_index];
        $sp_user_index++;

        return $rc;
    }

    # This is fairly quick, no need to parallelize
    foreach my $node (@$pair) {
        $conn          = $slobjs{CONN}{$node};
        $fw_prompts    = $slobjs{PROMPTS}{$node};
        my $in_sp      = 0;

        # get_sp_user() below requires the sp_user_index
        $sp_user_index = 0;

        $slobjs{MIDPLANES}{$node}  = 0;

        if ( need_to_set_model_on_midplane( $slobjs{CURRENT_MODEL}{$node} ) ) {
            #   login to the sp to get fru list from service processor to see how many Midplanes we have
            eval {
                $output = $conn->execute( command => "\cG",
                                              match_table => [
                                                  'Service Processor Login: ' => \&get_sp_user,
                                                  'Password: '                => 'netapp1!',
                                              ],
                                          timeout => 120,
                                          prompts => $sp_prompts );
            };
    
            if ( $@  ) {
                $Log->warn("Could not enter the Service Processor with the expected username/pw, trying no password");
                $sp_user_index = 0;
                eval {
                    $output = $conn->execute( command => "\cG",
                                                  match_table => [
                                                      'Service Processor Login: ' => \&get_sp_user,
                                                      'Password: '                => '',
                                                  ],
                                              timeout => 120,
                                              prompts => $sp_prompts );
                };
                if ( $@  ) {
                    push( @errors, "Could not enter the Service Processor for node $node with no password, Reason: $@" );
                } else {
                    $in_sp = 1;
                }
            } else {
                $in_sp = 1;
            }
        }

        if ( !@errors ) {
            if ( $in_sp ) {
                eval {
                    $output = $conn->execute( command => "system fru list",
                                              match_table => [ 'ERROR' => \&_declare_error ],
                                              timeout => 120,
                                              prompts => $sp_prompts );
                };
                if ( $@  ) {
                    push( @errors, "Could not query the system fru list for node $node, Reason: $@" );
                } else {
                    if ( $_errors ) {
                        push( @errors, "Could not query the system fru list for node $node, error: $output" );
                    } else {
                        my $num_midplanes = 1;
                        if ( $output =~ /2\s+Midplane/ ) {
                            $num_midplanes = 2;
                        }
            
                        $slobjs{MIDPLANES}{$node}  = $num_midplanes;
                    }
                }
            }
        }

        if ( !@errors && $in_sp ) {
            # We could be in MAINT or FIRMWARE
            my @prompts = (@$fw_prompts, $ChangeModel::maint_mode_prompt );
            # Note, when we exit from Service Processor, we will not get back
            # the bootloader prompt.  Even a \r after exit won't work because exit
            # is the last thing processed by the SP, and we would need the \r to be
            # processed by the bootloader.  And since there is no prompt of any kind
            # to signify that that you are out, we have to just timeout
            eval {
                     $conn->execute( command => "exit",
                                     timeout => 5,
                                     prompts => \@prompts );
            };
    
            # timeout is expected
            if ( $@ && ($@ !~ /timeout/i) ) {
                push( @errors, "Could not exit node $node from the Service Processor, Reason: $@" );
            }

            $conn->discardbuff();
        }
    }

    # This can take a little while... Subtest to parallelize
    @subtests=();
    foreach my $node (@$pair) {
        $conn          = $slobjs{CONN}{$node};
        $trans         = $slobjs{TRANSIT}{$node};
#print "\n\nIn BOOT_DIAGS block node = $node, state = ".$trans->get_state()."\n\n";
        if ( $trans->get_state( timeout => 900, maxidle => 900 ) eq 'FIRMWARE' ) {
            my $seq = ChangeModel::get_subtest_seq();
            my $st = subtest( sub { $conn->execute( command => "\r$boot_diags_cmd",
                                                    timeout => 900,
                                                    match_table => [
                                                        'Invalid command:' => $boot_diags_cmd,
                                                        'Continue with boot?' => 'yes',
                                                    ],
                                                    prompts => [$ChangeModel::maint_mode_prompt] ); },
                                                       "-bg",
                                                        "-onexit", sub { },
                                                       "-runid", $seq."_".$node."_boot_diags" );
    
            push( @subtests, $st );
        }
    }

    ChangeModel::get_subtest_results( subtests => \@subtests, errors => \@errors );

    if ( @errors ) {
        goto FINAL_TRANSIT;
    }

    foreach my $node (@$pair) {
        $conn          = $slobjs{CONN}{$node};
        $trans         = $slobjs{TRANSIT}{$node};

        $conn->execute(
            command     => "",
            timeout     => 900,
            match_table => [ 'Continue with boot?' => 'yes', ],
            prompts     => [$ChangeModel::maint_mode_prompt],
        );

        if ( $trans->get_state( timeout => 900, maxidle => 900 ) ne 'MAINT' ) {
            push( @errors, "Could not boot $node into Maintenance mode" );
        }
    }

    if ( @errors ) {
        goto FINAL_TRANSIT;
    }

    foreach my $node (@$pair) {
        $conn = $slobjs{CONN}{$node};

        if( $promoted ) {
            eval {
                $conn->execute(
                    command     => 'sldiag oem',
                    match_table => [ 'Please enter oem mode password:' => 'sldiagOEMon1' ],
                    timeout     => 120,
                    prompts     => [ 'OEM mode is enabled' ]
                );
            };
            if ( $@ ) {
                $Log->error( "Could not set sldiag oem: $@");
            }
        }

        if ( !@errors ) {
            eval {
                sldiag_login_mfg( conn => $conn );
            };

            if ( $@ ) {
                push( @errors, "Could not login node $node to sldiag mfg mode, Reason: $@" );
            }
        }

        if ( @errors ) {
            goto FINAL_TRANSIT;
        }
    }

    foreach my $node (@$pair) {
        my $conn          = $slobjs{CONN}{$node};
        my $target_model  = $slobjs{TARGET_MODEL}{$node};
        my $num_midplanes = $slobjs{MIDPLANES}{$node};

        my $retries = 3;
RETRY:
        $retries--;

        my $sldiag_system_model_executed;
        if( $sldiag_system_model_enabled ) {
            $conn->discardbuff();
            $output = $conn->execute(
                command     => "sldiag system model",
                match_table => [
                    'Choice' => '-1',
                    'ERROR' => \&_declare_error ],
                prompts     => [ $ChangeModel::maint_mode_prompt ]
            );
            # Only pull choices from output
            $output =~ s/(?s).*Choose from://;
            my $value;
            foreach my $line ( split(/\n/, $output) ) {
                if( $line =~ /$target_model/ ) {
                    $line =~ s/^\s+//;
                    $value = substr($line, 0, 1);
                    last;
                }
            }
            $output = $conn->execute(
                command     => "sldiag system model",
                match_table => [
                    'Choice' => $value,
                    'Proceed' => 'yes',
                    'ERROR' => \&_declare_error ],
                prompts     => [ $ChangeModel::maint_mode_prompt ]
            );
            $output = $conn->execute(
                timeout => 500,
                command => "halt",
            );
            $sldiag_system_model_executed = 1;
        }

        if( ! $sldiag_system_model_executed ) {
            # Controller is number 0, midplanes are 1 & 2
            my $slot = $num_midplanes;
            while ( $slot >= 0 ) {
                $conn->discardbuff();
                $output = $conn->execute( command => "sldiag system modify $slot product_name $target_model",
                                match_table => [ 'ERROR' => \&_declare_error,
                                                 'SLDIAG: the model name specified is not valid' => \&_declare_error,
	                                       ],
                                prompts => [$ChangeModel::maint_mode_prompt] );
                if ($_errors) {
                    if ( $output =~ /the model name specified is not valid/i ) {

                        # Looks like an SN or later build
                        my $system_model_output;
                        eval {
                            $system_model_output = $conn->execute(
                                command     => "sldiag system model",
                                match_table => [ 'Choice:' => '-1', ],
                                prompts => [ $ChangeModel::maint_mode_prompt, ],
                            );
                        };

                        if ($@) {
                            die "Command: \"sldiag system model\" failed";
                        }

                        my $reported_model;
                        if ( $system_model_output =~ /Current model:\s+(\S+)/i ) {
                            $reported_model = $1;
                        }

                        $Log->comment("Current model name: $reported_model");

                        my @lines = split( '\n', $system_model_output );
                        my @matches = map {
                            if ( $_ =~ /\s\d\)\s+([FAS|IBM]\S+)/i ) { $1; }
                        } @lines;
                        my @matched_names;
                        my $target_model_name = undef;

                        $Log->comment("Matched names from 'sldiag system model': @matches");

                        if ( $target_model =~ /netapp/i ) {
                            @matched_names = grep {/([FAS]\S+)/i} @matches;
                            if (@matched_names) {
                                $target_model_name = $matched_names[0];
                            }
                        } elsif ( $target_model =~ /ibm/i ) {
                            @matched_names = grep {/([IBM]\S+)/i} @matches;
                            if ( scalar(@matched_names) >= 1 ) {
                                # IBM 'filer' names are of the format IBM-dddd-CNd, which is what is needed
                                # instead of the IBM 'gateway' names, which are of the format IBM-dddd-DNd

                                foreach my $ibm_name (@matched_names) {
                                    if ( $ChangeModel::ibm_2_netapp{$ibm_name} ) {
                                        $target_model_name = $ibm_name;
                                        last;
                                    }
                                }
                                if ( !defined($target_model_name) ) {
                                    die
                                        "Error: could not match an IBM model long name:\n$system_model_output";
                                }
                            }
                        } elsif ( $target_model =~ /toggle/i ) {
                            if ( $reported_model =~ /FAS/i ) {
                                $target_model_name
                                    = $ChangeModel::netapp_2_ibm{$reported_model};
                            } elsif ( $reported_model =~ /IBM/i ) {
                                $target_model_name
                                    = $ChangeModel::ibm_2_netapp{$reported_model};
                            } else {
                                die
                                    "Error: could not determine the current model name:\n$system_model_output";
                            }
                        }

                        $Log->comment("Target model name to be used: $target_model_name");

                        $_errors = 0;
                        $output  = $conn->execute(
                            command =>
                                "sldiag system modify $slot product_name $target_model_name",
                            match_table => [ 'ERROR' => \&_declare_error ],
                            prompts => [$ChangeModel::maint_mode_prompt]
                        );
                        if ($_errors) {
                            die
                                "Command: \"sldiag system modify $slot product_name $target_model_name\" failed";
                        }
                    } else {
                        die
                            "Command: \"sldiag system modify $slot product_name $target_model\" failed";
                    }
                }
                $slot--;
            }


            $slot = $num_midplanes;
            my @product_name = ();
            while ( $slot >= 0 ) {
                $conn->discardbuff();
                $output = $conn->execute( command => "sldiag system commit $slot",
                                          match_table => [ 'ERROR' => \&_declare_error ],
                                          prompts => [$ChangeModel::maint_mode_prompt] );

                 if ( $_errors ) {
                     die "Command: \"sldiag system commit $slot\" failed with $output";
                 }

                $conn->discardbuff();
                $output = $conn->execute( command => "sldiag system show $slot",
                                          match_table => [ 'ERROR' => \&_declare_error ],
                                          prompts => [$ChangeModel::maint_mode_prompt] );
                 if ( $_errors ) {
                     die "Command: \"sldiag system show $slot\" failed with $output";
                 }

                $slot--;

                my $product_name;
                if ( $output =~ /product_name\s+:\s+([0-9a-zA-Z\-]+)/ ) {
                    $product_name = $1;
                    if ( !grep(/^$product_name$/, @product_name ) ) {
                        push( @product_name, $product_name );
                    }
                } else {
                    push(@errors, "Could not parse sldiag product_name for node $node");
                }
            }

            # There should be exactly one product name for all of the slots
            if ( scalar( @product_name ) != 1 ) {
                push(@errors, "One or more of the slots on node $node appears to have failed to change product_name ".Dumper(\@product_name));

                if ( $retries > 0 ) {
                    goto RETRY;
                } else {
                    goto BAIL;
                }
            }
        }
    }

    if ( @errors ) {
        goto BAIL;
    }

    # Currently Vespa requires that you login to the SP to cause the model name
    # to be refreshed so Ontap and Bootloader can see it.
    # Filed BURT 501891, they might change this requirement, but I'm not holding my breath.
    #
    # This is fairly quick, no need to parallelize
    foreach my $node (@$pair) {
        if ( requires_sp_model_refresh( $slobjs{CURRENT_MODEL}{$node} ) ) {
            $conn          = $slobjs{CONN}{$node};
            $fw_prompts    = $slobjs{PROMPTS}{$node};

            # get_sp_user() below requires the sp_user_index
            $sp_user_index = 0;

            # login to the sp to get refresh the SYS_MODEL so that loader and Ontap can see it
            eval {
                $output = $conn->execute( command => "\cG",
                                              match_table => [
                                                  'Service Processor Login: ' => \&get_sp_user,
                                                  'Password: '                => 'netapp1!',
                                              ],
                                          timeout => 120,
                                          prompts => $sp_prompts );
            };
    
            if ( $@  ) {
                $Log->warn("Could not enter the Service Processor with the expected username/pw, trying no password");
                $sp_user_index = 0;
                eval {
                    $output = $conn->execute( command => "\cG",
                                                  match_table => [
                                                      'Service Processor Login: ' => \&get_sp_user,
                                                      'Password: '                => '',
                                                  ],
                                              timeout => 120,
                                              prompts => $sp_prompts );
                };
                if ( $@  ) {
                    push( @errors, "Could not enter the Service Processor for node $node with no password, Reason: $@" );
                }
            }
    
            if ( @errors ) {
                push( @errors, "Was not able to refresh the acp model for node $node" );
                push( @errors, "Please login to the SP for both nodes ".$pair->[0].",".$pair->[1]." and execute the following:" );
                push( @errors, "    \"priv set diag; system acp model refresh\"" );
                goto BAIL;
            } else {
                eval {
                    $output = $conn->execute( command => "priv set diag; system acp model refresh",
                                              timeout => 120,
                                              prompts => $sp_prompts );
                };
                if ( $@  ) {
                    push( @errors, "Error trying to refresh the SYS_MODEL via the SP for node $node, Reason: $@" );
                }
            }

            if ( @errors ) {
                push( @errors, "Was not able to refresh the acp model for node $node" );
                push( @errors, "Please login to the SP for both nodes ".$pair->[0].",".$pair->[1]." and execute the following:" );
                push( @errors, "    \"priv set diag; system acp model refresh\"" );
                goto BAIL;
            } else {
                # We are most likely in MAINT but could be in FIRMWARE
                my @prompts = (@$fw_prompts, $ChangeModel::maint_mode_prompt );
                # Note, when we exit from Service Processor, we will not get back
                # the bootloader/maint prompt.  Even a \r after exit won't work because exit
                # is the last thing processed by the SP, and we would need the \r to be
                # processed by the bootloader/maint.  And since there is no prompt of any kind
                # to signify that that you are out, we have to just timeout
                eval {
                         $conn->execute( command => "exit",
                                         timeout => 5,
                                         prompts => \@prompts );
                };
        
                # timeout is expected
                if ( $@ && ($@ !~ /timeout/i) ) {
                    push( @errors, "Could not exit node $node from the Service Processor, Reason: $@" );
                }
                $conn->discardbuff();
            }
        }
    }

FINAL_TRANSIT:

    if ( $final_state ) {
        # This can take a little while...  Subtest to parallelize.
        @subtests=();
        foreach my $node (@$pair) {
            my $fstate = $slobjs{FINAL_STATE}{$node};
                $conn  = $slobjs{CONN}{$node};
                $trans = $slobjs{TRANSIT}{$node};
            my $seq = ChangeModel::get_subtest_seq();
            my $st = subtest( sub { ChangeModel::transit_to_final( transit_obj => $trans, conn_obj => $conn, state => $fstate ); },
                               "-runid", $seq."_".$node."_transit_to_final",
                               "-bg",
                               "-onexit", sub { },
                      );
            push( @subtests, $st );
        }
    
        ChangeModel::get_subtest_results( subtests => \@subtests, errors => \@errors );
        if ( !@errors ) {
            # Validate that the final model is what we expect
            my @final_models;
            foreach my $node (@$pair) {
                $conn          = $slobjs{CONN}{$node};
                $trans         = $slobjs{TRANSIT}{$node};
                $nodehost      = $slobjs{HOST}{$node};
                my $final_model = ChangeModel::get_current_model( transit_obj => $trans, conn => $conn, host => $nodehost );
                if ( !grep(/^$final_model$/, @final_models ) ) {
                    push( @final_models, $final_model );
                }
            }

            # There should be exactly one product name for all of the nodes
            if ( scalar( @final_models ) != 1 ) {
                push(@errors, "One or more of the nodes appears to have a different model name ".Dumper(\@final_models));
            } else {
                $Log->comment("All nodes are reporting model name |$final_models[0]|");
            }
        }
    }

BAIL:

    # using sldiag to change the model can mess up the filer if you don't do this
    # correctly (i.e. the midplane and controllers have to match for both nodes)
    # so sometimes we goto BAIL and not transit (i.e. reboot)

    if ( @errors ) {
        die fail_result("$whoami - ".join("\n", @errors));
    }
    return 1;
}

############################################################################
# Subroutine: sldiag_show_cmd
# Parameters: 
#   conn - NATE ConnectRec object reference
#   cmd  - sldiag command string to execute
# Returns: Output string
# Purpose: The sldiag commands often fork off the processes that process
#          commands.  Being it does this, you can often get back some data,
#          the command prompt, then some more data.  Normally CLIs send back
#          all of the data (especially for show commands) and then return the
#          prompt, which signifies that the command output is complete.
#          Instead, we send the "sldiag version" command (which appears to
#          run in the foreground) to act as our end of command marker.
#          Calls to "discardBuff(), labelled as "Palate cleanser" are
#          done to discard everything that is currently in the "say" buffers
#          so that we start out each step with a clearn I/O buffer.
#############################################################################
sub sldiag_show_cmd {
    my $whoami = (caller(0))[3];
    my %opt = (
                  conn => undef,
                  cmd  => undef,
                  @_
              );

    my $conn = $opt{conn};
    my $cmd  = $opt{cmd};

    if ( !$conn || !$cmd ) {
        die script_result("$whoami - Must provide conn and cmd options");
    }

    my $output = "";

    # Palate cleanser
    $conn->discardbuff();

    # SLDIAG often threads off the command, so the prompt can come back before
    # all of the output is finished.
    if ( $cmd =~ /sldiag version/ ) {
        $output = $conn->execute( command => $cmd,
                                  prompts => [$ChangeModel::maint_mode_prompt] );
    } else {
        $output = $conn->execute( command => $cmd,
                                  match_table => [
                                      $ChangeModel::maint_mode_prompt => 'sldiag version',
                                  ],
                                  prompts => ['System Level Diagnostic Version'] );
        $output =~ s/sldiag version//g;
    }

    $output =~ s/$ChangeModel::maint_mode_prompt //g;
    $output =~ s/$ChangeModel::maint_mode_prompt//g;
    #$output =~ s/\*> //g;
    #$output =~ s/\*>//g;

    # get rid of blank lines
    my @lines = split( /\n/, $output );
    my @tmp;
    foreach my $line (@lines) {
        if ( $line ) {
            push( @tmp, $line );
        }
    }

    $output = join( "\n", @tmp );

    return $output;
}

sub sldiag_login_mfg {
    my $whoami = (caller(0))[3];
    my %opt = (
                  conn => undef,
                  @_
              );

    my ($x1, $x2, $x3, $x4) = (0,0,0,0);
    my $conn = $opt{conn};

    if ( !$conn ) {
        die script_result("$whoami - Must provide conn option");
    }

    sldiag_show_cmd( conn => $conn, cmd => "sldiag" );

    my $cmd1        = pack('H*','736c646961672076657273696f6e');
    my $cmd2        = pack('H*','736c646961672064617465');
    my $cmd3        = pack('H*','736c64696167206d6667');
       $x1          = pack('H*','637479');
       $x3          = pack('H*','6d6667');
    my $cmd1_regex  = pack('H*','53797374656d204c6576656c20446961676e6f737469632056657'.
                                '273696f6e20285b612d7a412d5a5c2e302d395d2b29');
    my $cmd2_regex  = pack('H*','5b302d395d5b302d395d3a285b302d395d2b293a285b302d395d2b29');
    my $cmd3_regex  = pack('H*','506c6561736520656e746572206d616e75666163747572696e672'.
                                '06d6f64652070617373776f72643a');
    my $cmd3_match  = pack('H*','4d616e75666163747572696e67206d6f646520697320656e61626c6564');
    my $x5_limit    = pack('H*','3535');
    my $x5          = $x5_limit + 1;

    $Log->debug("cmd1($cmd1), cmd2($cmd2), cmd3($cmd3), cmd1_regex($cmd1_regex), ".
                "cmd2_regex($cmd2_regex), cmd3_regex($cmd3_regex), x1($x1), x3($x3)");

    my $cmd1_out = sldiag_show_cmd( conn => $conn, cmd => $cmd1 );

    if ( $cmd1_out =~ /$cmd1_regex/ ) {
        $x2 = $1;
        $x2 = join("", split(/\./, $x2));

        my $retries = 5;
        while ( $retries && ($x5 > $x5_limit) ) {
            my $cmd2_out = sldiag_show_cmd( conn => $conn, cmd => $cmd2 );

            if ( $cmd2_out =~ /$cmd2_regex/ ) {
                $x4 = $1;
                $x5 = $2;
            }

            if ( $x5 > $x5_limit ) {
                Tharn::sleep( 10 );
            } else {
                last;
            }
            $retries--;
        }

        my $string = $x1.$x2.$x3.$x4;

        # Palate cleanser
        $conn->discardbuff();

        $conn->execute( command     => $cmd3,
                        log_level   => "debug",
                        must_match  => $cmd3_match,
                        match_table => [
                            $cmd3_regex => $string,
                        ],
                        prompts => [$ChangeModel::maint_mode_prompt] );
    }
}

1;
