#Hey emacs, this is basically -*- perl -*-
# $Id$

#***********************************************************************
#
# NAME
#   ChangeModel::SysDiag
#
# SYNOPSIS
#   Rebrand a NetApp filer to be IBM, or vice versa, via sysdiag.
#
# DESCRIPTION
#
# AUTHOR
#   dl-coati-core
#***********************************************************************

package ChangeModel::SysDiag;

use strict;
use Tharn qw($Log fail_result script_result);
use Data::Dumper;


# Tried to abstract out as much of the platform differences
# to this part of the script to make maintenance easier.
my $main_exit_cmd           = "exit";
my $sub_menu_exit_selector  = "99";
my $sys_info_setup_selector = $sub_menu_exit_selector;
my $write_mb_model_selector = $sub_menu_exit_selector;
my $ok_to_save              = ' OK to save\?  Press \'y\' to confirm\.';

my $diag_main_menu_prompt   = 'Enter Diag, Command or Option:';
my $diag_sub_menu_prompt    = 'Select test or feature by number ';
my $enter_bp_serial         = 'Enter Backplane serial# ';
my $enter_bp_rev            = 'Enter Backplane rev ';
my $enter_bp_part           = 'Enter Backplane part# ';

# The enter_prod options are for Shuar
my $enter_prod_man          = 'Enter Product manufacturer ';
my $enter_prod_partnum      = 'Enter Product part number ';
my $enter_prod_serial       = 'Enter Product serial number ';
my $enter_prod_ver          = 'Enter Product version ';

# Enter Product name is Shuar-specific
my $choose_sys_model        = ['Choose system model\s+\[\S+\]', 'Enter Product name\s+\[\S+\]'] ;

my $boot_diags_cmd          = 'boot_diags';
my $mfg_mode_cmd            = 'ctmfgy';
my $motherboard_cmd         = 'mb';

my $shuar_models_regex = qr(FAS2020|FAS2040|FAS2050|IBM-2859-A|IBM-2862-A);

# Shuar has 'serial' in there and no "menu" at the end...
my $sysinfo_menu_regex    = qr/(\d+):\s+System (serial info|info) setup/;

# 3070-Backplane, 6080-Bkplane, "Write Product Info" is Shuar
my $write_backplane_regex = qr/(\d+):\s+Write\s+(Backplane|Bkplane|Product Info)/;
my $system_model_regex    = qr/System model\s+(\S+),/;
my $model_selector_regex  = qr/\s?(\d+):\s+(\S+)/;


## @change_model()
## @description Routine which will attempt to change the Model of a NetApp
##              filer to be that of it's IBM equivalent, or vice versa.
##
## @arg node                 Required: node name
## @arg requested_model_type Required: netapp|ibm|toggle
## @arg final_transit_state  Optional: Any NACL Transit.pm state
##                                         (i.e. "FIRMWARE", "CLI", etc...)
##
## @synopsis
##
##    ChangeModel::change_model( node=>$node, requested_model_type=>"toggle",
##                               final_transit_state => "FIRMWARE" );
##
## @return       :  1|die
##------------------------------------------------------------------------------
sub change_model {
    my $whoami = (caller(0))[3];
    my %opt = (
                  node                 => undef,
                  requested_model_type => undef,
                  final_transit_state  => undef,
                  @_
              );

    my $node        = $opt{node};
    my $model_type  = $opt{requested_model_type};
    my $final_state = $opt{final_transit_state};

    if ( !$node || !$model_type ) {
        die script_result("$whoami - Must specify node and requested_model_type options");
    }

    my ($current_model, $target_model) = ( "", "" );
    my @errors;
    my $rc = 0;
    my $shuar = 0;
    my %model_selector;
    my $output;

    my $conn = Tharn::connect($node.'#console');
    my $host = Tharn::host( $node );

    my $fw_prompts = $host->prompt_fw();

    $Log->debug("$node FIRMWARE PROMPTS = ".Dumper($fw_prompts));

    my $trans = NACL::Transit->new(name => $conn);

    # Running into problems where Transit doesn't seem to handle filers that are in
    # weird states very well.  Use gx_boot to cold_shutdown the node
    ChangeModel::cold_shutdown( nodes => [$node] );

    # Get the current model, if possible, by doing minimal transits
    $current_model = ChangeModel::get_current_model( transit_obj => $trans, host => $host, conn => $conn );

    if ( !$current_model ) {
        my $initial_transit_retries=3;
INITIAL_TRANSIT:
        $initial_transit_retries--;
        eval {
            $trans->change_state( to => 'FIRMWARE', 
                                  inhibit_takeover => 1,
                                  iftakenover => 'takeback',
                                  timeout => 900,
                                  maxidle => 900  );
        };
        if ( $@ ) {
            if ( $initial_transit_retries > 0 ) {
                $Log->warn("Error transiting to FIRMWARE $@ retrying in 2 minutes");
                Tharn::sleep( 120 );
                goto INITIAL_TRANSIT;
            } else {
                $Log->error("Error transiting to FIRMWARE $@");
                goto FINAL_TRANSIT;
            }
        }
        $current_model = ChangeModel::get_current_model( transit_obj => $trans, host => $host, conn => $conn );
        if ( !$current_model ) {
            push(@errors, "Could not find current model for $node");
            goto FINAL_TRANSIT;
        }
    }

    if ( $final_state =~ /CLI/ ) {
        $final_state = "FIRMWARE_THEN_CLI";
    }

    # Target Model of "" means we already ARE this model and there is no need to change
    eval {
        $target_model = ChangeModel::target_model( current_model => $current_model,
                                                   requested_model_type  => $model_type );
    };

    if ( !$@ ) {
        if ( !$target_model ) {
            $Log->comment("Node: $node, Current model: $current_model, Target: $model_type is a NOP");

            # Could be "FIRMWARE_THEN_CLI", if this is a nop, no point in rebooting
            if ( $final_state =~ /CLI/ ) {
                $final_state = "CLI";
            }

            goto FINAL_TRANSIT;
        } else {
            $Log->comment("Node: $node, Current model: $current_model, Target: $target_model");
        }
    } else {
        push(@errors, $@);
        goto FINAL_TRANSIT;
    }

    # We may already be here, but if we were at the CLI, we could have retrieved the model
    # from there to check if we needed to transit at all.
    $trans->change_state( to => 'FIRMWARE', 
                          inhibit_takeover => 1,
                          iftakenover => 'takeback',
                          timeout => 900,
                          maxidle => 900  );

    # Boot Diags and get into the motherboard sub-menu
              $conn->execute( command => $boot_diags_cmd,
                              prompts => [$diag_main_menu_prompt] );

              $conn->execute( command => $mfg_mode_cmd,
                              prompts => [$diag_main_menu_prompt] );

    $output = $conn->execute( command => $motherboard_cmd,
                              prompts => [$diag_sub_menu_prompt] );

    if ( $output =~ $sysinfo_menu_regex ) {
        $sys_info_setup_selector = $1;
    } else {
        push(@errors, "Could not find System info setup menu in Diags for node $node");
        goto EXIT_DIAGS;
    }

    $output = $conn->execute( command => $sys_info_setup_selector,
                              prompts => [$diag_sub_menu_prompt] );

    if ( $output =~ $write_backplane_regex ) {
        $write_mb_model_selector = $1;
    } else {
        push(@errors, "Could not find option to Write the Backplane Model Number in Diags for node $node");
        goto EXIT_DIAGS;
    }

    $output = $conn->execute( command => $write_mb_model_selector,
                              match_table => [
                                  $enter_prod_man     => '',
                                  $enter_prod_partnum => '',
                                  $enter_prod_serial  => '',
                                  $enter_prod_ver     => '',
                                  $enter_bp_serial    => '',
                                  $enter_bp_rev       => '',
                                  $enter_bp_part      => '',
                              ],
                              prompts => $choose_sys_model );

    # Shuar's diag menu works differently than the others, you specify the string, not
    # a selection number
    if ( $current_model =~ $shuar_models_regex ) {
        $shuar = 1;
    }

    # I suppose it is possible that the firmware and diags might think they are
    # a different model, and SHUAR actually doesn't list the available models
    if ( $output =~ $system_model_regex ) {
        if ( $current_model ne $1 ) {
            push( @errors, "Firmware or Ontap reported model($current_model) != Diags reported model($1)\n".
                           "Using Diags reported model");
            $current_model = $1;

            # Target Model of "" means we already ARE this model
            # i.e. we are FAS3070 and model_type was 'netapp'
            eval {
                $target_model = ChangeModel::target_model( current_model => $current_model,
                                                           requested_model_type  => $model_type );
            };
            if ( !$@ ) {
                if ( !$target_model ) {
                    # we are already in the diags menu, so will go ahead with the set command
                    $Log->comment("Current model: $current_model, Target: $model_type is a NOP");
                } else {
                    $Log->comment("Current model: $current_model, Target: $target_model");
                }
            } else {
                push(@errors, $@);
                $target_model = "";
            }

        }
        # Build model selector hash so we can select the proper model from the diags
        # menu.
        my @lines = split(/\n/, $output);
        foreach my $line ( @lines ) {
            if ( $line =~ $model_selector_regex ) {
                $model_selector{$2} = $1;
            }
        }
    } elsif ( !$shuar ) {
        # Not fatal, we still have the FIRMWARE current model, but error should be reported, or
        # if there are legitemate diag menu variants, this script needs to be updated to support.
        push(@errors, "Firmware or Ontap reported model($current_model), Could not parse the Diags model for node $node");
    }

    # Even if you specify a SPECIFIC_TARGET_MODEL string, that string has
    # to be in the diags menu, or we can't select it.
    if ( !$shuar && !defined $model_selector{$target_model} ) {
        # Sort of fatal, can't continue with change, but don't want to leave filer in the
        # diags menu...
        push(@errors, "Cannot find menu selector for target model($target_model), setting target($current_model) for node $node");

        # Leave the model as it is (you really can't exit out of the 'Choose system model" question with
        # a 99, you have to give a valid number, so we'll pick the model we currently are.)
        $target_model = $current_model;
    }

    my $target_model_selector = $target_model;

    # Shuar just uses the name, not a selection number
    if ( !$shuar ) {
        if ( defined $model_selector{$target_model} ) {
            $target_model_selector = $model_selector{$target_model};
        } else {
            push(@errors, "Cannot find menu selector for target model($target_model) for node $node");
        }
    }

    $Log->debug("Model Selector $target_model_selector, hash = ".Dumper(\%model_selector));

    my $save_response = 'y';

    if ( @errors ) {
        $save_response = 'n';
    }
    $output = $conn->execute( command => $target_model_selector,
                              match_table => [
                                  $enter_bp_serial    => '',
                                  $enter_prod_man     => '',
                                  $enter_prod_partnum => '',
                                  $enter_prod_serial  => '',
                                  $enter_prod_ver     => '',
                                  $enter_bp_rev       => '',
                                  $enter_bp_part      => '',
                                  $ok_to_save         => $save_response,
                              ],
                              prompts => [$diag_sub_menu_prompt] );
EXIT_DIAGS:

    # exit all the way out of diags
    $output = $conn->execute( command => $sub_menu_exit_selector,
                              timeout => 900,
                              match_table => [
                                  $diag_sub_menu_prompt  => $sub_menu_exit_selector,
                                  $diag_main_menu_prompt => $main_exit_cmd,
                              ],
                              prompts => $fw_prompts );

    # If we have already reported an error, we have not attempted the model change
    if ( !@errors ) {
        # Verify LOADER SYS_MODEL after change
        my $new_model = ChangeModel::get_loader_sys_model( host => $host, conn => $conn );
        if ( $new_model ) {
            # If target_model is not "", then we should have changed it.
            if ( $target_model && ($target_model ne $new_model) ) {
                push(@errors, "Attempt to change model to $target_model does not appear to be reflected in FIRMWARE($new_model) for node $node");
            }
        } else {
            push(@errors, "Could not find SYS_MODEL in ENV of FIRMWARE for node $node");
        }
    }

FINAL_TRANSIT:

    if ( $final_state ) {
        ChangeModel::transit_to_final( transit_obj => $trans, conn_obj => $conn, state => $final_state );
    }

    if ( @errors ) {
        die fail_result("$whoami - Issues reported: ".join("\n", @errors));
    }

    return 1;
}

1;
