# $Id: $
#
## @summary  Lib Checker module
## @author   dl-dpgqa-automation@netapp.com
## @status   Private
## @pod      here
## @description
## Checker module has routines that are being used by the checker correctness
## and the Checker error injection testcases

=head1 NAME

Checker.pm

=head1 AUTHOR

Dinesh Jos

=head1 DESCRIPTION

This module implements all of the checker fault injection routines

=head1 EXAMPLES

No examples needed. examples provided per routine

=cut


$| = 1; # flushes the buffers, Please comment this line out if
        # if there is timing issues when using this library.
use strict;
package Checker;

#####################################################
## package dp::snapmirror::paloma::Checker;
## non-standard way to declare a pkg name. add the lib
## path to the .ntest file. normally in perl you would
## use the "use lib" and specify the path in nate you
## should do this through the .ntest file
#####################################################
use GNOTE;
use GNOTE::GX;
use GNOTE::Volume;
use Server;
use Carp;   # confess and croak are loaded by default, no need for qw(confess)
use Switch;

use base qw( Exporter );

## "get_inode_info_orig" contains the get_inode dblade command implementation
## the out from the routine is needed and used by mod_inode and send_holes routines.
## the routine will be removed after they are re-implemented using wex
our @EXPORT_OK = qw(
        get_checker_logs
        get_inode_info
        get_inode_info_orig
        mod_inode
        punch_holes
        send_blocks
        parse_checker_logs
        delete_old_chkr_logs
	mod_dir
        );

=head1 ROUTINES

=back

=head2 DELETE OLD CHECKER LOGS

=item Routine name:
delete_old_chkr_logs

=item Description:

    Delete the old checker logs and the directory in which it was located

=item Arguments:

    vol_obj = volume object of type GNOTE
    test_dir = path to the file ( dir path ) if file is in a dir/subdir
    test_file = name of the file

=item Examples

    my $inode_info = get_inode_info(vol_obj => $vol_dst, test_dir => $test_dir, test_file => $test_file);
    my $inode_info = get_inode_info(vol_obj => $vol_dst, test_file => $test_file);

=item Return values

    Returns a hash refrence containing the inode info for success
    Return 0 for failure

=back

=cut

sub delete_old_chkr_logs
{
    NOTE::logcomment(" INFO -> entering delete_old_chkr_logs");
    my %args = @_;
    my $log_loc = "/mroot/etc/log/.snapmirrorChecker";
    my $filer;
    my $server_obj  = Server->new;
    my $origin_node;

    my @filer_list = $server_obj->servers;

    #-----------------------------------------------#
    # Delete all checker logs that are on the filer #
    #-----------------------------------------------#
    $origin_node = $server_obj->set_execution_server(-Server => $filer_list[0]);
    foreach $filer(@filer_list)
    {
            $server_obj->set_execution_server(-Server => $filer);
            my $shell_output = $server_obj->shell( cmd => "rm -rf $log_loc" );
    }
    $server_obj->set_execution_server(-Server => $origin_node);
    return 1;
}

=head2 GET INODE INFO

=item Routine name:
get_inode_info

=item Description:

    Get the contents of the inode for the file specified
    and return the contents of the inode in the form of a hash

=item Arguments

    vol_obj = volume object of type GNOTE
    test_dir = path to the file (dir path) if file is in a dir/subdir
    test_file = name of the file or diskid.inode pair ( see examples )
    NOTE : if you need the inode of directory pass the dir name as test_file argument

=item Examples

    my $inode_info = get_inode_info(vol_obj => $vol_dst, test_dir => $test_dir, test_file => $test_file);
    my $inode_info = get_inode_info(vol_obj => $vol_dst, test_file => $test_file);
    my $inode_info = get_inode_info(vol_obj => $vol_dst, test_file => $dir_name);
    my $inode_info = get_inode_info(test_file => $test_file);
    where test_file should be of format 1234.34 1234 being the disk id and 34 being the inode

=item Return values

    returns a hash refrence containing the inode info for success
    return 0 for failure

=item CLI arguments

Sierra Nevada build deprecates use of zsmcli command form making of destination volume writable for wex
command and requires use of '-noredirect' flag instead. Use B<SNDSTRW=1> or B<SN_DST_RW=1> cli argument for
all Sierra Nevada builds starting June 2012. See BURT 608802 for details.

=back

=cut

sub get_inode_info
{
    NOTE::logcomment(" INFO -> inside the routune get_inode_info");
    my %args = @_;
    my %inode_vals  =();    # hash with contents of inode init to empty
    my $vol_obj    = exists $args{vol_obj} ? $args{vol_obj} : undef ; # obj of type vol
    my $test_dir    = exists $args{test_dir} ? $args{test_dir} : undef; # test directory
    my $test_file    = exists $args{test_file} ? $args{test_file} : return 0; # name of the test file

    my ($server,       # server object
        $vol_name,     # volume name
        $file_inode,   # test file inode (on DP vol)
        $junct_inode,  #inode of the junction
        @inode_dump    # temp array to hold the inode dump
    );

    my $sndstrw = (Tharn::param('SNDSTRW') || Tharn::param('SN_DST_RW')) ? '-noredirect' : undef;	# BURT 608802

    $server = Server->new;

    # get the volume name
    if ( $vol_obj )
    {
        $vol_name = $vol_obj->name;

        eval{
            if ( $test_dir ) {
                $file_inode = $server->shell( cmd => "wex inode ${vol_name}${test_dir}/$test_file -dump all $sndstrw" );
            } else {
                $file_inode = $server->shell( cmd => "wex inode $vol_name/$test_file -dump all $sndstrw" );
            }
        };
        confess "$@" if $@;
    } else {
       eval {
           $file_inode = $server->shell( cmd => "wex inode $test_file -dump all $sndstrw" );
       };
       confess "$@" if $@;
    }


    # extracting the inode of the junction
    @inode_dump = split( /\012/, $file_inode); # spliting the output on new lines /\012/
    if ( $vol_obj ) {
        $junct_inode = shift(@inode_dump);         # pulling out the first line which contains the junt inode info
        shift(@inode_dump);
        $junct_inode = (split(/ /, $junct_inode))[1];
        $junct_inode =~ /\.(.*?)\//;
        $junct_inode = $1; # $1 is the special perl variable that holds the ret of the ( ) in the regxp
    }

    # build the hash for the inode values
    print "BURT 593949, more info needed:\n". $file_inode;
    %inode_vals = (
                $file_inode =~ /
                                (?:^|[\r\n]+)   # match new line sequence or absolute beginning of the line, but not tabs or spaces
                                \s*             # skip empty spaces, any spaces (tabs, newline, etc)
                                ([^=\r\n]+)     # capture anything of non '=' and non \r or \n nature
                                (?:[\011\040]*) # skip tabs and spaces, cannot use \s kz of newline character is \s too
                                =               # find the 'equality for all'
                                (?:[\011\040]*) # skip possible tabs or spaces but not new lines
                                ([^\r\n]+)      # capture rest of the line if
                                ?               # exists
                               /gsx		# mention extended RegEx of multiline nature with repetitive mode
              );


    if ($vol_obj ) {
        $inode_vals{junct_inode}= $junct_inode;
    }

    # checking if file has hard links if so return list of file names
    my $pos = 0;
    my $temp_var = join('=',@inode_dump);
    @inode_dump = split('=',$temp_var);
    delete $inode_vals{name};	# delete the key value holding a single name
    foreach (@inode_dump) {
       if($_ eq "name") {
           push( @ {$inode_vals {name} },$inode_dump[$pos + 1]);
       }
       $pos++;
    }

    return ( \%inode_vals);
}

=head2 MODIFY INODE CONTENTS

=item Routine name:
    mod_inode

=item Arguments

    vol_obj = (Required)volume object of type GNOTE
    test_dir = path to the file ( dir path ) if file is in a dir/subdir
    test_file = (Required)name of the file
    inode_chg = (Required) hash ref to containing the changes to the inode

=item Examples

	retval = mod_inode( vol_obj => $vol_dst, test_dir => $test_dir, test_file => $test_file, inode_chg => $inode_chg))
	retval = mod_inode( vol_obj => $vol_dst, test_file => $test_file, inode_chg => $inode_chg))

=item Return values

    returns 1 on success
    return 0 for failure

=item CLI arguments

Sierra Nevada build deprecates use of zsmcli command form making of destination volume writable for wex
command and requires use of '-noredirect' flag instead. Use B<SNDSTRW=1> or B<SN_DST_RW=1> cli argument for
all Sierra Nevada builds starting June 2012. See BURT 608802 for details.

=back

=cut

## POSSIBLE FIELDS OF CHANGE < PLEASE SEE WEX MANUAL SOME OPTIONS MAY NOT WORK >
##    id	size		epoch		atime
##    snapid	blockcount	indircount	mtime
##    inum	xinode		state		ctime
##    vbn	sinode		buftreeid	crtime
##    pvbn	inline		payload		location
##    dvmask	rdev		uid		generation
##    version	gid		used		name
##    type	i2p		fav		junction
##    subtype	owner		nlink		umask
##    level	tid		flags
##

sub mod_inode
{
    NOTE::logcomment(" INFO -> inside the routine mod_inode");
    use POSIX;

    my %args = @_;
    my $vol_obj = exists $args{vol_obj} ? $args{vol_obj} : return 0 ; # object of type volume
    my $test_dir = exists $args{test_dir} ? $args{test_dir} : undef ; # test directory
    my $test_file = exists $args{test_file} ? $args{test_file} : return 0; # name of the test file
    my $inode_chg = exists $args{inode_chg} ? $args{inode_chg} : return 0; # name of the test file

    my $sndstrw = (Tharn::param('SNDSTRW') || Tharn::param('SN_DST_RW')) ? '-noredirect' : undef;	# BURT 608802

    my ($server,       # server object
        $vol_name,     # dest volume name
        $vol_inst,     # dest volume instance to extract "Node name"
        $changes,      # list of changes to inode
        $ret_val,      # return of the mod_inode command
        $vol_uid,      # uuid of the volume
        $exec_node,    # node on which the vol resides.
        $origin_node,  # original node
        $snap_id,      # snapshot id
	$attrib        # key used to parse inode changes
        );
    my $is_file = 0;   # flag to check if the user passed a file name or inode number

    $server = Server->new;

    # get the dst volume name, uuid and node on which it is residing
    $vol_name = $vol_obj->name;
    $vol_inst = ($server->volume_show(volume => $vol_name))[0];
    $exec_node = $server->shell( cmd => "wex volume $vol_name -dump location" );
    $exec_node = (split( '=', $exec_node ))[1];
    $vol_uid = $vol_obj->uuid;


    NOTE::logcomment(" INFO -> vol_uid is ->$vol_uid<-");
    NOTE::logcomment(" INFO -> exec_node is ->$exec_node<-");

    # creating the command for get inode and extracting inode
# has been commented out since wex doesnt allow modifications of attributes after passing time attrib
# will be uncommented once the fix is done to wex
#    foreach (keys %$inode_chg) {
#        $changes .= "$_=$inode_chg->{$_},";
#    }
#    chop($changes);
#    NOTE::logcomment(" INFO -> changes is ->$changes<-");

    # Checking to see if file name or disk_id.inode num is passed?
    my $num;
    my $n_unparsed;
    $! = 0;
    ($num, $n_unparsed) = POSIX::strtod($test_file);
    if ( $n_unparsed ) {
       $is_file = 1;
    }


    # set execution to the node on which the volume resides since zsmcli
    # will run only on the local node
    $origin_node = $server->set_execution_server(-Server => $exec_node);

    unless( $sndstrw ) {	# BURT 608802
	# Get the redirect snapshot id
	$snap_id = zsmcli_changes( option => "get_snap_id", vol_uid =>$vol_uid );

	# Clear the data protection flag
	if (zsmcli_changes( option => "set_flag", value => "false", vol_uid => $vol_uid)){
	    return 0;
	}

	# Clear the redirect id to 0
	if (zsmcli_changes( option => "set_snap_id", vol_uid =>$vol_uid, value => 0 )){
	    return 0;
	}
    }

    # Use the wex command to alter the inode attribs
    if ( $is_file ) {
        eval{
            if ( $test_dir ) {
                foreach $attrib ( keys %$inode_chg) {
                    my $out = $server->shell( cmd => "wex inode ${vol_name}.${test_dir}/$test_file -set $attrib=$inode_chg->{$attrib} $sndstrw" );
		    confess $out if $out =~ /(unable to find name|unable to read data|error\s\d+|0 inodes changed)/si;
                }
            } else {
                foreach $attrib ( keys %$inode_chg) {
                    my $out = $server->shell( cmd => "wex inode ${vol_name}./${test_file} -set $attrib=$inode_chg->{$attrib} $sndstrw" );
		    confess $out if $out =~ /(unable to find name|unable to read data|error\s\d+|0 inodes changed)/si;
                }
            }
        };
        confess "$@" if $@;
    } else {
        eval {
            foreach $attrib ( keys %$inode_chg) {
                my $out = $server->shell( cmd => "wex inode $test_file -set $attrib=$inode_chg->{$attrib} $sndstrw" );
		die $out if $out =~ /(unable to find name|unable to read data|error\s\d+|0 inodes changed)/si;	# die is used due to eval, will be handled by confess...
            }
        };
        confess "$@" if $@;
    }

    unless( $sndstrw ) {    # BURT 608802
	# Set the data protection flag
	if (zsmcli_changes( option => "set_flag", value => "true", vol_uid => $vol_uid )){
	 return 0;
	}

	# set the redirect id to original id
	if (zsmcli_changes( option => "set_snap_id", vol_uid =>$vol_uid, value => $snap_id )){
	 return 0;
	}

	# set execution to back to the original node
	$server->set_execution_server(-Server => $origin_node);
	return 1;
    }

    return 1;

    # need to do an empty update after this. will be done in script
}

=head2 MODIFY DIR CONTENTS

=item Routine name:
    mod_dir

=item Synopsis :
 This routine is intented to cause changes in the directory by adding or deleteing files
 in the directory. This is a standin for the wex utity till we have one.
 If a list of file names is not provided the routine will add files appended by a number
 starting at 1, for as many files have been requested. if ref to list of name is provided
 then the list will be acted upon as requested

=item Arguments

    vol_obj = (Required)volume object of type GNOTE
    test_dir = path to the dir( dir path )
    change_type = add contents/delete contents [add|delete] (default : add)
    change_size = num of files to add/delete from the dir (default : 1)
    file_names  = (optional ) array refrence to list of files to delete/create

=item Examples

	mod_dir( vol_obj => $vol_dst, test_dir => "/src_dir", change_type => "delete", change_size => 1)
	mod_dir( vol_obj => $vol_dst, test_dir => $test_dir, change_type => "delete", file_names => $files)
	mod_dir( vol_obj => $vol_dst, test_dir => $test_dir, file_names => $files)
	mod_dir( vol_obj => $vol_dst, test_file => $test_file)

=item Return values
    will confess on failure with error string

=back

=cut

sub mod_dir
{
    NOTE::logcomment(" INFO -> inside the routine mod_dir");

    my %args = @_;
    my $vol_obj = exists $args{vol_obj} ? $args{vol_obj} : confess "vol_object needed "; # object of type volume
    my $test_dir = exists $args{test_dir} ? $args{test_dir} : undef ; # test directory
    my $change_type = exists $args{change_type} ? $args{change_type} : "add"; # name of the test file
    my $change_size = exists $args{change_size} ? $args{change_size} : 1; # name of the test file
    my $file_names = exists $args{file_names} ? $args{file_names} : undef; # name of the test file

    my ($server,       # server object
        $vol_name,     # dest volume name
        $vol_inst,     # dest volume instance to extract "Node name"
        $vol_uid,      # uuid of the volume
        $exec_node,    # node on which the vol resides.
        $origin_node,  # original node
        $snap_id,      # snapshot id
	$file          # file name that will be created/delete
        );
    $server = Server->new;
    $vol_name = $vol_obj->name;
    $exec_node = $server->shell( cmd => "wex volume $vol_name -dump location" );
    $exec_node = (split( '=', $exec_node ))[1];
    $vol_uid = $vol_obj->uuid;

    NOTE::logcomment(" INFO -> vol_uid is ->$vol_uid<-");
    NOTE::logcomment(" INFO -> exec_node is ->$exec_node<-");

    # convert the volume to a RW volume using zapi
    $origin_node = $server->set_execution_server(-Server => $exec_node);

    # Get the redirect snapshot id
    $snap_id = zsmcli_changes( option => "get_snap_id", vol_uid =>$vol_uid );

    # Clear the data protection flag
    if (zsmcli_changes( option => "set_flag", value => "false", vol_uid => $vol_uid)){
        confess "unable to set DP flag to false";
    }

    # Clear the redirect id to 0
    if (zsmcli_changes( option => "set_snap_id", vol_uid =>$vol_uid, value => 0 )){
        confess "unable to set snapid to 0";
    }

    # add files on the the directory
    if($change_type =~ /add/) {
        if($file_names){
            foreach (@$file_names) {
              $file = $test_dir . "/$_";
              $vol_obj->nfs->create("$file");
            }
        }
    else{
           for( my $i = 1 ; $i <= $change_size; $i++) {
              $file = $test_dir . "/file_$i";
              $vol_obj->nfs->create("$file");
           }
        }
    }

    # delete files from the the directory
    if($change_type =~ /delete/) {
        if($file_names){
            foreach (@$file_names) {
                $file = $test_dir . "/$_";
                eval { $vol_obj->nfs->unlink("$file"); };
                confess $@ if ($@);
            }
        }
        else{
           for( my $i = 1 ; $i <= $change_size; $i++) {
              $file = $test_dir . "/file_$i";
              $vol_obj->nfs->unlink("$file");
           }
        }
    }

    # convert the volume back to a DP volume using zapi
    # Set the data protection flag
    if (zsmcli_changes( option => "set_flag", value => "true", vol_uid => $vol_uid )){
        confess "unable to set DP flag back to true";
    }

    # set the redirect id to original id
    if (zsmcli_changes( option => "set_snap_id", vol_uid =>$vol_uid, value => $snap_id )){
        confess "unable to set snapid to original value";
    }

    # set execution to back to the original node
    $server->set_execution_server(-Server => $origin_node);

    # need to do an empty update after this. will be done in script
}

###############################
## zsmcli_changes
## needed due to burt 313255
## Description: make the zsmcli changes as per request
##    (Required)    volume id where changes are to be made
##    (Required)    option type for change or retriving info
##    (Required)    set value for flag of snapshot id
## output:
##    returns 0 or snapid on success
##    1 on failure
##############################
sub zsmcli_changes
{
    NOTE::logcomment(" INFO -> inside the zsmcli changes routine");
    my %args = @_;
    my $vol_uid = exists $args{vol_uid} ? $args{vol_uid} : undef;
    my $option = exists $args{option} ? $args{option} : undef;
    my $value = exists $args{value} ? $args{value} : 0;


    my $server = Server->new;
    my $cmd_output;
    my $ret_val = 1;

    #checking if the option is Flag set operation
    switch ($option) {
        case "set_flag"
        {
	    if( $value =~ /True/i){
		$cmd_output = $server->shell( cmd => "zsmcli d-volume-set-info id=$vol_uid desired-attrs=mirror-types-is-data-protection volume-attrs=\"[d-volume-info=[mirror-types=[d-mirror-types-info=[is-data-protection=true]]]]\"");
	    } elsif($value =~ /False/i){
		$cmd_output = $server->shell( cmd=> "zsmcli d-volume-set-info id=$vol_uid desired-attrs=mirror-types-is-data-protection volume-attrs=\"[d-volume-info=[mirror-types=[d-mirror-types-info=[is-data-protection=false]]]]\"");
	    }
        }

        case "get_snap_id"
        {
	    $cmd_output = $server->shell( cmd => "zsmcli d-volume-list-info id=$vol_uid desired-attrs=redirect-snapshot-id");

	    # extract the snap_id
	    $cmd_output = (split( /\012/, $cmd_output))[3];
	    my $r_pos = rindex ($cmd_output,"<");
	    my $l_pos = index ($cmd_output,">");
	    $l_pos++;
	    my $offs = $r_pos - $l_pos;
	    return(substr($cmd_output, $l_pos, $offs));
        }

        case "set_snap_id"
        {
	    $cmd_output = $server->shell( cmd => "zsmcli d-volume-set-info id=$vol_uid desired-attrs=redirect-snapshot-id volume-attrs='[d-volume-info=[redirect-snapshot-id=$value]]'");
        }
    }
    if ($cmd_output =~ /passed/) {
        $ret_val = 0;
    }
    return $ret_val;
}

=head2 PUNCH HOLES IN A FILE

=item Routine name:
    punch_holes

=item Description:

    punch holes in the file specified in the range of blocks provided

=item Arguments

    inode_info = (Required)reference to hash containing inode info
    start_blk =  start blk
    num_of_blks = num of blks each 4096 bytes

=item Examples

    punch_holes( inode_info => $inode_info, start_block => 0, num_of_blocks => 10 )

=item Return values

    returns 1 on success
    return 0 for failure

=back

=cut

sub punch_holes
{
    NOTE::logcomment(" INFO -> inside the routune punch_holes");
    my %args = @_;
    my %inode_vals =();    # hash with contents of inode initialize to empty
    my $file_id = exists $args{inode_info}{fileid} ?              # inode of the file
                         $args{inode_info}{fileid}:
                         return 0 ;
    my $generation = exists $args{inode_info}{generation} ?       # generation id of the file
                            $args{inode_info}{generation}:
                            return 0 ;
    my $fs_id = exists $args{inode_info}{fsid} ?                  # file system id
                       $args{inode_info}{fsid}         :
                       return 0 ;
    my $wafl_type = exists $args{inode_info}{wafl_type} ?         # wafl type
                           $args{inode_info}{wafl_type} :
                           return 0 ;
    my $node_name = exists $args{inode_info}{node_name} ?          # node on which the file exists
                           $args{inode_info}{node_name} :
                           return 0 ;
    my $start_blk = exists $args{start_block} ? $args{start_block} : 0;         # start blk
    my $num_of_blks = exists $args{num_of_blocks} ? $args{num_of_blocks} : 1;   # num of blks each 4096 bytes

    my $server = Server->new;
    my ($ret_val, $flag_val);

    $flag_val = _set_get_wafl_paloma_enable_flags($node_name,"get");

    _set_get_wafl_paloma_enable_flags($node_name,"set", 1);

    my $send_holes_cmd = "\"priv set test; send_holes -i $file_id " .
                        "-g $generation " .
                        "-f $fs_id " .
                        "-t $wafl_type ".
                        "-o $start_blk ".
                        "-s $num_of_blks ".
                        "\"";
    eval{
        $ret_val = $server->cli( cmd => "run -node $node_name -command $send_holes_cmd" );
    };
    confess "$@" if $@;

    _set_get_wafl_paloma_enable_flags($node_name,"set", $flag_val);

    if ( $ret_val =~ /Success/) {
        return 1;
    }
    else { return 0; }
}

## private routine to get or get the wafl_paloma_enable flag
sub _set_get_wafl_paloma_enable_flags{
    my ( $node_name, $flag_type, $val ) = @_;
    my $ret_flag;
    my $server = Server->new;
    my $show_flags_cmd = "\"priv set test; printflag wafl_paloma_enable\"";
    my $set_flags_cmd = "\"priv set test; setflag wafl_paloma_enable $val \"";
    if($flag_type =~ /set/i) {
	eval{ $server->cli( cmd => "run -node $node_name -command $set_flags_cmd" ); };
        confess "$@" if $@;
	return 1;
    }
    eval{ $ret_flag = $server->cli( cmd => "run -node $node_name -command $show_flags_cmd" ); };
        confess "$@" if $@;
    $ret_flag = (split(/=/, $ret_flag))[-1];
    return $ret_flag;
}

=head2 SEND BLOCKS IN A FILE

=item Routine name:
    send_blocks

=item Description:

    punch holes in the file specified in the range of blocks provided

=item Arguments

    inode_info = (Required)reference to hash containing inode info
    start_blk =  start blk : default 0
    num_of_blks = num of blks each 4096 bytes : default 1

=item Examples

    send_blocks( inode_info => $inode_info, offset => 0, bytes => 32768)

=item Return values

    returns 1 on success
    return 0 for failure

=back

=cut

sub send_blocks
{
    NOTE::logcomment(" INFO -> inside the routune send_blocks");
    my %args = @_;
    my %inode_vals =();    # hash with contents of inode initialize to empty
    my $file_id = exists $args{inode_info}{fileid} ?          # inode of the file
                         $args{inode_info}{fileid}:
                         return 1 ;
    my $generation = exists $args{inode_info}{generation} ?   # generation id of the file
                            $args{inode_info}{generation}:
                            return 1 ;
    my $fs_id = exists $args{inode_info}{fsid} ?              # file system id
                       $args{inode_info}{fsid}         :
                       return 1 ;
    my $wafl_type = exists $args{inode_info}{wafl_type} ?     # wafl type
                           $args{inode_info}{wafl_type} :
                           return 1 ;
    my $node_name = exists $args{inode_info}{node_name} ?     # node on which the file exists
                           $args{inode_info}{node_name} :
                           return 1 ;
    my $offset = exists $args{offset} ? $args{offset} : 0;    # startng offset
    my $bytes = exists $args{bytes} ? $args{bytes} : 1;       # number of bytes
    my $server = Server->new;
    my $ret_val;

    my $flag_val = _set_get_wafl_paloma_enable_flags($node_name,"get");

    _set_get_wafl_paloma_enable_flags($node_name,"set", 1);

    my $send_blks_cmd= "\"priv set test; send_blocks -i $file_id " .
                        "-g $generation " .
                        "-f $fs_id " .
                        "-t $wafl_type ".
                        "-o $offset ".
                        "-s $bytes ".
                        "\"";
    eval{
        $ret_val = $server->cli( cmd => "run -node $node_name -command $send_blks_cmd" );
    };
    confess "$@" if $@;

    _set_get_wafl_paloma_enable_flags($node_name,"set", $flag_val);
    if ( $ret_val =~ /Success/) {
        return 1;
    }
    else { return 0; }
}
=head2 PARSE CHECKER LOGS

=item Routine name:
    parse_checker_logs

=item Description:

    parse checker log and compare them with synthesized log.
    if the logs that are retrieved from the node matches the synthesized
    1 is returned else 0 is returned for all other errors -1 is returned

=item Arguments

    inode_info = (Required) Reference to hash containing inode info;
    node = (Required) node name on which the logs are present
    log_path = (Required) log location on the node ( path )
    log_type = (Required) log type ->type of coruption the log holds
    blocks = (Required) Reference to hash containing the data ranges

=item Examples

    (parse_checker_logs(log_type => "data",
                                blocks => \%ranges,
                                inode_info => $inode_info,
                                node => $ret_val->{node},
                                log_path => $ret_val->{path}))

=item Return values

    returns 1 on success
    return 0 for failure
    return -1 for othere failures

=back

=cut

sub parse_checker_logs
{
    NOTE::logcomment(" INFO -> entering parse_checker_logs");
    my %args = @_;
    my $inode_num = exists $args{inode_info} ? $args{inode_info}{fileid} : return -1;  #inode of the file
    my $node = exists $args{node} ? $args{node} : return -1;    # name of filer on whch the logs reside
    my $log_path = exists $args{log_path} ? $args{log_path} : return -1; # log location on node
    my $log_type = exists $args{log_type} ? $args{log_type} : return -1; # type of coruption the log holds
    my $ranges;    #reference to hash that contains the offsets for data coruption;
    if ($log_type eq 'data') {
        $ranges = exists $args{blocks} ? $args{blocks} : return -1;  #reference to hash with data ranges
    }

    my $orign_node;     # variable to hold the current execution node name
    my $offset;         # key to the ranges hash
    my %chkr_logs;      # hash to hold the sythesized and retrived error logs
    my $blk_cnt;        # number of blocks that have been altered
    my @ret_ary = ();   # array to hold the block offsets need to sythesize logs
    my $FLAG = 1;       # FLAG to determine a success or failure, will be set to 0 if logs dont match
    use constant block_size => 4096;

    my $server_obj    = Server->new;    # sever object


    # generate the log file based on error injected for data modification
    if ($log_type eq "data") {
        foreach $offset ( sort { $a <=> $b } keys %$ranges ) {
            push(@ret_ary, $offset);
            $blk_cnt = $ranges->{$offset}% block_size ?
                    ( $ranges->{$offset} / block_size ) + 1 :
                    ( $ranges->{$offset} / block_size );
            for( ;$blk_cnt >1 ; $blk_cnt--) {
               $offset+= block_size;
               push(@ret_ary,$offset);
            }
        }
        $chkr_logs{ synth_log } .= "DATA: Inode# $inode_num, Offset $_, Size 4096\n" foreach @ret_ary;
        chomp($chkr_logs{ synth_log });
    }

    if ($log_type eq "inode")    # PLACE HOLDER FOR INODE CHANGES LOG SYNTHESIZER
    {
        $FLAG = 1;
    }

    # retrive the logs from the nodes
    eval {
        $orign_node = $server_obj->set_execution_server(-Server => $node);
        $chkr_logs{$node} = $server_obj->shell( cmd => "sort -nk5 $log_path" );
        $server_obj->set_execution_server(-Server => $orign_node);
    };
    confess "$@" if $@;
    #code used for dev debug purposes this block will be skipped
#    {last;
    NOTE::logcomment(" sythesized logs ");
    NOTE::logcomment("$chkr_logs{synth_log}");
    NOTE::logcomment(" logs on filer ");
    NOTE::logcomment("$chkr_logs{$node}");
#    last;
#    };

    # compare the retrivied logs with the synthesized logs
    if($chkr_logs{synth_log} ne $chkr_logs{$node}) {
        $FLAG = 0;
    }
    return $FLAG;
}


# ROUTINES BELOW THIS ARE DEPRICIATED #
# AND WILL BE PHASED OUT SLOWLY      #

=head2 GET INODE INFO ORIGINAL

=item Routine name:
get_inode_info_orig

=item Description:

    Get the contents of the inode for the file specified
    and return the contents of the inode in the form of a hash
    this routine uses the get_inode dblade command and will be
    phased out for the routine that uses WEX

=item Arguments

    vol_obj = volume object of type GNOTE
    test_dir = path to the file (dir path) if file is in a dir/subdir
    test_file = name of the file
    sm_name = (Optional) name of the snapshot to extract the inode from, default is basesnapshot

=item Examples

    my $inode_info = get_inode_info(vol_obj => $vol_dst, test_dir => $test_dir, test_file => $test_file);
    my $inode_info = get_inode_info(vol_obj => $vol_dst, test_file => $test_file);

=item Return values

    returns a hash refrence containing the inode info for success
    return 0 for failure

=back

=cut

sub get_inode_info_orig
{
    NOTE::logcomment(" INFO -> inside the routune get_inode_info_orig");
    my %args = @_;
    my %inode_vals  =();     # hash with contents of inode initialize to empty
    my $vol_obj = exists $args{vol_obj} ? $args{vol_obj} : return 0; # object of type vol
    my $test_dir = exists $args{test_dir} ? $args{test_dir} : undef; # test dir
    my $test_file = exists $args{test_file} ? $args{test_file} : return 0; # name of the test file
    my $sm_name = exists $args{sm_name} ? $args{sm_name} : undef; # sm inst to extract "Base Snapshot Name"
    my ($server,               # server object
        $dst_vol_name,         # dest volume name
        $dst_vol_inst,         # dest volume instance to extract "Node name"
        $exec_node,            # execution node for run cmd
        $sm_inst,              # snapmirror name
        $get_inode_cmd,        # get_inode cli for $testfile
        $file_inode,           # test file inode (on DP vol)
        $origin_node           # original execution node
        );

    $server = Server->new;
    # get the node name on which the volume resides
    $dst_vol_name = $vol_obj->name;
    $dst_vol_inst = ($server->volume_show(volume => $dst_vol_name))[0];
    $exec_node = $server->shell( cmd => "wex volume $dst_vol_name -dump location" );
    $exec_node = (split( '=', $exec_node ))[1];
    unless ($sm_name)
    {
        $sm_inst = ($server->snapmirror_show( 'destination-volume' => $dst_vol_name,
                                                          -Timeout => 120,
                                                          instance => "", ))[0];
        $sm_name = $sm_inst->{"exported-snapshot"};
    }
    # change exec server to the host node for the volume
    $origin_node = $server->set_execution_server(-Server => $exec_node);
    $server->dblade_set_privilege(level=> "test", quiet=> "on");

    # creating the command for get inode and extracting inode
    if ($test_dir) {
        $get_inode_cmd = "get_inode /vol/$dst_vol_name/../${sm_name}${test_dir}/$test_file";
    }else {
        $get_inode_cmd = "get_inode /vol/$dst_vol_name/../${sm_name}/$test_file";
    }
    $file_inode = $server->dblade(cmd => $get_inode_cmd);
    $file_inode = join( ', ', (split( /\012/, $file_inode ))[1,2,3,4,9,10]); #need to work on lines 5,6,7,8
    %inode_vals = map { split( /=/ ) } split( ", ", $file_inode );
    $inode_vals{node_name} = $exec_node;
    #setting the execution server back to the origin
    $server->set_execution_server(-Server => $origin_node);
    return ( \%inode_vals);
}

1;
