#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/Client/Host/HPUXHost.pm#1 $

## @summary Provide an interface to a HOST machine
## @author knopp@netapp.com
## @status private
## @pod ../Host.pod

use strict;

# Package Name
##############
package San::SanClient::Host::HPUXHost;

# Package Requirements
######################
use POSIX;
use Config;
use English;
use Net::Ping;
use File::Basename;

use San::SanLog;
use Error;
use vars qw( @EXPORT @ISA );

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new HPUXHost object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new (\%)
{
    my ($class, %args) = @_;

    my ($device_str, @devices);
    my ($self,$conn,$nfs_mount);
    
    if (defined $args{-self}) {        
        $self=$args{-self};
    } else  {
        $self={};
        # bless 
        bless $self,$class;
    }

    # parse arguments
    foreach (keys %args) {
      if (/-?conn/) {
          if (not length($conn = $args{$_})) {
             $self->_set_err(
                   "Error: HPUXHost::new() Invalid value for -conn: $args{$_}");                 
             return $self;
          }
          next;
      }
      if (/-?self/) {
          if (not length($self = $args{$_})) {
             $self->_set_err(
                   "Error: HPUXHost::new() Invalid value for -self: $args{$_}");                 
             return $self;
          }
          next;
      }

      else {
          $self->_set_err(
                   "Error: HPUXHost::new() Invalid parameter specified \'$_\'");
          return $self; 
      }

    }

    unless (defined $conn and length $conn) {
        $self->_set_err("Error: HPUXHost::new()",
                        "Invalid value for argument '-conn'");
        return $self; 
    }

    $self->{conn}=$conn;

	$self->{nfs_mount}=undef;

    return($self);

}

################################################################################
# Subroutine Name: configure
# Objective: configure host machine
# Details: See POD documentation at the end of this file
################################################################################
sub configure {
	my ($self,%args)=@_;
	
	$self->_clear_err();

	#  This is a NOOP for hpux
	return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: discover 
# Objective: Discover devices and label if new.
# Details: See POD documentation at the end of this file
################################################################################
sub discover ($)
{
    my ($self) = @_;
    my (
    $cmd,
    @devices,
    $status, 
    @output,
    @hDevices,
    );

    $self->_clear_err();

    $self->sanlog(-msg => "Discover all devices");

    # execute ioscan (as root)
    $cmd = "/usr/sbin/ioscan -fn -C disk";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
                                     
    # determine if ioscan worked
    if ($status gt 0) {
        $self->_set_err(
                   "Error: HPUXHost::discover() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;                             
    } 

    # execute ioinit to create devices entries
    $cmd = "/usr/sbin/ioinit -i";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
                                     
    # determine if ioinit worked
    if ($status gt 0) {
        $self->_set_err(
                   "Error: HPUXHost::discover() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;                             
    } 

    # execute insf -e (to make sure kernel and devices are in sync)
    #$cmd = "/usr/sbin/insf -e";
    #($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
                                     
    # determine if insf worked
    #if ($status gt 0) {
    #    $self->_set_err(
    #               "Error: HPUXHost::discover() Command ($cmd) returned a bad status ($status)");
    #    return $ClientUtil::FALSE;                             
    #} 
     
    $self->sanlog(-msg => "Locate devices that are not part of a volume group",
                   -msg_level => 3);

    @devices= $self->_get_unlabeled_devices();

    $self->sanlog(-msg => scalar(@devices)." Devices are not part of a volume group",
                   -msg_level => 3);

    return $ClientUtil::FALSE if ($self->error);
    
    return $ClientUtil::TRUE ;

}
 
################################################################################
# Subroutine Name: format
# Objective:       Format devices with a label 
# Details:         See POD documentation at the end of this file
################################################################################
sub format{ 
    my ($self,%args)=@_;
    my (
        $cmd,
        $devices,
        $label,
        $status,
        @format_cmds,
        @used,
        @output,
        @hDevices,    
    );

	$self->_clear_err();

    # parse arguments

    foreach (keys %args) {
      if (/-?device/) {
          if (not length($devices = $args{$_})) {
              $self->_set_err("Error: HPUXHost::format() Invalid value for argument '-devices'");
              return ($ClientUtil::FALSE); 
          }        
          next;
      }
      else {
          $self->_set_err("Error: HPUXHost::format() Invalid parameter specified \'$_\'");
          return ($ClientUtil::FALSE);                             
      }
    }     

    #  Check that required parameter '-device' was provided
    unless (defined $devices and ref($devices) eq "ARRAY"){
    $self->_set_err("Error: HPUXHost::format() ",
            "Missing required argument '-device'");

        return 0;
    }

    # iterate through the devices 
    foreach my $device (@$devices) {    
        $self->sanlog(-msg => "Running pvcreate on device '$device'");        

        # formulate the pvcreate command
        $cmd = "/usr/sbin/pvcreate -f /dev/rdsk/$device";

        # run pvcreate on the device
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: HPUXHost::format() ",
                        "format returned a bad status ($status))");

            return 0;
        } 
        
    } 

    return 1;

}

################################################################################
# Method:    mkdir
# Objective: Create a directory on the client
# Details:   See POD documentation at the end of this file
################################################################################
sub mkdir {
    my ($self, %args) = @_;
    my ($cmd,
	$path,
	@output,
	$status);

    # parse arguments
    foreach (keys %args) {
	if (/-?path/) {
	    if (not length($path = $args{$_})) {
		$self->_set_err("Error: HPUXHost::mkdir() Invalid value for argument '-path'");
		return ($ClientUtil::FALSE); 
	    }        
	    next;
	} else {
	    $self->_set_err("Error: HPUXHost::format() Invalid parameter specified \'$_\'");
	    return ($ClientUtil::FALSE);                             
	}
    }     

    # check that required parameter '-path' was provided
    unless (defined($path) and length($path)){
	$self->_set_err("Error: HPUXHost::mkdir() Missing required argument '-path'");

        return $ClientUtil::FALSE;
    }

    $cmd = "mkdir $path";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    # determine if command worked
    if ($status gt 0) {
	$self->_set_err( "Error: HPUXHost::mkdir() Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;                             
    } 

    # verify existence of directory
    $cmd = "ls $path";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    # determine if command worked
    if ($status gt 0) {
	$self->_set_err( "Error: HPUXHost::mkdir() Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;                             
    } 
    
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: sanlun_version()
# Objective:       get sanlun version
# Details:         See POD documentation at the end of this file
################################################################################
sub sanlun_version {
    my ($self) = @_;
    my (
        $cmd, 
        @output, 
        $status,
        $version,
    );
    
    $self->_clear_err();

    $cmd = "/usr/sbin/swlist | grep -i santools | awk '{print \$2}'";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err( "Error: HPUXHost::sanlun_version() Command ($cmd) returned a bad status ($status)");
        return "";     
    } 
     
    unless (defined $output[0] and length $output[0]){
        $self->_set_err("Error: HPUXHost::sanlun_version(): Unable to get sanlun version");
        return "";
    }

    $version=$output[0];

    return $version;      

}

###############################################################################
# Method:    info
# Objective: Get client information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;
    my (
        $hsh,
        @output,
        );

    $self->_clear_err();

    push @output,
           "Host name:      ".$self->hostname(),
           "Host type:      ".$self->hosttype(),
           "OS version:     ".$self->os_version(),
           "Sanlun version: ".$self->sanlun_version(),
           "Protocol:       ".$self->get_protocol(),
           "\n";
           
   return (wantarray()) ? @output: \@output; 

}

###############################################################################
# Subroutine Name: get_mounted_device
# Objective:       returns a mounted device name when passed a filesystem
###############################################################################
sub get_mounted_device {
    my ($self,$mount) = @_;
    my (
	$cmd,
	$line,
	$status,
        @output,
	);

    $self->_clear_err();

    chop($mount) if ($mount =~ /\/$/);

    $mount=~s/\//\\\//g;

    # execute the command
    $cmd = "mount | grep $mount | awk '{print ".'$3'."}'";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0 or length("@output" ) eq 0) {
	$self->_set_err("Error: HPUXHost::get_mounted_device(): Mount command failed!", @output);
	return undef; 
    } 

    if (scalar(@output) > 1) {
	$self->_set_err("Error: HPUXHost::get_mounted_device(): Multiple mount entries found!", @output);
	return undef;
    }

    return $output[0];   
}

###############################################################################
# Method:    get_lvm_default
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lvm_default {
   my ($self) = @_;

   $self->_clear_err();

   return qw(nativelvm);

}

###############################################################################
# Method:    get_unsupported_fs
# Objective: Get list of unsupported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_unsupported_fs {
   my ($self) = @_;

   $self->_clear_err();

   return ("hfs");

}

###############################################################################
# Method:    get_supported_protocols
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_protocols {
   my ($self) = @_;

   $self->_clear_err();

   return qw(fcp iscsi nfs);

}

###############################################################################
# Method:    get_supported_fs_types
# Objective: Get list of supported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_fs_types {
   my ($self) = @_;

   $self->_clear_err();

   return qw(vxfs);

}

###############################################################################
# Method:    get_unsupported_fs
# Objective: Get list of unsupported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_unsupported_fs {
   my ($self) = @_;

   $self->_clear_err();

   return ("hfs");

}

################################################################################
# Subroutine Name: iostat
# Objective:       Format devices with a label 
# Details:         See POD documentation at the end of this file
################################################################################
sub iostat{ 
    my ($self,%args)=@_;
    my (
        $interval,
        $cmd,
        $status,
        @output,    
        $cnt,
        $count,
        @buf,
    );

    $self->_clear_err();

    #  Default interval is 2 seconds
    $interval=2;

    #  Default count is 1
    $count=2;

    # parse arguments
    foreach (keys %args) {
      if (/-?interval/) {
          if (not length($interval = $args{$_})) {
              $self->_set_err("Error: HPUXHost::iostat() Invalid value for argument '-interval'");
              return ($ClientUtil::FALSE); 
          }        
          next;
      }
      elsif (/-?count/) {
      if (not length($count = $args{$_})) {
          $self->_set_err("Error: HPUXHost::iostat() Missing value for argument '-count'");
          return ($ClientUtil::FALSE); 
      }
      next;
      }
      else {
          $self->_set_err("Error: HPUXHost::iostat() Invalid parameter specified \'$_\'");
          return ($ClientUtil::FALSE);                             
      }
    }     

    # iostat command for HP-UX
    $cmd="iostat $interval $count";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::iostat() ",
                    "iostat returned a bad status ($status))");

        return (wantarray()) ? () : [];
    } 

    push @output,"\n\n";
    return (wantarray()) ? @output : \@output;

}


################################################################################
# Subroutine Name: nfs_mount
# Objective:       NFS mount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_mount {
    my ($self,%args)=@_;
    my (
	$cmd,
	$directio,
	$hide,
	$resource,
	$fs_type,
	$mount_pt,
	$options,
	$nodename,
	$vol_name,
	$name,
	$status,
	@output,
	);
 
	$self->_clear_err();
        $fs_type="nfs";
        $directio=0;
	$hide=0;

    # parse arguments
    foreach (keys %args) {
		if (/-?resource/) {
			if (not length($resource = $args{$_})) {
				$self->_set_err("Error: HPUXHost::nfs_mount() Invalid value for argument '-device_mount'");
				return ($ClientUtil::FALSE); 
			}        
			next;
		}
		elsif (/-?directio/) {
		    if (not length($directio = $args{$_}) or $directio !~ /1|0/i) {
			$self->_set_err("Error: HPUXHost::nfs_mount() Missing value for argument '-directio'");
			return ($ClientUtil::FALSE);
		    }
		    next;
		}
		elsif (/-?mount_pt/) {
			if (not length($mount_pt = $args{$_})) {
			$self->_set_err("Error: HPUXHost::nfs_mount() Missing value for argument '-mount_pt'");
			return ($ClientUtil::FALSE); 
			}
			next;
		}
		elsif (/-?options/) {
			if (not length($options = $args{$_})) {
				$self->_set_err("Error: HPUXHost::nfs_mount() Missing value for argument '-options'");
				return ($ClientUtil::FALSE); 
			}
			next;
		}
		else {
			$self->_set_err("Error: HPUXHost::nfs_mount() Invalid parameter specified \'$_\'");
			return ($ClientUtil::FALSE);                       	  
		}
    }     

    
    #  Check that required parameter '-resource' was provided
    unless (defined $resource and length $resource){
		$self->_set_err("Error: HPUXHost::nfs_mount() ",
			"Missing required argument '-resource'");
		return $ClientUtil::FALSE;
    }

    #  Check that required parameter '-mount_pt' was provided
    unless (defined $mount_pt and length $mount_pt){
		$self->_set_err("Error: HPUXHost::nfs_mount() ",
			"Missing required argument '-mount_pt'");
		return $ClientUtil::FALSE;
    }

    #  Check that required parameter '-options' was provided
    unless (defined $options and length $options){
		$self->_set_err("Error: HPUXHost::nfs_mount() ",
			"Missing required argument '-options'");
		return $ClientUtil::FALSE;
    }

    #  check if the volume is already mounted
    $self->mount_table_verify($mount_pt);
    if ($self->error){

	    # It's not mounted so we need to do it now
	    #  Use default path if not defined
	    ($nodename = $resource) =~ s/(^\S+):\S+/$1/;

	    #Check for empty value
	    if ($nodename eq "") {
		    $self->_set_err("Error: HPUXHost::nfs_mount() Unable to determine filer ".
			            "name from device mount '$resource'", @output);

		    return $ClientUtil::FALSE;
	    }

	    $vol_name = basename($resource);

	    unless (defined $mount_pt) {
		    $mount_pt = "/mnt/qa/${nodename}/".$vol_name;
		    $self->sanlog(-msg => "Using default mount pt '$mount_pt'",      
                                  -msg_level => 3);
	    }
	    
	    #  Create the mount point and open up dir privs 
	    $cmd="mkdir -p $mount_pt; chmod -R 777 $mount_pt";
	    
	    @output = $self->execute($cmd,-hide=>$hide); 
	    if ($self->error) {
		    $self->sanlog(-msg => "WARNING: HPUXHost::nfs_mount()" .
                                          "Unable to create mount point "  .
                                          "directory for volume '$resource'".
			                  join "\n", @output);
	            $self->_clear_err();				      
	    } 


	    # Add directio option if requested and it was not
	    # passed in as part of the options string
	    if ($directio and $options!~/convosync=direct/i){    	
	       $options.=(defined $options and length $options)
		    ? ",convosync=direct"
		    : "convosync=direct";
	    }

	    # Remove option param if user passed it in as part of the 
	    # options string. Lib should add the tag.
	    $options=~s/-o //g;

	    # Mount the volume's filesystem
	    $cmd=($options)
	      ? "/usr/sbin/mount -o $options $resource $mount_pt"
	      : "/usr/sbin/mount $resource $mount_pt";


	    #  Create the NFS mount command.
	    #mount -o rw,hard,intr  gin:/vol/vol1/ /mnt
	    @output = $self->execute($cmd,-hide=>$hide); 
	    if ($self->error) {
		    $self->_set_err( "Error: HPUXHost::nfs_mount() Unable to NFS mount volume",
		    @output);
	    
		    return $ClientUtil::FALSE;                            
	    } 

	    $self->mount_table_verify($mount_pt);
		    
	    if ($self->error) {
		    $self->_set_err($self->errmsg());
		    return $ClientUtil::FALSE;
	    }

    }
    else {
	$self->sanlog(-msg => "Warning: HPUXHost::nfs_mount() volume " .
                              "($mount_pt) is already mounted");
	$self->_clear_err();
	return $ClientUtil::TRUE;
    }

=comment

	#  For now lets not add mount points into the fstab files.  Most test folks probaly 
	#  don't care about that.  However, keep the code around if someone changes mind.  


	#make sure the mount entry does not already exist
	$self->mount_entry_verify($mount_pt);

	if ($self->error) {
		# It's not mounted so we need to do it now
		#  Use default path if not defined

		$self->_clear_err();

		$self->mount_entry_add(
			-device_mount=>$resource,
			-mount_pt=>$mount_pt,
			-fs_type=>$fs_type,
			-mount_options=>"rw" . ',' . "hard" . ',' . "intr",
			);

		if ($self->error) {
			$self->_set_err($self->errmsg());
			return $ClientUtil::FALSE;
		}   

		#Check it again to make sure it's really there
		$self->mount_entry_verify($mount_pt);

		if ($self->error) {
			$self->_set_err($self->errmsg());
			return $ClientUtil::FALSE;
		}

	}
=cut   

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: nfs_unmount()
# Objective:       unmount NFS volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_unmount() {
    my ($self, %args) = @_;

	my(
		$cmd,
	        $hide,
		$mount_pt,
		$status,
		@output,
	);

	$self->_clear_err();
    $hide=0;

    # parse arguments
    foreach (keys %args) {
        if (/-?mount_pt/) {
	    #  Make sure mount point exists and is not '/'.
            if (not length($mount_pt = $args{$_}) and ($args{$_}!="/")) {
                $self->_set_err("Error: HPUXHost::nfs_unmount() Missing value for argument '-mount_pt'");
                return ($ClientUtil::FALSE); 
            }
            next;
        }
        else {
            $self->_set_err("Error: HPUXHost::nfs_mount() Invalid parameter specified \'$_\'");
            return ($ClientUtil::FALSE);                          
        }
    }     


    # Check that required parameter '-mount_pt' was provided
    unless (defined $mount_pt and length $mount_pt){
        $self->_set_err("Error: HPUXHost::nfs_mount() ",
                        "Missing required argument '-mount_pt'");
        return $ClientUtil::FALSE;
    }


    $cmd="umount ".$mount_pt;
    @output = $self->execute($cmd,-hide=>$hide); 
    if ($self->error) {
	$self->_set_err("Error: HPUXHost::nfs_unmount() ".
                        "Unable to unmount NFS volume ".$mount_pt." ",@output);
	return $ClientUtil::FALSE;
    }


=comment

    # Entry additions to the mount file have been commented out.  So deletion
    # of entries is not needed.

	$self->mount_entry_delete($nfs_mount);

	if ($self->error) {
		$self->_set_err($self->errmsg());
		return $ClientUtil::FALSE;
	}

	$self->mount_entry_verify($nfs_mount);

	unless ($self->error) {
		$self->_set_err("Error: HPUXHost::nfs_unmount() mount entry \"$nfs_mount\" was not removed from the mount file");
		return $ClientUtil::FALSE;
	}

	$self->_clear_err();

=cut

    #  Create the command to remove the NFS mount point 
    $cmd="rm -Rf ".$mount_pt;
    @output = $self->execute($cmd,-hide=>$hide); 
    if ($self->error) {
	$self->_set_err("Warning: HPUXHost::nfs_unmount() Unable to remove ".
                    "mount point directory '".$mount_pt."'",@output);

	return $ClientUtil::FALSE;
	
    } 


	return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: mount_entry_add()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_add {
    my ($self,%args)=@_;
    my (
	$cmd,
	$device_mount,
	$mount_pt,
	$fs_type,
	$mount_options,
	$status,
	@output,
	);

    $self->_clear_err();
    
    #  initialize variables 
    $mount_options="log,largefiles";

    # parse arguments
    foreach (keys %args) {
		if (/-?device_mount/) {
			if (not length($device_mount = $args{$_})) {
			$self->_set_err("Error: HPUXHost::mount_entry_add() Invalid value for argument '-device_mount'");
			return ($ClientUtil::FALSE); 
			}        
			next;
		}
		elsif (/-?mount_pt/) {
			if (not length($mount_pt = $args{$_})) {
			$self->_set_err("Error: HPUXHost::mount_entry_add() Missing value for argument '-mount_pt'");
			return ($ClientUtil::FALSE); 
			}
			next;
		}
		elsif (/-?fs_type/) {
			if (not length($fs_type = $args{$_})) {
			$self->_set_err("Error: HPUXHost::mount_entry_add() Missing value for argument '-fs_type'");
			return ($ClientUtil::FALSE); 
			}
			next;
		}
		elsif (/-?mount_options/) {
			if (not length($mount_options = $args{$_})) {
			$self->_set_err("Error: HPUXHost::mount_entry_add() Missing value for argument '-mount_options'");
			return ($ClientUtil::FALSE); 
			}
			next;
		}
		else {
			$self->_set_err("Error: HPUXHost::mount_entry_add() Invalid parameter specified \'$_\'");
			return ($ClientUtil::FALSE);                       	  
		}
    }     

    
    #  Check that required parameter '-device' was provided
    unless (defined $device_mount and length $device_mount){
	$self->_set_err("Error: HPUXHost::mount_entry_add() ",
			"Missing required argument '-device_mount'");
	return $ClientUtil::TRUE;
    }

    #  Check that required parameter '-device' was provided
    unless (defined $mount_pt and length $mount_pt){
	$self->_set_err("Error: HPUXHost::mount_entry_add() ",
			"Missing required argument '-mount_pt'");
	return $ClientUtil::TRUE;
    }

    #  Check that required parameter '-device' was provided
    unless (defined $fs_type and $fs_type=~/vxfs|hfs|nfs/i){
	$self->_set_err("Error: HPUXHost::mount_entry_add() ",
			"Missing required argument '-fs_type'");
	return $ClientUtil::TRUE;
    }

    $cmd="echo ".$device_mount."\t".$mount_pt."\t".$fs_type."\t".$mount_options."\t"."0"."\t"."0"."\t".">> /etc/fstab";

    $self->sanlog(-msg => "Adding ($fs_type) fs entry to mount table");	

    #  execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
         $self->_set_err("Error: HPUXHost::mount_entry_add Unable to add mount entry to /etc/fstab file",@output);
         return $ClientUtil::FALSE;     
    } 

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self,$mntpt)=@_;
    my (
        $cmd,
        $mount,
        $status,
	@output,
        );

    $self->_clear_err();

    #  Unmount the volume's filesystem
    $cmd="umount $mntpt";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
	$self->_set_err("Error:HPUXHost ::unmount_fs() Unable to unmount filesystem", @output);                       
    } 

    $self->mount_entry_delete($mntpt);

    $self->sanlog(-msg => "Removing mount point '$mntpt'");  

    #  Remove directory head
    $cmd = "rm -rf $self->{mount}";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
	$self->_set_err("Error: HPUXHost::unmount_fs() Unable to remove mount point '$mount'", @output);                           
    } 

    return $self;
}

################################################################################
# Subroutine Name: mount_entry_delete()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_delete {
	my ($self, $mount) = @_;
	my (
		$cmd,
		$status,
		@output,
		);

    $self->_clear_err();

    #  Clear vfstab mount entry
    $self->sanlog(-msg => "Clear /etc/fstab mount entry");

	unless (defined $mount and length($mount)) {
		$self->_set_err("Error: HPUXHost::mount_entry_delete Missing required paremeter '$mount'",@output);
         return $ClientUtil::FALSE;     
	}

    $mount=~s/\//\\\//g;

    $cmd="perl -e\'".
          "open(FH,\"< \/etc\/fstab\") or exit(1);".
          "open(FD,\"> \/etc\/fstab_new\") or exit(1);".
          "while(<FH>){\\\n".		     
          "print FD \$_ if (\$_ !~ /$mount\\s/);". 
          "}\\\n".
          "close FD;".
          "close FH;\\\n".
          "rename(\"/etc/fstab_new\", \"/etc/fstab\") or exit(1);".
          "\' ";
 
     #  execute the command
     ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
     if ($status > 0) {
         $self->_set_err("Error: HPUXHost::mount_entry_delete Unable to remove mount entry from /etc/fstab file",@output);
         return $ClientUtil::FALSE;     
     } 

    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: mount_entry_verify
# Objective:      get_mount_entries
###############################################################################
sub mount_entry_verify{
    my ($self,$mount) = @_;
    my (
	$cmd,
        $entry,
        $idx,
    	$status,
        $value,
	@values,
        @params,
        @output,
        @buf,
	);

    $self->_clear_err();

    # initialize
    $entry="";
    $idx=0;

    #$mount=~s/\//\\\//g;
    #  Check for mount point prifixed by tab or space
    $cmd = "cat /etc/fstab | grep -v '^#' | egrep '	$mount	| $mount'";

    # execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
	  $self->_set_err("Error: HPUXHost::mount_entry_verify(): Unable to verify mount file entry",@output);
	  return $entry;
    } 

    #$self->sanlog(-msg => \@output);

    if (scalar(@output)==0) {
      $self->_set_err("Error: HPUXHost::mount_entry_verify(): No mount file entry exists");
    }
    elsif(scalar(@output)>1){
      $self->_set_err("Error: HPUXHost::mount_entry_verify(): Multiple mount file entries exist",@output);	
    }
    else {
	  @params=("Device to mount:","Mount pt:","Fs type:","Mount Options:");
           
	  @values=split(/\s+/,$output[0]); 	
	  
	  foreach $idx (0..3) {
             push @buf, $params[$idx]." ".$values[$idx];
          }    

    }  

    return (wantarray()) ? @buf : \@buf;   
}

###############################################################################
# Subroutine Name: mount_table_verify
# Objective:       verify mount table
###############################################################################
sub mount_table_verify{
    my ($self,$mount) = @_;
    my (
        $status,
        @output,
    );

    $self->_clear_err();

    chop($mount) if ($mount =~ /\/$/);

    $mount=~s/\//\\\//g;

    # execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, "mount");
    if ($status > 0 or length("@output" ) eq 0) {
    $self->_set_err("Error: HPUXHost::mount_table_verify(): Unable to verify mount table", @output);
    return (wantarray()) ? @output : \@output;
    }

    unless (grep /$mount /, @output) {
    $self->_set_err("Error: HPUXHost::mount_table_verify(): Unable to verify mount table", @output);
    return (wantarray()) ? @output : \@output;
    }

    #  Need to parse the mount into some usefule structure.  For now just show
    #  that something was found.

    return (wantarray()) ? @output : \@output;
}

################################################################################
# Subroutine Name: filer_mount_entry_backup()
# Objective:       backup the filer /etc/exports file to /tmp for later restore
# Details:         See POD documentation at the end of this file
################################################################################
sub filer_mount_entry_backup {
	my ($self,$mount_pt)=@_;
    my (
	$cmd,
	$status,
	@output,
	);
    
    $self->_clear_err();

	unless(defined $mount_pt and length $mount_pt){
		$self->_set_err("Error: HPUXHost::filer_mount_entry_restore() mount_pt is a required argument");
		return $ClientUtil::FALSE;
	}
  
    #Save the original /etc/exports file for retore at cleanup
    $self->sanlog(-msg => "Backup the original /etc/exports file from \"$mount_pt\" for later restore");

	#backup original /etc/exports 
	$cmd="cp -p ${mount_pt}/etc/exports ${mount_pt}/etc/exports.ORIG";
	
    #  execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::filer_mount_entry_backup() Unable to backup \"${mount_pt}/etc/exports\"" , 
						"to \"${mount_pt}/etc/exports.ORIG\"",@output);
        return $ClientUtil::FALSE;     
    } 

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: filer_mount_entry_restore()
# Objective:       restore the filer /etc/exports file from /tmp 
# Details:         See POD documentation at the end of this file
################################################################################
sub filer_mount_entry_restore {
	my ($self,$mount_pt)=@_;
    my (
	$cmd,
	$status,
	@output,
	);
    
    $self->_clear_err();

	unless(defined $mount_pt and length $mount_pt){
		$self->_set_err("Error: HPUXHost::filer_mount_entry_restore() mount_pt is a required argument");
		return $ClientUtil::FALSE;
	}
  
    #Restore the original /etc/exports file
    $self->sanlog(-msg => "Restore the original /etc/exports file from \"$mount_pt\"");
	
	#restore original /etc/exports and remove old commented lines
	$cmd="cat ${mount_pt}/etc/exports.ORIG | grep -v '^#/vol' > ${mount_pt}/etc/exports";
	
    #  execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::filer_mount_entry_restore() Unable to restore \"${mount_pt}/etc/exports\"" , 
						"from \"${mount_pt}/etc/exports.ORIG\"",@output);
        return $ClientUtil::FALSE;     
    } 

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: config_paths
# Objective:       To run ntap_config_paths to setup proper path ordering
#                  for HP-UX.
# Details:         See POD documentation at the end of this file
################################################################################
sub config_paths { 
    my ($self)=@_;
    my (
        $cmd,
        $status,
        @output,    
    );

    $self->_clear_err();

    #  No Need to run if iscsi
    return if ($self->get_protocol()=~/iscsi/i);

    $self->sanlog(-msg => "Executing ntap_config_paths");

    # Run ntap_config_paths to setup proper path ordering
    $cmd = "ntap_config_paths";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 

    # determine if the ntap_config_paths worked.
    if ($status gt 0 and $status ne 203) {
        $self->_set_err("Error: Vg::_create Unable to run ntap_config_paths",
                        @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}
################################################################################
# Subroutine Name: raw_device_setup
# Objective:       Do whatever we need to do to use raw devices successfully
#                  after the call to discover().
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_setup {
    my ($self)=@_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: raw_device_cleanup
# Objective:       Do whatever we need to do to make sure the host is more
#                  happy when luns are removed.  For linux, we want to do
#                  an iscsi stop.
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_cleanup {
    my ($self)=@_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: orphaned_device_cleanup
# Objective:       When the luns are removed from the filer, we will have
#                  orphaned luns.  This causes big problems. If we do not
#                  remove these specail files we often fail setup in san Libs.
#                  
# Details:         See POD documentation at the end of this file
################################################################################
sub orphaned_device_cleanup {
    my ($self)=@_;
    my (
        $cmd,
        $device,
        $status,
        $hide,
        @output
        );

    $self->_clear_err();
    $hide = 0;

    # unconfigure devices
    $cmd = "ioscan -fn";
    $self->execute($cmd,-hide=>$hide); 	    

    # find all instances of unused NETAPP special files and, get the hardware path
    $cmd = "ioscan -fC disk | grep NETAPP | grep NO_HW | awk '{print  \$3}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # just ignore status and plow thru all special files
    foreach $device (@output)
    {
        # remove all instances of unused NETAPP special files    
        $cmd = "rmsf -H $device";
        $self->execute($cmd,-hide=>$hide);
    }

    return $ClientUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _get_devices
# Objective:       Get list of all devices on host
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_devices(){
   my ($self) = @_;
    my (
        $cnt,
        $cmd,
        $dev_str,
        $device,
        $status, 
        @output,
        $found,
        @hDevices,
    );

    # run ioscan getting information from the kernel (since this is allot 
    # faster, and ioscan -fn -C disk should have already been run earlier)
    $cmd = "/usr/sbin/ioscan -fnk -C disk";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
                                     
    # determine if ioscan command worked
    if ($status gt 0) {
        $self->_set_err(
              "Error: HPUXHost::_get_devices() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;                             
    } 

    $found = $ClientUtil::FALSE;
    $cnt = 0;
    foreach my $line (@output) {
       if ($line =~ /(netapp)|(ontap)/i and
           $line =~ /claimed/i) {
           $found = $ClientUtil::TRUE;
           next;
       }
    
       if ($found) {
           $device = (split(' ', $line))[1];
           $device =~ s#/dev/rdsk/##;

           # create hash of lun descriptions
           my $hDevs={};
           $hDevs->{path} = "/dev/rdsk";    
           $hDevs->{name} = $device;
           $hDevs->{type} = "NETAPP";
           $hDevs->{label} = "NA";
           $hDevs->{id} = $cnt;
           push @hDevices,$hDevs;
           $cnt++;
       }
       $found = $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => scalar(@hDevices)." devices found (includes multiple paths to devices)"); 

    return (wantarray()) ? @hDevices : \@hDevices; 

}

################################################################################
# Subroutine Name: _get_unlabeled_devices
# Objective:       Retrieve devices that do not have a label
# Details:         See POD documentation at the end of this file
################################################################################

sub _get_unlabeled_devices($){
    my ($self) = @_;
    my (
        @hDevices,
        @devices,
        $cmd,
        $status,
        @output,
    );

    #  Retrieve all host devices
    @hDevices = $self->_get_devices();
    return 0 if ($self->error);

    #  Determine which devices do not have 
    #  a label
    foreach my $device (@hDevices) {
        # Skip if this is not a NETAPP device
        next if ($device->{type} !~ /NETAPP/);

        # If the device is not in a VG (which is what
        # pvdisplay will how), mark it as "unlabeled"
        # and hence available.
        $cmd = "/usr/sbin/pvdisplay /dev/dsk/$device->{name}";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        if (grep(/Cannot display physical volume/i, @output) > 0) {
            push (@devices, $device->{name});
        }
    }
   
    #  Format the list of devices with label
    return (wantarray()) ? @devices : \@devices; 
}
  
1;

