#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/Client/Host/LinuxHost.pm#1 $

## @summary Provide an interface to a HOST machine
## @author knopp@netapp.com
## @status private
## @pod ../Host.pod

use strict;

# Package Name
##############
package Client::Host::LinuxHost;

# Package Requirements
######################
use POSIX;
use Config;
use English;
use Net::Ping;
use Tie::IxHash;
use Data::Dumper;
use File::Basename;

use Client::Mount::LinuxMount;

use San::SanLog;
use Error;
use vars qw( @EXPORT @ISA );

@ISA=qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new LinuxHost object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new (\%)
{
    my ($class, %args) = @_;

    my ($device_str, @devices, %hRaw);
    my ($self,$conn,$nfs_mount);

    if (defined $args{-self}) {		
	$self=$args{-self};
    } else  {
	$self={};
	bless $self,$class;
    }

    # parse arguments
    foreach (keys %args) {
	if (/-?conn/) {
	    if (not length($conn = $args{$_})) {
		$self->_set_err("Error: LinuxHost::new() Invalid value for -conn: $args{$_}");                 
		return $self;
	    }
	    next;
	}
	if (/-?self/) {
	    if (not length($self = $args{$_})) {
		$self->_set_err("Error: LinuxHost::new() Invalid value for -self: $args{$_}");                 
		return $self;
	    }
	    next;
	}
	else {
	    $self->_set_err("Error: LinuxHost::new() Invalid parameter specified \'$_\'");
	    return $self; 
	}

    }

    unless (defined $conn and length $conn) {
	$self->_set_err("Error: LinuxHost::new()",
			"Invalid value for argument '-conn'");
	return $self; 
    }

    $self->{conn}=$conn;

    #  Create, load, and manage hash of raw device bindings.
    #  Keep load order.
    tie %hRaw, "Tie::IxHash";

    $self->_load_raw_bindings(\%hRaw);
    $self->{raw_bindings}=\%hRaw;	

    return($self);
}

################################################################################
# Method:    configure
# Objective: configure host machine
# Details:   See POD documentation at the end of this file
################################################################################
sub configure {
    my ($self, %args)=@_;
    
    #  This is a NOOP for linux

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    discover 
# Objective: Discover devices
# Details:   See POD documentation at the end of this file
################################################################################
sub discover ($)
{
    my ($self) = @_;
    my (
	$cmd,
	@devices,
	$status, 
	@output,
	@hDevices,
	);
    
    $self->_clear_err();

    $self->sanlog(-msg => "Discover all devices");

    #
    # There are some serious problems that we can hit right here concerning
    # orphaned devices.  When we delete a lun on a filer, Linux still has
    # information about the device stored.  If these orphaned devices exist,
    # there are 2 problems: 1) running iscsi restart on SuSE will hang and
    # never return (it'll be trying to get rid of these devices and they hare
    # reference counts and will not be deleted....repeat forever).
    # 2) if you have these orphaned devices say /dev/sdc and /dev/sdd which used
    # to point to say a 5g lun.  Then you create 10g luns instead and restart or
    # reload the iscsi driver....the devices will be found and mapped to /dev/sdc
    # and /dev/sdd...only Linux will still think they are 5g luns.
    # Because of this, you need to make sure your system is free of orphaned
    # luns before you start your test.
    #
    # execute (as root)
    $cmd = "/etc/init.d/iscsi restart";
    # sleep(10);

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    # determine if command worked
    if ($status gt 0) {
	$self->_set_err( "Error: LinuxHost::discover() Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;                             
    } 

    # IMPORTANT NOTE:
    # Sometimes Linux takes a while to discover the devices.  Give it some time or its search 
    # will find nothing.  This sleep time may need to be increased depending on the number of 
    # new devices needing discovery.
    sleep(30);

    $self->sanlog(-msg => "Locate devices that are not part of a volume group",
                  -msg_level => 3);
    #  Allow iscsi driver time to discover new devices
    @devices= $self->_get_unlabeled_devices();
    $self->sanlog(-msg => scalar(@devices)." Devices are not part of a volume group",
                  -msg_level => 3);

    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: format
# Objective:       Format devices with a label 
# Details:         See POD documentation at the end of this file
################################################################################
sub format{ 
    my ($self,%args)=@_;
    my (
        $cmd,
        $devices,
        $label,
        $status,
        @format_cmds,
        @used,
        @output,
        @hDevices,    
	);

    # parse arguments
    foreach (keys %args) {
	if (/-?device/) {
	    if (not length($devices = $args{$_})) {
		$self->_set_err("Error: LinuxHost::format() Invalid value for argument '-devices'");
		return ($ClientUtil::FALSE); 
	    }        
	    next;
	}
	elsif (/-?label/) {  #  NOOP for compatability with Solaris
	    if (not length($label = $args{$_})) {
		$self->_set_err("Error: LinuxHost::format() Invalid value for argument '-devices'");
		return ($ClientUtil::FALSE); 
	    }        
	    next;
	}

	else {
	    $self->_set_err("Error: LinuxHost::format() Invalid parameter specified \'$_\'");
	    return ($ClientUtil::FALSE);                             
	}
    }     

    #  Check that required parameter '-device' was provided
    unless (defined $devices and ref($devices) eq "ARRAY"){
	$self->_set_err("Error: LinuxHost::format() ",
			"Missing required argument '-device'");

        return 0;
    }

    # iterate through the devices 
    foreach my $device (@$devices) {    
        $self->sanlog(-msg => "Running pvcreate on device '$device'");

        # formulate the pvcreate command
        $cmd = "pvcreate /dev/$device";

        # run pvcreate on the device
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: LinuxHost::format() ",
			    "format returned a bad status ($status))");

            return 0;
        } 
        
    } 

    return 1;
}

################################################################################
# Method:    mkdir
# Objective: Create a directory on the client
# Details:   See POD documentation at the end of this file
################################################################################
sub mkdir {
    my ($self, %args) = @_;
    my ($cmd,
	$path,
	@output,
	$status);

    # parse arguments
    foreach (keys %args) {
	if (/-?path/) {
	    if (not length($path = $args{$_})) {
		$self->_set_err("Error: LinuxHost::mkdir() Invalid value for argument '-path'");
		return ($ClientUtil::FALSE); 
	    }        
	    next;
	} else {
	    $self->_set_err("Error: LinuxHost::format() Invalid parameter specified \'$_\'");
	    return ($ClientUtil::FALSE);                             
	}
    }     

    # check that required parameter '-path' was provided
    unless (defined($path) and length($path)){
	$self->_set_err("Error: LinuxHost::mkdir() Missing required argument '-path'");

        return $ClientUtil::FALSE;
    }

    $cmd = "mkdir $path; chmod 777 $path";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    # determine if command worked
    if ($status gt 0) {
	$self->_set_err( "Error: LinuxHost::mkdir() Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;                             
    } 

    # verify existence of directory
    $cmd = "ls $path";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    # determine if command worked
    if ($status gt 0) {
	$self->_set_err( "Error: LinuxHost::mkdir() Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;                             
    } 
    
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: sanlun_version()
# Objective:       get sanlun version
# Details:         See POD documentation at the end of this file
################################################################################
sub sanlun_version {
    my ($self) = @_;
    my (
	$cmd, 
	@output, 
	$status,
	$version,
	);
    
    $self->_clear_err();

    $cmd = "sanlun version| awk '{print \$2}'";

    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err( "Error: LinuxHost::sanlun_version() Command ($cmd) returned a bad status ($status)");
	return "";     
    } 
    
    
    unless (defined $output[0] and length $output[0]){
	$self->_set_err("Error: LinuxHost::sanlun_version(): Unable to retrieve sanlun version");
	return "";
    }

    $version=$output[0];

    return $version;  
}

###############################################################################
# Method:    info
# Objective: Get client information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;
    my (
	$hsh,
	@output,
	);

    push @output,
    "Host name:      ".$self->hostname(),
    "Host type:      ".$self->hosttype(),
    "OS version:     ".$self->os_version(),
    "Vendor release: ".$self->_get_release(),
    "Sanlun version: ".$self->sanlun_version(),
    "\n";
    

    return (wantarray()) ? @output: \@output; 
}

################################################################################
# Subroutine Name: iostat
# Objective:       Format devices with a label 
# Details:         See POD documentation at the end of this file
################################################################################
sub iostat{ 
    my ($self,%args)=@_;
    my (
        $interval,
	$cmd,
	$status,
	@output,	
	$cnt,
	$count,
	@buf,
	);

    $self->_clear_err();

    #  Default interval is 2 seconds
    $interval=2;

    #  Default count is 1
    $count=1;

    # parse arguments
    foreach (keys %args) {
	if (/-?interval/) {
	    if (not length($interval = $args{$_})) {
		$self->_set_err("Error: LinuxHost::iostat() Invalid value for argument '-interval'");
		return ($ClientUtil::FALSE); 
	    }        
	    next;
	}
	elsif (/-?count/) {
	    if (not length($count = $args{$_})) {
		$self->_set_err("Error: LinuxHost::iostat() Missing value for argument '-count'");
		return ($ClientUtil::FALSE); 
	    }
	    next;
	}
	else {
	    $self->_set_err("Error: LinuxHost::iostat() Invalid parameter specified \'$_\'");
	    return ($ClientUtil::FALSE);                       	  
	}
    }     

    #  Remember that the first iostat cycle is wrong so add an extra cycle.
    #  Minimize the output return since 'z' flag doesn't kick in until second cycle.
    $count+=1;
    $cmd="iostat -x -t $interval $count | grep -v  '0.00   0.00  0.00  0.00    0.00    0.00     0.00     0.00     0.00     0.00    0.00   0.00   0.00'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
	$self->_set_err("Error: LinuxHost::iostat() ",
			"iostat returned a bad status ($status))");

	return (wantarray()) ? () : [];
    } 

    #$self->sanlog(-msg => "OUTPUT",\@output,"#"x50); 

    #  Clean up the output by removing the first iostat cycle

    $cnt=0;
    foreach my $line (@output){
	if ($line=~/Time/i){
	    $cnt++;	
	}

        next if ($cnt==1);
	
	push (@buf,$line);
	
    }

    push @buf,"\n\n";
    return (wantarray()) ? @buf : \@buf;
}

################################################################################
# Subroutine Name: reboot()
# Objective:       reboot host
# Details:         See POD documentation at the end of this file
################################################################################
sub reboot
{
    my ($self) = @_;

    my (
        $num_tries,
        $alive,
        $retry,
        @output,
        $status,
        $cmd,
        $machine
       );

    $self->_clear_err();

    $machine = $self->hostname(),

    $self->sanlog(-msg => "Rebooting Client $machine");

    # reboot the test machine
    $cmd = "/sbin/shutdown -r now";
    @output=&Tharn::say($self->{conn}, $cmd, '-blind');

    # sleep just enough to prime the reboot process
    sleep(180);

    # Ping the host every 30 seconds until it comes back
    $retry = 0;
    $alive = 0;
    $num_tries = 30;
    while ($retry < $num_tries) {
        my $ping = Net::Ping->new();

        # ping host to see if it is up
        if ($ping->ping($machine)) {
            $retry = $num_tries;
            $alive = 1;
        } else {
            &main::logcomment("($machine) not up sleeping 30 seconds");
            sleep(30);
            $retry++;
        }
    } # end while

    # check status of reboot and report
    #
    unless($alive){
        $self->_set_err("Machine never rebooted");
        return $ClientUtil::FALSE;
    }

    #
    # After we ping the box, it might be a little while before we
    # stop getting connection refused on the reconnect.
    #
    sleep(120);

    # reconnect to the client
    $self->reconnect();
    return $ClientUtil::FALSE if ($self->error);

    # run uname on the client so we know the connection/client are happy
    $self->sanlog(-msg => "Running Uname on $machine\n",
                  -msg_level => 3);
    ($status, @output) = ClientUtil::execute($self->{conn}, "uname -a");
    if ($status > 0) {
        $self->_set_err( "Error: LinuxHost::reboot() Command ($cmd) " .
                         "returned a bad status ($status)");
        return "";
    }
    
   $self->sanlog(-msg => "Hostname=".  join("\n", @output),
                 -msg_level => 3);

   $self->{reboot}=0;

   return $ClientUtil::PASS;
}

################################################################################
# Subroutine Name: is_nfs_mounted
# Objective:       Determine if NFS volume is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_nfs_mounted {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{nfs_mount} and length $self->{nfs_mount}){
        return $ClientUtil::TRUE;
    } else {
	return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: nfs_mount
# Objective:       NFS mount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_mount {
    my ($self,%args)=@_;
    my (
	$cmd,
	$directio,
	$hide,
	$resource,
	$mount_pt,
	$options,
	$nodename,
	$vol_name,
	$name,
	$status,
	@output,
	);
 
	$self->_clear_err();
        $options="rw,hard,intr";
        $hide=0;
	
    # parse arguments
    foreach (keys %args) {
	if (/-?resource/) {
	    if (not length($resource = $args{$_})) {
		$self->_set_err("Error: LinuxHost::nfs_mount() Invalid value for argument '-resource'");
		return ($ClientUtil::FALSE); 
	    }        
	    next;
	}
	elsif (/-?directio/) {
	    if (not length($directio = $args{$_} or $directio !~ /1|0/)) {
		$self->_set_err("Error: LinuxHost::nfs_mount() Missing value for argument '-directio'");
		return ($ClientUtil::FALSE);
	    }
	    next;
	}
	elsif (/-?mount_pt/) {
	    if (not length($mount_pt = $args{$_})) {
		$self->_set_err("Error: LinuxHost::nfs_mount() Missing value for argument '-mount_pt'");
		return ($ClientUtil::FALSE); 
	    }
	    next;
	}
	elsif (/-?options/) {
	    if (not length($options = $args{$_})) {
		$self->_set_err("Error: LinuxHost::nfs_mount() Missing value for argument '-options'");
		return ($ClientUtil::FALSE); 
	    }
	    next;
	}
	else {
	    $self->_set_err("Error: LinuxHost::nfs_mount() Invalid parameter specified \'$_\'");
	    return ($ClientUtil::FALSE);                       	  
	}
    }     

    
    # Check that required parameter '-resource' was provided
    unless (defined $resource and length $resource){
	$self->_set_err("Error: LinuxHost::nfs_mount() ",
			"Missing required argument '-resource'");
	return $ClientUtil::FALSE;
    }

    # Check that required parameter '-device' was provided
    unless (defined $mount_pt and length $mount_pt){
	$self->_set_err("Error: LinuxHost::nfs_mount() ",
			"Missing required argument '-mount_pt'");
	return $ClientUtil::FALSE;
    }


    # Check that required parameter '-options' was provided
    unless (defined $options and length $options){
	$self->_set_err("Error: LinuxHost::nfs_mount() ",
			"Missing required argument '-options'");
	return $ClientUtil::FALSE;
    }

    # check if the volume is already mounted
    $self->mount_table_verify($mount_pt);
    if ($self->error) {	    
	# It's not mounted so we need to do it now
	# Use default path if not defined
	$self->_clear_err();

	($nodename = $resource) =~ s/(^\S+):\S+/$1/;
	#Check for empty value
	if ($nodename eq "") {
	    $self->_set_err( "Error: LinuxHost::nfs_mount() Unable to determine filer ".
			     " name from device mount '$resource'",
			     @output);

	    return $ClientUtil::FALSE;
	}

	$vol_name = basename($resource);

	unless (defined $mount_pt) {
	    $mount_pt = "/mnt/qa/${nodename}/".$vol_name;
            $self->sanlog(-msg => "Using default mount pt '$mount_pt'",
                          -msg_level => 3);
	}
	
	#  Create the mount point and open up dir privs 
	$cmd="mkdir -p $mount_pt; chmod -R 777 $mount_pt";

	@output = $self->execute($cmd,-hide=>$hide); 	
	if ($self->error) {
	    $self->sanlog(-msg => "WARNING: LinuxHost::nfs_mount() Unable to create mount ".
			  "point directory for volume '$resource'\n"               .
			  join "\n", @output);
	    $self->_clear_err();
	} 	
	
	# Directio is not supported on Redhat versions (Ver ?) through mount options.  Its 
	# the application responsibility for setting up directio connectivity.
	if ($directio){  
	    
	    if ($options=~/forcedirectio|noquota/) {
		$self->sanlog(-msg => "INFO: LinuxHost::nfs_mount() option -forcedirectio and -noquota not supported.");
		$options=~s/forcedirectio,//g;
		$options=~s/noquota,//g;
	    }
	}

	#  Remove option param if user passed it in as part of the 
	#  options string. Lib should add the tag.
	$options=~s/-o //g;
	


	# Create the NFS mount command.
	if (defined $options){
	    $cmd= "mount -o $options $resource $mount_pt;chmod 777 $mount_pt";
	}
	else {
	    $cmd= "mount $resource $mount_pt;chmod 777 $mount_pt";
	}

	@output = $self->execute($cmd,-hide=>$hide); 
	if ($self->error) {
	    $self->_set_err("Error: LinuxHost::nfs_mount() Unable to NFS mount volume",
			    "Code: $status",@output);				 
	    return $ClientUtil::FALSE;
	} 

    }
    else {
	$self->sanlog(-msg => "Warning: LinuxHost::nfs_mount() volume ($mount_pt) is already mounted");
	$self->_clear_err();
	return $ClientUtil::TRUE;
    }
    
	
	return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: nfs_unmount()
# Objective:       unmount NFS volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_unmount() {
    my ($self, %args) = @_;
    my (
        $cmd,
        $force,
	$hide,
        $mount_pt,
        $name,
        $status,
        @output,
        );
    
    $self->_clear_err();
    
    $hide=0;

    # parse arguments
    foreach (keys %args) {
        if (/-?mount_pt/) {
	    #  Make sure mount point exists and is not '/'.
            if (not length($mount_pt = $args{$_}) and ($args{$_}!="/")) {
                $self->_set_err("Error: LinuxHost::nfs_unmount() Missing value for argument '-mount_pt'");
                return ($ClientUtil::FALSE); 
            }
            next;
        }
        else {
            $self->_set_err("Error: LinuxHost::nfs_mount() Invalid parameter specified \'$_\'");
            return ($ClientUtil::FALSE);                          
        }
    }     

    # Check that required parameter '-mount_pt' was provided
    unless (defined $mount_pt and length $mount_pt){
        $self->_set_err("Error: LinuxHost::nfs_mount() ",
                        "Missing required argument '-mount_pt'");
        return $ClientUtil::FALSE;
    }

    
    # Create the mount point and open up dir privs 
    $cmd="umount $mount_pt";
    @output = $self->execute($cmd,-hide=>$hide); 
    if ($self->error) {
	$self->_set_err("Error: LinuxHost::nfs_unmount() ".
                        "Unable to unmount NFS volume ".$mount_pt." ",@output);
	return $ClientUtil::FALSE;
    } 



   

=comment
    #  For now lets not delete mount points in the fstab files
    $self->mount_entry_delete($mount_pt);
    if ($self->error) {
	    $self->_set_err($self->errmsg());
	    return $ClientUtil::FALSE;
    }
=cut


    #  Create the command to remove the NFS mount point 
    $cmd="rm -Rf ".$mount_pt;
    @output = $self->execute($cmd,-hide=>$hide); 
    if ($self->error) {
	$self->_set_err("Warning: LinuxHost::nfs_unmount() Unable to remove ".
                        "mount point directory '".$mount_pt."'\n"            .
                        join "\n", @output);

	return $ClientUtil::FALSE;
	
    } 

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: mount_entry_add()
# Objective:       add entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_add {
    my ($self,%args)=@_;
    my (
	$cmd,
	$device_mount,
	$device_fsck,
	$label,
	$mount_pt,
        $fs_freq,
	$fs_type,
	$fsck_auto,
	$mount_boot,
	$mount_options,
	$status,
	@output,
	);

    $self->_clear_err();
    
    #  initialize variables 
    $fs_freq="0";
    $fsck_auto="0";	
    $mount_options="nocheck";

    # parse arguments
    foreach (keys %args) {
	if (/-?device_mount/) {
	    if (not length($device_mount = $args{$_})) {
		$self->_set_err("Error: LinuxHost::mount_entry_add() Invalid value for argument '-device_mount'");
		return ($ClientUtil::FALSE); 
	    }        
	    next;
	}
	elsif (/-?device_fsck/) {
	    #  Not supported
	    next;
	}
	elsif (/-?mount_pt/) {
	    if (not length($mount_pt = $args{$_})) {
		$self->_set_err("Error: LinuxHost::mount_entry_add() Missing value for argument '-mount_pt'");
		return ($ClientUtil::FALSE); 
	    }
	    next;
	}
	elsif (/-?label/) {
	    if (not length($label = $args{$_})) {
		$self->_set_err("Error: LinuxHost::mount_entry_add() Missing value for argument '-label'");
		return ($ClientUtil::FALSE); 
	    }
	    next;
	}
	elsif (/-?fs_type/) {
	    if (not length($fs_type = $args{$_})) {
		$self->_set_err("Error: LinuxHost::mount_entry_add() Missing value for argument '-fs_type'");
		return ($ClientUtil::FALSE); 
	    }
	    next;
	}
	elsif (/-?fsck_auto/) {
	    if (not length($fsck_auto = $args{$_})) {
		$self->_set_err("Error: LinuxHost::mount_entry_add() Missing value for argument '-fsck_auto'");
		return ($ClientUtil::FALSE); 
	    }
	    next;
	}
 	elsif (/-?mount_boot/) {
	    #  Not supported
	    next;
	}

	elsif (/-?mount_options/) {
	    if (not length($mount_options = $args{$_})) {
		$self->_set_err("Error: LinuxHost::mount_entry_add() Missing value for argument '-mount_options'");
		return ($ClientUtil::FALSE); 
	    }
	    next;
	}
	else {
	    $self->_set_err("Error: LinuxHost::mount_entry_add() Invalid parameter specified \'$_\'");
	    return ($ClientUtil::FALSE);                       	  
	}
    }     
    
    #  Check that required parameter '-device' was provided
    unless (defined $device_mount and length $device_mount){
	$self->_set_err("Error: LinuxHost::mount_entry_add() ",
			"Missing required argument '-device_mount'");
	return $ClientUtil::FALSE;
    }

    #  Check that required parameter '-mount_pt' was provided
    unless (defined $mount_pt and length $mount_pt){
	$self->_set_err("Error: LinuxHost::mount_entry_add() ",
			"Missing required argument '-mount_pt'");
	return $ClientUtil::FALSE;
    }

    #  Check that required parameter '-fs_type' was provided
    unless (defined $fs_type and $fs_type=~/ext3|nfs/i){
	$self->_set_err("Error: LinuxHost::mount_entry_add() ",
			"Missing required argument '-fs_type'");
	return $ClientUtil::FALSE;
    }

    #  Mounting a partition uses different semantics
    if (length $label) {
	$cmd="echo LABEL=$label"."\t".$mount_pt."\t".$fs_type."\t".
	    $mount_options."\t".$fs_freq."\t".$fsck_auto." ".">> /etc/fstab";
    } else {
	$cmd="echo ".$device_mount."\t".$mount_pt."\t".$fs_type."\t".
	    $mount_options."\t".$fs_freq."\t".$fsck_auto." ".">> /etc/fstab";
    }

    $self->sanlog(-msg => "Adding ($fs_type) fs entry to mount table");

    #  execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
	$self->_set_err("Error: LinuxHost::mount_entry_add Unable to add mount entry to /etc/fstab file",@output);
	return $ClientUtil::FALSE;     
    } 
}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self,$mntpt)=@_;
    my (
        $cmd,
        $mount,
        $status,
	@output,
        );

    $self->_clear_err();

    #  Unmount the volume's filesystem
    $cmd="umount $mntpt";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
	$self->_set_err("Error:LinuxHost ::unmount_fs() Unable to unmount filesystem", @output);                      
    } 

    # $self->mount_entry_delete($mntpt);

    $self->sanlog(-msg => "Removing mount point '$mntpt'");

    #  Remove directory head
    $cmd = "rm -rf $self->{mount}";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
	$self->_set_err("Error: LinuxHost::unmount_fs() Unable to remove mount point '$mount'", @output);  
    } 

    return $self;
}

################################################################################
# Subroutine Name: mount_entry_delete()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_delete {
    my ($self, $mount) = @_;
    my (
	$cmd,
	$status,
	@output,
	);

    $self->_clear_err();

    #  Clear vfstab mount entry
    $self->sanlog(-msg => "Clear /etc/filesystems mount entry");

    unless (defined $mount and length($mount)) {
	$self->_set_err("Error: LinuxHost::mount_entry_delete Missing required paremeter '$mount'",@output);
	return $ClientUtil::FALSE;     
    }

    $mount=~s/\//\\\//g;

    $cmd ="perl -e\'".
	"open(FH,\"< \/etc\/fstab\") or exit(1);".
	    "open(FD,\"> \/etc\/fstab_new\") or exit(1);".
		"while(<FH>){\\\n".		     
		    "print FD \$_ if (\$_ !~ /$mount/);". 
			"}\\\n".
			    "close FD;".
				"close FH;\\\n".
				    "rename(\"/etc/fstab_new\", \"/etc/fstab\") or exit(1);".
					"\' ";

    
    #  execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
	$self->_set_err("Error: LinuxHost::mount_entry_delete Unable to remove mount entry from /etc/fstab file",@output);
	return $ClientUtil::FALSE;     
    } 

    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: mount_entry_verify
# Objective:      get_mount_entries
###############################################################################
sub mount_entry_verify{
    my ($self,$mount, $dev) = @_;
    my (
	$cmd,
        $entry,
        $idx,
    	$status,
        $value,
		@values,
        @params,
        @output,
        @buf,
	);

    $self->_clear_err();

    # initialize
    $entry="";
    $idx=0;

    # get the /etc/fstab contents
    if ($dev) {
	$cmd = "cat /etc/fstab | grep -v '^#' | egrep '^$mount' ";
    } else {
	$cmd = "cat /etc/fstab | grep -v '^#' | egrep '	$mount	| $mount ' ";
    }

    # execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
	$self->_set_err("Error: LinuxHost::mount_entry_verify(): Unable to verify mount file entry",@output);
	return $entry;
    } 

    if (scalar(@output)==0) {
	$self->_set_err("Error: LinuxHost::mount_entry_verify(): No mount file entry exists");
    }
    elsif(scalar(@output)>1){
	$self->_set_err("Error: LinuxHost::mount_entry_verify(): Multiple mount file entries exist",@output);	
    }
    else {
	@params=("Device to mount:","Mount pt:","Fs type:", "Mount Options:");
	
	@values=split(/\s+/,$output[0]); 	
	
	foreach $idx (0..(scalar(@params)-2)) {
	    push @buf, $params[$idx]." ".$values[$idx];
	}

	my $startOpt=(scalar(@params)-1);
	my $endOpt=(scalar(@values)-1);

	push @buf, $params[$startOpt]." @values[$startOpt..$endOpt]";    

    }  

    (wantarray()) ? @buf : \@buf;   
}

###############################################################################
# Subroutine Name: mount_table_verify
# Objective:       verify mount table
###############################################################################
sub mount_table_verify{
    my ($self,$mount) = @_;
    my (
        $status,
        @output,
    );

    $self->_clear_err();

    chop($mount) if ($mount =~ /\/$/);

    $mount=~s/\//\\\//g;

    # execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, "mount");
    if ($status > 0 or length("@output" ) eq 0) {
	$self->_set_err("Error: LinuxHost::mount_table_verify(): Unable to verify mount table", @output);
	return (wantarray()) ? @output : \@output;
    }

    unless (grep /$mount /, @output) {
	$self->_set_err("Error: LinuxHost::mount_table_verify(): Unable to verify mount table", @output);
	return (wantarray()) ? @output : \@output;
    }

    #  Need to parse the mount into some usefule structure.  For now just show
    #  that something was found.

    return (wantarray()) ? @output : \@output;
}

################################################################################
# Subroutine Name: filer_mount_entry_backup()
# Objective:       backup the filer /etc/exports file to /tmp for later restore
# Details:         See POD documentation at the end of this file
################################################################################
sub filer_mount_entry_backup {
	my ($self,$mount_pt)=@_;
    my (
	$cmd,
	$status,
	@output,
	);
    
    $self->_clear_err();

	unless(defined $mount_pt and length $mount_pt){
		$self->_set_err("Error: LinuxHost::filer_mount_entry_restore() mount_pt is a required argument");
		return $ClientUtil::FALSE;
	}
  
    #Save the original /etc/exports file for retore at cleanup
    $self->sanlog(-msg => "Backup the original /etc/exports file from \"$mount_pt\" for later restore");

	#backup original /etc/exports 
	$cmd="cp -p ${mount_pt}/etc/exports ${mount_pt}/etc/exports.ORIG";
	
    #  execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::filer_mount_entry_backup() Unable to backup \"${mount_pt}/etc/exports\"" , 
						"to \"${mount_pt}/etc/exports.ORIG\"",@output);
        return $ClientUtil::FALSE;     
    } 

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: filer_mount_entry_restore()
# Objective:       restore the filer /etc/exports file from /tmp 
# Details:         See POD documentation at the end of this file
################################################################################
sub filer_mount_entry_restore {
	my ($self,$mount_pt)=@_;
    my (
	$cmd,
	$status,
	@output,
	);
    
    $self->_clear_err();

	unless(defined $mount_pt and length $mount_pt){
		$self->_set_err("Error: LinuxHost::filer_mount_entry_restore() mount_pt is a required argument");
		return $ClientUtil::FALSE;
	}
  
    #Restore the original /etc/exports file
    $self->sanlog(-msg => "Restore the original /etc/exports file from \"$mount_pt\"");
	
	#restore original /etc/exports and remove old commented lines
	$cmd="cat ${mount_pt}/etc/exports.ORIG | grep -v '^#/vol' > ${mount_pt}/etc/exports";
	
    #  execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::filer_mount_entry_restore() Unable to restore \"${mount_pt}/etc/exports\"" , 
						"from \"${mount_pt}/etc/exports.ORIG\"",@output);
        return $ClientUtil::FALSE;     
    } 

    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: get_mounted_device
# Objective:       returns a mounted device name when passed a filesystem
###############################################################################
sub get_mounted_device {
    my ($self,$mount) = @_;
    my (
	$cmd,
	$line,
	$status,
        @output,
	);

    $self->_clear_err();

    chop($mount) if ($mount =~ /\/$/);

    $mount=~s/\//\\\//g;

    # execute the command
    $cmd = "mount | grep $mount | awk '{print ".'$1'."}'";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0 or length("@output" ) eq 0) {
	$self->_set_err("Error: LinuxHost::get_mounted_device(): Mount command failed!", @output);
	return undef; 
    } 

    if (scalar(@output) > 1) {
	$self->_set_err("Error: LinuxHost::get_mounted_device(): Multiple mount entries found!", @output);
	return undef;
    }

    #  if the device is a partition, return the physical disk
    chop($output[0]) if ($output[0] =~ /dev\/sd.+\d$/);

    return $output[0];   
}

###############################################################################
# Method:    get_lvm_default
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lvm_default {
    my ($self) = @_;

    $self->_clear_err();

    return qw(nativelvm);
}

###############################################################################
# Method:    get_supported_protocols
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_protocols {
    my ($self) = @_;

    $self->_clear_err();

    return qw(iscsi nfs);
}

###############################################################################
# Method:    get_unsupported_fs
# Objective: Get list of unsupported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_unsupported_fs {
    my ($self) = @_;

    $self->_clear_err();

    return (undef);
}

###############################################################################
# Method:    get_supported_fs_types
# Objective: Get list of supported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_fs_types {
    my ($self) = @_;

    $self->_clear_err();

    return qw(ext3 ext2);
}

###############################################################################
# Method:    get_nics
# Objective: Get nic addresses
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_nics_OBSOLETE {
    my ($self,%args) = @_;
    my (
	$cmd,
	$protocol,
	$status,
	$wwpn,
	@output,
	@nics,
	);

    $self->_clear_err();

    $protocol="fcp";

    # parse arguments
    foreach (keys %args) {
	if (/-?protocol/) {
	    if (not length($protocol = $args{$_}) and $protocol !~ /fcp|iscsi/i) {
		$self->_set_err("Error: LinuxHost::get_nics() Invalid value for argument '-type'");
		return (wantarray()) ? () : [];
	    }
	    next;
	}	
	else {
	    $self->_set_err("Error: LinuxHost::get_nics() Invalid parameter specified \'$_\'");
	    return (wantarray()) ? () : [];                      	  
	}
    } 

    if ($protocol=~/fcp/i){
	# get the list of AIX hardware
	$cmd = "lscfg -vl fcs\* | grep 'Network Address'";
	
	# execute the command
	($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
	if ($status > 0) {
	    $self->_set_err("Error: LinuxHost::get_nics(): Unable to get nic nodenames",@output);
	    return (wantarray()) ? () : [];     
	} 

	# on each adapter, run tdutil and get wwpns.
	foreach my $line (@output) {       
	    # get the value
	    $wwpn = (split(/\.+/, $line))[1];
	    if (length $wwpn){
		$wwpn=~s/(..)(..)(..)(..)(..)(..)(..)(..)/$1:$2:$3:$4:$5:$6:$7:$8/;
	    }

	    push (@nics, $wwpn);
	}
    }
    else {
	$self->_set_err("Error: LinuxHost::get_nics(): iSCSI protocol is currently not supported",@output);
	return (wantarray()) ? () : [];    
    }

    return (wantarray()) ? @nics: \@nics; 
}

################################################################################
# Subroutine Name: config_paths
# Objective:       N/A for Linux, provide for compatibility with HP-UX
# Details:         See POD documentation at the end of this file
################################################################################
sub config_paths {
    my ($self)=@_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: raw_device_setup
# Objective:       Do whatever we need to do to use raw devices successfully
#                  after the call to discover().
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_setup {
    my ($self)=@_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: raw_device_cleanup
# Objective:       Do whatever we need to do to make sure the host is more
#                  happy when luns are removed.  For linux, we want to do
#                  an iscsi stop.
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_cleanup {
    my ($self, $protocol)=@_;
    my (
	$cmd,
	$status,
	@output,
	);

    $self->_clear_err();

    if ($protocol =~ /iscsi/i) {
        #
        # Stop iscsi.  We want to do this before the luns are removed and we
        # get orphaned devices.  Once you have orphaned devices, an iscsi
        # stop and/or a reboot (which will try to stop iscsi) will hang.
        #
        $cmd = "/etc/init.d/iscsi stop";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        
        # determine if command worked
        if ($status gt 0) {
	    $self->_set_err("Error: LinuxHost::orphaned_device_cleanup() " .
                            "Command ($cmd) returned a bad status ($status)");
	    return "";     
        } 
    } 

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: orphaned_device_cleanup
# Objective:       When the luns are removed from the filer, we will have
#                  orphaned luns.  This causes big problems.  We should
#                  have stopped iscsi before we removed the luns so we are
#                  going to reboot and do a iscsi restart here.
# Details:         See POD documentation at the end of this file
################################################################################
sub orphaned_device_cleanup {
    my ($self, $protocol)=@_;
    my (
	$cmd,
	$status,
	@output,
	);

    $self->_clear_err();

    #
    # Reboot the host (note that iscsi should be stopped here.  If it's not
    # and there are orphaned devices, this will fail.
    #
    $self->reboot();
    if ($self->error) {	    
	$self->_set_err("Error: LinuxHost::orphaned_device_cleanup() Reboot failed");
	return $ClientUtil::FALSE;                             
    }

    if ($protocol =~ /iscsi/i) {
        # restart the iscsi service (may already be up)
        $cmd = "/etc/init.d/iscsi restart";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        
        # determine if command worked
        if ($status gt 0) {
	    $self->_set_err( "Error: LinuxHost::orphaned_device_cleanup() " .
                             "Command ($cmd) returned a bad status ($status)");
	    return "";     
        } 
    } 

    return $ClientUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Method:    _load_nfs_mnt
# Objective: Recreate stale NFS mount entries from previously failed tests
# Details:   See POD documentation at the end of this file
################################################################################
sub _load_nfs_mnt {
    my ($self)=@_;
    my (@data_ip,
	@filers,
	$i,
	@mounts,
	$mntObj,
	$param,
	$path,
	$share,
	$status,
	$value,
	);
    
    $self->_clear_err();

    # Initialize mount array.  Any found will be pushed here
    @{$self->{mounts}} = ();

    # Retrive list of FILERS.  This may seem kind of round-a-bout but we need to
    # call Tharn::param which appears to set the $Tharn::Params hash for a
    # given param
    Tharn::param("NODES", "-default", undef);
    Tharn::param("FILER", "-default", undef);

    $param = $Tharn::Params{NODES}->{value};
    unless (defined($param)) {
	$param = $Tharn::Params{FILER}->{value};
    }

    # Create a list of Tharn host objects.  Each object contains data stored
    # in FILER tharnhost files
    foreach (split(",", $param)) {
	push @filers, Tharn::host($_);
    }
	
    # Retrieve list of DATA networks from tharnhosts files
    foreach my $f (@filers) {
	# This is primarily used for classic support or any scenario where the
	# user is NFS mounting directories over any non-DATA network
	$value = $f->{default_ip};
	
	push @data_ip, $value if (defined $value);
	
	for (1 .. 10) {
	    # Retrieve DATA_IP[X] which should be defined in filer tharnhosts file
	    $param = "DATA_IP" . $_;
	    if (defined($f->{$param})) {
		$value = $f->{$param};
		push @data_ip, $value;
	    } else {
		last;
	    }
	}
    }

    ($status, @mounts) = &ClientUtil::execute($self->{conn}, "mount | grep nfs"); 
    
    if ($status gt 0) {
	$self->_set_err("LinuxHost::_load_nfs_mnt() Command (mount | grep nfs) returned a bad status ($status)");
	return $ClientUtil::FALSE;                             
    } 

    # Iterate through system mounts discovered above
    for $i (0 .. $#mounts) {
	# Look in system mounts for any NFS DATA IPs in use
	foreach (@data_ip) {
	    if ($mounts[$i] =~ /$_/) {
		# Retrieve mount share and path
		my @mnt = split(' ', $mounts[$i]);
		$share = $mnt[0];
		$path = $mnt[2];

		# Create a mount object for each system mount on DATA network
		$mntObj = $self->create_mnt(-parent=>$self, -path=>$path, -share=>$share);
	    }
	}
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _create_raw_binding()
# Objective:       create raw device binding
# Details:         See POD documentation at the end of this file
################################################################################ 
sub _create_raw_binding {
    my ($self,$block_device)=@_;
    my (
	$cmd,
	$found,
	$hsh,
	$major,
	$minor,
	$raw_device,
	$status,
	@output,
	);

    #  Defaults  
    $found=0;
    $hsh=$self->{raw_bindings};	
    ($major,$minor)=split(":",$block_device);

    #  Check if already bound	
    foreach my $key (keys %$hsh){
	if ($hsh->{$key}{major} eq $major and 
	    $hsh->{$key}{minor} eq $minor) {
	    $raw_device=$key;
	    $found=1;
	    last;      }		
    }

    #  If not bound get next availabe device
    if (not $found) {
	foreach my $key (keys %$hsh){
	    next if ($hsh->{$key}{major} ne 0);
	    $hsh->{$key}{major}=$major;
	    $hsh->{$key}{minor}=$minor;
	    $raw_device=$key;
	    last;
	}
	
	$cmd="raw $raw_device $major $minor";

	($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
	
	# determine if command worked
	if ($status gt 0) {
	    $self->_set_err( "Error: LinuxHost::_create_raw_binding() Command ($cmd) returned a bad status ($status)");
	    return "";                             
	} 

    }

    return $raw_device;
}

################################################################################
# Subroutine Name: _delete_raw_binding()
# Objective:       remove raw device binding
# Details:         See POD documentation at the end of this file
################################################################################ 

sub _delete_raw_binding {
    my ($self,$key)=@_;
    my (
	$hsh
	);

    $hsh=$self->{raw_bindings};	

    $hsh->{$key}{major}=0;
    $hsh->{$key}{minor}=0;

    return $ClientUtil::True;
}

################################################################################
# Subroutine Name: _get_devices
# Objective:       Get list of all devices on host
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_devices(){
    my ($self) = @_;
    my (
	$cmd,
	$status,
	@buffer,
	@hDevices,
	@output,
	@lvm_output,
	@realdevs,
        $hosttype,
	);
    
    $hosttype = $self->hosttype();
    
    # use sanlun lun show to get a list of devices
    $cmd="sanlun lun show all | grep '/dev'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);   							 
    if ($status gt 0) {
        $self->_set_err( "Error: LinuxHost::_get_devices() Command ($cmd) " .
                         "returned a bad status ($status)");
        return $ClientUtil::FALSE;                             
    }

    # use lvmdiskscan to get information that'll allow us to tell if a lun
    # is in an LV
    $cmd="lvmdiskscan | grep '/dev/s'";
    ($status, @lvm_output) = &ClientUtil::execute($self->{conn}, $cmd);   							 
    if ($status gt 0) {
        $self->_set_err( "Error: LinuxHost::_get_devices() Command ($cmd) " .
                         "returned a bad status ($status)");
        return $ClientUtil::FALSE;                             
    }

    # sanlun output looks like this:
    #   filer: lun-pathname     device filename   adapter    protocol  lun size        lun state
    #
    # myfiler: /vol/blocks/lun1 /dev/sdb          <unknown>0 iSCSI     5g (5368709120) GOOD
    foreach my $line (@output){
        # hDevs is a hash to hold each device.  Will be pushed
        # onto hDevices if it's good
	my $hDevs={};

        (my $filer, my $lun_path, my $device , my $adapter, my $protocol, my $lun_size, my $lun_state) = split /\s+/, $line;
        $lun_size =~ s/g/ GB/;
        $lun_size =~ s/m/ MB/;
        my $short_dev = $device;
        $short_dev =~ s/\/dev\///;
        $self->sanlog(-msg => "Sanlun lun show Line is $line",
                      -msg_level => 7);
        $self->sanlog(-msg => "Filer    is $filer\n"     .
                              "Lun      is $lun_path\n"  .
                              "Device   is $device\n"    .
                              "Short D  is $short_dev\n" .
                              "Adapter  is $adapter\n"   .
                              "Protocol is $protocol\n"  .
                              "Size     is $lun_size\n"  .
                              "State    is $lun_state\n" .
                              "Hosttype is $hosttype"    ,
                      -indent    => 4,
                      -msg_level => 7);

        # Check to make sure we have an iscsi disk and make sure the
        # lun has not ben orphaned
        if ( ($protocol ne 'iSCSI') or ($lun_path !~ /^\/vol\//)){
            next;
        }
        $hDevs->{path}="/dev";	
        $hDevs->{name}=$short_dev;
        $hDevs->{type}="NETAPP";
        $hDevs->{size}=$lun_size;
        $hDevs->{id}=$short_dev;

        # figure out if the lun is in an lvm we need this
        # for the label field in the device hash.
        foreach my $lvm_line (@lvm_output) {
            $self->sanlog(-msg => "lvmdiskscan Line is $lvm_line",
                          -msg_level=> 7);
            $hDevs->{label}="free";
            next unless $lvm_line =~ /$device/;
            $self->sanlog(-msg => "Found device $device in LVM line",
                          -msg_level=> 7);
            # format is:
            # lvmdiskscan -- /dev/sdb  [       9.77 GB] free whole disk
            # lvmdiskscan -- /dev/sdc  [       9.77 GB] LVM physical volume
            if ($lvm_line =~ / LVM /) {
                $hDevs->{label}="USED";
            }
            last;  # skip to the end of lvm_output
        } # end foreach lvmdiskscan line

        # return this device
        $self->sanlog(-msg => "Pushing device $line\n",
                      -msg_level => 3);
        push @hDevices,$hDevs;

    } # end foreach sanlun line

    return (wantarray()) ? @hDevices : \@hDevices; 
}

################################################################################
# Subroutine Name: _get_unlabeled_devices
# Objective:       Retrieve devices that do not have a label
# Details:         See POD documentation at the end of this file
################################################################################

sub _get_unlabeled_devices($){
    my ($self) = @_;
    my (
	$cmd,
	$status,
	@hDevices,
	@all_devices,
	@devices,
	);

    #  Retrieve all host devices
    @hDevices = $self->_get_devices();
    return 0 if ($self->error);

    #  Determine which devices are not being used by a vg
    foreach my $device (@hDevices){
	push (@devices,$device->{name}) if ( $device->{label}=~/free|new/i); 
    }
    
    #  Format the list of devices with label
    return (wantarray()) ? @devices : \@devices; 
}

################################################################################
# Subroutine Name: _get_release
# Objective:       Get vendor release
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_release{
    my ($self) = @_;
    my (
	$cmd,
	$release,
	$status, 
	@output,
	@hDevices,
	);
    
    # execute (as root)
    
    $cmd="cat /etc/redhat-release";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);   							 

    # determine if lsdev command worked
    if ($status gt 0) {
	$self->_set_err( "Error: LinuxHost::_get_release() Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;                             
    } 

    $release=$output[0];

    return $release;
}

################################################################################
# Subroutine Name: _load_raw_bindings()
# Objective:       load raw device bindings
# Details:         See POD documentation at the end of this file
################################################################################
sub _load_raw_bindings {
    my ($self,$hsh) = @_;
    my (
	$cmd, 
	@output, 
	$status,
	);
    
    $self->_clear_err();

    #  Create a hash of raw devices
    foreach (1..255){		
	my $key="/dev/raw/raw".$_;
	$hsh->{$key}{major}=0;
	$hsh->{$key}{minor}=0;			
    }

    

    #  Load current raw device bindings. 
    $cmd = "raw -qa";

    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err( "Error: LinuxHost::load_raw_bindings() Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;    
    } 
    
    foreach my $line (@output){
	#/dev/raw/raw1:  bound to major 58, minor 0
	my ($key,$major,$minor)=$line=~/(\S+):\s+.*major\s+(\d+),\s+minor\s+(\d+)/;	
	
        $hsh->{$key}{major}=$major;
	$hsh->{$key}{minor}=$minor;
	
    }  

    return $ClientUtil::TRUE;  
}

1;


