#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/Client/Host/SolarisHost.pm#1 $

## @summary Provide an interface to a HOST machine
## @author knopp@netapp.com
## @status private
## @pod ../Host.pod

use strict;

# Package Name
##############
package Client::Host::SolarisHost;

# Package Requirements
######################
use POSIX;
use Config;
use English;
use Net::Ping;
use San::SanClient::ClientUtil;
use San::SanLog;
use Error;
use vars qw( @EXPORT @ISA $NEW_DEV_LABEL);

$NEW_DEV_LABEL="MyDev";

@ISA=qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new SolarisHost object
# Details:         See POD documentation at the end of this file
################################################################################

sub _new (\%)
{
    my ($class, %args) = @_;

    my ($device_str, @devices);
    my ($self,$conn);

    
    if (defined $args{-self}) {         
                $self=$args{-self};
    }
    else  {
                $self={};
                # bless 
                bless $self,$class;
    }


    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: SolarisHost::new() Invalid value for -conn: $args{$_}");                 
                return $self;
            }
            next;
        }
        if (/-?self/) {
            if (not length($self = $args{$_})) {
                $self->_set_err("Error: SolarisHost::new() Invalid value for -self: $args{$_}");                 
                return $self;
            }
            next;
        }

        else {
            $self->_set_err("Error: SolarisHost::new() Invalid parameter specified \'$_\'");
            return $self; 
        }

    }

    unless (defined $conn and length $conn) {
        $self->_set_err("Error: SolarisHost::new()",
                        "Invalid value for argument '-conn'");
        return $self; 
    }

    $self->{conn}=$conn;

    $self->{gdb}="/usr/local/bin/gdb";

    # get information about the remote host
    #$self->_get_host_info();

    # include Disk, and Filesystem FS libraries 
    # (needs to be here after HOST_TYPE is set)
    #require "SanClient/Disk.pm";
    #require "SanClient/FS.pm";

    return($self);

}
################################################################################
# Subroutine Name: asl_version()
# Objective:       get asl version
# Details:         See POD documentation at the end of this file
################################################################################
sub asl_version {
    my ($self) = @_;
    my (
        $cmd, 
        @output, 
        $status,
        $version,
        );
    
    $self->_clear_err();

    $cmd = "/usr/bin/pkginfo -l NTAPasl | grep 'VERSION' | awk '{print \$2}'";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err( "Error: SolarisHost::asl_version() Command ($cmd) returned a bad status ($status)");
        return "";     
    } 
    
    
    unless (defined $output[0] and length $output[0]){
        $self->_set_err("Error: SolarisHost::asl_version(): Unable to get filer name");
        return "";
    }

    ($version)=$output[0]=~/V(.*)/i;

    return $version;      

}

################################################################################
# Subroutine Name: configure
# Objective: configure host machine
# Details: See POD documentation at the end of this file
################################################################################
sub configure {
    my ($self,%args)=@_;
    my (
        %hsh1,
        %hsh2,
        );

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        $self->sanlog(-msg => "configure key: $_",
                      -msg_level => 9);
        if(/-?filer/i){
            $hsh1{-filers}=$args{$_};
            $hsh2{-filers}=$args{$_};
            next;
        }
        elsif (/-?lunObj/i) {
            $hsh2{-lunObj}=$args{$_};
            next;
        }       
        elsif(/-?sd/i){
            $hsh2{-sd}=$args{$_};
            next;
        }
        else {
            $self->_set_err("Error: SolarisHost::configure() Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;                            
        }
    } 

    $self->_verify_lpfc_conf(%hsh1);
    return $ClientUtil::FALSE if ($self->error);

    #foreach my $filerObj (@{$hsh1{-filers}}){
    #   $self->sanlog(-msg => $filerObj->get_name());
    #   foreach my $targ (@{$self->{cfmap}{$filerObj->get_name()}}){
    #           $self->sanlog(-msg => "\t$targ");
    #   }
    #}

    $self->_verify_sd_conf(%hsh2);
    return $ClientUtil::FALSE if ($self->error);
    

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: config_paths
# Objective:       N/A for Solaris, provide for compatibility with HP-UX
# Details:         See POD documentation at the end of this file
################################################################################
sub config_paths {
    my ($self)=@_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: discover 
# Objective: Discover devices and label if new.
# Details: See POD documentation at the end of this file
################################################################################

sub discover ($)
{
    my ($self) = @_;
    my (
        $cmd,
        @devices,
        $status, 
        @output,
        @hDevices,
        );

    $self->_clear_err();

    $self->sanlog(-msg => "Discover all devices");

    # execute devfsadm (as root) cleanup device links
    $cmd = "/usr/sbin/devfsadm -C";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    # determine if devfsadm command worked
    if ($status gt 0) {
        $self->_set_err( "Error: SolarisHost::discover() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;                             
    } 
    
    $self->sanlog(-msg => "Locate devices that are not part of a volume group",
                  -msg_level => 3);
    @devices= $self->_get_unlabeled_devices();
    $self->sanlog(-msg => scalar(@devices)." Devices are not part of a volume group",
                  -msg_level => 3);

    return $ClientUtil::FALSE if ($self->error);

    
    ### Label new devices
    if (scalar(@devices) > 0) {
      $self->sanlog(-msg => "Volume name '$NEW_DEV_LABEL'",
                    -msg_level => 3);  


      $self->format(-devices=>\@devices, -label=>"$NEW_DEV_LABEL"); 

   
                                                                   
       return $ClientUtil::FALSE if ($self->error);
    }


    return $ClientUtil::TRUE ;
}

################################################################################
# Subroutine Name: format
# Objective:       Format devices with a label 
# Details:         See POD documentation at the end of this file
################################################################################

sub format { 
    my ($self,%args)=@_;
    my (
        $cmd,
        $devices,
        $label,
        $status,
        @format_cmds,
        @used,
        @output,
        @hDevices,      
        );

    # parse arguments
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: SolarisHost::format() Invalid value for argument '-devices'");
                return ($ClientUtil::FALSE); 
            }        
            next;
        }
        elsif (/-?label/) {
            if (not length($label = $args{$_})) {
                $self->_set_err("Error: SolarisHost::format() Missing value for argument '-label'");
                return ($ClientUtil::FALSE); 
            }
            next;
        }
        else {
            $self->_set_err("Error: SolarisHost::format() Invalid parameter specified \'$_\'");
            return ($ClientUtil::FALSE);                          
        }
    }     

    ### Check that required parameter '-device' was provided
    unless (defined $devices and ref($devices) eq "ARRAY"){
        $self->_set_err("Error: SolarisHost::format() ",
                        "Missing required argument '-device'");

        return 0;
    }

    ### Check that required parameter '-label' was provided
    unless (defined $label and length $label){
        $self->_set_err("Error: SolarisHost::format() ",
                        "Missing required argument '-label'");

        return 0;
    }

    # define location of some temporary files to conduct formats on.
    my $filename    = "/usr/tmp/format_input.$PID";

    @format_cmds = $self->_get_format_cmds($label);
    ClientUtil::create_remote_file($self->{conn}, $filename, @format_cmds);

    # iterate through the devices and label
    foreach my $device(@$devices) {     
        $self->sanlog(-msg => "Creating volume name on device '$device'");                

        # formulate the format command
        $cmd = "/usr/sbin/format -d $device -f $filename -s";

        # format device
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: SolarisHost::format() ",
                            "format returned a bad status ($status))");

            ### Cleanup temporary format inout files                                    
            &ClientUtil::remove_remote_file($self->{conn}, $filename);

            return 0;
        } 
        
    } 

    ### Cleanup temporary format inout files    
    &ClientUtil::remove_remote_file($self->{conn}, $filename);
    
    return 1;

}

################################################################################
# Method:    mkdir
# Objective: Create a directory on the client
# Details:   See POD documentation at the end of this file
################################################################################
sub mkdir {
    my ($self, %args) = @_;
    my ($cmd,
	$path,
	@output,
	$status);

    # parse arguments
    foreach (keys %args) {
	if (/-?path/) {
	    if (not length($path = $args{$_})) {
		$self->_set_err("Error: SolarisHost::mkdir() Invalid value for argument '-path'");
		return ($ClientUtil::FALSE); 
	    }        
	    next;
	} else {
	    $self->_set_err("Error: SolarisHost::format() Invalid parameter specified \'$_\'");
	    return ($ClientUtil::FALSE);                             
	}
    }     

    # check that required parameter '-path' was provided
    unless (defined($path) and length($path)){
	$self->_set_err("Error: SolarisHost::mkdir() Missing required argument '-path'");

        return $ClientUtil::FALSE;
    }

    $cmd = "mkdir $path";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    # determine if command worked
    if ($status gt 0) {
	$self->_set_err( "Error: SolarisHost::mkdir() Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;                             
    } 

    # verify existence of directory
    $cmd = "ls $path";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    # determine if command worked
    if ($status gt 0) {
	$self->_set_err( "Error: SolarisHost::mkdir() Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;                             
    } 
    
    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: get_mounted_device
# Objective:       returns a mounted device name when passed a filesystem
###############################################################################
sub get_mounted_device {
    my ($self,$mount) = @_;
    my (
        $cmd,
        $line,
        $status,
        @output,
        );

    $self->_clear_err();

    chop($mount) if ($mount =~ /\/$/);

    $mount=~s/\//\\\//g;

    # execute the command
    $cmd = "mount | grep $mount | awk '{print ".'$3'."}'";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0 or length("@output" ) eq 0) {
        $self->_set_err("Error: SolarisHost::get_mounted_device(): Mount command failed!", @output);
        return undef; 
    } 

    if (scalar(@output) > 1) {
        $self->_set_err("Error: SolarisHost::get_mounted_device(): Multiple mount entries found!", @output);
        return undef;
    }

    return $output[0];   
}

###############################################################################
# Method:    get_lvm_default
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lvm_default {
    my ($self) = @_;

    $self->_clear_err();

    return qw(vxvm);

}

###############################################################################
# Method:    get_supported_protocols
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_protocols {
    my ($self) = @_;

    $self->_clear_err();

    return qw(fcp nfs);

}

###############################################################################
# Method:    get_supported_fs_types
# Objective: Get list of supported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_fs_types {
    my ($self) = @_;

    return qw(ufs vxfs);

}

###############################################################################
# Method:    get_unsupported_fs
# Objective: Get list of unsupported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_unsupported_fs_OBSOLETE {
    my ($self) = @_;

    return ("ufs");

}

###############################################################################
# Method:    get_nics_OLD
# Objective: Get nic addresses
# Details:   See POD documentation at the end of this file.
###############################################################################

sub get_nics_OBSOLETE {
    my ($self,%args) = @_;
    my (
        $cmd,
        $protocol,
        $status,
        $wwpn,
        @output,
        @nics,
        );

    $self->_clear_err();

    $protocol="fcp";

    # parse arguments
    foreach (keys %args) {
        if (/-?protocol/) {
            if (not length($protocol = $args{$_}) and $protocol !~ /fcp|iscsi/i) {
                $self->_set_err("Error: SolarisHost::get_nics() Invalid value for argument '-type'");
                return (wantarray()) ? () : [];
            }
            next;
        }       
        else {
            $self->_set_err("Error: SolarisHost::get_nics() Invalid parameter specified \'$_\'");
            return (wantarray()) ? () : [];                               
        }
    } 

    if ($protocol=~/fcp/i){

        # get the list of AIX hardware
        $cmd = "sanlun fcp show adapter | grep WWPN";
        
        # execute the command
        ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: SolarisHost::get_nics(): Unable to get nic nodenames",@output);
            return (wantarray()) ? () : [];     
        } 

        # on each adapter, run tdutil and get wwpns.
        foreach my $line (@output){
            ($wwpn)=$line=~/WWPN:(\S+)/;        
            $wwpn=~s/(..)(..)(..)(..)(..)(..)(..)(..)/$1:$2:$3:$4:$5:$6:$7:$8/ if (length $wwpn);
            
            push @nics,$wwpn;
        }

        
    }
    else {
        $self->_set_err("Error: SolarisHost::get_nics(): iSCSI protocol is currently not supported",@output);
        return (wantarray()) ? () : [];    
    }

    return (wantarray()) ? @nics: \@nics; 

}

###############################################################################
# Method:    info
# Objective: Get client information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;
    my (
        $hsh,
        @output,
        );

    push @output,
    "Host name:      ".$self->hostname(),
    "Host type:      ".$self->hosttype(),
    "OS version:     ".$self->os_version(),
    (
     "Sanlun version: ".$self->sanlun_version(),
     "ASL version:    ".$self->asl_version(),
    ) if ($self->get_protocol()=~/fcp|iscsi/i);
    
    push @output,"\n";

    return (wantarray()) ? @output: \@output; 
}

################################################################################
# Subroutine Name: iostat
# Objective:       Format devices with a label 
# Details:         See POD documentation at the end of this file
################################################################################

sub iostat{ 
    my ($self,%args)=@_;
    my (
        $interval,
        $cmd,
        $status,
        @output,        
        $cnt,
        $count,
        @buf,
        );

    $self->_clear_err();

    ### Default interval is 2 seconds
    $interval=2;

    ### Default count is 1
    $count=1;

    # parse arguments
    foreach (keys %args) {
        if (/-?interval/) {
            if (not length($interval = $args{$_})) {
                $self->_set_err("Error: SolarisHost::iostat() Invalid value for argument '-interval'");
                return ($ClientUtil::FALSE); 
            }        
            next;
        }
        elsif (/-?count/) {
            if (not length($count = $args{$_})) {
                $self->_set_err("Error: SolarisHost::iostat() Missing value for argument '-count'");
                return ($ClientUtil::FALSE); 
            }
            next;
        }
        else {
            $self->_set_err("Error: SolarisHost::iostat() Invalid parameter specified \'$_\'");
            return ($ClientUtil::FALSE);                          
        }
    }     

    ### Remember that the first iostat cycle is wrong so add an extra cycle.
    ### Minimize the output return since 'z' flag doesn't kick in until second cycle.
    $count+=1;
    $cmd="iostat -xnzt $interval $count | grep -v '0.0    0.0    0.0    0.0'";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SolarisHost::iostat() ",
                        "iostat returned a bad status ($status))");

        return (wantarray()) ? () : [];
    } 

    ### Clean up the output by removing the first iostat cycle

    $cnt=0;
    foreach my $line (@output){
        $cnt++ if ($line=~/extended device/i);

        if ($cnt>1){
            push (@buf,$line);
        }
    }

    push @buf,"\n\n";
    return (wantarray()) ? @buf : \@buf;

}
################################################################################
# Subroutine Name: mount_entry_add()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_add {
    my ($self,%args)=@_;
    my (
        $cmd,
        $device_mount,
        $device_fsck,
        $mount_pt,
        $fs_type,
        $fsck_auto,
        $mount_boot,
        $mount_options,
        $status,
        @output,
        );

    $self->_clear_err();
    
    ### initialize variables 
    $fsck_auto="-";     
    $mount_boot="yes";
    $mount_options="-";

    # parse arguments
    foreach (keys %args) {
        if (/-?device_mount/) {
            if (not length($device_mount = $args{$_})) {
                $self->_set_err("Error: SolarisHost::mount_entry_add() Invalid value for argument '-device_mount'");
                return ($ClientUtil::FALSE); 
            }        
            next;
        }
        elsif (/-?device_fsck/) {
            if (not length($device_fsck = $args{$_})) {
                $self->_set_err("Error: SolarisHost::mount_entry_add() Missing value for argument '-device_fsck'");
                return ($ClientUtil::FALSE); 
            }
            next;
        }
        elsif (/-?mount_pt/) {
            if (not length($mount_pt = $args{$_})) {
                $self->_set_err("Error: SolarisHost::mount_entry_add() Missing value for argument '-mount_pt'");
                return ($ClientUtil::FALSE); 
            }
            next;
        }
        elsif (/-?fs_type/) {
            if (not length($fs_type = $args{$_})) {
                $self->_set_err("Error: SolarisHost::mount_entry_add() Missing value for argument '-fs_type'");
                return ($ClientUtil::FALSE); 
            }
            next;
        }
        elsif (/-?fsck_auto/) {
            if (not length($fsck_auto = $args{$_})) {
                $self->_set_err("Error: SolarisHost::mount_entry_add() Missing value for argument '-fsck_auto'");
                return ($ClientUtil::FALSE); 
            }
            next;
        }
        elsif (/-?mount_boot/) {
            if (not length($mount_boot = $args{$_})) {
                $self->_set_err("Error: SolarisHost::mount_entry_add() Missing value for argument '-mount_boot'");
                return ($ClientUtil::FALSE); 
            }
            next;
        }

        elsif (/-?mount_options/) {
            if (not length($mount_options = $args{$_})) {
                $self->_set_err("Error: SolarisHost::mount_entry_add() Missing value for argument '-mount_options'");
                return ($ClientUtil::FALSE); 
            }
            next;
        }
        else {
            $self->_set_err("Error: SolarisHost::mount_entry_add() Invalid parameter specified \'$_\'");
            return ($ClientUtil::FALSE);                          
        }
    }     

    
    ### Check that required parameter '-device' was provided
    unless (defined $device_mount and length $device_mount){
        $self->_set_err("Error: SolarisHost::mount_entry_add() ",
                        "Missing required argument '-device_mount'");
        return $ClientUtil::TRUE;
    }

    ### Check that required parameter '-device' was provided
    unless (defined $device_fsck and length $device_fsck){
        $self->_set_err("Error: SolarisHost::mount_entry_add() ",
                        "Missing required argument '-device_fsck'");
        return $ClientUtil::TRUE;
    }

    ### Check that required parameter '-device' was provided
    unless (defined $mount_pt and length $mount_pt){
        $self->_set_err("Error: SolarisHost::mount_entry_add() ",
                        "Missing required argument '-mount_pt'");
        return $ClientUtil::TRUE;
    }

    ### Check that required parameter '-device' was provided
    unless (defined $fs_type and $fs_type=~/vxfs|ufs/i){
        $self->_set_err("Error: SolarisHost::mount_entry_add() ",
                        "Missing required argument '-fs_type'");
        return $ClientUtil::TRUE;
    }

    $cmd="echo ".
         $device_mount."\t".
         $device_fsck."\t".
         $mount_pt."\t".
         $fs_type."\t".
         $fsck_auto."\t".
         $mount_boot."\t".
         $mount_options. " ".
         ">> /etc/vfstab";

     $self->sanlog("Adding ($fs_type) fs entry to mount table ($mount_pt)");

     ### execute the command
     ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
     if ($status > 0) {
         $self->_set_err("Error: SolarisHost::mount_entry_add Unable to add mount entry to /etc/vfstab file",@output);
         return $ClientUtil::FALSE;     
     } 

}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {        
    my ($self,$mntpt)=@_;
    my (
        $cmd,
        $mount,
        $status,
        @output,
        );

    $self->_clear_err();

    ### Unmount the volume's filesystem
    $cmd="umount $mntpt";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
        $self->_set_err("Error:SolarisHost ::unmount_fs() Unable to unmount filesystem", @output);                      
    } 

    $self->mount_entry_delete($mntpt);

    $self->sanlog(-msg => "Removing mount point '$mntpt'");

    ### Remove directory head
    $cmd = "rm -rf $self->{mount}";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SolarisHost::unmount_fs() Unable to remove mount point '$mount'", @output);                           
    } 

    return $self;
}

################################################################################
# Subroutine Name: mount_entry_delete()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################

sub mount_entry_delete {
        my ($self, $mount) = @_;
        my (
                $cmd,
                $status,
                @output,
                );

    $self->_clear_err();

    ### Clear vfstab mount entry
    $self->sanlog(-msg => "Clear vfstab mount entry");

        unless (defined $mount and length($mount)) {
                $self->_set_err("Error: SanClient::mount_entry_delete Missing required paremeter '$mount'",@output);
         return $ClientUtil::FALSE;     
        }

        if ($mount=~/dev\/vx/i) {
                $self->_set_err("Error: SanClient2::mount_entry_delete '$mount' reflects device entry and not mount point",@output);
        return $ClientUtil::FALSE;     
        }

    $mount=~s/\//\\\//g;

    $cmd="perl -e\'".
          "open(FH,\"< \/etc\/vfstab\") or exit(1);".
          "open(FD,\"> \/etc\/vfstab_new\") or exit(1);".
          "while(<FH>){\\\n".                
          "print FD \$_ if (\$_ !~ /$mount\\s/);". 
          "}\\\n".
          "close FD;".
          "close FH;\\\n".
          "rename(\"/etc/vfstab_new\", \"/etc/vfstab\") or exit(1);".
          "\' ";

        
     ### execute the command
     ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
     if ($status > 0) {
         $self->_set_err("Error: SanClient2::mount_entry_delete Unable to remove mount entry from /etc/vfstab file",@output);
         return $ClientUtil::FALSE;     
     } 

    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: mount_entry_verify
# Objective:      get_mount_entries
###############################################################################
sub mount_entry_verify{
    my ($self,$mount) = @_;
    my (
        $cmd,
        $entry,
        $idx,
        $status,
        $value,
            @values,
        @params,
        @output,
        @buf,
        );

    $self->_clear_err();

    # initialize
    $entry="";
    $idx=0;

    #$mount=~s/\//\\\//g;

    # get the sd.conf content
    $cmd = "cat /etc/vfstab | grep -v '^#' | egrep '    $mount  | $mount ' ";

    # execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
          $self->_set_err("Error: SolarisHost::mount_entry_verify(): Unable to verify mount file entry",@output);
          return $entry;
    } 

    $self->sanlog(-msg => \@output, -msg_level => 9);

    if (scalar(@output)==0) {
      $self->_set_err("Error: SolarisHost::mount_entry_verify(): No mount file entry exists");
    }
    elsif(scalar(@output)>1){
      $self->_set_err("Error: SolarisHost::mount_entry_verify(): Multiple mount file entries exist",@output);   
    }
    else {
          @params=("Device to mount:","Device to fsck:","Mount pt:","Fs type:",
                       "Fsck pass:","Mount at boot:","Mount Options:");
           
          @values=split(/\s+/,$output[0]);      
          
          foreach $idx (0..(scalar(@params)-2)) {
         push @buf, $params[$idx]." ".$values[$idx];
      }

      my $startOpt=(scalar(@params)-1);
      my $endOpt=(scalar(@values)-1);

      push @buf, $params[$startOpt]." @values[$startOpt..$endOpt]";    

    }  

    return (wantarray()) ? @buf : \@buf;   
}

###############################################################################
# Subroutine Name: mount_table_verify
# Objective:       verify mount table
###############################################################################
sub mount_table_verify{
    my ($self,$mount) = @_;
    my (
        $line,
        $status,
        @output,
        );

    $self->_clear_err();

    chop($mount) if ($mount =~ /\/$/);

    $mount=~s/\//\\\//g;

    # execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, "mount");
    if ($status > 0 or length("@output" ) eq 0) {
        $self->_set_err("Error: AIXHost::mount_table_verify(): Unable to verify mount entry", @output);
        return (wantarray()) ? @output : \@output; 
    } 

    unless (grep /$mount/, @output) {
        $self->_set_err("Error: AIXHost::mount_table_verify(): Unable to verify mount entry", @output);
        return (wantarray()) ? @output : \@output;
    }
    
    ### Need to parse the mount into some usefule structure.  For now just show
    ### that something was found.

    return (wantarray()) ? @output : \@output;   
}

################################################################################
# Subroutine Name: nfs_mount
# Objective:       NFS mount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_mount {
    my ($self,%args)=@_;
    my (
        $cmd,
        $force,
        $mount_pt,
        $options,
        $name,
        $status,
        $resource,
        @output,
        );
    
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) {
        if (/-?resource/) {
            if (not length($resource = $args{$_})) {
                $self->_set_err("Error: SolarisHost::nfs_mount() Invalid value for argument '-resource'");
                return ($ClientUtil::FALSE); 
            }        
            next;
        }
        elsif (/-?mount_pt/) {
            if (not length($mount_pt = $args{$_})) {
                $self->_set_err("Error: SolarisHost::nfs_mount() Missing value for argument '-mount_pt'");
                return ($ClientUtil::FALSE); 
            }
            next;
        }
        elsif (/-?options/) {
            if (not length($options = $args{$_})) {
                $self->_set_err("Error: SolarisHost::nfs_mount() Missing value for argument '-options'");
                return ($ClientUtil::FALSE); 
            }
            next;
        }
        else {
            $self->_set_err("Error: SolarisHost::nfs_mount() Invalid parameter specified \'$_\'");
            return ($ClientUtil::FALSE);                          
        }
    }     

    
    # Check that required parameter '-resource' was provided
    unless (defined $resource and length $resource){
        $self->_set_err("Error: SolarisHost::nfs_mount() ",
                        "Missing required argument '-resource'");
        return $ClientUtil::FALSE;
    }

    # Check that required parameter '-mount_pt' was provided
    unless (defined $mount_pt and length $mount_pt){
        $self->_set_err("Error: SolarisHost::nfs_mount() ",
                        "Missing required argument '-mount_pt'");
        return $ClientUtil::FALSE;
    }

    
    # Create the mount point and open up dir privs 
    $cmd="mkdir -p $mount_pt";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);   
    if ($status gt 0) {
        $self->_set_err( "Error: SolarisHost::nfs_mount() Unable to create mount point directory for volume '$resource'",
                         "Code: $status",
                         @output);
        
        return $ClientUtil::FALSE;                            
    } 


    # Create the NFS mount command.
    if (defined $options){
        $cmd= "mount -F nfs -o $options $resource $mount_pt;chmod 777 $mount_pt";
    }
    else {
        $cmd= "mount -F nfs $resource $mount_pt";
    }

    $self->sanlog(-msg => "Cmd: ".$cmd);
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    if ($status gt 0) {
        unless ($status==33) {
            $self->_set_err( "Error: SolarisHost::nfs_mount() Unable to NFS mount volume",
                             "Code: $status",
                             @output);
            
            return $ClientUtil::FALSE;
        }
    } 
   

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: nfs_unmount()
# Objective:       unmount NFS volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_unmount() {
    my ($self,%args)=@_;
    my (
        $cmd,
        $force,
        $mount_pt,
        $name,
        $status,
        @output,
        );
    
    $self->_clear_err();
    
    # parse arguments
    foreach (keys %args) {
        if (/-?mount_pt/) {
            if (not length($mount_pt = $args{$_})) {
                $self->_set_err("Error: SolarisHost::nfs_unmount() Missing value for argument '-mount_pt'");
                return ($ClientUtil::FALSE); 
            }
            next;
        }
        else {
            $self->_set_err("Error: SolarisHost::nfs_mount() Invalid parameter specified \'$_\'");
            return ($ClientUtil::FALSE);                          
        }
    }     

    # Check that required parameter '-mount_pt' was provided
    unless (defined $mount_pt and length $mount_pt){
        $self->_set_err("Error: SolarisHost::nfs_mount() ",
                        "Missing required argument '-mount_pt'");
        return $ClientUtil::FALSE;
    }

    
    # Create the mount point and open up dir privs 
    $cmd="umount $mount_pt";
    $self->sanlog(-msg => "Cmd: ".$cmd);
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);   
    if ($status gt 0) {
        unless ("@output"=~/not mounted/i){

            $self->_set_err( "Error: SolarisHost::nfs_unmount() Unable to unmount directory '$mount_pt'",
                             "Code: $status",
                             @output);
            
            return $ClientUtil::FALSE; 
        }
    } 

=comment

    $self->mount_entry_delete($mount_pt);
    if ($self->error) {
	$self->_set_err($self->errmsg());
	return $ClientUtil::FALSE;
    }

    ## Create the command to remove the NFS mount point 
    $cmd="rm -Rf $mount_pt";
    $self->sanlog(-msg => "Cmd: ".$cmd);
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd,-hide=>0); 

    if ($status gt 0) {
	$self->sanlog(-msg => "Warning: SolarisHost::nfs_unmount() Unable to remove ".
                              "mount point directory \"$mount_pt\"". join "\n",@output);
	
    } 

=cut

    return $ClientUtil::TRUE;


}

################################################################################
# Subroutine Name: reboot()
# Objective:       reboot host
# Details:         See POD documentation at the end of this file
################################################################################
sub reboot {
    my ($self) = @_;
    my (
        $cmd, 
        @output, 
        $status,
        );
    
    $self->_clear_err();

    $cmd = "reboot -- -r";
    $self->sanlog(-msg => "CMD=".$cmd);
    @output=&Tharn::say($self->{conn}, $cmd, '-blind');
    $self->sanlog(-msg => "CMD=DONE", -msg_level => 7);

    foreach my $line (@output){
        $self->sanlog(-msg => "line=$line\n",
                      -indent    => 4,
                      -msg_level => 7);
    }

    ### sleep 2 minutes to make sure
    
    sleep(120);

    # Ping the host every 10 seconds until it comes back
    my $alive=0;

    $cmd= "ping ".$self->hostname();

    foreach( my $cnt=0; $cnt<36; $cnt++){
        $output[0]=`$cmd`;
        $self->sanlog(-msg =>"Ping output = ".$output[0],
                      -msg_level => 7);
        if ($output[0]=~/alive/) {
            $self->sanlog(-msg => "Machine is alive after reboot");
            $alive=1;
            last;
        }
        else {
            sleep(10);
        }
    }

    unless($alive){
        $self->_set_err("Machine never rebooted");
        return $ClientUtil::FALSE;
    }

    # sleep for a while to make sure the host is all the
    # way back up.  Sometimes ping will work before telnet
    # will work.
    sleep(120);

    # reconnect to the client
    $self->reconnect();
    return $ClientUtil::FALSE if ($self->error);

    # run uname on the client so we know the connection/client are happy
    ($status, @output) = ClientUtil::execute($self->{conn}, "uname -a");
    if ($status > 0) {
        $self->_set_err( "Error: SolarisHost::reboot() Command ($cmd) " .
                         "returned a bad status ($status)");
        return "";     
    } 

    my $tempmsg = "Uname output = ";
    $tempmsg .= join "\n", @output;
    $self->sanlog(-msg => $tempmsg);

    $self->{reboot}=0;

    return $ClientUtil::PASS;     

}


################################################################################
# Subroutine Name: sanlun_version()
# Objective:       get sanlun version
# Details:         See POD documentation at the end of this file
################################################################################
sub sanlun_version {
    my ($self) = @_;
    my (
        $cmd, 
        @output, 
        $status,
        $version,
        );
    
    $self->_clear_err();

    $cmd = "sanlun version";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err( "Error: SolarisHost::sanlun_version() Command ($cmd) returned a bad status ($status)");
        return "";     
    } 
    
    
    unless (defined $output[0] and length $output[0]){
        $self->_set_err("Error: SolarisHost::sanlun_version(): Unable to get filer name");
        return "";
    }

    $version = $output[0];
    $version =~  s/.*\sversion\s(\d+.\d+).*/$1/;

    return $version;      

}
################################################################################
# Subroutine Name: raw_device_setup
# Objective:       Do whatever we need to do to use raw devices successfully
#                  after the call to discover().
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_setup {
    my ($self)=@_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: raw_device_cleanup
# Objective:       Do whatever we need to do to make sure the host is more
#                  happy when luns are removed.  For linux, we want to do
#                  an iscsi stop.
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_cleanup {
    my ($self)=@_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: orphaned_device_cleanup
# Objective:       When the luns are removed from the filer, we will have
#                  orphaned luns.  This causes big problems.  We should
#                  have stopped iscsi before we removed the luns so we are
#                  going to reboot and do a iscsi restart here.
# Details:         See POD documentation at the end of this file
################################################################################
sub orphaned_device_cleanup {
    my ($self)=@_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}


################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Method:    _load_nfs_mnt
# Objective: Recreate stale NFS mount entries from previously failed tests
# Details:   See POD documentation at the end of this file
################################################################################
sub _load_nfs_mnt {
    my ($self)=@_;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_devices
# Objective:       Get list of all devices on host
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_devices(){
    my ($self) = @_;
    my (
        $cmd,
        $cmdstr,
        $status, 
        @output,
        @hDevices,
        );
    
    # execute devfsadm (as root)
    $cmd="/usr/sbin/format";

    #$cmdstr = sprintf("$cmd\n%s\n%s\n%s%s"," "x20,"0","\004","echo' '");
    $cmdstr = sprintf("$cmd\n%s\n%s\n%s%s"," "x40,"0","\004","echo ' '");

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmdstr); 
    
    # determine if devfsadm command worked
    if ($status gt 0) {
        $self->_set_err( "Error: SolarisHost::_get_devices() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;                             
    } 

    foreach my $line (@output){
        chomp $line;
        # 2. c4t0d0 <NETAPP-LUN-0.2 cyl 2046 alt 2 hd 16 sec 128> Label
        # 3. c4t1d1 <IBM-Nseries2866-A20-7200 cyl 5118 alt 2 hd 16 sec 256> Label
        if ($line =~ /\d+.\s+(\w+.)\s+\<(\w+)-(\w+).*\>\s*(.*)/){
            my $hDevs={};
            $hDevs->{path}="/dev/rdsk"; 
            $hDevs->{name}=$1;
            my $vid = $2;  # vendor  ID
            my $pid = $3;  # product ID
            if ($vid eq "IBM" and $pid =~ /Nseries/){
                $hDevs->{type}="IBM-Nseries";
            } else {
                $hDevs->{type}=$vid;
            }
            $hDevs->{label}=$4;
            $hDevs->{name} =~ m/^\w+d(\d+)/;
            $hDevs->{id}=$1;
            push @hDevices,$hDevs;
        }
    }

    $self->sanlog(-msg => scalar(@hDevices)." devices found (includes multiple paths to devices)");

    return (wantarray()) ? @hDevices : \@hDevices; 
}

################################################################################
# Subroutine Name: _get_format_cmds
# Objective:       Get list of commands to format device
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_format_cmds ($)
{
    my ($self,$label) = @_;

    my (@part_commands);

    # create an array with the commands to pass to format

    # write out the volume name and label
    push(@part_commands, "volname $label");
    push(@part_commands, "label");
    # push(@part_commands, "verify");
    push(@part_commands, "quit");

    # return the list of commands to be executed.
    return(@part_commands);

}

###############################################################################
# Subroutine Name: _get_lpfc_conf
# Objective:       Get the content of the sd_conf object
###############################################################################
sub _get_lpfc_conf {
    my ($self) = @_;
    my (
        $cmd, 
        @fcp_tags,
        $format,
        $status,
        @output,
        @target,
        );

    # get the lpfc.conf content
    $cmd = "cat \/kernel\/drv\/lpfc.conf | grep :lpfc | grep -v '#'";

    #$cmd = "cat lpfc.conf | grep :lpfc | grep -v '#'";

    # execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 1) {
        $self->_set_err("Error: SolarisHost::_get_lpfc_conf() Unable to read lpfc.conf file",@output);
        return (undef,undef);     
    } 
    elsif ($status==1){
        # Inform user if no no target entries are found.
        $self->sanlog(-msg => "Warning: lpfc.conf file has no target entries");
    }

    
    ### Check to make sure that no more than 1 lpfc config exists
    my @fcp_tags = grep {/fcp-bind-/} @output;
    
    if (scalar(@fcp_tags)> 1){
        $self->_set_err("Error: SolarisHost::_get_lpfc_conf() Multiple fcp-bind tags in configuration",@output);
        return (undef,undef);   
    }


    foreach my $line (@output){
        my ($target)=$line=~/\"(\S+)\"/;
        push @target,$target;
    }


    $format=join("\n",@output);

    return ($format,@target);
    

}

###############################################################################
# Method:    _get_lpfc_format()
# Objective: retrieve lpfc.conf format for targets
# Details:   See POD documentation at the end of this file.
###############################################################################

sub _get_lpfc_format {
    my ($self,$last,$target)=@_;
    my (
        $cnt,
        $lpfc_str,
        $targ,
        $time,
        @output,
        );
    
   my $time=strftime("%Y-%m-%d %H:%M:%S",localtime());;
  

   ### Comment out the last lpfc string
   foreach $targ (@$last){

        $cnt++;

        if ($cnt eq 1) {
            $lpfc_str.=sprintf("# fcp-bind-WWPN=\"%s\"%s\n",
                           $targ,
                           (scalar(@$last)>1) ? "," : ";"
                           );
        }
        else {
            $lpfc_str.=sprintf("#%s\"%s\"%s\n",
                           " "x15,
                           $targ,
                           ($cnt < scalar(@$last)) ? "," : ";"
                           );

        }
        
    }

    ### Append the new lpfc string with time of modification
    $cnt=0;
    foreach $targ (@$target){

        $cnt++;

        if ($cnt eq 1) {
            $lpfc_str.="\n# Updated by SanClient($time)\n";
            $lpfc_str.=sprintf(" fcp-bind-WWPN=\"%s\"%s\n",
                           $targ,
                           (scalar(@$target)>1) ? "," : ";"
                           );
        }
        else {
            $lpfc_str.=sprintf("%s\"%s\"%s\n",
                           " "x15,
                           $targ,
                           ($cnt < scalar(@$target)) ? "," : ";"
                           );

        }
        
        
    }

    return $lpfc_str;

}

###############################################################################
# Subroutine Name: _get_sd_conf
# Objective:       Get the content of the sd_conf object
###############################################################################
sub _get_sd_conf {
    my ($self) = @_;
    my (
        $cmd, 
        $status, 
        @output
        );


    # get the sd.conf content
    $cmd = "cat /kernel/drv/sd.conf";
    
    # execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SolarisHost::_get_sd_conf(): Unable to cat sd.conf file",@output);
        return (wantarray()) ? () : [];     
    } 
    
    
    return (wantarray()) ? @output : \@output;
    

}
################################################################################
# Subroutine Name: _get_unlabeled_devices
# Objective:       Retrieve devices that do not have a label
# Details:         See POD documentation at the end of this file
################################################################################

sub _get_unlabeled_devices($){
    my ($self) = @_;
    my (
        @hDevices,
        @devices,
        );

    ### Retrieve all host devices
    @hDevices = $self->_get_devices();
    return 0 if ($self->error);

    ### Determine which devices do not have 
    ### a label
    foreach my $device (@hDevices){
        next if ($device->{type}!~/NETAPP|IBM-Nseries/);
        if ($device->{label} eq ""){
            push (@devices,$device->{name});
        }
    }
    
    ### Format the list of devices with label
    return (wantarray()) ? @devices : \@devices; 
}


###############################################################################
# Method:    _recreate_sd_conf()
# Objective: recreate sd conf file 
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _recreate_sd_conf { 
    my ($self,%args)=@_;
    
    my (
        $line_start,
        $line_cnt,
        $mode,
        $buffer,
        $buflen,
        $bufsize,
        $file,
        $offset,
        $tot_lines,
        @buf,
        @chunk,
        );

    ### Parameter 'bufsize' is the default number of lines for each 
    ### file write.  It will be reduced if its length exceeds 'buflen'.
    ### Writes exceeding 2500 bytes are randomly corrupted by Nate
    ### communication mechanism.

    $bufsize=60;
    $buflen=2000;

    $line_start=0;
    $buffer=$args{-buffer}; 
    $file=$args{-file};

    
    my $tot_lines=scalar(@{$buffer});

    while ($line_start <= $tot_lines){
        

        ### Modify buffer size if its bigger 
        ### than amount of actual lines to write
        if ($line_start+$bufsize > $tot_lines){
            $bufsize=$tot_lines-$line_start+1; 
        }
        
        my $line_cnt=$line_start+$bufsize-1;
        
        $offset=0;

        while (1) { 

            @chunk=();

            push @chunk,join("\n",@{$buffer}[$line_start..$line_cnt-$offset]);
            if (length "@chunk" > $buflen) {
                $offset+=5;
                next;
            }

            last;
        }

        $mode=($line_start==0) ? ">" : ">>"; 

        $self->sanlog(-msg => "CHUNK:",
                      -msg_level => 9);
        $self->sanlog(-msg => \@chunk,
                      -msg_level => 9,
                      -indent => 4);

        @buf=$self->execute("echo \'@chunk\' $mode $file",-hide=>1);
        return $self if ($self->error);
        
        ### Advance the line_cnt to the next chunk
        ### to write
        $line_start+=$bufsize-$offset;  
        
    }

}
###############################################################################
# Subroutine Name: _update_drv
# Objective:       Update the driver
###############################################################################
sub _update_drv {
    my ($self) = @_;
    my (
        $cmd, 
        $status, 
        @output
        );


    # get the sd.conf content
    $cmd = "update_drv -f sd";
    
    # execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SolarisHost::_update_drv(): Unable to reload sd driver",@output);
        return $ClientUtil::FALSE;    
    } 
    
    
    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: _devfsadm
# Objective:       Update the driver
###############################################################################
sub _devfsadm {
    my ($self) = @_;
    my (
        $cmd, 
        $status, 
        @output
        );


    # execute devfsadm (as root)
    $cmd = "/usr/sbin/devfsadm";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    # determine if devfsadm command worked
    if ($status gt 0) {
        $self->_set_err( "Error: SolarisHost::_devfsadm() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;                             
    } 
    
    return $ClientUtil::TRUE;

}

###############################################################################
# Method:    _verify_lpfc_conf()
# Objective: verify/update the lpfc.conf with required target entries
# Details:   See POD documentation at the end of this file.
###############################################################################

sub _verify_lpfc_conf {
    my ($self,%args)=@_;
    my (
        $change,
        $cmd,
        $line,
        $lpfc_current,
        $lpfc_new,
        $initName,
        $status,
        $total,
        $wwpn,
        %hLpfcTarg,
        @cmd,
        @current,
        @filers,
        @initObjs,
        @lpfc_current,
        @lpfc_new,
        %lpfc_merge,
        @targObjs,
        @output,
        );
    

    $change=0;

    # parse arguments
    foreach (keys %args){

        if (/-?filer/i) {
            if (length($args{$_}) and ref($args{$_}) eq "ARRAY"){
                foreach my $obj (@{$args{$_}}){
                    unless (ref($obj) eq "SanFiler"){
                        $self->_set_err("Error: SolarisHost::_verify_lpfc_conf() Invalid value for argument '-filerObjs'");
                        return $ClientUtil::FALSE;
                    }
                }

                @filers=@{$args{$_}};   
            }
            else {                      
                $self->_set_err("Error: SolarisHost::_verify_lpfc_conf() Invalid value for argument '-filerObjs'");
                return $ClientUtil::FALSE;
            }

            next;
        }       
        else {
            $self->_set_err("Error: SolarisHost::_verify_lpfc_conf() Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;                            
        }
    } 

    unless (scalar(@filers)){
        $self->_set_err("Error: SolarisHost::_verify_lpfc_conf() Argument array is empty");
        return  $ClientUtil::FALSE;                               
    }

    $self->sanlog(-msg => "Retrieve current lpfc.conf targets");

    ### Retrieve current lpfc target format and array of 
    ### target entries from config file
    ($lpfc_current,@lpfc_current)=$self->_get_lpfc_conf();
    return $self if ($self->error);

    ### Determine max targ id of current lpfc.conf    
    foreach my $targId (@lpfc_current) {
            my ($name,$id)=$targId=~/:(\S+)t(\d+)$/;
            $hLpfcTarg{$name}=($id>$hLpfcTarg{$name})?$id:$hLpfcTarg{$name};
        }

    # $self->sanlog(-msg => "MAX targ=".$hsh{$name});

    ### Retrieve new lpfc targets 
    $self->sanlog(-msg => "Retrieve proposed lpfc.conf targets");
    foreach my $filerObj (@filers){
        ### Retrieve target port objects on filer       
        @targObjs=$filerObj->get_targets(-dmp=>1); 
        if ($filerObj->error){
            $self->_set_err($filerObj->errmsg);
            return $ClientUtil::FALSE;
        }       
        
        ### Retrieve initiator port objects on client 
        @initObjs=$self->get_initiators();
        return $ClientUtil::FALSE if ($self->error);
        
        ### initialize hash to hold unique Client Filer Target ids
        my %hctid;

        foreach my $initObj (@initObjs){

            ### Retrieve initiator information
            $initName=$initObj->get_name();
            $wwpn=$initObj->get_portname();        

            ### Iterate over existing target object and check whether
            ### an lpfc conf entry already exists.
            foreach my $targObj (@targObjs){
                my $target=undef;

                if (grep(/$wwpn/i, $targObj->get_initiator_wwpns())){

                    ### Check to see if target already has a persistent
                    ### binding
                    
                    my $portname=$targObj->get_portname();
                    my ($target) = grep/$portname/,@lpfc_current;

                    ### If entry is new then grab the next target id by
                    ### incrementing the current max target id
                    if (not defined $target) {
                
                        my $targId=(defined $hLpfcTarg{$initName})? $hLpfcTarg{$initName}+1 : 0;
                          
                        ### Keep track of target ids used for the filer
                        $hctid{$targId}=1;
                                                
                        $target=
                                $targObj->get_portname().":".
                                $initName."t".
                                $targId;        
                    

                        $hLpfcTarg{$initName}++;
                    }
                    else {
                        ### A target already exists.  Add this target
                        # to the hash map controller/target.
                        my ($name,$id)=$target=~/:(\S+)t(\d+)$/;

                        $hctid{$id}=1;
                    }

                    ### Add target entries to new lpfc array 
                    push @lpfc_new,$target;                         
                    
                }                                                       
                
            }

        }
        
        push my @cftid,(sort keys %hctid);
        $self->{cfmap}{$filerObj->get_name()}=[@cftid];
    }
    

    ### Copy content of current config to a hash with 
    ### key being such that it can be sorted on by 'c*t*'
    foreach my $item (@lpfc_current) {
          my ($targ,$lpfc)=$item=~/(\S+):(\S+)/;
          $lpfc_merge{$lpfc.$targ}=$item;
    }


    ### If any new lpfc entry is not in the current file
    ### than flag that a change has occurred
    foreach my $item (@lpfc_new){
        if (not grep(/^$item$/i, @lpfc_current)){
            $change=1;
            ### add the new item to merge hash 
            my ($targ,$lpfc)=$item=~/(\S+):(\S+)/;
            $lpfc_merge{$lpfc.$targ}=$item;
        }
    }


    ### If lpfc needs a change
    if ($change==1){
        
        # Display the changes in the error message

        #       $self->_set_err("Error: Configurations are different",
        #                           "CURRENT:",@lpfc_current,"",
        #                           "NEW:",@lpfc_new);



        ### Display to the user the new bindings, as well as identify which ones
        # are new.
        $self->sanlog(-msg => "Changes to lpfc.conf targets",
                      -msg_level => 3);
        @lpfc_new=();
        foreach my $item (sort keys %lpfc_merge){
            my $value=$lpfc_merge{$item};
            my ($state)=(grep (/$value/,@lpfc_current))? "" : "*NEW*";
            push @lpfc_new,$value;
            
            $self->sanlog(-msg => $value." ".$state,
                          -msg_level => 3,
                          -indent    => 4);                            
        }
        
        $self->sanlog(" ");

        # set the reboot flag so users know to reboot the client
        $self->{reboot}=1;
        $self->sanlog(-msg =>
                    "New Target entries have been added to the lpfc.conf.  A REBOOT",
                    "will be required to see these new entries\n\n"
                    );
        
        ### Create new lpfc bind string.  The string should comment
        # out old and append new string.
        $lpfc_new=$self->_get_lpfc_format(\@lpfc_current,\@lpfc_new);

        ### If no bind string exists that append to 
        # end of lpfc.conf file
        if (scalar(@lpfc_current) eq 0){
            @cmd=("echo '$lpfc_new' >> \/kernel\/drv\/lpfc.conf");
        }
        else { 
            ### Existing binding exists so replace with new
            @cmd=("perl -e \' ",
                  "open (FH,\"<\/kernel\/drv\/lpfc.conf\") or exit(1); ",
                  "\$buffer.=\$_ while <FH>;",
                  "open (FD,\">\/kernel\/drv\/lpfc.conf\") or exit(1); ",                            
                  "\$buffer=~"."s/".$lpfc_current."/".$lpfc_new."/;",
                  "print FD \$buffer;",
                  "close FD;",
                  "close FH;",
                  "\' ");
        }
        # execute the command
        ($status, @output) = ClientUtil::execute($self->{conn}, "@cmd");
        if ($status > 0) {
            $self->_set_err("Error: SolarisHost::configure() Unable to modify lpfc.conf file",@output);
            return (wantarray()) ? () : [];     
        } 

        return $ClientUtil::FALSE;
    }
    else {
        $self->sanlog(-msg => "Configuration has not changed");
    }

    
    return $ClientUtil::TRUE;

}


###############################################################################
# Method:    _verify_sd_conf()
# Objective: verify/update the sd.conf with required sd entries
# Details:   See POD documentation at the end of this file.
###############################################################################

sub _verify_sd_conf {
    my ($self,%args)=@_;
    my (
        $change,
        $cmd,
        $detail,
        $status,
        $sd,
        %hsh,
        %hfilerlun,
        @buf,
        @detail,
        @filerObjs,
        @header,
        @ids,
        @lunObjs,
        @output,
        @tmp,
        );
    

    ### Initialize 
    $change=0,
    $detail=0;


    # parse arguments
    foreach (keys %args) {
        #$self->sanlog(-msg => "arg=$_");
        if (/-?lunObj/i) {
            if (length($args{$_}) and ref($args{$_}) eq "ARRAY"){
                foreach my $obj (@{$args{$_}}){
                    unless (ref($obj) eq "Lun"){
                        $self->_set_err("Error: SolarisHost::_verify_sd_conf() Invalid value for argument '-lunObjs'");
                        return $ClientUtil::FALSE;
                    }
                }

                @lunObjs=@{$args{$_}};   
            }
            else {                      
                $self->_set_err("Error: SolarisHost::_verify_sd_conf() Invalid value for argument '-initObjs'");
                return $ClientUtil::FALSE;
            }

            next;
        }
        elsif (/-?filer/i) {
            if (length($args{$_}) and ref($args{$_}) eq "ARRAY"){
                foreach my $obj (@{$args{$_}}){
                    unless (ref($obj) eq "SanFiler"){
                        $self->_set_err("Error: SolarisHost::_verify_sd_conf() Invalid value for argument '-filerObjs'");
                        return $ClientUtil::FALSE;
                    }
                }

                @filerObjs=@{$args{$_}};   
            }
            else {                      
                $self->_set_err("Error: SolarisHost::_verify_sd_conf() Invalid value for argument '-filerObjs'");
                return $ClientUtil::FALSE;
            }

            next;
        }       
        elsif(/-?sd/i){
            if (not length($sd = $args{$_})) {                  
                $self->_set_err("Error: SolarisHost::_verify_sd_conf() Invalid value for argument '-sdExtraEntries'");
                return $ClientUtil::FALSE;
            }
            
            next;

        }
        else {
            $self->_set_err("Error: SolarisHost::_verify_sd_conf() Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;                            
        }
    } 

    if (not scalar(@lunObjs)) {
        $self->sanlog(-msg => 
                    "Info: SolarisHost::_verify_sd_conf() Verification of sd.conf is ",
                    "not required since no new lun entries were created");
        return $ClientUtil::TRUE;
    }


    unless (scalar(@filerObjs)){
        $self->_set_err("Error: SolarisHost::_verify_sd_conf() Required -FilerObj' array is empty");
        return  $ClientUtil::FALSE;                               
    }


    ### Retrieve content of sd.conf file
    @buf=$self->_get_sd_conf();
    
    ### Create a hash of 'existing' targets and lun numbers 
    foreach my $line (@buf){
        if ($line=~/name.*parent.*target.*lun.*/){              
            my ($target,$lun)=$line=~/target=(\d+)\s+lun=(\d+)/;                                
            $hsh{$target}{$lun}=1;
            $detail=1;
        }

        push @header,"$line" unless ($detail);

    }
  
    ### Seperate luns by what filer they below too.     
    if (@lunObjs){
        $self->sanlog(-msg => "lun total=".scalar(@lunObjs),
                      -msg_level => 9);

        foreach my $filerObj (@filerObjs){
            my $filername=$filerObj->get_name();
            $self->sanlog(-msg => "Filer: ".$filername,
                          -msg_level => 9);
            my @lunids;
            foreach my $lunObj (@lunObjs){
                $self->sanlog(-msg => "lun filername = ".$lunObj->get_filername(),
                          -indent    => 4,
                          -msg_level => 9);
                if ($lunObj->get_filername()=~/$filername/i){
                    $self->sanlog(-msg => "Lun: ".$lunObj->get_name()." Id: ".$lunObj->get_id(),
                          -indent    => 8,
                          -msg_level => 9);
                    push @lunids,$lunObj->get_id();     
                }
                
            }
            $hfilerlun{$filername}{cfmap}=$self->{cfmap}{$filername};
            $hfilerlun{$filername}{luns}=\@lunids;
            
        }
    }

    ### Add new lun ids to existing hash        
    foreach my $filername (keys %hfilerlun){

        # Determine the max lunid for the set of
        # luns created. 
        my $max=0;


        ### Determine if any of the sd entries are new.
        foreach my $id (@{$hfilerlun{$filername}{luns}}){
            foreach my $target (@{$hfilerlun{$filername}{cfmap}}){
                $max=($id>$max)?$id:$max;
                next if (exists $hsh{$target}{$id});
                $hsh{$target}{$id}=0;
                $change=1;
            }
            
        }

        ### Add additional '$sd' entries to the sd.conf so a 
        # a reboot is nor required when new luns are created.
        if (defined $sd and length($sd)){
            $self->sanlog(-msg => "Adding $sd additional scsi device entries to ",
                                  "each filer '$filername' target");
            foreach my $target (@{$hfilerlun{$filername}{cfmap}}){
                foreach my $entry (1..$sd){
                    next if (exists $hsh{$target}{$max+$entry});
                    $hsh{$target}{$max+$entry}=0;
                    $change=1;
                }
            }
        }

    }

    ### Recreate the sd.conf file if new sd entries are required
    if ($change) {
        
        my @target=();
        foreach my $target (sort {($a<=>$b)} keys %hsh){                
            foreach my $lun (sort {($a<=>$b)} keys %{$hsh{$target}}){
                push @target,"name=\"sd\" parent=\"lpfc\" target=$target lun=$lun;";
            }
        }                           

        #$self->sanlog(-msg => "TARGET=",@target);

        # Create copy of existing sd.conf
        $cmd="cp /kernel/drv/sd.conf /kernel/drv/sd.conf.bck";
        ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: SolarisHost::configure() Unable to modify sd.conf file",@output);
            return $ClientUtil::FALSE;   
        } 

        # Recreate the sd.conf file
        $self->_recreate_sd_conf(
                                 -buffer=>[@header,@target],
                                 -file=>"/kernel/drv/sd.conf",
                                 );
        return $ClientUtil::FALSE if ($self->error);

        
        if ($self->os_version() =~ /5.9/) {
            $self->_update_drv();
            return $ClientUtil::FALSE if ($self->error);

	    $self->sanlog(-msg => "Reloading device drivers");
            $self->_devfsadm();
            return $ClientUtil::FALSE if ($self->error);

        }
        else {

            $self->{reboot}=1;
            $self->sanlog(-msg => 
                        "Additional entries were added to sd.conf - A REBOOT will",
                        "be required to see all additional luns"               
                        );
	    
        }
	
    }


    return $ClientUtil::TRUE;

}


1;

  
=pod

=head1 Package Name

SolarisHost.pm

=head1 Author

Gary Knopp

=head1 Description

For description of methods see B<San::SanClient::Host.pod>.


=cut
