#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/Client/Mount/LinuxMount.pm#1 $

## @summary Encapsulates host mount functionality on Linux
## @author lippe@netapp.com
## @status private
## @pod here

# Package Name
##############
package Client::Mount::LinuxMount;

use strict;
use Error;
use Exporter;

use San::SanClient::ClientUtil;

use vars qw (
	     @ISA
	     );

@ISA = qw(Error);

###############################################################################
# Method:    _new
# Objective: Create new Mount object
###############################################################################
sub _new {
    my ($class, %args) = @_;
    my ($options,
	$parent,
	$path,
	$persist,
	$self,
	$share);

    $self = {};

    bless $self, $class;

    foreach (keys %args) {
	if (/-?parent$/) {
	    length($parent = $args{$_}) or do {
		$self->_set_err("LinuxMount::_new invalid value for \"parent\": $args{$_}");
		return $self;
	    };
	}
	elsif (/^-?share$/) {
	    length($share = $args{$_}) or do {
		$self->_set_err("LinuxMount::_new invalid value for \"share\": $args{$_}");
		return $self;
	    };
	}
	elsif (/^-?path$/) {
	    length($path = $args{$_}) or do {
		$self->_set_err("LinuxMount::_new invalid value for \"path\": $args{$_}");
		return $self;
	    };
	}
	elsif (/^-?options$/) {
	    length($options = $args{$_}) or do {
		$self->_set_err("LinuxMount::_new invalid value for \"options\": $args{$_}");
		return $self;
	    };
	}
	else {
	    $self->_set_err("LinuxMount::_new: Invalid parameter specified: $_");
	    return $self;
	}
    }
    
    # check for existance of required -share argument
    unless (grep (/^-share$/, (keys %args))) {
	$self->_set_err("Error: LinuxMount::_new() Missing required argument '-share'");
	return $self; 
    }

    $self->{parent} = $parent;
    $self->{conn} = $parent->{conn};
    $self->{log} = $parent->{log};

    $self->{share} = $share;
    $self->{type} = ($self->{share} =~ /\/dev\//) ? "fcp" : "nfs";

    # Use either user defined path or dynamically create one
    if (defined $path) {
	$self->{path} = $path;
    } else {
	# Supply default path based on share name
	$self->{path} = $self->_get_free_path($share, '/t/');
	if ($self->{error}) {
	    return $self;                             
	} 
    }

    $self->{options} = $options if (defined $options);

    $self->_mount unless $self->is_mounted;
    if ($self->{error}) {
	return $self;                             
    } 

    # Size can only be determined after volume is mounted
    $self->{size} = $self->_get_size;

    return $self;
} 

################################################################################
# Method:    destroy
# Objective: Unmount filesystem and remove mount point
################################################################################
sub destroy {
    my ($self) = @_;
    my ($cmd,
	@mounts,
	@output,
	$mntprop,
	$status);

    $self->_clear_err();

    # Unmount filesystem
    $cmd = "$main::SUDO_PATH umount -f $self->{path}";
    
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    if ($status gt 0) {
	$self->_set_err( "Error: LinuxMount::destroy() Command ($cmd) Error removing filesystem", @output);
	return $ClientUtil::FALSE;                     
    } 

    # Give the system a couple seconds to unmount
    sleep 2;

    if ($self->is_mounted) {
	$self->_set_err( "Error: LinuxMount::destroy() Command ($cmd) Filesystem still mounted");
	return $ClientUtil::FALSE;                     
    }

    # Remove mount point (try 3 times in case of failure)
    for (1 .. 3) {
	$cmd = "$main::SUDO_PATH rm -rf $self->{path}";
	
	($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
	
	if ($status gt 0) {  
	    # This is a ridiculous hack but sometimes the rm happens before the unmount is finished.
	    sleep 2;                
	} else {
	    sleep 2;
	    last;
	}
    }

    if ($status gt 0) {
	$self->_set_err( "Error: LinuxMount::destroy() Command ($cmd) Unable to remove directory", @output);
	return $ClientUtil::FALSE;                     
    } 

    # Remove mount object from parent list.  Parent is a Host object
    @mounts = ();
    foreach ($self->{parent}->get_mounts) {
	$mntprop = $_->get_properties;
	push @mounts, $_ unless ($self->{path} eq $mntprop->{path});
    }

    @{$self->{parent}{mounts}} = @mounts;

    $self->{path} = $self->{share} = $self->{options} = undef;

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    is_mounted
# Objective: Verify filesystem is mounted
################################################################################
sub is_mounted {
    my ($self) = @_;
    my (
	$cmd,
        $status,
        @output,
    );

    $self->_clear_err();

    $cmd = "$main::SUDO_PATH mount | grep $self->{path}";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    if ($status gt 0) {
	$self->_set_err("LinuxMount::is_mounted() Command ($cmd) failed");
	return $ClientUtil::FALSE;                     
    } 

    if (@output) {
	$cmd = "df $self->{path}";
	
	($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
	
	if ($status gt 0) {
	    $self->_set_err("LinuxMount::is_mounted() Command ($cmd) failed, possible stale mount.");
	    return $ClientUtil::FALSE;                     
	} else {
	    return $ClientUtil::TRUE;
	}
    } else {
	return $ClientUtil::FALSE;
    }
}

################################################################################
# Method:    get_properties
# Objective: Return hash of mount object properties
################################################################################
sub get_properties {
    my ($self) = @_;
    my ($properties);

    $self->_clear_err();

    $properties->{share} = $self->{share} if (defined $self->{share});
    $properties->{path} = $self->{path} if (defined $self->{path});
    $properties->{options} = $self->{options} if (defined $self->{options});
    $properties->{type} = $self->{type} if (defined $self->{type});
    $properties->{size} = $self->{size} if (defined $self->{size});

    return $properties;
}

################################################################################
# Method:    _get_free_path
# Objective: Return unused path in specified directory based on supplied share
################################################################################
sub _get_free_path {
    my ($self, $share, $dir) = @_;
    my ($base_path,
	$cmd,
	$default_path,
	$host,
	$i,
	@output,
	$path,
	$status);
    
    $self->_clear_err();

    if ($self->{type} eq "nfs") {
	unless ($share =~ /(.+):.+/) {
	    $self->_set_err( "Error: LinuxMount::_get_free_path() $share syntax invalid");
	    return $ClientUtil::FALSE;        
	}
	$host = $1;
	$base_path = "nfs." . $host . '.';
    } else {
	$base_path = "fcp." . $share . '.';
    }

    $cmd = "mkdir -p $dir";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    if ($status gt 0) {
	$self->_set_err( "Error: LinuxMount::_get_nfs_path() Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;                     
    } 

    $cmd = "ls $dir";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    if ($status gt 0) {
	$self->_set_err( "Error: LinuxMount::_get_nfs_path() Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;                     
    } 

    while (1) {
	# Look in /t for directories of type nfs.172.16.0.1.0
	$default_path = $base_path . $i++;
	last unless grep(/$default_path/, @output);
    }

    return ($dir . $default_path);
}

################################################################################
# Method:    _mount
# Objective: Mount filesystem with defined params
################################################################################
sub _mount {
    my ($self) = @_;
    my ($cmd,
	@output,
	$status);

    $self->_clear_err();

    # Create direcotry on client
    $cmd = "$main::SUDO_PATH mkdir -p $self->{path}; $main::SUDO_PATH chmod 777 $self->{path}";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    if ($status gt 0) {
	$self->_set_err("Error: LinuxMount::_mount() Command ($cmd): Unable to create directory");
	return $ClientUtil::FALSE;                     
    } 

    # Mount filesystem on client
    if ($self->{options}) {
	$cmd = "$main::SUDO_PATH mount -o $self->{options} $self->{share} $self->{path}";
    } else {
	$cmd = "$main::SUDO_PATH mount $self->{share} $self->{path}";
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    if ($status gt 0) {
	$self->_set_err("Error: LinuxMount::_mount() Command ($cmd): Unable to mount filesystem");
	return $ClientUtil::FALSE;                     
    } 

    return;
}

################################################################################
# Method:    _get_size
# Objective: Return the size of a given client mount in Megabytes.
################################################################################
sub _get_size {
    my ($self) = @_;
    my ($cmd,
	@output,
	$status,
	$size);

    $self->_clear_err();

    # Use df to determine mount size
    $cmd = "df $self->{path}";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    if ($status gt 0) {
	$self->_set_err("Error: LinuxMount::_get_size() Command ($cmd): Unable to return size");
	return $ClientUtil::FALSE;                     
    } 

    foreach (@output) {
	if ($_ =~ /(\d+)\s+\d+\s+\d+\s+\d+\%/) {
	    $size = $1 / 1024;
	}
    }

    if ($size) {
	# return a round number
	return sprintf("%.0f", $size);
    } else {
	$self->_set_err("Error: LinuxMount::_get_size() Command ($cmd): Unable to return size");
	return $ClientUtil::FALSE;                     
    }
}

1;
