# $Id: //depot/prod/test/main/lib/ClientMgr.pm#2 $ 

## @summary Provides ability to control different tools on a pool of clients
##
## @owner GX Automation Framework
##
## @author Michael Edmonds (michael.edmonds@netapp.com)
## @author GX Automation Framework (dl-afd@netapp.com)
##
#
# Copyright (c) 2007,2008 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
##
## @status public
## @pod here

package ClientMgr;
use strict;
use List::Util qw[min max];
use Data::Dumper;
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF CODEREF BOOLEAN);

use TracePkg;
use base qw(Class);

use Process::Load;

our $Trace;
BEGIN
{
    $Trace = TracePkg->new();
}
# end BEGIN

sub new
{
    $Trace->entry();

    my ($class) = shift;
    my $hash = {clients    => {type => SCALAR|ARRAYREF, optional => 1},
                error_type => {type => SCALAR, default => Err::DIE()},};
    my %params = validate(@_, $hash);
    my $self={};
    bless $self, $class;

    if(defined($params{clients}))
    {
        if(ref($params{clients}) ne "ARRAY")
        {
            $params{clients} = [$params{clients}];
        }
        $self->{client_list} = $params{clients};
    }
    $self->{io} = undef;
    $self->_configure_error(type => $params{error_type});

    $Trace->exit();
    return $self;
}
# end new()

sub umount
{
    $Trace->entry();
    my $self = shift;
    my $hash = {id       => {type => SCALAR|ARRAYREF, optional => 1},
                map      => {type => HASHREF,         optional => 1},
                mount_pt => {type => SCALAR,          optional => 1},
                clients  => {type => SCALAR|ARRAYREF, optional => 1},};
    my %params = validate(@_, $hash);

    my %options;
    if(exists($params{id}))
    {
        $options{id} = $params{id};
    }
    if(exists($params{map}))
    {
        $options{map} = $params{map};
    }
    if(exists($params{mount_pt}))
    {
        $options{nfs_path} = $params{mount_pt};
    }
    if(exists($params{clients}))
    {
        $options{clients} = $params{clients};
    }

    $self->cleanup(%options);

    $Trace->exit();
}
# end umount()

sub mount
{
    $Trace->entry();
    my $self = shift;
    my $hash = {id         => {type => SCALAR|ARRAYREF, optional => 1},
                clients    => {type => SCALAR|ARRAYREF, optional => 1},
                map        => {type => HASHREF},
                mount_pt   => {type => SCALAR, optional => 1},
                mount_opts => {type => SCALAR, optional => 1},
                transport  => {type => SCALAR, optional => 1},
                version    => {type => SCALAR, optional => 1}};
    my %params = validate(@_, $hash);

    my %options = (map => $params{map});
    if(exists($params{id}))
    {
        $options{id} = $params{id};
    }
    if(exists($params{mount_pt}))
    {
        $options{nfs_options}->{path} = $params{mount_pt};
    }
    if(exists($params{mount_opts}))
    {
        $options{nfs_options}->{options} = $params{mount_opts};
    }
    if(exists($params{transport}))
    {
        $options{nfs_options}->{transport} = $params{transport};
    }
    if(exists($params{version}))
    {
        $options{nfs_options}->{version} = $params{version};
    }
    if(exists($params{clients}))
    {
        $options{clients} = $params{clients};
    }

    $self->setup(%options);

    $Trace->exit();
}
# end mount()

sub cleanup
{
    $Trace->entry();
    my $self = shift;
    my $hash = {id         => {type => SCALAR|ARRAYREF, optional => 1},
                map        => {type => HASHREF,         optional => 1},
                nfs_path   => {type => SCALAR,          optional => 1},
                cifs_share => {type => SCALAR,          optional => 1},
                clients    => {type => SCALAR|ARRAYREF, optional => 1},
                cleanup    => {type => SCALAR,           default => 0}};
    my %params = validate(@_, $hash);
    if(!defined($params{id}))
    {
        $params{id} = [$self->get_ids()];
    }
    elsif(ref($params{id}) ne "ARRAY")
    {
        $params{id} = [$params{id}];
    }

    if((!(defined($params{clients}))) && (!(defined($self->{client_list}))))
    {
        if(!(defined($params{map})))
        {
            foreach my $id (@{$params{id}})
            {
                if($self->{io}->{$id}->{obj}->type() !~ qr/code/i)
                {
                    if(!(defined($self->{io}->{$id}->{obj}->{client_list})))
                    {
                        $self->_result(rc => $self->CONFIG(),
                                       msg => "cleanup() requires clients to cleanup");
                    }
                }
            }
        }
        else
        {
            my @map_clients = keys(%{$params{map}});
            $params{clients} = \@map_clients;
        }
    }

    if(defined($params{clients}) && (ref($params{clients}) ne "ARRAY"))
    {
        $params{clients} = [$params{clients}];
    }
    my %dir;
    if(defined($params{nfs_path}))
    {
        $dir{nfs_path} = $params{nfs_path};
    }
    if(defined($params{cifs_share}))
    {
        $dir{cifs_share} = $params{cifs_share};
    }

    my $all_info = {};
    my $i = 0;
    foreach my $id (@{$params{id}})
    {
        my $obj = $self->{io}->{$id}->{obj};
        if($obj->type() !~ qr/code/i)
        {
            
            my @client_list;
            if(defined($params{clients}))
            {
                @client_list = @{$params{clients}};
            }
            else
            {
                @client_list = @{$self->{io}->{$id}->{obj}->{client_list}};
            }
            my %map = %dir;
            if(defined($params{map}))
            {
                foreach my $client(@client_list)
                {
                    my @maps = @{$params{map}->{$client}};
                    my %info = %{$maps[$i % (int(@maps))]};
                    $all_info->{$id}->{$client} = \%info;
                    $i++;
                }
                $map{map} = $all_info->{$id};
            }
            $obj->cleanup(%map, clients => \@client_list,
                          cleanup => $params{cleanup});
            if ($obj->is_error())
            {
                $Trace->warn(msg => "ERROR " . Dumper($obj->result()));
                return;
            }
        }
    }
    $self->_result();
    $Trace->exit();
}
# end cleanup()

sub setup
{
    $Trace->entry();
    my $self = shift;
    my $hash = {id           => {type => SCALAR|ARRAYREF, optional => 1},
                clients      => {type => SCALAR|ARRAYREF, optional => 1},
                map          => {type => HASHREF},
                nfs_options  => {type => HASHREF, optional => 1},
                cifs_options => {type => HASHREF, optional => 1},
                cleanup      => {type => SCALAR,  default  => 0}};
    my %params = validate(@_, $hash);

    if(!defined($params{id}))
    {
        $params{id} = [$self->get_ids()];
    }
    if(ref($params{id}) ne "ARRAY")
    {
        $params{id} = [$params{id}];
    }
    if((!(defined($params{clients}))) && (!(defined($self->{client_list}))))
    {
        foreach my $id (@{$params{id}})
        {
            if($self->{io}->{$id}->{obj}->type() !~ qr/code/i)
            {
                if(!(defined($self->{io}->{$id}->{client_list})))
                {
                    $self->_result(rc  => $self->CONFIG(),
                                   msg => "setup() requires clients to setup");
                }
            }
        }
    }

    if(defined($params{clients}) && (ref($params{clients}) ne "ARRAY"))
    {
        $params{clients} = [$params{clients}];
    }
    
    $Trace->debug(msg=>"Map = " . Dumper($params{map}));

    my $all_info = {};
    my $i = 0;
    my %options;
    if(exists($params{nfs_options}))
    {
        $options{nfs_options} = $params{nfs_options};
    }
    if(exists($params{cifs_options}))
    {
        $options{cifs_options} = $params{cifs_options};
    }

    foreach my $id (@{$params{id}})
    {
        my $obj = $self->{io}->{$id}->{obj};
        if($obj->type() !~ qr/code/i)
        {
            my @client_list;
            if(defined($params{clients}))
            {
                @client_list = @{$params{clients}};
            }
            else
            {
                @client_list = @{$self->{io}->{$id}->{client_list}};
            }
            $Trace->debug(msg=>"Clients = " . join(",", @client_list));
            foreach my $client(@client_list)
            {
                my @maps = @{$params{map}->{$client}};
                my %map_info = %{$maps[$i % (int(@maps))]};
                $all_info->{$id}->{$client} = \%map_info;
            }

            $i++;

            $obj->setup(%options, map => $all_info->{$id},
                        clients => \@client_list,
                        cleanup => $params{cleanup});
            if ($obj->is_error())
            {
                $Trace->warn(msg => "ERROR " . Dumper($obj->result()));
                return;
            }
        }
    }
    $self->_result();
    $Trace->exit();
}
# end setup()

sub HAMMER
{
    return Process::Load::HAMMER();
}
# end HAMMER()

sub SIO
{
    return Process::Load::SIO();
}
# end SIO()

sub CTHON
{
    return Process::Load::CTHON();
}
# end CTHON()

sub RANDOM
{
    return Process::Load::RANDOM();
}
# end RANDOM()

sub add_load
{
    $Trace->entry();

    # I had to shift out the $self so I can use it on the @clients validation
    my $self = shift;

    my $hash = {type       => {type => SCALAR,
                               callbacks => {test => sub{(grep ($_ =~ $_[0],
                                                         Process::Load::valid_io_types()))}}},
                bin_params => {type => HASHREF, default => {}},
                clients    => {type => SCALAR|ARRAYREF,
                               optional => defined($self->{client_list}),
                               default => $self->{client_list}},
                interactive=> {type => SCALAR, default=> 0,
                               callbacks => {testinter => sub {(int($_[0]) == 0)
                                                                  ||
                                                          (int($_[0]) == 1)}}},
                };
    my %params = validate(@_, $hash);

    my %clients;
    if(defined($params{clients}))
    {
        if(ref($params{clients}) ne "ARRAY")
        {
            $params{clients} = [$params{clients}];
        }
        %clients = (clients => $params{clients});
    }

    my $id = max(keys(%{$self->{io}}));
    if(!(defined($id)))
    {
        $id = -1;
    }
    $id++;

    my $io_obj = Process::Load->new(error_type => $self->{error_type},
                                    type       => $params{type},
                                    id         => $id,
                                    interactive=> $params{interactive},
                                    %clients);
    if($io_obj->is_error())
    {
        $Trace->comment(msg=>"ERROR " . Dumper($io_obj->result()));
        return;
    }

    $io_obj->bin_conf(%{$params{bin_params}});
    if($io_obj->is_error())
    {
        $Trace->comment(msg=>"ERROR " . Dumper($io_obj->result()));
        return;
    }

    $self->{io}->{$id} = {obj => $io_obj, desc => $io_obj->type() . $id};
    if(defined($params{clients}))
    {
        $self->{io}->{$id}->{client_list} = $params{clients};
    }
    $self->_result();
    $Trace->exit();
    return $id;
}
# end add_load()

sub add_code
{
    $Trace->entry();

    # I had to shift out the $self so I can use it on the @clients validation
    my $self = shift;

    my $hash = {code    => {type => CODEREF},
                args    => {type => SCALAR|ARRAYREF, default=>[]},
                verify  => {type => SCALAR|ARRAYREF, default=>[]},
                clients => {type => SCALAR|ARRAYREF,
                            optional => defined($self->{client_list}),
                            default  => $self->{client_list}},
                };
    my %params = validate(@_, $hash);

    my %clients;
    if(defined($params{clients}))
    {
        if(ref($params{clients}) ne "ARRAY")
        {
            $params{clients} = [$params{clients}];
        }
        %clients = (clients => $params{clients});
    }

    my $id = max(keys(%{$self->{io}}));
    if(!(defined($id)))
    {
        $id = -1;
    }
    $id++;

    my $obj = Process::CodeRef->new(code       => $params{code},
                                    args       => $params{args},
                                    verify     => $params{verify},
                                    error_type => $self->{error_type},
                                    id         => $id, %clients);
    if($obj->is_error())
    {
        $Trace->comment(msg=>"ERROR " . Dumper($obj->result()));
        return;
    }

    $self->{io}->{$id} = {obj => $obj, desc => $obj->type() . $id};
    if(defined($params{clients}))
    {
        $self->{io}->{$id}->{client_list} = $params{clients};
    }
    $self->_result();
    $Trace->exit();
    return $id;
}
# end add_code()

sub status
{
    $Trace->entry();
    my $self = shift;
    my $hash = {id  => {type => SCALAR|ARRAYREF, optional => 1}};
    my %params = validate(@_, $hash);
    if(!(defined($params{id})))
    {
        $params{id} = [keys(%{$self->{io}})];
    }
    elsif(ref($params{id}) ne "ARRAY")
    {
        $params{id} = [$params{id}];
    }

    my @status_vals;
    foreach my $id (@{$params{id}})
    {
        my $obj = $self->{io}->{$id}->{obj};
        my $status_hash = $obj->status();
        $Trace->debug(msg=>"Status " . Dumper($status_hash));
        $status_hash->{id} = $id;
        push(@status_vals, $status_hash);
        if($obj->is_error())
        {
            $Trace->comment(msg=>"ERROR " . Dumper($obj->result()));
            return;
        }
    }
    $Trace->exit();
    return @status_vals;
}
# end status()

sub results
{
    $Trace->entry();
    my $self = shift;
    my $hash = {id    => {type => SCALAR|ARRAYREF, optional => 1},
                level => {type => SCALAR|ARRAYREF, default  => 2,
                           callbacks =>{test => sub {(int($_[0]) > -1)
                                                              &&
                                                    (int($_[0]) < 3)},}}
                };
    my %params = validate(@_, $hash);
    if(!(defined($params{id})))
    {
        $params{id} = [keys(%{$self->{io}})];
    }
    elsif(ref($params{id}) ne "ARRAY")
    {
        $params{id} = [$params{id}];
    }

    my %results;
    foreach my $id (@{$params{id}})
    {
        my $obj = $self->{io}->{$id}->{obj};
        my @obj_results = $obj->results(level=>$params{level});
        if(scalar(@obj_results) > 0)
        {
            $results{$id} = \@obj_results;
        }
        if($obj->is_error())
        {
            return;
        }
    }
    $Trace->exit();
    return %results;
}
# end results()

sub get_ids
{
    $Trace->entry();
    my $self = shift;
    my @ids = keys(%{$self->{io}});
    $Trace->exit();
    return @ids;
}
# end get_ids()

sub get_description
{
    $Trace->entry();
    my $self = shift;
    my $hash = {id => {type => SCALAR|ARRAYREF, optional => 1},};
    my %params = validate(@_, $hash);
    $Trace->exit();
    if(!(defined($params{id})))
    {
        $params{id} = [$self->get_ids()];
    }
    elsif(ref($params{id}) ne "ARRAY")
    {
        $params{id} = [$params{id}];
    }
    my @desc;
    foreach my $id (@{$params{id}})
    {
        push(@desc, {id => $id, desc => $self->{io}->{$id}->{desc}});
    }
    return @desc;
}
# end get_ids()

sub start
{
    $Trace->entry();
    my $self = shift;
    my $hash = {id      => {type => SCALAR|ARRAYREF, optional => 1},
                clients => {type => SCALAR|ARRAYREF, optional => 1},};
    my %params = validate(@_, $hash);
    if(!(exists($params{id})))
    {
        $params{id} = [keys(%{$self->{io}})];
    }
    elsif(ref($params{id}) ne "ARRAY")
    {
        $params{id} = [$params{id}];
    }

    my %clients;
    if(exists($params{clients}))
    {
        if(ref($params{clients}) ne "ARRAY")
        {
            $params{clients} = [$params{clients}];
        }
        $clients{clients} = $params{clients};
    }

    foreach my $id (@{$params{id}})
    {
        my $obj = $self->{io}->{$id}->{obj};
        $obj->start(%clients);
        if($obj->is_error())
        {
            $Trace->comment(msg=>"ERROR " . Dumper($obj->result()));
            return;
        }
    }
    $Trace->exit();
    return;
}
# end start()

sub pause
{
    $Trace->entry();
    my $self = shift;
    my $hash = {id => {type => SCALAR|ARRAYREF, optional => 1},};
    my %params = validate(@_, $hash);
    if(!(defined($params{id})))
    {
        $params{id} = [keys(%{$self->{io}})];
    }
    elsif(ref($params{id}) ne "ARRAY")
    {
        $params{id} = [$params{id}];
    }
    foreach my $id (@{$params{id}})
    {
        my $obj = $self->{io}->{$id}->{obj};
        $obj->pause();
        if($obj->is_error())
        {
            if($obj->error_code() != $obj->NA())
            {
                $Trace->comment(msg=>"ERROR " . Dumper($obj->result()));
                return;
            }
            $obj->clear_result();
        }
    }
    $Trace->exit();
    return;
}
# end pause()

sub resume
{
    $Trace->entry();
    my $self = shift;
    my $hash = {id => {type => SCALAR|ARRAYREF, optional => 1},};
    my %params = validate(@_, $hash);
    if(!(defined($params{id})))
    {
        $params{id} = [keys(%{$self->{io}})];
    }
    elsif(ref($params{id}) ne "ARRAY")
    {
        $params{id} = [$params{id}];
    }
    foreach my $id (@{$params{id}})
    {
        my $obj = $self->{io}->{$id}->{obj};
        $obj->resume();
        if($obj->is_error())
        {
            if($obj->error_code() != $obj->NA())
            {
                $Trace->comment(msg=>"ERROR " . Dumper($obj->result()));
                return;
            }
            $obj->clear_result();
        }
    }
    $Trace->exit();
    return;
}
# end resume()

sub stop
{
    $Trace->entry();
    my $self = shift;
    my $hash = {id           => {type => SCALAR|ARRAYREF, optional => 1},
                hard         => {type => SCALAR,          default  => 0,
                                    callbacks => {test => sub{($_[0] == 0) || ($_[0] == 1)}}},
                status_check => {type => SCALAR, default => 1,
                                    callbacks => {test => sub{($_[0] == 0) || ($_[0] == 1)}}},
                clients      => {type => SCALAR|ARRAYREF, optional => 1},
                };
    my %params = validate(@_, $hash);
    if(!(defined($params{id})))
    {
        $params{id} = [keys(%{$self->{io}})];
    }
    elsif(ref($params{id}) ne "ARRAY")
    {
        $params{id} = [$params{id}];
    }

    my %clients;
    if(exists($params{clients}))
    {
        if(ref($params{clients}) ne "ARRAY")
        {
            $params{clients} = [$params{clients}];
        }
        $clients{clients} = $params{clients};
    }

    foreach my $id (@{$params{id}})
    {
        my $obj = $self->{io}->{$id}->{obj};
        $obj->stop(hard => $params{hard}, status_check => $params{status_check},
                   %clients);
        if($obj->is_error())
        {
            if($obj->error_code() != $obj->NA())
            {
                $Trace->comment(msg=>"ERROR " . Dumper($obj->result()));
                return;
            }
            $obj->clear_result();
        }
    }
    $Trace->exit();
    return;
}
# end stop()

sub available_loads
{
    return Process::Load::valid_io_types();
}
# end available_loads()

1;


#############################################################################
#                                   POD
#############################################################################

=pod

=head1 Package Name

ClientMgr.pm

=head1 Author

Michael Edmonds

=head1 Owner

GX Automation Framework (dl-afd@netapp.com)

=head1 Description

ClientMgr (Client Manager) provides some mechanism for creating multiple
Process::Load events on a pool of clients.

=head1 See Also

Class.pm, Process.pm, Process::Load.pm, Process::CodeRef.pm

Package inherits from the following modules. See these modules for further 
interface detail.

Err.pm

Class.pm

=head1 Synopsis

Here are some examples. 

=over 4

    my $cMgr_obj = ClientMgr->new(clients => \@Clients);

    my $hmr_id = $cMgr_obj->add_load(type => $cMgr_obj->HAMMER());

    my $map = $Cfg->get_client_map();

    $cMgr_obj->setup(id => $hmr_id, map  => $map);
    $cMgr_obj->start();
    ...
    # Do Stuff
    ...
    $cMgr_obj->stop();
    $cMgr_obj->cleanup();

=head1 Public Methods

=head2 C<new(%params)>

=over 4

=item Description

Creates a new ClientMgr object.

=item Inputs

=item * clients => <\@Clients | $client>

(Optional)

Single Client, or an array reference of Clients

=item * error_type => <Err::DIE() | Err::RETURN()>

(Optional)

Default = Err::DIE()

this tells Error Handler to either do a die on an error or return no matter what
    and let the user handle the error

=item Outputs

ClientMgr object Reference.

=item Examples

    # Create a Client Manager Object 
    my $cMgr_obj = ClientMgr->new(clients=>\@Clients);
    if ($cMgr_obj->is_error())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }

=back

=head2 C<add_load(%params)>

=over 4

=item Description

Adds a type of load to the Client Manager that you wish to run in the background

=item Inputs

=item * type => <string>

Binary to run: "hammer", "sio", "cthon", ...

=item * bin_params => <HashRef>

The named parameters for "bin_params" are unique to each load type that you choose.

=item * clients => \@Clients

Optional (if clients passed in elswhere)

An array reference of Clients that this Load Generator will run on.
    (it can also take in a $singleClient, but wouldn't be common)

=item Outputs

(Integer) ID

=item Examples

    $cMgr_obj->add_load(type => $cMgr_obj->HAMMER());

    $cMgr_obj->add_load(type => $cMgr_obj->CTHON());

    --------------------------   OR   -------------------------------------

    my $hmr_conf = {bin => "/usr/local/test/bin/hammer",
                    filer_core => 0,
                    options => ""};

    my $hmr_id = $cMgr_obj->add_load(type => $cMgr_obj->HAMMER(),
                                     bin_params => $hmr_conf,
                                     clients => \@Clients);
    if ($cMgr_obj->is_error())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }
    
    my $cthon_conf = {bin_base => "/usr/software/test/bin/cthon04",
                      group    => 'a',
                      type     => 't'};

    my $cthon_id = $cMgr_obj->add_load(type => $cMgr_obj->CTHON(),
                                       bin_params => $cthon_conf,
                                       clients => \@Clients);
    if ($cMgr_obj->is_error())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }

=back

=head2 C<setup(%params)>

=over 4 

=item Description

This method sets up internal object information
and (if necessary) mounts/maps/... clients to the nodes for the load ids.

=item Inputs

=item * map => \%map

(Required)

This should be passed in from Knopp's Config::Cluster.pm,
    but it can be created by hand too

=item * clients => \@Clients

(Optional if passed in elsewhere)

An array reference of Clients that will be mounted.
    (it can also take in a $singleClient, but wouldn't be common)

=item * id => <$ID | \@IDs>

(Optional)

An integer or an arrayref of integers that was returned to you from the add_load() method.

=item * nfs_options => <\%HashRef>

(Optional)

$nfs_options->{path} = <Mount Point Path>
    (Default = the method generates the mount point path)
    Instead of generating a path, it will use this path to mount to.
        If this is Not an ABSOLUTE PATH, it will automatically prepend the path with:
        "/t/<thpl name>/"

$nfs_options->{options} = <Mount options>
    (Default = "hard,intr" + <possibly other options depeding on the Load Gen>)

$nfs_options->{transport} = <udp|tcp>
    (Default = "tcp")

$nfs_options->{version} = <1 .. 4>
    (Default = 3)

=item * cifs_options => <\%HashRef>

(Optional)

$cifs_options->{workgroup} = <workgroup of which the node would be part of>
    (Default = "WORKGROUP")

$cifs_options->{domain}    = <domain where the username exists>

$cifs_options->{username}  = <username on the domain/workgoup which has access to the share>

$cifs_options->{password}  = <password of the "username">

$cifs_options->{drive}     = <A-Z|a-z>
    (Default = "*" (mapping the share to next available drive letter))

=item * cleanup => <0 | 1>

(Optional)

Default = 0

Setting this to true would cause it to try to delete all files that are
    currently on the volume

=item Outputs

None

=item Examples

    $cMgr_obj->setup();

    ------------------------   OR   -----------------------------------

    $cMgr_obj->setup(id => $hmr_id, clients  => \@Clients, map => $map,
                     nfs_options => {version => 3, transport => "tcp"});
    if ($cMgr_obj->is_error())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }

=back

=head2 C<mount(%params)>

B<OBSOLETE: use setup() instead>

=over 4 

=item Description

This method mounts those clients for the load ids.

=item Inputs

=item * map => %map

(Required)

This should be passed in from Knopp's Config::Cluster.pm,
    but it can be created by hand too

=item * clients => \@Clients

(Optional if passed in elsewhere)

An array reference of Clients that will be mounted.
    (it can also take in a $singleClient, but wouldn't be common)

=item * id => <$ID | \@IDs>

(Optional)

An integer or an arrayref of integers that was returned to you from the add_load() method.

=item * mount_pt => $dir_path

(Optional)

This is a specific directory that you with to mount to.
If this is not passed in, one will be generated.

TODO:

Clean up the directories

this is not currently done in the tools we have now, but it should be done

=item * mount_opts => $options

(Optional)

default = "hard,intr"

Do not include the Transport and Version in this.
in the hop to make it easier programatically, those values are passed in as diff names

=item * transport => <"tcp" | "udp">

(Optional)

default = "tcp"

transport protocol

=item * transport => <1 | 2 | 3| 4>

(Optional)

default = 3

NFS protocol Version

=item Outputs

None

=item Examples

    $cMgr_obj->mount();

    ------------------------   OR   -----------------------------------

    $cMgr_obj->mount(id => $hmr_id, clients => \@Clients, map  => $map,
                    mount_opts => "hard,intr,noac,rsize=32768,wsize=32768",
                    version    => 3,
                    transport  => "tcp");
    if ($cMgr_obj->is_error())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }

=back

=head2 C<add_code(%params)>

=over 4

=item Description

Adds a Function / Method to the Client Manager that you wish to run in the
    background

=item Inputs

=item * code => <\&coderef>

(Required)

Can be a method from a class (\&Server::run_cmd) or just some function (\&func)

=item * args => <string | array_ref>

(Optional)

The parameters that will be passed to the code reference.

=item * verify => <string | array_ref>

(Optional)

The return output from the code refence will be verified against this stuff.
(If you want to verify a hash, push it into an array first)

=item * clients => \@Clients

Optional (if clients passed in elswhere)

An array reference of Clients that this code will run on.
    (it can also take in a $singleClient)

=item Outputs

(Integer) ID

=item Examples

    #This is similar to doing "my ($out1, $out2, $out3) = $some_obj->method($var1);"
    $cMgr_obj->add_code(code => \&SomeClass::method, args => [$some_obj, $var1],
                        verify => ["blah, 1, $something], clients => \@Clients);


=back

=head2 C<get_ids()>

=over 4

=item Description

This method returns all of the IDs of the objects that it knows about

=item Input

None

=item Outputs

Returns an Array of IDs that is has that represent the objects
(because I was told that returning objects made things "too complex")

=item Example

    my @ids = $cMgr_obj->get_ids();
    $cMgr_obj->setup(id => \@ids, clients => \@Clients, map  => $map);
    if ($cMgr_obj->is_error())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }

=back

=head2 C<get_description(%params)>

=over 4

=item Description

This method returns all of the IDs of the objects that it knows about

=item Inputs

=item * id => <$ID | \@IDs>

(Optional)

An integer or an arrayref of integers that was returned to you from the add_load() method.

=item Outputs

Returns an Array of hash references.

each element in the array will be like:

{id => <id>, desc => <description string>}

=item Example

    my @ids = $cMgr_obj->get_ids();
    foreach(@ids)
    {
        $cMgr_obj->start(id => $_);
        if ($cMgr_obj->is_error())
        {
            my $desc_hash = ($cMgr_obj->get_description(id => $_))[0];
            $Trace->fatal(msg => "problem starting " . $desc_hash->{desc});
        }
    }

=back


=head2 C<cleanup(%params)>

=over 4

=item Description

This method cleans up the object internals
and (if necessary) umounts/unmaps/... the clients for a particular id

=item Inputs

=item * id => <$ID | \@IDs>

(Optional)

An integer or an arrayref of integers that was returned to you from the add_load() method.

=item * clients => \@Clients

(Optional)

Default = all clients that it the ID knows about

Specifies the Client[s] to disconnect from the node

=item * map => \%map

(Optional)

Specifies some client/node maping

This should be passed in from Knopp's Config::Cluster.pm,
    but it can be created by hand too

=item * nfs_path => <full path of mount point>

(Optional)

This is if you want to umount a specific dir

=item * cifs_share => <\\\\server\\share>

(Optional)

This is if you want to unmap a specific map

=item * cleanup => <0 | 1>

(Optional)

Default = 0

Setting this to true would cause it to try to delete everything that this
    object wrote to the volume

=item Output

None

=item Examples

    $cMgr_obj->cleanup();
    if ($cMgr_obj->is_error())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }

=back

=head2 C<umount(%params)>

=over 4

B<OBSOLETE: use cleanup() instead>

=item Description

This method umounts all the clients for a particular id

=item Inputs

=item * id => <$ID | \@IDs>

(Optional)

An integer or an arrayref of integers that was returned to you from the add_load() method.

=item Output

None

=item Examples

    $cMgr_obj->umount();
    if ($cMgr_obj->is_error())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }

=back

=head2 C<start(%params)>

=over 4

=item Description

Start ID or IDs

=item Inputs

=item * id => <$ID | \@IDs>

(Optional)

An integer or an arrayref of integers that was returned to you from the add_load() method.

=item * clients => <\@Clients | $client>

(Optional)

Default = (All Clients that the ID knows about)
This Specifies which Clients in the ID start

=item Output

None

=item Example

    $cMgr_obj->start();
    if ($cMgr_obj->is_error())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }

=back

=head2 C<pause(%params)>

=over 4

=item Description

"potentially" pause the ID or IDs

=item Inputs

=item * id => <$ID | \@IDs>

(Optional)

An integer or an arrayref of integers that was returned to you from the add_load() method.

=item Output

None

=item Example

    $cMgr_obj->start();
    ...
    $cMgr_obj->pause();
    if ($cMgr_obj->is_error() && ($cMgr_obj->error_code() != $cMgr_obj->NA())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }
    ...
    $cMgr_obj->resume();
    if ($cMgr_obj->is_error() && ($cMgr_obj->error_code() != $cMgr_obj->NA())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }

=back

=head2 C<resume(%params)>

=over 4

=item Description

"potentially" resume the ID or IDs

=item Inputs

=item * id => <$ID | \@IDs>

(Optional)

An integer or an arrayref of integers that was returned to you from the add_load() method.

=item Output

None

=item Example

    $cMgr_obj->pause();
    ...
    $cMgr_obj->resume();
    if ($cMgr_obj->is_error() && ($cMgr_obj->error_code() != $cMgr_obj->NA())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }
    ...

=back

=head2 C<available_loads()>

=over 4

=item Description

returns an array of avalilable load generators

=item Input

None

=item Output

Array of Avialible Load Types

=item Examples

    my @availible_loads = ClientMgr::available_loads();
    my $cMgr_obj = ClientMgr->new(clients => \@Clients);
    foreach(@availible_loads)
    {
        $cMgr_obj->add_load(type=>$_);
    }
    ...

=back

=head2 C<status(%params)>

=over 4

=item Description

Status of clients for the ID or IDs

=item Inputs

=item * id => <$Load_ID | \@IDs>

(Optional)
An integer or an arrayref of integers that was returned to you from the add_load() method.

=item Output

Array of HashRefs :

    $hash = {id      => integer <load id>,
             clients => integer <quanitity of Clients for the ID>,
             running => @running_Clients,
             stopped => @stopped_Clients}

=item Examples

    my @status_hashes = $cMgr_obj->status();
    if ($cMgr_obj->is_error())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }

    foreach my $status (@status_hashes)
    {
        if(int($status->{stopped}) > 0)
        {
            foreach my $client (@{$status->{stopped}})
            {
                $Trace->fatal(msg=>"Load on " . $client . " was NOT running");
            }
        }
    }

=back

=head2 C<results(%params)>

=over 4

=item Description

results from the clients that are NOT "PASS"

=item Inputs

=item * id => <$ID | \@IDs>

(Optional)

An integer or an arrayref of integers that was returned to you from the add_load() method.

=item * level => <0 | 1 | 2>

(Optional)

default = 2

0 returns an Array of Array of Result Objects
    This is ALL Result Objects from every Client (if it has any).
    
1 returns an Array of WORST Result Objects
    This is the Worst result from each Client

2 returns an Array of bad Result Objects
    This if the worst result is Non-<PASS | INFO | PARAM>
    you kinda want this to be empty all of the time

**An empty array can be because it's not at a point of reporting a result.
    this *may* occur if this is called while it's running (for some things).

=item Output

Hash :

    %hash = ($id1 => \@result_objects,
             $id2 => \@result_objects,
             ...)

=item Examples

    my %bad = $cMgr_obj->results();
    if (defined(%bad))
    {
        foreach my $id (keys(%bad))
        {
            $Trace->comment(msg=>"ID = " . $id .
                                "\nResult = " . $bad{$id}->type());
        }
        $Trace->fatal(msg => Dumper(\%bad));
    }
    
    ... OR ...
    
    my %worst = $cMgr_obj->results(level => 1);
    if (defined(%worst))
    {
        foreach my $id (keys(%worst))
        {
            $Trace->debug(msg=>"ID = " . $id .
                                "\nResult = " . $worst{$id}->type());
        }
    }
    
    ... OR ...
    
    my %all = $cMgr_obj->results(level => 0);
    if (defined(%all))
    {
        foreach my $id (keys(%all))
        {
            foreach my $result(@{$all{$id}})
            {
                $Trace->debug(msg => "ID = " . $id .
                             "\nResult = " . $result->type());
            }
        }
    }

=back

=head2 C<stop(%params)>

=over 4

=item Description

Stop running ID or IDs

=item Inputs

=item * id => <$Load_ID | \@IDs>

(Optional)

An integer or an arrayref of integers that was returned to you from the add_load() method.

=item * hard => <0 | 1>

(Optional)

default 0

Should it try to do a hard stop or a graceful stop?
Hard stop just Kills the subtest

=item * status_check => <0 | 1>

(Optional)

default 1

Should it Check the status before it stops to make sure everything is running?

=item * clients => <\@Clients | $client>

(Optional)

Default = (All Clients that the IDs know about)
This Specifies which Clients of the id to stop

=item Output

None

=item Example

    $cMgr_obj->stop();
    if ($cMgr_obj->is_error())
    {
        $Trace->fatal(msg => $cMgr_obj->error_msg());
    }

=back

=head2 C<HAMMER()>

=over 4

=item Description

Used as a const (but easier to use w/ an IDE).

Just returns Process::Load::HAMMER()

=item Input

None

=item Output

String "hammer"

=item Examples

    my $Client_Mgr = ClientMgr->new();
    $ClientMgr_obj->add_load(type => $ClientMgr_obj->HAMMER());
    
=back

=head2 C<CTHON()>

=over 4

=item Description

Used as a const (but easier to use w/ an IDE).

Just returns Process::Load::CTHON()

=item Input

None

=item Output

String "cthon"

=item Example

    my $Client_Mgr = ClientMgr->new();
    $ClientMgr_obj->add_load(type => $ClientMgr_obj->CTHON());
    
=back

=head2 C<RANDOM()>

=over 4

=item Description

Used as a const (but easier to use w/ an IDE).

Just returns Process::Load::RANDOM()

=item Input

None

=item Output

String "random"

=item Example

    my $Client_Mgr = ClientMgr->new();
    $ClientMgr_obj->add_load(type => $ClientMgr_obj->RANDOM());
    
=back

=head2 C<SIO()>

=over 4

=item Description

Used as a const (but easier to use w/ an IDE).

Just returns Process::Load::SIO()

=item Input

None

=item Output

String "sio"

=item Example

    my $Client_Mgr = ClientMgr->new();
    $ClientMgr_obj->add_load(type => $ClientMgr_obj->SIO());
    
=back

=cut
