# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/ClientRPC/Locks.pm#3 $
## @summary     Interface class for inheriting methods supporting NATE3 RPC
## @author      dl-striping-qa@netapp.com
## @status      private
## @pod here
###############################################
##   NOT READY FOR PUBLIC CONSUMPTION
###############################################
package ClientRPC::CIFSACL;

use strict;

use Tharn qw ($Log);

### Inheritance order of package INTERFACE methods
### available for client execution via rpc.
### "ClientRPC" should always be first.

use base qw(
    ClientRPC
    ClientRPC::Intf::CIFSACL
    ClientRPC::Intf::Mount
);

### Packages with IMPLEMENTATIONS needed by client
### rpc server.
my $RpcWinLibs = << 'CODE';
    use Tharn qw(param $Log);
    use Win32::Perms;
    use ClientRPC::Impl::CIFSACL;
    use ClientRPC::Impl::Mount;
    $Log->comment("LIB: ".param('NATE_LIB'));
CODE

my $RpcUnixLibs = << 'CODE';
    use Tharn qw(param $Log);
    use Filelock::Unix;
    use ClientRPC::Impl::CIFSACL;
    use ClientRPC::Impl::Mount;
    $Log->comment("LIB: ".param('NATE_LIB'));
CODE

###############################################################################
# Method:    new()
# Objective: Create a new ClientRPC object to support interface methods
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class,%args) = @_;
    my (
        $code,
        $self,
        );
   
    ### Create ClientRPC object which
    ### starts up RPC server
    $self=$class->SUPER::_new(%args);
    return $self if ($self->{error});
   
    ### Determine what implementation libraries
    ### need to be imported onto the remote rpc
    ### server so it can execute test call
    ### requests
    $code=($self->hosttype()=~/win/)
        ? $RpcWinLibs        
        : $RpcUnixLibs;

    ### Load required libraries 
    eval { 
        $self->load_libs(code=>$code);  
    };
    if ($@){
       ($self->on_err()=~/die/)
            ? croak $@
            : $self->_set_err($@);
    }

    return $self; 
}

###############################################################################
# Method:    destroy()
# Objective: delete ClientRPC object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self) = @_;
   
    return $self->SUPER::_destroy();
        
}

1;

############################################################################
#  P O D
############################################################################
=pod

=head1 Package Name

ClientRPC::CIFSACL

=head1 Author

dl-striping-qa@netapp.com

=head1 Description

ClientRPC::CIFSACL provides an interface for executing file ACLS calls via
NATE3 STAF RPC mechansim. It's interface functionality is inherited from selected
packages in the ClientRPC Interface directory containing stub methods. These
stubs provide the logic for sending stringified code block requests via
rpc where they are executed and dealing with results when their returned.

As part of instantiation this interface package both starts and primes the RPC
server with the library imports needed to support the stringifed code requests
required by the stub packages when one of their methods is called.  

=head2 Notes

Each interface package inherits from 1 or more corresponding interface stub
packages in the ClientRPC Interface directory.  The stub packages provide
method and argument mappings to the location of the real implementation
package, as well as error handling on the implementation return call.  The
stub packages utilize ClientRPC methods to stringify the methods requests to
be sent to the remote RPC server. Stub interfaces provide no input validation
up front.  All argument validation should happen in the implementation libraries.

The package order in the base() array denotes the search order for a particular
package stub methods in the ClientRPC Interface directory. 

 use base qw(
    ClientRPC::Interface::FileLock  <== Add your abstract stub interfaces here
    ClientRPC::Interface::MyPkg1
    ClientRPC::Interface::MyPkg2
    ...
 );
 
The location of the  implementation packages that support the stub packages
included in this interface need to be specified under the OS heredoc variables
at the top of this interface package.  As an example ...

 my $RpcWinLibs = << 'CODE';
    use Tharn qw($Log); 
    use Filelock::Win32;       <== Add your implementaton libraries here
    use Dir1/MyPkg1;
    use Dir2/Dir3/MyPkg2;
 CODE

 my $RpcUnixLibs = << 'CODE';
    use Tharn qw($Log);
    use Filelock::Unix;           If implementations are same for both OS
    use MyLib1/MyPkg1         <== flavors then add same library to both
    use Dir4/MyPkg2;
 CODE

=head1 See Also

This package inherits from  the following modules.  See it's documention for
more detail

B<ClientRPC::Intf::Mount>

B<ClientRPC::Intf::Filelock>

=head1 Public Methods

=head2 B<new( B<client>, I<on_err>, I<uid> )>

Creates a new ClientRPC::Interface object for communicating to client
machines via NATE3 STAF RPC.

=over 4

=item Inputs

=over 4

=item * B<client>    

Name of client.   Must have a valid tharnhost file entry under
/x/eng/testbeds/nate_hosts 

=item * B<on_err> (default: "die")    

Error handler to use for returning errors back to the calling script.

=begin html

<ul>
<li><b>die</b>: All errors encountered through object are thrown by die 
<li><b>rtn</b>: All errors encountered are stored in the object.  It is the
responsibility of the caller to analyze errors through additional method calls.
See module <b>ERROR</b> for detail.
</ul><p>

=end html

=item * B<uid> (default: "rpc"I<incrementing nbr>)    

Unique name that can be used to tag object.  Useful when multiple
ClientRPC::Interface objects are created for the same client.

=back

=item Outputs

Returns 1 on success, throws error on failure

=item Examples

 # Example 1: Create  clientRPC::Interface objects for each client passed in

 foreach my $client (@clients) {
    $Test->nlog("- Creating clientRPC  object for ".$client);
    my $clientObj = new ClientRPC::Interface(client=>$client,on_err=>'rtn');
    if ($clientObj->error()){
        $Test->nlog($clientObj->errmsg());
        return $TCD::FAIL;
    }
    push @ClientObjs,$clientObj;
 }

=back

=head2 B<destroy()>

Stops the rpc server on the remote client.  Once it has been stopped it cannot
be restated unless a new ClientRPC::Interface object is created.

=over 4

=item Inputs

=over 4

None

=back

=item Outputs

Returns 1 on success, throws error on failure

=item Examples

 # Example 1: Destroy the ClientRPC object
 foreach my $clientRpcObj (@ClientRpcObjs) {
    $Test->nlog(" - destroy client object".
                " ".$clientRpcObj->hostname().
                " (".$clientRpcObj->uid().")");
    $rc=$clientRpcObj->destroy();
 }

=back

=cut
