# Hey emacs, this is basically -*- perl -*-
# $Id$

## @summary     Class for pathdiff
## @author      dl-striping-qa@netapp.com
## @status      private
## @pod here
###############################################
##   NOT READY FOR PUBLIC CONSUMPTION
###############################################
package ClientRPC::Impl::Pathdiff;

use strict;
use Carp qw(croak);
use Cwd;
use Data::Dumper;
use Params::Validate qw(validate SCALAR BOOLEAN OBJECT ARRAYREF HASHREF);
use Tharn qw ($Log);
use Pathdiff;

my $pathdiffObj;

###############################################################################
# Method:    pathdiff_init()
# Objective: Init a Pathdiff object with specific arguments
# Details:   See POD documentation at the end of this file.
###############################################################################  
sub pathdiff_init {
    my (%args)=@_;

    if ($pathdiffObj) {
        $Log->comment("pathdiff object already exists, not creating");
        return;
    }
    $Log->comment("Initing pathdiff object with args: ".Dumper({%args}));
    $pathdiffObj = Pathdiff->new(pathdiff_config => {%args});
}

sub _foc_pathdiff {
    if (!$pathdiffObj) {
        $Log->comment("Creating pathdiff Obj");
        $pathdiffObj = Pathdiff->new();
        return $pathdiffObj;
    } 
    
    $Log->comment("Reusing existing pathdiff Obj");
    return $pathdiffObj;
}

###############################################################################
# Method:    snapshot()
# Objective: Do a snapshot 
# Details:   See POD documentation at the end of this file.
###############################################################################  
sub snapshot {
    my (%args)=@_;
    
    ### Validate inout arguments
    (%args) = validate (
        @_,
        {
            "chdir"         => { type => SCALAR} ,
            directory       => { type => SCALAR} ,                
        }
    );
    $pathdiffObj = _foc_pathdiff();
    my $dbFile;

    $Log->debug("chdir: $args{chdir}");
    my $saved_pwd = getcwd;
    if (!chdir($args{"chdir"})) {
        $Log->error("Error chdir $args{chdir}"); 
        return;
    } 

    # Print out the path, its helpful if we need to find out who made the file
    $Log->comment("Starting pathdiff snapshot");
    $dbFile = $pathdiffObj->snapshot(directory => $args{directory}); 
    $Log->comment("Done with pathdiff snapshot, returned $dbFile");

    if (!chdir($saved_pwd)) {
        $Log->error("Error chdir $saved_pwd");
        return;
    }
    return $dbFile;
}

###############################################################################
# Method:    compare()
# Objective: Compare to snapshot 
# Details:   See POD documentation at the end of this file.
###############################################################################  
sub compare {
    my (%args)=@_;
    
    ### Validate inout arguments
    (%args) = validate (
        @_,
        {
            "chdir"    => { type => SCALAR },
            source     => { type => SCALAR },
            target     => { type => SCALAR },
            skip       => { type => ARRAYREF, 
                            optional => 1, 
                            default => [qw(ATime Dev Inode)],
                          },
        }
    );
    $Log->debug("Pathdiff chdir: $args{chdir}");
    my $saved_pwd = getcwd;
    if (!chdir($args{"chdir"})) {
        $Log->error("Error chdir $args{chdir}");
        return 0;
    } 
    $pathdiffObj = _foc_pathdiff();
    my $dbFile;
    $Log->debug("Starting pathdiff compare");
    $pathdiffObj->compare(source => $args{source} , 
                          target => $args{target},
                          skip   => $args{skip},
                         );
    $Log->debug("Done with pathdiff compare");
    if (!chdir($saved_pwd)) {
        $Log->error("Error chdir $saved_pwd");
        return;
    }
    return 1;
}

###############################################################################
# Method:    delete_db()
# Objective: Clean db file 
# Details:   See POD documentation at the end of this file.
###############################################################################  
sub delete_db {
    my (%args)=@_;
    
    ### Validate inout arguments
    (%args) = validate (
        @_,
        {
            db_file     => { type => SCALAR },
        }
    );

    $Log->debug("Deleting pathdiff db");
    unlink($args{db_file});
    $Log->debug("Pathdiff db deleted");
    return 1;
}



1;

############################################################################
#  P O D
############################################################################
=pod

=head1 Package Name

ClientRPC::Impl::Pathdiff

=head1 Author

dl-striping-qa@netapp.com

=head1 Description

ClientRPC::Impl::Pathdiff provides an API interface driving Pathdiff 

=head1 See Also

This package is accessible through instantiation of ClientRPC::Interface 
enabling easy execution of Pathdiff driver on remote host via RPC  

B<ClientRPC::Interface>

=head1 Public Methods

=head2 B<pathdiff_init ( B<%options> )>

Initialize a lib/Pathdiff.pm object for usage, pass options to configure

=over 4

=item Inputs

=over 4

=item * B<%options>    

See lib/Pathdiff.pm

=back

=item Outputs

None

=back

=head2 B<snapshot( B<volume> )>

Snapshot a directory for later comparison

=over 4

=item Inputs

=over 4

=item * B<chdir>    

Base path to ignore in the difference

=item * B<directory>

Directory to do save for later difference, relative to chdir

=back

=item Outputs

Returns path to the pathdiff db file 

=item Examples

# Example 1: Store a tree for later comparison 
    
eval {
   $rc = snapshot  (chdir=>"/t/9.0.2.39/root-tcp-3/vol1/path",
                    directory=>"testdir");
);
if ($@) {
   $Test->nlog($@);
   return $TCD:FAIL;
}   


=back

=head2 B<compare ( B<chdir> , B<source> , B<target> )>

Compare a prior snapshot or two directories using pathdiff

=over 4

=item Inputs

=over 4

=item * B<chdir>    

Base path to ignore in the difference

=item * B<source>    

The source, same as Pathdiff.pm, for comparison can be a pathdiff db.

=item * B<target>

The destination, same as Pathdiff.pm, for comparison

=back

=item Outputs

Returns 1 on success, throws error on failure

=item Examples

# Example 1: Compare trees 
    
eval {
   $rc = compare (chdir=>"/t/9.0.2.39/root-tcp-3/vol1/path",
                  source=>"$dbFileFromSnapshot",
                  target=>"testdir");
);
if ($@) {
   $Test->nlog($@);
   return $TCD:FAIL;
}   


=back

=head2 B<delete_db( B<db_file> )>

Delete the db file from Pathdiff

=over 4

=item Inputs

=over 4

=item * B<db_file>    

Full path to file that was earlier returned from pathdiff snapshot.

=back

=item Outputs

Returns 1 on success, throws error on failure

=item Examples

# Example 1: Delete the db file
    
eval {
   $rc = delete_db (
            db_file=>"$dbFileFromSnapshot",
        );
);
if ($@) {
   $Test->nlog($@);
   return $TCD:FAIL;
}   

=back

=cut
