#
# $Id: //depot/prod/test/Rsierranevada/lib/ClusterAgent.pm#1 $
#

## @summary
##   Cluster Agent.  Distributes commands to nodes in different clusters with
##   complete separation of Server objects through use of Subtest.  
## 
## @description
##   
## @synopsis
##
## @example
##
## @status in-progress
##
## @author ewing@netapp.com

## @pod here

=head1 NAME


ClusterAgent.pm - A module implemented Server instance containers with subtest.

=head1 DESCRIPTION

C<ClusterAgent> Allows multiple Server instances in the same memory space.
This module facilitates testing of the cluster peer infrastructure as well
as features which use it (i.e., snapmirror).

The ClusterAgent reference can be used just as a Server refrence can be used.
In fact, it mostly just passes parameters and results across process 
boundaries.

=head1 SYNOPSIS

    use ClusterAgent;

    $clus_a = new ClusterAgent(node => $params->{CLUSTER_A_NODE});

    $clus_b = new ClusterAgent(node => $params->{CLUSTER_B_NODE});

    # use the ClusterAgents just as you would use a Server reference
    my $a_id = $clus_a->cluster_identity_show();

    my $b_id = $clus_b->cluster_identity_show();

=head1 METHODS

Since the ClusterAgent is largely based on Server (which is largely based on
the 10-mode UI), Server documentation is to be considered authoritative.

The ClusterAgent supports a few additional method related to its special
abilities.  These additional methods are available:

B<TODO>

=cut

#
# Initial ClusterAgent package definition
#
# This section contains some of the constants and the "use" statements for
# the ClusterAgent.
#
package ClusterAgent;

use strict;
use warnings;

use Server;
use Options;
use TracePkg;
use Tharn;
use Data::Dumper;
use Carp qw(confess);
use AutoLoader qw/AUTOLOAD/;

my $Trace = new TracePkg();
$Trace->set_options(separateNewline => 1);

sub AGENT_TERMINATE
{
    return "ClusterAgent_ctrl_AGENT_TERMINATE";
}

sub AGENT_SUCCESS
{
    return "ClusterAgent_ctrl_AGENT_SUCCESS";
}

sub AGENT_FAILURE
{
    return "ClusterAgent_ctrl_AGENT_FAILURE";
}

#
# ClusterOp package defintiion.
#
# This object is used to pass data between the ClusterAgent and its worker
# subtest.  This package does some very basic verification that what's being
# passed is useful and some validation routines.
#

package ClusterOp;

#
# ClusterOp constructor
#
sub new 
{
    $Trace->entry();
    my $class = shift();
    my %p = @_;

    my $self = ();

    $self->{cmd} = $p{cmd} || Carp::confess();
    $self->{args} = $p{args};
    $self->{result} = ();
    $self->{output} = ();
    $self->{table} = ();
    $self->{eval_error} = ();

    $Trace->debug(msg => "Created new ClusterOp for:\n\tcmd: ".
        (exists($self->{cmd}) ? $self->{cmd} : "(none)"). 
        "\n\targs: ".
        (exists($self->{args}) ? $self->{args} : "(none)"));
    bless($self, $class);
    return $self;
}

#
# Utility to test for sucessful ClusterOps.
#
sub success
{
    my $self = shift();
    return ($self->{result} eq ClusterAgent::AGENT_SUCCESS);
}

sub failure
{
    my $self = shift();
    return !$self->success();
}

#
# Destructor for ClusterOp
#
sub DESTROY
{
    my $self = shift();
    $Trace->debug(msg => "Destroying ClusterOp for:\n\tcmd: ".
        (defined($self->{cmd}) ? $self->{cmd} : "(none)"). 
        "\n\targs: ".
        (defined($self->{args}) ? $self->{args} : "(none)"). 
        "\n\tresult: ".
        (defined($self->{result}) ? $self->{result} : "(none)"));
}

#
# ClusterAgent package definition continued
#
# This is the part of the agent which actually does the work of handing
# requests to a contained Server reference.
#
package ClusterAgent;

#
# ClusterAgent constructor
#
# Takes the named node spawns a subtest, verifies that the subtest is okay and
# returns control to the caller.  The subtest is ClusterAger::agent_exec() 
# defined below.
#
sub new 
{
    $Trace->entry();
    my $class = shift();
    my %p = @_;

    my $self = ();

    $self->{node} = $p{node} || Carp::confess();

    $Trace->debug(msg => "Starting cluster agent for ".$p{node});

    # Create the subtest which will issue commands to a specific cluster.
    $self->{subtest} = Tharn::subtest(
        sub { ClusterAgent::agent_exec(); },
        -bg,
        -ipc,
        -runid, 'ClusterAgent_'.$p{node},
        "NODES=".$p{node}
    );

    #
    # Try to make sure that the subtest is alright before returning
    # control to the caller.  Check agent_exec() below for the write_message()
    # that is being read here.
    #
    if ($self->{subtest}->ipc()->read_message() ne 'OK') {
        Carp::confess("Failed to contact subtest, exiting\n");
    }
    
    $Trace->debug(msg => "Created new ClusterAgent for ".$p{node});

    bless($self, $class);
    $Trace->exit();
    return $self;
}

#
# AUTLOAD for ClusterAgent.
#
# This method allows the ClusterAgent to pass just about anything it recieves
# to its agent_exec() subtest.
#
sub AUTOLOAD 
{
    my $method = our $AUTOLOAD;
    my $self = shift();
    my %p = @_;

    #
    # strip off the package name
    #
    $method =~ s/.*:://;
    $Trace->entry(msg => "AUTOLOAD entered for $method.");

    #
    # Call run with our passed-in command and our passed-in arguments.
    #
    my $op = $self->run(
        cmd => $method,
        args => \%p
    );

    #
    # If an eval_error is set by the subtest, it is rethrown here(?).
    #
    if ($op->{eval_error}) {
        die($op->{eval_error});
    }

    $Trace->exit(msg => "AUTOLOAD exited for $method.");
    return $op->{table};
}

#
# Takes a command and runs it.  This is the piece that actually communicates
# with the ClusterAgent::agent_exec() subtest.
#
sub run
{
    my $self = shift();
    my %p = @_;

    if (!defined($p{cmd})) {
        Carp::confess("Need a command to execute");
    }

    $Trace->entry(msg => "run entered for $p{cmd}.");

    #
    # Create the ClusterOp to issue to the agent_exec subtest.
    #
    my $in_op = new ClusterOp(
        cmd => $p{cmd},
        args => $p{args}
    );

    # Write the ClusterOp to the subtest (just command, no results).
    $Trace->debug(msg => "Write object to subtest:\n ".Dumper($in_op));
    $self->{subtest}->ipc()->write_obj($in_op);

    # Read a ClusterOp from the subtest (results and possibly the command too).
    my $out_op = $self->{subtest}->ipc()->read_obj();
    $Trace->debug(msg => "Read object from subtest:\n ".Dumper($out_op));

    if (!$out_op->success()) {
        Carp::confess("Failed to execute $p{cmd}: ".$out_op->{eval_error}."\n");
    }
    
    $Trace->exit(msg => "run exited for $p{cmd}.");
    return $out_op;
}

#
# ClusterAgent subtest function.  This is what allows us to issue commands
# to truly different Server objects.  This sub performs a basic loop that
#
#   1. Reads a ClusterOp from the parent.
#
#   2. Executes the command on its Server reference or executes a special 
#   ClusterAgent operation.
#
#   3. Writes the result back to the parent.
#
sub agent_exec
{
    my $self = shift();
    my $serv = new Server();

    #
    # Since this function is a subtest, this is needed.
    use Data::Dumper qw/Dumper/;

    $Trace->entry(msg => "agent_exec() entered");

    Subtest::ipc()->write_message("OK");

    #
    # This is the main event loop for agent_exec().
    #
    while (1) {
        my $op;

        #
        # Read a ClusterOp object from the parent.  This should be passed in
        # only from ClusterAgent::run().
        #
        eval {
            $op = Subtest::ipc()->read_obj();
        }; 
        if ($@) {
            $Trace->comment(msg => "Failed reading from parent.");
            return;
        }

        #
        # Certain special operations (like terminating the agent itself) are
        # handled.
        #
        if ($op->{cmd} eq  ClusterAgent::AGENT_TERMINATE) {
            $op->{result} = ClusterAgent::AGENT_SUCCESS;
            Subtest::ipc()->write_obj($op);
            last;
        }

        #
        # The general case is to issue the call directly on the server
        # reference.  This call is made within an eval closure so it doesn't
        # kill everything.  We'll pass the eval_error and re-thrown it in the
        # parent process.
        #
        eval {
            my $command = $op->{cmd};
            my $args = $op->{args};

            $Trace->debug(msg=> "Executing command: ".$command."\n".
                "with params: ".Dumper($args));

            #
            # Issue the command.
            #
            $op->{table} = $serv->$command(
                %$args,
                -Output => $op->{output}
            );
        };

        #
        # check for errors in the operation and set the ClusterOp result value.
        #
        if ($@) {
            $op->{result} = ClusterAgent::AGENT_FAILURE;
            $op->{eval_error} = $@;
        } else {
            $op->{result} = ClusterAgent::AGENT_SUCCESS;
        }

        Subtest::ipc()->write_obj($op);
    }

    $Trace->exit(msg => "agent_exec() exiting");
}

sub DESTROY
{
    my $self = shift();

    $Trace->entry(msg => "ClusterAgent destructor entered.");

    if (defined($self->{subtest}) && $self->{subtest}->is_alive()) {
        $Trace->debug(msg => "subtest is alive, stopping");
        defined($self->{subtest}) && $self->{subtest}->kill(message => 'ok');
    }
    $Trace->exit(msg => "Destroyed Agent for $self->{node}\n");
}

1;
