#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/ClusterLogging.pm#1

#***********************************************************************
#
# NAME
#   ClusterLogging.pm - cross cluster log gathering utility that utilizes
#       the scripting and clusterwide log viewer functionality available
#       in BoilerMaker and beyond.
#
# SYNOPSIS
#   Cross cluster log gathering utility that utilizes the scripting and
#   cluster wide log viewer functionality available in BoilerMaker.
#
# DESCRIPTION
#   This Perl module is meant for use in log gathering and troubleshooting
#   across the cluster.
#
# AUTHOR
#   weiss@netapp.com
#
#***********************************************************************

## @summary     Clusterwide Logging and Recording library     

package ClusterLogging;

#use statements
use strict;
use Data::Dumper;
use NACL::C::Node;
use TracePkg;
use Options;
use Params::Common;
use Sys::Hostname;
use NACL::APISet::Exceptions::NoMatchingEntriesException qw(:try);

#Global variables
my $Trace = new TracePkg();

# Create Node object
my $Node = NACL::C::Node->find();

# Create APISets to replace Server.pm
my $Systemshell = NACL::APISet->new(
    hostobj => $Node->hostrec(),
    category => "Node",
    interface => "CLI",
    set => "Systemshell"
);
my $NGSH = NACL::APISet->new(
    hostobj => $Node->hostrec(),
    category => "Node",
    interface => "CLI",
    set => "CMode"
);

#Methods

###############################################################################
#Sub: start_scripting
#Input Params: Hash defining the "scriptname" and the "size-limit" of the file
#Return values: Returns a scalar of the "scriptname"
#Summary: Enables scripting on a node with the specified filename and size.  If
#   no filename or size is specified, a default will be provided and returned.
###############################################################################
sub start_scripting {
    my %params = (
            scriptname => "log".time(),
            'size-limit' => "1G",
            @_,
    );
    
    $NGSH->system_script_start(filename=>$params{scriptname}, "size-limit"=>$params{'size-limit'});
    
    $Trace->debug(msg=>"Scripting started for: $params{scriptname}\n");
    return $params{scriptname};
}

###############################################################################
#Sub: stop_scripting
#Input Params: <None>
#Return values: <None>
#Summary: Stops the currently active script on a node
###############################################################################
sub stop_scripting {
    
    $NGSH->system_script_stop();
    
    $Trace->debug(msg=>"Scripting stopped.\n");
}

###############################################################################
#Sub: download_script
#Input Params: A hash defining the "scriptname" and a "destination" to upload
#   the script to.
#Return values: <None>
#Summary: Attempts to upload the script specified in the input parameters.  If
#   the destination provided is an http of ftp destination, the script upload
#   feature is utilized.  Otherwise, the method attempts to use rcp to copy the
#   script file to a path on the local host.
###############################################################################
sub download_script { 
    my %params = (
            scriptname=>undef,
            destination=>undef,
            @_,
    );
    my $output;
    if (!($params{scriptname})) {
        $Trace->fatal(msg=>"No script name specified for download");
    }
    if ($params{destination}) {
        my $flag =0;
        my $timer=0;
        while ($flag != 1) { #wait for script to stop recording   
            my $Response = $NGSH->system_script_show(filename=>$params{scriptname});
            $output = $Response->get_parsed_output();
            
            if ($output->[0]->{state} eq "file-closed") {
                $flag=1;
            }
            else { 
                $Trace->debug(msg=>"Waiting for script to finish recording...\n");
                sleep(15);
                $timer++;
            }
        }
        #use node upload feature if ftp or http destination
        if ($params{destination} =~ /ftp:\/\// || $params{destination} =~ /http:\/\//) {
            $NGSH->system_script_upload(filename=>$params{scriptname},
                                        destination=>$params{destination});
        }
        #else use rcp to copy the scripted log to the destination
        else {
            my $host=hostname;
            $Trace->debug(msg=>"Executing cmd: rcp \/mroot\/etc\/cliScripts\/admin\/$params{scriptname} root\@$host:$params{destination} \n");
            $Systemshell->execute_raw_command(command => "rcp \/mroot\/etc\/cliScripts\/admin\/$params{scriptname} root\@$host:$params{destination}");
        }
    }
    else {
        $Trace->fatal(msg=>"No destination specified, script will not be uploaded.");
    }
}

###############################################################################
#Sub: delete_script_file
#Input Params: A hash defining the "scriptname"
#Return values: <None>
#Summary: Attempts to delete and remove the script specified in the parameters
###############################################################################
sub delete_script_file {
     my %params = (
            scriptname=>undef,
            username=>undef,
            @_,
    );
    eval {
        if ($params{username}) {
            $NGSH->system_script_delete(username=>$params{username}, filename=>$params{scriptname});
        }
        else {
            $NGSH->system_script_delete(filename=>$params{scriptname});
        }
    };
}

###############################################################################
#Sub: retrieve_logs
#Input Params: A hash defining the following.  If values aren't specified,
#   default values will be used:
#            '-Timeout' - how long to wait for the cmd to return, default 600
#            files - specific logs to include, default is mgwd
#            timestamp - filter logs based on a timestamp or time frame
#            destination - destination to upload the logs to
#            node - specific node to filter logs on
#            sequence - specific sequence numbers to filter logs for
#            lines - specific text to filter the logs entries for
#            return_as_text - return the logs as one big scalar, default is false
#Return values: One of the four described below:
#   retrieve_logs will take in the parameters listed and either:
#   1) If a destination of ftp/http was specified, it will launch a script to
#       record the logs specified.  Then it will attempt to upload the file
#       to the specified ftp/http address.
#   2) If a destination other than ftp or http is specifed, it will attempt
#       to download the scripted file of the debug->log->show command to the
#       directory/file location specified.
#   3) If no destination is specified, it will return the output of the debug
#       log show command to the caller.  The return value is a ptr to an array
#       of hashes.
#   4) If no destination is specifed, and the return_as_text parameter was set
#       to true or yes, then the call will return the output of the debug log
#       show command as a scalar.
#Summary: Starts scripting on a node, configures the node to watch specific logs,
#   filters those logs based upon the parameters indicated, then uploads those
#   clusterwide log views or returns the clusterwide log view as data.
###############################################################################
sub retrieve_logs {
    my %params = (
            '-Timeout' => 600,
            files => "mgwd",
            timestamp => undef,
            destination => undef,
            node => undef,
            sequence => undef,
            lines => undef,
            return_as_text=>"false", #if true, will return a scalar
            @_,
    );
    
    my $scriptname = "log_".time();
    my $logoutput;
    my $output;
    my %optionsstring;
    #build the parameters to pass into the clusterwide log view command
    if ($params{files} eq "all") {
        $params{files}="*";
    }
    if ($params{'-Timeout'}) {
        $NGSH->set_timeout("connectrec-timeout" => $params{'-Timeout'});
    }
    if ($params{node}) {
        $optionsstring{node}=$params{node};
    }
    if ($params{timestamp}) {
        $params{timestamp} =~ s/(\D\D\D)\s\s(\d\s)/$1 0$2/g; #accounts for single digit day of month
        $optionsstring{timestamp}=$params{timestamp};
    }
    if ($params{sequence}) {
        $optionsstring{seq}=$params{sequence};
    }
    if ($params{lines}) {
        $optionsstring{'log-line'}=$params{lines};
    }
    while ((my $key,my $value)=each(%optionsstring)) {
        $Trace->debug(msg=>"Options String: $key - $value \n");
    }
    #start scripting and perform a clusterwide log view
    $NGSH->debug_log_files_modify("incl-files"=>$params{files});
    $NGSH->execute_raw_command(command => 'rows 0');
    if ($params{destination}) {
        $NGSH->system_script_start(filename=>$scriptname, "size-limit"=>"1G");
    }
    try {
        my $Response = $NGSH->debug_log_show(%optionsstring);
        $logoutput = $Response->get_parsed_output();
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        return undef;
    };
    
    if ($params{destination}) { #if a destination was specified to upload
        my $flag =0;
        my $timer=0;
        #stop scripting
        $NGSH->system_script_stop(); 

        while ($flag != 1) {  #wait for scripting to complete
            my $Response = $NGSH->system_script_show(filename=>$scriptname);
            $output = $Response->get_parsed_output();

            if ($output->[0]->{state} eq "file-closed") {
                $flag=1;
            }
            else { 
                $Trace->debug(msg=>"Waiting for script to finish recording...\n");
                sleep(15);
                $timer++;
            }
        }
        #if ftp or http destination, then use node's upload ability
        if ($params{destination} =~ /ftp:\/\// || $params{destination} =~ /http:\/\//) {
            $NGSH->system_script_upload(filename=>$scriptname,
                                        destination=>$params{destination});
        }
        else { #else, lets just copy the saved script to the specified destination on the network
            my $host=hostname;
            $Trace->debug(msg=>"Executing cmd: rcp \/mroot\/etc\/cliScripts\/admin\/$scriptname root\@$host:$params{destination} \n");
            $Systemshell->execute_raw_command(command => "rcp \/mroot\/etc\/cliScripts\/admin\/$scriptname root\@$host:$params{destination}");
        }
    }
    else { #if no destination, then we are returning the data to the caller
        if ($params{return_as_text} eq "yes" or $params{return_as_text} eq "true") {
            my $textversion;
            foreach my $logentry (@$logoutput) {
                $textversion=$textversion.$logentry->{timestamp}." ".$logentry->{node}." ".$logentry->{seq}." ".$logentry->{'log-line'}."\n";
            }
            return $textversion; #return a text glob of data (scalar)
        }
        else {
            return $logoutput; #return the data as a ref to an array of hashes
        }
    }
    #cleanup the scripts
    $NGSH->system_script_delete(filename=>$scriptname);
    $NGSH->clear_timeout();
}

###############################################################################
#Sub: get_logs_on_failure
#Input Params: A scalar defining the upload directory/destination.
#Return values: <None>, but method does upload log to specified destination
#Summary: A convenience method that when invoked basically grabs the last
#   5minutes worth of data, clusterwide, from the most popular log files: mgwd,
#   vifmgr, vldb, and command-history
###############################################################################
sub get_logs_on_failure {
    my $logdir=shift @_; #takes a directory to record the file to
    my %optionsstring = (timestamp=>'>5m',);
    $NGSH->set_timeout("connectrec-timeout" => 600);
    my $scriptname="clusterlog_".time().".log";
    #grab the clusterwide log view based upon the common params
    $NGSH->debug_log_files_modify("incl-files"=>"mgwd,vldb,vifmgr,command-history");
    $NGSH->execute_raw_command(command => 'set -showseparator "  "');
    $NGSH->execute_raw_command(command => "rows 0");
    $NGSH->system_script_start(filename=>$scriptname, "size-limit"=>"1G");
    my $Response = $NGSH->debug_log_show(%optionsstring);
    my $logoutput = $Response->get_parsed_output();
    
    #stop recording
    $NGSH->system_script_stop();
    $NGSH->execute_raw_command(command=>'set -showseparator "X;X"');
    my $flag=0;
    my $timer=0;
    my $output;
    while ($flag != 1) {   #wait for scrip to complete
        $Response = $NGSH->system_script_show(filename=>$scriptname);
        $output = $Response->get_parsed_output();
        if ($output->[0]->{state} eq "file-closed") {
            $flag=1;
        }
        else { 
            $Trace->debug(msg=>"Waiting for script to finish recording...\n");
            sleep(15);
            $timer++;
        }
    }
    #upload the scripted file to the destination
    my $host=hostname;
    $Trace->debug(msg=>"Executing cmd: rcp \/mroot\/etc\/cliScripts\/admin\/$scriptname root\@$host:$logdir \n");
    $Systemshell->execute_raw_command(command=>"rcp \/mroot\/etc\/cliScripts\/admin\/$scriptname root\@$host:$logdir");
    #should we keep these logs on the node????
    $NGSH->system_script_delete(filename=>$scriptname);
    $NGSH->clear_timeout();
}

###############################################################################
#Sub: retrieve_entries_between_text
#Input Params: A hash defining the following.  If values aren't specified,
#   default values will be used:
#            '-Timeout' - how long to wait for the cmd to return, default 600
#            files - specific logs to include, default is mgwd
#            timestamp - filter logs based on a timestamp or time frame
#            node - specific node to filter logs on
#            sequence - specific sequence numbers to filter logs for
#            lines - specific text to filter the logs entries for
#            begin_text - starting log entry to return, start with log entry
#                   containing this text
#            end_text - ending log entry to return, end with log entry
#                   containing this text
#            return_as_text - return the logs as one big scalar, default is false
#Return values: A scalar or a reference to an array of hashes contianing
#   the requested log entries.
#Summary: A convenience method that allows a user to specify a variety of
#   clusterwide log viewer filters, and search for specific log entries that
#   occurred between a starting log entry and an ending log entry.
###############################################################################
sub retrieve_entries_between_text {
    my %params = (
            '-Timeout' => 600,
            files => "mgwd",
            timestamp => undef,
            node => undef,
            sequence => undef,
            lines => undef,
            begin_text => undef,
            end_text => undef,
            return_as_text=>"false", #if true, will return a scalar
            @_,
    );
    
    my $logoutput;
    my $output;
    my %optionsstring;
    #build the parameter string for the log viewer command
    if ($params{files} eq "all") {
        $params{files}="*";
    }
    if ($params{'-Timeout'}) {
        $NGSH->set_timeout("connectrec-timeout" => $params{'-Timeout'});
    }
    if ($params{node}) {
        $optionsstring{node}=$params{node};
    }
    if ($params{timestamp}) {
        $params{timestamp} =~ s/(\D\D\D)\s\s(\d\s)/$1 0$2/g; #accounts for single digit day of month
        $optionsstring{timestamp}=$params{timestamp};
    }
    if ($params{sequence}) {
        $optionsstring{seq}=$params{sequence};
    }
    if ($params{lines}) {
        $optionsstring{'log-line'}=$params{lines};
    }
    while ((my $key,my $value)=each(%optionsstring)) {
        $Trace->debug(msg=>"Options String: $key - $value \n");
    }
    #configure the log viewer and perform a log show
    $NGSH->debug_log_files_modify("incl-files"=>$params{files});
    $NGSH->execute_raw_command(command=>"rows 0");
    my $Response = $NGSH->debug_log_show(%optionsstring);
    $logoutput = $Response->get_parsed_output();
    $NGSH->clear_timeout();
    
    #We are returning the data to the caller as a string
    if ($params{return_as_text} eq "yes" or $params{return_as_text} eq "true") {
        my $textversion;
        my $flag=0;
        foreach my $logentry (@$logoutput) { #grab entries between the text values
            if ($flag==1) {
                $textversion=$textversion.$logentry->{timestamp}." ".$logentry->{node}." ".$logentry->{seq}." ".$logentry->{'log-line'}."\n";  
            }
            if ($logentry->{'log-line'} =~ /.*$params{begin_text}.*/) {
                $textversion=$textversion.$logentry->{timestamp}." ".$logentry->{node}." ".$logentry->{seq}." ".$logentry->{'log-line'}."\n";  
                $flag=1;
            }
            if ($logentry->{'log-line'} =~ /.*$params{end_text}.*/) {
                $flag=0;
            }
        }
        return $textversion; #return a text glob of data (scalar)
    }
    else { #else return a reference to an array of hashes
        my $logdata;
        my $flag=0;
        foreach my $logentry (@$logoutput) { #grab entries between the text values
            if ($flag==1) {
               push(@$logdata,$logentry);
            }
            if ($logentry->{'log-line'} =~ /.*$params{begin_text}.*/) {
                push(@$logdata,$logentry);
                $flag=1;
            }
            if ($logentry->{'log-line'} =~ /.*$params{end_text}.*/) {
                $flag=0;
            }
        }
        return $logdata; #return the data as a ref to an array of hashes
    }
}

###############################################################################
#Sub: retrieve_log_rows
#Input Params: A hash defining the following.  If values aren't specified,
#   default values will be used:
#            '-Timeout' - how long to wait for the cmd to return, default 600
#            files - specific logs to include, default is mgwd
#            timestamp - filter logs based on a timestamp or time frame
#            node - specific node to filter logs on
#            sequence - specific sequence numbers to filter logs for
#            lines - specific text to filter the logs entries for
#            rows - number of rows to display (most recent)
#            return_as_text - return the logs as one big scalar, default is false
#Return values: A scalar or a reference to an array of hashes contianing
#   the requested log entries.
#Summary: A convenience method that allows a user view the last x number of
#   log entries recorded.
###############################################################################
sub retrieve_log_rows {
    my %params = (
            '-Timeout' => 600,
            files => "mgwd",
            timestamp => undef,
            node => undef,
            sequence => undef,
            lines => undef,
            rows => 100,
            return_as_text=>"false", #if true, will return a scalar
            @_,
    );
    
    my $logoutput;
    my $output;
    my %optionsstring;
    #build the paramters hash for the log viewer
    if ($params{files} eq "all") {
        $params{files}="*";
    }
    if ($params{'-Timeout'}) {
        $NGSH->set_timeout("connectrec-timeout" => $params{'-Timeout'});
    }
    if ($params{node}) {
        $optionsstring{node}=$params{node};
    }
    if ($params{timestamp}) {
        $params{timestamp} =~ s/(\D\D\D)\s\s(\d\s)/$1 0$2/g; #accounts for single digit day of month
        $optionsstring{timestamp}=$params{timestamp};
    }
    if ($params{sequence}) {
        $optionsstring{seq}=$params{sequence};
    }
    if ($params{lines}) {
        $optionsstring{'log-line'}=$params{lines};
    }
    while ((my $key,my $value)=each(%optionsstring)) {
        $Trace->debug(msg=>"Options String: $key - $value \n");
    }
    #configure the log viewer and perform a log show
    $NGSH->debug_log_files_modify("incl-files"=>$params{files});
    $NGSH->execute_raw_command(command=>"rows 0");
    my $Response = $NGSH->debug_log_show(%optionsstring);
    $logoutput = $Response->get_parsed_output();
    $NGSH->clear_timeout();
    
    #We are returning the data to the caller as a string
    if ($params{return_as_text} eq "yes" or $params{return_as_text} eq "true") {
        my $textversion;
        my $numrows=@$logoutput;
        my $startrow=$numrows-$params{rows};
        if ($startrow < 0) {
            $startrow=0;
        }
        for ($startrow;$startrow<$numrows;$startrow++) {
            my $logentry=$logoutput->[$startrow];
            $textversion=$textversion.$logentry->{timestamp}." ".$logentry->{node}." ".$logentry->{seq}." ".$logentry->{'log-line'}."\n";  
        }
        return $textversion; #return a text glob of data (scalar)
    }
    else { #return a reference to an array of hashes
        my $logdata;
        my $numrows=@$logoutput;
        my $startrow=$numrows-$params{rows};
        if ($startrow < 0) {
            $startrow=0;
        }
        for ($startrow;$startrow<$numrows;$startrow++) {
            my $logentry=$logoutput->[$startrow];
            push(@$logdata,$logentry);
        }   
        return $logdata; #return the data as a ref to an array of hashes
    }
}
=pod

=head1 NAME

ClusterLogging.pm - cross cluster log gathering utility that utilizes the
scripting and cluster wide log viewer functionality available in BoilerMaker.


=head1 SYNOPSIS

    use ClusterLogging;
    ClusterLogging::retrieve_logs(
            '-Timeout'=>"value" ,
            files=>"<log files>" ,
            timestamp=>"value" ,
            node=>"text" ,
            sequence=>"text" ,
            include=>"text" ,
            exclude=>"text" ,
            lines=>"text" ,
            return_as_text=>"true|false",
            destination=>"text"
            );


=head1 DESCRIPTION

The methods included in this library were developed in order to assist
engineers in the collection of log files throughout the cluster, in an
automated fashion.  These methods were written as a wrapper for the
existing library calls that can be made to invoke the scripting and
cluster wide log viewing functionality that already exists in BoilerMaker.
  
    
=head1 METHODS

    start_scripting();
        Description:
            This method will take a script name and start logging all cli access
            to the file.
        Inputs:
            scriptname=>"text" - A hash entry for the name of the script to
                record to.  Default is log<timestamp>.
            size-limit=>"value" - A hash entry for the max size of the script,
                defaults to 1G.
        Outputs:
            $scriptname - returns the name of the recording script as a scalar
        Examples:
            see full example below
           
            
    stop_scripting();
        Description:
            This method will stop a scripting process once was initiated.
        Inputs:
            None.  The script stop command is user dependent and will stop the
            only script that is being recorded.  Only one script can be
            recording per sessoion at any given time.
        Outputs:
            None
        Examples:
            see full example below
    
    
    download_script(scriptname=>"text",destination=>"text");
        Description:
            This method will download a script from the filer to the specifed
            destination.
        Inputs:
            scriptname=>"text" - A hash entry for the name of the script to
                download/upload.
            destination=>"text" - A hash entry for the location where the script
                should be downloaded to.  Scripts are download to the machine
                where the current perl script is being run.  So the destination
                should be a location local to this machine.  Otherwise, if you
                wish to download to a remote machine, you can specify the full
                details via an ftp or http url.
        Outputs:
            None
        Examples:
            see full example below
         
            
    delete_script_file(scriptname=>"text");
        Description:
            Removes the script file from the filer.
        Inputs:
            scriptname=>"text" - A hash entry for the the name of the script to
                remove
        Outputs:
            None
        Examples:
            see full example below
        
            
    retrieve_logs();
        Description:
            Retrieves the specified logs from the filers based upon the
            specified filters.  Then eiter returns the data or uploads the files
            to the specified destination.  Finally performs cleanup and deletes
            any used script from the filer.   
        Inputs:
            -Timeout=>"value" - Hash entry containing the timeout value for how
                long the call should wait.  Default is 10 minutes.
            files=>"text" - Hash entry defining which log files should be
                included in the output.  Default is the mgwd.log.
            destination=>"text" - Hash entry defining the location where the
                file should be uploaded or save to.  Save to path is local to
                the machine where the perl script is running.  If a remote
                machine is required, the full details can be specified using
                ftp or http addresses.  By default, this value is empty.
            return_as_text=>"true|false" - Hash entry detailing whether the
                output data should be returned as one big scalar.  Default is
                false.  Meaning that if a destination is not specified, the
                default return value is a reference to an array of hashes.
            FILTERS:
            timestamp=>"value" - Hash entry containing the desired time or time
                frame to filter the logs on.  Timestamp is typical format:
                Tue Jul 15 13:00:00 2008.
            node=>"text" - Hash entry detailing the node or nodes to filter the
                logs on.  Note that using this entry will return all log data
                then filter by this parameter.
            sequence=>"text" - Hash entry to filter based on sequence id within
                the logs.
            include=>"text" - Hash entry defining which nodes to specifically
                grab logs from.  All other nodes will be ignored.
            exclude=>"text" - Hash entry defining which nodes to exclude from
                the log view.  All other nodes will be included.
            lines=>"text" - Hash entry detailing what text or values to filter
                the log entries for.
        Outputs:
            Returns one of the following:
            1) If destination parameter is set to FTP/HTTP - filtered logs will
                be uploaded to the specified location
            2) If destination parameter is set, but contains local path -
                filtered logs will be copied to the specified location
            3) If return_as_text=true - filtered logs will be returned as one
                big scalar.
            4) If return_as_text=false, or no return_as_text parameter is
                specified - filtered logs will be returned as a reference to
                an array of hashes.
            *Note, currently if both the destination and return_as_text
                parameters are set, the destination parameter takes priority
                and the return_as_text parameter is ignored.
        Examples:
            see full example below
        **NOTE:
            If returning the data as a string or a reference to an array of
            hashes, be sure to set the NATE parameter CONNECTION_MAXBUFSIZE.
            By default this value is set to 1MB which is rather small if a
            large amount of log data is being returned. 
    
    
    retrieve_entries_between_text();
        Description:
            Retrieves all log entries that occur between the indicated start
            and stop text lines.
        Inputs:
            Takes the same has as indicated for retrieve_logs() with the
            addition of two new options.
                begin_text - log entry to start with.  Will start filtering when
                    the text here matches a log entry.
                end_text - log entry to stop with.  Will stop displaying log
                    entries when a log entry matches this text.
        Outputs:
            Returns either a scalar or a reference to an array of hashes,
                depending on the return_as_text hash option.
        Examples:
            see full example below
        **NOTE:
            If returning the data as a string or a reference to an array of
            hashes, be sure to set the NATE parameter CONNECTION_MAXBUFSIZE.
            By default this value is set to 1MB which is rather small if a
            large amount of log data is being returned. 
            
            
    retrieve_log_rows();
        Description:
            Retrieves the last X number of log entries, clusterwide wide,
            according to the filters specified.  Note that this returns the last
            X number of entries overall, not the last X entries from each log.
        Inputs:
            Takes the same has as indicated for retrieve_logs() with the
            addition of one new options.
                rows - the number of log entries to return total
        Outputs:
            Returns either a scalar or a reference to an array of hashes,
                depending on the return_as_text hash option.
        Examples:
            see full example below
        **NOTE:
            If returning the data as a string or a reference to an array of
            hashes, be sure to set the NATE parameter CONNECTION_MAXBUFSIZE.
            By default this value is set to 1MB which is rather small if a
            large amount of log data is being returned. 


    get_logs_on_failure($directory);
        Description:
            When called, automatically retrieves the last 5 mins worth of logs
            (vldb, mgwd, vifmgr, and command-history) and saves them in a file
            via the directory supplied.
        Inputs:
            <scalar> - indicating the directory to save the log to
        Outputs:
            <None>, but does record file to directory specified
        Examples:
            see full example below
            
            
=head1 EXAMPLES

Ex1: Retrieving last 5 minutes of the command-history log, clusterwide,
    and return the data as a scalar:

    my $cmd_data=ClusterLogging::retrieve_logs(files=>"command-history",
                                               timestamp=>">5m",
                                               return_as_text=>"true");
                                               
Ex2: Retrieving all log entries in mgwd within the past 1hr, saving to
    /u/weiss/LogViewer/mgwdlog.log on the local machine:

    ClusterLogging::retrieve_logs(files=>"mgwd",
                                timestamp=>">5m",
                                destination=>"/u/weiss/LogViewer/mgwdlog.log");

Ex3: Retrieving all log entries in mgwd related to ASUP for the
    specified nodes, and saving that data to
    /u/weiss/LogViewer/asup.txt:

    ClusterLogging::retrieve_logs(files=>"mgwd,command-history",
                                  include=>"rtpgx-3050-3,rtpgx-3050-4",
                                  lines=>'*asup*',
                                  destination=>"/u/weiss/LogViewer/asup.txt");
                                  
Ex4: Retrieving all command-history log entries between Tue July 15th,
    1-4:30pm, and saving that data to an ftp site:

    ClusterLogging::retrieve_logs(files=>"command-history",
                    timestamp=>'"Tue Jul 15 13:00:00 2008"..
                                    Tue Jul 15 16:30:00 2008"',
                    destination=>"ftp://<user>:<password>@<ip>/*.log");

Ex5: Using the start and stop scripting functionality and uploading:

    ClusterLogging::start_scripting(scriptname=>"abc.log",'-size-limit'=>"1G");
    <perform some cli commands>
    ClusterLogging::stop_scripting(scriptname=>"abc.log");
    ClusterLogging::download_script(scriptname=>"abc.log",destination="/a/b");
    ClusterLogging::delete_script_file(scriptname=>"abc.log");

Ex6: Use the -timestamp option to filter on log entries that occurred
    between the start of some commands and now.
        1)  In your script, record the current time:
                    my $starttime=localtime();
        2)  Perform the desired actions that may or may not generate
            log output
        3)  Now grab the desired logs entries that occurred between the
            start of that action and the completion of that action.
                    ClusterLogging::retrieve_logs(files=>"command-history",
                    timestamp=>"\"$starttime\"..\"localtime()\"",
                    destination=>"/u/weiss/LogViewer/myactions.log");
                    
Ex7: Retrieving all entries between a specific log entry and an ending
    log entry, based upon the text values of the log lines recorded.
    
    my $cmd_data=ClusterLogging::retrieve_entries_between_text(
                                                files=>"command-history,mgwd",
                                                begin_text=>"start",
                                                end_text=>"stop");
    
Ex8: Retrieve the last X number of rows from the logs specified (will
    retrieve the last X number of rows total, not X rows from each log)
        
    my $cmd_data=ClusterLogging::retrieve_log_rows(files=>"command-history,mgwd",
                                                    rows=>10);
                                                    
Ex9: Convenience method - get_logs_on_failure
    
    1)  Assume I run a test case and I'm checking for a failure.
    2)  I notice the failure occurred...I want to grab the logs
    3)  How convenient:
    
    ClusterLogging::get_logs_on_failure("/u/weiss/badthings");
    
=cut