# $Id: //depot/prod/test/main/lib/ClusterReboot.pm#1 $ 

## @summary Helper package to reboot entire clusters and handle
##          rebooting various nodes.
##
## @owner Cluster Infrastructure QA
##
## @author Brent Clausner (clausner@netapp.com)
## @author Cluster Infrastructure QA (dl-cluster-qa@netapp.com)
##
#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
##
## @pod here

=head1 NAME


ClusterReboot.pm - A module implemented to help assist in the feature of
cluster reboot.

=head1 DESCRIPTION

C<ClusterReboot> allows the caller to reboot any variety of nodes in a few
different manners. Reboot command, halt command, and simulated power outage
via RLM or vsim stop.

Usage of this is expected that if takeover occurs, forced giveback will be
done. In cases of halt and reboot, takeover is disabled.

=head1 SYNOPSIS

    use ClusterReboot;

    ClusterReboot->reboot(@nodelist);

=head1 METHODS

This helper library uses NACL for connections to the filer(s). It is expected
that all filers that will be used by this library are C-Mode.

reboot();
        Description:
            Reboots all the nodes passed to the method.
        Inputs:
            Array of C<NACL::C::Node>.
        Outputs:
            <none>
        Examples:
            <see synopsis above>

panic();
        Description:
            Reboots all the nodes passed to the method with the '-dump true'.
        Inputs:
            Array of C<NACL::C::Node>.
        Outputs:
            <none>
        Examples:
            ClusterReboot->panic(@nodelist);

rlm();
        Description:
            Simulate a power outage on the nodes for a specified duration.
        Inputs:
            Scalar integer and Array of C<NACL::C::Node>
        Outputs:
            <none>
        Examples:
            ClusterReboot->rlm(30, @nodelist);

halt();
        Description:
            Halt the specified nodes for a duration.
        Inputs:
            Scalar integer and Array of C<NACL::C::Node>
        Outputs:
            <none>
        Examples:
            ClusterReboot->halt(30, @nodelist);


=cut

package ClusterReboot;
use strict;
use Data::Dumper;
use NACL::C::Node;
use NACL::Transit;
use NATE::Process;
use NACL::APISet::Exceptions::ResponseException;
use NACL::APISet::Exceptions::ResponseException qw(:try);

use Params::Validate qw(validate SCALAR ARRAYREF BOOLEAN);

use base qw(Class);

sub reboot(@) {
    my $obj = shift;
    my @nodes = @_;
    my @procs = ();
    foreach my $node ( @nodes) {
        my %params = ();
        $params{'NODE'} = $node->name();
        $params{'FUNC'} = 'reboot';
        push(
                @procs,
                 NATE::Process->new(
                    codespec       => 'lib/ClusterReboot/reboot.thpl',
                    inherit_params => 1,
                    params         => \%params,
                    runid          => 'reboot-'.$node->name(),
                )->start()
            );
    }

    _wait_and_destroy(@procs);
    foreach my $n (@nodes) {
        $n->refresh_command_interface();
    }
    _wait_for_masters($nodes[0]);
}

sub panic(@) {
    my $obj = shift;
    my @nodes = @_;
    my @procs = ();
    foreach my $node ( @nodes) {
        my %params = ();
        $params{'NODE'} = $node->name();
        $params{'FUNC'} = 'panic';
        push(
                @procs,
                 NATE::Process->new(
                    codespec       => 'lib/ClusterReboot/reboot.thpl',
                    inherit_params => 1,
                    params         => \%params,
                    runid          => 'panic-'.$node->name(),
                )->start()
            );
    }

    _wait_and_destroy(@procs);

    #Deal with TO-GB
    sleep(15);
    print('Blindly trying to force giveback(if needed)\n');
    foreach my $n (NACL::C::Node->find()) {
        my @othernodes = ();
        my $name = $n->name();
        foreach my $other ( @nodes ) {
            if($other->name() ne $name) {
                push(@othernodes, $other);
            }
        }
        foreach my $othernode (@othernodes ) {
            eval {
                NACL::C::StorageFailover->giveback( command_interface => $n,
                                                    ofnode            => $othernode->name(),
                                                    'override-vetoes' => 'true' );
            };
        }
    }

    foreach my $n (@nodes) {
        $n->refresh_command_interface();
    }
    _wait_for_masters($nodes[0]);
}

sub halt {
    my $obj = shift;
    my ($timer,@nodes) = @_;
    my @procs = ();
    foreach my $node ( @nodes) {
        my %params = ();
        $params{'NODE'} = $node->name();
        $params{'FUNC'} = 'halt';
        $params{'TIME'} = $timer;
        push(
                @procs,
                 NATE::Process->new(
                    codespec       => 'lib/ClusterReboot/reboot.thpl',
                    inherit_params => 1,
                    params         => \%params,
                    runid          => 'halt-'.$node->name(),
                )->start()
            );
    }

    _wait_and_destroy(@procs);
    foreach my $n (@nodes) {
        $n->refresh_command_interface();
    }
    _wait_for_masters($nodes[0]);
}

sub rlm {
    my $obj = shift;
    my ($time, @nodes) = @_;
    my @procs = ();
    foreach my $node ( @nodes) {
        my %params = ();
        $params{'NODE'} = $node->name();
        $params{'FUNC'} = 'rlm';
        $params{'TIME'} = $time;
        push(
                @procs,
                 NATE::Process->new(
                    codespec       => 'lib/ClusterReboot/reboot.thpl',
                    inherit_params => 1,
                    params         => \%params,
                    runid          => 'rlm-'.$node->name(),
                )->start()
            );
    }

    _wait_and_destroy(@procs);
    foreach my $n (@nodes) {
        $n->refresh_command_interface();
    }

    #Deal with TO-GB
    sleep(15);
    print('Blindly trying to force giveback(if needed)\n');
    foreach my $n (NACL::C::Node->find()) {
        my @othernodes = ();
        my $name = $n->name();
        foreach my $other ( @nodes ) {
            if($other->name() ne $name) {
                push(@othernodes, $other);
            }
        }
        foreach my $othernode (@othernodes ) {
            eval {
                NACL::C::StorageFailover->giveback( command_interface => $n,
                                                    ofnode            => $othernode->name(),
                                                    'override-vetoes' => 'true' );
            };
        }
    }

    foreach my $n (@nodes) {
        my $transit_obj = NACL::Transit->new(name => $n->name());
        my $change = $transit_obj->change_state(to=>'USERNAME');
    }

    _wait_for_masters($nodes[0]);
    
}

sub _wait_and_destroy {
    my @procs = @_;
    my $staf = NATE::STAF::staf_global();

    foreach my $p (@procs) {
        $p->wait();
        $p->destroy();
    }
}

sub _wait_for_masters {
    my ($Node) = @_;
    my $loop_break = 0;
    my $loop_max   = 10;
    my $iter       = 0;

    while ( $loop_break == 0 && $iter < $loop_max ) {
        my $all_master = 1;
        try {
            foreach my $ring (
                NACL::C::ClusterRing->find( command_interface => $Node ) ) {
                if (   $ring->state()->master() eq '-'
                    || $ring->state()->master() eq '' ) {
                    $all_master = 0;
                }
            }
        } catch NATE::BaseException with {
            $all_master = 0;
        };

        if ( $all_master == 1 ) {
            $loop_break = 1;
        } else {
            sleep(5);
        }
        $iter++;
    }
}

1;
