#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/CmdsDriver.pm#5 $

# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide a set of attributes and methods to support commands testing
## @author pressley@netapp.com
## @status private
## @pod here

package CmdsDriver;

#
# Compiler directives.
#
use strict;

#
# Module import.
#
use Utils qw(make_host make_conn);
use Time::HiRes qw(gettimeofday);
use Exporter;
use NATE::Inc qw(get_file find_file);
use XML::Simple;
use Data::Dumper;
use Zapi;
use NaElement;
use Filer::Version qw(get_ontap_version);
use Tharn qw(host);
use San::ConnectionMgr;

use vars qw(
          @EXPORT
          @ISA
          %mapped_cmds
          );

#
# Inherit base class that encompasses Error, SanLog, and SanUtil as well
#
use base qw(San::CommonInterface);

#
# Constant variables
#
use constant ZAPI_USERNAME_7MODE => "root";
use constant ZAPI_USERNAME_CMODE => "admin";
use constant ZAPI_PASSWORD_CMODE => "netapp1!";
use constant CLI_MAXBUFSIZE      => "40000000";

# Environment variables
&Tharn::param("CMDSDRIVER_CMD_TIMING",  "-default", 0);
&Tharn::param("CMDSDRIVER_CMD_TIMEOUT", "-default", 600);

########################################################################
#                            Public Methods
########################################################################

########################################################################
# Method:    new
# Objective: Create new San object.
# Details:   See POD documentation at the end of this file.
########################################################################
sub new
{

    my($proto, %args) = @_;

    my $class = ref($proto) || $proto;

    #
    # Create a hash to store attributes and bless the hash to create an object.
    #
    my $self = {};
    bless $self, $class;

    #
    # Store the list of supported commands data/xml files
    # Also store the base command names (ie. "cli_volume")
    #
    @{$self->{data_files}} = ();
    @{$self->{cmds_loaded}} = ();

    #
    # The hash of commands created from the data files
    #
    $self->{commands} = undef;

    #
    # Set the debug_level.  The higher the level, the more information the user
    # should expect.
    #
    $self->{debug_level}   = 0;

    #
    # Set a default timeout
    #
    $self->{timeout} = &Tharn::param("CMDSDRIVER_CMD_TIMEOUT");

    #
    # Set a default maxbufsize for Tharn::say commands
    #
    $self->{maxbufsize} = CLI_MAXBUFSIZE;

    #
    # Enable or disable command timing
    #
    $self->{cmd_timing} = &Tharn::param("CMDSDRIVER_CMD_TIMING");

    #
    # Error message is not initially set, this will set the
    # following inherited attributes:
    #   $self->{error}  = 0;
    #   $self->{errmsg} = '';
    #
    $self->_clear_err();

    #
    # Attributes that affect how a command is executed.  All
    # of these has an override with the command parameter
    # and can be set via an accessor method.
    #
    $self->{filer}         = "";
    $self->{filer_version} = 0;
    $self->{filer_mode}    = "7mode";  # 7mode or Cmode supported
    $self->{cmd_interface} = "zapi";   # cli or zapi
    $self->{privilege}     = "";
    $self->{vfiler}        = "";
    $self->{vfiler_ip}     = undef;
    $self->{partner_mode}  = 0;        # 0 = no, 1 = yes
    $self->{show_cmd}      = 1;        # 0 = no, 1 = yes
    $self->{execution}     = 1;        # 0 = no, 1 = yes
    $self->{output_format} = "raw";    # raw | hash
    $self->{log_output}    = 0;        # always log command output
    $self->{connection}    = undef;
    $self->{log}           = undef;
    $self->{command_time}  = 0;
    $self->{depth}         = 0;        # used by _expand_complex_option
    $self->{vserver}       = "";       # default vserver for Cmode
    $self->{buildinfo}     = undef;

    #
    # The execute cmd will return a hash of results.  We will
    # store that so we can get to the last result.
    #
    $self->{resultsHash} = ();
    $self->{resultsHash}{result}       = 1;
    $self->{resultsHash}{reason}       = "";
    $self->{resultsHash}{output}       = "";
    $self->{resultsHash}{output_check} = 0;
    $self->{resultsHash}{result_check} = 0;

    #
    # We define the command files we want to load, the user can
    # override this so what we have here is a default list.
    #
    my $files_to_load = "volume, lun, igroup, fcp, iscsi";

    #
    # Now that all the defaults have been set, check if the
    # use sent any overrides
    #
    foreach (keys %args)
    {
        if (/-?debug_level$/) {
            length(my $debug_level = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::new() invalid value for " .
                                "\"debug_level\": " . $args{$_});
                return $self;
            };
            $self->set_debug_level($debug_level);
        }
        elsif (/-?log$/) {
            length($self->{log} = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::new() invalid value for " .
                                "\"log_handle\": " . $args{$_});
                return $self;
            };
        }
        elsif (/-?filer$/) {
            length($self->{filer} = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::new() invalid value for " .
                                "\"filer\": " . $args{$_});
                return $self;
            };
        }
        elsif (/-?cmds$/) {
            length($files_to_load = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::new() invalid value for \"cmds\":",
                                "$args{$_}");
                return $self;
            };
        }
        elsif (/-?cmd_interface$/) {
            length($self->{cmd_interface} = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::new() invalid value for " .
                                "\"cmd_interface\":" . $args{$_});
                return $self;
            };
        }
        elsif (/-?connection$/) {
            ref($self->connection($args{$_})) =~ /ConnectRec/i or
            do {
                $self->_set_err("CmdsDriver::new() invalid value for " .
                                "\"connection\":" . $args{$_});
                return $self;
            };
        }
        elsif (/^-?buildinfo$/) {
            #
            # An empty output is valid, don't check length
            #
            $self->{buildinfo} = $args{$_};
        }
        else {
            $self->_set_err("CmdsDriver::new() Invalid parameter specified: " .
                            "$_");
            return $self;
        }
    } # end foreach arg

    #
    # Check for required parameters
    #
    if (defined($self->{connection}))
    {
        $self->{filer} = $self->{connection}->hostp()->hostname();

        # we set node because if filer is cserver, we need it to
        # retrieve certain info that requires node param
        if ($self->filer_mode() =~ /Cmode/i) {
            $self->{node} = $self->{connection}->hostp()->{node};
        }
    }
    else
    {
        unless ($self->interface() eq "zapi")
        {
            if (length($self->{filer}))
            {
                #
                # make connection and set filer mode
                #
                my $conn_obj = $self->get_conn_obj();
                if ($conn_obj->error())
                {
                    $self->_set_err("Failed to create ConnectionMgr.");
                    return $self;
                }
                #
                # make connection and set connection for CmdsDriver
                #
                if ($conn_obj->make_connect(-filer => $self->{filer}))
                {
                    $self->connection(
                        $conn_obj->get_connection(-filer => $self->{filer}));
                    #
                    # set connection object
                    #
                    $self->{conn_obj} = $conn_obj;

                    #
                    # based on connect object, we know the mode ONTAP
                    # is running now, we can set filer_mode
                    #
                    $self->filer_mode($conn_obj->{mode});
                    unless ($conn_obj->{mode} eq "7mode")
                    {
                        #
                        # In Cmode, there are three environments (bsd, ngsh,
                        # and nodescope).
                        # We must set environment to ngsh in order to run Cmode
                        # commands
                        #
                        $self->set_shell_environment(-type => "ngsh");
                        if ($self->error())
                        {
                            $self->_set_err("failed to set ngsh environment",
                                            $self->errmsg());
                            return $self;
                        }
                    }

                    # we set node because if filer is cserver, we need it to
                    # retrieve certain info that requires node param
                    if ($self->filer_mode() =~ /Cmode/i) {
                        $self->{node} = $self->{connection}->hostp()->{node};
                    }
                } else {
                    $self->_set_err("Failed to make a connection to " .
                                    $self->{filer}, $conn_obj->errmsg());
                    return $self;
                }
            }
            else
            {
                #
                # we got here because the interface is CLI
                # and -filer and -connection is not passed in
                #
                $self->_set_err("Required parameter \"filer\"" .
                                " or \"connection\" for CLI interface");
                return $self;
            }
        }
    }

    #
    # Populate the commands hash
    #
    $self->_populate_commands($files_to_load);

    return $self;
}  # end of new method

#
# dave implement -async
#
# dave make sure we handle the case where options, required options,
# optional options is __NONE__
#
# command switches that are not attributes: async, expected_output, pass_thru
# new attributes - execute, command_time
#
#
###############################################################################
# Method:    execute_cmd
# Objective: Execute a command on a filer
# Details:   See POD documentation at the end of this file.
###############################################################################
sub execute_cmd
{
    my($self, %args) = @_;

    my $filer_version   = $self->filer_version();
    my $filer_mode      = $self->filer_mode();
    my $interface       = $self->interface();
    my $privilege       = $self->privilege();
    my $vfiler          = $self->vfiler();
    my $vfiler_ip       = $self->vfiler_ip();
    my $vserver         = $self->vserver();
    my $partner         = $self->partner_mode();
    my $output_format   = $self->output_format();
    my $connection      = $self->connection();
    my $log_output      = $self->log_output();
    my $timeout         = $self->timeout();
    my $maxbufsize      = $self->maxbufsize();
    my $cmd_timing      = $self->cmd_timing();
    my $check_output    = 0;
    my $check_result    = 0;
    my $expected_output = undef;
    my $expected_result = undef;
    my $filer           = $self->filer();
    my $cmd             = undef;
    my $pass_thru       = "";
    my $require_options = 1;
    my %cmd_args        = ();
    my $cmd_args_string = "";
    my $use_cmode_xml   = 0;
    my $donot_append_vserver  = 0;

    #
    # Clear errors.
    #
    $self->_clear_err;

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/^-?cmd$/) {
            length($cmd = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::execute_cmd() " .
                                "invalid value for \"cmd\":" .
                                $args{$_});
                return $self;
            };
        }
        elsif (/^-?cmd_filer_version$/) {
            length($filer_version = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::execute_cmd() "                .
                                "invalid value for \"cmd_filer_version\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/^-?cmd_interface$/) {
            $interface = $args{$_};
            if (not length($interface) or
                $interface !~ /zapi|cli|zapi_cli/i)
            {
                $self->_set_err("CmdsDriver::execute_cmd() "            .
                                "invalid value for \"cmd_interface\": " .
                                $args{$_});
                return $self;
            };

            #
            # In SanFiler, there's a zapi_cli interface that is used for
            # the system-cli zapi.  If one of those sneaks thru, we just
            # want to use the zapi interface
            #
            $interface =~ s/zapi_cli/zapi/;

        }
        elsif (/^-?cmd_privilege$/) {
            #
            # An empty privilege value is valid, don't check length
            #
            $privilege = $args{$_};
        }
        elsif (/^-?cmd_vserver$/) {
            #
            # An empty vserver value is valid, don't check length
            #
            $vserver = $args{$_};

            #
            # set vserver
            #
            $self->vserver($vserver);
        }
        elsif (/^-?cmd_vfiler$/) {
            #
            # An empty vfiler value is valid, don't check length
            #
            $vfiler = $args{$_};
        }
        elsif (/^-?cmd_vfiler_ip$/) {
            #
            # An empty vfiler_ip value is valid, don't check length
            #
            $vfiler_ip = $args{$_};
        }
        elsif (/^-?cmd_partner$/) {
            $partner = $args{$_};
            if (not length($partner) or
                $partner !~ /yes|no|1|0/i)
            {
                $self->_set_err("CmdsDriver::execute_cmd() "          .
                                "invalid value for \"cmd_partner\": " .
                                $args{$_});
                return $self;
            }

            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $partner = 1 if ($partner =~ /yes/i);
            $partner = 0 if ($partner =~ /no/i);
        }
        elsif (/^-?cmd_output_format$/) {
            length($output_format = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::execute_cmd() "            .
                                "invalid value for \"cmd_output_format\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/^-?cmd_log_output$/) {
            length($log_output = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::execute_cmd() "            .
                                "invalid value for \"cmd_log_output\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/^-?cmd_connection$/) {
            length($connection = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::execute_cmd() "            .
                                "invalid value for \"cmd_connection\": " .
                                $args{$_});
                return $self;
            };
            $self->connection($connection);
        }
        elsif (/^-?cmd_timeout$/) {
            length($timeout = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::execute_cmd() "        .
                                "invalid value for \"cmd_timeout\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/^-?cmd_maxbufsize$/) {
            length($maxbufsize = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::execute_cmd() "        .
                                "invalid value for \"cmd_maxbufsize\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/^-?cmd_timing$/) {
            length($cmd_timing = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::execute_cmd() "        .
                                "invalid value for \"cmd_timing\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/^-?cmd_filer$/) {
            length($filer = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::execute_cmd() "        .
                                "invalid value for \"cmd_filer\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/^-?cmd_pass_thru$/) {
            length($pass_thru = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::execute_cmd() "            .
                                "invalid value for \"cmd_pass_thru\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/^-?use_cmode_xml$/) {
            $use_cmode_xml = $args{$_};
            if (not length($use_cmode_xml) or
                $use_cmode_xml !~ /yes|no|1|0/i)
            {
                $self->_set_err("CmdsDriver::execute_cmd() "           .
                                "invalid value for \"use_cmode_xml\": " .
                                $args{$_});
                return $self;
            }

            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $use_cmode_xml = 1 if ($use_cmode_xml =~ /yes/i);
            $use_cmode_xml = 0 if ($use_cmode_xml =~ /no/i);
        }
        elsif (/^-?donot_append_vserver$/) {
            $donot_append_vserver = $args{$_};
            if (not length($donot_append_vserver) or
                $donot_append_vserver !~ /yes|no|1|0/i)
            {
                $self->_set_err("CmdsDriver::execute_cmd() "           .
                                "invalid value for \"donot_append_vserver\": " .
                                $args{$_});
                return $self;
            }

            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $donot_append_vserver = 1 if ($donot_append_vserver =~ /yes/i);
            $donot_append_vserver = 0 if ($donot_append_vserver =~ /no/i);
        }
        elsif (/^-?require_options$/) {
            $require_options = $args{$_};
            if (not length($require_options) or
                $require_options !~ /yes|no|1|0/i)
            {
                $self->_set_err("CmdsDriver::execute_cmd() "              .
                                "invalid value for \"require_options\": " .
                                $args{$_});
                return $self;
            }

            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $require_options = 1 if ($require_options =~ /yes/i);
            $require_options = 0 if ($require_options =~ /no/i);
        }
        elsif (/^-?check_output$/) {
            $check_output = $args{$_};
            if (not length($check_output) or
                $check_output !~ /yes|no|1|0/i)
            {
                $self->_set_err("CmdsDriver::execute_cmd() "           .
                                "invalid value for \"check_output\": " .
                                $args{$_});
                return $self;
            }

            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $check_output = 1 if ($check_output =~ /yes/i);
            $check_output = 0 if ($check_output =~ /no/i);
        }
        elsif (/^-?expected_output$/) {
            #
            # An empty output is valid, don't check length
            #
            $expected_output = $args{$_};
        }
        elsif (/^-?check_result$/) {
            $check_result = $args{$_};
            if (not length($check_result) or
                $check_result !~ /yes|no|1|0/i)
            {
                $self->_set_err("CmdsDriver::execute_cmd() "           .
                                "invalid value for \"check_result\": " .
                                $args{$_});
                return $self;
            }

            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $check_result = 1 if ($check_result =~ /yes/i);
            $check_result = 0 if ($check_result =~ /no/i);
        }
        elsif (/^-?expected_result$/) {
            length($expected_result = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::execute_cmd() "              .
                                "invalid value for \"expected_result\": " .
                                $args{$_});
                return $self;
            };
        }
        else {
            #
            # Anything else, we treat as a command argument and save
            #
            my $key = $_;
            my $value = $args{$key};
            $key =~ s/^-//; # remove the "-" prefix
            # If the filer is a Cmode filer then replace all _ with -
            # CLI translates this correctly already in CDef.pm::translate
            # Zapi requires this because the translation table looks
            #     for specific arguments that contain only - and not _
            if ($filer_mode =~ /Cmode/i) {
                $key=~ s/_/-/g; # Replace '_' with '-'
            }
            $cmd_args{$key} = $value;
            $self->sanlog(-msg => "Setting cmd_arg{$key} to $value",
                          -msg_level => 9);
            my $temp_val = $value;
            if (ref($value) =~ /ARRAY/)
            {
                $temp_val = Dumper($value);
            }
            $cmd_args_string .= "  $key = $temp_val\n";
        }
    } # end foreach arg


    #
    # Check for required arguments
    #
    unless (defined($cmd) and length $cmd)
    {
        $self->_set_err("CmdsDriver::execute_cmd(): " .
                        "missing required argument \"cmd\"");
        return ($self);
    }

    unless (defined($filer) and length $filer)
    {
        $self->_set_err("CmdsDriver::execute_cmd(): " .
                        "missing required argument \"filer\"");
        return ($self);
    }

    #
    # Log exactly what we are executing here.
    #
    my $exe_comment = "Execute Data\n";
    $exe_comment .= defined($cmd)             ? "  cmd             = $cmd\n" :
                                                "  cmd             = UNDEF\n";
    $exe_comment .= defined($interface)       ? "  interface       = $interface\n" :
                                                "  interface       = UNDEF\n";
    $exe_comment .= defined($filer_version)   ? "  filer_version   = $filer_version\n" :
                                                "  filer_version   = UNDEF\n";
    $exe_comment .= defined($filer_mode)      ? "  filer_mode      = $filer_mode\n" :
                                                "  filer_mode      = UNDEF\n";
    $exe_comment .= defined($use_cmode_xml)   ? "  use_cmode_xml   = $use_cmode_xml\n" :
                                                "  use_cmode_xml   = UNDEF\n";
    $exe_comment .= defined($privilege)       ? "  privilege       = $privilege\n" :
                                                "  privilege       = UNDEF\n";
    $exe_comment .= defined($vserver)         ? "  vserver         = $vserver\n" :
                                                "  vserver         = UNDEF\n";
    $exe_comment .= defined($vfiler)          ? "  vfiler          = $vfiler\n" :
                                                "  vfiler          = UNDEF\n";
    $exe_comment .= defined($vfiler_ip)       ? "  vfiler_ip       = $vfiler_ip\n" :
                                                "  vfiler_ip       = UNDEF\n";
    $exe_comment .= defined($partner)         ? "  partner         = $partner\n" :
                                                "  partner         = UNDEF\n";
    $exe_comment .= defined($output_format)   ? "  output_format   = $output_format\n" :
                                                "  output_format   = UNDEF\n";
    $exe_comment .= defined($log_output)      ? "  log_output      = $log_output\n" :
                                                "  log_output      = UNDEF\n";
    $exe_comment .= defined($check_output)    ? "  check_output    = $check_output\n" :
                                                "  check_output    = UNDEF\n";
    $exe_comment .= defined($check_result)    ? "  check_result    = $check_result\n" :
                                                "  check_result    = UNDEF\n";
    $exe_comment .= defined($expected_result) ? "  expected_result = $expected_result\n" :
                                                "  expected_result = UNDEF\n";
    $exe_comment .= defined($expected_output) ? "  expected_output = $expected_output\n" :
                                                "  expected_output = UNDEF\n";
    $exe_comment .= defined($filer)           ? "  filer           = $filer\n" :
                                                "  filer           = UNDEF\n";
    $exe_comment .= defined($timeout)         ? "  timeout         = $timeout\n" :
                                                "  timeout         = UNDEF\n";
    $exe_comment .= defined($maxbufsize)      ? "  maxbufsize      = $maxbufsize\n" :
                                                "  maxbufsize      = UNDEF\n";
    $exe_comment .= defined($cmd_timing)      ? "  cmd_timing      = $cmd_timing\n" :
                                                "  cmd_timing      = UNDEF\n";
    $exe_comment .= defined($pass_thru)       ? "  pass_thru       = $pass_thru\n" :
                                                "  pass_thru       = UNDEF\n";
    $exe_comment .= defined($require_options) ? "  require_options = $require_options\n" :
                                                "  require_options = UNDEF\n";
    $exe_comment .= "----Command Arguments----\n" . $cmd_args_string;
    $self->sanlog(-msg       => $exe_comment, -msg_level => 5);


    #
    # Create the hash key to look up.  We will find all commands
    # that match this key with an integer after it.  For example
    # "cli_lun_create" will match "cli_lun_create_0" and
    # "cli_lun_create_7.0".  We will use the filer_version to
    # figure out which entry to use.
    #
    my $cmd_key = $interface . "_" . $cmd;
    $cmd_key =~ s/ /_/g; # replace spaces with underscores

    my $cmd_found;
    my $cmd_hash;
    #
    # Since ZAPI is the same for both 7mode and Cmode, we only call
    # find_cdef_hash if cmd i/f is 'cli' and mode is 'Cmode'
    # otherwise use find_cmd_hash
    # If the user set '-use_cmode_xml', use XML data and find_cmd_hash()
    # to create the command even for Cmode. User must create XML data for
    # Cmode.
    #

    if ($filer_mode =~ /Cmode/i
        and $use_cmode_xml == 0)
    {
        ($cmd_found, $cmd_hash, %cmd_args) = $self->_find_cmode_hash($cmd_key,
                                                                     $filer_version,
                                                                     \%cmd_args
                                                                    );
    }
    else
    {
        ($cmd_found, $cmd_hash) = $self->_find_cmd_hash($cmd_key,
                                                        $filer_version);
    }

    unless ( (defined($cmd_found)) and (defined($cmd_hash)))
    {
        $self->_set_err("CmdsDriver::execute_cmd(): " .
                        "could not find command $cmd_key",
                        $self->errmsg());
        return ($self);
    }

    $self->sanlog(-msg       => "CmdsDriver::execute_cmd: " .
                                "command key is $cmd_key" .
                                " and command found is $cmd_found",
                  -msg_level => 3);

    my $dumpOut = Dumper($cmd_hash) if ($self->get_debug_level() >= 9);
    $self->sanlog(-msg       => "Command Hash ($cmd_found) is:\n$dumpOut",
                  -msg_level => 9);

    if ($interface =~ /cli/i)
    {
        my $cli_cmd = "";
        if ($filer_mode =~ /Cmode/i and $use_cmode_xml == 0)
        {
            $cli_cmd = $self->_get_cdef_cmd(-cmd_name        => $cmd_found,
                                            -cmd_hash        => $cmd_hash,
                                            -cmd_args        => \%cmd_args,
                                            -require_options => $require_options);
            if ($self->error())
            {
                return $self;
            }

            #
            # Now that we have the command.  Take care of priv set
            #
            if (length($privilege))
            {
                my $orig_privilege = length($self->privilege()) ? $self->privilege() : 'diag';
                $cli_cmd = "set -privilege $privilege ; " . $cli_cmd . " ; set -privilege $orig_privilege ; ";
            }
        }
        else
        {
            $cli_cmd = $self->_get_cli_cmd(-cmd_name        => $cmd_found,
                                           -cmd_hash        => $cmd_hash,
                                           -cmd_args        => \%cmd_args,
                                           -require_options => $require_options);
            if ($self->error())
            {
                return $self;
            }

            #
            # Now that we have the command.  Take care of vfiler context and/or
            # partner mode and/or priv set
            #
            if (length($vfiler) and ($vfiler ne "vfiler0"))
            {
                $cli_cmd = "vfiler run -q $vfiler " . $cli_cmd;
            }

            if ($partner)
            {
                $cli_cmd = "partner " . $cli_cmd;
            }
            if (length($privilege))
            {
        if ($partner)
        {
            $cli_cmd = "partner priv set $privilege ; " . $cli_cmd;
        }
        else
        {
                    $cli_cmd = "priv set $privilege ; " . $cli_cmd;
        }
            }
        }

        #
        # If the user has given us a pass_thru append it to the command line
        #
        if (length($pass_thru))
        {
            $cli_cmd .= " " . $pass_thru;
        }

        #
        # Now that we have the command created, execute it
        #
        $self->sanlog(-msg => "CmdsDriver::execute_cmd(): "  .
                              "Executing CLI command $filer" .
                              "::" . "$cli_cmd",
                      -use_loghdl => 0,
                      -msg_level => 1);
        my ($result, @output) = $self->filer_cli_cmd($cli_cmd,
                                                     -cmd_timeout    => $timeout,
                                                     -cmd_maxbufsize => $maxbufsize,
                                                     -cmd_timing     => $cmd_timing);
        if ($self->error())
        {
          return $self;
        }
        $self->sanlog(-msg => "$filer" . "::" . "$cli_cmd" .
                              " returned a $result",
                      -msg_level => 3);

        #
        # Fill the return hash
        #
        my $output_str = (join "\n", @output);
        $self->{resultsHash}{result} = $result;
        $self->{resultsHash}{reason} = "";
        $self->{resultsHash}{output} = $output_str;

    } # end if CLI interface
    else
    {
        #
        # Zapi interface
        #

        #
        # Steps to make this happen:
        # 1) Create a server (ZapiTest) object
        # 2) Find the Zapi command from the cmd_hash
        # 3) Check for required params
        # 4) Create an element (NaElement) for the command (api)
        # 5) Add to the element object (NaElement child_add) for each cmd arg
        # 6) invoke_elem to execute
        # 7) get results

        #
        # for Cmode, we need to handle vserver-scoped for certain commands
        #
        if ($filer_mode =~ /Cmode/i and $use_cmode_xml == 0)
        {
            #
            # These commands are outside vserver scoped
            #
            if ($cmd_found !~ /^(storage|net|vserver)/
                and $donot_append_vserver == 0)
            {
                #
                # For Cmode, we set Vfiler = Vserver for vserver scoped
                #
                if (defined $cmd_args{vserver}) {
                    $vfiler = delete $cmd_args{vserver};

                    #
                    # since we set vserver scoped, we have to strip out vserver
                    # option from required_options if it exists
                    #
                    $cmd_hash->{required_options} =~ s/vserver//g;
                } else {
                    $vfiler = $vserver;
                }
            }
        }

        #
        # Create a server (ZapiTest) object
        #
        my $server_obj = $self->_create_zapi_server($filer,
                                                    $vfiler,
                                                    $vfiler_ip);
        if ($self->error())
        {
          return $self;
        }

        #
        # Create an element (NaElement) for the command (api) and
        # load the element object with the cmd args
        #
        my $element_obj =
            $self->_get_zapi_element(-cmd_name        => $cmd_found,
                                     -cmd_hash        => $cmd_hash,
                                     -cmd_args        => \%cmd_args,
                                     -pass_thru       => $pass_thru,
                                     -cmd             => $cmd,
                                     -require_options => $require_options);
        if ($self->error())
        {
          return $self;
        }
        $self->sanlog(-msg       => "ZAPI element is:\n" . Dumper($element_obj),
                      -msg_level => 9);

        #
        # Execute and get results
        # Set an alarm around the ZAPI call because there are situations where this will
        # hang without the alarm.  Note we will use the timeout attribute which defaults
        # to 30 minutes.
        #
        my $orig_alarm = $SIG{ALRM};
        $SIG{ALRM} = sub { die "CMDSDRIVER_ZAPI_TIMEOUT"};
        my $output_obj;

        # set start time
        my $start_time = [Time::HiRes::gettimeofday()];

        eval
        {
            alarm($timeout);
            $output_obj = $server_obj->invoke_elem($element_obj,"any");
            alarm(0);
        };
        $SIG{ALRM} = $orig_alarm;

        #
        # If we timed out, we should return an error
        #
        if ($@ and $@ =~ /CMDSDRIVER_ZAPI_TIMEOUT/)
        {
            $self->_set_err("CmdsDriver::execute_cmd(): " .
                            "Timed out executing ZAPI command $cmd_key ($@)");
            return ($self);
        } elsif ($@) {

            #
            # else if another error occured
            #
            $self->_set_err("CmdsDriver::execute_cmd(): " .
                            "Error executing ZAPI command $cmd_key ($@)");
            return ($self);
        }

        #
        # Get end time and calculate the command execution time
        #
        my $elapsed_time =  Time::HiRes::tv_interval($start_time);
        $self->{command_time} = $elapsed_time;

        my $result = $output_obj->results_errno();
        my $reason = join("\n", $output_obj->results_reason());
        my $output = $output_obj->toString();

        #
        # Make sure we have something defined and readable
        # in the results hash and print.
        #
        $result = "" unless defined($result);
        $reason = "" unless defined($reason);
        $output = "" unless defined($output);

        my $log_level = 3;
        $log_level = 1 if ($log_output);
        $self->sanlog(-msg =>
                      "Zapi output:\n"        .
                      "  result  = $result\n" .
                      "  reason = $reason\n"  .
                      "  output = \n$output\n",
                      -msg_level => $log_level
                      );

        # log the duration of the zapi if requested by the user
        if ($cmd_timing == 1) {
            # determine singular or plural second
            my $sec_str = ($elapsed_time =~ /^(0|1)\./) ? "seconds" : "second";
            my $node;
            if (defined($self->{node})) {
                $node = $self->{node};
            } else {
                $node = $self->filer();
            }
            my $start = $start_time->[0] . "." . $start_time->[1];
            my $end = $start + $elapsed_time;
            $self->sanlog(-msg => ":CLV: node=($node) zapi=($cmd_found) elapsed=($elapsed_time $sec_str) start=($start) end=($end)");
        }

        #
        # If the debug level is 3 or higher, parse the output and dump it
        #
        my $current_debug_level = $self->get_debug_level();
        if ($current_debug_level > 2)
        {
            my $log_hash;
            eval
            {
                $log_hash = $self->_xml_to_hash($output);
            };
            if ($@)
            {
                #
                # Don't fail, just change the log, this is logging code
                # if the user wanted raw, it would be wrong to error
                #
                $self->sanlog(-msg =>"Unable to process XML into Hash " .
                                     "for logging",
                              -msg_level => 3);
            }
            else
            {
                my $dumpOut = Dumper($log_hash) if ($self->get_debug_level() >= 3);;
                $self->sanlog(-msg =>"Dumper of output hash:\n$dumpOut",
                              -msg_level => 3);
            }
        }

        #
        # Fill the return hash
        #
        $self->{resultsHash}{result} = $result;
        $self->{resultsHash}{reason} = $reason;

        if ($output_format =~ /raw/i)
        {
            $self->{resultsHash}{output} = $output;
        }
        elsif ($output_format =~ /system/i)
        {
            if ($cmd_found =~ /^zapi_system-cli_/i)
            {
                my $temp = $output_obj->child_get_string("cli-output");
                $self->{resultsHash}{output} = $temp;
            }
            else
            {
                $self->_set_err("CmdsDriver::execute_cmd():\n The command: " .
                $cmd_found . ", cannot use the output format 'system'. It is " .
                "only to be used for ZAPI command 'system-cli'.");
                return ($self);
            }
        }
        else
        {
            # For some reason, the zapi output for qtree_list_iter contains
            # a string <id>\d+</id> which can't get processed by XML2Hash.  So
            # I'm replacing id with qid to get the hash processed.
            if ($cmd eq "volume_qtree_show" or $cmd eq "qtree_list_iter") {
                $output =~ s/<id>(\d+)<\/id>/<qid>$1<\/qid>/g;
            }

            #
            # for Cmode, if there is failure, just return output str
            #
            if ($result and $filer_mode =~ /Cmode/i) {
                $self->{resultsHash}{output} = $output;
            } else {
                my $output_hash;
                eval {
                    $output_hash = $self->_xml_to_hash($output);
                };
                if ($@)
                {
                    $self->_set_err("CmdsDriver::execute_cmd():\n    $@\n " .
                        "Failure converting XML to hash.  One common cause of " .
                        "this is executing a system-cli ZAPI with a hash " .
                        " output format.\nTo fix, try executing with a 'system'".
                        "output format and processing the XML returned in your ".
                        "script");
                    return ($self);
                }
                $self->{resultsHash}{output} = $output_hash;
            }
        }
    } # end if ZAPI interface

    #
    # Do we check the output vs the expected output?
    #
    $self->{resultsHash}{output_check} = 0;
    if ($check_output)
    {
        my $output_type = "regex";
        unless (defined($expected_output))
        {
            ($expected_output, $output_type) =
                $self->_get_expected_output(-cmd_name => $cmd_found,
                                            -cmd_hash => $cmd_hash);
        }

        #
        # Default to no match and then check to see if we match
        #
        $self->{resultsHash}{output_check} = 0;

        #
        #
        #

        if ($output_type eq "regex" or
            $output_type eq "regex_neg")
        {
            if (ref($expected_output) =~ /HASH/)
            {
                $self->sanlog(-msg => "Translating expected output to string",
                              -msg_level => 7);
                #
                # For simple zapi expected output you'll get a hash from
                # something like: <results status="passed">^$</results> or
                # <results status="passed"><some_tag>\d+</some_tag></results>
                # We are going to special case these 2 cases and make it
                # be a string again and just do the regex check.
                #
                # We are expecting a hash that looks like this:
                # {
                #   'results' => {
                #     'status' => 'passed',
                #     'content' => '^$'    (or the some_tag)
                #   }
                # }
                #
                my $eo_hash = $expected_output;
                $expected_output = "";
                if (defined($eo_hash->{results}))
                {
                    $expected_output .= "<results status=\"";
                    my $results_hash = $eo_hash->{results};
                    if (defined($eo_hash->{results}{status}))
                    {
                        $expected_output .= $eo_hash->{results}{status};
                    }
                    $expected_output .= "\">";
                    if (defined($eo_hash->{results}{content}))
                    {
                        my $content = $eo_hash->{results}{content};
                        $content = "" if ($content eq "\^\$");
                        $expected_output .= $content;
                    }
                    else
                    {
                        foreach my $key (keys %{$eo_hash->{results}})
                        {
                            next if $key =~ /status/;
                            $expected_output .= "<" . $key . ">";
                            $expected_output .= $eo_hash->{results}{$key};
                            $expected_output .= "</" . $key . ">";
                        }
                    } # end if some_tag case
                    $expected_output .= "</results>";
                } # end if results hash defined
            } # end if expeced output is a HASH

            $self->sanlog(-msg => "Checking expected output.\n"            .
                                  "  Output type     = $output_type\n"     .
                                  "  Expected output = $expected_output\n" .
                                  "  Command output  = "                   .
                                        $self->{resultsHash}{output},
                          -msg_level => 7);

            if ($output_type eq "regex")
            {
                if ($self->{resultsHash}{output} =~ /$expected_output/)
                {
                    $self->sanlog(-msg => "Expected output matched",
                                  -msg_level => 7);
                    $self->{resultsHash}{output_check} = 1;
                }
                else
                {
                    $self->sanlog(-msg => "Expected output did not match",
                                  -msg_level => 7);
                }
            }
            else
            {
                if ($self->{resultsHash}{output} !~ /$expected_output/)
                {
                    $self->sanlog(-msg => "Expected output is correct",
                                  -msg_level => 7);
                    $self->{resultsHash}{output_check} = 1;
                }
                else
                {
                    $self->sanlog(-msg => "Expected output is not correct",
                                  -msg_level => 7);
                }
            }
        } # end if output_type is regex
        elsif ($output_type eq "regex_neg")
        {
            if ($self->{resultsHash}{output} !~ /$expected_output/)
            {
                $self->sanlog(-msg => "Expected output is correct",
                              -msg_level => 7);
                $self->{resultsHash}{output_check} = 1;
            }
            else
            {
                $self->sanlog(-msg => "Expected output is not correct",
                              -msg_level => 7);
            }
         }
        elsif ($output_type eq "xml")
        {
            $self->sanlog(-msg => "Checking Expected output type 'xml'",
                          -msg_level => 7);
            $self->{resultsHash}{output_check} =
                $self->_check_output_xml($expected_output,
                                  $self->{resultsHash}{output});
            # dave, implement something here
        }
        else
        {
            $self->_set_err("CmdsDriver::execute_cmd() " .
                            "unsupported expected output type: " .
                            $output_type);
            return $self;
        }
    } # end if check_output

    #
    # Do we check the result vs the expected result?
    #
    $self->{resultsHash}{result_check} = 0;
    if ($check_result)
    {
        unless (defined($expected_result))
        {
            $expected_result =
                $self->_get_expected_result(-cmd_name => $cmd_found,
                                            -cmd_hash => $cmd_hash);
        }

        #
        # Default to no match and then check to see if we match
        #
        $self->{resultsHash}{result_check} = 0;

        $self->sanlog(-msg => "Checking expected result.\n"          .
                              "  Expected result = $expected_result\n" .
                              "  Result is       = " .
                                        $self->{resultsHash}{result},
                      -msg_level => 7);

        #
        # The expected result can be a space separated list of integers.
        #
        my @result_list = split(/\s+/, $expected_result);
        foreach my $current_result (@result_list)
        {
            if ($current_result == ($self->{resultsHash}{result}))
            {
                $self->sanlog(-msg => "Expected result matched",
                              -msg_level => 7);
                $self->{resultsHash}{result_check} = 1;
            }
        }
        if ($self->{resultsHash}{result_check} == 0)
        {
            $self->sanlog(-msg => "Expected result did not match",
                          -msg_level => 7);
        }
    } # end if check_result

    #
    # Return a reference to the hash
    #
    return \%{$self->{resultsHash}};
} # end execute_cmd

###############################################################################
# Method:    filer_cli_cmd
# Objective: Issue arbitrary command to filer and get results.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub filer_cli_cmd
{
    my($self, $cmd, %args) = @_;
    my @raw_output;
    my $buffer;
    my $hide = 0;
    my $timeout = $self->timeout();
    my $maxbufsize = $self->maxbufsize();
    my $cmd_timing = $self->cmd_timing();
    my $errmsg;
    my $sentinal;
    my ($start_time, $elapsed_time, $sec_str);

    #
    # Clear errors.
    #
    $self->_clear_err;

    foreach (keys %args)
    {
        if (/-?hide/)
        {
            $hide = $args{$_};
            if (not length($hide))
            {
                $self->_set_err(
                    "CmdsDriver::filer_cli_cmd() invalid value for \"hide\": " .
                                $args{$_});
                return (1, []);
            }
            next;
        }
        elsif (/^-?cmd_timeout$/)
        {
            $timeout = $args{$_};
            if (not length($timeout))
            {
                $self->_set_err(
                    "CmdsDriver::filer_cli_cmd() invalid value for \"cmd_timeout\": " .
                                $args{$_});
                return (1, []);
            }
            next;
        }
        elsif (/^-?cmd_maxbufsize$/)
        {
            $maxbufsize = $args{$_};
            if (not length($maxbufsize))
            {
                $self->_set_err(
                    "CmdsDriver::filer_cli_cmd() invalid value for \"cmd_maxbufsize\": " .
                                $args{$_});
                return (1, []);
            }
            next;
        }
        elsif (/^-?cmd_timing$/)
        {
            $cmd_timing = $args{$_};
            next;
        }
    } # end foreach arg

    #
    # Make sure we were given a command
    #
    unless (length $cmd)
    {
        $self->_set_err("CmdsDriver::filer_cli_cmd() cmd: invalid CLI command");
        return (1, []);
    }
    #
    # Add a result call and sentinal to the cmd
    #
    $cmd .= ";result -l;echo DoNe" unless ($self->filer_mode eq 'Cmode');

    #
    # Log the command if we are supposed to show them
    #
    if ((!$hide) and $self->{show_cmd})
    {
        $self->sanlog(-msg => "CmdsDriver::filer_cli_cmd(): " .
                              "Executing Command: " . $cmd,
                      -use_loghdl => 0,
                      -msg_level  => 3);
    }

    return (1, []) unless ($self->{execution});

    #
    # Remove any carriage returns from the cli command
    #
    $cmd =~ s/[\n]/ /g;

    # get connection hosttype
    my $conn = $self->connection();

    # is connection cserver type?
    if ($conn->hostp->hosttype() =~ /cserver/i) {

        # check connection is alive. if not, reconnect
        unless ($conn->is_responsive()) {
            $conn = $self->reconnect();
            if ($self->error()) {
                $self->_set_err("failed to reconnect ", $self->errmsg());
                return (1, []);
            }
        }
    }

    #
    # Init the execution timer
    #
    # set start time
    $start_time = [Time::HiRes::gettimeofday()];

    my $timeout_error_msg = "";

    # Issue command and get results.
    (undef,undef,$buffer,undef)
        = Tharn::say($self->{connection},$cmd,
                     -timeout=>$timeout,
                     -maxbufsize=>$maxbufsize,
                     -eval=>sub{
                         if (@_) {
                             my $errmsg = join("",@_);
                             $errmsg =~ s/\n//g;
                             #print NTEST_STDOUT "$errmsg\n";
                             Tharn::logresult("CONFIG",abort=>0,"\n@_");
                             if ($errmsg =~ /say:\s*Err=TIMEOUT/i) {
                                 # burt506765: set an error on this method if a timeout occurred
                                 $timeout_error_msg = $errmsg;
                             }
                         }
                     });

    #
    # Get end time and calculate the command execution time
    #
    $elapsed_time =  Time::HiRes::tv_interval($start_time);
    $self->{command_time} = $elapsed_time;

    # log the duration of the command if requested by the user
    if ($cmd_timing == 1) {
        # determine singular or plural second
        $sec_str = ($elapsed_time =~ /^(0|1)\./) ? "seconds" : "second";
        my $node;
        if (defined($self->{node})) {
            $node = $self->{node};
        } else {
            $node = $self->filer();
        }
        my $start = $start_time->[0] . "." . $start_time->[1];
        my $end = $start + $elapsed_time;
        $self->sanlog(-msg => ":CLV: node=($node) cli=($cmd) elapsed=($elapsed_time $sec_str) start=($start) end=($end)");
    }

    #
    # Find and remove the sentinal DoNe and the result from the buffer.
    # Note, this makes an assumption.  It assumes that "DoNe\n" and
    # "result=\d+\n" get printed out intact.
    #
    $self->sanlog(-msg => "Raw Command Output:\n" .
                          "----------------------------------\n" .
                          $buffer                                .
                          "----------------------------------\n",
                  -msg_level => 9);

    my $result       = -1;  # Setting a default, should be updated later

    # Cmode doesn't not have result command so ignore the below checking
    if ($self->filer_mode eq '7mode') {
        $buffer =~ s/(DoNe\r?\n)//;
        my $done_found = $1;

        #
        # If we didn't find the DoNe sentinal, there's a chance that some
        # text got between the 'DoNe' and the end of the line.  We can try
        # to remove the two parts of the string separately.
        if ($done_found !~ /DoNe/)
        {
          $buffer =~ s/(.*)(DoNe)(.*)(\r?\n)(.*)/$1$3$5/;
          $done_found = $2;
        }

        $buffer =~ s/(result=\d+\r?\n)//;
        my $result_found = $1;

        #
        # If we didn't find the result, print something at least
        #
        if ($result_found =~ /result=(\d+)/)
        {
            $result = $1;
        }
        else
        {
            $self->sanlog(-msg => "Warning: CmdsDriver::filer_cli_cmd() Output " .
                                  "string did not contain expected result");
        }

        #
        # If we didn't find the sentinal DoNe, print something at least
        #
        unless ($done_found =~ /DoNe/)
        {
            $self->sanlog(-msg => "Warning: CmdsDriver::filer_cli_cmd() Output " .
                             "string did not contain expected sentinal 'DoNe'");
        }
    } else {
        # if no error in the buffer string, we're ok.
        $result = 0 unless ($buffer =~ /ERROR:/i);
    }
    #
    # Split the buffer into array
    #
    @raw_output = split(/\n/,$buffer);

    # if a timeout occurred, then set the result to positive, and set an error
    if (length($timeout_error_msg)) {
        $self->_set_err("A timeout occurred while executing the command $cmd (timeout: $timeout)" .
                        ". To adjust the timeout, increase the value of CMDSDRIVER_CMD_TIMEOUT" .
                        "\nError: $timeout_error_msg");
        return (1, @raw_output);
    }

    return ($result, @raw_output);

} # end filer_cli_cmd

###############################################################################
# Method:    discover_filer_version
# Objective: Find the filer version, format and save
# Details:   See POD documentation at the end of this file.
# dave - Make this take an option and will discover either from the filer (say)
#        or from the host file for the filer.
###############################################################################
sub discover_filer_version
{
    my($self, $filer_name) = @_;

    #
    # Find the version in X.X format
    #
    my $host    = host($filer_name);
    my $version = get_ontap_version($host);
    $self->sanlog(-msg       => "Filer::Version::get_ontap_version " .
                                "returned $version",
                  -msg_level => 3);

    #
    # Fromat and save the version
    #
    # dave $version = $self->_format_version($version);
    $self->filer_version($version);
} # end discover_filer_version

###############################################################################
#                               Accessor Methods
###############################################################################
sub timeout
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{timeout} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                   "timeout set to $value",
                      -msg_level => 3);
    }
    return $self->{timeout};
}

sub maxbufsize
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{maxbufsize} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                   "maxbufsize set to $value",
                      -msg_level => 3);
    }
    return $self->{maxbufsize};
}

sub cmd_timing
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{cmd_timing} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                   "cmd_timing set to $value",
                      -msg_level => 3);
    }
    return $self->{cmd_timing};
}

sub filer_version
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        # dave $self->{filer_version} = $self->_format_version($value);
        $self->{filer_version} = $value;
        $self->sanlog(-msg => ref($self) . "::" .
                              "filer_version set to $value",
                      -msg_level => 3);
    }
    return $self->{filer_version};
}

sub filer_mode
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        #
        # Normalize known values so user doesn't have to care about
        # capitalization (7Mode, 7MODE, 7MoDe all OK here)
        #
        $value = "7mode" if ($value =~ /7Mode/i);
        $value = "Cmode" if ($value =~ /CMode/i);
        $self->{filer_mode} = $value;
        $self->sanlog(-msg => ref($self) . "::" .
                              "filer_mode set to $value",
                      -msg_level => 3);
    }
    return $self->{filer_mode};
}

sub filer
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{filer} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                    "filer set to $value",
                      -msg_level => 3);
    }
    return $self->{filer};
}

sub interface
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{cmd_interface} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                    "cmd_interface set to $value",
                      -msg_level => 3);
    }
    return $self->{cmd_interface};
}

sub privilege
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{privilege} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                    "privilege set to $value",
                      -msg_level => 3);
    }
    return $self->{privilege};
}

sub vfiler
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{vfiler} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                    "vfiler set to $value",
                      -msg_level => 3);
    }
    return $self->{vfiler};
}

sub vfiler_ip
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        if (length($value))
        {
            $self->{vfiler_ip} = $value;
            $self->sanlog(-msg => ref($self) . "::" .
                                  "vfiler_ip set to $value",
                          -msg_level => 3);
        }
        else
        {
            $self->{vfiler_ip} = undef;
            $self->sanlog(-msg => ref($self) . "::" .
                                  "vfiler_ip set to undef",
                          -msg_level => 3);
        }
    }
    return $self->{vfiler_ip};
}

sub vserver
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{vserver} = $value;
        $self->sanlog(-msg => ref($self) . "::" .
                                    "vserver set to $value",
                      -msg_level => 3);
    }
    return $self->{vserver};
}

sub partner_mode
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{partner_mode} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                    "partner_mode set to $value",
                      -msg_level => 3);
    }
    return $self->{partner_mode};
}

sub show_cmd
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{show_cmd} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                    "show_cmd set to $value",
                      -msg_level => 3);
    }
    return $self->{show_cmd};
}

sub output_format
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{output_format} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                    "output_format set to $value",
                      -msg_level => 3);
    }
    return $self->{output_format};
}

sub log_output
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{log_output} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                    "log_output set to $value",
                      -msg_level => 3);
    }
    return $self->{log_output};
}

sub connection
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{connection} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                    "connection set",
                      -msg_level => 3);
    }

    #
    # When we are given a connection we are going to check it to see what the
    # mode is and set the filer_mode attribute based on what we find.  This
    # way users should be able to just create and CmdsDriver object and
    # set the connection and then get whatever mode their filer is in.
    #
    my $filer_mode = $self->_discover_filer_mode($self->{connection});
    $self->{filer_mode} = $filer_mode;
    $self->sanlog(-msg       => ref($self) . "::" .
                                "Setting filer mode to $filer_mode",
                  -msg_level => 3);
    return $self->{connection};
}

sub command_time
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{command_time} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                    "command_time set",
                      -msg_level => 3);
    }
    return $self->{command_time};
}

sub log
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{log} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                    "log set",
                      -msg_level => 3);
    }
    return $self->{log};
}

sub execution
{
    my ($self, $value) = @_;
    if (defined($value))
    {
        $self->{execution} = $value;
        $self->sanlog(-msg       => ref($self) . "::" .
                                    "execution set to $value",
                      -msg_level => 3);
    }
    return $self->{execution};
}

sub resultsHash
{
    my ($self) = @_;
    return \%{$self->{resultsHash}};
}

###############################################################################
#                               Private Methods
###############################################################################


###############################################################################
# Method:    _get_cli_cmd()
# Objective: Give a command name and a hash with data about that cmd, create
#            the basic command line for that command and return it
# Inputs:    -cmd_name - hash key with the command name
#            -cmd_hash - hash key with the cmd information hash
#            -cmd_args - hash key with the cmd arguments hash
#            -require_options - check for required options?
# Outputs:   command line to execute
#            On failure the error flag and message will be set
# Notes:
###############################################################################
sub _get_cli_cmd
{
    my($self, %args) = @_;
    my $full_cmd_name   = undef;
    my $cmd_hash        = undef;
    my $cmd_args        = undef;
    my $require_options = 1;

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?cmd_name$/) {
            length($full_cmd_name = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::_get_cli_cmd() "      .
                                "invalid value for \"cmd_name\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/-?cmd_hash$/) {
            length($cmd_hash = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::_get_cli_cmd() "      .
                                "invalid value for \"cmd_hash\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/-?cmd_args$/) {
            length($cmd_args = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::_get_cli_cmd() "      .
                                "invalid value for \"cmd_args\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/-?require_options$/) {
            length($require_options = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::_get_cli_cmd() "             .
                                "invalid value for \"require_options\": " .
                                $args{$_});
                return $self;
            };
        }
        else {
            $self->_set_err("CmdsDriver::_get_cli_cmd() " .
                            "Invalid parameter specified: $_");
            return $self;
        }
    } # end foreach arg

    #
    # Check for required arguments
    #
    unless (defined($cmd_hash))
    {
        $self->_set_err("CmdsDriver::_get_cli_cmd(): " .
                        "missing required argument \"cmd_hash\"");
        return ($self);
    }

    unless (defined($full_cmd_name))
    {
        $self->_set_err("CmdsDriver::_get_cli_cmd(): " .
                        "missing required argument \"cmd_name\"");
        return ($self);
    }

    unless (defined($cmd_args))
    {
        $self->_set_err("CmdsDriver::_get_cli_cmd(): " .
                        "missing required argument \"cmd_args\"");
        return ($self);
    }

    #
    # Let's pull some information out of the command hash
    #
    my $required_options = $cmd_hash->{required_options};
    if (ref($required_options) =~ /^HASH{/)
    {
        $required_options = "__NONE__";
    }
    $required_options =~ s/^\s+//;        # strip leading whitespce
    $required_options =~ s/\s+$//;        # strip trailing whtspce

    my $optional_options = $cmd_hash->{optional_options};
    if (ref($optional_options) =~ /^HASH{/)
    {
        $optional_options = "__NONE__";
    }
    $optional_options =~ s/^\s+//;        # strip leading whitespce
    $optional_options =~ s/\s+$//;        # strip trailing whtspce

    my $option_order = $cmd_hash->{option_order};
    if (ref($option_order) =~ /^HASH{/)
    {
        $option_order = "__NONE__";
    }
    $option_order =~ s/^\s+//;        # strip leading whitespce
    $option_order =~ s/\s+$//;        # strip trailing whtspce

    my $options_hash = $cmd_hash->{options};

    #
    # Let's make sure we have something for all the required options
    #
    if ($require_options)
    {
        foreach my $option (split " ", $required_options)
        {
            last if ($required_options eq "__NONE__");
            unless (defined($cmd_args->{$option}))
            {
                $self->_set_err("CmdsDriver::_get_cli_cmd(): The "     .
                                "$full_cmd_name command requires the " .
                                "argument \"$option\"");
                return ($self);
            }
        }
    }

    #
    # Now that we know we have all that is required, let's put the
    # command line together
    #
    my $cmd = $cmd_hash->{command};
    $cmd =~ s/^\s+//;        # strip leading whitespce
    $cmd =~ s/\s+$//;        # strip trailing whtspce
    foreach my $option (split " ", $option_order)
    {
        last if ($option_order eq "__NONE__");
        my $switch ="";
        if (defined($options_hash->{$option}))
        {
            $switch = $options_hash->{$option};
            if ($switch eq "__STRING__")
            {
                $switch ="";
            }
        }
        if (defined($cmd_args->{$option}))
        {
            my $value = $cmd_args->{$option};
            if (length($switch))
            {
                $cmd .= " $switch $value";
            }
            else
            {
                $cmd .= " $value";
            }
        }
    } # end for each option_order

    return $cmd;

} # end _get_cli_cmd

###############################################################################
# Method:    _get_cdef_cmd()
# Objective: Give a command name and a hash with data about that cmd, create
#            the basic command line for that command and return it
# Inputs:    -cmd_name - hash key with the command name
#            -cmd_hash - hash key with the cmd information hash
#            -cmd_args - hash key with the cmd arguments hash
#            -require_options - check for required options?
# Outputs:   command line to execute
#            On failure the error flag and message will be set
# Notes:
###############################################################################
sub _get_cdef_cmd
{
    my($self, %args) = @_;
    my $cmd_name        = undef;
    my $cmd_hash        = undef;
    my $cmd_args        = undef;
    my $cmd             = "";
    my $require_options = 1;
    my ($schema_obj);

    #
    # We now have cdef cmd and its cmd hash.
    # We could use them to translate to cli cmd but we should used cdefs_obj
    # to translate because it's implemented in cdef object
    #

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?cmd_name$/) {
            length($cmd_name = $args{$_}) or
            do {
                $self->_set_err("Invalid value for \"cmd_name\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/-?cmd_hash$/) {
            #
            # We don't need cmd_hash arg anymore so delete it after it is set
            #
            length($cmd_hash =  delete $args{$_}) or
            do {
                $self->_set_err("Invalid value for \"cmd_hash\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/-?cmd_args$/) {
            length($cmd_args = $args{$_}) or
            do {
                $self->_set_err("Invalid value for \"cmd_args\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/-?require_options$/) {
            length($require_options = $args{$_}) or
            do {
                $self->_set_err("Invalid value for \"require_options\": " .
                                $args{$_});
                return $self;
            };
        }
        else {
            $self->_set_err("Invalid parameter specified: $_");
            return $self;
        }
    } # end foreach arg

    #
    # schema object should already be loaded on _populate_commands() call
    # get schema object
    #
    if (defined($self->{cmode_schema}) and
        ref($self->{cmode_schema}) =~ /Schema/) {
        $schema_obj = $self->{cmode_schema};
    } else {
        $self->_set_err("Failed to find schema object");
        return "";
    }

    $cmd_args->{cdef} = $cmd_name;
    $cmd_args->{filer} = $self->get("filer");

    #
    # translate cdef and its options into Cmode cli cmd
    #
    $cmd = $schema_obj->translate(%{$cmd_args});
    if ($schema_obj->error()) {
        $self->_set_err("failed to translate cdef: " . $cmd_name,
                        $schema_obj->errmsg());
        return "";
    }

    return $cmd;

} # end _get_cdef_cmd

###############################################################################
# Method:    _create_zapi_server()
# Objective: Create a zapi server (ZapiTest) object.
# Inputs:    filer - name of the filer
# Outputs:   ZapiTest object
#            On failure the error flag and message will be set
# Notes:
###############################################################################
sub _create_zapi_server
{
    my($self, $filer, $vfiler, $vfiler_ip) = @_;

    my $server_obj;
    my $server_str;

    my $host     = host($filer);
    my $ip       = $host->{default_ip};
    my ($username, $password);
    #
    # set default user and password depned on filer_mode
    #
    if ($self->filer_mode() eq '7mode')
    {
        $username = $self->ZAPI_USERNAME_7MODE;
        $password = $host->{default_password};
        $server_str = "vfiler";
    }
    else
    {
        $username = $self->ZAPI_USERNAME_CMODE;
        $password = $self->ZAPI_PASSWORD_CMODE;
        $server_str = "vserver";
    }
    #
    # use $username and $password for ZAPI connection from host object
    # if they are set
    #
    if (defined $host->{connection_fields}{zapi}{username} and
        length($host->{connection_fields}{zapi}{username}))
    {
        $username = $host->{connection_fields}{zapi}{username};
    }
    if (defined $host->{connection_fields}{zapi}{password} and
        length($host->{connection_fields}{zapi}{password}))
    {
        $password = $host->{connection_fields}{zapi}{password};
    }

    # Use Vfiler_IP if defined
    # Check for vfiler_ip to be undef or an empty string
    if (length($vfiler) and defined $vfiler_ip and length($vfiler_ip)) {
            $ip = $vfiler_ip;
            $self->sanlog(-msg      => "Executing on vfiler $vfiler:$vfiler_ip",
                          -msg_level => 7);
    }

    $server_obj = new ZapiTest($ip, $username, $password, $self->timeout());
    unless (defined($server_obj))
    {
        $self->_set_err("CmdsDriver::_create_zapi_server: " .
                        "Unable to connect via ZAPI to filer $filer at $ip");
        return undef;
    }

    # Use Vfiler context switch if Vfiler_IP is not defined
    if (length($vfiler) and not length($vfiler_ip)) {
        # Set the vfiler context where the zapi commmand is to be run
        $server_obj->set_vfiler_server ($vfiler);

        # we need to display which server context the command is executed on
        $self->sanlog(-msg => "Executing on $server_str $vfiler",
                      -msg_level => 1);
    }

    # to resolve the long delay in Cmode, we need to set_legacy_http_conn
    if ($self->filer_mode() eq 'Cmode')
    {
        $server_obj->set_legacy_http_conn(0);
    }

    return $server_obj;
} # end _create_zapi_server

###############################################################################
# Method:    _get_zapi_element()
# Objective: Given a command name and a hash with data about that cmd, create
#            the NaElement object for that command and return it
# Inputs:    -cmd_name - hash key with the command name
#            -cmd_hash - hash key with the cmd information hash
#            -cmd_args - hash key with the cmd arguments hash
#            -pass_thru- space separated list of name=value pairs to be added
#                        to the element
#            -require_options - if not set, we don't check for required options
# Outputs:   command line to execute
#            On failure the error flag and message will be set
#
# Notes:     The NaElement returned can be as simple as an element created
#            from a single name/value pair or as complex as a hash of a hash
#            that contains an array of hash, etc.
###############################################################################
sub _get_zapi_element
{
    my($self, %args)    = @_;
    my $full_cmd_name   = undef;
    my $cmd_hash        = undef;
    my $cmd_args        = undef;
    my $cdef_cmd        = undef;
    my $debug_msg       = "";
    my $pass_thru       = "";
    my $require_options = 1;

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?cmd_name$/) {
            length($full_cmd_name = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::_get_zapi_element() " .
                                "invalid value for \"cmd_name\": " .
                                $args{$_});
                return $self;
            };
            $self->sanlog(-msg => "In _get_zapi_element with $full_cmd_name",
                          -msg_level => 9);
        }
        elsif (/-?cmd_hash$/) {
            length($cmd_hash = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::_get_zapi_element() " .
                                "invalid value for \"cmd_hash\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/-?cmd_args$/) {
            length($cmd_args = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::_get_zapi_element() " .
                                "invalid value for \"cmd_args\": " .
                                $args{$_});
                return $self;
            };
        }
        elsif (/-?cmd$/) {
            $cdef_cmd = $args{$_};
        }
        elsif (/-?pass_thru$/) {
            $pass_thru = $args{$_};
        }
        elsif (/-?require_options$/) {
            $require_options = $args{$_};
            if (not length($require_options) or
                $require_options !~ /yes|no|1|0/i)
            {
                $self->_set_err("CmdsDriver::_get_zapi_element() "        .
                                "invalid value for \"require_options\": " .
                                $args{$_});
                return $self;
            }

            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $require_options = 1 if ($require_options =~ /yes/i);
            $require_options = 0 if ($require_options =~ /no/i);
        }
        else {
            $self->_set_err("CmdsDriver::_get_zapi_element() " .
                            "Invalid parameter specified: $_");
            return $self;
        }
    } # end foreach arg

    #
    # Check for required arguments
    #
    unless (defined($cmd_hash))
    {
        $self->_set_err("CmdsDriver::_get_zapi_element(): " .
                        "missing required argument \"cmd_hash\"");
        return ($self);
    }

    unless (defined($full_cmd_name))
    {
        $self->_set_err("CmdsDriver::_get_zapi_element(): " .
                        "missing required argument \"cmd_name\"");
        return ($self);
    }

    unless (defined($cmd_args))
    {
        $self->_set_err("CmdsDriver::_get_zapi_element(): " .
                        "missing required argument \"cmd_args\"");
        return ($self);
    }

    #
    # Let's pull some information out of the command hash
    #
    my $cmd = $cmd_hash->{command};
    $cmd =~ s/^\s+//;        # strip leading whitespce
    $cmd =~ s/\s+$//;        # strip trailing whtspce
    $debug_msg = $cmd;

    my $required_options = $cmd_hash->{required_options};
    if (ref($required_options) =~ /^HASH{/)
    {
        $required_options = "__NONE__";
    }
    $required_options =~ s/^\s+//;        # strip leading whitespce
    $required_options =~ s/\s+$//;        # strip trailing whtspce

    my $optional_options = $cmd_hash->{optional_options};
    if (ref($optional_options) =~ /^HASH{/)
    {
        $optional_options = "__NONE__";
    }
    $optional_options =~ s/^\s+//;        # strip leading whitespce
    $optional_options =~ s/\s+$//;        # strip trailing whtspce

    my $option_order = $cmd_hash->{option_order};
    if (ref($option_order) =~ /^HASH{/)
    {
        $option_order = "__NONE__";
    }
    $option_order =~ s/^\s+//;        # strip leading whitespce
    $option_order =~ s/\s+$//;        # strip trailing whtspce

    my $options_hash = $cmd_hash->{options};

    #
    # Let's make sure we have something for all the required options
    #
    if ($require_options)
    {
        foreach my $option (split " ", $required_options)
        {
            last if ($required_options eq "__NONE__");

            $option =~ s/.ARRAY$//;

            unless (defined($cmd_args->{$option}))
            {
                $self->_set_err("CmdsDriver::_get_zapi_element(): The " .
                                "$full_cmd_name command requires "   .
                                "the argument \"$option\"");
                return ($self);
            }
        }
    }

    #
    # Now that we know we have all that is required, let's create the
    # NaElement object
    #
    my $element_obj = new NaElement($cmd);

    #
    # for Cmode and command is get-iter, build the query with sub elements
    #
    if ($self->filer_mode() =~ /Cmode/i and $cmd =~ /(.*)\-(get|modify)\-iter/) {

        my $object = $1;
        my $subcmd = $2;
        my $q_option = "query";
        my $a_option = "attributes";
        my $cmd_attr = "";

        # for volume-get-iter, it used 'name' instead of 'volume'
        if ($object =~ /^volume$/ and exists $cmd_args->{$object}) {
            $cmd_args->{name} = delete $cmd_args->{$object};
        }

        #
        # create new cmd_args input structure
        #
        foreach my $arg (keys %{$cmd_args})
        {
            #
            # ignore cmd and fields args. we will use field to parse from output
            #
            next if ($arg =~ /^(cmd|fields)$/i);

            # If the user passed in max-records for a zapi call, we need to
            # grab it here so it does not become part of the 'query' element.
            # We'll instantiate max-records as a new zapi argument, delete it
            # from the args list, and then go to the next arg.
            if ($arg =~ /^max-records$/) {
                $element_obj->child_add(new NaElement($arg,$cmd_args->{$arg}));
                delete $cmd_args->{$arg};
                next;
            }
            #
            # handle modify-iter command
            # not happy with this solution.
            # will have to think for a long term solution
            #
            if ($subcmd eq "modify" and $arg !~ /^($object|name)$/) {
                $cmd_attr = $object . "-attributes";
                if (ref($arg) =~ /HASH/) {
                    foreach (keys %{$arg}) {
                        $cmd_args->{$a_option}->{$cmd_attr}->{$_}->{$arg->{$_}} =
                            delete $cmd_args->{$_}->{$arg->{$_}};
                    }
                } else {
                    $cmd_args->{$a_option}->{$cmd_attr}->{$arg} =
                         delete $cmd_args->{$arg};
                }
                next;
            }
            #
            # set cmd sub attribute
            # we have to do this because ZAPI attributes are not consistent
            # some arguments are nested within a query tag. As a result, we use
            # the TranslationTable with specific nested query for specific cmd
            # see CmdsDriver/TranslationTable.pm for more info
            #
            if (exists $mapped_cmds{$cdef_cmd}{$q_option}) {
                my $mapped_arg = $mapped_cmds{$cdef_cmd}{$q_option};

                if (ref($mapped_arg) =~ /HASH/) {
                    foreach my $key (keys %{$mapped_arg}) {
                        # build nested arguments from hash
                        $cmd_args->{$q_option}->{$key}->{$mapped_arg->{$key}}->{$arg}
                            = delete $cmd_args->{$arg};
                    }
                } else {
                   $cmd_args->{$q_option}->{$mapped_arg}->{$arg} = delete $cmd_args->{$arg};
                }
            } else {
                $self->_set_err("CmdsDriver::_get_zapi_element(): Missing query " .
                                "of \"$cdef_cmd\" from TranslationTable.pm ");
                return ($self);
            }
        }

        #
        # add attribute element and its sub-elements for modify-iter
        #
        if (exists $cmd_args->{$a_option}) {
            my $attr_element = $self->_expand_complex_option($a_option,
                                                             $cmd_args->{$a_option},
                                                             $cmd_args->{$a_option});
            $element_obj->child_add($attr_element);
        }

        #
        # add query element and its sub-elements
        #
        if (exists $cmd_args->{$q_option}) {
            my $option_element = $self->_expand_complex_option($q_option,
                                                               $cmd_args->{$q_option},
                                                               $cmd_args->{$q_option});

            $element_obj->child_add($option_element);
            $debug_msg .= " (complex): " . Dumper($cmd_args);
        }
    }

    #
    # For debug purposes, dump out the options_hash from the ZAPI XML file
    # and the argument hash from the user
    #
    my $dumpOut = Dumper($options_hash) if ($self->get_debug_level() >= 9);;
    $self->sanlog(-msg       => "_get_zapi_element() Options Hash Output " .
                                "Dumper output is:\n$dumpOut",
                  -msg_level => 9);
    $dumpOut = Dumper($cmd_args) if ($self->get_debug_level() >= 9);;
    $self->sanlog(-msg       => "_get_zapi_element() Cmd Args Hash Output " .
                                "Dumper output is:\n$dumpOut",
                  -msg_level => 9);

    #
    # Add to the element_obj for each option given
    #
    foreach my $option (split " ", $option_order)
    {
        last if ($option_order eq "__NONE__");
        $dumpOut = Dumper($cmd_args) if ($self->get_debug_level() >= 9);
        $self->sanlog(-msg => "_get_zapi_element() checking option $option in: \n$dumpOut",
                      -msg_level => 9);
        next unless (defined $cmd_args->{$option}); # skip unless option given

        my $arg_value    = $cmd_args->{$option};
        my $option_value = $options_hash->{$option};
        $self->sanlog(-msg => "_get_zapi_element() did not skip option " .
                             "$option ($arg_value) ($option_value)",
                      -msg_level => 9);

        my $arg_ref    = ref($arg_value);
        my $option_ref = ref($option_value);

        $self->sanlog(-msg => "_get_zapi_element() arg_ref = $arg_ref " .
                              "option_ref = $option_ref",
                      -msg_level => 9);

        #
        # At this point, we are going to check the option and see which of
        # 3 types we support we are looking at: HASH/ARRAY,
        # simple ARRAY (example <args>arg.ARRAY</args>), simple type
        #
        if ($option_ref)
        {
            #
            # In this case we have a 'complex' object (HASH or ARRAY) and
            # we will expand this option.
            #
            $self->sanlog(-msg => "_get_zapi_element() option $option is " .
                                  "not simple",
                          -msg_level => 9);

            #
            # Since _expand_complex option is recursive, we reset the depth
            # here.  The depth is used for debug purposes and to prevent
            # endless loops.
            #
            $self->{depth}     = 0;

            #
            # crap! it seems _expand_complex_option() doesn't work if option_value
            # and arg_value are not the same. So what's the purpose? need to ask
            # Pressley. workaround for Cmode is set option_value = arg_value
            #
            if ($self->filer_mode() =~ /Cmode/i) {
                $option_value = $arg_value;
            }

            #
            # Get and element from the complex option.  Note that this element
            # will contain 1 or move sub-elements.
            #
            my $option_element = $self->_expand_complex_option($option,
                                                               $option_value,
                                                               $arg_value);

           $debug_msg .= " $option (complex)";
           $element_obj->child_add($option_element);
        }
        elsif ($option_value =~ /.*ARRAY$/)
        {
            #
            # Simple array looks like this: <args>arg.ARRAY</args>.  In this
            # case the arguement from the user should be and array and we
            # will add a  sub element ("arg" "arg_value" for example) for
            # each array member
            #
            my $option_element = $self->_expand_array_option(
                                                      $option,
                                                      $option_value,
                                                      $arg_value
                                                      );
            if ($self->error())
            {
                my $errmsg = join "\n", $self->errmsg();
                $self->_set_err("CmdsDriver::_get_zapi_element(): The "  .
                                "$full_cmd_name failed to expand a " .
                                " simple array type: $errmsg");
                return ($self);
            }

            $debug_msg .= " $option (array)";
            $element_obj->child_add($option_element);
        }
        else
        {
            #
            # Simple type, just add the child
            #
            my $switch ="";
            if (defined($options_hash->{$option}))
            {
                $switch = $option;
            }
            if (defined($cmd_args->{$option}))
            {
                my $value = $cmd_args->{$option};
                $debug_msg .= " $switch = $value";
                $element_obj->child_add(new NaElement($switch,$value));
            }
        }
    } # end for each option_order

    #
    # If we were given a pass_thru, add it to the element
    #
    if (length($pass_thru))
    {
        my @pass_thru_elements = split(/,/, $pass_thru);
        foreach my $pt_element (@pass_thru_elements)
        {
            my ($switch, $value) = split("=", $pt_element);
            $debug_msg .= " $switch = $value";
            $element_obj->child_add(new NaElement($switch,$value));
        }
    }

    $self->sanlog(-msg => "Executing ZAPI $debug_msg",
                  -use_loghdl => 0,
                  -msg_level  => 1);
    return $element_obj;
} # end _get_zapi_element

###############################################################################
# Method:    _expand_array_option()
# Objective: Options that are not simple name-value pairs have to be built
#            up as a tree.  One non-simple option is a simple array like
#            <args>arg.ARRAY</arg>.  This function will build the element
#            for an array
# Inputs:    option         - array option ("args" in example above)
#            name           - name of data ("arg" in example above)
#            arg_value      - the array of data
# Outputs:   new NaElement
# Notes:
###############################################################################
sub _expand_array_option
{
    my($self, $option, $name, $arg_value) = @_;

    $name =~ s/\.ARRAY//;  # strip off the ARRAY marker
    my $arg_ref    = ref($arg_value);
    unless ($arg_ref =~ /ARRAY/)
    {
        $self->_set_err("CmdsDriver::_expand_array_option(): Expected " .
                        "an ARRAY");
        return ($self);
    }
    my $option_element = new NaElement($option);
    $self->sanlog(-msg => "_expand_array_option() option ($option) " .
                          "is simple ARRAY",
                  -msg_level => 9);
    foreach my $data (@{$arg_value})
    {
        $option_element->child_add(new NaElement($name,$data));
        $self->sanlog(-msg => "_expand_array_option() add simple array " .
                              "element $name $data",
                      -msg_level => 9);
    } # end foreach arg_value
    return $option_element;
} # end _expand_array_option

###############################################################################
# Method:    _expand_complex_option()
# Objective: Options that are not simple name-value pairs have to be built
#            up as a tree.  To this we take the option from the XML and the
#            matching argument from the user input and build a complex element
# Inputs:    complex_option - option from the XML
#            complex_data   - argument data of matching option
# Outputs:   new NaElement
# Notes:
###############################################################################
sub _expand_complex_option
{
    my($self, $option, $option_value, $arg_value) = @_;

    if ($self->{depth} > 10)
    {
        $self->_set_err("CmdsDriver::_expand_complex_option() " .
                        "aborting option parsing, max depth exceeded");
        return $self;
    }

    my $arg_option = $option;
    $arg_option =~ s/\.ARRAY//; # strip off the .ARRAY marker



    my $arg_ref    = ref($arg_value);
    my $option_ref = ref($option_value);

    $self->sanlog(-msg => "_expand_complex_option() expanding option " .
                          "$option ($arg_option) depth is ".
                          $self->{depth} .
                          " arg_ref = $arg_ref option_ref = $option_ref",
                  -msg_level => 9);

    #
    # There are 4 things we support here:
    #     A Hash, an ARRAY, text with .ARRAY or simple text.
    #
    if ($arg_ref)
    {
        #
        # This should mean we have an HASH or an ARRAY
        #
        if ($arg_ref =~ /HASH/)
        {
            $self->sanlog(-msg => "_expand_complex_option() option $option " .
                                  "is a HASH ($arg_option)",
                          -msg_level => 9);
            my $dumpOut = Dumper($option_value) if ($self->get_debug_level() >= 9);
            $self->sanlog(-msg       => "_expand_complex_option() Options " .
                                        "Value Dumper output is:\n$dumpOut",
                          -msg_level => 9) ;
            $dumpOut = Dumper($arg_value) if ($self->get_debug_level() >= 9);
            $self->sanlog(-msg       => "_expand_complex_option() Arg " .
                                        "Value Dumper output is:\n$dumpOut",
                          -msg_level => 9);

            #
            # Create the top level element
            #
            my $option_element = new NaElement($arg_option);

            #
            # For each hash element given, create a sub element and add
            # that sub element to the option_element we will return
            #
            foreach my $option_key (keys(%{$option_value}))
            {

                my $arg_key = $option_key;
                $arg_key =~ s/\.ARRAY//;

                $self->sanlog(-msg => "_expand_complex_option() expanding " .
                                "option $option ($arg_option) key " .
                                "$option_key ($arg_key) $arg_value->{$arg_key}",
                              -msg_level => 9);

                next unless (defined($arg_value->{$arg_key}));

                my $complex_option = $option_value->{$option_key};
                my $complex_arg    = $arg_value->{$arg_key};

                $self->sanlog(-msg => "_expand_complex_option() " .
                                      "didn't skip $arg_key ($option_key)" .
                                      "recursively calling function",
                              -msg_level => 9);
                $self->{depth}++;
                my $sub_element = $self->_expand_complex_option($option_key,
                                                                $complex_option,
                                                                $complex_arg);
                $self->{depth}--;
                $self->sanlog(-msg       => "_expand_complex_option() " .
                  "after recursive call, ref option_element is "        .
                  ref($option_element)             .
                  "ref sub_element is "            .
                  ref($sub_element),
                  -msg_level => 9);

                if ($self->error())
                {
                    my $errmsg = join "\n", $self->errmsg();
                    $self->_set_err(
                       "CmdsDriver::_expand_complex_option(): Failed " .
                       "to expand option $option: $errmsg");
                    return ($self);
                }
                if (ref($sub_element) =~ /ARRAY/)
                {
                    $self->sanlog(-msg => "_expand_complex_option() adding " .
                                "an element for each element returned",
                              -msg_level => 9);
                    foreach my $element (@{$sub_element})
                    {
                        $option_element->child_add($element);
                    }
                }
                else
                {
                    $self->sanlog(-msg => "_expand_complex_option() adding " .
                                          "a single element",
                                  -msg_level => 9);

                    $option_element->child_add($sub_element);
                }
            } # end foreach option
            return($option_element);
        } # end if hash
        elsif ($arg_ref  =~ /ARRAY/)
        {
            #
            # If we get here we have an arg_value that's an array.  The
            # option value, will not be an array (it'll have a key that
            # thas .ARRAY in it, however).
            #
            $self->sanlog(-msg => "_expand_complex_option() expanding array " .
                                  "option $option is an ARRAY",
                          -msg_level => 9);
            my $dumpOut = Dumper($option_value) if ($self->get_debug_level() >= 9);
            $self->sanlog(-msg => "_expand_complex_option() " .
              "Options Value Dumper output is:\n$dumpOut",
                          -msg_level => 9);
            $dumpOut = Dumper($arg_value) if ($self->get_debug_level() >= 9);
            $self->sanlog(-msg => "_expand_complex_option() " .
              "Arg Value Dumper output is:\n$dumpOut",
                          -msg_level => 9);

            #
            # Create an array to hold the elements
            #
            my @element_array;

            #
            # For each element in the array, create a sub element and add
            # that sub element to the option_element we will return
            #
            foreach my $array_value (@{$arg_value})
            {
                $self->sanlog(-msg => "_expand_complex_option() " .
                  "expanding option $option ($arg_option)" .
                  "array value is $array_value ref is " . ref($array_value),
                              -msg_level => 9);
                #
                # As far as I know, all instances of this are a HASH.
                # What we need to do is create an element for the current
                # arg_option.  Then add each hash key to that element by
                # recursing back into _expand_complex_option.  Once we
                # have processed all hash keys we push the array_element
                # on the element_array stack.  We will return the stack
                # and we will need to go to both places that recurse and
                # check for an ARRAY ref and add each element in the array
                # to the tree
                #
                my $array_value_ref = ref($array_value);
                unless ($array_value_ref =~ /HASH/)
                {
                    $self->_set_err("CmdsDriver::_expand_complex_option(): " .
                                    "Failed to expand ARRAY $arg_option. "   .
                                    "Only a HASH reference is supported.");
                    return ($self);
                }

                #
                # Create an element representing the array
                #
                my $array_element = new NaElement($arg_option);
                $self->sanlog(-msg => "_expand_complex_option() " .
                  "Created element for $arg_option ref is" .
                  ref($array_element),
                              -msg_level => 3);

                #
                # Expand each array element
                #
                foreach my $array_key (keys %{$array_value})
                {
                    my $hash_value = $array_value->{$array_key};
                    my $option_hash_key = $array_key;
                    unless (defined ($option_value->{$array_key}))
                    {
                        $option_hash_key = $array_key . ".ARRAY";
                    }
                    my $option_hash_value = $option_value->{$option_hash_key};
                    $self->sanlog(-msg => "_expand_complex_option() " .
                                      "array_key is $array_key " .
                                      "hash value is $hash_value ref is "  .
                                      ref($hash_value)                     .
                                      " option_hash_key is $option_hash_key" .
                                      " array_element ref is " .
                                      ref($array_element),
                              -msg_level => 9);
                    $self->{depth}++;
                    my $sub_element = $self->_expand_complex_option(
                                                     $array_key,
                                                     $option_hash_value,
                                                     $hash_value);
                    $self->{depth}--;

                    if ($self->error())
                    {
                        my $errmsg = join "\n", $self->errmsg();
                        $self->_set_err(
                          "CmdsDriver::_expand_complex_option(): Failed " .
                          "to expand option $option: $errmsg");
                        return ($self);
                    }
                    if (ref($sub_element =~ /ARRAY/))
                    {
                        foreach my $element (@{$sub_element})
                        {
                            $array_element->child_add($element);
                        }
                    }
                    else
                    {
                        $array_element->child_add($sub_element);
                    }
                } # end for each hash key
                push @element_array, $array_element;
                $self->sanlog(-msg => "_expand_complex_option() " .
                    "pushing array_element on element_array" .
                    " ref a_e is " . ref($array_element) .
                    " and size element_array is " . scalar(@element_array),
                              -msg_level => 9);
            } # end foreach array value

            if (scalar(@element_array) >1 )
            {
                $self->sanlog(-msg => "_expand_complex_option() " .
                    "returning element_array num element is"      .
                    scalar(@element_array)      .
                    " ref is "                  .
                    ref(\@element_array),
                    -msg_level => 9);
                return(\@element_array);
            }
            else
            {
                $self->sanlog(-msg => "_expand_complex_option() " .
                    "returning element_array[0] ref is " .
                    ref($element_array[0]),
                    -msg_level => 9);

                return ($element_array[0]);
            }
        } # end if ARRAY
        else
        {
            $self->_set_err("CmdsDriver::_expand_complex_option(): " .
                            "unsupported type for option $option");
                return ($self);
        }
    } # end if arg_ref
    elsif ($option_value =~ /.*ARRAY$/)
    {
        my $option_element = $self->_expand_array_option(
                                                  $option,
                                                  $option_value,
                                                  $arg_value
                                                  );
        if ($self->error())
        {
            my $errmsg = join "\n", $self->errmsg();
            $self->_set_err("CmdsDriver::_expand_complex_option(): Failed " .
                            "to expand option $option: $errmsg");
            return ($self);
        }

        return($option_element);
    }
    else
    {
        #
        # Simple type, just add the child
        #
        my $switch = $option;
        if (defined($arg_value))
        {
            my $value = $arg_value;
            $self->sanlog(-msg => "_expand_complex_option() " .
                    "Adding simple element switch is $switch value is $value",
                              -msg_level => 9);
            return(new NaElement($switch,$value));
        }
    }
    $self->_set_err("CmdsDriver::_expand_complex_option(): Failed " .
                    "to expand option $option reached unexpected end");
    return ($self);
} # end _expand_complex_option

###############################################################################
# Method:    _get_expected_output()
# Objective: Given a command name and a hash with data about that cmd,
#            find the expected output
# Inputs:    -cmd_name - hash key with the command name
#            -cmd_hash - hash key with the cmd information hash
# Outputs:   expected output and output type
# Notes:     Defaults for expected output and output type are "*" and "regex"
###############################################################################
sub _get_expected_output
{
    my($self, %args) = @_;
    my $full_cmd_name   = undef;
    my $cmd_hash        = undef;
    my $expected_output = "";
    my $output_type     = "regex";
    my $debug_msg       = "";

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?cmd_name$/) {
            length($full_cmd_name = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::_get_expected_output() " .
                                "invalid value for \"cmd_name\": "    .
                                $args{$_});
                return $self;
            };
        }
        elsif (/-?cmd_hash$/) {
            length($cmd_hash = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::_get_expected_output() " .
                                "invalid value for \"cmd_hash\": "    .
                                $args{$_});
                return $self;
            };
        }
        else {
            $self->_set_err("CmdsDriver::_get_expected_output() " .
                            "Invalid parameter specified: $_");
            return $self;
        }
    } # end foreach arg

    #
    # Check for required arguments
    #
    unless (defined($cmd_hash))
    {
        $self->_set_err("CmdsDriver::_get_expected_output(): " .
                        "missing required argument \"cmd_hash\"");
        return ($self);
    }

    unless (defined($full_cmd_name))
    {
        $self->_set_err("CmdsDriver::_get_expected_output(): " .
                        "missing required argument \"cmd_name\"");
        return ($self);
    }

    #
    # Let's pull some information out of the command hash
    #
    my $cmd = $cmd_hash->{command};
    $cmd =~ s/^\s+//;        # strip leading whitespce
    $cmd =~ s/\s+$//;        # strip trailing whtspce
    $debug_msg = $cmd;

    my $eo_hash = $cmd_hash->{expected_output};

    #
    # Print the expected outputs hash
    #
    my $dumpOut = Dumper($eo_hash) if ($self->get_debug_level() >= 9);
    $self->sanlog(-msg       => "Expected Output Dumper output is:\n$dumpOut",
                  -msg_level => 9);

    #
    # Let's make sure we have something for all the required options
    #
    if (defined($eo_hash->{type}))
    {
        $output_type = $eo_hash->{type};
        $output_type =~ s/^\s+//;        # strip leading whitespce
        $output_type =~ s/\s+$//;        # strip trailing whtspce
    }
    if (defined($eo_hash->{data}))
    {
        $expected_output = $eo_hash->{data};
        $expected_output =~ s/^\s+//;        # strip leading whitespce
        $expected_output =~ s/\s+$//;        # strip trailing whtspce
        $expected_output = "" if ($expected_output eq "__NONE__");
    }
    return ($expected_output, $output_type);

} # end _get_expected_output

###############################################################################
# Method:    _get_expected_result()
# Objective: Given a command name and a hash with data about that cmd,
#            find the expected result
# Inputs:    -cmd_name - hash key with the command name
#            -cmd_hash - hash key with the cmd information hash
# Outputs:   expected result
# Notes:     Default for expected result is 0
###############################################################################
sub _get_expected_result
{
    my($self, %args) = @_;
    my $full_cmd_name   = undef;
    my $cmd_hash        = undef;
    my $expected_result = 0;
    my $debug_msg       = "";

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?cmd_name$/) {
            length($full_cmd_name = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::_get_expected_result() " .
                                "invalid value for \"cmd_name\": "    .
                                $args{$_});
                return $self;
            };
        }
        elsif (/-?cmd_hash$/) {
            length($cmd_hash = $args{$_}) or
            do {
                $self->_set_err("CmdsDriver::_get_expected_result() " .
                                "invalid value for \"cmd_hash\": "    .
                                $args{$_});
                return $self;
            };
        }
        else {
            $self->_set_err("CmdsDriver::_get_expected_result() " .
                            "Invalid parameter specified: $_");
            return $self;
        }
    } # end foreach arg

    #
    # Check for required arguments
    #
    unless (defined($cmd_hash))
    {
        $self->_set_err("CmdsDriver::_get_expected_result(): " .
                        "missing required argument \"cmd_hash\"");
        return ($self);
    }

    unless (defined($full_cmd_name))
    {
        $self->_set_err("CmdsDriver::_get_expected_result(): " .
                        "missing required argument \"cmd_name\"");
        return ($self);
    }

    #
    # Let's pull some information out of the command hash
    #
    my $cmd = $cmd_hash->{command};
    $cmd =~ s/^\s+//;        # strip leading whitespce
    $cmd =~ s/\s+$//;        # strip trailing whtspce
    $debug_msg = $cmd;

    if (defined($cmd_hash->{expected_result}))
    {
        $expected_result = $cmd_hash->{expected_result};
        $expected_result =~ s/^\s+//;        # strip leading whitespce
        $expected_result =~ s/\s+$//;        # strip trailing whtspce

        $self->sanlog(-msg => "Expected Result found is $expected_result",
                      -msg_level => 9);
    }
    else
    {
        $self->sanlog(-msg => "Expected Result NOT found using default, 0",
                      -msg_level => 9);
    }

    return ($expected_result);

} # end _get_expected_result

###############################################################################
# Method:    _xml_to_hash()
# Objective: Convert an XML string into a hash and return the hash
# Inputs:    string of XML to be converted
# Outputs:   hash created
# Notes:     You should put an eval around this when you call it
###############################################################################
sub _xml_to_hash
{
    my ($self, $output) = @_;

    my $XML_obj = new XML::Simple;
    my $data = $XML_obj->XMLin($output);

    return $data;
} # end _xml_to_hash

###############################################################################
# Method:    _populate_commands()
# Objective: Read the data files to create the commands hash
# Inputs:    None
# Outputs:   1 - success  0 - failure
#            On failure the error flag and message will be set
# Notes:
###############################################################################
sub _populate_commands
{
    my ($self, $load_str) = @_;

    my @l_cmds = split(/,\s*/, $load_str);
    my @cmds_to_load = ();
    my @temp = ();
    my $found = 0;
    my @current_load_list = ();

    #
    # For 7mode, load XML files
    #
    if ($self->filer_mode() eq '7mode') {
        #
        # Add the commands to load to the list
        #
        foreach my $c_name (@l_cmds)
        {
            #
            # If neither interface is specified, add
            # them both
            #
            if($c_name !~ /cli/ and $c_name !~ /zapi/)
            {
                push(@temp, "zapi_" . $c_name . ".xml");
                push(@temp, "cli_" . $c_name . ".xml");
            }
            else
            {
                push(@temp, $c_name . ".xml");
            }
        }

        $self->sanlog(-msg => "CmdsDriver::_populate_commands(): Loading files: " .
                              "@temp",
                      -msg_level => 9);

        #
        # Make sure we don't try to load a cmd that
        # has already been loaded
        #
        foreach my $c_name (@temp)
        {
            $found = 0;
            #
            # Check cmds_loaded for the cmd to be
            # added
            #
            foreach my $lc_name (@{$self->{cmds_loaded}})
            {
                  if($c_name eq $lc_name)
                  {
                      $found = 1;
                  }
            }

            #
            # If the cmd hasn't already been added
            # add it to the list of cmds to load
            #
            unless($found == 1)
            {
                push(@cmds_to_load, $c_name);
            }
        }

        #
        # Find the test/lib/CmdsDriver directory
        #
        my $base_dir = "CmdsDriver";
        if (! -d $base_dir )
        {
            #
            # Check NATE_LIB for the existence of it.
            #
            my @tharnlibs = split(/;/, $ENV{'NATE_LIB'});
            if (not scalar(@tharnlibs)) {
                @tharnlibs = split(/;/, $ENV{'NATE_LIB_UNIX'});
            }
            my $foundfile = 0;
            foreach my $tharnlib (@tharnlibs)
            {
                if ( -d "${tharnlib}/$base_dir")
                {
                    $base_dir = "${tharnlib}/$base_dir";
                    $foundfile = 1;
                    last;
                }
                elsif (-d "${tharnlib}/lib/$base_dir")
                {
                    $base_dir = "${tharnlib}/lib/$base_dir";
                    $foundfile = 1;
                    last;
                }
            } # end foreach tharnlib
            if (! $foundfile)
            {
                $self->sanlog(-msg =>
                      "The directory $base_dir doesn't exist in " .
                      "current directory or NATE_LIB: $ENV{'NATE_LIB'}",
                              -msg_level => 3);
            }
        }

        #
        # Create a list of files.
        #
        opendir(BASE_DIR, $base_dir); # dave check for error
        foreach my $dir (readdir(BASE_DIR))
        {
            next unless (-d "$base_dir/$dir"); # is this a sub directory?
            next if $dir =~ /^\./;             # ignore "dot" files
            next unless $dir =~ /^ONTAP_/;     # ignore without ONTAP_ prefix
            opendir(ONTAP_DIR, "$base_dir/$dir"); # dave check for error
            foreach my $file (readdir(ONTAP_DIR))
            {
                next unless $file =~ /.*\.xml$/; # only read in xml files

                foreach my $load_f (@cmds_to_load)
                {
                    if($file =~ /$load_f/)
                    {
                        push @{$self->{data_files}}, "$base_dir/$dir/$file";
                        push(@current_load_list, "$base_dir/$dir/$file");
                        $self->sanlog(-msg => "Adding data file $base_dir/$dir/$file",
                                      -msg_level => 9);
                    }
                }
            } # end for each file
        } # end for each dir

        #
        # Load the data from each command file (xml to hash)
        #
        foreach my $file_name (@current_load_list)
        {
            my $data = $self->_load_data_file($file_name);
            my $command = $file_name;
            $command =~ s/^CmdsDriver\///; # remove dir prefix
            $command =~ s/.xml//;          # remove file type suffix
            my @f_arr = split("/", $command);
            my $c = $f_arr[-1];
            push(@{$self->{cmds_loaded}}, $c);  # add it to list of added cmds
            $self->{commands}{$command} = $data;
        } # end foreach command

        #
        # Print the commands hash
        #
        my $dumpOut = Dumper($self->{commands}) if ($self->get_debug_level() >= 11);
        $self->sanlog(-msg       => "Dumper output is:\n$dumpOut",
                      -msg_level => 11);

    } else {  # For Cmode, create schema object and load cdefs files
        #
        # get schema obj. create one if it isn't exist.
        # store in CmdsDriver->{cmode_schema}
        #
        $self->{cmode_schema} = $self->get_schema_obj();
        if ($self->error()) {
            $self->_set_err("Failed to get cdef schema object",
                            $self->errmsg());
            return $SanUtil::FALSE;
        }

        #
        # load cdef if it has not loaded
        #
        unless ($self->{cmode_schema}->is_loaded(-filer => $self->filer())) {
            #
            # get ONTAP build information
            #
            unless ($self->get_shell_environment() =~ /ngsh/) {
                $self->set_shell_environment(-type => "ngsh");
            }

            # get build info if it doesn't exist
            $self->_get_build_info();
            if ($self->error()) {
                $self->_set_err("Failed to get build info",
                                 $self->errmsg());
                return $SanUtil::FALSE;
            }

            #
            # get the path that contained cdef files
            #
            my $cdef_path = $self->_get_cdef_path();
            if ($self->error()) {
                $self->_set_err("failed to get cdef path",
                                $self->errmsg());
                return $SanUtil::FALSE;
            }
            #
            # load cdef files using path and branch where they're resided
            # for more info, review San::Schema.pm
            #
            $self->{cmode_schema}->load(-path   => $cdef_path,
                                        -filer  => $self->filer(),
                                        -branch => $self->get("branch"));
            if ($self->{cmode_schema}->error()) {
                $self->_set_err("failed to load cdef schema",
                                $self->{cmode_schema}->errmsg());
                return $SanUtil::FALSE;
            }
        }
        #
        # get absolute path and package name
        #
        my $pkg = __PACKAGE__;
        $pkg =~ s|\:\:|\/|g;
        my $translate_table = $pkg . "/TranslationTable.pm";
        #
        # load translation table
        #
        require $translate_table;
    }

    return $SanUtil::TRUE;
} # end _populate_commands

###############################################################################
# Method:    _load_data_file()
# Objective: Create a hash of hash to represent the date in a command xml file
# Inputs:    file_path - name of xml file to load
# Outputs:   1 - success  0 - failure
#            On failure the error flag and message will be set
# Notes:     A new key will be added to the commands hash this key will be a
#            hash representing the data in the file
###############################################################################
sub _load_data_file
{
    my ($self, $file_path) = @_;

    $self->_clear_err();

    # my $file_name = "CmdsDriver/" . $command . ".xml";
    # my $file_path = $file_name;

    if (! -f $file_path )
    {
        #
        # Check NATE_LIB for the existence of it.
        my @tharnlibs = split(/;/, $ENV{'NATE_LIB'});
        if (not scalar(@tharnlibs)) {
            @tharnlibs = split(/;/, $ENV{'NATE_LIB_UNIX'});
        }
        my $foundfile = 0;
        foreach my $tharnlib (@tharnlibs)
        {
            if ( -f "${tharnlib}/$file_path")
            {
                $file_path = "${tharnlib}/$file_path";
                $foundfile = 1;
                last;
            }
        } # end foreach tharnlib
        if (! $foundfile)
        {
            $self->sanlog(-msg =>
                  "The file $file_path doesn't exist in " .
                  "current directory or NATE_LIB: $ENV{'NATE_LIB'}" );
        }
    }
    #
    # By this point we've either located the file or died
    #
    $self->sanlog(-msg => "Loading file $file_path",
                  -msg_level => 3);

    #
    # Create a hash and dump the data
    #

    #
    # create object
    #
    my $xmlObj = new XML::Simple;

    #
    # read XML file
    #
    my $data = $xmlObj->XMLin($file_path);

    return $data;
} # end _load_data_file


###############################################################################
# Method:    _find_cmd_hash()
# Objective: Given a command key and a version, find the appropriate hash
#            that defines the command
# Inputs:    cmd_key       - interface _ command
#            filer_version - which filer version we are using
# Outputs:   cmd_hash reference
#            On failure the error flag and message will be set
# Notes:     Algorithm:
#             -  set the current_version to 0
#             -  foreach command found
#                 - if ($cmd_version <= $filer_version AND
#                       $cmd_version > $current_version)
#                     -  $current_version = $cmd_version
#                     -  $cmd_hash = hash entry for current command
###############################################################################
sub _find_cmd_hash
{
    my ($self, $cmd_key, $filer_version) = @_;
    my $current_version = -1;
    my $cmd_hash        = undef;
    my $current_cmd     = undef;
    my $loaded = 0;

    #
    # Check if the command file is loaded
    #
    my ($temp) = split("-", $cmd_key);
    foreach my $loaded_cmd (@{$self->{cmds_loaded}})
    {
        $self->sanlog(-msg => "CmdsDriver::_find_cmd_hash(): Check file " .
                          "$temp vs $loaded_cmd and load file if different",
                     -msg_level => 9);

        if("$temp" eq "$loaded_cmd")
        {
            $self->sanlog(-msg => "CmdsDriver::_find_cmd_hash(): We think " .
                          "file $temp is the same as $loaded_cmd",
                  -msg_level => 7);
            $loaded = 1;
        }
    }

    #
    # If the command hasn't been loaded yet
    # repopulate command list
    #
    unless($loaded == 1)
    {
        $self->_populate_commands($temp);
    }

    $filer_version = $self->_format_version($filer_version);

    $self->sanlog(-msg => "Finding command with key $cmd_key and " .
                          "version $filer_version",
                  -msg_level => 7);
    #
    # Look at each file represented in the cmds hash
    #
    foreach my $file (keys %{$self->{commands}})
    {
        #
        # Within the file, look at each command and find a match
        # for the command key. We expect the file to have a
        # "_version" suffix
        #
        foreach my $cmd (keys %{$self->{commands}->{$file}})
        {
            if ($cmd =~ /^$cmd_key(_)(.*)$/)
            {
                my $version = $self->_format_version($2);
                if (($version <= $filer_version) and
                    ($version > $current_version))
                {
                    $current_version = $version;
                    $current_cmd     = $cmd;
                    $cmd_hash        = $self->{commands}->{$file}->{$cmd};
                }
            }
        } # end foreach cmd
    } # end foreach file
    return ($current_cmd, $cmd_hash);
} # end _find_cmd_hash

###############################################################################
# Method:    _find_cmode_hash()
# Objective: Given a command key and a version, find the appropriate hash
#            that defines the command
# Inputs:    cmd_key       - interface _ command
#            filer_version - which filer version we are using
# Outputs:   cmd_hash reference
#            On failure the error flag and message will be set
# Notes:     Algorithm:
#             -  set the current_version to 0
#             -  foreach command found
#                 - if ($cmd_version <= $filer_version AND
#                       $cmd_version > $current_version)
#                     -  $current_version = $cmd_version
#                     -  $cmd_hash = hash entry for current command
###############################################################################
sub _find_cmode_hash
{
    my ($self, $cmd_key, $filer_version, $cmd_args) = @_;
    my $current_version = -1;
    my $current_cmd     = undef;
    my $current_cmd2    = undef;
    my $cmd_hash        = {};
    my $cdef_obj        = {};
    my $interface       = $self->interface();
    my $schema_obj;

    # strip the first 3 or 4 chars and the underscore "_" in $cmd_key
    if ($cmd_key =~ /^(...?.)\_(.*)/)
    {
        $interface   = $1;
        $current_cmd = $2;
    }

    # obtain cdef schema object
    if (defined($self->{cmode_schema}) and
        ref($self->{cmode_schema}) =~ /Schema/) {
        $schema_obj = $self->{cmode_schema};
    } else {
        $self->_set_err("Failed to find schema object");
        return ();
    }

    #
    # assuming cdefs files are already loaded
    #
    # we now retrieve cdef obj that matched cdef_cmd from the Cmode schema obj
    # @cdefs_objs - $schema_obj->get_cdef_objs();
    #

    # Several cli commands were renamed as a result of BURT 490103. In order to
    # allow for the renaming (at least temporarily), BURT 681168 adds a
    # secondary command to look for.
    $current_cmd2 = $current_cmd;
    if ($interface =~ /cli/i) {
        if ($current_cmd eq 'lun_mapped_show') {
            $current_cmd2 = 'lun_mapping_show';
        } elsif ($current_cmd eq 'lun_unmap') {
            $current_cmd2 = 'lun_mapping_delete';
        } elsif ($current_cmd eq 'lun_map') {
            $current_cmd2 = 'lun_mapping_create';
        }
    }

    my $branch = $schema_obj->{$self->get("filer")};
    if (exists $schema_obj->{$branch}{$current_cmd}) {
        $cdef_obj = $schema_obj->{$branch}{$current_cmd};
    } elsif (exists $schema_obj->{$branch}{$current_cmd2}) {
        $cdef_obj = $schema_obj->{$branch}{$current_cmd2};
        $current_cmd = $current_cmd2;
    }

    # if interface=cli, find the matched cdef object and return cdef_cmd and hash
    # if interface=zapi, use translation table to map cdef to zapi_cmd and hash
    # if cdef object does not exist and interface is cli, we fail
    if ($interface =~ /cli/i and ref($cdef_obj) !~ /Cdef/i) {
        $self->_set_err("Failed to find cdef object for " . $current_cmd);
        return ();
    }

    #
    # if cdef obj exists, create a list of elements from cdef object
    #
    if (ref($cdef_obj) =~ /Cdef/i) {
        foreach my $element (qw/action alias_params file optional_params required_params/) {
            $cmd_hash->{$element} = $cdef_obj->{$element};
        }
    } else {
        # this block is for zapi only commands. This means we don't
        # have cdef object and faking it in translation table

        # remove filer argument if exist
        delete $cmd_args->{filer} if (exists $cmd_args->{filer});

        # fill required_params and alias_params with pass-in args key
        foreach my $key (keys %{$cmd_args}) {
            push @{$cmd_hash->{required_params}}, $key;
            $cmd_hash->{alias_params}->{$key} = "-$key";
        }
    }

    # if interface is zapi, translate to zapi cmd and attributes using TranslationTable.pm
    if ($interface =~ /zapi/i) {
        #
        # Translate cdef cmd, hash, and args to zapi/7glike
        #
        ($current_cmd, $cmd_hash, $cmd_args) = $self->_translate_cdef(cmd_name  => $current_cmd,
                                                                      cmd_hash  => $cmd_hash,
                                                                      cmd_args  => $cmd_args,
                                                                      interface => $interface);
        if ($self->error()) {
            $self->_set_err("Failed to translate cdef cmd: " . $current_cmd,
                            $self->errmsg());
            return ();
        }

        return ($current_cmd, $cmd_hash, %{$cmd_args});
    }

    return ($current_cmd, $cmd_hash, %{$cmd_args});
} # end _find_cmode_hash

###############################################################################
# Method:    _format_version()
# Objective: Given a version, return it as an integer
# Inputs:    version - which filer version we are using
# Outputs:   version
###############################################################################
sub _format_version
{
    my ($self, $version) = @_;

    my @version_array = split(/\./, $version);
    my $digits = scalar(@version_array);
    my $supported_digits = 4;
    for (my $i = $digits ; $i < $supported_digits ; $i++ )
    {
        push @version_array, 0;
    }
    $version = join("", @version_array);
return $version;
    #
    # Remove the dots
    #
    $version =~ s/\.//g;

    #
    # Brain dead brute force for now.
    #
    if ((0 == $version) or
        ($version > 999))
    {
        return $version
    }

    if ($version > 99)
    {
        return ($version * 10);
    }
    elsif ($version > 9)
    {
        return ($version * 100);
    }
    else
    {
        return ($version * 1000);
    }
} # end _format_version

###############################################################################
# Method:    _check_output_xml()
# Objective: Verify the output from a ZAPI matches the XML expected output
# Inputs:    expected_output, output - strings containing XML
# Outputs:   1 - check passed, 0 - check failed
###############################################################################
sub _check_output_xml
{
    my ($self, $expected_output, $output) = @_;

    $self->sanlog(-msg => "NOT IMPLEMENTED: Checking Output for XML\n" .
                  "Expected Output:\n" . $expected_output .
                  "\nActual Output:\n" . $output,
                  -msg_level => 3);
    return 1;
} # end _check_output_xml

###############################################################################
# Method:    _discover_filer_mode()
# Objective: Check the connection to see if we are 7 or C mode
# Inputs:    Connection object
# Outputs:   Mode: "7Mode" or "CMode"
###############################################################################
sub _discover_filer_mode
{
    my ($self, $connection) = @_;

    #
    # C-mode connections will have an environment field.  7-mode will not.
    #
    return "Cmode" if ($connection->{environment});
    return "7mode";
} # end _discover_filer_mode

###############################################################################
# Method:    _mapped_args()
# Objective: Map cdef args into zapi args
# Inputs:    CDEF command
# Outputs:   HASHREF of ZAPI args
###############################################################################
sub _mapped_args
{
    my ($self, %args) = @_;
    my $ret_args = {};
    my $interface = delete $args{interface};

    #
    # verify mapped_cmd table is exist for the cmd name
    #
    if (exists ($mapped_cmds{$args{cmd_name}}{cmd}{$interface}))
    {
        #
        # translate cmd to zapi
        #
        $ret_args->{cmd} = $mapped_cmds{$args{cmd_name}}{cmd}{$interface};
    }
    else
    {
        #
        # no table. so we retain the cmd_name
        #
        $ret_args->{cmd} = $args{cmd_name};
    }

    #
    # translate options as well
    #
    foreach my $option (keys %args) {
        #
        # we already translated cmd_name to zapi so ignore
        #
        next if ($option eq 'cmd_name');

        # check for valid option in translation table
        if (exists $mapped_cmds{$args{cmd_name}}{$option}{$interface})
        {
            my $mapped_arg = $mapped_cmds{$args{cmd_name}}{$option}{$interface};

            # translate value if it exists in table
            my $value;
            if (exists $mapped_cmds{$args{cmd_name}}{$option}{$args{$option}}) {
                $value = $mapped_cmds{$args{cmd_name}}{$option}{$args{$option}};
            } else {
                $value = $args{$option};
            }
            # Assign the value to the correct key (attribute)
            $ret_args = $self->_assign_hash_ref_value('ret_args'=> $ret_args,
                                                      'mapped_arg'=> $mapped_arg,
                                                      'value' => $value
                                                    );
        } else {
            #
            # if option does not exist, retain old option
            #
            $ret_args->{$option} = $args{$option};
        }
    }

    return $ret_args;
} # end _mapped_args

###############################################################################
# Method:    _mapped_hash()
# Objective: Map cdef hash into zapi hash
# Inputs:    CDEF command
# Outputs:   HASHREF of ZAPI hash
###############################################################################
sub _mapped_hash
{
    my ($self, %args) = @_;
    my $ret_hash = {};
    my $interface = delete $args{interface};

    # we need to map action first
    unless (defined $args{cmd_name}) {
        $self->_set_err("Command/action does not exist.");
        return {};
    }

    # first, map comand from translation table based on interface
    # if it does not exist, retain cmd_name
    my $cmd = $mapped_cmds{$args{cmd_name}}{cmd}{$interface};
    if ($cmd eq "") {
        $ret_hash->{command} =  $args{cmd_name};
    } else {
        $ret_hash->{command} =  $cmd;
    }

    #
    # translate aliases, required, and optional params
    #
    foreach my $option (keys %args) {
        # translate aliases to options
        if ($option eq 'alias_params') {

            # translate options hash to zapi style
            foreach my $element (keys %{$args{$option}}) {
                # get cmd options base on cmd_interface from the translation table
                my $cmd_option = $mapped_cmds{$args{cmd_name}}{$element}{$interface};

                # option could be HASH, so we need to grab the first element
                if (ref($cmd_option) =~ /HASH/i) {
                    my @elems = ();
                    foreach my $key (keys %{$cmd_option}) {
                        $ret_hash->{options}->{$key} =  $cmd_option;
                        push @elems, $key;
                    }

                    # join them
                    $cmd_option = join(" ", @elems);
                    $ret_hash->{option_order} .= "$cmd_option ";
                } else {

                    # if cmd_option exists, map options from the table
                    # otherwise retain old options
                    my $new_ele = ($cmd_option eq "") ? $element : $cmd_option;

                    $ret_hash->{options}->{$new_ele} =  $new_ele;

                    # as a workaround, we need to add option_order because
                    # _get_zapi_element needs it.
                    $ret_hash->{option_order} .= "$new_ele ";
                }
            }
        } else {
            # translate required and optional params hash array to zapi style
            my @params = ();
            if ($option =~ /(required_params|optional_params)/i) {
                my $options_arrayref = $args{$option};
                # match params from TranslationTable.pm
                foreach my $element (@{$options_arrayref}) {
                    my $param = $mapped_cmds{$args{cmd_name}}{$element}{$interface};

                    # param could be HASH, so we need to grab the first element
                    if (ref($param) =~ /HASH/i) {
                        my @elems = ();
                        foreach my $key (keys %{$param}) {
                            push @elems, $key;
                        }

                        # join them
                        $param = join(" ", @elems);
                    }

                    # if param exists, map options from the table
                    # otherwise retain old param
                    my $new_param = ($param eq "") ? $element : $param;

                    push @params, $new_param;
                }

                # we need to set option_order so let's use required_param
                if ($option =~ /required_params/i) {
                    # Need to make sure that we dont add any duplicates.
                    # We are looking for the complete word so we need to check
                    # the beginning of the string, the middle with spaces, and the end
                    # This protects against removing params that are unique but
                    # are contained in another param.
                    # Example: volume-create zapi
                    #                 params: volume and volume-raid-type
                    foreach my $param_to_add (@params) {
                                # Entire list matches this param
                        if (    $ret_hash->{option_order} !~ /^$param_to_add$/
                                # First param in list matches this param
                            and $ret_hash->{option_order} !~ /^$param_to_add\s/
                                # Middle param in list matches this param
                            and $ret_hash->{option_order} !~ /\s$param_to_add\s/
                                # Last param in list matches this param
                            and $ret_hash->{option_order} !~ /\s$param_to_add$/)
                        {
                            $ret_hash->{option_order} .= " $param_to_add";
                        }
                    }
                }

                # change required/optional_params to required/optional_options
                $option =~ s/params/options/g;

                #  join params and save in $option
                if (scalar(@params)) {
                    $ret_hash->{$option} = join (" ", @params);
                } else {
                    $ret_hash->{$option} = "__NONE__";
                }
            }
        }
    }

    return $ret_hash;
} # end _mapped_hash

###############################################################################
# Method:    _translate_cdef()
# Objective: Translate CDEF cmd, hash, and args to zapi/7glike
# Inputs:    CDEF command, hash, and args
# Outputs:   ZAPI or 7glike command, hash, and args
###############################################################################
sub _translate_cdef
{
    my ($self, %args) = @_;

    my $dumpOut = Dumper(\%args) if ($self->get_debug_level() >= 9);
    $self->sanlog(-msg       => "Args Hash for _translate_cdef is:\n$dumpOut",
                  -msg_level => 9);

    my @new_required_options = ();
    my $cmd_name  = $args{cmd_name};
    my $cmd_hash  = $args{cmd_hash};
    my $cmd_args  = $args{cmd_args};

    my $ret_cmd;

    # translate cdef cmd to zapi from translation table
    if (exists ($mapped_cmds{$cmd_name}{cmd}{$args{interface}})) {
        $ret_cmd = $mapped_cmds{$cmd_name}{cmd}{$args{interface}};
    }

    $dumpOut = Dumper($mapped_cmds{$cmd_name}) if ($self->get_debug_level() >= 9);
    $self->sanlog(-msg       => "Mapped Commands for ($cmd_name) is:\n$dumpOut",
                  -msg_level => 9);

    #
    # map cdef params to zapi from mapped table
    #
    $cmd_hash->{cmd_name} = $cmd_args->{cmd_name} = $cmd_name;
    $cmd_hash->{interface} = $cmd_args->{interface} = $args{interface};
    my $ret_args = $self->_mapped_args(%{$cmd_args});

    # map cmd_hash
    my $ret_hash = $self->_mapped_hash(%{$cmd_hash});

    # aggregate create cmd either disk-count or disks (disklist) but unfortunately
    # the cdef contained both as required params. if one exists, ignore the other
    # this logic should work for other cmd as well
    my @required_options = split(" ", $ret_hash->{required_options});
    foreach my $arg (keys %{$ret_args}) {
        push @new_required_options, grep { $_ =~ /^$arg$/i } @required_options;
    }

    # set required_options
    $ret_hash->{required_options} = join(" ", @new_required_options);

    $dumpOut = Dumper($ret_hash) if ($self->get_debug_level() >= 9);
    $self->sanlog(-msg       => "Return hash for _translate_cdef is:\n$dumpOut",
                  -msg_level => 9);

    $dumpOut = Dumper($ret_args) if ($self->get_debug_level() >= 9);
    $self->sanlog(-msg       => "Return args for _translate_cdef is:\n$dumpOut",
                  -msg_level => 9);

    return ($ret_cmd, $ret_hash, $ret_args);
} # end _translate_to_zapi



###############################################################################
# Method:       _assign_hash_ref_value()
# Objective:    Transfor a Hash of hashes into a hash of references.
# Inputs:       ret_args - The hash of references.
#               mapped_args - The hash of arguments from TranslationTable.
#               value - The value to assign to the hash.
# Outputs:   A hash of refrerences.
###############################################################################
sub _assign_hash_ref_value {
    my ($self, %args) = @_;
    # Locals
    my $ret_args = $args{'ret_args'};
    my $mapped_arg = $args{'mapped_arg'};
    my $value = $args{'value'};

    # If the mapped arguments are a hash, need to build it accordingly.
    if ( ref($mapped_arg) =~ /HASH/) {
        # If it is a hash of hashes, need to build it recursively.
        my $key = (keys %{$mapped_arg})[0];
        $ret_args->{$key} =  $self->_assign_hash_ref_value('ret_args'=> $ret_args->{$key},
                                                           'mapped_arg'=> $mapped_arg->{$key},
                                                           'value' => $value);
    } else {
        # If the mapped_args is a key, assign the value to it.
        $ret_args->{$mapped_arg} = $value;
    }

    return $ret_args;
}# end _assign_hash_ref_value

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

CmdsDriver.pm

=head1 Author

David Pressley

=head1 Description

CmdsDriver.pm provides an API for executing CLI or ZAPI commands on a filer

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm>

B<San/SanLog.pm>

B<Utils (make_host make_conn)>

B<Time::HiRes (gettimeofday)>

B<Error>

B<Exporter>

B<NATE::Inc (get_file find_file)>

B<XML::Simple>

B<Data::Dumper>

B<Zapi>

B<NaElement>

B<Filer::Version (get_ontap_version)>

B<Tharn (host)>

B<San::CommonInterface>

=head1 Attributes

=head2 B<data_files>

array = List of XML files to be read to create the commands hash.

=head2 B<commands>

hash = Commands hash created by trasnlating XML data files into hashes

=head2 B<debug_level>

integer = Current debug level (higher level increases verbosity
and >100 sprints call stacks with comments).  Default is 0.

=head2 B<timeout>

integer = Seconds to wait for commands to execute (defaulted to 600, 10 min)

=head2 B<maxbufsize>

integer = size of the buffer (in bytes) to hold output from the Tharn::say() (analogous to the -maxbufsize flag in Tharn::say())
default is 40000000 (40MB)

=head2 B<cmd_timing>

integer = 0 or 1, enables or disables the timing of commands that get displayed after every command runs
default is 0 (off)

=head2 B<filer>

string = filer name or cserver (Cmode) (if interface is cli, require either filer or connection)

=head2 B<filer_version>

integer = Ontap Version. Default is 0.

=head2 B<filer_mode>

string = Ontap Mode (7mode or Cmode). Default is 7mode.

=head2 B<node>

string = filer/node name (Cmode). The reason we need node parameter is when filer
is a cserver that manages a cluster, it needs to comunicate to a specific node.
Default is first node in a cluster.

=head2 B<cmd_interface>

string = <cli | zapi>.  Default is "zapi".

=head2 B<privilege>

string = Current user privilege.  If set "priv set $privilege" will be
prepeneded to CLI command lines.  Default is "".

=head2 B<vfiler>

string = Current vfiler context.  If set "vfiler run -q $vfiler" will be
prepended to CLI command lines.  For ZAPI we will use the vfiler_ip
address instead of the normal host object default_ip value.  Default is "".

=head2 B<vfiler_ip>

string = Ip address for the current vfiler context.  If we are using a
passed in or set vfiler context we will need to use this IP address instead
of the normal host object default_ip value.  Default is undef.

=head2 B<partner_mode>

boolean = <1 | 0>.  If set "partner" will be prepended to CLI command lines.
Default is 0.

=head2 B<show_cmd>

boolean = <1 | 0>.  If set will log CLI commands before they are executed.
Default is 1.

=head2 B<execution>

boolean = <1 | 0>. Allow CLI commands to be formed, but not executed.
Default is 0.

=head2 B<output_format>

string = <raw | hash>. Format of the output returned from executed command.
Currently only raw is supported. Default is "raw".

=head2 B<log_output>

boolean = <1 | 0>. If set, always log the command output for ZAPI commands
Default is 0.

=head2 B<connection>

Connection = Connection object sent to say for CLI commands.
Default is "undef". (if interface is cli, require either filer or connection)

=head2 B<command_time>

Integer = The time it took for the last command to execute.  Times are
determinedby calling gettimeofday. Default is 0.

=head2 B<resultsHash>

hash = Last results returned.  Hash contains the following keys:
reslut, reason, output, output_check, result_check.  The reason is
only set by ZAPI.  The output_check and result_check fields are set
when we are checking the expected output or result and the check passes.

=head2 B<errmsg> (from Error.pm)

array = list of error strings accessed via Error.pm

=head2 B<error> (from Error.pm)

boolean = <1 | 0> Flag to be set when an error occurs

=head2 B<log>

Filehandle = Filehandle to log output to

=head1 Public Methods

=head2 B<new()>

Creates a new CmdsDriver object.  Give default values to all supported attributes.

=over 4

=item Inputs

=item * I<-log> (default is undef)

File descriptor for a log file.

=item * I<-debug_level> (default is 0)

integer = current debug level (higher level increases verbosity
and >100 prints call stacks with comments and errors).

=item * I<-filer> (default is "")

string = filer name (if interface is cli, require either -filer or -connection)

=item * I<-connection> (default is undef)

object = connection object (if interface is cli, require either -filer or -connection)

=item * I<-cmds> (default is "volume, lun, iscsi, fcp, igroup")

string = list of file names to be loaded: ex.) zapi_volume will load
zapi_volume.xml, volume will load zapi_volume.xml and cli_volume.xml

=item Outputs

Returns a CmdsDriver object.

=item Examples

    #
    # Create a CmdsDriver object using filer name
    #
    $cmd_obj = new CmdsDriver(-log         => \*ntest::NTEST_STDOUT,
                              -filer       => "rtpqa-fas2020-3",
                              -debug_level => $debug_level);
    if ($cmd_obj->error())
    {
        logresult('FATAL', "Failed to instantiate CmdsDriver object" . $cmd_obj->errmsg());
    }


=back

=head2 B<execute_cmd(%args)>

Execute a command on a filer.  For CLI commands, we will use the stored
connection attribute for command execution.  For ZAPI, we will make a
connection using.  The ZAPI username and password will come from the host
object for the filer.  If the host object does not contain the zapi_username
and zapi_password, we will use defaults.  The 7Mode defaults are "root"
and the host{default_password} field.  The CMode defaults are "admin" and
"netapp1!".

=over 4

=item Inputs

=item * -cmd

The command to execute

=item * -cmd_filer

Name of the filer on which to execute the command

=item * -cmd_filer_version (default is the value of the filer_version attribute)

Filer version to use for cmd look up

=item * I<-cmd_interface> <cli | zapi> (default is the value of the interface attribute)

Interface to use.

=item * I<-cmd_connection> (default is the value of the connection attribute)

Filer connection to use for CLI commands.

=item * I<-cmd_output_format> <raw | hash | system> (default is the value of the output_format attribute)

How to process the output field of the resutlsHash returned.  For CLI commands only "raw"
is supported.  For ZAPI's we recommend "hash".  The "system" format is specifically for
the system-cli zapi.  It will get the output out of the hash and return it as a string
as if the command was executed with the CLI.

=item * I<-cmd_privilege> (default is the value of the privilege attribute)

User privilege (priv set) level to use for CLI commands.

=item * I<-cmd_vfiler> (default is the value of the vfiler attribute)

Vfiler context to use for CLI and ZAPI commands. For ZAPI commands there are two
 ways to set Vfiler context. To make ZAPI calls directly to a Vfiler IP, set
 -cmd_vfiler to the vfiler name and -cmd_vfiler_ip to the IP address of the
 Vfiler. On the other hand, to make ZAPI calls to a vfiler context, just set
 -cmd_vfiler only. The default behavior is to only use -cmd_vfiler context.

=item * I<-cmd_vfiler_ip> (default is the value of the vfiler_ip attribute)

IP address of the vfiler to use for ZAPI commands.

=item * I<-cmd_partner> (default is the value of the partner attribute)

Use partner mode (prepend "partner" to commands) for CLI commands.
Accepts no|yes|0|1.

=item * I<-cmd_timeout> (default is the value of the timeout attribute)

Timeout value for a command, overrides timeout attribute for this
command execution.  Value is an integer number of seconds.

=item * I<-cmd_maxbufsize> (default is the value of the maxbufsize attribute)

Number of bytes to pass into Tharn::say()'s -maxbufsize flag for CLI commands.
This allows you to control how much of the buffer output to keep in memory.

=item * I<-cmd_timing> (default is the value of the cmd_timing attribute)

Enable or disable command timing.

=item * I<-cmd_pass_thru> (default is "")

Add this data to the command.  For CLI commands the given string is simply
appended to the command.  For ZAPI, the format is a space separated list
of name=value pairs.  Each name, value pair will be added to the ZAPI call.
For example "lun=/vol/myvol/lun1 aggr=test_aggr".

=item * I<-use_cmode_xml> (default is 0)

Use XML files for Cmode.
Accepts no|yes|0|1

=item * I<-require_options> (default is 1)

Check for required options?  Use "0" or "no" to accomplish negative testing.
Accepts no|yes|0|1.

=item * I<-check_output> (defaults to 0)

Compare the output from the resultsHash to the expected result?
Accepts no|yes|0|1.

=item * I<-expected_output> (defaults to undef then the value from the commands XML, if available)

Regular expression used to compare the output in the resultsHash to the
expected output for a given command.  Ignored unless -check_output is set.

=item * I<-check_result> (defaults to 0)

Compare the result from the resultsHash to the expected result?
Accepts no|yes|0|1.

=item * I<-expected_result> (defaults to undef then the value from the commands XML, if available)

Regular expression used to compare the result in the resultsHash to the
expected result for a given command.  Ignored unless -check_result is set.

=item Outputs

Returns a hash (resultsHash) . The hash contains the following keys:
reslut, reason, output, output_check, result_check.  The reason is
only set by ZAPI.  The output_check and result_check fields are set
when we are checking the expected output or result and the check passes.
If an error occured the error flag is set, errmsg is set to the failure reason,
and $self is passed back as the return value.

=item Examples

    #
    # Instantiate a SanFiler object
    #
    $filer_obj = new SanFiler(-filer       => $Filer,
                              -timeout     => $Timeout,
                              -debug_level => $Debug_Level,
                             );
    if ($filer_obj->error()) {
        $Test->nlog("ERROR: Failed to create a SAN Filer object. ",
                    $filer_obj->errmsg());
        return $TCD::FAIL;
    }

    #
    # Instantiate a CmdsDriver object using SanFiler connection
    #
    $cmd_obj = new CmdsDriver(-log         => \*ntest::NTEST_STDOUT,
                              -connection  => $filer_obj->connection(),
                              -debug_level => $debug_level);
    if ($cmd_obj->error())
    {
        $Test->nlog("ERROR: Failed to instantiate CmdsDriver object",
                    $cmd_obj->errmsg());
        return $TCD::FATAL;
    }

    $resultsHash_ref =
      $cmd_obj->execute_cmd(
          -cmd               => "lun-create-by-size",
          -cmd_filer         => $filer,
          -cmd_filer_version => $version,
          -size              => "100m",
          -type              => "solaris",
          -check_output      => "yes",
          -check_result      => "yes",
          -expected_result   => "255 0",
          -path              => "/vol/cmdsTest/cli_lun_100m" . "$$");

    if ($cmd_obj->error())
    {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $Test->nlog("ERROR: Failed calling execute_cmd: " . $errmsg);
        return $TCD::FAIL;
    }

    #
    # NOTE: To run the same execute_cmd on Cmode, you need to understand CDEF
    # commands and use the CmdsDriver/TranslationTable.pm to translate CDEF command
    # and its options to ZAPI command. If command and options that do not
    # require translation, you don't need to add in the %mapped_cmd table.
    # See TranslationTable.pm document for more information.
    #
    # For example, ZAPI options for lun-create-by-size are size, type, and path
    # However, for CDEF command, lun_create_byzise options are size, ostype, and
    # lun-path. Therefore, you need to translate type to ostype and path to
    # lun-path as display below.

    %mapped_cmds = (
         # Cmode CDEF                 # ZAPI/7Glike
         'lun_create_bysize' => {
            'cmd' => { 'zapi' => 'lun-create-by-size', '7glike' => 'lun_create_bysize'},
            'ostype' => { 'zapi' => 'type', '7glike' => '-t'},
            'lun-path' => { 'zapi' => 'path', '7glike' => '-p'},
             '7g_order' => '-t -p',
            },
    );

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check} &&
        $resultsHash_ref->{output_check})
    {
        return $TCD::PASS;
    }
    else
    {
        return $TCD::FAIL;
    }


    #
    # Get information on the lun created
    #
    $resultsHash_ref =
      $cmd_obj->execute_cmd(-cmd       => "lun-list-info",
                            -cmd_filer => $filer,
                            -verbose   => "",
                            -path      => "/vol/cmdsTest/cli_lun_100m" . "$$");

    my $lun_info_string = $resultsHash_ref->{output};


=back

=head2 B<discover_filer_version(filer)>

Find the filer version, format, save.

=over 4

=item Inputs

=item * filer

Name of a filer.

=item Outputs

None

=item Examples

    #
    # Instantiate CmdsDriver Object
    #
    my $cmd_obj = new CmdsDriver(-log    => \*ntest::NTEST_STDOUT,
                                 -filer  => $Filer);
    $cmd_obj->discover_filer_version($Filer);
    print "Filer version is " . $cmd_obj->version();

=back

=head2 B<get_debug_level()>

Find the current debug level (inherited from Sanlog).

=over 4

=item Outputs

Returns a scalar that's the current debug level.

=item Examples

    #
    # Instantiate San Object and get debug level.
    #
    my $cmd_obj = San::new();
    my $debug = $cmd_obj->get_debug_level();

=back

=head2 B<set_debug_level(level)>

Set the current debug level (inherited from Sanlog).

=over 4

=item Inputs

=item * level

New debug level.

=item Outputs

Returns a scalar 1.

=item Examples

    #
    # Instantiate San Object and set debug level.
    #
    my $cmd_obj = San::new();
    my $debug = $cmd_obj->set_debug_level(3);

=back

=head2 B<timeout(value)>

Get or Set command timeout.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value

Number of seconds for timeout

=item Outputs

Returns the timeout value.

=item Examples

    my $timeout = $cmd_obj->timeout(3600);

=back

=head2 B<maxbufsize(value)>

Get or Set command maxbufsize.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value

Number of bytes to pass in as maxbufsize to tharn::say()

=item Outputs

Returns the maxbufsize

=item Examples

    # set maxbufsize to 10MB
    my $maxbufsize = $cmd_obj->maxbufsize(10000000);

=back

=head2 B<cmd_timing(value)>

Get or Set command timing attribute

=over 4

=item Inputs

=item * value

0 or 1, to disable or enable command timing respectively.

=item Outputs

Returns the maxbufsize

=item Examples

 # enable command timing
 $cmd_obj->cmd_timing(1);

 # disable command timing
 $cmd_obj->cmd_timing(0);

=back

=head2 B<filer_version(value)>

Get or Set filer_version attribute.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value

Appropriate value for a given attribute.

=item Outputs

Returns the attribute value.

=item Examples

    my $filer_version = $cmd_obj->filer_version(72);

=back

=head2 B<filer_mode(value)>

Get or Set filer_mode attribute.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value

Appropriate value for a given attribute.

=item Outputs

Returns the attribute value.

=item Examples

    my $filer_mode = $cmd_obj->filer_mode("Cmode");

=back

=head2 B<interface(value)>

Get or Set the interface attribute.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value (cli | zapi)

Interface to use to execute commands

=item Outputs

Returns the attribute value.

=item Examples

    my $interface = $cmd_obj->interface("zapi");

=back

=head2 B<privilege(value)>

Get or Set the privilege attribute.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value

Appropriate value for a CLI privilege level (test, diag, admin, etc.).

=item Outputs

Returns the attribute value.

=item Examples

    my $privilege = $cmd_obj->privilege("test");

=back

=head2 B<vfiler(value)>

Get or Set the vfiler attribute.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value

Appropriate name for a vfiler attribute.

=item Outputs

Returns the attribute value.

=item Examples

    my $vfiler = $cmd_obj->vfiler("vfiler14");

=back

=head2 B<vfiler_ip(value)>

Get or Set the vfiler_ip attribute.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value

Appropriate value for the vfiler_ip attribute.

=item Outputs

Returns the attribute value.

=item Examples

    my $vfiler_ip = $cmd_obj->vfiler_ip("10.60.123.124");

=back

=head2 B<partner_mode(value)>

Get or Set the partner_mode attribute.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value (0 | 1)

Appropriate value for the partner_mode attribute.

=item Outputs

Returns the attribute value.

=item Examples

    my $partner_mode = $cmd_obj->partner_mode(1);

=back

=head2 B<show_cmd(value)>

Get or Set the show_cmd attribute.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value (0 | 1)

Appropriate value for the show_cmd attribute.

=item Outputs

Returns the attribute value.

=item Examples

    my $show_cmd = $cmd_obj->show_cmd(1);

=back

=head2 B<output_format(value)>

Get or Set the output_format attribute.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value (raw | format)

Appropriate value for the output_format attribute.  Currently only raw is supported.

=item Outputs

Returns the attribute value.

=item Examples

    my $output_format = $cmd_obj->output_format("raw");

=back

=head2 B<connection(value)>

Get or Set the connection attribute.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value

Appropriate value for the connection attribute.

=item Outputs

Returns the attribute value.

=item Examples
    my $filer_conn = connect ($FILER_CONN);
    $cmd_obj->connection($filer_conn);

=back

=head2 B<command_time(value)>

Get or Set the command_time attribute.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value

Appropriate value for the command_time attribute.

=item Outputs

Returns the attribute value.

=item Examples

    my $command_time = $cmd_obj->command_time();

=back

=head2 B<log(value)>

Get or Set the log attribute.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value

Log handle used to multiplex results.

=item Outputs

Returns the attribute value.

=item Examples

    $cmd_obj->log($my_log_handle);

=back

=head2 B<execution(value)>

Get or Set the execution attribute.  If a value is given, the value is set.  The value is always returned.

=over 4

=item Inputs

=item * value (0 | 1)

Appropriate value for the execution attribute.

=item Outputs

Returns the attribute value.

=item Examples

    my $execution = $cmd_obj->execution();

=back

=head2 B<resultsHash(value)>

Get the resultsHash attribute.

=over 4

=item Inputs

=item * value

Appropriate value for the resultsHash attribute.

=item Outputs

Returns the attribute value.

=item Examples

    my $hashRef = $cmd_obj->resultsHash();

=back

=cut

1;
