package Cmode::HyA_Clone_Utils;

#------------------------------------------------------
#   Module Vars
#------------------------------------------------------
our ($VERSION);

#
# Define constants
#

use warnings;
use strict qw(vars subs refs);
use base   qw(Exporter);
use SFOTest  qw(run_nodescope_command );
use Utils::Units qw(convert_to_bytes);

#------------------------------------------------------
#   Imports
#------------------------------------------------------
use Tharn;
use Data::Dumper;
use Params::Validate qw(validate SCALAR ARRAYREF SCALARREF OBJECT);


#------------------------------------------------------
#     Prototypes
#------------------------------------------------------
sub disable_fixup;
sub enable_fixup;
sub add_ssd_disks;
sub enable_hya_flag;
sub wait_clones_split;
sub wait_clones_split_snap;
sub stop_traffic;
sub start_dt;
sub start_hammer;
sub get_md5sum;

BEGIN {

    #--------------------------------------------
    #   Exports
    #--------------------------------------------

    $VERSION    = '0.01';

    my @hya_clone_sub = qw(
                        &disable_fixup
                        &enable_fixup
                        &add_ssd_disks
                        &enable_hya_flag
                        &wait_clones_split
                        &wait_clones_split_snap
                        &stop_traffic
			&start_dt
			&start_hammer
			&get_md5sum
                    );

    our %EXPORT_TAGS = (
        'HyA_Clone_Utils'      => [ @hya_clone_sub ],
    );

    $EXPORT_TAGS{all} = [ map{ @{$_} } values %EXPORT_TAGS ];

    Exporter::export_ok_tags('all');
}


=head2 disable_fixup

Disable split on the node

=cut

sub disable_fixup
{
    my %args_validn_spec =
        (
            'node'     => {type => OBJECT},
        );

    my %opts = validate(@_, \%args_validn_spec);
    my $Node = $opts{'node'};

    my $out;

    $out = $Node->run(
                command => " \"priv set test;".
                           "printflag wafl_disable_sis_clone_fixup\"");

    if ( $out =~ /wafl_disable_sis_clone_fixup = 1/){
         return 0;
    }else{
        $out = $Node->run(
                    command => " \"priv set test;".
                               "setflag wafl_disable_sis_clone_fixup 1\"" );
        $out = $Node->run( command => " \"priv set test;".
                                  "printflag wafl_disable_sis_clone_fixup\"");
        if ( $out =~ /wafl_disable_sis_clone_fixup = 1/){
            return 0;
        }else{
            return 1;
        }
    }

}


=head2 enable_fixup

Enable split on the node

=cut

sub enable_fixup
{
    my %args_validn_spec =
        (
            'node'     => {type => OBJECT},
        );

    my %opts = validate(@_, \%args_validn_spec);
    my $Node = $opts{'node'};

    my $out;

    $out = $Node->run(
                command => " \"priv set test;".
                           "printflag wafl_disable_sis_clone_fixup \"" );

    if ( $out =~ /wafl_disable_sis_clone_fixup = 0/){
         return 0;
    }else{
        $out = $Node->run(
                        command => " \"priv set test;".
                                   "setflag wafl_disable_sis_clone_fixup 0\"" );
        $out = $Node->run(
                        command => " \"priv set test;".
                                   "printflag wafl_disable_sis_clone_fixup \"" );
        if ( $out =~ /wafl_disable_sis_clone_fixup = 0/){
            return 0;
        }else{
            return 1;
        }
    }
}

=head2 add_ssd_disks($)

Adds SSD disks to given aggregate

=cut

sub add_ssd_disks
{
    my %args_validn_spec =
        (
            'aggr'     => {type => SCALAR},
	    'api_obj'  => {type => OBJECT},
  	    'node'     => {type => OBJECT},
	    'num_ssds' => {type => SCALAR},
        );

    my %opts = validate(@_, \%args_validn_spec);
    my $aggr = $opts{'aggr'};
    my $apiobj = $opts{'api_obj'};
    my $Node = $opts{'node'};
    my $num_ssds = $opts{'num_ssds'};

    my $out = $apiobj->execute_raw_command( command =>
                        "disk show -aggregate $aggr -fields type");
    if($out =~ /SSD/){
        $Log->comment("Aggregate $aggr already has SSD's");
        return 0;
    }else{
        NACL::C::StorageAggregate->add_disks(
            command_interface   => $Node,
            aggregate           => $aggr,
            diskcount           => "$num_ssds",
            'disktype'          => "SSD",
        );
    }

    $out = $apiobj->execute_raw_command( command =>
                        "disk show -aggregate $aggr -fields type");
    if($out =~ /SSD/){
        $Log->comment("Added SSD's successfully...");
        return 0;
    }else{
        $Log->comment("Failed to add SSD's...");
        return 1;
    }

    return 0;
}

=head2 enable_hya_flag($)

Enables hya flag on the aggregate

=cut

sub enable_hya_flag
{
    my %args_validn_spec =
        (
            'aggr'     => {type => SCALAR},
            'api_obj'  => {type => OBJECT},
            'node'     => {type => OBJECT},
        );

    my %opts = validate(@_, \%args_validn_spec);
    my $aggr = $opts{'aggr'};
    my $apiobj = $opts{'api_obj'};
    my $Node = $opts{'node'};

    my $out = $apiobj->execute_raw_command( command =>
                          "aggr show $aggr -fields hybrid-enabled" );
    if($out =~ /true/){
        $Log->comment("Hybrid flag is already enabled on $aggr");
        return 0;
    } else {
        NACL::STask::Aggregate->modify(
            command_interface   => $Node,
            aggregate           => $aggr,
            'hybrid-enabled'    => "true",
        );
    }

    $out = $apiobj->execute_raw_command( command =>
                          "aggr show $aggr -fields hybrid-enabled" );
    if($out =~ /true/){
        $Log->comment("Enabled hybrid flag on $aggr");
        return 0;
    } else {
        $Log->comment("Failed to enable hybrid flag on $aggr");
        return 1;
    }
}


=head2 wait_clones_split

Wait for clones from AFS to be split

=cut
sub wait_clones_split
{
    
    my %args_validn_spec =
        (
            'vol'     => {type => SCALAR},
            'api_obj'  => {type => OBJECT},
            'node'     => {type => OBJECT},
            'timeout' => {type => SCALAR},
        );

    my %opts = validate(@_, \%args_validn_spec);
    my $vol = $opts{'vol'};
    my $apiobj = $opts{'api_obj'};
    my $Node = $opts{'node'};
    my $TIMEOUT = $opts{'timeout'};

    my ($out,@out1,$outstanding);

    my $end_time = time() + $TIMEOUT; # Get end time
    my $cur_time = time(); # Get current time
    while ($cur_time < $end_time) {
        $out = $apiobj->execute_raw_command( command =>
                        "run local \"priv set test;".
                        "cmf_test $vol header\"",
                        'timeout' => $TIMEOUT
            );

        @out1 = split(" ",$out);
        for (my $i = 0; $i <= $#out1; $i++){
            if ($out1[$i] =~ /num_entries/){
                $i++;
                      $outstanding = $out1[$i];
                    $Log->comment("no of unsplit entries = $outstanding");
            }
        }
        if($outstanding){
            $Log->comment("outstanding split entries = $outstanding");
        }else{
            $Log->comment("outstanding = $outstanding");
            return 0;
        }
        sleep(30); ## wait for some time before checking outstnading splits
        $cur_time = time();
    }

    $Log->comment("TIMEOUT - waiting for splits to complete");
    return 1;
}

=head2 wait_clones_split_snap

Wait for clones from snapshot to be split 

=cut
sub wait_clones_split_snap
{
    my %args_validn_spec =
        (
            'vol'     => {type => SCALAR},
            'api_obj'  => {type => OBJECT},
            'node'     => {type => OBJECT},
            'timeout' => {type => SCALAR},
        );

    my %opts = validate(@_, \%args_validn_spec);
    my $vol = $opts{'vol'};
    my $apiobj = $opts{'api_obj'};
    my $Node = $opts{'node'};
    my $TIMEOUT = $opts{'timeout'};

    my ($out,@out1,$outstanding);

    my $end_time = time() + $TIMEOUT; # Get end time
    my $cur_time = time(); # Get current time
    while ($cur_time < $end_time) {
        $out = $apiobj->execute_raw_command( command =>
                        "run local \"priv set test;".
                        "cmf_test $vol header\"",
                        'timeout' => $TIMEOUT
            );

        if( $out =~ /snapid/){
            @out1 = split(" ",$out);
            for (my $i = 0; $i <= $#out1; $i++){
                if ($out1[$i] =~ /snapid/){
                $i = $i + 5;
                      $outstanding = $out1[$i];
                    $Log->comment("no of snap unsplit entries = $outstanding");
                }
            }
            if($outstanding){
                $Log->comment("num of splits from snap remaining = $outstanding");
            }else{
                $Log->comment("All snap clones are split..");
                return 0;
            }
            sleep(30);
            $cur_time = time();
        }else {
            $Log->comment("No clones pending from snap...");
            return 0;
        }
    }

    $Log->comment("TIMEOUT - waiting for splits to complete from snap");
    return 1;
}


=head2 stop_traffic

Stops trafic on FILER_VOL

=cut

sub stop_traffic
{
    my %args_validn_spec =
       (
            'traffic' => {type => SCALARREF},
#	    'traffic_flag' => {type => SCALARREF},
       );

    my %opts = validate(@_, \%args_validn_spec);

    my $traffic_ref = $opts{'traffic'};
#    my $traffic_flag_ref = $opts{'traffic_flag'};
		
    eval {
        $$traffic_ref->stop();
    };
    if($@){
        $Log->comment("traffic possibly exited");
        sleep(5);
    }
#    $$traffic_flag_ref = 0;
    return 0;
}

=head2 start_dt

Starts dt on FILER_VOL

=cut

sub start_dt
{
    my %args_validn_spec =
        (
            'dt_traffic' => {type => SCALARREF},
	    'of'         => {type => SCALAR},
	    'client'     => {type => OBJECT},
            'pattern'    => {type => SCALAR},
            'records'    => {type => SCALAR},
            'depth'      => {type => SCALAR},
            'files'      => {type => SCALAR},
            'mnt_path'   => {type => SCALAR},
            'run_time'   => {type => SCALAR},
            'bs'         => {type => SCALAR , default  => "4k"},
            'dispose'    => {type => SCALAR , default  => "keep"},
            'sdirs'      => {type => SCALAR , default  => "2"},
       );

    my %opts = validate(@_, \%args_validn_spec);

    my $dt_traffic_ref = $opts{'dt_traffic'};
    my $dt_of = $opts{'of'};
    my $dt_pattern = $opts{'pattern'};
    my $dt_records= $opts{'records'};
    my $dt_files = $opts{'files'};
    my $dt_depth = $opts{'depth'};
    my $mnt_pt = $opts{'mnt_path'};
    my $bs = $opts{'bs'};
    my $dispose = $opts{'dispose'};
    my $dt_sdirs = $opts{'sdirs'};
    my $Client = $opts{'client'};
    my $traffic_run_time = $opts{'run_time'};

    my ($traffic,$file_md5sum,$raw_output,@output);

    $Log->comment("mnt point = $mnt_pt");
    $dt_of = "$mnt_pt/$dt_of";
    $Log->comment("dt file = $dt_of");

    $$dt_traffic_ref = NACL::MTask::Traffic->new(
        command_interface => $Client,
        paths => $mnt_pt
    );

    my $on_exception = sub {
        # The first argument to the method would be an object of type
        # NACL::STask::Client::RPC::Future
        my $future = shift;
        $Log->comment("start_dt : on_exception handler is invoked");
        if ($future->exception()) {
            # Exception occurred. Perform any exception handling
            # specific steps.
            $Log->comment("Error in running traffic");
            $Log->comment("Error is: \n". $future->exception());
            return 1;
        }
    };

    $Log->comment("Start the dt IO traffic");
    $$dt_traffic_ref->start(
            tool    => 'dt',
            of      => $dt_of,
            pattern => $dt_pattern,
            bs      => '4k',
            dispose => 'keep',
            sdirs   => $dt_sdirs,
            depth   => $dt_depth,
            records => $dt_records,
            runtime => $traffic_run_time,
            files   => $dt_files,
           on_exception => $on_exception,
    );
    sleep(10); ## let the traffic start

    return 0;
}

=head2 start_hammer

Starts hammer on FILER_VOL

=cut

sub start_hammer
{
    my %args_validn_spec =
        (
            'hammer_traffic'    => {type => SCALARREF},
            'file_size'         => {type => SCALAR},
	    'client'            => {type => OBJECT},
            'mnt_path'          => {type => SCALAR},
	    'run_time'	  	=> {type => SCALAR},
       );

    my %opts = validate(@_, \%args_validn_spec);

    my $hammer_traffic_ref = $opts{'hammer_traffic'};
    my $hammer_file_sz = $opts{'file_size'};
    my $Client = $opts{'client'};
    my $mnt_pt = $opts{'mnt_path'};
    my $traffic_run_time = $opts{'run_time'};

    my ($traffic_hammer,$file_md5sum,$raw_output,@output);
    my $size = NACL::C::UnitNormalization::convert_to_bytes($hammer_file_sz);

    $$hammer_traffic_ref = NACL::MTask::Traffic->new(
            command_interface => $Client,
            paths => $mnt_pt
    );

    my $on_exception = sub {
        # The first argument to the method would be an object of type
        # NACL::STask::Client::RPC::Future
        my $future = shift;
        $Log->comment("start_hammer : on_exception handler is invoked");
        if ($future->exception()) {
            # Exception occurred. Perform any exception handling
            # specific steps.
            $Log->comment("Error in running traffic");
            $Log->comment("Error is: \n". $future->exception());
            return 1;
        }
    };

    $Log->comment("Start the hammer traffic");
    $$hammer_traffic_ref->start(
        tool => 'hammer',
              runtime => $traffic_run_time,
              blocksize => "random",
              #maxfsize => "$size -nofilercore -nocleanup",
              maxfsize => "$size -nofilercore",
              mode    => "mixed",
       on_exception => $on_exception,
    );
    sleep(10); ##wait for few seconds,let the traffic start
#    $traffic_index++;

    return 0;
}

#------------------------------------------------------
#     Subroutines Definition
#------------------------------------------------------

=head2 get_md5sum()

Return md5sum of a file

=over 4

=item Arguments:

=over 4

Takes 2 scalars as inputs

=back

=item Returns:

Returns md5sum if successful, otherwise dies

=back

=cut

sub get_md5sum
{
    my %args_validn_spec =
        (
            'file'        => {type => SCALAR},
            'seconds'     => {type => SCALAR},
            'max_tries'   => {type => SCALAR},
            'client'     => {type => OBJECT},
            'timeout'     => {type => SCALAR},
        );

    my %opts = validate(@_, \%args_validn_spec);

    my $file = $opts{'file'};
    my $seconds = $opts{'seconds'};
    my $max_tries = $opts{'max_tries'};
    my $apiobj_client = $opts{'client'};
    my $TIMEOUT = $opts{'timeout'};

    my ($file_md5sum,$raw_output,@output);

    for (my $j = 0; $j < $max_tries+10; $j++){
        $raw_output = $apiobj_client->execute_raw_command(
                               command => "/usr/bin/md5sum $file",
                                timeout => $TIMEOUT);
        @output = split(' ',$raw_output);
        $file_md5sum = $output[0];
        my $len = length $file_md5sum;
        if($len ==  32){
            if($file_md5sum =~ /\d{1}/){
                $Log->comment("Got the md5sum...len - $len");
                return $file_md5sum;
            }else{
                $Log->comment("Did not get md5sum???...");
                $file_md5sum = "";
            }

            if(!$file_md5sum){
                if ( $j == ($max_tries - 1) ){
                    die "failed to get md5sum";
                }
                $Log->comment("Try num $j - Failed to get md5sum...retrying..");
                sleep(10);
            }else{
                $j = $max_tries;
            }
        }else{
            $Log->comment("got wrong value?? - $file_md5sum");
            sleep($seconds);
        }
    }
    die "failed to get md5sum";

}
