# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Cmode/Jobs.pm#1 $
#
# $Change: 1917575 $
#
# $DateTime: 2013/01/22 11:22:46 $
#
# Copyright (c) 2009 Network Appliance, Inc.
# All rights reserved.
#
#
# NAME
#      Jobs.pm - provide interfaces to C-Mode job cli command 
#
# DESCRIPTION
#      This perl module provide wrapper interfaces to c-mode job cli command
#
#
# AUTHOR
#       tarung@netapp.com
#
#
#
## @pod here
##
##

=pod

=head1 NAME

    Jobs.pm -  provide interfaces to C-Mode job cli command

=head1 SYNOPSIS

    use Cmode::Jobs qw(:all);
    or
    use Cmode::Jobs qw(:jobs);
    or
    use Cmode::Jobs qw(:schedule);

    use Server;

    my $err_msg;
    my $Server = Server->new() or die("Unable to connect to Server");
    my $jobId  = 123;

    job_delete(
        server => $Server,
        id     => $jobId,
        errmsg => \$err_msg,
    ) or die("job delete failed, error: $$err_msg");

    job_pause(
        server => $Server,
        id     => $jobId,
        errmsg => \$err_msg,
    ) or die("job pause failed, error: $$err_msg");

    job_resume(
        server => $Server,
        id     => $jobId,
        errmsg => \$err_msg,
    ) or die("job resume failed, error: $$err_msg");

    job_stop(
        server => $Server,
        id     => $jobId,
        errmsg => \$err_msg,
    ) or die("job stop failed, error: $$err_msg");

    job_watch_progress(
        server => $Server,
        id     => $jobId,
        errmsg => \$err_msg,
    ) or die("job watch progress failed, error: $$err_msg");

    my @job_list;

    job_show(
        server => $Server,
        errmsg => \$err_msg,
        cli_op => \@job_list;
    ) or die("job show failed, error: $$err_msg");

    OR
    job_show(
        server => $Server,
        jobtype => 'VOL_CLONE_SPLIT'
        errmsg => \$err_msg,
        cli_op => \@job_list;
    ) or die("job show failed, error: $$err_msg");

    job_history_show(
        server => $Server,
        node   => 'f3070-150-92'
        errmsg => \$err_msg,
        name   => 'Split parent_vol'
        cli_op => \@job_list;
    ) or die("job show failed, error: $$err_msg");


    job_schedule_cron_create(
        server => $Server,
        name   => '10min'
        minute => '10,20,30,40,50',
        errmsg => \$err_msg,
    ) or die("job schedule cron create failed, error: $$err_msg");


    job_schedule_cron_delete(
        server => $Server,
        name   => '10min',
        errmsg => \$err_msg,
    ) or die("job schedule cron delete failed, error: $$err_msg");

    job_schedule_cron_modify(
        server => $Server,
        name   => '10min',
        minute => '10,30,40',
        errmsg => \$err_msg,
    ) or die("job schedule cron modify failed, error: $$err_msg");

    job_schedule_cron_show(
        server => $Server,
        errmsg => \$err_msg,
        cli_op => \@job_list;
    ) or die("job_schedule_cron_show failed, error: $$err_msg");

    job_schedule_interval_create(
        server => $Server,
        name   => '10min10sec'
        minute => '10',
        seconds => '10',
        errmsg => \$err_msg,
    ) or die("job schedule interval create failed, error: $$err_msg");

    job_schedule_interval_delete(
        server => $Server,
        name   => '10min10sec',
        errmsg => \$err_msg,
    ) or die("job schedule interval delete failed, error: $$err_msg");

    job_schedule_interval_modify(
        server => $Server,
        name   => '10min10sec',
        seconds => '30',
        errmsg => \$err_msg,
    ) or die("job schedule interval modify failed, error: $$err_msg");

    job_schedule_interval_show(
        server => $Server,
        errmsg => \$err_msg,
        cli_op => \@job_list;
    ) or die("job_schedule_interval_show failed, error: $$err_msg");


=head1 DESCRIPTION

    This library implements wrapper interface around all job cli commands.

=head1 FUNCTIONS

Cmode Job Functions

=cut

package Cmode::Jobs ;

#------------------------------------------------------
#   Module Vars
#------------------------------------------------------
our ($Validation, $VERSION);

#
# Define constants
#
use constant JB_ERROR => undef;
use constant JB_TRUE  => 1;
use constant JB_FALSE => 0;


use warnings;
use strict qw(vars subs refs);
use base   qw(Exporter);

BEGIN {

    #--------------------------------------------
    #   Exports
    #--------------------------------------------

    $Validation = 1; # unused as of now
    $VERSION    = '0.01';

    my @jobs_sub = qw(
                        &job_delete
                        &job_pause
                        &job_resume
                        &job_stop
                        &job_watch_progress
                        &job_show
                        &job_history_show
                        &wait_for_job_completion
                        &wait_for_batch_jobs_completion
                        &cleanup_stray_jobs
                    );

    my @schedule_sub = qw(
                        &job_schedule_cron_create
                        &job_schedule_cron_delete
                        &job_schedule_cron_modify
                        &job_schedule_cron_show
                        &job_schedule_interval_create
                        &job_schedule_interval_delete
                        &job_schedule_interval_modify
                        &job_schedule_interval_show
                        );

    my @constants = qw(JB_ERROR JB_TRUE JB_FALSE);

    our %EXPORT_TAGS = (
        'jobs'      => [ @jobs_sub ],
        'schedule'  => [ @schedule_sub ],
        'constants' => [ @constants ],
    );

    $EXPORT_TAGS{all} = [ map{ @{$_} } values %EXPORT_TAGS ];

    Exporter::export_ok_tags('all');
}

#------------------------------------------------------
#   Imports
#------------------------------------------------------
use Tharn;
use Data::Dumper;
use Params::Validate     qw(validate validate_pos SCALAR ARRAYREF
                            SCALARREF OBJECT HASHREF CODEREF);

#------------------------------------------------------
#     Prototypes
#------------------------------------------------------
sub job_delete;
sub job_pause;
sub job_resume;
sub job_stop;
sub job_watch_progress;
sub job_show;
sub job_history_show;
sub wait_for_job_completion;
sub wait_for_batch_jobs_completion;
sub cleanup_stray_jobs;

sub job_schedule_cron_create;
sub job_schedule_cron_delete;
sub job_schedule_cron_modify;
sub job_schedule_cron_show;

sub job_schedule_interval_create;
sub job_schedule_interval_delete;
sub job_schedule_interval_modify;
sub job_schedule_interval_show;

#------------------------------------------------------
#     Local Variables
#------------------------------------------------------


#------------------------------------------------------
#     Subroutines Definition 
#------------------------------------------------------

=head2 job_delete

deletes the job by jobid

=over 4

=item Arguments:

=over 4

takes a hash as input with following key
    server  => server object
    id      => job id
    errmsg => referece to error variable
    timeout => timeout in seconds [60]

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_delete {
    my %options = validate(@_,
                    {
                        server      => {type => OBJECT},
                        id          => {type => SCALAR},
                        errmsg      => {type => SCALARREF},
                        timeout     => {
                                        type => SCALAR,
                                        optional =>1,
                                        default  => 60,
                                       },
                    });

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};
    $input_arg{'id'}       = $options{'id'};

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->job_delete( %input_arg );
    };

    if($@) {
        $$err_ref = "job_delete() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}


=head2 job_pause

pauses the job by jobid

=over 4

=item Arguments:

=over 4

takes a hash as input with following key
    server  => server object
    id      => job id
    errmsg => referece to error variable
    timeout => timeout in seconds [60]

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_pause {
    my %options = validate(@_,
                    {
                        server      => {type => OBJECT},
                        id          => {type => SCALAR},
                        errmsg      => {type => SCALARREF},
                        timeout     => {
                                        type => SCALAR,
                                        optional =>1,
                                        default  => 60,
                                       },
                    });

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};
    $input_arg{'id'}       = $options{'id'};

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->job_pause( %input_arg );
    };

    if($@) {
        $$err_ref = "job_pause() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}

=head2 job_resume

resumes the job by jobid

=over 4

=item Arguments:

=over 4

takes a hash as input with following key
    server  => server object
    id      => job id
    errmsg => referece to error variable
    timeout => timeout in seconds [60]

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_resume {
    my %options = validate(@_,
                    {
                        server      => {type => OBJECT},
                        id          => {type => SCALAR},
                        errmsg      => {type => SCALARREF},
                        timeout     => {
                                        type => SCALAR,
                                        optional =>1,
                                        default  => 60,
                                       },
                    });

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};
    $input_arg{'id'}       = $options{'id'};

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->job_resume( %input_arg );
    };

    if($@) {
        $$err_ref = "job_resume() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}

=head2 job_stop

stops the job by jobid

=over 4

=item Arguments:

=over 4

takes a hash as input with following key
    server  => server object
    id      => job id
    errmsg => referece to error variable
    timeout => timeout in seconds [60]

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_stop {
    my %options = validate(@_,
                    {
                        server      => {type => OBJECT},
                        id          => {type => SCALAR},
                        errmsg      => {type => SCALARREF},
                        timeout     => {
                                        type => SCALAR,
                                        optional =>1,
                                        default  => 60,
                                       },
                    });

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};
    $input_arg{'id'}       = $options{'id'};

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->job_stop( %input_arg );
    };

    if($@) {
        $$err_ref = "job_stop() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}

=head2 job_watch_progress

watch progress of job by jobid

=over 4

=item Arguments:

=over 4

takes a hash as input with following key
    server   => server object
    id       => job id
    interval => interval in seconds [60]
    errmsg  => referece to error variable
    timeout  => timeout in seconds [60]

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_watch_progress {
    my %options = validate(@_,
                    {
                        server      => {type => OBJECT},
                        id          => {type => SCALAR},
                        interval    => {
                                        type => SCALAR,
                                        optional => 1,
                                        default  => 60,
                                       },
                        errmsg      => {type => SCALARREF},
                        timeout     => {
                                        type => SCALAR,
                                        optional => 1,
                                        default  => 60,
                                       },
                    });

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};
    $input_arg{'id'}       = $options{'id'};
    $input_arg{'interval'} = $options{'interval'};

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->job_watch_progress( %input_arg );
    };

    if($@) {
        $$err_ref = "job_watch_progress() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}


=head2 job_show

show the jobs

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_show {

    my @state_list = qw( Initial Queued Running Waiting Pausing
                        Paused Quitting Success Failure Reschedule
                        Error Quit Dead Unknown Restart Dormant );

    my $states = join('|', @state_list);

    my $rgexp_valid_states = qr/^($states)$/;

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
            cli_op  => 1, 
        );

    my %args_validn_spec = 
        (
            server      => {type => OBJECT},
            name        => {type => SCALAR, optional => 1},
            id          => {type => SCALAR, optional => 1},
            description => {type => SCALAR, optional => 1},
            priority    => {
                            type => SCALAR,
                            optional =>1,
                            regex => qr/^(Low|Medium|High)$/,
                           },
            node        => {type => SCALAR, optional => 1},
            affinity    => {
                            type => SCALAR,
                            optional =>1,
                            regex => qr/^(Cluster|Node)$/,
                           },

            schedule    => {type => SCALAR, optional => 1},
            queuetime   => {type => SCALAR, optional => 1},
            starttime   => {type => SCALAR, optional => 1},
            endtime     => {type => SCALAR, optional => 1},
            deadtime    => {type => SCALAR, optional => 1},
            restarted   => {
                            type => SCALAR,
                            optional =>1,
                            regex => qr/^(true|false)$/,
                           },
            state       => {
                            type => SCALAR,
                            optional =>1,
                            regex  => $rgexp_valid_states,
                           },
            code        => {
                            type => SCALAR,
                            optional =>1,
                            regex => qr/^(\d+)$/,
                           },
            completion  => {type => SCALAR, optional => 1},
            jobtype     => {type => SCALAR, optional => 1},
            category    => {type => SCALAR, optional => 1},
            progress    => {type => SCALAR, optional => 1},
            errmsg      => {type => SCALARREF},
            cli_op      => {type => ARRAYREF},
            timeout     => {
                            type => SCALAR,
                            optional =>1,
                            default  => 300,
                           },
        );
                        

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = "\"" . $options{$key} . "\"";
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        @{$cmdOpRef} = $server->job_show( %input_arg );
    };

    if($@) {
        $$err_ref = "job_show() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}


=head2 job_history_show

show the jobs histroy

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_history_show {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
            cli_op  => 1, 
        );

    my %args_validn_spec = 
        (
            server      => {type => OBJECT},
            node        => {type => SCALAR, optional => 1},
            record      => {
                            type => SCALAR,
                            optional => 1,
                            regex => qr/^(\d+)$/,
                           },
            id          => {type => SCALAR, optional => 1},
            endtime     => {type => SCALAR, optional => 1},
            starttime   => {type => SCALAR, optional => 1},
            name        => {type => SCALAR, optional => 1},
            description => {type => SCALAR, optional => 1},
            code        => {
                            type => SCALAR,
                            optional =>1,
                            regex => qr/^(\d+)$/,
                           },
            completion  => {type => SCALAR, optional => 1},
            errmsg      => {type => SCALARREF},
            cli_op      => {type => ARRAYREF},
            timeout     => {
                            type => SCALAR,
                            optional =>1,
                            default  => 300,
                           },
        );
                        

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->job_history_show( %input_arg );
    };

    if($@) {
        $$err_ref = "job_history_show() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}


=head2 job_schedule_cron_create

create the cron job schedule 

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_schedule_cron_create {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            server      => {type => OBJECT},
            name        => {type => SCALAR},
            month       => {type => SCALAR, optional => 1},
            dayofweek   => {type => SCALAR, optional => 1},
            day         => {type => SCALAR, optional => 1},
            hour        => {type => SCALAR, optional => 1},
            minute      => {type => SCALAR},
            errmsg      => {type => SCALARREF},
            timeout     => {
                            type => SCALAR,
                            optional =>1,
                            default  => 60,
                           },
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->job_schedule_cron_create( %input_arg );
    };

    if($@) {
        $$err_ref = "job_schedule_cron_create() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}

=head2 job_schedule_cron_delete

deletes the cron job schedule by name

=over 4

=item Arguments:

=over 4

takes a hash as input with following key
    server  => server object
    name    => cron job name
    errmsg => referece to error variable
    timeout => timeout in seconds [60]

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_schedule_cron_delete {
    my %options = validate(@_,
                    {
                        server      => {type => OBJECT},
                        name        => {type => SCALAR},
                        errmsg      => {type => SCALARREF},
                        timeout     => {
                                        type => SCALAR,
                                        optional =>1,
                                        default  => 60,
                                       },
                    });

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};
    $input_arg{'name'}     = $options{'name'};

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->job_schedule_cron_delete( %input_arg );
    };

    if($@) {
        $$err_ref = "job_schedule_cron_delete() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}

=head2 job_schedule_cron_modify

modify the cron job schedule

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_schedule_cron_modify {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            server      => {type => OBJECT},
            name        => {type => SCALAR},
            month       => {type => SCALAR, optional => 1},
            dayofweek   => {type => SCALAR, optional => 1},
            day         => {type => SCALAR, optional => 1},
            hour        => {type => SCALAR, optional => 1},
            minute      => {type => SCALAR},
            errmsg      => {type => SCALARREF},
            timeout     => {
                            type => SCALAR,
                            optional =>1,
                            default  => 60,
                           },
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->job_schedule_cron_modify( %input_arg );
    };

    if($@) {
        $$err_ref = "job_schedule_cron_modify() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}

=head2 job_schedule_cron_show

show the cron job schedules

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_schedule_cron_show {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
            cli_op  => 1,
        );

    my %args_validn_spec =
        (
            server      => {type => OBJECT},
            name        => {type => SCALAR, optional => 1},
            month       => {type => SCALAR, optional => 1},
            dayofweek   => {type => SCALAR, optional => 1},
            day         => {type => SCALAR, optional => 1},
            hour        => {type => SCALAR, optional => 1},
            minute      => {type => SCALAR, optional => 1},
            description => {type => SCALAR, optional => 1},
            errmsg      => {type => SCALARREF},
            timeout     => {
                            type => SCALAR,
                            optional => 1,
                            default  => 60,
                           },
            cli_op      => {type => ARRAYREF},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        @{$cmdOpRef} = $server->job_schedule_cron_show( %input_arg );
    };

    if($@) {
        $$err_ref = "job_schedule_cron_show() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}

=head2 job_schedule_interval_create

create the interval job schedule

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_schedule_interval_create {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            server      => {type => OBJECT},
            name        => {type => SCALAR},
            days        => {type => SCALAR, optional => 1},
            hours       => {type => SCALAR, optional => 1},
            minutes     => {type => SCALAR, optional => 1},
            seconds     => {type => SCALAR, optional => 1},
            errmsg      => {type => SCALARREF},
            timeout     => {
                            type => SCALAR,
                            optional => 1,
                            default  => 60,
                           },
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->job_schedule_interval_create( %input_arg );
    };

    if($@) {
        $$err_ref = "job_schedule_interval_create() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}


=head2 job_schedule_interval_delete

deletes the interval job schedule by name

=over 4

=item Arguments:

=over 4

takes a hash as input with following key
    server  => server object
    name    => interval job name
    errmsg => referece to error variable
    timeout => timeout in seconds [60]

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_schedule_interval_delete {
    my %options = validate(@_,
                    {
                        server      => {type => OBJECT},
                        name        => {type => SCALAR},
                        errmsg      => {type => SCALARREF},
                        timeout     => {
                                        type => SCALAR,
                                        optional =>1,
                                        default  => 60,
                                       },
                    });

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};
    $input_arg{'name'}     = $options{'name'};

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->job_schedule_interval_delete( %input_arg );
    };

    if($@) {
        $$err_ref = "job_schedule_interval_delete() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}


=head2 job_schedule_interval_modify

modify the interval job schedule

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_schedule_interval_modify {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            server      => {type => OBJECT},
            name        => {type => SCALAR},
            days        => {type => SCALAR, optional => 1},
            hours       => {type => SCALAR, optional => 1},
            minutes     => {type => SCALAR, optional => 1},
            seconds     => {type => SCALAR, optional => 1},
            errmsg      => {type => SCALARREF},
            timeout     => {
                            type => SCALAR,
                            optional => 1,
                            default  => 60,
                           },
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->job_schedule_interval_modify( %input_arg );
    };

    if($@) {
        $$err_ref = "job_schedule_interval_modify() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}

=head2 job_schedule_interval_show

show the interval job schedules

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub job_schedule_interval_show {
    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
            cli_op  => 1,
        );

    my %args_validn_spec =
        (
            server      => {type => OBJECT},
            name        => {type => SCALAR, optional => 1},
            days        => {type => SCALAR, optional => 1},
            hours       => {type => SCALAR, optional => 1},
            minutes     => {type => SCALAR, optional => 1},
            seconds     => {type => SCALAR, optional => 1},
            description => {type => SCALAR, optional => 1},
            errmsg      => {type => SCALARREF},
            timeout     => {
                            type => SCALAR,
                            optional =>1,
                            default  => 60,
                           },
            cli_op      => {type => ARRAYREF},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        @{$cmdOpRef} = $server->job_schedule_interval_show( %input_arg );
    };

    if($@) {
        $$err_ref = "job_schedule_interval_show() failed, error: $@";
        return JB_FALSE;
    }

    return JB_TRUE;
}

=head2 wait_for_job_completion

wait for a job given by jobid to complete

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub wait_for_job_completion {

    my %args_validn_spec =
        (
            server     => {type => OBJECT},
            jobid      => {type => SCALAR},
            errmsg     => {type => SCALARREF},
            interval   => {type => SCALAR, default  => 1},
            timeout    => {type => SCALAR, default  => 60},
            job_info   => {type => HASHREF, optional =>1},
        );

    my %options  = validate(@_, \%args_validn_spec);

    my $server   = $options{'server'};  # server object
    my $jobid    = $options{'jobid'};   # job id
    my $timeout  = $options{'timeout'}; # timeout for cli command
    my $err_ref  = $options{'errmsg'};  # reference error msg variable
    my $interval = $options{'interval'}; # interval between two calls

    my $err_msg;
    my @jobs;
    my $match_regexp = qr/There are no entries matching your query/;

    # undef the job_info hash, if available
    if ( exists $options{'job_info'} ) {
        %{$options{'job_info'}} = ();
    }

    my $finish = time() + $timeout;

    while ( time() <=  $finish ) {

        if ( JB_FALSE == job_show(
                            server => $server,
                            id     => $jobid,
                            errmsg => \$err_msg,
                            cli_op => \@jobs) ) {
                if ( $err_msg =~ $match_regexp ) {
                    logcomment("No split job running");
                    return JB_TRUE;
                } else {
                    $$err_ref = "job show failed, error: $err_msg";
                    return JB_FALSE;
                }
            }

        if ( $jobs[0]->{'progress'} =~ /Complete:/i ) {
            logcomment("Job ID: $jobid is completed");
            if ( $options{'job_info'} ) {
                %{$options{'job_info'}} = %{$jobs[0]};
            }
            return JB_TRUE;
        } else {
            logcomment("Job ID: $jobid is running");
            sleep($interval);
        }
    }

    #
    # timeout occured
    #
    $$err_ref = "split job id: $jobid did not complete in $timeout seconds";

    return JB_FALSE;
}

=head2 wait_for_batch_jobs_completion

wait for all jobs of given jobtype to complete 

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub wait_for_batch_jobs_completion {

    my %args_validn_spec =
        (
            server     => {type => OBJECT},
            jobtype    => {type => SCALAR},
            errmsg     => {type => SCALARREF},
            interval   => {type => SCALAR, default  => 60},
            timeout    => {type => SCALAR, default  => 600},
        );

    my %options  = validate(@_, \%args_validn_spec);

    my $server   = $options{'server'};   # server object
    my $jobtype  = $options{'jobtype'};  # job type
    my $err_ref  = $options{'errmsg'};   # reference error msg variable
    my $interval = $options{'interval'}; # interval between two calls
    my $timeout  = $options{'timeout'};  # timeout for cli command

    my $err_msg;
    my @jobs;
    my $match_regexp = qr/There are no entries matching your query/;

    my $finish = time() + $timeout;

    #
    # disable showallfields to have less op and save time
    #
    $server->set('showallfields' => 'false');
    while ( time() <=  $finish ) {

        if ( JB_FALSE == job_show(
                            server  => $server,
                            jobtype => $jobtype,
                            state   => 'Running',
                            errmsg  => \$err_msg,
                            timeout => $timeout,
                            cli_op  => \@jobs) ) {
            if ( $err_msg =~ $match_regexp ) {
                logcomment("No $jobtype running job found");
                $server->set('showallfields' => 'true');
                return JB_TRUE;
            } else {
                logcomment("job_show() failed, error: $err_msg");
                $$err_ref = "job show failed, error: $err_msg";
                $server->set('showallfields' => 'true');
                return JB_FALSE;
            }
        } else {
            logcomment("$#jobs $jobtype jobs are running");
            logcomment("Sleeping for $interval second..");
            sleep($interval);
        }
    } ## end while

    #
    # enable showallfields again
    #
    $server->set('showallfields' => 'true');

    #
    # timeout occured
    #
    $$err_ref = "$jobtype jobs did not complete in $timeout seconds";

    return JB_FALSE;
}

=head2 cleanup_stray_jobs

delete all jobs of given jobtype

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub cleanup_stray_jobs {

    my %args_validn_spec =
        (
            server     => {type => OBJECT},
            jobtype    => {type => SCALAR},
            errmsg     => {type => SCALARREF},
            timeout    => {type => SCALAR, default  => 300},
        );

    my %options  = validate(@_, \%args_validn_spec);

    my $server   = $options{'server'};   # server object
    my $jobtype  = $options{'jobtype'};  # job type
    my $err_ref  = $options{'errmsg'};   # reference error msg variable
    my $timeout  = $options{'timeout'};  # timeout for cli command

    my $err_msg;
    my @jobs;
    my $failCount = 0;
    my $match_regexp = qr/There are no entries matching your query/;

    if ( JB_FALSE == job_show(
                        server  => $server,
                        jobtype => $jobtype,
                        errmsg  => \$err_msg,
                        timeout => $timeout,
                        cli_op  => \@jobs) ) {
        if ( $err_msg =~ $match_regexp ) {
            logcomment("No $jobtype job found for cleanup");
            return JB_TRUE;
        } else {
            logcomment("job_show() failed, error: $err_msg");
            $$err_ref = "job show failed, error: $err_msg";
            return JB_FALSE;
        }
    }

    logcomment("Cleaning up existing $jobtype type job");
    $$err_ref = "";
    foreach my $job ( @jobs ) {
        my $id = $job->{'id'};
        my $state = $job->{'state'};
        my $descr = $job->{'description'};
        logcomment("JobId: $id, State: $state, Descr: $descr");
        job_delete(
            server => $server,
            id     => $id,
            errmsg => \$err_msg,
            timeout => $timeout)
            or ( $failCount++ and
                $$err_ref .= "$id job delete failed, error: $err_msg");
    }

    if ( $failCount ) {
        logcomment("job_delete() failed, error: $$err_ref");
        return JB_FALSE;
    }

    return JB_TRUE;
}




1;


=head1 SEE ALSO

    None

=head1 STATUS

    Shared

=head1 AUTHOR

    Tarun Garg (tarung@netapp.com)
    $Id: //depot/prod/test/Rsierranevada/lib/Cmode/Jobs.pm#1 $
    Last $Change: 1917575 $
    Last Revised $DateTime: 2013/01/22 11:22:46 $

=head1 HISTORY

    2009-12-24 author: tarung      original lib created

=head1 COPYRIGHT

Copyright 2009, Network Appliance., All Rights Reserved

=cut

