# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Cmode/LunCmds.pm#1 $
#
# $Change: 1917575 $
#
# $DateTime: 2013/01/22 11:22:46 $
#
# Copyright (c) 2010 Network Appliance, Inc.
# All rights reserved.
#
#
# NAME
#      LunCmds.pm - cmode Lun commands library
#
# DESCRIPTION
#      This perl module provide wrapper interfaces to c-mode 
#       volume lun cli command
#
# AUTHOR
#       sladi@netapp.com
#
#
## @pod here
##
##

=pod

=head1 NAME

    LunCmds.pm - provide wrapper interfaces to C-Mode Lun operations cli command

=head1 SYNOPSIS

    use Cmode::LunCmds qw(:all);
    OR
    use Cmode::LunCmds qw(:lun);

=head1 DESCRIPTION

=head1 FUNCTIONS

Cmode Lun Operations Functions

=cut

package Cmode::LunCmds ;

#------------------------------------------------------
#   Module Vars
#------------------------------------------------------
our ($Validation, $VERSION);

#
# Define constants
#
use constant LUN_ERROR => undef;
use constant LUN_TRUE  => 1;
use constant LUN_FALSE => 0;

use warnings;
use strict qw(vars subs refs);

BEGIN {

    use base   qw(Exporter);

    #--------------------------------------------
    #   Exports
    #--------------------------------------------

    $Validation = 1; # unused as of now
    $VERSION    = '0.01';

    my @lun_sub = qw(
                        &lun_create
                        &lun_show
                        &lun_map
                        &lun_unmap
                        &lun_destroy
                        &lun_client_setup
                        &lun_filer_setup
                        &lun_igroup_create
                        &lun_igroup_add
                        &lun_igroup_show
                        &lun_igroup_delete
                        &vserver_iscsi_create
                        &vserver_iscsi_start
                        &vserver_iscsi_stop
                        &vserver_iscsi_delete
                        &vserver_iscsi_show
                    );

    my @constants = qw(LUN_ERROR LUN_TRUE LUN_FALSE);

    our %EXPORT_TAGS = (
        'lun'          => [ @lun_sub ],
        'constants'    => [ @constants ],
    );

    $EXPORT_TAGS{all} = [ map{ @{$_} } values %EXPORT_TAGS ];

    Exporter::export_ok_tags('all');
}

#------------------------------------------------------
#   Imports
#------------------------------------------------------
use Tharn;
use Data::Dumper;
use Client::Host::LinuxHost;
use San::SanClient;
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF SCALARREF OBJECT );
use Cmode::Flexclone qw(:all);
use Cmode::Quota qw(:all);

#------------------------------------------------------
#     Prototypes
#------------------------------------------------------
sub lun_create;
sub lun_show;
sub lun_map;
sub lun_destroy;
sub lun_client_setup;
sub lun_filer_setup;
sub lun_igroup_create;
sub lun_igroup_add; 
sub lun_igroup_show;
sub lun_igroup_delete;
sub vserver_iscsi_create; 
sub vserver_iscsi_start;
sub vserver_iscsi_stop;
sub vserver_iscsi_delete;
sub vserver_iscsi_show;

#------------------------------------------------------
#     Local Variables
#------------------------------------------------------


#------------------------------------------------------
#     Subroutines Definition
#------------------------------------------------------


=head2 lun_create()

creates lun on the specified volume 

=over 4

=item Arguments:

required_options => [qw( vserver volume lun size ostype )],

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub lun_create {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'     => {type => OBJECT},
            'vserver'    => {type => SCALAR},
            'volume'     => {type => SCALAR},
            'lun'        => {type => SCALAR},
            'size'       => {type => SCALAR},
            'ostype'     => {type => SCALAR},
            'lun-path'   => {type => SCALAR, optional => 1},
            'extent-size'=> {type => SCALAR, optional => 1},
            'suffix-size'=> {type => SCALAR, optional => 1},
            'option'     => {type => SCALAR, optional => 1},
            'file-path'  => {type => SCALAR, optional => 1},
            'errmsg'     => {type => SCALARREF},
            'timeout'    => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->lun_create( %input_arg );
    };

    if($@) {
        $$err_ref = "lun_create() failed, error: $@";
        return LUN_FALSE;
    }

    return LUN_TRUE;
}

=head2 lun_show()

shows lun on the specified volume 

=over 4

=item Arguments:

required_options => [qw( vserver volume lun size ostype )],
optional_options => [qw( lun-path space-resv state path qtree prefix-size
                    extent-size suffix-size option write file-path )]

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub lun_show {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'     => {type => OBJECT},
            'vserver'    => {type => SCALAR, optional => 1},
            'volume'     => {type => SCALAR, optional => 1},
            'lun'        => {type => SCALAR, optional => 1},
            'size'       => {type => SCALAR, optional => 1},
            'ostype'     => {type => SCALAR, optional => 1},
            'lun-path'   => {type => SCALAR, optional => 1},
            'extent-size'=> {type => SCALAR, optional => 1},
            'suffix-size'=> {type => SCALAR, optional => 1},
            'option'     => {type => SCALAR, optional => 1},
            'file-path'  => {type => SCALAR, optional => 1},
            'cli_op'     => {type => ARRAYREF},
            'errmsg'     => {type => SCALARREF},
            'timeout'    => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
         @{$cmdOpRef} = $server->lun_show( %input_arg );
    };

    if($@) {
        $$err_ref = "lun_show() failed, error: $@";
        return LUN_FALSE;
    }

    return LUN_TRUE;
}

=head2 lun_map()

maps lun on the specified volume 

=over 4

=item Arguments:

required_options => [qw( vserver volume lun igroup )],
optional_options => [qw( lun-path force lun-id lun-id-assigned path qtree )],

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub lun_map {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'         => {type => OBJECT},
            'vserver'        => {type => SCALAR},
            'volume'         => {type => SCALAR},
            'lun'            => {type => SCALAR},
            'igroup'         => {type => SCALAR},
            'lun-path'       => {type => SCALAR, optional => 1},
            'force'          => {type => SCALAR, optional => 1},
            'lun-id-assigned'=> {type => SCALAR, optional => 1},
            'path'           => {type => SCALAR, optional => 1},
            'qtree'          => {type => SCALAR, optional => 1},
            'errmsg'         => {type => SCALARREF},
            'timeout'        => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->lun_map( %input_arg );
    };

    if($@) {
        $$err_ref = "lun_map() failed, error: $@";
        return LUN_FALSE;
    }

    return LUN_TRUE;
}

=head2 lun_unmap()


unmaps lun on the specified volume 

=over 4

=item Arguments:

required_options => [qw( vserver volume lun igroup )],
optional_options => [qw( lun-path path qtree )]

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub lun_unmap {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'   => {type => OBJECT},
            'vserver'  => {type => SCALAR},
            'volume'   => {type => SCALAR},
            'lun'      => {type => SCALAR},
            'igroup'   => {type => SCALAR},
            'lun-path' => {type => SCALAR, optional => 1},
            'path'     => {type => SCALAR, optional => 1},
            'qtree'    => {type => SCALAR, optional => 1},
            'errmsg'   => {type => SCALARREF},
            'timeout'  => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->lun_unmap( %input_arg );
    };

    if($@) {
        $$err_ref = "lun_unmap() failed, error: $@";
        return LUN_FALSE;
    }

    return LUN_TRUE;
}


=head2 lun_destroy()


destroys lun on the specified volume 

=over 4

=item Arguments:

required_options => [qw( vserver volume lun )],
optional_options => [qw( lun-path force )],

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub lun_destroy {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'   => {type => OBJECT},
            'vserver'  => {type => SCALAR},
            'volume'   => {type => SCALAR},
            'lun'      => {type => SCALAR},
            'lun-path' => {type => SCALAR, optional => 1},
            'force'    => {type => SCALAR, optional => 1},
            'errmsg'   => {type => SCALARREF},
            'timeout'  => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->lun_destroy( %input_arg );
    };

    if($@) {
        $$err_ref = "lun_destroy() failed, error: $@";
        return LUN_FALSE;
    }

    return LUN_TRUE;
}

=head2 vserver_iscsi_show()


shows iscsi services for specified node /vserver 

=over 4

=item Arguments:

required_options => undef,
optional_options => [qw( vserver nodename alias status-admin  
max-error-recovery-level retain-timeout login-timeout max-conn-per-session 
max-ios-per-session tcp-window-size retain-timeout )],

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub vserver_iscsi_show {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'                   => {type => OBJECT},
            'vserver'                  => {type => SCALAR, optional => 1},
            'nodename'                 => {type => SCALAR, optional => 1},
            'alias'                    => {type => SCALAR, optional => 1},
            'status-admin'             => {type => SCALAR, optional => 1},
            'max-error-recovery-level' => {type => SCALAR, optional => 1},
            'retain-timeout'           => {type => SCALAR, optional => 1},
            'login-timeout'            => {type => SCALAR, optional => 1},
            'max-conn-per-session'     => {type => SCALAR, optional => 1},
            'max-ios-per-session'      => {type => SCALAR, optional => 1},
            'tcp-window-size'          => {type => SCALAR, optional => 1},
            'cli_op'                   => {type => ARRAYREF},
            'errmsg'                   => {type => SCALARREF},
            'timeout'                  => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }
    # reset the error message
    $$err_ref = "none";

    eval {
         @{$cmdOpRef}  = $server->vserver_iscsi_show( %input_arg );
    };

    if($@) {
        $$err_ref = "vserver_iscsi_show() failed, error: $@";
        return LUN_FALSE;
    }


    return LUN_TRUE;
}


=head2 vserver_iscsi_create()


create iscsi services for specified node /vserver 

=over 4

=item Arguments:

required_options => [qw( vserver )],
optional_options => [qw( nodename alias status-admin max-error-recovery-level 
retain-timeout login-timeout max-conn-per-session max-ios-per-session 
tcp-window-size start target-name target-alias )],

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub vserver_iscsi_create {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'                   => {type => OBJECT},
            'vserver'                  => {type => SCALAR},
            'nodename'                 => {type => SCALAR, optional => 1},
            'alias'                    => {type => SCALAR, optional => 1},
            'status-admin'             => {type => SCALAR, optional => 1},
            'max-error-recovery-level' => {type => SCALAR, optional => 1},
            'retain-timeout'           => {type => SCALAR, optional => 1},
            'login-timeout'            => {type => SCALAR, optional => 1},
            'max-conn-per-session'     => {type => SCALAR, optional => 1},
            'max-ios-per-session'      => {type => SCALAR, optional => 1},
            'tcp-window-size'          => {type => SCALAR, optional => 1},
            'start'                    => {type => SCALAR, optional => 1},
            'target-name'              => {type => SCALAR, optional => 1},
            'target-alias'             => {type => SCALAR, optional => 1},
            'errmsg'                   => {type => SCALARREF},
            'timeout'                  => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }
    # reset the error message
    $$err_ref = "none";

    eval {
        $server->vserver_iscsi_create( %input_arg );
    };

    if($@) {
        $$err_ref = "vserver_iscsi_create() failed, error: $@";
        return LUN_FALSE;
    }

    return LUN_TRUE;
}


=head2 vserver_iscsi_start()


starts iscsi services for specified node /vserver 

=over 4

=item Arguments:

required_options => [qw( vserver )],
optional_options => undef,

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub vserver_iscsi_start {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'                   => {type => OBJECT},
            'vserver'                  => {type => SCALAR},
            'errmsg'                   => {type => SCALARREF},
            'timeout'                  => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }
    # reset the error message
    $$err_ref = "none";

    eval {
        $server->vserver_iscsi_start( %input_arg );
    };

    if($@) {
        $$err_ref = "vserver_iscsi_start() failed, error: $@";
        return LUN_FALSE;
    }

    return LUN_TRUE;
}


=head2 vserver_iscsi_stop()


stop iscsi services for specified node /vserver 

=over 4

=item Arguments:

required_options => [qw( vserver )],
optional_options => undef,

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub vserver_iscsi_stop {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'                   => {type => OBJECT},
            'vserver'                  => {type => SCALAR},
            'errmsg'                   => {type => SCALARREF},
            'timeout'                  => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }
    # reset the error message
    $$err_ref = "none";

    eval {
        $server->vserver_iscsi_stop( %input_arg );
    };

    if($@) {
        $$err_ref = "vserver_iscsi_stop() failed, error: $@";
        return LUN_FALSE;
    }

    return LUN_TRUE;
}


=head2 vserver_iscsi_delete()


delete iscsi services for specified node /vserver 

=over 4

=item Arguments:

required_options => [qw( vserver )],
optional_options => undef,

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub vserver_iscsi_delete {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'                   => {type => OBJECT},
            'vserver'                  => {type => SCALAR},
            'errmsg'                   => {type => SCALARREF},
            'timeout'                  => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }
    # reset the error message
    $$err_ref = "none";

    eval {
        $server->vserver_iscsi_delete( %input_arg );
    };

    if($@) {
        $$err_ref = "vserver_iscsi_delete() failed, error: $@";
        return LUN_FALSE;
    }

    return LUN_TRUE;
}

=head2 lun_igroup_create ()


create a igroup 

=over 4

=item Arguments:

required_options => [qw( vserver igroup ostype initiator )],

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub lun_igroup_create {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'            => {type => OBJECT},
            'vserver'           => {type => SCALAR},
            'igroup'            => {type => SCALAR},
            'ostype'            => {type => SCALAR},
            'initiator'         => {type => SCALAR},
            'protocol-type'     => {type => SCALAR, optional => 1},
            'errmsg'    => {type => SCALARREF},
            'timeout'   => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }
    # reset the error message
    $$err_ref = "none";

    eval {
        $server->lun_igroup_create( %input_arg );
    };

    if($@) {
        $$err_ref = "lun_igroup_create() failed, error: $@";
        return LUN_FALSE;
    }

    return LUN_TRUE;
}


=head2 lun_igroup_add ()


Add initiator(s) to an initiator group 

=over 4

=item Arguments:

required_options => [qw( vserver igroup initiator )],
optional_options => [qw( force )],

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub lun_igroup_add {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'    => {type => OBJECT},
            'vserver'   => {type => SCALAR},
            'igroup'    => {type => SCALAR},
            'initiator' => {type => SCALAR},
            'force'     => {type => SCALAR, optional => 1},
            'errmsg'    => {type => SCALARREF},
            'timeout'   => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }
    # reset the error message
    $$err_ref = "none";

    eval {
        $server->lun_igroup_add( %input_arg );
    };

    if($@) {
        $$err_ref = "lun_igroup_add() failed, error: $@";
        return LUN_FALSE;
    }

    return LUN_TRUE;
}

=head2 lun_igroup_show ()


show igroup 

=over 4

=item Arguments:

  [ -instance | -v | -verbose | -fields <fieldname>, ... ]
  [ -vserver <vserver name> ]           Vserver Name (default: vs0)
  [[-igroup] <text (size 1..96)>]       Igroup Name
  [ -protocol-type {mixed|fcp|iscsi} ]  Protocol Type
  [ -ostype|-t {solaris|windows|hpux|aix|linux|netware|vmware|openvms|xen|hyper_v} ]
                                        OS type
  [ -portset|-a <text> ]                Portset binding igroup
  [ -uuid <UUID> ]                      UUID of the igroup
  [ -alua {true|false} ]                ALUA
  [ -init-details <text>, ... ]         Initiators


=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub lun_igroup_show {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'        => {type => OBJECT},
            'vserver'       => {type => SCALAR, optional => 1},
            'igroup'        => {type => SCALAR, optional => 1},
            'protocol-type' => {type => SCALAR, optional => 1},
            'ostype'        => {type => SCALAR, optional => 1},
            'portset'       => {type => SCALAR, optional => 1},
            'uuid'          => {type => SCALAR, optional => 1},
            'init-details'  => {type => SCALAR, optional => 1},
            'cli_op'        => {type => ARRAYREF},
            'errmsg'        => {type => SCALARREF},
            'timeout'       => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }
    # reset the error message
    $$err_ref = "none";

    eval {
        @{$cmdOpRef} = $server->lun_igroup_show( %input_arg );
    };

    if($@) {
        $$err_ref = "lun_igroup_show() failed, error: $@";
        return LUN_FALSE;
    }

    return LUN_TRUE;
}


=head2 lun_igroup_show ()


deletes igroup 

=over 4

=item Arguments:

required_options => [qw( vserver igroup )],
optional_options => [qw( force  )],

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub lun_igroup_delete {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'  => {type => OBJECT},
            'vserver' => {type => SCALAR},
            'igroup'  => {type => SCALAR},
            'force'   => {type => SCALAR, optional => 1},
            'errmsg'  => {type => SCALARREF},
            'timeout' => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }
    # reset the error message
    $$err_ref = "none";

    eval {
         $server->lun_igroup_delete( %input_arg );
    };

    if($@) {
        $$err_ref = "lun_igroup_delete() failed, error: $@";
        return LUN_FALSE;
    }

    return LUN_TRUE;
}

1;
