
#$Id: //depot/prod/test/Rsierranevada/lib/Cmode/ProtocolAccess.pm#1 $
## @summary
##    Accessing the volume using different protocols
##
## @description
##   Mount the volume on a client and accessing it using NFS,ISCSI protocol
##
## @author
##   anila@netapp.com
##
## @status
##   public
##
## @dependency
##  Lun size should be maximum of 1G.

package Cmode::ProtocolAccess;

use base qw(Exporter);

BEGIN {
    our @EXPORT = qw(access_protocol_nfs nfs_setup nfs_io nfs_cleanup iscsi_setup iscsi_io iscsi_cleanup);
}

use strict;
use Tharn;
use Data::Dumper;
use ClientMgr;
use Params::Validate qw(validate SCALAR OBJECT);
use CsiUtils qw(random_name_generate);


use constant LICENSE_CODE_NFS   => "TJFAEKPJYBFDAA";
use constant LICENSE_CODE_ISCSI => "RYAEFKPJYBFDAA";
use constant SIO_PATH           => "/usr/software/test/bin/sio";
use constant CLIENTMATCH        => "0.0.0.0/0";
use constant SUPERUSER          => "any";
use constant ANON               => "65534";
use constant RORULE             => "any";
use constant RWRULE             => "any";
use constant PROTOCOL_NFS       => "nfs";


## @Function
##      access_protocol_nfs- Does the NFS setup,mount the volume and access it
##
##
##
## @usage
##      use Protocol_access qw(access_protocol_nfs);
##      my $ret = access_protocol_nfs(server        => $Server,
##                                    vserver       => $vserver,
##                                    policyname    => "export_policy_nfs",
##                                    domain        => $domain,
##                                    'name-servers'=> $name_servers,
##                                    ip            => $ip,
##                                    client        => $client,
##                                    'mount-point' => $mount_pt);
##
## @arguments
##      Server                Required    Server object
##      vserver               Required    Vserver name
##      policyname            Required    Specifies the export policy that is to be created.
##      domain                Required    Domain name
##      name-servers          Required    IP address of DNS servers
##      ip                    Required    lif ip address
##      client                Required    Client on which volume to be mounted
##      mount-point           Required    Path at which volume to be mounted
##      junction-path         Required    Junction path
##      vol-name              Required    Volume creted for data access
##      root-vol              Required    Root volume name
##      file-size             Optional    Size of, file to be created during sio
##      file-quantity         Optional    Number of files to be created during sio
##      threads               Optional    Number of threads to be created during sio
##
## @return-value
##      0 on failure
##      1 on success
##
## @scope
##      Exported from this module
##


sub access_protocol_nfs {
    my %opts = validate(@_,
                        {server          => {type => OBJECT, isa => 'Server'},
                         vserver         => {type => SCALAR},
                         policyname      => {type => SCALAR},
                         domain          => {type => SCALAR},
                         'name-servers'  => {type => SCALAR},
                         ip              => {type => SCALAR},
                         client          => {type => SCALAR},
                         'mount-point'   => {type => SCALAR},
                         'junction-path' => {type => SCALAR},
                         'vol-name'      => {type => SCALAR},
                         'root-vol'      => {type => SCALAR},
                         'file-size'     => {type => SCALAR, optional => 1, default => "3m"}, 
                         'file-quantity' => {type => SCALAR, optional => 1, default => "1"},
                         threads         => {type => SCALAR, optional => 1, default => "1"}});

    my $Server        = $opts{server};
    my $vserver       = $opts{vserver};
    my $policyname    = $opts{policyname};
    my $domain        = $opts{domain};
    my $name_servers  = $opts{'name-servers'};
    my $ip            = $opts{ip};
    my $client        = $opts{client};
    my $mount_point   = $opts{'mount-point'};
    my $junction_path = $opts{'junction-path'};
    my $vol           = $opts{'vol-name'};
    my $root_vol      = $opts{'root-vol'};
    my $file_size     = $opts{'file-size'};
    my $file_quantity = $opts{'file-quantity'};
    my $threads       = $opts{'threads'};

    my $ret_value = 0;
    eval {


        logcomment("Verifying if NFS license exists..if not adding license");
        eval { $Server->system_license_show(feature => "NFS"); };
        if ($@ =~ /There are no entries matching your query/) {
            logcomment("NFS license not exists creating...");
            $Server->system_license_add('license-code' => LICENSE_CODE_NFS);
        } else {
            logcomment("NFS license already exists");
        }


        logcomment("Enabling and configuring NFS on a virtual server $vserver");
        $Server->vserver_nfs_create(vserver => $vserver);

        logcomment("Creating export-policy ");
        $Server->vserver_export_policy_create(vserver    => $vserver,
                                              policyname => $policyname);
        eval { $Server->vserver_export_policy_show(vserver => $vserver); };
        if ($@) {
            die("Failure in creating export policy");
        }
        eval { $Server->vserver_export_policy_rule_show(vserver => $vserver); };
        if ($@ =~ /There are no entries matching your query/){
            logcomment("Creating export-policy rule");
            $Server->vserver_export_policy_rule_create(vserver      => $vserver,
                                                   policyname   => $policyname,
                                                   clientmatch  => CLIENTMATCH,
                                                   rorule       => RORULE,
                                                   rwrule       => RWRULE,
                                                   protocol     => PROTOCOL_NFS,
                                                   anon         => ANON,
                                                   superuser    => SUPERUSER);
       
        }else{
            $Log->comment("Export-policy rule already exists for $vserver");
        }


        logcomment("Modifying the volume");
        $Server->volume_modify(volume  => $vol,
                               vserver => $vserver,
                               policy  => $policyname);

        $Server->volume_modify(volume             => $root_vol,
                               vserver            => $vserver,
                               policy             => $policyname,
                               'unix-permissions' => '---rwxrwxrwx');

        logcomment("Mounting the Volume ");
        $Server->volume_mount(volume          => $vol,
                              vserver         => $vserver,
                              'junction-path' => $junction_path);

        eval {
            $Server->vserver_services_dns_show(vserver => $vserver,
                                               domains => $domain);
        };
        if ($@ =~ /There are no entries matching your query/) {
            logcomment("Creating dns service");
            $Server->vserver_services_dns_create(vserver        => $vserver,
                                             state          => "enabled",
                                             domains        => $domain,
                                             'name-servers' => $name_servers);
        }else{
            $Log->comment("DNS already configured for $vserver");
        }

        logcomment(
               "Mount the volume and start accessing mount point $mount_point");
        $ret_value = _run_cm_load(ip              => $ip,
                                  'junc-path'     => $junction_path,
                                  'mount_point'   => $mount_point,
                                  client          => $client,
                                  'file-size'     => $file_size,
                                  'file-quantity' => $file_quantity,
                                  threads         => $threads);

    };
    if ($@) {
        die("Error  while doing NFS setup/mount or IO Reason:$@");
    }
    return ($ret_value);
} ## end sub access_protocol_nfs

sub _run_cm_load {
    my %params = @_;

    my $ip            = $params{ip};
    my $mount_point   = $params{'mount_point'};
    my $junction_path = $params{'junc-path'};
    my $client        = $params{client};
    my $file_size     = $params{'file-size'};
    my $file_quantity = $params{'file-quantity'};
    my $threads       = $params{threads};

    my $ret = 0;
    Tharn::snooze(5);
    my ($used_size_bfr_sio, $used_size_aft_sio);

    # Run the load...
    #

    my %clientMap = {};
    my @map_array;

    $map_array[0] = {data_ip => $ip,
                     path    => $mount_point,
                     volume  => {'junction-path' => $junction_path},};
    $clientMap{$client} = \@map_array;

    logcomment("Creating new client object");
    my $clientMgr_obj = ClientMgr->new(clients => $client);
    logcomment("Creating new client object... DONE");

    logcomment("Adding load object... ");
    my $id = $clientMgr_obj->add_load(type       => ClientMgr->SIO(),
                                      bin_params => {file_qty => $file_quantity,
                                                     file_size => $file_size,
                                                     threads   => $threads});

    logcomment("Setup ... ");
    $clientMgr_obj->setup(clients => $client, map => \%clientMap);
    logcomment("mapped ... ");

    if ($clientMgr_obj->is_error()) {
        die("Error msg : $clientMgr_obj->error_msg()\n");
    }
    logcomment("Setup ... DONE");

    logcomment("Establishing connection to client");
    my $host = host($client);
    my $conn = connect("#telnet");

    logcomment("Getting the used size of $mount_point from df -k before sio ");
    my $res1 =
      $conn->execute(command => "df -k $mount_point | awk \'{print \$3}\'");
    logcomment("Result df -k :$res1");
    my @array1 = split(/\n/, $res1);
    if ($array1[1] =~ /([0-9]+)/) {
        $used_size_bfr_sio = $1;
        $Log->comment("Used size before starting sio:$used_size_bfr_sio" . "K");
    } else {
        die("Unable to get used size from df -k before sio");
    }
    logcomment("Starting traffic for 10 minutes");
    $clientMgr_obj->start(id => $id, clients => $client);

    logcomment("Sleeping for 70 sec");
    sleep(70);

    logcomment("Getting the used size of mount_point from df -k after sio ");
    my $res =
      $conn->execute(command => "df -k $mount_point | awk \'{print \$3}\'");
    logcomment("Result df -k after sio:$res");
    my @array = split(/\n/, $res);
    if ($array[1] =~ /([0-9]+)/) {
        $used_size_aft_sio = $1 ;
        logcomment("Used size after sio:$used_size_aft_sio" . "K");
    } else {
        die("Unable to get used size from df -k after sio");
    }
    my $diff = $used_size_aft_sio - $used_size_bfr_sio;
    if (($used_size_aft_sio - $used_size_bfr_sio) > 0) {
        logcomment("Sio done successfully");
        $ret = 1;
    } else {
        die("Sio not done successfully");
    }

    $clientMgr_obj->cleanup(id       => $id,
                            clients  => $client,
                            nfs_path => $mount_point);
    disconnect($conn);
    return $ret;
} ## end sub _run_cm_load

## @Function
##      nfs_setup- Does the NFS setup,mount the volume 
##
##
##
## @usage
##      use Protocol_access qw(nfs_setup);
##      my $ret = nfs_setup (server        => $Server,
##                           vserver       => $vserver,
##                           policyname    => "export_policy_nfs",
##                           domain        => "rtp.netapp.com",
##                           'name-servers'=> 10.227.64.21,
##                           ip            => $lif_ip,
##                           client        => $CLIENT,
##                           'mount-point' => "/t/mnt_pt");
##
## @arguments
##      Server                Required    Server object
##      vserver               Required    Vserver name
##      policyname            Required    Specifies the export policy that is to be created.
##      domain                Required    Domain name
##      name-servers          Required    IP address of DNS servers
##      ip                    Required    lif ip address
##      client                Required    Client on which volume to be mounted
##      mount-point           Required    Path at which volume to be mounted
##      junction-path         Required    Junction path
##      vol-name              Required    Volume creted for data access
##      root-vol              Required    Root volume name
##
## @return-value
##      none 
##      
##
## @scope
##      Exported from this module
##

my $vol_mount=0;
my $nfs_mnt="/t/".(random_name_generate(type =>"mnt"));

sub nfs_setup {
    my %opts = validate(@_,
                        {server          => {type => OBJECT, isa => 'Server'},
                         'client-conn'   => {type => OBJECT,isa=>'Connectrec'},
                         vserver         => {type => SCALAR},
                         policyname      => {type => SCALAR},
                         domain          => {type => SCALAR},
                         'name-servers'  => {type => SCALAR},
                         ip              => {type => SCALAR},
                         'mount-point'   => {type => SCALAR,default => $nfs_mnt},
                         'junction-path' => {type => SCALAR},
                         'vol-name'      => {type => SCALAR},
                         'root-vol'      => {type => SCALAR}});


    my $Server        = $opts{server};
    my $client_conn   = $opts{'client-conn'};
    my $vserver       = $opts{vserver};
    my $policyname    = $opts{policyname};
    my $domain        = $opts{domain};
    my $name_servers  = $opts{'name-servers'};
    my $ip            = $opts{ip};
    my $mount_point   = $opts{'mount-point'};
    my $junction_path = $opts{'junction-path'};
    my $vol           = $opts{'vol-name'};
    my $root_vol      = $opts{'root-vol'};

    my $ret_value = 0;
    eval {


        logcomment("Verifying if NFS license exists..if not adding license");
        eval { $Server->system_license_show(feature => "NFS"); };
        if ($@ =~ /There are no entries matching your query/) {
            logcomment("NFS license not exists creating...");
            $Server->system_license_add('license-code' => LICENSE_CODE_NFS);
        } else {
            logcomment("NFS license already exists");
        }
        
        logcomment("Enabling and configuring NFS on a virtual server $vserver");
        eval {$Server->vserver_nfs_show(vserver => $vserver);};
        if ($@=~ /There are no entries matching your query/) {
            $Server->vserver_nfs_create(vserver => $vserver);
        } else {
            logcomment("Already enabled NFS on a virtual server $vserver");
        }

        eval { $Server->vserver_export_policy_show(vserver => $vserver,
                                                   policyname => $policyname); };
        if ($@=~ /There are no entries matching your query/) {
            logcomment("Creating export-policy ");
            $Server->vserver_export_policy_create(vserver    => $vserver,
                                                  policyname => $policyname);
        }else{
            $Log->comment("Export-policy $policyname already exists for $vserver");
        }
 
        eval { $Server->vserver_export_policy_rule_show(vserver      => $vserver,
                                                        policyname   => $policyname); };
        if ($@ =~ /There are no entries matching your query/){
            logcomment("Creating export-policy rule");
            $Server->vserver_export_policy_rule_create(vserver      => $vserver,
                                                       policyname   => $policyname,
                                                       clientmatch  => CLIENTMATCH,
                                                       rorule       => RORULE,
                                                       rwrule       => RWRULE,
                                                       protocol     => PROTOCOL_NFS,
                                                       anon         => ANON,
                                                       superuser    => SUPERUSER);
        }else{
            $Log->comment("Export-policy rule already exists for $vserver");
        }

        logcomment("Modifying the volume");
        $Server->volume_modify(volume  => $vol,
                               vserver => $vserver,
                               policy  => $policyname);

        $Server->volume_modify(volume             => $root_vol,
                               vserver            => $vserver,
                               policy             => $policyname,
                               'unix-permissions' => '---rwxrwxrwx');

        logcomment("Mounting the Volume ");
        $Server->volume_mount(volume          => $vol,
                              vserver         => $vserver,
                              'junction-path' => $junction_path);
        eval {
            $Server->vserver_services_dns_show(vserver => $vserver,
                                               domains => $domain);
        };
        if ($@ =~ /There are no entries matching your query/) {
            logcomment("Creating dns service");
            $Server->vserver_services_dns_create(vserver        => $vserver,
                                             state          => "enabled",
                                             domains        => $domain,
                                             'name-servers' => $name_servers);
        }else{
            $Log->comment("DNS already configured for $vserver");
        }

        logcomment("Mounting the volume");
         my (undef,undef,$mount_dir)=say($client_conn,"ls $mount_point",'-swallow');
        if($mount_dir =~/No such file or directory/){
            logcomment("Creating mount point $mount_point");
            say($client_conn,"mkdir -p $mount_point");
        }else{
            logcomment("Unmounting if mount point $mount_point already exists");
           say($client_conn,"umount  $mount_point");
        }

        my $cmd="mount $ip:$junction_path $mount_point -o proto=tcp";
        say($client_conn,$cmd,'-swallow');
        my (undef,undef,$mount)=say($client_conn,"mount | grep $mount_point");
        if($mount =~/$mount_point/){
           $vol_mount=1;
           logcomment("Successfully mounted");
        }else{
           die("Failed to mount the volume");
        }
    };
    if ($@) {
        die("Error  while doing NFS setup/mount  Reason:$@");
    }

}
## @Function
##      nfs_io- Does the io on mounted volume 
##
##
##
## @usage
##      use Protocol_access qw(nfs_io);
##      my $ret = nfs_io ('client-conn'   =>$client_conn,
##                        'mount-point    =>$mount_point,
##                        'file-size'     => "80m",
##                        'block-size'    => "4k",
##                        'run-time'      => 60,
##                        bg              => 0 );
##
## @arguments
##      client-conn           Required    Client connection object 
##      mount-point           Required    Path at which volume  mounted
##      file-size             Optional    Size of, file to be created during sio
##      block-size            Optional    Size of block to read/write during sio  
##      run-time              Optional    Duration of io 
##      bg                    Required    1- Sio process run in background else 0 
##
## @return-value
##    1- Success
##    0- Failure 
##
##
## @scope
##      Exported from this module
##
sub nfs_io {
    my %opts = validate(@_,
                        {
                         'client-conn'   => {type => OBJECT,isa=>'Connectrec'},
                         'mount-point'   => {type => SCALAR,default => $nfs_mnt},
                         'file-size'     => {type => SCALAR, optional => 1, default => "3m"},
                         'block-size'    => {type => SCALAR, optional => 1, default => "4k"},
                         'run-time'      => {type => SCALAR, optional => 1, default => "120"},
                          'bg'           => {type => SCALAR}});


    my $client_conn   = $opts{'client-conn'},
    my $mount_point   = $opts{'mount-point'};
    my $file_size     = $opts{'file-size'};
    my $block_size    = $opts{'block-size'};
    my $run_time      = $opts{'run-time'};
    my $bg            = $opts{bg};

    my $return = 0;

    eval {

        if($bg !=0 && $bg !=1){
            die("Accepted values of variable 'bg' are 0 or 1");
        }
        my $sio_file_name=random_name_generate(type =>"siofile");
        my $sio_file1="$mount_point"."/$sio_file_name";
        say($client_conn,"touch $sio_file1");
        my (undef,undef,$file)=say($client_conn,"ls $mount_point");
        if($file=~/$sio_file_name/){
            logcomment("File $sio_file_name created successfully");
        }else{
            logcomment("File $sio_file_name not created successfully");
            return 0; 
        }
        my $cmd1="ls -l $sio_file1 \| awk '{print \$5}'";
        my $cmd2=SIO_PATH." 0 0 $block_size 0 $file_size $run_time 2 $sio_file1 &";
        my $cmd3=SIO_PATH." 0 0 $block_size 0 $file_size $run_time 2 $sio_file1";

        if($bg){
            say($client_conn,"$cmd2");
            sleep(5);
            my (undef,undef,$val1) = say($client_conn,$cmd1,'-swallow');
            sleep(10);
            my (undef,undef,$val2) = say($client_conn,$cmd1,'-swallow');
            logcomment("Values are $val1 and $val2");
            if($val2 > $val1){
                $return =1;
            }
       }else{
            my (undef,undef,$val1) = say($client_conn,$cmd1,'-swallow');
            say($client_conn,"$cmd3",-timeout,10000);
            sleep(5);
            my (undef,undef,$val2) = say($client_conn,$cmd1,'-swallow');
            logcomment("Size of file before and after sio: $val1 and $val2");
            if($val2 > $val1){
                $return =1;
            }
      }
    };
    if ($@) {
        die("Error  while doing NFS IO Reason:$@");
    }
    return ($return);

}

sub nfs_cleanup{
    my %opts = validate(@_, {
                             'client-conn'       => {type => OBJECT,isa=>'Connectrec'},
                             'nfs-mountpt'       => {type => SCALAR,default =>$nfs_mnt }});
                             

    my $client_conn = $opts{'client-conn'};
    my $nfs_mount   = $opts{'nfs-mountpt'};
    eval{
        say($client_conn, "ps -ef \| grep sio");
        my $cmd =
          "ps -ef \| grep sio.*$nfs_mount \|grep -v \"grep\" \|awk '{print \$2}'";
        my (undef, undef, $pid) = say($client_conn, $cmd, '-swallow');
        if ($pid){
            logcomment("killing  nfs sio");
            say($client_conn, "kill -9 $pid", '-swallow');
        }
        sleep(5);
        say($client_conn,"ps -ef \| grep sio.*$nfs_mount",'-swallow');
        if($vol_mount){
           logcomment("Unmounting and deleting mount point");
           say($client_conn,"umount  $nfs_mount");
           sleep(5);

           say($client_conn,"rm -rf $nfs_mount");
           sleep(5);
           say($client_conn,"ls $nfs_mount",'-swallow');
        }
   };
   if($@){
       die("Nfs cleanup failed:Reason:$@");
   }

}

## @Function
##      iscsi_setup- Does the ISCSI setup,mount the lun
##
##
##
## @usage
##      use Protocol_access qw(iscsi_setup);
##      my $ret = iscsi_setup (server        => $Server,
##                             vserver       => $vserver,
##                             policyname    => "export_policy_nfs",
##                             domain        => "csiqa.rtp.netapp.com",
##                             'name-servers'=> "10.227.64.21",
##                             ip            => $lif_ip,
##                             client        => $CLIENT
##                             'mount-point' => "/t/mnt_pt");
##
## @arguments
##      Server                Required    Server object
##      client-conn           Required    Client connection object
##      vserver               Required    Vserver name
##      policyname            Required    Specifies the export policy that is to be created.
##      domain                Required    Domain name
##      name-servers          Required    IP address of DNS servers
##      ip                    Required    lif ip address
##      client                Required    Client on which volume to be mounted
##      mount-point           Required    Path at which volume to be mounted
##      vol-name              Required    Volume creted for data access
##      igroup                Optional    Igroup name
##      lun-size              Optional    Lun size
##      lun-name              Optional    Lun name
##
## @return-value
##      none
##
##
## @scope
##      Exported from this module
##
## @dependency
##  Lun size should be maximum of 1G.
##

my $lun_mount=0;
my $lun_igroup=0;
my $mnt = random_name_generate(type =>"mnt");
my $iscsi_mnt="/t/"."$mnt";

sub iscsi_setup{
    my %opts = validate(@_, {server       => {type => OBJECT,isa=>'Server'},
                             'client-conn'=> {type => OBJECT,isa=>'Connectrec'},
                             vserver      => {type => SCALAR},
                             policyname   => {type => SCALAR, default => "default"},
                             domain       => {type => SCALAR},
                            'name_servers'=> {type => SCALAR},
                             ip           => {type => SCALAR},
                             client       => {type => SCALAR},
                             mount_point  => {type => SCALAR, optional => 1, default => $iscsi_mnt},
                             'vol-name'   => {type => SCALAR},
                             igroup       => {type => SCALAR},
                            'lun-size'    => {type => SCALAR},
                            'lun-name'    => {type => SCALAR} });


  my $Server           = $opts{server};
  my $client_conn      = $opts{'client-conn'};
  my $vserver          = $opts{vserver};
  my $policyname       = $opts{policyname};
  my $domain           = $opts{domain} ;
  my $name_servers     = $opts{"name_servers"};
  my $ip               = $opts{ip};
  my $client           = $opts{client};
  my $mount_point      = $opts{'mount_point'};
  my $igroup           = $opts{'igroup'};
  my $lun_name         = $opts{'lun-name'};
  my $lun_size         = $opts{'lun-size'};
  my $vol              = $opts{'vol-name'};

  my ($return_val,$initiator);
  my $ostype="linux";
  my $protocol="any";

eval {
     my $size_lun;
     if($lun_size=~ /(\d+)[Gg]/){
        $size_lun=$1;
        logcomment("lunsize:$size_lun");
     }
     if($size_lun > 1){
         die("Size of lun should not be greater than 1G");
     }
     eval{$Server->system_license_show(feature      => "iSCSI");};
     if($@ =~ /There are no entries matching your query/){
         logcomment("iSCSI license not exists creating...");
         $Server->system_license_add('license-code' => LICENSE_CODE_ISCSI);
     } else {
         logcomment("iSCSI license already exists");
     }
     eval { $Server->vserver_export_policy_rule_show(vserver      => $vserver,
                                                     policyname   => $policyname); };
     if ($@ =~ /There are no entries matching your query/){
         logcomment("Creating export policy rule");
         $Server->vserver_export_policy_rule_create(vserver      => $vserver,
                                                    policyname   => $policyname,
                                                    clientmatch  => CLIENTMATCH,
                                                    rorule       => RORULE,
                                                    rwrule       => RWRULE,
                                                    protocol     => $protocol,
                                                    anon         => ANON,
                                                    superuser    => SUPERUSER);
    }else{
         $Log->comment("Export-policy rule already exists for $vserver");
    }

    eval {
       $Server->vserver_services_dns_show(vserver => $vserver,
                                          domains => $domain);
    };
    if ($@ =~ /There are no entries matching your query/) {
        logcomment("Creating dns service");
        $Server->vserver_services_dns_create(vserver        => $vserver,
                                             state          => "enabled",
                                             domains        => $domain,
                                            'name-servers'  => $name_servers );
    }else{
        $Log->comment("DNS already configured for $vserver");
    }

    logcomment("Taking backup of /etc/iscsi.conf");
    $client_conn->execute(
                        command => "cp -f /etc/iscsi.conf /etc/iscsi_conf_bkup",
                        prompts => ["$client"]);
    logcomment("Getting the Initiator name");
    $return_val =
          $client_conn->execute(command => "cat /etc/initiatorname.iscsi");

    if ($return_val =~ /InitiatorName=(.*)/s) {
        $initiator = $1;
        chomp $initiator;
        logcomment("iSCSI nodename of $client is: $initiator");
    }

    logcomment("Enable and configure ISCSI on virtual server");
    $Server->vserver_iscsi_create( vserver       =>$vserver,
                                  'target-alias' =>$vserver,
                                  'status-admin' =>"up");
     eval{
        $Server->vserver_iscsi_show(vserver =>$vserver);
        };
     if ($@){
        die(" Failure in configuring iSCSI on virtual server:$@");
     }

     logcomment("Creating igroup $igroup");
     my $cmd="lun igroup create -vserver $vserver -igroup $igroup -protocol iscsi  -ostype $ostype -initiator $initiator";
     $Server->cli(cmd=>"$cmd");
    eval{
       $Server->lun_igroup_show(vserver =>$vserver,igroup =>$igroup,initiator=>$initiator);
       $lun_igroup=1;
        };
     if ($@){
        die("Failure in creating igroup:$@");
     }
     logcomment("Creating lun");
     $Server->lun_create(vserver =>$vserver,volume=>$vol,lun=>$lun_name,size=>$lun_size,ostype=>$ostype,"-Verification" => "0");
    eval{
        $Server->lun_show(vserver =>$vserver,volume=>$vol, lun=>$lun_name);
        };
     if ($@){
        die("Failure in creating lun:$@");
     }

        logcomment("Mapping lun to igroup");
        $Server->lun_map(vserver =>$vserver,volume=>$vol,lun=>$lun_name,igroup=>$igroup);
        my $verify_map= $Server->lun_show(vserver =>$vserver,volume=>$vol,lun=>$lun_name);
        if($verify_map->[0]{mapped}=~/mapped/){
            logcomment("Lun $lun_name mapped to igroup $igroup");
        }else {
            die("Lun $lun_name  not mapped to igroup $igroup");
        }

        #-----------------------------------------------------------------------
        #   Giving Vserver lif IP address for Discovery IP address on
        #           /etc/iscsi.conf file and Restart iSCSI service in Client
        #-----------------------------------------------------------------------
        logcomment(  "Adding lif IP address address $ip"
                   . " in '/etc/iscsi.conf' file of Client($client) and "
                   . "restarting.. It should successfully restart the iSCSI "
                   . "service..");
        my $Discovery =
            "\nContinuous=no"
          . "\nHeaderDigest=never"
          . "\nDataDigest=never"
          . "\nImmediateData=yes"
          . "\nDiscoveryAddress=$ip";
        $client_conn->execute(
                           command => "echo \"$Discovery\" >> /etc/iscsi.conf");
        my $return_val =
          $client_conn->execute(
                     command => "grep \"$ip\" /etc/iscsi.conf",
                     prompts => ["$client"]);
        if ($return_val =~ /$ip/s) {
            logcomment("Discovery Address($ip) is added in"
                          . " iSCSI configuration file of client($client)...");
        } else {
            die "Failed to add Discovery Address($ip) in "
              . "iSCSI configuration file of client($client)...";
        }
        $client_conn->execute(command => "service iscsi stop",
                              prompts => ["$client"]);
        $client_conn->execute(command => "service iscsi start",
                              prompts => ["$client"]);
        $return_val =
          $client_conn->execute(command => "service iscsi status",
                                prompts => ["$client"]);
        if ($return_val =~ /running/s) {
            logcomment(
                        "iSCSI service is restarted in the client($client)...");
        } else {
            die "Failed to restart iSCSI service in the client($client)...";
        }
        #-----------------------------------------------------------------------
        #    Invoking 'sanlun lun show'
        #-----------------------------------------------------------------------
        logcomment(  "Invoking 'sanlun lun show' in Client($client).. "
                      . "It should list the vserver LUN");
        $return_val =
          $client_conn->execute(command => "/usr/sbin/sanlun lun show",
                                prompts => ["$client"]);
        my $device_path;
        if ($return_val =~ /$vserver.*$lun_name(\s*)(\/dev\/\w+)\s/s) {
            $device_path = $2;
            logcomment("LUN($lun_name) discovered from vserver "
                          . "and it is in the device($device_path)...");
        } else {
            die "Failed to discover the LUN($lun_name) in the client"
              . "($client)...";
        }

        #-----------------------------------------------------------------------
        #    Creating a file system on raw device file.
        #            Using mkfs -t ext2 /dev/xxx .
        #-----------------------------------------------------------------------
        logcomment( "Creating a file system on raw device file"
                     . "($device_path) and mounting the filesystem.. It should "
                     . "mount it successfully in $mount_point..");
        $return_val =
          $client_conn->execute(command     => "mkfs -t ext2 $device_path",
                                match_table => ["roceed anyway" => "y",],
                                prompts     => ["$client"],
                                timeout     => 70000);
        sleep(10);
        $client_conn->execute(  command =>
                               " mkdir -p $mount_point;"
                              . " mount -f $device_path $mount_point");

        sleep(5);
        $return_val =
          $client_conn->execute(command => 'mount | grep ' . "$mount_point",
                                prompts => ["$client"]);
        if ($return_val =~ /$mount_point/) {
            $lun_mount=1;
            logcomment("$lun_name is mounted on the client machine");
        } else {
            die " Failed to mount $lun_name on the client machine";
        }

 };
 if($@){
    my $temp=$@;
    logcomment("Calling cleanup");
    iscsi_cleanup(server=>$Server,'client-conn'=>$client_conn,'iscsi-mountpt'=>$mount_point,vserver=>$vserver,igroup=>$igroup);
    die("Error:Reason:$temp");
 }
}
## @Function
##      iscsi_sio- Does the io on mount point using sio
##
##
##
## @usage
##      use Protocol_access qw(iscsi_io);
##      my $ret =   iscsi_io ('client-conn' => $client_conn ,
##                            'mount-point' => $mount_pt,
##                             bg           => 1,
##                             'file-size'  => "20m",
##                             'block-size' => "4k",
##                             'run-time'   => 600
##                            );
##
## @arguments
##      client-conn           Required    Client connection object
##      mount-point           Optional    Path at which lun mounted
##      file-size             Optional    Size of, file to be created during sio
##      block-size            Optional    Size of block to read/write during sio
##      run-time              Optional    Number of seconds to read/write to file
##      bg                    Required    1- IO process should run in background else 0
##
##
## @return-value
##      0 on failure in data access
##      1 on successful data access
##
## @scope
##      Exported from this module
##
## @dependency
##
sub iscsi_io{
  my %opts = validate(@_, {
                             'client-conn'=> {type => OBJECT,isa=>'Connectrec'},
                             'mount-point'=> {type => SCALAR, default => $iscsi_mnt},
                             'file-size'  => {type => SCALAR, default => "80m"},
                             'block-size' => {type => SCALAR, default => "4k"} ,
                             'run-time'   => {type => SCALAR, default => "120"},
                             'bg'         => {type => SCALAR}});

  my $mount_point      = $opts{'mount-point'};
  my $client_conn      = $opts{'client-conn'};
  my $file_size        = $opts{'file-size'};
  my $block_size       = $opts{'block-size'};
  my $run_time         = $opts{'run-time'};
  my $bg               = $opts{bg};

  my $return=0;

  eval{
      if($bg !=0 && $bg !=1){
          die("Accepted values of variable 'bg' are 0 or 1");
      }
      my $sio_file_name=random_name_generate(type =>"siofile");
      my $sio_file1="$mount_point"."/$sio_file_name";
      say($client_conn,"touch $sio_file1");
      my (undef,undef,$file)=say($client_conn,"ls $mount_point");
      if($file=~/$sio_file_name/){
          logcomment("File $sio_file_name created successfully");
      }else{
          logcomment("File $sio_file_name not created successfully");
          return 0; 
      }

      my $cmd1="ls -l $sio_file1 \| awk '{print \$5}'";
      my $cmd2=SIO_PATH." 0 0 $block_size 0 $file_size $run_time 2 $sio_file1 &";
      my $cmd3=SIO_PATH." 0 0 $block_size 0 $file_size $run_time 2 $sio_file1";

      if($bg){
          say($client_conn,"$cmd2");
          sleep(5);
          my (undef,undef,$val1) = say($client_conn,$cmd1,'-swallow');
          sleep(10);
          my (undef,undef,$val2) = say($client_conn,$cmd1,'-swallow');
          if($val2 > $val1){
             $return =1;
          }
     }else{
          my (undef,undef,$val1) = say($client_conn,$cmd1,'-swallow');
          say($client_conn,"$cmd3",-timeout,10000);
          sleep(5);
          my (undef,undef,$val2) = say($client_conn,$cmd1,'-swallow');
          if($val2 > $val1){
             $return =1;
          }
     }
 };
 if($@){
    my $temp=$@;
    die("Error during io:Reason:$temp");
 }
 return $return;
}

## @Function
##      iscsi_cleanup- Does the cleanup like deleting igroup,unmounting mount point,kill the io process
##
##
##
## @usage
##      use Protocol_access qw(iscsi_cleanup);
##      my $ret =iscsi_cleanup(server         => $Server,
##                            'client-conn'   => $client_conn ,
##                            'iscsi-mountpt' => $mount_pt,
##                             vserver        => $vserver_name,
##                             'igroup'       => $igroup_name
##                            );
##
## @arguments
##      server                Required    Server object
##      client-conn           Required    Client connection object
##      iscsi-mountpt         Optional    Path at which lun  mounted
##      vserver               Required    Vserver name
##      igroup                Required    Igroup name
##
##
## @return-value
##     none
##
## @scope
##      Exported from this module
##
## @dependency
##


sub iscsi_cleanup{
    my %opts = validate(@_, {server              => {type => OBJECT,isa=>'Server'},
                             'client-conn'       => {type => OBJECT,isa=>'Connectrec'},
                             'iscsi-mountpt'     => {type => SCALAR,default =>$iscsi_mnt },
                             'vserver'           => {type => SCALAR},
                             'igroup'            => {type => SCALAR}});

    my $Server      = $opts{server};
    my $client_conn = $opts{'client-conn'};
    my $iscsi_mount = $opts{'iscsi-mountpt'};
    my $vserver     = $opts{vserver};
    my $igroup      = $opts{igroup};

    eval{
        logcomment("Retrieving the original /etc/iscsi.conf file");
        $client_conn->execute(
                        command => "mv -f /etc/iscsi_conf_bkup /etc/iscsi.conf");

        say($client_conn, "ps -ef \| grep sio");
        my $cmd =
          "ps -ef \| grep sio.*$iscsi_mount \|grep -v \"grep\" \|awk '{print \$2}'";
        my (undef, undef, $pid) = say($client_conn, $cmd, '-swallow');
        if ($pid){
            logcomment("killing sio");
            say($client_conn, "kill -9 $pid", '-swallow');
        }

        if($lun_mount){
            logcomment("Unmounting and deleting mount point");
            say($client_conn,"umount $iscsi_mount");
            sleep(2);

            say($client_conn,"rm -rf $iscsi_mount");
        }
        if($lun_igroup){
            logcomment("Deleting igroup $igroup");
            $Server->lun_igroup_delete(vserver => $vserver,igroup=>$igroup,force=>'true');
        }
   };
   if($@){
       die("Iscsi cleanup failed:Reason:$@");
   }

}
1;

