# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Cmode/Quota.pm#1 $
#
# $Change: 1917575 $
#
# $DateTime: 2013/01/22 11:22:46 $
#
# Copyright (c) 2009 Network Appliance, Inc.
# All rights reserved.
#
#
# NAME
#      Quota.pm - cmode quota library
#
# DESCRIPTION
#      This perl module provide wrapper interfaces to c-mode 
#       volume quota cli command
#
# AUTHOR
#       tarung@netapp.com
#
#
## @pod here
##
##

=pod

=head1 NAME

    Quota.pm - provide wrapper interfaces to C-Mode volume-quota cli command

=head1 SYNOPSIS

    use Cmode::Quota qw(:all);
    OR
    use Cmode::Quota qw(:quota);

=head1 DESCRIPTION

=head1 FUNCTIONS

Cmode Volume Quota Functions

=cut

package Cmode::Quota ;

#------------------------------------------------------
#   Module Vars
#------------------------------------------------------
our ($Validation, $VERSION);

#
# Define constants
#
use constant QUOTA_ERROR => undef;
use constant QUOTA_TRUE  => 1;
use constant QUOTA_FALSE => 0;

use warnings;
use strict qw(vars subs refs);
use base   qw(Exporter);

BEGIN {

    #--------------------------------------------
    #   Exports
    #--------------------------------------------

    $Validation = 1; # unused as of now
    $VERSION    = '0.01';

    my @quota_sub = qw(
                        &quota_modify
                        &quota_policy_copy
                        &quota_policy_create
                        &quota_policy_delete
                        &quota_policy_rename
                        &quota_policy_show
                        &quota_policy_rule_create
                        &quota_policy_rule_delete
                        &quota_policy_rule_modify
                        &quota_policy_rule_show
                        &quota_report
                        &quota_show
                    );

    my @util_sub = qw(
                        &parse_cmode_cli_op
                        &wait_for_quota_initialization
                    );

    my @constants = qw(QUOTA_ERROR QUOTA_TRUE QUOTA_FALSE);

    our %EXPORT_TAGS = (
        'quota'      => [ @quota_sub ],
        'util'       => [ @util_sub ],
        'constants'  => [ @constants ],
    );

    $EXPORT_TAGS{all} = [ map{ @{$_} } values %EXPORT_TAGS ];

    Exporter::export_ok_tags('all');
}

#------------------------------------------------------
#   Imports
#------------------------------------------------------
use Tharn;
use Data::Dumper;
use Params::Validate qw(validate SCALAR ARRAYREF SCALARREF OBJECT HASHREF);

#------------------------------------------------------
#     Prototypes
#------------------------------------------------------
sub quota_modify ;
sub quota_policy_copy ;
sub quota_policy_create ;
sub quota_policy_delete ;
sub quota_policy_rename ;
sub quota_policy_show ;
sub quota_policy_rule_create ;
sub quota_policy_rule_delete ;
sub quota_policy_rule_modify ;
sub quota_policy_rule_show ;
sub quota_report ;
sub quota_show ;

#
# utility functions
#
sub parse_cmode_cli_op;
sub wait_for_quota_initialization;


#------------------------------------------------------
#     Local Variables
#------------------------------------------------------


#------------------------------------------------------
#     Subroutines Definition
#------------------------------------------------------

=head2 quota_modify()

modify quota state for given flexvol and vserver

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub quota_modify {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
        );

    my %args_validn_spec =
        (
            'server'           => {type => OBJECT},
            'vserver'          => {type => SCALAR},
            'volume'           => {type => SCALAR},
            'state'            => {type => SCALAR, optional => 1},
            'logging'          => {type => SCALAR, optional => 1},
            'logging-interval' => {type => SCALAR, optional => 1},
            'foreground'       => {type => SCALAR, optional => 1},
            '-Verification'    => {type => SCALAR, default  => 1},
            'errmsg'           => {type => SCALARREF},
            '-Timeout'         => {type => SCALAR, default  => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_quota_modify( %input_arg );
    };

    if($@) {
        $$err_ref = "quota_modify() failed, error: $@";
        return QUOTA_FALSE;
    }

    return QUOTA_TRUE;
}


=head2 quota_policy_copy()

copy quota policy in given vserver

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub quota_policy_copy {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
        );

    my %args_validn_spec =
        (
            'server'           => {type => OBJECT},
            'vserver'          => {type => SCALAR},
            'policy-name'      => {type => SCALAR},
            'new-policy-name'  => {type => SCALAR},
            'errmsg'           => {type => SCALARREF},
            '-Verification'    => {type => SCALAR, default => 1},
            '-Timeout'         => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_quota_policy_copy( %input_arg );
    };

    if($@) {
        $$err_ref = "quota_policy_copy() failed, error: $@";
        return QUOTA_FALSE;
    }

    return QUOTA_TRUE;
}

=head2 quota_policy_create()

create quota policy in given flexvol and vserver 

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub quota_policy_create {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
        );

    my %args_validn_spec =
        (
            'server'        => {type => OBJECT},
            'vserver'       => {type => SCALAR},
            'policy-name'   => {type => SCALAR, optional => 1},
            'errmsg'        => {type => SCALARREF},
            '-Verification' => {type => SCALAR, default => 1},
            '-Timeout'      => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_quota_policy_create( %input_arg );
    };

    if($@) {
        $$err_ref = "quota_policy_create() failed, error: $@";
        return QUOTA_FALSE;
    }

    return QUOTA_TRUE;
}

=head2 quota_policy_delete()

delete quota policy in given vserver

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub quota_policy_delete {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
        );

    my %args_validn_spec =
        (
            'server'        => {type => OBJECT},
            'vserver'       => {type => SCALAR},
            'policy-name'   => {type => SCALAR, optional => 1},
            'errmsg'        => {type => SCALARREF},
            '-Verification' => {type => SCALAR, default => 1},
            '-Timeout'      => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_quota_policy_delete( %input_arg );
    };

    if($@) {
        $$err_ref = "quota_policy_delete() failed, error: $@";
        return QUOTA_FALSE;
    }

    return QUOTA_TRUE;
}

=head2 quota_policy_rename()

rename quota policy in given vserver

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub quota_policy_rename {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
        );

    my %args_validn_spec =
        (
            'server'           => {type => OBJECT},
            'vserver'          => {type => SCALAR},
            'policy-name'      => {type => SCALAR},
            'new-policy-name'  => {type => SCALAR},
            'errmsg'           => {type => SCALARREF},
            '-Verification'    => {type => SCALAR, default => 1},
            '-Timeout'         => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_quota_policy_rename( %input_arg );
    };

    if($@) {
        $$err_ref = "quota_policy_rename() failed, error: $@";
        return QUOTA_FALSE;
    }

    return QUOTA_TRUE;
}

=head2 quota_policy_show()

show quota policy

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub quota_policy_show {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            cli_op  => 1,
        );

    my %args_validn_spec =
        (
            'server'        => {type => OBJECT},
            'vserver'       => {type => SCALAR, optional => 1},
            'policy-name'   => {type => SCALAR, optional => 1},
            'policyid'      => {type => SCALAR, optional => 1},
            'last-modified' => {type => SCALAR, optional => 1},
            'errmsg'        => {type => SCALARREF},
            'cli_op'        => {type => ARRAYREF},
            '-Timeout'      => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        @{$cmdOpRef} = $server->volume_quota_policy_show( %input_arg );
    };

    if($@) {
        $$err_ref = "quota_policy_show() failed, error: $@";
        return QUOTA_FALSE;
    }

    return QUOTA_TRUE;
}


=head2 quota_policy_rule_create()

create rule for given quota policy and volume

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub quota_policy_rule_create {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
        );

    my %args_validn_spec =
        (
            'server'           => {type => OBJECT},
            'vserver'          => {type => SCALAR},
            'policy-name'      => {type => SCALAR},
            'volume'           => {type => SCALAR},
            'type'             => {type => SCALAR},
            'target'           => {type => SCALAR},
            'user-mapping'     => {type => SCALAR, optional => 1},
            'disk-limit'       => {type => SCALAR, optional => 1},
            'file-limit'       => {type => SCALAR, optional => 1},
            'threshold'        => {type => SCALAR, optional => 1},
            'soft-disk-limit'  => {type => SCALAR, optional => 1},
            'soft-file-limit'  => {type => SCALAR, optional => 1},
            'qtree'            => {type => SCALAR, optional => 1},
            'errmsg'           => {type => SCALARREF},
            '-Verification'    => {type => SCALAR, default => 1},
            '-Timeout'         => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_quota_policy_rule_create( %input_arg );
    };

    if($@) {
        $$err_ref = "quota_policy_rule_create() failed, error: $@";
        return QUOTA_FALSE;
    }

    return QUOTA_TRUE;
}


=head2 quota_policy_rule_delete()

delete rule for given quota policy and volume

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub quota_policy_rule_delete {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
        );

    my %args_validn_spec =
        (
            'server'           => {type => OBJECT},
            'vserver'          => {type => SCALAR},
            'policy-name'      => {type => SCALAR},
            'volume'           => {type => SCALAR},
            'type'             => {type => SCALAR},
            'target'           => {type => SCALAR},
            'errmsg'           => {type => SCALARREF},
            '-Verification'    => {type => SCALAR, default => 1},
            '-Timeout'         => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_quota_policy_rule_delete( %input_arg );
    };

    if($@) {
        $$err_ref = "quota_policy_rule_delete() failed, error: $@";
        return QUOTA_FALSE;
    }

    return QUOTA_TRUE;
}

=head2 quota_policy_rule_modify()

modify rule for given quota policy and volume

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub quota_policy_rule_modify {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
        );

    my %args_validn_spec =
        (
            'server'           => {type => OBJECT},
            'vserver'          => {type => SCALAR},
            'policy-name'      => {type => SCALAR},
            'volume'           => {type => SCALAR},
            'type'             => {type => SCALAR},
            'target'           => {type => SCALAR},
            'disk-limit'       => {type => SCALAR, optional => 1},
            'file-limit'       => {type => SCALAR, optional => 1},
            'threshold'        => {type => SCALAR, optional => 1},
            'soft-disk-limit'  => {type => SCALAR, optional => 1},
            'soft-file-limit'  => {type => SCALAR, optional => 1},
            'qtree'            => {type => SCALAR, optional => 1},
            'errmsg'           => {type => SCALARREF},
            '-Verification'    => {type => SCALAR, default => 1},
            '-Timeout'         => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_quota_policy_rule_modify( %input_arg );
    };

    if($@) {
        $$err_ref = "quota_policy_rule_modify() failed, error: $@";
        return QUOTA_FALSE;
    }

    return QUOTA_TRUE;
}

=head2 quota_policy_rule_show()

show quota rule

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub quota_policy_rule_show {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
        );

    my %args_validn_spec =
        (
            'server'           => {type => OBJECT},
            'vserver'          => {type => SCALAR, optional => 1},
            'policy-name'      => {type => SCALAR, optional => 1},
            'volume'           => {type => SCALAR, optional => 1},
            'type'             => {type => SCALAR, optional => 1},
            'target'           => {type => SCALAR, optional => 1},
            'disk-limit'       => {type => SCALAR, optional => 1},
            'file-limit'       => {type => SCALAR, optional => 1},
            'threshold'        => {type => SCALAR, optional => 1},
            'soft-disk-limit'  => {type => SCALAR, optional => 1},
            'soft-file-limit'  => {type => SCALAR, optional => 1},
            'errmsg'           => {type => SCALARREF},
            '-Verification'    => {type => SCALAR, default => 1},
            '-Timeout'         => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_quota_policy_rule_show( %input_arg );
    };

    if($@) {
        $$err_ref = "quota_policy_rule_show() failed, error: $@";
        return QUOTA_FALSE;
    }

    return QUOTA_TRUE;
}

=head2 quota_report()

show quota report for volumes

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub quota_report {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            cli_op  => 1,
        );

    my %args_validn_spec =
        (
            'server'           => {type => OBJECT},
            'vserver'          => {type => SCALAR, optional => 1},
            'volume'           => {type => SCALAR, optional => 1},
            'index'            => {type => SCALAR, optional => 1},
            'tree'             => {type => SCALAR, optional => 1},
            'quota-type'       => {type => SCALAR, optional => 1},
            'quota-target'     => {type => SCALAR, optional => 1},
            'disk-used'        => {type => SCALAR, optional => 1},
            'disk-limit'       => {type => SCALAR, optional => 1},
            'files-used'       => {type => SCALAR, optional => 1},
            'file-limit'       => {type => SCALAR, optional => 1},
            'threshold'        => {type => SCALAR, optional => 1},
            'soft-disk-limit'  => {type => SCALAR, optional => 1},
            'soft-file-limit'  => {type => SCALAR, optional => 1},
            'quota-specifier'  => {type => SCALAR, optional => 1},
            'errmsg'           => {type => SCALARREF},
            'cli_op'           => {type => ARRAYREF},
            '-Timeout'         => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        my $cli_op   = $server->volume_quota_report( %input_arg );

        if ( parse_cmode_cli_op(
                'string'  => $cli_op,
                'parseop' => $cmdOpRef) == QUOTA_FALSE) {
            die("cli op parsing failed");
        }
    };

    if($@) {
        $$err_ref = "quota_report() failed, error: $@";
        return QUOTA_FALSE;
    }

    return QUOTA_TRUE;
}

=head2 quota_show()

show quota state for volumes

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub quota_show {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            cli_op  => 1,
        );

    my %args_validn_spec =
        (
            'server'           => {type => OBJECT},
            'vserver'          => {type => SCALAR, optional => 1},
            'volume'           => {type => SCALAR, optional => 1},
            'state'            => {type => SCALAR, optional => 1},
            'scan-status'      => {type => SCALAR, optional => 1},
            'logging'          => {type => SCALAR, optional => 1},
            'logging-interval' => {type => SCALAR, optional => 1},
            'sub-status'       => {type => SCALAR, optional => 1},
            'last-error'       => {type => SCALAR, optional => 1},
            'errors'           => {type => SCALAR, optional => 1},
            'errmsg'           => {type => SCALARREF},
            'cli_op'           => {type => ARRAYREF},
            '-Timeout'         => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        @{$cmdOpRef} = $server->volume_quota_show( %input_arg );
    };

    if($@) {
        $$err_ref = "quota_show() failed, error: $@";
        return QUOTA_FALSE;
    }

    return QUOTA_TRUE;
}


=head2 parse_cmode_cli_op()

parse cmode cli output and return the array

=over 4

=item Arguments:

=over 4

takes a hash as input with following key

    string   => cli output string to be parsed 
    parseop  => array reference to return parsed output

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub parse_cmode_cli_op {

    my %options = validate(@_, {
                                 'string'  => {type => SCALAR},
                                 'parseop' => {type => ARRAYREF},
                               }
                    );

    my $string  = $options{'string'}; # input string

    my @op  = split ('\n', $string); # split the string and save it in o/p

    if ( scalar(@op) < 3 ) {
        logcomment("cli op is either incomplete or not in proper format");
        return QUOTA_FALSE;
    }

    my @key = split('X;X', shift @op); # hash key from o/p

    shift @op; # delete 2nd element in @op, it is useless

    my (@parseOp, @value);

    foreach my $i ( 0 .. $#op ) {
        @value = split('X;X', $op[$i]);
        foreach my $x (@key) {
           $parseOp[$i]->{$x} = shift @value;
        }
    }

    @{$options{'parseop'}} = @parseOp;

    return QUOTA_TRUE;
}

=head2 wait_for_quota_initialization

wait until quota is initialized for given volume

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub wait_for_quota_initialization {

    my %args_validn_spec =
        (
            server       => {type => OBJECT},
            vserver      => {type => SCALAR},
            volume       => {type => SCALAR},
            errmsg       => {type => SCALARREF},
            interval     => {type => SCALAR, default  => 1},
            timeout      => {type => SCALAR, default  => 300},
            status_info  => {type => HASHREF,optional =>1},
        );

    my %options  = validate(@_, \%args_validn_spec);

    my $server   = $options{'server'};  # server object
    my $vserver  = $options{'vserver'}; # vserver name
    my $volume   = $options{'volume'};   # volume name
    my $timeout  = $options{'timeout'}; # timeout for wait 
    my $err_ref  = $options{'errmsg'};  # reference error msg variable
    my $interval = $options{'interval'}; # interval between two calls

    my $err_msg;
    my @quota_status;

    # undef the job_info hash, if available
    if ( exists $options{'status_info'} ) {
        %{$options{'status_info'}} = ();
    }

    my $finish = time() + $timeout;

    while ( time() <=  $finish ) {

        if ( QUOTA_FALSE == quota_show(
                            server   => $server,
                            vserver  => $vserver,
                            volume   => $volume,
                            errmsg   => \$err_msg,
                            cli_op   => \@quota_status) ) {
                    
            $$err_ref = "quota_show() failed, error: $err_msg";
            return QUOTA_FALSE;
        }

        if ( exists $options{'status_info'} ) {
            %{$options{'status_info'}} = %{$quota_status[0]};
        }

        if ( $quota_status[0]->{'state'} eq 'initializing' ) {
            logcomment("Quota initialization is still active");
            sleep($interval);
        } else {
            logcomment("Quota initialization is not active");
            return QUOTA_TRUE;
        }
    }

    #
    # we are here due to timeout
    #
    $$err_ref = "quota init not completed in $timeout seconds";
    return QUOTA_FALSE;
}


1;


=head1 SEE ALSO

    None

=head1 STATUS

    Shared

=head1 AUTHOR

    Tarun Garg (tarung@netapp.com)
    $Id: //depot/prod/test/Rsierranevada/lib/Cmode/Quota.pm#1 $
    Last $Change: 1917575 $
    Last Revised $DateTime: 2013/01/22 11:22:46 $

=head1 HISTORY

    2010-02-23 author: tarung      original lib created

=head1 COPYRIGHT

Copyright 2009, Network Appliance., All Rights Reserved

=cut

