# $Id: //depot/prod/test/Rsierranevada/lib/Cmode/RbacCmdFetcher.pm#1 $
## @summary
##  RbacCmdFetcher - Fetches the allowed/denied commands from the commandset 
##                   according to the role specified from the commandset file.
##
## @description
##	Processes the vserver commandset file and forms a hash, from which the 
##  allowed/denied commands can be fetched using role and other optional 
##  arguments
##
## @keywords 
##  cmode
##
## @status 
##  public
##
## @dependencies
##  1) commandset flatfile should be available at the location specified in
##     constant 'COMMANDSFILE'
##	
## @author 
##  jaganj@netapp.com
##
package Cmode::RbacCmdFetcher;

use strict;
use warnings;
use Tharn;
use Params::Validate qw(validate SCALAR);
use constant COMMANDSFILE => "/x/eng/tools/test/test_update/lib/Cmode/flatfile.txt";
use base qw(Exporter);

BEGIN {
	our @EXPORT = qw( fetchCmds validateFlatfile);
}

## @Function
##   fetchCmds - Processes the flatfile to form hash and returns the 
##               allowed/denied commands according to the arguments received
##
## @synopsis
##      use Cmode::RbacCmdFetcher qw(fetchCmds);
##
##      my $ref     = fetchCmds(role=>'vsadmin-readonly');
##      my @allowed = @{$ref->{allowed}};
##      my @denied  = @{$ref->{denied}};
##
##      my $ref     = fetchCmds(role=>'vsadmin-readonly', type=>'allowed');
##      my @allowed = @{$ref->{allowed}};
##
##      my $ref = fetchCmds(role        =>'vsadmin-readonly',
##                          type        =>'allowed'
##                          commandset  =>'vserver'
##                          object      =>'vserver');
##      my @allowed = @{$ref->{allowed}};
##
## @usage
##      use Cmode::RbacCmdFetcher qw(fetchCmds);
##      my $ref = fetchCmds(role        =>'<role-name>',
##                          type        =>'<allowed|denied|both>'
##                          commandset  =>'<commandset-name>'
##                          object      =>'<object-name>'
##                          flatfile    =>'<absolute-path-of-commandset-file>'
##
## @arguments
##      role                Required    Role name of the user in vserver
##      type                Optional    type required (allowed|denied|both)
##      commandset          Optional    Commandset name
##      object              Optional    Object name
##      flatfile            Optional    To specify the user-defined commandset 
##                                      file path
##
## @return-value
##  arrayref in hashref or NULL on failure
##
## @scope
##  Exported from this module
##

sub fetchCmds {
    #---------------------------------------------------------------------------
    #   Sub Name      : fetchCmds()
    #   Functionality : Sub routine to return the requested allowed/denied 
    #                   commands depends on the argument.
    #   Inputs        : role=>'<role-name>',
    #                   type=>'<allowed|denied>'
    #                   commandset=>'<commandset-name>'
    #                   object=>'<object-name>'
    #   Return Value  : arrayref in hashref,
    #                   NULL on failure
    #   Scope         : Exported from this module
    #---------------------------------------------------------------------------
    my ($arg_role, $arg_type, $arg_cmdset, $arg_object, $arg_flatfile);
    my %opts = validate(@_, {role       => {type => SCALAR},
                             type       => {type => SCALAR, optional => 1, default => "both"},
                             commandset => {type => SCALAR, optional => 1, default => ""},
                             object     => {type => SCALAR, optional => 1, default => ""},
                             flatfile   => {type => SCALAR, optional => 1, default => COMMANDSFILE},});
    $arg_role     = $opts{'role'};
    $arg_type     = $opts{'type'};
    $arg_cmdset   = $opts{'commandset'};
    $arg_object   = $opts{'object'};
    $arg_flatfile = $opts{'flatfile'};
    die "No role specified as given in the argument role($arg_role)"
      if ($arg_role !~ /^(vsadmin|vsadmin-volume|vsadmin-protocol|vsadmin-readonly)$/);
    die "No access type specified as given in the argument type($arg_type)"
      if ($arg_type !~ /^(allowed|denied|both)$/);
    my @allowed = ();
    my @denied  = ();
    eval {
        my $h = &_readCommandDescriptionFile($arg_flatfile);
        foreach my $cmdset (keys(%{$h})) {
            ($cmdset = $arg_cmdset) if ($arg_cmdset ne "");
            foreach my $object (keys(%{$h->{$cmdset}})) {
                ($object = $arg_object) if ($arg_object ne "");
                foreach my $verb (keys(%{$h->{$cmdset}->{$object}})) {
                    if ($h->{$cmdset}->{$object}->{$verb}->{"$arg_role"} == 1) {
                        if ($arg_type =~ /^(allowed|both)$/) {
                            for (my $i = 0 ; $i < $#{$h->{$cmdset}->{$object}->{$verb}->{'commands'}} ; $i++) {
                                if($object =~ /misc/) {
                                    push(@allowed, $h->{$cmdset}->{$object}->{$verb} ->{'commands'}[$i]);
                                } else {
                                    push(@allowed, "$object " . $h->{$cmdset}->{$object}->{$verb} ->{'commands'}[$i]);
                                }
                            }
                        } ## end if ($arg_type =~ /^(allowed|both)$/)
                    } else {
                        if ($arg_type =~ /^(denied|both)$/) {
                            for (my $i = 0 ; $i < $#{$h->{$cmdset}->{$object}->{$verb}->{'commands'}}; $i++) {
                                if($object =~ /misc/) {
                                    push(@denied, $h->{$cmdset}->{$object}->{$verb} ->{'commands'}[$i]);
                                } else {
                                    push(@denied, "$object " . $h->{$cmdset}->{$object}->{$verb} ->{'commands'}[$i]);
                                }
                            }
                        }
                    } ## end else [ if ($h->{$cmdset}->{$object...
                } ## end foreach my $verb (keys(%{$h...
                last if ($arg_object ne "");
            } ## end foreach my $object (keys(%{...
            last if ($arg_cmdset ne "");
        } ## end foreach my $cmdset (keys(%{...
    };
    if ($@) {
        logcomment ("fetchCmds() sub-routine failed with the error: $@");
        return;
    }
    return {allowed => \@allowed} if ($arg_type =~ /^allowed$/);
    return {denied  => \@denied}  if ($arg_type =~ /^denied$/);
    return {allowed => \@allowed, denied => \@denied} if ($arg_type =~ /^both$/);
} ## end sub fetchCmds

##
## @Function
##   _readCommandDescriptionFile - to build the data structure from the
##                                 flat file mentioned in constant 
##                                 'COMMANDSFILE'.
##
## @synopsis
##
##  my $ref     = _readCommandDescriptionFile();
##
## @arguments
##  flatfilelocation   Required    Absolute location of the commandset flatfile
##                                 or else it will take it from COMMANDSFILE 
##                                 constant.
##
## @return-value
##  Datastructure in a hashref
##
## @scope
##  Private to this module
##
sub _readCommandDescriptionFile {
    #---------------------------------------------------------------------------
    #   Sub Name      : _readCommandDescriptionFile()
    #   Functionality : Sub routine to build the data structure from the
    #                   flat file mentioned.
    #   Input         : flatfile to be read
    #   Return Value  : hashref,
    #                   NULL on failure
    #   Scope         : Private to this module
    #---------------------------------------------------------------------------
    my $flatfile = $_[0];
    open(FH, $flatfile)
      or die "Unable to open file(" . $flatfile . ") for reading. $!";
    my ($hash, $ops, $cmd_set, $object, $verb, $role, $permission);
    my @opsArr = ();
    eval {
        while (<FH>) {
            chomp if ($_ !~ /^\n/);
            if ($_ =~ /commandset=(\w+)/) {
                $cmd_set = $1;
            } elsif ($_ =~ /object=(.*\w)/) {
                $object = $1;
            } elsif ($_ =~ /verb=(\w+)/) {
                $verb = $1;
            } elsif ($_ =~ /permission_([\w-]+)=(\d)/) {
                $role                                          = $1;
                $permission                                    = $2;
                $hash->{$cmd_set}->{$object}->{$verb}->{$role} = $permission;
            } else {
                $ops = $_;
                push @opsArr, $ops;
            }
            if ($_ =~ /^\s/) {
                $hash->{$cmd_set}->{$object}->{$verb}->{'commands'} = [@opsArr];
                @opsArr = ();
            }
        } ## end while (<FH>)
        push @opsArr, "";
        $hash->{$cmd_set}->{$object}->{$verb}->{'commands'} = [@opsArr];
        close FH;
    };
    if ($@) {
        close FH;
        logcomment ("_readCommandDescriptionFile() failed with the error: $@");
        return;
    }
    return $hash;
} ## end sub _readCommandDescriptionFile

## @Function
##   validateFlatfile - Validates the flatfile for the syntax error 
##
## @synopsis
##      use Cmode::RbacCmdFetcher qw(validateFlatfile);
##      my $rv  = validateFlatfile('/u/jaganj/flatfile.txt');
##
## @usage
##      use Cmode::RbacCmdFetcher qw(validateFlatfile);
##      my $rv  = validateFlatfile('<absolute-path-of-flatfile>');
##
## @arguments
##
## @return-value
##  1 on success and 0 on failure
##
## @scope
##  Exported from this module
##
sub validateFlatfile($) {

    #---------------------------------------------------------------------------
    #   Sub Name      : validateFlatfile()
    #   Functionality : Sub routine to verify the entries in the flat file
    #                   mentioned.
    #   Input         : flatfile to be read
    #   Return Value  : 1 on success,
    #                   0 on failure
    #   Scope         : Exported from this module
    #---------------------------------------------------------------------------
    my $flatfile = $_[0];
    open(FH, $flatfile)
      or die "Unable to open file(" . $flatfile . ") for reading. $!";
    my ($object, $line);
    eval {
        while ($line = <FH>) {
            if ($line =~ /^commandset=([\w ]+)/) {
                _verifyCommandset() or die "Format Err at Line No: $.";
            } elsif ($line =~ /^object=([\w ]+)/) {
                _verifyObject() or die "Format Err at Line No: $.";
            } elsif ($line =~ /^verb=verb[\d]+[\s]*$/) {
                _verifyVerb() or die "Format Err at Line No: $.";
            } elsif ($line =~ /^$/) {
                my $next_line = <FH>;
                if (!(    $next_line =~ /^commandset=([\w ]+)/
                       or $next_line =~ /^object=([\w ]+)/
                       or $next_line =~ /^verb=verb[\d]+[\s]*$/) ) {
                    die "Blank line is not followed by commandset/object/verb entity";
                }
                seek(FH, -length($next_line), 1);
                $.--;
            } ## end elsif ($line =~ /^$/)
        } ## end while ($line = <FH>)
        close FH;
    };
    if ($@) {
        close FH;
        print("verifyFlatfile() failed with the error: $@");
        return 0;
    }
    return 1;
} ## end sub validateFlatfile($)

sub _verifyVerb {

    #---------------------------------------------------------------------------
    #   Sub Name      : _verifyVerb
    #   Functionality : Sub routine to verify whether the verb is properly
    #                 : followed by the permissions and commands.
    #   Return Value  : 1 on success verification,
    #                   dies with reason on failure
    #   Scope         : Private to this module
    #---------------------------------------------------------------------------
    my $err_type;
    $err_type =
      "Verb is not followed by all/some permissions and commands at line # $.\n";
    if (   (<FH>) =~ /^permission_vsadmin=[01]+[\s]*$/
        && (<FH>) =~ /^permission_vsadmin-volume=[01]+[\s]*$/
        && (<FH>) =~ /^permission_vsadmin-protocol=[01]+[\s]*$/
        && (<FH>) =~ /^(permission_vsadmin-readonly=[01]+[\s]*)$/) {
        (return 1) if (<FH> =~ /\w+/);
        $err_type = "Commands are not followed by permissions at line # $.\n";
    } else {
        $err_type =
          "Permissions are not properly followed by verbs (or) permission definition error at line # $.\n";
    }
    die $err_type;
} ## end sub _verifyVerb

sub _verifyObject {

    #---------------------------------------------------------------------------
    #   Sub Name      : _verifyObject()
    #   Functionality : Sub routine to verify whether the object is properly
    #                 : followed by the Verb, permissions and commands.
    #   Return Value  : 1 on success verification,
    #                   dies with reason on failure
    #   Scope         : Private to this module
    #---------------------------------------------------------------------------
    if ((<FH>) =~ /^verb=verb[\d]+[\s]*$/) {
        (return 1) if (_verifyVerb());
    }
    die
      "Object is not followed by Verb, all/some permissions and commands at line # $.\n";
} ## end sub _verifyObject

sub _verifyCommandset {

    #---------------------------------------------------------------------------
    #   Sub Name      : _verifyCommandset()
    #   Functionality : Sub routine to verify whether the object is properly
    #                 : followed by the Object, Verb, Permissions and commands.
    #   Return Value  : 1 on success verification,
    #                   dies with reason on failure
    #   Scope         : Private to this module
    #---------------------------------------------------------------------------
    if ((<FH>) =~ /^object=([\w ]+)/) {
        (return 1) if (_verifyObject());
    }
    die
      "Commanset is not followed by Object, Verb, all/some permissions and commands at line # $.\n";
} ## end sub _verifyCommandset
1;
