# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Cmode/VolEfficiency.pm#1 $
#
# $Change: 1917575 $
#
# $DateTime: 2013/01/22 11:22:46 $
#
# Copyright (c) 2010 Network Appliance, Inc.
# All rights reserved.
#
#
# NAME
#      VolEfficiency.pm - cmode vol efficiency library
#
# DESCRIPTION
#      This perl module provide wrapper interfaces to c-mode 
#       volume efficiency cli command
#
# AUTHOR
#       tarung@netapp.com
#       sladi@netapp.com
#
#
## @pod here
##
##

=pod

=head1 NAME

    VolEfficiency.pm - provide wrapper interfaces to C-Mode volume efficiency cli command

=head1 SYNOPSIS

    use Cmode::VolEfficiency qw(:all);
    OR
    use Cmode::VolEfficiency qw(:efficiency);

=head1 DESCRIPTION

=head1 FUNCTIONS

Cmode Volume Efficiency Functions

=cut

package Cmode::VolEfficiency ;

#------------------------------------------------------
#   Module Vars
#------------------------------------------------------
our ($Validation, $VERSION);

#
# Define constants
#
use constant EFF_ERROR => undef;
use constant EFF_TRUE  => 1;
use constant EFF_FALSE => 0;

use warnings;
use strict qw(vars subs refs);

BEGIN {

    use base   qw(Exporter);

    #--------------------------------------------
    #   Exports
    #--------------------------------------------

    $Validation = 1; # unused as of now
    $VERSION    = '0.01';

    my @eff_sub = qw(
                        &volume_efficiency_config
                        &volume_efficiency_modify
                        &volume_efficiency_off
                        &volume_efficiency_on
                        &volume_efficiency_undo
                        &volume_efficiency_check
                        &volume_efficiency_stat
                        &volume_efficiency_policy_create
                        &volume_efficiency_policy_delete
                        &volume_efficiency_policy_modify
                        &volume_efficiency_policy_show
                        &volume_efficiency_start
                        &volume_efficiency_stop
                        &volume_efficiency_show
                        &wait_for_sis_scan_completion
                        &check_space_savings
                    );

    my @constants = qw(EFF_ERROR EFF_TRUE EFF_FALSE);

    our %EXPORT_TAGS = (
        'efficiency' => [ @eff_sub ],
        'constants'  => [ @constants ],
    );

    $EXPORT_TAGS{all} = [ map{ @{$_} } values %EXPORT_TAGS ];

    Exporter::export_ok_tags('all');
}

#------------------------------------------------------
#   Imports
#------------------------------------------------------
use Tharn;
use Data::Dumper;
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF SCALARREF OBJECT );
use Cmode::Flexclone qw(:all);
use Cmode::Quota qw(:all);

#------------------------------------------------------
#     Prototypes
#------------------------------------------------------
sub volume_efficiency_config;
sub volume_efficiency_modify;
sub volume_efficiency_off;
sub volume_efficiency_on;
sub volume_efficiency_undo;
sub volume_efficiency_check;
sub volume_efficiency_stat;
sub volume_efficiency_policy_create;
sub volume_efficiency_policy_delete;
sub volume_efficiency_policy_modify;
sub volume_efficiency_policy_show;
sub volume_efficiency_start;
sub volume_efficiency_stop;
sub volume_efficiency_show;
sub wait_for_sis_scan_completion;
sub check_space_savings;

#------------------------------------------------------
#     Local Variables
#------------------------------------------------------


#------------------------------------------------------
#     Subroutines Definition
#------------------------------------------------------

=head2 volume_efficiency_config()

sets efficiency configuration on given volume

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub volume_efficiency_config {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'   => {type => OBJECT},
            'vserver'  => {type => SCALAR},
            'volume'   => {type => SCALAR, optional => 1},
            'path'     => {type => SCALAR, optional => 1},
            'compression'            => {type => SCALAR, optional => 1},
            'inline-compression'     => {type => SCALAR, optional => 1},
            'min-blocks-shared'      => {type => SCALAR, optional => 1},
            'verify-trigger-rate'    => {type => SCALAR, optional => 1},
            'policy'   => {type => SCALAR, optional => 1},
            'schedule' => {type => SCALAR, optional => 1},
            'errmsg'   => {type => SCALARREF},
            'timeout'  => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_efficiency_config( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_config() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}

=head2 volume_efficiency_modify()

modifies efficiency configuration on given volume

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub volume_efficiency_modify {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'   => {type => OBJECT},
            'vserver'  => {type => SCALAR},
            'volume'   => {type => SCALAR, optional => 1},
            'path'     => {type => SCALAR, optional => 1},
            'compression'            => {type => SCALAR, optional => 1},
            'inline-compression'     => {type => SCALAR, optional => 1},
            'min-blocks-shared'      => {type => SCALAR, optional => 1},
            'verify-trigger-rate'    => {type => SCALAR, optional => 1},
            'policy'   => {type => SCALAR, optional => 1},
            'schedule' => {type => SCALAR, optional => 1},
            'errmsg'   => {type => SCALARREF},
            'timeout'  => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_efficiency_modify( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_modify() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}

=head2 volume_efficiency_off()

disable volume efficiency on a given volume

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub volume_efficiency_off {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'     => {type => OBJECT},
            'vserver'    => {type => SCALAR},
            'volume'     => {type => SCALAR, optional => 1},
            'path'       => {type => SCALAR, optional => 1},
            'errmsg'     => {type => SCALARREF},
            'timeout'    => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_efficiency_off( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_off() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}


=head2 volume_efficiency_undo()

enable volume efficiency on a given volume

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

Example
Usage:
    -vserver <vserver name>      Virtual Server Name
   { [-volume] <volume name>     Volume Name
   |  -path </vol/volume>  }     Path to volume
  [[-compression|-C] [true]]     *Decompress the data in the volume
  [ -dedupe|-D [true] ]          *Undo block sharing in the volume
  [ -inode|-i {1..2147483647} ]  *Inode number to undo sharing
  [ -undo-type|-t {all|wrong} ]  *Selective undo
  [ -log|-d [true] ]             *Only log incorrect savings

=item Returns:

Returns 1 if successful, otherwise 0

=back


=cut

sub volume_efficiency_undo {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'     => {type => OBJECT},
            'vserver'    => {type => SCALAR},
            'volume'     => {type => SCALAR},
            'path'       => {type => SCALAR, optional => 1},
            'compression'=> {type => SCALAR, optional => 1},
            'dedupe'     => {type => SCALAR, optional => 1},
            'inode'      => {type => SCALAR, optional => 1},
            'undo-type'  => {type => SCALAR, optional => 1},
            'log'        => {type => SCALAR, optional => 1}, 
            'errmsg'     => {type => SCALARREF},
            'timeout'    => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_efficiency_undo( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_undo() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}


=head2 volume_efficiency_check()

enable volume efficiency on a given volume

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

Example

Usage:
    -vserver <vserver name>               Virtual Server Name
   { [-volume] <volume name>              Volume Name
   |  -path </vol/volume>  }              Path to volume
  [[-delete-checkpoint|-d] {true|false}]  *Delete Checkpoint

=item Returns:

Returns 1 if successful, otherwise 0

=back


=cut

sub volume_efficiency_check {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'             => {type => OBJECT},
            'vserver'            => {type => SCALAR},
            'volume'             => {type => SCALAR},
            'path'               => {type => SCALAR, optional => 1},
            'delete-checkpoint'  => {type => SCALAR, optional => 1}, 
            'errmsg'             => {type => SCALARREF},
            'timeout'            => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_efficiency_check( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_check() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}




=head2 volume_efficiency_on()

enable volume efficiency on a given volume

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub volume_efficiency_on {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'     => {type => OBJECT},
            'vserver'    => {type => SCALAR},
            'volume'     => {type => SCALAR, optional => 1},
            'path'       => {type => SCALAR, optional => 1},
            'policy'     => {type => SCALAR, optional => 1},
            'errmsg'     => {type => SCALARREF},
            'timeout'    => {type => SCALAR, default => 60},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }
    # if path is defined, delete volume key
    delete $input_arg{'volume'} if ( exists $input_arg{'path'} );

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_efficiency_on( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_on() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}

=head2 volume_efficiency_policy_create()

create given volume efficiency policy for given vserver

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub volume_efficiency_policy_create {
    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'     => {type => OBJECT},
            'vserver'    => {type => SCALAR},
            'policy'     => {type => SCALAR},
            'schedule'   => {type => SCALAR},
            'duration'   => {type => SCALAR, optional => 1},
            'enabled'    => {type => SCALAR, optional => 1},
            'comment'    => {type => SCALAR, optional => 1},
            'errmsg'     => {type => SCALARREF},
            'timeout'    => {type => SCALAR, default => 60},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_efficiency_policy_create( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_policy_create() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}

=head2 volume_efficiency_policy_delete()

delete given volume efficiency policy for given vserver

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub volume_efficiency_policy_delete {
    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'     => {type => OBJECT},
            'vserver'    => {type => SCALAR},
            'policy'     => {type => SCALAR},
            'errmsg'     => {type => SCALARREF},
            'timeout'    => {type => SCALAR, default => 60},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_efficiency_policy_delete( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_policy_delete() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}

=head2 volume_efficiency_policy_modify()

modify given volume efficiency policy for given vserver

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub volume_efficiency_policy_modify {
    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'     => {type => OBJECT},
            'vserver'    => {type => SCALAR},
            'policy'     => {type => SCALAR},
            'schedule'   => {type => SCALAR},
            'duration'   => {type => SCALAR, optional => 1},
            'enabled'    => {type => SCALAR, optional => 1},
            'comment'    => {type => SCALAR, optional => 1},
            'errmsg'     => {type => SCALARREF},
            'timeout'    => {type => SCALAR, default => 60},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_efficiency_policy_modify( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_policy_modify() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}

=head2 volume_efficiency_policy_show()

show volume efficiency policy for given vserver

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub volume_efficiency_policy_show {
    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
            cli_op  => 1,
        );

    my %args_validn_spec =
        (
            'server'     => {type => OBJECT},
            'vserver'    => {type => SCALAR},
            'policy'     => {type => SCALAR},
            'schedule'   => {type => SCALAR},
            'duration'   => {type => SCALAR, optional => 1},
            'enabled'    => {type => SCALAR, optional => 1},
            'comment'    => {type => SCALAR, optional => 1},
            'errmsg'     => {type => SCALARREF},
            'timeout'    => {type => SCALAR, default => 60},
            'cli_op'     => {type => ARRAYREF},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        @{$cmdOpRef} = $server->volume_efficiency_policy_show( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_policy_show() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}

=head2 volume_efficiency_start()

starts volume efficiency operation on a given volume

=over 4

=item Arguments:

  [ -vserver <vserver name> ]           Vserver Name (default: vs0)
   { [-volume] <volume name>            Volume Name
   |  -path </vol/volume>  }            Volume Path
  [[-scan-old-data|-s] [true]]          Scan Old Data
  { [ -use-checkpoint|-p [true] ]       Use Checkpoint (if scanning old data)
  | [ -delete-checkpoint|-d [true] ] }  Delete Checkpoint
  [ -compression|-C [true] ]            *Start Compression (if scanning old
                                        data)
  [ -dedupe|-D [true] ]                 *Start Deduplication (if scanning old
                                        data)
  [ -shared-blocks|-a [true] ]          *Compress Shared Blocks (if scanning
                                        old data)
  [ -snapshot-blocks|-b [true] ]        *Compress Blocks In Snapshots (if
                                        scanning old data)
  [ -queue|-q [true] ]                  Operation Should Be Queued (if scanning
                                        old data)
  [ -force|-f [true] ]                  Force Operation

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub volume_efficiency_start {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'errmsg'              => {type => SCALARREF},
            'timeout'             => {type => SCALAR, default => 60},
            'server'              => {type => OBJECT},
            'vserver'             => {type => SCALAR},
            'volume'              => {type => SCALAR, optional => 1},
            'path'                => {type => SCALAR, optional => 1},
            'scan-old-data'       => {type => SCALAR, optional => 1},
            'compression'         => {type => SCALAR, optional => 1},
            'dedupe'              => {type => SCALAR, optional => 1},
            'shared-blocks'       => {type => SCALAR, optional => 1},
            'snapshot-blocks'     => {type => SCALAR, optional => 1},
            'queue'               => {type => SCALAR, optional => 1},
            'force'               => {type => SCALAR, optional => 1},
            'use-checkpoint'      => {type => SCALAR, optional => 1},
            'delete-checkpoint'   => {type => SCALAR, optional => 1},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_efficiency_start( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_start() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}

=head2 volume_efficiency_stop()

stop volume efficiency operation on a given volume

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub volume_efficiency_stop {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'     => {type => OBJECT},
            'vserver'    => {type => SCALAR},
            'volume'     => {type => SCALAR, optional => 1},
            'path'       => {type => SCALAR, optional => 1},
            'all'        => {type => SCALAR, optional => 1},
            'errmsg'     => {type => SCALARREF},
            'timeout'    => {type => SCALAR, default => 60},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        $server->volume_efficiency_stop( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_stop() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}

=head2 volume_efficiency_show()

show volume efficiency status on a given volume 

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub volume_efficiency_show {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            timeout => 1,
            cli_op  => 1,
        );

    my %args_validn_spec =
        (
            'server'    => {type => OBJECT},
            'vserver'   => {type => SCALAR, optional => 1},
            'volume'    => {type => SCALAR, optional => 1},
            'path'      => {type => SCALAR, optional => 1},
            'state'     => {type => SCALAR, optional => 1},
            'op-status' => {type => SCALAR, optional => 1},
            'progress'  => {type => SCALAR, optional => 1},
            'type'      => {type => SCALAR, optional => 1},
            'schedule'  => {type => SCALAR, optional => 1},
            'policy'    => {type => SCALAR, optional => 1},
            'min-blocks-shared'    => {type => SCALAR, optional => 1},
            'blks-skipped-sharing' => {type => SCALAR, optional => 1},
            'last-op-begin' => {type => SCALAR, optional => 1},
            'last-op-end'   => {type => SCALAR, optional => 1},
            'last-op-size'  => {type => SCALAR, optional => 1},
            'last-op-error' => {type => SCALAR, optional => 1},
            'changelog-usage'      => {type => SCALAR, optional => 1},
            'checkpoint-time'      => {type => SCALAR, optional => 1},
            'checkpoint-op-type'   => {type => SCALAR, optional => 1},
            'checkpoint-stage'     => {type => SCALAR, optional => 1},
            'checkpoint-sub-stage' => {type => SCALAR, optional => 1},
            'checkpoint-progress'  => {type => SCALAR, optional => 1},
            'logical-data-size'    => {type => SCALAR, optional => 1},
            'logical-data-limit'   => {type => SCALAR, optional => 1},
            'logical-data-percent' => {type => SCALAR, optional => 1},
            'queued-job'           => {type => SCALAR, optional => 1},
            'stale-fingerprint-percentage' => {type => SCALAR, optional => 1},
            'compression'       => {type => SCALAR, optional => 1},
            'inline'  => {type => SCALAR, optional => 1},
            'errmsg'  => {type => SCALARREF},
            'cli_op'  => {type => ARRAYREF},
            'timeout' => {type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);

    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'}; # cli output ref array

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {
        @{$cmdOpRef} = $server->volume_efficiency_show( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_show() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}

=head2 volume_efficiency_stat ()

check the sis status and space savings etc  .

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub volume_efficiency_stat {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            cli_op  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'                       =>{type => OBJECT},
            'vserver'                      =>{type => SCALAR},
            'volume'                       =>{type => SCALAR},
            'path'                         =>{type => SCALAR, optional => 1},
            'errmsg'                       =>{type => SCALARREF},
            'cli_op'                       =>{type => ARRAYREF},
            'timeout'                      =>{type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);
    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'};

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {

        @{$cmdOpRef} = $server->volume_efficiency_stat( %input_arg );
    };

    if($@) {
        $$err_ref = "volume_efficiency_stat() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}
=head2 wait_for_sis_scan_completion()

wait for sis scan to complete with in the time out .

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub wait_for_sis_scan_completion {

    my %args_validn_spec =
        (
            server     => {type => OBJECT},
            vserver    => {type => SCALAR},
            volume     => {type => SCALAR},
            scanfor    => {type => SCALAR},
            interval   => {type => SCALAR, default => 1},
            timeout    => {type => SCALAR, default => 1200},
            errmsg     => {type => SCALARREF},
        );

    my %options  = validate(@_, \%args_validn_spec);

    my $server    = $options{'server'};  # server object
    my $vserver   = $options{'vserver'}; # vserver name
    my $volume    = $options{'volume'};  # volume name
    my $scan_type = $options{'scanfor'}; # Type of scan .
    my $timeout   = $options{'timeout'}; # timeout for cli command
    my $err_ref   = $options{'errmsg'};  # reference error msg variable
    my $interval  = $options{'interval'}; # interval between two calls

    my @cmdOp;

    my $finish = time() + $timeout;

    while ( time() <=  $finish ) {
        if( $scan_type =~ m/eff_start/i ){
            if ( EFF_TRUE == volume_efficiency_show(
                                server => $server,
                                vserver => $vserver,
                                volume  => $volume,
                                cli_op  => \@cmdOp,
                                errmsg  => $err_ref) ) {
                if ( ${$cmdOp[0]}{'op-status'} =~ m/Idle/i ) {
                        logcomment("sis scan is completed");
                        return EFF_TRUE;
                } else {
                    logcomment("sis scan is still active");
                }
            } 
        }elsif ( $scan_type =~ m/eff_undo/i ) {    
            if ( EFF_TRUE == volume_efficiency_show(
                                server => $server,
                                vserver => $vserver,
                                volume  => $volume,
                                cli_op  => \@cmdOp,
                                errmsg  => $err_ref) ) {
                if ( ${$cmdOp[0]}{'op-status'} =~ m/Undoing/i &&
                     ${$cmdOp[0]}{'stage'} !~ m/Done/i ) {
                    logcomment("sis undo is still active");
                }
            } else {
                if( EFF_TRUE == volume_show( server => $server,
                                             vserver => $vserver,
                                             volume  => $volume,
                                             cli_op  => \@cmdOp,
                                             errmsg  => $err_ref)){
                    if (${$cmdOp[0]}{'dedupe-space-saved'} eq '0' ){
                        logcomment("There are no spcae savings on $volume");
                        return EFF_TRUE;
                    }
                }
            }
        }

    sleep($interval);
    }
    #
    # timeout occured
    #
    $$err_ref = "sis scan did not complete in $timeout seconds";

    return EFF_FALSE;
}

=head2 check_space_savings()

This function collects the space savings of both parent and clone vol .

=over 4

=item Arguments:

=over 4

takes a hash as input (see %args_validn_spec for valid keys)

=back

=item Returns:

Returns 1 if successful, otherwise 0

=back

=cut

sub check_space_savings{

    my (@cmdOp,$vol_savings,$clone_vol_savings);
    my %args_validn_spec =
        (
            server            => {type => OBJECT},
            vserver           => {type => SCALAR},
            parent_volume     => {type => SCALAR},
            clone_volume      => {type => SCALAR},
            timeout           => {type => SCALAR, default => 300},
            savings_array     => {type => ARRAYREF},
            errmsg            => {type => SCALARREF},
        );

    my %options  = validate(@_, \%args_validn_spec);

    my $server          = $options{'server'};  # server object
    my $vserver         = $options{'vserver'}; # vserver name
    my $parent_volume   = $options{'parent_volume'};  # volume name
    my $clone_volume    = $options{'clone_volume'};  # volume name
    my $savings_array   = $options{'savings_array'}; # reference to a savings array
    my $timeout         = $options{'timeout'}; # timeout for cli command
    my $err_ref         = $options{'errmsg'};  # reference error msg variable
#    my @vol_array       = ($parent_volume,$clone_volumer)
    foreach my $vol($parent_volume,$clone_volume){
        if ( EFF_TRUE == volume_efficiency_stat(
                                server => $server,
                                vserver => $vserver,
                                volume  => $vol,
                                cli_op  => \@cmdOp,
                                errmsg  => $err_ref) ) {

                    $vol_savings = ${$cmdOp[0]}{'space-saved'};
                    logcomment("The $vol have space savings $vol_savings");
                    push (@{$savings_array},{ volume   => $vol,
                                              savings  => $vol_savings});
        }else{
            $$err_ref = "volume efficiency stat failed"; 
            return EFF_FALSE;
        }
    }
    my @results = @{$savings_array};
    logcomment("$results[0]{volume} savings $results[0]{savings}\n"
               . "\n$results[1]{volume} savings $results[1]{savings}\n");
    if( $results[0]{savings} =~ m/^$results[1]{savings}$/ ){ 
        logcomment("The parent and clone vol have same space savings");
        return EFF_TRUE;
    } else {
        logcomment("The parent and clone vol dont have same space savings");
        $$err_ref = "The parent and clone vol dont have same space savings";
        return EFF_FALSE;
    }
}

1;


=head1 SEE ALSO

    None

=head1 STATUS

    Shared

=head1 AUTHOR

    Tarun Garg (tarung@netapp.com)
    $Id: //depot/prod/test/Rsierranevada/lib/Cmode/VolEfficiency.pm#1 $
    Last $Change: 1917575 $
    Last Revised $DateTime: 2013/01/22 11:22:46 $

=head1 HISTORY

    2010-08-04 author: tarung      original lib created

=head1 COPYRIGHT

Copyright 2010, Network Appliance., All Rights Reserved

=cut

sub volume_efficiency_old_stat {

    my %invalid_cmd_keys =
        (
            server  => 1,
            errmsg  => 1,
            cli_op  => 1,
            timeout => 1,
        );

    my %args_validn_spec =
        (
            'server'                       =>{type => OBJECT},
            'vserver'                      =>{type => SCALAR},
            'volume'                       =>{type => SCALAR},
            'path'                         =>{type => SCALAR, optional => 1},
            'errmsg'                       =>{type => SCALARREF},
            'cli_op'                       =>{type => HASHREF},
            'timeout'                      =>{type => SCALAR, default => 300},
        );

    my %options = validate(@_, \%args_validn_spec);
    my %input_arg; # hash of input argument to cli command

    my $server   = $options{'server'}; # server object
    my $err_ref  = $options{'errmsg'}; # refernce to error msg variable
    my $cmdOpRef = $options{'cli_op'};

    # set the input arguments to cli command
    $input_arg{'-Timeout'} = $options{'timeout'};

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    # reset the error message
    $$err_ref = "none";

    eval {

        my $cli_op = $server->volume_efficiency_stat( %input_arg );
        logcomment("The cli_output -$cli_op ");
        # chop the cli output based on newline char
        my @op_list = split('\n', $cli_op);

        # pick up the value from last line of cli output
        my @keyList = split('X;X', $op_list[0]);

        # pick up the value from last line of cli output
        my @valList = split('X;X', $op_list[-1]);

        foreach my $i ( 0..$#keyList ) {
            $cmdOpRef->{$keyList[$i]} = $valList[$i];
        }

    };

    if($@) {
        $$err_ref = "volume_efficiency_stat() failed, error: $@";
        return EFF_FALSE;
    }

    return EFF_TRUE;
}
