#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/7Mode/Aggregate.pm#1 $
#

## @summary Wrapper around NOTE::Aggregate.
##
## @description
## Wrapper around NOTE::Aggregate.
## The main reason for not using NOTE::Aggregate directly is to provide standardized output
##    and to enable specific processing on some failures, otherwise fatal.
##    This is required for handling compression-specific failures.
##    
## @synopsis
##
## use Compression::Aggregate;
## use vars qw($FILER $AGGREGATE );
## 
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $filer = Compression::Filer->new( filer => $FILER );
## my $obj = Compression::Aggregate->new(
## 	filer => $filer,
##      aggregate => $AGGREGATE,
##      debug => FALSE,
## );
##
## $obj->volume_is_contained( volume => $vol_name, expected => C_RC_FAILURE );
## $obj->offline_contained_volumes();
## $obj->online_contained_volumes();
## $obj->set_offline(force => TRUE, all => TRUE );
## $obj->set_offline();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 


package Compression::7Mode::Aggregate;

use strict;
use warnings;

use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Tharn;
use Compression::Constants qw(:all);
use Compression::Utils qw(detect_indirect_type_from_status
			  _c_interpret_logresult);
use IPU::7ModeIPU qw(aggr_upgrade_status_diag
                     aggr_upgrade_diag
                     aggr_upgrade_scan_stop
                     aggr_upgrade_scan_start);
use Aggregate qw(aggr_blockcheck aggr_disks
                 aggr_blockcheck_read_output
                 aggr_blockcheck_wait
                 aggr_disks);
use Data::Dumper;
use NACL::STask::Aggregate;

BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}



## @name ::new
##
## @summary Compression::Aggregate Constructor.
##
## @description
## Creates an object representing an aggregate on the specified filer.
##
## @return Object of type Compression::Aggregate
##
## @arg conn Required: Compression::Filer object corresponding to the hosting filer.
##
## @arg aggregate Optional: Aggregate information.
## It can be either a scalar (representing the name), or a hash reference.
## If not provided, uses the default values defined in Compression::Constants.
## 
## @arg ui Optional: Compression::UI object corresponding to the hosting filer.
## If not provided, it will be auto-generated.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
##
## my $filer = Compression::Filer->new( filer => $FILER );
## my $obj = Compression::Aggregate->new(
## 	filer => $filer,
##      aggregate => $AGGREGATE,
##      debug => TRUE,
## );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub new
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			conn 	=> {type => OBJECT,
                                    isa  => "NOTE::Filer",},
                        
                        aggregate => {type => SCALAR|HASHREF|UNDEF,
                                      optional => 1,},
                        
                        ui      => {type => OBJECT,
                                    isa  => "Compression::UI",
                                    optional => 1,},
                        
                        debug 	=> {type => SCALAR,
                                    optional => 1,
                                    default => FALSE,},
                        
                        attach => {type => SCALAR,
                                   optional => 1,
                                   default => FALSE,},
			
			disable_cleanup => {type => BOOLEAN,
                                    optional => 1,
                                    default => FALSE,
                                    },
						node => {type => SCALAR,
                                   optional => 1,},
		} );

	my $self = {};
        $self->{_debug} = $args{debug};
	$self->{_creator_pid} = $$;
	$self->{_disable_cleanup} = $args{disable_cleanup};
	$self->{_hosting_filer_has_changed} = FALSE; # For 7-mode this is not yet used
        
        logcomment("Aggregate: new: aggregate " . Dumper($args{aggregate})) if ($self->{_debug});
        
	# Construct the aggr hash as expected by the NOTE::Aggregate constructor
	my %aggr = ();
        
        my $aggregate_args = $args{aggregate};
        if ((defined($aggregate_args) && (ref($aggregate_args) eq "HASH")))
        {
                # Second-layer hash validation for aggregate args
                logcomment("Aggregate: new: aggregate attributes passed in as HASHREF") if ($self->{_debug});
                %aggr = validate_with
                        ( params => $aggregate_args,
                          spec   => {name => { type => SCALAR, default => C_DFLT_AGGREGATE },
                                     recycle => { type => SCALAR, default => C_DFLT_AGGR_PURGE },
                                     share => { type => SCALAR, default => C_DFLT_AGGR_PURGE },
                                     disks => { type => SCALAR, default => C_DFLT_AGGR_DISKS },
                                     snaplock => { type => SCALAR, default => C_ENABLE_SNAPLOCK },
                                     indirect_type => { type => SCALAR, default => $main::AGGR_IND_TYPE },
                                     },
                          allow_extra => 0,
                        );
        }
        else
        {
                logcomment("Aggregate: new: aggregate attribute passed in as SCALAR or undefined") if ($self->{_debug});
                $aggr{name} = (defined($aggregate_args)) ? $aggregate_args : C_DFLT_AGGREGATE;
                $aggr{recycle} = C_DFLT_AGGR_PURGE;
                $aggr{share} = C_DFLT_AGGR_PURGE;
                $aggr{disks} = C_DFLT_AGGR_DISKS;
                $aggr{snaplock} = C_ENABLE_SNAPLOCK;
                $aggr{indirect_type} = $main::AGGR_IND_TYPE;
        }
        logcomment("Aggregate: new: aggregate information hash to be passed to NOTE constructor:\n" . Dumper(%aggr)) if ($self->{_debug});
        
        # Pass the NOTE::Filer obj to reuse the connection
        my $conn = $args{conn};
	$aggr{Filer} = $conn;
        
        $self->{_snaplock} = $aggr{snaplock};
        # Need to initialize the compliance clock if snaplock aggregate
        # NOTE does not do this for us
        if ($self->{_snaplock} eq 'compliance')
        {
                logcomment("Initializing the compliance clock, to create aggregate of type " . $self->{_snaplock});
				$conn->check_license( 'snaplock' );
				$conn->say("priv set -q diag;options licensed_feature.snaplock.enable on");
                $conn->time(clock => 'compliance', set => 1);
        }
		elsif ($self->{_snaplock} eq 'enterprise')
        {
                logcomment("Initializing the compliance clock, to create aggregate of type " . $self->{_snaplock});
				$conn->check_license( 'snaplock_enterprise' );
				$conn->say("priv set -q diag;options licensed_feature.snaplock_enterprise.enable on");
                $conn->time(clock => 'compliance', set => 1);
        }
		elsif ($self->{_snaplock})
		{
			logcomment("Initializing the compliance clock, to create aggregate of type " . $self->{_snaplock});
			$conn->time(clock => 'compliance', set => 1);
		}
        
	# Create an aggregate if required, otherwise re-use it
        my $attach = $args{attach};
        if ($attach == TRUE)
        {
                $self->{_aggr} = NOTE::Aggregate->attach(name => $aggr{name},
                                                         Filer => $conn);
        }
        else
        {
				# Following fix (replacing NOTE::Aggregate->create() call
				# with NACL::STask::Aggregate->create() is a fix for
				# burt659093. -B (64|32) was recently removed from the
				# privilege mode admin. NACL handles these issues internally,
				# hence, better to rely on NACL.
				my $NACL_conn = NACL::C::Node->new(name => $args{node});
				NACL::STask::Aggregate->create(
						aggregate           => $aggr{name},
						command_interface   => $NACL_conn,
						diskcount           => $aggr{disks},
						"block-type"		=> $aggr{indirect_type}."-bit",
						if_exists           => 'reuse',
				);
				# Later we need to have NOTE object only. that's why the
				# following wrapper.
				$self->{_aggr} = NOTE::Aggregate->attach(
						name => $aggr{name},
						Filer => $conn,
				);
        }
	$self->{aggr_name} = $aggr{name};
        $self->{indirect_type} = $aggr{indirect_type};
	$self->{_recycle} = $aggr{recycle};
	$self->{_filer_name} = $conn->host();
	$self->{_block_check_scans} = {};
	$self->{_latest_block_check_id} = '';
        
        # UI
        if ( defined($args{ui}) )
        {
                $self->{_ui} = $args{ui};
        }
        else
        {
                # Auto-detect version of the hosting filer. Need this for UI.
                logcomment("Volume: UI object not provided. Auto-generating...") if ($self->{_debug});
                my $version = Compression::Version->new( conn => $conn, debug => $self->{_debug} );
                $self->{_ui} = Compression::UI->new( version => $version, debug => $args{debug}, boot_mode => C_ONTAP_7MODE );
        }

        bless $self, $class;
        
        #########################
	# Do extra setup steps
	#########################
        
        # Check status. If aggr is offline, attempt to bring online (it may happen if aggr was reused)
        my $aggr_status = $self->{_aggr}->status;
        my $aggr_state = $aggr_status->state();
        my $aggr_name = $self->{aggr_name};
        logcomment("State of aggregate $aggr_name is: $aggr_state") if ($self->{_debug});

        if ($aggr_state ne C_AGGR_STATE_ONLINE)
        {
                logcomment("Attempting to online aggregate $aggr_name ...") if ($self->{_debug});
                $self->set_online();
        }
        
        # Need to ensure here that the indirect type is as it should be...
        # NOTE::Aggregate does not ensure this for recycled aggregates
        my %current_type = $self->get_indirect_type();
        if ($current_type{indirect_type} != $self->{indirect_type})
        {
                logresult("NA", "NOT APPLICABLE: Aggregate $aggr_name is required to be " . $self->{indirect_type} . "-bit, " .
                                   "but it is a " . $current_type{indirect_type} . "-bit aggregate.");
        }
        
        logcomment("Using $aggr_name, a " . $self->{indirect_type} . "-bit aggregate.");
        
        # Need to ensure that the snaplock type is as it should be...
        # NOTE::Aggregate does not check this for recycled named aggregates
        my $aggr_snaplock_type = $self->{_aggr}->snaplock();
        my $required_snaplock_type = $self->{_snaplock};
        if ("$aggr_snaplock_type" ne "$required_snaplock_type")
        {
                logresult("NA", "NOT APPLICABLE: Aggregate $aggr_name is required to be a snaplock aggregate of type \"$required_snaplock_type\" " .
                                   "but it is of type \"$aggr_snaplock_type\"");
        }
		
		
	return $self;
}


## @name ::get_state
## @summary Gets the state of the aggregate.
## 
## @description
## Gets the state of the aggregate.
##
## @return state.
## 
## @examples
## my $aggr_state = $obj->get_state();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_state
{
	my $class = shift;
        my %args = validate(@_,
        {
                # None.		
	} );
        
        my $aggr_status = $class->{_aggr}->status();
        my $state = $aggr_status->{state};
        
        if (!defined($state))
        {
                logresult('fatal', "Unable to detect the aggregate state:\n" . Dumper($aggr_status));
        }
        
        return $state;
}


## @name ::get_indirect_type
## @summary Gets the indirect type of the aggregate (32/64 -bit).
## 
## @description
## Gets the indirect type of the aggregate (32/64 -bit).
##
## @return Hash with rc, indirect_type, and output (status hashref).
## 
## @examples
## $obj->get_indirect_type();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_indirect_type
{
	my $class = shift;
        my %args = validate(@_,
	{
		# None.		
	} );
        
        my $aggr_status = $class->{_aggr}->status();
        my $status = $aggr_status->{status};
        
        logcomment("Aggregate: Current status:\n" . Dumper($aggr_status) .
                   "Attempting to detect indirect type from: status: $status") if $class->{_debug};
        
        my $indirect_type = detect_indirect_type_from_status(status_string => $status);
        
        return (rc => C_RC_SUCCESS, indirect_type => $indirect_type, output => $aggr_status);
}


## @name ::volume_is_contained
## @summary Verifies that the specified volume exists in this aggregate.
## 
## @description
## Verifies that the specified volume exists in this aggregate.
##
## @return Hash with rc and output (list of existent volumes).
## 
## @arg volume Required: Name of the volume.
## 
## @arg expected Optional: Expected result (C_RC_SUCCESS/C_RC_FAILURE).
## If not provided, assumes that the volume should exist.
## C_RC_FAILURE_NON_FATAL can also be passed in as 'expected', in which case,
## the script will not be terminated upon any result.
## Instead, the result (C_RC_SUCCESS/C_RC_FAILURE) will be passed to the script for further processing.
##
## @examples
## $obj->volume_is_contained( volume => $vol, expected => C_RC_FAILURE );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub volume_is_contained
{
	my $class = shift;
        my %args = validate(@_,
	{
		volume   => { type => SCALAR, },
                expected  => { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },		
	} );
        
        my @vol_list = $class->{_aggr}->volumes();
        my $vol_to_find = $args{volume};
        my $rc = C_RC_FAILURE;
        
        logcomment("Aggregate: Looking for volume \"$vol_to_find\" in the list:\n" . Dumper(@vol_list)) if $class->{_debug};
        foreach my $existent_vol (@vol_list)
        {
                logcomment("Aggregate: Attempting to match volume \"$vol_to_find\" against \"$existent_vol\"") if $class->{_debug};
                if ($existent_vol eq $vol_to_find)
                {
                      logcomment("Aggregate: Volume $vol_to_find found: $vol_to_find=$existent_vol") if $class->{_debug};
                      $rc = C_RC_SUCCESS;
                      last;
                }
        }
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Verify that the volume \"$vol_to_find\" is contained by aggregate " . $class->{aggr_name},
                                                    output      => \@vol_list,);
}


## @name ::offline_contained_volumes
## @summary Offlines all volumes contained by this aggregate.
## 
## @description
## Offlines all volumes contained by this aggregate.
##
## @return Hash with rc and output (string containing ", "-delimited list of volumes that were offlined).
## 
## @arg skip_root_volume Optional: Boolean flag to decide if the root volume should be skipped.
## Defaults to TRUE.
##
## @examples
## $obj->offline_contained_volumes();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub offline_contained_volumes
{
	my $class = shift;
        my %args = validate(@_,
	{
		skip_root_volume => { type => BOOLEAN, optional => 1, default => TRUE, },	
	} );
        
        my $skip_root_flag = $args{skip_root_volume};
        my @volumes = $class->{_aggr}->Volumes();
        my $output = '';
        
        # Get the name of the root volume, to be skipped, if required.
        my $root_vol_name = $class->{_aggr}->Filer->vol_root();
        logcomment("Root volume is $root_vol_name") if ($class->{_debug});
        
        foreach my $volume (@volumes)
        {
                my $volume_name = $volume->name();
                logcomment("Aggregate: offline_contained_volumes: Processing volume $volume_name") if $class->{_debug};
                
                # Skip the root volume, if required
                if (($skip_root_flag == TRUE) && ($volume_name eq $root_vol_name))
                {
                        logcomment("Aggregate: offline_contained_volumes: Skipping root volume $volume_name") if ($class->{_debug});
                        next;
                }
                
                # Check if already offline
                my $vol_status = $volume->status();
                my $vol_state = $vol_status->{state};
                if ($vol_state eq 'offline')
                {
                        logcomment("Aggregate: Volume $volume_name is already offline. Skipping...") if $class->{_debug};
                        next;
                }
                
                logcomment("Aggregate: Bringing volume $volume_name offline...") if $class->{_debug};
                $volume->offline();
                $output .= "$volume_name, ";
        }
        
        # If we got here, NOTE did not encounter any error
        return (rc => C_RC_SUCCESS, output => $output );
}


## @name ::online_contained_volumes
## @summary Onlines all volumes contained by this aggregate.
## 
## @description
## Onlines all volumes contained by this aggregate.
##
## @return Hash with rc and output (string containing ", "-delimited list of volumes that were onlined).
## 
## @arg skip_root_volume Optional: Boolean flag to decide if the root volume should be skipped.
## Defaults to TRUE.
##
## @examples
## $obj->online_contained_volumes();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub online_contained_volumes
{
	my $class = shift;
        my %args = validate(@_,
	{
		skip_root_volume => { type => BOOLEAN, optional => 1, default => TRUE, },	
	} );
        
        my $skip_root_flag = $args{skip_root_volume};
        my @volumes = $class->{_aggr}->Volumes();
        my $output = '';
        
        # Get the name of the root volume, to be skipped, if required.
        my $root_vol_name = $class->{_aggr}->Filer->vol_root();
        logcomment("Root volume is $root_vol_name") if ($class->{_debug});
        
        foreach my $volume (@volumes)
        {
                my $volume_name = $volume->name();
                logcomment("Aggregate: online_contained_volumes: Processing volume $volume_name") if $class->{_debug};
                
                # Skip the root volume, if required
                if (($skip_root_flag == TRUE) && ($volume_name eq $root_vol_name))
                {
                        logcomment("Aggregate: online_contained_volumes: Skipping root volume $volume_name") if ($class->{_debug});
                        next;
                }
                
                # Check if already online
                my $vol_status = $volume->status();
                my $vol_state = $vol_status->{state};
                if ($vol_state eq 'online')
                {
                        logcomment("Aggregate: Volume $volume_name is already online. Skipping...") if $class->{_debug};
                        next;
                }
                
                logcomment("Aggregate: Bringing volume $volume_name online...") if $class->{_debug};
                $volume->online();
                $output .= "$volume_name, ";
        }
        
        # If we got here, NOTE did not encounter any error
        return (rc => C_RC_SUCCESS, output => $output );
}


## @name ::set_offline
## @summary Offlines this aggregate.
## 
## @description
## Offlines this aggregate. Wrapper around NOTE::Aggregate::offline.
## It accepts the same arguments as Aggregate::aggr_offline.
##
## @arg force Optional: Boolean flag to force offlining the aggregate.
##
## @arg cifsdelaytime Optional: Number of seconds to delay before taking the aggregate offline,
## during which time CIFS uses are warned of the pending loss of service.
##
## @arg all Optional: Boolean flag to offline aggregate with all its volume
##
## @return Hash with rc and output.
##
## @examples
## $obj->set_offline();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub set_offline
{
        my $class = shift;
        my %args = validate(@_,
	{
		force => { type => BOOLEAN, optional => 1, },
                cifsdelaytime => { type => SCALAR, optional => 1, },
                all => { type => BOOLEAN, optional => 1, },
	} );
        
        logcomment("Aggregate: set_offline called with arguments:\n" . Dumper(%args)) if ($class->{_debug});
        my $output = $class->{_aggr}->offline( %args );
        
        # If we got here, NOTE did not encounter any error
        return (rc => C_RC_SUCCESS, output => $output );
}


## @name ::set_online
## @summary Onlines this aggregate.
## 
## @description
## Onlines this aggregate. Wrapper around NOTE::Aggregate::online.
## It accepts the same arguments as Aggregate::aggr_online.
##
## @arg force Optional: Boolean flag to force onlining the aggregate.
##
## @return Hash with rc and output.
##
## @examples
## $obj->set_online();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub set_online
{
        my $class = shift;
        my %args = validate(@_,
	{
		force => { type => BOOLEAN, optional => 1, },
	} );
        
        logcomment("Aggregate: set_online called with arguments:\n" . Dumper(%args)) if ($class->{_debug});
        my $output = $class->{_aggr}->online( %args );
        
        # If we got here, NOTE did not encounter any error
        return (rc => C_RC_SUCCESS, output => $output );
}


## @name ::get_avail
## @summary Get available space
## 
## @description
## Returns available space in KB.
## Wrapper around NOTE::Aggregate
##
## @return available space in KB
##
## @examples
## my $avail = $obj->get_avail();
## 
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_avail
{
        my $class = shift;
        
        return $class->{_aggr}->avail();
}


## @name ::list_snapshots
## @summary Lists snapshots that exist on this aggregate.
## 
## @description
## Lists snapshots that exist on this aggregate. Wrapper around NOTE::Aggregate::snapshot_list.
##
## @return Hash with rc and output (reference to array with names of existent snapshots).
##  
## @examples
## $obj->list_snapshots();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub list_snapshots
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # none
                } );
        
        my @snapshots = $class->{_aggr}->snapshot_list();
        
        # If we got here, NOTE did not encounter any error. Return success.
        return (rc => C_RC_SUCCESS, output => \@snapshots);
}

## @name ::delete_snapshot
## @summary Deletes a snapshot.
## 
## @description
## Deletes a named snapshot. Wrapper around NOTE::Aggregate::snapshot_delete.
##
## @return Hash with rc and output (empty).
##
## @arg name Required: Name of the snapshot.
##  
## @examples
## $obj->delete_snapshot(name => $my_snapshot);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub delete_snapshot
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        name => { type => SCALAR, },
                } );
        
        $class->{_aggr}->snapshot_delete($args{name});
        
        # If we got here, NOTE did not encounter any error. Return success.
        return (rc => C_RC_SUCCESS, output => '');
}

## @name ::start_64_bit_upgrade
## @summary Starts 64-bit upgrade for this aggregate.
## 
## @description
## Starts 64-bit upgrade for this aggregate.
##
## @return Hash with rc and output.
##
## @arg mode Optional: Upgrade mode: check|normal|grow-reserved|grow-all.
## Default: none.
##  
## @examples
## $obj->start_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub start_64_bit_upgrade
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        mode => { type => SCALAR, optional => 1, default => C_LA_IPU_MODE_NORMAL},
                } );
        
        # TODO: switch to using NOTE for all the LA IPU methods once support is available
        my $aggr_name = $class->{aggr_name};
        my $mode = $args{mode};
        my $error_msg;
        
        my @allowed_modes = (C_LA_IPU_MODE_CHECK, C_LA_IPU_MODE_NORMAL,
                             C_LA_IPU_MODE_GROW_RESERVED, C_LA_IPU_MODE_GROW_ALL);
        
        if (!grep(/^$mode$/, @allowed_modes))
        {
                logresult('script', "Illegal \"mode\" parameter passed to start_64_bit_upgrade: $mode");
        }
        
        logcomment("Starting LA IPU on aggregate $aggr_name, mode: $mode");
        my $status = aggr_upgrade_diag(Fc => $class->{_aggr}->Filer->conn(),
                                       aggr_name => $aggr_name,
                                       mode => $mode,
                                       error_msg => \$error_msg);
        
        # There is not really a very good way to make it more reliable than this...
        # aggr_upgrade_diag will, in the most cases, return the raw string,
        # but for some cases it may return integers
        # For now, rely on "no error message"
        if (defined($error_msg))
        {
                logresult('fatal', "Failed to start LA IPU on aggregate $aggr_name: $error_msg");
        }
}


## @name ::wait_for_64_bit_upgrade
## @summary Waits for a previously started 64-bit upgrade to complete.
## 
## @description
## Waits for a previously started 64-bit upgrade to complete.
##
## @return Hash with rc.
##
## @arg timeout Optional: Timeout value, in seconds.
##
## @arg delay Optional: Time interval, in seconds, in between consecutive status checks.
##  
## @examples
## $obj->wait_for_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub wait_for_64_bit_upgrade
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # We need to have default values for timeout/delay until NOTE support is available for 7-mode
                        timeout => {type => SCALAR, optional => 1, default => C_DFLT_TIMEOUT},
                        delay => {type => SCALAR, optional => 1, default => 1},
                } );
        
        my $timeout = $args{timeout};
        my $delay = $args{delay};
        my $aggr_name = $class->{aggr_name};
        
        logcomment("Waiting for LA IPU of aggregate $aggr_name to complete");
        my %status_hash = $class->status_64_bit_upgrade(expected => C_RC_FAILURE_NON_FATAL);
        my $status = $status_hash{rc};
        
        while (($timeout > 0) && ($status != C_RC_LA_IPU_NOT_IN_PROGRESS))
        {
                sleep($delay);
                $timeout -= $delay;
                %status_hash = $class->status_64_bit_upgrade(expected => C_RC_FAILURE_NON_FATAL);
                $status = $status_hash{rc};
        }
        
        if ($timeout == 0)
        {
                logresult('fatal', "Timed-out while waiting for LA IPU of aggregate $aggr_name to complete:\n" . Dumper(%status_hash));
        }
        
        return (rc => C_RC_SUCCESS);
}


## @name ::stop_scan_64_bit_upgrade
## @summary Stops a previously started 64-bit upgrade.
## 
## @description
## Stops a previously started 64-bit upgrade .
##
## @return Hash with rc.
##  
## @examples
## $obj->stop_scan_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub stop_scan_64_bit_upgrade
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # None
                } );
        
        my $aggr_name = $class->{aggr_name};
        my $output;
        
        my $status = aggr_upgrade_scan_stop(Fc => $class->{_aggr}->Filer->conn(),
                                            aggr_name => $aggr_name,
                                            error_msg => \$output);
        
        if($status == 0)
        {
                logresult('fatal', "Failed to stop LA IPU scanner for aggregate $aggr_name");
        }
        
        return (rc => C_RC_SUCCESS);
}


## @name ::start_scan_64_bit_upgrade
## @summary Restarts a previously stopped 64-bit upgrade.
## 
## @description
## Restarts a previously stopped 64-bit upgrade.
##
## @return Hash with rc.
## @examples
## $obj->start_scan_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub start_scan_64_bit_upgrade
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # None
                } );
        
        my $aggr_name = $class->{aggr_name};
        my $output;
        
        my $status = aggr_upgrade_scan_start(Fc => $class->{_aggr}->Filer->conn(),
                                             aggr_name => $aggr_name,
                                             error_msg => \$output);
        
        if($status == 0)
        {
                logresult('fatal', "Failed to start LA IPU scanner for aggregate $aggr_name");
        }
        
        return (rc => C_RC_SUCCESS);
}



## @name ::status_64_bit_upgrade
## @summary Gets the status of the 64-bit upgrade for this aggregate.
## 
## @description
## Gets the status of the 64-bit upgrade for this aggregate.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_LA_IPU_STOPPED/C_RC_LA_IPU_NOT_IN_PROGRESS.
## (defined in Compression::Constants).
##
## @return Hash with rc and output. 
##  
## @examples
## my %res = $obj->status_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub status_64_bit_upgrade
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
                } );
	
        my $output;
        my $expected = $args{expected};
        my $aggr_name = $class->{aggr_name};
        my $rc = C_RC_FAILURE;
        
        # aggr_upgrade_status_diag will return 0/-1 if no status, a string is ipu is running...
        # Need to deal with different possible types...
        my $status = aggr_upgrade_status_diag(Fc => $class->{_aggr}->Filer->conn(),
                                              aggr_name => $aggr_name,
                                              error_msg => \$output);
        
        if ("$status" eq '0')
        {
                $rc = C_RC_LA_IPU_NOT_IN_PROGRESS;
        }
        elsif ("$status" ne '-1')
        {
                # Note: do not go through UI for this yet, since the status will be
                # made available directly once NOTE support is added
                if ($status =~ /$aggr_name\s+upgrading\s+stopped/)
                {
                        $rc = C_RC_LA_IPU_STOPPED;
                }
                else
                {
                        $rc = C_RC_SUCCESS;
                }
        }
        
        logcomment("Aggregate: status_64_bit_upgrade returning rc: $rc") if ($class->{_debug});
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $expected,
                                                    log_message => "Obtain 64-bit upgrade status for aggregate $aggr_name",
                                                    output      => $output);
}


## @name ::is_block_type_32bit
## @summary Checks if the aggregate is 32-bit.
## 
## @description
## Checks if the aggregate is 32-bit.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc.
##  
## @examples
## $obj->is_block_type_32bit(expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_block_type_32bit
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
                } );
	
	my %type = $class->get_indirect_type();
        my $rc = ($type{indirect_type} == C_AGGR_32_BIT) ? C_RC_SUCCESS : C_RC_FAILURE;
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Check if aggregate block type is 32-bit",
                                                    output      => $type{output});
}


## @name ::is_block_type_64bit
## @summary Checks if the aggregate is 64-bit.
## 
## @description
## Checks if the aggregate is 64-bit.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc.
##  
## @examples
## $obj->is_block_type_64bit(expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_block_type_64bit
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
                } );
	
	my %type = $class->get_indirect_type();
        my $rc = ($type{indirect_type} == C_AGGR_64_BIT) ? C_RC_SUCCESS : C_RC_FAILURE;
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Check if aggregate block type is 64-bit",
                                                    output      => $type{output});
}


## @name ::list_contained_volumes
## @summary List volumes contained by this aggregate.
## 
## @description
## List volumes contained by this aggregate.
##
## @return Array of volume names.
##  
## @examples
## my @vols = $obj->list_contained_volumes();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub list_contained_volumes
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        # None
                });
        
        return $class->{_aggr}->volumes();
}


## @name ::start_block_check
## @summary Runs block_check on this aggregate.
## 
## @description
## Runs block_check on this aggregate.
##
## @arg wait Optional: Flag to control if the method should wait for the scanner to complete.
## Default: TRUE.
## 
## @arg disks: Optional: Arrayref of disk names to run block check on.
## If none of the 'disks' and 'pvbn' arguments are passed in, then block check will be run on all the data disks of this aggregate.
## 
## @arg pvbn: Optional: Arrayref of pvbns to run block check against.
## If none of the 'disks' and 'pvbn' arguments are passed in, then block check will be run on all the data disks of this aggregate.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Default: C_RC_SUCCESS.
## 
## @arg expected_pvbn_file_path Optional: Expected file path to be listed by block check after completion.
## Only relevant if 'pvbn' is passed in.
##
## @return Hash with rc and output.
##  
## @examples
## $obj->start_block_check();
## $obj->start_block_check(pvbn => [0 .. 7],
##                         expected_pvbn_file_path => 'vol/my_vol/my_file');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub start_block_check
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        wait => {type => BOOLEAN, default => TRUE,},
			disks => {type => ARRAYREF, optional => 1,},
			pvbn => {type => ARRAYREF, optional => 1},
			expected => {type => SCALAR, default => C_RC_SUCCESS,},
			expected_pvbn_file_path => {type => SCALAR, optional => 1,},
                });
	
	my $wait = $args{wait}; 
	my $expected = delete($args{expected});
	my $expected_pvbn_file_path = delete($args{expected_pvbn_file_path});
	my $aggr_name = $class->{aggr_name};
	
	my @data_disks = ();
	if ((!exists($args{disks})) && (!exists($args{pvbn})))
	{
		@data_disks = $class->get_data_disk_names();
		$args{disks} = \@data_disks;
	}
	$args{filer} = $class->{_filer_name};
	$args{aggr} = $aggr_name;
	
	my $disks_checked = (exists($args{disks})) ? TRUE : FALSE;
	
	my ($errcode_or_scan_id, $logs) = Aggregate::aggr_blockcheck(%args);
	
	if ($wait == TRUE)
	{
		# $errcode_or_scan_id is the error code
		# $logs is log file path if success, error message if failure
		if ($errcode_or_scan_id == 1)
		{
			return $class->{_ui}->process_output_string(code => C_CMD_AGGR_BLOCK_CHECK_START,
								    output => $logs,
								    expected_rc => $expected,
								    log_message => "Start block_check on aggregate $aggr_name");
		}
		
		return $class->_parse_aggr_block_check_log(log => $logs,
							   disks_checked => $disks_checked,
							   pvbn => $args{pvbn},
							   expected => $expected,
							   expected_pvbn_file_path => $expected_pvbn_file_path);
	}
	else
	{
		# If the scanner completes too quickly, the scan id will not be retrieved
		my $scan_id = (defined($errcode_or_scan_id)) ? $errcode_or_scan_id : 'unknown';
		$class->{_block_check_scans}->{$scan_id}->{logs} = $logs;
		$class->{_block_check_scans}->{$scan_id}->{disks_checked} = $disks_checked;
		$class->{_block_check_scans}->{$scan_id}->{pvbn} = $args{pvbn};
		$class->{_block_check_scans}->{$scan_id}->{expected_pvbn_file_path} = $expected_pvbn_file_path;
		
		
		# $errcode_or_scan_id should be the scan id, but if the command fails, it is 1 (error code)
		# If the command succeedes but the scan id is 1, we may falsely fail here... should open a burt against Aggregate::aggr_blockcheck
		# For now, do our best to detect errors
		if ("$scan_id" eq "1")
		{
			return $class->{_ui}->process_output_string(code => C_CMD_AGGR_BLOCK_CHECK_START,
								    output => $logs,
								    expected_rc => $expected,
								    log_message => "Start block_check on aggregate $aggr_name");
		}
		
		$class->{_latest_block_check_id} = "$scan_id";
		
		return (rc => C_RC_SUCCESS, scan_id => $scan_id, output => "Error code: $scan_id\nLogs/error massage: $logs");
	}
}


## @name ::wait_for_block_check
## @summary Waits for already-started block_check scan to complete.
## 
## @description
## Waits for already-started block_check scan to complete.
##
## @arg scan_id Optional: Id of the block_check scan to wait for.
## If not provided, this will be the last-started scan.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Default: C_RC_SUCCESS.
##  
## @examples
## $obj->wait_for_block_check();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub wait_for_block_check
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        scan_id => {type => SCALAR, regex => qr/^\d+|unknown$/, default => $class->{_latest_block_check_id},},
			expected => {type => SCALAR, default => C_RC_SUCCESS,},
                });
	
	my $expected = $args{expected};
	my $aggr_name = $class->{aggr_name};
	my $scan_id = $args{scan_id};
	
	my %wait_args = (filer => $class->{_filer_name},
			 aggr => $aggr_name,);
	
	if (!exists($class->{_block_check_scans}->{$scan_id}))
	{
		logresult('script', "Aggr block_check scanner with id \"$scan_id\" was not started in the background through this library.\n" .
			            "No log information found:\n" . Dumper($class->{_block_check_scans}));
	}
	
	my $scan_id_info = $class->{_block_check_scans}->{$scan_id};
	$wait_args{ls_etc_log_before} = $scan_id_info->{logs};
	
	if ($scan_id =~ /^\d+$/)
	{ 
		$wait_args{scanid} = $scan_id;
	}
	
	my $log_path = Aggregate::aggr_blockcheck_wait(%wait_args);
	
	return $class->_parse_aggr_block_check_log(log => $log_path,
						   disks_checked => $scan_id_info->{disks_checked},
						   pvbn => $scan_id_info->{pvbn},
						   expected => $expected,
						   expected_pvbn_file_path => $scan_id_info->{expected_pvbn_file_path});
}


## @name ::get_data_disk_names
## @summary Gets data disks names of this aggregate.
## 
## @description
## Gets data disks names of this aggregate.
##
## @return
## Array of data disk names.
##  
## @examples
## my @data_disks = $obj->get_data_disk_names();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_data_disk_names
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			# None
		});
	
	my @disks_info = Aggregate::aggr_disks($class->{_aggr}->Filer->conn(), $class->{aggr_name});
	# Get the data disks
	my @disks = ();
	foreach my $disk_info (@disks_info)
	{
		if ($disk_info->{type} eq 'data')
		{
			push(@disks, $disk_info->{name});
		}
	}
	
	return @disks;
}


######################################
# Internal methods
######################################

sub _refresh
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		new_aggr => {type => SCALAR,},
	});
	
	my $new_aggr = $args{new_aggr};
	# Remember the filer
	my $conn = $class->{_aggr}->Filer;
	my $old_aggr = $class->{aggr_name};
	
	# Refresh the NOTE aggr object of this class.
	# If required to not recycle, destroy the old aggr first
	if ($class->{_recycle} == FALSE)
	{
		if (defined($ENV{'C_CORRUPTION_FOUND'}) && ($ENV{'C_CORRUPTION_FOUND'} != TRUE))
		{
			logcomment("Corruption found. Aggregate $old_aggr will not be destroyed");
		}
		else
		{
			logcomment("Purging aggregate $old_aggr");
			$class->{_aggr}->purge();
		}
	}
	
	# Attach to the new aggregate
	# At this point, ask for recycle and share -> the aggregate already exists
	$class->{_aggr} = NOTE::Aggregate->attach(name => $new_aggr,
						  Filer => $conn,
						  recycle => 1,
						  share => 1,);
	$class->{aggr_name} = $new_aggr;
	$class->{_block_check_scans} = {};
	$class->{_latest_block_check_id} = '';
}


sub _parse_aggr_block_check_log
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			log => {type => SCALAR,},
			disks_checked => {type => BOOLEAN,},
			pvbn => {type => UNDEF|ARRAYREF, optional => 1},
			expected => {type => SCALAR,},
			expected_pvbn_file_path => {type => UNDEF|SCALAR, optional => 1,},
		});
	
	my $log_path = $args{log};
	my $disks_checked = $args{disks_checked};
	my $expected_pvbn_file_path = $args{expected_pvbn_file_path};
	my $pvbn = $args{pvbn};
	my %return_hash = ();
	my $rc = C_RC_SUCCESS;
	
	if ($disks_checked == TRUE)
	{
		my ($scanned_disk,$vbn,$corrupted_disks,$snapshot_list,$lost_block_count)= Aggregate::aggr_blockcheck_read_output(filer => $class->{_filer_name},
																  path => $log_path,
																  disks => 1);
		
		$return_hash{output} = "Scanned disks:\n" . Dumper($scanned_disk) .
		                       "\n\nVbn:\n" . Dumper($vbn) .
				       "\n\nCorrupted disks:\n" . Dumper($corrupted_disks) .
				       "\n\nSnapshot list:\n" . Dumper($snapshot_list) .
				       "\n\nLost block count:\n" . Dumper($lost_block_count);
				       
		if (scalar(@$corrupted_disks) != 0)
		{
			$rc = C_RC_CORRUPTED_DISKS_FOUND;
		}
		
		if ($lost_block_count != 0)
		{
			$rc = C_RC_LOST_BLOCKS_FOUND;
		}
	}
	else
	{
		my $pvbn_file = aggr_blockcheck_read_output(filer => $class->{_filer_name},
							    path => $log_path,
							    pvbn => $pvbn);
		
		$return_hash{output} = Dumper($pvbn_file);
		
		# Aggr block_check skips offline/restricted volumes -> make sure all these pvbns were scanned...
		foreach my $pvbn_instance (@$pvbn)
		{
			if (!exists($pvbn_file->{$pvbn_instance}))
			{
				logcomment("Could not detect info for pvbn $pvbn_instance in the aggr block_check output");
				$rc = C_RC_PVBN_INFO_MISSING;
				$return_hash{pvbn_info}{$pvbn_instance} = 'MISSING';
			}
			else
			{
				my $info = $pvbn_file->{$pvbn_instance};
				chomp($info);
				# chomp seems to not be enough
				$info =~ s/\W*$//;
				$return_hash{pvbn_info}{$pvbn_instance} = $info;
				
				if ((defined($expected_pvbn_file_path)) && ($expected_pvbn_file_path ne $info))
				{
					logcomment("Detected incorrect file path for pvbn $pvbn_instance:\n" .
						   "Expected: $expected_pvbn_file_path\n" .
						   "Obtained: $info");
					$rc = C_RC_INCORRECT_PVBN_INFO;
				}
			}
		}
	}
	
	$class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
					     expected_rc => $args{expected},
					     log_message => "Run block_check on aggregate " . $class->{aggr_name},
					     output      => \%return_hash,);
	
	$return_hash{rc} = $rc;
	return %return_hash;
}

# Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
	
	if (($class->{_disable_cleanup} == FALSE) &&
            ($class->{_creator_pid} == $$))
	{
                my $aggr_name = $class->{aggr_name};
                logcomment("Executing 7-mode Aggregate clean-up for aggregate $aggr_name");

		eval
                {
                        if (($class->{_recycle} == FALSE) &&
			    (!defined($ENV{'C_CORRUPTION_FOUND'}) || ($ENV{'C_CORRUPTION_FOUND'} != TRUE)))
			{
				logcomment("Purging aggregate $aggr_name");
				$class->{_aggr}->purge();
			}
                };
                
                if( $@ )
                {
                        logcomment( "7-mode Aggregate: cleanup for aggregate $aggr_name: error: " . $@ );
                }
                else
                {
                        logcomment( "7-mode Aggregate: cleanup for aggregate $aggr_name: done" ) if ($class->{_debug});
                }
        }
	else
	{
                logcomment( "7-mode Aggregate: cleanup:  nothing done: script PID: " . $$) if ($class->{_debug});
	}
        
        if (defined($execution_error))
        {
                _c_interpret_logresult($execution_error);
        }
}


sub testme
{
	package main;
	
	require "drg/Compression/setup_core.pl"; 
	
	use vars qw($FILER $DEBUG $AGGREGATE $OFFLINE);
	param("DEBUG", "-default", FALSE);
	param("AGGREGATE", "-default", 'compr');
	param("OFFLINE", "-default", FALSE);
	
	my %res;
	my $string_res;
	
	my $filer = NOTE::Filer->connect( host => $FILER );
	my $obj = Compression::7Mode::Aggregate->new(conn => $filer,
						     aggregate => {name => $AGGREGATE,},
						     debug => $DEBUG);
	
	%res = $obj->list_snapshots();
	logcomment("list_snapshots returned:\n" . Dumper(%res));
	
	%res = $obj->get_indirect_type();
	logcomment("get_indirect_type returned:\n" . Dumper(%res));
	
	%res = $obj->volume_is_contained(volume => 'bogus_vol',
					 expected => C_RC_FAILURE);
	logcomment("volume_is_contained returned:\n" . Dumper(%res));
	
	%res = $obj->offline_contained_volumes();
	logcomment("offline_contained_volumes returned:\n" . Dumper(%res));
	
	%res = $obj->online_contained_volumes();
	logcomment("online_contained_volumes returned:\n" . Dumper(%res));
	
	if ($OFFLINE == TRUE)
	{
		%res = $obj->set_offline(force => TRUE, all => TRUE );
		logcomment("set_offline returned:\n" . Dumper(%res));
		
		%res = $obj->set_online();
		logcomment("set_online returned:\n" . Dumper(%res));
	}
	
	$string_res = $obj->get_avail();
	logcomment("get_avail returned: $string_res");
	
	logresult('pass', "Compression\:\:7Mode\:\:Aggregate\:\:testme passed");
}




1;
