#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/7Mode/CFO.pm#2 $
#
## @summary Compression CFO routines for 7-mode
##
## @description
##  This library provides functions specific
##	to CFO related automations
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
## @status private (internal)


package Compression::7Mode::CFO;

use strict;
use warnings;
use Tharn;
#use Module::Common;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use filer qw(filer_cluster_is_installed filer_cluster_license filer_cluster_is_licensed filer_cluster_enable filer_cluster_is_enabled filer_cluster_restore);
use Compression::Constants qw(:all);
use Compression::Version;
use Compression::UI;

BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}


# @name ::new
#
# @summary perform cluster initialization
#
# @description
# Perform all the basic cluster initialization required here
#
# @return Compression::Cluster object
#
# @arg filer Required: primary filer name
# @arg debug Optional: logs additional debug messages specific to this module
# @arg initialize Optional: Initialize clustering if needed (license, cf enable).
#
# @examples
# 	my $cluster  = Compression::Cluster->new(volume => $self->{volume}->{_vol},
#                                               client => $self->{_client}->{name} );
#
# @author dl-compression-qa@netapp.com, waroff@netapp.com
#

sub new
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		filer  => { type=>SCALAR },
		debug  => { type=>SCALAR, optional => 1, default=>FALSE },
		initialize => { type => BOOLEAN, optional => 1, default => TRUE },
	} );
	
	my $self = {};
	$self->{host_name} = $args{filer};
	
	$self->{ID} = $$;
	$self->{_debug} = $args{debug};
	
	if ($args{initialize} == TRUE) {
		# We are not connected to any filer yet. We need to initialize before creating the CFO obj.
		# Create a temporary NOTE object
		my $filer_obj = NOTE::Filer->connect(host => $self->{host_name}, mode => C_DFLT_FILER_CONN);
		my $conn = $filer_obj->conn;
		if (filer_cluster_is_installed($conn)) {
			my $version = Compression::Version->new(conn => $filer_obj,
								debug => $self->{_debug},
								boot_mode => C_ONTAP_7MODE);
			my $ui = Compression::UI->new(version => $version,
						      debug => $self->{_debug},
						      boot_mode => C_ONTAP_7MODE);
			my $license_needed = $ui->get_constant_value(constant_code => 'CFO_LICENSE_NEEDED');
			
			if (($license_needed == TRUE) && (!filer_cluster_is_licensed($conn))) {
				filer_cluster_license($conn)
			}
			
			if (!filer_cluster_is_enabled($conn)) {
				filer_cluster_enable($conn);
			}
			
			# Restore cluster if it is in takeover/giveback
			filer_cluster_restore($conn);
		}
		else {
			logresult('config', "CONFIG ERROR: Cluster initialization failed");
		}
	}
	
	$self->{pair} = NOTE::CFO->new(Filer => $self->{host_name});
	# TODO: remove
	$self->{_partner} = $self->{pair}->partner;
	$self->{partner_name} = $self->{_partner}->host;
	
	bless $self, $class;
	
	return $self;
}


# @name ::takeover
#
# @summary initiates a takeover
#
# @description
# Wrapper around NOTE::CFO::takeover
#
# @arg me			Optional: if TRUE, partner takes over
# @arg background	Optional: run as a background process
#
# @examples
# 	$cluster->takeover;
#
# @author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub takeover
{
	my $class = shift;

	# From lib/Compression/Cluster.pm's takeover() method, 'state' will always
	# be passed in the call to this takeover method and hence the validation
	# below will fail. To get around that, we need to ignore 'state' when
	# passed into this method call.
	# We cannot remove state from the list of arguments passed when the method
	# call is made because it is required for the other takeover() methods (
	# such as the one in CMode/CFO.pm )
	
	my %args = @_;
	delete $args{state};
	my @params = %args;
	
	%args = validate ( @params,
		{
			me => { type=>SCALAR, optional => 1, default=>FALSE }, # TODO: investigate: if no scripts use it, then remove: no support for c-mode
			background => { type=>SCALAR, optional => 1, default=>FALSE },
		},
	);
	
	my $me = $args{'me'};
	my $background	= $args{'background'};
	
	logcomment("\ntakeover\nme: $me") if ( $class->{_debug} );
	
	$class->{pair}->takeover(ME => $me, background => $background);
}

# @name ::giveback
#
# @summary initiates a giveback
#
# @description
# Wrapper around NOTE::CFO::giveback
#
# @arg background	Optional: run as a background process
# @arg force		Optional: use the "-f" force option
# @arg safe		Optional: if TRUE, giveback will not croak if the the current state of the CFO pair is NOT taken over
#
# @examples
# 	$cluster->giveback;
#
# @author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub giveback
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		background => { type=>SCALAR, optional => 1, default=>FALSE },
		force => { type=>SCALAR, optional => 1, default=>FALSE },
		safe => { type=>SCALAR, optional => 1, default=>FALSE },
	});
	
	my $background	= $args{'background'};
	my $force = $args{'force'};
	my $safe = $args{'safe'};
	
	$class->{pair}->giveback(background => $background, force => $force, safe => $safe);
}


# @name ::get_status
#
# @summary status
#
# @description
# Wrapper around NOTE::CFO::status
#
# @return
# A hash with status for "me" and "partner".
# Status can be 'UP', 'TAKEOVER', 'DEAD', 'UNREACH', 'DISABLED'
#
# @examples
# 	%res = $cluster->status;
#	%res includes the keys "me" and "partner"
#
# @author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub get_status
{
	my $class = shift;
	my (%args) = validate (@_,
	{
		# None
	});
	
	my %res;
	
	my ($me_filer_state, $partner_filer_state) = $class->{pair}->status;
	
	logcomment("\nFailover status:\n" .
		   "me: $me_filer_state\n" .
		   "partner: $partner_filer_state\n");
	
	$res{'me'} = $me_filer_state;
	$res{'partner'} = $partner_filer_state;
	return %res;
}


# @name ::monitor
#
# @summary monitor
#
# @description
# Wrapper around NOTE::CFO::monitor
#
# @return
# 
#
# @examples
# 	$cluster->monitor(me=>'TAKEOVER', partner=>'DEAD');
#	$cluster->monitor(me=>'UP', partner=>'UP');
#
# @author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub monitor
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		me => { type=>SCALAR },
		partner => { type=>SCALAR },
	});
	
	my $me		= $args{'me'};
	my $partner	= $args{'partner'};
	
	logcomment("\nMonitor waiting for:\n" .
		   "me: $me\n" .
		   "partner: $partner");
	
	$class->{pair}->monitor(ME => $me, PARTNER => $partner);
}


1;
