# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/7Mode/FileClone.pm#1 $
#
## @summary
##  Utility wrapper module for 7-mode file clone
##
## @author @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (internal)

package Compression::7Mode::FileClone;

BEGIN
{
	require "Compression/base_import.pl";

}

use strict;
use warnings;
use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Compression::Constants qw(:all);
use Dense::Clone qw(dense_sis_clone_start
                    dense_sis_clone_status
		    dense_sis_clone_clear
		    dense_sis_clone_stop
		    dense_sis_clone_wait_for_finish);




# Documentation in Compression::FileClone
sub new
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		filer => {
                        type => OBJECT,
			isa => "NOTE::Filer"},

                volume => {
			type => SCALAR,},

		debug => {
			type => BOOLEAN,
                        optional => 1,
			default => FALSE,},
                
                ui => {type => OBJECT,
		       isa => "Compression::UI",
		       },
	} );

	# Start building our object:
	my $self = {};
        $self->{_filer} = $args{filer};
	$self->{_conn} = $args{filer}->{_conn};
        $self->{_volume} = $args{volume};
        $self->{_debug} = $args{debug};
        $self->{_creator_pid} = $$;
	$self->{_ui} = $args{ui};
	$self->{_latest_id} = 0;
        

	bless($self, $class);

	return $self;
}


sub start
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                source => {type => SCALAR,},
		clone => {type => SCALAR, optional => 1,},
		expected => {type => SCALAR, default => C_RC_SUCCESS,},
		snapshot => {type => SCALAR, optional => 1,},
		blk_range => {type => SCALAR, optional => 1,},
        } );
        
	my $expected = delete($args{expected});
	my $source = $args{source};
	my $options = '';
	if (exists($args{snapshot}))
	{
		my $snap_name = delete($args{snapshot});
		$options .= "-s $snap_name ";
	}
	$args{options} = $options;
	
	if ((!exists($args{clone})) && (!exists($args{blk_range})))
	{
		logresult('script', "One of the \"clone\" or \"blk_range\" must be specified for the file clone start operation");
	}
	
	logcomment("\nStarting file clone for file: $source\n" . Dumper(%args));
	
	$args{conn} = $class->{_conn};
        my $output = Dense::Clone::dense_sis_clone_start(\%args);
	
	# Looks like dense_sis_clone_start returns the id of the clone oper in case of success.
	# It does not fail if the command does not execute successfully
	my %res = $class->{_ui}->process_output_string(code => C_CMD_FILE_CLONE_START,
						       output => $output,
						       expected_rc => $expected,
						       log_message => "File clone start for file: $source");
	
	if ($expected == C_RC_SUCCESS)
	{
		$res{id} = $output;
		$class->{_latest_id} = $output;
	}
	
	return %res;
}


sub status
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                id => {type => SCALAR, optional => 1,},
		latest_started => {type => SCALAR, default => FALSE,},
		expected => {type => SCALAR, default => C_RC_SUCCESS,},
        } );
        
	my $expected = delete($args{expected});
	my $rc = C_RC_SUCCESS;
	
	if ($expected != C_RC_SCANNER_IDLE)
	{
		$class->{_ui}->{version}->check_for_required_features(features => [C_FEATURE_NON_CTIME_FILE_CLONE,]);
	}
	
	my $latest = delete($args{latest_started});
	if ($latest == TRUE)
	{
		$args{id} = $class->{_latest_id};
	}
	
	$args{conn} = $class->{_conn};
	$args{vol} = $class->{_volume};
        my %status = Dense::Clone::dense_sis_clone_status(\%args);
	
	if (!keys(%status))
	{
		$rc = C_RC_SCANNER_IDLE;
	}
	
	$class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
					     expected_rc => $expected,
					     log_message => "Obtain file clone status for volume " . $class->{_volume},
					     output      => \%status);
	
	$status{rc} = $rc;
	return %status;
}


sub clear
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                id => {type => SCALAR, optional => 1, default => $class->{_latest_id},},
		expected => {type => SCALAR, default => C_RC_SUCCESS,},
        } );
        
	my $expected = $args{expected};
	my $id = $args{id};
	my $vol = $class->{_volume};
	
	$class->{_ui}->{version}->check_for_required_features(features => [C_FEATURE_NON_CTIME_FILE_CLONE,]);
	
	my $output = Dense::Clone::dense_sis_clone_clear(conn => $class->{_conn},
							 vol => $vol,
							 id => $id);
	
	return $class->{_ui}->process_output_string(code => C_CMD_FILE_CLONE_CLEAR,
						    output => $output,
						    expected_rc => $expected,
						    log_message => "Clear file clone operation for volume $vol, id $id");
}


sub stop
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                id => {type => SCALAR, optional => 1, default => $class->{_latest_id},},
		expected => {type => SCALAR, default => C_RC_SUCCESS,},
        } );
        
	my $expected = $args{expected};
	my $id = $args{id};
	my $vol = $class->{_volume};
	
	$class->{_ui}->{version}->check_for_required_features(features => [C_FEATURE_NON_CTIME_FILE_CLONE,]);
	
	my $output = Dense::Clone::dense_sis_clone_stop(conn => $class->{_conn},
							vol => $vol,
							id => $id);
	
	return $class->{_ui}->process_output_string(code => C_CMD_FILE_CLONE_STOP, 
						    output => $output,
						    expected_rc => $expected,
						    log_message => "Stop file clone operation for volume $vol, id $id");
}


sub wait_for_completion
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                id => {type => SCALAR, optional => 1, default => $class->{_latest_id},},
		expected => {type => SCALAR, default => C_RC_SUCCESS,},
        } );
        
	my $expected = $args{expected};
	my $id = $args{id};
	my $vol = $class->{_volume};
	
	logcomment("\nWaiting for file clone operation to complete:\n" .
		   "Volume: $vol\n" .
		   "Id: $id");
	
	Dense::Clone::dense_sis_clone_wait_for_finish(conn => $class->{_conn},
						      vol => $vol,
						      id => $id);
	
	return (rc => C_RC_SUCCESS);
}


sub is_bg_clone_in_progress
{
	my $class = shift;
        my (%args) = validate( @_,
        {
                inode_num => {type => SCALAR,},
		expected => {type => SCALAR, default => C_RC_SUCCESS,},
        } );
	
	$class->{_ui}->{version}->check_for_required_features(features => [C_FEATURE_CTIME_FILE_CLONE,]);
	
	my $vol = $class->{_volume};
	my $inum = $args{inode_num};
	my $expected = $args{expected};
	
	my $cmd_code = C_CMD_BG_FILE_CLONE_STATUS;
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $vol,);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_filer},
                                                            boot_mode => C_ONTAP_7MODE,
                                                            cmd => $command,);
	
	my $bg_clone_prefix = $class->{_ui}->get_constant_value(constant_code => 'BG_FILE_CLONE_ACTIVE_PREFIX');
	$bg_clone_prefix .= $inum;
	
	my $rc = ($output =~ /$bg_clone_prefix/) ? C_RC_SUCCESS : C_RC_SCANNER_IDLE;

	return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
						    expected_rc => $expected,
						    log_message => "Obtain background file clone status for volume " . $vol,
						    output      => $output);
}





1;