#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/7Mode/Lun.pm#1 $
#

# Documentation is found in Compression::Lun
# This module contains the filer-side methods needed for luns
package Compression::7Mode::Lun;

use strict;
use warnings;

use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult);
use blocks_filer;
use filer qw(filer_options_set
             filer_options_get);


BEGIN
{
	require "Compression/base_import.pl";

}


sub new
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		parent => {type => OBJECT,},

		name => {type => SCALAR,
			 optional => 1,},
		
		type => {type => SCALAR},
		
		size => {type=>SCALAR,},
		
		recycle => {type=>SCALAR,},
	});

	# Start building our object:
	my $self = {};
	my $parent = delete($args{parent});
	$self->{_debug} = $parent->{_debug};
	$self->{_disable_cleanup} = $parent->{_disable_cleanup};
	$self->{_creator_pid} = $$;
	$self->{_type} = $args{type};
	$self->{_mapped} = FALSE;
	
	$args{ostype} = $self->{_type};
	
	if ($parent->isa("Compression::VolClone"))
	{
		# NOTE::LUN does not like the Clone objects to be passed in
		$args{Parent} = NOTE::Volume->attach(Filer => $parent->{filer}->_get_underlying_filer_obj(),
						     name => $parent->{vol_name},);
	}
	elsif ($parent->isa("Compression::Compress"))
	{
		$args{Parent} = $parent->{volume}->_get_underlying_vol_obj();
	}
	else
	{
		logresult('script', "The parent parameter must be a Compress object or a VolClone object");
	}
	
	my $version = Compression::Version->new(conn => $args{Parent}->Filer, debug => $self->{_debug});
	$self->{_ui} = Compression::UI->new(version => $version, debug => $self->{_debug}, boot_mode => C_ONTAP_7MODE);
	
	$self->{_note_lun} = NOTE::LUN->create(%args);
	$self->{_name} = $self->{_note_lun}->name();
	$self->{_full_path} = $self->{_note_lun}->mount_path();
	$self->{_conn} = $self->{_note_lun}->Filer->conn();
	$self->{_filer_ip} = $self->{_note_lun}->Filer->defaultip;
	
	bless($self, $class);

	return $self;

}


sub license_and_start_iscsi
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		# None
	} );
	
	# For pre-RR, installing iscsi license is required
	my $ontap = $class->{_ui}->{version}->{ontap};
	
	logcomment("Licensing iscsi");
	$class->{_note_lun}->Filer->check_license(C_LICENSE_ISCSI);
	
	if ($ontap ge '8.1')
	{
		logcomment("Enabling licensed feature iscsi");
		my $option = $class->{_ui}->get_constant_value( constant_code => 'ENABLE_LICENSED_ISCSI' );
		if (filer::filer_options_get($class->{_conn}, $option) ne 'on')
		{
			# Do not verify if the option was set.
			# For some platforms, this will fail. Just log a warning instead.
			filer::filer_options_set($class->{_conn}, $option, 'on', 0);
			
			if (filer::filer_options_get($class->{_conn}, $option) ne 'on')
			{
				logcomment("Warning: The enable_licensed_iscsi option was not set to on");
			}
		}
	}
	
	blocks_filer::filer_iscsi_start($class->{_conn});
	
	return (rc => C_RC_SUCCESS, output => '');
}


sub create_igroup
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		igroup_name => {type => SCALAR,},
		initiator_name => {type => SCALAR,},
	} );
	
	my $igroup_name = $args{igroup_name};
	my $initiator_name = $args{initiator_name};
	
	# Note: this is currently having the following limitation:
	# if the igroup already exists, the initiator must be already added
	if (blocks_filer::filer_igroup_is_defined($class->{_conn}, $igroup_name))
	{
		logcomment("Igroup $igroup_name is already defined");
	}
	else
	{
		logcomment("Creating igroup name $igroup_name (initiator=$initiator_name)");
		blocks_filer::filer_iscsi_igroup_create_for_ostype($class->{_conn}, $igroup_name, $class->{_type}, $initiator_name);
	}
	
	$class->{_igroup_name} = $igroup_name;
	
	return (rc => C_RC_SUCCESS,)
}


sub map_to_igroup
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		lun_number => {type => SCALAR, optional => 1,},
	} );
	
	if (!defined($class->{_igroup_name}))
	{
		logresult('script', "Igroup must be created before mapping the lun")
	}
	
	logcomment("Mapping lun " . $class->{_name} . " to igroup " . $class->{_igroup_name});
	if (defined($args{lun_number}))
	{
		blocks_filer::filer_lun_map($class->{_conn},
					    $class->{_full_path},
					    $class->{_igroup_name},
					    $args{lun_number});
		$class->{_mapped_to_lun_no} = $args{lun_number};
	}
	else
	{
		$class->{_mapped_to_lun_no} = blocks_filer::filer_lun_map_low($class->{_conn},
									      $class->{_full_path},
									      $class->{_igroup_name});
	}
	
	$class->{_mapped} = TRUE;
	
	return (rc => C_RC_SUCCESS, mapped_to => $class->{_mapped_to_lun_no},);
}


sub unmap_from_igroup
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		# None
	} );
	
	if (!defined($class->{_igroup_name}))
	{
		logresult('script', "Igroup must be created before mapping/unmapping the lun")
	}
	
	logcomment("Unmapping lun " . $class->{_name} . " from igroup " . $class->{_igroup_name});
	blocks_filer::filer_lun_unmap($class->{_conn}, $class->{_full_path}, $class->{_igroup_name});
	$class->{_mapped} = FALSE;
}


sub get_iscsi_nodename
{
	my $class = shift;
	
	return blocks_filer::filer_iscsi_nodename_get($class->{_conn});
}


sub resize_lun
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		size => {type => SCALAR,},
	});

	$class->{_note_lun}->resize($args{size});
	
	return (rc => C_RC_SUCCESS);
}


sub online_lun
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		# None
	});

	$class->{_note_lun}->online();
	
	return (rc => C_RC_SUCCESS);
}


sub offline_lun
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		# None
	});

	$class->{_note_lun}->offline();
	
	return (rc => C_RC_SUCCESS);
}



############################################
# Internal methods
############################################

# Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
	
	if (($class->{_disable_cleanup} == FALSE) &&
            ($class->{_creator_pid} == $$))
	{
                logcomment("Executing 7-mode Lun clean-up for " . $class->{_name});
		
		if ($class->{_mapped} == TRUE)
		{
			eval
			{
				$class->unmap_from_igroup();
				$class->{_note_lun}->destroy();
			};
			if($@)
			{
				logcomment("Lun: 7-mode cleanup for lun " . $class->{_name} . ": error: " . $@);
			}
		}
        }
	else
	{
                logcomment( "Lun: 7-mode cleanup:  nothing done: script PID: " . $$) if ($class->{_debug});
	}
        
        if (defined($execution_error))
        {
                Compression::Utils::_c_interpret_logresult($execution_error);
        }
}


1;
