#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/7Mode/UpgradeRevert.pm#1 $
#

## @summary 7-mode upgrade/revert utility module.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (internal - to only be used by the compression library)

package Compression::7Mode::UpgradeRevert;

BEGIN
{
	require "Compression/base_import.pl";
}

use strict;
use warnings;

use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Tharn;

use Compression::Constants qw(:all);
use Compression::Utils qw(detect_indirect_type_from_status);
use Compression::CompressionUtilities;
use Compression::Decompress;
use Filer::Upgrade;
use Filer::Version qw(version_compare);
use Filer::Version qw(:releases);
use Filer::Transform;
use FS::Util::Utilities qw(revert_to);
use install;
use filer;
use Dense qw(dense_revert_to dense_status dense_start dense_wait_for_status);
use SnapMirror qw(snapmirror_break snapmirror_status_list snapmirror_quiesce snapmirror_enable snapmirror_release);
use SnapVault qw(snapvault_status_list);
use Volume::Clone qw(vol_clone_split_start);
use Volume qw(vol_status);


# Constructor
# Executes upgrade/revert initialization: bring filer from current_release to required_base_release
# Documentation in Compression::UpgradeRevert::initialize
sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
		filer => {type => SCALAR,},
                
                current_release => {type => SCALAR|UNDEF,
                                    default => $main::CURRENT_RELEASE},
                
                required_base_release => {type => SCALAR|UNDEF,},
                
                base_release_url => {type => SCALAR|UNDEF,
                                     default => $main::BASE_REL_URL,},
                
                ndu_testing => {type => SCALAR|UNDEF, 
                                optional => 1,},
                
                disable_cleanup => {type => BOOLEAN,
                                    default => $main::DISABLE_CLEANUP,},
                
                debug => {type => BOOLEAN,
                          default => $main::DEBUG,},
	} );
        
        my $required_base_rel = $args{required_base_release};
        if (!defined($required_base_rel) || ($required_base_rel !~ /^\d+\.\d+\.\d+$/))
        {
                logresult('script', "\n\nUndefined or illegal required_base_release argument value: \"$required_base_rel\"\n" .
                                    "This must be either passed in directly by the script or script must be run with correct command-line argument BASE_RELEASE\n" .
                                    "The string MUST contain the minor relase as well. Ex: 8.0.1\n\n");
        }
        
        my $current_release = $args{current_release};
        if (!defined($current_release) || ($current_release !~ /^\d+\.\d+\.\d+$/))
        {
                logresult('script', "\n\nAny upgrade/revert script must pass in a valid current_release parameter " .
                                    "or must be run with the command-line argument CURRENT_RELEASE set\n\n");
        }
	
	my $self = {};
        $self->{_filer} = $args{filer};
        $self->{_disable_cleanup} = $args{disable_cleanup};
        $self->{_required_base_release} = $required_base_rel;
        $self->{_base_release_url} = $args{base_release_url};
        
        $self->{_debug} = $args{debug};
        $self->{_creator_pid} = $$;
        
        # Connect to filer
        $self->{_conn} = NOTE::Filer->connect(host => $self->{_filer},);
        # Version
        $self->{_version} = Compression::Version->new(conn => $self->{_conn},
                                                      debug => $self->{_debug},
                                                      boot_mode => C_ONTAP_7MODE);
        
        $self->{_ndu_testing} = FALSE;
        if (filer_cluster_is_installed($self->{_conn}->conn()))
        {
		$self->{_ndu_testing} = TRUE;
        }
        
        if (defined($args{ndu_testing}))
        {
                if ($args{ndu_testing} != $self->{_ndu_testing})
                {
                        if ($args{ndu_testing} == TRUE)
                        {
                                logresult('config', "CONFIG ERROR: NDU testing was specifically required, but filer " . $self->{_filer} . " does not have CF installed");
                        }
                        else
                        {
                                # It means that this is a cluster, but the user specifically asked for non-NDU testing
                                $self->{_ndu_testing} = FALSE;
                        }
                }
        }
        
        if ($self->{_ndu_testing} == TRUE)
        {
                logcomment("NDU testing: Enabling clustering for filer " . $self->{_filer});
                my $initialize_status = filer_cluster_initialize($self->{_conn}->conn());
		if ($initialize_status != 1) {
			logresult('config', "CONFIG ERROR: Cluster initialization failed");
		}
                $self->{_cfo} = NOTE::CFO->new(Filer => $self->{_filer});
                $self->{_partner_conn} = $self->{_cfo}->partner();
                $self->{_partner} = $self->{_partner_conn}->host;
        }
        else
        {
                my $cluster_status = filer_cluster_is_enabled($self->{_conn}->conn());
                if ($cluster_status == 1)
                {
                        logcomment("Non-NDU testing required. Disabling clustering for filer " . $self->{_filer});
                        filer_cluster_disable($self->{_conn}->conn());
                }
        }
        
        bless $self, $class;
        
        if ($self->_check_current_release($current_release) == C_RC_FAILURE)
        {
                logresult('script', "\n\nError: user-defined current major release ($current_release) " .
                                    "does not match current major release of the filer " . $self->{_filer} . "\n" .
                                    "CURRENT_RELEASE was not set correctly\n" .
                                    "or script passed in incorrect current_release value to UpgradeRevert constructor\n\n");
        }
        
        $self->{_initial_release} = $current_release;
        
        
        $self->_install_base_release();
        
	return $self;
}


# Documentation in Compression::UpgradeRevert
sub upgrade_to_release
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                release => {type => SCALAR, regex => qr/^\d+\.\d+\.\d+$/,},
                url => {type => SCALAR|UNDEF, optional => 1, default => $main::UPGRADE_URL},
                kernel_type => {type => SCALAR, default => C_DFLT_KERNEL_TYPE,},
                reboot => {type => SCALAR, default => TRUE,},
                client => {type => SCALAR, default => C_CLIENT_NFS_LOCAL,},
                site => {type => SCALAR, default => $main::UPGRADE_SITE,},
        });
        
        my $reboot = delete($args{reboot});
        my $to_release = $args{release};
        
        
        if ($class->{_ndu_testing} == FALSE)
        {
                # Culstering is already disabled
                $args{node} = $class->{_conn};
                $class->_install_download_on_single_node(%args);
                
                if ($reboot == TRUE)
                {
                        logcomment("Rebooting filer " . $class->{_filer});
                        $class->{_conn}->reboot();
                        
                        # Check that after the upgrade, major releases match...
                        if ($class->_check_current_release($to_release) == C_RC_FAILURE)
                        {
                                logresult('fatal', "\n\nError: user-defined upgrade-target major release ($to_release) " .
                                                   "does not match current major release of the filer " . $class->{_filer} . " after the upgrade\n" .
                                                   "Upgrade may have failed\n" .
                                                   "or user-defined ontap URL may have been incorrect\n" .
                                                   "or script passed in incorrect release value to upgrade_to_release\n\n");
                        }
                }
        }
        else
        {
                $args{node} = $class->{_conn};
                $class->_install_download_on_single_node(%args);
                $args{node} = $class->{_partner_conn};
                $class->_install_download_on_single_node(%args);
                
                if ($reboot == TRUE)
                {
                        logcomment("Performing takeover of filer " . $class->{_filer});
                        $class->{_cfo}->takeover(ME => 1);
                        logcomment("Waiting for filer " . $class->{_filer} . " to be taken over...");
                        $class->{_cfo}->monitor(ME => 'DEAD', PARTNER => 'TAKEOVER');
                        
                        # It looks like, if the nvram version mismatches, takeover gets disabled because of version mismatch
                        # See burt 466777 for more details
                        # For these cases, attempting to do giveback to the partner to have it come up with the new version
                        # will fail and cf wil be disabled
                        my $timeout = C_DFLT_TIMEOUT;
                        my $my_cfo_state = 'DOWN';
                        my $partner_cfo_state = 'TAKEOVER';
                        
                        logcomment("Performing giveback to filer " . $class->{_filer});
                        $class->{_cfo}->giveback(background => 1);
                        
                        logcomment("Waiting for giveback to complete");
                        while (($my_cfo_state !~ /UP/) && ($my_cfo_state !~ /DISABLED/))
                        {
                                if ($timeout <= 0)
                                {
                                        logresult('fatal', "Timed-out while waiting for a known cluster state after the install (NDU testing).\n" .
                                                           "Last obtained status: $my_cfo_state, $partner_cfo_state");
                                }
                                
                                sleep(5);
                                $timeout -= 5;
                                ($my_cfo_state, $partner_cfo_state) = $class->{_cfo}->status();
                                logcomment("CFO STATE: $my_cfo_state, $partner_cfo_state");
                        }
                        
                        if ($my_cfo_state eq 'DISABLED')
                        {
                                logcomment("Rebooting the partner node, " . $class->{_partner});
                                $class->{_partner_conn}->reboot();
                        }
                        else
                        {
                                logcomment("Performing takeover of the partner node, " . $class->{_partner});
                                $class->{_cfo}->takeover();
                                logcomment("Waiting for the partner node to be taken over...");
                                $class->{_cfo}->monitor(ME => 'TAKEOVER', PARTNER => 'DEAD');
                                logcomment("Performing giveback to the partner node, " . $class->{_partner});
                                $class->{_cfo}->giveback;
                                
                        }
                        
                        # In both cases, after both filers are up with the newly-installed release, cf should be enabled...
                        logcomment("Waiting for both filers to be up...");
                        $class->{_cfo}->monitor(ME => 'UP', PARTNER => 'UP');
                        
                        # Check that after the upgrade, major releases match...
                        if ($class->_check_current_release($to_release) == C_RC_FAILURE)
                        {
                                logresult('fatal', "\n\nError: user-defined upgrade-target major release ($to_release) " .
                                                   "does not match current major release of the filer " . $class->{_filer} . " after the upgrade\n" .
                                                   "Upgrade may have failed\n" .
                                                   "or user-defined ontap URL may have been incorrect\n" .
                                                   "or script passed in incorrect release value to upgrade_to_release\n\n");
                        }
                }
        }
        
        return (rc => C_RC_SUCCESS,);
}


# Documentation in Compression::UpgradeRevert
sub revert_to_release
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                release => {type => SCALAR},
                force => {type => SCALAR, default => TRUE,},
                max_attempts => {type => SCALAR, regex => qr/^\d+$/, default => 5,},
                expected_requirements => {type => HASHREF, default => {},},
        });
        
        if ($class->{_ndu_testing} == FALSE)
        {
                # Culstering is already disabled
                $args{node} = $class->{_conn};
                $class->_revert_to_on_single_node(%args);
        }
        else
        {
                # First disable CFO
                filer_cluster_disable($class->{_conn}->conn());
                $args{node} = $class->{_conn};
                $class->_revert_to_on_single_node(%args);
                $args{node} = $class->{_partner_conn};
                $class->_revert_to_on_single_node(%args);
                # Re-enable CFO
                $class->{_conn}->transit_to( 'UP');
                $class->{_partner_conn}->transit_to( 'UP');
                filer_cluster_enable($class->{_conn}->conn());
        }
        
        
        
        
        return (rc => C_RC_SUCCESS);
}


################################################################
# Internal subroutines
################################################################

sub _install_download_on_single_node
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                node => {type => OBJECT, isa => "NOTE::Filer"},
                release => {type => SCALAR, regex => qr/^\d+\.\d+\.\d+$/,},
                url => {type => SCALAR|UNDEF, optional => 1, default => $main::UPGRADE_URL},
                kernel_type => {type => SCALAR, default => C_DFLT_KERNEL_TYPE,},
                client => {type => SCALAR, default => C_CLIENT_NFS_LOCAL,},
                site => {type => SCALAR|UNDEF,},
        });
        
        my $filer = $args{node};
        my $filer_name = $filer->host();
        my $to_release = $args{release};
        my $kernel_type = $args{kernel_type};
        my $url = $args{url};
        
        # If installing below br1, checl for 32-bit root aggregate
        if (Filer::Version::version_compare($to_release, C_MIN_REL_ALLOWING_64BIT_ROOT_VOL) < 0)
        {
                $class->_require_32bit_root_aggr($filer, $to_release);
        }
        
        # Filer::Upgrade does not currently have a way of passing in the url
        # It only uses an internally-generated path (/x/eng/rlse/DOT/<release>)
        # If the caller specified an url to be used for upgrading, we need to use a different api
        if (defined($url))
        {
                logcomment("\nInstalling new release on filer $filer_name: $to_release\n" .
                           "Current_release: " . $class->{_version}->{ontap} . "\n" .
                           "Detected user-provided URL to be used for upgrade:\n$url");
                
                install::copy_system_files($filer->conn->hostp(), $args{client},
                                           ONTAP_URL => $url,);
                install::do_download($filer->conn->hostp(), $args{client},);
        }
        else
        {
                logcomment("\nInstalling new release on filer $filer_name: $to_release\n" .
                           "Current_release: " . $class->{_version}->{ontap} . "\n" .
                           "Required kernel type: $kernel_type\n" .
                           "Site: " . $args{site});
                
                # Note: currently, the x.y.0 directory does not exist under /x/eng/rlse/DOT/
                # If this changes, then remove the following lines...
                # Removing the .0 to be able to upgrade
                if ($to_release =~ /^(\d+\.\d+)\.0$/)
                {
                        $to_release = $1;
                }
                
                my $upg = Filer::Upgrade->new($class->{_conn}->conn(),
                                              to_version => $to_release,);
                # Note: do_upgrade still reboots the filer if REBOOT is 0 -> pass y/n
                # Use install_release and download_kernel directly...
                $upg->install_release(kernel => $kernel_type,
                                      client => $args{client},
                                      SITE => $args{site});
                $upg->download_kernel();
        }
        
        return C_RC_SUCCESS;
}

sub _require_32bit_root_aggr
{
        my $class = shift;
        my $node = shift;
        my $attempted_rel = shift; # Just for logging purposes
        
        my $root_aggr = $node->aggr_root();
        
        logcomment("Checking type of root aggregate for filer " . $node->host . ": $root_aggr");
        my $root_aggr_obj = NOTE::Aggregate->attach(Filer => $node,
                                                    name => $root_aggr,);
        
        my $root_aggr_status = $root_aggr_obj->status();
        my $status = $root_aggr_status->{status};
        my $indirect_type = Compression::Utils::detect_indirect_type_from_status(status_string => $status);
        
        if ($indirect_type != C_AGGR_32_BIT)
        {
                logresult('config', "CONFIG ERROR: Upgrading/downgrading to $attempted_rel requires a 32-bit root volume. Aggregate $root_aggr is $indirect_type-bit");
        }
        
        return C_RC_SUCCESS;
}

sub _revert_to_on_single_node
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                node => {type => OBJECT, isa => "NOTE::Filer"},
                release => {type => SCALAR},
                force => {type => SCALAR, default => TRUE,},
                max_attempts => {type => SCALAR, regex => qr/^\d+$/,},
                expected_requirements => {type => HASHREF,},
        });
        
        my $to_release = $args{release};
        my $force = $args{force};
        my $max_attempts = $args{max_attempts};
        my $current_attempt = 1;
        my $target = $args{node};
        my $target_name = $target->host();
        my $expected_requirements = $args{expected_requirements};
        
        logcomment("Reverting filer $target_name to release: $to_release");
        
        while ($current_attempt <= $max_attempts)
        {
                eval
                {
                        FS::Util::Utilities::revert_to($target_name, $to_release,
                                                       force => $force);
                };
                
		my $error = $@;
                if ($error)
                {
                        if ($current_attempt == $max_attempts)
                        {
                                logresult('fatal', "Unable to execute revert_to $to_release after $max_attempts attempts:\n" . $error);
                        }
                        else
                        {
                                logcomment("\nAttempt $current_attempt to revert to $to_release failed\n" .
                                           "Addressing any requirements and retrying...");
                                $class->_address_revert_requirements(error => $error,
                                                                     node => $target,
                                                                     expected_requirements => $expected_requirements);
                        }
                }
                else
                {
                        if (scalar(keys(%$expected_requirements)) > 0)
                        {
                                logresult('fatal', "revert_to $to_release succeeded. Failed to match expected requirements to be addressed:\n" . Dumper($expected_requirements))
                        }
                        last;
                }
                $current_attempt++;
        }
        
        return C_RC_SUCCESS;
}

sub _address_revert_requirements
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                node => {type => OBJECT, isa => "NOTE::Filer"},
                error => {type => SCALAR},
                expected_requirements => {type => HASHREF,},
        });
        my $error = $args{error};
        
        logcomment("\nAttempting to address revert_to requirements:\n\n$error");
        
        while (defined($error) && ($error ne ''))
        {
                $error = $class->_process_revert_requirements(%args);
                $args{error} = $error;
        }
        
        return C_RC_SUCCESS;
        
}

sub _process_revert_requirements
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                node => {type => OBJECT, isa => "NOTE::Filer"},
                error => {type => SCALAR},
                expected_requirements => {type => HASHREF,},
        });
        my $unprocessed_error = $args{error};
        my $target = $args{node};
        my $expected_requirements = $args{expected_requirements};
        
        logcomment("Entered _process_revert_requirements:\n$unprocessed_error") if ($class->{_debug});
        my @known_requirements = qw(REVERT_SIS_METAFILES
                                    SNAPMIRROR_BREAK
                                    SNAPMIRROR_OFF
                                    DELETE_SNAPSHOTS
                                    DECOMPRESS_VOLUME
                                    DISABLE_COMPRESSION
                                    SNAPVAULT_OFF);
        
        my @lines = split(/\n/, $unprocessed_error);
        my $first_line = shift(@lines);
        
        foreach my $possible_req (@known_requirements)
        {
                my $info = $class->_get_revert_requirements_info($possible_req);
                my $message = $info->{message};
                if (!defined($message))
                {
                        logresult('script', "Library problem found: revert requirements for $possible_req are not correctly defined (message)");
                }
                
                logcomment("Looking for $possible_req\n" .
                           "Processing line:\n$first_line") if ($class->{_debug});
                if ($first_line =~ /$message/)
                {
                        logcomment("Found known revert_to requirement: $possible_req");
                        if (!defined($info->{action}))
                        {
                                logresult('script', "Library problem found: revert requirements for $possible_req are not correctly defined (action)");
                        }
                        
                        my $entities = [];
                        &{$info->{action}}(\$unprocessed_error, $target, $entities);
                        
                        # Check if this was one of the expected requirements
                        if (exists($expected_requirements->{$possible_req}))
                        {
                                logcomment("Found expected revert_to requirement: $possible_req");
                                if ((!defined($expected_requirements->{$possible_req})) || ($expected_requirements->{$possible_req} eq ''))
                                {
                                        # No extra check, at the entity-level is needed
                                        delete($expected_requirements->{$possible_req});
                                }
                                else
                                {
                                        my $required_entity = $expected_requirements->{$possible_req};
                                        if (grep(/^$required_entity/, @$entities))
                                        {
                                                logcomment("Found expected revert_to requirement: $possible_req for required entity \"$required_entity\":\n" .
                                                           "Entities found:\n" . Dumper($entities));
                                                delete($expected_requirements->{$possible_req});
                                        }
                                }
                        }
                        
                        return $unprocessed_error
                }
        }
        $unprocessed_error = join("\n", @lines);
        
        return $unprocessed_error;
}


sub _get_revert_requirements_info
{
        my $class = shift;
        my $code = shift;
        my $ontap = $class->{_version}->{ontap};
        
        my $info = {'base' => {DELETE_SNAPSHOTS => {message => 'Newer snapshots on .+ that must be deleted prior to reverting',
                                                    action => sub {my $output_ref = shift;
                                                                   my $target = shift;
                                                                   my $found_entity = shift;
                                                                   my $output = ${$output_ref};
                                                                   
                                                                   my @lines = split(/\n/, $output);
                                                                   foreach my $line (@lines)
                                                                   {
                                                                        if ($line =~ /Newer snapshots on (\S+) (\S+) /)
                                                                        {
                                                                                my $type = $1;
                                                                                my $name = $2;
                                                                                push(@$found_entity, $name);
                                                                                logcomment("Deleting snapshots on $type \"$name\"");
                                                                                
                                                                                if ($type =~ /aggregate/)
                                                                                {
                                                                                        filer::filer_aggr_snap_delete_all($target->conn(), $name);
                                                                                }
                                                                                else
                                                                                {
                                                                                        # Best-effort try. Only deal with some failures. Don't stop if failure
                                                                                        my $response = $target->say("snap delete -a -f $name");
                                                                                        if ($response =~ /Remaining snapshots are currently in use/)
                                                                                        {
                                                                                                my $vol_status = Volume::vol_status($target->conn,$name);
                                                                                                my $volume_status = $vol_status->status();
                                                                                                if ($volume_status =~ /Volume has clones: (\S+), Volume UUID/)
                                                                                                {
                                                                                                        my @clones = split(/,/, $1);
                                                                                                        foreach my $clone (@clones)
                                                                                                        {
                                                                                                                logcomment("Splitting clone $name, to unlock the snapshots");
                                                                                                                Volume::Clone::vol_clone_split_start(conn => $target->conn,
                                                                                                                                                     clone => $clone,
                                                                                                                                                     wait_to_complete => 1);
                                                                                                        }
                                                                                                }
                                                                                                
                                                                                                # Try again to delete snapshots
                                                                                                filer::filer_vol_snap_delete_all($target->conn(), $name);
                                                                                        }
                                                                                }
                                                                                $output =~ s/$line\n*//;
                                                                                ${$output_ref} = $output;
                                                                        }
                                                                   }
                                                                   },},
                               REVERT_SIS_METAFILES => {message => 'Please revert SIS metafiles on Volume',
                                                        action => sub {my $output_ref = shift;
                                                                       my $target = shift;
                                                                       my $found_entity = shift;
                                                                       my $output = ${$output_ref};
                                                                       my $filer = $target->host();
                                                                       
                                                                       my @lines = split(/\n/, $output);
                                                                       foreach my $line (@lines)
                                                                       {
                                                                                if ($line =~ /Please revert SIS metafiles on Volume (\S+) to (\S+)/)
                                                                                {
                                                                                        my $vol = $1;
                                                                                        my $rel = $2;
                                                                                        my $path = '/vol/' . $vol;
                                                                                        push(@$found_entity, $vol);
                                                                                        
                                                                                        logcomment("Reverting sis metafiles of volume $vol to release \"$rel\"");
                                                                                        
                                                                                        # Dense::dense_revert_to does not handle correctly the case of "SIS metafiles upgrade not yet done" error
                                                                                        # First get the status, to check for this case...
                                                                                        my %status = Dense::dense_status(host => $filer,
                                                                                                                         path => $path,
                                                                                                                         long => 1,);
                                                                                        
                                                                                        my $last_op_error = (exists($status{$path}->{'Last Operation Error'})) ? ($status{$path}->{'Last Operation Error'}) : '';
                                                                                        if (($last_op_error) && ($last_op_error =~ /SIS metafiles upgrade not yet done. Run \'sis start\' manually on the volume/))
                                                                                        {
                                                                                                logcomment("Sis metafiles upgrade not yet done on volume $vol. Running sis start");
                                                                                                Dense::dense_start(host=>$filer, path_name=>$path);
                                                                                                Dense::dense_wait_for_status(host=>$filer, path_name=>$path, status=>"IDLE",);
                                                                                        }
                                                                                        
                                                                                        Dense::dense_revert_to(host => $filer,
                                                                                                               path_name => $path,
                                                                                                               version => $rel,
                                                                                                               validate => 0,);
                                                                                        
                                                                                        # Dense::dense_revert_to does not wait for sis revert_to to complete
                                                                                        # Also, dense_wait_for_status does not recognize the "Downgrading" stage
                                                                                        # Wait here (best-effort, do not fail if timeout)...
                                                                                        my $status = 'Downgrading';
                                                                                        my $timeout = C_DFLT_SIS_REVERT_TIMEOUT;
                                                                                        while (($timeout >= 0) && ($status =~ /Downgrading/))
                                                                                        {
                                                                                                %status = Dense::dense_status(host => $filer,
                                                                                                                              path => $path,
                                                                                                                              long => 0);
                                                                                                $status = $status{$path}->{'Status'};
                                                                                                sleep(10);
                                                                                                $timeout = $timeout - 10;
                                                                                        }
                                                                                        
                                                                                        $output =~ s/$line\n*//;
                                                                                        ${$output_ref} = $output;
                                                                                }
                                                                        }
                                                                       },},
                               SNAPMIRROR_OFF => {message => 'SnapMirror must be turned off|Turn SnapMirror off',
                                                  action => sub {my $output_ref = shift;
                                                                 my $target = shift;
                                                                 my $found_entity = shift;
                                                                 my $output = ${$output_ref};
                                                                 my $filer = $target->host();
                                                                 
                                                                 # No entity to be added -> filer-wide requirement
                                                                 logcomment("Turning snapmirror off on filer $filer");
                                                                 filer::filer_snapmirror_off($target->conn());
                                                                 
                                                                 my @lines = split(/\n/, $output);
                                                                 # Remove the first line
                                                                 shift(@lines);
                                                                 ${$output_ref} = join("\n", @lines);
                                                                 },},
                               SNAPVAULT_OFF => {message => 'SnapVault must be turned off',
                                                 action => sub {my $output_ref = shift;
                                                                my $target = shift;
                                                                my $found_entity = shift;
                                                                my $output = ${$output_ref};
                                                                my $filer = $target->host();
                                                                
                                                                # No entity to be added -> filer-wide requirement
                                                                logcomment("Turning snapvault off on filer $filer");
                                                                filer::filer_options_set($target->conn(), 'snapvault.enable', 'off');
                                                                
                                                                my @lines = split(/\n/, $output);
                                                                # Remove the first line
                                                                shift(@lines);
                                                                ${$output_ref} = join("\n", @lines);
                                                                },},
                               DECOMPRESS_VOLUME => {message => 'Volume .+ has compressed data. Must decompress the volume',
                                                     action => sub {my $output_ref = shift;
                                                                    my $target = shift;
                                                                    my $found_entity = shift;
                                                                    my $output = ${$output_ref};
                                                                    
                                                                    my @lines = split(/\n/, $output);
                                                                    foreach my $line (@lines)
                                                                    {
                                                                             if ($line =~ /Volume (\S+) has compressed data/)
                                                                             {
                                                                                      my $vol = $1;
                                                                                      push(@$found_entity, $vol);
                                                                                      logcomment("Decompressing volume \"$vol\"");
                                                                                      
                                                                                      my $utilities = Compression::CompressionUtilities->new(conn => $target,
                                                                                                                                             volume => $vol,
                                                                                                                                             boot_mode => C_ONTAP_7MODE,);
                                                                                      $utilities->disable_compression();
                                                                                      
                                                                                      my $decompress = Compression::Decompress->new(conn => $target,
                                                                                                                                    volume => $vol,
                                                                                                                                    boot_mode => C_ONTAP_7MODE,);
                                                                                      $decompress->start();
                                                                                      $decompress->wait_for_percentage();
                                                                                      
                                                                                      $output =~ s/$line\n*//;
                                                                                      ${$output_ref} = $output;
                                                                             }
                                                                    }
                                                                   },},
                               DISABLE_COMPRESSION => {message => 'Volume .+ has compression enabled. Turn off the "compression"',
                                                       action => sub {my $output_ref = shift;
                                                                    my $target = shift;
                                                                    my $found_entity = shift;
                                                                    my $output = ${$output_ref};
                                                                    
                                                                    my @lines = split(/\n/, $output);
                                                                    foreach my $line (@lines)
                                                                    {
                                                                             if ($line =~ /Volume (\S+) has compression enabled/)
                                                                             {
                                                                                      my $vol = $1;
                                                                                      push(@$found_entity, $vol);
                                                                                      logcomment("Disabling compression for volume \"$vol\"");
                                                                                      
                                                                                      my $utilities = Compression::CompressionUtilities->new(conn => $target,
                                                                                                                                             volume => $vol,
                                                                                                                                             boot_mode => C_ONTAP_7MODE,);
                                                                                      $utilities->disable_compression();
                                                                                      
                                                                                      $output =~ s/$line\n*//;
                                                                                      ${$output_ref} = $output;
                                                                             }
                                                                    }
                                                                   },},
                               SNAPMIRROR_BREAK => {message => 'run either "snapmirror break" or "snapvault stop"|volume \S+ or remove it from the filer',
                                                    action => sub {my $output_ref = shift;
                                                                   my $target = shift;
                                                                   my $found_entity = shift;
                                                                   my $output = ${$output_ref};
                                                                   my $filer = $target->host();
                                                                   my $host_rec = $target->hostrec;
                                                                   
                                                                   SnapMirror::snapmirror_enable(host=>$host_rec);
                                                                   my @lines = split(/\n/, $output);
                                                                   my $addressed = 0;
                                                                   foreach my $line (@lines)
                                                                   {
                                                                        if (($line =~ /run either \"snapmirror break\" or \"snapvault stop\" on all snapmirrored qtrees in (\S+)\./) ||
                                                                            ($line =~ /volume (\S+) or remove it from the filer/))
                                                                        {
                                                                                my $vol = $1;
                                                                                my $path = '/vol/' . $vol;
                                                                                push(@$found_entity, $vol);
                                                                                if ($addressed == 0)
                                                                                {
                                                                                        my $check_sv = 0;
                                                                                        logcomment("Getting snapmirror relationships for volume \"$vol\"");
                                                                                        my $relationships = SnapMirror::snapmirror_status_list(host => $host_rec,);
                                                                                        foreach my $relationship (@$relationships)
                                                                                        {
                                                                                                my $dest_path = $relationship->{'destination-location'};
                                                                                                my $source_path = $relationship->{'source-location'};
                                                                                                
                                                                                                # Do this as a best-effort try
                                                                                                eval {
                                                                                                logcomment("Breaking snapmirror relationship with destination \"$dest_path\"");
                                                                                                SnapMirror::snapmirror_quiesce(host=>$host_rec, dstpath=>$dest_path);
                                                                                                SnapMirror::snapmirror_break(host => $host_rec, dstpath => $dest_path);
                                                                                                logcomment("Releasing snapmirror relationship with source \"$source_path\"");
                                                                                                SnapMirror::snapmirror_release(host => $host_rec, srcpath => $source_path);
                                                                                                };
                                                                                                if ($@)
                                                                                                {
                                                                                                        $check_sv = 1;
                                                                                                }
                                                                                        }
                                                                                        
                                                                                        if ($check_sv == 1)
                                                                                        {
                                                                                                # Try snapvault stop (still best-effort)
                                                                                                my $sv_list = SnapVault::snapvault_status_list($target->conn(),);
                                                                                                foreach my $sv_rel (@$sv_list)
                                                                                                {
                                                                                                        my $dest_system = $sv_rel->{'destination-system'};
                                                                                                        my $dest_path = $sv_rel->{'destination-path'};
                                                                                                        if ($dest_system eq $filer)
                                                                                                        {
                                                                                                                logcomment("Stopping snapvault backup relationship with destination path $dest_path");
                                                                                                                eval { SnapVault::sv_stop($target->conn(), $dest_path); };
                                                                                                        }
                                                                                                }
                                                                                        }
                                                                                        
                                                                                        $addressed = 1;
                                                                                }
                                                                                
                                                                                $output =~ s/$line\n*//;
                                                                                ${$output_ref} = $output;
                                                                        }
                                                                   }
                                                                   },},
                               },
                    };
        
        if (exists($info->{$ontap}))
        {
                return $info->{$ontap}->{$code};
        }
        else
        {
                return $info->{base}->{$code};
        }
}


sub _check_current_release
{
        my $class = shift;
        my $user_defined_current_rel = shift;
        
        $class->{_version}->refresh_sw_version(conn => $class->{_conn});
        my $filer_current_major_rel = $class->{_version}->{ontap};
        
        # $filer_current_major_rel will only be the major release.
        # Need to substract major release from the user-defined value,
        # otherwise version_compare will not evaluate correctly
        $user_defined_current_rel =~ /^(\d+\.\d+)/;
        my $user_defined_current_major_rel = $1;
        if (!defined($user_defined_current_major_rel))
        {
                return C_RC_FAILURE;
        }
        
        # Check the user-defined current release (ontap): 
        if (Filer::Version::version_compare($user_defined_current_major_rel, $filer_current_major_rel) != 0)
        {
                logcomment("\nFailed to match expected major release on the filer " . $class->{_filer} . "\n" .
                           "Current filer major release: $filer_current_major_rel\n" .
                           "Expected release: $user_defined_current_rel");
                return C_RC_FAILURE;
        }
        
        return C_RC_SUCCESS;
}


sub _install_base_release
{
        my $class = shift;
        
        my $required_base_release = $class->{_required_base_release};
        my $current_release = $class->{_initial_release};
        
        my $compare_result = Filer::Version::version_compare($required_base_release, $current_release);
        
        if ($compare_result == 0)
        {
                logcomment("\nNo initialization necessary for filer " . $class->{_filer} . " for upgrade/revert testing.\n" .
                           "Required base release: $required_base_release\n" .
                           "Current release is: $current_release");
                return C_RC_SUCCESS;
        }
        
        logcomment("\nInitializing filer " . $class->{_filer} . " for upgrade/revert testing.\n" .
                   "Required base release: $required_base_release\n" .
                   "Current release is: $current_release");
        
        if ($compare_result > 0)
        {
                # Simply an upgrade
                $class->upgrade_to_release(release => $required_base_release,
                                           url => $class->{_base_release_url},
                                           reboot => TRUE,);
        }
        else
        {
                logcomment("\nDowngrade operation required for upgrade/revert initialization.");
                
                # Determine if there is any custom intermediary release required
                my $downgrade_info = $class->_get_downgrade_info(from_release => $current_release,
                                                                 to_release => $required_base_release,);
                
                if ((defined($downgrade_info)) && (exists($downgrade_info->{intermediary_release})))
                {
                        logcomment("\nUpgrade/revert initialization: Upgrading to required intermediary release: " . $downgrade_info->{intermediary_release});
                        $class->upgrade_to_release(release => $downgrade_info->{intermediary_release},
                                                   url => $main::INTERM_REL_URL,
                                                   reboot => TRUE);
                }
                
                # Install the base release (no reboot) 
                logcomment("\nUpgrade/revert initialization: Downgrading to required base release: $required_base_release");
                $class->upgrade_to_release(release => $required_base_release,
                                           url => $class->{_base_release_url},
                                           reboot => FALSE);
                
                if ((defined($downgrade_info)) && (exists($downgrade_info->{clear_ndm_config})) && ($downgrade_info->{clear_ndm_config} == TRUE))
                {
                        ## See burt 386448 for details
                        $class->_clear_ndm_config($class->{_conn});
                        if ($class->{_ndu_testing} == TRUE)
                        {
                                $class->_clear_ndm_config($class->{_partner_conn});
                        }
                }
                
                # This is a downgrade, revert_to is needed...
                my $revert_to;
                if ((defined($downgrade_info)) && (exists($downgrade_info->{revert_to_needed})))
                {
                        $revert_to = $downgrade_info->{revert_to_needed};
                }
                
                if (!defined($revert_to))
                {
                        # Revert to the latest release allowed by the filer
                        my $transform = Filer::Transform->new($class->{_conn}->conn(),
                                                              to_version => $required_base_release,);
                        my $prev_releseas = $transform->prev_versions();
                        
                        # Obtain the latest release available, that is smaller than or equal to base_release
                        $required_base_release =~ /^(\d+\.\d+)/;
                        my $major_base_release = $1;
                        foreach my $prev_rel (@$prev_releseas)
                        {
                                if (Filer::Version::version_compare($major_base_release, $prev_rel) >= 0)
                                {
                                        if ((!defined($revert_to)) ||
                                            (defined($revert_to) && (Filer::Version::version_compare($prev_rel, $revert_to) > 0)))
                                        {
                                                $revert_to = $prev_rel;
                                        }
                                }
                        }
                        
                        if (!defined($revert_to))
                        {
                                logresult('config', "Unable to detect available version to revert to when installing the required base release");
                        }
                }
                
                logcomment("\nUpgrade/revert initialization: Reverting to release: $revert_to");
                $class->revert_to_release(release => $revert_to, force => TRUE);
                
                # Check that after the upgrade, major releases match...
                if ($class->_check_current_release($required_base_release) == C_RC_FAILURE)
                {
                        logresult('config', "\n\nError: user-defined base release ($required_base_release) " .
                                            "does not match current major release of the filer " . $class->{_filer} . " after the initialization\n\n");
                }
        }
        
        return C_RC_SUCCESS;
}


sub _clear_ndm_config
{
        my $class = shift;
        my $target = shift;
        my $target_name = $target->host();
        
        logcomment("\nUpgrade/revert initialization: Clearing the ndm.config file on filer $target_name");
        # For now have it like this. If any cmd change, this should be implemented in a release-specific manner.
        $target->say('priv set test; setflag wafl_metadata_visible 1; rm /vol/%s/metadir/vdisk/ndm.config', $target->vol_root(),);
}


sub _get_downgrade_info
{
        my $class = shift;
        my %args = validate(@_,
        {
		from_release => {type => SCALAR, regex => qr/^\d+\.\d+\.\d+$/,},
                to_release => {type => SCALAR, regex => qr/^\d+\.\d+\.\d+$/,},
	} );
        
        my $from_release = $args{from_release};
        my $to_release = $args{to_release};
        
        # First-level hash is keyed by "from_release"
        # Second-level hash is keyed by "to_release"
        my $downgrade_info = {'8.0.1' => {'8.0.0' => {intermediary_release => '8.1.0',
                                                      revert_to_needed => '8.0',
                                                      clear_ndm_config => TRUE,},},
                              '8.1.0' => {'8.0.0' => {revert_to_needed => '8.0',
                                                      clear_ndm_config => TRUE,},
					  '8.0.1' => {revert_to_needed => '8.0',
						      clear_ndm_config => TRUE,},},
			      '8.1.1' => {'8.0.0' => {revert_to_needed => '8.0',
						      clear_ndm_config => TRUE,},
					  '8.0.1' => {revert_to_needed => '8.0',
						      clear_ndm_config => TRUE,},},
                              };
        
        if (exists($downgrade_info->{$from_release}->{$to_release}))
        {
                return $downgrade_info->{$from_release}->{$to_release};
        }
        else
        {
                return undef;
        }
}


1;
