#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/7Mode/Vfiler.pm#1 $
#

## @summary Compression Vfiler routines (7-mode).
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (internal)


package Compression::7Mode::Vfiler;

use strict;
use warnings;

use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult _send_cmd_to_filer);


BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
        
        # Keep a global hash to keep track of instances created for each filer.
        # This is needed in the case of multiple objects per run
        # The second+ objects have to be recycled,
        # otherwise the first volume will be transfered to the hosting filer
        $Compression::7Mode::Vfiler::Instance_counts = {};
}


## @name ::new
## @summary Compression::Vfiler Constructor.
##
## @description
## Creates an object representing a vfiler.
##
## @return Object of type Compression::Vfiler.
##
## @arg name Required: Name of the vfiler.
##
## @arg ip Required: IP of the vfiler.
##
## @arg interface Required: Interface for the vfiler
##
## @arg root_vol_aggr Required: NOTE::Aggregate object for the vfiler's root aggregate.
##
## @arg root_vol_name Optional: Name for the vfiler's root volume.
##
## @arg root_vol_size Optional: Size for the vfiler's root volume.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @arg disable_cleanup Optional: Flag to disable any clean-up that is executed
## upon object destruction.
## Defaults to FALSE.
##
sub new
{
	my $class = shift;
	my (%args) = validate( @_,
		{
                        name => {type => SCALAR,
                                 },
                        
                        ip => {type => SCALAR,
                               },
                        
                        interface => {type => SCALAR,
                                      },
                        
                        root_vol_aggr => {type => OBJECT,
                                          isa => "NOTE::Aggregate",
                                         },
                        
                        root_vol_name => {type => SCALAR,
                                          optional => 1,
                                         },
                        
                        root_vol_size => {type => SCALAR,
                                          optional => 1,
                                          default => C_DFLT_VOL_SIZE,
                                          },
                        
                        debug 	=> {type => SCALAR,
				    optional => 1,
				    default => FALSE,
				    },
                        
                        disable_cleanup => {type     => BOOLEAN,
					    optional => 1,
					    default  => FALSE,
					    },
		} );

	my $self = {};
        $self->{_debug} = $args{debug};
        $self->{_disable_cleanup} = $args{disable_cleanup};
        $self->{_creator_pid} = $$;
        my $name = $args{name};
        $self->{name} = $name;
        $self->{_ip} = $args{ip};
        $self->{_interface} = $args{interface};
        
        my $root_vol = (defined($args{root_vol_name})) ? ($args{root_vol_name}) : ($self->{name} . '_root');
        $root_vol =~ s/-/_/g;
        $self->{_root_vol_name} = $root_vol;
        my $root_vol_aggr = $args{root_vol_aggr};
        
        # Decide if the vfiler should be recycled
        # Since we will usually have more than one object per script,
        # we need to recycle the second+ ones
        my $recycle = C_DFLT_RECYCLE_VFILER;
        logcomment("Vfiler: starting object creation for $name : Instance counts:\n" .
                   Dumper($Compression::7Mode::Vfiler::Instance_counts)) if ($self->{_debug});
        my $instance_count = $Compression::7Mode::Vfiler::Instance_counts->{$name};
        if ((defined($instance_count)) && ($instance_count > 0))
        {
		# Note: due to burt 442667, for now do not create more than one instance of the vfiler object
		return undef;
                # Override recycle if this is not the first instance
                $recycle = TRUE;
        }
        
        $self->{_recycled} = $recycle;
        
        if ($self->{_recycled} == FALSE)
        {
                my $existent_root_vol = NOTE::Volume->attach( name => $root_vol, Filer => $root_vol_aggr->Filer(), force => 1 );
                if ($existent_root_vol->exists())
                {
                        $existent_root_vol->purge();
                }
        }
        
        logcomment("Starting the vfiler creation process");
        
        logcomment("Adding required license");
        $root_vol_aggr->Filer->license(C_LICENSE_MULTISTORE);
	$self->{_filer_name} = $root_vol_aggr->Filer->host();
        
        # Looks like NOTE is purging vfiler by default, even if it is declared recycled
        my $purge_existing = ($recycle == TRUE) ? FALSE : TRUE;
        logcomment("Creating the vfiler root volume");
        $self->{_root_vol} = NOTE::Volume->create(Filer => $root_vol_aggr->Filer(),
                                                  Aggregate => $root_vol_aggr,
                                                  name => $root_vol,
                                                  size => $args{root_vol_size},
                                                  recycle => $recycle,
                                                  );
        
        logcomment("Creating/reusing the vfiler");
        $self->{_vfiler} = NOTE::Vfiler->create(name => $self->{name},
                                                ipaddr => $self->{_ip},
                                                interface => $self->{_interface},
                                                Rootvol => $self->{_root_vol},
                                                recycle => $recycle,
                                                purge_existing => $purge_existing, 
                                                );
        
        if (!defined($self->{_vfiler}))
        {
                logresult('config', "Failed to create vfiler " . $self->{name});
        }
	
	$self->{_password} = $self->{_vfiler}->password();
	$self->{_username} = $self->{_vfiler}->username();
	
        
        bless $self, $class;
        
        # Keep track of instances
        if (!defined($instance_count))
        {
                $instance_count = 1;
        }
        else
        {
                $instance_count++;
        }
        $Compression::7Mode::Vfiler::Instance_counts->{$name} = $instance_count;
        
	return $self;
}


## @name ::start_migrate
## @summary Starts vfiler migrate for this vfiler.
##
## @description
## Starts vfiler migrate for this vfiler.
## This operation has three logical steps, and
## should be performed in three stages. First, issue the "start" command which
## initiates the movement of the remote vfiler's storage to the local filer using
## SnapMirror. Then you can use the "status" command to monitor the status of this
## data movement. When this status for each path changes from "Being initialized"
## to "SnapMirrored", you can issue the "complete" command to finish the
## migration. When the "complete" command completes, the remote vfiler will have been
## moved to the local machine.
##
## @return Hash with rc and output.
##
## @arg dest_filer Required: Name of the destination filer.
##
## @arg interface Optional: Destination interface. Defaults to the vfiler's interface.
##
## @example
## N/A - to be used through the compress object ( $compress->start_vfiler_migrate(dest_filer => $MIGRATE_TO); )
##
sub start_migrate
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			dest_filer => {type => SCALAR,},
			interface => {type => SCALAR, optional => 1, default => $class->{_interface}},
		});
	
	my $dest_filer = $args{dest_filer};
	
	logcomment("Starting the vfiler migration process. Destination filer: $dest_filer");
	$class->{_dest_filer_obj} = NOTE::Filer->connect(host => $dest_filer,); 
	$class->{_dest_filer_name} = $dest_filer; 
	$class->{_dest_filer_version} = Compression::Version->new(conn => $class->{_dest_filer_obj}, debug => $class->{_debug},);
	$class->{_dest_filer_ui} = Compression::UI->new(version => $class->{_dest_filer_version}, debug => $class->{_debug}, boot_mode => C_ONTAP_7MODE,);
	$class->{_dest_interface} = $args{interface};
	
	logcomment("Adding required license on the destination filer");
	$class->{_dest_filer_obj}->license(C_LICENSE_MULTISTORE);
	
	my $return = $class->{_vfiler}->migrate_start(filer => $class->{_dest_filer_obj},
						      if_to_use => $class->{_dest_interface},);
	
	return (rc => C_RC_SUCCESS, output => $return);
}


## @name ::get_migrate_status
## @summary Gets migrate status for this vfiler.
##
## @description
## Gets migrate status for this vfiler.
##
## @return Hash with rc, migrate_status, and output.
##
## @example
## my %res = $obj->get_migrate_status();
##
sub get_migrate_status
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			# None
		});
	
	if (!defined($class->{_dest_filer_obj}))
	{
		logresult('script', "No vfiler migrate has been started yet by this library");
	}
	
	my $cmd_code = C_CMD_VFILER_MIGRATE_STATUS;
	my $command = $class->{_dest_filer_ui}->formulate_command(code   => $cmd_code,
								  filer => $class->{_filer_name},
								  vfiler => $class->{name},);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_dest_filer_obj},
                                                            boot_mode => C_ONTAP_7MODE,
                                                            cmd => $command,);
	
	return $class->{_dest_filer_ui}->process_output_string(code => $cmd_code,
							       output => $output,
							       expected_rc => C_RC_SUCCESS,
							       log_message => "Obtain vfiler migrate status for vfiler " . $class->{name},);
}


## @name ::wait_until_ready_for_complete_migrate
## @summary Wait until ready to issue the 'migrate complete' command.
##
## @description
## Wait until ready to issue the 'migrate complete' command.
##
## @return Hash with rc, migrate_status, and output.
##
## @arg timeout Optional: Timeout value. Defaults to C_DFLT_TIMEOUT.
##
## @example
## $obj->wait_until_ready_for_complete_migrate();
##
sub wait_until_ready_for_complete_migrate
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			timeout => {type => SCALAR, default => C_DFLT_TIMEOUT,},
		});
	
	my $timeout = $args{timeout};
	my $found_vol_not_ready = TRUE;
	my %res;
	
	while (($timeout >= 0) && (TRUE == $found_vol_not_ready))
	{
		%res = $class->get_migrate_status();
		my $status = $res{migrate_status};
		
		my $ready_to_complete = $class->{_dest_filer_ui}->get_constant_value(constant_code => 'VFILER_MIGRATE_IS_READY_FOR_COMPLETE');
		$found_vol_not_ready = FALSE;
		foreach my $vol (keys(%$status))
		{
			if ($status->{$vol}->{status} ne $ready_to_complete)
			{
				$found_vol_not_ready = TRUE;
				logcomment("Vol $vol is not ready for vfiler migrate completion (timeout=$timeout): " . $status->{$vol}->{status});
				last;
			}
		}
		$timeout--;
	}
	
	if (0 == $timeout)
	{
		logresult('fatal', "Timed-out while waiting for vfiler migrate to be ready for completion:\n\n" . Dumper(%res))
	}
	
	return %res;
}


## @name ::complete_migrate
## @summary Complete the vfiler migrate operation.
##
## @description
## Complete the vfiler migrate operation.
##
## @return Hash with rc and output.
##
## @example
## $obj->complete_migrate();
##
sub complete_migrate
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			# None
		});
	
	if (!defined($class->{_dest_filer_obj}))
	{
		logresult('script', "No vfiler migrate has been started yet by this library");
	}
	
	my $return = $class->{_vfiler}->migrate_complete(filer => $class->{_dest_filer_obj},);
	
	return (rc => C_RC_SUCCESS, output => $return);
}


## @name ::get_controller_name
## @summary Get name of the hosting filer.
##
## @description
## Get name of the hosting filer.
##
## @return Name of the hosting filer.
##
## @example
## my $controller = $obj->get_controller_name();
##
sub get_controller_name
{
	my $class = shift;
	
	return $class->{_filer_name};
}


## @name ::get_root_aggr_name
## @summary Get name of the aggregate hosting the vfiler root volume.
##
## @description
## Get name of the aggregate hosting the vfiler root volume.
##
## @return Name of the aggregate hosting the vfiler root volume.
##
## @example
## my $root_aggr_name = $obj->get_root_aggr_name();
##
sub get_root_aggr_name
{
	my $class = shift;
	
	return $class->{_root_vol}->Aggregate->name();
}




################################################################
# Internal subroutines
################################################################

sub _get_underlying_obj
{
        my $class = shift;
	my (%args) = validate( @_,
	{
		# None
	} );
        
        return $class->{_vfiler};
}

sub _attach_to_migrated_vfiler
{
	my $class = shift;
	
	if (!defined($class->{_dest_filer_obj}))
	{
		logresult('script', "No vfiler migrate has been started yet by this library");
	}
	
	# First, destroy the vfiler root volume left behind on the initial controller
	if ($class->{_recycled} == FALSE)
	{
		logcomment("Purging left-behind vfiler root volume");
		$class->{_root_vol}->purge( );
	}
	
	# Now attach to the migrated vfiler root volume
	$class->{_interface} = $class->{_dest_interface};
	$class->{_filer_name} = $class->{_dest_filer_name};
	$class->{_root_vol} = NOTE::Volume->attach(Filer => $class->{_filer_name},
						   name => $class->{_root_vol_name},);
	
	$class->{_vfiler} = NOTE::Vfiler->attach(name => $class->{name},
						 filer => $class->{_filer_name},
						 username => $class->{_password},
						 password => $class->{_username},);
	
	if (!defined($class->{_vfiler}))
        {
                logresult('config', "Failed to attach to migrated vfiler " . $class->{name});
        }
	
	$class->{_dest_filer_obj} = undef;
}


sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
        my $vfiler = $class->{name};
	
	if (($class->{_disable_cleanup} == FALSE) &&
                ($class->{_creator_pid} == $$) &&
                (!defined($ENV{'C_CORRUPTION_FOUND'}) || ($ENV{'C_CORRUPTION_FOUND'} != TRUE)))
	{
		logcomment("Executing Vfiler clean-up for \"$vfiler\"");
		
                eval
                {
                        # It looks like the root volume does not get cleaned-up by NOTE
                        if ($class->{_recycled} == FALSE)
                        {
				$class->{_vfiler}->{_recycle} = 0;
                                $class->{_vfiler}->DESTROY();
                                logcomment("Vfiler: cleanup: purging vfiler root volume") if ($class->{_debug});
                                $class->{_root_vol}->purge( );
                        }
                        
                };
                
                if( $@ )
                {
                        logcomment( "Vfiler cleanup for \"$vfiler\": failure due to : $@" );
                }
                else
                {
                        logcomment( "Vfiler cleanup for \"$vfiler\": done" ) if ($class->{_debug});
                }
                
                my $name = $class->{name};
                $Compression::7Mode::Vfiler::Instance_counts->{$name} = $Compression::7Mode::Vfiler::Instance_counts->{$name} - 1;
                logcomment("Vfiler: cleanup for vfiler $name finished: Instance counts:\n" .
                           Dumper($Compression::7Mode::Vfiler::Instance_counts)) if ($class->{_debug});
        }
	else
	{
                logcomment( "Vfiler cleanup for \"$vfiler\": nothing done: script PID: " . $$) if ($class->{_debug});
	}
        
        if (defined($execution_error))
        {
                _c_interpret_logresult($execution_error);
        }
}


#routine to test the module itself
sub testme
{

	package main;
        use Data::Dumper;
	
        require "drg/Compression/setup_core.pl";
        
        use vars qw($DEBUG $VOLUME $MIGRATE_TO $VFILER $VFILER_IP $VFILER_IF);
        
        param("DEBUG", "-default", FALSE);
	param("MIGRATE_TO", "-default", undef);
	param("VFILER", "-mesg", "Vfiler name");
	param("VFILER_IP", "-mesg", "Vfiler IP");
	param("VFILER_IF", "-mesg", "Vfiler interface");
        
        
        param("DEBUG", "-default", FALSE);
        
        my $compress = Compression::Compress->setup(filer => $FILER ,
                                                    aggregate => $AGGREGATE,
                                                    volume => {name => 'vfiler_testme_vol',
                                                               size => "3g",
                                                               },
						    vfiler => {name => $VFILER,
							       ip => $VFILER_IP,
							       interface => $VFILER_IF,
							       },
                                                    debug => $DEBUG,
                                                    );
	
	my $base = Compression::Compress->setup(filer => $FILER ,
						aggregate => $AGGREGATE,
						volume => {size => "3g",
							   },
						debug => $DEBUG,
						);
        
        
        my $large_file = 'large_file';
        my $large_file_size = 2 * GB; 
        my $large_file_savings = 1833980;
        my $modify_size = MB;
        my $modify_savings = 892;
        my $vol_mountpoint = $compress->{dataset}->get_mountpoint();
        
        logcomment("Testing basic compression functionality");
        $compress->wait_for_cp();
        $compress->is_enabled();
        $compress->disable();
        $compress->is_enabled(expected => C_RC_FAILURE);
        $compress->toggle();
        $compress->wait_for_cp();
        $compress->is_enabled();

        $compress->{dataset}->create(files => [{ name => $large_file, size => $large_file_size, },],);
	$base->{dataset}->create(files => [{ name => $large_file, size => $large_file_size, },],);
	my $test_path = $compress->{dataset}->get_path(filename => $large_file);
	my $base_path = $base->{dataset}->get_path(filename => $large_file);
        $compress->wait_for_cp();
        $compress->verify_savings(expected_compressed => $large_file_savings,
                                  allow_percentage_error => 20);

        # Set up vsr testing
        $compress->{filer}->license_add(service => C_LICENSE_SNAPRESTORE);
        my $snapshot = 'Vfiler_testme_sn';
        $compress->{volume}->create_snapshot(name => $snapshot);
        
        logcomment("Testing basic decompress functionality");
        $compress->disable();
        $compress->wait_for_cp();
        $compress->{decompress}->start();
        $compress->{decompress}->status();
        $compress->{decompress}->stop();
        $compress->wait_for_cp();
        $compress->{decompress}->status(expected => C_RC_NO_DECOMPRESSION_IN_PROGR);
        $compress->{decompress}->start();
        $compress->{decompress}->wait_for_percentage();
        $compress->verify_savings(expected_compressed => 0);
        
        logcomment("Testing basic compression scanner functionality");
        $compress->enable();
        $compress->wait_for_cp();
        $compress->{compress_scanner}->start();
        $compress->{compress_scanner}->status();
        $compress->{compress_scanner}->stop();
        $compress->wait_for_cp(num_cp => 5); 
        $compress->{compress_scanner}->status(expected => C_RC_SCANNER_IDLE);
        $compress->{compress_scanner}->wait_for_scanned_data();
        $compress->verify_savings(expected_compressed => $large_file_savings,
                                  allow_percentage_error => 20);
        
        logcomment("Testing basic dataset functionality");
        $compress->{dataset}->modify(files =>[{ name => $large_file, offset => 0, len => $modify_size},],);
        $compress->wait_for_cp();
        $compress->verify_savings(expected_compressed => $large_file_savings,
                                  allow_percentage_error => 20);
        $compress->{dataset}->truncate_file( filename => $large_file, size => $modify_size );
        $compress->wait_for_cp();
        $compress->verify_savings(expected_compressed => $modify_savings,
                                  allow_percentage_error => 20);
        my $replace_file = 'replace_file';
        $compress->{dataset}->create(files =>[{ name => $replace_file, offset => 0, len => $modify_size},],);
        my $replace_file_path = $compress->{dataset}->get_path(filename => $replace_file);
        $compress->wait_for_cp();
        $compress->verify_savings(expected_compressed => (2 * $modify_savings),
                                  allow_percentage_error => 20);
        $compress->{dataset}->replace_file(filename => $large_file,
                                           new_file => $replace_file_path);
        $compress->wait_for_cp();
        $compress->verify_savings(expected_compressed => (2 * $modify_savings),
                                  allow_percentage_error => 20);
        my $copy_file = $large_file . '_copy';
        $compress->{dataset}->copy_file(source => $replace_file_path,
                                        dest => $vol_mountpoint . "/$copy_file",);
        my $copy_file_path = $compress->{dataset}->get_path(filename => $copy_file);
        $compress->wait_for_cp();
        $compress->verify_savings(expected_compressed => (3 * $modify_savings),
                                  allow_percentage_error => 20);
        $compress->{volume}->flush_caches();
        $compress->{dataset}->compare_files(file_1=> $replace_file_path,
                                            file_2=> $copy_file_path);
        
        my $moved_file = $large_file . '_moved';
        $compress->{dataset}->move_file(source => $replace_file_path,
                                        dest => $vol_mountpoint . "/$moved_file",);
        $compress->wait_for_cp();
        $compress->verify_savings(expected_compressed => (3 * $modify_savings),
                                  allow_percentage_error => 20);
        my $moved_file_path = $compress->{dataset}->get_path(filename => $moved_file);
        $compress->{dataset}->compare_files(file_1=> $moved_file_path,
                                            file_2=> $copy_file_path);
        $compress->{dataset}->delete_file(filename => $moved_file);
        $compress->wait_for_cp();
        $compress->verify_savings(expected_compressed => (2 * $modify_savings),
                                  allow_percentage_error => 20);
        
        # Cannot be tested yet --> panic, burt 442267
        #logcomment("Testing volume snap restore");
        #$compress->{volume}->restore_snapshot(snapshot_name => $snapshot);
        #$compress->wait_for_cp();
        #$compress->is_enabled();
        #$compress->verify_savings(expected_compressed => $large_file_savings,
        #                          allow_percentage_error => 20);
        #
        #$compress->{dataset}->create(files => [{ name => $large_file, size => $modify_size, },],);
        #$compress->wait_for_cp();
        #$compress->verify_savings(expected_compressed => ($large_file_savings + $modify_savings),
        #                          allow_percentage_error => 20);
	
	if (defined($MIGRATE_TO))
	{
		logcomment("Testing vfiler migrate (destination controller: $MIGRATE_TO)");
		$compress->start_vfiler_migrate(dest_filer => $MIGRATE_TO);
		$compress->{vfiler}->wait_until_ready_for_complete_migrate();
		$compress->{vfiler}->complete_migrate();
		
		# This is the left-behind volume
		$compress->detach_from_vfiler();
		$compress->{decompress}->status(expected => C_RC_NO_DECOMPRESSION_IN_PROGR);
		$compress->{compress_scanner}->status(expected => C_RC_SCANNER_IDLE);
		$compress->get_savings();
		$compress->{dataset}->ls();
		
		# This is the migrated volume, on the new controller
		$compress->refresh_vfiler_controller();
		$compress->{dataset}->ls();
		$compress->{decompress}->status(expected => C_RC_NO_DECOMPRESSION_IN_PROGR);
		$compress->{compress_scanner}->status(expected => C_RC_SCANNER_IDLE);
		$compress->get_savings();
		$compress->{sis}->get_state();
		
		$compress->{dataset}->compare_files(file_1 => $base_path,
						    file_2 => $test_path,
						    expected_result => 'NOT_EQUAL',);
	}
        
        logresult('PASS', "Vfiler: testme: Done");
}


sub testme_cfo
{

	package main;
        use Data::Dumper;
	
        require "drg/Compression/setup_core.pl";
        
        
        my $cfo_pair = Compression::Cluster->new(filer => $FILER);
        
        my $local_vol = Compression::Compress->setup(filer => {name => $FILER,
                                                               require_cf_disabled => FALSE},
                                                     aggregate => {name => $AGGREGATE},
                                                     volume => {name => $VOLUME,
                                                                size => '5g',},);
        
        $cfo_pair->get_status;
        $local_vol->{dataset}->create(files => [{name => 'file', size => 2 * GB,},],);
        $local_vol->wait_for_cp();
        $local_vol->verify_savings(expected_compressed    => 1833980,
                                   allow_percentage_error => 10);
        
        $cfo_pair->takeover;
        $cfo_pair->monitor(me => 'TAKEOVER', partner => 'DEAD');
        $cfo_pair->get_status;
        
        $local_vol->disable();$local_vol->{decompress}->start();
        $local_vol->{decompress}->wait_for_percentage(min => 2, max => 55);
        $local_vol->{decompress}->stop();
        $local_vol->{decompress}->status(expected => C_RC_NO_DECOMPRESSION_IN_PROGR);
        $local_vol->{compress_scanner}->status(expected => C_RC_SCANNER_IDLE);
        $local_vol->is_enabled(expected => C_RC_FAILURE);
        $local_vol->enable();
        $local_vol->{compress_scanner}->status(expected => C_RC_SCANNER_IDLE);
        
        $cfo_pair->giveback;
        $cfo_pair->monitor(me => 'UP', partner => 'UP');
        $cfo_pair->get_status;
        
        logresult('PASS', "Vfiler: testme_cfo: Done");
        
}


1;
