#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/7Mode/VolCopy.pm#1 $
#

## @summary Volume copy routines.
##
## @description
##    This module provides all the capabilities
##    required for 7-mode volume copy.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (internal)

package Compression::7Mode::VolCopy;

use strict;
use warnings;

use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Tharn;

use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult);
use Volume::Copy qw(volcopy_throttle);


BEGIN
{
	require "Compression/base_import.pl";
}



## @name ::new
##
## @summary Compression::7Mode::VolCopy Constructor.
##
## @description
## Creates an Compression::7Mode::VolCopy object 
## (it does not start the vol copy relationship on the filer).
##
## @return Compression::7Mode::VolCopy object.
##
## @arg source Required: Source Compression::Compress object.
##
## @arg destination Required: Hashref of destination arguments (volume, aggregate, filer).
##
## @arg disable_cleanup Optional: Flag to disable any clean-up
## that would be executed upon object destruction.
## Defaults to $main::DISABLE_CLEANUP.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to $main::DEBUG.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
		source 	=> {type => OBJECT,
                            isa => "Compression::Compress",},
                
                destination => {type => HASHREF,},
                
                disable_cleanup => {type => BOOLEAN,
                                    optional => 1,
                                    default => $main::DISABLE_CLEANUP,},
                
                debug => {type => BOOLEAN,
                          optional => 1,
                          default => $main::DEBUG,},
		} );
	
	my $self = {};
        $self->{_source} = $args{source};
        $self->{_disable_cleanup} = $args{disable_cleanup};
        $self->{_debug} = $args{debug};
        $self->{_creator_pid} = $$;
        $self->{_boot_mode} = $self->{_source}->{boot_mode};
        $self->{_vserver} = $self->{_source}->{vserver};
        
        $self->{_src_vol_name} = $self->{_source}->{vol_name};
        $self->{_src_aggr_name} = $self->{_source}->{aggr_name};
        $self->{_src_host_name} = $self->{_source}->{host_name};
        
        my $dst = $args{destination};
        my %dst_args = validate_with
                (params => $dst,
                 spec   => {volume => { type => SCALAR|HASHREF|UNDEF, },
                            aggregate => { type => SCALAR|HASHREF, default => $self->{_src_aggr_name}, },
                            filer => { type => SCALAR|HASHREF, default => $self->{_src_host_name}, },},
                 allow_extra => 0,);
        
        $self->{_ui} = $self->{_source}->{_ui};
        
        $self->{_destination} = Compression::Compress->setup(filer => $dst_args{filer},
                                                             aggregate => $dst_args{aggregate},
                                                             volume 	=> $dst_args{volume},
                                                             debug => $self->{_debug},
                                                             disable_cleanup => $self->{_disable_cleanup});
        $self->{_destination}->wait_for_cp();
        
        $self->{_dst_vol_name} = $self->{_destination}->{vol_name};
        $self->{_dst_aggr_name} = $self->{_destination}->{aggr_name};
        $self->{_dst_host_name} = $self->{_destination}->{host_name};
        
        $self->{_copy} = NOTE::Volume::Copy->new(src => $self->{_source}->{volume}->_get_underlying_vol_obj(),
                                                 dst => $self->{_destination}->{volume}->_get_underlying_vol_obj(),);
        
        
        bless $self, $class; 

	return $self;
}


## @name ::attach_to_destination_vol
## @summary Gets the compress destination object.
## 
## @description
## Gets the compress destination object. It will return a compress object
## corresponding to a volume created with the default settings.
##
## @return Compression::Compress object corresponding to the destination volume.
##
## @examples
## my $dest = $obj->attach_to_destination_vol();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub attach_to_destination_vol
{
        my $class = shift;
        my %args = validate(@_,
                {
                        # None
                });
        
        return $class->{_destination};
}



## @name ::start
## @summary Starts the vol copy process.
## 
## @description
## Starts the vol copy process.
##
## @return Hash with rc.
## 
## @arg background Optional: Flag to run the transfer in the background.
## To not be set if the expected result is not success.
## Default: FALSE.
## 
## @arg copy_all Optional: Flag to include all the volume's snapshots.
## Default: FALSE.
## 
## @arg snapshot Optional: The name of the snapshot, on the source volume, to copy.
##
## @arg timeout Optional: Timeout, in seconds, for the command to complete.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @examples
## $obj->start();
## $obj->start(copy_all => TRUE);
## $obj->start(snapshot => 'my_snapshot');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub start
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        background => {type => SCALAR, optional => 1, default => FALSE,},
                        copy_all => {type => SCALAR, optional => 1, default => FALSE,},
                        snapshot => => {type => SCALAR, optional => 1,},
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
                } );
	
        my $expected = delete($args{expected});
        my $background = $args{background};
        
        if (($background == TRUE) && ($expected != C_RC_SUCCESS))
        {
                logresult('script', "Compression\:\:VolCopy\:\:start: background option should not be set " .
                                    "if expected is different than C_RC_SUCCESS");
        }
        
        my $source_vol = $class->{_src_vol_name};
        logcomment("Starting vol copy:\n" .
                   "Source: $source_vol (filer: " . $class->{_src_host_name} . "; aggregate: " . $class->{_src_aggr_name} . ")\n" .
                   "Destination: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . "; aggregate: " . $class->{_dst_aggr_name} . ")\n" .
                   "Background option: " . $args{background});
        
        eval
        {
                $class->{_copy}->start(%args);
        };
	
        my $output = $@;
        return  $class->{_ui}->process_output_string(code => C_CMD_VOL_COPY_START,
                                                     output => $output,
                                                     expected_rc => $expected,
                                                     log_message => "Start vol copy of volume $source_vol",);
}


## @name ::wait_for_completion
## @summary Wait for the background vol copy transfers to complete.
## 
## @description
## Wait for the background vol copy transfers to complete.
##
## @return Hash with rc.
##
## @examples
## $obj->wait_for_completion();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

sub wait_for_completion
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        # None
                } );
        
        
        logcomment("Waiting for vol copy to complete:\n" .
                   "Source: " . $class->{_src_vol_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                   "Destination: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . ")");
	
        $class->{_copy}->wait();
        
        return (rc => C_RC_SUCCESS);
}


## @name ::get_status
## @summary Gets the status of vol copy relationship.
## 
## @description
## Gets the status of vol copy relationship.
##
## @return Hash with rc, status, and transfer_progress. 
##
## @examples
## my %status = $obj->get_status();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub get_status
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        # None
                } );
        
        my $status = $class->{_copy}->status();
        my $copy_status = $status->{status};
        
        logcomment("Vol copy status: \"$copy_status\"\n" .
                   "Source: " . $class->{_src_vol_name} . " (filer: " . $class->{_src_host_name} . "; aggregate: " . $class->{_src_aggr_name} . ")\n" .
                   "Destination: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . "; aggregate: " . $class->{_dst_aggr_name});
        
        return (rc => C_RC_SUCCESS,
                status => $copy_status,
                transfer_progress => $status->{'transfer-progress'},);
}


## @name ::verify_status
## @summary Verifies the status of vol copy relationship.
## 
## @description
## Verifies the status of vol copy relationship.
##
## @arg status Optional: Expected status.
## Default: C_VOL_COPY_STATUS_TRANSFERRING.
##
## @return Hash with rc. 
##
## @examples
## $obj->verify_status(status => C_VOL_COPY_STATUS_IDLE);
## $obj->verify_status(); # Same as $obj->verify_status(status => C_VOL_COPY_STATUS_TRANSFERRING);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub verify_status
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        status => {type => SCALAR, optional => 1, default => C_VOL_COPY_STATUS_TRANSFERRING},
                } );
        
        my $expected_status = $args{status};
        
        logcomment("Verifying the vol copy status ($expected_status):\n" .
                   "Source: " . $class->{_src_vol_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                   "Destination: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . ")");
        
        my %current_status = $class->get_status();
        
        if ($current_status{status} ne $expected_status)
        {
                logresult('fatal', "Failed to detect expected vol copy transfer status ($expected_status):\n" . Dumper(%current_status));
        }
        
        return %current_status;
}


## @name ::wait_for_status
## @summary Wait for the vol copy relationship to reach specific state.
## 
## @description
## Wait for the vol copy relationship to reach specific state.
##
## @return Hash with rc.
##
## @arg status Required: Status to wait for (can be a regex).
##
# @arg source Optional: Flag to get the status from the source filer.
# Default: FALSE (destination filer).
#
# @arg restore Optional: Flag to get the status of the restore transfer.
# Default: FALSE. 
#
## @arg timeout Optional: Timeout, in seconds, for completion.
##
## @examples
## $obj->wait_for_status(status => C_VOL_COPY_STATUS_TRANSFERRING);
## $obj->wait_for_status(status => C_VOL_COPY_STATUS_IDLE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub wait_for_status
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        status => {type => SCALAR,},
                        # Do not expose them yet in 7-mode (not supported in C-mode)
                        #source => {type => SCALAR, optional => 1, default => FALSE,},
                        #restore => {type => SCALAR, optional => 1, default => FALSE,},
                        timeout => {type => SCALAR, optional => 1,},
                } );
        
        my $status = $args{status};
        my $timeout = $args{timeout};
        
        my $comment = "Waiting for vol copy status \"$status\":\n" .
                      "Source: " . $class->{_src_vol_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                      "Destination: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . ")";
        
        if (defined($timeout))
        {
                $comment .= "\nTimeout: $timeout";
        }
        
        logcomment($comment);
	
        $class->{_copy}->status_wait(%args);
        
        return (rc => C_RC_SUCCESS);
}


## @name ::abort
## @summary Aborts an active vol copy transfer.
## 
## @description
## Aborts an active vol copy transfer.
##
## @return Hash with rc.
##
## @examples
## $obj->abort();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub abort
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        # None
                } );
        
        logcomment("Aborting vol copy transfer:\n" .
                   "Source: " . $class->{_src_vol_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                   "Destination: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . ")");
        
        $class->{_copy}->abort();
        
        return (rc => C_RC_SUCCESS);
}


## @name ::throttle
## @summary Modifies the speed at which the vol copy transfer is being done.
## 
## @description
## Modifies the speed at which the vol copy transfer is being done.
##
## @arg value Optional: New speed value, to be set.
## The  value  ranges from 10 (full speed) to 1 (one-tenth of full speed).
## Default: C_DFLT_VOL_COPY_THROTTLE (10, full speed).
##
## @return Hash with rc. 
##
## @examples
## $obj->throttle(value => 1);
## $obj->throttle(); # To set back to default
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub throttle
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        value => {type => SCALAR, optional => 1, default => C_DFLT_VOL_COPY_THROTTLE},
                } );
        
        my $value = $args{value};
        
        my $op_num = $class->{_copy}->op_num();
        
        if (!defined($op_num))
        {
                logresult('script', "No volume copy operation was started");
        }
        
        logcomment("Modifying the speed of volume copy operation: new throttle value: $value\n" .
                   "Source: " . $class->{_src_vol_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                   "Destination: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . ")\n" .
                   "Operation number: $op_num");
        
        my $src_obj = $class->{_source}->{filer}->_get_underlying_filer_obj();
        my $src_conn = $src_obj->conn();
        
        Volume::Copy::volcopy_throttle($src_conn, $op_num, $value);
        
        return (rc => C_RC_SUCCESS);
}


## @name ::restore_destination
## @summary Restores the state of the destination volume.
## 
## @description
## Restores the state of the destination volume, after the copy transfer has completed 
## (brings it online and remounts it).
##
## @return Hash with rc. 
##
## @examples
## $obj->restore_destination();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub restore_destination
{
        my $class = shift;
        
        logcomment("Restoring the copy destination volume state:\n" .
                   "Volume: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . ")");
        
        # For a short time, the volume may be still busy with copy
        # Loop a few times, checking for the status first
        # (this is a best-effort try. If we finished)
        for (my $allowed_attempt = 1; $allowed_attempt <= 10; $allowed_attempt++)
        {
                my %dest_status = $class->{_destination}->{volume}->get_status();
                my $state = $dest_status{status}->{state};
                logcomment("VolCopy: Destination volume state: $state");
                if ($state ne C_VOL_STATE_BUSY)
                {
                        # Ok to proceed
                        last;
                }
        }
        
        logcomment("VolCopy: Bringing the destination volume online...");
        $class->{_destination}->{volume}->bring_online();
        logcomment("VolCopy: Remounting the destination volume...");
        $class->{_destination}->{volume}->nfs_umount();
        $class->{_destination}->{volume}->nfs_mount();
        
        $class->_reset_destination_dataset_info();
        
        return (rc => C_RC_SUCCESS);
}



########################
# Internal methods
########################

sub _reset_destination_dataset_info
{
        my $class = shift;
        
        my $src = $class->{_source};
        my $dst = $class->{_destination};
        
        $dst->{dataset}->attach_dataset_info(src_vol => $src,
                                             reset_info => TRUE,);
}


# Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
        
        if ($class->{_boot_mode} eq C_ONTAP_7MODE)
        {
                # TODO: we should not use the NOTE internal flag...
                # use it for now, until DPG team offers a spare method
                $class->{_copy}->{_spare_me} = 1;
        }
	
	if (($class->{_disable_cleanup} == FALSE) &&
            ($class->{_creator_pid} == $$))
        {
                logcomment("Executing Compression\:\:VolCopy cleanup for:\n" .
                           "Source: " . $class->{_src_vol_name} . "\n" .
                           "Destination: " . $class->{_dst_vol_name});
                
                eval
                {
                        # We need to do it this way because:
                        # NOTE seems to break the garbage collection basics (cleanup is done with an END block)
                        $class->{_copy}->destroy();
                };
                
                if( $@ )
                {
                        logcomment( "VolCopy: cleanup error: " . $@ );
                }
                
                undef($class->{_destination});
        }
	else
	{
                logcomment( "VolCopy: cleanup:  nothing done: script PID: " . $$) if ($class->{_debug});
	}
        
        if (defined($execution_error))
        {
                _c_interpret_logresult($execution_error);
        }
}


1;
