# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/7Mode/VolMove.pm#1 $
#
## @summary
##  Utility wrapper module for 7-mode vol move
##
## @author @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (internal)

package Compression::7Mode::VolMove;

BEGIN
{
	require "Compression/base_import.pl";

}

use strict;
use warnings;
use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Compression::Constants qw(:all);
use Dense::ndvm qw(vol_move_start vol_move_resume vol_move_abort);



# Documentation in Compression::VolMove
sub new
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		volume => {type => OBJECT,
			   isa => "NOTE::Volume",},
		
		dest_aggr => {type => SCALAR,},
                
                ui => {type => OBJECT,
		       isa => "Compression::UI",},
	});

	# Start building our object:
	my $self = {};
	$self->{_volume} = $args{volume};
	$self->{_dest_aggr} = $args{dest_aggr};
	$self->{_ui} = $args{ui};
	$self->{_debug} = $self->{_ui}->{_debug};
	$self->{_creator_pid} = $$;
	$self->{_vol_name} = $self->{_volume}->name();
	$self->{_filer_name} = $self->{_volume}->Filer->host();
	
	# Create the destination aggregate, if it does not exist yet
	# Use a temporary object for this
	logcomment("Initializing vol move: creating the destination aggregate, if not already present: " . $self->{_dest_aggr});
	my $aggr = NOTE::Aggregate->create(name => $self->{_dest_aggr},
					   Filer => $self->{_volume}->Filer,
					   recycle => 1,
					   share => 1,
					   disks => C_DFLT_AGGR_DISKS,
					   indirect_type => $main::AGGR_IND_TYPE,);
	
	
	bless($self, $class);
	
	return $self;
}


sub start
{
        my $class = shift;
        my (%args) = validate( @_,
        {
		no_cutover => {type => SCALAR, default => FALSE,}, 
		expected => {type => SCALAR, default => C_RC_SUCCESS,},
	});
        
	my $expected = delete($args{expected});
	my $vol_name = $class->{_vol_name};
	my $dest_aggr = $class->{_dest_aggr};
	
	logcomment("Starting move of volume $vol_name to aggregate $dest_aggr");
	$args{host} = $class->{_filer_name};
	$args{src_vol} = $vol_name;
	$args{dst_aggr} = $dest_aggr;
	
	eval {
		Dense::ndvm::vol_move_start(\%args);
	};
	my $output = $@;
	return $class->{_ui}->process_output_string(code => C_CMD_VOL_MOVE_START,
						    output => $output,
						    expected_rc => $expected,
						    log_message => "Start move of volume $vol_name to aggregate $dest_aggr",);
}


sub status
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                expected => {type => SCALAR, default => C_RC_SUCCESS,},
		required_state => {type => SCALAR, optional => 1,}
        });
        
	my $expected = $args{expected};
	my $vol_name = $class->{_vol_name};
	my $cmd_code = C_CMD_VOL_MOVE_STATUS;
	my $required_state = $args{required_state};
	
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
						       volume => $vol_name);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_volume}->Filer,
                                                            boot_mode => C_ONTAP_7MODE,
                                                            cmd => $command,);
	
	my %res = $class->{_ui}->process_output_string(code => $cmd_code,
						       output => $output,
						       required_return_hash_key => $vol_name,
						       expected_rc => $expected,
						       log_message => "Obtain move status for volume $vol_name",);
	
	if (defined($required_state))
	{
		my $status = $res{move_status};
		# For 7-mode, there is no cutover-deferred state -> translates to 'move'
		if (C_VOL_MOVE_STATE_CUTOVER_DEFERRED eq $required_state)
		{
			$required_state = C_VOL_MOVE_STATE_MOVE;
		}
		my $obtained_state = $status->{$vol_name}->{state};
		# For now compare directly. If output changes will be made to the status command,
		# state strings will have to be obtained through UI/Constants
		logcomment("Looking for required vol move state for volume $vol_name\n" .
			   "Required state: $required_state\n" .
			   "Obtained state: $obtained_state");
		if ($obtained_state ne $required_state)
		{
			logresult('fatal', "Failed to detect required vol move state for volume $vol_name\n" .
				           "Required state: $required_state\n" .
					   "Obtained state: $obtained_state\n" .
					   "\n\n" . Dumper(%res));
		}
	}
	
	return %res;
}


sub wait_for_completion
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                timeout => {type => SCALAR, regex => qr/^\d+$/, default => C_DFLT_TIMEOUT,},
		delay => {type => SCALAR, regex => qr/^\d+$/, default => 5,},
        });
        
	my $vol_name = $class->{_vol_name};
	my $timeout = $args{timeout};
	my $delay = $args{delay};
	
	logcomment("\nWaiting for vol move operation (volume $vol_name) to complete");
	my $rc = C_RC_SUCCESS;
	while ($rc != C_RC_VOL_IS_NOT_BEING_MOVED)
	{
		my %res = $class->status(expected => C_RC_FAILURE_NON_FATAL);
		$rc = $res{rc};
		sleep($delay);
		$timeout-=$delay;
		if ($timeout < 0)
		{
			logresult('fatal', "Timed-out while waiting for vol move to complete: volume: $vol_name");
		}
	}
	
	return (rc => C_RC_SUCCESS);
}


sub wait_for_state
{
        my $class = shift;
        my (%args) = validate( @_,
        {
		state => {type => SCALAR,},
                timeout => {type => SCALAR, regex => qr/^\d+$/, default => C_DFLT_TIMEOUT,},
		delay => {type => SCALAR, regex => qr/^\d+$/, default => 1,},
        });
        
	my $vol_name = $class->{_vol_name};
	my $timeout = $args{timeout};
	my $delay = $args{delay};
	# For 7-mode, there is no cutover-deferred state -> translates to 'move'
	my $state = (C_VOL_MOVE_STATE_CUTOVER_DEFERRED eq $args{state}) ? C_VOL_MOVE_STATE_MOVE : $args{state};
	
	logcomment("\nWaiting for vol move operation (volume $vol_name) to reach the state \"$state\"");
	while ($timeout >= 0)
	{
		# Expect success for status
		my %res = $class->status();
		my $status = $res{move_status};
		my $obtained_state = $status->{$vol_name}->{state};
		if ($obtained_state eq $state)
		{
			logcomment("\nFound required state for vol move operation (volume $vol_name): \"$state\"");
			last;
		}
		
		sleep($delay);
		$timeout-=$delay;
	}
	
	if ($timeout < 0)
	{
		logresult('fatal', "Timed-out while waiting for vol move to complete: volume: $vol_name");
	}
	
	return (rc => C_RC_SUCCESS);
}


sub initiate_cutover
{
        my $class = shift;
        my (%args) = validate( @_,
        {
		expected => {type => SCALAR, default => C_RC_SUCCESS,},
	});
        
	my $vol_name = $class->{_vol_name};
	my $expected = $args{expected};
	
	logcomment("Initiating move cutover volume $vol_name");
	
	my $cmd_code = C_CMD_VOL_MOVE_CUTOVER;
	
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
						       volume => $vol_name);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_volume}->Filer,
                                                            boot_mode => C_ONTAP_7MODE,
                                                            cmd => $command,);
	
	return $class->{_ui}->process_output_string(code => $cmd_code,
						    output => $output,
						    expected_rc => $expected,
						    log_message => "Initiate move cutover for volume $vol_name",);
}


sub pause
{
        my $class = shift;
        my (%args) = validate( @_,
        {
		expected => {type => SCALAR, default => C_RC_SUCCESS,},
	});
        
	my $vol_name = $class->{_vol_name};
	my $expected = $args{expected};
	
	logcomment("Pausing move of volume $vol_name");
	
	my $cmd_code = C_CMD_VOL_MOVE_PAUSE;
	
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
						       volume => $vol_name);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_volume}->Filer,
                                                            boot_mode => C_ONTAP_7MODE,
                                                            cmd => $command,);
	
	return $class->{_ui}->process_output_string(code => $cmd_code,
						    output => $output,
						    expected_rc => $expected,
						    log_message => "Pause move of volume $vol_name",);
}


sub is_paused
{
        my $class = shift;
        my (%args) = validate( @_,
        {
		expected => {type => SCALAR, default => C_RC_SUCCESS,},
	});
        
	my $vol_name = $class->{_vol_name};
	my $expected = $args{expected};
	
	logcomment("Checking if vol move is paused for volume $vol_name");
	
	my $cmd_code = C_CMD_VOL_MOVE_IS_PAUSED;
	
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
						       volume => $vol_name);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_volume}->Filer,
                                                            boot_mode => C_ONTAP_7MODE,
                                                            cmd => $command,);
	
	return $class->{_ui}->process_output_string(code => $cmd_code,
						    output => $output,
						    expected_rc => $expected,
						    log_message => "Check if vol move is paused for volume $vol_name",);
}


sub resume
{
        my $class = shift;
        my (%args) = validate( @_,
        {
		# None for now
	});
        
	my $vol_name = $class->{_vol_name};
	
	logcomment("Resuming vol move for volume $vol_name");
	$args{host} = $class->{_filer_name};
	$args{src_vol} = $vol_name;
	
	my $output = Dense::ndvm::vol_move_resume(\%args);
	
	return (rc => C_RC_SUCCESS, output => $output);
}


sub abort
{
        my $class = shift;
        my (%args) = validate( @_,
        {
		# None for now
	});
        
	my $vol_name = $class->{_vol_name};
	
	logcomment("Aborting vol move for volume $vol_name");
	$args{host} = $class->{_filer_name};
	$args{src_vol} = $vol_name;
	
	my $output = Dense::ndvm::vol_move_abort(\%args);
	
	return (rc => C_RC_SUCCESS, output => $output);
}



1;