#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/7Mode/Volume.pm#1 $
#

## @summary Compression Volume routines.
##
## @description
## Wrapper around NOTE::Volume.
## The main reason for not using NOTE::Volume directly is to provide standardized output
##    and to enable specific processing on some failures, otherwise fatal.
##    This is required for handling compression-specific failures.
##    
## @synopsis
##
## use Compression::Volume;
## use vars qw( $FILER $AGGREGATE $VOLUME );
## 
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $volume = Compression::Volume->new(filer => $filer_obj,
##                                       aggregate => $aggregate_obj,
##                                       volume => {name => 'flex_vol',
##                                                  size => '2g',
##                                                  recycle => FALSE,);
## OR
## my $volume = Compression::Volume->new(filer => $filer_obj,
##                                       volume => {name => 'trad_vol',
##                                                  type => C_VOL_TYPE_TRAD,
##                                                  disks => 3,
##                                                  recycle => FALSE,);
##
## @examples
## $volume->bring_offline();
## $volume->bring_online();
## $volume->set_option(service => C_OPTION_COMPRESSION, new_value => TRUE );
## $volume->verify_option( ( service => C_OPTION_COMPRESSION, expected => C_RC_FAILURE );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private


package Compression::7Mode::Volume;

use strict;
use warnings;

use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;
use GenericMount qw(umount);
use Compression::Constants qw(:all);
use Compression::Utils qw(detect_indirect_type_from_status
                          _send_cmd_to_filer
			  _c_interpret_logresult);

use wack;
use SnapCmds;
use Wreck qw(wreck);
use Volume qw(:all);
use filer qw(filer_vol_snap_reclaimspace_get);
use SayWrap qw(sayEOF);
use HybridBuftree qw(volume_is_64bit volume_is_hybrid buftree_check);
use IPU::7ModeIPU qw(vol_upgrade_diag);
use snaplock qw(set_retention_periods volume_expiry_date);
use SnapLock::Snaplock_auto_commit qw(set_vol_auto_commit);

use vars qw($WAFLIRON $WAFLIRON_IS_DONE);

BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}


## @name ::new
## @summary Compression::Volume Constructor.
##
## @description
## Creates an object representing a volume.
##
## @return Object of type Compression::Volume.
##
## @arg filer Required: NOTE::Filer object corresponding to the filer on which the volume will exist.
##
## @arg aggregate Optional: NOTE::Aggregate object corresponding to the aggregate which will contain the volume.
##
## @arg volume Optional: Volume information.
## It can be either a scalar (representing the name), or a hash reference.
## If not provided, uses the default values defined in Compression::Constants.
## If volume is passed in as a hashref, then the allowed keys/values are:
## - name => name of the volume,
## - type => volume type (flex/trad). Default: C_DFLT_VOL_TYPE,
## - size => volume size. Default: C_DFLT_VOL_SIZE,
## - recycle => flag to decide if the volume should be recycled,
## - tag => volume tag,
## - exportfs => Exportfs options for this volume. Default: C_DFLT_VOL_EXPORTFS,
## - disks => disk to use when creating a trad vol,
## - compress_enabled => Flag to decide if the volume should be compression-enabled. Default: C_ENABLE_COMPRESSION,
## - auto_mount => Flag to decide if the volume should be mounted on the localhost.
##
## @arg ui Optional: Compression::UI object corresponding to the hosting filer.
## If not provided, it will be auto-generated.
##
## @arg disable_cleanup Optional: Flag to disable any clean-up that is executed
## upon object destruction. Defaults to FALSE.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
##
## my $aggregate = NOTE::Aggregate->create(Filer => { host => $FILER },
##                                         name => C_DFLT_AGGREGATE,
##                                         recycle => TRUE,
##                                         share => TRUE,
##                                         disks => 3,
##                                         );
## 
## my $obj = Compression::Volume->new(aggregate => $aggregate,
##                                    volume => {name => 'test_vol',
##                                               size => "2g",
##                                               recycle => TRUE,);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub new
{
	my $class = shift;
	my (%args) = validate( @_,
		{
                        filer => {type => OBJECT,
				  isa  => "NOTE::Filer",
				  },
                        
                        aggregate => {type => OBJECT,
				      isa  => "NOTE::Aggregate",
				      optional => 1,
				      },
                        
                        volume 	=> {type => SCALAR|HASHREF|UNDEF,
				    optional => 1,
				    },
                        
                        vfiler => {type => OBJECT|UNDEF,
                                   optional => 1,
                                  },
                        
                        ui      => {type => OBJECT,
                                    isa  => "Compression::UI",
                                    optional => 1,
                                    },
                        
                        client  => {type => SCALAR,
                                    optional => 1,
                                    default => C_CLIENT_NFS_LOCAL,
                                    },
                        
                        disable_cleanup => {type     => BOOLEAN,
					    optional => 1,
					    default  => FALSE,
					    },
                        
                        debug 	=> {type => SCALAR,
				    optional => 1,
				    default => FALSE,
				    },
			
                        # Internal flag. To not be used outside the compression library
			attach => {type => OBJECT|SCALAR|UNDEF,
				   optional => 1,
				   },
			
			# Internal flag. To not be used outside the compression library
			_fake_mount => {type => SCALAR,
					default => FALSE,},
			
			# Internal flag. To not be used outside the compression library
			_added_to_vfiler => {type => SCALAR,
					     default => FALSE,},
		} );

	my $self = {};
        $self->{_debug} = $args{debug};
        $self->{_disable_cleanup} = $args{disable_cleanup};
        $self->{_creator_pid} = $$;
        
        logcomment("Volume: PID of the script creating this object: " . $self->{_creator_pid} .
                   ". New volume: " . Dumper($args{volume})) if ($self->{_debug});
        
	# Construct the vol hash as required by the NOTE::Volume constructor 
	my %vol = ();
        
        my $auto_mount_flag = C_DFLT_VOL_MOUNT;
        my $volume_args = $args{volume};
        if ((defined($volume_args) && (ref($volume_args) eq "HASH")))
        {
                # Second-layer hash validation for volume args
                logcomment("Volume: new: volume attributes passed in as HASHREF") if ($self->{_debug});
                %vol = validate_with
                        ( params => $volume_args,
                          spec   => {name => { type => SCALAR|UNDEF, default => C_DFLT_VOLUME },
                                     type => { type => SCALAR, default => C_DFLT_VOL_TYPE },
                                     size => { type => SCALAR, default => C_DFLT_VOL_SIZE },
                                     recycle => { type => SCALAR, default => $main::VOL_RECYCLE },
                                     tag => { type => SCALAR, optional => 1 },
                                     exportfs => { type => SCALAR, default => C_DFLT_VOL_EXPORTFS },
                                     disks => { type => SCALAR, optional => 1 },
                                     auto_mount => { type => BOOLEAN, default => C_DFLT_VOL_MOUNT },
                                     },
                          allow_extra => 0,
                        );
                
                if (!defined($vol{tag}))
                {
                        $vol{tag} = $vol{name};
                }
                $auto_mount_flag = $vol{auto_mount};
                delete($vol{auto_mount});
        }
        else
        {
                logcomment("Volume: new: volume attribute passed in as SCALAR or undefined") if ($self->{_debug});
                $vol{name} = (defined($volume_args)) ? $volume_args : C_DFLT_VOLUME;
                $vol{type} = C_DFLT_VOL_TYPE;
                $vol{size} = C_DFLT_VOL_SIZE;
                $vol{recycle} = $main::VOL_RECYCLE;
                $vol{tag} = $vol{name};
                $vol{exportfs} = C_DFLT_VOL_EXPORTFS;
        }
        
        $self->{_recycled} = $vol{recycle};
        $self->{_type} = $vol{type};
        
        # UI
        if ( defined($args{ui}) )
        {
                $self->{_ui} = $args{ui};
        }
        else
        {
                # Auto-detect version of the hosting filer. Need this for UI.
                logcomment("Volume: UI object not provided. Auto-generating...") if ($self->{_debug});
                my $version = Compression::Version->new( conn => $args{filer}, debug => $self->{_debug} );
                $self->{_ui} = Compression::UI->new( version => $version, debug => $args{debug}, boot_mode => C_ONTAP_7MODE );
        }
        
        if ($self->{_type} eq C_VOL_TYPE_TRAD)
        {
                logcomment("Volume: traditional volume requested: " . $self->{vol_name}) if ($self->{_debug});
                # NOTE::Volume does not like the size passed in for a trad volume
                delete($vol{size});
        }
        
        logcomment("Volume: new: volume information hash to be passed to NOTE constructor:\n" . Dumper(%vol)) if ($self->{_debug});

        my $conn = $args{filer};
        $vol{Filer} = $conn;
        if (defined($args{aggregate}))
        {
                logcomment("Volume: new: aggregate object was passed in") if ($self->{_debug});
                my $aggr = $args{aggregate};
                $vol{Aggregate} = $aggr;
        }
        # Done constructing the vol hash
	
	
	if (defined($args{attach}))
	{
                my $attach = $args{attach};
		if (ref($attach) ne '')
                {
                        if (!($attach->isa("NOTE::Volume")))
                        {
                                logresult('script', "If not a scalar, the attach argument must be a NOTE\:\:Volume object");
                        }
                        $self->{_vol} = $args{attach};
                }
                else
                {
                        # Scalar, representing the name of the volume to which to attach to
                        # This will first be renamed to $vol{name} (if this is defined)
                        my $vol_name = $vol{name};
                        
                        if ((defined($vol_name)) && ("$vol_name" ne "$attach"))
                        {
                                # First check if the volume $vol_name exists and destroy it
                                # Will happen if aggr copy is called twice in a row
                                my $renamed_vol = NOTE::Volume->attach(name => $vol_name,
                                                                       Filer => $conn,
                                                                       Aggregate => $args{aggregate},
                                                                       force => 1);
                                if ($renamed_vol->exists())
                                {
                                        $renamed_vol->purge();
                                }
                                
                                # Use Aggregate here without checking.
                                # This code path should only be used for flex vols currently...
                                my $temp_vol = NOTE::Volume->attach(name => $attach,
                                                                    Filer => $conn,
                                                                    Aggregate => $args{aggregate});
                                
                                $temp_vol->rename($vol_name);
                        }
                        $self->{_vol} = NOTE::Volume->attach(%vol);
                }
	}
	else
	{
		# First, see if volume already exists.
		# Use a temp NOTE volume for this. Forced attach to named volume only
		if (defined($vol{name}))
		{
			my $volume_name = $vol{name};
			$vol{force} = 1;
			my $existent_volume = NOTE::Volume->attach(%vol);
			delete $vol{force};
			
			if ($existent_volume->exists())
			{
				if ($vol{recycle} == FALSE)
				{
					# Volume should be recycled. Destroy it.
					logcomment("Volume recycling not requested. Destroying existent volume $volume_name") if ($self->{_debug});
					$existent_volume->purge( );
				}
				else
				{
					# Check status of existent volume. If offline, attempt to bring online.
					my $vol_status = $existent_volume->status();
					my $vol_state = $vol_status->{state};
					logcomment("Volume $volume_name will be reused. State: $vol_state") if ($self->{_debug});
					if ($vol_state eq C_VOL_STATE_OFFLINE)
					{
						logcomment("Volume $volume_name was offline. Bringing online...");
						$existent_volume->online();
					}
					
				}
			}
			else
			{
				logcomment("Volume $volume_name does not exist yet.") if ($self->{_debug});
				
			}
		}
        
		# Now create our vol object
		eval { $self->{_vol} = NOTE::Volume->create(%vol); };
		if ($@)
		{
			# TODO: apparently NOTE should be attempting to add disks too,
			# but has the error message not up-to-date?
			# Once this is settled, we should let NOTE add the disk for us...
			if (($@ =~ /there is not enough space in the given aggregate/) &&
			    (defined($args{aggregate})))
			{
				logcomment("There is not enough space to create the volume\n" .
					   "Attempting to add one disk to the containing aggregate, to allow volume creation");
				my $aggr = $args{aggregate};
				$aggr->add();
				logcomment("Volume: retrying volume creation") if ($self->{_debug});
				$self->{_vol} = NOTE::Volume->create(%vol);
			}
			else
			{
				logresult("config", "CONFIG ERROR: " . $@);
			}
		}
		
		# Double-check volume creation
		if (!defined($self->{_vol}))
		{
			logresult("config", "CONFIG ERROR: Error creating the volume");
		}
	}
	
        $self->{vol_name} = $self->{_vol}->name();
        $self->{_anonymous} = $self->{_vol}->is_anonymous();
        # Spare the volume, so NOTE does not destroy it and we get to control if we clean-up.
        $self->{_vol}->SPARE();
	
	# Set this to localhost by default.
	# This is because NOTE::Volume automatically mounts it: we need to umount as part of clean-up
	$self->{_was_mounted_at_setup} = $args{client};
        
        
	###################################
        # Vfiler
        ###################################
        $self->{_vfiler} = $args{vfiler};
        if (defined($self->{_vfiler}))
        {
		if (!($self->{_vfiler}->isa("Compression::7Mode::Vfiler")))
                {
                        logresult('script', "The vfiler argument is required to be a Compression\:\:7Mode\:\:Vfiler object");
                }
                
                $self->{__vfiler} = $self->{_vfiler}->_get_underlying_obj();
		if ($args{_added_to_vfiler} == FALSE)
		{
			logcomment("Adding the Compression volume as a resource to the vfiler");
			my $add_status = $self->{__vfiler}->add(vol => $self->{_vol});
			if ($add_status != 1)
			{
				logresult('fatal', "Failed to add the Compression volume as a resource to the vfiler");
			}
		}
                $self->{__vfiler}->exportfs(container => $self->{_vol});
	}

        bless $self, $class;
        
        ############################################
	# Do extra setup steps: mount
	############################################
        if ($auto_mount_flag == TRUE)
        {
		if ($args{_fake_mount} == FALSE)
		{
			$self->nfs_mount(client => $args{client});
		}
                $self->{_was_mounted_at_setup} = $args{client};
        }

	return $self;
}


## @name ::bring_offline
## @summary Sets volume offline.
## 
## @description
## Sets volume offline. Wrapper around NOTE::Volume::offline.
##
## @return Hash with rc and output.
## 
## @examples
##
## $obj->bring_offline();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub bring_offline
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # No args
                } );
        
        $class->{_vol}->offline();
        
        # If we got here, NOTE did not encounter any error. Return success.
        return (rc => C_RC_SUCCESS, output => "");
}


## @name ::bring_online
## @summary Sets volume online.
## 
## @description
## Sets volume online. Wrapper around NOTE::Volume::online.
##
## @arg expected Optional: expected result ( default is C_RC_SUCCESS ). 
##     Use C_RC_INSUFFICIENT_SPACE if there is not enough space to bring volume online.
##
## @return Hash with rc and output.
##  
## @examples
##
## $obj->bring_online();
## $obj->bring_online(expected=>C_RC_INSUFFICIENT_SPACE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub bring_online
{
        my $class = shift;
	my (%args) = validate( @_,
		{
			expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
			read_only => {type => SCALAR, default => FALSE,},
		} );
        
	my $vol_name = $class->{vol_name};
	my $expected = $args{'expected'};
	my $read_only = $args{read_only};
	my $output;
	my $cmd_code;
	
	if (TRUE == $read_only)
	{
		$cmd_code = C_CMD_VOL_ONLINE_RO;
		my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
							       volume => $vol_name,);
		$output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_vol}->Filer,
								 boot_mode => C_ONTAP_7MODE,
								 cmd => $command,);
	}
	else
	{
		$cmd_code = C_CMD_VOL_ONLINE;
		eval { $class->{_vol}->online(); };
		$output = $@;
	}

	return $class->{_ui}->process_output_string(code => $cmd_code,
                                                    output => $output,
                                                    expected_rc => $expected,
                                                    log_message => "Bring volume $vol_name online",);
}


## @name ::mark_restricted
## @summary Marks volume as restricted.
## 
## @description
## Marks volume as restricted. Wrapper around NOTE::Volume::restrict.
##
## @return Hash with rc and output.
##  
## @examples
##
## $obj->mark_restricted();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub mark_restricted
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # No args
                } );
        
        $class->{_vol}->restrict();
        
        # If we got here, NOTE did not encounter any error. Return success.
        return (rc => C_RC_SUCCESS, output => "");
}


## @name ::verify_option
## @summary Verifies that the desired volume option is set as expected.
## 
## @description
## Verifies that the desired volume option is set as expected. Logs fatal result if not.
##
## @return Hash with rc and output.
## 
## @arg service Required: feature that corresponds to desired option (defined in Compression::Constants).
##
## @arg expected Optional: expected return code ( C_RC_SUCCESS/C_RC_FAILURE ).
## Defaults to C_RC_SUCCESS.
##
## @examples
##
## $obj->verify_option( service => C_OPTION_GUARANTEE, expected => C_RC_SUCCESS );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub verify_option
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        service  => { type => SCALAR, },
			expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
                } );
        
        my $service = $args{service};
        my $expected = $args{expected};
        my $rc = C_RC_SUCCESS;
        my $val_to_set;
        
        my $option_def = $class->{_ui}->get_option_definition( option_code => $args{service} );
        my $option_name = $option_def->{name};
        logcomment("Obtained option definition for service $service: " . Dumper($option_def)) if ($class->{_debug});
        
        my $current_setting = $class->{_vol}->options($option_name);
        logcomment("Volume: $option_name is set to $current_setting") if ($class->{_debug});
        
        if ($current_setting eq $option_def->{disabled})
        {
                $rc = C_RC_FAILURE;
        }
        elsif ($current_setting ne $option_def->{enabled})
        {
                logresult("fatal", "Unexpected \"$option_name\" option for volume " . $class->{vol_name} . ": $current_setting");
        }
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Get option setting for $option_name",
                                                    output      => $current_setting);
}


## @name ::set_option
## @summary Sets a new value for the desired option for the volume.
## 
## @description
## Sets a new value for the desired option for the volume.
##
## @return Hash with rc and output.
## 
## @arg service Required: feature that corresponds to desired option (defined in Compression::Constants).
## 
## @arg new_value Required: tag that corresponds to new value to be set.
## For now, only TRUE/FALSE are handled, corresponding to enable/disable.
## In the case of C_OPTION_GUARANTEE, new_value can also be "file".
##
## @arg volume_name Optional: Name of volume for which to disable the option.
## If not provided, the desired option will be set for the volume for which the object was created.
##
## @arg check_if_already_set Optional: Flag to enable the library to first check the current setting.
## If the current setting matches the value to be set, the subroutine will not attempt to set it again.
## Instead, it will return success. The default value for check_if_already_set is TRUE.
##
## @arg expected Optional: expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @arg use_new_value_as_is Optional: Flag to enable the library to use the new_value string directly,
## even if a definition is already given in Options.pm. Default: FALSE.
## 
## @examples
## $obj->set_option( service => C_OPTION_GUARANTEE, new_value => TRUE, expected => 'file' );
## $obj->set_option( service => C_OPTION_GUARANTEE, new_value => TRUE, expected => FALSE );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub set_option
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        volume_name => { type => SCALAR|UNDEF, optional => 1 },
                        service => { type => SCALAR, },
			new_value => { type => SCALAR, },
                        check_if_already_set => { type => BOOLEAN, optional => 1, default => C_DFLT_CHECK_OPTION_FIRST },
                        expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
                        use_new_value_as_is => { type => SCALAR, optional => 1, default => FALSE },
                } );
        
        my $service = $args{service};
        my $new_value = $args{new_value};
        my $check_value_flag = $args{check_if_already_set};
        my $expected = $args{expected};
        my $rc = C_RC_FAILURE;
        my $val_to_set = $args{new_value}; # default
        my $volume_obj = $class->{_vol};   # default
	
	if (defined($args{volume_name}))
        {
                $volume_obj = NOTE::Volume->attach( name => $args{volume_name}, Filer => $class->{_vol}->Filer, force => 1 );
                if (!($volume_obj->exists()))
                {
                        return $class->{_ui}->check_for_expected_rc(obtained_rc => C_RC_VOLUME_NOT_FOUND,
                                                                    expected_rc => $args{expected},
                                                                    log_message => "Set option for service $service to $new_value",
                                                                    output      => '');
                }
        }
        
        # Get definition hash as defined in Options.pm
        my $option_def = $class->{_ui}->get_option_definition( option_code => $service );
        logcomment("Option definition for $service: " . Dumper($option_def)) if ($class->{_debug});
        if ($args{use_new_value_as_is} == FALSE)
        {
		my $val_true = TRUE;
		my $val_false = FALSE;
                if ("$new_value" eq "$val_true")
		{
			$val_to_set = $option_def->{enabled};
                }
		elsif ("$new_value" eq "$val_false")
		{
			$val_to_set = $option_def->{disabled};
                }
        }
        my $option_name = $option_def->{name};
        
        logcomment("Volume: Will set option $option_name to $val_to_set") if ($class->{_debug});
        
        if ($check_value_flag == TRUE)
        {
                my $current_setting = $volume_obj->options($option_name);
                logcomment("Volume: Current setting for option $option_name is: $current_setting") if ($class->{_debug});
                
                if ($current_setting eq $val_to_set)
                {
                        logcomment("Volume: Option $option_name already set to required value: $val_to_set") if ($class->{_debug});
                        return $class->{_ui}->check_for_expected_rc(obtained_rc => C_RC_SUCCESS,
                                                                    expected_rc => $args{expected},
                                                                    log_message => "Set option $option_name to $val_to_set",
                                                                    output      => $current_setting);
                }
        }

        my $new_setting;
        eval
	{
		$new_setting = $volume_obj->options( $option_name => $val_to_set );
	};
	
	if( $@ )
	{
                return $class->{_ui}->process_output_string(code => 'OPTION_SET',
                                                            output => $@,
                                                            expected_rc => $expected,
                                                            log_message => "Set option $option_name to $val_to_set",
                                                            );
	}
        elsif ($new_setting->{$option_name} eq $val_to_set)
        {
                $rc = C_RC_SUCCESS;
        }
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Set option $option_name to $val_to_set",
                                                    output      => $new_setting);
}


## @name ::get_mount_path
## @summary Returns the full volume name.
## 
## @description
## Returns the full volume name (ex: /vol/myvol). Wrapper around NOTE::mount_path.
##
## @return Full volume name.
##  
## @examples
## my $vol_mount_path = $obj->get_mount_path();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub get_mount_path
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # No args
                } );
        
        my $mount_path = $class->{_vol}->mount_path();
        
        return $mount_path;
}


## @name ::nfs_mount
## @summary NFS-mounts the volume on the specified client.
## 
## @description
## NFS-mounts the volume on the specified client.
##
## @return Hash with rc and output (resulted mountpoint).
## 
## @arg client Optional: Hostname of the client to mount the volume on.
## Defaults to 'localhost'.
## 
## @arg chmod Optional: Permissions to set on the mountpoint. 
## Defaults to '777'.
##  
## @examples
## $obj->nfs_mount( client => $my_client );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub nfs_mount
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        client => { type => SCALAR, optional => 1, default => C_CLIENT_NFS_LOCAL },
			chmod => { type => SCALAR|UNDEF, optional => 1, default => '777' },
                } );
        
        my $client = $args{client};
        my $volume = $class->{vol_name};
	my $chmod = $args{chmod};
        my %return_hash = ();
        
        # NOTE (the underlaying routine) is using 'exportfs -i -o ..'
        # For exportfs options to be persistent after a reboot,
        # this should actually be 'exportfs -p ...'. 
        # Call the exportfs wrapper here just to ensure that the exportfs options
        # are resynced to what they should be. Otherwise, mount will fail
        # if called after a reboot (chmod fails)
        $class->nfs_export_options();
        
        # Mount the volume
        logcomment("Volume: mounting volume \"$volume\" on  client \"$client\"") if ($class->{_debug});
        ## TODO: still need this logic after burts 441238/442667 are fixed?
        if (defined($class->{__vfiler}))
        {
                $return_hash{output} = $class->{__vfiler}->mount(container => $class->{_vol}, chmod => $chmod);
        }
        else
        {
		my %mount_args = (client => $client);
		if (defined($chmod))
		{
			$mount_args{chmod} = $chmod;
		}
                $return_hash{output} = $class->{_vol}->mount(%mount_args);
        }
        
	$class->{_mountpoint} = $return_hash{output};
        $return_hash{rc} = C_RC_SUCCESS;
        logcomment("Volume: nfs_mount: returning:\n" . Dumper(%return_hash)) if $class->{_debug};
        
        return %return_hash;
}

## @name ::nfs_umount
## @summary NFS-umounts the volume from the specified client.
## 
## @description
## NFS-umounts the volume from the specified client.
##
## @return Hash with rc and output (return value of NOTE::Volume::umount).
## 
## @arg client Optional: Hostname of the client to umount the volume from.
## Defaults to 'localhost'.
##  
## @examples
## $obj->nfs_umount( client => $my_client );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub nfs_umount
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        client => { type => SCALAR, optional => 1, default => C_CLIENT_NFS_LOCAL },
                } );
        
        my $client = $args{client};
        my $volume = $class->{vol_name};
        my %return_hash = ();
        
        # Umount the volume
        logcomment("7Mode Volume: umounting volume \"$volume\" from client \"$client\"") if $class->{_debug};
	if (defined($class->{__vfiler}))
        {
		GenericMount::umount($class->{_mountpoint}, client => $client);
	}
	else
	{
		$return_hash{output} = $class->{_vol}->umount(client => $client);
	}
        
        $return_hash{rc} = C_RC_SUCCESS;
        logcomment("7Mode Volume: nfs_umount: returning:\n" . Dumper(%return_hash)) if $class->{_debug};
        
        return %return_hash;
}


## @name ::get_available_space
## @summary Gets the available space (in KB) on this volume.
## 
## @description
## Gets the available space (in KB) on this volume.
##
## @return Hash with rc and output (available space in KB).
##  
## @examples
## $obj->get_available_space();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub get_available_space
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # No args
                } );
        
        # Note: NOTE::Volume::avail returns the value in KB
        # We may want to consider adding 'modifier' support
        my %return_hash = ();
        $return_hash{output} = $class->{_vol}->avail();
        $return_hash{rc} = C_RC_SUCCESS;
        
        logcomment("Volume: get_available_space. Returning: " . Dumper(%return_hash)) if $class->{_debug};
        return %return_hash;
}


## @name ::get_space_usage 
## @summary Gets disk usage of this volume.
## 
## @description
## Gets disk usage of this volume.
##
## @arg snapshot Optional: Flag to use if the snapshot statistics are needed.
## Default: FALSE.
##
## @return Hash with rc and output (available space in KB).
##  
## @examples
## $obj->get_space_usage(snapshot => TRUE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub get_space_usage
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        snapshot => { type => BOOLEAN, optional => 1, default => FALSE},
                } );
        
        my ($total ,$used, $avail, $capacity) = $class->{_vol}->disk_free(snapshot => $args{snapshot});
        
        return (rc => C_RC_SUCCESS,
                total => $total,
                used => $used,
                available => $avail,
                capacity => $capacity);
}


## @name ::get_status
## @summary Gets the status of the volume.
## 
## @description
## Gets the status of the volume. Wrapper around NOTE::Volume::status.
##
## @return Hash with rc and status, where status is the hash reference returned by NOTE::Volume::status.
##  
## @examples
## $obj->get_status();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub get_status
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # No args
                } );
        
        my %return_hash = ();
        $return_hash{status} = $class->{_vol}->status();
        $return_hash{rc} = C_RC_SUCCESS;
	$return_hash{container_aggr} = $class->{_vol}->status->parent;
        
        logcomment("Volume: get_status. Returning: " . Dumper(%return_hash)) if $class->{_debug};
        return %return_hash;
}


## @name ::create_snapshot
## @summary Creates a snapshot on the volume.
## 
## @description
## Creates a snapshot on the volume. Wrapper around NOTE::Volume::snapshot_create.
##
## @return Hash with rc and output, where output is the name of the newly-created snapshot.
##
## @arg name Optional: Name for the snapshot. If not provided, an anonymous name is used.
##  
## @examples
## $obj->create_snapshot(name => $my_snapshot);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub create_snapshot
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        name => { type => SCALAR, optional => 1, },
                } );
        
        my $snapshot = $class->{_vol}->snapshot_create($args{name});
        
        # If we got here, NOTE did not encounter any error. Return success.
        return (rc => C_RC_SUCCESS, output => $snapshot);
}


## @name ::delete_snapshot
## @summary Deletes a snapshot.
## 
## @description
## Deletes a named snapshot. Wrapper around NOTE::Volume::snapshot_delete.
##
## @return Hash with rc and output (empty).
##
## @arg name Required: Name of the snapshot.
##  
## @examples
## $obj->delete_snapshot(name => $my_snapshot);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub delete_snapshot
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        name => { type => SCALAR, },
                } );
        
        $class->{_vol}->snapshot_delete($args{name});
        
        # If we got here, NOTE did not encounter any error. Return success.
        return (rc => C_RC_SUCCESS, output => '');
}


## @name ::list_snapshots
## @summary Lists snapshots that exist on this volume.
## 
## @description
## Lists snapshots that exist on this volume. Wrapper around NOTE::Volume::snapshot_list.
##
## @return Hash with rc and output (reference to array with names of existent snapshots).
##  
## @examples
## $obj->list_snapshots();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub list_snapshots
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # none
                } );
        
        my @snapshots = $class->{_vol}->snapshot_list();
        
        # If we got here, NOTE did not encounter any error. Return success.
        return (rc => C_RC_SUCCESS, output => \@snapshots);
}


## @name ::verify_snapshot_exists
## @summary Verifies that a named snapshot exists on this volume.
## 
## @description
## Verifies that a named snapshot exists on this volume. Wrapper around NOTE::Volume::snapshot_exists.
##
## @return Hash with rc and output.
##
## @arg name Required: Name of the snapshot.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##  
## @examples
## $obj->verify_snapshot_exists(name => $existent_snapshot);
## $obj->verify_snapshot_exists(name => $inexistent_snapshot, expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub verify_snapshot_exists
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        name => { type => SCALAR, },
                        expected => { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },
                } );
        
        my $snapshot_name = $args{name};
        my $snapshot_exists = $class->{_vol}->snapshot_exists($snapshot_name);
        
        my $rc = ($snapshot_exists) ? C_RC_SUCCESS : C_RC_FAILURE;
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Verify snapshot $snapshot_name exists",
                                                    output      => $snapshot_exists);
}


## @name ::snapshot_schedule
## @summary Gets/sets the snapshot schedule for this volume.
## 
## @description
## Gets/sets the snapshot schedule for this volume. Wrapper around NOTE::Volume::snapshot_sched.
##
## @return Hash with rc and output (array reference with current schedule - only for getting the current setting).
##
## @arg disable Optional: Boolean flag indicating that the snapshot scheduler should be disabled for this volume.
## Only considered if set to TRUE.
##
## @arg schedule Optional: New snapshot schedule for this volume.
## If not specified, the current schedule is returned.
##
## @note
## If disable is set to TRUE, then the schedule argument is ignored.
##  
## @examples
## $obj->snapshot_schedule(disable => TRUE);
## $obj->snapshot_schedule(schedule => '1 1 0');
## my %current_setting = $obj->snapshot_schedule();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub snapshot_schedule
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        disable => { type => BOOLEAN, optional => 1, },
                        schedule => { type => SCALAR, optional => 1, },
                } );
        my $sched;
        
        logcomment("Volume: snapshot_schedule. Input:\n" . Dumper(%args)) if $class->{_debug};
        
        
        if (defined($args{disable}))
        {
                $sched = '0 0 0';
        }
        elsif (defined($args{schedule}))
        {
              $sched = $args{schedule};
        }
        
        my @snapshot_schedule = $class->{_vol}->snapshot_sched( sched => $sched );
        logcomment("Volume: snapshot_schedule. Current schedule:\n" . Dumper(@snapshot_schedule)) if $class->{_debug};
        
        return (rc => C_RC_SUCCESS, output => \@snapshot_schedule);
}


## @name ::restore_snapshot
## @summary Restores a snapshot of this volume.
## 
## @description
## Restores a snapshot of this volume.
##
## @arg snapshot_name Required: Snapshot name.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc and output.
##  
## @examples
## $obj->restore_snapshot(snapshot_name => 'my_snapshot');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub restore_snapshot
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        snapshot_name => { type => SCALAR, },
			expected => { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },
                } );
        
	eval
	{
		$class->{_vol}->snapshot_restore($args{snapshot_name});
	};
	
	return $class->{_ui}->process_output_string(code => C_CMD_VOL_SNAP_RESTORE,
						    output => $@,
						    expected_rc => $args{expected},
						    log_message => "Snapshot restore of volume " . $class->{vol_name},);
}


## @name ::get_snap_delta 
## @summary Gets snap delta between two given snapshots.
## 
## @description
## Gets snap delta between two given snapshots. Wrapper around NOTE::Volume::snap_delta.
##
## @return Hash with rc and snap_delta (reference to array of hashes).
##  
## @examples
## $obj->get_snap_delta(snapshot_start => 'my_snapshot1',
##                      snapshot_end => 'my_snapshot2');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub get_snap_delta
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        snapshot_start => { type => SCALAR, },
                        snapshot_end => { type => SCALAR, },
                } );
        
        my $raw_output = $class->{_vol}->snap_delta(beginning_snapshot => $args{snapshot_start},
                                                    ending_snapshot => $args{snapshot_end});
        
        logcomment("(NOTE) snap_delta returned:\n" . Dumper($raw_output)) if $class->{_debug};
        
        # NOTE return hash has keys taken directly from the output header.
        # Encapsulate here for now. If release-dependent processing is needed,
        # this should be moved in UI
        my $output = [];
        foreach my $index_data (@$raw_output)
        {
                my $tmp_hr = {};
                $tmp_hr->{kb_changed} = $index_data->{'KB changed'};
                $tmp_hr->{time} = $index_data->{'Time'};
                $tmp_hr->{rate} = $index_data->{'Rate (KB/hour)'};
                $tmp_hr->{from_snapshot} = $index_data->{'From Snapshot'};
                $tmp_hr->{to_snapshot} = $index_data->{'To'};
                
                # Check kb_changed, to catch any errors (including UI changes)
                if (!defined($tmp_hr->{kb_changed}))
                {
                        logresult('fatal', "Error processing snap delta. Undefined value for \"KB changed\". Obtained:\n" . Dumper($raw_output));
                }
                
                push(@$output, $tmp_hr);
        }

        return (rc => C_RC_SUCCESS, snap_delta => $output);
}


## @name ::get_snap_reclaimable 
## @summary Returns the amount of space reclaimed if a snapshot is deleted.
## 
## @description
## Returns the amount of space reclaimed if a snapshot is deleted (in bytes).
##
## @return Hash with rc and reclaimable.
##  
## @examples
## my %res = $obj->get_snap_reclaimable(snapshot_name => 'my_snapshot');
## my $value = $res{reclaimable};
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub get_snap_reclaimable
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        snapshot_name => { type => SCALAR, },
                } );
        
        my $output = filer_vol_snap_reclaimspace_get($class->{_vol}->Filer->conn, $class->{vol_name}, $args{snapshot_name});
        
        # If we got here, NOTE did not encounter any error. Return success.
        return (rc => C_RC_SUCCESS, reclaimable => $output);
}


## @name ::start_wafliron
## @summary Starts wafliron on the aggregate that contains this volume.
## 
## @description
## Starts wafliron on the aggregate that contains this volume Wrapper around NOTE::Volume::wafliron_start.
##
## @return Hash with rc and output.
# If the 'oc' argument is set to prev_cp, the return will also contain the reference to the cp hash provided by wafliron_start.
##
## @arg wait Optional: Flag to decide if start_wafliron should exit immediately or wait for the wafliron to complete.
## Defaults to TRUE.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Ignored if wait flag is 0. Legal values: C_RC_SUCCESS, C_RC_WAFLIRON_CORRUPTION_FOUND.
## Default value C_RC_SUCCESS (which means that wafliron found the filesystem clean).
##
## @arg force Optional: (command line option -f).
##
## @arg triage Optional: (command line option -t).
##
## @arg scan_speed Optional: Wafl scan speed.
##
## @arg inject_file Optional: (command line option -i <inject-file>).
##
## @arg timeout Optional: Wafliron timeout value.
##
## @arg volume_order Optional: Volume ordering string (command line option -v).
## Available starting IC.0.
##
# @arg oc Optional: Optional commit flag. Defaults to 0.
# Legal values: 0 (regular wafliron is run), 1 (wafliron is run in Optional Commit mode),
# prev_cp (wafliron is run in Optional Commit prev_cp mode).
# Available starting BR.0 and IC.1.
#
# @arg response_list Optional: Hash with key = vol/aggr name, value = reponse to be sent to the vol/aggr during prev_cp.
#
## @notes
## This is a wrapper around Wafliron::wafliron_start. More detailed argument documentation is available there.
## 
## @examples
## $obj->start_wafliron();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub start_wafliron
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # Arguments accepted by Wafliron::wafliron_start (NOTE will pass them down directly)
                        wait => { type => BOOLEAN, optional => 1, default => TRUE, },
                        expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS, },
                        force => { type => BOOLEAN, optional => 1, },
                        triage => { type => BOOLEAN, optional => 1, },
                        scan_speed => { type => SCALAR, optional => 1, },
                        inject_file => { type => SCALAR, optional => 1, },
                        timeout => { type => SCALAR, optional => 1, },
                        volume_order => { type => SCALAR, optional => 1, },
                        compression => { type => SCALAR, optional => 1, default => 0, },
                        #oc => { type => SCALAR, optional => 1, },
                        #response_list => { type => HASHREF, optional => 1, },
                } );
        
        # TODO: add/test/document oc related arguments once test code is available
        logcomment("Volume: start_wafliron called with following arguments:\n" . Dumper(%args)) if $class->{_debug};
        my $wait_flag = $args{wait};
        my $expected_rc = $args{expected};
        delete($args{expected});
        # TODO: wafliron -c has been deprecated.
        # For now, simply ignore the argument, to not break any script,
        # but the arg should be ultimatelly removed
        delete($args{compression});
        
        # Note: NOTE::Volume::wafliron_start will take care of the trad/flex handling
        my ($result, $message, $cp_hashref) = $class->{_vol}->wafliron_start( %args );
        
        if ($wait_flag == TRUE)
        {
                return $class->_process_wafliron_wait_result(corruption_found => $result,
                                                             message => $message,
                                                             cp_hashref => $cp_hashref,
                                                             expected => $expected_rc);
        }
        
        # We are not supposed to wait: $result is the wafliron id
        return ( rc => C_RC_SUCCESS, wafliron_id => $result, );
}


## @name ::wait_for_wafliron
## @summary Waits for wafliron to complete.
## 
## @description
## Waits for wafliron to complete. Wrapper around NOTE::Volume::wafliron_wait.
##
## @return Hash with rc and output.
## 
## @examples
## $obj->wait_for_wafliron();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub wait_for_wafliron
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        # No args
                } );
        
        logcomment("Volume: Waiting for wafliron to complete") if $class->{_debug};
        
        my ($corruption_found, $message, $cp_hashref) = $class->{_vol}->wafliron_wait();
        
        return $class->_process_wafliron_wait_result(corruption_found => $corruption_found,
                                                     message => $message,
                                                     cp_hashref => $cp_hashref,);
}


## @name ::get_wafliron_status
## @summary Gets the status of the wafliron process.
## 
## @description
## Gets the status of the wafliron process. Wrapper around NOTE::Volume::wafliron_status.
##
## @return Hash with rc and output (where output is the return code of NOTE::Volume::wafliron_status).
## rc can be currently: C_RC_SUCCESS/C_RC_WAFLIRON_IDLE/C_RC_NO_WAFLIRON_IN_PROGRESS.
## 
## @examples
## $obj->get_wafliron_status();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub get_wafliron_status
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        # No args
                } );
        
        my $status = $class->{_vol}->wafliron_status();
        my $rc = C_RC_FAILURE;
        if ($status == 1)
        {
                $rc = C_RC_SUCCESS;
        }
        elsif ($status == 2)
        {
                $rc = C_RC_WAFLIRON_IDLE;
        }
        elsif ($status == 0)
        {
		$rc = C_RC_NO_WAFLIRON_IN_PROGRESS;
	}
        
        if ($rc == C_RC_FAILURE)
        {
                logresult("fatal", "Failed to obtain wafliron status ($status). Volume: " . $class->{vol_name});
        }
        
        # TODO: add oc status ?
        return ( rc => $rc, output => $status );
}


## @name ::stop_wafliron
## @summary Terminates the wafliron process.
## 
## @description
## Terminates the wafliron process. Wrapper around NOTE::Volume::wafliron_stop.
##
## @return Hash with rc and output (where output is the return code of NOTE::Volume::wafliron_stop).
##
## @arg timeout Optional: Timeout value for wafliron to be stopped. 
## If not specified, Wafliron::wafliron_stop will use its own defaults (currently 2400).
##  
## @examples
## $obj->stop_wafliron( timeout => 3600 );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub stop_wafliron
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        timeout => { type => SCALAR, optional => 1, },
                } );
        
        my $stop_rc = $class->{_vol}->wafliron_stop(%args);
        my $rc = ($stop_rc == 1) ? C_RC_SUCCESS : C_RC_NO_WAFLIRON_IN_PROGRESS;
        
        return ( rc => $rc, output => $stop_rc );
}


## @name ::detect_indirect_type
## @summary Detects the indirect type of a flex volume (32/64 -bit).
## 
## @description
## Detects the indirect type of the volume.
## For flex volumes, this is the indirect type of the containing aggregate.
## For traditional volumes, the indirect type will be defaulted to 32.
##
## @return Hash with rc, indirect_type, and output (status hashref).
##  
## @examples
## $obj->detect_indirect_type();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub detect_indirect_type
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        # None
                } );
        
        if ($class->{_type} eq C_VOL_TYPE_TRAD)
        {
                logcomment("Volume: detect_indirect_type called for a traditional volume. Returning 32") if $class->{_debug};
                return (rc => C_RC_SUCCESS, indirect_type => C_AGGR_32_BIT, output => '');
        }
        
        my $container = $class->{_vol}->container;
        
        my $status = $container->status();
        
        my $indirect_type = detect_indirect_type_from_status(status_string => $status->{status});
        
        return (rc => C_RC_SUCCESS, indirect_type => $indirect_type, output => $status);
}


## @name ::run_wafl_check
## @summary Runs WAFL_check and boots up the filer.
## 
## @description
## Runs WAFL_check and boots up the filer. Wrapper around wack::wack. 
##
## @return Hash with rc and output.
##
## @arg target Optional: The entity to run wack on. Could be: '-all', aggregate name, volume name.
## Defaults to: the aggregate containing this volume (for flex volumes), or
## to the name of this volume (for traditional volumes).
##
## @arg timeout Optional: Timeout for WAFL_check to complete.
##
## @arg compression Optional: Flag to enable compression-related checks (command option '-c').
## Defaults to TRUE.
##
## @arg commit Optional: Flag to control if any wack changes will be commited.
## Defaults to FALSE.
##
## @arg bootimage Optional: Option for boot::transit_to when attempting to reach boot menu.
## (description of what kernel to boot).
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Defaults to C_RC_SUCCESS (which means that wack found the filesystem clean).
##
## @arg disable_boot_up Optional: Flag to be used if filer boot-up is not wanted after WAFL_check completes.
## Defaults to FALSE.
##
## @notes
## If the subroutine is called without the 'target' argument, then an extra check is done to ensure
## that wack is not run on a 64-bit aggregate (the script is terminated as not applicable). If the 'target' argument is specified,
## then no check for the indirect type is done.
##  
## @examples
## $obj->run_wafl_check();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub run_wafl_check
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        target => { type => SCALAR, optional => 1, },
                        timeout => { type => SCALAR, optional => 1, default => C_DFLT_WACK_TIMEOUT },
                        compression => { type => BOOLEAN, optional => 1, default => TRUE, },
                        #prev_cp => { type => BOOLEAN, optional => 1, default => FALSE, },
                        #prev_cp_selection => { type => BOOLEAN, optional => 1, default => undef, },
                        commit => { type => BOOLEAN, optional => 1, default => FALSE, },
                        bootimage => { type => SCALAR, optional => 1, },
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS, },
                        disable_boot_up => { type => BOOLEAN, optional => 1, default => FALSE, },
                } );
        
        logcomment("Volume: run_wafl_check input:\n" . Dumper(%args)) if $class->{_debug};
        
        my $rc = C_RC_FAILURE;
        my $output = '';
        my $host_rec = $class->{_vol}->Filer->hostrec();
        my $disable_boot_up = $args{disable_boot_up};
        my $wack_args = ($args{compression} == TRUE) ? '-c ' : '';
        # Note: Not enabling support for prev_cp for now.
        # When we will support it, we will have to deal with the following issue:
        # when running wack with the prev_cp option, using any inofile selection other than "r"
        # causes the "commit changes" question to pop up and wack returns with WACK_DIRTY.
        # Wack output needs to be directly parsed for errors to decide if there is any real corruption.
        #if ($args{prev_cp} == TRUE)
        #{
        #        $wack_args .= '-prev_cp '
        #}
        
        my $target = $args{target};
        if (!defined($target))
        {
                $target = ($class->{_type} eq C_VOL_TYPE_TRAD) ? $class->{vol_name} : $class->{_vol}->container->name();
                
                # Also check for 64-bit flex volumes. We should not run wack...
                my %current_type = $class->detect_indirect_type();
                if ($current_type{indirect_type} == C_AGGR_64_BIT)
                {
                        logresult ('NA', "NOT APPLICABLE: Target $target is 64-bit. Cannot run WAFL_check.");
                }
        }
        
        $wack_args .= $target;
        
        logcomment("Volume: WAFL_check args: $wack_args") if $class->{_debug};
        
        my $wack_result = wack($host_rec,
                               args => $wack_args,
                               commit => $args{commit},
                               bootimage => $args{bootimage},
                               timeout => $args{timeout},
                               #selection => $args{prev_cp_selection},
                               output => \$output,);
        if (!defined($wack_result))
        {
                logresult("fatal", "Undefined result returned by wack");
        }
        elsif ($wack_result eq WACK_CLEAN)
        {
                $rc = C_RC_SUCCESS;
        }
        elsif ($wack_result eq WACK_DIRTY)
        {
                $rc = C_RC_WACK_ERRORS_FOUND;
        }
        elsif ($wack_result eq WACK_NOTDONE)
        {
                $rc = C_RC_WACK_NOT_RUN;
        }
        elsif ($wack_result eq WACK_INFONEEDED)
        {
                $rc = C_RC_WACK_INFO_NEEDED;
        }
        elsif ($wack_result eq WACK_CONSOLENOTLOGGED)
        {
                $rc = C_RC_WACK_NOT_CONSOLE_LOGGED;
        }

        logcomment("Volume: WAFL_check complete: result=$wack_result. Disable boot-up flag is $disable_boot_up") if $class->{_debug};
        if ($disable_boot_up != TRUE)
        {
                logcomment("Volume: Booting up the filer") if $class->{_debug};
                $class->{_vol}->Filer->reboot();
        }

        my %return_hash = $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                               expected_rc => $args{expected},
                                                               log_message => "Run WAFL_check using arguments \"" . $wack_args . "\"",
                                                               output      => $output);
        
        # Before returning, we need to parse the output for any panic string.
        # wack.pm does not detect this for us
        # The && in this conditional may be too restrictive. It is intended to protect us from false alarms.
        if (($output =~ /PANIC:/i) && ($output =~ /DUMPCORE: /i))
        {
                logresult("fatal", "Panic was detected in the WAFL_check output:\n$output");
        }
        
        return %return_hash;
}


## @name ::snapshot_reserve
## @summary Get/Set the snap reserve for a particular volume 
## 
## @description
## Get/Set the snap reserve for a particular volume.  Wrapper around NOTE::Volume::snap_reserve
##
## @return Hash with rc and output which equals the original value
##
## @arg percent Optional: If present, sets the snap reserve to the percentage specified and returns the previous setting.
##		If not present, returns the current setting. 
##
##  
## @examples
##	 $obj->snapshot_reserve( percent=>0 ); # Set no snap reserve
##	 $obj->snapshot_reserve( percent=>20 ); # Set 20 percent snap reserve
##	 $obj->snapshot_reserve();
## 
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub snapshot_reserve
{
        my $class = shift;
	my (%args) = validate( @_,
		{
			percent => { type => SCALAR, optional => 1, },
        } );
        my $percent = $args{'percent'};
        
        logcomment("Volume: snapshot_reserve. Input:\n" . Dumper(%args)) if $class->{_debug};
        
        my $orig_setting = $class->{_vol}->snap_reserve( $percent );
        logcomment("Volume: snapshot_reserve. original setting: $orig_setting") if $class->{_debug};
        
        return (rc => C_RC_SUCCESS, output => $orig_setting);
}

## @name ::get_reserved_space
## @summary Gets the reserved space of the volume.
## 
## @description
## Gets the reserved space of the volume.
##
## @return Hash with rc and reserved ( in KB )
##  
## @examples
## $obj->get_reserved_space();
## 
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub get_reserved_space
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        # None
                } );
        my %return_hash = ();

		#my $status_obj = vol_status($class->{_vol}->Filer->conn, $class->{vol_name} );
		#my $df_obj = disk_free($class->{_vol}->Filer->conn, $class->{_vol}->mount_path, DF_RESERVE);
		my $df_obj = disk_free($class->{_vol}->Filer->conn, $class->{_vol}->mount_path, 'reserve');

        $return_hash{rc} = C_RC_SUCCESS;
		$return_hash{reserved} = $df_obj->{reserved};
        
        logcomment("Volume: get_reserved_space. Returning: " . Dumper(%return_hash)) if $class->{_debug};
        return %return_hash;
}


## @name ::get_snapshot_info
## @summary get snapshot info for a given snapshot on this volume
## 
## @description
## Get snapshot info for a given snapshot on this volume. Wrapper around SnapCmds::snapshot_list_info.
##
## @return Hash with rc and info which is a reference to a hash with keys as shown below.
##
## @arg name Required: name of snapshot
##
## @examples
## $obj->get_snapshot_info(snapshot=>$snapshot);
##
##	$res{info} example:
##	'access-time' => '1251495943',
##	'cumulative-percentage-of-used-blocks' => '0',
##	'cumulative-percentage-of-total-blocks' => '0',
##	'name' => 'ss_test',
##	'total' => '0',
##	'percentage-of-total-blocks' => '0',
##	'dependency' => '',
##	'cumulative-total' => '0',
##	'busy' => 'false',
##	'percentage-of-used-blocks' => '0'
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub get_snapshot_info
{
    my $class = shift;
	my (%args) = validate (@_, { name => { type=>SCALAR },});
	my $snapshot_name = $args{name};
	my %info;
	my @snapshots = snapshot_list_info( $class->{_vol}->Filer->conn, $class->{_vol}->name );
	logcomment("dumper: ". Dumper(@snapshots) )  if $class->{_debug};
	foreach my $ss (@snapshots) {
		if ($ss->{name} eq $snapshot_name) {
			%info = %{$ss};
			last;
		}
	}
	return (rc => C_RC_SUCCESS, info=>\%info);
}


## @name ::set_snap_autodelete
## @summary set snap autodelete state
## 
## @description
## Set snap autodelete on this volume. Wrapper around SnapCmds::snapshot_autodelete_cmd.
##
## @return Hash with rc.
##
## @arg state Required: state to set snap autodelete to
##
## @examples
## $obj->set_snap_autodelete(state=>'on');
## $obj->set_snap_autodelete(state=>'off');
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub set_snap_autodelete
{
    my $class = shift;
	my (%args) = validate (@_, { state => { type=>SCALAR, regex=> qr/^(on|off|reset)$/, }});
	my $state = $args{state};
	logcomment("Setting snap autodelete to $state");	
	snapshot_autodelete_cmd( $class->{_vol}->Filer->conn, $class->{_vol}->name, $state );
	return (rc => C_RC_SUCCESS);
}


## @name ::set_snap_autodelete_options
## @summary set snap autodelete options
## 
## @description
## Set snap autodelete options on this volume. Wrapper around SnapCmds::snapshot_autodelete_options.
##
## @return Hash with rc.
##
## @arg Optional:
##
## @examples
## $obj->set_snap_autodelete_options(
##						trigger=>'volume', 
##						target_free_space=>80,
##						delete_order=>'oldest_first',);
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub set_snap_autodelete_options
{
        my $class = shift;
        my (%args) = validate (@_,
        {
                commitment => { type=>SCALAR, optional=>1, regex=> qr/^(try|disrupt)$/, },
                trigger => { type=>SCALAR, optional=>1, regex=> qr/^(volume|snap_reserve|space_reserve)$/, },
                target_free_space => { type=>SCALAR, optional=>1, },
                delete_order => { type=>SCALAR, optional=>1, regex=> qr/^(newest_first|oldest_first)$/, },
                defer_delete => { type=>SCALAR, optional=>1, regex=> qr/^(scheduled|user_created|prefix|none)$/, },
                prefix => { type=>SCALAR, optional=>1, },
	});
        
	my $state = $args{state};
	my $options;
	foreach my $option (sort keys %args) {
		$options .= "$option: $args{$option}\n";
	}
	logcomment("set_snap_autodelete_options\n".$options);
	snapshot_autodelete_options( $class->{_vol}->Filer->conn, $class->{_vol}->name, %args );
	return (rc => C_RC_SUCCESS);
}

## @name ::inject_error
## @summary Inject corruption.
##
## @description
## Inject corruption. Wrapper around NOTE::Volume::wreck,
## with added functionality for generating the inject_script.
##
## @arg code Required: Code of the error to be injected.
## Ignored if inject_script arguments is defined.
##
## @arg filename Optional: Name of the file to inject the error to.
## Ignored for non file-specific corruptions and if inject_script is passed in directly.
##
## @arg bno Optional: Specific block number within a file to inject corruption at.
## Ignored for non file-specific corruptions and if inject_script is passed in directly.
##
## @arg start Optional: Start byte for corruption within a block.
## Ignored for non file-specific corruptions and if inject_script is passed in directly.
##
## @arg end Optional: End byte for corruption within a block.
## Ignored for non file-specific corruptions and if inject_script is passed in directly.
##
## @arg fill Optional: Byte to be used when corrupting a specific block pattern.
## Ignored for non file-specific corruptions and if inject_script is passed in directly.
##
## @arg cg_no Optional: CG number at which to injectthe corruption.
## This argument is only accounted for in the case of corrupting the CG (header, trailer).
##
## @arg new_bno Optional: Value to be used to corrupt an indirect block (ind_vol_bno).
##
## @arg new_bno Optional: Value to be used to corrupt an indirect block vvbn (ind_vol_vbno).
##
## @arg inject_script Optional: Text to be used when injecting the error.
##
## @return Hash with rc and output.
##  
## @examples
## $obj->inject_error(code => C_CORRUPT_FSINFO_INOS_USED);
## $obj->inject_error(inject_script => 'fsinfo_inos_used count=460');
##
## @note
## After injecting corruption, it will also set the wafl_iron_panic_on_corruption flag to 0.
##
## @note
## The way this method requires specific arguments varies by the corruption to be injected.
## Passing in more arguments than what is really needed for a certain corruption,
## will result in the extra arguments being ignored.
## Passing in less arguments than what is needed for a specific corruption,
## will result in script failure, asking for the parameter that is missing and cannot be auto-generated.
## Here are a few examples/rules:
## - If passing inject_script directly, all the other arguments are ignored.
## - For code=C_CORRUPT_FSINFO_INOS_USED: no other arguments are required.
## - For code=C_CORRUPT_FSINFO_BLKS_USED_BY_PLANE0: no other arguments are required.
## - For code=C_CORRUPT_FSINFO_BLKS_COMPRESS_HOLES: no other arguments are required.
## - For code=C_CORRUPT_INODE_BLK_CNT: filename (required).
## - For code=C_CORRUPT_REFCOUNT: filename (required), bno (optional->default: 1).
## - For code=C_CORRUPT_CHECKSUM: filename (required), bno (optional->default: 1), fill (optional->default: randomly generated).
## - For code=C_CORRUPT_CG_HEADER_VBN: filename (required), new_bno (required), cg_no (optional->default: 0)
## - For code=C_CORRUPT_CG_HEADER_VVBN: filename (required), new_vbno (required), cg_no (optional->default: 0)
## - For code=C_CORRUPT_CG_TRAILER_VBN: filename (required), new_bno (required), cg_no (optional->default: 0)
## - For code=C_CORRUPT_CG_TRAILER_VVBN: filename (required), new_bno (required), cg_no (optional->default: 0)
## - For code=C_CORRUPT_REC: filename (required), bno (optional->default: 0)
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub inject_error
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        inject_script => { type => SCALAR, optional => 1, },
                        code => { type => SCALAR, optional => 1, },
                        filename => { type => SCALAR, optional => 1, },
                        bno => { type => SCALAR|UNDEF, optional => 1, },
                        start => { type => SCALAR, optional => 1},
                        end => { type => SCALAR, optional => 1},
                        fill => { type => SCALAR|UNDEF, optional => 1},
                        cg_no => { type => SCALAR|UNDEF, optional => 1},
                        new_bno => { type => SCALAR|UNDEF, optional => 1},
                        new_vbno => { type => SCALAR|UNDEF, optional => 1},
                        level => { type => SCALAR|UNDEF, optional => 1},
                } );


        if (!defined($args{inject_script}))
        {
                if (!defined($args{code}))
                {
                        logresult("script", "None of the inject_script or code arguments have been passed to wreck");
                }
                
                $args{inject_script} = $class->{_ui}->get_wreck_inject_cmd(%args);
        }
        
        # TODO: Waiting for burt 375918. For the moment, use Wreck::wreck directly,
        # to avoid onlining the volume after wreck.
        #my $wreck_res = $class->{_vol}->wreck(inject_script => $args{inject_script});
        my $volume = $class->{_vol};
        
        # Looks like all the code in Wreck::wreck is placed within an eval block
        # Need to check here for success
        my $wreck_res = Wreck::wreck($volume->Filer->conn, $volume->name, type=>$volume->type, inject_script => $args{inject_script});

        if (!$wreck_res)
        {
                logresult('fatal', "Failed to wreck the volume " . $class->{vol_name} . ":\n" . $@);
        }

        return (rc => C_RC_SUCCESS, output => $wreck_res);
}


## @name ::size
## @summary 
## Set size of volume.
## 
## @description
## Set or get size of volume.  Wrapper around NOTE::Volume::size
## 
## @return rc and size
##
## @arg size Required: 
##
## @examples
## $obj->size(size=>'500m');
##
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub size
{
        my $class = shift;
        my (%args) = validate (@_,
                {
                        size => { type=>SCALAR, optional=>1 },
                });
        
        my $size = $args{size};
        my $vol_name = $class->{_vol}->name;
        my $new_size;
        if ( defined $args{size} )
        {
                logcomment("setting size to: $size");
                $new_size = $class->{_vol}->size($size);
        }
        else
        {
                $new_size = $class->{_vol}->size();
                logcomment("volume size of $vol_name is $new_size");
        }
        
        # If we got here, NOTE did not encounter any error. Return success.
        return (rc => C_RC_SUCCESS, size=>$new_size);
}


## @name ::nfs_export_options
## @summary Gets or sets NFS export options on the volume.
## 
## @description
## Gets or sets NFS export options on the volume. Wrapper around NOTE::Volume::exportfs.
##
## @arg new_setting Optional: New setting for the NFS export option.
## If not specified, the current setting is returned.
##
## @note
## If called multiple times,
## this function will not re-issue the exportfs command to the filer unless the exportfs option string changes.
## Also, even if called to get the current setting, if there is a difference between the current setting on the filer and
## what this setting should have been (last set value), then the filer setting is resynced to match this last-set value.
##
## @return Hash with rc and output (representing the current option string, if new_setting is not defined,
## and the NOTE vol obj if new_setting is defined).
##  
## @examples
## $obj->nfs_export_options($new_setting => C_DFLT_VOL_EXPORTFS)
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub nfs_export_options
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        new_setting => { type => SCALAR|UNDEF, optional => 1, },
                } );
        
        my $new_setting = $args{new_setting};
        my $exportfs;
        
        # TODO: still need this logic after burts 441238/442667 are fixed?
        if (defined($class->{__vfiler}))
        {
                $exportfs = $class->{__vfiler}->exportfs(container => $class->{_vol});
        }
        else
        {
                $exportfs = $class->{_vol}->exportfs($new_setting);
        }
        
        return (rc => C_RC_SUCCESS, output => $exportfs);
}


## @name ::set_autosize
## @summary set vol autosize state
## 
## @description
## Set vol autosize state
##
## @return Hash with rc.
##
## @arg state Required: state to set autosize to
##
## @examples
## $obj->set_autosize(state=>'on', size=>'30g');
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub set_autosize
{
        my $class = shift;
        my (%args) = validate (@_,
                {
                        state => { type=>SCALAR, regex=> qr/^(on|off|reset)$/, },
                        max_size  => { type=>SCALAR, },
                        increment  => { type=>SCALAR, },
                });
        
	my $state = $args{state};
	my $max_size = $args{max_size};
	my $increment = $args{increment};
	my $vol_name = $class->{_vol}->name;
        
	logcomment("Setting vol autosize to $state for volume: $vol_name\n".
                   "state: $state max_size: $max_size increment: $increment");
        
	my $cmd_code = C_CMD_AUTOSIZE;
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $vol_name,
                                                       size => $max_size,
                                                       increment => $increment,
                                                       state => $state);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_vol}->Filer,
                                                            boot_mode => C_ONTAP_7MODE,
                                                            cmd => $command,
                                                            node_run => FALSE,);
        
	logcomment("set_autosize: command output: $output") if ($class->{_debug});
	
	return $class->{_ui}->process_output_string(code => $cmd_code,
                                                    output => $output,
                                                    expected_rc => C_RC_SUCCESS,
                                                    log_message => "Set autosize for volume: $vol_name",);
}


## @name ::is_block_type_32bit
## @summary Checks if the volume is 32-bit.
## 
## @description
## Checks if the volume is 32-bit.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc.
##  
## @examples
## $obj->is_block_type_32bit(expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_block_type_32bit
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
                } );
	
	my $vol_name = $class->{_vol}->name;
        
	my $is_64bit = volume_is_64bit($class->{_vol}->Filer->conn(), $vol_name);
        my $rc = ($is_64bit == 1) ? C_RC_FAILURE : C_RC_SUCCESS;
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Check if volume block type vor $vol_name is 64-bit",
                                                    output      => $is_64bit);
}


## @name ::is_block_type_64bit
## @summary Checks if the volume is 64-bit.
## 
## @description
## Checks if the volume is 64-bit.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc.
##  
## @examples
## $obj->is_block_type_64bit(expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_block_type_64bit
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
                } );
	
        my $vol_name = $class->{_vol}->name;
        
	my $is_64bit = volume_is_64bit($class->{_vol}->Filer->conn(), $vol_name);
        my $rc = ($is_64bit == 1) ? C_RC_SUCCESS : C_RC_FAILURE;
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Check if volume block type vor $vol_name is 64-bit",
                                                    output      => $is_64bit);
}


## @name ::is_block_type_64bit
## @summary Checks if the volume is 64-bit.
## 
## @description
## Checks if the volume is 64-bit.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc.
##  
## @examples
## $obj->is_block_type_64bit(expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_block_type_upgrading
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
                } );
	
        my $vol_name = $class->{_vol}->name;
        
	my $is_upgrading = volume_is_hybrid($class->{_vol}->Filer->conn(), $vol_name);
        my $rc = ($is_upgrading == 1) ? C_RC_SUCCESS : C_RC_FAILURE;
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Check if volume block type vor $vol_name is upgrading",
                                                    output      => $is_upgrading);
}


## @name ::wait_for_block_type_64bit
## @summary Waits until the volume is marked as 64-bit.
## 
## @description
## Waits until the volume is marked as 64-bit.
##
## @arg timeout Optional: Timeout value (in seconds).
## Default is C_DFLT_TIMEOUT, as defined in Constants.pm.
##
## @arg check_interval Optional: Number of seconds between two consecutive status checks.
## Defaults to 1 sec.
##
## @return Hash with rc.
##  
## @examples
## $obj->wait_for_block_type_64bit(timeout => 600);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub wait_for_block_type_64bit
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        timeout => {type => SCALAR, optional => 1, default => C_DFLT_TIMEOUT,},
                        check_interval => {type => SCALAR, optional => 1, default => 1,},
                } );
	
        my $timeout = $args{timeout};
        my $check_interval = $args{check_interval};
        
        my $vol_name = $class->{_vol}->name;
	my $is_64bit = 0;
        my $is_upgrading = 1;
        
        logcomment("Waiting for volume $vol_name to be marked as 64-bit");
        
        while (($timeout > 0) && (($is_64bit != 1) || ($is_upgrading != 0)))
        {
                $is_64bit = volume_is_64bit($class->{_vol}->Filer->conn(), $vol_name);
                $is_upgrading = volume_is_hybrid($class->{_vol}->Filer->conn(), $vol_name);
                sleep($check_interval);
                $timeout -= $check_interval;
        }
        
        if ($timeout == 0)
        {
                logresult('fatal', "Timed-out while waiting for volume $vol_name to become 64-bit\n" .
                                   "is_upgrading: $is_upgrading\n" .
                                   "is_64bit: $is_64bit");
        }
        
        return (rc => C_RC_SUCCESS);
}


## @name ::start_64_bit_upgrade
## @summary Starts upgrading the volume to 64-bit.
## 
## @description
## Starts upgrading the volume to 64-bit.
##
## @arg mode Optional: Upgrade mode: check|normal.
## Default: none.
##  
## @examples
## $obj->start_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub start_64_bit_upgrade
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        mode => {type => SCALAR, optional => 1, default => C_LA_IPU_MODE_NORMAL},
                } );
        
        my $vol_name = $class->{vol_name};
        my $mode = $args{mode};
        my $error_msg;
        
        logcomment("Starting LA IPU on volume $vol_name, mode: $mode");
        my $result = IPU::7ModeIPU::vol_upgrade_diag(Fc => $class->{_vol}->Filer->conn(),
                                                     vol_name => $vol_name,
                                                     mode => $mode,
                                                     error_msg => \$error_msg);
        
        # There is not really a very good way to make it more reliable than this...
        # vol_upgrade_diag will, in the most cases, return the raw string,
        # but for some cases it may return integers
        # For now, rely on "no error message"
        if ((defined($error_msg)) || ("$result" eq "-1"))
        {
		my $fail_msg = '';
		if (defined($error_msg))
		{
			$fail_msg .= $error_msg;
		}
		if (defined($result))
		{
			$fail_msg .= "\n$result";
		}
		logresult('fatal', "Failed to start LA IPU on volume $vol_name: $fail_msg");
        }
}


## @name ::check_buftree
## @summary Runs buftree check on this volume.
## 
## @description
## Runs buftree check on this volume.
##
## @arg filename Optional: Name of file to run buftree check on.
##
## @arg priv Optional: Priv level (integer value).
##
## @arg dont_panic Optional: Flag to disable the filer panic
## in case the buftree checker finds an inconsistency.
##
## @arg timeout Optional: Timeout in seconds.
##
## @return Hash with rc and buftree_stats.
##  
## @examples
## $obj->check_buftree();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub check_buftree
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        filename => {type => SCALAR, optional => 1},
                        priv => {type => SCALAR, optional => 1},
                        dont_panic => {type => SCALAR, optional => 1},
                        timeout => {type => SCALAR, optional => 1},
                } );
        
        my $vol_name = $class->{vol_name};
        my $filename = delete($args{filename});
        my $priv = delete($args{priv});
        
        if ((!defined($filename)) && (!defined($priv)))
        {
                logresult('script', "check_buftree: One of the filename or priv argument needs to be passed in");
        }
        
        if ((defined($filename)) && (defined($priv)))
        {
                logresult('script', "check_buftree: Only one of the filename or priv argument needs to be passed in");
        }
        
        my $inodespec;
        if (defined($filename))
        {
                $inodespec = "path:/$filename";
        }
        else
        {
                $inodespec = "$priv";
        }
        
        logcomment("Starting buftree check on volume $vol_name");
        
        my $stats = HybridBuftree::buftree_check($class->{_vol}->Filer->conn(), $vol_name, $inodespec, %args);
        logcomment("Volume\:\:check_buftree: buftree_check returned:\n" . Dumper($stats)) if ($class->{_debug});
        
        # Process the stats to make the return hash consistent with c-mode
        my $processed_stats = {};
        $processed_stats->{total} = $stats->{total};
        
        my $levels_array = $stats->{levels};
        my $level = 0;
        foreach my $level_info (@$levels_array)
        {
                $processed_stats->{$level} = $level_info;
                $level++;
        }
        
        return (rc => C_RC_SUCCESS, buftree_stats => $processed_stats);
}


## @name ::set_snaplock_retention_period
## @summary Sets the snaplock retention period for this volume.
## 
## @description
## Sets the snaplock retention period for this volume.
##
## @arg default Optional: Value to be used to set the default retention period.
##
## @arg minimum Optional: Value to be used to set the minimum retention period.
##
## @arg maximum Optional: Value to be used to set the maximum retention period.
##
## @return @return Hash with rc and new retention periods.
##  
## @examples
## $obj->set_snaplock_retention_period(default => '90s',
##                                     minimum => '90s');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub set_snaplock_retention_period
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        default => {type => SCALAR, optional => 1,},
                        minimum => {type => SCALAR, optional => 1,},
                        maximum => {type => SCALAR, optional => 1,},
                } );
	
        my $vol_name = $class->{_vol}->name;
        
        logcomment("Setting snaplock retention periods for volume $vol_name to:\n" . Dumper(%args)); 
        my ($min, $default, $max) = snaplock::set_retention_periods($class->{_vol}->Filer->conn(), $vol_name, %args);
        
        return (rc => C_RC_SUCCESS,
                minimum => $min,
                default => $default,
                maximum => $max);
}


## @name ::set_snaplock_auto_commit
## @summary Sets the snaplock auto-commit period for this volume.
## 
## @description
## Sets the snaplock auto-commit period for this volume.
##
## @arg value Required: Value to be set for the auto-commit period.
##
## @return Hash with rc.
##  
## @examples
## $obj->set_snaplock_auto_commit(value => '10s');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub set_snaplock_auto_commit
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        value => {type => SCALAR,},
                } );
	
        my $vol_name = $class->{_vol}->name;
        my $value = $args{value};
        
        logcomment("Setting snaplock auto-commit period for volume $vol_name to: $value");
        
        my $output = SnapLock::Snaplock_auto_commit::set_vol_auto_commit($class->{_vol}, $value);
        
        return $class->{_ui}->process_output_string(code => C_CMD_SET_SNAPLOCK_AUTO_COMMIT,
						    output => $output,
                                                    expected_rc => C_RC_SUCCESS,
						    log_message => "Set snaplock auto-commit volume $vol_name to $value",);
}


## @name ::get_expiry_date
## @summary Gets the retention expiration date for this volume.
## 
## @description
## Gets the retention expiration date for this volume.
##
## @return Hash with rc and expiry_date.
##  
## @examples
## my %res = $obj->get_expiry_date();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_expiry_date
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        # None
                });
	
        my $vol_name = $class->{_vol}->name;
        
        my $output = snaplock::volume_expiry_date($class->{_vol}->Filer->conn(), $vol_name);
        
        return (rc => C_RC_SUCCESS,
                expiry_date => $output->{expiry_date});
}


## @name ::get_snaplock_containment_bits
## @summary Gets snaplock containment bits for this volume.
## 
## @description
## Gets snaplock containment bits for this volume.
##
## @return Hash with rc and snaplock_containment_stats.
##  
## @examples
## $obj->get_snaplock_containment_bits();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_snaplock_containment_bits
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        # None
                });
	
	my $vol_name = $class->{_vol}->name;
	
	my $cmd_code = C_CMD_GET_SNAPLOCK_CONTAINMENT_BITS;
	
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $vol_name,);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_vol}->Filer,
                                                            boot_mode => C_ONTAP_7MODE,
                                                            cmd => $command,
                                                            node_run => FALSE,);
	
	return $class->{_ui}->process_output_string(code => $cmd_code,
                                                    output => $output,
                                                    expected_rc => C_RC_SUCCESS,
                                                    log_message => "Get snaplock containment bits for volume: $vol_name",);
}


################################
# Internal subroutines
################################

# Keep this method internal. Renaming should be done only through Compress.
# This way, all members of the Compress object will have the volume name adjusted properly.
# Must return the new name
sub _rename
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        new_name => { type => SCALAR|UNDEF, },
                } );
        
        $class->{_vol}->rename($args{new_name});
        
        $class->{vol_name} = $class->{_vol}->name();
        
        # If we got here, NOTE did not encounter any error. 
        return ($class->{vol_name});
}

sub _refresh_aggregate
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		new_aggr => {type => OBJECT, isa => "NOTE::Aggregate"},
	});
	
	my $new_aggr = $args{new_aggr};
	my $vol_name = $class->{vol_name};
	
	$class->{_vol} = NOTE::Volume->attach(Filer => $new_aggr->Filer,
					      Aggregate => $new_aggr,
					      name => $vol_name,);
	$class->{_vol}->SPARE();
}


sub _refresh_vfiler
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		new_vfiler => {type => OBJECT|UNDEF,},
	});
	
	my $new_vfiler = $args{new_vfiler};
	
	$class->nfs_umount();
	$class->{_vfiler} = $new_vfiler;
	if (defined($new_vfiler))
        {
		if (!($new_vfiler->isa("Compression::7Mode::Vfiler")))
                {
                        logresult('script', "The vfiler argument is required to be a Compression\:\:7Mode\:\:Vfiler object");
                }
                
                $class->{__vfiler} = $new_vfiler->_get_underlying_obj();
                $class->{__vfiler}->exportfs(container => $class->{_vol});
	}
	else
	{
		$class->{__vfiler} = undef;
	}
	$class->nfs_mount();
}


sub _process_wafliron_wait_result
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        corruption_found => { type => SCALAR, },
                        message => { type => SCALAR, },
                        cp_hashref => { type => HASHREF|UNDEF, },
                        expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS, },
                } );
        
        logcomment("7Mode Volume: processing wafliron results:\n" . Dumper(%args)) if ($class->{_debug});
        
        my $rc = C_RC_SUCCESS;
        
        if ($args{corruption_found} == 1)
        {
                $rc = C_RC_WAFLIRON_CORRUPTION_FOUND;
        }
        
        my %return_hash = $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                               expected_rc => $args{expected},
                                                               log_message => "Process wafliron result",
                                                               output      => $args{message});
        
        if (defined($args{cp_hashref}))
        {
                $return_hash{cp_hashref} = $args{cp_hashref};
        }
        return %return_hash;
}

# Returns the underlying filer object (to be used with _send_cmd_to_filer)
sub _get_conn
{
	my $class = shift;
	return $class->{_vol}->Filer();
}


# Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
        my $volume = $class->{vol_name};
	
	if (($class->{_disable_cleanup} == FALSE) &&
            ($class->{_creator_pid} == $$) &&
            (!defined($ENV{'C_CORRUPTION_FOUND'}) || ($ENV{'C_CORRUPTION_FOUND'} != TRUE)))
	{
		logcomment("Executing 7Mode Volume clean-up for volume \"$volume\"");
		
                eval
                {
                        if ($class->{_recycled} == FALSE)
                        {
                                logcomment("7Mode Volume: cleanup: purging $volume") if ($class->{_debug});
                                $class->{_vol}->purge( );
                        }
                        
                };
                
                if( $@ )
                {
                        logcomment( "7Mode Volume: cleanup for volume \"$volume\": failure due to : $@" );
                }
                else
                {
                        logcomment( "7Mode Volume: cleanup for volume \"$volume\": done" ) if ($class->{_debug});
                }
        }
	else
	{
                logcomment( "7Mode Volume: cleanup for volume \"$volume\": nothing done: script PID: " . $$) if ($class->{_debug});
	}
        
        if (defined($execution_error))
        {
                _c_interpret_logresult($execution_error);
        }
}



1;
