#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/AggrCopy.pm#1 $
#

## @summary Aggregate copy routines.
##
## @description
##    This module provides all the capabilities
##	required for aggregate copy.
##
## @synopsis
## 
## BEGIN { require "drg/Compression/setup_core.pl"; }
##
## use Compression::AggrCopy;
## 
## my $obj = Compression::AggrCopy->new(source => $compress1,
##                                      destination => $compress2); 
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::AggrCopy;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Tharn;

use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult);
use Aggregate::Copy qw(aggrcopy_throttle);


BEGIN
{
	require "Compression/base_import.pl";
}



## @name ::new
##
## @summary Compression::AggrCopy Constructor.
##
## @description
## Creates an Compression::AggrCopy object 
## (it does not start the copy relationship on the filer).
##
## @return Compression::AggrCopy object.
##
## @arg source Required: Source Compression::Compress object.
##
## @arg destination Required: Destination Compression::Compress object.
##
## @arg disable_cleanup Optional: Flag to disable any clean-up that is executed
## upon object destruction.
## Defaults to $main::DISABLE_CLEANUP.
##
## @example
## my $aggr_copy = Compression::AggrCopy->new(source => $compress1,
##                                            destination => $compress2);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
		source 	=> {type => OBJECT,
                            isa => "Compression::Compress",},

		destination  => {type => OBJECT,
                                 isa => "Compression::Compress",},
                
                disable_cleanup => {type => BOOLEAN,
                                    optional => 1,
                                    default => $main::DISABLE_CLEANUP,},
                
                debug => {type => BOOLEAN,
                          optional => 1,
                          default => $main::DEBUG,},
		} );
	
	my $self = {};
        $self->{_source} = $args{source};
        $self->{_destination} = $args{destination};
        $self->{_disable_cleanup} = $args{disable_cleanup};
        $self->{_creator_pid} = $$;
        $self->{_boot_mode} = $self->{_source}->{boot_mode};
        $self->{_debug} = $args{debug};
        
        $self->{_src_aggr_name} = $self->{_source}->{aggr_name};
        $self->{_src_host_name} = $self->{_source}->{host_name};
        $self->{_dst_aggr_name} = $self->{_destination}->{aggr_name};
        $self->{_dst_host_name} = $self->{_destination}->{host_name};
        
        $self->{_ui} = $self->{_source}->{_ui};
        
        if ($self->{_boot_mode} eq C_ONTAP_CMODE)
        {
                logresult('NA', "NOT APPLICABLE: Aggregate copy is not supported in C-mode");
        }
        else
        {
                $self->{_copy} = NOTE::Aggregate::Copy->new(src => $self->{_source}->{aggregate}->_get_underlying_aggr_obj(),
                                                            dst => $self->{_destination}->{aggregate}->_get_underlying_aggr_obj(),);
        }
        
        
        bless $self, $class; 

	return $self;
}


## @name ::start
## @summary Starts the aggregate copy process.
## 
## @description
## Starts the aggregate copy process.
##
## @return Hash with rc.
## 
## @arg background Optional: Flag to run the transfer in the background.
## To not be set if the expected result is not success.
## Default: FALSE.
## 
## @arg copy_all Optional: Flag to include all the volume's snapshots.
## Default: FALSE.
## 
## @arg context Optional: Flag to decide if the -C option should be used
## (needed for aggregates with block reallocations).
## Default: FALSE.
## 
## @arg snapshot Optional: The name of the snapshot, on the source volume, to copy.
##
## @arg timeout Optional: Timeout, in seconds, for the command to complete.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @examples
## $obj->start();
## $obj->start(copy_all => TRUE);
## $obj->start(snapshot => 'my_snapshot');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub start
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        background => {type => SCALAR, optional => 1, default => FALSE,},
                        copy_all => {type => SCALAR, optional => 1, default => FALSE,},
			context => {type => SCALAR, optional => 1, default => FALSE,},
                        snapshot => => {type => SCALAR, optional => 1,},
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
                } );
	
        my $background = $args{background};
        my $expected = delete($args{expected});
        
        if (($background == TRUE) && ($expected != C_RC_SUCCESS))
        {
                logresult('script', "Compression\:\:AggrCopy\:\:start: background option should not be set " .
                                    "if expected is different than C_RC_SUCCESS");
        }
        
        # The attempt done by the underlying library to restrict the aggregate fails,
        # as the aggregate has online volumes. 
        # Work-around: first force-offline the aggregate
        $class->{_destination}->{aggregate}->set_offline(force => TRUE,
                                                         all => TRUE);
        
        my $source = $class->{_src_aggr_name};
        logcomment("Starting aggr copy:\n" .
                   "Source: $source (filer: " . $class->{_src_host_name} . ")\n" .
                   "Destination: " . $class->{_dst_aggr_name} . " (filer: " . $class->{_dst_host_name} . ")\n" .
                   "Background option: " . $background);
        
        eval
        {
                $class->{_copy}->start(%args);
        };
        
        return $class->{_ui}->process_output_string(code => C_CMD_AGGR_COPY_START,
                                                    output => $@,
                                                    expected_rc => $expected,
                                                    log_message => "Start aggregate copy from $source",);
}


## @name ::wait_for_completion
## @summary Wait for the background aggregate copy transfers to complete.
## 
## @description
## Wait for the background aggregate copy transfers to complete.
##
## @return Hash with rc.
##
## @examples
## $obj->wait_for_completion();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub wait_for_completion
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        restore_destination_state => {type => SCALAR, optional => 1, default => TRUE},
                } );
        
        my $restore_destination_state = $args{restore_destination_state};
        my $attach_to_copy_vol = $args{attach_to_copy_volume};
        
        logcomment("Waiting for aggregate copy to complete:\n" .
                   "Source: " . $class->{_src_aggr_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                   "Destination: " . $class->{_dst_aggr_name} . " (filer: " . $class->{_dst_host_name} . ")");
	
        $class->{_copy}->wait();
        
        return (rc => C_RC_SUCCESS);
}


## @name ::get_status
## @summary Gets the status of aggregate copy transfer.
## 
## @description
## Gets the status of aggregate copy transfer.
##
## @return Hash with rc and status.
##
## @examples
## my %status = $obj->get_status();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub get_status
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        # None
                } );
        
        my $status = $class->{_copy}->status();
        
        return (rc => C_RC_SUCCESS,
                status => $status->{status},
                );
}


## @name ::verify_status
## @summary Verifies the status of aggregate copy transfer.
## 
## @description
## Verifies the status of aggregate copy transfer.
##
## @arg status Optional: Expected status.
## Default: C_AGGR_COPY_STATUS_TRANSFERRING.
##
## @return Hash with rc. 
##
## @examples
## $obj->verify_status(status => C_AGGR_COPY_STATUS_IDLE);
## $obj->verify_status(); # Same as $obj->verify_status(status => C_AGGR_COPY_STATUS_TRANSFERRING);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub verify_status
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        status => {type => SCALAR, optional => 1, default => C_AGGR_COPY_STATUS_TRANSFERRING},
                } );
        
        my $expected_status = $args{status};
        
        logcomment("Verifying the aggregate copy status ($expected_status):\n" .
                   "Source: " . $class->{_src_aggr_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                   "Destination: " . $class->{_dst_aggr_name} . " (filer: " . $class->{_dst_host_name} . ")");
        
        my %current_status = $class->get_status();
        
        if ($current_status{status} ne $expected_status)
        {
                logresult('fatal', "Failed to detect expected aggregate copy transfer status ($expected_status):\n" . Dumper(%current_status));
        }
        
        return %current_status;
}


## @name ::wait_for_status
## @summary Wait for the aggregate copy transfer to reach specific state.
## 
## @description
## Wait for the aggregate copy transfer to reach specific state.
##
## @return Hash with rc.
##
## @arg status Required: Status to wait for (can be a regex).
##
## @arg source Optional: Flag to get the status from the source filer.
## Default: FALSE (destination filer).
##
## @arg restore Optional: Flag to get the status of the restore transfer.
## Default: FALSE. 
##
## @arg timeout Optional: Timeout, in seconds, for completion.
##
## @examples
## $obj->wait_for_status(status => C_AGGR_COPY_STATUS_TRANSFERRING);
## $obj->wait_for_status(status => C_AGGR_COPY_STATUS_IDLE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub wait_for_status
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        status => {type => SCALAR,},
                        source => {type => SCALAR, optional => 1, default => FALSE,},
                        restore => {type => SCALAR, optional => 1, default => FALSE,},
                        timeout => {type => SCALAR, optional => 1,},
                } );
        
        my $status = $args{status};
        my $timeout = $args{timeout};
        
        my $comment = "Waiting for aggregate copy status \"$status\":\n" .
                      "Source: " . $class->{_src_aggr_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                      "Destination: " . $class->{_dst_aggr_name} . " (filer: " . $class->{_dst_host_name} . ")";
        
        if (defined($timeout))
        {
                $comment .= "\nTimeout: $timeout";
        }
        
        logcomment($comment);
	
        $class->{_copy}->status_wait(%args);
        
        return (rc => C_RC_SUCCESS);
}


## @name ::abort
## @summary Aborts an active aggregate copy transfer.
## 
## @description
## Aborts an active aggregate copy transfer.
##
## @return Hash with rc.
##
## @examples
## $obj->abort();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub abort
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        # None
                } );
        
        logcomment("Aborting aggregate copy transfer:\n" .
                   "Source: " . $class->{_src_aggr_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                   "Destination: " . $class->{_dst_aggr_name} . " (filer: " . $class->{_dst_host_name} . ")");
        
        $class->{_copy}->abort();
        
        # We need to wait here for all the background processes, but ignore any errors.
        # This is because: this abort was of an operation started with the background option.
        # Since we aborted it, the underlying library will report it as a failure.
        # On the other hand, any call to wait_for_completion after this point will
        # also include this operation that was just aborted (by NOTE's design).
        # To eliminate any issues coming from this, wait now for the background process
        # (this will remove it from the "to wait for" list), but do not fail.
        eval
        {
                $class->{_copy}->wait();
        };
        
        if (($@) && ($class->{_debug}))
        {
                logcomment("Compression\:\:AggrCopy:\n" .
                           "Waited for background copy transfer that was aborted. Output:\n" . $@);
        }
         
        return (rc => C_RC_SUCCESS);
}


## @name ::throttle
## @summary Modifies the speed at which the aggregate copy transfer is being done.
## 
## @description
## Modifies the speed at which the aggregate copy transfer is being done.
##
## @arg value Optional: New speed value, to be set.
## The  value  ranges from 10 (full speed) to 1 (one-tenth of full speed).
## Default: C_DFLT_AGGR_COPY_THROTTLE (10, full speed).
##
## @return Hash with rc. 
##
## @examples
## $obj->throttle(value => 1);
## $obj->throttle(); # To set back to default
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub throttle
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        value => {type => SCALAR, optional => 1, default => C_DFLT_AGGR_COPY_THROTTLE},
                } );
        
        my $value = $args{value};
        
        my $op_num = $class->{_copy}->op_num();
        
        if (!defined($op_num))
        {
                logresult('script', "No aggregate copy operation was started");
        }
        
        logcomment("Modifying the speed of aggregate copy operation: new throttle value: $value\n" .
                   "Source: " . $class->{_src_aggr_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                   "Destination: " . $class->{_dst_aggr_name} . " (filer: " . $class->{_dst_host_name} . ")\n" .
                   "Operation number: $op_num");
        
        my $src_obj = $class->{_source}->{filer}->_get_underlying_filer_obj();
        
        my $src_conn = $src_obj->conn();
        
        Aggregate::Copy::aggrcopy_throttle($src_conn, $op_num, $value);
        
        return (rc => C_RC_SUCCESS);
}


## @name ::attach_to_copy_volume
## @summary Creates a Compression::Compress object corresponding to the copy of the source volume.
## 
## @description
## Creates a Compression::Compress object corresponding to the copy of the source volume.
##
## @return Compression::Compress object.
##
## @examples
## my $src_copy = $obj->attach_to_copy_volume();
##
## @note
## In the case when the source and destination aggregates are on the same filer,
## the SW will rename the copy volume to <src_vol>(1).
## Because of this naming convention, mounting the copy volume fails with syntax error.
## To overcome this, the library will first rename this copy to copy_1_of_<src_vol>,
## then attach to it.
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub attach_to_copy_volume
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        # None
                } );
        
        my $src = $class->{_source};
        my $dst = $class->{_destination};
        
        # First detect the copy name:
        # If the copy is on different filer, this will be the same as the source volume
        # If on same filer, it will be <source_vol_name>(1)
        my $src_vol_name = $src->{vol_name};
        my @dest_vols = $dst->{aggregate}->list_contained_volumes();
        my $copy_vol_name;
        foreach my $vol (@dest_vols)
        {
                if ($vol =~ /^$src_vol_name\(*1*\)*$/)
                {
                        $copy_vol_name = $vol;
                        last;
                }
        }
        
        if (!defined($copy_vol_name))
        {
                logresult('fatal', "Unable to detect the copy of volume $src_vol_name in the destination aggregate:\n" . Dumper(@dest_vols));
        }
        
        # If the copy volume is of form <source_vol_name>(1),
        # using it as is will not be possible, as mounting will fail with sintax error
        # Remane it to not contain the paranthesis
        my $rename_copy_vol = $copy_vol_name;
        if ($copy_vol_name =~ /^$src_vol_name\(1\)$/)
        {
                $rename_copy_vol = 'copy_1_of_' . $src_vol_name;
                logcomment("The copy of the volume $src_vol_name is named \"$copy_vol_name\".\n" .
                           "To be able to mount it through the library, this will be renamed to \"$rename_copy_vol\"");
        }
        
        return Compression::Compress->setup(filer => {name => $class->{_dst_host_name},},
                                            aggregate => {name => $class->{_dst_aggr_name},
                                                          indirect_type => $src->{aggregate}->{indirect_type},
                                                          snaplock => $src->{aggregate}->{_snaplock}},
                                            volume 	=> {name => $rename_copy_vol,
                                                            compress_enabled => C_FLAG_KEEP_VALUE_AS_IS,
                                                            asis => FALSE,},
                                            attach_to_volume => $copy_vol_name,
                                            debug => $class->{_debug},
                                            disable_cleanup => $class->{_disable_cleanup},
                                            _attach_dataset_info => $src,);
}


## @name ::restore_destination
## @summary Restores the state of the destination aggregate.
## 
## @description
## Restores the state of the destination aggregate (brings the aggregate online).
##
## @return Hash with rc.
##
## @examples
## my $src_copy = $obj->restore_destination();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub restore_destination
{
        my $class = shift;
        
        logcomment("Restoring destination aggregate state:\n" .
                   "Aggregate: " . $class->{_dst_aggr_name} . " (filer: " . $class->{_dst_host_name} . ")");
        
        my $dst = $class->{_destination};
        
        logcomment("AggrCopy: Bringing the destination aggregate online...");
        $dst->{aggregate}->set_online();
        
        # Informational...
        my @vols = $class->{_destination}->{aggregate}->list_contained_volumes();
        logcomment("The destination aggregate contains the volumes:\n" . Dumper(@vols));
        
        $class->_reset_dest_contents();
        
        return (rc => C_RC_SUCCESS);
}



########################
# Internal methods
########################

sub _reset_dest_contents
{
        my $class = shift;
        
        my $dst = $class->{_destination};
        
        # At this point, the volume that was initially created on the destination does not exist anymore
        # Disable its cleanup...
        $dst->_disable_vol_cleanup();       
}


# Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
        
        # TODO: we should not use the NOTE internal flag...
        # use it for now, until DPG team offers a spare method
        $class->{_copy}->{_spare_me} = 1;
	
	if (($class->{_disable_cleanup} == FALSE) &&
            ($class->{_creator_pid} == $$))
        {
                logcomment("Executing Compression\:\:AggrCopy cleanup for:\n" .
                           "Source: " . $class->{_src_aggr_name} . "\n" .
                           "Destination: " . $class->{_dst_aggr_name});

                eval
                {
                        # We need to do it this way because:
                        # NOTE seems to break the garbage collection basisc (cleanup is done with an END block)
                        $class->{_copy}->destroy();
                        undef($class->{_copy});
                };
                
                if( $@ )
                {
                        logcomment( "AggrCopy: cleanup error: " . $@ );
                }
        }
	else
	{
                logcomment( "AggrCopy: cleanup:  nothing done: script PID: " . $$) if ($class->{_debug});
	}
        
        if (defined($execution_error))
        {
                _c_interpret_logresult($execution_error);
        }
}


sub testme
{
        package main;
        
	require "drg/Compression/setup_core.pl";
        use Compression::AggrCopy;
        
        use vars qw($VOLUME $FILER_DST $FILER $AGGR_SRC $AGGR_DST);
        param("FILER", "-mesg", "Filer name");
        param("VOLUME", "-default", 'basic_script_vol');
        param("FILER_DST", "-default", $FILER);
        param("AGGR_SRC", "-default", 'compr');
        param("AGGR_DST", "-default", 'compr_copy');
        
        my %res;
        
        my $compress1 = Compression::Compress->setup(filer => {name => $FILER,},
                                                     aggregate => {name => $AGGR_SRC,},
                                                     volume 	=> {name => 'vol1',
                                                                    compress_enabled => FALSE,},);
        
        my $compress2 = Compression::Compress->setup(filer => {name => $FILER_DST,},
                                                     aggregate => {name => $AGGR_DST,},
                                                     volume 	=> {name => 'vol2',
                                                                    compress_enabled => FALSE,},);
        
        logcomment("Running self-test for module Compression\:\:AggrCopy");
        
        $compress1->{dataset}->create(files =>[{ name => 'file1', size => MB,},],);
        $compress2->{dataset}->create(files =>[{ name => 'file2', size => MB,},],);
        $compress1->wait_for_cp();
        
        $compress1->{dataset}->ls();
        $compress2->{dataset}->ls();
        
        my $aggr_copy = Compression::AggrCopy->new(source => $compress1,
                                                   destination => $compress2);
        
        $aggr_copy->verify_status(status => C_AGGR_COPY_STATUS_IDLE);
        
        $aggr_copy->start(background => TRUE);
        $aggr_copy->throttle(value => 9);
        
        $aggr_copy->verify_status();
        $aggr_copy->wait_for_status(status => C_AGGR_COPY_STATUS_TRANSFERRING,);
        $aggr_copy->abort();
        $aggr_copy->verify_status(status => C_AGGR_COPY_STATUS_IDLE);
        
        $aggr_copy->start(background => TRUE);
        $aggr_copy->wait_for_status(status => C_AGGR_COPY_STATUS_TRANSFERRING,);
        $aggr_copy->abort();
        $aggr_copy->start(background => TRUE);
        $aggr_copy->wait_for_completion();
        $compress1->wait_for_cp();
        
        $aggr_copy->start();
        $aggr_copy->verify_status(status => C_AGGR_COPY_STATUS_IDLE);
        $aggr_copy->restore_destination();
        
        my $copy = $aggr_copy->attach_to_copy_volume();
        $copy->{dataset}->ls();
        $copy->{dataset}->modify(files =>[{ name => 'file1', offset => 0, len => C_WAFL_BLOCK_SIZE},],);
        $copy->verify_savings(expected_compressed => 0);

logresult("PASS", "Self-test for Compression\:\:AggrCopy passed");
}


1;
