#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Aggregate.pm#1 $
#

## @summary Compression Aggregate utilities.
##
## @description
## Compression Aggregate utilities.
##    
## @synopsis
##
## use Compression::Aggregate;
## use vars qw($FILER $AGGREGATE );
## 
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $filer = NOTE::Filer->connect( host => $FILER );
## my $obj = Compression::Aggregate->new(
## 	filer => $filer,
##      aggregate => $AGGREGATE,
##      debug => FALSE,
## );
##
## $obj->volume_is_contained( volume => $vol_name, expected => C_RC_FAILURE );
## $obj->offline_contained_volumes();
## $obj->online_contained_volumes();
## $obj->set_offline(force => TRUE, all => TRUE );
## $obj->set_offline();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)


package Compression::Aggregate;

use strict;
use warnings;

use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Tharn;
use Compression::Constants qw(:all);
use Compression::7Mode::Aggregate;
use Compression::CMode::Aggregate;
use Compression::Utils qw(detect_indirect_type_from_status
			  _c_interpret_logresult); 
use Data::Dumper;

BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}



## @name ::new
##
## @summary Compression::Aggregate Constructor.
##
## @description
## Creates an object representing an aggregate on the specified filer.
##
## @return Object of type Compression::Aggregate
##
## @arg conn Required: NOTE::Filer object corresponding to the hosting filer.
##
## @arg aggregate Optional: Aggregate information.
## It can be either a scalar (representing the name), or a hash reference.
## If not provided, uses the default values defined in Compression::Constants.
## 
## @arg ui Optional: Compression::UI object corresponding to the hosting filer.
## If not provided, it will be auto-generated.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
##
## my $filer = NOTE::Filer->connect( host => $FILER );
## my $obj = Compression::Aggregate->new(
## 	filer => $filer,
##      aggregate => $AGGREGATE,
##      debug => TRUE,
## );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub new
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			conn 	=> {type => OBJECT,},
                        
                        aggregate => {type => SCALAR|HASHREF|UNDEF,
                                      optional => 1,},
			
			vserver => {type => SCALAR|UNDEF,
                                    optional => 1},
                        
                        boot_mode => {type => SCALAR,
                                      optional => 1,
                                      default => C_ONTAP_BOOT_MODE},
                        
                        ui      => {type => OBJECT,
                                    isa  => "Compression::UI",
                                    optional => 1,},
                        
                        debug 	=> {type => SCALAR,
                                    optional => 1,
                                    default => FALSE,},
                        
                        attach => {type => SCALAR,
                                   optional => 1,
                                   default => FALSE,},
			
			disable_cleanup => {type => BOOLEAN,
                                    optional => 1,
                                    default => FALSE,
                                    },
			node => {type => SCALAR,
                                   optional => 1,},
		} );

	my $self = {};
        $self->{_debug} = $args{debug};
	$self->{_boot_mode} = $args{boot_mode};
        
        delete($args{boot_mode});
        if ($self->{_boot_mode} eq C_ONTAP_CMODE)
	{
                $self->{_c_aggr} = Compression::CMode::Aggregate->new(%args);
	}
        else
        {
                delete($args{vserver});
                $self->{_c_aggr} = Compression::7Mode::Aggregate->new(%args);
        }
        
        $self->{aggr_name} = $self->{_c_aggr}->{aggr_name};
        $self->{_snaplock} = $self->{_c_aggr}->{_snaplock};
        $self->{indirect_type} = $self->{_c_aggr}->{indirect_type};
        $self->{_filer_name} = $self->{_c_aggr}->{_filer_name};
        
        bless $self, $class;
        
	return $self;
}


## @name ::get_state
## @summary Gets the state of the aggregate.
## 
## @description
## Gets the state of the aggregate.
##
## @return state.
## 
## @examples
## my $aggr_state = $obj->get_state();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_state
{
	my $class = shift;
        
        return $class->{_c_aggr}->get_state(@_);
}


## @name ::get_indirect_type
## @summary Gets the indirect type of the aggregate (32/64 -bit).
## 
## @description
## Gets the indirect type of the aggregate (32/64 -bit).
##
## @return Hash with rc, indirect_type, and output (status hashref).
## 
## @examples
## $obj->get_indirect_type();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_indirect_type
{
	my $class = shift;
        
        return $class->{_c_aggr}->get_indirect_type(@_);
}


## @name ::volume_is_contained
## @summary Verifies that the specified volume exists in this aggregate.
## 
## @description
## Verifies that the specified volume exists in this aggregate.
##
## @return Hash with rc and output (list of existent volumes).
## 
## @arg volume Required: Name of the volume.
## 
## @arg expected Optional: Expected result (C_RC_SUCCESS/C_RC_FAILURE).
## If not provided, assumes that the volume should exist.
## C_RC_FAILURE_NON_FATAL can also be passed in as 'expected', in which case,
## the script will not be terminated upon any result.
## Instead, the result (C_RC_SUCCESS/C_RC_FAILURE) will be passed to the script for further processing.
##
## @examples
## $obj->volume_is_contained( volume => $vol, expected => C_RC_FAILURE );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub volume_is_contained
{
	my $class = shift;
        
        return $class->{_c_aggr}->volume_is_contained(@_);
}


## @name ::offline_contained_volumes
## @summary Offlines all volumes contained by this aggregate.
## 
## @description
## Offlines all volumes contained by this aggregate.
##
## @return Hash with rc and output (string containing ", "-delimited list of volumes that were offlined).
## 
## @arg skip_root_volume Optional: Boolean flag to decide if the root volume should be skipped.
## Defaults to TRUE.
##
## @examples
## $obj->offline_contained_volumes();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub offline_contained_volumes
{
	my $class = shift;
        
        return $class->{_c_aggr}->offline_contained_volumes(@_);
}


## @name ::online_contained_volumes
## @summary Onlines all volumes contained by this aggregate.
## 
## @description
## Onlines all volumes contained by this aggregate.
##
## @return Hash with rc and output (string containing ", "-delimited list of volumes that were onlined).
## 
## @arg skip_root_volume Optional: Boolean flag to decide if the root volume should be skipped.
## Defaults to TRUE.
##
## @examples
## $obj->online_contained_volumes();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub online_contained_volumes
{
	my $class = shift;
        
        return $class->{_c_aggr}->online_contained_volumes(@_);
}


## @name ::set_offline
## @summary Offlines this aggregate.
## 
## @description
## Offlines this aggregate. 
##
## @arg force Optional: Boolean flag to force offlining the aggregate.
##
## @arg cifsdelaytime Optional: Number of seconds to delay before taking the aggregate offline,
## during which time CIFS uses are warned of the pending loss of service.
##
## @arg all Optional: Boolean flag to offline aggregate with all its volume
##
## @return Hash with rc and output.
##
## @examples
## $obj->set_offline();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub set_offline
{
        my $class = shift;
        
        return $class->{_c_aggr}->set_offline(@_);
}


## @name ::set_online
## @summary Onlines this aggregate.
## 
## @description
## Onlines this aggregate. 
##
## @arg force Optional: Boolean flag to force onlining the aggregate.
##
## @return Hash with rc and output.
##
## @examples
## $obj->set_online();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub set_online
{
        my $class = shift;
        
        return $class->{_c_aggr}->set_online(@_);
}


## @name ::get_avail
## @summary Get available space
## 
## @description
## Returns available space in KB.
##
## @return available space in KB
##
## @examples
## my $avail = $obj->get_avail();
## 
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_avail
{
	my $class = shift;
        
        return $class->{_c_aggr}->get_avail(@_);
}


## @name ::list_snapshots
## @summary Lists snapshots that exist on this aggregate.
## 
## @description
## Lists snapshots that exist on this aggregate. 
##
## @return Hash with rc and output (reference to array with names of existent snapshots).
##  
## @examples
## $obj->list_snapshots();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
## @status private (in-progress)
##

sub list_snapshots
{
        my $class = shift;
        
        return $class->{_c_aggr}->list_snapshots(@_);
}

## @name ::delete_snapshot
## @summary Deletes a snapshot.
## 
## @description
## Deletes a named snapshot.
##
## @return Hash with rc and output (empty).
##
## @arg name Required: Name of the snapshot.
##  
## @examples
## $obj->delete_snapshot(name => $my_snapshot);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
## @status private (in-progress)
##

sub delete_snapshot
{
        my $class = shift;
        
        return $class->{_c_aggr}->delete_snapshot(@_);
}


### @name ::start_64_bit_upgrade
## @summary Starts 64-bit upgrade for this aggregate.
## 
## @description
## Starts 64-bit upgrade for this aggregate.
##
## @return Hash with rc and output.
##
## @arg mode Optional: Upgrade mode: check|normal|grow-reserved|grow-all.
## Default: none.
##  
## @examples
## $obj->start_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub start_64_bit_upgrade
{
        my $class = shift;
        
        return $class->{_c_aggr}->start_64_bit_upgrade(@_);
}


## @name ::wait_for_64_bit_upgrade
## @summary Waits for a previously started 64-bit upgrade to complete.
## 
## @description
## Waits for a previously started 64-bit upgrade to complete.
##
## @return Hash with rc.
##
## @arg timeout Optional: Timeout value, in seconds.
##
## @arg delay Optional: Time interval, in seconds, in between consecutive status checks.
##  
## @examples
## $obj->wait_for_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub wait_for_64_bit_upgrade
{
        my $class = shift;
	
	return $class->{_c_aggr}->wait_for_64_bit_upgrade(@_);
}


## @name ::stop_scan_64_bit_upgrade
## @summary Stops a previously started 64-bit upgrade.
## 
## @description
## Stops a previously started 64-bit upgrade .
##
## @return Hash with rc.
##  
## @examples
## $obj->stop_scan_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub stop_scan_64_bit_upgrade
{
        my $class = shift;
	
	return $class->{_c_aggr}->stop_scan_64_bit_upgrade(@_);
}


## @name ::start_scan_64_bit_upgrade
## @summary Restarts a previously stopped 64-bit upgrade.
## 
## @description
## Restarts a previously stopped 64-bit upgrade.
##
## @return Hash with rc.
## @examples
## $obj->start_scan_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub start_scan_64_bit_upgrade
{
        my $class = shift;
	
	return $class->{_c_aggr}->start_scan_64_bit_upgrade(@_);
}


## @name ::status_64_bit_upgrade
## @summary Gets the status of the 64-bit upgrade for this aggregate.
## 
## @description
## Gets the status of the 64-bit upgrade for this aggregate.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_LA_IPU_STOPPED/C_RC_LA_IPU_NOT_IN_PROGRESS.
## (defined in Compression::Constants).
##
## @return Hash with rc and output. 
##  
## @examples
## my %res = $obj->status_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub status_64_bit_upgrade
{
        my $class = shift;
	
	return $class->{_c_aggr}->status_64_bit_upgrade(@_);
}


## @name ::is_block_type_32bit
## @summary Checks if the aggregate is 32-bit.
## 
## @description
## Checks if the aggregate is 32-bit.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc.
##  
## @examples
## $obj->is_block_type_32bit(expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_block_type_32bit
{
        my $class = shift;
	
	return $class->{_c_aggr}->is_block_type_32bit(@_);
}


## @name ::is_block_type_64bit
## @summary Checks if the aggregate is 64-bit.
## 
## @description
## Checks if the aggregate is 64-bit.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc.
##  
## @examples
## $obj->is_block_type_64bit(expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_block_type_64bit
{
        my $class = shift;
	
	return $class->{_c_aggr}->is_block_type_64bit(@_);
}


## @name ::list_contained_volumes
## @summary List volumes contained by this aggregate.
## 
## @description
## List volumes contained by this aggregate.
##
## @return Array of volume names.
##  
## @examples
## my @vols = $obj->list_contained_volumes();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub list_contained_volumes
{
        my $class = shift;
	
	return $class->{_c_aggr}->list_contained_volumes(@_);
}


## @name ::start_block_check
## @summary Runs block_check on this aggregate.
## 
## @description
## Runs block_check on this aggregate.
##
## @arg wait Optional: Flag to control if the method should wait for the scanner to complete.
## Default: TRUE.
## 
## @arg disks: Optional: Arrayref of disk names to run block check on.
## If none of the 'disks' and 'pvbn' arguments are passed in, then block check will be run on all the data disks of this aggregate.
## 
## @arg pvbn: Optional: Arrayref of pvbns to run block check against.
## If none of the 'disks' and 'pvbn' arguments are passed in, then block check will be run on all the data disks of this aggregate.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Default: C_RC_SUCCESS.
## 
## @arg expected_pvbn_file_path Optional: Expected file path to be listed by block check after completion.
## Only relevant if 'pvbn' is passed in.
##
## @return Hash with rc and output.
##  
## @examples
## $obj->start_block_check();
## $obj->start_block_check(pvbn => [0 .. 7],
##                         expected_pvbn_file_path => 'vol/my_vol/my_file');
##
## @note
## Extra information:
## http://wikid.netapp.com/w/WAFL_Availability/WAFL_recovery_avoidance/REP/REP_tool_info#aggr_block_check
## http://wikid.netapp.com/w/WAFL_Availability/WAFL_recovery_avoidance/REP/IC.2/Functional_Specification#Tools
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub start_block_check
{
	my $class = shift;
	return $class->{_c_aggr}->start_block_check(@_);
}


## @name ::wait_for_block_check
## @summary Waits for already-started block_check scan to complete.
## 
## @description
## Waits for already-started block_check scan to complete.
##
## @arg scan_id Optional: Id of the block_check scan to wait for.
## If not provided, this will be the last-started scan.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Default: C_RC_SUCCESS.
##  
## @examples
## $obj->wait_for_block_check();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub wait_for_block_check
{
	my $class = shift;
	return $class->{_c_aggr}->wait_for_block_check(@_);
}


## @name ::get_data_disk_names
## @summary Obtain the names of the data disks of this aggregate.
## 
## @description
## Obtain the names of the data disks of this aggregate.
##
## @return Array of disk names.
##  
## @examples
## my @data_disks = $obj->get_data_disk_names();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_data_disk_names
{
	my $class = shift;
	return $class->{_c_aggr}->get_data_disk_names(@_);
}



################################
# Internal subroutines
################################


sub _get_underlying_aggr_obj
{
        my $class = shift;
        return $class->{_c_aggr}->{_aggr};
}


sub _refresh
{
	my $class = shift;
	
	$class->{_c_aggr}->_refresh(@_);
	$class->{aggr_name} = $class->{_c_aggr}->{aggr_name};
	$class->{_hosting_filer_has_changed} = $class->{_c_aggr}->{_hosting_filer_has_changed};
	$class->{_filer_name} = $class->{_c_aggr}->{_filer_name};
}


sub testme
{
	package main;
	
	require "drg/Compression/setup_core.pl"; 
	
	use vars qw($FILER $DEBUG $AGGREGATE $OFFLINE $AGGR_BLOCK_CHECK);
	param("DEBUG", "-default", FALSE);
	param("AGGREGATE", "-default", 'compr');
	param("OFFLINE", "-default", FALSE);
	param("AGGR_BLOCK_CHECK", "-default", TRUE);
	
	
	my $compress = Compression::Compress->setup(filer => {name => $FILER,},
						    aggregate => {name => $AGGREGATE,},
						    volume 	=> {name => $VOLUME,
								    size => '10g',},
						    debug => $DEBUG,);
	
	my %res;
	my $scalar_res;
	my $filename = 'file1';
	
	%res = $compress->{aggregate}->list_snapshots();
	logcomment("list_snapshots returned:\n" . Dumper(%res));
	
	%res = $compress->{aggregate}->get_indirect_type();
	logcomment("get_indirect_type returned:\n" . Dumper(%res));
	
	%res = $compress->{aggregate}->volume_is_contained(volume => 'bogus_vol',
							   expected => C_RC_FAILURE);
	logcomment("volume_is_contained returned:\n" . Dumper(%res));
	
	%res = $compress->{aggregate}->offline_contained_volumes();
	logcomment("offline_contained_volumes returned:\n" . Dumper(%res));
	
	%res = $compress->{aggregate}->online_contained_volumes();
	logcomment("online_contained_volumes returned:\n" . Dumper(%res));
	
	if ($OFFLINE == TRUE)
	{
		%res = $compress->{aggregate}->set_offline(force => TRUE, all => TRUE );
		logcomment("set_offline returned:\n" . Dumper(%res));
		
		%res = $compress->{aggregate}->set_online();
		logcomment("set_online returned:\n" . Dumper(%res));
	}
	
	$scalar_res = $compress->{aggregate}->get_avail();
	logcomment("get_avail returned: $scalar_res");
	
	$compress->{dataset}->create(files =>[{ name => $filename, size => MB,},],);
	$compress->wait_for_cp();
	
	my $file_filer_path = $compress->{dataset}->get_filer_path(filename => $filename);
	my @pvbns = $compress->{dataset}->get_pvbns(filename => $filename, fbns => [1],);
	logcomment("get_pvbns returned:\n" . Dumper(@pvbns));
	
	if ($AGGR_BLOCK_CHECK == TRUE)
	{
		$compress->wait_for_cp();
		%res = $compress->{aggregate}->start_block_check(pvbn => \@pvbns,
								 expected_pvbn_file_path => $file_filer_path,);
		logcomment("start_block_check (pvbns, no corruption) returned:\n" . Dumper(%res));
		
		$compress->{aggregate}->start_block_check();
		logcomment("start_block_check (defaults, no corruption) returned:\n" . Dumper(%res));
		
		$compress->{aggregate}->start_block_check(wait => FALSE);
		%res = $compress->{aggregate}->wait_for_block_check();
		logcomment("start_block_check (defaults, no wait, no corruption) returned:\n" . Dumper(%res));
		
		$compress->{filer}->set_flag(code => C_FLAG_WAFL_IRON_PANIC_ON_CORRUPTION,
					     value => 0);
		$compress->{filer}->set_flag(code => C_FLAG_WAFL_DECOMPRESS_FAIL_TESTING,
					     value => 1);
		$compress->{filer}->set_flag(code => C_FLAG_WAFL_INCONSISTENT_PANIC_TESTING,
					     value => 1);
		$compress->{volume}->wreck(code => C_CORRUPT_CHECKSUM,
					   filename => $filename,);
		$compress->{volume}->bring_online();
		$compress->{volume}->start_wafliron();
		$compress->wait_for_cp();
		%res = $compress->{aggregate}->start_block_check(expected => C_RC_LOST_BLOCKS_FOUND,);
		logcomment("start_block_check (corruption) returned:\n" . Dumper(%res));
	}
	
	logresult('pass', "Compression\:\:Aggregate\:\:testme passed");
}




1;
