#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/CMode/Aggregate.pm#1 $
#

## @summary Compression utilities for aggegate.
##
## @description
## Compression utilities for aggegate.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 


package Compression::CMode::Aggregate;

use strict;
use warnings;

use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Tharn;
use Compression::Constants qw(:all);
use Compression::Utils qw(detect_indirect_type_from_status
			  _c_interpret_logresult);
use NACL::APISet;
use FS::Aggr_block_check;
use Data::Dumper;
use NACL::STask::Aggregate;
use NACL::CS::Vserver;

BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}



## @name ::new
##
## @summary Compression::CMode::Aggregate Constructor.
##
## @description
## Compression::CMode::Aggregate Constructor.
##
## @return Object of type Compression::CMode::Aggregate
##
## @arg conn Required: NOTE::GX::Node object corresponding to the hosting filer.
##
## @arg aggregate Optional: Aggregate information.
## It can be either a scalar (representing the name), or a hash reference.
## If not provided, uses the default values defined in Compression::Constants.
##
## @arg vserver Required: vserver name.
## 
## @arg ui Optional: Compression::UI object corresponding to the hosting filer.
## If not provided, it will be auto-generated.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
##
## my $filer = Compression::Filer->new( filer => $FILER );
## my $obj = Compression::Aggregate->new(
## 	filer => $filer,
##      aggregate => $AGGREGATE,
##      debug => TRUE,
## );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub new
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			conn 	=> {type => OBJECT,
                                    isa  => "NOTE::GX::Node",},
                        
                        aggregate => {type => SCALAR|HASHREF|UNDEF,
                                      optional => 1,},
                        
                        vserver => {type => SCALAR},
                        
                        ui => {type => OBJECT,
                               isa  => "Compression::UI",
                               optional => 1,},
                        
                        debug => {type => SCALAR,
                                  optional => 1,
                                  default => FALSE,},
                        
                        attach => {type => SCALAR,
                                   optional => 1,
                                   default => FALSE,},
			
			disable_cleanup => {type => BOOLEAN,
					    optional => 1,
					    default => FALSE,
					    },
			node => {type => SCALAR,
                                   optional => 1,},
		} );

	my $self = {};
        $self->{_debug} = $args{debug};
	$self->{_creator_pid} = $$;
	$self->{_disable_cleanup} = $args{disable_cleanup};
	$self->{_hosting_filer_has_changed} = FALSE;
        
        logcomment("Aggregate: new: aggregate " . Dumper($args{aggregate})) if ($self->{_debug});
        
	# Construct the aggr hash as expected by the NOTE::Aggregate constructor
	my %aggr = ();
        
        my $aggregate_args = $args{aggregate};
        if ((defined($aggregate_args) && (ref($aggregate_args) eq "HASH")))
        {
                # Second-layer hash validation for aggregate args
                logcomment("Aggregate: new: aggregate attributes passed in as HASHREF") if ($self->{_debug});
                %aggr = validate_with
                        ( params => $aggregate_args,
                          spec   => {name => { type => SCALAR, default => C_DFLT_AGGREGATE },
                                     recycle => { type => SCALAR, default => C_DFLT_AGGR_PURGE },
                                     share => { type => SCALAR, default => C_DFLT_AGGR_PURGE },
                                     disks => { type => SCALAR, default => C_DFLT_AGGR_DISKS },
                                     snaplock => { type => SCALAR, default => C_ENABLE_SNAPLOCK },
                                     indirect_type => { type => SCALAR, default => $main::AGGR_IND_TYPE },
                                     },
                          allow_extra => 0,
                        );
        }
        else
        {
                logcomment("Aggregate: new: aggregate attribute passed in as SCALAR or undefined") if ($self->{_debug});
                $aggr{name} = (defined($aggregate_args)) ? $aggregate_args : C_DFLT_AGGREGATE;
                $aggr{recycle} = C_DFLT_AGGR_PURGE;
                $aggr{share} = C_DFLT_AGGR_PURGE;
                $aggr{disks} = C_DFLT_AGGR_DISKS;
                $aggr{snaplock} = C_ENABLE_SNAPLOCK;
                $aggr{indirect_type} = $main::AGGR_IND_TYPE;
        }
        
        # Pass the NOTE obj to reuse the connection
        my $conn = $args{conn};
	$aggr{Node} = $conn;
        $aggr{Vserver} = $args{vserver};
        
        logcomment("Aggregate: new: aggregate information hash to be passed to GNOTE constructor:\n" . Dumper(%aggr)) if ($self->{_debug});
        
	# Create an aggregate if required, otherwise re-use it
        my $attach = $args{attach};
        if ($attach == TRUE)
        {
                $self->{_aggr} = NOTE::GX::Aggregate->attach(name => $aggr{name},
                                                             Node => $conn);
        }
        else
        {
				# The below is the Fix for Burt 649016.
				# Replacing NOTE::GX::Aggregate->create() with NACL code.
				# Other details are mentioned in the burt itself.
				my $NACL_conn = NACL::C::Node->new(name => $args{node});
				NACL::STask::Aggregate->create(
									aggregate			=> $aggr{name},
									command_interface	=> $NACL_conn,
									diskcount			=> $aggr{disks},
									"block-type"		=> $aggr{indirect_type}."-bit",
									if_exists			=> 'reuse',
							);

				# Later we need to have NOTE object only. that's why the
				# following wrapper.
				$self->{_aggr} = NOTE::GX::Aggregate->attach(
									name => $aggr{name},
									Node => $conn,
								);
				# If $aggr is not present in the aggr-list of the vserver,
				# then add it to the aggr-list.
				my $detail = NACL::CS::Vserver->fetch(
									command_interface	=> $NACL_conn,
									requested_fields  	=> [qw(aggr-list)],
									filter				=> 
											{ vserver	=> $aggr{Vserver}},
									);
				my @aggrlist = $detail->aggr_list();
				#If the array list is empty, $detail->aggr_list() returns a "-". Need to ignore it
				if (grep($_ eq "-" , @aggrlist ))
				{
					logcomment("All aggregates in the vserver");
				}
				elsif (! grep($_ eq  $aggr{name}, @aggrlist )) {
					push(@aggrlist, $aggr{name} );
					NACL::C::Vserver->modify (
							command_interface   => $NACL_conn,
							vserver				=> $aggr{Vserver},
							'aggr-list'			=> \@aggrlist,
					);
				}
				# End of fix
								
        }
	$self->{aggr_name} = $aggr{name};
        $self->{indirect_type} = $aggr{indirect_type};
	$self->{_recycle} = $aggr{recycle};
	if ($self->{_snaplock})
	{
		logresult('NA', "NOT APPLICABLE: Snaplock is not supported on C-mode");
	}
	
	$self->{_snaplock} = $aggr{snaplock};
	$self->{_filer_name} = $conn->name();
	$self->{_block_check_scans} = {};
	$self->{_latest_block_check_id} = '';
	$self->{_conn} = $conn;
        
        # UI
        if ( defined($args{ui}) )
        {
                $self->{_ui} = $args{ui};
        }
        else
        {
                # Auto-detect version of the hosting filer. Need this for UI.
                logcomment("Volume: UI object not provided. Auto-generating...") if ($self->{_debug});
                my $version = Compression::Version->new( conn => $conn, debug => $self->{_debug} );
                $self->{_ui} = Compression::UI->new( version => $version, debug => $args{debug}, boot_mode => C_ONTAP_CMODE );
        }

        bless $self, $class;
        
        #########################
	# Do extra setup steps
	#########################
        
        # Need to ensure here that the indirect type is as it should be...
        # NOTE::Aggregate does not ensure this for recycled aggregates
        my %current_type = $self->get_indirect_type();
        if ($current_type{indirect_type} != $self->{indirect_type})
        {
                logresult("NA", "NOT APPLICABLE: Aggregate " . $self->{aggr_name} . " is required to be " . $self->{indirect_type} . "-bit, " .
                                   "but it is a " . $current_type{indirect_type} . "-bit aggregate.");
        }
        
        logcomment("Using " . $self->{aggr_name} . ", a " . $self->{indirect_type} . "-bit aggregate.");
        
        # Check status. If aggr is offline, attempt to bring online (it may happen if aggr was reused)
        my $aggr_status = $self->{_aggr}->status();
        my $aggr_state = $aggr_status->{'state'};
        logcomment("State of aggregate " . $self->{aggr_name} . " is: $aggr_state") if ($self->{_debug});

        if ($aggr_state ne C_AGGR_STATE_ONLINE)
        {
                logcomment("Attempting to online aggregate " . $self->{aggr_name} . " ...") if ($self->{_debug});
                $self->set_online();
        }

	return $self;
}


## @name ::get_state
## @summary Gets the state of the aggregate.
## 
## @description
## Gets the state of the aggregate.
##
## @return state.
## 
## @examples
## my $aggr_state = $obj->get_state();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_state
{
	my $class = shift;
        my %args = validate(@_,
        {
                # None.		
	} );
        
        my $aggr_status = $class->{_aggr}->status();
        my $state = $aggr_status->{state};
        
        if (!defined($state))
        {
                logresult('fatal', "Unable to detect the aggregate state:\n" . Dumper($aggr_status));
        }
        
        return $state;
}


## @name ::get_indirect_type
## @summary Gets the indirect type of the aggregate (32/64 -bit).
## 
## @description
## Gets the indirect type of the aggregate (32/64 -bit).
##
## @return Hash with rc, indirect_type, and output (status hashref).
## 
## @examples
## $obj->get_indirect_type();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_indirect_type
{
	my $class = shift;
        my %args = validate(@_,
	{
		# None.		
	} );
        
        my $aggr_status = $class->{_aggr}->status();
        my $block_type = $aggr_status->{'block-type'};
        
        logcomment("Aggregate: Current status:\n" . Dumper($aggr_status) .
                   "Attempting to detect indirect type from: status: $block_type") if $class->{_debug};
        
        my $indirect_type = detect_indirect_type_from_status(status_string => $block_type);
        
        return (rc => C_RC_SUCCESS, indirect_type => $indirect_type, output => $aggr_status);
}


## @name ::volume_is_contained
## @summary Verifies that the specified volume exists in this aggregate.
## 
## @description
## Verifies that the specified volume exists in this aggregate.
##
## @return Hash with rc and output (list of existent volumes).
## 
## @arg volume Required: Name of the volume.
## 
## @arg expected Optional: Expected result (C_RC_SUCCESS/C_RC_FAILURE).
## If not provided, assumes that the volume should exist.
## C_RC_FAILURE_NON_FATAL can also be passed in as 'expected', in which case,
## the script will not be terminated upon any result.
## Instead, the result (C_RC_SUCCESS/C_RC_FAILURE) will be passed to the script for further processing.
##
## @examples
## $obj->volume_is_contained( volume => $vol, expected => C_RC_FAILURE );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub volume_is_contained
{
	my $class = shift;
        my %args = validate(@_,
	{
		volume   => { type => SCALAR, },
                expected  => { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },		
	} );
        
        my @vol_list = $class->{_aggr}->volumes();
        my $vol_to_find = $args{volume};
        my $rc = C_RC_FAILURE;
        
        logcomment("Aggregate: Looking for volume \"$vol_to_find\" in the list:\n" . Dumper(@vol_list)) if $class->{_debug};
        if (grep(/^$vol_to_find$/, @vol_list))
        {
                $rc = C_RC_SUCCESS;
        }
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Verify that the volume \"$vol_to_find\" is contained by aggregate " . $class->{aggr_name},
                                                    output      => \@vol_list,);
}


## @name ::offline_contained_volumes
## @summary Offlines all volumes contained by this aggregate.
## 
## @description
## Offlines all volumes contained by this aggregate.
##
## @return Hash with rc and output (string containing ", "-delimited list of volumes that were offlined).
## 
## @arg skip_root_volume Optional: Boolean flag to decide if the root volume should be skipped.
## Defaults to TRUE.
## 
## @arg skip_vsroot_volumes Optional: Boolean flag to decide if the vsroot volumes should be skipped.
## Defaults to TRUE.
##
## @examples
## $obj->offline_contained_volumes();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub offline_contained_volumes
{
	my $class = shift;
        my %args = validate(@_,
	{
		skip_root_volume => { type => BOOLEAN, optional => 1, default => TRUE, },
                skip_vsroot_volumes => { type => BOOLEAN, optional => 1, default => TRUE, },
	} );
        
        my $skip_root_flag = $args{skip_root_volume};
        my $skip_vsroot_flag = $args{skip_vsroot_volumes};
        my @volumes = $class->{_aggr}->Volumes();
        my $output = '';
        
        # See if this is mroot aggr (if it is, we will need to check for the mroot volume)
        my $mroot_vol_name;
        my $mroot_aggr = $class->{_aggr}->server->storage_aggregate_show('has-mroot' => 'true',
                                                                         'nodes' => $class->{_aggr}->Node->name);
        my $root_aggr_name = $mroot_aggr->[0]->{aggregate};
        logcomment("Mroot aggregate is $root_aggr_name");
        
        my $detected_mroot_vol = 0;
        foreach my $volume (@volumes)
        {
                my $volume_name = $volume->name();
                logcomment("Aggregate: offline_contained_volumes: Processing volume $volume_name") if $class->{_debug};
                
                # Skip the root volume, if required
                if (($skip_root_flag == TRUE) && ($root_aggr_name eq $class->{aggr_name}) && ($detected_mroot_vol == 0))
                {
                        logcomment("Aggregate: offline_contained_volumes: Working on the mroot aggregate.\n" .
                                   "Looking for the mroot volume name") if ($class->{_debug});
                        my $cmd_code = C_CMD_DETECT_ROOT_VOL;
                        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                                       volume => $volume_name);
                        
                        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_aggr}->Node,
                                                                            boot_mode => C_ONTAP_CMODE,
                                                                            cmd => $command,
                                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
                        
                        my %processed_hash = $class->{_ui}->process_output_string(code => $cmd_code,
                                                                                  output => $output,
                                                                                  expected_rc => C_RC_FAILURE_NON_FATAL,
                                                                                  log_message => "Obtain dblade-level volume status for $volume_name",);
                        
                        if ($processed_hash{rc} == C_RC_ROOT_VOLUME_DETECTED)
                        {
                                $detected_mroot_vol = 1;
                                logcomment("Aggregate: offline_contained_volumes: Working on the mroot aggregate.\n" .
                                           "Skipping the mroot volume: $volume_name");
                                next;
                        }
                        
                }
                
                # Skip the vsroot volumes, if required
                if (($skip_vsroot_flag == TRUE) && ($class->{_aggr}->server->is_root_volume(volume => $volume_name)))
                {
                        logcomment("Aggregate: offline_contained_volumes: Skipping vsroot volume: $volume_name");
                        next;
                }
                
                # Check if already offline
                my $vol_state = $volume->status("state");
                if ($vol_state eq C_VOL_STATE_OFFLINE)
                {
                        logcomment("Aggregate: offline_contained_volumes: Skipping offline volume $volume_name");
                        next;
                }
                
                logcomment("Aggregate: Bringing volume $volume_name offline...") if $class->{_debug};
                $volume->offline();
                $output .= "$volume_name, ";
        }
        
        # If we got here, NOTE did not encounter any error
        return (rc => C_RC_SUCCESS, output => $output );
}


## @name ::online_contained_volumes
## @summary Onlines all volumes contained by this aggregate.
## 
## @description
## Onlines all volumes contained by this aggregate.
##
## @return Hash with rc and output (string containing ", "-delimited list of volumes that were onlined).
## 
## @arg skip_root_volume Optional: Boolean flag to decide if the root volume should be skipped.
## Defaults to TRUE.
## 
## @arg skip_vsroot_volumes Optional: Boolean flag to decide if the vsroot volume sshould be skipped.
## Defaults to TRUE.
##
## @examples
## $obj->online_contained_volumes();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub online_contained_volumes
{
	my $class = shift;
        my %args = validate(@_,
	{
		skip_root_volume => { type => BOOLEAN, optional => 1, default => TRUE, },
                skip_vsroot_volumes => { type => BOOLEAN, optional => 1, default => TRUE, }, 
	} );
        
        my $skip_root_flag = $args{skip_root_volume};
        my $skip_vsroot_flag = $args{skip_vsroot_volumes};
        my @volumes = $class->{_aggr}->Volumes();
        my $output = '';
        
        # See if this is mroot aggr (if it is, we will need to check for the mroot volume)
        my $mroot_aggr = $class->{_aggr}->server->storage_aggregate_show('has-mroot' => 'true',
                                                                         'nodes' => $class->{_aggr}->Node->name);
        my $root_aggr_name = $mroot_aggr->[0]->{aggregate};
        logcomment("Mroot aggregate is $root_aggr_name");
        
        my $detected_mroot_vol = 0;
        
        foreach my $volume (@volumes)
        {
                my $volume_name = $volume->name();
                logcomment("Aggregate: online_contained_volumes: Processing volume $volume_name") if $class->{_debug};
                
                # Skip the root volume, if required
                if (($skip_root_flag == TRUE) && ($root_aggr_name eq $class->{aggr_name}) && ($detected_mroot_vol == 0))
                {
                        logcomment("Aggregate: online_contained_volumes: Working on the mroot aggregate.\n" .
                                   "Looking for the mroot volume name") if ($class->{_debug});
                        my $cmd_code = C_CMD_DETECT_ROOT_VOL;
                        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                                       volume => $volume_name);
                        
                        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_aggr}->Node,
                                                                            boot_mode => C_ONTAP_CMODE,
                                                                            cmd => $command,
                                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
                        
                        my %processed_hash = $class->{_ui}->process_output_string(code => $cmd_code,
                                                                                  output => $output,
                                                                                  expected_rc => C_RC_FAILURE_NON_FATAL,
                                                                                  log_message => "Obtain dblade-level volume status for $volume_name",);
                        
                        if ($processed_hash{rc} == C_RC_ROOT_VOLUME_DETECTED)
                        {
                                $detected_mroot_vol = 1;
                                logcomment("Aggregate: online_contained_volumes: Working on the mroot aggregate.\n" .
                                           "Skipping the mroot volume: $volume_name");
                                next;
                        }
                        
                }
                
                # Skip the vsroot volumes, if required
                if (($skip_vsroot_flag == TRUE) && ($class->{_aggr}->server->is_root_volume(volume => $volume_name)))
                {
                        logcomment("Aggregate: online_contained_volumes: Skipping vsroot volume: $volume_name");
                        next;
                }
                
                # Check if already online
                my $vol_state = $volume->status("state");
                if ($vol_state eq C_VOL_STATE_ONLINE)
                {
                        logcomment("Aggregate: online_contained_volumes: Skipping online volume $volume_name");
                        next;
                }
                
                logcomment("Aggregate: Bringing volume $volume_name online...") if $class->{_debug};
                $volume->online();
                $output .= "$volume_name, ";
        }
        
        # If we got here, NOTE did not encounter any error
        return (rc => C_RC_SUCCESS, output => $output );
}


## @name ::set_offline
## @summary Offlines this aggregate.
## 
## @description
## Offlines this aggregate. Wrapper around NOTE::Aggregate::offline.
## It accepts the same arguments as Aggregate::aggr_offline.
##
## @arg force Optional: Boolean flag to force offlining the aggregate.
##
## @arg cifsdelaytime Optional: Number of seconds to delay before taking the aggregate offline,
## during which time CIFS uses are warned of the pending loss of service.
##
## @arg all Optional: Boolean flag to offline aggregate with all its volume
##
## @return Hash with rc and output.
##
## @examples
## $obj->set_offline();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub set_offline
{
        my $class = shift;
        my %args = validate(@_,
	{
		force => { type => BOOLEAN, optional => 1, },
                cifsdelaytime => { type => SCALAR, optional => 1, },
                all => { type => BOOLEAN, optional => 1, },
	} );
	
	my $all_volumes = $args{all};
	
	logcomment("Aggregate: set_offline called with arguments:\n" . Dumper(%args)) if ($class->{_debug});
	
	my %offline_args = (aggregate => $class->{aggr_name});
	if ((defined($all_volumes)) && ($all_volumes == TRUE))
	{
		$offline_args{'unmount-volumes'} = 'true';
	}
	
	# Use Server.pm directly. Looks like GNOTE is now using "aggr modify", which does not
	# provide any option to unmount all the contained volumes
	$class->{_aggr}->server->storage_aggregate_offline(%offline_args);
	
	# If we got here, underlying call did not encounter any error
	return (rc => C_RC_SUCCESS,);
}


## @name ::set_online
## @summary Onlines this aggregate.
## 
## @description
## Onlines this aggregate. Wrapper around NOTE::Aggregate::online.
## It accepts the same arguments as Aggregate::aggr_online.
##
## @arg force Optional: Boolean flag to force onlining the aggregate.
##
## @return Hash with rc and output.
##
## @examples
## $obj->set_online();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub set_online
{
        my $class = shift;
        my %args = validate(@_,
	{
		force => { type => BOOLEAN, optional => 1, },
	} );

        logcomment("Aggregate: set_online called with arguments:\n" . Dumper(%args)) if ($class->{_debug});
        $class->{_aggr}->online( %args );
        
        # If we got here, NOTE did not encounter any error
        return (rc => C_RC_SUCCESS,);
}


## @name ::get_avail
## @summary Get available space
## 
## @description
## Returns available space in KB.
## Wrapper around NOTE::Aggregate
##
## @return available space in KB
##
## @examples
## my $avail = $obj->get_avail();
## 
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_avail
{
        my $class = shift;
		my $avail = $class->{_aggr}->avail();
		$avail = int($avail/KB);
        return $avail;
}


## @name ::list_snapshots
## @summary Lists snapshots that exist on this aggregate.
## 
## @description
## Lists snapshots that exist on this aggregate. Wrapper around NOTE::Aggregate::snapshot_list.
##
## @return Hash with rc and output (reference to array with names of existent snapshots).
##  
## @examples
## $obj->list_snapshots();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##
sub list_snapshots
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # none
                } );
        
        my $cmd_code = C_CMD_AGGR_SNAP_LIST;
        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $class->{aggr_name});
	
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_aggr}->Node,
                                                            boot_mode => C_ONTAP_CMODE,
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
	
	my %processed_hash = $class->{_ui}->process_output_string(code => $cmd_code,
								  output => $output,
								  expected_rc => C_RC_SUCCESS,
								  log_message => "List snapshots of aggregate " . $class->{aggr_name},);
        logcomment("Processed aggregate snapshot list:\n" . Dumper(%processed_hash)) if ($class->{_debug});
        
        my $snapshot_entires = $processed_hash{snapshots};
        my $snap_list = [];
        foreach my $index (keys %$snapshot_entires)
        {
                push(@$snap_list, $snapshot_entires->{$index});
        }
        
        return (rc => C_RC_SUCCESS, output => $snap_list);
}


## @name ::delete_snapshot
## @summary Deletes a snapshot.
## 
## @description
## Deletes a named snapshot. Wrapper around NOTE::Aggregate::snapshot_delete.
##
## @return Hash with rc and output (empty).
##
## @arg name Required: Name of the snapshot.
##  
## @examples
## $obj->delete_snapshot(name => $my_snapshot);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##
sub delete_snapshot
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        name => { type => SCALAR, },
                } );
        
        my $snapshot_name = $args{name};
        my $cmd_code = C_CMD_AGGR_SNAP_DELETE;
        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $class->{aggr_name},
                                                       snapshot => $snapshot_name);
	
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_aggr}->Node,
                                                            boot_mode => C_ONTAP_CMODE,
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
	
	return $class->{_ui}->process_output_string(code => $cmd_code,
                                                    output => $output,
                                                    expected_rc => C_RC_SUCCESS,
                                                    log_message => "Delete snapshot $snapshot_name of aggregate " . $class->{aggr_name},);
}


## @name ::start_64_bit_upgrade
## @summary Starts 64-bit upgrade for this aggregate.
## 
## @description
## Starts 64-bit upgrade for this aggregate.
##
## @return Hash with rc and output.
##
## @arg mode Optional: Upgrade mode: check|normal|grow-reserved|grow-all.
## Default: none.
##  
## @examples
## $obj->start_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub start_64_bit_upgrade
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        mode => { type => SCALAR, optional => 1, default => C_LA_IPU_MODE_NORMAL },
                } );
        
        my $mode = $args{mode};
        my $aggr_name = $class->{aggr_name};
        my %return = (rc => C_RC_SUCCESS);
        
        my @allowed_modes = (C_LA_IPU_MODE_CHECK, C_LA_IPU_MODE_NORMAL,
                             C_LA_IPU_MODE_GROW_RESERVED, C_LA_IPU_MODE_GROW_ALL);
        
        if (!grep(/^$mode$/, @allowed_modes))
        {
                logresult('script', "Illegal \"mode\" parameter passed to start_64_bit_upgrade: $mode\n" .
                                    "Allowed values:\n" . Dumper(@allowed_modes));
        }
        
        logcomment("Starting LA IPU on aggregate $aggr_name, mode: $mode");
        
        my $result = $class->{_aggr}->upgrade_start(%args);
        
        if ((defined($mode)) && ($mode eq 'check'))
        {
                # Note: according to the documentation, GNOTE will, at some later time,
                # return the result in this case as an array of hashrefs
                # For now, it is returned as a string
                $return{output} = $result;
        }
        
        logcomment("Aggregate: start_64_bit_upgrade returning:\n" . Dumper(%return)) if ($class->{_debug});
        return %return;
}


## @name ::wait_for_64_bit_upgrade
## @summary Waits for a previously started 64-bit upgrade to complete.
## 
## @description
## Waits for a previously started 64-bit upgrade to complete.
##
## @return Hash with rc.
##
## @arg timeout Optional: Timeout value, in seconds.
##
## @arg delay Optional: Time interval, in seconds, in between consecutive status checks.
##  
## @examples
## $obj->wait_for_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub wait_for_64_bit_upgrade
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        timeout => {type => SCALAR, optional => 1,},
                        delay => {type => SCALAR, optional => 1,},
                } );
        
        my $wait_status = $class->{_aggr}->upgrade_wait(%args); 
        if ($wait_status != 0)
        {
                logresult('fatal', "LA IPU for aggregate " . $class->{aggr_name} . "took more than the allowed timeout");
        }
        
        return (rc => C_RC_SUCCESS);
}


## @name ::stop_scan_64_bit_upgrade
## @summary Stops a previously started 64-bit upgrade.
## 
## @description
## Stops a previously started 64-bit upgrade .
##
## @return Hash with rc.
##  
## @examples
## $obj->stop_scan_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub stop_scan_64_bit_upgrade
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # None
                } );
        
        $class->{_aggr}->upgrade_scan_stop(); 
        
        # If we got here, GNOTE did not encounter any error
        return (rc => C_RC_SUCCESS);
}


## @name ::start_scan_64_bit_upgrade
## @summary Restarts a previously stopped 64-bit upgrade.
## 
## @description
## Restarts a previously stopped 64-bit upgrade.
##
## @return Hash with rc.
## @examples
## $obj->start_scan_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub start_scan_64_bit_upgrade
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # None
                } );
        
        $class->{_aggr}->upgrade_scan_start();
        
        # If we got here, GNOTE did not encounter any error
        return (rc => C_RC_SUCCESS);
}



## @name ::status_64_bit_upgrade
## @summary Gets the status of the 64-bit upgrade for this aggregate.
## 
## @description
## Gets the status of the 64-bit upgrade for this aggregate.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_LA_IPU_STOPPED/C_RC_LA_IPU_NOT_IN_PROGRESS.
## (defined in Compression::Constants).
##
## @return Hash with rc and output. 
##  
## @examples
## my %res = $obj->status_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub status_64_bit_upgrade
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
                } );
	
        my $status_ref;
        my $output = '';
        my $expected = $args{expected};
        my $rc = C_RC_FAILURE;
        my $aggr_name = $class->{aggr_name};
        
	eval { $status_ref = $class->{_aggr}->upgrade_status(); };
        
        my %return = $class->{_ui}->process_output_string(code => C_CMD_LA_IPU_STATUS,
                                                          output => $@,
                                                          expected_rc => C_RC_FAILURE_NON_FATAL,
                                                          log_message => "Obtain 64-bit upgrade status for aggregate $aggr_name",);
        
        $rc = $return{rc};
        if ($rc == C_RC_SUCCESS)
        {
                # We should get a valid $status_ref - look for our aggr
                foreach my $status_entry (@$status_ref)
                {
                        logcomment("Aggregate: status_64_bit_upgrade processing:\n" . Dumper($status_entry)) if ($class->{_debug});
                        if (($status_entry->{aggregate} eq $aggr_name) && ($status_entry->{'scanner-status'} eq 'stopped'))
                        {
                                $rc = C_RC_LA_IPU_STOPPED;
                                last;
                        }
                        
                }
                $output = Dumper($status_ref);
        }
        else
        {
                $output = $@;
        }
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $expected,
                                                    log_message => "Obtain 64-bit upgrade status for aggregate $aggr_name",
                                                    output      => $output);
}


## @name ::is_block_type_32bit
## @summary Checks if the aggregate is 32-bit.
## 
## @description
## Checks if the aggregate is 32-bit.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc.
##  
## @examples
## $obj->is_block_type_32bit(expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_block_type_32bit
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
                } );
	
        my $rc;
        
	my $is_32bit = $class->{_aggr}->is_32bit();
        $rc = ($is_32bit == 1) ? C_RC_SUCCESS : C_RC_FAILURE;
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Check if aggregate block type is 32-bit",
                                                    output      => $is_32bit);
}


## @name ::is_block_type_64bit
## @summary Checks if the aggregate is 64-bit.
## 
## @description
## Checks if the aggregate is 64-bit.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc.
##  
## @examples
## $obj->is_block_type_64bit(expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_block_type_64bit
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
                } );
	
        my $rc;
        
	my $is_64bit = $class->{_aggr}->is_64bit();
        $rc = ($is_64bit == 1) ? C_RC_SUCCESS : C_RC_FAILURE;
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Check if aggregate block type is 64-bit",
                                                    output      => $is_64bit);
}


## @name ::list_contained_volumes
## @summary List volumes contained by this aggregate.
## 
## @description
## List volumes contained by this aggregate.
##
## @return Array of volume names.
##  
## @examples
## my @vols = $obj->list_contained_volumes();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub list_contained_volumes
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        # None
                });
        
        return $class->{_aggr}->volumes();
}


## @name ::get_data_disk_names
## @summary Gets data disks names of this aggregate.
## 
## @description
## Gets data disks names of this aggregate.
##
## @return
## Array of data disk names.
##  
## @examples
## my @data_disks = $obj->get_data_disk_names();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_data_disk_names
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			# None
		});
	
	my @disks = ();
	my $cmd_code = C_CMD_GET_DATA_DISKS;
	
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
						       aggregate => $class->{aggr_name},);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
							    boot_mode => C_ONTAP_CMODE,
							    cmd => $command,
							    node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
	
	my $data_disk_prefix = $class->{_ui}->get_constant_value(constant_code => 'AGGR_DATA_DISK_PREFIX');
	
	my @lines = split(/\n/, $output);
	
	foreach my $line (@lines)
	{
		if ($line =~ /^$data_disk_prefix(\S+)/)
		{
			push(@disks, $1);
		}
	}
	
	if (scalar(@disks) == 0)
	{
		logresult('fatal', "Could not detect data disks for aggregate " . $class->{aggr_name})
	}
	
	return @disks;
}


## @name ::start_block_check
## @summary Runs block_check on this aggregate.
## 
## @description
## Runs block_check on this aggregate.
##
## @arg wait Optional: Flag to control if the method should wait for the scanner to complete.
## Default: TRUE.
## 
## @arg disks: Optional: Arrayref of disk names to run block check on.
## If none of the 'disks' and 'pvbn' arguments are passed in, then block check will be run on all the data disks of this aggregate.
## 
## @arg pvbn: Optional: Arrayref of pvbns to run block check against.
## If none of the 'disks' and 'pvbn' arguments are passed in, then block check will be run on all the data disks of this aggregate.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Default: C_RC_SUCCESS.
## 
## @arg expected_pvbn_file_path Optional: Expected file path to be listed by block check after completion.
## Only relevant if 'pvbn' is passed in.
##
## @return Hash with rc and output.
##  
## @examples
## $obj->start_block_check();
## $obj->start_block_check(pvbn => [0 .. 7],
##                         expected_pvbn_file_path => 'vol/my_vol/my_file');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub start_block_check
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			wait => {type => BOOLEAN, default => TRUE,},
			disks => {type => ARRAYREF, optional => 1,},
			pvbn => {type => ARRAYREF, optional => 1},
			expected => {type => SCALAR, default => C_RC_SUCCESS,},
			expected_pvbn_file_path => {type => SCALAR, optional => 1,},
		});
	
	my $wait = $args{wait};
	my $expected = delete($args{expected});
	my $expected_pvbn_file_path = delete($args{expected_pvbn_file_path});
	my $aggr_name = $class->{aggr_name};
	
	my @data_disks = ();
	if ((!exists($args{disks})) && (!exists($args{pvbn})))
	{
		@data_disks = $class->get_data_disk_names();
		$args{disks} = \@data_disks;
	}
	
	my $disks_checked = (exists($args{disks})) ? TRUE : FALSE;
	
	$class->_ensure_nacl_api_obj();
	
	$args{api_obj} = $class->{_nacl_api_obj};
	$args{aggr} = $aggr_name;
	
	my ($errcode_or_scan_id, $logs) = FS::Aggr_block_check::aggr_blockcheck(%args);
	
	if ($wait == TRUE)
	{
		# $errcode_or_scan_id is the error code
		# $logs is log file path if success, error message if failure
		if ($errcode_or_scan_id == 1)
		{
			return $class->{_ui}->process_output_string(code => C_CMD_AGGR_BLOCK_CHECK_START,
								    output => $logs,
								    expected_rc => $expected,
								    log_message => "Start block_check on aggregate $aggr_name");
		}
		
		return $class->_parse_aggr_block_check_log(log => $logs,
							   disks_checked => $disks_checked,
							   pvbn => $args{pvbn},
							   expected => $expected,
							   expected_pvbn_file_path => $expected_pvbn_file_path);
	}
	else
	{
		# If the scanner completes too quickly, the scan id will not be retrieved
		my $scan_id = (defined($errcode_or_scan_id)) ? $errcode_or_scan_id : 'unknown';
		$class->{_block_check_scans}->{$scan_id}->{logs} = $logs;
		$class->{_block_check_scans}->{$scan_id}->{disks_checked} = $disks_checked;
		$class->{_block_check_scans}->{$scan_id}->{pvbn} = $args{pvbn};
		$class->{_block_check_scans}->{$scan_id}->{expected_pvbn_file_path} = $expected_pvbn_file_path;
		
		if ("$scan_id" eq "1")
		{
			return $class->{_ui}->process_output_string(code => C_CMD_AGGR_BLOCK_CHECK_START,
								    output => $logs,
								    expected_rc => $expected,
								    log_message => "Start block_check on aggregate $aggr_name");
		}
		
		$class->{_latest_block_check_id} = "$scan_id";
		
		return (rc => C_RC_SUCCESS, scan_id => $scan_id, output => "Error code: $scan_id\nLogs/error massage: $logs");
	}
}


## @name ::wait_for_block_check
## @summary Waits for already-started block_check scan to complete.
## 
## @description
## Waits for already-started block_check scan to complete.
##
## @arg scan_id Optional: Id of the block_check scan to wait for.
## If not provided, this will be the last-started scan.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Default: C_RC_SUCCESS.
##  
## @examples
## $obj->wait_for_block_check();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub wait_for_block_check
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			scan_id => {type => SCALAR, regex => qr/^\d+|unknown$/, default => $class->{_latest_block_check_id},},
			expected => {type => SCALAR, default => C_RC_SUCCESS,},
		});
	
	my $expected = $args{expected};
	my $aggr_name = $class->{aggr_name};
	my $scan_id = $args{scan_id};
	
	$class->_ensure_nacl_api_obj();
	
	my %wait_args = (api_obj => $class->{_nacl_api_obj},
			 aggr => $aggr_name,);
	
	if (!exists($class->{_block_check_scans}->{$scan_id}))
	{
		logresult('script', "Aggr block_check scanner with id \"$scan_id\" was not started in the background through this library.\n" .
			            "No log information found:\n" . Dumper($class->{_block_check_scans}));
	}
	
	my $scan_id_info = $class->{_block_check_scans}->{$scan_id};
	$wait_args{ls_etc_log_before} = $scan_id_info->{logs};
	
	if ($scan_id =~ /^\d+$/)
	{
		$wait_args{scanid} = $scan_id;
	}
	
	my $log_path = FS::Aggr_block_check::aggr_blockcheck_wait(%wait_args);
	
	return $class->_parse_aggr_block_check_log(log => $log_path,
						   disks_checked => $scan_id_info->{disks_checked},
						   pvbn => $scan_id_info->{pvbn},
						   expected => $expected,
						   expected_pvbn_file_path => $scan_id_info->{expected_pvbn_file_path});
}



######################################
# Internal methods
######################################

sub _ensure_nacl_api_obj
{
	my $class = shift;
	
	if (!defined($class->{_nacl_api_obj}))
	{
		$class->{_nacl_api_obj} = NACL::APISet->new(name => $class->{_filer_name},
							    category  => "Node",
							    interface => "CLI",
							    set       => "CMode", );
	}
}

sub _parse_aggr_block_check_log
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			log => {type => SCALAR,},
			disks_checked => {type => BOOLEAN,},
			pvbn => {type => UNDEF|ARRAYREF, optional => 1},
			expected => {type => SCALAR,},
			expected_pvbn_file_path => {type => UNDEF|SCALAR, optional => 1,},
		});
	
	my $log_path = $args{log};
	my $disks_checked = $args{disks_checked};
	my $expected_pvbn_file_path = $args{expected_pvbn_file_path};
	my $pvbn = $args{pvbn};
	my %return_hash = ();
	my $rc = C_RC_SUCCESS;
	
	$class->_ensure_nacl_api_obj();
	
	if ($disks_checked == TRUE)
	{
		my ($scanned_disk,$vbn,$corrupted_disks,$levels,$fbn,$error_code,$snapshot_list,$lost_block_count)= FS::Aggr_block_check::aggr_blockcheck_read_output(api_obj => $class->{_nacl_api_obj},
																				      path => $log_path,
																				      disks => 1);
		
		$return_hash{output} = "Scanned disks:\n" . Dumper($scanned_disk) .
		                       "\n\nVbn:\n" . Dumper($vbn) .
				       "\n\nCorrupted disks:\n" . Dumper($corrupted_disks) .
				       "\n\nLevels:\n" . Dumper($levels) .
				       "\n\nFbns:\n" . Dumper($fbn) .
				       "\n\nErrors:\n" . Dumper($error_code) .
				       "\n\nSnapshot list:\n" . Dumper($snapshot_list) .
				       "\n\nLost block count:\n" . Dumper($lost_block_count);
		
		if (scalar(keys(%$corrupted_disks)) != 0)
		{
			$rc = C_RC_CORRUPTED_DISKS_FOUND;
		}
		
		if (scalar(keys(%$lost_block_count)) != 0)
		{
			$rc = C_RC_LOST_BLOCKS_FOUND;
		}
	}
	else
	{
		my ($pvbn_file,$level,$error_code) = FS::Aggr_block_check::aggr_blockcheck_read_output(api_obj => $class->{_nacl_api_obj},
												       path => $log_path,
												       pvbn => $pvbn);
		
		$return_hash{output} = "Pvbn info:\n" . Dumper($pvbn_file) .
		                       "\n\nLevels:\n" . Dumper($level) .
				       "\n\nError codes:\n" . Dumper($error_code);
		
		# Aggr block_check skips offline/restricted volumes -> make sure all these pvbns were scanned...
		foreach my $pvbn_instance (@$pvbn)
		{
			if (!exists($pvbn_file->{$pvbn_instance}))
			{
				logcomment("Could not detect info for pvbn $pvbn_instance in the aggr block_check output");
				$rc = C_RC_PVBN_INFO_MISSING;
				$return_hash{pvbn_info}{$pvbn_instance} = 'MISSING';
			}
			else
			{
				my $info = $pvbn_file->{$pvbn_instance};
				chomp($info);
				# chomp seems to not be enough
				$info =~ s/\W*$//;
				$return_hash{pvbn_info}{$pvbn_instance} = $info;
				
				# This is currently a dblade command. Need to remove the /vol
				$info =~ s/^\/vol//;
				
				if ((defined($expected_pvbn_file_path)) && ($expected_pvbn_file_path ne $info))
				{
					logcomment("Detected incorrect file path for pvbn $pvbn_instance:\n" .
						   "Expected: $expected_pvbn_file_path\n" .
						   "Obtained: $info");
					$rc = C_RC_INCORRECT_PVBN_INFO;
				}
			}
		}
		
		# Check for detected errors
		foreach my $pvbn_checked (keys(%$error_code))
		{
			if ($error_code->{$pvbn_checked} != 0)
			{
				logcomment("Detected corruption: pvbn $pvbn_checked, error code " . $error_code->{$pvbn_checked});
				$rc = C_RC_LOST_BLOCKS_FOUND;
			}
		}
	}
	
	$class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
					     expected_rc => $args{expected},
					     log_message => "Run block_check on aggregate " . $class->{aggr_name},
					     output      => \%return_hash,);
	
	$return_hash{rc} = $rc;
	return %return_hash;
}

sub _refresh
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		new_aggr => {type => SCALAR,},
	});
	
	my $new_aggr = $args{new_aggr};
	my $old_aggr = $class->{aggr_name};
	
	# Refresh the NOTE aggr object of this class.
	# If required to not recycle, destroy the old aggr first
	if ($class->{_recycle} == FALSE)
	{
		if (defined($ENV{'C_CORRUPTION_FOUND'}) && ($ENV{'C_CORRUPTION_FOUND'} != TRUE))
		{
			logcomment("Corruption found. Aggregate $old_aggr will not be destroyed");
		}
		else
		{
			logcomment("Purging aggregate $old_aggr");
			$class->{_aggr}->purge();
		}
	}
	
	my $old_node_name = $class->{_conn}->name();
	
	# Attach to the new aggregate
	$class->{_aggr} = NOTE::GX::Aggregate->attach(name => $new_aggr,
						      Node => $class->{_conn});
	
	my $node_name = $class->{_aggr}->Node->name();
	
	if ($node_name ne $old_node_name)
	{
		logcomment("Compression aggregate is now hosted by new filer: $node_name (used to be $old_node_name)");
		$class->{_hosting_filer_has_changed} = TRUE;
		$class->{_filer_name} = $node_name;
		$class->{_block_check_scans} = {};
		$class->{_latest_block_check_id} = '';
		$class->{_conn} = $class->{_aggr}->Node;
		my $version = Compression::Version->new( conn => $class->{_conn}, debug => $class->{_debug}, boot_mode => C_ONTAP_CMODE );
		$class->{_ui} = Compression::UI->new( version => $version, debug => $class->{_debug}, boot_mode => C_ONTAP_CMODE );
	}
	
	$class->{aggr_name} = $new_aggr;
	$class->{_block_check_scans} = {};
	$class->{_latest_block_check_id} = '';
}


# Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
	
	if (($class->{_disable_cleanup} == FALSE) &&
            ($class->{_creator_pid} == $$))
	{
                my $aggr_name = $class->{aggr_name};
                logcomment("Executing C-mode Aggregate clean-up for aggregate $aggr_name");

		eval
                {
                        if (($class->{_recycle} == FALSE) &&
			    (!defined($ENV{'C_CORRUPTION_FOUND'}) || ($ENV{'C_CORRUPTION_FOUND'} != TRUE)))
			{
				logcomment("Purging aggregate $aggr_name");
				$class->{_aggr}->purge();
			}
                };
                
                if( $@ )
                {
                        logcomment( "C-mode Aggregate: cleanup for aggregate $aggr_name: error: " . $@ );
                }
                else
                {
                        logcomment( "C-mode Aggregate: cleanup for aggregate $aggr_name: done" ) if ($class->{_debug});
                }
        }
	else
	{
                logcomment( "C-mode Aggregate: cleanup:  nothing done: script PID: " . $$) if ($class->{_debug});
	}
        
        if (defined($execution_error))
        {
                _c_interpret_logresult($execution_error);
        }
}



1;
