#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/CMode/Filer.pm#1 $
#

# @summary Wrapper around NOTE::Filer.
# 
# @description
# 
# Wrapper around NOTE::Filer.
# The main reason for not using NOTE::Filer directly is to provide standardized output,
# and to enable specific processing on some failures, otherwise fatal.
# This is required for handling compression-specific failures.
# 
# @synopsis
#
# use Compression::Filer;
# use vars qw($FILER );
# 
# BEGIN { require "Compression/setup_core.pl"; }
# 
# my $obj = Compression::Filer->new( filer => $FILER );
# my $obj = Compression::Filer->new( filer => { name => $FILER, conn_type => '#rsh', stats => FALSE } );
#
# @examples 
#
# my $obj = Compression::Filer->new( filer => $FILER );
# $obj->license_add( service => C_LICENSE_COMPRESSION, expected => C_RC_NO_NEARSTORE_LICENSE );
# $obj->license_remove( service => C_LICENSE_COMPRESSION, expected => C_RC_COMPRESSION_NOT_TURNED_OFF );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 


package Compression::CMode::Filer;

use strict;
use warnings;

use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;
use Date::Manip;

use Tharn;
use LogReader::Messages;
use Compression::Filer::License;
use Compression::Fiji;
use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult
                          detect_indirect_type_from_status);
;


BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
        
        # Keep a global hash to keep track of CP state per each filer.
        # This way, we do not depend on the number and usage of Compression objects created by a script.
        # Extra cleaning is needed if timer-triggered CP was disabled: only clearing the Fiji point is not enough.
        $Compression::Filer::CP_is_disabled = {};
}


# @name ::new
# @summary Compression::Filer Constructor.
#
# @description
# Creates an object representing a filer.
#
# @return Object of type Compression::Filer.
#
# @arg filer Required: Scalar (filer name) or hash of filer information.
# If filer is passed in as a hashref, then the allowed keys/values are:
# - name => name of the filer,
# - mode => mode to connect to the filer. Default: rsh,
# - stats => flag to enable recording/printing statistics. Default: FALSE,
# - license_ready => flag to enable installing the compression license. Default: C_DFLT_LICENSE_READY,
# - require_debug_build => Flag to enable checking for a debug build. If a debug build is required, but not found
#   then the test will be terminated as not applicable. Default: C_DFLT_REQUIRE_DEBUG_BUILD,
# - require_hw_platform => Flag to enable checking for hardware platform. If a simulator is found instead,
#   then the test will be terminated as not applicable. Default: C_DFLT_REQUIRE_HW_PLATFORM,
# - require_cf_disabled => Flag to be used to disable cf, if needed. By using this flag, no further action is needed within the script.
#   The library will take care of re-enabling cf at the end of the script.
# 
# @arg debug Optional: Controls verbosity of this module.
# Defaults to FALSE.
#
# @example
# my $obj = Compression::Filer->new( filer => $FILER, debug => TRUE );
# my $obj = Compression::Filer->new( filer => { name => $FILER, conn_type => '#rsh', stats => TRUE } );
# my $obj = Compression::Filer->new( filer => { name => $FILER, license_ready => FALSE, require_debug_build => TRUE, require_hw_platform => TRUE, } );
#
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub new
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			filer 	=> {type => SCALAR|HASHREF,
				    },
                        
                        vserver => {type => SCALAR,},
                        
                        debug 	=> {type => SCALAR,
                                    optional => 1,
				    default => FALSE,
				    },
                        
                        required_features => {type => ARRAYREF|UNDEF,
                                              optional => 1,
                                              },
		} );
        
        my $self = {};
        $self->{_debug} = $args{debug};
        $self->{_creator_pid} = $$;
        $self->{_start_vol_count} = 0;
        $self->{_start_vol_list} = '';
        $self->{_boot_mode} = $args{boot_mode};
        
        logcomment("CMode Filer: new called with arguments: " . Dumper(%args)) if ($self->{_debug});
        
        # First, construct the filer hash
        my $filer_args = $args{filer};
	my %filer = ( );
        
        my $auto_license_flag = C_DFLT_LICENSE_READY;
        my $debug_build_needed = C_DFLT_REQUIRE_DEBUG_BUILD;
        my $hw_platform_needed = C_DFLT_REQUIRE_HW_PLATFORM;
        my $cf_disabled_needed = C_DFLT_CF_DISABLE;
        if (ref($filer_args) eq "HASH")
        {
                # Second-layer hash validation for filer args
                logcomment("CMode Filer: new: filer attributes passed in as HASHREF") if ($self->{_debug});
                %filer = validate_with
                        ( params => $filer_args,
                          spec   => {name => { type => SCALAR },
                                     stats => { type => SCALAR, default => C_DFLT_FILER_CONN_STATS },
                                     license_ready => { type => SCALAR, default => C_DFLT_LICENSE_READY },
                                     require_debug_build => { type => SCALAR, default => C_DFLT_REQUIRE_DEBUG_BUILD },
                                     require_hw_platform => { type => SCALAR, default => C_DFLT_REQUIRE_HW_PLATFORM },
                                     require_cf_disabled => { type => SCALAR, default => C_DFLT_CF_DISABLE }, },
                          allow_extra => 0,
                        );
                
                $auto_license_flag = $filer{license_ready};
                $debug_build_needed = $filer{require_debug_build};
                $hw_platform_needed = $filer{require_hw_platform};
                $cf_disabled_needed = $filer{require_cf_disabled};
                $self->{host_name} = $filer{name};
        }
        else
        {
                logcomment("CMode Filer: new: filer attribute passed in as SCALAR") if ($self->{_debug});
                $self->{host_name} = $filer_args;
                #$filer{mode} = C_DFLT_10MODE_FILER_CONN;
                $filer{stats} = C_DFLT_FILER_CONN_STATS;
        }
        
        # Connect to filer
        $self->{_conn} = NOTE::GX::Node->connect(name => $self->{host_name},
                                                 stats => $filer{stats});
        
        $self->{_cluster_enable_nedded} = FALSE;
        $self->{_ha_was_configured} = FALSE;
        my $node_build_type = $self->{_conn}->build_type();
        $node_build_type =~ s/^\s*//;
        $node_build_type =~ s/\s*$//;
        $self->{_build_type} = $node_build_type;
        
        # Check if filer is required to not be a simulator
        my $cmode_build_sim = C_10MODE_BUILD_TYPE_SIM;
        if (($hw_platform_needed == TRUE) && ($self->{_build_type} =~ /$cmode_build_sim/))
        {
                logresult ('NA', "NOT APPLICABLE: Filer " . $self->{host_name} . " is a simulator. Script required HW platform.\n" .
                                 "Detected build type: " . $self->{_build_type});
        }
        
        # Check if debug build is requested
        $self->{_has_debug_kernel} = FALSE;
        my $cmode_build_type_debug = C_10MODE_BUILD_TYPE_DEBUG;
        if ($self->{_build_type} =~ /$cmode_build_type_debug/ )
        {
                $self->{_has_debug_kernel} = TRUE;
        }
        
        if (($debug_build_needed == TRUE) && ($self->{_has_debug_kernel} == FALSE))
        {
                logresult ('NA', "NOT APPLICABLE: Could not detect a debug kernel on node " . $self->{host_name} . ". Script required debug kernel.\n" .
                                 "Detected build type: " . $self->{_build_type});
        }

        # Version
        $self->{version} = Compression::Version->new(conn => $self->{_conn},
                                                     debug => $self->{_debug},
                                                     boot_mode => C_ONTAP_CMODE); 
        if (defined($args{required_features}))
        {
                $self->{version}->check_for_required_features(features => $args{required_features});
        }
        
        # UI
        $self->{_ui} = Compression::UI->new(version => $self->{version},
                                            debug => $self->{_debug},
                                            boot_mode => C_ONTAP_CMODE); 

        bless $self, $class;
        
        ############################################
	# Do extra setup steps: licensing
	############################################
	if($auto_license_flag == TRUE)
	{
		logcomment( "Ensuring Compression licensing" ) if ($self->{_debug});
		$self->_license_compression();
	}
        
        logcomment("Filer: completed object creation for \"" . $self->{host_name} . "\"") if ($self->{_debug});
        
        ############################################
	# Do extra setup steps: disable cfo
	############################################
        if ($cf_disabled_needed == TRUE)
        {
                $self->_require_standalone_filer();
        }
        
	return $self;
}


# @name ::get_ontap_version
# @summary Returns the ontap version string.
# 
# @description
# Returns the ontap version string.
#
# @return Ontap version string.
# 
# @examples
# my $ontap = $obj->get_ontap_version();
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub get_ontap_version
{
	my $class = shift;
        my %args = validate(@_,
	{
		# No args
	} );
        
        return $class->{_ui}->{version}->{ontap};
}


# @name ::get_indirect_type_of_root_vol
# @summary Returns the indirect type of the root volume.
# 
# @description
# Returns the indirect type of the root volume.
#
# @return Indirect type of the root volume (32/64).
# 
# @examples
# my $root_type = $obj->get_indirect_type_of_root_vol();
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub get_indirect_type_of_root_vol
{
	my $class = shift;
        my %args = validate(@_,
	{
		# No args
	} );
        
        my $mroot_aggr = $class->{_conn}->server->storage_aggregate_show('has-mroot' => 'true',
                                                                         'nodes' => $class->{host_name});
        my $root_aggr_block_type = $mroot_aggr->[0]->{'block-type'};
        return detect_indirect_type_from_status(status_string => $root_aggr_block_type);
}


# @name ::license_add
# @summary Adds license for specified feature.
# 
# @description
# Adds license for specified feature.
#
# @return Hash with rc and output.
# 
# @arg service Required: Feature to be licensed.
# 
# @arg expected Optional: Expected return code (defined in Compression::Constants).
# If not provided, assumes the command should complete successfully.
#
# @examples
#
# $obj->license_add( service => C_LICENSE_COMPRESSION );
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub license_add
{
	my $class = shift;
        my %args = validate(@_,
	{
		service   => { type => SCALAR, },
		expected  => { type => SCALAR, optional => 1, default => C_RC_SUCCESS, },
	} );
        
        my $node = $class->{_conn};
        my $service = $args{service};
        my $output = "";
        
        my $cmode_license_codes = $class->_get_cmode_license_code(service => $service);
        foreach my $code (@$cmode_license_codes)
        {
                logcomment("Licensing \"$code\"");
                $node->Cluster->license($code, 1);
        }
        
        # For now, do not process any failure
        return (rc => C_RC_SUCCESS);
}

# @name ::license_remove
# @summary Removes license for specified feature.
# 
# @description
# Removes license for specified feature. Can handle compression-specific failures.
#
# @return Hash with rc and output.
# 
# @arg service Required: Feature to be unlicensed.
# 
# @arg expected Optional: Expected return code (defined in Compression::Constants).
# If not provided, assumes the command should complete successfully.
#
# @examples
#
# $obj->license_remove( service => C_LICENSE_COMPRESSION, expected => C_RC_COMPRESSION_NOT_TURNED_OFF );
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub license_remove
{
	my $class = shift;
        my %args = validate(@_,
	{
		service   => { type => SCALAR, },
                expected  => { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },	
	} );
        
        my $node = $class->{_conn};
        my $service = $args{service};
        my $output = "";
        
        my $cmode_license_codes = $class->_get_cmode_license_code(service => $service);
        foreach my $code (@$cmode_license_codes)
        {
                logcomment("Licensing \"$code\"");
                $node->Cluster->license($code, 0);
        }
        
        # For now, do not process any failure
        return (rc => C_RC_SUCCESS);
}


# @name ::license_verify
# @summary Verifies if filer is licensed for specified feature.
# 
# @description
# Verifies if filer is licensed for specified feature. 
#
# @return Hash with rc and output.
# 
# @arg service Required: Feature to be verified.
# 
# @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
# (defined in Compression::Constants).
# 
# @arg site_license Optional: Boolean flag to enable extra checking for site license.
# Defaults to FALSE.
#
# @examples
#
# $obj->license_verify( service => C_LICENSE_COMPRESSION, expected => C_RC_SUCCESS );
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub license_verify
{
	my $class = shift;
        my %args = validate(@_,
	{
		service => { type => SCALAR, },
                expected => { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },
                site_license => { type => SCALAR, optional => 1, },
	} );
        
        my $node = $class->{_conn};
        my $service = $args{service};
        my $expected = $args{expected};
        my $expected_setting = ($expected == C_RC_SUCCESS) ? 1 : 0;
        my $output = "";
        
        my $cmode_license_codes = $class->_get_cmode_license_code(service => $service);
        foreach my $code (@$cmode_license_codes)
        {
                logcomment("Verifying license for \"$code\"");
                my $current_setting = $node->Cluster->license($code);
                if ($current_setting != $expected_setting)
                {
                        logresult('fatal', "Verify license for $service:\n" .
                                           "Expected setting: $expected_setting\n" .
                                           "Current setting: $current_setting");
                }
        }
        
        return (rc => C_RC_SUCCESS);
        
        
        
        
#        my %ret_hash = ();
#        my $rc = C_RC_SUCCESS;
#        my $filer = $class->{_conn};
#        my $service = $args{service};
#	my $site_license_required = $args{site_license};
#        my $output = "";
#	
#        my $conn_rec = $class->{_conn}->conn();
#        my $is_licensed = filer_service_is_licensed($conn_rec, $service);
#        if (($is_licensed == 0) ||
#	   (defined($site_license_required) && ($site_license_required == TRUE) && ($is_licensed != 2)) ||
#	   (defined($site_license_required) && ($site_license_required == FALSE) && ($is_licensed != 1)))
#        {
#                $rc = C_RC_FAILURE;
#        }
#        
#        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
#                                                    expected_rc => $args{expected},
#                                                    log_message => "Verify license for $service",
#                                                    output      => $is_licensed);
}


# @name ::set_option_on_all_volumes
# @summary Sets an option to all the volumes existent on the filer.
# 
# @description
# Retrieves the list of volumes existent on the filer and sets the desired option for all volumes.
#
# @return Hash with rc and output.
# Output is a hashref with key (= volume name) => value (= new option setting).
# 
# @arg service Required: Feature that corresponds to desired option (defined in Compression::Constants).
# 
# @arg new_value Required: Tag that corresponds to new value to be set.
# For now, only TRUE/FALSE are handled, corresponding to enable/disable.
#
# @arg skip_root_volume Optional: Boolean flag to decide if the option should be attempted to be set for the root volume.
# Defaults to TRUE.
#
# @arg skip_offline_restricted Optional: Boolean flag to decide if the option should be attempted to be set for offline/restricted volumes.
# Defaults to TRUE.
#
# @arg check_if_already_set Optional: Flag to enable the library to first check the current setting.
# If the current setting matches the value to be set, the subroutine will not attempt to set it again.
# Defaults to TRUE.
#
# @notes
# Currently it does not handle compression-specific failures.
#
# @examples
# $obj->set_option_on_all_volumes( service => C_OPTION_COMPRESSION, new_value => FALSE, skip_root_volume => TRUE );
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub set_option_on_all_volumes
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        service          => { type => SCALAR, },
                        new_value        => { type => SCALAR, },
                        skip_root_volume => { type => BOOLEAN, optional => 1, default => TRUE, },
                        skip_offline_restricted => { type => BOOLEAN, optional => 1, default => TRUE, },
                        check_if_already_set => { type => BOOLEAN, optional => 1, default => TRUE },
                } );
        
        logresult('script', "This method is not available for c-mode");
}


# @name ::volume_exists
# @summary Verifies that the specified volume exists on this filer.
# 
# @description
# Verifies that if the specified volume exists on this filer.
#
# @return Hash with rc and output (list of existent volumes).
# 
# @arg volume Required: Name of the volume.
# 
# @arg expected Optional: Expected result (C_RC_SUCCESS/C_RC_FAILURE).
# If not provided, assumes that the volume should exist.
# C_RC_FAILURE_NON_FATAL can also be passed in as 'expected', in which case,
# the script will not be terminated upon any result.
# Instead, the result (C_RC_SUCCESS/C_RC_FAILURE) will be passed to the script for further processing.
#
# @examples
#
# $obj->volume_exists( volume => $vol, expected => C_RC_FAILURE );
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub volume_exists
{
	my $class = shift;
        my %args = validate(@_,
	{
		volume   => { type => SCALAR, },
                expected  => { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },		
	} );
        
        my $vol_to_find = $args{volume};
        my $rc = C_RC_FAILURE;
        
        my %existent_vols = $class->get_status_of_all_volumes();
        if (exists($existent_vols{status}{$vol_to_find}))
        {
                logcomment("CMode Filer: Volume $vol_to_find found: $vol_to_find:\n" . Dumper(%existent_vols)) if $class->{_debug};
                $rc = C_RC_SUCCESS;
        }
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Verify that the volume \"$vol_to_find\" exists on filer " . $class->{host_name},
                                                    output => \%existent_vols);
}


# @name ::get_status_of_all_volumes
# @summary Gets the status of all volumes existent on this filer.
# 
# @description
# Gets the status of all volumes existent on this filer.
#
# @return Hash with rc and status (which is a hash with key<->values: vol_name<->hash (parent, options, status, type, state)) 
#
# @examples
#
# $obj->get_status_of_all_volumes();
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub get_status_of_all_volumes
{
	my $class = shift;
        my %args = validate(@_,
	{
                # No args
	} );
        
        my %return_hash = ();
        
        # First, get the aggeregates associated with this node
        my @aggregates = $class->{_conn}->Aggregates();
        
        foreach my $aggregate (@aggregates)
        {
                my @vol_objs = $aggregate->Volumes();
        
                foreach my $existent_vol (@vol_objs)
                {
                        my $vol_name = $existent_vol->name();
                        my $vol_status = $existent_vol->status();
                        logcomment("CMode Filer: get_status_of_all_volumes: Processing volume $vol_name: status:\n" . Dumper($vol_status)) if $class->{_debug};
                        # Need to process every key-value. Otherwise, upon return, status of every volume will be the ref to the last obtained status.
                        my @status_keys = qw(parent type state); # Some 7-mode keys are not available in c-mode (options status validation)
                        foreach my $key (@status_keys)
                        {
                                $return_hash{status}{$vol_name}{$key} = $vol_status->{$key};
                        } 
                }
        }
        
        $return_hash{rc} = C_RC_SUCCESS;
        
        return %return_hash;
}


# @name ::count_existent_volumes
# @summary Obtains the number of volumes existent on this filer.
# 
# @description
# Obtains the number of volumes existent on this filer.
#
# @return Hash with rc, count, and output (", "-separated list of volumes).
# 
# @arg state Optional: Volume state, to be used to filter the existent volumes.
# If specified, then the subroutine will return the number of volumes in the desired state.
# If not specified, then the count of all volumes is returned.
# Volume states are defined in Compression::Constants.
#
# @examples
#
# $obj->count_existent_volumes( state => C_VOL_STATE_ONLINE );
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub count_existent_volumes
{
	my $class = shift;
        my %args = validate(@_,
	{
		state   => { type => SCALAR, optional => 1, },	
	} );
        
        my $required_state = $args{state};
        my $vol_count = 0;
        my $vol_list = "";
        
        # First, get the aggeregates associated with this node
        my @aggregates = $class->{_conn}->Aggregates();
        
        if (!defined($required_state))
        {
                my @volumes = ();
                foreach my $aggr_obj (@aggregates)
                {
                        my @vol_list = $aggr_obj->volumes();
                        logcomment("Aggregate " . $aggr_obj->name() . ": volume list\n" . Dumper(@vol_list)) if $class->{_debug};
                        foreach my $vol (@vol_list)
                        {
                                $vol_count++;
                                push(@volumes, $vol);
                        }
                        
                }
                
                $vol_list = join(", ",@volumes);
                return (rc => C_RC_SUCCESS, count => $vol_count, output => $vol_list);
        }
        
        # Volume required to be in a specific state
        logcomment("Filer: count_existent_volumes: required state: $required_state") if $class->{_debug};

        foreach my $aggregate (@aggregates)
        {
                my @vol_objs = $aggregate->Volumes();
        
                foreach my $existent_vol (@vol_objs)
                {
                        my $vol_name = $existent_vol->name();
                        my $vol_state = $existent_vol->status("state");
                        logcomment("Filer: count_existent_volumes: Processing volume $vol_name: state: $vol_state") if $class->{_debug};
                        if ($vol_state eq $required_state)
                        {
                                $vol_count++;
                                $vol_list .= "$vol_name, ";
                                logcomment("Filer: count_existent_volumes: Match found ($vol_name): current count: $vol_count") if $class->{_debug};
                        }
                }
        }
        
        return (rc => C_RC_SUCCESS, count => $vol_count, output => $vol_list);
}


# @name ::reboot_and_wait
# @summary Reboots filer and waits until it comes up.
# 
# @description
# Reboots filer and waits until it comes up.
#
# @return Hash with rc and output (return of the NOTE::Filer::reboot).
# 
# @arg timeout Optional: Time in seconds to wait for the filer to come up.
# Defaults to C_DFLT_BOOTUP_WAIT_TIME (defined in Compression::Constants).
#
# @examples
#
# $obj->reboot_and_wait( timeout => 500 );
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub reboot_and_wait
{
	my $class = shift;
        my %args = validate(@_,
	{
		timeout  => { type => SCALAR, regex => qr/^\d+$/, optional => 1, default => C_DFLT_BOOTUP_WAIT_TIME, },	
	} );
        
        # GNOTE does not accept a timeout yet
        $class->{_conn}->reboot();
        
        return ( rc => C_RC_SUCCESS);
}


# @name ::powercycle_and_wait
# @summary Powercycles filer and waits until it comes up.
# 
# @description
# Powercycles filer and waits until it comes up.
#
# @return Hash with rc and output.
# 
# @arg new_state Optional: State to bring the filer in after powercycling (described in boot.pm).
# Defaults to 'UP'.
# 
# @arg timeout Optional: Time in seconds to wait for the filer to come up.
# Defaults to C_DFLT_BOOTUP_WAIT_TIME (defined in Compression::Constants).
#
# @examples
#
# $obj->powercycle_and_wait( new_state => 'BOOT_MENU', timeout => 500 );
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub powercycle_and_wait
{
	my $class = shift;
        my %args = validate(@_,
	{
                new_state => { type => SCALAR, optional => 1, default => 'UP', },
		timeout  => { type => SCALAR, regex => qr/^\d+$/, optional => 1, default => C_DFLT_BOOTUP_WAIT_TIME, },	
	} );
        
        my $new_state = $args{new_state};
        delete($args{new_state});
        
        if ($new_state ne 'UP')
        {
                logresult('script', "This method only supports new_state=UP for C-mode");
        }
        
        # GNOTE does not support a timeout yet for this method. For now ignore the timeout arg.
        # If timeout will be needed, we'll need to open a RFF
        $class->{_conn}->power_cycle();
        
        return ( rc => C_RC_SUCCESS, output => '' );
}

# @name ::panic_and_wait
# @summary Panics filer and waits until it comes up.
# 
# @description
# Panics filer and waits until it comes up.
#
# @return Hash with rc and output.
# 
# @arg new_state Optional: State to bring the filer in after panic (described in boot.pm).
# Defaults to 'UP'.
# 
# @arg timeout Optional: Time in seconds to wait for the filer to be brought in desired state.
# This time is an extra to the 320 sec of sleep after the panic (filer::filer_panic.pm).
# Defaults to C_DFLT_BOOTUP_WAIT_TIME (defined in Compression::Constants).
#
# @notes
# The normal behavior for a filer is to come back 'UP' after panic,
# without the library having to take any action. Specifying a 'new_state' other than 'UP'
# means that the filer's state succession will be panic->up->new_state.
#
# @examples
#
# $obj->panic_and_wait( new_state => 'BOOT_MENU', timeout => 500 );
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub panic_and_wait
{
	my $class = shift;
        my %args = validate(@_,
	{
                #new_state => { type => SCALAR, optional => 1, default => 'UP', }, # no c-mode support yet
		timeout  => { type => SCALAR, regex => qr/^\d+$/, optional => 1, default => C_DFLT_BOOTUP_WAIT_TIME, },
	} );
        
        #my $new_state = $args{new_state};
        #delete($args{new_state});

	# 03/08/2012 : Added to avoid coredump time, causing timeouts for CMode testcases
	$class->{_conn}->server->shell( cmd => "sudo sysctl debug.dont_dump_core=1");

        $class->{_conn}->panic();

        return ( rc => C_RC_SUCCESS, );
}

# @name ::wait_until_up
# @summary Waits until the filer comes up.
# 
# @description
# Waits until the filer comes up. Wrapper around NOTE::Filer::transit_to.
#
# @return Hash with rc and output.
# 
# @arg timeout Optional: Time in seconds to wait for the filer to be brought in desired state.
# This time is an extra to the 320 sec of sleep after the panic (filer::filer_panic).
# If not defined, boot::transit_to will use its own default value.
#
# @arg powercycle Optional: Boolean flag to decide if the filer should be power-cycled in case of failure
# (if transit_to fails once, then power-cycle the filer and try once more to bring it to desired state).
# This is useful, as sometimes transit_to fails to detect (without retrying) that the filer came up.
# Defaults to FALSE.
#
# @arg min_duration Optional: Minimum duration imposed to the "bring up" process.
# Useful if we want to ensure that the filer actually was down (for example, it panicked).
# If set to a value greater than 0, the test will fail if the filer was in an 'UP' state in less than min_duration.
# Defaults to 0.
#
# @examples
# $obj->wait_until_up( min_duration => 120 );
# $obj->wait_until_up( timeout => 500, powercycle => FALSE );
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub wait_until_up
{
	my $class = shift;
        my %args = validate(@_,
	{
		timeout  => { type => SCALAR, regex => qr/^\d+$/, optional => 1, default => C_DFLT_BOOTUP_WAIT_TIME},
                powercycle => { type => BOOLEAN, optional => 1, default => FALSE, },
                min_duration => { type => SCALAR, regex => qr/^\d+$/, optional => 1, default => 0 },
	} );
        
        my $host_name = $class->{host_name};
        my $min_duration = $args{min_duration};
        delete($args{min_duration});
        delete($args{powercycle}); # No c-mode support

        my $start_time = time();
        logcomment("Filer: wait_until_up: start time: $start_time") if $class->{_debug};

        logcomment("Calling GNOTE transit_to with the following arguments:\n" . Dumper(%args)) if $class->{_debug};
        
        # Have this in an eval block. This method will be called, in most of the cases, after a script-generated panic
        # The underlying library will fail when detecting the panic.
        # Ignore here the panic, as this is the 7-mode behavior.
        eval
        {
                $class->{_conn}->transit_to('UP', %args);
        };
        
        if (($@) && ($@ !~ /panicked/))
        {
                logresult('fatal', "Failure bringing the system up:\n" . $@);
        }
        
        # The first attempt will fail if the filer was purposely panicked,
        # but it will leave the filer still down. Try again
        $class->{_conn}->transit_to('UP', %args);
        
        $class->{_conn}->reconnect();

        my $end_time = time();
        my $elapsed_sec = $end_time - $start_time;
        logcomment("Filer: wait_until_up: end time: $start_time\n" .
                   "Elapsed time: $elapsed_sec seconds") if $class->{_debug};

        if (($min_duration > 0) && ($elapsed_sec < $min_duration))
        {
                # Stay around and check if filer ever went down, then fail
                my $current_time = time();
                my $timeout = $args{timeout};
                my $host_went_down = 0;
                while (($current_time - $start_time < $timeout) && ($host_went_down == 0))
                {
                        sleep(30);
                        my $ping_res = $class->{_conn}->ping();
                        logcomment("Filer: wait_until_up: ping returned $ping_res (current time: $current_time)") if $class->{_debug};
                        if ($ping_res == 0)
                        {
                                $host_went_down = 1;
                        }
                        $current_time = time();
                }
                if ($host_went_down == 1)
                {
                        my $total_time = $current_time - $start_time;
                        logcomment("Filer $host_name went down after $total_time seconds.");
                }
                else
                {
                        logcomment("Filer $host_name did not went down (waited $timeout seconds).");
                }
                logresult("fatal", "Filer $host_name was brought up in $elapsed_sec seconds, which is less than the minimum required of $min_duration");
        }

        return ( rc => C_RC_SUCCESS);
}


# @name ::set_fiji_point
# @summary Sets Fiji point on the filer.
# 
# @description
# Sets Fiji point on the filer.
#
# @return The Compression::Fiji object corresponding to this Fiji point.
# 
# @arg tag Required: Tag corresponding to the fiji point to be set.
# Fiji point tags are defined in Compression::Constants.
#
# @arg volname Optional: Name of volume, to be used for FIJI points enabled for a specific volume.
#
# @arg sleep_stage Optional: Sleep stage, to be used for the dense.sleep fiji points.
#
# @arg sleep_time Optional: Sleep time, to be used for the dense.sleep fiji points.
# Default: C_DFLT_FIJI_DENSE_SLEEP_TIME.
#
# @arg error_value Optional: Error value, to be used for the dense.compression_metafile.err fiji point
# (default C_FIJI_DFLT_REP_METAFILE_ERROR).
##
## @arg inode_num Optional: Inode num, to be used for the C_FIJI_TAG_SUSPEND_SCAN_64BIT_UPGRADE fiji point
## (required for this specific fiji).
##
## @arg level Optional: Error value, to be used for the C_FIJI_TAG_SUSPEND_SCAN_64BIT_UPGRADE fiji point
## (defaults to 0).
##
## @arg fbn Optional: Error value, to be used for the C_FIJI_TAG_SUSPEND_SCAN_64BIT_UPGRADE fiji point
## (required for this specific fiji and should be multiple of 510).
# 
# @examples
#
# my $fiji = $obj->set_fiji_point(tag => C_FIJI_TAG_DISABLE_CP);
# $fiji->get_slot();
# $fiji->check_if_active();
#
# @note
# This module will also keep track of any C_FIJI_TAG_DISABLE_CP/C_FIJI_TAG_ENABLE_CP that were set on the filer.
# This is to enable cleanup subroutine to re-enable the timer-triggered CP, if needed.
#
# @note
# This subroutine creates a new Compression::Fiji object and sets the Fiji point on the filer
# (calls Compression::Fiji->new and Compression::Fiji->set_point).
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub set_fiji_point
{
	my $class = shift;
        my %args = validate(@_,
	{
                tag => { type => SCALAR, },
                volname => { type => SCALAR, optional => 1, },
                sleep_stage => {type => SCALAR|UNDEF, optional => 1,},
                sleep_time => {type => SCALAR|UNDEF, optional => 1,},
                error_value => {type => SCALAR|UNDEF, optional => 1,},
                inode_num => {type => SCALAR|UNDEF, optional => 1,}, 
                level => {type => SCALAR|UNDEF, optional => 1,},
                fbn => {type => SCALAR|UNDEF, optional => 1,},
	} );
        
        my $fiji_tag = $args{tag};
        my $host_name = $class->{host_name};
        
        $args{conn} = $class->{_conn};
        $args{ui} = $class->{_ui};
        $args{boot_mode} = C_ONTAP_CMODE;

        my $fiji = Compression::Fiji->new(%args);

        $fiji->set_point();

        if ($fiji_tag eq C_FIJI_TAG_DISABLE_CP)
        {
                logcomment( "Filer: Disabled timer-triggered CP") if ($class->{_debug});
                $Compression::Filer::CP_is_disabled->{$host_name} = TRUE;
        }
        elsif ($fiji_tag eq C_FIJI_TAG_ENABLE_CP)
        {
                logcomment( "Filer: Enabled timer-triggered CP") if ($class->{_debug});
                $Compression::Filer::CP_is_disabled->{$host_name} = FALSE;
        }

        logcomment( "Filer ($host_name): set_fiji_point: Dumping CP_is_disabled:\n" . Dumper($Compression::Filer::CP_is_disabled)) if ($class->{_debug});

        return $fiji;
}


# @name ::sync_blocks_to_disk
# @summary Syncs dirty FS blocks to disk.
# 
# @description
# Syncs dirty FS blocks to disk. Wrapper around NOTE::Filer::sync.
#
# @return Hash with rc and output (void).
# 
# @arg wait Optional: Wait for sync to complete. Defaults to TRUE.
#
# @examples
#
# $obj->sync_blocks_to_disk();
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub sync_blocks_to_disk
{
	my $class = shift;
        my %args = validate(@_,
	{
                wait => { type => SCALAR, optional => 1, default => TRUE, },
	} );
        
        # No available functionality in GNOTE yet
        my $cmd_code = C_CMD_SYNC;
        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,);
	
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => C_ONTAP_CMODE,
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
	
	return $class->{_ui}->process_output_string(code => $cmd_code,
						    output => $output,
						    expected_rc => C_RC_SUCCESS,
						    log_message => "Sync blocks to disk",);

        return ( rc => C_RC_SUCCESS, output => '' );
}


# @name ::reconnect
# @summary Reconnects to the filer with specified connection method.
# 
# @description
# Reconnects to the filer with specified connection method.
#
# @return Hash with rc and output (void).
# 
# @arg mode Optional: The connection method to use.
#
# @examples
#
# $obj->reconnect( mode => '#console' );
# 
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub reconnect
{
	my $class = shift;
        my %args = validate(@_,
	{
                mode => { type => SCALAR, optional => 1, },
	} );
        
        if (defined($args{mode}))
        {
                # For backwards compatibility with 7-mode, allow mode to be '#...'
                my $mode = $args{mode};
                delete($args{mode});
                $mode =~ s/#//g;
                $args{list} = $mode;
        }

        $class->{_conn}->reconnect( %args );

        return ( rc => C_RC_SUCCESS, output => '' );
}


# @name ::get_root_vol_name
# @summary obtains the name of root volume
#
# @description
# Obtains the root volume name
# 
# @return root volume name
#
# @examples
#
# $root_vol_name = $obj->get_root_vol_name();
#
# @author dl-compression-qa@netapp.com
# 
sub get_root_vol_name
{
        my $class = shift;
        
        my $root_vol_name;
        
        # First get the name of the aggregate hosting the mroot volume
        my $root_aggr_name = $class->get_root_aggr_name();
        my @root_aggr_list = $class->{_conn}->Aggregates(aggregate => $root_aggr_name); # only one elem
        my $root_aggr_obj = shift(@root_aggr_list);
        my @root_aggr_volumes = $root_aggr_obj->volumes();
        if (scalar(@root_aggr_volumes) == 1)
        {
               $root_vol_name = shift(@root_aggr_volumes);
               return $root_vol_name ;
        }
        else
        {
                # Loop through the volume list and get dblade-level vol status
                foreach my $vol (@root_aggr_volumes)
                {
                        my $cmd_code = C_CMD_DETECT_ROOT_VOL;
                        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                                       volume => $vol);
                        
                        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                                            boot_mode => C_ONTAP_CMODE,
                                                                            cmd => $command,
                                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
                        
                        my %processed_hash = $class->{_ui}->process_output_string(code => $cmd_code,
                                                                                  output => $output,
                                                                                  expected_rc => C_RC_FAILURE_NON_FATAL,
                                                                                  log_message => "Obtain dblade-level volume status for $vol",);
                        
                        if ($processed_hash{rc} == C_RC_ROOT_VOLUME_DETECTED)
                        {
                                return $vol;
                        }
                }
        }

        # We were not able to detect the mroot volume if we got here
        logresult('fatal', "Unable to detect the mroot volume name");
}

# @name ::get_root_aggr_name
# @summary obtains the name of root aggregate
#
# @description
# Obtains the root aggregate name
#
# @return root aggregate name
#
# @examples
#
# $root_aggr_name = $obj->get_root_aggr_name();
#
# @author dl-compression-qa@netapp.com
#
sub get_root_aggr_name
{
        my $class = shift;
        
        my $mroot_aggr = $class->{_conn}->server->storage_aggregate_show('has-mroot' => 'true',
                                                                         'nodes' => $class->{host_name});
        
        my $root_aggr_name = $mroot_aggr->[0]->{aggregate};
        logcomment("Root aggregate is $root_aggr_name \n") if ($class->{_debug});
        return $root_aggr_name ;
}

# @name ::set_flag
# @summary Sets the required filer option.
#
sub set_flag
{
        my $class = shift;
        my %args = validate(@_,
	{
                code => { type => SCALAR, },
                value => { type => SCALAR, },
                privilege => { type => SCALAR, optional => 1, default => C_PRIV_DIAG },
	} );
        
        my $flag_name = $class->{_ui}->get_constant_value(constant_code => $args{code});
        my $value_to_set = $args{value};
        
        # TODO: GNOTE does not have a way to pass in the privilege,
        # It sets it to test always. File RFF?
        logcomment("Setting flag: $flag_name to ") if ($class->{_debug});
        $class->{_conn}->setflag(name => $flag_name, value => $value_to_set);
        
        # No errors if we got here
        return (rc => C_RC_SUCCESS);
}

# @name ::wait_for_console_message
# @summary Tails the messages file until the required messages are seen.
# 
# @description
# Tails the messages file until the required messages are seen.
#
# @arg message_codes Required: Arrayref of codes corresponding to the
# message strings to be serched for in the messages file (defined in Constants.pm).
#
# @arg timeout Optional: Maximum allowed time, in seconds, to wait for match.
# Default: C_DFLT_CONSOLE_WAIT_TIMEOUT (defined in Constants.pm).
#
# @arg match_type Optional: Type of match to be performed. Legal values: ANY, ALL.
# Default: ALL.
#
# @return Hash with rc and output.
#  
# @examples
# $obj->wait_for_console_message(message_codes => [C_CONSOLE_OUT_OF_SPACE_SFSR_ABORTED,
#                                                  C_CONSOLE_INSUFFICIENT_SPACE_FOR_SFSR],
#                                match_type => 'ANY');
#
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private 
#
sub wait_for_console_message
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        timeout => { type => SCALAR|UNDEF, optional => 1, default => C_DFLT_CONSOLE_WAIT_TIMEOUT, },
                        message_codes => { type => ARRAYREF, },
                        match_type => { type => SCALAR, regex => qr/^ANY|ALL$/, default => 'ALL', },
                } );
        
        my $timeout = $args{timeout};
        my $msg_codes = $args{message_codes};
        my $match_type = $args{match_type};
        my $host_name = $class->{host_name};
        
        my $ui = $class->{_ui};
        my $patterns = $ui->{console_messages}->get_message_format(message_codes => $msg_codes);
        # We need to reconnect here, for any bg subtests. Otherwise, the output gets clogged
        $class->{_conn}->reconnect();
        my $tail = LogReader::MlogMessages->new($class->{_conn}->server());
        my $required_match_count = scalar(@$patterns);
        my $matches = {};
        my $output = '';
        my $sleep_interval = 5;
        my $done = 0;
        
        logcomment("Waiting for the following strings in the messages file:\n" . Dumper($patterns));
        $tail->start();
        while ($done == 0)
        {
                $tail->stop();
                my $lines = $tail->to_log();
                # remove what MlogMessages prepends, keep only messages
                $lines =~ s/^\nMessages Log:\n\n//;
                if ($lines ne '')
                {
                        $output .= $lines;
                        logcomment("$lines");
                }
                logcomment("timeout=$timeout; lines:\n" . Dumper($output)) if ($class->{_debug});
                
                sleep($sleep_interval);
                $timeout -= $sleep_interval;
                
                if ($timeout < 0)
                {
			logresult('fatal', "\n====================================================================\n" .
                                           "Timed-out while waiting for required string(s) in the messages file.\n" .
                                           "Timeout: " . $args{timeout} . " seconds\n" .
                                           "====================================================================\n" .
				           "Waited for $match_type of the strings:\n" . Dumper($patterns) . "\n" .
                                           "====================================================================\n" .
                                           "Matches found:\n" . Dumper(keys(%$matches)) . "\n" .
                                           "====================================================================\n" .
                                           "Accumulated output:\n$output");
		}
                
                foreach my $pattern (@$patterns)
                {
                        if (exists($matches->{$pattern}))
                        {
                                # Already found
                                next;
                        }
                        my $pattern_matches = $tail->matches($pattern);
                        if (scalar(@$pattern_matches) > 0)
                        {
                                $matches->{$pattern} = 1;
                        }
                }
                
                my $matches_count = scalar(keys(%$matches));
                if (($match_type eq 'ANY') && ($matches_count > 0))
                {
                        $done = 1;
                }
                elsif ($matches_count == $required_match_count)
                {
                        $done = 1; 
                }
                else
                {
                        $tail->start();
                }
        }
        
        return (rc => C_RC_SUCCESS, output => $output);
}


# @name ::initialize_compliance_clock
# @summary Initializes the filer's compliance clock.
#
# @description
# Initializes the filer's compliance clock.
#
# @return Hash with rc.
#
# @examples
# $obj->initialize_compliance_clock();
#
#
# @author voicu@netapp.com, dl-compression-qa@netapp.com
#
sub initialize_compliance_clock
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        # None
                } );
        
        logresult('script', "Snaplock is not supported on C-mode");
}


## @name ::get_system_time
## @summary Gets the time on this filer. 
##
## @description
## Gets the time on this filer. 
##
## @return Hash with rc, time (integer), gmt (Unix-format gmt).
##
## @examples
## $obj->get_system_time();
##
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_system_time
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        # None
                } );
        
        my $filer_name = $class->{host_name};
        my @filer_date = $class->{_conn}->get_date();
	my $utc_date_and_time = $filer_date[0]->{utcdateandtime};
	$utc_date_and_time =~ s/\"//g;
	
	my $filer_date = Date::Manip::UnixDate( $utc_date_and_time, '%s' );
	my $gmt = gmtime($filer_date);
        
        logcomment("\nFiler: $filer_name\nDate: $filer_date (GMT: $gmt)");
        return (rc => C_RC_SUCCESS, time => $filer_date, gmt => $gmt);
}


################################
# Internal subroutines
################################
sub _require_standalone_filer
{
        my $class = shift;
        my $cluster_status = 0;
        
        # sfo_enable will fail if service is not configured.
        # We still need to consider this case as a $cluster_status == 0
        eval { $cluster_status = $class->{_conn}->sfo_enable(); };
        if ((!$@) && ($cluster_status == 1))
        {
                logcomment("Disabling clustering for filer " . $class->{host_name});
                # First check if ha is configured
                $class->{_ha_was_configured} = $class->{_conn}->ha_configure();
                if ($class->{_ha_was_configured} == TRUE)
                {
                        $class->{_conn}->server->cluster_ha_modify( configured => 'false');
                }
                $class->{_conn}->sfo_enable(0);
                $class->{_cluster_enable_nedded} = TRUE;
        }
}

# Internal method
# Checks if the current kernel on this filer is debug and
# resets the value of the internal class attribute _has_debug_kernel.
sub _set_debug_kernel
{
	my $class = shift;
	
	my %build_info = $class->{_conn}->build_info();
	my $node_build_type = $build_info{TYPE};
	$node_build_type =~ s/^\s*//;
	$node_build_type =~ s/\s*$//;
	
	$class->{_build_type} = $node_build_type;
	
	my $cmode_build_type_debug = C_10MODE_BUILD_TYPE_DEBUG;
	if ($class->{_build_type} =~ /$cmode_build_type_debug/)
	{
		$class->{_has_debug_kernel} = TRUE;
	}
	else
	{
		$class->{_has_debug_kernel} = FALSE;
	}
}

# This method is used to decide what licensing is needed for compression
sub _license_compression
{
        my $class = shift;
        my $license = Compression::Filer::License->new(version => $class->{version},
                                                       boot_mode => C_ONTAP_CMODE);

        my $needed_licenses = $license->get_licensing_info();

        logcomment("Setting up the filer with the required licenses for compression.\n" .
                   "Licenses needed:\n" . Dumper($needed_licenses)) if ($class->{_debug});
        
        my $temp_license_cmds = {'a_sis' => "license add OSFYRVH",
                                  'compression' => "license add QTEUWOJ",};

        my $srvr = $class->{_conn}->server();
        foreach my $license_code (@$needed_licenses)
        {
                logcomment("Installing license: $license_code") if ($class->{_debug});
                
                # TODO: this is just a temporary solution, until the SE licensing is covered by the base license
                # At that point, only the base license should be listed in License.pm and installing the license could be done through license_add
                #$class->license_add(service => $license_code);
                $srvr->system_node_run(node => 'local',
                                       command => $temp_license_cmds->{$license_code},);
        }
        
        # TODO: this should be a temporary solution also
        # Check if any 'licensed_feature' option needs to be on
        my $need_licensed_option = (defined($main::LICENSED_FEATURE_OPTION)) ? $main::LICENSED_FEATURE_OPTION : FALSE;
        if ($need_licensed_option == TRUE)
        {
                my $needed_licensed_option = $license->_get_licensed_feature_option_info();
                
                foreach my $license_option (@$needed_licensed_option)
                {
                        logcomment("Setting option: $license_option") if ($class->{_debug});
                        $srvr->system_node_run(node => 'local',
                                               command => "\"priv set -q advanced; registry set options.$license_option on\"",);
                }
        }
}

# This method is used for providing backwards compatibility with 7-mode licensing codes
sub _get_cmode_license_code
{
        my $class = shift;
        my (%args) 	= validate( @_,
		{
                        service => { type => SCALAR, },
		} );
        
        my $service = $args{service};
        
        my $cmode_codes = [];
        
        if ($service eq C_LICENSE_NFS)
	{
		$cmode_codes = ['NFS'];
	}
	elsif ($service eq C_LICENSE_CIFS)
	{
		$cmode_codes = ['CIFS'];
	}
	elsif ($service eq C_LICENSE_SNAPMIRROR)
	{
		$cmode_codes = ['SnapMirror_DP', 'SnapMirror_LS'];
	}
	elsif ($service eq C_LICENSE_SNAPRESTORE)
	{
		$cmode_codes = ['SnapRestore'];
        }
	elsif ($service eq C_LICENSE_CMODE_BASE)
	{
		$cmode_codes = ['Base'];
	}
        
        return $cmode_codes;
}


# Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests start to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
        
        my $host_name = $class->{host_name};
        
	if ($class->{_creator_pid} == $$)
	{
                logcomment("Executing CMode Filer clean-up for host $host_name");
     
                logcomment("Filer: cleanup for host $host_name :Dumping CP_is_disabled:\n" .
                           Dumper($Compression::Filer::CP_is_disabled)) if ($class->{_debug});
                my $cp_is_disabled_flag = $Compression::Filer::CP_is_disabled->{$host_name};
                if (defined($cp_is_disabled_flag) && ($cp_is_disabled_flag == TRUE))
                {
                        logcomment( "Filer: $host_name: cleanup: Timer-triggered CP was disabled. Re-enabling...") if ($class->{_debug});
                        $class->set_fiji_point(tag => C_FIJI_TAG_ENABLE_CP);
			# Wait for 1 cp before NOTE will clear the fiji point
			my $wait_time = ONE_CP_TIME + 1;
			sleep($wait_time);
                }
                
                if ($class->{_cluster_enable_nedded} == TRUE)
                {
                        logcomment( "Filer: $host_name: cleanup: CF was disabled during this test. Re-enabling...") if ($class->{_debug});
                        $class->{_conn}->sfo_enable(1);
                        $class->{_cluster_enable_nedded} = FALSE;
                        if ($class->{_ha_was_configured} == TRUE)
                        {
                                $class->{_conn}->server->cluster_ha_modify( configured => 'true');
                        }
                }        
        }
	else
	{
                logcomment( "CMode Filer: $host_name: cleanup: nothing done: script PID: " . $$) if ($class->{_debug});
	}
        
        if (defined($execution_error))
        {
                _c_interpret_logresult($execution_error);
        }
}


sub testme
{
	package main;
	use Compression::Constants qw(:all);
	use Compression::Compress;
	use Compression::Decompress;
        use filer;
	use vars qw($FILER);

	require "Compression/setup_core.pl";

	logcomment("Filer: testme: Filer=$FILER");

        my $obj = Compression::Filer->new( filer => $FILER );
        $obj->license_add( service => C_LICENSE_SERVICE_DEPENDENCY );
        $obj->license_add( service => C_LICENSE_COMPRESSION );
        $obj->set_option_on_all_volumes( service => C_OPTION_COMPRESSION, new_value => FALSE, );
        $obj->license_remove( service => C_LICENSE_COMPRESSION );
        $obj->license_remove( service => C_LICENSE_SERVICE_DEPENDENCY );
        
        logcomment("Filer: testme: done");
}


1;
