#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/CMode/SFO.pm#1 $
#
## @summary Compression C-mode SFO routines 
##
## @description
##  This library provides functions specific
##	to c-mode sfo
##
## @author dl-compression-qa@netapp.com, voicu@netapp.com
## @status private (internal)


package Compression::CMode::SFO;

use strict;
use warnings;
use Tharn;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Compression::Constants qw(:all);

BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}


# @name ::new
#
# @summary Class constructor: perform cluster initialization
#
# @arg filer Required: primary filer name
# @arg debug Optional: logs additional debug messages specific to this module
# @arg initialize Optional: Initialize clustering if needed (license, cf enable).

sub new
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		filer  => { type=>SCALAR },
		debug  => { type=>SCALAR, optional => 1, default=>FALSE },
		initialize => { type => BOOLEAN, optional => 1, default => TRUE },
	});
	
	my $self = {};
	$self->{host_name} = $args{filer};
	
	$self->{ID} = $$;
	$self->{_debug} = $args{debug};
	
	$self->{_node} = NOTE::GX::Node->connect(name => $self->{host_name});
	if ($args{initialize} == TRUE)
	{
		$self->{_node}->sfo_enable(1);
	}
	
	$self->{partner_name} = $self->{_node}->sfo_status->{'partner-name'};
	
	$self->{_version} = Compression::Version->new(conn => $self->{_node},
						      debug => $self->{_debug},
						      boot_mode => C_ONTAP_CMODE);
	
	$self->{_ui} = Compression::UI->new(version => $self->{_version},
                                            debug => $self->{_debug},
                                            boot_mode => C_ONTAP_CMODE); 
	
	bless $self, $class;
	
	return $self;
}


# @name ::takeover
#
# @summary Initiates a takeover
#
# @arg me Optional: if TRUE, partner takes over
# @arg background Optional: run as a background process

sub takeover
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		#me => { type=>SCALAR, optional => 1, default=>FALSE }, TODO
		background => { type=>SCALAR, optional => 1, default=>FALSE },
		state 	   => { type => SCALAR, optional => 1, default => 'In takeover' },
	});

	$class->{_node}->takeover(%args);
}

# @name ::giveback
#
# @summary Initiates a giveback
#
# @arg background	Optional: run as a background process
# @arg force		Optional: use the "-f" force option
# Default: TRUE (such that 'override-vetoes' will be set to true (needed in case of failed disks on the node)).
# @arg safe		Optional: if TRUE, giveback will not croak if the the current state of the CFO pair is NOT taken over

sub giveback
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		background => { type => SCALAR, optional => 1, default => FALSE },
		force => { type => SCALAR, optional => 1, default => TRUE, },
		#safe => { type=>SCALAR, optional => 1, default=>FALSE },
		state 	   => { type => SCALAR, optional => 1, default => 'Connected to \S+' },
		} );

	logcomment("Wait for the partner to come in 'Waiting for giveback' state");
	my $count = 0; # Wait for a consistent state, 
		       # - verify the state is stable by checking it gives consistent value for 3 consequtive checks
	my $partner_state = $class->{_node}->sfo_status('partner-state');
	logcomment("Partner State: $partner_state");
	while (!(($partner_state =~ /Waiting for giveback/) && ($count > 2))) {
		sleep 10;
		$partner_state = $class->{_node}->sfo_status('partner-state');
		logcomment("Partner State: $partner_state");
		$count++ if ($partner_state =~ /Waiting for giveback/);
	}
	
	$class->{_node}->giveback(%args);
}


# @name ::get_status
#
# @return
# A hash with status for "me" and "partner".
# Status can be 'UP', 'TAKEOVER', 'DEAD', 'UNREACH', 'DISABLED'

sub get_status
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		# None
	});
	
	# TODO: complete this list 
	my $known_states = {'UP' => 'FAILOVER_STATE_UP',
			    'TAKEOVER' => 'FAILOVER_STATE_TO',
			    'DEAD' => 'FAILOVER_STATE_REBOOT',
			    };
	
	my $status = $class->{_node}->sfo_status();
	my $me_state = $status->{'local-firmware-state'};
	my $partner_state = $status->{'partner-firmware-state'};
	
	# Best effort to translate the states to be backward-compatible with 7-mode
	foreach my $known_state (keys %$known_states)
	{
		my $state_string = $class->{_ui}->get_constant_value(constant_code => $known_states->{$known_state});
		if ($state_string eq $me_state)
		{
			$me_state = $known_state;
		}
		
		if ($state_string eq $partner_state)
		{
			$partner_state = $known_state;
		}
	}
	
	logcomment("\nFailover status:\n" .
		   "me: $me_state\n" .
		   "partner: $partner_state\n");
	
	return (me => $me_state, partner => $partner_state, output => $status);
}

# @name ::monitor
#
# @summary monitor

sub monitor
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		me => { type=>SCALAR },
		partner => { type=>SCALAR },
	});
	
	my $me_state = $args{me};
	
	my $supported_states = {'TAKEOVER' => 'In takeover',
				'UP' => 'Connected to \S+'};
	
	# TODO: looks like GNOTE is only monitoring the 'me' node
	# Do we need to monitor the partner too?
	# For now, ignore the partner argument
	
	my $cmode_state = $supported_states->{$me_state};
	if (!defined($cmode_state))
	{
		logresult('script', "State $me_state is not supported for C-mode failover monitoring");
	}
	
	$class->{_node}->sfo_monitor(state => $cmode_state);
}


1;
