#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/CMode/SnapMirror.pm#1 $
#
## @summary Compression CMode SnapMirror routines
##  
##
## @description
## Compression CMode SnapMirror routines
##
## @author dl-compression-qa@netapp.com, weili@netapp.com
## @status private (internal)

package Compression::CMode::SnapMirror;

use strict;
use warnings;
use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;


BEGIN
{
	require "Compression/base_import.pl";
}

use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult);
use Carp qw(croak);  


## @name:: create
## @summary create a SnapMirror relationship
##	
##
## @description
##	Create a 1 x 1 or 1 x N snapmirror relationship(s) involving a combination of DP and LS mirrors.
##
## @return a Object of type Compression::CMode::SnapMirror.
##
## @arg src Required: source volume object(Compression::Compress), or an array reference to an array of these objects 
## @arg dp Optional: dp type destination volume object (Compression::Compress), or an array reference to an array of these objects
## @arg ls Optional: dp type destination volume object (Compression::Compress), or an array reference to an array of these objects
##  at least one of dp or ls arguments must exists
## @arg debug Optional: set debug
## @arg expected Optional: expected return, default C_RC_SUCCESS
##
## @example
## my $sm_obj = Compression::CMode::SnapMirror->create( src => $src_vol,
##							dp  => $dst_vol, );
## my $sm_obj = Compression::CMode::SnapMirror->create( src => $src_vol,  
##   		                                       	dp  => \@dp_volumes,
##                                              	ls  => \@ls_volumes, );
##                                               
sub create
{
	logcomment ("Start create snapmirror object, Compression::CMode::SnapMirror");
	my $class = shift;
	my $self = {};
	
	my (%args) = validate (@_, {src 	=> { type =>OBJECT, isa => "Compression::Compress" },
				    dp 		=> { type =>OBJECT|ARRAYREF, optional => 1,  },
				    ls 		=> { type =>OBJECT|ARRAYREF, optional => 1,  },
				    xdp         => { type =>OBJECT|ARRAYREF, optional => 1,  },
				    debug 	=> { type =>SCALAR, optional => 1, default=>FALSE },
				    expected 	=> { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },
	} );

	if ( (! defined $args{dp} ) && (! defined $args{ls} ) && (! defined $args{xdp} )  ) {
		logresult('script', "Must have at least 1 destination object.");
	}
	
	my $src = $args{src};
	delete($args{src});		
	my $dst_dp = $args{dp};
	delete($args{dp});
	my $dst_ls = $args{ls};
	delete($args{ls});
	my $dst_xdp = delete($args{xdp});
	$self->{_xdp} = FALSE;
	
	my $src_obj = $src->{volume}->_get_underlying_vol_obj();
	my $src_compr_obj = $src;

	my @dst_dp_volume;
	if ( defined $dst_dp){	
		@dst_dp_volume = $class->_get_volume_objects(dst=>$dst_dp);
	}
		
	my @dst_ls_volume;
	if ( defined $dst_ls){
		@dst_ls_volume = $class->_get_volume_objects(dst=>$dst_ls);
	}
	
	# TODO: For now, SnapMirrorSet does not support xdp destinations.
	# To work around this, we are using SnapMirror directly for the moment.
	# As a result, not all methods of this class will work for xdp (ex: attach).
	# Once the dpg team adds xdp support in SnapMirrorSet, we will switch back to using it for all three kinds of relationships.
	# This should also eliminate the current restrictions associated with xdp (only one dest volume, no dp/ls).
	# Also, the use of $self->{_xdp} should no longer be needed (and the related logic in verify_last_transfer_size).
	my @dst_xdp_volume;
	if (defined($dst_xdp)) {
		if ((defined $args{dp}) || (defined $args{ls})) {
			logresult('script', "SnapMirror XDP is not yet supported along with other destination types");
		}
		@dst_xdp_volume = $class->_get_volume_objects(dst=>$dst_xdp);
		if (scalar(@dst_xdp_volume) != 1) {
			logresult('script', "SnapMirror XDP is not yet supported for more than one destination volume");
		}
		$self->{_xdp} = TRUE;
	}

	$self->{ID} 	= $$;
	$self->{_debug}	= $args{debug};
	delete($args{debug});
	# This assumes that the command will execute on the source filer
	# or that both src and dst are running the same version.
	$self->{_ui} = $src_compr_obj->{_ui};
			
	my $sm_obj;	
	eval {
		if (defined($dst_xdp)) {
			my $dst_xdp_obj = shift(@dst_xdp_volume);
			$sm_obj = NOTE::GX::SnapMirror->create(src => $src_obj, dst => $dst_xdp_obj, type => 'XDP', cmd_only => 1);
		}
		else {
			$sm_obj = NOTE::SnapMirrorSet->create(src=>$src_obj,dp=>\@dst_dp_volume, ls=>\@dst_ls_volume, %args );
		}
	};
		
	$src_compr_obj->{_ui}->process_output_string(code 	 => C_CMD_SNAPMIRROR_INITIALIZE,
						     output 	 => $@,
						     expected_rc => $args{expected},
						     log_message => "SnapMirror create",);
	
	if (!defined($sm_obj)) {
		logresult('fatal', "Undefined object returned by NOTE\:\:SnapMirror\:\:initialize");
	}
		
	$self->{_src} = $src;
	$self->{_dst_dp} = $dst_dp;
	$self->{_dst_ls} = $dst_ls;
	$self->{_snapmirror} = $sm_obj;

	bless($self, $class);

	logcomment("SnapMirror: Setup Complete..\n\n") if ($self->{_debug});
	return $self;
}

## @name:: initialize
## @summary initialize a SnapMirror relationship
##	
## @description
##	initialize a SnapMirror relationship
##
## @return Hash with rc and output.
##
## @arg background Optional: Initialize snapmirror in background
## @arg timeout Optional: timeout, default 900
## @arg expected Optional: expected return, default C_RC_SUCCESS
##
## @example
## my $obj = Compression::SnapMirror->initialize (background => 1);
##
sub initialize
{
	my $class = shift;
	my (%args) = validate (@_, {background 	=> { type=>BOOLEAN, optional => 1,  default=>0},
				    timeout	=> { type=>SCALAR, optional => 1, default=>900 },
				    expected 	=> { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },
	});
	
	# GNOTE seems to have issues with setting the foreground option, so set it ourselves
	if ($args{background} == 0) {
		$args{foreground} = 'true';
	}
	eval {
		$class->{_snapmirror}->initialize(%args);
	};
	
	my $output = $@;
	logcomment("SnapMirror: Initialize Complete..\n\n") if ($class->{_debug});
	
	my $res = $class->{_ui}->process_output_string(code => C_CMD_SNAPMIRROR_INITIALIZE,
						     output => $output,
						     expected_rc => $args{expected},
						     log_message => "SnapMirror initialize",);

	return $res;

}
 
## @name:: update
## @summary SnapMirror update
##	
##
## @description
##	Starts an incremental transfer for SnapMirror relationships
##
## @return Hash with rc and output.
##
## @arg dst Optional: destination volume object, if given, will update only this mirror
## @arg type Optional: mirror type, if set to 'ls', will update all ls mirrors
## @arg source-snapshot Optional: snap-shot name,  if given, will update this snapshot
## @arg background Optional: if set, runs update in the background, default 0
## @arg expected Optional: expected return code, default C_RC_SUCCESS
##
## @example
# Update all destinations in this 1 x N snapmirror relationships
##  $obj->update();
##
## Update only the mirror with $dst_vol destination volume
##  $sms->update(dst=>$dst_vol);
##
## Update all LS mirrors, using 'update-ls-set' cmnd for multiple LS mirrors
##  $sms->update(type=>"ls");
##
## Update in background and wait for update to complete
## $obj->update(background=>1);
## $obj->monitor();
sub update
{
	my $class = shift;
	my %args = validate (@_, {
				dst 		  => { type => SCALAR, optional => 1 },
				type		  => { type => SCALAR, optional => 1 },
				'source-snapshot' => { type => SCALAR, optional => 1 },
				background 	  => { type => SCALAR, optional => 1, default => FALSE },
				expected 	  => { type => SCALAR, optional => 1, default => C_RC_SUCCESS, },
	} );
	
	# GNOTE seems to have issues with setting the foreground option, so set it ourselves
	if ($args{background} == 0) {
		$args{foreground} = 'true';
	}
	
	eval {
		$class->{_snapmirror}->update(%args);
	};
	
	return $class->{_ui}->process_output_string(code => C_CMD_SNAPMIRROR_UPDATE,
						    output => $@,
						    expected_rc => $args{expected},
						    log_message => "SnapMirror update",);
}

## @name:: resync
## @summary SnapMirror resync
##
## @description
##	snapmirror resync.
##
## @return Hash with rc and output.
##
## @example
## $obj->resync();
sub resync
{
	my $class = shift;
	my (%args) = validate (@_, { } );

	$class->{_snapmirror}->resync();
	return (rc => C_RC_SUCCESS);
}


## @name:: quiesce
## @summary SnapMirror quiesce
##	
##
## @description
##	Quiesce SnapMirror
##
## @return Hash with rc and output.
##
## @example
## $obj->quiesce();
sub quiesce
{
	my $class = shift;
	my (%args) = validate (@_, { } );
	
	$class->{_snapmirror}->quiesce();
	return (rc => C_RC_SUCCESS);
}

## @name:: resume
## @summary SnapMirror resume
##	
##
## @description
##	Resumes quiesced transfers.
##
## @return Hash with rc and output.
##
## @example
## $obj->resume();
sub resume
{
	my $class = shift;
	my (%args) = validate (@_, { } );

	$class->{_snapmirror}->resume();
	return (rc => C_RC_SUCCESS);
}

## @name:: abort
## @summary SnapMirror abort
##	
##
## @description
##	snapmirror abort.
##
## @return Hash with rc and output.
##
## @example
## $obj->abort();
sub abort
{
	my $class = shift;
	my (%args) = validate (@_, { } );

	$class->{_snapmirror}->abort();
	return (rc => C_RC_SUCCESS);
}


## @name:: break
## @summary SnapMirror break
##	
##
## @description
##	Break SnapMirror relationship
##
## @arg background Optional: Flag to execute in the background. Default 0
##
## @return Hash with rc and output.
##
## @example
## $obj->break();
sub break
{
	my $class = shift;
	my (%args) = validate (@_, {background => {type => SCALAR, optional => 1,},} );

	$class->{_snapmirror}->break(%args);
	return (rc => C_RC_SUCCESS);
}

## @name:: delete
## @summary SnapMirror delete
##	
##
## @description
##	Delete SnapMirror relationship
##
## @return Hash with rc and output.
##
## @example
## $obj->delete();
sub delete
{
	my $class = shift;
	my (%args) = validate (@_, { } );

	$class->{_snapmirror}->delete();
	return (rc => C_RC_SUCCESS);
}



## @name:: promote
## @summary promot the snapmirror relationship
##
## @return new promoted snapmirror object.
##
## @description
##	Performs a 'snapmirror promote'. 
##
## @arg dst Required: the promoted destination object (Compression::Compress)
## 
## @return Hash with rc and output.
##
## @example
## $obj->promote();
sub promote
{
	my $class = shift;
	my (%args) = validate (@_, { dst => { type =>OBJECT, isa => "Compression::Compress" },
				    });

	my $name = $args{dst}->{volume}->_get_underlying_vol_obj()->name ;
	my $mirror = $class->_mirror( name => $name );
	$class->{_snapmirror}->{_src} = $mirror->promote(dst=>$name);

	return $class;
}

## @name:: modify
## @summary Modify the snapmirror relationship
##	
## @description
##	Modify snapmirror relationship
##
## @return Hash with rc and output.
##
## @example
## $obj->modify( schedule => $my_sched );
sub modify
{
	my $class = shift;
	my %args = validate (@_, { 
		schedule => { type=>SCALAR, optional => 1},
	} );

	$class->{_snapmirror}->modify(%args);
	return (rc => C_RC_SUCCESS);
}

## @name:: schedule_create
## @summary Create a schedule for snapmirror
##	
##
## @description
##	Create a job for snapmirror schedule
##
## @return a job.
##
## @arg filer_src Required: Source filer name
## @arg name Required: schedule's name
## @arg schedule Required: the schedule
##
## @example
## 
## $sm_obj->schedule_create(	filer_src=>$FILER_SRC,
##				schedule_name=> 'my_sched',
##				sched=>'0,5,10,15,20,25,30,35,40,45,50,55' ); 
## $sm_obj->modify( schedule => 'my_sched' );  #asociates sched to the snapmirror object
sub schedule_create
{
	my $class = shift;
	my (%args) = validate (@_, { 
		filer_src => { type=>SCALAR },
		name => { type=>SCALAR },
		schedule => { type=>SCALAR },
	} );
	
	my $cluster = NOTE::GX::Cluster->connect( Node => $args{filer_src} );
	my $job = NOTE::GX::Job->new(Cluster => $cluster);
	$job->schedule_add(name   => $args{name},
			   minute => $args{schedule} );
	$class->{_job}= $job;
	return $job;
}

## @name:: schedule
## @summary SnapMirror schedule 
##	
##
## @description
##	Get, Clear or Restore the mirror schedule that was set during mirror creation or with the 'modify' method.
##	The 'clear' option also records the schedule locally for later 	'restore'. 
##
## @arg action Optional: Default get 
## 	action=>"clear" 	- Remove schedule from filer and record it in this object.
##	action=>"restore" 	-- Restore the schedule 	
##
## @example
##    $obj->schedule(action=>C_SNAPMIRROR_SCHEDULE_CLEAR); # remove the schedule
##    $obj->schedule(action=>C_SNAPMIRROR_SCHEDULE_RESTORE); # restore the schedule
sub schedule
{
	my $class = shift;
	my (%args) = validate (@_, { 
			action => { type=>SCALAR, optional=>1, default => C_SNAPMIRROR_SCHEDULE_GET },
		     } );
	
	my $res = $class->{_snapmirror}->schedule(cmd=>$args{action});
	return $res;
}

## @name:: check
## @summary SnapMirror check
##	
##
## @description
##	Run SnapMirror checker
##
## @return Hash with rc and output.
##
## @example
## $obj->check();
sub check
{
	my $class = shift;
	my (%args) = validate (@_, { } );

	$class->{_snapmirror}->check();
	return (rc => C_RC_SUCCESS);
}

## @name:: wait_for
## @summary SnapMirror wait for status to happen
##	
##
## @description
##	Wait for the snapmirror relationship to reach a certain conidition
##      by specifying values for columns in the 'snapmirror show' output
##
## @arg match Optional: BOOLEAN, Determine if waiting for condition to match or not. Default: 1
## @arg current Optional: A comma separated list of 'snapmirror show' column names.
##			  This allows for the specification of using the current
##			  values of these columns instead of giving a value.
## @arg timeout Optional: Timeout in seconds to wait for specified condition. Default: 3600 (1 hour)
## @arg interval Optional: Polling interval to check for specified condition. Default: 15
##
## @return Hash with rc and output.
##
## @example
## $obj->wait_for(state=>C_SNAPMIRROR_STATE_SNAPMIRRORED, status=>C_SNAPMIRROR_STATUS_IDLE);
## $obj->wait_for(match=>0, current=>C_SNAPMIRROR_STATUS); ## return if current status changes 
sub wait_for
{
	my $class = shift;
	my %args = validate (@_, {
			match	 => { type => BOOLEAN, optional => 1 },
			state	 => { type => SCALAR, optional => 1 },
			status	 => { type => SCALAR, optional => 1 },
			current	 => { type => SCALAR, optional => 1 },
			timeout	 => { type => SCALAR, optional => 1, default  => 3600 },
			interval => { type => SCALAR, optional => 1, default  => 15 },
	} );

	$class->{_snapmirror}->wait_for(%args);

	return (rc => C_RC_SUCCESS);
}

## @name:: monitor
## @summary monitor SnapMirror 
##	
##
## @description
##	Monitors the snapmirror action by checking until the snapmirror reaches a
##	specific state/status combination.
##
## @arg state Optional: Wait until mirror is at this state. Default: snapmirrored
## @arg status Optional: Wait until mirror is at this status. Default: idle
## @arg timeout Optional: Timeout in seconds. Default: 3600 (1 hour)
## @arg interval Optional: Pause interval in between check, in seconds. Default: 5
##
## @return Hash with rc and output.
##
## @example
## $obj->monitor();
## $obj->monitor(timeout=>60*60, interval=>15, state=>C_SNAPMIRROR_STATE_SNAPMIRRORED, status=>C_SNAPMIRROR_STATUS_IDLE);
## $obj->monitor(status=>C_SNAPMIRROR_STATUS_TRANSFERRING);
sub monitor
{
	my $class = shift;
	my %args = validate (@_, {
			state	 => { type => SCALAR, optional => 1 },
			status	 => { type => SCALAR, optional => 1 },
			timeout	 => { type => SCALAR, optional => 1, default  => 3600 },
			interval => { type => SCALAR, optional => 1, default  => 15 },
	} );

	$class->{_snapmirror}->monitor(%args);
	return (rc => C_RC_SUCCESS);
}

## @name:: status
## @summary SnapMirror get status
##	
##
## @description
##	get SnapMirror status
##
## @return snapmirror status hash.
##
## @example
## $obj->status();
sub status
{
	my $class = shift;
	my (%args) = validate (@_, { } );
	my $res = $class->{_snapmirror}->status();
	return $res;
}

## @name:: attach
## @summary attach an existing snapmirror relationship to a SnapMirror object
##
## @description
##	attach an existing snapmirror relationship to a SnapMirror(set) object
##
## @arg type Required: src, snapmirror source
## @arg dp Optional: dp type destination volume object (NOTE::GX::Volume), or an array reference to an array of these objects
## @arg ls Optional: ls type destination volume object (NOTE::GX::Volume), or an array reference to an array of these objects
##  NOTE: at least one of dp and ls must exist
##
## @return a snapmirror object.
##
## @example
## $obj->attach(src => $src_vol,  
##              dp => \@dst_vols,);
##
## NOTE: This method is comment as "internal" method for GNOTE::GX::SnapMirrorSet.pm even though paloma scripts are using it as external method.
##       be cautious when using this method.
##       dpg team will make this method public later. Please refer to burt 486272 
sub attach
{
	my $class = shift;
	my (%args) = validate (@_, { 
		src => { type =>OBJECT, isa => "Compression::Compress" },
		dp  => { type=>OBJECT|ARRAYREF, optional => 1},
		ls  => { type=>OBJECT|ARRAYREF, optional => 1},
	} );

	my %opt;
	$opt{src} = $args{src}->{volume}->_get_underlying_vol_obj();

	my (@vol_dp_objs, @vol_ls_objs);
	if (defined $args{dp}){
		@vol_dp_objs = $class->_get_volume_objects(dst=>$args{dp});
		$opt{dp} = \@vol_dp_objs;
	}
	
	if (defined $args{ls}){
		@vol_ls_objs = $class->_get_volume_objects(dst=>$args{ls});
		$opt{ls} = \@vol_ls_objs;
	}

	$class->{_snapmirror} = $class->{_snapmirror}->attach(%opt);
	return $class;
}


## @name:: verify_last_transfer_size
## @summary SnapMirror: Verify the last transfer size.
##
## @description SnapMirror: Verify the last transfer size.
##
## @arg min Optional: Minimum addmissible value for the last transfer size, in bytes.
## @arg max Optional: Maximum addmissible value for the last transfer size, in bytes.
##
## @return Last transfer size.
##
## @example
## $obj->verify_last_transfer_size();
##
sub verify_last_transfer_size
{
	my $class = shift;
	my (%args) = validate (@_, {min => {type => SCALAR, optional => 1,},
				    max => {type => SCALAR, optional => 1,},} );
	
	my $min = $args{min};
	my $max = $args{max};
	my $entries_found = 0;
	
	if ((!defined($min)) && (!defined($max)))
	{
		logresult('script', "SnapMirror\:\:verify_last_transfer_size : at least one of the arguments must be passed in (min, max)");
	}
	
	my $last_transfer_size = 0;
	my $status = $class->{_snapmirror}->status('last-transfer-size');
	
	if ($class->{_xdp} == TRUE) {
		# TODO: This logic should not be permanent. See comment in 'create'.
		# GNOTE::GX::SnapMirror::status returns the size string directly
		if ($status =~ /^\d+$/) {
			$last_transfer_size = $status;
		}
		$entries_found = 1;
	}
	else {
		# GNOTE::SnapMirrorSet::status returns a hashref, where the keys are the GNOTE::Volume destination objects
		foreach my $vol (keys(%$status)) {
			$entries_found++;
			# In some cases, may be '-'
			if ($status->{$vol} =~ /^\d+$/) {
				$last_transfer_size = $status->{$vol};
				}
		}
	}
	
	if ($entries_found != 1)
	{
		logresult('fatal', "SnapMirror: Unexpected result returned by GNOTE\:\:SnapMirrorSet\:\:status\n" . Dumper($status));
	}
	logcomment("SnapMirror: Last transfer size: $last_transfer_size");
	
	if ((defined($min)) && ($last_transfer_size < $min))
	{
		logresult('fatal', "SnapMirror: Last transfer size is less than the minimum admissible value:\n" .
			           "Obtained value: $last_transfer_size\n" .
				   "Minimum required value: $min");
	}
	
	if ((defined($max)) && ($last_transfer_size > $max))
	{
		logresult('fatal', "SnapMirror: Last transfer size is more than the maximum admissible value:\n" .
			           "Obtained value: $last_transfer_size\n" .
				   "Maximum required value: $max");
	}
	
	return $last_transfer_size;
}


################################
# Internal subroutines
################################
## get the SnapMirror object
sub _mirror
{
	my $class = shift;
	my (%args) = @_; 

	return $class->{_snapmirror}->mirror(%args);
}

## get volume objects
sub _get_volume_objects
{
	my $class = shift;
	my (%args) = @_;
	my $dst = $args{dst};
	my @volume_objs;
	
	if (ref($dst) eq "ARRAY") {
		foreach my $obj (@$dst){
			if( $obj->isa("Compression::Compress")) {	
				my $vol_obj = $obj->{volume}->_get_underlying_vol_obj();
				push (@volume_objs, $vol_obj);
			}
			else {
			      croak "destination object should be Compression::Compress object or an array referrence to these objects\n $!";
			}
		}
	} elsif ($dst->isa("Compression::Compress")){
		my $vol_obj = $dst->{volume}->_get_underlying_vol_obj();
		push (@volume_objs, $vol_obj);
	} else {
		croak "destination object should be Compression::Compress object or a array referrence to these objects \n $!";
	}

	return @volume_objs;
}

# TODO: Need to investigate to clean up staled snapmirror relationship and offline/restricted volumes in DESTROY method

1;