#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/CMode/VolCopy.pm#1 $
#

## @summary Volume copy routines.
##
## @description
##    This module provides all the capabilities
##    required for volume copy. This is an internal module.
##    To not be used directly by scripts

## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (internal)

package Compression::CMode::VolCopy;

use strict;
use warnings;

use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Tharn;

use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult);


BEGIN
{
	require "Compression/base_import.pl";
}



## @name ::new
##
## @summary Compression::CMode::VolCopy Constructor.
##
## @description
## Creates an Compression::CMode::VolCopy object 
## (it does not start the vol copy relationship on the filer).
##
## @return Compression::CMode::VolCopy object.
##
## @arg source Required: Source Compression::Compress object.
##
## @arg destination Required: Destination GNOTE::GX::Volume object.
##
## @arg disable_cleanup Optional: Flag to disable any clean-up
## that would be executed upon object destruction.
## Defaults to FALSE.
##
## @example
## my $vol_copy = Compression::VolCopy->new(source => $compress1,
##                                          destination => $compress2);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
		source 	=> {type => OBJECT,
                            isa => "Compression::Compress",},

		destination => {type => HASHREF,},
                
                vserver => {type => SCALAR,},
                
                disable_cleanup => {type => BOOLEAN,
                                    optional => 1,
                                    default => FALSE,},
                
                debug => {type => BOOLEAN,
                          optional => 1,
                          default => $main::DEBUG,},
		} );
	
	my $self = {};
        $self->{_source} = $args{source};
        $self->{_disable_cleanup} = $args{disable_cleanup};
        $self->{_debug} = $args{debug};
        $self->{_creator_pid} = $$;
        $self->{_boot_mode} = $self->{_source}->{boot_mode};
        $self->{_vserver} = $args{vserver};
        
        $self->{_src_vol_name} = $self->{_source}->{vol_name};
        $self->{_src_aggr_name} = $self->{_source}->{aggr_name};
        $self->{_src_host_name} = $self->{_source}->{host_name};
        
        my $dst = $args{destination};
        my $src = $self->{_source};
        my %dst_args = validate_with
                (params => $dst,
                 spec   => {volume => { type => SCALAR|HASHREF|UNDEF, },
                            aggregate => { type => SCALAR|HASHREF, default => $self->{_src_aggr_name}, },
                            filer => { type => SCALAR|HASHREF, default => $self->{_src_host_name}, },},
                 allow_extra => 0,);

        my $dst_vol_args = $dst_args{volume};
        my $dflt_dst_vol_name = 'copy_of_' . $self->{_src_vol_name};
        if ((defined($dst_vol_args) && (ref($dst_vol_args) eq "HASH")))
        {
                $self->{_dst_vol_name} = (defined($dst_vol_args->{name})) ? $dst_vol_args->{name} : $dflt_dst_vol_name;
                $self->{_dst_vol_args} = $dst_vol_args;
                $self->{_dst_vol_args}->{name} = $self->{_dst_vol_name};
                $self->{_dst_vol_args}->{compress_enabled} = C_FLAG_KEEP_VALUE_AS_IS;
                $self->{_dst_vol_args}->{asis} = FALSE;
        }
        else
        {
                $self->{_dst_vol_name} = (defined($dst_vol_args)) ? $dst_vol_args : $dflt_dst_vol_name;
                $self->{_dst_vol_args} = {name => $self->{_dst_vol_name},
                                          compress_enabled => C_FLAG_KEEP_VALUE_AS_IS,
                                          asis => FALSE,};
        }
        
        my $dst_aggr_args = $dst_args{aggregate};
        if (ref($dst_aggr_args) eq "HASH")
        {
                $self->{_dst_aggr_name} = (defined($dst_aggr_args->{name})) ? $dst_aggr_args->{name} : $self->{_src_aggr_name};
                $self->{_dst_aggr_args} = $dst_aggr_args;
                $self->{_dst_aggr_args}->{name} = $self->{_dst_aggr_name};
                $self->{_dst_aggr_args}->{indirect_type} = (exists($dst_aggr_args->{indirect_type})) ? $dst_aggr_args->{indirect_type} : $src->{aggregate}->{indirect_type};
                $self->{_dst_aggr_args}->{snaplock} = (exists($dst_aggr_args->{snaplock})) ? $dst_aggr_args->{snaplock} : $src->{aggregate}->{snaplock};
        }
        else
        {
                $self->{_dst_aggr_name} = $dst_aggr_args;
                $self->{_dst_aggr_args} = {name => $dst_aggr_args,
                                           indirect_type => $src->{aggregate}->{indirect_type},
                                           snaplock => $src->{aggregate}->{_snaplock},};
        }
        
        my $dst_filer_args = $dst_args{filer};
        if (ref($dst_filer_args) eq "HASH")
        {
                $self->{_dst_host_name} = (defined($dst_filer_args->{name})) ? $dst_filer_args->{name} : $self->{_src_host_name};
                $self->{_dst_filer_args} = $dst_filer_args;
                $self->{_dst_filer_args}->{name} = $self->{_dst_host_name};
        }
        else
        {
                $self->{_dst_host_name} = $dst_filer_args;
                $self->{_dst_filer_args} = {name => $dst_filer_args};
        }
        
        $self->{_ui} = $self->{_source}->{_ui};
        $self->{_transfer_job_id} = -1;
        

        bless $self, $class; 

	return $self;
}


## @name ::attach_to_destination_vol
## @summary Gets the compress destination object.
## 
## @description
## Gets the compress destination object. It will return a compress object
## corresponding to the volume created as a result of the volume copy operation.
##
## @return Compression::Compress object corresponding to the destination volume.
##
## @examples
## my $dest = $obj->attach_to_destination_vol();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub attach_to_destination_vol
{
        my $class = shift;
        my %args = validate(@_,
                {
                        # None
                });
        
        if ($class->{_transfer_job_id} == -1)
        {
                logresult('NA', "NOT APPLICABLE: In C-mode, the destination volume will be created by the volume copy process " .
                                "and it can only be used after the copy transfer was completed");
        }
        
        $class->_attach_to_destination();
        
        return $class->{_destination};
}



## @name ::start
## @summary Starts the vol copy process.
## 
## @description
## Starts the vol copy process.
##
## @return Hash with rc.
## 
## @arg background Optional: Flag to run the transfer in the background.
## To not be set if the expected result is not success.
## Default: FALSE.
## 
## @arg copy_all Optional: Flag to include all the volume's snapshots.
## Default: FALSE.
## 
## @arg snapshot Optional: The name of the snapshot, on the source volume, to copy.
##
## @arg timeout Optional: Timeout, in seconds, for the command to complete.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @arg restore_destination_state Optional: Flag to control if, after the copy,
## the destination volume's state should be restored (onlined and remounted).
## If defaults to TRUE, if the background option is not true.
##
## @examples
## $obj->start();
## $obj->start(copy_all => TRUE);
## $obj->start(snapshot => 'my_snapshot');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub start
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        background => {type => SCALAR, optional => 1, default => FALSE,},
                        copy_all => {type => SCALAR, optional => 1, default => FALSE,},
                        snapshot => => {type => SCALAR, optional => 1,},
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
                } );
	
        
        my $expected = $args{expected};
        my $background = $args{background};
        my $foreground = ($background == TRUE) ? 'false' : 'true';
        my %res = (rc => C_RC_SUCCESS);
        
        if (($args{copy_all} == TRUE) || (defined($args{snapshot})))
        {
                logresult('NA', "NOT APPLICABLE: Volume copy: arguments \"copy_all\" and \"snapshot\" are not available in C-mode");
        }
        
        my $source_vol = $class->{_src_vol_name};
        logcomment("Starting vol copy:\n" .
                   "Source: $source_vol (filer: " . $class->{_src_host_name} . "; aggregate: " . $class->{_src_aggr_name} . ")\n" .
                   "Destination: " . $class->{_dst_vol_name} . " (aggregate: " . $class->{_dst_aggr_name} . ")\n" .
                   "Background option: " . $args{background});
        
        my $src_obj = $class->{_source}->{volume}->_get_underlying_vol_obj();
        
        if ($class->{_transfer_job_id} != -1)
        {
                # Attempting to resume the job fails (it was stopped - pause is not supported)
                logresult('NA', "NOT APPLICABLE: In C-mode, vol copy cannot be restarted");
        }
        else
        {
                my $attempt_job_id;
                eval
                {
                        $attempt_job_id = $src_obj->server->volume_copy_start('volume' => $class->{_src_vol_name},
                                                                              'vserver' => $class->{_vserver},
                                                                              'destination-volume' => $class->{_dst_vol_name},
                                                                              'destination-aggregate' => $class->{_dst_aggr_name},
                                                                              'foreground' => $foreground);
                };
                
                if (!$@)
                {
                        $class->{_transfer_job} = NOTE::GX::Job->new(Cluster => $src_obj->Cluster,
                                                                     id => $attempt_job_id,);
                        $class->{_transfer_job_id} = $attempt_job_id;
                }
                
                %res =  $class->{_ui}->process_output_string(code => C_CMD_VOL_COPY_START,
                                                             output => $@,
                                                             expected_rc => $expected,
                                                             log_message => "Start vol copy of volume $source_vol",);
        }
        
        if (($expected == C_RC_SUCCESS) && ($background == FALSE))
        {
                $class->_attach_to_destination();
        }
        
        return %res;
}


## @name ::wait_for_completion
## @summary Wait for the background vol copy transfers to complete.
## 
## @description
## Wait for the background vol copy transfers to complete.
##
## @return Hash with rc.
##
## @examples
## $obj->wait_for_completion();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

sub wait_for_completion
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        # None
                } );
        
        
        logcomment("Waiting for vol copy to complete:\n" .
                   "Source: " . $class->{_src_vol_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                   "Destination: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . ")");
        
        if ($class->{_transfer_job_id} == -1)
        {
                logresult('script', "Compression\:\:AggrCopy: wait_for_completion must be called after the transfer has been started");
        }
        
        $class->{_transfer_job}->wait();
        
        $class->_attach_to_destination();
        
        return (rc => C_RC_SUCCESS);
}


## @name ::get_status
## @summary Gets the status of vol copy relationship.
## 
## @description
## Gets the status of vol copy relationship.
##
## @return Hash with rc, status, and transfer_progress. 
##
## @examples
## my %status = $obj->get_status();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub get_status
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        # None
                } );
        
        my $status = C_VOL_COPY_STATUS_IDLE;
        
        if ($class->{_transfer_job_id} != -1)
        {
                my $job_status = $class->{_transfer_job}->status(id => $class->{_transfer_job_id});
                
                # status will return 0 if there is no corresponding job entry anymore
                if ((ref($job_status) eq "HASH") && ($job_status->{state} =~ /running/i))
                {
                        $status = C_VOL_COPY_STATUS_TRANSFERRING;
                }
        }
        
        logcomment("Vol copy status: \"$status\"\n" .
                   "Source: " . $class->{_src_vol_name} . " (filer: " . $class->{_src_host_name} . "; aggregate: " . $class->{_src_aggr_name} . ")\n" .
                   "Destination: " . $class->{_dst_vol_name} . " (aggregate: " . $class->{_dst_aggr_name});
        
        return (rc => C_RC_SUCCESS,
                status => $status,);
}


## @name ::verify_status
## @summary Verifies the status of vol copy relationship.
## 
## @description
## Verifies the status of vol copy relationship.
##
## @arg status Optional: Expected status.
## Default: C_VOL_COPY_STATUS_TRANSFERRING.
##
## @return Hash with rc. 
##
## @examples
## $obj->verify_status(status => C_VOL_COPY_STATUS_IDLE);
## $obj->verify_status(); # Same as $obj->verify_status(status => C_VOL_COPY_STATUS_TRANSFERRING);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub verify_status
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        status => {type => SCALAR, optional => 1, default => C_VOL_COPY_STATUS_TRANSFERRING},
                } );
        
        my $expected_status = $args{status};
        
        logcomment("Verifying the vol copy status ($expected_status):\n" .
                   "Source: " . $class->{_src_vol_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                   "Destination: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . ")");
        
        my %current_status = $class->get_status();
        
        if ($current_status{status} ne $expected_status)
        {
                logresult('fatal', "Failed to detect expected vol copy transfer status ($expected_status):\n" . Dumper(%current_status));
        }
        
        return %current_status;
}


## @name ::wait_for_status
## @summary Wait for the vol copy relationship to reach specific state.
## 
## @description
## Wait for the vol copy relationship to reach specific state.
##
## @return Hash with rc.
##
## @arg status Required: Status to wait for (can be a regex).
#
# @arg source Optional: Flag to get the status from the source filer.
# Default: FALSE (destination filer).
#
# @arg restore Optional: Flag to get the status of the restore transfer.
# Default: FALSE. 
##
## @arg timeout Optional: Timeout, in seconds, for completion.
##
## @examples
## $obj->wait_for_status(status => C_VOL_COPY_STATUS_TRANSFERRING);
## $obj->wait_for_status(status => C_VOL_COPY_STATUS_IDLE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub wait_for_status
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        status => {type => SCALAR,},
                        # Not supported in C-mode
                        #source => {type => SCALAR, optional => 1, default => FALSE,},
                        #restore => {type => SCALAR, optional => 1, default => FALSE,},
                        timeout => {type => SCALAR, optional => 1, default => C_DFLT_TIMEOUT},
                } );
        
        my $status = $args{status};
        my $timeout = $args{timeout};
        
        my $comment = "Waiting for vol copy status \"$status\":\n" .
                      "Source: " . $class->{_src_vol_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                      "Destination: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . ")";
        
        if (defined($timeout))
        {
                $comment .= "\nTimeout: $timeout";
        }
        
        logcomment($comment);
        
        my %current_status = $class->get_status();
        
        while ($current_status{status} ne $status)
        {
                $timeout--;
                if ($timeout <= 0)
                {
                        logresult('fatal', "Timed-out while waiting for volume copy status \"$status\"\n" .
                                           "Last obtained status: " . $current_status{status});
                }
                
                %current_status = $class->get_status();
        }
        
        if (($class->{_transfer_job_id} != -1) && ($status eq C_VOL_COPY_STATUS_IDLE))
        {
                # Most likely we are waiting for copy to complete
                # (it has been started sometime and now it is idle)
                $class->_attach_to_destination();
        }
        
        return (rc => C_RC_SUCCESS);
}


## @name ::abort
## @summary Aborts an active vol copy transfer.
## 
## @description
## Aborts an active vol copy transfer.
##
## @return Hash with rc.
##
## @examples
## $obj->abort();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub abort
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        # None
                } );
        
        logcomment("Aborting vol copy transfer:\n" .
                   "Source: " . $class->{_src_vol_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                   "Destination: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . ")");
        
        if ($class->{_transfer_job_id} == -1)
        {
                logresult('NA', "NOT APPLICABLE: In C-mode, abort cannot be called unless the copy was previously started");
        }
        
        my $src_obj = $class->{_source}->{volume}->_get_underlying_vol_obj();
        $src_obj->server->job_stop(id => $class->{_transfer_job_id});
        
        return (rc => C_RC_SUCCESS);
}


## @name ::throttle
## @summary Modifies the speed at which the vol copy transfer is being done.
## 
## @description
## Modifies the speed at which the vol copy transfer is being done.
##
## @arg value Optional: New speed value, to be set.
## The  value  ranges from 10 (full speed) to 1 (one-tenth of full speed).
## Default: C_DFLT_VOL_COPY_THROTTLE (10, full speed).
##
## @return Hash with rc. 
##
## @examples
## $obj->throttle(value => 1);
## $obj->throttle(); # To set back to default
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

sub throttle
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        value => {type => SCALAR, optional => 1, default => C_DFLT_VOL_COPY_THROTTLE},
                } );
        
        logresult('NA', "NOT APPLICABLE: Vol copy throttle is not available in C-mode");
}


## @name ::restore_destination
## @summary Restores the state of the destination volume.
## 
## @description
## Restores the state of the destination volume, after the copy transfer has completed 
## (brings it online and remounts it).
##
## @return Hash with rc. 
##
## @examples
## $obj->restore_destination();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub restore_destination
{
        my $class = shift;
        
        logcomment("Restoring the copy destination volume state:\n" .
                   "Volume: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . ")");
        
        $class->_attach_to_destination();
        # For a short time, the volume may be still busy with copy
        # Loop a few times, checking for the status first
        # (this is a best-effort try. If we finished)
        for (my $allowed_attempt = 1; $allowed_attempt <= 10; $allowed_attempt++)
        {
                my %dest_status = $class->{_destination}->{volume}->get_status();
                my $state = $dest_status{status}->{state};
                logcomment("VolCopy: Destination volume state: $state");
                if ($state ne C_VOL_STATE_BUSY)
                {
                        # Ok to proceed
                        last;
                }
        }
        
        logcomment("VolCopy: Bringing the destination volume online...");
        $class->{_destination}->{volume}->bring_online();
        logcomment("VolCopy: Remounting the destination volume...");
        $class->{_destination}->{volume}->nfs_umount();
        $class->{_destination}->{volume}->nfs_mount();
        
        $class->_reset_destination_dataset_info();
        
        return (rc => C_RC_SUCCESS);
}



########################
# Internal methods
########################

sub _reset_destination_dataset_info
{
        my $class = shift;
        
        $class->_attach_to_destination();
        $class->{_destination}->{dataset}->attach_dataset_info(src_vol => $class->{_source},
                                                               reset_info => TRUE,);
}

sub _attach_to_destination
{
        my $class = shift;
        
        if (!defined($class->{_destination}))
        {
                $class->{_destination} = Compression::Compress->setup(filer => $class->{_dst_filer_args},
                                                                      aggregate => $class->{_dst_aggr_args},
                                                                      volume => $class->{_dst_vol_args},
                                                                      attach_to_volume => $class->{_dst_vol_name},
                                                                      debug => $class->{_debug},
                                                                      disable_cleanup => $class->{_disable_cleanup},);
        }
}

## Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
        
        if (($class->{_disable_cleanup} == FALSE) &&
            ($class->{_creator_pid} == $$))
        {
                logcomment("Executing Compression\:\:CMode\:\:VolCopy cleanup for:\n" .
                           "Source: " . $class->{_src_vol_name} . "\n" .
                           "Destination: " . $class->{_dst_vol_name});
                
                eval
                {
                        my %active = $class->get_status();
                        if ($active{status} eq C_VOL_COPY_STATUS_TRANSFERRING)
                        {
                                $class->abort();
                        }
                };
                
                if( $@ )
                {
                        logcomment( "VolCopy: cleanup error: " . $@ );
                }
                
                undef($class->{_destination});        
        }
	else
	{
                logcomment( "VolCopy: c-mode cleanup:  nothing done: script PID: " . $$) if ($class->{_debug});
	}
        
        if (defined($execution_error))
        {
                _c_interpret_logresult($execution_error);
        }
}


1;
