# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/CMode/VolMove.pm#1 $
#
## @summary
##  Utility wrapper module for C-mode vol move
##
## @author @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (internal)

package Compression::CMode::VolMove;

BEGIN
{
	require "Compression/base_import.pl";

}

use strict;
use warnings;
use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Compression::Constants qw(:all);



# Documentation in Compression::VolMove
sub new
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		volume => {type => OBJECT,
			   isa => "NOTE::GX::Volume",},
		
		dest_aggr => {type => SCALAR,},
                
                ui => {type => OBJECT,
		       isa => "Compression::UI",},
	});

	# Start building our object:
	my $self = {};
	$self->{_volume} = $args{volume};
	$self->{_dest_aggr} = $args{dest_aggr};
	$self->{_ui} = $args{ui};
	$self->{_debug} = $self->{_ui}->{_debug};
	$self->{_creator_pid} = $$;
	$self->{_vol_name} = $self->{_volume}->name();
	$self->{_filer_name} = $self->{_volume}->Node->name();
	$self->{_vserver} = $self->{_volume}->Vserver->name();
	$self->{_job_id} = -1;
	
	# Create the destination aggregate, if it does not exist yet
	# Use a temporary object for this
	logcomment("Initializing vol move: creating the destination aggregate, if not already present: " . $self->{_dest_aggr});
	my $aggr = NOTE::GX::Aggregate->create(name => $self->{_dest_aggr},
					       Node => $self->{_volume}->Node(),
					       recycle => 1,
					       share => 1,
					       disks => C_DFLT_AGGR_DISKS,
					       indirect_type => $main::AGGR_IND_TYPE,);
	
	
	bless($self, $class);
	
	return $self;
}


sub start
{
        my $class = shift;
        my (%args) = validate( @_,
        {
		no_cutover => {type => SCALAR, default => FALSE,},
                expected => {type => SCALAR, default => C_RC_SUCCESS,},
	});
        
	my $expected = delete($args{expected});
	my $vol_name = $class->{_vol_name};
	my $dest_aggr = $class->{_dest_aggr};
	my $no_cutover = delete($args{no_cutover});
	if (TRUE == $no_cutover)
	{
		$args{'cutover-action'} = 'wait';
	}
	
	logcomment("Starting move of volume $vol_name to aggregate $dest_aggr");
	$args{'destination-aggregate'} = $class->{_dest_aggr};
	$args{volume} = $vol_name;
	$args{vserver} = $class->{_vserver};
	# For backwards compatibility with 7-mode
	$args{foreground} = 'false';
	
	eval {
		$class->{_job_id} = $class->{_volume}->server->volume_move_start(%args);
	};
	
	my %res = $class->{_ui}->process_output_string(code => C_CMD_VOL_MOVE_START,
						       output => $@,
						       expected_rc => $expected,
						       log_message => "Start move of volume $vol_name to aggregate $dest_aggr",);
	
	if (-1 == $class->{_job_id})
	{
		$class->{_job_id} = $class->{_volume}->Node->job_id($@);
	}
	
	return %res;
}


sub status
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                expected => {type => SCALAR, default => C_RC_SUCCESS,},
		required_state => {type => SCALAR, optional => 1,}
        });
        
	my $expected = $args{expected};
	my $vol_name = $class->{_vol_name};
	my $cmd_code = C_CMD_VOL_MOVE_STATUS;
	my $required_state = $args{required_state};
	
	my $job = NOTE::GX::Job->new(Cluster => $class->{_volume}->Cluster,
				     id =>$class->{_job_id},);
	my $status = $job->status();
	my $completion = $status->{completion};
	my $rc = ($completion eq '""') ? C_RC_SUCCESS : C_RC_VOL_IS_NOT_BEING_MOVED;
	
	my %res = $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
						       expected_rc => $expected,
						       log_message => "Obtain move status for volume $vol_name",
						       output      => $status);
	$res{progress} = $status->{progress};
	$res{move_status} = $status;
	
	
	if (defined($required_state))
	{
		my $obtained_state = $class->_determine_state($status->{progress});
		if ($obtained_state ne $required_state)
		{
			logresult('fatal', "Failed to detect required vol move state for volume $vol_name\n" .
				           "Required state: $required_state\n" .
					   "Obtained state: $obtained_state\n" .
					   "\n\n" . Dumper($status));
		}
	}
	
	return %res;
}


sub wait_for_completion
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                timeout => {type => SCALAR, regex => qr/^\d+$/, default => C_DFLT_TIMEOUT,},
		delay => {type => SCALAR, regex => qr/^\d+$/, default => 5,},
        });
        
	my $vol_name = $class->{_vol_name};
	my $timeout = $args{timeout};
	my $delay = $args{delay};
	
	logcomment("\nWaiting for vol move operation (volume $vol_name) to complete");
	$class->{_volume}->Node->job_wait(id =>$class->{_job_id},);
	
	return (rc => C_RC_SUCCESS);
}


sub wait_for_state
{
        my $class = shift;
        my (%args) = validate( @_,
        {
		state => {type => SCALAR,},
                timeout => {type => SCALAR, regex => qr/^\d+$/, default => C_DFLT_TIMEOUT,},
		delay => {type => SCALAR, regex => qr/^\d+$/, default => 1,},
        });
        
	my $vol_name = $class->{_vol_name};
	my $timeout = $args{timeout};
	my $delay = $args{delay};
	my $state = $args{state};
	
	logcomment("\nWaiting for vol move operation (volume $vol_name) to reach the state \"$state\"");
	while ($timeout >= 0)
	{
		# Expect success for status
		my %res = $class->status();
		my $progress = $res{progress};
		my $obtained_state = $class->_determine_state($progress);
		logcomment("Current state: $obtained_state (progress: $progress)"); 
		if ($obtained_state eq $state)
		{
			logcomment("\nFound required state for vol move operation (volume $vol_name): \"$state\"");
			last;
		}
		
		sleep($delay);
		$timeout-=$delay;
	}
	
	if ($timeout < 0)
	{
		logresult('fatal', "Timed-out while waiting for vol move to complete: volume: $vol_name");
	}
	
	return (rc => C_RC_SUCCESS);
}


sub initiate_cutover
{
        my $class = shift;
        my (%args) = validate( @_,
        {
		expected => {type => SCALAR, default => C_RC_SUCCESS,},
	});
        
	my $vol_name = $class->{_vol_name};
	my $expected = $args{expected};
	
	logcomment("Initiating move cutover volume $vol_name");
	
	eval {
		$class->{_volume}->server->volume_move_trigger_cutover(volume => $vol_name,
								       vserver => $class->{_vserver},);
	};
	
	return $class->{_ui}->process_output_string(code => C_CMD_VOL_MOVE_CUTOVER,
						    output => $@,
						    expected_rc => $expected,
						    log_message => "Initiate move cutover for volume $vol_name",);
}


sub pause
{
        my $class = shift;
        
	logresult('NA', "NOT APPLICABLE: Vol move pause is not available in C-mode");
}


sub is_paused
{
        my $class = shift;
        
	logresult('NA', "NOT APPLICABLE: Vol move pause/is_paused is not available in C-mode");
}


sub resume
{
        my $class = shift;
        
	logresult('NA', "NOT APPLICABLE: Vol move resume is not available in C-mode");
}


sub abort
{
        my $class = shift;
        my (%args) = validate( @_,
        {
		# None for now
	});
        
	my $vol_name = $class->{_vol_name};
	
	logcomment("Aborting vol move for volume $vol_name");
	
	$class->{_volume}->server->job_stop(id => $class->{_job_id},);
	
	# Allow the job cleanup to be completed 
	my $timeout = 20;
	my %status;
	while ($timeout > 0)
	{
		%status = $class->status(expected => C_RC_FAILURE_NON_FATAL);
		if (C_RC_VOL_IS_NOT_BEING_MOVED == $status{rc})
		{
			last;
		}
		sleep(1);
		$timeout--;
	}
	
	if ($timeout <= 0)
	{
		logresult('fatal', "Timed-out while waiting for the vol move operation to be aborted\n" . Dumper(%status))
	}
	
	return (rc => C_RC_SUCCESS,);
}


################################################################
# Internal subroutines
################################################################

sub _determine_state
{
	my $class = shift;
	my $progress_str = shift;
	
	my $obtained_state = 'unknown';
	my $vol_name = $class->{_vol_name};
	
	my %progress = $class->{_ui}->process_output_string(code => C_CMD_FIND_VOL_MOVE_STATE,
							    output => $progress_str,
							    expected_rc => C_RC_FAILURE_NON_FATAL,
							    log_message => "Obtain move state for volume $vol_name from status:\$progress_str",);
	
	my $progress_rc = $progress{rc};
	if (_C_RC_VOL_MOVE__SETUP_STATE == $progress_rc)
	{
		$obtained_state = C_VOL_MOVE_STATE_SETUP;
	}
	elsif (_C_RC_VOL_MOVE__MOVE_STATE == $progress_rc)
	{
		$obtained_state = C_VOL_MOVE_STATE_MOVE;
	}
	elsif (_C_RC_VOL_MOVE__CUTOVER_STATE == $progress_rc)
	{
		$obtained_state = C_VOL_MOVE_STATE_CUTOVER;
	}
	elsif (_C_RC_VOL_MOVE__CUTOVER_DEFERRED_STATE == $progress_rc)
	{
		$obtained_state = C_VOL_MOVE_STATE_CUTOVER_DEFERRED;
	}
	
	return $obtained_state;
}


1;