#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Cluster.pm#1 $
#
## @summary Compression Cluster/SFO routines 
##
## @description
##  This library provides functions specific
##	to cluster(7-mode)/sfo(c-mode)  related automations
##
## @synopsis
##
## use Compression::Cluster;
## use vars qw($FILER $VOLUME $CLIENT $ENABLE $AGGREGATE); 
## 
## BEGIN { require "drg/Compression/setup_core.pl"; }
## 
## my $base_vol = Compression::Compress->setup(
## 				filer		=> { name => $FILER },
## 				aggregate	=> { name => $AGGREGATE },
## );
##
##	my $pair = Compression::Cluster->new(filer=>$FILER);
##	%res = $pair->get_status;
##	$pair->takeover;
##	$pair->monitor(me=>'TAKEOVER', partner=>'DEAD');
##	%res = $pair->get_status;
##	$pair->giveback;
##	$pair->monitor(me=>'UP', partner=>'UP');
##	%res = $pair->get_status;
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
## @status private (in-progress)


package Compression::Cluster;

use strict;
use warnings;
use Tharn;
#use Module::Common;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Compression::Constants qw(:all);
use Compression::7Mode::CFO;
use Compression::CMode::SFO;

BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}


## @name ::new
##
## @summary perform cluster/sfo initialization
##
## @description
## Perform all the basic cluster/sfo initialization required here
##
## @return Compression::Cluster object
##
## @arg filer Required: primary filer/node name
## @arg debug Optional: logs additional debug messages specific to this module
## @arg initialize Optional: Initialize clustering if needed (license, enable).
##
## @examples
## 	$cluster  = Compression::Cluster->new(filer => $FILER );
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##

sub new
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		filer  => { type=>SCALAR },
		debug  => { type=>SCALAR, optional => 1, default=>FALSE },
		initialize => { type => BOOLEAN, optional => 1, default => TRUE },
	} );
	
	my $FILER = $args{filer};
	my $self 	= {};
	
	$self->{ID} 			= $$;
	$self->{Description} 	= "Cluster Library";
	$self->{_debug} = $args{debug};
	
	if (defined($main::VSERVER))
	{
		$self->{_boot_mode} = C_ONTAP_CMODE;
		$self->{_failover} = Compression::CMode::SFO->new(%args);
	}
	else
	{
		$self->{_boot_mode} = C_ONTAP_7MODE;
		$self->{_failover} = Compression::7Mode::CFO->new(%args);
	}
	
	$self->{partner_name} = $self->{_failover}->{partner_name};
	$self->{host_name} = $self->{_failover}->{host_name};
	
	bless $self, $class;
	
	logcomment("Cluster Initialized with ". Dumper($self) ) if ( $self->{_debug} );
	logcomment("Partner is: ". $self->{partner_name} );
	return $self;
}

## @name ::takeover
##
## @summary Initiates takeover.
##
## @description
## Initiates takeover of the partner.
##
## @arg me Optional: if TRUE, partner takes over
## @arg background Optional: run as a background process
##
## @examples
## 	$cluster->takeover();
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub takeover
{
	my $class = shift;
	my (%args) = validate (@_,
	{
		state => { type => SCALAR, optional => 1, default => 'In takeover' },
	} );
	
	$class->{_failover}->takeover(%args);
}

## @name ::giveback
##
## @summary Initiates giveback.
##
## @description
## Initiates giveback to the partner.
##
## @arg background	Optional: run as a background process
## @arg force		Optional: use the "-f" force option
## @arg safe		Optional: if TRUE, giveback will not croak if the the current state of the CFO pair is NOT taken over
##
## @examples
## 	$cluster->giveback();
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub giveback
{
	my $class = shift;
	
	$class->{_failover}->giveback(@_);
}


## @name ::get_status
##
## @summary Obtain failover status.
##
## @description
## Obtain failover status.
##
## @return
## A hash with status for "me" and "partner".
## Status can be 'UP', 'TAKEOVER', 'DEAD', 'UNREACH', 'DISABLED'
##
## @examples
## 	%res = $cluster->status;
##	%res keys: "me" and "partner"
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub get_status
{
	my $class = shift;
	
	return $class->{_failover}->get_status(@_);
}

## @name ::monitor
##
## @summary monitor
##
## @description Monitors failover until both members reach the desired state 
##
## @examples
## 	$cluster->monitor(me => 'TAKEOVER', partner => 'DEAD');
##	$cluster->monitor(me => 'UP', partner => 'UP');
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub monitor
{
	my $class = shift;
	
	$class->{_failover}->monitor(@_);
}


1;
