#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Compress.pm#1 $
#

## @summary Compression routines
##
## @description
##    This module provides capabilities
##	required for core compression work flow.
##
## @synopsis
##
## use Compression::Constants;
## use vars qw($FILER $AGGREGATE $VOLUME $CLIENT);
##
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $obj = Compression::Compress->setup(
##       filer => { name => $FILER, license_ready => FALSE, require_debug_build => TRUE, require_hw_platform => TRUE, },
## 	aggregate => { name => $AGGREGATE },
##      volume => { name => $VOLUME, size => "2g", compress_enabled => FALSE, },
##      );
##
## @examples
##
## $obj->enable();
## $obj->disable();
## $obj->is_enabled( expected => C_RC_FAILUE );
## $obj->toggle();
## $obj->verify_savings( expected_compressed => 8 );
## $obj->verify_savings( expected_compressed => 8, allow_percentage_error => 15 );
## $obj->{filer}->license_add( service => C_LICENSE_COMPRESSION );
## $obj->{filer}->license_remove( service => C_LICENSE_COMPRESSION );
## $obj->{filer}->license_verify( service => C_LICENSE_COMPRESSION, expected => C_RC_FAILURE );
## $obj->{filer}->license_verify( service => C_LICENSE_COMPRESSION, site_license => TRUE );
## $obj->{filer}->reboot_and_wait( timeout => 500 );
## $obj->{filer}->powercycle_and_wait( timeout => 500 );
## $obj->{filer}->panic_and_wait( timeout => 500 );
## $obj->{filer}->reconnect( mode => '#console' );
## $obj->{filer}->volume_exists( volume => $vol, expected => C_RC_FAILURE );
## $obj->{filer}->get_status_of_all_volumes();
## $obj->{filer}->count_existent_volumes( state => C_VOL_STATE_ONLINE );
## my $fiji = $obj->{filer}->set_fiji_point(tag => C_FIJI_TAG_DISABLE_CP);
## Note: The aggregate member of the Compression::Compress object is not available for traditional-volume setup.
## $obj->{aggregate}->volume_is_contained( volume => $vol, expected => C_RC_FAILURE );
## $obj->{aggregate}->set_offline(force => TRUE, all => TRUE );
## $obj->{aggregate}->set_offline();
## $obj->{aggregate}->offline_contained_volumes();
## $obj->{aggregate}->online_contained_volumes();
## $obj->{volume}->bring_offline();
## $obj->{volume}->bring_online();
## $obj->{volume}->mark_restricted();
## $obj->{volume}->nfs_mount( client => $my_client );
## $obj->{volume}->nfs_umount( client => $my_client );
## $obj->{volume}->get_available_space();
## $obj->{volume}->get_status();
## $obj->{volume}->create_snapshot(name => $my_snapshot);
## $obj->{volume}->delete_snapshot(name => $my_snapshot);
## $obj->{volume}->verify_snapshot_exists(name => $my_snapshot, expected => C_RC_FAILURE);
## $obj->{volume}->list_snapshots();
## $obj->{volume}->start_wafliron();
## $obj->{volume}->start_wafliron( wait => FALSE );
## $obj->{volume}->wait_for_wafliron();
## $obj->{volume}->get_wafliron_status();
## $obj->{volume}->stop_wafliron();
## $obj->{volume}->run_wafl_check();
## $obj->{decompress}->status( expected => C_RC_NO_DECOMPRESSION_IN_PROGR );
## $obj->{decompress}->start( timeout => 1 );
## $obj->{decompress}->wait_for_percentage(min => 15, max => 50);
## $obj->{decompress}->stop( disable_status_check => TRUE );
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::Compress;

use strict;
use warnings;

use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult compare_values);
use Compression::Version;
use Compression::Dataset;
use Compression::Decompress;
use Compression::Sis;
use Compression::FileFolding;
use Compression::VolClone;
use Compression::CompressionScanner;
use Compression::CompressionUtilities;
use Compression::Filer::PerfStats;
use Compression::FileClone;
use Compression::VolMove;
use Compression::SNMP;
use Compression::Performance;
use NACL::C::SystemNode;

our @ISA = qw(Compression);


my $cleanup = FALSE;



BEGIN
{
	require "Compression/base_import.pl";
}



## @name ::setup
##
## @summary Compression::Compress constructor.
##
## @description
## Configure and setup pre-requisites for starting the test for compression:
## - Connect to the filer;
## - Detect sw version on the filer;
## - Create aggregate, if required;
## - Create volume, if required;
## - mount volume, if required;
## - license compression, if required;
## - set default configurations like UI type and others for internal use;
##
## @return Compression::Compress object
##
## @arg filer Scalar (name of filer) or hashref of filer information.
## If filer is passed in as a hashref, then the allowed keys are:
## - name => name of the filer,
## - mode => mode to connect to the filer. Default: rsh,
## - perf_stats => used to enable recording/printing statistics. Represents the type of statistics. Default: undefined,
## - license_ready => flag to enable installing the compression license. Default: C_DFLT_LICENSE_READY,
## - require_debug_build => Flag to enable checking for a debug build. If a debug build is required, but not found
##   then the test will be terminated as not applicable. Default: C_DFLT_REQUIRE_DEBUG_BUILD,
## - require_hw_platform => Flag to enable checking for hardware platform. If a simulator is found instead,
##   then the test will be terminated as not applicable. Default: C_DFLT_REQUIRE_HW_PLATFORM,
## - require_cf_disabled => Flag to be used to disable cf, if needed. By using this flag, no further action is needed within the script.
##   The library will take care of re-enabling cf at the end of the script. Currently, the default value is TRUE.
##   This means that any CFO script will need to specifically set it to FALSE.
##
## @arg aggregate Optional: Scalar (name of aggregate) or hashref of aggregate information.
## For example:
## {
##	name => $aggregate,
##      recycle => $reuse,
##	share => TRUE (default),
##      snaplock => TRUE,
##      disks => $num_disks,
## }
##
## @note If setup is called with a request for a traditional volume,
##  the 'aggregate' argument will not be allowed to be passed in,
##  and the aggregate member of the new compress object will not be avilable.
##
## @arg volume Optional: Scalar (name of volume) or hashref of information about the volume:
## name, type, connection type definitions for the volume.
## If not provided, uses the default values.
## For example:
## {    name => $volume,
##      type => $flex_vol_type,
##      size => $volsize,
##	recycle => FALSE,
##      dedupe => FALSE,
##	compress_enabled => TRUE,
##      auto_mount => FALSE,
## }
##
## @note For traditional volumes (type => C_VOL_TYPE_TRAD),
## the volume argument hash can contain the number of disks for the new volume, (ex: disks => 3).
## Also, compress_enabled should be set to FALSE or C_FLAG_KEEP_VALUE_AS_IS, as compression is not available for traditional volumes.
##
## @arg client Scalar (name of client) or hashref of client information.
## For example:
## {
##	name => $client,
##      type => $type_of_client
## }
##
## @arg vfiler Optional: Hashref of vfiler information. The default is:
## {name => $main::VFILER, ip => $main::VFILER_IP, interface => $main::VFILER_IF,},
## which means that, if the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then a vfiler will be created and the compression-related commands,
## will be executed in the vfiler context, if available.
##
## @arg disable_cleanup Optional: Flag to disable any clean-up that is executed
## upon object destruction.
## Defaults to FALSE.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @arg required_features Optional: Arrayref of features that should be supported.
## If any of the listed features is not supported in the releaase under test,
## then the script is terminated with 'N/A'.
##
## @arg supportability Optional: Flag to control if a 'supportability' member
## is required to be created for the compress object (to be used for supportability testing).
## @note Supportability testing can be achieved transparently with a script, by using the
## command-line argument SUPPORTABILITY=1. In this case, supportability checks are
## executed by the library after each of the compression basic actions: enable, disable,
## compress start/stop/completed, decompress start/stop/completed.
## By design, if the script is explicitly passing the supportability flag to setup, then $main::SUPPORTABILITY
## is ignored, and no extra check is done by the library. Any required check needs to be done in the script.
##
## @note
## If $main::CHECK_CONTAINMENT_VERSION is 1, then setup will also check the containment version
## for the volume, if compression is enabled at setup.
##
##
## @examples
## my $compress = Compression::Compress->setup(
##	filer => $FILER,
##      aggregate => $AGGREGATE,
##	volume 	=>
##                 { name => $vol_name,
##		     size => "2g",
##		     compress_enabled => TRUE,
##		     recycle => TRUE, },
##      disable_cleanup => TRUE,
##      debug => TRUE,
## );
##
## @example
## my $compress = Compression::Compress->setup(
##	filer => $FILER,
##	volume 	=> { name => C_DFLT_TRAD_VOLUME,
##		     type => C_VOL_TYPE_TRAD,
##		     compress_enabled => C_FLAG_KEEP_VALUE_AS_IS, },
## );
##
##
## @example
## my $compress = Compression::Compress->setup(
##     filer => {name => $filer,
##               conn_type => $connection_type,
##               license_ready => TRUE,
##               require_debug_build => TRUE,
##               require_hw_platform => TRUE},
##      aggregate => $AGGREGATE,
##	volume 	=> {name => $vol_name,
##		    size => "2g",
##		    compress_enabled => TRUE,
##		    recycle => TRUE, },
##       perf_stats => C_WAFL_STATS,);
##
## @example
## my $compress = Compression::Compress->setup(
##      filer => {name => $filer,},
##      aggregate => $AGGREGATE,
##	volume 	=> {name => $vol_name,},
##      required_features => [C_FEATURE_PRE_CP]);
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub setup
{
	my $class = shift;
	
	my (%args) = validate( @_,
		{
		filer 	=> {
			type => SCALAR|HASHREF,
			},
		
		vserver => {
			type => SCALAR|UNDEF,
			optional => 1,
			default => $main::VSERVER,
			},

		aggregate => {
			type => SCALAR|HASHREF,
			optional => 1,
			},

		volume 	=> {
			type => SCALAR|HASHREF|UNDEF,
			optional => 1,
			},

		client => {
			type => SCALAR|HASHREF,
			optional => 1,
			},
		disable_cleanup => {
			type => BOOLEAN,
                        optional => 1,
			default => $main::DISABLE_CLEANUP,
			},

		debug => {
			type => BOOLEAN,
                        optional => 1,
			default => $main::DEBUG,
			},

                file_folding => {
                    	type => SCALAR,
			optional => 1,
			},
		
		required_features => {
                    	type => ARRAYREF,
			optional => 1,
			},
		
		perf_stats => {type => SCALAR|UNDEF,
			       optional => 1,
			       },
		
		vfiler => {type => HASHREF,
			   optional => 1,
			   default => {name => $main::VFILER,
				       ip => $main::VFILER_IP,
				       interface => $main::VFILER_IF,}
			   },
		
		supportability => {
			type => SCALAR,
			optional => 1,
			},
		
		attach_to_volume => {
			type => SCALAR,
			optional => 1,
			},
		
		file_clone => {
			type => SCALAR,
			optional => 1,
			},
		
		snmp => {
			type => SCALAR,
			optional => 1,
			},
		
		performance => {
			type => SCALAR,
			optional => 1,
			},
		
		_attach_dataset_info => {
			type => OBJECT,
			isa => "Compression::Compress",
			optional => 1,
			},
	});
	
	# WARNING!!!
	# The VolClone class is inheriting from the Compress class.
	# When adding new members to the Compress object, make sure they are also made available to the VolClone object.

	# Start building our object:
	my $self 	= {};
	$self->{boot_mode} = C_ONTAP_7MODE;

	my $skip_mount = FALSE;    ## Skip mount the volume (used for snapmirror destination volume)
	if ( (defined ($args{volume})) && (ref($args{volume}) eq "HASH")
	     &&  (defined ($args{volume}->{auto_mount}) )  && (!$args{volume}->{auto_mount}) ){
			$skip_mount = TRUE;
	}
	
	if (defined($args{vserver}))
	{
		$self->{vserver} = $args{vserver};
		$self->{boot_mode} = C_ONTAP_CMODE;
	}
	$self->{_debug} = $args{debug};
	$self->{_disable_cleanup} = $args{disable_cleanup};
	if (defined($args{attach_to_volume}))
	{
		$self->{_attach_only} = TRUE;
		$self->{_attach_to_volume} = $args{attach_to_volume};
	}
	else
	{
		$self->{_attach_only} = FALSE;
		$self->{_attach_to_volume} = undef;
	}
	$self->{_creator_pid} = $$;
	$self->{_count_volumes} = (defined($main::COUNT_VOLUMES)) ? $main::COUNT_VOLUMES : FALSE;
	logcomment("Compress: setup called with args: " . Dumper(%args)) if ($self->{_debug});
	##########################
	# CLIENT
	##########################
	my $client_args = $args{client};
        if ((defined($client_args) && (ref($client_args) eq "HASH")))
        {
                # Second-layer hash validation for client args
                logcomment("Compress: new: client attributes passed in as HASHREF") if ($self->{_debug});
                my %client_validation = validate_with
                        ( params => $client_args,
                          spec   => {name => { type => SCALAR, default => $main::CLIENT },
                                     type => { type => SCALAR, default => C_DFLT_CLIENT_TYPE },
                                     },
                          allow_extra => 0,
                        );

                $self->{_client}->{name} = $client_validation{name};
		$self->{_client}->{_required_type} = $client_validation{type};
        }
        else
        {
                logcomment("Compress: new: client attribute passed in as SCALAR or undefined") if ($self->{_debug});
                $self->{_client}->{name} = (defined($client_args)) ? $client_args : $main::CLIENT;
                $self->{_client}->{_required_type} = C_DFLT_CLIENT_TYPE;
        }
	logcomment("Compress: client:\n" . Dumper($self->{_client})) if ($self->{_debug});

	my $client_host = host($self->{_client}->{name});
	if ($client_host->hosttype eq 'windows')
	{
		$self->{_client}->{type} = C_CLIENT_CIFS_TAG;
	}
	else
	{
		$self->{_client}->{type} = C_CLIENT_NFS_TAG;
		# TODO: the Dataset tree should be cleaned-up to rely on already obtained client type
		# instead of retrieving it again
		# TODO: This should be removed as soon as we support nfs on remote clients
		$self->{_client}->{name} = C_CLIENT_NFS_LOCAL;
	}
	
	if (($self->{_client}->{_required_type} ne C_CLIENT_TYPE_ANY) &&
	    ($self->{_client}->{_required_type} ne $self->{_client}->{type}))
	{
		logresult ('NA', "NOT APPLICABLE: Required client type is: " . $self->{_client}->{_required_type} . "\n" .
			         "Attempted to run on client of type: " . $self->{_client}->{type});
	}
	
	##########################
	# FILER
	##########################
	$self->{filer} = Compression::Filer->new(filer => $args{filer},
						 debug => $self->{_debug},
						 vserver => $self->{vserver},
						 boot_mode => $self->{boot_mode},
						 required_features => $args{required_features},
						 count_volumes => $self->{_count_volumes},
						 );
	$self->{host_name} = $self->{filer}->{host_name};
	
	$self->{_ui} = $self->{filer}->{_ui};
	
	my $conn = $self->{filer}->_get_underlying_filer_obj();
	
	logcomment("Created connection object to filer \"" . $self->{host_name} . "\"") if ($self->{_debug});
	
	##########################
	# AGGREGATE AND/OR VOLUME
	##########################
	my $volume_args = $args{volume};
	# Let Compression::Sis handle the a-sis initialization
	my $asis = $main::ENABLE_DEDUPE;
	if ((defined($volume_args) && (ref($volume_args) eq "HASH")) && (exists($volume_args->{asis})))
	{
		$asis = $volume_args->{asis};
		delete($volume_args->{asis});
	}
	
	my $configure_scheduler = C_DFLT_CONFIGURE_SCHEDULER;
	if ((defined($volume_args) && (ref($volume_args) eq "HASH")) && (exists($volume_args->{configure_scheduler})))
	{
		$configure_scheduler = $volume_args->{configure_scheduler};
		delete($volume_args->{configure_scheduler});
	}
	
	my $vol_compression_status = C_ENABLE_COMPRESSION;
	if ((defined($volume_args) && (ref($volume_args) eq "HASH")) && (exists($volume_args->{compress_enabled})))
	{
		$vol_compression_status = $volume_args->{compress_enabled};
		delete($volume_args->{compress_enabled});
	}
	
	# We need to check first if traditional volume is needed => we should not create an aggregate
	if (defined($volume_args) && (ref($volume_args) eq "HASH") &&
	    defined($args{volume}->{type}) && ($args{volume}->{type} eq C_VOL_TYPE_TRAD))
	{
		if ($self->{boot_mode} eq C_ONTAP_CMODE)
		{
			logresult ('NA', "NOT APPLICABLE: Trad vols are not supported in C-mode");
		}
		
		logcomment("Compress: new: will be creating a traditional volume") if ($self->{_debug});
		$self->{_trad_vol} = TRUE;

		# For traditional volumes we should not have an aggregate defined
		if (defined($args{aggregate}))
		{
			logresult("script", "Traditional-volume setup should not have an aggregate specified:\n" . Dumper(%args));
		}
		
		if ($self->{_attach_only} == TRUE)
		{
			logresult('script', "The attach-only feature is not supported for traditional volumes yet");
		}

		$self->{volume} = Compression::Volume->new(filer => $conn,
							   volume => $args{volume},
							   vserver => $self->{vserver},
							   boot_mode => $self->{boot_mode},
							   ui => $self->{_ui},
							   client => $self->{_client}->{name},
							   debug => $self->{_debug},
							   disable_cleanup => $self->{_disable_cleanup},
							   _fake_mount => TRUE, 
							   );
	}
	else
	{
		$self->{_trad_vol} = FALSE;
		$self->{aggregate} = Compression::Aggregate->new(conn => $conn,
								 ui => $self->{_ui},
								 aggregate => $args{aggregate},
								 vserver => $self->{vserver},
								 boot_mode => $self->{boot_mode},
								 debug => $self->{_debug},
								 attach => $self->{_attach_only},
								 disable_cleanup => $self->{_disable_cleanup},
								 node	=> $self->{host_name},
								 );
		
		$self->{aggr_name} = $self->{aggregate}->{aggr_name};
		logcomment("Created aggregate object for aggregate \"" . $self->{aggr_name}) if ($self->{_debug});
		
		###################################
		# Vfiler
		###################################
		my $vfiler_info = $args{vfiler};
		my $fake_mount = TRUE; # volume will be mounted in Dataset
		# Note: because of burt 442667, we are only attaching one volume to the vfiler
		# To make sure we are attaching a test volume, do not create the vfiler if compression is required to be off (base vol)
		# Once this burt is fixed, the compression-related conditional should be removed
		if ((defined($vfiler_info->{name})) && ($vol_compression_status == TRUE))
		{
			use Compression::7Mode::Vfiler;
			
			if ($self->{boot_mode} eq C_ONTAP_CMODE)
			{
				logresult ('NA', "NOT APPLICABLE: Vfilers are not supported in C-mode");
			}
			
			if ((!defined($vfiler_info->{ip})) || (!defined($vfiler_info->{interface})))
			{
				logresult('script', "If VFILER is used, then VFILER_IP and VFILER_IF must be passed in as well");
			}
			$vfiler_info->{root_vol_aggr} = $self->{aggregate}->_get_underlying_aggr_obj();
			$vfiler_info->{disable_cleanup} = $self->{_disable_cleanup};
			
			$self->{vfiler} = Compression::7Mode::Vfiler->new(%$vfiler_info);
			# Note: this is needed because of burt 442667 - once this is fixed, add the base volume to vfiler as well
			# (conditional should not be needed)
			if (defined($self->{vfiler}))
			{
				$self->{_vfiler_name} = $vfiler_info->{name};
				$fake_mount = FALSE;
			}
		}
		$self->{volume} = Compression::Volume->new(filer => $conn,
							   aggregate => $self->{aggregate}->_get_underlying_aggr_obj(),
							   volume => $args{volume},
							   vserver => $self->{vserver},
							   vfiler => $self->{vfiler},
							   boot_mode => $self->{boot_mode},
							   ui => $self->{_ui},
							   client => $self->{_client}->{name},
							   debug => $self->{_debug},
							   disable_cleanup => $self->{_disable_cleanup},
							   attach => $self->{_attach_to_volume},
							   _fake_mount => $fake_mount, 
							   );
	}
	$self->{vol_name} = $self->{volume}->{vol_name};
	logcomment("Created volume object for volume \"" . $self->{vol_name} . "\"") if ($self->{_debug});
	
	################
	# SUPPORTABILITY
	################
	# If the script specifically passed in a value,
	# only create the supportability object and leave it
	# up the the script to use it as needed. No extra checks from the library.
	# If the script did not pass anything, check the SUPPORTABILITY flag
	# If this is true, then automatically perform extra checks, where applicable.
	if (defined($args{supportability}))
	{
		if ($args{supportability} == TRUE)
		{
			use Compression::Supportability;
			$self->{supportability} = Compression::Supportability->new(conn => $conn,
										   boot_mode => $self->{boot_mode},
										   ui => $self->{_ui},
										   debug => $self->{_debug},);
		}
	}
	else
	{
		if ($main::SUPPORTABILITY == TRUE)
		{
			use Compression::Supportability;
			$self->{_supportability} = Compression::Supportability->new(conn => $conn,
										    boot_mode => $self->{boot_mode},
										    ui => $self->{_ui},
										    debug => $self->{_debug},);
		}
	}

	#########################################
	# Dataset, Decompress, CompressionScanner
	#########################################
	
	my $vfiler_mountpoint = (defined($self->{vfiler})) ? ($self->{volume}->_get_mountpoint()) : undef;
	$self->{dataset} = Compression::Dataset->new(volume => $self->{volume}->_get_underlying_vol_obj(),
						     vserver => $self->{vserver},
						     client => $self->{_client}->{name},
						     skip_mount => $skip_mount,
						     boot_mode => $self->{boot_mode},
						     ui => $self->{_ui},
						     debug => $self->{_debug},
						     has_debug_kernel => $self->{filer}->{_has_debug_kernel},
						     vfiler_mountpoint => $vfiler_mountpoint);

	# Skip get mount point only for CMode when auto_mount set to false, for DP type volumes
	# In 7Mode, we can mount the volume
	if  ( (!($skip_mount) ) ||        
	      ( ($skip_mount) && ( $self->{boot_mode} eq C_ONTAP_7MODE ) ) ) {   
		$self->{_vol_mount_path} = $self->{volume}->get_mount_path();
	}

	if ($args{_attach_dataset_info})
	{
		$self->{dataset}->attach_dataset_info(src_vol => $args{_attach_dataset_info},);
	}

	$self->{decompress} = Compression::Decompress->new(conn => $conn, 
							   volume => $self->{vol_name},
							   vserver => $self->{vserver},
							   boot_mode => $self->{boot_mode},
							   ui => $self->{_ui},
							   vfiler => $self->{_vfiler_name},
							   disable_cleanup => $self->{_disable_cleanup},
							   debug => $self->{_debug},
							   supportability => $self->{_supportability},);
	$self->{compress_scanner} = Compression::CompressionScanner->new(conn => $conn, 
									 volume => $self->{vol_name},
									 vserver => $self->{vserver},
									 boot_mode => $self->{boot_mode},
									 ui => $self->{_ui},
									 vfiler => $self->{_vfiler_name},
									 disable_cleanup => $self->{_disable_cleanup},
									 debug => $self->{_debug},
									 supportability => $self->{_supportability},);
	$self->{_utilities} = Compression::CompressionUtilities->new(conn => $conn, 
								     volume => $self->{vol_name},
								     vserver => $self->{vserver},
								     boot_mode => $self->{boot_mode},
								     ui => $self->{_ui},
								     vfiler => $self->{_vfiler_name},
								     disable_cleanup => $self->{_disable_cleanup},
								     debug => $self->{_debug},
								     supportability => $self->{_supportability},);
	
	#$self->{_utilities}->{_debug} = 1;
	###################################
	# Sis
	###################################

	if (!defined($asis))
	{
		# Get the release-dependent default
		$asis = $self->{_ui}->get_constant_value(constant_code => 'C_DFLT_INITIALIZE_SIS');
	}

	$self->{sis} = Compression::Sis->new(filer => $conn,
					     volume_path => $self->{_vol_mount_path},
					     volume => $self->{volume}->_get_underlying_vol_obj(),
					     vserver => $self->{vserver},
					     boot_mode => $self->{boot_mode},
					     configure_scheduler => $configure_scheduler,
					     debug => $self->{_debug},
					     disable_cleanup => $self->{_disable_cleanup},
					     ui => $self->{_ui},
					     initialize => $asis);

	
        ###################################
        # FileFolding
        ###################################
        if (defined($args{file_folding}))
        {
		$self->{file_folding} = Compression::FileFolding->new(filer => $conn,
								      volume => $self->{vol_name},
								      vserver => $self->{vserver},
								      boot_mode => $self->{boot_mode},
								      ui => $self->{_ui},
								      debug => $self->{_debug});
        }
	
	###################################
        # PerfStats
        ###################################
        if (defined($args{perf_stats}))
        {
		$self->{perf_stats} = Compression::Filer::PerfStats->new(filer => $self->{host_name},
									 stats_type => $args{perf_stats},
									 ui => $self->{_ui},
									 debug => $self->{_debug});
        }
	
	###################################
        # FileClone
        ###################################
        if ((exists($args{file_clone})) && ($args{file_clone} == TRUE))
        {
		$self->{file_clone} = Compression::FileClone->new(filer => $conn,
								  volume => $self->{vol_name},
								  vserver => $self->{vserver},
								  boot_mode => $self->{boot_mode},
								  ui => $self->{_ui},
								  debug => $self->{_debug});
        }
	
	###################################
	# SNMP
	###################################
	if ((exists($args{snmp})) && ($args{snmp} == TRUE))
	{
		$self->{snmp} = Compression::SNMP->new(filer => $self->{host_name},
						       volume => $self->{vol_name},
						       ontap => $self->{_ui}->{version}->{ontap},
						       boot_mode => $self->{boot_mode},
						       debug => $self->{_debug},);
	}
	
	###################################
	# Performance
	###################################
	if ((exists($args{performance})) && ($args{performance} == TRUE))
	{
		$self->{performance} = Compression::Performance->new(filer => $self->{host_name},
								     boot_mode => $self->{boot_mode},
								     debug => $self->{_debug},);
	}
	
	bless($self, $class);
	
	
	###################################
	# Do extra setup steps: snaplock
	###################################
	if ($self->{aggregate}->{_snaplock})
	{
		$self->{volume}->set_snaplock_retention_period(default => C_VOL_DFLT_SNAPLOCK_RETENTION,
							       minimum => C_VOL_MIN_SNAPLOCK_RETENTION,
							       maximum => C_VOL_MAX_SNAPLOCK_RETENTION);
	}
	
	############################################
	# Do extra setup steps: enable compression
	############################################
	if ($vol_compression_status == TRUE)
	{
		logcomment( "Compression requested to be on..." ) if ($self->{_debug});
		$self->enable(volume => $self->{vol_name}, check_if_already_set => TRUE);
	}
	elsif (($vol_compression_status == FALSE))
	{
		# This will cover the case when the volume was recycled, but disable even if not recycled
		logcomment( "Compression requested to be off..." ) if ($self->{_debug});
		$self->disable(volume => $self->{vol_name}, check_if_already_set => TRUE);
	}
	elsif ($vol_compression_status == C_FLAG_KEEP_VALUE_AS_IS)
	{
		logcomment( "Compression option will be kept as is..." ) if ($self->{_debug});
	}
	else
	{
		logresult("script", "Volume: Invalid value for \"compress_enabled\" flag: $vol_compression_status");
	}
	
	logcomment("Compress: Setup Complete..\n\n") if ($self->{_debug});
	return $self;

}


## @name ::toggle
## @summary Flips the compression setting for the volume.
##
## @description
## Flips the compression setting for the volume.
##
## @return Hash with rc and output.
##
## @examples
## $compress->toggle();
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##

sub toggle
{
	my $class 	= shift;
	my (%args) 	= validate( @_,
			{
				# No args
			} );

	my %current_setting = $class->is_enabled(expected => C_RC_FAILURE_NON_FATAL);
	
	# For now ignore inline
	if ($current_setting{rc} == _C_RC_COMPR_OFF)
	{
		return $class->enable();
	}
	else
	{
		return $class->disable();
	}
}


## @name ::enable
## @summary Enables compression for the volume.
##
## @description
## Enables compression for the volume.
##
## @return Hash with rc and output.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the Compress object was created.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @arg check_if_already_set Optional: Flag to enable the library to first check the current setting.
## If compression is already enabled, the subroutine will not attempt to enable it again.
## Instead, it will return success. The default value for check_if_already_set is C_DFLT_CHECK_OPTION_FIRST.
##
## @arg inline Optional: Flag to control if inline compression should be enabled as well (RR and after).
## Legal values: TRUE/FALSE. Default value: release-dependent.
##
## @arg background Optional: Flag to control if background compression should be enabled (RR and after).
## Legal values: TRUE/FALSE. Default value: TRUE.
## It is highly recommended to use check_if_already_set false whenever background is false.
## WARNING: Setting background to false means that compression will not be enabled.
## This flag should only be used for negative-testing.
##
## @arg check_containment_version Optional: Flag to enable containment version check after enabling compression.
## Legal values: TRUE/FALSE. Default value: $main::CHECK_CONTAINMENT_VERSION.
##
## @examples
## $compress->enable( expected => C_RC_NO_COMPRESSION_LICENSE );
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @note
## If the SUPPORTABILITY cmd-line argument is 1, then,
## along with this action, the correspondent supportability message will be also checked.
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub enable
{
	my $class = shift;
	my (%args) 	= validate( @_,
		{
			volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
			vserver => { type => SCALAR|UNDEF, optional => 1, default => $class->{vserver} },
			expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
			check_if_already_set => { type => BOOLEAN, optional => 1, default => C_DFLT_CHECK_OPTION_FIRST },
			inline => { type => BOOLEAN, optional => 1, default => $main::ENABLE_INLINE },
			background => { type => SCALAR, optional => 1, default => TRUE },
			check_containment_version => { type => SCALAR, optional => 1, default => $main::CHECK_CONTAINMENT_VERSION },
		} );
	
	my $check_containment_version = delete($args{check_containment_version});
	
	my %return = $class->{_utilities}->enable_compression(%args);
	
	if (($check_containment_version == TRUE) && ($args{expected} == C_RC_SUCCESS))
	{
		$class->verify_containment_version(compression_state => C_STATE_AT_LEAST_ONCE_ENABLED);
	}
	
	return %return;
}


## @name ::disable
## @summary Disables the compression for the volume.
##
## @description
## Disables the compression for the volume.
##
## @return Hash with rc and output.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @arg volume Optional: Name of volume for which to disable compression.
## Defaults to the volume for which the Compress object was created.
##
## @arg all_volumes Optional: Flag to disable compression on all existent volumes.
##
## @arg skip_root_volume Optional: Only considered if 'all_volumes' option is TRUE.
## Boolean flag to decide if compression should be attempted to be disabled for the root volume.
## Defaults to TRUE.
##
## @arg skip_offline_restricted Optional: Only considered if 'all_volumes' option is TRUE.
## Boolean flag to decide if compression should be attempted to be disabled for offline/restricted volumes.
## Defaults to TRUE.
##
## @arg check_if_already_set Optional: Flag to enable the library to first check the current setting.
## If compression is already disabled, the subroutine will not attempt to disable it again.
## Instead, it will return success. The default value for check_if_already_set is C_DFLT_CHECK_OPTION_FIRST.
## If the all_volumes argument is TRUE, then check_if_already_set is automatically set to TRUE.
##
## @arg inline Optional: Flag to control if only inline compression should be disabled.
## Legal values: TRUE/FALSE. Default value: release-dependent.
##
## @notes
## If all_volumes flag is set to true, the expected result is considered to always be success (no 'expected' handling).
##
## @examples
## $compress->disable();
## $compress->disable( all_volumes => TRUE );
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @note
## If the SUPPORTABILITY cmd-line argument is 1, then,
## along with this action, the correspondent supportability message will be also checked.
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub disable
{
	my $class 	= shift;
	
	return $class->{_utilities}->disable_compression(@_);
}


## @name ::is_enabled
## @summary Verifies that compression is enabled for the volume.
##
## @description
## Verifies that compression is enabled for the volume.
##
## @return Hash with rc and output.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes that compression be enabled.
## Legal values: C_RC_SUCCESS, C_RC_FAILURE
##
## @arg inline Optional: Expected return code (defined in Compression::Constants) for inline compression check.
## Legal values: C_RC_SUCCESS, C_RC_FAILURE.
##
## @examples
## $compress->is_enabled( expected => C_RC_SUCCESS );
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_enabled
{
        my $class = shift;
	
	return $class->{_utilities}->is_compression_enabled(@_);
}


## @name ::wait_for_cp
## @summary Wait for specified number of CPs to occur
##
## @description
## Wait for specified number of CPs to occur, usually used after enabling compression on the volume.
##
## @return Hash with rc and output (empty).
##
## @arg num_cp Optional: Number of CPs to wait.
## Default value: 1.
##
## @arg wait_time Optional: Extra time to wait, in seconds.
## It will be additional wait time after num_cp has elapsed.
## Default value: 1.
##
## @examples
## $compress->wait_for_cp( num_cp => 2, wait_time => 5 );
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub wait_for_cp
{
	my $class = shift;
	
	return $class->{_utilities}->wait_for_CP(@_);
}


## @name ::get_savings
## @summary Retrieves the overall savings of the volume.
##
## @description
## Retrieves the overall savings of the volume using the 'df -S<modifier>' command.
##
## @return Hash with rc, output, and savings.
##
## @arg modifier Optional: Output modifier to be used when retrieving savings.
## Defaults to 'k' (which means that the command to obtain the savings will be 'df -Sk').
##
## @arg all_volumes Optional: Flag to decide if savings should be retrieved for all volumes.
## Defaults to FALSE.
##
## @arg overwrite_args_string Optional: String to be passed as argument to the df command.
## The command that will be sent to the filer will be 'df -S<modifier> <overwrite_args_string>.
## If this argument is used, any successful command output will be processed such that
## the 'savings' hash will have as keys the exact string that identifies the data line.
## For example, if overwrite_args_string is 'vol_name' , the corresponding savings key hash will be
## '/vol/vol_name/', instead of 'vol_name', as it would be when getting the savings for the associated volume.
## Another difference from the case of the associated volume is that
## no checking is done on the output to ensure that there indeed was an entry in the output for 'vol_name'.
## If this is desired, then the script should specifically check $result{savings}->{vol_name}.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @examples
## $compress->get_savings( modifier => 'g' );
## $compress->get_savings( all_volumes => TRUE, modifier => 'm' );
## $compress->get_savings( overwrite_args_string => 'vol1 vol2', expected => C_RC_INVALID_COMMAND_ARGUMENT );
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_savings
{
	my $class = shift;
	
	return $class->{_utilities}->get_space_savings(@_);
}


## @name ::verify_savings
## @summary Verifies space savings for the volume based on expected  values.
##
## @description
## Verifies space savings for the volume based on expected  values.
##
## @return Hash with rc, output, and obtained savings.
## Logs fatal result if values (expected vs obtained) do not match.
##
## @arg expected_compressed Optional: Expected compression savings on the volume.
##
## @arg expected_used Optional: Expected used space on the volume.
##
## @arg expected_deduplicated Optional: Expected a-sis savings on the volume.
##
## @arg modifier Optional: Output modifier to be used when retrieving savings.
## Defaults to 'k' (which means that the command to obtain the savings will be 'df -Sk').
## The expected_compressed value should be passed in according to the modifier.
##
## @arg allow_percentage_error Optional: Allow the obtained savings value to differ from the expected value
## by this percentage. Defaults to 0.
##
## @arg run_scanner_if_compressed_mismatch Optional: Run the compression scanner if the
## compression savings do not match the expected value.
## Legal values: TRUE/FALSE. Default value: release-dependent (FALSE for pre-RR, TRUE for RR).
##
## @examples
## $compress->verify_savings( modifier => 'k', expected_compressed => 8 );
## $compress->verify_savings( expected_compressed => 8, allow_percentage_error => 15 );
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub verify_savings
{
	my $class = shift;
	my (%args) 	= validate( @_,
			{
				expected_compressed => { type => SCALAR, optional => 1, regex => qr/^\d+$/, },
				expected_used => { type => SCALAR, optional => 1, regex => qr/^\d+$/, },
				expected_deduplicated => { type => SCALAR, optional => 1, regex => qr/^\d+$/, },
				modifier => { type => SCALAR, optional => 1, default => 'k' },
				allow_percentage_error => { type => SCALAR, regex => qr/^\d+$/, default => 0, },
				run_scanner_if_compressed_mismatch => { type => BOOLEAN, default => $main::SCAN_FOR_SAVINGS, },
				_iteration => { type => SCALAR, regex => qr/^\d+$/, default => 1, },
			} );

	logcomment("Compress: Verify Savings: " . Dumper(%args) ) if ($class->{_debug});

	my %result = ();
	my %expected;
	my $allow_percentage_error = $args{allow_percentage_error};
	$expected{compressed} = $args{expected_compressed} if defined($args{expected_compressed});
	$expected{used} = $args{expected_used} if defined($args{expected_used});
	$expected{deduplicated} = $args{expected_deduplicated} if defined($args{expected_deduplicated});
	my $volume = $class->{vol_name};
	my $current_iteration = $args{_iteration};
	my $max_allowed_iterations = 3;

	if ((!defined($expected{compressed})) && (!defined($expected{used})) && (!defined($expected{deduplicated}))) {
		logresult("script", "verify_savings: Must specify at least expected_compressed, expected_deduplicated, or expected_used");
	}

	my %obtained_savings = $class->get_savings( modifier => $args{modifier} );
	logcomment("Obtained savings: " . Dumper(%obtained_savings)) if ($class->{_debug});
	$result{output} = $obtained_savings{savings};

	foreach my $field (sort keys %expected)
	{
		my $obtained = $obtained_savings{savings}{$volume}{$field};
		my $expected_val = $expected{$field};
		# Decide if we need to run the compression scanner if mismatch
		# - only for compression savings
		# - based on release, unless directly passed in
		my $run_compr_scanner = ($field eq 'compressed') ? $args{run_scanner_if_compressed_mismatch} : FALSE;
		if (!defined($run_compr_scanner))
		{
			# Get the release-dependent default
			$run_compr_scanner = $class->{_ui}->get_constant_value(constant_code => 'RUN_SCANNER_IF_SAVINGS_MISMATCH');
		}
		
		# Still... there are some releases for which we want this flag to be ignored
		my $ignore_run_compr_scanner = $class->{_ui}->get_constant_value(constant_code => 'IGNORE_RUN_SCANNER_FLAG');
		if ($ignore_run_compr_scanner == TRUE)
		{
			$run_compr_scanner = FALSE;
		}
		
		logcomment("Verifying $field:\n" .
			   "Expected: $expected_val\n" .
			   "Obtained: $obtained\n" .
			   "Allowed percentage error: $allow_percentage_error");
		
		my $non_fatal_failure = $run_compr_scanner || ($current_iteration <= $max_allowed_iterations);
		
		my $rc = compare_values(obtained_value => $obtained,
                                        expected_value => $expected_val,
                                        allow_percentage_error => $allow_percentage_error,
                                        values_meaning => "verify_savings $field",
                                        non_fatal_failure => $non_fatal_failure,
                                        verbose => $class->{_debug});
		
		if ($rc == C_RC_FAILURE)
		{
			if ($current_iteration <= $max_allowed_iterations)
			{
				# Give it more chances. May be just timing
				logcomment("verify_savings: failure to match $field\n" .
					   "Attempt: $current_iteration out of $max_allowed_iterations\n" .
					   "Waiting for one CP and trying again...");
				$class->wait_for_cp(); 
				
				$args{_iteration} = $current_iteration + 1;
				return $class->verify_savings(%args);
			}
			
			# We should never fail here, but better safe than sorry
			if (($field ne 'compressed') || ($run_compr_scanner == FALSE))
			{
				logresult("fatal", "Expected verify_savings $field mismatch:\n" .
					           "Obtained $obtained, expected $expected_val\n" .
						   "Allowed percentage error: $allow_percentage_error");
			}
			
			logcomment("Expected verify_savings $field mismatch:\n" .
				   "Obtained $obtained, expected $expected_val\n" .
				   "Allowed percentage error: $allow_percentage_error\n" .
				   "Running the compression scanner...");
			
			eval
			{
				$class->{compress_scanner}->start();
				$class->{compress_scanner}->wait_for_scanned_data();
			};
			
			if ($@)
			{
				logcomment("Failure while running compression scanner:\n" . $@);
			}
			
			logcomment("Checking compression savings again...");
			$class->{filer}->sync_blocks_to_disk();
			$class->wait_for_cp();
			%obtained_savings = $class->get_savings( modifier => $args{modifier} );
			$result{output} = $obtained_savings{savings};
			$obtained = $obtained_savings{savings}{$volume}{$field};
			
			$rc = compare_values(obtained_value => $obtained,
                                             expected_value => $expected_val,
                                             allow_percentage_error => $allow_percentage_error,
                                             values_meaning => "verify_savings $field",
                                             non_fatal_failure => FALSE,
                                             verbose => $class->{_debug});
		}
		
		$result{rc} = $rc;
	}
	return %result;
}


## @name ::verify_reserved_space
##
## @summary verify reserved space
##
## @description
## verify reserved space based on expected values
##
## @return hash of status (rc)
## Logs fatal result if values (expected vs obtained) do not match.
##
## @arg expected_reserved Required: Expected reserved on the volume in KB
##
## @arg allow_percentage_error Optional: Allow the obtained reserved value to differ from the expected value
## by this percentage. Defaults to 0.
##
## @examples
##  $compress->verify_reserved_space( expected_reserved => 10, allow_percentage_error => 2 );
##
sub verify_reserved_space
{
	my $class 	= shift;
	my (%args) 	= validate( @_,
			{
				expected_reserved => { type => SCALAR, regex => qr/^\d+$/, },
				allow_percentage_error => { type => SCALAR, regex => qr/^\d+$/, default => 0, },
			} );
	my %result = ();
	my %res = $class->{volume}->get_reserved_space;
	
	my $max_attempts = 5;
	for (my $i = 1; $i <= $max_attempts; $i++)
	{
		my $non_fatal_failure = ($i == $max_attempts) ? FALSE : TRUE;
		logcomment("\nVerifying reserved space\n" .
			   "Expected: " . $args{expected_reserved} . "\n" .
			   "Obtained: " . $res{reserved} . "\n" .
			   "Allowed percentage error: " . $args{allow_percentage_error});
		$result{rc} = compare_values(obtained_value => $res{reserved},
					     expected_value => $args{expected_reserved},
					     allow_percentage_error => $args{allow_percentage_error},
					     values_meaning => 'reserved',
					     non_fatal_failure => $non_fatal_failure,
					     verbose => $class->{_debug});
		
		if ($result{rc} == C_RC_SUCCESS)
		{
			last;
		}
		$class->wait_for_cp();
	}
	return %result;
}

## @name ::verify_snapshot_info
##
## @summary verify snapshot info
##
## @description
## verify snapshot info
##
## @return hash of status (rc)
## Logs fatal result if values (expected vs obtained) do not match.
##
## @arg name	Required: Name of snapshot
## @arg expected_percent_used Required: Expected used by snapshot
## @arg allow_percentage_error Optional: Allow the obtained reserved value to differ from the expected value
## by this percentage. Defaults to 0.
##
## @examples
##  $compress->verify_snapshot_info( name=>$SNAPSHOT, expected_percent_used=>10, allow_percentage_error=>2 );
##
sub verify_snapshot_info
{
	my $class 	= shift;
	my (%args) 	= validate( @_,
			{
				name => { type=>SCALAR },
				expected_percent_used => { type => SCALAR, regex => qr/^\d+$/, },
				allow_percentage_error => { type => SCALAR, regex => qr/^\d+$/, default => 0, },
			} );
	my %result = ();
	my %expected;
	$expected{'percentage-of-used-blocks'}	= $args{expected_percent_used} if defined($args{expected_percent_used});

	my %res = $class->{volume}->get_snapshot_info(name=>$args{'name'});
	foreach my $field (sort keys %expected)
	{
		my $obtained = $res{info}->{$field};
		logcomment("verify_snapshot_info: $args{name}\n".
					"$field\nexpected:$expected{$field} obtained:$obtained ".
					"allow_percentage_error:$args{allow_percentage_error}" );
		$result{rc} = compare_values(obtained_value => $obtained,
					     expected_value => $expected{$field},
					     allow_percentage_error => $args{allow_percentage_error},
					     values_meaning => 'snapshot info ' . $field ,
					     verbose => $class->{_debug});
	}
	return %result;
}

## @name ::rename_volume
## @summary Renames the volume associated to the object.
##
## @description
## Renames the volume associated to the object.
##
## @return Hash with rc, output (new volume name).
##
## @arg new_name Optional: New name for the volume.
##
## @examples
## $compress->rename_volume( new_name => 'renamed_vol' );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub rename_volume
{
	my $class = shift;
	my (%args) 	= validate( @_,
			{
				new_name => { type => SCALAR, optional => 1, },
			} );

	logcomment("Compress: rename_volume input: " . Dumper(%args) ) if ($class->{_debug});

	$class->{volume}->_rename(new_name => $args{new_name});
	my $new_name = $class->{volume}->{vol_name};
	$class->{_vol_mount_path} = $class->{volume}->get_mount_path();
	$class->{decompress}->_update_volume_name(new_name => $new_name);
	$class->{compress_scanner}->_update_volume_name(new_name => $new_name);
	$class->{_utilities}->_update_volume_name(new_name => $new_name);
	$class->{sis}->_update_volume_path(new_path => $class->{_vol_mount_path});
	$class->{vol_name} = $new_name;

	return (rc => C_RC_SUCCESS, output => $class->{vol_name});
}


## @name ::clone_volume
## @summary Clones the volume associated to the object.
##
## @description
## Clones the volume associated to the object.
##
## @return A Compress::VolClone object.
##
## @arg clone_name Optional: Name for the clone of the volume.
##
## @arg snapshot Optional: Name of existing snapshot of parent volume, to serve
## as the basis of the clone volume.
##
## @arg space_reservation Optional: Space reservation setting for the clone volume
## Legal values: none/file/volume. 
##
## @examples
## my $clone = $compress->clone_volume( clone_name => 'clone_vol' );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub clone_volume
{
	my $class 	= shift;
	my (%args) 	= validate( @_,
			{
				clone_name => { type => SCALAR, optional => 1, },
				snapshot => {type => SCALAR, optional => 1, },
				space_reservation => {type => SCALAR, optional => 1, },
			} );

	logcomment( "Compress: volume_clone\n". Dumper(%args) ) if ($class->{_debug});

	return Compression::VolClone->new(source => $class,
					  name => $args{clone_name},
					  snapshot => $args{snapshot},
					  space_reservation => $args{space_reservation},);
}


## @name ::get_statistics
## @summary Displays the statistics for the volume.
##
## @description
## Displays the statistics for the volume.
##
## @arg volume Optional: Name of volume for which to display statistics.
## Defaults to the volume for which the Compress object was created.
##
## @return Hash with rc, statistics, and output.
##
## @examples
## my %res = $compress->get_statistics();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_statistics
{
	my $class = shift;
	
	return $class->{_utilities}->get_compression_statistics(@_);
}


## @name ::get_compression_attempted_bits
## @summary Obtains the "compression attempted" fingerprint bits.
##
## @description
## Obtains the "compression attempted" fingerprint bits.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the Compress object was created.
##
## @arg inode Optional: Inode for which to get the fingerprints.
##
## @arg fbns Optional: Fbns for which to get the fingerprints.
## Can be: specific value (ex: 8), range (ex: "3-7"), or "all".
## Default: "all".
##
## @arg aggregate Optional: Boolean flag to look for aggregate-level fingerprints.
## Default: FALSE.
##
## @arg vvol Optional: Boolean flag to look for vvol-level fingerprints.
## Default: FALSE.
##
## @arg changelog Optional: Boolean flag to look for changelog fingerprints.
## Default: TRUE.
##
## @arg remove_stale_entries Optional: Boolean flag to use if only the non-stale entries are wanted
## (latest cp-count). Default: FALSE.
##
## @arg expected Optional: Optional: Expected return code (C_RC_SUCCESS/C_RC_FAILURE).
## Default: C_RC_SUCCESS.
##
## @return Hash with rc, count, aggregate, vvol, checking_stats, changelog_0, changelog_1, and output.
##
## @examples
## $compress->get_compression_attempted_bits(inode => 96,
##                                           fbns => "4-7",
##                                           changelog => FALSE,
##                                           aggregate => TRUE);
## $compress->get_compression_attempted_bits(inode => 96,
##                                           fbns => 8,
##                                           remove_stale_entries => TRUE,
##                                           expected => C_RC_FAILURE);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_compression_attempted_bits
{
	my $class = shift;
	
	return $class->{_utilities}->obtain_compression_attempted_bits(@_);
}


## @name ::verify_compression_attempted
## @summary Verifies the "compression attempted" fingerprint bits.
##
## @description
## Verifies the "compression attempted" fingerprint bits.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the Compress object was created.
##
## @arg inode Optional: Inode for which to get the fingerprints.
##
## @arg fbns Optional: Fbns for which to get the fingerprints.
## Can be: specific value (ex: 8), range (ex: "3-7"), or "all".
## Default: "all".
##
## @arg aggregate Optional: Boolean flag to look for aggregate-level fingerprints.
## Default: FALSE.
##
## @arg vvol Optional: Boolean flag to look for vvol-level fingerprints.
## Default: FALSE.
##
## @arg changelog Optional: Boolean flag to look for changelog fingerprints.
## Default: TRUE.
##
## @arg remove_stale_entries Optional: Boolean flag to use if only the non-stale entries are wanted
## (latest cp-count). Default: FALSE.
##
## @arg expected Optional: Expected return code (C_RC_COMPRESSION_ATTEMPTED/C_RC_COMPRESSION_NOT_ATTEMPTED).
## Default: C_RC_COMPRESSION_ATTEMPTED.
##
## @return Hash with rc, aggregate, vvol, checking_stats, changelog_0, changelog_1, and output.
##
## @examples
## my $compress->verify_compression_attempted(inode => 96,
##                                            fbns => "all",
##                                            expected => C_RC_COMPRESSION_NOT_ATTEMPTED);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub verify_compression_attempted
{
	my $class = shift;
	
	return $class->{_utilities}->verify_compression_attempted_bits(@_);
}


## @name ::get_containment_version
## @summary Obtains containment version for the volume.
##
## @description
## Obtains containment version for the volume.
##
## @return Hash with rc, containment_version, and output.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the Compress object was created.
##
## @examples
## $compress->get_containment_version();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_containment_version
{
	my $class = shift;
	
	return $class->{_utilities}->get_compression_containment_version(@_);
}


## @name ::verify_containment_version
## @summary Verifies containment version for the volume.
##
## @description
## Verifies containment version for the volume.
##
## @return Hash with rc, containment_version, and output.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the Compress object was created.
##
## @arg containment_version Optional: Expected value for the current containment version.
##
## @arg highest_containment_version Optional: Expected value for the highest containment version.
##
## @arg compression_state Optional: Only valid if none of the containment_version and highest_containment_version are specified.
## State of compression at the moment the containment version is verified. The release-dependent expected values are calculated
## depending on this. Legal values: C_STATE_AT_LEAST_ONCE_ENABLED (compression was at least once enabled for this volume),
## C_STATE_NEVER_ENABLED, defined in Constants.pm.
## Default value: C_STATE_WAS_ENABLED.
##
## @note
## If none of the containment_version and highest_containment_version are specified,
## then both fields are checked against the release-dependent expected values.
## If any of the above fields are specified, then only the ones that were specified
## are checked against the passed-in values.
##
## @examples
## $compress->verify_containment_version();
## $compress->verify_containment_version(compression_state => C_STATE_NEVER_ENABLED);
## $compress->verify_containment_version(compression_state => C_STATE_AT_LEAST_ONCE_ENABLED);
## $compress->verify_containment_version(highest_containment_version => 0);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub verify_containment_version
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
			containment_version => { type => SCALAR, optional => 1, },
			highest_containment_version => { type => SCALAR, optional => 1, },
			compression_state => { type => SCALAR, optional => 1, default => C_STATE_AT_LEAST_ONCE_ENABLED},
		} );
	
	my $volume = delete($args{volume});
	my $compression_state = delete($args{compression_state});

	# If none of the extected values are defined, then check both against the release-dependent values
	# Otherwise, check only the one that was passed in
	if ((!defined($args{containment_version})) && (!defined($args{highest_containment_version})))
	{
		my @allowed_compression_states = (C_STATE_AT_LEAST_ONCE_ENABLED, C_STATE_NEVER_ENABLED);
		if (!grep(/^$compression_state$/, @allowed_compression_states))
		{
			logresult('script', "Illegal compression_state argument value passed to verify_containment_version: $compression_state");
		}
		
		$args{containment_version} = $class->{_ui}->get_constant_value(constant_code => "CONTAINMENT_VERSION_$compression_state");
		$args{highest_containment_version} = $class->{_ui}->get_constant_value(constant_code => "HIGHEST_CONTAINMENT_VERSION_$compression_state");
	}
	
	my %current_values = $class->get_containment_version(volume => $volume);
	logcomment("get_containment_version returned:\n" . Dumper(%current_values)) if ($class->{_debug});
	my $obtained_vals = $current_values{containment_version};
	
	foreach my $field (keys (%args))
	{
		my $expected = $args{$field};
		my $obtained = $obtained_vals->{$field};
		
		logcomment("Verifying $field for volume $volume\n" .
			   "Expected value: $expected\n" .
			   "Obtained value: $obtained");
		
		Compression::Utils::compare_values(obtained_value => $obtained,
						   expected_value => $expected,
						   allow_percentage_error => 0,
						   values_meaning => "$field for volume $volume",
						   non_fatal_failure => FALSE,
						   verbose => $class->{_debug});
	}
	
	# RC in %current_values is already success
	return %current_values;
}


## @name ::get_compression_stats
## @summary Obtains compression stats for the volume.
##
## @description
## Obtains compression stats for the volume.
##
## @return Hash with rc, compression_stats, and output.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the Compress object was created.
##
## @examples
## $compress->get_compression_stats();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_compression_stats
{
	my $class = shift;
	
	return $class->{_utilities}->show_compression_stats(@_);
}

## @name ::compression_stats_start_collection
## @summary Start compression statistics collection
##
## @description
## Start compression statistics collection. This  subcommand  must be used
## before the stats stop command.
##
## @return Hash with rc.
##
## @examples
## $compress->compression_stats_start_collection();
##
## @author weili@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub compression_stats_start_collection
{
	my $class = shift;
	
	return $class->{_utilities}->compression_stats_start_collection(@_);
}

## @name ::compression_stats_stop_collection
## @summary Stop compression  statistics collection.
##
## @description
## Stop compression statistics collection
##
## @return Hash with rc, compression_stats, and output.
##
## @examples
## $compress->compression_stats_stop_collection();
##
## @author weili@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub compression_stats_stop_collection
{
	my $class = shift;
	
	return $class->{_utilities}->compression_stats_stop_collection(@_);
}


## @name ::verify_compression_stats
## @summary Verifies compression stats for the volume.
##
## @description
## Verifies compression stats for the volume.
##
## @return Hash with rc, compression_stats, and output.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the Compress object was created.
##
## @arg fields Required: Hashref of name=>value for fields to check. Field names are
## the same as the keys used in the compression_stats returned by  get_compression_stats.
## Available fields: num_blocks_saved, num_blocks_cdata, num_blocks_uncompressed,
## num_buf_decompressed, num_cg_decompressed, num_partial_write, num_valid_buf,
## num_invalid_raid_read, num_valid_buf_suspended, num_cow_cg, num_blocking_io,
## num_vol_decomp_ahead_busy, num_vol_decomp_no_resource, num_vol_decomp_zero_holes,
## num_decompress_suspends, avg_vol_decompress_time, avg_decompress_cg_time, num_compress_attempts,
## num_compress_fail, num_not_enough_savings, num_no_mem, num_compress_suspends,
## num_cgs_frz_attempts, num_cgs_frz_failed, num_buf_attempts, num_nocompress_files,
## num_min_input, num_cgs_with_holes, num_next_cp, num_wait_for_compressor, avg_compress_cg_time
## num_extra_cp_reads, num_compress_ahead, num_overload, hole_to_data, hole_to_compressed,
## data_to_hole, num_write_uncompressed, num_eos_bufs, num_cgs_not_fully_dirty, num_wafl_decompress,
## num_decompress_read_bad_blks, num_decompress_mark_hw_bad, num_decompress_mark_sw_bad,
## num_compress_mark_hw_bad, num_compress_mark_sw_bad, num_compressed_inline,
## num_compressed_bg, num_partial_cg_modifies, num_partial_cg_new.
##
## @examples
## $compress->verify_compression_stats(fields => {'num_partial_write' => 0,
##                                                'num_cg_decompressed' => 0});
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub verify_compression_stats
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
			fields => { type => HASHREF, },
		} );
	
	my $volume = $args{volume};
	my $fields = $args{fields};
	
	my %current = $class->get_compression_stats();
	my $current_stats = $current{compression_stats};
	
	foreach my $field (keys %$fields)
	{
		if (!exists($current_stats->{$field}))
		{
			logresult('script', "verify_compression_stats: field \"$field\" is not available for compression stats.\n" .
				            "Available fields are:\n" . Dumper(keys(%$current_stats)));
		}
		
		my $expected = $fields->{$field};
		my $obtained = $current_stats->{$field};
		logcomment("Checking compression stats for volume $volume\n" .
			   "Field: $field\n" .
			   "Expected value: $expected\n" .
			   "Obtained value: $obtained");
		
		Compression::Utils::compare_values(obtained_value => $obtained,
						   expected_value => $expected,
						   allow_percentage_error => 0,
						   values_meaning => "compression stats ($field) for volume $volume",
						   non_fatal_failure => FALSE,
						   verbose => $class->{_debug});
	}
	
	return %current;
}


## @name ::decompress_revert
## @summary Reverts the containment version values.
##
## @description
## Runs decompress revert to the containment version values.
##
## @return Hash with rc, and output.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the Compress object was created.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @arg check_containment_version Optional: Flag to enable containment version check after revert.
## Legal values: TRUE/FALSE. Default value: $main::CHECK_CONTAINMENT_VERSION.
##
## @examples
## $compress->decompress_revert();
##
## @note
## Scripts performing this action should use required_features => [C_FEATURE_DECOMPRESS_REVERT]
## when calling setup, as this command is currently only available for BR.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub decompress_revert
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
			expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
			check_containment_version => { type => SCALAR, optional => 1, default => $main::CHECK_CONTAINMENT_VERSION },
		} );
	
	my $check_containment_version = delete($args{check_containment_version});
	
	my %return = $class->{_utilities}->run_decompress_revert(@_);
	
	if (($check_containment_version == TRUE) && ($args{expected} == C_RC_SUCCESS))
	{
		# Verify that containment version is reset
		$class->verify_containment_version(compression_state => C_STATE_NEVER_ENABLED);
	}
	
	return %return;
}


## @name:: start_vol_move_to_aggr
## @summary Starts vol move of the compress object to the specified aggregate.
##
## @description
## Starts vol move of the compress object to the specified aggregate.
## The destination aggregate is created automatically, if it does not already exist.
## After this action, the compress object will have a new member, vol_move,
## which can be used for any subsequent vol move actions.
##
## @arg dest_aggr Required: Name of the destination aggregate.
##
## @arg no_cutover Optional: Flag to disable cutover during the move. Default: FALSE.
##
## @arg retain_src Optional: Flag to retain the source volume after the move is complete. Defaulf: FALSE.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @return Hash with rc and output.
##
## @example
## $compress->start_vol_move_to_aggr(dest_aggr => C_DFLT_AGGREGATE_DEST);
## $compress->{vol_move}->status();
## $compress->{vol_move}->wait_for_completion();
##
sub start_vol_move_to_aggr
{
	my $class = shift;
	my (%args) = validate_with(params => \@_,
				   allow_extra => 1,
				   spec   =>
	{
		dest_aggr => {type => SCALAR,},
	});
	
	my $dest_aggr = delete($args{dest_aggr});
	
	if ((!defined($class->{vol_move})) ||
	    ($class->{vol_move}->{_dest_aggr} ne $dest_aggr))
	{
		$class->{vol_move} = Compression::VolMove->new(volume => $class->{volume}->_get_underlying_vol_obj(),
							       dest_aggr => $dest_aggr,
							       ui => $class->{_ui},);
	}
	
	return $class->{vol_move}->start(%args);
}


## @name:: refresh_container_aggr
## @summary Refreshes internal information in case the volume is now contained by a new aggregate.
##
## @description
## Refreshes internal information in case the volume is now contained by a new aggregate
## (ex: after vol move).
##
## @arg fail_if_not_changed Optional: Flag to enable failing in the case the container aggregate is still the same.
## Default: FALSE.
##
##@return Hash with rc.
##
## @example
## $compress->refresh_container_aggr(fail_if_not_changed => TRUE);
##
sub refresh_container_aggr
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			fail_if_not_changed => {type => SCALAR, default => FALSE,},
		});
	
	my $fail_if_not_changed = $args{fail_if_not_changed};
	my %status = $class->{volume}->get_status();
	my $container_aggr = $status{container_aggr};
	my $vol_name = $class->{vol_name};
	
	if (!defined($container_aggr))
	{
		logresult('script', "Failed to detect the container aggregate for volume $vol_name\n\n" . Dumper(%status));
	}
	
	if ($container_aggr eq $class->{aggr_name})
	{
		my $comment = "\nVolume $vol_name is contained by aggregate $container_aggr." .
		              "\nNo aggregate change detected.";
		if (TRUE == $fail_if_not_changed)
		{
			logresult('fatal', "$comment\n\n" . Dumper(%status));
		}
		logcomment($comment);
		return (rc => C_RC_SUCCESS,);
	}
	
	logcomment("\nContainer aggregate for volume $vol_name changed:\n" .
		   "Current aggregate: $container_aggr\n" .
		   "Old-known aggregate: " . $class->{aggr_name} . "\n" .
		   "Refreshing...");
	
	$class->{aggregate}->_refresh(new_aggr => $container_aggr,);
	
	if ($class->{aggregate}->{_hosting_filer_has_changed} == TRUE)
	{
		# Refresh the filer member (no need to go back to the aggregate -- GNOTE is doing it for us)
		$class->{filer} = Compression::Filer->new(filer => $class->{aggregate}->{_filer_name},
							  debug => $class->{_debug},
							  vserver => $class->{vserver},
							  boot_mode => $class->{boot_mode},
							  count_volumes => $class->{_count_volumes},
							  );
		$class->{host_name} = $class->{filer}->{host_name};
		$class->{_ui} = $class->{filer}->{_ui};
		my $conn = $class->{filer}->_get_underlying_filer_obj();
		$class->{decompress}->_update_conn($conn);
		$class->{compress_scanner}->_update_conn($conn);
		$class->{_utilities}->_update_conn($conn);
	}
	
	# Does not refresh UI yet (if new node has a different release)
	$class->{volume}->_refresh_aggregate(new_aggr => $class->{aggregate}->_get_underlying_aggr_obj(),); 
	
	# Note: The other members of the compress object seem to be ok without any update.
	# If this will change in the future (and any other member will need to execute any aggregate-level action),
	# we will have to address individual cases in here.
	
	return (rc => C_RC_SUCCESS,);
}


## @name:: start_vfiler_migrate
## @summary Starts vfiler migrate.
##
## @description
## Starts vfiler migrate.
## This operation has three logical steps, and
## should be performed in three stages. First, issue the "start" command which
## initiates the movement of the remote vfiler's storage to the local filer using
## SnapMirror. Then you can use the "status" command to monitor the status of this
## data movement. When this status for each path changes from "Being initialized"
## to "SnapMirrored", you can issue the "complete" command to finish the
## migration. When the "complete" command completes, the remote vfiler will have been
## moved to the local machine.
##
## @return Hash with rc and output.
##
## @arg dest_filer Required: Name of the destination filer.
##
## @arg interface Optional: Destination interface. Defaults to the vfiler's interface.
##
## @example
## $compress->start_vfiler_migrate(dest_filer => $MIGRATE_TO);
##
sub start_vfiler_migrate
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			dest_filer => {type => SCALAR,},
			interface => {type => SCALAR, optional => 1,},
		});
	
	if (!defined($class->{vfiler}))
	{
		logresult('script', "This method is only available for volumes created as resources to vfilers");
	}
	
	$class->{vfiler}->start_migrate(%args);
}


## @name:: detach_from_vfiler
## @summary 
## Detach from the "vfiler-resource" volume and attach to the "left-behind" volume after vfiler migrate is completed.
##
## @description
## Detach from the "vfiler-resource" volume and attach to the "left-behind" volume after vfiler migrate is completed.
##
## @example
## $compress->detach_from_vfiler();
##
## @note
## Some dataset actions will not be available on the files in this volume.
## 
sub detach_from_vfiler
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			# None
		});
	
	if (!defined($class->{vfiler}))
	{
		logresult('script', "This method is only available for volumes created as resources to vfilers");
	}
	
	$class->{volume}->_update_vfiler(new_vfiler => undef,);
	$class->{dataset}->_update_vfiler_mountpoint();
	$class->{decompress}->_update_vfiler();
	$class->{compress_scanner}->_update_vfiler();
	$class->{_utilities}->_update_vfiler();
}


## @name:: refresh_vfiler_controller
## @summary Attach to the "vfiler-resource" volume that was just migrated to a new controller.
##
## @description
## Attach to the "vfiler-resource" volume that was just migrated to a new controller.
##
## @example
## $compress->refresh_vfiler_controller();
## 
sub refresh_vfiler_controller
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			# None
		});
	
	if (!defined($class->{vfiler}))
	{
		logresult('script', "This method is only available for volumes created as resources to vfilers");
	}
	
	logcomment("Attaching the compress object to the new vfiler controller");
	$class->{vfiler}->_attach_to_migrated_vfiler();
	my $new_controller = $class->{vfiler}->get_controller_name();
	
	$class->{filer} = Compression::Filer->new(filer => $new_controller,
						  debug => $class->{_debug},
						  vserver => $class->{vserver},
						  boot_mode => $class->{boot_mode},
						  count_volumes => $class->{_count_volumes},
						  );
	$class->{host_name} = $class->{filer}->{host_name};
	$class->{_ui} = $class->{filer}->{_ui};
	
	my $conn = $class->{filer}->_get_underlying_filer_obj();
	my $new_aggr = $class->{vfiler}->get_root_aggr_name();
	$class->{aggregate} = Compression::Aggregate->new(conn => $conn,
							  ui => $class->{_ui},
							  aggregate => $new_aggr,
							  vserver => $class->{vserver},
							  boot_mode => $class->{boot_mode},
							  debug => $class->{_debug},
							  attach => TRUE,
							  disable_cleanup => $class->{_disable_cleanup},
							  );
	
	$class->{volume} = Compression::Volume->new(filer => $conn,
						    aggregate => $class->{aggregate}->_get_underlying_aggr_obj(),
						    volume => {name => $class->{vol_name},
							       auto_mount => FALSE,},
						    vserver => $class->{vserver},
						    boot_mode => $class->{boot_mode},
						    ui => $class->{_ui},
						    client => $class->{_client}->{name},
						    debug => $class->{_debug},
						    disable_cleanup => $class->{_disable_cleanup},
						    attach => $class->{vol_name},
						    vfiler => $class->{vfiler},
						    _added_to_vfiler => TRUE,
						    );
	# Do not mount the volume as part of object creation
	# (garbage collection for the old vol object would unmount it - they both have the same mountpoint)
	$class->{volume}->nfs_mount();
	
	# Note: supportability, file clone, file folding not updated 
	my $vfiler_name = $class->{vfiler}->{name};
	$class->{dataset}->_update_vol_obj($class->{volume}->_get_underlying_vol_obj());
	$class->{dataset}->_update_vfiler_mountpoint($class->{volume}->_get_mountpoint);
	$class->{decompress}->_update_conn($conn);
	$class->{decompress}->_update_vfiler($vfiler_name);
	$class->{compress_scanner}->_update_conn($conn);
	$class->{compress_scanner}->_update_vfiler($vfiler_name);
	$class->{_utilities}->_update_conn($conn);
	$class->{_utilities}->_update_vfiler($vfiler_name);
	$class->{sis}->_update_conn(new_filer => $conn);
}


## @name ::refresh_sw_version
## @summary Refresh sw version information for this object.
## 
## @description
## Refresh sw version information for this object.
## Useful when the files was upgraded/downgraded during testing.
## 
## @examples
## $compress->refresh_sw_version();
##
sub refresh_sw_version
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			# None
		});
	
	$class->{filer}->refresh_sw_version();
	$class->{dataset}->_set_debug_kernel($class->{filer}->{_has_debug_kernel});
	if ($class->{snmp})
	{
		$class->{snmp}->_set_ontap($class->{_ui}->{version}->{ontap});
	}
}


## @name ::configure_idd
## @summary Enables/disables IDD.
##
## @description
## Enables/disables IDD.
##
## @return Hash with rc, and output.
##
## @arg volume Optional: Name of volume for which to configure IDD.
## Defaults to the volume for which the Compress object was created.
##
## @arg enabled Optional: Flag to control if IDD should be enabled or disabled.
## Default: TRUE.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @examples
## $compress->configure_idd();
## $compress->configure_idd(enabled => FALSE);
##
sub configure_idd
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			volume => {type => SCALAR, optional => 1, default => $class->{vol_name},},
			enabled => {type => SCALAR, optional => 1, default => TRUE,},
			expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
		});
	
	return $class->{_utilities}->configure_IDD(@_);
}


## @name ::is_idd_enabled
## @summary Checks if IDD is enabled for the volume.
##
## @description
## Checks if IDD is enabled for the volume.
##
## @return Hash with rc, and output.
##
## @arg volume Optional: Name of volume for which to check IDD.
## Defaults to the volume for which the Compress object was created.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Legal values: C_RC_IDD_ON, C_RC_IDD_OFF. Default: C_RC_IDD_ON.
##
## @examples
## $compress->is_idd_enabled();
## $compress->is_idd_enabled(expected => C_RC_IDD_OFF);
##
sub is_idd_enabled
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			volume => {type => SCALAR, optional => 1, default => $class->{vol_name},},
			expected => {type => SCALAR, optional => 1, default => C_RC_IDD_ON,},
		});
	
	return $class->{_utilities}->is_IDD_enabled(@_);
}

## @name ::wait_for_commit
## @summary Checks if Autocommit scan on a volume has completed.
##
## @description
## Checks if Autocommit scan on a volume has completed.
##
## @return none.
##
## $compress->wait_for_commit();
##

sub wait_for_commit
{
	my $class = shift;
	my ($raw_output,$count);
	my (@output,@out);
	my $node = NACL::C::SystemNode->new( name => $class->{filer}->{_c_filer}->{host_name} );
	my $Api_Set_Obj = $node->get_7m_or_nodescope_apiset;
	my $previous = 0;
	while ($count < 15){
		$raw_output = $Api_Set_Obj->execute_command(command => "wafl scan status ".$class->{volume}->{vol_name});
		@output = split(/\n/, $raw_output);
		foreach (@output) {
			next if ($_ !~ /autocommit volume scan/);
			@out = split(/\s+/, $_);
			last ;
		}
		if ($out[5] > $previous){
			last;
		}
		$previous = $out[5];
		sleep 10;
		$count++;
	}
	
	if ($count == 20){
		logcomment("No commit even after 150 seconds");
	}
	return;
}



################################
# Internal subroutines
################################

# Disable any volume-level cleanup
# Needed for cases like aggr copy
sub _disable_vol_cleanup
{
	my $class = shift;
	
	$class->{volume}->{_disable_cleanup} = TRUE;
	$class->{decompress}->{_disable_cleanup} = TRUE;
	$class->{compress_scanner}->{_disable_cleanup} = TRUE;
}


# Any necessary clean-up upon object destruction
# Note: we only need this to ensure the right order for the member destructors
sub DESTROY
{
	my $class = shift @_;
	# WARNING: We need to store any eventual error message and fail it at the end of any destructor.
	# Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
	# At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
	# WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
	# logresult ('NA', "NOT APPLICABLE: .....");
	
	my $execution_error = $@;
	if (($class->{_disable_cleanup} == FALSE) &&
	    ($class->{_creator_pid} == $$))
	{
		my $vol_name = $class->{vol_name};
		logcomment("Executing Compress clean-up for volume $vol_name");
		logcomment("Set the overlap requests to default value(on)");
		$class->{filer}->overlap_requests(option => "on");
		# For some reason, it looks like the garbage collection for
		# all the class members does not happen in the right order
		# Ensure the order in here
		undef($class->{perf_stats});
		undef($class->{file_folding});
		undef($class->{sis});
		undef($class->{_utilities});
		undef($class->{compress_scanner});
		undef($class->{decompress});
		undef($class->{dataset});
		# This is for the case when the compression object is a clone (C-mode)
		if ((defined($class->{_split_job_id})) && ($class->{_split_job_id} != -1))
		{
			$class->clone_split_stop;
		}
		undef($class->{volume});
		undef($class->{vfiler});
		undef($class->{aggregate});
		# Make sure some disruptive flags are reset (even if they were not necessarily set during testing... )
		# Do this on best-effort basis.
		eval
		{
			my $flags_to_reset = {0 => [C_FLAG_WAFL_DISABLE_FILE_CLONE_BG_SHARING]};
			foreach my $val (keys %$flags_to_reset)
			{
				my $flags = $flags_to_reset->{$val};
				foreach my $flag (@$flags)
				{
					$class->{filer}->set_flag(code => $flag,
								  value => $val);
				}
			}
		};
		undef($class->{filer});
		
		logcomment( "Compress: cleanup for volume $vol_name: done" ) if ($class->{_debug});
	}
	else
	{
		logcomment( "Compress: cleanup:  nothing done: script PID: " . $$) if ($class->{_debug});
	}
	
	if (defined($execution_error))
	{
		_c_interpret_logresult($execution_error);
	}
}


# routine to test the module itself
sub testme
{
	package main;
	use Compression::Constants qw(:all);
	use Compression::Compress;
	use vars qw($FILER $AGGREGATE $VOLUME );

	require "drg/Compression/setup_core.pl";


	my $obj = Compression::Compress->setup(filer => $FILER,
					       aggregate => $AGGREGATE,
					       volume => $VOLUME,);
	my $vol_name = $obj->{vol_name};

	my $inexistent_vol_name = 'somevolthatshouldnotexist';
	my $inexistent_vol = NOTE::Volume->attach(Filer => $FILER,
						  name => $inexistent_vol_name,
						  force => 1,);
	if ($inexistent_vol->exists())
	{
		$inexistent_vol->purge();
	}

	my $expected_rcs = $obj->_get_testme_expected_results();

	$obj->{filer}->volume_exists( volume => $vol_name );
	$obj->{filer}->volume_exists( volume => $inexistent_vol_name, expected => C_RC_FAILURE );
	$obj->{aggregate}->volume_is_contained( volume => $vol_name );
	$obj->{aggregate}->volume_is_contained( volume => $inexistent_vol_name, expected => C_RC_FAILURE );
	$obj->enable( volume => $inexistent_vol_name, expected => C_RC_VOLUME_NOT_FOUND );

	$obj->{volume}->bring_offline();
	$obj->enable(expected => $expected_rcs->{enable_on_offlined_vol});
	$obj->disable(expected => $expected_rcs->{disable_on_offlined_vol});
	$obj->{volume}->bring_online();

	$obj->disable(all_volumes => TRUE);

	$obj->disable();
	$obj->is_enabled( expected => C_RC_FAILURE );
	$obj->toggle();
	$obj->is_enabled( expected => C_RC_SUCCESS );
	$obj->toggle();
	$obj->is_enabled( expected => C_RC_FAILURE );
	$obj->enable();
	$obj->is_enabled( expected => C_RC_SUCCESS );

	$obj->verify_savings( expected_compressed => 0 );
	$obj->wait_for_cp();
	$obj->{dataset}->create(files => [ { name => 'test_file', size => C_CG_SIZE }, ], );
	$obj->wait_for_cp();

	my $expected_compressed = (C_MAX_CG_BLKS-1) * 4;
	$obj->verify_savings( expected_compressed => $expected_compressed );
	$obj->verify_savings( expected_compressed => $expected_compressed - 1,
			      allow_percentage_error => 10);
	$obj->verify_savings( expected_deduplicated => 0 );
	$obj->get_savings( overwrite_args_string => 'vol1 vol2', expected => C_RC_INVALID_COMMAND_ARGUMENT );
	$obj->get_savings( overwrite_args_string => $inexistent_vol_name, expected => C_RC_VOLUME_NOT_FOUND );

	$obj->rename_volume();
	$obj->verify_savings( expected_compressed => $expected_compressed );
	$obj->is_enabled( expected => C_RC_SUCCESS );
	$obj->{decompress}->status(expected => C_RC_NO_DECOMPRESSION_IN_PROGR);
	$obj->{compress_scanner}->status(expected => C_RC_SCANNER_IDLE);
	
	my $new_vol_name = 'renamed_comp_vol';
	$obj->rename_volume(new_name => $new_vol_name);
	if (($obj->{vol_name} ne $new_vol_name) ||
	    ($obj->{decompress}->{vol_name} ne $new_vol_name) ||
	    ($obj->{_utilities}->{vol_name} ne $new_vol_name) ||
	    ($obj->{compress_scanner}->{vol_name} ne $new_vol_name))
	{
		logresult("fatal", "Failed to properly update the volume name to \"$new_vol_name\"");
	}

	logresult('PASS', "Compress: testme: Done" );
}

sub _get_testme_expected_results
{
	my $class = shift;
	
	my $ontap = $class->{_ui}->{version}->{ontap};
	
	if ($ontap eq '7.3')
	{
		return {enable_on_offlined_vol => C_RC_VOLUME_NOT_ONLINE,
			disable_on_offlined_vol => C_RC_VOLUME_NOT_ONLINE,
			enable_without_license => C_RC_NO_COMPRESSION_LICENSE,
			};
	}
	elsif ($ontap eq '8.0')
	{
		return {enable_on_offlined_vol => C_RC_VOLUME_NOT_ONLINE,
			disable_on_offlined_vol => C_RC_VOLUME_NOT_ONLINE,
			enable_without_license => C_RC_NO_COMPRESSION_LICENSE,
			};
	}
	else
	{
		return {enable_on_offlined_vol => C_RC_VOLUME_NOT_FOUND,
			disable_on_offlined_vol => C_RC_VOLUME_NOT_FOUND,
			enable_without_license => C_RC_NO_COMPRESSION_LICENSE,
			};
	}
}



1;
