#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/CompressionScanner.pm#1 $
#

## @summary CompressionScanner routines
##
## @description
## This module provides all the capabilities
## required for compression scanner work flow.
##
## @synopsis
## use Compression::CompressionScanner;
## use Compression::Constants qw(:all);
## use NOTE::Filer;
## use vars qw($FILER); 
## 
## BEGIN { require "Compression/base_import.pl"; }
## 
## my $obj = Compression::CompressionScanner->new(conn => $conn, volume => 'vol1');
##
## @examples
##
## my $obj = Compression::Compress->setup(filer => $FILER ,
##                                        aggregate => $AGGREGATE,
##                                        volume => {name => $VOLUME1,
##                                                   size => "8g",
##                                                   compress_enabled => FALSE,},
##                                        );
##
##$obj->{compress_scanner}->start(expected => C_RC_SUCCESS); 
##$obj->{compress_scanner}->stop();
##$obj->{compress_scanner}->status(expected => C_RC_SCANNER_IDLE );
##$obj->{compress_scanner}->start(deduped_blocks => TRUE, snapshotted_blocks => TRUE, delete_checkpoint =>TRUE); 
##$obj->{compress_scanner}->stop(no_checkpoint => TRUE);
##$obj->{compress_scanner}->wait_for_scanned_data( min_scanned => 357480, max_scanned => 1157480);
##$obj->{compress_scanner}->verify_checkpoint(); 

## @author dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::CompressionScanner;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;
use Tharn;
#use filer;
use Compression::Constants qw(:all);
use Compression::Version;
use Compression::Utils qw(_c_interpret_logresult _send_cmd_to_filer process_data_size_string);
use SayWrap qw(sayEOF);

BEGIN
{
        require "Compression/base_import.pl";
        require Exporter;
        our @ISA = qw(Exporter);
}

## @name ::new
## @summary Compression::CompressionScanner Constructor.
##
## @description
## Creates an object representing compression scanner utility.
##
## @return Compression::CompressionScanner object.
##
## @arg conn Required: NOTE::Filer object.
##
## @arg volume Optional: name of volume to be associated with this object.
## 
## @arg vfiler Optional: Name of vfiler.
##
## @arg ui Optional: Compression::UI object corresponding to the hosting filer.
## If not provided, it will be auto-generated.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
## 
## @arg supportability Optional: Compression::Supportability object,
## to be passed in if, along with the command completion, is it required
## to also check the supportability messages.
##
## @example
##
## my $obj = Compression::CompressionScanner->new( conn  => $filer, volume => 'my_vol');
##
## @author dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub new
{
        my $class = shift;
        my %args = validate(@_,
                {
                        conn => {type => OBJECT,
                                 },
                        
                        volume => {type => SCALAR,
                                   optional => 1,
                                   },
                        
                        ui => {type => OBJECT,
                               isa => "Compression::UI",
                               optional => 1,
                               },
                        
                        disable_cleanup => {type => BOOLEAN,
                                            optional => 1,
                                            default => FALSE,
                                            },
                        
                        debug => {type => SCALAR,
                                  optional => 1,
                                  default => FALSE,
                                  },
                        
                        vserver => {type => SCALAR|UNDEF,
                                    optional => 1},
                        
                        boot_mode => {type => SCALAR,
                                      optional => 1,
                                      default => C_ONTAP_BOOT_MODE},
                        
                        vfiler => {type => SCALAR|UNDEF,
                                   optional => 1,},
                        
                        supportability => {type => UNDEF|OBJECT,
                                           optional => 1,},
                } );
        
        my $self = {};
        $self->{vol_name} = $args{volume} if (defined $args{volume});
        $self->{_conn} = $args{conn};
       
        $self->{_debug} = $args{debug};
        $self->{_disable_cleanup} = $args{disable_cleanup};
        $self->{_creator_pid} = $$;
        $self->{_boot_mode} = $args{boot_mode};
        $self->{vserver} = $args{vserver};
        $self->{_vfiler} = $args{vfiler};
        $self->{_supportability} = $args{supportability};
        
        if ($self->{_boot_mode} eq C_ONTAP_7MODE)
        {
                if  (!($self->{_conn}->isa("NOTE::Filer")))
                {
                        logresult('script', "For 7-mode, the conn argument is required to be a NOTE\:\:Filer object");
                }
                
                $self->{host_name} = $self->{_conn}->host();
        }
        elsif ($self->{_boot_mode} eq C_ONTAP_CMODE)
        {
                if (!($self->{_conn}->isa("NOTE::GX::Node")))
                {
                        logresult('script', "For C-mode, the conn argument is required to be a NOTE\:\:GX\:\:Node object");
                }
                
                $self->{host_name} = $self->{_conn}->name();
        }
        
        if ((defined($self->{_supportability})) && (!($self->{_supportability}->isa("Compression::Supportability"))))
        {
                logresult('script', "The supportability argument, if passed in, must be a Compression\:\:Supportability object");
        }
        
        logcomment("CompressionScanner: new: PID of the script creating this object: " . $self->{_creator_pid}) if ($self->{_debug});
        
        ####################################################
        # Dependency on other internal modules
        ####################################################
        
        # UI
        if ( defined($args{ui}) )
        {
                $self->{_ui} = $args{ui};
        }
        else
        {
                # Auto-detect version of the hosting filer. Need this for UI.
                logcomment("CompressionScanner: UI object not provided. Auto-generating...") if ($self->{_debug});
                my $version = Compression::Version->new( conn => $args{conn}, debug => $self->{_debug} );
                $self->{_ui} = Compression::UI->new( version => $version, debug => $args{debug}, boot_mode => $self->{_boot_mode} );
        }
        
        bless $self, $class;
        return $self;
}


## @name ::start
## @summary Starts compression scanner on a specific volume.
## 
## @description
## Starts compression scanner on a specific volume.
##
## @return Hash with rc and output
## 
## @arg volume Optional: Name of the volume.
## If not provided, uses volume associated with the object, if available.
## If not provided and the object does not have an associated volume, logs script error.
##
## @arg deduped_blocks Optional: Boolean flag to start the scanner with the option to compress deduplicated blocks.
## Default: FALSE.
##
## @arg snapshotted_blocks Optional: Boolean flag to start the scanner with the option to
## compress blocks in active filesystem that are shared with snapshots.
## Default: FALSE.
##
## @arg delete_checkpoint Optional: Boolean flag to start the scanner with the option to delete checkpoint, if it exists.
## Default: FALSE.
##
## @arg timeout Optional: Timeout, in seconds, for the command to complete.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
## 
## @arg scan Optional: Boolean flag for starting the scanner with the option to scan already existent data
## (for RR: -s option).
## 
## @arg scan_to_compress Optional: Boolean flag for starting the scanner with the option to compress already existent data
## (for RR: -C option).
## 
## @arg scan_to_dedupe Optional: Boolean flag for starting the scanner with the option to dedupe already existent data
## (for RR: -D option).
##
## @examples
##
## $obj->start();
## $obj->start(deduped_blocks => TRUE, snapshotted_blocks=> TRUE, delete_checkpoint =>TRUE);
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @note
## If the SUPPORTABILITY cmd-line argument is 1, then,
## along with this action, the correspondent supportability message will be also checked.
## 
## @author dl-compression-qa@netapp.com
## @status private (in-progress)

sub start
{
        my $class = shift;
        
        my (%args) = validate( @_,
                {
                        volume => {type => SCALAR,
                                   optional => 1,
                                   default  => $class->{vol_name}, },
                        
                        vserver => {type => SCALAR,
                                    optional => 1,
                                    default  => $class->{vserver}, },
                        
                        timeout => {type => SCALAR,
                                    regex => qr/^\d+$/,
                                    optional => 1, },
                        
                        expected => {type => SCALAR,
                                     optional => 1,
                                     regex => qr/^\d+$/,
                                     default  => C_RC_SUCCESS, },
                        
                        scan => {type => BOOLEAN,
                                 optional => 1,
                                 default => FALSE, },

                        scan_to_compress => {type => BOOLEAN,
                                             optional => 1, },

                        scan_to_dedupe => {type => BOOLEAN,
                                           optional => 1, },
                        
                        deduped_blocks => {type => BOOLEAN,
                                           optional => 1, },
                        
                        snapshotted_blocks => {type => BOOLEAN,
                                               optional => 1, },
                        
                        delete_checkpoint => {type => BOOLEAN,
                                              optional => 1,
                                              default  => FALSE, },
                } );
        
        my $vol_name = $args{volume};
        my $expected = $args{expected};
        if (!defined($vol_name))
        {
                logresult( "script", "Script attempted to start the compression scanner without an associated volume name" );
        }
        
        my $rc = C_RC_FAILURE;
        my $cmd_code = C_CMD_COMPRESS_START;
        my $need_persistent_connection = 0;
        
        # TODO: For BR, for -a, -s, we need persistent connections to handle the 'Are you sure you want to proceed?' string
        # We should move this knowledge inside UI (more than one command string; need for persistent connections)
        # For now, do it this way...
        my $ontap = $class->{_ui}->{version}->{ontap};
        if (("$ontap" eq '8.0') &&
            ($expected == C_RC_SUCCESS) &&
            (((exists($args{deduped_blocks})) && ($args{deduped_blocks} == TRUE)) ||
             ((exists($args{snapshotted_blocks})) && ($args{snapshotted_blocks} == TRUE))))
        {
                $need_persistent_connection = 1;
        }
        
        if ((defined($class->{_supportability})) && ($expected == C_RC_SUCCESS))
        {
                $class->{_supportability}->start_tailing(events => [{code => C_EVENT_COMPRESS_STARTED,
                                                                     options => {deduped_blocks => $args{deduped_blocks},
                                                                                 snapshotted_blocks => $args{snapshotted_blocks},
                                                                                 delete_checkpoint => $args{delete_checkpoint},
                                                                                 scan => $args{scan},
                                                                                 scan_to_compress => $args{scan_to_compress},
                                                                                 scan_to_dedupe => $args{scan_to_dedupe},},},
                                                                    ],
                                                         required_identifier => $vol_name,
                                                         );
        }
        
        my $command = $class->{_ui}->formulate_command(code => $cmd_code,
                                                       volume => $vol_name,
                                                       vserver => $args{vserver},
                                                       options => {deduped_blocks => $args{deduped_blocks},
                                                                   snapshotted_blocks => $args{snapshotted_blocks},
                                                                   delete_checkpoint => $args{delete_checkpoint},
                                                                   scan => $args{scan},
                                                                   force => ($args{scan}) || ($args{delete_checkpoint}), # so we don't need persistent connections
                                                                   scan_to_compress => $args{scan_to_compress},
                                                                   scan_to_dedupe => $args{scan_to_dedupe},
                                                                   });
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            timeout => $args{timeout},
                                                            need_persistent_connection => $need_persistent_connection,
                                                            # TODO: these strings should be in UI (they are only used for BR)
                                                            waitstring => "Are you sure you want to proceed",
                                                            replaystring => "y",
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
                                                            zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
        
        my %res = $class->{_ui}->process_output_string(code => $cmd_code,
                                                       output => $output,
                                                       expected_rc => $expected,
                                                       log_message => "Start compression scanner for volume $vol_name",);
        
        if ((defined($class->{_supportability})) && ($expected == C_RC_SUCCESS))
        {
                $class->{_supportability}->stop_and_match();
        }
        
        return %res;
}


## @name ::status
## @summary Reports the status of volume compression.
## 
## @description
## Reports the status of volume compression.
##
## @return Hash with rc, output, and compress_statistics. 
## 
## @arg volume Optional: Name of the volume.
## If not provided, uses volume associated to the object, if available.
## If not provided and the object does not have an associated volume, reports compress status for all volumes.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes there should be compression in progress.
##
## @arg expected_stage Optional: Expected stage of the scanner (meaningful for RR).
## If not defined, the check is only made that the scanner is active.
## Legal values: C_STAGE_COMPRESSING, C_STAGE_SAVING, C_STAGE_GATHERING, C_STAGE_DONE
## (defined in Constants.pm). Default value: none.
##
## @examples
## $obj->status(expected => C_RC_SCANNER_IDLE);
## $obj->status();
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
## 
## @author dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub status
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        volume => {type => SCALAR,
                                   optional => 1,
                                   default => $class->{vol_name},},
                        
                        vserver => {type => SCALAR|UNDEF,
                                    optional => 1,
                                    default  => $class->{vserver}, },
                        
                        expected => {type => SCALAR,
                                     optional => 1,
                                     default => C_RC_SUCCESS,},
                        
                        expected_stage => {type => SCALAR|UNDEF,
                                           optional => 1,}, 
		} ); 
        
        my $vol_name = $args{volume};
        my $expected = $args{expected};
        my $expected_stage = $args{expected_stage};
        my $cmd_code = C_CMD_COMPRESS_STATUS;
        my $zapi = $class->{_ui}->is_zapi_available(code => $cmd_code);
        
        if ((defined($expected_stage)) && ($main::UI_TYPE eq C_UI_ZAPI) && ($zapi == TRUE))
        {
                logresult('NA', "NOT APPLICABLE: Scanner stage is not available through ZAPI");
        }
        
        my $command = $class->{_ui}->formulate_command(code => $cmd_code,
                                                       volume => $vol_name,
                                                       vserver => $args{vserver});
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            timeout => $args{timeout},
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
                                                            zapi_available => $zapi,);
        
        my %status = $class->{_ui}->process_output_string(code => $cmd_code,
                                                          output => $output,
                                                          required_return_hash_key => $vol_name,
                                                          expected_rc => C_RC_FAILURE_NON_FATAL,
                                                          log_message => "Obtain compression scanner status",);
        
        # For pre-RR, once we obtained a legal status output, we automatically set the status/stage to active/compressing
        # (even if the command output does not include them).
        # Otherwise, we also need to check the status/stage (but only for a specific volume)
        # If the volume is not specified (either as argument, or at object creation), we declare success only for getting a legal output
        if (($status{rc} == C_RC_SUCCESS) && (defined($vol_name)))
        {
                # process_output_string already ensured that we have the status for our volume
                my $vol_status = $status{compress_statistics}{$vol_name};
                if (!defined($expected_stage))
                {
                        # Only look for an active state
                        my $active_status = $class->{_ui}->get_constant_value(constant_code => 'C_STATUS_ACTIVE');
                        my $initializing_status = $class->{_ui}->get_constant_value(constant_code => 'C_STATUS_INITIALIZING');
                        my $obtained_status = $vol_status->{status};
                        
                        if (($obtained_status ne $active_status) && ($obtained_status ne $initializing_status))
                        {
                                logcomment("Compression scanner is not active");
                                $status{rc} = C_RC_SCANNER_IDLE;
                        }
                }
                else
                {
                        my $obtained_stage = $vol_status->{stage};
                        my $expected_stage_constant = $class->{_ui}->get_constant_value(constant_code => $expected_stage);
                        if ($obtained_stage ne $expected_stage_constant)
                        {
                                logcomment("Unexpected compression scanner stage:\n" .
                                           "Obtained: $obtained_stage\n" .
                                           "Expected: $expected_stage_constant");
                                $status{rc} = C_RC_UNEXPECTED_SCANNER_STAGE;
                        }
                }
        }
        
        $class->{_ui}->check_for_expected_rc(obtained_rc => $status{rc},
                                             expected_rc => $expected,
                                             log_message => "Obtain compression scanner status",
                                             output      => $output);
        
        return %status;      
}


## @name ::stop
## @summary Stops compression scanner for a specific volume.
## 
## @description
## Stops compression scanner for a specific volume (checkpoint is created by default).
##
## @return Hash with rc and output.
## 
## @arg volume Optional: Name of the volume.
## If not provided, uses volume associated to the object, if available.
## If not provided and the object does not have an associated volume, logs script error.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Defaults to C_RC_SUCCESS.
##
## @arg no_checkpoint Optional: Boolean flag to stop the scanner with the option to
## not create any checkpoint and to delete any existing checkpoint.
## Defaults to FALSE. Not available for RR.
##
## @examples
## $obj->stop();
## $obj->stop(no_checkpoint => TRUE);
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @note
## If the SUPPORTABILITY cmd-line argument is 1, then,
## along with this action, the correspondent supportability message will be also checked.
## 
## @author dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub stop
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        volume => {type => SCALAR,
                                   optional => 1,
                                   default => $class->{vol_name},},
                        
                        vserver => {type => SCALAR,
                                    optional => 1,
                                    default  => $class->{vserver}, },
                        
                        expected => {type => SCALAR,
                                     optional => 1,
                                     default => C_RC_SUCCESS,},
                        
                        no_checkpoint => {type => BOOLEAN,
                                          optional => 1,
                                          default  => FALSE},
                } );
        
        my $vol_name = $args{volume};
        if ( !defined($vol_name) )
        {
                logresult( "script", "Script attempted to stop the compression scanner without an associated volume name");
        }
        
        my $expected = $args{expected};
        my $cmd_code = C_CMD_COMPRESS_STOP;
        
        if ((defined($class->{_supportability})) && ($expected == C_RC_SUCCESS))
        {
                $class->{_supportability}->start_tailing(events => [{code => C_EVENT_COMPRESS_ABORTED,
                                                                     options => {no_checkpoint => $args{no_checkpoint},},},
                                                                    ],
                                                         required_identifier => $vol_name,
                                                         );
        }
        
        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $vol_name,
                                                       vserver => $args{vserver},
                                                       options => {no_checkpoint => $args{no_checkpoint},});
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
                                                            zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
        logcomment("CompressionScanner: stop:\n" .
                   "command: $command\n" .
                   "output: $output") if ($class->{_debug});
        
        my %res = $class->{_ui}->process_output_string(code => $cmd_code,
                                                       output => $output,
                                                       expected_rc => $expected,
                                                       log_message => "Stop compression scanner for volume $vol_name",);
        
        if ($expected == C_RC_SUCCESS)
        {
                # Wait for scanner to become idle. Post-BR it takes some time...
                logcomment("Compression scanner has been stopped. Wait for it to become idle...");
                $class->wait_for_scanned_data(timeout => C_DFLT_MAX_SCANNER_ABORT_TIME,
                                              _supportability => FALSE);
                
                if (defined($class->{_supportability}))
                {
                        $class->{_supportability}->stop_and_match();
                }
        }
        
        return %res;
}


## @name ::wait_for_scanned_data
##
## @summary Checks for the given amount of scanned data within a mix and a max boundary.
##
## @description
## Checks for the given amount of scanned data within a mix and a max boundary.
##
## @return Hash with rc and output 
## 
## @arg volume Optional: Name of the volume.
## If not provided, uses volume associated to the object, if available.
## If not provided and the object does not have an associated volume, logs script error.
##
## @arg min_scanned Optional: Lower limit for data scanned. It can be a value (in bytes)
## or 'all' (meaning that this method will wait until the compress scanner completes).
## Default: 'all'.
##
## @arg max_scanned Optional: Upper limit for data scanned. It can be a value (in bytes).
## Default: 0.
##
## @arg check_interval Optional: Number of seconds between two consecutive status checks.
## Defaults to C_DFLT_COMPRESS_INTERVAL (defined in Compression::Constants).
##
## @arg timeout Optional: Number of seconds to wait for required percentage.
## Defaults to C_DFLT_COMPRESS_TIMEOUT (defined in Compression::Constants).
##
## @arg stage Optional: Stage that the scanner is required to be in.
## NOTE: if min_scanned is 'all' (which is the default behavior), then
## this method will wait until the required stage is completed.
##
## @examples
##
## $obj->wait_for_scanned_data(min_scanned => 11210, max_scanned => 357480);
## $obj->wait_for_scanned_data();
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @note
## If the SUPPORTABILITY cmd-line argument is 1, then,
## when waiting for all the data to be scanned, the correspondent supportability message will be also checked.
## 
## @author dl-compression-qa@netapp.com
## @status private (in-progress)

sub wait_for_scanned_data
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        volume => {type => SCALAR,
                                   optional => 1,
                                   default => $class->{vol_name},},
                        
                        vserver => {type => SCALAR,
                                    optional => 1,
                                    default  => $class->{vserver}, },
                        
                        min_scanned => {type => SCALAR,
                                        regex => qr/^\d+$|^all$/,
                                        default => 'all',},
                        
                        max_scanned => {type => SCALAR,
                                        optional => 1,
                                        regex => qr/^\d+$/,
                                        default => 0},
                        
                        timeout => {type => SCALAR,
                                    optional => 1,
                                    regex => qr/^\d+$/,
                                    default => C_DFLT_COMPRESS_TIMEOUT}, 
                        
                        check_interval => {type => SCALAR,
                                           optional => 1,
                                           regex => qr/^\d+$/,},
                        
                        stage => {type => SCALAR,
                                  optional => 1,
                                  },
                        
                        _supportability => {type => SCALAR,
                                            optional => 1,
                                            default => TRUE,
                                            },
                } );
        
        my $vol_name = $args{volume};
        if ( !defined($vol_name) )
        {
                logresult( "script", "Compression scanner: wait_for_scanned_data called without an associated volume name");
        }
        
        my $expected_stage = $args{stage};
        my $min = $args{min_scanned};
        my $max = $args{max_scanned};
        # Don't allow 0 for check interval (will be used for sleep) 
	my $check_interval = ($args{check_interval}) ? $args{check_interval} : C_DFLT_COMPRESS_INTERVAL;
	my $timeout= $args{timeout};
        my $_supportability = $args{_supportability};
	my $elapsed_time = 0;
	my %status_hash = ();
	my $rc = C_RC_FAILURE;
        my $seen_expected_stage = (defined($expected_stage)) ? FALSE : TRUE;
	
	if (("$min" ne 'all') && ($min > $max))
        {
		logresult( "script", "min_scanned($min) > max_scanned($max)" );
	}
        
        my $comment = "Waiting for data processed by the compress scanner:\n" .
                      "Min scanned: $min bytes\n" .
                      "Max scanned: $max bytes\n" .
                      "Timeout: $timeout\n" .
                      "Check interval: $check_interval\n";
        if (defined($expected_stage))
        {
                $comment .= "Waiting for stage: $expected_stage";
        }
        logcomment($comment);
        
        if (($_supportability == TRUE) && (defined($class->{_supportability})) && ("$min" eq 'all'))
        {
                $class->{_supportability}->start_tailing(events => [{code => C_EVENT_COMPRESS_COMPLETED,},],
                                                         required_identifier => $vol_name,
                                                         );
        }
	
	while ($elapsed_time <= $timeout)
        {
                logcomment("CompressionScanner :: looping: elapsed time: $elapsed_time\n") if ($class->{_debug});
		%status_hash = $class->status(volume => $vol_name,
                                              vserver => $args{vserver},
                                              expected => C_RC_FAILURE_NON_FATAL,
                                              expected_stage => $expected_stage);
		logcomment("CompressionScanner: elapsed_time=$elapsed_time (timeout=$timeout):\nstatus returned:\n" . Dumper(%status_hash) . "\n") if ($class->{_debug});
                
		if ($status_hash{rc} != C_RC_SUCCESS)
		{
			$rc = $status_hash{rc};
                        if (($rc == C_RC_SCANNER_IDLE) || ($rc == C_RC_UNEXPECTED_SCANNER_STAGE))
                        {
                                if ($seen_expected_stage == TRUE)
                                {
                                        if ("$min" eq 'all')
                                        {
                                                $rc = C_RC_SUCCESS;
                                        }
                                        else
                                        {
                                                $rc = C_RC_PERCENTAGE_MISSED;
                                                last;
                                        }
                                }
                                else
                                {
                                        logcomment("CompressionScanner :: did not see expected stage yet. " .
                                                   "Sleeping $check_interval second(s)...\n") if ($class->{_debug});
                                        sleep($check_interval);
                                        $elapsed_time += $check_interval;
                                        next;
                                }
                        }
                        last;
		}
                
                $seen_expected_stage = TRUE;
                
                if ("$min" ne 'all')
                {
                        my $current_scanned = $status_hash{compress_statistics}{$vol_name}{data_scanned};
                        if (!defined($current_scanned))
                        {
                                # Do not fail here. Some stages do not have "data processed"
                                # We may want to see "any stage", but some amount of data processed.
                                logcomment("CompressionScanner :: could not detect data_scanned value. Setting it to 1") if ($class->{_debug});
                                $current_scanned = 1;
                        }
                        logcomment("CompressionScanner :: currently scanned: $current_scanned KB. Time $elapsed_time\n") if ($class->{_debug});
                        
                        my $current_scanned_bytes = $current_scanned * KB;
                        if (($current_scanned_bytes >= $min) && ($current_scanned_bytes <= $max))
                        {
                                logcomment("CompressionScanner :: wait_for_scanned_data SUCCESS: scanned: $current_scanned_bytes") if ($class->{_debug});
                                $rc = C_RC_SUCCESS;
                                last;
                        }
		
                        if ($current_scanned_bytes > $max)
                        {
                                $rc = C_RC_PERCENTAGE_MISSED;
                                last;
                        }    
                }
		
		logcomment("CompressionScanner :: sleeping $check_interval second(s)...\n") if ($class->{_debug});
		sleep($check_interval);
		$elapsed_time += $check_interval;
	}
	
        my %check_rc = $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                            expected_rc => C_RC_SUCCESS,
                                                            log_message => "Wait for data processed by the compression scanner to be in between $min and $max (bytes) - elapsed_time=$elapsed_time(timeout=$timeout)",
                                                            output      => $status_hash{output});
        
        if (($_supportability == TRUE) && (defined($class->{_supportability})) && ("$min" eq 'all'))
        {
                $class->{_supportability}->stop_and_match();
        }
        
        $status_hash{rc} = $check_rc{rc};
        return %status_hash;
}


## @name ::wait_until_active
##
## @summary Waits until the scanner becomes active.
##
## @description
## Waits until the scanner becomes active.
##
## @return Hash with rc and output 
## 
## @arg volume Optional: Name of the volume.
## If not provided, uses volume associated to the object, if available.
## If not provided and the object does not have an associated volume, logs script error.
##
## @arg check_interval Optional: Number of seconds between two consecutive status checks.
## Defaults to C_DFLT_COMPRESS_INTERVAL (defined in Compression::Constants).
##
## @arg timeout Optional: Number of seconds to wait.
## Defaults to C_DFLT_COMPRESS_TIMEOUT (defined in Compression::Constants).
##
## @examples
## $obj->wait_until_active();
## $obj->wait_until_active(timeout => 600);
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.

sub wait_until_active
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        volume => {type => SCALAR,
                                   optional => 1,
                                   default => $class->{vol_name},},
                        
                        vserver => {type => SCALAR,
                                    optional => 1,
                                    default  => $class->{vserver}, },
                        
                        timeout => {type => SCALAR,
                                    optional => 1,
                                    regex => qr/^\d+$/,
                                    default => C_DFLT_COMPRESS_TIMEOUT},
                        
                        check_interval => {type => SCALAR,
                                           optional => 1,
                                           regex => qr/^\d+$/,},
                } );
        
        my $vol_name = $args{volume};
        if ( !defined($vol_name) )
        {
                logresult( "script", "Compression scanner: wait_until_active called without an associated volume name");
        }
        
        # Don't allow 0 for check interval (will be used for sleep)
        my $check_interval = ($args{check_interval}) ? $args{check_interval} : C_DFLT_COMPRESS_INTERVAL;
        my $timeout= $args{timeout};
        my $elapsed_time = 0;
        my %status_hash = ();
        my $rc = C_RC_SCANNER_IDLE;
        
        logcomment("Waiting for the compress scanner on volume $vol_name to become active:\n" .
                   "Timeout: $timeout\n" .
                   "Check interval: $check_interval\n");
        
        while ($rc == C_RC_SCANNER_IDLE)
        {
                if ($elapsed_time > $timeout)
                {
                        logresult('fatal', "Timed-out while waiting for the compression scanner to become active:\n" .
                                           "Latest obtained status:\n" . Dumper(%status_hash));
                }
                
                sleep($check_interval);
                $elapsed_time += $check_interval;
                
                %status_hash = $class->status(volume => $vol_name,
                                              vserver => $args{vserver},
                                              expected => C_RC_FAILURE_NON_FATAL);
                $rc = $status_hash{rc};
                logcomment("CompressionScanner\:\:wait_until_active: elapsed_time=$elapsed_time (timeout=$timeout):\n" .
                           "status returned:\n" . Dumper(%status_hash) . "\n") if ($class->{_debug});
        }
        
        # Now verify that the scanner is active (and no other error occured)
        $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                             expected_rc => C_RC_SUCCESS,
                                             log_message => "Wait for compression scanner to become active",
                                             output      => $status_hash{output},);
        
        return %status_hash;
}


## @name ::get_checkpoint
## @summary Gets the checkpoint info.
## 
## @description
## Gets the checkpoint info. It can be used for basic checks for the checkpoint.
## For more complicated ones, use verify_checkpoint (for example, to check for specific stages,
## or for a certain amount of data scanned).
##
## @return Hash with rc, output, checkpoint_creation_time, total_bytes_processed,
## and checkpoint_info (if available).
## 
## @arg volume Optional: Name of the volume.
## If not provided, uses volume associated with the object, if available.
## If not provided and the object does not have an associated volume, logs script error.
##
## @arg timeout Optional: Timeout, in seconds, for the command to complete.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## By default, it is not defined (no expected result check is done).
##
## @arg allow_scanner_idle Optional: Flag to enable "best effort" attempt to
## check for the checkpoint file existence even if the scanner is idle.
## If this flag is not used, in the case when the scanner is idle,
## no extra check on the checkpoint file is performed.
## Default value: FALSE.
##
## @examples
## $obj->get_checkpoint(expected => C_RC_CHECKPOINT_FILE_EMPTY);
##
## @author dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_checkpoint
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        volume => {type => SCALAR,
                                   optional => 1,
                                   default => $class->{vol_name},},
                        
                        vserver => {type => SCALAR|UNDEF,
                                    optional => 1,
                                    default  => $class->{vserver}, },
                        
                        timeout => {type => SCALAR|UNDEF,
                                    optional => 1,},
                        
                        expected => {type => SCALAR,
                                     optional => 1,},
                        
                        allow_scanner_idle => {type => BOOLEAN,
                                               optional => 1,
                                               default => FALSE,},
		} );
        
        my $expected = $args{expected};
        my $vol_name = $args{volume};
        my $allow_scanner_idle = $args{allow_scanner_idle};
        if ( !defined($vol_name))
        {
                logresult( "script", "Volume name was neither passed during initialization nor to stop api" );
        }
        
        my $cmd_code = C_CMD_COMPRESS_STATUS_CHECKPOINT;
        my $rc = C_RC_SUCCESS;
        # Information that we need to extract to return to the caller (gathered from different hashes, depending on releases)
        my $checkpoint_creation_time = '';
        my $kb_data_processed = 0;
 
        my $command = $class->{_ui}->formulate_command(code => $cmd_code,
                                                       volume => $vol_name,
                                                       vserver => $args{vserver},);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            timeout => $args{timeout},
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
        
        my %return =  $class->{_ui}->process_output_string(code => $cmd_code,
                                                           output => $output,
                                                           required_return_hash_key => $vol_name,
                                                           expected_rc => C_RC_FAILURE_NON_FATAL,
                                                           log_message => "Obtain checkpoint info for volume $vol_name",);
        
        logcomment("CompressionScanner: get_checkpoint: process_output_string returned:\n" . Dumper(%return)) if ($class->{_debug});
        
        $rc = $return{rc};
        # For RR we will get success even if the checkpoint file is empty, or if the scanner is idle
        # We need to double-check
        if ($rc == C_RC_SUCCESS)
        {
                my $scanner_status = $return{checkpoint_info}{$vol_name}{status};
                my $idle_status = $class->{_ui}->{constants}->get_constant_value(constant_code => C_STATUS_IDLE);
                $checkpoint_creation_time =  $return{checkpoint_info}{$vol_name}{checkpoint_creation_time};
                if ((defined($scanner_status)) && ("$scanner_status" eq "$idle_status"))
                {
                        $rc = C_RC_SCANNER_IDLE;
                        logcomment("Compression scanner is idle");
                }
                else
                {
                        $rc = $class->{_ui}->parse_string_for_rc(code => _C_CMD_PARSE_CHECKPOINT_TIME,
                                                                 string => $checkpoint_creation_time);
                        logcomment("After checking the checkpoint info string, rc = $rc") if ($class->{_debug});
                }
        }
        
        # Extract "total kb processed"
        # Try first pulling it out of the compress_statistics hash (BR1: if there is no checkpoint)
        if (exists($return{compress_statistics}{$vol_name}{data_scanned}))
        {
                $kb_data_processed = $return{compress_statistics}{$vol_name}{data_scanned};
        }
        # Look in the checkpoint_info hash: the most relevant info
        if (defined($return{checkpoint_info}{$vol_name}{total_data_scanned}))
        {
                $kb_data_processed = $return{checkpoint_info}{$vol_name}{total_data_scanned};
        }
        
        $return{scanner_stage} = $class->{_ui}->{constants}->get_constant_value(constant_code => C_STAGE_NONE);
        if (defined($return{checkpoint_info}{$vol_name}{stage}))
        {
                $return{scanner_stage} = $return{checkpoint_info}{$vol_name}{stage};
        }
        elsif ($rc != C_RC_SCANNER_IDLE)
        {
                # For BR: the only stage is "compressing"
                $return{scanner_stage} = $class->{_ui}->{constants}->get_constant_value(constant_code => C_STAGE_COMPRESSING);
        }
        
        $return{checkpoint_stage} = $class->{_ui}->{constants}->get_constant_value(constant_code => C_CHECKPOINT_STAGE_NONE);
        if (defined($return{checkpoint_info}{$vol_name}{checkpoint_stage}))
        {
                $return{checkpoint_stage} = $return{checkpoint_info}{$vol_name}{checkpoint_stage};
        }
        
        if (($allow_scanner_idle == TRUE) && ($rc == C_RC_SCANNER_IDLE))
        {
                # This will cover BR (we do not get checkpoint info if the scanner is idle)
                $rc = C_RC_CHECKPOINT_FILE_EMPTY;
                
                # For RR and after, we need to check
                if ((defined($checkpoint_creation_time)) && ($checkpoint_creation_time ne ''))
                {
                        $rc = $class->{_ui}->parse_string_for_rc(code => _C_CMD_PARSE_CHECKPOINT_TIME,
                                                                 string => $checkpoint_creation_time);
                        logcomment("allow_scanner_idle true, after checking the checkpoint info string, rc = $rc") if ($class->{_debug});
                }
        }
        
        if (defined($expected))
        {
                $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                     expected_rc => $expected,
                                                     log_message => "Obtain checkpoint info for volume $vol_name",
                                                     output      => $output);
        }
        
        $return{rc} = $rc;
        $return{checkpoint_creation_time} = $checkpoint_creation_time;
        $return{total_bytes_processed} = $kb_data_processed * KB;
        logcomment("CompressionScanner: get_checkpoint returning:\n" . Dumper(%return)) if ($class->{_debug});
        
        return %return;
}


# TODO: Hourly checkpoint not in consideration yet
##
## @name ::verify_checkpoint
## @summary Checks for existence of checkpoint. 
## 
## @description
## Checks for existence of checkpoint. 
##
## @return Hash with rc, output, and checkpoint_info
## 
## @arg volume Optional: Name of the volume.
## If not provided, uses volume associated with the object, if available.
## If not provided and the object does not have an associated volume, logs script error.
##
## @arg timeout Optional: Timeout, in seconds, for the command to complete.
## NOTE: if wait_for_requirements is TRUE, then the timeout is the total time
## to wait for the requirements to be met.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Default: C_RC_SUCCESS.
## 
## @arg expected_stage Optional: Stage in which the scanner is expected to be (defined in Compression::Constants).
## Default: C_STAGE_COMPRESSING.
## 
## @arg expected_checkpoint_stage Optional: Expected checkpoint stage (defined in Compression::Constants).
## Default: none (no checkpoint stage check is executed).
## 
## @arg min_bytes_processed Optional: Min no. of bytes that the scanner is expected to have processed.
## 
## @arg max_bytes_processed Optional: Max no. of bytes that the scanner is expected to have processed.
##
## @arg wait_for_requirements Optional: Boolean flag to enable looping until the requirements are met before
## verifying the checkpoint (expected_stage, expected_checkpoint_stage, min_bytes_processed, max_bytes_processed).
## Default: FALSE.
##
## @examples
## $obj->verify_checkpoint();
##
## @author dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub verify_checkpoint
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        volume => {type => SCALAR,
                                   optional => 1,
                                   default => $class->{vol_name},},
                        
                        vserver => {type => SCALAR|UNDEF,
                                    optional => 1,
                                    default  => $class->{vserver}, },
                        
                        expected => {type => SCALAR,
                                     optional => 1,
                                     default => C_RC_SUCCESS,},
                        
                        expected_stage => {type => SCALAR,
                                           default => C_STAGE_COMPRESSING,
                                           optional => 1,},
                        
                        expected_checkpoint_stage => {type => SCALAR,
                                                      optional => 1,},
                        
                        min_bytes_processed => {type => SCALAR,
                                                regex => qr/^\d+$/,
                                                optional => 1,},
                        
                        max_bytes_processed => {type => SCALAR,
                                                regex => qr/^\d+$/,
                                                optional => 1,},
                        
                        timeout => {type => SCALAR,
                                    regex => qr/^\d+$/,
                                    optional => 1,},
                        
                        wait_for_requirements => {type => BOOLEAN,
                                                  optional => 1,
                                                  default => FALSE,}, 
		} );
        
        my $expected = $args{expected};
        my $vol_name = $args{volume};
        my $vserver = $args{vserver};
        my $expected_stage = $class->{_ui}->{constants}->get_constant_value(constant_code => $args{expected_stage});
        my $expected_checkpoint_stage;
        if (defined($args{expected_checkpoint_stage}))
        {
                $expected_checkpoint_stage = $class->{_ui}->{constants}->get_constant_value(constant_code => $args{expected_checkpoint_stage});
        }
        my $min_bytes_processed = $args{min_bytes_processed};
        my $max_bytes_processed = $args{max_bytes_processed};
        my $timeout = $args{timeout};
        my $wait_for_requirements = $args{wait_for_requirements};
        my %checkpoint;
        my $rc;
        
        if ($wait_for_requirements == TRUE)
        {
                # Set wait_for_requirements to false and delete timeout
                $args{wait_for_requirements} = FALSE;
                delete($args{timeout});
                $args{expected} = C_RC_FAILURE_NON_FATAL;
                if (!defined($timeout))
                {
                        $timeout = C_DFLT_COMPRESS_TIMEOUT;
                }
                
                my $comment = "Waiting for the scanner running on volume $vol_name to meet the requirements:\n" .
                              "Scanner stage: $expected_stage\n";
                $comment .= "Checkpoint stage: $expected_checkpoint_stage\n" if (defined($expected_checkpoint_stage));
                $comment .= "Min bytes processed: $min_bytes_processed\n" if (defined($min_bytes_processed));
                $comment .= "Max bytes processed: $max_bytes_processed\n" if (defined($max_bytes_processed));
                $comment .= "Timeout: $timeout";
                
                logcomment($comment);
                
                my $poll_rc = C_RC_UNEXPECTED_SCANNER_STAGE;
                while (($timeout >= 0) && (($poll_rc == C_RC_UNEXPECTED_SCANNER_STAGE) ||
                                          ($poll_rc == C_RC_UNEXPECTED_CHECKPOINT_STAGE) ||
                                          ($poll_rc == C_RC_DATA_RANGE_MISSED) ||
                                          ($poll_rc == C_RC_SCANNER_IDLE)))
                {
                        if ($timeout == 0)
                        {
                                logresult('fatal', "Timed-out while waiting for required stage or total data scanned:\n" . Dumper(%args) .
                                                   "\n\nLast obtained status:\n" . Dumper(%checkpoint));
                        }
                        
                        %checkpoint = $class->verify_checkpoint(%args);
                        logcomment("CompressionScanner: verify_checkpoint (recursive): timeout=$timeout") if ($class->{_debug});
                        $poll_rc = $checkpoint{rc};
                        sleep(1);
                        $timeout--;
                }
                
                $rc = $poll_rc;
        }
        else
        {
                %checkpoint = $class->get_checkpoint(volume => $vol_name,
                                                     vserver => $vserver,
                                                     timeout => $timeout);
                
                $rc = $checkpoint{rc};
                
                my $scanner_stage = $checkpoint{scanner_stage};
                if ($scanner_stage ne $expected_stage)
                {
                        logcomment("Failed to detect the required scanner stage:\n" .
                                   "Required stage: \"$expected_stage\"\n" .
                                   "Detected stage: \"$scanner_stage\"");
                        $rc = C_RC_UNEXPECTED_SCANNER_STAGE;
                }
                
                my $checkpoint_stage = $checkpoint{checkpoint_stage};
                if (!defined($checkpoint_stage))
                {
                        logresult('fatal', "Failed to detect the checkpoint stage");
                }
                if (defined($expected_checkpoint_stage) && ($expected_checkpoint_stage ne $checkpoint_stage))
                {
                        logcomment("Failed to detect the required checkpoint stage:\n" .
                                   "Required stage: \"$expected_checkpoint_stage\"\n" .
                                   "Detected stage: \"$checkpoint_stage\"");
                        $rc = C_RC_UNEXPECTED_CHECKPOINT_STAGE;
                }
                
                my $total_bytes_processed = $checkpoint{total_bytes_processed};
                if (defined($min_bytes_processed) && ($total_bytes_processed < $min_bytes_processed))
                {
                        logcomment("Data processed by scanner is smaller than the required lower limit:\n" .
                                   "Lower limit: \"$min_bytes_processed\"\n" .
                                   "Currently processed: \"$total_bytes_processed\"");
                        $rc = C_RC_DATA_RANGE_MISSED;
                }
                
                if (defined($max_bytes_processed) && ($total_bytes_processed > $max_bytes_processed))
                {
                        logcomment("Data processed by scanner is bigger than the required upper limit:\n" .
                                   "Upper limit: \"$max_bytes_processed\"\n" .
                                   "Currently processed: \"$total_bytes_processed\"");
                        $rc = C_RC_DATA_RANGE_MISSED;
                }
        }
        
        $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                             expected_rc => $expected,
                                             log_message => "Verify checkpoint for volume $vol_name",
                                             output      => $checkpoint{output});
        
        $checkpoint{rc} = $rc;
        logcomment("CompressionScanner: verify_checkpoint returning:\n" . Dumper(%checkpoint)) if ($class->{_debug});
        
        return %checkpoint;
}


## @name ::get_last_operation_size
## @summary Get the last operation size. 
## 
## @description
## Get the size (in KB) processed by the last-running scanner. 
##
## @return The size in KB.
## 
## @arg volume Optional: Name of the volume.
## If not provided, uses volume associated with the object, if available.
## If not provided and the object does not have an associated volume, logs script error.
## 
## @arg vserver Optional: Name of the vserver (for c-mode testing).
## If not provided, uses vserver associated with the object, if available.
##
## @examples
## my $size = $obj->get_last_operation_size();
##
sub get_last_operation_size
{
        my $class = shift;
        my %args = validate( @_,
        {
                volume => {type => SCALAR,
                           optional => 1,
                           default => $class->{vol_name},},
                
                vserver => {type => SCALAR|UNDEF,
                            optional => 1,
                            default  => $class->{vserver}, }, 
	});
        
        my $volume = $args{volume};
        
        $args{expected} = C_RC_FAILURE_NON_FATAL;
        my %res = $class->status(%args);
        
        my $status = $res{compress_statistics};
        
        if (!exists($status->{$volume}->{last_operation_size}))
        {
                logresult('fatal', "Failed to detect the last operation size of compress scanner for volume $volume:\n" . Dumper(%res));
        }
        
        my $size = $status->{$volume}->{last_operation_size};
        
        # For c-mode, the last operation size is not in human-friendly format -> it is simply the size in B
        # In this case, there is no extra processing needed
        if ($size =~ /\D/)
        {
                # Remove any spaces to be processed correctly by process_data_size_string
                $size =~ s/\s//g;
                %res = Compression::Utils::process_data_size_string(size_string => $size,);
                $size = $res{size_in_bytes};
        }
        
        $size = $size / KB;
        logcomment("\nCompress scanner: volume: $volume\n" .
                   "Last operation size: $size KB");
        
        return $size;
}



################################################################
# Internal subroutines
################################################################

sub _update_volume_name
{
        my $class = shift;
	my (%args) = validate( @_,
	{
		new_name => { type => SCALAR, },
	} );
        
        $class->{vol_name} = $args{new_name};
        
        logcomment("CompressionScanner: class volume was renamed to: " . $class->{vol_name}) if $class->{_debug};
}

sub _update_vfiler
{
	my $class = shift;
	my $new_vfiler = shift;
	
	$class->{_vfiler} = $new_vfiler;
}


sub _update_conn
{
	my $class = shift;
	my $new_conn = shift;
	
	# This assumes that the new filer is running the same version 
	if (($class->{_boot_mode} eq C_ONTAP_7MODE) && !($new_conn->isa("NOTE::Filer")))
        {
                logresult('script', "For 7-mode, the conn argument is required to be a NOTE\:\:Filer object");
        }
        elsif (($class->{_boot_mode} eq C_ONTAP_CMODE) && !($new_conn->isa("NOTE::GX::Node")))
        {
                logresult('script', "For 10-mode, the conn argument is required to be a NOTE\:\:GX\:\:Node object");
        }
	$class->{_conn} = $new_conn;
}


# Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift @_;
	# WARNING: We need to store any eventual error message and fail it at the end of any destructor.
	# Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
	# At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
	# WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        
	my $execution_error = $@;
	if ((defined $class->{vol_name}) &&
            ($class->{_disable_cleanup} == FALSE) &&
            ($class->{_creator_pid} == $$))
        {
                my $vol_name = $class->{vol_name};
                logcomment("Executing CompressionScanner clean-up for volume $vol_name");

                eval
                {
                        my %compress_status = $class->status(expected => C_RC_FAILURE_NON_FATAL);
                        if ($compress_status{rc} == C_RC_SUCCESS)
                        {
                                $class->stop();
                        }
                };

                if( $@ )
                {
                        logcomment( "CompressionScanner: cleanup for volume $vol_name: error: " . $@ );
                }
                else
                {
                        logcomment( "CompressionScanner: cleanup for volume $vol_name: done" ) if ($class->{_debug});
                }
        }
        else
        {
                logcomment( "CompressionScanner: cleanup:  nothing done: script PID: " . $$) if ($class->{_debug});
        }
        
        if (defined($execution_error))
        {
                _c_interpret_logresult($execution_error);
        }
}


#routine to test the module itself
sub testme
{

	package main;
        use Data::Dumper;
	
        require "drg/Compression/setup_core.pl";
        use Compression::Constants qw(:all);
        
        use vars qw($FILER $AGGREGATE $VOLUME $DEBUG);
        
        param("DEBUG", "-default", FALSE);
        
        my $compress = Compression::Compress->setup(filer => $FILER ,
                                                    aggregate => $AGGREGATE,
                                                    volume => {name => $VOLUME,
                                                               size => "3g",
                                                               compress_enabled => FALSE,
                                                               },
                                                    debug => $DEBUG,
                                                    );
        
        my $filename = 'large_file';
        
        my $expected_rcs = $compress->{compress_scanner}->_get_testme_expected_results();

        $compress->wait_for_cp();
        $compress->{dataset}->create(files => [{ name => $filename, size => 2 * GB, },],);
        $compress->wait_for_cp();
        $compress->verify_savings(expected_compressed => 0);
        
        $compress->{compress_scanner}->get_checkpoint(expected => C_RC_SCANNER_IDLE);
        $compress->{compress_scanner}->status(expected => C_RC_SCANNER_IDLE);
        $compress->{compress_scanner}->start(scan_to_compress => TRUE,
                                             scan => TRUE,
                                             expected => C_RC_COMPRESSION_NOT_TURNED_ON);
        $compress->{compress_scanner}->start(scan_to_compress => TRUE,
                                             scan => TRUE,
                                             deduped_blocks => TRUE,
                                             expected => C_RC_COMPRESSION_NOT_TURNED_ON);
        $compress->{compress_scanner}->start(scan_to_compress => TRUE,
                                             scan => TRUE,
                                             snapshotted_blocks => TRUE,
                                             expected => C_RC_COMPRESSION_NOT_TURNED_ON);

        $compress->enable();
        $compress->wait_for_cp();
        
        $compress->{compress_scanner}->start();
        $compress->{compress_scanner}->start(expected => C_RC_COMPRESSION_STILL_IN_PROGRESS);
        $compress->{compress_scanner}->verify_checkpoint(expected => $expected_rcs->{checkpoint_for_less_than_1g},
                                                         expected_stage => C_STAGE_COMPRESSING,
                                                         min_bytes_processed => MB,
                                                         max_bytes_processed => 900 * MB,
                                                         wait_for_requirements => TRUE,
                                                         );
        
        $compress->{compress_scanner}->status();
        
        $compress->{compress_scanner}->verify_checkpoint(expected => C_RC_SUCCESS,
                                                         expected_stage => C_STAGE_COMPRESSING,
                                                         min_bytes_processed => GB,
                                                         max_bytes_processed => 2 * GB,
                                                         wait_for_requirements => TRUE,
                                                         );
        
        $compress->{compress_scanner}->stop();
        $compress->wait_for_cp(num_cp => 5); 
        $compress->{compress_scanner}->status(expected => C_RC_SCANNER_IDLE);

        $compress->wait_for_cp();
        $compress->{compress_scanner}->start();
        $compress->{compress_scanner}->wait_for_scanned_data();
        $compress->{compress_scanner}->status(expected => C_RC_SCANNER_IDLE);

        $compress->wait_for_cp();
        $compress->verify_savings(expected_compressed => 1833980, allow_percentage_error => 20);
        
        logresult('PASS', "CompressionScanner: testme: Done");
}


sub _get_testme_expected_results
{
	my $class = shift;
	
	my $ontap = $class->{_ui}->{version}->{ontap};
	
	if ($ontap eq '7.3')
        {
                return {checkpoint_for_less_than_1g => C_RC_CHECKPOINT_FILE_EMPTY,
                        };
	}
        elsif ($ontap eq '8.0')
        {
                return {checkpoint_for_less_than_1g => C_RC_CHECKPOINT_FILE_EMPTY,
                        };
	}
        else
        {
                return {checkpoint_for_less_than_1g => C_RC_SUCCESS,
                        };
	}
}


1;

