#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/CompressionUtilities.pm#1 $
#

## @summary Compression utilities module.
##
## @description
##    This module provides the basic compression utilities.
##
## @synopsis
## use Compression::CompressionUtilities;
## use Compression::Constants qw(:all);
## use NOTE::Filer;
## use vars qw($FILER); 
## 
## BEGIN { require "Compression/setup_core.pl"; }
## 
## my $obj = Compression::CompressionUtilities->new( conn => $conn, volume => 'vol1' );	    
##
## @examples
##
## Used independently:
## my $conn = NOTE::Filer->connect( host => $FILER );
## my $obj = Compression::CompressionUtilities->new( conn => $conn, volume => 'vol1' );
## TODO: documentation throughout
## @examples
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::CompressionUtilities;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Tharn;

use Compression::Constants qw(:all);
use Compression::Version;
use Compression::Utils qw(_c_interpret_logresult compare_values _send_cmd_to_filer);
use SayWrap qw(sayEOF);



BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}



## @name ::new
##
## @summary Compression::CompressionUtilities Constructor.
##
## @description
## Creates an Compression::CompressionUtilities object.
##
## @return Compression::CompressionUtilities object.
##
## @arg filer Required: NOTE::Filer object.
##
## @arg volume Optional: name of volume to be associated with this object.
## 
## @arg vfiler Optional: Name of vfiler.
##
## @arg ui Optional: Compression::UI object corresponding to the hosting filer.
## If not provided, it will be auto-generated.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
## 
## @arg supportability Optional: Compression::Supportability object,
## to be passed in if, along with the command completion, is it required
## to also check the supportability messages.
##
## @example
##
## my $obj = Compression::CompressionUtilities->new( conn => $conn, volume => 'vol1' );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
		conn 	=> {type => OBJECT,
                            },

		volume  => {type => SCALAR,
                            optional => 1,
                            },
                
                ui      => {type => OBJECT,
                            isa => "Compression::UI",
                            optional => 1,
                            },
                
                disable_cleanup => {type => BOOLEAN,
                                    optional => 1,
                                    default => FALSE,
                                    },
                
                debug  => {type => SCALAR,
                           optional => 1,
                           default => FALSE,
                           },
                
                vserver => {type => SCALAR|UNDEF,
                            optional => 1},
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default => C_ONTAP_BOOT_MODE},
		
                vfiler => {type => SCALAR|UNDEF,
                           optional => 1,},
                
                supportability => {type => UNDEF|OBJECT,
                                   optional => 1,},
                } );
	
	my $self = {};
        $self->{vol_name} = $args{volume} if (defined $args{volume});
	$self->{_conn} = $args{conn};
        $self->{_debug} = $args{debug};
        $self->{_disable_cleanup} = $args{disable_cleanup};
        $self->{_creator_pid} = $$;
        $self->{_boot_mode} = $args{boot_mode};
        $self->{vserver} = $args{vserver};
	$self->{_vfiler} = $args{vfiler};
        $self->{_supportability} = $args{supportability};
        
        if (($self->{_boot_mode} eq C_ONTAP_7MODE) && !($self->{_conn}->isa("NOTE::Filer")))
        {
                logresult('script', "For 7-mode, the conn argument is required to be a NOTE\:\:Filer object");
        }
        elsif (($self->{_boot_mode} eq C_ONTAP_CMODE) && !($self->{_conn}->isa("NOTE::GX::Node")))
        {
                logresult('script', "For 10-mode, the conn argument is required to be a NOTE\:\:GX\:\:Node object");
        }
        
        if ((defined($self->{_supportability})) && (!($self->{_supportability}->isa("Compression::Supportability"))))
        {
                logresult('script', "The supportability argument, if passed in, must be a Compression\:\:Supportability object");
        }
        
        logcomment("CompressionUtilities: new: PID of the script creating this object: " . $self->{_creator_pid}) if ($self->{_debug});
	
	####################################################
	# Dependency on other internal modules
	####################################################
	
        # UI
        if ( defined($args{ui}) )
        {
                $self->{_ui} = $args{ui};
        }
        else
        {
                # Auto-detect version of the hosting filer. Need this for UI.
                logcomment("CompressionUtilities: UI object not provided. Auto-generating...") if ($self->{_debug});
                my $version = Compression::Version->new( conn => $args{conn}, debug => $self->{_debug} );
                $self->{_ui} = Compression::UI->new( version => $version, debug => $args{debug}, boot_mode => $self->{_boot_mode} );
        }
	
	bless $self, $class;

	return $self;
}


## @name ::toggle_compression
## @summary Flips the compression setting for the volume.
##
## @description
## Flips the compression setting for the volume.
##
## @return Hash with rc and output.
##
## @examples
## $obj->toggle_compression();
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub toggle_compression
{
	my $class 	= shift;
	my (%args) 	= validate( @_,
			{
				# No args
			} );

	my %current_setting = $class->is_compression_enabled(expected => C_RC_FAILURE_NON_FATAL);
	
	# For now ignore inline
	if ($current_setting{rc} == _C_RC_COMPR_OFF)
	{
		return $class->enable();
	}
	else
	{
		return $class->disable();
	}
}


## @name ::enable_compression
## @summary Enables compression for the volume.
##
## @description
## Enables compression for the volume.
##
## @return Hash with rc and output.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the Compress object was created.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @arg check_if_already_set Optional: Flag to enable the library to first check the current setting.
## If compression is already enabled, the subroutine will not attempt to enable it again.
## Instead, it will return success. The default value for check_if_already_set is C_DFLT_CHECK_OPTION_FIRST.
##
## @arg inline Optional: Flag to control if inline compression should be enabled as well (RR and after).
## Legal values: TRUE/FALSE. Default value: release-dependent.
##
## @arg background Optional: Flag to control if background compression should be enabled (RR and after).
## Legal values: TRUE/FALSE. Default value: TRUE.
## It is highly recommended to use check_if_already_set false whenever background is false.
## WARNING: Setting background to false means that compression will not be enabled.
## This flag should only be used for negative-testing.
##
## @examples
## $compress->enable( expected => C_RC_NO_COMPRESSION_LICENSE );
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub enable_compression
{
	my $class 	= shift;
	my (%args) 	= validate( @_,
		{
                        volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
                        vserver => { type => SCALAR|UNDEF, optional => 1, default => $class->{vserver} },
                        expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
                        check_if_already_set => { type => BOOLEAN, optional => 1, default => C_DFLT_CHECK_OPTION_FIRST },
                        inline => { type => BOOLEAN, optional => 1, default => $main::ENABLE_INLINE },
                        background => { type => SCALAR, optional => 1, default => TRUE },
		} );

	logcomment( "CompressionUtilities: Enable Compression\n". Dumper(%args) ) if ($class->{_debug});
	
	my $volume = $args{volume};
        my $vserver = $args{vserver};
	my $expected = $args{expected};
	my $check_if_already_set = $args{check_if_already_set};
	my $inline = $args{inline};
	if (!defined($inline))
	{
		$inline = $class->{_ui}->get_constant_value( constant_code => 'C_DFLT_ENABLE_INLINE' );
	}
	
	# Warning: background has been added only for negative-testing
	# If background is false, we basically do not enable compression
	# In this case, don't change $wanted_setting (only allow background false with non-success rc)
	my $wanted_setting = $expected;
	if ($expected == C_RC_SUCCESS)
	{
		$wanted_setting = ($inline == TRUE) ? _C_RC_COMPR_ON_FULLY_FUNCTIONAL : _C_RC_COMPR_ON_NO_INLINE;
	}
	
	if ($check_if_already_set == TRUE)
	{
		my %res = $class->is_compression_enabled(volume => $volume,
                                                         vserver => $vserver,
                                                         expected => C_RC_FAILURE_NON_FATAL);
		
		logcomment("Compression required to be enabled to: $wanted_setting\n" .
			   "Current setting: " . $res{rc}) if ($class->{_debug});
		
		if ($res{rc} == $wanted_setting)
		{
			logcomment( "CompressionUtilities: compression is already enabled\n") if ($class->{_debug});
			return %res;
		}
	}
        
        my $cmd_code = C_CMD_ENABLE_COMPRESSION;
        
        if ((defined($class->{_supportability})) && ($expected == C_RC_SUCCESS))
        {
                $class->{_supportability}->start_tailing(events => [{code => C_EVENT_COMPRESSION_ENABLED,
                                                                     options => {inline => $inline,},},
                                                                    ],
                                                         required_identifier => $volume,
                                                         );
        }
	
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
						       volume => $volume,
                                                       vserver => $vserver,
						       options => {inline => $inline,
								   background => $args{background},},);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code));
	
	my %enable = $class->{_ui}->process_output_string(code => $cmd_code,
							  output => $output,
							  expected_rc => C_RC_FAILURE_NON_FATAL,
							  log_message => "Enable compression for volume $volume",);
	
	my $ontap = $class->{_ui}->{version}->{ontap};
	if (($enable{rc} == C_RC_SUCCESS) && ($expected == C_RC_SUCCESS))
	{
                if (defined($class->{_supportability}))
                {
                        $class->{_supportability}->stop_and_match();
                }
                
		logcomment("Compression was enabled without errors. Double-checking...");
		my $enable_inline = ($inline == TRUE) ? C_RC_SUCCESS : C_RC_FAILURE;
		return $class->is_compression_enabled(volume => $volume,
                                                      expected => C_RC_SUCCESS,
                                                      inline => $enable_inline);
	}
	
	return $class->{_ui}->check_for_expected_rc(obtained_rc => $enable{rc},
						    expected_rc => $wanted_setting,
						    log_message => "Enable compression for volume $volume",
						    output      => $output);
}


## @name ::disable_compression
## @summary Disables the compression for the volume.
##
## @description
## Disables the compression for the volume.
##
## @return Hash with rc and output.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @arg volume Optional: Name of volume for which to disable compression.
## Defaults to the volume for which the CompressionUtilities object was created.
##
## @arg all_volumes Optional: Flag to disable compression on all existent volumes.
##
## @arg skip_root_volume Optional: Only considered if 'all_volumes' option is TRUE.
## Boolean flag to decide if compression should be attempted to be disabled for the root volume.
## Defaults to TRUE.
##
## @arg skip_offline_restricted Optional: Only considered if 'all_volumes' option is TRUE.
## Boolean flag to decide if compression should be attempted to be disabled for offline/restricted volumes.
## Defaults to TRUE.
##
## @arg check_if_already_set Optional: Flag to enable the library to first check the current setting.
## If compression is already disabled, the subroutine will not attempt to disable it again.
## Instead, it will return success. The default value for check_if_already_set is C_DFLT_CHECK_OPTION_FIRST.
## If the all_volumes argument is TRUE, then check_if_already_set is automatically set to TRUE.
##
## @arg inline Optional: Flag to control if only inline compression should be disabled.
## Legal values: TRUE/FALSE. Default value: release-dependent.
##
## @notes
## If all_volumes flag is set to true, the expected result is considered to always be success (no 'expected' handling).
##
## @examples
## $obj->disable_compression();
## $obj->disable_compression( all_volumes => TRUE );
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub disable_compression
{
	my $class 	= shift;
	my (%args) 	= validate( @_,
		{
                        volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
                        vserver => { type => SCALAR|UNDEF, optional => 1, default => $class->{vserver} },
                        expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
                        all_volumes => { type => BOOLEAN, optional => 1, default => FALSE },
                        skip_root_volume => { type => BOOLEAN, optional => 1, default => TRUE, },
                        skip_vsroot_volumes => { type => BOOLEAN, optional => 1, default => TRUE, }, # TODO document
                        skip_offline_restricted => { type => BOOLEAN, optional => 1, default => TRUE, },
                        check_if_already_set => { type => BOOLEAN, optional => 1, default => C_DFLT_CHECK_OPTION_FIRST },
                        inline => { type => BOOLEAN, optional => 1,}, 
		} );

	logcomment( "CompressionUtilities: Disable Compression\n". Dumper(%args) ) if ($class->{_debug});
	
	my $skip_offline_restricted_flag = $args{skip_offline_restricted};
	my $skip_root_flag = $args{skip_root_volume};
        my $skip_vsroot_flag = $args{skip_vsroot_volumes};
	my $inline = $args{inline};
	if (!defined($inline))
	{
		$inline = $class->{_ui}->get_constant_value( constant_code => 'C_DFLT_DISABLE_INLINE' );
	}

	if ($args{all_volumes} == TRUE)
	{
                if ($class->{_boot_mode} eq C_ONTAP_CMODE)
                {
                        my $node_obj = $class->{_conn};
                        my $node_name = $node_obj->name();
                        my $detected_mroot_vol = 0;
                        # Detect the name of the mroot aggregate first
                        my $mroot_aggr = $node_obj->server->storage_aggregate_show('has-mroot' => 'true',
                                                                                   'nodes' => $node_name);
                        my $root_aggr_name = $mroot_aggr->[0]->{aggregate};
                        logcomment("Mroot aggregate of node $node_name is $root_aggr_name");
                        
                        # Loop through this node's aggregates
                        my @aggregates = $node_obj->Aggregates();
                        foreach my $aggregate (@aggregates)
                        {
                                my @vol_objs = $aggregate->Volumes();
                                my $aggr_name = $aggregate->name();
                                foreach my $volume (@vol_objs)
                                {
                                        my $volume_name = $volume->name();
                                        # Skip the root volume, if required
                                        if (($skip_root_flag == TRUE) && ($root_aggr_name eq $aggr_name) && ($detected_mroot_vol == 0))
                                        {
                                                logcomment("CompressionUtilities: disable_compression: Processing the mroot aggregate.\n" .
                                                           "Looking for the mroot volume name") if ($class->{_debug});
                                                
                                                my $cmd_code = C_CMD_DETECT_ROOT_VOL;
                                                my $command = $class->{_ui}->formulate_command(code => $cmd_code,
                                                                                               volume => $volume_name);
                                                
                                                my $output = Compression::Utils::_send_cmd_to_filer(conn => $node_obj,
                                                                                                    boot_mode => C_ONTAP_CMODE,
                                                                                                    cmd => $command,
                                                                                                    node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
												    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code));
                                                
                                                my %processed_hash = $class->{_ui}->process_output_string(code => $cmd_code,
                                                                                                          output => $output,
                                                                                                          expected_rc => C_RC_FAILURE_NON_FATAL,
                                                                                                          log_message => "Obtain dblade-level volume status for $volume_name",);
                                                
                                                if ($processed_hash{rc} == C_RC_ROOT_VOLUME_DETECTED)
                                                {
                                                        $detected_mroot_vol = 1;
                                                        logcomment("CompressionUtilities: disable_compression: Skipping the mroot volume: $volume_name");
                                                        next;
                                                }
                                        }
                                        
                                        # Skip the vsroot volumes, if required
                                        if (($skip_vsroot_flag == TRUE) && ($node_obj->server->is_root_volume(volume => $volume_name)))
                                        {
                                                logcomment("CompressionUtilities: disable_compression: Skipping vsroot volume: $volume_name");
                                                next;
                                        }
                                        
                                        # If required, skip volume if offline or restricted
                                        if ($skip_offline_restricted_flag == TRUE)
                                        {
                                                my $vol_state = $volume->status("state");
                                                if ($vol_state =~ /offline|restricted/)
                                                {
                                                        logcomment("CompressionUtilities: disable_compression: Skipping $vol_state volume $volume_name");
                                                        next;
                                                }
                                        }
                                        
                                        $class->_disable_compression_on_one_volume(volume => $volume_name,
                                                                                   expected => C_RC_SUCCESS,
                                                                                   check_if_already_set => TRUE,
                                                                                   inline => $inline);
                                }
                        }
                }
                else
                {
                        my $filer_obj = $class->{_conn};
                        # Get the name of the root volume, to be skipped, if required.
                        my $root_vol_name = $filer_obj->vol_root();
                        logcomment("Root volume is $root_vol_name") if ($class->{_debug});
                        
                        my @vol_list = $filer_obj->Volumes;
                        foreach my $volume (@vol_list)
                        {
                                my $volume_name = $volume->name();
                                
                                # Skip the root volume, if required
                                if (($skip_root_flag == TRUE) && ($volume_name eq $root_vol_name))
                                {
                                        logcomment("Skipping root volume $volume_name");
                                        next;
                                }
                                
                                # If required, skip volume if offline or restricted
                                if ($skip_offline_restricted_flag == TRUE)
                                {
                                        my $vol_status = $volume->status();
                                        my $vol_state = $vol_status->{state};
                                        if ( $vol_state =~ /offline|restricted/ )
                                        {
                                                logcomment("Skipping volume \"$volume_name\", in state: \"$vol_state\" ");
                                                next;
                                        }
                                }
                                
                                $class->_disable_compression_on_one_volume(volume => $volume_name,
                                                                           expected => C_RC_SUCCESS,
                                                                           check_if_already_set => TRUE,
                                                                           inline => $inline);
                        }
                }
		return (rc => C_RC_SUCCESS);
	}
	else
	{
		return $class->_disable_compression_on_one_volume(volume => $args{volume},
                                                                  vserver => $args{vserver},
								  expected => $args{expected},
								  check_if_already_set => $args{check_if_already_set},
								  inline => $inline);
	}
}


## @name ::is_compression_enabled
## @summary Verifies that compression is enabled for the volume.
##
## @description
## Verifies that compression is enabled for the volume.
##
## @return Hash with rc and output.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes that compression be enabled.
## Legal values: C_RC_SUCCESS, C_RC_FAILURE
##
## @arg inline Optional: Expected return code (defined in Compression::Constants) for inline compression check.
## Legal values: C_RC_SUCCESS, C_RC_FAILURE.
##
## @examples
## $obj->is_compression_enabled( expected => C_RC_SUCCESS );
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_compression_enabled
{
        my $class = shift;
        my (%args)      = validate( @_,
                {
                        volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
                        vserver => { type => SCALAR|UNDEF, optional => 1, default => $class->{vserver} },
                        expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
                        inline => { type => SCALAR, optional => 1, default => $main::IS_INLINE_ENABLED},
                } );

        logcomment("CompressionUtilities: is_compression_enabled:\n". Dumper(%args) ) if ($class->{_debug});
	my $volume = $args{volume};
        my $vserver = $args{vserver};
	my $inline = $args{inline};
	if (!defined($inline))
	{
		$inline = $class->{_ui}->get_constant_value( constant_code => 'C_DFLT_IS_INLINE_ENABLED' );
	}
	
	my $expected = $args{expected};
	if ($expected == C_RC_FAILURE)
	{
		$expected = _C_RC_COMPR_OFF;
	}
	elsif ($expected == C_RC_SUCCESS)
	{
		$expected = ($inline == C_RC_FAILURE) ? _C_RC_COMPR_ON_NO_INLINE : _C_RC_COMPR_ON_FULLY_FUNCTIONAL;
	}
	
	my $cmd_code = C_CMD_IS_COMPRESSION_ENABLED;
	
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $volume,
                                                       vserver => $vserver);
	
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code));
	
	return $class->{_ui}->process_output_string(code => $cmd_code,
						    output => $output,
						    expected_rc => $expected,
						    log_message => "Check if compression is enabled for volume $volume",);
}


## @name ::get_space_savings
## @summary Retrieves the overall savings of the volume.
##
## @description
## Retrieves the overall savings of the volume using the 'df -S<modifier>' command.
##
## @return Hash with rc, output, and savings.
##
## @arg modifier Optional: Output modifier to be used when retrieving savings.
## Defaults to 'k' (which means that the command to obtain the savings will be 'df -Sk').
##
## @arg all_volumes Optional: Flag to decide if savings should be retrieved for all volumes.
## Defaults to FALSE.
##
## @arg overwrite_args_string Optional: String to be passed as argument to the df command.
## The command that will be sent to the filer will be 'df -S<modifier> <overwrite_args_string>.
## If this argument is used, any successful command output will be processed such that
## the 'savings' hash will have as keys the exact string that identifies the data line.
## For example, if overwrite_args_string is 'vol_name' , the corresponding savings key hash will be
## '/vol/vol_name/', instead of 'vol_name', as it would be when getting the savings for the associated volume.
## Another difference from the case of the associated volume is that
## no checking is done on the output to ensure that there indeed was an entry in the output for 'vol_name'.
## If this is desired, then the script should specifically check $result{savings}->{vol_name}.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @examples
## $obj->get_space_savings( modifier => 'g' );
## $obj->get_space_savings( all_volumes => TRUE, modifier => 'm' );
## $obj->get_space_savings( overwrite_args_string => 'vol1 vol2', expected => C_RC_INVALID_COMMAND_ARGUMENT );
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub get_space_savings
{
	my $class = shift;
	my (%args) 	= validate( @_,
		{
                        modifier => { type => SCALAR, optional => 1, default => "k" },
                        all_volumes => { type => BOOLEAN, optional => 1, default => FALSE },
                        overwrite_args_string => { type => SCALAR, optional => 1, },
                        expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
		} );

	logcomment("CompressionUtilities: arguments for get_space_savings:\n ". Dumper(%args) ) if ($class->{_debug});

	my $overwrite_args_string = $args{overwrite_args_string};
	if (($args{all_volumes} == TRUE) && (defined($overwrite_args_string)))
	{
		logresult("script", "CompressionUtilities: get_space_savings: Arguments \"all_volumes\" and \"overwrite_args_string\" cannot be used togheter");
	}

	my $volume = $class->{vol_name};
	if (($args{all_volumes} == TRUE) || (defined($overwrite_args_string)))
	{
		$volume = undef;
	}

	my $modifier = $args{modifier};
	my $cmd_code = C_CMD_GET_SAVINGS;
        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $volume,
						       modifier => $modifier);

	if (defined($overwrite_args_string))
	{
		$command .= " $overwrite_args_string";
	}

        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);

	my $log_message = 'Obtain space savings';
	if (defined($volume))
	{
		$log_message .= " for volume $volume";
	}

	my %ui_args = (code => $cmd_code,
		       output => $output,
		       required_return_hash_key => $volume,
		       modifier => $modifier,
		       expected_rc => $args{expected},
		       log_message => $log_message,);

	if (defined($overwrite_args_string))
	{
		$ui_args{disable_data_identifier_processing} = TRUE;
	}

	return $class->{_ui}->process_output_string( %ui_args );
}


## @name ::get_compression_statistics
## @summary Displays the statistics for the volume.
##
## @description
## Displays the statistics for the volume.
##
## @arg volume Optional: Name of volume for which to display statistics.
## Defaults to the volume for which the CompressionUtilities object was created.
##
## @return Hash with rc, statistics, and output.
##
## @examples
## my %res = $obj->get_compression_statistics();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_compression_statistics
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
                        vserver => { type => SCALAR|UNDEF, optional => 1, default => $class->{vserver} },
		} );

	my $cmd_code = C_CMD_GET_STATS;
	my $volume = $args{volume};
	
	my $command = $class->{_ui}->formulate_command(code => $cmd_code,
                                                       vserver => $args{vserver},
                                                       volume => $volume,);
	
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
	
	return $class->{_ui}->process_output_string(code => $cmd_code,
						    output => $output,
						    required_return_hash_key => $volume,
						    expected_rc => C_RC_SUCCESS,
						    log_message => "Obtain statistics for volume $volume",);
}


## @name ::obtain_compression_attempted_bits
## @summary Obtains the "compression attempted" fingerprint bits.
##
## @description
## Obtains the "compression attempted" fingerprint bits.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the CompressionUtilities object was created.
##
## @arg inode Optional: Inode for which to get the fingerprints.
##
## @arg fbns Optional: Fbns for which to get the fingerprints.
## Can be: specific value (ex: 8), range (ex: "3-7"), or "all".
## Default: "all".
##
## @arg aggregate Optional: Boolean flag to look for aggregate-level fingerprints.
## Default: FALSE.
##
## @arg vvol Optional: Boolean flag to look for vvol-level fingerprints.
## Default: FALSE.
##
## @arg changelog Optional: Boolean flag to look for changelog fingerprints.
## Default: TRUE.
##
## @arg remove_stale_entries Optional: Boolean flag to use if only the non-stale entries are wanted
## (latest cp-count). Default: FALSE.
##
## @arg expected Optional: Optional: Expected return code (C_RC_SUCCESS/C_RC_FAILURE).
## Default: C_RC_SUCCESS.
##
## @return Hash with rc, count, aggregate, vvol, checking_stats, changelog_0, changelog_1, and output.
##
## @examples
## $compress->obtain_compression_attempted_bits(inode => 96,
##                                             fbns => "4-7",
##                                             changelog => FALSE,
##                                             aggregate => TRUE);
## $compress->obtain_compression_attempted_bits(inode => 96,
##                                             fbns => 8,
##                                             remove_stale_entries => TRUE,
##                                             expected => C_RC_FAILURE);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub obtain_compression_attempted_bits
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
                        vserver => { type => SCALAR|UNDEF, optional => 1, default => $class->{vserver} },
			inode => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
			fbns => { type => SCALAR, default => 'all',},
			aggregate => { type => BOOLEAN, default => FALSE, },
			vvol => { type => BOOLEAN, default => FALSE, },
			changelog => { type => BOOLEAN, default => TRUE,},
			remove_stale_entries => { type => BOOLEAN, default => FALSE },
			expected => { type => SCALAR, default => C_RC_SUCCESS },
		} );

	my $cmd_code = C_CMD_VERIFY_COMPRESSION_ATTEMPTED;
	my $volume = $args{volume};
	my $inode = $args{inode};
	my $fbns = $args{fbns};
	my $aggregate = $args{aggregate};
	my $vvol = $args{vvol};
	my $changelog = $args{changelog};
	my $count = 0;
	my $rc = C_RC_SUCCESS;
	my $fingerprints;
	my %triaged;
	
	if (($fbns !~ /^\d+$/) && ($fbns !~ /^\d+-\d+$/) && ($fbns !~ /^all$/))
	{
		logresult('script', "obtain_compression_attempted_bits: invalid fbns argument: $fbns\n" .
			            "Legal values:\n" .
				    "- an individual value (ex: 8)\n" .
				    "- a range (ex: 1-7)\n" .
				    "- all");
	}
	
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $volume,
                                                       vserver => $args{vserver});
	
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
	
	my %processed_hash = $class->{_ui}->process_output_string(code => $cmd_code,
								  output => $output,
								  expected_rc => C_RC_SUCCESS,
								  log_message => "Obtain the compression-attempted bits for volume $volume",);
	
	# Return only what we are asked to
	my %return_hash = (rc => $processed_hash{rc}, output => $processed_hash{output});
	
	if ($aggregate == TRUE)
	{
		$fingerprints = $processed_hash{fingerprints};
		%triaged = $class->_triage_fingerprints(fingerprint_hash => $fingerprints,
							inode => $inode,
							fbns => $fbns,
							remove_stale => $args{remove_stale_entries},
							hash_meaning => "aggregate fingerprints for volume $volume",);
		$return_hash{aggregate} = $triaged{triaged};
		$count += $triaged{count};
	}
	
	if ($vvol == TRUE)
	{
		$fingerprints = $processed_hash{fingerprints_vvol};
		%triaged = $class->_triage_fingerprints(fingerprint_hash => $fingerprints,
							inode => $inode,
							fbns => $fbns,
							remove_stale => $args{remove_stale_entries},
							hash_meaning => "vvol fingerprints for volume $volume",);
		$return_hash{vvol} = $triaged{triaged};
		$count += $triaged{count};
	}
	
	if ($changelog == TRUE)
	{
		$fingerprints = $processed_hash{changelog_buckets};
		# Changelog buckets are extra nested by buckets
		foreach my $bucket (keys %$fingerprints)
		{
			my $return_key = 'changelog_' . $bucket;
			%triaged = $class->_triage_fingerprints(fingerprint_hash => $fingerprints->{$bucket},
								inode => $inode,
								fbns => $fbns,
								remove_stale => $args{remove_stale_entries},
								hash_meaning => "changelog fingerprints for volume $volume",);
			$return_hash{$return_key} = $triaged{triaged};
			$count += $triaged{count};
		}
	}
	
	if ($count == 0)
	{
		$rc = C_RC_FAILURE;
	}
	
	$class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
					     expected_rc => $args{expected},
					     log_message => "Get fingerprints to correspond to:\n" . Dumper(%args),
					     output      => \%processed_hash);
	
	$return_hash{count} = $count;
	return %return_hash;
}


## @name ::verify_compression_attempted
## @summary Verifies the "compression attempted" fingerprint bits.
##
## @description
## Verifies the "compression attempted" fingerprint bits.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the CompressionUtilities object was created.
##
## @arg inode Optional: Inode for which to get the fingerprints.
##
## @arg fbns Optional: Fbns for which to get the fingerprints.
## Can be: specific value (ex: 8), range (ex: "3-7"), or "all".
## Default: "all".
##
## @arg aggregate Optional: Boolean flag to look for aggregate-level fingerprints.
## Default: FALSE.
##
## @arg vvol Optional: Boolean flag to look for vvol-level fingerprints.
## Default: FALSE.
##
## @arg changelog Optional: Boolean flag to look for changelog fingerprints.
## Default: TRUE.
##
## @arg remove_stale_entries Optional: Boolean flag to use if only the non-stale entries are wanted
## (latest cp-count). Default: FALSE.
##
## @arg expected Optional: Expected return code (C_RC_COMPRESSION_ATTEMPTED/C_RC_COMPRESSION_NOT_ATTEMPTED).
## Default: C_RC_COMPRESSION_ATTEMPTED.
##
## @return Hash with rc, aggregate, vvol, checking_stats, changelog_0, changelog_1, and output.
##
## @examples
## my $obj->verify_compression_attempted_bits(inode => 96,
##                                            fbns => "all",
##                                            expected => C_RC_COMPRESSION_NOT_ATTEMPTED);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub verify_compression_attempted_bits
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
                        vserver => { type => SCALAR|UNDEF, optional => 1, default => $class->{vserver} },
			inode => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
			fbns => { type => SCALAR, default => 'all',},
			aggregate => { type => BOOLEAN, default => FALSE, },
			vvol => { type => BOOLEAN, default => FALSE, },
			changelog => { type => BOOLEAN, default => TRUE,},
			remove_stale_entries => { type => BOOLEAN, default => FALSE },
			expected => { type => SCALAR, default => C_RC_COMPRESSION_ATTEMPTED },
		} );

	my $volume = $args{volume};
	# Remember expected, then set the expected args to success, to send it down to obtain_compression_attempted_bits
	my $expected = $args{expected};
	$args{expected} = C_RC_SUCCESS;
	
	# First get only the entries that are required
	my %processed_info = $class->obtain_compression_attempted_bits(%args);
	my $output = $processed_info{output};
	# Delete 'status' hash entries
	delete($processed_info{rc});
	delete($processed_info{count});
	delete($processed_info{output});
	logcomment("Checking compression-attempted bits:\n" . Dumper(%processed_info)) if ($class->{_debug});
	
	my $check_rc = C_RC_FAILURE;
	my $found_null_bit = FALSE;
	my $found_non_null_bit = FALSE;
	foreach my $key (keys %processed_info)
	{
		my $fingerprint_hash = $processed_info{$key};
		
		foreach my $fingerprint (keys %$fingerprint_hash)
		{
			my $occurence_hash = $fingerprint_hash->{$fingerprint};
			
			foreach my $occurence (keys %$occurence_hash)
			{
				my $compression_bit = $occurence_hash->{$occurence}->{compression_attempted_bit};
				logcomment( "CompressionUtilities: checking compression bit for fingerprint $fingerprint: $compression_bit") if ($class->{_debug});
				
				if (!defined($compression_bit))
				{
					logresult('script', "Library problem found: unexpected hash layout for compression-attempted bits");
				}
				elsif ($compression_bit == 0)
				{
					$found_null_bit = TRUE;
					$check_rc = C_RC_COMPRESSION_NOT_ATTEMPTED;
					
					if ($found_non_null_bit == TRUE)
					{
						return $class->{_ui}->check_for_expected_rc(obtained_rc => C_RC_MIXED_COMPRESSION_BITS,
											    expected_rc => $expected,
											    log_message => "Verify the compression-attempted bits for volume $volume",
											    output      => $output);
					}
				}
				elsif ($compression_bit == 1)
				{
					$found_non_null_bit = TRUE;
					$check_rc = C_RC_COMPRESSION_ATTEMPTED;
					
					if ($found_null_bit == TRUE)
					{
						return $class->{_ui}->check_for_expected_rc(obtained_rc => C_RC_MIXED_COMPRESSION_BITS,
											    expected_rc => $expected,
											    log_message => "Verify the compression-attempted bits for volume $volume",
											    output      => $output);
					}
				}
				else
				{
					logresult('fatal', "Unexpected value for compression-attempted bit:\n" . Dumper($occurence));
				}
			}
		}
	}
	
	$class->{_ui}->check_for_expected_rc(obtained_rc => $check_rc,
					     expected_rc => $expected,
					     log_message => "Verify the compression-attempted bits for volume $volume",
					     output      => $output);
	
	# Add the output back to return it
	$processed_info{output} = $output;
	$processed_info{rc} = $check_rc;
	
	return %processed_info;
}


## @name ::wait_for_CP
## @summary Wait for specified number of CPs to occur
##
## @description
## Wait for specified number of CPs to occur, usually used after enabling compression on the volume.
##
## @return Hash with rc and output (empty).
##
## @arg num_cp Optional: Number of CPs to wait.
## Default value: 1.
##
## @arg wait_time Optional: Extra time to wait, in seconds.
## It will be additional wait time after num_cp has elapsed.
## Default value: 1.
##
## @examples
## $obj->wait_for_CP( num_cp => 2, wait_time => 5 );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub wait_for_CP
{
	my $class = shift;
	my %args = validate( @_,
			    {
				num_cp    => { type => SCALAR, regex => qr/^\d+$/, optional => 1, default => 1 },
				wait_time => { type => SCALAR, regex => qr/^\d+$/, optional => 1, default => C_DFLT_EXTRA_WAIT },
			    } );
	my %ret_hash = ();

	logcomment("Compress: wait_for_CP input:\n". Dumper(%args) ) if ($class->{_debug});
	my $num_cp = $args{num_cp};
	my $extra_wait = $args{wait_time};

        logcomment( "Waiting for $num_cp CP(s) ..." );
	for (my $counter = 0; $counter < $num_cp; $counter++)
	{
		logcomment( "Waiting for ONE_CP_TIME (". ONE_CP_TIME .") to elapse" ) if ($class->{_debug});
		sleep(ONE_CP_TIME);
	}

	if ($extra_wait > 0)
	{
		logcomment( "Waiting for $extra_wait extra second(s) to elapse..." );
		sleep($extra_wait);
	}

	logcomment( "Compress: wait_for_CP: Done" ) if ($class->{_debug});

	$ret_hash{rc} = C_RC_SUCCESS;
	$ret_hash{output} = "";
	return %ret_hash;
}


## @name ::get_compression_containment_version
## @summary Obtains containment version for the volume.
##
## @description
## Obtains containment version for the volume.
##
## @return Hash with rc, containment_version, and output.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the Compress object was created.
##
## @examples
## $obj->get_compression_containment_version();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_compression_containment_version
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
                } );
        
        my $cmd_code = C_CMD_GET_CONTAINMENT_VERSION;
        my $volume = $args{volume};
        
        my $command = $class->{_ui}->formulate_command(code => $cmd_code,
                                                       volume => $volume,);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
        
        return $class->{_ui}->process_output_string(code => $cmd_code,
                                                    output => $output,
                                                    expected_rc => C_RC_SUCCESS,
                                                    log_message => "Obtain compression containment version for $volume",);
}


## @name ::show_compression_stats
## @summary Obtains compression stats for the volume.
##
## @description
## Obtains compression stats for the volume.
##
## @return Hash with rc, compression_stats, and output.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the Compress object was created.
##
## @examples
## $compress->show_compression_stats();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub show_compression_stats
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
                } );
        
        my $cmd_code = C_CMD_GET_COMPRESSION_STATS;
        my $volume = $args{volume};
        
        my $command = $class->{_ui}->formulate_command(code => $cmd_code,
                                                       volume => $volume,);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
        
        return $class->{_ui}->process_output_string(code => $cmd_code,
                                                    output => $output,
                                                    expected_rc => C_RC_SUCCESS,
                                                    log_message => "Obtain compression stats for $volume",);
}

## @name ::compression_stats_start_collection
## @summary Start compression statistics collection
##
## @description
## Start compression statistics collection
##
## @return Hash with rc, compression_stats, and output.
##
## @arg expected Optional: expected return
## Defaults: C_RC_SUCCESS
##
## @examples
## $compress->compression_stats_start_collection();
##
## @author weili@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub compression_stats_start_collection
{
        my $class = shift;
	my %args  = validate( @_,
			{
				expected => { type => SCALAR, default => C_RC_SUCCESS },
			} );
        
        my $cmd_code = C_CMD_COMPRESSION_STATS_START;
        my $volume = $class->{vol_name};
        
        my $command = $class->{_ui}->formulate_command(code => $cmd_code,
                                                       volume => $volume,);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
        
        return $class->{_ui}->process_output_string(code => C_CMD_COMPRESSION_STATS_START,
                                                    output => $output,
                                                    expected_rc => $args{expected},
                                                    log_message => "Start to collect compression statistics");
}

## @name ::compression_stats_stop_collection
## @summary Stop compression statistics collection
##
## @description
## Stop compression statistics collection
##
## @return Hash with rc, compression_stats, and output.
##
## @arg expected Optional: expected return
## Defaults: C_RC_SUCCESS
##
## @examples
## $compress->compression_stats_stop_collection();
##
## @author weili@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub compression_stats_stop_collection
{
        my $class = shift;
	my %args  = validate( @_,
			{
				expected => { type => SCALAR, default => C_RC_SUCCESS }
			} );
        
        my $cmd_code = C_CMD_COMPRESSION_STATS_STOP;
        my $volume = $class->{vol_name};
        
        my $command = $class->{_ui}->formulate_command(code => $cmd_code,
                                                       volume => $volume,);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
        
        return $class->{_ui}->process_output_string(code => C_CMD_GET_COMPRESSION_STATS,
                                                    output => $output,
                                                    expected_rc => $args{expected},
                                                    log_message => "Stop to collect compression statistics");
}

## @name ::run_decompress_revert
## @summary Reverts the containment version values.
##
## @description
## Runs decompress revert to the containment version values.
##
## @return Hash with rc, and output.
##
## @arg volume Optional: Name of volume for which to enable compression.
## Defaults to the volume for which the Compress object was created.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @examples
## $compress->run_decompress_revert(check_containment_version => FALSE);
##
## @note
## This command is only available in BR.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub run_decompress_revert
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
                        expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
                } );
        
        my $cmd_code = C_CMD_DECOMPRESS_REVERT;
        my $volume = $args{volume};
        
        my $command = $class->{_ui}->formulate_command(code => $cmd_code,
                                                       volume => $volume,);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
        
        return $class->{_ui}->process_output_string(code => $cmd_code,
                                                    output => $output,
                                                    expected_rc => $args{expected},
                                                    log_message => "Run decompress revert for $volume",);
}


## @name ::configure_IDD
## @summary Enables/disables IDD.
##
## @description
## Enables/disables IDD.
##
## @return Hash with rc, and output.
##
## @arg volume Optional: Name of volume for which to configure IDD.
## Defaults to the volume for which the object was created.
##
## @arg vserver Optional: Vserver name (for C-mode only).
## Defaults to the vserver for which the object was created.
##
## @arg enabled Optional: Flag to control if IDD should be enabled or disabled.
## Legal values: TRUE(1)/FALSE(0). Default: TRUE.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @examples
## $obj->configure_IDD();
## $obj->configure_IDD(enabled => FALSE);
##
sub configure_IDD
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			volume => {type => SCALAR, optional => 1, default => $class->{vol_name}},
			vserver => {type => SCALAR|UNDEF, optional => 1, default => $class->{vserver}},
			enabled => {type => SCALAR, optional => 1, default => TRUE,},
			expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS},
		});
	
	my $cmd_code = C_CMD_CONFIGURE_IDD;
	my $volume = $args{volume};
	my $enabled_string = $class->{_ui}->get_constant_value(constant_code => 'IDD_ENABLED_' . $args{enabled});
	
	my $command = $class->{_ui}->formulate_command(code => $cmd_code,
						       volume => $volume,
						       vserver => $args{vserver},
						       modifier => $enabled_string,);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
							    boot_mode => $class->{_boot_mode},
							    cmd => $command,
							    node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
	
	return $class->{_ui}->process_output_string(code => $cmd_code,
						    output => $output,
						    expected_rc => $args{expected},
						    log_message => "Configure IDD ($enabled_string) for $volume",);
}


## @name ::is_IDD_enabled
## @summary Checks if IDD is enabled for the volume.
##
## @description
## Checks if IDD is enabled for the volume.
##
## @return Hash with rc, and output.
##
## @arg volume Optional: Name of volume for which to check IDD.
## Defaults to the volume for which the object was created.
##
## @arg vserver Optional: Vserver name (for C-mode only).
## Defaults to the vserver for which the object was created.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Legal values: C_RC_IDD_ON, C_RC_IDD_OFF. Default: C_RC_IDD_ON.
##
## @examples
## $compress->is_IDD_enabled();
## $compress->is_IDD_enabled(expected => C_RC_IDD_OFF);
##
sub is_IDD_enabled
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			volume => {type => SCALAR, optional => 1, default => $class->{vol_name}},
			vserver => { type => SCALAR|UNDEF, optional => 1, default => $class->{vserver} },
			expected => {type => SCALAR, optional => 1, default => C_RC_IDD_ON},
		});
	
	my $cmd_code = C_CMD_IS_IDD_ENABLED;
	my $volume = $args{volume};
	
	my $command = $class->{_ui}->formulate_command(code => $cmd_code,
						       volume => $volume,
						       vserver => $args{vserver},);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
							    boot_mode => $class->{_boot_mode},
							    cmd => $command,
							    node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
	
	return $class->{_ui}->process_output_string(code => $cmd_code,
						    output => $output,
						    expected_rc => $args{expected},
						    log_message => "Verify if IDD is enabled for $volume",);
}


################################################################
# Internal subroutines
################################################################

sub _update_volume_name
{
        my $class = shift;
	my (%args) = validate( @_,
	{
		new_name => { type => SCALAR, },
	} );
        
        $class->{vol_name} = $args{new_name};
        
        logcomment("CompressionUtilities: class volume was renamed to: " . $class->{vol_name}) if ($class->{_debug});
}


sub _update_vfiler
{
	my $class = shift;
	my $new_vfiler = shift;
	
	$class->{_vfiler} = $new_vfiler;
}


sub _update_conn
{
	my $class = shift;
	my $new_conn = shift;
	
	# This assumes that the new filer is running the same version 
	if (($class->{_boot_mode} eq C_ONTAP_7MODE) && !($new_conn->isa("NOTE::Filer")))
        {
                logresult('script', "For 7-mode, the conn argument is required to be a NOTE\:\:Filer object");
        }
        elsif (($class->{_boot_mode} eq C_ONTAP_CMODE) && !($new_conn->isa("NOTE::GX::Node")))
        {
                logresult('script', "For 10-mode, the conn argument is required to be a NOTE\:\:GX\:\:Node object");
        }
	$class->{_conn} = $new_conn;
}


sub _triage_fingerprints
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			fingerprint_hash => { type => HASHREF|UNDEF, },
			inode => { type => SCALAR|UNDEF, optional => 1, },
			fbns => { type => SCALAR, },
			remove_stale => { type => BOOLEAN, default => FALSE, },
			hash_meaning => { type => SCALAR, },
		} );

	logcomment("CompressionUtilities: _triage_fingerprints input: " . Dumper(%args) ) if ($class->{_debug});

	my $fingerprint_hash = $args{fingerprint_hash};
	my $inode = $args{inode};
	my $fbns = $args{fbns};
	my $remove_stale = $args{remove_stale};
	my $hash_meaning = $args{hash_meaning};
	
	if (!defined($fingerprint_hash))
	{
		return (count => 0);
	}
	
	my %triaged_hash = ();
	my $count = 0;
	my $triage_by_fbn = $class->{_ui}->get_constant_value(constant_code => 'TRIAGE_FP_BY_FBN');
	
	foreach my $fingerprint (keys %$fingerprint_hash)
	{
		my $fingerprint_info = $fingerprint_hash->{$fingerprint};
		my $non_stale_entry_index = 0;
		my $max_cp_count = -1;
		my $add_non_stale_entry_to_triaged = TRUE;
		
		foreach my $fingerprint_entry (keys %$fingerprint_info)
		{
			my $add_to_triaged_hash = TRUE;
			my $entry_info = $fingerprint_info->{$fingerprint_entry};
			logcomment("CompressionUtilities: triaging $hash_meaning:\n" .
				   "fingerprint value: $fingerprint: entry: $fingerprint_entry\n" . Dumper($entry_info) ) if ($class->{_debug});
			
			if (defined($inode) && ($entry_info->{inode} != $inode))
			{
				logcomment("CompressionUtilities: triaging $hash_meaning:\n" .
					   "Ignoring entry for inode " . $entry_info->{inode} . " (only looking for inode $inode)") if ($class->{_debug});
				$add_to_triaged_hash = FALSE;
			}
			
			if (($fbns =~ /^\d+$/) && ($entry_info->{fbn} != $fbns))
			{
				logcomment("CompressionUtilities: triaging $hash_meaning:\n" .
					   "Ignoring entry for fbn " . $entry_info->{fbn} . " (only looking for fbn $fbns)") if ($class->{_debug});
				$add_to_triaged_hash = FALSE;
			}
			elsif($fbns =~ /^(\d+)-(\d+)$/)
			{
				my $min_fbn = $1;
				my $max_fbn = $2;
				my $current_fbn = $entry_info->{fbn};
				
				if (($current_fbn < $min_fbn) || ($current_fbn > $max_fbn))
				{
					logcomment("CompressionUtilities: triaging $hash_meaning:\n" .
						   "Ignoring entry for fbn " . $entry_info->{fbn} . " (only looking for fbn range $min_fbn-$max_fbn)") if ($class->{_debug});
					$add_to_triaged_hash = FALSE;
				}
			}
			
			# If we need to triage by fbns, consider in here all the fingerprints,
			# except for the ones that are not for the required inode or fbn
			if (($remove_stale == TRUE) && ($triage_by_fbn != TRUE))
			{
				# Just keep track of the non-stale entry
				if ($entry_info->{cp_count} > $max_cp_count)
				{
					$non_stale_entry_index = $fingerprint_entry;
					$max_cp_count = $entry_info->{cp_count};
					$add_non_stale_entry_to_triaged = $add_to_triaged_hash;
				}
			}
			elsif ($add_to_triaged_hash == TRUE)
			{
				$triaged_hash{triaged}{$fingerprint}{$fingerprint_entry} = $entry_info;
			}
		}
		
		if (($remove_stale == TRUE) && ($add_non_stale_entry_to_triaged == TRUE))
		{
			logcomment("CompressionUtilities: triaging $hash_meaning:\n" .
				   "Removal of stale fingerprints required. Adding entry to return hash:\n" .
				   "Fingerprint $fingerprint\n" .
				   "Cp count: $max_cp_count\n" .
				   "Hash index: $non_stale_entry_index") if ($class->{_debug});
			$triaged_hash{triaged}{$fingerprint}{0} = $fingerprint_info->{$non_stale_entry_index};
		}
		
	}
	
	if (($remove_stale == TRUE) && ($triage_by_fbn == TRUE))
	{
		# Need to process this case differently because different fbns may have the same fingerprint
		# Also, the fingerprint of the stale and non-stale entries may differ
		# First process the triaged hash (which at this point may contain stale entries) to group information by fbn
		my %return_hash = ();
		my $fingerprints_grouped_by_fbn = {};
		my $triaged_fps = $triaged_hash{triaged};
		
		foreach my $triaged_fp (keys(%$triaged_fps))
		{
			my $triaged_fp_info = $triaged_fps->{$triaged_fp};
			foreach my $triaged_fp_entry (keys(%$triaged_fp_info))
			{
				my $triaged_entry_info = $triaged_fp_info->{$triaged_fp_entry};
				my $triaged_fbn = $triaged_entry_info->{fbn};
				$fingerprints_grouped_by_fbn->{$triaged_fbn}->{$triaged_fp}->{$triaged_fp_entry}->{info} = $triaged_entry_info;
				$fingerprints_grouped_by_fbn->{$triaged_fbn}->{$triaged_fp}->{$triaged_fp_entry}->{fingerprint} = $triaged_fp;
				$fingerprints_grouped_by_fbn->{$triaged_fbn}->{$triaged_fp}->{$triaged_fp_entry}->{cp_count} = $triaged_entry_info->{cp_count};
				$fingerprints_grouped_by_fbn->{$triaged_fbn}->{$triaged_fp}->{$triaged_fp_entry}->{fp_entry} = $triaged_fp_entry;
			}
		}
		
		# Now we have all the fingerprints grouped by fbns
		foreach my $fbn_to_consider (keys(%$fingerprints_grouped_by_fbn))
		{
			my $fbns_found = $fingerprints_grouped_by_fbn->{$fbn_to_consider};
			# For each fbn, filter the one with the latest cp count
			my $max_cp_count = -1;
			my $max_cp_count_fp_info = {};
			foreach my $fbn_fingerprint (keys(%$fbns_found))
			{
				my $fbn_fingerprint_entries = $fbns_found->{$fbn_fingerprint};
				foreach my $fbn_fingerprint_entry (keys(%$fbn_fingerprint_entries))
				{
					my $fbn_fingerprint_info = $fbn_fingerprint_entries->{$fbn_fingerprint_entry};
					if ($fbn_fingerprint_info->{cp_count} > $max_cp_count)
					{
						$max_cp_count_fp_info->{fingerprint} = $fbn_fingerprint_info->{fingerprint};
						$max_cp_count_fp_info->{fp_entry} = $fbn_fingerprint_info->{fp_entry};
						$max_cp_count_fp_info->{info} = $fbn_fingerprint_info->{info};
					}
				}
			}
			
			# Now add the filtered fbn to the return hash
			my $max_cp_count_fp_val = $max_cp_count_fp_info->{fingerprint};
			my $max_cp_count_fp_entry = $max_cp_count_fp_info->{fp_entry};
			$return_hash{triaged}{$max_cp_count_fp_val}{$max_cp_count_fp_entry} = $max_cp_count_fp_info->{info};
			
			$count++;
		}
		
		$return_hash{count} = $count;
		
		return %return_hash;
	}
	else
	{
		my $triaged = $triaged_hash{triaged};
		if (defined($triaged_hash{triaged}))
		{
			$count = scalar keys (%$triaged);
		}
		$triaged_hash{count} = $count;
		
		return %triaged_hash;
	}
}


sub _disable_compression_on_one_volume
{
	my $class = shift;
	my (%args) 	= validate( @_,
		{
			volume => { type => SCALAR, optional => 1, default => $class->{vol_name} },
                        vserver => { type => SCALAR|UNDEF, optional => 1, default => $class->{vserver} },
			expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
			check_if_already_set => { type => BOOLEAN, optional => 1, default => C_DFLT_CHECK_OPTION_FIRST },
			inline => { type => BOOLEAN },
		} );
	
	logcomment("CompressionUtilities: _disable_compression_on_one_volume input: " . Dumper(%args) ) if ($class->{_debug});
	
	my $volume = $args{volume};
        my $vserver = $args{vserver};
	my $expected = $args{expected};
	my $check_if_already_set = $args{check_if_already_set};
	my $inline = $args{inline};
	my $wanted_setting = ($inline == TRUE) ? _C_RC_COMPR_ON_NO_INLINE : _C_RC_COMPR_OFF;
	
	if ($check_if_already_set == TRUE)
	{
		my %res = $class->is_compression_enabled(volume => $volume,
                                                         vserver => $vserver,
                                                         expected => C_RC_FAILURE_NON_FATAL);
		
		logcomment("Compression required to be disabled to: $wanted_setting\n" .
			   "Current setting: " . $res{rc}) if ($class->{_debug});
		
		if ($res{rc} == $wanted_setting)
		{
			logcomment( "CompressionUtilities: compression is already disabled\n") if ($class->{_debug});
			return %res;
		}
	}
	
	
	my $cmd_code = C_CMD_DISABLE_COMPRESSION;
        
        if ((defined($class->{_supportability})) && ($expected == C_RC_SUCCESS))
        {
                $class->{_supportability}->start_tailing(events => [{code => C_EVENT_COMPRESSION_DISABLED,
                                                                     options => {inline => $inline,},},
                                                                    ],
                                                         required_identifier => $volume,
                                                         );
        }
	
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       vserver => $vserver,
                                                       volume => $volume,
						       options => {disable_inline_only => $inline,},);
	
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
	
	my %disable = $class->{_ui}->process_output_string(code => $cmd_code,
							   output => $output,
							   expected_rc => C_RC_FAILURE_NON_FATAL,
							   log_message => "Disable compression for volume $volume",);
	
	if (($disable{rc} == C_RC_SUCCESS) && ($expected == C_RC_SUCCESS))
	{
                if (defined($class->{_supportability}))
                {
                        $class->{_supportability}->stop_and_match();
                }
		logcomment("Compression was disabled without errors. Double-checking...");
		my $is_inline = ($inline == TRUE) ? C_RC_FAILURE : C_RC_SUCCESS;
		return $class->is_compression_enabled(volume => $volume,
                                                      vserver => $vserver,
                                                      expected => $wanted_setting,
                                                      inline => $is_inline);
	}
	
	return $class->{_ui}->check_for_expected_rc(obtained_rc => $disable{rc},
						    expected_rc => $args{expected},
						    log_message => "Disable compression for volume $volume",
						    output      => $output);
}



1;
