#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Dataset.pm#1 $
#
## @summary Compression Dataset routines 
##
## @description
##  This library provides all the capabilities
##	required for create/modify/delete Compression dataset
##  this routine makes the dataset transparent to the scripts
##
## @synopsis
##
## use Compression::Dataset;
## use vars qw($FILER $VOLUME $CLIENT $ENABLE $AGGREGATE); 
## 
## BEGIN { require "Compression/setup_core.pl"; }
## 
## my $base_vol = Compression::Compress->setup(
## 				filer		=> { name => $FILER },
## 				aggregate	=> { name => $AGGREGATE },
## );
##
## my $dataset = $base_vol->{dataset};
##
## @examples 
##	%res = $dataset->create( 
##		files => [
##					{ name => "f1", size => ( 4 * KB) },
##				],
##	);
##
##	%res = $dataset->modify( 
##		files => [
##					{ name => "f1", offset => (4 * KB), len => (4* KB), pattern=>'b' },
##					{ name => "f1", offset => (8 * KB), len => (4* KB), pattern=>'c' },
##				],
##	);
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com, kozel@netapp.com
## @status private (in-progress)


package Compression::Dataset;

use strict;
use warnings;

use Tharn;
use Compression::Dataset::Common;
use Compression::Dataset::CIFS;
use Compression::Dataset::NFS;
use Module::Common;
use Data::Dumper;
use SPARSE::Resource::Volume;
use SPARSE::Resource::Filer;
use SPARSE::Util::FileIO;
use SPARSE::Util::Logger;
use SPARSE::Util::Utilities;
use FS::Util::Utilities qw(file_reserve dumpblock);
use Dense::Utils;
use Dense::TestData;
use cat qw(do_shell);
use File::Copy;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Compression::Constants qw(:all);
use Compression::Utils qw(_send_cmd_to_filer);
use File::Compare;
use File::Basename;
use Carp qw(croak);
use FileManip;
use filer;
use Compression::Dataset::Constants qw(:all);
use Digest::SHA qw(sha1);
use SnapCmds;


BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}


## @name ::new
##
## @summary perform dataset initialization
##
## @description
## Perform all the basic dataset initialization required here
##
## @return Compression::Dataset object
##
## @arg volume Required: NOTE Volume object
## @arg client Optional: name of the client
## @arg debug Optional: logs additional debug messages specific to this module
##
## @examples
## 	$self->{dataset}  = Compression::Dataset->new(volume => $self->{volume}->{_vol},
##						      client => $self->{_client}->{name} );
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
##

sub new
{
	my $class 	= shift;
	my (%args) = validate (@_, {volume    => { type=>OBJECT, },
				    vserver   => { type=>SCALAR|UNDEF, optional => 1 },
				    client    => { type=>SCALAR, optional => 1 },
				    debug     => { type=>SCALAR, optional => 1, default=>FALSE },
				    skip_mount=> { type=>BOOLEAN, optional => 1, default=>FALSE },
				    ui        => {type => OBJECT, isa => "Compression::UI"},
				    boot_mode => {type => SCALAR, optional => 1, default => C_ONTAP_BOOT_MODE},
				    has_debug_kernel => {type => SCALAR, optional => 1, default => TRUE}, # So that scripts can continue
				    vfiler_mountpoint => {type => SCALAR|UNDEF, optional => 1,}, 
		} );	
	my $self = {};
	$self->{ID} = $$;
	$self->{Description} = "Dataset create/modify";
	$self->{_vserver} = $args{vserver};
	$self->{client} = ( defined $args{client} )? $args{client} : "localhost";
	$self->{client_host} = host($self->{client});
	$self->{client_conn} = connect("#rsh");
	$self->{client_hosttype} = $self->{client_host}->hosttype;
	$self->{_ui} = $args{ui};
	$self->{_debug} = $args{debug};
	$self->{ds_type} = C_DFLT_COMPRESSION_DATASET;
	$self->{_has_debug_kernel} = $args{has_debug_kernel};	
	$self->{_boot_mode} = $args{boot_mode};
	$self->{_skip_mount} = $args{skip_mount};
		
	if ($self->{client_hosttype} eq 'windows') {
		$self->{_c_dataset} = Compression::Dataset::CIFS->new(%args);
	} else {
		$self->{_c_dataset} = Compression::Dataset::NFS->new(%args);
	}

	$self->{files} = {};
	$self->{_removed_files} = {};
	$self->{sfsr_will_restore_file_as_compressed} = $self->{_ui}->get_constant_value( constant_code => 'SFSR_WILL_RESTORE_FILE_AS_COMPRESSED' );
	$self->{sfsr_uses_clone_engine} = $self->{_ui}->get_constant_value( constant_code => 'SFSR_USES_CLONE_ENGINE' );
	

	bless $self, $class;

	if ( !$self->{_skip_mount} ){ 
		# Default testdir path
		$self->{testdir_path} = $self->get_mountpoint();
	} else {
		$self->{testdir_path} = '';
	}
	logcomment("Dataset Initialized with ". Dumper($self) ) if ( $self->{_debug} );
	return $self;

}


## @name ::create
##
## @summary create the dataset or set of files on the volume
##
## @description
## Create set of files or specific dataset for the volume
##
## @return hash of status and output
##
##
## @arg	files		Required: Array of hashrefs ( one for each file )
##		files => [
##					{ 
##						name	=>	"f1", 
##						offset	=>	0, #default=0
##						size	=> ( 4 * KB), 
##						type	=> 'uncompressable|compressable|existing_data|write_bytes|blocksondisk|large', #default='write_bytes'
##						compression	=> 0,
##						blocksondisk	=> 0,
##						pattern	=> 'x', #default='a'
##						data	=>	$data, #only use if type=existing_data
##					},
##				]
##
## @arg	client		Optional: Client
## @arg	path		Optional: Defaults to root of volume
## @arg	testdir		Optional: Test directory
## @arg	ds_type		Optional: Dataset Type
## @arg compression	Optional: File compressability (0 - 100)
## @arg blocksondisk	Optional: Number of blocks per compression group written on disk
## @arg	offset		Optional: Offset to for writing files ( overriden by offset for individual file, default is 0 ).
## @arg	len			Optional: Length of data to be written ( overriden by size of individual file )
## @arg	pattern		Optional: Default 'a'
## @arg	background	Optional: Run in background ( TRUE|FALSE )
## @arg dont_die_on_error	Optional: If volume is full, script won't die ( TRUE|FALSE ).  Setting this to TRUE has same effect as 
##		setting expected_result to C_RC_FAILURE_NON_FATAL.
## @arg expected_result	Optional: Croak if not expected ( C_RC_SUCCESS(default)|C_RC_INSUFFICIENT_SPACE|C_RC_FAILURE_NON_FATAL). If 
##						dont_die_on_error is TRUE, expected_result defaults to C_RC_FAILURE_NON_FATAL )
##
## @examples
##
##
## %res = $dataset->create( 
##		files => [
##					{ 
##						name => "f1", 
##						size => GB, 
##						type=>'uncompressable', #default='write_bytes' 
##					},
##				],
##	);
##	
##
## %res = $dataset->create( 
##		files => [
##					{ name => "f1", offset=>0, size => ( 4 * KB) },
##				],
##		background => TRUE,
##	);
## $dataset->wait_finish( filename => 'f1' );
##
##
## my $data = $dataset->get_uncompressable_data(size=>4 * KB);
## %res = $dataset->create( 
##		files => [
##					{ name => "f1", data=>$data, type=>'existing_data' },
##				],
##	);
## 
##	Notes:
##	type can be:
##		1. write_bytes		-(default) repeats a pattern
##		2. existing_data	-need to call get_uncompressable_data() first if using this option
##		3. uncompressable	-create uncompressable data
##		4. compressable		-creates a file with a certain percentage of compressable blocks
##		5. blocksondisk		-each compression group in a file compresses to a certain number of blocks
##		4. large			-originally for large files ( will be depreciated )
##	pattern needs to be specified if type is write_bytes or large ( default is 'a' )
##	data only should be used if type=existing_data
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com, kozel@netapp.com
##

sub create
{
	my $class = shift;
	return $class->{_c_dataset}->create(@_);
}




## @name ::modify
##
## @summary modify the dataset or set of files on the volume
##
## @description
## Modify set of files or specific dataset for the volume
##
## @return hash of status and output
##
##
## @arg	files		Required: Array of hashrefs ( one for each file )
##		files => [
##					{ 
##						name	=> "f1", 
##						offset	=> ( 4 * KB),
##						len		=> ( 8 * KB), 
##						type	=> 'uncompressable|existing_data|write_bytes', #default='write_bytes'
##						pattern	=> 'x', #default='a'
##						data	=>	$data, #only use if type=existing_data
##					},
##				]
## @arg	client		Optional: Client
## @arg	path		Optional: Defaults to root of volume
## @arg	ds_type		Optional: Dataset Type
## @arg	offset		Optional: Offset to for writing files ( overriden by offset for individual file, default is 0 ).
## @arg	len			Optional: Length of data to be written ( overriden by len of individual file )
## @arg	pattern		Optional: Default 'a'
## @arg	background	Optional: Run in background ( TRUE|FALSE )
## @arg dont_die_on_error	Optional: If volume is full, script won't die ( TRUE|FALSE ).  Setting this to TRUE has same effect as 
##		setting expected_result to C_RC_FAILURE_NON_FATAL.
## @arg expected_result	Optional: Croak if not expected ( C_RC_SUCCESS(default)|C_RC_INSUFFICIENT_SPACE|C_RC_FAILURE_NON_FATAL). If 
##						dont_die_on_error is TRUE, expected_result defaults to C_RC_FAILURE_NON_FATAL )
##
## @examples
##
## %res = $dataset->modify( 
##		files => [
##					{ name => "f1", offset=>( 4 * KB), len => ( 8 * KB), },
##				],
##		background => TRUE,
##	);
## $dataset->wait_finish( filename => 'f1' );
##
##
## %res = $dataset->modify( 
##		files => [
##					{ name => "f1", offset=>( 4 * KB), len => ( 8 * KB), type=>'uncompressable' },
##				],
##	);
##
##
## my $data = $dataset->get_uncompressable_data(size=>4 * KB);
## %res = $dataset->modify( 
##		files => [
##					{ name => "f1", data=>$data, type=>'existing_data' },
##				],
##	);
## 
##	Notes:
##	type can be:
##		1. write_bytes		-(default) repeats a pattern
##		2. existing_data	-need to call get_uncompressable_data() first if using this option
##		3. uncompressable	-create uncompressable data
##		4. large			-originally for large files ( will be depreciated )
##		5. holey			-add holey blocks
##	
##	data only should be used if type=existing_data
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com, kozel@netapp.com
##
sub modify
{
	my $class = shift;
	return $class->{_c_dataset}->modify(@_);
}


## @name ::delete_file
##
## @summary delete the file
##
## @description
## delete the file
##
## @return hash of status
##
## @arg filename	Required: name of file to delete
##
## @examples
##  $dataset->delete_file( filename=>'f1' );
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
##
sub delete_file
{
        my $class = shift;
        return $class->{_c_dataset}->delete_file(@_);
}

## @name ::compare_datasets
##
## @summary compare the files on volume vs baseline dataset
##
## @description
## file compare on the volume with baseline dataset
##
## @return
## %res;
##	$res{'rc'}		= 'C_RC_SUCCESS';
##
## @arg dataset_2	Required: 2nd dataset to compare
## @arg expected_result Optional: EQUAL or NOT_EQUAL. Default is EQUAL. Will croak if result is not expected.
## @arg ignore_snapshots Optional: don't compare snapshots.  Default is TRUE.
##
## @examples
##	%res = $dataset_1->compare_datasets( 
##		dataset_2 => $dataset_2,
##	);
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
##
sub compare_datasets
{
        my $class = shift;
        return $class->{_c_dataset}->compare_datasets(@_);
}


## @name ::display_file_contents
##
## @summary display the contents of a file
##
## @description
## display contents of file.  This is more just a utility for 
## debugging/visual inspection.
##
## @return none
##
## @arg	Required:	filename
##
## @examples 
##	$dataset->display_file_contents( filename=>'f1' );
##
sub display_file_contents
{
        my $class = shift;
        return $class->{_c_dataset}->display_file_contents(@_);
}

## @name ::get_path
##
## @summary return path to file
##
## @description
## return path to file
##
## @return full path to the file
##
## @arg	filename Required: name of file.
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset.
##
## @examples
##  my $filepath = $dataset->get_path( filename=>'f1' );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_path
{
        my $class = shift;
        return $class->{_c_dataset}->get_path(@_);		
}

## @name ::get_mountpoint
##
## @summary return mount point of volume
##
## @description
## wrapper around $vol->mountpoint
##
## @return $vol->mountpoint
##
## @examples
##  my $mountpoint = $dataset->get_mountpoint;
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_mountpoint
{
        my $class = shift;
        return $class->{_c_dataset}->get_mountpoint(@_);
}


## @name ::get_filer_path
##
## @summary return path to file on filer
##
## @description
## return path to file on filer
##
## @return the filer's path to the file ( e.g. /vol/vol1/f1 )
##
## @arg	filename Required: name of file
##
## @examples
##  my $filer_path = $dataset->get_filer_path( filename=>'f1' );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_filer_path
{
        my $class = shift;
        return $class->{_c_dataset}->get_filer_path(@_);
}


## @name ::replace_file
##
## @summary replace file with some specified file
##
## @description
## replace file with some specified file
##
## @return hash with success code
##
## @arg filename	Required:	name of file to replace
## @arg new_file	Required:	path to new file
##
## @example
##  $dataset->replace_file( filename=>'f1', new_file=>'path_to_new_file' );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub replace_file
{
        my $class = shift;
        return $class->{_c_dataset}->replace_file(@_);
}

## @name ::compare_files
##
## @summary compares 2 files
##
## @description
## compares 2 files.  Croaks if not expected result.
##
## @return hash of status and output
##	$res{'result'} = EQUAL or NOT_EQUAL
## 
## @arg file_1	Required: path to first file
## @arg file_2	Required: path to second file
## @expected_result	Optional: EQUAL|NOT_EQUAL ( default is 'EQUAL' )
##
## @example
##
##	%res = $dataset->compare_files( 
##						file_1=> $dataset_1->get_path( filename=>'f1' ), 
##						file_2=> $dataset_2->get_path( filename=>'f1' ),
##						expected_result=>EQUAL|NOT_EQUAL #default:EQUAL
##						);
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub compare_files
{
        my $class = shift;
        return $class->{_c_dataset}->compare_files(@_);
}


## @name ::copy_file
##
## @summary copies a file
##
## @description
## Copies a file
##
## @return hash with status
## 
## @arg source	Required: path to source file
## @arg dest	Required: path to destination file
## @arg	new_dataset	Optional: use this only when copying to a new dataset ( i.e. new volume )
## @arg background	Optional: default is FALSE ( if TRUE, use with wait_finish )
##
## @examples
##
##	%res = $dataset->copy_file( 
##						source		=> $dataset_1->get_path( filename=>'f1' ), 
##						dest		=> $destination_file
##						new_dataset	=> $dataset_2  #optional: only use if copying file to different volume
##  );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub copy_file
{
        my $class = shift;
        return $class->{_c_dataset}->copy_file(@_);
}


## @name ::move_file
##
## @summary move a file
##
## @description
## Moves a file
##
## @return hash of status
## 
## @arg source	Required: path to source file
## @arg dest	Required: path to destination file
## @arg	new_dataset	Optional: use this only when copying to a new dataset ( i.e. new volume )
## @arg background	Optional: default is FALSE ( if TRUE, use with wait_finish )
##
## @examples
##
##	%res = $dataset->move_file( 
##						source		=> $dataset_1->get_path( filename=>'f1' ), 
##						dest		=> $destination_file
##						new_dataset	=> $dataset_2  #optional: only use if moving file to different volume
##  );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub move_file
{
        my $class = shift;
        return $class->{_c_dataset}->move_file(@_);
}


## @name ::truncate_file
##
## @summary truncate file to specified size
##
## @description
## truncate file to specified size
##
## @return hash of status
##
## @arg	filename	Required: name of file to truncate
## @arg size		Required: size to truncate file to
##
## @examples
##  $dataset->truncate_file( filename=>'f1', size=>1024 );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub truncate_file
{
	my $class = shift;
	return $class->{_c_dataset}->truncate_file(@_);
}


## @name ::get_file_size
##
## @summary get size of file in bytes
##
## @description
## get size of file in bytes
##
## @return size of file in bytes
##
## @arg filename	Required: name of file
## @arg snapshot	Optional: name of snapshot ( if file is in snapshot )
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset
## 
## @examples
##  my $file_size = $dataset->get_file_size( filename=>'f1');
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_file_size
{
        my $class = shift;
        return $class->{_c_dataset}->get_file_size(@_);
}

## @name ::get_timestamp
##
## @summary get timestamp of file
##
## @description
## get timestamp of file
##
## @return timestamp (last modify time) of file
##
## @arg filename	Required: name of file
## @arg snapshot	Optional: name of snapshot ( if file is in snapshot )
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset
## 
## @examples
##  my $timestamp = $dataset->get_timestamp( filename=>'f1');
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_timestamp
{
        my $class = shift;
        return $class->{_c_dataset}->get_timestamp(@_);
}


## @name ::get_uncompressable_data
##
## @summary get uncompressable data
##
## @description create data that is uncompressable.  Only use this when you need to add
##  the exact same data to different files. Warning: out-of-memory error on large files.
##
## @return data
##
## @arg size	Required:	size of data to create
## 
## @examples
##  my $data = $dataset->get_uncompressable_data(size=>4 * KB);
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com, kozel@netapp.com
##
sub get_uncompressable_data
{
        my $class = shift;
        return $class->{_c_dataset}->get_uncompressable_data(@_);
}

## @name ::wait_finish
##
## @summary wait for background process to finish
##
## @description
## wait for background process to finish from create,
##  modify, copy_file, or move_file
##
## @return hash of status
## 
## @examples
##  $dataset->wait_finish( filename=>'f1' );
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub wait_finish
{
        my $class = shift;
        return $class->{_c_dataset}->wait_finish(@_);
}


## @name ::is_alive
##
## @summary Checks if a background process is alive.
##
## @description
## Checks if a background process is alive,
## modify, copy_file, or move_file
##
## @return Boolean value: 0/1
## 
## @examples
##  if ($dataset->is_alive(filename=>'f1')) {logcomment("The background process is alive");}
## 	
##
## @author dl-compression-qa@netapp.com
##
sub is_alive
{
        my $class = shift;
        return $class->{_c_dataset}->is_alive(@_);
}


## @name ::restore_file_info
##
## @summary Restore internal info of a file that was previously removed.
##
## @description
## Restore internal info of a file that was previously removed.
## To be used after operations that cause files to reappear (for example: snapshot restore).
##
## @arg files Required: Arrayref of files for which to restore info.
## 
## @examples
##  $dataset->restore_file_info(files => ['f1', 'f2', 'f3',] );
##
## @author dl-compression-qa@netapp.com
##
sub restore_file_info
{
	my $class = shift;
	my (%args) = validate (@_, {
		files => { type=>ARRAYREF, },
	} );
	
	my $files = $args{files};
	
	logcomment("Restoring file info for files:\n" . Dumper($files)) if ($class->{_debug});
	foreach my $filename (@$files) {
		$class->{_c_dataset}->{_common}->_re_add_file(filename => $filename);
	}
	
	return (rc => C_RC_SUCCESS);
}


## @name ::generate_path_for_attempt_file_clone
##
## @summary Generates path for attempted file clone destination.
##
## @description
## Generates path for attempted file clone destination.
## Also, it generates all the internal dataset information needed for the clone.
## This method must be completed with a call to restore_file_clone_info after the successful file clone operation
##
## @arg file_name Required: Name of the file-clone source file.
##
## @arg clone_name Required: Name for the new file clone destination.
##
## @arg allow_removed Optional: flag to enable operation to be permitted for file-clone source files that were previosly removed by dataset.
## 
## @examples
##  my $clone_path = $dataset->generate_path_for_attempt_file_clone(file_name => 'my_file', clone_name => 'my_clone');
##
## @author dl-compression-qa@netapp.com
##
sub generate_path_for_attempt_file_clone
{
	my $class = shift;
	my (%args) = validate (@_, {
		file_name => {type => SCALAR,},
		allow_removed => {type => SCALAR, default => FALSE,},
		clone_name => {type => SCALAR,},
	} );
	
	my $file_name = $args{file_name};
	my $clone_name = $args{clone_name};
	
	return $class->{_c_dataset}->{_common}->_add_attempt_file_clone_dest(filename => $file_name,
									     clonename => $clone_name,
									     allow_removed => $args{allow_removed},);
}


## @name ::restore_file_clone_info
##
## @summary Restores internal dataset information for the file clone destination.
##
## @description
## Restores internal dataset information for the file clone destination.
## This method must be preceeded by generate_path_for_attempt_file_clone (before starting the file clone operation).
##
## @arg clone_name Required: Name for the file clone destination.
## 
## @examples
## $dataset->restore_file_clone_info(clone_name => 'my_clone');
##
## @author dl-compression-qa@netapp.com
##
sub restore_file_clone_info
{
	my $class = shift;
	my (%args) = validate (@_, {
		clone_name => {type=>SCALAR,},
	} );
	
	my $file_name = $args{file_name};
	my $clone_name = $args{clone_name};
	
	return $class->{_c_dataset}->{_common}->_add_file_clone_dest(clonename => $clone_name);
}


## @name ::get_vbn_hash
##
## @summary get vbn hash for some file
##
## @description
## get vbn hash for some file
##
## @arg filename Required: File name.
## @arg snapshot Optional: Name of the snapshot, if file is in snapshot.
## @arg fbn      Optional: fbn to compare ( or fbn range e.g. 8-16 )
##
## @return hash of status and vbn hash
## 
## @examples
##  my %res = $dataset->get_vbn_hash( filename=>'f1' );
##  my $vbns_info = $res{'result'};
##  logcomment("vbn for fbn 3 is "	. $vbns_info->{3}->{vbn} );
##  logcomment("vvbn for fbn 3 is "	. $vbns_info->{3}->{vvbn} );
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub get_vbn_hash
{
	my $class = shift;
	return $class->{_c_dataset}->{_common}->get_vbn_hash(@_);
}


## @name ::compare_vbn_hashes
##
## @summary Compare the vbn hashes for two files.
##
## @description
## Compare the vbn hashes for two files.
##
## @arg file_1 Required: It can be either a string, representing the name of the first file,
## or a reference to an already obtained vbn hash.
##
## @arg file_2 Required: It can be either a string, representing the name of the second file,
## or a reference to an already obtained vbn hash.
##
## @arg snapshot_1 Optional: Name of the snapshot, if file_1 is in a snapshot.
##
## @arg snapshot_2 Optional: Name of the snapshot, if file_2 is in a snapshot.
##
## @arg expected Optional: EQUAL|NOT_EQUAL. Default: 'EQUAL'.
##
## @arg compare_field Optional: all|vbn|vvbn default is 'all'
##
## @return hash of rc and vbn hashes that were compared.
## 
## @examples
##  $dataset->compare_vbn_hashes(file_1 => $file,
##                               file_2 => $file,
##                               snapshot_2 => $snapshot_name);
##
## @author dl-compression-qa@netapp.com, voicu@netapp.com
##
sub compare_vbn_hashes
{
	my $class = shift;
	return $class->{_c_dataset}->{_common}->compare_vbn_hashes(@_);
}


## @name ::compare_vbns
##
## @summary compare vbns from 2 files
##
## @description
## for 2 files, given an fbn ( or range of fbns )
## this compares the vbn values.   Croaks if not expected result.
##
## @return hash of status
##
## @arg file_1	Required: first file to compare
## @arg file_2	Required: second file to compare
## @arg fbn		Optional: fbn to compare ( or fbn range e.g. 8-16 )
## @arg expected	Optional: EQUAL|NOT_EQUAL default is 'EQUAL'
##
## @examples
##  $dataset->compare_vbns( file_1		=>'f1', 
##							file_2		=>'f2', 
##							fbn			=>'0-31', # or '3'
##							expected	=>'EQUAL', or 'NOT_EQUAL'
##							);
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub compare_vbns
{
	my $class = shift;
	return $class->{_c_dataset}->{_common}->compare_vbns(@_);
}


## @name ::compare_vvbns
##
## @summary compare vvbns from 2 files
##
## @description
## for 2 files, given an fbn ( or range of fbns )
## this compares the vvbn values.  Croaks if not expected result.
##
## @return hash of status
##
## @arg file_1	Required: first file to compare
## @arg file_2	Required: second file to compare
## @arg fbn		Optional: fbn to compare ( or fbn range e.g. 8-16 )
## @arg expected	Optional: EQUAL|NOT_EQUAL default is 'EQUAL'
##
## @examples
##  $dataset->compare_vvbns(file_1		=>'f1', 
##							file_2		=>'f2', 
##							fbn			=>'0-31', # or '3'
##							expected	=>'EQUAL', or 'NOT_EQUAL'
##							);
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub compare_vvbns
{
	my $class = shift;
	return $class->{_c_dataset}->{_common}->compare_vvbns(@_);
}


## @name ::compare_vbns_within_file
##
## @summary compare vbns within a file
##
## @description
## for a file, given a range of fbns,
## this compares the vbn values.   Croaks if not expected result.
##
## @return hash of status
##
## @arg filename        Required: name of file
## @arg fbn             Optional: fbn range e.g. 8-16
## @arg expected        Optional: EQUAL|NOT_EQUAL default is 'EQUAL'
##		EQUAL means every vbn in this range is equal
##		NOT_EQUAL means 1 or more vbn is different than the other
##
## @examples
##  $dataset->compare_vbns_within_file( filename =>'f1', 
##                                      fbn =>'0-31',
##                                      expected =>'EQUAL', or 'NOT_EQUAL'
##                                    );
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub compare_vbns_within_file
{
	my $class = shift;
	return $class->{_c_dataset}->{_common}->compare_vbns_within_file(@_);
}


## @name ::mk_dir
##
## @summary Make a directory in volume
##
## @description
## Make a directory in volume
##
## @return hash of status
## 
## @arg	name	Required: name of directory
## @arg path	Optional: complete path to the directory the folder will be created in
##
## @examples
##
##	%res = $dataset->mkdir(
##							name => 'd1',
##						 );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub mk_dir
{
        my $class = shift;
        return $class->{_c_dataset}->mk_dir(@_);
}

## @name ::dump_file_info
##
## @summary display paths associated with file
##
## @description
## Just for debugging purposes.
##
## @return none
## 
## @example
##		$dataset->dump_file_info()
##
## displays something like:
## 
##		Dumping file info for f5
##		relative_path: xx/f5
##		filer_path: /vol/c1/xx/f5
##		path: /t/172.18.47.30/vol/c1/shared/tcp_v3/xx/f5
##
sub dump_file_info
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->dump_file_info(@_);
}


## @name ::read_file
##
## @summary verify that a file can be read
##
## @description
## This function is to verify that a file can be read.
## Croaks if it can not read file.
##
## @return status
##
## @arg	filename	Required:
## @arg snapshot	Optional: name of snapshot ( if file is in snapshot )
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset
##
## @examples 
##	$dataset->read_file( filename=>'f1', snapshot=>'ss1', allowed_removed=>TRUE );
##
sub read_file
{
        my $class = shift;
        return $class->{_c_dataset}->read_file(@_);
}


## @name ::file_reservation_enable
##
## @summary file reservation enable
##
## @description
## Enable file reservation
##
## @return status
##
## @arg	Required:	filename
##
## @examples 
##	$dataset->file_reservation_enable( filename=>'f1' );
##
sub file_reservation_enable
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->file_reservation_enable(@_);
}

## @name ::file_reservation_disable
##
## @summary file reservation disable
##
## @description
## Disable file reservation.
##
## @return status
##
## @arg	Required:	filename
##
## @examples 
##	$dataset->file_reservation_disable( filename=>'f1' );
##
sub file_reservation_disable
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->file_reservation_disable(@_);
}


## @name ::verify_compressed
##
## @summary verify that file is compressed
##
## @description
## From the output of showvbns, verify that file is compressed.
## Croaks if not as expected.
##
## @return hash of status
##
## @arg filename					Required: file to verify
## @arg fbn							Optional: fbn to compare ( or fbn range e.g. 8-16 )
## @arg expected					Optional: COMPRESSED|NOT_COMPRESSED default is 'COMPRESSED'
## @arg max_allowed_uncompressed	Optional: default is 2
## @arg method						Optional: SHOWVBNS|DUMPBLOCK default is 'SHOWVBNS'
## @arg snapshot					Optional: name of snapshot, default is undef
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset
##
## @examples
##  $dataset->verify_compressed(	filename	=>'f1', 
##									fbn			=>'0-31', # or '3'
##									expected	=>'COMPRESSED', # or 'NOT_COMPRESSED'
##									max_allowed_uncompressed => { type=>SCALAR, default=>2 },
##							);
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub verify_compressed
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->verify_compressed(@_);
}


## @name ::get_block_layout
##
## @summary get block layout
##
## @description
## From the output of showvbns, get pattern of headers, data, trailers 
##
## @return hash reference with block pattern keyed by fbn
##  e.g. $block_layout_1->{
##							0=>'H',
##							1=>'D',
##							2=>'T',
##							3=>'T',
##							4=>'T',
##							5=>'T',
##							6=>'T',
##							7=>'T',
##							
##
## @arg filename	Required: file to verify
## @arg fbn			Optional: fbn to compare ( or fbn range e.g. 8-16 )
##
## @examples
##  $block_pattern = $dataset->get_block_pattern(	filename	=>'f1', 
##													fbn			=>'0-31',
##							);
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub get_block_layout
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->get_block_layout(@_);
}

## @name ::compare_block_layouts
##
## @summary compares 2 block layouts
##
## @description
## This function compares the block layouts from 2 files.
## Use this after get_block_layout ( see examples below ).
## Croaks if not expected result.
##
## @return hash of status and output
##	$res{'result'} = EQUAL or NOT_EQUAL
## 
## @arg block_layout_1	Required: block layout of first file
## @arg block_layout_2	Required: block layout of second file
## @expected_result	Optional: EQUAL|NOT_EQUAL ( default is 'EQUAL' )
##
## @example
##
##	my $block_layout_1 = $dataset_1->get_block_layout(	filename	=>'f1', 
##														fbn			=>'0-7',
##								);
##
##	my $block_layout_2 = $dataset_1->get_block_layout(	filename	=>'f2', 
##														fbn			=>'0-7',
##								);
##
##	%res = $dataset->compare_block_layouts( 
##						block_layout_1=> $block_layout_1, 
##						block_layout_2=> $block_layout_2,
##						expected_result=>EQUAL|NOT_EQUAL #default:EQUAL
##						);
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub compare_block_layouts
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->compare_block_layouts(@_);
}

## @name ::verify_checksums
##
## @summary verifies data integrity by comparing checksums
##
## @description
## computes SHA1 checksums for all compression groups and compares them to the precalculated ones
##
## @return hash of status
##
## @arg files		Optional: hash of the filenames
## @arg path		Optional: path to a file
##
## Note, one or the other must be defined
##
## @examples
##  $dataset->verify_checksums( files =>
##								[	{ name=> 'f1' }
##									{ name=> 'f2' }
##								]);
##
##
## @author kozel@netapp.com, dl-compression-qa@netapp.com
##

sub verify_checksums
{
        my $class = shift;
        return $class->{_c_dataset}->verify_checksums(@_);
}

## @name ::set_debug
##
## @summary turn on or off debug flag
##
## @description
## Turn on or off debug flag.
##
## @return hash of status
##
## @arg debug	Optional: TRUE or FALSE ( default TRUE )
##
## @examples
##  $dataset->set_debug( debug=>TRUE );
##
sub set_debug
{
	my $class 	= shift;
	my (%args) = validate (@_, { debug => { type=>SCALAR, default=>TRUE }} );
	my $debug = $args{debug};
	$class->{_debug} = $debug;
	$class->{_c_dataset}->{_debug} = $debug;
	$class->{_c_dataset}->{_common}->{_debug} = $debug;
	logcomment("debug set to : ". $class->{_debug} );
}

## @name ::set_file_reservation
##
## @summary 
##
## @description
## 
##
## @return hash of status
##
## @arg filename	Required: name of file
## @arg enabled		Optional: TRUE or FALSE (default TRUE )
##
## @examples
##  %res = $src_vol->{dataset}->set_file_reservation(filename=>$FILENAME);
##  %res = $src_vol->{dataset}->set_file_reservation(filename=>$FILENAME, enabled=>FALSE);
sub set_file_reservation
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->set_file_reservation(@_);
}


## @name ::get_file_reservation
##
## @summary  Get the current file reservation setting.
##
## @description Get the current file reservation setting.
##
## @return Hash of status.
##
## @arg filename Required: Name of the file.
##
## @arg allow_removed Optional: Flag to allow setting to be retrieved for a file that was previously removed from this dataset.
## Default value: FALSE.
##
## @arg expected Optional: Expected setting (C_FILE_RESERVATION_ON/C_FILE_RESERVATION_OFF).
## Default value: none.
##
## @examples
## $obj->get_file_reservation(filename => 'my_file',
##                            expected => C_FILE_RESERVATION_ON);
##  
sub get_file_reservation
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->get_file_reservation(@_);
}


## @name ::single_file_snap_restore
##
## @summary 
##	Single file snap restore (SFSR) reverts a file in the active file system to the state contained in a snapshot
##
## @description
##	Single file snap restore (SFSR) reverts a file in the active file system to the state contained in a snapshot 
##
## @return hash of status
##
## @arg filename	Required: name of file
## @arg snapshot	Required: name of snapshot to restore from
## @arg wait_for_completion	Optional: don't return until file is completely restored ( default: TRUE )
## @arg expected Optional: Expected return code ( default: C_RC_SUCCESS )
##
## @examples
##	%res = $src_vol->{dataset}->single_file_snap_restore(filename=>$FILENAME, snapshot=>$SNAPSHOT);
sub single_file_snap_restore
{
        my $class = shift;
        return $class->{_c_dataset}->single_file_snap_restore(@_);
}


## @name ::wait_for_single_file_snap_restore
##
## @summary 
##	Waits until a file is completely restored from a snapshot.
##
## @description
##	Waits until a file is completely restored from a snapshot.
##
## @return hash of status
##
## @arg filename Required: Name of file.
## @arg snapshot Required: Name of the snapshot from which the file is restored.
## @arg timeout Optional: Timeout in seconds to wait for the file to be restored. Default: 300.
##
## @examples
## %res = $src_vol->{dataset}->wait_for_single_file_snap_restore(filename=>$FILENAME, snapshot=>$SNAPSHOT);
##
sub wait_for_single_file_snap_restore
{
        my $class = shift;
        return $class->{_c_dataset}->wait_for_single_file_snap_restore(@_);
}


## @name ::fill_remaining_space
##
## @summary 
##	
##
## @description
##	
##
## @return hash of status
##
## @arg filename	Required: name of file
## @arg leave_space	Optional: amount of available space to leave ( currently to the nearest 4KB, default is zero )
##
## @examples
##	%res = $src_vol->{dataset}->fill_remaining_space(filename=>$FILENAME);
sub fill_remaining_space
{
        my $class = shift;
        return $class->{_c_dataset}->fill_remaining_space(@_);
}


## @name ::get_data_range
##
## @summary get specified range of content from a file
##	
##
## @description get specified range of content from a file
##	
## @return data content
##
## @arg filename	Required: name of file
## @arg leave_space	Optional: amount of available space to leave ( currently to the nearest 4KB, default is zero )
##
## @examples
##			my $data = $class->get_data_range(	path	=> $path,
##												offset	=> $offset,
##												len		=> $len,
##			);
sub get_data_range
{
        my $class = shift;
        return $class->{_c_dataset}->get_data_range(@_);
}


## @name ::verify_block_type
##
## @summary verify block type
##
## @description
## From the output of showvbns, verifies that the type of block is HEADER, TRAILER, or DATA.
## Croaks if not as expected.
##
## @arg filename					Required: file to verify
## @arg fbn							Required: fbn to compare
## @arg expected					Required: HEADER|TRAILER|DATA
## @arg snapshot					Optional: name of snapshot, default is undef
##
## @examples
##  $dataset->verify_block_type(	filename	=>'f1', 
##									fbn			=> 3,
##									expected	=>'HEADER',
##							);
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub verify_block_type
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->verify_block_type(@_);
}


## @name ::get_dumpblock_info
##
## @summary  Get dumpblock info.
##
## @description Wrapper around FS::Utilities::dumpblock.
##
## @return Hash with status and output.
##  e.g. $res{output} => {
##						'pvbn-of-parent' => '35128867',
##						'pvbn' => '281474976710652',
##						'dbn' => '0',
##						'disk' => '(unknown)',
##						'status' => 'ok',
##						'vvbn-of-parent' => '4151',
##						'vvbn' => '281474976710652',
##						'pvbn-of-parent-from-container' => '35128867',
##						'pvbn-from-parent' => '281474976710652',
##						'pvbn-from-container' => '281474976710652'
##
## @arg filename Required: Name of the file.
## @arg fbn Required: fbn of the file.
## @arg allow_removed Optional: Flag to allow setting to be retrieved for a file that was previously removed from this dataset.
##		Default value: FALSE.
## @arg level Optional: Block level of the file being used for dumpblock.
## @arg snapname Optional: Name of snapshot if file is in a snapshot.
##
## @examples
## $obj->get_dumpblock_info( filename => $FILENAME, fbn => '3');
##  
sub get_dumpblock_info
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->get_dumpblock_info(@_);
}

## @name ::modify_all_CGs
##
## @summary  modify all CGs in a file
##
## @description modify a certain range within every CG in a file.  As long as the file has compressed CGs, this
## function will find them and overwrite the specified range in each CG.
##
## @return Hash with status.
##
## @arg filename Required: Name of the file.
## @arg offet Required: offset within each CG
## @arg len	Required: number of bytes to write within each CG
## @arg pattern Required: character to write
## @arg CG_ranges Optional: reference to array of ranges as obtained from get_CG_ranges. If undef,
## this function will call get_CG_ranges to get them.
## @arg estimate_boundries Optional: estimate CG ranges based on C_CG_SIZE ( default FALSE )
## Only set this if you're overwriting parts that are uncompressed (i.e. not real CGs ).
##
## @examples
## $obj->modify_all_CGs(filename=>$FILENAME, offset=>4*KB, len=>2*KB, pattern=>'c' )
##  
sub modify_all_CGs
{
        my $class = shift;
        return $class->{_c_dataset}->modify_all_CGs(@_);
}


## @name ::get_CG_ranges
##
## @summary  get CG ranges
##
## @description get all the CG ranges in a file, using VBNs
##
## @return Hash with status and "CG_ranges" which is an array of ranges in the form 0-7, 8-15, etc.
##
## @arg filename Required: Name of the file.
##
## @examples
## 		%res = $obj->get_CG_ranges( filename=>$filename );
##		my @CG_ranges = @{$res{CG_ranges}};
##  
sub get_CG_ranges
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->get_CG_ranges(@_);
}


## @name ::attach_dataset_info
##
## @summary  attach_dataset_info
##
## @description wrapper around _attach_to_cloned_dataset
##
## @arg src_vol Required: Compression::Compress object from src volume
##
## @examples
## 
sub attach_dataset_info
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->attach_dataset_info(@_);

}

## @name ::get_vbn_constants
##
## @summary get the vbn/vvbn constants
##
## @description
## get the vbn/vvbn constants
##
## @return hash of status and output
##	
## 
## @example
##	%res = $dataset->get_vbn_constants();
##  if (grep(/^$x$/, @{$res{vbn_headers}} ) ) {
##		logcomment("$x is a valid vbn header");
##	}
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_vbn_constants {
	my $class = shift;
	return $class->{_c_dataset}->get_vbn_constants(@_);
}

## @name ::ls
##
## @summary ls
##
## @description
## ls
##
## @example
##  $dataset->ls();
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub ls
{
        my $class = shift;
        return $class->{_c_dataset}->ls(@_);
}


## @name ::get_file_atime
##
## @summary Gets the atime of the file (on the client).
##
## @description
## Gets the atime of the file (on the client). 
##
## @return Hash with rc, atime (# of seconds since the start of the epoch), gmt_atime
##
## @arg filename Required: Name of file.
##
## @example
## $dataset->get_file_atime(filename => 'my_file');
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_file_atime
{
        my $class = shift;
        return $class->{_c_dataset}->get_file_atime(@_);
}


## @name ::set_file_atime
##
## @summary Sets the atime of the file (on the client).
##
## @description
## Sets the atime of the file (on the client).
##
## @arg filename Required: Name of file.
##
## @arg new_atime Required: New atime to set (# of seconds since the start of the epoch).
##
## @return The return of the get_file_atime method, after the new atime has been set.
##
## @example
## my %res = $compress->{dataset}->get_file_atime(filename => 'my_file');
## my $new_atime = $res{atime} + 30;
## $dataset->set_file_atime(filename => 'my_file', $new_atime);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub set_file_atime
{
        my $class = shift;
        return $class->{_c_dataset}->set_file_atime(@_);
}


## @name ::verify_file_is_committed
##
## @summary Verify that the file is committed.
##
## @description
## Verify that the file is committed.
##
## @arg filename Required: Name of file.
##
## @example
## $compress->{dataset}->verify_file_is_committed(filename => 'my_file');
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub verify_file_is_committed
{
	my $class = shift;
	my (%args) = validate (@_, { filename => { type=>SCALAR },} );
	
	# TODO: for now just verify that the file cannot be deleted
	# Add chmod
	return $class->{_c_dataset}->delete_file(filename => $args{filename},
						 expected => C_RC_READ_ONLY_FILESYSTEM);
}


## @name ::get_file_inode_num
##
## @summary Gets the file inode num.
##
## @description
## Gets the file inode num.
##
## @arg filename Required: File name.
##
## @return File inode num..
##
## @example
##  my $inode_num = $dataset->get_file_inode_num(filename => 'my_file');
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_file_inode_num
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->_get_file_id(@_);
}


## @name ::get_pvbns
##
## @summary Gets pvbns for the given fbns of the file.
##
## @description
## Gets pvbns for the given fbns of the file.
##
## @arg filename Required: File name.
##
## @arg fbns Required: Arrayref of fbns for which to obtain the pvbns.
##
## @return Array of pvbns.
##
## @example
##  my @pvbns = $dataset->get_pvbns(filename => 'my_file', fbns => [0 .. 7]);
##
sub get_pvbns
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        filename => {type => SCALAR,},
			fbns =>{type => ARRAYREF,},
                });
	
	my $filename = $args{filename};
	my $fbns = $args{fbns};
	my @pvbns = ();
	
	foreach my $fbn (@$fbns)
	{
		my %fbn_info = $class->{_c_dataset}->{_common}->_get_vbn_info_single(filename => $filename,
										     fbn => $fbn);
		push(@pvbns, $fbn_info{vbn});
	}
	
	return @pvbns;
}


## @name ::attach_dataset_info
##
## @summary Creates internal information for files appearing after cloning a lun. 
##
## @description Creates internal information for files appearing after cloning a lun.
##
## @arg source_lun Required: Compression::Lun object corresponding to the source lun.
##
## @arg clone_lun Required: Compression::Lun object corresponding to the clone lun.
##
## @examples
## 
sub attach_lun_clone_info
{
        my $class = shift;
	return $class->{_c_dataset}->{_common}->_attach_lun_clone_info(@_);
}


## @name ::wafl_punch
##
## @summary Punch holes in a file using "wafl punch". 
##
## @description Punch holes in a file using "wafl punch". 
##
## @arg filename Required: Name of the file.
##
## @arg fbn Required: Fbn to be used for punching.
##
## @arg count Required: Number of holes to be punched.
##
## @examples
## $obj->wafl_punch(filename => 'my_file', fbn => 1, count => 4);
## 
sub wafl_punch
{
	my $class = shift;
	return $class->{_c_dataset}->{_common}->wafl_punch(@_);
}


## @name ::generate_data_for_quick_file_check
##
## @summary Generates data that is incompressible by quick file check.
##
## @description Generates data that is incompressible by quick file check.
##
## @arg incompressible_size Optional: Size of the first chunk of data in the cg, incompressible.
## Default: 4k.
##
## @arg pattern Optional: Pattern to fill the rest of the cg with.
## Default: a.
##
## @arg total_len Optional: Total len for the data.
## Default: C_CG_SIZE.
##
## @return Generated data.
##
## @examples
## my $data = $obj->generate_data_for_quick_file_check(incompressible_size => 4 * KB + 1);
## 
sub generate_data_for_quick_file_check
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		incompressible_size => {type => SCALAR, regex => qr/^\d+$/, optional => 1, default => 4 * KB,},
		pattern => {type => SCALAR, optional => 1, default => "a",},
		total_len => {type => SCALAR, optional => 1, default => C_CG_SIZE,},
	});
	
	my $incompressible_size = $args{incompressible_size};
	my $pattern = $args{pattern};
	my $total_len = $args{total_len};
	my $pattern_size = $total_len - $incompressible_size;
	
	if ($total_len > C_CG_SIZE) {
		logresult('script', "This method only generates data up to 1 CG");
	}
	
	if ($incompressible_size > $total_len) {
		logresult('script', "Incompressible_size must be smaller than the total len");
	}
	
	logcomment("Generating incompressible data for quick file check:\n" .
		   "Incompressible size: $incompressible_size\n" .
		   "Pattern: \"$pattern\"\n" .
		   "Total len: $total_len");
	
	my $data = $class->get_uncompressable_data(size => $incompressible_size);
	if ($pattern_size > 0) {
		$data = $data . ($pattern x $pattern_size);
	}
	
	return $data;
}


## @name ::verify_vbn_zero
##
## @summary Verifies that a specified block is VBN_ZERO. 
##
## @description Verifies that a specified block is VBN_ZERO.  
##
## @arg filename Required: Name of the file.
##
## @arg fbns Required: Fbn to be verified.
## Can be a single fbn or a range (ex: 1, or 2-5).
##
## @arg expected Required: Expected result: VBN_ZERO or NOT_VBN_ZERO.
## Default: VBN_ZERO.
##
## @examples
## $obj->verify_vbn_zero(filename => 'my_file', fbns => 1, expected => "NOT_VBN_ZERO");
## $obj->verify_vbn_zero(filename => 'my_file', fbns => "2-7");
##  
sub verify_vbn_zero
{
	my $class = shift;
	return $class->{_c_dataset}->{_common}->verify_vbn_zero(@_);
}


################################################################
# Internal subroutines
################################################################

# Internal method
# Resets the value of the internal class attribute _has_debug_kernel (to the caller-provided value).
sub _set_debug_kernel
{
	my $class = shift;
	my $debug_build = shift;
	
	$class->{_has_debug_kernel} = $debug_build;
	$class->{_c_dataset}->{_has_debug_kernel} = $debug_build;
	$class->{_c_dataset}->{_common}->{_has_debug_kernel} = $debug_build;
}

# Internal subroutine, used only by VolClone
# to allow processing of the files whithin the cloned volume
sub _attach_to_cloned_dataset
{
        my $class = shift;
        return $class->{_c_dataset}->{_common}->_attach_to_cloned_dataset(@_);
}

# internal function currently used by Compression::SnapVault::restore
# for the case when restoring to a different qtree
sub _replace_qtree_name {
        my $class = shift;
        return $class->{_c_dataset}->_replace_qtree_name(@_);
}

# internal function currently used by Compression::SnapMirror::_update_dataset_hash
sub _get_qtree_file_hash {
        my $class = shift;
        return $class->{_c_dataset}->_get_qtree_file_hash(@_);
}

sub _update_vfiler_mountpoint
{
	my $class = shift;
	my $new_vfiler_mountpoint = shift;
	
	$class->{_c_dataset}->{_vfiler_mountpoint} = $new_vfiler_mountpoint;
}


sub _update_vol_obj
{
	my $class = shift;
	$class->{_c_dataset}->_update_vol_obj(@_);
}


1;
