#
# $Id: //depot/prod/test/main/lib/Compression/Dataset/CIFS.pm $
#
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)


package Compression::Dataset::CIFS;

use strict;
use warnings;

use Tharn;

use Compression::Dataset::Common;
use Module::Common;
use Data::Dumper;
use SPARSE::Resource::Volume;
use SPARSE::Resource::Filer;
use SPARSE::Util::FileIO;
use SPARSE::Util::Logger;
use SPARSE::Util::Utilities;
use FS::Util::Utilities qw(file_reserve dumpblock);
use Dense::Utils;
use Dense::TestData;
use cat qw(do_shell);
use File::Copy;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Compression::Constants qw(:all);
use Compression::Utils qw(_send_cmd_to_filer);
use File::Compare;
use File::Basename;
use Carp qw(croak);
use FileManip;
use filer;
use Compression::Dataset::Constants qw(:all);
use Digest::SHA qw(sha1);
use SnapCmds;
use TestEnv::MinWin qw(execute_cmd unmap_drive map_drive);

BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}


## @name ::new
##
## @summary perform dataset initialization
##
## @description
## Perform all the basic dataset initialization required here
##
## @return Compression::Dataset object
##
## @arg volume Required: NOTE Volume object
## @arg client Optional: name of the client
## @arg debug Optional: logs additional debug messages specific to this module
##
## @examples
## 	$self->{dataset}  = Compression::Dataset->new(volume => $self->{volume}->{_vol},
##						      client => $self->{_client}->{name} );
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
##

sub new
{
	my $class 	= shift;
	my (%args) = validate (@_, {volume    => { type=>OBJECT, },
				    vserver   => { type=>SCALAR|UNDEF, optional => 1 },
				    client    => { type=>SCALAR, },
				    skip_mount     => { type=>BOOLEAN, optional => 1, default=>FALSE },
				    debug     => { type=>SCALAR, optional => 1, default=>FALSE },
				    ui        => {type => OBJECT, isa => "Compression::UI"},
				    boot_mode => {type => SCALAR, optional => 1, default => C_ONTAP_BOOT_MODE},
				    has_debug_kernel => {type => SCALAR, optional => 1, default => TRUE}, # So that scripts can continue
				    vfiler_mountpoint => {type => SCALAR|UNDEF, optional => 1,}, # Not used for cifs yet
		} );	
	my $self = {};
	$self->{ID} = $$;
	$self->{Description} = "Dataset::CIFS";
	$self->{volume} = $args{volume};
	$self->{_vserver} = $args{vserver};
	$self->{client} = ( defined $args{client} )? $args{client} : "localhost";
	$self->{client_host} = host($self->{client});
	$self->{client_conn} = connect("#rsh");
	$self->{client_hosttype} = $self->{client_host}->hosttype;
	$self->{client_ip} = $self->{client_host}->default_ip;
	$self->{client_name} = $self->{client};
	$self->{_ui} = $args{ui};
	$self->{_debug} = $args{debug};	
	$self->{_has_debug_kernel} = $args{has_debug_kernel};
	$self->{skip_mount} = $args{skip_mount};
        delete ($args{skip_mount});
		
	$self->{_boot_mode} = $args{boot_mode};
	if ($self->{_boot_mode} eq C_ONTAP_7MODE) 
	{
		if (!($self->{volume}->isa("NOTE::Volume")))
		{
			logresult('script', "For 7-mode, the volume argument is required to be a NOTE\:\:Volume object");
		}
		$self->{_conn} = $self->{volume}->Filer();
		$self->{vol_name} = $self->{volume}->name;
		$self->{_client} = NOTE::Client->new(name => $self->{client}, storage => $self->{volume} );		
	}
	elsif ($self->{_boot_mode} eq C_ONTAP_CMODE) 
	{
		if (!($self->{volume}->isa("NOTE::GX::Volume")))
		{
			logresult('script', "For 10-mode, the volume argument is required to be a NOTE\:\:GX\:\:Volume object");
		}
		$self->{_conn} = $self->{volume}->Node();
		$self->{vol_name} = $self->{volume}->name;

		# Skip mounting so we specifically mount rw (just as Volume does), to get the .admin mountpoint
		$self->{_client} = NOTE::Client->new(name => $self->{client}, storage => $self->{volume}, skip_remount => 1 );
		if (!$self->{skip_mount}){	# skip mount when auto_mount set to false
			$self->{volume}->mount( client => $self->{client}, RW => 1 );		
		}
	}
	else
	{
		logresult('script', "Unknown boot_mode: " . $self->{_boot_mode});
	}
	
	delete $args{client};
	$self->{_common} = Compression::Dataset::Common->new(%args);
	
	$self->{sfsr_will_restore_file_as_compressed} = $self->{_ui}->get_constant_value( constant_code => 'SFSR_WILL_RESTORE_FILE_AS_COMPRESSED' );
	
	bless $self, $class;

	$self->{drive_letter} = $self->{_client}->mountpoint;
	unmap_drive($self->{client_name},$self->{drive_letter});
	$self->cifs_map_share();	

	if (!$self->{skip_mount}){
		$self->{_common}->{mountpoint} = $self->get_mountpoint();
		$self->{_common}->{testdir_path} = $self->{_common}->{mountpoint};
	} else {
		$self->{_common}->{mountpoint} = '';
		$self->{_common}->{testdir_path} = '';
	}

	logcomment("Dataset Initialized with ". Dumper($self) ) if ( $self->{_debug} );

	return $self;

}

## @name ::cifs_map_share
##
## @summary cifs_map_share
##
## @description
## cifs_map_share
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub cifs_map_share {
		my $class = shift;
		my $client_name = $class->{client_name};
		my $drive_letter = $class->{drive_letter};
		my $cifs_share_path = $class->get_cifs_share_path_unc();
		logcomment("cifs_map_share: $cifs_share_path on $drive_letter");
		my $username = C_CIFS_ADMIN_USER;
		my $password = C_CIFS_ADMIN_PASSWORD;
		my $bg_map = subtest("lib/Compression/Dataset/CIFS/Scripts/cifs_map_share.thpl",
							"DRIVE=$drive_letter", "SHARE=$cifs_share_path",
							"USERNAME=$username", "PASSWORD=$password",
				'-client',$client_name
		);
}


## @name ::get_cifs_share_path_unc
##
## @summary get_cifs_share_path_unc
##
## @description
## get_cifs_share_path_unc
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_cifs_share_path_unc {
	my $class = shift;
	my $volume = $class->{volume};
	my $server_ip = $volume->Filer->defaultip;
	my $vol_name = $volume->name;
	my $cifs_share_path = '\\\\'. $server_ip .'\\'. $vol_name;
	return $cifs_share_path;
}

## @name ::get_cifs_share_path_local
##
## @summary get_cifs_share_path_local
##
## @description
## get_cifs_share_path_local
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_cifs_share_path_local {
	my $class = shift;
	my $drive_letter = $class->{drive_letter};
	my $cifs_share_path = $drive_letter;
	return $cifs_share_path;
}


## @name ::create
##
## @summary create the dataset or set of files on the volume
##
## @description
## Create set of files or specific dataset for the volume
##
## @return hash of status and output
##
##
## @arg	files		Required: Array of hashrefs ( one for each file )
##		files => [
##					{ 
##						name	=>	"f1", 
##						offset	=>	0, #default=0
##						size	=> ( 4 * KB), 
##						type	=> 'uncompressable|compressable|existing_data|write_bytes|blocksondisk|large', #default='write_bytes'
##						compression	=> 0,
##						blocksondisk	=> 0,
##						pattern	=> 'x', #default='a'
##						data	=>	$data, #only use if type=existing_data
##					},
##				]
##
## @arg	client		Optional: Client
## @arg	path		Optional: Defaults to root of volume
## @arg	testdir		Optional: Test directory
## @arg	ds_type		Optional: Dataset Type
## @arg compression	Optional: File compressability (0 - 100)
## @arg blocksondisk	Optional: Number of blocks per compression group written on disk
## @arg	offset		Optional: Offset to for writing files ( overriden by offset for individual file, default is 0 ).
## @arg	len			Optional: Length of data to be written ( overriden by size of individual file )
## @arg	pattern		Optional: Default 'a'
## @arg	background	Optional: Run in background ( TRUE|FALSE )
## @arg dont_die_on_error	Optional: If volume is full, script won't die ( TRUE|FALSE ).  Setting this to TRUE has same effect as 
##		setting expected_result to C_RC_FAILURE_NON_FATAL.
## @arg expected_result	Optional: Croak if not expected ( C_RC_SUCCESS(default)|C_RC_INSUFFICIENT_SPACE|C_RC_FAILURE_NON_FATAL). If 
##						dont_die_on_error is TRUE, expected_result defaults to C_RC_FAILURE_NON_FATAL )
##
## @examples
##
##
## %res = $dataset->create( 
##		files => [
##					{ 
##						name => "f1", 
##						size => GB, 
##						type=>'uncompressable', #default='write_bytes' 
##					},
##				],
##	);
##	
##
## %res = $dataset->create( 
##		files => [
##					{ name => "f1", offset=>0, size => ( 4 * KB) },
##				],
##		background => TRUE,
##	);
## $dataset->wait_finish( filename => 'f1' );
##
##
## my $data = $dataset->get_uncompressable_data(size=>4 * KB);
## %res = $dataset->create( 
##		files => [
##					{ name => "f1", data=>$data, type=>'existing_data' },
##				],
##	);
## 
##	Notes:
##	type can be:
##		1. write_bytes		-(default) repeats a pattern
##		2. existing_data	-need to call get_uncompressable_data() first if using this option
##		3. uncompressable	-create uncompressable data
##		4. compressable		-creates a file with a certain percentage of compressable blocks
##		5. blocksondisk		-each compression group in a file compresses to a certain number of blocks
##		4. large			-originally for large files ( will be depreciated )
##	pattern needs to be specified if type is write_bytes or large ( default is 'a' )
##	data only should be used if type=existing_data
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com, kozel@netapp.com
##

sub create
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		files           => { type=>ARRAYREF },
		client		=> { type=>SCALAR, optional => 1 },
		qtree		=> { type=>OBJECT, optional => 1, isa => "Compression::Qtree", },
		path		=> { type=>SCALAR, optional => 1 },
		testdir		=> { type=>SCALAR, optional => 1 },
		ds_type		=> { type=>SCALAR, optional => 1 },
		compression	=> { type=>SCALAR, optional => 1 },
		blocksondisk	=> { type=>SCALAR, optional => 1 },
		offset		=> { type=>SCALAR, optional => 1 },
		len             => { type=>SCALAR, optional => 1 },
		pattern		=> { type=>SCALAR, optional => 1 },
		background	=> { type=>SCALAR, optional => 1, default=>FALSE },
		dont_die_on_error => { type=>SCALAR, optional=>1, default=>FALSE },
		disable_checksums => { type=>SCALAR, optional=>1, default=>TRUE },
		expected_result => { type=>SCALAR, optional=>1, default=>C_RC_SUCCESS },
		lun => {type => OBJECT, isa => "Compression::Lun", optional => 1,}, 
	} );
	
	my @files               = @{$args{files}};
	my $vol_mountpoint      = $class->get_mountpoint();
	my $client              = ( defined $args{client} )? $args{client} : "localhost";
	my $testdir		= ( defined $args{testdir} )? $args{testdir} : $class->{_common}->{testdir};
	my $ds_type		= ( defined $args{ds_type} )? $args{ds_type} : C_DFLT_COMPRESSION_DATASET;
	my $compression	        = ( defined $args{compression} )? $args{compression} : -1;
	my $blocksondisk        = ( defined $args{blocksondisk} )? $args{blocksondisk} : -1;
	my $offset 		= ( defined $args{offset} )? $args{offset} : 0;
	my $len			= ( defined $args{len} )? $args{len} : "1024";
	my $pattern		= ( defined $args{pattern} )? $args{pattern} : "a";
	my $background	        = ( defined $args{background} )? $args{background} : FALSE;
	my $path 		= ( defined $args{path} )? $args{path} : $vol_mountpoint;
	my $dont_die_on_error	= $args{dont_die_on_error};
	my $disable_checksums	= $args{disable_checksums};
	my $expected_result     = $args{expected_result};
	my $qtree               = $args{qtree};
	my $lun                 = $args{lun};
	if (defined($lun)) {
		logresult('script', "Lun support is not available for cifs yet");
	}

	logcomment("In dataset create\n");
	if (($background == TRUE) && ($dont_die_on_error == TRUE)) {
		logresult('script', '"background" and "dont_die_on_error" can not both be TRUE');
	}

	$expected_result = C_RC_FAILURE_NON_FATAL if $dont_die_on_error == TRUE;
	$dont_die_on_error = TRUE if $expected_result == C_RC_FAILURE_NON_FATAL;
	

	#################################################################
	# debugging....
	$class->ls();
	#################################################################
	my %res;
	if ( defined $qtree ) {
		$path .= '\\'. $qtree->{name};
	}

	# 
	# Create testdir if needed
	#
	my $testdir_path = $path;
	if ($testdir) {
		$testdir_path .= $testdir;
		$class->mk_dir(name=>$testdir, path=>$path);
	}

	#
	# Check filenames unique
	#
	my %seen_filenames;
	foreach my $file ( @files )
	{
		if ( $seen_filenames{ $file->{name} } ) {
			croak "filenames must be unique $!";
		}
		$seen_filenames{$file->{name}} = 1;
	}

	#
	# Create object with path info for each file
	#
	foreach my $file ( @files )
	{
		my $filename = $file->{name};
		$class->_new_file(filename=>$filename,
				  testdir=>$testdir,
				  path=>$path,
		);
		$file->{path}          = $class->{_common}->{files}->{$filename}->{path};
		$file->{filer_path}    = $class->{_common}->{files}->{$filename}->{filer_path};
		$file->{relative_path} = $class->{_common}->{files}->{$filename}->{relative_path};
	}

	#
	# Save other class info
	#
	$class->{_common}->{ds_type} = $ds_type;
	$class->{_common}->{testdir} = $testdir;
	$class->{_common}->{testdir_path} = $testdir_path;

	if ($ds_type eq C_DATASET_SPARSE)
	{
		logcomment("In dataset create ds_type = $ds_type");
		eval
		{
			foreach my $file ( @files )
			{
				my $filename = $file->{name};
				logcomment("adding file = $file->{path}");
				
				my $offset	= ( defined $file->{offset} )	? $file->{offset} 	: $offset;
				my $len		= ( defined $file->{len} )		? $file->{len} 		: $file->{size};
				my $type	= ( defined $file->{type} )		? $file->{type} 	: 'write_bytes';
				my $compression = ( defined $file->{compression} ) ? $file->{compression}	: $compression;
				my $blocksondisk = ( defined $file->{blocksondisk} ) ? $file->{blocksondisk}	: $blocksondisk;
				my $pattern	= ( defined $file->{pattern} )	? $file->{pattern} 	: $pattern;
				my $data	= ( defined $file->{data} )		? $file->{data} 	: '';
				my $display_pattern = substr($pattern, 0, 12) . '.....';
				my $expected_result	= ( defined $file->{expected_result} )	? $file->{expected_result} 	: $expected_result;
				
				$type = 'uncompressable' if $compression == 0;
				$type = 'blocksondisk' if $blocksondisk != -1;
				
				my $file_path = $file->{path};
				
				my $comment = "\n" .
				              "\t\tcreate file $file_path\n" .
					      "\t\toffset: $offset\n" .
					      "\t\ttype: $type\n" .
					      "\t\texpected_result: " . $expected_result . "\n";
				$comment .= "\t\tlen: $len\n" if $len;
				$comment .= "\t\tpattern: $display_pattern\n" if $type ne 'uncompressable';
				$comment .= "\t\tcompression: $compression\n" if $compression != -1;
				$comment .= "\t\tblocksondisk: $blocksondisk\n" if $blocksondisk != -1;
				logcomment($comment);
				
				if ($background == TRUE) {
					$class->{_common}->{files}->{$filename}->{bg_subtest} = $class->create_file_cifs(path => $file_path,
															 offset => $offset,
															 len => $len,
															 pattern => $pattern,
															 bg => TRUE,
															 data => $data,
															 type => $type,);							
				}
				else {
					$class->create_file_cifs(path => $file_path,
								 offset => $offset,
								 len => $len,
								 pattern => $pattern,
								 data => $data,
								 type => $type,);
				}
			} 
		};
		
		if( $@ )
		{
			logresult("fatal", "$@");
		}
	}
	else
	{
		logresult( "fatal", "Unsupported dataset type $ds_type" );
	} 
	
	if ($class->{_debug}) { $class->ls(); }
	
	return %res;
}

## @name ::modify
##
## @summary modify the dataset or set of files on the volume
##
## @description
## Modify set of files or specific dataset for the volume
##
## @return hash of status and output
##
##
## @arg	files		Required: Array of hashrefs ( one for each file )
##		files => [
##					{ 
##						name	=> "f1", 
##						offset	=> ( 4 * KB),
##						len		=> ( 8 * KB), 
##						type	=> 'uncompressable|existing_data|write_bytes', #default='write_bytes'
##						pattern	=> 'x', #default='a'
##						data	=>	$data, #only use if type=existing_data
##					},
##				]
## @arg	client		Optional: Client
## @arg	path		Optional: Defaults to root of volume
## @arg	ds_type		Optional: Dataset Type
## @arg	offset		Optional: Offset to for writing files ( overriden by offset for individual file, default is 0 ).
## @arg	len			Optional: Length of data to be written ( overriden by len of individual file )
## @arg	pattern		Optional: Default 'a'
## @arg	background	Optional: Run in background ( TRUE|FALSE )
## @arg dont_die_on_error	Optional: If volume is full, script won't die ( TRUE|FALSE ).  Setting this to TRUE has same effect as 
##		setting expected_result to C_RC_FAILURE_NON_FATAL.
## @arg expected_result	Optional: Croak if not expected ( C_RC_SUCCESS(default)|C_RC_INSUFFICIENT_SPACE|C_RC_FAILURE_NON_FATAL). If 
##						dont_die_on_error is TRUE, expected_result defaults to C_RC_FAILURE_NON_FATAL )
##
## @examples
##
## %res = $dataset->modify( 
##		files => [
##					{ name => "f1", offset=>( 4 * KB), len => ( 8 * KB), },
##				],
##		background => TRUE,
##	);
## $dataset->wait_finish( filename => 'f1' );
##
##
## %res = $dataset->modify( 
##		files => [
##					{ name => "f1", offset=>( 4 * KB), len => ( 8 * KB), type=>'uncompressable' },
##				],
##	);
##
##
## my $data = $dataset->get_uncompressable_data(size=>4 * KB);
## %res = $dataset->modify( 
##		files => [
##					{ name => "f1", data=>$data, type=>'existing_data' },
##				],
##	);
## 
##	Notes:
##	type can be:
##		1. write_bytes		-(default) repeats a pattern
##		2. existing_data	-need to call get_uncompressable_data() first if using this option
##		3. uncompressable	-create uncompressable data
##		4. large			-originally for large files ( will be depreciated )
##		5. holey			-add holey blocks
##	
##	data only should be used if type=existing_data
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com, kozel@netapp.com
##
sub modify
{
	my $class 	= shift;
#	logcomment("Dumping class: " . Dumper($class) );
	my (%args) = validate (@_, {	
		files		=> { type=>ARRAYREF },
		client		=> { type=>SCALAR, optional => 1 },
		path		=> { type=>SCALAR, optional => 1 },
		fill		=> { type=>SCALAR, optional => 1 },
		ds_type		=> { type=>SCALAR, optional => 1 },
		compression	=> { type=>SCALAR, optional => 1 },
		blocksondisk	=> { type=>SCALAR, optional => 1 },
		offset		=> { type=>SCALAR, optional => 1 },
		len			=> { type=>SCALAR, optional => 1 },
		pattern		=> { type=>SCALAR, optional => 1 },
		background	=> { type=>SCALAR, optional => 1, default=>FALSE },
		dont_die_on_error => { type=>SCALAR, optional=>1, default=>FALSE },
		disable_checksums => { type=>SCALAR, optional=>1, default=>TRUE },
		expected_result => { type=>SCALAR, optional=>1, default=>C_RC_SUCCESS },
		clone_lun => {type => SCALAR|UNDEF, optional => 1,},
	} );
	
	my @files             = @{$args{files}};
	my $vol_mountpoint    = $class->get_mountpoint();
	my $client            = ( defined $args{client} )? $args{client} : "localhost";
	my $path              = ( defined $args{path} )? $args{path} : $vol_mountpoint;
	my $fill              = ( defined $args{fill} )? $args{fill} : 0;
	my $offset            = ( defined $args{offset} )? $args{offset} : 0;
	my $len               = ( defined $args{len} )? $args{len} : "1024";
	my $compression       = ( defined $args{compression} )? $args{compression} : -1;
	my $blocksondisk      = ( defined $args{blocksondisk} )? $args{blocksondisk} : -1;
	my $pattern           = ( defined $args{pattern} )? $args{pattern} : "a";
	my $samedata          = ( defined $args{samedata} )? $args{samedata} : 0;
	my $background	      = ( defined $args{background} )? $args{background} : FALSE;
	my $dont_die_on_error = $args{dont_die_on_error};
	my $disable_checksums = $args{disable_checksums};
	my $expected_result   = $args{expected_result};
	my $ds_type           = $class->{_common}->{ds_type};
	my $lun                 = $args{lun};
	if (defined($lun)) {
		logresult('script', "Lun support is not available for cifs yet");
	}
	
	logcomment("In dataset modify\n");
	if (($background == TRUE) && ($dont_die_on_error == TRUE)) {
		logresult('script', '"background" and "dont_die_on_error" can not both be TRUE');
	}
	
	$expected_result = C_RC_FAILURE_NON_FATAL if $dont_die_on_error == TRUE;
	$dont_die_on_error = TRUE if $expected_result == C_RC_FAILURE_NON_FATAL;

	my %res;

	#################################################################
	# debugging....
	$class->ls();
	#################################################################

	if( $samedata )
	{
		$pattern = "a";
	}
	elsif ( ! $pattern )
	{
		$pattern = "b";
	}

	
	if ($ds_type eq C_DATASET_SPARSE)
	{
		logcomment("In dataset modify C_DATASET_SPARSE\n");
		eval
		{
			foreach my $file ( @files )
			{
				my $filename = $file->{name};
				my $file_path =  $class->{_common}->{files}->{$filename}->{path};
				my $relative_path = $class->{_common}->{files}->{$filename}->{relative_path};
				
				my $offset	= ( defined $file->{offset} )	? $file->{offset} 	: $offset;
				my $len		= ( defined $file->{len} )		? $file->{len} 		: $len;
				my $type	= ( defined $file->{type} )		? $file->{type} 	: 'write_bytes';
				my $compression	= ( defined $file->{compression} ) 	? $file->{compression}	: $compression;
				my $blocksondisk = 	( defined $file->{blocksondisk} )	? $file->{blocksondisk}	: $blocksondisk;
				my $pattern	= ( defined $file->{pattern} )	? $file->{pattern} 	: $pattern;
				my $data	= ( defined $file->{data} )		? $file->{data} 	: '';
				my $display_pattern = substr($pattern, 0, 12) . '...';
				my $shift	= ( defined $file->{'shift'} ) ? $file->{'shift'} : FALSE;
				my $expected_result	= ( defined $file->{expected_result} )	? $file->{expected_result} 	: $expected_result;
				
				$type = 'uncompressable' if $compression == 0;
				$type = 'blocksondisk' if $blocksondisk != -1;
				
				if (($shift == TRUE)&&($type !~/\b(write_bytes|holey)\b/)) {
					logresult("script", "shifting data currently only works if type is write_bytes(default) " .
						            "or 'holey'");
				}
				
				my $comment = "\n" .
				              "\t\tmodify file $file_path\n" .
					      "\t\tfilename: $filename\n" .
					      "\t\toffset: $offset\n" .
					      "\t\ttype: $type\n";
				$comment .= "\t\tlen: $len\n" if $type ne 'existing_data';
				$comment .= "\t\tpattern: $display_pattern\n" if $type eq 'write_bytes' || $type eq 'large';
				$comment .= "\t\tcompression: $compression\n" if $compression != -1;
				$comment .= "\t\tblocksondisk: $blocksondisk\n" if $blocksondisk != -1;
				$comment .= "\t\tshift: $shift\n" if $shift == TRUE;
				logcomment($comment);
				
				if ($background == TRUE) {
					$class->{_common}->{files}->{$filename}->{bg_subtest} = $class->modify_file_cifs(path => $file_path,
															 offset => $offset,
															 len => $len,
															 pattern => $pattern,
															 bg => TRUE,
															 type => $type,
															 data => $data,
															 shift => $shift,);
				}
				else {
					$class->modify_file_cifs(path => $file_path,
								 offset => $offset,
								 len => $len,
								 pattern => $pattern,
								 type => $type,
								 data => $data,
								 shift => $shift,);
				}
			} 
		};
		
		if( $@ )
		{
			logresult("fatal", "$@");
		}
	}
	else
	{
		logresult( "fatal", "Unsupported dataset type $ds_type" );
	} 
	
	if ($class->{_debug}) { $class->ls(); }
	
	return %res;
}


## @name ::get_path
##
## @summary return path to file
##
## @description
## return path to file
##
## @return full path to the file
##
## @arg	filename Required: name of file.
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset.
##
## @examples
##  my $filepath = $dataset->get_path( filename=>'f1' );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_path
{
	my $class = shift;
	return $class->{_common}->get_path(@_);
}

## @name ::delete_file
##
## @summary delete the file
##
## @description
## delete the file
##
## @return hash of status
##
## @arg filename	Required: name of file to delete
##
## @examples
##  $dataset->delete_file( filename=>'f1' );
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
##
sub delete_file
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename => { type=>SCALAR },
		expected => { type => SCALAR, default => C_RC_SUCCESS,},} );
	
	my %res;
	my $filename = $args{filename};
	
	if ($args{expected} != C_RC_SUCCESS) {
		logresult('script', "Expected support is not available yet for CIFS for delete_file")
	}
	$class->_filename_exists( filename=>$filename, expected=>TRUE );
	
	my $client_name = $class->{client_name};
	my $file_path = $class->_generate_full_path(filename => $filename);
	my $bg = subtest("lib/Compression/Dataset/CIFS/Scripts/delete_file.thpl",
						"PATH=$file_path",
			'-client',$client_name
	);

	$class->_remove_file(filename => $filename);
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::compare_datasets
##
## @summary compare the files on volume vs baseline dataset
##
## @description
## file compare on the volume with baseline dataset
##
## @return
## %res;
##	$res{'rc'}		= 'C_RC_SUCCESS';
##
## @arg dataset_2	Required: 2nd dataset to compare
## @arg expected_result Optional: EQUAL or NOT_EQUAL. Default is EQUAL. Will croak if result is not expected.
## @arg ignore_snapshots Optional: don't compare snapshots.  Default is TRUE.
##
## @examples
##	%res = $dataset_1->compare_datasets( 
##		dataset_2 => $dataset_2,
##	);
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
##
sub compare_datasets
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		dataset_2 => { type=>OBJECT },
		expected_result => { type => SCALAR, optional => 1, regex => qr/^(EQUAL|NOT_EQUAL)$/, default => 'EQUAL'},
		ignore_snapshots => { type => SCALAR, optional => 1, default => TRUE},	
	} );

	my $dataset_2 	= ( defined $args{dataset_2} )? $args{dataset_2} : undef;
	my $expected_result = $args{'expected_result'};
	my $ignore_snapshots = $args{'ignore_snapshots'};
	my %res;
	
	my $path_1 = $class->get_mountpoint();
	my $path_2 = $dataset_2->{_c_dataset}->get_mountpoint;

	my $op = "Comparing\n" .
	         "         $path_1\n" .
	         "    with of $path_2 \n" .
			 "ignore_snapshots: $ignore_snapshots \n" .
			 "expected_result: $expected_result\n";

	logcomment($op);

	my $client_name = $class->{client_name};
	my $subtest = subtest("lib/Compression/Dataset/CIFS/Scripts/compare_datasets.thpl",
						"PATH_1=$path_1", "PATH_2=$path_2",
						"EXPECTED_RESULT=$expected_result",
						"IGNORE_SNAPSHOTS=$ignore_snapshots",
			'-client',$client_name
	);


	$res{'rc'} = 'C_RC_SUCCESS';
	return %res;

}


## @name ::compare_files
##
## @summary compares 2 files
##
## @description
## compares 2 files.  Croaks if not expected result.
##
## @return hash of status and output
##	$res{'result'} = EQUAL or NOT_EQUAL
## 
## @arg file_1	Required: path to first file
## @arg file_2	Required: path to second file
## @expected_result	Optional: EQUAL|NOT_EQUAL ( default is 'EQUAL' )
##
## @example
##
##	%res = $dataset->compare_files( 
##						file_1=> $dataset_1->get_path( filename=>'f1' ), 
##						file_2=> $dataset_2->get_path( filename=>'f1' ),
##						expected_result=>EQUAL|NOT_EQUAL #default:EQUAL
##						);
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub compare_files
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		file_1 => { type=>SCALAR },
		file_2 => { type=>SCALAR },
		expected_result => { type => SCALAR, optional => 1, regex => qr/^(EQUAL|NOT_EQUAL)$/, default => 'EQUAL'},
	});
	
	my $file_1 = $args{file_1};
	my $file_2 = $args{file_2};
	my $expected_result = $args{expected_result};
	my %res;
	my $result;

	my $client_name = $class->{client_name};
	my $subtest = subtest("lib/Compression/Dataset/CIFS/Scripts/compare_files.thpl",
						"FILE_1=$file_1", "FILE_2=$file_2",
						"EXPECTED_RESULT=$expected_result",
			'-client',$client_name
	);	

	$res{'rc'} = C_RC_SUCCESS;
	$res{'result'} = $result;
	return %res;
}

## @name ::copy_file
##
## @summary copies a file
##
## @description
## Copies a file
##
## @return hash with status
## 
## @arg source	Required: path to source file
## @arg dest	Required: path to destination file
## @arg	new_dataset	Optional: use this only when copying to a new dataset ( i.e. new volume )
## @arg background	Optional: default is FALSE ( if TRUE, use with wait_finish )
##
## @examples
##
##	%res = $dataset->copy_file( 
##						source		=> $dataset_1->get_path( filename=>'f1' ), 
##						dest		=> $destination_file
##						new_dataset	=> $dataset_2  #optional: only use if copying file to different volume
##  );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub copy_file
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		source => { type=>SCALAR },
		dest => { type=>SCALAR },
		new_dataset => { type=>HASHREF, optional=>1 },
		background => { type=>SCALAR, optional => 1, default=>FALSE },
	} );
	
	my $background	= $args{background};
	my $src_file = $args{source};
	my $dst_file = $args{dest};
	$dst_file =~ s#/#\\#g;
	my $new_dataset = $args{new_dataset};
	fileparse_set_fstype("MSWin32");
	my $src_filename = basename($src_file);
	my $dst_filename = basename($dst_file);

	my $dst_vol = $class->{volume}; 
	my $dst_dataset = $class;
	if ( $new_dataset ) {
		$dst_vol = $new_dataset->{volume};
		$dst_dataset = $new_dataset;
	}
	my $dst_vol_mountpoint = $dst_dataset->get_mountpoint;
	#
	# If dataset is specified, verify that dst path is on correct volume.
	#
	if ( $new_dataset && ( $dst_file !~ /$dst_vol_mountpoint/ ) ) {
		croak "dst_file: $dst_file needs to include $dst_vol_mountpoint in its path $!";
	}
	
	my %res;
	logcomment("copy_file\n$src_file\n$dst_file\n");

	$dst_dataset->_new_file( filename=>$dst_filename );
	#
	# Update path info
	#
	$dst_dataset->{_common}->{files}->{$dst_filename}->{path} = $dst_file;
	if ( $dst_file =~ /$dst_vol_mountpoint/ ) {
		($dst_dataset->{_common}->{files}->{$dst_filename}->{relative_path} ) = $dst_file =~ /$dst_vol_mountpoint\\(.+)$/;
		$dst_dataset->{_common}->{files}->{$dst_filename}->{filer_path} = $dst_vol->mount_path .'/'.	
				$dst_dataset->{_common}->{files}->{$dst_filename}->{relative_path};
	} else {
		# file is being copied off the filer
		$dst_dataset->{_common}->{files}->{$dst_filename}->{relative_path} = undef;
		$dst_dataset->{_common}->{files}->{$dst_filename}->{filer_path}    = undef;
	}

	my $client_name = $class->{client_name};
	if ($background == TRUE) {
			$dst_dataset->{_common}->{files}->{$dst_filename}->{bg_subtest} 
					= subtest("lib/Compression/Dataset/CIFS/Scripts/copy_file.thpl",
								"SRC=$src_file","DST=$dst_file",
								'-client',$client_name,
								'-bg',
								'-runid','copy_file_'. $dst_filename,							
			);
	} else {
		my $bg = subtest("lib/Compression/Dataset/CIFS/Scripts/copy_file.thpl",
							"SRC=$src_file","DST=$dst_file",
							'-client',$client_name
		);		
	}
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}


## @name ::move_file
##
## @summary move a file
##
## @description
## Moves a file
##
## @return hash of status
## 
## @arg source	Required: path to source file
## @arg dest	Required: path to destination file
## @arg	new_dataset	Optional: use this only when copying to a new dataset ( i.e. new volume )
## @arg background	Optional: default is FALSE ( if TRUE, use with wait_finish )
##
## @examples
##
##	%res = $dataset->move_file( 
##						source		=> $dataset_1->get_path( filename=>'f1' ), 
##						dest		=> $destination_file
##						new_dataset	=> $dataset_2  #optional: only use if moving file to different volume
##  );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub move_file
{
	my $class 	= shift;
	my (%args) = validate (@_, {	source		=> { type=>SCALAR },
									dest		=> { type=>SCALAR },
									new_dataset => { type=>HASHREF, optional=>1 },
									background	=> { type=>SCALAR, optional => 1, default=>FALSE },
	} );
	my $background	= $args{background};
	my $src_file = $args{source};
	my $dst_file = $args{dest};
	$dst_file =~ s#/#\\#g;
	my $new_dataset = $args{new_dataset};
	fileparse_set_fstype("MSWin32");
	my $src_filename = basename($src_file);
	my $dst_filename = basename($dst_file);
	my %res;
	logcomment("move_file\n$src_file\n$dst_file\n");

	my $dst_vol = $class->{volume}; 
	my $dst_dataset = $class;
	if ( $new_dataset ) {
		$dst_vol = $new_dataset->{volume};
		$dst_dataset = $new_dataset;
	}
	my $dst_vol_mountpoint = $dst_dataset->get_mountpoint;
	# If dataset is specified, verify that dst path is on correct volume.
	if ( $new_dataset && ( $dst_file !~ /$dst_vol_mountpoint/ ) ) {
		croak "dst_file: $dst_file needs to include $dst_vol_mountpoint in its path $!";
	}
	
	$class->_remove_file(filename => $src_filename);
	
	$dst_dataset->_new_file( filename=>$dst_filename );
	#
	# Update path info
	#	
	$dst_dataset->{_common}->{files}->{$dst_filename}->{path} = $dst_file;
	if ( $dst_file =~ /$dst_vol_mountpoint/ ) {
		($dst_dataset->{_common}->{files}->{$dst_filename}->{relative_path} ) = $dst_file =~ /$dst_vol_mountpoint\\(.+)$/;
		$dst_dataset->{_common}->{files}->{$dst_filename}->{filer_path} = $dst_vol->mount_path .'/'. $dst_dataset->{_common}->{files}->{$dst_filename}->{relative_path};
	} else {
		# file is being moved off the filer
		$dst_dataset->{_common}->{files}->{$dst_filename}->{relative_path} = undef;
		$dst_dataset->{_common}->{files}->{$dst_filename}->{filer_path}    = undef;
	}

	
	my $client_name = $class->{client_name};
	if ($background == TRUE) {
			$dst_dataset->{_common}->{files}->{$dst_filename}->{bg_subtest} 
					= subtest("lib/Compression/Dataset/CIFS/Scripts/move_file.thpl",
								"SRC=$src_file","DST=$dst_file",
								'-client',$client_name,
								'-bg',
								'-runid','move_file_'. $dst_filename,							
			);
	} else {
		my $bg = subtest("lib/Compression/Dataset/CIFS/Scripts/move_file.thpl",
							"SRC=$src_file","DST=$dst_file",
							'-client',$client_name
		);		
	}	
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::get_file_size
##
## @summary get size of file in bytes
##
## @description
## get size of file in bytes
##
## @return size of file in bytes
##
## @arg filename	Required: name of file
## @arg snapshot	Optional: name of snapshot ( if file is in snapshot )
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset
## 
## @examples
##  my $file_size = $dataset->get_file_size( filename=>'f1');
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_file_size
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename => { type=>SCALAR },
		snapshot => { type=>SCALAR|UNDEF, optional=>1, },
		allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
	} );
	
	my $filename = $args{filename};
	my $snapshot = $args{snapshot};

	my $file_path = $class->_generate_full_path(filename => $filename,
							    snapshot => $snapshot,
							    allow_removed => $args{allow_removed});
	

	my $client_name = $class->{client_name};
	my $st = subtest("lib/Compression/Dataset/CIFS/Scripts/file_size.thpl",
						"PATH=$file_path",
			'-client',$client_name
	);
	
	my $file_size = $st->message_get(type => "size");

	logcomment("file size of: " . $file_path ." is ". $file_size) if $class->{_debug};
	
	return $file_size;
}

## @name ::get_timestamp
##
## @summary get timestamp of file
##
## @description
## get timestamp of file
##
## @return timestamp (last modify time) of file
##
## @arg filename	Required: name of file
## @arg snapshot	Optional: name of snapshot ( if file is in snapshot )
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset
## 
## @examples
##  my $timestamp = $dataset->get_timestamp( filename=>'f1');
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_timestamp
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename => { type=>SCALAR },
		snapshot => { type=>SCALAR|UNDEF, optional=>1, },
		allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
	} );
	
	my $filename = $args{filename};
	my $snapshot = $args{snapshot};
	
	my $file_path = $class->_generate_full_path(filename => $filename,
							    snapshot => $snapshot,
							    allow_removed => $args{allow_removed});

	my $client_name = $class->{client_name};
	my $st = subtest("lib/Compression/Dataset/CIFS/Scripts/get_timestamp.thpl",
						"PATH=$file_path",
			'-client',$client_name
	);
	
	my $mtime = $st->message_get(type => "mtime");

	logcomment("timestamp of: " . $file_path ." is ". $mtime);
	
	return $mtime;
}

## @name ::get_uncompressable_data
##
## @summary get uncompressable data
##
## @description create data that is uncompressable.  Only use this when you need to add
##  the exact same data to different files. Warning: out-of-memory error on large files.
##
## @return data
##
## @arg size	Required:	size of data to create
## 
## @examples
##  my $data = $dataset->get_uncompressable_data(size=>4 * KB);
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com, kozel@netapp.com
##
sub get_uncompressable_data
{
	my $class	= shift;
	my (%args)	= validate (@_, { size => { type=>SCALAR, regex => qr/^\d+$/ } } );
	my $size	= $args{size};
	my @charset =
              ( 0 .. 9, 'a' .. 'z', 'A' .. 'Z', '#', ',',
                qw(~ ! @ $ % ^ & * ( ) _ + = - { } | : " < > ? / . ' ; ] [ \ `)
            );
	my $data = join '', map $charset[rand @charset], 1..$size;
	return $data;
}



## @name ::read_file
##
## @summary verify that a file can be read
##
## @description
## This function is to verify that a file can be read.
## Croaks if it can not read file.
##
## @return status
##
## @arg	filename	Required:
## @arg snapshot	Optional: name of snapshot ( if file is in snapshot )
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset
##
## @examples 
##	$dataset->read_file( filename=>'f1', snapshot=>'ss1', allowed_removed=>TRUE );
##
sub read_file
{
	my $class 	= shift;
    my (%args) = validate (@_, {	filename => { type=>SCALAR },
									snapshot => { type=>SCALAR|UNDEF, optional=>1, },
									allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
	} );
	my %res;
	my $filename = $args{filename};
	my $snapshot = $args{snapshot};
	my $file_path = $class->_generate_full_path(filename => $filename,
							    snapshot => $snapshot,
							    allow_removed => $args{allow_removed});

	my $client_name = $class->{client_name};
	my $bg = subtest("lib/Compression/Dataset/CIFS/Scripts/read_file.thpl",
						"PATH=$file_path",
			'-client',$client_name
	);


	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::verify_checksums
##
## @summary verifies data integrity by comparing checksums
##
## @description
## computes SHA1 checksums for all compression groups and compares them to the precalculated ones
##
## @return hash of status
##
## @arg files		Optional: hash of the filenames
## @arg path		Optional: path to a file
##
## Note, one or the other must be defined
##
## @examples
##  $dataset->verify_checksums( files =>
##								[	{ name=> 'f1' }
##									{ name=> 'f2' }
##								]);
##
##
## @author kozel@netapp.com, dl-compression-qa@netapp.com
##

sub verify_checksums
{
	my $class = shift if ref($_[0]) eq "Compression::Dataset";
	my %args	= validate (@_, {	files	=> { type=>ARRAYREF, optional=>1 },
									path	=> { type=>SCALAR, optional=>1},
									die_on_error	=> { type=>SCALAR, optional=>1, default=>TRUE},
	});

	logresult("script", "need to implement for cifs");
}

## @name ::modify_all_CGs
##
## @summary  modify all CGs in a file
##
## @description modify a certain range within every CG in a file.  As long as the file has compressed CGs, this
## function will find them and overwrite the specified range in each CG.
##
## @return Hash with status.
##
## @arg filename Required: Name of the file.
## @arg offet Required: offset within each CG
## @arg len	Required: number of bytes to write within each CG
## @arg pattern Required: character to write
## @arg CG_ranges Optional: reference to array of ranges as obtained from get_CG_ranges. If undef,
## this function will call get_CG_ranges to get them.
## @arg estimate_boundries Optional: estimate CG ranges based on C_CG_SIZE ( default FALSE )
## Only set this if you're overwriting parts that are uncompressed (i.e. not real CGs ).
##
## @examples
## $obj->modify_all_CGs(filename=>$FILENAME, offset=>4*KB, len=>2*KB, pattern=>'c' )
##  
sub modify_all_CGs
{
	my $class 	= shift;
    my (%args) = validate (@_, {	filename	=> { type=>SCALAR },
									pattern		=> { type=>SCALAR },
									offset		=> { type=>SCALAR },
									len			=> { type=>SCALAR },
									CG_ranges	=> { type=>ARRAYREF, optional=>1, default=>undef },
									estimate_boundries	=> { type=>SCALAR, optional=>1, default=>FALSE },
	} );
	my %res;
	my $filename	= $args{filename};
	my $pattern		= $args{pattern};
	my $offset_each_CG		= $args{offset};
	my $len_each_CG			= $args{len};
	my $estimate_boundries			= $args{estimate_boundries};
	my $file_size = $class->get_file_size( filename => $filename );
	my $file_path = $class->get_path( filename=>$filename );
	my @CG_ranges = undef;
	my $num_CG_ranges = 0;
	if (defined $args{CG_ranges} ) {
		@CG_ranges = @{$args{CG_ranges}};
		$num_CG_ranges = scalar @CG_ranges;
	}
	
	my $num_CGs = int($file_size/ C_CG_SIZE);
	my $summary_text = "modify_all_CGs:\n".
				"filename: $filename\n".
				"offet: $offset_each_CG\n".
				"len: $len_each_CG\n".
				"pattern: $pattern\n";
	if ($num_CG_ranges>0) {
		$summary_text .= "CG_ranges: $num_CG_ranges\n";
	} else {
		$summary_text .= "CG_ranges: undef\n";
	}
	logcomment($summary_text);
	if (($offset_each_CG+$len_each_CG)> C_CG_SIZE ) {
		logresult("script", "the sum of offset and len need to be less than a CG:". C_CG_SIZE);
	}
	
	my $client_name = $class->{client_name};

	if ( ! $estimate_boundries ) {
		if (! $num_CG_ranges) {
			%res = $class->{_common}->get_CG_ranges( filename=>$filename );
			@CG_ranges = @{$res{CG_ranges}};
		}
		
		my $st = subtest("lib/Compression/Dataset/CIFS/Scripts/modify_all_CGs.thpl",
							"PATH=$file_path","OFFSET=$offset_each_CG","LEN=$len_each_CG","PATTERN=$pattern",
				'-client',$client_name
		);
		$st->message_put(message=>\@CG_ranges, type => "CG_ranges");

	} else { # estimate_boundries based on C_CG_SIZE


	}
	$res{'CG_ranges'} = \@CG_ranges;
	$res{'rc'} = C_RC_SUCCESS;
	return %res;

}

## @name ::single_file_snap_restore
##
## @summary 
##	Single file snap restore (SFSR) reverts a file in the active file system to the state contained in a snapshot
##
## @description
##	Single file snap restore (SFSR) reverts a file in the active file system to the state contained in a snapshot 
##
## @return hash of status
##
## @arg filename	Required: name of file
## @arg snapshot	Required: name of snapshot to restore from
## @arg wait_for_completion	Optional: don't return until file is completely restored ( default: TRUE )
## @arg expected Optional: Expected return code ( default: C_RC_SUCCESS )
##
## @examples
##	%res = $src_vol->{dataset}->single_file_snap_restore(filename=>$FILENAME, snapshot=>$SNAPSHOT);
sub single_file_snap_restore
{
	my $class 	= shift;
	my (%args) = validate (@_, { filename => { type=>SCALAR },
				     snapshot => { type=>SCALAR },
				     restore_as_path => { type=>SCALAR|UNDEF, optional=>1 },
				     wait_for_completion => { type=>SCALAR, optional=>1, default=>TRUE },
				     expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
	});
	my %res;
	my $filename = $args{filename};
	my $snapshot = $args{snapshot};
	my $wait_for_completion = $args{wait_for_completion};
	my $restore_as_path = $args{restore_as_path};
	my $rc = C_RC_SUCCESS;
	
	my $comment_str = "single_file_snap_restore\n" .
	                  "Filename: $filename\n" .
			  "Snapshot: $snapshot\n";
			  
	if (defined($restore_as_path)) {
		$comment_str .= "Restoring to path: $restore_as_path\n";
	}
	
	logcomment($comment_str);
	
	my $file_already_exists = $class->_filename_exists( filename=>$filename, expected=>C_RC_FAILURE_NON_FATAL );
	my $file_path = $class->_generate_file_path(filename => $filename,
						    allow_removed => TRUE);
	
	if ($class->{_boot_mode} eq C_ONTAP_CMODE) {
		# No cdefs for file restore yet. Send comand directly
		# If still no cdefs in the future, we should probably send the command directly for 7mode too
		my $cmd_code = (defined($restore_as_path)) ? C_CMD_SFSR_TO_PATH_START : C_CMD_SFSR_START;
		my $vol_name = $class->{volume}->name();
		my $vol_mount_path = $class->_get_vol_mount_path(); 
	
		my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
							       volume => $vol_name,
							       vserver => $class->{_vserver},
							       snapshot => $snapshot,
							       file_path => $filename,
							       restore_path => $restore_as_path);
		
		my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
								    boot_mode => $class->{_boot_mode},
								    cmd => $command,
								    timeout => C_DFLT_SFSR_TIMEOUT,
								    node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
		
		my %sfsr_res = $class->{_ui}->process_output_string(code => $cmd_code,
								    output => $output,
								    expected_rc => $args{expected},
								    log_message => "SFSR of file $file_path from snapshot $snapshot",);
		
		$rc = $sfsr_res{rc};
		if ($rc != C_RC_SUCCESS) {
			return %sfsr_res;
		}
	}
	else {
		my $sfsr_output = snaprestore_file($class->{volume}->Filer->conn, $file_path, $snapshot, restore_as_path => $restore_as_path);
		# snaprestore_file should return 1 in case of success
		# in case of error, it dies (inside an eval block)
		if ((!defined($sfsr_output)) || ($sfsr_output != 1)) {
			return $class->{_ui}->process_output_string(code => C_CMD_SFSR_START,
								    output => $@,
								    expected_rc => $args{expected},
								    log_message => "Start SFSR for file $filename",);
		}
	}
	
	my $restored_filename = $filename;
	if ((defined($restore_as_path)) && ($restore_as_path ne $file_path)) {
		$restored_filename = basename($restore_as_path);
		# We are restoring to a different path, add the new file to $class->{files}
		$class->_new_file(filename => $restored_filename);
	}
	elsif ($file_already_exists == FALSE) {
		# If this was a deleted/moved file, re-add it to $class->{files}
		$class->_re_add_file(filename=>$filename);
	}
	
	if ($wait_for_completion == TRUE) {
		$class->wait_for_single_file_snap_restore(filename => $restored_filename,
							  restored_from_file => $filename,
							  snapshot => $snapshot);
	}
	
	sleep 3;
	
	$res{'rc'} = $rc;
	return %res;
}

## @name ::wait_for_single_file_snap_restore
##
## @summary 
##	Waits until a file is completely restored from a snapshot.
##
## @description
##	Waits until a file is completely restored from a snapshot.
##
## @return hash of status
##
## @arg filename Required: Name of file.
## @arg snapshot Required: Name of the snapshot from which the file is restored.
## @arg timeout Optional: Timeout in seconds to wait for the file to be restored. Default: 300.
##
## @examples
## %res = $src_vol->{dataset}->wait_for_single_file_snap_restore(filename=>$FILENAME, snapshot=>$SNAPSHOT);
##
sub wait_for_single_file_snap_restore
{
	my $class 	= shift;
	my (%args) = validate( @_,
		{
		 filename => { type => SCALAR },
		 snapshot => { type => SCALAR },
		 restored_from_file => { type => SCALAR|UNDEF, optional => 1 },
		 timeout => { type => SCALAR, regex => qr/^\d+$/, optional => 1, default => C_DFLT_SFSR_TIMEOUT },
                } );
	
	my $filename = $args{filename};
	my $restored_from_filename = (defined($args{restored_from_file})) ? $args{restored_from_file} : $filename;
	my $snapshot = $args{snapshot};
	my $timeout = $args{timeout};
	my $file_path;
	
	logcomment("Waiting for single file snap restore to complete:\n".
		   "Filename: $filename\n" .
		   "Restored from file: $restored_from_filename\n" .
		   "Snapshot: $snapshot\n" .
		   "Timeout: $timeout");
	
	$file_path = $class->_generate_file_path(filename => $filename);
	
	my $ss_file_size = $class->get_file_size(filename=>$restored_from_filename,
						 snapshot=>$snapshot,
						 allow_removed => TRUE);
	my $ss_timestamp = $class->get_timestamp(filename=>$restored_from_filename,
						 snapshot=>$snapshot,
						 allow_removed => TRUE);
	my $file_size = 0;
	my $timestamp = -1;
	my $sleep_interval = 5;
	
	while (($file_size != $ss_file_size) || ($timestamp != $ss_timestamp)) {
		if ( $class->file_exists(filename=>$filename) ) {
			$file_size = $class->get_file_size(filename=>$filename);
			$timestamp = $class->get_timestamp(filename=>$filename);
		}
		sleep($sleep_interval);
		$timeout -= $sleep_interval;
		
		if ($timeout < 0) {
			logresult('fatal', "Timed-out while waiting for single file snap restore to complete\n" .
				           "Filename: $filename\n" .
					   "Snapshot: $snapshot\n" .
					   "File size: $file_size\n" .
					   "Snapshot file size: $ss_file_size\n" .
					   "File timestamp: $timestamp\n" .
					   "Snapshot file timestamp: $ss_timestamp\n");
		}
	}
	
	return (rc => C_RC_SUCCESS);
}

## @name ::fill_remaining_space
##
## @summary 
##	
##
## @description
##	
##
## @return hash of status
##
## @arg filename	Required: name of file
## @arg leave_space	Optional: amount of available space to leave ( currently to the nearest 4KB, default is zero )
##
## @examples
##	%res = $src_vol->{dataset}->fill_remaining_space(filename=>$FILENAME);
sub fill_remaining_space
{
	my $class 	= shift;
	my (%args) = validate (@_,  {	filename => { type=>SCALAR },
									leave_space	=> { type=>SCALAR, optional=>1, default=>0},
	} );
	my $filename = $args{filename};
	my $leave_space = $args{leave_space};
	
	my $vol = $class->{volume}; 
	my $avail = $vol->avail * KB;

	$class->_new_file(	filename=>$filename,);
	my $path = $class->{_common}->{files}->{$filename}->{path};
	
	logcomment("fill_remaining_space\n$path\navail: $avail\nleave_space:$leave_space");



	my $client_name = $class->{client_name};
	my $st = subtest("lib/Compression/Dataset/CIFS/Scripts/fill_remaining_space.thpl",
						"PATH=$path","AVAIL=$avail","LEAVE_SPACE=$leave_space",
			'-client',$client_name
	);
}

## @name ::get_data_range
##
## @summary get specified range of content from a file
##	
##
## @description get specified range of content from a file
##	
## @return data content
##
## @arg filename	Required: name of file
## @arg leave_space	Optional: amount of available space to leave ( currently to the nearest 4KB, default is zero )
##
## @examples
##			my $data = $class->get_data_range(	path	=> $path,
##												offset	=> $offset,
##												len		=> $len,
##			);
sub get_data_range
{
	my $class 	= shift;
	my (%args) = validate (@_,  {	path		=> { type=>SCALAR },
									offset		=> { type=>SCALAR, optional=>1, default=>0 },
									len			=> { type=>SCALAR, optional=>1, default=>0 },
	} );
	my $path		= $args{path};
	my $offset		= $args{offset};
	my $len			= $args{len};


	my $client_name = $class->{client_name};
	my $st = subtest("lib/Compression/Dataset/CIFS/Scripts/get_data_range.thpl",
						"PATH=$path","OFFSET=$offset","LEN=$len",
			'-client',$client_name
	);

	my $data = $st->message_get(type => "data");
	return $data;
}

## @name ::display_file_contents
##
## @summary display the contents of a file
##
## @description
## display contents of file.  This is more just a utility for 
## debugging/visual inspection.
##
## @return none
##
## @arg	Required:	filename
##
## @examples 
##	$dataset->display_file_contents( filename=>'f1' );
##
sub display_file_contents
{
	my $class   = shift;
	my (%args) = validate (@_, { filename => { type=>SCALAR },
								offset	 => { type=>SCALAR, optional=>1 },
								len		 => { type=>SCALAR, optional=>1 },								
	} );
	my $filename = $args{filename};
	$class->_filename_exists( filename=>$filename, expected=>TRUE );
	my $file_size = $class->get_file_size(filename=>$filename);
	my $offset = $args{offset} || 0;
	my $len = $args{len} || $file_size;
	my $file_path = $class->{_common}->{files}->{$filename}->{path};
	logcomment("display_file_contents: $filename\n".
				"$file_path\n".
				"offset: $offset\n".
				"len: $len\n");

	my $client_name = $class->{client_name};
	my $bg = subtest("lib/Compression/Dataset/CIFS/Scripts/display_file_contents.thpl",
						"PATH=$file_path",
						"OFFSET=$offset",
						"LEN=$len",
			'-client',$client_name
	);
}


## @name ::write_bytes_cifs
##
## @summary write_bytes_cifs
##
## @description
## write_bytes_cifs
##
## @return none
##
## @arg	Required:	filename
##
## @examples 
##	
##
sub write_bytes_cifs {
	my $class = shift;
	my (%args) = validate( @_,
	{
        path => {type => SCALAR,},
		offset => {type => SCALAR, optional=>1, default=>0},
		len => {type => SCALAR,},
		pattern => {type => SCALAR, optional=>1, default=>'a'},
     } );
	my $client_name = $class->{client_name};
	
	logcomment("write_bytes_cifs");
	my $bg_map = subtest("lib/Compression/Dataset/CIFS/Scripts/write_bytes.thpl",
						"PATH=$args{path}", "OFFSET=$args{offset}",
						"LEN=$args{len}", "PATTERN=$args{pattern}",
			'-client',$client_name
	);
}

## @name ::create_file_cifs
##
## @summary create_file_cifs
##
## @description
## create_file_cifs
##
## @return none
##
## @arg	Required:	filename
##
## @examples 
##	
##
sub create_file_cifs {
	my $class = shift;
	my (%args) = validate( @_,
	{
        path => {type => SCALAR,},
		offset => {type => SCALAR, optional=>1, default=>0},
		len => {type => SCALAR|UNDEF, optional=>1, default=>UNDEF},
		pattern => {type => SCALAR, optional=>1, default=>'a'},
		bg => {type => SCALAR, optional=>1, default=>FALSE},
		data => {type => SCALAR|UNDEF, optional=>1, default=>UNDEF},
		type => {type => SCALAR, optional=>1, regex => qr/^(write_bytes|existing_data|uncompressable)$/, default=>'write_bytes'},
     } );
	my $client_name = $class->{client_name};
	my $runid = 'create_file_'. basename( $args{path} );
	my $len = $args{len} || '0';
	my $type = $args{type};

	logcomment("create_file_cifs");

	if (( ! defined $args{len} )&&( ! defined $args{data} )) {
		logresult("script", "need to specify either len or data");
	}

	my $subtest;
	if ($args{'bg'} == TRUE) {
		$subtest = subtest("lib/Compression/Dataset/CIFS/Scripts/create_file.thpl",
							"PATH=$args{path}", "OFFSET=$args{offset}",
							"LEN=$len", "PATTERN=$args{pattern}",
							"TYPE=$type",
				'-client',$client_name,
				'-bg',
				'-runid',$runid,	
		);
	} else {
		$subtest = subtest("lib/Compression/Dataset/CIFS/Scripts/create_file.thpl",
							"PATH=$args{path}", "OFFSET=$args{offset}",
							"LEN=$len", "PATTERN=$args{pattern}",
							"TYPE=$type",
				'-client',$client_name,
				'-bg',
		);
	}

	if ( $type eq 'existing_data' ) {
		logcomment("sending data to cifs client");
		$subtest->message_put(message=>$args{data}, type => "existing_data");		
	}
	$subtest->wait_finish();
	return $subtest;
}

## @name ::modify_file_cifs
##
## @summary modify_file_cifs
##
## @description
## modify_file_cifs
##
## @return none
##
## @arg	Required:	filename
##
## @examples 
##	
##
sub modify_file_cifs {
	my $class = shift;
	my (%args) = validate( @_,
	{
        path => {type => SCALAR,},
		offset => {type => SCALAR, optional=>1, default=>0},
		len => {type => SCALAR,},
		pattern => {type => SCALAR, optional=>1, default=>'b'},
		bg => {type => SCALAR, optional=>1, default=>FALSE},
		data => {type => SCALAR|UNDEF, optional=>1, default=>UNDEF},
		type => {type => SCALAR, optional=>1, regex => qr/^(write_bytes|existing_data|uncompressable|holey)$/, default=>'write_bytes'},
		shift => {type => SCALAR, optional=>1, default=>FALSE},
     } );
	my $client_name = $class->{client_name};
	my $runid = 'create_file_'. basename($args{path});
	my $type = $args{type};

	logcomment("modify_file_cifs");
	my $subtest;
	if ($args{'bg'} == TRUE) {
		$subtest = subtest("lib/Compression/Dataset/CIFS/Scripts/modify_file.thpl",
							"PATH=$args{path}", "OFFSET=$args{offset}",
							"LEN=$args{len}", "PATTERN=$args{pattern}",
							"TYPE=$args{type}", "SHIFT=$args{shift}",
				'-client',$client_name,
				'-bg',
				'-runid',$runid,	
		);
	} else {
		$subtest = subtest("lib/Compression/Dataset/CIFS/Scripts/modify_file.thpl",
							"PATH=$args{path}", "OFFSET=$args{offset}",
							"LEN=$args{len}", "PATTERN=$args{pattern}",
							"TYPE=$args{type}", "SHIFT=$args{shift}",
				'-client',$client_name,
				'-bg',
		);
	}

	if ( $type eq 'existing_data' ) {
		logcomment("sending data to cifs client");
		$subtest->message_put(message=>$args{data}, type => "existing_data");		
	}
	$subtest->wait_finish();
	return $subtest;
}


## @name ::get_mountpoint
##
## @summary return mount point of volume
##
## @description
## wrapper around $vol->mountpoint
##
## @return $vol->mountpoint
##
## @examples
##  my $mountpoint = $dataset->get_mountpoint;
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_mountpoint
{
	my $class 	= shift;
	return $class->get_cifs_share_path_local;
}

## @name ::get_filer_path
##
## @summary return path to file on filer
##
## @description
## return path to file on filer
##
## @return the filer's path to the file ( e.g. /vol/vol1/f1 )
##
## @arg	filename Required: name of file
##
## @examples
##  my $filer_path = $dataset->get_filer_path( filename=>'f1' );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_filer_path
{
	my $class 	= shift;
	my (%args) = validate (@_, { filename => { type=>SCALAR },} );
	my $filename = $args{filename};
	$class->_filename_exists( filename=>$filename, expected=>TRUE );
	my %res;
	return $class->{_common}->{files}->{$filename}->{filer_path};
}

## @name ::replace_file
##
## @summary replace file with some specified file
##
## @description
## replace file with some specified file
##
## @return hash with success code
##
## @arg filename	Required:	name of file to replace
## @arg new_file	Required:	path to new file
##
## @example
##  $dataset->replace_file( filename=>'f1', new_file=>'path_to_new_file' );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub replace_file
{
	my $class 	= shift;
	my (%args) = validate (@_, { filename => { type=>SCALAR },
								 new_file => { type=>SCALAR },} );
	my %res;
	my $filename = $args{filename};
	my $new_file = $args{new_file};
	$class->_filename_exists( filename=>$filename, expected=>TRUE );
	my $file_path = $class->{_common}->{files}->{$filename}->{path};

	my $client_name = $class->{client_name};
	my $bg = subtest("lib/Compression/Dataset/CIFS/Scripts/replace_file.thpl",
						"PATH=$file_path","NEW_FILE_PATH=$new_file",
						'-client',$client_name
	);

	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::ls
##
## @summary ls
##
## @description
## ls
##
## @example
##  $dataset->ls();
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub ls
{
	my $class 	= shift;
	logcomment("Check that mount is good");
	my $path = $class->get_mountpoint();
	my $client_name = $class->{client_name};
	my $subtest = subtest("lib/Compression/Dataset/CIFS/Scripts/ls_dir.thpl",
					"PATH=$path",
					'-client',$client_name
	);

}

## @name ::wait_finish
##
## @summary wait for background process to finish
##
## @description
## wait for background process to finish from create,
##  modify, copy_file, or move_file
##
## @return hash of status
## 
## @examples
##  $dataset->wait_finish( filename=>'f1' );
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub wait_finish
{
	my $class 	= shift;
	my (%args) = validate (@_, { filename => { type=>SCALAR },} );
	my %res;
	my $filename = $args{filename};

	logcomment("waiting for $filename background process to complete");
	$class->{_common}->{files}->{$filename}->{'bg_subtest'}->wait_finish();
	logcomment("$filename bg_subtest has finished");

	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}


## @name ::is_alive
##
## @summary Checks if a background process is alive.
##
## @description
## Checks if a background process is alive,
## modify, copy_file, or move_file
##
## @return Boolean value: 0/1
## 
## @examples
##  if ($dataset->is_alive(filename=>'f1')) {logcomment("The background process is alive");}
##
## @author dl-compression-qa@netapp.com
##
sub is_alive
{
        my $class = shift;
	my (%args) = validate (@_, {filename => { type=>SCALAR},} );
	
	my $filename = $args{filename};
	
	return $class->{_common}->{files}->{$filename}->{'bg_subtest'}->is_alive();
}


## @name ::mk_dir
##
## @summary Make a directory in volume
##
## @description
## Make a directory in volume
##
## @return hash of status
## 
## @arg	name	Required: name of directory
## @arg path	Optional: complete path to the directory the folder will be created in
##
## @examples
##
##	%res = $dataset->mkdir(
##							name => 'd1',
##						 );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub mk_dir
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		name => { type=>SCALAR },
		path => { type=>SCALAR, optional=>1 },
	} );
	
	my $dir_name = $args{name};
	my $vol_mountpoint = $class->get_mountpoint();
	my $dir_path = (defined $args{path}) ? ($args{path} .'\\'. $dir_name) : ($vol_mountpoint .'\\'. $dir_name);
	my %res;
	
	my $client_name = $class->{client_name};
	my $bg = subtest("lib/Compression/Dataset/CIFS/Scripts/mk_dir.thpl",
						"PATH=$dir_path",
				'-client',$client_name
	);
	
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::truncate_file
##
## @summary truncate file to specified size
##
## @description
## truncate file to specified size
##
## @return hash of status
##
## @arg	filename	Required: name of file to truncate
## @arg size		Required: size to truncate file to
##
## @examples
##  $dataset->truncate_file( filename=>'f1', size=>1024 );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub truncate_file
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename => { type=>SCALAR },
		size => { type=>SCALAR },
	} );
	
	my %res;
	my $filename = $args{filename};
	my $size = $args{size};

	my $client_name = $class->{client_name};
	my $file_path = $class->{_common}->{files}->{$filename}->{path};
	my $bg = subtest("lib/Compression/Dataset/CIFS/Scripts/truncate_file.thpl",
						"PATH=$file_path","SIZE=$size",
			'-client',$client_name
	);

	$res{'rc'} = C_RC_SUCCESS;
	if ($class->{_debug}) { $class->ls(); }
	return %res;
}

## @name ::file_exists
##
## @summary does file exist
##
## @description
## does file exist? returns TRUE or FALSE
##
## @return hash of status
##
## @arg	filename	Required: name of file 
##
## @examples
##  my $result = $dataset->file_exists( filename=>'f1' );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub file_exists
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename => { type=>SCALAR },
	} );
	
	my %res;
	my $filename = $args{filename};

	my $client_name = $class->{client_name};
	my $file_path = $class->{_common}->{files}->{$filename}->{path};
	my $st = subtest("lib/Compression/Dataset/CIFS/Scripts/file_exists.thpl",
						"PATH=$file_path",
			'-client',$client_name
	);
	my $file_exists = $st->message_get(type => "file_exists");

	return $file_exists;
}

## @name ::get_vbn_constants
##
## @summary get the vbn/vvbn constants
##
## @description
## get the vbn/vvbn constants
##
## @return hash of status and output
##	
## 
## @example
##	%res = $dataset->get_vbn_constants();
##  if (grep(/^$x$/, @{$res{vbn_headers}} ) ) {
##		logcomment("$x is a valid vbn header");
##	}
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_vbn_constants {
	my $class = shift;
	return $class->{_common}->get_vbn_constants(@_);
}

# @name ::get_file_atime
#
# @summary Gets the atime of the file (on the client).
#
# @description
# Gets the atime of the file (on the client). 
#
# @return Hash with rc, atime (# of seconds since the start of the epoch), gmt_atime
#
# @arg filename Required: Name of file.
#
# @example
#  $dataset->get_file_atime(filename => 'my_file');
#
# @author voicu@netapp.com, dl-compression-qa@netapp.com
#
sub get_file_atime
{
        my $class = shift;
	
	logresult('script', "Snaplock support is not available for CIFS yet");
}

# @name ::set_file_atime
#
# @summary Sets the atime of the file (on the client).
#
# @description
# Sets the atime of the file (on the client).
#
# @arg filename Required: Name of file.
#
# @arg new_atime Required: New atime to set (# of seconds since the start of the epoch).
#
# @return The return of the get_file_atime method, after the new atime has been set.
#
# @example
# my %res = $compress->{dataset}->get_file_atime(filename => 'my_file');
# my $new_atime = $res{atime} + 30;
# $dataset->set_file_atime(filename => 'my_file', $new_atime);
#
# @author voicu@netapp.com, dl-compression-qa@netapp.com
#
sub set_file_atime
{
        my $class = shift;
	
	logresult('script', "Snaplock support is not available for CIFS yet");
}


################################################################
# Internal subroutines
################################################################

# Internal function
# name ::_new_file
#
# summary (INTERNAL) create a new file object
#
# description
# Creates a new file object.
# Doesn't create actual file.
#
# example
#  $dataset->_new_file( filename=>'f2' );
# 
# File objects contain the folling:
# $file->{path}			= full path on client
# $file->{filer_path}		= path for filer commands
# $file->{relative_path}	= path relative to volume
# 
# note: both filer_path and relative_path are only for
# commands run on filer
sub _new_file
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		filename => { type=>SCALAR },
		testdir => { type=>SCALAR|UNDEF, optional=>1 },
		path => { type=>SCALAR|UNDEF, optional=>1 },
	} );
	my $file;
	my $filename = $args{filename};
	my $vol_mountpoint = $class->get_mountpoint();
	my $testdir_path  = ( defined $args{path} )? $args{path} : $vol_mountpoint;
	$class->{_common}->{testdir} = $args{testdir};

	if (defined $args{testdir}) {
		$testdir_path .= '\\' . $class->{_common}->{testdir};
	}
	$class->{_common}->{testdir_path} = $testdir_path;

	$file->{name} = $filename;
	$file->{path} = $testdir_path .'\\'. $file->{name};

	$file->{relative_path} = $file->{path};
	$file->{relative_path} =~ s#^$vol_mountpoint\\##;
	$file->{relative_path} =~ s#\\#/#g;
	$file->{filer_path} = $class->_get_vol_mount_path() .'/'. $file->{relative_path};

	if ( $file->{path} !~ /$vol_mountpoint/ ) {
		# file is not on a volume
		$file->{relative_path} = undef;
		$file->{filer_path}    = undef;
	}

	my $comment = "creating new file object for $filename\n";
	$comment .= "path: ". $file->{path} ."\n";
	$comment .= "relative_path: ". $file->{relative_path} ."\n";
	$comment .= "filer_path: ". $file->{filer_path} ."\n";
	logcomment($comment);

	$class->{_common}->_new_file(file=>$file);
}

sub _get_vol_mount_path
{
	my $class 	= shift;
	return $class->{volume}->mount_path();
}

# Internal function
# name ::_filename_exists
#
# summary (INTERNAL) Checks that file exists in the dataset object
#
# description
# Used in error checking by various functions
#
# example
#		$class->_filename_exists( filename=>$filename, expected=>FALSE ); 
#
sub _filename_exists
{
	my $class = shift;
	return $class->{_common}->_filename_exists(@_);
}

# Internal function
# name ::_remove_file
#
# summary (INTERNAL) Moves the file entry from the $class->{files} hash
#  into the $class->{_removed_files} one.
#
# example
#		$class->_remove_file( filename=>$filename ); 
#
sub _remove_file
{
	my $class = shift;
	return $class->{_common}->_remove_file(@_);
}

# Internal function
# name ::_generate_file_path
#
# summary (INTERNAL)
# If snapshot is defined, generates the complete path of a file within a snapshot.
# If snapshot is not defined, returns the file path.
#
# description
# Generates the complete path of a file (within a snapshot, if required).
#
# example
#  my $snapshot_file_path = $dataset->_generate_file_path(filename=>'f2',
#                                                         snapshot => 'my_snapshot');
# 
#
#
sub _generate_file_path
{
	my $class = shift;
	return $class->{_common}->_generate_file_path(@_);
}



# Internal function
# name ::_generate_relative_path
#
# summary (INTERNAL)
# If snapshot is defined, generates the relative path of a file within a snapshot.
# If snapshot is not defined, returns the relative file path.
#
# description
# Generates the relative path of a file (within a snapshot, if required).
#
# example
#  my $relative_path = $dataset->_generate_relative_path(filename=>'f2',
#                                                        snapshot => 'my_snapshot');
# 
#
#
sub _generate_relative_path
{
	my $class = shift;
	return $class->{_common}->_generate_relative_path(@_);
}

# Internal function
# name ::_generate_full_path
#
# summary (INTERNAL)
# This is same as _generate_file_path except returns full path as seen when mounted on a client.
# If snapshot is defined, generates the complete path of a file within a snapshot.
# If snapshot is not defined, returns the file path.
#
# description
# Generates the complete path of a file (within a snapshot, if required).
#
# example
#  my $snapshot_full_path = $dataset->_generate_full_path(filename=>'f2',
#                                                         snapshot => 'my_snapshot');
# 
#
#
sub _generate_full_path
{
	my $class 	= shift;
	my (%args) = validate (@_,
			       { 
					filename => { type=>SCALAR },
					snapshot => { type=>SCALAR|UNDEF, optional=>1 },
					allow_removed => { type=>BOOLEAN|UNDEF, optional=>1, },
				} );
	
	my $filename = $args{filename};
	my $snapshot = $args{snapshot};
	my $allow_removed = $args{allow_removed};
	my $file_path;
	
	if (!defined($allow_removed)) {
		$allow_removed = (defined($snapshot)) ? TRUE : FALSE;
	}
	
	my $file_already_exists = $class->_filename_exists( filename=>$filename, expected=>C_RC_FAILURE_NON_FATAL );
	if ($file_already_exists == FALSE) {
		if ($allow_removed == FALSE) {
			logresult('fatal', "File $filename does not exist");
		}
		$class->_file_was_removed(filename=>$filename);
		$file_path = $class->{_common}->{_removed_files}->{$filename}->{path};
	}
	else {
		$file_path = $class->{_common}->{files}->{$filename}->{path};
	}
	
	if (defined($snapshot)) {
		my $vol_mountpoint = $class->get_mountpoint;
		$file_path =~ s#^$vol_mountpoint#$vol_mountpoint\\.snapshot\\$snapshot#;
	}
	return $file_path;
}

# Internal function
# name ::_file_was_removed
#
# summary (INTERNAL) Checks that file was ever known and has been removed by the dataset object
#
# example
#		$class->_file_was_removed( filename=>$filename ); 
#
sub _file_was_removed
{
	my $class = shift;
	return $class->{_common}->_file_was_removed(@_);
}

# Internal function
# name ::_re_add_file
#
# summary (INTERNAL) Re-adds the file hash into $class->{files}.
#
# example
#		$class->_re_add_file( filename=>$filename ); 
#
sub _re_add_file
{
	my $class = shift;
	return $class->{_common}->_re_add_file(@_);
}

# internal function currently used by Compression::SnapVault::restore
# for the case when restoring to a different qtree
sub _replace_qtree_name {
	my $class = shift;
	my (%args) = validate( @_,
	{
        old_qtree_name => {type => SCALAR,},
		new_qtree_name => {type => SCALAR,},
     } );
	
	my $old_qtree_name = $args{'old_qtree_name'};
	my $new_qtree_name = $args{'new_qtree_name'};
	my $files = $class->{_common}->{files};
	my $mountpoint = $class->get_mountpoint();
	my $vol = $class->{volume};
	my $vol_mount_path = $vol->mount_path;

	foreach my $file (keys %$files)
	{
		$files->{$file}->{relative_path} =~ s/^$old_qtree_name/$new_qtree_name/;
		$files->{$file}->{path} =~ s#^$mountpoint\\$old_qtree_name#$mountpoint\\$new_qtree_name#;
		$files->{$file}->{filer_path} =~ s#^$vol_mount_path\\$old_qtree_name#$vol_mount_path\\$new_qtree_name#;
	}
}

# internal function currently used by Compression::SnapMirror::_update_dataset_hash
sub _get_qtree_file_hash {
	my $class = shift;
	my (%args) = validate( @_,
	{
        qtree_name => {type => SCALAR,},
    } );
	my $qtree_name = $args{'qtree_name'};
	my $files = $class->{_common}->{files};

	my $qtree_files = {};
	foreach my $file (keys %$files)
	{
		if ($files->{$file}->{relative_path} =~ /^$qtree_name/ ) {
			$qtree_files->{$file} = $files->{$file};
		}
	}
	return $qtree_files;
}

sub _update_vol_obj
{
	my $class = shift;
	
	logresult('script', "This method is not currently supported for CIFS");
}


1;
