#
# $Id: //depot/prod/test/main/lib/Compression/Dataset/Common.pm $
#
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)


package Compression::Dataset::Common;

use strict;
use warnings;

use Tharn;

use Module::Common;
use Data::Dumper;
use FS::Util::Utilities qw(file_reserve dumpblock);
use Dense::Utils;
use Dense::TestData;
use cat qw(do_shell);
use File::Copy;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Compression::Constants qw(:all);
use Compression::Utils qw(_send_cmd_to_filer);
use File::Compare;
use File::Basename;
use Carp qw(croak);
use FileManip;
use filer;
use Compression::Dataset::Constants qw(:all);
use Digest::SHA qw(sha1);
use SnapCmds;

BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}


## @name ::new
##
## @summary perform dataset initialization
##
## @description
## Perform all the basic dataset initialization required here
##
## @return Compression::Dataset object
##
## @arg volume Required: NOTE Volume object
## @arg debug Optional: logs additional debug messages specific to this module
##
## @examples
## 	$self->{dataset}  = Compression::Dataset->new(volume => $self->{volume}->{_vol},
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
##

sub new
{
	my $class 	= shift;
	my (%args) = validate (@_, {volume    => { type=>OBJECT, },
				    vserver   => { type=>SCALAR|UNDEF, optional => 1 },
				    debug     => { type=>SCALAR, optional => 1, default=>FALSE },
				    ui        => {type => OBJECT, isa => "Compression::UI"},
				    boot_mode => {type => SCALAR, optional => 1, default => C_ONTAP_BOOT_MODE},
				    has_debug_kernel => {type => SCALAR, optional => 1, default => TRUE}, # So that scripts can continue
				    vfiler_mountpoint => {type => SCALAR|UNDEF, optional => 1,}, 
		} );	
	my $self = {};
	$self->{ID} = $$;
	$self->{Description} = "Dataset::Common";
	$self->{volume} = $args{volume};
	$self->{_vserver} = $args{vserver};
	$self->{_ui} = $args{ui};
	$self->{_debug} = $args{debug};
	$self->{ds_type} = C_DFLT_COMPRESSION_DATASET;
	$self->{_has_debug_kernel} = $args{has_debug_kernel};
		
	$self->{_boot_mode} = $args{boot_mode};
	if ($self->{_boot_mode} eq C_ONTAP_7MODE) 
	{
		if (!($self->{volume}->isa("NOTE::Volume")))
		{
			logresult('script', "For 7-mode, the volume argument is required to be a NOTE\:\:Volume object");
		}
		$self->{_conn} = $self->{volume}->Filer();
		$self->{vol_name} = $self->{volume}->name;
		
	}
	elsif ($self->{_boot_mode} eq C_ONTAP_CMODE) 
	{
		if (!($self->{volume}->isa("NOTE::GX::Volume")))
		{
			logresult('script', "For 10-mode, the volume argument is required to be a NOTE\:\:GX\:\:Volume object");
		}
		$self->{_conn} = $self->{volume}->Node();
		$self->{vol_name} = $self->{volume}->name;
		

	}
	else
	{
		logresult('script', "Unknown boot_mode: " . $self->{_boot_mode});
	}
	
	$self->{files} = {};
	$self->{_removed_files} = {};
	$self->{_lun_clones} = {};
	$self->{sfsr_will_restore_file_as_compressed} = $self->{_ui}->get_constant_value( constant_code => 'SFSR_WILL_RESTORE_FILE_AS_COMPRESSED' );
	
	bless $self, $class;
	
	logcomment("Dataset::Common Initialized with ". Dumper($self) ) if ( $self->{_debug} );

	return $self;

}

## @name ::get_path
##
## @summary return path to file
##
## @description
## return path to file
##
## @return full path to the file
##
## @arg	filename Required: name of file.
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset.
##
## @examples
##  my $filepath = $dataset->get_path( filename=>'f1' );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_path
{
	my $class 	= shift;
	my (%args) = validate (@_,
		{
			filename => {type => SCALAR},
			allow_removed => {type => BOOLEAN, optional => 1, default => FALSE},
			clone_lun => {type => SCALAR|UNDEF, optional => 1,},
		});
	
	my $filename = $args{filename};
	my $allow_removed = $args{allow_removed};
	my $clone_lun = $args{clone_lun};
	my $path;
	
	my $file_already_exists = $class->_filename_exists(filename => $filename, expected => C_RC_FAILURE_NON_FATAL, clone_lun => $clone_lun);
	if ($file_already_exists == FALSE) {
		if ($allow_removed == FALSE) {
			logresult('fatal', "get_path: File $filename does not exist");
		}
		$class->_file_was_removed(filename=>$filename, clone_lun => $clone_lun);
		$path = (defined($clone_lun)) ? ($class->{_lun_clones}->{$clone_lun}->{removed_files}->{$filename}->{path}) : $class->{_removed_files}->{$filename}->{path};
	}
	else {
		$path = (defined($clone_lun)) ? ($class->{_lun_clones}->{$clone_lun}->{files}->{$filename}->{path}) : $class->{files}->{$filename}->{path};
	}
	
	
	if (!($path)) {
		croak "Path doesn't exist for $filename: $!";
	}

	return $path;		
}


## @name ::dump_file_info
##
## @summary display paths associated with file
##
## @description
## Just for debugging purposes.
##
## @return none
## 
## @example
##		$dataset->dump_file_info()
##
## displays something like:
## 
##		Dumping file info for f5
##		relative_path: xx/f5
##		filer_path: /vol/c1/xx/f5
##		path: /t/172.18.47.30/vol/c1/shared/tcp_v3/xx/f5
##
sub dump_file_info
{
	my $class 	= shift;
	foreach my $file ( sort keys %{$class->{files} } ) {
		my $text = "Dumping file info for $file\n";
		if ( $class->{files}->{$file}->{relative_path} ) {
			$text .= "relative_path: $class->{files}->{$file}->{relative_path}\n";
		} else {
			$text .= "relative_path: none\n";
		}
		if ($class->{files}->{$file}->{filer_path}) {
			$text .= "filer_path: $class->{files}->{$file}->{filer_path}\n";
		} else {
			$text .= "filer_path: none\n";
		}
		if ($class->{files}->{$file}->{path}) {
			$text .= "path: $class->{files}->{$file}->{path}\n";
		} else {
			$text .= "path: none"
		}
		logcomment($text);
	}
}

## @name ::get_vbn_hash
##
## @summary get vbn hash for some file
##
## @description
## get vbn hash for some file
##
## @arg filename Required: File name.
## @arg snapshot Optional: Name of the snapshot, if file is in snapshot.
## @arg fbn      Optional: fbn to compare ( or fbn range e.g. 8-16 )
## @arg panic_expected Optional: Flag to be set if the command is expected to panic the filer.
##                     In this case, no output matching is done, but the method returns with C_RC_SUCCESS 
##                     right after issuing the command. Default value: FALSE.
##                     Note: this flag is not considered if "fbn" is defined.
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Ignored if fbn is passed in or if panic_expected is TRUE.
## Default: C_RC_SUCCESS.
##
##
## @return hash of status and vbn hash
## 
## @examples
##  my %res = $dataset->get_vbn_hash( filename=>'f1' );
##  my $vbns_info = $res{'result'};
##  logcomment("vbn for fbn 3 is "	. $vbns_info->{3}->{vbn} );
##  logcomment("vvbn for fbn 3 is "	. $vbns_info->{3}->{vvbn} );
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub get_vbn_hash
{
	my $class 	= shift;
	my (%args) = validate( @_,
		{
		 filename => { type=>SCALAR },
		 snapshot => { type=>SCALAR|UNDEF, optional=>1 },
		 fbn	  => { type=>SCALAR, optional=>1 },
		 panic_expected => { type=>SCALAR, optional=>1, default => FALSE },
		 expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
                } );
	
	my %res;
	my $filename = $args{filename};
	my $fbn = $args{fbn};
	my $vbns_info;
	
	if ( defined($fbn) ) {
		my ( $fbn_start, $fbn_end );
		if ( $fbn =~ /^(\d+)$/ ) {
			$fbn_start = $1;
			$fbn_end = $1;
		} elsif ( $fbn =~ /^(\d+)\s*-\s*(\d+)$/ ) {
			$fbn_start = $1;
			$fbn_end = $2;
		} else {
			logresult('script', "unknown fbn range $!");
		}
		for (my $fbn = $fbn_start; $fbn <= $fbn_end; $fbn++) {
			my %fbn_info = $class->_get_vbn_info_single(filename => $filename,
								    snapshot => $args{snapshot},
								    fbn => $fbn);
			$vbns_info->{$fbn}->{vbn} = $fbn_info{vbn};
			$vbns_info->{$fbn}->{vvbn} = $fbn_info{vvbn};
		}
	} else {
		$vbns_info = $class->_get_vbn_info_all(filename => $filename,
						       snapshot => $args{snapshot},
						       panic_expected => $args{panic_expected},
						       expected => $args{expected},);
	}

	$res{'result'} = $vbns_info;
	$res{'rc'} = C_RC_SUCCESS;
	
	return %res;
}

## @name ::compare_vbn_hashes
##
## @summary Compare the vbn hashes for two files.
##
## @description
## Compare the vbn hashes for two files.
##
## @arg file_1 Required: It can be either a string, representing the name of the first file,
## or a reference to an already obtained vbn hash.
##
## @arg file_2 Required: It can be either a string, representing the name of the second file,
## or a reference to an already obtained vbn hash.
##
## @arg snapshot_1 Optional: Name of the snapshot, if file_1 is in a snapshot.
##
## @arg snapshot_2 Optional: Name of the snapshot, if file_2 is in a snapshot.
##
## @arg expected Optional: EQUAL|NOT_EQUAL. Default: 'EQUAL'.
##
## @arg compare_field Optional: all|vbn|vvbn default is 'all'
##
## @return hash of rc and vbn hashes that were compared.
## 
## @examples
##  $dataset->compare_vbn_hashes(file_1 => $file,
##                               file_2 => $file,
##                               snapshot_2 => $snapshot_name);
##
## @author dl-compression-qa@netapp.com, voicu@netapp.com
##
sub compare_vbn_hashes
{
	my $class 	= shift;
	my (%args) = validate( @_,
		{
		 file_1		=> { type=>SCALAR|HASHREF },
		 file_2		=> { type=>SCALAR|HASHREF },
		 snapshot_1 => { type=>SCALAR|UNDEF, optional => 1 },
		 snapshot_2 => { type=>SCALAR|UNDEF, optional => 1 },
		 expected	=> { type=>SCALAR, default=>'EQUAL', regex=> qr/^(EQUAL|NOT_EQUAL)$/ },
		 compare_field => { type=>SCALAR, optional => 1, regex=>qr/^(all|vbn|vvbn)$/, default=>'all' },
                } );
	
	my $expected = $args{expected};
	my @compare_fields = ($args{compare_field});
	my $file_1 = $args{file_1};
	my $file_2 = $args{file_2};
	my %return_hash = ();
	$return_hash{rc} = C_RC_SUCCESS;
	
	my $vbns_file_1;
	if (ref($file_1) eq "HASH")
	{
		$vbns_file_1 = $file_1;
	}
	else
	{
		my %res = $class->get_vbn_hash(filename => $args{file_1}, snapshot => $args{snapshot_1});
		$vbns_file_1 = $res{result};
	}
	$return_hash{vbns_file_1} = $vbns_file_1;
	
	my $vbns_file_2;
	if (ref($file_2) eq "HASH")
	{
		$vbns_file_2 = $file_2;
	}
	else
	{
		my %res = $class->get_vbn_hash(filename => $args{file_2}, snapshot => $args{snapshot_2});
		$vbns_file_2 = $res{result};
	}
	$return_hash{vbns_file_2} = $vbns_file_2;
	
	logcomment("Starting to compare the vbn hashes:\n" .
		   "########## vbn hash of file_1:\n" . Dumper($vbns_file_1) . "\n" .
		   "########## vbn hash of file_2:\n" . Dumper($vbns_file_2));
	
	# First see if elements count is the same
	my $fbns_file_1 = keys %$vbns_file_1;
	my $fbns_file_2 = keys %$vbns_file_2;
	
	if (($fbns_file_1 != $fbns_file_2))
	{
		if ($expected  eq 'EQUAL')
		{
			logresult('fatal', "The two vbn hashes have different element counts: $fbns_file_1 != $fbns_file_2");
		}
		else
		{
			return %return_hash;
		}
	}
	
	if ($args{compare_field} eq 'all')
	{
		my $first_fbn_hash = $vbns_file_1->{0};
		@compare_fields = keys %$first_fbn_hash;
	}
	
	my $found_unequal_values = 0;
	
	foreach my $fbn (sort keys %$vbns_file_1)
	{
		my $fbn_hash_file_1 = $vbns_file_1->{$fbn};
		my $fbn_hash_file_2 = $vbns_file_2->{$fbn};
		foreach my $field (@compare_fields)
		{
			if (!defined($fbn_hash_file_1->{$field}) || !defined($fbn_hash_file_2->{$field}))
			{
				logresult('fatal', "Unable to obtain $field value for fbn $fbn");
			}
			
			logcomment("FBN $fbn: $field-s: file_1=" . $fbn_hash_file_1->{$field} .
				   ", file_2=" . $fbn_hash_file_2->{$field});
			
			if ($fbn_hash_file_1->{$field} != $fbn_hash_file_2->{$field})
			{
				if ($expected  eq 'EQUAL')
				{
					logresult('fatal', "$field value mismatch at fbn $fbn: " .
						           $fbn_hash_file_1->{$field} . " != " . $fbn_hash_file_2->{$field});
				}
				else
				{
					return %return_hash;
				}
				$found_unequal_values++;
			}
		}
	}
	
	if (($expected  eq 'NOT_EQUAL') && ($found_unequal_values == 0))
	{
		logresult('fatal', "The two hashes are equal for all required fields");
	}

	return %return_hash;
}

## @name ::compare_vbns
##
## @summary compare vbns from 2 files
##
## @description
## for 2 files, given an fbn ( or range of fbns )
## this compares the vbn values.   Croaks if not expected result.
##
## @return hash of status
##
## @arg file_1	Required: first file to compare
## @arg file_2	Required: second file to compare
## @arg fbn		Optional: fbn to compare ( or fbn range e.g. 8-16 )
## @arg expected	Optional: EQUAL|NOT_EQUAL default is 'EQUAL'
##
## @examples
##  $dataset->compare_vbns( file_1		=>'f1', 
##							file_2		=>'f2', 
##							fbn			=>'0-31', # or '3'
##							expected	=>'EQUAL', or 'NOT_EQUAL'
##							);
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub compare_vbns
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		file_1   => { type=>SCALAR },
		file_2   => { type=>SCALAR },
		fbn      => { type=>SCALAR, optional=>1 },
		expected => { type=>SCALAR, default=>'EQUAL' },
	
	} );
	
	my %res;
	my $file_1   = $args{file_1};
	my $file_2   = $args{file_2};
	my $fbn      = $args{fbn};
	my $expected = $args{expected};

	if ( ! defined($fbn) ) {
		$class->compare_vbn_hashes(file_1 => $file_1,
					   file_2 => $file_2,
					   compare_field => 'vbn',
					   expected => $expected);
	  
	} else {
		my ( $fbn_start, $fbn_end );
		if ( $fbn =~ /^(\d+)$/ ) {
			$fbn_start = $1;
			$fbn_end = $1;
		} elsif ( $fbn =~ /^(\d+)\s*-\s*(\d+)$/ ) {
			$fbn_start = $1;
			$fbn_end = $2;
		} else {
			logresult('script', "Illegal fbn range: $fbn");
		}
		$class->_compare_vbns_range(file_1 => $file_1,
					    file_2 => $file_2,
					    fbn_1 => $fbn_start,
					    fbn_2 => $fbn_end,
					    expected => $expected,
					    );
	}
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::compare_vvbns
##
## @summary compare vvbns from 2 files
##
## @description
## for 2 files, given an fbn ( or range of fbns )
## this compares the vvbn values.  Croaks if not expected result.
##
## @return hash of status
##
## @arg file_1	Required: first file to compare
## @arg file_2	Required: second file to compare
## @arg fbn		Optional: fbn to compare ( or fbn range e.g. 8-16 )
## @arg expected	Optional: EQUAL|NOT_EQUAL default is 'EQUAL'
##
## @examples
##  $dataset->compare_vvbns(file_1		=>'f1', 
##							file_2		=>'f2', 
##							fbn			=>'0-31', # or '3'
##							expected	=>'EQUAL', or 'NOT_EQUAL'
##							);
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub compare_vvbns
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		file_1 => { type=>SCALAR },
		file_2 => { type=>SCALAR },
		fbn => { type=>SCALAR, optional=>1 },
		expected => { type=>SCALAR, default=>'EQUAL' },
	
	} );
	
	my %res;
	my $file_1		= $args{file_1};
	my $file_2		= $args{file_2};
	my $fbn			= $args{fbn};
	my $expected	= $args{expected};

	if ( ! defined($fbn) ) {
		$class->compare_vbn_hashes(file_1 => $file_1,
					   file_2 => $file_2,
					   compare_field => 'vvbn',
					   expected => $expected);
	} else {
		my ( $fbn_start, $fbn_end );
		if ( $fbn =~ /^(\d+)$/ ) {
			$fbn_start = $1;
			$fbn_end = $1;
		} elsif ( $fbn =~ /^(\d+)\s*-\s*(\d+)$/ ) {
			$fbn_start = $1;
			$fbn_end = $2;
		} else {
			croak "unknown fbn range $!";
		}
		
		$class->_compare_vvbns_range(file_1 =>$file_1,
					     file_2 =>$file_2,
					     fbn_1 =>$fbn_start,
					     fbn_2 =>$fbn_end,
					     expected	=>$expected,
					     );
	}
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::compare_vbns_within_file
##
## @summary compare vbns within a file
##
## @description
## for a file, given a range of fbns,
## this compares the vbn values.   Croaks if not expected result.
##
## @return hash of status
##
## @arg filename        Required: name of file
## @arg fbn             Optional: fbn range e.g. 8-16
## @arg expected        Optional: EQUAL|NOT_EQUAL default is 'EQUAL'
##		EQUAL means every vbn in this range is equal
##		NOT_EQUAL means 1 or more vbn is different than the other
##
## @examples
##  $dataset->compare_vbns_within_file( filename =>'f1', 
##                                      fbn =>'0-31',
##                                      expected =>'EQUAL', or 'NOT_EQUAL'
##                                    );
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub compare_vbns_within_file
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename  => { type=>SCALAR },
		fbn       => { type=>SCALAR,},
		expected  => { type=>SCALAR, default=>'EQUAL', regex => qr/^EQUAL|NOT_EQUAL$/ },
	
	} );
	
	my %res;
	my $filename = $args{filename};
	my $fbn = $args{fbn};
	my $expected = $args{expected};

	logcomment("compare_vbns_within_file:\n".
		   "file: $filename\n".
		   "fbn: $fbn\n".
		   "expected: $expected\n");

	my ( $fbn_start, $fbn_end );
	if ( $fbn =~ /^(\d+)\s*-\s*(\d+)$/ ) {
		$fbn_start = $1;
		$fbn_end = $2;
	} else {
		croak "unknown fbn range $!";
	}

	my $last_vbn;
	my $result = 'EQUAL';
	for (my $fbn = $fbn_start; $fbn<=$fbn_end; $fbn++) {
		my %fbn_info = $class->_get_vbn_info_single(filename => $filename,
							    fbn => $fbn);
		my $current_vbn = $fbn_info{fbn_info}->{$fbn}->{vbn};
		if (!defined($current_vbn)) {
			logresult('fatal', "Failed to obtain vbn for file $filename, fbn $fbn\n" . Dumper(%fbn_info));
		}
		if (($fbn > $fbn_start) && ($current_vbn != $last_vbn)) {
			$result = 'NOT_EQUAL';
			last;
		}
		$last_vbn = $current_vbn;
	}

	if ($result ne $expected) {
		logresult('fatal', "\nRESULT: $result\nEXPECTED RESULT: $expected $!");
	}

	$res{'rc'} = C_RC_SUCCESS;
	$res{'result'} = $result;
	return %res;
}

## @name ::file_reservation_enable
##
## @summary file reservation enable
##
## @description
## Enable file reservation
##
## @return status
##
## @arg	Required:	filename
##
## @examples 
##	$dataset->file_reservation_enable( filename=>'f1' );
##
sub file_reservation_enable
{
	my $class 	= shift;
	my (%args) = validate (@_, {	filename => { type=>SCALAR },} );
	
	return $class->set_file_reservation(filename => $args{filename},
					    enabled => TRUE);
}

## @name ::file_reservation_disable
##
## @summary file reservation disable
##
## @description
## Disable file reservation.
##
## @return status
##
## @arg	Required:	filename
##
## @examples 
##	$dataset->file_reservation_disable( filename=>'f1' );
##
sub file_reservation_disable
{
	my $class 	= shift;
	my (%args) = validate (@_, { filename => { type=>SCALAR },} );
	
	return $class->set_file_reservation(filename => $args{filename},
					    enabled => FALSE);
}

## @name ::verify_compressed
##
## @summary verify that file is compressed
##
## @description
## From the output of showvbns, verify that file is compressed.
## Croaks if not as expected.
##
## @return hash of status
##
## @arg filename					Required: file to verify
## @arg fbn							Optional: fbn to compare ( or fbn range e.g. 8-16 )
## @arg expected					Optional: COMPRESSED|NOT_COMPRESSED default is 'COMPRESSED'
## @arg max_allowed_uncompressed	Optional: default is 2
## @arg method						Optional: SHOWVBNS|DUMPBLOCK default is 'SHOWVBNS'
## @arg snapshot					Optional: name of snapshot, default is undef
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset
##
## @examples
##  $dataset->verify_compressed(	filename	=>'f1', 
##									fbn			=>'0-31', # or '3'
##									expected	=>'COMPRESSED', # or 'NOT_COMPRESSED'
##									max_allowed_uncompressed => { type=>SCALAR, default=>2 },
##							);
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub verify_compressed
{
	my $class 	= shift;
	my (%args) = validate (@_, {filename	=> { type=>SCALAR, optional=>1 },
				    file_id		=> { type=>SCALAR, optional=>1 },
				    fbn			=> { type=>SCALAR, optional=>1 },
				    expected	=> { type=>SCALAR, default=>'COMPRESSED', regex => qr/^COMPRESSED|NOT_COMPRESSED$/ },
				    max_allowed_uncompressed => { type=>SCALAR, optional=>1 },
				    snapshot => { type=>SCALAR|UNDEF, optional=>1 },
				    method	=> { type=>SCALAR, default=>'SHOWVBNS', regex => qr/^SHOWVBNS|DUMPBLOCK$/ },
				    allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
	
	} );
	my %res;
	my $filename	= $args{filename};
	my $file_id		= $args{file_id};
	my $fbn			= $args{fbn};
	my $expected	= $args{expected};
	my $snapshot	= $args{snapshot};
	my $max_allowed_uncompressed	= $args{max_allowed_uncompressed};
	my $method = $args{method};
	my $allow_removed = $args{allow_removed};

	my $comment = "\nverify_compressed\n";
	$comment .= "filename: $filename\n" if $filename;
	$comment .= "file_id: $file_id\n" if $file_id;
	$comment .= "fbn: $fbn\n" if $fbn;
	$comment .= "expected: $expected\n" if $expected;
	$comment .= "snapshot: $snapshot\n" if $snapshot;
	$comment .= "max_allowed_uncompressed: $max_allowed_uncompressed\n" if $max_allowed_uncompressed;
	$comment .= "method: $method\n" if $method;
	logcomment($comment);

	if ((! $filename) && (! $file_id)) {
		logresult("script", "Need to specify either filename or file_id");
	}


	if ( ! defined($fbn) ) {
	  croak("verify_compressed_all() not yet implemented");
	  $class->verify_compressed_all(filename =>$filename,
					expected =>$expected,
					snapshot =>$snapshot,
								);
	} else {
		if ($method eq 'DUMPBLOCK') {
			if ($fbn !~ /^\d+$/) {
				logresult("script", "For DUMPBLOCK, please specify one fbn");
			}
			if ($max_allowed_uncompressed) {
				logresult("script", "max_allowed_uncompressed is not an option for DUMPBLOCK");
			}
			
			$class->_verify_compressed_from_dumpblock(filename	=>$filename,
								  file_id       =>$file_id,
								  fbn           =>$fbn,
								  expected	=>$expected,
								  snapshot	=>$snapshot,
								  allow_removed => $allow_removed,
								  );
		} elsif ( $method eq 'SHOWVBNS' ) {
			$max_allowed_uncompressed = $max_allowed_uncompressed || 2;
			my ( $fbn_start, $fbn_end );
			if ( $fbn =~ /^(\d+)$/ ) {
				$fbn_start = $1;
				$fbn_end = $1;
			} elsif ( $fbn =~ /^(\d+)\s*-\s*(\d+)$/ ) {
				$fbn_start = $1;
				$fbn_end = $2;
			} else {
				croak "unknown fbn range $!";
			}
			$class->_verify_compressed_range(filename=>$filename,
							 fbn_1		=>$fbn_start,
							 fbn_2		=>$fbn_end,
							 expected	=>$expected,
							 max_allowed_uncompressed => $max_allowed_uncompressed,
							 snapshot	=>$snapshot,
							 allow_removed => $allow_removed,
							 );
		} else {
			croak "unknown method $!";
		}
	}
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::get_block_layout
##
## @summary get block layout
##
## @description
## From the output of showvbns, get pattern of headers, data, trailers 
##
## @return hash reference with block pattern keyed by fbn
##  e.g. $block_layout_1->{
##							0=>'H',
##							1=>'D',
##							2=>'T',
##							3=>'T',
##							4=>'T',
##							5=>'T',
##							6=>'T',
##							7=>'T',
##							
##
## @arg filename	Required: file to verify
## @arg fbn			Optional: fbn to compare ( or fbn range e.g. 8-16 )
##
## @examples
##  $block_pattern = $dataset->get_block_pattern(	filename	=>'f1', 
##													fbn			=>'0-31',
##							);
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub get_block_layout
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		filename => { type=>SCALAR },
		fbn => { type=>SCALAR, optional=>1 },
	} );
	
	my %res;
	my $filename	= $args{filename};
	my $fbn			= $args{fbn};
	my $block_layout;

	logcomment("\nget_block_layout\n" .
		   "filename: $filename\n" .
		   "fbn: $fbn\n");				

	if ( ! defined($fbn) ) {
	  croak("get_block_layout_all() not yet implemented");
	  $block_layout = $class->get_block_layout_all(	filename	=>$filename, 
								);
	} else {
		my ( $fbn_start, $fbn_end );
		if ( $fbn =~ /^(\d+)$/ ) {
			$fbn_start = $1;
			$fbn_end = $1;
		} elsif ( $fbn =~ /^(\d+)\s*-\s*(\d+)$/ ) {
			$fbn_start = $1;
			$fbn_end = $2;
		} else {
			croak "unknown fbn range $!";
		}
		$block_layout = $class->_get_block_layout_range(filename =>$filename,
								fbn_1 =>$fbn_start,
								fbn_2 =>$fbn_end,);
	}
	return $block_layout;
}

## @name ::compare_block_layouts
##
## @summary compares 2 block layouts
##
## @description
## This function compares the block layouts from 2 files.
## Use this after get_block_layout ( see examples below ).
## Croaks if not expected result.
##
## @return hash of status and output
##	$res{'result'} = EQUAL or NOT_EQUAL
## 
## @arg block_layout_1	Required: block layout of first file
## @arg block_layout_2	Required: block layout of second file
## @expected_result	Optional: EQUAL|NOT_EQUAL ( default is 'EQUAL' )
##
## @example
##
##	my $block_layout_1 = $dataset_1->get_block_layout(	filename	=>'f1', 
##														fbn			=>'0-7',
##								);
##
##	my $block_layout_2 = $dataset_1->get_block_layout(	filename	=>'f2', 
##														fbn			=>'0-7',
##								);
##
##	%res = $dataset->compare_block_layouts( 
##						block_layout_1=> $block_layout_1, 
##						block_layout_2=> $block_layout_2,
##						expected_result=>EQUAL|NOT_EQUAL #default:EQUAL
##						);
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub compare_block_layouts
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		block_layout_1 => { type=>HASHREF },
		block_layout_2 => { type=>HASHREF },
		expected_result => { type => SCALAR, optional => 1, regex => qr/^(EQUAL|NOT_EQUAL)$/, default => 'EQUAL'},
	});
	
	my $block_layout_1 = $args{block_layout_1};
	my $block_layout_2 = $args{block_layout_2};
	my $expected_result = $args{expected_result};
	my %res;
	my $result = 'EQUAL';

	my $block_1_size = scalar(keys %{$block_layout_1});
	my $block_2_size = scalar(keys %{$block_layout_2});

	logcomment("compare_block_layouts\n".
		   "expected_result: $expected_result\n".
		   "file_1: $block_1_size blocks\n".
		   "file_2: $block_2_size blocks");

	if ($block_1_size != $block_2_size) {
		logresult('script', "Blocks sizes are different: block_1:$block_1_size block_2:$block_2_size");
	}

	foreach my $fbn (sort {$a <=> $b} keys %{$block_layout_1} ) {
		logcomment("fbn:$fbn  file_1:$block_layout_1->{$fbn} file_1:$block_layout_2->{$fbn}") if $class->{_debug};
		if ( $block_layout_1->{$fbn} ne $block_layout_2->{$fbn} ) {
			$result = 'NOT_EQUAL';
			logcomment("NOT_EQUAL at fbn $fbn");
			last;
		}
	}

	# compare with expected
	if ( defined($expected_result) && ($result ne $expected_result) ) {
		croak "RESULT: $result\nEXPECTED RESULT: $expected_result $!";
	}

	$res{'rc'} = C_RC_SUCCESS;
	$res{'result'} = $result;
	return %res;
}

## @name ::set_file_reservation
##
## @summary 
##
## @description
## 
##
## @return hash of status
##
## @arg filename	Required: name of file
## @arg enabled		Optional: TRUE or FALSE (default TRUE )
##
## @examples
##  %res = $src_vol->{dataset}->set_file_reservation(filename=>$FILENAME);
##  %res = $src_vol->{dataset}->set_file_reservation(filename=>$FILENAME, enabled=>FALSE);
sub set_file_reservation
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename => { type=>SCALAR },
		enabled  => { type=>SCALAR, optional=>1, default=>TRUE},
	});
	
	my $filename = $args{filename};
	my $enabled  = $args{enabled};
	
	my $cmd_code = C_CMD_FILE_RESERVATION;
	my $is_dblade_cmd = $class->{_ui}->is_dblade_level_command(code => $cmd_code);
	my $file_path = $class->_generate_file_path(filename => $filename,
						    need_vol_prepended => $is_dblade_cmd);
	
	my $command = $class->{_ui}->formulate_command(code => $cmd_code,
						       file_path => $file_path,
						       options => {enabled => $enabled},);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
							    boot_mode => $class->{_boot_mode},
							    cmd => $command,
							    node_run => $is_dblade_cmd,);
	
	my $expected_rc = ($enabled == TRUE) ? C_RC_FILE_RESERVATION_ON : C_RC_FILE_RESERVATION_OFF;
	return $class->{_ui}->process_output_string(code => $cmd_code,
						    output => $output,
						    expected_rc => $expected_rc,
						    log_message => "Set file reservation (enabled=$enabled) for file $file_path",);
}

## @name ::get_file_reservation
##
## @summary  Get the current file reservation setting.
##
## @description Get the current file reservation setting.
##
## @return Hash of status.
##
## @arg filename Required: Name of the file.
##
## @arg allow_removed Optional: Flag to allow setting to be retrieved for a file that was previously removed from this dataset.
## Default value: FALSE.
##
## @arg expected Optional: Expected setting (C_FILE_RESERVATION_ON/C_FILE_RESERVATION_OFF).
## Default value: none.
##
## @examples
## $obj->get_file_reservation(filename => 'my_file',
##                            expected => C_FILE_RESERVATION_ON);
##  
sub get_file_reservation
{
	my $class 	= shift;
	my (%args) = validate (@_, { filename => { type=>SCALAR },
				     allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
				     expected => { type=>SCALAR, optional => 1 },});
	
	my $filename = $args{filename};
	my $expected_setting = $args{expected};
	my $expected_rc;
	
	if ($expected_setting eq C_FILE_RESERVATION_ON)
	{
		$expected_rc = C_RC_FILE_RESERVATION_ON;
	}
	elsif ($expected_setting eq C_FILE_RESERVATION_OFF)
	{
		$expected_rc = C_RC_FILE_RESERVATION_OFF;
	}
	else
	{
		logresult('script', "Unknown expected file reservation setting: $expected_setting");
	}
	
	my $cmd_code = C_CMD_FILE_RESERVATION;
	my $dblade_level = $class->{_ui}->is_dblade_level_command(code => $cmd_code);
	my $file_path = $class->_generate_file_path(filename => $filename,
						    allow_removed => $args{allow_removed},
						    need_vol_prepended => $dblade_level);
	my $command = $class->{_ui}->formulate_command(code => $cmd_code,
						       file_path => $file_path,);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
							    boot_mode => $class->{_boot_mode},
							    cmd => $command,
							    node_run => $dblade_level,);
	
	return $class->{_ui}->process_output_string(code => $cmd_code,
						    output => $output,
						    expected_rc => $expected_rc,
						    log_message => "Verify file reservation \"$expected_setting\" for file $file_path",);
}

## @name ::verify_block_type
##
## @summary verify block type
##
## @description
## From the output of showvbns, verifies that the type of block is HEADER, TRAILER, or DATA.
## Croaks if not as expected.
##
## @arg filename					Required: file to verify
## @arg fbn							Required: fbn to compare
## @arg expected					Required: HEADER|TRAILER|DATA
## @arg snapshot					Optional: name of snapshot, default is undef
##
## @examples
##  $dataset->verify_block_type(	filename	=>'f1', 
##									fbn			=> 3,
##									expected	=>'HEADER',
##							);
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub verify_block_type
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		filename => { type=>SCALAR, },
		fbn => { type=>SCALAR, },
		expected => { type=>SCALAR, regex => qr/^HEADER|TRAILER|DATA$/ },
		snapshot => { type=>SCALAR|UNDEF, optional=>1 },
	} );
	
	my $filename	= $args{filename};
	my $fbn			= $args{fbn};
	my $expected	= $args{expected};
	my $snapshot	= $args{snapshot};

	my $comment = "\nverify_block_type\n";
	$comment .= "filename: $filename\n";
	$comment .= "fbn: $fbn\n";
	$comment .= "expected: $expected\n";
	$comment .= "snapshot: $snapshot\n" if $snapshot;
	logcomment($comment);	
	
	my %res = $class->_get_vbn_info_single(	filename =>$filename,
					       fbn =>$fbn,
					       snapshot	=>$snapshot,
					       );
	my $block_type = $class->_get_block_type(vbn =>$res{'vbn'},
						 vvbn =>$res{'vvbn'},
						 );	

	if ($block_type ne $expected) {
		croak "block_type:$block_type expected:$expected $!";
	} else {
		logcomment("block_type:$block_type expected:$expected");
	}


}

## @name ::get_dumpblock_info
##
## @summary  Get dumpblock info.
##
## @description Wrapper around FS::Utilities::dumpblock.
##
## @return Hash with status and output.
##  e.g. $res{output} => {
##						'pvbn-of-parent' => '35128867',
##						'pvbn' => '281474976710652',
##						'dbn' => '0',
##						'disk' => '(unknown)',
##						'status' => 'ok',
##						'vvbn-of-parent' => '4151',
##						'vvbn' => '281474976710652',
##						'pvbn-of-parent-from-container' => '35128867',
##						'pvbn-from-parent' => '281474976710652',
##						'pvbn-from-container' => '281474976710652'
##
## @arg filename Required: Name of the file.
## @arg fbn Required: fbn of the file.
## @arg allow_removed Optional: Flag to allow setting to be retrieved for a file that was previously removed from this dataset.
##		Default value: FALSE.
## @arg level Optional: Block level of the file being used for dumpblock.
## @arg snapname Optional: Name of snapshot if file is in a snapshot.
##
## @examples
## $obj->get_dumpblock_info( filename => $FILENAME, fbn => '3');
##  
sub get_dumpblock_info
{
	my $class = shift;
	my (%args) = validate (@_,
	{
		filename => { type=>SCALAR },
		allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
		fbn => { type=>SCALAR, },
		level => { type=>SCALAR, optional => 1 },
		snapname => { type=>SCALAR, optional => 1 },
	});
	
	my $filename = $args{filename};
	my $fbn = $args{fbn};
	my $block_info;
	
	my $file_path = $class->_generate_file_path(filename => $filename,
						    allow_removed => $args{allow_removed});
	if ($class->{_boot_mode} eq C_ONTAP_CMODE)
	{
		if (defined($args{snapname})) {
			logresult ('NA', "NOT APPLICABLE: Snapname not supported in cmode");
		}
		my $volume_name = $class->{vol_name};
		my $level = (defined($args{level})) ? ($args{level}) : 0;
		my $cmd_code = ($level == 1) ? C_CMD_DUMPBLOCK_INFO_L1 : C_CMD_DUMPBLOCK_INFO;
		my $file_id = $class->_get_file_id(filename => $filename,
						   allow_removed => $args{allow_removed});
		
		my $command = $class->{_ui}->formulate_command(code => $cmd_code,
							       volume => $volume_name,
							       file_path => $file_id,
							       fbn => $fbn,
							       level => $level,
							       );
		
		my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
								    boot_mode => $class->{_boot_mode},
								    cmd => $command,
								    node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
		
		my %processed_output = $class->{_ui}->process_output_string(code => $cmd_code,
									    output => $output,
									    expected_rc => C_RC_SUCCESS,
									    log_message => "Get dumpblock info for file id $file_id, fbn $fbn, volume $volume_name",);
		$block_info = $processed_output{dumpblock_info};
	}
	else
	{
		my %options;
		$options{level} = $args{level} if defined $args{level};
		$options{snapname} = $args{snapname} if defined $args{snapname};
		$block_info = dumpblock(conn => $class->{volume}->Filer->conn,
					block_type	=> 'info',
					volume		=> $class->{vol_name},
					blockno		=> $fbn,
					file		=> $file_path,
					%options);
	}
	
	return (rc => C_RC_SUCCESS, output => $block_info);
}

## @name ::get_CG_ranges
##
## @summary  get CG ranges
##
## @description get all the CG ranges in a file, using VBNs
##
## @return Hash with status and "CG_ranges" which is an array of ranges in the form 0-7, 8-15, etc.
##
## @arg filename Required: Name of the file.
##
## @examples
## 		%res = $obj->get_CG_ranges( filename=>$filename );
##		my @CG_ranges = @{$res{CG_ranges}};
##  
sub get_CG_ranges
{
	my $class 	= shift;
    my (%args) = validate (@_, {	filename	=> { type=>SCALAR },
	} );
	my %res;
	my $filename	= $args{filename};
	my %constants = $class->get_vbn_constants();
	my @vbn_headers = @{$constants{'vbn_headers'}};
	my @vbn_trailers = @{$constants{'vbn_trailers'}};
	%res = $class->get_vbn_hash( filename=>$filename );
	my $vbns_info = $res{'result'};
	my $current_CG_range;
	my @CG_ranges;

	foreach my $fbn (sort { $a <=> $b } keys %{$vbns_info}) {
		if ( grep(/^$vbns_info->{$fbn}->{vbn}$/, @vbn_headers ) ) {
				$current_CG_range = $fbn .'-';
			} elsif ( ( grep(/^$vbns_info->{$fbn}->{vbn}$/, @vbn_trailers) ) &&
					  ( ! grep(/^$vbns_info->{$fbn+1}->{vbn}$/, @vbn_trailers) ) )	
			{
				$current_CG_range .= $fbn;
				if ( $current_CG_range =~ /\d+\-\d+/ ) {
					push(@CG_ranges, $current_CG_range);
					undef $current_CG_range;
				} else {
					logresult("fatal", "incomplete current_CG_range: $current_CG_range");
				}				
			}
	}

	$res{'CG_ranges'} = \@CG_ranges;
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::attach_dataset_info
##
## @summary  Attach dataset info from a different Compress object.
##
## @description Attach dataset info from a different Compress object.
##
## @arg src_vol Required: Compression::Compress object from src volume
##
## @examples
## 
sub attach_dataset_info
{
	my $class = shift;
	my (%args) = validate (@_, {	
		src_vol => { type=>OBJECT, isa  => "Compression::Compress", },
		reset_info => {type => SCALAR, optional => 1, default => FALSE,},
	} );
	my $src_vol = $args{'src_vol'};

	$class->_attach_to_cloned_dataset(files => $src_vol->{dataset}->{_c_dataset}->{_common}->{files},
					  removed_files => $src_vol->{dataset}->{_c_dataset}->{_common}->{_removed_files},
					  ds_type => $src_vol->{dataset}->{_c_dataset}->{_common}->{ds_type},
					  testdir => $src_vol->{dataset}->{_c_dataset}->{_common}->{testdir},
					  testdir_path => $src_vol->{dataset}->{_c_dataset}->{_common}->{testdir_path},
					  parent_path => $src_vol->{volume}->get_mount_path(),
					  parent_mountpoint => $src_vol->{dataset}->get_mountpoint(),
					  reset_info => $args{reset_info},);

}

## @name ::get_uncompressable_data
##
## @summary get uncompressable data
##
## @description create data that is uncompressable.  Only use this when you need to add
##  the exact same data to different files. Warning: out-of-memory error on large files.
##
## @return data
##
## @arg size	Required:	size of data to create
## 
## @examples
##  my $data = $dataset->get_uncompressable_data(size=>4 * KB);
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com, kozel@netapp.com
##
sub get_uncompressable_data
{
	my $class	= shift;
	my (%args)	= validate (@_, { size => { type=>SCALAR, regex => qr/^\d+$/ } } );
	my $size	= $args{size};

	my $blocksize = C_WAFL_BLOCK_SIZE;
	my $numblocks = int ($size / $blocksize);
	my $remainder = $size % $blocksize;
	my $shortblock	= $blocksize - 20;
	
	logcomment "get_uncompressable_data size=$size";
	
	my $data = '';
	my $randcg8 = `dd if=/dev/urandom count=1 bs=$shortblock`;
	my $randcg7 = `dd if=/dev/urandom count=1 bs=$shortblock`;
	$randcg8 .= `dd if=/dev/urandom count=7 bs=$blocksize`;
	$randcg7 .= `dd if=/dev/urandom count=6 bs=$blocksize`;

	my $sha_sum = sha1($randcg8);
	$randcg8	= $sha_sum.$randcg8;
	$sha_sum = sha1($randcg7);
	$randcg7	= $sha_sum.$randcg7;

	# Add as many complete compression groups as possible
	my $cg_counter = 1;
	while ( $numblocks >= 7 ) {
		if ( $cg_counter == 32 ) {
			$data	.= $randcg7;
			$cg_counter = 1;
			$numblocks -= 7;
		} elsif ( $numblocks >= 8 ) {
			$data	.= $randcg8;
			$cg_counter++;
			$numblocks -= 8;
		} else {
			last;
		}
	}

	# create the final compression group if needed
	my $final_cg = '';
	if ( $numblocks > 0 || $remainder > 20) {
		$final_cg = `dd if=/dev/urandom count=1 bs=$shortblock` if $numblocks != 0;
		$numblocks--;
		$final_cg	.= `dd if=/dev/urandom count=$numblocks bs=$blocksize` if $numblocks != -1;
		$final_cg	.= `dd if=/dev/urandom count=1 bs=$remainder` if $remainder != 0;
	
		$sha_sum	= sha1($final_cg);
		$final_cg	= $sha_sum.$final_cg;
		$data	.= $final_cg;
	} else {
		my $known_pattern = "01234567890123456789";
		$final_cg = substr($known_pattern, 0, $remainder);
	}
	
	return $data;
}

## @name ::get_vbn_constants
##
## @summary get the vbn/vvbn constants
##
## @description
## get the vbn/vvbn constants
##
## @return hash of status and output
##	
## 
## @example
##	%res = $dataset->get_vbn_constants();
##  if (grep(/^$x$/, @{$res{vbn_headers}} ) ) {
##		logcomment("$x is a valid vbn header");
##	}
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_vbn_constants {
	my $class = shift;
	my (%args) = validate( @_,{} );
	my $ontap_ver = $class->{_ui}->{version}->{ontap};	

	my @constant_codes = qw(
			C_VBN_HEADER
			C_VBN_TRAILER
			C_VVBN_HEADER
			C_VVBN_TRAILER
	);
	if (($ontap_ver ne '7.3')&&($ontap_ver ne '8.0')) {
		push(@constant_codes, 'C_VVBN_HEADER_B');	
	}

	my @vbn_headers;
	my @vvbn_headers;
	my @vbn_trailers;
	my @vvbn_trailers;
	my %res;
	foreach my $constant_code (@constant_codes) {
		my $value = $class->{_ui}->get_constant_value( constant_code => $constant_code );
		if ($constant_code =~ /^C_VBN_HEADER/ ) {
			push(@vbn_headers, $value);
		} elsif ($constant_code =~ /^C_VBN_TRAILER/ ) {
			push(@vbn_trailers, $value);
		} elsif ($constant_code =~ /^C_VVBN_HEADER/ ) {
			push(@vvbn_headers, $value);
		} elsif ($constant_code =~ /^C_VVBN_TRAILER/ ) {
			push(@vvbn_trailers, $value);
		}
	}
	$res{vbn_headers} = \@vbn_headers;
	$res{vbn_trailers} = \@vbn_trailers;
	$res{vvbn_headers} = \@vvbn_headers;
	$res{vvbn_trailers} = \@vvbn_trailers;
	return %res;
}


## @name ::wafl_punch
##
## @summary Punch holes in a file using "wafl punch". 
##
## @description Punch holes in a file using "wafl punch". 
##
## @arg filename Required: Name of the file.
##
## @arg fbn Required: Fbn to be used for punching.
##
## @arg count Required: Number of holes to be punched.
##
## @examples
## $obj->wafl_punch(filename => 'my_file', fbn => 1, count => 4);
## 
sub wafl_punch
{
	my $class = shift;
	my (%args) = validate (@_, {filename => {type=>SCALAR,},
				    fbn => {type=>SCALAR,},
				    count => {type=>SCALAR,},});
	
	my $filename = $args{filename};
	my $fbn = $args{fbn};
	my $count = $args{count};
	
	my $cmd_code = C_CMD_WAFL_PUNCH;
	my $dblade_level = $class->{_ui}->is_dblade_level_command(code => $cmd_code);
	my $file_path = $class->_generate_file_path(filename => $filename,
						    need_vol_prepended => $dblade_level);
	
	my $command = $class->{_ui}->formulate_command(code => $cmd_code,
						       file_path => $file_path,
						       fbn => $fbn,
						       increment => $count,);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
							    boot_mode => $class->{_boot_mode},
							    cmd => $command,
							    node_run => $dblade_level,);
	
	return $class->{_ui}->process_output_string(code => $cmd_code,
						    output => $output,
						    expected_rc => C_RC_SUCCESS,
						    log_message => "Wafl punch of file $filename (fbn $fbn, count $count)",);
}


## @name ::verify_vbn_zero
##
## @summary Verifies that a specified block is VBN_ZERO. 
##
## @description Verifies that a specified block is VBN_ZERO.  
##
## @arg filename Required: Name of the file.
##
## @arg fbns Required: Fbn to be verified.
## Can be a single fbn or a range (ex: 1, or 2-5).
##
## @arg expected Required: Expected result: VBN_ZERO or NOT_VBN_ZERO.
## Default: VBN_ZERO.
##
## @examples
## $obj->verify_vbn_zero(filename => 'my_file', fbns => 1, expected => "NOT_VBN_ZERO");
## $obj->verify_vbn_zero(filename => 'my_file', fbns => "2-7");
## 
sub verify_vbn_zero
{
	my $class = shift;
	my (%args) = validate (@_, {filename => {type => SCALAR,},
				    fbns => {type => SCALAR,},
				    expected => {type => SCALAR, default => 'VBN_ZERO', regex => qr/^VBN_ZERO|NOT_VBN_ZERO$/},
	});
	
	my $filename = $args{filename};
	my $fbns = $args{fbns};
	my $expected = $args{expected};
	my $vbn_zero = $class->{_ui}->get_constant_value(constant_code => 'VBN_ZERO' );
	
	my ($fbn_start, $fbn_end);
	if ($fbns =~ /^(\d+)$/) {
		$fbn_start = $1;
		$fbn_end = $1;
	}
	elsif ($fbns =~ /^(\d+)\s*-\s*(\d+)$/) {
		$fbn_start = $1;
		$fbn_end = $2;
	}
	else {
		logresult('script', "Illegal fbns passed to verify_vbn_zero: $fbns");
	}
	
	for (my $fbn = $fbn_start; $fbn <= $fbn_end; $fbn++) {
		my %res = $class->_get_vbn_info_single(filename	=> $filename,
						       fbn => $fbn,);
		
		my $vbn = $res{vbn};
		my $vvbn = $res{vvbn};
		if (("$vbn" eq "$vbn_zero") && ("$vvbn" eq "$vbn_zero")) {
			if ($expected ne 'VBN_ZERO') {
				logresult('fatal', "Detected unexpected VBN_ZERO at fbn $fbn:\n" . Dumper(%res));
			}
			logcomment("FBN $fbn: detected VBN_ZERO: $vbn, $vvbn");
		}
		else {
			if ($expected eq 'VBN_ZERO') {
				logresult('fatal', "Did not detect expected VBN_ZERO at fbn $fbn:\n" . Dumper(%res));
			}
			logcomment("FBN $fbn: did not detect VBN_ZERO: $vbn, $vvbn");
		}
	}
	
	return (rc => C_RC_SUCCESS,);
}

################################################################
# Internal subroutines
################################################################


# Internal function
# name ::_new_file
#
# summary (INTERNAL) create a new file object
#
# description
# Creates a new file object.
# Doesn't create actual file.
#
# example
#  $dataset->_new_file( filename=>'f2' );
# 
# File objects contain the folling:
# $file->{path}			= full path on client
# $file->{filer_path}		= path for filer commands
# $file->{relative_path}	= path relative to volume
# 
# note: both filer_path and relative_path are only for
# commands run on filer
sub _new_file
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		file => { type=>HASHREF },
		clone_lun => { type=>SCALAR|UNDEF, optional=>1 },
	} );
	my $file = $args{file};
	my $filename = $file->{name};
	my $clone_lun = $args{clone_lun};
	if (defined($clone_lun)) {
		$class->{_lun_clones}->{$clone_lun}->{files}->{$filename} = $file;
	}
	else {
		$class->{files}->{$filename} = $file;
	}
}

# Internal function
# name ::_get_vbn_info_single
#
# summary (INTERNAL) 
#
# description
# 
#
# example
#		$class->_get_vbn_info_single(	filename	=>$filename, 
#										fbn			=>$fbn,
#									);	
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _get_vbn_info_single
{
	my $class 	= shift;
	my (%args) = validate (@_, {filename => { type=>SCALAR },
				    fbn => { type=>SCALAR },
				    snapshot => { type=>SCALAR|UNDEF, optional=>1, },
				    allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
	} );
	my %res;
	my $filename = $args{filename};
	my $fbn = $args{fbn};
	my $snapshot = $args{snapshot};
	
	if ($class->{_has_debug_kernel} == FALSE)
	{
		logresult ('NA', "NOT APPLICABLE: The showvbns command is only available on debug builds");
	}

	my $cmd_code = C_CMD_GET_VBN_INFO;
	my $is_dblade_cmd = $class->{_ui}->is_dblade_level_command(code => $cmd_code);
	my $file_path = $class->_generate_file_path(filename => $filename,
						    snapshot => $snapshot,
						    allow_removed => $args{allow_removed},
						    need_vol_prepended =>$is_dblade_cmd);
	
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
						       file_path => $file_path,
						       fbn => $fbn);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
							    node_run => $is_dblade_cmd,);
	
	
	my %vbn_info = $class->{_ui}->process_output_string(code => $cmd_code,
							    output => $output,
							    required_return_hash_key => $fbn,
							    expected_rc => C_RC_SUCCESS,
							    log_message => "Obtain vbn info for file $file_path, fbn $fbn",);
	
	my $vbn = $vbn_info{vbn_info}->{$fbn}->{vbn};
	my $vvbn = $vbn_info{vbn_info}->{$fbn}->{vvbn};
	
	if ((!defined($vbn)) || (!defined($vvbn)))
	{
		logresult('fatal', "Failed to obtain vbn/vvbn for file $file_path, fbn $fbn:\n\n" . Dumper(%vbn_info));
	}
	
	$res{'vbn'} = $vbn;
	$res{'vvbn'} = $vvbn;
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

# Internal function
# name ::_get_vbn_info_all
# summary (INTERNAL)
#
# example
# $class->_get_vbn_info_all(filename =>$filename, 
#                           snapshot => $snapshot);	
#
# author dl-compression-qa@netapp.com, voicu@netapp.com
#
sub _get_vbn_info_all
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename => { type=>SCALAR },
		snapshot => { type=>SCALAR|UNDEF, optional=>1, },
		allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
		panic_expected => { type=>SCALAR, optional=>1, default => FALSE },
		expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
	} );
	
	my $filename	= $args{filename};
	my $snapshot	= $args{snapshot};
	my $panic_expected = $args{panic_expected};
	my $expected = $args{expected};
	
	if ($class->{_has_debug_kernel} == FALSE)
	{
		logresult ('NA', "NOT APPLICABLE: The showvbns command is only available on debug builds");
	}

	my $cmd_code = C_CMD_GET_VBN_INFO;
	my $is_dblade_cmd = $class->{_ui}->is_dblade_level_command(code => $cmd_code);
	my $file_path = $class->_generate_file_path(filename => $filename,
						    snapshot => $snapshot,
						    allow_removed => $args{allow_removed},
						    need_vol_prepended =>$is_dblade_cmd);
	
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
						       file_path => $file_path,);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
							    maxbufsize => 10 * MB,
							    node_run => $is_dblade_cmd,
							    panic_expected => $panic_expected);
	
	if ($panic_expected == TRUE)
	{
		logcomment("Get vbn info: command is expected to panic the filer. Partial output:\n$output") if ($class->{_debug});
		return {rc => C_RC_SUCCESS, output => $output}
	}
	
	if ($expected != C_RC_SUCCESS)
	{
		return $class->{_ui}->process_output_string(code => C_CMD_GET_VBN_INFO_ERROR,
							    output => $output,
							    expected_rc => $expected,
							    log_message => "Obtain vbn info for corrupted file $file_path",);
	}
	
	my %vbn_info = $class->{_ui}->process_output_string(code => $cmd_code,
							    output => $output,
							    expected_rc => C_RC_SUCCESS,
							    log_message => "Obtain vbn info for file $file_path",);
	
	if (!defined($vbn_info{vbn_info}->{'0'}))
	{
		logresult('script', "Incomplete buffer, try using a smaller file:\n" . Dumper(%vbn_info));
	}
	
	return $vbn_info{vbn_info};
}

# Internal function
# name ::_get_vol_mount_path
# summary (INTERNAL)
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _get_vol_mount_path
{
	my $class 	= shift;
	return $class->{volume}->mount_path();
}

# Internal function
# name ::_filename_exists
#
# summary (INTERNAL) Checks that file exists in the dataset object
#
# description
# Used in error checking by various functions
#
# example
#		$class->_filename_exists( filename=>$filename, expected=>FALSE ); 
#
sub _filename_exists
{
	my $class 	= shift;
	my (%args) = validate (@_, {filename => { type=>SCALAR },
				    expected => { type=>SCALAR, default=>TRUE, },
				    clone_lun => {type => SCALAR|UNDEF, optional => 1,},
	} );
	my $filename = $args{filename};
	my $expected = $args{expected};
	my $clone_lun = $args{clone_lun};
	my $files = (defined($clone_lun)) ? ($class->{_lun_clones}->{$clone_lun}->{files}) : $class->{files};
	my $result = FALSE;
	
	if (exists($files->{$filename})) {
		$result = TRUE;
		if ($expected == FALSE ) {
			croak "filename $filename already exists $!";
		}
	}
	elsif ($expected == TRUE) {
		croak "filename $filename does NOT exist $!";
	}
	return $result;
}

# Internal function
# name ::_remove_file
#
# summary (INTERNAL) Moves the file entry from the $class->{files} hash
#  into the $class->{_removed_files} one.
#
# example
#		$class->_remove_file( filename=>$filename ); 
#
sub _remove_file
{
	my $class 	= shift;
	my (%args) = validate (@_, {
				    filename => { type=>SCALAR },
				    } );
	
	my $filename = $args{filename};
	
	$class->{_removed_files}->{$filename} = $class->{files}->{$filename};
	delete($class->{files}->{$filename});
}

# Internal function
# name ::_generate_file_path
#
# summary (INTERNAL)
# If snapshot is defined, generates the complete path of a file within a snapshot.
# If snapshot is not defined, returns the file path.
#
# description
# Generates the complete path of a file (within a snapshot, if required).
#
# example
#  my $snapshot_file_path = $dataset->_generate_file_path(filename=>'f2',
#                                                         snapshot => 'my_snapshot');
# 
#
#
sub _generate_file_path
{
	my $class 	= shift;
	my (%args) = validate (@_,
			       { 
					filename => { type=>SCALAR },
					snapshot => { type=>SCALAR|UNDEF, optional=>1 },
					allow_removed => { type=>BOOLEAN|UNDEF, optional=>1, },
					need_vol_prepended => { type => BOOLEAN, default => FALSE },
				} );
	
	my $filename = $args{filename};
	my $snapshot = $args{snapshot};
	my $allow_removed = $args{allow_removed};
	my $file_path;
	
	if (!defined($allow_removed)) {
		$allow_removed = (defined($snapshot)) ? TRUE : FALSE;
	}
	
	my $file_already_exists = $class->_filename_exists( filename=>$filename, expected=>C_RC_FAILURE_NON_FATAL );
	if ($file_already_exists == FALSE) {
		if ($allow_removed == FALSE) {
			logresult('fatal', "File $filename does not exist");
		}
		$class->_file_was_removed(filename=>$filename);
		$file_path = $class->{_removed_files}->{$filename}->{filer_path};
	}
	else {
		$file_path = $class->{files}->{$filename}->{filer_path};
	}
	
	if (defined($snapshot)) {
		my $vol_mountpath = $class->_get_vol_mount_path();
		$file_path =~ s#^$vol_mountpath#$vol_mountpath/.snapshot/$snapshot#;
	}
	
	if ($args{need_vol_prepended} == TRUE)
	{
		if ($file_path !~ /^\/vol\//)
		{
			$file_path =~ s/^\//\/vol\//;
		}
	}
	
	return $file_path;
}



# Internal function
# name ::_generate_relative_path
#
# summary (INTERNAL)
# If snapshot is defined, generates the relative path of a file within a snapshot.
# If snapshot is not defined, returns the relative file path.
#
# description
# Generates the relative path of a file (within a snapshot, if required).
#
# example
#  my $relative_path = $dataset->_generate_relative_path(filename=>'f2',
#                                                        snapshot => 'my_snapshot');
# 
#
#
sub _generate_relative_path
{
	my $class 	= shift;
	my (%args) = validate (@_,
			       { 
					filename => { type=>SCALAR },
					snapshot => { type=>SCALAR|UNDEF, optional=>1 },
					allow_removed => { type=>BOOLEAN|UNDEF, optional=>1, },
				} );
	
	my $filename = $args{filename};
	my $snapshot = $args{snapshot};
	my $allow_removed = $args{allow_removed};
	my $relative_path;
	
	if (!defined($allow_removed)) {
		$allow_removed = (defined($snapshot)) ? TRUE : FALSE;
	}
	
	my $file_already_exists = $class->_filename_exists( filename=>$filename, expected=>C_RC_FAILURE_NON_FATAL );
	if ($file_already_exists == FALSE) {
		if ($allow_removed == FALSE) {
			logresult('fatal', "File $filename does not exist");
		}
		$class->_file_was_removed(filename=>$filename);
		$relative_path = $class->{_removed_files}->{$filename}->{relative_path};
	}
	else {
		$relative_path = $class->{files}->{$filename}->{relative_path};
	}
	
	if (defined($snapshot)) {
		$relative_path = ".snapshot/$snapshot/". $relative_path;
	}
	
	return $relative_path;
}


# Internal function
# name ::_file_was_removed
#
# summary (INTERNAL) Checks that file was ever known and has been removed by the dataset object
#
# example
#		$class->_file_was_removed( filename=>$filename ); 
#
sub _file_was_removed
{
	my $class 	= shift;
	my (%args) = validate (@_, {
				    filename => { type=>SCALAR },
				    clone_lun => {type => SCALAR|UNDEF, optional => 1,},
				    } );
	
	my $filename = $args{filename};
	my $clone_lun = $args{clone_lun};
	my $removed_files = (defined($clone_lun)) ? ($class->{_lun_clones}->{$clone_lun}->{removed_files}) : $class->{_removed_files};
	
	if (!exists($removed_files->{$filename})) {
		logresult('fatal', "Filename $filename was not created and removed by dataset");
	}
	
	return TRUE;
}

# Internal function
# name ::_re_add_file
#
# summary (INTERNAL) Re-adds the file hash into $class->{files}.
#
# example
#		$class->_re_add_file( filename=>$filename ); 
#
sub _re_add_file
{
	my $class 	= shift;
	my (%args) = validate (@_, {
				    filename => { type=>SCALAR },
				    } );
	
	my $filename = $args{filename};
	
	$class->{files}->{$filename} = $class->{_removed_files}->{$filename};
	delete($class->{_removed_files}->{$filename});
}

# Internal function
# name ::_get_filename
#
# summary (INTERNAL) giving a path, returns the filename
#
sub _get_filename
{
	my $class 	= shift;
	my (%args) = validate (@_, { path => { type=>SCALAR }, } );
	my $path = $args{path};
	my $filename;
	foreach my $file ( sort keys %{$class->{files} } ) {
		if ( $class->{files}->{$file}->{path} eq $path ) {
			$filename = $file;
			last;
		}
	}
	return $filename;
}

# Internal function
# name ::_compare_vbns_range
#
# summary (INTERNAL) compare vbns range from 2 files
#
# description
# for 2 files, given a range of fbns,
# this compares the vbn values
#
# example
#  $class->_compare_vbns_range(file_1   => $file_1, 
#                              file_2   => $file_2,
#                              fbn_1    => $fbn_start,
#                              fbn_2    => $fbn_end,
#                              expected => $expected,
#                              );
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _compare_vbns_range
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		file_1   => { type=>SCALAR },
		file_2   => { type=>SCALAR },
		fbn_1    => { type=>SCALAR },
		fbn_2    => { type=>SCALAR },
		expected => { type=>SCALAR, default=>'EQUAL' },
	
	} );
	my %res;
	my $file_1		= $args{file_1};
	my $file_2		= $args{file_2};
	my $fbn_1		= $args{fbn_1};
	my $fbn_2		= $args{fbn_2};
	my $expected	= $args{expected};

	for (my $fbn = $fbn_1; $fbn <= $fbn_2; $fbn++) {
		$class->_compare_single_vbn(file_1 =>$file_1,
					    file_2 =>$file_2,
					    fbn =>$fbn,
					    expected =>$expected,
					    );
	}

	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

# Internal function
# name ::_compare_single_vbn
#
# summary (INTERNAL) compare vbns from 2 files
#
# description
# for 2 files, given a single fbn,
# this compares the vbn values
#
# example
#		$class->_compare_single_vbn(	file_1		=>$file_1, 
#										file_2		=>$file_2,
#										fbn			=>$fbn,
#										expected	=>$expected,
#								);
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _compare_single_vbn
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		file_1    => { type=>SCALAR },
		file_2    => { type=>SCALAR },
		fbn       => { type=>SCALAR },
		expected  => { type=>SCALAR, default=>'EQUAL' },
	
	} );
	
	my %res;
	my $file_1   = $args{file_1};
	my $file_2   = $args{file_2};
	my $fbn      = $args{fbn};
	my $expected = $args{expected};
	
	my %vbn_info = $class->_get_vbn_info_single(filename => $file_1,
						    fbn => $fbn);
	
	my $vbn_1 = $vbn_info{vbn_info}->{$fbn}->{vbn};
	if (!defined($vbn_1)) {
		logresult('fatal', "Failed to detect vbn for file $file_1, fbn $fbn" . Dumper(%vbn_info));
	}
	
	%vbn_info = $class->_get_vbn_info_single(filename => $file_2,
						    fbn => $fbn);
	
	my $vbn_2 = $vbn_info{vbn_info}->{$fbn}->{vbn};
	if (!defined($vbn_2)) {
		logresult('fatal', "Failed to detect vbn for file $file_2, fbn $fbn" . Dumper(%vbn_info));
	}
	
	my $result = ($vbn_1 == $vbn_2) ? 'EQUAL' : 'NOT_EQUAL';
	
	logcomment( "\nfbn: $fbn\n" .
		   "vbn $file_1: $vbn_1 \n".
		   "vbn $file_2: $vbn_2 \n".
		   "result: $result \n".
		   "expected: $expected \n");
	
	if ($result ne $expected) {
		logresult('fatal', "Failed to compare vbns for files $file_1 and $file_2, fbn $fbn:\nexpected: $expected\nactual: $result");
	}

	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}


# Internal function
# name ::_compare_vvbns_range
#
# summary (INTERNAL) compare vvbns range from 2 files
#
# description
# for 2 files, given a range of fbns,
# this compares the vvbn values
#
# example
#		$class->compare_vvbns_range(	file_1		=>$file_1, 
#										file_2		=>$file_2,
#										fbn_1		=>$fbn_start,
#										fbn_2		=>$fbn_end,
#										expected	=>$expected,
#								);
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _compare_vvbns_range
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		file_1 => { type=>SCALAR },
		file_2 => { type=>SCALAR },
		fbn_1 => { type=>SCALAR },
		fbn_2 => { type=>SCALAR },
		expected => { type=>SCALAR, default=>'EQUAL' },
	
	} );
	
	my %res;
	my $file_1		= $args{file_1};
	my $file_2		= $args{file_2};
	my $fbn_1		= $args{fbn_1};
	my $fbn_2		= $args{fbn_2};
	my $expected	= $args{expected};

	for (my $fbn = $fbn_1; $fbn <= $fbn_2; $fbn++) {
		$class->_compare_single_vvbn(file_1 => $file_1,
					     file_2 =>$file_2,
					     fbn =>$fbn,
					     expected =>$expected,
					     );		
	}

	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}


# Internal function
# name ::_compare_single_vvbn
#
# summary (INTERNAL) compare vvbns from 2 files
#
# description
# for 2 files, given a single fbn,
# this compares the vvbn values
#
# example
#		$class->_compare_single_vvbn(	file_1		=>$file_1, 
#										file_2		=>$file_2,
#										fbn			=>$fbn,
#										expected	=>$expected,
#									);	
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _compare_single_vvbn
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		file_1 => { type=>SCALAR },
		file_2 => { type=>SCALAR },
		fbn => { type=>SCALAR },
		expected => { type=>SCALAR, default=>'EQUAL' },
	} );
	
	my %res;
	my $file_1		= $args{file_1};
	my $file_2		= $args{file_2};
	my $fbn			= $args{fbn};
	my $expected	= $args{expected};
	
	my %vbn_info = $class->_get_vbn_info_single(filename => $file_1,
						    fbn => $fbn);
	
	my $vvbn_1 = $vbn_info{vbn_info}->{$fbn}->{vvbn};
	if (!defined($vvbn_1)) {
		logresult('fatal', "Failed to detect vvbn for file $file_1, fbn $fbn:\n" . Dumper(%vbn_info));
	}
	
	%vbn_info = $class->_get_vbn_info_single(filename => $file_2,
						 fbn => $fbn);
	my $vvbn_2 = $vbn_info{vbn_info}->{$fbn}->{vvbn};
	if (!defined($vvbn_2)) {
		logresult('fatal', "Failed to detect vvbn for file $file_2, fbn $fbn:\n" . Dumper(%vbn_info));
	}
	
	my $result = ($vvbn_1 == $vvbn_2) ? 'EQUAL' : 'NOT_EQUAL';
	logcomment("\nfbn: $fbn\n" .
		   "vvbn $file_1: $vvbn_1\n" .
		   "vvbn $file_2: $vvbn_2\n" .
		   "result: $result\n" .
		   "expected: $expected\n");
	
	if ($result ne $expected) {
		logresult('fatal', "Failed to compare vvbns for files $file_1 and $file_2, fbn $fbn\nexpected: $expected\nactual: $result");
	}

	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

# Internal function
# name ::_verify_compressed_range
#
# summary (INTERNAL) from showvbns, verify that a fbn range is compressed
#
# description
# from showvbns, verify that a fbn range is compressed
#
# example
#		$class->_compare_vbns_range(	filename	=>$filename,
#										fbn_1		=>$fbn_start,
#										fbn_2		=>$fbn_end,
#										expected	=>$expected, ( 'COMPRESSED' or 'NOT_COMPRESSED' )
#										max_allowed_uncompressed => 2 ( default=2 )
#								);
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _verify_compressed_range
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename	=> { type=>SCALAR },
		fbn_1		=> { type=>SCALAR },
		fbn_2		=> { type=>SCALAR },
		expected	=> { type=>SCALAR, default=>'COMPRESSED' },
		max_allowed_uncompressed => { type=>SCALAR, default=>2 },
		snapshot => { type=>SCALAR|UNDEF, optional=>1, },
		allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
	} );
	
	my %res;
	my $filename = $args{filename};
	my $fbn_1 = $args{fbn_1};
	my $fbn_2 = $args{fbn_2};
	my $expected = $args{expected};
	my $snapshot = $args{snapshot};
	my $max_allowed_uncompressed = $args{max_allowed_uncompressed};
	my $allow_removed_file = $args{allow_removed};
	my $max_allowed_compressed_data_blocks = C_MAX_CG_BLKS - 2;
	my $max_allowed_trailer_blocks = C_MAX_CG_BLKS - 2;
	my $num_headers = 0;
	my $num_trailers = 0;

	if ($class->{_debug} ) {
		logcomment("\n_verify_compressed_range\n".
			   "filename: $filename\n".
			   "fbn_1: $fbn_1\n".
			   "fbn_2: $fbn_2\n".
			   "expected: $expected\n".
			   "snapshot: $snapshot\n".
			   "max_allowed_uncompressed: $max_allowed_uncompressed\n".
			   "max_allowed_compressed_data_blocks: $max_allowed_compressed_data_blocks\n".
			   "max_allowed_trailer_blocks: $max_allowed_trailer_blocks\n".
			   "allow_removed_file: $allow_removed_file\n");
	}

	if ($fbn_2 <= $fbn_1) {
		logresult( "script", "Second fbn needs to be greater than first fbn." );
	}

	my $last_block_type = 'NONE';
	my $num_uncompressed_blocks_found = 0;
	my $num_consecutive_data_blocks = 0;
	my $num_consecutive_trailer_blocks = 0;
	my $current_data_location = 'UNKNOWN';

	for (my $fbn = $fbn_1; $fbn <= $fbn_2; $fbn++) {
		my %res = $class->_get_vbn_info_single(filename	=> $filename,
						       fbn => $fbn,
						       snapshot	=> $snapshot,
						       allow_removed => $allow_removed_file
						       );
		my $current_block_type = $class->_get_block_type(vbn => $res{'vbn'},
								 vvbn => $res{'vvbn'},
								 );

		if ($current_block_type eq 'HEADER')
		{
			$num_headers++;
			$num_consecutive_data_blocks = 0;
			$num_consecutive_trailer_blocks = 0;
			$current_data_location = 'IN_CG';
			
			if ($expected eq 'NOT_COMPRESSED') {
				croak "There should not be a HEADER in non compressed data";
			} elsif ($last_block_type eq 'HEADER') {
				croak "2 consecutive HEADER blocks were found $!";
			}
		}
		elsif ($current_block_type eq 'TRAILER')
		{
			$num_trailers++;
			$num_consecutive_data_blocks = 0;
			$num_consecutive_trailer_blocks++;
			
			if ($expected eq 'NOT_COMPRESSED') {
				croak "There should not be a TRAILER in non compressed data";
			}
			if ($num_consecutive_trailer_blocks > $max_allowed_trailer_blocks ) {
				croak "Exceeded max_allowed_trailer_blocks ( $max_allowed_trailer_blocks )\n".
				"$num_consecutive_trailer_blocks trailer blocks were found $!";
			}
		}
		elsif ($current_block_type eq 'DATA')
		{
			$num_consecutive_trailer_blocks = 0;
			$num_consecutive_data_blocks++;
			last if $expected eq 'NOT_COMPRESSED';
			
			if ($last_block_type eq 'TRAILER') {
				$current_data_location = 'NOT_IN_CG';
			}
			
			if ($current_data_location eq 'NOT_IN_CG') {
				logcomment("data: NOT_IN_CG") if $class->{_debug};
				$num_uncompressed_blocks_found++;
			}
			
			if ($expected eq 'COMPRESSED') {
				if ($num_uncompressed_blocks_found > $max_allowed_uncompressed ) {
					croak "Exceeded max_allowed_uncompressed ( $max_allowed_uncompressed )\n".
					"$num_uncompressed_blocks_found uncompressed blocks were found $!";
				}
				if ($num_consecutive_data_blocks > $max_allowed_compressed_data_blocks ) {
					croak "Exceeded max_allowed_compressed_data_blocks ( $max_allowed_compressed_data_blocks )\n".
					"$num_consecutive_data_blocks consecutive compressed blocks were found $!";
				}
			}
		}

		if ($class->{_debug} ) {
					logcomment(	
					"\ncurrent: $current_block_type\n".
					"num_uncompressed_blocks_found: $num_uncompressed_blocks_found\n".
					"num_consecutive_data_blocks: $num_consecutive_data_blocks\n".
					"num_consecutive_trailer_blocks: $num_consecutive_trailer_blocks\n".
					"current_data_location: $current_data_location\n"
					);
		}
		$last_block_type = $current_block_type;
		
	}
	
	if ($expected eq 'COMPRESSED') {
		if ($num_headers == 0) {
			croak "Compressed data range needs at least one header. None were found in this range. $!";
		}
		if ($num_trailers == 0) {
			croak "Compressed data range needs at least one trailer. None were found in this range. $!";
		}

	}

	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

# Internal function
# name ::_verify_compressed_from_dumpblock
#
# summary (INTERNAL) from dumpblock, verify that block is compressed
#
# description
# from dumpblock, verify that block is compressed
#
# example
#		$class->_verify_compressed_from_dumpblock(	filename	=>$filename,
#													file_id		=>$file_id,
#													fbn			=>$fbn,
#													expected	=>$expected,
#											);
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _verify_compressed_from_dumpblock
{
	my $class 	= shift;
	my (%args) = validate (@_, {filename	=> { type=>SCALAR|UNDEF, optional=>1 },
				    file_id		=> { type=>SCALAR|UNDEF, optional=>1 },
				    fbn			=> { type=>SCALAR },
				    expected	=> { type=>SCALAR, default=>'COMPRESSED' },
				    snapshot	=> { type=>SCALAR|UNDEF, optional=>1 },
				    allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
	} );
	my %res;
	my $filename	= $args{filename};
	my $file_id		= $args{file_id};
	my $fbn			= $args{fbn};
	my $expected	= $args{expected};
	my $snapshot	= $args{snapshot};

	my $result;

	if ($class->{_debug} ) {
		my $comment = "\n_verify_compressed_from_dumpblock\n";
		$comment .=	"filename: $filename\n" if $filename;
		$comment .=	"file_id: $file_id\n" if $file_id;
		$comment .=	"fbn: $fbn\n" if $fbn;
		$comment .=	"snapshot: $snapshot\n" if $snapshot;
		$comment .=	"expected: $expected\n" if $expected;
		logcomment($comment);
	}
	if ((! $filename) && (! $file_id)) {
		logresult("script", "Need to specify either filename or file_id");
	}
	
	if (! $file_id) {
		$file_id = $class->_get_file_id(filename => $filename,
						snapshot=>$snapshot,
						allow_removed => $args{allow_removed});
	}
	
	my $volume_name = $class->{vol_name}; 
	my $cmd_code = C_CMD_DUMPBLOCK_FILE;
	my $command = $class->{_ui}->formulate_command(code => $cmd_code,
						       volume => $volume_name,
						       file_path => $file_id,
						       fbn => $fbn);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
							    boot_mode => $class->{_boot_mode},
							    cmd => $command,
							    node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
	
	my $expected_rc = ($expected eq 'COMPRESSED') ? C_RC_DUMPED_COMPRESSED_BLOCK : C_RC_DUMPED_UNCOMPRESSED_BLOCK;
	return $class->{_ui}->process_output_string(code => $cmd_code,
						    output => $output,
						    expected_rc => $expected_rc,
						    log_message => "Dump $expected block for file id $file_id, fbn $fbn, volume $volume_name",);
}

# Internal function
# name ::_get_file_id
#
# summary (INTERNAL) get file id
#
# description
# get file id
#
# example
#		$class->_get_file_id(	filename	=>$filename, );
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _get_file_id
{
	my $class 	= shift;
	my (%args) = validate (@_, {filename	=> { type=>SCALAR },
				    snapshot	=> { type=>SCALAR|UNDEF, optional=>1, },
				    allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
	} );
	my $filename	= $args{filename};
	my $snapshot	= $args{snapshot};
	if ($class->{_debug} ) {
		logcomment("\n_get_file_id\n".
			   "filename: $filename\n"
		);
	}
	
	my $cmd_code = C_CMD_GET_FILE_ID;
	my $is_dblade_level = $class->{_ui}->is_dblade_level_command(code => $cmd_code);
	my $file_path = $class->_generate_file_path(filename => $filename,
						    snapshot => $snapshot,
						    allow_removed => $args{allow_removed},
						    need_vol_prepended => $is_dblade_level);
	
	my $command = $class->{_ui}->formulate_command(code => $cmd_code,
						       file_path => $file_path);
	
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
							    boot_mode => $class->{_boot_mode},
							    cmd => $command,
							    node_run => $is_dblade_level,);
	
	my %processed_res = $class->{_ui}->process_output_string(code => $cmd_code,
								 output => $output,
								 expected_rc => C_RC_SUCCESS,
								 log_message => "Get file id for file $file_path",);
	
	my $file_id = $processed_res{fh_info}->{fileid};
	
	return hex($file_id);
}




# Internal function
# name ::_verify_compressed_header
#
# example
#		$class->_verify_compressed_header(	vbn				=>$vbn, 
#											vvbn			=>$vvbn,
#											expected_state	=>COMPRESSED|NOT_COMPRESSED ( default COMPRESSED )
#									);	
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _verify_compressed_header
{
	my $class 	= shift;
	my (%args) = validate (@_, {vbn				=> { type=>SCALAR },
								vvbn			=> { type=>SCALAR },
								expected_state	=> { type=>SCALAR, default=>'COMPRESSED' },
	
	} );
	my %res;
	my $vbn				= $args{vbn};
	my $vvbn			= $args{vvbn};
	my $expected_state	= $args{expected_state};

	logcomment("\n".
				"verify_compressed_header\n".
				"vbn: $vbn\n".
				"vvbn: $vvbn\n".
				"expected_state: $expected_state\n");
	
	my %constants = $class->get_vbn_constants();
	my @vbn_headers = @{$constants{'vbn_headers'}};
	my @vvbn_headers = @{$constants{'vvbn_headers'}};

	if ($expected_state eq 'COMPRESSED') {
		if ( ! grep(/^$vbn$/, @vbn_headers) ) {
			croak "vbn:$vbn for header block should be one of: " . join(',', @vbn_headers);
		}
		if ( ! grep(/^$vvbn$/, @vvbn_headers) ) {
			croak "vvbn:$vvbn for header block should be one of: " . join(',', @vvbn_headers);
		}
	} else {
		if ( grep(/^$vbn$/, @vbn_headers) ) {
			croak "vbn:$vbn should not be one of: " . join(',', @vbn_headers);
		}
		if ( grep(/^$vvbn$/, @vvbn_headers) ) {
			croak "vvbn:$vvbn should not be one of: " . join(',', @vvbn_headers);
		}
	}
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

# Internal function
# name ::_verify_compressed_trailer
#
# example
#		$class->_verify_compressed_trailer(	vbn				=>$vbn, 
#											vvbn			=>$vvbn,
#											expected_state	=>COMPRESSED|NOT_COMPRESSED ( default COMPRESSED )
#									);	
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _verify_compressed_trailer
{
	my $class 	= shift;
	my (%args) = validate (@_, {vbn				=> { type=>SCALAR },
								vvbn			=> { type=>SCALAR },
								expected_state	=> { type=>SCALAR, default=>'COMPRESSED' },
	
	} );
	my %res;
	my $vbn				= $args{vbn};
	my $vvbn			= $args{vvbn};
	my $expected_state	= $args{expected_state};

	logcomment("\n".
				"verify_compressed_trailer\n".
				"vbn: $vbn\n".
				"vvbn: $vvbn\n".
				"expected_state: $expected_state\n");
	
	my %constants = $class->get_vbn_constants();
	my @vbn_trailers = @{$constants{'vbn_trailers'}};
	my @vvbn_trailers = @{$constants{'vvbn_trailers'}};

	if ($expected_state eq 'COMPRESSED') {
		if ( ! grep(/^$vbn$/, @vbn_trailers) ) {
			croak "vbn:$vbn for trailer block should be one of: " . join(',', @vbn_trailers);
		}
		if ( ! grep(/^$vvbn$/, @vvbn_trailers) ) {
			croak "vvbn:$vvbn for trailer block should be one of: " . join(',', @vvbn_trailers);
		}
	} else {
		if ( grep(/^$vbn$/, @vbn_trailers) ) {
			croak "vbn:$vbn should not be one of: " . join(',', @vbn_trailers);
		}
		if ( grep(/^$vvbn$/, @vvbn_trailers) ) {
			croak "vvbn:$vvbn should not be one of: " . join(',', @vvbn_trailers);
		}
	}
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

# Internal function
# name ::_get_block_type
#
# summary (INTERNAL) 
#	get block type
#
# description
#   get block type based on vbn and vvbn 
#
# example
#		$current_block_type = $class->_get_block_type(	vbn		=>$res{'vbn'}, 
#														vvbn	=>$res{'vvbn'},
#													);
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _get_block_type
{
	my $class 	= shift;
	my (%args) = validate (@_, {vbn	 => { type=>SCALAR },
								vvbn => { type=>SCALAR },
	
	} );
	my %res;
	my $vbn	 = $args{vbn};
	my $vvbn = $args{vvbn};
	my $block_type;

	my %constants = $class->get_vbn_constants();
	my @vbn_headers		= @{$constants{'vbn_headers'}};
	my @vvbn_headers	= @{$constants{'vvbn_headers'}};
	my @vbn_trailers	= @{$constants{'vbn_trailers'}};
	my @vvbn_trailers	= @{$constants{'vvbn_trailers'}};

	if ( grep(/^$vbn$/, @vbn_headers) ){
		if (grep(/^$vvbn$/, @vvbn_headers)) {
			$block_type = 'HEADER';
		} else {
			croak "mismatch between vbn and vvbn $!";
		}
	} elsif ( grep(/^$vbn$/, @vbn_trailers) ){
		if ( grep(/^$vvbn$/, @vvbn_trailers) ) {
			$block_type = 'TRAILER';
		} else {
			croak "mismatch between vbn and vvbn $!";
		}
	} else {
		$block_type = 'DATA';
	}
	return $block_type;
}

# Internal function
# name ::_verify_real_data
#
# summary (INTERNAL) 
#	verify block is real data
#
# description
#   based on vbn and vvbn, verify block is real data (i.e. not header or trailer )
#
# example
#			$class->_verify_real_data(	vbn		=>$res{'vbn'}, 
#										vvbn	=>$res{'vvbn'},
#									);
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _verify_real_data
{
	my $class 	= shift;
	my (%args) = validate (@_, {vbn	 => { type=>SCALAR },
								vvbn => { type=>SCALAR },
	
	} );
	my %res;
	my $vbn	 = $args{vbn};
	my $vvbn = $args{vvbn};

	logcomment("\n".
				"verify_real_data\n".
				"vbn: $vbn\n".
				"vvbn: $vvbn\n");

	my %constants = $class->get_vbn_constants();
	my @vbn_headers		= @{$constants{'vbn_headers'}};
	my @vvbn_headers	= @{$constants{'vvbn_headers'}};
	my @vbn_trailers	= @{$constants{'vbn_trailers'}};
	my @vvbn_trailers	= @{$constants{'vvbn_trailers'}};

	if ( grep(/^$vbn$/, @vbn_headers) ) {
		croak "for real data, vbn:$vbn should not be one of: " . join(',', @vbn_headers);
	}
	if ( grep(/^$vvbn$/, @vvbn_headers) ) {
		croak "for real data, vvbn:$vvbn should not be one of: " . join(',', @vvbn_headers);
	}
	if ( grep(/^$vbn$/, @vbn_trailers) ) {
		croak "for real data, vbn:$vbn should not be one of: " . join(',', @vbn_trailers);
	}
	if ( grep(/^$vvbn$/, @vvbn_trailers) ) {
		croak "for real data, vvbn:$vvbn should not be one of: " . join(',', @vvbn_trailers);
	}


}


# Internal function
# name ::_get_block_layout_range
#
# summary (INTERNAL) from showvbns, get block layout
#
# description
# from showvbns, get block layout
#
# example
#		$class->_get_block_layout_range(filename	=>$filename,
#										fbn_1		=>$fbn_start,
#										fbn_2		=>$fbn_end,
#								);
#
# author dl-compression-qa@netapp.com, waroff@netapp.com
#
sub _get_block_layout_range
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename => { type=>SCALAR },
		fbn_1 => { type=>SCALAR },
		fbn_2 => { type=>SCALAR },
	
	} );
	
	my $block_layout;
	my $filename = $args{filename};
	my $fbn_1 = $args{fbn_1};
	my $fbn_2 = $args{fbn_2};

	my %block_type_codes;
	$block_type_codes{'HEADER'}  = 'H';
	$block_type_codes{'TRAILER'} = 'T';
	$block_type_codes{'DATA'}    = 'D';


	if ($fbn_2 <= $fbn_1) {
		logresult( "script", "Second fbn needs to be greater than first fbn." );
	}

	for (my $fbn = $fbn_1; $fbn <= $fbn_2; $fbn++) {
		my %res = $class->_get_vbn_info_single(filename =>$filename,
						       fbn =>$fbn,);
		
		my $current_block_type = $class->_get_block_type(vbn =>$res{'vbn'},
								 vvbn =>$res{'vvbn'},);

		$block_layout->{$fbn} = $block_type_codes{$current_block_type};
		
	}
	return $block_layout;
}

# Internal subroutine, used only by VolClone
# to allow processing of the files whithin the cloned volume
sub _attach_to_cloned_dataset
{
	my $class = shift;
	my (%args) = validate( @_,
	{
                files => {type => HASHREF,},
		removed_files => {type => HASHREF,},
		ds_type => {type => SCALAR,},
		testdir => {type => SCALAR|UNDEF,},
		testdir_path => {type => SCALAR,},
		parent_path => {type => SCALAR,},
		parent_mountpoint => {type => SCALAR,},
		reset_info => {type => SCALAR, default => FALSE,},
         } );
	
	logcomment("Dataset: _attach_to_cloned_dataset input:\n" . Dumper(%args)) if ($class->{_debug});
	
	my $files = $args{files};
	my $removed_files = $args{removed_files};
	my $parent_path = $args{parent_path};
	my $parent_mountpoint = $args{parent_mountpoint};
	my $parent_testdir_path = $args{testdir_path};
	my $reset_info = $args{reset_info};
	
	
	# We need now to substitute the parent's mountpoint, filer path with the clone's paths
	$class->_import_file_hash_from_clone(files => $files,
					     key_to_clasify => 'files',
					     parent_path => $parent_path,
					     parent_mountpoint => $parent_mountpoint,
					     reset_info => $reset_info,);
	
	$class->_import_file_hash_from_clone(files => $removed_files,
					     key_to_clasify => '_removed_files',
					     parent_path => $parent_path,
					     parent_mountpoint => $parent_mountpoint,
					     reset_info => $reset_info,);
	
	if ($parent_testdir_path =~ /^$parent_mountpoint/)
	{
		my $clone_mountpoint = $class->{mountpoint};
		$parent_testdir_path =~ s/^$parent_mountpoint/$clone_mountpoint/;
	}
	else
	{
		logcomment("_attach_to_cloned_dataset: Not processing testdir_path \"$parent_testdir_path\":\n" .
			   "It does not start with the volume mountpoint, as expected: $parent_mountpoint") if ($class->{_debug});
	}
	$class->{testdir_path} = $parent_testdir_path;
	
	$class->{ds_type} = $args{ds_type};
	$class->{testdir} = $args{testdir};
	
	if ($class->{_debug})
	{
		my $comment = "Dataset: _attach_to_cloned_dataset: finished attaching to dataset:\n" .
		              "\nFiles hash:\n" . Dumper($class->{files}) .
			      "\nRemoved files hash:\n" . Dumper($class->{_removed_files}) .
			      "\nds_type: " . $class->{ds_type} .
			      "\ntestdir_path: " . $class->{testdir_path};
		if (defined($class->{testdir}))
		{
			$comment .= "\ntestdir: " . $class->{testdir};
		}
		logcomment($comment);
	}
}


# Internal subroutine, used only by VolClone
sub _import_file_hash_from_clone
{
	my $class = shift;
	my (%args) = validate( @_,
	{
                files => {type => HASHREF,},
		parent_path => {type => SCALAR,},
		parent_mountpoint => {type => SCALAR,},
		key_to_clasify => {type => SCALAR,},
		reset_info => {type => SCALAR, default => FALSE,},
         } );
	
	logcomment("Dataset: _import_file_hash_from_clone input\n" . Dumper(%args)) if ($class->{_debug});
	
	my $files = $args{files};
	my $parent_path = $args{parent_path};
	my $clone_path = $class->_get_vol_mount_path();
	my $parent_mountpoint = $args{parent_mountpoint};
	my $clone_mountpoint = $class->{mountpoint};
	my $dataset_key = $args{key_to_clasify};
	my $reset_info = $args{reset_info};
	
	if ($reset_info == TRUE) {
		$class->{$dataset_key} = {};
	}
	
	foreach my $file (keys %$files)
	{
		logcomment("Dataset: _import_file_hash_from_clone input: importing file \"$file\"") if ($class->{_debug});
		
		# Process the filer path
		my $filer_path = $files->{$file}->{filer_path};
		logcomment("Dataset: _import_file_hash_from_clone input: processing: $filer_path") if ($class->{_debug});
		if ($filer_path =~ /^$parent_path/)
		{
			$filer_path =~ s/^$parent_path/$clone_path/;
			logcomment("Dataset: _import_file_hash_from_clone input: processed: $filer_path") if ($class->{_debug});
		}
		else
		{
			logcomment("_import_file_hash_from_clone: Not processing filer_path \"$filer_path\":\n" .
				   "It does not start with the volume path, as expected: $parent_path") if ($class->{_debug});
		}
		
		$class->{$dataset_key}->{$file}->{filer_path} = $filer_path;
		
		# Process the path on which the file was mounted on the client
		my $mounted_path = $files->{$file}->{path};
		logcomment("Dataset: _import_file_hash_from_clone input: processing: $mounted_path") if ($class->{_debug});
		if ($mounted_path =~ /^$parent_mountpoint/)
		{
			$mounted_path =~ s/^$parent_mountpoint/$clone_mountpoint/;
			logcomment("Dataset: _import_file_hash_from_clone: processed: $mounted_path") if ($class->{_debug});
		}
		else
		{
			logcomment("_import_file_hash_from_clone: Not processing path \"$mounted_path\":\n" .
				   "It does not start with the volume mountpoint, as expected: $parent_mountpoint") if ($class->{_debug});
		}
		
		$class->{$dataset_key}->{$file}->{path} = $mounted_path;
		
		# Entries that do not need to be changed
		$class->{$dataset_key}->{$file}->{name} = $files->{$file}->{name};
		$class->{$dataset_key}->{$file}->{relative_path} = $files->{$file}->{relative_path};
	}
}


sub _add_attempt_file_clone_dest
{
	my $class = shift;
	my (%args) = validate( @_,
	{
                filename => {type => SCALAR,},
		clonename => {type => SCALAR,},
		allow_removed => {type => SCALAR, default => FALSE,},
         } );
	
	my $filename = $args{filename};
	my $clonename = $args{clonename};
	my $allow_removed = $args{allow_removed};
	my $file;
	
	my $file_source_exists = $class->_filename_exists(filename => $filename, expected => C_RC_FAILURE_NON_FATAL);
	if ($file_source_exists == FALSE) {
		if ($allow_removed == FALSE) {
			logresult('fatal', "File $filename does not exist");
		}
		$class->_file_was_removed(filename => $filename);
		$file = $class->{_removed_files}->{$filename};
	}
	else {
		$file = $class->{files}->{$filename};
	}
	
	
	
	$class->_filename_exists(filename => $clonename, expected => FALSE);
	
	my $clone = {};
	$clone->{name} = $clonename;
	my $path = $file->{path};
	$path =~ s/$filename$/$clonename/;
	$clone->{path} = $path;
	
	$path = $file->{relative_path};
	$path =~ s/$filename$/$clonename/;
	$clone->{relative_path} = $path;
	
	$path = $file->{filer_path};
	$path =~ s/$filename$/$clonename/;
	$clone->{filer_path} = $path;

	my $comment = "creating new clone info for $clonename\n";
	$comment .= "path: ". $clone->{path} ."\n";
	$comment .= "relative_path: ". $clone->{relative_path} ."\n";
	$comment .= "filer_path: ". $clone->{filer_path} ."\n";
	logcomment($comment);
	
	$class->{attempt_clones}->{$clonename} = $clone;

	return $clone->{filer_path};
}


sub _add_file_clone_dest
{
	my $class = shift;
	my (%args) = validate( @_,
	{
                clonename => {type => SCALAR,},
         } );
	
	my $clonename = $args{clonename};
	
	if (!exists($class->{attempt_clones}->{$clonename}))
	{
		logresult('script', "Clone info is not present. Path was not generated by the compression library...")
	}
	
	$class->_new_file(file => $class->{attempt_clones}->{$clonename});
	
	delete($class->{attempt_clones}->{$clonename});
}


sub _attach_lun_clone_info
{
        my $class = shift;
        my (%args) = validate( @_,
	{
		source_lun => {type => OBJECT, isa => "Compression::Lun"},
		clone_lun => {type => OBJECT, isa => "Compression::LunClone"},
		source_dataset => {type => OBJECT, isa => "Compression::Dataset", optional => 1,},
		_lun_of_cloned_vol => {type => SCALAR, default => FALSE,},
        });
	
	my $source_lun = $args{source_lun};
	my $clone_lun = $args{clone_lun};
	my $source_dataset = $args{source_dataset};
	my $lun_of_cloned_vol = $args{_lun_of_cloned_vol};
	my $files = (defined($source_dataset)) ? ($source_dataset->{_c_dataset}->{_common}->{files}) : $class->{files};
	my $removed_files = (defined($source_dataset)) ? ($source_dataset->{_c_dataset}->{_common}->{_removed_files}) : $class->{_removed_files};
	my ($file, $file_path);
	
	my $source_lun_mount = $source_lun->get_mountpoint();
	my $clone_lun_mount = $clone_lun->get_mountpoint();
	my $clone_lun_name = $clone_lun->{name};
	
	foreach $file (keys(%$files)) {
		$file_path = $files->{$file}->{path};
		if ($file_path =~ /^$source_lun_mount/) {
			$file_path =~ s/^$source_lun_mount/$clone_lun_mount/;
			if (TRUE == $lun_of_cloned_vol) {
				$files->{$file}->{path} = $file_path;
			}
			else {
				$class->{_lun_clones}->{$clone_lun_name}->{files}->{$file}->{name} = $file;
				$class->{_lun_clones}->{$clone_lun_name}->{files}->{$file}->{path} = $file_path;
				$class->{_lun_clones}->{$clone_lun_name}->{files}->{$file}->{filer_path} = undef;
				$class->{_lun_clones}->{$clone_lun_name}->{files}->{$file}->{relative_path} = undef;
			}
		}
	}
	
	foreach $file (keys(%$removed_files)) {
		$file_path = $removed_files->{$file}->{path};
		if ($file_path =~ /^$source_lun_mount/) {
			$file_path =~ s/^$source_lun_mount/$clone_lun_mount/;
			if (TRUE == $lun_of_cloned_vol) {
				$removed_files->{$file}->{path} = $file_path;
			}
			else {
				$class->{_lun_clones}->{$clone_lun_name}->{removed_files}->{$file}->{name} = $file;
				$class->{_lun_clones}->{$clone_lun_name}->{removed_files}->{$file}->{path} = $file_path;
				$class->{_lun_clones}->{$clone_lun_name}->{removed_files}->{$file}->{filer_path} = undef;
				$class->{_lun_clones}->{$clone_lun_name}->{removed_files}->{$file}->{relative_path} = undef;
			}
		}
	}
	
	if (TRUE == $lun_of_cloned_vol)
	{
		my $testdir_path = (defined($source_dataset)) ? ($source_dataset->{_c_dataset}->{_common}->{testdir_path}) : $class->{testdir_path};
		
		if ((defined($testdir_path)) && ($testdir_path =~ /^$source_lun_mount/)) {
			$testdir_path =~ s/^$source_lun_mount/$clone_lun_mount/;
			if (defined($source_dataset)) {
				$source_dataset->{_c_dataset}->{_common}->{testdir_path} = $testdir_path;
			}
			else {
				$class->{testdir_path} = $testdir_path;
			}
		}
	}
	
}

sub _update_vol_obj
{
	my $class = shift;
	my $new_vol = shift;
	
	if ($class->{_boot_mode} eq C_ONTAP_7MODE)
	{
		if (!($new_vol->isa("NOTE::Volume")))
		{
			logresult('script', "For 7-mode, the volume argument is required to be a NOTE\:\:Volume object");
		}
		$class->{volume} = $new_vol;
		$class->{_conn} = $class->{volume}->Filer();
		$class->{vol_name} = $class->{volume}->name;
	}
	elsif ($class->{_boot_mode} eq C_ONTAP_CMODE)
	{
		if (!($new_vol->isa("NOTE::GX::Volume")))
		{
			logresult('script', "For 10-mode, the volume argument is required to be a NOTE\:\:GX\:\:Volume object");
		}
		$class->{volume} = $new_vol;
		$class->{_conn} = $class->{volume}->Node();
		$class->{vol_name} = $class->{volume}->name;
	}
}


1;
