#
# $Id: //depot/prod/test/main/lib/Compression/Dataset/NFS.pm $
#
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)


package Compression::Dataset::NFS;

use strict;
use warnings;

use Tharn;

use Compression::Dataset::Common;
use Module::Common;
use Data::Dumper;
use SPARSE::Resource::Volume;
use SPARSE::Resource::Filer;
use SPARSE::Util::FileIO;
use SPARSE::Util::Logger;
use SPARSE::Util::Utilities;
use FS::Util::Utilities qw(file_reserve dumpblock);
use Dense::Utils;
use Dense::TestData;
use snaplock qw(client_get_atime system_time_get client_set_atime);
use cat qw(do_shell);
use File::Copy;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Compression::Constants qw(:all);
use Compression::Utils qw(_send_cmd_to_filer);
use File::Compare;
use File::Basename;
use Carp qw(croak);
use FileManip;
use filer;
use Compression::Dataset::Constants qw(:all);
use Digest::SHA qw(sha1);
use SnapCmds;

BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}


## @name ::new
##
## @summary perform dataset initialization
##
## @description
## Perform all the basic dataset initialization required here
##
## @return Compression::Dataset object
##
## @arg volume Required: NOTE Volume object
## @arg client Optional: name of the client
## @arg debug Optional: logs additional debug messages specific to this module
##
## @examples
## 	$self->{dataset}  = Compression::Dataset->new(volume => $self->{volume}->{_vol},
##						      client => $self->{client}->{name} );
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
##

sub new
{
	my $class 	= shift;
	my (%args) = validate (@_, {volume    => { type=>OBJECT, },
				    vserver   => { type=>SCALAR|UNDEF, optional => 1 },
				    client    => { type=>SCALAR, },
				    skip_mount     => { type=>BOOLEAN, optional => 1, default=>FALSE },
				    debug     => { type=>SCALAR, optional => 1, default=>FALSE },
				    ui        => {type => OBJECT, isa => "Compression::UI"},
				    boot_mode => {type => SCALAR, optional => 1, default => C_ONTAP_BOOT_MODE},
				    has_debug_kernel => {type => SCALAR, optional => 1, default => TRUE}, # So that scripts can continue
				    vfiler_mountpoint => {type => SCALAR|UNDEF, optional => 1,},  
		} );	
	my $self = {};
	$self->{ID} = $$;
	$self->{Description} = "Dataset create/modify";
	$self->{volume} = $args{volume};
	$self->{_vserver} = $args{vserver};
	$self->{client} = ( defined $args{client} )? $args{client} : "localhost";
	$self->{_ui} = $args{ui};
	$self->{_debug} = $args{debug};
	$self->{_has_debug_kernel} = $args{has_debug_kernel};
	$self->{skip_mount} = $args{skip_mount};
	delete ($args{skip_mount});
	$self->{_vfiler_mountpoint} = delete($args{vfiler_mountpoint}); 
	
	$self->{_boot_mode} = $args{boot_mode};
	if ($self->{_boot_mode} eq C_ONTAP_7MODE) 
	{
		if (!($self->{volume}->isa("NOTE::Volume")))
		{
			logresult('script', "For 7-mode, the volume argument is required to be a NOTE\:\:Volume object");
		}
		$self->{_conn} = $self->{volume}->Filer();
		$self->{vol_name} = $self->{volume}->name;
		# At this moment, creatig a NOTE::Client object after the vfiler creation is not possible
		# Volume mount fails as the volume is exported by the vfiler
		if (!defined($self->{_vfiler_mountpoint}))
		{
			$self->{_client} = NOTE::Client->new(name => $self->{client}, storage => $self->{volume} );
		}
	}
	elsif ($self->{_boot_mode} eq C_ONTAP_CMODE) 
	{
		if (!($self->{volume}->isa("NOTE::GX::Volume")))
		{
			logresult('script', "For 10-mode, the volume argument is required to be a NOTE\:\:GX\:\:Volume object");
		}
		$self->{_conn} = $self->{volume}->Node();
		$self->{vol_name} = $self->{volume}->name;
		# Skip mounting so we specifically mount rw (just as Volume does), to get the .admin mountpoint
		$self->{_client} = NOTE::Client->new(name => $self->{client}, storage => $self->{volume}, skip_remount => 1 );
		param('NOTE_NFSMOUNTOPTS', "-set", "hard,intr,proto=tcp");
		if (!$self->{skip_mount}){	# skip mount when auto_mount set to false
			$self->{volume}->mount( client => $self->{client}, RW => 1 );
		}
	}
	else
	{
		logresult('script', "Unknown boot_mode: " . $self->{_boot_mode});
	}
	
	
	delete $args{client};
	$self->{_common} = Compression::Dataset::Common->new(%args);
	$self->{sfsr_will_restore_file_as_compressed} = $self->{_ui}->get_constant_value( constant_code => 'SFSR_WILL_RESTORE_FILE_AS_COMPRESSED' );
	
	bless $self, $class;

	if (!$self->{skip_mount}){
		$self->{_common}->{mountpoint} = $self->get_mountpoint();
		$self->{_common}->{testdir_path} = $self->{_common}->{mountpoint};
	}else {
		$self->{_common}->{mountpoint} = '';
		$self->{_common}->{testdir_path} = '';
	}
	logcomment("Dataset Initialized with ". Dumper($self) ) if ( $self->{_debug} );

	return $self;

}




## @name ::create
##
## @summary create the dataset or set of files on the volume
##
## @description
## Create set of files or specific dataset for the volume
##
## @return hash of status and output
##
##
## @arg	files		Required: Array of hashrefs ( one for each file )
##		files => [{name	=> "f1", 
##                         offset => 0, #default=0
##                         size => ( 4 * KB),
##                         type => 'uncompressable|compressable|existing_data|write_bytes|blocksondisk|large', #default='write_bytes'
##                         compression => 0,
##                         blocksondisk => 0,
##                         pattern => 'x', #default='a'
##                         data => $data, #only use if type=existing_data
##                         },
##                        ]
##
## @arg	client		Optional: Client
## @arg	path		Optional: Defaults to root of volume
## @arg	testdir		Optional: Test directory
## @arg	ds_type		Optional: Dataset Type
## @arg compression	Optional: File compressability (0 - 100)
## @arg blocksondisk	Optional: Number of blocks per compression group written on disk
## @arg	offset		Optional: Offset to for writing files ( overriden by offset for individual file, default is 0 ).
## @arg	len             Optional: Length of data to be written ( overriden by size of individual file )
## @arg	pattern		Optional: Default 'a'
## @arg	background	Optional: Run in background ( TRUE|FALSE )
## @arg dont_die_on_error	Optional: If volume is full, script won't die ( TRUE|FALSE ).  Setting this to TRUE has same effect as 
##		setting expected_result to C_RC_FAILURE_NON_FATAL.
## @arg expected_result	Optional: Croak if not expected ( C_RC_SUCCESS(default)|C_RC_INSUFFICIENT_SPACE|C_RC_FAILURE_NON_FATAL). If 
##						dont_die_on_error is TRUE, expected_result defaults to C_RC_FAILURE_NON_FATAL )
## @arg	lun		Optional: Compression::Lun object corresponding to the lun in which the file needs to be created.
##
## @examples
##
## %res = $dataset->create(files => [{name => "f1", size => GB, type=>'uncompressable',},],);
##
## %res = $dataset->create(files => [{ name => "f1", offset=>0, size => ( 4 * KB) },],
##                         background => TRUE,);
## $dataset->wait_finish(filename => 'f1');
##
## my $data = $dataset->get_uncompressable_data(size=>4 * KB);
## %res = $dataset->create(files => [{ name => "f1", data => $data, type => 'existing_data' },],);
## 
## @notes
##	type can be:
##		1. write_bytes		-(default) repeats a pattern
##		2. existing_data	-need to call get_uncompressable_data() first if using this option
##		3. uncompressable	-create uncompressable data
##		4. compressable		-creates a file with a certain percentage of compressable blocks
##		5. blocksondisk		-each compression group in a file compresses to a certain number of blocks
##		4. large			-originally for large files ( will be depreciated )
##	pattern needs to be specified if type is write_bytes or large ( default is 'a' )
##	data only should be used if type=existing_data
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com, kozel@netapp.com
##

sub create
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		files           => { type=>ARRAYREF },
		client		=> { type=>SCALAR, optional => 1 },
		qtree		=> { type=>OBJECT, optional => 1, isa => "Compression::Qtree", },
		path		=> { type=>SCALAR, optional => 1 },
		testdir		=> { type=>SCALAR, optional => 1 },
		ds_type		=> { type=>SCALAR, optional => 1 },
		compression	=> { type=>SCALAR, optional => 1 },
		blocksondisk	=> { type=>SCALAR, optional => 1 },
		offset		=> { type=>SCALAR, optional => 1 },
		len             => { type=>SCALAR, optional => 1 },
		pattern		=> { type=>SCALAR, optional => 1 },
		background	=> { type=>SCALAR, optional => 1, default=>FALSE },
		dont_die_on_error => { type=>SCALAR, optional=>1, default=>FALSE },
		disable_checksums => { type=>SCALAR, optional=>1, default=>TRUE },
		expected_result => { type=>SCALAR, optional=>1, default=>C_RC_SUCCESS },
		lun => {type => OBJECT, isa => "Compression::Lun", optional => 1,}, 
	} );
	
	my @files               = @{$args{files}};
	my $vol_mountpoint      = $class->get_mountpoint();
	my $client              = ( defined $args{client} )? $args{client} : "localhost";
	my $testdir		= ( defined $args{testdir} )? $args{testdir} : $class->{_common}->{testdir};
	my $ds_type		= ( defined $args{ds_type} )? $args{ds_type} : C_DFLT_COMPRESSION_DATASET;
	my $compression	        = ( defined $args{compression} )? $args{compression} : -1;
	my $blocksondisk        = ( defined $args{blocksondisk} )? $args{blocksondisk} : -1;
	my $offset 		= ( defined $args{offset} )? $args{offset} : 0;
	my $len			= ( defined $args{len} )? $args{len} : "1024";
	my $pattern		= ( defined $args{pattern} )? $args{pattern} : "a";
	my $background	        = ( defined $args{background} )? $args{background} : FALSE;
	my $path 		= ( defined $args{path} )? $args{path} : $vol_mountpoint;
	my $dont_die_on_error	= $args{dont_die_on_error};
	my $disable_checksums	= $args{disable_checksums};
	my $expected_result     = $args{expected_result};
	my $qtree               = $args{qtree};
	my $lun                 = $args{lun};
	my $clone_lun;
	if (defined($lun)) {
		if (defined($args{path})) {
			logresult('script', "Dataset\:\:create: path cannot be used along with lun");
		}
		if (($lun->isa("Compression::LunClone")) && (FALSE == $lun->{_cloned_volume})) {
			$clone_lun = $lun->{name};
		}
		my $lun_mountpoint = $lun->get_mountpoint();
		$path = $lun_mountpoint;
	}

	logcomment("In dataset create\n");
	if (($background == TRUE) && ($dont_die_on_error == TRUE)) {
		logresult('script', '"background" and "dont_die_on_error" can not both be TRUE');
	}

	$expected_result = C_RC_FAILURE_NON_FATAL if $dont_die_on_error == TRUE;
	$dont_die_on_error = TRUE if $expected_result == C_RC_FAILURE_NON_FATAL;
	

	#################################################################
	# debugging....
	$class->ls(path => $path); 
	#################################################################
	my %res;
	if ( defined $qtree ) {
		$path .= '/'. $qtree->{name};
	}

	# 
	# Create testdir if needed
	#
	my $testdir_path = $path; 
	if ($testdir) {
		$testdir_path .= '/'. $testdir;
		if (! -e $testdir_path ) {
			logcomment("testdir_path: $testdir_path");
			do_shell("sudo mkdir -m 777 $testdir_path");
		}
	}

	#
	# Check filenames unique
	#
	my %seen_filenames;
	foreach my $file ( @files )
	{
		if ( $seen_filenames{ $file->{name} } ) {
			croak "filenames must be unique $!";
		}
		$seen_filenames{$file->{name}} = 1;
	}

	#
	# Create object with path info for each file
	#
	foreach my $file ( @files )
	{
		my $filename = $file->{name};
		$class->_new_file(filename=>$filename,
				  testdir=>$testdir,
				  path=>$path,
				  clone_lun => $clone_lun,
		);
		$file->{path} = $class->get_path(filename => $filename, clone_lun => $clone_lun);
	}

	#
	# Save other class info
	#
	$class->{_common}->{ds_type} = $ds_type;
	$class->{_common}->{testdir} = $testdir;
	$class->{_common}->{testdir_path} = $testdir_path;

	if ($ds_type eq C_DATASET_SPARSE)
	{
		logcomment("In dataset create ds_type = $ds_type");
		eval
		{
			foreach my $file ( @files )
			{
				my $filename = $file->{name};
				logcomment("adding file = $file->{path}");
				
				my $offset	= ( defined $file->{offset} )	? $file->{offset} 	: $offset;
				my $len		= ( defined $file->{len} )		? $file->{len} 		: $file->{size};
				my $type	= ( defined $file->{type} )		? $file->{type} 	: 'write_bytes';
				my $compression = ( defined $file->{compression} ) ? $file->{compression}	: $compression;
				my $blocksondisk = ( defined $file->{blocksondisk} ) ? $file->{blocksondisk}	: $blocksondisk;
				my $pattern	= ( defined $file->{pattern} )	? $file->{pattern} 	: $pattern;
				my $data	= ( defined $file->{data} )		? $file->{data} 	: '';
				my $display_pattern = substr($pattern, 0, 12) . '.....';
				my $expected_result	= ( defined $file->{expected_result} )	? $file->{expected_result} 	: $expected_result;
				
				$type = 'uncompressable' if $compression == 0;
				$type = 'blocksondisk' if $blocksondisk != -1;
				
				my $file_path = $file->{path};
				
				my $comment = "\n".
				              "\t\tcreate file $file_path\n".
					      "\t\toffset: $offset\n".
					      "\t\ttype: $type\n".
					      "\t\texpected_result: " . $expected_result ."\n";
				$comment .= "\t\tlen: $len\n" if $len;
				$comment .= "\t\tpattern: $display_pattern\n" if $type ne 'uncompressable';
				$comment .= "\t\tcompression: $compression\n" if $compression != -1;
				$comment .= "\t\tblocksondisk: $blocksondisk\n" if $blocksondisk != -1;
				logcomment($comment);
				
				my ($res_1, $res_2);
				eval {
					logcomment("touch $file_path");
					$res_1 = `touch $file_path 2>&1`;
					logcomment("chmod 0777 $file_path");
					$res_2 = `chmod 0777 $file_path 2>&1`;
				};
				
				if ($res_1 =~ /No space left on device/i) {
					$res{'rc'} = C_RC_INSUFFICIENT_SPACE;
					$res{'error'} = 'No space left on device';
					logcomment('No space left on device');
					if (( $expected_result != C_RC_INSUFFICIENT_SPACE) &&
					    ( $expected_result != C_RC_FAILURE_NON_FATAL )) {
								croak "No space left on device $!";
					}
					return %res;
				}
				
				if ($background == TRUE) {
					$class->{_common}->{files}->{$filename}->{bg_subtest} = subtest( sub{
						$class->_write_data(path => $file_path,
								    offset => $offset,
								    len => $len,
								    pattern => $pattern,
								    type => $type,
								    compression	=> $compression,
								    blocksondisk => $blocksondisk,
								    data => $data,
								    disable_checksums => $disable_checksums,
								    expected_result => $expected_result,
								    );
						},
						'-bg',
						'-runid','create_file_'. $filename,
						);
				}
				else {
					%res = $class->_write_data(path	=> $file_path,
								   offset => $offset,
								   len => $len,
								   pattern => $pattern,
								   type	=> $type,
								   compression	=> $compression,
								   blocksondisk	=> $blocksondisk,
								   data	=> $data,
								   disable_checksums => $disable_checksums,
								   expected_result => $expected_result,
								   );
							
				}
			} 
		};
		
		if( $@ )
		{
			logresult("fatal", "$@");
		}
	}
	else
	{
		logresult( "fatal", "Unsupported dataset type $ds_type" );
	} 
	
	if ($class->{_debug}) { $class->ls(); }
	
	return %res;
}
## @name ::modify
##
## @summary modify the dataset or set of files on the volume
##
## @description
## Modify set of files or specific dataset for the volume
##
## @return hash of status and output
##
##
## @arg	files		Required: Array of hashrefs ( one for each file )
##		files => [
##					{ 
##						name	=> "f1", 
##						offset	=> ( 4 * KB),
##						len		=> ( 8 * KB), 
##						type	=> 'uncompressable|existing_data|write_bytes', #default='write_bytes'
##						pattern	=> 'x', #default='a'
##						data	=>	$data, #only use if type=existing_data
##					},
##				]
## @arg	client		Optional: Client
## @arg	path		Optional: Defaults to root of volume
## @arg	ds_type		Optional: Dataset Type
## @arg	offset		Optional: Offset to for writing files ( overriden by offset for individual file, default is 0 ).
## @arg	len			Optional: Length of data to be written ( overriden by len of individual file )
## @arg	pattern		Optional: Default 'a'
## @arg	background	Optional: Run in background ( TRUE|FALSE )
## @arg dont_die_on_error	Optional: If volume is full, script won't die ( TRUE|FALSE ).  Setting this to TRUE has same effect as 
##		setting expected_result to C_RC_FAILURE_NON_FATAL.
## @arg expected_result	Optional: Croak if not expected ( C_RC_SUCCESS(default)|C_RC_INSUFFICIENT_SPACE|C_RC_FAILURE_NON_FATAL). If 
##						dont_die_on_error is TRUE, expected_result defaults to C_RC_FAILURE_NON_FATAL )
##
## @examples
##
## %res = $dataset->modify( 
##		files => [
##					{ name => "f1", offset=>( 4 * KB), len => ( 8 * KB), },
##				],
##		background => TRUE,
##	);
## $dataset->wait_finish( filename => 'f1' );
##
##
## %res = $dataset->modify( 
##		files => [
##					{ name => "f1", offset=>( 4 * KB), len => ( 8 * KB), type=>'uncompressable' },
##				],
##	);
##
##
## my $data = $dataset->get_uncompressable_data(size=>4 * KB);
## %res = $dataset->modify( 
##		files => [
##					{ name => "f1", data=>$data, type=>'existing_data' },
##				],
##	);
## 
##	Notes:
##	type can be:
##		1. write_bytes		-(default) repeats a pattern
##		2. existing_data	-need to call get_uncompressable_data() first if using this option
##		3. uncompressable	-create uncompressable data
##		4. large			-originally for large files ( will be depreciated )
##		5. holey			-add holey blocks
##	
##	data only should be used if type=existing_data
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com, kozel@netapp.com
##
sub modify
{
	my $class 	= shift;
	my (%args) = validate (@_, {	
		files		=> { type=>ARRAYREF },
		client		=> { type=>SCALAR, optional => 1 },
		path		=> { type=>SCALAR, optional => 1 },
		fill		=> { type=>SCALAR, optional => 1 },
		ds_type		=> { type=>SCALAR, optional => 1 },
		compression	=> { type=>SCALAR, optional => 1 },
		blocksondisk	=> { type=>SCALAR, optional => 1 },
		offset		=> { type=>SCALAR, optional => 1 },
		len             => { type=>SCALAR, optional => 1 },
		pattern		=> { type=>SCALAR, optional => 1 },
		background	=> { type=>SCALAR, optional => 1, default=>FALSE },
		dont_die_on_error => { type=>SCALAR, optional=>1, default=>FALSE },
		disable_checksums => { type=>SCALAR, optional=>1, default=>TRUE },
		expected_result => { type=>SCALAR, optional=>1, default=>C_RC_SUCCESS },
		clone_lun => {type => SCALAR|UNDEF, optional => 1,},
	} );
	
	my @files             = @{$args{files}};
	my $vol_mountpoint    = $class->get_mountpoint();
	my $client            = ( defined $args{client} )? $args{client} : "localhost";
	my $path              = ( defined $args{path} )? $args{path} : $vol_mountpoint;
	my $fill              = ( defined $args{fill} )? $args{fill} : 0;
	my $offset            = ( defined $args{offset} )? $args{offset} : 0;
	my $len               = ( defined $args{len} )? $args{len} : "1024";
	my $compression       = ( defined $args{compression} )? $args{compression} : -1;
	my $blocksondisk      = ( defined $args{blocksondisk} )? $args{blocksondisk} : -1;
	my $pattern           = ( defined $args{pattern} )? $args{pattern} : "a";
	my $samedata          = ( defined $args{samedata} )? $args{samedata} : 0;
	my $background	      = ( defined $args{background} )? $args{background} : FALSE;
	my $dont_die_on_error = $args{dont_die_on_error};
	my $disable_checksums = $args{disable_checksums};
	my $expected_result   = $args{expected_result};
	my $ds_type           = $class->{_common}->{ds_type};
	
	logcomment("In dataset modify\n");
	if (($background == TRUE) && ($dont_die_on_error == TRUE)) {
		logresult('script', '"background" and "dont_die_on_error" can not both be TRUE');
	}
	
	$expected_result = C_RC_FAILURE_NON_FATAL if $dont_die_on_error == TRUE;
	$dont_die_on_error = TRUE if $expected_result == C_RC_FAILURE_NON_FATAL;

	my %res;

	#################################################################
	# debugging....
	$class->ls();
	#################################################################

	if( $samedata )
	{
		$pattern = "a";
	}
	elsif ( ! $pattern )
	{
		$pattern = "b";
	}

	
	if ($ds_type eq C_DATASET_SPARSE)
	{
		logcomment("In dataset modify C_DATASET_SPARSE\n");
		eval
		{
			foreach my $file ( @files )
			{
				my $filename = $file->{name};
				my $file_path =  $class->{_common}->get_path(filename => $filename, clone_lun => $args{clone_lun});
				
				my $offset	= ( defined $file->{offset} )	? $file->{offset} 	: $offset;
				my $len		= ( defined $file->{len} )		? $file->{len} 		: $len;
				my $type	= ( defined $file->{type} )		? $file->{type} 	: 'write_bytes';
				my $compression	= ( defined $file->{compression} ) 	? $file->{compression}	: $compression;
				my $blocksondisk = ( defined $file->{blocksondisk} )	? $file->{blocksondisk}	: $blocksondisk;
				my $pattern	= ( defined $file->{pattern} )	? $file->{pattern} 	: $pattern;
				my $data	= ( defined $file->{data} )		? $file->{data} 	: '';
				my $display_pattern = substr($pattern, 0, 12) . '...';
				my $shift	= ( defined $file->{'shift'} ) ? $file->{'shift'} : FALSE;
				my $expected_result	= ( defined $file->{expected_result} )	? $file->{expected_result} 	: $expected_result;
				
				$type = 'uncompressable' if $compression == 0;
				$type = 'blocksondisk' if $blocksondisk != -1;
				
				if (($shift == TRUE)&&($type !~/\b(write_bytes|holey)\b/)) {
					logresult("script", "shifting data currently only works if type is write_bytes(default) " .
						            "or 'holey'");
				}
				
				my $comment = "\n" .
				              "\t\tmodify file $file_path\n" .
					      "\t\tfilename: $filename\n" .
					      "\t\toffset: $offset\n" .
					      "\t\ttype: $type\n";
				$comment .= "\t\tlen: $len\n" if $type ne 'existing_data';
				$comment .= "\t\tpattern: $display_pattern\n" if $type eq 'write_bytes' || $type eq 'large';
				$comment .= "\t\tcompression: $compression\n" if $compression != -1;
				$comment .= "\t\tblocksondisk: $blocksondisk\n" if $blocksondisk != -1;
				$comment .= "\t\tshift: $shift\n" if $shift == TRUE;
				logcomment($comment);
				
				if ($background == TRUE) {
					$class->{_common}->{files}->{$filename}->{bg_subtest} = subtest( sub{
						$class->_write_data(path => $file_path,
								    offset => $offset,
								    len => $len,
								    pattern => $pattern,
								    type => $type,
								    compression	=> $compression,
								    blocksondisk => $blocksondisk,
								    data => $data,
								    disable_checksums => $disable_checksums,
								    shift => $shift,
								    expected_result => $expected_result,
								    );
						},
						'-bg',
						'-runid','modify_file_'. $filename,);
				}
				else {
					%res = $class->_write_data(path	=> $file_path,
								   offset => $offset,
								   len => $len,
								   pattern => $pattern,
								   type	=> $type,
								   compression => $compression,
								   blocksondisk => $blocksondisk,
								   data	=> $data,
								   disable_checksums => $disable_checksums,
								   shift => $shift,
								   expected_result => $expected_result,
								   );
				}
			} 
		};
		
		if( $@ )
		{
			logresult("fatal", "$@");
		}
	}
	else
	{
		logresult( "fatal", "Unsupported dataset type $ds_type" );
	} 
	
	if ($class->{_debug}) { $class->ls(); }
	
	return %res;
}

## @name ::get_path
##
## @summary return path to file
##
## @description
## return path to file
##
## @return full path to the file
##
## @arg	filename Required: name of file.
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset.
##
## @examples
##  my $filepath = $dataset->get_path( filename=>'f1' );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_path
{
	my $class = shift;
	return $class->{_common}->get_path(@_);		
}

## @name ::delete_file
##
## @summary delete the file
##
## @description
## delete the file
##
## @return hash of status
##
## @arg filename	Required: name of file to delete
##
## @examples
##  $dataset->delete_file( filename=>'f1' );
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
##
sub delete_file
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename => { type=>SCALAR },
		expected => { type => SCALAR, default => C_RC_SUCCESS,},} );
	my %res;
	my $filename = $args{filename};
	$class->_filename_exists( filename=>$filename, expected=>TRUE );
	
	my $file_path = $class->_generate_full_path(filename => $filename);
	logcomment("deleting file: $file_path");
	if (! -e $file_path) {
		return $class->{_ui}->check_for_expected_rc(obtained_rc => C_RC_FILE_DOES_NOT_EXIST,
							    expected_rc => $args{expected},
							    log_message => "Delete file $file_path",
							    output      => $!);
	}
	unlink($file_path);
	if (-e $file_path) {
		return $class->{_ui}->process_output_string(code => C_CMD_DELETE_FILE,
							    output => $!,
							    expected_rc => $args{expected},
							    log_message => "Delete file $file_path",);
	}

	$class->_remove_file(filename => $filename);
	return $class->{_ui}->check_for_expected_rc(obtained_rc => C_RC_SUCCESS,
						    expected_rc => $args{expected},
						    log_message => "Delete file $file_path",
						    output      => '');
}

## @name ::compare_datasets
##
## @summary compare the files on volume vs baseline dataset
##
## @description
## file compare on the volume with baseline dataset
## ( This is a wrapper around DataManip::compare_trees )
##
## @return
## %res;
##	$res{'rc'}		= 'C_RC_SUCCESS';
##	$res{'differ'}	= reference to array of files which are different
##	$res{'in1'}		= reference to array of files only in first dataset
##	$res{'in2'}		= reference to array of files only in 2nd dataset
##
## @arg dataset_2	Required: 2nd dataset to compare
## @arg expected_result Optional: EQUAL or NOT_EQUAL. Default is EQUAL. Will croak if result is not expected.
## @arg ignore_snapshots Optional: don't compare snapshots.  Default is TRUE.
##
## @examples
##	%res = $dataset_1->compare_datasets( 
##		dataset_2 => $dataset_2,
##	);
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
##
sub compare_datasets
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		dataset_2 => { type=>OBJECT },
		expected_result => { type => SCALAR, optional => 1, regex => qr/^(EQUAL|NOT_EQUAL)$/, default => 'EQUAL'},
		ignore_snapshots => { type => SCALAR, optional => 1, default => TRUE},	
	} );

	my $dataset_2 	= ( defined $args{dataset_2} )? $args{dataset_2} : undef;
	my $expected_result = $args{'expected_result'};
	my $ds_type = $class->{_common}->{ds_type};
	my $result;
	my %res;
	
	my $path_1 = $class->get_mountpoint();
	my $path_2 = $dataset_2->{_c_dataset}->get_mountpoint;
	if ($class->{_common}->{testdir} ) {
		$path_1 .= '/'. $class->{_common}->{testdir};
	}
	if ($dataset_2->{testdir} ) {
		$path_2 .= '/'. $dataset_2->{_common}->{testdir};
	}


	my $op = "Comparing\n" .
	         "         " . $path_1 . "\n" .
	         "    with of " . $path_2 . "\n" .
			 "ignore_snapshots: ". $args{'ignore_snapshots'} . "\n" .
		 "expected_result: $expected_result\n";

	my ($in1, $in2, $differ, $subpaths);
	if ($ds_type eq C_DATASET_SPARSE)
	{
		logcomment($op);
		($in1, $in2, $differ, $subpaths) = DataManip::compare_trees("localhost",
									    $path_1,
									    $path_2,);
	}
	else
	{
		logresult( "fatal", "Unsupported dataset type $ds_type" );	
	}
	
	if ($args{'ignore_snapshots'} == TRUE) {
		my @old_in1 = @{$in1};
		my @new_in1;
		foreach my $file (@old_in1) {
			if ($file !~ /^\.snapshot/ ) {
				push(@new_in1, $file);
			}
		}
		$in1 = \@new_in1;

		my @old_in2 = @{$in2};
		my @new_in2;
		foreach my $file (@old_in2) {
			if ($file !~ /^\.snapshot/ ) {
				push(@new_in2, $file);
			}
		}
		$in2 = \@new_in2;
	}

	if (
		( (scalar @{$differ}) == 0 ) &&
		( (scalar @{$in1}) == 0 ) &&
		( (scalar @{$in2}) == 0 ) ) {
		$result = 'EQUAL';
	} else {
		$result = 'NOT_EQUAL';
	}

	if ( defined($expected_result) && ($result ne $expected_result) ) {
		logresult('FATAL', "compare_datasets RESULT: $result\nEXPECTED RESULT: $expected_result");
	}

	$res{'rc'} = 'C_RC_SUCCESS';
	$res{'differ'} = $differ;
	$res{'in1'} = $in1;
	$res{'in2'} = $in2;
	return %res;
}


## @name ::compare_files
##
## @summary compares 2 files
##
## @description
## compares 2 files.  Croaks if not expected result.
##
## @return hash of status and output
##	$res{'result'} = EQUAL or NOT_EQUAL
## 
## @arg file_1	Required: path to first file
## @arg file_2	Required: path to second file
## @expected_result	Optional: EQUAL|NOT_EQUAL ( default is 'EQUAL' )
##
## @example
##
##	%res = $dataset->compare_files( 
##						file_1=> $dataset_1->get_path( filename=>'f1' ), 
##						file_2=> $dataset_2->get_path( filename=>'f1' ),
##						expected_result=>EQUAL|NOT_EQUAL #default:EQUAL
##						);
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub compare_files
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		file_1 => { type=>SCALAR },
		file_2 => { type=>SCALAR },
		expected_result => { type => SCALAR, optional => 1, regex => qr/^(EQUAL|NOT_EQUAL)$/, default => 'EQUAL'},
	});
	
	my $file_1 = $args{file_1};
	my $file_2 = $args{file_2};
	my $expected_result = $args{expected_result};
	my %res;
	my $result;

	if (! -e $file_1) {
		croak "file: $file_1 does not exist $!";
	}
	if (! -e $file_2) {
		croak "file: $file_2 does not exist $!";
	}
	my $compare_result = compare($file_1, $file_2);
	if ($compare_result == 0) {
		$result = 'EQUAL';
		logcomment("The following 2 files are equal:\n".
			$file_1 . "\n" . $file_2 );
	} elsif ($compare_result == 1) {
		$result = 'NOT_EQUAL';
		logcomment("The following 2 files are NOT equal:\n".
			$file_1 . "\n" . $file_2 );
	}
	else {
		logresult('FATAL', "Error when attempting to compare files: $!\n" .
			           "File 1: $file_1\n" .
				   "File 2: $file_2");
	}
	# compare with expected
	if ( defined($expected_result) && ($result ne $expected_result) ) {
		logresult('FATAL', "compare_files failed:\n$file_1\n$file_2\nRESULT: $result\nEXPECTED RESULT: $expected_result");
	}
	

	$res{'rc'} = C_RC_SUCCESS;
	$res{'result'} = $result;
	return %res;
}

## @name ::copy_file
##
## @summary copies a file
##
## @description
## Copies a file
##
## @return hash with status
## 
## @arg source	Required: path to source file
## @arg dest	Required: path to destination file
## @arg	new_dataset	Optional: use this only when copying to a new dataset ( i.e. new volume )
## @arg background	Optional: default is FALSE ( if TRUE, use with wait_finish )
##
## @examples
##
##	%res = $dataset->copy_file( 
##						source		=> $dataset_1->get_path( filename=>'f1' ), 
##						dest		=> $destination_file
##						new_dataset	=> $dataset_2  #optional: only use if copying file to different volume
##  );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub copy_file
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		source => { type=>SCALAR },
		dest => { type=>SCALAR },
		new_dataset => { type=>HASHREF, optional=>1 },
		background => { type=>SCALAR, optional => 1, default=>FALSE },
	} );
	
	my $background	= $args{background};
	my $src_file = $args{source};
	my $dst_file = $args{dest};
	my $new_dataset = $args{new_dataset};
	my $src_filename = basename($src_file);
	my $dst_filename = basename($dst_file);

	my $dst_vol = $class->{volume}; 
	my $dst_dataset = $class;
	if ( $new_dataset ) {
		$dst_vol = $new_dataset->{_c_dataset}->{volume};
		$dst_dataset = $new_dataset->{_c_dataset};
	}
	my $dst_vol_mountpoint = $dst_vol->mountpoint;
	#
	# If dataset is specified, verify that dst path is on correct volume.
	#
	if ( $new_dataset && ( $dst_file !~ /$dst_vol_mountpoint/ ) ) {
		croak "dst_file: $dst_file needs to include $dst_vol_mountpoint in its path $!";
	}
	
	my %res;
	logcomment("copy_file\n$src_file\n$dst_file\n");

	$dst_dataset->_new_file( filename=>$dst_filename );
	#
	# Update path info
	#
	$dst_dataset->{_common}->{files}->{$dst_filename}->{path} = $dst_file;
	if ( $dst_file =~ /$dst_vol_mountpoint/ ) {
		($dst_dataset->{_common}->{files}->{$dst_filename}->{relative_path} ) = $dst_file =~ /$dst_vol_mountpoint\/(.+)$/;
		$dst_dataset->{_common}->{files}->{$dst_filename}->{filer_path} = $dst_vol->mount_path .'/'. $dst_dataset->{_common}->{files}->{$dst_filename}->{relative_path};
	} else {
		# file is being copied off the filer
		$dst_dataset->{_common}->{files}->{$dst_filename}->{relative_path} = undef;
		$dst_dataset->{_common}->{files}->{$dst_filename}->{filer_path}    = undef;
	}

	if (! -e $src_file) {
		croak "file: $src_file does not exist $!";
	}
	if (-e $dst_file) {
		croak "file: $dst_file already exists $!";
	}

	if ($background == TRUE) {
			$dst_dataset->{_common}->{files}->{$dst_filename}->{bg_subtest} = subtest( sub{
					copy($src_file, $dst_file) or croak "Copy failed: $!";
				},
				'-bg',
				'-runid','copy_file_'. $dst_filename,							
			);
	} else {		
			copy($src_file, $dst_file) or croak "Copy failed: $!";;
	}
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::move_file
##
## @summary move a file
##
## @description
## Moves a file
##
## @return hash of status
## 
## @arg source	Required: path to source file
## @arg dest	Required: path to destination file
## @arg	new_dataset	Optional: use this only when copying to a new dataset ( i.e. new volume )
## @arg background	Optional: default is FALSE ( if TRUE, use with wait_finish )
##
## @examples
##
##	%res = $dataset->move_file( 
##						source		=> $dataset_1->get_path( filename=>'f1' ), 
##						dest		=> $destination_file
##						new_dataset	=> $dataset_2  #optional: only use if moving file to different volume
##  );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub move_file
{
	my $class 	= shift;
	my (%args) = validate (@_, {	source		=> { type=>SCALAR },
									dest		=> { type=>SCALAR },
									new_dataset => { type=>HASHREF, optional=>1 },
									background	=> { type=>SCALAR, optional => 1, default=>FALSE },
	} );
	my $background	= $args{background};
	my $src_file = $args{source};
	my $dst_file = $args{dest};
	my $new_dataset = $args{new_dataset};
	my $src_filename = basename($src_file);
	my $dst_filename = basename($dst_file);
	my %res;
	logcomment("move_file\n$src_file\n$dst_file\n");

	my $dst_vol = $class->{volume}; 
	my $dst_dataset = $class;
	if ( $new_dataset ) {
		$dst_vol = $new_dataset->{volume};
		$dst_dataset = $new_dataset;
	}
	my $dst_vol_mountpoint = $dst_vol->mountpoint;
	# If dataset is specified, verify that dst path is on correct volume.
	if ( $new_dataset && ( $dst_file !~ /$dst_vol_mountpoint/ ) ) {
		croak "dst_file: $dst_file needs to include $dst_vol_mountpoint in its path $!";
	}
	
	$class->_remove_file(filename => $src_filename);
	
	$dst_dataset->_new_file( filename=>$dst_filename );
	#
	# Update path info
	#	
	$dst_dataset->{_common}->{files}->{$dst_filename}->{path} = $dst_file;
	if ( $dst_file =~ /$dst_vol_mountpoint/ ) {
		($dst_dataset->{_common}->{files}->{$dst_filename}->{relative_path} ) = $dst_file =~ /$dst_vol_mountpoint\/(.+)$/;
		$dst_dataset->{_common}->{files}->{$dst_filename}->{filer_path} = $dst_vol->mount_path .'/'. $dst_dataset->{_common}->{files}->{$dst_filename}->{relative_path};
	} else {
		# file is being moved off the filer
		$dst_dataset->{_common}->{files}->{$dst_filename}->{relative_path} = undef;
		$dst_dataset->{_common}->{files}->{$dst_filename}->{filer_path}    = undef;
	}

	if (! -e $src_file) {
		croak "file: $src_file does not exist $!";
	}
	if (-e $dst_file) {
		croak "file: $dst_file already exists $!";
	}
	
	if ($background == TRUE) {
			$dst_dataset->{_common}->{files}->{$dst_filename}->{bg_subtest} = subtest( sub{
					move($src_file, $dst_file) or croak "Move failed: $!";
				},
				'-bg',
				'-runid','move_file_'. $dst_filename,							
			);
	} else {		
			move($src_file, $dst_file) or croak "Move failed: $!";;
	}
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::get_file_size
##
## @summary get size of file in bytes
##
## @description
## get size of file in bytes
##
## @return size of file in bytes
##
## @arg filename	Required: name of file
## @arg snapshot	Optional: name of snapshot ( if file is in snapshot )
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset
## 
## @examples
##  my $file_size = $dataset->get_file_size( filename=>'f1');
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_file_size
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename => { type=>SCALAR },
		snapshot => { type=>SCALAR|UNDEF, optional=>1, },
		allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
	} );
	
	my $filename = $args{filename};
	my $snapshot = $args{snapshot};

	my $relative_path = $class->_generate_relative_path(filename => $filename,
							    snapshot => $snapshot,
							    allow_removed => $args{allow_removed});
	my $file_size;
	if (defined($relative_path)) {
		my $file = $class->_file_nfs_attach(relative_path => $relative_path);
		$file_size = $file->size;
	}
	else {
		if (defined($snapshot)) {
			logresult('script', "snapshot is not supported for files that are not in the volume"); 
		}
		my $path = $class->get_path(filename => $filename);
		my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat($path);
		$file_size = $size;
	}
	logcomment("file size of: $filename is " . $file_size);
	
	return $file_size;
}

## @name ::get_timestamp
##
## @summary get timestamp of file
##
## @description
## get timestamp of file
##
## @return timestamp (last modify time) of file
##
## @arg filename	Required: name of file
## @arg snapshot	Optional: name of snapshot ( if file is in snapshot )
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset
## 
## @examples
##  my $timestamp = $dataset->get_timestamp( filename=>'f1');
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_timestamp
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename => { type=>SCALAR },
		snapshot => { type=>SCALAR|UNDEF, optional=>1, },
		allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
	} );
	
	my $filename = $args{filename};
	my $snapshot = $args{snapshot};
	my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks);
	
	my $relative_path = $class->_generate_relative_path(filename => $filename,
							    snapshot => $snapshot,
							    allow_removed => $args{allow_removed});
	if (defined($relative_path)) {
		my $file = $class->_file_nfs_attach(relative_path => $relative_path);
		($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = $file->stat;
		
	}
	else {
		if (defined($snapshot)) {
			logresult('script', "snapshot is not supported for files that are not in the volume"); 
		}
		my $path = $class->get_path(filename => $filename);
		($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat($path);
	}
	logcomment("timestamp of: $filename is ". $mtime);
	
	return $mtime;
}

## @name ::get_uncompressable_data
##
## @summary get uncompressable data
##
## @description create data that is uncompressable.  Only use this when you need to add
##  the exact same data to different files. Warning: out-of-memory error on large files.
##
## @return data
##
## @arg size	Required:	size of data to create
## 
## @examples
##  my $data = $dataset->get_uncompressable_data(size=>4 * KB);
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com, kozel@netapp.com
##
sub get_uncompressable_data
{
	my $class = shift;
	return $class->{_common}->get_uncompressable_data(@_);
}



## @name ::read_file
##
## @summary verify that a file can be read
##
## @description
## This function is to verify that a file can be read.
## Croaks if it can not read file.
##
## @return status
##
## @arg	filename	Required:
## @arg snapshot	Optional: name of snapshot ( if file is in snapshot )
## @arg allow_removed	Optional: flag to enable operation to be permitted on files that were previosly removed by dataset
##
## @examples 
##	$dataset->read_file( filename=>'f1', snapshot=>'ss1', allowed_removed=>TRUE );
##
sub read_file
{
	my $class 	= shift;
    my (%args) = validate (@_, {	filename => { type=>SCALAR },
									snapshot => { type=>SCALAR|UNDEF, optional=>1, },
									allow_removed => { type=>BOOLEAN, optional => 1, default => FALSE },
	} );
	my %res;
	my $filename = $args{filename};
	my $snapshot = $args{snapshot};
	my $file_path = $class->_generate_full_path(filename => $filename,
							    snapshot => $snapshot,
							    allow_removed => $args{allow_removed});


	if (! -e $file_path) {
		croak "file $file_path does not exist $!";
	}
	logcomment("read_file $file_path");
	open ( FH, $file_path ) or croak "can't open $file_path for reading $!";
	binmode(FH);
	my $total_bytes = 0;
	while ( my $bytes = read(FH, my $buffer,512) ) 
	{
		$total_bytes += $bytes;
	}
	close FH;
	logcomment("total bytes: $total_bytes");
	if ($total_bytes == 0) {
		croak "zero bytes were read";
	}
	

	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::verify_checksums
##
## @summary verifies data integrity by comparing checksums
##
## @description
## computes SHA1 checksums for all compression groups and compares them to the precalculated ones
##
## @return hash of status
##
## @arg files		Optional: hash of the filenames
## @arg path		Optional: path to a file
##
## Note, one or the other must be defined
##
## @examples
##  $dataset->verify_checksums( files =>
##								[	{ name=> 'f1' }
##									{ name=> 'f2' }
##								]);
##
##
## @author kozel@netapp.com, dl-compression-qa@netapp.com
##

sub verify_checksums
{
	my $class = shift if ref($_[0]) eq "Compression::Dataset";
	my %args	= validate (@_, {	files	=> { type=>ARRAYREF, optional=>1 },
									path	=> { type=>SCALAR, optional=>1},
									die_on_error	=> { type=>SCALAR, optional=>1, default=>TRUE},
	});

	my $die_on_error = $args{die_on_error};
	my @files = ( defined $args{files} )? @{$args{files}} : {};;
	my $path = $args{path};
	if ( defined $path ) {
		shift @files;
		push ( @files, {filepath=>$path} );
	}

	my $new_sha_sum	= '';
	my $old_sha_sum	= '';
	foreach my $file ( @files )	{
		my $filename;
		my $len;
		my $path;

		if ( defined $file->{filepath} ) {
			$path = $file->{filepath};
			$len = -s $path;
		} else {
			$filename = $file->{name};
			$len	= $class->get_file_size( filename => $filename );
			$path = $class->{_common}->{files}->{$filename}->{path};
		}

		logcomment "Verifying checksums of $path";
		open(FH, "+<$path") || die("Could not open file $path: $!");

		my $cg_data	= '';
		my $counter	= 1;
		my $cg_counter = 0;
		while ( $len > 0 ) {
			if ( $counter == 32 && $len >= 7 * C_ONE_WAFL_BLOCK ) {
				read(FH, $old_sha_sum, 20);
				read(FH, $cg_data, 7 * C_ONE_WAFL_BLOCK - 20);
				$len	-= 7 * C_ONE_WAFL_BLOCK;
				$new_sha_sum = sha1($cg_data);
				$cg_counter++;
				if ( $old_sha_sum ne $new_sha_sum ) {
					if ( $die_on_error ) {
						croak "Checksum failure in compression group $cg_counter. Modified files will fail!";
					} else {
						logcomment "Checksum failure in compression group $cg_counter. Modified files will fail!";
					}
				}
				$counter = 1;
			} elsif ($len >= 8 * C_ONE_WAFL_BLOCK) {
				read(FH, $old_sha_sum, 20);
				read(FH, $cg_data, 8 * C_ONE_WAFL_BLOCK - 20);
					$len	-= 8 * C_ONE_WAFL_BLOCK;
				$new_sha_sum = sha1($cg_data);
				$cg_counter++;
				if ( $old_sha_sum ne $new_sha_sum ) {
					if ( $die_on_error ) {
						croak "Checksum failure in compression group $cg_counter. Modified files will fail!";
					} else {
						logcomment "Checksum failure in compression group $cg_counter. Modified files will fail!";
					}
				}
				$counter++;
			} elsif ( $len > 20 ) {
				read(FH, $old_sha_sum, 20);
				read(FH, $cg_data, $len - 20);
				$len	-= $len;
				$new_sha_sum = sha1($cg_data);
				$cg_counter++;
				if ( $old_sha_sum ne $new_sha_sum ) {
					if ( $die_on_error ) {
						croak "Checksum failure in compression group $cg_counter. Modified files will fail!";
					} else {
						logcomment "Checksum failure in compression group $cg_counter. Modified files will fail!";
					}
				}

			} else {

				my $known_pattern = "01234567890123456789";
				read(FH, $cg_data, $len);
				my $check_data = substr($known_pattern, 0, $len);
				if ( $cg_data ne $check_data ) {
					if ( $die_on_error ) {	
						croak "Checksum failure at end of file. Modified files will fail!";
					} else {
						logcomment "Checksum failure at end of file. Modified files will fail!";
					}
				}
				$len -= $len;
			}
		}
		close(FH);
		logcomment "Checksums verified. Data intact.";
	}
}

## @name ::modify_all_CGs
##
## @summary  modify all CGs in a file
##
## @description modify a certain range within every CG in a file.  As long as the file has compressed CGs, this
## function will find them and overwrite the specified range in each CG.
##
## @return Hash with status.
##
## @arg filename Required: Name of the file.
## @arg offet Required: offset within each CG
## @arg len	Required: number of bytes to write within each CG
## @arg pattern Required: character to write
## @arg CG_ranges Optional: reference to array of ranges as obtained from get_CG_ranges. If undef,
## this function will call get_CG_ranges to get them.
## @arg estimate_boundries Optional: estimate CG ranges based on C_CG_SIZE ( default FALSE )
## Only set this if you're overwriting parts that are uncompressed (i.e. not real CGs ).
##
## @examples
## $obj->modify_all_CGs(filename=>$FILENAME, offset=>4*KB, len=>2*KB, pattern=>'c' )
##  
sub modify_all_CGs
{
	my $class 	= shift;
    my (%args) = validate (@_, {	filename	=> { type=>SCALAR },
									pattern		=> { type=>SCALAR },
									offset		=> { type=>SCALAR },
									len			=> { type=>SCALAR },
									CG_ranges	=> { type=>ARRAYREF, optional=>1, default=>undef },
									estimate_boundries	=> { type=>SCALAR, optional=>1, default=>FALSE },
	} );
	my %res;
	my $filename	= $args{filename};
	my $pattern		= $args{pattern};
	my $offset_each_CG		= $args{offset};
	my $len_each_CG			= $args{len};
	my $estimate_boundries			= $args{estimate_boundries};
	my $file_size = $class->get_file_size( filename => $filename );
	my $file_path = $class->get_path( filename=>$filename );
	my @CG_ranges = undef;
	my $num_CG_ranges = 0;
	if (defined $args{CG_ranges} ) {
		@CG_ranges = @{$args{CG_ranges}};
		$num_CG_ranges = scalar @CG_ranges;
	}
	
	

	my $num_CGs = int($file_size/ C_CG_SIZE);
	my $summary_text = "modify_all_CGs:\n".
				"filename: $filename\n".
				"offet: $offset_each_CG\n".
				"len: $len_each_CG\n".
				"pattern: $pattern\n";
	if ($num_CG_ranges>0) {
		$summary_text .= "CG_ranges: $num_CG_ranges\n";
	} else {
		$summary_text .= "CG_ranges: undef\n";
	}
	logcomment($summary_text);
	if (($offset_each_CG+$len_each_CG)> C_CG_SIZE ) {
		logresult("script", "the sum of offset and len need to be less than a CG:". C_CG_SIZE);
	}
	
	if ( ! $estimate_boundries ) {
		if (! $num_CG_ranges) {
			%res = $class->{_common}->get_CG_ranges( filename=>$filename );
			@CG_ranges = @{$res{CG_ranges}};
		}
		foreach my $CG_range ( @CG_ranges ) {
			if ($CG_range !~ /\d+\-\d+/ ) {
				logresult("fatal", "invalid CG_range: $CG_range");
			} else {
				my ( $start_fbn, $end_fbn) = $CG_range =~ /(\d+)\-(\d+)$/;
				my $CG_start = $start_fbn * C_WAFL_BLOCK_SIZE;
				my $offset = $CG_start + $offset_each_CG;
				my $len = $len_each_CG;
				my $CG_size = (($end_fbn - $start_fbn )+ 1) * C_WAFL_BLOCK_SIZE;
				if ($len > ($CG_size-$offset_each_CG) ) {
					$len = $CG_size-$offset_each_CG;
				}
				%res = $class->_write_bytes(	path		=> $file_path,
												offset		=> $offset,
												len			=> $len,
												pattern		=> $pattern,
				);
			}
		}

	} else { # estimate_boundries based on C_CG_SIZE
		logcomment("There are $num_CGs full CGs in $filename.") if $class->{_debug};
		if (($offset_each_CG + $len_each_CG) > C_CG_SIZE ) {
			logresult("script", "The sum of offset + len needs to be less than the size of a CG.");
		}

		my $leftover_bytes = $file_size - ( $num_CGs * C_CG_SIZE );

		my $offset;
		my $CG_start;
		for (my $i=0; $i<$num_CGs; $i++) {
			$CG_start = $i * C_CG_SIZE;
			$offset = $CG_start + $offset_each_CG;
			%res = $class->_write_bytes(	path		=> $file_path,
											offset		=> $offset,
											len			=> $len_each_CG,
											pattern		=> $pattern,
			);
		}

		$CG_start = $num_CGs * C_CG_SIZE;
		$offset = $CG_start + $offset_each_CG;	
		my $byte_places_left_in_file = $file_size - $offset;
		my $leftover_length;
		if ($byte_places_left_in_file > 0) {
			$leftover_length = ($byte_places_left_in_file>=$len_each_CG)? $len_each_CG : $byte_places_left_in_file;
			%res = $class->_write_bytes(	path		=> $file_path,
											offset		=> $offset,
											len			=> $leftover_length,
											pattern		=> $pattern,
			);
		}

	}
	$res{'CG_ranges'} = \@CG_ranges;
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::single_file_snap_restore
##
## @summary 
##	Single file snap restore (SFSR) reverts a file in the active file system to the state contained in a snapshot
##
## @description
##	Single file snap restore (SFSR) reverts a file in the active file system to the state contained in a snapshot 
##
## @return hash of status
##
## @arg filename	Required: name of file
## @arg snapshot	Required: name of snapshot to restore from
## @arg wait_for_completion	Optional: don't return until file is completely restored ( default: TRUE )
## @arg expected Optional: Expected return code ( default: C_RC_SUCCESS )
##
## @examples
##	%res = $src_vol->{dataset}->single_file_snap_restore(filename=>$FILENAME, snapshot=>$SNAPSHOT);
sub single_file_snap_restore
{
	my $class 	= shift;
	my (%args) = validate (@_, { filename => { type=>SCALAR },
				     snapshot => { type=>SCALAR },
				     restore_as_path => { type=>SCALAR|UNDEF, optional=>1 },
				     wait_for_completion => { type=>SCALAR, optional=>1, default=>TRUE },
				     expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
	});
	my %res;
	my $filename = $args{filename};
	my $snapshot = $args{snapshot};
	my $wait_for_completion = $args{wait_for_completion};
	my $restore_as_path = $args{restore_as_path};
	my $rc = C_RC_SUCCESS;
	
	my $comment_str = "single_file_snap_restore\n" .
	                  "Filename: $filename\n" .
			  "Snapshot: $snapshot\n";
			  
	if (defined($restore_as_path)) {
		$comment_str .= "Restoring to path: $restore_as_path\n";
	}
	
	logcomment($comment_str);
	
	my $file_already_exists = $class->_filename_exists( filename=>$filename, expected=>C_RC_FAILURE_NON_FATAL );
	my $file_path = $class->_generate_file_path(filename => $filename,
						    allow_removed => TRUE);
	
	if ($class->{_boot_mode} eq C_ONTAP_CMODE) {
		# No cdefs for file restore yet. Send comand directly
		# If still no cdefs in the future, we should probably send the command directly for 7mode too
		my $cmd_code = (defined($restore_as_path)) ? C_CMD_SFSR_TO_PATH_START : C_CMD_SFSR_START;
		my $vol_name = $class->{volume}->name();
		my $vol_mount_path = $class->_get_vol_mount_path(); 
	
		my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
							       volume => $vol_name,
							       vserver => $class->{_vserver},
							       snapshot => $snapshot,
							       file_path => $filename,
							       restore_path => $restore_as_path);
		
		my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
								    boot_mode => $class->{_boot_mode},
								    cmd => $command,
								    timeout => C_DFLT_SFSR_TIMEOUT,
								    node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
		
		my %sfsr_res = $class->{_ui}->process_output_string(code => $cmd_code,
								    output => $output,
								    expected_rc => $args{expected},
								    log_message => "SFSR of file $file_path from snapshot $snapshot",);
		
		$rc = $sfsr_res{rc};
		if ($rc != C_RC_SUCCESS) {
			return %sfsr_res;
		}
	}
	else {
		my $sfsr_output = snaprestore_file($class->{volume}->Filer->conn, $file_path, $snapshot, restore_as_path => $restore_as_path);
		# snaprestore_file should return 1 in case of success
		# in case of error, it dies (inside an eval block)
		my $res = $@;
		if ((!defined($sfsr_output)) || ($sfsr_output != 1)) {
			return $class->{_ui}->process_output_string(code => C_CMD_SFSR_START,
								    output => $res,
								    expected_rc => $args{expected},
								    log_message => "Start SFSR for file $filename",);
		}
	}
	
	my $restored_filename = $filename;
	if ((defined($restore_as_path)) && ($restore_as_path ne $file_path)) {
		$restored_filename = basename($restore_as_path);
		# We are restoring to a different path, add the new file to $class->{files}
		$class->_new_file(filename => $restored_filename);
	}
	elsif ($file_already_exists == FALSE) {
		# If this was a deleted/moved file, re-add it to $class->{files}
		$class->_re_add_file(filename=>$filename);
	}
	
	if ($wait_for_completion == TRUE) {
		$class->wait_for_single_file_snap_restore(filename => $restored_filename,
							  restored_from_file => $filename,
							  snapshot => $snapshot);
	}
	
	sleep 3;
	
	$res{'rc'} = $rc;
	return %res;
}

## @name ::wait_for_single_file_snap_restore
##
## @summary 
##	Waits until a file is completely restored from a snapshot.
##
## @description
##	Waits until a file is completely restored from a snapshot.
##
## @return hash of status
##
## @arg filename Required: Name of file.
## @arg snapshot Required: Name of the snapshot from which the file is restored.
## @arg timeout Optional: Timeout in seconds to wait for the file to be restored. Default: 300.
##
## @examples
## %res = $src_vol->{dataset}->wait_for_single_file_snap_restore(filename=>$FILENAME, snapshot=>$SNAPSHOT);
##
sub wait_for_single_file_snap_restore
{
	my $class 	= shift;
	my (%args) = validate( @_,
		{
		 filename => { type => SCALAR },
		 snapshot => { type => SCALAR },
		 restored_from_file => { type => SCALAR|UNDEF, optional => 1 },
		 timeout => { type => SCALAR, regex => qr/^\d+$/, optional => 1, default => C_DFLT_SFSR_TIMEOUT },
                } );
	
	my $filename = $args{filename};
	my $restored_from_filename = (defined($args{restored_from_file})) ? $args{restored_from_file} : $filename;
	my $snapshot = $args{snapshot};
	my $timeout = $args{timeout};
	my $file_path;
	
	logcomment("Waiting for single file snap restore to complete:\n".
		   "Filename: $filename\n" .
		   "Restored from file: $restored_from_filename\n" .
		   "Snapshot: $snapshot\n" .
		   "Timeout: $timeout");
	
	$file_path = $class->_generate_file_path(filename => $filename);
	
	my $ss_file_size = $class->get_file_size(filename=>$restored_from_filename,
						 snapshot=>$snapshot,
						 allow_removed => TRUE);
	my $ss_timestamp = $class->get_timestamp(filename=>$restored_from_filename,
						 snapshot=>$snapshot,
						 allow_removed => TRUE);
	my $file_size = 0;
	my $timestamp = -1;
	my $sleep_interval = 5;
	
	while (($file_size != $ss_file_size) || ($timestamp != $ss_timestamp)) {
		if (-e $class->{_common}->{files}->{$filename}->{path} ) {
			$file_size = $class->get_file_size(filename=>$filename);
			$timestamp = $class->get_timestamp(filename=>$filename);
		}
		sleep($sleep_interval);
		$timeout -= $sleep_interval;
		
		if ($timeout < 0) {
			logresult('fatal', "Timed-out while waiting for single file snap restore to complete\n" .
				           "Filename: $filename\n" .
					   "Snapshot: $snapshot\n" .
					   "File size: $file_size\n" .
					   "Snapshot file size: $ss_file_size\n" .
					   "File timestamp: $timestamp\n" .
					   "Snapshot file timestamp: $ss_timestamp\n");
		}
	}
	
	return (rc => C_RC_SUCCESS);
}

## @name ::fill_remaining_space
##
## @summary 
##	
##
## @description
##	
##
## @return hash of status
##
## @arg filename	Required: name of file
## @arg leave_space	Optional: amount of available space to leave ( currently to the nearest 4KB, default is zero )
##
## @examples
##	%res = $src_vol->{dataset}->fill_remaining_space(filename=>$FILENAME);
sub fill_remaining_space
{
	my $class 	= shift;
	my (%args) = validate (@_,  {	filename => { type=>SCALAR },
									leave_space	=> { type=>SCALAR, optional=>1, default=>0},
	} );
	my $vol = $class->{volume}; 
	my $filename = $args{filename};
	my $leave_space = $args{leave_space};
	my $block_size = C_WAFL_BLOCK_SIZE * 10;
	my $mountpoint = $vol->mountpoint;
	$class->_new_file(	filename=>$filename,);
	my $path = $class->{_common}->{files}->{$filename}->{path};
	my $avail = $vol->avail * KB;

	logcomment("fill_remaining_space\n$path\navail: $avail\nleave_space:$leave_space");
	eval {
		my $offset = 0;
		my $output = '';
		my $num_tries = 0;
		while () {
			$output = `dd if=/dev/urandom of=$path count=1 bs=$block_size seek=$offset 2>&1`;
			if (
					($output =~ /No space left on device/i ) ||
					($avail <= $leave_space) )
				
			{
				# If we get the no space error, it might be still compressing so we need
				# to sleep and retry.
				$num_tries++;
				sleep 10;
				last if $num_tries > 3;
				if ($num_tries == 1) {
					$avail = $vol->avail * KB;
				}
			} else {
				$num_tries = 0;
				$avail = $avail - $block_size;
				$offset++;
			}
		}
		if ($avail) {
			my $remaining = $avail - $leave_space;
			return if $remaining <= 0;
			my $seek = $offset * $block_size;
			$output = `dd if=/dev/urandom of=$path count=$remaining bs=1 seek=$seek 2>&1`;
			logcomment("out: $output");
		}
	};
}

## @name ::get_data_range
##
## @summary get specified range of content from a file
##	
##
## @description get specified range of content from a file
##	
## @return data content
##
## @arg filename	Required: name of file
## @arg leave_space	Optional: amount of available space to leave ( currently to the nearest 4KB, default is zero )
##
## @examples
##			my $data = $class->get_data_range(	path	=> $path,
##												offset	=> $offset,
##												len		=> $len,
##			);
sub get_data_range
{
	my $class 	= shift;
	my (%args) = validate (@_,  {	path		=> { type=>SCALAR },
									offset		=> { type=>SCALAR },
									len			=> { type=>SCALAR, optional=>1, default=>undef },
	} );
	my $path		= $args{path};
	my $offset		= $args{offset};
	my $len			= $args{len};
	my $data		= '';

	if (! -e $path) {
		croak "file $path does not exist $!";
	}
	open ( FH, $path ) or die "can't open $path for reading $!";
	binmode(FH);
	seek(FH,$offset,0);
	my $total_bytes = 0;
	my $buffer_size = 512;
	my $current_buffer_size = $buffer_size;
	if ((defined $len)&&($len < $buffer_size)) {
		$current_buffer_size = $len;
	}
	while ( my $bytes = read(FH, my $buffer, $current_buffer_size) ) 
	{
		$data .= $buffer;
		$total_bytes += $bytes;
		$len -= $buffer_size if defined $len;
		if ((defined $len)&&($len>0)&&($len <= $buffer_size)) {
			$current_buffer_size = $len;
		}
		if ((defined $len)&&($len <= 0)) {
			last;
		}
	}
	close FH;
	return $data;
}

## @name ::display_file_contents
##
## @summary display the contents of a file
##
## @description
## display contents of file.  This is more just a utility for 
## debugging/visual inspection.
##
## @return none
##
## @arg	Required:	filename
##
## @examples 
##	$dataset->display_file_contents( filename=>'f1' );
##
sub display_file_contents
{
	my $class   = shift;
	my (%args) = validate (@_, { filename => { type=>SCALAR },
								offset	 => { type=>SCALAR, optional=>1 },
								len		 => { type=>SCALAR, optional=>1 },								
	} );
	my $filename = $args{filename};
	$class->_filename_exists( filename=>$filename, expected=>TRUE );
	my $file_size = $class->get_file_size(filename=>$filename);
	my $offset = $args{offset} || 0;
	my $len = $args{len} || $file_size;
	my $file_path = $class->{_common}->{files}->{$filename}->{path};
	if (! -e $file_path) {
		croak "file $file_path does not exist $!";
	}
	logcomment("display_file_contents: $filename\n".
				"$file_path\n".
				"offset: $offset\n".
				"len: $len\n");
	my $file_contents = '';
	### FIXME out-of-memory error size > 1GB
	open ( FH, $file_path ) or die "can't open $file_path for reading $!";
	binmode(FH);
	seek(FH,$offset,0);
	my $total_bytes = 0;
	my $buffer_size = 512;
	my $num_bytes_remaining = $len;
	$buffer_size = $len if $buffer_size > $len;
	while ( my $bytes = read(FH, my $buffer, $buffer_size) ) 
	{
		$total_bytes += $bytes;
		$file_contents .= $buffer;
		if ($len > 0 ) {
			$num_bytes_remaining -= $bytes;
			if ($num_bytes_remaining < $buffer_size) {
				$buffer_size = $num_bytes_remaining;
			}
			if ($total_bytes == $len) {
				last;
			}
		}
	}
	close FH;
	logcomment("total_bytes read: $total_bytes " .
		   "--------------------------------------------------------------\n" .
		   "$file_contents\n" .
		   "--------------------------------------------------------------");
}


## @name ::get_mountpoint
##
## @summary return mount point of volume
##
## @description
## wrapper around $vol->mountpoint
##
## @return $vol->mountpoint
##
## @examples
##  my $mountpoint = $dataset->get_mountpoint;
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_mountpoint
{
	my $class 	= shift;
	
	if (defined($class->{_vfiler_mountpoint})) {
		return $class->{_vfiler_mountpoint};
	}
	return $class->{volume}->mountpoint;
}

## @name ::get_filer_path
##
## @summary return path to file on filer
##
## @description
## return path to file on filer
##
## @return the filer's path to the file ( e.g. /vol/vol1/f1 )
##
## @arg	filename Required: name of file
##
## @examples
##  my $filer_path = $dataset->get_filer_path( filename=>'f1' );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_filer_path
{
	my $class 	= shift;
	my (%args) = validate (@_, { filename => { type=>SCALAR },} );
	my $filename = $args{filename};
	$class->_filename_exists( filename=>$filename, expected=>TRUE );
	my %res;
	return $class->{_common}->{files}->{$filename}->{filer_path};
}

## @name ::replace_file
##
## @summary replace file with some specified file
##
## @description
## replace file with some specified file
##
## @return hash with success code
##
## @arg filename	Required:	name of file to replace
## @arg new_file	Required:	path to new file
##
## @example
##  $dataset->replace_file( filename=>'f1', new_file=>'path_to_new_file' );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub replace_file
{
	my $class 	= shift;
	my (%args) = validate (@_, { filename => { type=>SCALAR },
								 new_file => { type=>SCALAR },} );
	my %res;
	my $filename = $args{filename};
	my $new_file = $args{new_file};
	$class->_filename_exists( filename=>$filename, expected=>TRUE );
	my $file_path = $class->{_common}->{files}->{$filename}->{path};

	logcomment("Replacing file: $file_path\n with $new_file");
	if (! -e $new_file) {
		croak "new file: $new_file does not exist $!";
	}
	if (! -e $file_path) {
		croak "file: $file_path does not exist $!";
	}	
	copy($new_file, $file_path) or croak "Copy failed: $!";

	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}
## @name ::ls
##
## @summary ls
##
## @description
## ls
##
## @example
##  $dataset->ls();
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub ls
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		path => {type=>SCALAR, optional => 1,},
	});
	
	my $path = (defined($args{path})) ? $args{path} : $class->get_mountpoint();
	
	logcomment("Listing contents of $path");
	
	my $df_output = `df $path`;
	my $ls_output = `ls -sail $path`;
	
	logcomment( "df $path\n".
		   "$df_output\n".
		   "ls -sail $path\n".
		   "$ls_output");

}

## @name ::wait_finish
##
## @summary wait for background process to finish
##
## @description
## wait for background process to finish from create,
##  modify, copy_file, or move_file
##
## @return hash of status
## 
## @examples
##  $dataset->wait_finish( filename=>'f1' );
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
##
sub wait_finish
{
	my $class 	= shift;
	my (%args) = validate (@_, { filename => { type=>SCALAR },} );
	my %res;
	my $filename = $args{filename};

	logcomment("waiting for $filename background process to complete");
	$class->{_common}->{files}->{$filename}->{'bg_subtest'}->wait_finish();
	logcomment("$filename bg_subtest has finished");

	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}


## @name ::is_alive
##
## @summary Checks if a background process is alive.
##
## @description
## Checks if a background process is alive,
## modify, copy_file, or move_file
##
## @return Boolean value: 0/1
## 
## @examples
##  if ($dataset->is_alive(filename=>'f1')) {logcomment("The background process is alive");}
##
## @author dl-compression-qa@netapp.com
##
sub is_alive
{
        my $class = shift;
	my (%args) = validate (@_, {filename => { type=>SCALAR},} );
	
	my $filename = $args{filename};
	
	return $class->{_common}->{files}->{$filename}->{'bg_subtest'}->is_alive();
}


## @name ::mk_dir
##
## @summary Make a directory in volume
##
## @description
## Make a directory in volume
##
## @return hash of status
## 
## @arg	name	Required: name of directory
## @arg path	Optional: complete path to the directory the folder will be created in
##
## @examples
##
##	%res = $dataset->mkdir(
##							name => 'd1',
##						 );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub mk_dir
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		name => { type=>SCALAR },
		path => { type=>SCALAR, optional=>1 },
	} );
	
	my $dir_name = $args{name};
	my $vol_mountpoint = $class->get_mountpoint();
	my $dir_path = (defined $args{path}) ? ($args{path} .'/'. $dir_name) : ($vol_mountpoint .'/'. $dir_name);
	my %res;
	
	if (! -e $dir_path ) {
		logcomment("mk_dir path: $dir_path");
		do_shell("sudo mkdir -m 777 $dir_path");
	}
	
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
}

## @name ::truncate_file
##
## @summary truncate file to specified size
##
## @description
## truncate file to specified size
##
## @return hash of status
##
## @arg	filename	Required: name of file to truncate
## @arg size		Required: size to truncate file to
##
## @examples
##  $dataset->truncate_file( filename=>'f1', size=>1024 );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub truncate_file
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		filename => { type=>SCALAR },
		size => { type=>SCALAR, regex => qr/^\d+$/, },
	} );
	
	my %res;
	my $filename = $args{filename};
	my $size = $args{size};
	if ($size < 0)
	{
		logresult('script', "truncate size must not be negative: $size")
	}

	my $try_using_absolute_path = 0;
	my $relative_path = $class->_generate_relative_path(filename => $filename);
	if (defined($relative_path)) {
		logcomment("Truncating file: $filename\n" .
			   "Relative path: $relative_path\n" .
			   "Size: $size");
		eval {
			my $file = $class->_file_nfs_attach(relative_path => $relative_path);
			$file->truncate(size=>$size);
		};
		if ($@) {
			logcomment("Failed to truncate $filename using the relative path: " . $@);
			$try_using_absolute_path = 1;
		}
	}
	else {
		$try_using_absolute_path = 1;
	}
	if ($try_using_absolute_path == 1) {
		my $path = $class->get_path(filename => $filename);
		logcomment("Truncating file: $filename\n" .
			   "Absolute path: $path\n" .
			   "Size: $size");
		truncate($path, $size);
	}

	$res{'rc'} = C_RC_SUCCESS;
	if ($class->{_debug}) { $class->ls(); }
	return %res;
}

## @name ::get_vbn_constants
##
## @summary get the vbn/vvbn constants
##
## @description
## get the vbn/vvbn constants
##
## @return hash of status and output
##	
## 
## @example
##	%res = $dataset->get_vbn_constants();
##  if (grep(/^$x$/, @{$res{vbn_headers}} ) ) {
##		logcomment("$x is a valid vbn header");
##	}
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
##
sub get_vbn_constants {
	my $class = shift;
	return $class->{_common}->get_vbn_constants(@_);
}

## @name ::get_file_atime
##
## @summary Gets the atime of the file (on the client).
##
## @description
## Gets the atime of the file (on the client). 
##
## @return Hash with rc, atime (# of seconds since the start of the epoch), gmt_atime
##
## @arg filename Required: Name of file.
##
## @example
##  $dataset->get_file_atime(filename => 'my_file');
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_file_atime
{
        my $class = shift;
	my (%args) = validate (@_, {
		filename => { type=>SCALAR },
	} );
	
	my $filename = $args{filename};
	my ($atime, $gmt_atime);
	
	my $relative_path = $class->_generate_relative_path(filename => $filename);
	if (defined($relative_path)) {
		my $file = $class->_file_nfs_attach(relative_path => $relative_path);
		$atime = $file->atime();
	}
	else {
		my $path = $class->get_path(filename => $filename);
		my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$a_time,$mtime,$ctime,$blksize,$blocks) = stat($path);
		$atime = $a_time;
	}
	$gmt_atime = gmtime($atime);
	
	logcomment("File: $filename\n" .
		   "atime: $atime (GMT: $gmt_atime)");
	
	return (rc => C_RC_SUCCESS,
		atime => $atime,
		gmt_atime => $gmt_atime,);
}

## @name ::set_file_atime
##
## @summary Sets the atime of the file (on the client).
##
## @description
## Sets the atime of the file (on the client).
##
## @arg filename Required: Name of file.
##
## @arg new_atime Required: New atime to set (# of seconds since the start of the epoch).
##
## @return The return of the get_file_atime method, after the new atime has been set.
##
## @example
## my %res = $compress->{dataset}->get_file_atime(filename => 'my_file');
## my $new_atime = $res{atime} + 30;
## $dataset->set_file_atime(filename => 'my_file', $new_atime);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub set_file_atime
{
        my $class = shift;
        my (%args) = validate (@_, {
		filename => { type=>SCALAR },
		new_atime => { type=>SCALAR, },
	} );
	
	my $filename = $args{filename};
	my $new_atime = $args{new_atime};
	
	my $relative_path = $class->_generate_relative_path(filename => $filename);
	my $file = $class->_file_nfs_attach(relative_path => $relative_path);
	
	$file->touch(atime => $new_atime);
	
	# Return the newly-set atime
	return $class->get_file_atime(filename => $filename);
}


################################################################
# Internal subroutines
################################################################

# Internal function
# name ::_new_file
#
# summary (INTERNAL) create a new file object
#
# description
# Creates a new file object.
# Doesn't create actual file.
#
# example
#  $dataset->_new_file( filename=>'f2' );
# 
# File objects contain the folling:
# $file->{path}			= full path on client
# $file->{filer_path}		= path for filer commands
# $file->{relative_path}	= path relative to volume
# 
# note: both filer_path and relative_path are only for
# commands run on filer
sub _new_file
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		filename => { type=>SCALAR },
		testdir => { type=>SCALAR|UNDEF, optional=>1 },
		path => { type=>SCALAR|UNDEF, optional=>1 },
		clone_lun => { type=>SCALAR|UNDEF, optional=>1 },
	} );
	my $file;
	my $filename = $args{filename};
	my $vol_mountpoint = $class->get_mountpoint();
	my $testdir_path  = ( defined $args{path} )? $args{path} : $vol_mountpoint;
	$class->{_common}->{testdir} = $args{testdir};

	if (defined $args{testdir}) {
		$testdir_path .= '/' . $class->{_common}->{testdir};
	}
	$class->{_common}->{testdir_path} = $testdir_path;

	
	$file->{name} = $filename;
	$file->{path} = $testdir_path .'/'. $filename;
	
	$file->{relative_path} = $file->{path};
	$file->{relative_path} =~ s#^$vol_mountpoint/##;
	$file->{relative_path} =~ s#\\#/#;
	$file->{filer_path} = $class->_get_vol_mount_path() .'/'. $file->{relative_path};

	if ( $file->{path} !~ /$vol_mountpoint/ ) {
		# file is not on a volume
		$file->{relative_path} = undef;
		$file->{filer_path}    = undef;
	}

	my $comment = "creating new file object for $filename\n";
	$comment .= "path: ". $file->{path} ."\n";
	$comment .= "relative_path: ". $file->{relative_path} ."\n" if (defined($file->{relative_path}));
	$comment .= "filer_path: ". $file->{filer_path} ."\n" if (defined($file->{filer_path}));
	logcomment($comment);

	$class->{_common}->_new_file(file=>$file,
				     clone_lun => $args{clone_lun});
}

sub _get_vol_mount_path
{
	my $class 	= shift;
	return $class->{volume}->mount_path();
}

# Internal function
# name ::_filename_exists
#
# summary (INTERNAL) Checks that file exists in the dataset object
#
# description
# Used in error checking by various functions
#
# example
#		$class->_filename_exists( filename=>$filename, expected=>FALSE ); 
#
sub _filename_exists
{
	my $class = shift;
	return $class->{_common}->_filename_exists(@_);
}

# Internal function
# name ::_remove_file
#
# summary (INTERNAL) Moves the file entry from the $class->{files} hash
#  into the $class->{_removed_files} one.
#
# example
#		$class->_remove_file( filename=>$filename ); 
#
sub _remove_file
{
	my $class = shift;
	return $class->{_common}->_remove_file(@_);
}

# Internal function
# name ::_generate_file_path
#
# summary (INTERNAL)
# If snapshot is defined, generates the complete path of a file within a snapshot.
# If snapshot is not defined, returns the file path.
#
# description
# Generates the complete path of a file (within a snapshot, if required).
#
# example
#  my $snapshot_file_path = $dataset->_generate_file_path(filename=>'f2',
#                                                         snapshot => 'my_snapshot');
# 
#
#
sub _generate_file_path
{
	my $class = shift;
	return $class->{_common}->_generate_file_path(@_);
}



# Internal function
# name ::_generate_relative_path
#
# summary (INTERNAL)
# If snapshot is defined, generates the relative path of a file within a snapshot.
# If snapshot is not defined, returns the relative file path.
#
# description
# Generates the relative path of a file (within a snapshot, if required).
#
# example
#  my $relative_path = $dataset->_generate_relative_path(filename=>'f2',
#                                                        snapshot => 'my_snapshot');
# 
#
#
sub _generate_relative_path
{
	my $class = shift;
	return $class->{_common}->_generate_relative_path(@_);
}

# Internal function
# name ::_generate_full_path
#
# summary (INTERNAL)
# This is same as _generate_file_path except returns full path as seen when mounted on a client.
# If snapshot is defined, generates the complete path of a file within a snapshot.
# If snapshot is not defined, returns the file path.
#
# description
# Generates the complete path of a file (within a snapshot, if required).
#
# example
#  my $snapshot_full_path = $dataset->_generate_full_path(filename=>'f2',
#                                                         snapshot => 'my_snapshot');
# 
#
#
sub _generate_full_path
{
	my $class 	= shift;
	my (%args) = validate (@_,
			       { 
					filename => { type=>SCALAR },
					snapshot => { type=>SCALAR|UNDEF, optional=>1 },
					allow_removed => { type=>BOOLEAN|UNDEF, optional=>1, },
				} );
	
	my $filename = $args{filename};
	my $snapshot = $args{snapshot};
	my $allow_removed = $args{allow_removed};
	my $file_path;
	
	if (!defined($allow_removed)) {
		$allow_removed = (defined($snapshot)) ? TRUE : FALSE;
	}
	
	my $file_already_exists = $class->_filename_exists( filename=>$filename, expected=>C_RC_FAILURE_NON_FATAL );
	if ($file_already_exists == FALSE) {
		if ($allow_removed == FALSE) {
			logresult('fatal', "File $filename does not exist");
		}
		$class->_file_was_removed(filename=>$filename);
		$file_path = $class->{_common}->{_removed_files}->{$filename}->{path};
	}
	else {
		$file_path = $class->{_common}->{files}->{$filename}->{path};
	}
	
	if (defined($snapshot)) {
		my $vol_mountpoint = $class->get_mountpoint;
		$file_path =~ s#^$vol_mountpoint#$vol_mountpoint/.snapshot/$snapshot#;
	}
	return $file_path;
}

# Internal function
# name ::_file_was_removed
#
# summary (INTERNAL) Checks that file was ever known and has been removed by the dataset object
#
# example
#		$class->_file_was_removed( filename=>$filename ); 
#
sub _file_was_removed
{
	my $class = shift;
	return $class->{_common}->_file_was_removed(@_);
}

# Internal function
# name ::_re_add_file
#
# summary (INTERNAL) Re-adds the file hash into $class->{files}.
#
# example
#		$class->_re_add_file( filename=>$filename ); 
#
sub _re_add_file
{
	my $class = shift;
	return $class->{_common}->_re_add_file(@_);
}

# Internal function
# name ::_write_data
#
# summary (INTERNAL) creates a file on a volume
#
# description
# This is called by the create function.
#
# example
#		$class->_write_data(	path	=> $file->{path},
#								offset	=> $offset,
#								len		=> $len,
#								pattern	=> $pattern,
#								type	=> 'write_bytes|existing_data|uncompressable|compressable',
#								compression => $compression,
#								blocksondisk	=> $blocksondisk,
#								data	=> $data,
#								expected_result => C_RC_SUCCESS(default)|C_RC_INSUFFICIENT_SPACE|C_RC_FAILURE_NON_FATAL,
#							);
# 
sub _write_data
{
	my $class 	= shift;
	my (%args) = validate (@_,  {	path	=> { type=>SCALAR },
									offset	=> { type=>SCALAR },
									len		=> { type=>SCALAR|UNDEF },
									pattern	=> { type=>SCALAR|UNDEF },
									type	=> { type=>SCALAR },
									compression => { type=>SCALAR },
									blocksondisk	=> { type=>SCALAR },
									data	=> { type=>SCALAR, default=>'' },
									shift	=> { type=>SCALAR, default=> FALSE },
									disable_checksums => { type=>SCALAR, optional=>1, default=>TRUE },
									expected_result => { type=>SCALAR, optional=>1, default=>C_RC_SUCCESS },
	} );
	my $path	= $args{path};
	my $offset	= $args{offset};
	my $len		= $args{len};
	my $pattern = $args{pattern};
	my $type	= $args{type};
	my $data	= $args{data};
	my $compression	= $args{compression};
	my $blocksondisk	= $args{blocksondisk};
	my $expected_result = $args{expected_result};
	my $disable_checksums	= $args{disable_checksums};
	#TODO: disable_checksums currently only affects _write_bytes; still need to update other functions.
	my $shift = $args{'shift'};
	my $remaining_data;
	my %res;

	if (( $expected_result == C_RC_INSUFFICIENT_SPACE ) && ($type ne 'uncompressable')) {
		logresult("script", "expected_result=>C_RC_INSUFFICIENT_SPACE can only be caught currently when type=>uncompressable\n".
					"due to burt #372834");
	}

	my $logcomment = "_write_data\n";
	$logcomment .= "path: $path\n" if $path;
	$logcomment .= "offset: $offset\n" if $offset;
	$logcomment .= "len: $len\n" if $len;
	$logcomment .= "type: $type\n" if $type;
	$logcomment .= "shift: $shift\n" if $shift;
	$logcomment .= "expected_result: $expected_result\n" if $expected_result;
	logcomment($logcomment);
	if (( $type eq 'write_bytes' )&&($shift == TRUE)) {
		$remaining_data = $class->get_data_range(path => $path,
							 offset => $offset,
							 );
	}
	if ( $type eq 'holey' ) {
		$remaining_data = $class->get_data_range(path => $path,
							 offset => $offset + $len,
							 );		
	}

	if ($type eq 'write_bytes')
	{
		%res = $class->_write_bytes(path => $path,
					    offset => $offset,
					    len => $len,
					    pattern => $pattern,
					    disable_checksums => $disable_checksums,
					    expected_result => $expected_result,
					    );
		if ($shift == TRUE) {
			%res = $class->_write_existing_data(path => $path,
							    offset => $offset + $len,
							    data => $remaining_data,
							    );
		}
	}
	elsif ($type eq 'large')
	{
		# will probably depreciate this case
		%res = $class->_write_bytes(path => $path,
					    offset => $offset,
					    len => $len,
					    pattern => $pattern,
					    disable_checksums => $disable_checksums,
					    expected_result	=> $expected_result,
					    );
	}
	elsif ($type eq 'existing_data')
	{
		logresult("script", "'data' needs to be specified when type='existing_data'") if ! $data;
		%res = $class->_write_existing_data(path => $path,
						    offset => $offset,
						    data => $data,
						    expected_result => $expected_result,
						    );
	}
	elsif ($type eq 'uncompressable')
	{
		if ($expected_result == C_RC_INSUFFICIENT_SPACE ) {
			%res = $class->_write_uncompressable_data_alternate(path => $path,
									    offset => $offset,
									    len => $len,
									    expected_result => $expected_result,
									    );
		}
		else {
			%res = $class->_write_uncompressable_data(path => $path,
								  offset => $offset,
								  len => $len,
								  expected_result => $expected_result,
								  );
		}
	}
	elsif ($type eq 'blocksondisk')
	{
		%res = $class->_write_blocks_on_disk(path => $path,
						     offset => $offset,
						     len => $len,
						     pattern => $pattern,
						     blocksondisk => $blocksondisk,
						     expected_result => $expected_result,
						     );
	}
	elsif ($type eq 'compressable')
	{
		%res = $class->_write_compressable_data(path => $path,
							offset => $offset,
							len => $len,
							pattern	=> $pattern,
							compression => $compression,
							expected_result	=> $expected_result,
							);
	}
	elsif ($type eq 'holey')
	{
		my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$orig_file_size,$atime,$mtime,$ctime,$blksize,$blocks) = stat($path);
		my $new_file_size = $orig_file_size - $len;
		truncate($path, $offset);
		if ($remaining_data) {
			if ($shift == TRUE) {
				%res = $class->_write_existing_data(path => $path,
								    offset => $offset,
								    data => $remaining_data,);
				truncate($path, $new_file_size);
			}
			else {
				%res = $class->_write_existing_data(path => $path,
								    offset => $offset + $len,
								    data => $remaining_data,);				
			}
		}
		else {
			if ($shift == FALSE) {
				# creating sparse blocks at end of file by adding a byte at the end
				# and then truncating it off.
				%res = $class->_write_bytes(path => $path,
							    offset => $offset + $len,
							    len => 1,
							    pattern => 'x',);
				truncate($path, $orig_file_size);
			}
		}
	}
	else
	{
		croak "unknown create type $!";
	}
	
	return %res;
}

# Internal function
# name ::_write_bytes
#
# summary (INTERNAL) write bytes to a file
#
# description
# This is called from the _write_data function
#
# example
#			$class->_write_bytes(	path	=> $path,
#									offset	=> $offset,
#									len		=> $len,
#									pattern	=> $pattern,
#									expected_result => C_RC_SUCCESS,
#			);
# 
sub _write_bytes
{
	my $class 	= shift;
	my (%args) = validate (@_,  {	path		=> { type=>SCALAR },
									offset		=> { type=>SCALAR },
									len			=> { type=>SCALAR },
									pattern		=> { type=>SCALAR },
									disable_checksums => { type=>SCALAR, optional=>1, default=>TRUE },
									expected_result => { type=>SCALAR, optional=>1, default=>C_RC_SUCCESS },
	} );
	my $path		= $args{path};
	my $offset		= $args{offset};
	my $len			= $args{len};
	my $pattern		= $args{pattern};
	my $disable_checksums	= $args{disable_checksums};
	my $expected_result	= $args{expected_result};
	my $result;
	my %res;

	my $blocksize	= C_WAFL_BLOCK_SIZE;
	my $numblocks	= int ($len / $blocksize);
	my $remainder	= $len % $blocksize;
	my $shortblock	= $blocksize - 20;

	# create compression groups of pattern characters
	my $patternblock = $pattern x $blocksize;
	my $patterncg8 = $patternblock x 7;
	my $patterncg7 = $patternblock x 6;
	$patterncg8	.= $pattern x $shortblock;
	$patterncg7	.= $pattern x $shortblock;

	my $sha_sum;
	if ($disable_checksums == TRUE) {
		$patterncg8	= ($pattern x 20) . $patterncg8;
		$patterncg7	= ($pattern x 20) . $patterncg7;
	} else {
		$sha_sum = sha1($patterncg8);
		$patterncg8	= $sha_sum.$patterncg8;
		$sha_sum = sha1($patterncg7);
		$patterncg7	= $sha_sum.$patterncg7;
	}
	
	eval {
		# open the file for writing
		open(FH, "+<$path") || die("Could not open file $path: $!");
		seek(FH,$offset,0);

		# write as many complete compression groups as possible
		my $cg_counter = 1;
		my $sleep_counter = 0;
		while ( $numblocks >= 7 ) {
			$sleep_counter++;
			if ($sleep_counter == 500) {
				sleep 1;
				$sleep_counter = 0;
			}
			if ( $cg_counter == 32 ) {
				print FH $patterncg7 or die "$!";
				$cg_counter = 1;
				$numblocks -= 7;
			} elsif ( $numblocks >= 8 ) {
				print FH $patterncg8 or die "$!";
				$cg_counter++;
				$numblocks -= 8;
			} else {
				last;
			}
		}

		# create the final compression group if needed
		if ( $numblocks > 0 || $remainder > 0 ) {
			my $final_cg = '';
			if ($disable_checksums == TRUE) {
				$final_cg = $pattern x ($numblocks * $blocksize) if $numblocks > 0;
				$final_cg .= $pattern x $remainder if $remainder > 0;
			} else {
				if ( $remainder <= 20 && $numblocks == 0 ) {
					my $known_pattern = "01234567890123456789";
					$final_cg = substr( $known_pattern, 0, $remainder);
				} else {
					$final_cg	= $pattern x $shortblock if $numblocks != 0;
					$numblocks--;
					$final_cg	.= $patternblock x $numblocks if $numblocks != -1;
					$remainder -= 20 if  $numblocks == -1;
					$final_cg	.= $pattern x $remainder if $remainder != 0;
					$sha_sum	= sha1($final_cg);
					$final_cg	= $sha_sum.$final_cg;
				}
			}
			print FH $final_cg or die "$!";
		}
		close(FH);
	};

	if ($@) {
		$result = C_RC_FAILURE;
		$res{'error'} = $@;
	} else {
		$result = C_RC_SUCCESS;
	}
	
	if ( $result != $expected_result ) {
		croak "result: $result, expected: $expected_result $!" unless $expected_result == C_RC_FAILURE_NON_FATAL;
	}

	$res{'rc'} = $result;
	return %res;
}

# Internal function
# name ::_write_existing_data
#
# summary (INTERNAL) write data to a file
#
# description
# This is called by the _write_data function and 
# is for the case where you need to write the 
# same random data to more than one file. 
#
# example
#			$class->_write_existing_data(	path	=> $path,
#											offset	=> $offset,
#											data	=> $data,
# 
sub _write_existing_data
{
	my $class 	= shift;
	my (%args) = validate (@_,  {	path	=> { type=>SCALAR },
									offset	=> { type=>SCALAR },
									data	=> { type=>SCALAR },
									expected_result => { type=>SCALAR, optional=>1, default=>C_RC_SUCCESS },
	} );
	my $path	= $args{path};
	my $offset	= $args{offset};
	my $data	= $args{data};	
	my $expected_result	= $args{expected_result};
	my $result;
	my %res;

	eval {
		open(FH, "+<$path") || die("Could not open file $path: $!");
		seek(FH,$offset,0);
		print FH $data;
		close(FH);
	};

	if ($@) {
		$result = C_RC_FAILURE;
		$res{'error'} = $@;
	} else {
		$result = C_RC_SUCCESS;
	}
	
	if ( $result != $expected_result ) {
		croak "result: $result, expected: $expected_result $!" unless $expected_result == C_RC_FAILURE_NON_FATAL;
	}

	$res{'rc'} = $result;
	return %res;
}

# Internal function
# name ::_write_uncompressable_data
#
# summary (INTERNAL) write data to a file
#
# description
# This is called by the _write_data function and 
# is for the case where you need to write uncompressable 
# data to a file 
#
# example
#			$class->_write_uncompressable_data(	path	=> $path,
#											offset	=> $offset,
#											len		=> $len,
# 
sub _write_uncompressable_data
{
	my $class 	= shift;
	my (%args) = validate (@_,  {	path	=> { type=>SCALAR },
									offset	=> { type=>SCALAR },
									len	=> { type=>SCALAR },
									expected_result => { type=>SCALAR, optional=>1, default=>C_RC_SUCCESS },
	} );
	my $len		= $args{len};
	my $path	= $args{path};
	my $offset	= $args{offset};
	my $expected_result	= $args{expected_result};
	my $result;
	my %res;
	logcomment("_write_uncompressable_data");

	my $blocksize	= C_WAFL_BLOCK_SIZE;
	my $numblocks	= int ($len / $blocksize);
	my $remainder	= $len % $blocksize;
	my $shortblock	= $blocksize - 20;

	# create uncompressable compression groups
	my $data = '';
	my $placeholder = $shortblock + 7 * $blocksize;
	my $randcg8 = `dd if=/dev/urandom count=1 bs=$placeholder`;
	$placeholder -= $blocksize;
	my $randcg7 = `dd if=/dev/urandom count=1 bs=$placeholder`;

	my $sha_sum = sha1($randcg8);
	$randcg8	= $sha_sum.$randcg8;
	$sha_sum = sha1($randcg7);
	$randcg7	= $sha_sum.$randcg7;
	
	eval {
	# open the file for writing
	open(FH, "+<$path") || die("Could not open file $path: $!");
	seek(FH,$offset,0);

	# write as many complete compression groups as possible
	my $cg_counter = 1;
	my $sleep_counter = 0;
	while ( $numblocks >= 7 ) {
		$sleep_counter++;
		if ($sleep_counter == 500) {
			sleep 1;
			$sleep_counter = 0;
		}
		if ( $cg_counter == 32 ) {
			print FH $randcg7;
			$cg_counter = 1;
			$numblocks -= 7;
		} elsif ( $numblocks >= 8 ) {
			print FH $randcg8;
			$cg_counter++;
			$numblocks -= 8;
		} else {
			last;
		}
	}

	# create the final compression group if needed
	my $final_cg = '';
	if ( $numblocks > 0 || $remainder > 20 ) {
		if ( $numblocks > 0 ) {
			$placeholder = $shortblock + ($numblocks - 1) * $blocksize if $numblocks !=0;
			$final_cg = `dd if=/dev/urandom count=1 bs=$placeholder`;
		}
		$remainder -= 20 if $numblocks == 0;
		$final_cg	.= `dd if=/dev/urandom count=1 bs=$remainder`;
		$sha_sum	= sha1($final_cg);
		$final_cg	= $sha_sum.$final_cg;
	} else {
		my $known_pattern = "01234567890123456789";
		$final_cg = substr( $known_pattern, 0, $remainder);
	}
	$data	.= $final_cg;
	print FH $data;

	close(FH);
	};

	if ($@) {
		$result = C_RC_FAILURE;
		$res{'error'} = $@;
	} else {
		$result = C_RC_SUCCESS;
	}
	
	if ( $result != $expected_result ) {
		croak "result: $result, expected: $expected_result $!" unless $expected_result == C_RC_FAILURE_NON_FATAL;
	}

	$res{'rc'} = $result;
	return %res;
}

# Internal function
# name ::_write_compressable_data
#
# summary (INTERNAL) write data to a file
#
# description
# This is called by the _write_data function and 
# is for the case where you need to write compressable 
# data to a file 
#
# example
#			$class->_write_compressable_data(	path	=> $path,
#											offset	=> $offset,
#											compression	=> $compression,
#											len		=> $len,
sub _write_compressable_data
{
		
	my $class 	= shift;
	my (%args) = validate (@_,  {	path	=> { type=>SCALAR },
									offset	=> { type=>SCALAR },
									len		=> { type=>SCALAR },
									pattern	=> { type=>SCALAR },
									compression	=> { type=>SCALAR },
									expected_result => { type=>SCALAR, optional=>1, default=>C_RC_SUCCESS },
	} );

	my $path	= $args{path};
	my $offset	= $args{offset};
	my $compression	= $args{compression};
	my $pattern	= $args{pattern};
	my $len		= $args{len};
	my $expected_result	= $args{expected_result};
	my $result;

	my $large_cg = C_DFLT_CG_BLKS;
	my $small_cg = 7;

	my $cg_counter = 0;
	my $len_copy = $len;
	my $num_cg8 = 0;
	my $num_cg7 = 0; 
	## Count number and type of compression groups
	while ( $len_copy > 0 ) {
		if ( $cg_counter == 31 && $len_copy >= $small_cg * C_WAFL_BLOCK_SIZE ) {
			$num_cg7++;
			$cg_counter = 0;
			$len_copy -= $small_cg * C_WAFL_BLOCK_SIZE;
		} elsif ( $len_copy >= $large_cg * C_WAFL_BLOCK_SIZE ) {
			$num_cg8++;
			$cg_counter++;
			$len_copy -= $large_cg * C_WAFL_BLOCK_SIZE;
		} else { 
			last;
		}
	}
	my $max_compression = 0;
	$max_compression = (7/$large_cg) * $num_cg8 + (6/7) * $num_cg7;
	$max_compression = 100 * $max_compression / ( $num_cg8 + $num_cg7 ) if $len >= $large_cg * C_WAFL_BLOCK_SIZE;

	croak 'The value of compression must be declared for type compressable' if $compression == -1;
	croak "$compression is not a valid value for compression (0 - $max_compression)" if $compression < -1; 
	
	my $blocksize = C_WAFL_BLOCK_SIZE;
	my $numblocks = int ( $len / $blocksize );
	my $remainder = $len % $blocksize;
	if ( $len < C_MIN_DATA_TO_COMPRESS ) {
		logcomment ("File size too small to compress.");
		open(FH, "+<$path") || die("Could not open file $path: $!");
		seek(FH,$offset,0);

		if ( $len > 20 ) {
			$len -= 20;
			my $data = `dd if=/dev/urandom count=1 bs=$len`;
			print FH sha1($data);
			print FH $data;
		} else {
		my $known_pattern = "01234567890123456789";
		print FH substr($known_pattern, 0, $len);
		}
		close FH;
		goto ENDING;
	}

	my $emptyblocks = int($numblocks * $compression / 100);

	# Generate expected results
	my $fullblocks		= $numblocks - $emptyblocks;
	my $expected_savings	= $emptyblocks * $blocksize / KB;
	my $expected_compression = sprintf("%.3f",($expected_savings * KB / $len * 100));
	my $expected_error	= sprintf("%.3f",(abs($compression - $expected_compression)));

	# Check compression error of +- 1 "0" blocks looking for closer result (above four lines compressed into one)
	my $error_check_low	= abs($compression - ((($emptyblocks - 1)*4)*KB/$len*100));
	my $error_check_high = abs($compression - ((($emptyblocks + 1)*4)*KB/$len*100));

	if ( $expected_error > $error_check_low ) {
		if ( $error_check_high > $error_check_low && $emptyblocks > 0 ) {
			$emptyblocks -= 1;
		}
	} elsif ( $expected_error > $error_check_high && $numblocks > $emptyblocks + 1 ) {
			$emptyblocks += 1;
	}

	# Check for < 25% savings in first block
	my $blocks_in_first_cg = C_MAX_CG_BLKS;
	$blocks_in_first_cg	= $numblocks if $numblocks < C_MAX_CG_BLKS;

	my $num_comp_in_first_cg = $blocks_in_first_cg;
	$num_comp_in_first_cg	= $emptyblocks if $emptyblocks < $blocks_in_first_cg;
	
	if ($num_comp_in_first_cg / $blocks_in_first_cg < 0.25) {
		 logcomment "All Compression groups under 25% compressable, unable to supply $compression% compression for a $len Byte file.";
	}
	
	# Recalculate in case number of empty blocks was changed
	$fullblocks = $numblocks - $emptyblocks;
	$expected_savings = $emptyblocks * C_ONE_WAFL_BLOCK / KB;
	while( 100 * $emptyblocks / $numblocks  > $max_compression) {
		$emptyblocks--;
		$fullblocks = $numblocks - $emptyblocks;
		$expected_savings = $emptyblocks * C_ONE_WAFL_BLOCK / KB;
	}

	eval {
	# Begin writing the file
	open(FH, "+<$path") || die("Could not open file $path: $!");
	seek(FH,$offset,0);

	# Create compression groups of zeros with checksums
	my $shortblock = $blocksize - 20;
	my $placeholder = $shortblock + ( 7 * $blocksize );
	my $zerocg8	= $pattern x $placeholder;
	$placeholder -= $blocksize;
	my $zerocg7 = $pattern x $placeholder;
	
	my $sha_sum = sha1($zerocg8);
	$zerocg8	= $sha_sum.$zerocg8;
	$sha_sum	= sha1($zerocg7);
	$zerocg7	= $sha_sum.$zerocg7;

	# Add blocks of zeros until unable make a full cg
	$cg_counter = 1;
	while ( $emptyblocks >= $small_cg )	{
		if ( $cg_counter == 32 )
		{
			print FH $zerocg7;
			$emptyblocks -= 6;
			$fullblocks--;
			$cg_counter = 1;
		} elsif ( $emptyblocks >= $large_cg ) {
			print FH $zerocg8;
			$emptyblocks -= $small_cg;
			$fullblocks--;
			$cg_counter++;
		} elsif ( $emptyblocks == $small_cg ) { 
			last;
		}
	}

	# Prep for the half compressable and half uncompressable cg
	$numblocks = $emptyblocks + $fullblocks;
	my $cg_max_size = $large_cg;
	$cg_max_size = $small_cg if $cg_counter == 32;

	my $cg_size = $cg_max_size;
	$cg_size = $numblocks if $numblocks < $cg_max_size;

	# Adjust estimated savings
	# MAGIC: email kozel@netapp.com if you understand why this is needed!
	if ($cg_size < $cg_max_size && $remainder > 0) {
			$expected_savings += 4;
	} elsif ($emptyblocks / $cg_size < 0.25) {
		$expected_savings -= $emptyblocks * C_ONE_WAFL_BLOCK / KB;
	}

	# Create the mixed compression group
	$placeholder = $shortblock + ( $emptyblocks * $blocksize );
	my $mixed_cg = $pattern x $placeholder;
	$fullblocks--; # from one '0' block that gets to disk

	my $num_rand = $cg_size - ( $emptyblocks + 1 );
	$num_rand	= $fullblocks if $fullblocks < $num_rand;
	$mixed_cg	.= `dd if=/dev/urandom count=$num_rand bs=$blocksize` if $num_rand != 0;;
	if ( $cg_size < $cg_max_size ) {
		$mixed_cg	.= `dd if=/dev/urandom count=1 bs=$remainder`;
	}

	$sha_sum	= sha1($mixed_cg);
	$mixed_cg	= $sha_sum.$mixed_cg;

	$emptyblocks	-= $emptyblocks;
	$fullblocks		-= $num_rand;

	$cg_counter++;
	$cg_counter = 1 if $cg_counter == 33;
	print FH $mixed_cg;
	if ( $cg_size < $cg_max_size ) {
		close FH;
		goto ENDING;
	}
	
	#Add compression groups of uncompressable data until unable to make a complete group
	$placeholder = $shortblock + ( 7 * $blocksize );
	my $randcg8	= `dd if=/dev/urandom count=1 bs=$placeholder`;
	$placeholder	-= $blocksize;
	my $randcg7	= `dd if=/dev/urandom count=1 bs=$placeholder`;
	$sha_sum	= sha1($randcg8);
	$randcg8	= $sha_sum.$randcg8;
	$sha_sum	= sha1($randcg7);
	$randcg7	= $sha_sum.$randcg7;

	while ( $fullblocks >= $small_cg ) {
		if ($cg_counter == 32 ) {
			print FH $randcg7;
			$cg_counter = 1;
			$fullblocks -= $small_cg;
		} elsif ( $fullblocks >= $large_cg ) {
			print FH $randcg8;
			$cg_counter++;
			$fullblocks -= $large_cg;
		} else {
			last;
		}
	}

	my $remains = '';
		
	if ( $remainder <= 20 ) {
		$placeholder = ($fullblocks - 1) * $blocksize + $shortblock;
		$remains = `dd if=/dev/urandom count=1 bs=$placeholder` if $fullblocks != 0;;
		my $known_pattern = "01234567890123456789";
		$remains	.= substr( $known_pattern, 0, $remainder );
		if ( $fullblocks > 0 ) {
			$sha_sum = sha1($remains);
			$remains = $sha_sum.$remains;
		}
	} elsif ( $remainder != 0 ) {
		$remains	= `dd if=/dev/urandom count=$fullblocks bs=$blocksize` if $fullblocks != 0;	
		$remainder	-= 20;
		$remains	.= `dd if=/dev/urandom count=1 bs=$remainder`;
		$sha_sum	= sha1($remains);
		$remains	= $sha_sum.$remains;
	}
	
	print FH $remains;

	close(FH);
	};

	ENDING:

	$expected_savings = 0 if !defined($expected_savings);
	$expected_compression = sprintf("%.3f",($expected_savings * KB / $len * 100));
	$expected_error = sprintf("%.3f",(abs($compression - $expected_compression)));
	#logcomment  "$expected_compression is above the maximum compression. ($max_compression% smaller)" if $expected_compression > $max_compression;
	my $size = sprintf("%.3f",($len / KB));	
	logcomment ("\tDesired compression is $compression\n".
				"\t\t\tDesired file size is $size KB\n".
				"\t\t\tExpected savings is $expected_savings KB\n".
				"\t\t\tExpected compression is $expected_compression\n".
				"\t\t\tExpected error is $expected_error\n");

	my %return = (
				expected_savings	=> $expected_savings,
				expected_compression	=> $expected_compression,
				expected_error		=> $expected_error,
	);

	if ($@) {
		$result = C_RC_FAILURE;
		$return{'error'} = $@;
	} else {
		$result = C_RC_SUCCESS;
	}	
	if ( $result != $expected_result ) {
		croak "result: $result, expected: $expected_result $!" unless $expected_result == C_RC_FAILURE_NON_FATAL;
	}
	$return{'rc'} = $result;

	return %return;
}
 
# Internal function
# name ::_write_blocks_on_disk
#
# summary (INTERNAL) write data to a file
#
# description
# This is called by the _write_data function and 
# is for the case where you need to write a certain
# number of blocks per compression group to disk after 
# compression
#
# example
#			$class->_write_blocks_on_disk(	path	=> $path,
#											offset	=> $offset,
#											blocksondisk	=> $blocksondisk,
#											len		=> $len,
# 
sub _write_blocks_on_disk
{
	my $class = shift;
	my (%args) = validate (@_, {	path    => { type=>SCALAR },
									offset  => { type=>SCALAR },
									len		=> { type=>SCALAR },
									pattern	=> { type=>SCALAR },
									blocksondisk	=> { type=>SCALAR },
									expected_result => { type=>SCALAR, optional=>1, default=>C_RC_SUCCESS },
	} );

	my $path        = $args{path};
	my $offset      = $args{offset};
	my $len         = $args{len};
	my $pattern		= $args{pattern};
	my $expected_result	= $args{expected_result};
	my $result;
	my %res;

	my $blocksize   = C_WAFL_BLOCK_SIZE;
	my $blocksondisk = $args{blocksondisk};

	my $large_cg = C_DFLT_CG_BLKS;
	my $small_cg = 7;
	
	if ( $blocksondisk < -1 || $blocksondisk == 0 || $blocksondisk > 8 ){
		croak "$blocksondisk is not a valid value for blocks per cg (1 - 8)";
	}
	if ( $blocksondisk == 6 && $len > 31 * $large_cg * C_ONE_WAFL_BLOCK ){
		logcomment("Compression groups of 7 blocks will be 7 blocks on disk. (less than 25% savings)");
	}
	if ( $blocksondisk >= 7 ) {
		logcomment("All compression groups will be uncompressable");
	}

	# Create compression groups of 8 and 7 blocks
	my $sha_sum		= '';
	my $sum_size	= 20;
	my $randblocks	= $blocksondisk - 1;
	my $shortblock	= $blocksize - $sum_size;

	my $placeholder = 0;
	my $zeroblocks = -1;
	my $block8 = '';
	my $block7 = '';
	if ( $randblocks > 0) {
		$placeholder = $shortblock + ($randblocks - 1) * $blocksize; 
		$block8	= `dd if=/dev/urandom count=1 bs=$placeholder`;
		$block7	= `dd if=/dev/urandom count=1 bs=$placeholder`;
		$zeroblocks	= $large_cg - $randblocks;
	} else {
		$zeroblocks	= $large_cg;
		$block8	= $pattern x $shortblock;
		$block7	= $pattern x $shortblock;
		$zeroblocks--;
	}
	$block8	.= $pattern x ( $zeroblocks * $blocksize );
	$zeroblocks--;
	$block7	.= $pattern x ( $zeroblocks * $blocksize );

	# Add SHA1 checksums to blocks
	$sha_sum	= sha1($block8);
	$block8		= $sha_sum.$block8;
	$sha_sum	= sha1($block7);
	$block7		= $sha_sum.$block7;
	
	eval {
	# Open file for writing
	open(FH, "+<$path") || die("Could not open file $path: $!");
	seek(FH,$offset,0);

	# Create the file from the generic blocks
	my $numblocks	= int($len / $blocksize);
	while ( $numblocks >= 255 )	{
		print FH $block8 x 31;
		print FH $block7;
		$numblocks	-= 255;
	}

	while ( $numblocks >= $large_cg ) {
		print FH $block8;
		$numblocks	-= $large_cg;
	}
	# deal with a possible final incomplete compression group
	my $remainder = $len % $blocksize;
	my $temp_data = '';

	if ($numblocks >= $blocksondisk){
		if ( $randblocks > 0 ) {
			$placeholder = $shortblock + ($randblocks - 1) * $blocksize; 
			$temp_data = `dd if=/dev/urandom count=1 bs=$placeholder`;
			$numblocks -= $randblocks;
		} else {
			$temp_data = $pattern x $shortblock;
			$numblocks--;
		}
		$temp_data	.= $pattern x ($numblocks * $blocksize + $remainder);

		$sha_sum	= sha1($temp_data);
		$temp_data	= $sha_sum.$temp_data;
	} elsif ( $numblocks > 0 || $remainder > 20 ) {
		$placeholder = $numblocks * $blocksize + $remainder - 20;
		$temp_data = `dd if=/dev/urandom count=1 bs=$placeholder`;
		$sha_sum	= sha1($temp_data);
		$temp_data	= $sha_sum.$temp_data;
	} else {
		my $known_data = "01234567890123456789";	
		$temp_data = substr($known_data, 0, $remainder);
	}

	print FH $temp_data;
	close(FH);
	};

	if ($@) {
		$result = C_RC_FAILURE;
		$res{'error'} = $@;
	} else {
		$result = C_RC_SUCCESS;
	}
	
	if ( $result != $expected_result ) {
		croak "result: $result, expected: $expected_result $!" unless $expected_result == C_RC_FAILURE_NON_FATAL;
	}

	$res{'rc'} = $result;
	return %res;
}

sub _file_nfs_attach
{
	my $class 	= shift;
	my (%args) = validate (@_,
	{
		relative_path => { type=>SCALAR },
	});
	
	my $relative_path = $args{relative_path};
	
	if ($class->{_boot_mode} eq C_ONTAP_CMODE)
	{
		# For C-mode, we need to always mount the volume with RW, to get the .admin mountpoint
		# It looks like GNOTE is umounting/remounting the volume when creating the NFS object
		return $class->{volume}->nfs(RW => 1)->attach($relative_path);
	}
	else
	{
		return $class->{volume}->nfs->attach($relative_path);
	}
}

# Internal function
# repeat pattern to exact number of bytes
# This will probably be depreciated.  It's no longer used.  
sub _fill_pattern_to_size
{
	my $class 	= shift;
	my (%args) = validate (@_, { 
							size => { type=>SCALAR, regex => qr/^\d+$/ },
							pattern => { type=>SCALAR },							
							} );
	my $final_size = $args{size};
	my $pattern = $args{pattern};
	my @tmp =  split (//, $pattern);
	my $pattern_size = scalar @tmp;
	my $content = '';
	my $current_size = 0;
	for (my $i=1; $i<=$final_size; $i++) {
		$content .= $pattern;
		$current_size += $pattern_size;
		last if $current_size >= $final_size;
	}
	if ( $current_size > $final_size ) {
		my $counter = 0;
		my $temp_content;
		foreach my $byte (split //, $content) {
			$counter++;
			$temp_content .= $byte;
			if ($counter >= $final_size ) {
				$content = $temp_content;
				last;
			}
		}
	} elsif ( $current_size < $final_size ) {
		croak "ERROR: should never be here $!";
	}
	return $content;
}

# Internal function
# name ::_write_uncompressable_data_alternate
#
# summary (INTERNAL) write data to a file
#
# description
# This is an alternate version of _write_uncompressable_data
# that can be used when we need to catch the error for
# "No space left on device" which isn't caught in some cases ( see burt #372834 )
#
# example
#			$class->_write_uncompressable_data_alternate(	path	=> $path,
#											offset	=> $offset,
#											len		=> $len,
# 
sub _write_uncompressable_data_alternate
{
	my $class 	= shift;
	my (%args) = validate (@_,  {	path	=> { type=>SCALAR },
									offset	=> { type=>SCALAR },
									len	=> { type=>SCALAR },
									expected_result => { type=>SCALAR, optional=>1, default=>C_RC_SUCCESS },

	} );
	my $len		= $args{len};
	my $path	= $args{path};
	my $offset	= $args{offset};
	my $expected_result = $args{expected_result};
	my %res;
	my $result = C_RC_SUCCESS;

	logcomment("_write_uncompressable_data_alternate");
	my $blocksize	= C_WAFL_BLOCK_SIZE * 10;
	my $numblocks	= int ($len / $blocksize);
	my $remainder	= $len % $blocksize;

	eval {
		my $output;
		my $new_offset;
		for ($new_offset = $offset; $new_offset < $numblocks; $new_offset++) {
			$output = `dd if=/dev/urandom of=$path count=1 bs=$blocksize seek=$new_offset conv=notrunc 2>&1`;
			if ($output =~ /No space left on device/i) {
				$result = C_RC_INSUFFICIENT_SPACE;
				last;
			}
		}
		if ($remainder >0 ) {
			my $remainder_offset = $new_offset * $blocksize;
			$output = `dd if=/dev/urandom of=$path count=$remainder bs=1 seek=$remainder_offset conv=notrunc 2>&1`;
			if ($output =~ /No space left on device/i) {
				$result = C_RC_INSUFFICIENT_SPACE;
			}
		}	
	};

	logcomment("_write_uncompressable_data_alternate\n".
				"result: $result, expected: $expected_result");

	$res{'rc'} = $result;
	if ( $result != $expected_result ) {
		croak "result: $result, expected: $expected_result $!" unless $expected_result == C_RC_FAILURE_NON_FATAL;
	}

	return %res;
}

# internal function currently used by Compression::SnapVault::restore
# for the case when restoring to a different qtree
sub _replace_qtree_name {
	my $class = shift;
	my (%args) = validate( @_,
	{
        old_qtree_name => {type => SCALAR,},
		new_qtree_name => {type => SCALAR,},
     } );
	
	my $old_qtree_name = $args{'old_qtree_name'};
	my $new_qtree_name = $args{'new_qtree_name'};
	my $files = $class->{_common}->{files};
	my $mountpoint = $class->get_mountpoint();
	my $vol = $class->{volume};
	my $vol_mount_path = $vol->mount_path;

	foreach my $file (keys %$files)
	{
		$files->{$file}->{relative_path} =~ s/^$old_qtree_name/$new_qtree_name/;
		$files->{$file}->{path} =~ s#^$mountpoint/$old_qtree_name#$mountpoint/$new_qtree_name#;
		$files->{$file}->{filer_path} =~ s#^$vol_mount_path/$old_qtree_name#$vol_mount_path/$new_qtree_name#;
	}
}

# internal function currently used by Compression::SnapMirror::_update_dataset_hash
sub _get_qtree_file_hash {
	my $class = shift;
	my (%args) = validate( @_,
	{
        qtree_name => {type => SCALAR,},
    } );
	my $qtree_name = $args{'qtree_name'};
	my $files = $class->{_common}->{files};

	my $qtree_files = {};
	foreach my $file (keys %$files)
	{
		if ($files->{$file}->{relative_path} =~ /^$qtree_name/ ) {
			$qtree_files->{$file} = $files->{$file};
		}
	}
	return $qtree_files;
}

sub _update_vol_obj
{
	my $class = shift;
	
	$class->{_common}->_update_vol_obj(@_);
	$class->{volume} = $class->{_common}->{volume};
	$class->{_conn} = $class->{_common}->{_conn};
	$class->{vol_name} = $class->{_common}->{vol_name};
	
	if (defined($class->{_client}))
	{
		if ($class->{_boot_mode} eq C_ONTAP_7MODE)
		{
			$class->{_client} = NOTE::Client->new(name => $class->{client}, storage => $class->{volume} );
		}
		elsif ($class->{_boot_mode} eq C_ONTAP_CMODE)
		{
			$class->{_client} = NOTE::Client->new(name => $class->{client}, storage => $class->{volume}, skip_remount => 1 );
		}
	}
}

1;
