#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Decompress.pm#1 $
#

## @summary Compression Decompress routines
##
## @description
##    This module provides all the capabilities
##	required for decompression work flow.
##
## @synopsis
## use Compression::Decompress;
## use Compression::Constants qw(:all);
## use NOTE::Filer;
## use vars qw($FILER); 
## 
## BEGIN { require "Compression/setup_core.pl"; }
## 
## my $obj = Compression::Decompress->new( conn => $conn, volume => 'vol1' );	    
##
## @examples
##
## Used independently:
## my $conn = NOTE::Filer->connect( host => $FILER );
## my $obj = Compression::Decompress->new( conn => $conn, volume => 'vol1' );
## $obj->start( expected => C_RC_COMPRESSION_NOT_TURNED_OFF );
## $obj->status( expected => C_RC_NO_DECOMPRESSION_IN_PROGR );
## $obj->wait_for_percentage( min => 45, max => 95);
## $obj->stop();
##
## @notes
## Any Compression::Compress object will have a Decompress member, 'decompress'.
## If using Compression::Compress, all available Decompress methods are accessible through this member.
##
## @examples
##
## Used through a Compress object:
## my $compress = Compression::Compress->setup( filer => $FILER, volume => 'vol1' );
## $compress->{decompress}->start( expected => C_RC_NO_DATA_TO_DECOMPRESS );
## $compress->{decompress}->wait_for_percentage( min => 45, max => 95 );
## $compress->{decompress}->stop();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::Decompress;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use POSIX qw(floor);
use Data::Dumper;

use Tharn;

use Compression::Constants qw(:all);
use Compression::Version;
use Compression::Utils qw(_c_interpret_logresult
                          _send_cmd_to_filer
			  process_data_size_string);
use SayWrap qw(sayEOF);
# Only needed for the burt362962 work-around. To be removed!
use filer qw(filer_cli_advanced_say);



BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}



## @name ::new
##
## @summary Compression::Decompress Constructor.
##
## @description
## Creates an object representing decompress utility.
##
## @return Compression::Decompress object.
##
## @arg conn Required: NOTE::Filer object.
##
## @arg volume Optional: name of volume to be associated with this object.
## 
## @arg vfiler Optional: Name of vfiler.
##
## @arg ui Optional: Compression::UI object corresponding to the hosting filer.
## If not provided, it will be auto-generated.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
## 
## @arg supportability Optional: Compression::Supportability object,
## to be passed in if, along with the command completion, is it required
## to also check the supportability messages.
##
## @example
##
## my $obj = Compression::Decompress->new( conn => $conn, volume => 'vol1' );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
		conn 	=> {type => OBJECT,
                            },

		volume  => {type => SCALAR,
                            optional => 1,
                            },
                
                ui      => {type => OBJECT,
                            isa => "Compression::UI",
                            optional => 1,
                            },
                
                disable_cleanup => {type => BOOLEAN,
                                    optional => 1,
                                    default => FALSE,
                                    },
                
                debug  => {type => SCALAR,
                           optional => 1,
                           default => FALSE,
                           },
                
                vserver => {type => SCALAR|UNDEF,
                            optional => 1},
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default => C_ONTAP_BOOT_MODE},
                
                vfiler => {type => SCALAR|UNDEF,
			   optional => 1,},
                
                supportability => {type => UNDEF|OBJECT,
                                   optional => 1,},
		} );
	
	my $self = {};
        $self->{vol_name} = $args{volume} if (defined $args{volume});
	$self->{_conn} = $args{conn};
        $self->{_debug} = $args{debug};
        $self->{_disable_cleanup} = $args{disable_cleanup};
        $self->{_creator_pid} = $$;
        $self->{_boot_mode} = $args{boot_mode};
        $self->{vserver} = $args{vserver};
        $self->{_total_kb_to_scan} = {};
        $self->{_refresh_total_kb_to_scan} = TRUE;
        $self->{_vfiler} = $args{vfiler};
        $self->{_supportability} = $args{supportability};
        
        if (($self->{_boot_mode} eq C_ONTAP_7MODE) && !($self->{_conn}->isa("NOTE::Filer")))
        {
                logresult('script', "For 7-mode, the conn argument is required to be a NOTE\:\:Filer object");
        }
        elsif (($self->{_boot_mode} eq C_ONTAP_CMODE) && !($self->{_conn}->isa("NOTE::GX::Node")))
        {
                logresult('script', "For 10-mode, the conn argument is required to be a NOTE\:\:GX\:\:Node object");
        }
        
        if ((defined($self->{_supportability})) && (!($self->{_supportability}->isa("Compression::Supportability"))))
        {
                logresult('script', "The supportability argument, if passed in, must be a Compression\:\:Supportability object");
        }
        
        logcomment("Decompress: new: PID of the script creating this object: " . $self->{_creator_pid}) if ($self->{_debug});
	
	####################################################
	# Dependency on other internal modules
	####################################################
	
        # UI
        if ( defined($args{ui}) )
        {
                $self->{_ui} = $args{ui};
        }
        else
        {
                # Auto-detect version of the hosting filer. Need this for UI.
                logcomment("Decompress: UI object not provided. Auto-generating...") if ($self->{_debug});
                my $version = Compression::Version->new( conn => $self->{_conn}, debug => $self->{_debug} );
                $self->{_ui} = Compression::UI->new( version => $version, debug => $self->{_debug}, boot_mode => $self->{_boot_mode} );
        }
        
        $self->{scanner_will_restart_after_filer_down} = $self->{_ui}->get_constant_value( constant_code => 'DECOMPRESS_RESTARTS_AFTER_REBOOT' );
	
	bless $self, $class;

	return $self;
}


## @name ::start
## @summary Starts decompression of a specific volume.
## 
## @description
## Starts decompression of a specific volume.
##
## @return Hash with rc and output
## 
## @arg volume Optional: Name of the volume.
## If not provided, uses volume associated to the object, if available.
## If not provided and the object does not have an associated volume, logs script error.
##
## @arg timeout Optional: Timeout, in seconds, for the command to complete.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
## 
## @arg target_decompress Optional: Boolean flag to start the scanner to only decompress
## (for RR: -C option). Legal values: undef, TRUE, false. Default value: TRUE.
## 
## @arg target_dedupe Optional: Boolean flag to start the scanner to only dedupe
## (for RR: -D option). Legal values: undef, TRUE, false. Default value: undef.
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @note
## If the SUPPORTABILITY cmd-line argument is 1, then,
## along with this action, the correspondent supportability message will be also checked.
##
## @examples
##
## $obj->start( volume => 'vol2', expected => C_RC_COMPRESSION_NOT_TURNED_OFF );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

sub start
{
	
	my $class 	= shift;
	my (%args) = validate( @_,
		{
			volume => { type => SCALAR, optional => 1, default  => $class->{vol_name}, },
                        vserver => {type => SCALAR, optional => 1, default  => $class->{vserver}, },
                        timeout => { type => SCALAR, regex => qr/^\d+$/, optional => 1, },
			expected => { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },
                        target_decompress => { type => SCALAR|UNDEF, optional => 1, default  => TRUE, },
                        target_dedupe => { type => SCALAR|UNDEF, optional => 1, },
			#vol_autosize => { type => BOOLEAN, optional => 1, default => FALSE, },	
		} );
	
        # TODO
	#my $autosize = $args{vol_autosize};  # if true then it will increase the volume size by # of compressed blocks so that decompress can succeed

	my $vol_name = $args{volume};
        my $expected = $args{expected};
	if ( !defined($vol_name) ) {
		logresult( "script", "Volume name was neither passed during initialization nor to start api" );
	}
        
        my $rc = C_RC_FAILURE;
        my $cmd_code = C_CMD_DECOMPRESS_START;
        
        if ((defined($class->{_supportability})) && ($expected == C_RC_SUCCESS))
        {
                $class->{_supportability}->start_tailing(events => [{code => C_EVENT_DECOMPRESS_STARTED,
                                                                     options => {target_decompress => $args{target_decompress},
                                                                                 target_dedupe => $args{target_dedupe},},},
                                                                    ],
                                                         required_identifier => $vol_name,
                                                         );
        }
        
        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $vol_name,
                                                       vserver => $args{vserver},
                                                       options => {target_decompress => $args{target_decompress},
                                                                   target_dedupe => $args{target_dedupe},
                                                                   });

        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            timeout => $args{timeout},
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
        
        logcomment("Decompress: start: command output: $output") if ($class->{_debug});
        
        my %res =  $class->{_ui}->process_output_string(code => $cmd_code,
                                                        output => $output,
                                                        expected_rc => $expected,
                                                        log_message => "Start decompression of volume $vol_name",);
        
        if ((defined($class->{_supportability})) && ($expected == C_RC_SUCCESS))
        {
                $class->{_supportability}->stop_and_match();
        }
        
        return %res;
}


## @name ::status
## @summary Reports the status of volume decompression.
## 
## @description
## Reports the status of volume decompression.
##
## @return Hash with rc, output, and percent 
## 
## @arg volume Optional: Name of the volume.
## If not provided, uses volume associated to the object, if available.
## If not provided and the object does not have an associated volume, reports decompress status for all volumes.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes there should be decompression in progress.
##
## @examples
## $obj->status( volume => 'vol2', expected => C_RC_NO_DECOMPRESSION_IN_PROGR );
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub status
{
	my $class 	= shift;
	my (%args) 	= validate( @_,
		{
			volume => {type => SCALAR, optional => 1, default => $class->{vol_name},},
                        vserver => {type => SCALAR|UNDEF, optional => 1, default  => $class->{vserver}, },
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
		} );
	
	# Allow undefined volume => in this case, return the status for all volumes
	my $vol_name = $args{volume};
        my $expected = $args{expected};
        my $cmd_code = C_CMD_DECOMPRESS_STATUS;
        # For RR and after, the status gives KB processed, as opposed to percentage
        # Calculate the percentage based on the currently-used space
        # Start with a negative value (to get the used space only once)
        my %status_hash = ();
        
        my $command = $class->{_ui}->formulate_command(code => $cmd_code,
                                                       vserver => $args{vserver},
                                                       volume => $vol_name,);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
        
        logcomment("Decompress: status: command output: $output") if ($class->{_debug});
        
        # Tricky here because...
        # The 'vol decompress status' command seems to ignore the volume name, if given (for pre-RR)
        # For RR, 'sis status -l' information is different.
        # For Pre-RR, we will get the parsed info in a 'percent' hash,
        # For RR, we will get it in a 'decompress_statistics' hash.
        # Both need extra processing
        
        # First, process the output using release-specific rules
        %status_hash = $class->{_ui}->process_output_string(code => $cmd_code,
                                                            output => $output,
                                                            expected_rc => C_RC_FAILURE_NON_FATAL,
                                                            log_message => "Obtain decompression status",);
        logcomment("Formatted status:\n" . Dumper(%status_hash)) if ($class->{_debug});
        
        # For RR and after, we do not have a percentage
        # Loop through all the volumes and calculate it
        if (exists($status_hash{decompress_statistics}))
        {
                my $volumes_status = $status_hash{decompress_statistics};
                foreach my $volume (keys %$volumes_status)
                {
                        my $vol_status_hash = $volumes_status->{$volume};
                        
                        my $expected_status = $class->{_ui}->{constants}->get_constant_value(constant_code => C_STATUS_UNDOING);
			my $initializing_status = $class->{_ui}->get_constant_value(constant_code => 'C_STATUS_INITIALIZING');
                        my $obtained_status = $vol_status_hash->{status};
                        if (("$expected_status" ne "$obtained_status") && ("$initializing_status" ne "$obtained_status"))
                        {
                                logcomment("Decompress: status: unexpected scanner status for $vol_name: $obtained_status\n" .
                                           "Expected status: $expected_status") if ($class->{_debug});
                                next;
                        }
			elsif ("$initializing_status" eq "$obtained_status")
			{
				logcomment("Decompress: status: scanner is initializing") if ($class->{_debug});
				$status_hash{percent}{$volume} = 0;
			}
                        else
                        {
                                my $kb_to_scan_entries = $class->{_total_kb_to_scan};
                                if (($class->{_refresh_total_kb_to_scan} == TRUE) || (!defined($kb_to_scan_entries->{$volume})))
                                {
                                        $kb_to_scan_entries->{$volume} = $class->_get_total_kb_to_process(volume => $volume);
                                }
                                
                                my $total_kb_to_scan = $kb_to_scan_entries->{$volume};
                                my $kb_processed = 1;
                                if (defined($status_hash{decompress_statistics}{$volume}{data_processed}))
                                {
                                        $kb_processed = $status_hash{decompress_statistics}{$volume}{data_processed};
                                }
                                
                                my $percentage_scanned = floor(($kb_processed / $total_kb_to_scan) * 100);
                                
                                # TODO: follow-up: burt?
                                # There are cases when the scanner reports more KB scanned then the calculated $total_kb_to_scan
                                # In these cases, bring the percentage down to 99, since the scanner did not complete yet
                                if ($percentage_scanned > 99)
                                {
                                        logcomment("Obtained percentage scanned: $percentage_scanned\n" .
                                                   "Setting it to 99") if ($class->{_debug});
                                        $percentage_scanned = 99;
                                }
                                
                                logcomment("Decompress status for volume $volume:\n" .
                                           "Total KB to scan: $total_kb_to_scan\n" .
                                           "KB processed: $kb_processed\n" .
                                           "Percentage scanned: $percentage_scanned");
                                
                                $status_hash{percent}{$volume} = $percentage_scanned;
                        }
                }
        }
        
        # Then look for status for specific volumes, if needed.
        my $rc = $status_hash{rc};
        if (($rc == C_RC_SUCCESS) && (defined($vol_name)))
        {
                if (!defined($status_hash{percent}{$vol_name}))
                {
                        logcomment("Decompress: no decompression in progress for volume $vol_name") if ($class->{_debug});
                        $rc = C_RC_NO_DECOMPRESSION_IN_PROGR;
                }
        }
        
        # Check against expected
        my %checked_status = $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                                  expected_rc => $expected,
                                                                  log_message => "Obtain decompression status",
                                                                  output      => $output);
        
        # Return the status hash, has more information.
        $status_hash{rc} = $checked_status{rc};
        return %status_hash;
}


## @name ::wait_for_percentage
## @summary Waits until the status of volume decompression is within specified limits.
## 
## @description
## Waits until the status of volume decompression is within specified limits.
##
## @return Hash with rc and output 
## 
## @arg volume Optional: Name of the volume.
## If not provided, uses volume associated to the object, if available.
## If not provided and the object does not have an associated volume, logs script error.
##
## @arg min Optional: Lower limit for expected decompress percentage.
## Defaults to 100.
##
## @arg max Optional: Upper limit for expected decompress percentage.
## Defaults to 100.
##
## @arg check_interval Optional: Number of seconds between two consecutive status checks.
## Defaults to C_DFLT_DECOMPRESS_INTERVAL (defined in Compression::Constants).
##
## @arg timeout Optional: Number of seconds to wait for required percentage.
## Defaults to C_DFLT_DECOMPRESS_TIMEOUT (defined in Compression::Constants).
##
## @examples
##
## $obj->wait_for_percentage( volume => "vol2", min => 15, max => 50, check_interval => 3, timeout => 600);
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @note
## If the SUPPORTABILITY cmd-line argument is 1, then,
## when waiting for the scanner to complete, the correspondent supportability message will be also checked.
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub wait_for_percentage
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			volume => {type => SCALAR, optional => 1, default  => $class->{vol_name},},
			vserver => {type => SCALAR|UNDEF, optional => 1, default  => $class->{vserver}, },
			min => {type => SCALAR, regex => qr/^\d+$/, optional => 1, default  => 100,},
                        max => {type => SCALAR, regex => qr/^\d+$/, optional => 1, default => 100,},
                        check_interval => {type => SCALAR, optional => 1, default => C_DFLT_DECOMPRESS_INTERVAL,},
                        timeout  => {type => SCALAR, optional => 1, default => C_DFLT_DECOMPRESS_TIMEOUT,},
                        disable_wait_after_completion => {type => BOOLEAN, optional => 1, default => FALSE,},
		} );

	my $min_percent = $args{min};
	my $max_percent = $args{max};
	my $check_interval = $args{check_interval};
	my $timeout= $args{timeout};
	my $elapsed_time = 0;
	my %status_hash = ();
	my $rc = C_RC_FAILURE;
	
	my $vol = $args{volume};
	if ( !defined($vol) ) {
		logresult( "script", "Volume name was neither passed during initialization nor to wait_for_percentage api" );
	}
	
	if ($min_percent > $max_percent) {
		logresult( "script", "min_percent($min_percent) > max_percent($max_percent)" );
	}
        
        if ((defined($class->{_supportability})) && ($min_percent == 100))
        {
                $class->{_supportability}->start_tailing(events => [{code => C_EVENT_DECOMPRESS_COMPLETED,},],
                                                         required_identifier => $vol,
                                                         );
        }
        
        logcomment("Waiting for the decompression scanner for volume $vol:\n" .
                   "Min percent: $min_percent\n" .
                   "Max percent: $max_percent\n" .
                   "Timeout: $timeout\n" .
                   "Check interval: $check_interval");
        
        # Reset _total_kb_to_scan, to be freshly retrieved (only once within this method)
        $class->{_refresh_total_kb_to_scan} = FALSE;
        delete($class->{_total_kb_to_scan}->{$vol});
	
	while ($elapsed_time <= $timeout) {
		%status_hash = $class->status(volume => $vol, vserver => $args{vserver}, expected => C_RC_FAILURE_NON_FATAL);
		logcomment("Decompression status returned:\n" . Dumper(%status_hash)) if ($class->{_debug});
                
		if ($status_hash{rc} != C_RC_SUCCESS)
		{
			$rc = $status_hash{rc};
			if (($min_percent == 100) && ($rc == C_RC_NO_DECOMPRESSION_IN_PROGR)) {
				# we wanted to wait until decompress is complete
				$rc = C_RC_SUCCESS;
                                
                                # TODO: To be removed!
                                # Note: this is only a work-around for burt362962
                                if (($class->{_boot_mode} eq C_ONTAP_7MODE) && ($args{disable_wait_after_completion} != TRUE))
                                {
                                        my $conn_rec = $class->{_conn}->conn();
                                        my $walf_scan_status = filer_cli_advanced_say($conn_rec, "wafl scan status $vol");
                                        logcomment("Decompress: burt362962 work-around: wafl scan status $vol:\n$walf_scan_status") if ($class->{_debug});
                                        if ($walf_scan_status =~ /blocks decompressed \[\d+ compressed blocks remaining/)
                                        {
                                                sleep($check_interval);
                                                $elapsed_time += $check_interval;
                                                logcomment("Decompress: burt362962 work-around:" .
                                                           " wafl scan status still shows decompression in progress") if ($class->{_debug});
                                                next;
                                        } 
                                }
                                # End work-around for burt362962
			}
			last;
		}
		
		my $current_percent = $status_hash{percent}{$vol};
		if (!defined($current_percent))
		{
			$rc = C_RC_FAILURE;
			last;
		}
		logcomment("percent complete $current_percent. Time $elapsed_time\n") if ($class->{_debug});
		
		if (($current_percent >= $min_percent) && ($current_percent <= $max_percent)) {
			logcomment("Decompress :: wait_for_percentage_complete SUCCESS: percentage: $current_percent\n") if ($class->{_debug});
			$rc = C_RC_SUCCESS;
			last;
		}
		
		if ($current_percent > $max_percent) {
			$rc = C_RC_PERCENTAGE_MISSED;
			last;
		}
		
		if (($min_percent - $current_percent)  <= ($max_percent - $min_percent)/2) {
			# Getting close to the targeted percentage: minimize sleep interval to not miss it
			$check_interval = 1;
			logcomment ("sleep interval reduced to $check_interval\n") if ($class->{_debug});
		}
		
		logcomment("Sleeping $check_interval seconds...\n") if ($class->{_debug});
		sleep($check_interval);
		$elapsed_time += $check_interval;
		logcomment("Looping again: elapsed time: $elapsed_time\n") if ($class->{_debug});
	}
        
        # Reset _refresh_total_kb_to_scan
        $class->{_refresh_total_kb_to_scan} = TRUE;
        
        if ((defined($class->{_supportability})) && ($min_percent == 100))
        {
                $class->{_supportability}->stop_and_match();
        }
	
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => C_RC_SUCCESS,
                                                    log_message => "Wait for decompress progress to be in between $min_percent% and $max_percent%",
                                                    output      => $status_hash{output});
}

## @name ::stop
## @summary Stops decompression for a specific volume.
## 
## @description
## Stops decompression for a specific volume.
##
## @return Hash with rc and output.
## 
## @arg volume Optional: Name of the volume.
## If not provided, uses volume associated to the object, if available.
## If not provided and the object does not have an associated volume, logs script error.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Defaults to C_RC_SUCCESS.
##
## @examples
## $obj->stop( volume => 'vol2' );
## $obj->stop( expected => C_RC_NO_DECOMPRESSION_IN_PROGR );
##
## @note
## If the VFILER, VFILER_IP, VFILER_IF command-line parameters
## are specified, then this action will be executed in the vfiler context.
##
## @note
## If the SUPPORTABILITY cmd-line argument is 1, then,
## along with this action, the correspondent supportability message will be also checked.
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub stop
{
	my $class 	= shift;
	my (%args) = validate( @_,
		{
			volume => {type => SCALAR, optional => 1, default => $class->{vol_name},},
                        vserver => {type => SCALAR|UNDEF, optional => 1, default  => $class->{vserver}, },
                        expected => {type => SCALAR, optional => 1, default => C_RC_SUCCESS,},
		} );

	my $vol_name = $args{volume};
        my $expected = $args{expected};
        
	if ( !defined($vol_name) ) {
		logresult( "script", "Volume name was neither passed during initialization nor to stop api" );
	}
        
        my $rc = C_RC_FAILURE;
        my $cmd_code = C_CMD_DECOMPRESS_STOP;
        
        if ((defined($class->{_supportability})) && ($expected == C_RC_SUCCESS))
        {
                $class->{_supportability}->start_tailing(events => [{code => C_EVENT_DECOMPRESS_ABORTED,},],
                                                         required_identifier => $vol_name,
                                                         );
        }
        
        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       vserver => $args{vserver},
                                                       volume => $vol_name);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
        
        logcomment("Decompress: stop: command output: $output") if ($class->{_debug});
        
        my %res = $class->{_ui}->process_output_string(code => $cmd_code,
                                                       output => $output,
                                                       expected_rc => $expected,
                                                       log_message => "Stop decompression of volume $vol_name",);
        
        if ((defined($class->{_supportability})) && ($expected == C_RC_SUCCESS))
        {
                $class->{_supportability}->stop_and_match();
        }
        
        return %res;
}


## @name ::get_last_operation_size
## @summary Get the last operation size. 
## 
## @description
## Get the size (in KB) processed by the last-running scanner. 
##
## @return The size in KB.
## 
## @arg volume Optional: Name of the volume.
## If not provided, uses volume associated with the object, if available.
## If not provided and the object does not have an associated volume, logs script error.
## 
## @arg vserver Optional: Name of the vserver (for c-mode testing).
## If not provided, uses vserver associated with the object, if available.
##
## @examples
## my $size = $obj->get_last_operation_size();
##
sub get_last_operation_size
{
        my $class = shift;
        my %args = validate( @_,
        {
                volume => {type => SCALAR,
                           optional => 1,
                           default => $class->{vol_name},},
                
                vserver => {type => SCALAR|UNDEF,
                            optional => 1,
                            default  => $class->{vserver}, }, 
	});
        
        my $volume = $args{volume};
        
        $args{expected} = C_RC_FAILURE_NON_FATAL;
        my %res = $class->status(%args);
        
        my $status = $res{decompress_statistics};
        
        if (!exists($status->{$volume}->{last_operation_size}))
        {
                logresult('fatal', "Failed to detect the last operation size of decompress scanner for volume $volume:\n" . Dumper(%res));
        }
        
        my $size = $status->{$volume}->{last_operation_size};
        
	# For c-mode, the last operation size is not in human-friendly format -> it is simply the size in B
	# In this case, there is no extra processing needed
        if ($size =~ /\D/)
        {
		# Remove any spaces to be processed correctly by process_data_size_string
		$size =~ s/\s//g;
		%res = Compression::Utils::process_data_size_string(size_string => $size,);
		$size = $res{size_in_bytes};
	}
        
        $size = $size / KB;
	logcomment("\nDecompress scanner: volume: $volume\n" .
                   "Last operation size: $size KB");
        
        return $size;
}

################################################################
# Internal subroutines
################################################################

sub _update_volume_name
{
        my $class = shift;
	my (%args) = validate( @_,
	{
		new_name => { type => SCALAR, },
	} );
        
        $class->{vol_name} = $args{new_name};
        
        logcomment("Decompress: class volume was renamed to: " . $class->{vol_name}) if $class->{_debug};
}


sub _update_vfiler
{
	my $class = shift;
	my $new_vfiler = shift;
	
	$class->{_vfiler} = $new_vfiler;
}


sub _get_total_kb_to_process
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                volume => { type => SCALAR, },
	} );
        
        my $vol_name = $args{volume};
        my $modifier = 'k';
        my $cmd_code = C_CMD_GET_SAVINGS;
        
        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $vol_name,
						       modifier => $modifier);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
							    zapi_available => $class->{_ui}->is_zapi_available(code => $cmd_code),);
        
        my %savings = $class->{_ui}->process_output_string(code => $cmd_code,
                                                           output => $output,
                                                           required_return_hash_key => $vol_name,
                                                           modifier => $modifier,
                                                           expected_rc => C_RC_SUCCESS,
                                                           log_message => 'Obtain space usage to be used to calculate decompress scanner progress',);
        
        my $total_saved = $savings{savings}{$vol_name}{total_saved};
        my $used = $savings{savings}{$vol_name}{used};
	
        return ($total_saved + $used);
}


sub _update_conn
{
	my $class = shift;
	my $new_conn = shift;
	
	# This assumes that the new filer is running the same version 
	if (($class->{_boot_mode} eq C_ONTAP_7MODE) && !($new_conn->isa("NOTE::Filer")))
        {
                logresult('script', "For 7-mode, the conn argument is required to be a NOTE\:\:Filer object");
        }
        elsif (($class->{_boot_mode} eq C_ONTAP_CMODE) && !($new_conn->isa("NOTE::GX::Node")))
        {
                logresult('script', "For 10-mode, the conn argument is required to be a NOTE\:\:GX\:\:Node object");
        }
	$class->{_conn} = $new_conn;
}


# Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
	
	if (( defined $class->{vol_name} ) &&
            ($class->{_disable_cleanup} == FALSE) &&
            ($class->{_creator_pid} == $$))
	{
                my $vol_name = $class->{vol_name};
                logcomment("Executing Decompress clean-up for volume $vol_name");

		logcomment( "Decompress: cleanup for volume $vol_name: stopping decompression") if ($class->{_debug});
                eval
                {
                        my %decompress_status = $class->status( expected => C_RC_FAILURE_NON_FATAL );
                        if ($decompress_status{rc} == C_RC_SUCCESS)
                        {
                                $class->stop();
                        }
                };
                
                if( $@ )
                {
                        logcomment( "Decompress: cleanup for volume $vol_name: error: " . $@ );
                }
                else
                {
                        logcomment( "Decompress: cleanup for volume $vol_name: done" ) if ($class->{_debug});
                }
        }
	else
	{
                logcomment( "Decompress: cleanup:  nothing done: script PID: " . $$) if ($class->{_debug});
	}
        
        if (defined($execution_error))
        {
                _c_interpret_logresult($execution_error);
        }
}

 
# routine to test the module itself
sub testme
{
	package main;
	use Compression::Constants qw(:all);
	use Compression::Compress;
	use Compression::Decompress;
        use filer;
	use vars qw($FILER $AGGREGATE $VOLUME1 $VOLUME2);

	require "Compression/setup_core.pl";
        
        param('VOLUME1', "-default", C_DFLT_VOLUME);
        param('VOLUME2', "-default", 'named_comp_vol');

	my $compress1 = Compression::Compress->setup(filer => $FILER ,
                                                     aggregate => $AGGREGATE,
                                                     volume => {name => $VOLUME1,
                                                                size => "4g",},
                                                     );
        
        my $compress2 = Compression::Compress->setup(filer => $FILER ,
                                                     aggregate => $AGGREGATE,
                                                     volume => {name => $VOLUME2,
                                                                size => "4g",},
                                                     );

        my $filer = NOTE::Filer->connect( host => $FILER );
        my $filename = 'large_file';
        my $vol_name1 = $compress1->{vol_name};
        my $vol_name2 = $compress2->{vol_name};
        
        if ($vol_name2 ne 'named_comp_vol')
        {
                logresult("fatal", "Failed to create named volume. Volume: $vol_name2 was created instead");
        }
        
        # Decompress object with no volume asociated
	my $decompress_no_vol = Compression::Decompress->new( conn  => $filer, );
        
        $compress1->{decompress}->start( expected => C_RC_COMPRESSION_NOT_TURNED_OFF );
        
        $compress1->{dataset}->create(
		files =>[ 
			{ name => $filename, size => 3 * GB, },
			],
	);
        
        $compress2->{dataset}->create(
		files =>[ 
			{ name => $filename, size => 3 * GB, },
			],
	);
        
        $compress1->wait_for_cp();

	$compress1->disable();
        $compress2->disable();
        
	$compress1->{decompress}->status( expected => C_RC_NO_DECOMPRESSION_IN_PROGR );
        $compress2->{decompress}->status( expected => C_RC_NO_DECOMPRESSION_IN_PROGR );
        
        $compress1->{decompress}->start(expected => C_RC_SUCCESS);
        $compress2->{decompress}->start(expected => C_RC_SUCCESS);
        
        my %result = $decompress_no_vol->status();
        # Verify here that we got savings for both volumes:
        if (( !defined($result{percent}{$vol_name1}) )  || ( !defined($result{percent}{$vol_name2}) ))
        {
            logresult("fatal", "Decompress status (both volumes) incomplete");
        }
        
        # Get the volume that has decompression going on
        my $dec_vol = ($result{percent}{$vol_name1} > 0) ? $compress1 : $compress2;
        my $dec_vol_name = $dec_vol->{vol_name};
        my $other_vol = ($result{percent}{$vol_name1} > 0) ? $compress2 : $compress1;
        my $other_vol_name = $other_vol->{vol_name};
        
        $dec_vol->{decompress}->wait_for_percentage(min => 1, max => 50);
        $dec_vol->{decompress}->stop();
        
        %result = $other_vol->{decompress}->status( expected => C_RC_SUCCESS );
        if ( !defined($result{percent}{$other_vol_name}) )
        {
            logresult("fatal", "Decompress status for $other_vol_name incomplete");
        }
        
        $dec_vol->{decompress}->status( expected => C_RC_NO_DECOMPRESSION_IN_PROGR );
        
	logcomment( "Decompress: testme completed" );
}

1;
