#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Fiji.pm#1 $
#

## @summary Wrapper around NOTE::Fiji.
##
## @description
## Wrapper around NOTE::Fiji.
## The main reason for not using NOTE::Fiji directly is to provide standardized output,
## and to allow operations with compression-specific fiji points.
##
## @synopsis
##
## use Compression::Fiji;
## use vars qw($FILER );
## 
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $fiji = Compression::Fiji->new(conn => $note_filer, tag => 'Simulate nvlog out of space',);
##
## @examples
## $fiji->set_point();
## $fiji->get_slot();
## $fiji->check_if_active();
## $fiji->clear_point();
##
## @notes
## Fiji points corresponding to C_FIJI_TAG_DISABLE_CP and C_FIJI_TAG_ENABLE_CP
## present a special behavior:
## 1.) They are each-other's oposite (C_FIJI_TAG_ENABLE_CP should be set to anihilate the effects of C_FIJI_TAG_DISABLE_CP).
## 2.) Clearing them does not mean that their effect is also terminated.
##    On the opposite, even if they are not cleared, their effect may be already terminated (if the 'oposite' Fiji point was set in the meantime).
## 3.) As a result of 2, check_if_active is not relevant to the effect of these two Fiji points.
## 4.) As a result of 2, wait_for_point should not be used for these two Fiji points.
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::Fiji;

use strict;
use warnings;

use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Tharn;

use Compression::Constants qw(:all);



BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}



## @name ::new
## @summary Compression::Fiji Constructor.
##
## @description
## Creates an object representing a Fiji point. It does not set the fiji point on the filer.
##
## @return Compression::Fiji object.
##
## @arg conn Required: NOTE::Filer object.
##
## @arg tag Required: Tag corresponding to the fiji point (defined in Compression::Constants).
##
## @arg ui Optional: Compression::UI object corresponding to the hosting filer.
## If not provided, it will be auto-generated.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @arg volname Optional: Name of volume, to be used for FIJI points enabled for a specific volume.
##
## @arg sleep_stage Optional: Sleep stage, to be used for the dense.sleep fiji points.
##
## @arg sleep_time Optional: Sleep time, to be used for the dense.sleep fiji points.
##
## @example
## my $obj = Compression::Fiji->new( conn => $conn, tag => 'Simulate nvlog out of space' );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub new
{
	my $class = shift;
        my %args = validate_with
                (params => \@_,
                 spec   => {conn => {type => OBJECT,},
                            tag => {type => SCALAR,},
                            ui => {type => OBJECT, isa => "Compression::UI", optional => 1,},
                            debug => {type => SCALAR, optional => 1, default => FALSE,},
                            volname => {type => SCALAR|UNDEF, optional => 1,},
                            boot_mode => {type => SCALAR, default => C_ONTAP_7MODE,},
                            sleep_stage => {type => SCALAR|UNDEF, optional => 1,},
                            sleep_time => {type => SCALAR|UNDEF, optional => 1,},
                            error_value => {type => SCALAR|UNDEF, optional => 1,},
                            inode_num => {type => SCALAR|UNDEF, optional => 1,},
                            level => {type => SCALAR|UNDEF, optional => 1,},
                            fbn => {type => SCALAR|UNDEF, optional => 1,},
                            },
                 # Needed for Fiji points that use params.
                 allow_extra => 1,
                 );
	
	my $self = {};
	$self->{_conn} = $args{conn};
        $self->{_debug} = $args{debug};
        $self->{_creator_pid} = $$;
        $self->{_tag} = $args{tag};
        $self->{_boot_mode} = $args{boot_mode};
        
        logcomment("Fiji: new: PID of the script creating this object: " . $self->{_creator_pid}) if ($self->{_debug});
	
	####################################################
	# Dependency on other internal modules
	####################################################
	
        # UI
        if ( defined($args{ui}) )
        {
                $self->{_ui} = $args{ui};
        }
        else
        {
                # Auto-detect version of the hosting filer. Need this for UI.
                logcomment("Fiji: UI object not provided. Auto-generating...") if ($self->{_debug});
                my $version = Compression::Version->new( conn => $args{conn}, debug => $self->{_debug} );
                $self->{_ui} = Compression::UI->new( version => $version, debug => $args{debug}, boot_mode => $self->{_boot_mode} );
        }
        
        my %create_args = (tag => $self->{_tag});
        
        if (defined($args{volname}))
        {
                $create_args{volname} = $args{volname};
        }
        
        if (defined($args{sleep_stage}))
        {
                # Get the release-dependent string
                my $sleep_stage = $self->{_ui}->{constants}->get_constant_value(constant_code => $args{sleep_stage});
                $create_args{sleep_stage} = $sleep_stage;
        }
        
        if (defined($args{sleep_time}))
        {
                $create_args{sleep_time} = $args{sleep_time};
        }
        
        if (defined($args{error_value}))
        {
                $create_args{error_value} = $args{error_value};
        }
        
        if (defined($args{inode_num}))
        {
                $create_args{inode_num} = $args{inode_num};
        }
        
        if (defined($args{level}))
        {
                $create_args{level} = $args{level};
        }
        
        if (defined($args{fbn}))
        {
                $create_args{fbn} = $args{fbn};
        }
        
        if ($self->{_boot_mode} eq C_ONTAP_CMODE) 
        {
                if (!($self->{_conn}->isa("NOTE::GX::Node")))
                {
                        logresult('script', "For C-mode, the conn parameter must be an object of type NOTE\:\:GX\:\:Node");
                }
                
                $self->{_host_name} = $self->{_conn}->name();
                # GNOTE wants the Filer arg to be a server obj
                $create_args{Filer} = $self->{_conn}->server();
                
                $self->{_fiji} = NOTE::GX::Fiji->new(%create_args);
        }
        else
        {
                if (!($self->{_conn}->isa("NOTE::Filer")))
                {
                        logresult('script', "For 7-mode, the conn parameter must be an object of type NOTE\:\:Filer");
                }
                
                $self->{_host_name} = $self->{_conn}->host();
                $create_args{Filer} = $self->{_conn};
                
                $self->{_fiji} = NOTE::Fiji->new(%create_args);
        }
        
        logcomment("Fiji: new: created Fiji object for filer " . $self->{_host_name} . ". Tag: " . $self->{_tag}) if ($self->{_debug});
	
	bless $self, $class;

	return $self;
}

## @name ::set_point
## @summary Sets the fiji point on the filer.
##
## @description
## Sets the fiji point on the filer.
##
## @return Hash with rc and output.
##
## @example
## $obj->set_point();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub set_point
{
        my $class = shift;
	my (%args) 	= validate( @_,
		{
			# No args
		} );
        
        my $output = $class->{_fiji}->set();
        
        return (rc => C_RC_SUCCESS, output => $output);
}

## @name ::nvset_point
## @summary Sets the fiji point in NVRAM on the filer.
##
## @description
## Sets the fiji point in NVRAM on the filer.
##
## @return Hash with rc and output.
##
## @example
## $obj->nvset_point();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub nvset_point
{
        my $class = shift;
	my (%args) 	= validate( @_,
		{
			# No args
		} );
        
        my $output = $class->{_fiji}->nvset();
        
        return (rc => C_RC_SUCCESS, output => $output);
}


## @name ::clear_point
## @summary Clears the fiji point on the filer.
##
## @description
## Clears the fiji point on the filer.
##
## @return Hash with rc and output.
##
## @example
## $obj->clear_point();
##
## @notes
## Fiji points corresponding to C_FIJI_TAG_DISABLE_CP and C_FIJI_TAG_ENABLE_CP
## present a special behavior:
## 1.) They are each-other's oposite (C_FIJI_TAG_ENABLE_CP should be set to anihilate the effects of C_FIJI_TAG_DISABLE_CP).
## 2.) Clearing them does not mean that their effect is also terminated.
##    On the opposite, even if they are not cleared, their effect may be already terminated (if the 'oposite' Fiji point was set in the meantime).
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub clear_point
{
        my $class = shift;
	my (%args) 	= validate( @_,
		{
			# No args
		} );
        
        my $output = $class->{_fiji}->clear();
        
        return (rc => C_RC_SUCCESS, output => $output);
}


## @name ::get_slot
## @summary Gets the slot in the Fiji table corresponding to this point.
##
## @description
## Gets the slot in the Fiji table corresponding to this point.
##
## @return Hash with rc and output.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes that the fiji point should be active.
## Legal values: C_RC_SUCCESS, C_RC_FAILURE (fiji point not active), C_RC_FAILURE_NON_FATAL.
##
## @example
## my %res = $obj->get_slot();
## my $slot = $res{output};
##
## @notes
## Fiji points corresponding to C_FIJI_TAG_DISABLE_CP and C_FIJI_TAG_ENABLE_CP
## present a special behavior:
## 1.) They are each-other's oposite (C_FIJI_TAG_ENABLE_CP should be set to anihilate the effects of C_FIJI_TAG_DISABLE_CP).
## 2.) Clearing them does not mean that their effect is also terminated.
##    On the opposite, even if they are not cleared, their effect may be already terminated (if the 'oposite' Fiji point was set in the meantime).
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub get_slot
{
        my $class = shift;
	my (%args) 	= validate( @_,
		{
			expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
		} );
        
        my $rc = C_RC_SUCCESS;
        my $slot = $class->{_fiji}->slot();
        
        if (!defined($slot))
        {
                $rc = C_RC_FAILURE;
        }
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Check fiji point \"" . $class->{_tag} ."\"",
                                                    output      => $slot);
}


## @name ::check_if_active
## @summary Checks if the Fiji point is still active.
##
## @description
## Checks if the Fiji point is still active.
##
## @return Hash with rc and output.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes that the fiji point should be active.
## Legal values: C_RC_SUCCESS, C_RC_FAILURE (fiji point not active), C_RC_FAILURE_NON_FATAL.
##
## @example
## $obj->check_if_active( expected => C_RC_FAILURE );
##
## @notes
## Fiji points corresponding to C_FIJI_TAG_DISABLE_CP and C_FIJI_TAG_ENABLE_CP
## present a special behavior:
## 1.) They are each-other's oposite (C_FIJI_TAG_ENABLE_CP should be set to anihilate the effects of C_FIJI_TAG_DISABLE_CP).
## 2.) Clearing them does not mean that their effect is also terminated.
##    On the opposite, even if they are not cleared, their effect may be already terminated (if the 'oposite' Fiji point was set in the meantime).
## 3.) As a result of 2, check_if_active is not relevant to the effect of these two Fiji points.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub check_if_active
{
        my $class = shift;
	my (%args) 	= validate( @_,
		{
			expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
		} );
        
        my $status = $class->{_fiji}->active();
        my $rc = ($status == TRUE) ? C_RC_SUCCESS : C_RC_FAILURE;
        
        return $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
                                                    expected_rc => $args{expected},
                                                    log_message => "Check if fiji point \"" . $class->{_tag} ."\" is active",
                                                    output      => $status);
}


## @name ::wait_for_point
## @summary Waits until the Fiji point is no longer active.
##
## @description
## Waits until the Fiji point is no longer active.
##
## @return Hash with rc and output.
##
## @arg interval Optional: Polling interval.
##
## @example
## $obj->wait_for_point();
##
## @notes
## Fiji points corresponding to C_FIJI_TAG_DISABLE_CP and C_FIJI_TAG_ENABLE_CP present a special behavior:
## 1.) They are each-other's oposite (C_FIJI_TAG_ENABLE_CP should be set to anihilate the effects of C_FIJI_TAG_DISABLE_CP).
## 2.) Clearing them does not mean that their effect is also terminated.
##    On the opposite, even if they are not cleared, their effect may be already terminated (if the 'oposite' Fiji point was set in the meantime).
## 3.) As a result of 2, wait_for_point should not be used for these two Fiji points.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub wait_for_point
{
        my $class = shift;
	my (%args) 	= validate( @_,
		{
			interval => { type => SCALAR, optional => 1, },
		} );
        
        my $rc = $class->{_fiji}->wait(%args);
        
        return (rc => C_RC_SUCCESS, output => $rc);
}


## @name ::spare_point
## @summary Sets the Fiji point to not be cleared upon script exit.
##
## @description
## Sets the Fiji point to not be cleared upon script exit.
##
## @return Hash with rc and output.
##
## @example
## $obj->spare_point();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub spare_point
{
        my $class = shift;
	my (%args) 	= validate( @_,
		{
			# No args
		} );
        
        my $rc = $class->{_fiji}->SPARE();
        
        return (rc => C_RC_SUCCESS, output => $rc);
}


1;
