# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/FileClone.pm#1 $
#
## @summary
##  Utility wrapper module for file clone
##
## @description
##  Utility wrapper module for file clone.
##  Functional specs:
##   - RR: http://wikid.netapp.com/w/WAFL/WART/capacity/Constant_time_sis_clones_FS
##
## @synopsis
##
## use Compression::FileClone;
## use vars qw( $FILER  );
##
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $obj = Compression::FileClone->new(host => $FILER, volume => $vol_name, debug => TRUE,);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::FileClone;

BEGIN
{
	require "Compression/base_import.pl";

}

use strict;
use warnings;
use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Compression::Constants qw(:all);
use Compression::7Mode::FileClone;
use Compression::CMode::FileClone;




## @name:: new
## @summary Compression::FileClone Constructor.
##
## @description
## Creates an object representing FileClone utilities.
##
## @return Object of type Compression::FileClone.
##
## @arg filer Required: NOTE/GNOTE filer/node object.
##
## @arg volume Required: Name of the volume.
##
## @arg vserver Optional: Name of the vserver (only relevant in c-mode).
##
## @arg boot_mode Optional: Filer boot mode (7M/CM). Default: 7M.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @arg ui Required: Compression::UI object for this filer.
##
## @example
## my $obj = Compression::FileClone->new (filer => $conn,
##                                        volume => $vol_name,
##                                        debug => TRUE, );

sub new
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		filer => {
                        type => OBJECT,},

                volume => {
			type => SCALAR,},

		debug => {
			type => BOOLEAN,
                        optional => 1,
			default => FALSE,},
                
                vserver => {type => SCALAR|UNDEF,
                            optional => 1},
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default => C_ONTAP_BOOT_MODE},
                
                ui => {type => OBJECT,
		       isa => "Compression::UI",
		       },
	} );

	# Start building our object:
	my $self = {};
        $self->{_filer} = $args{filer};
	$self->{_boot_mode} = delete($args{boot_mode});
        
        if ($self->{_boot_mode} eq C_ONTAP_7MODE) 
        {
		delete($args{vserver});
		$self->{_file_clone} = Compression::7Mode::FileClone->new(%args);
        }
        elsif ($self->{_boot_mode} eq C_ONTAP_CMODE)
        {
		$self->{_file_clone} = Compression::CMode::FileClone->new(%args);
        }


	bless($self, $class);

	logcomment("FileClone: Setup Complete..\n\n") if ($self->{_debug});
	return $self;
}


## @name:: start
## @summary Starts a clone operation.
##
## @description
## Starts a clone operation.
##
## @arg source Required: Filer path of the source file (Ex: /vol/my_vol/my_file).
##
## @arg clone Optional: Filer path of the clone file (Ex: /vol/my_vol/my_clone).
## Either clone or blk_range must be specified.
##
## @arg blk_range Optional: Block range for file cloning (<src_fbn>:<dest_fbn>:<fbn_cnt> ...).
## Either clone or blk_range must be specified.
##
## @arg snapshot Optional: Snapshot name to be used for file cloning.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @return Hash with rc and output. On 7-mode only: if the command succeedes, the hash will also contain id
##         (the id of the file clone operation).
##
## @example
## my %res = $obj->start(source => $path, clone => $clone_path, snapshot => 'snap1');
## my $id = $res{id};
##
sub start
{
        my $class = shift;
        
        return $class->{_file_clone}->start(@_);
}


## @name:: status
## @summary Gets the status of running and failed clone operations for the volume.
##
## @description
## Gets the status of running and failed clone operations for the volume.
##
## @arg id Optional: Id of the file clone operation for this volume.
##
## @arg latest_started Optional: Flag to decide if the status should be retrieved for the latest-started file clone operation.
## Note that if this is TRUE, then any passed-in id argument is overwritten with the id of the latest-started file clone operation for this volume.
## Default: FALSE.
##
## @note
## Starting with RR, this command is a no-op.
##
## @example 
## my %status = $compress->{file_clone}->status();
## $compress->{file_clone}->status(latest_started => TRUE,
##                                 expected => C_RC_SCANNER_IDLE);
##
sub status
{
        my $class = shift;
        
        return $class->{_file_clone}->status(@_);
}


## @name:: clear
## @summary Clears information of a failed clone operation for the volume.
##
## @description
## Clears information of a failed clone operation for the volume.
##
## @arg id Optional: Id of the file clone operation for this volume.
## If not specified, defaults to the latest-started file clone operation.
##
## @return Hash with rc and output.
##
## @note
## Starting with RR, this command is a no-op.
##
## @example
## $obj->clear();
## $obj->clear(id => $my_id);
##
sub clear
{
        my $class = shift;
        
        return $class->{_file_clone}->clear(@_);
}


## @name:: stop
## @summary Stops a running clone operation.
##
## @description
## Stops a running clone operation.
##
## @arg id Optional: Id of the file clone operation for this volume.
## If not specified, defaults to the latest-started file clone operation.
##
## @return Hash with rc.
##
## @note
## Starting with RR, this command is a no-op.
##
## @example
## $obj->stop();
## $obj->stop(id => $my_id);
##
sub stop
{
        my $class = shift;
        
        return $class->{_file_clone}->stop(@_);
}


## @name:: wait_for_completion
## @summary Waits for a clone operation to complete.
##
## @description
## Waits for a clone operation to complete.
##
## @arg id Optional: Id of the file clone operation for this volume.
## If not specified, defaults to the latest-started file clone operation.
##
## @return Hash with rc.
##
## @example
## $obj->wait_for_completion();
## $obj->wait_for_completion(id => $my_id);
##
sub wait_for_completion
{
	my $class = shift;
        
        return $class->{_file_clone}->wait_for_completion(@_);
}


## @name:: is_bg_clone_in_progress
## @summary Checks if the background file clone is still in process.
##
## @description
## Checks if the background file clone is still in process.
##
## @arg inode_num Required: Inode number of the clone.
##
## @return Hash with rc.
##
## @example
## $obj->is_bg_clone_in_progress(inode_num => $inum, expected => C_RC_SCANNER_IDLE);
##
sub is_bg_clone_in_progress
{
	my $class = shift;
        
        return $class->{_file_clone}->is_bg_clone_in_progress(@_);
}


## @name:: wait_for_bg_clone
## @summary Waits for the background file clone to complete.
##
## @description
## Waits for the background file clone to complete.
##
## @arg inode_num Required: Inode number of the clone.
##
## @return Hash with rc.
##
## @example
## $obj->is_bg_clone_in_progress(inode_num => $inum, expected => C_RC_SCANNER_IDLE);
##
sub wait_for_bg_clone
{
	my $class = shift;
	my (%args) = validate( @_,
        {
                inode_num => {type => SCALAR,},
		timeout => {type => SCALAR, default => C_DFLT_TIMEOUT}, 
        } );
	
	my $vol = $class->{_file_clone}->{_volume};
	my $inum = $args{inode_num};
	my $timeout = $args{timeout};
	
	my $rc = C_RC_SUCCESS;
	my %res;
	
	while ($rc != C_RC_SCANNER_IDLE)
	{
		if ($timeout < 0)
		{
			logresult('fatal', "Timed-out while waiting for the background file clone to complete for volume $vol, inode $inum\n" .
				           Dumper(%res));
		}
		
		%res = $class->is_bg_clone_in_progress(inode_num => $inum,
						       expected => C_RC_FAILURE_NON_FATAL);
		$rc = $res{rc};
		sleep(5);
		$timeout-=5;
		logcomment("wait_for_bg_clone: timeout = $timeout...");
	}
}


sub testme
{

	package main;
        use Data::Dumper;
	
        require "drg/Compression/setup_core.pl";
	
        
        use vars qw($FILER $AGGREGATE $VOLUME $DEBUG);
        
        param("DEBUG", "-default", FALSE);
        
        my $compress = Compression::Compress->setup(filer => {name => $FILER ,
							      require_debug_build => 1,
							      },
                                                    aggregate => $AGGREGATE,
                                                    volume => {name => $VOLUME,
                                                               size => "7g",
                                                               compress_enabled => FALSE,
                                                               },
                                                    debug => $DEBUG,
						    file_clone => TRUE,
                                                    );
	
	my $ontap = $compress->{_ui}->{version}->{ontap};
	
	my $filename = 'file';
	my $clonename1 = 'clone1';
	my $clonename2 = 'clone2';
	my $clonename3 = 'clone3';
	
	# Only BR and after
	if ($ontap eq '8.0')
	{
		my $filesize = 3 * GB;
		
		$compress->{dataset}->create(files =>[{ name => $filename, size => $filesize,},],);
		$compress->wait_for_cp();
		$compress->{volume}->create_snapshot(name => 'snap1');
		$compress->wait_for_cp();
		
		# No snapshot, stop
		my $path = $compress->{dataset}->get_filer_path(filename => $filename,);
		my $clone1_path = $compress->{dataset}->generate_path_for_attempt_file_clone(file_name => $filename,
											     clone_name => $clonename1);
		
		$compress->{file_clone}->start(source => $path,
                               clone => $clone1_path,);
		$compress->{file_clone}->status(latest_started => TRUE);
		$compress->{file_clone}->stop();
		$compress->{file_clone}->status(latest_started => TRUE,
						expected => C_RC_SCANNER_IDLE);
		$compress->{file_clone}->clear();
		
		# Snapshot, no stop
		my $clone2_path = $compress->{dataset}->generate_path_for_attempt_file_clone(file_name => $filename,
											     clone_name => $clonename2);
		
		$compress->{file_clone}->start(source => $path,
					       clone => $clone2_path,
					       snapshot => 'snap1',);
		$compress->{file_clone}->status(latest_started => TRUE);
		$compress->{file_clone}->wait_for_completion();
		$compress->{file_clone}->status(latest_started => TRUE,
						expected => C_RC_SCANNER_IDLE);
	}
	else
	{
		my $filesize = KB;
		$compress->{dataset}->create(files =>[{ name => $filename, size => $filesize,},],);
		$compress->wait_for_cp();
		
		$compress->{volume}->create_snapshot(name => 'snap1');
		$compress->wait_for_cp();
		
		# No snapshot, no suspend
		my $path = $compress->{dataset}->get_filer_path(filename => $filename,);
		my $clone1_path = $compress->{dataset}->generate_path_for_attempt_file_clone(file_name => $filename,
											     clone_name => $clonename1);
		
		$compress->{file_clone}->start(source => $path,
					       clone => $clone1_path,);
		$compress->{dataset}->restore_file_clone_info(clone_name => $clonename1);
		$compress->{file_clone}->wait_for_completion();
		
		# No snapshot, suspend bg clone
		my $clone2_path = $compress->{dataset}->generate_path_for_attempt_file_clone(file_name => $filename,
											     clone_name => $clonename2);
		
		$compress->{filer}->set_flag(code => C_FLAG_WAFL_DISABLE_FILE_CLONE_BG_SHARING,
					     value => 1);
		$compress->wait_for_cp();
		$compress->{file_clone}->start(source => $path,
					       clone => $clone2_path,);
		$compress->{dataset}->restore_file_clone_info(clone_name => $clonename2);
		
		my $clone2_inode = $compress->{dataset}->get_file_inode_num(filename => $clonename2);
		$compress->{file_clone}->is_bg_clone_in_progress(inode_num => $clone2_inode);
		
		$compress->{filer}->set_flag(code => C_FLAG_WAFL_DISABLE_FILE_CLONE_BG_SHARING,
					     value => 0);
		
		$compress->{file_clone}->wait_for_bg_clone(inode_num => $clone2_inode);
		
		# Snapshot, suspend bg clone
		my $clone3_path = $compress->{dataset}->generate_path_for_attempt_file_clone(file_name => $filename,
											     clone_name => $clonename3);
		
		$compress->{filer}->set_flag(code => C_FLAG_WAFL_DISABLE_FILE_CLONE_BG_SHARING,
					     value => 1);
		$compress->wait_for_cp();
		$compress->{file_clone}->start(source => $path,
					       clone => $clone3_path,
					       snapshot => 'snap1',);
		$compress->{dataset}->restore_file_clone_info(clone_name => $clonename3);
		
		my $clone3_inode = $compress->{dataset}->get_file_inode_num(filename => $clonename3);
		$compress->{file_clone}->is_bg_clone_in_progress(inode_num => $clone3_inode);
		
		$compress->{filer}->set_flag(code => C_FLAG_WAFL_DISABLE_FILE_CLONE_BG_SHARING,
					     value => 0);
		
		$compress->{file_clone}->wait_for_bg_clone(inode_num => $clone3_inode);
	}
        
        logresult('PASS', "FileClone: testme: Done");
}




1;