# -*- perl -*-
# $Id: $
#
## @summary
##  Utility wrapper module for file folding
##
## @description
##  wafl file folding utilities. wafl scan fold (start, status, abort)
##  http://web.netapp.com/engineering/design-depot/wafl/filefolding.html
##
## @synopsis
##
## use Compression::FileFolding;
## use vars qw( $FILER  );
##
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $obj = Compression::FileFolding->new (host => $FILER,
##	 			            volume_path => $vol_mount_path,
##				            initialize => TRUE );
##
## @author dl-compression-qa@netapp.com, raval@netapp.com
## @status private (in-progress)

package Compression::FileFolding;

use strict;
use warnings;
use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult
                          _send_cmd_to_filer);


BEGIN
{
	require "Compression/base_import.pl";

}


## @name:: new
## @summary Compression::FileFolding Constructor.
##
## @description
## Creates an object representing FileFolding utilities.
##
## @return Object of type Compression::FileFolding.
##
## @arg filer Required: NOTE::Filer object.
##
## @arg volume Required: Volume name.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::FileFolding->new (filer => $conn,
##                                          volume => $self->{volume}->{volume},
##                                          debug => $args{debug} );

sub new
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		filer => {
                        type => OBJECT,},

                volume => {
			type => SCALAR,},

		debug => {
			type => BOOLEAN,
                        optional => 1,
			default => FALSE,},
                
                vserver => {type => SCALAR|UNDEF,
                            optional => 1},
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default => C_ONTAP_BOOT_MODE},
                
                ui => {type => OBJECT,
		       isa => "Compression::UI",
		       },
	} );

	# Start building our object:
	my $self = {};
        $self->{_conn} = $args{filer}->{_conn};
        $self->{_filer} = $args{filer};
        $self->{_volume} = $args{volume};
        $self->{_debug} = $args{debug};
        $self->{_creator_pid} = $$;
        $self->{_boot_mode} = $args{boot_mode};
        $self->{vserver} = $args{vserver};
        $self->{_ui} = $args{ui};
        
        if (($self->{_boot_mode} eq C_ONTAP_7MODE) && !($self->{_filer}->isa("NOTE::Filer")))
        {
                logresult('script', "For 7-mode, the filer argument is required to be a NOTE\:\:Filer object");
        }
        elsif (($self->{_boot_mode} eq C_ONTAP_CMODE) && !($self->{_filer}->isa("NOTE::GX::Node")))
        {
                logresult('script', "For 10-mode, the filer argument is required to be a NOTE\:\:GX\:\:Node object");
        }


	bless($self, $class);

	logcomment("FileFolding: Setup Complete..\n\n") if ($self->{_debug});
	return $self;
}


## @name:: start
## @summary initiates a file_fold on a single file
##
## @arg file_name Required: scalar - file_name
##
## @arg wait_for_completion Optional: scalar  ( TRUE|FALSE )
##      - TRUE, calls FS::Util::file_folding::do_fold()
##          - Waits till file folding finishes
##          - Doesn't check for active file folding scan but dies on error
##      - FALSE, calls FS::Util::file_folding::file_fold_start()
##          - Does not wait for file folding to complete
##          - Checks for active file folding scan otherwise dies
##          - Returns scan_id which can be used for ff_wafl_scan_abort()
##
## @return $scan_id (while wait_for_completion => FALSE) otherwise none
##
## @example
## $test_vol->{file_folding}->start(file_name => 'f1');
## $test_vol->{file_folding}->start(file_name => 'f1', wait_for_completion => 1);

sub start
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                file_name => { type => SCALAR },
                wait_for_completion => { type => SCALAR, optional => 1 },
        } );
        
        my $wait_for_completion = $args{wait_for_completion};
        my $filename = $args{file_name};
        my $volume = $class->{_volume};
        my $ff_scan_id;
        
        # This is not completely reliable, as it's not atomic,
        # but it's better than nothing
        # First, see if there are any file folding scans already running,
        # so we know which scanner id is ours
        my %pre_start_status = $class->_get_non_fatal_status();
        my $pre_start_scans = $pre_start_status{scan_ids};
        
        my $cmd_code = C_CMD_FILE_FOLDING_START;
        
        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $volume,
                                                       file_path => $filename);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_filer},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
        
        $class->{_ui}->process_output_string(code => $cmd_code,
                                             output => $output,
                                             expected_rc => C_RC_SUCCESS,
                                             log_message => "Start file folding for file $filename, volume $volume",);
        
        my %status = $class->_get_non_fatal_status();
        
        if ($wait_for_completion == TRUE)
        {
                my $timeout = C_DFLT_COMPRESS_TIMEOUT;
                while (($status{rc} == C_RC_SUCCESS) && ($timeout > 0))
                {
                        sleep(1);
                        $timeout--;
                        %status = $class->_get_non_fatal_status();
                }
                
                if ($status{rc} == C_RC_SUCCESS)
                {
                        logresult('fatal', "Timed-out while waiting for file folding for volume " . $class->{_volume});
                }
        }
        else
        {
                logcomment("File folding for file $filename, volume $volume was started.\n" .
                           "Double-checking that the scanner is active");
                my $active_scans = $status{scan_ids};
                
                foreach my $active_scan (@$active_scans)
                {
                        if (!grep(/^$active_scan$/, @$pre_start_scans))
                        {
                                if (defined($ff_scan_id))
                                {
                                        logresult('fatal', "More than one new file folding scan detected:\n" . Dumper($active_scans) .
                                                           "\n\nScans before starting file folding:\n" . Dumper($pre_start_scans));
                                }
                        }
                        
                        $ff_scan_id = $active_scan;
                }
                
                $status{scan_id} = $ff_scan_id;
        }
        
        $status{rc} = C_RC_SUCCESS;
        return %status;
}


## @name:: status
## @summary Returns the status of a scan
##
## @arg object Optional: vol / aggr - Name of the volume or aggregate whose scan
##       status is to be returned
##
## @return %scans: hash of hashs by Scan Id -> type[Volume|Aggregate], name, status and progress
##
## @example
## my %scans = $test_vol->{file_folding}->ff_wafl_scan_status();

sub status
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                # None
        } );
        
        my %status = $class->_get_non_fatal_status();
        
        if ($status{rc} != C_RC_SUCCESS)
        {
                logresult('fatal', "Could not detect active file folding scans\n" . Dumper(%status));
        }
        
        return %status;
}


## @name:: abort
## @summary Terminate a scan by id
##
## @arg scan_id: ID of the scan to be terminated
##
## @return: none
##
## @example
## $test_vol->{file_folding}->ff_wafl_scan_abort(scan_id => $scan_id);

sub abort
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                scan_id => {type => SCALAR},
        } );
        
        my $scan_id = $args{scan_id};
        
        my $cmd_code = C_CMD_FILE_FOLDING_ABORT;
        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       modifier => $scan_id);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_filer},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
        
        return $class->{_ui}->process_output_string(code => $cmd_code,
                                                    output => $output,
                                                    expected_rc => C_RC_SUCCESS,
                                                    log_message => "Abort file folding scan with id $scan_id",);
}


################################
# Internal methods
################################

sub _get_non_fatal_status
{
        my $class = shift;

        my $volume = $class->{_volume};
        my $ff_scan_type = $class->{_ui}->get_constant_value(constant_code => 'FILE_FOLDING_SCAN_TYPE');
        my $ff_scan_ids = [];

        my $cmd_code = C_CMD_FILE_FOLDING_STATUS;
        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $volume,);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_filer},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
        
        my %status = $class->{_ui}->process_output_string(code => $cmd_code,
                                                          output => $output,
                                                          expected_rc => C_RC_FAILURE_NON_FATAL,
                                                          log_message => "Check file folding status for volume $volume",);

        # Look for the scan id of file folding
        my $active_scans = (defined($status{wafl_scans})) ? $status{wafl_scans} : {};
        foreach my $scan_id (keys %$active_scans)
        {
                if ($active_scans->{$scan_id}->{scan_type} =~ /$ff_scan_type/)
                {
                        push(@$ff_scan_ids, $scan_id);
                }
        }
        
        $status{rc} = (scalar(@$ff_scan_ids) == 0) ? C_RC_SCANNER_IDLE : C_RC_SUCCESS;
        $status{scan_ids} = $ff_scan_ids;
        return %status;
}





1;