#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Filer.pm#1 $
#

## @summary Compression Filer utilities.
## 
## @description
## Compression Filer utilities.
## 
## @synopsis
##
## use Compression::Filer;
## use vars qw($FILER );
## 
## BEGIN { require "Compression/setup_core.pl"; }
## 
## my $obj = Compression::Filer->new( filer => $FILER );
## my $obj = Compression::Filer->new( filer => { name => $FILER, conn_type => '#rsh', stats => FALSE } );
##
## @examples 
##
## my $obj = Compression::Filer->new( filer => $FILER );
## $obj->license_add( service => C_LICENSE_COMPRESSION, expected => C_RC_NO_NEARSTORE_LICENSE );
## $obj->license_remove( service => C_LICENSE_COMPRESSION, expected => C_RC_COMPRESSION_NOT_TURNED_OFF );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)


package Compression::Filer;

use strict;
use warnings;

use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Tharn;
use filer;
use Compression::Filer::License;
use Compression::Fiji;
use Compression::Constants qw(:all);
use Compression::7Mode::Filer;
use Compression::CMode::Filer;
use NACL::C::SystemNode;
use Compression::Utils qw(_c_interpret_logresult
                          detect_indirect_type_from_status
                          compare_hash_entries);
;


BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
        
        # Keep a global hash to keep track of CP state per each filer.
        # This way, we do not depend on the number and usage of Compression objects created by a script.
        # Extra cleaning is needed if timer-triggered CP was disabled: only clearing the Fiji point is not enough.
        $Compression::Filer::CP_is_disabled = {};
}


## @name ::new
## @summary Compression::Filer Constructor.
##
## @description
## Creates an object representing a filer.
##
## @return Object of type Compression::Filer.
##
## @arg filer Required: Scalar (filer name) or hash of filer information.
## If filer is passed in as a hashref, then the allowed keys/values are:
## - name => name of the filer,
## - mode => mode to connect to the filer. Default: rsh,
## - stats => flag to enable recording/printing statistics. Default: FALSE,
## - license_ready => flag to enable installing the compression license. Default: C_DFLT_LICENSE_READY,
## - require_debug_build => Flag to enable checking for a debug build. If a debug build is required, but not found
##   then the test will be terminated as not applicable. Default: C_DFLT_REQUIRE_DEBUG_BUILD,
## - require_hw_platform => Flag to enable checking for hardware platform. If a simulator is found instead,
##   then the test will be terminated as not applicable. Default: C_DFLT_REQUIRE_HW_PLATFORM,
## - require_cf_disabled => Flag to be used to disable cf, if needed. By using this flag, no further action is needed within the script.
##   The library will take care of re-enabling cf at the end of the script.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::Filer->new( filer => $FILER, debug => TRUE );
## my $obj = Compression::Filer->new( filer => { name => $FILER, conn_type => '#rsh', stats => TRUE } );
## my $obj = Compression::Filer->new( filer => { name => $FILER, license_ready => FALSE, require_debug_build => TRUE, require_hw_platform => TRUE, } );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub new
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			filer 	=> {type => SCALAR|HASHREF,},
                        
                        vserver => {type => SCALAR|UNDEF,
                                    optional => 1},
                        
                        boot_mode => {type => SCALAR,
                                      optional => 1,
                                      default => C_ONTAP_BOOT_MODE},
                        
                        debug 	=> {type => SCALAR,
                                    optional => 1,
				    default => FALSE,
				    },
                        
                        required_features => {type => ARRAYREF|UNDEF,
                                              optional => 1,
                                              },
                        
                        count_volumes 	=> {type => SCALAR,
                                            optional => 1,
                                            default => FALSE,
                                            },
		} );
        
        my $self = {};
        $self->{_debug} = $args{debug};
        $self->{_creator_pid} = $$;
        $self->{_boot_mode} = $args{boot_mode};
        $self->{_count_volumes} = $args{count_volumes};
        delete($args{count_volumes});
        # Initializing _wafl_scan_speed_initial_val to 0.
        # This means that wafl scan speed was not changed during the test
        $self->{_wafl_scan_speed_initial_val} = 0; 
        
        logcomment("Filer: new called with arguments: " . Dumper(%args)) if ($self->{_debug});
        
        delete($args{boot_mode});
        if ($self->{_boot_mode} eq C_ONTAP_CMODE)
	{
                $self->{_c_filer} = Compression::CMode::Filer->new(%args);
	}
        else
        {
                delete($args{vserver});
                $self->{_c_filer} = Compression::7Mode::Filer->new(%args);
        }
        
        $self->{host_name} = $self->{_c_filer}->{host_name};
        $self->{_ui} = $self->{_c_filer}->{_ui};
        $self->{_has_debug_kernel} = $self->{_c_filer}->{_has_debug_kernel};
        
        bless $self, $class;
        
        ##############################################
	# Internal data: count volumes 'to start with'
	##############################################
        if ($self->{_count_volumes} == TRUE)
        {
                my %vol_count = $self->count_existent_volumes();
                
                $self->{_start_vol_count} = $vol_count{count};
                $self->{_start_vol_list} = $vol_count{output};
                logcomment("\n###################################################\n" .
                           "===================================================\n" .
                           "Start volume count: " . $self->{_start_vol_count} . "\n" .
                           "Start volume list: " . $self->{_start_vol_list} . "\n" .
                           "===================================================\n" .
                           "###################################################");
        }
        
	return $self;
}


## @name ::refresh_sw_version
## @summary Gets the sw version (ontap, wafl) version from the filer.
## 
## @description
## Gets the sw version (ontap, wafl) version from the filer.
##
## @return Hash with ontap_ver and wafl_ver.
## 
## @examples
## $obj->refresh_sw_version();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub refresh_sw_version
{
	my $class = shift;
	
	my %return =  $class->{_ui}->{version}->refresh_sw_version(conn => $class->_get_underlying_filer_obj());
	$class->{_ui}->{cli}->{messages}->_refresh_constants();
	$class->{_c_filer}->_set_debug_kernel();
	$class->{_has_debug_kernel} = $class->{_c_filer}->{_has_debug_kernel};
	
	return %return;
}


## @name ::get_ontap_version
## @summary Returns the ontap version string.
## 
## @description
## Returns the ontap version string.
##
## @return Ontap version string.
## 
## @examples
## my $ontap = $obj->get_ontap_version();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub get_ontap_version
{
	my $class = shift;
        
        return $class->{_c_filer}->get_ontap_version(@_);
}


## @name ::get_indirect_type_of_root_vol
## @summary Returns the indirect type of the root volume.
## 
## @description
## Returns the indirect type of the root volume.
##
## @return Indirect type of the root volume (32/64).
## 
## @examples
## my $root_type = $obj->get_indirect_type_of_root_vol();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub get_indirect_type_of_root_vol
{
	my $class = shift;
        
        return $class->{_c_filer}->get_indirect_type_of_root_vol(@_);
}


## @name ::license_add
## @summary Adds license for specified feature.
## 
## @description
## Adds license for specified feature.
##
## @return Hash with rc and output.
## 
## @arg service Required: Feature to be licensed.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @examples
##
## $obj->license_add( service => C_LICENSE_COMPRESSION );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub license_add
{
	my $class = shift;
        
        return $class->{_c_filer}->license_add(@_);
}

## @name ::license_remove
## @summary Removes license for specified feature.
## 
## @description
## Removes license for specified feature. Can handle compression-specific failures.
##
## @return Hash with rc and output.
## 
## @arg service Required: Feature to be unlicensed.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @examples
##
## $obj->license_remove( service => C_LICENSE_COMPRESSION, expected => C_RC_COMPRESSION_NOT_TURNED_OFF );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub license_remove
{
	my $class = shift;
        
        return $class->{_c_filer}->license_remove(@_);
}


## @name ::license_verify
## @summary Verifies if filer is licensed for specified feature.
## 
## @description
## Verifies if filer is licensed for specified feature. 
##
## @return Hash with rc and output.
## 
## @arg service Required: Feature to be verified.
## 
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
## 
## @arg site_license Optional: Boolean flag to enable extra checking for site license.
## Defaults to FALSE.
##
## @examples
##
## $obj->license_verify( service => C_LICENSE_COMPRESSION, expected => C_RC_SUCCESS );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub license_verify
{
	my $class = shift;
        
        return $class->{_c_filer}->license_verify(@_);
}


## @name ::set_option_on_all_volumes
## @summary Sets an option to all the volumes existent on the filer.
## 
## @description
## Retrieves the list of volumes existent on the filer and sets the desired option for all volumes.
##
## @return Hash with rc and output.
## Output is a hashref with key (= volume name) => value (= new option setting).
## 
## @arg service Required: Feature that corresponds to desired option (defined in Compression::Constants).
## 
## @arg new_value Required: Tag that corresponds to new value to be set.
## For now, only TRUE/FALSE are handled, corresponding to enable/disable.
##
## @arg skip_root_volume Optional: Boolean flag to decide if the option should be attempted to be set for the root volume.
## Defaults to TRUE.
##
## @arg skip_offline_restricted Optional: Boolean flag to decide if the option should be attempted to be set for offline/restricted volumes.
## Defaults to TRUE.
##
## @arg check_if_already_set Optional: Flag to enable the library to first check the current setting.
## If the current setting matches the value to be set, the subroutine will not attempt to set it again.
## Defaults to TRUE.
##
## @notes
## Currently it does not handle compression-specific failures.
##
## @examples
## $obj->set_option_on_all_volumes( service => C_OPTION_COMPRESSION, new_value => FALSE, skip_root_volume => TRUE );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub set_option_on_all_volumes
{
        my $class = shift;
        
        return $class->{_c_filer}->set_option_on_all_volumes(@_);
}


## @name ::volume_exists
## @summary Verifies that the specified volume exists on this filer.
## 
## @description
## Verifies that if the specified volume exists on this filer.
##
## @return Hash with rc and output (list of existent volumes).
## 
## @arg volume Required: Name of the volume.
## 
## @arg expected Optional: Expected result (C_RC_SUCCESS/C_RC_FAILURE).
## If not provided, assumes that the volume should exist.
## C_RC_FAILURE_NON_FATAL can also be passed in as 'expected', in which case,
## the script will not be terminated upon any result.
## Instead, the result (C_RC_SUCCESS/C_RC_FAILURE) will be passed to the script for further processing.
##
## @examples
##
## $obj->volume_exists( volume => $vol, expected => C_RC_FAILURE );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub volume_exists
{
	my $class = shift;
        
        return $class->{_c_filer}->volume_exists(@_);
}


## @name ::get_status_of_all_volumes
## @summary Gets the status of all volumes existent on this filer.
## 
## @description
## Gets the status of all volumes existent on this filer.
##
## @return Hash with rc and status (which is a hash with key<->values: vol_name<->hash (parent, options, status, type, state)) 
##
## @examples
##
## $obj->get_status_of_all_volumes();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub get_status_of_all_volumes
{
	my $class = shift;
        
        return $class->{_c_filer}->get_status_of_all_volumes(@_);
}


## @name ::count_existent_volumes
## @summary Obtains the number of volumes existent on this filer.
## 
## @description
## Obtains the number of volumes existent on this filer.
##
## @return Hash with rc, count, and output (", "-separated list of volumes).
## 
## @arg state Optional: Volume state, to be used to filter the existent volumes.
## If specified, then the subroutine will return the number of volumes in the desired state.
## If not specified, then the count of all volumes is returned.
## Volume states are defined in Compression::Constants.
##
## @examples
##
## $obj->count_existent_volumes( state => C_VOL_STATE_ONLINE );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub count_existent_volumes
{
	my $class = shift;
        
        return $class->{_c_filer}->count_existent_volumes(@_);
}


## @name ::reboot_and_wait
## @summary Reboots filer and waits until it comes up.
## 
## @description
## Reboots filer and waits until it comes up.
##
## @return Hash with rc and output (return of the NOTE::Filer::reboot).
## 
## @arg timeout Optional: Time in seconds to wait for the filer to come up.
## Defaults to C_DFLT_BOOTUP_WAIT_TIME (defined in Compression::Constants).
##
## @examples
##
## $obj->reboot_and_wait( timeout => 500 );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub reboot_and_wait
{
	my $class = shift;
        
        return $class->{_c_filer}->reboot_and_wait(@_);
}


## @name ::powercycle_and_wait
## @summary Powercycles filer and waits until it comes up.
## 
## @description
## Powercycles filer and waits until it comes up.
##
## @return Hash with rc and output.
## 
## @arg new_state Optional: State to bring the filer in after powercycling (described in boot.pm).
## Defaults to 'UP'.
## 
## @arg timeout Optional: Time in seconds to wait for the filer to come up.
## Defaults to C_DFLT_BOOTUP_WAIT_TIME (defined in Compression::Constants).
##
## @examples
##
## $obj->powercycle_and_wait( new_state => 'BOOT_MENU', timeout => 500 );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub powercycle_and_wait
{
	my $class = shift;
        
        return $class->{_c_filer}->powercycle_and_wait(@_);
}

## @name ::panic_and_wait
## @summary Panics filer and waits until it comes up.
## 
## @description
## Panics filer and waits until it comes up.
##
## @return Hash with rc and output.
## 
## @arg new_state Optional: State to bring the filer in after panic (described in boot.pm).
## Defaults to 'UP'.
## 
## @arg timeout Optional: Time in seconds to wait for the filer to be brought in desired state.
## This time is an extra to the 320 sec of sleep after the panic (filer::filer_panic.pm).
## Defaults to C_DFLT_BOOTUP_WAIT_TIME (defined in Compression::Constants).
##
## @notes
## The normal behavior for a filer is to come back 'UP' after panic,
## without the library having to take any action. Specifying a 'new_state' other than 'UP'
## means that the filer's state succession will be panic->up->new_state.
##
## @examples
##
## $obj->panic_and_wait( new_state => 'BOOT_MENU', timeout => 500 );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub panic_and_wait
{
	my $class = shift;
        
        return $class->{_c_filer}->panic_and_wait(@_);
}

## @name ::wait_until_up
## @summary Waits until the filer comes up.
## 
## @description
## Waits until the filer comes up. Wrapper around NOTE::Filer::transit_to.
##
## @return Hash with rc and output.
## 
## @arg timeout Optional: Time in seconds to wait for the filer to be brought in desired state.
## This time is an extra to the 320 sec of sleep after the panic (filer::filer_panic).
## If not defined, boot::transit_to will use its own default value.
##
## @arg powercycle Optional: Boolean flag to decide if the filer should be power-cycled in case of failure
## (if transit_to fails once, then power-cycle the filer and try once more to bring it to desired state).
## This is useful, as sometimes transit_to fails to detect (without retrying) that the filer came up.
## Defaults to FALSE.
##
## @arg min_duration Optional: Minimum duration imposed to the "bring up" process.
## Useful if we want to ensure that the filer actually was down (for example, it panicked).
## If set to a value greater than 0, the test will fail if the filer was in an 'UP' state in less than min_duration.
## Defaults to 0.
##
## @examples
## $obj->wait_until_up( min_duration => 120 );
## $obj->wait_until_up( timeout => 500, powercycle => FALSE );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub wait_until_up
{
	my $class = shift;
        
        return $class->{_c_filer}->wait_until_up(@_);
}


## @name ::set_fiji_point
## @summary Sets Fiji point on the filer.
## 
## @description
## Sets Fiji point on the filer.
##
## @return The Compression::Fiji object corresponding to this Fiji point.
## 
## @arg tag Required: Tag corresponding to the fiji point to be set.
## Fiji point tags are defined in Compression::Constants.
##
## @arg volname Optional: Name of volume, to be used for FIJI points enabled for a specific volume.
##
## @arg sleep_stage Optional: Sleep stage, to be used for the dense.sleep fiji points.
##
## @arg sleep_time Optional: Sleep time, to be used for the dense.sleep fiji points.
## Default: C_DFLT_FIJI_DENSE_SLEEP_TIME.
##
## @arg inode_num Optional: Inode num, to be used for the C_FIJI_TAG_SUSPEND_SCAN_64BIT_UPGRADE fiji point
## (required for this specific fiji).
##
## @arg level Optional: Error value, to be used for the C_FIJI_TAG_SUSPEND_SCAN_64BIT_UPGRADE fiji point
## (defaults to 0).
##
## @arg fbn Optional: Error value, to be used for the C_FIJI_TAG_SUSPEND_SCAN_64BIT_UPGRADE fiji point
## (required for this specific fiji and should be multiple of 510).
## 
## @examples
##
## my $fiji = $obj->set_fiji_point(tag => C_FIJI_TAG_DISABLE_CP);
## $fiji->get_slot();
## $fiji->check_if_active();
##
## @note
## This module will also keep track of any C_FIJI_TAG_DISABLE_CP/C_FIJI_TAG_ENABLE_CP that were set on the filer.
## This is to enable cleanup subroutine to re-enable the timer-triggered CP, if needed.
##
## @note
## This subroutine creates a new Compression::Fiji object and sets the Fiji point on the filer
## (calls Compression::Fiji->new and Compression::Fiji->set_point).
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub set_fiji_point
{
	my $class = shift;
        
        return $class->{_c_filer}->set_fiji_point(@_);
}


## @name ::sync_blocks_to_disk
## @summary Syncs dirty FS blocks to disk.
## 
## @description
## Syncs dirty FS blocks to disk. Wrapper around NOTE::Filer::sync.
##
## @return Hash with rc and output (void).
## 
## @arg wait Optional: Wait for sync to complete. Defaults to TRUE.
##
## @examples
##
## $obj->sync_blocks_to_disk();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub sync_blocks_to_disk
{
	my $class = shift;
        
        return $class->{_c_filer}->sync_blocks_to_disk(@_);
}


## @name ::reconnect
## @summary Reconnects to the filer with specified connection method.
## 
## @description
## Reconnects to the filer with specified connection method.
##
## @return Hash with rc and output (void).
## 
## @arg mode Optional: The connection method to use.
## Defaults to C_DFLT_FILER_CONN (defined in Compression::Constants).
##
## @examples
##
## $obj->reconnect( mode => '#console' );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub reconnect
{
	my $class = shift;
        
        return $class->{_c_filer}->reconnect(@_);
}


## @name ::wait_for_console_message
## @summary Tails the messages file until the required messages are seen.
## 
## @description
## Tails the messages file until the required messages are seen.
##
## @arg message_codes Required: Arrayref of codes corresponding to the
## message strings to be serched for in the messages file (defined in Constants.pm).
##
## @arg timeout Optional: Maximum allowed time, in seconds, to wait for match.
## Default: C_DFLT_CONSOLE_WAIT_TIMEOUT (defined in Constants.pm).
##
## @arg match_type Optional: Type of match to be performed. Legal values: ANY, ALL.
## Default: ALL.
##
## @return Hash with rc and output.
##  
## @examples
## $obj->wait_for_console_message(message_codes => [C_CONSOLE_OUT_OF_SPACE_SFSR_ABORTED,
##                                                  C_CONSOLE_INSUFFICIENT_SPACE_FOR_SFSR],
##                                match_type => 'ANY');
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub wait_for_console_message
{
        my $class = shift;
        
        return $class->{_c_filer}->wait_for_console_message(@_);
}

## @name ::get_root_vol_name
## @summary obtains the name of root volume
##
## @description
## Obtains the root volume name
## 
## @return root volume name
##
## @examples
##
## $root_vol_name = $obj->get_root_vol_name();
##
## @author dl-compression-qa@netapp.com
## 
sub get_root_vol_name
{
        my $class = shift;

        return $class->{_c_filer}->get_root_vol_name(@_);
}

## @name ::get_root_aggr_name
## @summary obtains the name of root aggregate
##
## @description
## Obtains the root aggregate name
##
## @return root aggregate name
##
## @examples
##
## $root_aggr_name = $obj->get_root_aggr_name();
##
## @author dl-compression-qa@netapp.com
##
sub get_root_aggr_name
{
        my $class = shift;
        
        return $class->{_c_filer}->get_root_aggr_name(@_);
}

## @name ::set_flag
## @summary Sets the required flag.
##
## @description
## Sets the required flag.
##
## @arg code Required: Code corresponding to the
## flag to be set (defined in Compression::Constants).
##
## @arg value Required: New value for the
## flag to be set.
##
## @arg privilege Required: Privilege that this flag requires.
## Default: C_PRIV_DIAG (defined in Compression::Constants).
##
## @return Hash with rc.
##
## @examples
##
## $obj->set_flag(code => C_FLAG_WAFL_INCONSISTENT_PANIC_TESTING,
##                value => 0);
## $obj->set_flag(code => C_FLAG_WAFL_DECOMPRESS_FAIL_TESTING,
##                value => 1);
##
##
## @author dl-compression-qa@netapp.com
##
sub set_flag
{
        my $class = shift;
        
        return $class->{_c_filer}->set_flag(@_);
}


## @name ::initialize_compliance_clock
## @summary Initializes the filer's compliance clock.
##
## @description
## Initializes the filer's compliance clock.
##
## @return Hash with rc.
##
## @examples
## $obj->initialize_compliance_clock();
##
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub initialize_compliance_clock
{
        my $class = shift;
        
        return $class->{_c_filer}->initialize_compliance_clock(@_);
}


## @name ::get_system_time
## @summary Gets the time on this filer. 
##
## @description
## Gets the time on this filer. 
##
## @return Hash with rc.
##
## @examples
## $obj->get_system_time();
##
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_system_time
{
        my $class = shift;
        
        return $class->{_c_filer}->get_system_time(@_);
}


## @name ::set_wafl_scan_speed
## @summary Sets the wafl scan speed. 
##
## @description
## Sets the wafl scan speed.
##
## @arg new_value Optional: New value to set.
## Defaults to C_DFLT_WAFL_SCAN_SPEED.
##
## @return Hash with rc, output.
##
## @examples
## $obj->wafl_scan_speed(new_value => 1000);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub set_wafl_scan_speed
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        new_value => { type => SCALAR, optional => 1, regex => qr/^\d+$/, default => C_DFLT_WAFL_SCAN_SPEED, },
                } );
        
        my $new_value = $args{new_value};
        my $cmd_code = C_CMD_SET_WAFL_SCAN_SPEED;
        
        # We need to get the current value first, to change back to at cleanup
        my %temp_data = $class->get_wafl_scan_speed();
        my $current_val = $temp_data{wafl_scan_speed}->{current_value};
        
        if ($current_val == $new_value)
        {
                logcomment("No change made wo wafl scan speed: current value is $current_val");
                return %temp_data;
        }
        
        logcomment("Setting wafl scan speed to $new_value");
        
        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       modifier => $new_value,);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_c_filer}->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
        
        %temp_data =  $class->{_ui}->process_output_string(code => $cmd_code,
                                                           output => $output,
                                                           expected_rc => C_RC_SUCCESS,
                                                           log_message => "Set wafl scan speed to $new_value on filer " . $class->{host_name},);
        
        # If we got here, it means the command executed successfuly
        # Remember what was the initial value of the scan speed, to reset it at cleanup
        # Only set it if it was not already set.
        # This is to cover the case when the speed is changed at least twice during the execution.
        if ($class->{_wafl_scan_speed_initial_val} == 0)
        {
                $class->{_wafl_scan_speed_initial_val} = $current_val;
        }
        
        return %temp_data;
}


## @name ::get_wafl_scan_speed
## @summary Gets the current value of wafl scan speed. 
##
## @description
## Gets the current value of wafl scan speed. 
##
## @return Hash with rc, output, wafl_scan_speed.
##
## @examples
## my %current_setting = $obj->wafl_scan_speed();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_wafl_scan_speed
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        # None
                } );
        
        my $cmd_code = C_CMD_GET_WAFL_SCAN_SPEED;
        
        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,);
        
        my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_c_filer}->{_conn},
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
        
        return $class->{_ui}->process_output_string(code => $cmd_code,
                                                    output => $output,
                                                    expected_rc => C_RC_SUCCESS,
                                                    log_message => "Get wafl scan speed setting on filer " . $class->{host_name},);
}


## @name ::get_wafl_stats
## @summary Get wafl stats.
## 
## @description
## Get wafl stats.
##
## @return Hash with rc and wafl_stats.
##
## @arg expected_values Optional: Hashref of key-values of expected stats.
## If this parameter is used, then all the required fields are checked.
## It retries in case of mismatch (up to 5 attempts).
## Available fields: cgs_resolved_in_KC, cgs_resolved_in_Bento, truncate_suspends,
## precp_overwritten, flippity_suspends, flippity_skipped, decompress_nomem', 
## buffers_cacheable_in_EC, non_stripe_writes, partial_cg_throttled.
## 
## @examples
## my %res = $compress->{filer}->get_wafl_stats(); # no checks done
## $compress->{filer}->get_wafl_stats(expected_values => {truncate_suspends => 0,
##                                                        flippity_skipped => 0,},);
##
## @note Because of the size of the output, currently only the compression stats are parsed and returned.
##
sub get_wafl_stats
{
	my $class = shift;
	my %args = validate( @_,
	{
		expected_values => {type => HASHREF, optional => 1,},
	});
	
	my $cmd_code = C_CMD_GET_WAFL_STATS;
	my $expected_values = $args{expected_values};
	my $filer_name = $class->{host_name};
	
	my $command = $class->{_ui}->formulate_command(code => $cmd_code,);
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_c_filer}->{_conn},
							    boot_mode => $class->{_boot_mode},
							    cmd => $command,
							    node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
	my %current = $class->{_ui}->process_output_string(code => $cmd_code,
							   output => $output,
							   expected_rc => C_RC_SUCCESS,
							   log_message => "Get wafl stats for filer $filer_name",);
	
	if (defined($expected_values))
	{
		for (my $i = 1; $i <= 5; $i++)
		{
			logcomment("Checking wafl stats values: attempt $i");
			my $fatal = ($i == 5) ? TRUE : FALSE;
			my $match = Compression::Utils::compare_hash_entries(obtained => $current{wafl_stats},
									     expected => $expected_values,
									     values_meaning => "wafl stats for filer $filer_name",
									     fatal_failure => $fatal,);
			if ($match == TRUE)
			{
				last;
			}
			
			sleep(ONE_CP_TIME);
			$output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_c_filer}->{_conn},
									 boot_mode => $class->{_boot_mode},
									 cmd => $command,
									 node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
			%current = $class->{_ui}->process_output_string(code => $cmd_code,
									output => $output,
									expected_rc => C_RC_SUCCESS,
									log_message => "Get wafl stats for filer $filer_name",);
		}
	}
	
	return %current;
}


################################
# Internal subroutines
################################


sub _get_underlying_filer_obj
{
        my $class = shift;
        return $class->{_c_filer}->{_conn};
}


# Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests start to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
        
        my $host_name = $class->{host_name};
        if ($class->{_creator_pid} == $$)
	{
                eval
                {
                        my $wafl_scan_speed_initial_val = $class->{_wafl_scan_speed_initial_val};
                        if ($wafl_scan_speed_initial_val != 0)
                        {
                                logcomment("Filer: $host_name: cleanup: wafl scan speed was changed during execution. Resetting...");
                                $class->set_wafl_scan_speed(new_value => $wafl_scan_speed_initial_val);
                        }
                        
                        if ($class->{_count_volumes} == TRUE)
                        {
                                my %vol_count = $class->count_existent_volumes();
                                my $end_vol_count = $vol_count{count};
                                my $start_vol_count = $class->{_start_vol_count};
                                
                                my $vol_status_msg = "\n###################################################\n" .
                                                     "===================================================\n" .
                                                     "End volume count: $end_vol_count\n" .
                                                     "End volume list: " . $vol_count{output} . "\n";
                                
                                if ($end_vol_count != $start_vol_count)
                                {
                                        $vol_status_msg .= "WARNING: start/end volume count mismatch:\n" .
                                                           "Start volume count: $start_vol_count\n" .
                                                           "Start volume list: " . $class->{_start_vol_list} . "\n";
                                }
                                
                                $vol_status_msg .= "===================================================\n" .
                                                   "###################################################";
                                                   
                                logcomment($vol_status_msg);
                        }
                };
                
                $class->{_c_filer} = undef;
        }

        if (defined($execution_error))
        {
                _c_interpret_logresult($execution_error);
        }
}


sub testme
{
	package main;
        
	use Compression::Constants qw(:all);
        use Compression::Filer;
        
        use vars qw($FILER $DEBUG $REBOOT);
        
        param("DEBUG", "-default", FALSE);
        param("REBOOT", "-default", FALSE);
        
        
        my %res;
        my $string_res;
        my $obj = Compression::Filer->new(filer => $FILER,
                                          debug => $DEBUG);
        
        $string_res = $obj->get_ontap_version();
        logcomment("got ontap: $string_res");
        
        $string_res = $obj->get_indirect_type_of_root_vol();
        logcomment("got indirect type of root vol: $string_res");
        
        %res = $obj->license_add(service => C_LICENSE_CIFS);
        logcomment("license add returned:\n" . Dumper(%res));
        
        %res = $obj->license_verify(service => C_LICENSE_CIFS);
        logcomment("license verify returned:\n" . Dumper(%res));
        
        %res = $obj->license_remove(service => C_LICENSE_CIFS);
        logcomment("license remove returned:\n" . Dumper(%res));
        
        %res = $obj->license_verify(service => C_LICENSE_CIFS,
                                    expected => C_RC_FAILURE);
        logcomment("license verify returned:\n" . Dumper(%res));
        
        %res = $obj->volume_exists(volume => 'bogus_vol',
                                   expected => C_RC_FAILURE);
        logcomment("volume_exists returned:\n" . Dumper(%res));
        
        %res = $obj->get_status_of_all_volumes();
        logcomment("get_status_of_all_volumes returned:\n" . Dumper(%res));
        
        %res = $obj->count_existent_volumes();
        logcomment("count_existent_volumes returned:\n" . Dumper(%res));
        
        if ($REBOOT == TRUE)
        {
                %res = $obj->reboot_and_wait();
                logcomment("reboot_and_wait returned:\n" . Dumper(%res));
        }
        
        $string_res = $obj->get_root_aggr_name();
        logcomment("get_root_aggr_name returned: $string_res");
        
        $string_res = $obj->get_root_vol_name();
        logcomment("get_root_vol_name returned: $string_res");
                
        logresult('PASS', "Filer: testme: complete");
}


sub overlap_requests
{
	my $class = shift;
	my (%args) = validate( @_,
		{
			option 	=> {type => SCALAR}
		});
	my $node = NACL::C::SystemNode->new( name => $class->{_c_filer}->{host_name} );
	my $Api_Set_Obj = $node->get_7m_or_nodescope_apiset;
	$Api_Set_Obj->execute_command(command => "options sis.overlap_requests ".$args{option});
	return;
}

sub minutely_schedule
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        option   => {type => SCALAR}
                });
        my $node = NACL::C::SystemNode->new( 
				name => $class->{_c_filer}->{host_name} );
        my $Api_Set_Obj = $node->get_7m_or_nodescope_apiset;
        $Api_Set_Obj->execute_command(
				command => "options sis.minutely_schedule ".$args{option});
        return;
}

1;
