#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Filer/PerfStats.pm#1 $
#

## @summary
## Utility module for wafl statistics.
##
## @description
## Utility module for wafl statistics.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 

package Compression::Filer::PerfStats;


use strict;
use warnings;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Tharn;
use Perf::Stats;
use Compression::Constants qw(:all);
use Compression::Utils qw(compare_values);

BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}


## @name ::new
##
## @summary Compression::Filer::PerfStats Constructor.
##
## @return Object of type Compression::Filer::PerfStats.
##
## @arg filer Required: Name of the filer (or c-mode node) to connect to.
##
## @arg ui Required: Compression::UI object for this filer.
##
## @arg stats_type Optional: Type of stats to track (wafl_stats, stats, cpstats).
## Defaults to wafl_stats.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##

sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
                filer => {type => SCALAR,
                          },
                
                stats_type => {type => SCALAR,
                               regex => qr/^wafl_stats|stats|cpstats$/,
                               default => 'wafl_stats',
                               },
                
                ui => {type => OBJECT,
                       isa => "Compression::UI",
                       },

                debug => {type => SCALAR,
                          optional => 1,
                          default  => FALSE,},      
	} );

	my $self = {};
        $self->{_debug} = $args{debug};
        $self->{_node_name} = $args{filer};
        $self->{_stats_type} = $args{stats_type};
        $self->{_ui} = $args{ui};
        $self->{_tracked_counters} = [];
        
        $self->{_stats_obj} = Perf::Stats->new(node => $self->{_node_name},
                                               debug => $self->{_debug},
                                               source => $args{stats_type},);
        
        
	bless $self, $class;
	return $self;
}


## @name ::add_counters_to_be_tracked
## @summary Adds the specified counters to the list of 'tracked counters'.
## 
## @description
## Adds the specified counters to the list of 'tracked counters'.
##
## @arg counters Required: Arrayref of codes corresponding to the counters to be added
## (defined in Compression::Constants).
##
## @arg stats_type Optional: Type of stats to track (wafl_stats, stats, cpstats).
## Defaults to the stats type for which the object was first created.
##
## @return Hash with rc.
##
## @examples
##
## $obj->add_counters_to_be_tracked(counters => [C_PERF_STATS_COUNTER_BUFFERS_CACHEABLE_IN_EC]);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##
sub add_counters_to_be_tracked
{
	my $class = shift;
        my (%args) = validate( @_,
		{
                        counters => {type => ARRAYREF,},
                        stats_type => {type => SCALAR, default => $class->{_stats_type},},
                } );
        
        my $counter_definitions = $args{counters};
        my $stats_type = $args{stats_type};
        my $counters_to_add = [];
        my $class_tracked_counters = $class->{_tracked_counters};
        
        # TODO: test c-mode
        foreach my $counter_def (@$counter_definitions)
        {
                my $counter_str = $class->{_ui}->get_perf_stats_counter(code => $counter_def);
                push(@$counters_to_add, $counter_str);
                if ($stats_type eq $class->{_stats_type})
                {
                        push(@$class_tracked_counters, $counter_str);
                }
        }
        
        $class->{_stats_obj}->add($counters_to_add, source => $stats_type);
        
        return (rc => C_RC_SUCCESS,);
}


## @name ::collect_stats
## @summary Collects statistics.
## 
## @description
## Collects statistics.
##
## @return Hash with rc and perf_stats.
##
## @examples
##
## $obj->collect_stats();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##
sub collect_stats
{
	my $class = shift;
        my (%args) = validate( @_,
		{
                        # None
                } );
        
        # TODO: test c-mode 
        my $sample_stats = $class->{_stats_obj}->sample();
        
        return (rc => C_RC_SUCCESS, perf_stats => $sample_stats);
}


## @name ::verify_stats
## @summary Collects statistics and checks the sampled values against the expected values.
## 
## @description
## Collects statistics and checks the sampled values against the expected values.
##
## @arg expected_values Required: Hashref of 'counter code'/'expected value' pairs to be checked.
##
## @return Hash with rc and perf_stats.
##
## @examples
##
## $obj->verify_stats(expected_values => {C_PERF_STATS_COUNTER_BUFFERS_CACHEABLE_IN_EC => 0});
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##
sub verify_stats
{
	my $class = shift;
        my (%args) = validate( @_,
		{
                        expected_values => {type => HASHREF},
                } );
        
        my $expected_values = $args{expected_values};
        my $stats_type = $class->{_stats_type};
        my $counters_to_check = {};
        
        # TODO: test c-mode 
        my $sampled_stats = $class->{_stats_obj}->sample();
        
        foreach my $counter (keys(%$expected_values))
        {
                my $counter_str = $class->{_ui}->get_perf_stats_counter(code => $counter);
                $counters_to_check->{$counter_str} = $expected_values->{$counter};
        }
        
        foreach my $counter_to_check (keys(%$counters_to_check))
        {
                my $expected_value = $counters_to_check->{$counter_to_check};
                my $obtained_value = $sampled_stats->{$counter_to_check};
                if (!defined($obtained_value))
                {
                        logresult('fatal', "Failed to detect value for \"$counter_to_check\"");
                }
                
                logcomment("Checking $stats_type value for \"$counter_to_check\"\n" .
                           "Expected value: $expected_value\n" .
                           "Obtained value: $obtained_value");
                
                if ($obtained_value != $expected_value)
                {
                        logresult('fatal', "$stats_type value mismatch for \"$counter_to_check\":\n" .
                                           "Obtained $obtained_value, expected $expected_value:\n" . Dumper($sampled_stats));
                }
        }
        
        return (rc => C_RC_SUCCESS, perf_stats => $sampled_stats);
}


## @name ::store_checkpoint
## @summary Collects statistics and stores them to be used for calculate_delta/verify_delta.
## 
## @description
## Collects statistics and stores them to be used for calculate_delta/verify_delta.
##
## @return Hash with rc.
##
## @examples
##
## $obj->store_checkpoint();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##
sub store_checkpoint
{
	my $class = shift;
        my (%args) = validate( @_,
		{
                        # None
                } );
        
        # TODO: test c-mode 
        $class->{_stats_obj}->checkpoint();
        
        return (rc => C_RC_SUCCESS,);
}


## @name ::calculate_delta
## @summary Returns delta from the last sample stat saved by store_checkpoint.
## 
## @description
## Returns delta from the last sample stat saved by store_checkpoint.
##
## @return Hash with rc and delta.
##
## @examples
##
## my %res = $obj->calculate_delta();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##
sub calculate_delta
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        # None
                } );
        
        # TODO: need to support passed-in samples?
        my $delta = $class->{_stats_obj}->delta(hide_zeros => 0);
        
        return (rc => C_RC_SUCCESS, delta => $delta);
}


## @name ::verify_delta
## @summary Verifies the value of delta from the last sample stat saved by store_checkpoint.
## 
## @description
## Verifies the value of delta from the last sample stat saved by store_checkpoint.
##
## @arg expected_values Optional: Hashref of 'counter code'/'expected delta' value pairs to be checked
## If not defined, all the counters that were added by add_counters_to_be_tracked (for the class stats_type)
## will be checked and expected to be 0.
##
## @return Hash with rc and delta.
##
## @examples
##
## my %res = $obj->verify_delta(expected_values => {C_PERF_STATS_COUNTER_BUFFERS_CACHEABLE_IN_EC => 0);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##
sub verify_delta
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        expected_values => {type => HASHREF, optional => 1},
                } );
        
        my $expected_values = $args{expected_values};
        my $counters_to_check = {};
        my $stats_type = $class->{_stats_type};
        
        my $delta = $class->{_stats_obj}->delta(hide_zeros => 0);
        
        if (defined($expected_values))
        {
                foreach my $counter (keys(%$expected_values))
                {
                        my $counter_str = $class->{_ui}->get_perf_stats_counter(code => $counter);
                        $counters_to_check->{$counter_str} = $expected_values->{$counter};
                }
        }
        else
        {
                my $class_tracked_counters = $class->{_tracked_counters};
                foreach my $counter (@$class_tracked_counters)
                {
                        $counters_to_check->{$counter} = 0;
                }
        }
        
        foreach my $counter_to_check (keys(%$counters_to_check))
        {
                my $expected_value = $counters_to_check->{$counter_to_check};
                my $obtained_value = $delta->{$counter_to_check};
                if (!defined($obtained_value))
                {
                        logresult('fatal', "Failed to detect delta for \"$counter_to_check\"");
                }
                
                logcomment("Checking $stats_type delta for counter \"$counter_to_check\"\n" .
                           "Expected value: $expected_value\n" .
                           "Obtained value: $obtained_value");
                
                if ($obtained_value != $expected_value)
                {
                        logresult('fatal', "$stats_type delta mismatch for \"$counter_to_check\":\n" .
                                           "Obtained $obtained_value, expected $expected_value:\n" . Dumper($delta));
                }
        }
        
        return (rc => C_RC_SUCCESS, delta => $delta);
}


## @name ::start_stats_monitor
## @summary Monitors the stats in the background.
## 
## @description
## Monitors the stats in the background.
##
## @arg poll_interval Optional: Poling interval (default: 5s).
##
## @return The background subtest pid.
##
## @examples
##
## $obj->start_stats_monitor(poll_interval => 10);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##
sub start_stats_monitor
{
	my $class = shift;
        my (%args) = validate( @_,
		{
                        poll_interval => { type => SCALAR|UNDEF, optional => 1, },
                } );
        
        # TODO: test c-mode 
        return $class->{_stats_obj}->monitor(interval => $args{poll_interval});
}


## @name ::stop_stats_monitor
## @summary Stops the background monitor.
## 
## @description
## Stops the background monitor.
##
## @arg allowed_variations Optional: Allowed variation (in percentage) for two consecutive collected values
## (the check will be performed for all the 'tracked counters' for the class stats_type).
## If this argument is not defined, no value check is performed.
##
## @return Hash with rc and output.
##
## @examples
##
## my %res = $obj->stop_stats_monitor();
## my %res = $obj->stop_stats_monitor(allowed_variations => 0);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private 
##
sub stop_stats_monitor
{
	my $class = shift;
        my (%args) = validate( @_,
		{
                        allowed_variations => {type => SCALAR, regex => qr/^\d+$/, optional => 1 }, 
                } );
        
        my $class_tracked_counters = $class->{_tracked_counters};
        my $allowed_variations = $args{allowed_variations};
        
        # TODO: test c-mode 
        my $collected_samples = $class->{_stats_obj}->stop();
        
        if (defined($allowed_variations))
        {
                my $class_stats_type = $class->{_stats_type};
                foreach my $counter (@$class_tracked_counters)
                {
                        my $previous_value = $collected_samples->[0]->{$counter};
                        
                        foreach my $sample (@$collected_samples)
                        {
                                my $current_value = $sample->{$counter};
                                
                                logcomment("Checking \"$class_stats_type\" variation for counter \"$counter\"\n" .
                                           "Verifying value: $current_value\n" .
                                           "Previous value: $previous_value\n");
                                
                                compare_values(obtained_value => $current_value,
                                               expected_value => $previous_value,
                                               allow_percentage_error => $allowed_variations,
                                               values_meaning => "allowed \"$class_stats_type\" variation for counter \"$counter\"",
                                               non_fatal_failure => FALSE,
                                               verbose => $class->{_debug});
                                
                                $previous_value = $current_value;
                        }
                }
        }
        
        return (rc => C_RC_SUCCESS, output => $collected_samples);
}	



1;
