#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Lun.pm#1 $
#

## @summary Compression routines specific to luns.
##
## @description
## Compression routines specific to luns.
##
## @examples
##
## use Compression::Constants;
## use vars qw($FILER $AGGREGATE $VOLUME $CLIENT);
##
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $obj = Compression::Compress->setup(
##       filer => { name => $FILER, },
## 	aggregate => { name => $AGGREGATE },
##      volume => { name => $VOLUME, size => "20g",},
##      );
## my $lun = Compression::Lun->new(parent => $compress,
##				   name => 'test_lun',
##				   size => '5g',);
##
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::Lun;

use strict;
use warnings;

use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;
use Utils qw(make_conn);
use SayWrap qw(sayEOF);

use Compression::Constants qw(:all);
use Compression::7Mode::Lun;


BEGIN
{
	require "Compression/base_import.pl";

}



## @name ::new 
##
## @summary Compression::Lun constructor.
##
## @description
## Compression::Lun constructor.
## Creates and configures a lun.
## Currently, this module only supports linux.
##
## @return Compression::Lun object
##
## @arg parent Required: Compression object corresponding to the parent.
## 
## @arg name Optional: Name of the lun. If not provided, it will be auto-generated.
##
## $arg size Optional: Size of the lun. Default: C_DFLT_LUN_SIZE (defined in Constants.pm).
##
## @arg recycle Optional: Flag to decide if the lun should be recycled. Default: FALSE.
## 
## @arg mount Optional: Flag to decide if the lun should be mounted. Default: TRUE.
## 
## @arg initiator_name:Optional: Initiator name. If not passed in, the library will attemt to
## auto-detect it. Failure to detect the initiator name will result in logging config error.
##
## @arg igroup_name Optional: Igroup name to be used. If not passed in, this will default to the
## hostname of the client.
##
##
## @arg iscsiadm_path Optional: Absolute path for the iscsiadm command.
##
## @examples
## my $lun = Compression::Lun->new(parent => $compress,
##				   name => 'test_lun1',
##				   size => '5g',);
##
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub new
{
	my $class = shift;
	my (%args) = validate( @_,
		{
		parent => {type => OBJECT,
			   },

		name => {type => SCALAR,
			 optional => 1,
			 },
		
		size => {type=>SCALAR,
			 default => C_DFLT_LUN_SIZE,
			 },
		
		recycle => {type=>SCALAR,
			    default => FALSE,
			    },
		
		mount => {type=>SCALAR,
			  default => TRUE,
			  },
		
		initiator_name => {type => SCALAR,
				   optional => 1,
				   },
		
		igroup_name => {type => SCALAR,
				optional => 1,
				},
		
		iscsiadm_path => {type => SCALAR,
				  optional => 1,
				  },
		
		port => {type => SCALAR,
			 regex => qr/^\d+$/,
			 default => 3260,
			 },
	});

	# Start building our object:
	my $self = {};
	my $parent = $args{parent};
	my $mount = delete($args{mount});
	
	$self->{_debug} = $parent->{_debug};
	$self->{_disable_cleanup} = $parent->{_disable_cleanup};
	$self->{_creator_pid} = $$;
	$self->{boot_mode} = $parent->{boot_mode};
	$self->{_client_name} = $parent->{_client}->{name};
	$self->{_client_conn} = Utils::make_conn($self->{_client_name},"#rsh-persist");
	$self->{_client_hosttype} = $parent->{dataset}->{client_hosttype};
	$self->{_initiator_name} = delete($args{initiator_name});
	$self->{_igroup_name} = delete($args{igroup_name});
	$self->{_parent_vol} = $parent->{vol_name};
	my $tools_dir = ($main::ISCSI_TOOLS_DIR) ? $main::ISCSI_TOOLS_DIR : '/sbin';
	$tools_dir .= "/";
	$tools_dir =~ s/\/\//\//g;
	$self->{_tools_dir} = $tools_dir;
	$self->{_iscsiadm_path} = ($args{iscsiadm_path}) ? $args{iscsiadm_path} : $self->{_tools_dir} . 'iscsiadm';
	delete($args{iscsiadm_path});
	$self->{_filer_port} = delete($args{port});
	$self->{_logged_in} = FALSE;
	$self->{_mounted} = FALSE;
	$self->{_present_disks} = {};
	
	if (!($parent->isa("Compression::Compress")) && !($parent->isa("Compression::VolClone")))
	{
		logresult('script', "The parent parameter needs to be a Compress or a VolClone object")
	}
	
	$self->{_type} = $self->{_client_hosttype};
	if ($self->{_client_hosttype} eq 'unix-linux')
	{
		# Redhat linux
		$self->{_type} = 'linux'; 
	}
	
	if ($self->{_type} ne 'linux')
	{
		logresult("script", "Lun support is implemented and tested only on linux");
	}
	$args{type} = $self->{_type};
	
	if ($self->{boot_mode} eq C_ONTAP_CMODE)
	{
		logresult('script', "Lun support for C-mode is not available yet")
	}
	else
	{
		$self->{_lun} = Compression::7Mode::Lun->new(%args);
	}
	
	$self->{name} = $self->{_lun}->{_name};
	$self->{_filer_ip} = $self->{_lun}->{_filer_ip};
	$self->{_full_path} = $self->{_lun}->{_full_path};
	
	bless($self, $class);
	
	$self->_configure();
	
	if ($mount == TRUE)
	{
		$self->mount_lun();
	}

	return $self;

}


## @name ::mount_lun
## @summary Mounts the lun on the client.
##
## @description
## Mounts the lun on the client.
##
## @arg chmod Optional: Permissions to set to the new mountpath. Default: 777.
##
## @return Hash with rc and mountpoint. 
##
## @example
## $lun->mount_lun();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub mount_lun
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		chmod => {type => SCALAR, regex => qr/^\d+$/, default => 777,},
	} );
	
	my $chmod = $args{chmod};
	my $device = $class->{_device};
	my $conn = $class->{_client_conn};
	my $mountpoint = $class->{_mountpoint};
	my $name = $class->{name};
	my $cmd = '';
	my $output = '';
	my $vol_name = $class->{_parent_vol};
	
	if (TRUE == $class->{_mounted})
	{
		# mountpoint cannot be undefined... ok to print
		logresult('script', "Lun $name is already mounted on $mountpoint")
	}
	
	# mountpoint will be undefined, if this is the lun is mounted for the first time
	if (!defined($mountpoint))
	{
		my $filer_ip = $class->{_filer_ip};
		$mountpoint = "/t/$filer_ip/luns/$vol_name/$name";
		$cmd = "sudo mkdir -p $mountpoint";
		$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT});
		if ($output =~ /mkdir: .+/)
		{
			logresult('config', "Creating mountpoint for the lun failed: $output");
		}
		$class->{_mountpoint} = $mountpoint;
	}
	
	logcomment("Mounting lun $name ($device) on $mountpoint");
	$cmd = "sudo mount $device $mountpoint";
	$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT});
	if ($output =~ /mount: .+/)
	{
		logresult('fatal', "Mounting the lun failed: $output");
	}
	
	logcomment("Changing permissions to $mountpoint to $chmod");
	$cmd = "sudo chmod $chmod $mountpoint";
	$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT});
	if ($output =~ /chmod: .+/)
	{
		logresult('fatal', "Changing permissions to $mountpoint failed: $output");
	}
	
	$class->{_mounted} = TRUE;
	
	return (rc => C_RC_SUCCESS, mountpoint => $mountpoint,)
}


## @name ::unmount_lun
## @summary Unmounts the lun from the client.
##
## @description
## Unmounts the lun from the client.
##
## @return Hash with rc.
##
## @example
## $lun->unmount_lun();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub unmount_lun
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		# None
	} );
	
	my $mountpoint = $class->{_mountpoint};
	my $name = $class->{name};
	
	if (FALSE == $class->{_mounted})
	{
		logresult('script', "Lun $name is not mounted")
	}
	
	logcomment("Unmounting lun $name ($mountpoint)");
	my $cmd = "sudo umount $mountpoint";
	my $output = sayEOF($class->{_client_conn}, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT});
	if ($output =~ /umount: .+/)
	{
		logresult('fatal', "Unmounting the lun failed: $output");
	}
	
	$class->{_mounted} = FALSE;
	
	return (rc => C_RC_SUCCESS);
}


## @name ::get_mountpoint
## @summary Get lun's mountpoint.
##
## @description
## Get lun's mountpoint.
##
## @return Muntpoint.
##
## @example
## my $mountpoint = $lun->get_mountpoint();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub get_mountpoint
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		# None
	} );
	
	if (FALSE == $class->{_mounted})
	{
		logresult('script', "Lun " . $class->{name} . " is not mounted")
	}
	
	return $class->{_mountpoint};
}


## @name ::resize_lun
## @summary Resize the lun.
##
## @description
## Resize the lun.
##
## @arg size Required: New lun size.
##
## @return Hash with rc.
##
## @example
## my $mountpoint = $lun->resize_lun(size => '6g');
##
sub resize_lun
{
	my $class = shift;
	
	return $class->{_lun}->resize_lun(@_);
}


## @name ::online_lun
## @summary Online the lun.
##
## @description
## Online the lun.
##
## @return Hash with rc.
##
## @example
## $lun->online_lun();
##
sub online_lun
{
	my $class = shift;
	
	return $class->{_lun}->online_lun(@_);
}


## @name ::offline_lun
## @summary Offline the lun.
##
## @description
## Offline the lun.
##
## @return Hash with rc.
##
## @example
## $lun->offline_lun();
##
sub offline_lun
{
	my $class = shift;
	
	return $class->{_lun}->offline_lun(@_);
}


############################################
# Internal methods
############################################

sub _configure
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		igroup_name => {type => SCALAR, optional => 1,},
		initiator_name => {type => SCALAR, optional => 1,},
	} );
	
	logcomment("Configuring lun " . $class->{name});
	$class->{_lun}->license_and_start_iscsi();
	
	if (!defined($class->{_igroup_name}))
	{
		$class->{_igroup_name} = $class->_generate_igroup_name();
	}
	
	if (!defined($class->{_initiator_name}))
	{
		$class->{_initiator_name} = $class->_detect_initiator_name();
	}
	
	$class->{_lun}->create_igroup(igroup_name => $class->{_igroup_name},
				      initiator_name => $class->{_initiator_name},);
	
	$class->{_lun}->map_to_igroup();
	
	$class->{_iscsi_nodename} = $class->{_lun}->get_iscsi_nodename();
	
	$class->_iscsi_discover_and_login();
	
	$class->_format_and_make_fs();
	
	return C_RC_SUCCESS;
}

sub _format_and_make_fs
{
	my $class = shift;
	my $conn = $class->{_client_conn};
	my $filer_ip = $class->{_filer_ip};
	my $name = $class->{name};
	my $tools_dir = $class->{_tools_dir};
	my $cmd = '';
	my $output = '';
	my $disk;
	
	logcomment("Attempting to detect lun's disk");
	$cmd = "sudo " . $tools_dir . "fdisk -l"; 
	my $done = 0;
	my $timeout = 300;
	while ($done == 0)
	{
		$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT});
		if ($output =~ /Disk (\S+) doesn't contain a valid partition table/)
		{
			$disk = $1;
			$done = 1;
			# Now remember what disks we found. We may need this if clones will be created
			my @lines = split(/\n/, $output);
			foreach my $line (@lines)
			{
				if ($line =~ /Disk (\S+): /)
				{
					$class->{_present_disks}->{$1} = 1;
				}
			}
		}
		else
		{
			sleep(10);
			$timeout -= 10;
			
			if ($timeout <= 0)
			{
				logresult('fatal', "Timed-out while attemting to detect lun's disk")
			}
		}
	}
	
	if ($disk !~ /^\/dev/)
	{
		logresult('config', "Failed to parse the output to detect the disk: $disk");
	}
	
	logcomment("Formatting the disk $disk");
	$cmd = "sudo " . $tools_dir . "fdisk $disk";
	Tharn::say($conn, $cmd,'-noprompt', '-wait', "Command .+:");
	Tharn::say($conn, 'n','-noprompt', '-wait', "primary partition");
	Tharn::say($conn, 'p','-noprompt', '-wait', "Partition number .+:");
	Tharn::say($conn, '1','-noprompt', '-wait', "First cylinder.+:");
	Tharn::say($conn, "\n",'-noprompt', '-wait', "Last cylinder.+:");
	Tharn::say($conn, "\n",'-noprompt', '-wait', "Command .+:");
	(undef, undef, $output) = Tharn::say($conn, "w", '-timeout', C_DFLT_ISCSI_CMD_TIMEOUT);
	if ($output !~ /Syncing disks/)
	{
		logresult('fatal', "Formatting $disk failed: $output");
	}
	
	my $device;
	$cmd = "sudo " . $tools_dir . "fdisk -l";
	$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT});
	if ($output =~ /($disk\S+).+Linux/)
	{
		$device = $1;
	}
	
	if (!defined($device))
	{
		logresult('config', "Unable to detect the new device: $output")
	}
	
	logcomment("Making the file system on $device");
	$cmd = "sudo " . $tools_dir . "mke2fs $device";
	$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT});
	if ($output !~ /Writing superblocks and filesystem accounting information: done/)
	{
		logresult('fatal', "Making the filesystem on $device failed: $output");
	}
	
	$class->{_device} = $device;
	
	return C_RC_SUCCESS;
}


sub _iscsi_discover_and_login
{
	my $class = shift;
	
	my $iscsiadm = $class->{_iscsiadm_path};
	my $filer_ip = $class->{_filer_ip};
	my $filer_port = $class->{_filer_port};
	my $conn = $class->{_client_conn};
	my $filer_iscsi_nodename = $class->{_iscsi_nodename};
	my $cmd = '';
	my $output = '';
	
	$cmd = "sudo $iscsiadm -m discovery -t st -p $filer_ip:$filer_port";
	$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT} );
	if ($output !~ /$filer_iscsi_nodename/)
	{
		logresult('config', "Iscsi discovery failed: $output");
	}
	
	$cmd = "sudo $iscsiadm -m discovery -t sendtargets -p $filer_ip:$filer_port -P 1";
	$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT} ); 
	if ($output !~ /$filer_iscsi_nodename/)
	{
		logresult('config', "Iscsi discovery sendtargets failed: $output");
	}
	
	$cmd = "sudo $iscsiadm -m node -T $filer_iscsi_nodename -p $filer_ip:$filer_port -l";
	$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT} ); 
	if ($output =~ /Login to.+ successful/)
	{
		$class->{_logged_in} = TRUE;
	}
	elsif ($output =~ /initiator reported error .+ already exists/)
	{
		logcomment("Iscsi login already exists.");
		# Detect the session id, to attempt to rescan
		$cmd = "sudo $iscsiadm -m session";
		$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT});
		if ($output =~ /\[(\d+)\] .+ $filer_iscsi_nodename/)
		{
			my $session_id = $1;
			logcomment("Attempting to rescan. Session id: $session_id");
			$cmd = "sudo $iscsiadm -m session -P 1 -r $session_id -R";
			$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT});
			sleep(20);
		}
		else
		{
			logcomment("Could not detect the session id. No rescanning will be attempted.");
		}
	}
	else
	{
		logresult('config', "Iscsi login failed: $output");
	}
	
	return C_RC_SUCCESS;
}

sub _iscsi_logout
{
	my $class = shift;
	my $filer_iscsi_nodename = $class->{_iscsi_nodename};
	my $filer_ip = $class->{_filer_ip};
	my $filer_port = $class->{_filer_port};
	my $iscsiadm = $class->{_iscsiadm_path};
	
	my $cmd = "sudo $iscsiadm -m node -T $filer_iscsi_nodename -p $filer_ip:$filer_port --logout";
	my $output = sayEOF($class->{_client_conn}, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT} );
	
	if ($output !~ /Logout of .+ successful/)
	{
		logresult('config', "Iscsi logout failed: $output");
	}
	
	$class->{_logged_in} = FALSE;
	return C_RC_SUCCESS;
}

sub _detect_initiator_name
{
	my $class = shift;
	
	logcomment("Initiator name was not passed in when creating the Compression\:\:Lun object. Attempting to detect it");
	my $output = SayWrap::sayEOF($class->{_client_conn}, "sudo cat /etc/iscsi/initiatorname.iscsi");
	if ($output =~ /InitiatorName=(\S+)/)
	{
		return $1;
	}
	logresult('config', "Initiator name was not passed in. Unable to detect it. Cannot continue")
}

sub _generate_igroup_name
{
	my $class = shift;
	
	my $hostname = SayWrap::sayEOF($class->{_client_conn}, "hostname");
	# Remove the domain
	$hostname =~ s/\..*//;
	if (!defined($hostname))
	{
		logcomment("Not able to generate igroup name. Using the client name (warning: this may be localhost)");
		$hostname = $class->{_client_name};
	}
	
	return $hostname;
}

# Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
	
	if (($class->{_disable_cleanup} == FALSE) &&
            ($class->{_creator_pid} == $$))
	{
                logcomment("Executing Lun clean-up for " . $class->{name});
		eval
		{
			if ($class->{_mounted} == TRUE)
			{
				$class->unmount_lun();
			}
			
			if ($class->{_logged_in} == TRUE)
			{
				$class->_iscsi_logout();
			}
		};
		
		if($@)
		{
			logcomment("Lun: cleanup for lun " . $class->{name} . ": error: " . $@);
		}
		
		$class->{_lun} = undef;
        }
	else
	{
                logcomment( "Lun: cleanup:  nothing done: script PID: " . $$) if ($class->{_debug});
	}
        
        if (defined($execution_error))
        {
                Compression::Utils::_c_interpret_logresult($execution_error);
        }
}

# routine to test the module itself
sub testme
{
	package main;
	
	require "drg/Compression/setup_core.pl";
	
	use Compression::Lun;
	use Compression::LunClone;
	
	my %res;
	
	my $compress = Compression::Compress->setup(filer => $FILER,
						    aggregate => $AGGREGATE,
						    volume => {size => '20g',
							       compress_enabled => FALSE},
						    );
	my $volname = $compress->{vol_name};
	my $file_size = MB;
	my $truncate_size = int($file_size/2);
	my $allow_savings_error = 20;
	
	logcomment("Adding compressible data to volume");
	$compress->{dataset}->create(files => [{name => 'file1', size => $file_size,},
					       {name => 'file2', size => $file_size,},],);
	
	logcomment("Adding holey file to volume");
	$compress->{dataset}->create(files => [{name => 'holey1', size => $file_size, type => 'holey',},],);
	$compress->wait_for_cp();
	
	my $lun = Compression::Lun->new(parent => $compress,
					size => '5g',
					);
	
	logcomment("Adding data to lun");
	$compress->{dataset}->create(files => [{name => 'lunfile1', size => $file_size,},
					       {name => 'lunfile2', size => $file_size,},],
				     lun => $lun,);
	
	logcomment("Adding holey file to lun");
	$compress->{dataset}->create(files => [{name => 'lunholey1', size => $file_size, type => 'holey',},],
				     lun => $lun,);
	$compress->wait_for_cp();
	
	my $file1_path = $compress->{dataset}->get_path(filename => 'file1');
	my $file2_path = $compress->{dataset}->get_path(filename => 'file2');
	my $lunfile1_path = $compress->{dataset}->get_path(filename => 'lunfile1');
	my $lunfile2_path = $compress->{dataset}->get_path(filename => 'lunfile2');
	
	logcomment("Comparing the files (no file operations yet)");
	$compress->{dataset}->compare_files(file_1 => $file1_path,
					    file_2 => $file2_path,);
	$compress->{dataset}->compare_files(file_1 => $file1_path,
					    file_2 => $lunfile1_path,);
	$compress->{dataset}->compare_files(file_1 => $lunfile1_path,
					    file_2 => $lunfile2_path,);
	
	logcomment("Taking a snapshot of the volume");
	$compress->{volume}->create_snapshot(name => 'snap1');
	
	logcomment("Testing file modify");
	$compress->{dataset}->modify(files =>[{ name => 'file2', offset => 0, len => $truncate_size, pattern => 'b'},],);
	$compress->{dataset}->modify(files =>[{ name => 'lunfile2', offset => 0, len => $truncate_size, pattern => 'b'},],);
	$compress->wait_for_cp();
	
	logcomment("Comparing the files after modify");
	$compress->{dataset}->compare_files(file_1 => $file2_path,
					    file_2 => $lunfile2_path,);
	$compress->{dataset}->compare_files(file_1 => $file1_path,
					    file_2 => $lunfile2_path,
					    expected_result => 'NOT_EQUAL');
	
	logcomment("Testing file truncate");
	$compress->{dataset}->truncate_file(filename => 'file2', size => $truncate_size);
	$compress->{dataset}->truncate_file(filename => 'lunfile2', size => $truncate_size);
	$compress->wait_for_cp();
	
	logcomment("Testing file delete");
	$compress->{dataset}->delete_file(filename => 'file2',);
	$compress->{dataset}->delete_file(filename => 'lunfile2',);
	
	logcomment("Testing file timestamp");
	$compress->{dataset}->get_timestamp( filename => 'file1');
	$compress->{dataset}->get_timestamp( filename => 'lunfile1');
	
	logcomment("Testing file size");
	my $vol_file_size = $compress->{dataset}->get_file_size( filename => 'file1');
	my $lun_file_size = $compress->{dataset}->get_file_size( filename => 'lunfile1');
	
	if ($vol_file_size != $lun_file_size)
	{
		logresult('fatal', "Size of the volume file ($vol_file_size) does not match the size of the lun file ($lun_file_size)")
	}
	
	logcomment("Testing file atime");
	$compress->{dataset}->get_file_atime(filename => 'file1',);
	$compress->{dataset}->get_file_atime(filename => 'lunfile1',);
	
	logcomment("Testing lun clone");
	my $clone = Compression::LunClone->new(source => $lun);
	my $clone_name = $clone->{name};
	
	$compress->{dataset}->create(files => [{name => 'clonefile1', size => $file_size,},],
				     lun => $clone,);
	$compress->{dataset}->attach_lun_clone_info(source_lun => $lun,
						    clone_lun => $clone,);
	
	my $added_clonefile1_path = $compress->{dataset}->get_path(filename => 'clonefile1', clone_lun => $clone_name);
	$compress->{dataset}->compare_files(file_1 => $file1_path,
					    file_2 => $added_clonefile1_path,);
	
	my $clone_of_lunfile1_path = $compress->{dataset}->get_path(filename => 'lunfile1', clone_lun => $clone_name);
	$compress->{dataset}->compare_files(file_1 => $lunfile1_path,
					    file_2 => $clone_of_lunfile1_path,);
	
	logcomment("Testing file modify - cloned lun");
	$compress->{dataset}->modify(files =>[{ name => 'lunfile1', offset => 0, len => $truncate_size, pattern => 'c'},],
				     clone_lun => $clone_name);
	$compress->{dataset}->compare_files(file_1 => $lunfile1_path,
					    file_2 => $clone_of_lunfile1_path,
					    expected_result => 'NOT_EQUAL');
	
	
	$clone->split_start(wait => FALSE);
	$clone->split_status();
	$clone->split_wait();
	$clone->split_status(expected => C_RC_SCANNER_IDLE);
	
	logresult('pass', "Compression\:\:Lun: testme: Done" );
}



1;
