#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/LunClone.pm#1 $
#

## @summary Compression routines specific to lun clones.
##
## @description
## Compression routines specific to lun clones.
##
## @examples
##
## use Compression::Constants;
## use vars qw($FILER $AGGREGATE $VOLUME $CLIENT);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::LunClone;

use strict;
use warnings;

use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;
use Scalar::Util qw(blessed);

use Utils qw(make_conn);
use SayWrap qw(sayEOF);

use Compression::Constants qw(:all);
use Compression::Lun;
use Compression::7Mode::LunClone;

our @ISA = ("Compression::Lun");

BEGIN
{
	require "Compression/base_import.pl";

}


## @name ::new 
##
## @summary Compression::LunClone constructor.
##
## @description
## Compression::LunClone constructor.
## Creates and mounts a lun clone.
##
## @return Compression::LunClone object.
##
## @arg source Required: Compression::Lun object corresponding to the source lun.
##
## @arg parent Optional: Compression object corresponding to the parent volume.
## If not specified, the clone will be created on the same volume as the source lun.
## 
## @arg name Optional: Name of the lun. If not provided, it will be auto-generated.
##
## $arg mount Optional: Flag to decide if the clone lun will be mounted. Default: TRUE.
##
##
## @examples
## my $clone = Compression::LunClone->new(source => $lun);
##
sub new
{
	my $class = shift;
	my (%args) = validate( @_,
		{
		source => {type => OBJECT,
			   isa => "Compression::Lun",
			   },
		
		parent => {type => OBJECT,
			   isa  => "Compression::Compress", 
			   optional => 1,
			   },

		name => {type => SCALAR,
			 optional => 1,
			 },
		
		mount => {type=>SCALAR,
			  default => TRUE,
			  },
	});

	# Start building our object:
	my $self = {};
	my $parent = $args{parent};
	my $lun_source = $args{source};
	my $mount = delete($args{mount});
	
	$self->{_debug} = $lun_source->{_debug};
	$self->{_disable_cleanup} = $lun_source->{_disable_cleanup};
	$self->{_creator_pid} = $$;
	$self->{boot_mode} = $lun_source->{boot_mode};
	$self->{_client_name} = $lun_source->{_client_name};
	$self->{_client_conn} = $lun_source->{_client_conn};
	$self->{_client_hosttype} = $lun_source->{_client_hosttype};
	$self->{_initiator_name} = $lun_source->{_initiator_name};
	$self->{_igroup_name} = $lun_source->{_igroup_name};
	$self->{_tools_dir} = $lun_source->{_tools_dir};
	$self->{_iscsiadm_path} = $lun_source->{_iscsiadm_path};
	$self->{_filer_port} = $lun_source->{_filer_port};
	$self->{_filer_ip} = $lun_source->{_filer_ip};
	$self->{_iscsi_nodename} = $lun_source->{_iscsi_nodename};
	$self->{_present_disks} = $lun_source->{_present_disks};
	$self->{_parent_vol} = (defined($parent)) ? ($parent->{vol_name}) : $lun_source->{_parent_vol};
	$self->{_cloned_volume} = FALSE; # clone of a lun, not lun inside of a volume clone
	
	$self->{_type} = $lun_source->{_type};
	$self->{_logged_in} = FALSE;
	$self->{_mounted} = FALSE;
	
	if ($self->{boot_mode} eq C_ONTAP_CMODE)
	{
		logresult('script', "Lun support for C-mode is not available yet")
	}
	else
	{
		$self->{_lun} = Compression::7Mode::LunClone->new(%args);
	}
	
	$self->{name} = $self->{_lun}->{_name};
	$self->{_full_path} = $self->{_lun}->{_full_path};
	
	bless($self, $class);
	
	$self->_configure();
	
	if ($mount == TRUE)
	{
		$self->mount_lun();
	}

	return $self;

}


## @name ::split_start
##
## @summary Start splitting this lun clone.
##
## @description
## Start splitting this lun clone.
##
## @arg wait Optional: Flag to decide if the the method should wait for the split to complete. Default: TRUE.
##
## @examples
## $clone->split_start();
## $clone->split_start(wait => FALSE);
##
sub split_start
{
	my $class = shift;
	
	return $class->{_lun}->split_start(@_);
}


## @name ::split_wait
##
## @summary Wait for split to complete.
##
## @description
## Wait for split to complete.
##
## @examples
## $clone->split_wait();
##
sub split_wait
{
	my $class = shift;
	
	return $class->{_lun}->split_wait(@_);
}


## @name ::split_status
##
## @summary Check the status of the split operation.
##
## @description
## Check the status of the split operation.
##
## @arg expected Optional: Expected return code for the operation. Default: C_RC_SUCCESS.
##
## @examples
## $clone->split_status();
## $clone->split_status(expected => C_RC_SCANNER_IDLE);
##
sub split_status
{
	my $class = shift;
	
	return $class->{_lun}->split_status(@_);
}


## @name ::attach_to_lun_in_cloned_volume 
##
## @summary Attach to the lun contained by a cloned volume.
##
## @description
## Attach to the lun contained by a cloned volume.
##
## @return Compression::LunClone object.
##
## @arg source Required: Compression::Lun object corresponding to the source lun.
##
## @arg cloned_vol Required: VolClone object corresponding to the cloned volume.
## If not specified, the clone will be created on the same volume as the source lun.
## 
## $arg mount Optional: Flag to decide if the clone lun will be mounted. Default: TRUE.
##
##
## @examples
## my $clone = Compression::LunClone->attach_to_lun_in_cloned_volume(source => $lun,
#                                                                    cloned_vol => $clone,);
##
sub attach_to_lun_in_cloned_volume
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		source => {type => OBJECT,
			   isa => "Compression::Lun",
			   },
		
		cloned_vol => {type => OBJECT,
			       isa  => "Compression::VolClone",
			       },
		
		mount => {type=>SCALAR,
			  default => TRUE,
			  },
	});
	
	if (blessed($class))
	{
		logresult('script', "This method should not be called on existent objects");
	}
	
	# Start building our object:
	my $self = {};
	my $parent = $args{cloned_vol};
	my $lun_source = $args{source};
	my $mount = delete($args{mount});
	
	$self->{_debug} = $lun_source->{_debug};
	$self->{_disable_cleanup} = $lun_source->{_disable_cleanup};
	$self->{_creator_pid} = $$;
	$self->{boot_mode} = $lun_source->{boot_mode};
	$self->{_client_name} = $lun_source->{_client_name};
	$self->{_client_conn} = $lun_source->{_client_conn};
	$self->{_client_hosttype} = $lun_source->{_client_hosttype};
	$self->{_initiator_name} = $lun_source->{_initiator_name};
	$self->{_igroup_name} = $lun_source->{_igroup_name};
	$self->{_parent_vol} = $parent->{vol_name};
	$self->{_tools_dir} = $lun_source->{_tools_dir};
	$self->{_iscsiadm_path} = $lun_source->{_iscsiadm_path};
	$self->{_filer_port} = $lun_source->{_filer_port};
	$self->{_filer_ip} = $lun_source->{_filer_ip};
	$self->{_iscsi_nodename} = $lun_source->{_iscsi_nodename};
	$self->{_present_disks} = $lun_source->{_present_disks};
	$self->{_cloned_volume} = TRUE; # clone contained by a volume clone
	
	$self->{_type} = $lun_source->{_type};
	$self->{_logged_in} = FALSE;
	$self->{_mounted} = FALSE;
	
	if ($self->{boot_mode} eq C_ONTAP_CMODE)
	{
		logresult('script', "Lun support for C-mode is not available yet")
	}
	else
	{
		$self->{_lun} = Compression::7Mode::LunClone->_attach_to_lun_in_cloned_volume(%args);
	}
	
	$self->{name} = $self->{_lun}->{_name};
	$self->{_full_path} = $self->{_lun}->{_full_path};
	
	bless($self, $class);
	
	$self->_configure(online => TRUE,);
	
	if ($mount == TRUE)
	{
		$self->mount_lun();
		$parent->{dataset}->attach_lun_clone_info(source_lun => $lun_source,
							  clone_lun => $self,
							  _lun_of_cloned_vol => TRUE,);
	}
	
	return $self;
}




############################################
# Internal methods
############################################

sub _configure
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		igroup_name => {type => SCALAR, optional => 1,},
		initiator_name => {type => SCALAR, optional => 1,},
		online => {type => SCALAR, default => FALSE,}
	} );
	
	logcomment("Configuring lun clone " . $class->{name});
	$class->{_lun}->map_to_igroup();
	
	if (TRUE == $args{online})
	{
		$class->online_lun();
	}
	$class->_format_and_make_fs();
	
	return C_RC_SUCCESS;
}

sub _format_and_make_fs
{
	my $class = shift;
	my $conn = $class->{_client_conn};
	my $filer_ip = $class->{_filer_ip};
	my $name = $class->{name};
	my $tools_dir = $class->{_tools_dir};
	my $iscsiadm = $class->{_iscsiadm_path};
	my $filer_iscsi_nodename = $class->{_iscsi_nodename};
	my $cmd = '';
	my $output = '';
	my $disk;
	my $session_id;
	
	logcomment("Attempting to detect the clone lun");
	
	# Detect the session id
	$cmd = "sudo $iscsiadm -m session";
	$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT});
	if ($output =~ /\[(\d+)\] .+ $filer_iscsi_nodename/)
	{
		$session_id = $1;
	}

	if (!defined($session_id))
	{
		logresult('config', "Failed to detect the iscsi session id");
	}
	
	$cmd = "sudo $iscsiadm -m session -P 1 -r $session_id -R";
	$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT});
	if ($output =~ /(\/dev\/\S+).+Linux/)
	{
		$class->{_device} = $1;
	}
	else
	{
		# Output was not comprehensive enough
		# Try to find the "newly-appeared" disk
		# Add some delay to allow the disk to pop up
		sleep(20);
		$cmd = "sudo " . $tools_dir . "fdisk -l"; 
		$output = sayEOF($conn, $cmd, {timeout => C_DFLT_ISCSI_CMD_TIMEOUT});
		my @lines = split(/\n/, $output);
		my @new_disks = ();
		foreach my $line (@lines)
		{
			if ($line =~ /Disk (\S+): /)
			{
				my $disk = $1;
				if (!exists($class->{_present_disks}->{$1}))
				{
					push(@new_disks, $1);
				}
			}
		}
		
		if (scalar(@new_disks) != 1)
		{
			logresult('config', "Failed to detect the clone lun. New disks:\n" . Dumper(@new_disks))
		}
		
		my $clone_lun = $new_disks[0];
		if ($output =~ /($clone_lun\d+)/)
		{
			$class->{_device} = $1;
		}
	}
	
	if (!defined($class->{_device}))
	{
		logresult('config', "Unable to detect the new device: $output")
	}
	
	return C_RC_SUCCESS;
}


1;
