#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Performance.pm#1 $
#

## @summary Utility module to be used for performance testing.
##
## @description
## Utility module to be used for performance testing.
## Intended to be used as a member of the Compress object.
##
## @example
##
## my $compress = Compression::Compress->setup(filer => {name => $FILER,},
##                                             aggregate => {name => $AGGREGATE,},
##                                             volume => {name => $VOLUME,},
##                                             performance => TRUE,);
## my $volname = $compress->{vol_name};
##
## $compress->{performance}->start_timer(action => "my test step");
## ...
## $compress->{performance}->stop_timer(max_allowed_duration => 5,);

##
## @note
## This module is intended to be only used internally by the Compression library.
## Interfaces may change at any point.
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
## @status private (internal)


package Compression::Performance;

use strict;
use warnings;

use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;
use Tharn;
use Compression::Constants qw(:all);
use Compression::Utils qw(process_performance_info);



BEGIN
{
	require "Compression/base_import.pl";
}



## @name ::new
## @summary Compression::Performance Constructor.
##
## @description Compression::Performance Constructor.
##
## @return  Compression::Performance object.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
## 
## @arg filer Required: Name of the filer.
## 
## @arg boot_mode Required: Boot mode of the hosting filer (C_ONTAP_7MODE or C_ONTAP_CMODE).
##
sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
		filer => {type => SCALAR,},
		debug => {type => SCALAR, default  => FALSE,},
		boot_mode => {type => SCALAR,},
	});
	
	my $self = {};
	$self->{_filer} = $args{filer};
	$self->{_debug} = $args{debug};
	$self->{_boot_mode} = $args{boot_mode};
	$self->{_start_time} = 0;
	$self->{_timer_purpose} = 'unspecified';
	
	
	bless $self, $class;
	
	return $self;
}

## @name ::start_timer
## @summary Start the timer.
##
## @description
## Start the timer. This starting time will be used to calculate the duration of a certain operation
## as part of stop_timer call.
##
## @arg action Required: Short description of the action for which the timer is started.
## Used for logging purposes.
##
## @return Hash with rc and start_time.
##
## @example
## $compress->{performance}->start_timer(action => "snapvault update");
##
sub start_timer
{
	my $class = shift;
	my %args = validate(@_,
	{
		action => {type => SCALAR,},
	});
	
	$class->{_timer_purpose} = $args{action};
	
	logcomment("\nStarted performance timer for action:\n" . $args{action});
	$class->{_start_time} = time();
	
	return (rc => C_RC_SUCCESS);
}


## @name ::stop_timer
## @summary Stops the timer.
##
## @description
## Stops the timer. Uses the starting time obtained by start_timer to calculate the duration of
## a certain operation. 
##
## @return Hash with rc and duration.
##
## @arg failure_handling Optional: Flag to control how failures should be handled.
## Allowed values: none, fail (log fatal result),
## panic (panic the 'generate_cores_on' filers to generate cores), or panic_and_fail.
## Default: fail.
##
## @arg max_allowed_duration Optional: Maximum value allowed for the transfer duration (s).
## If this argument is used, the obtained duration is compared against it and
## if it is bigger, then 'failure_handling' action is executed.
##
## @arg min_allowed_throughput Optional: Minimum value allowed for the throughput (KB/s).
## If this argument is used, the obtained throughput is compared against it and
## if it is smaller, then 'failure_handling' action is executed.
##
## @arg processed_data_size Optional: If min_allowed_throughput is provided, then this
## parameter should also be provided, to specify the amount of data (in KB) that was processed.
##
## @arg generate_cores_on Optional: This parameter can be used to specify the filer names on which
## to generate cores in case of degraded performace. In case this is not provided,
## cores will be generated on the filer that corresponds to this object.
##
## @example
## $compress->{performance}->stop_timer(); # will just log the duration.
## $compress->{performance}->stop_timer(max_allowed_duration => 5,); # will ensure that the duration was no longer than 5 sec.
## $compress->{performance}->stop_timer(min_allowed_throughput => 100,
##                                      processed_data_size => 200); # will ensure that the throughput was no smaller than 100.
## $compress->{performance}->stop_timer(max_allowed_duration => 5,
##                                      failure_handling => 'panic',
##                                      generate_cores_on => [filer_src, filer_dst],); # to generate cores in case of failure on src and dst
##
sub stop_timer
{
	my $class = shift;
	my %args = validate( @_,
	{
		max_allowed_duration => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
		min_allowed_throughput => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
		processed_data_size => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
		failure_handling => {type => SCALAR, regex => qr/^none|fail|panic|panic_and_fail$/, default => 'fail',},
		generate_cores_on => {type => ARRAYREF, optional => 1,},
	});
	
	if ($class->{_start_time} == 0)
	{
		logresult('script', "Timer was not previously started. Please call start_timer before the action that needs to be timed.");
	}
	
	my $stop_time = time();
	my $duration = $stop_time - $class->{_start_time};
	my $filers_to_panic = (defined($args{generate_cores_on})) ? $args{generate_cores_on} : [$class->{_filer},];
	my $description = $class->{_timer_purpose};
	
	$class->{_start_time} = 0;
	$class->{_timer_purpose} = 'unspecified';
	
	return Compression::Utils::process_performance_info(duration => $duration,
							    description => $description,
							    boot_mode => $class->{_boot_mode},
							    max_allowed_duration => $args{max_allowed_duration},
							    min_allowed_throughput => $args{min_allowed_throughput},
							    processed_data_size => $args{processed_data_size},
							    failure_handling => $args{failure_handling},
							    generate_cores_on => $filers_to_panic,);
}




1;
