#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/SNMP.pm#1 $
#

## @summary Compression SNMP module
##
## @description
##    This module provides SNMP support for compression volumes.
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (internal)
##
## @note
## This module is only intended to be used internally by the Compression library.
## Therefore, changes to some APIs may happen at any time.

package Compression::SNMP;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;
use Tharn;

use Compression::Constants qw(:all);
use Compression::Utils qw(convert_data_size_value);
use NACL::APISet;



BEGIN
{
	require "Compression/base_import.pl";
}



# @name ::new
# @summary Compression::SNMP Constructor.
# @return Compression::SNMP object.
#
# @arg filer Required: Filer name.
# @arg ontap Required: Ontap version of the filer.
# @arg boot_mode Optional: Boot mode of the filer. Defaults to 7-mode.
# @arg debug Optional: Controls the verbosity of this module. Defaults to FALSE.
# @arg volume Required: Volume name.
#
# @example
#
# my $obj = Compression::SNMP->new(filer => $FILER,
#                                  volume => $VOLUME,
#                                  ontap => $ontap,
#                                  boot_mode => C_ONTAP_7MODE,
#                                  debug => TRUE,);
#
sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
		filer 	=> {type => SCALAR,},
		ontap => {type => SCALAR,},
		boot_mode => {type => SCALAR, default => C_ONTAP_BOOT_MODE,},
		debug  => {type => SCALAR, default => FALSE,},
		volume  => {type => SCALAR,},
	});
	
	my $self = {};
	$self->{_filer} = $args{filer};
	$self->{_boot_mode} = $args{boot_mode};
	$self->{_debug} = $args{debug};
	$self->{_volume} = $args{volume};
	$self->{_ontap} = $args{ontap};
	
	if ($self->{_ontap} eq '8.0')
	{
		logresult('script', "SNMP support is not currently available for BR");
	}
	
	my $nacl_set_type = "7Mode";
	if ($self->{_boot_mode} eq C_ONTAP_CMODE)
	{
		$nacl_set_type = "CMode";
	}
	
	$self->{_hostobj} = host($self->{_filer});
	$self->{_snmp} = NACL::APISet->new(hostobj   => $self->{_hostobj},
					   category  => "Node",
					   interface => "SNMP",
					   set       => $nacl_set_type,
					   cdef_path => $main::SNMP_CDEFS);
	
	
	bless $self, $class;
	
	return $self;
}


## @name ::verify_savings
## @summary Verify space savings for the volume.
##
## @description
## Verify space savings for the volume.
##
## @return Hash with rc and output.
##
## @arg expected_result Required: Hashref representing the expected values.
## This should be either the 'output' entry of the return of 'verify_savings'
## or the 'savings' entry of the output of 'get_savings'.
##
## @arg check_values Required: Arrayref representing the list of values to be checked.
## Allowed values: compressed, deduplicated, used.
##
## @arg modifier Optional: Data modifier to be used when retrieving savings.
## Defaults to 'k'.
##
## @examples
## my %res = $compress->verify_savings(expected_compressed => 0,);
## $compress->{snmp}->verify_savings(expected_result => $res{output},
##                                   check_values => ['compressed',],);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub verify_savings
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		expected_result => {type => HASHREF,},
		check_values => {type => ARRAYREF,},
		modifier => {type => SCALAR, optional => 1, default => 'k',}, 
	});
	
	my $cmd_code = C_CMD_GET_SAVINGS;
	my $baseoid = $class->_get_baseoid($cmd_code);
	
	my $response = $class->{_snmp}->execute_raw_command('method' => 'snmp_walk',
							    'baseoid' => $baseoid,
							    'return-response-object' => 1,);
	my $parsed_response = $response->get_parsed_output();
	
	return $class->_check_for_expected_values(code => $cmd_code,
						  response => $parsed_response,
						  expected_result => $args{expected_result},
						  values_to_check => $args{check_values},
						  modifier => $args{modifier},);
}


## @name ::is_compression_enabled
## @summary Verify if compression is enabled on the volume.
##
## @description
## Verify if compression is enabled on the volume.
##
## @return Hash with rc and output.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes that compression be enabled. Legal values: C_RC_SUCCESS, C_RC_FAILURE.
##
## @arg inline Optional: Expected return code (defined in Compression::Constants) for inline compression check.
## Legal values: C_RC_SUCCESS, C_RC_FAILURE. Default: $main::IS_INLINE_ENABLED.
##
## @examples
## $compress->{snmp}->is_compression_enabled();
## $compress->{snmp}->is_compression_enabled(inline => C_RC_FAILURE);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub is_compression_enabled
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
		inline => { type => SCALAR, optional => 1, default => $main::IS_INLINE_ENABLED},
	});
	
	my $expected = $args{expected};
	my $inline = $args{inline};
	my $expected_state;
	
	if ($expected == C_RC_FAILURE)
	{
		$expected_state = 'COMPR_OFF';
	}
	elsif ($expected == C_RC_SUCCESS)
	{
		$expected_state = 'COMPR_ON_FULLY_FUNCTIONAL';
		if ((defined($inline)) && ($inline == C_RC_FAILURE))
		{
			$expected_state = 'COMPR_ON_NO_INLINE';
		}
	}
	else
	{
		logresult('script', "Argument expected has an invalid value: $expected\n" .
			            "Please only use C_RC_SUCCESS or C_RC_FAILURE");
	}
	
	my $cmd_code = C_CMD_IS_COMPRESSION_ENABLED;
	my $baseoid = $class->_get_baseoid($cmd_code);
	
	my $expected_result = $class->_generate_expected_result_by_state(state => $expected_state,
									 code => $cmd_code,);
	
	my $response = $class->{_snmp}->execute_raw_command('method' => 'snmp_walk',
							    'baseoid' => $baseoid,
							    'return-response-object' => 1,);
	my $parsed_response = $response->get_parsed_output();
	
	return $class->_check_for_expected_values(code => $cmd_code,
						  response => $parsed_response,
						  expected_result => $expected_result,
						  values_to_check => ['is_compression_on', 'is_inline_compression_on'],);
}


## @name ::verify_compress_status
## @summary Verify compression scanner status on the volume.
##
## @description
## Verify compression scanner status on the volume.
##
## @return Hash with rc and output.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes that scanner should be running. Legal values: C_RC_SUCCESS, C_RC_SCANNER_IDLE.
##
## @examples
## $compress->{snmp}->verify_compress_status();
## $compress->{snmp}->verify_compress_status(expected => C_RC_SCANNER_IDLE);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub verify_compress_status
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
	});
	
	my $expected = $args{expected};
	if (($expected != C_RC_SCANNER_IDLE) && ($expected != C_RC_SUCCESS))
	{
		logresult('script', "Invalid value for expected argument: $expected. Allowed: C_RC_SUCCESS, C_RC_SCANNER_IDLE");
	}
	
	my $expected_scanner_state = ($expected == C_RC_SUCCESS) ? 'C_STATUS_ACTIVE' : 'C_STATUS_IDLE';
	my $cmd_code = C_CMD_COMPRESS_STATUS;
	my $baseoid = $class->_get_baseoid($cmd_code);
	
	my $expected_result = $class->_generate_expected_result_by_state(state => $expected_scanner_state,
									 code => $cmd_code,);
	
	my $response = $class->{_snmp}->execute_raw_command('method' => 'snmp_walk',
							    'baseoid' => $baseoid,
							    'return-response-object' => 1,);
	my $parsed_response = $response->get_parsed_output();
	
	return $class->_check_for_expected_values(code => $cmd_code,
						  response => $parsed_response,
						  expected_result => $expected_result,
						  values_to_check => ['scanner_status'],);
}


## @name ::verify_decompress_status
## @summary Verify decompression scanner status on the volume.
##
## @description
## Verify decompression scanner status on the volume.
##
## @return Hash with rc and output.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes that scanner should be running. Legal values: C_RC_SUCCESS, C_RC_NO_DECOMPRESSION_IN_PROGR.
##
## @examples
## $compress->{snmp}->verify_decompress_status();
## $compress->{snmp}->verify_decompress_status(expected => C_RC_NO_DECOMPRESSION_IN_PROGR);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub verify_decompress_status
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
	});
	
	my $expected = $args{expected};
	if (($expected != C_RC_NO_DECOMPRESSION_IN_PROGR) && ($expected != C_RC_SUCCESS))
	{
		logresult('script', "Invalid value for expected argument: $expected. Allowed: C_RC_SUCCESS, C_RC_NO_DECOMPRESSION_IN_PROGR");
	}
	
	my $expected_scanner_state = ($expected == C_RC_SUCCESS) ? 'C_STATUS_ACTIVE' : 'C_STATUS_IDLE';
	my $cmd_code = C_CMD_DECOMPRESS_STATUS;
	my $baseoid = $class->_get_baseoid($cmd_code);
	
	my $expected_result = $class->_generate_expected_result_by_state(state => $expected_scanner_state,
									 code => $cmd_code,);
	
	my $response = $class->{_snmp}->execute_raw_command('method' => 'snmp_walk',
							    'baseoid' => $baseoid,
							    'return-response-object' => 1,);
	my $parsed_response = $response->get_parsed_output();
	
	return $class->_check_for_expected_values(code => $cmd_code,
						  response => $parsed_response,
						  expected_result => $expected_result,
						  values_to_check => ['scanner_status'],);
}


## @name ::verify_volume_status
## @summary Verify status of the volume.
##
## @description
## Verify status of the volume.
##
## @return Hash with rc and output.
##
## @arg expected_values Optional: Hashref of pairs key-value to be checked in volume status.
## The default behavior is to check the volume state (online).
## Legal keys: vol_state, vol_fsid, vol_aggr_name.
##
## @examples
## $compress->{snmp}->verify_volume_status();
## $compress->{snmp}->verify_volume_status(expected_values => {vol_aggr_name => $AGGREGATE});
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub verify_volume_status
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		expected_values => {type => HASHREF, optional => 1,},
	});
	
	my $expected_values = $args{expected_values};
	my $values_to_check = [];
	my $cmd_code = C_CMD_VOL_STATUS;
	
	if (!defined($expected_values))
	{
		$expected_values = $class->_generate_expected_result_by_state(state => 'VOL_IS_ONLINE',
									      code => $cmd_code,);
	}
	
	foreach my $key (keys(%$expected_values))
	{
		push(@$values_to_check, $key);
	}
	
	my $baseoid = $class->_get_baseoid($cmd_code);
	
	my $response = $class->{_snmp}->execute_raw_command('method' => 'snmp_walk',
							    'baseoid' => $baseoid,
							    'return-response-object' => 1,);
	my $parsed_response = $response->get_parsed_output();
	
	return $class->_check_for_expected_values(code => $cmd_code,
						  response => $parsed_response,
						  expected_result => $expected_values,
						  values_to_check => $values_to_check,);
}




################################################################
# Internal subroutines
################################################################


# Resets the ontap version for this object
# Input: new ontap version string.
sub _set_ontap
{
	my $class = shift;
	my $new_ontap = shift;
	
	$class->{_ontap} = $new_ontap;
}


# Get the baseoid for a specific command
# Input: command code (as defined in Constants.pm)
# Output: baseoid string
sub _get_baseoid
{
	my $class = shift;
	my $cmd = shift;
	my $ontap = $class->{_ontap};
	
	my $supported_oids =
	{
		'base' => {GET_SPACE_SAVINGS => '.1.3.6.1.4.1.789.1.5.4.1',
			   IS_COMPRESSION_ENABLED => '.1.3.6.1.4.1.789.1.23.2',
			   COMPRESS_STATUS => '.1.3.6.1.4.1.789.1.23.2',
			   DECOMPRESS_STATUS => '.1.3.6.1.4.1.789.1.23.2',
			   VOL_STATUS => '1.3.6.1.4.1.789.1.5.8',
			   },
	};
	
	if (exists($supported_oids->{$ontap}->{$cmd}))
	{
		return $supported_oids->{$ontap}->{$cmd};
	}
	elsif (exists($supported_oids->{'base'}->{$cmd}))
	{
		return $supported_oids->{'base'}->{$cmd};
	}
	
	logresult('script', "Library problem found: no baseoid defined for command $cmd (release $ontap)");
}


# For each of the 'values_to_check', loop through the entries of 'response'
# and check if all the 'values_to_check' have the expected value.
# Input:
# - code: command code (as defined in Constants.pm)
# - response: The parsed response returned by NACL 
# - expected_result: The hashref representing the expected result,
#   as passed in by the caller of the high-level method (see documentation for each of these external methods)
# - values_to_check: The arrayref that defines which of the entries in the expected_result should be checked, 
#   as passed in by the caller of the high-level method (see documentation for each of these external methods).
#   Some of these values could also be internally generated, by _generate_expected_result_by_state, to match the expected behavior.
#   For example, to check if compression is enabled, we will look at the values for both is_compression_on and is_inline_compression_on.
# - modifier: any data modifier needed for data conversion.
# Output:
#   Hash with rc and output.
sub _check_for_expected_values
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		code => {type => SCALAR,},
		response => {type => ARRAYREF,},
		expected_result => {type => HASHREF,},
		values_to_check => {type => ARRAYREF,},
		modifier => {type => SCALAR, optional => 1,}, 
	});
	
	my $cmd_code = $args{code};
	my $snmp_response = $args{response};
	my $values_to_check = $args{values_to_check};
	my $modifier = $args{modifier};
	my $expected_result = {};
	
	my $processing_info = $class->_get_processing_information($cmd_code);
	
	# Make sure that what we will have to process is a hashref, keyed by entity
	# This will make processing easier
	if ((!defined($processing_info->{expected_result_has_multiple_entries})) ||
	    ($processing_info->{expected_result_has_multiple_entries} == FALSE))
	{
		my $entity = $processing_info->{expected_result_entity};
		if (!defined($entity))
		{
			logresult('script', "Library problem found: command code $cmd_code is missing information " .
				            "on which entity the values are obtained for:\n" . Dumper($processing_info));
		}
		my $volume = $class->{$entity};
		$expected_result->{$volume} = $args{expected_result};
	}
	else
	{
		$expected_result = $args{expected_result};
	}
	
	# Loop through the entities in the expected result
	# Detect the correspondent entity in the snmp output
	# For each of the values_to_check, verify the expected value against the snmp-obtained value
	foreach my $entity_name (keys(%$expected_result))
	{
		my $snmp_format_info = $processing_info->{snmp_response_info};
		my $expected_entity_result = $expected_result->{$entity_name};
		logcomment("SNMP: processing: $entity_name\nExpected result:\n" . Dumper($expected_entity_result) .
			   "\n\nLooking for snmp output defined by:\n" . Dumper($snmp_format_info)) if ($class->{_debug});
		
		my $entity_defined_by = $snmp_format_info->{entity_defined_by};
		my $expected_entity_name = $entity_name;
		if (!defined($entity_defined_by))
		{
			logresult('script', "Library problem found: command code $cmd_code: missing information: entity_defined_by");
		}
		
		my $detected_entity = 0;
		foreach my $obtained_entities (@$snmp_response)
		{
			my $obtained_entity_name = $obtained_entities->{$entity_defined_by}->{value};
			
			if (defined($snmp_format_info->{entity_value_prefix}))
			{
				my $prefix = $snmp_format_info->{entity_value_prefix};
				$obtained_entity_name =~ s/^$prefix//;
			}
			
			if (defined($snmp_format_info->{entity_value_sufix}))
			{
				my $sufix = $snmp_format_info->{entity_value_sufix};
				$obtained_entity_name =~ s/$sufix$//;
			}
			
			if ("$obtained_entity_name" ne "$expected_entity_name")
			{
				next;
			}
			
			$detected_entity = 1;
			
			foreach my $value_to_check (@$values_to_check)
			{
				my $expected_value = $expected_entity_result->{$value_to_check};
				if (!defined($expected_value))
				{
					logresult('script', "Invalid check_values value: $value_to_check\n" .
						            "Could not detect a correspondent entry in:\n" . Dumper($expected_entity_result));
				}
				
				my $value_to_check_defined_by = $snmp_format_info->{expected_result_tokens_defined_by}->{$value_to_check}->{key};
				my $obtained_value = $obtained_entities->{$value_to_check_defined_by}->{value};
				if (!defined($obtained_value))
				{
					logresult('fatal', "Failed to detect \"$value_to_check\" snmp information for entity $entity_name\n" . Dumper($obtained_entities));
				}
				
				if (defined($modifier))
				{
					my $snmp_value_modifier = $snmp_format_info->{expected_result_tokens_defined_by}->{$value_to_check}->{retrieved_value_modifier};
					if (!defined($snmp_value_modifier))
					{
						logresult('script', "\nLibrary problem found: modifier processing required,\n" .
							            "but no retrieved_value_modifier information defined for:\n" .
								    "command: $cmd_code\n" .
								    "value to check: $value_to_check");
					}
					
					$obtained_value = Compression::Utils::convert_data_size_value(value => $obtained_value,
												      current_modifier => $snmp_value_modifier,
												      required_modifier => $modifier,);
				}
				
				logcomment("\nChecking SNMP information for \"$entity_name\":\n" .
					   "Checking value of \"$value_to_check\":\n" .
					   "Expected value: $expected_value\n" .
					   "Obtained value: $obtained_value");
				
				if ("$expected_value" ne "$obtained_value")
				{
					logresult('fatal', "Expected SNMP value mismatch for $entity_name\n" .
						           "Expected $value_to_check: $expected_value\n" .
							   "Obtained $value_to_check: $obtained_value");
				}
			}
			
			last;
		}
		
		if ($detected_entity == 0)
		{
			logresult('fatal', "Failed to detect information for entity \"$entity_name\" in the SNMP output:\n" . Dumper($snmp_response));
		}
	}
	
	return (rc => C_RC_SUCCESS, output => $snmp_response);
}


# Get processing information for a specific command code.
# This will be used to process the parsed output returned by the NACL library.
# Input:
# - command code (as defined in Constants.pm).
# Output:
# - Hashref representing release-specific processing information.
#
#   Keys in this hash represent:
#   - expected_result_has_multiple_entries: flag to decide if the expected_result hash passed in by the caller is expected to have multiple entries
#     (For example, space savings for more than our volume).
#   - expected_result_entity: in case expected_result_has_multiple_entries is FALSE, the expected_result_entity key is used to decide
#     for which of the class members we just obtained the output
#     (for example, for IS_COMPRESSION_ENABLED, we will be obtaining the output for $class->{_volume}, thus expected_result_entity gets set to '_volume').
#   - snmp_response_info: where to find the needed information in the parsed snmp output:
#     - entity_defined_by: which key should be used to identify the entity that we are interested in
#       (for example, for GET_SPACE_SAVINGS, the name of the volume will be detected by looking at the dfFileSys entry).
#     - entity_value_prefix: any "to be discarded" prefix that may be contained by the value of entity_defined_by
#       (for example, for GET_SPACE_SAVINGS, the value of the dfFileSys entry will be /vol/<volume_name>... , thus we will set entity_value_prefix to '/vol/').
#     - entity_value_sufix: any "to be discarded" sufix that may be contained by the value of entity_defined_by
#       (for example, for 7-mode GET_SPACE_SAVINGS, the value of the dfFileSys entry will be ...<volume_name>/ , thus we will set entity_value_sufix to '/').
#     - expected_result_tokens_defined_by: hashref defining which key should be used to identify each of the possible values required to be checked by expected_result/values_to_check:
#       For each of these possible values_to_check, processing will be defined by:
#       - key: which key should be used to identify it
#         (for example, the 'compressed' field for space savings will be detected by looking at the df64CompressSaved entry).
#       - retrieved_value_modifier: for any data size, this will determine how this data is returned within the snmp output
#         (for example, for GET_SPACE_SAVINGS, the 'compressed' value will be returned in kbytes, thus retrieved_value_modifier is set to 'k').
sub _get_processing_information
{
	my $class = shift;
	my $cmd = shift;
	my $ontap = $class->{_ontap};
	
	my $info =
	{
		'base' => {GET_SPACE_SAVINGS => {expected_result_has_multiple_entries => TRUE,
						 snmp_response_info => {entity_defined_by => 'dfFileSys',
									entity_value_prefix => '/vol/',
									entity_value_sufix => '/', 
									expected_result_tokens_defined_by => {compressed => {key => 'df64CompressSaved',
															     retrieved_value_modifier => 'k',},
													      deduplicated => {key => 'df64DedupeSaved',
															       retrieved_value_modifier => 'k',},
													      used => {key => 'dfKBytesUsed',
														       retrieved_value_modifier => 'k',},},},},
			   IS_COMPRESSION_ENABLED => {expected_result_has_multiple_entries => FALSE,
						      expected_result_entity => '_volume',
						      snmp_response_info => {entity_defined_by => 'sisPath',
									     entity_value_prefix => '/vol/',
									     expected_result_tokens_defined_by => {is_compression_on => {key => 'sisIsCompressionEnabled',},
														   is_inline_compression_on => {key => 'sisIsInlineEnabled',},},},},
			   VOL_STATUS => {expected_result_has_multiple_entries => FALSE,
					  expected_result_entity => '_volume',
					  snmp_response_info => {entity_defined_by => 'volName',
								 expected_result_tokens_defined_by => {vol_state => {key => 'volState',},
												       vol_fsid => {key => 'volFSID',},
												       vol_aggr_name => {key => 'volAggrName',},},},},
												       # Unless TCs require it, for now, do not expose the following information
												       # It may need some extra parsing capabilities (ex: regex instead of match)
												       #clone_of => {key => 'volCloneOf',},
												       #space_guarantee_enabled => {key => 'volSpaceGuaranteeEnabled',},
												       #space_guarantee_type => {key => 'volSpaceGuarantee',},
												       #clone_children_list => {key => 'volCloneChildrenList',},
												       #vol_vserver => {key => 'volVserver',},
												       #vol_is_clone => {key => 'volClone',},
												       #vol_type => {key => 'volType',},
			   COMPRESS_STATUS => {expected_result_has_multiple_entries => FALSE,
					       expected_result_entity => '_volume',
					       snmp_response_info => {entity_defined_by => 'sisPath',
								      entity_value_prefix => '/vol/',
								      expected_result_tokens_defined_by => {scanner_status => {key => 'sisStatus',},},},},
			   DECOMPRESS_STATUS => {expected_result_has_multiple_entries => FALSE,
						 expected_result_entity => '_volume',
						 snmp_response_info => {entity_defined_by => 'sisPath',
									entity_value_prefix => '/vol/',
									expected_result_tokens_defined_by => {scanner_status => {key => 'sisStatus',},},},},
			   },
	};
	
	if (exists($info->{$ontap}->{$cmd}))
	{
		return $info->{$ontap}->{$cmd};
	}
	elsif (exists($info->{'base'}->{$cmd}))
	{
		return $info->{'base'}->{$cmd};
	}
	
	logresult('script', "Library problem found: no processing information defined for command $cmd (release $ontap)");
}


# For some of the API's of this module, keep track internally of which are values to be checked and what are the expected returns
# (for example, when checking that compression is disabled, check that both is_compression_on and is_inline_compression_on are 0).
# Input:
# - code: command code (as defined in Constants.pm).
# - state: what state should we expect to be in
#   (for example: when checking for compression setting: we can be in three different states: compression is off, only bg compression, bg_inline compression).
# Output:
# - Hashref representing the expected result (expected_result).
sub _generate_expected_result_by_state
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		state => {type => SCALAR,},
		code => {type => SCALAR,},
	});
	
	my $ontap = $class->{_ontap};
	my $cmd = $args{code};
	my $state = $args{state};
	
	my $info =
	{
		'base' => {IS_COMPRESSION_ENABLED => {COMPR_OFF => {is_compression_on => 0,
								    is_inline_compression_on => 0,},
						      COMPR_ON_FULLY_FUNCTIONAL => {is_compression_on => 1,
										    is_inline_compression_on => 1,},
						      COMPR_ON_NO_INLINE => {is_compression_on => 1,
									     is_inline_compression_on => 0,},},
			   COMPRESS_STATUS => {C_STATUS_ACTIVE => {scanner_status => 'active',},
					       C_STATUS_IDLE => {scanner_status => 'idle',},},
			   DECOMPRESS_STATUS => {C_STATUS_ACTIVE => {scanner_status => 'undoing',},
						 C_STATUS_IDLE => {scanner_status => 'idle',},},
			   VOL_STATUS => {VOL_IS_ONLINE => {vol_state => 'online',},},
			   },
	};
	
	if (exists($info->{$ontap}->{$cmd}->{$state}))
	{
		return $info->{$ontap}->{$cmd}->{$state};
	}
	elsif (exists($info->{'base'}->{$cmd}->{$state}))
	{
		return $info->{'base'}->{$cmd}->{$state};
	}
	
	logresult('script', "Library problem found: _generate_expected_result_by_state: missing information for command $cmd, state $state (release $ontap)");
}



 
# routine to test the module itself
sub testme
{
	package main;
	
	require "Compression/setup_core.pl";
	
	param('VOLUME', "-default", 'test_pkg_vol'); 
	
	my %res;
	my $compress = Compression::Compress->setup(filer => $FILER ,
						    aggregate => $AGGREGATE,
						    volume => {name => $VOLUME,
							       size => '2g',},
						    snmp => TRUE,);
	
	logcomment("Testing verify_savings (no data yet)");
	%res = $compress->verify_savings(expected_compressed => 0,);
	$compress->{snmp}->verify_savings(expected_result => $res{output},
					  check_values => ['compressed',],);
	
	my $filename = 'file';
	$compress->{dataset}->create(files =>[{name => $filename, size => MB,},],);
	$compress->wait_for_cp();
	
	logcomment("Testing verify_savings after adding 1m of data");
	%res = $compress->verify_savings(expected_compressed => 892,);
	$compress->{snmp}->verify_savings(expected_result => $res{output},
					  check_values => ['compressed',],);
	
	logcomment("Testing is_compression_enabled");
	$compress->is_enabled();
	$compress->{snmp}->is_compression_enabled();
	
	$compress->disable();
	$compress->is_enabled(expected => C_RC_FAILURE);
	$compress->{snmp}->is_compression_enabled(expected => C_RC_FAILURE);
	
	$compress->{dataset}->create(files =>[{name => $filename . "_1", size => 200 * MB,},],);
	$compress->wait_for_cp();
	
	logcomment("Testing verify_compress_status");
	$compress->enable();
	$compress->{compress_scanner}->start();
	$compress->{compress_scanner}->status();
	$compress->{snmp}->verify_compress_status();
	$compress->{compress_scanner}->wait_for_scanned_data();
	$compress->{snmp}->verify_compress_status(expected => C_RC_SCANNER_IDLE);
	
	logcomment("Testing verify_decompress_status");
	$compress->disable();
	$compress->{decompress}->start();
	$compress->{decompress}->status();
	$compress->{snmp}->verify_decompress_status();
	$compress->{decompress}->wait_for_percentage();
	$compress->{snmp}->verify_decompress_status(expected => C_RC_NO_DECOMPRESSION_IN_PROGR);
	
	logcomment("Testing verify_volume_status");
	$compress->{snmp}->verify_volume_status();
	$compress->{snmp}->verify_volume_status(expected_values => {vol_aggr_name => $AGGREGATE});
	
	
	logresult('pass', "SNMP: testme completed" );
}



1;
