#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Sis.pm#1 $
#

## @summary Utility module for dense(sis) operations.
##
## @description
## Utility module for dense(sis) operations. 
##    
## @synopsis
##
## use Compression::Sis;
## use vars qw( $FILER  );
## 
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $dense = = Compression::Sis->new(host => $FILER,
##	 			       volume_path => $vol_mount_path,
##				       initialize => TRUE);
## 
## @author dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::Sis;

use strict;
use warnings;

use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult
			  process_data_size_string);
use Dense qw(dense_is_licensed
             dense_set_license
             dense_toggle
             dense_start
             dense_stop
             dense_verify_state
             dense_off_undo
             dense_config
             dense_check
             dense_wait_for
             dense_get_state);
use Dense::Utils qw(dense_parse_showvbns);


BEGIN
{
	require "Compression/base_import.pl";
	
}



## @name ::new
## @summary Compression::Sis Constructor.
##
## @description
## Creates an object representing a dense utility for the volume.
##
## @return Object of type Compression::Sis.
##
## @arg filer Required: NOTE::Filer object.
##
## @arg volume_path Required: Volume path name.
##
## @arg initialize Optional: Flag to enable extra setup (license, turn sis on).
## Default: FALSE.
## @arg configure_scheduler Optional: Flag to control if the scheduler should be configured
## (to the passed-in schedule argument or to the default schedule). Only used if initialize is TRUE.
## Default: C_DFLT_CONFIGURE_SCHEDULER (defined in Constants.pm).
##
## @arg schedule Optional: String to configure the sis scheduler. Only used if initialize is TRUE.
## Default: -.
##
## @arg disable_cleanup Optional: Flag to disable any clean-up that is executed
## upon object destruction. Defaults to FALSE.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $dense = Compression::Sis->new(host => $FILER,
##	 			     volume_path => $vol_mount_path,
##				     initialize => TRUE); 

sub new
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		filer => { 
                        type => OBJECT,
                        },
                
                volume_path => {
			type => SCALAR|UNDEF,
                        optional => 1,
		},
                
                volume => {
                        type => OBJECT,
                        optional => 1,
		},
                        	
		disable_cleanup => {
			type => BOOLEAN,
                        optional => 1,
			default => FALSE,
		},
		
		debug => {
			type => BOOLEAN,
                        optional => 1,
			default => FALSE,
                        },
                
                initialize => {
			type => BOOLEAN,
                        optional => 1,
			default => FALSE,
                        },
                
                configure_scheduler => {
			type => BOOLEAN,
                        optional => 1,
			default => C_DFLT_CONFIGURE_SCHEDULER,
		},
                
                schedule => {
			type => SCALAR,
                        optional => 1,
			default => C_DFLT_SIS_SCHEDULE,
                        },
		
		vserver => {type => SCALAR|UNDEF,
			    optional => 1},
		
		boot_mode => {type => SCALAR,
			      optional => 1,
			      default => C_ONTAP_BOOT_MODE},
		
		ui => {type => OBJECT,
		       isa => "Compression::UI",
		       optional => 1,
		       },
	});
	
	# Start building our object:
	my $self = {};
        $self->{_initialized} = FALSE;
        $self->{_filer} = $args{filer};
        $self->{_path} = $args{volume_path};
        $self->{_volume} = $args{volume};
        $self->{_disable_cleanup} = $args{disable_cleanup};
        $self->{_debug} = $args{debug};
        $self->{_creator_pid} = $$;
	$self->{_boot_mode} = $args{boot_mode};
        $self->{_vserver} = $args{vserver};
        
        
        if ($self->{_boot_mode} eq C_ONTAP_7MODE)
        {
                if  (!($self->{_filer}->isa("NOTE::Filer")))
                {
                        logresult('script', "For 7-mode, the filer argument is required to be a NOTE\:\:Filer object");
                }
                
                if (!defined($self->{_path}))
                {
                        logresult('script', "For 7-mode, the volume_path argument is required");
                }
                
                $self->{_host} = $self->{_filer}->host();
        }
        elsif ($self->{_boot_mode} eq C_ONTAP_CMODE)
        {
                if (!defined($self->{_volume}))
                {
                        logresult('script', "For C-mode, the volume argument is required");
                }
                
                if ((!($self->{_filer}->isa("NOTE::GX::Node"))) || (!($self->{_volume}->isa("NOTE::GX::Volume"))))
                {
                        logresult('script', "For C-mode, the filer argument is required to be a NOTE\:\:GX\:\:Node object\n" .
                                            "and the volume argument needs to be a NOTE\:\:GX\:\:Volume object");
                }
                
                $self->{_host} = $self->{_filer}->name();
                $self->{_vol_name} = $self->{_volume}->name();
        }
	
	# UI
        if ( defined($args{ui}) )
        {
                $self->{_ui} = $args{ui};
        }
        else
        {
                # Auto-detect version of the hosting filer. Need this for UI.
                logcomment("Sis: UI object not provided. Auto-generating...") if ($self->{_debug});
                my $version = Compression::Version->new( conn => $args{filer}, debug => $self->{_debug} );
                $self->{_ui} = Compression::UI->new( version => $version, debug => $self->{_debug}, boot_mode => $self->{_boot_mode} );
        }

	bless($self, $class);
        
        if ($args{initialize} == TRUE)
        {
                logcomment("Licensing and configuring asis...");
                $self->initialize();
                
                if ($args{configure_scheduler} == TRUE)
                {
                        my $schedule = $args{schedule};
                        logcomment("Configuring sis scheduler to \"$schedule\"");
                        $self->config(schedule_string => $schedule);
                }
        }
	
	logcomment("Sis: Setup Complete..\n\n") if ($self->{_debug});
	return $self;

}


## @name ::initialize
## @summary Perform setup steps for Sis (license, turn sis on). 
##
## @description
## Perform setup steps for Sis (license, turn sis on). 
##
## @return Hash with rc.
##
## @example
## $dense->initialize();

sub initialize
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        # None
		});
        
	my $need_license = $class->{_ui}->get_constant_value(constant_code => 'INSTALL_SIS_LICENSE');
        if (($class->{_boot_mode} eq C_ONTAP_7MODE) && ($need_license))
        {
                if (Dense::dense_is_licensed(host => $class->{_host}))
                {
                        logcomment("Asis is already licensed on this filer");
                } else
                {
                        Dense::dense_set_license(host => $class->{_host});
                }
        }
        elsif ($class->{_boot_mode} eq C_ONTAP_CMODE)
        {
                # Don't do anything here.
                # Sis license will be included in the base license
        }
        
        $class->set_state(state => C_DEDUPE_STATE_ON);
                
        $class->{_initialized} = TRUE;
        
        return (rc => C_RC_SUCCESS);
}


## @name ::set_state
## @summary Turns sis on or off. 
##
## @description
## Turns sis on or off. 
##
## @arg state Optional: New state for sis. Default: on.
##
## @arg volume Optional: Name of volume to set sis state (ignored if 'all_volumes' is TRUE). 
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @arg all_volumes Optional: Flag to set state on all existent volumes.
##
## @arg skip_root_volume Optional: Only considered if 'all_volumes' option is TRUE.
## Boolean flag to decide if root volume should be skipped.
## Defaults to TRUE.
##
## @arg skip_vsroot_volumes Optional: Only considered if 'all_volumes' option is TRUE.
## Boolean flag to decide if vsroot volumes should be skipped.
## Defaults to TRUE.
##
## @arg skip_offline_restricted Optional: Only considered if 'all_volumes' option is TRUE.
## Boolean flag to decide offline restricted volumes should be skipped.
## Defaults to TRUE.
##
## @return Hash with rc.
##
## @example
## $dense->set_state();

sub set_state
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        state => {type => SCALAR, optional => 1, default => C_DEDUPE_STATE_ON},
                        volume => {type => SCALAR, optional => 1},
                        expected => {type => SCALAR, regex => qr/^\d+$/, default  => C_RC_SUCCESS,},
			all_volumes => { type => BOOLEAN, optional => 1, default => FALSE },
			skip_root_volume => { type => BOOLEAN, optional => 1, default => TRUE, },
                        skip_vsroot_volumes => { type => BOOLEAN, optional => 1, default => TRUE, },
                        skip_offline_restricted => { type => BOOLEAN, optional => 1, default => TRUE, },
		});

	my $state = $args{state};
	my $all_volumes = $args{all_volumes};
	my $skip_offline_restricted_flag = $args{skip_offline_restricted};
	my $skip_root_flag = $args{skip_root_volume};
        my $skip_vsroot_flag = $args{skip_vsroot_volumes};

	my $path = $class->{_path};
        my $output = '';
        
	if ($class->{_boot_mode} eq C_ONTAP_7MODE)
        {
				if ($all_volumes == TRUE) {
					my $root_vol_name = $class->{_filer}->vol_root();
					my @vol_list = $class->{_filer}->Volumes;
					foreach my $vol (@vol_list) {
						my $vol_name = $vol->name;
						my $vol_path = $vol->mount_path;
						if (($vol_name eq $root_vol_name)&&($skip_root_flag == TRUE)) {
							next;
						}
						# If required, skip volume if offline or restricted
						if ($skip_offline_restricted_flag == TRUE)
						{
								my $vol_status = $vol->status();
								my $vol_state = $vol_status->{state};
								if ( $vol_state =~ /offline|restricted/ )
								{
										logcomment("Skipping volume \"$vol_name\", in state: \"$vol_state\" ");
										next;
								}
						}
						my $current_state = $class->get_state(vol_path=>$vol_path);
						if ( $current_state eq 'ENABLED' ) {
							next if $state eq C_DEDUPE_STATE_ON;
						} elsif ( $current_state ne 'DISABLED' ) {
							next if $state eq C_DEDUPE_STATE_OFF;
						} else {
							if ($state eq C_DEDUPE_STATE_OFF) {
								# don't turn off if no status
								next;
							}
						}
						eval
						{
								Dense::dense_toggle(host => $class->{_host},
                                                                                    path_name => $vol_path,
                                                                                    state => $state);
						};                
						$output = $@;
						$class->{_ui}->process_output_string(code => C_CMD_DEDUPE_SET_STATE,
						    output => $output,
						    expected_rc => $args{expected},
						    log_message => "Set sis state to $state on path $vol_path");
					}
					return (rc => C_RC_SUCCESS);
				} else {
                                        my $set_path = (defined($args{volume})) ? ('/vol/' . $args{volume}) : $path;
					eval
					{
							Dense::dense_toggle(host => $class->{_host},
                                                                            path_name => $set_path,
                                                                            state => $state);
					};                
					$output = $@;
				}
        }
        elsif ($class->{_boot_mode} eq C_ONTAP_CMODE)
        {
			if ($all_volumes == TRUE) {
					my $node_obj = $class->{_filer};
					my $node_name = $node_obj->name();
					my $detected_mroot_vol = 0;
					# Detect the name of the mroot aggregate first
					my $mroot_aggr = $node_obj->server->storage_aggregate_show('has-mroot' => 'true',
                                                                                                   'nodes' => $node_name);
					my $root_aggr_name = $mroot_aggr->[0]->{aggregate};
					logcomment("Mroot aggregate of node $node_name is $root_aggr_name");
					
					# Loop through this node's aggregates
					my @aggregates = $node_obj->Aggregates();
					foreach my $aggregate (@aggregates)
					{
							my @vol_objs = $aggregate->Volumes();
							my $aggr_name = $aggregate->name();
							foreach my $volume (@vol_objs)
							{
									my $volume_name = $volume->name();
									# Skip the root volume, if required
									if (($skip_root_flag == TRUE) && 
										($root_aggr_name eq $aggr_name) && 
										($detected_mroot_vol == 0))
									{
										logcomment("Compression::Sis::set_state: Processing the mroot aggregate.\n" .
												   "Looking for the mroot volume name") if ($class->{_debug});
										
										my $cmd_code = C_CMD_DETECT_ROOT_VOL;
										my $command = $class->{_ui}->formulate_command(code => $cmd_code,
																					   volume => $volume_name);
										
										my $output = Compression::Utils::_send_cmd_to_filer(conn => $node_obj,
																							boot_mode => C_ONTAP_CMODE,
																							cmd => $command,
																							node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
										
										my %processed_hash = $class->{_ui}->process_output_string(code => $cmd_code,
																								  output => $output,
																								  expected_rc => C_RC_FAILURE_NON_FATAL,
																								  log_message => "Obtain dblade-level volume status for $volume_name",);
										
										if ($processed_hash{rc} == C_RC_ROOT_VOLUME_DETECTED)
										{
											$detected_mroot_vol = 1;
											logcomment("Compression::Sis::set_state: Skipping the mroot volume: $volume_name");
											next;
										}
									}

									# Skip the vsroot volumes, if required
									if (($skip_vsroot_flag == TRUE) && 
										($node_obj->server->is_root_volume(volume => $volume_name)))
									{
										logcomment("Compression::Sis::set_state: Skipping vsroot volume: $volume_name");
										next;
									}
									
									# If required, skip volume if offline or restricted
									if ($skip_offline_restricted_flag == TRUE)
									{
										my $vol_state = $volume->status("state");
										if ($vol_state =~ /offline|restricted/)
										{
												logcomment("Compression::Sis::set_state: Skipping $vol_state volume $volume_name");
												next;
										}
									}
									eval
									{											
											if ($state eq C_DEDUPE_STATE_OFF)
											{
													$volume->efficiency_off();
											}
											else
											{
													$volume->efficiency_on(); 
											}
									};
									$output = $@;
									$class->{_ui}->process_output_string(code => C_CMD_DEDUPE_SET_STATE,
										output => $output,
										expected_rc => $args{expected},
										log_message => "Set sis state to $state on path $path");
			
							}
					}
			} else {
                eval
                {
                        my $vol;
                        if (defined($args{volume}))
                        {
                                $vol = NOTE::GX::Volume->attach(name => $args{volume},
                                                                Node => $class->{_volume}->Node);
                        }
                        else
                        {
                             $vol = $class->{_volume}; 
                        }
                        if ($state eq C_DEDUPE_STATE_OFF)
                        {
                                $vol->efficiency_off();
                        }
                        else
                        {
                                $vol->efficiency_on(); 
                        }
                };
                $output = $@;
			}
        }
        
        return $class->{_ui}->process_output_string(code => C_CMD_DEDUPE_SET_STATE,
						    output => $output,
						    expected_rc => $args{expected},
						    log_message => "Set sis state to $state on path $path");
}

## @name ::verify_state
## @summary verifies if sis is enabled or disabled on the volume
##
## @description
## Verifies the state of the sis volume. 
##
## @return Hash with rc and output.
##
## @example
## $dense->verify_state();

sub verify_state
{
        my $class = shift;
        my (%args) = validate( @_,
                {
				state => { type=>SCALAR|ARRAYREF },
                                expected_error => { type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/ } ,
                                partner => { type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/ } }
                          );

	$args{host} = $class->{_host};
        $args{path_name} = $class->{_path};
        my $output;
        my $expected_state = $args{state};
        my $expect_error = $args{expected_error};
        
        if ($class->{_boot_mode} eq C_ONTAP_7MODE)
        {
                $output = Dense::dense_verify_state(\%args);
        }
        else
        {
                # Using server directly for now. sis_status seems to be using the wrong command mapping
                my $disabled_str = $class->{_ui}->get_constant_value(constant_code => 'C_SIS_DISABLED_STR');
                my $res = $class->{_volume}->efficiency_show();
                $output = (defined($res)) ? ($res->{state}) : $disabled_str;
                
                if ($expected_state eq C_DEDUPE_STATE_ENABLED)
                {
                        my $enabled_str = $class->{_ui}->get_constant_value(constant_code => 'C_SIS_ENABLED_STR');
                        if ((($output ne $enabled_str) && ($expect_error == 0)) ||
                            (($output eq $enabled_str) && ($expect_error == 1)))
                        {
                                logresult('fatal', "Unexpected sis state: $output.\nRequired state: $enabled_str, expected error: $expect_error");
                        }
                }
                else # disabled
                {
                        if ((($output ne $disabled_str) && ($expect_error == 0)) ||
                            (($output eq $disabled_str) && ($expect_error == 1)))
                        {
                                logresult('fatal', "Unexpected sis state: $output.\nRequired state: $disabled_str, expected error: $expect_error");
                        }
                }
        }

        return (rc => C_RC_SUCCESS, output => $output);
}

## @name ::get_state
## @summary get state of sis on the volume.
##
## @description
## Get the state of the sis volume. 
## return can be one of the following ENABLED, DISABLED and NOSTATUS
##
## @return Hash with rc and output.
##
## @example
## my $state = $dense->get_state();
sub get_state
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		vol_path => { type => SCALAR, optional => 1, },
	});
	my $vol_path = ( defined $args{vol_path} )? $args{vol_path} : $class->{_path};
	my $output = Dense::dense_get_state(host=>$class->{_host}, path_name=>$vol_path);
	return $output;
}


## @name ::start
## @summary Starts sis on this volume. 
##
## @description
## Start sis on this volume. 
##
## @arg scan Optional: Flag to use the -s -f option with sis start. Default: 0.
##
## @arg nosync Optional: Flag to disable syncing before sis start. Default: 0.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @return Hash with rc and output.
##
## @example
## $dense->start();

sub start
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        scan => { type => BOOLEAN, optional => 1, },
                        nosync => { type => BOOLEAN, optional => 1, },
			expected => {type => SCALAR, regex => qr/^\d+$/, default  => C_RC_SUCCESS, },
		});
	
        $args{host} = $class->{_host};
        $args{path_name} = $class->{_path};
	my $output;
	my $expected = $args{expected};
	delete($args{expected});
        
	if ($class->{_boot_mode} eq C_ONTAP_7MODE)
        {
                eval
                {
                        $output = Dense::dense_start(\%args);
                };
        }
        else
        {
                eval
                {
                        my $scan = ($args{scan}) ? 1 : 0;
                        # Looks like GNOTE is setting snapshot-blocks to true by default. Set to false for backwards compatibility with 7-mode
                        $class->{_volume}->efficiency_start('scan-old-data' => $scan,
                                                            'snapshot-blocks' => 0,
                                                            'background' => 1);
                };
                $output = $@;
        }

	return $class->{_ui}->process_output_string(code => C_CMD_DEDUPE_START,
						    output => $@,
						    expected_rc => $expected,
						    log_message => "Start sis scanner on path " . $args{path_name});
}

## @name ::policy_create
## @summary Create dedup policy.
##
## @description
## Create a dedup policy.
##
## @arg policy, Required: Name of the policy to create, a string of 1 to 32 characters.
##
## @arg schedule, Required: The name of a job schedule.
##                         The value passed here must match the name of an existing job schedule.
##
## @arg duration, Optional: The length of time during which the SE engine should be active before going
##                         back to an inactive state and waiting for the scheduler to next activate it.
##                         The value is an integer number of hours in the range 1 to 999.
##		  Default: 1
##
## @arg enabled, Optional: Whether or not to enable this policy immediately upon creating it. The default
##                Default: true
##
## @arg comment, Optional: A comment regarding this policy. Limited to 255 characters
##
## @return Hash with rc and output.
##
## @example
## $compression->{sis}->policy_create(policy => 'policy_name_1',
##                                schedule => 'hourly',
##                                duration => 1,
##                                comment => "\"Policy for hourly schedule\"");
sub policy_create{
        my $class = shift;
        my (%args) = validate( @_, {
                        policy   => { type => SCALAR},
                        schedule => { type => BOOLEAN },
                        duration => { type => SCALAR,  optional => 1, default  => 1 },
                        enabled  => { type => BOOLEAN, optional => 1, default  => TRUE },
                        comment  => { type => SCALAR,  optional => 1},

                });

        if ($class->{_boot_mode} eq C_ONTAP_CMODE) {
                $class->{_volume}->efficiency_policy_create(%args);
        } else {
                logresult('script', "policy_create method is for CMODE only");
        }

        return (rc => C_RC_SUCCESS);
}

## @name ::policy_delete
## @summary Deletea dedup policy.
##
## @description
## Delete a dedup policy.
##
## @arg policy, Required: Name of the policy to be deleted
##
## @return Hash with rc.
##
## @example
## $compression->{sis}->policy_delete(policy => 'policy_name_1');
##
sub policy_delete{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        policy   => { type => SCALAR},
                });

        if ($class->{_boot_mode} eq C_ONTAP_CMODE) {
                $class->{_volume}->efficiency_policy_delete(%args);
        } else {
                logresult('script', "policy_delete method is for CMODE only");
        }

        return (rc => C_RC_SUCCESS);
}

## @name ::policy_show
## @summary Show dedup policy.
##
## @description
## Show a dedup policy.
##
## @arg policy, Optional: Name of the policy to be shown
##
## @return Policy hash ref.
##
## @example
## $compression->{sis}->policy_show(policy => 'policy_name_1');
##
sub policy_show{
        my $class = shift;
        my (%args) = validate( @_, {
                        policy   => { type => SCALAR, optional=>1,},
                });

        my $res;
        if ($class->{_boot_mode} eq C_ONTAP_CMODE) {
                $res = $class->{_volume}->efficiency_policy_show(%args);
        } else {
                logresult('script', "policy_show method is for CMODE only");
        }

        return $res;
}

## @name ::verify_efficiency_status
## @summary verify_efficiency_status dedup status.
##
## @description
## verify_efficiency_status dedup status.
##
## @arg name, Required: Name of the efficiency status item
##
## @arg expected_string, Required: The value of efficiency status item
##
## @return efficiency status hash ref.
## The hash is in format :
##   fingerprint => {
##              'fingerprint_vvol' => {  ## volume FPDB
##                             'blocks' => '400',
##                             'files' => '5.',
##                             'used' => '396',
##                             'size' => '1572KB'
##                             },
##              'fingerprint' => {    ## aggregate FPDB
##                             'blocks' => '395',
##                             'files' => '1.',
##                             'used' => '395',
##                             'size' => '1572KB'
##                             } };
##
## @example
## $compression->{sis}->verify_efficiency_status(name=>'policy', expected_string => '-');
##
sub verify_efficiency_status{
        my $class = shift;
        my (%args) = validate( @_, {
                                name => { type=>SCALAR },
                                expected_string => { type=>SCALAR, optional=>1, } }
                          );

        my $res;
        if ($class->{_boot_mode} eq C_ONTAP_CMODE) {

                $res = $class->{_volume}->efficiency_show();
                if ($res->{$args{name}} =~ /^$args{expected_string}$/){
                        return (rc => C_RC_SUCCESS);
                } else {
                        logresult('fatal', "Storage efficiency state $res->{$args{name}} does not match the expected: \nRequired state: $args{expected_string}, Expected error: $args{expect_error}");
                }
        } else {
                logresult('script', "verify_efficiency_status method is for CMODE only");
        }

}

## @name ::get_fingerprint_db
## @summary Get fingerprint db.
##
## @description
## Get "sis ls /vol/vol_name" output, parse it and return fingerprint status.
##
## @arg volume, Optional: volume name, default: object volume name
##
## @arg vserver, Optional: vserver name, default: object vserver name
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @return fingerprint DB data hash.
##
## @example
## $compression->{sis}->get_fingerprint_db();
##
sub get_fingerprint_db{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        volume => { type => SCALAR, optional => 1, default => $class->{_vol_name} },
                        vserver => { type => SCALAR|UNDEF, optional => 1, default => $class->{_vserver} },
                        expected => { type => SCALAR, default => C_RC_SUCCESS },
                } );

        my $cmd_code = C_CMD_VERIFY_SIS_LS;
        my $volume = $args{volume};

        my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                       volume => $volume,
                                                       vserver => $args{vserver});

        my $conn = $class->{_filer};

        my $output = Compression::Utils::_send_cmd_to_filer(conn => $conn,
                                                            boot_mode => $class->{_boot_mode},
                                                            cmd => $command,
                                                            node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),
                                                            vfiler => $class->{_vfiler},
                                                            vfiler_available => $class->{_ui}->is_vfiler_available_command(code => $cmd_code),);

        my %ret = $class->{_ui}->process_output_string(code => C_CMD_VERIFY_SIS_LS,
                                                       output => $output,
                                                       expected_rc => C_RC_SUCCESS,
                                                       log_message => "get fingerprint hash with sis ls for $volume",);

        logcomment("Convert Aggregate and volume fingerprint sizes into bytes");
	my %result_hash;
        if ($ret{fingerprints}->{'fingerprint.vvol'}){
                %result_hash = process_data_size_string( size_string => $ret{fingerprints}->{'fingerprint.vvol'}->{size} );
                $ret{fingerprints}->{'fingerprint.vvol'}->{size} = $result_hash{size_in_bytes};
        }

        if ($ret{fingerprints}->{'fingerprint'}){
                %result_hash = process_data_size_string( size_string => $ret{fingerprints}->{fingerprint}->{size} );
                $ret{fingerprints}->{'fingerprint'}->{size} = $result_hash{size_in_bytes};
        }
        return %ret;
}

## @name ::verify_fingerprint_db
## @summary Verify fingerprint db
##
## @description
## Verify fingerprint db for aggregate FPDB and/or volume FPDB
##
## @arg fingerprint, Required: fingerprint DB to compare with
##
## @arg aggregate, Optional: aggregate FPDB exists and check, default: TRUE
##
## @arg volume, Optional: volume FPDB exists and check, default: TRUE
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @return fingerprint DB data hash.
##
## @example
## $compression->{sis}->verify_fingerprint_db(fingerprint => \%fingerprints, aggregate=>FALSE);
##
sub verify_fingerprint_db{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        fingerprint => { type => SCALAR|HASHREF} ,
                        aggregate  =>{ type => BOOLEAN, optional=>1, default => TRUE },
                        volume  =>{ type => BOOLEAN, optional=>1, default => TRUE },
                        expected => { type => SCALAR, optional=>1, default => C_RC_SUCCESS },
                } );
        my %ret = $class->get_fingerprint_db();

        if ( $args{aggregate} && $args{volume} ) {  # Check both aggregate and volume FPDB are exists
             if ( ($ret{fingerprints}->{fingerprint}->{size} ne 0)  &&
                  ($ret{fingerprints}->{fingerprint}->{blocks_used} ne 0) &&
                  ($args{fingerprint}->{fingerprints}->{'fingerprint.vvol'}->{size} eq $ret{fingerprints}->{'fingerprint.vvol'}->{size}) &&
                  ($args{fingerprint}->{fingerprints}->{'fingerprint.vvol'}->{blocks_used} eq $ret{fingerprints}->{'fingerprint.vvol'}->{blocks_used})   ) {
                        return (rc => C_RC_SUCCESS);

                }
        } elsif ( !$args{aggregate}  &&  $args{volume} ){  ## aggregate FPDB not exist, Check volume FPDB
                if ( (!$ret{fingerprints}->{fingerprint} )  &&
                     ($args{fingerprint}->{fingerprints}->{'fingerprint.vvol'}->{size} eq $ret{fingerprints}->{'fingerprint.vvol'}->{size}) &&
                     ($args{fingerprint}->{fingerprints}->{'fingerprint.vvol'}->{blocks_used} eq $ret{fingerprints}->{'fingerprint.vvol'}->{blocks_used})   ) {
                        return (rc => C_RC_SUCCESS);

                }
        } elsif ( $args{aggregate}  &&  !$args{volume} ){ ## volume FPDB not exist, Check aggregate FPDB
                if ( ( $ret{fingerprints}->{fingerprint}->{size} ne 0)  &&
                     ( $ret{fingerprints}->{fingerprint}->{blocks_used} ne 0)  &&
                     ( !$ret{fingerprints}->{'fingerprint.vvol'} )    ) {
                        return (rc => C_RC_SUCCESS);

                }
        } else {  ## no FPDB exists
                if ( (!$ret{fingerprints}->{'fingerprint.vvol'} ) &&
                     (!$ret{fingerprints}->{'fingerprint'} )   ) {
                        return (rc => C_RC_SUCCESS);

                }
        }
        logresult('fatal', "Checking fingerprint DB dismatch\n" .Dumper(%ret)) unless ($args{expected} eq C_RC_FAILURE );

}

## @name ::wait_for
## @summary Waits for a specific sis status or state. 
##
## @description
## Waits for a specific sis status or state. 
##
## @arg wait_on Optional: Area to wait for (C_DEDUPE_STAGE or C_DEDUPE_STATUS, defined in Constants.pm).
## Default: C_DEDUPE_STATUS.
##
## @arg wait_for Optional: What to wait for (defined in Constants.pm):
## For C_DEDUPE_STATUS: C_DEDUPE_STATUS_INITIALIZING, C_DEDUPE_STATUS_IDLE, C_DEDUPE_STATUS_ACTIVE,
##                      C_DEDUPE_STATUS_UNDOING, C_DEDUPE_STATUS_NOSTATUS, C_DEDUPE_STATUS_DONE, C_DEDUPE_STATUS_DASH.
## For C_DEDUPE_STAGE: C_DEDUPE_STAGE_GATHING, C_DEDUPE_STAGE_SORTING, C_DEDUPE_STAGE_CHECK_P1,
##                     C_DEDUPE_STAGE_CHECK_P2, C_DEDUPE_STAGE_SAVING, C_DEDUPE_STAGE_DONE, C_DEDUPE_STAGE_DASH.
##
## @arg time_out Optional: Timeout value.
##
## @arg time_gap Optional: Time interval between two consecutive polls.
##
## @arg expected_error Optional: Error string to check on "Current Operation Error". Default is "-".
##
## @return Hash with rc and output.
##
## @example
## $dense->wait_for();
## $dense->wait_for(wait_on => C_DEDUPE_STATUS, wait_for => C_DEDUPE_STATUS_ACTIVE);

sub wait_for
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        wait_on => { type => SCALAR, optional => 1, default => C_DEDUPE_STATUS},
                        wait_for => { type => SCALAR, optional => 1, default => C_DEDUPE_STATUS_IDLE },
                        time_out => { type => SCALAR, regex => qr/^\d+$/, optional => 1, default => C_DFLT_COMPRESS_TIMEOUT, },
                        time_gap => { type=>SCALAR, optional=>1, },
                        expected_error => { type=>SCALAR, optional=>1, },
		});
        
        $args{host} = $class->{_host};
        $args{path_name} = $class->{_path};
        my $output = '';
        
        if ($class->{_boot_mode} eq C_ONTAP_7MODE)
        {
                $output = Dense::dense_wait_for(\%args);
        }
        else
        {
                my $wait_on = $args{wait_on};
                my $wait_for = $args{wait_for};
                
                if (($wait_on ne C_DEDUPE_STATUS) || ($args{expected_error}))
                {
                        logresult('script', "Sis\:\:wait_for support for C-mode is only provided for:\n" .
                                            "wait_on => C_DEDUPE_STATUS\n" .
                                            "expected_error => 0 or not defined");
                }
                
                my $timeout = C_DFLT_COMPRESS_TIMEOUT; 
                my $wanted_status = $class->{_ui}->get_constant_value(constant_code => $wait_for);
                my $no_status = $class->{_ui}->get_constant_value(constant_code => C_DEDUPE_STATUS_NOSTATUS);
                my $sis_status = $class->{_volume}->efficiency_show();
                my $current_status = (defined($sis_status)) ? ($sis_status->{'op-status'}) : $no_status;
                logcomment("Sis scanner status: $current_status");
                while (($timeout > 0) && ($current_status !~ /$wanted_status/i))
                {
                        sleep(1);
                        $sis_status = $class->{_volume}->efficiency_show();
                        $current_status = (defined($sis_status)) ? ($sis_status->{'op-status'}) : $no_status;
                        logcomment("Sis scanner status: $current_status");
                        $timeout--;
                }
                
                if ($current_status !~ /$wanted_status/i)
                {
                        logresult('fatal', "Timed-out while waiting for sis status \"$wanted_status\"\n" .
                                           "Last seen status: $current_status")
                }
        }
        
        return (rc => C_RC_SUCCESS, output => $output);
}


## @name ::stop
## @summary Abort the de-duplication on this volume.
##
## @description
## Abort the de-duplication process or inflation process on path. 
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @arg partner Optional: Default: 0.
##
## @return Hash with rc and output.
##
## @example
## $dense->stop();

sub stop
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                       partner => { type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/ },
		       expected => {type => SCALAR, regex => qr/^\d+$/, default  => C_RC_SUCCESS, },
                } );

        $args{host} = $class->{_host};
        $args{path_name} = $class->{_path};
	my $expected = $args{expected};
	delete($args{expected});
	my $output;
	
	if ($class->{_boot_mode} eq C_ONTAP_7MODE)
        {
                eval
                {
                        $output = Dense::dense_stop(\%args);
                };
        }
        else
        {
                eval
                {
                        $class->{_volume}->efficiency_stop();
                };
                
                $output = $@;
	}

        return $class->{_ui}->process_output_string(code => C_CMD_DEDUPE_STOP,
						    output => $@,
						    expected_rc => $expected,
						    log_message => "Stop sis scanner on path " . $args{path_name});
}

## @name ::off_undo
## @summary Converts dense volume into normal volume.
##
## @description
## The dedup operation on volume is unconfigured, and all the data sharings in the volume are reverted. 
##
##
## @arg force Optional: Set to 1 will force sis undo to run even when volume is full, debug only option, defaults to 0.
##
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @arg expected_sis_off Optional: Expected return code when executing set_state (defined in Compression::Constants,).
## If not provided, assumes the command should complete successfully.
##
## @return Hash with rc and output.
##
## @example
## $dense->off_undo();

sub off_undo
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                force => { type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/ },
		expected => {type => SCALAR, regex => qr/^\d+$/, default  => C_RC_SUCCESS, },
		expected_sis_off => {type => SCALAR, regex => qr/^\d+$/, default  => C_RC_SUCCESS, },
                });

        $args{host} = $class->{_host};
        $args{path_name} = $class->{_path};
	my $expected = delete($args{expected});
	my $expected_sis_off = delete($args{expected_sis_off});
        my $output;
        
        if ($class->{_boot_mode} eq C_ONTAP_7MODE)
        {
		eval
                {
			$output = Dense::dense_off_undo(\%args);
		};
        }
        else
        {
		eval
		{
			$class->set_state(state => C_DEDUPE_STATE_OFF, expected => $expected_sis_off);
			$class->{_volume}->efficiency_undo(background => 1);
		};
		$output = $@;
        }
		$output = $@;
	return $class->{_ui}->process_output_string(code => C_CMD_DEDUPE_OFF_UNDO,
						    output => $output,
						    expected_rc => $expected,
						    log_message => "Start sis undo on path " . $args{path_name});
}

## @name ::config
## @summary Configures a schedule of de-duplication of a volume
##
## @description
## Used to configure the schedule of de-duplication process of a volume.
##
## @arg schedule_string Optional:  string of schedule if you do not want to spcify in day, hour or minute options (requested). If this is not null, day, hour, munite, time order, ignore_undef are ignored all together, only the schedule_string will be passed. Default is null.
##
## @arg policy Optional: policy name (This argument is for CMODE only)
##
## @return Hash with rc and output.
##
## @example
## $dense->config();
## $compression->{sis}->config(policy=>$policy_name);

sub config
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                 schedule_string => { type=>SCALAR | UNDEF, default=>"-", optional=>1 },
		 policy => { type=>SCALAR|UNDEF, optional=>1 },
                });

        $args{host} = $class->{_host};
        $args{path_name} = $class->{_path};
        my ($output, $policy, $schedule);
	##Only one of policy and schedule arguments can appear in efficiency config
        if (defined $args{policy}){
               $policy = 'policy';
               $schedule = '';
	       if (defined $args{schedule_string}){
		  logcomment("Only one of policy and schedule arguments can appear in efficiency config.
			      Deleting schedule_string parameter");
		  delete $args{schedule_string};
	       }
        } else {
               $schedule = 'schedule';
        }

        if ($class->{_boot_mode} eq C_ONTAP_7MODE)
        {
		if ( defined $args{policy} ){
		   logresult('script', "policy is not a 7Mode parameter");
		}
                $output = Dense::dense_config(\%args);
        }
        else
        {
                $class->{_volume}->server->volume_efficiency_config($schedule => $args{schedule_string},
                                                                    volume => $class->{_vol_name},
                                                                    vserver => $class->{_vserver},
								    $policy => $args{policy},);
        }

        return (rc => C_RC_SUCCESS, output => $output);
}


## @name ::get_refcount_hash
## @summary Gets the refcount hash for a specific file. 
##
## @description
## Gets the refcount hash for a specific file. 
##
## @arg filename Required: Name of file (not containing the volume mount path).
##
## @arg fbn_range Optional: Range of fbns to obtain the refcounts for (single fbn, range (ex: 5:10), or empty string (for all fbns)).
##
## @return Hash with rc and refcount.
##
## @example
## $dense->get_refcount_hash(filename => $file_name);

sub get_refcount_hash
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        filename => { type=>SCALAR },
                        fbn_range => { type=>SCALAR, optional=>1 },
		});
        
        $args{file_name} = $class->{_path} . '/' . $args{filename};
        delete($args{filename});
        my $res = {};
        
        if ($class->{_boot_mode} eq C_ONTAP_7MODE)
        {
                $args{conn} = $class->{_filer}->conn();
                my %dense_vbn_hash = Dense::Utils::dense_parse_showvbns(\%args);
                logcomment("dense_parse_showvbns returned:\n" . Dumper(%dense_vbn_hash)) if ($class->{_debug});
                
                # Keep only the refcounts. Vbn/vvbns available through dataset
                foreach my $key (keys %dense_vbn_hash)
                {
                        $res->{$key} = $dense_vbn_hash{$key}{refcount};
                }
        }
        else
        {
                my $cmd_code = C_CMD_GET_REFCOUNT;
                my $filepath = $args{file_name};
                
                my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                               file_path => '/vol' . $filepath);
                
                my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_filer},
                                                                    boot_mode => $class->{_boot_mode},
                                                                    cmd => $command,
                                                                    node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
                
                my %refcount_info = $class->{_ui}->process_output_string(code => $cmd_code,
                                                                         output => $output,
                                                                         expected_rc => C_RC_SUCCESS,
                                                                         log_message => "Obtain vbn info for file $filepath",);
                
                my $refcount_hash = $refcount_info{refcount_info};
                foreach my $fbn (keys %$refcount_hash)
                {
                        $res->{$fbn} = $refcount_hash->{$fbn}->{refcount};
                }
        }
        
        return (rc => C_RC_SUCCESS, refcount => $res);
}


## @name ::check
## @summary Verifies and updates the fingerprint database for this volume. 
##
## @description
## Verifies and updates the fingerprint database for this volume. 
##
## @arg count Optional: Flag to use the -c option with sis check. Default: FALSE.
##
## @arg values Optional: Flag to use the -v option with sis check. Default: FALSE.
##
## @arg delete_stale Optional: Flag to use the -d option with sis check. Default: FALSE.
##
## @return Hash with rc and output.
##
## @example
## $dense->check();

sub check
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        count => { type => BOOLEAN, optional => 1, default => FALSE },
                        values => { type => BOOLEAN, optional => 1, default => FALSE },
                        delete_stale => { type => BOOLEAN, optional => 1, default => FALSE },
		});
        
        my $output;
        
        if ($class->{_boot_mode} eq C_ONTAP_7MODE)
        {
                my $cmd_options = '';
                if ($args{count})
                {
                        $cmd_options .= '-c ';
                }
                if ($args{values})
                {
                        $cmd_options .= '-v ';
                }
                if ($args{delete_stale})
                {
                        $cmd_options .= '-d ';
                }
                
                $output = Dense::dense_check(host => $class->{_host},
                                             path_name => $class->{_path},
                                             cmd_opt => $cmd_options);
        }
        else
        {
                # Note: for 7-mode, Dense.pm does not currently do any kind of output parsing
                # It just checks that the command executed with no errors
                # Doing the same here for c-mode
                # If output parsing is needed, then this is done for RR and after through Compress.pm
                my $cmd_code = C_CMD_SIS_CHECK;
                my $volume = $class->{_vol_name};
                
                my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
                                                               volume => $volume,
                                                               options => \%args);
                
                $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_filer},boot_mode => $class->{_boot_mode},
                                                                 cmd => $command,
                                                                 node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
                
                $class->{_ui}->process_output_string(code => $cmd_code,
                                                     output => $output,
                                                     expected_rc => C_RC_SUCCESS,
                                                     log_message => "Sis check for volume $volume",);
        }

        return (rc => C_RC_SUCCESS, output => $output);
}


################################################################
# Internal subroutines
################################################################

sub _update_volume_path
{
        my $class = shift;
	my (%args) = validate( @_,
	{
		new_path => { type => SCALAR, },
	} );
        
        $class->{_path} = $args{new_path};
        
        logcomment("Sis: class path was changed to: " . $class->{_path}) if $class->{_debug};
}

sub _update_conn
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		new_filer => { type => OBJECT, },
	} );
	
	$class->{_filer} = $args{new_filer};
	
	if ($class->{_boot_mode} eq C_ONTAP_7MODE)
        {
		if (!($class->{_filer}->isa("NOTE::Filer")))
		{
			logresult('script', "For 7-mode, the conn argument is required to be a NOTE\:\:Filer object");
		}
		$class->{_host} = $class->{_filer}->host();
        }
	elsif ($class->{_boot_mode} eq C_ONTAP_CMODE)
        {
		
		logresult('script', "This method is not supported in c-mode")
        }
}


# We do not need to overload the destructor here.
# If cleanup is not disabled for the compress object,
# Then sis cleanup would be taken care of when destroying the volume,
# by Volume::vol_offline


1;
