#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/SnapMirror.pm#1 $
#
## @summary Compression SnapMirror routines
##  
##
## @description
## Wrapper around NOTE::SnapMirror.
## The main reason for not using NOTE::SnapMirror directly is to provide standardized output
##    and to enable specific processing on some failures, otherwise fatal.
##    This is required for handling compression-specific failures.
##
## @synopsis
## use Compression::SnapMirror;
## use vars qw( $FILER  );
##
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $sm = Compression::SnapMirror->initialize(src=>$src_vol, dst=>$dst_vol);
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
## @status private (in-progress)

package Compression::SnapMirror;

use strict;
use warnings;
use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult process_performance_info);
use Carp qw(croak);

BEGIN
{
	require "Compression/base_import.pl";

}


## @name:: initialize
## @summary initialize a SnapMirror relationship
##	
##
## @description
##	initialize a SnapMirror relationship
##
## @return Object of type Compression::SnapMirror.
##
## @arg src Required: source volume or qtree object (Compression::Compress or Compress::Qtree)
## @arg dst Required: destination volume or qtree object (Compression::Compress or Compress::Qtree)
## @arg debug Optional: Controls verbosity of this module (defaults to FALSE)
##
## @example
## my $obj = Compression::SnapMirror->initialize (	src => $src,
##											dst => $dst,
##											debug => $args{debug} );
##
## Note: this can be called with no args to reinitialize the snapmirror
##
sub initialize
{
	my $class 	= shift;
	my (%args) = validate (@_, {src => { type=>OBJECT, optional => 1,  },
				    dst => { type=>OBJECT, optional => 1,  },
				    debug => { type=>SCALAR, optional => 1, default=>FALSE },
				    expected => { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },
	} );

	if ((! defined $args{src} ) && (! defined $args{dst} ) ) {
		logcomment("Re-initializing the existing snapmirror relationship.");
		$class->{_snapmirror}->initialize();
		$class->_update_dataset_hash();
		return;
	} else {	
	
		my $src = $args{src};
		my $dst = $args{dst};
		
		my $src_obj;
		my $src_compr_obj;
		if ( $src->isa("Compression::Qtree") ) {
			$src_obj = $src->{_qtree};
			$src_compr_obj = $src->{_compress};
		} elsif ($src->isa("Compression::Compress") ) {
			$src_obj = $src->{volume}->_get_underlying_vol_obj();
			$src_compr_obj = $src;
		} else {
			croak "unable to get src object $!";
		}

		my $dst_obj;
		if ( $dst->isa("Compression::Qtree") ) {
			$dst_obj = $dst->{_qtree};
		} elsif ( $dst->isa("Compression::Compress") ) {
			$dst_obj = $dst->{volume}->_get_underlying_vol_obj();
		} else {
			croak "unable to get dst object $!";
		}
	
		if ($src_compr_obj->{boot_mode} eq C_ONTAP_7MODE) {
			$src_obj->Filer->say("options snapmirror.access all");
			$dst_obj->Filer->say("options snapmirror.access all");
		}
		
		my $self = {};
		$self->{ID} 	= $$;
		$self->{_debug}	= $args{debug};
		# This assumes that the command will execute on the source filer
		# or that both src and dst are running the same version.
		$self->{_ui} = $src_compr_obj->{_ui};
		
		
		my $sm_obj;
		
		eval {
			$sm_obj = NOTE::SnapMirror->initialize(src=>$src_obj, dst=>$dst_obj );
		};
		
		if ($@) {
			$src_compr_obj->{_ui}->process_output_string(code => C_CMD_SNAPMIRROR_INITIALIZE,
								     output => $@,
								     expected_rc => $args{expected},
								     log_message => "SnapMirror initialize",);
			return undef;
		}
		
		if (!defined($sm_obj)) {
			logresult('fatal', "Undefined object returned by NOTE\:\:SnapMirror\:\:initialize");
		}
		
		$self->{_src} = $src;
		$self->{_dst} = $dst;

		$self->{_snapmirror} = $sm_obj;

		bless($self, $class);

		_update_dataset_hash($self);

		logcomment("SnapMirror: Setup Complete..\n\n") if ($self->{_debug});
		return $self;

	}

}

## @name:: update
## @summary SnapMirror update
##	
##
## @description
##	Starts an incremental transfer for a SnapMirror relationship
##
## @return Hash with rc and output.
##
## @arg background Optional: if TRUE, runs update in the background
##
## @example
## $obj->update();
## or
## $obj->update(background=>TRUE);
## $obj->wait_for_update();

sub update
{
	my $class 	= shift;
	my (%args) = validate (@_, { 
		background => { type=>SCALAR, optional => 1, default=>FALSE },
		throttle => { type=>SCALAR, optional => 1 },
	} );
	# todo: fix background, currently it results in an error:
	# Transfer failed: replication transfer failed to complete
	my %res;
	$res{'rc'} = C_RC_SUCCESS;
	my $throttle = $args{'throttle'};

	my %params;
	foreach my $arg (sort keys %args) {
		next if $arg eq 'background';
		if ( defined $args{$arg} ) {
			$params{$arg} = $args{$arg};
		}
	}

	my $src_path = $class->{_snapmirror}->src_path();
	my $dst_path = $class->{_snapmirror}->dst_path();

	my $dst_compr_obj;
	if ($class->{_dst}->isa("Compression::Qtree")) {
		$dst_compr_obj = $class->{_dst}->{_compress};
	} elsif ($class->{_dst}->isa("Compression::Compress")) {
		$dst_compr_obj = $class->{_dst};
	} else {
		croak "unknown dst object $!";
	}


	my $cli;
	if ($throttle) {
		$cli = "snapmirror update -S $src_path -k $throttle $dst_path";
	} else {
		$cli = "snapmirror update -S $src_path $dst_path";
	}

	if ($args{'background'} == TRUE ) {
		$class->{update_subtest} = subtest( sub{
			my $vol_obj = $dst_compr_obj->{volume}->_get_underlying_vol_obj();
			say($vol_obj->Filer->conn, "priv set -q diag; $cli" );
		},
				'-bg',
				'-runid','snapmirror_update',							
		);
	} else {
		$class->{_snapmirror}->update(%params);
	}

	$class->_update_dataset_hash();

	return %res;
}



## @name:: wait_for_update
## @summary wait for SnapMirror update
##	
##
## @description
##	If update was done in the background, wait for it to complete.
##
## @return Hash with rc and output.
##
## @example
## $obj->update(background=>TRUE);
## $obj->wait_for_update();

sub wait_for_update
{
	my $class 	= shift;
	my (%args) = validate (@_, { } );

	my %res;
	$res{'rc'} = C_RC_SUCCESS;

	$class->{update_subtest}->wait();

	return %res;
}

## @name:: break
## @summary SnapMirror break
##	
##
## @description
##	Break SnapMirror relationship
##
## @return Hash with rc and output.
##
## @example
## $obj->break();

sub break
{
	my $class 	= shift;
	my (%args) = validate (@_, { } );
	my %res;
	$res{'rc'} = C_RC_SUCCESS;
	$class->{_snapmirror}->break();
	return %res;
}

## @name:: release
## @summary SnapMirror release
##	
##
## @description
##	Release SnapMirror relationship
##
## @return Hash with rc and output.
##
## @example
## $obj->release();

sub release
{
	my $class 	= shift;
	my (%args) = validate (@_, { } );
	my %res;
	$res{'rc'} = C_RC_SUCCESS;
	$class->{_snapmirror}->release();
	return %res;
}


## @name:: quiesce
## @summary SnapMirror quiesce
##	
##
## @description
##	Quiesce SnapMirror
##
## @return Hash with rc and output.
##
## @example
## $obj->quiesce();
sub quiesce
{
	my $class 	= shift;
	my (%args) = validate (@_, { } );
	my %res;
	$res{'rc'} = C_RC_SUCCESS;
	$class->{_snapmirror}->quiesce();
	return %res;
}

## @name:: resume
## @summary SnapMirror resume
##	
##
## @description
##	Resumes quiesced transfers.
##
## @return Hash with rc and output.
##
## @example
## $obj->resume();
sub resume
{
	my $class 	= shift;
	my (%args) = validate (@_, { } );
	my %res;
	$res{'rc'} = C_RC_SUCCESS;
	$class->{_snapmirror}->resume();
	return %res;
}


## @name:: abort
## @summary SnapMirror abort
##	
##
## @description
##	snapmirror abort.
##
## @return Hash with rc and output.
##
## @example
## $obj->abort();
sub abort
{
	my $class 	= shift;
	my (%args) = validate (@_, { } );
	my %res;
	$res{'rc'} = C_RC_SUCCESS;
	$class->{_snapmirror}->abort();
	return %res;
}


## @name:: resync
## @summary SnapMirror resync
##	
##
## @description
##	snapmirror resync.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @return Hash with rc and output.
##
## @example
## $obj->resync();
sub resync
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		src => { type=>OBJECT, optional => 1,  },
		dst => { type=>OBJECT, optional => 1,  },
		expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
	} );
	
	my %res;
	$res{'rc'} = C_RC_SUCCESS;
	
	my $expected = $args{expected};
	my $ui;

	# This should be done in a nicer way...
	if ((! defined $args{src} ) && (! defined $args{dst} ) ) {
		eval {
			$class->{_snapmirror}->resync();
		};
		
		$ui = $class->{_ui};
		%res = $ui->process_output_string(code => C_CMD_SNAPMIRROR_RESYNC,
						  output => $@,
						  expected_rc => $expected,
						  log_message => "Snapmirror resync",
						  );
		
		if ($expected == C_RC_SUCCESS) {
			$class->_update_dataset_hash();
		}
		return %res;
	} else {	

		my $src = $args{src};
		my $dst = $args{dst};

		my $src_obj;
		if ( $src->isa("Compression::Qtree") ) {
			$src_obj = $src->{_qtree};
			$ui = $src->{_compress}->{_ui};
			logcomment("qtree: ui:\n" . Dumper($ui));
		} elsif ($src->isa("Compression::Compress") ) {
			$src_obj = $src->{volume}->_get_underlying_vol_obj();
			$ui = $src->{_ui};
			logcomment("compress: ui:\n" . Dumper($ui));
		} else {
			croak "unable to get src object $!";
		}

		my $dst_obj;
		if ( $dst->isa("Compression::Qtree") ) {
			$dst_obj = $dst->{_qtree};
		} elsif ( $dst->isa("Compression::Compress") ) {
			$dst_obj = $dst->{volume}->_get_underlying_vol_obj();
		} else {
			croak "unable to get dst object $!";
		}
		
		eval {
			$class->{_snapmirror}->resync(src=>$src_obj, dst=>$dst_obj );
		};
		
		if (!defined($ui)) {croak "was not able to init ui";}
		%res = $ui->process_output_string(code => C_CMD_SNAPMIRROR_RESYNC,
					   output => $@,
					   expected_rc => $expected,
					   log_message => "Snapmirror resync",
					   );

		if ($expected == C_RC_SUCCESS) {
			$class->{_src} = $src;
			$class->{_dst} = $dst;

			$class->_update_dataset_hash();
		}
	}
	
	return %res;
}

## @name:: check
## @summary SnapMirror check
##	
##
## @description
##	Run SnapMirror checker
##
## @return Hash with rc and output.
##
## @example
## $obj->check();
sub check
{
	my $class 	= shift;
	my (%args) = validate (@_, { } );
	my %res;
	$res{'rc'} = C_RC_SUCCESS;
	$class->{_snapmirror}->check();
	return %res;
}


## @name:: treecompare
## @summary treecompare
##	
##
## @description
##	Run treecompare
##
## @return Hash with rc and output.
##
## @example
## $obj->treecompare();
sub treecompare
{
	my $class 	= shift;
	my (%args) = validate (@_, { } );
	my %res;
	$res{'rc'} = C_RC_SUCCESS;
	$class->{_snapmirror}->compare(treecompare=>1);
	return %res;
}

## @name:: status_wait
## @summary SnapMirror status_wait
##	
## @description
##	Wait for SnapMirror to be a certain status.
##
## @return Hash with rc and output.
##
## @example
## $obj->status_wait();
sub status_wait {
	my $class 	= shift;
	my (%args) = validate (@_, {
			status => { type=>SCALAR, optional=>1, default=>'idle' },
			ignore_transfer_error => { type=>SCALAR, optional=>1, default=>FALSE },
	} );
	my $status = $args{'status'};
	my $ignore_transfer_error = $args{'ignore_transfer_error'};
	$class->{_snapmirror}->status_wait(status => $status, ignore_transfer_error=>$ignore_transfer_error);
}


## @name:: wait_sync
## @summary SnapMirror wait_sync
##	
## @description
##	Wait for SnapMirror to be synced.
##
## @return Hash with rc and output.
##
## @example
## $obj->wait_sync();
sub wait_sync {
	my $class 	= shift;
	my (%args) = validate (@_, {} );
	$class->{_snapmirror}->wait_sync();
}


## @name:: check_performance
## @summary Check SnapMirror performance.
##	
## @description
## Check SnapMirror performance. Obtain the last-transfer-duration and the amount of data transfered
## from the snapmirror status and calculate the throughput.
##
## @arg max_allowed_duration Optional: Maximum value allowed for the transfer duration (s).
## If this argument is used, the obtained duration is compared against it and
## if it is bigger, then 'failure_handling' is executed.
##
## @arg min_allowed_throughput Optional: Minimum value allowed for the throughput (KB/s).
## If this argument is used, the obtained throughput is compared against it and
## if it is smaller, then 'failure_handling' is executed.
##
## @arg failure_handling Optional: Flag to determine what action should be taken if
## the throughput was detected to be smaller than min_allowed_throughput.
## This argument is ignored when none of min_allowed_throughput or max_allowed_duration are passed in.
## Allowed values: none, fail, panic, panic_and_fail. Default: fail.
##
## @return Hash with rc and duration.
##
## @example
## $obj->check_performance();
## $obj->check_performance(max_allowed_duration => 10,);
## $obj->check_performance(min_allowed_throughput => 50,
##                         fail_action => 'panic');
##
sub check_performance
{
	my $class = shift;
	my (%args) = validate (@_,
	{
		max_allowed_duration => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
		min_allowed_throughput => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
		failure_handling => {type => SCALAR, regex => qr/^none|panic|fail|panic_and_fail$/, default => 'fail',}, 
	});
	
	my $status = $class->{_snapmirror}->status();
	
	my $last_transfer_duration = $status->{'last-transfer-duration'};
	my $last_transfer_size = $status->{'last-transfer-size'};
	my $last_transfer_type = (exists($status->{'last-transfer-type'})) ? $status->{'last-transfer-type'} : 'unknown';
	
	my $src = $class->{_snapmirror}->src->Filer->host();
	my $dst = $class->{_snapmirror}->dst->Filer->host();
	my $filers_to_panic = ($src eq $dst) ? [$src] : [$src, $dst];
	return Compression::Utils::process_performance_info(duration => $last_transfer_duration,
							    description => "SnapMirror ($last_transfer_type)",
							    boot_mode => C_ONTAP_7MODE,
							    max_allowed_duration => $args{max_allowed_duration},
							    min_allowed_throughput => $args{min_allowed_throughput},
							    processed_data_size => $last_transfer_size,
							    failure_handling => $args{failure_handling},
							    generate_cores_on => $filers_to_panic,);
}


################################
# Internal methods
################################

sub _update_dataset_hash {
	my $class 	= shift;
	my (%args) = validate (@_, {} );

	my $src = $class->{_src};
	my $dst = $class->{_dst};	

	my $src_compr_obj;
	my $dst_compr_obj;

	if ($src->isa("Compression::Qtree")) {
		$src_compr_obj = $src->{_compress};
	} elsif ($src->isa("Compression::Compress")) {
		$src_compr_obj = $src;
	} else {
		croak "unknown src object $!";
	}

	if ($dst->isa("Compression::Qtree")) {
		$dst_compr_obj = $dst->{_compress};
	} elsif ($dst->isa("Compression::Compress")) {
		$dst_compr_obj = $dst;
	} else {
		croak "unknown dst object $!";
	}

	if ($src->isa("Compression::Qtree") && $dst->isa("Compression::Qtree") ) {
		
		my $qtree_files = $src_compr_obj->{dataset}->_get_qtree_file_hash(qtree_name=>$src->{name});

		$dst_compr_obj->{dataset}->_attach_to_cloned_dataset(files => $qtree_files,
							removed_files => $src_compr_obj->{dataset}->{_c_dataset}->{_common}->{_removed_files},
							ds_type => $src_compr_obj->{dataset}->{_c_dataset}->{_common}->{ds_type},
							testdir => $src_compr_obj->{dataset}->{_c_dataset}->{_common}->{testdir},
							testdir_path => $src_compr_obj->{dataset}->{_c_dataset}->{_common}->{testdir_path},
							parent_path => $src_compr_obj->{volume}->get_mount_path(),
							parent_mountpoint => $src_compr_obj->{dataset}->{_c_dataset}->{_common}->{mountpoint});

		$dst_compr_obj->{dataset}->_replace_qtree_name(old_qtree_name=>$src->{name}, 
														new_qtree_name=>$dst->{name}, );
	} else {
		$dst_compr_obj->{dataset}->_attach_to_cloned_dataset(files => $src_compr_obj->{dataset}->{_c_dataset}->{_common}->{files},
							removed_files => $src_compr_obj->{dataset}->{_c_dataset}->{_common}->{_removed_files},
							ds_type => $src_compr_obj->{dataset}->{_c_dataset}->{_common}->{ds_type},
							testdir => $src_compr_obj->{dataset}->{_c_dataset}->{_common}->{testdir},
							testdir_path => $src_compr_obj->{dataset}->{_c_dataset}->{_common}->{testdir_path},
							parent_path => $src_compr_obj->{volume}->get_mount_path(),
							parent_mountpoint => $src_compr_obj->{dataset}->{_c_dataset}->{_common}->{mountpoint});
	}
}


1;