#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/SnapVault.pm#1 $
#
## @summary Compression SnapVault routines
##  
##
## @description
## Wrapper around NOTE::SnapVault.
## The main reason for not using NOTE::SnapVault directly is to provide standardized output
##    and to enable specific processing on some failures, otherwise fatal.
##    This is required for handling compression-specific failures.
##
## @synopsis
## use Compression::SnapVault;
## use vars qw( $FILER  );
##
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $obj = Compression::SnapVault->new (host => $FILER,
##				            initialize => TRUE );
##
## @author dl-compression-qa@netapp.com, waroff@netapp.com
## @status private (in-progress)

package Compression::SnapVault;

use strict;
use warnings;
use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult process_performance_info);
use fileSystem qw(treecompare);
use Carp qw(croak);

BEGIN
{
	require "Compression/base_import.pl";

}


## @name:: start
## @summary 
##
## @description start a snapvault relationship
## 
##
## @return Object of type Compression::SnapVault.
##
## @arg src Required: src object from Compression::Qtree
## @arg dst Required: dst object from Compression::Qtree
## @arg restart Optional: Flag for the start command to resync a relationship (equivalent of -r option).
## @arg debug Optional: Controls verbosity of this module (defaults to FALSE)
##
## @example
## my $obj = Compression::SnapVault->start ( src   => $qt_src,
##										     dst   => $qt_dst, );
##
sub start
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		src => { type=>OBJECT, isa => "Compression::Qtree", },
		dst => { type=>OBJECT, isa => "Compression::Qtree", },
		restart => {type => SCALAR, optional => 1},
		debug => { type=>SCALAR, optional => 1, default=>FALSE },
	} );
	my $src = $args{src};
	my $dst = $args{dst};

	$src->{_qtree}->Filer->snapvault_enable();
	$dst->{_qtree}->Filer->snapvault_enable();

	my $sv_obj = NOTE::SnapVault->start(src=>$src->{_qtree},
					    dst=>$dst->{_qtree},
					    restart => $args{restart},
					    );

	my $self = {};
	$self->{ID} 	= $$;
	# This assumes that the command will execute on the source filer
	# or that both src and dst are running the same version.
	$self->{_ui} = $src->{_compress}->{_ui};
	$self->{_debug}	= $args{debug};
	$self->{_snapvault} = $sv_obj;

	$self->{_src} = $src;
	$self->{_dst} = $dst;

	bless($self, $class);

	logcomment("SnapVault: Setup Complete..\n\n") if ($self->{_debug});
	return $self;
}

## @name:: update
## @summary update
##
## @description Starts an incremental transfer for a SnapVault relationship
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @return 
##
## @example
## %res = $sv->update();
##
sub update
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		background	=> { type=>SCALAR, optional => 1, default=>FALSE, },
		throttle	=> { type=>SCALAR, optional => 1,  },
		expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
	} );
	
	my $expected = delete($args{expected});
	
	eval {
		$class->{_snapvault}->update(%args);
	};
	
	my $output = $@;
	
	return $class->{_ui}->process_output_string(code => C_CMD_SNAPVAULT_UPDATE,
						    output => $output,
						    expected_rc => $expected,
						    log_message => "Snapvault update",
						    );
}


## @name:: status_wait
## @summary Wait for the SnapVault relationship to reach specific states.
##
## @description 
## Wait for the SnapVault relationship to reach specific states.
##
## @arg status Required: Status to wait for.
##
## @arg ignore_transfer_error Optional: Flag to not fail if the status contains the C<current-transfer-error> element.
## Default: FALSE.
##
## @arg timeout Optional: Maximum time allowed for status_wait to complete, in seconds.
## 
## @return Hash with rc and output.
##
## @example
## %res = $sv->status_wait(status => 'idle');
##
sub status_wait
{
	my $class 	= shift;
	my (%args) = validate (@_, {
		status => { type=>SCALAR },
		ignore_transfer_error => { type=>SCALAR, optional=>1, default=>FALSE },
		timeout => {type=>SCALAR, optional => 1,},
	} );
	
	
	$class->{_snapvault}->status_wait(%args);
	
	return (rc => C_RC_SUCCESS);
	
}


## @name:: stop
## @summary stop
##
## @description Stops a SnapVault relationship
## 
##
## @return 
##
## @example
## %res = $sv->stop();
##
sub stop
{
	my $class 	= shift;
	my (%args) = validate (@_, {} );
	my %res;
	
	$class->{_snapvault}->stop();
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
	
}


## @name:: release
## @summary release
##
## @description Releases the source-side resources being held for a particular SnapVault destination
## 
##
## @return 
##
## @example
## %res = $sv->release();
##
sub release
{
	my $class 	= shift;
	my (%args) = validate (@_, {} );
	my %res;
	
	$class->{_snapvault}->release();
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
	
}

## @name:: restore
## @summary restore
##
## @description restore
## 
## @return Hash with rc and output.
##
## @example
## %res = $sv->restore(snapshot => 'ss1');
##
sub restore
{
	my $class 	= shift;
	my (%args) = validate (@_, {
			qtree	 => { type=>OBJECT, isa => "Compression::Qtree", },
			snapshot => { type=>SCALAR },
			force => { type => SCALAR, optional => 1, default => TRUE, },
	} );
	my %res;
	my $snapshot = $args{'snapshot'};
	my $qtree = $args{'qtree'};

	if ( $qtree->{_compress}->{volume}->get_mount_path() ne $class->{_src}->{_compress}->{volume}->get_mount_path() ) {
		my $old_qtree_name = $class->{_src}->{name};
		my $new_qtree_name = $qtree->{name};
		logcomment("SnapVault restore to new source qtree: $new_qtree_name");
		$qtree->{_compress}->{dataset}->_attach_to_cloned_dataset(files => $class->{_src}->{_compress}->{dataset}->{_c_dataset}->{_common}->{files},
						    removed_files => $class->{_src}->{_compress}->{dataset}->{_c_dataset}->{_common}->{_removed_files},
						    ds_type => $class->{_src}->{_compress}->{dataset}->{_c_dataset}->{_common}->{ds_type},
						    testdir => $class->{_src}->{_compress}->{dataset}->{_c_dataset}->{_common}->{testdir},
						    testdir_path => $class->{_src}->{_compress}->{dataset}->{_c_dataset}->{_common}->{testdir_path},
						    parent_path => $class->{_src}->{_compress}->{volume}->get_mount_path(),
						    parent_mountpoint => $class->{_src}->{_compress}->{dataset}->get_mountpoint());
		$qtree->{_compress}->{dataset}->_replace_qtree_name(old_qtree_name=>$old_qtree_name, 
															new_qtree_name=>$new_qtree_name, );
	} else {
		logcomment("SnapVault restore to original source qtree.");
	}
	
	$class->{_snapvault}->restore(dst =>$qtree->{_qtree}, snapshot => $snapshot, force => $args{force});
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
	
}


## @name:: snap_sched
## @summary snap_sched
##
## @description snap_sched
## 
## @return Hash with rc and output.
##
## @example
## %res = $sv->snap_sched(snapshot => 'ss1');
##
sub snap_sched
{
	my $class 	= shift;
	my (%args) = validate (@_, {
			snapshot => { type=>SCALAR, optional=>1 },
			sched    => { type=>SCALAR, optional=>1 },
			xfer     => { type=>SCALAR, optional=>0, default=>0 },
			source   => { type=>SCALAR, optional=>0, default=>0 },
	} );
	my %res;
	
	my %params;
	foreach my $arg (sort keys %args) {
		if ( defined $args{$arg} ) {
			$params{$arg} = $args{$arg};
		}
	}

	$class->{_snapvault}->snap_sched(%params);
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
	
}

## @name:: abort
## @summary abort
##
## @description abort
## 
## @return Hash with rc and output.
##
## @example
## %res = $sv->abort();
##
sub abort
{
	my $class 	= shift;
	my (%args) = validate (@_, {} );
	my %res;
	
	$class->{_snapvault}->abort();
	$res{'rc'} = C_RC_SUCCESS;
	return %res;
	
}


## @name:: check
## @summary SnapVault check
##	
##
## @description
##	Run SnapVault checker
##
## @return Hash with rc and output.
##
## @example
## $obj->check();
sub check
{
	my $class 	= shift;
	my (%args) = validate (@_, { } );
	my %res;
	$res{'rc'} = C_RC_SUCCESS;
	$class->{_snapvault}->check();
	return %res;
}

## @name:: treecompare
## @summary treecompare
##	
##
## @description
##	Run treecompare
##
## @return Hash with rc and output.
##
## @example
## $obj->treecompare();
sub treecompare
{
	my $class 	= shift;
	my (%args) = validate (@_, { } );
	my %res;
	$res{'rc'} = C_RC_SUCCESS;
    my $base_snap = $class->{_snapvault}->status->{'base-snapshot'};
	my $dst_snap = $class->{_snapvault}->dst_snapshot;
    my $src_path = $class->{_src}->{_compress}->{_vol_mount_path} . "/.snapshot/$dst_snap/" . $class->{_src}->{name};
    my $dst_path = $class->{_dst}->{_compress}->{_vol_mount_path} . "/.snapshot/$base_snap/" . $class->{_dst}->{name};
	my $src_vol = $class->{_src}->{_compress}->{volume}->_get_underlying_vol_obj();
	my $dst_vol = $class->{_dst}->{_compress}->{volume}->_get_underlying_vol_obj();
	my $src_conn = $src_vol->Filer->conn;
	my $dst_conn = $dst_vol->Filer->conn;
    my $ret = fileSystem::treecompare($src_conn, $src_path, $dst_path, filer_dst=>$dst_conn);
	if ($ret != 1) {
		croak "treecompare failed $!";
	}
	return %res;
}


## @name:: check_performance
## @summary Check SnapVault performance.
##	
## @description
## Check SnapVault performance. Obtain the last-transfer-duration and the amount of data transfered
## from the status and calculate the throughput.
##
## @arg max_allowed_duration Optional: Maximum value allowed for the transfer duration (s).
## If this argument is used, the obtained duration is compared against it and
## if it is bigger, then 'failure_handling' is executed.
##
## @arg min_allowed_throughput Optional: Minimum value allowed for the throughput (KB/s).
## If this argument is used, the obtained throughput is compared against it and
## if it is smaller, then 'failure_handling' is executed.
##
## @arg failure_handling Optional: Flag to determine what action should be taken if
## the throughput was detected to be smaller than min_allowed_throughput.
## This argument is ignored when none of min_allowed_throughput or max_allowed_duration are passed in.
## Allowed values: none, fail, panic, panic_and_fail. Default: fail.
##
## @return Hash with rc and duration.
##
## @example
## $obj->check_performance();
## $obj->check_performance(max_allowed_duration => 10,);
## $obj->check_performance(min_allowed_throughput => 50,
##                         fail_action => 'none');
##
sub check_performance
{
	my $class = shift;
	my (%args) = validate (@_,
	{
		max_allowed_duration => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
		min_allowed_throughput => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
		failure_handling => {type => SCALAR, regex => qr/^none|panic|fail|panic_and_fail$/, default => 'fail',}, 
	});
	
	my $status = $class->{_snapvault}->status();
	
	my $last_transfer_duration = $status->{'last-transfer-duration'};
	my $last_transfer_size = $status->{'last-transfer-size'};
	my $last_transfer_type = (exists($status->{'last-transfer-type'})) ? $status->{'last-transfer-type'} : 'unknown';
	
	my $src = $class->{_snapvault}->src->Filer->host();
	my $dst = $class->{_snapvault}->dst->Filer->host();
	my $filers_to_panic = ($src eq $dst) ? [$src] : [$src, $dst]; 
	return Compression::Utils::process_performance_info(duration => $last_transfer_duration,
							    description => "SnapVault ($last_transfer_type)",
							    boot_mode => C_ONTAP_7MODE,
							    max_allowed_duration => $args{max_allowed_duration},
							    min_allowed_throughput => $args{min_allowed_throughput},
							    processed_data_size => $last_transfer_size,
							    failure_handling => $args{failure_handling},
							    generate_cores_on => $filers_to_panic,);
}


## @name:: get_last_transfer_size
## @summary Get the size of the last transfer.
##	
## @description
## Get the size of the last transfer.
##
## @return The size in KB.
##
## @example
## my $data_transferred = $obj->get_last_transfer_size();
##
sub get_last_transfer_size
{
	my $class = shift;
	my (%args) = validate (@_,
	{
		# None
	});
	
	my $status = $class->{_snapvault}->status();
	my $size = $status->{'last-transfer-size'};
	
	if (!defined($size))
	{
		logresult('fatal', "Could not detect the size of the last transfer:\n" . Dumper($status));
	}
	
	logcomment("Size of the last snapvault transfer: $size");
	
	return $size;
}



1;