#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Supportability.pm#1 $
#

## @summary Utility module to be used for supportability testing.
##
## @description
## Utility module to be used for supportability testing.
## Executes release and mode agnostic testing for supportability.
## For BR, it ensures that the events are logged in /etc/log/messages.
## For RR 7-mode, it ensures that they are loged in /etc/log/sis.
## For RR c-mode, it ensures that they are loged in /mroot/etc/log/sis.
## Intended to be used as a member of the Compress object.
##
## @example
##
## my $compress = Compression::Compress->setup(filer => {name => $FILER,},
##                                             aggregate => {name => $AGGREGATE,},
##                                             volume => {name => $VOLUME,},
##                                             supportability => TRUE,);
## my $volname = $compress->{vol_name};
## $compress->{supportability}->start_tailing(required_identifier => $volname,
##                                            events => [{code => C_EVENT_COMPRESSION_DISABLED,},
##                                                       {code => C_EVENT_COMPRESSION_ENABLED,
##                                                        options => {inline => TRUE,},},
##                                                       ],);
## $compress->disable();
## $compress->wait_for_cp();
## $compress->enable(inline => TRUE);
## $compress->{supportability}->stop_and_match();


package Compression::Supportability;

use strict;
use warnings;

use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Tharn;

use Compression::Constants qw(:all);
use LogReader::Base;



BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}



## @name ::new
## @summary Compression::Supportability Constructor.
##
## @description Compression::Supportability Constructor.
##
## @return  Compression::Supportability object.
##
## @arg conn Required: NOTE::Filer (for 7-mode) or (G)NOTE::GX::Node (for c-mode) object .
##
## @arg ui Required: Compression::UI object corresponding to the hosting filer.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
## 
## @arg boot_mode Optional: Boot mode of the hosting filer (C_ONTAP_7MODE or C_ONTAP_CMODE).
## Defaults to C_ONTAP_7MODE.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub new
{
	my $class = shift;
        my %args = validate_with
                (params => \@_,
                 spec   => {conn => {type => OBJECT,},
                            ui => {type => OBJECT, isa => "Compression::UI", },
                            debug => {type => SCALAR, optional => 1, default => FALSE,},
                            boot_mode => {type => SCALAR, default => C_ONTAP_7MODE,},
                            },
                 );
	
	my $self = {};
	$self->{_conn} = $args{conn};
        $self->{_debug} = $args{debug};
        $self->{_ui} = $args{ui};
        $self->{_boot_mode} = $args{boot_mode};
        $self->{ontap_ver} = $self->{_ui}->{version}->{ontap};
        
        if ($self->{_boot_mode} eq C_ONTAP_7MODE)
        {
                if  (!($self->{_conn}->isa("NOTE::Filer")))
                {
                        logresult('script', "For 7-mode, the conn argument is required to be a NOTE\:\:Filer object");
                }
                
                $self->{_server} = $self->{_conn}->conn();
        }
        elsif ($self->{_boot_mode} eq C_ONTAP_CMODE)
        {
                if (!($self->{_conn}->isa("NOTE::GX::Node")))
                {
                        logresult('script', "For C-mode, the conn argument is required to be a NOTE\:\:GX\:\:Node object");
                }
                
                $self->{_server} = $self->{_conn}->server();
        }
        
        $self->{_tail} = LogReader::Base->new($self->{_server});
        $self->{_started_tailing} = 0;
        $self->{_expected_messages} = [];
	
	bless $self, $class;
        
        $self->{log_file} = $self->_get_log_file();
        
	return $self;
}

## @name ::start_tailing
## @summary Initiate tailing of the supportability message for specific action(s).
##
## @description Initiate tailing of the supportability message for specific action(s).
##
## @arg events Required: Arrayref of events information (in the form of hashrefs)
## for which the supportability message is targeted.
##
## @arg required_identifier Optional: Entity for which the message check is expected to be printed.
## (Ex: the volume name). If not specified, then the targeted supportability message is allowed to
## be printed for any entity.
##
## @return Hash with rc.
##
## @example
## $compress->{supportability}->start_tailing(required_identifier => $volname,
##                                            events => [{code => C_EVENT_COMPRESSION_DISABLED,},
##                                                       {code => C_EVENT_COMPRESSION_ENABLED,
##                                                        options => {inline => TRUE,},},
##                                                       ],);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub start_tailing
{
        my $class = shift;
        my (%args) 	= validate( @_,
                {
                        events => {type => ARRAYREF,},
                        required_identifier => {type => SCALAR|UNDEF, optional => 1,},
                } );
        
        my $events = $args{events};
        $class->{_required_identifier} = $args{required_identifier};
        
        if ($class->{_started_tailing} == 1)
        {
                logresult('script', "Supportability\:\:start_tailing cannot be called twice in a row. Use stop_and_match in between calls");
        }
        
        my @tailed_files = ();
        foreach my $event (@$events)
        {
                my %event_info = validate_with
                        (params => $event,
                         spec   => {code => { type => SCALAR,},
                                    options => { type => HASHREF|UNDEF, optional => 1},
                                    },
                         allow_extra => 0,
                         );
                        
                $class->_attach_tailing_info(%event_info);
        }
        
        my $results = $class->{_tail}->rdfile($class->{log_file});

	$class->{start_linecount} = scalar(@$results);
        
        $class->{_started_tailing} = 1;
        
        logcomment("Supportability: initialized tailing:\n" .
                   "Log file: " . $class->{log_file} . "\n" .
                   "Expected messages:\n" . Dumper($class->{_expected_messages})) if $class->{_debug};
        
        return (rc => C_RC_SUCCESS);
}


## @name ::stop_and_match
## @summary Completes a previously-started tailing, by searching for the expected messages.
##
## @description Completes a previously-started tailing, by searching for the expected messages.
##
## @return Hash with rc and output.
##
## @example
## $compress->{supportability}->stop_and_match();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub stop_and_match
{
        my $class = shift;
        my (%args) 	= validate( @_,
                {
                        # No args
                } );
        
        if ($class->{_started_tailing} == 0)
        {
                logresult('script', "Supportability\:\:start_tailing needs to be called before calling stop_and_match");
        }
        
        my $output = '';

        my $results = $class->{_tail}->rdfile($class->{log_file});
        
        my $stop_linecount = scalar(@$results);
        my $start_linecount = $class->{start_linecount};
        # To watch for: if log file rotates: is this enough? Should we consider looking in the .x log file?
        if ($stop_linecount < $start_linecount)
        {
                logcomment("Supportability testing:\n" .
                           "Log file " . $class->{log_file} . " must have been rotated\n" .
                           "Parsing from the first line");
                $start_linecount = 0;
        }
        
        logcomment("Supportability: stop_and_match: starting processing:\n" .
                   "Start linecount: $start_linecount\n" .
                   "Stop linecount: $stop_linecount\n" .
                   "Attempting to match messages:\n" . Dumper($class->{_expected_messages})) if $class->{_debug};
        for (my $index = $start_linecount; $index < $stop_linecount; $index++)
        {
                my $line = $results->[$index];
                if ($line eq "\n")
                {
                        logcomment("Supportability: processing empty line: \"$line\"\n") if ($class->{_debug});
                        next;
                }
                $output .= "$line\n";
                
                my $expected_messages = $class->{_expected_messages};
                if (scalar(@$expected_messages) == 0)
                {
                        # No more messages to look for 
                        last;
                }
                logcomment("Supportability: processing line: \"$line\"\n" .
                           "Looking for any of the expected messages:\n" .
                           Dumper($expected_messages)) if ($class->{_debug});
                
                my $i = 0;
                foreach my $expected_message (@$expected_messages)
                {
                        logcomment("Supportability ($i): trying to match: $expected_message") if $class->{_debug};
                        if ($line =~ /$expected_message/)
                        {
                                logcomment("Supportability testing:\n" .
                                           "Matched expected message in " . $class->{log_file} . ":\n$line");
                                splice(@$expected_messages, $i, 1);
                                # Exit the 'foreach $expected_message' loop
                                last;
                        }
                $i++;
                }
        }
        
        my $unmatched_messages = $class->{_expected_messages};
        logcomment("Supportability: stop_and_match completed unmatched messages:\n" . Dumper($unmatched_messages)) if ($class->{_debug});
        
        # Re-init the tailing info
        $class->{_started_tailing} = 0;
        $class->{_expected_messages} = [];
        
        if (scalar(@$unmatched_messages) != 0)
        {
                logresult('fatal', "Supportability: failed to match all expected messages (in " . $class->{log_file} . "):\n" .
                                   "Output:\n\"$output\"\n\n" .
                                   "Unmatched messages:\n" . Dumper($unmatched_messages));
        }
        
        return (rc => C_RC_SUCCESS, output => $output);
}


###################
# Internal methods
###################

sub _get_log_file
{
        my $class = shift;
        my $boot_mode = $class->{_boot_mode};
        my $ontap = $class->{ontap_ver};
        
        if ($boot_mode eq C_ONTAP_7MODE)
        {
                if ($ontap eq '8.0')
                {
                        return '/etc/log/messages';
                }
                else
                {
                        # This should be the latest release with changes to supportability mechanism
                        return '/etc/log/sis';
                }
        }
        elsif ($boot_mode eq C_ONTAP_CMODE)
        {
                return '/mroot/etc/log/sis';
        }
        else
        {
                logresult('script', "Unrecognized boot_mode: \"$boot_mode\"");
        }
}

sub _attach_tailing_info
{
        my $class = shift;
        my %args = @_;
        
        # Note: currently we do not need to decide on boot_mode
        # The log messages are identical for both modes
        #my $boot_mode = $class->{_boot_mode};
        my $ontap = $class->{ontap_ver};
        my $attached_info = FALSE;
        
        if ($ontap eq '8.0')
	{
		$attached_info = $class->_event_info_8_0(%args);
        }
        
        if ($attached_info == FALSE)
        {
                logcomment("Could not detect release-specific supportability information. Using the defaults") if ($class->{_debug});
                $attached_info = $class->_event_info_defaults(%args);
        }
        
        if ($attached_info == FALSE)
        {
                logresult("script", "Could not detect supportability information (release $ontap):\n" . Dumper(%args));
        }
}

sub _attach_supported_info
{
        my $class = shift;
        my %args = validate(@_,
	{       
                info  => {type 	 => HASHREF,},
                options => {type => HASHREF|UNDEF,},
	} );
        
        my $info = $args{info};
        my $options = $args{options};
        my $expected_messages = $class->{_expected_messages};
        my $attach_main_message = 1;
        
        if (defined($options))
        {
                foreach my $passed_in_option (keys %$options)
                {
                        # Currently only TRUE/FALSE (1/0) handling...
                        my $option_value = $options->{$passed_in_option};
                        if ((defined($option_value)) &&
                            (exists($info->{options}->{$passed_in_option}->{$option_value})))
                        {
                                my $option_info = $info->{options}->{$passed_in_option}->{$option_value};
                                # Look for options which may cause extra log message to be printed
                                if (exists($option_info->{logs_extra_message}))
                                {
                                        my $message = $class->_substitute_required_identifier($option_info->{logs_extra_message});
                                        push(@$expected_messages, $message);
                                }
                                
                                # Look for options which may cause the log message to not be printed
                                if ((exists($option_info->{suppresses_log_message})) &&
                                    ($option_info->{suppresses_log_message} == TRUE))
                                {
                                        $attach_main_message = 0;
                                }
                        }
                }
        }
        
        if (($attach_main_message == 1) && (exists($info->{logs_message})))
        {
                my $message = $class->_substitute_required_identifier($info->{logs_message});
                push(@$expected_messages, $message);
        }
        
        return C_RC_SUCCESS;
}

sub _substitute_required_identifier
{
        my $class = shift;
        my $message = shift;
        
        my $identifier = $class->{_required_identifier};
        
        if ($message =~ /<identifier>/)
        {
                if (defined($identifier))
                {
                        $message =~ s/<identifier>/$identifier/;
                }
                else
                {
                        # Be content with 'any identifier' (not including spaces) 
                        my $any = '\S+';
                        $message =~ s/<identifier>/$any/;
                }
        }
        
        return $message;
}

sub _event_info_8_0
{
        my $class = shift;
        my %args = @_;
        
        my $event_code = $args{code};
        
        my $info = {'COMPRESSION_ENABLED' => {logs_message => 'Compression is now turned On for volume \"<identifier>\"',},
                    'COMPRESSION_DISABLED' => {logs_message => 'Compression is now turned Off for volume \"<identifier>\"',},
                    'COMPRESS_STARTED' => {logs_message => 'Starting volume compression on volume <identifier>',},
                    'COMPRESS_ABORTED' => {logs_message => 'Compression of existing data for volume \"<identifier>\" is aborted',},
                    'COMPRESS_COMPLETED' => {logs_message => 'Compression of existing data for volume \"<identifier>\" is done',},
                    'DECOMPRESS_STARTED' => {logs_message => 'Starting volume decompression on volume <identifier>',},
                    'DECOMPRESS_COMPLETED' => {logs_message => 'decompression of volume \"<identifier>\" is done',},
                    'DECOMPRESS_ABORTED' => {logs_message => 'decompression of volume \"<identifier>\" is aborted',},
                    };
        
        if (defined($info->{$event_code}))
        {
                $class->_attach_supported_info(info => $info->{$event_code},
                                               options => $args{options});
                return TRUE;
        }
        
        return FALSE;
}

sub _event_info_defaults
{
        my $class = shift;
        my %args = @_;
        
        my $event_code = $args{code};
        my $options = $args{options};
        
        my $info = {'COMPRESSION_ENABLED' => {logs_message => '/vol/<identifier>\s+\[sid:\s+\d+\]\s+Compression Enabled',
                                              options => {inline => {'1' => {suppresses_log_message => FALSE,
                                                                             logs_extra_message => '/vol/<identifier>\s+\[sid:\s+\d+\]\s+Inline Compression Enabled',},},},},
                    'COMPRESSION_DISABLED' => {logs_message => '/vol/<identifier>\s+\[sid:\s+\d+\]\s+Compression Disabled',
                                               options => {inline => {'1' => {suppresses_log_message => TRUE,
                                                                              logs_extra_message => '/vol/<identifier>\s+\[sid:\s+\d+\]\s+Inline Compression Disabled',},
                                                                      '0' => {suppresses_log_message => FALSE,
                                                                              logs_extra_message => '/vol/<identifier>\s+\[sid:\s+\d+\]\s+Inline Compression Disabled',},},},},
                    'COMPRESS_STARTED' => {logs_message => '/vol/<identifier>\s+\[sid:\s+\d+\]\s+Begin\s+\(sis\s+start\)',
                                           options => {scan => {'1' => {suppresses_log_message => TRUE,
                                                                        logs_extra_message => '/vol/<identifier>\s+\[sid:\s+\d+\]\s+Begin\s+\(sis\s+start\s+scan\)',},},},},
                    'COMPRESS_ABORTED' => {logs_message => '/vol/<identifier>\s+\[sid:\s+\d+\].+Operation was stopped',},
                    'COMPRESS_COMPLETED' => {logs_message => '/vol/<identifier>\s+\[sid:\s+\d+\]\s+End\s+\(\d+\s+KB\)',},
                    'DECOMPRESS_STARTED' => {logs_message => '/vol/<identifier>\s+\[sid:\s+\d+\]\s+Undo',},
                    'DECOMPRESS_COMPLETED' => {logs_message => '/vol/<identifier>\s+\[sid:\s+\d+\]\s+End\s+\(\d+\s+KB\)',},
                    'DECOMPRESS_ABORTED' => {logs_message => '/vol/<identifier>\s+\[sid:\s+\d+\].+Operation was stopped',},
		    'DECOMPRESS_SPACE_FULL' => {logs_message => '/vol/<identifier>\s+\[sid:\s+\d+\].+Not enough disk space for the operation',},
		    'COMPRESSED_VOL_IS_OFFLINED' => {logs_message => 'Volume \'<identifier>\' is compressed and is being taken offline. Compression is not supported in this version of Data ONTAP',},
                    };
        
        if (defined($info->{$event_code}))
        {
                $class->_attach_supported_info(info => $info->{$event_code},
                                               options => $options);
                return TRUE;
        }
        
        return FALSE;
}



1;
