#
# $Id: //depot/prod/test/main/lib/Compression/Tape.pm#2 $
#

## @summary Tape operations.
##
## @description
## Tape operations.
##
## @examples
##
## my $obj = Compression::Compress->setup(
##             filer => { name => $FILER, },
## 	       aggregate => { name => $AGGREGATE },
##             volume => { name => $VOLUME, size => "20g",},
##      );
## my $tape = Compression::Tape->new(src => $src_vol,  
##                                   dst => $dst_vol,
##				     tape_name => $TAPE_NAME,
##				     tape_protocol=>$TAPE_BACKUP_PROTOCOL,
##				     tapefiler=>$TAPEFILER);
##
## NOTE:
## 1. This library is a library for Ndmp, dump/restore and smtape. 
##  if TAPE_BACKUP_PROTOCOL is not specified default to "ndmp" if tape is available otherwise default to "dumprestore"
##  if TAPE_BACKUP_PROTOCOL="ndmp" and tape is not available then result is "NA"
##  if TAPE_BACKUP_PROTOCOL="dumprestore" and tape is available the operations should be dump/restore (not using ndmp)
##  if TAPE_BACKUP_PROTOCOL="smtape" and tape is available the operations should be "smtape" backup/restore; if tape not available result is "NA"
## 2. For NDMP, source filer and destination filer could be different than tapefiler. (Tapefiler is the filer that tape drive attached to)
##     If they are different, 3way ndmp backup/restore will be performed, otherwise, it will be local backup/restore
## 3. For Dump/restore, source filer, destination filer and tapefiler should be same
##     If tape_protocol is dump/restore but there is not tape drive available, the dump/restore protocol will be performed without tape
## 4. For ndmpcopy, there is no need to set any TAPE_BACKUP_PROTOCOL
##  
## @author weili@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::Tape;

BEGIN
{
	require "Compression/base_import.pl";
}

use strict;
use warnings;

use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use NOTE::NDMP;
use FSR::DR::DRutils;
use Tapeinfo;
use dp::dumprestore::DumpRestore  qw( dump_data restore_data );

use Compression::Constants qw(:all);

## @name ::new 
##
## @summary Compression::Tape constructor.
##
## @description
## Compression:Tape constructor.
## Creates a Tape object.
##
## @return Compression::Tape object
##
## @arg src Required: Tape source. Must be a Compression object
##
## @arg dst Required: Tape destination. Must be a Compression object
##
## @arg tape_name Optional: Name of the tape where the backup/restore will be fromed
##      Default:  first non rewindable tape name got from "sysconfig -t"
## 
## @arg tapefiler Optional: Name of the filer with tape drive attached
##	Default: Source filer
##
## @arg tape_protocol Optional: use ndmp or dump/restore backup method (ndmp, dumprestore, smtape)
##
## @examples
## my $Tape = Compression::Tape->new(src => $src_vol,  
##                                   dst => $dst_vol,
##				     tape_name => $TAPE_NAME,
##				     tape_protocol=>$TAPE_BACKUP_PROTOCOL,
##				     tapefiler=>$TAPEFILER);
##
sub new
{
	my $class = shift;
	
	my (%args) = validate (@_, {src 	  => { type =>OBJECT, isa => "Compression::Compress" },
				    dst		  => { type =>OBJECT, isa => "Compression::Compress" },
				    tapefiler	  => { type =>SCALAR, optional => 1,  },
				    debug 	  => { type =>SCALAR, optional => 1, default=>FALSE },
				    tape_protocol => { type => SCALAR|UNDEF, optional => 1},
				    tape_name 	  => { type => SCALAR|UNDEF, optional => 1},
				    boot_mode 	  => {type => SCALAR, optional => 1, default => C_ONTAP_BOOT_MODE},
	} );

	logcomment ("Create new Tape object in Compression::Tape");
	# Start building our object:
	my $self = {};
        $self->{_boot_mode} = $args{boot_mode};
	if ($self->{_boot_mode} eq C_ONTAP_CMODE)
	{
		logresult('Script', "Tape library does not support CMODE for now");

	}
	
	$self->{_debug}	= $args{debug};
	$self->{_src_vol} = $args{src}->{volume}->_get_underlying_vol_obj();
	$self->{_dst_vol} = $args{dst}->{volume}->_get_underlying_vol_obj();
	$self->{_src_filer} = $args{src}->{filer}->_get_underlying_filer_obj();
	$self->{_dst_filer} = $args{dst}->{filer}->_get_underlying_filer_obj();
	$self->{_tape_protocol} = $args{tape_protocol};
	$self->{_src_filer_conn} = $self->{_src_filer}->conn;
	$self->{_dst_filer_conn} = $self->{_dst_filer}->conn;
	$self->{_src_host} = $args{src}->{filer}->{host_name};
	$self->{_dst_host} = $args{dst}->{filer}->{host_name};
	
	if( $args{tapefiler} ) {
		$self->{_tapefiler} = NOTE::Filer->connect( host => $args{tapefiler} );
		$self->{_tapefiler_conn } = $self->{_tapefiler}->conn;
	} else {
		$self->{_tapefiler}=$self->{_src_filer};
		$self->{_tapefiler_conn } = $self->{_src_filer_conn};
	}
	
	bless($self, $class);
	
	$self->{_tape_name} = $self->_get_tape();
	if( $self->{_tape_protocol} eq C_SMTAPE_TAPE_PROTOCOL) {
		$self->{_tape_name} = (defined $args{tape_name}) ? $args{tape_name} : $self->{_tape_name};
		$self->{_tape} = NOTE::ImageBackup->new(storage=>$self->{_src_vol}, device=>$self->{_tape_name});
	} else {
		$self->{_tape}= NOTE::NDMP->new(src => $self->{_src_filer}, dst => $self->{_dst_filer});
	}

	if( ($self->{_tape_protocol} eq C_NDMP_TAPE_PROTOCOL) ||
	    (!$self->{_tape_protocol} && $self->{_tape_name}) ){
		if (!$self->{_tape_name}){
			logresult ('NA', "NDMP protocol requested, but tape is not available");
		} else {
			$self->{_tape_protocol} = C_NDMP_TAPE_PROTOCOL;
			$self->{_tape}->enable();
			$self->{_tape}->enable(destination  => 1);
			$self->{_tape}->set_access(access=>C_NDMP_ACCESS);
			$self->{_tape}->set_access(destination=>1, access=>C_NDMP_ACCESS);
			$self->{_tape}->set_auth_type(authtype=>C_NDMP_AUTHTYPE);
			$self->{_tape}->set_auth_type(destination=>1, authtype=>C_NDMP_AUTHTYPE);
			$self->{_tape}->set_version(destination=>1, version=>C_NDMP_VERSION);
			$self->{_tape}->set_version(destination=>1, version=>C_NDMP_VERSION);
			$self->{_tape_name} = (defined $args{tape_name})? $args{tape_name} : $self->{_tape_name};
		}
	} elsif( $self->{_tape_protocol} eq C_SMTAPE_TAPE_PROTOCOL) {
		if (!$self->{_tape_name}){
			logresult('NA', "Smtape protocol requested, but tape is not available");
		} else {
			$self->{_tape_protocol} = C_SMTAPE_TAPE_PROTOCOL;  
			$self->{_tape_name} = defined $args{tape_name} ? $args{tape_name}:$self->_get_tape();
		}
	} else {
		$self->{_tape_protocol} = C_DUMP_TAPE_PROTOCOL;
		if ($self->{_tape_name}){
			$self->{_tape_name} = (defined $args{tape_name}) ? $args{tape_name} : $self->{_tape_name};
		}
	}
		
	return $self;

}

## @name ::backup 
##
## @summary Does a tape backup
##
## @description
##   Does a tape backup
##
## @return Hash with rc 
##
## @arg src Optional: Compression object of the volume to be backed up
## 	Default:  source volume
##
## @arg version Optional: Ndmp version used for doing tape backup
##	Default:  4
##
## @arg mover_record Optional: mover record size
##	Default:  64512
##
## @arg mover_window  Optional: mover window size
##	Default:  -1
##
## @arg isIncr  Optional: Is increment backup
##	Default:  0
##
## @arg snapshot  Optional: snap name on the source volume
##		    If given then src path given for ndmp backup will be /vol/<volname>/.snapshot/<snapshot>
##	Default:  0
##
## @arg dump_dir  Optional: for dump/restore only, the directory to be dumped to 
##	Default:  dump_dir
##
## @arg dump_level  Optional: for dump/restore only, dumped level
##	Default:  0
##
## @arg dump_opt  Optional: for dump/restore only, dump options
##	Default:  'uf'
##
## @examples
##	     $tape->backup(src=>$src_vol);
##
sub backup
{
	my $class = shift;
	my (%args) = validate (@_, {src 	  => { type => OBJECT, isa => "Compression::Compress", optional => 1, },
				    version  	  => { type => SCALAR, optional => 1, },  
				    mover_record  => { type => SCALAR, optional => 1, },  
				    mover_window  => { type => SCALAR, optional => 1, },  
				    isIncr 	  => { type => BOOLEAN, optional => 1, },  
				    snapshot 	  => { type => SCALAR, optional => 1, },
				    dump_dir	  => { type => SCALAR, optional => 1, default  => C_NDMP_DUMP_DIR, }, 
				    dump_level	  => { type => SCALAR, optional => 1, default  => 0, },
				    dump_opt	  => { type => SCALAR, optional => 1, default  => C_NDMP_DUMP_OPTION, },  
				    expected 	  => { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },
	} );

	if($args{src}) {
		$args{src}=$args{src}->{volume}->_get_underlying_vol_obj();
	} else {	
		$args{src}=$class->{_src_vol};
	}
	
	my $vol_name = $args{src}->name;
	$args{tapefiler} = $class->{_tapefiler};
	$args{data_filer} = $class->{_src_host};
	$args{tape} =  $class->{_tape_name};
	
	my ($dump_dst,$cmd,$dump_src, $device);
	if ( $class->{_tape_protocol} eq C_NDMP_TAPE_PROTOCOL ){
		logcomment("Tape found, perform ndmp backup");
		$class->{_tape}->backup(%args);
	}elsif($class->{_tape_protocol} eq C_SMTAPE_TAPE_PROTOCOL ){
		my %opt;
                $opt{src_path} = "\/vol\/".$vol_name;
                $opt{dst_path} = $class->{_tape_name};
		TapeRewind($class->{_tapefiler_conn}, $args{tape});
                $class->{_tape}->backup( %opt);
	}else {
		my $level = $args{dump_level}.$args{dump_opt};
		$dump_src = "\/vol\/".$vol_name;
		if ($class->{_tape_name}) {
			$device = $class->{_tape_name};
			TapeRewind($class->{_tapefiler_conn}, $args{tape});
			logcomment("Tape found, perform dumprestore backup");
		} else {
			$device = "\/vol\/".$vol_name."\/". $args{dump_dir};
			logcomment("Tape not found, perform dumprestore to $device");
		}
	
		dump_data($class->{_tapefiler_conn}, $args{dump_level}, $device, $dump_src, $args{dump_opt});
	}
	
	return (rc => C_RC_SUCCESS);
}

## @name ::restore 
##
## @summary Does a tape restore
##
## @description
##   Does a tape restore.
##
## @return Hash with rc 
##
## @arg dst Optional: Compression object of the volume to be restored
## 	Default:  destination volume
##
## @arg version Optional: Ndmp version used for doing ndmp restore
##	Default:  4
##
## @arg mover_record Optional: mover record size
##	Default:  64512
##
## @arg mover_window  Optional: mover window size
##	Default:  -1
##
## @arg isIncr  Optional: Is increment restore
##	Default:  0
##
## @arg dump_dir Optional: for dump/restore only, the dir to be restore to 
##	Default:  dump_dir
##
## @arg break Optional: for smtape only, break the snapmirror relationship after restore 
##	Default:  1
##
## @arg restore_opt  Optional: for dump/restore only, restore options
##	Default:  'rDf'
##
## @examples
##	     $tape->restore();
##
sub restore
{
	my $class = shift;
	my (%args) = validate (@_, {dst 	  => { type =>OBJECT, isa => "Compression::Compress", optional => 1,},
				    version 	  => { type => SCALAR, optional => 1, },  
				    mover_record  => { type => SCALAR, optional => 1, },  
				    mover_window  => { type => SCALAR, optional => 1, },  
				    isIncr 	  => { type => BOOLEAN, optional => 1, },  
				    dump_dir	  => { type => SCALAR, optional => 1, default  => C_NDMP_DUMP_DIR, },
				    break	  => { type => BOOLEAN, optional => 1, default  => 1 },
				    expected 	  => { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },
	} );
	
	if($args{dst}) { 
		$args{dst}=$args{dst}->{volume}->_get_underlying_vol_obj();
	} else {
		$args{dst}=$class->{_dst_vol};
	}
	
	$args{tapefiler} = $class->{_tapefiler};
	$args{data_filer} = $class->{_dst_host};

	my $vol_name = $args{dst}->name;
	$args{tape} =  $class->{_tape_name};
	my $restore_dst;
	
	if ( $class->{_tape_protocol} eq C_NDMP_TAPE_PROTOCOL ){
		logcomment("Tape found, perform ndmp restore");
		TapeRewind($class->{_dst_filer_conn}, $args{tape});
		eval {
			$class->{_tape}->restore(%args);
		};
	} elsif ($class->{_tape_protocol} eq C_SMTAPE_TAPE_PROTOCOL) {
		my %opt;
                $opt{dst_path} = "\/vol\/".$vol_name;
                $opt{src_path} = $class->{_tape_name};
                $args{dst}->restrict();
		eval {
			$class->{_tape}->restore( %opt);
		};
	} else {
		$restore_dst="\/vol\/".$vol_name;
		my $device;
		if ( $class->{_tape_name} )  {
			logcomment("Tape found, perform dumprestore restore");
			TapeRewind($class->{_dst_filer_conn}, $args{tape});
			$device = $class->{_tape_name};
		} else {
			$device = "\/vol\/".$class->{_src_vol}->name."\/". $args{dump_dir};
			logcomment("Tape not found, perform ndmp dumprestore, restore to $device");
		}
		eval {
			restore_data($class->{_src_filer_conn}, $device, $restore_dst);
		}
	}
	
	if ($@ ){
		if( $args{expected} eq C_RC_FAILURE){
			logcomment(C_NDMP_TAPE_PROTOCOL." restore failed as expected \n $@ ");
			return (rc => C_RC_SUCCESS);
		} else {
			logresult("fatal", C_NDMP_TAPE_PROTOCOL." restore failed: \n $@");
		}
	}
	
	if ($class->{_tape_protocol} eq C_SMTAPE_TAPE_PROTOCOL) {
		$class->{_src_vol}->online();
		if($args{break}){
			$class->_mstape_snapmirror_break(dst=>$args{dst});
		}
	}
	
	return (rc => C_RC_SUCCESS);
}

## @name ::ndmpcopy 
##
## @summary perform a ndmpcopy
##
## @description
##   Does a NDMP copy from src volume to dst volume 
##
## @return Hash with rc 
##
## @arg src_vol Optional: Compression object of the ndmpcopy source volume
## 	Default:  source volume
##
## @arg dst_vol Optional: Compression object of the ndmpcopy destination volume
## 	Default:  destination volume
##
## @arg snapshot Optional: snap name on the source volume,
##		 If given then src path given for ndmp copy will be /vol/<volname>/.snapshot/<snapshot>
##
## @arg timeout Optional: ndmpcopy timeout,
##	Default:  -1 (no time out)
##
## @arg isIncr Is an incremental copy,
##	Default:  0 
##
## @examples
##	     $tape->ndmpcopy(src_vol=>$src_vol, dst_vol=>$dst_vol);
##
sub ndmpcopy
{
	my $class = shift;
	my (%args) = validate (@_, {src_vol 	  => { type =>OBJECT, isa => "Compression::Compress", optional => 1,},
				    dst_vol  	  => { type =>OBJECT, isa => "Compression::Compress", optional => 1,  },
				    snapshot 	  => { type => BOOLEAN, optional => 1, },
				    timeout	  => { type => SCALAR, optional => 1, default  => -1 },
				    isIncr	  => { type => SCALAR, optional => 1, default  => 0 },
				    expected 	  => { type => SCALAR, optional => 1, default  => C_RC_SUCCESS, },
	} );
	
	my ($srcVol, $dstVol, $srcVol_conn, $dstVol_conn, $srcUser, $srcPassword, $dstUser,$dstPassword,
	     $src_host, $dst_host, $src_path, $dst_path, $ret);
	if($args{src_vol}) {
		$srcVol_conn=$args{src_vol}->{filer}->_get_underlying_filer_obj->conn;
		$srcVol=$args{src_vol}->{volume}->_get_underlying_vol_obj();
		$src_host = $args{src_vol}->{filer}->{host_name};
	} else {
		$srcVol_conn=$class->{_src_filer_conn};
		$srcVol=$class->{_src_vol};
		$src_host = $class->{_src_host};
	}
	
	if($args{dst_vol}) {
		$dstVol_conn=$args{dst_vol}->{filer}->_get_underlying_filer_obj->conn;
		$dstVol=$args{dst_vol}->{volume}->_get_underlying_vol_obj();
		$dst_host = $args{dst_vol}->{filer}->{host_name};
	} else {
		$dstVol_conn=$class->{_dst_filer_conn};
		$dstVol=$class->{_dst_vol};
		$dst_host = $class->{_dst_host};
	}
	
	$srcUser=$srcVol_conn->{username};
	$srcPassword=$srcVol_conn->{password};
	$dstUser=$dstVol_conn->{username};
	$dstPassword=$dstVol_conn->{password};
	$dst_path = $dstVol->mount_path;

	if ($class->{_tape_protocol} ne C_NDMP_TAPE_PROTOCOL){
		$class->{_tape}->enable();
		$class->{_tape}->enable(destination  => 1);
		$class->{_tape}->set_access(access=>C_NDMP_ACCESS);
		$class->{_tape}->set_access(destination=>1, access=>C_NDMP_ACCESS);
		$class->{_tape}->set_auth_type(authtype=>C_NDMP_AUTHTYPE);
		$class->{_tape}->set_auth_type(destination=>1, authtype=>C_NDMP_AUTHTYPE);
		$class->{_tape}->set_version(destination=>1, version=>C_NDMP_VERSION);
		$class->{_tape}->set_version(destination=>1, version=>C_NDMP_VERSION);
	}
	
	if($args{snapshot})
	{
		$src_path=$srcVol->mount_path."\/\.snapshot\/".$args{snapshot};
	}
	else { $src_path=$srcVol->mount_path; }
    

	if ($src_host eq $dst_host)
        {
		logcomment( "Calling local NDMP Copy, both src and dst are on $class->{_src_host};  $src_path ");
		$ret = NdmpCmds::localNdmpCopy($srcVol_conn,$src_path,$dst_path,
					$args{timeout} ,0,$args{isIncr});
        }
	else {
		logcomment( "Calling remote  NDMP Copy, src is on $src_host and dst is on $dst_host");
		$ret = NdmpCmds::remoteNdmpCopy($srcVol_conn, $src_host, $dst_host,
					 $srcUser,$dstUser, $srcPassword,$dstPassword,$src_path,$dst_path,
					 $args{timeout} ,0,$args{isIncr});
        }

	if ($ret ) {  
		return (rc => C_RC_SUCCESS);
	} else {		
		logresult("fatal", "Failed to make ndpmcopy from $src_path to $dst_path \n");
	}	
}

## @name ::is_tape_available 
##
## @summary Check if tape is available
##
## @description
##   Check if tape is available
##
## @return Hash with and tape name array 
##
##
## @examples
##    $tape->is_tape_available;
##
sub is_tape_available{
	my $class = shift;
	my (%args) = validate (@_, { } );

	my $filer_conn=$class->{_tapefiler}->{_conn};
	my @tapes = getTapes($filer_conn);

	my %ret;

	if (@tapes) {
		$ret{rc} = C_RC_SUCCESS;
		$ret{tape} = \@tapes;
	} else {
		return $ret{rc} = C_RC_FAILURE;
	}
	return %ret;
}

## internal method
sub _get_tape{
	my $class = shift;
	my (%args) = validate (@_, { } );
	
	my %tapes = $class->is_tape_available();
	my $count = 0;
	my $tape;
	if ($tapes{rc} eq C_RC_SUCCESS){
		foreach ( @{$tapes{tape}} ) {
			if ($_ =~ /^n\w+$/) {
				return $_;
			}
		}
		logresult("fatal", "No non-rewindable tape found, check tape drive and tape status");

	} else {
		return undef;
	}
}

sub _mstape_snapmirror_break {
	my $class = shift;
	my (%args) = validate (@_, {dst	  => { type =>OBJECT, isa => "NOTE::Volume" } } );

	SnapMirror::snapmirror_break(
			host=>$args{dst}->Filer->hostrec,
			dstpath=>$args{dst}->hostpath);
}


1;
