#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI.pm#1 $
#

## @summary
## This module provides UI (CLI, ZAPI) utilities for Compression modules.
##
## @description
## This module is one place for all the compression commands, and output parsing capabilities.
## It will provide support for CLI, ZAPI, 7Mode, CMode for all compression-related commands.
## Its main purpose is to act as a tool for the Compression modules.
## Scripts should not use this module directly.
## It can be used, though, by other libraries, to formulate and parse the output of compression commands.
##
## @synopsis
##
## use Compression::Constants;
## use Compression::UI;
## use Compression::Version;
## use NOTE::Filer;
##
## my $filer = NOTE::Filer->connect( host => $FILER );
## my $version = Compression::Version->new( conn => $filer );
## my $obj = Compression::UI->new( version => $version );
##
## @examples 
## my $command = $obj->formulate_command(code => C_CMD_DECOMPRESS_START, volume => $vol_name );
## # Use own means to send command to the filer...
## my $output = ...
## $obj->process_output_string(code => C_CMD_DECOMPRESS_START,
##                             output => $output,
##                             expected_rc => C_RC_NO_DATA_TO_DECOMPRESS,
##                             log_message => "Start decompression of volume $vol_name",);
## 
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::UI;


use strict;
use warnings;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Tharn;
use Module::Common;
use Compression::Constants qw(:all);
use Compression::UI::Constants;
use Compression::UI::ConsoleMessages;
use Compression::UI::CLI;
use Compression::UI::CLI::Commands;
use Compression::UI::ZAPI;


BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}


## @name ::new
##
## @summary Compression::UI Constructor.
##
## @return Object of type Compression::UI.
##
## @arg version Required: Compression::Version object for the hosting filer.
## 
## @arg boot_mode Optional: Boot mode of the hosting filer.
## Defaults to 7M.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::UI->new( version => $version );
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa => "Compression::Version",},
                
                boot_mode => {type     => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,},
                
                debug => {type => SCALAR,
                          optional => 1,
                          default  => FALSE,},      
	} );

	my $self = {};
        $self->{_debug} = $args{debug};
        $self->{version} = $args{version};
        $self->{boot_mode} = $args{boot_mode};
        # This flag 
        $self->{_vfiler} = FALSE;
        
	$self->{cli} = Compression::UI::CLI->new(version => $self->{version},
                                                 boot_mode => $self->{boot_mode},
                                                 debug => $self->{_debug},
                                                );
	$self->{constants} = Compression::UI::Constants->new(version => $self->{version},
                                                 boot_mode => $self->{boot_mode},
                                                 debug => $self->{_debug},
                                                );
        
        $self->{console_messages} = Compression::UI::ConsoleMessages->new(version => $self->{version},
                                                                          boot_mode => $self->{boot_mode},
                                                                          debug => $self->{_debug},
                                                                          );
        
	$self->{zapi} = Compression::UI::ZAPI->new(version => $self->{version},
                                                   boot_mode => $self->{boot_mode},
                                                   debug => $self->{_debug},
                                                   );

	bless $self, $class;
	return $self;
}


## @name ::formulate_command
## @summary Generates release-specific CLI/ZAPI command.
##
## @description
## Wrapper for generating CLI/ZAPI commands for 7G/7M/CM.
##
## @return The command string.
##
## @arg ui Optional: UI type. Defaults to CLI.
##
## @arg code Required: Code corresponding to the operation for which the command is generated.
## Operation codes are defined in Compression::Constants.
##
## @arg volume Optional: Name of the volume, to be used for volume-specific commands.
##
## @arg modifier Optional: Modifier, to be used for space-specific commands (k/m/g).
##
## @examples 
## my $command = $obj->formulate_command(code => C_CMD_DECOMPRESS_START, volume => $vol_name );
## my $command = $obj->formulate_command(code => C_CMD_GET_SAVINGS, modifier => 'k');
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub formulate_command
{
	my $class = shift;
	my %args = validate(@_,
	{
		ui 	  => { type => SCALAR, optional => 1, default => $main::UI_TYPE, },
                code      => { type => SCALAR, },
                volume    => { type => SCALAR|UNDEF, optional => 1, },
                vserver   => { type => SCALAR|UNDEF, optional => 1, },
                file_path => { type => SCALAR|UNDEF, optional => 1, },
                fbn       => { type => SCALAR|UNDEF, optional => 1, },
                modifier  => { type => SCALAR|UNDEF, optional => 1, },
                size      => { type => SCALAR|UNDEF, optional => 1, },
                increment => { type => SCALAR|UNDEF, optional => 1, },
                state     => { type => SCALAR|UNDEF, optional => 1, },
                options   => { type => HASHREF|UNDEF, optional => 1, },
		snapshot  => { type => SCALAR|UNDEF, optional => 1, },
		restore_path => { type => SCALAR|UNDEF, optional => 1, },
                level => { type => SCALAR|UNDEF, optional => 1, },
		aggregate => { type => SCALAR|UNDEF, optional => 1, },
		vfiler => { type => SCALAR|UNDEF, optional => 1, },
		filer => { type => SCALAR|UNDEF, optional => 1, },
	});
        
	# Cover the case when setup_core is not included
        my $ui = (defined($args{ui})) ? $args{ui} : C_DFLT_UI_TYPE;
	delete($args{ui});
	my $cmd_code = $args{code};
	my $raw_cmd = C_STR_ZAPI_NOT_AVAILABLE;
	my $cmd;
        
        if (defined($args{vserver}) && ($class->{boot_mode} ne C_ONTAP_CMODE))
        {
                logresult('script', "The vserver argument should only be used for C-mode");
        }

	logcomment("UI: formulate_command: ". Dumper(%args) ) if ($class->{_debug});
	
	if (($ui ne C_UI_CLI) && ($ui ne C_UI_ZAPI))
	{
		logresult( "script", "Invalid UI type: \"$ui\"\nLegal values: " . C_UI_CLI . ", " . C_UI_ZAPI);
	}
	
	if (($ui eq C_UI_ZAPI) && ($class->is_zapi_available(code => $cmd_code)))
	{
		$cmd = $class->{zapi}->get_command_format(%args);
	}
	else 
	{
		# Get the release specific command syntax
		$raw_cmd = $class->{cli}->{commands}->get_command_format( code => $cmd_code, );
                
                my $cmd_options = $class->{cli}->{commands}->get_options_string( code => $cmd_code,
                                                                                 options => $args{options});
                
                # Populate applicable command fields
		$cmd = $class->_replace_vars_with_values( cmd => $raw_cmd,
                                                          volume => $args{volume},
                                                          vserver => $args{vserver},
                                                          file_path => $args{file_path},
                                                          fbn => $args{fbn},
                                                          modifier => $args{modifier},
                                                          size => $args{size},
                                                          increment => $args{increment},
                                                          state => $args{state},
                                                          options => $cmd_options,
							  snapshot => $args{snapshot},
							  restore_path => $args{restore_path},
                                                          level => $args{level},
							  aggregate => $args{aggregate},
							  vfiler => $args{vfiler},
							  filer => $args{filer},);
	}
	
        logcomment("UI: formulate_command: returning: $cmd") if ($class->{_debug});
	return $cmd;
}

## @name ::get_option_definition
## @summary Generates the release-specific option name and legal values.
##
## @description
## Wrapper for generating option strings and their legal values.
##
## @return Ref to hash with option name and legal values to set the option to.
##
## @arg ui Optional: UI type. Defaults to CLI.
##
## @arg option_code Required: Code corresponding to the option.
## Option codes are defined in Compression::Constants.
##
## @examples 
## my $option_def = $obj->get_option_definition( option_code => C_OPTION_COMPRESSION );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_option_definition
{
	my $class = shift;
	my %args = validate(@_,
	{
		ui => { type => SCALAR, optional => 1, default => C_DFLT_UI_TYPE, },
                option_code => { type => SCALAR, },
	} );
        
        my $ui = $args{ui};

	logcomment("UI: get_option_def input: ". Dumper(%args) ) if ($class->{_debug});

	if( ( $ui eq C_UI_CLI )) 
	{
		return $class->{cli}->{options}->get_option_definition( option_code => $args{option_code} );
	}
	elsif( ( $ui eq C_UI_ZAPI ))
	{
		logresult( "script", "ZAPI is not supported for options" );
	}
	else
	{
		logresult( "script", "Invalid UI type: " . $args{ui} . ". Available: ". C_UI_CLI );
	}
}


## @name ::get_fiji_point_definition
## @summary Generates the release-specific fiji point definition.
##
## @description
## Wrapper for generating fiji point definition.
##
## @return Ref to hash with fiji point tag, prog, param.
##
## @arg ui Optional: UI type. Defaults to CLI.
##
## @arg tag Required: Tag corresponding to the fiji point.
## Fiji tags are defined in Compression::Constants.
##
## @examples 
## my $fiji_def = $obj->get_fiji_point_definition( tag => C_FIJI_TAG_DISABLE_CP );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_fiji_point_definition
{
	my $class = shift;
	my %args = validate(@_,
	{
		ui => { type => SCALAR, optional => 1, default => C_DFLT_UI_TYPE, },
                tag => { type => SCALAR, },
	} );
        
        my $ui = $args{ui};

	logcomment("UI: get_fiji_point_definition input: ". Dumper(%args) ) if ($class->{_debug});

	if( ( $ui eq C_UI_CLI )) 
	{
		return $class->{cli}->{fiji_points}->get_definition_hash( tag => $args{tag} );
	}
	elsif( ( $ui eq C_UI_ZAPI ))
	{
		logresult( "script", "ZAPI is not supported for fiji points" );
	}
	else
	{
		logresult( "script", "Invalid UI type: " . $args{ui} . ". Available: ". C_UI_CLI );
	}
}

## @name ::get_constant_value
## @summary Generates the release-specific value
##
## @description
## Wrapper for generating release-specific option values
##
## @return constant value
##
## @arg ui Optional: UI type. Defaults to CLI.
##
## @arg constant_code Required: Code corresponding to the constant.
##
## @examples 
## my $constant = $obj->get_constant_value( constant_code => 'C_OPTION_COMPRESSION' );
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_constant_value
{
	my $class = shift;
	my %args = validate(@_,
	{
		constant_code => { type => SCALAR, },
	} );
        
	logcomment("UI: get_constant_value input: ". Dumper(%args) ) if ($class->{_debug});

	return $class->{constants}->get_constant_value( constant_code => $args{constant_code} );
}


## @name ::check_for_expected_rc
## @summary Checks if obtained/expected return codes match.
##
## @description
## Output processing tool. Provides standardized subroutine output and
## pass/fail behavior for the Compression library.
## Logs fatal failure if 'obtained_rc' is different than 'expected_rc'.
## This behavior can be disabled by passing 'expected_rc' as C_RC_FAILURE_NON_FATAL.
## In this case, a failure to match the two return codes will not result in script termination.
## Instead, the obtained rc is passed to calling subroutine, for further processing.
##
## @return Hash with rc and output.
##
## @arg obtained_rc Required: Obtained return code.
##
## @arg expected_rc Required: Expected return code.
##
## @arg log_message Required: Message to be logged in case of failure.
##
## @arg output Required: Command output.
## In case of successful match, the output will be returned to the script.
## In case of failure, the output will be attached to the log message.
##
## @examples 
## my $option_def = $obj->check_for_expected_rc(obtained_rc => $rc,
##                                              expected_rc => $expected,
##                                              log_message => "Set option $option_name to $val_to_set",
##                                              output      => $new_setting);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub check_for_expected_rc
{
	my $class = shift;
	my %args = validate(@_,
	{
		obtained_rc => { type => SCALAR, },
                expected_rc => { type => SCALAR, },
                log_message => { type => SCALAR, },
                output      => { type => UNDEF|SCALAR|ARRAYREF|HASHREF, },  # May need to remove type checking for output if it gets too complicated.
                                                                            # At this point, output is just for logging failures.
	} );
        
        my %return_hash = ();
        my $rc = $args{obtained_rc};
        my $expected = $args{expected_rc};

	logcomment("UI: check_for_expected_rc: ". Dumper(%args) ) if ($class->{_debug});

	if ($expected == C_RC_FAILURE_NON_FATAL)
        {
                # Non-fatal failures, needed for operations like toggle to work
                # Do nothing here. Need $rc to stay unchanged
        }
        elsif ($rc == $expected)
        {
                $rc = C_RC_SUCCESS;
        }
        else
        {
                if (($rc == C_RC_WAFLIRON_CORRUPTION_FOUND) || ($rc == C_RC_WACK_ERRORS_FOUND))
                {
                        logcomment("##########################################################\n" .
                                   "##########################################################\n" .
                                   "##########################################################\n\n" .
                                   "Unexpected filesystem corruption found. Disabling cleanup.\n\n" .
                                   "##########################################################\n" .
                                   "##########################################################\n" .
                                   "##########################################################\n");
                        $ENV{'C_CORRUPTION_FOUND'} = TRUE;
                }
                my $fail_comment = "Failure while executing \"" . $args{log_message} . "\" . Expected $expected, got $rc";
                if (defined($args{output}))
                {
                        $fail_comment .= "\n\n\nCommand output:\n\n" . Dumper($args{output});
                }
                
                logresult("fatal", $fail_comment);
        }
        
        $return_hash{rc} = $rc;
        $return_hash{output} = $args{output};
        
        return %return_hash;
}


## @name ::process_output_string
## @summary Wrapper for command output processing.
##
## @description
## Wrapper for command output processing.
## Checks wether the output is defined and fails failure accordingly.
## Checks if the command should be formatted, and if it is, the output is processed
## and all the information is automatically formatted into a hash with the predefined keys.
## For unformatted command output, the output string is matched against a list of known strings
## and the corresponding return code is obtained.
## The obtained return code is then checked against 'expected_rc' according to the rules described in
## check_for_expected_rc.
##
## @return Hash with rc and output.
##
## @arg ui Optional: UI type. Defaults to CLI.
##
## @arg code Required: Code corresponding to the operation for which the output is parsed.
## Operation codes are defined in Compression::Constants.
##
## @arg modifier Optional: Modifier, to be used for space-specific commands (k/m/g).
##
## @arg required_return_hash_key Optional: For formatted command output that need to have an entry
## for a specific entity (for example, space savings for a specific volume). If required_return_hash_key
## is defined, but the command output did not contain an entry for the specified entity,
## then the script is terminated with a fatal error.
##
## @arg disable_data_identifier_processing Optional: Flag to disable any processing on the data identifier
## (First entry in the data line, for example: /vol/vol_name for 'df -S').
##
## @arg expected_rc Required: Expected return code.
##
## @arg log_message Required: Message to be logged in case of failure.
##
## @arg output Required: Command output.
## In case of successful match, the output will be returned to the calling subroutine.
## In case of failure, the output will be attached to the log message.
##
## @examples
## my %result = $obj->process_output_string(code => C_CMD_GET_SAVINGS,
##                                          output => $output,
##                                          required_return_hash_key => $volume,
##                                          modifier => $modifier,
##                                          log_message => "Obtain space savings for volume $volume",);
## my %result = $obj->process_output_string(code => C_CMD_DECOMPRESS_START
##                                          output => $output,
##                                          expected_rc => $expected,
##                                          log_message => "Start decompression of volume $vol_name",);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub process_output_string
{
	my $class = shift;
	my %args = validate(@_,
	{
		ui => { type => SCALAR, optional => 1, default => $main::UI_TYPE, },
                code => { type => SCALAR, },
                output => { type => SCALAR|UNDEF|OBJECT, },
                modifier => {type => SCALAR, optional => 1,},
                required_return_hash_key => {type => SCALAR|UNDEF, optional => 1,},
                disable_data_identifier_processing => {type => BOOLEAN, optional => 1, default => FALSE, },
                expected_rc => { type => SCALAR, optional => 1, default => C_RC_SUCCESS, },
                log_message => { type => SCALAR, },
	} );
        
        # Cover the case when setup_core is not included
        my $ui = (defined($args{ui})) ? $args{ui} : C_DFLT_UI_TYPE;
        my $code = $args{code};
        my $expected_rc = $args{expected_rc};
        my $log_message = $args{log_message};
	delete($args{ui});
        delete($args{expected_rc});
        delete($args{log_message});
        
        my %return_hash = ( rc => C_RC_FAILURE, output => $args{output});

	logcomment("UI: format_output: " . Dumper(%args) ) if ($class->{_debug});
	
	if (($ui ne C_UI_CLI) && ($ui ne C_UI_ZAPI))
	{
		logresult( "script", "Invalid UI type: \"$ui\"\nLegal values: " . C_UI_CLI . ", " . C_UI_ZAPI);
	}
	
	if (($ui eq C_UI_ZAPI) && ($class->is_zapi_available(code => $code)))
	{
		%return_hash = $class->{zapi}->process_output(%args);
	}
	else
	{
                %return_hash = $class->{cli}->cli_process_output_string(%args);
	}
	
	my %checked_result = $class->check_for_expected_rc(obtained_rc => $return_hash{rc},
							   expected_rc => $expected_rc,
							   output      => $args{output},
							   log_message => $log_message,);
	# Over-write the return rc with the 'checked' rc
	$return_hash{rc} = $checked_result{rc};
        
        logcomment("UI: process_output_string returning:\n" . Dumper(%return_hash)) if ($class->{_debug});
        return %return_hash;
}


## @name ::parse_string_for_rc
## @summary Parse a string to get correspondent return code.
##
## @description
## Parse a string to get correspondent return code.
##
## @return Return code.
##
## @arg ui Optional: UI type. Defaults to CLI.
##
## @arg code Required: Code corresponding to the operation for which the output is parsed.
## Operation codes are defined in Compression::Constants.
##
## @arg string Required: String to be parsed.
##
## @examples
## my %result = $obj->parse_string_for_rc(code => C_CMD_GET_SAVINGS,
##                                        string => $my_string,);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub parse_string_for_rc
{
	my $class = shift;
	my %args = validate(@_,
	{
		ui => { type => SCALAR, optional => 1, default => $main::UI_TYPE, },
                code => { type => SCALAR, },
                string => { type => SCALAR, },
	} );
        
        # Cover the case when setup_core is not included
	my $ui = (defined($args{ui})) ? $args{ui} : C_DFLT_UI_TYPE;
	my $code = $args{code};

	logcomment("UI: parse_string_for_rc: " . Dumper(%args) ) if ($class->{_debug});
	
	if (($ui ne C_UI_CLI) && ($ui ne C_UI_ZAPI))
	{
		logresult( "script", "Invalid UI type: \"$ui\"\nLegal values: " . C_UI_CLI . ", " . C_UI_ZAPI);
	}

	if (($ui eq C_UI_ZAPI) && ($class->is_zapi_available(code => $code)))
	{
		return $class->{zapi}->{messages}->get_return_code(output => $args{string},
								   code => $args{code});
	}
	else
	{
		return $class->{cli}->{messages}->get_return_code(output => $args{string},
								  code => $args{code});
	}
}


## @name ::get_wreck_inject_cmd
## @summary Generates the injection command to use for wreck.
##
## @description
## Generates the injection command to use for wreck.
##
## @return Injection command string.
##
## @arg code Required: Code corresponding to the error to be injected (defined in Constants.pm).
##
## @arg filename Optional: Name of the file to inject the error to.
## Ignored for non file-specific corruptions.
##
## @arg bno Optional: Specific block number within a file to inject corruption at.
## Ignored for non file-specific corruptions.
##
## @arg start Optional: Start byte for corruption within a block.
## Ignored for non file-specific corruptions. Will be auto-generated if not passed in (and needed).
##
## @arg end Optional: End byte for corruption within a block.
## Ignored for non file-specific corruptions. Will be auto-generated if not passed in (and needed).
##
## @arg fill Optional: Byte to be used when corrupting a specific block pattern.
## Ignored for non file-specific corruptions.  Will be auto-generated if not passed in (and needed).
##
## @arg cg_no Optional: CG number at which to injectthe corruption.
## This argument is only accounted for in the case of corrupting the CG (header, trailer).
##
## @arg new_bno Optional: Value to be used to corrupt an indirect block (ind_vol_bno).
##
## @arg new_bno Optional: Value to be used to corrupt an indirect block vvbn (ind_vol_vbno).
##
## @note
## For more detail about accepted/required argument combinations, please see Compression::Volume::wreck.
##
## @examples 
## my $inject_cmd = $obj->get_wreck_inject_cmd(code => C_INJECTION_CMD_CODE_FSINFO_INOS_USED);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub get_wreck_inject_cmd
{
	my $class = shift;
	my %args = validate(@_,
	{
		code => { type => SCALAR, },
                filename => { type => SCALAR, optional => 1,},
                bno => { type => SCALAR|UNDEF, optional => 1},
                start => { type => SCALAR, optional => 1},
                end => { type => SCALAR, optional => 1},
                fill => { type => SCALAR|UNDEF, optional => 1},
                cg_no => { type => SCALAR|UNDEF, optional => 1},
                new_bno => { type => SCALAR|UNDEF, optional => 1},
                new_vbno => { type => SCALAR|UNDEF, optional => 1},
		level => { type => SCALAR|UNDEF, optional => 1},
	} );
        
        my $injection_cmd;
        my $code = $args{code};
        my $boot_mode = $class->{boot_mode};
        
        # For now, we keep track of commands in here.
        # If release-dependent processing becomes necessary, this method will become the ui wrapper
        
	if ($code eq C_CORRUPT_FSINFO_INOS_USED)
	{
		$injection_cmd = 'fsinfo_inos_used count=' . int(rand(10000));
	}
	elsif ($code eq C_CORRUPT_FSINFO_BLKS_USED_BY_PLANE0)
	{
		$injection_cmd = 'fsinfo_blks_used_by_plane0 count=' . int(rand(10000));
	}
	elsif ($code eq C_CORRUPT_FSINFO_BLKS_COMPRESS_HOLES)
	{
		$injection_cmd = 'fsinfo_blks_compress_holes count=' . int(rand(10000));
	}
	elsif ($code eq C_CORRUPT_INODE_BLK_CNT)
	{
		if (!defined($args{filename}))
		{
			logresult("script", "Wreck command $code requires a filename");
                }
		$injection_cmd = 'inode_blk_cnt inum=path:/' . $args{filename} . ' count=' . int(rand(100));
	}
	elsif ($code eq C_CORRUPT_REFCOUNT)
	{
		if (!defined($args{filename}))
		{
			logresult("script", "Wreck command $code requires a filename");
                }
		my $bno = (defined($args{bno})) ? $args{bno} : 1;
		$injection_cmd = 'refcount_set inum=path:/' . $args{filename} . ' bno=' . $bno . ' count=' . int(rand(1000));
        }
	elsif ($code eq C_CORRUPT_CHECKSUM)
	{
		if (!defined($args{filename}))
		{
			logresult("script", "Wreck command $code requires a filename");
                }
		my $bno = (defined($args{bno})) ? $args{bno} : 1;
		my $start = C_CHECKSUM_START_BYTE;
		my $end = C_CHECKSUM_END_BYTE;
		my $fill = (defined($args{fill})) ? $args{fill} : int(rand(255));
		$injection_cmd = 'block_pattern inum=path:/' . $args{filename} . " bno=$bno start=$start end=$end fill=$fill";
        }
	elsif ($code eq C_CORRUPT_CG_HEADER_VBN)
	{
		if (!defined($args{filename}) || !defined($args{new_bno}))
		{
			logresult("script", "Wreck command $code requires the following attributes: filename, new_bno");
                }
		my $cg_no = (defined($args{cg_no})) ? $args{cg_no} : 0;
		my $bno = $cg_no * C_DFLT_CG_BLKS;
		$injection_cmd = 'ind_vol_bno inum=path:/' . $args{filename} . " level=1 bno=$bno new_bno=" . $args{new_bno} . " index=0";
        }
	elsif ($code eq C_CORRUPT_CG_HEADER_VVBN)
	{
		if (!defined($args{filename}) || !defined($args{new_vbno}))
		{
			logresult("script", "Wreck command $code requires the following attributes: filename, new_vbno");
                }
		my $cg_no = (defined($args{cg_no})) ? $args{cg_no} : 0;
		my $bno = $cg_no * C_DFLT_CG_BLKS;
		$injection_cmd = 'ind_vol_vbno inum=path:/' . $args{filename} . " level=1 bno=$bno new_vbno=" . $args{new_vbno} . " index=0";
        }
	elsif ($code eq C_CORRUPT_CG_TRAILER_VBN)
	{
		if (!defined($args{filename}) || !defined($args{new_bno}))
		{
			logresult("script", "Wreck command $code requires the following attributes: filename, new_bno");
                }
		my $cg_no = (defined($args{cg_no})) ? $args{cg_no} : 0;
		my $bno = $cg_no * C_DFLT_CG_BLKS; # The block number is the one of the header. The index will point to the trailer.
		my $trailer_index = C_DFLT_CG_BLKS - 1;
		$injection_cmd = 'ind_vol_bno inum=path:/' . $args{filename} . " level=1 bno=$bno new_bno=" . $args{new_bno} . " index=$trailer_index";
        }
	elsif ($code eq C_CORRUPT_CG_TRAILER_VVBN)
	{
		if (!defined($args{filename}) || !defined($args{new_vbno}))
		{
			logresult("script", "Wreck command $code requires the following attributes: filename, new_vbno");
		}
		my $cg_no = (defined($args{cg_no})) ? $args{cg_no} : 0;
		my $bno = $cg_no * C_DFLT_CG_BLKS; # The block number is the one of the header. The index will point to the trailer.
		my $trailer_index = C_DFLT_CG_BLKS - 1;
		$injection_cmd = 'ind_vol_vbno inum=path:/' . $args{filename} . " level=1 bno=$bno new_vbno=" . $args{new_vbno} . " index=$trailer_index";
        }
	elsif ($code eq C_CORRUPT_CG_DATA_VBN)
	{
		if (!defined($args{filename}) || !defined($args{new_bno}))
		{
			logresult("script", "Wreck command $code requires the following attributes: filename, new_bno");
		}
		my $cg_no = (defined($args{cg_no})) ? $args{cg_no} : 0;
		my $bno = $cg_no * C_DFLT_CG_BLKS; # The block number is the one of the header. The index will point to the first data block.
		my $trailer_index = 1;
		$injection_cmd = 'ind_vol_bno inum=path:/' . $args{filename} . " level=1 bno=$bno new_bno=" . $args{new_bno} . " index=$trailer_index";
        }
	elsif ($code eq C_CORRUPT_CG_DATA_VVBN)
	{
		if (!defined($args{filename}) || !defined($args{new_vbno}))
		{
			logresult("script", "Wreck command $code requires the following attributes: filename, new_vbno");
                }
		my $cg_no = (defined($args{cg_no})) ? $args{cg_no} : 0;
		my $bno = $cg_no * C_DFLT_CG_BLKS; # The block number is the one of the header. The index will point to the first data block.
		my $trailer_index = 1;
		$injection_cmd = 'ind_vol_vbno inum=path:/' . $args{filename} . " level=1 bno=$bno new_vbno=" . $args{new_vbno} . " index=$trailer_index";
        }
	elsif ($code eq C_CORRUPT_REP)
	{
		if (!defined($args{filename}))
		{
			logresult("script", "Wreck command $code requires a filename");
                }
		my $bno = (defined($args{bno})) ? $args{bno} : 0;
		my $level = (defined($args{level})) ? $args{level} : 1;
		$injection_cmd = 'pseudo_bad inum=path:/' . $args{filename} . " level=$level bno=$bno";
        }
	else
	{
		logresult("script", "Unsupported error injection command code: $code");
	}
        
        logcomment("get_wreck_inject_cmd: returning injection cmd: $injection_cmd") if ($class->{_debug});
        
        return $injection_cmd;
}


## @name ::get_corruption_messages
## @summary Generates the corruption messages needed for error detection using wafliron, wack.
##
## @description
## Wrapper for generating corruption messages needed for error detection using wafliron, wack.
##
## @return Ref to array of corruption messages.
##
## @arg ui Optional: UI type. Defaults to CLI.
##
## @arg code Required: Code corresponding to the corruption (defined in Constants.pm).
##
## @examples 
## my $messages_array_ref = $obj->get_corruption_messages( code => C_CORRUPT_FSINFO_INOS_USED);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_corruption_messages
{
	my $class = shift;
	my %args = validate(@_,
	{
		ui => { type => SCALAR, optional => 1, default => C_DFLT_UI_TYPE, },
                detection_method => { type => SCALAR, regex => qr/^(wack|wafliron)$/ },
                code => { type => SCALAR, },
	} );
        
        my $ui = $args{ui};
        my $corruption_detection_method = $args{detection_method};

	logcomment("UI: get_corruption_messages input: ". Dumper(%args) ) if ($class->{_debug});

	if( ( $ui eq C_UI_CLI )) 
	{
                if ($corruption_detection_method eq 'wafliron')
                {
                        return $class->{cli}->{corruption_messages}->get_wafliron_corruption_message(code => $args{code});
                }
                elsif ($corruption_detection_method eq 'wack')
                {
                        return $class->{cli}->{corruption_messages}->get_wack_corruption_message( code => $args{code} );
                }
                else
                {
                        # Should never get here...
                        logresult( "script", "Unsupported corruption detection method: $corruption_detection_method" );
                }
	}
	elsif( ( $ui eq C_UI_ZAPI ))
	{
		logresult( "script", "ZAPI not yet implemented" );
	}
	else
	{
		logresult( "script", "Invalid UI type: " . $args{ui} . ". Available: ". C_UI_CLI );
	}
}


## @name ::get_perf_stats_counter
## @summary Generates the perf stats counter string for this release.
##
## @description
## Generates the perf stats counter string for this release.
##
## @return The perf stats counter string for this release.
##
## @arg code Required: Code corresponding to the perf stats counter (defined in Constants.pm).
##
## @examples 
## my $messages_array_ref = $obj->get_perf_stats_counter( code => );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_perf_stats_counter
{
        my $class = shift;
	my %args = validate(@_,
	{
		code => { type => SCALAR, },
                filename => { type => SCALAR, optional => 1,},
                bno => { type => SCALAR|UNDEF, optional => 1},
                start => { type => SCALAR, optional => 1},
                end => { type => SCALAR, optional => 1},
                fill => { type => SCALAR|UNDEF, optional => 1},
                cg_no => { type => SCALAR|UNDEF, optional => 1},
                new_bno => { type => SCALAR|UNDEF, optional => 1},
                new_vbno => { type => SCALAR|UNDEF, optional => 1},
	} );
        
        my $code = $args{code};
        my $boot_mode = $class->{boot_mode};
        
        # For now, we keep track of strings in here.
        # If release-dependent processing becomes necessary, this method will become the ui wrapper
        # TODO: need specific c-mode defs?
        
	my $counter_defs = {'C_PERF_STATS_COUNTER_BUFFERS_CACHEABLE_IN_EC' => 'Buffers cacheable in EC',
                            'C_PERF_STATS_COUNTER_DECOMPRESS_NOMEM' => 'Decompress Nomem'
                            };
        
        my $counter_str = $counter_defs->{$code};
        if (!defined($counter_str))
        {
                logresult("script", "Unsupported perf stats counter code: $code");
        }
        
        logcomment("get_perf_stats_counter returning: $counter_str") if ($class->{_debug});
        
        return $counter_str;
}


## @name ::is_dblade_level_command
## @summary Decide if a certain command needs to be executed at d-blade level (C-mode node run).
##
## @description
## Decide if a certain command needs to be executed at d-blade level (C-mode node run).
##
## @return TRUE/FALSE.
##
## @arg ui Optional: UI type. Defaults to CLI.
##
## @arg code Required: Code corresponding to the operation for which the command needs to be executed.
## Operation codes are defined in Compression::Constants.
##
## @examples
## my $need_node_run = $obj->is_dblade_level_command(code => C_CMD_GET_SAVINGS,);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_dblade_level_command
{
	my $class = shift;
	my %args = validate(@_,
	{
		ui 	  => { type => SCALAR, optional => 1, default => C_DFLT_UI_TYPE, }, 
                code      => { type => SCALAR, },
	} );
        
        my $ui = $args{ui};
        
	if( ( $ui eq C_UI_CLI )) 
	{
		return $class->{cli}->{commands}->is_dblade_level_cmd( code => $args{code}, );
	}
	elsif( ( $ui eq C_UI_ZAPI ))
	{
		return FALSE;
	}
	else
	{
		logresult( "script", "Invalid UI Type, user defined=[" . $args{ui} . "], dflt=[". C_DFLT_UI_TYPE . "]" );
	}
}


## @name ::is_vfiler_available_command
## @summary Decide if a certain command is available in the vfiler context.
##
## @description
## Decide if a certain command is available in the vfiler context.
##
## @return TRUE/FALSE.
##
## @arg code Required: Code corresponding to the command.
## Operation codes are defined in Compression::Constants.
##
## @examples
## my $is_vfiler_avail = $obj->is_vfiler_available_command(code => C_CMD_GET_SAVINGS,);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_vfiler_available_command
{
        my $class = shift;
        my %args = validate(@_,
	{
                ui 	  => { type => SCALAR, optional => 1, default => C_DFLT_UI_TYPE, },
                code      => { type => SCALAR, },
	} );
        
        my $ui = $args{ui};
        
        if( ( $ui eq C_UI_CLI ))
        {
                return $class->{cli}->{commands}->is_vfiler_available_cmd( code => $args{code}, );
	}
        elsif( ( $ui eq C_UI_ZAPI ))
        {
                return $class->{zapi}->{commands}->is_vfiler_available_cmd(code => $args{code},);
        }
        else
        {
                logresult( "script", "Invalid UI Type, user defined=[" . $args{ui} . "], dflt=[". C_DFLT_UI_TYPE . "]" );
	}
        
        
}


## @name ::is_zapi_available
## @summary Decide if a certain command is available via ZAPI.
##
## @description
## Decide if a certain command is available via ZAPI.
##
## @return TRUE/FALSE.
##
## @arg code Required: Code corresponding to the command.
## Operation codes are defined in Compression::Constants.
##
## @examples
## my $can_do_zapi = $obj->is_zapi_available(code => C_CMD_GET_SAVINGS,);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_zapi_available
{
	my $class = shift;
        my %args = validate(@_,
	{
                code      => { type => SCALAR, },
	} );
        
	my $cmd_code = $args{code};
	my $cmd = $class->{zapi}->{commands}->get_command_format(code => $cmd_code);
	if ($cmd ne C_STR_ZAPI_NOT_AVAILABLE)
	{
		return TRUE;
	}
        
	return FALSE;
}


#######################
# Internal subroutines
#######################

# Note: argument list for this method will need to be in-sync with Commands.pm
sub _replace_vars_with_values
{
        my $class = shift;
	my %args = validate(@_,
	{
		cmd 	  => { type => SCALAR, },
                volume    => { type => SCALAR|UNDEF, optional => 1, default => undef },
                vserver   => { type => SCALAR|UNDEF, optional => 1, default => undef },
                modifier  => { type => SCALAR|UNDEF, optional => 1, default => undef },
                file_path => { type => SCALAR|UNDEF, optional => 1, default => undef },
                fbn       => { type => SCALAR|UNDEF, optional => 1, default => undef },
                size      => { type => SCALAR|UNDEF, optional => 1, default => undef },
                increment => { type => SCALAR|UNDEF, optional => 1, default => undef },
                state     => { type => SCALAR|UNDEF, optional => 1, default => undef },
                options   => { type => SCALAR|UNDEF, optional => 1, default => undef },
		snapshot  => { type => SCALAR|UNDEF, optional => 1, default => undef },
		restore_path => { type => SCALAR|UNDEF, optional => 1, },
                level => { type => SCALAR|UNDEF, optional => 1, },
		aggregate => { type => SCALAR|UNDEF, optional => 1, },
		vfiler => { type => SCALAR|UNDEF, optional => 1, },
		filer => { type => SCALAR|UNDEF, optional => 1, },
	} );

	my $cmd = $args{cmd};
        delete($args{cmd});

	logcomment("UI: _replace_vars_with_values: ". Dumper(%args) ) if ($class->{_debug});

        # Now the args hash contains only pairs 'var name' - value.
        # Loop through the hash keys to substitute applicable fields in the command.
	foreach  my $key (keys %args)
	{
		my $p = "Key=[$key]";
		my $val = $args{$key};
		if ( $cmd =~ /<$key>/ )
		{
                        # Required command component.
                        # Fail if the value to replace the component is undefined
                        if ( defined $val )
                        {
                                $cmd =~ s/\<$key\>/$val/g;
                                $p .= ", replaced <$key> with \"$val\"";
                        }
                        else
                        {
                                logresult("script", "UI: Library problem found: required command component \"$key\" was not passed in. " .
                                          "Faulty command: \"$cmd\"");
                        }
		}
		elsif ( $cmd =~ /[$key]/ )
		{
                        # Optional command component.
                        # Do not fail if the value to replace the component is undefined, but still strip out the place-holder
                        if ( defined $val )
                        {
                                $cmd =~ s/\[$key\]/$val/g;
                                $p .= ", replaced [$key] with [$val]";
                        }
                        else
                        {
                                $cmd =~ s/\S*\[$key\]//g;
                                $p .= ", removed [$key]";
                        }
		}
		logcomment("UI: $p") if ($class->{_debug});
                
	}

	logcomment("UI: Cmd after resolving vars: $cmd") if ($class->{_debug});
	return $cmd;
}


1;
