#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/CLI.pm#1 $
#

## CLI utilities for Compression modules.
##
## @description
##    This module provides basic functionality for Compression CLI (7G, 7M, 10M)
##
## @synopsis
##
## use Compression::Constants;
## use Compression::UI::CLI;
## use Compression::Version;
## use NOTE::Filer;  # To obtain filer version.
##
## my $filer = NOTE::Filer->connect( host => $FILER );
## my $version = Compression::Version->new( conn => $filer );
## my $obj = Compression::UI::CLI->new( version => $version );
##
## @examples 
## $obj->cli_process_output_string(code => C_CMD_DECOMPRESS_START,
##                                 output => $output,
##                                 expected_rc => C_RC_NO_DATA_TO_DECOMPRESS,
##                                 log_message => "Start decompression of volume $vol_name",);
## 
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::UI::CLI;

use strict;
use warnings;

use Tharn;
use Module::Common;
use Data::Dumper;
use Params::Validate qw( validate SCALAR HASHREF ARRAYREF OBJECT UNDEF BOOLEAN);

use Compression::Constants qw(:all);
use Compression::UI::CLI::Commands;
use Compression::UI::CLI::Formats;
use Compression::UI::CLI::Options;
use Compression::UI::CLI::Messages;
use Compression::UI::CLI::FijiPoints;
use Compression::UI::CLI::CorruptionMessages;


BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}


## @name ::new
## @summary Compression::UI::CLI Constructor.
##
## @return Object of type Compression::UI::CLI.
##
## @arg version Required: Compression::Version object for the hosting filer.
## 
## @arg boot_mode Optional: Boot mode of the hosting filer.
## Defaults to 7M.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::UI::CLI->new( version => $version );
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub new
{
	my $class = shift;
	my %args = validate(@_,
	{	
		version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type     => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug => {type => SCALAR,
                          optional => 1,
                          default => FALSE,
                          },

	} );
	
	my $self = {};
        $self->{_debug} = $args{debug};

	$self->{name} = "CLI";
	$self->{description} = "Compression CLI commands";
	$self->{id} = $$;
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};
	
	$self->{commands} = Compression::UI::CLI::Commands->new(version => $args{version},
                                                                boot_mode => $self->{_boot_mode},
                                                                debug => $self->{_debug} );
	$self->{formats} = Compression::UI::CLI::Formats->new(version => $args{version},
                                                              boot_mode => $self->{_boot_mode},
                                                              debug => $self->{_debug});
        $self->{options} = Compression::UI::CLI::Options->new(version => $args{version},
                                                              boot_mode => $self->{_boot_mode},
                                                              debug => $self->{_debug});
        $self->{messages} = Compression::UI::CLI::Messages->new(version => $args{version},
                                                                boot_mode => $self->{_boot_mode},
                                                                debug => $self->{_debug});
        $self->{fiji_points} = Compression::UI::CLI::FijiPoints->new(version => $args{version},
                                                                     boot_mode => $self->{_boot_mode},
                                                                     debug => $self->{_debug});
        $self->{corruption_messages} = Compression::UI::CLI::CorruptionMessages->new(version => $args{version},
                                                                                     boot_mode => $self->{_boot_mode},
                                                                                     debug => $self->{_debug});
        
	bless $self, $class;

	return $self;
}


## @name ::cli_process_output_string
## @summary Wrapper for CLI command output processing.
##
## @description
## Wrapper for CLI command output processing.
## Checks wether the output is defined and fails failure accordingly.
## Checks if the command should be formatted, and if it is, the output is processed
## and all the information is automatically formatted into a hash with the predefined keys.
## For unformatted command output, the output string is matched against a list of known strings
## and the corresponding return code is obtained.
##
## @return Hash with rc and output.
##
## @arg code Required: Code corresponding to the operation for which the output is parsed.
## Operation codes are defined in Compression::Constants.
##
## @arg output Required: Command output.
##
## @arg modifier Optional: Modifier, to be used for space-specific commands (k/m/g).
##
## @arg required_return_hash_key Optional: For formatted command output that need to have an entry
## for a specific entity (for example, space savings for a specific volume). If required_return_hash_key
## is defined, but the command output did not contain an entry for the specified entity,
## then the script is terminated with a fatal error.
##
## @arg disable_data_identifier_processing Optional: Flag to disable any processing on the data identifier
## (First entry in the data line, for example: /vol/vol_name for 'df -S').
##
## @examples
## my %result = $obj->cli_process_output_string(code => C_CMD_GET_SAVINGS,
##                                              output => $output,
##                                              required_return_hash_key => $volume,
##                                              modifier => $modifier,
##                                              log_message => "Obtain space savings for volume $volume",);
## my %result = $obj->cli_process_output_string(code => C_CMD_DECOMPRESS_START
##                                              output => $output,
##                                              expected_rc => $expected,
##                                              log_message => "Start decompression of volume $vol_name",);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub cli_process_output_string
{
        my $class = shift;
	my %args = validate(@_,
	{
                code      => { type => SCALAR, },
                output    => { type => SCALAR|UNDEF, },
                modifier  => {type => SCALAR, optional => 1,},
                required_return_hash_key    => {type => SCALAR|UNDEF, optional => 1,},
                disable_data_identifier_processing => {type => BOOLEAN, optional => 1, default => FALSE, },
	} );
        
        my $code = $args{code};
        my $output = $args{output};
        my %return_hash = ( rc => C_RC_FAILURE, output => $output );
        
        # Check if output is expected to be in a formatted form or just a plain massage string
        %return_hash = $class->{formats}->check_and_process_output_if_formatted( %args );
		
        if ((defined($return_hash{rc})) && ($return_hash{rc} == C_RC_SUCCESS))
        {
                return %return_hash;
        }
        
        # We get here if:
        # - Output is expected to be just a message string.
        # - Output is expected to be formatted, but was not. I this case, still continue: try to make better sense of failure.
        $return_hash{rc} = $class->{messages}->get_return_code(output => $output,
                                                               code => $code);
        
        return %return_hash;
}


1;
