#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/CLI/Commands.pm#1 $
#

## @summary CLI Commands template for Compression.
##
## @description
##    This library provides CLI command templates for compression functionality (7G, 7M, CM).
##
## @synopsis
##
## use Compression::Constants;
## use Compression::UI::CLI::Commands;
## use Compression::Version;
## use NOTE::Filer;  # To obtain filer version.
##
## my $filer = NOTE::Filer->connect( host => $FILER );
## my $version = Compression::Version->new( conn => $filer );
## my $obj = Compression::UI::CLI::Commands->new( version => $version );
##
## @examples 
## $obj->get_command_format(code => C_CMD_DECOMPRESS_START, volume => $vol_name );
## 
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)




package Compression::UI::CLI::Commands;

use Tharn;
use Module::Common;

use strict;
use warnings;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Compression::Constants qw(:all);
use Compression::Version;


BEGIN
{
	require Exporter;
	our @ISA = qw(Exporter);
	require "Compression/base_import.pl";
};


## @name ::new
##
## @summary Compression::UI::CLI::Commands Constructor.
##
## @return Object of type Compression::UI::CLI::Commands.
##
## @arg version Required: Compression::Version object for the hosting filer.
## 
## @arg boot_mode Optional: Boot mode of the hosting filer.
## Defaults to 7M.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::UI::CLI::Commands->new( version => $version );
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
##

sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type     => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug  => {  
			type 	 => SCALAR,
			optional => 1,
                        default  => FALSE,
			},
		} );
        
	my $self = {};
        
        $self->{_debug} = $args{debug};
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};
	
	bless $self, $class;
	return $self;
}

## @name ::get_command_format
##
## @description
## Generates release agnostic/specific commands based on the template definition
## without populating any place-holders.
##
## @return Raw command string (any place-holders are left in place).
## 
## @arg code Required: Code corresponding to the operation for which the command is generated.
## Operation codes are defined in Compression::Constants.
## 
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
##

sub get_command_format
{
	my $class = shift;
	my %args = validate(@_,
	{	
		code      => { type => SCALAR, },
                
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $boot_mode = $class->{_boot_mode};
        my $command_code = $args{code};
        my $return_cmd;
        
        logcomment("Commands: get_command_format. Command lookup for code=$command_code," .
                   " ontap version $ontap_ver, boot_mode $boot_mode") if ($class->{_debug});
        
        if ($boot_mode eq C_ONTAP_7MODE)
        {
                if ($ontap_ver eq '7.3')
		{
			$return_cmd = $class->_7mode_commands_7_3( code => $command_code );
                }
		elsif ($ontap_ver eq '8.0')
		{
			$return_cmd = $class->_7mode_commands_8_0( code => $command_code );
                }
		elsif ($ontap_ver eq '8.1')
		{
			$return_cmd = $class->_7mode_commands_8_1( code => $command_code );
                }
		elsif ($ontap_ver eq '8.2')
		{
			$return_cmd = $class->_7mode_commands_8_2( code => $command_code );
                }
		else
		{
			logcomment("Commands: Specific support for sw_version $ontap_ver not found. Will be using base commands") if ($class->{_debug});
                }
        
                if ( !defined($return_cmd) )
                {
                        logcomment( "Commands: Could not find command specific for 7-mode $ontap_ver. Defaulting to base commands. " ) if ($class->{_debug});
                        $return_cmd =  $class->_7mode_base_commands( code => $command_code );
                }
        }
        elsif ($boot_mode eq C_ONTAP_CMODE)
        {
                if ($ontap_ver eq '8.1')
		{
			$return_cmd = $class->_Cmode_commands_8_1( code => $command_code );
                }
        
                if ( !defined($return_cmd) )
                {
                        logcomment( "Commands: Could not find command specific for C-mode $ontap_ver. Defaulting to base commands. " ) if ($class->{_debug});
                        $return_cmd =  $class->_Cmode_base_commands( code => $command_code );
                }
        }
        else
        {
                logresult('script', "Unrecognized boot_mode: \"$boot_mode\"");
        }

        logcomment("Commands: get_command_format. Returning command for code=$command_code (boot_mode=$boot_mode): \"$return_cmd\"")
        if ($class->{_debug});
        
        return $return_cmd;
}


## @name ::get_options_string
##
## @description
## Generates release agnostic/specific command options string.
##
## @return String representing the command options.
## 
## @arg code Required: Code corresponding to the operation for which the command options are needed.
## Operation codes are defined in Compression::Constants.
## 
## @arg options Optional: Hashref of command options, as accepted by the highest-level interface. 
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##

sub get_options_string
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => { type => SCALAR, },
                options => { type => HASHREF|UNDEF, optional => 1, },
        } );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $boot_mode = $class->{_boot_mode};
        my $command_code = $args{code};
        my $options_to_set = $args{options};
        my $allowed_options = {};
        my $return_str = '';
        
        logcomment("Commands: get_options_string. Options lookup for code=$command_code, ontap version $ontap_ver\n" .
                   "Options required to be set:\n" . Dumper($options_to_set) ) if ($class->{_debug});
        
        if ($boot_mode eq C_ONTAP_7MODE)
        {
                if ($ontap_ver eq '7.3')
		{
			$allowed_options = $class->_7mode_cmd_options_7_3( code => $command_code );
                }
		elsif ($ontap_ver eq '8.0')
		{
			$allowed_options = $class->_7mode_cmd_options_8_0( code => $command_code );
                }
		elsif ($ontap_ver eq '8.1')
		{
			$allowed_options = $class->_7mode_cmd_options_8_1( code => $command_code );
                }
		elsif ($ontap_ver eq '8.2')
		{
			$allowed_options = $class->_7mode_cmd_options_8_2( code => $command_code );
                }
		else
		{
			logcomment("Commands: get_options_string: Specific support for sw_version $ontap_ver not found. Will be using base options") if ($class->{_debug});
			$allowed_options = $class->_7mode_base_cmd_options( code => $command_code );
                }
        }
        elsif ($boot_mode eq C_ONTAP_CMODE)
        {
                if ($ontap_ver eq '8.1')
		{
			$allowed_options = $class->_Cmode_cmd_options_8_1( code => $command_code );
                }
		elsif ($ontap_ver eq '8.2')
		{
			$allowed_options = $class->_Cmode_cmd_options_8_2( code => $command_code );
                }
		else
		{
			logcomment("Commands: get_options_string: Specific support for sw_version $ontap_ver not found. Will be using base options") if ($class->{_debug});
			$allowed_options = $class->_Cmode_base_cmd_options( code => $command_code );
                }
        }
        else
        {
                logresult('script', "Unrecognized boot_mode: \"$boot_mode\"");
        }
        
        logcomment("Commands: get_options_string. Allowed options for command code=$command_code: " . Dumper($allowed_options)) if ($class->{_debug});
        
        foreach my $option_symbol (keys %$options_to_set)
        {
                my $option_value = $options_to_set->{$option_symbol};
                if (!defined($option_value))
                {
                        logcomment("Option $option_symbol had undefined value. Skipping... ") if ($class->{_debug});
                        next;
                }
                logcomment("Option $option_symbol required to be set to $option_value") if ($class->{_debug});
                
                if (defined($allowed_options->{$option_symbol}))
                {
                        my $allowed_option_values = $allowed_options->{$option_symbol};
                        if ($option_value == TRUE)
			{
				if (defined($allowed_option_values->{TRUE}))
				{
					logcomment("Option $option_symbol is allowed to be set to $option_value: " .
						   $allowed_option_values->{TRUE}) if ($class->{_debug});
					$return_str .= $allowed_option_values->{TRUE} . " ";
				}
                        }
			elsif ($option_value == FALSE)
			{
				if (defined($allowed_option_values->{FALSE}))
				{
					logcomment("Option $option_symbol is allowed to be set to $option_value: " .
						   $allowed_option_values->{FALSE}) if ($class->{_debug});
					$return_str .= $allowed_option_values->{FALSE} . " ";
                                }
                        }
			else
			{
				logcomment("Option $option_symbol is currently not allowed to be set to $option_value: ") if ($class->{_debug});
                        }
                }
                else
                {
                        logcomment("Commands: get_options_string:\n" .
                                   "Command code: $command_code\n" .
                                   "Attempt to set unsupported option for this release: $option_symbol\n" .
                                   "Ignoring...") if ($class->{_debug});
                }
        }
        
        # Remove the extra space at the end, if any
        $return_str =~ s/\s$//;
        
        logcomment("Commands: get_options_string: returning command options for code $command_code: $return_str") if ($class->{_debug});
        return $return_str;
}


## @name ::is_dblade_level_cmd
## @summary Decide if a certain command needs to be executed at d-blade level (C-mode node run).
##
## @description
## Decide if a certain command needs to be executed at d-blade level (C-mode node run).
##
## @return TRUE/FALSE.
##
## @arg code Required: Code corresponding to the operation for which the command needs to be executed.
## Operation codes are defined in Compression::Constants.
##
## @examples
## my $need_node_run = $obj->is_dblade_level_cmd(code => C_CMD_GET_SAVINGS,);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub is_dblade_level_cmd
{
        my $class = shift;
	my %args = validate(@_,
	{	
		code      => { type => SCALAR, },
        } );

        my $code = $args{code};
        my $ontap_ver = $class->{_version}->{ontap};
        my $default = FALSE;
        
        if ($class->{_boot_mode} ne C_ONTAP_CMODE)
        {
                return $default;
        }
        
        my $cmode_entries = {'8.1' => {GET_SPACE_SAVINGS => TRUE,
                                       GET_VBN_INFO => TRUE,
                                       DUMPBLOCK_FILE => TRUE,
                                       FILE_RESERVATION => TRUE,
                                       GET_FILE_ID => TRUE,
                                       VERIFY_COMPR_ATTEMPTED_BIT => TRUE,
				       VERIFY_SIS_LS => TRUE,
                                       SYNC => TRUE,
                                       GET_SNAP_RECLAIMABLE => TRUE,
                                       DISK_FREE => TRUE,
                                       SNAP_DELTA => TRUE,
                                       GET_RESERVED => TRUE,
                                       AGGR_SNAP_LIST => TRUE,
                                       AGGR_SNAP_DELETE => TRUE,
                                       DETECT_ROOT_VOL => TRUE,
                                       FILE_FOLDING_START => TRUE,
                                       FILE_FOLDING_STATUS => TRUE,
                                       FILE_FOLDING_ABORT => TRUE,
                                       DUMPBLOCK_INFO => TRUE,
                                       DUMPBLOCK_INFO_L1 => TRUE,
                                       GET_REFCOUNT => TRUE,
                                       SIS_CHECK => TRUE,
                                       GET_CONTAINMENT_VERSION => TRUE,
                                       GET_COMPRESSION_STATS => TRUE,
				       COMPRESSION_STATS_START => TRUE,
				       COMPRESSION_STATS_STOP => TRUE,
				       BUFTREE_CHECK => TRUE,
                                       SET_WAFL_SCAN_SPEED => TRUE,
                                       GET_WAFL_SCAN_SPEED => TRUE,
                                       GET_SNAPLOCK_CONTAINMENT_BITS => TRUE,
				       BG_FILE_CLONE_STATUS => TRUE,
				       GET_DATA_DISKS => TRUE,
				       WAFL_PUNCH => TRUE,
				       GET_VOLUME_STATS => TRUE,
				       GET_WAFL_STATS => TRUE,
                                       },
			     '8.2' => {GET_SPACE_SAVINGS => TRUE,
				       GET_VBN_INFO => TRUE,
				       DUMPBLOCK_FILE => TRUE,
				       FILE_RESERVATION => TRUE,
				       GET_FILE_ID => TRUE,
				       VERIFY_COMPR_ATTEMPTED_BIT => TRUE,
				       VERIFY_SIS_LS => TRUE,
				       SYNC => TRUE,
				       GET_SNAP_RECLAIMABLE => TRUE,
				       DISK_FREE => TRUE,
				       SNAP_DELTA => TRUE,
				       GET_RESERVED => TRUE,
				       AGGR_SNAP_LIST => TRUE,
				       AGGR_SNAP_DELETE => TRUE,
				       DETECT_ROOT_VOL => TRUE,
				       FILE_FOLDING_START => TRUE,
				       FILE_FOLDING_STATUS => TRUE,
				       FILE_FOLDING_ABORT => TRUE,
				       DUMPBLOCK_INFO => TRUE,
				       DUMPBLOCK_INFO_L1 => TRUE,
				       GET_REFCOUNT => TRUE,
				       SIS_CHECK => TRUE,
				       GET_CONTAINMENT_VERSION => TRUE,
				       GET_COMPRESSION_STATS => TRUE,
				       COMPRESSION_STATS_START => TRUE,
				       COMPRESSION_STATS_STOP => TRUE,
				       BUFTREE_CHECK => TRUE,
				       SET_WAFL_SCAN_SPEED => TRUE,
				       GET_WAFL_SCAN_SPEED => TRUE,
				       GET_SNAPLOCK_CONTAINMENT_BITS => TRUE,
				       BG_FILE_CLONE_STATUS => TRUE,
				       GET_DATA_DISKS => TRUE,
				       WAFL_PUNCH => TRUE,
				       GET_VOLUME_STATS => TRUE,
				       GET_WAFL_STATS => TRUE,
				       },
					 '8.3' => {GET_SPACE_SAVINGS => TRUE,
				       GET_VBN_INFO => TRUE,
				       DUMPBLOCK_FILE => TRUE,
				       FILE_RESERVATION => TRUE,
				       GET_FILE_ID => TRUE,
				       VERIFY_COMPR_ATTEMPTED_BIT => TRUE,
				       VERIFY_SIS_LS => TRUE,
				       SYNC => TRUE,
				       GET_SNAP_RECLAIMABLE => TRUE,
				       DISK_FREE => TRUE,
				       SNAP_DELTA => TRUE,
				       GET_RESERVED => TRUE,
				       AGGR_SNAP_LIST => TRUE,
				       AGGR_SNAP_DELETE => TRUE,
				       DETECT_ROOT_VOL => TRUE,
				       FILE_FOLDING_START => TRUE,
				       FILE_FOLDING_STATUS => TRUE,
				       FILE_FOLDING_ABORT => TRUE,
				       DUMPBLOCK_INFO => TRUE,
				       DUMPBLOCK_INFO_L1 => TRUE,
				       GET_REFCOUNT => TRUE,
				       SIS_CHECK => TRUE,
				       GET_CONTAINMENT_VERSION => TRUE,
				       GET_COMPRESSION_STATS => TRUE,
				       COMPRESSION_STATS_START => TRUE,
				       COMPRESSION_STATS_STOP => TRUE,
				       BUFTREE_CHECK => TRUE,
				       SET_WAFL_SCAN_SPEED => TRUE,
				       GET_WAFL_SCAN_SPEED => TRUE,
				       GET_SNAPLOCK_CONTAINMENT_BITS => TRUE,
				       BG_FILE_CLONE_STATUS => TRUE,
				       GET_DATA_DISKS => TRUE,
				       WAFL_PUNCH => TRUE,
				       GET_VOLUME_STATS => TRUE,
				       GET_WAFL_STATS => TRUE,
				       },
                             };
        
        my $specific_entry = $cmode_entries->{$ontap_ver}->{$code};
        
        if (defined($specific_entry))
        {
                return $specific_entry;
        }
        else
        {
                return $default;
        }
}


## @name ::is_vfiler_available_cmd
## @summary Decide if a certain command is available in the vfiler context.
##
## @description
## Decide if a certain command is available in the vfiler context.
##
## @return TRUE/FALSE.
##
## @arg code Required: Code corresponding to the command.
## Operation codes are defined in Compression::Constants.
##
## @examples
## my $need_node_run = $obj->is_vfiler_available_cmd(code => C_CMD_GET_SAVINGS,);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub is_vfiler_available_cmd
{
        my $class = shift;
	my %args = validate(@_,
	{	
		code      => { type => SCALAR, },
        } );

        my $code = $args{code};
        my $ontap_ver = $class->{_version}->{ontap};
        my $absolute_default = FALSE;
        
        my $vfiler_avail = {'default' => {ENABLE_COMPRESSION => TRUE,
                                          IS_COMPRESSION_ENABLED => TRUE,
                                          DISABLE_COMPRESSION => TRUE,
                                          GET_SPACE_SAVINGS => TRUE,
                                          GET_STATS => FALSE, 
                                          VERIFY_COMPR_ATTEMPTED_BIT => FALSE,
                                          DECOMPRESS_START => TRUE,
                                          DECOMPRESS_STATUS => TRUE,
                                          DECOMPRESS_STOP => TRUE,
                                          COMPRESS_START => TRUE,
                                          COMPRESS_STATUS => TRUE,
                                          COMPRESS_STOP => TRUE,
                                          },
                            '8.1' => {ENABLE_COMPRESSION => TRUE,
                                      },
                            };
        
        if (defined($vfiler_avail->{$ontap_ver}->{$code}))
        {
                return $vfiler_avail->{$ontap_ver}->{$code};
        }
        elsif (defined($vfiler_avail->{default}->{$code}))
        {
                return $vfiler_avail->{default}->{$code};
        }
        else
        {
                return $absolute_default;
        }
}


########################################
# Internal soubroutines: command strings
########################################
sub _7mode_base_commands
{
	my $class = shift;
	my %args = validate(@_,
	{	
		code      => { type => SCALAR, },
        } );

        my $code = $args{code};
	logcomment("Commands: _7mode_base_commands input: ". Dumper(%args) ) if( $class->{_debug} );

	my %base_cmds = (
		GET_SPACE_SAVINGS          => "df -S[modifier] [volume]",
		DECOMPRESS_START           => "vol decompress start <volume>",
		DECOMPRESS_STATUS          => "vol decompress status [volume]",
		DECOMPRESS_STOP            => "vol decompress stop <volume>",
		AUTOSIZE                   => "vol autosize <volume> -m <size> -i <increment> <state>",
                IS_COMPRESSION_ENABLED     => "vol options <volume>",
                ENABLE_COMPRESSION         => "vol options <volume> compression on",
                DISABLE_COMPRESSION        => "vol options <volume> compression off",
                GET_VBN_INFO               => "priv set -q diag; wafl showvbns <file_path> [fbn]",
                DUMPBLOCK_FILE             => "priv set -q diag; dumpblock file <volume> <file_path> <fbn>",
                FILE_RESERVATION           => "file reservation <file_path> [options]",
                GET_FILE_ID                => "priv set -q diag; showfh <file_path>",
                FILE_FOLDING_START         => "priv set -q diag; wafl scan fold /vol/<volume>/<file_path>",
                FILE_FOLDING_STATUS        => "priv set -q diag; wafl scan status -V <volume>",
                FILE_FOLDING_ABORT         => "priv set -q diag; wafl scan abort <modifier>",
                GET_CONTAINMENT_VERSION    => "priv set -q diag; stats show volume:<volume>",
                GET_COMPRESSION_STATS      => "priv set -q diag; stats show compression:<volume>",
		COMPRESSION_STATS_START    => "priv set -q diag; stats start compression",
		COMPRESSION_STATS_STOP     => "priv set -q diag; stats stop",
                SET_WAFL_SCAN_SPEED        => "priv set -q diag; wafl scan speed <modifier>",
                GET_WAFL_SCAN_SPEED        => "priv set -q diag; wafl scan speed",
                GET_SNAPLOCK_CONTAINMENT_BITS => "priv set -q diag; stats show volume:<volume>",
		BG_FILE_CLONE_STATUS       => "priv set -q test;cmf_test <volume> list",
		VOL_MOVE_STATUS            => "vol move status <volume>",
		VOL_MOVE_CUTOVER           => "vol move cutover <volume>",
		VOL_MOVE_PAUSE             => "vol move pause <volume>",
		VOL_MOVE_IS_PAUSED         => "vol move status <volume>",
		VFILER_MIGRATE_STATUS      => "vfiler migrate status <vfiler>\@<filer>",
		VOL_ONLINE_RO              => "priv set -q test; vol online <volume> -ro",
		WAFL_PUNCH                 => "priv set -q diag; wafl punch <file_path> <fbn> <increment>",
		GET_VOLUME_STATS           => "priv set -q diag; stats show volume:<volume>",
		GET_WAFL_STATS             => "priv set -q diag; wafl stats",
		CONFIGURE_IDD              => "sis config -Q <modifier> /vol/<volume>",
		IS_IDD_ENABLED             => "sis status -l /vol/<volume>",
	);

	my $cmd = $base_cmds{$code};
        
        if (!defined($cmd))
        {
                logresult("script", "This library only maintains compression-specific commands. Command code \"$code\" is not supported.");
        }
	logcomment("Commands: _7mode_base_commands. Returning command for code=$code: \"$cmd\"") if ($class->{_debug});

	return $cmd;
}

sub _7mode_commands_7_3
{
	my $class = shift;
	my %args = validate(@_,
	{	
		code      => { type => SCALAR, },
	} );
        
        logcomment("Commands: _7mode_commands_7_3 input: ". Dumper(%args)) if ($class->{_debug});
        
	my %cmds_7_3 = ();

        my $code = $args{code};
        
	return $cmds_7_3{$code};
}

sub _7mode_commands_8_0
{
	my $class = shift;
	my %args = validate(@_,
	{	
		code      => { type => SCALAR, },
	} );

	logcomment("Commands: _7mode_commands_8_0 input: ". Dumper(%args)) if ($class->{_debug});
        
	my %cmds_BR_0 = (
                COMPRESS_START             => "vol compress start [options] <volume>",
                COMPRESS_STATUS_CHECKPOINT => "priv set -q test; vol compress status -l <volume>",
                COMPRESS_STATUS            => "vol compress status [volume]",
                COMPRESS_STOP              => "vol compress stop [options] <volume>",
                DECOMPRESS_REVERT          => "vol decompress revert <volume>"
                );
        
        my $code = $args{code};
        
        return $cmds_BR_0{$code};
}

sub _7mode_commands_8_1
{
	my $class = shift;
	my %args = validate(@_,
	{	
		code      => { type => SCALAR, },
	} );

	logcomment("Commands: _7mode_commands_8_1 input: ". Dumper(%args)) if ($class->{_debug});
        
	my %cmds_8_1 = (
                IS_COMPRESSION_ENABLED     => "sis status -l /vol/<volume>",
                ENABLE_COMPRESSION         => "sis config [options] /vol/<volume>",
                DISABLE_COMPRESSION        => "sis config [options] -I false /vol/<volume>",
                COMPRESS_START             => "priv set -q advanced; sis start [options] /vol/<volume>",
                COMPRESS_STATUS            => "priv set -q diag; sis status -l /vol/[volume]",
                COMPRESS_STOP              => "sis stop /vol/<volume>",
                COMPRESS_STATUS_CHECKPOINT => "priv set -q diag; sis status -l /vol/<volume>",
                DECOMPRESS_START           => "priv set -q diag; sis undo /vol/<volume> [options]",
                DECOMPRESS_STATUS          => "priv set -q diag; sis status -l /vol/[volume]",
                DECOMPRESS_STOP            => "sis stop /vol/<volume>",
                GET_STATS                  => "priv set -q diag; sis stat -lv /vol/<volume>",
                VERIFY_COMPR_ATTEMPTED_BIT => "priv set -q diag; sis check -cv /vol/<volume>",
                );
        
        my $code = $args{code};
        
        return $cmds_8_1{$code};
}

sub _7mode_commands_8_2
{
	my $class = shift;
	my %args = validate(@_,
	{	
		code      => { type => SCALAR, },
	} );

	logcomment("Commands: _7mode_commands_8_2 input: ". Dumper(%args)) if ($class->{_debug});
        
	my %cmds = (
                IS_COMPRESSION_ENABLED     => "sis status -l /vol/<volume>",
                ENABLE_COMPRESSION         => "sis config [options] /vol/<volume>",
                DISABLE_COMPRESSION        => "sis config [options] -I false /vol/<volume>",
                COMPRESS_START             => "priv set -q advanced; sis start [options] /vol/<volume>",
                COMPRESS_STATUS            => "priv set -q diag; sis status -l /vol/[volume]",
                COMPRESS_STOP              => "sis stop /vol/<volume>",
                COMPRESS_STATUS_CHECKPOINT => "priv set -q diag; sis status -l /vol/<volume>",
                DECOMPRESS_START           => "priv set -q diag; sis undo /vol/<volume> [options]",
                DECOMPRESS_STATUS          => "priv set -q diag; sis status -l /vol/[volume]",
                DECOMPRESS_STOP            => "sis stop /vol/<volume>",
                GET_STATS                  => "priv set -q diag; sis stat -lv /vol/<volume>",
                VERIFY_COMPR_ATTEMPTED_BIT => "priv set -q diag; sis check -cv /vol/<volume>",
                );
        
        my $code = $args{code};
        
        return $cmds{$code};
}

sub _Cmode_base_commands
{
	my $class = shift;
	my %args = validate(@_,
	{	
		code      => { type => SCALAR, },
        } );

        my $code = $args{code};
	logcomment("Commands: _Cmode_base_commands input: ". Dumper(%args) ) if( $class->{_debug} );

	my %base_cmds = (
		GET_SPACE_SAVINGS          => "df -S[modifier] [volume]",
                IS_COMPRESSION_ENABLED     => "volume efficiency show -l -vserver <vserver> -volume <volume>",
                ENABLE_COMPRESSION         => "volume efficiency config -vserver <vserver> -volume <volume> [options]",
                DISABLE_COMPRESSION        => "volume efficiency config -vserver <vserver> -volume <volume> [options] -inline false",
                COMPRESS_START             => "volume efficiency start -vserver <vserver> -volume <volume> [options]",
                COMPRESS_STATUS            => "volume efficiency show -l -vserver <vserver> -volume <volume>",
                COMPRESS_STOP              => "volume efficiency stop -vserver <vserver> -volume <volume>",
                COMPRESS_STATUS_CHECKPOINT => "volume efficiency show -l -vserver <vserver> -volume <volume>",
                DECOMPRESS_START           => "volume efficiency undo -vserver <vserver> -volume <volume> [options]",
                DECOMPRESS_STATUS          => "volume efficiency show -l -vserver <vserver> -volume <volume>",
                DECOMPRESS_STOP            => "volume efficiency stop -vserver <vserver> -volume <volume>",
                GET_STATS                  => "volume efficiency stat -l -vserver <vserver> -volume <volume>",
                VERIFY_COMPR_ATTEMPTED_BIT => "priv set -q diag; sis check -cv /vol/<volume>",
		VERIFY_SIS_LS 		   => "priv set -q diag; sis ls /vol/<volume>",
                GET_VBN_INFO               => "priv set -q diag; wafl showvbns <file_path> [fbn]",
                DUMPBLOCK_FILE             => "priv set -q diag; dumpblock file <volume> <file_path> <fbn>",
                FILE_RESERVATION           => "file reservation <file_path> [options]",
                GET_FILE_ID                => "priv set -q diag; showfh <file_path>",
                SYNC                       => "priv set -q diag; sync; sync",
                SFSR_START                 => "volume snapshot restore-file -vserver <vserver> -volume <volume> -snapshot <snapshot> -path /<file_path>",
                SFSR_TO_PATH_START         => "volume snapshot restore-file -vserver <vserver> -volume <volume> -snapshot <snapshot> -path /<file_path> -restore-path /<restore_path>",
                GET_SNAP_RECLAIMABLE       => "snap reclaimable <volume> <snapshot>",
                DISK_FREE                  => "df -k <volume>",
                SNAP_DELTA                 => "snap delta <volume> <snapshot> <state>",
                GET_RESERVED               => "df -r -k <volume>",
                AGGR_SNAP_LIST             => "snap list -A <volume>",
                AGGR_SNAP_DELETE           => "snap delete -A <volume> <snapshot>",
                DETECT_ROOT_VOL            => "vol status <volume>",
                FILE_FOLDING_START         => "priv set -q diag; wafl scan fold /vol/<volume>/<file_path>",
                FILE_FOLDING_STATUS        => "priv set -q diag; wafl scan status -V <volume>",
                FILE_FOLDING_ABORT         => "priv set -q diag; wafl scan abort <modifier>",
                DUMPBLOCK_INFO             => "priv set -q diag; dumpblock info <volume> <file_path> <fbn>",
                DUMPBLOCK_INFO_L1          => "priv set -q diag; dumpblock info <volume> <file_path> <fbn> 1",
                GET_REFCOUNT               => "priv set -q test;wafl showvbns -s <file_path>",
                SIS_CHECK                  => "priv set -q diag; sis check [options] /vol/<volume>",
                GET_CONTAINMENT_VERSION    => "priv set -q diag; stats show volume:<volume>",
                GET_COMPRESSION_STATS      => "priv set -q diag; stats show compression:<volume>",
		COMPRESSION_STATS_START    => "priv set -q diag; stats start compression",
		COMPRESSION_STATS_STOP     => "priv set -q diag; stats stop",
		BUFTREE_CHECK              => "priv set -q test; wafl buftree check <volume> <file_path> [options]",
                SET_WAFL_SCAN_SPEED        => "priv set -q diag; wafl scan speed <modifier>",
                GET_WAFL_SCAN_SPEED        => "priv set -q diag; wafl scan speed",
                GET_SNAPLOCK_CONTAINMENT_BITS => "priv set -q diag; stats show volume:<volume>",
		BG_FILE_CLONE_STATUS       => "priv set -q test;cmf_test <volume> list",
		GET_DATA_DISKS             => "aggr status -r <aggregate>",
		WAFL_PUNCH                 => "priv set -q diag; wafl punch <file_path> <fbn> <increment>",
		GET_VOLUME_STATS           => "priv set -q diag; stats show volume:<volume>",
		GET_WAFL_STATS             => "priv set -q diag; wafl stats",
		CONFIGURE_IDD              => "volume efficiency config -idd <modifier> -vserver <vserver> -volume <volume>",
		IS_IDD_ENABLED             => "volume efficiency show -l -vserver <vserver> -volume <volume>",
	);

	my $cmd = $base_cmds{$code};
        
        if (!defined($cmd))
        {
                logresult("script", "This library only maintains compression-specific commands. Command code \"$code\" is not supported in C-mode.");
        }
	logcomment("Commands: _Cmode_base_commands. Returning command for code=$code: \"$cmd\"") if ($class->{_debug});

	return $cmd;
}

sub _Cmode_commands_8_1
{
	my $class = shift;
	my %args = validate(@_,
	{	
		code      => { type => SCALAR, },
	} );
        
        logcomment("Commands: _Cmode_commands_8_1 input: ". Dumper(%args)) if ($class->{_debug});
        
	my %cmd_strings = (
		DISABLE_COMPRESSION        => "volume efficiency config -vserver <vserver> -volume <volume> [options] -I false",
	);

        my $code = $args{code};
        
	return $cmd_strings{$code};
}



########################################
# Internal soubroutines: command options
########################################

# Note: the hash keys in the "command option" methods
# have to correspond to the arguments that represent command options
# in the highest-level APIs.
# For example, the high-level method corresponding to COMPRESS_STOP (CompressionScanner::stop),
# has the option argument 'no_checkpoint', thus, the options hash key is still 'no_checkpoint'.

sub _7mode_base_cmd_options
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => { type => SCALAR, },
        } );
        
        my $code = $args{code};
        logcomment("Commands: _7mode_base_cmd_options input: ". Dumper(%args) ) if ($class->{_debug});
        
        my $base_cmd_options = {
                FILE_RESERVATION => {enabled => {TRUE => 'enable',
                                                 FALSE => 'disable'},},
                };
        
        return $base_cmd_options->{$code};
}


sub _7mode_cmd_options_7_3
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => { type => SCALAR, },
        } );
        
        logcomment("Commands: _7mode_cmd_options_7_3 input: ". Dumper(%args)) if ($class->{_debug});
        
        my $cmd_options_7_3 = {
                FILE_RESERVATION => {enabled => {TRUE => 'enable',
                                                 FALSE => 'disable'},},
                };
        
        my $code = $args{code};
        return $cmd_options_7_3->{$code};
}


sub _7mode_cmd_options_8_0
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => { type => SCALAR, },
        } );
        
        logcomment("Commands: _7mode_cmd_options_8_0 input: ". Dumper(%args)) if ($class->{_debug});
        
        my $cmd_options_BR_0 = {
                COMPRESS_START => {deduped_blocks => {TRUE => '-a',},
                                   snapshotted_blocks => {TRUE => '-s',},
                                   delete_checkpoint => {TRUE => '-d',}},
                COMPRESS_STOP => {no_checkpoint => {TRUE => '-d',},},
                FILE_RESERVATION => {enabled => {TRUE => 'enable',
                                                 FALSE => 'disable'},},
                };
        
        my $code = $args{code};
        return $cmd_options_BR_0->{$code};
}


sub _7mode_cmd_options_8_1
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => { type => SCALAR, },
        } );
        
        logcomment("Commands: _7mode_cmd_options_8_1 input: ". Dumper(%args)) if ($class->{_debug});
        
        my $cmd_options_8_1 = {
                ENABLE_COMPRESSION => {inline => {TRUE => '-I true',
                                                  FALSE => '-I false'},
                                       background => {TRUE => '-C true',
                                                      FALSE => '-C false'},},
                DISABLE_COMPRESSION => {disable_inline_only => {TRUE => '-C true',
                                                                FALSE => '-C false'},},
                COMPRESS_START => {scan => {TRUE => '-s',},
                                   force => => {TRUE => '-f',},
                                   scan_to_compress => => {TRUE => '-C',},
                                   scan_to_dedupe => => {TRUE => '-D',},
                                   deduped_blocks => {TRUE => '-a',},
                                   snapshotted_blocks => {TRUE => '-b',},
                                   delete_checkpoint => {TRUE => '-d',}},
                DECOMPRESS_START => {target_decompress => {TRUE => '-C',},
                                     target_dedupe => {TRUE => '-D',},},
                FILE_RESERVATION => {enabled => {TRUE => 'enable',
                                                 FALSE => 'disable'},},
                };
        
        my $code = $args{code};
        return $cmd_options_8_1->{$code};
}


sub _7mode_cmd_options_8_2
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => { type => SCALAR, },
        } );
        
        logcomment("Commands: _7mode_cmd_options_8_2 input: ". Dumper(%args)) if ($class->{_debug});
        
        my $cmd_options = {
                ENABLE_COMPRESSION => {inline => {TRUE => '-I true',
                                                  FALSE => '-I false'},
                                       background => {TRUE => '-C true',
                                                      FALSE => '-C false'},},
                DISABLE_COMPRESSION => {disable_inline_only => {TRUE => '-C true',
                                                                FALSE => '-C false'},},
                COMPRESS_START => {scan => {TRUE => '-s',},
                                   force => => {TRUE => '-f',},
                                   scan_to_compress => => {TRUE => '-C',},
                                   scan_to_dedupe => => {TRUE => '-D',},
                                   deduped_blocks => {TRUE => '-a',},
                                   snapshotted_blocks => {TRUE => '-b',},
                                   delete_checkpoint => {TRUE => '-d',}},
                DECOMPRESS_START => {target_decompress => {TRUE => '-C',},
                                     target_dedupe => {TRUE => '-D',},},
                FILE_RESERVATION => {enabled => {TRUE => 'enable',
                                                 FALSE => 'disable'},},
                };
        
        my $code = $args{code};
        return $cmd_options->{$code};
}


sub _Cmode_base_cmd_options
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => { type => SCALAR, },
        } );
        
        my $code = $args{code};
        logcomment("Commands: _Cmode_base_cmd_options input: ". Dumper(%args) ) if ($class->{_debug});
        
        my $base_cmd_options = {
                ENABLE_COMPRESSION => {inline => {TRUE => '-inline true',
                                                  FALSE => '-inline false'},
                                       background => {TRUE => '-compression true',
                                                      FALSE => '-compression false'},},
                DISABLE_COMPRESSION => {disable_inline_only => {TRUE => '-compression true',
                                                                FALSE => '-compression false'},},
                COMPRESS_START => {scan => {TRUE => '-s true',},
                                   force => => {TRUE => '-f true',},
                                   scan_to_compress => => {TRUE => '-C true',
                                                           FALSE => '-C false'},
                                   scan_to_dedupe => => {TRUE => '-D true',
                                                         FALSE => '-D false'},
                                   deduped_blocks => {TRUE => '-a true',
                                                      FALSE => '-a false'},
                                   snapshotted_blocks => {TRUE => '-b true',
                                                          FALSE => '-b false'},
                                   delete_checkpoint => {TRUE => '-d true',}},
                DECOMPRESS_START => {target_decompress => {TRUE => '-C true',
                                                           FALSE => '-C false'},
                                     target_dedupe => {TRUE => '-D true',
                                                       FALSE => '-D false'},},
                FILE_RESERVATION => {enabled => {TRUE => 'enable',
                                                 FALSE => 'disable'},},
                SIS_CHECK => {count => {TRUE => '-c'},
                              values => {TRUE => '-v',},
                              delete_stale => {TRUE => '-d',},},
		BUFTREE_CHECK => {filename => {TRUE => '-n'}},
                };
        
        return $base_cmd_options->{$code};
}

sub _Cmode_cmd_options_8_1
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => { type => SCALAR, },
        } );
        
        logcomment("Commands: _Cmode_cmd_options_8_1 input: ". Dumper(%args)) if ($class->{_debug});
        
        my $cmd_options = {
                ENABLE_COMPRESSION => {inline => {TRUE => '-inline true',
                                                  FALSE => '-inline false'},
                                       background => {TRUE => '-compression true',
                                                      FALSE => '-compression false'},},
                DISABLE_COMPRESSION => {disable_inline_only => {TRUE => '-compression true',
                                                                FALSE => '-compression false'},},
                COMPRESS_START => {scan => {TRUE => '-s true',},
                                   force => => {TRUE => '-f true',},
                                   scan_to_compress => => {TRUE => '-C true',
                                                           FALSE => '-C false'},
                                   scan_to_dedupe => => {TRUE => '-D true',
                                                         FALSE => '-D false'},
                                   deduped_blocks => {TRUE => '-a true',
                                                      FALSE => '-a false'},
                                   snapshotted_blocks => {TRUE => '-b true',
                                                          FALSE => '-b false'},
                                   delete_checkpoint => {TRUE => '-d true',}},
                DECOMPRESS_START => {target_decompress => {TRUE => '-C true',
                                                           FALSE => '-C false'},
                                     target_dedupe => {TRUE => '-D true',
                                                       FALSE => '-D false'},},
                FILE_RESERVATION => {enabled => {TRUE => 'enable',
                                                 FALSE => 'disable'},},
                SIS_CHECK => {count => {TRUE => '-c'},
                              values => {TRUE => '-v',},
                              delete_stale => {TRUE => '-d',},},
		BUFTREE_CHECK => {filename => {TRUE => '-n'}},
                };
        
        my $code = $args{code};
        return $cmd_options->{$code};
}

sub _Cmode_cmd_options_8_2
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => { type => SCALAR, },
        } );
        
        logcomment("Commands: _Cmode_cmd_options_8_2 input: ". Dumper(%args)) if ($class->{_debug});
        
        my $cmd_options = {
                ENABLE_COMPRESSION => {inline => {TRUE => '-inline true',
                                                  FALSE => '-inline false'},
                                       background => {TRUE => '-compression true',
                                                      FALSE => '-compression false'},},
                DISABLE_COMPRESSION => {disable_inline_only => {TRUE => '-compression true',
                                                                FALSE => '-compression false'},},
                COMPRESS_START => {scan => {TRUE => '-s true',},
                                   force => => {TRUE => '-f true',},
                                   scan_to_compress => => {TRUE => '-C true',
                                                           FALSE => '-C false'},
                                   scan_to_dedupe => => {TRUE => '-D true',
                                                         FALSE => '-D false'},
                                   deduped_blocks => {TRUE => '-a true',
                                                      FALSE => '-a false'},
                                   snapshotted_blocks => {TRUE => '-b true',
                                                          FALSE => '-b false'},
                                   delete_checkpoint => {TRUE => '-d true',}},
                DECOMPRESS_START => {target_decompress => {TRUE => '-C true',
                                                           FALSE => '-C false'},
                                     target_dedupe => {TRUE => '-D true',
                                                       FALSE => '-D false'},},
                FILE_RESERVATION => {enabled => {TRUE => 'enable',
                                                 FALSE => 'disable'},},
                SIS_CHECK => {count => {TRUE => '-c'},
                              values => {TRUE => '-v',},
                              delete_stale => {TRUE => '-d',},},
		BUFTREE_CHECK => {filename => {TRUE => '-n'}},
                };
        
        my $code = $args{code};
        return $cmd_options->{$code};
}


1;
