#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/CLI/CorruptionMessages.pm#1 $
#

## @summary CLI template for filesystem corruption messages.
##
## @description
##    This library keeps track corruption messages needed for error detection using wafliron, wack.
##
## @synopsis
##
## use Compression::UI::CLI::CorruptionMessages;
## use Compression::CLI;
## use Compression::Version;
## use NOTE::Filer;  # To obtain filer version.
##
## my $filer = NOTE::Filer->connect( host => $FILER );
## my $version = Compression::Version->new( conn => $filer );
## my $obj = Compression::UI::CLI::CorruptionMessages->new( version => $version );
##
## @examples 
## $obj->get_wafliron_messages(code => C_CORRUPT_FSINFO_INOS_USED);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)



package Compression::UI::CLI::CorruptionMessages;


use strict;
use warnings;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Tharn;
use Compression::Constants qw(:all);
use Compression::Version;


BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
        
};


## @name ::new
##
## @summary Compression::UI::CLI::CorruptionMessages Constructor.
##
## @return Object of type Compression::UI::CLI::CorruptionMessages.
##
## @arg version Required: Compression::Version object for the hosting filer.
## 
## @arg boot_mode Optional: Boot mode of the hosting filer.
## Defaults to 7M.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::UI::CLI::CorruptionMessages->new( version => $version );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##

sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug  => {  
			type 	 => SCALAR,
			optional => 1,
                        default  => FALSE,
			},
	} );
        
	my $self = {};
        
        $self->{_debug} = $args{debug};
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};
        
        logcomment("Created new CorruptionMessages object:\n" . Dumper($self)) if ($self->{_debug});
	
	bless $self, $class;
	return $self;
}


## @name ::get_wafliron_corruption_message
## @summary Generates corruption messages to be expected in the wafliron output.
## 
## @description
## Generates corruption messages to be expected in the wafliron output if a particular error was detected/fixed.
##
## @return
## Reference to array of corruption messages.
## 
## @arg code Required: Code corresponding to the corruption (defined in Constants.pm).
## 
## @examples
## $obj->get_wafliron_corruption_message(code => C_CORRUPT_FSINFO_INOS_USED);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##

sub get_wafliron_corruption_message
{
        my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $code = $args{code};
        my $ret_hash = {};
        
        if ($ontap_ver eq '7.3')
	{
		$ret_hash = $class->_wafliron_messages_7_3( code => $code );
	}
	elsif ($ontap_ver eq '8.0')
	{
		$ret_hash = $class->_wafliron_messages_8_0( code => $code );
	}
	elsif ($ontap_ver eq '8.1')
	{
		$ret_hash = $class->_wafliron_messages_8_1( code => $code );
	}
	elsif ($ontap_ver eq '8.2')
	{
		$ret_hash = $class->_wafliron_messages_8_2( code => $code );
	}
	else
	{
		logcomment("CorruptionMessages: Specific support for sw_version $ontap_ver not found. Will be using base wafliron messages");
	}
        
        if ( !defined($ret_hash) || !(scalar @$ret_hash) )
        {
                logcomment("CorruptionMessages: Could not find wafliron corruption messages definition specific for $ontap_ver. " .
                           "Defaulting to base wafliron messages. " ) if ($class->{_debug});
       		$ret_hash =  $class->_base_wafliron_messages( code => $code );
        }

        return $ret_hash;
}


## @name ::get_wack_corruption_message
## @summary Generates corruption messages to be expected in the wack output.
## 
## @description
## Generates corruption messages to be expected in the wack output if a particular error was detected/fixed.
##
## @return
## Reference to array of corruption messages.
## 
## @arg code Required: Code corresponding to the corruption (defined in Constants.pm).
## 
## @examples
## $obj->get_wack_corruption_message(code => C_CORRUPT_FSINFO_INOS_USED);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##

sub get_wack_corruption_message
{
        my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $code = $args{code};
        my $ret_hash = {};
        
        if ($ontap_ver eq '7.3')
	{
		$ret_hash = $class->_wack_messages_7_3( code => $code );
	}
	elsif ($ontap_ver eq '8.0')
	{
		$ret_hash = $class->_wack_messages_BR_0( code => $code );
	}
	else
	{
		logcomment("CorruptionMessages: Specific support for sw_version $ontap_ver not found. Will be using base wack messages");
	}
        
        if ( !defined($ret_hash) || !(scalar @$ret_hash) )
        {
                logcomment("CorruptionMessages: Could not find wack corruption messages definition specific for $ontap_ver. " .
                           "Defaulting to base wack messages. " ) if ($class->{_debug});
       		$ret_hash =  $class->_base_wack_messages( code => $code );
        }

        return $ret_hash;
}


###################################
# Internal subroutines
###################################

######## Wafliron ##########
sub _base_wafliron_messages
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );

	logcomment("CorruptionMessages: _base_wafliron_messages: ". Dumper(%args) ) if ($class->{_debug});

	my $base_wafliron_messages_hash = {
                fsinfo_inos_used => ['Total inodes used incorrect'],
		fsinfo_blks_used_by_plane0 => ['plane 0: shows blocks used'],
		fsinfo_blks_compress_holes => ['FS info shows total compression saved blocks to be \d+. Setting to \d+.'],
		# Commenting out the old msgs
                #inode_blk_cnt => ['Inode \d+: \w+ block count from \d+ to \d+ to match size \d+',
                #                  'Inode \d+, blocks was \d+, should be \d+'],
		# 08/06/2012 : Wafliron msg coming for 'inode_blk_cnt' coruption
		# 		Inode 96, block count 0 is too small.  Adjusting up...
		# 		Inode 96, block cnt is 515 less than actual value. Correcting it to 515.
		inode_blk_cnt => ['Inode \d+, block count \d+'],
                corrupt_checksum => ['Inode \d+, blocks was \d+, should be \d+',
                                     'Inode \d+, Compression saved blocks was \d+, should be \d+',
                                     'Invalid compress entries in indirect blocks'],
                corrupt_cg_header_vbn => ['Clearing invalid CG. Inode \d+, indirect level 1, block \d+, index \d+: ',],
                corrupt_cg_header_vvbn => ['Inode \d+, indirect level 1, block \d+, index \d+: Clearing .+ vbn',],
                corrupt_cg_trailer_vbn => ['Clearing invalid CG. Inode \d+, indirect level 1, block \d+, index \d+: ',],
                corrupt_cg_trailer_vvbn => ['Clearing invalid CG. Inode \d+, indirect level 1, block \d+, index \d+: ',],
                corrupt_cg_data_vbn => ['Inode \d+, indirect level 1, block \d+, index \d+: Clearing compressed vbn',],
                corrupt_cg_data_vvbn => ['Inode \d+, indirect level 1, block \d+, index \d+: Clearing compressed vbn',],
                refcount_set => ['Refcount for block \d+ is incorrect: \d+. Correcting to \d+'],
		corrupt_rep => ['Block had an unrecoverable error'],
	};

	my $code = $args{code};
	my $wafliron_messages = $base_wafliron_messages_hash->{$code};
        
        if (!defined($wafliron_messages))
        {
                logresult("script", "No wafliron corruption message for code \"$code\" was found.");
        }
	logcomment("_base_wafliron_messages: Corruption code=$code, wafliron messages:\n" . Dumper($wafliron_messages)) if ($class->{_debug});

	return $wafliron_messages;
}

sub _wafliron_messages_7_3
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        logcomment("CorruptionMessages: _wafliron_messages_7_3: ". Dumper(%args) ) if ($class->{_debug});
        
	my $wafliron_messages_hash_7_3 = {};
	my $code = $args{code};
        
        return $wafliron_messages_hash_7_3->{$code}
}

sub _wafliron_messages_8_0
{
        my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        logcomment("CorruptionMessages: _wafliron_messages_8_0: ". Dumper(%args) ) if ($class->{_debug});
        
	my $wafliron_messages_hash_8_0 = {};
	my $code = $args{code};
        
        return $wafliron_messages_hash_8_0->{$code}
}

sub _wafliron_messages_8_1
{
        my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        logcomment("CorruptionMessages: _wafliron_messages_8_1: ". Dumper(%args) ) if ($class->{_debug});
        
	my $wafliron_messages_hash = {
		fsinfo_blks_used_by_plane0 => ['Adjusting blocks used by plane 0'],
		};
	my $code = $args{code};
        
        return $wafliron_messages_hash->{$code}
}

sub _wafliron_messages_8_2
{
        my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        logcomment("CorruptionMessages: _wafliron_messages_8_2: ". Dumper(%args) ) if ($class->{_debug});
        
	my $wafliron_messages_hash = {
		fsinfo_blks_used_by_plane0 => ['Adjusting blocks used by plane 0'],
		};
	my $code = $args{code};
        
        return $wafliron_messages_hash->{$code}
}


######### Wack ###########
sub _base_wack_messages
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );

	logcomment("CorruptionMessages: _base_wack_messages: ". Dumper(%args) ) if ($class->{_debug});

	my $base_wack_messages_hash = {
                fsinfo_inos_used => ['Total inodes corrected'],
		fsinfo_blks_used_by_plane0 => ['blocks used for \d+ planes corrected'],
		fsinfo_blks_compress_holes => ['FS info shows total compression saved blocks to be \d+. Setting to \d+'],
                inode_blk_cnt => ['Block counts corrected: \d+'],
                corrupt_checksum => ['Clearing compressed vbn in a file \d+',
                                     'Compression saved count is \d+.  Setting to \d+'],
                corrupt_cg_header_vbn => ['Inode \d+, indirect level 1, block \d+, index \d+: Clearing compressed vbn',
                                          'Invalid compress entries in indirect blocks: \d+'],
                corrupt_cg_header_vvbn => ['Inode \d+, indirect level 1, block \d+, index \d+: Clearing out of range vbn',
                                           'Out of range entries in indirect blocks: \d+'],
                corrupt_cg_trailer_vbn => ['Inode \d+, indirect level 1, block \d+, index \d+: Clearing compressed vbn',
                                           'Invalid compress entries in indirect blocks'],
                corrupt_cg_trailer_vvbn => ['Inode \d+, indirect level 1, block \d+, index \d+: Clearing compressed vbn',
                                            'Out of range entries in indirect blocks: \d+'],
                corrupt_cg_data_vbn => ['Inode \d+, indirect level 1, block \d+, index \d+: Clearing compressed vbn',],
                corrupt_cg_data_vvbn => ['Inode \d+, indirect level 1, block \d+, index \d+: Clearing compressed vbn',],
                refcount_set => ['Refcount for block \d+ is incorrect: \d+. Correcting to \d+'],
	};

	my $code = $args{code};
	my $wack_messages = $base_wack_messages_hash->{$code};
        
        if (!defined($wack_messages))
        {
                logresult("script", "No wack corruption message for code \"$code\" was found.");
        }
	logcomment("_base_wack_messages: Corruption code=$code, wack messages:\n" . Dumper($wack_messages)) if ($class->{_debug});

	return $wack_messages;
}



sub _wack_messages_7_3
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        logcomment("CorruptionMessages: _wack_messages_7_3: ". Dumper(%args) ) if ($class->{_debug});
        
	my $wack_messages_hash_7_3 = {};
	my $code = $args{code};
        
        return $wack_messages_hash_7_3->{$code}
}

sub _wack_messages_BR_0
{
        my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        logcomment("CorruptionMessages: _wack_messages_BR_0: ". Dumper(%args) ) if ($class->{_debug});
        
	my $wack_messages_hash_BR_0 = {};
	my $code = $args{code};
        
        return $wack_messages_hash_BR_0->{$code}
}


1;
