#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/CLI/FijiPoints.pm#1 $
#

## @summary CLI template for Fiji strings.
##
## @description
##    This library keeps track of Fiji strings needed for compression scripts (7G, 7M, 10M).
##
## @synopsis
##
## use Compression::UI::CLI::FijiPoints;
## use Compression::CLI;
## use Compression::Version;
## use NOTE::Filer;  # To obtain filer version.
##
## my $filer = NOTE::Filer->connect( host => $FILER );
## my $version = Compression::Version->new( conn => $filer );
## my $obj = Compression::UI::CLI::FijiPoints->new( version => $version );
##
## @examples 
## $obj->get_definition_hash(tag => C_FIJI_TAG_DISABLE_CP);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)



package Compression::UI::CLI::FijiPoints;


use strict;
use warnings;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Tharn;
use Compression::Constants qw(:all);
use Compression::Version;


BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
        
};


## @name ::new
## @summary Compression::UI::CLI::FijiPoints Constructor.
##
## @return Object of type Compression::UI::CLI::FijiPoints.
##
## @arg version Required: Compression::Version object for the hosting filer.
## 
## @arg boot_mode Optional: Boot mode of the hosting filer.
## Defaults to 7M.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::UI::CLI::FijiPoints->new( version => $version );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug  => {  
			type 	 => SCALAR,
			optional => 1,
                        default  => FALSE,
			},
	} );
        
	my $self = {};
        
        $self->{_debug} = $args{debug};
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};
	$self->{_tags} = [];
        
        logcomment("Created new FijiPoints object:\n" . Dumper($self)) if ($self->{_debug});
	
	bless $self, $class;
	
	$self->_register_fiji_points();
        
	return $self;
}


## @name ::get_definition_hash
## 
## @description
## Generates release agnostic/specific definition hash, to be used to register the Fiji point with NOTE::Fiji.
##
## @return
## Hashref with tag, progr, param.
## 
## @arg tag Required: Tag corresponding to the Fiji point for which the definition is needed.
## 
## @examples
## $obj->get_definition_hash(tag => C_FIJI_TAG_DISABLE_CP);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_definition_hash
{
        my $class = shift;
	my %args = validate(@_,
	{
                tag  => { type => SCALAR, },
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $fiji_tag = $args{tag};
        my $ret_hash = {};
        
	if ($ontap_ver eq '7.3')
	{
		$ret_hash = $class->_fiji_points_7_3( tag => $fiji_tag );
	}
	elsif ($ontap_ver eq '8.0')
	{
		$ret_hash = $class->_fiji_points_BR_0( tag => $fiji_tag );
	}
	elsif ($ontap_ver eq '8.1')
	{
		$ret_hash = $class->_fiji_points_8_1( tag => $fiji_tag );
	}
	elsif ($ontap_ver eq '8.2')
	{
		$ret_hash = $class->_fiji_points_8_2( tag => $fiji_tag );
	}
	else
	{
		logcomment("FijiPoints: Specific support for sw_version $ontap_ver not found. Will be using base definitions") if ($class->{_debug});
	}
        
        if ( !defined($ret_hash) || !(scalar %$ret_hash) )
        {
                logcomment( "FijiPoints: Tag=$fiji_tag: Could not find fiji point definition specific for $ontap_ver." .
                           " Defaulting to base definitions. " ) if ($class->{_debug});
       		$ret_hash =  $class->_base_fiji_points( tag => $fiji_tag );
        }
        
        # Add the tag to the definition hash
        $ret_hash->{tag} = $fiji_tag;

        return $ret_hash;
}


###################################
# Internal subroutines
###################################

# We need this internal method and the _generate_tags_only arg
# to be able to only register with NOTE the fiji points that are allowed/supported for this release,
# as they vary from one release to another
sub _register_fiji_points
{
	my $class = shift;
	my %args = validate(@_,
	{
                # No args
	} );
	
	my $ontap_ver = $class->{_version}->{ontap};
	
	# First get all the tags that are legal for this release
	# Get the release-specific first, then the base ones
	if ($ontap_ver eq '7.3')
	{
		$class->_fiji_points_7_3( _generate_tags_only => TRUE );
	}
	elsif ($ontap_ver eq '8.0')
	{
		$class->_fiji_points_BR_0( _generate_tags_only => TRUE );
	}
	elsif ($ontap_ver eq '8.1')
	{
		$class->_fiji_points_8_1( _generate_tags_only => TRUE );
	}
	elsif ($ontap_ver eq '8.2')
	{
		$class->_fiji_points_8_2( _generate_tags_only => TRUE );
	}
	
	$class->_base_fiji_points( _generate_tags_only => TRUE );
	
	my $tags = $class->{_tags};
	logcomment("FijiPoints: registering Fiji points for release $ontap_ver\n" . Dumper($tags)) if ($class->{_debug});
	
	foreach my $tag (@$tags)
	{
		my $fiji_def = $class->get_definition_hash( tag => $tag );
		if ($class->{_boot_mode} eq C_ONTAP_CMODE)
		{
			NOTE::GX::Fiji->register( %$fiji_def );
		}
		else
		{
			NOTE::Fiji->register( %$fiji_def );
		}
	}
}


sub _base_fiji_points
{
	my $class = shift;
	my %args = validate(@_,
	{
                tag  => { type => SCALAR, optional => 1},
		_generate_tags_only => { type => SCALAR, optional => 1, default => FALSE },
	} );

	logcomment("FijiPoints: _base_fiji_points: ". Dumper(%args) ) if ($class->{_debug});

	my $base_fiji_points_hash = {
                'Disable timer-triggered CP' => {prog => '(wafl.disable.timer.CP(eq @yesno yes)(return 1))',
                                                 param => {},
                                                 quiet => 1,
                                                 },
                
                'Enable timer-triggered CP' => {prog => '(wafl.disable.timer.CP(eq @yesno no)(return 1))',
                                                param => {},
                                                quiet => 1,
                                                },
                
                'Panic on compress' => {prog => '(wafl.vvol.compression.panic(&& (eq @when when_compressing) (eq @vol %s)) (return 1))',
                                                param => ['volname'],
                                                quiet => 1,
                                                },
                
                'Panic on decompress' => {prog => '(wafl.vvol.compression.panic(&& (eq @when when_decompressing) (eq @vol %s)) (return 1))',
                                                param => ['volname'],
                                                quiet => 1,
                                                },
                
                'Corrupt checksum on compress' => {prog => '(wafl.vvol.compress.checksum(eq @volume %s) (return 1))',
                                                param => ['volname'],
                                                quiet => 1,
                                                },
                
                'Corrupt checksum on decompress' => {prog => '(wafl.vvol.decompress.checksum(eq @volume %s) (return 1))',
                                                param => ['volname'],
                                                quiet => 1,
                                                },
	};
	
	if ($args{_generate_tags_only} == TRUE)
	{
		my $tags = $class->{_tags};
		
		foreach my $tag (keys %$base_fiji_points_hash)
		{
			if (!grep(/^$tag$/, @$tags))
			{
				push(@$tags, $tag);
			}
		}
	}
	else
	{
		my $fiji_tag = $args{tag};
		if (!defined($fiji_tag))
		{
			logresult('script', "Library problem found: _base_fiji_points called with no tag")
		}
		my $fiji_def = $base_fiji_points_hash->{$fiji_tag};
		
		if (!defined($fiji_def))
		{
			logresult('script', "This library only maintains compression-specific Fiji points.\n" .
				            "Fiji tag \"$fiji_tag\" is not supported for this release.");
		}
		logcomment("FijiPoints: _base_fiji_points: Tag=$fiji_tag, definition hash: " . Dumper($fiji_def)) if ($class->{_debug});
	
		return $fiji_def;
	}
}



sub _fiji_points_7_3
{
	my $class = shift;
	my %args = validate(@_,
	{
                tag  => { type => SCALAR, optional => 1 },
		_generate_tags_only => { type => SCALAR, optional => 1, default => FALSE },
	} );
        
        logcomment("FijiPoints: _fiji_points_7_3: ". Dumper(%args) ) if ($class->{_debug});
        
	my $fiji_points_hash_7_3 = {};
	
	if ($args{_generate_tags_only} == TRUE)
	{
		my $tags = $class->{_tags};
		foreach my $tag (keys %$fiji_points_hash_7_3)
		{
			if (!grep(/^$tag$/, @$tags))
			{
				push(@$tags, $tag);
			}
		}
	}
	else
	{
		my $fiji_tag = $args{tag};
		if (!defined($fiji_tag))
		{
			logresult('script', "Library problem found: _fiji_points_7_3 called with no tag")
		}
        
		return $fiji_points_hash_7_3->{$fiji_tag};
	}
}

sub _fiji_points_BR_0
{
        my $class = shift;
	my %args = validate(@_,
	{
                tag  => { type => SCALAR, optional => 1 },
		_generate_tags_only => { type => SCALAR, optional => 1, default => FALSE },
	} );
        
        logcomment("FijiPoints: _fiji_points_BR_0: ". Dumper(%args) ) if ($class->{_debug});
        
	my $fiji_points_hash_BR_0 = {
                'Suppress decompress completion ems' => {prog => '(wafl.vvol.decompress.terminate (eq @volume %s) (return 1))',
                                                param => ['volname'],
                                                quiet => 1,
                                                },
                'Sleep during compress scanner' => {prog => '(wafl.wp1m.delay.backdoor (eq @vol %s) (return %d))',
                                                    param => ['volname', {sleep_time => C_DFLT_FIJI_COMPR_SCANNER_SLEEP_TIME}],
                                                    quiet => 1,
                                                    },
        };
	
	if ($args{_generate_tags_only} == TRUE)
	{
		my $tags = $class->{_tags};
		foreach my $tag (keys %$fiji_points_hash_BR_0)
		{
			if (!grep(/^$tag$/, @$tags))
			{
				push(@$tags, $tag);
			}
		}
	}
	else
	{
		my $fiji_tag = $args{tag};
		if (!defined($fiji_tag))
		{
			logresult('script', "Library problem found: _fiji_points_BR_0 called with no tag")
		}
		return $fiji_points_hash_BR_0->{$fiji_tag};
	}
}

sub _fiji_points_8_1
{
        my $class = shift;
	my %args = validate(@_,
	{
                tag  => { type => SCALAR, optional => 1 },
		_generate_tags_only => { type => SCALAR, optional => 1, default => FALSE },
	} );
        
        logcomment("FijiPoints: _fiji_points_8_1: ". Dumper(%args) ) if ($class->{_debug});
        
	my $fiji_points_hash_8_1 = {
                'Suppress decompress completion ems' => {prog => '(wafl.vvol.decompress.terminate (eq @volume %s) (return 1))',
                                                param => ['volname'],
                                                quiet => 1,
                                                },
                'Dense sleep at stage' => {prog => '(dense.sleep.stage (&& (eq @sleep_stage %s) (eq @vol %s)) (return %d))',
                                                param => ['sleep_stage', 'volname', {sleep_time => C_DFLT_FIJI_DENSE_SLEEP_TIME}],
                                                quiet => 1,
                                                },
                'Dense sleep before record checkpoint' => {prog => '(dense.sleep.before.record.checkpoint (&& (eq @sleep_stage %s) (eq @vol %s)) (return %d))',
                                                param => ['sleep_stage', 'volname', {sleep_time => C_DFLT_FIJI_DENSE_SLEEP_TIME}],
                                                quiet => 1,
                                                },
                'Dense sleep after record checkpoint' => {prog => '(dense.sleep.after.record.checkpoint (&& (eq @sleep_stage %s) (eq @vol %s)) (return %d))',
                                                param => ['sleep_stage', 'volname', {sleep_time => C_DFLT_FIJI_DENSE_SLEEP_TIME}],
                                                quiet => 1,
                                                },
                'Simulate REP error on compression metafile' => {prog => '(dense.compression_metafile.err(eq @path %s) (return %d))',
                                                                 param => ['volname', {error_value => C_FIJI_DFLT_REP_METAFILE_ERROR}],
                                                                 quiet => 1,
                                                                 },
                'Simulate REP error while reading changelog' => {prog => '(dense.compression_filter.err(eq @path %s) (return %d))',
                                                                 param => ['volname', {error_value => C_FIJI_DFLT_REP_METAFILE_ERROR}],
                                                                 quiet => 1,
                                                                 },
                'Suspend 64-bit upgrade scan' => {prog => '(repeat 999999 (wafl.scan.abort(&& (eq @scan_type SCAN_64BIT_UPGRADE)(== @inode_num %d) (== @is_priv 0) (== @level %d) (== @fbn %d) (eq @volname %s))(return(true))))',
                                                param => ['inode_num', {level => 0}, 'fbn', 'volname'],
                                                quiet => 1,
                                                },
                'Suspend file clone operation' => {prog => '(repeat 999 (wafl.sis.clone.sleep.before.blkcount(&& (eq @vol %s) (== @fileid %d)) (return true)))',
                                                param => ['volname', 'inode_num'],
                                                quiet => 1,
                                                },
        };
	
	if ($args{_generate_tags_only} == TRUE)
	{
		my $tags = $class->{_tags};
		foreach my $tag (keys %$fiji_points_hash_8_1)
		{
			if (!grep(/^$tag$/, @$tags))
			{
				push(@$tags, $tag);
			}
		}
	}
	else
	{
		my $fiji_tag = $args{tag};
		if (!defined($fiji_tag))
		{
			logresult('script', "Library problem found: _fiji_points_8_1 called with no tag")
		}
		return $fiji_points_hash_8_1->{$fiji_tag};
	}
}


sub _fiji_points_8_2
{
        my $class = shift;
	my %args = validate(@_,
	{
                tag  => { type => SCALAR, optional => 1 },
		_generate_tags_only => { type => SCALAR, optional => 1, default => FALSE },
	} );
        
        logcomment("FijiPoints: _fiji_points_8_2: ". Dumper(%args) ) if ($class->{_debug});
        
	my $fiji_points_hash_8_2 = {
                'Suppress decompress completion ems' => {prog => '(wafl.vvol.decompress.terminate (eq @volume %s) (return 1))',
                                                param => ['volname'],
                                                quiet => 1,
                                                },
                'Dense sleep at stage' => {prog => '(dense.sleep.stage (&& (eq @sleep_stage %s) (eq @vol %s)) (return %d))',
                                                param => ['sleep_stage', 'volname', {sleep_time => C_DFLT_FIJI_DENSE_SLEEP_TIME}],
                                                quiet => 1,
                                                },
                'Dense sleep before record checkpoint' => {prog => '(dense.sleep.before.record.checkpoint (&& (eq @sleep_stage %s) (eq @vol %s)) (return %d))',
                                                param => ['sleep_stage', 'volname', {sleep_time => C_DFLT_FIJI_DENSE_SLEEP_TIME}],
                                                quiet => 1,
                                                },
                'Dense sleep after record checkpoint' => {prog => '(dense.sleep.after.record.checkpoint (&& (eq @sleep_stage %s) (eq @vol %s)) (return %d))',
                                                param => ['sleep_stage', 'volname', {sleep_time => C_DFLT_FIJI_DENSE_SLEEP_TIME}],
                                                quiet => 1,
                                                },
                'Simulate REP error on compression metafile' => {prog => '(dense.compression_metafile.err(eq @path %s) (return %d))',
                                                                 param => ['volname', {error_value => C_FIJI_DFLT_REP_METAFILE_ERROR}],
                                                                 quiet => 1,
                                                                 },
                'Simulate REP error while reading changelog' => {prog => '(dense.compression_filter.err(eq @path %s) (return %d))',
                                                                 param => ['volname', {error_value => C_FIJI_DFLT_REP_METAFILE_ERROR}],
                                                                 quiet => 1,
                                                                 },
                'Suspend 64-bit upgrade scan' => {prog => '(repeat 999999 (wafl.scan.abort(&& (eq @scan_type SCAN_64BIT_UPGRADE)(== @inode_num %d) (== @is_priv 0) (== @level %d) (== @fbn %d) (eq @volname %s))(return(true))))',
                                                param => ['inode_num', {level => 0}, 'fbn', 'volname'],
                                                quiet => 1,
                                                },
                'Suspend file clone operation' => {prog => '(repeat 999 (wafl.sis.clone.sleep.before.blkcount(&& (eq @vol %s) (== @fileid %d)) (return true)))',
                                                param => ['volname', 'inode_num'],
                                                quiet => 1,
                                                },
        };
	
	if ($args{_generate_tags_only} == TRUE)
	{
		my $tags = $class->{_tags};
		foreach my $tag (keys %$fiji_points_hash_8_2)
		{
			if (!grep(/^$tag$/, @$tags))
			{
				push(@$tags, $tag);
			}
		}
	}
	else
	{
		my $fiji_tag = $args{tag};
		if (!defined($fiji_tag))
		{
			logresult('script', "Library problem found: _fiji_points_8_2 called with no tag")
		}
		return $fiji_points_hash_8_2->{$fiji_tag};
	}
}



1;
