#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/CLI/Formats.pm#1 $
#

## @summary Utility module for parsing formatted CLI output.
##
## @description
## This module provides functionality for parsing of CLI command output
## that is expected to be in a formatted form.
## For example, the output of the 'df -S' command:
## Filesystem                used       compressed         a-sis     %saved
## /vol/test_vol/        401160KB        2750968KB           0KB        87%.
##
## @synopsis
##
## use Compression::Constants;
## use Compression::UI::CLI::Formats;
## use Compression::Version;
## use NOTE::Filer;  # To obtain filer version.
##
## my $filer = NOTE::Filer->connect( host => $FILER );
## my $version = Compression::Version->new( conn => $filer );
## my $obj = Compression::UI::CLI::Formats->new( version => $version );
##
## @examples 
## $obj->check_and_process_output_if_formatted(code => C_CMD_GET_SAVINGS,
##                                             output => $output,
##                                             required_return_hash_key => $volume,
##                                             modifier => $modifier,);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)



package Compression::UI::CLI::Formats;

use strict;
use warnings;

use Tharn;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Compression::Constants qw(:all);
use Compression::UI::CLI::Formats::Type1;
use Compression::UI::CLI::Formats::Type2;


BEGIN
{
	require Exporter;
	our @ISA = qw(Exporter);
}


## @name ::new
## @summary Compression::UI::CLI::Formats Constructor.
##
## @return Object of type Compression::UI::CLI::Formats.
##
## @arg version Required: Compression::Version object for the hosting filer.
## 
## @arg boot_mode Optional: Boot mode of the hosting filer.
## Defaults to 7M.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::UI::CLI::Formats->new( version => $version );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub new
{
	my $class = shift;
        my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug  => {  
			type 	 => SCALAR,
			optional => 1,
                        default  => FALSE,
			},
	} );
        
	my $self = {};
        
        $self->{_debug} = $args{debug};
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};
        
        $self->{formats_type1} = Compression::UI::CLI::Formats::Type1->new(version => $self->{_version},
                                                                           boot_mode => $self->{_boot_mode},
                                                                           debug => $self->{_debug});
        $self->{formats_type2} = Compression::UI::CLI::Formats::Type2->new(version => $self->{_version},
                                                                           boot_mode => $self->{_boot_mode},
                                                                           debug => $self->{_debug});

	logcomment("Created new Formats object:\n" . Dumper($self)) if ($self->{_debug});
	bless $self, $class;

	return $self;
}


## @name ::check_and_process_output_if_formatted.
## @summary Wrapper for processing of formatted CLI command output.
##
## @description
## Wrapper for processing of CLI command output that is expected to be in a formatted form.
## If, for the specific release, the command output is not expected to be formatted,
## or it is supposed to be formatted, but the format is not as expected,
## it returns C_RC_FAILURE, and the command output.
## If the command output is formatted, the output is processed
## and all the information is automatically formatted into a hash with the predefined keys.
##
## @return Hash with rc, output, and pre-defined key for the values hash.
##
## @arg code Required: Code corresponding to the operation for which the output is parsed.
## Operation codes are defined in Compression::Constants.
##
## @arg output Required: Command output.
##
## @arg modifier Optional: Modifier, to be used for space-specific commands (k/m/g).
##
## @arg required_return_hash_key Optional: For formatted command output that need to have an entry
## for a specific entity (for example, space savings for a specific volume). If required_return_hash_key
## is defined, but the command output did not contain an entry for the specified entity,
## then the script is terminated with a fatal error.
##
## @arg disable_data_identifier_processing Optional: Flag to disable any processing on the data identifier
## (First entry in the data line, for example: /vol/vol_name for 'df -S').
##
## @examples
## my %result = $obj->check_and_process_output_if_formatted(code => C_CMD_GET_SAVINGS,
##                                                          output => $output,
##                                                          required_return_hash_key => $volume,
##                                                          modifier => $modifier,);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub check_and_process_output_if_formatted
{
        
        my $class = shift;
	my %args = validate(@_,
	{
                code     => { type => SCALAR, },
                output   => { type => SCALAR|UNDEF, },
                modifier => {type => SCALAR|UNDEF, optional => 1,},
                required_return_hash_key => {type => SCALAR|UNDEF, optional => 1,},
                disable_data_identifier_processing => {type => BOOLEAN, optional => 1, default => FALSE},
	} );
        
        logcomment("Formats: check_and_process_output_if_formatted input:\n" . Dumper(%args)) if ($class->{_debug});
        
        my $code = $args{code};
        my $output = $args{output};
        
        my %return_hash = ( rc => C_RC_FAILURE, output => $output );

        # If command corresponding to this code does not expect a formatted output, bail out imediatelly
        my $format;
        my $boot_mode = $class->{_boot_mode};
        if ($boot_mode eq C_ONTAP_7MODE)
        {
                $format = $class->_check_if_7mode_formatted_output_expected( code => $code );
        }
        elsif($boot_mode eq C_ONTAP_CMODE)
        {
                $format = $class->_check_if_Cmode_formatted_output_expected( code => $code );
        }
        else
        {
                logresult('script', "Unrecognized boot_mode: \"$boot_mode\"");
        }
        if ( $format->{format_expected} == FALSE )
        {
                logcomment("Formats: formatted output is not expected for code $code\n") if ($class->{_debug});
                return %return_hash;
        }
        
        # If variable format is allowed, then we will do our best-effort to parse the output,
        # but even if parsing fails, we will still return success. The upper-level subroutine should handle the failure
        my $variable_format = (defined($format->{variable_format})) ? $format->{variable_format} : FALSE;
        if ($variable_format == TRUE)
        {
                $return_hash{rc} = C_RC_SUCCESS;
        }
        
        if (!defined($output) || ($output eq ''))
        {
                logcomment("Formats: formatted output is expected for code $code, but output is empty\n") if ($class->{_debug});
                return %return_hash;
        }
        
        my $expected_format_types = $format->{format_types};
        my $expected_format_types_count = scalar(@$expected_format_types);
        my $expect_multiple_formats = ($expected_format_types_count > 1) ? TRUE : FALSE;
        my $return_hash_keys = $format->{return_hash_keys};
        my @return_hash_keys_array = (defined($return_hash_keys) && (ref($return_hash_keys) eq 'ARRAY')) ? @$return_hash_keys : ();
        my $output_to_be_processed = $output;
        # Remove multiple new lines
        $output_to_be_processed =~ s/\n\n+/\n/g;
        # Remove any empty lines before the output
        $output_to_be_processed =~ s/^\n//;
        
        logcomment("Formats: formatted output is expected for code $code. Format types:\n" . Dumper($expected_format_types)) if ($class->{_debug});
        my $index = 1;
        foreach my $format_type (@$expected_format_types)
        {
                my $attempt_return_hash_key = shift(@return_hash_keys_array);
                my $return_hash_key = $attempt_return_hash_key || "formatted_output_$index";
                if (!defined($output_to_be_processed) || ($output_to_be_processed eq ''))
                {
                        if ($variable_format == TRUE)
                        {
                                logcomment("Code $code: variable output allowed: output is incomplete\n".
                                           "Missing formatted output of type $format_type\n" .
                                           "Output:\n$output");
                                return %return_hash;
                        }
                        
                        logresult('fatal', "Formatted output is expected for code $code, but output is incomplete\n".
                                           "Missing formatted output of type $format_type\n" .
                                           "Output:\n$output");
                }
                
                my %processed = $class->_process_formatted_output_by_type(code => $code,
                                                                          format_type => $format_type,
                                                                          allow_multiple_formats => $expect_multiple_formats,
                                                                          output => $output_to_be_processed,
                                                                          modifier => $args{modifier},
                                                                          required_return_hash_key => $args{required_return_hash_key},
                                                                          disable_data_identifier_processing => $args{disable_data_identifier_processing},
                                                                          variable_format => $variable_format,
                                                                          );
                logcomment("Formats: finished processing format type $format_type. Partially processed output:\n" .
                           Dumper(%processed)) if ($class->{_debug});
                
                # Delete the hash entries that are for internal use
                my $rc = $processed{rc};
                $output_to_be_processed = $processed{unprocessed_output};
                delete($processed{rc});
                delete($processed{unprocessed_output});
                
                if ($rc != C_RC_SUCCESS)
                {
                        logcomment("Formats: Non-fatal error encountered while processing format type $format_type") if ($class->{_debug});
                        return %return_hash;
                }
                
                # What is left is only "processed data". Add it to the return hash
                foreach my $key (keys(%processed))
                {
                       $return_hash{$return_hash_key}{$key} =  $processed{$key};
                }
                
                $index++;
                logcomment("Formats: partial return hash (after processing format type $format_type):\n" . Dumper(%return_hash)) if ($class->{_debug});
        }
        
        if (!defined($output_to_be_processed) || ($output_to_be_processed eq ''))
        {
                $return_hash{rc} = C_RC_SUCCESS;
        }
        else 
        {
                # Print a warning. This may be a valid case, when expecting non-success rc,
                # But it may also mean that some formatted output was incomplete.
                logcomment("Expected formatted output could not be completely processed\n" .
                           "Remaining output:\n$output_to_be_processed") if ($class->{_debug});
                
                # If variable format, make an extra effort to parse the rest of the output.
                # Still, do not change the return code.
                if (($variable_format == TRUE) && (defined($format->{extra_variable_output})))
                {
                        my %extra_variable_output = $class->_process_extra_variable_output(extra_variable_formats => $format->{extra_variable_output},
                                                                                           output => $output_to_be_processed,
                                                                                           code => $code);
                        foreach my $extra_variable_key (keys %extra_variable_output)
                        {
                                $return_hash{$extra_variable_key} = $extra_variable_output{$extra_variable_key};
                        }
                }
        }
        
        logcomment("Formats: check_and_process_output_if_formatted returning:\n" . Dumper(%return_hash)) if ($class->{_debug});
        return %return_hash;
}


#########################################
# Internal subroutines
#########################################

sub _process_extra_variable_output
{
        my $class = shift;
	my %args = validate(@_,
	{
                code => { type => SCALAR, },
                output => { type => SCALAR, },
                extra_variable_formats => { type => ARRAYREF, },
	} );
        
        my $code = $args{code};
        my $output = $args{output};
        my $extra_variable_formats = $args{extra_variable_formats};
        my $output_to_be_processed = $output;
        my $extra_processing_rc = C_RC_SUCCESS;
        my %return_hash = ();
        
        logcomment("Formats: _process_extra_variable_output called with:\n" .
                   Dumper(%args)) if ($class->{_debug});
        
        foreach my $extra_variable_type (@$extra_variable_formats)
        {
                logcomment("Formats: code $code: attempting to process extra variable output:\n" .
                           Dumper($extra_variable_type)) if ($class->{_debug});
                
                my $data_identifier_prefix = (defined($extra_variable_type->{data_identifier}->{prefix})) ? $extra_variable_type->{data_identifier}->{prefix} : '';
                my $data_identifier_suffix = (defined($extra_variable_type->{data_identifier}->{suffix})) ? $extra_variable_type->{data_identifier}->{suffix} : '';
                my $return_hash_key = (defined($extra_variable_type->{return_hash_key})) ? $extra_variable_type->{return_hash_key} : 'extra_formatted_output';
                
                # First remove any 'unwanted' lines from the beginning of the output
                while (($output_to_be_processed =~ /^$data_identifier_prefix(.+)$data_identifier_suffix/) &&
                       ($extra_processing_rc == C_RC_SUCCESS))
                {
                        my $data_identifier = $1;
                        $data_identifier =~ s/^\s*//;
                        $data_identifier =~ s/\s*$//;
                        
                        logcomment("Formats: code $code: processing extra variable output:$output_to_be_processed\n") if ($class->{_debug});
                        
                        my %processed_hash = $class->_process_formatted_output_by_type(code => $extra_variable_type->{code},
                                                                                       output => $output_to_be_processed,
                                                                                       allow_multiple_formats => TRUE,
                                                                                       format_type => $extra_variable_type->{type},
                                                                                       variable_format => FALSE);
                        
                        logcomment("Formats: code $code: processing extra variable output:\n" .
                                   "Data identifier: $data_identifier\n" .
                                   "Partially processed output:\n" . Dumper(%processed_hash)) if ($class->{_debug});
                        
                        $extra_processing_rc = $processed_hash{rc};
                        delete($processed_hash{rc});
                        $output_to_be_processed = $processed_hash{unprocessed_output};
                        delete($processed_hash{unprocessed_output});
                        
                        foreach my $processed_entry (keys %processed_hash)
                        {
                                logcomment("Formats: code $code: adding entry to return hash:\n" .
                                           "Key: $return_hash_key\nEntry: $processed_entry") if ($class->{_debug});
                                if ((defined($data_identifier)) && ($data_identifier ne ''))
                                {
                                        $return_hash{$return_hash_key}{$data_identifier}{$processed_entry} = $processed_hash{$processed_entry};
                                }
                                else
                                {
                                        $return_hash{$return_hash_key}{$processed_entry} = $processed_hash{$processed_entry};
                                }
                                
                        }
                }
                
                if (!(defined($output_to_be_processed)) || ($output_to_be_processed eq ''))
                {
                        last;
                }
        }
        
        logcomment("Formats: code $code: _process_extra_variable_output returning:\n" . Dumper(%return_hash)) if ($class->{_debug});
        return %return_hash;
}

sub _process_formatted_output_by_type
{
        my $class = shift;
	my %args = validate(@_,
	{
                code => { type => SCALAR, },
                output => { type => SCALAR, },
                allow_multiple_formats => { type => BOOLEAN, },
                format_type => { type => SCALAR, },
                modifier => {type => SCALAR|UNDEF, optional => 1,},
                required_return_hash_key => {type => SCALAR|UNDEF, optional => 1,},
                disable_data_identifier_processing => {type => BOOLEAN, optional => 1, default => FALSE},
                variable_format => {type => BOOLEAN, optional => 1, default => FALSE},
	} );
        
        my $code = $args{code};
        my $format_type = $args{format_type};
        delete($args{format_type});
        
        # Note:
        # For any type of formats, we should have a standard _process_output method
        # Requirements for these methods are:
        # - they should permit the same arguments as _process_formatted_output_by_type (except for format_type).
        #   Depending on the format type, some arguments may be no-op
        # - They need to return a hash with:
        #   - rc: a non-success rc means that "expected formatted" parsing will stop
        #         and all the output will be sent down to Messages for "error message" parsing.
        #         Use this for parsing errors that can be considered "non-fatal" (could be legal error messages).
        #   - unprocessed_output: this is the substring of the output that could not be processed
        #         for the current format type
        #   - any hash entry besides rc and unprocessed_output is interpreted as the succesfully-parsed result
        #         and will be added to the return hash of check_and_process_output_if_formatted.
        if ($format_type == C_FORMAT_TYPE_1)
	{
		return $class->{formats_type1}->_process_output(%args)
	}
	elsif ($format_type == C_FORMAT_TYPE_2)
	{
		return $class->{formats_type2}->_process_output(%args)
	}
	else
	{
		logresult('script', "Library problem found: unsupported format type required by command code $code: $format_type");
	}
}


sub _check_if_7mode_formatted_output_expected
{
        my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $code = $args{code};
        my $return_hash = {};
        
	if ($ontap_ver eq '7.3')
	{
		$return_hash = $class->_7mode_formats_7_3( code => $code );
	}
	elsif ($ontap_ver eq '8.0')
	{
		$return_hash = $class->_7mode_formats_8_0( code => $code );
	}
	elsif ($ontap_ver eq '8.1')
	{
		$return_hash = $class->_7mode_formats_8_1( code => $code );
        }
	elsif ($ontap_ver eq '8.2')
	{
		$return_hash = $class->_7mode_formats_8_2( code => $code );
        }
	else
	{
		logcomment("Formats: Specific support for 7mode sw_version $ontap_ver not found. Will be using base formats") if ($class->{_debug});
	}
        
        if ( !defined($return_hash) || !(scalar %$return_hash) )
        {
                logcomment( "Formats: Could not find 7mode format definition specific for $ontap_ver. Checking base formats.") if ($class->{_debug});
       		$return_hash =  $class->_7mode_base_formats( code => $code );
        }

        return $return_hash;
}




sub _7mode_base_formats
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );

	logcomment("Formats: _7mode_base_formats input:\n". Dumper(%args) ) if( $class->{_debug} );

	my $base_formats_hash = {
                GET_SPACE_SAVINGS => {format_expected => TRUE,
                                      format_types => [C_FORMAT_TYPE_1],
                                      return_hash_keys => ['savings'],
                                      },
                DECOMPRESS_STATUS => {format_expected => TRUE,
                                      format_types => [C_FORMAT_TYPE_1],
                                      return_hash_keys => ['percent'],
                                      },
                GET_VBN_INFO      => {format_expected => TRUE,
                                      format_types => [C_FORMAT_TYPE_1],
                                      return_hash_keys => ['vbn_info'],
                                      },
                GET_FILE_ID       => {format_expected => TRUE,
                                      format_types => [C_FORMAT_TYPE_2],
                                      return_hash_keys => ['fh_info'],
                                      },
                FILE_FOLDING_STATUS => {format_expected => TRUE,
                                        format_types => [C_FORMAT_TYPE_1],
                                        variable_format => TRUE,
                                        return_hash_keys => ['wafl_scans'],
                                        },
                GET_CONTAINMENT_VERSION => {format_expected => TRUE,
                                            format_types => [C_FORMAT_TYPE_2],
                                            return_hash_keys => ['containment_version'],
                                            },
                GET_COMPRESSION_STATS => {format_expected => TRUE,
                                          format_types => [C_FORMAT_TYPE_2],
                                          return_hash_keys => ['compression_stats'],
                                          },
                GET_WAFL_SCAN_SPEED => {format_expected => TRUE,
                                        format_types => [C_FORMAT_TYPE_2],
                                        return_hash_keys => ['wafl_scan_speed'],
                                        },
                GET_SNAPLOCK_CONTAINMENT_BITS => {format_expected => TRUE,
                                                  format_types => [C_FORMAT_TYPE_2],
                                                  return_hash_keys => ['snaplock_containment_stats'],
                                                  },
		VOL_MOVE_STATUS => {format_expected => TRUE,
				    format_types => [C_FORMAT_TYPE_1],
				    return_hash_keys => ['move_status'],
				    },
		VFILER_MIGRATE_STATUS => {format_expected => TRUE,
					  format_types => [C_FORMAT_TYPE_1],
					  return_hash_keys => ['migrate_status'],
					  },
		GET_VOLUME_STATS => {format_expected => TRUE,
				     format_types => [C_FORMAT_TYPE_2],
				     return_hash_keys => ['vol_stats'],
				     },
		GET_WAFL_STATS => {format_expected => TRUE,
				   format_types => [C_FORMAT_TYPE_2],
				   return_hash_keys => ['wafl_stats'],},
	};

	my $code = $args{code};
	my $format = $base_formats_hash->{$code};
        
        if (!defined($format))
        {
                $format = { format_expected => FALSE };
                logcomment("Code \"$code\": Command output is not expected to be formatted.") if ($class->{_debug});
        }
	logcomment("Formats: _7mode_base_formats returning: Code=$code, option hash:\n" . Dumper($format)) if ($class->{_debug});

	return $format;
}

sub _7mode_formats_7_3
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        logcomment("Formats: _7mode_formats_7_3:\n". Dumper(%args) ) if ($class->{_debug});
        
	my $formats_hash_7_3 = {};
	my $code = $args{code};
        
        return $formats_hash_7_3->{$code};
}

sub _7mode_formats_8_0
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        logcomment("Formats: _7mode_formats_8_0:\n". Dumper(%args) ) if ($class->{_debug});
        
	my $formats_hash_8_0 = {
		COMPRESS_STATUS => {format_expected => TRUE,
                                    format_types => [C_FORMAT_TYPE_1],
                                    return_hash_keys => ['compress_statistics'],
                                    },
                COMPRESS_STATUS_CHECKPOINT => {format_expected => TRUE,
                                               format_types => [C_FORMAT_TYPE_1,
                                                                C_FORMAT_TYPE_2],
                                               return_hash_keys => ['compress_statistics',
                                                                    'checkpoint_info'],
                                               },
	};
	my $code = $args{code};
        
        return $formats_hash_8_0->{$code};
}

sub _7mode_formats_8_1
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        logcomment("Formats: _7mode_formats_8_1:\n". Dumper(%args) ) if ($class->{_debug});
        
	my $formats_hash_8_1 = {
                COMPRESS_STATUS => {format_expected => TRUE,
                                    format_types => [C_FORMAT_TYPE_2],
                                    return_hash_keys => ['compress_statistics'],
                                    },
		COMPRESS_STATUS_CHECKPOINT => {format_expected => TRUE,
                                               format_types => [C_FORMAT_TYPE_2],
                                               return_hash_keys => ['checkpoint_info'],
                                               },
                DECOMPRESS_STATUS => {format_expected => TRUE,
                                      format_types => [C_FORMAT_TYPE_2],
                                      return_hash_keys => ['decompress_statistics'],
                                      },
                GET_STATS => {format_expected => TRUE,
                              format_types => [C_FORMAT_TYPE_2],
                              return_hash_keys => ['statistics'],
                              },
                VERIFY_COMPR_ATTEMPTED_BIT => {format_expected => TRUE,
                                               format_types => [C_FORMAT_TYPE_1,
                                                                C_FORMAT_TYPE_1,
                                                                C_FORMAT_TYPE_2],
                                               return_hash_keys => ['fingerprints',
                                                                    'fingerprints_vvol',
                                                                    'checking_stats'],
                                               variable_format => TRUE,
                                               extra_variable_output => [{code => _C_CMD_CHANGELOG_FINGERPRINTS,
                                                                          type => C_FORMAT_TYPE_1,
                                                                          return_hash_key => 'changelog_buckets',
                                                                          data_identifier => {prefix => '\s*contents of bucket:\s*',
                                                                                              suffix => '\s*\(number of entries: \d+\)',},},],
                                               },
	};
	my $code = $args{code};
        
        return $formats_hash_8_1->{$code};
}


sub _7mode_formats_8_2
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        logcomment("Formats: _7mode_formats_8_2:\n". Dumper(%args) ) if ($class->{_debug});
        
	my $formats_hash = {
                COMPRESS_STATUS => {format_expected => TRUE,
                                    format_types => [C_FORMAT_TYPE_2],
                                    return_hash_keys => ['compress_statistics'],
                                    },
		COMPRESS_STATUS_CHECKPOINT => {format_expected => TRUE,
                                               format_types => [C_FORMAT_TYPE_2],
                                               return_hash_keys => ['checkpoint_info'],
                                               },
                DECOMPRESS_STATUS => {format_expected => TRUE,
                                      format_types => [C_FORMAT_TYPE_2],
                                      return_hash_keys => ['decompress_statistics'],
                                      },
                GET_STATS => {format_expected => TRUE,
                              format_types => [C_FORMAT_TYPE_2],
                              return_hash_keys => ['statistics'],
                              },
                VERIFY_COMPR_ATTEMPTED_BIT => {format_expected => TRUE,
                                               format_types => [C_FORMAT_TYPE_1,
                                                                C_FORMAT_TYPE_1,
																C_FORMAT_TYPE_1,
                                                                C_FORMAT_TYPE_2],
                                               return_hash_keys => ['fingerprints',
                                                                    'fingerprints_vvol',
                                                                    'checking_stats'],
                                               variable_format => TRUE,
                                               extra_variable_output => [{code => _C_CMD_CHANGELOG_FINGERPRINTS,
                                                                          type => C_FORMAT_TYPE_1,
                                                                          return_hash_key => 'changelog_buckets',
                                                                          data_identifier => {prefix => '\s*contents of bucket:\s*',
                                                                                              suffix => '\s*\(number of entries: \d+\)',},},],
                                               },
	};
	my $code = $args{code};
        
        return $formats_hash->{$code};
}


sub _check_if_Cmode_formatted_output_expected
{
        my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $code = $args{code};
        my $return_hash = {};
        
	if ($ontap_ver eq '8.1')
	{
		$return_hash = $class->_Cmode_formats_8_1( code => $code );
        }
	else
	{
		logcomment("Formats: Specific support for Cmode sw_version $ontap_ver not found. Will be using base formats") if ($class->{_debug});
	}
        
        if ( !defined($return_hash) || !(scalar %$return_hash) )
        {
                logcomment( "Formats: Could not find Cmode format definition specific for $ontap_ver. Checking base formats.") if ($class->{_debug});
       		$return_hash =  $class->_Cmode_base_formats( code => $code );
        }

        return $return_hash;
}


sub _Cmode_base_formats
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );

	logcomment("Formats: _Cmode_base_formats input:\n". Dumper(%args) ) if( $class->{_debug} );

	my $base_formats_hash = {
                GET_SPACE_SAVINGS => {format_expected => TRUE,
                                      format_types => [C_FORMAT_TYPE_1],
                                      return_hash_keys => ['savings'],
                                      },
                COMPRESS_STATUS => {format_expected => TRUE,
                                    format_types => [C_FORMAT_TYPE_2],
                                    return_hash_keys => ['compress_statistics'],
                                    },
                GET_FILE_ID => {format_expected => TRUE,
                                format_types => [C_FORMAT_TYPE_2],
                                return_hash_keys => ['fh_info'],
                                },
                GET_STATS => {format_expected => TRUE,
                              format_types => [C_FORMAT_TYPE_2],
                              return_hash_keys => ['statistics'],
                              },
                VERIFY_COMPR_ATTEMPTED_BIT => {format_expected => TRUE,
                                               format_types => [C_FORMAT_TYPE_1,
                                                                C_FORMAT_TYPE_1,
                                                                C_FORMAT_TYPE_2],
                                               return_hash_keys => ['fingerprints',
                                                                    'fingerprints_vvol',
                                                                    'checking_stats'],
                                               variable_format => TRUE,
                                               extra_variable_output => [{code => _C_CMD_CHANGELOG_FINGERPRINTS,
                                                                          type => C_FORMAT_TYPE_1,
                                                                          return_hash_key => 'changelog_buckets',
                                                                          data_identifier => {prefix => '\s*contents of bucket:\s*',
                                                                                              suffix => '\s*\(number of entries: \d+\)',},},],
                                               },
                COMPRESS_STATUS_CHECKPOINT => {format_expected => TRUE,
                                               format_types => [C_FORMAT_TYPE_2],
                                               return_hash_keys => ['checkpoint_info'],
                                               },
                GET_VBN_INFO      => {format_expected => TRUE,
                                      format_types => [C_FORMAT_TYPE_1],
                                      return_hash_keys => ['vbn_info'],
                                      },
		DECOMPRESS_STATUS => {format_expected => TRUE,
				      format_types => [C_FORMAT_TYPE_2],
				      return_hash_keys => ['decompress_statistics'],
				      },
                GET_SNAP_RECLAIMABLE => {format_expected => TRUE,
                                         format_types => [C_FORMAT_TYPE_2],
                                         return_hash_keys => ['reclaimable_space'],
                                         },
                DISK_FREE => {format_expected => TRUE,
                              format_types => [C_FORMAT_TYPE_1],
                              return_hash_keys => ['df'],
                              },
                GET_RESERVED => {format_expected => TRUE,
                                 format_types => [C_FORMAT_TYPE_1],
                                 return_hash_keys => ['reserved'],
                                 },
                AGGR_SNAP_LIST => {format_expected => TRUE,
                                   format_types => [C_FORMAT_TYPE_1],
                                   variable_format => TRUE,
                                   return_hash_keys => ['snapshots'],
                                   },
                FILE_FOLDING_STATUS => {format_expected => TRUE,
                                        format_types => [C_FORMAT_TYPE_1],
                                        variable_format => TRUE,
                                        return_hash_keys => ['wafl_scans'],
                                        },
                
                DUMPBLOCK_INFO => {format_expected => TRUE,
                                   format_types => [C_FORMAT_TYPE_2],
                                   return_hash_keys => ['dumpblock_info'],
                                   },
                
                DUMPBLOCK_INFO_L1 => {format_expected => TRUE,
                                      format_types => [C_FORMAT_TYPE_2],
                                      return_hash_keys => ['dumpblock_info'],
                                      },
                
                GET_REFCOUNT => {format_expected => TRUE,
                                 format_types => [C_FORMAT_TYPE_1],
                                 return_hash_keys => ['refcount_info'],
                                 },
                
                GET_CONTAINMENT_VERSION => {format_expected => TRUE,
                                            format_types => [C_FORMAT_TYPE_2],
                                            return_hash_keys => ['containment_version'],
                                            },
                GET_COMPRESSION_STATS => {format_expected => TRUE,
                                          format_types => [C_FORMAT_TYPE_2],
                                          return_hash_keys => ['compression_stats'],
                                          },
		BUFTREE_CHECK => {format_expected => TRUE,
				  format_types => [C_FORMAT_TYPE_1],
				  variable_format => TRUE,
				  return_hash_keys => ['buftree_stats'],
				  },
                GET_WAFL_SCAN_SPEED => {format_expected => TRUE,
                                        format_types => [C_FORMAT_TYPE_2],
                                        return_hash_keys => ['wafl_scan_speed'],
                                        },
                GET_SNAPLOCK_CONTAINMENT_BITS => {format_expected => TRUE,
                                                  format_types => [C_FORMAT_TYPE_2],
                                                  return_hash_keys => ['snaplock_containment_stats'],
                                                  },
		VERIFY_SIS_LS => {format_expected => TRUE,
                                  format_types => [C_FORMAT_TYPE_1,],
				  variable_format => TRUE,
                                  return_hash_keys => ['fingerprints',],
                                  },
		GET_VOLUME_STATS => {format_expected => TRUE,
				     format_types => [C_FORMAT_TYPE_2],
				     return_hash_keys => ['vol_stats'],
				     },
		GET_WAFL_STATS => {format_expected => TRUE,
				   format_types => [C_FORMAT_TYPE_2],
				   return_hash_keys => ['wafl_stats'],},
	};

	my $code = $args{code};
	my $format = $base_formats_hash->{$code};
        
        if (!defined($format))
        {
                $format = { format_expected => FALSE };
                logcomment("Code \"$code\": Command output is not expected to be formatted.") if ($class->{_debug});
        }
	logcomment("Formats: _Cmode_base_formats returning: Code=$code, option hash:\n" . Dumper($format)) if ($class->{_debug});

	return $format;
}


sub _Cmode_formats_8_1
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        logcomment("Formats: _Cmode_formats_8_1:\n". Dumper(%args) ) if ($class->{_debug});
        
	my $formats_hash_8_1 = {
                
	};
	my $code = $args{code};
        
        return $formats_hash_8_1->{$code};
}


1;
