#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/CLI/Formats/Type1.pm#1 $
#

## @summary Utility module for parsing formatted CLI output of type 1.
##
## @description
## This module provides functionality for parsing of CLI command output
## that is expected to be in a 'table-like' formatted form.
## For example, the output of the 'df -S' command:
## Filesystem                used       compressed         a-sis     %saved
## /vol/test_vol/        401160KB        2750968KB           0KB        87%
## /vol/base_vol/          1160KB              0KB           0KB         0%
##
## @synopsis
## N/A: To be only used internally by the Compression library.
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)



package Compression::UI::CLI::Formats::Type1;

use strict;
use warnings;

use Tharn;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Compression::Constants qw(:all);
use Compression::UI::Constants;


BEGIN
{
	require Exporter;
	our @ISA = qw(Exporter);
}


## @name ::new
## @summary Compression::UI::CLI::Formats Constructor.
##
## @return Object of type Compression::UI::CLI::Formats.
##
## @arg version Required: Compression::Version object for the hosting filer.
## 
## @arg boot_mode Optional: Boot mode of the hosting filer.
## Defaults to 7M.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::UI::CLI::Formats->new( version => $version );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub new
{
	my $class = shift;
        my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug  => {  
			type 	 => SCALAR,
			optional => 1,
                        default  => FALSE,
			},
	} );
        
	my $self = {};
        
        $self->{_debug} = $args{debug};
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};

	logcomment("Created new Formats Type1 object:\n" . Dumper($self)) if ($self->{_debug});
	bless $self, $class;

	return $self;
}


sub _process_output
{
    my $class = shift;
	my %args = validate(@_,
	{
                code => { type => SCALAR, },
                output => { type => SCALAR, },
                allow_multiple_formats => { type => BOOLEAN, },
                modifier => {type => SCALAR|UNDEF, optional => 1,},
                required_return_hash_key => {type => SCALAR|UNDEF, optional => 1,},
                disable_data_identifier_processing => {type => BOOLEAN, optional => 1, default => FALSE},
                variable_format => {type => SCALAR, optional => 1, default => FALSE},
	} );
        my $code = $args{code};
        my $output = $args{output};
        my $allow_multiple_formats = $args{allow_multiple_formats};
        my $variable_format = $args{variable_format};
        # This is used for sanity checks. If passed in,
        # we will check that the return hash contains this key.
        # This check is useful for commands for a specific entity (for example, volume),
        # when we want to make sure the output was for that specific entity
        # For this format type, it is expected to be found as (part of) the first entry in the output line
        my $required_return_key = $args{required_return_hash_key};
        my $modifier = $args{modifier};
        my $disable_data_identifier_processing = $args{disable_data_identifier_processing};
	my $data_index = 1;
        my $found_required_return_key = 1;
        if (defined($required_return_key))
        {
                $found_required_return_key = 0;
        }

        my %return_hash = (rc => C_RC_FAILURE);
        # If variable format is allowed, then we will do our best-effort to parse the output,
        # but even if parsing fails, we will still return success. The upper-level subroutine should handle the failure
        if ($variable_format == TRUE)
        {
                $return_hash{rc} = C_RC_SUCCESS;
        }
        
        my $format = $class->_get_format_info(code => $code);
        logcomment("Formats Type 1:\n" .
                   "Code $code. Processing output:\n$output\n" .
                   "Allowed multiple formats: $allow_multiple_formats\n" .
                   "Expected format info\n:" . Dumper($format)) if ($class->{_debug});

        # Look for header first.
        my $header = $format->{header};
        if (!defined($header))
        {
                logresult('script', "Formats Type 1: header not defined for code $code.");
        }
        logcomment("Formats Type 1: looking for header:$header\n") if ($class->{_debug});

        my @lines = split(/\n/, $output);
        logcomment("Formats Type 1: Starting processing output:\n" . Dumper(@lines)) if ($class->{_debug});
        
        if ($header ne '')
        {
                my $header_line = shift(@lines);
                if ($header_line !~ /^$header$/i)
                {
                        logcomment("Formats Type1: code $code. Did not find expected header: $header") if ($class->{_debug});
                        # Exit with failure, to allow detecting of other eventual error messages.
                        $return_hash{unprocessed_output} = $output;
                        return %return_hash;
                }
        }
        elsif ($class->{_debug})
        {
                logcomment("Formats Type1: code $code. Output is not expected to contain a header. Starting to process data lines");
        }

        # We already extracted the header. What is left is only data. Start procesing, line by line.
        # Each line will result in a sub-hash, with the key defined by the first data field in the line.
        # The rest of the data fields in the line will become the values of the sub-hash,
        # where the sub-hash keys are defined by the 'return_data_keys' array in the format definition for the specific release.
        my $data_delimiter = $format->{data_delimiter} || '\s+';
        my $data_identifier = $format->{data_identifier};
        my $output_hash_keys = $format->{return_data_keys};
        my $data_patterns_to_discard = $format->{data_patterns_to_discard};
        my $required_data_line_layout = $format->{required_data_line_layout};
        my $ignore_lines = $format->{ignore_lines};
        my $duplicate_entries_allowed = $format->{duplicate_entries_allowed};
        my $attach_constant_keys_if_success = $format->{attach_constant_return_keys};
	my $single_data_field_allowed = (exists($format->{single_data_field_allowed})) ? $format->{single_data_field_allowed} : FALSE;
        my $duplicate_entries_indexes = {};
        my @unprocessed_lines = @lines;
        
        logcomment("Formats Type 1: starting processing data lines:\n" . Dumper(@lines)) if ($class->{_debug});
        foreach my $data_line ( @lines )
        {
                logcomment("Formats Type 1: processing data line: \"$data_line\"") if ($class->{_debug});
                
                if ((defined($required_data_line_layout)) && ($data_line !~ /^$required_data_line_layout$/))
                {
                        if (($allow_multiple_formats == FALSE) && ($variable_format == FALSE))
                        {
                                logresult('fatal', "Unexpected output format for code $code: \n" .
                                          "Multiple formats are not allowed. Expected format of type 1.\n" .
                                          "Data line does not have the required layout: \"$required_data_line_layout\"\n" .
                                          "Line with error: \"$data_line\"\n\n" .
                                          "Output:\n$output");
                        }
                        else
                        {
                                # Check if line should be ignored
                                if (defined($ignore_lines))
                                {
                                        logcomment("Formats Type 1: Multiple or variable formats allowed.\n" .
                                                   "Data line does not have the required layout: \"$required_data_line_layout\"\n" .
                                                   "Line processed:\"$data_line\"\n" .
                                                   "Checking if the line should be ignored:\n" . Dumper($ignore_lines)) if ($class->{_debug});
                                        my $line_should_be_ignored = FALSE;
                                        foreach my $line_to_ignore (@$ignore_lines)
                                        {
                                                if ($data_line =~ /$line_to_ignore/)
                                                {
                                                        $line_should_be_ignored = TRUE;
                                                        last;
                                                }
                                        }
                                        
                                        if ($line_should_be_ignored == TRUE)
                                        {
                                                shift(@unprocessed_lines);
                                                next;
                                        }
                                }
                                
                                # Hit the first line that is not of type 1 format
                                logcomment("Formats Type 1: Multiple or variable formats allowed.\n" .
                                           "Data line does not have the required layout: \"$required_data_line_layout\"\n" .
                                           "Skipping processing.\n" .
                                           "Line with error:\n $data_line") if ($class->{_debug}); 
                                last;
                        }
                }
                
                # Discard any required patterns
                if (defined $data_patterns_to_discard)
                {
                        foreach my $discard_pattern ( @$data_patterns_to_discard )
                        {
                                logcomment("Formats: discarding pattern \"$discard_pattern\" from data line") if ($class->{_debug});
                                $data_line =~ s/$discard_pattern//g;
                        }
                }
                
                my @data_fields = split(/$data_delimiter/, $data_line);
                logcomment("Formats Type 1: processing individual data fields for this line:\n" . Dumper(@data_fields)) if ($class->{_debug});
                # First data field should match the data identifier
                # This will become the hash key for this data line
                my $data_key = shift(@data_fields);
                if (!defined($data_key))
                {
                        if ($variable_format == TRUE)
                        {
                                # Hit the first line that is not of type 1 format
                                logcomment("Formats Type 1: Variable formats allowed.\n" .
                                           "Header detected, but no data found: skipping processing.\n" .
                                           "Line not in format type 1:\n $data_line") if ($class->{_debug}); 
                                last;
                        }
                        logresult('fatal', "Formats: unexpected output format. Header detected, but no data found: $output");
                }
                logcomment("Formats: processing first data field: $data_key") if ($class->{_debug});
                if (defined($data_identifier) && ($disable_data_identifier_processing == FALSE))
                {
                        if ($data_key =~ /^$data_identifier/)
                        {
                                $data_key =~ s/^$data_identifier//;
                                # Remove any eventual known tailing patterns
                                $data_key =~ s/\/$//;

                                logcomment("Formats Type 1: finished processing first data field: $data_key") if ($class->{_debug});
                        }
                        else
                        {
                                if (($allow_multiple_formats == FALSE) && ($variable_format == FALSE))
                                {
                                        logresult('fatal', "Unexpected output format for code $code: \n" .
                                                           "Multiple formats are not allowed. Expected format of type 1.\n" .
                                                           "Failed to match data key \"$data_key\" with data identifier pattern: \"$data_identifier\"\n" .
                                                           "Line with error: \"$data_line\"\n\n" .
                                                           "Output:\n$output");
                                }
                                else
                                {
                                        # Hit the first line that is not of type 1 format
                                        logcomment("Formats Type 1: Multiple formats allowed. Skipping processing.\n" .
                                                   "Failed to match data key \"$data_key\" with data identifier pattern: \"$data_identifier\"\n" .
                                                   "Line with error:\n $data_line") if ($class->{_debug}); 
                                        last;
                                }
                        }
                }
		
		if (($data_key eq '') && ($single_data_field_allowed == TRUE))
		{
			$data_key = "$data_index";
			$data_index++;
		}
                
                # Now check that the hash key is what we expected, if required
                if ((defined($required_return_key)) &&
                    ("$data_key" ne "$required_return_key"))
                {
                        # Not interested in this entry
                        # But do not consider it unprocessed, we are still within the legal format type
                        logcomment("Formats Type 1: Skipping data line for (not required entry) data key \"$data_key\"\n") if ($class->{_debug}); 
                        shift(@unprocessed_lines);
                        next;
                }
                $found_required_return_key = 1;
                
                # Discard any data modifiers, if required
                if (defined($modifier))
                {
                        # Look for modifier changes in command output
                        my $output_data_modifier = $format->{modifier_changes}->{$modifier};
                        if (!defined($output_data_modifier))
                        {
                                $output_data_modifier = $modifier;
                        }
                        
                        logcomment("Formats Type 1: discarding output data modifier: $output_data_modifier") if ($class->{_debug});
                        foreach my $data_field ( @data_fields )
                        {
                                $data_field =~ s/$output_data_modifier$//;
                        }
                }

                # Found the (if) requested return hash key.
                # Already extracted the hash key for this data line.
                # What is left are the values to be added to the return hash.
                if (defined($output_hash_keys))
                {
                        logcomment("Formats Type 1: starting formatting return hash with keys:\n" . Dumper($output_hash_keys)) if ($class->{_debug});
                        my $next_index_if_duplicate_entries = 0;
                        if ((defined($duplicate_entries_allowed)) && ($duplicate_entries_allowed == TRUE))
                        {
                                if (defined($duplicate_entries_indexes->{$data_key}))
                                {
                                        $next_index_if_duplicate_entries = $duplicate_entries_indexes->{$data_key};
                                }
                                
                                $duplicate_entries_indexes->{$data_key} = $next_index_if_duplicate_entries + 1;
                        }
                        
                        foreach my $output_key ( @$output_hash_keys )
                        {
				logcomment("Formats Type 1: looking for value for return key $output_key in:\n" . Dumper(@data_fields)) if ($class->{_debug});
                                my $output_value = shift(@data_fields);
                                chomp($output_value);
                                if (!defined($output_value))
                                {
                                        logresult('fatal', "Unexpected output format for code $code: \n" .
                                                           "Expected format of type 1.\n" .
                                                           "Did not find value that corresponds to: $output_key\n" .
                                                           "Line with error: $data_line\n\n" .
                                                           "Output:\n$output");
                                }
                                logcomment("Formats Type 1: adding $output_key ($output_value) to return hash") if ($class->{_debug});
                                
                                if ((defined($duplicate_entries_allowed)) && ($duplicate_entries_allowed == TRUE))
                                {
                                        $return_hash{$data_key}{$next_index_if_duplicate_entries}{$output_key} = $output_value;
                                }
                                else
                                {
                                        $return_hash{$data_key}{$output_key} = $output_value;
                                }
                        }
                }
                else
                {
                        # Add only the first value left in this line (no extra hash nesting)
                        # Don't handle duplicate entries in this case
			# Remove any tailing spaces
			my $data_value = shift(@data_fields);
			$data_value =~ s/\s*$//;
                        $return_hash{$data_key} = $data_value;
                }
                
                if (defined($attach_constant_keys_if_success))
                {
                        my $constants = Compression::UI::Constants->new(version => $class->{_version},
                                                                        boot_mode => $class->{_boot_mode},
                                                                        debug => $class->{_debug},);
                        foreach my $key (keys(%$attach_constant_keys_if_success))
                        {
                                $return_hash{$data_key}{$key} = $constants->get_constant_value(constant_code => $attach_constant_keys_if_success->{$key});
                        }
                }
                
                shift(@unprocessed_lines);
        }

        if ($found_required_return_key == 0)
        {
                logresult('fatal', "Error in processing output of type 1 for code $code:\n" .
                                   "Did not find line to correspond to required entity \"$required_return_key\"\n" .
                                   "Output:\n$output\n");
        }

        $return_hash{rc} = C_RC_SUCCESS;
        $return_hash{unprocessed_output} = join("\n", @unprocessed_lines); 

        logcomment("Formats Type 1: _process_output returning:\n" . Dumper(%return_hash)) if ($class->{_debug});
        
        return %return_hash;
}


sub _get_format_info
{
        my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $code = $args{code};
        my $return_hash = {};
        
        if ($class->{_boot_mode} eq C_ONTAP_7MODE)
        {
                if ($ontap_ver eq '7.3')
		{
			$return_hash = $class->_7mode_formats_7_3( code => $code );
                }
		elsif ($ontap_ver eq '8.0')
		{
			$return_hash = $class->_7mode_formats_8_0( code => $code );
                }
		elsif ($ontap_ver eq '8.1')
		{
			$return_hash = $class->_7mode_formats_8_1( code => $code );
                }
		elsif ($ontap_ver eq '8.2')
		{
			$return_hash = $class->_7mode_formats_8_2( code => $code );
                }
		else
		{
			logcomment("Formats Type 1: Specific support for sw_version $ontap_ver not found. Will be using base formats") if ($class->{_debug});
                }
                
                if ( !defined($return_hash) || !(scalar %$return_hash) )
                {
                        logcomment( "Formats Type 1: Could not find format definition specific for $ontap_ver. Checking base formats.") if ($class->{_debug});
                        $return_hash =  $class->_7mode_base_formats( code => $code );
                }
        }
        elsif ($class->{_boot_mode} eq C_ONTAP_CMODE)
        {
                if ($ontap_ver eq '8.1')
		{
			$return_hash = $class->_Cmode_formats_8_1( code => $code );
                }
		else
		{
			logcomment("Formats Type 1: Specific support for C-mode sw_version $ontap_ver not found. Will be using base formats") if ($class->{_debug});
                }
                
                if ( !defined($return_hash) || !(scalar %$return_hash) )
                {
                        logcomment( "Formats Type 1: Could not find format definition specific for $ontap_ver. Checking base formats.") if ($class->{_debug});
                        $return_hash =  $class->_Cmode_base_formats( code => $code );
                }
        }

        return $return_hash;
}



################# 7-mode #############
sub _7mode_base_formats
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );

	logcomment("Formats Type 1: _7mode_base_formats input:\n". Dumper(%args) ) if( $class->{_debug} );

	my $base_formats_hash = {
                GET_SPACE_SAVINGS => {header => 'Filesystem\s+used\s+total-saved\s+%total-saved\s+deduplicated\s+%deduplicated\s+compressed\s+%compressed',
                                      data_delimiter => '\s+',
                                      data_identifier => '/vol/',
                                      return_data_keys => [ 'used', 'total_saved', 'percentage_saved_total', 'deduplicated', 'percentage_deduplicated', 'compressed', 'percentage_compressed' ],
                                      data_patterns_to_discard => [ '%' ],
                                      modifier_changes => {'k' => 'KB', 'm' => 'MB', 'g' => 'GB', 't' => 'TB' },
                                      },
                
                DECOMPRESS_STATUS => {header => 'Filesystem\s+Progress',
                                      data_delimiter => '\s+',
                                      data_identifier => '/vol/',
                                      data_patterns_to_discard => [ '%' ],
                                      },
                
                GET_VBN_INFO => {header => '',
                                 required_data_line_layout => 'fbn\s+\d+\s+is\s+vbn\s+\d+,\s+vvbn\s+\d+.*',
                                 data_patterns_to_discard => ['fbn\s+', 'is vbn', ', vvbn'],
                                 data_delimiter  => '\s+',
                                 return_data_keys => ['vbn', 'vvbn'],
                                 },
		FILE_FOLDING_STATUS => {header => 'Volume \w+:',
					required_data_line_layout => '\s*\d+\s+.*',
					ignore_lines => ['\s*Scan id\s+',
							 ],
					data_delimiter => '\s\s+',
					data_patterns_to_discard => ['^\s*',],
					return_data_keys => [ 'scan_type', 'progress' ],
					},
		VOL_MOVE_STATUS => {header => 'Source\s+Destination\s+CO Attempts\s+CO Time\s+State',
				    data_delimiter => '\s+',
				    return_data_keys => ['dest_aggr', 'cutover_attempts', 'cutover_time', 'state'],
				    },
		VFILER_MIGRATE_STATUS => {header => 'Storage Path\s+Status',
				    data_delimiter => '\s+',
				    return_data_keys => ['status',],
				    },
	};

	my $code = $args{code};
	my $format = $base_formats_hash->{$code};
        
        if (!defined($format))
        {
                $format = { format_expected => FALSE };
                logresult('script', "Library problem found: Processing required for format type 1, but no format info defined. Command code: $code");
        }

	return $format;
}

sub _7mode_formats_7_3
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
	my $formats_hash_7_3 = {
                GET_SPACE_SAVINGS => {header          => 'Filesystem\s+used\s+compressed\s+a-sis\s+%saved',
                                      data_delimiter  => '\s+',
                                      data_identifier => '/vol/',
                                      return_data_keys => [ 'used', 'compressed', 'deduplicated', 'percentage_saved_total' ],
                                      data_patterns_to_discard => [ '%' ],
                                      modifier_changes => {'k' => 'KB', 'm' => 'MB', 'g' => 'GB', 't' => 'TB' },
                                      },
        };
	my $code = $args{code};
        
        return $formats_hash_7_3->{$code}
}

sub _7mode_formats_8_0
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
	my $formats_hash_BR_0 = {
		COMPRESS_STATUS => {header => 'Path\s+Data scanned\(KB\)\s+Snapshot skipped data\(KB\)\s+Shared skipped data\(KB\)\s+Rate\(KB\/s\)',
				    data_delimiter  => '\s+',
				    data_identifier => '/vol/',
				    return_data_keys => [ 'data_scanned', 'snapshot_skipped', 'shared_skipped_block', 'rate'],
                                    attach_constant_return_keys => {stage => 'C_STAGE_COMPRESSING',
                                                                    status => 'C_STATUS_ACTIVE',
                                                                    },
                                    },
                
                COMPRESS_STATUS_CHECKPOINT => {header => 'Path\s+Data scanned\(KB\)\s+Snapshot skipped data\(KB\)\s+Shared skipped data\(KB\)\s+Rate\(KB\/s\)',
                                               data_delimiter  => '\s+',
                                               data_identifier => '/vol/',
                                               return_data_keys => [ 'data_scanned', 'snapshot_skipped', 'shared_skipped_block', 'rate'],
                                               },
	};
	my $code = $args{code};
        
        return $formats_hash_BR_0->{$code}
}

sub _7mode_formats_8_1
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
	my $formats_hash_8_1 = {
                VERIFY_COMPR_ATTEMPTED_BIT => {header => 'Checking\s(fingerprint|fingerprint\.vvol)\s+\.+\s+\d+\s+records',
                                               required_data_line_layout => '\s*\w+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+',
                                               data_delimiter  => '\s+',
                                               data_patterns_to_discard => [ '^\s+' ],
                                               return_data_keys => [ 'inode', 'fbn', 'generation_count', 'cp_count', 'compression_attempted_bit'],
                                               duplicate_entries_allowed => TRUE,
                                               },
                _CHANGELOG_FINGERPRINTS => {header => '\s*contents of bucket: \d+ \(number of entries: \d+\)',
                                            required_data_line_layout => '\s*\w+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+',
                                            data_delimiter  => '\s+',
                                            data_patterns_to_discard => [ '^\s+' ],
                                            return_data_keys => [ 'inode', 'fbn', 'generation_count', 'cp_count', 'compression_attempted_bit'],
                                            duplicate_entries_allowed => TRUE,
                                            },
		VERIFY_SIS_LS => {header => 'SIS metadata for flexible volume \S+ in aggregate \S+:\s+',
                                  required_data_line_layout => '^fingerprint.+',
                                  ignore_lines => ['.*'],
                                  data_delimiter  => '\s+',
                                  data_patterns_to_discard => [ '^\s+' ],
                                  return_data_keys => [ 'size', 'blocks_used'],
                                  },
	};
	my $code = $args{code};
        
        return $formats_hash_8_1->{$code}
}

sub _7mode_formats_8_2
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
	my $formats_hash = {
                VERIFY_COMPR_ATTEMPTED_BIT => {header => 'Checking\s(fingerprint|fingerprint\.vvol|fingerprint\.vvol\.delta|duplicate\.p1|duplicate\.p2)\s+\.+\s+\d+\s+records',
					       required_data_line_layout => '\s*\w+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+',
					       ignore_lines => ['Number of Segments:',
								'Number of Records:',
								'Contents of Segment:',
								'Checking\s(duplicate\.p1|duplicate\.p2)\s+\.+\s+\d+\s+records',
								'^\s+\d+\s+\d+\s+\d+\s+\d+$',
								'^\s*\w+\s+\d+\s+\d+$'],
					       data_delimiter  => '\s+',
					       data_patterns_to_discard => [ '^\s+' ],
					       return_data_keys => [ 'inode', 'fbn', 'vvbn_info', 'generation_count', 'cp_count', 'compression_attempted_bit', 'unprocessed_bit'],
					       duplicate_entries_allowed => TRUE,
					       },
                _CHANGELOG_FINGERPRINTS => {header => '\s*contents of bucket: \d+ \(number of entries: \d+\)',
                                            required_data_line_layout => '\s*\w+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+',
                                            data_delimiter  => '\s+',
                                            data_patterns_to_discard => [ '^\s+' ],
                                            return_data_keys => [ 'inode', 'fbn', 'vvbn_info', 'generation_count', 'cp_count', 'compression_attempted_bit', 'unprocessed_bit'],
                                            duplicate_entries_allowed => TRUE,
                                            },
		VERIFY_SIS_LS => {header => 'SIS metadata for flexible volume \S+ in aggregate \S+:\s+',
                                  required_data_line_layout => '^fingerprint.+',
                                  ignore_lines => ['.*'],
                                  data_delimiter  => '\s+',
                                  data_patterns_to_discard => [ '^\s+' ],
                                  return_data_keys => [ 'size', 'blocks_used'],
                                  },
	};
	my $code = $args{code};
        
        return $formats_hash->{$code}
}


########### C-mode ####################
sub _Cmode_base_formats
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );

	logcomment("Formats Type 1: _Cmode_base_formats input:\n". Dumper(%args) ) if( $class->{_debug} );

	my $base_formats_hash = {
                GET_SPACE_SAVINGS => {header => 'Filesystem\s+used\s+total-saved\s+%total-saved\s+deduplicated\s+%deduplicated\s+compressed\s+%compressed',
                                      data_delimiter  => '\s+',
                                      data_identifier => '/vol/',
                                      return_data_keys => [ 'used', 'total_saved', 'percentage_saved_total', 'deduplicated', 'percentage_deduplicated', 'compressed', 'percentage_compressed' ],
                                      data_patterns_to_discard => [ '%' ],
                                      modifier_changes => {'k' => 'KB', 'm' => 'MB', 'g' => 'GB', 't' => 'TB' },
                                      },
                VERIFY_COMPR_ATTEMPTED_BIT => {header => 'Checking\s(fingerprint|fingerprint\.vvol)\s+\.+\s+\d+\s+records',
                                               required_data_line_layout => '\s*\w+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+',
                                               ignore_lines => ['Number of Segments:',
                                                                'Number of Records:',
                                                                'Contents of Segment:',
								'^\s*\w+\s+\d+\s+\d+$'],
                                               data_delimiter  => '\s+',
                                               data_patterns_to_discard => [ '^\s+' ],
                                               return_data_keys => ['inode', 'fbn', 'vvbn_info', 'generation_count', 'cp_count', 'compression_attempted_bit', 'unprocessed_bit'],
                                               duplicate_entries_allowed => TRUE,
                                               },
                _CHANGELOG_FINGERPRINTS => {header => '\s*contents of bucket: \d+ \(number of entries: \d+\)',
                                            required_data_line_layout => '\s*\w+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+',
                                            data_delimiter  => '\s+',
                                            data_patterns_to_discard => [ '^\s+' ],
                                            return_data_keys => ['inode', 'fbn', 'vvbn_info', 'generation_count', 'cp_count', 'compression_attempted_bit', 'unprocessed_bit'],
                                            duplicate_entries_allowed => TRUE,
                                            },
                GET_VBN_INFO => {header => '',
                                 required_data_line_layout => 'fbn\s+\d+\s+is\s+vbn\s+\d+,\s+vvbn\s+\d+\s*',
                                 data_patterns_to_discard => ['fbn\s+', 'is vbn', ', vvbn', '\s*$'],
                                 data_delimiter  => '\s+',
                                 return_data_keys => ['vbn', 'vvbn'],
                                 },
		DISK_FREE => {header => 'Filesystem\s+total\s+used\s+avail\s+capacity\s+Mounted on',
			      data_delimiter => '\s+',
			      data_identifier => '/vol/',
			      return_data_keys => [ 'total', 'used', 'avail', 'capacity', 'mounted_on' ],
			      data_patterns_to_discard => [ '%', 'KB' ],
			      },
		GET_RESERVED => {header => 'Filesystem\s+total\s+used\s+avail\s+reserved\s+Mounted on',
				 data_delimiter => '\s+',
				 data_identifier => '/vol/',
				 return_data_keys => [ 'total', 'used', 'avail', 'reserved', 'mounted_on' ],
				 data_patterns_to_discard => [ '%', 'KB', '\(', '\)', ],
			      },
		AGGR_SNAP_LIST => {header => 'Aggregate.+',
				   required_data_line_layout => '\s*\d+\%\s+\(\s*\d+\%\)\s+\d+\%\s+\(\s*\d+\%\)\s+\w+\s+\d+\s+\d+\:\d+\s*.+',
				   ignore_lines => ['working',
                                                    '\s*\%\/used\s+\%\/total\s+date\s+name\s*',
                                                    '---',
                                                    ],
				   data_delimiter => '\s*\d+\%\s+\(\s*\d+\%\)\s+\d+\%\s+\(\s*\d+\%\)\s+\w+\s+\d+\s+\d+\:\d+\s*',
				   single_data_field_allowed => TRUE,
                                   },
		FILE_FOLDING_STATUS => {header => 'Volume \w+:',
					required_data_line_layout => '\s*\d+\s+.*',
					ignore_lines => ['\s*Scan id\s+',
							 ],
					data_delimiter => '\s\s+',
					data_patterns_to_discard => ['^\s*',],
					return_data_keys => [ 'scan_type', 'progress' ],
					},
                GET_REFCOUNT => {header => '',
                                 required_data_line_layout => 'fbn\s+\d+\s+is\s+vbn\s+\d+,\s+vvbn\s+\d+,\s+refcount\s+\d+',
                                 data_patterns_to_discard => ['fbn\s+', 'is\s+vbn\s+\d+,\s+vvbn\s+\d+,\s+refcount\s+'],
                                 data_delimiter  => '\s+',
                                 return_data_keys => ['refcount'],
                                 },
		BUFTREE_CHECK => {header => '',
				  data_delimiter  => '\s+',
				  data_patterns_to_discard => [ '^\s+' ],
				  required_data_line_layout => '\s*\S+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s*',
				  return_data_keys => ['blks32', 'holes32', 'blks64', 'holes64', 'total'],
				  ignore_lines => ['-+\s+-+\s+-+\s+-+\s+-+\s+-+',
						   'level\s+32-bit\s+blks\s+32-bit\s+holes\s+64-bit\s+blks\s+64-bit\s+holes\s+total',
						   'wafl buftree check: Suspending until', 
						   ],
                                 },
		VERIFY_SIS_LS => {header => 'SIS metadata for flexible volume \S+ in aggregate \S+:\s+',
                                  required_data_line_layout => '^fingerprint.+',
                                  ignore_lines => ['.*'],
                                  data_delimiter  => '\s+',
                                  data_patterns_to_discard => [ '^\s+' ],
                                  return_data_keys => [ 'size', 'blocks_used'],
                                  },
	};

	my $code = $args{code};
	my $format = $base_formats_hash->{$code};
        
        if (!defined($format))
        {
                $format = { format_expected => FALSE };
                logresult('script', "Library problem found: Processing required for format type 1, but no format info defined. Command code: $code");
        }

	return $format;
}


sub _Cmode_formats_8_1
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
	my $formats_hash_8_1 = {
		GET_VBN_INFO => {header => '',
				 required_data_line_layout => 'fbn\s+\d+\s+is\s+vbn\s+\d+,\s+vvbn\s+\d+',
				 data_patterns_to_discard => ['fbn\s+', 'is vbn', ', vvbn'],
				 data_delimiter  => '\s+',
				 return_data_keys => ['vbn', 'vvbn'],
				 },
		VERIFY_COMPR_ATTEMPTED_BIT => {header => 'Checking\s(fingerprint|fingerprint\.vvol)\s+\.+\s+\d+\s+records',
                                               required_data_line_layout => '\s*\w+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+',
                                               ignore_lines => ['Number of Segments:',
                                                                'Number of Records:',
                                                                'Contents of Segment:'],
                                               data_delimiter  => '\s+',
                                               data_patterns_to_discard => [ '^\s+' ],
                                               return_data_keys => [ 'inode', 'fbn', 'generation_count', 'cp_count', 'compression_attempted_bit'],
                                               duplicate_entries_allowed => TRUE,
                                               },
                _CHANGELOG_FINGERPRINTS => {header => '\s*contents of bucket: \d+ \(number of entries: \d+\)',
                                            required_data_line_layout => '\s*\w+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+',
                                            data_delimiter  => '\s+',
                                            data_patterns_to_discard => [ '^\s+' ],
                                            return_data_keys => [ 'inode', 'fbn', 'generation_count', 'cp_count', 'compression_attempted_bit'],
                                            duplicate_entries_allowed => TRUE,
                                            },
	};
	my $code = $args{code};
        
        return $formats_hash_8_1->{$code}
}


1;
