#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/CLI/Messages.pm#1 $
#

## @summary Utility module to allow output message interpretation.
##
## @description
##    This module enables interpretation of non-formatted output message after command execution.
##
##
## @synopsis
##
## use Compression::Constants;
## use Compression::UI::CLI::Messages;
## use Compression::Version;
## use NOTE::Filer;  # To obtain filer version.
##
## my $filer = NOTE::Filer->connect( host => $FILER );
## my $version = Compression::Version->new( conn => $filer );
## my $obj = Compression::UI::CLI::Messages->new( version => $version );
##
## @examples 
## $obj->get_return_code(output => $output,
##                       code => C_CMD_DECOMPRESS_START);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::UI::CLI::Messages;

use strict;
use warnings;

use Tharn;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;


use Compression::Constants qw(:all);
use Compression::Version;
use Compression::UI::CLI::Messages::Constants_7_3;
use Compression::UI::CLI::Messages::Constants_BR_0;
use Compression::UI::CLI::Messages::Constants_8_1;
use Compression::UI::CLI::Messages::Constants_8_2;


BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}


## @name ::new
##
## @summary Compression::UI::CLI::Messages Constructor.
##
## @return Object of type Compression::UI::CLI::Messages.
##
## @arg version Required: Compression::Version object for the hosting filer.
## 
## @arg boot_mode Optional: Boot mode of the hosting filer.
## Defaults to 7M.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::UI::CLI::Messages->new( version => $version );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##

sub new
{
	my $class = shift;
        my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug  => {  
			type 	 => SCALAR,
			optional => 1,
                        default  => FALSE,
			},
	} );
        
	my $self = {};
        $self->{_debug} = $args{debug};
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};
        
        bless $self, $class;
        
        $self->_refresh_constants();

	logcomment("Massages: Created new object:\n" . Dumper($self)) if ($self->{_debug});
	

	return $self;
}


## @name ::get_return_code
## 
## @description
## Parses the given string and returns the corresponding return code.
##
## @return
## Return code corresponding to the parsed string, if a match is found and corresponds to the given command code.
## If a match is found, but corresponds to a different command code, the return code will be C_RC_FAILURE.
## If no match is found, the return code will be C_RC_FAILURE.
## 
## @arg output Required: Command output.
## 
## @arg code Required: Code corresponding to the command for which the output was obtained.
## It is used to look-up known output messages for the specific command.
## If a match is found within outputs corresponding to the specific command,
## the corresponding return code is returned, including C_RC_SUCCESS.
## If not, all other known messages are searched, but an eventual C_RC_SUCCESS code will be changed into C_RC_FAILURE.
## 
## @examples
## $obj->get_return_code(output => $output,
##                       code => C_CMD_DECOMPRESS_START);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##

sub get_return_code
{
	
	my $class 	= shift;
	my (%args) = validate( @_,
	{
		output => { type => SCALAR|UNDEF, },
		code => { type => SCALAR, },
	} );
        
        # TODO? : plugin variable substitution into message (ex: volume name)
        my $output = $args{output};
        my $code = $args{code};
	
	my $constants = $class->{_constants}->{messages};
        my $search_result;
        
        # First try to find a match through messages corresponding to this command code
        my $command_messages = $constants->{$code};
        
        # Handle any empty output
        my $legal_empty_output_rc = $class->_check_for_allowed_empty_output(output => $output,
                                                                            search_set => $command_messages,
                                                                            code => $code);
        if ($legal_empty_output_rc != C_RC_FAILURE)
        {
                return $legal_empty_output_rc;
        }
        
        # Passed-in code may not have an entry in the constants hash
        if (defined($command_messages))
        {
                logcomment("Messages:\n Searching for \"$output\"\n through messages for passed-in command code $code") if $class->{_debug};
                $search_result = $class->_search_for_code_specific_message(output => $output,
                                                                           search_set => $command_messages,
                                                                           );
                if ($search_result->{match_found} == TRUE)
                {
                        return $search_result->{rc};
                }
        }
        
        # Loop through the rest of messages, but:
        # since an eventual 'success' message would correspond to a different command,
        # turn any 'success' into 'failure'
        for my $attempt_code ( keys %$constants )
        {
                if ($attempt_code eq $code)
                {
                        next;
                }
                
                $command_messages = $constants->{$attempt_code};
                logcomment("Messages:\n Searching for \"$output\"\n through messages for command code $attempt_code") if $class->{_debug};
                $search_result = $class->_search_for_code_specific_message(output => $output,
                                                                           search_set => $command_messages,
                                                                           );
                if ($search_result->{match_found} == TRUE)
                {
                        if ($search_result->{rc} == C_RC_SUCCESS)
                        {
                                logcomment("Messages:\n Command output for code $code, command output $output\n" .
                                           "Match found against success-indicating message for command code $attempt_code\n" .
                                           "Changing to failure...") if $class->{_debug};
                                return C_RC_FAILURE;
                        }
                        logcomment("Messages:\n Look-up for code $code, command output: $output\n" .
                                   "Returning " . $search_result->{rc}) if $class->{_debug};
                        return $search_result->{rc};
                }
        }
        
        logcomment("Messages:\n Look-up for code $code, command output: $output\n" .
                   "Did not know how to interpret message. No match found. Failing...") if $class->{_debug};
        return C_RC_FAILURE;
}

############################
# Internal subroutines
############################

sub _check_for_allowed_empty_output
{
        my $class 	= shift;
	my (%args) = validate( @_,
	{
                output => { type => SCALAR|UNDEF, },
		search_set => { type => HASHREF|UNDEF, },
                code => { type => SCALAR, },
	} );
        
        my $search_set = $args{search_set};
        my $output = $args{output};
        my $code = $args{code};
        my $output_should_be_empty = FALSE;
        my $output_is_empty = FALSE;
        
        if (!defined($output) || ($output eq ''))
        {
                logcomment("Messages:\n Detected empty output for command code $code") if $class->{_debug};
                $output_is_empty = TRUE;
        }
        
        if (defined($search_set) && (defined($search_set->{''})))
        {
                $output_should_be_empty = TRUE;
        }
        
        if ($output_is_empty == TRUE)
        {
                if ($output_should_be_empty == TRUE)
                {
                        logcomment("Messages: detected (allowed) undefined or empty output string") if ($class->{_debug});
                        return $search_set->{''};
                }
                else
                {
                        logresult('fatal', "Unexpected empty output for command code $code");
                }
        }
        
        return C_RC_FAILURE;
}



sub _search_for_code_specific_message
{
        my $class 	= shift;
	my (%args) = validate( @_,
	{
                output => { type => SCALAR, },
		search_set => { type => HASHREF, },
	} );
        
        my $search_set = $args{search_set};
        my $output = $args{output};
        
        my $return_hash = { rc => C_RC_FAILURE, match_found => FALSE, };
        
        logcomment("Messages:\n Searching for \"$output\"\n through:\n" . Dumper($search_set)) if $class->{_debug};
        
        for my $message ( keys %$search_set )
        {
                if ($message ne '')
                {
                        my $rc = $search_set->{$message};
                        logcomment("Messages:\n trying: $message\n against: $output ") if $class->{_debug};
                        if ($output =~ /$message/i)
                        {
                                logcomment("Messages: SUCCESS\n matched: $message\n against: $output\nReturning $rc") if $class->{_debug};
                                $return_hash->{rc} = $rc;
                                $return_hash->{match_found} = TRUE;
                                last;
                        }
                }
        }
        
        return $return_hash;
}

sub _refresh_constants
{
        my $class = shift;
        my $ontap_ver = $class->{_version}->{ontap};
        
        if ($ontap_ver eq '7.3')
        {
                $class->{_constants} = Compression::UI::CLI::Messages::Constants_7_3->new(debug => $class->{_debug},
                                                                                          boot_mode => $class->{_boot_mode});
	}
        elsif ($ontap_ver eq  '8.0')
        {
                $class->{_constants} = Compression::UI::CLI::Messages::Constants_BR_0->new(debug => $class->{_debug},
                                                                                           boot_mode => $class->{_boot_mode});
        }
        elsif ($ontap_ver eq  '8.1')
        {
                $class->{_constants} = Compression::UI::CLI::Messages::Constants_8_1->new(debug => $class->{_debug},
                                                                                          boot_mode => $class->{_boot_mode});
	}
	elsif ($ontap_ver eq  '8.2')
	{
		$class->{_constants} = Compression::UI::CLI::Messages::Constants_8_2->new(debug => $class->{_debug},
											  boot_mode => $class->{_boot_mode});
	}
        else
        {
                # WARNING: this should always point to the latest 'specifically-supported' version (latest case above)
                logcomment("\n\n\nWARNING!!!\n" .
                           "Detected ontap version \"$ontap_ver\"\n" .
                           "Using the default messages definitions (different ontap version).\n" .
                           "This may result in output matching failures.\n" .
                           "If this is the case, please add correspondent module in Compression/UI/CLI/Messages");
                $class->{_constants} = Compression::UI::CLI::Messages::Constants_8_2->new(debug => $class->{_debug},
                                                                                          boot_mode => $class->{_boot_mode});
	}
}



1;
