#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/CLI/Options.pm#1 $
#

## @summary CLI template for options names and legal value strings.
##
## @description
##    This library keeps track of names and legal value strings for options needed for compression scripts (7G, 7M, 10M).
##
## @synopsis
##
## use Compression::UI::CLI::Options;
## use Compression::CLI;
## use Compression::Version;
## use NOTE::Filer;  # To obtain filer version.
##
## my $filer = NOTE::Filer->connect( host => $FILER );
## my $version = Compression::Version->new( conn => $filer );
## my $obj = Compression::UI::CLI::Options->new( version => $version );
##
## @examples 
## $obj->get_option_definition(option_code => C_OPTION_COMPRESSION);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)



package Compression::UI::CLI::Options;


use strict;
use warnings;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Tharn;
use Compression::Constants qw(:all);
use Compression::Version;


BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
        
};


## @name ::new
##
## @summary Compression::UI::CLI::Options Constructor.
##
## @return Object of type Compression::UI::CLI::Options.
##
## @arg version Required: Compression::Version object for the hosting filer.
## 
## @arg boot_mode Optional: Boot mode of the hosting filer.
## Defaults to 7M.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::UI::CLI::Options->new( version => $version );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##

sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug  => {  
			type 	 => SCALAR,
			optional => 1,
                        default  => FALSE,
			},
	} );
        
	my $self = {};
        
        $self->{_debug} = $args{debug};
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};
        
        logcomment("Created new Options object:\n" . Dumper($self)) if ($self->{_debug});
	
	bless $self, $class;
	return $self;
}


## @name ::get_option_definition
## 
## @description
## Generates release agnostic/specific option definition, including name and legal values to be set.
##
## @return
## Hash with 'name', 'enabled' value to be set, 'disabled' value to be set.
## 
## @arg option_code Required: Code corresponding to the option for which the definition is needed.
## 
## @examples
## $obj->get_option_definition(option_code => C_OPTION_COMPRESSION);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##

sub get_option_definition
{
        my $class = shift;
	my %args = validate(@_,
	{
                option_code  => { type => SCALAR, },
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $option_code = $args{option_code};
        my $boot_mode = $class->{_boot_mode};
        my $ret_hash = {};
        
        if ($boot_mode eq C_ONTAP_7MODE)
        {
                if ($ontap_ver eq '7.3')
		{
			$ret_hash = $class->_7mode_options_7_3( option_code => $option_code );
                }
		elsif ($ontap_ver eq '8.0')
		{
			$ret_hash = $class->_7mode_options_8_0( option_code => $option_code );
                }
                
                if (!defined($ret_hash) || !(scalar %$ret_hash))
                {
                        logcomment("Options: Specific support for sw_version $ontap_ver not found. Will be using base definitions") if ($class->{_debug});
                        $ret_hash =  $class->_7mode_base_options( option_code => $option_code );
                }
        }
        elsif ($boot_mode eq C_ONTAP_CMODE)
        {
                $ret_hash =  $class->_Cmode_base_options( option_code => $option_code );
        }
        else
        {
                logresult('script', "Unrecognized boot_mode: \"$boot_mode\"");
        }
        
        return $ret_hash;
}


###################################
# Internal subroutines
###################################

sub _7mode_base_options
{
	my $class = shift;
	my %args = validate(@_,
	{
                option_code  => { type => SCALAR, },
	} );

	logcomment("Options: _7mode_base_options: ". Dumper(%args) ) if ($class->{_debug});

	my $base_options_hash =
        {
                GUARANTEE => {name => 'guarantee',
                              enabled => 'volume',
                              disabled => 'none',
                              file => 'file',
                              },
                FRACTIONAL_RESERVE =>{name =>'fractional_reserve',
                                      disabled => 0,
                                      },
                TRY_FIRST =>{name =>'try_first',
                             volume_grow => 'volume_grow',
                             snap_delete => 'snap_delete',
                             },
	};

	my $code = $args{option_code};
	my $option = $base_options_hash->{$code};
        
        if (!defined($option))
        {
                logresult("script", "This library only maintains compression-specific options.\n" .
                                    "Option code \"$code\" is not supported for this release.");
        }
	logcomment("_7mode_base_options: Option code=$code, option hash: " . Dumper($option)) if ($class->{_debug});

	return $option;
}



sub _7mode_options_7_3
{
	my $class = shift;
	my %args = validate(@_,
	{
                option_code  => { type => SCALAR, },
	} );
        
        logcomment("Options: _7mode_options_7_3: ". Dumper(%args) ) if ($class->{_debug});
        
	my $options_hash_7_3 =
        {
                COMPRESSION => {name => 'compression',
                                enabled => 'on',
                                disabled => 'off',
                                },
        };
	my $code = $args{option_code};
        
        return $options_hash_7_3->{$code}
}

sub _7mode_options_8_0
{
        my $class = shift;
	my %args = validate(@_,
	{
                option_code  => { type => SCALAR, },
	} );
        
        logcomment("Options: _7mode_options_8_0: ". Dumper(%args) ) if ($class->{_debug});
        
	my $options_hash_8_0 =
        {
                COMPRESSION => {name => 'compression',
                                enabled => 'on',
                                disabled => 'off',
                                },
        };
	my $code = $args{option_code};
        
        return $options_hash_8_0->{$code}
}

sub _Cmode_base_options
{
	my $class = shift;
	my %args = validate(@_,
	{
                option_code  => { type => SCALAR, },
	} );

	logcomment("Options: _Cmode_base_options: ". Dumper(%args) ) if ($class->{_debug});

	my $base_options_hash =
        {
                GUARANTEE => {name => 'space-guarantee',
                              enabled => 'volume',
                              disabled => 'none',
                              file => 'file',
                              },
                FRACTIONAL_RESERVE =>{name =>'fractional-reserve',
                                      disabled => 0,
                                      },
                TRY_FIRST =>{name =>'space-mgmt-try-first',
                             volume_grow => 'volume_grow',
                             snap_delete => 'snap_delete',
                             },
	};

	my $code = $args{option_code};
	my $option = $base_options_hash->{$code};
        
        if (!defined($option))
        {
                logresult("script", "This library only maintains compression-specific options.\n" .
                                    "Option code \"$code\" is not supported for this release.");
        }
	logcomment("_Cmode_base_options: Option code=$code, option hash: " . Dumper($option)) if ($class->{_debug});

	return $option;
}



1;
