#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/ConsoleMessages.pm#1 $
#

## @summary Utility module to allow console message interpretation.
##
## @description
##    Utility module to allow console message interpretation.
##
## @synopsis
##
## use Compression::UI::ConsoleMessages;
## use Compression::Version;
## use NOTE::Filer;  # To obtain filer version.
##
## my $filer = NOTE::Filer->connect( host => $FILER );
## my $version = Compression::Version->new( conn => $filer );
## my $obj = Compression::UI::CLI::ConsoleMessages->new( version => $version );
##
## @examples 
## my $messages = $obj->get_message_format(message_codes => [C_CONSOLE_VOLUME_FULL,
##                                                           C_CONSOLE_SFSR_ABORTED]);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)



package Compression::UI::ConsoleMessages;


use strict;
use warnings;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Tharn;
use Compression::Constants qw(:all);
use Compression::Version;



BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
        
};


## @name ::new
##
## @summary Compression::UI::ConsoleMessages Constructor.
##
## @return Object of type Compression::UI::CLI::ConsoleMessages.
##
## @arg version Required: Compression::Version object for the hosting filer.
## 
## @arg boot_mode Optional: Boot mode of the hosting filer.
## Defaults to 7M.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::UI::ConsoleMessages->new( version => $version );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##

sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug  => {  
			type 	 => SCALAR,
			optional => 1,
                        default  => FALSE,
			},
	} );
        
	my $self = {};
        
        $self->{_debug} = $args{debug};
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};
        
        logcomment("Created new ConsoleMessages object:\n" . Dumper($self)) if ($self->{_debug});
	
	bless $self, $class;
	return $self;
}


## @name ::get_message_format
## 
## @description
## Generates the console message(s) correspondent to one or more codes.
##
## @return
## Reference to array of message strings.
## 
## @arg message_code Required: Array ref of codes corresponding to required messages.
## 
## @examples
## $obj->get_message_format(message_codes => [C_CONSOLE_VOLUME_FULL,
##                                            C_CONSOLE_SFSR_ABORTED]);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##

sub get_message_format
{
        my $class = shift;
	my %args = validate(@_,
	{
                message_codes  => { type => ARRAYREF, },
	} );
        
        my $message_codes = $args{message_codes};
        my $return_array_ref = [];
        
        foreach my $message_code (@$message_codes)
        {
                logcomment( "ConsoleMessages: Processing code $message_code" ) if ($class->{_debug});
                my $message_string =  $class->_get_individual_message_format( message_code => $message_code );
                push(@$return_array_ref, $message_string);
        }
        
        return $return_array_ref;
}



###################################
# Internal subroutines
###################################

sub _get_individual_message_format
{
        my $class = shift;
	my %args = validate(@_,
	{
                message_code  => { type => SCALAR, },
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $message_code = $args{message_code};
        my $message_string;
        
        if ($ontap_ver eq '8.0')
	{
		$message_string = $class->_console_messages_8_0( message_code => $message_code );
	}
	elsif ($ontap_ver eq '8.1')
	{
		$message_string = $class->_console_messages_8_1( message_code => $message_code );
	}
	else
	{
		logcomment("ConsoleMessages: Specific support for sw_version $ontap_ver not found. Will be using base messages") if ($class->{_debug});
	}
        
        if ( !defined($message_string) )
        {
                logcomment( "ConsoleMessages: Could not find message specific for $ontap_ver. Defaulting to base messages. " ) if ($class->{_debug});
       		$message_string =  $class->_base_console_messages( message_code => $message_code );
        }
        
        return $message_string;
}

sub _base_console_messages
{
	my $class = shift;
	my %args = validate(@_,
	{
                message_code  => { type => SCALAR, },
	} );

	logcomment("ConsoleMessages: _base_console_messages: ". Dumper(%args) ) if ($class->{_debug});

	my $base_messages_hash = {
                C_CONSOLE_VOLUME_FULL => 'file system on volume \S+ is full',
                C_CONSOLE_OUT_OF_SPACE_SFSR_ABORTED => 'Out of disk space. Stopping snap restore for inode \d+ on volume \S+',
                C_CONSOLE_INSUFFICIENT_SPACE_FOR_SFSR => 'Not enough disk space for snap restore on volume \S+',
		C_CONSOLE_INSUFFICIENT_SPACE_TO_GROW_VOLUME => 'Unable to grow volume \S+ to recover space',
                C_CONSOLE_VOL_RESTORE_STARTED => 'Reverting volume \S+ to a previous snapshot',
                C_CONSOLE_DECOMPRESS_ABORTED => 'decompression of volume \S+ is aborted',
		C_CONSOLE_DECOMPRESS_COMPLETED => 'decompression of volume \S+ is done',
                C_CONSOLE_COMPRESS_CHECKPOINT_DELETED => 'Checkpoint deleted for compression of existing data for volume \S+',
                C_CONSOLE_COMPRESSION_ABORTED_AS_SNAPSHOT_IS_CREATED => 'Compression of existing data for volume \S+ aborting as snapshot is created',
                C_CONSOLE_OUT_OF_SPACE_SIS_ABORTED => 'Undoing shared blocks on volume \S+ has aborted because there is insufficient free space',
		C_CONSOLE_WAFL_INCONSISTANT => 'WAFL inconsistent: bad block at VBN \d+',
	};

	my $code = $args{message_code};
	my $message = $base_messages_hash->{$code};
        
        if (!defined($message))
        {
                logresult("script", "Console message code \"$code\" is not supported yet.");
        }
	logcomment("_base_console_messages: Code=$code,\nmessage: $message") if ($class->{_debug});

	return $message;
}


sub _console_messages_8_0
{
        my $class = shift;
	my %args = validate(@_,
	{
                message_code  => { type => SCALAR, },
	} );
        
        logcomment("ConsoleMessages: _console_messages_8_0: ". Dumper(%args) ) if ($class->{_debug});
        
	my $messages_hash_8_0 = {};
	my $code = $args{message_code};
        
        return $messages_hash_8_0->{$code};
}


sub _console_messages_8_1
{
        my $class = shift;
	my %args = validate(@_,
	{
                message_code  => { type => SCALAR, },
	} );
        
        logcomment("ConsoleMessages: _console_messages_8_1: ". Dumper(%args) ) if ($class->{_debug});
        
	my $messages_hash = {
		C_CONSOLE_DECOMPRESS_ABORTED => 'SIS undo operation for \S+ has stopped',
	};
	my $code = $args{message_code};
        
        return $messages_hash->{$code};
}



1;
