#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/Constants.pm#1 $
#

package Compression::UI::Constants;

use strict;
use warnings;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Tharn;
use Compression::Constants qw(:all);
use Compression::Version;


BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter); 
};


## @name ::new
##
## @summary 
##
## @return Object of type Compression::UI::Constants.
##
## @arg version Required: Compression::Version object for the hosting filer.
## 
## @arg boot_mode Optional: Boot mode of the hosting filer.
## Defaults to 7M.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::UI::Constants->new( version => $version );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
##

sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa  => "Compression::Version",},
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,},
                
                debug => {type => SCALAR,
                          optional => 1,
                          default  => FALSE,},
	} );
        
        my $self = {};
        
        $self->{_debug} = $args{debug};
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};
        
        logcomment("Created new Constants object:\n" . Dumper($self)) if ($self->{_debug});
        
        bless $self, $class;
        return $self;
}


## @name ::get_constant_value
## 
## @description
## Generates release agnostic/specific constant definition, including name and legal values to be set.
##
## @return
## 
## 
## @arg constant_code Required: code for constant for which the definition is needed.
## 
## @examples
## $obj->get_constant_value(constant_code => 'C_VBN_HEADER');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com, waroff@netapp.com
##

sub get_constant_value
{
        my $class = shift;
	my %args = validate(@_,
	{
                constant_code  => { type => SCALAR, },
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $constant_code = $args{constant_code};
        my $boot_mode = $class->{_boot_mode};
        my $constant_value;
        
        if ($boot_mode eq C_ONTAP_7MODE)
        {
                if ("$ontap_ver" eq '7.3')
                {
                        $constant_value = $class->_constants_7_3_7mode(constant_code => $constant_code);
                }
                elsif ("$ontap_ver" eq '8.0')
                {
                        $constant_value = $class->_constants_8_0_7mode(constant_code => $constant_code);
                }
                elsif ("$ontap_ver" eq '8.1')
                {
                        $constant_value = $class->_constants_8_1_7mode(constant_code => $constant_code);
                }
		elsif ("$ontap_ver" eq '8.2')
                {
                        $constant_value = $class->_constants_8_2_7mode(constant_code => $constant_code);
                }
                
                if (!defined($constant_value))
                {
                        logcomment("Constants: Specific support for sw_version $ontap_ver not found. Trying 7-mode base constants") if ($class->{_debug});
                        $constant_value = $class->_base_constants_7mode( constant_code => $constant_code );
                }
        }
        elsif ($boot_mode eq C_ONTAP_CMODE)
        {
		if ("$ontap_ver" eq '8.1')
		{
			$constant_value = $class->_constants_8_1_Cmode(constant_code => $constant_code);
		}
		
		if (!defined($constant_value))
		{
			logcomment("Constants: Specific support for sw_version $ontap_ver not found. Trying C-mode base constants") if ($class->{_debug});
			$constant_value = $class->_base_constants_Cmode(constant_code => $constant_code);
		}
        }
        else
        {
                logresult('script', "Unrecognized boot_mode: \"$boot_mode\"");
        }
        
        if (!defined($constant_value))
        {
                logcomment( "Constants: Could not find constants definition specific for $ontap_ver. Defaulting to base options. " ) if ($class->{_debug});
                $constant_value = $class->_base_constants( constant_code => $constant_code );
        }

        return $constant_value;
}


###################################
# Internal subroutines
###################################

sub _base_constants
{
	my $class = shift;
	my %args = validate(@_,
	{
                constant_code  => { type => SCALAR, },
	} );

	logcomment("Constants: _base_constants: ". Dumper(%args) ) if ($class->{_debug});

	my $base_constants_hash = {
                'C_DFLT_INITIALIZE_SIS' => TRUE,
                'C_DFLT_IS_INLINE_ENABLED' => C_RC_SUCCESS,
                'C_DFLT_ENABLE_INLINE' => TRUE,
                'C_DFLT_DISABLE_INLINE' => FALSE,
                'C_STAGE_COMPRESS_PREPROC' => 'Compress_Preproc',
                'C_STAGE_COMPRESSING' => 'Compressing',
                'C_STAGE_CHECKING_P1' => 'Checking_P1',
                'C_STAGE_CHECKING_P2' => 'Checking_P2',
                'C_STAGE_NONE' => '-',
                'C_STAGE_SAVING' => 'Saving',
                'C_STAGE_SORTING' => 'Sorting',
                'C_STAGE_GATHERING' => 'Gathering',
		'C_STAGE_INITIALIZING' => 'Initializing',
                'C_STAGE_DONE' => 'Done',
                'C_STATUS_IDLE' => 'Idle',
                'C_STATUS_ACTIVE' => 'Active',
		'C_STATUS_INITIALIZING' => 'Initializing',
                'C_STATUS_UNDOING' => 'Undoing',
                'C_CHECKPOINT_STAGE_NONE' => '-',
                'C_CHECKPOINT_STAGE_SORTING' => 'Sorting',
                'C_CHECKPOINT_STAGE_SAVING_PASS1' => 'Saving_pass1',
                'C_CHECKPOINT_STAGE_SAVING_PASS2' => 'Saving_pass2',
                'C_CHECKPOINT_STAGE_SAVING_SHARING' => 'saving_sharing',
                'C_CHECKPOINT_STAGE_SAVING_END' => 'Saving_end',
                'C_CHECKPOINT_STAGE_COMPRESSING' => 'Compressing',
                'RUN_SCANNER_IF_SAVINGS_MISMATCH' => TRUE,
                'IGNORE_RUN_SCANNER_FLAG' => FALSE,
		'C_SIS_ENABLED_STR' => 'Enabled',
		'C_SIS_DISABLED_STR' => 'Disabled',
                'INITIALIZING' => 'Initializing',
                'IDLE' => 'Idle',
                'ACTIVE' => 'Active',
                'UNDOING' => 'Undoing',
                'NOSTATUS' => 'NoStatus',
                'DONE' => 'Done',
                'DASH' => '-',
                'FILE_FOLDING_SCAN_TYPE' => 'file folding',
                'FAILOVER_STATE_UP' => 'SF_UP',
                'FAILOVER_STATE_TO' => 'SF_TO',
                'FAILOVER_STATE_REBOOT' => 'SF_REBOOT',
                'WAFL_INCONSISTENT_PANIC_TESTING' => 'wafl_inconsistent_panic_testing',
                'WAFL_DECOMPRESS_FAIL_TESTING' => 'wafl_decompress_fail_testing',
                'WAFL_64BIT_UPGRADE_SCAN_DISABLE' => 'wafl_64bit_upgrade_scan_disable',
                'WAFL_IRON_PANIC_ON_CORRUPTION' => 'wafl_iron_panic_on_corruption',
		'DISABLE_FILE_CLONE_BG_SHARING' => 'wafl_disable_sis_clone_fixup',
		'BG_FILE_CLONE_ACTIVE_PREFIX' => 'Clone-inum = ',
                'DECOMPRESS_RESTARTS_AFTER_REBOOT' => FALSE,
                'SFSR_WILL_RESTORE_FILE_AS_COMPRESSED' => FALSE,
                'CONTAINMENT_VERSION_WAS_ENABLED' => 1,
                'HIGHEST_CONTAINMENT_VERSION_WAS_ENABLED' => 1,
                'CONTAINMENT_VERSION_NEVER_ENABLED' => 0,
                'HIGHEST_CONTAINMENT_VERSION_NEVER_ENABLED' => 0,
		'AGGR_DATA_DISK_PREFIX' => '\s*data\s+',
		'ENABLE_LICENSED_ISCSI' => 'licensed_feature.iscsi.enable',
		'VFILER_MIGRATE_IS_READY_FOR_COMPLETE' => 'snapmirrored',
		'CFO_LICENSE_NEEDED' => FALSE,
		'IDD_ENABLED_0' => 'false',
		'IDD_ENABLED_1' => 'true',
		'VBN_ZERO' => '281474976710644',
		'TRIAGE_FP_BY_FBN' => TRUE,
		'INSTALL_SIS_LICENSE' => FALSE,
                };

	my $code = $args{constant_code};
	my $constant = $base_constants_hash->{$code};
        
        if (!defined($constant))
        {
                logresult("script", "This library only maintains compression-specific constants. Option code \"$code\" is not supported.");
        }
	logcomment("_base_constants: Constant code=$code, constant hash: " . Dumper($constant)) if ($class->{_debug});

	return $constant;
}


sub _base_constants_7mode
{
	my $class = shift;
	my %args = validate(@_,
	{
                constant_code  => { type => SCALAR, },
	} );

	logcomment("Constants: _base_constants_7mode: ". Dumper(%args) ) if ($class->{_debug});
        my $code = $args{constant_code};

	my $constants_hash = {
                'SFSR_USES_CLONE_ENGINE' => FALSE,
        };

	return $constants_hash->{$code}
}


sub _base_constants_Cmode
{
	my $class = shift;
	my %args = validate(@_,
	{
                constant_code  => { type => SCALAR, },
	} );

	logcomment("Constants: _base_constants_Cmode: ". Dumper(%args) ) if ($class->{_debug});
        my $code = $args{constant_code};

	my $constants_hash = {
                'C_VBN_HEADER' => '281474976710654',
                'C_VBN_TRAILER' => '281474976710652',
                'C_VVBN_HEADER' => '281474976710654',
                'C_VVBN_TRAILER' => '281474976710652',
                'C_VVBN_HEADER_B' => '281474976710646',
                'C_DFLT_INITIALIZE_SIS' => TRUE,
                'C_DFLT_IS_INLINE_ENABLED' => C_RC_SUCCESS,
                'C_DFLT_ENABLE_INLINE' => TRUE,
                'C_DFLT_DISABLE_INLINE' => FALSE,
                'IGNORE_RUN_SCANNER_FLAG' => FALSE,
                'C_FIJI_DENSE_SLEEP_STAGE_NOP' => 'DENSE_STAGE_NOP',
                'C_FIJI_DENSE_SLEEP_STAGE_LOGGING' => 'DENSE_STAGE_LOGGING',
                'C_FIJI_DENSE_SLEEP_STAGE_WOC_SORTING' => 'DENSE_STAGE_WOC_SORTING', 
                'C_FIJI_DENSE_SLEEP_STAGE_WOC_COMPRESSING' => 'DENSE_STAGE_WOC_COMPRESSING', 
                'C_FIJI_DENSE_SLEEP_STAGE_GATHERING' => 'DENSE_STAGE_GATHERING',
                'C_FIJI_DENSE_SLEEP_STAGE_SORTING' => 'DENSE_STAGE_SORTING',
                'C_FIJI_DENSE_SLEEP_STAGE_DEDUPING_P1' => 'DENSE_STAGE_DEDUPING_P1', 
                'C_FIJI_DENSE_SLEEP_STAGE_DEDUPING_P2' => 'DENSE_STAGE_DEDUPING_P2',
                'C_FIJI_DENSE_SLEEP_STAGE_DEDUPING_SHARING' => 'DENSE_STAGE_DEDUPING_SHARING',
                'C_FIJI_DENSE_SLEEP_STAGE_CHECKING' => 'DENSE_STAGE_CHECKING',
                'C_FIJI_DENSE_SLEEP_STAGE_CHECKING_P1' => 'DENSE_STAGE_CHECKING_P1',
                'C_FIJI_DENSE_SLEEP_STAGE_CHECKING_P2' => 'DENSE_STAGE_CHECKING_P2',
                'C_FIJI_DENSE_SLEEP_STAGE_DONE' => 'DENSE_STAGE_DONE',
                'SFSR_USES_CLONE_ENGINE' => TRUE,
		'SFSR_WILL_RESTORE_FILE_AS_COMPRESSED' => TRUE,
        };

	return $constants_hash->{$code}
}


sub _constants_7_3_7mode
{
	my $class = shift;
	my %args = validate(@_,
	{
                constant_code  => { type => SCALAR, },
	} );
        
        logcomment("constant: _constants_7_3_7mode: ". Dumper(%args) ) if ($class->{_debug});
        my $code = $args{constant_code};
        
	my $constants_hash_7_3 = {
                'C_VBN_HEADER'		=> '4294967294',
                'C_VBN_TRAILER'		=> '4294967292',
                'C_VVBN_HEADER'		=> '4294967294',
                'C_VVBN_TRAILER'	=> '4294967292',
                'C_DFLT_INITIALIZE_SIS' => FALSE,
                'IGNORE_RUN_SCANNER_FLAG' => TRUE,
                'DECOMPRESS_RESTARTS_AFTER_REBOOT' => TRUE,
                'SFSR_WILL_RESTORE_FILE_AS_COMPRESSED' => TRUE,
		'CFO_LICENSE_NEEDED' => TRUE,
		'TRIAGE_FP_BY_FBN' => FALSE,
		'INSTALL_SIS_LICENSE' => TRUE,
	};
        
        return $constants_hash_7_3->{$code}
}

sub _constants_8_0_7mode
{
        my $class = shift;
	my %args = validate(@_,
	{
                constant_code  => { type => SCALAR, },
	} );
        
        logcomment("Constants: _constants_8_0_7mode: ". Dumper(%args) ) if ($class->{_debug});
        
	my $constants_hash_BR_0 = {
                'C_VBN_HEADER'		=> '281474976710654',
                'C_VBN_TRAILER'		=> '281474976710652',
                'C_VVBN_HEADER'		=> '281474976710654',
                'C_VVBN_TRAILER'	=> '281474976710652',
                'IGNORE_RUN_SCANNER_FLAG' => TRUE,
                'DECOMPRESS_RESTARTS_AFTER_REBOOT' => TRUE,
                'SFSR_WILL_RESTORE_FILE_AS_COMPRESSED' => TRUE,
		'CFO_LICENSE_NEEDED' => TRUE,
		'TRIAGE_FP_BY_FBN' => FALSE,
		'INSTALL_SIS_LICENSE' => TRUE,
	};
	my $code = $args{constant_code};
        
    return $constants_hash_BR_0->{$code}
}

sub _constants_8_1_7mode
{
        my $class = shift;
        my %args = validate(@_,
	{
                constant_code  => { type => SCALAR, },
	} );
        
        logcomment("Constants: _constants_8_1_7mode: ". Dumper(%args) ) if ($class->{_debug});
        
	my $constants_hash_8_1 = {
                'C_VBN_HEADER' => '281474976710654',
                'C_VBN_TRAILER' => '281474976710652',
                'C_VVBN_HEADER' => '281474976710654',
                'C_VVBN_TRAILER' => '281474976710652',
                'C_VVBN_HEADER_B' => '281474976710646',
                'C_DFLT_INITIALIZE_SIS' => TRUE,
                'C_DFLT_IS_INLINE_ENABLED' => C_RC_SUCCESS,
                'C_DFLT_ENABLE_INLINE' => TRUE,
                'C_DFLT_DISABLE_INLINE' => FALSE,
                'IGNORE_RUN_SCANNER_FLAG' => FALSE,
                'C_FIJI_DENSE_SLEEP_STAGE_NOP' => 'DENSE_STAGE_NOP',
                'C_FIJI_DENSE_SLEEP_STAGE_LOGGING' => 'DENSE_STAGE_LOGGING',
                'C_FIJI_DENSE_SLEEP_STAGE_WOC_SORTING' => 'DENSE_STAGE_WOC_SORTING', 
                'C_FIJI_DENSE_SLEEP_STAGE_WOC_COMPRESSING' => 'DENSE_STAGE_WOC_COMPRESSING', 
                'C_FIJI_DENSE_SLEEP_STAGE_GATHERING' => 'DENSE_STAGE_GATHERING',
                'C_FIJI_DENSE_SLEEP_STAGE_SORTING' => 'DENSE_STAGE_SORTING',
                'C_FIJI_DENSE_SLEEP_STAGE_DEDUPING_P1' => 'DENSE_STAGE_DEDUPING_P1', 
                'C_FIJI_DENSE_SLEEP_STAGE_DEDUPING_P2' => 'DENSE_STAGE_DEDUPING_P2',
                'C_FIJI_DENSE_SLEEP_STAGE_DEDUPING_SHARING' => 'DENSE_STAGE_DEDUPING_SHARING',
                'C_FIJI_DENSE_SLEEP_STAGE_CHECKING' => 'DENSE_STAGE_CHECKING',
                'C_FIJI_DENSE_SLEEP_STAGE_CHECKING_P1' => 'DENSE_STAGE_CHECKING_P1',
                'C_FIJI_DENSE_SLEEP_STAGE_CHECKING_P2' => 'DENSE_STAGE_CHECKING_P2',
                'C_FIJI_DENSE_SLEEP_STAGE_DONE' => 'DENSE_STAGE_DONE',
		'CFO_LICENSE_NEEDED' => TRUE,
		'TRIAGE_FP_BY_FBN' => FALSE,
		'INSTALL_SIS_LICENSE' => TRUE,
	};
	my $code = $args{constant_code};
        
        return $constants_hash_8_1->{$code}
}


sub _constants_8_2_7mode
{
        my $class = shift;
        my %args = validate(@_,
	{
                constant_code  => { type => SCALAR, },
	} );
        
        logcomment("Constants: _constants_8_2_7mode: ". Dumper(%args) ) if ($class->{_debug});
        
	my $constants_hash = {
                'C_VBN_HEADER' => '281474976710654',
                'C_VBN_TRAILER' => '281474976710652',
                'C_VVBN_HEADER' => '281474976710654',
                'C_VVBN_TRAILER' => '281474976710652',
                'C_VVBN_HEADER_B' => '281474976710646',
                'C_DFLT_INITIALIZE_SIS' => TRUE,
                'C_DFLT_IS_INLINE_ENABLED' => C_RC_SUCCESS,
                'C_DFLT_ENABLE_INLINE' => TRUE,
                'C_DFLT_DISABLE_INLINE' => FALSE,
                'IGNORE_RUN_SCANNER_FLAG' => FALSE,
                'C_FIJI_DENSE_SLEEP_STAGE_NOP' => 'DENSE_STAGE_NOP',
                'C_FIJI_DENSE_SLEEP_STAGE_LOGGING' => 'DENSE_STAGE_LOGGING',
                'C_FIJI_DENSE_SLEEP_STAGE_WOC_SORTING' => 'DENSE_STAGE_WOC_SORTING', 
                'C_FIJI_DENSE_SLEEP_STAGE_WOC_COMPRESSING' => 'DENSE_STAGE_WOC_COMPRESSING', 
                'C_FIJI_DENSE_SLEEP_STAGE_GATHERING' => 'DENSE_STAGE_GATHERING',
                'C_FIJI_DENSE_SLEEP_STAGE_SORTING' => 'DENSE_STAGE_SORTING',
                'C_FIJI_DENSE_SLEEP_STAGE_DEDUPING_P1' => 'DENSE_STAGE_DEDUPING_P1', 
                'C_FIJI_DENSE_SLEEP_STAGE_DEDUPING_P2' => 'DENSE_STAGE_DEDUPING_P2',
                'C_FIJI_DENSE_SLEEP_STAGE_DEDUPING_SHARING' => 'DENSE_STAGE_DEDUPING_SHARING',
                'C_FIJI_DENSE_SLEEP_STAGE_CHECKING' => 'DENSE_STAGE_CHECKING',
                'C_FIJI_DENSE_SLEEP_STAGE_CHECKING_P1' => 'DENSE_STAGE_CHECKING_P1',
                'C_FIJI_DENSE_SLEEP_STAGE_CHECKING_P2' => 'DENSE_STAGE_CHECKING_P2',
                'C_FIJI_DENSE_SLEEP_STAGE_DONE' => 'DENSE_STAGE_DONE',
	};
	my $code = $args{constant_code};
        
        return $constants_hash->{$code}
}


sub _constants_8_1_Cmode
{
	my $class = shift;
	my %args = validate(@_,
	{
		constant_code  => { type => SCALAR, },
	} );
	
	logcomment("Constants: _constants_8_1_Cmode: ". Dumper(%args) ) if ($class->{_debug});
        
	my $constants_hash = {
                'TRIAGE_FP_BY_FBN' => FALSE,
	};
	my $code = $args{constant_code};
	
	return $constants_hash->{$code}
}



1;
