#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/ZAPI.pm#1 $
#

## ZAPI utilities for Compression modules.
##
## @description
##    This module provides basic functionality for Compression ZAPI (7M, 10M)
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (internal)

package Compression::UI::ZAPI;

use strict;
use warnings;

BEGIN
{
	require "Compression/base_import.pl";
}

use Data::Dumper;
use Params::Validate qw( validate SCALAR HASHREF ARRAYREF OBJECT UNDEF BOOLEAN);

use Tharn;
use Compression::Constants qw(:all);
use Compression::UI::ZAPI::Commands;
use Compression::UI::ZAPI::Formats;
use Compression::UI::ZAPI::Messages;



# @name ::new
# @summary Compression::UI::ZAPI Constructor.
#
# @return Object of type Compression::UI::ZAPI.
#
# @arg version Required: Compression::Version object for the hosting filer.
# 
# @arg boot_mode Optional: Boot mode of the hosting filer.
# Defaults to 7M.
#
# @arg debug Optional: Controls verbosity of this module.
# Defaults to FALSE.
#
# @example
# my $obj = Compression::UI::ZAPI->new( version => $version );
#
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private (internal)
#
sub new
{
	my $class = shift;
	my %args = validate(@_,
	{	
		version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type     => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug => {type => SCALAR,
                          optional => 1,
                          default => FALSE,
                          },

	} );
	
	my $self = {};
        $self->{_debug} = $args{debug};
	$self->{id} = $$;
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};
	
	$self->{commands} = Compression::UI::ZAPI::Commands->new(version => $args{version},
                                                                 boot_mode => $self->{_boot_mode},
                                                                 debug => $self->{_debug} );
	$self->{formats} = Compression::UI::ZAPI::Formats->new(version => $args{version},
							       boot_mode => $self->{_boot_mode},
							       debug => $self->{_debug});
	$self->{messages} = Compression::UI::ZAPI::Messages->new(version => $args{version},
								 boot_mode => $self->{_boot_mode},
								 debug => $self->{_debug});
        
	bless $self, $class;

	return $self;
}


# @name ::get_command_format
# Returns the raw-xml command to be sent to the filer.
# This raw-xml command already includes any options.
# The expected format command definitions is explained in UI/ZAPI/Commands.pm
#
# Arguments:
# - code: the code corresponding to the command for which we need to get the layout (defined in Constants.pm).
# - the rest of the arguments are "tokens" for any command arguments thet we need to plug-in.
#   Ex: volume - token for volume name for a command that accepts the volume name as an argument.
#
# @status private (internal)
#
sub get_command_format
{
	my $class = shift;
	my %args = validate(@_,
	{
		code      => { type => SCALAR, },
		volume    => { type => SCALAR|UNDEF, optional => 1, },
		vserver   => { type => SCALAR|UNDEF, optional => 1, },
		file_path => { type => SCALAR|UNDEF, optional => 1, },
		fbn       => { type => SCALAR|UNDEF, optional => 1, },
		modifier  => { type => SCALAR|UNDEF, optional => 1, },
		size      => { type => SCALAR|UNDEF, optional => 1, },
		increment => { type => SCALAR|UNDEF, optional => 1, },
		state     => { type => SCALAR|UNDEF, optional => 1, },
		options   => { type => HASHREF|UNDEF, optional => 1, },
		snapshot  => { type => SCALAR|UNDEF, optional => 1, },
		restore_path => { type => SCALAR|UNDEF, optional => 1, },
		level => { type => SCALAR|UNDEF, optional => 1, },
		aggregate => { type => SCALAR|UNDEF, optional => 1, },
		vfiler => { type => SCALAR|UNDEF, optional => 1, },
		filer => { type => SCALAR|UNDEF, optional => 1, },
	});
	
	my $cmd_code = delete($args{code});
	my $options = delete($args{options});
	# Remaining args should be only command params (eg: volume name, aggregate, etc)
	
	my $raw_cmd = $class->{commands}->get_command_format(code => $cmd_code);
	if ($raw_cmd eq C_STR_ZAPI_NOT_AVAILABLE)
	{
		logresult('script', "Library problem found: get_command_format called for command not supported for zapi: $cmd_code");
	}
	
	# Extract any child elements
	my $children = "<>";
	if ($raw_cmd =~ /\S+\s+(.+)/)
	{
		$children = $1;
		$raw_cmd =~ s/\s+.+//;
	}
	
	my $cmd = "";
	
	$children =~ s/^\<//;
	$children =~ s/\>$//;
	my @children_to_add = split(/\s+/, $children);
	foreach my $child_to_add (@children_to_add)
	{
		if ($child_to_add =~ /^\<(\S+)\>$/)
		{
			# Mandatory child
			my $child = $1;
			if ($child =~ /(\S+)=(\S+)/)
			{
				# Need naming conversion
				my $actual_name = $1;
				my $required_args = $2;
				if ($required_args =~ /\<(\S+)\>/)
				{
					my $needed_arg = $1;
					my $arg_value = $args{$needed_arg};
					
					if (!defined($arg_value))
					{
						logresult('script', "Library problem found: required command component \"$needed_arg\" was not passed in.");
					}
					
					$required_args =~ s/\<$needed_arg\>/$arg_value/;
					
					$cmd .= "<$actual_name>$required_args</$actual_name>";
				}
				else
				{
					logresult('script', "Malformed zapi format for command code $cmd_code: $child" );
				}
			}
			else
			{
				# Direct addition
				my $arg_value = $args{$child};
				if (!defined($arg_value))
				{
					logresult('script', "Library problem found: required command component \"$child\" was not passed in.");
				}
				
				$cmd .= "<$child>$arg_value</$child>";
			}
		}
		elsif ($child_to_add =~ /^\[(\S+)\]$/)
		{
			# Non-mandatory child
			my $child = $1;
			if ($child =~ /(\S+)=(\S+)/)
			{
				# Need naming conversion
				my $actual_name = $1;
				my $required_args = $2;
				if ($required_args =~ /\[(\S+)\]/)
				{
					my $needed_arg = $1;
					my $arg_value = $args{$needed_arg};
					
					if (defined($arg_value))
					{
						$required_args =~ s/\[$needed_arg\]/$arg_value/;
						$cmd .= "<$actual_name>$required_args</$actual_name>";
					}
				}
				else
				{
					logresult('script', "Malformed zapi format for command code $cmd_code: $child" );
				}
			}
			else
			{
				# Direct addition
				my $arg_value = $args{$child};
				if (defined($arg_value))
				{
					$cmd .= "<$child>$arg_value</$child>";
				}
			}
		}
		elsif ($child_to_add =~ /^(\S+)=(\S+)$/)
		{
			# Internal-use children, to be set to pre-defined value (eg: verbose=true)
			$cmd .= "<$1>$2</$1>";
		}
		# Ignore the rest...
	}
	my $cmd_options = $class->{commands}->get_options_string(code => $cmd_code, options => $options,);
	
	$cmd = "<$raw_cmd>" . $cmd . $cmd_options . "</$raw_cmd>";
	
	return $cmd;
}


# @name ::process_output
# @description
# Wrapper for ZAPI command output processing.
# Please see documentation for arguments/behavior in UI.pm
# @status private (internal)
#
sub process_output
{
        my $class = shift;
	my %args = validate(@_,
	{
                code      => {type => SCALAR,},
                output    => {type => SCALAR|OBJECT,}, 
                modifier  => {type => SCALAR, optional => 1,},
                required_return_hash_key => {type => SCALAR|UNDEF, optional => 1,},
                disable_data_identifier_processing => {type => BOOLEAN, optional => 1, default => FALSE,},
	} );
        
        my $code = $args{code};
        my $output = $args{output};
        my %return_hash = (rc => C_RC_FAILURE, output => $output);
        
        # Check if output is expected to be in a formatted form or just a plain massage string
        %return_hash = $class->{formats}->check_and_process_output_if_formatted(%args);
	
	logcomment("ZAPI output:\n" . $return_hash{output});
		
        if ((defined($return_hash{rc})) && ($return_hash{rc} == C_RC_SUCCESS))
        {
                return %return_hash;
        }
	
        # We get here if:
        # - Output is expected to be just a message string.
        # - Output is expected to be formatted, but was not. In this case, still continue: try to make better sense of failure.
	$return_hash{rc} = $class->{messages}->get_return_code(output => $return_hash{output},
                                                               code => $code);
        
        return %return_hash;
}


1;
