#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/ZAPI/Commands.pm#1 $
#

## @summary ZAPI Commands template for Compression.
##
## @description
##    This library provides ZAPI command templates for compression functionality (7M, CM).
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (internal)




package Compression::UI::ZAPI::Commands;

use strict;
use warnings;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);

use Tharn;
use Module::Common;

use Compression::Constants qw(:all);
use Compression::Version;


BEGIN
{
	require Exporter;
	our @ISA = qw(Exporter);
	require "Compression/base_import.pl";
};


# @name ::new
#
# @summary Compression::UI::ZAPI::Commands Constructor.
#
# @return Object of type Compression::UI::ZAPI::Commands.
#
# @arg version Required: Compression::Version object for the hosting filer.
# 
# @arg boot_mode Optional: Boot mode of the hosting filer.
# Defaults to 7M.
#
# @arg debug Optional: Controls verbosity of this module.
# Defaults to FALSE.
#
# @example
# my $obj = Compression::UI::ZAPI::Commands->new( version => $version );
#
sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type     => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug  => {  
			type 	 => SCALAR,
			optional => 1,
                        default  => FALSE,
			},
		} );
        
	my $self = {};
        
        $self->{_debug} = $args{debug};
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};
	
	bless $self, $class;
	return $self;
}

# @name ::get_command_format
# Generates release agnostic/specific commands based on the template definition
# without populating any place-holders.
#
# @return Raw command string (any place-holders are left in place).
# 
# @arg code Required: Code corresponding to the operation for which the command is generated.
# Operation codes are defined in Compression::Constants.
#
# Command format is: "zapi-command <extra command arguments>",
# where "extra command arguments" can contain any of the folowing space-delimited types of arguments:
# - <required_arg> : a required argument. Its value needs to be passed in by the caller. Ex: <volume>
# - <required_actual_arg=<required_arg>> : a required argument. Its value needs to be passed in by the caller, but using a different identifier.
#   Ex: <path=/vol/<volume>> (the caller will pass the volume, but the xml command will be formatted by using path = /vol/volume).
# - [optional_arg] : a required argument. Its value may be passed in by the caller. Ex: [volume]
# - arg=value : internal-use argument (the value will always be set to the lib-defined content, cannot be changed by the caller).
#
sub get_command_format
{
	my $class = shift;
	my %args = validate(@_,
	{	
		code => { type => SCALAR, },
                
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $boot_mode = $class->{_boot_mode};
        my $command_code = $args{code};
        my $return_cmd;
        
        logcomment("ZAPI Commands: get_command_format. Command lookup for code=$command_code," .
                   " ontap version $ontap_ver, boot_mode $boot_mode") if ($class->{_debug});
        
        if ($boot_mode eq C_ONTAP_7MODE)
        {
                if ($ontap_ver eq '8.0')
                {
			$return_cmd = $class->_7mode_commands_8_0(code => $command_code);
                }
                elsif ($ontap_ver eq '8.1')
                {
                        $return_cmd = $class->_7mode_commands_8_1(code => $command_code);
                }
                
                if (!defined($return_cmd))
                {
                        logcomment( "ZAPI Commands: Could not find command specific for 7-mode $ontap_ver. Defaulting to base commands. " ) if ($class->{_debug});
                        $return_cmd =  $class->_7mode_base_commands(code => $command_code);
                }
        }
        elsif ($boot_mode eq C_ONTAP_CMODE)
        {
		if ($ontap_ver eq '8.1')
		{
			$return_cmd = $class->_Cmode_commands_8_1(code => $command_code);
		}
		
		if (!defined($return_cmd))
		{
			logcomment( "ZAPI Commands: Could not find command specific for C-mode $ontap_ver. Defaulting to base commands. " ) if ($class->{_debug});
			$return_cmd =  $class->_Cmode_base_commands(code => $command_code);
		}
        }
        else
        {
                logresult('script', "Unrecognized boot_mode: \"$boot_mode\"");
        }
        
        logcomment("ZAPI Commands: get_command_format. Returning command for code=$command_code (boot_mode=$boot_mode): \"$return_cmd\"") if ($class->{_debug});
        
        return $return_cmd;
}


# @name ::get_options_string
# Generates release agnostic/specific ZAPI command options.
#
# @return Xml-formatted string representing the command options.
# 
# @arg code Required: Code corresponding to the operation for which the command options are needed.
# Operation codes are defined in Compression::Constants.
# 
# @arg options Optional: Hashref of command options, as accepted by the highest-level interface. 
#
# Expected format for command options: token=value
#

sub get_options_string
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => { type => SCALAR, },
                options => { type => HASHREF|UNDEF, optional => 1, },
        } );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $boot_mode = $class->{_boot_mode};
        my $command_code = $args{code};
        my $options_to_set = $args{options};
        my $allowed_options;
	my $raw_options = '';
        
        logcomment("ZAPI Commands: get_options: Options lookup for code=$command_code, ontap version $ontap_ver\n" .
                   "Options required to be set:\n" . Dumper($options_to_set) ) if ($class->{_debug});
        
        if ($boot_mode eq C_ONTAP_7MODE)
        {
                if ($ontap_ver eq '8.0')
                {
                        $allowed_options = $class->_7mode_cmd_options_8_0( code => $command_code );
                }
                elsif ($ontap_ver eq '8.1')
                {
                        $allowed_options = $class->_7mode_cmd_options_8_1( code => $command_code );
                }
		
                if (!defined($allowed_options))
                {
                        logcomment("ZAPI Commands: get_options: Specific support for sw_version $ontap_ver not found. Will be using base options") if ($class->{_debug});
                        $allowed_options = $class->_7mode_base_cmd_options(code => $command_code);
                }
        }
        elsif ($boot_mode eq C_ONTAP_CMODE)
        {
		# Only base-cmds for c-mode for now
                $allowed_options = $class->_Cmode_base_cmd_options(code => $command_code);
        }
        else
        {
                logresult('script', "Unrecognized boot_mode: \"$boot_mode\"");
        }
        
        logcomment("Commands: get_options_string. Allowed options for command code=$command_code:\n" . Dumper($allowed_options)) if ($class->{_debug});
        
        foreach my $option_symbol (keys(%$options_to_set))
        {
                my $option_value = $options_to_set->{$option_symbol};
                if (!defined($option_value))
                {
                        logcomment("Option $option_symbol had undefined value. Skipping... ") if ($class->{_debug});
                        next;
                }
                logcomment("Option $option_symbol required to be set to $option_value") if ($class->{_debug});
                
                if (defined($allowed_options->{$option_symbol}))
                {
                        my $allowed_option_values = $allowed_options->{$option_symbol};
                        if ($option_value == TRUE)
			{
				if (defined($allowed_option_values->{TRUE}))
				{
					logcomment("Option $option_symbol is allowed to be set to $option_value: " .
						   $allowed_option_values->{TRUE}) if ($class->{_debug});
					$raw_options .= $allowed_option_values->{TRUE} . " ";
				}
                        }
			elsif ($option_value == FALSE)
			{
				if (defined($allowed_option_values->{FALSE}))
				{
					logcomment("Option $option_symbol is allowed to be set to $option_value: " .
						   $allowed_option_values->{FALSE}) if ($class->{_debug});
					$raw_options .= $allowed_option_values->{FALSE} . " ";
				}
                        }
			else
			{
				logcomment("Option $option_symbol is currently not allowed to be set to $option_value. Skipping");
                        }
                }
                else
                {
                        logcomment("Commands: get_options_string:\n" .
                                   "Command code: $command_code\n" .
                                   "Attempt to set unsupported option for this release: $option_symbol\n" .
                                   "Ignoring...") if ($class->{_debug});
                }
        }
	
	my $return_str = '';
	my @unprocessed_options = split(/\s+/, $raw_options);
	foreach my $unprocessed_option (@unprocessed_options)
	{
		if ($unprocessed_option =~ /^(\S+)=(\S+)$/)
		{
			$return_str .= "<$1>$2</$1>";
		}
		else
		{
			logresult('script', "\nLibrary problem found:\nMalformed ZAPI command options for command $command_code\n" .
				            "Illegal option: $unprocessed_option\n" .
					    "Allowed format is: option_name=option_value");
		}
	}
	
	logcomment("Commands: get_options_string: returning command options for code $command_code: $return_str") if ($class->{_debug});
        return $return_str;
}

########################################
# Internal soubroutines: command strings
########################################
sub _7mode_base_commands
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => { type => SCALAR, },
        });
        
        my $code = $args{code};
        logcomment("ZAPI Commands: _7mode_base_commands input: ". Dumper(%args) ) if( $class->{_debug} );
        
        my %base_cmds = (
		IS_COMPRESSION_ENABLED     => "sis-status <verbose=true <path=/vol/<volume>>>",
		ENABLE_COMPRESSION         => "sis-set-config <<path=/vol/<volume>>>",
		DISABLE_COMPRESSION        => "sis-set-config <enable-inline-compression=false <path=/vol/<volume>>>",
		GET_SPACE_SAVINGS          => "volume-list-info <[volume]>",
		COMPRESS_START             => "sis-start <<path=/vol/<volume>>>",
		COMPRESS_STATUS            => "sis-status <verbose=true <path=/vol/<volume>>>",
		COMPRESS_STOP              => "sis-stop <<path=/vol/<volume>>>",
		DECOMPRESS_STATUS          => "sis-status <verbose=true <path=/vol/<volume>>>",
		DECOMPRESS_STOP            => "sis-stop <<path=/vol/<volume>>>",
		COMPRESS_STATUS_CHECKPOINT => "sis-status <verbose=true <path=/vol/<volume>>>",
		_PARSE_CHECKPOINT_TIME     => "<internal_no_op>", # No command is sent to the filer, but declare it available, so parsing can be done
		CONFIGURE_IDD              => "sis-set-config <<enable-idd=<modifier>> <path=/vol/<volume>>>",
		IS_IDD_ENABLED             => "sis-status <verbose=true <path=/vol/<volume>>>",
	);
        
        my $cmd = $base_cmds{$code};
        
        if (!defined($cmd))
        {
                return C_STR_ZAPI_NOT_AVAILABLE;
        }
        logcomment("ZAPI Commands: _7mode_base_commands. Returning command for code=$code: \"$cmd\"") if ($class->{_debug});
        
        return $cmd;
}

sub _7mode_commands_8_0
{
	my $class = shift;
	my %args = validate(@_,
	{	
		code      => { type => SCALAR, },
	} );

	logcomment("ZAPI Commands: _7mode_commands_8_0 input: ". Dumper(%args)) if ($class->{_debug});
        
	my %cmds = ();
        
        my $code = $args{code};
        
        #return $cmds{$code};
	# Br is not supported for zapi yet
	return C_STR_ZAPI_NOT_AVAILABLE;
}


sub _7mode_commands_8_1
{
        my $class = shift;
        my %args = validate(@_,
        {
                code      => { type => SCALAR, },
        } );
        
        logcomment("ZAPI Commands: _7mode_commands_8_1 input: ". Dumper(%args)) if ($class->{_debug});
        
        my $code = $args{code};
        my %cmds = (
		CONFIGURE_IDD => C_STR_ZAPI_NOT_AVAILABLE,
		IS_IDD_ENABLED => C_STR_ZAPI_NOT_AVAILABLE);
        
        return $cmds{$code};
}

sub _Cmode_base_commands
{
	my $class = shift;
	my %args = validate(@_,
	{
		code => { type => SCALAR, },
	});
	
	my $code = $args{code};
	logcomment("ZAPI Commands: _Cmode_base_commands input: ". Dumper(%args) ) if( $class->{_debug} );
	
	my %base_cmds = (
		IS_COMPRESSION_ENABLED     => "sis-status <verbose=true <path=/vol/<volume>> <vserver>>",
		ENABLE_COMPRESSION         => "sis-set-config <<path=/vol/<volume>> <vserver>>",
		DISABLE_COMPRESSION        => "sis-set-config <enable-inline-compression=false <path=/vol/<volume>> <vserver>>",
		GET_SPACE_SAVINGS          => "volume-get-iter", # No volume-level filtering allowed by this api
		COMPRESS_STATUS            => "sis-status <verbose=true <path=/vol/<volume>> <vserver>>",
		COMPRESS_STATUS_CHECKPOINT => "sis-status <verbose=true <path=/vol/<volume>> <vserver>>",
		_PARSE_CHECKPOINT_TIME     => "<internal_no_op>", # No command is sent to the filer, but declare it available, so parsing can be done
		COMPRESS_START             => "sis-start <<path=/vol/<volume>> <vserver>>",
		COMPRESS_STOP              => "sis-stop <<path=/vol/<volume>> <vserver>>",
		DECOMPRESS_STATUS          => "sis-status <verbose=true <path=/vol/<volume>> <vserver>>",
		DECOMPRESS_STOP            => "sis-stop <<path=/vol/<volume>> <vserver>>",
		CONFIGURE_IDD              => "sis-set-config <<enable-idd=<modifier>> <path=/vol/<volume>> <vserver>>",
		IS_IDD_ENABLED             => "sis-status <verbose=true <path=/vol/<volume>> <vserver>>",
	);
	
	my $cmd = $base_cmds{$code};
	
	if (!defined($cmd))
	{
		return C_STR_ZAPI_NOT_AVAILABLE;
	}
	logcomment("ZAPI Commands: _Cmode_base_commands. Returning command for code=$code: \"$cmd\"") if ($class->{_debug});
	
	return $cmd;
}

sub _Cmode_commands_8_1
{
	my $class = shift;
	my %args = validate(@_,
	{
		code => { type => SCALAR, },
	});
	
	logcomment("ZAPI Commands: _Cmode_commands_8_1 input: ". Dumper(%args)) if ($class->{_debug});
	
	my $code = $args{code};
	my %cmds = (
		CONFIGURE_IDD              => C_STR_ZAPI_NOT_AVAILABLE,
		IS_IDD_ENABLED             => C_STR_ZAPI_NOT_AVAILABLE,
		);
	
	return $cmds{$code};
}


#########################################
## Internal soubroutines: command options
#########################################

sub _7mode_base_cmd_options
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => {type => SCALAR,},
        } );
        
        my $code = $args{code};
        logcomment("ZAPI Commands: _7mode_base_cmd_options input: ". Dumper(%args) ) if ($class->{_debug});
        
        my $base_cmd_options = {
		ENABLE_COMPRESSION => {inline => {TRUE => 'enable-inline-compression=true',
						  FALSE => 'enable-inline-compression=false'},
				       background => {TRUE => 'enable-compression=true',
						      FALSE => 'enable-compression=false'},},
		DISABLE_COMPRESSION => {disable_inline_only => {TRUE => 'enable-compression=true',
								FALSE => 'enable-compression=false'},},
		COMPRESS_START => {scan => {TRUE => 'scan=true',},
				   delete_checkpoint => {TRUE => 'delete-checkpoint=true',},},
                };
	
	my $return = (exists($base_cmd_options->{$code})) ? $base_cmd_options->{$code} : {};
        
        return $return;
}


sub _7mode_cmd_options_8_0
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => {type => SCALAR,},
        } );
        
        logcomment("Commands: _7mode_cmd_options_8_0 input: ". Dumper(%args)) if ($class->{_debug});
        
        my $cmd_options = {};
        
        my $code = $args{code};
        return $cmd_options->{$code};
}


sub _7mode_cmd_options_8_1
{
        my $class = shift;
        my %args = validate(@_,
        {
                code => {type => SCALAR,},
        } );
        
        logcomment("Commands: _7mode_cmd_options_8_1 input: ". Dumper(%args)) if ($class->{_debug});
        
        my $cmd_options = {};
        
        my $code = $args{code};
        return $cmd_options->{$code};
}

sub _Cmode_base_cmd_options
{
	my $class = shift;
	my %args = validate(@_,
	{
		code => {type => SCALAR,},
	});
	
	my $code = $args{code};
	logcomment("ZAPI Commands: _Cmode_base_cmd_options input: " . Dumper(%args)) if ($class->{_debug});
	
	my $base_cmd_options = {
		ENABLE_COMPRESSION => {inline => {TRUE => 'enable-inline-compression=true',
						  FALSE => 'enable-inline-compression=false'},
				       background => {TRUE => 'enable-compression=true',
						      FALSE => 'enable-compression=false'},},
		DISABLE_COMPRESSION => {disable_inline_only => {TRUE => 'enable-compression=true',
								FALSE => 'enable-compression=false'},},
		COMPRESS_START => {scan => {TRUE => 'scan=true',},
				   delete_checkpoint => {TRUE => 'delete-checkpoint=true',},},
                };
	
	my $return = (exists($base_cmd_options->{$code})) ? $base_cmd_options->{$code} : {};
	return $return;
}


# Leave this undocummented (only used internally)
sub is_vfiler_available_cmd
{
        my $class = shift;
	my %args = validate(@_,
	{	
		code      => { type => SCALAR, },
        } );

        # No vfiler support for zapi
	return FALSE;
}




1;
