#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/ZAPI/Formats.pm#1 $
#

## @summary Utility module for parsing formatted ZAPI output (when more than one value needs to be extracted).
##
## @description
## This module provides functionality for parsing of ZAPI command output
## when more than one value needs to be extracted.
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (internal)



package Compression::UI::ZAPI::Formats;

use strict;
use warnings;

BEGIN
{
	require "Compression/base_import.pl";
}

use Tharn;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Compression::Constants qw(:all);
use Compression::Utils qw(convert_data_size_value);




# @name ::new
# @summary Compression::UI::ZAPI::Formats Constructor.
#
# @return Object of type Compression::UI::ZAPI::Formats.
#
# @arg version Required: Compression::Version object for the hosting filer.
# 
# @arg boot_mode Optional: Boot mode of the hosting filer.
# Defaults to 7M.
#
# @arg debug Optional: Controls verbosity of this module.
# Defaults to FALSE.
#
# @example
# my $obj = Compression::UI::ZAPI::Formats->new( version => $version );
#
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private (internal)
#
sub new
{
	my $class = shift;
        my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug  => {  
			type 	 => SCALAR,
			optional => 1,
                        default  => FALSE,
			},
	} );
        
	my $self = {};
        
        $self->{_debug} = $args{debug};
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};

	logcomment("Created new ZAPI Formats object:\n" . Dumper($self)) if ($self->{_debug});
	bless $self, $class;

	return $self;
}


# @name ::check_and_process_output_if_formatted.
# Wrapper for processing of ZAPI command output when more than one value needs to be extracted.
# This method will process the zapi output and extract information according to the base or release-specific format.
# Expected format definitions:
# - nesting_tokens : a hashref containing all the xml tokens that nest the infomation of interest (ex: sis status for an individual volume)
# - entity_identifier : once the output is filtered at the nesting_tokens level, entity_identifier is the token that defines the entity (ex: volume name)
# - return_keys : at the entity level: what information should be returned and what are the xml tokens that define it
# - subnesting : useful for information that is sub-nested by different xml tokens, at the entity level
#   (ex: logical-data is a sub-nested entity of volume-level sis status)
# For both nesting and subnesting, extra processing for the obtained information can be done by using one of the following keys:
# - retrieved_value_modifier : for values that need modifier changes to be returned as required by the caller
#   (ex: for space savings, 'used' is returned by zapi in bytes, but the caller expects kbytes)
# - extract_information : flag used to extract extra information from the zapi output for a specific token
#   (ex: for sis status, the progress contains the amount of data that was already scanned. This will be extracted in a separate return entry)
# - assign_value : useful for commands that do not have a consistent output, but the caller expects a consistent return hash
#   (ex: in c-mode, for sis status: no checkpoint information is returned if checkpoint was not yet created.
#        In this case, some default values will be assigned to this missing info)
#
sub check_and_process_output_if_formatted
{
        my $class = shift;
	my %args = validate(@_,
	{
                code     => { type => SCALAR, },
                output   => { type => SCALAR|OBJECT, },
                modifier => {type => SCALAR|UNDEF, optional => 1,},
                required_return_hash_key => {type => SCALAR|UNDEF, optional => 1,}, 
                disable_data_identifier_processing => {type => BOOLEAN, optional => 1, default => FALSE},
	});
        
        logcomment("ZAPI Formats: check_and_process_output_if_formatted input:\n" . Dumper(%args)) if ($class->{_debug});
        
        my $code = $args{code};
        my $output = $args{output};
	my $modifier = $args{modifier};
	my $required_return_key = $args{required_return_hash_key};
	my $found_required_return_key = (defined($required_return_key)) ? 0 : 1;
	
	# Note: for now, we do not need to deal with duplicate_entries_allowed 
	
        my %return_hash = (rc => C_RC_FAILURE, output => $output);
	
	if (ref($output) eq '') 
	{
		logcomment("ZAPI Formats: output for command code $code was not formatted\n") if ($class->{_debug});
		return %return_hash;
	}
	
	$return_hash{output} = $output->get_raw_output();

        # If command corresponding to this code does not expect a formatted output, bail out imediatelly
        my $format;
        my $boot_mode = $class->{_boot_mode};
        if ($boot_mode eq C_ONTAP_7MODE)
        {
                $format = $class->_check_if_7mode_formatted_output_expected( code => $code );
        }
        elsif($boot_mode eq C_ONTAP_CMODE)
        {
                $format = $class->_check_if_Cmode_formatted_output_expected( code => $code );
        }
        else
        {
                logresult('script', "Unrecognized boot_mode: \"$boot_mode\"");
        }
	
        if ( $format->{format_expected} == FALSE )
        {
                logcomment("ZAPI Formats: formatted output is not expected for code $code\n") if ($class->{_debug});
                return %return_hash;
        }
	
	# NACL will return an arrayref of hashrefs
	my $parsed_output = $output->get_parsed_output();
	
	my $return_hash_keys = $format->{return_hash_keys};
	foreach my $return_hash_key (keys(%$return_hash_keys))
	{
		my $return_key_info = $return_hash_keys->{$return_hash_key};
		my $output_to_process = $parsed_output;
		
		logcomment("ZAPI Formats: processing code $code, return key \"$return_hash_key\"\n" .
			   "Expected output info:\n" . Dumper($return_key_info)) if ($class->{_debug});
		
		my $nesting_tokens = $return_key_info->{nesting_tokens};
		my $found = 1;
		if ((defined($nesting_tokens)) && (ref($nesting_tokens) eq 'ARRAY'))
		{
			# Subtract only the information that is required (within the nested identifiers)
			foreach my $nesting_id (@$nesting_tokens)
			{
				if (ref($output_to_process) ne 'ARRAY')
				{
					logresult('script', "ZAPI Formats: unexpected output type for command code $code.\n" .
						            "Don't know how to process:\n\n" . Dumper($output_to_process));
				}
				$found = 0;
				foreach my $output_to_process_entry (@$output_to_process)
				{
					if (exists($output_to_process_entry->{$nesting_id}))
					{
						$output_to_process = $output_to_process_entry->{$nesting_id};
						$found = 1;
						last;
					}
				}
				
				if ($found == 0)
				{
					logresult('fatal', "\nNested ZAPI output is expected for code $code, but output is incomplete\n" .
						  "Expected information nested by $nesting_id\n" .
						  "Obtained output:\n" . Dumper($output));
				}
			}
		}
		
		my $entity_identifier = $return_key_info->{entity_identifier};
		if (!(defined($entity_identifier)))
		{
			logresult('script', "Library problem found: missing entity identifier for code $code");
		}
		
		if (ref($output_to_process) ne 'ARRAY')
		{
			logresult('script', "ZAPI Formats: unexpected output type for command code $code (entity-level processing).\n" .
				  "Don't know how to process:\n\n" . Dumper($output_to_process));
		}
		
		foreach my $entity (@$output_to_process)
		{
			my $entity_identifier_token = $entity_identifier->{token};
			my $entity_name_info = $entity;
			my $entity_return_hash = {};
			my $detected_complete_info_for_entity = 1;
			if (exists($entity_identifier->{subnesting}))
			{
				my $entity_name_subnesting = $entity_identifier->{subnesting};
				foreach my $entity_name_subnesting_token (@$entity_name_subnesting)
				{
					if (ref($entity_name_info) eq 'HASH')
					{
						$entity_name_info = $entity_name_info->{$entity_name_subnesting_token};
						if (ref($entity_name_info) eq 'ARRAY')
						{
							# For now, just processed based on the idea that this sould only contain one entry
							# (since it identifies the entry)
							# If this does not hold in the future, we will fail the parsing later on, by not finding the required name
							$entity_name_info = $entity_name_info->[0];
						}
						else
						{
							logresult('fatal', "ZAPI: Unexpected entity-level information (array expected). Unable to obtain entity name:\n" . Dumper($entity_name_info));
						}
					}
					else
					{
						logresult('fatal', "ZAPI: Unexpected entity-level information (hash expected). Unable to obtain entity name:\n" . Dumper($entity_name_info));
					}
				}
			}
			
			if (!(defined($entity_name_info->{$entity_identifier_token})))
			{
				logresult('fatal', "\nUnable to detect the \"$entity_identifier_token\" value in the ZAPI output\n" .
						   "Faulty information:\n" . Dumper($entity) . "\n\n" .
						   "ZAPI output:\n" . Dumper($output));
			}
			
			my $entity_name = $entity_name_info->{$entity_identifier_token};
			
			if (exists($entity_identifier->{prefix}))
			{
				my $prefix = $entity_identifier->{prefix};
				logcomment("ZAPI Formats: processing code $code:\n" .
					   "Removing entity identifier prefix \"$prefix\" " .
					   "from entity name $entity_name") if ($class->{_debug});
				$entity_name =~ s/^$prefix//;
			}
			
			if ((defined($required_return_key)) && ("$entity_name" ne "$required_return_key"))
			{
				# Not interested in this entry
				logcomment("ZAPI Formats: Skipping entry \"$entity_name\"\n") if ($class->{_debug});
				next;
			}
			
			# See if we need to extract any return data at this level (before sub-nesting)
			my $return_keys = $return_key_info->{return_keys};
			if (defined($return_keys))
			{
				foreach my $key (keys(%$return_keys))
				{
					my %extracted_information = $class->_extract_information(entity => $entity,
												 requirements => $return_keys->{$key},
												 modifier => $modifier,
												 cmd_code => $code,
												 return_key => $key);
					
					foreach my $detected_info (keys(%extracted_information))
					{
						$entity_return_hash->{$detected_info} = $extracted_information{$detected_info};
					}
				}
			}
			
			my $subnesting = $return_key_info->{subnesting};
			if (defined($subnesting))
			{
				foreach my $subnesting_info (@$subnesting)
				{
					my $subnesting_tokens = $subnesting_info->{subnesting_tokens};
					my @temp_subnesting_tokens = @$subnesting_tokens;
					my $subnesting_return_keys = $subnesting_info->{return_keys};
					my $subnested_entity;
					
					# Extract only the information that is required (within the sub-nested identifiers)
					# The first sub-nesting id needs to be processed diferently:
					# It is just the key in the hash
					my $subnesting_id = shift(@temp_subnesting_tokens);
					if (exists($entity->{$subnesting_id}))
					{
						$subnested_entity = $entity->{$subnesting_id};
					}
					else
					{
						logcomment("\nMissing sub-nested ZAPI information for entity $entity_name (code $code)\n" .
							   "Could not find information subnested by $subnesting_id\n" .
							   "ZAPI output:\n" . Dumper($output));
						$detected_complete_info_for_entity = 0;
						# Do not process any other nesting info. This entity will not be added to return hash.
						last;
					}
					
					# Now we are processing arrayrefs
					foreach $subnesting_id (@temp_subnesting_tokens)
					{
						if (ref($subnested_entity) ne 'ARRAY')
						{
							logresult('script', "ZAPI Formats: unexpected sub-nested output type for command code $code.\n" .
								            "Don't know how to process:\n\n" . Dumper($subnested_entity));
						}
						
						$found = 0;
						foreach my $entity_entry (@$subnested_entity)
						{
							if (exists($entity_entry->{$subnesting_id}))
							{
								$subnested_entity = $entity_entry->{$subnesting_id};
								$found = 1;
								last;
							}
						}
						
						if ($found == 0)
						{
							logresult('fatal', "\nSub-nested ZAPI output is expected for code $code, but output is incomplete\n" .
								           "Expected information nested by $subnesting_id\n" .
									   "Obtained output:\n" . Dumper($subnested_entity));
						}
					}
					
					if (ref($subnested_entity) eq 'ARRAY')
					{
						if (scalar(@$subnested_entity) == 1)
						{
							$subnested_entity = $subnested_entity->[0];
						}
						else
						{
							logresult('fatal', "\nSub-nested ZAPI output is expected for code $code, but multiple entries were found\n" .
								           "Don't know how to process\n" .
									   "Obtained output:\n" . Dumper($subnested_entity));
						}
					}
					
					foreach my $subnested_key (keys(%$subnesting_return_keys))
					{
						my %extracted_information = $class->_extract_information(entity => $subnested_entity,
													 requirements => $subnesting_return_keys->{$subnested_key},
													 modifier => $modifier,
													 cmd_code => $code,
													 return_key => $subnested_key);
						
						foreach my $detected_info (keys(%extracted_information))
						{
							$entity_return_hash->{$detected_info} = $extracted_information{$detected_info};
						}
					}
				}
			}
			
			# Add extracted info to return hash
			if ($detected_complete_info_for_entity == 1)
			{
				$found_required_return_key = 1;
				foreach my $return_key (keys(%$entity_return_hash))
				{
					$return_hash{$return_hash_key}{$entity_name}{$return_key} = $entity_return_hash->{$return_key};
				}
			}
		}
	}
        
	if ($found_required_return_key == 0)
	{
		logresult('fatal', "Error in processing ZAPI output for code $code:\n" .
                                   "Did not find information for required entity \"$required_return_key\"\n" .
                                   "Output:\n" . Dumper($output));
        }
	
	$return_hash{rc} = C_RC_SUCCESS;
        
        logcomment("ZAPI Formats: check_and_process_output_if_formatted returning:\n" . Dumper(%return_hash)) if ($class->{_debug});
        return %return_hash;
}


#########################################
# Internal subroutines
#########################################

sub _extract_information
{
        my $class = shift;
	my %args = validate(@_,
	{
                entity => {type => HASHREF,},
		requirements => {type => HASHREF,},
		modifier => {type => SCALAR|UNDEF,},
		cmd_code => {type => SCALAR,},
		return_key => {type => SCALAR,},
	} );
	
	my $entity = $args{entity};
	my $requirements = $args{requirements};
	my $modifier = $args{modifier};
	my $code = $args{cmd_code};
	my $return_key = $args{return_key};
	my %return = ();
	
	my $token = $requirements->{token};
	my $value;
	if (!exists($entity->{$token}))
	{
		if (exists($requirements->{assign_value}))
		{
			# Allow missing information and assign a default value
			$value = $requirements->{assign_value};
		}
		else
		{
			logresult('fatal', "Missing information in the ZAPI format for code $code: unable to detect $token\n" . Dumper($entity));
		}
	}
	else
	{
		$value = $entity->{$token};
	}
	
	if ((defined($modifier)) &&
	    (exists($requirements->{retrieved_value_modifier})))
	{
		$value = Compression::Utils::convert_data_size_value(value => $value,
								     current_modifier => $requirements->{retrieved_value_modifier},
								     required_modifier => $modifier,);
	}
	
	$return{$return_key} = $value;
	
	if (exists($requirements->{extract_information}))
	{
		my $extract_extra_info = $requirements->{extract_information};
		
		foreach my $info_to_extract (@$extract_extra_info)
		{
			my $regex = $info_to_extract->{regex};
			logcomment("ZAPI Formats: extra processing for \"$value\" (cmd code $code)\n" .
				   "Attempting to extract info:\n" . Dumper($info_to_extract)) if ($class->{_debug});
			
			if ($value !~ /^($regex)/)
			{
				# Exit this loop to stop extracting extra info
				logcomment("ZAPI Formats: Failed to match leading regex: $regex (cmd code $code)\n" .
					   "No more extra info will be extracted from:\n$value") if ($class->{_debug});
				last;
			}
			
			my $extra_info_data = $1;
			my $extra_info_output_hash_key = $info_to_extract->{return_key};
			if (defined($extra_info_output_hash_key))
			{
				logcomment("ZAPI Formats: Adding extra info to the return hash:\n" .
					   "$extra_info_output_hash_key => $extra_info_data") if ($class->{_debug});
				$return{$extra_info_output_hash_key} = $extra_info_data;
			}
			
			$value =~ s/^$extra_info_data//;
		}
	}
	
	
	return %return;
}

sub _check_if_7mode_formatted_output_expected
{
        my $class = shift;
	my %args = validate(@_,
	{
                code  => {type => SCALAR,},
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $code = $args{code};
        my $return_hash = {};
        
	if ($ontap_ver eq '8.0')
	{
		$return_hash = $class->_7mode_formats_8_0(code => $code);
	}
        
        if ((!defined($return_hash)) || (!(scalar %$return_hash)))
        {
                logcomment( "ZAPI Formats: Could not find 7mode format definition specific for $ontap_ver. Checking base formats.") if ($class->{_debug});
       		$return_hash =  $class->_7mode_base_formats(code => $code);
        }

        return $return_hash;
}




sub _7mode_base_formats
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => {type => SCALAR,},
	});

	logcomment("ZAPI Formats: _7mode_base_formats input:\n". Dumper(%args)) if ($class->{_debug});

	my $base_formats_hash =
	{
		GET_SPACE_SAVINGS => {format_expected => TRUE,
				      return_hash_keys => {'savings' => {nesting_tokens => ['volumes', 'volume-info',],
									 entity_identifier => {token => 'name',},
									 return_keys => {used => {token => 'size-used',
												  retrieved_value_modifier => 'b',},},
									 subnesting => [{subnesting_tokens => ['sis', 'sis-info',],
											 return_keys => {total_saved => {token => 'total-saved',
															 retrieved_value_modifier => 'k',},
													 percentage_saved_total => {token => 'percent-total-saved'},
													 deduplicated => {token => 'dedup-saved',
															  retrieved_value_modifier => 'k',},
													 percentage_deduplicated => {token => 'percent-dedup-saved'},
													 compressed => {token => 'compress-saved',
															retrieved_value_modifier => 'k',},
													 percentage_compressed => {token => 'percent-compress-saved'},},},],
									 },},
				      },
		COMPRESS_STATUS => {format_expected => TRUE,
				    return_hash_keys => {'compress_statistics' => {nesting_tokens => ['sis-object', 'dense-status',],
										   entity_identifier => {token => 'path',
													 prefix => '/vol/',},
										   return_keys => {state => {token => 'state',},
												   compression => {token => 'is-compression-enabled',},
												   inline_compression => {token => 'is-inline-compression-enabled',},
												   status => {token => 'status',},
												   progress => {token => 'progress',
														extract_information => [{regex => '\d+',
																	 return_key => 'data_scanned',
																	 },],},
												   type => {token => 'type',},
												   schedule => {token => 'schedule',},
												   minimum_blocks_shared => {token => 'minimum-blocks-shared',},
												   blocks_skipped_sharing => {token => 'blocks-skipped-sharing',},
												   last_operation_begin => {token => 'last-operation-begin-timestamp',},
												   last_operation_end => {token => 'last-operation-end-timestamp',},
												   last_operation_size => {token => 'last-operation-size-bytes',},
												   last_operation_error => {token => 'last-operation-state',},
												   changelog_usage => {token => 'changelog-used-percent',},
												   queued_jobs => {token => 'queued-job-type',},
												   checkpoint_time => {token => 'checkpoint-time',},
												   checkpoint_op_type => {token => 'checkpoint-op-type',},
												   checkpoint_stage => {token => 'checkpoint-stage',},
												   checkpoint_substage => {token => 'checkpoint-sub-stage',},
												   checkpoint_progress => {token => 'checkpoint-progress',},
												   blocks_deduplicated => {token => 'minimum-blocks-shared',},},
										   subnesting => [{subnesting_tokens => ['logical-data', 'sis-logical-data',],
												   return_keys => {logical_data => {token => 'logical-data-size',},},},],
									 },},},
		DECOMPRESS_STATUS => {format_expected => TRUE,
				      return_hash_keys => {'decompress_statistics' => {nesting_tokens => ['sis-object', 'dense-status',],
										       entity_identifier => {token => 'path',
													     prefix => '/vol/',},
										       return_keys => {state => {token => 'state',},
												       compression => {token => 'is-compression-enabled',},
												       inline_compression => {token => 'is-inline-compression-enabled',},
												       status => {token => 'status',},
												       progress => {token => 'progress',
														    extract_information => [{regex => '\d+',
																	     return_key => 'data_processed',},
																	    {regex => '\s+KB\s+',},
																	    {regex => '\w+',
																	     return_key => 'processing_phase',},],},
												       type => {token => 'type',},
												       schedule => {token => 'schedule',},
												       minimum_blocks_shared => {token => 'minimum-blocks-shared',},
												       blocks_skipped_sharing => {token => 'blocks-skipped-sharing',},
												       last_operation_begin => {token => 'last-operation-begin-timestamp',},
												       last_operation_end => {token => 'last-operation-end-timestamp',},
												       last_operation_size => {token => 'last-operation-size-bytes',},
												       last_operation_error => {token => 'last-operation-state',},
												       changelog_usage => {token => 'changelog-used-percent',},
												       queued_jobs => {token => 'queued-job-type',},
												       checkpoint_time => {token => 'checkpoint-time',},
												       checkpoint_op_type => {token => 'checkpoint-op-type',},
												       checkpoint_stage => {token => 'checkpoint-stage',},
												       checkpoint_substage => {token => 'checkpoint-sub-stage',},
												       checkpoint_progress => {token => 'checkpoint-progress',},
												       blocks_deduplicated => {token => 'minimum-blocks-shared',},},
										       subnesting => [{subnesting_tokens => ['logical-data', 'sis-logical-data',],
												       return_keys => {logical_data => {token => 'logical-data-size',},},},],
										       },},},
		COMPRESS_STATUS_CHECKPOINT => {format_expected => TRUE,
					       return_hash_keys => {'checkpoint_info' => {nesting_tokens => ['sis-object', 'dense-status',],
											  entity_identifier => {token => 'path',
														prefix => '/vol/',},
											  return_keys => {state => {token => 'state',},
													  compression => {token => 'is-compression-enabled',},
													  inline_compression => {token => 'is-inline-compression-enabled',},
													  status => {token => 'status',},
													  progress => {token => 'progress',
														       extract_information => [{regex => '\d+',
																		return_key => 'total_data_scanned',},],},
													  type => {token => 'type',},
													  schedule => {token => 'schedule',},
													  minimum_blocks_shared => {token => 'minimum-blocks-shared',},
													  blocks_skipped_sharing => {token => 'blocks-skipped-sharing',},
													  last_operation_begin => {token => 'last-operation-begin-timestamp',},
													  last_operation_end => {token => 'last-operation-end-timestamp',},
													  last_operation_size => {token => 'last-operation-size-bytes',},
													  last_operation_error => {token => 'last-operation-state',},
													  changelog_usage => {token => 'changelog-used-percent',},
													  queued_jobs => {token => 'queued-job-type',},
													  checkpoint_creation_time => {token => 'checkpoint-time',},
													  checkpoint_op_type => {token => 'checkpoint-op-type',},
													  checkpoint_stage => {token => 'checkpoint-stage',},
													  checkpoint_substage => {token => 'checkpoint-sub-stage',},
													  checkpoint_progress => {token => 'checkpoint-progress',},
													  blocks_deduplicated => {token => 'minimum-blocks-shared',},},
											  subnesting => [{subnesting_tokens => ['logical-data', 'sis-logical-data',],
													  return_keys => {logical_data => {token => 'logical-data-size',},},},],
											  },},},
	};

	my $code = $args{code};
	my $format = $base_formats_hash->{$code};
        
        if (!defined($format))
        {
                $format = {format_expected => FALSE};
                logcomment("Code \"$code\": Command output is not expected to be formatted.") if ($class->{_debug});
        }
	logcomment("ZAPI Formats: _7mode_base_formats returning: Code=$code, option hash:\n" . Dumper($format)) if ($class->{_debug});

	return $format;
}

sub _7mode_formats_8_0
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => {type => SCALAR,},
	} );
        
        logcomment("ZAPI Formats: _7mode_formats_8_0:\n". Dumper(%args)) if ($class->{_debug});
        
	my $formats_hash = {
	};
	my $code = $args{code};
        
        return $formats_hash->{$code};
}


sub _check_if_Cmode_formatted_output_expected
{
        my $class = shift;
	my %args = validate(@_,
	{
                code  => {type => SCALAR,},
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $code = $args{code};
        
	# Only base-formats for c-mode for now
	return $class->_Cmode_base_formats(code => $code);
}


sub _Cmode_base_formats
{
	my $class = shift;
	my %args = validate(@_,
	{
		code  => {type => SCALAR,},
	});
	
	logcomment("ZAPI Formats: _Cmode_base_formats input:\n". Dumper(%args)) if ($class->{_debug});

	my $base_formats_hash = {
		GET_SPACE_SAVINGS => {format_expected => TRUE,
				      return_hash_keys => {'savings' => {nesting_tokens => ['attributes-list', 'volume-attributes',],
									 entity_identifier => {token => 'name',
											       subnesting => ['volume-id-attributes',],},
									 subnesting => [{subnesting_tokens => ['volume-sis-attributes',],
											 return_keys => {total_saved => {token => 'total-space-saved',
															 retrieved_value_modifier => 'b',},
													 percentage_saved_total => {token => 'percentage-total-space-saved'},
													 deduplicated => {token => 'deduplication-space-saved',
															  retrieved_value_modifier => 'b',},
													 percentage_deduplicated => {token => 'percentage-deduplication-space-saved'},
													 compressed => {token => 'compression-space-saved',
															retrieved_value_modifier => 'b',},
													 percentage_compressed => {token => 'percentage-compression-space-saved'},},},
											{subnesting_tokens => ['volume-space-attributes',],
											 return_keys => {used => {token => 'size-used',
														  retrieved_value_modifier => 'b',},},},],
									 },},
				      },
		COMPRESS_STATUS => {format_expected => TRUE,
				    return_hash_keys => {'compress_statistics' => {nesting_tokens => ['sis-object', 'dense-status',],
										   entity_identifier => {token => 'path',
													 prefix => '/vol/',},
										   return_keys => {state => {token => 'state',},
												   compression => {token => 'is-compression-enabled',},
												   inline_compression => {token => 'is-inline-compression-enabled',},
												   status => {token => 'status',},
												   progress => {token => 'progress',
														extract_information => [{regex => '\d+',
																	 return_key => 'data_scanned',},],},
												   type => {token => 'type',},
												   schedule => {token => 'schedule',},
												   minimum_blocks_shared => {token => 'minimum-blocks-shared',},
												   blocks_skipped_sharing => {token => 'blocks-skipped-sharing',},
												   last_operation_begin => {token => 'last-operation-begin-timestamp',},
												   last_operation_end => {token => 'last-operation-end-timestamp',},
												   last_operation_size => {token => 'last-operation-size-bytes',},
												   last_operation_error => {token => 'last-operation-state',},
												   changelog_usage => {token => 'changelog-used-percent',},
												   checkpoint_time => {token => 'checkpoint-time',
														       assign_value => 0,},
												   checkpoint_op_type => {token => 'checkpoint-op-type',
															  assign_value => '-',},
												   checkpoint_stage => {token => 'checkpoint-stage',
															assign_value => '-',},
												   checkpoint_substage => {token => 'checkpoint-sub-stage',
															   assign_value => '-',},
												   checkpoint_progress => {token => 'checkpoint-progress',
															   assign_value => '-',},
												   blocks_deduplicated => {token => 'minimum-blocks-shared',},},
										   subnesting => [{subnesting_tokens => ['logical-data', 'sis-logical-data',],
												   return_keys => {logical_data => {token => 'logical-data-size',},},},],
										   },},},
		DECOMPRESS_STATUS => {format_expected => TRUE,
				      return_hash_keys => {'decompress_statistics' => {nesting_tokens => ['sis-object', 'dense-status',],
										       entity_identifier => {token => 'path',
													     prefix => '/vol/',},
										       return_keys => {state => {token => 'state',},
												       compression => {token => 'is-compression-enabled',},
												       inline_compression => {token => 'is-inline-compression-enabled',},
												       status => {token => 'status',},
												       progress => {token => 'progress',
														    extract_information => [{regex => '\d+',
																	     return_key => 'data_processed',},
																	    {regex => '\s+KB\s+',},
																	    {regex => '\w+',
																	     return_key => 'processing_phase',},],},
												       type => {token => 'type',},
												       schedule => {token => 'schedule',},
												       minimum_blocks_shared => {token => 'minimum-blocks-shared',},
												       blocks_skipped_sharing => {token => 'blocks-skipped-sharing',},
												       last_operation_begin => {token => 'last-operation-begin-timestamp',},
												       last_operation_end => {token => 'last-operation-end-timestamp',},
												       last_operation_size => {token => 'last-operation-size-bytes',},
												       last_operation_error => {token => 'last-operation-state',},
												       changelog_usage => {token => 'changelog-used-percent',},
												       checkpoint_time => {token => 'checkpoint-time',
															   assign_value => 0,},
												       checkpoint_op_type => {token => 'checkpoint-op-type',
															      assign_value => '-',},
												       checkpoint_stage => {token => 'checkpoint-stage',
															    assign_value => '-',},
												       checkpoint_substage => {token => 'checkpoint-sub-stage',
															       assign_value => '-',},
												       checkpoint_progress => {token => 'checkpoint-progress',
															       assign_value => '-',},
												       blocks_deduplicated => {token => 'minimum-blocks-shared',},},
										       subnesting => [{subnesting_tokens => ['logical-data', 'sis-logical-data',],
												       return_keys => {logical_data => {token => 'logical-data-size',},},},],
										       },},},
		COMPRESS_STATUS_CHECKPOINT => {format_expected => TRUE,
					       return_hash_keys => {'checkpoint_info' => {nesting_tokens => ['sis-object', 'dense-status',],
											  entity_identifier => {token => 'path',
														prefix => '/vol/',},
											  return_keys => {state => {token => 'state',},
													  compression => {token => 'is-compression-enabled',},
													  inline_compression => {token => 'is-inline-compression-enabled',},
													  status => {token => 'status',},
													  progress => {token => 'progress',
														       extract_information => [{regex => '\d+',
																		return_key => 'total_data_scanned',},],},
													  type => {token => 'type',},
													  schedule => {token => 'schedule',},
													  minimum_blocks_shared => {token => 'minimum-blocks-shared',},
													  blocks_skipped_sharing => {token => 'blocks-skipped-sharing',},
													  last_operation_begin => {token => 'last-operation-begin-timestamp',},
													  last_operation_end => {token => 'last-operation-end-timestamp',},
													  last_operation_size => {token => 'last-operation-size-bytes',},
													  last_operation_error => {token => 'last-operation-state',},
													  changelog_usage => {token => 'changelog-used-percent',},
													  checkpoint_creation_time => {token => 'checkpoint-time',
																       assign_value => 0,},
													  checkpoint_op_type => {token => 'checkpoint-op-type',
																 assign_value => '-',},
													  checkpoint_stage => {token => 'checkpoint-stage',
															       assign_value => '-',},
													  checkpoint_substage => {token => 'checkpoint-sub-stage',
																  assign_value => '-',},
													  checkpoint_progress => {token => 'checkpoint-progress',
																  assign_value => '-',},
													  blocks_deduplicated => {token => 'minimum-blocks-shared',},},
											  subnesting => [{subnesting_tokens => ['logical-data', 'sis-logical-data',],
													  return_keys => {logical_data => {token => 'logical-data-size',},},},],
											  },},},
	};
	
	my $code = $args{code};
	my $format = $base_formats_hash->{$code};
	
	if (!defined($format))
        {
		$format = {format_expected => FALSE};
		logcomment("Code \"$code\": Command output is not expected to be formatted.") if ($class->{_debug});
        }
	logcomment("ZAPI Formats: _Cmode_base_formats returning: Code=$code, option hash:\n" . Dumper($format)) if ($class->{_debug});
	
	return $format;
}




1;
