#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/ZAPI/Messages.pm#1 $
#

## @summary Utility module to allow ZAPI output message interpretation. 
##
## @description
##    This module enables interpretation of ZAPI output message after command execution.
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (internal)

package Compression::UI::ZAPI::Messages;

use strict;
use warnings;

BEGIN
{
	require "Compression/base_import.pl";
}

use Tharn;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;


use Compression::Constants qw(:all);
use Compression::Version;
use Compression::UI::ZAPI::Messages::Constants_8_1;
use Compression::UI::ZAPI::Messages::Constants_8_2;






# @name ::new
#
# @summary Compression::UI::ZAPI::Messages Constructor.
#
# @return Object of type Compression::UI::ZAPI::Messages.
#
# @arg version Required: Compression::Version object for the hosting filer.
# 
# @arg boot_mode Optional: Boot mode of the hosting filer.
# Defaults to 7M.
#
# @arg debug Optional: Controls verbosity of this module.
# Defaults to FALSE.
#
# @example
# my $obj = Compression::UI::ZAPI::Messages->new( version => $version );
#
sub new
{
	my $class = shift;
        my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug  => {  
			type 	 => SCALAR,
			optional => 1,
                        default  => FALSE,
			},
	} );
        
	my $self = {};
        $self->{_debug} = $args{debug};
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};
        
        bless $self, $class;
        
        $self->_refresh_constants();

	logcomment("Massages: Created new object:\n" . Dumper($self)) if ($self->{_debug});
	

	return $self;
}


# @name ::get_return_code
# Parses the given output and returns the corresponding return code.
# The code is used to look-up known output messages for the specific command.
# If a match is found within outputs corresponding to the specific command,
# the corresponding return code is returned, including C_RC_SUCCESS.
# If not, all other known messages are searched, but an eventual C_RC_SUCCESS code will be changed into C_RC_FAILURE.
#
sub get_return_code
{
	
	my $class 	= shift;
	my (%args) = validate( @_,
	{
		output => { type => SCALAR, },
		code => { type => SCALAR, },
	} );
        
        my $output = $args{output};
        my $code = $args{code};
	my $constants = $class->{_constants}->{messages};
        my $search_result;
	
	# Remove the white spaces from the xml format, for easier parsing
	$output =~ s/>\s+\t*</></g;
        
        # First try to find a match through messages corresponding to this command code
        my $command_messages = $constants->{$code};
        
        # Handle any empty output
        my $legal_empty_output_rc = $class->_check_for_allowed_empty_output(output => $output,
                                                                            search_set => $command_messages,
                                                                            code => $code);
        if ($legal_empty_output_rc != C_RC_FAILURE)
        {
                return $legal_empty_output_rc;
        }
        
        # Passed-in code may not have an entry in the constants hash
        if (defined($command_messages))
        {
                logcomment("Messages:\n Searching for \"$output\"\n through messages for passed-in command code $code") if $class->{_debug};
                $search_result = $class->_search_for_code_specific_message(output => $output,
                                                                           search_set => $command_messages,
                                                                           );
                if ($search_result->{match_found} == TRUE)
                {
                        return $search_result->{rc};
                }
        }
        
        # Loop through the rest of messages, but:
        # since an eventual 'success' message would correspond to a different command,
        # turn any 'success' into 'failure'
        for my $attempt_code ( keys %$constants )
        {
                if ($attempt_code eq $code)
                {
                        next;
                }
                
                $command_messages = $constants->{$attempt_code};
                logcomment("ZAPI Messages:\n Searching for \"$output\"\n through messages for command code $attempt_code") if $class->{_debug};
                $search_result = $class->_search_for_code_specific_message(output => $output,
                                                                           search_set => $command_messages,
                                                                           );
                if ($search_result->{match_found} == TRUE)
                {
                        if ($search_result->{rc} == C_RC_SUCCESS)
                        {
                                logcomment("Messages:\n Command output for code $code, command output $output\n" .
                                           "Match found against success-indicating message for command code $attempt_code\n" .
                                           "Changing to failure...") if $class->{_debug};
                                return C_RC_FAILURE;
                        }
                        logcomment("Messages:\n Look-up for code $code, command output: $output\n" .
                                   "Returning " . $search_result->{rc}) if $class->{_debug};
                        return $search_result->{rc};
                }
        }
        
        logcomment("ZAPI Messages:\n Look-up for code $code, command output: $output\n" .
                   "Did not know how to interpret message. No match found. Failing...") if $class->{_debug};
        return C_RC_FAILURE;
}

############################
# Internal subroutines
############################

sub _check_for_allowed_empty_output
{
        my $class 	= shift;
	my (%args) = validate( @_,
	{
                output => { type => SCALAR|UNDEF, },
		search_set => { type => HASHREF|UNDEF, },
                code => { type => SCALAR, },
	} );
        
        my $search_set = $args{search_set};
        my $output = $args{output};
        my $code = $args{code};
        my $output_should_be_empty = FALSE;
        my $output_is_empty = FALSE;
        
        if (!defined($output) || ($output eq ''))
        {
                logcomment("Messages:\n Detected empty output for command code $code") if $class->{_debug};
                $output_is_empty = TRUE;
        }
        
        if (defined($search_set) && (defined($search_set->{''})))
        {
                $output_should_be_empty = TRUE;
        }
        
        if ($output_is_empty == TRUE)
        {
                if ($output_should_be_empty == TRUE)
                {
                        logcomment("Messages: detected (allowed) undefined or empty output string") if ($class->{_debug});
                        return $search_set->{''};
                }
                else
                {
                        logresult('fatal', "Unexpected empty output for command code $code");
                }
        }
        
        return C_RC_FAILURE;
}



sub _search_for_code_specific_message
{
        my $class 	= shift;
	my (%args) = validate( @_,
	{
                output => { type => SCALAR, },
		search_set => { type => HASHREF, },
	} );
        
        my $search_set = $args{search_set};
        my $output = $args{output};
        
        my $return_hash = { rc => C_RC_FAILURE, match_found => FALSE, };
        
        logcomment("Messages:\n Searching for \"$output\"\n through:\n" . Dumper($search_set)) if $class->{_debug};
        
        for my $message ( keys %$search_set )
        {
                if ($message ne '')
                {
                        my $rc = $search_set->{$message};
                        logcomment("Messages:\n trying: $message\n against: $output ") if $class->{_debug};
                        if ($output =~ /$message/i)
                        {
                                logcomment("Messages: SUCCESS\n matched: $message\n against: $output\nReturning $rc") if $class->{_debug};
                                $return_hash->{rc} = $rc;
                                $return_hash->{match_found} = TRUE;
                                last;
                        }
                }
        }
        
        return $return_hash;
}

sub _refresh_constants
{
        my $class = shift;
        my $ontap_ver = $class->{_version}->{ontap};
        
        if ($ontap_ver eq  '8.1')
	{
		$class->{_constants} = Compression::UI::ZAPI::Messages::Constants_8_1->new(debug => $class->{_debug},
											   boot_mode => $class->{_boot_mode});
	}
	elsif ($ontap_ver eq  '8.2')
	{
		$class->{_constants} = Compression::UI::ZAPI::Messages::Constants_8_2->new(debug => $class->{_debug},
											   boot_mode => $class->{_boot_mode});
	}
        else
	{
		# WARNING: this should always point to the latest 'specifically-supported' version (latest case above)
		logcomment("\n\n\nWARNING!!!\n" .
			   "Detected ontap version \"$ontap_ver\"\n" .
			   "Using the default ZAPI messages definitions (different ontap version).\n" .
			   "This may result in output matching failures for any ZAPI commands.\n" .
			   "If this is the case, please add correspondent module in Compression/UI/ZAPI/Messages");
		$class->{_constants} = Compression::UI::ZAPI::Messages::Constants_8_2->new(debug => $class->{_debug},
											   boot_mode => $class->{_boot_mode});
	}
}



1;
