#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UpgradeRevert.pm#1 $
#

## @summary Upgrade/revert utilities module.
##
## @description
## Upgrade/revert utilities module.
##
## @synopsis
## 
## BEGIN { require "drg/Compression/setup_core.pl"; }
##
## use Compression::UpgradeRevert;
##
## my $upgrade = Compression::UpgradeRevert->initialize(filer => $FILER,
##                                                      required_base_release => '8.0.0',
##                                                      base_release_url => 'http://10.57.83.10//engineering/rlse/DOT/RboilermakerN_100311_0300/final/bedrock/export/x86_64.debug/tarball/image.tgz',);
## $upgrade->upgrade_to_release(release => '8.1.0',
##                              url => 'http://10.57.83.10//engineering/rlse/DOT/mainN/final/bedrock/export/x86_64.debug/tarball/image.tgz');
## $upgrade->upgrade_to_release(release => '8.0.2',
##                              url => 'http://10.57.83.10//engineering/rlse/DOT/RboilermakerN/final/bedrock/export/x86_64.debug/tarball/image.tgz',
##                              reboot => FALSE,);
## $upgrade->revert_to_release(release => '8.0');
##
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::UpgradeRevert;

BEGIN
{
	require "Compression/base_import.pl";
}

use strict;
use warnings;

use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Tharn;

use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult);
use Compression::7Mode::UpgradeRevert;




## @name ::initialize
##
## @summary Compression::UpgradeRevert Constructor.
##
## @description
## Creates an Compression::UpgradeRevert object.
## It executes basic upgrade/revert initialization:
## - disables clustering;
## - installs required base release;
##
## @return Compression::UpgradeRevert object.
##
## @arg filer Required: Filer name.
##
## @arg vserver Optional: Vserver name, for c-mode testing (currently not supported).
##
## @arg current_release Required: Current release on the filer (at the time the constructor is called).
## It MUST be in the form x.y.z.
## It can be either passed in directly from the script or given as a command-line argument: CURRENT_RELEASE.
##
## @arg required_base_release Required: Base release to which the filer must be intialised.
## It MUST be in the form x.y.z.
## It can be either passed in directly from the script or given as a command-line argument: BASE_RELEASE.
##
## @arg base_release_url Optional: URL to be used when installing the base release.
## If not provided, it will default to /x/eng/rlse/DOT/<base_release>.
## It can be either passed in directly from the script or given as a command-line argument: BASE_REL_URL.
##
## @arg ndu_testing Optional: Flag overwrite the default upgrade type (if NDU should be performed or not).
## By default, the library will check if clustering is installed on the filer.
## If it is, then clustering will be enabled and the upgrade type will be set to NDU.
## If the filer is a stand-alone node, then a regular (non NDU) upgrade is performed.
## Setting this flag to TRUE means that NDU testing is required. If the flag is specifically set to TRUE
## and the filer is not part of a cluster, then the script will be terminated with a config error.
## Setting this flag to FALSE means that NDU testing is specifically avoided. This will cause the library to execute a regular upgrade
## (note that in this case, even if the filer is part of a cluster, the following will happen:
##  clustering will be disabled and only the specified filer will be upgraded. The partner node will be left running its current release).
##
## @arg disable_cleanup Optional: Flag to disable any clean-up
## that would be executed upon object destruction.
## Defaults to $main::DISABLE_CLEANUP.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to $main::DEBUG.
##
## @notes
## For some cases, going from the current release to the base release requires installing an intermediary release.
## For example, for current release BR1, base release BR0, the intermediary release will be RR0.
## For these cases, it is possible to pass in an URL to be used when instaling this intermediary release.
## This can be achieved by using the command-line argument INTERM_REL_URL.
## Example: INTERM_REL_URL='http://10.57.83.10//engineering/rlse/DOT/mainN_110126_1315/final/bedrock/export/x86_64.optimize/tarball/image.tgz'
##
## @example
## my $upgrade = Compression::UpgradeRevert->initialize(filer => $FILER,
##                                                     required_base_release => '8.0.0',
##                                                     base_release_url => 'http://10.57.83.10//engineering/rlse/DOT/RboilermakerN_100311_0300/final/bedrock/export/x86_64.optimize/tarball/image.tgz',);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub initialize
{
	my $class = shift;
	my %args = validate(@_,
	{
		filer => {type => SCALAR,},
                
                vserver => {type => SCALAR|UNDEF,
                            default => $main::VSERVER,},
                
                current_release => {type => SCALAR,
                                    regex => qr/^\d+\.\d+\.\d+$/,
                                    default => $main::CURRENT_RELEASE},
                
                required_base_release => {type => SCALAR,
                                          regex => qr/^\d+\.\d+\.\d+$/,
                                          default => $main::BASE_RELEASE},
                
                base_release_url => {type => SCALAR,
                                     default => $main::BASE_REL_URL,},
                
		ndu_testing => {type => SCALAR|UNDEF,
				default => $main::NDU_TESTING,},
                
                disable_cleanup => {type => BOOLEAN,
                                    default => $main::DISABLE_CLEANUP,},
                
                debug => {type => BOOLEAN,
                          default => $main::DEBUG,},
	} );
	
        my $self = {};
        $self->{_creator_pid} = $$;
        $self->{_boot_mode} = C_ONTAP_7MODE;
	if (defined($args{vserver}))
	{
		$self->{vserver} = $args{vserver};
		$self->{boot_mode} = C_ONTAP_CMODE;
	}
        
        if ($self->{_boot_mode} eq C_ONTAP_CMODE)
        {
                logresult ('script', "C-mode support for upgrade/revert is not available yet");
        }
        else
        {
                delete($args{vserver});
                $self->{_upg_rev} = Compression::7Mode::UpgradeRevert->new(%args);
        }
        
        
        bless $self, $class; 

	return $self;
}


## @name ::upgrade_to_release
## @summary Installs and downloads the required release.
## 
## @description
## Installs and downloads the required release. If required, reboots the filer.
##
## @arg release Required: Release to be installed.
## It MUST be in the form x.y.z.
##
## @arg url Optional: URL to be used for the software install.
## Default: $main::UPGRADE_URL (not defined, by default).
##
## @arg kernel_type Optional: Kernel type to be used for the software install (used for auto-generating the url, if not provided).
## Ignored if the url is passed in. Default: C_DFLT_KERNEL_TYPE (debug).
##
## @arg reboot Optional: Flag to decide if the filer will be rebooted after the install and download.
## In case of NDU testing, instead of reboot, the library will perform takeover/giveback.
## Default: TRUE.
##
## @arg client Optional: Client from which to execute the upgrade.
## Default: localhost.
##
## @arg site Optional: Site from which the upgrade is executed (used for auto-generating the url, if not provided).
## Ignored if the url is passed in. Default: svl.
##
## @return Hash with rc.
##  
## @examples
## $obj->upgrade_to_release(release => '8.1.0',
##                          url => 'http://10.57.83.10//engineering/rlse/DOT/mainN/final/bedrock/export/x86_64.debug/tarball/image.tgz');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub upgrade_to_release
{
        my $class = shift;
        
        return $class->{_upg_rev}->upgrade_to_release(@_);
}


## @name ::revert_to_release
## @summary Reverts to required release.
## 
## @description
## Reverts to required release. Attempts to deal with revert_to requirements, on a best-effort basis (remove snapshots, sis revert, etc).
##
## @arg release Required: Release to revert to.
## It MUST be in the form x.y.
##
## @arg force Optional: Flag to decide if force revert.
## Default: TRUE.
##
## @arg max_attempts Optional: Maximum number of retries.
## Default: 5.
##
## @arg expected_requirements Optional: Hashref of expected requirements that the library must encounter while attempting to execute revert_to.
## Legal values for hash keys are: DELETE_SNAPSHOTS, REVERT_SIS_METAFILES, SNAPMIRROR_OFF, DECOMPRESS_VOLUME, DISABLE_COMPRESSION.
## The corresponding hash values may be: either '' if the requirement is not expected for a specific entity (volume, aggregate, etc),
## or a specific entity name if the requirement must be encountered for a specific entity (ex: DISABLE_COMPRESSION must be specifically required for volume 'my_test_vol').
##
## @return Hash with rc.
##  
## @examples
## $obj->revert_to_release(release => '8.0');
## $obj->revert_to_release(release => '8.0',
##                         expected_requirements => {REVERT_SIS_METAFILES => '',
##                                                   DISABLE_COMPRESSION => $vol_name},);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub revert_to_release
{
        my $class = shift;
        
        return $class->{_upg_rev}->revert_to_release(@_);
}



1;
