#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Utils.pm#1 $
#

## @summary Compression utility module
##
## @description
##    This module provides utility functions required for compression work flow
##
## @synopsis
##
## use Compression::Constants;
## use Compression::Utils;
## 
## BEGIN { require "Compression/setup_core.pl"; }
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::Utils;

use strict;
use warnings;

use Tharn;
use Module::Common;
use Data::Dumper;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);

use Compression::Constants qw(:all);
use SayWrap qw(sayEOF);
use NACL::APISet;
use NACL::APISet::Exceptions::ResponseException qw(:try);

BEGIN {
        
        use vars ('@ISA','@EXPORT');
        require Exporter;
        @ISA = ('Exporter');
        @EXPORT = qw(process_data_size_string
                     detect_indirect_type_from_status
                     compare_values
		     convert_data_size_value
		     process_performance_info
		     compare_hash_entries
                     _c_interpret_logresult
		     _send_cmd_to_filer
		     _send_cmd_7mode
		     _send_cmd_Cmode);
}



## @name ::compare_values
## @summary Compares two values, within an allowed percentage of error.
## 
## @description
## Compares two values, within an allowed percentage of error.
## In case of failure, logs fatal result, unless non_fatal_failure is set to TRUE.
##
## 
## @return C_RC_SUCCESS/C_RC_FAILURE.
##
## @arg obtained_value Required: Value to be compared.
##
## @arg expected_value Required: Value to be compared against.
##
## @arg allow_percentage_error Optional: Allowed percentage of error. Defaults to 0.
##
## @arg values_meaning Required: String to be used to log an eventual failure.
##
## @arg non_fatal_failure Optional: Boolean to decide if an eventual failure should be fatal.
## Defaults to FALSE.
##
## @arg verbose Required: Boolean to decide the verbosity. Defaults to FALSE.
##
## @examples
## my $are_equal = compare_values(obtained_value => $obtained,
##                                expected_value => $expected,
##                                values_meaning => 'compression savings');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub compare_values
{
	my (%args) = validate( @_,
		{
			obtained_value => { type => SCALAR, regex => qr/^\d+$/, },
			expected_value => { type => SCALAR, regex => qr/^\d+$/, },
			allow_percentage_error => { type => SCALAR, regex => qr/^\d+$/, default => 0, },
			values_meaning => { type => SCALAR, },
			non_fatal_failure => { type => BOOLEAN, default => FALSE },
                        verbose => { type => BOOLEAN, default => FALSE },
		} );

	my $obtained = $args{obtained_value};
	my $expected = $args{expected_value};
	my $jitter = $args{allow_percentage_error};
	my $val_meaning = $args{values_meaning};
	my $non_fatal_failure = $args{non_fatal_failure};
        my $verbose = $args{verbose};
        
        logcomment("Compression Utils: compare_values input: " . Dumper(%args) ) if ($verbose);

	if (($jitter < 0) || ($jitter > 100))
	{
		logresult("script", "Illegal allow_percentage_error value: $jitter. Should be between 0 and 100:\n");
	}

	my $min = $expected - int($expected * $jitter / 100 );
	my $max = $expected + int($expected * $jitter / 100 );

	logcomment("Compress: _compare_values: obtained=$obtained, expected=$expected, min=$min, max=$max") if ($verbose);

        if (($obtained < $min) || ($obtained > $max))
        {
		if ($non_fatal_failure == TRUE)
		{
			return C_RC_FAILURE;
		}
		
                logresult("fatal", "Expected $val_meaning mismatch: obtained $obtained, expected value limits of min=$min, max=$max");
        }

	return C_RC_SUCCESS;
}


## @name ::process_data_size_string
## @summary Processes the data size input string.
## 
## @description
## Processes the data size input string to separate the data modifier and to convert the size into bytes.
## 
## @return Hash with modifier, size (with no modifier appended), size_in_bytes.
##
## @arg size_string Required: Data size string.
##
## @examples
## my %result_hash = process_data_size_string( size_string => '20m' );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub process_data_size_string
{
	my (%args) = validate( @_,
		{
			size_string => {type => SCALAR},
		} );
	
	my $size_string = $args{size_string};
	my %return_hash = ();
	my ($modifier, $size_in_bytes);
	
	# Extract modifier
	if ($size_string =~ /^\d+(\w+)$/)
	{
		$modifier = $1;
	}
	else
	{
		logresult("script", "Unsupported data size string: $size_string");
	}
	
	$size_string =~ s/$modifier$//;
	
	$return_hash{modifier} = $modifier;
	$return_hash{size} = $size_string;
	
	if (($modifier eq 'k') || ($modifier eq 'KB'))
	{
		$size_in_bytes = $size_string * KB;
	}
	elsif ( ($modifier eq 'm') || ($modifier eq 'MB'))
	{
		$size_in_bytes = $size_string * MB;
	}
	elsif (($modifier eq 'g') || ($modifier eq 'GB'))
	{
		$size_in_bytes = $size_string * GB;
	}
	else
	{
		logresult("script", "Unsupported modifier for data size ($size_string): $modifier");
	}
	$return_hash{size_in_bytes} = $size_in_bytes;
	
	return %return_hash;
}


## @name ::convert_data_size_value
## @summary Convert data size value according to modifier change.
## 
## @description
## Convert data size value according to modifier change.
## 
## @return Converted data size value.
##
## @arg value Required: Data size value.
##
## @arg current_modifier Required: Data size value.
##
## @arg required_modifier Required: Data size value.
##
## @examples
## my $converted_val = convert_data_size_value(value => 892, current_modifier => 'k', required_modifier => 'g');
##
sub convert_data_size_value
{
	my (%args) = validate( @_,
		{
			value => {type => SCALAR, regex => qr/^\d+$/,},
			current_modifier => {type => SCALAR, regex => qr/^b|k|m|g$/,},
			required_modifier => {type => SCALAR, regex => qr/^b|k|m|g$/,},
		});
	
	my $value = $args{value};
	my $current_modifier = $args{current_modifier};
	my $required_modifier = $args{required_modifier};
	
	if ($current_modifier eq $required_modifier)
	{
		return $value;
	}
	
	if ($current_modifier eq 'k')
	{
		$value = $value * KB;
	}
	elsif ($current_modifier eq 'm')
	{
		$value = $value * MB;
	}
	elsif ($current_modifier eq 'g')
	{
		$value = $value * GB;
	}
	
	if ($required_modifier eq 'k')
	{
		$value = int($value / KB);
	}
	elsif ($required_modifier eq 'm')
	{
		$value = int($value / MB);
	}
	elsif ($required_modifier eq 'g')
	{
		$value = int($value / GB);
	}
	
	return $value;
}


## @name ::detect_indirect_type_from_status
## @summary Detect the indirect type (32/64 -bit).
## 
## @description
## Parses the status string to detect the indirect type (32/64 -bit).
## 
## @return Indirect type (32/64).
##
## @arg status_string Required: Status string.
##
## @examples
## my $indirect_type = parse_status_string_for_indirect_type( status_string => $status );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
# Note: if an API will not be made available to us
# and if the status output changes in time,
# the logic in here should be moved to UI, to make it release-idependent.
# It is done here for now because NOTE provides half-way support.
sub detect_indirect_type_from_status
{
	my (%args) = validate( @_,
		{
			status_string => {type => SCALAR},
		} );
	
	my $status = $args{status_string};
	my $indirect_type = C_DFLT_AGGR_INDIRECT_TYPE; 
	
        if ($status =~ /.*(\d\d)-bit/)
        {
               $indirect_type = $1; 
        }
        else
        {
		# Note: conform documentation, we should always be able to detect the type here.
		# If this changes, we will want to only log a message here and return the default value.
                logresult("fatal", "Could not detect the indirect type in the status string \"$status\"");
        }
        
	return $indirect_type;
}


## @name ::process_performance_info
## @summary Process performance information.
##
## @description
## Process performance information. In case of failure, execute 'failure_handling' action.
##
## @return Hash with rc and duration.
##
## @arg duration Required: Duration of the action for which performance is checked.
##
## @arg description Required: Short description of the action for which performance is checked.
## Used for logging purposes.
##
## @arg failure_handling Optional: Flag to control how failures should be handled.
## Allowed values: none, fail (log fatal result),
## panic (panic the 'generate_cores_on' filers to generate cores), or panic_and_fail.
## Default: fail.
##
## @arg max_allowed_duration Optional: Maximum value allowed for the duration.
## If this argument is used, the obtained duration is compared against it and
## if it is bigger, then 'failure_handling' action is executed.
##
## @arg min_allowed_throughput Optional: Minimum value allowed for the throughput.
## If this argument is used, the obtained throughput is compared against it and
## if it is smaller, then 'failure_handling' action is executed.
##
## @arg processed_data_size Optional: If min_allowed_throughput is provided, then this
## parameter should also be provided, to specify the amount of data (in KB) that was processed.
##
## @arg generate_cores_on Optional: This parameter can be used to specify the filer names on which
## to generate cores in case of degraded performace. In case this is not provided,
## cores will be generated on the filer that corresponds to this object.
##
## @arg boot_mode Required: Ontap boot mode of the filer(s) to generate panic on.
##
## @example
## my %res = process_performance_info(duration => $duration,
##                                    description => "decompress scanner for volume $my_vol",
##                                    boot_mode => C_ONTAP_7MODE,
##                                    max_allowed_duration => 10,
##                                    min_allowed_throughput => 100,
##                                    processed_data_size => 2050,
##                                    failure_handling => 'panic',
##                                    generate_cores_on => [$FILER_SRC, $FILER_DST],);
##
sub process_performance_info
{
	my (%args) = validate( @_,
	{
		duration => {type => SCALAR|UNDEF,},
		description => {type => SCALAR,},
		boot_mode => {type => SCALAR,},
		max_allowed_duration => {type => SCALAR|UNDEF, optional => 1,},
		min_allowed_throughput => {type => SCALAR|UNDEF, optional => 1,},
		processed_data_size => {type => SCALAR|UNDEF, optional => 1,},
		failure_handling => {type => SCALAR, regex => qr/^none|fail|panic|panic_and_fail$/, default => 'fail',},
		generate_cores_on => {type => ARRAYREF, optional => 1,},
	});
	
	my $duration = $args{duration};
	my $max_allowed_duration = $args{max_allowed_duration};
	my $min_allowed_throughput = $args{min_allowed_throughput};
	my $processed_data_size = $args{processed_data_size};
	my $failure_handling = $args{failure_handling};
	my $generate_cores_on = $args{generate_cores_on};
	my $description = $args{description};
	my $boot_mode = $args{boot_mode};
	my $bad_performance = FALSE;
	my $bad_performance_reason = '';
	
	if (!defined($duration))
	{
		$bad_performance = TRUE;
		$bad_performance_reason = "Unable to detect the duration of operation \"$description\"";
	}
	else
	{
		logcomment("\nPerformance statistics for action:\n$description\n" .
			   "Duration: $duration\n");
		
		if ((defined($max_allowed_duration)) && ($duration > $max_allowed_duration))
		{
			$bad_performance = TRUE;
			$bad_performance_reason .= "\nPerformace was detected unacceptable for action:\n$description\n" .
			                           "Maximum allowed duration: $max_allowed_duration\n" .
						   "Actual duration: $duration\n";
		}
		
		if (defined($processed_data_size))
		{
			if ($duration == 0)
			{
				# round up to 1 sec
				$duration = 1;
			}
			my $obtained_throughput = $processed_data_size / $duration;
			logcomment("\nPerformance statistics for action:\n$description\n" .
				   "Processed data size: $processed_data_size\n" .
				   "Throughput: $obtained_throughput");
			
			if ((defined($min_allowed_throughput)) && ($obtained_throughput < $min_allowed_throughput))
			{
				$bad_performance = TRUE;
				$bad_performance_reason .= "\nPerformace was detected unacceptable for action:\n$description\n" .
				                           "Minimum allowed throughput: $min_allowed_throughput\n" .
							   "Actual throughput: $obtained_throughput\n";
			}
			
		}
		elsif (defined($min_allowed_throughput))
		{
			$bad_performance = TRUE;
			$bad_performance_reason .= "\nPerformace check failure for:\n$description\n" .
			                           "Minimum allowed throughput was required to be checked: $min_allowed_throughput\n" .
						   "But the size of the processed data could not be detected\n";
		}
	}
	
	if ($bad_performance == TRUE)
	{
		if ($failure_handling eq 'fail')
		{
			logresult('fatal', $bad_performance_reason);
		}
		elsif ($failure_handling =~ /panic/)
		{
			if (!defined($generate_cores_on))
			{
				logcomment("$bad_performance_reason\n\nCannot generate cores. No filer names were provided");
			}
			else
			{
				logcomment("$bad_performance_reason\n\nGenerating core files...");
				foreach my $filer (@$generate_cores_on)
				{
					logcomment("Issuing panic on filer $filer");
					if ($boot_mode eq C_ONTAP_CMODE)
					{
						my $conn = NOTE::GX::Node->connect(name => $filer,);
						$conn->panic();
					}
					else
					{
						my $conn = NOTE::Filer->connect(host => $filer,);
						$conn->panic();
						$conn->transit_to('UP');
					}
				}
			}
			
			if ($failure_handling eq 'panic_and_fail')
			{
				logresult('fatal', $bad_performance_reason);
			}
		}
		else
		{
			# none
			logcomment($bad_performance_reason);
		}
	}
	
	return (rc => C_RC_SUCCESS, duration => $duration,);	
}


## @name ::compare_hash_entries
## @summary Compare specific hash entries.
##
## @description
## Compare specific hash entries.
##
## @return TRUE/FALSE.
##
## @arg obtained Required: Hashref of obtained results.
##
## @arg expected Required: Hashref of expected results.
##
## @arg values_meaning Required: Short description of what the results hash represents
## (Used for logging purposes).
##
## @arg fatal_failure Optional: Flag to control if a mismatch is considered fatal.
## Default: TRUE.
##
## @example
## Compression::Utils::compare_hash_entries(obtained => $current_stats,
##                                          expected => $expected_values,
##                                          values_meaning => "volume stats for $volume");
##
sub compare_hash_entries
{
	my %args = validate( @_,
	{
		obtained => {type => HASHREF,},
		expected => {type => HASHREF,},
		values_meaning => {type => SCALAR,},
		fatal_failure => {type => SCALAR, default => TRUE,},
	});
	
	my $obtained = $args{obtained};
	my $expected = $args{expected};
	my $values_meaning = $args{values_meaning};
	my $fatal_failure = $args{fatal_failure};
	
	foreach my $field (keys %$expected)
	{
		if (!exists($obtained->{$field}))
		{
			logresult('script', "Cannot compare $values_meaning:\n" .
				            "Field \"$field\" is not available.\n" .
					    "Available fields are:\n" . Dumper(keys(%$obtained)));
		}
		
		my $expected_val = $expected->{$field};
		my $obtained_val = $obtained->{$field};
		logcomment("Checking $values_meaning\n" .
			   "Field: $field\n" .
			   "Expected value: $expected_val\n" .
			   "Obtained value: $obtained_val");
		
		if ("$expected_val" ne "$obtained_val")
		{
			my $message = "Mismatch in $values_meaning:\n" .
			              "Field: $field\n" .
				      "Expected value: $expected_val\n" .
				      "Obtained value: $obtained_val";
				      
			if ($fatal_failure == TRUE)
			{
				logresult('fatal', $message);
			}
			
			logcomment($message);
			return FALSE;
		}
	}
	
	return TRUE;
}


# Subroutines internal to the compression library (to be used only within the library)
sub _c_interpret_logresult
{
	my $message_string = shift;
	
	if ($message_string =~ /^NOT APPLICABLE: /)
        {
                logresult('NA', "$message_string");
        }
        elsif($message_string =~ /^CONFIG ERROR: /)
        {
                logresult('config', "$message_string");
        }
        else
        {
                logresult('fatal', "$message_string");
        }
}


sub _send_cmd_to_filer
{
	my (%args) = validate( @_,
		{
			conn => {type => OBJECT},
			boot_mode => {type => SCALAR,},
			cmd => {type => SCALAR|OBJECT},
			timeout => {type => SCALAR|UNDEF, optional => 1,},
			need_persistent_connection => {type => SCALAR, default => FALSE},
			waitstring => {type => SCALAR, optional => 1},
			replaystring => {type => SCALAR, optional => 1},
                        node_run => {type => BOOLEAN, default => FALSE},
                        maxbufsize => {type => SCALAR, optional => 1},
                        vfiler => {type => SCALAR|UNDEF, optional => 1,},
                        vfiler_available => {type => BOOLEAN, default => FALSE},
                        panic_expected => { type=>SCALAR, optional=>1, default => FALSE },
			zapi_available => {type => BOOLEAN, default => FALSE},
		} );
	
	my $boot_mode = $args{boot_mode};
	delete($args{boot_mode});
	
	if (($args{need_persistent_connection} == TRUE) && (!(defined($args{waitstring})) ||
						!(defined($args{replaystring}))))
	{
		logresult('script', "The waitstring and replystring arguments need to be passed in if a persistent connection is needed");
	}
	
	if ($boot_mode eq C_ONTAP_7MODE)
	{
		_send_cmd_7mode(%args);
	}
	elsif ($boot_mode eq C_ONTAP_CMODE)
	{
		_send_cmd_Cmode(%args);
	}
	else
	{
		logresult('script', "Library problem found: unrecognized boot_mode passed in: \"$boot_mode\"");
	}
	
}

sub _send_cmd_7mode
{
	my (%args) = validate( @_,
		{
			conn => {type => OBJECT, isa => "NOTE::Filer"},
			cmd => {type => SCALAR|OBJECT},
			timeout => {type => SCALAR|UNDEF, optional => 1,},
			need_persistent_connection => {type => SCALAR, default => FALSE},
			waitstring => {type => SCALAR, optional => 1},
			replaystring => {type => SCALAR, optional => 1},
                        node_run => {type => BOOLEAN, default => FALSE}, # Not used for 7-mode
                        maxbufsize => {type => SCALAR|UNDEF, optional => 1},
                        vfiler => {type => SCALAR|UNDEF, optional => 1,},
                        vfiler_available => {type => BOOLEAN, default => FALSE},
                        panic_expected => { type=>SCALAR, optional=>1, default => FALSE },
			zapi_available => {type => BOOLEAN, default => FALSE}, 
		} );
	
	# Notes:
	# - timeout is only considered for now for non-persistent connections
	# - waitstring/replaystring are only considered for persistent connections
	my $conn = $args{conn};
	my $filer_name = $conn->host();
	my $cmd = $args{cmd};
	my $timeout = $args{timeout};
	my $need_persistent_conn = $args{need_persistent_connection};
	my $waitstr_if_persistent_conn = $args{waitstring};
	my $replay_if_persistent_conn = $args{replaystring};
        my $maxbufsize = $args{maxbufsize};
	my $vfiler = $args{vfiler};
	my $vfiler_available = $args{vfiler_available};
	my $zapi_available = $args{zapi_available};
        my $panic_expected = $args{panic_expected};
	my $output;
	
	my $ui_type = (defined($main::UI_TYPE)) ? $main::UI_TYPE : C_DFLT_UI_TYPE;
	if (($ui_type eq C_UI_ZAPI) && ($zapi_available == TRUE))
	{
		my $Host = host($filer_name);
		my $Api_Set_Obj = NACL::APISet->new(hostobj   => $Host,
						    category  => "Node",
						    interface => "ZAPI",
						    set       => "7Mode",
						);
		
		try
		{
			$output = $Api_Set_Obj->execute_raw_command(command => $cmd,'return-response-object' => 1);
		}
		catch NACL::APISet::Exceptions::ResponseException with
		{
			my $exception_object = shift;
			my $resp = $exception_object->get_response_object();
			$output = $resp->get_raw_output();
		}
		
		logcomment("ZAPI command:\n$cmd");
		# Bail out here - no extra cases handled for zapi
		return $output; 
	}
	
	if (defined($vfiler) && ($vfiler_available == TRUE))
	{
		# Detect needed privilege, to send with the vfiler run command
		my $privilege = '';
		if ($cmd =~ /(priv\s+set\s+.*;)/)
		{
			$privilege = $1;
			$cmd =~ s/$privilege//;
		}
		
		$cmd = "$privilege vfiler run $vfiler $cmd";
	}
	
        if ($panic_expected == TRUE)
        {
                my $conn_rec = $conn->conn();
                (undef, undef, $output) = say($conn_rec, $cmd, '-blind');
        }
        elsif (defined($maxbufsize))
        {
                # Use say directly, SayWrap does not allow changing the buffer size
                my $conn_rec = $conn->conn();
                (undef, undef, $output) = say($conn_rec, $cmd, "-maxbufsize", $maxbufsize);
        }
	elsif ($need_persistent_conn == FALSE)
	{
		if (defined($timeout) && ($timeout =~ /^\d+$/) && ($timeout > 0))
		{
			# NOTE::Filer:say does not correctly send the timeout to sayEOF
			# Until this is fixed, use sayEOF directly
			my $conn_rec = $conn->conn();
			$output = sayEOF($conn_rec, $cmd, {timeout => $timeout} );
		}
		else
		{
                        $output = $conn->say($cmd);
		}
	}
	else
	{
		my $filer_name = $conn->host();
		host("$filer_name");
                my $conn_tel = connect("#telnet");
		
                my (undef, undef, $output) = say($conn_tel, $cmd,'-noprompt', '-wait', $waitstr_if_persistent_conn);
                my ($match, $before, $after, $prompt) = say($conn_tel, $replay_if_persistent_conn);
                $output = $match;
	}
	
	if (defined($vfiler))
	{
		$output =~ s/^\s*=+\s+$vfiler\n*//;
	}
		
	return $output;
}

sub _send_cmd_Cmode
{
	my (%args) = validate( @_,
		{
			conn => {type => OBJECT, isa => "NOTE::GX::Node"},
			cmd => {type => SCALAR},
			timeout => {type => SCALAR|UNDEF, optional => 1,},
			need_persistent_connection => {type => SCALAR, default => FALSE},
			waitstring => {type => SCALAR, optional => 1},
			replaystring => {type => SCALAR, optional => 1},
                        node_run => {type => BOOLEAN, default => FALSE},
                        maxbufsize => {type => SCALAR|UNDEF, optional => 1},
                        vfiler => {type => SCALAR|UNDEF, optional => 1,}, # Not used for C-mode
                        vfiler_available => {type => BOOLEAN, default => FALSE}, # Not used for C-mode
                        panic_expected => { type=>SCALAR, optional=>1, default => FALSE },
			zapi_available => {type => BOOLEAN, default => FALSE}, 
		} );
	
	my $conn = $args{conn};
	my $cmd = $args{cmd};
	my $filer_name = $conn->name();
	my $timeout = defined($args{timeout}) ? $args{timeout} : C_DFLT_CMD_TIMEOUT;
	my $need_persistent_conn = $args{need_persistent_connection};
	my $waitstr_if_persistent_conn = $args{waitstring};
	my $replay_if_persistent_conn = $args{replaystring};
        my $need_node_run = $args{node_run};
	my $zapi_available = $args{zapi_available};
        my $panic_expected = $args{panic_expected};
	my $output;
	
	my $ui_type = (defined($main::UI_TYPE)) ? $main::UI_TYPE : C_DFLT_UI_TYPE;
	if (($ui_type eq C_UI_ZAPI) && ($zapi_available == TRUE))
	{
		my $Host = host($filer_name);
		my $Api_Set_Obj = NACL::APISet->new(hostobj   => $Host,
						    category  => "Node",
						    interface => "ZAPI",
						    set       => "CMode",
						   );
		
		my %execute_args = ('return-response-object' => 1,);
		my $comment = "ZAPI command:\n";
		
		# If the command needs to be executed in vserver context, we need to process it differently
		if ($cmd =~ /<vserver>(\S+)<\/vserver>/)
		{
			my $vserver_name = $1;
			$execute_args{'vserver-override-name'} = $vserver_name;
			$cmd =~ s/<vserver>$vserver_name<\/vserver>//;
			$comment .= "Vserver context: $vserver_name\n";
		}
		
		$execute_args{command} = $cmd;
		$comment .= $cmd;
		
		try
		{
			$output = $Api_Set_Obj->execute_raw_command(%execute_args);
		}
		catch NACL::APISet::Exceptions::ResponseException with
		{
			my $exception_object = shift;
			my $resp = $exception_object->get_response_object();
			$output = $resp->get_raw_output();
		}
		
		logcomment($comment);
		
		# Bail out here - no extra cases handled for zapi
		return $output; 
	}
        
        # TODO
        if ($panic_expected == TRUE)
        {
                logresult('script', "panic_expected is not supported on c-mode yet");
        }
	
	# TODO:
	# - maxbufsize ?
	if ($need_node_run == TRUE)
        {
                eval
                {
                        $output = $conn->server->system_node_run(node => 'local',
                                                                 command => "\"$cmd\"",);
                };
                
                if ($@)
                {
                        $output .= $@;
                }
        }
        else
        {
                my %cli_args = (cmd => $cmd,
                                -Ignore_errors => 1,);
                if (defined($timeout) && ($timeout =~ /^\d+$/) && ($timeout > 0))
                {
                        $cli_args{'-Timeout'} = $timeout;
                }
                
                $output = $conn->server->cli(%cli_args);
        }
	
	return $output;
}



1;
