#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Version.pm#1 $
#

## @summary Module to provide versioning capabilities for the Compression modules.
##
## @description
##    Provides ontap and wafl versioning capabilites
##    (s/w features supportability and others).
##    Although this module can be used independently, its purpose is to support the Compression modules 
##
## @synopsis
##
## use Compression::Version;
## use vars qw($FILER); 
## 
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $filer = NOTE::Filer->connect( host => $FILER );
## my $obj = Compression::Version->new( conn => $filer );
## $obj->refresh_sw_version( conn => $filer );
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::Version;

use strict;
use warnings;

use Data::Dumper;
use Module::Common;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);

use Tharn;
use Compression::Constants qw(:all);


BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);

};

## @name ::new
##
## @summary Compression::Version Constructor.
##
## @description
## Creates a Compression::Version object.
##
## @return Object of type Compression::Version.
##
## @arg conn Required: NOTE::Filer object corresponding to the hosting filer.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
## my $obj = Compression::Version->new( conn => $filer, debug => TRUE );
##
## @author sureshb@netapp.com, dl-compression-qa@netapp.com
##

sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
		conn 	=> { 
			type => OBJECT,
			},
		
		debug 	=> {
                        type => SCALAR,
			optional => 1,
			default => FALSE,
			},
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default => C_ONTAP_7MODE},
		} );
	
	my $self = {};
        
        $self->{_debug} = $args{debug};
        $self->{_boot_mode} = $args{boot_mode};
        
        # Placeholders for ontap and wafl version strings
        $self->{ontap} = "";
        $self->{wafl} = "";

	bless $self, $class;
        
        $self->refresh_sw_version( conn => $args{conn} );
        
	return $self;
}


## @name ::refresh_sw_version
## @summary Gets the sw version (ontap, wafl) version from the filer.
##
## @description
## Gets the sw version (ontap, wafl) version from the filer.
##
## @return
## Hash with ontap_ver and wafl_ver.
##
## @notes
## This subroutine also updates the ontap and wafl version values of the object.
##
## @examples
## $obj->refresh_sw_version( conn => $filer );
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##

sub refresh_sw_version
{
	my $class = shift;
        my %args = validate(@_,
	{
		conn 	=> { 
			type => OBJECT,
			},
		} );
        
	my %sw = ();

        my $conn = $args{conn};
        my $boot_mode = $class->{_boot_mode};
        my $filer_name;
        
        if ($boot_mode eq C_ONTAP_7MODE)
        {
                if (!($conn->isa("NOTE::Filer")))
                {
                        logresult('script', "For 7-mode, the conn argument is required to be a NOTE\:\:Filer object");
                }
                $filer_name = $conn->host();
                
                $sw{ontap_ver} = $conn->version();
                $class->{ontap} = $sw{ontap_ver};
                
                $sw{wafl_ver} = $conn->wafl_version();
                $class->{wafl} = $sw{wafl_ver};
        }
        elsif ($boot_mode eq C_ONTAP_CMODE)
        {
                if (!($conn->isa("NOTE::GX::Node")))
                {
                        logresult('script', "For C-mode, the conn argument is required to be a NOTE\:\:GX\:\:Node object");
                }
                
                my $build_version = $conn->build_version();
                # Only keep the major release string
                $build_version =~ s/^.*__//;
                $build_version =~ s/\D*$//;
                if ($build_version =~ /\d+\.\d+\.\w+$/)
                {
                        $build_version =~ s/\.\w+$//;
                }
		
		if ($build_version =~ /\d+\.\d+RC.+$/)
		{
			$build_version =~ s/RC.+$//;
		}
		
		if ($build_version =~ /\d+\.\d+X\d+$/)
		{
			$build_version =~ s/X\d+$//;
		}
                
                if ($build_version !~ /^\d+\.\d+$/)
                {
                        logresult('script', "Could not substract the major release string from the build version: $build_version");
                }
                $filer_name = $conn->name();
                
                $sw{ontap_ver} = $build_version;
                $class->{ontap} = $build_version;
                
                # TODO: wafl version for c-mode?
                #$sw{wafl_ver} = 
                #$class->{wafl} = 
        }
        else
        {
                logresult('script', "Unrecognized boot mode: $boot_mode");
        }
	
	
	logcomment("SW version of filer \"$filer_name\" is:\n". Dumper(%sw) ."\n");
	return %sw;
}


## @name ::is_feature_supported
## @summary Checks if a specific feature is supported in the current release.
##
## @description
## Checks if a specific feature is supported in the current release.
##
## @arg feature Required: Feature code to be checked for (codes are defined in Compression::Constants).
##
## @return
## TRUE/FALSE (defined in Constants.pm).
##
## @examples
## my $is_supported = $obj->is_feature_supported(feature => C_FEATURE_PRE_CP);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##

sub is_feature_supported
{
	my $class = shift;
        my %args = validate(@_,
	{
		feature => {type => SCALAR,},
	} );
        
	my $feature = $args{feature};

        my $ontap = $class->{ontap};
        my $boot_mode = $class->{_boot_mode};
	
	logcomment("Checking the \"$feature\" feature for ontap version: $ontap") if ($class->{_debug});
	
        # For now only use ontap version to decide
        # WARNING: Any entry in here MUST have the default value defined
        my $features = {};
        
        if ($boot_mode eq C_ONTAP_CMODE)
        {
                $features = {
                             'PRE_CP' => {'default' => TRUE,
                                          '7.3' => FALSE,
                                          '8.0' => FALSE,
                                          },
                             'LA_IPU' => {'default' => TRUE,
                                          '7.3' => FALSE,
                                          '8.0' => FALSE,
                                          },
                             'DECOMPRESS_REVERT' => {'default' => FALSE,
                                                     '8.0' => TRUE,
                                                     },
                             'WAFLIRON_WHILE_DECOMPRESS' => {'default' => FALSE,
                                                             '7.3' => TRUE,
                                                             '8.0' => TRUE,
                                                             },
                             'INLINE_COMPRESSION_ON_PARTIAL_CG' => {default => FALSE,
                                                                    '7.3' => TRUE,
                                                                    '8.0' => TRUE,
                                                                    },
                             'SNAPLOCK' => {default => FALSE,
                                           },
			     'CTIME_FILE_CLONE' => {'default' => TRUE,
						    },
			     'NON_CTIME_FILE_CLONE' => {'default' => FALSE,
							},
			     'IDD' => {'default' => TRUE,
				       '8.1' => FALSE,
				       },
			     'BCE' => {'default' => TRUE,
				       '8.1' => FALSE,
				       },
			     'VBN_ZERO' => {'default' => TRUE,
					    '8.1' => FALSE,
					    },
			     };
        }
        else
        {
                $features = {
                             'PRE_CP' => {'default' => TRUE,
                                          '7.3' => FALSE,
                                          '8.0' => FALSE,
                                          },
                             'LA_IPU' => {'default' => TRUE,
                                          '7.3' => FALSE,
                                          '8.0' => FALSE,
                                          },
                             'DECOMPRESS_REVERT' => {'default' => FALSE,
                                                     '8.0' => TRUE,
                                                     },
                             'WAFLIRON_WHILE_DECOMPRESS' => {'default' => FALSE,
                                                             '7.3' => TRUE,
                                                             '8.0' => TRUE,
                                                             },
                             'INLINE_COMPRESSION_ON_PARTIAL_CG' => {default => FALSE,
                                                                    '7.3' => TRUE,
                                                                    '8.0' => TRUE,
                                                                    },
                             'SNAPLOCK' => {default => TRUE,
                                            },
			     'CTIME_FILE_CLONE' => {'default' => TRUE,
						    '7.3' => FALSE,
						    '8.0' => FALSE,
						    },
			     'NON_CTIME_FILE_CLONE' => {'default' => FALSE,
							'7.3' => TRUE,
							'8.0' => TRUE,
							},
			     'IDD' => {'default' => TRUE,
				       '7.3' => FALSE,
				       '8.0' => FALSE,
				       '8.1' => FALSE,
				       },
			     'BCE' => {'default' => TRUE,
				       '7.3' => FALSE,
				       '8.0' => FALSE,
				       '8.1' => FALSE,
				       },
			     'VBN_ZERO' => {'default' => TRUE,
					    '7.3' => FALSE,
					    '8.0' => FALSE,
					    '8.1' => FALSE,
					    },
                             };
        }
        
        if (!exists($features->{$feature}))
        {
                logresult('script', "Feature $feature is not maintained by this library");
        }
        
        my $availability_hash = $features->{$feature};
        if (defined($availability_hash->{$ontap}))
        {
                return $availability_hash->{$ontap};
        }
        
        return $availability_hash->{default};
}


## @name ::check_for_required_features
## @summary Checks if all the required features are supported in the current release.
##
## @description
## Checks if all the required features are supported in the current release.
## Logs N/A if any of the features in the list is not available.
##
## @arg features Required: Arrayref of feature codes to be checked for
## (codes are defined in Compression::Constants).
##
## @examples
## $obj->check_for_required_features(features => [C_FEATURE_PRE_CP]);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##

sub check_for_required_features
{
	my $class = shift;
        my %args = validate(@_,
	{
		features => {type => ARRAYREF,},
	} );
        
	my $required_features = $args{features};

        foreach my $feature (@$required_features)
        {
                my $is_supported = $class->is_feature_supported(feature => $feature);
                if ($is_supported != TRUE)
                {
                        logresult ('NA', "NOT APPLICABLE: Required feature \"$feature\" is not supported on this release (" . $class->{_boot_mode} . ")");
                }
        }
        
        return C_RC_SUCCESS;
}



# @name ::is_compression_supported
#
# @synopsis check if compression support
#
# @description
# check if compression feature is supported on the specified sw release version
#
# Returns boolean TRUE/FALSE
#
# @synopsis
# use Compression::Version;
# use vars qw($FILER $ENABLE); 
# 
# BEGIN { require "Compression/setup_core.pl"; }
# 
# my $obj = Compression::Version;
# $obj->is_compression_supported();
#
# @author sureshb@netapp.com, dl-compression-qa@netapp.com
#

#sub is_compression_supported
#{
#	my $class = shift;
#	my (%args) = @_;
#	$class->Detect_Sw_Version( conn => $args{host} ) 
#		if( !defined $class->{_ver} );
#
#	my $msg;
#	my $ret = Filer::Version::supported_feature( 
#			$args{conn}->hostp(), COMPRESSION, \$msg );
#	logresult("config", "$msg") if( !defined $ret or !$ret );
#
#	return 1;
#}



1;
