#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/VolClone.pm#1 $
#

## @summary Compression routines specific to clone volumes.
##
## @description
## Compression routines specific to clone volumes.
## The VolClone class inherits from the Compress class.
##
## @examples
##
## use Compression::Constants;
## use vars qw($FILER $AGGREGATE $VOLUME $CLIENT);
##
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $obj = Compression::Compress->setup(
##       filer => { name => $FILER, },
## 	aggregate => { name => $AGGREGATE },
##      volume => { name => $VOLUME, size => "2g", compress_enabled => FALSE, },
##      );
## my $clone = $obj->clone_volume(clone_name => 'clone_vol');
## $clone->is_enabled();
## $clone->get_savings();
## $clone->clone_verify();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::VolClone;

use strict;
use warnings;

use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Compression::Compress;
use Compression::Constants qw(:all);

our @ISA = ("Compression::Compress");

BEGIN
{
	require "Compression/base_import.pl";

}



## @name ::new 
##
## @summary Compression::VolClone constructor.
##
## @description
## Compression::VolClone constructor.
##
## @return Compression::VolClone object
##
## @arg source Required: Compression parent object.
## 
## @arg name Optional: Name of the clone. If not provided, it will be auto-generated.
##
## @arg snapshot Optional: Name of existing snapshot of parent volume, to serve
## as the basis of the clone volume.
##
## @arg space_reservation Optional: Space reservation setting for the clone volume
## Legal values: none/file/volume. 
##
## @examples
## my $clone = Compression::VolClone->new(source => $compr_obj,
##					  name => $args{clone_name});
##
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub new
{
	my $class = shift;
	my (%args) = validate( @_,
		{
		source => {
			   type => OBJECT,
			   isa  => "Compression::Compress",
			   },

		name => {
			 type => SCALAR|UNDEF,
			 optional => 1,
			 },
		
		snapshot => {type => SCALAR|UNDEF,
			     optional => 1,
			     },
		
		space_reservation => {type => SCALAR|UNDEF,
				      optional => 1,
				      },
	});

	# Start building our object:
	my $self 	= {};
	my $parent = $args{source};
	$self->{_debug} = $parent->{_debug};
	$self->{_disable_cleanup} = $parent->{_disable_cleanup};
	$self->{_creator_pid} = $$;
	$self->{_client}->{name} = $parent->{_client}->{name};
	$self->{_client}->{type} = $parent->{_client}->{type};
	$self->{filer} = $parent->{filer}; 
	$self->{host_name} = $parent->{host_name};
	$self->{_ui} = $parent->{_ui};
	$self->{aggregate} = $parent->{aggregate};
	$self->{aggr_name} = $parent->{aggr_name};
	$self->{boot_mode} = $parent->{boot_mode};
        $self->{vserver} = $parent->{vserver};
	$self->{_count_volumes} = $parent->{_count_volumes};
	
	my %create_args = ();
	my $clone;
	my $conn = $self->{filer}->{_c_filer}->{_conn};
	
	if ($self->{boot_mode} eq C_ONTAP_CMODE)
	{
		# GNOTE does not have clone support for c-mode yet
		my $parent_underlying_obj = $parent->{volume}->_get_underlying_vol_obj();
		$create_args{flexclone} = (defined($args{name})) ? $args{name} : $parent_underlying_obj->newName();
		$create_args{vserver} = $self->{vserver};
		$create_args{'parent-volume'} = $parent->{vol_name};
		$create_args{foreground} = 'true';
		if (defined($args{snapshot}))
		{
			$create_args{'parent-snapshot'} = $args{snapshot};
		}
		else
		{
			# In C-mode, the snapshot is a required field
			# Create a snapshot now
			$create_args{'parent-snapshot'} = '_snap_' . $create_args{flexclone};
			$parent_underlying_obj->snapshot_create($create_args{'parent-snapshot'});
			sleep(10);
			
		}
		
		if (defined($args{space_reservation}))
		{
			$create_args{'space-guarantee'} = $args{space_reservation};
		}
		
		logcomment("Licensing FlexClone");
		$parent_underlying_obj->Cluster->license('FlexClone', 1);
		
		$parent_underlying_obj->server->volume_clone_create(%create_args);
		
		$clone = NOTE::GX::Volume->attach(name => $create_args{flexclone},
						  Node => $conn);
		
		# Save the status info necessary for clone_verify
		my $parent_status = $parent_underlying_obj->status();
		$self->{_parent_size} = $parent_status->{size};
		$self->{_parent_used} = $parent_status->{used};
		$self->{_parent_name} = $create_args{'parent-volume'};
		$self->{_base_snapshot} = $create_args{'parent-snapshot'};
	}
	else
	{
		$create_args{name} = $args{name};
		if (defined($args{snapshot}))
		{
			$create_args{snapname} = $args{snapshot};
		}
		
		if (defined($args{space_reservation}))
		{
			$create_args{spcres} = $args{space_reservation};
		}
	
		$clone = $parent->{volume}->{_c_volume}->{_vol}->clone(%create_args);
	}
	
	$self->{volume} = Compression::Volume->new(filer => $conn,
						   aggregate => $self->{aggregate}->{_c_aggr}->{_aggr},
						   volume => {name => $args{name},},
						   vserver => $self->{vserver},
						   boot_mode => $self->{boot_mode},
						   attach => $clone,
						   ui => $parent->{_ui},
						   client => $parent->{_client}->{name},
						   debug => $self->{_debug},
						   disable_cleanup => $self->{_disable_cleanup},
						   );

	# Not checking the volume type here.
	# NOTE does it for us before creating the clone.
	
	$self->{vol_name} = $self->{volume}->{vol_name};
	$self->{_vol_mount_path} = $self->{volume}->get_mount_path();
	logcomment("Created volume clone object for volume \"" . $self->{vol_name}) if ($self->{_debug});
	
	if (defined($parent->{supportability}))
	{
		$self->{supportability} = Compression::Supportability->new(conn => $conn,
									   boot_mode => $self->{boot_mode},
									   ui => $self->{_ui},
									   debug => $self->{_debug},);
	}
	elsif (defined($parent->{_supportability}))
	{
		$self->{_supportability} = Compression::Supportability->new(conn => $conn,
									    boot_mode => $self->{boot_mode},
									    ui => $self->{_ui},
									    debug => $self->{_debug},);
	}

	$self->{dataset}  = Compression::Dataset->new(volume => $self->{volume}->{_c_volume}->{_vol},
						      client => $self->{_client}->{name},
						      boot_mode => $self->{boot_mode},
						      ui => $self->{_ui} );
	
	$self->{dataset}->attach_dataset_info(src_vol => $parent);
	
	$self->{decompress} = Compression::Decompress->new(conn => $conn,
							   volume => $self->{vol_name},
							   vserver => $self->{vserver},
							   boot_mode => $self->{boot_mode},
							   ui => $self->{_ui},
							   disable_cleanup => $self->{_disable_cleanup},
							   debug => $self->{_debug},
							   supportability => $self->{_supportability},);
	
	$self->{compress_scanner} = Compression::CompressionScanner->new(conn => $conn,
									 volume => $self->{vol_name},
									 vserver => $self->{vserver},
									 boot_mode => $self->{boot_mode},
									 ui => $self->{_ui},
									 disable_cleanup => $self->{_disable_cleanup},
									 debug => $self->{_debug},
									 supportability => $self->{_supportability},);
	
	$self->{_utilities} = Compression::CompressionUtilities->new(conn => $conn,
								     volume => $self->{vol_name},
								     vserver => $self->{vserver},
								     boot_mode => $self->{boot_mode},
								     ui => $self->{_ui},
								     disable_cleanup => $self->{_disable_cleanup},
								     debug => $self->{_debug},
								     supportability => $self->{_supportability},);

	$self->{sis} = Compression::Sis->new(filer => $conn,,
					     volume_path => $self->{volume}->get_mount_path(),
					     volume => $self->{volume}->_get_underlying_vol_obj(),
					     vserver => $self->{vserver},
					     boot_mode => $self->{boot_mode},
					     ui => $self->{_ui},
					     debug => $self->{_debug},
					     disable_cleanup => $self->{_disable_cleanup});
	$self->{sis}->{_initialized} = $parent->{sis}->{_initialized};


        if (defined($parent->{file_folding}))
        {
		$self->{file_folding} = Compression::FileFolding->new(filer => $conn,
								      volume => $self->{vol_name},
								      vserver => $self->{vserver},
								      boot_mode => $self->{boot_mode},
								      ui => $self->{_ui},
								      debug => $self->{_debug});
        }
	
	$self->{perf_stats} = $parent->{perf_stats};
	
	bless($self, $class);

	logcomment("VolClone: Setup Complete..\n\n") if ($self->{_debug});
	return $self;

}


## @name ::clone_verify
## @summary Verifies this clone of the Compress object.
##
## @description
## Verifies this clone of the Compress object. Wrapper around NOTE::Volume::Clone::verify.
## It will verify if the volume is a clone: if parent and clone have same hosting aggregate,
## if base snapshot is un-removable, if snap sched is the same, if parentvol cannot be destroyed,
## if the data in parent and clone are the same.
##
## @return Hash with rc and output.
##
## @examples
## $obj->clone_verify();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub clone_verify
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		# None
	} );
	
	my $cloned_vol = $class->{volume}->_get_underlying_vol_obj();
	if ($class->{boot_mode} eq C_ONTAP_CMODE)
	{
		# Verify what Volume::Clone is verifying in 7-mode
		#- if parent and clone have same hosting aggregate
		#- if snap sched is the same
		#- if the data in parent and clone are same
		# TODO ?
		#- if base snapshot is un-removable
		#- if parentvol cannot be destroyed
		my $cloned_vol_status = $cloned_vol->server->volume_clone_show(flexclone => $class->{vol_name},
									       vserver => $class->{vserver});
		
		if ((($class->{aggr_name}) ne ($cloned_vol_status->[0]->{aggregate})) ||
		    # TODO: these fail: bug?
		    #(($class->{_parent_size}) != ($cloned_vol_status->[0]->{size})) ||
		    #(($class->{_parent_used}) != ($cloned_vol_status->[0]->{used})) ||
		    (($class->{_parent_name}) ne ($cloned_vol_status->[0]->{'parent-volume'})) ||
		    (($class->{_base_snapshot}) ne ($cloned_vol_status->[0]->{'parent-snapshot'})))
		{
			logresult('fatal', "Clone verify failed:\n" .
				           "Parent aggregate name: " . $class->{aggr_name} . "\n" .
					   "Parent volume size: " . $class->{_parent_size} . "\n" .
					   "Parent volume used space: " . $class->{_parent_used} . "\n" .
					   "Base snapshot: " . $class->{_base_snapshot} . "\n" .
					   "Cloned volume status:\n" . Dumper($cloned_vol_status));
		}
	}
	else
	{
		$cloned_vol->verify();
	}
	
	return (rc => C_RC_SUCCESS, output => '');
}


## @name ::clone_split
## @summary Split this clone of the Compress object.
##
## @description
## Split this clone of the Compress object.
##
## @arg background Optional: Wait for split to complete in the background.
## Default: FALSE.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Defaults to C_RC_SUCCESS
##
## @return Hash with rc and output.
##
## @examples
## $obj->clone_split();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub clone_split
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		background => { type => BOOLEAN, optional => 1, default => FALSE },
		expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
	} );
	
	my $background = $args{background};
	my $cloned_vol = $class->{volume}->_get_underlying_vol_obj();
	
	if ($class->{boot_mode} eq C_ONTAP_CMODE)
	{
		eval
		{
			my $foreground = ($background == TRUE) ? 'false' : 'true';
			$cloned_vol->server->volume_clone_split_start(flexclone => $class->{vol_name},
								      vserver => $class->{vserver},
								      foreground => $foreground);
		};
	}
	else
	{
		eval
		{
			$cloned_vol->split(background => $background);
		};
	}
	my $output = $@;

	return $class->{_ui}->process_output_string(code => C_CMD_VOL_CLONE_SPLIT_START,
                                                    output => $output,
                                                    expected_rc => $args{expected},
                                                    log_message => "Split clone " . $class->{vol_name},);
}


## @name ::clone_split_stop
## @summary Abort splitting this clone of the Compress object.
##
## @description
## Abort splitting this clone of the Compress object.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Defaults to C_RC_SUCCESS
##
## @return Hash with rc and output.
##
## @examples
## $obj->clone_split_stop();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub clone_split_stop
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
	} );
	
	my $output;
	my $cloned_vol = $class->{volume}->_get_underlying_vol_obj();
	
	if ($class->{boot_mode} eq C_ONTAP_CMODE)
	{
		eval
		{
			$cloned_vol->server->volume_clone_split_stop(flexclone => $class->{vol_name},
								     vserver => $class->{vserver},);
		};
	}
	else
	{
		eval
		{
			$cloned_vol->split_stop();
		};
	}
	$output = $@;

	return $class->{_ui}->process_output_string(code => C_CMD_VOL_CLONE_SPLIT_STOP,
                                                    output => $output,
                                                    expected_rc => $args{expected},
                                                    log_message => "Stop split clone " . $class->{vol_name},);
}


## @name ::clone_split_status
## @summary Get the split status.
##
## @description
## Get the split status.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Legal values: C_RC_SUCCESS,C_RC_VOLUME_IS_NOT_BEING_SPLIT.
## Defaults to C_RC_SUCCESS (which means that split is in progress for this clone).
##
## @return Hash with rc and status.
##
## @examples
## $obj->clone_split_status();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub clone_split_status
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		expected => { type => SCALAR, optional => 1, default => C_RC_SUCCESS },
	} );
	
	my $rc = C_RC_FAILURE;
	my $status;
	my $cloned_vol = $class->{volume}->_get_underlying_vol_obj();
	
	if ($class->{boot_mode} eq C_ONTAP_CMODE)
	{
		eval
		{
			my $split_status = $cloned_vol->server->volume_clone_split_show(flexclone => $class->{vol_name},
											vserver => $class->{vserver},);
			$status = $split_status->[0];
		};
		
	}
	else
	{
		$status = $cloned_vol->split_status();
	}
	$rc = (defined($status)) ? C_RC_SUCCESS : C_RC_VOLUME_IS_NOT_BEING_SPLIT;
	
	my %return_status = $class->{_ui}->check_for_expected_rc(obtained_rc => $rc,
								 expected_rc => $args{expected},
								 log_message => "Clone split status for volume " . $class->{vol_name},
								 output      => $status);
	
	return (rc => $return_status{rc}, status => $status);
}


## @name ::wait_for_clone_split
## @summary Wait for volume clone split to complete.
##
## @description
## Wait for volume clone split to complete.
##
## @arg timeout Optional: Timeout, in seconds.
##
## @return Hash with rc and status.
##
## @examples
## $obj->wait_for_clone_split();
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub wait_for_clone_split
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		timeout => { type => SCALAR, optional => 1, regex => qr/^\d+$/, default => C_DFLT_CLONE_SPLIT_TIMEOUT }, 
	} );
	
	my $timeout = $args{timeout};
	
	logcomment("Waiting for completion of clone split for volume " . $class->{vol_name});
	
	my %res = $class->clone_split_status(expected => C_RC_FAILURE_NON_FATAL);
	logcomment("Clone split status:\n" . Dumper(%res)) if ($class->{_debug});
	my $status_rc = $res{rc};
	
	my $time_elapsed = 0;
	while ($time_elapsed <= $timeout)
	{
		if ($status_rc == C_RC_VOLUME_IS_NOT_BEING_SPLIT)
		{
			return {rc => C_RC_SUCCESS, status => $res{status}};
		}
		
		# Split did not complete yet
		sleep(10);
		$time_elapsed += 10;
		%res = $class->clone_split_status(expected => C_RC_FAILURE_NON_FATAL);
		$status_rc = $res{rc};
	}
	
	logresult('fatal', "Timed-out while waiting for volume clone split to complete for volume "  . $class->{vol_name} . 
		           "\nTimeout value: $timeout");
}


################################################
# Parent methods that need to become unsupported
################################################

sub setup
{
	my $class = shift;
	
	$class->_not_supported();
}



############################################
# Internal methods
############################################

sub _not_supported
{
	my $class = shift;
	
	logresult("script", "This method is not supported for a clone object");
}


# routine to test the module itself
sub testme
{
	package main;
	use Compression::Constants qw(:all);
	use Compression::Compress;
	use vars qw($FILER $AGGREGATE $VOLUME );

	require "drg/Compression/setup_core.pl";


	my $compress = Compression::Compress->setup(filer => $FILER,
						    aggregate => $AGGREGATE,
						    volume => {name => $VOLUME,
							       size => '4g'},);

	$compress->wait_for_cp();
	
	my $file_size = GB;
	my $expected_compressed_per_file = 916988;
	my $allow_savings_error = 20;

	$compress->{dataset}->create(files =>[{ name => 'file1', size => $file_size,},],);
	$compress->{dataset}->create(files =>[{ name => 'file2', size => $file_size,},],);
	$compress->{dataset}->create(files =>[{ name => 'file3', size => $file_size,},],);

	$compress->wait_for_cp();
	$compress->verify_savings(expected_compressed => 3 * $expected_compressed_per_file,
				  allow_percentage_error => $allow_savings_error);

	$compress->{dataset}->delete_file(filename => 'file2',);

	$compress->wait_for_cp();
	$compress->verify_savings(expected_compressed => 2 * $expected_compressed_per_file,
				  allow_percentage_error => $allow_savings_error);

	my $clone = $compress->clone_volume(); 
	$clone->wait_for_cp();
	$clone->is_enabled();
	$clone->verify_savings(expected_compressed => 2 * $expected_compressed_per_file,
			       allow_percentage_error => $allow_savings_error);
	$clone->clone_verify();
	$clone->clone_split_status(expected => C_RC_VOLUME_IS_NOT_BEING_SPLIT);
	$clone->clone_split_stop(expected => C_RC_VOLUME_IS_NOT_BEING_SPLIT);
	
	$clone->wait_for_cp();
	$clone->clone_split_status(expected => C_RC_VOLUME_IS_NOT_BEING_SPLIT);
	$clone->clone_split(background => TRUE);
	$clone->clone_split_status();
	$clone->wait_for_clone_split();
	$clone->clone_split_status(expected => C_RC_VOLUME_IS_NOT_BEING_SPLIT);
	$clone->clone_split_stop(expected => C_RC_VOLUME_IS_NOT_A_CLONE);

	# Double the size of file1
	$clone->{dataset}->modify(files =>[{ name => 'file1', offset => $file_size, len => $file_size},],);
	$clone->wait_for_cp();
	$clone->verify_savings(expected_compressed => 3 * $expected_compressed_per_file,
			       allow_percentage_error => $allow_savings_error);
	$compress->verify_savings(expected_compressed => 2 * $expected_compressed_per_file,
				  allow_percentage_error => $allow_savings_error);

	$compress->{dataset}->delete_file(filename => 'file3',);
	$compress->wait_for_cp();
	$compress->verify_savings(expected_compressed => $expected_compressed_per_file,
				  allow_percentage_error => $allow_savings_error);
	$clone->verify_savings(expected_compressed => 3 * $expected_compressed_per_file,
			       allow_percentage_error => $allow_savings_error);

	$clone->{dataset}->delete_file(filename => 'file3',);
	$clone->wait_for_cp();
	$clone->verify_savings(expected_compressed => 2 * $expected_compressed_per_file,
			       allow_percentage_error => $allow_savings_error);
	$compress->verify_savings(expected_compressed => $expected_compressed_per_file,
				  allow_percentage_error => $allow_savings_error);
	
	$clone->disable();
	$clone->{decompress}->start();
	$clone->{decompress}->wait_for_percentage();
	$clone->wait_for_cp();
	$clone->verify_savings(expected_compressed => 0);
	$compress->verify_savings(expected_compressed => $expected_compressed_per_file,
				  allow_percentage_error => $allow_savings_error);
	
	$clone->is_enabled(expected => C_RC_FAILURE);
	$compress->is_enabled();
	
	logcomment( "VolClone: testme: Done" );
}



1;
