#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/VolCopy.pm#1 $
#

## @summary Volume copy routines.
##
## @description
##    This module provides all the capabilities
##    required for volume copy.
##
## @synopsis
## 
## BEGIN { require "drg/Compression/setup_core.pl"; }
##
## use Compression::VolCopy;
## 
## my $compress1 = Compression::Compress->setup(filer => {name => $FILER_SRC,},
##                                              aggregate => {name => $AGGREGATE,},
##                                              volume => {name => 'vol1',},);
## my $vol_copy = Compression::VolCopy->new(source => $compress1,
##                                          destination => {volume => 'vol2',
##                                                          aggregate => $AGGREGATE,
##                                                          filer => $FILER_DST},);
##
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

package Compression::VolCopy;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Tharn;

use Compression::Constants qw(:all);
use Compression::Utils qw(_c_interpret_logresult);
use Compression::7Mode::VolCopy;
use Compression::CMode::VolCopy;


BEGIN
{
	require "Compression/base_import.pl";
}



## @name ::new
##
## @summary Compression::VolCopy Constructor.
##
## @description
## Creates an Compression::VolCopy object 
## (it does not start the vol copy relationship on the filer).
##
## @return Compression::VolCopy object.
##
## @arg source Required: Source Compression::Compress object.
##
## @arg destination Required: Hashref of destination arguments (volume, aggregate, filer).
##
## @arg disable_cleanup Optional: Flag to disable any clean-up
## that would be executed upon object destruction.
## Defaults to $main::DISABLE_CLEANUP.
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to $main::DEBUG.
##
## @example
## my $vol_copy = Compression::VolCopy->new(source => $compress1,
##                                          destination => {volume => 'vol2',
##                                                          aggregate => $AGGREGATE,
##                                                          filer => $FILER_DST},);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
##
sub new
{
	my $class = shift;
	my %args = validate(@_,
	{
		source 	=> {type => OBJECT,
                            isa => "Compression::Compress",},
                
                destination => {type => HASHREF,},
                
                disable_cleanup => {type => BOOLEAN,
                                    optional => 1,
                                    default => $main::DISABLE_CLEANUP,},
                
                debug => {type => BOOLEAN,
                          optional => 1,
                          default => $main::DEBUG,},
		} );
	
	my $self = {};
        my $source = $args{source};
        $self->{_disable_cleanup} = $args{disable_cleanup};
        $self->{_debug} = $args{debug};
        $self->{_creator_pid} = $$;
        $self->{_boot_mode} = $source->{boot_mode};
        $self->{_vserver} = $source->{vserver};
        
        
        if ($self->{_boot_mode} eq C_ONTAP_CMODE)
        {
                $args{vserver} = $self->{_vserver};
                $self->{_copy} = Compression::CMode::VolCopy->new(%args);
        }
        else
        {
                $self->{_copy} = Compression::7Mode::VolCopy->new(%args);
        }
        
        $self->{_src_vol_name} = $self->{_copy}->{_src_vol_name};
        $self->{_src_aggr_name} = $self->{_copy}->{_src_aggr_name};
        $self->{_src_host_name} = $self->{_copy}->{_src_host_name};
        $self->{_dst_vol_name} = $self->{_copy}->{_dst_vol_name};
        $self->{_dst_aggr_name} = $self->{_copy}->{_dst_aggr_name};
        $self->{_dst_host_name} = $self->{_copy}->{_dst_host_name};
        
        
        bless $self, $class; 

	return $self;
}


## @name ::attach_to_destination_vol
## @summary Gets the compress destination object.
## 
## @description
## Gets the compress destination object.
## For 7-mode, this method can be called at any time. It will return a compress object
## corresponding to a volume created with the default settings.
## For C-mode, it can only be called after the copy transfer completed. It will return a compress object
## corresponding to the volume created as a result of the volume copy operation.
##
## @return Compression::Compress object corresponding to the destination volume.
##
## @examples
## my $dest = $obj->attach_to_destination_vol();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

sub attach_to_destination_vol
{
	
	my $class = shift;
	
        return $class->{_copy}->attach_to_destination_vol(@_);
}


## @name ::start
## @summary Starts the vol copy process.
## 
## @description
## Starts the vol copy process.
##
## @return Hash with rc.
## 
## @arg background Optional: Flag to run the transfer in the background.
## To not be set if the expected result is not success.
## Default: FALSE.
## 
## @arg copy_all Optional: Flag to include all the volume's snapshots.
## Default: FALSE.
## 
## @arg snapshot Optional: The name of the snapshot, on the source volume, to copy.
##
## @arg timeout Optional: Timeout, in seconds, for the command to complete.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @arg restore_destination_state Optional: Flag to control if, after the copy,
## the destination volume's state should be restored (onlined and remounted).
## If defaults to TRUE, if the background option is not true.
##
## @examples
## $obj->start();
## $obj->start(copy_all => TRUE);
## $obj->start(snapshot => 'my_snapshot');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

sub start
{
	
	my $class = shift;
	
        return $class->{_copy}->start(@_);
}


## @name ::wait_for_completion
## @summary Wait for the background vol copy transfers to complete.
## 
## @description
## Wait for the background vol copy transfers to complete.
##
## @return Hash with rc.
##
## @examples
## $obj->wait_for_completion();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

sub wait_for_completion
{
	
	my $class = shift;
	
        return $class->{_copy}->wait_for_completion(@_);
}


## @name ::get_status
## @summary Gets the status of vol copy relationship.
## 
## @description
## Gets the status of vol copy relationship.
##
## @return Hash with rc, status, and transfer_progress. 
##
## @examples
## my %status = $obj->get_status();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

sub get_status
{
	
	my $class = shift;
	
        return $class->{_copy}->get_status(@_);
}


## @name ::verify_status
## @summary Verifies the status of vol copy relationship.
## 
## @description
## Verifies the status of vol copy relationship.
##
## @arg status Optional: Expected status.
## Default: C_VOL_COPY_STATUS_TRANSFERRING.
##
## @return Hash with rc. 
##
## @examples
## $obj->verify_status(status => C_VOL_COPY_STATUS_IDLE);
## $obj->verify_status(); # Same as $obj->verify_status(status => C_VOL_COPY_STATUS_TRANSFERRING);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

sub verify_status
{
	
	my $class = shift;
	my (%args) = validate( @_,
                {
                        status => {type => SCALAR, optional => 1, default => C_VOL_COPY_STATUS_TRANSFERRING},
                } );
        
        my $expected_status = $args{status};
        
        logcomment("Verifying the vol copy status ($expected_status):\n" .
                   "Source: " . $class->{_src_vol_name} . " (filer: " . $class->{_src_host_name} . ")\n" .
                   "Destination: " . $class->{_dst_vol_name} . " (filer: " . $class->{_dst_host_name} . ")");
        
        my %current_status = $class->get_status();
        
        if ($current_status{status} ne $expected_status)
        {
                logresult('fatal', "Failed to detect expected vol copy transfer status ($expected_status):\n" . Dumper(%current_status));
        }
        
        return %current_status;
}


## @name ::wait_for_status
## @summary Wait for the vol copy relationship to reach specific state.
## 
## @description
## Wait for the vol copy relationship to reach specific state.
##
## @return Hash with rc.
##
## @arg status Required: Status to wait for (can be a regex).
##
# @arg source Optional: Flag to get the status from the source filer.
# Default: FALSE (destination filer).
#
# @arg restore Optional: Flag to get the status of the restore transfer.
# Default: FALSE. 
#
## @arg timeout Optional: Timeout, in seconds, for completion.
##
## @examples
## $obj->wait_for_status(status => C_VOL_COPY_STATUS_TRANSFERRING);
## $obj->wait_for_status(status => C_VOL_COPY_STATUS_IDLE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

sub wait_for_status
{
	
	my $class = shift;
	
        return $class->{_copy}->wait_for_status(@_);
}


## @name ::abort
## @summary Aborts an active vol copy transfer.
## 
## @description
## Aborts an active vol copy transfer.
##
##
## @arg restore_destination_state Optional: Flag to control if, after the abort,
## the destination volume's state should be restored (onlined and remounted).
##
## @return Hash with rc.
##
## @examples
## $obj->abort();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

sub abort
{
	
	my $class = shift;
	
        return $class->{_copy}->abort(@_);
}


## @name ::throttle
## @summary Modifies the speed at which the vol copy transfer is being done.
## 
## @description
## Modifies the speed at which the vol copy transfer is being done.
##
## @arg value Optional: New speed value, to be set.
## The  value  ranges from 10 (full speed) to 1 (one-tenth of full speed).
## Default: C_DFLT_VOL_COPY_THROTTLE (10, full speed).
##
## @return Hash with rc. 
##
## @examples
## $obj->throttle(value => 1);
## $obj->throttle(); # To set back to default
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)

sub throttle
{
	
	my $class = shift;
	
        return $class->{_copy}->throttle(@_);
}


## @name ::restore_destination
## @summary Restores the state of the destination volume.
## 
## @description
## Restores the state of the destination volume, after the copy transfer has completed 
## (brings it online and remounts it).
##
## @return Hash with rc. 
##
## @examples
## $obj->restore_destination();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##
sub restore_destination
{
        my $class = shift;
	
        return $class->{_copy}->restore_destination(@_);
}



########################
# Internal methods
########################

# Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
	
	if (($class->{_disable_cleanup} == FALSE) &&
            ($class->{_creator_pid} == $$))
        {
                undef($class->{_copy});
        }
	else
	{
                logcomment( "VolCopy: cleanup:  nothing done: script PID: " . $$) if ($class->{_debug});
	}
        
        if (defined($execution_error))
        {
                _c_interpret_logresult($execution_error);
        }
}


sub testme
{
        package main;
        
	require "drg/Compression/setup_core.pl";
        use Compression::VolCopy;
        
        use vars qw($VOLUME $FILER_DST $FILER);
        
        param("FILER", "-mesg", "Filer name");
        param("VOLUME", "-default", 'basic_script_vol');
        param("FILER_DST", "-default", $FILER);
        
        my $compress1 = Compression::Compress->setup(filer => {name => $FILER ,},
                                                     aggregate => {name => $AGGREGATE,},
                                                     volume 	=> {name => 'vol1',
                                                                    compress_enabled => FALSE,},);
        
        my $boot_mode = $compress1->{boot_mode};
        
        # Need to differentiate here because the functionality differs
        if ($boot_mode == C_ONTAP_CMODE)
        {
                logcomment("Running C-mode testme...");
                
                my $vol_copy = Compression::VolCopy->new(source => $compress1,
                                                         destination => {volume => 'vol2',
                                                                         aggregate => $AGGREGATE,
                                                                         filer => $FILER_DST},);
                
                $compress1->{dataset}->create(files =>[{ name => 'file1', size => MB,},],);
                $compress1->wait_for_cp();
                $compress1->{dataset}->ls();
                
                $vol_copy->start(background => TRUE);
                $vol_copy->verify_status();
                $vol_copy->wait_for_status(status => C_VOL_COPY_STATUS_TRANSFERRING);
                $vol_copy->wait_for_completion();
                $vol_copy->verify_status(status => C_VOL_COPY_STATUS_IDLE);
                $vol_copy->restore_destination();
                
                my $compress2 = $vol_copy->attach_to_destination_vol();
                
                $compress1->{dataset}->ls();
                $compress2->{dataset}->ls();
                $compress2->{dataset}->modify(files =>[{ name => 'file1', offset => 0, len => C_WAFL_BLOCK_SIZE},],);
                $compress2->wait_for_cp();
                $compress1->verify_savings(expected_compressed => 0);
                $compress2->verify_savings(expected_compressed => 0);

                
        }
        else
        {
                logcomment("Running 7-mode testme...");
                
                my $vol_copy = Compression::VolCopy->new(source => $compress1,
                                                         destination => {volume => 'vol2',
                                                                         aggregate => $AGGREGATE,
                                                                         filer => $FILER_DST},);
                my $compress2 = $vol_copy->attach_to_destination_vol();
                
                $compress1->{dataset}->create(files =>[{ name => 'file1', size => MB,},],);
                $compress2->{dataset}->create(files =>[{ name => 'file2', size => MB,},],);
                $compress1->wait_for_cp();
                $compress1->{dataset}->ls();
                
                $vol_copy->start(background => TRUE);
                $vol_copy->throttle(value => 1);
                $vol_copy->verify_status();
                $vol_copy->wait_for_status(status => C_VOL_COPY_STATUS_TRANSFERRING);
                $vol_copy->abort();
                $vol_copy->verify_status(status => C_VOL_COPY_STATUS_IDLE);
                $vol_copy->start();
                # This should exit immediatelly
                $vol_copy->wait_for_status(status => C_VOL_COPY_STATUS_IDLE);
                
                $vol_copy->start(background => TRUE);
                $vol_copy->wait_for_completion();
                
                $vol_copy->restore_destination();
                
                $compress1->{dataset}->ls();
                $compress2->{dataset}->ls();
                $compress2->{dataset}->modify(files =>[{ name => 'file1', offset => 0, len => C_WAFL_BLOCK_SIZE},],);
                $compress2->wait_for_cp();
                $compress1->verify_savings(expected_compressed => 0);
                $compress2->verify_savings(expected_compressed => 0);
        }
        
        logresult('PASS', "Compression\:\:VolCopy testme passed");
}


1;
