# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/VolMove.pm#1 $
#
## @summary
##  Utility wrapper module for vol move.
##
## @description
##  Utility wrapper module for vol move.
##
## @note
## This module is only intended to be used as internal to the Compression library
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (internal)

package Compression::VolMove;

BEGIN
{
	require "Compression/base_import.pl";

}

use strict;
use warnings;
use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;

use Compression::Constants qw(:all);
use Compression::7Mode::VolMove;
use Compression::CMode::VolMove;




## @name:: new
## @summary Compression::VolMove constructor.
##
## @description
## Creates an object representing VolMove utilities.
## Creates the destination aggregate, if it does not exist yet.
##
## @return Object of type Compression::VolMove.
##
## @arg volume Required: NOTE/GNOTE::Volume object.
##
## @arg dest_aggr Required: Name of the destination aggregate.
##
## @arg vserver Optional: Name of the vserver (only relevant in c-mode).
##
## @arg ui Required: Compression::UI object for this filer.
##
## @example
## N/A : To not be used outside the Compression library.
##
sub new
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		volume => {type => OBJECT,},
		
		dest_aggr => {type => SCALAR,},
                
                ui => {type => OBJECT,
		       isa => "Compression::UI",},
	} );

	# Start building our object:
	my $self = {};
        $self->{_boot_mode} = $args{ui}->{boot_mode};
	$self->{_debug} = $args{ui}->{_debug};
	$self->{_dest_aggr} = $args{dest_aggr};
        
        if ($self->{_boot_mode} eq C_ONTAP_7MODE) 
        {
		$self->{_vol_move} = Compression::7Mode::VolMove->new(%args);
        }
        elsif ($self->{_boot_mode} eq C_ONTAP_CMODE)
        {
		$self->{_vol_move} = Compression::CMode::VolMove->new(%args);
        }


	bless($self, $class);

	logcomment("VolMove: Setup Complete..\n\n") if ($self->{_debug});
	return $self;
}


## @name:: start
## @summary Starts a vol move operation.
##
## @description
## Starts a vol move operation.
##
## @arg no_cutover Optional: Flag to disable cutover during the move. Default: FALSE.
##
## @arg retain_src Optional: Flag to retain the source volume after the move is complete. Defaulf: FALSE.
## 
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @return Hash with rc and output.
##
## @example
## $obj->start();
##
sub start
{
        my $class = shift;
        
        return $class->{_vol_move}->start(@_);
}


## @name:: status
## @summary Gets the status of vol move operation.
##
## @description
## Gets the status of vol move operation.
##
## @arg required_state Optional: Required vol move state.
## If not defined, no state-related check will be executed.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @example 
## $obj->status();
## $obj->status(required_state => C_VOL_MOVE_STATE_SETUP);
## $obj->status(expected => C_RC_VOL_IS_NOT_BEING_MOVED);
##
sub status
{
        my $class = shift;
        
        return $class->{_vol_move}->status(@_);
}


## @name:: wait_for_completion
## @summary Waits for a vol move operation to complete.
##
## @description
## Waits for a vol move operation to complete.
##
## @arg timeout Optional: Timeout value. Default: C_DFLT_TIMEOUT.
##
## @arg delay Optional: Time delay between two consecutive checks. Default: 5.
##
## @return Hash with rc.
##
## @example
## $obj->wait_for_completion();
##
sub wait_for_completion
{
	my $class = shift;
        
        return $class->{_vol_move}->wait_for_completion(@_);
}


## @name:: wait_for_state
## @summary Waits for a vol move operation to reach the required state.
##
## @description
## Waits for a vol move operation to reach the required state.
##
## @arg state Required: State to be reached.
## Legal values: C_VOL_MOVE_STATE_SETUP, C_VOL_MOVE_STATE_MOVE, C_VOL_MOVE_STATE_CUTOVER.
##
## @arg timeout Optional: Timeout value. Default: C_DFLT_TIMEOUT.
##
## @arg delay Optional: Time delay between two consecutive checks.
## Default: 1.
##
## @return Hash with rc.
##
## @example
## $obj->wait_for_state(state => C_VOL_MOVE_STATE_MOVE);
##
sub wait_for_state
{
	my $class = shift;
        
        return $class->{_vol_move}->wait_for_state(@_);
}


## @name:: initiate_cutover
## @summary Initiate cutover of volume during its vol move.
##
## @description
## Initiate cutover of volume during its vol move.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @return Hash with rc and output.
##
## @example
## $obj->initiate_cutover();
##
sub initiate_cutover
{
	my $class = shift;
        
        return $class->{_vol_move}->initiate_cutover(@_);
}


## @name:: pause
## @summary Pause vol move of volume.
##
## @description
## Pause vol move of volume.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @return Hash with rc and output.
##
## @example
## $obj->pause();
##
sub pause
{
	my $class = shift;
        
        return $class->{_vol_move}->pause(@_);
}


## @name:: is_paused
## @summary Checks if vol move is paused for this volume.
##
## @description
## Checks if vol move is paused for this volume.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @return Hash with rc and output.
##
## @example
## $obj->is_paused();
##
sub is_paused
{
	my $class = shift;
        
        return $class->{_vol_move}->is_paused(@_);
}


## @name:: resume
## @summary Resumes vol move for this volume.
##
## @description
## Resumes vol move for this volume.
##
## @return Hash with rc and output.
##
## @example
## $obj->resume();
##
sub resume
{
	my $class = shift;
        
        return $class->{_vol_move}->resume(@_);
}


## @name:: abort
## @summary Aborts vol move for this volume.
##
## @description
## Aborts vol move for this volume.
##
## @return Hash with rc and output.
##
## @example
## $obj->abort();
##
sub abort
{
	my $class = shift;
        
        return $class->{_vol_move}->abort(@_);
}



################################################################
# Internal subroutines
################################################################

sub testme
{

	package main;
        use Data::Dumper;
	
        require "drg/Compression/setup_core.pl";
	
        
        use vars qw($FILER $AGGREGATE $VOLUME $DEBUG $DEST_AGGR);
        
        param("DEBUG", "-default", FALSE);
	param("DEST_AGGR", "-default", C_DFLT_AGGREGATE_DEST);
        
        my $compress = Compression::Compress->setup(filer => {name => $FILER ,
							      require_debug_build => 1,
							      },
                                                    aggregate => $AGGREGATE,
                                                    volume => {name => $VOLUME,
                                                               size => "7g",
                                                               compress_enabled => FALSE,
                                                               },
                                                    debug => $DEBUG,);
	
	my $filename = 'file1';
	
	$compress->{dataset}->create(files =>[{name => $filename, size => GB,},],);
	$compress->{dataset}->ls();
	$compress->wait_for_cp();
	
	logcomment("Testing basic vol move");
	$compress->start_vol_move_to_aggr(dest_aggr => $DEST_AGGR,);
	$compress->{vol_move}->status(required_state => C_VOL_MOVE_STATE_SETUP);
	$compress->{vol_move}->wait_for_state(state => C_VOL_MOVE_STATE_MOVE);
	$compress->{vol_move}->wait_for_completion();
	$compress->{vol_move}->status(expected => C_RC_VOL_IS_NOT_BEING_MOVED);
	$compress->wait_for_cp(num_cp => 2);
	
	logcomment("Testing vol move with manual cutover");
	$compress->start_vol_move_to_aggr(dest_aggr => $AGGREGATE, no_cutover => TRUE);
	$compress->{vol_move}->status();
	$compress->{vol_move}->wait_for_state(state => C_VOL_MOVE_STATE_CUTOVER_DEFERRED);
	$compress->{vol_move}->initiate_cutover();
	$compress->{vol_move}->wait_for_completion();
	$compress->{vol_move}->status(expected => C_RC_VOL_IS_NOT_BEING_MOVED);
	$compress->wait_for_cp(num_cp => 2);
	
	logcomment("Testing vol move with pause/resume/abort");
	$compress->start_vol_move_to_aggr(dest_aggr => $DEST_AGGR,);
	$compress->{vol_move}->wait_for_state(state => C_VOL_MOVE_STATE_MOVE);
	if ($compress->{boot_mode} eq C_ONTAP_CMODE)
	{
		# abort/resume are not available
		$compress->{vol_move}->abort();
	}
	else
	{
		$compress->{vol_move}->pause();
		$compress->{vol_move}->status(required_state => C_VOL_MOVE_STATE_MOVE);
		$compress->{vol_move}->is_paused();
		$compress->{vol_move}->resume();
		$compress->{vol_move}->wait_for_completion();
	}
	$compress->{vol_move}->status(expected => C_RC_VOL_IS_NOT_BEING_MOVED);
	
	my $aggr_changed = ($compress->{boot_mode} eq C_ONTAP_CMODE) ? FALSE : TRUE;
	$compress->refresh_container_aggr(fail_if_not_changed => $aggr_changed);
	
	$compress->{dataset}->modify(files =>[{ name => $filename, offset => 0, len => 2 * KB},],);
	
	$compress->{dataset}->ls();
        
        logresult('PASS', "VolMove: testme: Done");
}




1;