#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/Volume.pm#1 $
#

## @summary Compression Volume routines.
##
## @description
## Wrapper around NOTE::Volume.
## The main reason for not using NOTE::Volume directly is to provide standardized output
##    and to enable specific processing on some failures, otherwise fatal.
##    This is required for handling compression-specific failures.
##    
## @synopsis
##
## use Compression::Volume;
## use vars qw( $FILER $AGGREGATE $VOLUME );
## 
## BEGIN { require "Compression/setup_core.pl"; }
##
## my $volume = Compression::Volume->new(filer => $filer_obj,
##                                       aggregate => $aggregate_obj,
##                                       volume => {name => 'flex_vol',
##                                                  size => '2g',
##                                                  recycle => FALSE,);
## OR
## my $volume = Compression::Volume->new(filer => $filer_obj,
##                                       volume => {name => 'trad_vol',
##                                                  type => C_VOL_TYPE_TRAD,
##                                                  disks => 3,
##                                                  recycle => FALSE,);
##
## @examples
## $volume->bring_offline();
## $volume->bring_online();
## $volume->set_option(service => C_OPTION_COMPRESSION, new_value => TRUE );
## $volume->verify_option( ( service => C_OPTION_COMPRESSION, expected => C_RC_FAILURE );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)


package Compression::Volume;

use strict;
use warnings;

use Tharn;
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Data::Dumper;
use Compression::Constants qw(:all);
use Compression::7Mode::Volume;
use Compression::CMode::Volume;
use Compression::Utils qw(detect_indirect_type_from_status
                          compare_hash_entries
                          _send_cmd_to_filer
			  _c_interpret_logresult);

use wack;
use SnapCmds;
use Wreck qw(wreck);
use Volume qw(:all);
use filer qw(filer_vol_snap_reclaimspace_get);
use SayWrap qw(sayEOF);

use vars qw($WAFLIRON $WAFLIRON_IS_DONE);

BEGIN
{
	require "Compression/base_import.pl";
	require Exporter;
	our @ISA = qw(Exporter);
}


## @name ::new
## @summary Compression::Volume Constructor.
##
## @description
## Creates an object representing a volume.
##
## @return Object of type Compression::Volume.
##
## @arg filer Required: NOTE::Filer object corresponding to the filer on which the volume will exist.
##
## @arg aggregate Optional: NOTE::Aggregate object corresponding to the aggregate which will contain the volume.
##
## @arg volume Optional: Volume information.
## It can be either a scalar (representing the name), or a hash reference.
## If not provided, uses the default values defined in Compression::Constants.
## If volume is passed in as a hashref, then the allowed keys/values are:
## - name => name of the volume,
## - type => volume type (flex/trad). Default: C_DFLT_VOL_TYPE,
## - size => volume size. Default: C_DFLT_VOL_SIZE,
## - recycle => flag to decide if the volume should be recycled,
## - tag => volume tag,
## - exportfs => Exportfs options for this volume. Default: C_DFLT_VOL_EXPORTFS,
## - disks => disk to use when creating a trad vol,
## - compress_enabled => Flag to decide if the volume should be compression-enabled. Default: C_ENABLE_COMPRESSION,
## - auto_mount => Flag to decide if the volume should be mounted on the localhost.
##
## @arg ui Optional: Compression::UI object corresponding to the hosting filer.
## If not provided, it will be auto-generated.
##
## @arg disable_cleanup Optional: Flag to disable any clean-up that is executed
## upon object destruction. Defaults to FALSE.
## 
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @example
##
## my $aggregate = NOTE::Aggregate->create(Filer => { host => $FILER },
##                                         name => C_DFLT_AGGREGATE,
##                                         recycle => TRUE,
##                                         share => TRUE,
##                                         disks => 3,
##                                         );
## 
## my $obj = Compression::Volume->new(aggregate => $aggregate,
##                                    volume => {name => 'test_vol',
##                                               size => "2g",
##                                               recycle => TRUE,);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub new
{
	my $class = shift;
	my (%args) = validate( @_,
		{
                        filer => {type => OBJECT,},
                        
                        aggregate => {type => OBJECT,
				      optional => 1,},
                        
                        volume 	=> {type => SCALAR|HASHREF|UNDEF,
				    optional => 1,},
                        
                        vserver => {type => SCALAR|UNDEF,
                                    optional => 1},
                        
                        boot_mode => {type => SCALAR,
                                      optional => 1,
                                      default => C_ONTAP_BOOT_MODE},
                        
                        ui      => {type => OBJECT,
                                    isa  => "Compression::UI",
                                    optional => 1,},
                        
                        client  => {type => SCALAR,
                                    optional => 1,
                                    default => C_CLIENT_NFS_LOCAL,},
                        
                        disable_cleanup => {type     => BOOLEAN,
					    optional => 1,
					    default  => FALSE,},
                        
                        debug 	=> {type => SCALAR,
				    optional => 1,
				    default => FALSE,},
			
			# Internal flag. To not be used outside the compression library
			attach => {type => OBJECT|SCALAR|UNDEF,
				   optional => 1,},
                        
                        vfiler => {type => OBJECT|UNDEF,
                                   optional => 1,},
			
			# Internal flag. To not be used outside the compression library
			_fake_mount => {type => SCALAR,
					default => FALSE,},
			
			# Internal flag. To not be used outside the compression library
			_added_to_vfiler => {type => SCALAR,
					     default => FALSE,},
		} );

	my $self = {};
        $self->{_debug} = $args{debug};
        $self->{_disable_cleanup} = $args{disable_cleanup};
        $self->{_creator_pid} = $$;
        $self->{_boot_mode} = $args{boot_mode};
        
        delete($args{boot_mode});
        if ($self->{_boot_mode} eq C_ONTAP_CMODE)
	{
                if (defined($args{vfiler}))
                {
                        logresult ('NA', "NOT APPLICABLE: Vfilers are not supported in C-mode");
                }
                delete($args{vfiler});
		delete($args{_added_to_vfiler});
		delete($args{_mounted_on});
                $self->{_c_volume} = Compression::CMode::Volume->new(%args);
	}
        else
        {
                delete($args{vserver});
                $self->{_c_volume} = Compression::7Mode::Volume->new(%args);
        }
	
        $self->{vol_name} = $self->{_c_volume}->{vol_name};
        
        # Array of clients this volume gets mounted on during its lifespan
        $self->{_mounted_on} = [ ];
        
        if (defined($self->{_c_volume}->{_was_mounted_at_setup}))
        {
                my $client_list = $self->{_mounted_on};
                push(@$client_list, $self->{_c_volume}->{_was_mounted_at_setup});
        }
        
        $self->{_ui} = $self->{_c_volume}->{_ui};
        $self->{_vfiler_name} = $self->{_c_volume}->{_vfiler_name};
        $self->{_recycled} = $self->{_c_volume}->{_recycled};
        
        bless $self, $class;
        
	return $self;
}


## @name ::bring_offline
## @summary Sets volume offline.
## 
## @description
## Sets volume offline. Wrapper around NOTE::Volume::offline.
##
## @return Hash with rc and output.
## 
## @examples
##
## $obj->bring_offline();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub bring_offline
{
        my $class = shift;
        
        return $class->{_c_volume}->bring_offline(@_);
}


## @name ::bring_online
## @summary Sets volume online.
## 
## @description
## Sets volume online. Wrapper around NOTE::Volume::online.
##
## @arg expected Optional: expected result ( default is C_RC_SUCCESS ). 
##     Use C_RC_INSUFFICIENT_SPACE if there is not enough space to bring volume online.
##
## @return Hash with rc and output.
##  
## @examples
##
## $obj->bring_online();
## $obj->bring_online(expected=>C_RC_INSUFFICIENT_SPACE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub bring_online
{
        my $class = shift;
        
        return $class->{_c_volume}->bring_online(@_);
}


## @name ::mark_restricted
## @summary Marks volume as restricted.
## 
## @description
## Marks volume as restricted. Wrapper around NOTE::Volume::restrict.
##
## @return Hash with rc and output.
##  
## @examples
##
## $obj->mark_restricted();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub mark_restricted
{
        my $class = shift;
        
        return $class->{_c_volume}->mark_restricted(@_);
}


## @name ::verify_option
## @summary Verifies that the desired volume option is set as expected.
## 
## @description
## Verifies that the desired volume option is set as expected. Logs fatal result if not.
##
## @return Hash with rc and output.
## 
## @arg service Required: feature that corresponds to desired option (defined in Compression::Constants).
##
## @arg expected Optional: expected return code ( C_RC_SUCCESS/C_RC_FAILURE ).
## Defaults to C_RC_SUCCESS.
##
## @examples
##
## $obj->verify_option( service => C_OPTION_GUARANTEE, expected => C_RC_SUCCESS );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub verify_option
{
        my $class = shift;
        
        return $class->{_c_volume}->verify_option(@_);
}


## @name ::set_option
## @summary Sets a new value for the desired option for the volume.
## 
## @description
## Sets a new value for the desired option for the volume.
##
## @return Hash with rc and output.
## 
## @arg service Required: feature that corresponds to desired option (defined in Compression::Constants).
## 
## @arg new_value Required: tag that corresponds to new value to be set.
## For now, only TRUE/FALSE are handled, corresponding to enable/disable.
## In the case of C_OPTION_GUARANTEE, new_value can also be "file".
##
## @arg volume_name Optional: Name of volume for which to disable the option.
## If not provided, the desired option will be set for the volume for which the object was created.
##
## @arg check_if_already_set Optional: Flag to enable the library to first check the current setting.
## If the current setting matches the value to be set, the subroutine will not attempt to set it again.
## Instead, it will return success. The default value for check_if_already_set is TRUE.
##
## @arg expected Optional: expected return code (defined in Compression::Constants).
## If not provided, assumes the command should complete successfully.
##
## @arg use_new_value_as_is Optional: Flag to enable the library to use the new_value string directly,
## even if a definition is already given in Options.pm. Default: FALSE.
## 
## @examples
## $obj->set_option( service => C_OPTION_GUARANTEE, new_value => TRUE, expected => 'file' );
## $obj->set_option( service => C_OPTION_GUARANTEE, new_value => TRUE, expected => FALSE );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub set_option
{
        my $class = shift;
        
        return $class->{_c_volume}->set_option(@_);
}


## @name ::get_mount_path
## @summary Returns the full volume name.
## 
## @description
## Returns the full volume name (ex: /vol/myvol). Wrapper around NOTE::mount_path.
##
## @return Full volume name.
##  
## @examples
## my $vol_mount_path = $obj->get_mount_path();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_mount_path
{
        my $class = shift;
        
        return $class->{_c_volume}->get_mount_path(@_);
}


## @name ::nfs_mount
## @summary NFS-mounts the volume on the specified client.
## 
## @description
## NFS-mounts the volume on the specified client.
##
## @notes
## Internally stores the list of the clients this volume is mounted on during its lifespan, and
## uses it upon object destruction, to automatically umount the volume.
##
## @return Hash with rc and output (resulted mountpoint).
## 
## @arg client Optional: Hostname of the client to mount the volume on.
## Defaults to 'localhost'.
## 
## @arg chmod Optional: Permissions to set on the mountpoint. Only supported for 7-mode.
## Defaults to '777'.
##  
## @examples
## $obj->nfs_mount( client => $my_client );
##
## @note
## For c-mode, the chmod argument is not supported.
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub nfs_mount
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        client => { type => SCALAR, optional => 1, default => C_CLIENT_NFS_LOCAL },
			chmod => { type => SCALAR|UNDEF, optional => 1, default => '777' },
                } );
        
        my $client = $args{client};
        my $volume = $class->{vol_name};
        
        # Mount the volume
        my %return_hash = $class->{_c_volume}->nfs_mount(@_);
        
        # Remember the host we just mounted the volume on
        my $client_list = $class->{_mounted_on};
        
        logcomment("Volume: list of hosts that $volume is mounted on:\n" . Dumper($client_list) .
                   "\nAdding $client to this list.") if $class->{_debug};
        push(@$client_list, $client);
        logcomment("Volume: new list of hosts that $volume is mounted on:\n" . Dumper($client_list)) if $class->{_debug};
        
        $return_hash{rc} = C_RC_SUCCESS;
        logcomment("Volume: nfs_mount: returning:\n" . Dumper(%return_hash)) if $class->{_debug};
        
        return %return_hash;
}

## @name ::nfs_umount
## @summary NFS-umounts the volume from the specified client.
## 
## @description
## NFS-umounts the volume from the specified client.
##
## @return Hash with rc and output (return value of NOTE::Volume::umount).
## 
## @arg client Optional: Hostname of the client to umount the volume from.
## Defaults to 'localhost'.
##  
## @examples
## $obj->nfs_umount( client => $my_client );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub nfs_umount
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        client => { type => SCALAR, optional => 1, default => C_CLIENT_NFS_LOCAL },
                } );
        
        my $client = $args{client};
        my $volume = $class->{vol_name};
        
        # Umount the volume
        my %return_hash = $class->{_c_volume}->nfs_umount(@_);
        
        # Remove the host from the list of clients the volume is mounted on (including multiple entries)
        my $client_list = $class->{_mounted_on};
        logcomment("Volume: Current list of hosts that $volume is mounted on:\n" . Dumper($client_list) .
                   "\nRemoving $client from this list") if $class->{_debug};
        my $i = 0;
        while (defined($client_list->[$i]))
        {
                my $listed_client = $client_list->[$i];
                logcomment("Volume: processing index $i: $listed_client") if $class->{_debug};
                if ($listed_client eq $client)
                {
                        logcomment("Volume: removing $listed_client from the list") if $class->{_debug};
                        splice(@$client_list, $i, 1);
                        # Removed an entry, the whole array shifted back, need to process this index again
                        $i--;
                }
                $i++;
        }
        logcomment("Volume: new list of hosts that $volume is mounted on:\n" . Dumper($client_list)) if $class->{_debug};      
        
        $return_hash{rc} = C_RC_SUCCESS;
        logcomment("Volume: nfs_umount: returning:\n" . Dumper(%return_hash)) if $class->{_debug};
        
        return %return_hash;
}


## @name ::get_available_space
## @summary Gets the available space (in KB) on this volume.
## 
## @description
## Gets the available space (in KB) on this volume.
##
## @return Hash with rc and output (available space in KB).
##  
## @examples
## $obj->get_available_space();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_available_space
{
        my $class = shift;
        
        return $class->{_c_volume}->get_available_space(@_);
}


## @name ::get_space_usage 
## @summary Gets disk usage of this volume.
## 
## @description
## Gets disk usage of this volume.
##
## @arg snapshot Optional: Flag to use if the snapshot statistics are needed.
## Default: FALSE.
##
## @return Hash with rc and output (available space in KB).
##  
## @examples
## $obj->get_space_usage(snapshot => TRUE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_space_usage
{
        my $class = shift;
        
        return $class->{_c_volume}->get_space_usage(@_);
}


## @name ::verify_space_usage
## @summary Verifies the disk usage of this volume.
## 
## @description
## Verifies the disk usage of this volume.
##
## @arg expected_values Optional: Hashref of values that are expected.
## Legal hash keys: used, total, available, capacity.
##
## @arg snapshot Optional: Flag to use if the snapshot statistics are needed to be verified.
## Default: FALSE.
##
## @arg allow_percentage_error Optional: Percentage of error that is allowed for the obtained values.
## Default: 0.
##
## @return Hash with rc and output (available space in KB).
##  
## @examples
## $obj->->verify_space_usage(expected_values => {used => 3, total => 16384, available => 16288, capacity => 1 },
##                            snapshot => TRUE);
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub verify_space_usage
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        snapshot => { type => BOOLEAN, optional => 1, default => FALSE },
                        expected_values => { type => HASHREF },
                        allow_percentage_error => { type => SCALAR, optional => 1, default => 0 },
                } );
        
        my $expected_vals_ref = $args{expected_values};
        my $allow_percentage_error = $args{allow_percentage_error};
        my $snapshot = $args{snapshot};
        
         my %expected_vals = validate_with
                        ( params => $expected_vals_ref,
                          spec   => {total => { type => SCALAR, regex => qr/^\d+$/, optional => 1, },
                                     used => { type => SCALAR, regex => qr/^\d+$/, optional => 1, },
                                     available => { type => SCALAR, regex => qr/^\d+$/, optional => 1, },
                                     capacity => { type => SCALAR, regex => qr/^\d+$/, optional => 1, },},
                          allow_extra => 0,
                        );
                        
        if (keys(%expected_vals) == 0)
        {
                logresult('script', 'verify_disk_usage called with empty expected_values hash');
        }
        
        my %ret_hash = $class->get_space_usage(snapshot => $snapshot);
        
        my %min_allowed_errors = (used => C_VOL_USED_SPACE_MIN_JITTER,
                                  total => 0,
                                  available => 0,
                                  capacity => 0,);
        
        foreach my $key (keys(%expected_vals))
        {
                my $expected_val = $expected_vals{$key};
                
                my $min_jitter = $min_allowed_errors{$key};
                my $user_allowed_jitter = int(($expected_val * $allow_percentage_error) / 100);
                my $allowed_error = ($min_jitter > $user_allowed_jitter) ? $min_jitter : $user_allowed_jitter;
                
                my $min_expected_val = $expected_val - $allowed_error;
                my $max_expected_val = $expected_val + $allowed_error;
                my $obtained_val = $ret_hash{$key};
                
                logcomment("Verifying disk usage for volume " . $class->{vol_name} . " (snapshot=$snapshot):\n" .
                           " Field: $key\n Obtained value: $obtained_val\n" .
                           " Minimum allowed value: $min_expected_val\n Maximum allowed value: $max_expected_val");
                
                if (($obtained_val < $min_expected_val) || ($obtained_val > $max_expected_val))
                {
                        logresult('fatal', "Volume " . $class->{vol_name} . ": unexpected disk usage value for field \"$key\"\n" .
                                           "Obtained value: $obtained_val\n" .
                                           "Minimum allowed value: $min_expected_val\n" .
                                           "Maximum allowed value: $max_expected_val");
                }
        }
        
        return %ret_hash;
}


## @name ::flush_caches
## @summary Flushes all caches (NFS, filer).
## 
## @description
## Flushes all caches (NFS, filer). 
##
## @return Hash with rc and output (empty).
##  
## @examples
## $obj->flush_caches();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub flush_caches
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        # No args
                } );
        
        logcomment("Clearing filer caches");
        $class->bring_offline();
        $class->bring_online();
        
        logcomment("Clearing NFS caches");
        $class->nfs_umount();
        $class->nfs_mount();
        
        return (rc => C_RC_SUCCESS, output => '');
}


## @name ::get_status
## @summary Gets the status of the volume.
## 
## @description
## Gets the status of the volume. Wrapper around NOTE::Volume::status.
##
## @return Hash with rc and status, where status is the hash reference returned by NOTE::Volume::status.
##  
## @examples
## $obj->get_status();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_status
{
        my $class = shift;
        
        return $class->{_c_volume}->get_status(@_);
}


## @name ::create_snapshot
## @summary Creates a snapshot on the volume.
## 
## @description
## Creates a snapshot on the volume. Wrapper around NOTE::Volume::snapshot_create.
##
## @return Hash with rc and output, where output is the name of the newly-created snapshot.
##
## @arg name Optional: Name for the snapshot. If not provided, an anonymous name is used.
##  
## @examples
## $obj->create_snapshot(name => $my_snapshot);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub create_snapshot
{
        my $class = shift;
        
        return $class->{_c_volume}->create_snapshot(@_);
}


## @name ::delete_snapshot
## @summary Deletes a snapshot.
## 
## @description
## Deletes a named snapshot. Wrapper around NOTE::Volume::snapshot_delete.
##
## @return Hash with rc and output (empty).
##
## @arg name Required: Name of the snapshot.
##  
## @examples
## $obj->delete_snapshot(name => $my_snapshot);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub delete_snapshot
{
        my $class = shift;
        
        return $class->{_c_volume}->delete_snapshot(@_);
}


## @name ::list_snapshots
## @summary Lists snapshots that exist on this volume.
## 
## @description
## Lists snapshots that exist on this volume. Wrapper around NOTE::Volume::snapshot_list.
##
## @return Hash with rc and output (reference to array with names of existent snapshots).
##  
## @examples
## $obj->list_snapshots();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub list_snapshots
{
        my $class = shift;
        
        return $class->{_c_volume}->list_snapshots(@_);
}


## @name ::verify_snapshot_exists
## @summary Verifies that a named snapshot exists on this volume.
## 
## @description
## Verifies that a named snapshot exists on this volume. Wrapper around NOTE::Volume::snapshot_exists.
##
## @return Hash with rc and output.
##
## @arg name Required: Name of the snapshot.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##  
## @examples
## $obj->verify_snapshot_exists(name => $existent_snapshot);
## $obj->verify_snapshot_exists(name => $inexistent_snapshot, expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub verify_snapshot_exists
{
        my $class = shift;
        
        return $class->{_c_volume}->verify_snapshot_exists(@_);
}


## @name ::snapshot_schedule
## @summary Gets/sets the snapshot schedule for this volume.
## 
## @description
## Gets/sets the snapshot schedule for this volume. Wrapper around NOTE::Volume::snapshot_sched.
##
## @return Hash with rc and output (array reference with current schedule - only for getting the current setting).
##
## @arg disable Optional: Boolean flag indicating that the snapshot scheduler should be disabled for this volume.
## Only considered if set to TRUE.
##
## @arg schedule Optional: New snapshot schedule for this volume.
## If not specified, the current schedule is returned.
##
## @note
## If disable is set to TRUE, then the schedule argument is ignored.
##  
## @examples
## $obj->snapshot_schedule(disable => TRUE);
## $obj->snapshot_schedule(schedule => '1 1 0');
## my %current_setting = $obj->snapshot_schedule();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub snapshot_schedule
{
        my $class = shift;
        
        return $class->{_c_volume}->snapshot_schedule(@_);
}


## @name ::remove_all_snapshots
## @summary Removes all existent for this volume.
## 
## @description
## Removes all existent for this volume.
##
## @return Hash with rc and output (reference to array containing the removed snapshots if verify is set to FALSE.
## If verify is set to TRUE, then te array will be empty (snapshots remaining after the removal)).
##
## @arg verify Optional: Boolean flag to verify that there were no snaphots left
## (which may happen if snap sched was not previously disabled). Default: FALSE.
##  
## @examples
## $obj->remove_all_snapshots(verify => TRUE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub remove_all_snapshots
{
        my $class = shift;
	my (%args) = validate( @_,
		{
                        verify => { type => BOOLEAN, optional => 1, default => FALSE }
                } );
        
        my %res = $class->list_snapshots();
        my $existent_snapshots = $res{output};

        foreach my $snapshot (@$existent_snapshots)
        {
                logcomment("Deleting existent snapshot $snapshot");
                $class->delete_snapshot(name => $snapshot);
        }

        if ($args{verify} == TRUE)
        {
                logcomment("Verifying that all existent snapshots (if any) of the volume have been removed") if $class->{_debug};
                %res = $class->list_snapshots();
                $existent_snapshots = $res{output};
                if (scalar(@$existent_snapshots) > 0)
                {
                        logresult('script', "Library problem: Failed to remove all existent snapshots:\n" . Dumper($existent_snapshots));
                }
        }
        
        return (rc => C_RC_SUCCESS, output => $existent_snapshots);
}


## @name ::restore_snapshot
## @summary Restores a snapshot of this volume.
## 
## @description
## Restores a snapshot of this volume.
##
## @arg snapshot_name Required: Snapshot name.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc and output.
##  
## @examples
## $obj->restore_snapshot(snapshot_name => 'my_snapshot');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub restore_snapshot
{
        my $class = shift;
        
        return $class->{_c_volume}->restore_snapshot(@_);
}


## @name ::get_snap_delta 
## @summary Gets snap delta between two given snapshots.
## 
## @description
## Gets snap delta between two given snapshots. Wrapper around NOTE::Volume::snap_delta.
##
## @return Hash with rc and snap_delta (reference to array of hashes).
##  
## @examples
## $obj->get_snap_delta(snapshot_start => 'my_snapshot1',
##                      snapshot_end => 'my_snapshot2');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_snap_delta
{
        my $class = shift;
        
        return $class->{_c_volume}->get_snap_delta(@_);
}


## @name ::get_snap_reclaimable 
## @summary Returns the amount of space reclaimed if a snapshot is deleted.
## 
## @description
## Returns the amount of space reclaimed if a snapshot is deleted (in bytes).
##
## @return Hash with rc and reclaimable.
##  
## @examples
## my %res = $obj->get_snap_reclaimable(snapshot_name => 'my_snapshot');
## my $value = $res{reclaimable};
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_snap_reclaimable
{
        my $class = shift;
        
        return $class->{_c_volume}->get_snap_reclaimable(@_);
}


## @name ::start_wafliron
## @summary Starts wafliron on the aggregate that contains this volume.
## 
## @description
## Starts wafliron on the aggregate that contains this volume Wrapper around NOTE::Volume::wafliron_start.
##
## @return Hash with rc and output.
# If the 'oc' argument is set to prev_cp, the return will also contain the reference to the cp hash provided by wafliron_start.
##
## @arg wait Optional: Flag to decide if start_wafliron should exit immediately or wait for the wafliron to complete.
## Defaults to TRUE.
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Ignored if wait flag is 0. Legal values: C_RC_SUCCESS, C_RC_WAFLIRON_CORRUPTION_FOUND.
## Default value C_RC_SUCCESS (which means that wafliron found the filesystem clean).
##
## @arg force Optional: (command line option -f).
##
## @arg triage Optional: (command line option -t).
##
## @arg scan_speed Optional: Wafl scan speed.
##
## @arg inject_file Optional: (command line option -i <inject-file>).
##
## @arg timeout Optional: Wafliron timeout value.
##
## @arg volume_order Optional: Volume ordering string (command line option -v).
## Available starting IC.0.
##
# @arg oc Optional: Optional commit flag. Defaults to 0.
# Legal values: 0 (regular wafliron is run), 1 (wafliron is run in Optional Commit mode),
# prev_cp (wafliron is run in Optional Commit prev_cp mode).
# Available starting BR.0 and IC.1.
#
# @arg response_list Optional: Hash with key = vol/aggr name, value = reponse to be sent to the vol/aggr during prev_cp.
#
## @notes
## This is a wrapper around Wafliron::wafliron_start. More detailed argument documentation is available there.
## 
## @examples
## $obj->start_wafliron();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub start_wafliron
{
        my $class = shift;
        
        return $class->{_c_volume}->start_wafliron(@_);
}


## @name ::wait_for_wafliron
## @summary Waits for wafliron to complete.
## 
## @description
## Waits for wafliron to complete. Wrapper around NOTE::Volume::wafliron_wait.
##
## @return Hash with rc and output.
## 
## @examples
## $obj->wait_for_wafliron();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub wait_for_wafliron
{
        my $class = shift;
        
        return $class->{_c_volume}->wait_for_wafliron(@_);
}


## @name ::get_wafliron_status
## @summary Gets the status of the wafliron process.
## 
## @description
## Gets the status of the wafliron process. Wrapper around NOTE::Volume::wafliron_status.
##
## @return Hash with rc and output (where output is the return code of NOTE::Volume::wafliron_status).
## rc can be currently: C_RC_SUCCESS/C_RC_WAFLIRON_IDLE/C_RC_NO_WAFLIRON_IN_PROGRESS.
## 
## @examples
## $obj->get_wafliron_status();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_wafliron_status
{
        my $class = shift;
        
        return $class->{_c_volume}->get_wafliron_status(@_);
}


## @name ::stop_wafliron
## @summary Terminates the wafliron process.
## 
## @description
## Terminates the wafliron process. Wrapper around NOTE::Volume::wafliron_stop.
##
## @return Hash with rc and output (where output is the return code of NOTE::Volume::wafliron_stop).
##
## @arg timeout Optional: Timeout value for wafliron to be stopped. 
## If not specified, Wafliron::wafliron_stop will use its own defaults (currently 2400).
##  
## @examples
## $obj->stop_wafliron( timeout => 3600 );
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub stop_wafliron
{
        my $class = shift;
        
        return $class->{_c_volume}->stop_wafliron(@_);
}


## @name ::detect_indirect_type
## @summary Detects the indirect type of a flex volume (32/64 -bit).
## 
## @description
## Detects the indirect type of the volume.
## For flex volumes, this is the indirect type of the containing aggregate.
## For traditional volumes, the indirect type will be defaulted to 32.
##
## @return Hash with rc, indirect_type, and output (status hashref).
##  
## @examples
## $obj->detect_indirect_type();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub detect_indirect_type
{
        my $class = shift;
        
        return $class->{_c_volume}->detect_indirect_type(@_);
}


## @name ::run_wafl_check
## @summary Runs WAFL_check and boots up the filer.
## 
## @description
## Runs WAFL_check and boots up the filer. Wrapper around wack::wack. 
##
## @return Hash with rc and output.
##
## @arg target Optional: The entity to run wack on. Could be: '-all', aggregate name, volume name.
## Defaults to: the aggregate containing this volume (for flex volumes), or
## to the name of this volume (for traditional volumes).
##
## @arg timeout Optional: Timeout for WAFL_check to complete.
##
## @arg compression Optional: Flag to enable compression-related checks (command option '-c').
## Defaults to TRUE.
##
## @arg commit Optional: Flag to control if any wack changes will be commited.
## Defaults to FALSE.
##
## @arg bootimage Optional: Option for boot::transit_to when attempting to reach boot menu.
## (description of what kernel to boot).
##
## @arg expected Optional: Expected return code (defined in Compression::Constants).
## Defaults to C_RC_SUCCESS (which means that wack found the filesystem clean).
##
## @arg disable_boot_up Optional: Flag to be used if filer boot-up is not wanted after WAFL_check completes.
## Defaults to FALSE.
##
## @notes
## If the subroutine is called without the 'target' argument, then an extra check is done to ensure
## that wack is not run on a 64-bit aggregate (the script is terminated as not applicable). If the 'target' argument is specified,
## then no check for the indirect type is done.
##  
## @examples
## $obj->run_wafl_check();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub run_wafl_check
{
        my $class = shift;
        
        return $class->{_c_volume}->run_wafl_check(@_);
}


## @name ::snapshot_reserve
## @summary Get/Set the snap reserve for a particular volume 
## 
## @description
## Get/Set the snap reserve for a particular volume.  Wrapper around NOTE::Volume::snap_reserve
##
## @return Hash with rc and output which equals the original value
##
## @arg percent Optional: If present, sets the snap reserve to the percentage specified and returns the previous setting.
##		If not present, returns the current setting. 
##
##  
## @examples
##	 $obj->snap_reserve( percent=>0 ); # Set no snap reserve
##	 $obj->snap_reserve( percent=>20 ); # Set 20 percent snap reserve
##	 $obj->snap_reserve();
## 
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub snapshot_reserve
{
        my $class = shift;
        
        return $class->{_c_volume}->snapshot_reserve(@_);
}

## @name ::get_reserved_space
## @summary Gets the reserved space of the volume.
## 
## @description
## Gets the reserved space of the volume.
##
## @return Hash with rc and reserved ( in KB )
##  
## @examples
## $obj->get_reserved_space();
## 
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_reserved_space
{
        my $class = shift;
        
        return $class->{_c_volume}->get_reserved_space(@_);
}


## @name ::get_snapshot_info
## @summary get snapshot info for a given snapshot on this volume
## 
## @description
## Get snapshot info for a given snapshot on this volume. Wrapper around SnapCmds::snapshot_list_info.
##
## @return Hash with rc and info which is a reference to a hash with keys as shown below.
##
## @arg name Required: name of snapshot
##
## @examples
## $obj->get_snapshot_info(snapshot=>$snapshot);
##
##	$res{info} example:
##	'access-time' => '1251495943',
##	'cumulative-percentage-of-used-blocks' => '0',
##	'cumulative-percentage-of-total-blocks' => '0',
##	'name' => 'ss_test',
##	'total' => '0',
##	'percentage-of-total-blocks' => '0',
##	'dependency' => '',
##	'cumulative-total' => '0',
##	'busy' => 'false',
##	'percentage-of-used-blocks' => '0'
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_snapshot_info
{
        my $class = shift;
        
        return $class->{_c_volume}->get_snapshot_info(@_);
}


## @name ::set_snap_autodelete
## @summary set snap autodelete state
## 
## @description
## Set snap autodelete on this volume. Wrapper around SnapCmds::snapshot_autodelete_cmd.
##
## @return Hash with rc.
##
## @arg state Required: state to set snap autodelete to
##
## @examples
## $obj->set_snap_autodelete(state=>'on');
## $obj->set_snap_autodelete(state=>'off');
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub set_snap_autodelete
{
        my $class = shift;
        
        return $class->{_c_volume}->set_snap_autodelete(@_);
}


## @name ::set_snap_autodelete_options
## @summary set snap autodelete options
## 
## @description
## Set snap autodelete options on this volume. Wrapper around SnapCmds::snapshot_autodelete_options.
##
## @return Hash with rc.
##
## @arg Optional:
##
## @examples
## $obj->set_snap_autodelete_options(
##						trigger=>'volume', 
##						target_free_space=>80,
##						delete_order=>'oldest_first',);
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub set_snap_autodelete_options
{
        my $class = shift;
        
        return $class->{_c_volume}->set_snap_autodelete_options(@_);
}

## @name ::wreck
## @summary Inject corruption.
##
## @description
## Inject corruption. Wrapper around NOTE::Volume::wreck,
## with added functionality for generating the inject_script.
##
## @arg code Required: Code of the error to be injected.
## Ignored if inject_script arguments is defined.
##
## @arg filename Optional: Name of the file to inject the error to.
## Ignored for non file-specific corruptions and if inject_script is passed in directly.
##
## @arg bno Optional: Specific block number within a file to inject corruption at.
## Ignored for non file-specific corruptions and if inject_script is passed in directly.
##
## @arg start Optional: Start byte for corruption within a block.
## Ignored for non file-specific corruptions and if inject_script is passed in directly.
##
## @arg end Optional: End byte for corruption within a block.
## Ignored for non file-specific corruptions and if inject_script is passed in directly.
##
## @arg fill Optional: Byte to be used when corrupting a specific block pattern.
## Ignored for non file-specific corruptions and if inject_script is passed in directly.
##
## @arg cg_no Optional: CG number at which to injectthe corruption.
## This argument is only accounted for in the case of corrupting the CG (header, trailer).
##
## @arg new_bno Optional: Value to be used to corrupt an indirect block (ind_vol_bno).
##
## @arg new_bno Optional: Value to be used to corrupt an indirect block vvbn (ind_vol_vbno).
##
## @arg inject_script Optional: Text to be used when injecting the error.
##
## @return Hash with rc and output.
##  
## @examples
## $obj->wreck(code => C_CORRUPT_FSINFO_INOS_USED);
## $obj->wreck(inject_script => 'fsinfo_inos_used count=460');
##
## @note
## After injecting corruption, it will also set the wafl_iron_panic_on_corruption flag to 0.
##
## @note
## The way this method requires specific arguments varies by the corruption to be injected.
## Passing in more arguments than what is really needed for a certain corruption,
## will result in the extra arguments being ignored.
## Passing in less arguments than what is needed for a specific corruption,
## will result in script failure, asking for the parameter that is missing and cannot be auto-generated.
## Here are a few examples/rules:
## - If passing inject_script directly, all the other arguments are ignored.
## - For code=C_CORRUPT_FSINFO_INOS_USED: no other arguments are required.
## - For code=C_CORRUPT_FSINFO_BLKS_USED_BY_PLANE0: no other arguments are required.
## - For code=C_CORRUPT_FSINFO_BLKS_COMPRESS_HOLES: no other arguments are required.
## - For code=C_CORRUPT_INODE_BLK_CNT: filename (required).
## - For code=C_CORRUPT_REFCOUNT: filename (required), bno (optional->default: 1).
## - For code=C_CORRUPT_CHECKSUM: filename (required), bno (optional->default: 1), fill (optional->default: randomly generated).
## - For code=C_CORRUPT_CG_HEADER_VBN: filename (required), new_bno (required), cg_no (optional->default: 0)
## - For code=C_CORRUPT_CG_HEADER_VVBN: filename (required), new_vbno (required), cg_no (optional->default: 0)
## - For code=C_CORRUPT_CG_TRAILER_VBN: filename (required), new_bno (required), cg_no (optional->default: 0)
## - For code=C_CORRUPT_CG_TRAILER_VVBN: filename (required), new_bno (required), cg_no (optional->default: 0)
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub wreck
{
        my $class = shift;
        
        return $class->{_c_volume}->inject_error(@_);
}


## @name ::detect_and_fix_injected_error
## @summary Runs the required data corruption detection tool to detect and fix injected errors.
## 
## @description
## Runs the required data corruption detection tool (wafliron/wack) to detect and fix injected errors.
## Runs the required data corruption detection tool (wafliron/wack) again to ensure errors are no longer detected.
##
## @arg corruptions Required: Array ref of codes of the errors to be detected and corrected.
## Corruption codes are defined in Constants.pm.
##
## @arg method Optional: Corruption detection tool (wafliron/wack) to be used to detect and fix the error.
##
## @arg options Optional: Hashref of options for the corruption detection tool.
## Please see documentation for start_wafliron and run_wafl_check for legal options.
## To be noted that wait (for wafliron) and commit (for wack) are overriden to TRUE, for error correction.
##
## @return Hash with rc and output.
##  
## @examples
## $obj->wreck(code => C_CORRUPT_FSINFO_INOS_USED);
## $obj->wreck(inject_script => 'fsinfo_inos_used count=460');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub detect_and_fix_injected_error
{
        my $class = shift;
        my (%args) = validate( @_,
		{
                        method => { type => SCALAR, regex => qr/^(wack|wafliron)$/, default => 'wafliron', },
                        options => { type => HASHREF, default => {compression => TRUE} },
                        corruptions => { type => ARRAYREF },
                } );
        
        my $detection_method = $args{method};
        my $corruptions_expected = $args{corruptions};
        my $detection_options = $args{options};
        my $detection_output;
        my %detection_res_hash;
        
        if ($detection_method eq 'wafliron')
        {
                # Force wafliron wait and expected error
                $detection_options->{wait} = TRUE;
                $detection_options->{expected} = C_RC_WAFLIRON_CORRUPTION_FOUND;
                %detection_res_hash = $class->start_wafliron(%$detection_options);
        }
        elsif ($detection_method eq 'wack')
        {
                # Force wack commit and expected error
                $detection_options->{commit} = TRUE;
                $detection_options->{expected} = C_RC_WACK_ERRORS_FOUND;
                %detection_res_hash = $class->run_wafl_check(%$detection_options);
        }
        else
        {
                logresult('script', "Unsupported corruption detection method: $detection_method");
        }

        $detection_output = $detection_res_hash{output};
        logcomment("$detection_method completed: output:\n$detection_output") if ($class->{_debug});
        
        my $corruptions_not_found = '';
        foreach my $corruption_code (@$corruptions_expected)
        {
                logcomment("Verifying that corruption \"$corruption_code\" was detected by $detection_method");
                my $expected_corruption_messages = $class->{_ui}->get_corruption_messages(code => $corruption_code,
                                                                                          detection_method => $detection_method);
                logcomment("Starting to look for expected corruption messages in the $detection_method output. Looking for:\n" .
                           Dumper($expected_corruption_messages)) if ($class->{_debug});
        
                my $found_corruption = FALSE;
                foreach my $error_string (@$expected_corruption_messages)
                {
                        if ($detection_output =~ /$error_string/i)
                        {
                                logcomment("Success: corruption message \"$error_string\" was found in the $detection_method output");
                                $found_corruption = TRUE;
                        }
                }
        
                if ($found_corruption == FALSE)
                {
                        logcomment("Failed to detect corruption message for \"$corruption_code\" in the $detection_method output.\n" .
                                   "Looked for:\n" . Dumper($expected_corruption_messages));
                        $corruptions_not_found .= "$corruption_code, "
                }
        
        }
        
        if ($corruptions_not_found ne '')
        {
                logresult('fatal', "$detection_method failed to detect the following corruption(s): $corruptions_not_found\n\n" .
                                   "$detection_method output:\n$detection_output\n");
        }
        
        logcomment("Running $detection_method again, to verify that the corruption was fixed");
        
        if ($detection_method eq 'wafliron')
        {
                # Force wafliron wait
                $detection_options->{wait} = TRUE;
                $detection_options->{expected} = C_RC_SUCCESS;
                %detection_res_hash = $class->start_wafliron(%$detection_options);
        }
        elsif ($detection_method eq 'wack')
        {
                # Set wack commit to FALSE
                $detection_options->{commit} = FALSE;
                $detection_options->{expected} = C_RC_SUCCESS;
                %detection_res_hash = $class->run_wafl_check(%$detection_options);
        }
        
        return %detection_res_hash;
}


## @name ::size
## @summary 
## Set size of volume.
## 
## @description
## Set or get size of volume.  Wrapper around NOTE::Volume::size
## 
## @return rc and size
##
## @arg size Required: 
##
## @examples
## $obj->size(size=>'500m');
##
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub size
{
        my $class = shift;
        
        return $class->{_c_volume}->size(@_);
}


## @name ::nfs_export_options
## @summary Gets or sets NFS export options on the volume.
## 
## @description
## Gets or sets NFS export options on the volume. Wrapper around NOTE::Volume::exportfs.
##
## @arg new_setting Optional: New setting for the NFS export option.
## If not specified, the current setting is returned.
##
## @note
## If called multiple times,
## this function will not re-issue the exportfs command to the filer unless the exportfs option string changes.
## Also, even if called to get the current setting, if there is a difference between the current setting on the filer and
## what this setting should have been (last set value), then the filer setting is resynced to match this last-set value.
##
## @return Hash with rc and output (representing the current option string, if new_setting is not defined,
## and the NOTE vol obj if new_setting is defined).
##  
## @examples
## $obj->nfs_export_options($new_setting => C_DFLT_VOL_EXPORTFS)
##
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub nfs_export_options
{
        my $class = shift;
        
        return $class->{_c_volume}->nfs_export_options(@_);
}


## @name ::set_autosize
## @summary set vol autosize state
## 
## @description
## Set vol autosize state
##
## @return Hash with rc.
##
## @arg state Required: state to set autosize to
##
## @examples
## $obj->set_autosize(state=>'on', size=>'30g');
##
## @author waroff@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub set_autosize
{
        my $class = shift;
        
	return $class->{_c_volume}->set_autosize(@_);
}


## @name ::is_block_type_32bit
## @summary Checks if the volume is 32-bit.
## 
## @description
## Checks if the volume is 32-bit.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc.
##  
## @examples
## $obj->is_block_type_32bit(expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_block_type_32bit
{
        my $class = shift;
	
	return $class->{_c_volume}->is_block_type_32bit(@_);
}


## @name ::is_block_type_64bit
## @summary Checks if the volume is 64-bit.
## 
## @description
## Checks if the volume is 64-bit.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc.
##  
## @examples
## $obj->is_block_type_64bit(expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_block_type_64bit
{
        my $class = shift;
	
	return $class->{_c_volume}->is_block_type_64bit(@_);
}


## @name ::is_block_type_upgrading
## @summary Checks if the volume is 'upgrading'.
## 
## @description
## Checks if the volume is is_block_type_upgrading.
##
## @arg expected Optional: Expected return code: C_RC_SUCCESS/C_RC_FAILURE
## (defined in Compression::Constants).
##
## @return Hash with rc.
##  
## @examples
## $obj->is_block_type_upgrading(expected => C_RC_FAILURE);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub is_block_type_upgrading
{
        my $class = shift;
	
	return $class->{_c_volume}->is_block_type_upgrading(@_);
}


## @name ::wait_for_block_type_64bit
## @summary Waits until the volume is marked as 64-bit.
## 
## @description
## Waits until the volume is marked as 64-bit.
##
## @arg timeout Optional: Timeout value (in seconds).
## Default is C_DFLT_TIMEOUT, as defined in Constants.pm.
##
## @arg check_interval Optional: Number of seconds between two consecutive status checks.
## Defaults to 1 sec.
##
## @return Hash with rc.
##  
## @examples
## $obj->wait_for_block_type_64bit(timeout => 600);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub wait_for_block_type_64bit
{
        my $class = shift;
	
	return $class->{_c_volume}->wait_for_block_type_64bit(@_);
}


## @name ::start_64_bit_upgrade
## @summary Starts upgrading the volume to 64-bit.
## 
## @description
## Starts upgrading the volume to 64-bit.
##
## @arg mode Optional: Upgrade mode: check|normal|grow-reserved|grow-all.
## Default: none.
##  
## @examples
## $obj->start_64_bit_upgrade();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub start_64_bit_upgrade
{
        my $class = shift;
        
        return $class->{_c_volume}->start_64_bit_upgrade(@_);
}


## @name ::check_buftree
## @summary Runs buftree check on this volume.
## 
## @description
## Runs buftree check on this volume.
##
## @arg filename Optional: Name of file to run buftree check on.
##
## @arg priv Optional: Priv level.
##
## @arg dont_panic Optional: Flag to disable the filer panic
## in case the buftree checker finds an inconsistency.
##
## @arg timeout Optional: Timeout in seconds.
##
## @return Hash with rc and buftree_stats.
##  
## @examples
## $obj->check_buftree();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub check_buftree
{
        my $class = shift;
        
        return $class->{_c_volume}->check_buftree(@_);
}


## @name ::verify_buftree_layout
## @summary Runs buftree check on this volume
## and verifies individual results.
## 
## @description
## Runs buftree check on this volume and verifies individual results.
##
## @arg filename Optional: Name of file to run buftree check on.
##
## @arg priv Optional: Priv level.
##
## @arg expected_values Required: Hashref of expected values.
##
## @arg dont_panic Optional: Flag to disable the filer panic
## in case the buftree checker finds an inconsistency.
##
## @arg timeout Optional: Timeout in seconds.
##
## @return Hash with rc and buftree_stats.
##  
## @examples
## $obj->verify_buftree_layoutverify_buftree_layout(filename => 'file1',
##                                                  expected_values => {'0' => {holes32 => 0},
##                                                                      'total' => {holes64 => 0,
##                                                                                  holes32 => 0}});
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub verify_buftree_layout
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        filename => {type => SCALAR, optional => 1},
                        priv => {type => SCALAR, optional => 1},
                        dont_panic => {type => SCALAR, optional => 1},
                        timeout => {type => SCALAR, optional => 1},
                        expected_values => {type => HASHREF},
                        _non_fatal_failure => {type => SCALAR, default => FALSE}
                } );
        
        my $expected_vals = delete($args{expected_values});
        my $non_fatal_failure = delete($args{_non_fatal_failure});
        
        my %res = $class->check_buftree(%args);
        my $current_layout = $res{buftree_stats};
        
        logcomment("Current buftree layout:\n" . Dumper($current_layout));
        
        foreach my $level (keys %$expected_vals)
        {
                my $expected = $expected_vals->{$level};
                my $obtained = $current_layout->{$level};
                
                foreach my $field (keys %$expected)
                {
                        logcomment("Verifying buftree layout: field \"$field\", level \"$level\":\n" .
                                   "Expected: " . $expected->{$field} . "\n" .
                                   "Obtained: " . $obtained->{$field});
                        
                        if ((!(defined($obtained->{$field}))) || ($obtained->{$field} != $expected->{$field}))
                        {
                                if ($non_fatal_failure == TRUE)
                                {
                                        $res{rc} = C_RC_FAILURE;
                                        return (%res);
                                }
                                logresult('fatal', "Buftree check: values mismatch for field \"$field\", level \"$level\":\n" .
                                                   "Obtained: " . $obtained->{$field} . "\n" .
                                                   "Expected: " . $expected->{$field});
                        }
                }
        }
        
        return %res;
}


## @name ::wait_for_buftree_layout
## @summary Waits until buftree check shows specific values.
## 
## @description
## Waits until buftree check shows specific values.
##
## @arg filename Optional: Name of file to run buftree check on.
##
## @arg priv Optional: Priv level.
##
## @arg expected_values Required: Hashref of expected values.
##
## @arg dont_panic Optional: Flag to disable the filer panic
## in case the buftree checker finds an inconsistency.
##
## @arg timeout Optional: Timeout in seconds. Default: C_DFLT_TIMEOUT.
##
## @return Hash with rc and buftree_stats.
##  
## @examples
## $obj->verify_buftree_layoutverify_buftree_layout(filename => 'file1',
##                                                  expected_values => {'0' => {holes32 => 0},
##                                                                      'total' => {holes64 => 0,
##                                                                                  holes32 => 0}});
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub wait_for_buftree_layout
{
        my $class = shift;
        my (%args) = validate( @_,
                {
                        filename => {type => SCALAR, optional => 1},
                        priv => {type => SCALAR, optional => 1},
                        dont_panic => {type => SCALAR, optional => 1},
                        timeout => {type => SCALAR, default => C_DFLT_TIMEOUT},
                        expected_values => {type => HASHREF},
                } );
        
        my $timeout = delete($args{timeout});
        $args{_non_fatal_failure} = TRUE;
        my $rc = C_RC_FAILURE;
        my $delay = 5;
        my %res;
        
        logcomment("Waiting for buftree layout.\nTimeout: $timeout\nExpected values:\n" . Dumper($args{expected_values}));
        while (($timeout >= 0) && ($rc != C_RC_SUCCESS))
        {
                logcomment("Waiting for buftree layout: timeout = $timeout");
                %res = $class->verify_buftree_layout(%args);
                $rc = $res{rc};
                sleep($delay);
                $timeout -= $delay;
        }
        
        if ($timeout < 0)
        {
                logresult('fatal', "Timed-out while waiting for the buftree layout:\n" . Dumper(%res))
        }
        
        return %res;
}


## @name ::set_snaplock_retention_period
## @summary Sets the snaplock retention period for this volume.
## 
## @description
## Sets the snaplock retention period for this volume.
##
## @arg default Optional: Value to be used to set the default retention period.
##
## @arg minimum Optional: Value to be used to set the minimum retention period.
##
## @arg maximum Optional: Value to be used to set the maximum retention period.
##
## @return Hash with rc and new retention periods.
##  
## @examples
## $obj->set_snaplock_retention_period(default => '90s',
##                                     minimum => '90s');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub set_snaplock_retention_period
{
        my $class = shift;
	
	return $class->{_c_volume}->set_snaplock_retention_period(@_);
}


## @name ::set_snaplock_auto_commit
## @summary Sets the snaplock auto-commit period for this volume.
## 
## @description
## Sets the snaplock auto-commit period for this volume.
##
## @arg value Required: Value to be set for the auto-commit period.
##
## @return Hash with rc.
##  
## @examples
## $obj->set_snaplock_auto_commit(value => '10s');
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub set_snaplock_auto_commit
{
        my $class = shift;
	
	return $class->{_c_volume}->set_snaplock_auto_commit(@_);
}


## @name ::get_expiry_date
## @summary Gets the retention expiration date for this volume.
## 
## @description
## Gets the retention expiration date for this volume.
##
## @return Hash with rc and expiry_date.
##  
## @examples
## my %res = $obj->get_expiry_date();
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub get_expiry_date
{
        my $class = shift;
	
	return $class->{_c_volume}->get_expiry_date(@_);
}


## @name ::verify_snaplock_containment_bits
## @summary Verifies snaplock containment bits for this volume.
## 
## @description
## Verifies snaplock containment bits for this volume.
##
## @arg containment_version_slc Optional: Expected containment_version_slc value.
##
## @arg containment_version_sle Optional: Expected containment_version_sle value.
##
## @arg containment_version_highest_slc Optional: Expected containment_version_highest_slc value.
##
## @arg containment_version_highest_sle Optional: Expected containment_version_highest_sle value.
##
## @note
## At least one of the arguments must be specified, otherwise no check is done.
##
## @return Hash with rc and snaplock_containment_stats.
##  
## @examples
## $obj->verify_snaplock_containment_bits(containment_version_slc => 0,
##                                        containment_version_highest_slc => 0);
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)
##

sub verify_snaplock_containment_bits
{
        my $class = shift;
        my (%args) = validate( @_,
        {
                containment_version_slc => {type => SCALAR, optional => 1,},
                containment_version_sle => {type => SCALAR, optional => 1,},
                containment_version_highest_slc => {type => SCALAR, optional => 1,},
                containment_version_highest_sle => {type => SCALAR, optional => 1,},
        });
        
        my %res = $class->{_c_volume}->get_snaplock_containment_bits();
        
        foreach my $field (keys %args)
        {
                my $obtained = $res{snaplock_containment_stats}->{$field};
                my $expected = $args{$field};
                
                logcomment("Verifying snaplock containment bit: $field\n" .
                           "Obtained value: $obtained\n" .
                           "Expected value: $expected");
                if (!defined($obtained) || ($obtained != $expected))
                {
                        logresult('fatal', "Unexpected snaplock containment bit value: $field\n" .
                                           "Expected: $expected\n" .
                                           "Obtained: $obtained");
                }
        }
        
        return %res
}


## @name ::get_stats
## @summary Obtain volume stats.
## 
## @description
## Obtain volume stats.
##
## @arg expected_values Optional: Hashref of key-values of expected volume stats.
## If this parameter is used, then all the required fields are checked.
## It retries in case of mismatch (up to 5 attempts).
## Available fields: instance_uuid, parent_aggr, avg_latency, total_ops,
## read_data, read_latency, read_ops, write_data, write_latency, write_ops,
## other_latency, other_ops, internal_msgs, read_blocks, write_blocks,
## synchronous_frees, asynchronous_frees, df_worker,
## nfs_read_data, nfs_read_latency, nfs_read_ops, nfs_write_data, nfs_write_latency,
## nfs_write_ops, nfs_other_latency, nfs_other_ops,
## cifs_read_data, cifs_read_latency, cifs_read_ops, cifs_write_data, cifs_write_latency,
## cifs_write_ops, cifs_other_latency, cifs_other_ops,
## san_read_data, san_read_latency, san_read_ops, san_write_data, san_write_latency',
## san_write_ops, san_other_latency, san_other_ops,
## fcp_read_data, fcp_read_latency, fcp_read_ops, fcp_write_data,
## fcp_write_latency, fcp_write_ops, fcp_other_latency, fcp_other_ops,
## iscsi_read_data, iscsi_read_latency, iscsi_read_ops, iscsi_write_data, iscsi_write_latency,
## iscsi_write_ops, iscsi_other_latency, iscsi_other_ops,
## flexcache_read_data, flexcache_read_ops, flexcache_write_data, flexcache_write_ops,
## flexcache_other_ops, flexcache_send_data, flexcache_receive_data,
## wv_fsid, wv_vol_type, wv_fsinfo_fs_version, wv_volinfo_fs_options, wv_volinfo_fs_flags,
## wv_fsinfo_blks_total, wv_fsinfo_blks_reserve', wv_fsinfo_blks_used, wv_fsinfo_blks_used_by_plane0,
## wv_fsinfo_blks_blks_rsrv_holes_cifs, wv_fsinfo_blks_blks_rsrv_holes, wv_fsinfo_blks_blks_rsrv_overwrite,
## wv_fsinfo_blks_rsrv_absents, wv_fsinfo_blks_snap_reserve_pct, wv_fsinfo_blks_res_state,
## wv_fsinfo_blks_overwrite_slider_pct, wv_fsinfo_blks64_blks_rsrv_holes_cifs, wv_fsinfo_blks64_blks_rsrv_holes,
## wv_fsinfo_public_inos_total, wv_fsinfo_public_inos_reserve, wv_fsinfo_public_inos_used, wv_fsinfo_private_inos_total,
## wv_fsinfo_private_inos_reserve, wv_fsinfo_private_inos_used, wv_fsinfo_blkr_cp, wv_fsinfo_blks_vvol_dbys_df_cnt,
## wvblk_past_eof, wvblk_past_eof64, wvblk_snap_reserve, wvblk_zombie_blks, wv_fsinfo_blks_rsrv_parent,
## wvblk_spcres_in_parent, wvblk_rsrv_parent_holes, wvblk_rsrv_parent_overwrite, wvblk_rsrv_parent_overwrite_always,
## wvbd_active_frees, wvbd_whole_frees, wvbd_active_frees_y, wvbd_owner_changed_y, volume:\S+:wvbd_whole_frees_o,
## wvbd_whole_frees_o, wvblk_saved_fsinfo_public_inos_total, wvblk_saved_fsinfo_public_inos_reserve, wvblk_saved_fsinfo_public_inos_used,
## wvblk_saved_fsinfo_private_inos_total, wvblk_saved_fsinfo_private_inos_reserve, wvblk_saved_fsinfo_private_inos_used,
## wvblk_delalloc, wvblk_rsrv_delalloc, wvblk_rsrv_holes_cifs, wvblk_rsrv_holes, wvblk_rsrv_overwrite, wvblk_rsrv_absents,
## wvblk_rsrv_holes_cifs64, wvblk_rsrv_holes64, extent_size, wvdf_enabled, wvdf_inconsistent_scores, wvdf_last_fbn,
## wvdf_water_mark, wvdf_max_frees, wvdf_total_score, wvblk_parent_to_be_reclaimed, wvip_vvol_container_wi_blk_cnt,
## wvip_vvol_container_wi_size, wvip_public_inofile_wi_blk_cnt, wvip_private_inofile_wi_blk_cnt, wvip_vvol_container_indirects,
## wvol_number_suspended, clone_storage_blocks, wvblk_reclaim_time_start, wvblk_reclaim_time_reset, wvblk_reclaim_time_done,
## wvblk_reclaim_time_abort, wvzmb_num_zmsgs_inuse, wvblk_ind_delalloc, wvsblk_vvrd_last_fbn, wvsblk_vvrd_spcflags,
## wvsblk_vvrd_spc_clone_inherited, wvsblk_space_tax, wvsblk_lev0_over_nominal, wvsblk_vvrd_vol_size, wvsblk_vvrd_flags,
## wv_fsinfo_containment_version_slc, wv_fsinfo_containment_version_sle, wv_fsinfo_containment_version_compression,
## wv_fsinfo_containment_version_vmalign, wv_fsinfo_containment_version_spare1, wv_fsinfo_containment_version_spare2,
## wv_fsinfo_containment_version_highest_slc, wv_fsinfo_containment_version_highest_sle, wv_fsinfo_containment_version_highest_compression,
## wv_fsinfo_containment_version_highest_vmalign, wv_fsinfo_containment_version_highest_spare1, wv_fsinfo_containment_version_highest_spare2,
## wvi2p_wip_wi_size, bad_user_blk_read_error_propagated, bad_user_blk_read_error_not_propagated, bad_fixable_blk_read_error_propagated,
## bad_fixable_blk_read_error_not_propagated, bad_non_fixable_blk_read_error_propagated, bad_non_fixable_blk_read_error_not_propagated,
## bad_fixable_metafile_blk_read_error_propagated, bad_fixable_metafile_blk_read_error_not_propagated, bad_zombie_ind_blk_read_err_propagated,
## bad_zombie_ind_blk_read_err_not_propagated, bad_container_user_blk_read_error_propagated, bad_container_user_blk_read_error_not_propagated,
## bad_container_fixable_afs_blk_read_error_propagated, bad_container_fixable_afs_blk_read_error_not_propagated,
## bad_container_fixable_snap_blk_read_error_propagated, bad_container_fixable_snap_blk_read_error_not_propagated,
## bad_container_non_fixable_blk_read_error_propagated, bad_container_non_fixable_blk_read_error_not_propagated,
## wv_playlist_hits, wv_playlist_misses, wv_playlist_not_present, wv_playlist_entries, wv_playlist_reqs, wv_playlist_no_raidbufs,
## wv_playlist_getbuf_failures, wv_playlist_cont_indirects, wv_playlist_vvbn_holes, wv_playlist_prefetch_not_started,
## wv_playlist_prefetch_start_time, wv_playlist_prefetch_end_time, wv_playlist_load_end_time, wv_playlist_apfi_used_slots,
## wv_playlist_apfi_collision_inserts, wv_playlist_apfi_collision_accesses, wvsnap_ondisk_count, wvsnap_incore_count, wvsnap_loaded_total,
## delete_log_dlog_in_use, delete_log_dlog_ind_blk_cnt, delete_log_dlog_blk_cnt, delete_log_purge_ind_blk_cnt, delete_log_purge_blk_cnt',
## delete_log_dlog_hole_cnt, delete_log_dlog_cap, delete_log_overflow_in_use, delete_log_overflow0_blk_cnt,
## delete_log_overflow1_blk_cnt, delete_log_del_log, delete_log_purge_send, delete_log_purge_block,
## delete_log_purge_vbns, delete_log_dlog_frees, delete_log_dlog_vbn_clears, delete_log_dlog_keys, delete_log_pre_split_failed,
## delete_log_pre_bufs_loaded, delete_log_purge_overflow, delete_log_purge_overflow_susp, delete_log_purge_overflow_no_split,
## delete_log_trunc_overflow_of_append, delete_log_dlog_overuse_mem, delete_log_dlog_overflow, delete_log_overflow_total,
## delete_log_act_dirty, delete_log_pre_overflow, delete_log_overflow_vbns, delete_log_overflow_frees, delete_log_overflow_frees_ind,
## delete_log_overflow_commits, delete_log_overflow_send, delete_log_overflow_susp, delete_log_overflow_throttle, msgs_rejected_in_nvfailed_state,
## msgs_allowed_in_nvfailed_state, clone_inodes, clone_lookups, clone_lookup_hits, clone_blks_copied, clone_eio_blks,
## clone_num_entries, clone_max_entries, clone_max_hierarchy, clone_unsplit_snap_entries, clone_num_share_stopped,
## clone_read_redirected, clone_split_ra, clone_split_ra_lag, clone_inline_split_kireeti_in_progress, clone_inline_split_stage_size_limitation,
## clone_inline_split_beyond_eof, clone_inline_split_source_destination_dirty, clone_inline_split_source_spec_vbn,
## clone_inline_split_enospc, clone_inline_split_edquot, nonzero_dbys_cnt.
##
## @return Hash with rc and vol_stats.
##  
## @examples
## my %stats = $obj->get_stats();  # No expected-values comparison done
## $obj->get_stats(expected_values => {wv_fsinfo_blks_used => 44,},);
##
sub get_stats
{
	my $class = shift;
	my (%args) = validate( @_,
	{
		expected_values => {type => HASHREF, optional => 1,},
	});
	
	my $expected_values = $args{expected_values};
	my $volume = $class->{vol_name};
	my $cmd_code = C_CMD_GET_VOLUME_STATS;
	
	my $command = $class->{_ui}->formulate_command(code   => $cmd_code,
						       volume => $volume,);
	my $output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_c_volume}->_get_conn(),
							    boot_mode => $class->{_boot_mode},
							    cmd => $command,
							    node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
	my %current = $class->{_ui}->process_output_string(code => $cmd_code,
							   output => $output,
							   expected_rc => C_RC_SUCCESS,
							   log_message => "Obtain volume stats for: $volume",);
	
	if (defined($expected_values))
	{
		for (my $i = 1; $i <= 5; $i++)
		{
			logcomment("Checking volume stats: attempt $i");
			my $fatal = ($i == 5) ? TRUE : FALSE;
			my $match = Compression::Utils::compare_hash_entries(obtained => $current{vol_stats},
									     expected => $expected_values,
									     values_meaning => "volume stats for $volume",
									     fatal_failure => $fatal,);
			if ($match == TRUE)
			{
				last;
			}
			
			sleep(ONE_CP_TIME);
			$output = Compression::Utils::_send_cmd_to_filer(conn => $class->{_c_volume}->_get_conn(),
									 boot_mode => $class->{_boot_mode},
									 cmd => $command,
									 node_run => $class->{_ui}->is_dblade_level_command(code => $cmd_code),);
			%current = $class->{_ui}->process_output_string(code => $cmd_code,
									output => $output,
									expected_rc => C_RC_SUCCESS,
									log_message => "Obtain volume stats for: $volume",);
		}
	}
	
	return %current;
}




################################
# Internal subroutines
################################


sub _get_underlying_vol_obj
{
        my $class = shift;
        return $class->{_c_volume}->{_vol};
}


# Keep this method internal. Renaming should be done only through Compress.
# This way, all members of the Compress object will have the volume name adjusted properly.
sub _rename
{
        my $class = shift;
        
        my $new_name = $class->{_c_volume}->_rename(@_);
        $class->{vol_name} = $new_name;
        
        return (rc => C_RC_SUCCESS, output => "");
}


sub _refresh_aggregate
{
	my $class = shift;
	$class->{_c_volume}->_refresh_aggregate(@_);
}

sub _update_vfiler
{
	my $class = shift;
	
	$class->{_c_volume}->_refresh_vfiler(@_);
	$class->{_vfiler_name} = $class->{_c_volume}->{_vfiler_name};
}

sub _get_mountpoint
{
        my $class = shift;
        return $class->{_c_volume}->{_mountpoint};
}



# Any necessary clean-up upon object destruction
sub DESTROY
{
	my $class = shift;
        
        # WARNING: We need to store any eventual error message and fail it at the end of any destructor.
        # Otherwise, it looks like ntest will evaluate the 'destructor execution' and any real error will be lost.
        # At this moment, our fix is reliable, but, if tests starts to not detect failures, this is the place to change.
        # WARNING: To enable the clean-up rubroutines to detect NA results, the logresult message should begin with 'NOT APPLICABLE: ':
        # logresult ('NA', "NOT APPLICABLE: .....");
        my $execution_error = $@;
        my $volume = $class->{vol_name};
        
        $class->{_c_volume}->{_disable_cleanup} = $class->{_disable_cleanup};
        
        if (($class->{_disable_cleanup} == FALSE) &&
            ($class->{_creator_pid} == $$) &&
            (!defined($ENV{'C_CORRUPTION_FOUND'}) || ($ENV{'C_CORRUPTION_FOUND'} != TRUE)))
	{
		logcomment("Executing Volume clean-up for volume \"$volume\"");
		
                eval
                {
                        if ((defined($ENV{'WAFLIRON_FLAG'})) && ($ENV{'WAFLIRON_FLAG'} == TRUE))
                        {
                                logcomment("Volume: cleanup: should run wafliron") if ($class->{_debug});
                                if (!(defined($ENV{'WAFLIRON_IS_DONE'})) || ($ENV{'WAFLIRON_IS_DONE'} == FALSE))
                                {
                                        logcomment("Volume: cleanup: waiting for wafliron...");
                                        $class->start_wafliron( wait => TRUE );
                                        $ENV{'WAFLIRON_IS_DONE'} = TRUE;
                                }
                        }
                        
                        logcomment("Volume: cleanup: unmounting $volume from all known hosts\n" . Dumper($class->{_mounted_on})) if ($class->{_debug});
                        while (defined($class->{_mounted_on}->[0]))
                        {
                                $class->nfs_umount(client => $class->{_mounted_on}->[0]);
                        }
                };
                
                if( $@ )
                {
                        logcomment( "Volume: cleanup for volume \"$volume\": failure due to : $@" );
                }
                else
                {
                        logcomment( "Volume: cleanup for volume \"$volume\": done" ) if ($class->{_debug});
                }
                
                $class->{_c_volume} = undef;
        }
	else
	{
                logcomment( "Volume: cleanup for volume \"$volume\": nothing done: script PID: " . $$) if ($class->{_debug});
	}
        
        if (defined($execution_error))
        {
                _c_interpret_logresult($execution_error);
        }
}


# routine to test the module itself
sub testme
{
	package main;
	use Compression::Constants qw(:all);
	use Compression::Compress;
	use vars qw($FILER $AGGREGATE $VOLUME );

	require "Compression/setup_core.pl";

	my $compress = Compression::Compress->setup( 
			filer => $FILER,
			aggregate => $AGGREGATE,
			volume => $VOLUME, 
		);

	$compress->{volume}->bring_offline();
        $compress->{volume}->bring_online();
        #$compress->{volume}->set_option( service => C_OPTION_COMPRESSION, new_value => FALSE );
        #$compress->{volume}->verify_option( service => C_OPTION_COMPRESSION, expected => C_RC_FAILURE );
        $compress->{volume}->nfs_mount();
        $compress->{volume}->nfs_umount();

	logcomment( "Volume: testme: Done" );
}


1;
