#!perl -w
# Hey emacs, this is basically -*- perl -*-
#
#Copyright (c) 2009 Network Appliance, Inc.
#All rights reserved.
#
## @summary Ivan 
##
## @description
##
## Below is the middle portion of Ivan as described by the Ivan The Intilizer Test Bed 
## http://wikid.netapp.com/w/Storage_Efficiency/Compression/streets_testbed_project
##
## @usage
##
## @keywords StressBed TestBed Ivan 
## @dependencies
## @author pade@netapp.com, kozel@netapp.com
##
## @status private
##

package Compression::tools::Initializer;

use strict;
use warnings;

use Tharn;
use XML::Simple qw(XMLin);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Carp qw(croak);
use QtreeCmds qw(qtreeList qtreeCreate);
use Admin::License qw(license :constants);
use SnapMirror qw(snapmirror_enable snapmirror_initialize snapmirror_status snapmirror_status_list snapmirror_break snapmirror_release);
use SnapMirror::Helper qw(sm_set_config);
use SnapCmds qw(snapsched_volume);
use Volume qw(vol_restrict vol_options vol_online vol_offline vol_restrict vol_container vol_list vol_size vol_status vol_root vol_destroy vol_create_flex vol_exists vol_wait_change);
use SnapVault qw(sv_toggle sv_start sv_stop snapvault_release);
use Utils::Units qw(:constants convert_units convert_to_bytes);
use Utils qw(make_host make_conn);
use Aggregate qw(aggr_purge aggr_wait_change aggr_rename aggr_add aggr_list aggr_disks aggr_root aggr_create aggr_mirror);
use filer qw(filer_disk_capacity_get);
use Dense qw(dense_toggle);


#CONSTANTS
my $TIMES_TO_PING_EACH_HOST = 1;          #default 10 seconds
my $TIMEOUT_FOR_EACH_HOST   = 2;          #default 2 seconds 
my $TIMEOUT_FOR_EACH_FILER  = 5;          #default 5 seconds 
my $TIMEOUT_FOR_EACH_VOLUME = 5;          #default 5 seconds 
my $TIMEOUT_FOR_SSH_CHECK_ON_FILER = 10;  #default 10 seconds
my $TIMEOUT_FOR_FREE_SPACE_ON_FILER = 10; #default 10 seconds
my $TIMEOUT_FOR_VOL_CREATION = 3600;	  #default 3600 seconds
my $TERABYTE_IN_BYTES = 1099511627776;
my $GIGABYTE_IN_BYTES = 1073741824;
my $MEGABYTE_IN_BYTES = 1048576;
my $MIN_DISKS_FOR_RAIDDP = 3;

## @name ::run
##
## @summary main subroutine
##
## @description
## Parses the xml and calls subroutines and libraries to handle the setup
##
## @return 1
##
## @arg configfile Required: path to the configuration xml file
## @arg reuse Optional: flags the reuse of volumes when possible
##
## @examples
## use Compression::Testbed::Initializer;
## run( configfile => "/path/to/testbed.xml" );
##
## @author kozel@netapp.com, pade@netapp.com, dl-compression-qa@netapp.com

sub run
{
	my (%args) = validate (@_, {	configfile => { type => SCALAR },
									reuse => { type => SCALAR, optional => 1},
									DESTROYALLAGGREGATES => { type => SCALAR, optional => 1, default=>0 }
							});
	
	my $configFile = $args{configfile};
	#my $reuse = $args{reuse};
	my $destroyAllAggregates = $args{DESTROYALLAGGREGATES};
	my $reuse = 'FALSE';
	my $summary = '';

	# Parse the xml into hashes, forcearray turns the quoted tags into arrays for easy looping
	my $config = XMLin( $configFile, forcearray => ['filer', 'aggregate', 'volume', 'option', 'voldefaults', 'feature'], keyattr => [] );
	
	my $testbedName = $config->{name};
	$summary .= ("\n\n$configFile parsed. Testbed $testbedName will be setup.\n");
	 
	$summary .= &cleanup_filers( configuration => $config, reuse => $reuse );

	&create_aggregates( configuration => $config , destroy => $destroyAllAggregates);

	$summary .= &create_volumes( configuration => $config, reuse => $reuse );

	# license and set defined options
	$summary .= &configure_volume_options( configuration => $config );
	$summary .= &create_qtrees( configuration => $config );

	# license and establish relationships
	$summary .= &configure_relationships( configuration => $config );

	logcomment $summary;
	return 1;
}

# Internal function
# name ::configure_volume_options
#
# summary (INTERNAL) call subroutines to set options
#
# description
# Called by the main segement to traverse the configuration
# and call subroutines to handle setting up options on volumes
#
# @arg configuration	Configuration data structure of the parsed XML
# 
sub configure_volume_options
{
	my (%args)	= validate (@_,	{	configuration	=> { type => HASHREF },
								} );
	my $config	= $args{configuration};
	my $summary	= '';

	## Process each filer for this configuration
	foreach my $filer (@{$config->{filer}}) {
		my $hostname = $filer->{hostname};
		logcomment("\n\nHandling options and qtrees on $hostname\n");
		
		## Process each aggregate for this filer
		foreach my $aggregate (@{$filer->{aggregate}}) {
			my $aggregateName = $aggregate->{name};
		
			my $defaultSize;
			my %defaultOptions = ();
			## Process volume defaults for this aggregate
			foreach my $volDefault (@{$aggregate->{voldefaults}}) {
				logcomment("Processing default values for $aggregateName\n");
				$defaultSize = $volDefault->{size};
				logcomment("\tDefault volume size is $defaultSize\n");
				foreach my $option (@{$volDefault->{option}}) {
					$defaultOptions{$option->{name}} = $option->{value};
					logcomment("\tDefault value for $option->{name} is $option->{value}\n");
				}
			}

			## Process each volume for this aggregate
			foreach my $volume (@{$aggregate->{volume}}) {
				my $volName = $volume->{name};
				logcomment("Processing $volName in $aggregateName on $hostname\n");

				## Configure each option
				# Merge defaultOptions with volume specific options
				my %volOptions = %defaultOptions;
				foreach my $option (@{$volume->{option}}) {
					$volOptions{$option->{name}} = $option->{value};
				}

				while ( my ($name, $value) = each(%volOptions) ) {
					&_configure_option($name, $value, $hostname, $volName);
					logcomment("\tSetting option $name to $value\n");
				}
			}
		}
	}
	return $summary;
}

# Internal function
# name ::create_qtrees
#
# summary (INTERNAL) Creates Qtrees from configuration
#
# description
# Called by the main segement to traverse the configuration
# and call subroutines to handle creating qtrees.
#
# @arg configuration	Configuration data structure of the parsed XML
# 
sub create_qtrees
{
	my (%args)	= validate (@_,	{	configuration	=> { type => HASHREF },
								} );
	my $config	= $args{configuration};
	my $summary	= '';

	## Process each filer for this configuration
	foreach my $filer (@{$config->{filer}}) {
		my $hostname = $filer->{hostname};
		logcomment("\n\nCreating qtrees on $hostname\n");
		
		## Process each aggregate for this filer
		foreach my $aggregate (@{$filer->{aggregate}}) {
		
			my $defaultNumQtrees;
			## Process volume defaults for this aggregate
			foreach my $volDefault (@{$aggregate->{voldefaults}}) {
				$defaultNumQtrees = $volDefault->{qtrees};
				logcomment("\tDefault number of qtrees is $defaultNumQtrees\n");
			}

			## Process each volume for this aggregate
			foreach my $volume (@{$aggregate->{volume}}) {
				my $volName = $volume->{name};
		
				## Create qtree
				if (defined($volume->{qtrees}) || defined($defaultNumQtrees)) {
					my $host = host($1) if ($hostname =~ m/([0-9a-zA-z_-]+).+/);
					my $conn = connect("$1#rsh") if ($hostname =~ m/([0-9a-zA-z_-]+).+/);#XXX Add support for already shortname
					my $numQtrees = $volume->{qtrees};
					$numQtrees = $defaultNumQtrees unless defined $numQtrees;

					logcomment("\tCreating $numQtrees qtree(s) on $volName\n");
					for ( my $num = 1; $num <= $numQtrees; $num++ ) {
						QtreeCmds::qtreeCreate($conn, "/vol/$volName/qtree$num");
					}
				}
			}
		}
	}
	return $summary;
}

# Internal function
# name ::_configure_volume_relationships
#
# summary (INTERNAL) call subroutines to set relationships
#
# description
# Called by the main segement to traverse the configuration
# and call subroutines to handle setting up of relationships
# between volumes
#
# @arg configuration	Configuration data structure of the parsed XML
#
sub configure_relationships
{
	my (%args)	= validate (@_,	{	configuration	=> { type => HASHREF },
									reuse			=> { type => SCALAR, optional => 1 },
								} );
	my $config	= $args{configuration};
	my $reuse	= $args{reuse};
	my $summary	= '';

	foreach my $filer (@{$config->{filer}}) {
		my $hostname = $filer->{hostname};
		logcomment("\n\nConfiguring Relationships for $hostname\n");

		foreach my $aggregate (@{$filer->{aggregate}}) {
			my $aggregateName = $aggregate->{name};

			foreach my $volume (@{$aggregate->{volume}}) {
				my $volName = $volume->{name};
				
				foreach my $feature (@{$volume->{feature}}) {
					my $name = $feature->{name};
					my $srcHostname = $feature->{source}->{hostname};
					my $srcVol = $feature->{source}->{volume}->[0];
					logcomment("Configuring $name relationship between source $srcHostname:$srcVol and $hostname:$volName\n");
					if ($name eq 'VSM' || $name eq 'QSM') {
						&_configure_snapmirror($name, $feature, $hostname, $volName);
					} elsif ($name eq 'SV') {
						&_configure_snapvault($feature, $hostname, $volName);
					} elsif ($name eq 'RSM') {
						&_configure_raidsyncmirror($feature, $hostname, $aggregateName);
					} else {
						logcomment("$name is not a supported feature. Only VSM, QSM, SV, and RSM are avaliable.\n");
					}
				}
			}
		}
	}
	return $summary;
}

# Internal function
# name ::_configure_options
#
# summary (INTERNAL) handles licensing and setting options
#
# description
# Called by configure_volume_options to set an option to a given value.
# Deals with licensing in a limited fashion.
#
# @arg name		Name of the option to set
# @arg value	Value to set the option to
# @arg hostname	Hostname of the target filer (same as nate_hosts file)
# @arg volName	Name of the volume to set the option on
#
sub _configure_option
{
	my ($name, $value, $hostname, $volName) = @_;

	# Create connection object. [regex selects short hostname]
	my $host = host($1) if ($hostname =~ m/([0-9a-zA-z_-]+).+/);
	my $conn = connect("$1#rsh") if ($hostname =~ m/([0-9a-zA-z_-]+).+/); #XXX Add support for already shortname

	# Use the library by default
	my $useLibrary = 1;

	# Deal with options that require licenses or are not supported by Volume.pm
	if ($name eq 'compression') {
		license($conn, $name) or die $@ if ($value eq 'on');

	} elsif ($name eq 'dedupe' || $name eq 'a_sis') {
		license($conn, 'a_sis') or die $@ if ($value eq 'on');
#		Dense::dense_toggle(host => $host, path_name => "/vol/$volName" , state => $value);
		say($conn, "sis $value /vol/$volName"); 
		$useLibrary = 0;		

	#autogrow
	} elsif ($name eq 'autogrow') {
		say($conn, "vol option autogrow $volName $value");
		$useLibrary = 0;
	} elsif ($name eq 'autosize') {
		#XXX#put the say command here for autosize of find the lib to do it
		#XXX#not supported in vol_options
		logcomment("Printing autosize: vol autosize $volName $value\n");
		say($conn, "vol autosize $volName on");
		say($conn, "vol autosize $volName $value");
		$useLibrary = 0; #XXX# but not really
	}
	
	if ($useLibrary) {
		Volume::vol_options($conn, $volName , $name => $value);
	}
}

# Internal function
# name ::_configure_snapmirror
#
# summary (INTERNAL) handles licensing and establishing snapmirror
#
# description
# Called by configure_relationships to create a snapmirror relationship
# 
# @arg name		Either 'VSM' or 'QSM' to differentiate how to treat qtrees
# @arg feature	Hash containing source hostname and volume name
# @arg hostname	Hostname of the destination filer (same as nate_hosts file)
# @arg volName	Name of the destination volume
#
sub _configure_snapmirror
{
	my ($name, $feature, $hostname, $volName) = @_;

	my $source = $feature->{source};

	my $dstFiler = $1 if ($hostname =~ m/([0-9a-zA-z_-]+).+/);
	my $srcFiler = $1 if ($source->{hostname} =~ m/([0-9a-zA-z_-]+).+/);

	my $srcVol = $source->{srcvol}; #  "->[0]" needed because all xml elements called volume are made arrays for foreach looping
	my $dstVol = $volName;

	my $srcHost = host($srcFiler);
	my $dstHost = host($dstFiler);

	my $srcConn = connect("$srcFiler#rsh");
	my $dstConn = connect("$dstFiler#rsh");

	license( $srcConn, "snapmirror" );
	license( $dstConn, "snapmirror" );

	SnapMirror::snapmirror_enable( host => $dstHost );
	SnapMirror::snapmirror_enable( host => $srcHost );

	Volume::vol_online( $dstConn, $dstVol );
	Volume::vol_restrict( $dstConn, $dstVol ) unless $name eq 'QSM';

	my ($src, $dst);
	if ($name eq 'VSM') {
		$src = "$srcFiler:$srcVol";
		$dst = "$dstFiler:$dstVol";

		SnapMirror::Helper::sm_set_config( host => $dstHost, src => $src, dst => $dst );
		SnapMirror::snapmirror_initialize( host => $dstHost, src => $src, dst => $dst );
		SnapMirror::snapmirror_status( host => $dstHost, src => $src, dst => $dst );

	} elsif ($name eq 'QSM') {
		my @qtrees = QtreeCmds::qtreeList($srcConn, $srcVol);
		foreach my $qtree (@qtrees) {
			$src = "$srcFiler:".$qtree;
			$dst = "$dstFiler:/vol/$volName/".$1 if ($qtree =~ m/.\w+.\w+.(\w+)/); # $1 contains only the qtree name

			SnapMirror::Helper::sm_set_config( host => $dstHost, src => $src, dst => $dst );
			SnapMirror::snapmirror_initialize( host => $dstHost, src => $src, dst => $dst );
			SnapMirror::snapmirror_status( host => $dstHost, src => $src, dst => $dst );
		}
	}
}

# Internal function
# name ::_configure_snapvault
#
# summary (INTERNAL) handles licensing and establishing snapvault
#
# description
# Called by configure_relationships to create a snapvault relationship
# 
# @arg feature	Hash containing source hostname and volume name
# @arg hostname	Hostname of the destination filer (same as nate_hosts file)
# @arg volName	Name of the destination volume
#
sub _configure_snapvault
{
	my ($feature, $hostname, $volName) = @_;
	
	my $source = $feature->{source};

	my $dstFiler = $1 if ($hostname =~ m/([0-9a-zA-z_-]+).+/);
	my $srcFiler = $1 if ($source->{hostname} =~ m/([0-9a-zA-z_-]+).+/);
	
	my $srcVol = $source->{srcvol};
	my $dstVol = $volName;

	my $srcHost = host($srcFiler);
	my $dstHost = host($dstFiler);

	my $dstConn = connect("$dstFiler#rsh");
	my $srcConn = connect("$srcFiler#rsh");

	license( $srcConn, "sv_ontap_pri" );
	license( $dstConn, "sv_ontap_sec" );

	SnapVault::sv_toggle($srcConn, "on");
	SnapVault::sv_toggle($dstConn, "on");
	
	my	$host = make_host($hostname);
	my	$conn = make_conn($host);
	say($conn, "options snapvault.access all"); 

	my @qtrees = QtreeCmds::qtreeList($srcConn, $srcVol);
	foreach my $qtree (@qtrees) {
		SnapVault::sv_start( $dstConn, "$dstFiler:/vol/$dstVol/$1", SOURCE => "$srcFiler:$qtree" ) if ($qtree =~ m/.\w+.\w+.(\w+)/);
	}	
}

# Internal function
# name ::_configure_raidsyncmirror
#
# summary (INTERNAL) handles licensing and establishing raid syncmirror
#
# description
# Called by configure_relationships to create a raid syncmirror relationship
# 
# @arg feature	Hash containing source hostname and volume name
# @arg hostname	Hostname of the destination filer (same as nate_hosts file)
# @arg dstAggr	Name of the destination aggregate
#
sub configure_raidsyncmirror
{
	my ($feature, $hostname, $dstAggr) = @_;
	my $source = $feature->{source};

	my $dstFiler = $1 if ($hostname =~ m/([0-9a-zA-z_-]+).+/);
	my $srcFiler = $1 if ($source->{hostname} =~ m/([0-9a-zA-z_-]+).+/);
	
	my $srcVol = $source->{srcvol};

	my $srcHost = host($srcFiler);
	my $dstHost = host($dstFiler);

	my $dstConn = connect("$dstFiler#rsh");
	my $srcConn = connect("$srcFiler#rsh");

	# Get name of aggregate to overwrite as the mirror destination
	my $srcAggr = Volume::vol_container($srcConn, $srcVol);
	Aggregate::aggr_mirror($dstConn, aggr => $srcAggr, victim_aggr => $dstAggr );
}

# Internal function
# name ::cleanup_filers
#
# summary (INTERNAL) 
#
# description
# 
# 
# @arg configuration	Configuration data structure of parsed XML
# @arg reuse			Flag to preserve existing relationships if also in the configuration
#
sub cleanup_filers
{
	## XXX   Doesn't support reuse
	my (%args) = validate (@_, {	configuration	=> { type => HASHREF },
									reuse			=> { type => SCALAR, optional => 1 },
								} );
	my $config = $args{configuration};
	my $summary = '';

	foreach my $filer (@{$config->{filer}}) {
		my $hostname = $filer->{hostname};	
		my $shortHostname = $1 if ($hostname =~ m/([0-9a-zA-z_-]+).+/);
		logcomment("Cleaning up $shortHostname\n");
	
		my $filerHost = host($shortHostname);
		my $filerConn = connect("$shortHostname#rsh");

		## Remove active SnapMirror relationships
		# NOTE: Doesn't handle states: uninitilized, unknown, or broken-off
		my $statuses = SnapMirror::snapmirror_status_list(host=>$filerHost);
		foreach my $status (@{$statuses}) {
			if ($status->{state} eq 'snapmirrored') {
				my $dstPath = $status->{"destination-location"};
				SnapMirror::snapmirror_break( host => $filerHost, dstpath => $dstPath, hard => 1 );
				logcomment("\tBreaking snapmirror destination $shortHostname:$dstPath\n");
			} elsif ($status->{state} eq 'source') {
				my $srcPath = $status->{"source-location"};
				my $dstPath = $status->{"destination-location"};
				SnapMirror::snapmirror_release( host =>$filerHost, srcpath => $srcPath, dst => $dstPath);
				logcomment("\tBreaking snapmirror source $shortHostname:$srcPath to $dstPath\n");
			}
		}
  
  		## Remove active SnapVault relationships
		# NOTE: Doesn't handle states: uninitilized, unknown, or broken-off
		$statuses = SnapVault::snapvault_status_list($filerConn);
		foreach my $status (@{$statuses}) {
			if ( $status->{state} eq 'snapvaulted' ) { 
				my $dstPath = $shortHostname.':'.$status->{'destination-path'};
				SnapVault::sv_stop($filerConn, $dstPath, COMPLETE => 'true');
				logcomment("\tStopping snapvault destination $dstPath\n");
			} elsif ( $status->{state} eq 'source' ) {
				my $srcPath = $status->{'source-path'};
				my $dstPath = $status->{'destination-system'}.':'.$status->{'destination-path'};
				SnapVault::snapvault_release($filerConn, $srcPath, $dstPath);
				logcomment("\tBreaking snapvault connection between $srcPath and $dstPath\n");
			}
		}
	
		## Remove exising qtrees
		my @qtrees = QtreeCmds::qtreeList($filerConn); 
		foreach my $qtree (@qtrees) {
			say ($filerConn, "priv set test; qtree delete $qtree");
			logcomment("\tDeleting $shortHostname:$qtree\n");
		}
	}
	return $summary;
}

# Internal function
# name ::create_volumes
#
# summary (INTERNAL) Creates or reuses volumes
#
# description
# Traverses the configuration deleting existing volumes, creating new ones,
# or reusing them based on the reuse flag.
# 
# @arg configuration	Configuration data structure of parsed XML
# @arg reuse			Flag to preserve existing relationships if also in the configuration
#
sub create_volumes
{
	my (%args) = validate (@_, {	configuration	=> { type => HASHREF },
									reuse	=> { type => SCALAR, optional => 1},
								});

	my $config = $args{configuration};
	my $reuseVolumes = $args{reuse};
	my $summary = '';

	foreach my $filer (@{$config->{filer}}) {
		my $hostname = $filer->{hostname};	
		my $shortHostname = $1 if ($hostname =~ m/([0-9a-zA-z_-]+).+/);
		logcomment("\n\nCreating volumes on $shortHostname\n");

		my $filerHost = host($shortHostname);
		my $filerConn = connect("$shortHostname#rsh");

		my (@volsToCreate, %volsToDelete);
		## Handle reusing volumes	
		if ( $reuseVolumes =~ m/true/i ) {

			# Add a hash of all volumes on filer
			%volsToDelete = map {$_ => 1 } Volume::vol_list($filerConn);
			delete $volsToDelete{Volume::vol_root($filerConn)};

			foreach my $aggregate (@{$filer->{aggregate}}) {
				my $aggrName = $aggregate->{name};

				## Get default volume size
				# NOTE: foreach needed because of xml. Only one voldefaults should exist.
				my ($defaultSize, $defaultSnapShotSched);
				foreach my $voldefaults (@{$aggregate->{voldefaults}}) {
					$defaultSize = $voldefaults->{size};
					$defaultSnapShotSched = $voldefaults->{snapshotsched};
				}

				foreach my $volume (@{$aggregate->{volume}}) {
					my $desiredName = $volume->{name};
					my $desiredSize	= $volume->{size};
					$desiredSize = $defaultSize if !defined($desiredSize);

					my $snapShotSched = $volume->{snapshotsched};
					$snapShotSched = $defaultSnapShotSched if !defined($snapShotSched);
					
					logcomment("Looking for reusable volume for $shortHostname:$aggrName:$desiredName of size $desiredSize\n");
					my $desiredSizeInKB = Utils::Units::convert_to_bytes($desiredSize) / 1024;	# Volume:vol_size returns size in kB

					if ( $volsToDelete{$desiredName} ) {
						my $existingSize	= Volume::vol_size( $filerConn, $desiredName ); 
						my $status_object	= Volume::vol_status( $filerConn, $desiredName );
						my $existingAggregate = $status_object->parent();
						
						## If a volume with the same name and size is contained in the correct aggregate, remove it from deletion
						if ( $existingSize eq $desiredSizeInKB && $existingAggregate eq $aggrName ) {
							logcomment("\tReusing $shortHostname:$aggrName:$desiredName\n");
							delete $volsToDelete{$desiredName};
						} else {
							logcomment("\tAdding $shortHostname:$aggrName:$desiredName with size $desiredSize to Creation list\n");
							push ( @volsToCreate, {name => $desiredName, size => $desiredSize, aggr => $aggrName, snapshotsched => $snapShotSched});
						}
					} else { ## Else add it to creation 
						push ( @volsToCreate, {name => $desiredName, size => $desiredSize, aggr => $aggrName, snapshotsched => $snapShotSched});
						logcomment("\tAdding $shortHostname:$aggrName:$desiredName with size $desiredSize to Creation list\n");
					}
				}
			}

		## Not reusing volumes	
		} else {
			logcomment("Reusing flag not set, adding all volumes to Deletion List\n");
			## Mark all volumes for deletion
			%volsToDelete = map {$_ => 1 } Volume::vol_list($filerConn);
			delete $volsToDelete{Volume::vol_root($filerConn)};

			## All all vols to be created to array
			foreach my $aggregate (@{$filer->{aggregate}}) {
				my $aggrName = $aggregate->{name};

				## Get default volume size
				# NOTE: foreach needed because of xml. Only one volDefaults should exist.
				my ($defaultSize, $defaultSnapSched);
				foreach my $voldefaults (@{$aggregate->{voldefaults}}) {
					$defaultSize = $voldefaults->{size};
					$defaultSnapSched = $voldefaults->{snapsched};
				}

				## Add all volumes to array
				foreach my $volume (@{$aggregate->{volume}}) {
					my $volName = $volume->{name};
					my $volSize = $volume->{size};
					$volSize = $defaultSize if !defined($volSize);

					my $snapSched = $volume->{snapsched};
					$snapSched = $defaultSnapSched if !defined($snapSched);

					logcomment("Adding $shortHostname:$aggrName:$volName with size $volSize to Creation List\n");
					push( @volsToCreate, {name => $volName, size => $volSize, aggr => $aggrName, snapsched => $snapSched});
				}
			}
		}

		local $SIG{ALRM} = 'IGNORE';
		## Delete all volumes defined in %volsToDelete
		logcomment("Deleting Volumes\n");
		for my $volName ( keys %volsToDelete ) {
			logcomment("\tDeleting $volName\n");
			Volume::vol_offline($filerConn, $volName);
			Volume::vol_destroy($filerConn, $volName);
		}

		## Create all volumes defined in @volsToCreate
		logcomment("Creating Volumes\n");
		foreach my $volToCreate (@volsToCreate) {
			my $volName = $volToCreate->{name};
			my $volSize = $volToCreate->{size};
			my $snapShotSched = $volToCreate->{snapshotsched};
			my $aggrName = $volToCreate->{aggr};

			logcomment("\tCreating $shortHostname:$aggrName:$volName with size $volSize\n");
			Volume::vol_create_flex($filerConn, $volName, $aggrName, $volSize);

			#volume validation because the one in Volume.pm I couldn't get to work used their code
            Volume::vol_exists($filerConn, $volName) or croak("$@"); 
			Volume::vol_wait_change($filerConn, $volName,
								    want    => {
										state => qr(^online$),
										status => qr(normal|mirrored|raid)
									}
		    ) or croak("$@");

			if (defined($snapShotSched)) {
				logcomment("\t\tSetting Snap Schedule for $shortHostname:$aggrName:$volName\n");
				SnapCmds::snapsched_volume($filerConn, $volName, $snapShotSched);					
			}
		}
		local $SIG{ALRM} = 'DEFAULT';
	}
	return $summary;
}

## @name ::create_aggregates
##
## @summary Creates aggregates 
##
## @description
## creates all the aggregates required by the XML
##
## @return 1 on success otherwise reports errors
##
## @arg config file Required: path to the configuration xml file

sub create_aggregates {

	my (%args) = validate (@_, { configuration => { type => HASHREF },
							     destroy => { type => SCALAR }
	});

	my $config = $args{configuration};
	my $destroyAllAggregates = $args{destroy};

	#aggregates to be created will a hash with the keys being the names of the aggregates
	#	and the values being an array with [0] = number of disks or arraylist of them, [1] = addresslength, [2] = raidtype
	#aggregates to be destroyed will be a hash with the keys being the names of the aggregate
	#	and the values being the number of disks in that aggregate
	my (%aggregatesToBeCreated, %aggregatesToBeDestroyed, %ret);
    my (@aggregatesToRename, @ret, @disksToBeCreated);
	my ($shortHostName, $host, $conn);

	foreach my $filer (@{$config->{filer}}) 
	{
		#obtains a short version for connect()
		my $filerName = $filer->{hostname};
		$shortHostName = $1 if($filerName =~ m/([0-9a-zA-z_-]+).+/);

		#test all the given hosts
		&_host_testing( configuration => $config );

		#test all the given filers
		&_filer_testing( configuration => $config );

		#figure out the aggregates to be made
		%aggregatesToBeCreated = &_aggregates_to_be_created( configuration => $config,
														     filer => $filer,
															 shortHostName => $shortHostName
		);



		#figure out the aggregates to be destroyed
		@ret = &_aggregates_to_be_destroyed($config,
											$shortHostName, 
											%aggregatesToBeCreated
		);

		%aggregatesToBeCreated = %{$ret[0]};
		%aggregatesToBeDestroyed = %{$ret[1]}; 
		
		#figure out the aggreagates to be Renamed
		@aggregatesToRename = &_aggregates_to_be_renamed($config, 
														 %aggregatesToBeCreated, 
														 %aggregatesToBeDestroyed
		); 

		#turn off alarm signals to prevent script for exiting
		local $SIG{ALRM} = 'IGNORE';
	
		$host = make_host($shortHostName);
		$conn = make_conn($host);
		
		if($destroyAllAggregates) 
		{
			#removing aggregates
			logcomment("Aggregate Destruction\n");
			foreach my $aggrName (keys %aggregatesToBeDestroyed) {
				logcomment("Destroying Aggregate: $aggrName\n");
				Aggregate::aggr_purge($conn, $aggrName);
			}	
		}	
		logcomment("Aggregate Creation\n");
		foreach my $aggrName (keys %aggregatesToBeCreated) 
		{
			logcomment("Creating $aggregatesToBeCreated{$aggrName}[1] bit Aggregate: $aggrName with $aggregatesToBeCreated{$aggrName}[0] disks of types $aggregatesToBeCreated{$aggrName}[2] \n");
			#if you'd like more options later, then you must use an %opts hash as per Aggregate.pm
			Aggregate::aggr_create($conn, $aggrName, $aggregatesToBeCreated{$aggrName}[0], 
									"indirect_type", $aggregatesToBeCreated{$aggrName}[1],
									"raidtype", 	 $aggregatesToBeCreated{$aggrName}[2]);
		} #foreach

		#Polls till all the aggreagates are created
		foreach my $aggrName (keys %aggregatesToBeCreated) 
		{
			logcomment("Waiting on Aggregate: $aggrName  to be created\n");
			Aggregate::aggr_wait_change($conn, $aggrName);
		} #foreach

		logcomment("Aggregate Rename\n");
		foreach my $aggrToRename (@aggregatesToRename) 
		{
			my $existingName = $aggrToRename->{existingName};
			my $desiredName = $aggrToRename->{desiredName};
			my $numDisksToAdd = $aggrToRename->{numDisksToAdd};
			logcomment("Renaming Aggregate: $existingName to Aggregate: $desiredName adding $numDisksToAdd disks\n");
			Aggregate::aggr_rename($conn, $existingName, $desiredName);
			Aggregate::aggr_add($conn, $desiredName, $numDisksToAdd);		
		} #foreach

	} #foreach

	#alarm signal back on
	local $SIG{ALRM} = 'DEFAULT';
	#nice gesture zero all spares
	say($conn, "disk zero spares");
	logcomment("\nCompleted Aggregate Setup\n\n");
	return 1;
}		

## Internal function
# name ::_aggregates_to_be_renamed
#
# summary (INTERNAL) compares the two lists (aggregatesToBeCreated and aggregatesToBeDestroyed)
#		to find aggregates with the same name and disks to be renamed. It remvoes these from
#		both lists as it's the same aggregate. It will also find aggregates to be reused to 
#		save the time of having to zero disks.
#
# description
# Called by create_aggregates to create a list of aggregates to be created 
#
# @return array with the disks to rename(old and new name) and the amount of disks to add
#
# @arg configuration Configuration data structure of parsed XML 
# @arg aggregatesToBeCreated hash of all the aggregates to created
# @arg aggregatesToBeDestroyed hash of all the aggregates to destroy
#
sub _aggregates_to_be_renamed {

	#XXX# unable to validate the aggregatesToBeCreated and destroyed as there types are unknown
	#	my %args = validate (@_, { configuration => { type => HASHREF },
	#						aggregatesToBeCreated => { type => HASHREF },
	#						aggregatesToBeDestroyed => { type => HASHREF },
	#	});
	#   my $config = $args{configuration};
	#	my %aggregatesToBeCreated = $args{aggregatesToBeCreated};
	#	my %aggregatesToBeDestroyed = $args{aggregatesToBeDestroyed};

	my ($config, %aggregatesToBeDestroyed, %aggregatesToBeCreated) = @_;

	my @aggregatesToRename;
	foreach my $desiredName (keys %aggregatesToBeCreated) {
		my $tempExistingName;
		my $tempNumDisksToAdd = 100; #100 because you can't have more than 100 disks in an aggregate
		logcomment("Looking for reusable aggregate for $desiredName with <= $aggregatesToBeCreated{$desiredName}[0]");

		foreach my $existingName (keys %aggregatesToBeDestroyed) {
			if ($aggregatesToBeCreated{$desiredName}[0] >= $aggregatesToBeDestroyed{$existingName}) {
				my $numDisksToAdd = $aggregatesToBeCreated{$desiredName}[0] - $aggregatesToBeDestroyed{$existingName};
				logcomment("$existingName has $aggregatesToBeDestroyed{$existingName} disks, $numDisksToAdd would have to be added");
				#check below is to find the best possible case, to prevent inefficient allocation
				if ($numDisksToAdd < $tempNumDisksToAdd) { 
					$tempNumDisksToAdd = $numDisksToAdd;
					$tempExistingName = $existingName;
				}
			}

			if ($tempNumDisksToAdd != 100) {
				#rename aggr in destroyed hash
				push(@aggregatesToRename, {existingName => $tempExistingName, desiredName => $desiredName, numDisksToAdd => $tempNumDisksToAdd});
				#remove from destroyed hash
				delete $aggregatesToBeDestroyed{$tempExistingName};
				#remove aggr from creation hash
				delete $aggregatesToBeCreated{$desiredName};
			}
		}
	}
	return @aggregatesToRename;
}

# Internal function
# name ::_aggregates_to_be_destroyed
#
# summary (INTERNAL) obtains a given list of aggregates
#		to be created and their disk size from the filer
#
# description
# Called by create_aggregates to create a list of aggregates to be destroyed 
#
# @return hash with all the aggregates on the filer
#
# @arg configuration Configuration data structure of parsed XML 
# @arg shorthostname the shortened verion of the host name for nate_hosts
# @arg %aggregatesToBeCreated hash pointing to an array with [0] being the disks
#			and [1] being the type
sub _aggregates_to_be_destroyed {

	#XXX# unable to get aggregatesToBeCreated hash of arrays to validate as it's
	#XXX# a differant being 
	#my %args = validate (@_, { configuration => {type => HASHREF},
	#					       shorthostname => {type => SCALAR},
	#						   aggregatesToBeCreated => {type => UNDEF}
	#});
	#my $config = $args{configuration};
	#my $shortHostName = $args{shorthostname};
	#my %aggregatesToBeCreated = $args{aggregatesToBeCreated}; 
	my ($config, $shortHostName, %aggregatesToBeCreated) = @_;
	my %tempAggrListOfDisks = ();
	my @ret;


	my %aggregatesToBeDestroyed = ();
	my $count = 0;
	my $conn = connect("$shortHostName#rsh");

	#maps every aggregate's name on the filer to the hash
	%aggregatesToBeDestroyed = map {$_ => []} Aggregate::aggr_list($conn);

	#finds the number of disks each aggregate takes
	foreach my $aggrName (keys %aggregatesToBeDestroyed) {
		my @disks = Aggregate::aggr_disks($conn, $aggrName);
		my $numOfDisks = @disks;
		$aggregatesToBeDestroyed{$aggrName}[0] = $numOfDisks;
	}

	foreach my $aggrName (keys %aggregatesToBeDestroyed) {
		my @disks = Aggregate::aggr_disks($conn, $aggrName);
		my @diskList;
		my $count = 0;
		foreach my $disk (@disks) {
			my $diskName = $disk->{name};
			$diskList[$count++] = $diskName;
		}
		$tempAggrListOfDisks{$aggrName} = \@diskList;
	}

	#remove root aggregate
	my $root = Aggregate::aggr_root($conn);
	delete $aggregatesToBeDestroyed{$root};
	delete $tempAggrListOfDisks{$root};

	#reusing aggrs with two cases for integer disks number and a list of disk 

	#if aggr has same name and size remove from both lists
	#the test for name is implicit when searching for the hash
	foreach my $aggrName (keys %aggregatesToBeCreated) {
		if (defined($aggregatesToBeCreated{$aggrName}[0]) && defined($aggregatesToBeDestroyed{$aggrName})) { #make sure they both exist to be checked
			if ($aggregatesToBeCreated{$aggrName}[0] eq $aggregatesToBeDestroyed{$aggrName}) {
				logcomment("Reusing Aggregate: $aggrName\n");
				delete $aggregatesToBeCreated{$aggrName};
				delete $aggregatesToBeDestroyed{$aggrName};
			}
		}
	}

	#if aggr has the same name and same disks remove from both lsits
	foreach my $aggrName (keys %aggregatesToBeCreated) {
		my @aggrToBeCreated = @{$aggregatesToBeCreated{$aggrName}[0]};
		my @listsOnDisk;
		if (defined($tempAggrListOfDisks{$aggrName}) ) {
			@listsOnDisk = @{$tempAggrListOfDisks{$aggrName}};
		}
		my %bothLists;
		my @itemsOnBothLists;
		my @itemsNotOnBothLists;
		foreach my $item (@aggrToBeCreated, @listsOnDisk) {
			$bothLists{$item}++;
		}
		foreach my $item (keys %bothLists) {
			if ($bothLists{$item} == 2) {
				push (@itemsOnBothLists, $item);
			} else {
				push (@itemsNotOnBothLists, $item);
			}
		}
		my $size = @itemsNotOnBothLists;
		if ( $size == 0 ) { 
			delete($aggregatesToBeCreated{$aggrName});
			delete($aggregatesToBeDestroyed{$aggrName});
		}
	}

	$ret[0] = \%aggregatesToBeCreated;
	$ret[1] = \%aggregatesToBeDestroyed;
	return @ret; 
}
	
# Internal function
# name :_aggregates_to_be_created
#
# summary (INTERNAL) obtains from the XML a given list of
#		aggregates to be created and their disk size
#
# description
# Called by create_aggregates to create a list of aggregates to be created 
#
# @return hash with all the on the XML
#
# @arg configuration Configuration data structure of parsed XML 
sub _aggregates_to_be_created {
	my %args = validate (@_, {filer => {type => HASHREF}, configuration => {type => HASHREF}, shortHostName => {type => SCALAR}});

  	my $filer		= $args{filer};
	my $configFile	= $args{configuration};
	my $shortHostName = $args{shortHostName};

	my %aggregatesToBeCreated = ();
	my $filerName	= $filer->{hostname};
	foreach my $aggregate (@{$filer->{aggregate}}) {
		my $aggregateName	= $aggregate->{name};
		my $aggregateType	= $aggregate->{type};
		my $aggregateRaidType = $aggregate->{raidtype};
		my $disksInXML		= $aggregate->{disks};
		my @DiskList		= $disksInXML;

		my ($volDefaultSize, $volDefaultSched);
		foreach my $voldefaults (@{$aggregate->{voldefaults}}) {
			$volDefaultSize = $voldefaults->{size};
			$volDefaultSched = $voldefaults->{sched};
		}

		#finding the aggregate type
		if($aggregateType !~ /(32-bit|64-bit)/i) {
			logcomment("Aggregate type can be set to '32-bit' or '64-bit' in $configFile. Using 32-bit\n"); 
			$aggregatesToBeCreated{$aggregateName}[1] = '32';
		} elsif ($aggregateType eq '32-bit') {
			$aggregatesToBeCreated{$aggregateName}[1] = '32';
		} else {
			$aggregatesToBeCreated{$aggregateName}[1] = '64';
		}

		#finding the raid type
		if($aggregateRaidType !~/(raid4|raid_dp)/i) {
			logcomment("Aggregate Raid type can be set to 'raid4' or 'raid_dp' in $configFile. Using raid_dp\n");
			$aggregatesToBeCreated{$aggregateName}[2] = 'raid_dp';
		} elsif ($aggregateRaidType eq 'raid4') {
			$aggregatesToBeCreated{$aggregateName}[2] = 'raid4';
		} else {
			$aggregatesToBeCreated{$aggregateName}[2] = 'raid_dp';
		}

		my ($totalSpaceRequested, $sizeOfDisksOnFiler, $numberOfDisks);
		if(defined($disksInXML)){ 
			#the user knows how many disks they need
			#creating lists of disks for aggregate

			my @disksToBeCreated;
			my $count = 0;
			my $scalarFlag = 0;
			my $arrayFlag = 0;
			foreach my $disks (@{$DiskList[0]}) {
				if ($disks =~ / +/) { #case where they are seperated by spaces
					my @multipleDisks = split(/ /, $disks);
					$arrayFlag = 1;
					foreach my $part (@multipleDisks) {
						$disksToBeCreated[$count++] = $part;	
					}
				} elsif ($disks =~ /-+/) { #case where they are to be expanded
					$arrayFlag = 1;
					if($disks =~ /(\d{1}\w{1}).(\d{2,3})-\d{1}\w{1}.(\d{2,3})/) {
						for(my $i = $2; $i<= $3; $i++) {
							$disksToBeCreated[$count++] = "$1.$i";
						}
					}
				} elsif ($disks =~ /(\d{1}\w{1}.\d{1,3})/) { #single case
					$arrayFlag = 1;
					$disksToBeCreated[$count++] = $1; 
				} elsif ($disks =~ /(\d{1})/) { #case where they just give use a number
					croak("doesnt support scalars yet");
					$scalarFlag = 1;
					$aggregatesToBeCreated{$aggregateName}[0] = $1;
					#can't make an aggregate of raiddp less than 3 disks, will give error in Aggregate.pm 
					if($aggregatesToBeCreated{$aggregateName}[0] < $MIN_DISKS_FOR_RAIDDP) {		
						$aggregatesToBeCreated{$aggregateName}[0] = $MIN_DISKS_FOR_RAIDDP;
					}
				}
				if(!$scalarFlag) { #if it's not a scalar then set the array list as our disk list
					$aggregatesToBeCreated{$aggregateName}[0] = \@disksToBeCreated;
				}
			}
#			logcomment("$scalarFlag $arrayFlag");
#			if($scalarFlag && $arrayFlag) {
#				croak("Can't ask for a # of disks and specific disks\n");
#			}

		} else {
			#I'll figure out how many disks they need
			$sizeOfDisksOnFiler = &_disksize_on_filer(shortHostName => $shortHostName);
			#finds the total space requested to make the Aggregate
			$totalSpaceRequested = &_space_requested_per_aggregate(aggregate => $aggregate ,volDefaultSize => $volDefaultSize);
			$numberOfDisks = &_roundup(number => ($totalSpaceRequested / $sizeOfDisksOnFiler));
			$aggregatesToBeCreated{$aggregateName}[0] = $numberOfDisks;
			#can't make an aggregate of raiddp less than 3 disks, will give error in Aggregate.pm 
			if($aggregatesToBeCreated{$aggregateName}[0] < $MIN_DISKS_FOR_RAIDDP) {		
				$aggregatesToBeCreated{$aggregateName}[0] = $MIN_DISKS_FOR_RAIDDP;
			}
		}
	}
	return %aggregatesToBeCreated;
}
# Internal function
# name ::_filer_testing
#
# summary (INTERNAL) tests that the given filers are all online
#
# description
# Called by create_aggregates to test the filers
#
# @return nothing
#
# @arg configuration Configuration data structure of parsed XML 
sub _filer_testing {
    my %args = validate (@_, {configuration => {type => HASHREF}});
    my $config = $args{configuration};

	logcomment("\nFiler Testing\n");
	foreach my $filer (@{$config->{filer}}) {
		my $filerName = $filer->{hostname};
		my $shortHostName = $1 if($filerName =~ m/([0-9a-zA-z_-]+).+/);
		my $conn = connect("$shortHostName#rsh");

		#Availability check on filers
		my $timeoutSignal = 0;
		my $answer;
		eval {
			local $SIG{ALRM} = sub { $timeoutSignal++; die "alarm\n";};
			alarm($TIMEOUT_FOR_EACH_FILER);
			logcomment("Connecting to $filerName filer\n");
			$answer = say($conn, "?");
		};
		if($timeoutSignal) {
			logcomment("$filerName : FAILED to connect\n");
			croak("Recheck filer: $filerName unavailable to connect\n");
		} else {
			logcomment("$filerName is online\n");
		}

		#SSH check on filers
		my $sshSignal = 0;
		my $response;
		eval {
			local $SIG{ALRM} = sub { $sshSignal++; die "alarm\n";};
			alarm($TIMEOUT_FOR_SSH_CHECK_ON_FILER);
			logcomment("Testing SSH on $filerName filer\n");
			$answer = say($conn, "secureadmin status");
		};
		if($sshSignal || $response =~ m/ssh2	- inactive/) {
			logcomment("$filerName : Failed SSH test\n");
			croak("Recheck filer: $filerName SSH is disabled\n");
		} else {
			logcomment("$filerName has SSH online\n");
		}
	}
}

# Internal function
# name ::_host_testing
#
# summary (INTERNAL) tests that the given hosts are all online
#
# description
# Called by create_aggregates to test the hosts
#
# @return nothing
#
# @arg configuration Configuration data structure of parsed XML 
sub _host_testing {
    my %args = validate (@_, {configuration => {type => HASHREF}});
    my $config = $args{configuration};

    logcomment("\nHost Testing\n");
	foreach my $host (@{$config->{host}}) {
		my $hostname = $host->{hostname};
		my $output = `ping $hostname -c $TIMES_TO_PING_EACH_HOST -W $TIMEOUT_FOR_EACH_HOST`; 
		if($output =~ /, (\d+)% packet loss,/) {
			if($1 > 0) {
				croak("$hostname didn't have 100% packet transmittion\n"
						. "Remove the unavailable host from the list\n");
			} else {
				logcomment("$hostname pinged successfully\n");
			}
		}
	}
}

## Internal function
# name ::_disksize_on_filer
#
# summary (INTERNAL) finds the size of a disk on the filer
#
# description
# 	Called by _aggregates_to_be_created
#
# @return sizeOfDisk size of a disk on the filer 
#
# @arg filername the name of the filer to find the disk size of
sub _disksize_on_filer {
	my %args = validate (@_, {shortHostName => {type=> SCALAR}});

	my $shortHostName = $args{shortHostName};

	#finding total disks to use
	my $host = host($shortHostName);
	my $conn = connect("#rsh");

	my $diskSize = filer::filer_disk_capacity_get($conn);
	return int(Utils::Units::convert_units( $diskSize, UNIT_B, UNIT_MB ));
}

## Internal function
# name ::_space_requested_per_aggregate
#
# summary (INTERNAL) finds how much space is requested per aggregate 
#
# description
# 	Called by _aggregates_to_be_created
#
# @return $totalSpaceRequested the total space asked for by the configuration file
#
# @arg aggregate the aggregate object to search through
# @arg voldefaultsize used incase the volSize isn't defined
sub _space_requested_per_aggregate {
	my %args = validate (@_, {aggregate => {type => HASHREF}, volDefaultSize => {type => SCALAR}});

	my $aggregate = $args{configuartion};
	my $volDefaultSize = $args{volDefaultSize};

	my $totalSpaceRequested = 0;
	foreach my $volume (@{$aggregate->{volume}}) {
		my $volSize = $volume->{size};
		$volSize = $volDefaultSize if !defined($volSize);

		$totalSpaceRequested += convert_to_bytes($volSize);
	}

	# Convert to MB because disk sizes are in MB
	$totalSpaceRequested = convert_units($totalSpaceRequested, UNIT_B, UNIT_MB );
	return $totalSpaceRequested;
}

## Internal function
# name ::_roundup
#
# summary (INTERNAL) rounds up the next whole number 
#
# description
# 	Called by _aggregates_to_be_created
#
# @return the given number rounded up
#
# @arg number you want rounded up
sub _roundup {
	my %args = validate (@_, {number => {type => SCALAR}});
    my $n = $args{number};
	return (($n == int($n)) ? $n : int($n+1));
}

1;
