package ConfigBroker;

use strict;
use warnings;
use DBI qw();
use Tharn qw($Log);
use Tharn;
our @ISA = qw (Exporter);
use Params::Validate qw(validate validate_with SCALAR OBJECT HASHREF BOOLEAN ARRAYREF UNDEF);
use Module::Common qw(validate_mixed);
use ConfigBroker::ResourceGroup;
use ConfigBroker::Tasks;
use ConfigBroker::CbUtils;
use ConfigBroker::MySqld;
use Data::Dumper;
use Options;
use ConfigBroker::Schema;
use cluster::enterprise::TestTriage::DiscoverConfiguration qw(getObjectList discoverObject);
use File::Basename;
use File::Path;
#use NATE::Events qw(listener_add call_on_fork_add);
use NACL::C::Node;
use ConfigBroker::ResourceGroup::NASGroup;

our $AUTOLOAD;
our $DEFAULT_DB_SCHEMA_VERSION = 6.0;
my $cb_version_table = "CB_Version_Table";


## @summary ConfigBroker
## @author veena@netapp.com
## @status public
## @pod here

=head1 NAME

ConfigBroker - ConfigBroker main module 

=head1 SYNOPSIS

  use ConfigBroker; 
  
  my $CB = new ConfigBroker(); 
  my $CB = new ConfigBroker($testbed_path); 

  "new" creates a ConfigBroker object.
 
=head1 DESCRIPTION

Top level ConfigBroker class and object methods.

=cut


#call_on_fork_add({callback     => \&db_forked,
#                  keep_on_fork => 1});

sub new
{
   $Log->enter();
   my ($pkg, $testbeddir, $basic) = @_;
   my $class = ref($pkg) || $pkg ;
   my $self = {};

   ### the db will exist in the topdir of the testbed.  Get the testbed
   ### topdir from the LOGDIR nate value.  The name of the DB will be
   ### "ConfigBroker.db"

   my $is_windows = 0;
   if ($^O eq 'MSWin32') {
      $is_windows = 1;
   }
   my $db;  
 
   my $cb_dir_name  = ($is_windows) ? '\\ConfigBroker' : '/ConfigBroker';
   
   if ((! defined $testbeddir) || ($testbeddir eq "")) {
      # If we are here, this was not passed in...get it from tharn
      $testbeddir  = ($is_windows) ? Tharn::param('TOP_LOGDIR_WINDOWS') : Tharn::param('TOP_LOGDIR_UNIX');
      chomp $testbeddir;
      my $dir_adjuster = ($is_windows) ? '\\..' : '/..';

      # Check up three levels for the ConfigBroker directory
      # if it isn't there, then die;
      $db = $testbeddir;
      my $found_cb = 0;
      foreach my $backdir (1 .. 3) { 
          # end the loop if the ConfigBroker directory is found
          if (-e $db . $cb_dir_name) { 
              $found_cb = 1;
              last;
          }
          # otherwise, check up one level next time
          $db .= $dir_adjuster;
      }

      # We need to know if this is the CTL PB environment. CTL PB does this a little
      # different when it comes to where logs are kept. When it comes to the CTL, it 
      # doesn't matter where we put the database, it's not going to be intended to 
      # out live the ntest invocation because the TB is created and destroyed dynamically.

      # Their scheme for the log location is:
      # <testsuite_name>_<execution_attempt_id (in database)>/<index for tc (starts at 1)>
      #  _<testcase_name>_<testcase_execution_id (in database)>

      # The CTL PB porting areas use a different value for TESTBED


      # We may not find a directory at all. This probably means we would need to create 
      # a database. The CTL PB does things differently.
      if (not $found_cb) { 
          # if we are in the pushbutton or porting area of CTL, we need to put stuff in 
          # different places
          my $tb = Tharn::param('TESTBED');
          my $ctlpb = (($tb =~ /^pushbutton$/i) or ($tb =~ /^porting/i));
          
          # If this is the PB area, keep the ConfigBroker directory one level up. 
          # a regular testbed, we should keep this two levels up.
          if ($ctlpb) { 
              $db = $testbeddir . $dir_adjuster . $cb_dir_name;

          } else { 
              $db = $testbeddir . $dir_adjuster . $dir_adjuster . $cb_dir_name;
          }


      } else { 
          $db = $db . $cb_dir_name;
      }

   } else { 
      $db = $testbeddir . $cb_dir_name;
   }

 


   $Log->debug("looking for CB database in $db");
   $self->{'dbname'} = "CB";
   $self->{'db_base_dir'} = $db;
   my $sqld = new ConfigBroker::MySqld(base_dir => $db);
   $self->{'mysqld'}     = $sqld;
   my $server_host       = $sqld->{'host'};
   my $server_port       = $sqld->{config}->{'port'};
   $self->{'dsn'}        = $sqld->dsn(dbname => $self->{'dbname'}, host => $server_host, port => $server_port);
   $Log->debug("DSN IS $self->{'dsn'}\n");

   if ($is_windows) {
      $self->{'path'}       = "$db\\var\\$self->{'dbname'}";
   }
   else {
      $self->{'path'}       = "$db/var/$self->{'dbname'}";
   }
   $self->{'user'}       = undef;
   $self->{'auth'}       = undef;
   $self->{'dbh'}        = '';
   $self->{'resourceschema'}     = undef;
   bless ($self, $class);
   return $self if ($basic); 


   ### get the default schema or from a schemafile if exists.
   #my $schemafile = "$testbeddir/cb.schema";
   #if (-e $schemafile) {
    #  $self->{'resourceschema'} = new ConfigBroker::Schema($schemafile);
   #}
   #else {
    #  $self->{'resourceschema'} = new ConfigBroker::Schema();
   #}
   $self->{CB_Version_Table} = $cb_version_table;   

   ### first check if the DB exists.  If it does, no need to initialize tables etc.  Just need to connect.
   my $db_exists = 0;
   $db_exists = 1 if (-e $self->{'path'});
   ### Connect to db.  Creates new if it does not exist.
   my $umask = umask 0;
   my %connect_opts = (AutoCommit => 1, PrintError => 0, RaiseError => 1);
   $self->_connect(\%connect_opts);
   opendir(DIR, $self->{'path'});
   my @dbfiles = readdir(DIR); 
   $Log->debug(Dumper(\@dbfiles));
   if (scalar(@dbfiles) <= 2) {
      $self->_init_db();

      ### BELOW IS A TEMPORARY HACK FOR SCHEMA VERSION CHANGE WARNING TO WORK CORRECTLY.
      ### WILL BE REMOVED SOON.
      ### WE ADD VERSION ONLY IF THE DB IS CREATED FRESH.
      my $add_record = qq[INSERT INTO $self->{CB_Version_Table} (package, version) ];
      $add_record .= qq[values ('$class', '$DEFAULT_DB_SCHEMA_VERSION')];
      $Log->comment("ADD record IS $add_record\n");
      $self->{dbh}->do($add_record);
   }

   $self->_verify_and_store_schema_or_die($class, $DEFAULT_DB_SCHEMA_VERSION);
  
   $self->{'dbh'}->{InactiveDestroy} = 1;
   closedir DIR;
   $Log->exit();
   return $self;
}

sub get_dbh {
   my $self = shift;
   return $self->{'dbh'};
}

sub AUTOLOAD {
     my $self = shift;
     my %opts = @_;

     my $name = $AUTOLOAD;
     $name =~ s/.*://;   # strip fully-qualified portion

     return if $name eq "DESTROY";

     my $resource_package ="ConfigBroker::ResourceGroup::$name";

     $Log->debug("Auto loading $resource_package");

     eval "use $resource_package;";

     die($@) if $@;

     my $version = eval ("\$$resource_package"."::SCHEMA_VERSION");

     $self->_verify_and_store_schema_or_die($name, $version);

     $opts{'dbh'}  = $self->{'dbh'};
     $opts{'name'} = "$name";

     my $rg_obj = $resource_package->new(%opts);
     
     return $rg_obj;     
}

sub stop {
   $Log->enter();
    my ($self) = shift;

    my $sqld = $self->{'mysqld'};
    my $host       = $sqld->{'host'};
    my $pid        = $sqld->{'pid'};
    $sqld->stop($host, $pid);
    $Log->exit();
}

sub obliterate {
   $Log->enter();
    my ($self) = shift;

    $self->stop();
    rmtree($self->{'db_base_dir'});
    $Log->exit();
}
       

sub db_dump {
    my ($self, %opts) = validate_mixed( @_,
                  [{type => OBJECT}],
                   {"format"  => { type => SCALAR, optional => 1, default => 'CSV'},
                    "output_dir"  => {type => SCALAR},
                    "output_file" => {type => SCALAR, optional => 1},}
                  );

    my $sqld = $self->{'mysqld'};
    my $server_host       = $sqld->{'host'};
    my $server_port       = $sqld->{config}->{'port'};
    my $outfile;
    my $dump_cmd_args;
    my $outdir = $opts{output_dir};
    if ($opts{'format'} eq "CSV") {
        my $csvdir = "$outdir/CSV";
        if (! -d $csvdir) {
            if(! mkdir($csvdir)) {
                die "mkdir failed to create dir $csvdir\n";
            }
        }
        $dump_cmd_args = "-t --tab --fields-terminated-by=, -T $csvdir CB";
    } 
    elsif ($opts{'format'} eq "SQL") {
       $outfile = "$outdir/cb_backup.sql";
       $outfile = "$outdir/$opts{output_file}" if (defined $opts{output_file});
       $dump_cmd_args    = "CB \> $outfile";
    }
    elsif ($opts{'format'} eq "XML") {
       $outfile = "$outdir/cb_backup.xml";
       $outfile = "$outdir/$opts{output_file}" if (defined $opts{output_file});
       $dump_cmd_args    = "--xml CB \> $outfile";
    }
 
    my $dump_cmd = "mysqldump --host $server_host --port $server_port $dump_cmd_args" ;
    my $ret = system($dump_cmd);
    if ($ret != 0) {
        die "sqldump failed\n";
    }
}

sub db_forked {
    my $self = shift;
    $self->{'dbh'}->{InactiveDestroy} = 1;
    $self->{'dbh'} = $self->{'dbh'}->clone();
    return $self;
}

sub _connect {
   $Log->enter();
   my ($self, $options) = shift;
   #DBI->trace(3);
   $self->{'dbh'} = DBI->connect($self->{'dsn'}, '', '', $options)
        || $Log->debug($DBI::errstr);
   if (($DBI::errstr) && ($DBI::errstr =~ /Can't connect to MySQL server/)) {
      my $ret = $self->{'mysqld'}->verify_and_cleanup();    
      if ($ret != 0) {
         die "Existing MySQL server instance on $self->{'mysqld'}->{'host-name'} could not be stopped and re-started.\n";
      }
      else {
         ### Restart mysqld now.
         $self->{'mysqld'} = new ConfigBroker::MySqld(base_dir => $self->{'db_base_dir'});
         my $server_host       = $self->{'mysqld'}->{'host'};
         $self->{'dsn'}        = $self->{'mysqld'}->dsn(dbname => $self->{'dbname'}, host => $server_host);

         ### Now try to connect again and fail if you cannot connect.
         $self->{'dbh'} = DBI->connect($self->{'dsn'}, '', '', $options)
                          || $Log->debug($DBI::errstr);
      }
   }
   $self->{'dbh'}->{'mysql_auto_reconnect'} = 1; 
   $Log->exit();
   return 0;
}

sub _init_db(){
   $Log->enter();
   my ($self) = shift;

   ### Add the tables that we need.
   $self->_create_resourcegroup_table();

   ### create table type definiton table.
   $self->_create_typedefn_table();
   $self->_populate_typedefn_table();

   ### create super group table.
   $self->_create_supergroup_table();

   ### create resource tables for each type of resource.
   ### currently supported resources are: aggr, vserver, lif, igroup, volume, qtree, lun  
   $self->_create_resource_tables();
  
   $self->_create_task_table();
   # Create configbroker version table 
   $self->_create_version_table();

   $Log->exit();
   return 0;
}

sub _create_task_table {
	$Log->enter;
	my $self = shift;
	my $schema = &ConfigBroker::Schema::get_tasks_table_schema(); 
	my $attributes  = $schema->{'tasks'};
	my $attributestr .= '('.$attributes.')';
	&ConfigBroker::CbUtils::create_table($self->{'dbh'},'tasks',$attributestr);
}

sub _create_resourcegroup_table {
   $Log->enter();
   my ($self) = shift;
   $Log->enter();
   #my $table_defn = "name TEXT PRIMARY KEY, reuse TEXT, share TEXT, dsn TEXT, configfile TEXT, configengine TEXT, cleanupfile TEXT, cmdlinestring TEXT, status TEXT";
   my $table_defn = "(name VARCHAR(100) PRIMARY KEY, reuse VARCHAR(15), share VARCHAR(15), dsn VARCHAR(255), configfile VARCHAR(512), configengine VARCHAR(20), cleanupfile VARCHAR(255), cmdlinestring VARCHAR(512), status VARCHAR(20))";
   &ConfigBroker::CbUtils::create_table ($self->{'dbh'}, "resourcegroup", $table_defn);
   $Log->exit();
   return 0;
}

sub _create_typedefn_table {
   $Log->enter();
   my ($self) = shift;
   $Log->enter();
   #my $table_defn = "typename TEXT PRIMARY KEY, typefields TEXT";
   my $table_defn = "(typename VARCHAR(64) PRIMARY KEY, typefields TEXT, name TEXT)";
   &ConfigBroker::CbUtils::create_table ($self->{'dbh'}, "typedefn", $table_defn);
   $Log->exit();
   return 0;
}

sub _create_supergroup_table {
   $Log->enter();
   my ($self) = shift;
   $Log->enter();
   #my $table_defn = "name TEXT PRIMARY KEY, groups TEXT";
   my $table_defn = "(name VARCHAR(100) PRIMARY KEY, groups VARCHAR(512))";
   &ConfigBroker::CbUtils::create_table ($self->{'dbh'}, "supergroup", $table_defn);
   $Log->exit();
   return 0;
}

sub delete_resource_type {
    $Log->enter();
    my ($self, %opts) =
          validate_mixed(@_,
                  [{type => OBJECT},],
                  {name    => {type => SCALAR,},},);

   my $ret = &ConfigBroker::CbUtils::drop_table($self->{'dbh'}, $opts{name});
   $self->_delete_typedefn($opts{name});
   
   return $ret;
}

sub _create_resource_table { 
    $Log->enter();
    my ($self, %opts) =
          validate_mixed(@_,
                  [{type => OBJECT},],
                  {name    => {type => SCALAR,},
                   defcomplete => {type => SCALAR, optional => 1,},
                   primarykey  => {type => ARRAYREF,},
                   tabledefn   => {type => ARRAYREF,},},);

    my @typefieldarray;
    my $string = &ConfigBroker::Schema::get_default_table_attrs();
    @typefieldarray = split(' ', $string) if (not $opts{defcomplete});
    foreach my $def (@{$opts{tabledefn}}) { 
       push (@typefieldarray, $def);
    }
    
    # add the field type information 
    my $tabledefn;
    foreach my $field (@typefieldarray){
       if ($field eq "uniq_resource_id") {
          $tabledefn .= $field . " VARCHAR(500), " ;
       }
       else {
          $tabledefn .= $field . " TEXT, " ;
       }
    }

    my $primarykeys = "PRIMARY KEY (uniq_resource_id)";
  
    # The typedefn table maintains a list of the column headers. Just stick the raw 
    # column header data in there
    $self->_add_typedefn($opts{name}, join(' ', @typefieldarray))  if (not $opts{defcomplete});

    # The final sqlinsert should resemble this:
    # (name VARCHAR(100), resourcegroup VARCHAR(100), resource_category TEXT, resource_status TEXT, resource_priority TEXT,
    # resource_comment TEXT, hosttype TEXT, os_ver TEXT,  PRIMARY KEY (name,resourcegroup));
    my $sqlinsert = "(" . $tabledefn . ' ' . $primarykeys . ")";
	
    $Log->debug("Creating table $opts{name} ; The typefields are $sqlinsert\n");
    &ConfigBroker::CbUtils::create_table ($self->{'dbh'}, $opts{name}, $sqlinsert);
	
   $Log->exit();
   return 0;
}

sub _create_resource_tables {
   my ($self) = shift;
   $Log->enter();

   my $schemamap = $self->{'resourceschema'}->get_resource_schema();
   foreach my $typename (keys(%$schemamap))
   {
      my $typefield = $schemamap->{$typename};
      my @typefieldarray = split(" ", $typefield);
      my $pkeys = &ConfigBroker::Schema::get_primary_keys($typename);
      $self->_create_resource_table(name => $typename,
                                   primarykey => $pkeys,
                                   defcomplete => 1,
                                   tabledefn => \@typefieldarray);

      ### now create triggers that will operate on inserts and updates on this table.
      $self->_create_resource_table_triggers(resourcetype => $typename, primarykey => $pkeys);

      ### Now create a resource-group cross reference table with foreign key constraint
      ### The resourcegroup table already should exist.
      $self->_create_crossref_table(resourcetype => $typename);
   }
   $Log->exit();
   return 0;
}

sub _create_resource_table_triggers {
    my ($self, %opts) =
          validate_mixed(@_,
                  [{type => OBJECT},],
                  {resourcetype => {type => SCALAR,},
                   primarykey   => {type => ARRAYREF,},},);

    my @primary = @{$opts{primarykey}};
    ### create triggers to get the uniqid column for this table using the values of primary keys
    ### The triggers will operate on inserts and updates.
    my $sqlconcat = "SET NEW.uniq_resource_id=CONCAT_WS(\'_\'\,";
    foreach (@primary) {
       $sqlconcat .= "NEW.$_," ;
    }
    $sqlconcat =~ s/,$/)/ ;
    my $triggername = $opts{resourcetype} . "_insert";
    my $triggercmd = "CREATE TRIGGER $triggername BEFORE INSERT ON $opts{resourcetype} FOR EACH ROW $sqlconcat" ;
    $Log->debug("TRIGGERCMD IS $triggercmd\n");
    $self->{'dbh'}->do($triggercmd);
    $triggername = $opts{resourcetype} . "_update";
    $triggercmd = "CREATE TRIGGER $triggername BEFORE UPDATE ON $opts{resourcetype} FOR EACH ROW $sqlconcat" ;
    $Log->debug("TRIGGERCMD IS $triggercmd\n");
    $self->{'dbh'}->do($triggercmd);
} 

sub _create_crossref_table {
    $Log->enter();
    my ($self, %opts) =
          validate_mixed(@_,
                  [{type => OBJECT},],
                  {resourcetype => {type => SCALAR,},},);

    ### There will be one crossreference table per resourcetype.  It will be named as
    ### "$resourcetype" . "crossref", for example, for type "volume" the crossref table will be called "volumecrossref"
    my $table = $opts{resourcetype};
    my $reftable = "$table" . "crossref" ;

    my $tabledefn = "(resourceid VARCHAR(500), resourcegroup VARCHAR(100), PRIMARY KEY (resourceid, resourcegroup), FOREIGN KEY (resourceid) REFERENCES $opts{resourcetype} (uniq_resource_id) on delete cascade, FOREIGN KEY (resourcegroup) REFERENCES resourcegroup (name) on delete cascade)";
    &ConfigBroker::CbUtils::create_table ($self->{'dbh'}, $reftable, $tabledefn);

}

sub create_resource_type {
    $Log->enter();
    my ($self, %opts) =
          validate_mixed(@_,
                  [{type => OBJECT},],
                  {name    => {type => SCALAR,},
                   defcomplete => {type => SCALAR, optional => 1,},
                   primarykey  => {type => ARRAYREF,},
                   tabledefn   => {type => ARRAYREF,},},);


    if (ConfigBroker::CbUtils::verify_table_exists($self->{dbh}, $opts{name}) ) {
       $Log->comment("Resource typpe $opts{name} already exists.  Not creating again\n");
       return;
    }

    $self->_create_resource_table(%opts); 

    ### now create triggers that will operate on inserts and updates on this table.
    $self->_create_resource_table_triggers(resourcetype => $opts{name}, primarykey => $opts{primarykey});

    ### Now create a resource-group cross reference table with foreign key constraint
    ### The resourcegroup table already should exist.
    $self->_create_crossref_table(resourcetype => $opts{name});
}

sub _create_resource_tables_old {
   $Log->enter();
   my ($self) = shift;
   $Log->enter();

   my %tablefields;
   
   ### The database string should look like this :
   ### "name TEXT PRIMARY KEY, resourcegroup TEXT, num_disks TEXT, state TEXT" and so on including all of the fields.

   my $schemamap = $self->{'resourceschema'}->get_resource_schema();
   foreach my $typename (keys(%$schemamap))
   {
      my $pkeys = &ConfigBroker::Schema::get_primary_keys($typename);
      my @primary = @$pkeys;
      my $typefield = $schemamap->{$typename};
      my @typefieldarray = split(" ", $typefield);
      my $tabledefn;
      foreach my $field (@typefieldarray){
         my $isprimary = 0;
         foreach (@primary) {
            if ($_ eq $field){
               $isprimary = 1;
               last;
            }
         }
         if ($isprimary) {
            $tabledefn .= $field . " VARCHAR(100), " ; 
         }
         else {
            $tabledefn .= $field . " TEXT, " ; 
         }
      } 
      #my $primarykeys = &ConfigBroker::Schema::get_primary_keys($typename);
      my $primarykeys = "PRIMARY KEY (";
      my $count = scalar(@primary);
      foreach (@primary) {
         $count --;
         if ($count != 0) {
            $primarykeys .= "$_, ";
         }
         else {
            $primarykeys .= "$_)";
         }
      }
      $tabledefn .= $primarykeys;
      $tablefields{$typename} = "($tabledefn)";
   }

   foreach my $typname (keys(%tablefields)) {
      $Log->debug("Creating table $typname ; The typefields are $tablefields{$typname} \n");
      &ConfigBroker::CbUtils::create_table ($self->{'dbh'}, $typname, $tablefields{$typname});
   }
   $Log->exit();
   return 0;
}
 
sub _add_typedefn() {
   $Log->enter();
   my ($self, $typename, $typefields) = @_;
   $Log->enter();

   eval {
      #my $insert_cmd = "INSERT INTO \"typedefn\" (typename, typefields) " .
      my $insert_cmd = "INSERT INTO typedefn (typename, typefields, name) " .
                       "VALUES('$typename', '$typefields', '$typename')";
      $self->{'dbh'}->do($insert_cmd);
   };
   if ($@) {
      $Log->warn("Error inserting row into table \"typedefn\". Error is $@");
   }

   $Log->exit();
   return 0;
}

sub _delete_typedefn() {
   $Log->enter();
   my ($self, $typename) = @_;
   $Log->enter();

   eval {
      my $delete_cmd = "DELETE FROM typedefn " .
                       "WHERE typename='$typename'" ;
      $self->{'dbh'}->do($delete_cmd);
   };
   if ($@) {
      $Log->warn("Error delete row from table \"typedefn\". Error is $@");
   }

   $Log->exit();
   return 0;
}
sub get_typedefn() {

   my ($self) = @_;

   my $get_fields_cmd = "SELECT typefields FROM typedefn";
   my $fields = $self->{'dbh'}->selectcol_arrayref($get_fields_cmd);
   my @reskeys = split(" ", @$fields[0]);
}


## Get all the attributes of the table to display as form to get user input
sub get_customtype_fields() {

   my ($self,$typename) = @_;
   my $get_fields_cmd = "SELECT typefields FROM typedefn WHERE typename=\'$typename\'";
   $Log->comment("$get_fields_cmd");
   my $fields = $self->{'dbh'}->selectcol_arrayref($get_fields_cmd);
   my @reskeys = split(" ", @$fields[0]);
   return \@reskeys;
}


## Get all resource types from the typedefn table

sub get_typename{
	  my ($self) = @_;
  my $get_fields_cmd = "SELECT typename FROM typedefn";
  my $types = $self->{'dbh'}->selectcol_arrayref($get_fields_cmd);
	my @reskeys = @$types;
	return \@reskeys;
}

## Get all resource fields from the table for a particular resource type.

sub get_resname_type{
	$Log->enter();
	  my ($self,$typename) = @_;
  my $get_fields_cmd = "SELECT * FROM" ." $typename";
  $Log->comment($get_fields_cmd);
  my $types_names =  $self->{'dbh'}->selectall_arrayref($get_fields_cmd, {Slice => {}});
	return $types_names;
}

## Get all resource fields for a particular resource type and resource name

sub get_resource_fields_name{
	$Log ->enter();
	my ($self,$restype,$resname)=@_;
	my $get_fields_cmd="SELECT * FROM $restype WHERE $resname";
	$Log->comment($get_fields_cmd);
	my $resarrayref = $self->{'dbh'}->selectall_arrayref($get_fields_cmd, {Slice => {}});
	return $resarrayref->[0];
}


sub _populate_typedefn_table {
   $Log->enter();
   my ($self) = shift;
   $Log->enter();

   my $testbeddir = dirname ($self->{'path'});
   my $schemafile = "$testbeddir/cb.schema";
   my $schema_object;
   if (-e $schemafile) { 
      $schema_object = new ConfigBroker::Schema($schemafile);
   }
   else {
      $schema_object = new ConfigBroker::Schema();
   }
   $self->{'resourceschema'} = $schema_object;
   my $schemamap = $schema_object->get_resource_schema();
   foreach my $typename (keys(%$schemamap))
   {
      #my $typefields = "$schemamap->{$typename}" . "$engenx_fields";
      my $typefields = $schemamap->{$typename};
      $self->_add_typedefn($typename, $typefields);
   }
   $Log->exit();
   return 0;
}

sub delete_resource_group_from_db() {
   my ($self, $name) = @_;
   $Log->enter();

   eval {
      my $delete_cmd = "DELETE FROM resourcegroup " .
                       "WHERE name='$name'";
      $self->{'dbh'}->do($delete_cmd);
   };
   if ($@) {
      $Log->warn ("Error deleting row from table \"resourcegroup\". Error is $@");
   }

   $Log->exit();
   return 0;
}
  
sub create_supergroup() {
   $Log->enter();
   my ($self, %opts) = 
			validate_mixed(@_,
				[{type => OBJECT},],
				{groupname => {type => SCALAR,},
				subgroups_arrayref => {type => ARRAYREF,},},);


   ### Check if the supergroup exist.
   if (&ConfigBroker::CbUtils::does_supergroup_exist($self->{'dbh'}, $opts{'groupname'})) {
      $Log->error("Super group $opts{'groupname'} already exists.  Cannot create a new one\n");
      exit 1;
   }
   
   my $SG = new SuperGroup($opts{'groupname'}, $self->{'dbh'}, $opts{'subgroups_arrayref'});
   $Log->exit();
   return $SG;
}
  
sub create_resource_group() {
   $Log->enter();
   my ($self, %opts) = 
		validate_mixed(@_,
		[{type => OBJECT},],
		{name => {type => SCALAR,},
		configfile => {type => SCALAR,optional => 1},
		configengine => {type => SCALAR , optional => 1},
		reuse => {type => SCALAR , optional => 1, default => "REUSE" },
		share => {type => SCALAR, optional => 1, default => "SHARE" },
		cleanupfile => {type => SCALAR, optional => 1, ,default => ""},
		cmdlinestring => {type => SCALAR , optional =>1 },
		discover => {type => SCALAR, optional => 1,default => 0},
		is7mode => {type => SCALAR, optional => 1,default => 0},
		status => {type => SCALAR, optional => 1,default => "PENDING"},
		discoverobjs => {type => ARRAYREF , optional => 1},
		excludeobjs  => {type => ARRAYREF , optional => 1},
		node => {type => ARRAYREF, optional => 1,},},);

   if ($opts{'discover'} ==1){
      if (scalar(@{$opts{'discoverobjs'}}) < 1) {
         $opts{'discoverobjs'} = undef;
      }
   }
   

    my $name = $opts{'name'};
    $Log->debug("Create resourcegroup $name");
    my $configshare=$opts{'share'};
    my $configreuse = $opts{'reuse'};
    my $configfile = $opts{'configfile'};
    my $cleanupfile = $opts{'cleanupfile'};
    my $configengine = $opts{'configengine'};
    my $discover = $opts{'discover'}; 
    my $status = $opts{'status'};
     
    if ($discover == 1) {
         if(! defined $opts{'node'}) {
            $Log->error("Node is required for discovery.  Exiting...\n");
            exit 1;
         }
    } 
   
   my $cmdlinestring = "";
   $cmdlinestring = $opts{'cmdlinestring'}  if (defined $opts{'cmdlinestring'});
   
   ### Verify if the config exist. If so do the logic to decide if it is
   ### sharable, reusable etc. If not create a new config and return the
   ### config object to the caller.

   ### validate the input and create the resource group.
   if(! defined $opts{'configfile'}) {
      if (! defined $opts{'configengine'}) {
         $configengine = "";
          $configfile = "";
      }
      else {
         $Log->error("Configfile should be defined with config engine.  ERROR..\n");
         exit 1;
      }
   }
   else {
      if (! defined $opts{'configengine'}) {
         $Log->error("Configfile should be defined with configengine.  ERROR..\n");
         exit 1;
      }
      else
      {
        $configengine = $opts{'configengine'};
         $configfile   = $opts{'configfile'};
        # $cleanupfile  = $opts{'cleanupfile'} if defined ($opts{'cleanupfile'});
      }
   }
   ### Whether to share this config with others. Default is "SHARE"
   if( $opts{'share'} ) {
      if (
         ($opts{'share'} eq "SHARE") || 
         ($opts{'share'} eq "EXCLUSIVE")) {
             $configshare = $opts{'share'};
         }
      else {
         ### ERROR
         $Log->error("Bad argument for configshare\n");
         exit 1;
      }
   }
   ### Whether to reuse the existing config or not. Default is "NOREUSE"
   if( $opts{'reuse'} ) {
      if (($opts{'reuse'} eq "REUSE") || 
          ($opts{'reuse'} eq "NOREUSE") || 
          ($opts{'reuse'} eq "CREATENEW") ) { 
             $configreuse = $opts{'reuse'};
      }
      else {
         ### ERROR
         $Log->error("Bad argument for configreuse\n");
         exit 1;
      }
   }

   ### see if the resourcegroup exists and if so what to do based on the reuse/share values.
   ### existing must be SHARE & requesting must be REUSE to make use of the existing config.
   my $rgh = &ConfigBroker::CbUtils::get_resource_group($self->{'dbh'}, $name);
   if (defined $rgh)
   {
      my %old_rg = %{$rgh} ;
      $Log->debug("Existing $name resourcegroup: " . Dumper(\%old_rg));
      $Log->debug("Resource group $name exists\n");
      if ($old_rg{'name'} eq $name)
      {
         if ($configreuse eq "REUSE") {
            if ($old_rg{'share'} eq "SHARE") {
               $Log->debug("WILL REUSE EXISTING config $name\n");
               ### Check if the status is pending
               my $alarmval = 600;  # ten minutes
               eval {
                  local $SIG{ALRM} = sub { die "alarm\n" };
                  alarm $alarmval;
                  until ($rgh->{'status'} ne "PENDING") {
                     $Log->comment("Existing RG is in PENDING state. Sleeping and waiting for status change.\n");
                     sleep 10;
                     $rgh = &ConfigBroker::CbUtils::get_resource_group($self->{'dbh'}, $name);
                  }
                  alarm 0;
               };
               if ($@) {
                  if ($@ eq "alarm\n") {
                     $Log->error("Timed out waiting for PENDING state to change. \n");
                  }
                  else {
                     $Log->error("UNKNOWN ERROR\n");
                  }
                  exit 1;
               }
               if ($old_rg{'status'} =~ /INUSE/) {
                  $Log->error("WILL NOT REUSE EXISTING config $name because existing config status is RESERVED and is IN USE.  EXITING\n");
                  exit 1;
               }
	      my $RG = $self->get_resource_group_object($name);
               ### update the status to whatever was passed in.
               if ((defined $opts{'status'}) && ($opts{'status'} =~ /INUSE/)){ 
		$RG->update_status(newstatus => "READY");
               }
               #### else leave it alone as is.
               $Log->exit();
               return $RG;
            }
            else {
               $Log->error("WILL NOT REUSE EXISTING config $name because existing config is $old_rg{'share'} and not SHAREABLE.  EXITING\n");
               exit 1;
            }
         }
         elsif ($configreuse eq "NOREUSE") {
            $Log->error("WILL NOT REUSE EXISTING config $name because NOREUSE requested.  EXITING\n");
            exit 1;
         }
         elsif (($configreuse eq "CREATENEW") || ($configreuse eq "REUSE")) {
            ### Check if the status is pending
            if (($old_rg{'status'} eq "PENDING") || ($old_rg{'status'} =~ /INUSE/)) {
               $Log->error("Existing RG is in PENDING or INUSE state. Cannot delete\n");
               exit 1;
            }
            $Log->comment("Deleting existing resource group $name\n");
            $self->delete_resource_group(groupname => $opts{'name'});
         }
      }
   }

   ### If we had an existing resourcegroup and we are reusing it, we've already returned.
   ### If we are here, we are creating a new one.
      
   ### create a new resource group.
   my %res_info;
   $res_info{'dbh'}           = $self->{'dbh'};
   $res_info{'name'}          = $name;
   $res_info{'reuse'}         = $configreuse;
   $res_info{'share'}         = $configshare;
   $res_info{'configfile'}    = $configfile;
   $res_info{'configengine'}  = $configengine;
   $res_info{'cleanupfile'}   = $cleanupfile;
   $res_info{'cmdlinestring'} = $cmdlinestring if ($cmdlinestring ne "");
   $res_info{'dsn'}           = $self->{'dsn'};
   $res_info{'status'}        = $status;
   my $RG = new ConfigBroker::ResourceGroup(%res_info); 
	
   if ($discover) {
      my $is7mode = $opts{'is7mode'} ? $opts{'is7mode'} : 0;
      ### if we are reusing and rediscovering partial, we want to delete
      ##i# the specific type of resources.
      if (($configreuse eq "REUSE") && (defined $opts{'discoverobjs'})) {
         my $objsref = $opts{'discoverobjs'};
          my @objs = @$objsref;
         foreach my $type (@objs) {
            ### delete this type of resource in this resourcegroup
            $RG->delete_resources_for_type_from_group(type=>$type);
         }
      }
      ### we do discovery if we are not reusing OR
      ### if we are reusing and only doing partial discovery of certain type of resources. 
      if ( $configreuse ne "NOREUSE" ) {
         my $node = $opts{'node'};
         my %engineinfo;
         if ($is7mode) {
		### Do the 7-mode discovery 
		%engineinfo = (
			'enginename'        => "Discover7M",
			'discover'          => 1,
			'resourcegroup'     => $name,
			'node' 		   => $node,
			'discoverobjs'      => $opts{'discoverobjs'},
			'excludeobjs'       => $opts{'excludeobjs'},
			'dbh'               => $self->{'dbh'}
		);
	
         }
	
         else {
         ### Do the c-mode discovery 
         %engineinfo = (
            'enginename'        => "Discover",
            'discover'          => 1,
            'resourcegroup'     => $name,
            'node' 		   => $node->[0],
            'discoverobjs'      => $opts{'discoverobjs'},
	    'excludeobjs'       => $opts{'excludeobjs'},
            'dbh'               => $self->{'dbh'}
         );
     }
         require ConfigBroker::Engine;
         my $engine_obj = ConfigBroker::Engine::get_engine_object(\%engineinfo);
          my $rc = $engine_obj->run_discovery();
	  if ($rc != 0) {
	 $Log->error("Failure returned from engine discovery.\n");
	    $Log->comment("Deleteing resource group $name that may not be complete and exiting.\n");
	    $self->delete_resource_group_from_db($name);
	    $Log->comment("Done cleaning up. EXITING.\n");
	    exit 1;
	
         }
      }
   }
   else {
      if ($configengine ne "") {
         #my $rgobj = $self->get_resource_group_object($self->{name});
         my %engineinfo = (
            'enginename' 	=> $configengine,
            'configfile'   	=> $configfile,
            'cmdlinestring'	=> $cmdlinestring,
            'resourcegroup'     => $name,
            'dbh'               => $self->{'dbh'}
         );
         require ConfigBroker::Engine;
         my $engine_obj = ConfigBroker::Engine::get_engine_object(\%engineinfo);
         my $rc = $engine_obj->run_setup();
         ### Check to see if the setup failed.  If so, remove the newly created group and any resources
	 if ($rc != 0) {
	    $Log->error("Failure returned from engine setup.\n");
            $Log->comment("Deleteing resource group $name that may not be complete and exiting.\n");
	    $self->delete_resource_group_from_db($name);
	    $Log->comment("Done cleaning up. EXITING.\n");
	    exit 1;
	 } 
	 my $cleanupfile = $engine_obj->get_cleanupfile();
         $RG->update_cleanupfilename($cleanupfile);
      } 
   }
   $RG->update_status(newstatus => "READY");
   ### update the status to "READY"
   if ((defined $opts{'status'}) && ($opts{'status'} =~ /INUSE/)) {
	$RG->update_status(newstatus => "READY");
   }
   $Log->exit();
   return $RG;
  
}

sub does_group_exist_already {
   $Log->enter();
   my ($self, $name) = @_;

   my $rgh = &ConfigBroker::CbUtils::get_resource_group($self->{'dbh'}, $name);
   if (defined $rgh) {
      $Log->debug("Resource Group $name exists". Dumper($rgh));
      $Log->comment("Resource group $name exists\n");
      return 1;
   }
   else {
      $Log->comment("Resource group $name does not exist\n");
      return 0;
   }
} 

sub get_tasks_object {
   $Log->enter();
   my ($self) = @_;

   my %dbinfo; 
   $dbinfo{'dbh'} = $self->{'dbh'};
   $dbinfo{'dsn'} = $self->{'dsn'};
   my $TasksObj = new ConfigBroker::Tasks(%dbinfo);
   return $TasksObj;
}

sub get_resource_group_object {
   $Log->enter();
   my ($self, $groupname) = @_;

   my $RG = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $groupname, $self->{'dsn'});

   $Log->exit();
   return $RG;
   
}


sub get_resource_group {
    $Log->enter();
    my ($self, %opts) =
          validate_mixed(@_,
                  [{type => OBJECT},],
		  {name => {type => SCALAR,},},);

    my $rg = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $opts{'name'}, $self->{'dsn'});
    $Log->exit();
    return $rg;
}

sub get_all_resource_groups {
   $Log->enter();
   my ($self) = @_;

   my @rg_list = &ConfigBroker::CbUtils::get_all_resource_groups($self->{'dbh'});
   $Log->exit();
   return (@rg_list);
}

sub delete_all_resource_groups {
    $Log->enter();
    my ($self) = @_;

    my @rg_list = &ConfigBroker::CbUtils::get_all_resource_groups($self->{'dbh'});

    foreach my $name (@rg_list) {
        $Log->debug("Deleting RG $name\n");
        $self->delete_resource_group(groupname => $name);
    }
}

sub get_resourcegroup_from_type {
   my ($self, %opts) = 
		validate_mixed(@_,
			[{type => OBJECT},],
			{type => {type => SCALAR,},},);

   my $maxrecords = 1;
   my $rgname = &ConfigBroker::CbUtils::get_resourcegroup_name($self->{'dbh'}, $opts{'type'}, $maxrecords);
   $Log->exit();
   return $rgname;
}

sub delete_resource_group
{
   $Log->enter();
   my ($self, %opts) =
        validate_mixed(@_,
               [{type => OBJECT},],
               {groupname => {type => SCALAR,},
                cleanupfile => {type => SCALAR, optional => 1,},},);

   my $rg_record = &ConfigBroker::CbUtils::get_resource_group($self->{'dbh'}, $opts{groupname});
   if (!defined $rg_record)
   { 
      $Log->comment("Resource group $opts{groupname} doesn't exist. No need to delete anything.\n");
      $Log->exit();
      return 0;
   }

   my $rgh = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $opts{groupname});

   if ((defined $rgh->{'configengine'}) && (length ($rgh->{'configengine'} ne ""))) {
      my $clnup;
      if ((defined $opts{cleanupfile}) && (length ($opts{cleanupfile} ne "")))  {
         $clnup = $opts{cleanupfile};
      } 
      elsif (defined $rgh->{'cleanupfile'}) {
         $clnup = $rgh->{'cleanupfile'};
      }
      elsif (defined $rgh->{'configfile'}) {
         $clnup = $rgh->{'configfile'};
      }
      else {
         $Log->debug("Cleanup file is not provided or stored for this resourcegroup..\n");
      }

      my %engineinfo = (
         'enginename'        => $rgh->{'configengine'},
         'cleanupfile'       => $clnup,
         'resourcegroup'     => $opts{groupname},
         'dbh'               => $self->{'dbh'}
      );
      ### Override stored cleanup file if one provided now.
      require ConfigBroker::Engine;
      my $engine_obj = ConfigBroker::Engine::get_engine_object(\%engineinfo);
      $engine_obj->run_cleanup();
   }
   my $RG = $self->get_resource_group_object($opts{groupname}); 
   $Log->debug("Deleting resources from DB for group $RG->{'name'}\n");
   $RG->delete_all_resources_from_group();

   $self->delete_resource_group_from_db($opts{groupname});
   $Log->exit();
   return 0;
}

sub _find_name_from_nacl_node {
   $Log->enter();
    my ($self, %opts) =
          validate_mixed(@_,
                  [{type => OBJECT},],
                  {nodeobj => {type => OBJECT, optional => 1,},},);
 
   # just get the first node
   my $nodeobj = ($opts{'nodeobj'}) ? $opts{'nodeobj'} : $self->_get_nacl_node_objects()->[0];

   $Log->exit();
   return $nodeobj->hostrec->id();

}


sub _get_nacl_node_objects { 
  $Log->enter();
   my ($self, %opts) =
         validate_mixed(@_,
                 [{type => OBJECT},],
                 {name => {type => SCALAR, optional => 1,},},);

my @nodes = NACL::C::Node->find();
my @return;
	
	      foreach my $node (@nodes) {
				 #$node->apiset_must(interface => 'ZAPI');
				 if (not $opts{name}) { 
	
            push (@return, $node);
	
	
        } elsif (($opts{name}) and ($node->hostrec->id() =~ /$opts{name}/i)) {

            push (@return, $node);
        }
	      }
	        $Log->warn("The node named $opts{name} was not found") if (not scalar(@return));

    $Log->exit();

    return \@return;


}

sub discover_and_create_resource_group
{
   $Log->enter();
   my ($self, %opts) =
         validate_mixed(@_,
                [{type => OBJECT},],
                {nodename => {type => SCALAR, optional => 1,},
                 reuse => {type => SCALAR, optional => 1, default => 'REUSE'},
		 discoverobjs => {type => ARRAYREF, optional => 1,default => []},
		 excludeobjs => {type => ARRAYREF, optional => 1,default => []},
                 groupname => {type => SCALAR, optional => 1},
                 is7mode => {type => SCALAR, optional => 1},
                 discoverclients => {type => SCALAR, optional => 1,default => 0},
		 clients => {type => ARRAYREF, optional => 1,default => []},
                 discoverusers => {type => SCALAR, optional => 1,default => 0},
                 share => {type => SCALAR, optional => 1,default => "SHARE"},},);

   my $discoverobjs = $opts{'discoverobjs'};
   my $excludeobjs = $opts{'excludeobjs'};
   my $is7mode = ($opts{'is7mode'}) ? $opts{'is7mode'} : 0;

   my $nodename;
   my $node;
   if ($is7mode) {
      $nodename = ($opts{'nodename'}) ? $opts{'nodename'} : $self->_find_name_from_nacl_node();
      $node = $self->_get_nacl_node_objects();
   }
   else {
      $nodename = ($opts{'nodename'}) ? $opts{'nodename'} : $self->_find_name_from_nacl_node();
      $node = $self->_get_nacl_node_objects(name => $nodename);
   }

   my $default_rgname = "discover_$nodename";
   my $rgname = ($opts{'groupname'}) ? $opts{'groupname'} : $default_rgname;

   my %discoverinfo = ( 
      'node'     => $node,
      'name'     => $rgname,
      'reuse'    => $opts{reuse},
      'share'    => $opts{share},
      'discover' => 1,
      'discoverobjs' => $discoverobjs,
      'excludeobjs'  => $excludeobjs,
      'is7mode'  => $is7mode,
   );

   my $RG = $self->create_resource_group(%discoverinfo);
   my $ret = 0;
   if ((defined $opts{discoverclients}) && ($opts{discoverclients} == 1)) {
      if (defined $opts{clients})  {
         my $clients = $opts{clients};
         $ret = $RG->discover_clients_into_group(clients => $clients);
      }
      else {
         $ret = $RG->discover_clients_into_group();
      }
   }
   if ($ret) {
      $Log->error("Client Discovery failed. Exiting\n");
      exit 1;
   }
   if ((defined $opts{discoverusers}) && ($opts{discoverusers} == 1)) {
       $ret = $RG->discover_users_into_group();
   }
   if ($ret) {
      $Log->error("Users Discovery failed. Exiting\n");
      exit 1;
   }

   $Log->exit(Dumper($RG));
   return $RG;

}

sub discover_and_create_client_resource_group
{
   $Log->enter();
   my ($self, %opts) =
         validate_mixed(@_,
                [{type => OBJECT},],
                {reuse => {type => SCALAR, optional => 1, default => 'REUSE'},
                 groupname => {type => SCALAR, optional => 1},
		 clients => {type => ARRAYREF, optional => 1,default => []},},);

   my $default_rgname = "discover_clients";
   my $rgname = ($opts{'groupname'}) ? $opts{'groupname'} : $default_rgname;

   my %groupinfo = ( 
      'name'     => $rgname,
      'reuse'    => $opts{reuse},
   );

   my $RG = $self->create_resource_group(%groupinfo);
   my $ret = 0;
   if (defined $opts{clients})  {
      my $clients = $opts{clients};
      $ret = $RG->discover_clients_into_group(clients => $clients);
   }
   else {
      $ret = $RG->discover_clients_into_group();
   }
   if ($ret) {
      $Log->error("Client Discovery failed. Exiting\n");
      exit 1;
   }

   $Log->exit(Dumper($RG));
   return $RG;

}

sub discover_users_into_resource_group {
    $Log->enter();
    my ($self, %opts) = validate_mixed(
        @_,
        [{type => OBJECT},],
        {groupname => {type => SCALAR, optional => 1}},
    );

    my $default_rgname = "cluster";
    my $rgname = ($opts{'groupname'}) ? $opts{'groupname'} : $default_rgname;

    my $RG = $self->get_resource_group(name => $rgname);
    my $ret = 0;
    $ret = $RG->discover_users_into_group();
    if ($ret) {
        $Log->error("User Discovery failed. Exiting\n");
        exit 1;
    }

    $Log->exit(Dumper($RG));
    return $RG;

} ## end sub discover_users_into_resource_group

sub get_default_discovery_objects {
   my $self = shift;

   my $default_disco_objs = &ConfigBroker::Schema::get_nacl_resource_names();
   return $default_disco_objs;
}

sub disconnect {
   my $self = shift;
   $self->{'dbh'}->{InactiveDestroy} = 0;
   $self->{'dbh'}->disconnect()
    or warn "Disconnection failed: $DBI::errstr\n";
}

sub DESTROY {
   my ($self) = @_;
   $self->disconnect();
}

sub get_nasgroup_object()
{
    my $self = shift;
    my %opts = validate_with(
                    params => \@_,
                    spec => {},
                    allow_extra => 1,
                );

    # add the valuable items from CB to incoming opts and let NASGroup judge
    # the validity of the options
    $opts{'name'} = "NASGroup";
    $opts{'dbh'}  = $self->{'dbh'};

    $self->_verify_and_store_schema_or_die( $opts{'name'}, $ConfigBroker::ResourceGroup::NASGroup::SCHEMA_VERSION);

    my $NASRG = new ConfigBroker::ResourceGroup::NASGroup(%opts);
    return $NASRG;
}

sub _create_version_table {
   my ($self) = @_;

   my $table_defn = "(package VARCHAR(100) PRIMARY KEY, version TEXT)";
   ConfigBroker::CbUtils::create_table ($self->{'dbh'}, $self->{CB_Version_Table}, $table_defn);
}


sub _verify_and_store_schema_or_die {
   my ($self, $package_name, $package_version) = @_;

   # Check for the presence of version file to guage if the version table was created during 
   # the database initailization  
   if (! ConfigBroker::CbUtils::verify_table_exists($self->{dbh}, $self->{CB_Version_Table}) ) {
      die config_result(qq[Unable to verify ConfigBroker version because main versoning ].
                        qq[table '$cb_version_table' was not created during database initialization!]);

   }

   my $get_res_cmd = "SELECT version FROM $self->{CB_Version_Table} ";
   $get_res_cmd .=   "WHERE package=\'$package_name\'";
   $Log->debug("$self->{CB_Version_Table}  query $get_res_cmd");


   #Presuming that the versioning table exist
   my ($cb_stored_version) = $self->{dbh}->selectrow_arrayref($get_res_cmd);

   # If the versioning table exists, but the package shcema version entry does not exist, 
   # the returned value ($stored_version) will be undef 
   my $stored_version = ref($cb_stored_version) eq "ARRAY" ? $cb_stored_version->[0] : $cb_stored_version;

   # First time version is being logged
   if (! defined $stored_version ) {
      if ($package_name eq "ConfigBroker") { 
         ### If it is the main package and the main schema, we are adding it after init_db for now.
         ### So, we should never have undef for this.
         ### BELOW IS ANOTHER TEMPORARY HACK FOR VERSION CHECK.                             
         ### WE ARE ADDING THE VERSION WEHNEVER WE INIT THE DB.
         my $msg ="There is a Configbroker database schema incompatability for the '$package_name' schema/package!\n";
         $msg  .= "The exsisting version is '$stored_version' and current version is '$package_version':\n";
         $msg  .= "Before re-running your test, you *MUST* cleanup the existing database by running the following:\n";
         $msg .= "\tntest system_test/workload/tools/cb.thpl CB_STOP=1 CB_OBLITERATE=1\n";
         die config_result($msg);
      }
      #### For all other packages, proceed to add.
      $Log->debug(qq[Adding '$package_name' schema/package version '$package_version' to version table $self->{CB_Version_Table}]);
   }
   # Compare the versions and die with message to cleanup and rerun
   elsif ($stored_version != $package_version) {

      my $msg ="There is a Configbroker database schema incompatability for the '$package_name' schema/package!\n";
      $msg  .= "The exsisting version is '$stored_version' and current version is '$package_version':\n";
      $msg  .= "Before re-running your test, you *MUST* cleanup the existing database by running the following:\n";
      $msg .= "\tntest system_test/workload/tools/cb.thpl CB_STOP=1 CB_OBLITERATE=1\n";
      die config_result($msg);
   }
   else {
                    
      $Log->debug("Schema version check for class $package_name is consistent -  version is $package_version");
      return 0;
   }

   # Store the record
   my $add_record = qq[INSERT INTO $self->{CB_Version_Table} (package, version) ];
   $add_record .= qq[values ('$package_name', '$package_version')];
   $Log->debug("ADD record IS $add_record\n");
   $Log->comment("ADD record IS $add_record\n");
   $self->{dbh}->do($add_record);
}



=head1 METHODS

=head2 create_resource_group

    Object method

=head3 Usage

    $cb->create_resource_group(%opts)
    $opts{'name'} - name of the resource group.  This is a mandatory parameter.
    $opts{'configfile'} - name of the input file used by the config engine 
    $opts{'configengine'} - The config engine to use to create the actual resources on the cluster/client/..etc. 
    $opts{'reuse'}  - Whether the resource group can be reused if it was already in the DB. 
    $opts{'share'} - Whether this resource group can be shared. 
    $opts{'cmdlinestring'} - Any command line string that needs to be passed in for the engine used.
    $opts('discover'} - Whether to search for an existing resource group in the DB.
    $opts{'status'}- Status of the resource group to be set.
    $opts{'node'} - Name of the node to search for resource groups . This is mandatory only when we are discovering existing resources.
    $opts('discoverobjs'} - Array of resource objects to be discovered .	
    All of the above are optional except the "name".  If no configengine/configfile are given, just a resourcegroup is created
    with no resources configured.  Resources can be added to the group using the resourcegroup object.

=head3 Return

    The ResourceGroup object.

=head2 create_resource_type

    Object method

=head3 Usage

    $cb->create_resource_type(%opts)
    $opts{'name'} - name of the resource type.  This is a mandatory parameter.
    $opts{'defcomplete'} - if this is a new custom type the definition is not complete in CB. Hence this should be 0.
    $opts{'primarykey'}  - the fields that form the primary key for this type of resource.  This is mandatory ARRAYREF parameter.
    $opts{'tabledefn'}  - the attribute fields for this new type of resource.  This is mandatory ARRAYREF parameter.

=head2 delete_resource_group

    Object method

=head3 Usage

    $cb->delete_resource_group(%opts)
    $opts{'groupname'} - name of the resource type.  This is a mandatory parameter.
    $opts{'cleanupfile'} - optional cleanup file name. .

=head3 Return

    0=Success

=head2 delete_all_resource_groups

    Object method

=head3 Usage

    my $CB = new ConfigBroker();
    $CB->delete_resource_group();

=head2 discover_and_create_resource_group

    Object method

=head3 Usage

    my $CB = new ConfigBroker();
    $CB->discover_and_create_resource_group(%opts);
    $opts{'nodename'} - the node to connect to for discovering resources.
    $opts{'share'} - if the group can be shared.
    $opts{'reuse'} - if the group can be reused if already exists and shareable.
    $opts{'groupname'} - optional. If not provided, the name will be discover_<nodename>
    $opts{'is7mode'}   - optional. If not set to 1, the default is 0 and the discovery is for cmode. 
    $opts{'discoverclients'}   - optional. If not set to 1, the default is 0 and clients (windows/*nix) discovery will not be done. 
    $opts{'clients'} - Reference to an array of the clientnames to discover. Example: ["rhel5-016-162","win7-43-061"] . Default is empty.
                       This value can also be passed in through command line Options CLIENT or CLIENTS or CLIENTFILE.
    $opts{'discoverobjs'} - Reference to an array of the type of resources to discover instead of the default list ("volume", "aggregate",.. etc.). 
    $opts{'excludeobjs'}  - Reference to an array of the type of resources NOT to discover by default ("volume", "aggregate",.. etc.). 

=head3 Return

    The ResourceGroup object.

=head2 discover_and_create_client_resource_group

    Object method

=head3 Usage

    my $CB = new ConfigBroker();
    $CB->discover_and_create_client_resource_group(%opts);
    $opts{'reuse'} - if the group can be reused if already exists and shareable.
    $opts{'groupname'} - optional. If not provided, the name will be "discover_clients"
    $opts{'clients'} - Reference to an array of the clientnames to discover. Example: ["rhel5-016-162","win7-43-061"] . Default is empty.
                       This value can also be passed in through command line Options CLIENT or CLIENTS or CLIENTFILE.

=head3 Return

    The ResourceGroup object.

=head2 get_resource_group

    Object method

=head3 Usage

    my $CB = new ConfigBroker();
    my $RG = $CB->get_resource_groupt(%opts);
    %opts{'groupname'} - Name of the resource group to be found.

=head3 Return

    The resourcegroup object with the name $groupname that has already been created and exists.

=head2 get_tasks_object

    Object method

=head3 Usage

    my $CB = new ConfigBroker();
    my $tasks_object = $CB->get_tasks_object();

=head3 Return

    The Tasks object. 

=head2 get_default_discovery_objects

    Object method

=head3 Usage

    my $CB = new ConfigBroker();
    my $default_disco_objs = $CB->get_default_discovery_objects();

=head3 Return

    An arrayref containing all the filer objects that would be discovered by default. 

=head2 get_all_resource_groups

    Object method

=head3 Usage

    my $CB = new ConfigBroker();
    my @resourcegroups = $CB->get_all_resource_groups();

=head3 Return

    An array of all the existing resourcegroup names. 

=head2 create_super_group

    Object method

=head3 Usage

    $cb->create_supergroup(%opts)
    %opts{'supergroupname'}=Name of the supergroup to be created.
    %opts{'subgroups'} - Array of resource groups to be added to the supergroup.
=head3 Return

    The SuperGroup object. (OR) 1 - if failure.

=head2 get_resourcegroup_from_type

Retrieves Resource groups based on type .

=head3 USAGE 
     
    $cb->get_resourcegroup_from_type(%opts)
    %opts{'type'} - Type of the resource group to be retieved .	

=cut

=head2 db_dump

    Object method
    Dumps the database into a CSV format files for each table or into a SQL format file 

=head3 Usage

    $cb->db_dump(%opts);
    $opts{'format'}      - Optional scalar values of "CSV" or "SQL" to be passed in for the format of the dump output. Default is "CSV".
    $opts{'output_dir'}  - Required scalar value of full directory path of where the dump output should be directed to.
    $opts{'output_file'} - Optional scalar value of a file name to write the SQL format output. The output will be directed to this file in the output dir
                           Default file is "cb_backup.sql"

=head2 stop

    Object method
    Kills the database mysqld server process and removes the pid/host/port files that CB uses.
    This method does NOT remove the entire database.  Just removes the server process.

=head3 Usage

    $cb->stop;

=head2 obliterate 

    Object method
    Kills the database mysqld server process and removes the database.
    CAUTION: This method WILL remove the entire database and your DATA WILL BE LOST.

=head3 Usage

    $cb->obliterate;


=head2 get_nasgroup_object

    Object method

=head3 Usage

    $cb->get_nasgroup_object(%opts);
    $opts{'protocols'} - Optional arrayref of protocols can be passed in.  Default is a list containing "nfs" and "cifs".
    $opts{'timeout'} - Optional scalar integer for number of seconds to wait for protocols to appear in CB before dying.
	  
=head3 Return

    The NASGroup object. Exits upon any failure.

=cut

1;
