package ConfigBroker::Engine::EngineDiscoverUsers;

use strict;
use warnings;
use Tharn qw($Log script_result host);
use Data::Dumper;
our @ISA = qw (ConfigBroker::Engine Exporter);
use ConfigBroker::CbUtils;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF BOOLEAN ARRAYREF UNDEF);
use Module::Common qw(validate_mixed);
use Options;
use Params::Common;
use Hostrec;
use NATE::ParamSet;
use XML::Simple;
use NATE::Inc qw(find_file);

sub new {
    my ($pkg, $configinfo) = @_;
    $Log->enter();
    my $class = ref($pkg) || $pkg;
    my $self = $class->SUPER::new($configinfo);
    bless($self, $class);

    $Log->exit();
    return $self;
} ## end sub new

sub run_discovery() {
    my ($self) = @_;
    $Log->enter();

    # Get the connection object for the nodes
    my $ci = NACL::C::Node->find();

    my $ret = 0;

    # Get the tale definition
    my $get_fields_cmd = "SELECT typename FROM typedefn";
    my $fields         = $self->{'dbh'}->selectcol_arrayref($get_fields_cmd);
    my $RG = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'},
        $self->{'resourcegroup'});

    # Get the path to the user config and the desired user config ids
    my $Options = new Options;
    $Options->process();
    my $options_hash = $Options->get_hash();

    # Get the discovered home directory search paths
    my $search_paths
        = $RG->get_resources(type => 'vservercifshomedirectorysearchpath');

    # Get the directories in each homedir search path. The assumption being
    # that each dir is a homedir named as the users name
    my $homedirs = {};
    if (ref($search_paths) && @{$search_paths}) {

        # For each search path get a list of user dirs
        foreach my $sp (@{$search_paths}) {
            my $path    = $sp->{path};
            my $vserver = $sp->{vserver};

            # Get the discovered volumes for this vserver
            my $volumes = $RG->get_resources(
                type          => 'volume',
                match_filters => {vserver => $vserver}
            );

            # Find the relative path from the last volume
            my ($volume, $relative_path)
                = _get_relative_volume_path($path, $volumes);

            # get the DSID from the volume
            my @matches = grep { $_->{volume} eq $volume } @{$volumes};
            my $dsid = $matches[0]->{dsid};

            $Log->comment(
                "Searching for user directories for $vserver in $dsid.$path");

            my @dirs = NACL::C::Volume->explore(
                command_interface => $ci,
                format            => 'dir',
                scope             => "$dsid.$relative_path",
            );

            # Save each users homedir path in a hash
            if (@dirs) {
                foreach my $dir (@dirs) {
                    my $dir_name = $dir->{name}->[0];
                    if ($dir_name =~ /\.$|\."$/) { next; }
                    $Log->debug("Saving $vserver:$path:$dir_name");
                    $homedirs->{$vserver}->{$dir_name} = "$path/$dir_name";
                }
            }
        }
    }

    # Read the users from the XML
    my $users = [];

    # Get the user realted params
    my $cb_user_config = $options_hash->{CB_USER_CONFIG};
    my $cb_users       = $options_hash->{CB_USERS};

    # If we have both params read the users from the file
    if ($cb_user_config && $cb_user_config) {
        $Log->debug("Reading CB_USER_CONFIG($cb_user_config)");

        # Use NATE to search for the file and return the full path
        my $file = find_file($cb_user_config);
        if (!$file) {
            die script_result(
                "unable to find $options_hash->{CB_USER_CONFIG}");
        }

        # Read the XML
        my $xml = new XML::Simple;
        my $user_config = $xml->XMLin($file, forcearray => ['users']);

        # set the first user id
        my $id = 1;

        # Walk through all specified user config ids and and retrieve the
        # the users from the XML
        foreach my $user_config_id (split(/,/, $cb_users)) {
            $Log->comment("Reading user config id $user_config_id");

            # Users key should be an arrayref
            my @uc;
            if (defined($user_config->{$user_config_id}->{users})) {
                @uc = @{$user_config->{$user_config_id}->{users}};
                if (ref($user_config->{$user_config_id}->{users}) ne 'ARRAY')
                {
                    push(@uc, $user_config->{$user_config_id}->{users});
                }
            }

            # Walk through every user specification and add it to the users
            # array
            foreach my $user (@uc) {
                my $ad           = $user->{ad};
                my $count        = $user->{count};
                my $host         = $user->{host};
                my $hosttype     = $user->{hosttype};
                my $nis          = $user->{nis};
                my $password     = $user->{password};
                my $user_name    = $user->{user_name};
                my $zero_padding = $user->{zero_padding};
                my $protocol     = $user->{protocol};

                # Create each users entry in the array
                if ($count) {

                    # If the count tag is specified do the work to create that
                    # many users
                    for (my $i = 1; $i <= $count; $i++) {

                        # Copy the counter var so we can pad it later if
                        # needed
                        my $idx = $i;

                        # Pad the user index if required
                        if ($zero_padding) {
                            $idx = sprintf("%0*d", $zero_padding, $i);
                        }

                        # If the password tag is not specified in the XML then
                        # use the user name as the password
                        if (!defined($user->{password})) {
                            $password = $user_name . $idx;
                        }
                        my $name = $user_name . $idx . '_' . $id;
                        if (defined($host)) { $name .= '_' . $host; }

                        # Push the user info into the users AoH
                        push(
                            @{$users},
                            {   id        => $id,
                                name      => $name,
                                user_name => $user_name . $idx,
                                ad        => $ad,
                                host      => $host,
                                hosttype  => $hosttype,
                                nis       => $nis,
                                password  => $password,
                                protocol  => $protocol
                            }
                        );

                        # Increment the user id for database uniqueness
                        $id++;
                    }
                } else {

                    # If the password tag is not specified in the XML then
                    # use the user name as the password
                    if (!$password) {
                        $password = $user_name;
                    }

                    my $name = $user_name . '_' . $id;
                    if (defined($host)) { $name .= '_' . $host; }

                    # Push the user info into the users AoH
                    push(
                        @{$users},
                        {   id        => $id,
                            name      => $name,
                            user_name => $user_name,
                            ad        => $ad,
                            host      => $host,
                            hosttype  => $hosttype,
                            nis       => $nis,
                            password  => $password,
                            protocol  => $protocol
                        }
                    );

                    # Increment the user id for database uniqueness
                    $id++;
                }
            }
        }

        if ($cb_users =~ /discovered/) {
            $Log->comment("Discovering CIFS CM local users");

            # Get the disacovered vserver info so we can know the cifs_server
            my $vservers = $RG->get_resources(type => 'vservercifs');
            my %cifs_servers;
            foreach my $vs (@{$vservers}) {
                $cifs_servers{$vs->{vserver}} = $vs->{cifs_server};
            }

            # Get the discovered local users that are active
            my $local_users = $RG->get_resources(
                type          => 'vservercifsusersandgroupslocaluser',
                match_filters => {is_account_disabled => 'false'}
            );

            if (@{$local_users}) {

                # Walk through the list of local users
                foreach my $lu (@{$local_users}) {

                    if (!defined($lu->{full_name})) { next; }

                    # Strip the host from the user_name value
                    my $user_name = $lu->{user_name};
                    my $vserver   = $lu->{vserver};
                    my $host      = $cifs_servers{$vserver};
                    $user_name =~ s/$host\\//g;
                    my $name
                        = $user_name . '_' 
                        . $id . '_' 
                        . $host . '_'
                        . $vserver;

                    # Add the users homedir path if we found it in an
                    # earlier step
                    my $homedir;
                    if (defined($homedirs->{$vserver}->{$user_name})) {
                        $homedir = $homedirs->{$vserver}->{$user_name};
                        $Log->debug(
                            "Found homedir for local user $vserver:$user_name:$homedir"
                        );
                    } else {
                        $Log->debug(
                            "No homedir found for local user $vserver:$user_name"
                        );
                    }

                    # Push the user info into the users AoH
                    push(
                        @{$users},
                        {   id        => $id,
                            name      => $name,
                            user_name => $user_name,
                            ad        => '',
                            homedir   => $homedir,
                            host      => $host,
                            hosttype  => $ci->hostrec()->hosttype(),
                            nis       => '',
                            password  => $lu->{full_name},
                            vserver   => $vserver,
                            protocol  => 'cifs'
                        }
                    );

                    # Increment the user id for database uniqueness
                    $id++;
                }
            }
            $Log->comment("Discovering 7M useradmin users");

            # Get the discovered local users
            $local_users = $RG->get_resources(type => 'useradmin',);

            if (@{$local_users}) {

                # Walk through the list of local users
                foreach my $lu (@{$local_users}) {

                    my $name
                        = $lu->{'name'} . '_' 
                        . $id . '_'
                        . $lu->{'node'} . '_'
                        . $lu->{'vfiler'};

                    my $cifs_info = $RG->get_resources(
                        type          => 'cifs',
                        match_filters => {vfiler => $lu->{'vfiler'}}
                    );
                    my $host = $cifs_info->[0]->{netbios_servername};

                    my $hosttype;
                    my $node = $lu->{'node'};
                    eval { $hosttype = host($node)->hosttype(); };
                    if ($@) { $Log->comment($@); }

                    # Push the user info into the users AoH
                    push(
                        @{$users},
                        {   id        => $id,
                            name      => $name,
                            user_name => $lu->{'name'},
                            ad        => '',
                            homedir   => $lu->{'homedir'},
                            host      => $host,
                            hosttype  => $hosttype,
                            nis       => '',
                            password  => $lu->{full_name},
                            vserver   => $lu->{'vfiler'},
                            protocol  => 'cifs,nfs'
                        }
                    );

                    # Increment the user id for database uniqueness
                    $id++;
                }
            }
        }

        # Here we perform a second pass to match up discovered homedirs with
        # users
        foreach my $vserver (keys(%{$homedirs})) {
            foreach my $user_name (keys(%{$homedirs->{$vserver}})) {
                for (my $i = 0; $i < @{$users}; $i++) {

                    # Skip if the user name does not match
                    if ($users->[$i]->{user_name} ne $user_name) {
                        next;
                    }

                    # Skip if the vserver is defined and it does not match
                    if (defined($users->[$i]->{vserver})
                        && $users->[$i]->{vserver} ne $vserver) {
                        next;
                    }

                    # Skip if homedir is defined
                    if (defined($users->[$i]->{homedir})) { next; }

                    # Add the homedir path and vserver to the user info
                    $users->[$i]->{homedir}
                        = $homedirs->{$vserver}->{$user_name};
                    $users->[$i]->{vserver} = $vserver;
                    if (!defined($users->[$i]->{host})) {
                        $users->[$i]->{name} .= '_' . $vserver;
                    }
                }
            }
        }

        # Turn NATE output back on
        local $NATE::Log::CommEvent::CommLogLevel = "debug";

        # Add the users to ConfigBroker
        if (@{$users}) {
            my $objtype = "user";
            if (scalar(@{$users}) > 0) {
                $Log->debug("Adding objecttype $objtype\n");
                $RG->add_resources(
                    resource_type    => lc($objtype),
                    resourcearrayref => $users
                );
            }
        }
    }
    $Log->exit();
    return $ret;
} ## end sub run_discovery()

sub _get_relative_volume_path {

    $Log->enter();
    my ($path, $volumes) = @_;

    $Log->debug("Getting volume from homedir search path $path");

    # Try the exact match
    $Log->debug("Searching volume junction paths for an exact match");
    my @matches = grep { $_->{junction_path} eq $path } @{$volumes};

    # Keep deconstructing the path and comparing it against known
    # junction-paths until we find a matching volume. I'm assuming
    # that junction paths are unique
    if (!@matches) {
        $Log->debug("No exact junction_path match found "
                . "for $path, deconstructing");
        my @elements = split(/\//, $path);
        if (@elements) {
            while (@elements) {
                pop(@elements);
                my $current_path = join('/', @elements);

                # If the current path is undef, this means we are down to the
                # two elements / and <dir>. Reset the current path to be /
                # so we match the root volume.
                if (!$current_path) { $current_path = '/'; }
                $Log->debug(
                    "Checking for junction paths matching $current_path");
                @matches = grep { $_->{junction_path} eq $current_path }
                    @{$volumes};
                if (@matches) { last; }
            }
        }
    }

    # Now that we have a volume, create the relative path
    my @elements = split(/\//, $path);
    while (@elements && $elements[0] ne $matches[0]->{volume}) {
        shift(@elements);
    }

    # Shift off the volume
    my $volume = shift(@elements);

    my $relative_path = join('/', @elements);
    if ($relative_path !~ /\/$/) { $relative_path .= '/'; }
    $Log->exit();
    return ($volume, $relative_path);
} ## end sub _get_relative_volume_path

1;
