package ConfigBroker::MySqld;

use strict;
use warnings;

use Cwd qw(abs_path);
use File::Spec::Functions qw(catdir);
use File::Temp qw(tempdir);
use Params::Validate qw(validate SCALAR HASHREF);

#use lib qw("/usr/software/test/share/nate/dbserver/last/lib");
use NATE::Child::Exec   ();
use NATE::Util::Sockets qw(loopback_address);
use NATE::Events        qw(eventloop_drain);
use NATE::Log           qw(log_global);
use Data::Dumper;
use Tharn qw($Log);
use NATE::Util::Sockets qw(get_address);

my @PATHS_TO_SEARCH = ('/usr/local/mariadb', '/usr/local/mysql', '/usr', '/usr/software/mariadb', '/usr/local', '/usr/software');
my $hostname = `hostname`;
chomp $hostname;
my $host = get_address($hostname);
$Data::Dumper::Useqq = 1; 

sub new {
    my $class = shift;
    my %opts = validate(@_, {
        auto_start       => {type=>SCALAR,  default  => 1},
        auto_setup       => {type=>SCALAR,  default  => 1},
        auto_init        => {type=>SCALAR,  default  => 1},
        skip_grant_tables=> {type=>SCALAR,  default  => 1},
        config           => {type=>HASHREF, default  => {}},
        port             => {type=>SCALAR,  optional => 1},
        base_dir         => {type=>SCALAR,  optional => 1},
        mysqld           => {type=>SCALAR,  optional => 1},
        mysql_install_db => {type=>SCALAR,  optional => 1},
        install_path     => {type=>SCALAR,  optional => 1},
        pid              => {type=>SCALAR,  optional => 1},
    });
    my $port = delete($opts{port});

    my $self = \%opts;

    bless $self, $class;
    $self->port($port) if($port);

    # Initialize the MySqld instance
    $self->init() if $self->{auto_init};

    if($self->{auto_start}) {
        $self->discover_start();
    }
    elsif($self->{auto_setup}) {
        $self->setup(); # just setup w/o starting
    }
    return $self;
}

sub discover_start {
    my $self = shift;
    my $port = $self->port(0)||3306;
    my $error;
    for (0..30) {
        eval {
            $port = $self->setup(start_port=>$port);
            $self->start();
        };
        $error = $@;
        return $port if(!$error);
        die $error if($error !~/already in use|no ports available/i);
        $port = $self->port(0); # reset port
        $port ||= 3305; # if this line executes, we were likley booted out of setup() and have no port
        $port++;
    }
    die($error); # must have never found any ports after 31 attempts
}

sub init {
    my $self = shift;
    if ( $self->{base_dir} ) {
        $self->{base_dir} = abs_path($self->{base_dir});
    }
    else {
        $self->{base_dir} = tempdir( CLEANUP => $ENV{NATE_MYSQLD_PRESERVE} ? undef : 1 );
    }

    $self->{config}->{socket}          ||= catdir($self->{base_dir}, 'tmp', 'mysql.sock');
    $self->{config}->{datadir}         ||= catdir($self->{base_dir}, 'var');
    $self->{config}->{'pid-file'}      ||= catdir($self->{base_dir}, "tmp", 'mysql.pid');
    $self->{'host-file'}               ||= catdir($self->{base_dir}, "tmp", 'mysql.host');
    $self->{'port-file'}               ||= catdir($self->{base_dir}, "tmp", 'mysql.port');
    $self->{config}->{max_connections} ||= 10000;


  if (-e $self->{config}->{'pid-file'}) {
        if ($self->{auto_start} || $self->{auto_setup})  {
           ### set the auto_start and auto_setup to 0 so we just return the object with existing info
           $self->{auto_start} = 0;
           $self->{auto_setup} = 0;
        }
        ### If we are not going to re-setup or re-start, just initialize all the rest from existing running instance
        ### and return the object
        open (PID_FD, "<$self->{'config'}->{'pid-file'}");
        my @running_pid = <PID_FD>;
        close PID_FD;
        $self->{'pid'} =  $running_pid[0];
        $Log->debug("Server PID is $self->{'pid'}\n");
        open (HOST_FD, "<$self->{'host-file'}");
        my @running_host = <HOST_FD>;
        close HOST_FD;
        $self->{'host'} =  $running_host[0];
        $Log->debug("Server running on host $self->{'host'}\n");
        $Log->debug("NOTE: This may be different from the current host $hostname; hostip $host.\n");
        open (PORT_FD, "<$self->{'port-file'}");
        my @running_port = <PORT_FD>;
        close PORT_FD;
        $self->{config}->{'port'} =  $running_port[0];
        $Log->debug("Server running on port $self->{config}->{'port'}\n");
  }
  else {
    if(!($self->{mysql_install_db} && $self->{mysqld})) {
        unshift(@PATHS_TO_SEARCH, $self->{install_path}) if($self->{install_path});
        if (!$self->{mysql_install_db}) {
            my $file = _search_file('mysql_install_db-5.3.5', 'libexec', 'bin','scripts');
            if (-e $file) {
                $self->{mysql_install_db} = $file;
            }
            else {
               $file = _search_file('mysql_install_db-5.5.27', 'libexec', 'bin','scripts');
                if (-e $file) {
                    $self->{mysql_install_db} = $file;
                }
                else {
                    my $msg = "Cannot find mysql_install_db for ConfigBroker\n";
                    $msg .= "Please ensure either mysql_install_db-5.3.5 or mysql_install_db-5.5.27";
                    $msg .= " is available in one of the following paths.\n";
                    foreach my $path (@PATHS_TO_SEARCH) {
                       $msg .= "$path/bin\n";
                       $msg .= "$path/libexec\n";
                       $msg .= "$path/scripts\n";
                    }
                    die $msg;
                }
            }
        }

        if (!$self->{mysqld}) {
            my $file = _search_file('mysqld-5.3.5', 'libexec', 'bin');
            if (-e $file) {
                $self->{mysqld} = $file;
            }
            else {
                $file = _search_file('mysqld-5.5.27', 'libexec', 'bin');
                if (-e $file) {
                    $self->{mysqld} = $file;
                }
                else {
                    my $msg = "Can not find mysqld for ConfigBroker\n";
                    $msg .= "Please ensure either mysqld-5.3.5 or mysqld-5.5.27";
                    $msg .= " is available in one of the following paths.\n";
                    foreach my $path (@PATHS_TO_SEARCH) {
                       $msg .= "$path/bin\n";
                       $msg .= "$path/libexec\n";
                    }
                    die $msg;
                }
            }
        }
    }
  }
}

sub port {
    my $self = shift;
    my $port = $self->{config}->{'port'};
    if(@_) {
        ($self->{config}->{'port'}) = @_;
    }
    return $port;
}

sub setup {
    my $self = shift;
    my %opts = (start_port => 3306, @_);

    # create directory structure
    mkdir $self->{base_dir} unless -d $self->{base_dir};
    foreach my $subdir ('etc','tmp', 'var', catdir('var','test'), catdir('var','mysql'), catdir('var','CB')) 
    {
        my $dir = catdir($self->{base_dir}, $subdir);
        mkdir($dir);
        chmod(0777, $dir);
    }
    my $port = $self->{config}->{'port'};
    if(!$port)
    {
        require NATE::Util::Sockets;
        local $@;
        $port = eval{NATE::Util::Sockets::get_available_port(start_port => $opts{start_port});};
        my $err = $@;
        if ((!$port) && $err) {
            # We look for this error message, so change it in discover_start() if it changes
            die "No ports available for starting Mysql server: $err\n";
        }
        $self->{config}->{'port'} = $port;
    }
    # my.cnf
    my $fh;
    open($fh, '>', catdir($self->{base_dir}, 'etc', 'my.cnf')) or die "failed to create file:" . catdir($self->{base_dir}, 'etc', 'my.cnf') . ":$!";
    print $fh "[mysqld]\n";
    print $fh map {
        my $v = $self->{config}->{$_};
        defined $v && length $v
            ? "$_=$v\n"
            : "$_\n";
    } sort keys %{$self->{config}};
    close $fh;
    my $cfgfile = catdir($self->{base_dir}, 'etc', 'my.cnf');

    chmod 0644, catdir($self->{base_dir}, 'etc', 'my.cnf') or warn $!;

    # mysql_install_db
    if (! -d catdir($self->{base_dir}, 'var', 'mysql')) {
        my @cmd = ($self->{mysql_install_db},
                   '--defaults-file="'.catdir($self->{base_dir}, 'etc', 'my.cnf').'"',
                   '--force',);

        if(my $base_install = $self->{install_path}) {
            push(@cmd, '--basedir='.$base_install);
        }
        else{
            my $mysql_base_dir = $self->{mysql_install_db};
            if ($mysql_base_dir =~ s|/[^/]+/mysql_install_db$||) {
                push(@cmd, '--basedir='.$mysql_base_dir);
            }
        }

        my $complete = 0;
        my $log = catdir($self->{base_dir}, 'tmp', 'mysql_install_db.log');
        my $mysql_install_db = NATE::Child::Exec->new(username             => '',
                                                      command              => \@cmd,
                                                      endpoint             => 'local',
                                                      logfile              => $log,
                                                      kill_if_parent_exits => 0,
                                                      inherit_env          => 1,
                                                      on_termination       => [\&_install_db_complete, \$complete, \@cmd],
                                                      #                       511 == 0777  : XXX UMASK_DIR is Not honored!
                                                      env                  => {UMASK => 511, UMASK_DIR => 511,},
                                                      );

        for(0..120) {
            eventloop_drain(timeout=>1);
            return $port if($complete);
        }
        $mysql_install_db->kill(1);
        die(join(' ', @cmd)." did not complete within 120 seconds.\nSee $log for more info\n");
    }
    open (FD, ">$self->{'port-file'}");
    print FD $port;
    close FD;
    return $port;

}

sub start {
    my $self = shift;
    return if defined $self->{pid};

    my @cmd =  ($self->{mysqld},
                '--defaults-file="' . catdir($self->{base_dir}, 'etc', 'my.cnf').'"',
                '--connect_timeout=1200',
                '--wait_timeout=28800',
                '--interactive_timeout=28800',
                '--max_connections=10000',
                '--max_user_connections=10000',
                '--max_allowed_packet=1073741824',
                #'--max_allowed_packet=32M',
                '--user=root',         # Keep this here eventhough it throws a warning for cycl machines
                );
    push(@cmd, '--skip-grant-tables') if($self->{skip_grant_tables}); # skip grant tables MUST Be last
    my $log = catdir($self->{base_dir}, 'tmp', 'mysqld.log');
    my $mysqld = NATE::Child::Exec->new(username             => '',
                                        command              => \@cmd,
                                        endpoint             => 'local',
                                        logfile              => $log,
                                        kill_if_parent_exits => 0,
                                        inherit_env          => 0,
                                        on_termination       => [\&_mysqld_ended],
                                        #                        511 == 0777
                                        env                  => {UMASK => 511, UMASK_DIR => 511,},
                                       );
    while (! -e $self->{config}->{'pid-file'}) {
        eventloop_drain(timeout=>.1);
        if (defined($mysqld->{exit_status})) {
            my $fh;
            open($fh, $log);
            local $\ = undef;
            my $out = <$fh>;
            close $fh;
            die "Failed to launch mysqld\nSee $log for more information:\n$out\n";
        }
    }
    $self->{pid}  = $mysqld->{pid};
    $self->{host} = $host;
    open (FD, ">$self->{'host-file'}");
    print FD $host;
    close FD;

    $self->{mysqld_process} = $mysqld;
}

sub DESTROY_NOT {
    my $self = shift;
    $self->stop if(defined($self->{pid}));
}

sub dsn {
    my ($self, %args) = @_;
    $Log->comment(Dumper(\%args));
    $args{port} ||= $self->{config}->{port}
        if $self->{config}->{port};
    if (defined $args{port}) {
        $args{host} ||= $self->{'host'} || $self->{config}->{'bind-address'} || loopback_address();
    } else {
        $args{mysql_socket} ||= $self->{config}->{socket};
    }
    $args{user}   ||= 'root';
    $args{dbname} ||= 'test';
    return 'DBI:mysql:' . join(';', map { "$_=$args{$_}" } sort keys %args);
}

sub get_host {
    my ($self) = @_;

    my $ret = "";
    open (FD, "<$self->{'host-file'}");
    my @hostinfo = <FD>;
    close FD;
    $ret = $hostinfo[0] if (scalar(@hostinfo) > 0) ;
    return $ret;
}

sub get_pid {
    my ($self) = @_;

    my $ret = "";
    open (FD, "<$self->{config}->{'pid-file'}");
    my @pidinfo = <FD>;
    close FD;
    $ret = $pidinfo[0] if (scalar(@pidinfo) > 0) ;
    return $ret;
}

sub verify_and_cleanup{
    my ($self) = @_;

    my $host = $self->get_host();
    $Log->debug("Current host is $host\n");
    return -1 if ($host eq "");
    my $pid  = $self->get_pid();
    $Log->debug("Current PID is $pid\n");
    return -1 if ($pid eq "");
    my $ps_string = "";
    my $cmd = "rsh $host ps auex | grep $pid | grep sqld";
    my @pslist = `rsh $host ps auex` ;
    @pslist = grep(/$pid/, @pslist); 
    @pslist = grep (/sqld/, @pslist) ;
    if (scalar (@pslist) > 0) {
        $Log->warn("Process $pid still running on host $host.\nNot cleaning up");
        return -1;
    }
    else {
       unlink $self->{config}->{'pid-file'};
       unlink $self->{'host-file'};
       unlink $self->{'port-file'};
    }
    return 0;
} 

sub stop {
    my ($self, $host, $pid) = @_;
    # might remain for example when sending SIGKILL
    if (defined $self->{mysqld_process}) {
       my $mysqld = $self->{mysqld_process};
       $mysqld->kill();
    }
    else {
       my $output = `sudo rsh $host kill -9 $pid`;
    }
    unlink $self->{config}->{'pid-file'};
    unlink $self->{'host-file'};
    unlink $self->{'port-file'};
}

sub _search_file {
    my ($file, @subdirs) = @_;

    # XXX: This is confusing to me
    for my $mysql (map { catdir($_, 'bin', 'mysql') } @PATHS_TO_SEARCH) {
        if (-x $mysql) {
            for my $subdir (@subdirs) {
                my $path = $mysql;
                if ($path =~ s|/bin/mysql$|/$subdir/$file| and -x $path) {
                    return $path;
                }
            }
        }
    }
}

sub _install_db_complete {
    my ($child, $flag, $cmd_array) = @_;
    $$flag = 1;# Yes, this is a pointer so we don't have to create an anonymous sub
    return if(!$child->{exit_status});
    my $error=join(' ', @$cmd_array)."\nexited ".NATE::Child::decode_status_adv($child->{exit_status})."\n";
    die($error);
}

sub _mysqld_ended {
    my $child = shift;
    my $message='mysqld exited '.NATE::Child::decode_status_adv($child->{exit_status})."\n";
    log_global()->debug($message);
}

1;

=head1 NAME

ConfigBroker::MySqld - ConfigBroker Database MySql Server

=head1 SYNOPSIS

 use ConfigBroker::MySqld;
 my $mysqld = new ConfigBroker::MySqld( config => { port => ... }, base_dir => ... );

=head1 DESCRIPTION

This module can be used to create, start, stop a mysqld instance.
It uses C<< NATE::Child::Fork >> to create a new mysqld process.

=head1 METHODS

=head2 new

This is the constructor method that creates a new MySqld object.

=head2 setup

setup() performs basic setup needed for running mysqld.
It creates the directory structure, configuration file,
and also executes mysql_install_db to create all mysql
specific databases.

=head2 start

This method starts the mysqld daemon by exec'ing
a new process and it also creates the 'test' database.

=head2 dsn

This method return the dsn to be used for mysql connection.

=head2 stop

Kills the mysqld daemon by sending SIGINT.

=head1 AUTHOR/MAINTAINER

 NATE Development (dl-nate-dev@netapp.com)

=cut

# Local Variables:
# indent-tabs-mode: nil
# End:
