package ConfigBroker::ResourceGroup::NASGroup;

use strict;
use warnings;
use DBI qw();
use File::Path;
use File::Basename;
use Tharn qw($Log);
use Tharn;
use NATE::Result;
use Data::Dumper;
our @ISA = qw (ConfigBroker::ResourceGroup Exporter);
use ConfigBroker::CbUtils;
use Module::Common qw(validate_mixed);
use Params::Validate
    qw(validate validate_with SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);

my $private_rg_prefix= "_private_";

our $SCHEMA_VERSION = 1.0;


sub new
{
    # we specifically use validate_with and allow_extra to let unknown 
    # "free riding" params be available downstream
    my $pkg = shift;
    my $class = ref($pkg) || $pkg;

    my %opts = validate_with(
                    params => \@_,
                    spec => { 
                        protocols => {type => ARRAYREF,
                                      optional => 1,
                                      default => ["nfs", "cifs"],},
                        timeout   => {type => SCALAR,
                                      optional => 1,
                                      default => -1,},
                    },
                    allow_extra => 1,
                );
    
    $Log->enter();
    my $self = $class->SUPER::new(%opts);
    bless ($self, $class);

    foreach my $key (keys(%opts)) {
        $self->{$key} = $opts{$key};
    }

    $Log->debug("NASGroup structure: " . Dumper($self));

    # check CB for our protos
    # if timeout was specified, die when we hit it
    my $all_fnd = 0;
    my $start_time = time();
    my $delta_time = -1;
    my $warn_counter = 0;
    my (%protos_fnd);
    while (!$all_fnd && $delta_time <= $self->{timeout}) {
        # get all RGs from CB & check for ours
        my @current_groups = &ConfigBroker::CbUtils::get_all_resource_groups($self->{'dbh'});
        %protos_fnd = ();
        foreach my $proto (@{$self->{protocols}}) {
            $protos_fnd{$proto} = 0;
            foreach my $thisgroup (@current_groups) {
                if ($thisgroup eq ($proto . "_activity")) {
                    $protos_fnd{$proto} = 1;
                    last;
                }
            }
        }
        # assume we found all unless a proto says otherwise
        $all_fnd = 1;
        foreach my $proto (keys(%protos_fnd)) {
            if (!$protos_fnd{$proto}) {
                $all_fnd = 0;
                last;
            }
        }
        
        # sleep a while before trying again
        Tharn::sleep(10);

        $delta_time = time() - $start_time;
        
        # warn every 5 minutes
        if (int($delta_time / 30) > $warn_counter) {
            my $msg = ($self->{timeout} == -1) ? "infinite" : "$self->{timeout} secs";
            $Log->warn("ConfigBroker NASGroup does not yet contain the requested protocol activities:\n" . Dumper(\%protos_fnd) . "(timeout = $msg)");
            $warn_counter++;
        }
        
        # clear timer for infinite behavior if that was the caller request
        $delta_time = -1 if ($self->{timeout} == -1);
    }

    if (!$all_fnd) {
        die config_result("ConfigBroker NASGroup did not find the requested "
                        . "protocol activities within the $self->{timeout} "
                        . "timeout:\n" . Dumper(\%protos_fnd));
    } else {
        $Log->debug("ConfigBroker NASGroup found all requested protocols in ConfigBroker: " . join(',', @{$self->{protocols}}));
    }

    $Log->exit();
    return $self;
}


# return an array ref of hashrefs of all nas activity data for the specfied protocol
# caller can use same protocols as provided in new or specify the protocol(s), multiple protocols are treated with
# *OR* logic (activity performed by CIFS *OR* NFS)
# caller can provide regex filter for the CB activity table  
#
# example: ensure CB contains NFS & CIFS traffic, ask for all CIFS activity w/o a filter
# my $cb = ConfigBroker->new();
# my $NASRG = $cb->get_nasgroup_object(timeout => 300);
# my $active = $NASRG->activity(protocol => 'cifs');
sub all_activity {
    my ($self, %opts) =
        validate_mixed(@_,
                [{type => OBJECT},],
                { protocols       => {type => ARRAYREF,
                                      optional => 1,},
                  activity_filter => { type => HASHREF,
                                       optional => 1,
                                       default => {},},
                },
        );

    # use protocols if provided by caller, otherwise use those that established in new
    my @protos = defined($opts{protocols}) ? @{$opts{protocols}} : @{$self->{protocols}};

    $Log->debug(  "requested protocols:\n" . Dumper(@protos)
                . "activity_filter:\n" . Dumper($opts{activity_filter})
               );

    # Pull activity data from the indicated proto activity tables
    my @rows;
    foreach my $p (@protos) {
        my $rgname = ($p =~ /nfs/i) ? "nfs_activity" : "cifs_activity" ;
        my $rg = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $rgname);
                
        my $filter = $opts{activity_filter};
        my @rgs = @{$rg->get_resources(
                 type          => $rgname,
                 regex_filters => $filter,)};

        push (@rows, @rgs);
    }

    $Log->exit("returning " . scalar(@rows) . " activity entries");
    return \@rows;
}


# return an array ref of hashrefs of nas activity data that are currently ongoing
# caller can use same protocols as provided in new or specify the protocol(s), multiple protocols are treated with
# *OR* logic (activity performed by CIFS *OR* NFS)
# caller can provide regex filter for the CB activity table  
#
# example: ensure CB contains NFS traffic, ask for current NFS activity from workitems with name dt*
# my $cb = ConfigBroker->new();
# my $NASRG = $cb->get_nasgroup_object(protocols => ['nfs'], timeout => 300);
# my $act_filter = {workitem => "dt*"};
# my $active = $NASRG->current_activity(protocol => 'nfs', activity_filter => $act_filter);
#
sub current_activity {
    my ($self, %opts) =
        validate_mixed(@_,
                [{type => OBJECT},],
                { protocols       => {type => ARRAYREF,
                                      optional => 1,},
                  activity_filter => { type => HASHREF,
                                       optional => 1,
                                       default => {},},
                },
        );
    $Log->enter();

    # use protocols if provided by caller, otherwise use those that established in new
    my @protos = defined($opts{protocols}) ? @{$opts{protocols}} : @{$self->{protocols}};

    $Log->debug(  "requested protocols:\n" . Dumper(@protos)
                . "activity_filter:\n" . Dumper($opts{activity_filter})
               );

    # add stoptime to filter to get only active jobs
    my $filter = $opts{activity_filter};
    $filter->{stoptime} = '-';
    my $rgs = $self->all_activity(
                protocols     => \@protos,
                regex_filters => $filter,);

    $Log->exit("returning " . scalar(@$rgs) . " current activity entries");
    return $rgs;
}

# return an array ref of hashrefs of volumes that are currently experiencing IO
# caller can use same protocols as provided in new or specify the protocol(s), multiple protocols are treated with
# *OR* logic (volumes receiving IO via NFS *OR* CIFS)
# caller can provide regex filters for the CB activity table and/or the CB volume table 
#
# example: ensure CB contains CIFS & NFS traffic, ask for volumes with security style mixed or ntfs on testaggr_2
# that are receiving CIFS IO from tools matching *sdstress*
# my $cb = ConfigBroker->new();
# my $NASRG = $cb->get_nasgroup_object(protocols => ['nfs', 'cifs'], timeout => 300);
# my $act_filter = {workitem => "*sdstress*"};
# my $vol_filter = {aggregate => "testaggr_2", security_style => "mixed|ntfs" };
# my $active = $NASRG->current_active_volumes(protocols => ['cifs'], activity_filter => $act_filter, volume_filter => $vol_filter);
#
sub current_active_volumes {
    my ($self, %opts) =
        validate_mixed(@_,
                [{type => OBJECT},],
                { protocols       => { type => ARRAYREF,
                                       optional => 1, },
                  activity_filter => { type => HASHREF,
                                       optional => 1,
                                       default => {}, },
                  volume_filter   => { type => HASHREF,
                                       optional => 1,
                                       default => {}, },
                },
        );
    $Log->enter();
    
    # use protocols if provided by caller, otherwise use those that established in new
    my @protos = defined($opts{protocols}) ? @{$opts{protocols}} : @{$self->{protocols}};

    $Log->debug(  "requested protocols:\n" . Dumper(@protos)
                . "activity_filter:\n" . Dumper($opts{activity_filter})
                . "volume_filter:\n" . Dumper($opts{volume_filter}) 
               );

    # Pull activity data from the indicated proto activity tables
    my @raw_path_rows;
    foreach my $p (@protos) {
        my $rgname = ($p =~ /nfs/i) ? "nfs_activity" : "cifs_activity" ;
        my $rg = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $rgname);

        my $filter = $opts{activity_filter};
        $filter->{stoptime} = '-';
        my @rgs = @{$rg->get_resources(
                 type          => $rgname,
                 regex_filters => $filter,)};

        push (@raw_path_rows, @rgs);
    }

    # Make a new set without any duplicate vserver-volume pairs
    my @unique_path_rows;
    my %paths;
    foreach my $row (@raw_path_rows) {
        if (not $paths{$row->{volume}}{$row->{vserver}}) {
            $paths{$row->{volume}}{$row->{vserver}} = 1;
            push (@unique_path_rows, $row);
        }
    }
    $Log->debug("unique_path_rows:\n" . Dumper(\@unique_path_rows));

    # Pull the true volume rows from CB. This is more valuable to the caller
    # than the <proto>_activity rows. Use the caller's filter to narrow down
    # the query results
    my $filter = $opts{volume_filter};
    my @return = $self->_get_true_rows('volume', \@unique_path_rows, $filter);

    $Log->exit("found " . scalar(@return) . " volumes");
    return \@return;

}

# return an array ref of hashrefs of shares that are currently experiencing IO
# only CIFS IO is considered (naturally)
# caller can provide regex filters for the CB activity table and/or the CB share table 
#
# example: ensure CB contains CIFS traffic, ask for shares on testaggr_2 with traffic
# my $cb = ConfigBroker->new();
# my $NASRG = $cb->get_nasgroup_object(protocols => ['cifs']);
# my $act_filter = {workitem => "*sdstress*"};
# my $sha_filter = {aggregate => "testaggr_2", security_style => "mixed|ntfs" };
# my $active = $NASRG->current_active_volumes(activity_filter => $act_filter, share_filter => $sha_filter);
#
sub current_active_shares {
    my ($self, %opts) =
        validate_mixed(@_,
                [{type => OBJECT},],
                { protocols       => { type => ARRAYREF,
                                       optional => 1, },
                  activity_filter => { type => HASHREF,
                                       optional => 1,
                                       default => {}, },
                  share_filter    => { type => HASHREF,
                                       optional => 1,
                                       default => {}, },
                },
        );
    $Log->enter();

    # set protocols to cifs because nothing else makes sense 
    my @protos = ("cifs");

    $Log->debug(  "activity_filter:\n" . Dumper($opts{activity_filter})
                . "share_filter:\n" . Dumper($opts{share_filter})
               );

    # Pull activity data from the cifs_activity table b/c nothing else makes sense
    my @raw_path_rows;
    my $rgname = "cifs_activity";
    my $rg = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $rgname);

    my $filter = $opts{activity_filter};
    $filter->{stoptime} = '-';
    my @rgs = @{$rg->get_resources(
                type          => $rgname,
                regex_filters => $filter,)};

    push (@raw_path_rows, @rgs);

    # Make a new set without any duplicate vserver-share pairs
    my @unique_path_rows;
    my %paths;
    foreach my $row (@raw_path_rows) {
        my $share_extract = (split(/:/, $row->{resourcename}))[1];
        if (not $paths{$share_extract}{$row->{vserver}}) {
            $paths{$share_extract}{$row->{vserver}} = 1;
            # stack the extracted share onto the row data to make it available in get_true_rows
            # no harm b/c this cifs_activity data is only intermediate (not returned to caller)
            $row->{vservercifsshare} = $share_extract;
            push (@unique_path_rows, $row);
        }
    }
    $Log->debug("unique_path_rows:\n" . Dumper(\@unique_path_rows));

    # Pull the true share rows from CB. This is more valuable to the caller
    # than the <proto>_activity rows. Use the caller's filter to narrow down
    # the query results
    $filter = $opts{share_filter};
    my @return = $self->_get_true_rows('vservercifsshare', \@unique_path_rows, $filter);

    $Log->exit("found " . scalar(@return) . "shares");
    return \@return;

}


# return an array ref of hashrefs of interfaces that are currently experiencing IO
# caller can use same protocols as setup in new or specify the protocol(s)
# caller can provide regex filters for the CB activity table and/or the CB interface table 
#
# example: ensure CB contains CIFS & NFS traffic, ask for interfaces on e1* with CIFS sdstress traffic
# my $cb = ConfigBroker->new();
# my $NASRG = $cb->get_nasgroup_object(protocols => ['nfs', 'cifs']);
# my $act_filter = {workitem => "*sdstress*"};
# my $int_filter = {home_port => "e1" };
# my $active = $NASRG->current_active_volumes(protocols => ['cifs'], activity_filter => $act_filter, interface_filter => $int_filter);
#
sub current_active_interfaces {
    my ($self, %opts) =
        validate_mixed(@_,
                [{type => OBJECT},],
                { protocols        => { type => ARRAYREF,
                                        optional => 1, },
                  activity_filter  => { type => HASHREF,
                                        optional => 1,
                                        default => {}, },
                  interface_filter => { type => HASHREF,
                                        optional => 1,
                                        default => {}, },
                },
        );
    $Log->enter();

    # use protocols if provided by caller, otherwise use those that established in new
    my @protos = defined($opts{protocols}) ? @{$opts{protocols}} : @{$self->{protocols}};

    $Log->debug(  "requested protocols:\n" . Dumper(@protos)
                . "activity_filter:\n" . Dumper($opts{activity_filter})
                . "interface_filter:\n" . Dumper($opts{interface_filter})
               );

    # Pull activity data from the indicated proto activity tables
    my @raw_path_rows;
    foreach my $p (@protos) {
        my $rgname = ($p =~ /nfs/i) ? "nfs_activity" : "cifs_activity" ;
        my $rg = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $rgname);

        my $filter = $opts{activity_filter};
        $filter->{stoptime} = '-';
        my @rgs = @{$rg->get_resources(
                 type          => $rgname,
                 regex_filters => $filter,)};
    
        push (@raw_path_rows, @rgs);
    }
                
    # Make a new set without any duplicate dest_ip
    my @unique_path_rows;
    my %paths;
    foreach my $row (@raw_path_rows) {
        if (not $paths{$row->{destip}}) {
            $paths{$row->{destip}} = 1;
            # stack the dest_ip onto the row data as 'address' to make it available in get_true_rows
            # no harm b/c this cifs_activity data is only intermediate (not returned to caller)
            $row->{address} = $row->{destip};
            push (@unique_path_rows, $row);
        }
    }   
    $Log->debug("unique_path_rows:\n" . Dumper(\@unique_path_rows));
        
    # Pull the true interface rows from CB. This is more valuable to the caller
    # than the <proto>_activity rows. Use the caller's filter to narrow down
    # the query results
    my $filter = $opts{interface_filter};
    my @return = $self->_get_true_rows('networkinterface', \@unique_path_rows, $filter);

    $Log->exit("found " . scalar(@return) . " interfaces");
    return \@return;
    
}   

 
sub _get_true_rows {
    my ($self, $type, $activity_rows, $input_filter) = @_;    

    $Log->enter("type $type");
  
    # lookup each activity row in the type table, build array to return 
    my @return;
    my $clust_rg = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, 'cluster');
    foreach my $row (@{$activity_rows}) {
        # use the caller's filter with our vserver & $type name; therefore caller cannot ask 
        # "what are the volumes (or shares) on vserverX that have NFS traffic right now".
        # Allowing the user to specify vserver produces unpredictable & confusing behavior.
        # Possible to revisit when foreign keys are implemented in CB
        my $filter;
        $filter->{vserver} = $row->{vserver};
        if ($type eq 'networkinterface') {
            $filter->{address} = $row->{address};
        } else {
            $filter->{name} = $row->{$type};
        }
        
        my @found = @{$clust_rg->get_resources(
                 type          => $type,
                 regex_filters => $input_filter,
                 match_filters => $filter,)};
        
        $Log->debug("found $type match:\n" . Dumper(@found));
        push (@return, @found);
    }
    $Log->debug("returning true $type" . "s: \n" . Dumper(\@return));

    $Log->exit("found " . scalar(@return) . " $type" . "s");
    return @return;
}

sub remove_workload_from_group {
    my ($self, %opts) =
        validate_mixed(@_,
                [{type => OBJECT},],
                { group 	   => { type => SCALAR,
                                        optional => 0, },
                  workload 	   => { type => SCALAR,
                                        optional => 0, },
                  protocol         => { type => SCALAR,
                                        optional => 0, },
                },
        );
    $Log->enter();

    my $rg = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $opts{group});
 
    my $workload_type = $opts{protocol} . "_workloads" ;
    my $workload = $rg->get_resources(type => $workload_type , match_filters => {name => $opts{workload}});
    if (defined $workload) {
      $rg->delete_resources(type => $workload_type, match_filters => {name => $opts{workload}});
    }
    else {
       die "Did not find any workloads matching $opts{workload} in group $opts{group} for protocol $opts{protocol}\n";
    }
    $Log->exit();
}

sub add_workload_to_group {
    my ($self, %opts) =
        validate_mixed(@_,
                [{type => OBJECT},],
                { group            => { type => SCALAR,
                                        optional => 0, },
                  workload         => { type => SCALAR,
                                        optional => 0, },
                  protocol         => { type => SCALAR,
                                        optional => 0, },
                },
        );
    $Log->enter();

    my $rg = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $opts{group});

    my $workload_type = $opts{protocol} . "_workloads" ;
    my $workload = { 
      'name' => $opts{workload},
      'workload' => $opts{workload},
    } ;
    my @workloads;
    push (@workloads, $workload);
    $rg->add_resources(resource_type => $workload_type, resourcearrayref => \@workloads);
    $Log->exit();
}

# marr: this should be dropped, only user as of 8/23/12 is                       
# NASActivityCounter.pm demo thpl
#
# get all the rows from all the activity tables,
# no matter what the progress
sub get_volume_activity {
    my ($self) = @_;
    $Log->enter();

    my @protos = @{$self->{'protocols'}};

    # Pull activity data from two different tables
    my @return;
    foreach my $p (@protos) {
        my $rgname = ($p =~ /nfs/i) ? "nfs_activity" : "cifs_activity" ;
        $Log->debug("Get resource group object for group $rgname");
        my $rg = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $rgname);
        if (not $rg) {
            $Log->warn("Failed to retrieve resourcegroup  $rgname!");
            next;
        }

        my $rows;
        $Log->debug("Get entries in resourcegroup $rgname of the same type");
        $rows = $rg->get_resources(type => $rgname);
        if (not $rows) {
            $Log->comment($@);
            next;
         }

        foreach my $row (@{$rows}) {
           # Pass on the entry row inserted by the TOPO lib
           next if ($row->{volume} =~ /TheBeginning/i);
           push (@return, $row);
        }

    }
    $Log->exit("found " . scalar(@return) . " volumes");
    return \@return;
}


sub remove_io_client {
    my ($self, %opts) =
        validate_mixed(@_,
                [{type => OBJECT},],
                  { 
                  client 	   => { type => ARRAYREF,
                                        optional => 0, },
                  protocol         => { type => ARRAYREF,
                                        optional => 1, },
                   group            => { type => SCALAR,
                                        optional => 1, },
                },
        );                             
    $Log->enter();

    # Set the defaults if no protocol is passed in
    $opts{protocol} = exists $opts{protocol} ? $opts{protocol} : $self->{protocols};
    
    foreach my $protocol (@{$opts{protocol}}) {
    
       my $client_proto_table = "${protocol}_clients";
       my $client_type = $client_proto_table;

       foreach my $client ( @{$opts{client}} ) {

          # Get the records that contain the client wanted
          my $resource_group_aref = &ConfigBroker::CbUtils::get_resourcegroups_for_type($self->{'dbh'}, $client_proto_table, $client);
          
          if ($resource_group_aref) {
          
             # Get all the records and respective resource group
             foreach my $client_record ( @{$resource_group_aref} ) {

               my $rg = $client_record->{resourcegroup};

               # Remove only specified rg client if rg is specified 
               if ($opts{group} && $rg ne $opts{group} ) {
                  next;
               }
               # Don't remove the private resource group with the client information necessary 
               # to restore a client 
               elsif ($rg =~/^$private_rg_prefix/) {
                  next;
               }

               my $client_group = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $rg);
               my $client_record = $client_group->get_resources(type => $client_type, match_filters => {name=>$client});

               $Log->comment( qq[Removing client '$client' from resource group $rg] );
               $client_group->delete_resources(type => $client_type, match_filters => {name=>$client});

               # Add the client to a private resource group to be reused when a client is 
               # to be added back for IO
               my $private_group = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, "$private_rg_prefix$rg");
               $private_group->add_resources(resource_type =>$client_type, resourcearrayref => $client_record);
            }
          }
       }
    }

    $Log->exit();
}


sub restore_io_client {
    my ($self, %opts) =
        validate_mixed(@_,
                [{type => OBJECT},],
                  { 
                  client 	   => { type => ARRAYREF,
                                        optional => 0, },
                  protocol         => { type => ARRAYREF,
                                        optional => 1, },
                   group            => { type => SCALAR,
                                        optional => 1, },
                },
        );
    $Log->enter();

    # Set the defaults if no protocol is passed in
    $opts{protocol} = exists $opts{protocol} ? $opts{protocol} : [qw(nfs cifs san)]; 

    foreach my $protocol (@{$opts{protocol}}) {
    
       my $client_proto_table = "${protocol}_clients";
       my $client_type = "${protocol}_clients";

       foreach my $client ( @{$opts{client}} ) {

          # Get the records that contain the client wanted
          my $resource_group_aref = &ConfigBroker::CbUtils::get_resourcegroups_for_type($self->{'dbh'}, $client_proto_table, $client);
          
          if ($resource_group_aref) {
          
             # Get all the records and respective resource group
             foreach my $client_record ( @{$resource_group_aref} ) {

               my $rg = $client_record->{resourcegroup};

               # Restore only the rg client if an rg is specified 
               if ($opts{group} && $rg ne "$private_rg_prefix$opts{group}" ) {
                  next;
               }
               # Exclude records that clients weren't removed from  
               elsif($rg !~/^$private_rg_prefix/) {
                  next;
               }

               # Get the original client record
               my $private_group = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $rg);
               my $client_ahref = $private_group->get_resources(type => $client_type, match_filters => {name=>$client});

               $rg =~ s/^$private_rg_prefix//;

               my $client_group = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $rg);

               $Log->comment(qq[Adding client $client for $protocol IO to resource group ($rg).]);
               $client_group->add_resources(resource_type => $client_type, resourcearrayref => $client_ahref);
 
               # Delete private rg  
               $private_group->delete_resources(type => $client_type, match_filters => {name=>$client});
            }
          }
       }
    }

    $Log->exit();
}

sub client_io_running {
    my ($self, %opts) =
        validate_mixed(@_,
                [{type  => OBJECT},],
               {
                   client 	   => { type => ARRAYREF, optional => 0, },
                   protocol   => { type => ARRAYREF, optional => 1, },
                   timeout =>  { type => SCALAR, optional => 1, },
                   delay =>  { type => SCALAR, optional => 1, }
               }
        );

    $Log->enter();
    
    # use protocols if provided by caller, otherwise use those that established in new
    my @protos = defined($opts{protocols}) ? @{$opts{protocols}} : @{$self->{protocols}};

    $Log->debug("Ckeck if IO is running giving the following:\n". Dumper \%opts);

    my $still_running = 0;
    my $time_at_beginning = time;

    $opts{timeout} ||=1;

    while (time - $time_at_beginning < $opts{timeout} ) {   

       # Pull activity data from the indicated proto activity tables
       my @stopped_activity_data;
       foreach my $protocol (@protos) {
           my $rgname = "${protocol}_activity";
           my $rg = &ConfigBroker::CbUtils::get_resource_group_object($self->{'dbh'}, $rgname);

           my $filter = $opts{activity_filter};
           $filter->{stoptime} = '-';
           my $resource_ahref = $rg->get_resources(
                        type          => $rgname,
                        regex_filters => {stoptime => "-"}
                     );

           push (@stopped_activity_data, @{$resource_ahref});
       }

       my %stopped_client;
       foreach my $row_href (@stopped_activity_data) {
          $stopped_client{$row_href->{client}}++;
       }

       $still_running = 0;
       foreach my $client ( @{$opts{client}} ) {

           if ( !exists $stopped_client{$client} ) {
               $still_running = 1;
           }   
       }

       my $msg = "IO is still running on at least one of the specified clients."; 
       if (!$still_running) {
         $msg = "IO has stopped running on all the specified clients.";
         $Log->comment($msg);
         last;
       }

       $Log->debug($msg);

       # Sleep delayed time if specified
       if ($opts{timeout}) {
           # sleep 1/10 the maximum timeout or 1 sec if no delay is specifed
           $opts{delay} = $opts{delay} || int ($opts{timeout}/10) || 1;
                       
           sleep ($opts{delay});
       }
    }

    return $still_running;
}


1;

#### re-define some of the base class functions to empty functions because 
#### this custom group is not a traditional RG but a combination of a few.



=head2 get_active_volumes_

   Object method; Get a list of all the active volumes.

=head3 Usage

    $cb = new ConfigBroker;
    $NASRG = $cb->get_nasgroup_object();
    $NASRG->get_active_volumes(%opts);
    $opts{'protocols'} - OPTIONAL protocol to use.  Can be "nfs" or "cifs" ;

=head3 Return

    An array of active volumes.

=head2 most_used_volume

   Object method; Get the most used volume as a hashref.

=head3 Usage

    $cb = new ConfigBroker;
    $NASRG = $cb->get_nasgroup_object();
    $mostused = $NASRG->most_used_volume();

=head3 Return

   Returns the most used volume as a hashref.

=head2 least_used_volume

   Object method; Get the least used volume as a hashref.

=head3 Usage

    $cb = new ConfigBroker;
    $NASRG = $cb->get_nasgroup_object();
    $leastused = $NASRG->least_used_volume();

=head3 Return

   Returns the least used volume as a hashref.


=head2 restore_io_client

   restore an io client in order to resume IO

=head3 Usage

    $cb = new ConfigBroker;
    $NASRG = $cb->get_nasgroup_object();
    $NASRG->restore_io_client(%opt);

    where %opts:
    $opts{'client'} - array reference of the list of clients to add for the next workload 
    $opts{'protocols'} - OPTIONAL the list of protocols to use: "san", "nfs" or "cifs", default is the default user 
                         protocols (set at object creation)
    $opts{'group'} - OPTIONAL the specific client group from which to remove client from IO activity 

=head3 Return

   Returns the most used volume as a hashref.


=head2 remove_io_client

   remove an IO client in order to stop IO

=head3 Usage

    $cb = new ConfigBroker;
    $NASRG = $cb->get_nasgroup_object();
    $NASRG->remove_io_client(%opts);

    where %opts:
    $opts{'client'} - array reference of the list of clients to remove from the next workload 
    $opts{'protocols'} - OPTIONAL the list of protocols to use: "san", "nfs" or "cifs" , default is all the protocols
    $opts{'group'} - OPTIONAL the specific client group from which to remove client from IO activity 

=head3 Return

 
=head2 client_io_running

   Check if a list of client is running IO

=head3 Usage

    $cb = new ConfigBroker;
    $NASRG = $cb->get_nasgroup_object();
    $NASRG->client_io_running(%opts);

    where %opts:
    $opts{'client'} - array reference of the list of clients to add for the next workload 
    $opts{'protocols'} - OPTIONAL the list of protocols to use: "san", "nfs" or "cifs" 
    $opts{'timeout'}  - OPTIONAL the maximum time to wait for of client IO to stop running"
    $opts{'delay'}  - OPTIONAL the delay time in seconds between querries (or frequency) for checking if IO is stopped"

=head3 Return

   Returns 0 or 1: 
      0 if all IO on the specified clients have stopped running
      1 if IO on at least one specified client is still running
  
