## @summary ConfigBroker::ResourceGroup::SFOGroup
## @author robbieh@netapp.com
## @status public
## @pod here

=head1 NAME

ConfigBroker::ResourceGroup::SFOGroup  

=head1 SYNOPSIS

 use ConfigBroker;

 my $cb= new ConfigBroker(); 

 my $sfo_grp = $cb->SFOGroup(%opt); 
 # Note: %opt (optional) can be to used set the maximum number of seconds 
 # to wait before the resource group exists (the default is 240). Example: wait=>300  

 my $partner_status = $sfo_grp->partner_status();


=head1 DESCRIPTION

Class module for SFO resource group 

<B>Note<B>: This package is intended to be used in conjunction with ConfigBorker, the ConfigBroker 
object will be used to get resoucrce group object 

=cut

package ConfigBroker::ResourceGroup::SFOGroup;


use strict;
use Carp;
use warnings;
use DBI qw();
use File::Path;
use File::Basename;
use Tharn qw($Log);
use Tharn;
use NATE::Result;
use Data::Dumper;
our @ISA = qw (ConfigBroker::ResourceGroup Exporter);
use ConfigBroker::CbUtils;
use Module::Common qw(validate_mixed);
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use ConfigBroker;
use NATE::ParamSet;

our @EXPORT  = qw($SFOResourceGroup);

our $var = NATE::ParamSet->new(global=>1);
our $AUTOLOAD;
our $SFOResourceGroup = 'ConfigBroker::ResourceGroup::SFOGroup';
our $SCHEMA_VERSION = 1.0;

my $resource_groupname="sfo_activity";

use constant MAX_RG_WAIT_TIME => 240;

my %sfo_autoload = ( 
   status => undef,
);


sub new {
   $Log->enter();

   my ($pkg) = shift;
   my %opts = @_;
   my $class = ref($pkg) || $pkg;

   $opts{groupname} ||= $resource_groupname;
   $opts{name} = $opts{groupname}; 
  
   $Log->debug( "Getting resource group object '$opts{groupname}'\n".  Dumper(\%opts) );

   if (!ConfigBroker::CbUtils::verify_table_exists($opts{dbh}, $opts{groupname}) ) {
       $Log->debug("The table for $opts{name} does not exist.");
       return; 
   }

   my $self = $class->SUPER::new(%opts);
   foreach my $key (keys %opts) {
      $self->{$key} = $opts{$key};
   }

   $self->{rg_obj} = $self->_wait_for_rg();
        
   bless ($self, $class); 

   $Log->exit();
   return $self;
}

sub AUTOLOAD {
     my $self = shift;
     my %opts = @_;
     my $type = ref($self) or croak "$self is not an object";

     my $name = $AUTOLOAD;
     $name =~ s/.*://;   # strip fully-qualified portion

     $Log->debug("Get SFO activity data via : $name");

     my $rg_sources_ahref = $self->{rg_obj}->get_resources(type=>$self->{name});

     $name =~ s/^get_//;

     # If resource group exists ensure that the group has data to access 
     if ( @$rg_sources_ahref && !(exists $sfo_autoload{$name} || exists $rg_sources_ahref->[0]->{$name}) ) {

        # Get all the public allowed fields
        my $allowed = join "\n", ( keys %{$rg_sources_ahref->[0]}, keys %sfo_autoload);
        die config_result(qq{Accessor method "$name" is not one of the allowed methods for "$rg_sources_ahref->[0]->{resourcegroup}: Allowable calls:\n$allowed\n"});
     }
    
     my ($sfo_record, %all_record);
     my %timed_record;

     # Get the last record for each node
     foreach my $record (reverse @$rg_sources_ahref) {

        $timed_record{$record->{time_updated}} = $record;

        # Get the most recent record
        if (exists $opts{node} && $record->{node} eq $opts{node} ) {

           $sfo_record = $record;
           last;
        }
        elsif (!$opts{node} && !exists $all_record{$record->{node}} ) {

             $all_record{$record->{node}} = $record;
        }
     }

     if ( $opts{node} ) {

        # Return the record with all the status field for the particular node 
        if ($name eq "status") {

           $Log->comment("Return the status of the last record for node $opts{node}");
           return $sfo_record;
        }
        else {

            $Log->comment(qq[Return $name value for SFO node "$opts{node}".]) ;
            return $sfo_record->{$name}; 
        }
     }
     else {

        if ( $opts{all} == 1  ) {

           $Log->comment("Return the last record for all SFO nodes") ;
           return \%all_record;
        }
        else {

           #Sort in descending order to get the time for the last updated record
           my ($last_currenttime) = sort{ $b <=> $a} keys %timed_record;

           # get the node data from the latest record
           $sfo_record = $timed_record{$last_currenttime};
                      
           if (exists $sfo_record->{$name}) {
              $Log->comment(qq[Return $name value for SFO node "$sfo_record->{node}".]) ;
              return $sfo_record->{$name};
           }
           else {
              $Log->comment(qq[Return the last record for SFO node $sfo_record->{node}".]) ;
              return $sfo_record;
           }
        }
     }
}


sub _wait_for_rg {
   my $self = shift; 
   my $wait_time = $self->{"wait"} || MAX_RG_WAIT_TIME;   

   $Log->comment("Get '$self->{groupname}' resource group object or wait a maximum of $wait_time seconds for resource group to exist.");  

   my $rg;
   while (!$rg && $wait_time >= 0) {
      my @groups = &ConfigBroker::CbUtils::get_all_resource_groups($self->{'dbh'});
      foreach my $cb_rg (@groups) { 
       if ($cb_rg eq $self->{'groupname'}) {                      
          $rg  = ConfigBroker::CbUtils::get_resource_group_object($self->{dbh}, $self->{groupname}, $self->{dsn});
          last;
       }
      }
      sleep (10);
      $wait_time -= 10;
   }


   if (!defined $rg) {
      die config_result ("After $wait_time seconds, the SFO resource group object does not exist!");
   }
   

   return $rg;
}

# Explicity added so that AUTOLOAD doesn't catch it
sub DESTROY {}

return 1;

=head1 METHODS

The following are the methods to use to retrieve SFO resource group properties: 
The methods are similar to the SFO resource group properties: 

=over

=item * 

Methods/Properties:

   - node
   - node_operation
   - node_status
   - node_status_raw
   - next_operation 
   - partner
   - partner_status
   - partner_status_raw    
   - time_until_autogiveback
   - time_updated
   - schedule 
   - auto_takeover_schedhule
   - refresh_rate

=item * 

Usage:

   $objet-><property>(%options);
   Where %options take one of the following: 
    - node=>$node   (optional)

=back

=over

=item *

Examples:

   # Get the SFO operation that the specified node has executed 
   my $operation = $sfo_grp->node_operation(node=>$node);


   # Get the status of partner for the specified node  
   my $status = $sfo_grp->partner_status(node=>$node);

   # Get the status of partner for the specified node  
   my $status = $sfo_grp->partner_status(node=>$node);

   # NOTE: If a node is not specified, then the property/data will be retrived
   # from the last updated updated node record. This is is useful when the nodes are randomly chosen 

   # Get the status of partner for the specified last updated node  
   my $status = $sfo_grp->partner_status();

   # Get the last updated node  
   my $last_update_node = $sfo_grp->node();

=back 

=head2 get_status

   $sfo_grp->get_status(%opts); 
   %opts:
      node=>$node (optional)
      all=>1 (optional)
        
=over

=item * 

Usage

1. $sfo_grp->get_status(node=>$node); 
Returns a hash reference of the SFO resource group (last updated) record for the specified node:

Example:

C<< my $node_record_href = $sfo_grp->get_status(node=>'nst-vsa-163') >>;

Sample record: 

   $node_record_href =  {
      'node' => 'nst-vsa-163',
      'node_operation' => 'GIVEBACK',
      'node_status' => 'UP',
      'node_status_raw' => 'In takeover',
      'partner' => 'nst-vsa-162'           
      'partner_status' => "'Waiting_for_cluster_applications"',
      'partner_status_raw' => '"Waiting for cluster applications to come online on the local node"',
      'time_until_autogiveback' => '',
      'is_random_node_' => '1',
      'refresh_rate' => '30',
      'time_updated' => '1350939983',
      'schedule' => '1350940221',
      'next_operation' => 'TAKEOVER',
      'resourcegroup' => 'sfo_activity',
   }

2. C<< $sfo_grp->get_status(); >> 

Returns a hash reference of the SFO resource group record for the node that was
B<last> updated:
This is is useful when the nodes are randomly chosen,  

Example:

C<< my $node_record_href = $sfo_grp->get_status() >>;

Sample record: 

   $node_record_href = {
      'node' => 'nst-vsa-163',
      'node_operation' => 'GIVEBACK',
      'node_status' => 'UP',
      'node_status_raw' => 'In takeover',
      'partner' => 'nst-vsa-162'           
      'partner_status' => "'Waiting_for_cluster_applications"',
      'partner_status_raw' => '"Waiting for cluster applications to come online on the local node"',
      'time_until_autogiveback' => '',
      'is_random_node_' => '1',
      'refresh_rate' => '30',
      'time_updated' => '1350939983',
      'schedule' => '1350940221',
      'next_operation' => 'TAKEOVER',
      'resourcegroup' => 'sfo_activity
   }

3. C<< $sfo_grp->get_status(all=>1); >> 

Returns a hash reference of the last updated record for all the nodes subjected to SFO activity:

Example:

C<< my $nodes_record_href = $sfo_grp->get_status(all=>1) >>;

Sample record: 

   $nodes_record_href = {

     'nst-vsa-162' => {
         'node' => 'nst-vsa-162',
         'node_operation' => 'GIVEBACK',
         'node_status' => 'UP',
         'node_status_raw' => 'In takeover',
         'partner' => 'nst-vsa-163'           
         'partner_status' => "'Waiting_for_cluster_applications"',
         'partner_status_raw' => '"Waiting for cluster applications to come online on the local node"',
         'time_until_autogiveback' => '',
         'is_random_node_' => '1',
         'refresh_rate' => '30',
         'time_updated' => '1350939983',
         'schedule' => '1350940221',
         'next_operation' => 'TAKEOVER',
         'resourcegroup' => 'sfo_activity',
      },

      'nst-vsa-164' => {
         'node' => 'nst-vsa-164',
         'node_operation' => 'GIVEBACK',
         'node_status' => 'UP',
         'node_status_raw' => 'In takeover',
         'partner' => 'nst-vsa-165'           
         'partner_status' => "'Waiting_for_cluster_applications"',
         'partner_status_raw' => '"Waiting for cluster applications to come online on the local node"',
         'time_until_autogiveback' => '',
         'is_random_node_' => '1',
         'refresh_rate' => '30',
         'time_updated' => '1351183610',
         'schedule' => '1351183586',
         'next_operation' => 'TAKEOVER',
         'resourcegroup' => 'sfo_activity',
      }

   }

=back


=head2 node

   The node executing the SFO operation: TAKEOVER or GIVEBACK 

=head2 node_operation

   This is the current "script" SFO operation operation: TAKEOVER or GIVEBACK 

=head2 node_status

   This will return the node status like below:  
    - takeover 
         Status string contains "In takeover"  or Takeover in progress"
    - unknown  
         Status string contains "Unknown"
    - unreachable
         Status string contains "Unreachable"
    - up  
         Status string contains "Connected to" 
    - takeover_scheduled 
         Status string contains "In takeover"  or Takeover in progress"
    - waiting_for_cluster_applications 
         Status string contains "Waiting for cluster applications to come online on the local node"


=head2 node_status_raw

   This will return the raw output of the node status, for example:    
      - In takeover, Auto giveback will be initiated in 237 seconds
      - "Takeover scheduled in 162 seconds"  

=head2 partner_status

   This will return the partner status like below:  
    - takeover 
         Status string contains "In takeover" or Takeover in progress"
    - unknown  
         Status string contains"Unknown"
    - unreachable
         Status string contains "Unreachable"
    - up   
         Status string contains "Connected to" or "Up"
    - waiting_for_giveback
         Status string contains "Waiting for giveback"
    - giveback_failed
         Status string contains "giveback failed due to ..."
    - relocating_sfo_aggregates
         Status string contains "relocating SFO aggregates in preparation of takeover"
    - partial_giveback   
         Status string contains "Connected to ... Partial giveback"
         
=head2 partner
 
   The partner of the SFO node being takenover or given back

=head2 partner_status_raw

   This will return the raw output of the status, for example:    
      - Relocated SFO aggregates to nst-vsa-162. Waiting for nst-vsa-162 to takeover
      - "Waiting for cluster applications to come online on the local node"  


=head2 next_operation
 
   The next operation scheduled to be performed by the SFO script: takover or giveback

=head2 time_until_autogiveback

   The time in seconds for autogiveback to occur.  

=head2 auto_takeover_schedhule

   The time in seconds before takeover is automatically initiated.  

=head2 is_random_node

   The value that specifies that SFO nodes are chosen randomly; this is dependent on the SFO_RANDOMIZE_NODES setting
   in the main SFO script: value is 0 or 1 - default 0  

=head2 auto_takeover_schedhule

   The time in seconds before takeover is automatically initiated.  

=head2 refresh_rate

   The time in seconds of how often the SFO nodes are polled for to get their status  

=head2 time_updated
 
   The time (epochs seocnds) when a node's SFO status is updated   

=head2 schedule
 
   This will return the time in epoch seconds when the next "script" SFO operation is schedule to occur 

=cut



