## @summary ConfigBroker::Events
## @author amritan@netapp.com
## @status public
## @pod here

=head1 NAME

ConfigBroker::Tasks

=head1 SYNOPSIS

        use ConfigBroker;
	use ConfigBroker::Tasks;

        my $cb = new ConfigBroker;
	my $tasks_obj = $cb->get_tasks_object();
        $tasks_obj->add_tasks(<parameters>);
        $tasks_obj->get_tasks(<parameters>);

=head1 DESCRIPTION

	Class module for Tasks object methods.

=cut


package ConfigBroker::Tasks;

use strict;
use warnings;
use DBI qw();
use Tharn qw($Log);
use Data::Dumper;
our @ISA = qw (Exporter);
use Params::Validate qw(validate SCALAR OBJECT HASHREF BOOLEAN ARRAYREF UNDEF);
use Module::Common qw(validate_mixed);
use POSIX qw (isdigit);
use ConfigBroker;

our @EXPORT = qw(add_tasks get_tasks delete_tasks);

my $TABLE_NAME = "tasks";


sub new{
        my ($pkg, %dbinfo) = @_;
	my $class = ref($pkg) || $pkg;
        my $self = {};
        $self->{'dbh'} = $dbinfo{'dbh'};
        $self->{'dsn'} = $dbinfo{'dsn'};
	bless($self,$class);
	return $self;
}

#function" to add tasks to the databse.
sub add_tasks{
	$Log->enter();
	my($self, %opts) = 
		validate_mixed(@_,
			[{type => OBJECT},],
			{recordsarrayref => {type => ARRAYREF, optional => 0,},},);
	$self->_add_tasks_bulk_mysql($opts{recordsarrayref});
	$Log->exit();
	return 0;
}

#helper function: That generates the insert query and exeutes it.
##                Insert one or more recrods.
sub _add_tasks_bulk_mysql{
	$Log->enter();
	my($self, $records) = @_;
	my $noofrecords = scalar(@{$records});
	my $count = 0;
	my $recordsprocessed = 0;
	my $insertquery = qq {INSERT INTO $TABLE_NAME};
	my $fieldname;
	my $fieldvalues;
	my $recordsquery;
	my $keycount;
	my $totalkeys;
	my $return_code = 0; 


	foreach (@$records){
		$recordsprocessed++;
		my %record;
		if(!keys %{$_}){
			$Log->debug("empty record");
			#bypass the record if an empty data record is passed in.
			next;
		}else{
			%record = %{$_};
		}
		if(!defined $record{task_name} || !defined $record{resource_id} || !defined $record{reporting_module}){
			#bypass this record because key fields not present
			next;
		}
		if(!defined $record{timestamp}){
			$record{timestamp} = time();
		}else{
			if(!isdigit($record{timestamp})){
				#bypass this record because timestamp is not in UTC
				$Log->debug("Timestmap must be in UTC! Skipping record.");
				next;
			}					
		}
		if(!defined $record{task_action}){
			#default task action is "begin"
			$record{task_action} = "begin";
		}

		if(!defined $record{task_type}){
			$record{task_type} = " ";
		}
		
		if(!defined $record{resource_type}){
			$record{resource_type} = " ";
		}
		$count++;
		#sort the keys
		my @sorted_fields = sort {$a cmp $b} (keys %record);			
	
		if($count == 1){
			$fieldname = "(";
			$fieldvalues = "VALUES (";
			$keycount = 0;
			$totalkeys = keys %record;
			foreach my $field (@sorted_fields){
				$keycount++;
				$fieldname .= $field;
				$fieldvalues .= "\'$record{$field}\'";
				if($keycount == $totalkeys){
					$fieldname .= ") ";
					$fieldvalues .= ") ";
				}else{
					$fieldname .= ", ";
					$fieldvalues .= ",";
				}							
			}
			$insertquery .= $fieldname;
		}else{
			$fieldvalues = "(";
			$keycount = 0;
			$totalkeys = keys %record;
			foreach my $field (@sorted_fields){
				$keycount++;
				$fieldvalues .= "\'$record{$field}\'";
				if($keycount == $totalkeys){
					$fieldvalues .= ") ";
				}else{
					$fieldvalues .= ",";
				}							
			}
		}

		$recordsquery .= $fieldvalues;
		if($recordsprocessed < $noofrecords){
			$recordsquery .= ", ";
		}				

	}

	if($recordsquery && $recordsquery ne ""){
		$insertquery .= $recordsquery;
		$Log->debug("insert query is: $insertquery");
		eval{
			$self->{'dbh'}->do($insertquery);
		};

		if($@){
			$Log->comment("Insert query failed with: $@");
			$return_code = 1;
		}
	}
	$Log->exit();
	return $return_code;
}

#helper function to create the filter string for the delete_tasks and get_tasks function
sub _create_filter_string{
	my ($self, %filter) = validate_mixed(@_,
					[{type => OBJECT},],
					{table_name       => {type => SCALAR, optional => 1},
					 task_name        => {type => SCALAR, optional => 1},
					 task_type        => {type => SCALAR, optional => 1},
					 task_action      => {type => SCALAR, optional => 1},
					 timestamp        => {type => SCALAR, optional => 1},
					 resource_id      => {type => SCALAR, optional => 1},
					 reporting_module => {type => SCALAR, optional => 1},
					},); 

	my $filterstring;

	#create the filter string
	foreach my $attr (keys %filter){
		if($attr eq 'timestamp'){
			if(!$filter{$attr} || !isdigit($filter{$attr})){
				next;
			}else{
				$filterstring .= "$attr  = \'$filter{$attr}\' AND ";
			}	
		}else{
			$filterstring .= "$attr  = \'$filter{$attr}\' AND "; 	
		}
	}
	
	#Strip out the last AND from the query string.
	if($filterstring){
		$filterstring =~ s/AND\s+$//;
	}
	
	return $filterstring;
}

#function: to query for records from the database.
sub get_tasks{
	$Log->enter();
	my ($self ,  %filter) = @_; 
	my $filterquery = $self->_create_filter_string(%filter);
	my $selectquery = "SELECT * from $TABLE_NAME ";
	my $whereclause = "WHERE ";
	my $data;
	my $query;

	#create the query
	if($filterquery){
		$query .= $selectquery . $whereclause . $filterquery;
	}else{
		$query .= $selectquery;
	}			
	
	$Log->debug("The select query is: $query");

	#execute the query
	eval{
		$data = $self->{'dbh'}->selectall_arrayref($query, {Slice => {}});
	};

	if($@){
		$Log->comment("Select query failed with: $@");
		return 1;
	}
	
	$Log->exit();
	return $data;
}

#function: delete tasks from the databast
sub delete_tasks{
	$Log->enter();
	my ($self, %filter) = @_;
	my $filterquery = $self->_create_filter_string(%filter);
	my $deletequery = " DELETE from $TABLE_NAME ";
	my $whereclause = "WHERE ";
	my $query;
	my $return_code = 0;
	
	if($filterquery){
		$query .= $deletequery. $whereclause . $filterquery;
	}else{
		$query .= $deletequery;
	}			
	
	$Log->debug("delete query is: $query");
	#execute the query
	eval{
		$self->{'dbh'}->do($query);
	};      
	 
	if($@){
		$Log->comment("Delete query failed with: $@");
		$return_code = 1;
	}       

	$Log->exit();
	return $return_code;
}

1;

=head1 METHODS

	The following are the methods that the Tasks library provdes.

=over

=item *

Methods:
	
	- add_tasks
	- get_tasks
	- delete_tasks


=item *
	
Examples:
		
	#add tasks 
	my $return_code = $eventobj->add_tasks(records => [$data1, $data2...]);

	#get tasks 
	my $results = $eventobj->get_tasks(%filter);

	#delete tasks 
	my $return_code = $eventobj->delete_tasks(%filter);

=item *

Filters:
	
	task_name            = Type: String. Optional in: get_tasks and delete_tasks. Not Optional in: add_tasks
	task_type            = Type: String. Optional in all functions. 
	task_action          = Type: String. Optional in all functions. Defaults to "begins" in add_tasks.
	timestamp            = Type: String and always UTC. Optional in all fucntions. CB will insert a UTC time value if not passed to the add_tasks function. 
	resource_id          = Type: String. Optional in: get_tasks and delete_tasks. Not Optional in: add_tasks
	resource_type        = Type: String. Optional in all functions. 
	reporting_module     = Type: String. Optional in: get_tasks and delete_tasks. Not Optional in: add_tasks

=back

=head2 add_tasks

	Object method; add task entries to the database.

=over

=item *

Usage:
	
	my $return_code = $eventobj->add_tasks(records => [$data1, $data2...]);
	records (not optional) -  An array of one or more data records to be inserted.

=item *

Sample data record passed in to be inserted:

	my $data1 = {
		task_name        => "volume create",
		task_type        => "create",
		task_action      => "begin",
		timestamp        => "1354898725",
		resource_id      => "vol1",
		resource_type    => "volume",
		reporting_module => "cifs"
	};
	
	"task_name", "resource_id" and "reporting_module" are essential fields. The record will be skipped if these are not passed.
	"timestamp" must be in UTC if passed in, else CB will pass the timestamp for the user. The record will be bypassed if ia timestamp passed in contains anything but digits.
	"task_action" if not specified, defaults to "begin".
	"task_type" and "resource_type" are optional fields and can be NULL.

=item *

Return:

	0 - if record inserted sucessfully; 1 - if error encountered while executing the query.

=back

=head2 get_tasks
	
	Object method; returns an array ref of record hashes of task entries.

=over

=item *

Usage:

	my $records = $eventobj-get_tasks(%filter);
	%filter(optional)  - a hash containing filters to query the database. If not specified all records are fetched.

=item *

Sample filter passed in to fetch records:

	my %filter = (
		task_name     => "volume create",
		task_action   => "begin",
		timestamp     => "1354898725",
	);

	Any combination of the keys can be specified.
	"timestamp" must be in UTC. If not it will not be considered as a filter while quering data.

=item * 

Return:

	Returns an array ref of record hashes.
	Sample returned data: 

	 $VAR1 = [
	 	{
	  		"reporting_module" => "cifs",
	  		"resource_id" => "vol1",
	 		 "timestamp" => "1354907497",
	 		 "task_type" => "create",
	 		 "task_name" => "volume create",
	 		 "task_action" => "end",
	 		 "resource_type" => "volume"
	  	},
	  	{
	 		 "reporting_module" => "cifs",
	  		 "resource_id" => "vol1",
	 		 "timestamp" => "11354907500",
	 		 "task_type" => "create",
	  		 "task_name" => "volume create",
	  		 "task_action" => "begin",
	  		 "resource_type" => "volume"
	  	}
	  ];

=back

=head2 delete_tasks

	Object method; deletes records from the table.

=over

=item *

Usage:

	my $return_code = $eventobj->delete_tasks(%filter);
	%filter(optional)  - a hash containing filters to delete a specific record(s). If not specified all records are deleted.

=item *

Sample filter passed in to fetch records:

	my %filter = (
		task_name     => "volume create",
		task_action   => "begin",
		timestamp     => "1354898725",
	);

	Any combination of the keys can be specified.
	"timestamp" must be in UTC. If not it will not be considered as a filter.

=item *

Return:

	0 - if record(s) deleted sucessfully; 1 - if error encountered while executing the query.

=back

=cut
