
package Confipv6;
## @summary Library of routines to check ipv6 is enabled on a cluster pair
##
## @synopsis
##
## use Confipv6;
##
## ontap_ipv6_support();
## is_ipv6_ra_enable();
## enable_ipv6_ra();    
## get_rlm_version();
## is_rlm_ipv6_supported();
## rlm_ipv6_configured();
## verify_cf_hwassist_values();
##
use strict;
use vars ('@ISA','@EXPORT');
require Exporter;
@ISA = ('Exporter');
@EXPORT =('ontap_ipv6_support', 'is_ipv6_ra_enable', 'enable_ipv6_ra', 'get_rlm_version',
           'is_rlm_ipv6_supported', 'rlm_ipv6_configured','ipv6_configured','check_ipv6_enabled');

use Tharn qw(say script_result host logcomment logresult param);
use boot;
use cluster;
use constant GB_time => 540;

#-------------------------------------------------------------------------------
# Function prototypes
#-------------------------------------------------------------------------------
sub ontap_ipv6_support($);
sub is_ipv6_ra_enable ($);
sub enable_ipv6_ra ($);
sub get_rlm_version($);
sub is_rlm_ipv6_supported($);
sub rlm_ipv6_configured($);
sub verify_cf_hwassist_values($);

#-------------------------------------------------------------------------------
# Prechecks to verify if the nodes are ipv6 configured 
#------------------------------------------------------------------------------- 

#-------------------------------------------------------------------------------
#Function to verify ontap ipv6 is enabled.
#-------------------------------------------------------------------------------

   sub ontap_ipv6_support($) {
      my ($my_conn) = @_;
      my ( $rlm_status, $rlm_version, $rlm_global_ip, $rlm_router_assigned_ip );
      my ( undef, undef, $ontap_ipv6 ) = say( $my_conn, "options ip.v6.enable" );

      #Verify ontap ipv6 is enabled

      if ( $ontap_ipv6 !~ /ip.v6.enable\s+on/ ) {
         return 0;
      }

      return 1;
   }

#-------------------------------------------------------------------------------
#Function to verify router assigned ipv6 is enabled.
#-------------------------------------------------------------------------------

   sub is_ipv6_ra_enable ($) {
      my ($my_conn) = @_;
      my ( undef, undef, $ipv6_ra ) = say( $my_conn, "options ip.v6.ra_enable" );

      # Verify router assigned ipv6 enabled

      if ( $ipv6_ra !~ /ip.v6.ra_enable\s+on/ ) {
         return 0;
      }

      return 1;
   }

#-------------------------------------------------------------------------------
#Function to Enable router  assigned ipv6
#-------------------------------------------------------------------------------

   sub enable_ipv6_ra ($) {
      my ($my_conn) = @_;

      my ( undef, undef, $ra_ipv6 ) = say( $my_conn, "options ip.v6.ra_enable on" );
      return;
   }

#-------------------------------------------------------------------------------
#Function to get RLM version
#-------------------------------------------------------------------------------

   sub get_rlm_version($) {
      my ($my_conn) = @_;
      my ( undef, undef, $rlm_status ) = say( $my_conn, "rlm status" );
      my $rlm_version;

      # Get RLM FW version

      $rlm_status =~ /.*Firmware Version:\s+(\d)/;
      $rlm_version = $1;

      return $rlm_version;
   }

#-------------------------------------------------------------------------------
#Function to verify rlm_ipv6 is supported
#-------------------------------------------------------------------------------

   sub is_rlm_ipv6_supported($) {
      my ($my_conn) = @_;
      my ( undef, undef, $rlm_status ) = say( $my_conn, "rlm status" );

      # Verify RLM ipv6 is enabled or disabled.

      if ( $rlm_status =~ /IPv6 configuration:\s+Disabled/ ) {
         return 0;
      }
      return 1;
   }

#-------------------------------------------------------------------------------
# Verify either GA / RA ip is assigned to RLM
#-------------------------------------------------------------------------------

   sub rlm_ipv6_configured($) {
      my ($my_conn) = @_;
      my ( undef, undef, $rlm_status ) = say( $my_conn, "rlm status" );
	sleep(5);

      my ( $rlm_global_ip, $rlm_router_assigned_ip );

      # If enabled,Verify RLM ipv6 is configured.

      if ( $rlm_status =~ /Global IP:\s+(\S+)/ ) {
         $rlm_global_ip = $1;
      }

      if ( $rlm_status =~ /Router Assigned IP:\s+(\S+)/ ) {
         $rlm_router_assigned_ip = $1;
         # Fail if RLM ipv6 is not configured.

         if ( !($rlm_global_ip) || !($rlm_router_assigned_ip) ) {
            logresult( 'FAIL', "RLM ipv6 is not configured; Please configure RLM ipv6" );
            return 0;
         }

         logcomment("RLM Global ip = $rlm_global_ip Router assigned ip: $rlm_router_assigned_ip");
         return 1;
      }
   }
1;
