# this is basically -*- perl -*-
#  vim: filetype=perl
#  $Id: //depot/prod/test/main/lib/ConsolelessSetup.pm#1
#  
#  Copyright (c) 2013 NetApp, Inc. All rights reserved. 
#
## @summary
##   ConsolelessSetup - Library for testing Consoleless Setup
##                     (Simple Setup, Big Easy, Cluster Setup Simplicity)
##
## @description
##   This file provides common functions for Consoleless cluster setup tests.
##
## @synopsis
##
## @changes
##
## @keywords setup, cluster setup, simplicity, big easy
##
## @Author : dl-cluster-qa
#

package ConsolelessSetup;

# Standard Libs

use strict;
use warnings;
use Data::Dumper;
use LWP::UserAgent;
use Net::Ping;
use HTTP::Request;
use HTTP::Response;
use XML::LibXML;
use DataUtils;

# NetApp Libs

use Tharn qw($Log host snooze);
use Hostrec qw();
use gx_licenses;
use NACL::APISet;
use NACL::Transit;
use NACL::C::Node;
use NACL::STask::Cluster;
use NACL::APISet::Exceptions::NoMatchingEntriesException qw(:try);
use NACL::APISet::Exceptions::ResponseException qw(:try);

################################################################################
#                                GLOBALS
################################################################################

my $Base_License  = gx_licenses::get_base_license( version => 2 );
my $AUTO_MGMT_LIF = "mgmt_auto";

################################################################################
# Subroutine: create_console_APISet()
# Parameters: node object
# Returns:    console apiset object
# Purpose:    create a NACL APISet connection to the console of the given node
################################################################################
sub create_console_APISet {
    my $self = shift;

    my %args = (
        node => undef,
        @_
    );

    my $node = $args{node};
    my $conn = Tharn::connect("$node#console", -timeout, "5");

    my @prompts = $conn->prompts();
    push(@prompts, '::>');
    $conn->prompts(@prompts);

    # Creating an APISet object
    my $apiset = NACL::APISet->new(
        connection => $conn,
        category   => "Node",
        interface  => "CLI",
        set        => 'CMode'
    );
    return $apiset;
}

################################################################################
# Subroutine: create_zapi_APISet()
# Parameters: node object and password
# Returns:    zapi apiset object
# Purpose:    create an apiset object for the node and password provided.
#             admin user assumed.
################################################################################
sub create_zapi_APISet {
    my $self = shift;

    my %args = (
        node     => undef,
        password => undef,
        @_
    );

    my $node      = $args{'node'};
    my $password  = $args{'password'};
    my $host      = host($node);

    my $zapi_apiset = NACL::APISet->new(
        hostobj   => $host,
        category  => "Node",
        interface => "ZAPI",
        set       => "CMode",
        password  => $password,
    );

    $zapi_apiset->set_transport_type("type" => "HTTPS");

    return $zapi_apiset;
}

################################################################################
# Subroutine: exit_wizard()
# Parameters: an array of node objects
# Returns:    none
# Purpose:    In the testcases we will use gx_boot to wipeconfig and 24/6 the
#             nodes, but will be left sitting at the setup wizard. This will
#             exit out and get us to the ::> prompt.
################################################################################
sub exit_wizard {
    my $self = shift;

    my %args = (
        nodes => undef,
        @_
    );

    my $nodes = $args{'nodes'};

    foreach my $node (@$nodes) {
        my $conn = Tharn::connect("$node#console", -timeout, "5");
        $conn->execute(
            command     => "\rexit",
            timeout     => 30,
            match_table => [ "login"    => "admin\r",
                             "Password" => "\r",],
            prompts     => ['::>']
        );
    }
}

################################################################################
# Subroutine: rerole_cluster_ports()
# Parameters: console apiset object
# Returns:    an array listing the configured cluster ports
# Purpose:    Many lab filers have non-platform default ports connected to the
#             cluster switches. gx_boot normally re-roles the ports based on
#             info in the NATE host file, so we'll have to do something similar.
################################################################################
# TODO IP Spaces will have some impact on this in FS.0, (port roles are depricated?) I imagine.
sub rerole_cluster_ports {
    my $self = shift;

    my %args = (
        'console-apiset' => undef,
        @_
    );

    my $apiset = $args{'console-apiset'};
    my @clustports;

    # Obtain port role info from hostrec
    my $host_obj = Hostrec->new(id=>$apiset->get_name());

    my @filter = ({nettype => "cluster", family => "inet4"});
    my @cluster_lifinfo = $host_obj->net_filter(include => \@filter, exclude => []);

    # First set all non-mgmt ports to data w/ default MTU of 1500.
    # This prevents us from leaving ports not referenced in nate_host files roled as cluster, etc.
    $apiset->network_port_modify(node=>'local',port=>'!e0M',role=>'data',mtu=>'1500');

    # Now role the cluster ports per values obtained from hostrec
    # If the hostrec does not specify cluster ports, default to e0a and e0b
    my @default_clustports = ('e0a','e0b');
    @clustports = ();
    my $i = 0;
    foreach my $cluster_lif (@cluster_lifinfo) {
        if (!$cluster_lif->{'interface'}) {
            $cluster_lif->{'interface'} = $default_clustports[$i];
        }
        $i++;
        $apiset->network_port_modify(node => "local",
                                     port => $cluster_lif->{'interface'},
                                     role => $cluster_lif->{'nettype'},
                                     mtu  => $cluster_lif->{'mtu'});
        push(@clustports, $cluster_lif->{'interface'});
    }

    return @clustports;
}

################################################################################
# Subroutine: get_mgmt_lif()
# Parameters: console apiset object, dhcp-avail (boolean)
# Returns:    lif_info
# Purpose:    check to see if a node-mgmt LIF was auto-generated from DHCP,
#             and if not, create one that NACL can use.
################################################################################
sub get_mgmt_lif {
    my $self = shift;

    my %args = (
        'console-apiset' => undef,
        'dhcp-avail'     => undef,
        @_
    );

    my $apiset     = $args{'console-apiset'};
    my $DHCP_Avail = $args{'dhcp-avail'};
    my $node       = $apiset->get_name();
    my $lif_info;
    my $routing_group;

    # Obtain port role info from hostrec
    my $host_obj = Hostrec->new(id=>$node);
    my @filter = ({nettype => "management", family => "inet4"});
    my @mgmt_lifinfo = $host_obj->net_filter(include => \@filter, exclude => []);

    $Log->comment("Hostrec defines node-mgmt IP $mgmt_lifinfo[0]->{'address'}" .
                  "on port $mgmt_lifinfo[0]->{'interface'} for $node");

    # Check to see if we have a node-mgmt LIF that matches what is defined
    # in the hostrec (so that NACL and other things will be expected to work)
    # If not, create one
    my $response;
    my $create_lif = 0;
    try {
        $response = $apiset->network_interface_show(role    => 'node-mgmt',
                                                    address => $mgmt_lifinfo[0]->{'address'});
        $lif_info = $response->get_parsed_output()->[0];
    }
    catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        $Log->comment("Failed to find an existing node-mgmt LIF that matches hostrec definition");
        $create_lif++;
    };

    # If we don't already have the expected node-mgmt LIF, create it
    if (!$create_lif) {
        $Log->comment("Found a node-mgmt LIF on $node that matches hostrec " .
                      "defined LIF ($mgmt_lifinfo[0]->{'address'})");
    } else {
        $Log->comment("Creating a node-mgmt LIF with address $mgmt_lifinfo[0]->{'address'}");
        $response = $apiset->network_interface_create(
            'lif'       => 'mgmt_ntest',
            'role'      => 'node-mgmt',
            'home-node' => 'localhost',
            'home-port' => $mgmt_lifinfo[0]->{'interface'},
            'address'   => $mgmt_lifinfo[0]->{'address'},
            'netmask'   => $mgmt_lifinfo[0]->{'netmask'});
        $response = $apiset->network_interface_show(
            'address' => $mgmt_lifinfo[0]->{'address'},
            'fields'  => 'routing-group,address');
        $lif_info = $response->get_parsed_output()->[0];
        $routing_group = $response->get_parsed_output()->[0]->{'routing-group'};
        my $create_route = 0;
        try {
            $response = $apiset->network_routing_groups_route_show(
                'routing-group' => $routing_group,
                'gateway'       => $mgmt_lifinfo[0]->{'gateway'});
        }
        catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
            $Log->comment("Failed to find an existing default route for routing" .
                          "group $routing_group");
            $create_route++;
        };
        if (!$create_route) {
            $Log->comment("Found a default route in routing group $routing_group");
        }
        else {
            $Log->comment("Creating default route using gateway $mgmt_lifinfo[0]->{'gateway'}");
            $response = $apiset->network_routing_groups_route_create(
                'routing-group' => $routing_group,
                'destination'   => '0.0.0.0/0',
                'gateway'       => $mgmt_lifinfo[0]->{'gateway'});
        }

        my $ping_tries = 5;
        my $ping = Net::Ping->new('external');
        $ping->ping($mgmt_lifinfo[0]->{'address'});
        while ((!$ping->ping($mgmt_lifinfo[0]->{'address'})) && ($ping_tries)) {
            $Log->warn("Ping attempt to node-mgmt IP $mgmt_lifinfo[0]->{'address'} failed");
            $Log->warn("Deleting and re-creating default route as workaround for reachability issue (INC10164539)");
            $apiset->network_routing_groups_route_delete('routing-group' => $routing_group,
                                                         'destination'   => '0.0.0.0/0');
            $apiset->network_routing_groups_route_create('routing-group' => $routing_group,
                                                         'destination'   => '0.0.0.0/0',
                                                         'gateway'       => $mgmt_lifinfo[0]->{'gateway'});
            $ping_tries--;
        }
    }

    return $lif_info;
}

################################################################################
# Subroutine: http_get_na_admin()
# Parameters: ip reference
# Returns:    na_admin URL on success, 0 on failure
# Purpose:    perfrom the http get on the na_admin URL vi $IP provided
################################################################################
sub http_get_na_admin {
    my $self = shift;

    my %args = (
        ip => undef,
        @_
    );

    my $IP = $args{ip};
    my $result;

    my $uri = "http://$IP/na_admin";
    my $ua  = LWP::UserAgent->new( agent => "RSA" );
    my $req = HTTP::Request->new( GET => $uri );

    $req->authorization_basic( 'admin', '' );

    $result = $ua->request($req);

    my $http_status = $result->status_line();
    if ( $http_status !~ /200/ ) {
        $Log->warn("Expected 200 Success but received: $http_status");
        $Log->comment( Dumper($result) );
        return 0;
    }
    return $result;
}

################################################################################
# Subroutine: verify_pre_cluster_http_result()
# Parameters: console apiset object and pre cluster http result content
# Returns:    1 on success, 0 on failure
# Purpose:    validate the http get content on na_admin pre cluster
################################################################################
sub verify_pre_cluster_http_result {
    my $self = shift;

    my %args = (
        'content'         => undef,
        'console-apiset'  => undef,
        @_
    );

    my $content = $args{'content'};
    my $apiset  = $args{'console-apiset'};
    my %hash;
    my $hp      = \%hash;
    my $status  = _parse_xml($hp, $content);

    if ( $status != 1 ) {
        $Log->comment("XML Parsing failed");
        return 0;
    }
    $Log->comment( "\nHP " . Dumper($hp) );

    # Verify nvramid, model, serialnumber
    my $response      = $apiset->system_node_show();
    my @parsed_output = @{ $response->get_parsed_output() };

    $Log->comment( "parse content: " . Dumper(@parsed_output) );

    my $exp_value = $parsed_output[0]->{'nvramid'};
    my $act_value = $hp->{'nvram-id'};

    $Log->comment("Expected nvram-id of $exp_value,  Found $act_value");

    if ( $exp_value ne $act_value ) {
        $Log->comment("Expected nvram-id of $exp_value,  Found $act_value");
        return 0;
    }
    $exp_value = $parsed_output[0]->{'model'};
    $act_value = $hp->{'model'};

    $Log->comment("Expected model of $exp_value,  Found $act_value");

    if ( $exp_value ne $act_value ) {
        $Log->comment("Expected model of $exp_value,  Found $act_value");
        return 0;
    }

    $exp_value = $parsed_output[0]->{'serialnumber'};
    $act_value = $hp->{'serial-number'};

    $Log->comment("Expected serialnumber of $exp_value,  Found $act_value");

    if ( $exp_value ne $act_value ) {
        $Log->comment("Expected serialnumber of $exp_value,  Found $act_value");
        return 0;
    }

    # Verify version
    $response      = $apiset->system_node_image_show();
    @parsed_output = @{ $response->get_parsed_output() };

    $Log->comment( "parse content: " . Dumper(@parsed_output) );

    $exp_value = $parsed_output[0]->{'version'};
    $act_value = $hp->{'version'};

    if ( $exp_value ne $act_value ) {
        $Log->comment("Expected version to be $exp_value,  Found $act_value");
        return 0;
    }

    # Verify partner-nvram-id
    $response      = $apiset->storage_failover_show();
    @parsed_output = @{ $response->get_parsed_output() };

    $Log->comment( "\nNODE INFO: " . Dumper(@parsed_output) );

    $exp_value = $parsed_output[0]->{'partner-nvramid'};
    $act_value = $hp->{'partner-nvram-id'};

    if ( $exp_value ne $act_value ) {
        $Log->comment("Expected partner-nvram-id of $exp_value, Found $act_value");
        return 0;
    }

    return 1;
}

################################################################################
# Subroutine: verify_post_cluster_http_result()
# Parameters: console apiset object and post cluster http result content
# Returns:    1 on success, 0 on failure
# Purpose:    validate the http get content on na_admin post cluster
################################################################################
sub verify_post_cluster_http_result {
    my $self = shift;

    my %args = (
        'content'         => undef,
        'console-apiset'  => undef,
        @_
    );

    my $content = $args{'content'};
    my $apiset  = $args{'console-apiset'};
    my %hash;
    my $hp      = \%hash;
    my $status  = _parse_xml($hp, $content);

    if ( $status != 1 ) {
        return 0;
    }

    my $response      = $apiset->cluster_identity_show();
    my @parsed_output = @{ $response->get_parsed_output() };

    $Log->comment( "parse content: " . Dumper(@parsed_output) );

    my $exp_value = $parsed_output[0]->{'name'};
    my $act_value = $hp->{'cluster-name'};

    if ( $exp_value ne $act_value ) {
        $Log->comment(
            "Expected cluster-name value of $exp_value,  Found $act_value");
        return 0;
    }

    $exp_value = "Data ONTAP Cluster-Mode";
    $act_value = $hp->{'version'};

    if ( $exp_value ne $act_value ) {
        $Log->comment("Expected post cluster version value of $exp_value, Found $act_value");
        return 0;
    }

    return 1;
}

################################################################################
# Subroutine: create_cluster_lifs()
# Parameters: zapi apiset object and console apiset object
# Returns:    number of failures
# Purpose:    create 2 cluster lifs on node referenced in argument $apiset
#             Use console apiset to get the port for lifs
################################################################################
sub create_cluster_lifs {
    my $self = shift;

    my %args = (
        'zapi-apiset'    => undef,
        'console-apiset' => undef,
        @_
    );

    my $zapi_apiset     = $args{'zapi-apiset'};
    my $console_apiset  = $args{'console-apiset'};
    my $failures        = 0;
    my $response;
    my @clus_ports;
    my $status          = _get_cluster_ports($console_apiset, \@clus_ports);

    if ( $status != 1 ) {
        $Log->comment("Cluster port retrieval failed");
        return 1;
    }

    if ( scalar(@clus_ports) < 2 ) {
        $Log->comment("2 CLUSTER PORTS REQUIRED");
        return 1;
    }

    # Exactly 2 cluster ports needed
    while ( scalar(@clus_ports) > 2 ) {
        pop(@clus_ports);
    }

    foreach my $port (@clus_ports) {
        $response = $zapi_apiset->net_interface_create(
            'interface-name'     => "clus_" . $port,
            role                 => "cluster",
            'home-node'          => "localhost",
            'home-port'          => $port,
            vserver              => "localhost",
            'is-ipv4-link-local' => "true",
        );

        if ( $response->get_execution_status() ne 'passed' ) {
            $Log->error( "Cluster lifs creation pre-cluster failed with "
                  . "'is-ipv4-link-local' equal to true.  " );

            $failures++;
        }
    }
    return $failures;
}

################################################################################
# Subroutine: delete_cluster_lifs()
# Parameters: console apiset object
# Returns:    none
# Purpose:    cleanup cluster lifs  on the ndoe via apiset connection
################################################################################
sub delete_cluster_lifs {
    my $self = shift;

    my %args = (
        'console-apiset' => undef,
        @_
    );

    my $apiset       = $args{'console-apiset'};
    my $response     = $apiset->network_interface_show(role => "cluster");
    my $cluster_lifs = $response->get_parsed_output();

    foreach my $lif (@$cluster_lifs) {
        $response = $apiset->network_interface_delete(
            lif      => $lif->{'lif'},
            vserver  => "localhost",
        );
    }
}

################################################################################
# Subroutine: ping_lifs()
# Parameters: source/destination console apiset objects
# Returns:    1 on success, 0 on failure
# Purpose:    ping cluster lifs on nodes connection via the apiset objs provided
#             full mesh ping
################################################################################
sub ping_lifs {
    my $self = shift;

    my %args = (
        'console-apiset-src' => undef,
        'console-apiset-dst' => undef,
        @_
    );

    my $console_apiset_src = $args{'console-apiset-src'};
    my $console_apiset_dst = $args{'console-apiset-dst'};

    my $response1 =
      $console_apiset_src->network_interface_show( role => "cluster" );
    my $cluster_lifs_src = $response1->get_parsed_output();

    my $response2 =
      $console_apiset_dst->network_interface_show( role => "cluster" );
    my $cluster_lifs_dst = $response2->get_parsed_output();

    my $response;
    my $failures = 0;
    foreach my $src (@$cluster_lifs_src) {
        foreach my $dest (@$cluster_lifs_dst) {
            try {
                $response = $console_apiset_src->network_ping(
                    lif         => $src->{'lif'},
                    'lif-owner' => 'localhost',
                    destination => $dest->{'address'}
                );
            } catch NACL::APISet::Exceptions::ResponseException with {
                $Log->warn("Ping failed from " . $src->{'lif'} . " to " .
                           $dest->{'address'});
                $failures++;
            };
        }
    }

    if ($failures) {
        return 0;
    }

    return 1;
}

################################################################################
# Subroutine: reboot_precluster()
# Parameters: console apiset object
# Returns:    none
# Purpose:    Use Transit to reboot and tansition to the cli
################################################################################
sub reboot_precluster {
    my $self = shift;

    my %args = (
        'console-apiset' => undef,
        @_
    );

    my $apiset = $args{'console-apiset'};

    my $transit_obj = NACL::Transit->new( name => $apiset->get_connection() );
    $transit_obj->reboot();
    $transit_obj->change_state( 'to' => 'CLI' );
}

################################################################################
# Subroutine: create_cluster()
# Parameters: zapi apiset object and console apiset object
# Returns:    1 on success, 0 on failure
# Purpose:    Create cluster lifs and a cluster(Both via zapi) on the node
#             connection via the apisets provided
################################################################################
sub create_cluster {
    my $self = shift;

    my %args = (
        'zapi-apiset'    => undef,
        'console-apiset' => undef,
        @_
    );

    my $zapi_apiset    = $args{'zapi-apiset'};
    my $console_apiset = $args{'console-apiset'};   # used to check cluster health post create.

    # Create cluster LIFs on node
    ConsolelessSetup->create_cluster_lifs('zapi-apiset'    => $zapi_apiset,
                                          'console-apiset' => $console_apiset);

    ###############################################################
    #$Log->comment("Creating cluster on $Nodes[0] using create ZAPI");
    ###############################################################
    my $clust_name = $console_apiset->get_name()."_".time();
    my $response = $zapi_apiset->cluster_create(
        license        => $Base_License,
        'cluster-name' => $clust_name
    );

    if ($response->get_execution_status() !~ /pass/) {
       $Log->comment("ZAPI test execution status failed doing cluster create");
       return 0;
    }

    my $max_time;
    my $timeout = 120;

    ################################################################
    $Log->comment("Polling create progress status using status ZAPI");
    ################################################################
    $max_time = time() + $timeout;
    my $cj_complete = 'false';
    while ((time() < $max_time) && ($cj_complete ne 'true')) {
        try {
            $response    = $zapi_apiset->cluster_create_join_progress_get();
            $cj_complete = $response->get_parsed_output()->[0]->{'attributes'}->[0]->{'cluster-create-join-progress-info'}->[0]->{'is-complete'};
        } catch NACL::APISet::Exceptions::ResponseException with {
            $Log->comment("ZAPI response failure due to SSL certificate re-generation");
            Tharn::snooze(2);
        };
    }

    if ($cj_complete ne 'true') {
        $Log->comment("Cluster Create did not complete successfully)");
        return 0;
    }

    # Check the cluster create succeeded
      #enable ssh
    $response = $console_apiset->security_login_create(
        authmethod          => "password",
        application         => "ssh",
        username            => "admin",
    );

    my $node = NACL::C::Node->new(name => $zapi_apiset->get_name());

    # Ignore these Health Monitor Checking alerts which are not being tested for
    # the purpose of the script. 
    # 1. 'ClusterSwitchMissing Alert',
    # 2. 'ControllerToShelfIomA_Alert'
    try {
        NACL::STask::Cluster->cluster_health_check(command_interface=>$node);
    }
    catch NACL::Exceptions::ClusterHealthCheckError with {
        my $exceptionObj = shift;
        
	$Log->comment("Error message is " . $exceptionObj->text());

        if ($exceptionObj->text() =~ /ClusterSwitchMissing_Alert/) {
            $Log->warn("Error ignored: Health Monitor throwed ClusterSwitchMissing_Alert.");
        } elsif ($exceptionObj->text() =~ /ControllerToShelfIomA_Alert/) {
            $Log->warn("Error ignored: Health Monitor throwed ControllerToShelfIomA_Alert");
        } else {
            $Log->error("Unable to pass cluster health check");
            $exceptionObj->throw();
        }
    }
    catch NACL::C::Exceptions::TimeNotSynchronised with {
        $Log->warn("Ignoring time not synchronized error since NTP is unconfigured");
    };

    return 1;
}

################################################################################
# Subroutine: cluster_join()
# Parameters: zapi-apiset1(node1), zapi-apiset2(node2) and console-apiset
# Returns:    1 on success, 0 on failure
# Purpose:    join node2 to the node1 that is already in the cluster
################################################################################
sub cluster_join {
    my $self = shift;

    my %args = (
        'zapi-apiset1'   => undef,
        'zapi-apiset2'   => undef,
        'console-apiset' => undef,
        @_
    );

    my $zapi_apiset1   = $args{'zapi-apiset1'};
    my $zapi_apiset2   = $args{'zapi-apiset2'};
    my $console_apiset = $args{'console-apiset'};  # used to check the status post join

    # Create cluster LIFs on node
    ConsolelessSetup->create_cluster_lifs('zapi-apiset'    => $zapi_apiset2,
                                          'console-apiset' => $console_apiset);

    # Get a cluster IP from node 1 to use in the join
    my $cluster_ip = _get_lif_ip($zapi_apiset1, "cluster");

    ############################################################
    $Log->comment("Joining cluster using join ZAPI");
    ############################################################
    my $response = $zapi_apiset2->cluster_join('cluster-ip-address'=>$cluster_ip);

    if($response->get_execution_status() !~ /pass/) {
        $Log->comment("ZAPI test execution status failed doing cluster_create");
        return 0;
    }

    my $max_time;
    my $timeout = 120;

    ################################################################
    $Log->comment("Polling create progress status using status ZAPI");
    ################################################################
    $max_time = time() + $timeout;

    my $cj_complete = 'false';
    while ((time() < $max_time) && ($cj_complete ne 'true')) {
        try {
            $response    = $zapi_apiset2->cluster_create_join_progress_get();
            $cj_complete = $response->get_parsed_output()->[0]->{'attributes'}->[0]->{'cluster-create-join-progress-info'}->[0]->{'is-complete'};
        } catch NACL::APISet::Exceptions::ResponseException with {
            $Log->comment("ZAPI response failure due to SSL certificate re-generation");
            Tharn::snooze(2);
        };
    }

    if ($cj_complete ne 'true') {
        $Log->comment("Cluster Join did not complete successfully)");
	return 0;
    }

    ########################################################
    $Log->comment("Verifying successful addition to cluster");
    ########################################################

    # This is a workaround.   Sometimes the sec log table is not fully
    # propogated to the join node, so the ssh auth method is not visible
    # even though it exists in the cluster.
    # Try to create it the login method.   If its a dup entry then ignore.
    # If it is NOT a dup and succeseds emit a warning
    # All other errors are errors.
    my $failures = 0;
    try {
        $response = $console_apiset->security_login_create(
            authmethod          => "password",
            application       => "ssh",
            username          => "admin",
          );
          $Log->warn("Should be a duplicate propogated from join node.");
          #ignore for now
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exceptionObj = shift;
        my $error_reason = $exceptionObj->text();

        if ($error_reason =~ /duplicate/) {
            $Log->comment("It is expected that the join node inherits the SSH role from the cluster");
        } else {
            $Log->error("Unable to allow ssh access for admin");
            $failures++;
        }
    };

    if ($failures) {
	return 0;
    }

    my $node = NACL::C::Node->new(name => $zapi_apiset2->get_name());

    # Ignore these Health Monitor Checking alerts which are not being tested for
    # the purpose of the script. 
    # 1. 'ClusterSwitchMissing Alert',
    # 2. 'ControllerToShelfIomA_Alert'
    try {
        NACL::STask::Cluster->cluster_health_check(command_interface=>$node);
    }
    catch NACL::Exceptions::ClusterHealthCheckError with {
        my $exceptionObj = shift;
        
	$Log->comment("Error message is " . $exceptionObj->text());
        
	if ($exceptionObj->text() =~ /ClusterSwitchMissing_Alert/) {
            $Log->warn("Error ignored: Health Monitor throwed ClusterSwitchMissing_Alert.");
        } elsif ($exceptionObj->text() =~ /ControllerToShelfIomA_Alert/) {
            $Log->warn("Error ignored: Health Monitor throwed ControllerToShelfIomA_Alert");
        } else {
            $Log->error("Unable to pass cluster health check");
            $exceptionObj->throw();
        }
    }
    catch NACL::C::Exceptions::TimeNotSynchronised with {
        $Log->warn("Ignoring time not synchronized error since NTP is unconfigured");
    };

    return 1;
}

################################################################################
# Subroutine: verify_precluster_certs()
# Parameters: console apiset object and zapi apiset object
# Returns:    1 on success, 0 on failure
# Purpose:    Verify creation of pre-cluster SSL certificates for ontapi
################################################################################
sub verify_precluster_certs {
    my $self = shift;

    my %args = (
        'console-apiset' => undef,
        'zapi-apiset'    => undef,
        @_
    );

    my $console_apiset = $args{'console-apiset'};
    my $zapi_apiset    = $args{'zapi-apiset'};
    my $response;
    my $preclus_dir;

    my $transit_obj = NACL::Transit->new( name => $console_apiset->get_connection() );
    $transit_obj->change_state( 'to' => 'SYSTEMSHELL' );

    $response = $console_apiset->execute_raw_command(
        command => "ls /mroot/etc/vserver_4294967295/certificates/ssl/server/");

    if ($response =~ /(0\+\w+\+preclus)/) {
        $preclus_dir = $1;
        $Log->comment("The pre-cluster SSL cert directory is: $preclus_dir");
    }
    else {
        $Log->error("Failed to find a pre-cluster SSL cert directory");
        return 0;
    }

    $response = $console_apiset->execute_raw_command(
        command => "ls /mroot/etc/vserver_4294967295/certificates/ssl/server/$preclus_dir");
    if ($response !~ /server.key/ || $response !~ /server.crt/) {
        $Log->error("Failed to find SSL certificates in preclus directory");
        return $TCD::FAIL;
    }

    # Need to escape the +'s so we can match contents of Apache conf file
    my $escaped_preclus_dir =~ s/\+/\\+/g;
    $response = $console_apiset->execute_raw_command(command=>"cat /var/etc/httpd-vserver.conf");

    if (($response !~ /$escaped_preclus_dir\/server.crt/) ||
        ($response !~ /$escaped_preclus_dir\/server.key/)) {
        $Log->error("Did not find preclus certs in httpd-vserver.conf");
        return 0;
    }

    $transit_obj->change_state( 'to' => 'CLI' );

    return 1;
}

################################################################################
# Subroutine: verify_postcluster_certs()
# Parameters: console apiset object and zapi apiset object
# Returns:    1 on success, 0 on failure
# Purpose:    Verify creation of post-cluster SSL certificates for ontapi
#             and removal of pre-cluster certificates
################################################################################
sub verify_postcluster_certs {
    my $self = shift;

    my %args = (
        'console-apiset' => undef,
        'zapi-apiset'    => undef,
        @_
    );

    my $console_apiset = $args{'console-apiset'};
    my $zapi_apiset    = $args{'zapi-apiset'};
    my $response;
    my $postclus_dir;

    my $transit_obj = NACL::Transit->new( name => $console_apiset->get_connection() );
    $transit_obj->change_state( 'to' => 'SYSTEMSHELL' );

    snooze(3);
    $response = $console_apiset->execute_raw_command(command=>"pwd");
    $response = $console_apiset->execute_raw_command(
        command => "ls /mroot/etc/vserver_4294967295/certificates/ssl/server/");

    if ($response =~ /preclus/) {
        $Log->error("Found a pre-cluster certificate directory in a post-cluster state");
        return 0;
    }

    if ($response =~ /(\w\+\w{8}\+\w{10}\.cert\.\w{10})/) {
        $postclus_dir = $1;
        $Log->comment("The post-cluster SSL cert directory is: $postclus_dir");
    }
    else {
        $Log->error("Failed to find a post-cluster SSL cert directory");
        return 0;
    }

    $response = $console_apiset->execute_raw_command(
        command => "ls /mroot/etc/vserver_4294967295/certificates/ssl/server/$postclus_dir");
    if ($response !~ /server.key/ || $response !~ /server.crt/) {
        $Log->error("Failed to find SSL certificates in preclus directory");
        return $TCD::FAIL;
    }

    # Need to escape the +'s so we can match contents of Apache conf file
    my $escaped_postclus_dir = $postclus_dir;
    $escaped_postclus_dir =~ s/\+/\\+/g;
    $response = $console_apiset->execute_raw_command(command=>"cat /var/etc/httpd-vserver.conf");
    if (($response !~ /$escaped_postclus_dir\/server.crt/) ||
        ($response !~ /$escaped_postclus_dir\/server.key/)) {
        $Log->error("Did not find post-cluster certs in httpd-vserver.conf");
        return 0;
    }
    if ($response =~ /preclus/) {
        $Log->error("Found reference to pre-cluster certificates in httpd-vserver.conf");
        return 0;
    }

    $transit_obj->change_state( 'to' => 'CLI' );

    return 1;
}

################################################################################
# Subroutine: verify_ssl_zapis()
# Parameters: zapi apiset object
# Returns:    1 on success, 0 on failure
# Purpose:    Verify ability to execute a ZAPI with and without SSL
################################################################################
sub verify_ssl_zapis {
    my $self = shift;

    my %args = (
        'zapi-apiset' => undef,
        @_
    );

    my $zapi_apiset = $args{'zapi-apiset'};
    my $response;
    my $exceptionObj;
    my $error_reason;
    my $failures = 0;

    # First we try the ZAPI over unencrypted HTTP
    $zapi_apiset->set_transport_type("type" => "HTTP");

    try {
        $response = $zapi_apiset->net_port_get_iter();
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        $exceptionObj = shift;
        $error_reason = $exceptionObj->text();
        $failures++;
    };

    if ($failures) {
        $Log->error("ZAPI execution failed with reason: $error_reason");
        return 0;
    }

    # Now try over HTTPS/SSL
    $failures = 0;
    $zapi_apiset->set_transport_type("type" => "HTTPS");

    try {
        $response = $zapi_apiset->net_port_get_iter();
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        $exceptionObj = shift;
        $error_reason = $$exceptionObj->text();
        $failures++;
    };

    if ($failures) {
        $Log->error("ZAPI execution failed with reason: $error_reason");
        return 0;
    }

    return 1;
}

################################################################################
# Subroutine: check_for_DHCP_node_mgmt_lif()
# Parameters: console apiset object
# Returns:    DHCP node management lif on success, 0 on failure
# Purpose:    validate auto generation of the node mgmt lif
################################################################################
sub check_for_DHCP_node_mgmt_lif {
    my $self = shift;

    my %args = (
        'console-apiset' => undef,
        @_
    );

    my $apiset = $args{'console-apiset'};
    my $response;
    my $lif;

    #If none found, error otherwise okay
    $response = $apiset->network_interface_show(
        role=>'node-mgmt',
        lif=>$AUTO_MGMT_LIF
    );

    $Log->comment ("Found Node Mgmt lif as expected");

    $lif = $response->get_parsed_output();
    my $routing_group = @$lif[0]->{'routing-group'};
    $response = $apiset->network_routing_groups_route_show(
        'routing-group' => $routing_group
    );
    $Log->comment ("Found Route lif as expected");

    my $ping = Net::Ping->new('external');
    my $mgmt_ip = @$lif[0]->{'address'};
    $ping->ping($mgmt_ip);

    if (!$ping->ping($mgmt_ip)) {
        $Log->error("Auto-generated node-mgmt LIF (" . $mgmt_ip .
            ") is unreachable from test controller!");
        return 0;
    }
    $Log->comment ("Ping Succeeded");

    return $lif;
}

#======================================================#
#                Private Subroutines
#======================================================#

#validate $content is syntactically correct
#return parsed output in $hp
sub _parse_xml {

    my $hp      = shift;
    my $content = shift;
    my $status  = 1;

    $content =~ /model\>(.*)\<\/model/;
    $hp->{'model'} = $1;
    $content =~ /serial-number\>(.*)\<\/serial-number/;
    $hp->{'serial-number'} = $1;
    $content =~ /\<nvram-id\>(.*)\<\/nvram-id\>/;
    $hp->{'nvram-id'} = $1;
    $content =~ /partner-nvram-id\>(.*)\<\/partner/;
    $hp->{'partner-nvram-id'} = $1;
    $content =~ /cluster-name\>(.*)\<\/cluster-name/;
    $hp->{'cluster-name'} = $1;
    $content =~ /version\>(.*)\<\/version/;
    $hp->{'version'} = $1;

    return $status;
}

#get a lif's IP via the apiset provided
sub _get_lif_ip {

    my $apiset = shift;
    my $role   = shift;

    my $response;
    my $retries = 0;

    $response = $apiset->network_interface_show( role => $role );

    my @lif = @{ $response->get_parsed_output() };

    $Log->comment( Dumper( \@lif ) );

    # Get IP
    my $IP = $lif[0]->{'address'};
    return $IP;
}

# returns an array of cluster ports(via $port_list) on the node referenced in apiset
sub _get_cluster_ports {

    my $apiset    = shift;
    my $port_list = shift;

    my @ports;

    my $response = $apiset->network_port_show( role => "cluster", link => "up" );
    my $cluster_ports_up = $response->get_parsed_output();
    if ( scalar(@$cluster_ports_up) < 2 ) {
        $Log->error( "Only " . scalar(@$cluster_ports_up)
              . " cluster port has link, so we will fail this test run." );
        return 0;
    }
    foreach my $port (@$cluster_ports_up) {
        push( @$port_list, $port->{port} );
    }

    $Log->comment( Dumper($port_list) );
    return 1;
}

1;
