###############################################################################
# Id: //depot/prod/test/main/lib/Coral.pm#65 $
#
# Copyright (c) 2008-2009 NetApp, Inc.
# All rights reserved.
###############################################################################
#
# Coral Utilities Package
#
# Provides methods for setting up, testing and destroying coral sets constructs
#
###############################################################################

package Coral;

use strict;
use Server;
use TracePkg;
use Options;
use Params::Common; # for aggregate and vserver definitions
use Data::Dumper;

###############################################################################
#
# my $coral = Coral->new();
#
# ...initializes the coral package if necessary, and returns a reference to
# the singleton.  Interface is idempotent.
#
###############################################################################

my %coral;

sub new
{
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $coral = \%coral;
    bless($coral, $class);

    if (not defined $coral{init}) {
        $coral{init} = 1;
        $coral{trace} = new TracePkg;
        $coral{options} = new Options;
        $coral{format} = "coral";
    }

    return $coral;
}

sub options { return $coral{options}; }
sub server { return $coral{server}; }
sub trace { return $coral{trace}; }
sub format { return $coral{format}; }

###############################################################################
#
# Coral->setFormat("acro");
#
# ...used to tell Coral to make Acro-format volumes or Coral-format volumes.
#
###############################################################################

sub setFormat
{
    my $coral = shift;
    $coral{format} = shift;
    die_script("Coral.pm: setFormat must take either \"acro\" or \"coral\"")
        unless $coral{format} eq "coral" or $coral{format} eq "acro";
}

###############################################################################
#
# die_config("reason")
# die_script("reason")
# die_fatal("reason")
# die_fail("reason")
#
# ...kills the script, reporting the supplied text and logging an appropriate
# NATE condition.
#
###############################################################################

sub die_config   # initial test configuration discovered to be unsound
{
    my $coral = shift;
    $coral{trace}->fatal(msg=>shift, result=>"CONFIG");
}

sub die_script   # script needs to be updated to match product output
{
    my $coral = shift;
    $coral{trace}->fatal(msg=>shift, result=>"SCRIPT");
}

sub die_fatal    # unexpected failure in running external tool etc
{
    my $coral = shift;
    $coral{trace}->fatal(msg=>shift, result=>"FATAL");
}

sub die_fail     # test discovered flaw in product
{
    my $coral = shift;
    $coral{trace}->fatal(msg=>shift, result=>"FAIL");
}

###############################################################################
#
# $coral->read_command_line();
#
# ...adds the following command-line options to the global Options object:
#
#        volname=coral_unit_test
#        ...volume name of coral set to construct (default "coral_unit_test").
#
#        volsize=1g
#        ...size of coral set constituent volumes (default "10g")
#
#        aggregate=stripedaggr
#        ...name of striped aggregate to use (default determined automatically).
#
#        juncname=coral_unit_test
#        ...name of junction from root to new coral set (default taken from volname)
#
#        vserver=vs0
#        ...name of vserver for new coral set (default determined automatically
#           if not specified)
#
###############################################################################

sub read_command_line
{
    my $coral = shift;

    my $defname = ($coral{format} eq "acro" ? "acro_unit_test" : "coral_unit_test");

    $coral{options}->parameters(
            definitions=>[
                {
                    name => "volname",
                    description => "volume name of coral set to construct (default \"$defname\")",
                    default => $defname,
                    spec => "s",
                },
                {
                    name => "volsize",
                    description => "size of coral set (total of all constituent volumes; default 10g)",
                    default => "10g",
                    spec => "s",
                },
                {
                    name => "juncname",
                    description => "name of junction from root to new coral set (default taken from volname)",
                    spec => "s",
                },
                {
                    name => "panic_on_error",
                    description => "Trigger panic when phase 2 errors occur",
                    spec => "boolean",
                    default => 1,
                }
            ]);
}

###############################################################################
#
# my $server = Server->new();
# $coral->specify_server($server);
#
# ...just lets coral know what server(s) you'll be working with.  Call this
# as soon as you know.
#
###############################################################################

sub specify_server
{
    my $coral = shift;
    $coral{server} = shift unless defined $coral{server};
    my $release = $coral{server}->get_release(format=>"cooked");
    #diag user requires us to use "sudo" for most filer commands in COT
    if ($release !~ /tricky/i) {
        $coral{sudo} = "sudo ";
    }
}

###############################################################################
#
# $coral->setup_test();
#
# ...constructs a new coral set (destroying the old one if necessary) and
# prepares it for your test.
#
###############################################################################

sub setup_test
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $volname = $args{volname};
    my $volsize = $args{volsize};
    my $juncname = $args{juncname};
    my $vserver = $args{vserver};
    my $aggregate = $args{aggregate};
    my $vol_params = $args{vol_params};
    $aggregate = $aggregate->[0] if (ref($aggregate) eq "ARRAY");

    $coral->die_config("must call Coral::specify_server before Coral::setup_test\n")
        unless defined $coral{server};

    # Make sure coral's ready to do its work
    $coral->assert_licensed();

    # Turn on panic-on-phase2-error on all the servers
    if ($args{panic_on_error}) {
        foreach my $filername (@coral{server}->servers()) {
            $coral->call_coraltest(server=>$filername, command=>"panic_on_error -p 1");
            }
    }

    # Pick a vserver if one wasn't specified
    $vserver = $coral->pick_vserver() unless defined $vserver;

    # Select a striped aggregate to use, in case one wasn't provided
    $aggregate = $coral->pick_aggregate(aggregate=>$aggregate);

    # Construct a new coral set
    $coral->create_coralset(
                        vserver=>$vserver,
                        volname=>$volname,
                        volsize=>$volsize,
                        aggregate=>$aggregate,
                        format=>$coral{format},
                        vol_params=>$vol_params);

    # Mount the coral set
    $coral->create_junction(
                        vserver=>$vserver,
                        volname=>$volname,
                        juncname=>$juncname);
}

###############################################################################
#
# $coral->exit_test();
#
# ...destroys a coral set and junction.
#
###############################################################################

sub exit_test
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $vserver = $args{vserver};
    my $volname = $args{volname};

    $coral->die_config("must call Coral::specify_server before Coral::exit_test\n")
        unless defined $coral{server};

    # Pick a vserver if one wasn't specified
    $vserver = $coral->pick_vserver(volname=>$volname) unless defined $vserver;

    # If there's already a junction or coral set, delete it
    $coral->delete_coralset(vserver=>$vserver, volname=>$volname);

    # Turn off panic-on-phase2-error on all the servers
    if ($args{panic_on_error}) {
        foreach my $filername (@coral{server}->servers()) {
            $coral->call_coraltest(server=>$filername, command=>"panic_on_error -p 0");
        }
    }
}

###############################################################################
#
# $coral->enable();
#
# ...ensures that coral is running on all filers, even if there's no coral
# volumes on one of them.  (Some tests, like gridiron's smoke test, might try
# to use 'coraltest readinode' to check out stuff on flexvols.  That stuff
# won't work unless coral is awake.)
#
###############################################################################

sub enable
{
    my $coral = shift;

    foreach my $filername (@coral{server}->servers()) {
        $_ = $coral->call_coraltest(server=>$filername, command=>"enable");
        $coral->die_fatal("failed to enable coral on $filername:\n$_")
            unless /succeeded/;
    }
}

###############################################################################
#
# $coral->assert_licensed();
#
# ...determines whether coral is licensed on all filers, and logs a fatal
# error if not.
#
###############################################################################

sub assert_licensed
{
    my $coral = shift;

    foreach my $filername (@coral{server}->servers()) {
        # TODO: soon we'll probe ngsh to see if licensing is enabled
        # $coral->die_config("cannot run test; coral is not licensed on $filername:\n$_")
        #     unless /coral is enabled/;
    }
}

###############################################################################
#
# my $success = $coral->delete_coralset(
#                    vserver=>"vs0",
#                    volname=>"coral_volume");
#
# ...deletes the constituent volumes that compose the specified coral set,
# along with any vldb records referring to the volumes or coral set
#
###############################################################################

sub delete_coralset
{
    my ($coral, %args) = (shift, timeout=>60, %{$coral{options}->get_hash()}, @_);
    my $vserver = $args{vserver};
    my $volname = $args{volname};
    my $timeout = $args{timeout};

    # Pick a vserver if one wasn't specified
    $vserver = $coral->pick_vserver(volname=>$volname) unless defined $vserver;

    # volume_purge will unmount, offline and delete the volume for us
    $coral{server}->volume_purge(
                        vserver=>$vserver,
                        volume=>$volname,
                        "-Timeout"=>$timeout);
}

###############################################################################
#
# my $aggregate = $coral->pick_aggregate(aggregate=>"suggested_aggr_name");
#
# ...returns the name of a striped aggregate for constructing a coral set.
# If an aggregate is suggested on input, it will be verified to be a striped
# aggregate and returned.
#
###############################################################################

sub pick_aggregate
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $aggregate = $args{aggregate};
    $aggregate = $aggregate->[0] if (ref($aggregate) eq "ARRAY");

    if (defined $aggregate)
    {
        my @rc = $coral{server}->storage_aggregate_show(aggregate=>$aggregate);
        my $aggr = $rc[0];
        my %aggr = %$aggr;
        return $aggr{aggregate} if $aggr{voltype} eq "striped";
        $coral->die_config("$aggregate is not a striped aggregate");
    }
    else
    {
        my @rc = $coral{server}->storage_aggregate_show();
        for (my $ii = 0; $ii < scalar(@rc); ++$ii) {
            my $aggr = @rc[$ii];
            my %aggr = %$aggr;
            return $aggr{aggregate} if $aggr{voltype} eq "striped";
        }
        $coral->die_config("you must configure a striped aggregate before running this test");
    }
}

###############################################################################
#
# my $vserver = $coral->pick_vserver(volname=>$volname);
#
# ...returns the name of the first vserver in the cluster; or, if a volume
# is specified, returns its vserver
#
###############################################################################

sub pick_vserver
{
    my ($coral, %args) = @_;
    my $volname = $args{volname};

    if (defined $volname)
    {
        my @rc = $coral{server}->volume_show(volume=>$volname);
        my $vs = @rc[0];
        my %vs = %$vs;
        return $vs{vserver};
    }
    else
    {
        my $role = $coral{server}->has_change(change => "vserver-role-change") ?
            "cluster" : "storage";
        my @rc = $coral{server}->vserver_show(role => $role);
        for (my $ii = 0; $ii < scalar(@rc); ++$ii) {
            my $vs = @rc[$ii];
            my %vs = %$vs;
            return $vs{vserver};
        }
        $coral->die_config("no vservers in cluster apparently; try creating one.");
    }
}

###############################################################################
#
# $coral->create_coralset(
#                    vserver=>"vs1",
#                    volname=>"testcoralset",
#                    volsize=>"10g",
#                    aggregate=>"stripedaggr",
#                    format=>"coral");
#
# ...constructs vldb entries and constituent volumes for a new coral set.
#
###############################################################################

sub create_coralset
{
    my ($coral, %args) = (shift, timeout=>60, %{$coral{options}->get_hash()}, @_);
    my $vserver = $args{vserver};
    my $volname = $args{volname};
    my $volsize = $args{volsize};
    my $aggregate = $args{aggregate};
    $aggregate = $aggregate->[0] if (ref($aggregate) eq "ARRAY");
    my $timeout = $args{timeout};
    my $format = $args{format};
    $format = "coral" unless defined $format;
    my $extra_vol_params = $args{vol_params};
    my @extra_vol_params = defined $extra_vol_params ? @$extra_vol_params : ();

    # Pick a vserver if one wasn't specified
    $vserver = $coral->pick_vserver() unless defined $vserver;

    my $minsize = $coral{server}->find_min_max_volume_size($aggregate, "minperm");
    my $vs_in_bytes = $volsize;
    my $ms_in_bytes = $minsize;
    #get the aggregate we're going to use
    my $aggrh = $coral{server}->storage_aggregate_show(aggregate => $aggregate);
    $coral{options}->unit_to_base($vs_in_bytes, $ms_in_bytes);
    if ($vs_in_bytes < $ms_in_bytes) {
        $coral{trace}->comment(msg=>"Volsize being adjusted from $volsize to $minsize");
        $volsize = $minsize;
    }

    my %vol_params = (
            vserver => $vserver,
            volume => $volname,
            size => $volsize,
            aggregate => $aggregate,
            "-Timeout"=>$timeout,
            @extra_vol_params);

    my $release = $coral{server}->get_release(format=>"cooked");
    if ($release !~ /tricky/i) {
        $vol_params{'stripe-format'} = ($format eq "acro") ? "full" : "data",
    }

    if ($format eq "acro") {
        my $stripe_count = $vol_params{'stripe-count'};
        if (not defined $stripe_count) {
            $stripe_count = 6;
            $vol_params{'stripe-count'} = $stripe_count;
        }
        # TODO: Fix this properly!
        # find_min_max_volume_size only accounts for the number of nodes, not
        # for the stripe count.  It therefore gets the min volume size wrong
        # for acro as we currently hardwire stripe-count to 6 (and we do it
        # *after* we've called find_min_max_volume_size!).
        # Quick workaround is to do the min size calculation manually here.
        if ($release !~ /tricky/i && $coral{server}->has_change(change => "min_dv_size_100mb")) {
            $minsize = (100 * $stripe_count) . "MB"; #100MB * stripe-count
        } else {
            $minsize = (1228.8 * $stripe_count) . "MB"; #1.2GB * stripe-count
        }
        $vs_in_bytes = $volsize;
        $ms_in_bytes = $minsize;
        $coral{options}->unit_to_base($vs_in_bytes, $ms_in_bytes);
        if ($vs_in_bytes < $ms_in_bytes) {
            $coral{trace}->comment(msg=>"Volsize being adjusted from $volsize to $minsize");
            $vol_params{'size'} = $minsize;
        }
    } else {
        #Get the nodes that are in the aggregate to determine the stripe count
        my $nodes = $aggrh->[0]{nodes};
        #Turn it into an array in case we get a comma separated scalar.
        if (ref($nodes) ne "ARRAY") {
            my @nodes = split(",",$nodes);
            $nodes = \@nodes;
        }
        if (int(@{$nodes}) < 2)  {
            $coral{trace}->comment(msg => "Default stripe count of ".int(@{$nodes}).  " is below the minimum (2) Adjusting.");
            $vol_params{'stripe-count'} = 2;
        } elsif (not defined $vol_params{'stripe-count'}) {
            $vol_params{'stripe-count'} = int(@{$nodes});
        }
    }

    eval {
        $coral{server}->volume_new(%vol_params);
    };

    #Force panic if we timeout during volume creation.
    if ($@ && $@ =~ /(TIMEOUT waiting for waitstrings|Job failed: Waiting for volume)/) {
        $coral{trace}->comment(msg=>"Timeout creating striped volume. Possibly BURT 303181 match:$1  ");
        $coral{server}->node_reboot(node => '*', dump => 'true');
        $coral->die_fatal("Timeout creating striped volume. Possibly BURT 303181 - Forcing panic $@");
    } elsif ($@) {
        $coral->die_fatal("Error creating striped volume: $@");
    }

    my @volshowrc = $coral{server}->volume_show(volume=>$volname);
}

###############################################################################
#
# $coral->create_junction(
#                     vserver=>"vs0",
#                     volname=>"coralvolume",
#                     juncname=>"/coral");
#
# ...constructs a junction from the root of the specified vserver to
# the specified child volume.
#
###############################################################################

sub create_junction
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $vserver = $args{vserver};
    my $volname = $args{volname};
    my $juncname = $args{juncname};
    $juncname = $volname unless defined $juncname;

    # Pick a vserver if one wasn't specified
    $vserver = $coral->pick_vserver(volname=>$volname) unless defined $vserver;

    my @rc = $coral{server}->vserver_show(vserver=>$vserver);
    my $data = $rc[0];
    my %data = %$data;
    my $rootvolume = $data{rootvolume};

    $juncname = "/$juncname" unless $juncname =~ /^\//;
    $coral{server}->volume_mount(
                        vserver => $vserver,
                        volume => $volname,
                        parent => $rootvolume,
                        "junction-path" => $juncname);
}

###############################################################################
#
# $coral->delete_junction(
#                     vserver=>"vs0",
#                     volname=>"coralvolume");
#
# ...remove a junction associated with a volume
#
###############################################################################

sub delete_junction
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $vserver = $args{vserver};
    my $volname = $args{volname};

    # Pick a vserver if one wasn't specified
    $vserver = $coral->pick_vserver(volname=>$volname) unless defined $vserver;

    $coral{server}->volume_unmount(
                        vserver => $vserver,
                        volume => $volname);
}

###############################################################################
#
# if ($coral->is_online(volname=>$volname)) { ... }
#
# ...returns nonzero if the specified volume is online
#
###############################################################################

sub is_online
{
    my ($coral, %args) = @_;
    my $volname = $args{volname};

    my @rc = $coral{server}->volume_show(volume=>$volname);
    my $vs = @rc[0];
    my %vs = %$vs;
    return ($vs{state} =~ /online/i) ? 1 : 0;
}


###############################################################################
#
# $coral->online(
#                vserver=>"vs0",
#                volname=>"coralvolume",
#                juncname=>"/coral");
#
# ...takes the specified coral volume online, or dies trying.
#
###############################################################################

sub online
{
    my ($coral, %args) = (shift, timeout=>60, %{$coral{options}->get_hash()}, @_);
    my $vserver = $args{vserver};
    my $volname = $args{volname};
    my $timeout = $args{timeout};
    my $juncname = $args{juncname};
    $juncname = $volname unless defined $juncname;


    # Pick a vserver if one wasn't specified
    $vserver = $coral->pick_vserver(volname=>$volname) unless defined $vserver;

    eval {
        $coral{server}->volume_modify(
                        vserver => $vserver,
                        volume => $volname,
                        state => "online",
                        "-Timeout"=>$timeout);
    };

    my $error = $@;

    $coral->die_fatal("unable to online volume: " . $error) unless (not $error or $error =~ /The volume or plex is online/);

    $juncname = "/$juncname" unless $juncname =~ /^\//;
    eval {
        $coral{server}->volume_mount(
                                     vserver => $vserver,
                                     volume => $volname,
                                     "junction-path" => $juncname);
    };
    if ($@) {
        # perform a show to aid in debugging
        $coral{server}->volume_show(volume=>$volname);
        $coral->die_fatal("unable to mount volume: " . $@);
    }
}

###############################################################################
#
# $coral->offline(vserver=>"vs0", volname=>"coralvolume");
#
# ...takes the specified coral volume offline, or dies trying.
#
###############################################################################

sub offline
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $vserver = $args{vserver};
    my $volname = $args{volname};

    # Pick a vserver if one wasn't specified
    $vserver = $coral->pick_vserver(volname=>$volname) unless defined $vserver;

    $coral{server}->volume_unmount(
                        vserver => $vserver,
                        volume => $volname);

    $coral{server}->volume_modify(
                        vserver => $vserver,
                        volume => $volname,
                        state => "offline");
}

###############################################################################
#
# $coral->create_flex_aggr(
#                    aggregate=>"aggrname",
#                    diskcount=>4);
#
# ...constructs the named flex aggregate on an arbitrary node.
#
###############################################################################

sub create_flex_aggr
{
    my ($coral, %args) = (shift, timeout=>60, %{$coral{options}->get_hash()}, @_);
    my $aggregate = $args{aggregate};
    $aggregate = $aggregate->[0] if (ref($aggregate) eq "ARRAY");
    my $diskcount = $args{diskcount};
    my $timeout = $args{timeout};

    eval {
        $coral{server}->storage_aggregate_new(
                                    aggregate => $aggregate,
                                    diskcount => $diskcount,
                                    raidtype => "raid4",
                                    voltype => "flex",
                                    "-Timeout" => $timeout);
    };
    my $origerr = $@;

    # The aggregate may not be created quickly enough and this command may fail.  We should hang out for a little
    # while and then check on the aggregate creation

    sleep 5;
    for (my $loop = 0; $loop < 6; ++$loop) {
        my $rc = $coral{server}->storage_aggregate_show (aggregate => $aggregate);
        return if (defined($rc) && ($rc->[0]->{state} !~ /creating/i));
        sleep 5;
    }

    $coral->die_fatal("unable to create flex aggregate: " . $origerr);
}


###############################################################################
#
# $coral->create_striped_aggr(
#                    aggregate=>"aggrname",
#                    diskcount=>4);
#
# ...constructs the named striped aggregate across as many nodes as you
# have in the cluster.
#
###############################################################################

sub create_striped_aggr
{
    my ($coral, %args) = (shift, timeout=>60, %{$coral{options}->get_hash()}, @_);
    my $aggregate = $args{aggregate};
    $aggregate = $aggregate->[0] if (ref($aggregate) eq "ARRAY");
    my $diskcount = $args{diskcount};
    my $nodelist = join(",",$coral->{server}->servers());
    my $timeout = $args{timeout};

    $nodelist .= "," . $nodelist if (scalar($coral->{server}->servers()) == 1);

    eval {
        $coral{server}->storage_aggregate_new(
                                    aggregate => $aggregate,
                                    diskcount => $diskcount,
                                    nodes => $nodelist,
                                    raidtype => "raid4",
                                    voltype => "striped",
                                    "-Timeout" => $timeout);
    };
    my $origerr = $@;
    # The aggregate may not be created quickly enough and this command may fail.  We should hang out for a little
    # while and then check on the aggregate creation

    sleep 5;
    for (my $loop = 0; $loop < 6; ++$loop) {
        my $rc = $coral{server}->storage_aggregate_show (aggregate => $aggregate);
        return if (defined($rc) && ($rc->[0]->{state} !~ /creating/i));
        sleep 5;
    }

    $coral->die_fatal("unable to create striped aggregate: " . $origerr);
}

###############################################################################
#
# $coral->delete_aggr(aggregate=>"aggrname");
#
# ...offlines and deletes the specified aggregate.
#
###############################################################################

sub delete_aggr
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $aggregate = $args{aggregate};
    $aggregate = $aggregate->[0] if (ref($aggregate) eq "ARRAY");

    $coral{server}->storage_aggregate_modify(
                                aggregate => $aggregate,
                                state => "offline");

    $coral{server}->storage_aggregate_delete(
                                aggregate => $aggregate);
}

###############################################################################
#
# my $aggrname = $coral->find_aggr_holding_constituent(volname=>"test_001");
#
# ...returns the name of the aggregate holding a particular constituent volume.
# Remember that the MDV has the same name as the overall coral set.
#
###############################################################################

sub find_aggr_holding_constituent
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $volname = $args{volname};
    my $vserver = $args{vserver};
    # Pick a vserver if one wasn't specified
    $vserver = $coral->pick_vserver(volname=>$volname) unless defined $vserver;

    my @rc = $coral{server}->volume_member_show(
                                vserver => $vserver,
                                volume => $volname);
    my $out = $rc[0];
    my %out = %$out;
    if (wantarray) {
        my $aggregates = $coral{server}->storage_aggregate_member_show(
                                                                       aggregate=>$out{aggregate});
        return ($out{aggregate}, $aggregates->[0]);
    } else {
        return $out{aggregate};
    }
}

###############################################################################
#
# my @dvs = $coral->find_volume_dvs(volname=>"test_001");
#
# ...returns the names of the DVs in the volume.
#
###############################################################################

sub find_volume_dvs
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $volname = $args{volname};
    my $vserver = $args{vserver};
    # Pick a vserver if one wasn't specified
    $vserver = $coral->pick_vserver(volname=>$volname) unless defined $vserver;

    my $volume_array = $coral{server}->volume_show(
                                                   vserver => $vserver,
                                                   volume => $volname);
    my @rc = $coral{server}->volume_member_show(msid => $volume_array->[0]{msid});
    my @dvs = map({$_->{volume}} grep($_->{volume} =~ /.+_\d\d\d/, @rc));
    return @dvs;
}

###############################################################################
#
# my $output = $coral->call_local(
#                   command=>"ls -l file",
#                   nonfatal=>1);
#
# ...returns the text output from running the specified command, run against
# whichever machine is executing this test script.  If you don't specify
# nonfatal=>1, then the script will die_fatal if the command returns nonzero.
#
###############################################################################

sub call_local
{
    my ($coral, %args) = @_;
    my $command = $args{command};
    my $hostname = `hostname`; chop $hostname;
    my $whoami = `whoami`; chop $whoami;

    $coral{trace}->debug(msg=>"executing '$command' here on $hostname (running as $whoami)");
    $_ = `/usr/software/bin/bash -c "$command 2>&1"`;
    $coral{trace}->debug(msg=>$_);

    if (not defined $args{nonfatal}) {
        $coral->die_fatal("'$command' failed:\n$_")
            unless (($? >> 8) == 0);
    }

    return $_;
}

###############################################################################
#
# my $output = $coral->call_dbladecli(
#                   server=>"ddeux-0",
#                   retries=>3,
#                   command=>"wafl_susp -w");
#
# ...returns the text output from running a command in dbladecli, run against
# the specified machine.
#
###############################################################################

sub call_dbladecli
{
    my ($coral, %args) = @_;
    my $server = $args{server};
    my $retries = $args{retries} || 5;
    my $timeout = $args{timeout};  $timeout = 60 unless defined $timeout;

    while ($retries--) {
        $coral{trace}->debug(msg=>"executing '$args{command}' on $server");
        $_ = $coral{server}->dblade(
                                    cmd=>$args{command},
                                    Server=>$server,
                                    "-Timeout" => $timeout);
        if ($_ =~ /^(syntax error|usage:)/i) {
            $coral->die_fatal("'$args{command}' failed: $_");
        }
        last if ($_ !~ /Already Connected/);
        sleep 1;
    }
    if ($! =~ /Already Connected/) {
        $coral->die_fatal("'$args{command}' failed: $_");
    }

    return $_;
}

###############################################################################
#
# my $output = $coral->call_vldbtest(
#                   command=>"dump -v",
#                   nonfatal=>1,
#                   timeout=>60);
#
# ...returns the text output from the vldbtest command, run against an
# arbitrary filer in the cluster.  The command is actually invoked *on*
# that filer, so /usr/sbin/vldbtest had better be there.
#
###############################################################################

sub call_vldbtest
{
    my ($coral, %args) = @_;
    my @servers = $coral{server}->servers();
    my $servername = $servers[0];
    my $timeout = $args{timeout};  $timeout = 60 unless defined $timeout;
    my $cmdline = $coral{sudo}."vldbtest -H localhost $args{command}";

    $coral{trace}->debug(msg=>"executing '$cmdline' on $servername");
    $_ = $coral{server}->shell(
                cmd=>$cmdline,
                Server=>$servername,
                "-Timeout" => $timeout);

    if (not defined $args{nonfatal}) {
        $coral->die_fatal("'$cmdline' failed:\n$_")
            unless /\bresult:\s*(\d+)\b/ and !$1;
        $coral->die_fatal("'$cmdline' failed:\n$_")
            if /call failed:/;
    }

    return $_;
}

###############################################################################
#
# my $output = $coral->call_zsmcli(
#                   server=>"ddeux-0",
#                   command=>"d-volume...",
#                   nonfatal=>1,
#                   timeout=>60);
#
# ...returns the text output from the zsmcli command, run against a specified
# filer.  The command is actually invoked *on* that filer, so
# /usr/sbin/zsmcli had better be there.
#
###############################################################################

sub call_zsmcli
{
    my ($coral, %args) = @_;
    my $servername = $args{server};
    my $timeout = $args{timeout};  $timeout = 60 unless defined $timeout;
    my $cmdline = $coral{sudo}."zsmcli localhost root \"\" $args{command}";
    my $retries = 5;
    my $delay = 1;

    while ($retries--) {
        $coral{trace}->debug(msg=>"executing '$cmdline' on $servername");
        $_ = $coral{server}->shell(
                        cmd=>$cmdline,
                        Server=>$servername,
                        "-Timeout" => $timeout);

        # If we got success, all done
        last if (/\bresults status=.*passed\b/);

        if (/\berrno=.*13158\b/) {
            sleep $delay;
            $delay *= 2;
        } elsif (not defined $args{nonfatal}) {
            $coral->die_fatal("'$cmdline' failed:\n$_");
        }
    }

    return $_;
}

###############################################################################
#
# my %output = $coral->call_wex(
#                   command=>"inode acro1.98",
#                   dump=>"size,blockcount",
#                   timeout=>60);
# printf ("got $output{size}, $output{blockcount}\n");
#
# ...returns the specified values as a hash for easy consultation.
#
###############################################################################

sub call_wex
{
    my ($coral, %args) = @_;
    my @servers = $coral{server}->servers();
    my $servername = $servers[0];
    my $timeout = $args{timeout};  $timeout = 60 unless defined $timeout;
    my $cmdline = $coral{sudo}."wex $args{command}";
    $cmdline .= " -find $args{find}" if defined $args{find};
    $cmdline .= " -dump $args{dump}" if defined $args{dump};
    $cmdline .= " +list $args{list}" if defined $args{list};
    $cmdline .= $args{filter} if defined $args{filter};
    $cmdline .= " | grep =" if defined $args{dump};

    my $rc = 0;
    $coral{trace}->debug(msg=>"executing '$cmdline' on $servername");
    my $text = $coral{server}->shell(
                    cmd=>$cmdline,
                    Server=>$servername,
                    "-Timeout" => $timeout);
                    # exit_code => \$rc    # TODO: eats output; bug in Server.pm!

    if (not defined $args{nonfatal}) {
        $coral->die_fatal("'$cmdline' failed:\n$text")
            unless $rc == 0;
    }
    $coral{trace}->debug(msg => "Received output:".$text);


    # If the caller asked for dump output, return an array of parsed LHS=RHS
    # data; the caller can use that either as an array of pairs, or as a
    # hash.  Otherwise, just return the raw text.
    if (defined $args{dump}) {
        my @array = ();
        foreach (split(/\n/,$text)) {
            # If a given line doesn't have an '=' sign, just ignore it.
            #
            # This is needed to work around the output from a gcov run
            # (which adds a line with out an '=' sign)
            #
            # Otherwise, split the key,value pair, and add it to the
            # results to be returned.
            chomp $_;
            if ($_ =~ m/=/) {
                $coral{trace}->debug(msg => "Splitting line:".$_);
                push(@array,split(/=/,$_));
            } else {
                $coral{trace}->debug(msg => "Ignoring line:".$_);
            }
        }
        $coral{trace}->debug(msg => "Returning:".Dumper(\@array));
        return @array;
    } else {
        return $text;
    }
}

###############################################################################
#
# my $output = $coral->call_coraltest(
#                   server=>"ddeux-0",
#                   command=>"online -m 213192381",
#                   timeout=>60);
#
# ...returns the text output from the coraltest command, run against a
# specified filer.  The command is actually invoked *on* that filer, so
# /usr/sbin/coraltest had better be there.
#
###############################################################################

sub call_coraltest
{
    my ($coral, %args) = @_;
    my $servername = $args{server};
    my $timeout = $args{timeout};  $timeout = 60 unless defined $timeout;
    my $cmdline = $coral{sudo}."coraltest $args{command} -f localhost";

    $coral{trace}->debug(msg=>"executing '$cmdline' on $servername");
    $_ = $coral{server}->shell(
                    cmd=>$cmdline,
                    Server=>$servername,
                    "-Timeout" => $timeout);
    if ($_ =~ /illegal option/) {
        die "'$cmdline' failed with:$_";
    }

    return $_;
}

###############################################################################
#
# my $uuid = $coral->create_uuid();
#
# ...returns 32 random hex digits.  Not the right algorithm, but should work
# for such simple tests.  Ideally we'd be calling management to do this.
#
###############################################################################

sub create_uuid
{
    return sprintf("%04X%04X%04X%04X%04X%04X%04X%04X",
                rand(65536), rand(65536), rand(65536), rand(65536),
                rand(65536), rand(65536), rand(65536), rand(65536));
}

###############################################################################
#
# my $uuid = $coral->format_uuid(uuid=>"12938120931312131238120398123242");
#
# ...returns a pretty-formatted "12938120-9313-1213-1238-120398123242".
#
###############################################################################

sub format_uuid
{
    my ($coral, %args) = @_;

    $_ = $args{uuid};
    return undef unless /^(........)(....)(....)(....)(............)$/;
    return $1."-".$2."-".$3."-".$4."-".$5;
}

###############################################################################
#
# my $iplist = $coral->find_data_addresses(
#                    {vserver=>"vs0" | volname=>"volname"});
#
# ...returns a semicolon-delimited list of all IP addresses through which the
# specified vserver can be mounted.
#
###############################################################################

sub find_data_addresses
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);

    # Check the incoming parameters
    my $vserver = $args{vserver};
    $vserver = $coral->pick_vserver(volname=>$args{volname}) unless defined $vserver;

    # Find the data addresses assigned to this vserver
    my @rc = $coral{server}->network_interfaces_show(server=>$vserver, role=>"data");

    my @addresses = ( );
    foreach my $interface (@rc) {
        my %interface = %$interface;
        push @addresses, $interface{address};
    }

    $coral->die_config("unable to find any addresses for vserver $vserver")
        unless scalar(@addresses) >= 1;

    return join(";", @addresses);
}

###############################################################################
#
# my $testpath = $coral->mount(
#                   vserver=>"vs0",
#                   volname=>"volumename",
#                   juncname=>"/coralset",
#                   protocol=>"tcp",
#                   version=>3);
#
# ...mounts the specified volume on a new mount point and returns the target
# path.  Defaults to TCPv3 unless told otherwise.
#
###############################################################################

sub mount
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);

    # Check the incoming parameters
    my $vserver = $args{vserver};
    my $volname = $args{volname};
    my $juncname = $args{juncname};
    my $protocol = $args{protocol};
    my $version = $args{version};
    $juncname = $volname unless defined $juncname;
    $juncname = "/$juncname" unless $juncname =~ /^\//;
    $protocol = "tcp" unless defined $protocol;
    $version = 3 unless defined $version;

    # Pick a vserver if one wasn't specified
    $vserver = $coral->pick_vserver(volname=>$volname) unless defined $vserver;

    # Find the data addresses assigned to this vserver; we'll just pick
    # one at random.
    my $iplist = $coral->find_data_addresses(vserver=>$vserver, volname=>$volname);
    my @addresses = split /;/,$iplist;
    my $address = $addresses[ rand(scalar(@addresses)) ];

    # Now, finally, mount the volume here on this filer.  Don't forget to
    # make it writable by non-superusers, just for testing convenience.
    my $mountpoint = "/t/coraltest/" . create_uuid();

    # Added the following packet capture to aid in investigation of
    # BURT 254575.

    my $capture_log = $main::TOP_LOGDIR_UNIX."/trace_pcap";
    $coral{trace}->comment(msg=>"Starting tcpdump in the background and storing trace at $capture_log");
    system("sudo /usr/sbin/tcpdump -w $capture_log -s 0 -i any -W 10 -C 10 host $address&");

    $coral->call_local(command=>"sudo mkdir -p $mountpoint");
    $coral->call_local(command=>"sudo chmod 777 $mountpoint");
    $coral->call_local(command=>"sudo mount -o soft,intr,proto=$protocol,vers=$version,actimeo=0 $address:/ $mountpoint");
    $coral->call_local(command=>"sudo chmod 777 $mountpoint");
    $coral->call_local(command=>"sudo chmod 777 $mountpoint$juncname");
    $coral->call_local(command=>"sudo ls -lad $mountpoint");
    $coral->call_local(command=>"sudo ls -lad $mountpoint$juncname");

    # Added the following for BURT 254575.

    $coral{trace}->comment(msg=>"Killing all tcpdump processes.");
    system("sudo killall tcpdump");

    return $mountpoint . $juncname;
}

###############################################################################
#
# $coral->unmount(testpath=>$path_returned_from_mount);
#
# ...unmounts and removes the specified directory, reversing a 'mount' call.
#
###############################################################################

sub unmount
{
    my ($coral, %args) = @_;
    my $testpath = $args{testpath};
    my $mountpoint = $testpath;
    $mountpoint =~ s/\/[^\/]*$//g;

    $coral->call_local(command=>"sudo umount $mountpoint", nonfatal=>1);
    $coral->call_local(command=>"sudo rmdir $mountpoint", nonfatal=>1);
}

###############################################################################
#
# my $geom = $coral->get_geometry(volname=>"coral_volume_name");
#
# ...returns a geometry object that encapsulates data about a coral set.
# useful for calling some of the other routines around here.
#
###############################################################################

sub get_geometry
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $volname = $args{volname};
    # TODO: Remove this arg; always get the striping table
    my $striping_table = $args{striping_table};

    # Prepare an output object
    my %geom;

    # Find the MSID of the coral set
    my $msid = $coral->find_volume_msid_by_name(volname=>$volname);
    $coral->die_config("unable to find coral set $volname in order to read its geometry")
        unless defined $msid;
    $geom{msid} = $msid;

    # Find that msid's vldb record
    my @dvs;
    my @dvservers;
    $_ = $coral->call_vldbtest(command=>"msidlookup -m $msid");
    foreach $_ (split /[\n\r]/) {
        if (/\bdsid: (\d+)/i) {
            my $dsid = $1;
            my $server = $coral->find_filer_by_dsid(dsid=>$dsid);
            if ($coral{format} ne "acro" && ! defined $geom{mdv_dsid}) {
                $geom{mdv_dsid} = $dsid;
                $geom{mdv_filer} = $server;
            } else {
                push @dvs, $dsid;
                push @dvservers, $server;
            }
        } elsif (/stripe width: (\d+)/) {
            $geom{width} = $1;
        } elsif (/stripe algorithm: (\d+)/) {
            $geom{algorithm} = $1;
        }
    }

    $coral->die_config("unable to parse coral set $volname vldb record")
        unless defined $geom{width}
        and defined $geom{algorithm}
        and (scalar(@dvs) >= 1);

    if (defined $striping_table) {
        my $msid_hex = sprintf('0x%08X', $msid);

        $_ = $coral->call_zsmcli(
            server=>$dvservers[0],
            command=>"d-striping-table-list-iter-next tag=$msid_hex:1:1:1 maximum=1");

        # parse
        $coral->die_config("unable to find striping table for msid $msid")
            unless /<records>1<\/records>.*<algorithm>3<\/algorithm>.*<striping-table>(.+)<\/striping-table>/s;

        my $table = $1;
        # strip out the xml tags
        $table =~ s/<\/?data>//g;
        my @table = split(/\s+/, $table);
        shift @table;

        # convert hex to decimal
        for (my $i = 0; $i < scalar @table; $i++) {
            $table[$i] = hex $table[$i];
        }

        $geom{striping_table} = \@table;
    }

    $geom{dvlist} = join(";",@dvs);
    $geom{dvfilers} = join(";",@dvservers);
    return \%geom;
}

###############################################################################
#
# my $inum = $coral->get_file_inum(
#                   volname=>"coral_sandbox",
#                   filename=>"/path/in/volume");
#
# ...returns the inode number of a particular file within the specified
# volume.  Works for directories et al as well.
#
###############################################################################

sub get_file_inum
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $filename = $args{filename};
    my $volname = $args{volname};
    $filename =~ s/^\///g;

    # The caller might have supplied what we need to know in order to find
    # the inum (or he might not).

    my $server = $args{server};
    my $dbladename = $args{dbladename};

    if (!defined $server || !defined $dbladename) {

        # Find out what server hosts this volume (or this volume's MDV)

        my $dsid = $coral->find_volume_dsid_by_name(volname=>$volname);

        $coral->die_config("unable to find volume $volname vldb record")
            unless defined $dsid;

        $server = $coral->find_filer_by_dsid(dsid=>$dsid);
        $coral->die_config("unable to find filer hosting $dsid")
            unless defined $server;

        # Find out what the D-blade calls this volume

        $dbladename = $coral->find_volume_dbladename_by_dsid(dsid=>$dsid);
    }

    # Now ask the dblade what that file handle is.  This part is
    # pretty iffy, since the dbladcli is a crappy way to get data; it often
    # simply fails, or takes so long to clear its backlog of output data
    # that the command times out.  So we'll try several times.

    my $inumhex = undef;
    my $gen = undef;
    for (my $retry = 0; ($retry < 5) && (!defined $inumhex); ++$retry) {
        $_ = $coral->call_dbladecli(
                        server=>$server,
                        command=>"priv set diag ; showfh /vol/$dbladename/$filename");
        $inumhex = $1 if (/fileid=(0x[0-9a-f]*)/i);
        $gen = $1 if (/gen=(0x[0-9a-f]*)/i);
    }

    $coral->die_config("unable to find file $filename within volume $volname ($server)")
        unless defined $inumhex;

    return (wantarray) ? (hex $inumhex, hex $gen) : hex $inumhex;
}

###############################################################################
#
# my %inode = $coral->read_inode(dsid=>$dsid, inum=>$inum, [geom=>$geom]);
#
# ...reads an inode directly from a coral set's constituent volume, returning
# a hash for all relevant fields.  Supply a geometry object for faster work;
# otherwise three calls are necessary instead of one.
#
###############################################################################

$Coral::INODE_STATE_MISSING = 0;   # don't have such an inode cached at all
$Coral::INODE_STATE_INVALID = 1;   # inode is there, but it's marked invalid
$Coral::INODE_STATE_VALID = 2;     # inode is there and cahce is marked valid

sub read_inode
{
    my ($coral, %args) = @_;
    my $dsid = $args{dsid};
    my $inum = $args{inum};
    my $geom = $args{geom};  # optional

    # Find the filer that hosts that volume
    my $filer = $coral->find_filer_by_dsid(dsid=>$dsid, geom=>$geom);
    $coral->die_config("unable to find filer hosting $dsid")
        unless defined $filer;

    # Direct a coraltest call there, to read the inode directly
    $_ = $coral->call_coraltest(
                    server=>$filer,
                    command=>"readinode -d $dsid -i $inum -v");
    $coral->die_fatal("unable to read inode $dsid.$inum:\n$_")
        if /readinode failed/;

    # Crack the output parameters.  This part is straight-forward.
    my %inode = ( );

    if (/is not cached/) {
        $inode{state} = $Coral::INODE_STATE_MISSING;
    } elsif (/is cached and valid/) {
        $inode{state} = $Coral::INODE_STATE_VALID;
    } else {
        $inode{state} = $Coral::INODE_STATE_INVALID;
    }

    if ($inode{state} != $Coral::INODE_STATE_MISSING) {
        foreach $_ (split /[\n\r]/) {
            if (/(\S+)\s*:\s*(\S+)/) {
                my $lhs = $1;
                my $rhs = $2;
                $rhs =~ s/\///g;    # timestamps are just big ugly numbers
                $rhs = eval($rhs);  # turn hex strings into usable decimals
                $inode{$lhs} = $rhs;
            }
        }
    }

    return %inode;
}

###############################################################################
#
# my ($role, $isowner, $nextstripe) = $coral->query_role(
#                                                 geom=>$geom,
#                                                 dsid=>$dsid,
#                                                 inum=>$inum,
#                                                 offset=>$offset);
#
# ...returns whether the specified volume is the MDV, FAV or just another DV;
# if an "offset" parameter is supplied, will also indicate whether the
# specified volume is responsible for that stripe offset, and the offset where
# the next stripe begins.
#
###############################################################################

$Coral::VOLUME_ROLE_MDV = 0;
$Coral::VOLUME_ROLE_FAV = 1;
$Coral::VOLUME_ROLE_DV = 2;

sub query_role
{
    my ($coral, %args) = @_;
    my $geom = $args{geom};
    my %geom = %$geom;
    my $dsid = $args{dsid};
    my $inum = $args{inum};
    my $offset = $args{offset};

    # Find the filer that hosts that volume
    my $filer = $coral->find_filer_by_dsid(dsid=>$dsid, geom=>$geom);
    $coral->die_config("unable to find filer hosting $dsid")
        unless defined $filer;

    # Direct a coraltest call there, to read the inode directly
    my $cmdline = "queryrole -m $geom{msid} -d $dsid -i $inum";
    $cmdline .= " -o $offset -l 1" if defined $offset;
    $_ = $coral->call_coraltest(server=>$filer, command=>$cmdline);
    $coral->die_fatal("failed to query role of $dsid in volume $geom{msid}:\n$_")
        if /reports error/;

    # Crack the output from the command
    my $role;
    if ($dsid == $geom{mdv_dsid}) {
        $role = $Coral::VOLUME_ROLE_MDV;
    } elsif (/serves as the fav/i) {
        $role = $Coral::VOLUME_ROLE_FAV;
    } else {
        $role = $Coral::VOLUME_ROLE_DV;
    }

    my $isowner = undef;
    if (/owns the stripe/) {
        $isowner = 1;
    } elsif (/does not own the stripe/) {
        $isowner = 0;
    }

    my $nextstripe = undef;
    if (/begins at offset (\d+)/) {
        $nextstripe = $1;
    }

    return ($role, $isowner, $nextstripe);
}

###############################################################################
#
# $coral->verify_inode(geometry=>$geom, inum=>$inum, flags=>$verify_flags);
#
# ...reads inodes directly from a coral set's constituent volumes, and ensures
# that they match your expectations.  Pass a combination of the flags below
# to indicate what should be tested.
#
###############################################################################

$Coral::VERIFY_MDV_VALID       = 0x00000001;
$Coral::VERIFY_MDV_MISSING     = 0x00000002;
$Coral::VERIFY_FAV_VALID       = 0x00000010;
$Coral::VERIFY_FAV_INVALID     = 0x00000020;
$Coral::VERIFY_FAV_MISSING     = 0x00000040;
$Coral::VERIFY_DV_VALID        = 0x00000100;
$Coral::VERIFY_DV_INVALID      = 0x00000200;
$Coral::VERIFY_DV_MISSING      = 0x00000400;
$Coral::VERIFY_TYPE_DIRECTORY  = 0x00001000;
$Coral::VERIFY_TYPE_REGULAR    = 0x00002000;

sub verify_inode
{
    my ($coral, %args) = @_;

    # Crack the input args, and assemble a list of all constituent volumes
    my $geom = $args{geometry};
    my %geom = %$geom;
    my $inum = $args{inum};
    my $flags = $args{flags};
    my @dsids = split /;/,$geom{dvlist};
    unshift @dsids, $geom{mdv_dsid};

    # As we go through the cached inodes, we'll be comparing meta-data caches.
    my %cached;

    # Start looking through those volumes, and verify everything we can
    foreach my $dsid (@dsids) {
        my %inode = $coral->read_inode(dsid=>$dsid, inum=>$inum, geom=>$geom);
        my ($role, $isowner, $nextstripe) = $coral->query_role(geom=>$geom, dsid=>$dsid, inum=>$inum);

        # If the caller thinks he knows the MDV's cache state, check it
        if (($role == $Coral::VOLUME_ROLE_MDV) && ($flags & ($Coral::VERIFY_MDV_VALID | $Coral::VERIFY_MDV_MISSING))) {
            $coral->die_fail("mdv shows inode $dsid.$inum missing unexpectedly")
                if ($inode{state} == $Coral::INODE_STATE_MISSING) and !($flags & $Coral::VERIFY_MDV_MISSING);
        }


        # If the caller thinks he knows the FAV's cache state, check it
        if (($role == $Coral::VOLUME_ROLE_FAV) && ($flags & ($Coral::VERIFY_FAV_VALID | $Coral::VERIFY_FAV_INVALID | $Coral::VERIFY_FAV_MISSING))) {
            $coral->die_fail("fav shows inode $dsid.$inum missing unexpectedly")
                if ($inode{state} == $Coral::INODE_STATE_MISSING) and !($flags & $Coral::VERIFY_FAV_MISSING);
            $coral->die_fail("fav shows inode $dsid.$inum marked invalid unexpectedly")
                if ($inode{state} == $Coral::INODE_STATE_INVALID) and !($flags & $Coral::VERIFY_FAV_INVALID);
            $coral->die_fail("fav shows inode $dsid.$inum valid unexpectedly")
                if ($inode{state} == $Coral::INODE_STATE_VALID) and !($flags & $Coral::VERIFY_FAV_VALID);
        }

        # If the caller thinks he knows the DV's cache state, check it
        if (($role == $Coral::VOLUME_ROLE_DV) && ($flags & ($Coral::VERIFY_DV_VALID | $Coral::VERIFY_DV_INVALID | $Coral::VERIFY_DV_MISSING))) {
            $coral->die_fail("dv shows inode $dsid.$inum missing unexpectedly")
                if ($inode{state} == $Coral::INODE_STATE_MISSING) and !($flags & $Coral::VERIFY_DV_MISSING);
            $coral->die_fail("dv shows inode $dsid.$inum marked invalid unexpectedly")
                if ($inode{state} == $Coral::INODE_STATE_INVALID) and !($flags & $Coral::VERIFY_DV_INVALID);
            $coral->die_fail("dv shows inode $dsid.$inum valid unexpectedly")
                if ($inode{state} == $Coral::INODE_STATE_VALID) and !($flags & $Coral::VERIFY_DV_VALID);
        }

        # Compare the cached data here against any data cached elsewhere
        if ($inode{state} == $Coral::INODE_STATE_VALID) {
            foreach my $lhs (keys %geom) {
                my $rhs = $geom{$lhs};

                # We won't compare FAV-owned meta-data or block counts
                next if ($lhs=="atime" or $lhs=="crtime" or $lhs=="ctime" or $lhs=="mtime" or $lhs=="size" or $lhs=="blocks");

                # Otherwise, all cached meta-data must be identical
                if (not defined $cached{$lhs}) {
                    $cached{$lhs} = $rhs;
                } else {
                    $coral->die_fail("inode $inum in volume $geom{msid} has different cached values for $lhs ($cached{$lhs} and $inode{$lhs})")
                        unless $cached{$lhs} == $inode{$lhs};
                }
            }

            # If the caller thinks he knows the inode's type, we should check it
            if ($flags & ($Coral::VERIFY_TYPE_DIRECTORY | $Coral::VERIFY_TYPE_REGULAR)) {
                $coral->die_fail("inode $dsid.$inum is type $inode{wafl_type}, should be a directory")
                    if ($flags & $Coral::VERIFY_TYPE_DIRECTORY) and ($inode{wafl_type} != 2);
                $coral->die_fail("inode $dsid.$inum is type $inode{wafl_type}, should be a regular file")
                    if ($flags & $Coral::VERIFY_TYPE_REGULAR) and ($inode{wafl_type} != 1);
            }
        }
    }

    $coral{trace}->debug(msg=>"verify-inode $inum (flags ".sprintf("0x%X",$flags).") passed");
}

###############################################################################
#
# $coral->create_snapshot(
#        vserver=>"vs0",
#        volname=>"coral_sandbox",
#        snapname=>"snapshot_name");
#
# ...just snapshots the specified volume, and dies if it fails.
#
###############################################################################

sub create_snapshot
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $vserver = $args{vserver};
    my $volname = $args{volname};
    my $snapname = $args{snapname};

    # Pick a vserver if one wasn't specified
    $vserver = $coral->pick_vserver(volname=>$volname) unless defined $vserver;

    $coral{server}->volume_snapshot_new(
                    vserver=>$vserver,
                    volume=>$volname,
                    snapshot=>$snapname);
}

###############################################################################
#
# my $filername = $coral->find_filer_by_dsid(dsid=>1027, [geom=>$geom]);
#
# ...returns the name of the filer currently hosting the specified volume.
# If you supply a geometry object, then the lookup is done just by consulting
# that object if possible--much faster.
#
###############################################################################

sub find_filer_by_dsid
{
    my ($coral, %args) = @_;
    my $dsid = $args{dsid};
    my $geom = $args{geom};

    if (defined $geom) {
        my %geom = %$geom;
        return $geom{mdv_filer} if ($dsid == $geom{mdv_dsid});

        my @dsids = split /;/,$geom{dvlist};
        my @filers = split /;/,$geom{dvfilers};
        for (my $ii = 0; $ii < scalar(@dsids); ++$ii) {
            return $filers[$ii] if ($dsid == $dsids[$ii]);
        }
    }

    $_ = $coral->call_vldbtest(command=>"dsidlookup -d $dsid");
    return undef unless (/dbladeid:\s*(\S+)\b/i);
    my $dbladeid = $1;

    $_ = $coral->call_vldbtest(command=>"getbladeinfo -b $dbladeid");
    return undef unless (/hostName:\s*(\S+)\b/i);
    return $1;
}

###############################################################################
#
# my $filername = $coral->find_filer_by_aggregate(aggregate=>"aggr1s");
#
# ...returns the name of the filer currently hosting the specified aggregate.
# If it's a striped aggregate, returns the filer hosting its first constituent.
#
###############################################################################

sub find_filer_by_aggregate
{
    my ($coral, %args) = @_;
    my $aggregate = $args{aggregate};
    my $aggregate_000 = $aggregate . "_000";

    $_ = $coral->call_vldbtest(command=>"dump -a");

    my $dbladeid = undef;
    $dbladeid = $1 if (/aggrName: $aggregate\b.*bladeID: (\S+)/i);
    $dbladeid = $1 if (/aggrName: $aggregate_000\b.*bladeID: (\S+)/i);

    return undef unless defined $dbladeid;

    $_ = $coral->call_vldbtest(command=>"getbladeinfo -b $dbladeid");
    return undef unless (/hostName:\s*(\S+)\b/i);
    return $1;
}

###############################################################################
#
# my $volname = $coral->find_mroot_dbladename(server=>$server);
#
# ...returns the "UUID_12391238903218239_mroot" string used by the D-blade
# as a volume name, for this server's mroot volume.
#
###############################################################################

sub find_mroot_dbladename
{
    my ($coral, %args) = @_;
    my $server = $args{server};
    my $timeout = $args{timeout};  $timeout = 60 unless defined $timeout;

    $_ = $coral{server}->shell(
                    cmd=>"mount",
                    Server=>$server,
                    "-Timeout" => $timeout);

    $coral->die_fatal("unable to find mroot on $server:\n$_")
        unless /\S+:0.0,(0x\S+) on \/mroot/;

    return $coral->find_volume_dbladename_by_dsid (server=>$server, dsid=>hex $1);
}

###############################################################################
#
# my $volname = $coral->find_volume_dbladename_by_dsid(dsid=>$dsid);
#
# ...returns the "UUID_12391238903218239_volname" string used by the D-blade
# as a volume name, given an input dsid.
#
###############################################################################

sub find_volume_dbladename_by_dsid
{
    my ($coral, %args) = @_;
    my $dsid = $args{dsid};
    my $server = $args{server};

    # Which filer does this volume belong to?  That's the filer we'll have to
    # ask in order to find out the name.
    $server = $coral->find_filer_by_dsid(dsid=>$dsid) unless defined $server;

    # Grab the name with a little zapi call
    my $dsid_16 = sprintf("0x%08X", $dsid);
    $_ = $coral->call_zsmcli(server=>$server, command=>"d-volume-list-info id=$dsid_16 desired-attrs=name");

    $coral->die_fatal("unable to find d-blade volname for dsid $dsid")
        unless /<name>(\S+)<\/name>/;

    return $1;
}

###############################################################################
#
# my $msid = $coral->find_volume_msid_by_name(volname=>"coral_unit_test");
#
# ...returns a decimal-encoded msid for the specified volume, if there's a
# vldb entry for it at all.
#
###############################################################################

sub find_volume_msid_by_name
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $volname = $args{volname};

    $_ = $coral->call_vldbtest(command=>"dump -v");
    return undef unless /\bmsid:\s*(\S+)\b.*name:\s+$volname\b/;

    return $1;
}

###############################################################################
#
# my $dsid = $coral->find_volume_dsid_by_name(volname=>"coral_unit_test");
#
# ...returns a decimal-encoded dsid for the specified volume, if there's a
# vldb entry for it at all.  If it's a coral set, returns the dsid of the mdv.
#
###############################################################################

sub find_volume_dsid_by_name
{
    my ($coral, %args) = (shift, %{$coral{options}->get_hash()}, @_);
    my $volname = $args{volname};

    $_ = $coral->call_vldbtest(command=>"dump -v");

    return undef unless /\bdsid:\s*(\d+).*name:\s+$volname\b/;
    return $1;
}

###############################################################################
#
# $coral->runWafliron([timeout=>300],
#                     [filer=>filername],
#                     [wafl_scan_speed=>speed]);
# runWafliron - Invokes wafliron on an aggregate and waits for it to finish
#
#    This routine will run against flex and coral volumes. It might make more
#     sense to move the non-striped support into a different module.
#
###############################################################################

sub runWafliron
{
    my $coral = shift;
    my %args = (timeout=>900, @_);
    my $timeout = $args{timeout};
    my $filer = $args{filer};
    my $wafl_scan_speed = $args{wafl_scan_speed};

    # the orginal code only worked on a single aggregate. Setup the array
    # to track a single aggregate or a list of multiple aggregates
    my @aggregates = ();
    if (ref($args{aggregate}) eq "ARRAY") {
        @aggregates = @{$args{aggregate}};
    } else {
        @aggregates = ($args{aggregate});
    }
    my %errors;

    # find the current wafl_scan_speed
    my $original_wafl_scan_speed;
    if ($wafl_scan_speed) {
        my $out = $coral->call_dbladecli(
                                         server => $filer,
                                         command => "priv set diag ; wafl scan speed");
        if ($out =~ /WAFL scan speed is (\d+)/) {
            $original_wafl_scan_speed = $1;
        }
        $coral{trace}->trace(msg=>"Found the original wafl scan speed to be '$original_wafl_scan_speed'");
        $coral{trace}->trace(msg=>"Setting the new wafl scan speed to be $wafl_scan_speed");
        $coral->call_dbladecli(
                               server => $filer,
                               command => "priv set diag ; wafl scan speed $wafl_scan_speed");
    }

    # start wafliron on the aggregates
    foreach my $aggr (@aggregates) {
        my $aggregate = (ref($aggr) eq "HASH") ? $aggr->{aggregate} : $aggr;
        if (!$filer) {
            $filer = $coral->find_filer_by_aggregate (aggregate=>$aggregate);
        }
        $coral{trace}->comment(msg=>"starting wafliron on $aggregate (filer $filer)\n");

        $errors{$aggregate} = $coral->call_dbladecli(
                                                     timeout => 180,
                                                     server => $filer,
                                                     command => "priv set diag ; aggr wafliron start $aggregate");
    }

    # wait on the irons to finish
    my $stoptime = time() + $timeout;
    my $iron_seen = undef;
    while (time() < $stoptime)
    {
        my $output = $coral->call_dbladecli(
                                            server => $filer,
                                            command => "priv set diag ; aggr wafliron status");
        my $iron = undef;
        foreach my $line (split(/\n/, $output)) {
            if ($line =~ /wafliron is active on aggregate:? (\S+)/i) {
                my $aggregate = $1;
                $errors{$aggregate} = undef;
                $iron_seen = 1;
                $iron = 1;
            }
        }
        last if (!$iron);
        sleep 15;
    }

    if ($original_wafl_scan_speed) {
        $coral{trace}->trace(msg=>"Setting the wafl scan speed back to the original value of '$original_wafl_scan_speed'");
        $coral->call_dbladecli(
                               server => $filer,
                               command => "priv set diag ; wafl scan speed $original_wafl_scan_speed");
    }

    my $errors = "";
    foreach my $error (values %errors) {
        $errors .= "$error\n" if (defined($error));
    }
    my $aggregate = join(",", @aggregates);

    if (!$iron_seen) {
        $coral->die_fatal("Apparantly, no wafliron was started on $filer");
    } elsif ($errors) {
        $coral->die_fatal("wafliron failed: $errors");
    } elsif (time() >= $stoptime) {
        $coral->die_fatal("wafliron on aggregates '$aggregate' took too long");
    }
}

###############################################################################
#
# $coral->run_parallel_waflirzon([timeout=>600],
#                     aggregate=>[aggr1,aggr2],
#                     [wafl_scan_speed=>speed]);
# Invokes wafliron on a list of aggregates in parallel
#
###############################################################################
sub run_parallel_wafliron
{
    my $coral = shift;
    my %args = (
                aggregate => undef,
                timeout => 600,
                wafl_scan_speed => undef,
                @_);

    $coral{trace}->entry(msg=>"@{$args{aggregate}}, $args{timeout}");

    my %members;

    foreach my $aggregate (@{$args{aggregate}}) {
        my $rc = $coral{server}->storage_aggregate_show(aggregate => $aggregate);
        my $aggr_hash = $rc->[0];

        $coral{trace}->trace(msg=>"$aggregate is of type \'$aggr_hash->{'volume-style'}\'");
        if ($aggr_hash->{'volume-style'} eq "flex") {
            my $node = $aggr_hash->{nodes};
            $coral{trace}->trace(msg=>"$aggregate is on node '$node'");
            $members{$node} = [] if (!$members{$node});
            push(@{$members{$node}}, $aggregate);
        } else {
            $coral{trace}->trace(msg=>"find the member aggregates of $aggregate");
            my $members = $coral{server}->storage_aggregate_member_show(aggregate=>"${aggregate}_*");
            foreach my $member_hash (@$members) {
                my $node = $member_hash->{node};
                my $aggregate = $member_hash->{aggregate};
                $coral{trace}->trace(msg=>"$aggregate is on node '$node'");
                $members{$node} = [] if (!$members{$node});
                push(@{$members{$node}}, $aggregate);
            }
        }
    }

    foreach my $node (keys %members) {
        $coral{trace}->trace(msg=>"start the wafliron subprocess on $node");
        main::subtest(sub {$coral->runWafliron(timeout => $args{timeout},
                                               wafl_scan_speed => $args{wafl_scan_speed},
                                               aggregate => $members{$node},
                                               filer => $node)},
                      -runid, $node,
                      -bg,
                      -onexit, sub {});
    }
    Subtest::wait_finish(timeout=>$args{timeout});
    my @subtests = Subtest::get_all();
    my @results = map({$_->results()} @subtests);
    my $worst_result = NATE::Result::worst(@results);
    if ($worst_result && $worst_result->type ne NATE::Result::Pass->type()) {
        die $worst_result;
    }
}
1;


=pod

=head1 NAME

Coral.pm - Some pod should go here

=cut
