
#  $Id: //depot/prod/test/Rsierranevada/lib/Corruption.pm#1 $

## @summary Mini-framework for automating corruption tests
## @author dl-johny@netapp.com (FSS WAFL Quality)
## @cf     false
## @status public
## @pod here
## @Change History
##      Santhosh Paul /  12-Jun-2010 / sanpaul@netapp.com   [BURT 408462] 

=head1 NAME

 Corruption - Mini-framework for automating corruption tests

=head1 SYNOPSIS

 test_case_1.thpl:

 use Corruption;

 my $name = "unique_test_case_name";
 my $params = Corruption::get_params($name, "aggr", "flex", "clone", "trad");
 my $aggr = $$params{aggr};
 my $flex = $$params{flex};
 my $clone = $$params{clone};
 my $trad = $$params{trad};

 Corruption::test_case($name, {

    create_resources => sub {
        Corruption::new_flex($flex)
            or die("new_flex: $@");
        Corruption::new_clone($clone, $flex)
            or die("new_clone: $@");
        Corruption::new_trad($trad)
            or die("new_trad: $@");
        create_foobar($whatever);
    },

    destroy_resources => sub {
        destroy_foobar($whatever);
    },

    inject_script => {
        type => 'flex'
        vol => $flex,
        lines => "inject_command arg=val ...",
        comment => "Changing foobar to baz"
    },

    check_wack_output => sub {
        my $outref = shift;
        if ($$outref =~ /broken foobar, correcting/) {
            logresult("pass", "$name: Wack output " .
                  "references corrupted foobar");
        } else {
            logresult("fail", "$name: Wack output does " .
                  "not reference corrupted foobar");
        }
    },

    check_iron_output => sub {
        my $outref = shift;
        if ($$outref =~ /broken foobar, correcting/) {
            logresult("pass", "$name: Iron output " .
                  "references corrupted foobar");
        } else {
            logresult("fail", "$name: Iron output does " .
                  "not reference corrupted foobar");
        }
    },

    check_data => sub {
        if (foobar_looks_ok()) {
            logresult("pass", "$name: foobar is ok");
        } else {
            logresult("fail", "$name: foobar is not ok");
        }
    }

    });

 batch_1.thpl:

 use Corruption;
 Corruption::batch(
 "$::ONTAPDIR/path/test_case_1.thpl",
 "$::ONTAPDIR/path/test_case_2.thpl",
 "$::ONTAPDIR/path/batch_2.thpl",
 "$::ONTAPDIR/path/batch_3.thpl"
 );

 Command line:

 ntest
    [MODE=<combined|separate|corrupt_only|corrupt_one>]
    [METHODS=[<wack|iron>[,<wack|iron>][;..]]]
    [RANDOM_SEED=<number>]
    [PSEUDO_BAD=1]
    <batch.thpl|test_case_1.thpl|..>


=head1 DESCRIPTION

This package is a mini-library designed to support the writing and
execution of test cases whose purpose is to introduce corruption into
WAFL filesystems and test whether or not that corruption is fixed
by running WAFL_check and wafliron.

=head1 TEST EXECUTION

In general, this framework operates by creating resources; corrupting them;
using WAFL_check or wafliron to fix them; verifying that it found corruption
and that its output references the specific corruption that was applied;
running WAFL_check or wafliron a second time; verifying that it found no
corruption the second time (evidence that it successfully fixed it all the
first time); verifying that the resources look right after being repaired;
repeating this entire process with various different mixes of repair methods,
each time with a freshly re-created and re-corrupted set of resources;
and finally destroying the resources before exiting.

Some variations in this process can be induced through the use of the
MODE, METHODS, and RANDOM_SEED options.  If omitted, they are assigned
default values.

=over 4

=item MODE

combined (default) - All test cases being executed are organized into
phases, with all resources being created at once, then all corruption
applied at once, etc. with wack/iron checking performed only once on
the entire batch of resources/corruption.

separate - If more than one test case is being executed, they are
all executed one after the other as if they had been executed individually
with many different invocations of the ntest command.

corrupt_many - Resources are created and corrupted, but are not wacked,
ironed, or destroyed.  ntest exits with the resources still present and
still in a corrupted state.  In effect, this mode is a tool for generating
corrupted resources.

corrupt_one - Like corrupt_many, except that all corruption from all the
test cases is combined into just one volume.  This nontrivial mode is not
fully implemented, and it may take some thought and further development to
complete it.

=item METHODS

This string determines the sequence of repair methods that are used.
The format of this value is a semicolon-delimited list of zero or more
"sets," each of which is a comma-delimited list of one or two "methods,"
each of which is the string "wack" or the string "iron".

Each set is run against freshly re-created and re-corrupted resources and
involves using the set's first method to attempt to fix the corruption
(pass if corruption is found), then using the set's second method (if any)
to verify that the corruption was fixed (pass if no corruption is found).

For example, the default value of this option is

 wack,wack;wack,iron;iron,wack;iron,iron

This value tests all four permutations of the two available methods
through the following series of actions:

=over 4

=item Purge and re-create resources, and corrupt them

=item Run wack to fix corruption

=item Run wack to verify that corruption was fixed

=item Purge and re-create resources, and corrupt them

=item Run wack to fix corruption

=item Run iron to verify that corruption was fixed

=item Purge and re-create resources, and corrupt them

=item Run iron to fix corruption

=item Run wack to verify that corruption was fixed

=item Purge and re-create resources, and corrupt them

=item Run iron to fix corruption

=item Run iron to verify that corruption was fixed

=back

If this value is used

 METHODS=wack;iron;wack,iron

Then these actions will be performed:

=over 4

=item Purge and re-create resources, and corrupt them

=item Run wack to fix corruption

=item Purge and re-create resources, and corrupt them

=item Run iron to fix corruption

=item Purge and re-create resources, and corrupt them

=item Run wack to fix corruption

=item Run iron to verify that corruption was fixed

=back

If this value is used

 METHODS=ioc;iron;iron,ioc

Then these actions will be performed:

=over 4

=item Purge and re-create resources, and corrupt them

=item ioc to fix corruption

=item Purge and re-create resources, and corrupt them

=item Run iron to fix corruption

=item Purge and re-create resources, and corrupt them

=item Run iron to fix corruption

=item Run ioc to verify that corruption was fixed.If Corruption found, perform run_reject

=back

=item RANDOM_SEED

This number is used to seed perl's random number generator when
Corruption.pm is first loaded; immediately before eval'ing each test script
during a call to batch(); and immediately before calling each
test-case-supplied function reference.

This means that the sequence of random numbers generated by a given test
case's calls to perl's rand() function will be the same during the initial
evaluation of the test script code and during each subsequent call to a
test-script-supplied function.

This also makes it possible to exactly repeat a previous run of a test
case or batch involving random numbers, by manually supplying the
RANDOM_SEED value read from the previous run's output.

If not supplied by the user, RANDOM_SEED defaults initially to a
random number generated by perl's rand(), and then remains
constant through the rest of the test execution process.

=back

=item PSEUDO_BAD

This flag if set to 1, will call 'wreck' with -pseudo_bad flag. It means
along with the corruption script being executed, the block that was read
by wreck will be marked pseudo bad on disk. Note that pseudo bad tagging
on blocks only works on block checksummed disks with parity initialized. 

If not supplied by the user, defaults to value 0. Wreck will not be called
with -pseudo_bad flag in that case.
 
=back

=head1 TEST CASES

Each individual test case is represented by a .thpl file that has a unique
$name string, calls get_params($name, ..) to be assigned resource names
to work with, and then calls test_case($name, {}), passing a big hash of
callback functions and corruption instructions that the corruption
framework will use to operate the test case.  Some of the hash's
members are optional and can simply be left out.

=over 4

=item create_resources

A mandatory reference to a function creating the resources that the test
case will later corrupt.  Failures in this process should be reported with
die(), since the test cannot continue unless this phase succeeds.

=item destroy_resources

An optional reference to a corresponding function destroying those resources.
This is usually omitted, since the framework automatically creates and
destroys the parent aggregate, and automatically destroys any tradvols
that were created with the new_trad() function.

=item inject_script

This mandatory hash reference describes an injection script that will be passed
to the 'wreck' command in order to corrupt the resources.

=over 4

=item type

The type of the volume on which 'wreck' should be run -
'aggr', 'flex', or 'trad'.

=item vol

The name of the volume on which 'wreck' should be run.

=item lines

The text of the injection script that should be passed to 'wreck'.

=item comment

A comment explaining the corruption that is created thereby.

=back

Instead of a single hash reference, you may optionally supply a reference to
a list of two or more such hash references, or (if the inject script data must
be generated based on dynamic data after create_resources() has already been
called) a reference to a function that returns such a hash or list reference.

If a function reference is supplied, it will be called
anew, refreshing the inject script data, after each call to
create_resources().

=item check_wack_output

An optional reference to a function that receives a reference to WAFL_check
output and should scan it to see if the corruption was detected by WAFL_check,
calling logresult("pass", "msg") if it was or logresult("fail", "msg") if it
was not.

=item check_iron_output

An optional reference to a similar function for checking wafliron output.

=item check_data

An optional reference to a function that examines the resources directly to
determine whether or not the corruption was fixed.  Depending on how the
intended fixing is supposed to work, the expected appearance of the repaired
resources may be different from how they appeared before they were corrupted.
Like the check_output functions, this one should call logresult("pass",
"msg") if the resources look right or logresult("fail", "msg") if they
do not.

=back

If such a test
case .thpl file is individually executed by ntest, the call to test_case()
will not only register the test case with the framework but will also
cause it to be executed.

=head1 TEST BATCHES

A .thpl file representing a batch of test cases should make a single
call to batch(), passing in a list of absolute paths of one or more
child .thpl files.  Each child .thpl file can be either a test case or
another batch, in which case all its children are loaded in turn.
The same test case .thpl file may not appear more than once in any
part of a batch.

Once the entire tree of specified files has been loaded,
all the test cases it incorporates are then executed in a way determined
by the MODE and METHODS options.

=head1 FUNCTIONS FOR ORGANIZING TEST CASES

=cut

BEGIN {
    use vars qw($TESTNAME $METHODS $MODE $AGGR_SIZE $NAME_FILE
            $SNAP_FILE $LOST_PREFIX $FILER_CLONE_VOL $MOUNT_OPTS
            $EXPORT_OPTS $SNAPDIR $LOST_FOUND_DIR $ZOMBIE_FILE_SIZE
            $ZOMBIE_FILE_COUNT $SPACE_GUARANTEE $CORRUPTION_AGGR 
                    $OPTIONAL_COMMIT $PSEUDO_BAD $IRON_ON_ROOT $ASIS_ENABLE 
                    $WORM_RETENTION $SNAPLOCK_AGGR);
    param("TESTNAME", "-mesg", "Full path of test script");
    param("METHODS",
        "-default", "ioc,ioc;ioc,iron;iron,ioc;iron,iron");
    param("MODE", "-default", "combined");
    param("AGGR_SIZE", "-default", "3");
    param("NAME_FILE", "-default", "name");
    param("SNAP_FILE", "-default", "snap");
    param("LOST_PREFIX", "-default", 
              "(lost)|(vol_\\w{8}_\\w{4}_\\w{4})|(vol-\\w{8}-\\w{4}-\\w{4})");
    param("FILER_CLONE_VOL", "-default", "clone");
    param("MOUNT_OPTS", "-default", "noac,hard,intr,noquota,rw");
    param("EXPORT_OPTS", "-default", "anon=0,rw");
    param("SNAPDIR", "-default", ".snapshot");
    param("LOST_FOUND_DIR", "-default", "lost+found");
    param("ZOMBIE_FILE_SIZE", "-default", "8000");
    param("ZOMBIE_FILE_COUNT", "-default", "10");
    param("RANDOM_SEED", "-default", int(rand(0xffffffff)));
    param("DESTROY_CORRUPT_VOLUMES", "-default", 1);
    param("SPACE_GUARANTEE", "-default", "none");
    param("OPTIONAL_COMMIT", "-default", "none");
    param("CORRUPTION_AGGR", "-default", "wafl_check_iron_aggr");
    param("PSEUDO_BAD", "-default", 0);
    param("IRON_ON_ROOT", "-default", undef);
    param("ASIS_ENABLE", "-default", undef);           
    ## SNAPLOCK_AGGR is required to determine the snaplock aggregate type
    ## SnapLock Compliance or SnapLock Enterprise.
    param("SNAPLOCK_AGGR", "-default", undef);
    ## WORM_RETENTION is required to lock the snaplock file.
    param("WORM_RETENTION", "-default", "180s" );
    require "FS/commonParams.pl";
};


package Corruption;

use strict;

use constant BOOT_TIMEOUT => 2400;

use vars ('@ISA','@EXPORT');

require Exporter;
@ISA = ('Exporter');
@EXPORT = (
    'get_params',
    'test_case',
    'batch',
    'new_flex',
    'new_clone',
    'find_lost_flex',
    'mark_flex_for_recovery',
    'vol_is_online',
    'vol_exists',
    'vol_is_unrecoverable',
    'mount_volume',
    'list_entries',
    'make_file',
    'check_file',
    'make_file_nfs',
    'check_file_nfs',
    'remove_file',
    'inode_number',
    'find_lost_file',
    'make_dir',
    'make_snapshots',
    'list_snapshots',
    'check_snapshot',
    'populate_zombie_test_files', 
    'remove_zombie_test_files', 
    'sfsr_zombie_test_files',
    'add_license',
    'fiji_add_filter',
    'fiji_clear_filters',
    'run_floppy_boot_cmd'
);

use Carp qw(croak);
use Tharn;
use boot;
use Aggregate;
use Volume;
use Volume::Clone;
use Wreck;
use wack;
use CLI::FileIO qw(rdfile wrfile);
use Admin::License qw(license is_licensed);
use mount;
use ExportFS;
use SnapCmds;
use FS::Util::Utilities qw(write_file);
use FS::Util::Setup qw(mount_vol umount_vol);
use FS::Util::Setup qw(create_vol_and_mount);
use filer qw(filer_options_get filer_options_set);
use Wafliron qw(wafliron_start wafliron_runcommit wafliron_runreject 
                wafliron_on_root wafliron_pick_new_root 
                wafliron_create_temp_root);
use Dense qw(dense_toggle dense_status dense_start);
use compression qw(enable_compression_option);
use Filer::Version qw(get_wafl_version
                      WV_BOILERMAKER_MIN 
              WV_BOILERMAKER_MAX );
use Data::Dumper;
use Utils qw(make_conn);
use snaplock qw(
    compliance_clock_initialize
    COMPLIANCE_CLOCK_ALREADY_SET
    set_default_retention_period
);
use SPARSE::Util::Logger;

use constant IRON_CLEAN     => 'CLEAN';
use constant IRON_DIRTY     => 'DIRTY';
use constant IRON_NOTDONE   => 'NOTDONE';
use constant IRON_COREDUMP  => 'CORE'; 
use constant IRON_UNEXPECTED=> 'UNEXPECTED';

# ----------------------------------------------------------------------
# Function Declaration
# ----------------------------------------------------------------------
sub get_injects($);

# Core API for organizing corruption test execution

# Host, console, and rsh
# Console is used for console-only stuff (like wacking at the boot menu).
# Rsh is used for everything else.

$Corruption::OUR_FILER_AGGR = $::CORRUPTION_AGGR;
$Corruption::OUR_TEMP_AGGR  = $::CORRUPTION_AGGR."_temp";
$Corruption::OUR_TEMP_VOL   = "vol0_temp";

# Keep track of tree level so we know when to execute our accumulated
# group of test cases.
#
# Each call to batch() increments script_level at the beginning and
# decrements it at the end.
#
# If batch() exits with script_level set to 0, that means it was the
# top-level call to batch(), so the batch is considered complete and is
# executed with run().  Otherwise it's considered to be a sub-batch and
# the test cases are not executed yet.
#
# If test_case() is called with script_level set to 0, it's not being
# called by batch(), so it must be a standalone execution of a test case
# and is executed immediately with run().

$Corruption::script_level = 0;

# Hash and list of all the test cases.  A case (often referred to in
# singular by scalar $case or in groups by list ref $cases) is one of
# the anonymous hash refs passed as the second parameter of test_case()
# calls.

%Corruption::cases = ();
@Corruption::cases = ();
$Corruption::vol_mounts = {};

%Corruption::result = ();
$Corruption::result{inject_str} = "";
$Corruption::result{fix} = "";
$Corruption::result{verify} = "";
$Corruption::result{fixresult} = "";
$Corruption::result{verifyresult} = "";

# List of all trad vols that were created by new_trad().
# At the end, we automatically remember to destroy them.

@Corruption::trads = ();


# Apply the random seed at least once before evaluating any test script code.

logcomment("Corruption::Random seed: $::RANDOM_SEED");
srand($::RANDOM_SEED);


# Build a display string composed of the names of a set of test cases
# separated by newlines.

sub
get_name_str($) {
    my $cases = shift;
    if (scalar(@$cases) == 1) {
        my $case = $$cases[0];
        return $$case{name};
    } else {
        my $names = "";
        for my $case(@$cases) {
            $names .= $$case{name} . "\n";
        }
        return $names;
    }
}

#  Print (returns string) contents of "one-level" hashtable.
#  -- Cheapo version of Data::Dumper
sub print_hash_1($) {
    my ($hashref) = @_;
    my $str = "(\n";
    while ( my ($vol,$mnt) = each(%$hashref) ) {
        $str .= "    ".sprintf("%-30s", $vol)."=> $mnt\n";
    }
    $str .= ")\n";
    return($str);
}

#  Returns string rep. of 'vol_mounts' -- all mounted vols.
sub print_vol_mounts() {
    return( print_hash_1($Corruption::vol_mounts) );
}

sub vol_mount_aux($) {
    my ($vol) = @_;
    return eval {
        # logcomment("Corruption::vol_mount_aux:  Mounting volume \"$vol\"");
        my $mnt = mount_vol(filer => $Corruption::rsh, vol => $vol);
        if ($@ || !$mnt) {
            croak("Corruption::vol_mount_aux:  mount_vol $vol  Error: $@");
        }
        $Corruption::vol_mounts->{$vol} = $mnt;
        # logcomment("Corruption::vol_mount_aux:   vol_mounts:\n\n".print_vol_mounts());
        return($mnt);
    };
}

#  Unmounts given vol
sub unmount_vol($) {
    my ($vol) = @_;
    my $mnt = $Corruption::vol_mounts->{$vol};
    return eval {
    if (!$mnt) {
        # croak("Corruption::unmount_vol:  No mount found for volume $vol");
        # logcomment("Corruption::unmount_vol:  No mount found for volume $vol");
        return(0);
    }
    my $stat = umount_vol($mnt);
    if ($@ || !$stat) {
        croak("Corruption::unmount_vol:  'umount_vol' $vol  Error: $@");
    }
    };
}

#  Unmounts all mounted vols
sub delete_all_mounts() {
    my $err = 0;
    while ( my ($vol, $mnt) = each(%$Corruption::vol_mounts) ) {
        # logcomment("Corruption::delete_all_mounts:  Unmounting volume $vol");
        my $stat = umount_vol($mnt);
        unless ($stat) {
            logcomment("Corruption::delete_all_mounts:  'umount_vol' $vol  Error: $@");
            $err++;
        }
    }
    $Corruption::vol_mounts = {};
    return ($err ? 0 : 1);
}

# Return a list of all the inject scripts required by a given test case.
# Each returned inject is a hash ref like the inject_script member of a
# case hash.
#
# If the test case has a sub as its inject_script field, the random seed
# is applied before the sub is called.

sub
get_injects($) {
    my ($ref) = @_;
    if ($ref =~ /^HASH/) {
        return ($ref);
    } elsif ($ref =~ /^ARRAY/) {
        return (@$ref);
    } elsif ($ref =~ /^CODE/) {
        srand($::RANDOM_SEED);
        my $ref1 = &$ref();
        return get_injects($ref1);
    } else {
        die("Corruption::get_injects - inject_script must be ref to hash, array, or sub: $ref");
    }
}


# Blend together the injects required by a group of test cases
# and return them within a single hash ref, keyed by volume name.

sub
combine_injects($) {
    my $cases = shift;
    my $injects = {};

    for my $case(@$cases) {
        my $ref = $$case{inject_script};
        my @injs = get_injects($ref);
        for my $inj (@injs) {
            my $inj2 = $$injects{$$inj{vol}};
            if (defined($inj2)) {
                if ($$inj2{type} ne $$inj{type}) {
                    my $names = get_name_str($$inj2{cases});
                    logresult("fatal", 
                        "Corruption::combine_injects - ".
                        "Type conflict for volume " .
                        "$$inj{vol}: It is given " .
                        "type $$inj{type} by " .
                        "$$case{name}, and type " .
                        "$$inj2{type} by $names");
                }
                $$inj2{lines} .= "\n" . $$inj{lines} . "\n";
                $$inj2{comment} .= "$$case{name}: $$inj{comment}\n";
                my $tmp = $$inj2{cases};
                push(@$tmp, $case);
            } else {
                $$injects{$$inj{vol}} = {
                        type => $$inj{type},
                        vol => $$inj{vol},
                        lines => $$inj{lines},
                        comment => "$$case{name}: $$inj{comment}\n",
                        cases => [$case]
                };
            }
        }
    }
    return $injects;
}


# Add a new case to the internal hash and list.
# Test it to make sure it has the necessary parts.

sub
add_case($$) {
    my ($name, $case) = @_;

    $$case{name} = $name;

    if ($Corruption::cases{$name}) {
        logresult("fatal", "Corruption::add_case - Duplicate test case name: $name");
    }

    if (!defined($$case{create_resources})) {
        logresult("fatal", "Corruption::add_case - Missing create_resources: $name");
    }

    if (!defined($$case{inject_script})) {
        logresult("fatal", "Corruption::add_case - Missing inject_script: $name");
    }

    $Corruption::cases{$name} = $case;
    push(@Corruption::cases, $case);
}


# Do initial stuff to get the specified host into a usable state.

sub
init() {
    my $timeout = 2200;
    logcomment("
        FILER              =    $::FILER 
        METHODS            =    $::METHODS 
        OPTIONAL COMMIT    =    $::OPTIONAL_COMMIT");

    logcomment("Initializing and connecting to host: $::FILER");
    logcomment("Corruption::init - Initializing and connecting to host: $::FILER");
    $Corruption::host = host($::FILER);
    $Corruption::con = connect("#console", -timeout, $timeout);
    $Corruption::rsh = connect("#rsh", -timeout, $timeout);
    if ($Corruption::host->hosttype =~ /simulator/) {
        $Corruption::rsh->timeout(300);
    } else {
        $Corruption::rsh->timeout(1900);
    }

    $Corruption::rshsocket = connect("#rsh-socket", -timeout, $timeout);
    my $boot_timeout = $main::BOOT_TIMEOUT || BOOT_TIMEOUT;
    transit_to($Corruption::host, UP, prefix=>"FILER", timeout => $boot_timeout)
        or die("Corruption::init - transit_to: $@");
        
    # If METHODS contains WACK & Ontap Version is > BR, then FAIL the 
    #   test case because WACK is not supported POST-BR
    my $ver = get_wafl_version(host($::FILER));
    if (($::METHODS =~/wack/) && ($ver > WV_BOILERMAKER_MAX)) {
        logresult("fatal", "Corruption::init - WACK is not supported in this Ontap Release ");
    } else {
        logcomment("Corruption::init - Confirmed WAFL version & WACK Compatibility");
    }
        
    Admin::License::license($Corruption::rsh, "nfs")
                or die("Corruption::init - Can't license nfs: $@");
    # Remember the original root aggr and root volume
    $Corruption::ORIG_ROOT_VOLUME =Volume::vol_root($Corruption::rsh);
    $Corruption::ORIG_ROOT_AGGR = Aggregate::aggr_root($Corruption::rsh);

    # Check & Configure the HTTPD options of the Filer
    _check_and_configure_httpd();

} # END of sub()


# Apply the random seed, then call create_resources() of a single test case.
# If it fails, it's supposed to die() inside, killing the whole test.

sub
create_one($) {
    my $case = shift;
    my $sub = $$case{'create_resources'};
    srand($::RANDOM_SEED);
    &$sub();
}

sub 
check_aggr_restrict()
{
    my $status_obj = Aggregate::aggr_status($Corruption::rsh, $Corruption::OUR_FILER_AGGR);
    if ($status_obj->state() =~ /restricted/) {
        # should not bring an inconsistent aggr online
        if ($status_obj->state() =~ /restricted/) {
            Aggregate::aggr_offline($Corruption::rsh, $Corruption::OUR_FILER_AGGR, force => 1, all => 1) or
            logresult("CONFIG","Corruption::check_aggr_restrict - $@");
            Aggregate::aggr_purge($Corruption::rsh, $Corruption::OUR_FILER_AGGR)
                       or logresult("CONFIG","Corruption::check_aggr_restrict -  $@");
            logcomment("Corruption::check_aggr_restrict - Creating base aggregate: $Corruption::OUR_FILER_AGGR");
            ## Create snaplock aggregate if SNAPLOCK_AGGR is specified.
            if ( $::SNAPLOCK_AGGR eq 'snaplock' or $::SNAPLOCK_AGGR eq 'snaplock_enterprise' ) {
                    Aggregate::aggr_create($Corruption::rsh, $Corruption::OUR_FILER_AGGR, $::AGGR_SIZE , $::SNAPLOCK_AGGR => 1 ) or logresult("CONFIG","Corruption::check_aggr_restrict -  $@");
            } else {
                    Aggregate::aggr_create($Corruption::rsh, $Corruption::OUR_FILER_AGGR, $::AGGR_SIZE) or logresult("CONFIG","Corruption::check_aggr_restrict -  $@");
            }
        }
    }
}

sub
check_aggr_ironing($)
{
    my ($recreate) = @_;

    my(undef,undef,$output) = say($Corruption::rsh,
        "priv set -q test; aggr wafliron status $Corruption::OUR_FILER_AGGR");
    if ($output =~ /is\s+active/) {
        say($Corruption::rsh, "priv set -q test; aggr wafliron stop $Corruption::OUR_FILER_AGGR");
        my $endt = time() + 1200;
        while ((time() < $endt) and ($output =~ /is\s+active/)) {
            sleep(3);
            (undef,undef,$output) = say($Corruption::rsh,
                "priv set -q test; aggr wafliron status $Corruption::OUR_FILER_AGGR");
        }
        Aggregate::aggr_purge($Corruption::rsh, $Corruption::OUR_FILER_AGGR) 
                   or logresult("CONFIG","Corruption::check_aggr_ironing - $@");
        if ($recreate == 1) {
            logcomment("Corruption::check_aggr_ironing - Creating base aggregate: $Corruption::OUR_FILER_AGGR");
            ## Create snaplock aggregate if SNAPLOCK_AGGR specified.
            if ( $::SNAPLOCK_AGGR eq 'snaplock' or $::SNAPLOCK_AGGR eq 'snaplock_enterprise'  ) {
               Aggregate::aggr_create($Corruption::rsh, $Corruption::OUR_FILER_AGGR, $::AGGR_SIZE , $::SNAPLOCK_AGGR => 1 ) or logresult("CONFIG","Corruption::check_aggr_ironing - $@");
            } else {
               Aggregate::aggr_create($Corruption::rsh, $Corruption::OUR_FILER_AGGR, $::AGGR_SIZE) or logresult("CONFIG","Corruption::check_aggr_ironing - $@");
            }
        }
    }
}

# Create framework-level resources like the base aggr.
# Then call create_resources() of a group of test cases.
# If any fail, they're supposed to die() inside, killing the whole test.

sub
create($) {
    my $cases = shift;
    my %aggr_args;
    logcomment("Corruption::create - Creating resources");
    if (!Aggregate::aggr_exists($Corruption::rsh, 
        $Corruption::OUR_FILER_AGGR)) {
        logcomment("Corruption::create - Creating base aggregate: $Corruption::OUR_FILER_AGGR");
        ## Create snaplock aggregate if SNAPLOCK_AGGR specified.
        if ( $::SNAPLOCK_AGGR eq 'snaplock' or $::SNAPLOCK_AGGR eq 'snaplock_enterprise' ) {
            ## Make sure that snaplock/snaplock_enterprise license exists.
            if ( $::SNAPLOCK_AGGR eq 'snaplock' ) {
                license($Corruption::rsh, 'snaplock');
            } elsif ( $::SNAPLOCK_AGGR eq 'snaplock_enterprise' ) {
                license($Corruption::rsh, 'snaplock_enterprise' );
            }
            ## Initialize compliance clock if not initialized before.
            logcomment("Corruption::create - compliance clock initialize");
            my ($cc_time, $did_init) = eval { return compliance_clock_initialize($Corruption::rsh, strict => 1); };
            if ($@ and (index($@, COMPLIANCE_CLOCK_ALREADY_SET) == -1)) {
                # die if the error is not COMPLIANCE_CLOCK_ALREADY_SET
                die $@;
            }
            Aggregate::aggr_create($Corruption::rsh, $Corruption::OUR_FILER_AGGR, $::AGGR_SIZE ,  $::SNAPLOCK_AGGR => 1 ) or logresult("CONFIG","Corruption::create - $@");
        } else {
            Aggregate::aggr_create($Corruption::rsh, $Corruption::OUR_FILER_AGGR, $::AGGR_SIZE) or logresult("CONFIG","Corruption::create - $@");
        } 
    } else {
        logcomment("Corruption::create - Aggregate $Corruption::OUR_FILER_AGGR exists");
        check_aggr_ironing(1);
        check_aggr_restrict();
    }

    if ($::MODE eq 'corrupt_one') {
        logcomment("Corruption::create - Creating target flexvol: $::FILER_VOL");
        if(Volume::vol_exists($Corruption::rsh, $::FILER_VOL)) {
            Volume::vol_purge($Corruption::rsh, $::FILER_VOL);
        }
        Volume::vol_create_flex($Corruption::rsh, $::FILER_VOL,
                $Corruption::OUR_FILER_AGGR, $::VOL_SIZE,
                {guarantee => $::SPACE_GUARANTEE,skip_volume_special_options_hook => 1});
        asis_enable($::FILER_VOL) if ($::ASIS_ENABLE);
        compression_enable($::FILER_VOL)    if ($::COMPRESSION_ENABLE);                 
        if ($::SNAPLOCK_AGGR eq "snaplock" or $::SNAPLOCK_AGGR eq "snaplock_enterprise" ) {
            set_default_retention_period($Corruption::rsh, $::WORM_RETENTION , $::FILER_VOL ) ;
        }  
        
    } else {
        logcomment("Corruption::create - Telling test cases to create their resources");
        for my $case(@$cases) {
            create_one($case);
        }
    }
}


# If a given test case has destroy_resources(), apply the random seed and
# then call it.

sub
destroy_one($) {
    my $case = shift;
    my $sub = $$case{'destroy_resources'};
    if (defined($sub)) {
        srand($::RANDOM_SEED);
        &$sub();
    }
}


# Call destroy_resources() of all test cases that have it,
# then destroy framework-level resources.
# do not destroy OUR_FILER_AGGR 
sub
destroy($) {
    my $cases = shift;

    logcomment("Corruption::destroy - Purging resources");
    for my $case(@$cases) {
        destroy_one($case);
    }
    delete_all_mounts();

    if (Aggregate::aggr_exists($Corruption::rsh, $Corruption::OUR_FILER_AGGR)) {
        check_aggr_ironing(0);
    }
    if (Aggregate::aggr_exists($Corruption::rsh, $Corruption::OUR_FILER_AGGR)) {
        if ($::CLEANUP eq 'all') {
            logcomment("Corruption::destroy - Purging base aggregate: $Corruption::OUR_FILER_AGGR");
        Aggregate::aggr_purge($Corruption::rsh, $Corruption::OUR_FILER_AGGR);
        } else {
            my $aggr_s = Aggregate::aggr_status($Corruption::rsh,
                        $Corruption::OUR_FILER_AGGR);
            my $vols = $aggr_s->volumes();
            foreach my $vol (@{$vols}) {
                # Do not try to purge vol0
                    next if ($vol eq 'vol0');
                    logcomment("Corruption::destroy - Purging volume $vol");
                    if ( $::SNAPLOCK_AGGR eq "snaplock" ) {
                        ## Online the volume if it is offlined by test.
                        Volume::vol_online($Corruption::rsh, $vol);
                        my @sleep_time = split(/s/, $::WORM_RETENTION ); 
                        logcomment("Corruption::destroy - $vol is SnapLock compliance volume, sleeping $sleep_time[0]+100 seconds for WORM file to get expire\n");
                        sleep($sleep_time[0]+100);
                    }
                Volume::vol_purge($Corruption::rsh, $vol) or logcomment("Corruption::destroy - vol_purge error: $@");
            }
        }
    }

    for my $trad (@Corruption::trads) {
    logcomment("Corruption::destroy - Purging tradvol: $trad");
    if ( $::SNAPLOCK_AGGR eq "snaplock" ) {
            ## Online the volume if it is offlined by test.
            Volume::vol_online($Corruption::rsh, $trad);
            my @sleep_time = split(/s/, $::WORM_RETENTION );
            logcomment("Corruption::destroy - $trad is SnapLock compliance volume, sleeping $sleep_time[0]+100 seconds for WORM file to get expire\n");
            sleep($sleep_time[0]+100);
    }
    Volume::vol_purge($Corruption::rsh, $trad);
    }
}


# Run wreck on one volume.  The argument is an inject hash ref.

sub
run_wreck_vol($) {
    my $inj = shift;
    logcomment("Corruption::run_wreck_vol - Running wreck on: $$inj{type}: $$inj{vol}\n"
           . $$inj{comment});
    # turn off dbg printf messages 
    say($Corruption::rshsocket, "priv set -q diag; dbg to off");  
    say($Corruption::rsh, "priv set -q test; setflag wafl_inconsistent_panic_testing 1");       
    # If we are doing iron_on_root and the corruption type is 'aggr'
    #  then we have to create a temporary root aggr so we
    #  can corrupt the existing root aggr, CORRUPTION_AGGR should be set to 'aggr0'
    if ($::IRON_ON_ROOT) {
        if ($$inj{type} eq 'aggr') {
            logcomment("Corruption::run_wreck_vol - Switching to new root volume because IRON_ON_ROOT and corruption type=aggr");
            wafliron_create_temp_root($Corruption::rsh, aggr=>$Corruption::OUR_TEMP_AGGR, vol=>$Corruption::OUR_TEMP_VOL)or logresult("CONFIG","Corruption::run_wreck_vol - $@");
            # After the create_temp_root all aggrs and volumes are restricted
            # get our aggr back online
            Aggregate::aggr_online($Corruption::rsh, $::CORRUPTION_AGGR);
        }
    }
    wreck($Corruption::rshsocket, $$inj{vol},
          "type" => $$inj{type},
          "inject_script" => $$inj{lines},
          "pseudo_bad" => $::PSEUDO_BAD)
    or logresult("fatal", "Corruption::run_wreck_vol - Wreck failed: $@");

    $Corruption::result{injectstr} = $$inj{lines};
}


# Run wreck with an entire hash full of injects
# (as returned by combine_injects()).

sub
run_wreck($) {
    my ($injects) = @_;
    for my $vol(sort(keys(%$injects))) {
        run_wreck_vol($$injects{$vol});
    }
}


# Clear, create, and corrupt resources as required by a group of test cases.

sub
prepare($) {
    my ($cases) = @_;
    destroy($cases);
    create($cases);
    my $injects = combine_injects($cases);
    run_wreck($injects);
    return $injects;
}


# Apply the random seed, then ask one test case to check wack output.
# It should call logresult("pass") or logresult("fail").

sub
check_wack_one($$) {
    my ($case, $outref) = @_;
    my $sub = $$case{check_wack_output};
    if (defined($sub)) {
        srand($::RANDOM_SEED);
        &$sub($outref);
    }
}


# Ask a group of test cases to check wack output.
# Each should call logresult("pass) or logresult("fail").

sub
check_wack($$) {
    my ($cases, $outref) = @_;
    for my $case(@$cases) {
        check_wack_one($case, $outref);
    }
}


# Apply the random seed, then ask one test case to check iron output.
# It should call logresult("pass") or logresult("fail").

sub
check_iron_one($$) {
    my ($case, $outref) = @_;
    my $sub = $$case{check_iron_output};
    if (defined($sub)) {
        srand($::RANDOM_SEED);
        &$sub($outref);
    }
}


# Ask a group of test cases to check iron output.
# Each should call logresult("pass) or logresult("fail").

sub
check_iron($$) {
    my ($cases, $outref) = @_;
    for my $case(@$cases) {
        check_iron_one($case, $outref);
    }
}


# Apply the random seed, then ask one test case to check actual
# post-repair data.  It should call logresult("pass") or logresult("fail").

sub
check_data_one($) {
    my $case = shift;
    my $sub = $$case{check_data};
    if (defined($sub)) {
        srand($::RANDOM_SEED);
        &$sub();
    }
}


# Bring online the base aggregate and any volumes that would have been
# wrecked and fixed by a group of test cases.  (If it times out doing this,
# look for a core in the simdir from around that time - the sim may have
# crashed trying to bring a volume online after failing to fix its
# corruption.)
#
# Then ask the test cases to check actual post-repair data.
# Each should call logresult("pass") or logresult("fail").

sub
check_data($) {
    my $cases = shift;
    my $boot_timeout = $main::BOOT_TIMEOUT || BOOT_TIMEOUT;
    transit_to($Corruption::con, UP, prefix=>"FILER", timeout => $boot_timeout)
        or die("Corruption::check_data - transit_to: $@");
    Aggregate::aggr_online($Corruption::rsh, $Corruption::OUR_FILER_AGGR);

    for my $case(@$cases) {
        my $ref = $$case{inject_script};
        my @injs = get_injects($ref);
        for my $inj (@injs) {
            my $type = $$inj{type};
            my $vol = $$inj{vol};
            if ($type ne 'aggr') {
                Volume::vol_online($Corruption::rsh, $vol);
            }
        }
        check_data_one($case);
    }
}


# Calculate which volumes need to be wacked, given the corruption caused
# by a group of test cases.  The answer is returned as a space-delimited
# string.

sub
calc_wacks($) {
    my $cases = shift;
    my %wacks = ();

    for my $case (@$cases) {
        my $ref = $$case{inject_script};
        my @injs = get_injects($ref);
        for my $inj (@injs) {
            my $type = $$inj{type};
            my $vol = (($type eq 'flex') ?
            $Corruption::OUR_FILER_AGGR : $$inj{vol});
                $wacks{$vol} = 1;
        }   
    }
    return join(" ", sort(keys(%wacks)));
}


# Attempt to use wack to fix corruption.
# If it doesn't find anything, that counts as a failure.
# Then have a group of test cases check wack's output.

sub
fix_wack($) {
    my $cases = shift;
    my $wacks = calc_wacks($cases);
    my $output;

    logcomment("Corruption::fix_wack - Using wack to fix corruption");

    my $ret = wack::wack($Corruption::con,
            prefix=>"FILER",
            args => $wacks,
            commit => 1,
                        timeout => 1800,
            output => \$output);

    if ($ret eq WACK_DIRTY) {
        $Corruption::result{fixresult} = "PASS";
        logresult("pass", "Corruption::fix_wack - WACK found corruption");
    } else {
        # it's possible that wack didn't get run, try again
        $ret = wack::wack($Corruption::con,
                prefix=>"FILER",
                args => $wacks,
                commit => 1,
                timeout => 1800,
                output => \$output);
        if ($ret eq WACK_DIRTY) {
            $Corruption::result{fixresult} = "PASS";
            logresult("pass", "Corruption::fix_wack - WACK found corruption");
        } else {
            $Corruption::result{fixresult} = "FAIL";
            logresult("fail", "Corruption::fix_wack - WACK did not find corruption");
            my $boot_timeout = $main::BOOT_TIMEOUT || BOOT_TIMEOUT;
            transit_to($Corruption::con, UP, prefix=>"FILER", timeout => $boot_timeout)
                or die("Corruption::fix_wack - transit_to: $@");
        }
    }
    check_wack($cases, \$output);
}


# Attempt to use wack to verify that corruption is gone.
# If it finds any corruption, that counts as a failure.

sub
verify_wack($) {
    my $cases = shift;
    my $wacks = calc_wacks($cases);

    logcomment("Corruption::verify_wack - Using wack to verify that corruption was fixed");

    my $ret = wack::wack($Corruption::con, 
            prefix => "FILER",
            args => $wacks,
            commit => 1);

    if ($ret eq WACK_CLEAN) {
        $Corruption::result{verifyresult} = "PASS";
        logresult("pass", "Corruption::verify_wack - WACK thinks corruption was fixed");
    } else {
        $Corruption::result{verifyresult} = "FAIL";
        logresult("fail", "Corruption::verify_wack - WACK thinks corruption was not fixed");
    }
}


# Calculate which runs of wafliron are required to fix corruption created
# by a group of test cases, and also which cases are involved with each run
# of iron.
#
# The answer is returned as a hash ref keyed by volume name, whose values
# are hash refs whose fields are type (aggr/vol), vol (name), and
# cases (list ref).

sub
calc_irons($) {
    my $injects = shift;
    my $irons = {};

    for my $vol (sort(keys(%$injects))) {
        my $inj = $$injects{$vol};
        my ($iron_type, $iron_vol) = (($$inj{type} eq 'flex') ?
             ('aggr', $Corruption::OUR_FILER_AGGR) : ($$inj{type}, $vol));

        my $tmp = $$irons{$iron_vol};
        if (defined($tmp)) {
            if ($$tmp{type} ne $iron_type) {
                logresult("fatal", "Corruption::calc_irons - Conflicting types for " .
                  "volume $iron_vol: $$tmp{type}: " .
                  "$iron_type");
            }
            my $tmp_cases = $$tmp{cases};
            my $inj_cases = $$inj{cases};
            push(@$tmp_cases, @$inj_cases);
        } else {
            $tmp = {};
            $$tmp{type} = $iron_type;
            $$tmp{vol} = $iron_vol;
            my $inj_cases = $$inj{cases};
            $$tmp{cases} = [@$inj_cases];
            $$irons{$iron_vol} = $tmp;
        }
    }

    return $irons;
}


# Attempt to use iron to fix one volume's corruption.
# If it doesn't find any corruption, that counts as a failure.
# Then tell all test cases related to that volume to check iron's output.

sub
fix_iron_vol($) {
    my $iron = shift;
    my $type = (($$iron{type} eq 'aggr') ? 'aggr' : 'vol');
    my $vol = $$iron{vol};
    my $output = "";
    my $ret;
    my $to;
    my $found = 0;
    my $tmp_root_done = 0;
        
    my $injs = {};
    my $case;
    my $cases = $$iron{cases};
    if (scalar(@$cases) == 1) {
        $case = $$cases[0];
        $injs = $$case{inject_script};
    } else {
        $Corruption::result{fixresult} = "FAIL";
        logresult("fail", "Corruption::fix_iron_vol - Cases is not array of one: \n".Dumper($iron)); 
    }

    if ($Corruption::host->hosttype =~ /simulator/) {
        $to = 2400;
    } else {
        $to = 4800;
    }
        
    logcomment("Corruption::fix_iron_vol - Optional commit =\"$::OPTIONAL_COMMIT\"");
    # IRON_ON_ROOT does not work with trad volumes only flex volumes
    if ($::IRON_ON_ROOT && $$injs{type} ne 'trad') {
        logcomment("Corruption::fix_iron_vol - IRON_ON_ROOT =\"$::IRON_ON_ROOT\"");
        if ($::IRON_ON_ROOT =~ /core/) {
            my ($object, $vol_name);
            # The injection type dictates what volume or aggr we pass to
            #  vol_pick_root
            if ($$injs{type} eq 'aggr') {
                logcomment("Corruption::fix_iron_vol - inject_type is: $$injs{type} vol is: $$iron{vol}");
                my $name_str = get_name_str($cases);
                $vol_name = 'testvol_flex_'.$name_str;
                $object = $$iron{vol};
            } else {
                logcomment("Corruption::fix_iron_vol - inject_type is: $$injs{type} vol is: $$injs{vol}");
                $vol_name = $$injs{vol};
                $object = $$iron{vol};
            }
            # Pick corrupted aggr and volume as new root
            wafliron_pick_new_root($Corruption::con, $object, $vol_name, reboot=> 0, tmp_root=> 1);
            $tmp_root_done = 1;
            # Run wafliron_start which will call wafliron_start_on_root
            eval {($found, $output) = wafliron_on_root($Corruption::con, $object, root_vol=>$vol_name, oc => $::OPTIONAL_COMMIT, timeout => $to,
                                   wait => 1, tmp_root=>$tmp_root_done, type=> "aggr");
            };
            if ($@) {
                die ("Corruption::fix_iron_vol - wafliron_on_root failed with: $@");
            }
        } elsif ($::IRON_ON_ROOT =~ /multi|spare/) {
            $tmp_root_done =
                ($$injs{type} eq 'aggr') ? 1 : 0;
        
            if ($::OPTIONAL_COMMIT =~ /none/) {
                eval { ($found, $output) = wafliron_start($Corruption::rsh, $vol, timeout=>$to, oc => 0, tmp_root=>$tmp_root_done); };
            } elsif ($::OPTIONAL_COMMIT =~ /commit/) {
                eval { ($found, $output) = wafliron_runcommit($Corruption::rsh, $vol, timeout=>$to, tmp_root=>$tmp_root_done); };
            } elsif ($::OPTIONAL_COMMIT =~ /reject/) {
                eval { ($found, $output) = wafliron_runreject($Corruption::rsh, $vol, timeout=>$to, tmp_root=>$tmp_root_done); };
            }
        } else {
            die ("Corruption::fix_iron_vol - Unknown IRON_ON_ROOT value");
        }

        if ($tmp_root_done) {
            # Need to return to using original root volume and aggr
            # before we return to the caller
            wafliron_pick_new_root($Corruption::con, $Corruption::ORIG_ROOT_AGGR, $Corruption::ORIG_ROOT_VOLUME, reboot=> 1);
        }
    } else {
        
        if ($::OPTIONAL_COMMIT =~ /none/) {
                eval { ($found, $output) = wafliron_start($Corruption::rsh, $vol, timeout=>$to, oc => 0); };
        } elsif ($::OPTIONAL_COMMIT =~ /commit/) {
                eval { ($found, $output) = wafliron_runcommit($Corruption::rsh, $vol, timeout=>$to); };
        } elsif ($::OPTIONAL_COMMIT =~ /reject/) {
                eval { ($found, $output) = wafliron_runreject($Corruption::rsh, $vol, timeout=>$to); };
        }
    }

    if ($@) {
       $Corruption::result{fixresult} = "FATAL";
       logresult("fatal", "Corruption::fix_iron_vol - Execution Error while Fixing with Iron : $@"); 
    } else {
        if ($found) {
           $Corruption::result{fixresult} = "PASS";
           logresult("pass", "Corruption::fix_iron_vol - IRON found corruption");
        } else { 
           $Corruption::result{fixresult} = "FAIL";
           logresult("fail", "Corruption::fix_iron_vol - IRON did not find corruption");
           logcomment("Corruption::fix_iron_vol - IRON Output : $output");
        }
    }
 
    check_iron($$iron{cases}, \$output);
}

# Attempt to use ioc to fix one volume's corruption.
# If it doesn't find any corruption, that counts as a failure.
# Then tell all test cases related to that volume to check iron's output.

sub
fix_ioc_vol($) {
    my $iron = shift;
    my $type = (($$iron{type} eq 'aggr') ? 'aggr' : 'vol');
    my $vol = $$iron{vol};
    my $output = "";
    my $ret;
    my $to;
    my $found = 0;
    my $tmp_root_done = 0;

    my $injs = {};
    my $case;
    my $cases = $$iron{cases};
    if (scalar(@$cases) == 1) {
        $case = $$cases[0];
        $injs = $$case{inject_script};
    } else {
        $Corruption::result{fixresult} = "FAIL";
        logresult("fail", "Corruption::fix_ioc_vol - Cases is not array of one: \n".Dumper($iron));
    }

    if ($Corruption::host->hosttype =~ /simulator/) {
        $to = 2400;
    } else {
        $to = 4800;
    }
    logcomment("Perform IOC on the $type");
    # IRON_ON_ROOT does not work with trad volumes only flex volumes
    if ($::IRON_ON_ROOT && $$injs{type} ne 'trad') {
        logcomment("IRON_ON_ROOT =\"$::IRON_ON_ROOT\"");
        if ($::IRON_ON_ROOT =~ /core/) {
            my ($object, $vol_name);
            # The injection type dictates what volume or aggr we pass to
            #  vol_pick_root
            if ($$injs{type} eq 'aggr') {
                logcomment("inject_type is: $$injs{type} vol is: $$iron{vol}");
                my $name_str = get_name_str($cases);
                $vol_name = 'testvol_flex_'.$name_str;
                $object = $$iron{vol};
            } else {
                logcomment("inject_type is: $$injs{type} vol is: $$injs{vol}");
                $vol_name = $$injs{vol};
                $object = $$iron{vol};
            }
            # Pick corrupted aggr and volume as new root
            wafliron_pick_new_root($Corruption::con, $object, $vol_name, reboot=> 0, tmp_root=> 1);
            $tmp_root_done = 1;
            # Run wafliron_start which will call wafliron_start_on_root
            eval {($found, $output) = wafliron_on_root($Corruption::con, $object, root_vol=>$vol_name, oc => 'commit', timeout => $to,
                                   wait => 1, tmp_root=>$tmp_root_done);
            };
        } elsif ($::IRON_ON_ROOT =~ /multi|spare/) {
            $tmp_root_done =
                ($$injs{type} eq 'aggr') ? 1 : 0;

            # Run IOC on Commit Mode
            eval { ($found, $output) = wafliron_runcommit($Corruption::rsh, $vol, timeout=>$to, tmp_root=>$tmp_root_done); };
        } else {
            die ("Corruption::fix_ioc_vol - Unknown IRON_ON_ROOT value");
        }

        if ($tmp_root_done) {
            # Need to return to using original root volume and aggr
            # before we return to the caller
            wafliron_pick_new_root($Corruption::con, $Corruption::ORIG_ROOT_AGGR, $Corruption::ORIG_ROOT_VOLUME, reboot=> 1);
        }
    } else {
        # Run IOC on Commit Mode
        eval { ($found, $output) = wafliron_runcommit($Corruption::rsh, $vol, timeout=>$to); };
    }

    if ($@) {
        $Corruption::result{fixresult} = "FATAL";
        logresult("fatal", "Corruption::fix_ioc_vol - Execution Error while Fixing with IOC: $@");
    } else {
        if ($found) {
           $Corruption::result{fixresult} = "PASS";
           logresult("pass", "Corruption::fix_ioc_vol - IOC found corruption");
        } else {
           $Corruption::result{fixresult} = "FAIL";
           logresult("fail", "Corruption::fix_ioc_vol - IOC did not find corruption");
           logcomment("Corruption::fix_ioc_vol - IOC Output : $output");
        }
    }
    check_iron($$iron{cases}, \$output);
}

# Attempt to use iron to verify that one volume's corruption was fixed.
# If it finds any corruption, that counts as a failure.

sub
verify_iron_vol($) {
    my $iron = shift;
    my $type = (($$iron{type} eq 'aggr') ? 'aggr' : 'vol');
    my $vol = $$iron{vol};
     
    my ($found, $output);

    if ($::OPTIONAL_COMMIT =~ /none/) {
        eval { ($found, $output) = wafliron_start($Corruption::rsh, $vol, oc => 0); };
    } elsif ($::OPTIONAL_COMMIT =~ /commit/) {
        eval { ($found, $output) = wafliron_runcommit($Corruption::rsh, $vol); };
    } elsif ($::OPTIONAL_COMMIT =~ /reject/) {
        eval { ($found, $output) = wafliron_runreject($Corruption::rsh, $vol); };
    }

    if ($@) {
        $Corruption::result{verifyresult} = "FATAL";
        logresult("fatal", "Corruption::verify_iron_vol - Execution Error while Verifing with IRON: $@");
    } else {
        if ($found) {
            $Corruption::result{verifyresult} = "FAIL"; 
            logresult("fail", "Corruption::verify_iron_vol - IRON thinks corruption was not fixed");
            logcomment("Corruption::verify_iron_vol - IRON Output : $output");
        } else {
            $Corruption::result{verifyresult} = "PASS";
            logresult("pass", "Corruption::verify_iron_vol - IRON thinks corruption was fixed");
        }
    }
}


# Add the verify_ioc_vol() subroutine to support IOC among METHODS
#
# Attempt to use ioc to verify that one volume's corruption was fixed.
# If it finds any corruption, that counts as a failure & leave the 
# corruption as it is for manual debugging

sub
verify_ioc_vol($) {
    my $ioc = shift;
    my $type = (($$ioc{type} eq 'aggr') ? 'aggr' : 'vol');
    my $vol = $$ioc{vol};

    my ($found, $output);

    # Run WaflIron in Optional Commit so that it fixes the corruptio for 
    #     the test bed to be stable for ffurther test case execution 
    eval { ($found, $output) = wafliron_start($Corruption::rsh, $vol, oc => 1); };

    if ($@) {
        $Corruption::result{verifyresult} = "FATAL";
        logresult("fatal", "Corruption::verify_ioc_vol - Execution Error while Verifing with IOC: $@");
    } else {
        if ($found) {
           $Corruption::result{verifyresult} = "FAIL";
           logresult("fail", "Corruption::verify_ioc_vol - IOC thinks corruption was not fixed");
           logcomment("Corruption::verify_ioc_vol - IOC Output : $output");
        } else {
           $Corruption::result{verifyresult} = "PASS";
           logresult("pass", "Corruption::verify_ioc_vol - IOC thinks corruption was fixed");
        }
    }
}

# Attempt to use multiple runs of iron to fix the corruption created by
# a group of injects.  Tell each test case involved with a given inject
# (refs to the cases are kept within the injects) to check iron's output 
# from ironing that volume.

sub
fix_iron($) {
    my $injects = shift;
    my $irons = calc_irons($injects);

    logcomment("Corruption::fix_iron - Using iron to fix corruption");

    my $boot_timeout = $main::BOOT_TIMEOUT || BOOT_TIMEOUT;
    transit_to($Corruption::con, UP, prefix=>"FILER", timeout => $boot_timeout)
        or die("Corruption::fix_iron - transit_to: $@");

    for my $vol(sort(keys(%$irons))) {
        fix_iron_vol($$irons{$vol});
    }
}

# Add the fix_ioc() subroutine to support IOC among METHODS
#
# Attempt to use multiple runs of iron to fix the corruption created by
# a group of injects.  Tell each test case involved with a given inject
# (refs to the cases are kept within the injects) to check iron's output
# from ironing that volume.

sub
fix_ioc($) {
    my $injects = shift;
    my $irons = calc_irons($injects);

    logcomment("Corruption::fix_ioc - Using IOC to fix corruption");

    my $boot_timeout = $main::BOOT_TIMEOUT || BOOT_TIMEOUT;
    transit_to($Corruption::con, UP, prefix=>"FILER", timeout => $boot_timeout)
        or die("Corruption::fix_ioc - transit_to: $@");

    for my $vol(sort(keys(%$irons))) {
        fix_ioc_vol($$irons{$vol});
    }
}


# Attempt to use multiple runs of iron to verify that the corruption created
# by a group of injects has been fixed.  If any run of iron finds corruption,
# that counts as a test failure.

sub
verify_iron($) {
    my $injects = shift;
    my $irons = calc_irons($injects);

    logcomment("Corruption::verify_iron - Using iron to verify that corruption was fixed");

    my $boot_timeout = $main::BOOT_TIMEOUT || BOOT_TIMEOUT;
    transit_to($Corruption::con, UP, prefix=>"FILER", timeout => $boot_timeout)
        or die("Corruption::verify_iron - transit_to: $@");

    for my $vol(sort(keys(%$irons))) {
        verify_iron_vol($$irons{$vol});
    }
}

# Add the verify_ioc() subroutine to support IOC among METHODS
#
# Attempt to use multiple runs of ioc to verify that the corruption created
# by a group of injects has been fixed.  If any run of ioc finds corruption,
# that counts as a test failure & leave the corruption as it is for manual 
# debugging

sub
verify_ioc($) {
    my $injects = shift;
    my $irons = calc_irons($injects);

    logcomment("Corruption::verify_ioc - Using IOC to verify that corruption was fixed");

    my $boot_timeout = $main::BOOT_TIMEOUT || BOOT_TIMEOUT;
    transit_to($Corruption::con, UP, prefix=>"FILER", timeout => $boot_timeout)
        or die("Corruption::verify_ioc - transit_to: $@");

    for my $vol(sort(keys(%$irons))) {
        verify_ioc_vol($$irons{$vol});
    }
}

# Using a given method, attempt to fix corruption created by a group of
# test cases (for wack) or injects (for iron), and check that method's
# output.

sub
fix($$$) {
    my ($method, $cases, $injects) = @_;
    if ($method eq 'wack') {
        fix_wack($cases);
    } elsif ($method eq 'iron') {
        fix_iron($injects);
    } elsif ($method eq 'ioc') {
        fix_ioc($injects);
    } else {
        logresult("fatal", "Corruption::fix - Illegal method: $method");
    }
}


# Using a given method, attempt to verify that the corruption created by
# a group of test cases (for wack) or injects (for iron), and check that
# method's output.

sub
verify($$$) {
    my ($method, $cases, $injects) = @_;
    if ($method eq 'wack') {
        verify_wack($cases);
    } elsif ($method eq 'iron') {
        verify_iron($injects);
    } elsif ($method eq 'ioc') {
        verify_ioc($injects);
    } else {
        logresult("fatal", "Corruption::verify - Illegal method: $method");
    }
}


# Given a group of test cases, a fix method, and optionally a verify method,
# create resources, corrupt them, attempt to repair them, check the repair
# method's output, optionally attempt to verify using a second repair method
# that they were successfully repaired, and then check the post-repair
# filesystems to verify that they look right.

sub
run_set($@) {
    my ($cases, $fix, $verify) = @_;
    $Corruption::result{fix} = $fix;
    $Corruption::result{verify} = $verify;

    logcomment("PHASE 1 :: Inject Corruption");
    my $injects = prepare($cases);

    logcomment("PHASE 2 :: Fix Corruption using '$fix'");
    fix($fix, $cases, $injects);

    if (defined($verify)) {
        logcomment("PHASE 3 :: Verify Fix using '$verify'");
        verify($verify, $cases, $injects);
    } else {
        logcomment("PHASE 3 :: No Verify Method Defined");
    }

    logcomment("PHASE 4 :: Online Aggr & Vol"); 
    check_data($cases);

    logcomment("    -- RESULT --
    =====================================================
        Data Injection : $Corruption::result{injectstr}
        Fix with       : $Corruption::result{fix}     
        Fix Result     : $Corruption::result{fixresult}
        Verify with    : $Corruption::result{verify}
        Verify Result  : $Corruption::result{verifyresult}
    =====================================================");
}


# Given a group of test cases, run a sequence of such sets as defined
# by a METHODS string.

# >> For Methods = WACK & IRON - Do not care OPTIONAL COMMIT (IRON can be IRON / IOC)
# >> For Methods = WACK & IOC  - Do not care OPTIONAL COMMIT (IOC as run_reject)
# >> For METHODS = WACK & WACK - Do not care OPTIONAL COMMIT

# >> For METHODS = IRON & IOC  - Set OPTIONAL COMMIT = NONE for the run set. Run IRON & IOC as reject
# >> For METHODS = IOC  & IRON - Set OPTIONAL COMMIT = NONE for the run set. Run IOC as Commit & IRON
# >> For METHODS = IRON & IRON - Set OPTIONAL COMMIT = NONE for the run set. Run IRON & IRON
# >> For METHODS = IOC  & IOC  - Set OPTIONAL COMMIT = NONE for the run set. Run IOC as Commit & IOC as Reject

sub
run_methods($) {
    my ($cases) = @_;
    my @sets = split(/\s?;\s?/, $::METHODS);
    my $TEMP_OPTIONAL_COMMIT = $::OPTIONAL_COMMIT;
    for my $set (@sets) {
        logcomment("Corruption::run_methods - Running set: $set");
        my ($fix, $verify) = split(/\s?,\s?/, $set);

        # If (fix / verify =  IRON or IOC) & OPTIONAL_COMMIT = commit / reject
        #    Move the Value of OPTIONAL_COMMIT to TEMP variable
        #    Set the Value of OPTIONAL_COMMIT to NONE
        #    Run the Set & upon Completion, reset the value to Original
        #    TEMP_OPTIONAL_COMMIT = OPTIONAL_COMMIT
                
        # For WACK IOC - Set $OPTIONAL_COMMIT = NONE
        if (($fix ne "wack")||($verify ne "wack")) {
            $TEMP_OPTIONAL_COMMIT = $::OPTIONAL_COMMIT;
            $::OPTIONAL_COMMIT = "none";
            logcomment("Corruption::run_methods - REASSIGN OPTIONAL COMMIT = $::OPTIONAL_COMMIT");
        }

        run_set($cases, $fix, $verify);
        logcomment("Corruption::run_methods - Finished set: $set");
        # Restore the Original Value for OPTIONAL_COMMIT
        logcomment("Corruption::run_methods - Restore The Original Value to $::OPTIONAL_COMMIT");
        $::OPTIONAL_COMMIT = $TEMP_OPTIONAL_COMMIT;
    }
}


# Given a group of test cases and a METHODS string, run them in a coordinated
# way - within each fix/verify set, all resources will be created at once,
# all corruption performed at once, all fixing performed at once, etc.

sub
run_combined($) {
    my $cases = shift;
    my $name_str = get_name_str($cases);

    logcomment("Corruption::run_combined - Running combined batch:\n$name_str");
    init();

    my $snmp_value = filer_options_get($Corruption::rsh,"snmp.enable");
    filer_options_set($Corruption::rsh,"snmp.enable","off");
    run_methods($cases);
    destroy($cases);
    filer_options_set($Corruption::rsh,"snmp.enable",$snmp_value);
    logcomment("Corruption::run_combined - Finished batch:\n$name_str");
}


# Run a single test case according to a METHODS string.

sub
run_single($) {
    my $cases = shift;
    my $name_str = get_name_str($cases);

    logcomment("Corruption::run_single - Running test case: $name_str");
    run_methods($cases);

    logcomment("Corruption::run_single - Finished test case:\n$name_str");
}


# Given a group of test cases and a METHODS string, run them in series -
# each test case will be performed by itself and will complete before the
# next one begins.
#
# In practice, this mode is not used much, because one test failure anywhere
# along the line can devastate the sim, leaving all later tests to struggle
# and fail.

sub
run_separate($) {
    my $cases = shift;
    my $name_str = get_name_str($cases);

    logcomment("Corruption::run_separate - Running separate batch:\n$name_str");
    init();
    for my $case (@$cases) {
        run_single([$case]);
    }
    destroy($cases);
    logcomment("Corruption::run_separate - Finished batch:\n$name_str");
}


# Given a group of test cases, create and corrupt their resources, but
# don't attempt to fix them and don't destroy them afterward.  Just leave
# them around, restricted and in a corrupted state.  This mode is
# effectively a tool for generating a bunch of corrupted resources.

sub
run_corrupt($) {
    my $cases = shift;
    my $name_str = get_name_str($cases);

    logcomment("Corruption::run_corrupt - Creating and corrupting resources only:\n$name_str");
    init();
    prepare($cases);
    logcomment("Corruption::run_corrupt - Resources were created and are now being left " .
               "behind in a corrupted state:\n$name_str");
}


# Run an accumulated group of test cases in a way controlled by the
# MODE and METHODS options.

sub
run(;$) {
    my $cases = (scalar(@_) ? shift : \@Corruption::cases);

    if ($::MODE eq 'combined') {
        run_combined($cases);
    } elsif ($::MODE eq 'separate') {
        run_separate($cases);
    } elsif ($::MODE =~ /^corrupt(_one|_many)$/) {
        run_corrupt($cases);
    } else {
        logresult("fatal", "Corruption::run - Invalid mode: $::MODE");
    }
}


=head2 get_params($test_case_name, $parmname, $parmname, ...)

Each test case calls this to request from the framework the
resource names it should use.  A hash reference is returned
with the requested resource identifiers pointing to the resource
names selected by the framework.

The following resource identifiers can be requested:

=over 4

=item aggr

The name of the parent aggregate whose creation and destruction are
controlled by the framework itself.

=item flex

The name of the flexvol that the test case should use.  The test case
itself is responsible for creating the flexvol (though not for destroying it,
since the framework will automatically destroy the parent aggregate).

=item clone

The name of the clone vol that the test case should use.  The test case
itself is responsible for creating the clone (though not for destroying it,
since the framework will automatically destroy the parent aggregate).

=item trad

The name of the tradvol that the test case should use.  The test case
is responsible for creating this trad with a call to new_trad(),
but the test case does not need to destroy the trad, because the framework
will remember it and will destroy it automatically.

=back

=cut

sub
get_params($@) {
    my ($name, @params) = @_;
    my $ret = {};
    my $suffix = (($::MODE eq 'corrupt_one') ? "" : "_$name");

    for my $param (@params) {
        if ($param eq 'aggr') {
            $$ret{$param} = $Corruption::OUR_FILER_AGGR;
        } elsif ($param eq 'flex') {
            $$ret{$param} = $::FILER_VOL . $suffix;
        } elsif ($param eq 'clone') {
            $$ret{$param} = $::FILER_CLONE_VOL . $suffix;
        } elsif ($param eq 'trad') {
            $$ret{$param} = $::FILER_VOL . "_trad" . $suffix;
        } else {
            logresult("fatal", "Corruption::get_params - Test case $name requested " .
                  "invalid parameter: $param");
        }
    }
    return $ret;
}


=head2 test_case($name, {...})

Each test case calls this to register with the framework a hash full of
corruption instructions and callback functions that the framework will
use to operate the test case.  A detailed explanation of how this function
works and what contents the hash should have can be found under TEST CASES,
above.

=cut

sub
test_case($$) {
    my ($name, $case) = @_;
    add_case($name, $case);
    if ($Corruption::script_level == 0) {
        run();
    }
}


# Apply the random seed, then assimilate another test script, parsing and
# checking it and adding it to the global hash and list.  After this is
# done for all test cases in a batch, run() is called.

sub
add_script($) {
    my $path = shift;

    open(SCRIPT, $path)
        or logresult("fatal", "Corruption::add_script - Can't open script: $path");
    my @lines = <SCRIPT>;
    my $script = join("\n", @lines);
    close(SCRIPT);

    srand($::RANDOM_SEED);
    if (!defined(eval($script))) {
        logresult("fatal", "Corruption::add_script - $path: $@");
    }
}


=head2 batch($relpath1, $relpath2, ...)

As described in TEST BATCHES, a batch .thpl file calls this function
to supply a list of absolute paths of the sub-batch and sub-test-case
.thpl files it comprises.  The entire tree of referenced .thpl files are
loaded and eval()ed, building up a set of test cases in memory.  Before
returning, the top-level call to batch() executes this built-up set of
test cases in a manner determined by MODE and METHODS (see TEST EXECUTION).

=cut

sub
batch(@) {
    my @scripts = @_;

    $Corruption::script_level ++;

    for my $script (@scripts) {
        add_script($script);
    }

    $Corruption::script_level --;
    if ($Corruption::script_level == 0) {
        run();
    }
}


=head1 API FOR TEST CASES

This framework also includes a small supporting API for use by test cases'
callback functions.

=head2 new_flex($volname, size => '50m')

Creates a new flexvol of the given size within the framework's automatically
created parent aggregate.  If size is omitted, it uses the default size
from the VOL_SIZE default parameter, which at the time of this writing is
30m.  To permit many different test cases to create their own flexvols
within the same aggregate, the flexvol is created with no space guarantee.
Enables asis or compression on the volume if corresponding parameters are set.
Return true on success, or undef on failure (with error string in $@).

=cut

sub
new_flex($@) {
    my ($vol, %opts) = @_;
    my $size = (defined($opts{size}) ? $opts{size} : $::VOL_SIZE);

    Volume::vol_create_flex($Corruption::rsh, $vol, 
        $Corruption::OUR_FILER_AGGR,
        $size, {guarantee => $::SPACE_GUARANTEE,skip_volume_special_options_hook => 1}) or croak ($@);
    if ($::SNAPLOCK_AGGR eq "snaplock" or $::SNAPLOCK_AGGR eq "snaplock_enterprise" ) {
        set_default_retention_period($Corruption::rsh, $::WORM_RETENTION , $vol ) ;
    }

    asis_enable($vol, %opts) if ($::ASIS_ENABLE);
    compression_enable($vol, %opts)     if ($::COMPRESSION_ENABLE);
    return 1;                  
}

=head2 asis_enable($volname, nfiles => 3)

Adds nearstore and asis licenses if filer is not licenced.
Enables ASIS on the volume, and create nfiles 1M each. 
Return true on success, or undef on failure (with error string in $@).

=cut


sub
asis_enable($@) {
    my ($vol, %opts) = @_;
    my $vol_path = "/vol/$vol"; 
    my $nfiles = (defined($opts{nfiles}) ? $opts{nfiles} : 2);
    logcomment("Corruption::asis_enable - Enabling ASIS on the volume $vol");

    unless(is_licensed($Corruption::rsh, 'nearstore_option')&& is_licensed($Corruption::rsh, 'a_sis')){
        logcomment ("Corruption::asis_enable - ASIS is not licensed ,so adding nearstore and a_sis licenses");
        license($Corruption::rsh, 'nearstore_option') or croak($@);
        license($Corruption::rsh, 'a_sis') or croak($@);    
    }
    dense_toggle(host=>$Corruption::host, path_name=> $vol_path, state=>"on") 
        or die ("Corruption::asis_enable - asis_on: $@");

    foreach (1 .. $nfiles) {
        make_file($vol, "asis_file", size => 1000000)
            or die ("Corruption::asis_enable - make_file: $@");   
    }
    return 1;
}

=head2 asis_start($volname, timeout => 2400)

Starts  deduplication process on the volume, and wait till it done.
Return true on success, or undef on failure (with error string in $@).

=cut
sub
asis_start($@) {
    my ($vol, %opts) = @_;
    my $vol_path = "/vol/$vol"; 
    my $timeout = (defined($opts{timeout}) ? $opts{timeout} : 4800);
    my $endtime = time() + $timeout;
    my $current_status;
    
    dense_start(host=>$Corruption::host, path_name=>$vol_path)
        or die("Corruption::asis_start - $@");
    while(time() < $endtime) {
        my %status = dense_status(host=>$::FILER, path=>$vol_path);
        $current_status = $status{$vol_path}{Status};
        logcomment ("Corruption::asis_start - asis status  = $current_status");
        if($current_status =~/Idle/i){
            return 1;    
        }else{
            sleep(30);
        }
    }   
    croak("Corruption::asis_start - asis timeout");
}

=head2 compression_enable($volname, timeout => 2400)

Enable ccompression on the specified volume

=cut 

sub
compression_enable($@) {
    my ($vol, %opts) = @_;
    unless(is_licensed($Corruption::con, 'nearstore_option')&& is_licensed($Corruption::con, 'compression')){
        logcomment ("Corruption::compression_enable - Compression is not licensed ,so adding nearstore and compression licenses");
        license($Corruption::rsh, 'nearstore_option') or croak("Corruption::compression_enable - Error[nearstore_option] : $@");
        license($Corruption::rsh, 'compression') or croak("Corruption::compression_enable - Error[compression] : $@");  
    }
    enable_compression_option(filer=>$::FILER, volname=>$vol)
        or croak ("Corruption::compression_enable - Unable to turn compression on $vol");
                    
}   

=head2 new_trad($volname, size => '2')

Creates a new tradvol using the given number of disks.
If size is omitted, it uses the default size from the TRAD_SIZE
default parameter, which at the time of this writing is 2.
If creation succeeds, the framework will remember that this tradvol
was created and will automatically destroy it during cleanup.
Return true on success, or undef on failure (with error string in $@).

=cut

sub
new_trad($@) {
    my ($vol, %opts) = @_;
    my $size = (defined($opts{size}) ? $opts{size} : $::VOL_SIZE);

    return eval {

        Volume::vol_purge($Corruption::rsh, $vol);

            if (Volume::vol_create($Corruption::rsh, $vol, $size)) {
            push(@Corruption::trads, $vol);
            asis_enable($vol, %opts) if ($::ASIS_ENABLE);
            compression_enable($vol, %opts) if ($::COMPRESSION_ENABLE);
            ## Set the volume retention if the volume is in snaplock aggregate.
            if ($::SNAPLOCK_AGGR eq "snaplock" or $::SNAPLOCK_AGGR eq "snaplock_enterprise" ) {
                set_default_retention_period($Corruption::rsh, $::WORM_RETENTION , $vol ) ;
            }
            return 1;
        } else {
            croak("Corruption::new_trad - vol_create: $vol - $@");  
        }
    };
}

=head2 new_clone($clone_volname, $parent_volname, snapname => 'hourly.1')

Creates a new clone with the given parent flex vol within the framework's
automatically created parent aggregate. If snapshot is not given, a new 
base snapshot will be created for the clone.
If the clone vol already exists, it will be purged and newly created.
Return true on success, or undef on failure (with error string in $@).

=cut

sub
new_clone($@) {
    my ($clone_vol, $parent_vol, %opts) = @_;
    my $snap = (defined($opts{snapname}) ? $opts{snapname} : "");
    my $force_clone = ( $opts{force_clone} ? 1 : 0);
    add_license("flex_clone");
    return Volume::Clone::vol_clone_create(
                    conn => $Corruption::rsh,
                    parent => $parent_vol,
                    clone => $clone_vol,
                    snapname => $snap,
                    force_clone => $force_clone,
                    purge_existing => 1);
}

=head2 vol_is_online($volname)

Return a true value if the named volume is online, or a false value
if it is not.

=cut

sub
vol_is_online($) {
    my $vol = shift;

    return eval {
        my $obj = Volume::vol_status($Corruption::rsh, $vol)
            or croak("Corruption::vol_is_online - vol_status: $@");

        my $state = $obj->state();
        if ($state !~ /online/i) {
            croak("Corruption::vol_is_online - state is not online: $state");
        }

        return 1;
    };
}

=head2 vol_is_unrecoverable($volname)

Return a true value if the named volume is marked offline and unrecoverable, 
or a false value if it is not.

=cut

sub
vol_is_unrecoverable($) {
    my $vol = shift;

    return eval {
        my $obj = Volume::vol_status($Corruption::rsh, $vol)
            or croak("vol_status: $@");

        my $state = $obj->state();
        my $status = $obj->status();
        if ($state !~ /offline/i || $status !~ /unrecoverable/i ||
            $status !~ /wafl inconsistent/i) {
            croak("Corruption::vol_is_online - Not offline/unrecoverable: $state/$status");
        }

        return 1;
    };
}

=head2 mark_flex_for_recovery($flexname)

Mark a flexvol for identification after lost flexvol recovery,
by creating in the flexvol's top directory a file whose name is
the value of NAME_FILE and whose contents are the flexvol's name.
Return true on success, or undef on failure (with error string in $@).

=cut

sub
mark_flex_for_recovery($) {
    my ($vol) = @_;
    return wrfile($Corruption::rsh, "/vol/$vol/$::NAME_FILE", $vol);
}

=head2 find_lost_flex($prevname_regex)

Search all volumes whose names begin with the value of LOST_PREFIX
for a top-directory file whose name is the value of NAME_FILE and
whose contents (the volume's pre-corruption name) match the specified
regular expression.  Such a file was presumably created by a
pre-corruption call to mark_flex_for_recovery().  If a matching
volume is found, its name is returned; if not, undef is returned.

=cut

sub
find_lost_flex($) {
    my ($rex) = @_;
    my @vols = Volume::vol_list($Corruption::rsh);
    for my $vol (@vols) {
        next unless ($vol =~ /^$::LOST_PREFIX/);
        my $data = rdfile($Corruption::rsh, "/vol/$vol/$::NAME_FILE");
        if ($data =~ $rex) {
            return $vol;
        }
    }
    return undef;
}

=head2 vol_exists($volname)

Return a true value if the named volume exists, or a false value
if it is not.

=cut

sub
vol_exists($) {
    my $vol = shift;
    return Volume::vol_exists($Corruption::rsh, $vol);
}

=head2 mount_volume($volname)

Export and mount the named volume.  If successful, return the path of the
mount point; otherwise return undef, with error string in $@.

=cut

sub
mount_volume($) {
    my ($vol) = @_;
    return eval {
        my $path = "/vol/$vol";
        ExportFS::export_path($Corruption::rsh, $path,
                      OPTIONS => $::EXPORT_OPTS,
                      WRITE_ETC => 1)
            or croak("Corruption::mount_volume : Error $@");
        my $mpt = mount::mount($Corruption::host->default_ip(),
                       "/vol/$vol",
                       mountopts => $::MOUNT_OPTS,
                       chmod => "0777")
            or croak("Corruption::mount_volume : Error $@");
        return $mpt;
    };
}

=head2 list_entries($vol, $dirpath)

Return a list of the entries in the given directory within the
given volume.  The directory path is given relative to the
volume's root, omitting the initial "/".  "." and "..",
if present, are included in the resulting list.  On failure,
return a null list and leave any explanation in $@.

=cut

sub
list_entries($$) {
    my ($vol, $dirpath) = @_;
    my $path = "/vol/$vol/$dirpath";

    return eval {
        my(undef,undef,$output) = say($Corruption::rsh,
                "priv set -q test; ls $path");

        if ($output =~ /can't/i) {
            croak("Corruption::list_entries - ls: $output");
        }

        my @files = split(/[\r\n]+/, $output);

        for my $i (0 .. $#files) {
            $files[$i] =~ s/\s*$//;
        }

        return @files;
    };
}

=head2 make_file($vol, $file, <contents => $contents | size => $size> { scramble => $scramble } )

Create a file with given contents or size in the root directory of the
given volume.  If size is given, the file will be created by the mkfile
command and will contain all zeroes. If ASIS_ENABLE starts deduplication
so, fingerprints database is updated for each file.Return true for 
success, or undef for failure (with error string in $@).

=cut

sub
make_file($$@) {
    my ($vol, $file, %opts) = @_;
    my $filer_path        = "/vol/$vol/$file";
    my $mountpoint        = $Corruption::vol_mounts->{$vol};
    my $contents          = $opts{contents};
    my $size              = $opts{size};
    my $nfiles            = defined($opts{nfiles}) ? $opts{nfiles} : 1;
    my $asis_flag         = $::ASIS_ENABLE;
    my $compression_flag  = $::COMPRESSION_ENABLE;
    my $scramble = ( defined($opts{scramble})  ?  $opts{scramble} :
                                            ( $asis_flag ? 0.5 :
                                       ( $compression_flag ? 40 : 0 ) ) );
    return eval {

        # logcomment("Corruption::make_file\n\nVol: $vol\nFile: $file\nScramble: $scramble\nContents: ".
        #       ($contents ? "<-- CONTENTS  Length: ".length($contents)." -->" : "") ."\n".
        #       "Size: $size\nNFiles: $nfiles");
        logcomment(" Corruption::make_file\n\nVol: $vol\nFile: $file\nNFiles: $nfiles");

        if (defined($size) && defined($contents)) {
            logcomment("Corruption::make_file - Both 'size' and 'contents' options were given, ".
                     "ignoring 'size' option");
            $size = undef;
        } elsif (!defined($contents) && !defined($size)) {
            croak("Corruption::make_file - Must provide either contents or size option");
        }
        my $scramble = 0;
        my $rval;

        unless ($mountpoint) {
            $mountpoint = vol_mount_aux($vol);
            croak("Corruption::make_file - Could not mount volue $vol:  Error: $@") if (!$mountpoint);
        }
        if (defined($contents)) {
            # logcomment("Corruption::make_file:\n".  "NOTE:  File 'contents' option was specified\n".
            #              "Length of contents: ".length($contents));
        
            ## Commit the file to WORM by changing file mode to 0444
            if ( $::SNAPLOCK_AGGR eq 'snaplock' or $::SNAPLOCK_AGGR eq 'snaplock_enterprise' ) {
                my $mode = "0444";
                $rval = write_file( agent   => 'localhost',
                        vol_path    => $filer_path,
                        mountpoint  => $mountpoint,
                        contents    => $contents,
                        numfiles    => $nfiles,
                        mode    => $mode,
                        verify  => 1,
                     );
            } else {

                $rval = write_file( agent   => 'localhost',
                        vol_path    => $filer_path,
                        mountpoint  => $mountpoint,
                        contents    => $contents,
                        numfiles    => $nfiles,
                        verify  => 1,
                     );
            }   
            if ($@ || !$rval) {
                croak("Corruption::make_file - write_file error: $@");
            }
            return 1;

        } elsif (defined($size)) {
            # logcomment("Corruption::make_file:\n".  "File 'size' was specified: $size);
            ## Commit the file to WORM by changing file mode to 0444
            if ( $::SNAPLOCK_AGGR eq 'snaplock' or $::SNAPLOCK_AGGR eq 'snaplock_enterprise' ) {
                my $mode = "0444";
                $rval = write_file( agent   => 'localhost',
                        vol_path    => $filer_path,
                        mountpoint  => $mountpoint,
                        size    => $size,
                        scramble    => $scramble,
                        numfiles    => $nfiles,
                        mode    => $mode,
                        verify  => 1,
                         );
            } else {
                $rval = write_file( agent   => 'localhost',
                        vol_path    => $filer_path,
                        mountpoint  => $mountpoint,
                        size    => $size,
                        scramble    => $scramble,
                        numfiles    => $nfiles,
                        verify  => 1,
                    );
            }
            if ($@ || !$rval) {
                croak("Corruption::make_file - write_file error: $@");
            }

            if ($::ASIS_ENABLE){
                # logcomment("Corruption::make_file:  "'ASIS_ENABLE' is set -- Calling 'asis_start' ....");
                asis_start($vol, %opts) 
                     or croak("Corruption::make_file -  asis_start error: $@");   
                # logcomment("Corruption::make_file:  ".  "... Returned from 'asis_start'");
            }

            return 1;
        }
    };
}

=head2 check_file($vol, $file, match => $rex)

Check that a given file with contents matching a given regular expression
exists in the root directory of the given volume.  Return true if it does,
or undef if it does not (with any explanation in $@).

=cut

sub
check_file($$@) {
    my ($vol, $file, %opts) = @_;
    my $path = "/vol/$vol/$file";
    my $match = $opts{match};
    my $size = $opts{size};

    return eval {
        if (defined($match)) {
            my $tmp = rdfile($Corruption::rsh, $path);
            if ($tmp !~ $match) {
                croak("Corruption::check_file - $path contents do not match " .
                      "[$match]: [$tmp]");
            }
            return 1;
        } else {
            croak("Corruption::check_file - Must provide match");
        }
    };
}

=head2 make_file_nfs($path, $str, $count)

Use NFS to create a file at the given path whose contents are
$count repetitions of $str.  Return true on success or
undef on failure, with error string in $@.

=cut

sub
make_file_nfs($$$) {
    my ($path, $str, $count) = @_;

    return eval {

        open(MAKE_FILE_NFS, ">$path")
            or croak("Corruption::make_file_nfs - open ($path): $!");

        for (my $i = 0; $i < $count; $i++) {
            print(MAKE_FILE_NFS $str)
                or croak("Corruption::make_file_nfs - print = ($path): $!");
        }

        close(MAKE_FILE_NFS)
            or croak("Corruption::make_file_nfs - close ($path): $!");

        return 1;
    };
}

=head2 check_file_nfs($path, $str, $count)

Use NFS to check a file at the given path whose contents are
expected to be $count repetitions of $str.  Return true if the file
exists and matches or undef otherwise, with explanation in $@.

=cut

sub
check_file_nfs($$$) {
    my ($path, $str, $count) = @_;
    my $tmp;
    my $len = length($str);
    my $read;

    return eval {

        open(CHECK_FILE_NFS, "<$path")
            or croak("Corruption::check_file_nfs - open ($path): $!");

        for (my $i = 0; $i < $count; $i++) {

            $read = read(CHECK_FILE_NFS, $tmp, $len);
            if ($read != $len) {
                croak("Corruption::check_file_nfs - $i - read: $read/$len: $!");
            }

            if ($tmp ne $str) {
                croak("Corruption::check_file_nfs - $i - mismatch: $tmp/$str");
            }
        }

        $read = read(CHECK_FILE_NFS, $tmp, $len);
        if ($read != 0) {
            croak("Corruption::check_file_nfs - data after expected eof: $read: $tmp");
        }

        close(CHECK_FILE_NFS)
            or croak("Corruption::check_file_nfs - close ($path): $!");

        return 1;
    };
}

=head2 remove_file($vol, $file)

Remove a given file from the root directory of the given volume.
Return true for success, or undef for failure (with error
string in $@).

=cut

sub
remove_file($$) {
    my ($vol, $file) = @_;
    my $path = "/vol/$vol/$file";

    return eval {
        my $output = SayWrap::sayEOF($Corruption::rsh,
            "priv set test; rm $path");
        if ($output !~ /^\s*$/) {
            croak("Corruption::remove_file - rm: $output");
        }
        return 1;
    };
}

=head2 find_lost_file($vol, $rex)

Search the given volume's $::LOST_FOUND_DIR for a file whose
contents match the given regular expression.  If found, return
the file's name.  Otherwise, return undef (with any explanation in $@).

=cut

sub
find_lost_file($$) {
    my ($vol, $rex) = @_;

    return eval {
        my @files = list_entries($vol, $::LOST_FOUND_DIR)
            or croak("Corruption::find_lost_file - list_entries: $@");

        for my $file (@files) {
            next if ($file =~ /^\.+$/);
            if (check_file($vol, "\"$::LOST_FOUND_DIR/$file\"",
                    match => $rex)) {
                return $file;
            }
        }

        croak("Corruption::find_lost_file - no match: @files");
    };
}

=head2 make_dir($path)

Create a directory with a given absolute path (from the client's
perspective, not the filer's).  The volume must already
be mounted on the client, and the path must include its mount point.
Return true for success, or undef for failure (with error string in $@).

=cut

sub
make_dir($) {
    my ($path) = @_;
    if (mkdir($path)) {
        if ( $::SNAPLOCK_AGGR eq 'snaplock' or $::SNAPLOCK_AGGR eq 'snaplock_enterprise' ) {
            ## Create one worm file on the new directory 
            ## Get Volume name & file path from mount path  
            my @path_split = split(/\//, $path, 5);
            my @vol_split  = split(/\_/, $path_split[3], 2); 
            make_file($vol_split[1], "$path_split[4]/file_tmp" , size => "100k" );  
        }
        return 1;
    } else {
        $@ = $!;
        return undef;
    }
}

sub
make_one_snapshot($$) {
    my ($vol, $snap) = @_;

    return eval {
        make_file($vol, $::SNAP_FILE, contents => $snap)
            or croak("Corruption::make_one_snapshot - make_file: $@");
        SnapCmds::snapshot_create($Corruption::rsh, $vol, $snap)
            or croak("Corruption::make_one_snapshot - snapshot_create: $@");
        remove_file($vol, $::SNAP_FILE)
            or croak("Corruption::make_one_snapshot - remove_file: $@");
        return 1;
    };
}

=head2 make_snapshots($vol, @snapnames)

Create a series of one or more snapshots with specified names.
Each snapshot will capture a newly created file in its root directory whose
name is $::SNAP_FILE and whose contents are the snapshot's name.  Return
true if all succeed or undef if any fails (with error string in $@).

=cut

sub
make_snapshots($@) {
    my ($vol, @snaps) = @_;

    return eval {
        for my $snap (@snaps) {
            make_one_snapshot($vol, $snap)
                or croak("Corruption::make_snapshots - $@");
        }
        return 1;
    };
}

=head2 list_snapshots($vol)

Return a list of the snapshots within a given volume,
or a null list on failure (with error string in $@).

=cut

sub
list_snapshots($) {
    my ($vol) = @_;
    return SnapCmds::snapshot_list($Corruption::rsh, $vol);
}

=head2 check_snapshot($vol, $snap, $rex)

Check whether the given snapshot has in its root directory
a file whose name is $::SNAP_FILE and whose contents match $rex.
Return true if this is so or undef if it is not so (with explanation
in $@).

=cut

sub
check_snapshot($$$) {
    my ($vol, $snap, $rex) = @_;

    return eval {
        my $path = "/vol/$vol/$::SNAPDIR/$snap/$::SNAP_FILE";
        my $contents = rdfile($Corruption::rsh, $path);
        if ($contents !~ $rex) {
            croak("Corruption::check_snapshot - $path contents do not match: [$contents]");
        }
        return 1;
    };
}

=head2 find_snapshot($vol, $rex)

Search all snapshots on the given volume and determine whether
any of them has in its root directory a file whose name is $::SNAP_FILE
and whose contents match $rex.  Return the name of the matching snapshot,
or undef if not found (with any explanation in $@).

=cut

sub
find_snapshot($$) {
    my ($vol, $rex) = @_;
    return eval {
        my @snaps = Corruption::list_snapshots($vol);
        for my $snap (@snaps) {
            if (check_snapshot($vol, $snap, $rex)) {
                return $snap;
            }
        }
        return undef;
    };
}

=head2 populate_zombie_test_files($vol)

Populate files to be used in zombie metafile tests.
Return true for success, or undef for failure (with error string in $@).

=cut

sub
populate_zombie_test_files($) {
    my ($vol) = @_;
    return eval {
        for (my $i = 0;  $i < $::ZOMBIE_FILE_COUNT; $i++) {
            my $file = "zombie_file.$i";
            make_file($vol, $file, size => $::ZOMBIE_FILE_SIZE)
                or croak("Corruption::populate_zombie_test_files - make_file: $@");
        }
        return 1;
    };
}

=head2 remove_zombie_test_files($vol)

Remove files populated for zombie metafile tests.
Return true for success, or undef for failure (with error string in $@).

=cut

sub
remove_zombie_test_files($) {
    my ($vol) = @_;
    return eval {
        for (my $i = 0;  $i < $::ZOMBIE_FILE_COUNT; $i++) {
            my $file = "zombie_file.$i";
            remove_file($vol, $file)
                or croak("Corruption::remove_zombie_test_files - remove_file: $@");
        }
        return 1;
    };
}

=head2 sfsr_zombie_test_files($vol, $snap)

Do sfsr on all zombie test files.
Return true for success, or undef for failure (with error string in $@).

=cut

sub
sfsr_zombie_test_files($$) {
    my ($vol, $snap) = @_;
    return eval {
        for (my $i = 0;  $i < $::ZOMBIE_FILE_COUNT; $i++) {
            my $file = "/vol/$vol/zombie_file.$i";
            SnapCmds::snaprestore_file($Corruption::rsh, $file,
                           $snap)
                or croak("Corruption::sfsr_zombie_test_files - snaprestore_file: $@");
        }
        return 1;
    };
}


=head2 run_floppy_boot_cmd($vol, $cmd)

Reboot filer in maintenance mode (22/7), execute the given command
and bring up filer again.
Return true for success, or undef for failure (with error string in $@).

=cut

sub
run_floppy_boot_cmd($$) {
    my ($vol, $cmd) = @_;
    return eval {

        my $cmdtimeout = 180;
        # need this for sfsr zombie tests to flush metafile
        SayWrap::sayEOF($Corruption::rsh,
                "priv set test; sync; sync");

        # floppy boot 22/7 and enter the command $cmd
        # at a boot prompt
        transit_to($Corruption::con, BOOT_MENU, prefix=>"FILER")
            or die ("Corruption::run_floppy_boot_cmd - transit_to: $@");
        SayWrap::sayWAIT($Corruption::con, "22/7", "choose one");
        SayWrap::sayWAIT($Corruption::con, $cmd, 
                 "successfully enqueued", TIMEOUT=>$cmdtimeout);

        # bring up filer
        my $boot_timeout = $main::BOOT_TIMEOUT || BOOT_TIMEOUT;
        transit_to($Corruption::con, UP, prefix=>"FILER", timeout => $boot_timeout)
            or die("Corruption::run_floppy_boot_cmd - transit_to: $@");

        return 1;
    };
}


=head2 add_license($str)

License the given service.
Return true for success, or undef for failure (with error string in $@).

=cut

sub
add_license($) {
    my ($lic) = @_;

    return eval {
        Admin::License::license($Corruption::rsh, $lic)
                                or croak("Corruption::admin_license - $@"); 
        return 1;
    };

}

=head2 fiji_add_filter($filter_str,%opts)

Add a fiji filter as specified in $filter_str.
Return true for success, or undef for failure (with error string in $@).
If 'timeout' option is given, it is used in the 'say' call for the 'fiji'
command. Default is 300 sec. If timeout of '0' is specified, then no
'timeout' option is used with the fiji command.

Options:

timeout: Optional. Default is 300 sec.


=cut

sub
fiji_add_filter($;%) {
    my ($filter_str, %opts) = @_;
    my %say_opts = ();
    if (defined($opts{timeout})) {
        $say_opts{TIMEOUT} = $opts{timeout};
    } else {
        $say_opts{TIMEOUT} = 300;
    }
    return eval {
        my $output = SayWrap::sayEOF($Corruption::rsh,
                "priv set test; fiji filter \"$filter_str\"", %say_opts);
        if (!($output =~ /filter set/i)) {
            croak("Corruption::fiji_add_filter - fiji filter: $output");
        }
        return 1;
    };
}

=head2 fiji_clear_filters(%opts)

Clear all fiji filters.
Return true for success, or undef for failure (with error string in $@).
If 'timeout' option is given, it is used in the 'say' call for the 'fiji'
command. Default is 300 sec. If timeout of '0' is specified, then no
'timeout' option is used with the fiji command.

Options:

timeout: Optional. Default is 300 sec.

=cut

sub fiji_clear_filters($;%) {
    my (%opts) = @_;
    my %say_opts = ();
    if (defined($opts{timeout})) {
        $say_opts{TIMEOUT} = $opts{timeout};
    } else {
        $say_opts{TIMEOUT} = 300;
    }
        ## @step Clear the fiji filter
        logStep("Clear the fiji filter");
    my $output = eval { SayWrap::sayEOF($Corruption::rsh,
                                  "priv set test; fiji clear",
                                          %say_opts)
                        };
    if (!($output =~ /fiji.*filter.*cleared/i)) {
        logresult('WARN', "Corruption::fiji_clear_filters - Clear fiji filter: $output");
        return 0;
    }
    return 1;
}

## Check & Configure HTTPD on the Filer  

=head2 _check_and_configure_httpd() 

Check the HTTPD Status on the Filer & Enables it if OFF

=cut

sub _check_and_configure_httpd() {

    my $data; #Temp Variable

    $data = filer_options_get($Corruption::rsh,"httpd.enable");
    if($data eq "off")
    {
        logcomment("Corruption::_check_and_configure_httpd - HTTPD is not Enabled, Enabling it ");
        filer_options_set($Corruption::rsh,"httpd.enable","on",1);
    }

    $data = filer_options_get($Corruption::rsh,"httpd.admin.enable");
    if($data eq "off")
    {
        logcomment("Corruption::_check_and_configure_httpd - HTTPD ADMIN is not enabled . Enabling it");
        filer_options_set($Corruption::rsh,"httpd.admin.enable","on",1);
    }

}




# iron_console() is moved from Iron.pm

sub iron_console ($$@) {
    my ($con, $name, %opts) = @_;
    my $type = (exists($opts{type}) ? $opts{type} : 'aggr');
    my $timeout = (exists($opts{timeout}) ? $opts{timeout} : 600);
    my $out1;
    my $out2;
    my $outref = (defined($opts{output}) ? $opts{output} : \$out1);

    (undef, undef, $$outref) = say($con,
        "priv set test; $type wafliron start $name",
        '-timeout', $timeout);

    if ($$outref !~ /Starting wafliron/i) {
        my (undef,undef, $text) = say($con, 
            "priv set test; $type wafliron status $name",
            '-timeout', $timeout);
        if ($text =~ /is active/) {
            logcomment("Corruption::iron_console - Even no Starting statement, wafliron seems running");
                }
        else {
            return IRON_NOTDONE;
                }
    }
            
    (undef, $out2) = say($con, "",
        '-noprompt',
        '-timeout', $timeout*2,
        '-wait', "$name, wafliron completed", '-eval_error');

        logcomment("Corruption::iron_console - out2: $out2 ============ ");

    my (undef,undef, $after) = say($con, 
        "priv set test; $type wafliron status $name",
        -timeout, $timeout);
    if ($after =~ /is active/) {
        $$outref .= $out2;
        (undef, $out2) = say($con, "",
            '-noprompt',
            '-timeout', 900,
            '-wait', "$name, wafliron completed", '-eval_error');
        my (undef,undef, $after) = say($con, 
            "priv set test; $type wafliron status $name",
            '-timeout', $timeout);
        if ($after =~ /is active/) {
            return IRON_NOTDONE;
        }
    }
        elsif ($after !~ /is not currently active/i) {
            $$outref .= $out2;
                if ($$outref =~ /dumpcore/ ) {
                    return IRON_COREDUMP;
                }
                else {
                    return IRON_UNEXPECTED;
                }
        }

    $$outref .= $out2;
        logcomment("Corruption::iron_console - output: $$outref");

        if ($$outref =~ /dumpcore/ ) {
                return IRON_COREDUMP;
        }
    if ($$outref =~ /wafl.check.info:error/ || 
            $$outref =~ /wafl.iron.unrecoverable:CRITICAL/) {
        return IRON_DIRTY;
    } else {
        return IRON_CLEAN;
    }
}
sub inode_number ($)
{
 my ($inum) = @_;
    my $ver = get_wafl_version(host($::FILER));
    
    if ($ver >= WV_BOILERMAKER_MIN && $inum < 96) {
        return ("priv:$inum");
    }
    else {
        return ($inum);
    }
}

sub make_zombie_file($$@) {
my ($vol, $file, %opts) = @_;
    my $ver = get_wafl_version(host($::FILER));
    if ($ver >= WV_BOILERMAKER_MIN) {
      make_meta_file($vol, $file, contents => "zombie", file_id=>600 )
       or die("make_file: $@");
    }
    else {
       make_file($vol, $file, size => 8000 )
       or die("make_file: $@");
    } 
} 
 
sub make_meta_file($$@) {
 my ($vol, $file, %opts) = @_;
 my $path = "/vol/$vol/metadir/$file";
 my $contents = defined($opts{contents})? $opts{contents} : "metadata";
 my $file_id = defined($opts{file_id})? $opts{file_id} : 600;
 my $cmd = $path. " -private -i ".$file_id;
 return eval {
  say($Corruption::con, "priv set test; setflag wafl_metadata_visible 1");
  wrfile($Corruption::con, $cmd , $contents) or die("make_meta_file: $@");
  return 1;
 } 
}

1
