# $Id: //depot/prod/test/Rsierranevada/lib/CounterManager.pm#1 $ #
## @summary     Provides the ability to create, and increment counter instances in the  
##              process of testing the Counter Manager feature.
## @author      cwilcox@netapp.com,dl-cluster-qa@netapp.com
## @pod here

=head1 NAME

CounterManager.pm

=head1 SYNOPSIS

=over 4

 use Server;
 use CounterManager;

 my $CM_obj = new CounterManager;

 $CM_obj->verify_test_counter(
     type => "managed",
     num_instances => 10,
     num_increments => 10,
     cleanup => 1,
 );


 Creating RPC client.
 Sucessfully created RPC client.
 /*******************************************/
 /* TEST OUT perf_object_list_info */
 /*******************************************/
 \tCalling perf_object_list_info_1 (debug)
 \tCalled: print_perf_object_list_info_results
 \tResult:
 \t\tNum Objects: 2
 \t\tobject_name:"voltable" privelege_level:"basic"
 \t\tobject_name:"zapi" privelege_level:"basic"

Computed Excerpt:

 Creating RPC client.
 Sucessfully created RPC client.
 /**************************************/
 /* TEST OUT perf_object_get_instances */
 /**************************************/
 \tCalling perf_object_get_instances (testcomputed, debug)
 \tCalled: print_perf_object_get_instances_results
 \tResult:
 \t\tTimestamp: Fri Jan  8 19:56:53 2010
 \t\tNum Instances: 10
 \t\tobject_name:"testcomputed" instance_name:"cm_testcomputed_1" counter_name:"total_raw" value:"10"
 \t\tobject_name:"testcomputed" instance_name:"cm_testcomputed_1" counter_name:"time_sec" value:"1262998613"
 \t\tobject_name:"testcomputed" instance_name:"cm_testcomputed_1" counter_name:"total_avg" value:"10"
 \t\tobject_name:"testcomputed" instance_name:"cm_testcomputed_1" counter_name:"total_pct" value:"10"
 \t\tobject_name:"testcomputed" instance_name:"cm_testcomputed_1" counter_name:"time_sec_str" value:"Fri Jan  8 19:56:53 2010

Managed Excerpt:

 Creating RPC client.
 Sucessfully created RPC client.
 /**************************************/
 /* TEST OUT perf_object_get_instances */
 /**************************************/
 \tCalling perf_object_get_instances (testmanagedsimple, debug)
 \tCalled: print_perf_object_get_instances_results
 \tResult: 
 \t\tTimestamp: Fri Jan  8 19:50:20 2010
 \t\tNum Instances: 10
 \t\t\tobject_name:"testmanagedsimple" instance_name:"cm_testmanagedsimple_1" counter_name:"total_delta" value:"10"
 \t\t\tobject_name:"testmanagedsimple" instance_name:"cm_testmanagedsimple_1" counter_name:"time_rate" value:"10"

Regular Excerpt:

 Creating RPC client.
 Sucessfully created RPC client.
 /**************************************/
 /* TEST OUT perf_object_get_instances */
 /**************************************/
 \tCalling perf_object_get_instances (testregular, debug)
 \tCalled: print_perf_object_get_instances_results
 \tResult: 
 \t\tTimestamp: Fri Jan  8 19:59:29 2010
 \t\tNum Instances: 10
 \t\t\tobject_name:"testregular" instance_name:"cm_testregular_1" counter_name:"total" value:"10"
 \t\t\tobject_name:"testregular" instance_name:"cm_testregular_1" counter_name:"latency_ms" value:"699"
 \t\t\tobject_name:"testregular" instance_name:"cm_testregular_1" counter_name:"latency_hist" value:"0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0"

=head1 DESCRIPTION

C<CounterManager> Provides the ability to create, and increment test counter instances in the
                  process of testing the Counter Manager feature.

=cut

## @summary     CounterManager Library

package CounterManager;

#use statements
use strict;
use warnings;
use Data::Dumper;
use Switch qw(fallthrough);
use Server;
use Tharn qw($Log);
use Options;
use Params::Common;
use Params::Validate qw(:all);
use Carp;

my $Server = new Server();
my $Options = new Options();

my @ComputedCounters = qw(total_raw total_avg total_pct);
my @ManagedCounters = qw(total_delta time_rate);
my @RegularCounters = qw(total);

=pod

=head2 new

=over 4

B<Description>
Creates the CounterManager object.

B<Input>

B<Output>

=over 4

=item *

Blessed object reference

=back

=back

=cut

sub new {
    my $class = shift;
    my $self = {};
   
    bless $self, $class;

    return $self;
}

=pod

=head2 verify_test_counter

=over 4

B<Description>
This method is used to verify test counters for managed, computed, and regular by verifying instance
creation, and subsequent increments. A dev implemented test counter will be leveraged to ensure the
underlying infrastructure, and code base remain sane. 

B<Input>

=over 4

=item *

B<type> The type of counter to verify.

=item * 

B<num_instances> The number of instances to create.

=item *

B<num_increments> The number of times each instance's counters will be incremented.

=item *

B<cleanup> Boolean to specify whether counters will be cleaned up. Default is true. This argument exists
           solely for the ability to manually check the status after the test has been run.

=back

B<Output>

=over 4

None

=back

=back

=cut

sub verify_test_counter {
    my $self = shift;
    my %params = validate(
    @_,
    {
       type => { type  => SCALAR,
                 regex => qr/managed|computed|regular/,
                 default => "managed",
               },
       num_instances  => { type  => SCALAR,
                           optional => 1,
                           default => 1,
                         }, 
       num_increments => { type => SCALAR,
                           optional => 1,
                           default => 1,
                         },
       cleanup => { type => SCALAR,
                    optional => 1,
                    default => 1,
                  },
    }
    );
    my $type;
    if ($params{type} eq "managed") {
        $type = "testmanagedsimple";
    } elsif ($params{type} eq "computed") {
        $type = "testcomputed";
    } elsif ($params{type} eq "regular") {
        $type = "testregular";
    } elsif($params{type} eq "all" ) {
        $type = "all";
        my $objects = $Server->shell(cmd=>"cm_client -p mgwd -a $type");
        my $status = $objects ne q{} ? 1:0;
        return $status;
    }

    my $instance_prefix = "cm_".$type."_"; 
    my $num_instances = $params{num_instances};
    my $num_increments = $params{num_increments};

    # Ensure there are no pre-existing counter instances
    my $instance_precheck = $Server->shell(cmd=>"cm_client -p mgwd -a object-instance-list-info -o $type");
    my @precheck_output_array = split ('\n',$instance_precheck);

    foreach my $line (@precheck_output_array) {
        if ($line =~ /object_name:\"$type\"\sinstance_name:\"($instance_prefix\d+)/) {
            $Log->debug("Instance $1 should not exist yet. Delete, and proceed.");
            $Server->cli(cmd=>"set -priv test ; debug counter delete-instance -object $type -instance $1");
        }
    }

    my @instance_array;
    my $instance_list_output;
    my $instance_name;

    # Create the test counter instances per the num_instances parameter
    for (my $i=1;$i<=$num_instances;$i++) {
        $instance_name = $instance_prefix.$i;
        # Create test counter instance
        $Server->cli(cmd=>"set -priv test ; debug counter create-instance -object $type -instance $instance_name");
        # Verify number of created instances
        $instance_list_output = $Server->shell(cmd=>"cm_client -p mgwd -a object-instance-list-info -o $type");
        $instance_list_output =~ /Num\sInstances:\s(\d+)/;
        # If the number of instance is not correct, throw an error
        if ($1 != $i) {
            die "ERROR: We were expecting the number of created $type instances to be $i, but got $1 instead.";
        } else {
            $Log->debug("Verified $type instances: $i.");
        }
        # Verify the instance was created
        if ($instance_list_output !~ /object_name:\"$type\"\sinstance_name:\"$instance_name/) {
            die "ERROR: Could not find created instance: $instance_name in the object-instance-list-info output string.";
            
        } else {
            $Log->debug("Verified creation of the $instance_name instance.");
            push @instance_array, $instance_name;
        }

        # Verify the counters per type instance
        my @CounterArray;
        if ($params{type} eq "managed") {
            @CounterArray = @ManagedCounters;
        } elsif ($params{type} eq "computed") {
            @CounterArray = @ComputedCounters;
        } elsif ($params{type} eq "regular") {
            @CounterArray = @RegularCounters;
        }

        # Loop through each counter and increment per num_increments
        foreach my $counter (@CounterArray) {
            # Increment test counter instances
            for (my $c=1;$c<=$num_increments;$c++) {
                $Server->cli(cmd=>"set -priv test ; debug counter increment-counter -object $type -instance $instance_name -counter $counter");
            }
            # Parse the instance output and verify counters were incremented
            my $get_instance_output = $Server->shell(cmd=>"cm_client -p mgwd -a object-get-instances -o $type");
            # object_name:"testcomputed" instance_name:"cm_testcomputed_1" counter_name:"total_raw" value:"0"
            if ($get_instance_output !~ /instance_name:\"$instance_name\"\scounter_name:\"$counter\"\svalue:\"$num_increments\"/) {
                die "ERROR: Counter value for $counter, did not increment as expected.";
            }
            # Verify bucket array counters for regular object instances
            if ($params{type} eq "regular") {
                $get_instance_output =~ /instance_name:\"$instance_name\"\scounter_name:\"latency_ms\"\svalue:\"(\d+)\"/;
                my $ms_value = $1;

                $get_instance_output =~ /instance_name:\"$instance_name\"\scounter_name:\"latency_hist\"\svalue:\"(.+)\"/;
                my $bucket_string = $1;

                _test_latency($ms_value,$bucket_string);
            }
        }       
    }
    # Delete the counter instances
    if ($params{cleanup}) {
        while (scalar(@instance_array) > 0) {
            my $instance = shift(@instance_array);
            $Server->cli(cmd=>"set -priv test ; debug counter delete-instance -object $type  -instance $instance");
            # Verify deletion
            my $total_instances = scalar(@instance_array);
            $instance_list_output = $Server->shell(cmd=>"cm_client -p mgwd -a object-instance-list-info -o $type");
            $instance_list_output =~ /Num\sInstances:\s(\d+)/;
            # If the number of instances do not reflect the deletion, throw an error
            if ($1 != $total_instances) {
                die "ERROR: We were expecting the number of created $type instances to be $total_instances, but got $1 instead.";
            } else {
                $Log->debug("Verified $type instance, $instance has been deleted.");

            }
        }
    }
}
# Private Methods

sub _test_latency {
    my ($ms_value, $bucket_list) = @_;
    # Split the bucket string into an array
    my @bucket_array = split (",",$bucket_list);

    # If the latency ms value falls into the case range, and the correct bucket is filled,
    # the hist counter is verified.
    switch ( $ms_value ) {
        case [0] {if ($bucket_array[0] != 1) {
            die "ERROR: Did not find $ms_value in bucket 0 - <1ms"; } }
        case [1] {if ($bucket_array[1] != 1) {
            die "ERROR: Did not find $ms_value in bucket 1 - <2ms"; } }
        case [2..3] {if ($bucket_array[2] != 1) {
            die "ERROR: Did not find $ms_value in bucket 2 - <4ms";  } }
        case [4..5] {if ($bucket_array[3] != 1) {
            die "ERROR: Did not find $ms_value in bucket 4 - <6m";  } }#4 - <6ms
        case [6..7] {if ($bucket_array[4] != 1) {
            die "ERROR: Did not find $ms_value in bucket 6 - <8ms";  } }#6 - <8ms
        case [8..9] {if ($bucket_array[5] != 1) {
            die "ERROR: Did not find $ms_value in bucket 8 - <10ms";  } }#8 - <10ms
        case [10..11] {if ($bucket_array[6] != 1) {
            die "ERROR: Did not find $ms_value in bucket 10 - <12ms";  } }#10 - <12ms
        case [12..15] {if ($bucket_array[7] != 1) {
            die "ERROR: Did not find $ms_value in bucket 12 - <16ms";  } }#12 - <16ms
        case [16..19] {if ($bucket_array[8] != 1) {
            die "ERROR: Did not find $ms_value in bucket 16 - <20ms";  } }#16 - <20ms
        case [20..29] {if ($bucket_array[9] != 1) {
            die "ERROR: Did not find $ms_value in bucket 20 - <30ms";  } }#20 - <30ms
        case [30..39] {if ($bucket_array[10] != 1) {
            die "ERROR: Did not find $ms_value in bucket 30 - <40ms";  } }#30 - <40ms
        case [40..49] {if ($bucket_array[11] != 1) {
            die "ERROR: Did not find $ms_value in bucket 40 - <50ms";  } }#40 - <50ms
        case [50..59] {if ($bucket_array[12] != 1) {
            die "ERROR: Did not find $ms_value in bucket 50 - <60ms";  } }#50 - <60ms
        case [60..69] {if ($bucket_array[13] != 1) {
            die "ERROR: Did not find $ms_value in bucket 60 - <70ms";  } }#60 - <70ms
        case [70..79] {if ($bucket_array[14] != 1) {
            die "ERROR: Did not find $ms_value in bucket 70 - <80ms";  } }#70 - <80ms
        case [80..89] {if ($bucket_array[15] != 1) {
            die "ERROR: Did not find $ms_value in bucket 80 - <90ms";  } }#80 - <90ms
        case [90..99] {if ($bucket_array[16] != 1) {
            die "ERROR: Did not find $ms_value in bucket 90 - <100ms";  } }#90 - <100ms
        case [100..119] {if ($bucket_array[17] != 1) {
            die "ERROR: Did not find $ms_value in bucket 100 - <120ms";  } }#100 - <120ms
        case [120..139] {if ($bucket_array[18] != 1) {
            die "ERROR: Did not find $ms_value in bucket 120 - <140ms";  } }#120 - <140ms
        case [140..159] {if ($bucket_array[19] != 1) {
            die "ERROR: Did not find $ms_value in bucket 140 - <160ms";  } }#140 - <160ms
        case [160..179] {if ($bucket_array[20] != 1) {
            die "ERROR: Did not find $ms_value in bucket 160 - <180ms";  } }#160 - <180ms
        case [180..199] {if ($bucket_array[21] != 1) {
            die "ERROR: Did not find $ms_value in bucket 180 - <200ms";  } }#180 - <200ms
        case [200..399] {if ($bucket_array[22] != 1) {
            die "ERROR: Did not find $ms_value in bucket 200 - <400ms";  } }#200 - <400ms
        case [400..599] {if ($bucket_array[23] != 1) {
            die "ERROR: Did not find $ms_value in bucket 400 - <600ms";  } }#400 - <600ms
        case [600..799] {if ($bucket_array[24] != 1) {
            die "ERROR: Did not find $ms_value in bucket 600 - <800ms";  } }#600 - <800ms
        case [800..999] {if ($bucket_array[25] != 1) {
            die "ERROR: Did not find $ms_value in bucket 800 - <1000ms";  } }#800 - <1000ms
        case [1000..1499] {if ($bucket_array[26] != 1) {
            die "ERROR: Did not find $ms_value in bucket 1000 - <1500ms";  } }#1000 - <1500ms
        case [1500..1999] {if ($bucket_array[27] != 1) {
            die "ERROR: Did not find $ms_value in bucket 1500 - <2000ms";  } }#1500 - <2000ms
        case [2000..2499] {if ($bucket_array[28] != 1) {
            die "ERROR: Did not find $ms_value in bucket 2000 - <2500ms";  } }#2000 - <2500ms
        case [2500..2999] {if ($bucket_array[29] != 1) {
            die "ERROR: Did not find $ms_value in bucket 2500 - <3000ms";  } }#2500 - <3000ms
        case [3000..3499] {if ($bucket_array[30] != 1) {
            die "ERROR: Did not find $ms_value in bucket 3000 - <3500ms";  } }#3000 - <3500ms
        case [3500..3999] {if ($bucket_array[31] != 1) {
            die "ERROR: Did not find $ms_value in bucket 3500 - <4000ms";  } }#3500 - <4000ms
        case [4000..7999] {if ($bucket_array[32] != 1) {
            die "ERROR: Did not find $ms_value in bucket 4000 - <8000ms";  } }#4000 - <8000ms
        case [8000..11999] {if ($bucket_array[33] != 1) {
            die "ERROR: Did not find $ms_value in bucket 8000 - <12000ms";  } }#8000 - <12000ms
        case [12000..15999] {if ($bucket_array[34] != 1) {
            die "ERROR: Did not find $ms_value in bucket 12000 - <16000ms";  } }#12000 - <16000ms
        case [16000..99999] {if ($bucket_array[35] != 1) {
            die "ERROR: Did not find $ms_value in bucket >16000ms";  } }#>16000ms
        #case [16000..2147483647] gives a memory wrap failure#>16000ms
    }
}
1

