# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/CsiUtils.pm#1 $
#
#Copyright (c) 2009 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines for performing various kinds of validation on cluster aggregates/volumes.
##
## @description
##    This library helps you in performing various kinds of validation on cluster aggregates/volumes .
##
## @synopsis
##    use CsiUtils;
##
## @examples
##    use CsiUtils qw(validate_items);
##    $Server->volume_new( vserver    => $Vserv_name,
##                         volume     => $Vol_name,
##                         aggregate  => $Aggr_name,
##                        ) or die "Unable to create volume";
##
##    &validate_items('perform_validation' => "volume",
##                     server              => $Server,
##                     volume              => $Vol_name,
##                     vserver             => $Vserv_name,
##                     aggregate           => $Aggr_name
##                    ) or die "Validation of Volume failed:$@";
##
## @author neelkanth.raju@netapp.com,preethi.kumar@netapp.com

package CsiUtils;

use strict;
use Tharn;
use Data::Dumper;
use Exporter qw(import);
use filer qw(filer_cli_priv filer_cli_say filer_cli_diag_say);
use File::Temp;
use Os::OsUtils qw(myConnectPlus myDisconnectPlus startShell);
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF HASHREF UNDEF GLOBREF);

BEGIN {
    our @EXPORT = qw(&chk_aggr
      &chk_license
      &chk_vserv
      &cmd_wait_or_check_state
      &execute_in_shell
      &get_members
      &nfs_mount_delete
      &nfs_mount_setup
      &random_name_generate
      &validate_items
      &filer_cmd
      &filer_diag_cmd);

      use vars qw(
        $Destination
      );

      $Destination = "0.0.0.0/0";

} ## end BEGIN

#-------------------------------------------------------------------------------------------
# name ::       validate_items
# summary  :   Performs exhaustive validation on aggregates or volumes (both flex and acro)
#	       and member aggregates or volumes based on request.
#
# Required options:
#               server => Server object
#               perform_validation => volume, aggregate, member_volumes, member_aggregates
#			Type of validation to perform.
#			Default : volume
#
# description :
#     'perform_validation' => "volume"
#             The subroutine will validate the below fields :
#             "size dsid msid state type volume-style volume-ownership and security-style"
#             in addition to the parameters which are passed to validate_items().
#             if any of the obove mentioned parameters are passed then the value recieved
#             is used to override the default values.
#     'perform_validation' => "aggregate"
#             The subroutine will validate the below fields :
#             "size uuid availsize usedsize diskcount chksumenabled inconsistent volume-style ha-policy"
#             in addition to the parameters which are passed to validate_items().
#             if any of the obove mentioned parameters are passed then the value recieved
#             is used to override the default values.
#
#     'perform_validation' => "member_volumes"
#	      values for below items are first taken from parent volume and then,
#	      used to validate all the corresponding member volumes.
#	      "size dsid msid state type volume-style volume-ownership and security-style"
#             if any of the obove mentioned parameters are passed then the value recieved
#             is used to override the default values.
#
#     'perform_validation' => "member_aggregates"
#             values for below items are first taken from parent aggregate and then,
#             used to validate all the corresponding member aggregates.
#             "size uuid availsize usedsize diskcount chksumenabled inconsistent volume-style ha-policy"
#             if any of the obove mentioned parameters are passed then the value recieved
#             is used to override the default values.
#
# Returns :     1 on success . "die's" on failure along with appropriate error message.
#-------------------------------------------------------------------------------------------
sub validate_items {
    my $Default_array_ref = [];

    #validating input parameters recieved with all possible valid keys.
    my %opts = validate(@_,
                        {"server" => {type => OBJECT, isa => 'Server'},
                         "perform_validation" =>
                           {type => SCALAR, default => "volume", optional => 1},
                         "aggregate" =>
                           {type => SCALAR, default => "", optional => 1},
                         "vserver" =>
                           {type => SCALAR, default => "", optional => 1},
                         "volume" =>
                           {type => SCALAR, default => "", optional => 1},
                         "size" =>
                           {type => SCALAR, default => "", optional => 1},
                         "name-ordinal" =>
                           {type => SCALAR, default => "", optional => 1},
                         "dsid" =>
                           {type => SCALAR, default => "", optional => 1},
                         "msid" =>
                           {type => SCALAR, default => "", optional => 1},
                         "state" =>
                           {type => SCALAR, default => "", optional => 1},
                         "type" =>
                           {type => SCALAR, default => "", optional => 1},
                         "volume-style" =>
                           {type => SCALAR, default => "", optional => 1},
                         "volume-ownership" =>
                           {type => SCALAR, default => "", optional => 1},
                         "policy" =>
                           {type => SCALAR, default => "", optional => 1},
                         "user" =>
                           {type => SCALAR, default => "", optional => 1},
                         "group" =>
                           {type => SCALAR, default => "", optional => 1},
                         "security-style" =>
                           {type => SCALAR, default => "", optional => 1},
                         "unix-permissions" =>
                           {type => SCALAR, default => "", optional => 1},
                         "junction-path" =>
                           {type => SCALAR, default => "", optional => 1},
                         "junction-path-source" =>
                           {type => SCALAR, default => "", optional => 1},
                         "junction-active" =>
                           {type => SCALAR, default => "", optional => 1},
                         "parent" =>
                           {type => SCALAR, default => "", optional => 1},
                         "vsroot" =>
                           {type => SCALAR, default => "", optional => 1},
                         "comment" =>
                           {type => SCALAR, default => "", optional => 1},
                         "available" =>
                           {type => SCALAR, default => "", optional => 1},
                         "total" =>
                           {type => SCALAR, default => "", optional => 1},
                         "used" =>
                           {type => SCALAR, default => "", optional => 1},
                         "percent-used" =>
                           {type => SCALAR, default => "", optional => 1},
                         "autosize" =>
                           {type => SCALAR, default => "", optional => 1},
                         "max-autosize" =>
                           {type => SCALAR, default => "", optional => 1},
                         "autosize-increment" =>
                           {type => SCALAR, default => "", optional => 1},
                         "files" =>
                           {type => SCALAR, default => "", optional => 1},
                         "files-used" =>
                           {type => SCALAR, default => "", optional => 1},
                         "maxdir-size" =>
                           {type => SCALAR, default => "", optional => 1},
                         "space-guarantee" =>
                           {type => SCALAR, default => "", optional => 1},
                         "space-guarantee-enabled" =>
                           {type => SCALAR, default => "", optional => 1},
                         "min-readahead" =>
                           {type => SCALAR, default => "", optional => 1},
                         "atime-update" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snapdir-access" =>
                           {type => SCALAR, default => "", optional => 1},
                         "percent-snapshot-space" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snapshot-space-used" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snapshot-policy" =>
                           {type => SCALAR, default => "", optional => 1},
                         "create-time" =>
                           {type => SCALAR, default => "", optional => 1},
                         "language" =>
                           {type => SCALAR, default => "", optional => 1},
                         "stripe-count" =>
                           {type => SCALAR, default => "", optional => 1},
                         "stripe-width" =>
                           {type => SCALAR, default => "", optional => 1},
                         "striping-epoch" =>
                           {type => SCALAR, default => "", optional => 1},
                         "lookup-cache" =>
                           {type => SCALAR, default => "", optional => 1},
                         "properly-striped" =>
                           {type => SCALAR, default => "", optional => 1},
                         "concurrency" =>
                           {type => SCALAR, default => "", optional => 1},
                         "optimize" =>
                           {type => SCALAR, default => "", optional => 1},
                         "foreground" =>
                           {type => SCALAR, default => "", optional => 1},
                         "clone-volume" =>
                           {type => SCALAR, default => "", optional => 1},
                         "antivirus-on-access-policy" =>
                           {type => SCALAR, default => "", optional => 1},
                         "node" =>
                           {type => SCALAR, default => "", optional => 1},
                         "clone-volume-name" =>
                           {type => SCALAR, default => "", optional => 1},
                         "cloned-snapshot" =>
                           {type => SCALAR, default => "", optional => 1},
                         "uuid" =>
                           {type => SCALAR, default => "", optional => 1},
                         "nvram-failed" =>
                           {type => SCALAR, default => "", optional => 1},
                         "filesys-size-fixed" =>
                           {type => SCALAR, default => "", optional => 1},
                         "inode-parent-path-trans-enabled" =>
                           {type => SCALAR, default => "", optional => 1},
                         "extent-enabled" =>
                           {type => SCALAR, default => "", optional => 1},
                         "fractional-reserve" =>
                           {type => SCALAR, default => "", optional => 1},
                         "auto-snapshots-enabled" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snap-mirror-enabled" =>
                           {type => SCALAR, default => "", optional => 1},
                         "ignore-inconsistent" =>
                           {type => SCALAR, default => "", optional => 1},
                         "stripe-format" =>
                           {type => SCALAR, default => "", optional => 1},
                         "index-dir-enabled" =>
                           {type => SCALAR, default => "", optional => 1},
                         "load-sharing-source" =>
                           {type => SCALAR, default => "", optional => 1},
                         "move-target" =>
                           {type => SCALAR, default => "", optional => 1},
                         "max-write-alloc-blocks" =>
                           {type => SCALAR, default => "", optional => 1},
                         "fsid" =>
                           {type => SCALAR, default => "", optional => 1},
                         "language-str" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snapshot-clone-dependency" =>
                           {type => SCALAR, default => "", optional => 1},
                         "create-ucode" =>
                           {type => SCALAR, default => "", optional => 1},
                         "convert-ucode" =>
                           {type => SCALAR, default => "", optional => 1},
                         "svo-enabled" =>
                           {type => SCALAR, default => "", optional => 1},
                         "svo-checksum-enabled" =>
                           {type => SCALAR, default => "", optional => 1},
                         "svo-reject-errors" =>
                           {type => SCALAR, default => "", optional => 1},
                         "svo-rman-allowed" =>
                           {type => SCALAR, default => "", optional => 1},
                         "disk-root" =>
                           {type => SCALAR, default => "", optional => 1},
                         "compression-enabled" =>
                           {type => SCALAR, default => "", optional => 1},
                         "space-mgmt-try-first" =>
                           {type => SCALAR, default => "", optional => 1},
                         "read-realloc" =>
                           {type => SCALAR, default => "", optional => 1},
                         "sched-snap-name" =>
                           {type => SCALAR, default => "", optional => 1},
                         "volume-check-status" =>
                           {type => SCALAR, default => "", optional => 1},
                         "volume-check-estimated-time-left" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snap-autodelete-enabled" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snap-autodelete-commitment" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snap-autodelete-defer-delete" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snap-autodelete-delete-order" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snap-autodelete-prefix" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snap-autodelete-target-free-space" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snap-autodelete-trigger" =>
                           {type => SCALAR, default => "", optional => 1},
                         "availsize" =>
                           {type => SCALAR, default => "", optional => 1},
                         "usedsize" =>
                           {type => SCALAR, default => "", optional => 1},
                         "nodes" => {type     => ARRAYREF,
                                     default  => $Default_array_ref,
                                     optional => 1},
                         "diskcount" =>
                           {type => SCALAR, default => "", optional => 1},
                         "disklist" => {type     => ARRAYREF,
                                        default  => $Default_array_ref,
                                        optional => 1},
                         "volcount" =>
                           {type => SCALAR, default => "", optional => 1},
                         "plexes" => {type     => ARRAYREF,
                                      default  => $Default_array_ref,
                                      optional => 1},
                         "raidgroups" => {type     => ARRAYREF,
                                          default  => $Default_array_ref,
                                          optional => 1},
                         "raidtype" =>
                           {type => SCALAR, default => "", optional => 1},
                         "maxraidsize" =>
                           {type => SCALAR, default => "", optional => 1},
                         "raidstatus" =>
                           {type => SCALAR, default => "", optional => 1},
                         "chksumenabled" =>
                           {type => SCALAR, default => "", optional => 1},
                         "chksumstatus" =>
                           {type => SCALAR, default => "", optional => 1},
                         "chksumstyle" =>
                           {type => SCALAR, default => "", optional => 1},
                         "inconsistent" =>
                           {type => SCALAR, default => "", optional => 1},
                         "ignore-inconsistent" =>
                           {type => SCALAR, default => "", optional => 1},
                         "raid-block-checksum" =>
                           {type => SCALAR, default => "", optional => 1},
                         "raid-zoned-checksum" =>
                           {type => SCALAR, default => "", optional => 1},
                         "thorough-scrub" =>
                           {type => SCALAR, default => "", optional => 1},
                         "voltype" =>
                           {type => SCALAR, default => "", optional => 1},
                         "disksize" =>
                           {type => SCALAR, default => "", optional => 1},
                         "disktype" =>
                           {type => SCALAR, default => "", optional => 1},
                         "diskrpm" =>
                           {type => SCALAR, default => "", optional => 1},
                         "allow-mixed" =>
                           {type => SCALAR, default => "", optional => 1},
                         "has-mroot" =>
                           {type => SCALAR, default => "", optional => 1},
                         "has-partner-mroot" =>
                           {type => SCALAR, default => "", optional => 1},
                         "files-total" =>
                           {type => SCALAR, default => "", optional => 1},
                         "files-used" =>
                           {type => SCALAR, default => "", optional => 1},
                         "is-snaplock" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snaplock-type" =>
                           {type => SCALAR, default => "", optional => 1},
                         "mirror-status" =>
                           {type => SCALAR, default => "", optional => 1},
                         "plex-count" =>
                           {type => SCALAR, default => "", optional => 1},
                         "snap-mirrored" =>
                           {type => SCALAR, default => "", optional => 1},
                         "filesys-size-fixed" =>
                           {type => SCALAR, default => "", optional => 1},
                         "auto-snapshots" =>
                           {type => SCALAR, default => "", optional => 1},
                         "resync-snap-time" =>
                           {type => SCALAR, default => "", optional => 1},
                         "root" =>
                           {type => SCALAR, default => "", optional => 1},
                         "iron-status" =>
                           {type => SCALAR, default => "", optional => 1},
                         "iron-scan-percentage" =>
                           {type => SCALAR, default => "", optional => 1},
                         "iron-last-start-errno" =>
                           {type => SCALAR, default => "", optional => 1},
                         "iron-last-start-error-info" =>
                           {type => SCALAR, default => "", optional => 1},
                         "volcount-quiesced" =>
                           {type => SCALAR, default => "", optional => 1},
                         "volcount-not-online" =>
                           {type => SCALAR, default => "", optional => 1},
                         "mirror-count-ls-destinations" =>
                           {type => SCALAR, default => "", optional => 1},
                         "mirror-count-dp-destinations" =>
                           {type => SCALAR, default => "", optional => 1},
                         "mirror-count-move-destinations" =>
                           {type => SCALAR, default => "", optional => 1},
                         "mirror-count-dp-qtree-destinations" =>
                           {type => SCALAR, default => "", optional => 1},
                         "ha-policy" =>
                           {type => SCALAR, default => "", optional => 1},
                         "mount-state" =>
                           {type => SCALAR, default => "", optional => 1},
                         "block-type" =>
                           {type => SCALAR, default => "", optional => 1},
                         "home-id" =>
                           {type => SCALAR, default => "", optional => 1},
                         "home-name" =>
                           {type => SCALAR, default => "", optional => 1},
                         "owner-id" =>
                           {type => SCALAR, default => "", optional => 1},
                         "owner-name" =>
                           {type => SCALAR, default => "", optional => 1},
                         "striping" =>
                           {type => SCALAR, default => "", optional => 1},
                         "isphantom" =>
                           {type => SCALAR, default => "", optional => 1},
                         "rawvolcount" =>
                           {type => SCALAR, default => "", optional => 1},
                         "striped-volcount" =>
                           {type => SCALAR, default => "", optional => 1},
                         "collective-volcount" =>
                           {type => SCALAR, default => "", optional => 1}});
    my %received_opts = @_;

    eval {
        if ($opts{perform_validation} eq "volume") {
            delete $received_opts{perform_validation}
              if (exists $received_opts{perform_validation});
            validate_volume(%received_opts);

        } elsif ($opts{perform_validation} eq "aggregate") {
            delete $received_opts{perform_validation};
            validate_aggregate(%received_opts);

        } elsif ($opts{perform_validation} eq "member_volumes") {
            delete $received_opts{perform_validation};
            validate_member_volumes(%received_opts);

        } elsif ($opts{perform_validation} eq "member_aggregates") {
            delete $received_opts{perform_validation};
            validate_member_aggregates(%received_opts);

        } else {
            die "Improper Value perform_validation => $opts{perform_validation}"
              . " passed in call to subroutine CsiUtils::validate_items";
        }
    };
    return 0 if ($@);
    return 1;
} ## end sub validate_items

#===============================================================================
#  populate_defaults
#  - internally called by validate_volume,validate_member_volumes, etc.
#  - Returns a populated hash of items to be validated based on input parameters
#===============================================================================
sub populate_defaults {
    my %opts   = @_;
    my $server = $opts{server};

    if (exists $opts{aggregate}) {
        if ($opts{populate} eq "standard") {

            #default parameters to validate an aggregate
            return ("size"          => "\\d+",
                    "uuid"          => '[\w\-]+',
                    "availsize"     => "\\d+",
                    "usedsize"      => "\\d+",
                    "diskcount"     => "\\d+",
                    "chksumenabled" => "true|false",
                    "inconsistent"  => "true|false",
                    "volume-style"  => "flex|striped",
                    "ha-policy"     => "sfo|cfo",);
        } ## end if ($opts{populate} eq...
        if ($opts{populate} eq "true") {

            # some parameters will be populated from $ops{aggregate} to validate its members
            # rest will be taken as default
            my $aggr_output;
            eval {
                $aggr_output = $server->storage_aggregate_show(
                                                 aggregate => $opts{aggregate});
            };
            die "Aggregate show failed: $@" if ($@);

            my %ret_hash;
            my $aggr = $opts{aggregate} . "_";

            #populating member names from plexes data field.
            my @members =
              get_members(aggregate => $opts{aggregate},
                          server    => $server)
              or die
              "Could not find any members for aggregate :  $opts{aggregate} : $@";

            my $no_of_members = @members;

            $ret_hash{members} = \@members;
            $ret_hash{"size"}  = "\\d+";
            $ret_hash{"state"} = ${$$aggr_output[0]}{state};
            eval {
                $ret_hash{"node"} = join("|", @{${$$aggr_output[0]}{nodes}});
            };
            if ($@ =~ /ARRAY ref/) {
                $ret_hash{"node"} = ${$$aggr_output[0]}{nodes};
            }
            $ret_hash{"disklist"} = ${$$aggr_output[0]}{disklist};
            $ret_hash{"diskcount"} =
              ${$$aggr_output[0]}{diskcount} / $no_of_members;
            $ret_hash{"chksumenabled"} = ${$$aggr_output[0]}{chksumenabled};
            $ret_hash{"inconsistent"}  = ${$$aggr_output[0]}{inconsistent};
            $ret_hash{"ha-policy"}     = ${$$aggr_output[0]}{"ha-policy"};
            return %ret_hash;

        } ## end if ($opts{populate} eq...
    } elsif (exists $opts{volume}) {
        if ($opts{populate} eq "standard") {

            #default parameters to validate a volume
            return ("size"         => "\\d+",
                    "dsid"         => "\\d+",
                    "msid"         => "\\d+",
                    "state"        => "online|restricted|offline|force-online",
                    "type"         => "RW|LS|DP|TMP",
                    "volume-style" => "flex|striped",
                    "volume-ownership" => "node|cluster",
                    "security-style"   => "unix|ntfs|mixed",);
        } ## end if ($opts{populate} eq...
        if ($opts{populate} eq "true") {

            # parameters will be populated from $ops{volume} to validate its members
            # rest will be taken as default
            my $vol_output;
            eval {
                $vol_output = $server->volume_show(volume => $opts{volume});
            };
            die "Volume show failed : $@" if ($@);

            my %ret_hash;
            my @members =
              get_members(volume => $opts{volume},
                          server => $server)
              or die
              "Could not find any members for volume :  $opts{volume} : $@";

            $ret_hash{members} = \@members;
            $ret_hash{dsid}    = "\\d+",
              $ret_hash{msid}  = "\\d+",
              $ret_hash{size}  = "\\d+";
            $ret_hash{state}          = ${$$vol_output[0]}{state};
            $ret_hash{"volume-style"} = ${$$vol_output[0]}{"volume-style"};
            $ret_hash{type}           = ${$$vol_output[0]}{type};
            $ret_hash{"volume-ownership"} =
              ${$$vol_output[0]}{"volume-ownership"};
            $ret_hash{"security-style"} = ${$$vol_output[0]}{"security-style"};
            return %ret_hash;
        } ## end if ($opts{populate} eq...
    } ## end elsif (exists $opts{volume...
} ## end sub populate_defaults

#===============================================================================
#  validate_volume
#  - internally called by validate_items, to perform volume parameters validation
#===============================================================================
sub validate_volume {

    my %params = @_;
    my $server = $params{server};
    delete $params{server};

    die "Please provide parameter <volume>" unless (exists $params{volume});

    my %defaults = populate_defaults(volume   => $params{volume},
                                     server   => $server,
                                     populate => "standard") or die "$@";
    my @err_strings;
    my ($exp_val, $field, $act_val);
    my $map_opts = $server->volume_show(volume  => $params{volume},
                                        vserver => $params{vserver});

    # validate for parameters requested by user
    foreach $field (keys(%params)) {
        $exp_val = $params{$field};
        $act_val = ${$$map_opts[0]}{$field};
        unless ($act_val =~ /$exp_val/) {
            push @err_strings, "Expected Parameter [ $field => $exp_val ] did"
              . "not match with \nActual Parameter [ $field => $act_val ]";
        } else {
            logcomment("Expected Parameter [ $field => $exp_val ] "
                    . "matched with \nActual Parameter [ $field => $act_val ]");
        }

        #removing parameter from default expected list if already present.
        delete $defaults{$field} if (exists $defaults{$field});
    } ## end foreach $field (keys(%params...

    # validate for parameters set as default
    foreach $field (keys(%defaults)) {
        $exp_val = $defaults{$field};
        $act_val = ${$$map_opts[0]}{$field};
        unless ($act_val =~ /^$exp_val$/) {
            push @err_strings, "Expected Parameter [ $field => $exp_val ] did"
              . "not match with \nActual Parameter [ $field => $act_val ]";
        } else {
            logcomment("Expected Parameter [ $field => $exp_val ] "
                    . "matched with \nActual Parameter [ $field => $act_val ]");
        }
    } ## end foreach $field (keys(%defaults...
    die "@err_strings" if (@err_strings);
    return 1;
} ## end sub validate_volume

#===============================================================================
#  validate_member_volumes
#  - internally called by validate_items, to perform volume menber parameters validation
#===============================================================================
sub validate_member_volumes {

    my %params = @_;
    my $server = $params{server};
    delete $params{server};

    my %mapped_opts = "";
    my %defaults = populate_defaults(volume   => $params{volume},
                                     server   => $server,
                                     populate => "true") or die "$@";

    my @members = @{$defaults{members}};
    delete $defaults{members};
    my @err_strings;

    #Deleting name of volume from parameters to be matched.
    delete $params{volume};

    #Validating all member volumes
    foreach my $member (@members) {
        logcomment("Validating parameters of member Volume : $member\n");

        my $map_opts = $server->volume_member_show(volume => $member);

        # validate for parameters requested by user
        foreach my $field (keys(%params)) {
            my $exp_val = $params{$field};

            #removing parameter from default expected list if already present.
            delete $defaults{$field} if (exists $defaults{$field});
            my $act_val = ${$$map_opts[0]}{$field};
            unless ($act_val =~ /^$exp_val$/) {
                push @err_strings,
                    "Expected Parameter [ $field => $exp_val ] did"
                  . "not match with \nActual Parameter [ $field => $act_val ]"
                  . " For Member Volume: $member\n";
            } else {
                logcomment("Expected Parameter [ $field => $exp_val ] matched "
                         . "with \n Actual Parameter [ $field => $act_val ]\n");
            }
        } ## end foreach my $field (keys(%params...

        # validate for parameters set as default
        foreach my $field (keys(%defaults)) {
            my $exp_val = $defaults{$field};
            my $act_val = ${$$map_opts[0]}{$field};
            unless ($act_val =~ /^$exp_val$/) {
                push @err_strings,
                    "Expected Parameter [ $field => $exp_val ] did"
                  . "not match with \nActual Parameter [ $field => $act_val ]"
                  . " For Member Volume: $member\n";
            } else {
                logcomment("Expected Parameter [ $field => $exp_val ] matched "
                         . "with \n Actual Parameter [ $field => $act_val ]\n");
            }
        } ## end foreach my $field (keys(%defaults...
    } ## end foreach my $member (@members)
    die "@err_strings" if (@err_strings);
    return 1;
} ## end sub validate_member_volumes

#===============================================================================
#  validate_aggregate
#  - internally called by validate_items, to perform aggregate parameters validation
#===============================================================================
sub validate_aggregate {

    my %params = @_;
    my $server = $params{server};
    delete $params{server};

    die "Please provide parameter <aggregate>"
      unless (exists $params{aggregate});

    my %defaults = populate_defaults(aggregate => $params{aggregate},
                                     server    => $server,
                                     populate  => "standard") or die "$@";
    my @err_strings;
    my ($exp_val, $field, $act_val, $disk, $node);

    my $map_opts =
      $server->storage_aggregate_show(aggregate => $params{aggregate});

    # validate for parameters requested by user
    #validate for reference type parameter disklist in param hash
    if (exists $params{disklist}) {
        logcomment(
             "Validating parameter disklist of aggregate : $params{aggregate}\n"
        );
        my $act_d_list = join(" | ", @{${$$map_opts[0]}{disklist}});
        foreach my $disk (@{$params{disklist}}) {
            unless ($act_d_list =~ /$disk/) {
                push @err_strings,
                    "Expected Parameter [ disklist => $disk ]"
                  . "is absent in \n Actual Parameter [ disklist => "
                  . "$act_d_list ] For aggregate: $params{aggregate}\n";
            } else {
                logcomment(  "Actual Parameter [ disklist =>"
                           . " $act_d_list ] contains \n"
                           . "Expected Parameter [ disklist => $disk ]\n");
            }
        } ## end foreach my $disk (@{$params...
        delete $params{disklist};
    } ## end if (exists $params{disklist...

    #validate for reference type parameter nodes in param hash
    if (exists $params{nodes}) {
        logcomment(
              "Validating parameter nodes of aggregate : $params{aggregate}\n");
        my $act_n_list = join("|", @{${$$map_opts[0]}{nodes}});
        foreach $node (@{$params{nodes}}) {
            unless ($act_n_list =~ s/$node//) {
                push @err_strings,
                    "Expected Parameter [ nodes => $node ]"
                  . "is absent in \n Actual Parameter [ nodes => "
                  . " @{${$$map_opts[0]}{nodes}} ] For aggregate: $params{aggregate}\n";
            } else {
                logcomment(  "Actual Parameter [ nodes =>"
                           . " @{${$$map_opts[0]}{nodes}} ] contains "
                           . "\n Expected Parameter [ nodes => $node ]\n");
            }
        } ## end foreach $node (@{$params{nodes...
        delete $params{nodes};
    } ## end if (exists $params{nodes...

    foreach $field (keys(%params)) {
        $exp_val = $params{$field};

        #removing parameter from default expected list if already present.
        delete $defaults{$field} if (exists $defaults{$field});
        $act_val = ${$$map_opts[0]}{$field};
        unless ($act_val =~ /^$exp_val$/) {
            push @err_strings, "Expected Parameter [ $field => $exp_val ] did"
              . "not match with \nActual Parameter [ $field => $act_val ]";
        } else {
            logcomment("Expected Parameter [ $field => $exp_val ] "
                    . "matched with \nActual Parameter [ $field => $act_val ]");
        }
    } ## end foreach $field (keys(%params...

    # validate for parameters set as default
    foreach $field (keys(%defaults)) {
        $exp_val = $defaults{$field};
        $act_val = ${$$map_opts[0]}{$field};
        unless ($act_val =~ /^$exp_val$/) {
            push @err_strings, "Expected Parameter [ $field => $exp_val ] did"
              . "not match with \nActual Parameter [ $field => $act_val ]";
        } else {
            logcomment("Expected Parameter [ $field => $exp_val ] "
                    . "matched with \nActual Parameter [ $field => $act_val ]");
        }
    } ## end foreach $field (keys(%defaults...
    die "@err_strings" if (@err_strings);
    return 1;
} ## end sub validate_aggregate

#===============================================================================
#  validate_member_aggregates
#  - internally called by validate_items, to perform aggregatic memebers parameters validation
#===============================================================================
sub validate_member_aggregates {

    my %params = @_;
    my $server = $params{server};
    delete $params{server};

    die "Please provide parameter <aggregate>"
      unless (exists $params{aggregate});

    my %defaults = populate_defaults(aggregate => $params{aggregate},
                                     server    => $server,
                                     populate  => "true") or die "$@";

    my @members = @{$defaults{members}};
    delete $defaults{members};
    my @err_strings;

    #Deleting name of aggregate from parameters to be matched.
    delete $params{aggregate};

    #Validating all member aggregates
    foreach my $member (@members) {
        logcomment(
               "Validating parameter disklist of member aggregate : $member\n");

        my $map_opts =
          $server->storage_aggregate_member_show(aggregate => $member)
          or die "$@";

        # validate for parameters requested by user

        # validate for reference type parameter disklist in parameter hash.
        if (exists $params{disklist}) {
            logcomment(
                "Validating parameter disklist of aggregate : $params{aggregate}\n"
            );
            my $act_d_list = join(" | ", @{${$$map_opts[0]}{disklist}});
            foreach my $disk (@{$params{disklist}}) {
                unless ($act_d_list =~ /$disk/) {
                    push @err_strings,
                        "Expected Parameter [ disklist => $disk ]"
                      . "is absent in \n Actual Parameter [ disklist => "
                      . "$act_d_list ] For aggregate: $params{aggregate}\n";
                } else {
                    logcomment(  "Actual Parameter [ disklist =>"
                               . " $act_d_list ] contains \n"
                               . "Expected Parameter [ disklist => $disk ]\n");
                }
            } ## end foreach my $disk (@{$params...
            delete $defaults{disklist} if (exists $defaults{disklist});
        } ## end if (exists $params{disklist...

        #validate for reference type parameter disklist in default hash
        if (exists $defaults{disklist}) {
            my $default_d_list = join("|", @{$defaults{disklist}});
            foreach my $disk (@{${$$map_opts[0]}{disklist}}) {
                unless ($default_d_list =~ /$disk/) {
                    push @err_strings,
                        "Actual Parameter [ disklist => $disk ]"
                      . "is absent in \n Expected Parameter [ disklist => "
                      . "$default_d_list ] For Member aggregate: $member\n";
                } else {
                    logcomment(  "Expected Parameter [ disklist =>"
                               . " $default_d_list ] contains \n"
                               . "Actual Parameter [ disklist => $disk ]\n");
                }
            } ## end foreach my $disk (@{${$$map_opts...
        } ## end if (exists $defaults{disklist...
    } ## end foreach my $member (@members)

    delete $params{disklist}   if (exists $params{disklist});
    delete $defaults{disklist} if (exists $defaults{disklist});

    foreach my $member (@members) {
        logcomment("Validating parameters of member aggregate : $member\n");

        my $map_opts =
          $server->storage_aggregate_member_show(aggregate => $member)
          or die "$@";

        # validate for parameters requested by user

        foreach my $field (keys(%params)) {
            my $exp_val = $params{$field};

            #removing parameter from default expected list if already present.
            delete $defaults{$field} if (exists $defaults{$field});
            my $act_val = ${$$map_opts[0]}{$field};
            unless ($act_val =~ /^$exp_val$/) {
                push @err_strings,
                    "Expected Parameter [ $field => $exp_val ] did"
                  . "not match with \nActual Parameter [ $field => $act_val ]"
                  . " For Member aggregate: $member\n";
            } else {
                logcomment("Expected Parameter [ $field => $exp_val ] matched "
                         . "with \n Actual Parameter [ $field => $act_val ]\n");
            }
        } ## end foreach my $field (keys(%params...

        # validate for parameters set as default
        foreach my $field (keys(%defaults)) {
            my $exp_val = $defaults{$field};
            my $act_val = ${$$map_opts[0]}{$field};
            unless ($act_val =~ /^$exp_val$/) {
                push @err_strings,
                    "Expected Parameter [ $field => $exp_val ] did"
                  . "not match with \nActual Parameter [ $field => $act_val ]"
                  . " For Member aggregate: $member\n";
            } else {
                logcomment(" Expected Parameter [ $field => $exp_val ] matched "
                         . "with \n Actual Parameter [ $field => $act_val ]\n");
            }
        } ## end foreach my $field (keys(%defaults...
    } ## end foreach my $member (@members)

    die "validate_member_aggregates() FAILED:\n@err_strings" if (@err_strings);
    return 1;
} ## end sub validate_member_aggregates

#===============================================================================
#  get_members
#===============================================================================
sub get_members {
    my %opts   = @_;
    my $server = $opts{server};
    delete $opts{server};

    my @members;
    my @member_output;
    if (exists $opts{aggregate}) {
        eval {
            @member_output = ();
            @member_output = $server->storage_aggregate_member_show();
        };
        die "$@" if ($@);
        foreach my $member (@member_output) {
            if (${$member}{aggregate} =~ /^$opts{aggregate}_\d\d\d$/) {
                push(@members, ${$member}{aggregate});
            }
        }
        die "There are no entries matching your query" unless (@members);
        return @members;
    } ## end if (exists $opts{aggregate...
    if (exists $opts{volume}) {
        eval {
            @member_output = ();
            @member_output = $server->volume_member_show();
        };
        die "$@" if ($@);
        foreach my $member (@member_output) {
            push(@members, ${$member}{volume})
              if (${$member}{volume} eq $opts{volume});
            if (${$member}{volume} =~ /^$opts{volume}_\d\d\d$/) {
                push(@members, ${$member}{volume});
            }
        }
        die "There are no entries matching your query" unless (@members);
        return @members;
    } ## end if (exists $opts{volume...
} ## end sub get_members

#---------------------------------------------------------------------------------------------------
## name ::     random_name_generate()
## summary  :  This subroutine returns random name prefixed with the argument.
## Usage       random_name_generate(type=>"aggr")
## Required options:  Prefix string that needs to be appended to the randomly generated string.
## Returns     random string. For eg: "<prefix_string>_<random_string>"
#---------------------------------------------------------------------------------------------------
sub random_name_generate {
    my ($name, %opts, $file_name, $file);
    (%opts) = validate(@_, {type => {type => SCALAR}});
    $name = $opts{type};

    #tmpnam returns the complete path of the file which is going to be created
    $file = tmpnam();

    ($file_name) = $file =~ m/.*\/(.*)$/;

    return "$name" . '_' . "$file_name";
} ## end sub random_name_generate

#--------------------------------------------------------------------------------
# name ::       cmd_wait_or_check_state
# summary  :    Executes a particular type of show cmd and polls or check for the given 'field'
#               to reach the given 'value' depending on value of parameter 'wait' .
# Required options:
#               server => Server object
#               cmd    => command to execute
#               cmd implemented so far:
#                       aggregate_show => storage_aggregate_show
#                       volume_show    => volume_show
#                       aggregate_mem_show => aggregate_member_show
#                       volume_mem_show => volume_member_show
#               options=> reference of arguments to be passed to the show function
#               field  => attribute of show for which polling has to be done
#               value  => state of field to wait for..
#               wait   => 1 if polling required
#                         0 if only check required.(default)
#-------------------------------------------------------------------------------
sub cmd_wait_or_check_state (%) {
    my %opts    = @_;
    my $server  = $opts{server};
    my $field   = $opts{field};
    my $value   = $opts{value};
    my $cmd     = $opts{cmd};
    my $options = $opts{options};
    my ($t_val, $output, $flag) = (0, "", 1);
    my ($timeout);
    my @option = %$options;

    if ($opts{wait}) {
        $timeout = $opts{timeout} ? $opts{timeout} : 300;
    } else {
        $timeout = 0;
    }
    if ($cmd =~ /aggregate_show/) {
        eval { $output = $server->storage_aggregate_show(@option); };
    } elsif ($cmd =~ /volume_show/) {
        eval { $output = $server->volume_show(@option); };
    } elsif ($cmd =~ /volume_mem_show/) {
        eval { $output = $server->volume_member_show(@option); };
    } elsif ($cmd =~ /aggregate_mem_show/) {
        eval { $output = $server->storage_aggregate_member_show(@option); };
    }
    die "$@" if ($@);

    $t_val = time;
    $flag = 0 if (${$$output[0]}{$field} =~ /$value/);
    $timeout += $t_val;

    while ($flag and $t_val < $timeout) {
        if ($cmd =~ /aggregate_show/) {
            eval { $output = $server->storage_aggregate_show(@option); };
        } elsif ($cmd =~ /volume_show/) {
            eval { $output = $server->volume_show(@option); };
        } elsif ($cmd =~ /volume_mem_show/) {
            eval { $output = $server->volume_member_show(@option); };
        } elsif ($cmd =~ /aggregate_mem_show/) {
            eval { $output = $server->storage_aggregate_member_show(@option); };
        }

        die "$@" if ($@);
        $flag = 0 if (${$$output[0]}{$field} =~ /$value/);
        $t_val = time;
        sleep 5;    # delay between polling
    } ## end while ($flag and $t_val <...

    if ($flag and $t_val >= $timeout) {
        logcomment("Waiting for $field to reach $value Timed Out ");
        return 0;
    }
    return 1;
} ## end sub cmd_wait_or_check_state (%)
################################################################################

#--------------------------------------------------------------------------------
# name ::       chk_aggr
# summary  :    This subroutine returns the list of aggregate based on the
#               volume style of the aggregate
# Required options:
#               server => Server object
#               style  => Volume style
#-------------------------------------------------------------------------------
sub chk_aggr {
    my %opts          = @_;
    my $style         = $opts{style};
    my $server        = $opts{server};
    my $aggr_del_flag = 0;
    my (@aggr_present, @aggr_name, $aggr_list, $aggr_name);
    my $len = 0;
    eval {
        @aggr_present =
          $server->storage_aggregate_show("volume-style" => $style,
                                          root           => "false",
                                          aggregate      => "!p_*");
        foreach my $name (@aggr_present) {
            push(@aggr_name, $name->{aggregate});
        }
        $len = @aggr_name;

    };
    if (($@) || ($len == 0)) {
        if ($style eq "flex") {
            eval {
                logcomment("Generating Aggregate Name");
                $aggr_name = random_name_generate(type => "TEST_AGGR");
                $server->storage_aggregate_new(aggregate => $aggr_name,
                                               diskcount => 4);

            };

            logcomment("===Waiting till the Aggregate Name is Online===");
            cmd_wait_or_check_state(server  => $server,
                                    cmd     => "storage_aggregate_show",
                                    options => {aggregate => $aggr_name},
                                    wait    => 1,
                                    field   => 'state',
                                    value   => 'online',
                                    timeout => 2000
                                   ) or die "Failed the get the aggregate $aggr_name online";
        } elsif ($style eq "striped") {
            eval {
                my (@nodes, @Nodes);
                @nodes = $server->servers();
                if ($#nodes == 0) {
                    @Nodes = ($nodes[0], $nodes[0]);
                } else {
                    @Nodes = ($nodes[0], $nodes[1]);
                }
                logcomment("Generating Aggregate Name");
                $aggr_name = random_name_generate(type => "TEST_AGGR");
                $server->storage_aggregate_new(aggregate      => $aggr_name,
                                               nodes          => \@Nodes,
                                               diskcount      => 6,
                                               "volume-style" => $style
                                              ) or die " Unable to create aggregate ";

            };

            logcomment("===Waiting till the Aggregate Name is Online===");
            cmd_wait_or_check_state(server  => $server,
                                    cmd     => "storage_aggregate_show",
                                    options => {aggregate => $aggr_name},
                                    wait    => 1,
                                    field   => 'state',
                                    value   => 'online',
                                    timeout => 2000
                                   ) or die "Failed the get the aggregate $aggr_name online";

        } ## end elsif ($style eq "striped")
        $aggr_del_flag = 1;
        push(@aggr_name, $aggr_name);
    } ## end if (($@) || ($len == 0...

    return $aggr_del_flag, @aggr_name;

} ## end sub chk_aggr
################################################################################

#--------------------------------------------------------------------------------
# name ::       chk_vserv
# summary  :    This subroutine returns the list of vserver of type cluster
#               and the vserver with rootvolume
# Required options:
#               server => Server object
#-------------------------------------------------------------------------------
sub chk_vserv {
    my %opts   = @_;
    my $server = $opts{server};

    my (@vser_det,      @vser_name,    $vser_list, $vser_name,
        $root_vol_name, @aggr_present, @aggr_name, $aggr_list,
        $aggr_name);
    my (@rvol, $count, $len, @aggr_list);
    my $aggr_del_flag = 0;
    my $vser_del_flag = 0;
    eval {
        @vser_det = $server->vserver_show("type"   => "cluster",
                                           vserver => "!p_*");
        $vser_list     = $vser_det[0]->{vserver};
        $root_vol_name = $vser_det[0]->{rootvolume};
        if ($root_vol_name ne "-") {
            push(@vser_name, $vser_list);
        }
        $len = @vser_name;
        logcomment("The list of Vservers: @vser_name");

    };
    if (($@) || ($len == 0)) {
        ($aggr_del_flag, @aggr_list) =
          chk_aggr(server => $server,
                   style  => "flex");
        $aggr_name = $aggr_list[0];
        eval {
            logcomment("Generating Vserver Name");
            $vser_name = random_name_generate(type => "TEST_VSERVER");
            $root_vol_name = "TEST_ROOTVOL" . "$vser_name";
            $server->vserver_new(vserver                     => $vser_name,
                                 rootvolume                  => $root_vol_name,
                                 aggregate                   => $aggr_name,
                                 "ns-switch"                 => "file",
                                 "rootvolume-security-style" => "unix"
                                ) or die "Unable to create vserver";
        };
        $vser_del_flag = 1;
        push(@vser_name, $vser_name)

    } ## end if (($@) || ($len == 0... 
    
    return $vser_del_flag, $root_vol_name, @vser_name;

} ## end sub chk_vserv
################################################################################

#--------------------------------------------------------------------------------
# name ::       chk_license
# summary  :    This subroutine checks if the license is existing else will
#               add the license.
# Required options:
#               server => Server object
#-------------------------------------------------------------------------------
sub chk_license {
    my %opts   = @_;
    my $server = $opts{server};
    my ($out, $hash_val);
    my $field   = 'license-code';
    my @lic_arr = "";
    my $count   = 0;

    eval {
        logcomment("Checking whether the Required license is existing");
        $out = $server->system_license_exists();
        foreach $hash_val (@$out) {
            $hash_val = ${$$out[$count]}{$field};
            $count    = $count + 1;
            push(@lic_arr, $hash_val);
        }

        # Checking for the Striped Volume License.
        logcomment("Checking for the Striped Volume License");
        if (grep(/DTYSFKPJYBFDAA/, @lic_arr)) {
            logcomment("The Striped Volume license is existing ");
        } else {
            logcomment("Adding the Striped Volume license");
            $server->system_license_add('license-code' => " DTYSFKPJYBFDAA ",
                                        verification   => 0);
        }

        # Checking for the GX_FlexVol_HPO License.
        logcomment("Checking for the  GX_FlexVol_HPO License");
        if (grep(/XFOSBKPJYBFDAA/, @lic_arr)) {
            logcomment("The GX_FlexVol_HPO license is existing ");
        } else {
            logcomment("Adding the GX_FlexVol_HPO license");
            $server->system_license_add('license-code' => "XFOSBKPJYBFDAA",
                                        verification   => 0);
        }
    };

    if ($@) {
        logcomment("License Check exited with error:$@");
        return 0;
    }
    return 1;
} ## end sub chk_license
###############################################################################

############### _mount_setup ##################################################
# This sub routine performes mounting a volume on the client.
###############################################################################
sub nfs_mount_setup {
    my %opts = @_;

    my ($routing_group, @response, @cmd_output);
    my $server  = $opts{server};
    my @nodes   = $server->cluster_show;
    my $Node    = $nodes[1]->{node};
    my $Host    = host($Node) or die($@);
    my $Address = $Host->{DATA_IP1};
    my $Port    = $Host->{DATA_PORT1};
    my $Netmask = $Host->{DATA_MASK1};
    my $Gateway = $Host->{DATA_GATEWAY1};
    my $lif     = $opts{lif} ? $opts{lif} : "data1";
    $Destination  = $opts{destination} if( $opts{destination} );

    $Log->comment("====== \$Node ====> $Node\n");
    $Log->comment("====== \$Port ====> $Port\n");
    $Log->comment("====== \$Address ====> $Address\n");
    $Log->comment("====== \$Netmask ====> $Netmask\n");
    $Log->comment("====== \$Gateway ====> $Gateway\n");

    $Log->comment(  "====== Verifying existance of lif data1 ."
                  . " if present deleting the same.");
    eval { $server->network_interface_show(lif => "data1"); };
    if (!$@) {
        eval { $server->network_interface_delete(server => "*", lif => $lif); };
    }

    ###creating network interface
    eval {
        $server->network_interface_create(server         => $opts{vserver},
                                          lif            => $lif,
                                          role           => "data",
                                          'home-node'    => $Node,
                                          'home-port'    => $Port,
                                          address        => $Address,
                                          netmask        => $Netmask,
                                          'status-admin' => "up");
    };

    #if unable to create interface with data_ip1 then try using data_ip2.
    if ($@) {
        $Address = $Host->{DATA_IP2};
        $Port    = $Host->{DATA_PORT1};
        eval {
            $server->network_interface_create(server         => $opts{vserver},
                                              lif            => $lif,
                                              role           => "data",
                                              'home-node'    => $Node,
                                              'home-port'    => $Port,
                                              address        => $Address,
                                              netmask        => $Netmask,
                                              'status-admin' => "up");

        };
        die "FAIL: network interface Creation failed." if ($@);
    } ## end if ($@)

    @response = $server->network_routing_groups_show(server => $opts{vserver});
    $routing_group = ${$response[0]}{'routing-group'};
    $Log->comment("======The routing group is $routing_group");

    #### Creating Routing-Group route.

    eval {
        $server->network_routing_groups_route_create(
                                              server          => $opts{vserver},
                                              'routing-group' => $routing_group,
                                              destination     => $Destination,
                                              gateway         => $Gateway);
    };

    #####creating export policy
    eval {
        $server->vserver_export_policy_rule_create(vserver    => $opts{vserver},
                                                   policyname => "default",
                                                   clientmatch  => $Destination,
                                                   rorule       => "any",
                                                   rwrule       => "any",
                                                   superuser    => "any",
                                                   anon         => "anon",
                                                   'allow-suid' => "true",
                                                   'allow-dev'  => "true");

    };

    $server->system_license_add('license-code' => "TJFAEKPJYBFDAA");

    $server->vserver_nfs_create(vserver => $opts{vserver}, access => "true");

    return $Address;

} ## end sub nfs_mount_setup
################################################################################

####### nfs_mount_delete #######################################################
sub nfs_mount_delete {
    my %opts   = @_;
    my $server = $opts{server};
    my $lif    = $opts{lif} ? $opts{lif} : "data1";
    my (@response, $routing_group);

    ###deleting network interface
    eval {
        $server->network_interface_delete(server => $opts{vserver},
                                          lif    => $lif);
    };

    ##### Getting the value of Routing-group
    eval {
        @response =
          $server->network_routing_groups_show(server => $opts{vserver});
        $routing_group = ${$response[0]}{'routing-group'};
        $Log->comment("======The routing group is $routing_group");
    };

    #### Deleting Route
    eval {
        $server->network_routing_groups_route_delete(
                                              server          => $opts{vserver},
                                              'routing-group' => $routing_group,
                                              destination     => $Destination);

    };

    #### Deleting Routing-group
    eval {
        $server->network_routing_groups_delete(server => $opts{vserver},
                                               'routing-group' => $routing_group
                                              );
    };

    eval { $server->vserver_nfs_delete(vserver => $opts{vserver}); };

    return 1;
} ## end sub nfs_mount_delete
###############################################################################

#------------------------------------------------------------------------------
# name ::       execute_in_shell
# summary  :    Executes the command passed in systemshell
#               and validates  the output at users request.
#
# Required options:
#       common :
#               node => Node name on which to execute.
#               cmd => The command to execute in systemshell
# Optional options:
#       common :
#               exp_result => the regex to match with the output
#                             returned by executing "cmd"
#                             default : "";
#       7-mode :
#               conntype => type of connection to use to log in to filer.
#                           default : ssh
#               user     => username to log into system shell
#                           default : diag
#               password => Password to log in to systemshell
#                           default : netapp1!
# Returns : $result  if exp_result matches with $result.
#------------------------------------------------------------------------------

sub execute_in_shell {

    my %opts   = @_;

    # node parameter has to be passed if
    my $node = $opts{node} ? $opts{node}
               : die "Parameter node is not passed to subroutine";

    logcomment("Creating Host Object");
    my $host = host($node) or die($@);

    my $cmd =
        $opts{cmd}
      ? $opts{cmd}
      : die "command to execute not passed to subroutine";
    my $result;

    if ($host->hosttype() =~ m/filer-ng/) {
        my $server_obj = Server->new(nodes => $opts{node} ,  
                                     no_cli => 1,
                                     nodes_on_demand => 1 );

        $result = $server_obj->shell(cmd => $cmd);

        $server_obj->handle_connections(close => [qw(console_connection shell_connection)]);

        if (exists($opts{exp_result})) {
            if ($result !~ /$opts{exp_result}/) {
                die "FAIL: Incorrect verification response detected: $result";
            } else {
                logcomment("PASS: Verification Completed successfully");
            }
        }
        return $result;
    } ## end if ($host->hosttype() ...

    # Below steps run for 7-mode setup
    my ($connection, $user, $password, $conObj);

    $connection = $opts{conntype} ? $opts{conntype} : "ssh";
    $user       = $opts{user}     ? $opts{user}     : "diag";
    $password   = $opts{password} ? $opts{password} : "netapp1!";

    $conObj =
      myConnectPlus($node, "root", $host->default_password(), 1, $connection);
    filer_cli_priv ($conObj , "test");

    if (startShell($conObj, $node, $user, $password)) {
        eval { (undef,undef,$result,undef) = say($conObj, $cmd, -prompt ,"$node% ");
             };
        if (exists($opts{exp_result})) {
            if ($result !~ /$opts{exp_result}/) {
                die "FAIL: Incorrect verification response detected: $result";
            } else {
                logcomment("PASS: Verification Completed successfully");
            }
        }
        myDisconnectPlus($conObj, $connection);
        return $result;
    } ## end if (startShell($conObj...

    myDisconnectPlus($conObj, $connection);
    die "Unable to start System shell";

} ## end sub execute_in_shell
################################################################################
#------------------------------------------------------------------------------
# name ::       filer_cmd
# summary  :    Executes the filer cli command passed in and returns the output
#               of the CLI command 
#
# Required options:
#               conn_obj => Filer connection object
#               cmd  => The filer cli command to execute.
#
# Returns : $text 
#------------------------------------------------------------------------------

sub filer_cmd {

    my %opts   = @_;
    # node parameter has to be passed if
    my $conn_obj =
        $opts{conn_obj}
      ? $opts{conn_obj}
      : die "Parameter Connection Object is not passed to subroutine";

    my $cmd = 
        $opts{cmd}
      ? $opts{cmd}
      : die "Parameter Command is not passed to subroutine";
        
    my($text,$result) = filer_cli_say($conn_obj,$cmd);
    return $text;

} ## end sub filer_cmd
################################################################################
#------------------------------------------------------------------------------
# name ::       filer_diag_cmd
# summary  :    Executes the filer cli diag command passed in and returns the
#               output of the CLI command
#
# Required options:
#               conn_obj => Filer connection object
#               cmd  => The filer cli diag command to execute.
#
# Returns : $text
#------------------------------------------------------------------------------

sub filer_diag_cmd {

    my %opts   = @_;
    # node parameter has to be passed if
    my $conn_obj =
        $opts{conn_obj}
      ? $opts{conn_obj}
      : die "Parameter Connection Object is not passed to subroutine";

    my $cmd =
        $opts{cmd}
      ? $opts{cmd}
      : die "Parameter Command is not passed to subroutine";

    my($text,$result) = filer_cli_diag_say($conn_obj,$cmd);
    return $text;

} ## end sub filer_diag_cmd
################################################################################
return 1;
