#
#  $Id: $
#
#  Copyright (c) 2008 Network Appliance, Inc.
#  All rights reserved.
#

# -----------------------------------------------------------
##
## @summary Access functions for the d-snapshot-* DBlade ZAPIs
##
## @description
##	This file contains routines that let you exercise the
##	internal API of the filer using ssh access to the FreeBSD
##	shell command zsmcli.
##
##	Contains interface functions to d-blade Zapis that take MSIDs
##	or DSIDs as arguments or return values.
##
## @author shenry@netapp.com

=head Summary

Support routines for d-blade Zapi d-snapshot tests.

=head Description

This file contains routines that let you exercise the internal
(d-blade) snapshot API of the filer using ssh access to the FreeBSD
shell command zsmcli.

=head2 Author

shenry@netapp.com, dl-waflqa-art@netapp.com

=cut

package DBlade::dSnapshot;

use	strict;
use	Tharn;
use	Socket;
use	LWP::UserAgent;
use	XML::Parser;
use	NaElement;
use	SPARSE::Util::Logger qw( logDbg );
use	Data::Dumper;
use	DBlade::dZapi qw( fail_log );

use	vars ('@ISA', '@EXPORT');


=head2 new

  use DBlade::dSnapshot;
  my $zapi = new DBlade::dSnapshot($FILER, $user, $password);

Create a new instance of the dblade snapshot zapi interface.
All access is through member functions.

=over

=item $FILER

(Required): Hostspec or name of the converged filer.

=item $user

(Optional): Username to login to the filer's BSD shell with.
Defaults to "root".

=item $password

(Optional): password for the login account. Takes the password from
the nate_host file if not specified.

=back

=cut

#' emacs-fodder

sub new {
    my ($class, $server, $user, $password) = @_;

    my $zapi = new DBlade::dZapi($server, $user, $password);

    my $self = {
                zapi => $zapi,
               };
    bless $self, $class;
    return $self;
}


=head2 d_snapshot_create

  use DBlade::dSnapshot;
  my $zapi = new DBlade::dSnapshot($FILER, $user, $password);
  my $lsnapno = d_snapshot_create(%args);

Create a snapshot with the specified identity and characteristics.

Returns the logical-snap-id for the snapshot.

=over

=item 'async'

(Optional): If TRUE, returns immediately to caller. Defaults to FALSE

=item 'containing-aggregate-uuid'

(Optional): the identity of the aggregate hosting the volume.

=item 'is-hidden-snapshot'

(Optional): Snapshot created is invisible to clients. defaults to FALSE.
Only valid for constituent volume of striped volume.

=item 'logical-snap-id'

(Optional): The logical snapshot ID, visible in filehandles, etc.
Default is next available ID. Dies if ID is already in use.

=item 'quiesce'

(Optional): Whether the volume should be quiesced before taking the
snapshot.

=item 'snap-name'

(Required): The new snapshot's human-readable name.

=item 'vol-id'

(Required): DSID or UUID of the volume to create the snapshot in.

=back

=cut

sub d_snapshot_create ($%)
{
    my ($self, %args) = @_;

    my $ne = $self->{zapi}->invoke('d-snapshot-create', \%args);

    if (! $ne->results_ok() ) {
        fail_log( "d-snapshot-create()", $ne);
    }
    return $ne->child_get('logical-snap-id')->{content};
}

=head2 d_snapshot_delete

  use DBlade::dSnapshot;
  my $zapi = new DBlade::dSnapshot($FILER, $user, $password);
  my $ret = d_snapshot_delete(%args);

Delete a snapshot with the specified identity and characteristics.

=over

=item 'containing-aggregate-uuid'

(Optional): the identity of the aggregate hosting the volume.

=item 'snap-id'

(Optional): The DSID of the snapshot to delete. Either this or the
snap-name field must be provided.

=item 'snap-name'

(Optional): The snapshot name to delete. If this is provided then
vol-id is required.

=item 'vol-id'

(Required): DSID or UUID of the volume to delete the snapshot in.

=back

=cut

sub d_snapshot_delete ($%)
{
    my ($self, %args) = @_;

    my $ne = $self->{zapi}->invoke('d-snapshot-delete', \%args);

    if (! $ne->results_ok() ) {
        fail_log( "d-snapshot-delete()", $ne);
    }
}

=head2 d_snapshot_list_info

  use DBlade::dSnapshot;
  my $zapi = new DBlade::dSnapshot($FILER, $user, $password);
  my $refinfo = d_snapshot_list_info(%args);
  my %snapinfo = %$refinfo;

Return attributes for exactly one snapshot.

=over

=item 'containing-aggregate-uuid'

(Optional): the identity of the aggregate hosting the volume of the
snapshot.

=item 'desired-attrs'

(Optional): Comma-separated list of desired attributes.
Defaults to all non-expensive-to-compute values.

=item 'snap-id'

(Optional): The DSID of the snapshot to list. Either this or the
snap-name field must be provided.

=item 'snap-name'

(Optional): The snapshot name to list. If this is provided then
vol-id is required.

=item 'vol-id'

(Optional): DSID or UUID of the volume to delete the snapshot in.
Required if snapshot-name is used instead of snapshot-id.

=back

=cut

sub d_snapshot_list_info ($%)
{
    my ($self, %args) = @_;

    my $ne = $self->{zapi}->invoke('d-snapshot-list-info', \%args);

    if (! $ne->results_ok() ) {
        fail_log( "d-snapshot-list-info()", $ne);
    }

    my ($elem, $vattrs);
    $elem = $ne->child_get("snapshot-attrs");
    if ($elem) {
        $vattrs = _parse_snapshot_info($elem->child_get("d-snapshot-info"));
        if (defined($vattrs) and ref($vattrs)) {
            return $vattrs;
        }
    return undef;
    }
}

=head2 d_snapshot_list_info_iter

  use DBlade::dSnapshot;
  my $zapi = new DBlade::dSnapshot($FILER, $user, $password);
  my @snapinfo = d_snapshot_list_info_iter(%args);

Return attributes for many snapshots.

=over

=item 'containing-aggregate-uuid'

(Optional): the identity of the aggregate hosting the volume of the
snapshot.

=item 'desired-attrs'

(Optional): Comma-separated list of desired attributes.
Defaults to all non-expensive-to-compute values.

=item 'maximum-records'

(Optional): The maximum number of records to retrieve per iteration.
Defaults to 5.

=item 'vol-id'

(Optional): DSID or UUID of the volume to delete the snapshot in.
Required if snapshot-name is used instead of snapshot-id.

=back

=cut

sub d_snapshot_list_info_iter ($%)
{
    my ($self, %args) = @_;

    $args{"desired-attrs"} = "all" unless defined $args{"desired-attrs"};

    $args{"maximum-records"} = 50 unless defined $args{"maximum-records"};
    my $max_recs = $args{"maximum-records"};
    my $ne = $self->{zapi}->invoke("d-snapshot-list-info-iter-start", \%args);

    if (! $ne->results_ok() ) {
        fail_log( "d-snapshot-list-info-iter-start()", $ne);
    }

    my (@ret, $elem, $chld, $vattrs, $fail);
    my ($tag, $recs, $islast, $child, $num0);

    while (! $islast) {
        $tag = $ne->child_get("next-tag")->{content};
        $recs = $ne->child_get("num-records-retrieved")->{content};
        $islast = $ne->child_get("is-end-of-list")->{content};
        $chld = $ne->child_get("snapshot-attrs");

        eval { $num0 = $chld->child_max_number(); } or return undef;

        for (my $i=0; $i < $num0; $i++) {
            $elem = $chld->child_get_nth($i);
            $vattrs = _parse_snapshot_info($elem);
            logDbg("d-snap...-next: num0=$num0, i=$i, elem=".Dumper(\$vattrs));
            if (defined($vattrs) and ref($vattrs)) {
                push(@ret, $vattrs);
            }
        }
        last if $islast;

        logDbg("d-snap...-next: max_recs=$max_recs, tag='$tag'");
        $ne = $self->{zapi}->invoke("d-snapshot-list-info-iter-next",
                            {'maximum-records' => $max_recs,
                             'tag' => $tag});

#        logDbg("d-snap...-next returns: ".Dumper($ne));
        if (! $ne->results_ok() ) {
            $fail = $ne;
            last;
        }
    }
    fail_log("d-snapshot-list-info-iter-next()", $fail) if $fail;

    if (defined(@ret)) {
        return @ret;
    }
    return undef;
}

# internal support function for the 2 d-snapshot-list-info* functions

sub _parse_snapshot_info ($)
{
    my $elem = shift;

    my @simple = qw( access-time busy cp-count cumulative-total
                     cumulative-percentage-of-total-blocks
                     cumulative-percentage-of-used-blocks dependency
                     fs-version index logical-snap-id name
                     percentage-of-total-blocks
                     percentage-of-used-blocks snap-id snap-msid total
                     uniq-volume-uuid volume-dsid );
    my (%ret, $e, $f, $c);
    foreach my $k (@simple) {
        $f = $elem->child_get($k);
        $ret{$k} = $f->{content} if $f;
    }
    $e = $elem->child_get("owners");
    if ($e) {
        $f = $e->child_get("d-snapshot-owner");
        my @v = ();
        my $n = $e->child_max_number();
        for (my $i=0; $i < $n; $i++) {
            $c = $f->child_get_nth($i);
            push(@v, $c->{content}) if $c;
        }
        $ret{"owners"} = \@v if @v;
    }

    return \%ret;
}

=head2 d_snapshot_set_info

  use DBlade::dSnapshot;
  my $zapi = new DBlade::dSnapshot($FILER, $user, $password);
  my %unset = d_snapshot_set_info(%args);

Set info for a snapshot specified by snap-id, or snap-name and vol-id

=over

=item 'containing-aggregate-uuid'

(Optional): the identity of the aggregate hosting the volume.

=item 'vol-id'

(Optional): DSID or UUID of the volume to set snapshot info in

=item 'snap-name'

(Optional): Name of the desired snapshot

=item 'snap-id'

(Optional): DSID of the specified snapshot.

=item 'desired-attrs'

(Required): list of attributes to set. Setable attributes are:
logical-snap-id,name,snap-id,snap-msid.

=item 'snapshot-attrs'

(Required): list of attribute values to set.

=back

=cut

sub d_snapshot_set_info ($%)
{
    my ($self, %args)  = @_;

    my $val = $args{'snapshot-attrs'};
    $args{'snapshot-attrs'} = {'d-snapshot-info' => $val};
    my $ne = $self->{zapi}->invoke("d-snapshot-set-info", \%args);

    if (! $ne->results_ok() ) {
        fail_log( "d-snapshot-set-info", $ne);
    }
    my %ret;
    foreach my $k (qw( first-unset-attr first-unset-attr-reason )) {
        my $e = $ne->child_get($k);
        $ret{$k} = $e->{content} if $e;
    }
    return %ret;
}

=head2 read_snapshot_trans_table

  use DBlade::dSnapshot;
  my $zapi = new DBlade::dSnapshot($FILER, $user, $password);
  my %snapinfo = read_snapshot_trans_table(%args);

Return logical and physical snap-IDs for many snapshots on one volume.

=over

=item 'containing-aggregate-uuid'

(Optional): the identity of the aggregate hosting the volume.

=item 'vol-id'

(Required): DSID or UUID of the volume to read the snapshot translation
table from.

=back

=cut

sub read_snapshot_trans_table ($%)
{
    my ($self, %args) = @_;

    $args{'desired-attrs'} = "name,index,logical-snap-id,snap-id";
    $args{'maximum-records'} = 256;
    my @arr = $self->d_snapshot_list_info_iter(%args);

    # mapping: index is physical, value is logical
    my %mapping;
    foreach my $aref (@arr) {
        my %ahash = %{$aref};
        $mapping{$ahash{"index"}} = $ahash{"logical-snap-id"};
    }
    return %mapping
}

=head2 d_snapshot_delta

  use DBlade::dSnapshot;
  my $zapi = new DBlade::dSnapshot($FILER, $user, $password);
  my %info = $zapi->d_snapshot_delta(%args);

=over

=item 'containing-aggregate-uuid'

(Optional): the UUID of the aggregate containing the volume.

=item 'snap-id-1'

(Optional): The DSID of the first snapshot to compare.
Either this or snap-name-1 and vol-id are required.

=item 'snap-id-2'

(Optional): The DSID of the first snapshot to compare.
Either this or snap-name-2 and vol-id are required.

=item 'snap-name-1'

(Optional): The name of the snapshot for comparison.
If this is supplied then vol-id is required.

=item 'snap-name-2'

(Optional): The name of the snapshot for comparison.
If this is supplied then vol-id is required.

=item 'vol-id'

(Optional): The DSID or the UUID of the volume containing the snapshots to compare.

=back

=cut

sub d_snapshot_delta($@)
{
    my ($self, %args) = @_;

    my $ne = $self->{zapi}->invoke('d-snapshot-delta', \%args);

    if (! $ne->results_ok() ) {
        fail_log( "d-snapshot-delta()", $ne);
    }

    my %ret;
    my $elem = $ne->child_get("delta");
    if ($elem) {
        my $e = $elem->child_get("d-snapshot-delta-info");
        if ($e) {
            for my $f (qw( consumed-size elapsed-time )) {
                $ret{$f} = $e->child_get($f)->{content};
            }
        }
        return %ret;
    }
    return undef;
}

=head2 d_snapshot_reclaimable

  use DBlade::dSnapshot;
  my $zapi = new DBlade::dSnapshot($FILER, $user, $password);
  my %info = $zapi->d_snapshot_reclaimable(%args);

=over

=item 'containing-aggregate-uuid'

(Optional): the UUID of the aggregate containing the volume.

=item 'snap-ids'

(Optional): Array of DSID of the snapshots to consider.
Either this or snap-names and vol-id are required.

=item 'snap-names'

(Optional): Array of names of the snapshots to consider.
If this is supplied then vol-id is required.

=item 'vol-id'

(Optional): The DSID or the UUID of the volume containing the snapshots to compare.

=back

=cut

sub d_snapshot_reclaimable($@)
{
    my ($self, %args) = @_;

    if ($args{'snap-ids'}) {
        my @snap_ids = @{$args{'snap-ids'}};
        logDbg("d_snapshot_reclaimable: snap_ids = @snap_ids");
        my $str = '[d-snapshot-id='.join(' d-snapshot-id=', @snap_ids).']';
        logDbg("d_snapshot_reclaimable: str=$str");
        $args{'snap-ids'} = $str;
    }
    my $ne = $self->{zapi}->invoke('d-snapshot-reclaimable', \%args);

    if (! $ne->results_ok() ) {
        fail_log( "d-snapshot-reclaimable()", $ne);
    }

    return $ne->child_get('reclaimable-size')->{content};

}

1;
