#
#  $Id: $
#
#  Copyright (c) 2008 Network Appliance, Inc.
#  All rights reserved.
#

# -----------------------------------------------------------
##
## @summary Access functions for the d-volume-* DBlade ZAPIs
##
## @description
##	This file contains routines that let you exercise the
##	internal API of the filer using ssh access to the FreeBSD
##	shell command zsmcli.
##
##	Contains interface functions to d-blade Zapis that take MSIDs
##	or DSIDs as arguments or return values.
##
## @author shenry@netapp.com

=head Summary

Support routines for d-blade Zapi d-volume tests.

=head Description

This file contains routines that let you exercise the internal
(d-blade) d-volume API of the filer using ssh access to the FreeBSD
shell command zsmcli.

=head2 Author

shenry@netapp.com, dl-waflqa-art@netapp.com

=cut

package DBlade::dVolume;

use strict;
use Tharn;
use Socket;
use LWP::UserAgent;
use XML::Parser;
use NaElement;
use SPARSE::Util::Logger qw( logWarn logDbg logTrc );
use Data::Dumper;
use DBlade::dZapi qw( fail_log );

=head2 new

  use DBlade::dVolume;
  my $zapi = new DBlade::dVolume($FILER, $user, $password);

Create a new instance of the dblade volume zapi interface.
All access is through member functions.

=over

=item $FILER

(Required): Hostspec or name of the converged filer.

=item $user

(Optional): Username to login to the filer's BSD shell with.
Defaults to "root".

=item $password

(Optional): password for the login account. Takes the password from
the nate_host file if not specified.

=back

=cut

#' emacs-fodder

sub new {
    my ($class, $server, $user, $password) = @_;

    my $zapi = new DBlade::dZapi($server, $user, $password);

    my $self = {zapi => $zapi,};
    bless $self, $class;
    return $self;
}

=head2 d_volume_list_info_iter

  use DBlade::dZapi;
  my $dzapi = new DBlade::dZapi($FILER, $user, $passwd);
  my @vol_info_list = $dzapi->d_volume_list_info_iter($attrs);

Returns an array of hashref of all volume info on this d-blade.

=over

=item $attrs

(Optional): A comma separated list of desired attributes to be
returned. Defaults to all attributes.

=back

Dies with $@ if the routine will not complete for some reason.

=cut

sub d_volume_list_info_iter ($@) {
    my $self  = shift;
    my $attrs = shift;
    $attrs = "all" unless $attrs;

    my $max_recs = 5;    # maximum records to retrieve per call
    my $ne =
      $self->{zapi}
      ->invoke("d-volume-list-info-iter-start", {'desired-attrs' => $attrs});

    if (!$ne->results_ok()) {
        fail_log("d-volume-list-info-iter-start()", $ne);
    }

    my $tag = $ne->child_get("next-tag")->{content};

    my (@ret, $elem, $chld, $vattrs, $fail);

    my $recs = $max_recs;
    while ($recs >= $max_recs) {
        logDbg("d-vol...-next: max_recs=$max_recs, tag='$tag'");
        $ne = $self->{zapi}->invoke("d-volume-list-info-iter-next",
                                    {'maximum-records' => $max_recs,
                                     'tag'             => $tag});

        #        logDbg("d-vol...-next returns: ".Dumper($ne));
        if (!$ne->results_ok()) {
            $fail = $ne;
            last;
        }

        $tag  = $ne->child_get("next-tag")->{content};
        $recs = $ne->child_get("num-records-retrieved")->{content};
        $chld = $ne->child_get("volume-attrs");
        my $num0;

        # logDbg("d_volume_list_info_iter(): max_recs=$max_recs, recs=$recs");
        eval {
            $num0 = $chld->child_max_number();
        }; if($@) {
            if($recs == 0){
                last;
            } else {
                return undef;
            }
        }

        #        logDbg("d_volume_list_info_iter() returns $num0 volume records");
        for (my $i = 0 ; $i < $num0 ; $i++) {
            $elem = $chld->child_get_nth($i);
            logDbg("d-vol...-next: num0=$num0, i=$i, elem=$elem");
            $vattrs = _parse_volume_info($elem);
            if (defined($vattrs) and ref($vattrs)) {
                push(@ret, $vattrs);
            }
        }
    } ## end while ($recs >= $max_recs)

    $ne = $self->{zapi}->invoke("d-volume-list-info-iter-end", {'tag' => $tag});

    fail_log("d-volume-list-info-iter-next()", $fail) if $fail;

    if (!$ne->results_ok()) {
        fail_log("d-volume-list-info-iter-end()", $ne);
    }

    return @ret;

} ## end sub d_volume_list_info_iter ($@)

=head2 d_volume_list_info

  use DBlade::dZapi;
  my $dzapi = new DBlade::dZapi($FILER, $user, $passwd);
  my %vol_info_list = d_volume_list_info_iter(%args);

Returns the volume information specified in $attrs for the volume
described by $id in containing aggregate $aggr.

=over

=item id

(Required): The hex-encoded DSID or the formatted UUID of the volume
whose information is requested.

=item desired-attrs

(Optional): A comma separated list of desired attributes to be
returned or the string "all". Defaults to all attributes.

=item containing-aggr-uuid

(Optional): The UUID of the aggregate in which the volume is supposed
to reside.

=back

Returns a hash of the attributes of the specified volume.
Dies with $@ if the routine will not complete for some reason.

=cut

sub d_volume_list_info ($%) {
    my ($self, %args) = @_;

    my $ne = $self->{zapi}->invoke("d-volume-list-info", \%args);

    if (!$ne->results_ok()) {
        fail_log("d-volume-list-info()", $ne);
    }

    my ($elem, $vattrs);
    $elem = $ne->child_get("volume-attrs");
    if ($elem) {
        $vattrs = _parse_volume_info($elem->child_get("d-volume-info"));
        if (defined($vattrs) and ref($vattrs)) {
            return $vattrs;
        }
    }
    return undef;
} ## end sub d_volume_list_info ($%)

# internal support function for the 2 d-volume-list-info* functions

sub _parse_volume_info ($) {
    my $elem = shift;

    my @simple = qw( clone-child-count containing-aggregate
      containing-aggregate-uuid creation-time data-version
      dsid extent-enabled files-total files-used
      fractional-reserve fs-size-fixed i2p-enabled
      ignore-inconsistent index-dir-enabled is-home
      is-inconsistent is-invalid is-mroot is-snaplock
      is-unrecoverable language-code max-dir-size min-ra
      mirror-type msid name no-atime-update no-snap
      no-snap-dir nvfail ondisk-quiesced owning-vserver-uuid
      percentage-used quiesced quota-init size
      size-available size-total size-used snapmirrored
      space-guarantee space-guarantee-enabled state uuid );
    my (%ret, $e, $f, $g, $c, $d, $n);

    foreach my $k (@simple) {
        $f = $elem->child_get($k);
        $ret{$k} = $f->{content} if $f;
    }
    $e = $elem->child_get("autosize");
    if ($e) {
        $f = $e->child_get("d-autosize-info");
        $n = $e->child_max_number();
        my @a = ();
        for (my $i = 0 ; $i < $n ; $i++) {
            $g = $f->child_get_nth($i);
            my %v = ();
            foreach my $i (qw( increment-size is-enabled maximum-size )) {
                $c = $f->child_get($i);
                $v{$i} = $c->{content} if $c;
            }
            push(@a, \%v) if %v;
        }
        $ret{"autosize"} = \@a if @a;
    } ## end if ($e)
    $e = $elem->child_get("clone-child");
    if ($e) {
        $f = $e->child_get("d-clone-child-info");
        my %v = ();
        foreach my $i (qw( dsid msid name owning-vserver-uuid uuid )) {
            $c = $f->child_get($i);
            $v{$i} = $c->{content} if $c;
        }
        $ret{"clone-child"} = \%v if %v;
    }
    $e = $elem->child_get("clone-parent");
    if ($e) {
        $f = $e->child_get("d-clone-parent-info");
        my %v = ();
        foreach
          my $i (qw( dsid msid name owning-vserver-uuid snapshot-id uuid )) {
            $c = $f->child_get($i);
            $v{$i} = $c->{content} if $c;
        }
        $ret{"clone-parent"} = \%v if %v;
    } ## end if ($e)
    $g = $elem->child_get("security");
    if ($g) {
        $f = $g->child_get("d-security-info");
        if ($f) {
            my %v = ();
            $c = $f->child_get("style");
            $v{style} = $c->{content} if $c;
            $c = $f->child_get("unix")->child_get("d-security-unix-info");
            if ($c) {
                my %w = ();
                foreach my $i (qw( groupid permissions userid )) {
                    $d = $c->child_get($i);
                    $w{$i} = $d->{content} if $d;
                }
                $v{"unix"} = \%w if %w;
            }
            $ret{"security"} = \%v if %v;
        } ## end if ($f)
    } ## end if ($g)
    $e = $elem->child_get("snapshot-autodelete");
    if ($e) {
        $f = $e->child_get("d-snapshot-autodelete-info");
        my %v = ();
        foreach my $i (qw( commitment defer-delete delete-order enabled
                       prefix target-free-space-percent trigger )
          ) {
            $c = $f->child_get($i);
            $v{$i} = $c->{content} if $c;
        }
        $ret{"snapshot-autodelete"} = \%v if %v;
    } ## end if ($e)
    $e = $elem->child_get("snapshot-reserve");
    if ($e) {
        $f = $e->child_get("d-snapshot-reserve-info");
        my %v = ();
        foreach my $i (qw( percentage size )) {
            $c = $f->child_get($i);
            $v{$i} = $c->{content} if $c;
        }
        $ret{"snapshot-reserve"} = \%v if %v;
    }
    $e = $elem->child_get("snapshot-schedule");
    if ($e) {
        $f = $e->child_get("d-snapshot-schedule-info");
        my %v = ();
        foreach my $i (qw( days hours minutes weeks which-hours which-minutes ))
        {
            $c = $f->child_get($i);
            $v{$i} = $c->{content} if $c;
        }
        $ret{"snapshot-schedule"} = \%v if %v;
    } ## end if ($e)
    $e = $elem->child_get("striping");
    if ($e) {
        $f = $e->child_get("d-striping-info");
        my %v = ();
        foreach
          my $i (qw( collective-state is-striped-representative stripe-type )) {
            $c = $f->child_get($i);
            $v{$i} = $c->{content} if $c;
        }
        $ret{"striping"} = \%v if %v;
    } ## end if ($e)
    $e = $elem->child_get("wafliron");
    if ($e) {
        $f = $e->child_get("d-wafliron-info");
        my %v = ();
        foreach my $i (qw( percent-inodes-scanned state )) {
            $c = $f->child_get($i);
            $v{$i} = $c->{content} if $c;
        }
        $ret{"wafliron"} = \%v if %v;
    }

    return \%ret;
} ## end sub _parse_volume_info ($)

=head2 d_volume_set_info

  use DBlade::dZapi;
  my $dzapi = new DBlade::dZapi($FILER, $user, $passwd);
  my %ret = $zapi->d_volume_set_info(%args);

Returns the error when setting volume information specified in $attrs
for the volume described by $id in containing aggregate $aggr.

=over

=item 'id'

(Required): The hex-encoded DSID or the formatted UUID of the volume
whose information is requested to be set.

=item 'desired-attrs'

(Required): A comma separated list of desired attributes to be
set.

=item 'containing-aggr-uuid'

(Optional): The UUID of the aggregate in which the volume is supposed
to reside.

=item 'reason'

(Optional): log why the change is being made. Appears to be ignored.

=item 'target-set'

(Optional): boolean -- if the target is the striped volume as a whole,
or just a specific volume.

=item 'volume-attrs'

(Required): a d-volume-info structure in which the fields identified in
'desired-attrs' have been set with the new values.

Probably only really works with the simple settings.

=back

Returns a hash of the first attribute not set (in the order supplied in
'desired-attrs') in the specified volume.
Dies with $@ if the routine will not complete for some reason.

=cut

sub d_volume_set_info ($%) {
    my ($self, %args) = @_;

    my $val = $args{'volume-attrs'};
    $args{'volume-attrs'} = {'d-volume-info' => $val};
    my $ne = $self->{zapi}->invoke("d-volume-set-info", \%args);

    if (!$ne->results_ok()) {
        fail_log("d-volume-set-info", $ne);
    }
    my %ret;
    foreach my $k (qw( first-unset-attr first-unset-attr-reason )) {
        my $p = $ne->child_get($k);
        $ret{$k} = $p->{content} if $p;
    }
    return %ret;
} ## end sub d_volume_set_info ($%)

=head2 d_volume_create

  use DBlade::dZapi;
  my $dzapi = new DBlade::dZapi($FILER, $user, $passwd);
  $bool = $zapi->d_volume_create(%args);

Creates a volume on a specified vserver. Dies on failure.

Full ZAPI docs at:
http://web.netapp.com/engineering/design-depot/appliance-mgmt/zephyr/zapidoc.cache/ontapng/mainngN/ontap/web_ontap_d/d-volume/index.html#d-volume-create

=over

=item 'containing-aggr-name'

(Optional): Name of aggregate in which to create the volume, if not
a clone.

=item 'containing-aggr-uuid'

(Optional): UUID of the aggregate in which to create the volume, if not
a clone.

=item 'dsid'

(Required): new DSID for the volume or clone.

=item 'language-code'

(Optional): Specifies the language to use for the new volume/clone.
Defaults to "C" (Posix).

=item 'max-dir-size'

(Optional): Maximum size in kbytes to which any directory in this volume
can grow. If this is a clone, the value MUST match the parent's value.
Defaults to '10240'.


=item 'msid'

(Required): The MSID value to assign to this volume/clone.

=item 'name'

(Required): The name of this volume.

=item 'owning-vserver-uuid'

(Required): the formatted UUID of the vserver which will own this volume.

=item 'parent-snapshot-id'

(Optional): DSID of snapshot that will be the base of the clone.
'parent-volume-id' must also appear.

=item 'parent-snapshot-name'

(Optional): snapshot name in the parent vol that will be the base of
the clone. 'parent-volume-id' must also appear.

=item 'parent-volume-id'

(Optional): DSID or UUID of the parent volume if this is a clone.

=item 'size'

(Optional): Size of the volume to be created. Defaults to '30m'.

=item 'uuid'

(Required): the UUID to be assigned to this volume/clone.

=back

As we are bypassing the m-host when creating this volume, the system
may become unstable. It is recommended to minimize use of ngsh commands
until the volume is destroyed using 'd_volume_destroy()'.

=cut

sub d_volume_create ($@) {
    my ($self, %args) = @_;

    my $ne = $self->{zapi}->invoke("d-volume-create", \%args);

    if (!$ne->results_ok()) {
        fail_log("d-volume-create", $ne);
    }
    return 0;
} ## end sub d_volume_create ($@)

=head2 d_volume_destroy

  use DBlade::dZapi;
  my $dzapi = new DBlade::dZapi($FILER, $user, $passwd);
  $bool = $zapi->d_volume_destroy(%args);

Destroys a volume on a specified vserver. Dies on failure.

Full ZAPI docs at:
http://web.netapp.com/engineering/design-depot/appliance-mgmt/zephyr/zapidoc.cache/ontapng/mainngN/ontap/web_ontap_d/d-volume/index.html#d-volume-destroy

=over

=item 'containing-aggr-uuid'

(Optional): UUID of the aggregate in which to find the volume.

=item 'id'

(Required): The identity of the volume to be destroyed. The only legal
values are the DSID and the UUID.

=item 'reason'

(Optional): The reason for the offline. Ignored for non-striped volumes.

=back

As we are bypassing the m-host when destroying this volume, this
function should only be called on volumes created directly with
'd_volume_create()'.

=cut

sub d_volume_destroy ($%) {
    my ($self, %args) = @_;

    my $ne = $self->{zapi}->invoke("d-volume-destroy", \%args);

    if (!$ne->results_ok()) {
        fail_log("d-volume-destroy", $ne);
    }
    return 0;
} ## end sub d_volume_destroy ($%)

=head2 d_volume_offline

  use DBlade::dZapi;
  my $dzapi = new DBlade::dZapi($FILER, $user, $passwd);
  $bool = $zapi->d_volume_offline(%args);

Offlines a volume on a specified vserver. Dies on failure.

Full ZAPI docs at:
http://web.netapp.com/engineering/design-depot/appliance-mgmt/zephyr/zapidoc.cache/ontapng/mainngN/ontap/web_ontap_d/d-volume/index.html#d-volume-offline

=over

=item 'containing-aggr-uuid'

(Optional): UUID of the aggregate in which to find the volume.

=item 'id'

(Required): The identity of the volume to be offlined. The only legal
values are the DSID and the UUID.

=back

As we are bypassing the m-host when manipulating this volume, this
function should only be called on volumes created directly with
'd_volume_create()'.

=cut

sub d_volume_offline ($%) {
    my ($self, %args) = @_;

    my $ne = $self->{zapi}->invoke("d-volume-offline", \%args);

    if (!$ne->results_ok()) {
        fail_log("d-volume-offline", $ne);
    }
    return 0;
} ## end sub d_volume_offline ($%)

=head2 d_volume_online

  use DBlade::dZapi;
  my $dzapi = new DBlade::dZapi($FILER, $user, $passwd);
  $bool = $zapi->d_volume_online(%args);

Onlines a volume on a specified vserver. Dies on failure.

Full ZAPI docs at:
http://web.netapp.com/engineering/design-depot/appliance-mgmt/zephyr/zapidoc.cache/ontapng/mainngN/ontap/web_ontap_d/d-volume/index.html#d-volume-online

=over

=item 'containing-aggr-uuid'

(Optional): UUID of the aggregate in which to find the volume.

=item 'id'

(Required): The identity of the volume to be onlined. The only legal
values are the DSID and the UUID.

=back

As we are bypassing the m-host when manipulating this volume, this
function should only be called on volumes created directly with
'd_volume_create()'.

=cut

sub d_volume_online ($%) {
    my ($self, %args) = @_;

    my $ne = $self->{zapi}->invoke("d-volume-online", \%args);

    if (!$ne->results_ok()) {
        fail_log("d-volume-online", $ne);
    }
    return 0;
} ## end sub d_volume_online ($%)

=head2 d_volume_restrict

  use DBlade::dZapi;
  my $dzapi = new DBlade::dZapi($FILER, $user, $passwd);
  $bool = $zapi->d_volume_restrict(%args);

Restricts a volume on a specified vserver. Dies on failure.

Full ZAPI docs at:
http://web.netapp.com/engineering/design-depot/appliance-mgmt/zephyr/zapidoc.cache/ontapng/mainngN/ontap/web_ontap_d/d-volume/index.html#d-volume-restrict

=over

=item 'containing-aggr-uuid'

(Optional): UUID of the aggregate in which to find the volume.

=item 'id'

(Required): The identity of the volume to be restricted. The only legal
values are the DSID and the UUID.

=back

As we are bypassing the m-host when manipulating this volume, this
function should only be called on volumes created directly with
'd_volume_create()'.

=cut

sub d_volume_restrict ($%) {
    my ($self, %args) = @_;

    my $ne = $self->{zapi}->invoke("d-volume-restrict", \%args);

    if (!$ne->results_ok()) {
        fail_log("d-volume-restrict", $ne);
    }
    return 0;
} ## end sub d_volume_restrict ($%)

=head2 volume_name_to_dsid

  use DBlade::dZapi;
  my $dzapi = new DBlade::dZapi($FILER, $user, $passwd);
  $vol_dsid = $zapi->volume_name_to_dsid(%args);

Looks through the volumes on the specified vserver by name and returns
the DSID of the volume.

=over

=item 'vserver'

(Required): The UUID of the vserver to search for the volume by name.

=item 'name'

(Required): The name of the volume to be located.

=back

=cut

sub volume_name_to_dsid($%) {
    my ($self, %args) = @_;

    my @arr = $self->d_volume_list_info_iter("name,dsid,owning-vserver-uuid");
    if(undef(@arr)) { return undef; }
    foreach my $aref (@arr) {
	 my %retl = %{$aref};

	logDbg(
            "volume_name_to_dsid: name=$retl{name}, vserver=$retl{'owning-vserver-uuid'}"
            );
        if (    $retl{"name"} eq $args{"name"}
            and $retl{"owning-vserver-uuid"} =~ /$args{"vserver"}/i) {
            return $retl{"dsid"};
        }
    } ## end foreach my $aref (@arr)
    return undef;
} ## end sub volume_name_to_dsid($%)

1;
