#
#  $Id: $
#
#  Copyright (c) 2008 Network Appliance, Inc.
#  All rights reserved.
#

# -----------------------------------------------------------
##
## @summary Support routines for DBlade ZAPI access functions
##
## @description
##	This file contains routines that let you exercise the
##	internal API of the filer using ssh access to the FreeBSD
##	shell command zsmcli.
##
##	Contains interface functions to d-blade Zapis that take MSIDs
##	or DSIDs as arguments or return values.
##
## @author shenry@netapp.com

=head Summary

Support routines for d-blade Zapi tests.

=head Description

This file contains routines that let you exercise the internal
(d-blade) API of the filer using ssh access to the FreeBSD shell
command zsmcli.

=head2 Author

shenry@netapp.com

=cut

package DBlade::dZapi;

use	strict;
use	Tharn;
use	Socket;
use	LWP::UserAgent;
use	XML::Parser;
use	NaElement;
use	SPARSE::Util::Logger qw( logWarn logDbg logTrc );
use	Data::Dumper;
use	Utils qw( make_conn );
use	Server;

use	vars ('@ISA', '@EXPORT');
require Exporter;
@ISA    = qw(Exporter);
@EXPORT = qw(invoke fail_log);


=head2 new

  use DBlade::dZapi;
  my $zapi = new DBlade::dZapi($FILER, $user, $password);

Create a new instance of the dblade zapi interface.
All access is through member functions.

=over

=item $FILER

(Required): Hostspec or name of the converged filer.

=item $user

(Optional): Username to login to the filer's BSD shell with.
Defaults to "root".

=item $password

(Optional): password for the login account. Takes the password from
the nate_host file if not specified.

=back

=cut

#' emacs-fodder


sub new {
    my ($class, $node, $user, $password) = @_;

    my $host = host($node);

    $Log->step("## new dZapi($class, $node, $user, $password");
    # If admin username is not provided, we assume we're using
    # the default BSD diag account.

    # force total override...
    $user = "diag";

    # If I can figure out how to tell a login failure, I can try
    # the new system default password, too.

    # If the password is not provided, and the node name
    # is not an IP address, then we can look up the default password for it

    if ( !defined($password)
         && ! ( "$node" =~ /[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+/ ) ) {

        $password = $host->{default_password};
    }

    my $server = new Server();

    my $self = {
                server => $server,
		node => $node,
		user   => $user,
		password => $password,
		host => $host,
		transport_type => "SSH",
               };

    bless $self, $class;
    return $self;
}

=head2 node

  $my_hostspec = $zapi->node();

Convenience function. Can only be set on "new";

=cut

sub node {
    my $self = shift;
    return ( $self->{node} );
}

=head2 user

  $my_user = $zapi->user();

Convenience function. Can only be set on "new";

=cut

sub user {
    my $self = shift;
    return ( $self->{user} );
}

=head2 password

  $my_password = $zapi->password();

Convenience function. Can only be set on "new";

=cut

sub password {
    my $self = shift;
    return ( $self->{password} );
}


## @name $zapi->fail_response($errno, $reason);
## @description
##	Member function returning an NaElement which is a <results>
##	containing a failure.  This is used when the transmission path
##	fails, and we don't actually get back any XML from the node.
## @arg $errno
##	(Required): The errno to be returned for this response.
## @arg $reason
##	(Required): A text version describing the error in more detail.
##
## Normally only used by internal functions.

sub fail_response {
    my $self	= shift;
    my $errno	= shift;
    my $reason	= shift;

    my $n = new NaElement("results");
    $n->attr_set("status","failed");
    $n->attr_set("reason","$reason");
    $n->attr_set("errno","$errno");
    return $n;
}

## @name fail_log($name, $elem);
## @summary log the failure
## @arg $name
##	name of the function to be in log
## @arg $elem
##	The NaElement of the result with the failure info

sub fail_log($$@)
{
    my $name = shift;
    my $ne   = shift;

    if ($ne->name() eq "results") {
        my $s = $ne->attr_get('status');
        my $e = $ne->attr_get('errno');
        my $r = $ne->attr_get('reason');
        die "$name FAILed: status=>$s, errno=>$e, reason=>$r";
    } else {
        die "$name FAILed for unknown reason";
    }

}

=head2 invoke

  my $response = $zapi->invoke($command, $options);

Member function that submits a d-blade ZAPI request using zsmcli.
Prepares an API request, submits it with ssh to zsmcli,
and returns an NaElement object which contains all the results.

=over

=item $command

(Required): The zapi command string (eg: "d-volume-list-info")

=item $options

A reference to a hash containg arguments to the $command.
May be optional or required depending on the command.

=back

=cut

sub invoke ( $$@ )
{
    my $self	= shift;
    my $cmd	= shift;
    my $req	= shift;

    my $server = $self->{server};

    $Log->debug("send to _xform_args: ".Dumper(\$req));
    my $cmdstring = "sudo ".$self->_xform_args($cmd, $req);
    $Log->debug("invoke() server=$server, sends='$cmdstring'");
    my $ref = $server->shell(node => $self->{host}->hostname(),
                             cmd => $cmdstring);
    logDbg("invoke::shell() returns ref=".Dumper(\$ref));
    $ref =~ s/^ *(sudo )?zsmcli.*\r\r\n//m;

    my $ret = $self->parse_xml($ref,undef,undef);
    #logDbg("invoke() returns: ".$ret->toString());
    return $ret;
}

## @name	::_xform_args
## @synopsis	internal function to convert perl arg format to zsmcli
##		command string.
## @description	Takes a command name and a ref to a hash of args to that
##		d-blade Zapi command and converts it to a string of the
##		command to be executed.
## @param $cmd	commmand name to execute
## @param $req	ref to hash of args to $cmd

sub _xform_args ( $$$ )
{
    my $self	= shift;
    my $cmd	= shift;
    my $req	= shift;

    my $str = "zsmcli $cmd";
    if (ref($req) eq "HASH") {
        for my $key (keys(%$req)) {
            my $val = ${$req}{$key};
            if ($key eq "tag") {
                #$val =~ s/ /\\ /g;
                #$val =~ s/\"/\\\"/g;
                $str .= " ".$key."='".$val."'";
            } elsif (ref($val)) {
                $str .= " ".$key."='["._xform_helper($val)."]'";
            } else {
                $str .= " ".$key."='".$val."'";
            }
        }
        $str =~ s/ \\cH//g;
    }

    logDbg("_xform_args() returns: $str");
    return $str;
}

sub _xform_helper
{
    my ($val) = @_;

    my $str;
    logDbg("type of val is ".ref($val)) if ref($val);
    if (ref($val) eq 'HASH') {
        my %sub = %$val;
        foreach my $k2 (keys %sub) {
            if (ref($sub{$k2})) {
                $str .= $k2."=["._xform_helper($sub{$k2})."] ";
            } else {
                $str .= $k2."=".$sub{$k2}." ";
            }
        }
    } elsif (ref($val) eq "ARRAY") {
        die "DBlade::dZapi::_xform_args(): ARRAY type unimplemented";
    } elsif (ref($val)) {
        die "DBLADE::dZapi::_xform_args(): unknown reference type";
    } else {
        $str .= $val;
    }
    return $str;
}

=head2 netohash

  my %hash = $zapi->netohash($invoke_output);

Designed to take the output of invoke() and turn it into a herarchy
of perl hash refs and array refs.

=over

=item invoke_output

(Required): the output of the invoke() member function, or other
NaElement structure.

=back

Unimplelented.

=cut

sub netohash ( $$@ )
{
    my $self = shift;
    my $ne  = shift;

    if (! $ne->results_ok() ) {
        fail_log("internal: netohash", $ne);
    }

    my %ret;
    my $ar0 = $ne->child_get("results")->{children};
    for (my $i1=0; $i1 < $#$ar0; $i1++) {
        my @arr1;
        my $ar1 = $ar0->[$i1]->{children};
        for (my $i2=0; $i2 <= $#$ar1; $i2++) {
        }
    }

}


## @name ::parse_xml
## @synopsis	Take an XML response and parse it.
## @description Prepares an XML API response, parses it,
##		and returns an NaElement object which
##		contains all the results.
##		Internal function.
## @param $xml		complete XML response

sub parse_xml {

    my $self	= shift;
    my $xml		= shift;

    if ( $::ZAPI_snoop > 0 ) {
        $self->{log}->("Zapi parse_xml xml=".$xml."\n");
    }

    $::ZAPI_stack = [];

    my $p = new XML::Parser(ErrorContext => 2);
    $p->setHandlers(
                    Start => \&server_start_handler,
                    Char => \&server_char_handler,
                    End => \&server_end_handler
                   );
    $p->parse($xml);

    if ( $#$::ZAPI_stack < 0 ) {
        return $self->fail_response(13001,
                                    "Zapi::parse_xml - no elements on stack");
    }
    my $r = pop(@$::ZAPI_stack);

    if ( $::ZAPI_snoop > 0 ) {
        $self->{log}->("Zapi::invoke_reserved_port raw output="
                       .$r->toString()."\n");
    }

    # We return the first element, which
    # should be <results>.  This only works if the original request
    # has only a single request.
    my $c;
    if ( $r->name() ne "results" ) {
        return $self->fail_response(13001,
                                    "Zapi::parse_xml - Unexpected element "
                                    ."in output - expected results, got ".$r->name());
    }

    return $r;
}

sub server_start_handler ($$@) {
	my $xp = shift;
	my $el = shift;

	my $n = new NaElement("$el");
	push(@$::ZAPI_stack,$n);

	my $sz = $#$::ZAPI_stack;

	%::ZAPI_atts = ();
	while ( @_ ) {
		my $att = shift;
		my $val = shift;
		$::ZAPI_atts{$att} = $val;
		$n->attr_set($att,$val);
	}
}

sub server_char_handler {
	my $xp = shift;
	my $data = shift;

	my $i = $#$::ZAPI_stack;
	$::ZAPI_stack->[$i]->content_add($data);
}

sub server_end_handler {
	my $xp = shift;
	my $el = shift;

	# We leave the last element on the stack.
	if ( $#$::ZAPI_stack > 0 ) {
		my $sz = $#$::ZAPI_stack;

		# Pop the element and add it as a child
		# to its parent.
		my $n = pop(@$::ZAPI_stack);
		my $ns = $n->toString();
		my $i = $#$::ZAPI_stack;

		$::ZAPI_stack->[$i]->child_add($n);
	}
}

1;
