# $Id:
#
# Copyright (c) 2010 NetApp.
# All rights reserved.
#
#########################################################
## @summary Library of functions to retreive info from HyperV
##
##
## @author Prasad S (prasads@netapp.com)
##
## @status public
##
## @pod here
#########################################################

# Package Name
##############
package hyperv;

require Exporter;
@ISA    = qw/Exporter/;
@EXPORT = qw/get_virtualmachines get_hostcluster get_hostsystems
    create_virtualmachine delete_virtualmachine start_virtualmachine
    stop_virtualmachine rename_virtualmachine/;

use Tharn;
use strict;
use Error;
use XML::Simple;
use Data::Dumper;
use vars qw(@ISA);
@ISA = qw(Error);

# Declare global variables
use vars qw($class $hostId $hostType $userName $passWord);
use vars qw($hostObj $connObj);

=head NAME

hyperv 
- Functions to access HyperV entities & to perform create/delete VM operations

=head1 SYNOPSIS

use hyperv; 

my $obj = new hyperv( <hostId>, <hostType>, <userName>, <passWord> );


=head1 DESCRIPTION

This package offers the functions to perform operations on HyperV objects using
SCVMM and to get the storage mapping of the VMs using SDCLI

While creating the hyperv I<object>, I<hostType>, I<hostId>, I<userName> and  
I<passWord> of the SCVMM needs to be provided,

my $obj = new hyperv( <hostId>, <hostType>, <userName>, <passWord> );

Example:
my $obj = new hyperv( 10.10.10.12, "windows", "admin", "passWord" );

=head1 FUNCTIONS

All the get functions will return reference to array of reference to hash, 
i.e. each value in array is reference to an hash.
More details are explained in each function as below.

=cut

###############################################################################
# P U B L I C   M E T H O D S
###############################################################################

###############################################################################
# Subroutine Name: new()
# Objective:       Create a new hyperv object
###############################################################################
sub new {
    ( $class, $hostId, $hostType, $userName, $passWord ) = @_;

    my $self = {};
    bless( $self, $class );
    eval {
        $hostObj = host("$hostId;hosttype=$hostType");
        $connObj
            = connect("#rsh-persay;username=$userName;password=$passWord");
    };
    if ($@) {
        $Log->comment($@);
        $self->_clear_err();
        $self->_set_err($@);
    }
    return $self;
} ## end sub new

=head2 B<get_virtualmachines()>

=over 4

=item Inputs

No inputs are needed. It just makes call to SCVMM to fetch the VM list

=item Returns

Return reference to array of references to hashes, 
i.e. each value in array is reference to an hash, 

where each hash is a set of {key:value} pair containing the various properties
of a VM like Name, VHDs etc.

=item Examples

    
    my $obj = new hyperv( $hostId, $hostType, $userName, $passWord );
    
    my $out = $obj->get_virtualmachines();
    
    #print properties(like Name, OS, VHDs) of each VM from the DataStructure
    foreach (@$out) {
        print(
            "Name            :    $_->{'Name'}",
            "OperatingSystem    :    $_->{'OperatingSystem'}",
            "Location        :    $_->{'Location'}",
            "VirtualHardDisks    :    $_->{'VirtualHardDisks'}",
            "HostName        :    $_->{'HostName'}",
        );
    }
    


=cut

###############################################################################
# Subroutine Name: get_virtualmachines()
# Objective:       To get the list of hyper VM(s) & its properties
###############################################################################
sub get_virtualmachines {

    my $self     = shift;
    my $cmd      = "get-vm ";
    my $param    = "-vmmserver $hostId";             #Command to get VMs
    my $isPScmd  = 1;
    my $fileName = "file_get_virtualmachines.txt";

    my $out = $self->_run_functions( $isPScmd, $cmd, $param, $fileName );
    return $out;
} ## end sub get_virtualmachines

=head2 B<get_hostcluster()>

=over 4

=item Inputs

No inputs are needed. 
It just makes call to SCVMM to fetch the cluster list & its properties

=item Returns

Return reference to array of references to hashes, 
i.e. each value in array is reference to an hash, 

where each hash is a set of {key:value} pair containing the various 
properties of a cluster like  Name, ClusterName, Nodes etc.

=item Examples
    
    my $obj = new hyperv( $hostId, $hostType, $userName, $passWord );
    
    my $out = $obj->get_hostcluster();
    
    #print properties(like Name, ClusterName, Nodes) 
	of each cluster from the DataStructure
    foreach (@$out) {
        print(
            "Name            :    $_->{'Name'}",
            "ClusterName        :    $_->{'ClusterName'}",
            "Nodes            :    $_->{'Nodes'}",
        );
    }
    



=cut

###############################################################################
# Subroutine Name: get_hostcluster()
# Objective:       To get the list of hyper cluster(s) & its properties
###############################################################################
sub get_hostcluster {

    my $self     = shift;
    my $cmd      = "Get-VMHostCluster ";
    my $param    = "-vmmserver $hostId";
    my $isPScmd  = 1;
    my $fileName = "file_get_hostcluster.txt";

    my $out = $self->_run_functions( $isPScmd, $cmd, $param, $fileName );
    return $out;
} ## end sub get_hostcluster

=head2 B<get_hostsystems()>

=over 4

=item Inputs

No inputs are needed. 
It just makes call to SCVMM to fetch the cluster list & its properties

=item Returns

Return reference to array of references to hashes, 
i.e. each value in array is reference to an hash, 

where each hash is a set of {key:value} pair containing the various 
properties of a Hosts like  Name, HostCluster etc.

=item Examples
    
    my $obj = new hyperv( $hostId, $hostType, $userName, $passWord );

    my $out = $obj->get_hostsystems();
    
    foreach (@$out) {
        print(
            "Name        :    $_->{'Name'}",
            "HostCluster    :    $_->{'HostCluster'}",
        );
    }

=cut

###############################################################################
# Subroutine Name: get_hostsystems()
# Objective: To get the list of hyperv parent hosts & its properties(if needed)
###############################################################################

sub get_hostsystems {

    my $self     = shift;
    my $cmd      = "get-vmhost ";
    my $param    = "-vmmserver $hostId";
    my $isPScmd  = 1;
    my $fileName = "file_get_hostsystems.txt";

    #error
    #my $param        = "-bVMMServer $hostId";

    my $out = $self->_run_functions( $isPScmd, $cmd, $param, $fileName );
    return $out;
} ## end sub get_hostsystems

=head2 B<create_virtualmachine()>

=over 4

=item Inputs

=item * $vmFilename, $vmName, $vmDescription, $vmOwner, 
$vmVMHost, $vmPath [mandatory]

provide the following parameters 

vmFilename, the name of VHD
vmName, the name of the Virtual Machine
vmDescription, some desccription for the Virtual Machine
vmOwner, the SCVMM who will be the Owner of the Virtual Machine
vmVMHost, the physical hostting HyperV server
vmPath, the path where 



=item Returns

Returns the created VM object & its properties

=item Examples
    
    my ($vmFilename, $vmName, $vmDescription, $vmOwner, $vmVMHost, $vmPath) = 
    ("hyperv5_disk1", "hyperv5", "test", 'SMQA\administrator', 
	'w2k8-184-153.smqa.lab.eng.btc.netapp.in', 'C:\ClusterStorage\Volume1');

    my $out 
	    = $obj->create_virtualmachine($vmFilename, $vmName, $vmDescription, 
		$vmOwner, $vmVMHost, $vmPath);
    
    foreach (@$out) {
        my ($Name, $Owner, $VirtualHardDisks, $VMHost ) = 
            ($_->{'Name'}, $_->{'Owner'}, 
			$_->{'VirtualHardDisks'}, $_->{'VMHost'});
        print(
            "\n $i ==== Inside thpl_create_virtualmachine === \n",
            "Name                :\t\t$_->{'Name'}",
            "Owner               :\t\t$_->{'Owner'}",
            "VirtualHardDisks    :\t\t$_->{'VirtualHardDisks'}",
            "VMHost              :\t\t$_->{'VMHost'}",
        );
        
        
        if (($vmName eq $Name) && ($vmVMHost eq $VMHost) ) {
            print ("\n\n VM create is successful \n\n");
        }    
    }    
    


=cut

###############################################################################
# Subroutine Name: create_virtualmachine()
# Objective:       To create Hyperv VM & set its properties(if needed)
###############################################################################

sub create_virtualmachine {

    my ( $self, $vmFilename, $vmName, $vmDescription, $vmOwner, $vmVMHost,
        $vmPath )
        = @_;
    my $cmd
        = '$TemplateJobGroup = [System.Guid]::NewGuid() ; New-VirtualDiskDrive -VMMServer localhost -IDE -Bus 0 -LUN 0 -JobGroup $TemplateJobGroup -Size 40960 -Dynamic -Filename '
        . $vmFilename
        . ' ; New-VM -VMMServer localhost -Name '
        . $vmName
        . ' -Description '
        . $vmDescription
        . ' -Owner '
        . $vmOwner
        . ' -VMHost '
        . $vmVMHost
        . ' -Path '
        . $vmPath
        . ' -JobGroup $TemplateJobGroup -RunAsynchronously  -RunAsSystem -StartAction NeverAutoTurnOnVM -StopAction SaveVM ';
    my $param    = " ";
    my $isPScmd  = 1;
    my $fileName = "file_create_vm_$vmName.txt";

    my $out = $self->_run_functions( $isPScmd, $cmd, $param, $fileName );
    return $out;

} ## end sub create_virtualmachine

=head2 B<delete_virtualmachine()>

=over 4

=item Inputs

=item * $vmName, $vmVMHost [mandatory]

provide the following parameters 

vmName, the name of the Virtual Machine
vmVMHost, the physical hosting HyperV server


=item Returns

Returns the deleted VM object & its properties

=item Examples

    my $out = $obj->delete_virtualmachine($vmName, $vmVMHost);
    
    foreach (@$out) {
    my ($Name, $VMHost ) = ($_->{'Name'}, $_->{'VMHost'} );
        print(
            "Name    :    $_->{'Name'}",
            "VMHost    :    $_->{'VMHost'}",
        );
        
        
        if (($vmName eq $Name) && ($vmVMHost eq $VMHost) ) {
            print ("\n\n VM delete is successful \n\n");
        }
    }    
    
=cut

###############################################################################
# Subroutine Name: delete_virtualmachine()
# Objective:       To delete Hyperv VM
###############################################################################

sub delete_virtualmachine {

    my ( $self, $vmName, $vmVMHost ) = @_;
    my $cmd
        = '$VM = '
        . "Get-VM -VMMServer $hostId"
        . '  | where { $_.VMHost.Name -eq \"'
        . $vmVMHost
        . '\" -and $_.Name -eq \"'
        . $vmName
        . '\" } ; Remove-VM $VM';
    my $param    = "  ";
    my $isPScmd  = 1;
    my $fileName = "file_delete_virtualmachine.txt";

    my $out = $self->_run_functions( $isPScmd, $cmd, $param, $fileName );
    return $out;

} ## end sub delete_virtualmachine

=head2 B<start_virtualmachine()>

=over 4

=item Inputs


=item * $vmName, $vmVMHost [mandatory]

provide the following parameters 

vmName, the name of the Virtual Machine
vmVMHost, the physical hosting HyperV server


=item Returns

Returns the VM object of the started VM with its properties

=item Examples

    my $out = $obj->start_virtualmachine($vmName, $vmVMHost);
    
    foreach (@$out) {
	    my ($Name, $VMHost, $StatusString ) 
		    = ($_->{'Name'}, $_->{'VMHost'}, $_->{'StatusString'} );
        print(
            "Name         :    $_->{'Name'}",
            "VMHost       :    $_->{'VMHost'}",
            "StatusString : $_->{'StatusString'}"
        );
        
 
        if (($vmName eq $Name) && ($vmVMHost eq $VMHost) && ($StatusString eq "Running")) {
            print ("\n\n VM start is successful \n\n");
        }
    }    
    
=cut

###############################################################################
# Subroutine Name: start_virtualmachine()
# Objective:       To start a Hyperv VM
###############################################################################
sub start_virtualmachine {

    my ( $self, $vmName, $vmVMHost ) = @_;
    my $cmd
        = '$VM = '
        . "Get-VM -VMMServer $hostId"
        . '  | where { $_.VMHost.Name -eq \"'
        . $vmVMHost
        . '\" -and $_.Name -eq \"'
        . $vmName
        . '\" } ; start-vm $VM';
    my $param    = "  ";
    my $isPScmd  = 1;
    my $fileName = "file_start_virtualmachine.txt";

    my $out = $self->_run_functions( $isPScmd, $cmd, $param, $fileName );
    return $out;

} ## end sub start_virtualmachine

=head2 B<stop_virtualmachine()>

=over 4

=item Inputs

=item * $vmName, $vmVMHost [mandatory]

provide the following parameters 

vmName, the name of the Virtual Machine
vmVMHost, the physical hosting HyperV server


=item Returns

Returns the VM object of the stopped VM with its properties

=item Examples

    my $out = $obj->stop_virtualmachine($vmName, $vmVMHost);
    
    foreach (@$out) {
	    my ($Name, $VMHost, $StatusString ) 
		    = ($_->{'Name'}, $_->{'VMHost'}, $_->{'StatusString'} );
        print(
            "Name         :    $_->{'Name'}",
            "VMHost       :    $_->{'VMHost'}",
            "StatusString : $_->{'StatusString'}"
        );
        
        
        if (($vmName eq $Name) && ($vmVMHost eq $VMHost) && ($StatusString eq "Stopped")) {
            print ("\n\n VM stop is successful \n\n");
        }
    }    
    
=cut

###############################################################################
# Subroutine Name: stop_virtualmachine()
# Objective:       To stop a Hyperv VM
###############################################################################

sub stop_virtualmachine {

    my ( $self, $vmName, $vmVMHost ) = @_;
    my $cmd
        = '$VM = '
        . "Get-VM -VMMServer $hostId"
        . '  | where { $_.VMHost.Name -eq \"'
        . $vmVMHost
        . '\" -and $_.Name -eq \"'
        . $vmName
        . '\" } ; stop-vm $VM';
    my $param    = "  ";
    my $isPScmd  = 1;
    my $fileName = "file_start_virtualmachine.txt";

    my $out = $self->_run_functions( $isPScmd, $cmd, $param, $fileName );
    return $out;

} ## end sub stop_virtualmachine

=head2 B<rename_virtualmachine()>

=over 4

=item Inputs


=item * $vmName, $vmVMHost, $newVMName [mandatory]

provide the following parameters 

vmName, the name of the Virtual Machine
vmVMHost, the physical hosting HyperV server
newVMName, the new name of the Virtual Machine

=item Returns

Returns the VM object of the renamed VM with its properties

=item Examples

    my $out = $obj->rename_virtualmachine($vmName, $vmVMHost, $newVMName);
    
    foreach (@$out) {
    my ($Name, $VMHost ) = ($_->{'Name'} );
        print(
            "Name         :    $_->{'Name'}",
            "VMHost       :    $_->{'VMHost'}"
        );
        
    
        if ( ($newVMName eq $Name) && ($vmVMHost eq $VMHost) ) {
            print ("\n\n VM rename is successful \n\n");
        }
    }    
    
=cut

###############################################################################
# Subroutine Name: rename_virtualmachine()
# Objective:       To stop a Hyperv VM
###############################################################################

sub rename_virtualmachine {

    my ( $self, $vmName, $vmVMHost, $newVMName ) = @_;
    my $cmd
        = '$VM = '
        . "Get-VM -VMMServer $hostId"
        . '  | where { $_.VMHost.Name -eq \"'
        . $vmVMHost
        . '\" -and $_.Name -eq \"'
        . $vmName
        . '\" } ; set-vm -VM $VM -name \"'
        . $newVMName . '\"';
    my $param    = "  ";
    my $isPScmd  = 1;
    my $fileName = "file_rename_virtualmachine.txt";

    my $out = $self->_run_functions( $isPScmd, $cmd, $param, $fileName );
    return $out;

} ## end sub rename_virtualmachine

###############################################################################
# P R I V A T E   M E T H O D S
###############################################################################

###############################################################################
# Subroutine Name: _run_functions()
# Objective      : To consolidate & run a set of methods
#                  to get the output of the command
###############################################################################

sub _run_functions {

    my ( $self, $isPScmd, $cmd, $param, $fileName ) = @_;
    my $err
        = ' ; $err_start=\"vimqa_error_start\" ; $err_end=\"vimqa_error_end\" ; $err_start; $ERROR ; $err_end ';
    $param = $param . $err;
    my @rawRshOutput = _invokeRsh( $isPScmd, "\"$cmd $param\"" );
    open WRFILE, ">$fileName";
    print WRFILE @rawRshOutput;
    close WRFILE;
    $Log->debug("In debug mode, the raw rsh output of PS command is logged");
    $Log->debug(@rawRshOutput);
    $self->_clear_err();

    if ( $rawRshOutput[0]
        =~ m/vimqa_error_start(\r+)(\n+)vimqa_error_end/ )
    {
    } else {
        $Log->comment(
            "\nThe Error msg start = \n  $rawRshOutput[0] \nError msg end \n  "
        );
        $self->_set_err("$rawRshOutput[0]");
    } ## end else [ if ( $rawRshOutput[0] ...)]

    my $processedRshOutput = _processFile($fileName);
    $Log->debug("Processed raw rsh output of PS command is logged");
    $Log->debug($processedRshOutput);

    if ( unlink($fileName) == 0 ) {
        $Log->comment("\nFile $fileName was not deleted.\n");
    }

    my $out = _parseOutput($processedRshOutput);
    return $out;

} ## end sub _run_functions

###############################################################################
# Subroutine Name: _invokeRsh()
# Objective:       To invoke a rsh call to the host & get the output
###############################################################################
sub _invokeRsh {

    my $isPS = shift;
    my $cmd  = shift;
    my $psc
        = 'C:\Windows\System32\WindowsPowerShell\v1.0\powershell.exe -PSConsoleFile '
        . '"%programFiles%\Microsoft System Center Virtual Machine Manager 2008 R2\bin\cli.psc1" ';
    if ($isPS) { $cmd = $psc . $cmd; }
    my ( undef, undef, $out, undef )
        = say( $connObj, "cmd /c echo \\n | $cmd ",
        -timeout, 3600, "-noverbose" );
    $Log->comment("cmd /c echo \\n | $cmd ");

    return $out;
} ## end sub _invokeRsh

###############################################################################
# Subroutine Name: _processFile()
# Objective:       To process the file containing the rsh output & clean it up
###############################################################################
sub _processFile {

    my $fileName = $_[0];
    my $textfor;
    my $uncleanstr;
    my $clenstr;

    open( RDFILE, "<$fileName" );
    while (<RDFILE>) {
        if ( $_ eq "\n" ) {
            $textfor = $textfor . sprintf("\n********");
            next;
        }
        chomp;
        if (/^ /) {
            $uncleanstr = $_;
            $clenstr    = _cleanup( \$uncleanstr );
            $textfor    = $textfor . sprintf($clenstr);
        } else {
            $uncleanstr = $_;
            $clenstr    = _cleanup( \$uncleanstr );
            $textfor    = $textfor . "\n" . sprintf($clenstr);
        }
    } ## end while (<RDFILE>)
    close(RDFILE);
    return $textfor;

} ## end sub _processFile

###############################################################################
# Subroutine Name: _cleanup()
# Objective      : To do some cleanup of extra \r, \n & for easier regex
###############################################################################
sub _cleanup {

    my @strref = @_;
    my $str    = $strref[0];
    $$str =~ s/:\s+/::/;
    $$str =~ s/\s+//g;
    $$str =~ s/\\r\\r\\n//g;
    return ($$str);

} ## end sub _cleanup

###############################################################################
# Subroutine Name: _parseOutput()
# Objective      : To parse the raw rsh output
###############################################################################
sub _parseOutput {

    my $recordsList = $_[0];
    my @recordsSplit;
    my @lineSplit;
    my @myfinal;

    @recordsSplit = split( /\n\n/, $recordsList );
    $Log->debug("In case of some issue with RSH path, look here");
    $Log->debug( $recordsSplit[0] );
    if ( not $recordsSplit[0] ) {
        shift(@recordsSplit);
    }
    pop(@recordsSplit);
    foreach (@recordsSplit) {
        my %mydata;
        @lineSplit = split( /\n/, $_ );
        foreach (@lineSplit) {
            my ( $key, $value ) = split( /::/, $_ );

            #Remove leading & trailing braces from values if present
            $value =~ s/^{//g;
            $value =~ s/}$//g;
            $mydata{$key} = $value;
        } ## end foreach (@lineSplit)
        push @myfinal, \%mydata;
    } ## end foreach (@recordsSplit)
    return \@myfinal;

} ## end sub _parseOutput

1;

