# $Id:
#
# Copyright (c) 2010 NetApp.
# All rights reserved.
#
#########################################################
## @summary Library of functions to retreive info from HyperV
##
##
## @author Prasad S (prasads@netapp.com)
##
## @status public
##
## @pod here
#########################################################

# Package Name
##############
package stor_map_hyperv;

require Exporter;
@ISA    = qw/Exporter/;
@EXPORT = qw/getStorageDetails/;

use Tharn;
use strict;
use Error;
use XML::Simple;
use Data::Dumper;
use hyperv;
use filerutils;
use vars qw(@ISA);
@ISA = qw(Error);


# Declare global variables
use vars qw($class $hostId $hostType $userName $passWord);
use vars qw($hostObj $connObj);


###############################################################################
# P U B L I C   M E T H O D S
###############################################################################

###############################################################################
# Subroutine Name: new()
# Objective:       Create a new hyperv object
###############################################################################
sub new {

	($class, $hostId, $hostType, $userName, $passWord) = @_;
    my $self = {};
    bless( $self, $class );
    return $self;
} ## end sub new


=head2 B<getStorageDetails()>

=over 4

=item Inputs

No inputs are needed. It just makes call to SDW to fetch the sdcli disk list

=item Returns

Return reference to array of references to hashes, 
i.e. each value in array is reference to an hash, 

where each hash is a set of {key:value} pair containing the various properties
of a VM like Name, VHDs etc.

=item Examples

    my $storageMap = new stor_map_hyperv( );( $hostId, $hostType, $userName, $passWord );
    
    my $StorageDetails = $storageMap->getStorageDetails('10.225.239.26');
    if ( $storageMap->error ) {
        $Test->nlog( $storageMap->errmsg );
        return $TCD::FAIL;
    }
	$Test->nlog(  Dumper($StorageDetails) );
	
	foreach (@$StorageDetails) {
        $Test->nlog("LUNPath:$_->{'LUNPath'}");
		$Test->nlog("StorageSystem:$_->{'StorageSystem'}");
		my $vms1 = $_->{'Hyper-VVMsName'} ;
		$Test->nlog( @$vms1);
    }

=cut

################################################################################
# Subroutine Name: getStorageDetails()
# Objective:       To get the storage object from the host
################################################################################

sub getStorageDetails {

    #my ($self ) = @_;
	my $self = shift(@_);
	my $filers = shift(@_);
    my $cmd          = "sdcli disk list";
	my $fileName     = "sdcli_disk_list_$hostId.txt";
	my $filerutils  = new filerutils();
	
	if ( !$filers ) {
        $self->_set_err("Filers details are missing");
        return undef;
    }
	
	$self->_connect_host( $hostId, $hostType, $userName, $passWord );
    my $out = $self->_run_functions( $cmd, $fileName );
	$Log->comment(ref($filers));
    foreach(@$out) {
     if (defined $_->{'LUNPath'}) {
	    my $filer = $self->_get_filer($filers,$_->{'IPAddresses'});
		if ($filer) {
		my $vol = $self->_get_path($_->{'LUNPath'});
        my $aggrname = $filerutils->get_aggr($filer,$vol);
		$_->{'Aggr'} = $aggrname; 
		} else {
		  $_->{'Aggr'} = undef;
		}
     }
    }
    return $out;

}


################################################################################
# Subroutine Name: _get_path
# Objective:       To extract volume/lun path from snapdrive output
################################################################################
sub _get_path {
    my $self = shift(@_);
    my $path = shift(@_);
    if ( $path =~ /.*:(.*)\// ) {
        return $1;
    }
}

################################################################################
# Subroutine Name: _get_filer
# Objective:       To get the corresponding filer from array of hashes.
################################################################################
sub _get_filer {
    my $self       = shift(@_);
    my $filers     = shift(@_);
    my $filer_name = shift(@_);

    foreach (@$filers) {
        if ( $_->{filerName} eq $filer_name ) {
            return $_;
        }
    }
} 

#sub getStorageDetailsFiler {
#
#	my ($self, $input ) = @_;
#}

###############################################################################
# P R I V A T E   M E T H O D S
###############################################################################

###############################################################################
# Subroutine Name: _connect_host()
# Objective      : To initiate a rsh connection to Windows host
###############################################################################

sub _connect_host {

	my ( $self, $hostId, $hostType, $userName, $passWord ) = @_;
	$Log->debug("Inside _connect_host");
	$Log->debug("class $self, hostId $hostId, hostType $hostType, userName $userName, passWord $passWord");
	sleep (5);
    eval {
        $hostObj = host("$hostId;hosttype=$hostType");
        $connObj
            = connect("#rsh-persay;username=$userName;password=$passWord");
    };
    if ($@) {
        $Log->comment($@);
        $self->_clear_err();
        $self->_set_err($@);
    }

}

###############################################################################
# Subroutine Name: _run_functions()
# Objective      : To consolidate & run a set of methods
#                  to get the output of the command
###############################################################################
sub _run_functions {

    my ( $self, $cmd, $fileName ) = @_;
    
    my @rawRshOutput = _invokeRsh( $cmd );
    open WRFILE, ">$fileName";
    print WRFILE @rawRshOutput;
    close WRFILE;
    $Log->debug("In debug mode, the raw rsh output of PS command is logged");
    $Log->comment(@rawRshOutput);
    $self->_clear_err();

    my $processedRshOutput = _processFile($fileName);
    $Log->debug("Processed raw rsh output of command is logged");
    $Log->debug($processedRshOutput);

    if ( unlink($fileName) == 0 ) {
        $Log->comment("\nFile $fileName was not deleted.\n");
    }

    my $out = _parseOutput($processedRshOutput);
    return $out;

} ## end sub _run_functions

###############################################################################
# Subroutine Name: _invokeRsh()
# Objective:       To invoke a rsh call to the host & get the output
###############################################################################
sub _invokeRsh {

    my $cmd  = shift;

    my ( undef, undef, $out, undef )
        = say( $connObj, "cmd /c echo \\n | $cmd ",
        -timeout, 3600, "-noverbose" );
    $Log->comment("cmd /c echo \\n | $cmd ");

    return $out;
} ## end sub _invokeRsh

###############################################################################
# Subroutine Name: _processFile()
# Objective:       To process the file containing the rsh output & clean it up
###############################################################################
sub _processFile {

    my $fileName = $_[0];
    my $textfor;
    my $uncleanstr;
    my $clenstr;

    open( RDFILE, "<$fileName" );
    while (<RDFILE>) {
        if ( $_ eq "\n" ) {
            $textfor = $textfor . sprintf("\n********");
            next;
        }
        chomp;
        if (/^ /) {
            $uncleanstr = $_;
            $clenstr    = _cleanup( \$uncleanstr );
            $textfor    = $textfor . sprintf($clenstr);
        } else {
            $uncleanstr = $_;
            $clenstr    = _cleanup( \$uncleanstr );
            $textfor    = $textfor . "\n" . sprintf($clenstr);
        }
    } ## end while (<RDFILE>)
    close(RDFILE);
    return $textfor;

} ## end sub _processFile

###############################################################################
# Subroutine Name: _cleanup()
# Objective      : To do some cleanup of extra \r, \n & for easier regex
###############################################################################
sub _cleanup {

    my @strref = @_;
    my $str    = $strref[0];
    $$str =~ s/:\s+/::/;
    $$str =~ s/\s+//g;
    $$str =~ s/\\r\\r\\n//g;
    return ($$str);

} ## end sub _cleanup

###############################################################################
# Subroutine Name: _parseOutput()
# Objective      : To parse the raw rsh output
###############################################################################
sub _parseOutput {

	my ( $processedRshOutput ) = $_[0];
    my ( @recordsSplit, @lineSplit, @myfinal );

    @recordsSplit = split( /UNCPath/, $processedRshOutput );
    $Log->debug("In case of some issue with RSH path, look here");
    $Log->debug( $recordsSplit[0] );
	$Log->debug ("\n recordsSplit[0] =-$recordsSplit[0]-\n");	
	chomp($recordsSplit[0]);
    if ( not $recordsSplit[0] ) {
		$Log->debug ("\n After chomp recordsSplit[0] =-$recordsSplit[0]-\n");	
        shift(@recordsSplit);
    }
    
    foreach (@recordsSplit) {
        my %mydata;
        @lineSplit = split( /\n/, $_ );
        foreach (@lineSplit) {
            my ( $key, $value ) = split( /::/, $_ );
			if ( ($key eq 'Hyper-VVMsName')|| ( $key eq 'Hyper-VVMName') ) { 
	# To over come the singular and plural of the sdcli verbose output while listing the vm(s)
				$key =~ s/Hyper-VVMName/Hyper-VVMsName/g;
				$Log->debug("\nInside if loop.  $key :: $value\n");
				my @vms = split( /,/, $value ); 
				$Log->debug("\nArray is @vms\n");
				$mydata{$key} = \@vms;
			}
			else {
				$mydata{$key} = $value;
			}
        } ## end foreach (@lineSplit)
        push @myfinal, \%mydata;
    } ## end foreach (@recordsSplit)

    return \@myfinal;
 
} ## end sub _parseOutput

1;

