# $Id:
#
# Copyright (c) 2010 NetApp.
# All rights reserved.
#
################################################################################
## @summary Library of functions to access DFM Soap APIs to perform
##          the OM(operation manager) operations.
##
## @author Chetna Samant (chetna@netapp.com)
##
## @status public
##
## @pod here
################################################################################

#
# Package Name
#
package OpsMgrSoap;

#
# Imported modules
#
#use lib '/usr/software/test/share/nate/latest/lib';
#use strict;
use MyInterfaces::DfmService::DfmPort;
use SOAP::Transport::HTTP;
use Data::Dumper;
use XML::Simple;
use NATE::Log;
use misc;
use Tharn;
use Error;

#
#  Global variables declaration
#

use vars qw($interface);

#
#Inheriting from base class
#
@ISA=qw(Error);


=head NAME

 OM- This module is to access DFM soap APIs to perform OM(operation manager) operations.

=head1 SYNOPSIS

use DFM;
my $dfmObj = new dfm(<hostname>,<username>,<password>,[port no]);

=head1 DESCRIPTION

This package offers the functions to perform operations on DFM objects which internally calls
DFM SOAP APIs.
While creating the dfm I<object>, I<hostname>, I<username> and  I<password> of the DFM host has to be passed as arguments.
The optional parameter I<port no> refers to Soap services port no for DFM. This option is used only when there is change in the
DFM Soap service port no in future. So we can create op object as follows,

my $dfmObj = new dfm(<hostname>,<username>,<password>); 

=head1 FUNCTIONS

All the get functions will return reference array, each value in  array is reference to hash.
More details are explained in each function as below.

=cut

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new Soap object
################################################################################

sub new {
    my $class = shift;
    my $self  = {
        _hostName => shift,
        _userName => shift,
        _password => shift,
        _port     => shift,
    };

    $self->{_port} = "8488" if not defined $self->{_port};

    *SOAP::Transport::HTTP::Client::get_basic_credentials = sub {
        return ( $self->{_userName}, $self->{_password} );
    };

    $interface = MyInterfaces::DfmService::DfmPort->new(
        { proxy => "https://$self->{_hostName}:$self->{_port}/apis/soap/v1" } );
    if ( !$interface ) {
        die(
    "Could not create an instance for accessing Web Service. Check if DFM services are running on the server.\n"
        );
    }
    $Log->comment("#######Interface = $interface############");
    bless( $self, $class );
    return $self;
}

################################################################################
# Subroutine Name: get_events_report_query()
# Objective:       Get the listed events from API call: report-get-data
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  If NO INPUTS specified, mandatory params given default values
#                  QueryName:    query-event-list
#                  OutputFormat: json
################################################################################

sub get_events_report_query {
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $output );    #variable initializations

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "IncludeTotalRowCount" ) ) {
            $modHash->{IncludeTotalRowCount} = $value;
        }
        elsif ( !( $key cmp "OutputFormat" ) ) {
            $modHash->{OutputFormat} = $value;
        }
        elsif ( !( $key cmp "QueryAttributeName" ) ) {
            $modHash->{QueryAttributeNameList}{QueryAttributeName} = $value;
        }
        elsif ( !( $key cmp "QueryFilterName" ) ) {
            $modHash->{QueryFilterList}{QueryFilter}{QueryFilterName} = $value;
        }
        elsif ( !( $key cmp "QueryFilterOperator" ) ) {
            $modHash->{QueryFilterList}{QueryFilter}{QueryFilterOperator} =
              $value;
        }
        elsif ( !( $key cmp "QueryFilterValue" ) ) {
            $modHash->{QueryFilterList}{QueryFilter}{QueryFilterValue} = $value;
        }
        elsif ( !( $key cmp "QueryFilterLogicalOperator" ) ) {
            $modHash->{QueryFilterLogicalOperator} = $value;
        }
        elsif ( !( $key cmp "QueryName" ) ) {
            $modHash->{QueryName} = $value;
        }
        elsif ( !( $key cmp "QuerySortDirection" ) ) {
            $modHash->{QuerySortList}{QuerySort}{QuerySortDirection} = $value;
        }
        elsif ( !( $key cmp "QuerySortName" ) ) {
            $modHash->{QuerySortList}{QuerySort}{QuerySortName} = $value;
        }
        elsif ( !( $key cmp "RowCount" ) ) {
            $modHash->{RowCount} = $value;
        }
        elsif ( !( $key cmp "StartIndex" ) ) {
            $modHash->{StartIndex} = $value;
        }
        else {
            $Log->comment("INVALID INPUT PASSED!");     #Exception Handling TBD!
        }
    }
    
    #passing mandatory parameters
    if (not exists $modHash->{OutputFormat}) {
        $modHash->{OutputFormat} = "json";
    }
    if (not exists $modHash->{QueryName}) {
        $modHash->{QueryName} = "query-event-list";
    }

    #printing out the data structure to be passed in APIcall
    $Log->comment( Dumper($modHash) );
    
    #access Web Service
    $result = $interface->ReportQueryGetData($modHash);
    $Log->comment($result);
    $Log->comment( Dumper($result) );
    
    #error handling
    if (not $result) {
        my $e = $result->get_faultstring();
        #$Log->comment("FAULTSTRING: $e");
        my $error = "Error calling SOAP method:" . _get_fault_string($e);
	$self->_clear_err();
	$self->_set_err($error);
    }
    #if (not $result) { return _check_for_error($result) ; }
    
    #process output and return
    my $TOTALROWS = $result->get_TotalRows();
    $Log->comment("$result->get_QueryResult()");
    $Log->comment("-------------------->>>>>>>>>$TOTALROWS");

}

################################################################################
# Subroutine Name: alarm_create()
# Objective:       Create a DFM alarm based on the parameters passed
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT: 
#                 { EmailAddress=>"<email-id>"}
################################################################################

sub alarm_create {
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "AdminName" ) ) {
            $modHash->{AlarmInfo}{AdminLoginNames}{AdminName} = $value;
        }
        elsif ( !( $key cmp "AlarmId" ) ) {
            $modHash->{AlarmInfo}{AlarmId} = $value;
        }
        elsif ( !( $key cmp "AlarmScript" ) ) {
            $modHash->{AlarmInfo}{AlarmScript} = $value;
        }
        elsif ( !( $key cmp "AlarmScriptRunas" ) ) {
            $modHash->{AlarmInfo}{AlarmScriptRunas} = $value;
        }
        elsif ( !( $key cmp "EmailAddress" ) ) {
            $modHash->{AlarmInfo}{EmailAddresses}{EmailAddress} = $value;
        }
        elsif ( !( $key cmp "EventClass" ) ) {
            $modHash->{AlarmInfo}{EventClass} = $value;
        }
        elsif ( !( $key cmp "EventName" ) ) {
            $modHash->{AlarmInfo}{EventName} = $value;
        }
        elsif ( !( $key cmp "EventSeverity" ) ) {
            $modHash->{AlarmInfo}{EventSeverity} = $value;
        }
        elsif ( !( $key cmp "GroupId" ) ) {
            $modHash->{AlarmInfo}{GroupId} = $value;
        }
        elsif ( !( $key cmp "GroupName" ) ) {
            $modHash->{AlarmInfo}{GroupName} = $value;
        }
        elsif ( !( $key cmp "IsDisabled" ) ) {
            $modHash->{AlarmInfo}{IsDisabled} = $value;
        }
        elsif ( !( $key cmp "IsRepeatNotify" ) ) {
            $modHash->{AlarmInfo}{IsRepeatNotify} = $value;
        }        
        elsif ( !( $key cmp "AdminName" ) ) {
            $modHash->{AlarmInfo}{AdminName} = $value;
        }
        elsif ( !( $key cmp "RepeatInterval" ) ) {
            $modHash->{AlarmInfo}{RepeatInterval} = $value;
        }
        elsif ( !( $key cmp "TimeFrom" ) ) {
            $modHash->{AlarmInfo}{TimeFrom} = $value;
        }
        elsif ( !( $key cmp "TimeTo" ) ) {
            $modHash->{AlarmInfo}{TimeTo} = $value;
        }
        elsif ( !( $key cmp "TrapCommunity" ) ) {
            $modHash->{AlarmInfo}{TrapDestinations}{TrapDestination}{TrapCommunity} = $value;
        }
        elsif ( !( $key cmp "TrapHost" ) ) {
            $modHash->{AlarmInfo}{TrapDestinations}{TrapDestination}{TrapHost} = $value;
        }
        elsif ( !( $key cmp "TrapPort" ) ) {
            $modHash->{AlarmInfo}{TrapDestinations}{TrapDestination}{TrapPort} = $value;
        }
        else {
            $Log->comment("INVALID INPUT PASSED!");     #Exception Handling TBD!
        }
    }

    #printing out the data structure to be passed in APIcall
    $Log->comment( Dumper($modHash) );
    
    #access Web Service
    $result = $interface->AlarmCreate($modHash);
    #$Log->comment($result);
    #$Log->comment( Dumper($result) );
    
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else {
        
        #process output and return
        
        my $attrib = $result->get_AlarmId();
        $hash->{AlarmId} = "$attrib";
        push (@arr_hashes,$hash);
        return \@arr_hashes;
    }
    
}
    
    
################################################################################
# Subroutine Name: alarm_modify()
# Objective:       Modify a DFM alarm based on the parameters passed
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {AlarmId=>"<alarm-id>",}
################################################################################

sub alarm_modify { 
my ( $self, $inputFields ) = @_;
my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

#clear any previous Error set
$self->_clear_err();

#forming the data structure

    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "AdminName" ) ) {
            $modHash->{AlarmInfo}{AdminLoginNames}{AdminName} = $value;
        }
        elsif ( !( $key cmp "AlarmId" ) ) {
            $modHash->{AlarmInfo}{AlarmId} = $value;
        }
        elsif ( !( $key cmp "AlarmScript" ) ) {
            $modHash->{AlarmInfo}{AlarmScript} = $value;   
        }
        elsif ( !( $key cmp "AlarmScriptRunas" ) ) {
            $modHash->{AlarmInfo}{AlarmScriptRunas} = $value;
        }
        elsif ( !( $key cmp "EmailAddress" ) ) {
            $modHash->{AlarmInfo}{EmailAddresses}{EmailAddress} = $value;
        }
        elsif ( !( $key cmp "EventClass" ) ) {
            $modHash->{AlarmInfo}{EventClass} = $value;
        }
        elsif ( !( $key cmp "EventName" ) ) {
            $modHash->{AlarmInfo}{EventName} = $value;
        }
        elsif ( !( $key cmp "EventSeverity" ) ) {
            $modHash->{AlarmInfo}{EventSeverity} = $value;
        }
        elsif ( !( $key cmp "GroupId" ) ) {
            $modHash->{AlarmInfo}{GroupId} = $value;
        }
        elsif ( !( $key cmp "GroupName" ) ) {
            $modHash->{AlarmInfo}{GroupName} = $value;
        }
        elsif ( !( $key cmp "IsDisabled" ) ) {
            $modHash->{AlarmInfo}{IsDisabled} = $value;
        }
        elsif ( !( $key cmp "IsRepeatNotify" ) ) {
            $modHash->{AlarmInfo}{IsRepeatNotify} = $value;
        }        
        elsif ( !( $key cmp "AdminName" ) ) {
            $modHash->{AlarmInfo}{AdminName} = $value;
        }
        elsif ( !( $key cmp "RepeatInterval" ) ) {
            $modHash->{AlarmInfo}{RepeatInterval} = $value;
        }
        elsif ( !( $key cmp "TimeFrom" ) ) {
            $modHash->{AlarmInfo}{TimeFrom} = $value;
        }
        elsif ( !( $key cmp "TimeTo" ) ) {
            $modHash->{AlarmInfo}{TimeTo} = $value;
        }
        elsif ( !( $key cmp "TrapCommunity" ) ) {
            $modHash->{AlarmInfo}{TrapDestinations}{TrapDestination}{TrapCommunity} = $value;
        }
        elsif ( !( $key cmp "TrapHost" ) ) {
            $modHash->{AlarmInfo}{TrapDestinations}{TrapDestination}{TrapHost} = $value;
        }
        elsif ( !( $key cmp "TrapPort" ) ) {
            $modHash->{AlarmInfo}{TrapDestinations}{TrapDestination}{TrapPort} = $value;
        }
        else {
            $Log->comment("INVALID INPUT PASSED!");     #Exception Handling TBD!
        }
    }   

    #printing out the data structure to be passed in APIcall
    $Log->comment( Dumper($modHash) );
    
    #access Web Service
    $result = $interface->AlarmModify($modHash);
    $Log->comment($result);
    #$Log->comment( Dumper($result) );
    
    #error handling
    
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else {
        return[{status=>"SUCCESS"}];
    }
}

################################################################################
# Subroutine Name: alarm_destroy()
# Objective:       Destroy a DFM alarm
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {AlarmId=>"<AlarmId>"}
################################################################################
sub alarm_destroy { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "AlarmId" ) ) {
            $modHash->{AlarmId} = $value;
        }
    }

    #printing out the data structure to be passed in APIcall
    $Log->comment( Dumper($modHash) );
    
    #access Web Service
    $result = $interface->AlarmDestroy($modHash);
    $Log->comment($result);
    #$Log->comment( Dumper($result) );
    
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else {
        return[{status=>"SUCCESS"}];
    }
}
  
################################################################################
# Subroutine Name: alarm_test()
# Objective:       Test an alarm by performing it's trigger actions.
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {AlarmId =>"<alarm-id>",}
################################################################################

sub alarm_test { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure not reqd
    $modHash = $inputFields;

    #printing out the data structure to be passed in APIcall
    $Log->comment( Dumper($modHash) );

    #access Web Service
    $result = $interface->AlarmTest($modHash);
    
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else {
        return[{status=>"SUCCESS"}];
    }
}  
  
################################################################################
# Subroutine Name: group_add_member()
# Objective:       Add one or more members to an existing group
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {GroupNameorId=>"<Groupid>", MemberNameOrId=>"<MemberId>",}
################################################################################

sub group_add_member { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure not reqd
    $modHash = $inputFields;

    #printing out the data structure to be passed in APIcall
    $Log->comment( Dumper($modHash) );

    #access Web Service
    $result = $interface->GroupAddMember($modHash);
    
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else {
        return[{status=>"SUCCESS"}];
    }
}

################################################################################
# Subroutine Name: group_create()
# Objective:       Create a new group
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {GroupName=>"<group-name>",}
################################################################################

sub group_create { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );
    
    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure not reqd
    $modHash = $inputFields;

    #printing out the data structure to be passed in APIcall
    $Log->comment( Dumper($modHash) );

    #access Web Service
    $result = $interface->GroupCreate($modHash);
    
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else {
        my $attrib = $result->get_GroupId();
        $hash->{GroupId} = "$attrib";
        push (@arr_hashes,$hash);
        return \@arr_hashes;
    }
}

################################################################################
# Subroutine Name: group_delete_member()
# Objective:       Delete members from a group
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {GroupNameorId=>"<group-name>",MemberNameOrId=>"<member">,}
################################################################################
sub group_delete_member { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure not reqd
    $modHash = $inputFields;

    #printing out the data structure to be passed in APIcall
    $Log->comment( Dumper($modHash) );

    #access Web Service
    $result = $interface->GroupDeleteMember($modHash);
    
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else {
        return[{status=>"SUCCESS"}];
    }
}

################################################################################
# Subroutine Name: group_destroy()
# Objective:       Destroy one or more groups and their subgroups
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {GroupNameOrId=>"<group-name>"}
################################################################################

sub group_destroy { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure not reqd
    $modHash = $inputFields;

    #printing out the data structure to be passed in APIcall
    $Log->comment( Dumper($modHash) );

    #access Web Service
    $result = $interface->GroupDestroy($modHash);
    
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else {
        return[{status=>"SUCCESS"}];
    }
}

################################################################################
# Subroutine Name: group_list()
# Objective:       List groups or the members of a group
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT: None
#                  By default, all groups would be listed
################################################################################
sub group_list { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();

    #forming the data structure not reqd
    if ($inputFields) {
        $modHash = $inputFields;
        
        #printing out the data structure to be passed in APIcall
        my ( $key, $value );
        while ( ( $key, $value ) = each %$modHash ) {
            $Log->comment("$key = $value; \n");
        }
        $Log->comment( Dumper($modHash) );
        
        #access Web Service
        $result = $interface->GroupListIterStart($modHash);
    }
    else {
        $result = $interface->GroupListIterStart();
    }
    
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }

    my $tag = $result->get_Tag(); 
    my $max = $result->get_Records();
      #stores the tag and no. of records which need to be passed to
      #GroupListIterStartResult
    $Log->comment("The maximum records obtained: $max");
                  
    #access Web Service
    $result =
      $interface->GroupListIterNext({ Maximum => "$max", Tag => $tag},,);
      
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    $Log->comment($result->get_Records());
    my $xmlOutput = $result->get_GroupList();
    $Log->comment($xmlOutput);
    
    #open(TFILE,">dummy.xml")||die("can't open file to write");
    #open TFILE, ">", "dummy.xml";
    #print TFILE "$xmlOutput";
    #close TFILE;
    open(CMD, "print $xmlOutput |< dumb.xml "); 
    #open(CMD, "| print $xmlOutput > dumb.xml");

    #open (T1FILE,"dummy.xml");
    # create object
    
    #my $xobj = new XML::SIMPLE();
    #my $xobj = $xobj->XMLin('dummy.xml');
    #print $xobj;
    
    #my $booklist = XMLin('booklist.xml');
    #my $fico = <T1FILE>;
    #$Log->comment("$fico");
    
    my $groupList = XMLin('CMD');
    $Log->comment("ERROR after XMLin!");
    
    #my $xmlref = $xobj->XMLout($groupList);
    $Log->(comment(Dumper($groupList)));
    
    my $xs = new XML::Simple();
    my $ref = $xs->XMLin("<p>Hello world!</p>");
    my $xml = $xs->XMLout($ref);
    print $xml;

    $interface->GroupListIterEnd({Tag=>$tag});
    unlink("dummy.xml");
}

################################################################################
# Subroutine Name: group_get()
# Objective:       Get the options for a group
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT: 
#                  {GroupNameOrId=>"<group-name>"}
################################################################################
sub group_get { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure not reqd
    $modHash = $inputFields;
        
    #printing out the data structure to be passed in APIcall
    $Log->comment( Dumper($modHash) );
        
    #access Web Service
    $result = $interface->GroupGetOptions($modHash);

    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else  {
        $Log->comment($result->get_GroupOptionInfo());
        $result
    }
}

################################################################################
# Subroutine Name: event_generate()
# Objective:       Generate events in the DFM system
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {EventName=>"<event-name>",Source=>"source"}
################################################################################
sub event_generate { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure not reqd
    $modHash = $inputFields;
    
    #printing out the data structure to be passed in APIcall
    $Log->comment(Dumper($modHash) );

    #access Web Service
    $result = $interface->EventGenerate($modHash);
    $Log->comment($result);
    
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else {
        return[{status=>"SUCCESS"}];
    }
}
    
################################################################################
# Subroutine Name: event_acknowledge()
# Objective:       Acknowledge events.
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {EventId=>"<event-id>"}    or
#                  {EventIdType=>"<event-ids>"}
################################################################################
sub event_acknowledge { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure

    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "EventId" ) ) {
            $modHash->{EventId} = $value;
        }
    elsif ( !( $key cmp "EventIdType" ) ) {
        $modHash->{EventIdList}{EventIdType} = $value;
        }
    else {
            $Log->comment("INVALID INPUT PASSED!");     #Exception Handling TBD!
        }
    }
    
    #printing out the data structure to be passed in APIcall
    $Log->comment(Dumper($modHash) );

    #access Web Service
    $result = $interface->EventAcknowledge($modHash);
    $Log->comment($result);
   
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    #if (not $result) { return _check_for_error($result) ; }
    
    else {
        if ($modHash->{"EventId"})  {
            my $attrib = $result->get_EventAcknowledged();
            $hash->{EventAcknowledged} = "$attrib";
            $Log->comment("here is the hash:   $hash->{EventAcknowledged}");
            push (@arr_hashes,$hash);
        }
        elsif ($modHash->{EventIdList}{EventIdType})  {
            my $attrib = $result->get_EventsAcknowledgedList();
            my $attrib1 = $attrib->get_EventActionInfo();
            my $attrib2 = $attrib1->get_EventId();
            $hash->{EventId} = "$attrib2";
            $attrib2 = $attrib1->get_Timestamp();
            $hash->{Timestamp} = "$attrib2";
            $attrib2 = $attrib1->get_ErrorCode();
            $hash->{ErrorCode} = "$attrib2";
            $attrib2 = $attrib1->get_ErrorMessage();
            $hash->{ErrorMessage} = "$attrib2";
            push (@arr_hashes,$hash);
        }
    }
return \@arr_hashes;    #return the reference to array of hash
}

################################################################################
# Subroutine Name: event_delete()
# Objective:       Delete events.
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {EventId=>"<event-id>"}    or
#                  {EventIdType=>"<event-ids>"}
################################################################################
sub event_delete {
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure

    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "EventId" ) ) {
            $modHash->{EventId} = $value;
        }
    elsif ( !( $key cmp "EventIdType" ) ) {
        $modHash->{EventIdList}{EventIdType} = $value;
        }
    else {
            $Log->comment("INVALID INPUT PASSED!");     #Exception Handling TBD!
        }
    }
    
    #printing out the data structure to be passed in APIcall
    $Log->comment(Dumper($modHash) );

    #access Web Service
    $result = $interface->EventDelete($modHash);
    $Log->comment($result);
   
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else {
        if ($modHash->{"EventId"})  {
            my $attrib = $result->get_EventDeleted();
            $hash->{EventDeleted} = "$attrib";
            $Log->comment("here is the hash:   $hash->{EventDeleted}");
            push (@arr_hashes,$hash);
        }
        elsif ($modHash->{EventIdList}{EventIdType})  {
            
            #form hash element of the array whose reference is to be returned
            
            my $attrib = $result->get_EventsDeletedList();
            my $attrib1 = $attrib->get_EventActionInfo();
            my $attrib2 = $attrib1->get_EventId();
            $hash->{EventId} = "$attrib2";
            $attrib2 = $attrib1->get_Timestamp();
            $hash->{Timestamp} = "$attrib2";
            $attrib2 = $attrib1->get_ErrorCode();
            $hash->{ErrorCode} = "$attrib2";
            $attrib2 = $attrib1->get_ErrorMessage();
            $hash->{ErrorMessage} = "$attrib2";
            push (@arr_hashes,$hash);
        }
    }
return \@arr_hashes;    #return the reference to array of hash
}  
    
################################################################################
# Subroutine Name: host_add()
# Objective:       Add new managed host to the DFM.
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {HostNameOrId =>  "<host-name>",}
################################################################################
sub host_add { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure

    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "AddNetwork" ) ) {
            $modHash->{AddNetwork} = $value;
        }
        elsif ( !( $key cmp "AdminPort" ) ) {
            $modHash->{AdminPort} = $value;
        }
        elsif ( !( $key cmp "BlockCmode" ) ) {
            $modHash->{BlockCmode} = $value;
        }
        elsif ( !( $key cmp "HostNameOrId" ) ) {
            $modHash->{HostNameOrId} = $value;
        }
        else {
            $Log->comment("INVALID INPUT PASSED!");     #Exception Handling TBD!
        }
    }

    #printing out the data structure to be passed in APIcall
    $Log->comment(Dumper($modHash) );

    #access Web Service
    $result = $interface->HostAdd($modHash);
    $Log->comment($result);
    
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else {  #form hash element of the array whose reference is to be returned
        
        my $attrib = $result->get_HostAddress();
        if ($attrib) {
            $Log->comment("Host already exists in DFM db");
            #$hash->{status=>"Host already exists in DFM db"};
            last;
        }
        $hash->{HostAddress} = "$attrib";
        $attrib = $result->get_AgentId();
        $hash->{AgentId} = "$attrib";
        $attrib = $result->get_ApplianceId();
        $hash->{ApplianceId} = "$attrib";
    }
push (@arr_hashes,$hash);
return \@arr_hashes;    #return the reference to array of hash
}

################################################################################
# Subroutine Name: host_modify()
# Objective:       Modify attributes  of a host stored in DFM database.
#                  Achieves functionality of 'dfm host set' CLI
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  Any one option to be set along with
#                  {HostNameOrId=>"<host-name>"}
################################################################################
sub host_modify { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure

    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "AdminPort" ) ) {
            $modHash->{HostModifyInfo}{AdminPort} = $value;
        }
        elsif ( !( $key cmp "HostAddress" ) ) {
            $modHash->{HostModifyInfo}{HostAddress} = $value;
        }
        elsif ( !( $key cmp "HostDescription" ) ) {
            $modHash->{HostModifyInfo}{HostDescription} = $value;
        }
        elsif ( !( $key cmp "IsDpIgnored" ) ) {
            $modHash->{HostModifyInfo}{HostDpModifyInfo}{IsDpIgnored} = $value;
        }
        elsif ( !( $key cmp "SnapvaultMaxBackupThreads" ) ) {
            $modHash->{HostModifyInfo}{HostDpModifyInfo}{SnapvaultMaxBackupThreads} = $value;
        }
        elsif ( !( $key cmp "HostName" ) ) {
            $modHash->{HostModifyInfo}{HostName} = $value;
        }
        elsif ( !( $key cmp "HostPassword" ) ) {
            $modHash->{HostModifyInfo}{HostPassword} = $value;
        }
        elsif ( !( $key cmp "HostUsername" ) ) {
            $modHash->{HostModifyInfo}{HostUsername} = $value;
        }
        elsif ( !( $key cmp "NdmpPassword" ) ) {
            $modHash->{HostModifyInfo}{NdmpCredentials}{NdmpPassword} = $value;
        }
        elsif ( !( $key cmp "NdmpPort" ) ) {
            $modHash->{HostModifyInfo}{NdmpCredentials}{NdmpPort} = $value;
        }
        elsif ( !( $key cmp "NdmpUsername" ) ) {
            $modHash->{HostModifyInfo}{NdmpCredentials}{NdmpUsername} = $value;
        }
        elsif ( !( $key cmp "PerfAdvisorTransport" ) ) {
            $modHash->{HostModifyInfo}{PerfAdvisorTransport} = $value;
        }
        elsif ( !( $key cmp "UseHostsEquiv" ) ) {
            $modHash->{HostModifyInfo}{UseHostsEquiv} = $value;
        }
        elsif ( !( $key cmp "HostNameOrId" ) ) {
            $modHash->{HostNameOrId} = $value;
        }
        else {
            $Log->comment("INVALID INPUT PASSED!");     #Exception Handling TBD!
        }
    }

    #printing out the data structure to be passed in APIcall
    $Log->comment(Dumper($modHash) );

    #access Web Service
    $result = $interface->HostModify($modHash);
    $Log->comment($result);
    
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    else {
        return[{status=>"SUCCESS"}];
    }
}
    

################################################################################
# Subroutine Name: host_get_attribute()
# Objective:       Retrieve all of the attributes associated with a host
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {EmailAddress=>"<email-id>"}
################################################################################
sub host_get_attribute { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #forming the data structure not reqd
    $modHash = $inputFields;
    
    #printing out the data structure to be passed in APIcall
    $Log->comment(Dumper($modHash) );

    #access Web Service
    $result = $interface->HostGetAttributes($modHash);
    $Log->comment($result);
    
    #error handling
    if (not $result) {
        my $e = $result->get_faultstring();
        #$Log->comment("FAULTSTRING: $e");
        my $error = "Error calling SOAP method:" . _get_fault_string($e);
	#$self->_clear_err();
	#$self->_set_err($error);
    }
    #if (not $result) { return _check_for_error($result) ; }
    {
        my $element = "Address";
        #$Log->comment($element);
        my $funcName = "\$result->get_$element";
        $Log->comment($funcName);
        #my $funcRef = *$funcName;
        #$Log->comment(&$funcRef);
        my $elementUsual = $result->get_Address();
        $Log->comment("$elementUsual");
        $element = &$funcName.();
        
        #my $RESULT1;
        #eval {$RESULT1 = $result->get_Address() };
        #$Log->comment($RESULT1);
    }
}



################################################################################
# Subroutine Name: group_move()
# Objective:       Generate events in the DFM system
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {GroupNameOrId=>"<group-name>",
#                   ParentGroupNameOrId=>"group-name"}
################################################################################
sub group_move { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure not reqd
    $modHash = $inputFields;
    
    #printing out the data structure to be passed in APIcall
    $Log->comment(Dumper($modHash) );

    #access Web Service
    $result = $interface->GroupMove($modHash);
    $Log->comment($result);
    
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else {
        return[{status=>"SUCCESS"}];
    }
}

################################################################################
# Subroutine Name: dfm_object_refresh()
# Objective:       Request monitors be scheduled to run
# Input:           Hash containing all inputs; keys in accordance to wsdl file
#                  Mandatory INPUT:
#                  {EmailAddress=>"<email-id>"}
################################################################################

sub dfm_object_refresh { 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );

    #clear any previous Error set
    $self->_clear_err();
    
    #forming the data structure

    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ChildType" ) ) {
            $modHash->{ChildType} = $value;
        }
        elsif ( !( $key cmp "MonitorName" ) ) {
            $modHash->{MonitorNames}{MonitorName} = $value;
        }
        elsif ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{ObjectNameOrId} = $value;
        }
        else {
            $Log->comment("INVALID INPUT PASSED!");     #Exception Handling TBD!
        }
    }
    
    #printing out the data structure to be passed in APIcall
    $Log->comment(Dumper($modHash) );

    #access Web Service
    $result = $interface->DfmObjectRefresh($modHash);
    $Log->comment($result);
    
    #error handling
    if (not $result) {
        return $self->_check_for_error($result) ;
    }
    
    else {
        return[{status=>"SUCCESS"}];
    }
}


sub dfm_about { 
    
    #forming the data structure
    $result = $interface->DfmAbout();
    $Log->comment($result);
}


################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _check_for_error()
# Objective:       Get the fault details from message response
################################################################################

sub _check_for_error
{
    my ($self,$result) = @_;
    #$Log->comment("result OBJECT passed: $result");
    if (not $result) {
        my $errorString = $result->get_faultstring();
        #$Log->comment("FAULTSTRING: $errorString");
        my $error = "Error calling SOAP method:"._get_fault_string($errorString);
	$self->_set_err($error);
        return $result;
    }
}

################################################################################
# Subroutine Name: _get_fault_string()
# Objective:       Get the fault details from message response
################################################################################

sub _get_fault_string
{
    my($errorMessage) = @_;
    my $xmlData;
    if(($xmlData) = $errorMessage =~ m/Error.*\n.*\n.*\n.*\n(.*\n.*)/) {
        ;
    }
    else {
        return $errorMessage;
    }
    open FILE, ">", "dummy.xml";
    print FILE $xmlData;
    close FILE;
    open FILE, "<", "dummy.xml";
    # create object
    my $xml = new XML::Simple;
   
    # read XML file
    my $data = $xml->XMLin("dummy.xml");
    unlink("dummy.xml");
   
    return $data->{'SOAP-ENV:Body'}->{'SOAP-ENV:Fault'}->{'faultstring'};
}


1;

  

