#!/usr/software/bin/perl

# $Id:
#
# Copyright (c) 2010 NetApp.
# All rights reserved.
#
#########################################################
## @summary Library of functions to access UM Soap APIs to perform
##          PA (Performance Advisor) operations.
##
## @author SreeHari Cheerla (sreeharich@netapp.com)
##
## @status public
##
## @pod here
#########################################################

# Package Name
##############
package PALib;

require Exporter;

use strict;
use warnings;
use List::MoreUtils qw(uniq);

#Module Import
######################
#use lib '/usr/software/test/share/nate/latest/lib';

use MyInterfaces::DfmService::DfmPort;
use SOAP::Transport::HTTP;
use Data::Dumper;
use XML::Simple;
use NATE::Log;
use Tharn;

# Package Requirements

# Declare global variables

use vars qw(@ISA $interface);



#PALib - Functions to access UM soap APIs to perform PA (Performance Advisor) operations.

=head1 SYNOPSIS

use PALib;
my $PALibObj=new PALib(<hostname>,<username>,<password>,[port no]);

=head1 DESCRIPTION

This package offers the functions to perform operations on PA objects which internally calls
UM SOAP APIs.
While creating the PALib I<object>, I<hostname>, I<username> and  I<password> of the UM host has to be passed as arguments.
The optional parameter I<port no> refers to Soap services port no for UM. This option is used only when there is change in the
UM Soap service port no in future. So we can create PALib object as follows,

my $PALibObj = new PALib(<hostname>,<username>,<password>);

=head1 FUNCTIONS

The parameter for each function is a Hash value and all the get functions will return reference array, each value in  array is reference to hash.
More details are explained in each function as below.

=cut

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new Soap object
################################################################################
sub new
{
    my $class = shift;
    my $self = {
        _hostName => shift,
        _userName => shift,
        _password => shift,
        _port => shift,
    };

    $self->{_port} = "8488" if not defined $self->{_port};

   *SOAP::Transport::HTTP::Client::get_basic_credentials  = sub {
        return ($self->{_userName}, $self->{_password});
    };
    $interface = MyInterfaces::DfmService::DfmPort->new({proxy=>"https://$self->{_hostName}:$self->{_port}/apis/soap/v1"});
    bless($self,$class);
    return $self;
}

=head2 B<get_events(I<ObjNameOrId>)>

This function gets the events information from UM.

=over 4

=item Inputs

=item * ObjNameOrId [optional]

provide the ObjNameOrId, the values can be I<groupName , groupID , HostName , HostID>.

=item Returns

Returns a reference to an array.Values in array contains references to hashs of the events(refer examples).
Datastore attributes are I<>

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    #if we are not specifying any arguments, then pass empty string.
    $events = $PALibObj->get_events("");

    #if we are specifying ObjNameOrId
    $events = $PALibObj->get_events(<ObjNameOrId>);

    #listing the elements from the array reference, here is the example for listing events info .
    foreach (@$eventName) {
        print $_->{EventId};
        print $_->{EventOriginatingId};
        print $_->{EventSourceId};
        print $_->{EventSourceType};
        print $_->{EventType};
        print $_->{IsDeleted};
        print $_->{EventSourceName};
        print $_->{EventName};
        print $_->{EventTime};
        print $_->{EventSeverity};
        
    }

    #function will return undef if there is an error in getting datastores
    if(not $events) {
        $Test->nlog("Error has occured while retrieving datastores");
        return TCD::FAIL;
    }


=cut

################################################################################
                                                                                   
sub get_events {

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "EventSourceId" ) ) {
            $modHash->{EventSourceId} = $value;
        }
		if ( !( $key cmp "EventSeverities" ) ) {
            $modHash->{EventSeverities}{ObjStatus} = $value;
        }
        if ( !( $key cmp "StartTime" ) ) {
            $modHash->{TimeRange}{EventTimestampRange}{StartTime} = $value;
        }
        if ( !( $key cmp "EndTime" ) ) {
            $modHash->{TimeRange}{EventTimestampRange}{EndTime} = $value;
        }
    }

    $result = $interface->EventListIterStart($modHash);
	
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
        }
 
    my $tag =$result->get_Tag();
    #Get the number of events present
    my $max = $result->get_Records();
    #Get the object of EventListIterNext in $resultEvent
    my $resultEvent = $interface->EventListIterNext({Maximum => "$max", Tag => $tag});
    if (not $resultEvent) {
                my $error =  _check_for_error($resultEvent);
                return $error;
    }
    my $eventsInfo = $resultEvent->get_Events();
    my $event = $eventsInfo->get_EventInfo();
    my @eventList;
    foreach (@$event) {
        my %eventDetails;
        print $_->get_EventSourceName() . "\n";
        $eventDetails{EventId} = $_->get_EventId();
        $eventDetails{EventOriginatingId} = $_->get_EventOriginatingId();
        $eventDetails{EventSourceId} = $_->get_EventSourceId();
        $eventDetails{EventSourceType} = $_->get_EventSourceType();
        $eventDetails{EventType} = $_->get_EventType();
        $eventDetails{IsDeleted} = $_->get_IsDeleted();
        $eventDetails{EventSourceName} = $_->get_EventSourceName();
		$eventDetails{EventName} = $_->get_EventName();
        $eventDetails{EventTime} = $_->get_EventTime();
        $eventDetails{EventSeverity} = $_->get_EventSeverity();
		my $src = $_->get_EventSourceName();
		my $name = $_->get_EventName();
		my $sev = $_->get_EventSeverity();
	    push(@eventList, \%eventDetails);
    }
    $interface->EventListIterEnd({Tag => $tag});
    return \@eventList;
} #end sub get_events

=head2 B<get_hostDetails(I<ObjNameOrId>)>

This function gets the host information from UM.

=over 4

=item Inputs

=item * ObjNameOrId [optional]

provide the ObjNameOrId, the values can be I<groupName , groupID , HostName , HostID>.

=item Returns

Returns a reference to an array.Values in array contains references to hashs of the host Details(refer examples).
Datastore attributes are I<>

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)


=cut

################################################################################



sub get_hostDetails {

  my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{ObjectNameOrId} = $value;
        }
    }
    $result = $interface->HostListInfoIterStart($modHash);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
        }

    my $tag =$result->get_Tag();
    #Get the number of events present
    my $max = $result->get_Records();
    #Get the object of EventListIterNext in $resultEvent
    my $resultHosts = $interface->HostListInfoIterNext({Maximum => "$max", Tag => $tag});
    if (not $resultHosts) {
                my $error =  _check_for_error($resultHosts);
                return $resultHosts;
    }
    my @hostinfos = $resultHosts->get_Hosts()->get_HostInfo();
    my @hostList;
    foreach (@hostinfos) {
        my %hostDetail;
        $hostDetail{HostId} = $_->get_HostId();
        $hostDetail{HostAddress} = $_->get_HostAddress();
        $hostDetail{HostStatus} = $_->get_HostObjStatus();
        $hostDetail{HostName} = $_->get_HostName();
        $hostDetail{PerfStatus} = $_->get_HostPerfInfo()->get_HostPerfStatus();
        push(@hostList, \%hostDetail);
    }
    $interface->HostListInfoIterEnd({Tag => $tag});
    return \@hostList;
} #end sub get_hostDetails

## Adding function get_perf_status 
sub get_perf_status {

  my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "Objects" ) ) {
            $modHash->{Objects}{ObjNameOrId} = $value;
        }
    }
    $result = $interface->DfmObjectsGetStatus($modHash);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
        }

    my @hostinfos = $result->get_ObjectsStatus()->get_ObjStatusInfo();
    my @hostList;
    foreach (@hostinfos) {
        my %hostDetail;
        $hostDetail{ObjectName} = $_->get_ObjectNameOrId();
        $hostDetail{HostStatus} = $_->get_ObjectStatus();
        $hostDetail{PerfStatus} = $_->get_ObjectPerfStatus();
        push(@hostList, \%hostDetail);
    }
    return \@hostList;
} #end sub get_perf_status

## Adding function get_unit
sub get_unit {

  my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ObjectName" ) ) {
            $modHash->{ObjectName} = $value;
        }
    }
    $result = $interface->PerfObjectCounterListInfo($modHash);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
        }

    my @hostinfos = $result->get_Counters()->get_PerfCounter();
   $Log->log("hosts:@hostinfos");
    my @hostList;
    foreach (@hostinfos) {
        my %hostDetail;
	my $name =  $_->get_Desc();
	my $unit = $_->get_Unit();
#        $hostDetail{Name} = $_->get_Name();
 #       $hostDetail{Unit} = $_->get_Unit();
	$Log->log("Name:$name , Unit: $unit");
        push(@hostList, \%hostDetail);
    }
    return \@hostList;
} #end s
=head2 B<collect_clientstats(I<ObjNameOrId,CollectionPeriod>)>

This funtion collects the client stats information for a storage device .

=over 4

=item Inputs  

=item  * ObjNameOrId

=item  * CollectionPeriod [optional]

provide the ObjNameOrId, If the collection Period is not passing then by default it takes as 15 secs.

=item Returns

Returns a scalar value that will have the STATID of the collection (refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    #if we are .

    $result = $PALibObj->PerfCollectClientOperationStatistics($modHash);

    # Getting the value of the StatID

    $statID = $result . 	

=cut


sub collect_clientstats {
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "CollectionPeriod" ) ) {
            $modHash->{CollectionPeriod} = $value;
        }elsif ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{ObjectNameOrId} = $value;
        }
    }
    $result = $interface->PerfCollectClientOperationStatistics($modHash);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
	}
    my $statID = $result->get_StatId();
    return $statID ;
}


=head2 B<list_clientstats(I<ObjNameOrId,StatId,EndTime,LastOnly,StartTime>)>

=over 4

=item Inputs 

=item * ObjNameOrId

=item * StatId [optional]

=item * EndTime [optional]

=item * LastOnly [optional]

=item * StartTime [optional]

	provide the ObjNameOrId,StatId , EndTime , LastOnly or StartTime 

=item Returns

Returns a reference to an array.Values in array contains references to hashes of the clients stats by client IP (refer examples).
Datastore attributes are I<>

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);


     Getting the values of Stats for each client 


	my $objName = "10.72.184.160";
        my $input = {ObjectNameOrId => $objName};
        my $dn_um = $soapObj->list_clientstats($input);

        foreach (@$dn_um){
                my $id = $_->{StatId};
                my $count = $_->{count};
                $Log->comment("Count in test = $count");
                for (my $i=1;$i <= $count;$i++ ){
                           my $addr = $_->{"ClientIp".$i};
                           my $cifs = $_->{"CifsOther".$i} + $_->{"CifsRead".$i} + $_->{"CifsWrite".$i};
                           my $nfs = $_->{"NfsOther".$i} + $_->{"NfsRead".$i}+ $_->{"NfsWrite".$i};
                           $Log->comment("ID = $id IP = $addr NFS = $nfs  CIFS = $cifs\n");
                }

        }
	
=cut

sub list_clientstats{
 
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "StatId" ) ) {
            $modHash->{StatId} = $value;
        } elsif ( !( $key cmp "ObjectNameOrId" ) ) {
             $modHash->{ObjectNameOrId} = $value;
        } elsif ( !( $key cmp "EndTime" ) ) {
            $modHash->{EndTime} = $value;
        } elsif ( !( $key cmp "LastOnly" ) ) {
            $modHash->{LastOnly} = $value;
        } elsif ( !( $key cmp "StartTime" ) ) {
            $modHash->{StartTime} = $value;
        }
    }
    $result = $interface->PerfClientStatsListInfoIterStart($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
    }

    my $tag =$result->get_Tag();
    #Get the number of client stats present
    my $max = $result->get_Records();
    #Get the object of PerfClientStatsListInfoIterNext in $resultEvent
    my $resultEvent = $interface->PerfClientStatsListInfoIterNext({Maximum => "$max", Tag => $tag});
    if (not $resultEvent) {
                my $error =  _check_for_error($resultEvent);
                return $resultEvent;
    }

    my $hostClients = $resultEvent->get_HostsClientStats()->get_HostClientStats();

    my @clientStatList;

    foreach (@$hostClients) {
        my %clientStatDetails;
	my $count = 1;
        $clientStatDetails{CollectionPeriod} = $_->get_CollectionPeriod();
        $clientStatDetails{ObjectId} = $_->get_ObjectId();
        $clientStatDetails{ObjectName} = $_->get_ObjectName();
        $clientStatDetails{StartTime} = $_->get_StartTime();
        $clientStatDetails{StatId} = $_->get_StatId();
        my $stats = $_->get_Stats()->get_ClientStats();
        foreach my $client (@$stats){
	    my $ip = $client->get_ClientNameOrIp();
  	    my $cifsOther = $client->get_CifsOpStats()->get_OtherOperations();
	    my $cifsRead = $client->get_CifsOpStats()->get_ReadOperations();
	    my $cifsWrite = $client->get_CifsOpStats()->get_WriteOperations();
	    my $nfsOther = $client->get_NfsOpStats()->get_OtherOperations();
	    my $nfsRead = $client->get_NfsOpStats()->get_ReadOperations();
	    my $nfsWrite = $client->get_NfsOpStats()->get_WriteOperations();

	    $clientStatDetails{"ClientIp".$count} = $ip;
	    $clientStatDetails{"CifsOther".$count} = $cifsOther;
	    $clientStatDetails{"CifsRead".$count} = $cifsRead;
	    $clientStatDetails{"CifsWrite".$count} = $cifsWrite;
	    $clientStatDetails{"NfsOther".$count} = $nfsOther;
	    $clientStatDetails{"NfsRead".$count} = $nfsRead;
	    $clientStatDetails{"NfsWrite".$count} = $nfsWrite; 
	    $count = $count + 1;
	}
	$clientStatDetails{count}=$count-1;
	push(@clientStatList, \%clientStatDetails);
    }
    $interface->PerfClientStatsListInfoIterEnd({Tag => $tag});
    return \@clientStatList;
}

=head2 B<purge_clientstat(I<ObjNameOrId,StatId,EndTime,StartTime>)>

=item Inputs 

=item * ObjNameOrId [optional]

=item * StatId [optional]

=item * StartTime [optional]

=item * EndTime [optional]

	The input to this function is a hash  . Either ObjactNameOrId or CollectionPeriod is Mandatory .
	If we are not providing the StatID then it removes all the clientStat info for the Object .

=item Returns

Returns a scalar value that will have the Number of Deleted Client Stats information (refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    if we are .

    $result = $PALibObj->PerfClientStatsPurge($modHash);

     Getting the values of the count

	my $objName = "10.72.184.160";
        my $input = {ObjectNameOrId => $objName};
        my $count =  $soapObj->purge_clientStats();
        $Log->comment( "COUNT = $count");

=cut

sub purge_clientstats{
	
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "StatId" ) ) {
            $modHash->{StatId} = $value;
        } elsif ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{ObjectNameOrId} = $value;
        } elsif ( !( $key cmp "EndTime" ) ) {
            $modHash->{EndTime} = $value;
        } elsif ( !( $key cmp "StartTime" ) ) {
            $modHash->{StartTime} = $value;
        }
    }
    $result = $interface->PerfClientStatsPurge($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
    }
    my $count = $result->get_DeletedStatsCount();
    return $count;	 	
}

=head2 B<threshold_create(I<ObjNameOrId,ThresholdInfo2>)>

=item Inputs

=item * ObjNameOrId 

=item * ThresholdInfo2


        The input to this function is a hash with all the values of ThresholdInfo2 .

=item Returns

Returns a scalar value that will have the ThresholdId (refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    if we are .

    $result = $PALibObj->PerfThresholdCreate2($modHash);

     Getting the values of the count

        my $ObjName = "10.72.184.160" ;
        my $ThresholdEventName = "sreehari-test-event";
        my $ThresholdInterval = 60;
        my $ThresholdType = "Upper";
        my $ThresholdUnit = "percent";
        my $ThresholdValue = 70;
        my $thresholdObjectType = "system";
        my $CounterName = "cpu_busy";

        my $input = {ObjectNameOrId => $ObjName,ThresholdEventName=>$ThresholdEventName,ThresholdInterval=>$ThresholdInterval,ThresholdType=>$ThresholdType,ThresholdUnit=>$ThresholdUnit,ThresholdValue=>$ThresholdValue,ThresholdObjectType=>$thresholdObjectType,CounterName=>$CounterName};

        my $thresholdID =  $soapObj->threshold_create($input);

        $Log->comment("Thireshold Id IS : $thresholdID");

=cut


sub threshold_create{

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{ObjectNameOrId} = $value;
        }
	if ( !( $key cmp "Id" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Id} = $value;
        }

	if ( !( $key cmp "MetricResult" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{MetricResult} = $value;
        }

	if ( !( $key cmp "Period" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Period} = $value;
        }

	if ( !( $key cmp "Slide" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Slide} = $value;
        }

	if ( !( $key cmp "Type" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Type} = $value;
        }

	if ( !( $key cmp "CounterDescription" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{CounterDescription} = $value;
        }

	if ( !( $key cmp "CounterName" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{CounterName} = $value;
        }

	if ( !( $key cmp "IsCalculatedStats" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{IsCalculatedStats} = $value;
        }

	if ( !( $key cmp "LabelNames" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{LabelNames} = $value;
        }

	if ( !( $key cmp "ThresholdObjectType" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{ObjectType} = $value;
        }
	#This one is perfLables LableNamames which is different to the above one .
	if ( !( $key cmp "perfLabelNames" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{PerfLabels}{PerfLabel}{LabelNames} = $value;
        }

	if ( !( $key cmp "TimeConsolidationMethod" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{TimeConsolidationMethod} = $value;
        }

	if ( !( $key cmp "ThresholdType" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdType} = $value;
        }

	if ( !( $key cmp "ThresholdUnit" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdUnit} = $value;
        }

	if ( !( $key cmp "ThresholdValue" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdValue} = $value;
        }

	if ( !( $key cmp "IsEnabled" ) ) {
            $modHash->{ThresholdInfo2}{IsEnabled} = $value;
        }

	if ( !( $key cmp "IsIndirect" ) ) {
            $modHash->{ThresholdInfo2}{IsIndirect} = $value;
        }

	if ( !( $key cmp "ObjectFullName" ) ) {
            $modHash->{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectFullName} = $value;
        }

	if ( !( $key cmp "ObjectId" ) ) {
            $modHash->{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectId} = $value;
	}

	if ( !( $key cmp "ObjectName" ) ) {
            $modHash->{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectName} = $value;
	}

	if ( !( $key cmp "ObjectType" ) ) {
            $modHash->{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectType} = $value;
	}

	if ( !( $key cmp "Name" ) ) {
            $modHash->{ThresholdInfo2}{ResourcePropertyConditions}{ResourcePropertyCondition}{Name} = $value;
        }

	if ( !( $key cmp "Value" ) ) {
            $modHash->{ThresholdInfo2}{ResourcePropertyConditions}{ResourcePropertyCondition}{Value} = $value;
	}

	if ( !( $key cmp "TemplateId" ) ) {
            $modHash->{ThresholdInfo2}{TemplateId} = $value;
        }

	if ( !( $key cmp "TemplateName" ) ) {
            $modHash->{ThresholdInfo2}{TemplateName} = $value;
	}

	if ( !( $key cmp "ThresholdEventName" ) ) {
            $modHash->{ThresholdInfo2}{ThresholdEventName} = $value;
	}

	if ( !( $key cmp "ThresholdId" ) ) {
            $modHash->{ThresholdInfo2}{ThresholdId} = $value;
	}

	if ( !( $key cmp "ThresholdInterval" ) ) {
            $modHash->{ThresholdInfo2}{ThresholdInterval} = $value;
	}
    }
    $result = $interface->PerfThresholdCreate2($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
    }
    my $thresholdId = $result->get_ThresholdId();
    return $thresholdId;
}

=head2 B<get_thresholdList(I<ObjNameOrId,ThresholdId>)>

=item Inputs

=item * ObjNameOrId

=item * ThresholdId


        The input to this function is a hash with ObjNameorId or ThresholdId  .

=item Returns

Returns a array of hashes with the details of a Threshold for a particular Object or a particular threshold (refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    if we are .

    $result = $PALibObj->PerfThresholdListInfo2IterNext($modHash);

     Getting the values  

        my $objName = "10.72.184.160";
        my $input = {ObjectNameOrId => $objName};

        my $ds_um = $soapObj->get_thresholdList($input);
        foreach(@$ds_um) {
           my $ds_name = $_->{ObjectType};
           my $eventID = $_->{ThresholdId};
           my $counter = $_->{CounterName};
           $Log->comment("Obj Type = $ds_name : ID = $eventID : Counter = $counter\n");
         }

=cut

sub get_thresholdList {

    my ( $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ThresholdId" ) ) {
            $modHash->{ThresholdId} = $value;
			$Log->log("Threshold id :  $value");
        } elsif ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{ObjectNameOrId} = $value;
        }
    }	
    $result = $interface->PerfThresholdListInfo2IterStart($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
    }
    my $tag =$result->get_Tag();
    #Get the number of thresholds present
    my $max = $result->get_Records();
    #Get the object of thresholdInfo2 in $resultEvent
    my $resultEvent = $interface->PerfThresholdListInfo2IterNext({Maximum => "$max", Tag => $tag});
   # print $resultEvent;
    if (not $resultEvent) {
                my $error =  _check_for_error($resultEvent);
                return $resultEvent;
    }
    my $thresholds = $resultEvent->get_AttributesList()->get_ThresholdInfo2();
    $Log->comment("Result is : @$thresholds");
    my @thresholdinfos;
    foreach (@$thresholds) {
	my %thresholdDetails;
     
		if( $_->get_CounterThresholds()->get_CounterThresholdInfo()->get_PerfObjectCounter() ) {
             $thresholdDetails{CounterDescription} =$_->get_CounterThresholds()->get_CounterThresholdInfo()->get_PerfObjectCounter()->get_CounterDescription(); 
             $thresholdDetails{CounterName} = $_->get_CounterThresholds()->get_CounterThresholdInfo()->get_PerfObjectCounter()->get_CounterName();
             $thresholdDetails{CounterObjectType} = $_->get_CounterThresholds()->get_CounterThresholdInfo()->get_PerfObjectCounter()->get_ObjectType();
		}
        $thresholdDetails{ThresholdType} = $_->get_CounterThresholds()->get_CounterThresholdInfo()->get_ThresholdType();
        $thresholdDetails{ThresholdUnit} = $_->get_CounterThresholds()->get_CounterThresholdInfo()->get_ThresholdUnit();
        $thresholdDetails{ThresholdValue} = $_->get_CounterThresholds()->get_CounterThresholdInfo()->get_ThresholdValue();
        $thresholdDetails{IsEnabled} = $_->get_IsEnabled();
        $thresholdDetails{IsIndirect} = $_->get_IsIndirect();
	$thresholdDetails{ThresholdId}=$_->get_ThresholdId();
	if ($_->get_ObjectsInfo()){
	$thresholdDetails{ObjectFullName} = $_->get_ObjectsInfo()->get_ObjectInfo()->get_ObjectFullName();
        $thresholdDetails{ObjectId} = $_->get_ObjectsInfo()->get_ObjectInfo()->get_ObjectId();
        $thresholdDetails{ObjectType} = $_->get_ObjectsInfo()->get_ObjectInfo()->get_ObjectType();
	} else{
	}      
        $thresholdDetails{ThresholdEventName} = $_->get_ThresholdEventName();
        $thresholdDetails{ThresholdInterval} = $_->get_ThresholdInterval();
        push(@thresholdinfos, \%thresholdDetails);
    }
    $interface->PerfThresholdListInfo2IterEnd({Tag => $tag});
    return \@thresholdinfos;
}

=head2 B<threshold_modify(I<ObjNameOrId,ThresholdInfo2>)>

=item Inputs

=item * ObjNameOrId

=item * ThresholdInfo2


        The input to this function is a hash with all the values of ThresholdInfo2 .

=item Returns

Returns a empty XML file if thresholds modification successfull(refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    if we are .

    $result = $PALibObj->PerfThresholdModify2($modHash);

        my $ObjName = "10.72.184.160" ;
        my $ThresholdEventName = "dabba";
        my $ThresholdInterval = 60;
        my $ThresholdType = "Upper";
        my $ThresholdUnit = "percent";
        my $ThresholdValue = 70;
        my $thresholdObjectType = "system";
        my $CounterName = "cpu_busy";
        my $thresholdId = 7;
#       my $input = {ObjectNameOrId => $ObjName,ThresholdEventName=>$ThresholdEventName,ThresholdInterval=>$ThresholdInterval,ThresholdType=>$ThresholdType,ThresholdUnit=>$ThresholdUnit,ThresholdValue=>$ThresholdValue,ThresholdObjectType=>$thresholdObjectType,CounterName=>$CounterName,ThresholdId=>$thresholdId};

        my $input = {ThresholdEventName=>$ThresholdEventName,ThresholdId=>$thresholdId};
        my $thresholdID =  $soapObj->threshold_modify($input);

=cut


sub threshold_modify
{
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    my $thresholds = get_thresholdList($inputFields);
    #forming the data structure
    foreach my $retValue (@$thresholds){
        while ( ( my $key, my $value ) = each %$retValue ) {

	    if ( !( $key cmp "CounterObjectType" ) ) {
                $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{ObjectType} = $value;
            }

            if ( !( $key cmp "CounterDescription" ) ) {
                $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{CounterDescription} = $value;
            }

            if ( !( $key cmp "CounterName" ) ) {
                $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{CounterName} = $value;
            }

            if ( !( $key cmp "ThresholdType" ) ) {
                $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdType} = $value;
            }

            if ( !( $key cmp "ThresholdUnit" ) ) {
                $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdUnit} = $value;
            }

            if ( !( $key cmp "ThresholdValue" ) ) {
                $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdValue} = $value;
            }

            if ( !( $key cmp "IsEnabled" ) ) {
                $modHash->{ThresholdInfo2}{IsEnabled} = $value;
            }
	    if ( !( $key cmp "IsIndirect" ) ) {
                $modHash->{ThresholdInfo2}{IsIndirect} = $value;
            }

            if ( !( $key cmp "ObjectFullName" ) ) {
                $modHash->{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectFullName} = $value;
            }

            if ( !( $key cmp "ObjectId" ) ) {
                $modHash->{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectId} = $value;
            }

            if ( !( $key cmp "ObjectType" ) ) {
                $modHash->{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectType} = $value;
            }

            if ( !( $key cmp "ThresholdEventName" ) ) {
                $modHash->{ThresholdInfo2}{ThresholdEventName} = $value;
            }
            if ( !( $key cmp "ThresholdInterval" ) ) {
              $modHash->{ThresholdInfo2}{ThresholdInterval} = $value;
            }
     }
}
    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{ObjectNameOrId} = $value;
 	}       
        if ( !( $key cmp "Id" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Id} = $value;
        }

        if ( !( $key cmp "MetricResult" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{MetricResult} = $value;
        }

        if ( !( $key cmp "Period" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Period} = $value;
        }


        if ( !( $key cmp "Slide" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Slide} = $value;
        }

        if ( !( $key cmp "Type" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Type} = $value;
        }

        if ( !( $key cmp "CounterDescription" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{CounterDescription} = $value;
        }

        if ( !( $key cmp "CounterName" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{CounterName} = $value;
        }

        if ( !( $key cmp "IsCalculatedStats" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{IsCalculatedStats} = $value;
        }

        if ( !( $key cmp "LabelNames" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{LabelNames} = $value;
        }
        if ( !( $key cmp "ThresholdObjectType" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{ObjectType} = $value;
        }
        #This one is perfLables LableNamames which is different to the above one .
        if ( !( $key cmp "perfLabelNames" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{PerfLabels}{PerfLabel}{LabelNames} = $value;
        }

        if ( !( $key cmp "TimeConsolidationMethod" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{TimeConsolidationMethod} = $value;
        }

        if ( !( $key cmp "ThresholdType" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdType} = $value;
        }

        if ( !( $key cmp "ThresholdUnit" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdUnit} = $value;
        }

        if ( !( $key cmp "ThresholdValue" ) ) {
            $modHash->{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdValue} = $value;
        }

        if ( !( $key cmp "IsEnabled" ) ) {
            $modHash->{ThresholdInfo2}{IsEnabled} = $value;
        }

        if ( !( $key cmp "IsIndirect" ) ) {
            $modHash->{ThresholdInfo2}{IsIndirect} = $value;
        }

        if ( !( $key cmp "ObjectFullName" ) ) {
            $modHash->{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectFullName} = $value;
        }

        if ( !( $key cmp "ObjectId" ) ) {
            $modHash->{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectId} = $value;
        }

        if ( !( $key cmp "ObjectName" ) ) {
            $modHash->{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectName} = $value;
        }

        if ( !( $key cmp "ObjectType" ) ) {
            $modHash->{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectType} = $value;
        }

        if ( !( $key cmp "Name" ) ) {
            $modHash->{ThresholdInfo2}{ResourcePropertyConditions}{ResourcePropertyCondition}{Name} = $value;
        }

        if ( !( $key cmp "Value" ) ) {
            $modHash->{ThresholdInfo2}{ResourcePropertyConditions}{ResourcePropertyCondition}{Value} = $value;
        }

        if ( !( $key cmp "TemplateId" ) ) {
            $modHash->{ThresholdInfo2}{TemplateId} = $value;
        }

        if ( !( $key cmp "TemplateName" ) ) {
            $modHash->{ThresholdInfo2}{TemplateName} = $value;
        }

        if ( !( $key cmp "ThresholdEventName" ) ) {
            $modHash->{ThresholdInfo2}{ThresholdEventName} = $value;
        }

        if ( !( $key cmp "ThresholdId" ) ) {
            $modHash->{ThresholdInfo2}{ThresholdId} = $value;
        }

        if ( !( $key cmp "ThresholdInterval" ) ) {
            $modHash->{ThresholdInfo2}{ThresholdInterval} = $value;
        }


    }
    $result = $interface->PerfThresholdModify2($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                $Log->comment("Error is : $error");
                return $result;
    }
    return $result;
}


=head2 B<threshold_delete(I<ThresholdId>)>

=item Inputs

=item * ThresholdId


        The input to this function is a hash with ThresholdId .

=item Returns

Returns Null value if the ThresholdDelete is successful (refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    if we are .

    $result = $PALibObj->PerfThresholdDelete($modHash);

     Getting the values of the count

        my $thresholdId = 4;
        my $input = {ThresholdId=>$thresholdId};
        my $thresholdID =  $soapObj->threshold_delete($input);

=cut

sub threshold_delete {

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ThresholdId" ) ) {
            $modHash->{ThresholdId} = $value;
        }
    }

    $result = $interface->PerfThresholdDelete($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
    }
    return $result;
}


=head2 B<thresholdTemplate_create(I<ObjNameOrId,ThresholdInfo2,TemplateName>)>

=item Inputs

=item * ObjNameOrId

=item * ThresholdInfo2

=item * TemplateName

        The input to this function is a hash .

=item Returns

Returns a scalar value that will have the ThresholdTemplateId (refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    if we are .

    $result = $PALibObj->PerfThresholdTemplateCreate($modHash);

     Getting the values of the count

        my $ObjectName = "f3050-184-160.lab.eng.btc.netapp.in" ;
        my $ThresholdEventName = "sreehari-test-event";
        my $ThresholdInterval = 60;
        my $ThresholdType = "Upper";
        my $ThresholdUnit = "percent";
        my $ThresholdValue = 70;
        my $thresholdObjectType = "system";
        my $CounterName = "cpu_busy";
        my $TemplateName = "Hari-Templatetestsree";
        my $ObjectId = 80 ;
        my $input = {ObjectFullName => $ObjectName,ThresholdEventName=>$ThresholdEventName,ThresholdInterval=>$ThresholdInterval,ThresholdType=>$ThresholdType,ThresholdUnit=>$ThresholdUnit,ThresholdValue=>$ThresholdValue,ThresholdObjectType=>$thresholdObjectType,CounterName=>$CounterName,TemplateName=>$TemplateName,ObjectId=>$ObjectId};

        my $templateID =  $soapObj->thresholdTemplate_create($input);

        $Log->comment("Threshold template Id IS : $templateID");

=cut

sub thresholdTemplate_create {

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "Id" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Id} = $value;
        }

        if ( !( $key cmp "MetricResult" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{MetricResult} = $value;
        }

        if ( !( $key cmp "Period" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Period} = $value;
        }


        if ( !( $key cmp "Slide" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Slide} = $value;
        }

        if ( !( $key cmp "Type" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Type} = $value;
        }

        if ( !( $key cmp "CounterDescription" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{CounterDescription} = $value;
        }


        if ( !( $key cmp "CounterName" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{CounterName} = $value;
        }
        if ( !( $key cmp "IsCalculatedStats" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{IsCalculatedStats} = $value;
        }

        if ( !( $key cmp "LabelNames" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{LabelNames} = $value;
        }


        if ( !( $key cmp "ThresholdObjectType" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{ObjectType} = $value;
        }
        #This one is perfLables LableNamames which is different to the above one .
        if ( !( $key cmp "perfLabelNames" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{PerfLabels}{PerfLabel}{LabelNames} = $value;
        }

        if ( !( $key cmp "TimeConsolidationMethod" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{TimeConsolidationMethod} = $value;
        }

        if ( !( $key cmp "ThresholdType" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdType} = $value;
        }

        if ( !( $key cmp "ThresholdUnit" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdUnit} = $value;
        }

        if ( !( $key cmp "ThresholdValue" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdValue} = $value;
        }

        if ( !( $key cmp "IsEnabled" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{IsEnabled} = $value;
        }

        if ( !( $key cmp "IsIndirect" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{IsIndirect} = $value;
        }

        if ( !( $key cmp "ObjectFullName" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectFullName} = $value;
        }

        if ( !( $key cmp "ObjectId" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectId} = $value;
        }

        if ( !( $key cmp "ObjectName" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectName} = $value;
        }

        if ( !( $key cmp "ObjectType" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectType} = $value;
        }

        if ( !( $key cmp "Name" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ResourcePropertyConditions}{ResourcePropertyCondition}{Name} = $value;
        }

        if ( !( $key cmp "Value" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ResourcePropertyConditions}{ResourcePropertyCondition}{Value} = $value;
        }


        if ( !( $key cmp "ThresholdTemplateId" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{TemplateId} = $value;
        }

        if ( !( $key cmp "ThresholdTemplateName" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{TemplateName} = $value;
        }

        if ( !( $key cmp "ThresholdEventName" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ThresholdEventName} = $value;
        }

        if ( !( $key cmp "ThresholdId" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ThresholdId} = $value;
        }

        if ( !( $key cmp "ThresholdInterval" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ThresholdInterval} = $value;
        }
	if ( !( $key cmp "TemplateDescription" ) ) {
            $modHash->{TemplateDescription} = $value;
        }
	if ( !( $key cmp "TemplateName" ) ) {
            $modHash->{TemplateName} = $value;
        }

    }

    $result = $interface->PerfThresholdTemplateCreate($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                $Log->comment("Error is : $error");
                return $error;
    }
    my $templateId = $result->get_TemplateId();
    return $templateId; 
}



=head2 B<get_thresholdTemplateList(I<TemplateNameorId>)>

=item Inputs

=item * TemplateNameorId


        The input to this function is a hash with TemplateNameorId  .

=item Returns

Returns a array of hashes with the details of a Thresholds which are there in a Template (refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    if we are .

    $result = $PALibObj->PerfThresholdTemplateListInfoIterNext($modHash);

     Getting the values

        my $objName = "10.72.184.160";
        my $isinclude = "true";
        my $host = "true";
        my $input = {IncludeThresholds=>$isinclude,IncludeObjects=>$host};

         my $ds_um = $soapObj->get_thresholdTemplateList($input);
         foreach(@$ds_um) {
           my $ds_name = $_->{TemplateName};
           my $eventID = $_->{TemplateId};
           my $counter = $_->{CounterName};
           my $event = $_->{ThresholdEventName};
           my $id = $_->{ObjectId};
           $Log->comment("Obj Type = $ds_name : ID = $eventID Counter = $counter Event = $event Object = $id\n");
        }

=cut

sub get_thresholdTemplateList
{
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call
    my $includeObjects = 0;
    my $includeThresholds = 0;
    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "TemplateNameOrId" ) ) {
            $modHash->{TemplateNameOrId} = $value;
        }
        if ( !( $key cmp "IncludeObjects" ) ) {
            $modHash->{IncludeObjects} = $value;
	    $includeObjects = 1;
        }
        if ( !( $key cmp "IncludeThresholds" ) ) {
            $modHash->{IncludeThresholds} = $value;
	    $includeThresholds = 1;
        }
    }
    $result = $interface->PerfThresholdTemplateListInfoIterStart($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $error;
        }

    my $tag =$result->get_Tag();
    #Get the number of threshold templates present
    my $max = $result->get_Records();
    #Get the object of thresholdTemplateInfo2 in $resultEvent
    my $resultEvent = $interface->PerfThresholdTemplateListInfoIterNext({Maximum => "$max", Tag => $tag});
   # print $resultEvent;
    if (not $resultEvent) {
                my $error =  _check_for_error($resultEvent);
                return $error;
        }
    my $thresholdTemplates = $resultEvent->get_ThresholdTemplates()->get_ThresholdTemplateInfo();
    my @thresholdTemplateInfos;

    foreach (@$thresholdTemplates) {
        my %thresholdTemplateDetails;
        $thresholdTemplateDetails{TemplateIsEnabled} =$_->get_IsEnabled();
	
        $thresholdTemplateDetails{TemplateDescription} =$_->get_TemplateDescription();
        $thresholdTemplateDetails{TemplateId} =$_->get_TemplateId();
        $thresholdTemplateDetails{TemplateName} =$_->get_TemplateName();
        if ( $includeThresholds eq 1 and $_ =~ m/ThresholdInfo2/){
        $thresholdTemplateDetails{CounterDescription} =$_->get_MemberThresholds()->get_ThresholdInfo2()->get_CounterThresholds()->get_CounterThresholdInfo()->get_PerfObjectCounter()->get_CounterDescription();

        $thresholdTemplateDetails{CounterName} =$_->get_MemberThresholds()->get_ThresholdInfo2()->get_CounterThresholds()->get_CounterThresholdInfo()->get_PerfObjectCounter()->get_CounterName();

        $thresholdTemplateDetails{ThresholdObjectType} =$_->get_MemberThresholds()->get_ThresholdInfo2()->get_CounterThresholds()->get_CounterThresholdInfo()->get_PerfObjectCounter()->get_ObjectType();

        $thresholdTemplateDetails{ThresholdType} =$_->get_MemberThresholds()->get_ThresholdInfo2()->get_CounterThresholds()->get_CounterThresholdInfo()->get_ThresholdType();
        $thresholdTemplateDetails{ThresholdUnit} =$_->get_MemberThresholds()->get_ThresholdInfo2()->get_CounterThresholds()->get_CounterThresholdInfo()->get_ThresholdUnit();
        $thresholdTemplateDetails{ThresholdValue} =$_->get_MemberThresholds()->get_ThresholdInfo2()->get_CounterThresholds()->get_CounterThresholdInfo()->get_ThresholdValue();

        $thresholdTemplateDetails{ThresholdIsEnabled} =$_->get_MemberThresholds()->get_ThresholdInfo2()->get_IsEnabled();


        $thresholdTemplateDetails{ThresholdEventName} = $_->get_MemberThresholds()->get_ThresholdInfo2()->get_ThresholdEventName();
        $thresholdTemplateDetails{ThresholdId} = $_->get_MemberThresholds()->get_ThresholdInfo2()->get_ThresholdId();
        $thresholdTemplateDetails{ThresholdInterval} = $_->get_MemberThresholds()->get_ThresholdInfo2()->get_ThresholdInterval();
        }
	
	if ($includeObjects eq 1 and $_ =~ m/AppliedObjects/){	

        $thresholdTemplateDetails{ObjectFullName} = $_->get_AppliedObjects()->get_ObjectInfo()->get_ObjectFullName();
        $thresholdTemplateDetails{ObjectId} = $_->get_AppliedObjects()->get_ObjectInfo()->get_ObjectId();
	}
	elsif($includeObjects eq 0 or ! ($_ =~ m/AppliedObjects/)){
	$thresholdTemplateDetails{ObjectFullName} = "NULL";
        $thresholdTemplateDetails{ObjectId} = "NULL";
	}

        push(@thresholdTemplateInfos, \%thresholdTemplateDetails);

    }
    $interface->PerfThresholdTemplateListInfoIterEnd({Tag => $tag});
    return \@thresholdTemplateInfos;

}


=head2 B<thresholdTemplate_delete(I<TemplateNameOrId>)>

=item Inputs

=item * TemplateNameOrId


        The input to this function is a hash with TemplateNameOrId .

=item Returns

Returns the hash with the TemplateId and TemplateName (refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    if we are .

    $result = $PALibObj->PerfThresholdTemplateDelete($modHash);

     Getting the values of the count

        my $objName = 6;
        my $input = {TemplateNameOrId=>$objName};

         my $ds_um = $soapObj->thresholdTemplate_delete($input);

        my $id =  $ds_um->{TemplateId};
        my $name = $ds_um->{TemplateName};
        $Log->comment ("id = $id Name = $name\n");

=cut

sub thresholdTemplate_delete {
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "TemplateNameOrId" ) ) {
            $modHash->{TemplateNameOrId} = $value;
        }
    }
    $result = $interface->PerfThresholdTemplateDelete($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result ) {
                my $error =  _check_for_error($result);
                return $error;
    }
    my %thresholdTemplateDetails;
    $thresholdTemplateDetails{TemplateId} =$result->get_TemplateId();
    $thresholdTemplateDetails{TemplateName} =$result->get_TemplateName();
    return \%thresholdTemplateDetails;
}

=head2 B<thresholdTemplate_attachObjects(I<TemplateNameOrId , ObjectId>)>

=item Inputs

=item * TemplateNameOrId

=item * ObjectId


        The input to this function is a hash with TemplateNameOrId and ObjectId .

=item Returns

Returns the emptx XML format if the attachment of the Object to Template is successful (refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    if we are .

    $result = $PALibObj->PerfThresholdTemplateAttachObjects($modHash);

     Getting the values of the count

        my $templateId= 6;
	my $objId = 82;
        my $input = {TemplateNameOrId=>$templateId , ObjectId=>$objId};

         my $result = $soapObj->thresholdTemplate_AttachObjects($input);

        $Log->comment ("Result = $result\n");

=cut

sub thresholdTemplate_attachObjects {

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ObjectFullName" ) ) {
            $modHash->{NewObjects}{ObjectInfo}{ObjectFullName} = $value;
        }
        if ( !( $key cmp "ObjectId" ) ) {
            $modHash->{NewObjects}{ObjectInfo}{ObjectId} = $value;
        }
        if ( !( $key cmp "ObjectName" ) ) {
            $modHash->{NewObjects}{ObjectInfo}{ObjectName} = $value;
        }
        if ( !( $key cmp "ObjectType" ) ) {
            $modHash->{NewObjects}{ObjectInfo}{ObjectType} = $value;
        }
        if ( !( $key cmp "TemplateNameOrId" ) ) {
            $modHash->{TemplateNameOrId} = $value;
        }
    }
    $result = $interface->PerfThresholdTemplateAttachObjects($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result ) {
                my $error =  _check_for_error($result);
                return $error;
    }
    return $result ;
}

=head2 B<thresholdTemplate_detachObjects(I<TemplateNameOrId , ObjectId>)>

=item Inputs

=item * TemplateNameOrId

=item * ObjectId


        The input to this function is a hash with TemplateNameOrId and ObjectId .

=item Returns

Returns the emptx XML format if the attachment of the Object to Template is successful (refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    if we are .

    $result = $PALibObj->PerfThresholdTemplateDetachObjects($modHash);

     Getting the values 

        my $templateId= 6;
        my $objId = 82;
        my $input = {TemplateNameOrId=>$templateId , ObjectId=>$objId};

         my $result = $soapObj->thresholdTemplate_detachObjects($input);

        $Log->comment ("Result = $result\n");

=cut



sub thresholdTemplate_detachObjects{

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ObjectFullName" ) ) {
            $modHash->{ObjectFullName} = $value;
        }
        if ( !( $key cmp "ObjectId" ) ) {
            $modHash->{ObjectId} = $value;
        }
        if ( !( $key cmp "ObjectName" ) ) {
            $modHash->{ObjectName} = $value;
        }
        if ( !( $key cmp "ObjectType" ) ) {
            $modHash->{ObjectType} = $value;
        }
        if ( !( $key cmp "TemplateNameOrId" ) ) {
            $modHash->{TemplateNameOrId} = $value;
        }
    }
    $result = $interface->PerfThresholdTemplateDetachObjects($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result ) {
                my $error =  _check_for_error($result);
                return $error;
    }
    return $result ;
}

=head2 B<thresholdTemplate_modify(I<ObjNameOrId,ThresholdInfo2,TemplateId>)>

=item Inputs

=item * ObjNameOrId

=item * ThresholdInfo2

=item * TemplateId

        The input to this function is a hash .

=item Returns

Returns empty XML file if the modification is successful (refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    if we are .

    $result = $PALibObj->PerfThresholdTemplateCreate($modHash);

     Getting the values 

        my $ObjectName = "f3050-184-160.lab.eng.btc.netapp.in" ;
        my $ThresholdEventName = "sreehari-test-event";
        my $ThresholdInterval = 60;
        my $ThresholdType = "Upper";
        my $ThresholdUnit = "percent";
        my $ThresholdValue = 70;
        my $thresholdObjectType = "system";
        my $CounterName = "cpu_busy";
        my $TemplateName = "Hari-Templatetestsree";
        my $ObjectId = 80 ;
 my $id = 8;
        my $desc = "oopopas";

        my $input = {TemplateId=>$id,ThresholdEventName=>$ThresholdEventName,ThresholdInterval=>$ThresholdInterval,ThresholdType=>$ThresholdType,ThresholdUnit=>$ThresholdUnit,ThresholdValue=>$ThresholdValue,ThresholdObjectType=>$thresholdObjectType,CounterName=>$CounterName,TemplateDescription=>$desc};

         my $ds_um = $soapObj->thresholdTemplate_modify($input);

=cut

sub thresholdTemplate_modify {

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call


    my $thresholds = get_thresholdTemplateList($inputFields);
    #forming the data structure
  foreach my $retValue (@$thresholds){
    while ( ( my $key, my $value ) = each %$retValue ) {
	
	if ( !( $key cmp "CounterDescription" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{CounterDescription} = $value;
        }


        if ( !( $key cmp "CounterName" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{CounterName} = $value;
        }
	if ( !( $key cmp "ThresholdObjectType" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{ObjectType} = $value;
        }
        if ( !( $key cmp "ThresholdType" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdType} = $value;
        }

        if ( !( $key cmp "ThresholdUnit" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdUnit} = $value;
        }

        if ( !( $key cmp "ThresholdValue" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdValue} = $value;
        }
        if ( !( $key cmp "ThresholdIsEnabled" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{IsEnabled} = $value;
        }

        if ( !( $key cmp "ThresholdEventName" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ThresholdEventName} = $value;
        }

        if ( !( $key cmp "ThresholdId" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ThresholdId} = $value;
        }

        if ( !( $key cmp "ThresholdInterval" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ThresholdInterval} = $value;
        }
    }
  }

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "Id" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Id} = $value;
        }

        if ( !( $key cmp "MetricResult" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{MetricResult} = $value;
        }

        if ( !( $key cmp "Period" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Period} = $value;
        }
      if ( !( $key cmp "Slide" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Slide} = $value;
        }

        if ( !( $key cmp "Type" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{Metrics}{Metric}{Type} = $value;
        }

        if ( !( $key cmp "CounterDescription" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{CounterDescription} = $value;
        }


        if ( !( $key cmp "CounterName" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{CounterName} = $value;
        }
        if ( !( $key cmp "IsCalculatedStats" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{IsCalculatedStats} = $value;
        }

        if ( !( $key cmp "LabelNames" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{LabelNames} = $value;
        }


        if ( !( $key cmp "ThresholdObjectType" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{ObjectType} = $value;
        }
        #This one is perfLables LableNamames which is different to the above one .
        if ( !( $key cmp "perfLabelNames" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{PerfLabels}{PerfLabel}{LabelNames} = $value;
        }

        if ( !( $key cmp "TimeConsolidationMethod" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{PerfObjectCounter}{TimeConsolidationMethod} = $value;
        }

        if ( !( $key cmp "ThresholdType" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdType} = $value;
        }

        if ( !( $key cmp "ThresholdUnit" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdUnit} = $value;
        }

        if ( !( $key cmp "ThresholdValue" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{CounterThresholds}{CounterThresholdInfo}{ThresholdValue} = $value;
        }

        if ( !( $key cmp "IsEnabled" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{IsEnabled} = $value;
        }

        if ( !( $key cmp "IsIndirect" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{IsIndirect} = $value;
        }

        if ( !( $key cmp "ObjectFullName" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectFullName} = $value;
        }

        if ( !( $key cmp "ObjectId" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectId} = $value;
        }

        if ( !( $key cmp "ObjectName" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectName} = $value;
        }

        if ( !( $key cmp "ObjectType" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ObjectsInfo}{ObjectInfo}{ObjectType} = $value;
        }

        if ( !( $key cmp "Name" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ResourcePropertyConditions}{ResourcePropertyCondition}{Name} = $value;
        }

        if ( !( $key cmp "Value" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ResourcePropertyConditions}{ResourcePropertyCondition}{Value} = $value;
        }

        if ( !( $key cmp "ThresholdTemplateId" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{TemplateId} = $value;
        }

        if ( !( $key cmp "ThresholdTemplateName" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{TemplateName} = $value;
        }
        if ( !( $key cmp "ThresholdEventName" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ThresholdEventName} = $value;
        }

        if ( !( $key cmp "ThresholdId" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ThresholdId} = $value;
        }

        if ( !( $key cmp "ThresholdInterval" ) ) {
            $modHash->{MemberThresholds}{ThresholdInfo2}{ThresholdInterval} = $value;
        }
        if ( !( $key cmp "TemplateDescription" ) ) {
            $modHash->{TemplateDescription} = $value;
        }
        if ( !( $key cmp "TemplateName" ) ) {
            $modHash->{TemplateName} = $value;
        }
        if ( !( $key cmp "TemplateId" ) ) {
            $modHash->{TemplateId} = $value;
        }

    }

    $result = $interface->PerfThresholdTemplateModify($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                $Log->comment("Error is : $error");
                return $error;
        }
    $Log->comment("Result = $result");
    return $result;
}

=head2 B<get_dataList(I<ObjectNameOrId>)>

=item Inputs

=item * ObjectNameOrId


        The input to this function is a hash  .

=item Returns

Returns  array of hash with all the counter information for diff counter groups (refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    if we are .

    $result = $PALibObj->PerfCounterGroupListInfo($modHash);

     Getting the values


        my $ObjectNameOrId = 82;
        my $input = {ObjectNameOrId=>$ObjectNameOrId};

        my $ds_um = $soapObj->get_dataList($input);
        foreach(@$ds_um) {
                my $cg = $_->{CounterGroupName};
                my $count = $_->{count};
                for (my $i=1;$i<$count;$i++ ){
                           my $addr = $_->{"ApplianceNameOrId.$i"};
                           my $cn = $_->{"CounterName.$i"};
                           $Log->comment("CG = $cg ID = $addr CN = $cn\n");
                           $count = $count -1 ;
                }
        }

=cut

sub get_dataList {

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "CounterGroupName" ) ) {
            $modHash->{CounterGroupName} = $value;
        }
        if ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{ObjectNameOrId} = $value;
        }
        if ( !( $key cmp "ConsolidateHosts" ) ) {
            $modHash->{ConsolidateHosts} = $value;
        }
        if ( !( $key cmp "CounterInfoOnly" ) ) {
            $modHash->{CounterInfoOnly} = $value;
        }
        if ( !( $key cmp "CustomOnly" ) ) {
            $modHash->{CustomOnly} = $value;
        }
        if ( !( $key cmp "DisabledCounterInfoOnly" ) ) {
            $modHash->{DisabledCounterInfoOnly} = $value;
        }
        if ( !( $key cmp "GroupNameOrId" ) ) {
            $modHash->{GroupNameOrId} = $value;
        }
        if ( !( $key cmp "IncludeCalculatedStats" ) ) {
            $modHash->{IncludeCalculatedStats} = $value;
        }
        if ( !( $key cmp "MetaDataOnly" ) ) {
            $modHash->{MetaDataOnly} = $value;
        }
     }	
    $result = $interface->PerfCounterGroupListInfo($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                $Log->comment("Error is : $error");
                return $result;
    }

    $Log->comment("Result = $result");
    my $counterGroups = $result->get_CounterGroups()->get_PerfCounterGroup();
    my @counterInfos;

    foreach (@$counterGroups) {
        my %counterDetails;
        $counterDetails{CounterGroupName} =$_->get_CounterGroupName();
	my $dataSrcs = $_->get_DataSources()->get_PerfDataSource();
	my $count = 1;
	foreach my $perf (@$dataSrcs){
		$counterDetails{"ApplianceNameOrId.$count"} = $perf->get_ApplianceNameOrId();
		$counterDetails{"CounterName.$count"} = $perf->get_CounterName();
		$counterDetails{"InstanceName.$count"} = $perf->get_InstanceName();
		$counterDetails{"ObjectName.$count"} = $perf->get_ObjectName();
		$count = $count+1;
	}
	$counterDetails {count} = $count-1;
	push(@counterInfos, \%counterDetails);
     }	
     return \@counterInfos;

}

=head2 B<perfData_copy(I<ObjectNameOrId , SourceHostNameOrId>)>

=item Inputs

=item * SourceHostNameOrId

=item * ObjectNameOrId


        The input to this function is a hash with SourceHostNameOrId and ObjectNameOrId .

=item Returns

Returns the emptx XML format if the  copy of the perf data is successful (refer examples).

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);

    if we are .

    $result = $PALibObj->PerfCopyCounterConfiguration($modHash);

     Getting the values

        my $SourceHostNameOrId= 6;
        my $objId = 82;
        my $input = {SourceHostNameOrId=>$SourceHostNameOrId , ObjectId=>$objId};

         my $result = $soapObj->perfData_copy($input);

        $Log->comment ("Result = $result\n");

=cut

sub perfData_copy {

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ObjNameOrId" ) ) {
            $modHash->{DestinationObjects}{ObjNameOrId} = $value;
        }
        if ( !( $key cmp "Preview" ) ) {
            $modHash->{Preview} = $value;
        }
        if ( !( $key cmp "SourceHostNameOrId" ) ) {
            $modHash->{SourceHostNameOrId} = $value;
        }
        if ( !( $key cmp "SrcIsDefault" ) ) {
            $modHash->{SrcIsDefault} = $value;
        }
    }
    $result = $interface->PerfCopyCounterConfiguration($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                $Log->comment("Error is : $error");
                return $result;
    }
    $Log->comment("Result = $result");
    return $result;
}

sub get_hostIdsByVersion{

    my %details ;
    $details{'82'}{'agentTimestamp'} = '';
    $details{'82'}{'asisIsLicensed'} = 'Yes';
    $details{'82'}{'autoClientStatEnabled'} = '';
    $details{'82'}{'autoSupportIsEnabled'} = '1';
    $details{'82'}{'cacheConfigVersion'} = '';
    $details{'82'}{'capacityKBytesTotal'} = '4015570944';
    $details{'82'}{'capacityKBytesUsed'} = '3620584448';
    $details{'82'}{'ccTimestamp'} = '';
    $details{'82'}{'cfCannotTakeoverCause'} = '3';
    $details{'82'}{'cfInterconnectStatus'} = '3';

    return \%details;
}

=head2 B<customview_create(I<CounterName,ObjectName,ChartName,PerfChart_CounterName,PerfChart_ObjectName,ViewName>)>

=item Inputs
=item * CounterName
=item * ObjectName
=item * ChartName
=item * PerfChart_CounterName
=item * PerfChart_ObjectName
=item * ViewName
The input to this function is a hash with all the values of PerfViewCreate .

=item Returns
returns an xml if custom view created successfully.
Returns I<undef> if there is error in getting creating view.Error will be printed by library in logs(refer examples)

=item Examples
    $PALibObj = new PALib($dfmserverIP, $user, $password);
    $result = $PALibObj->PerfViewCreate($modHash);
     Getting the values    	
 	 my $counterName = "cifs_ops"; 	
 	 my $objectName = "volume";	
 	 my $ChartName = "Chart1";
     my $ViewName = "chnage";

my $input = {CounterName=>$counterName,ObjectName=>$objectName,ChartName=>$ChartName,PerfChart_CounterName=>$counterName,PerfChart_ObjectName=>$objectName, ViewName => 
$ViewName};

my $result1 =  $soapObj->customview_create($input);
$Log->comment("RESULT IS : $result1");
=cut

sub customview_create{

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {

        if ( !( $key cmp "ApplianceNameOrId" ) ) {
            $modHash->{DataSources}{PerfDataSource}{ApplianceNameOrId} = $value;
        }

        if ( !( $key cmp "CounterName" ) ) {
            $modHash->{DataSources}{PerfDataSource}{CounterName} = $value;
        }

        if ( !( $key cmp "GroupNameOrId" ) ) {
            $modHash->{DataSources}{PerfDataSource}{GroupNameOrId} = $value;
        }

        if ( !( $key cmp "InstanceName" ) ) {
            $modHash->{DataSources}{PerfDataSource}{InstanceName} = $value;
        }

        if ( !( $key cmp "LabelNames" ) ) {
            $modHash->{DataSources}{PerfDataSource}{LabelNames} = $value;
        }

        if ( !( $key cmp "ObjectFullName" ) ) {
            $modHash->{DataSources}{PerfDataSource}{ObjectFullName} = $value;
        }


        if ( !( $key cmp "ObjectName" ) ) {
            $modHash->{DataSources}{PerfDataSource}{ObjectName} = $value;
        }
        if ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{DataSources}{PerfDataSource}{ObjectNameOrId} = $value;
        }

        if ( !( $key cmp "ObjectType" ) ) {
            $modHash->{DataSources}{PerfDataSource}{ObjectType} = $value;
        }

        if ( !( $key cmp "RealTime" ) ) {
            $modHash->{RealTime} = $value;
        }
        #This one is perfview details which is different from the above one .

        if ( !( $key cmp "PerfViewApplianceNameOrId" ) ) {
            $modHash->{View}{PerfView}{ApplianceNameOrId} = $value;
        }

        if ( !( $key cmp "AreEventsShown" ) ) {
            $modHash->{View}{PerfView}{AreEventsShown} = $value;
        }

        if ( !( $key cmp "PerfAssocObjType" ) ) {
            $modHash->{View}{PerfView}{AssocObjType}{PerfAssocObjType} = $value;
        }

        if ( !( $key cmp "ChartName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{ChartName} = $value;
        }

        if ( !( $key cmp "Color" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{Color} = $value;
        }

  #This one is perfchart details which is different from the above one .

        if ( !( $key cmp "PerfChart_ApplianceNameOrId" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ApplianceNameOrId} = $value;
        }

        if ( !( $key cmp "PerfChart_CounterName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{CounterName} = $value;
        }

        if ( !( $key cmp "PerfChart_GroupNameOrId" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{GroupNameOrId} = $value;
        }
        if ( !( $key cmp "PerfChart_InstanceName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{InstanceName} = $value;
        }
        if ( !( $key cmp "PerfChart_LabelNames" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{LabelNames} = $value;
        }

        if ( !( $key cmp "PerfChart_ObjectFullName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ObjectFullName} = $value;
        }

        if ( !( $key cmp "PerfChart_ObjectName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ObjectName} = $value;
        }

        if ( !( $key cmp "PerfChart_ObjectNameOrId" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ObjectNameOrId} = $value;
        }

        if ( !( $key cmp "PerfChart_ObjectType" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ObjectType} = $value;
        }

        if ( !( $key cmp "MaximumInstances" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{MaximumInstances} = $value;
        }

        if ( !( $key cmp "SelectAllInstances" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{SelectAllInstances} = $value;
        }

        if ( !( $key cmp "SortOrder" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{SortOrder} = $value;
        }

        if ( !( $key cmp "TimeInterval" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{TimeInterval} = $value;
        }
         #This one is perfline details which is different from the above one .

        if ( !( $key cmp "PerfLine_Color" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{Color} = $value;
        }
	if ( !( $key cmp "DataGroupMethod" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataGroupMethod} = $value;
        }
	if ( !( $key cmp "PerfLine_ApplianceNameOrId" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ApplianceNameOrId}= $value;
        }
        if ( !( $key cmp "PerfLine_CounterName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{CounterName}= $value;
        }
 	if ( !( $key cmp "PerfLine_GroupNameOrId" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{GroupNameOrId}= $value;
        }
	if ( !( $key cmp "PerfLine_InstanceName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{InstanceName}= $value;
        }
	if ( !( $key cmp "PerfLine_LabelNames" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{LabelNames}= $value;
        }
	if ( !( $key cmp "PerfLine_ObjectFullName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ObjectFullName}= $value;
        }
	if ( !( $key cmp "PerfLine_ObjectName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ObjectName}= $value;
        }
	if ( !( $key cmp "PerfLine_ObjectNameOrId" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ObjectNameOrId}= $value;
        }
	if ( !( $key cmp "PerfLine_ObjectType" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ObjectType}= $value;
        }
	if ( !( $key cmp "LineName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{LineName}= $value;
        }
	if ( !( $key cmp "PerfLine_Type" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{Type}= $value;
        }
	if ( !( $key cmp "MaximumY" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{MaximumY}= $value;
        }

	if ( !( $key cmp "MaximumY" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Type}= $value;
        }
	if ( !( $key cmp "CounterGroupName" ) ) {
            $modHash->{View}{PerfView}{CounterGroupName}= $value;
        }
	if ( !( $key cmp "PerfView_GroupNameOrId" ) ) {
            $modHash->{View}{PerfView}{GroupNameOrId}= $value;
        }
	if ( !( $key cmp "IsDefault" ) ) {
            $modHash->{View}{PerfView}{IsDefault}= $value;
        }
	if ( !( $key cmp "IsStopped" ) ) {
            $modHash->{View}{PerfView}{IsStopped}= $value;
        }
	if ( !( $key cmp "SampleBuffer" ) ) {
            $modHash->{View}{PerfView}{SampleBuffer}= $value;
        }	
	if ( !( $key cmp "SampleRate" ) ) {
            $modHash->{View}{PerfView}{SampleRate}= $value;
        }
	if ( !( $key cmp "ViewName" ) ) {
            $modHash->{View}{PerfView}{ViewName}= $value;
        }
	if ( !( $key cmp "ViewType" ) ) {
            $modHash->{View}{PerfView}{ViewType}= $value;
        }
    }

    $result = $interface->PerfViewCreate($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                $Log->comment("Error is : $error");
				return $result;
     }
     return $result;
}

=head2 B<get_viewlist(I<ApplianceNameOrId>)>

=item Inputs

=item * ApplianceNameOrId(Optional)

The input to this function is a hash, there is no mandatory parameter.

=item Returns

Returns a array of hashes with the list and deatils of all views (refer examples).

Returns I<undef> if there is error in getting viewlist.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);
    $result = $PALibObj->PerfViewListIterStart($modHash);
    Getting the values  
	my $ApplianceNameOrId = "87";
    my $result1 =  $soapObj->get_viewlist();
    $Log->comment("RESULT IS : $result1");
    foreach(@$result1) {
           my $ViewName = $_->{ViewName};
           $Log->comment("Viewname = $ViewName\n");
         }

=cut

sub get_viewlist{
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call
    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ApplianceNameOrId" ) ) {
            $modHash->{ApplianceNameOrId} = $value;
        }
	if ( !( $key cmp "CustomOnly" ) ) {
            $modHash->{CustomOnly}= $value;
        }
	if ( !( $key cmp "GroupNameOrId" ) ) {
            $modHash->{GroupNameOrId}= $value;
        }
	if ( !( $key cmp "IncludeEmptyViews" ) ) {
            $modHash->{IncludeEmptyViews}= $value;
        }
	if ( !( $key cmp "InstanceName" ) ) {
            $modHash->{InstanceName}= $value;
        }
	if ( !( $key cmp "ObjectName" ) ) {
            $modHash->{ObjectName}= $value;
        }
	if ( !( $key cmp "ViewName" ) ) {
            $modHash->{ViewName}= $value;
        }
    }

    $result = $interface->PerfViewListIterStart($modHash);
	#return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                $Log->comment("Error is : $error");
    }
    my $tag =$result->get_Tag();
    my $records =$result->get_Records();

    my $viewdetails = $interface->PerfViewListIterNext({Maximum => $records, Tag => $tag});
    if (not $viewdetails) {
                my $error =  _check_for_error($viewdetails);
                return $viewdetails;
    }
    my $Viewinfo = $viewdetails->get_Views->get_PerfView();
	
   $Log->comment("RESULT  = $Viewinfo");    
   my @viewdata;
  
	
    foreach (@$Viewinfo) {
        my %viewlist;
        $viewlist{AreEventsShown} = $_->get_AreEventsShown();
        my $assocobjtype =  $_->get_AssocObjType();
		foreach my $assocobjtypes (@$assocobjtype) {
			$viewlist{PerfAssocObjType} = $assocobjtypes->get_PerfAssocObjType();
		}
       	    my $perfcharts = $_->get_Charts()->get_PerfChart();
	    foreach my $chart(@$perfcharts){
		my $lines = $chart->get_Lines();
		foreach my $line (@$lines) { 
			my $perflines=$line->get_PerfLine();
			foreach my $perfline (@$perflines) {
				$viewlist{LineName} = $perfline->get_LineName();
				my $datagroups = $perfline->get_DataGroup();
                		foreach  my $datagroup(@$datagroups) {
					my $perfdatagrps = $datagroup->get_PerfDataGroup();
					foreach my $perfdatagrp(@$perfdatagrps) {
			                        $viewlist{CounterName} = $perfdatagrp->get_DataSources()->get_PerfDataSource()->get_CounterName();
                        			$viewlist{ObjectName} =  $perfdatagrp->get_DataSources()->get_PerfDataSource()->get_ObjectName();
					}
                		}
			}
		}
                $viewlist{ChartName} = $chart->get_ChartName();
		my $dynamicdatasources = $chart->get_DynamicDataSources();
     		foreach my $dynamicdatasourcess (@$dynamicdatasources) {
		        my $datasources =$chart->get_DynamicDataSources()->get_PerfDynamicDataSources();
	 	        foreach my $datasource(@$datasources) {
				$viewlist{MaximumInstances} =$datasource->get_MaximumInstances();
				$viewlist{CounterName} = $datasource->get_DataSource()->get_PerfDataSource()->get_CounterName();
			        $viewlist{ObjectName} =  $datasource->get_DataSource()->get_PerfDataSource()->get_ObjectName();

			}
    		}
    }

        $viewlist{CounterGroupName} = $_->get_CounterGroupName();
        $viewlist{IsDefault} = $_->get_IsDefault();
        $viewlist{IsStopped} = $_->get_IsStopped();
	$viewlist{SampleBuffer} =  $_->get_SampleBuffer();
        $viewlist{SampleRate} =  $_->get_SampleRate();
	$viewlist{ViewName} =  $_->get_ViewName();
	$viewlist{ViewType} =  $_->get_ViewType();
        push(@viewdata, \%viewlist);
    }
    $interface->PerfViewListIterEnd({Tag => $tag});
    return \@viewdata;
}

=head2 B<view_destroy(I<CounterName,ObjectName,ChartName,PerfChart_CounterName,PerfChart_ObjectName,ViewName>)>

=item Inputs

=item * ViewName

=item * ApplianceNameOrId

The input to this function is a hash with all the values of PerfViewDestroy .

=item Returns

returns an xml if custom view destroyed successfully.

Returns I<undef> if there is error in  destroying view.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);
    $result = $PALibObj->PerfViewDestroy($modHash);
    Getting the values of the count     
    my $ApplianceNameOrId = "87";
    my $IsCounterGroupAlsoDeleted = "true";
    my $ViewName = "new1";
    my $input = {ViewName => $ViewName};
    my $result1 =  $soapObj->view_destroy($input);
    $Log->comment("RESULT IS : $result1");

=cut

sub view_destroy{
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call
    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ApplianceNameOrId" ) ) {
            $modHash->{ApplianceNameOrId} = $value;
        }
		 if ( !( $key cmp "GroupNameOrId" ) ) {
            $modHash->{GroupNameOrId} = $value;
        }
		 if ( !( $key cmp "IsCounterGroupAlsoDeleted" ) ) {
            $modHash->{IsCounterGroupAlsoDeleted} = $value;
        }
		 if ( !( $key cmp "ViewName" ) ) {
            $modHash->{ViewName} = $value;
        }
        }

    $result = $interface->PerfViewDestroy($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
    }
    return $result;
}

	
=head2 B<view_modify(I<ViewType,PerfChart_CounterName,PerfChart_InstanceName,PerfChart_ObjectName,ChartName,ViewName,View_ViewName,AreEventsShown>)>

=item Inputs
=item * ViewType
=item * PerfChart_CounterName
=item * PerfChart_InstanceName
=item * PerfChart_ObjectName
=item * ChartName
=item * ViewName
=item * View_ViewName
=item * AreEventsShown

The input to this function is a hash with all the values of PerfViewModify .

=item Returns
returns an xml if custom view modified successfully.
Returns I<undef> if there is error in  modifying view.Error will be printed by library in logs(refer examples)

=item Examples
    $PALibObj = new PALib($dfmserverIP, $user, $password);
    $result =  $PALibObj->PerfViewModify($modHash);
    Getting the values    	
 	my $counterName = "http_ops";
 	my $objectName = "system";
 	my $AreEventsShown = "1";
 	my $ChartName = "Chart1";
 	my $InstanceName = "f3020-225-54";
 	my $ViewName = "new";
 	my $ViewType = "custom_view_instances";

my $input = {ViewType=>$ViewType,PerfChart_CounterName=>$counterName,PerfChart_InstanceName=>$InstanceName,PerfChart_ObjectName=>$objectName,ChartName=>$ChartName,ViewName=>$ViewName,View_ViewName=>$ViewName,AreEventsShown=>$AreEventsShown};
        my $result1 =  $soapObj->view_modify($input);
        $Log->comment("RESULT IS : $result1");

=cut

sub view_modify{

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {

        if ( !( $key cmp "ApplianceNameOrId" ) ) {
            $modHash->{ApplianceNameOrId} = $value;
        }
      
        if ( !( $key cmp "GroupNameOrId" ) ) {
            $modHash->{GroupNameOrId} = $value;
        }

        #This one is perfview details which is different from the above one .

        if ( !( $key cmp "PerfView_ApplianceNameOrId" ) ) {
            $modHash->{View}{PerfView}{ApplianceNameOrId} = $value;
        }

        if ( !( $key cmp "AreEventsShown" ) ) {
            $modHash->{View}{PerfView}{AreEventsShown} = $value;
        }

        if ( !( $key cmp "PerfAssocObjType" ) ) {
            $modHash->{View}{PerfView}{AssocObjType}{PerfAssocObjType} = $value;
        }

        if ( !( $key cmp "ChartName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{ChartName} = $value;
        }

        if ( !( $key cmp "Color" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{Color} = $value;
        }

  #This one is perfchart details which is different from the above one .
        if ( !( $key cmp "PerfChart_ApplianceNameOrId" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ApplianceNameOrId} = $value;
        }

        if ( !( $key cmp "PerfChart_CounterName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{CounterName} = $value;
        }

        if ( !( $key cmp "PerfChart_GroupNameOrId" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{GroupNameOrId} = $value;
        }
        if ( !( $key cmp "PerfChart_InstanceName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{InstanceName} = $value;
        }
        if ( !( $key cmp "PerfChart_LabelNames" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{LabelNames} = $value;
        }

        if ( !( $key cmp "PerfChart_ObjectFullName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ObjectFullName} = $value;
        }

        if ( !( $key cmp "PerfChart_ObjectName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ObjectName} = $value;
        }

        if ( !( $key cmp "PerfChart_ObjectNameOrId" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ObjectNameOrId} = $value;
        }

        if ( !( $key cmp "PerfChart_ObjectType" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ObjectType} = $value;
        }


        if ( !( $key cmp "MaximumInstances" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{MaximumInstances} = $value;
        }

        if ( !( $key cmp "SelectAllInstances" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{SelectAllInstances} = $value;
        }

        if ( !( $key cmp "SortOrder" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{SortOrder} = $value;
        }

        if ( !( $key cmp "TimeInterval" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{DynamicDataSources}{PerfDynamicDataSources}{TimeInterval} = $value;
        }
  #This one is perfline details which is different from the above one .

        if ( !( $key cmp "PerfLine_Color" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{Color} = $value;
        }
	if ( !( $key cmp "DataGroupMethod" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataGroupMethod} = $value;
        }
	if ( !( $key cmp "PerfLine_ApplianceNameOrId" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ApplianceNameOrId}= $value;
        }
        if ( !( $key cmp "PerfLine_CounterName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{CounterName}= $value;
        }
 	if ( !( $key cmp "PerfLine_GroupNameOrId" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{GroupNameOrId}= $value;
        }
	if ( !( $key cmp "PerfLine_InstanceName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{InstanceName}= $value;
        }
	if ( !( $key cmp "PerfLine_LabelNames" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{LabelNames}= $value;
        }
	if ( !( $key cmp "PerfLine_ObjectFullName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ObjectFullName}= $value;
        }
	if ( !( $key cmp "PerfLine_ObjectName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ObjectName}= $value;
        }
	if ( !( $key cmp "PerfLine_ObjectNameOrId" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ObjectNameOrId}= $value;
        }
	if ( !( $key cmp "PerfLine_ObjectType" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ObjectType}= $value;
        }
	if ( !( $key cmp "LineName" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{LineName}= $value;
        }
	if ( !( $key cmp "PerfLine_Type" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Lines}{PerfLine}{Type}= $value;
        }
	if ( !( $key cmp "MaximumY" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{MaximumY}= $value;
        }
	if ( !( $key cmp "Type" ) ) {
            $modHash->{View}{PerfView}{Charts}{PerfChart}{Type}= $value;
        }
	if ( !( $key cmp "CounterGroupName" ) ) {
            $modHash->{View}{PerfView}{CounterGroupName}= $value;
        }
	if ( !( $key cmp "PerfView_GroupNameOrId" ) ) {
            $modHash->{View}{PerfView}{GroupNameOrId}= $value;
        }
	if ( !( $key cmp "IsDefault" ) ) {
            $modHash->{View}{PerfView}{IsDefault}= $value;
        }
	if ( !( $key cmp "IsStopped" ) ) {
            $modHash->{View}{PerfView}{IsStopped}= $value;
        }
	if ( !( $key cmp "SampleBuffer" ) ) {
            $modHash->{View}{PerfView}{SampleBuffer}= $value;
        }	
	if ( !( $key cmp "SampleRate" ) ) {
            $modHash->{View}{PerfView}{SampleRate}= $value;
        }
	if ( !( $key cmp "ViewName" ) ) {
            $modHash->{View}{PerfView}{ViewName}= $value;
        }
	if ( !( $key cmp "ViewType" ) ) {
            $modHash->{View}{PerfView}{ViewType}= $value;
        }
	if ( !( $key cmp "View_ViewName" ) ) {
            $modHash->{ViewName}= $value;
        }
    }

    $result = $interface->PerfViewModify($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                $Log->comment("Error is : $error");
		return $result;
    }
    return $result;
}
	
	
=head2 B<view_start(I<CounterGroupName>)>

=item Inputs
=item * CounterGroupName

The input to this function is a hash with all the values of PerfCounterGroupStart.

=item Returns
returns an xml if custom view started successfully.
Returns I<undef> if there is error in starting view.Error will be printed by library in logs(refer examples)

=item Examples
    $PALibObj = new PALib($dfmserverIP, $user, $password);
    $result =  $PALibObj->PerfCounterGroupStart($modHash);
    Getting the values  
  	
    my $countergroupName = "ZAPIview";
    my $counterName = "cifs_ops";
    my $objectName = "volume";
    
    my $input = {CounterGroupName=>$countergroupName};
    my $result1 =  $soapObj->view_start($input);
    $Log->comment("RESULT IS : $result1");

=cut

sub view_start{

    
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {

        if ( !( $key cmp "ApplianceNameOrId" ) ) {
            $modHash->{ApplianceNameOrId} = $value;
        }

	if ( !( $key cmp "GroupNameOrId" ) ) {
            $modHash->{GroupNameOrId} = $value;
        }
	if ( !( $key cmp "CounterGroupName" ) ) {
            $modHash->{CounterGroupName} = $value;
        }
    }

    $result = $interface->PerfCounterGroupStart($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
    }
    return $result;
}

=head2 B<view_stop(I<CounterGroupName>)>
=item Inputs
=item * CounterGroupName
The input to this function is a hash with all the values of PerfCounterGroupStop.

=item Returns
returns an xml if custom view started successfully.
Returns I<undef> if there is error in getting creating view.Error will be printed by library in logs(refer examples)
=item Examples
    $PALibObj = new PALib($dfmserverIP, $user, $password);
    $result =  $PALibObj->PerfCounterGroupStop($modHash);
    Getting the values  
  	
    my $countergroupName = "ZAPIview";
    my $counterName = "cifs_ops";
    my $objectName = "volume";
    
    my $input = {CounterGroupName=>$countergroupName};
 

    my $result1 =  $soapObj->view_stop($input);
    $Log->comment("RESULT IS : $result1");

=cut

sub view_stop{
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {

        if ( !( $key cmp "ApplianceNameOrId" ) ) {
            $modHash->{ApplianceNameOrId} = $value;
        }
	if ( !( $key cmp "GroupNameOrId" ) ) {
            $modHash->{GroupNameOrId} = $value;
        }
	if ( !( $key cmp "CounterGroupName" ) ) {
            $modHash->{CounterGroupName} = $value;
        }
    }

    $result = $interface->PerfCounterGroupStop($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
    }
    return $result;
}

=head2 B<view_associate_object(I<ObjectNameOrId,ViewName>)>

=item Inputs
=item * CounterGroupName

The input to this function is a hash with all the values of PerfViewObjectAssociationAdd.

=item Returns
returns an xml if object assosiated successfully in view.
Returns I<undef> if there is error in getting creating view.Error will be printed by library in logs(refer examples)

=item Examples
    $PALibObj = new PALib($dfmserverIP, $user, $password);
    $result =  $PALibObj->PerfViewObjectAssociationAdd($modHash);
    Getting the values  
  	
    my $ViewName = "ZAPIview";
    my $objectName = "volume";
    
    my $input = {ObjectNameOrId=>$objectName,ViewName=>$ViewName};
 
    my $result1 =  $soapObj->view_associate_object($input);
    $Log->comment("RESULT IS : $result1");
=cut

sub view_associate_object{
    
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {

        if ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{ObjectNameOrId} = $value;
        }

	if ( !( $key cmp "ViewName" ) ) {
            $modHash->{ViewName} = $value;
        }
    }

    $result = $interface->PerfViewObjectAssociationAdd($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
    }
    return $result;
}

=head2 B<view_countergroupgetdata(I<CounterGroupName,DataGroup_CounterName,DataGroup_ObjectName>)>

=item Inputs
=item * CounterGroupName
=item * DataGroup_CounterName
=item * DataGroup_ObjectName

The input to this function is a hash with all the values of PerfCounterGroupGetData.

=item Returns
returns a hash reference having details about the counter.
Returns I<undef> if there is error in getting counterdata.Error will be printed by library in logs(refer examples)

=item Examples
    $PALibObj = new PALib($dfmserverIP, $user, $password);
    $result =  $PALibObj->PerfCounterGroupGetData($modHash);
    Getting the values  
  	
   my $countergroupName = "ZAPIview";
   my $counterName = "cifs_ops";
   my $objectName = "volume";

my $input = {CounterGroupName=>$countergroupName,DataGroup_CounterName=>$counterName,DataGroup_ObjectName=>$objectName};
 

        my $result1 =  $soapObj->view_countergroupgetdata($input);
        my $unit = $result1->{Unit};
        $Log->comment("RESULT IS :$unit ");

=cut

sub view_countergroupgetdata{
    
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {

        if ( !( $key cmp "ApplianceNameOrId" ) ) {
            $modHash->{ApplianceNameOrId} = $value;
        }

        if ( !( $key cmp "CounterGroupName" ) ) {
            $modHash->{CounterGroupName} = $value;
        }
	if ( !( $key cmp "DataGroupMethod" ) ) {
            $modHash->{DataGroup}{PerfDataGroup}{DataGroupMethod} = $value;
        }
	if ( !( $key cmp "DataGroup_ApplianceNameOrId" ) ) {
            $modHash->{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ApplianceNameOrId} = $value;
        }
	if ( !( $key cmp "DataGroup_CounterName" ) ) {
            $modHash->{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{CounterName} = $value;
        }
	if ( !( $key cmp "DataGroup_GroupNameOrId" ) ) {
            $modHash->{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{GroupNameOrId} = $value;
        }
	if ( !( $key cmp "DataGroup_InstanceName" ) ) {
            $modHash->{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{InstanceName} = $value;
        }
	if ( !( $key cmp "DataGroup_LabelNames" ) ) {
            $modHash->{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{LabelNames} = $value;
        }
	if ( !( $key cmp "DataGroup_ObjectFullName" ) ) {
            $modHash->{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ObjectFullName} = $value;
        }
	if ( !( $key cmp "DataGroup_ObjectName" ) ) {
            $modHash->{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ObjectName} = $value;
        }

	if ( !( $key cmp "DataGroup_ObjectNameOrId" ) ) {
            $modHash->{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ObjectNameOrId} = $value;
        }
	if ( !( $key cmp "DataGroup_ObjectType" ) ) {
            $modHash->{DataGroup}{PerfDataGroup}{DataSources}{PerfDataSource}{ObjectType} = $value;
        }
	if ( !( $key cmp "EndTime" ) ) {
            $modHash->{EndTime} = $value;
        }
	if ( !( $key cmp "GroupNameOrId" ) ) {
            $modHash->{GroupNameOrId} = $value;
        }
	if ( !( $key cmp "SampleRate" ) ) {
            $modHash->{SampleRate} = $value;
        }
	if ( !( $key cmp "StartTime" ) ) {
            $modHash->{StartTime} = $value;
        }
	if ( !( $key cmp "TimeConsolidationMethod" ) ) {
            $modHash->{TimeConsolidationMethod} = $value;
        }
    }

    $result = $interface->PerfCounterGroupGetData($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
    }
    $Log->comment("Result is = $result");
    my %counterdetail;
    $counterdetail{CounterData}=$result->get_CounterData();
    $counterdetail{Unit}=$result->get_Unit();
    return \%counterdetail;
}


=head2 B<view_countergroupgetdynamicdata(I<CounterGroupName,Dynamic_CounterName,Dynamic_ObjectName>)>

=item Inputs
=item * CounterGroupName
=item * Dynamic_CounterName
=item * Dynamic_ObjectName

The input to this function is a hash with all the values of PerfCounterGroupGetDynamicDataSources.

=item Returns
returns a xml with result of API.
Returns I<undef> if there is error in getting DynamicData.Error will be printed by library in logs(refer examples)

=item Examples
    $PALibObj = new PALib($dfmserverIP, $user, $password);
    $result =  $PALibObj->PerfCounterGroupGetDynamicDataSources($modHash);
    Getting the values  
  	
   my $countergroupName = "ZAPIview";
   my $counterName = "cifs_ops";
   my $objectName = "volume";

my $input = {CounterGroupName=>$countergroupName,DataGroup_CounterName=>$counterName,DataGroup_ObjectName=>$objectName};
 

        my $result1 =  $soapObj->PerfCounterGroupGetDynamicDataSources($input);
        my $unit = $result1->{Unit};
        $Log->comment("RESULT IS :$unit ");

=cut
sub view_countergroupgetdynamicdata{
  
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call
    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {

        if ( !( $key cmp "ApplianceNameOrId" ) ) {
            $modHash->{ApplianceNameOrId} = $value;
        }

	if ( !( $key cmp "CounterGroupName" ) ) {
            $modHash->{CounterGroupName} = $value;
        }
	if ( !( $key cmp "Color" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{Color} = $value;
        }
	if ( !( $key cmp "Dynamic_ApplianceNameOrId" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ApplianceNameOrId} = $value;
        }
	if ( !( $key cmp "Dynamic_CounterName" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{CounterName} = $value;
        }
	if ( !( $key cmp "Dynamic_GroupNameOrId" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{GroupNameOrId} = $value;
        }

	if ( !( $key cmp "Dynamic_InstanceName" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{InstanceName} = $value;
        }
	if ( !( $key cmp "Dynamic_LabelNames" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{LabelNames} = $value;
        }
	if ( !( $key cmp "Dynamic_ObjectFullName" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ObjectFullName} = $value;
        }
	if ( !( $key cmp "Dynamic_ObjectName" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ObjectName} = $value;
        }

	if ( !( $key cmp "Dynamic_ObjectNameOrId" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ObjectNameOrId} = $value;
        }
	if ( !( $key cmp "Dynamic_ObjectType" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{DataSource}{PerfDataSource}{ObjectType} = $value;
        }
	if ( !( $key cmp "MaximumInstances" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{MaximumInstances} = $value;
        }
	if ( !( $key cmp "SelectAllInstances" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{SelectAllInstances} = $value;
        }
	if ( !( $key cmp "SortOrder" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{SortOrder} = $value;
        }
	if ( !( $key cmp "TimeInterval" ) ) {
            $modHash->{DynamicDataSources}{PerfDynamicDataSources}{TimeInterval} = $value;
        }

	if ( !( $key cmp "GroupNameOrId" ) ) {
            $modHash->{GroupNameOrId} = $value;
        }
    }

    $result = $interface->PerfCounterGroupGetDynamicDataSources($modHash);
    if (not $result) {
                my $error =  _check_for_error($result);
                return $result;
    }
    return $result;
}
=head2 <B> perf_get_counter_data_metrics I<ObjType ObjectNameorId>)>

This function return the counter data for a particular counter.

=over 4

=item Inputs

=item * ObjNameOrId 

=item * Duration 

=item * CounterName

=item * objectType

=item * SampleRate

provide the ObjNameOrId, the values can be I<groupName , groupID , HostName , HostID>.

=item Returns

Returns a reference to an array.Values in array contains references to hashs of the events(refer examples).
Datastore attributes are I<>

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);


    my $input = {ObjectNameOrId =>$ObjName,Duration=>$duration,CounterName=>$counterName,Type=>$statMethod1,ObjectType=>$objectType,SampleRate=>$sampleRate1,Id=>"0"};
    my $result =  $PALibObj->perf_get_counter_data($input);
    my $counterData = $result->{CounterData};
    my $metric = $result->{Metrics};
    my @data = split(",",$counterData);
    my @counterValues;
    foreach (@data){
        my @val= split(":","$_");
        push(@counterValues,$val[1]);
    }


=cut

################################################################################

sub perf_get_counter_data_metrics {
	
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure

    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "Duration" ) ) {
            $modHash->{Duration} = $value;
        }
        if ( !( $key cmp "Direction" ) ) {
            $modHash->{Direction} = $value;
        }

        if ( !( $key cmp "EndTime" ) ) {
            $modHash->{EndTime} = $value;
        }
        if ( !( $key cmp "CounterDescription" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{CounterDescription} = $value;
        }
        if ( !( $key cmp "CounterName" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{CounterName} = $value;
        }
        if ( !( $key cmp "IsCalculatedStats" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{IsCalculatedStats} = $value;
        }
        if ( !( $key cmp "LabelNames" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{LabelNames} = $value;
        }
        if ( !( $key cmp "Id" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{Id} = $value;
        }
        if ( !( $key cmp "MetricResult" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{MetricResult} = $value;
        }
        if ( !( $key cmp "Period" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{Period} = $value;
        }
        if ( !( $key cmp "Slide" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{Slide} = $value;
        }
        if ( !( $key cmp "Value" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{Value} = $value;
        }
        if ( !( $key cmp "Type" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{Type} = $value;
        }

        if ( !( $key cmp "ObjectType" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{ObjectType} = $value;
        }
        if ( !( $key cmp "LabelNames" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{PerfLabels}{PerfLabel}{LabelNames} = $value;
        }
        if ( !( $key cmp "TimeConsolidationMethod" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{TimeConsolidationMethod} = $value;
        }
        if ( !( $key cmp "InstanceName" ) ) {
            $modHash->{InstanceCounterInfo}{InstanceName} = $value;
        }

        if ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{InstanceCounterInfo}{ObjectNameOrId} = $value;
        }
        if ( !( $key cmp "NumberSamples" ) ) {
            $modHash->{NumberSamples} = $value;
        }

        if ( !( $key cmp "SampleRate" ) ) {
            $modHash->{SampleRate} = $value;
        }
        if ( !( $key cmp "StartTime" ) ) {
            $modHash->{StartTime} = $value;
        }
        if ( !( $key cmp "Day" ) ) {
            $modHash->{TimeFilter}{Days}{Day} = $value;
        }
        if ( !( $key cmp "Month" ) ) {
            $modHash->{TimeFilter}{Months}{Months} = $value;
        }
        if ( !( $key cmp "From" ) ) {
            $modHash->{TimeFilter}{TimeRanges}{TimeRange}{From} = $value;
        }
        if ( !( $key cmp "To" ) ) {
            $modHash->{TimeFilter}{TimeRanges}{TimeRange}{To} = $value;
        }

    }


    $result = $interface->PerfGetCounterData($modHash);
    #return _check_for_error($self, $result) if (not $result);
    $Log->log("Result is $result");
    if (not $result ) {
                my $error = _check_for_error($result);
		$Log->log("Error is $result");
                return $result;
    }
    my @perfData;
    my $perfInstCouData = $result->get_PerfInstances()->get_PerfInstanceCounterData();
    foreach (@$perfInstCouData) {
        my %counterData;
    $counterData{CounterData} =$_->get_Counters()->get_PerfCounterData()->get_CounterData();
    $counterData{CounterName} =$_->get_Counters()->get_PerfCounterData()->get_CounterName();
    $counterData{Unit} =$_->get_Counters()->get_PerfCounterData()->get_Unit();
    $counterData{ObjectId} =$_->get_ObjectId();
    $counterData{InstanceName} =$_->get_InstanceName();
    $counterData{Metrics} =$_->get_Counters()->get_PerfCounterData()->get_Metrics()->get_Metric()->get_MetricResult();
    push(@perfData, \%counterData);
    }
    return \@perfData;
}

=head2 <B> perf_get_counter_data I<ObjType ObjectNameorId>)>

This function return the counter data for a particular counter.

=over 4

=item Inputs

=item * ObjNameOrId 

=item * Duration 

=item * CounterName

=item * objectType

=item * SampleRate

provide the ObjNameOrId, the values can be I<groupName , groupID , HostName , HostID>.

=item Returns

Returns a reference to an array.Values in array contains references to hashs of the events(refer examples).
Datastore attributes are I<>

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

    $PALibObj = new PALib($dfmserverIP, $user, $password);


    my $input = {ObjectNameOrId =>$ObjName,Duration=>$duration,CounterName=>$counterName,Type=>$statMethod1,ObjectType=>$objectType,SampleRate=>$sampleRate1,Id=>"0"};
    my $result =  $PALibObj->perf_get_counter_data($input);
    my $counterData = $result->{CounterData};
    my $metric = $result->{Metrics};
    my @data = split(",",$counterData);
    my @counterValues;
    foreach (@data){
        my @val= split(":","$_");
        push(@counterValues,$val[1]);
    }


=cut

################################################################################

sub perf_get_counter_data {
	
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure

    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "Duration" ) ) {
            $modHash->{Duration} = $value;
        }
        if ( !( $key cmp "Direction" ) ) {
            $modHash->{Direction} = $value;
        }

        if ( !( $key cmp "EndTime" ) ) {
            $modHash->{EndTime} = $value;
        }
        if ( !( $key cmp "CounterDescription" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{CounterDescription} = $value;
        }
        if ( !( $key cmp "CounterName" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{CounterName} = $value;
        }
        if ( !( $key cmp "IsCalculatedStats" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{IsCalculatedStats} = $value;
        }
        if ( !( $key cmp "LabelNames" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{LabelNames} = $value;
        }
        if ( !( $key cmp "Id" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{Id} = $value;
        }
        if ( !( $key cmp "MetricResult" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{MetricResult} = $value;
        }
        if ( !( $key cmp "Period" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{Period} = $value;
        }
        if ( !( $key cmp "Value" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{Value} = $value;
        }

        if ( !( $key cmp "Type" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{Type} = $value;
        }

        if ( !( $key cmp "ObjectType" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{ObjectType} = $value;
        }
        if ( !( $key cmp "LabelNames" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{PerfLabels}{PerfLabel}{LabelNames} = $value;
        }
        if ( !( $key cmp "TimeConsolidationMethod" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{TimeConsolidationMethod} = $value;
        }
        if ( !( $key cmp "InstanceName" ) ) {
            $modHash->{InstanceCounterInfo}{InstanceName} = $value;
        }

        if ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{InstanceCounterInfo}{ObjectNameOrId} = $value;
        }
        if ( !( $key cmp "NumberSamples" ) ) {
            $modHash->{NumberSamples} = $value;
        }

        if ( !( $key cmp "SampleRate" ) ) {
            $modHash->{SampleRate} = $value;
        }
        if ( !( $key cmp "StartTime" ) ) {
            $modHash->{StartTime} = $value;
        }
        if ( !( $key cmp "Day" ) ) {
            $modHash->{TimeFilter}{Days}{Day} = $value;
        }
        if ( !( $key cmp "Month" ) ) {
            $modHash->{TimeFilter}{Months}{Months} = $value;
        }
        if ( !( $key cmp "From" ) ) {
            $modHash->{TimeFilter}{TimeRanges}{TimeRange}{From} = $value;
        }
        if ( !( $key cmp "To" ) ) {
            $modHash->{TimeFilter}{TimeRanges}{TimeRange}{To} = $value;
        }

    }


    $result = $interface->PerfGetCounterData($modHash);
    #return _check_for_error($self, $result) if (not $result);
    $Log->log("Result is $result");
    if (not $result ) {
                my $error = _check_for_error($result);
		$Log->log("Error is $result");
                return $result;
    }
    my @perfData;
    my $perfInstCouData = $result->get_PerfInstances()->get_PerfInstanceCounterData();
    foreach (@$perfInstCouData) {
        my %counterData;
    $counterData{CounterData} =$_->get_Counters()->get_PerfCounterData()->get_CounterData();
    $counterData{CounterName} =$_->get_Counters()->get_PerfCounterData()->get_CounterName();
    $counterData{Unit} =$_->get_Counters()->get_PerfCounterData()->get_Unit();
    $counterData{ObjectId} =$_->get_ObjectId();
    $counterData{InstanceName} =$_->get_InstanceName();
    push(@perfData, \%counterData);
    }
    return \@perfData;
}

=head2 <B> get_instances I<ApplianceNameOrId CounterName ObjectType>)>

This function return the instances that are available for the specified ObjectType 

=over 4

=item Inputs

=item * ApplianceNameOrId 

=item * CounterName 

=item * ObjectType 

=item Returns

Returns an array 

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

        my $name1 = "131";
        my $cou = "total_ops";
        my $insta;
        my @arr;
        my $input = {ApplianceNameOrId=>$name1,CounterName=>$cou ,ObjectType=>"volume"};
        my @ds_um = $soapObj->perf_counter_group_list_info($input);
	    $Log->log("Arr = @ds_um");
=cut

sub get_instances {
	
    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure

    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "Duration" ) ) {
            $modHash->{Duration} = $value;
        }
        if ( !( $key cmp "Direction" ) ) {
            $modHash->{Direction} = $value;
        }

        if ( !( $key cmp "EndTime" ) ) {
            $modHash->{EndTime} = $value;
        }
        if ( !( $key cmp "CounterDescription" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{CounterDescription} = $value;
        }
        if ( !( $key cmp "CounterName" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{CounterName} = $value;
        }
        if ( !( $key cmp "IsCalculatedStats" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{IsCalculatedStats} = $value;
        }
        if ( !( $key cmp "LabelNames" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{LabelNames} = $value;
        }
        if ( !( $key cmp "Id" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{Id} = $value;
        }
        if ( !( $key cmp "MetricResult" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{MetricResult} = $value;
        }
        if ( !( $key cmp "Period" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{Period} = $value;
        }
        if ( !( $key cmp "Value" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{Value} = $value;
        }

        if ( !( $key cmp "Type" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{Metrics}{Metric}{Type} = $value;
        }

        if ( !( $key cmp "ObjectType" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{ObjectType} = $value;
        }
        if ( !( $key cmp "LabelNames" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{PerfLabels}{PerfLabel}{LabelNames} = $value;
        }
        if ( !( $key cmp "TimeConsolidationMethod" ) ) {
            $modHash->{InstanceCounterInfo}{CounterInfo}{PerfObjectCounter}{TimeConsolidationMethod} = $value;
        }
        if ( !( $key cmp "InstanceName" ) ) {
            $modHash->{InstanceCounterInfo}{InstanceName} = $value;
        }

        if ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{InstanceCounterInfo}{ObjectNameOrId} = $value;
        }
        if ( !( $key cmp "NumberSamples" ) ) {
            $modHash->{NumberSamples} = $value;
        }

        if ( !( $key cmp "SampleRate" ) ) {
            $modHash->{SampleRate} = $value;
        }
    }

    $result = $interface->PerfGetCounterData($modHash);
    #return _check_for_error($self, $result) if (not $result);
    $Log->log("Result is $result");
    if (not $result ) {
                my $error = _check_for_error($result);
		$Log->log("Error is $result");
                return $result;
    }
    my @instances;
    my $perfInstCouData = $result->get_PerfInstances()->get_PerfInstanceCounterData();
    foreach (@$perfInstCouData) {
	my $name = $_->get_InstanceName();
   	push(@instances, "$name");
    }
    return @instances;
}

=head2 <B> get_enabled_counters I<ApplianceNameOrId CounterGroupName>)>

This function return the enabled counters for an object of a counter group

=over 4

=item Inputs

=item * ApplianceNameOrId 

=item * CounterGroupName 


=item Returns

Returns an array 

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

        my $name1 = "131";
        my $cou = "volume:vserver";
        my $insta;
        my @arr;
        my $input = {ApplianceNameOrId=>$name1,CounterGroupName=>$cou };
        my @ds_um = $soapObj->perf_counter_group_list_info($input);
	    $Log->log("Arr = @ds_um");
=cut

################################################################################
sub get_enabled_counters{

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure

    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ApplianceNameOrId" ) ) {
            $modHash->{ApplianceNameOrId} = $value;
        }
        if ( !( $key cmp "CounterGroupName" ) ) {
            $modHash->{CounterGroupName} = $value;
        }
    }
    	$modHash->{CounterInfoOnly} = "true";
	$modHash->{IncludeCalculatedStats}="false";
	$modHash->{MetaDataOnly} = "true";
	
    $result = $interface->PerfCounterGroupListInfo($modHash);
    #return _check_for_error($self, $result) if (not $result);
    $Log->log("Result is $result");
    if (not $result ) {
            my $error = _check_for_error($result);
   	    $Log->log("Error is $result");
            return $result;
    }
	 my @counters;
	 my @counter;
	 my $perfCouData = $result->get_CounterGroups()->get_PerfCounterGroup()->get_CounterInfo()->get_PerfObjectCounter();
	 foreach (@$perfCouData) {
		  my %counterData;
		$counterData{CounterName} = $_->get_CounterName();
		my $name = $_->get_CounterName();
		push(@counters, "$name");
	
	 }
	 $Log->log("Enabled Counters : @counters");
	 return @counters;
}

=head2 <B> perf_copy_counter_configuration I<ObjectNameOrId SourceHostNameOrId>)>

This function return the array of hasshes of hosts to where data copied

=over 4

=item Inputs

=item * ObjectNameOrId 

=item * SourceHostNameOrId 


=item Returns

Returns an array of haseh

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

=cut


sub perf_copy_counter_configuration{

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure

    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ObjNameOrId" ) ) {
            $modHash->{DestinationObjects}{ObjNameOrId} = $value;
        }
        if ( !( $key cmp "SourceHostNameOrId" ) ) {
            $modHash->{SourceHostNameOrId} = $value;
        }
        if ( !( $key cmp "SrcIsDefault" ) ) {
            $modHash->{SrcIsDefault} = "true";
        }
    }
	$result = $interface->PerfCopyCounterConfiguration($modHash);
    #return _check_for_error($self, $result) if (not $result);
    $Log->log("Result is $result");
    if (not $result ) {
            my $error = _check_for_error($result);
	    $Log->log("Error is $result");
	    return $result;
    }
	 my @zapi_results;
	 my $perfResults = $result->get_CopyResult()->get_DestinationHostInfo();
	 $Log->log("Result is PALIB = $result");
	 foreach (@$perfResults) {
	        my %hostData;
		if ($_->get_ErrorReason()){
		  	$hostData{Error} = $_->get_ErrorReason();
		} else {
			$hostData{Error} = "";
		}
		$hostData{HostId} = $_->get_HostId();
		$hostData{HostName} = $_->get_HostName();
		push(@zapi_results, \%hostData);
	 }
	return \@zapi_results;
}

=head2 <B> get_counter_groups I<ApplianceNameOrId>)>

This function return the array of hasshes of counter groups

=over 4

=item Inputs

=item * ApplianceNameOrId 


=item Returns

Returns an array of haseh

Returns I<undef> if there is error in getting events.Error will be printed by library in logs(refer examples)

=item Examples

=cut


sub get_counter_groups {

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );    #to store object returned from API call

    #forming the data structure

    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "ApplianceNameOrId" ) ) {
            $modHash->{ApplianceNameOrId} = $value;
        }
        if ( !( $key cmp "ConsolidateHosts" ) ) {
            $modHash->{ConsolidateHosts} = $value;
        }
        if ( !( $key cmp "CounterGroupName" ) ) {
            $modHash->{CounterGroupName} = $value;
        }
        if ( !( $key cmp "CounterInfoOnly" ) ) {
            $modHash->{CounterInfoOnly} = $value;
        }
        if ( !( $key cmp "CustomOnly" ) ) {
            $modHash->{CustomOnly} = $value;
        }
        if ( !( $key cmp "DisabledCounterInfoOnly" ) ) {
            $modHash->{DisabledCounterInfoOnly} = $value;
        }
        if ( !( $key cmp "GroupNameOrId" ) ) {
            $modHash->{GroupNameOrId} = $value;
        }
        if ( !( $key cmp "IncludeCalculatedStats" ) ) {
            $modHash->{IncludeCalculatedStats} = $value;
        }
        if ( !( $key cmp "MetaDataOnly" ) ) {
            $modHash->{MetaDataOnly} = $value;
        }
    }
    $result = $interface->PerfCounterGroupListInfo($modHash);
    $Log->log("Result is : $result");

    if (not $result ) {
            my $error = _check_for_error($result);
	    $Log->log("Error is $result");
	    return $result;
    }
	 my @counter_groups;
	 my $perfResults = $result->get_CounterGroups()->get_PerfCounterGroup();
	 foreach (@$perfResults) {
		my %counterData;
		my $count = 1;
		$counterData{CounterGroupName} = $_->get_CounterGroupName();
                my $counters = $_->get_DataSources()->get_PerfDataSource();
	        foreach(@$counters) {
			$counterData{"HostNameOrId".$count} = $_->get_ApplianceNameOrId();
			$counterData{"CounterName".$count} = $_->get_CounterName();
			$counterData{"InstanceName".$count} = $_->get_InstanceName();
			$counterData{"ObjectName".$count} = $_->get_ObjectName();
			$count = $count + 1;
		      }
		$counterData{count} = $count - 1 ;
  		push(@counter_groups, \%counterData);
	 }
	return \@counter_groups;
}



sub workload_list {

    my ( $self, $inputFields ) = @_;
    my ( $result, $modHash, $hash, @arr_hashes );   
    my $include="";

    #forming the data structure
    while ( ( my $key, my $value ) = each %$inputFields ) {
        if ( !( $key cmp "IncludeStreams" ) ) {
            $modHash->{IncludeStreams} = $value;
	    $include = $value;
        } elsif ( !( $key cmp "ObjectNameOrId" ) ) {
            $modHash->{ObjectNameOrId} = $value;
        } elsif ( !( $key cmp "WorkloadClass" ) ) {
            $modHash->{WorkloadClass} = $value;
        }
    }	
    $result = $interface->QosWorkloadListInfoIterStart($modHash);
    #return _check_for_error($self, $result) if (not $result);
    if (not $result) {
                my $error =  _check_for_error($result);
		$Log->log("In start error $result");
                return $result;
    }
    my $tag =$result->get_Tag();
    my $max = $result->get_Records();
    my $resultWorkload = $interface->QosWorkloadListInfoIterNext({Maximum => "$max", Tag => $tag});

    if (not $resultWorkload) {
                my $error =  _check_for_error($resultWorkload);
		$Log->log("In error");
                return $resultWorkload;
    }
    my $workloads = $resultWorkload->get_Workloads()->get_WorkloadInfo();
    $Log->comment("Result is : @$workloads");
    
    my @workloadInfo;
    foreach (@$workloads) {
	my %workloadDetails;
     	my $hash = $_;
        $workloadDetails{HostId} =$_->get_HostId(); 
        $workloadDetails{HostName} =$_->get_HostName(); 
        $workloadDetails{WorkloadClass} =$_->get_WorkloadClass(); 
        $workloadDetails{WorkloadId} =$_->get_WorkloadId(); 
        $workloadDetails{WorkloadName} =$_->get_WorkloadName(); 
        $workloadDetails{WorkloadFullName} =$_->get_WorkloadFullName(); 
	if($include ne ""){
		my $streams = $_->get_Streams()->get_StreamInfo();
		my @stream_arr;
		foreach(@$streams) {
			my %stream_info;
			my $hash1 = $_;
			my $objects = $hash1->get_StorageObjectInfo()->get_ObjectInfo();
			my @object_arr;
			foreach(@$objects) {
				my %streamsDetails;
				$streamsDetails{ObjectFullName} = $_->get_ObjectFullName();
				$streamsDetails{ObjectId} = $_->get_ObjectId();
				$streamsDetails{ObjectName} = $_->get_ObjectName();
				$streamsDetails{ObjectType} = $_->get_ObjectType();
				push(@object_arr,\%streamsDetails);
			}
			$stream_info{ObjectsInfo} = \@object_arr;
			push(@stream_arr,\%stream_info);
		}
		$workloadDetails{Streams} = \@stream_arr;
	}
	push(@workloadInfo , \%workloadDetails);
    }
	$interface->QosWorkloadListInfoIterEnd({Tag => $tag});
	return \@workloadInfo;
}
################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _check_for_error()
# Objective:       Get the fault details from message response
################################################################################

sub _check_for_error
{
    my($result) = @_;

    if (not $result) {
    my $e = $result->get_faultstring();
    my $error = "Error calling SOAP method:" . _get_fault_string($e);
    return $error;
    }
}
################################################################################
# Subroutine Name: _get_fault_detail()
# Objective:       Get the fault details from message response
################################################################################

sub _get_fault_string
{
    my($errorMessage) = @_;
    my $xmlData;
    if(($xmlData) = $errorMessage =~ m/Error.*\n.*\n.*\n.*\n(.*\n.*)/) {
        ;
    }
    else {
        return $errorMessage;
    }
    my $xml = new XML::Simple;
    # read XML data
    my $data = $xml->XMLin($xmlData);
    return $data->{'SOAP-ENV:Body'}->{'SOAP-ENV:Fault'}->{'faultstring'};
}

1;
