# $Id:
#
# Copyright (c) 2010 NetApp.
# All rights reserved.
#
#########################################################
## @summary Library of functions to access Group Soap APIs to perform
##          VIM(Virtual Infrastructure Management) operations.
##
## @author Nachiket (nachiket@netapp.com)
##
## @status public
##
## @pod here
#########################################################

# Package Name
##############
package groups;

require Exporter;
@ISA = qw/Exporter/;

@EXPORT = qw/group_create
  group_add_member
  group_delete_member
  group_destroy
  group_list
  group_member_list
  /;
use Tharn;
use strict;

#Module Import
######################
use UMInterfaces::DfmService::DfmPort;
use SOAP::Transport::HTTP;
use Data::Dumper;
use XML::Simple;

# Package Requirements
######################
use Error;

# Declare global variables
use vars qw(@ISA $interface);

@ISA = qw(Error);

=head2 new

This subroutine is used to create soap object

B<Input:>

url, usrname,password

B<Returns:>

soap object

B<Usage:>

new groups("<server name or IP of DFM Server>","<username>","<password>");

B<Example:>

my $test = new groups("<server name or IP of DFM Server>","<username>","<password>");

=cut

sub new {
    my $class = shift;
    my $self  = {
        _hostName => shift,
        _userName => shift,
        _password => shift,
        _port     => shift,
    };
    $self->{_port} = "8488" if not defined $self->{_port};

    *SOAP::Transport::HTTP::Client::get_basic_credentials = sub {
        return ( $self->{_userName}, $self->{_password} );
    };
    $interface = UMInterfaces::DfmService::DfmPort->new(
        { proxy => "https://$self->{_hostName}:$self->{_port}/apis/soap/v1" } );
    bless( $self, $class );
    return $self;
} ## end sub new

=head2 group_create

This subroutine is used to create DFM groups
B<Input:>

An reference to an hash containing the group name which
is to be created

B<Returns:>

group id
 
B<Usage:>

group_create({GroupName =>"<group name>"})
 
B<Example:>

my $groupid = $soap->group_create({GroupName =>"mygroup"})

=cut

sub group_create {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->GroupCreate($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $groupid = $result->get_GroupId();
    return $groupid;
}

=head2 group_add_member

This subroutine is used to add a member to a given group

B<Input:>

An reference to an hash containing the group name or id
and member name of id

B<Returns:>

none
 
B<Usage:>

group_add_member({GroupNameOrId=>'<group name or id>',MemberNameOrId =>'<member name or id>'})
 
B<Example:>

my $result = $soap->group_add_member({GroupNameOrId=>'test',MemberNameOrId =>'vim3'})

=cut

sub group_add_member {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->GroupAddMember($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
}

=head2 group_delete_member

This subroutine is used to delete a member in a given group

B<Input:>

An reference to an hash containing the group name or id
and member name of id

B<Returns:>

none
 
B<Usage:>

group_delete_member({GroupNameOrId=>'<group name or id>',MemberNameOrId =>'<member name or id>'})
 
B<Example:>

my $result = $soap->group_delete_member({GroupNameOrId=>'test',MemberNameOrId =>'vim3'})

=cut

sub group_delete_member {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->GroupDeleteMember($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
}

=head2 group_destroy

This subroutine is used to destroy a given group

B<Input:>

An reference to an hash containing the group name or id


B<Returns:>

none
 
B<Usage:>

group_destroy({GroupNameOrId=>'<group name or id>'})
 
B<Example:>

my $result = $soap->group_destroy({GroupNameOrId=>'test'})

=cut

sub group_destroy {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->GroupDestroy($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
}

=head2 group_list

This subroutine is used to list all the groups

B<Input:>

An reference to an hash containing the group name or id.
This arg is optional.
 
B<Returns:>

Returns a reference to an array of hashes containing the group names;
 
B<Usage:>

group_list({GroupNameOrId=>'<group name or id>'})
 
B<Example:>

my $result = $soap->list({GroupNameOrId=>'test'})

=cut

sub group_list {
    my ( $self, $args ) = @_;
    my @grouparray;
    $self->_clear_err();
    my $result = $interface->GroupListIterStart($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }

    my $tag = $result->get_Tag();
    my $max = $result->get_Records() . int;

    my $groups =
      $interface->GroupListIterNext( { Maximum => $max, Tag => $tag } );
    my $grouplist = $groups->get_GroupList();
    my $groupinfo = $grouplist->get_GroupInfo();

    foreach (@$groupinfo) {
        my %group;
        $group{Name}          = $_->get_Name();
        $group{Id}            = $_->get_Id();
        $group{ShortName}     = $_->get_ShortName();
        $group{GroupStatus}   = $_->get_GroupStatus();
        $group{SubgroupCount} = $_->get_SubgroupCount();
        $group{ParentGroupId} = $_->get_ParentGroupId();
        push( @grouparray, \%group );
    }

    $interface->GroupMemberListIterEnd( { Tag => $tag } );
    return \@grouparray;

}

=head2 group_member_list

This subroutine is used to list all the members within an group

B<Input:>

An reference to an hash containing another hash which the the group name of id.
 
B<Returns:>

Returns a reference to an array of hashes containing the group members;
 
B<Usage:>

group_list({{Group => {Group=>'<group name or id>'}})
 
B<Example:>

my $result = $soap->group_list({Groups =>{Group => 'test'}})

=cut

sub group_member_list {
    my ( $self, $args ) = @_;
    my @groupmemarray;
    $self->_clear_err();
    my $result = $interface->GroupMemberListIterStart($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }

    my $tag = $result->get_Tag();
    my $max = $result->get_Records() . int;
    $Log->comment("**********Records = $max****************");
    if ( $max == 00 ) {
        return \@groupmemarray;
    }
    my $grouplist =
      $interface->GroupMemberListIterNext( { Maximum => $max, Tag => $tag } );
    my $groupmemberlist = $grouplist->get_GroupMemberList();
    my $groupmem        = $groupmemberlist->get_GroupMemberInfo();

    foreach (@$groupmem) {
        my %groupmem;
        $groupmem{Name} = $_->get_Name();
        $groupmem{Id}   = $_->get_Id();
        push( @groupmemarray, \%groupmem );
    }
    $interface->GroupMemberListIterEnd( { Tag => $tag } );
    return \@groupmemarray;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _check_for_error()
# Objective:       Get the fault details from message response
################################################################################
sub _check_for_error {
    my ($result) = @_;

    if ( not $result ) {
        my $e     = $result->get_faultstring();
        my $error = "Error calling SOAP method:" . _get_fault_string($e);
        return $error;
    }
} ## end sub _check_for_error

################################################################################
# Subroutine Name: _get_fault_detail()
# Objective:       Get the fault details from message response
################################################################################
sub _get_fault_string {
    my ($errorMessage) = @_;
    my $xmlData;

    if ( ($xmlData) = $errorMessage =~ m/Error.*\n.*\n.*\n.*\n(.*\n.*)/ ) {
        ;
    }
    else {
        return $errorMessage;
    }

    # create object
    my $xml = new XML::Simple;

    # read XML data
    my $data = $xml->XMLin($xmlData);
    return $data->{'SOAP-ENV:Body'}->{'SOAP-ENV:Fault'}->{'faultstring'};
} ## end sub _get_fault_string
