# $Id:
#
# Copyright (c) 2010 NetApp.
# All rights reserved.
#
#########################################################
## @summary Library of functions to access RBAC Soap APIs to perform
##          VIM(Virtual Infrastructure Management) operations.
##
## @author Satish Hunasikatti (satishh@netapp.com)
##
## @status public
##
## @pod here
#########################################################

# Package Name
##############
package rbac;

require Exporter;
@ISA = qw/Exporter/;

@EXPORT = qw/
    rbac_role_add
    rbac_role_capability_add
    rbac_admin_role_add
    rbac_admin_role_remove
    rbac_role_capability_remove
    rbac_role_delete
    rbac_access_check
    rbac_operation_add
    rbac_operation_delete
    rbac_role_inherit
    /;
use Tharn;
use strict;

#Module Import
######################
use UMInterfaces::DfmService::DfmPort;
use SOAP::Transport::HTTP;
use Data::Dumper;
use XML::Simple;

# Package Requirements
######################
use Error;

# Declare global variables
use vars qw(@ISA $interface);

@ISA = qw(Error);

=head1 NAME

rbac - Functions to access RBAC Soap APIs to perform VIM(Virtual
Infrastructure Management) operations.

=head1 SYNOPSIS

use rbac;
my $rbacObj = new rbac(<hostname>,<username>,<password>,[port no]);

=head1 DESCRIPTION

This package offers the functions to perform RBAC operations on VIM objects which
internally calls UM SOAP APIs.
While creating the rbac I<object>, I<hostname>, I<username> and
I<password> of the UM host has to be passed as arguments.
The optional parameter I<port no> refers to Soap services port no for UM. This
option is used only when there is change in the UM Soap service port no in
future. So we can create rbac object as follows,

my $rbacObj = new rbac(<hostname>,<username>,<password>);

=head1 FUNCTIONS

More details are explained in each function as below.

=cut

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new Soap object
################################################################################
sub new {
    my $class = shift;
    my $self  = {
        _hostName => shift,
        _userName => shift,
        _password => shift,
        _port     => shift,
    };
    $self->{_port} = "8488" if not defined $self->{_port};

    *SOAP::Transport::HTTP::Client::get_basic_credentials = sub {
        return ( $self->{_userName}, $self->{_password} );
    };
    $interface
        = UMInterfaces::DfmService::DfmPort->new(
        { proxy => "https://$self->{_hostName}:$self->{_port}/apis/soap/v1" }
        );
    bless( $self, $class );
    return $self;
} ## end sub new

=head2 B<rbac_role_add(I<%args>)>

This function adds RBAC role to DFM.

=over 4

=item Inputs

=item * %args

provide the following key<->value pairs as input.
I<RoleName> => I<Role Name To add :string[mandatory]>
I<OwnerNameOrId> => I<Owner role Name To add :string[optional]>
I<Description> => I<Description of the role :string[optional]>

=item Returns

Returns I<Role id>(refer examples).
Returns I<undef> if there is an error(refer examples).

=item Examples

	#create rbac object
    $rbacObj = new rbac(<hostname>,<username>,<password>);

    #create new RBAC role.
    my $rbacId = $rbacObj->rbac_role_add({ RoleName => 'role_name' });

    my $rbacId = $rbacObj->rbac_role_add({
										   RoleName => 'role_name',
	                                       OwnerNameOrId => 'testAdmin'
										   });

    my $rbacId = $rbacObj->rbac_role_add({
	                                       RoleName => 'role_name',
										   OwnerNameOrId => 'dfmAdmin',
										   Description => ''
										   });

    #function will set error, if there is an error.
    if ($rbacObj->error) {
        $Test->nlog($vimsoap->errmsg);
        return $TCD::FAIL;
    }

=cut

################################################################################
# Subroutine Name: rbac_role_add()
# Objective:       create a new RBAC role by DFM
################################################################################
sub rbac_role_add {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->RbacRoleAdd($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $roleId = $result->get_RoleId();
    return $roleId;
} ## end sub rbac_role_add

=head2 B<rbac_role_capability_add(I<%args>)>

This function adds capabilty on DFM functionalities to RBAC role.

=over 4

=item Inputs

=item * %args

provide the following key<->value pairs as input.
I<Operation> => I<An existing Operation To add To the specified role :string[mandatory]>
I<RoleNameOrId> => I<Role Name or object Id of the role To add the Capability :string[mandatory]>
I<Resource> => I<The Resource associated with the given Operation : hash as follows[mandatory]>

    Resource =>  {
      ResourceIdentifier =>  { # DFMTypes::ResourceIdentifier
        Aggregate =>  { # DFMTypes::WrapperOfAggregateResource
          AggregateResource =>  { # DFMTypes::AggregateResource
            AggregateResourceNameOrId =>  { # DFMTypes::WrapperOfAggregateNameOrId
              AggregateNameOrId =>  { # DFMTypes::AggregateNameOrId
                AggregateId =>  $some_value, # unsignedInt
                AggregateName =>  $some_value, # string
              },
            },
            FilerIdentifier =>  { # DFMTypes::WrapperOfFilerResource
              FilerResource =>  { # DFMTypes::FilerResource
                FilerId =>  $some_value, # unsignedInt
                FilerName =>  $some_value, # string
              },
            },
          },
        },
        Dataset =>  { # DFMTypes::WrapperOfDatasetResource
          DatasetResource =>  { # DFMTypes::DatasetResource
            DatasetId =>  $some_value, # unsignedInt
            DatasetName =>  $some_value, # string
          },
        },
        Filer =>  { # DFMTypes::WrapperOfFilerResource
          FilerResource =>  { # DFMTypes::FilerResource
            FilerId =>  $some_value, # unsignedInt
            FilerName =>  $some_value, # string
          },
        },
        Group =>  { # DFMTypes::WrapperOfGroupResource
          GroupResource =>  { # DFMTypes::GroupResource
            GroupId =>  $some_value, # unsignedInt
            GroupName =>  $some_value, # string
          },
        },
        Host =>  { # DFMTypes::WrapperOfHostResource
          HostResource =>  { # DFMTypes::HostResource
            HostId =>  $some_value, # unsignedInt
            HostName =>  $some_value, # string
          },
        },
        Lun =>  { # DFMTypes::WrapperOfLunResource
          LunResource =>  { # DFMTypes::LunResource
            HostIdentifier =>  { # DFMTypes::WrapperOfHostResource
              HostResource =>  { # DFMTypes::HostResource
                HostId =>  $some_value, # unsignedInt
                HostName =>  $some_value, # string
              },
            },
            LunIdentifierNameOrId =>  { # DFMTypes::WrapperOfLunNameOrId
              LunNameOrId =>  { # DFMTypes::LunNameOrId
                LunId =>  $some_value, # unsignedInt
                LunName =>  $some_value, # string
              },
            },
            VolumeIdentifier =>  { # DFMTypes::WrapperOfVolumeResource
              VolumeResource =>  { # DFMTypes::VolumeResource
                AggregateIdentifier =>  { # DFMTypes::WrapperOfAggregateResource
                  AggregateResource =>  { # DFMTypes::AggregateResource
                    AggregateResourceNameOrId =>  { # DFMTypes::WrapperOfAggregateNameOrId
                      AggregateNameOrId =>  { # DFMTypes::AggregateNameOrId
                        AggregateId =>  $some_value, # unsignedInt
                        AggregateName =>  $some_value, # string
                      },
                    },
                    FilerIdentifier =>  { # DFMTypes::WrapperOfFilerResource
                      FilerResource =>  { # DFMTypes::FilerResource
                        FilerId =>  $some_value, # unsignedInt
                        FilerName =>  $some_value, # string
                      },
                    },
                  },
                },
                HostIdentifier =>  { # DFMTypes::WrapperOfHostResource
                  HostResource =>  { # DFMTypes::HostResource
                    HostId =>  $some_value, # unsignedInt
                    HostName =>  $some_value, # string
                  },
                },
                VfilerIdentifier =>  { # DFMTypes::WrapperOfVfilerResource
                  VfilerResource =>  { # DFMTypes::VfilerResource
                    VfilerId =>  $some_value, # unsignedInt
                    VfilerNameOrUuid =>  $some_value, # string
                  },
                },
                VolumeIdentifierNameOrId =>  { # DFMTypes::WrapperOfVolumeNameOrId
                  VolumeNameOrId =>  { # DFMTypes::VolumeNameOrId
                    VolumeId =>  $some_value, # unsignedInt
                    VolumeName =>  $some_value, # string
                  },
                },
              },
            },
          },
        },
        Policy =>  { # DFMTypes::WrapperOfPolicyResource
          PolicyResource =>  { # DFMTypes::PolicyResource
            PolicyId => $some_value, # ObjId
            PolicyName => $some_value, # ObjName
          },
        },
        Qtree =>  { # DFMTypes::WrapperOfQtreeResource
          QtreeResource =>  { # DFMTypes::QtreeResource
            HostIdentifier =>  { # DFMTypes::WrapperOfHostResource
              HostResource =>  { # DFMTypes::HostResource
                HostId =>  $some_value, # unsignedInt
                HostName =>  $some_value, # string
              },
            },
            QtreeIdentifierNameOrId =>  { # DFMTypes::WrapperOfQtreeNameOrId
              QtreeNameOrId =>  { # DFMTypes::QtreeNameOrId
                QtreeId => $some_value, # ObjId
                QtreeName => $some_value, # ObjName
              },
            },
            VolumeIdentifier =>  { # DFMTypes::WrapperOfVolumeResource
              VolumeResource =>  { # DFMTypes::VolumeResource
                AggregateIdentifier =>  { # DFMTypes::WrapperOfAggregateResource
                  AggregateResource =>  { # DFMTypes::AggregateResource
                    AggregateResourceNameOrId =>  { # DFMTypes::WrapperOfAggregateNameOrId
                      AggregateNameOrId =>  { # DFMTypes::AggregateNameOrId
                        AggregateId =>  $some_value, # unsignedInt
                        AggregateName =>  $some_value, # string
                      },
                    },
                    FilerIdentifier =>  { # DFMTypes::WrapperOfFilerResource
                      FilerResource =>  { # DFMTypes::FilerResource
                        FilerId =>  $some_value, # unsignedInt
                        FilerName =>  $some_value, # string
                      },
                    },
                  },
                },
                HostIdentifier =>  { # DFMTypes::WrapperOfHostResource
                  HostResource =>  { # DFMTypes::HostResource
                    HostId =>  $some_value, # unsignedInt
                    HostName =>  $some_value, # string
                  },
                },
                VfilerIdentifier =>  { # DFMTypes::WrapperOfVfilerResource
                  VfilerResource =>  { # DFMTypes::VfilerResource
                    VfilerId =>  $some_value, # unsignedInt
                    VfilerNameOrUuid =>  $some_value, # string
                  },
                },
                VolumeIdentifierNameOrId =>  { # DFMTypes::WrapperOfVolumeNameOrId
                  VolumeNameOrId =>  { # DFMTypes::VolumeNameOrId
                    VolumeId =>  $some_value, # unsignedInt
                    VolumeName =>  $some_value, # string
                  },
                },
              },
            },
          },
        },
        RbacRole =>  { # DFMTypes::WrapperOfRbacRoleResource
          RbacRoleResource =>  { # DFMTypes::RbacRoleResource
            RbacRoleId =>  $some_value, # unsignedInt
            RbacRoleName =>  $some_value, # string
          },
        },
        ResourceId =>  $some_value, # unsignedInt
        ResourcePool =>  { # DFMTypes::WrapperOfResourcePoolResource
          ResourcePoolResource =>  { # DFMTypes::ResourcePoolResource
            ResourcePoolId =>  $some_value, # unsignedInt
            ResourcePoolName =>  $some_value, # string
          },
        },
        StorageService =>  { # DFMTypes::WrapperOfStorageServiceResource
          StorageServiceResource =>  { # DFMTypes::StorageServiceResource
            StorageServiceId => $some_value, # ObjId
            StorageServiceName => $some_value, # ObjName
          },
        },
        Vfiler =>  { # DFMTypes::WrapperOfVfilerResource
          VfilerResource =>  { # DFMTypes::VfilerResource
            VfilerId =>  $some_value, # unsignedInt
            VfilerNameOrUuid =>  $some_value, # string
          },
        },
        VfilerTemplate =>  { # DFMTypes::WrapperOfVfilerTemplateResource
          VfilerTemplateResource =>  { # DFMTypes::VfilerTemplateResource
            VfilerTemplateId => $some_value, # ObjId
            VfilerTemplateName => $some_value, # ObjName
          },
        },
        Volume =>  { # DFMTypes::WrapperOfVolumeResource
          VolumeResource =>  { # DFMTypes::VolumeResource
            AggregateIdentifier =>  { # DFMTypes::WrapperOfAggregateResource
              AggregateResource =>  { # DFMTypes::AggregateResource
                AggregateResourceNameOrId =>  { # DFMTypes::WrapperOfAggregateNameOrId
                  AggregateNameOrId =>  { # DFMTypes::AggregateNameOrId
                    AggregateId =>  $some_value, # unsignedInt
                    AggregateName =>  $some_value, # string
                  },
                },
                FilerIdentifier =>  { # DFMTypes::WrapperOfFilerResource
                  FilerResource =>  { # DFMTypes::FilerResource
                    FilerId =>  $some_value, # unsignedInt
                    FilerName =>  $some_value, # string
                  },
                },
              },
            },
            HostIdentifier =>  { # DFMTypes::WrapperOfHostResource
              HostResource =>  { # DFMTypes::HostResource
                HostId =>  $some_value, # unsignedInt
                HostName =>  $some_value, # string
              },
            },
            VfilerIdentifier =>  { # DFMTypes::WrapperOfVfilerResource
              VfilerResource =>  { # DFMTypes::VfilerResource
                VfilerId =>  $some_value, # unsignedInt
                VfilerNameOrUuid =>  $some_value, # string
              },
            },
            VolumeIdentifierNameOrId =>  { # DFMTypes::WrapperOfVolumeNameOrId
              VolumeNameOrId =>  { # DFMTypes::VolumeNameOrId
                VolumeId =>  $some_value, # unsignedInt
                VolumeName =>  $some_value, # string
              },
            },
          },
        },
      },
    },


=item Returns

No return value.

=item Examples

	#create rbac object
    $rbacObj = new rbac(<hostname>,<username>,<password>);

	my $resourceVdisk = {
					       ResourceIdentifier => {
							   ResourceId => $vdiskId,
					       },
				       };
	$rbacObj->rbac_role_capability_add({
	                                  Operation => 'DFM.Database.Read',
	                                  Resource => $resourceVdisk,
									  RoleNameOrId => 'role_name'
									  });

    #function will set error, if there is an error.
    if ($rbacObj->error) {
        $Test->nlog($vimsoap->errmsg);
        return $TCD::FAIL;
    }

=cut

################################################################################
# Subroutine Name: rbac_role_capability_add()
# Objective:       add capability to role by DFM
################################################################################
sub rbac_role_capability_add {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->RbacRoleCapabilityAdd($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
} ## end sub rbac_role_capability_add

=head2 B<rbac_admin_role_add(I<%args>)>

This function adds role to DFM admin.

=over 4

=item Inputs

=item * %args

provide the following key<->value pairs as input.
I<AdminNameOrId> => I<The admin or usergroup Name or object Id of an admin
                            or usergroup To add role To :string[mandatory]>
I<RoleNameOrId> => I<A role Name or object Id of a role To add To an administrator
                                                   or usergroup :string[mandatory]>

=item Returns

Returns I<AdminNameOrId> (refer examples).
Returns I<undef> if there is an error(refer examples).

=item Examples

	#create rbac object
    $rbacObj = new rbac(<hostname>,<username>,<password>);

    #Add a RBAC role to DFM Admin.
    $rbacObj->rbac_admin_role_add({ AdminNameOrId => 'dfmAdmin', RoleNameOrId => 'role' });

    #function will set error, if there is an error.
    if ($rbacObj->error) {
        $Test->nlog($vimsoap->errmsg);
        return $TCD::FAIL;
    }

=cut

################################################################################
# Subroutine Name: rbac_admin_role_add()
# Objective:       add capability to role by DFM
################################################################################
sub rbac_admin_role_add {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->RbacAdminRoleAdd($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $adminId = $result->get_AdminNameOrId();
    return $adminId;
} ## end sub rbac_admin_role_add

=head2 B<rbac_admin_role_remove(I<%args>)>

This function remove role from DFM admin.

=over 4

=item Inputs

=item * %args

provide the following key<->value pairs as input.
I<AdminNameOrId> => I<An admin or usergroup Name or object Id of an admin or usergroup
                                                      To remove role From :string[mandatory]>
I<DeleteAll> => I<If TRUE, removes all the roles for given administrator :boolean[optional]>
I<RoleNameOrId> => I<A role Name or object Id of a role To add To an administrator
                                                              or usergroup :string[optional]>

=item Returns

Returns I<AdminNameOrId> (refer examples).
Returns I<undef> if there is an error(refer examples).

=item Examples

	#create rbac object
    $rbacObj = new rbac(<hostname>,<username>,<password>);

    #Remove a RBAC role to DFM Admin.
    $rbacObj->rbac_admin_role_remove({ AdminNameOrId => 'dfmAdmin', RoleNameOrId => 'role' });

    #function will set error, if there is an error.
    if ($rbacObj->error) {
        $Test->nlog($vimsoap->errmsg);
        return $TCD::FAIL;
    }

=cut

################################################################################
# Subroutine Name: rbac_admin_role_remove()
# Objective:       remove role for admin by DFM
################################################################################
sub rbac_admin_role_remove {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->RbacAdminRoleRemove($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
} ## end sub rbac_admin_role_remove

=head2 B<rbac_role_capability_remove(I<%args>)>

This function removes capabilty on DFM functionalities to RBAC role.

=over 4

=item Inputs

=item * %args

provide the following key<->value pairs as input.
I<DeleteAll> => I<If TRUE, removes all the Capabilities for given role :boolean[optional]>
I<Operation> => I<An existing Operation To remove from the specified role :string[optional]>
I<RoleNameOrId> => I<Role Name or object Id of the role To remove the Capability :string[mandatory]>
I<Resource> => I<The Resource associated with the given Operation : hash as follows[optional]>

    Resource =>  {
      ResourceIdentifier =>  { # DFMTypes::ResourceIdentifier
        Aggregate =>  { # DFMTypes::WrapperOfAggregateResource
          AggregateResource =>  { # DFMTypes::AggregateResource
            AggregateResourceNameOrId =>  { # DFMTypes::WrapperOfAggregateNameOrId
              AggregateNameOrId =>  { # DFMTypes::AggregateNameOrId
                AggregateId =>  $some_value, # unsignedInt
                AggregateName =>  $some_value, # string
              },
            },
            FilerIdentifier =>  { # DFMTypes::WrapperOfFilerResource
              FilerResource =>  { # DFMTypes::FilerResource
                FilerId =>  $some_value, # unsignedInt
                FilerName =>  $some_value, # string
              },
            },
          },
        },
        Dataset =>  { # DFMTypes::WrapperOfDatasetResource
          DatasetResource =>  { # DFMTypes::DatasetResource
            DatasetId =>  $some_value, # unsignedInt
            DatasetName =>  $some_value, # string
          },
        },
        Filer =>  { # DFMTypes::WrapperOfFilerResource
          FilerResource =>  { # DFMTypes::FilerResource
            FilerId =>  $some_value, # unsignedInt
            FilerName =>  $some_value, # string
          },
        },
        Group =>  { # DFMTypes::WrapperOfGroupResource
          GroupResource =>  { # DFMTypes::GroupResource
            GroupId =>  $some_value, # unsignedInt
            GroupName =>  $some_value, # string
          },
        },
        Host =>  { # DFMTypes::WrapperOfHostResource
          HostResource =>  { # DFMTypes::HostResource
            HostId =>  $some_value, # unsignedInt
            HostName =>  $some_value, # string
          },
        },
        Lun =>  { # DFMTypes::WrapperOfLunResource
          LunResource =>  { # DFMTypes::LunResource
            HostIdentifier =>  { # DFMTypes::WrapperOfHostResource
              HostResource =>  { # DFMTypes::HostResource
                HostId =>  $some_value, # unsignedInt
                HostName =>  $some_value, # string
              },
            },
            LunIdentifierNameOrId =>  { # DFMTypes::WrapperOfLunNameOrId
              LunNameOrId =>  { # DFMTypes::LunNameOrId
                LunId =>  $some_value, # unsignedInt
                LunName =>  $some_value, # string
              },
            },
            VolumeIdentifier =>  { # DFMTypes::WrapperOfVolumeResource
              VolumeResource =>  { # DFMTypes::VolumeResource
                AggregateIdentifier =>  { # DFMTypes::WrapperOfAggregateResource
                  AggregateResource =>  { # DFMTypes::AggregateResource
                    AggregateResourceNameOrId =>  { # DFMTypes::WrapperOfAggregateNameOrId
                      AggregateNameOrId =>  { # DFMTypes::AggregateNameOrId
                        AggregateId =>  $some_value, # unsignedInt
                        AggregateName =>  $some_value, # string
                      },
                    },
                    FilerIdentifier =>  { # DFMTypes::WrapperOfFilerResource
                      FilerResource =>  { # DFMTypes::FilerResource
                        FilerId =>  $some_value, # unsignedInt
                        FilerName =>  $some_value, # string
                      },
                    },
                  },
                },
                HostIdentifier =>  { # DFMTypes::WrapperOfHostResource
                  HostResource =>  { # DFMTypes::HostResource
                    HostId =>  $some_value, # unsignedInt
                    HostName =>  $some_value, # string
                  },
                },
                VfilerIdentifier =>  { # DFMTypes::WrapperOfVfilerResource
                  VfilerResource =>  { # DFMTypes::VfilerResource
                    VfilerId =>  $some_value, # unsignedInt
                    VfilerNameOrUuid =>  $some_value, # string
                  },
                },
                VolumeIdentifierNameOrId =>  { # DFMTypes::WrapperOfVolumeNameOrId
                  VolumeNameOrId =>  { # DFMTypes::VolumeNameOrId
                    VolumeId =>  $some_value, # unsignedInt
                    VolumeName =>  $some_value, # string
                  },
                },
              },
            },
          },
        },
        Policy =>  { # DFMTypes::WrapperOfPolicyResource
          PolicyResource =>  { # DFMTypes::PolicyResource
            PolicyId => $some_value, # ObjId
            PolicyName => $some_value, # ObjName
          },
        },
        Qtree =>  { # DFMTypes::WrapperOfQtreeResource
          QtreeResource =>  { # DFMTypes::QtreeResource
            HostIdentifier =>  { # DFMTypes::WrapperOfHostResource
              HostResource =>  { # DFMTypes::HostResource
                HostId =>  $some_value, # unsignedInt
                HostName =>  $some_value, # string
              },
            },
            QtreeIdentifierNameOrId =>  { # DFMTypes::WrapperOfQtreeNameOrId
              QtreeNameOrId =>  { # DFMTypes::QtreeNameOrId
                QtreeId => $some_value, # ObjId
                QtreeName => $some_value, # ObjName
              },
            },
            VolumeIdentifier =>  { # DFMTypes::WrapperOfVolumeResource
              VolumeResource =>  { # DFMTypes::VolumeResource
                AggregateIdentifier =>  { # DFMTypes::WrapperOfAggregateResource
                  AggregateResource =>  { # DFMTypes::AggregateResource
                    AggregateResourceNameOrId =>  { # DFMTypes::WrapperOfAggregateNameOrId
                      AggregateNameOrId =>  { # DFMTypes::AggregateNameOrId
                        AggregateId =>  $some_value, # unsignedInt
                        AggregateName =>  $some_value, # string
                      },
                    },
                    FilerIdentifier =>  { # DFMTypes::WrapperOfFilerResource
                      FilerResource =>  { # DFMTypes::FilerResource
                        FilerId =>  $some_value, # unsignedInt
                        FilerName =>  $some_value, # string
                      },
                    },
                  },
                },
                HostIdentifier =>  { # DFMTypes::WrapperOfHostResource
                  HostResource =>  { # DFMTypes::HostResource
                    HostId =>  $some_value, # unsignedInt
                    HostName =>  $some_value, # string
                  },
                },
                VfilerIdentifier =>  { # DFMTypes::WrapperOfVfilerResource
                  VfilerResource =>  { # DFMTypes::VfilerResource
                    VfilerId =>  $some_value, # unsignedInt
                    VfilerNameOrUuid =>  $some_value, # string
                  },
                },
                VolumeIdentifierNameOrId =>  { # DFMTypes::WrapperOfVolumeNameOrId
                  VolumeNameOrId =>  { # DFMTypes::VolumeNameOrId
                    VolumeId =>  $some_value, # unsignedInt
                    VolumeName =>  $some_value, # string
                  },
                },
              },
            },
          },
        },
        RbacRole =>  { # DFMTypes::WrapperOfRbacRoleResource
          RbacRoleResource =>  { # DFMTypes::RbacRoleResource
            RbacRoleId =>  $some_value, # unsignedInt
            RbacRoleName =>  $some_value, # string
          },
        },
        ResourceId =>  $some_value, # unsignedInt
        ResourcePool =>  { # DFMTypes::WrapperOfResourcePoolResource
          ResourcePoolResource =>  { # DFMTypes::ResourcePoolResource
            ResourcePoolId =>  $some_value, # unsignedInt
            ResourcePoolName =>  $some_value, # string
          },
        },
        StorageService =>  { # DFMTypes::WrapperOfStorageServiceResource
          StorageServiceResource =>  { # DFMTypes::StorageServiceResource
            StorageServiceId => $some_value, # ObjId
            StorageServiceName => $some_value, # ObjName
          },
        },
        Vfiler =>  { # DFMTypes::WrapperOfVfilerResource
          VfilerResource =>  { # DFMTypes::VfilerResource
            VfilerId =>  $some_value, # unsignedInt
            VfilerNameOrUuid =>  $some_value, # string
          },
        },
        VfilerTemplate =>  { # DFMTypes::WrapperOfVfilerTemplateResource
          VfilerTemplateResource =>  { # DFMTypes::VfilerTemplateResource
            VfilerTemplateId => $some_value, # ObjId
            VfilerTemplateName => $some_value, # ObjName
          },
        },
        Volume =>  { # DFMTypes::WrapperOfVolumeResource
          VolumeResource =>  { # DFMTypes::VolumeResource
            AggregateIdentifier =>  { # DFMTypes::WrapperOfAggregateResource
              AggregateResource =>  { # DFMTypes::AggregateResource
                AggregateResourceNameOrId =>  { # DFMTypes::WrapperOfAggregateNameOrId
                  AggregateNameOrId =>  { # DFMTypes::AggregateNameOrId
                    AggregateId =>  $some_value, # unsignedInt
                    AggregateName =>  $some_value, # string
                  },
                },
                FilerIdentifier =>  { # DFMTypes::WrapperOfFilerResource
                  FilerResource =>  { # DFMTypes::FilerResource
                    FilerId =>  $some_value, # unsignedInt
                    FilerName =>  $some_value, # string
                  },
                },
              },
            },
            HostIdentifier =>  { # DFMTypes::WrapperOfHostResource
              HostResource =>  { # DFMTypes::HostResource
                HostId =>  $some_value, # unsignedInt
                HostName =>  $some_value, # string
              },
            },
            VfilerIdentifier =>  { # DFMTypes::WrapperOfVfilerResource
              VfilerResource =>  { # DFMTypes::VfilerResource
                VfilerId =>  $some_value, # unsignedInt
                VfilerNameOrUuid =>  $some_value, # string
              },
            },
            VolumeIdentifierNameOrId =>  { # DFMTypes::WrapperOfVolumeNameOrId
              VolumeNameOrId =>  { # DFMTypes::VolumeNameOrId
                VolumeId =>  $some_value, # unsignedInt
                VolumeName =>  $some_value, # string
              },
            },
          },
        },
      },
    },


=item Returns

No return value.

=item Examples

	#create rbac object
    $rbacObj = new rbac(<hostname>,<username>,<password>);

	my $resourceVdisk = {
					       ResourceIdentifier => {
							   ResourceId => $vdiskId,
					       },
				       };
	$rbacObj->rbac_role_capability_remove({
	                                  Operation => 'DFM.Database.Read',
	                                  Resource => $resourceVdisk,
									  RoleNameOrId => 'role_name'
									  });

    #function will set error, if there is an error.
    if ($rbacObj->error) {
        $Test->nlog($vimsoap->errmsg);
        return $TCD::FAIL;
    }

=cut

################################################################################
# Subroutine Name: rbac_role_capability_remove()
# Objective:       remove capability to role by DFM
################################################################################
sub rbac_role_capability_remove {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->RbacRoleCapabilityRemove($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
} ## end sub rbac_role_capability_remove

=head2 B<rbac_role_delete(I<%args>)>

This function deletes RBAC role from DFM.

=over 4

=item Inputs

=item * %args

provide the following key<->value pairs as input.
I<RoleName> => I<Role Name To remove :string[mandatory]>

=item Returns

no return value.

=item Examples

	#create rbac object
    $rbacObj = new rbac(<hostname>,<username>,<password>);

    #Delete RBAC role.
    $rbacObj->rbac_role_delete({ RoleName => 'role_name' });

    #function will set error, if there is an error.
    if ($rbacObj->error) {
        $Test->nlog($vimsoap->errmsg);
        return $TCD::FAIL;
    }

=cut

################################################################################
# Subroutine Name: rbac_role_delete()
# Objective:       delete a new role by DFM
################################################################################
sub rbac_role_delete {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->RbacRoleDelete($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
} ## end sub rbac_role_delete

=head2 B<rbac_access_check(I<%args>)>

This function Checks whether the given admin or usergroup
has access To the specified Resource.

=over 4

=item Inputs

=item * %args

provide the following key<->value pairs as input.
I<AdminNameOrId> => I<The admin or usergroup Name or object Id of an admin
                                  or usergroup To check access :string[mandatory]>
I<Operation> => I<Operation requested on given Resource :string[mandatory]>
I<Resource> => I<The specific Resource To check access :hash as follows[mandatory]>

=item Returns

Returns a Hash containing value for following keys,
I<GlobalUsergroupStatus:string>
I<LocalUsergroupStatus:string>
I<Result:string> - Result of whether or not the given admin or usergroup is
                      allowed To perform the specified Action on the given Resource.

=item Examples

	#create rbac object
    $rbacObj = new rbac(<hostname>,<username>,<password>);

    #Check RBAC access.
	my $resourceVdisk = {
					       ResourceIdentifier => {
							   ResourceId => $vdiskId,
					       },
				       };
	my $rbacInfo = $rbacObj->rbac_access_check({
	                                  Operation => 'DFM.Database.Read',
	                                  Resource => $resourceVdisk,
									  AdminNameOrId => 'dfmAdmin'
									  });

	my 	GlobalUsergroupStatus = $rbacInfo->{GlobalUsergroupStatus};
	my  LocalUsergroupStatus = $rbacInfo->{LocalUsergroupStatus};
	my  Result = $rbacInfo->{Result};

    #function will set error, if there is an error.
    if ($rbacObj->error) {
        $Test->nlog($vimsoap->errmsg);
        return $TCD::FAIL;
    }

=cut

################################################################################
# Subroutine Name: rbac_access_check()
# Objective:       check rbac access by DFM
################################################################################
sub rbac_access_check {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->RbacAccessCheck($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }

    my $output;

    $output->{GlobalUsergroupStatus} = $result->get_GlobalUsergroupStatus();
    $output->{LocalUsergroupStatus}  = $result->get_LocalUsergroupStatus();
    $output->{Result}                = $result->get_Result();

    return $output;
} ## end sub rbac_access_check

=head2 B<rbac_operation_add(I<%args>)>

This function adds a new Operation To the RBAC system.

=over 4

=item Inputs

=item * %args

provide the following key<->value pairs as input.
I<Operation> => I<Operation To be added :hash as follows[mandatory]>

$Operation =  { # DFMTypes::WrapperOfRbacOperation
   RbacOperation =>  { # DFMTypes::RbacOperation
     OperationName =>  $some_value, # string
     OperationNameDetails =>  { # DFMTypes::WrapperOfRbacOperationNameDetails
       RbacOperationNameDetails =>  { # DFMTypes::RbacOperationNameDetails
         OperationDescription =>  $some_value, # string
         OperationSynopsis =>  $some_value, # string
         ResourceType =>  $some_value, # string
       },
     },
   },
 },

=item Returns

No return value.

=item Examples

	#create rbac object
    $rbacObj = new rbac(<hostname>,<username>,<password>);

    #Add new operation to RBAC.
    $rbacObj->rbac_operation_add({ Operation => $Operation });


    #function will set error, if there is an error.
    if ($rbacObj->error) {
        $Test->nlog($vimsoap->errmsg);
        return $TCD::FAIL;
    }

=cut

################################################################################
# Subroutine Name: rbac_operation_add()
# Objective:       rbac operation add by DFM
################################################################################
sub rbac_operation_add {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->RbacOperationAdd($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
} ## end sub rbac_operation_add

=head2 B<rbac_operation_delete(I<%args>)>

This function deletes a new Operation To the RBAC system.

=over 4

=item Inputs

=item * %args

provide the following key<->value pairs as input.
I<Operation> => I<Operation To be deleted :string[mandatory]>


=item Returns

No return value.

=item Examples

	#create rbac object
    $rbacObj = new rbac(<hostname>,<username>,<password>);

    #Delete existing operation from RBAC.
    $rbacObj->rbac_operation_delete({ Operation => 'operation' });


    #function will set error, if there is an error.
    if ($rbacObj->error) {
        $Test->nlog($vimsoap->errmsg);
        return $TCD::FAIL;
    }

=cut

################################################################################
# Subroutine Name: rbac_operation_delete()
# Objective:       rbac operation delete by DFM
################################################################################
sub rbac_operation_delete {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->RbacOperationDelete($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
} ## end sub rbac_operation_delete

=head2 B<rbac_role_inherit(I<%args>)>

This function Inherit From a role. The Effect is that the affected role will
                                          gain the Capabilities From the inherited role.

=over 4

=item Inputs

=item * %args

provide the following key<->value pairs as input.
I<InheritedRoleNameOrId> => I<A role Name or object Id of a role To inherit From :string[mandatory]>
I<RoleNameOrId> = I<A role Name or object Id of a role To modify :string[mandatory]>

=item Returns

No return value.

=item Examples

	#create rbac object
    $rbacObj = new rbac(<hostname>,<username>,<password>);

    #Inherit from a role.
    $rbacObj->rbac_role_inherit({ InheritedRoleNameOrId => 'role_to_inherit_from', RoleNameOrId => 'role_name' });


    #function will set error, if there is an error.
    if ($rbacObj->error) {
        $Test->nlog($vimsoap->errmsg);
        return $TCD::FAIL;
    }

=cut

################################################################################
# Subroutine Name: rbac_role_inherit()
# Objective:       rbac role inherit by DFM
################################################################################
sub rbac_role_inherit {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->RbacRoleInherit($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
} ## end sub rbac_role_inherit

=head2 B<rbac_role_disinherit(I<%args>)>

Disinherit one or more roles.

=over 4

=item Inputs

=item * %args

provide the following key<->value pairs as input.
I<DisinheritAll> = I<If TRUE, disinherits all the roles that given RoleNameOrId inherits.>
I<InheritedRoleNameOrId> => I<A role Name or object Id of a role To inherit From :string[mandatory]>
I<RoleNameOrId> = I<A role Name or object Id of a role To modify :string[mandatory]>

=item Returns

No return value.

=item Examples

	#create rbac object
    $rbacObj = new rbac(<hostname>,<username>,<password>);

    #Inherit from a role.
    $rbacObj->rbac_role_disinherit({ InheritedRoleNameOrId => 'role_to_inherit_from', RoleNameOrId => 'role_name' });


    #function will set error, if there is an error.
    if ($rbacObj->error) {
        $Test->nlog($vimsoap->errmsg);
        return $TCD::FAIL;
    }

=cut

################################################################################
# Subroutine Name: rbac_role_disinherit()
# Objective:       rbac role disinherit by DFM
################################################################################
sub rbac_role_disinherit {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->RbacRoleDisinherit($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
} ## end sub rbac_role_disinherit

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _check_for_error()
# Objective:       Get the fault details from message response
################################################################################
sub _check_for_error {
    my ($result) = @_;

    if ( not $result ) {
        my $e     = $result->get_faultstring();
        my $error = "Error calling SOAP method:" . _get_fault_string($e);
        return $error;
    }
} ## end sub _check_for_error

################################################################################
# Subroutine Name: _get_fault_detail()
# Objective:       Get the fault details from message response
################################################################################
sub _get_fault_string {
    my ($errorMessage) = @_;
    my $xmlData;

    if ( ($xmlData) = $errorMessage =~ m/Error.*\n.*\n.*\n.*\n(.*\n.*)/ ) {
        ;
    } else {
        return $errorMessage;
    }

    # create object
    my $xml = new XML::Simple;

    # read XML data
    my $data = $xml->XMLin($xmlData);
    return $data->{'SOAP-ENV:Body'}->{'SOAP-ENV:Fault'}->{'faultstring'};
} ## end sub _get_fault_string
