# $Id:
#
# Copyright (c) 2010 NetApp.
# All rights reserved.
#
#########################################################
## @summary Library of functions to access UM Soap APIs to perform
##          VIM(Virtual Infrastructure Management) operations.
##
## @author Satish Hunasikatti (satishh@netapp.com)
##
## @status public
##
## @pod here
#########################################################

# Package Name
##############
package vimsoap;

require Exporter;
@ISA = qw/Exporter/;

@EXPORT
    = qw/get_datastores register_dpagent unregister_dpagent discover_dpagent
    get_dpagents get_storageresources get_hypervisors get_virtualdisks
    get_virtualcenters get_datacenters get_hypervclusters get_events
    get_dpagentstoragesystemlist get_hsplugin_info compare authorize_dpagent
    configure_dpagent get_job/;
use Tharn;
use strict;

#Module Import
######################
use UMInterfaces::DfmService::DfmPort;
use SOAP::Transport::HTTP;
use Data::Dumper;
use XML::Simple;

# Package Requirements
######################
use Error;

# Declare global variables
use vars qw(@ISA $interface);

@ISA = qw(Error);

=head1 NAME

vimsoap - Functions to access UM soap APIs to perform VIM(Virtual
Infrastructure Management) operations.

=head1 SYNOPSIS

use vimsoap;
my $vimsoapObj = new vimsoap(<hostname>,<username>,<password>,[port no]);

=head1 DESCRIPTION

This package offers the functions to perform operations on VIM objects which
internally calls UM SOAP APIs.
While creating the vimsoap I<object>, I<hostname>, I<username> and
I<password> of the UM host has to be passed as arguments.
The optional parameter I<port no> refers to Soap services port no for UM. This
option is used only when there is change in the UM Soap service port no in
future. So we can create vimsoap object as follows,

my $vimsoapObj = new vimsoap(<hostname>,<username>,<password>);

=head1 FUNCTIONS

All the get functions will return reference array, each value in  array is
reference to hash.
More details are explained in each function as below.

=cut

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new Soap object
################################################################################
sub new {
    my $class = shift;
    my $self  = {
        _hostName => shift,
        _userName => shift,
        _password => shift,
        _port     => shift,
    };
    $self->{_port} = "8488" if not defined $self->{_port};

    *SOAP::Transport::HTTP::Client::get_basic_credentials = sub {
        return ( $self->{_userName}, $self->{_password} );
    };
    $interface
        = UMInterfaces::DfmService::DfmPort->new(
        { proxy => "https://$self->{_hostName}:$self->{_port}/apis/soap/v1" }
        );
    bless( $self, $class );
    return $self;
} ## end sub new

=head2 B<get_events(I<%args>)>

This function gets the event information from UM.

=over 4

=item Inputs

=item * %args [optional]

provide the "EventSourceId" as key to hash input, the value should be
I<DatacenterId>.


=item Returns

Returns a reference to an array.Values in array contains references to hashs
of the datastores(refer examples).
Datastore attributes are
I<EventAbout, EventSourceName, EventCondition, EventSourceId>

Returns I<undef> if there is error in getting datastores.Error will be printed
 by library in logs(refer examples)

=item Examples

    $vimsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #Specifying EventSourceId
    my $event = $soap->get_events({ EventSourceId => $_->{DatacenterId} });

    #listing the elements from the array reference, here is the example for
    #listing event details.

	foreach (@$event) {
		$Test->nlog( $_->{EventSourceName} );
		$Test->nlog( $_->{EventCondition} );
		$Test->nlog( $_->{EventAbout} );
		$Test->nlog( $_->{EventSourceId} );
	}

    #function will set error, if there is an error in getting datastores
    if ($vimsoap->error) {
        $Test->nlog($vimsoap->errmsg);
        return $TCD::FAIL;
    }

=cut

################################################################################
# Subroutine Name: get_events()
# Objective:       Get the event details for a object
################################################################################

sub get_events {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->EventListIterStart($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag = $result->get_Tag();
    my $max = $result->get_Records() . int;
    $max = 5;
    my $resultEvents
        = $interface->EventListIterNext( { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultEvents . "\n" );
    $Log->debug( "######################result :"
            . Dumper($resultEvents)
            . "Max: $max\n" );

    if ( not $resultEvents ) {
        my $error = _check_for_error($resultEvents);
        $self->_clear_err();
        $self->_set_err($error);
        return $resultEvents;
    } ## end if ( not $resultEvents)
    my $eventsInfo = $resultEvents->get_Events();
    my $events = $eventsInfo->get_EventInfo();
    my @eventList;

    foreach (@$events) {
        my %eventsDetails;
        $eventsDetails{EventAbout}      = $_->get_EventAbout();
        $eventsDetails{EventSourceName} = $_->get_EventSourceName();
        $eventsDetails{EventCondition}  = $_->get_EventCondition();
        $eventsDetails{EventSourceId}   = $_->get_EventSourceId();
        $eventsDetails{EventTime}       = $_->get_EventTime();
        $eventsDetails{EventSeverity}   = $_->get_EventSeverity();

        #$eventsDetails{eventSeverity}   = $_->eventSeverity();
        #Add any new fields as needed
        push( @eventList, \%eventsDetails );
    } ## end foreach (@$events)
    $interface->EventListIterEnd( { Tag => $tag } );
    return \@eventList;
} ## end sub get_events

=head2 B<get_datastores(I<%args>)>

This function gets the datastore information from UM.

=over 4

=item Inputs

=item * %args [optional]

provide the "ObjNameOrId" as key to hash input, the values can be
I<Virtual Center,Data Center,Virtual Machine,ESX server>.

OR

provide boolean value  as "true" for key "IncludeDeleted" to List all
including deleted datastores.

=item Returns

Returns a reference to an array.Values in array contains references to hashs
of the datastores(refer examples).
Datastore attributes are
I<DatastoreName, AllocatedCapacity, DatastoreId, UsedCapacity, DatastoreType>

Returns I<undef> if there is error in getting datastores.Error will be printed
 by library in logs(refer examples)

=item Examples

    $vimsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #Not specifying any arguments.
    $datastores = $vimsoapObj->get_datastores();

    #Specifying ObjNameOrId
    $datastores = $vimsoapObj->get_datastores({ObjNameOrId => <ObjNameOrId>
    #List all including deleted datastores
    $datastores = $vimsoapObj->get_datastores({IncludeDeleted => 'true'});

    #listing the elements from the array reference, here is the example for
    #listing datastore names.

    foreach (@$datastores) {
        print $_->{DatastoreName};
        print $_->{AllocatedCapacity};
        print $_->{DatastoreId};
        print $_->{UsedCapacity};
        print $_->{DatastoreType};
    }

    #function will set error, if there is an error in getting datastores
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }


=cut

################################################################################
# Subroutine Name: get_datastores()
# Objective:       Get the datastores discovered by DFM
################################################################################
sub get_datastores {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->ViDatastoreListInfoIterStart($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag             = $result->get_Tag();
    my $max             = $result->get_Records() . int;
    my $resultDatastore = $interface->ViDatastoreListInfoIterNext(
        { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultDatastore . "\n" );

    if ( not $resultDatastore ) {
        my $error = _check_for_error($resultDatastore);
        $self->_clear_err();
        $self->_set_err($error);
        return $resultDatastore;
    } ## end if ( not $resultDatastore)
    my $datastoresInfo = $resultDatastore->get_Datastores();
    my $datastore      = $datastoresInfo->get_Datastore();
    my @datastoreList;

    foreach (@$datastore) {
        my %datastoreDetails;
        $datastoreDetails{DatastoreName}     = $_->get_DatastoreName();
        $datastoreDetails{AllocatedCapacity} = $_->get_AllocatedCapacity();
        $datastoreDetails{DatastoreId}       = $_->get_DatastoreId();
        $datastoreDetails{UsedCapacity}      = $_->get_UsedCapacity();
        $datastoreDetails{DatastoreType}     = $_->get_DatastoreType();
        push( @datastoreList, \%datastoreDetails );
    } ## end foreach (@$datastore)
    $interface->ViDatastoreListInfoIterEnd( { Tag => $tag } );
    return \@datastoreList;
} ## end sub get_datastores

=head2 B<get_virtualmachines(I<%args>)>

This function gets the virtual machine information from UM.

=over 4

=item Inputs

=item * %args [optional]

Define the following keys in the hash input.

"ObjNameOrId", the values can be
I<Virtual Center,Data Center,Datastore,Hypervisor>.

"VirtualInfrastructureType", the values can be
I<NetApp.VMwareManagement, NetApp.HyperVManagement>.

OR


provide boolean value  as "true" for key "IncludeDeleted" to List all
including deleted virtual machines.

=item Returns

Returns a reference to an array.Values in array contains references to hashs
 of the virtual machines(refer examples).
virtual machine attributes are I<VirtualMachineName,
 GuestOsType, IsProtected,>
I<VirtualInfrastructureType, VirtualMachineId>

Returns I<undef> if there is error in getting virtualMachines.Error will be
 printed by library in logs(refer examples)

=item Examples

    $unsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #Not specifying any arguments.
    $virtualMachines = $vimsoapObj->get_virtualmachines();

    #Specifying ObjNameOrId
    $virtualMachines = $vimsoapObj->get_virtualmachines({
                              ObjNameOrId => <ObjNameOrId>
                       });

    #Specifying VirtualInfrastructureType
    $virtualMachines = $vimsoapObj->get_virtualmachines({
                VirtualInfrastructureType => <VirtualInfrastructureType>
                       });

    #Specifying both ObjNameOrId and VirtualInfrastructureType
    $virtualMachines = $vimsoapObj->get_virtualmachines({
                              ObjNameOrId => <ObjNameOrId>,
               VirtualInfrastructureType => <VirtualInfrastructureType>
                       });

    #List all including deleted virtual machines.
    $virtualMachines = $vimsoapObj->get_virtualmachines({
                                                IncludeDeleted => 'true'
                                    });

    #listing the elements from the array reference, here is the example for
    listing virtual machine names.
    foreach (@$virtualMachines) {
        print $_->{VirtualMachineName};
        print $_->{GuestOsType};
        print $_->{IsProtected};
		print $_->{Status};
    }

    #function will set error, if there is an error in getting virtual machines.
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }


=cut

################################################################################
# Subroutine Name: get_virtualmachines()
# Objective:       Get the virtual machines discovered by DFM
################################################################################
sub get_virtualmachines {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->ViVirtualMachineListInfoIterStart($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag      = $result->get_Tag();
    my $max      = $result->get_Records() . int;
    my $resultVm = $interface->ViVirtualMachineListInfoIterNext(
        { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultVm . "\n" );

    if ( not $resultVm ) {
        my $error = _check_for_error($resultVm);
        $self->_clear_err();
        $self->_set_err($error);
        return $resultVm;
    } ## end if ( not $resultVm )
    my $vmInfo = $resultVm->get_VirtualMachines();
    my $vm     = $vmInfo->get_VirtualMachine();
    my @vmList;

    foreach (@$vm) {
        my %vmDetails;
        $vmDetails{VirtualMachineName} = $_->get_VirtualMachineName();
        $vmDetails{GuestOsType}        = $_->get_GuestOsType();
        $vmDetails{IsProtected}        = $_->get_IsProtected();
        $vmDetails{VirtualInfrastructureType}
            = $_->get_VirtualInfrastructureType();
        $vmDetails{VirtualMachineId} = $_->get_VirtualMachineId();
        $vmDetails{Status}           = $_->get_Status();
        push( @vmList, \%vmDetails );
    } ## end foreach (@$vm)
    $interface->ViVirtualMachineListInfoIterEnd( { Tag => $tag } );
    return \@vmList;
} ## end sub get_virtualmachines

=head2 B<register_dpagent(I<%args>)>

This function registers HS with UM.

=over 4

=item Inputs

=item * %args

Define the following keys in the hash input.

"NameOrIpAddress" [mandatory]
provide the name or IP Address of the host agent.

"AdminPort" [optional]
provide the admin port for the host agent.

provide the AdminPort.

=item Returns

Id of the of job created to discover resources managed by the Host Service.
Returned only if the Host Service resgistration is successful.

=item Examples

    $unsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #register dp agent
    $jobId = $vimsoapObj->register_dpagent({
                             NameOrIpAddress => <NameOrIpAddress>,
                                          AdminPort => <AdminPort>
                          });

    #function will set error, if there is an error in registering dp agent.
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: register_dpagent()
# Objective:       Register Host Agent with the DFM
################################################################################
sub register_dpagent {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->HostServiceRegister($args);
    $Log->debug( "result :" . $result . "\n" );

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }

    #my $jobId = $result->get_JobId();
    my $jobId = $result->get_HostServiceId();
    return $jobId;
} ## end sub register_dpagent

=head2 B<authorize_dpagent(I<%args>)>

This function authorizes HS with UM.

=over 4

=item Inputs

=item * %args

Define the following keys in the hash input.

"HostService Id" [mandatory]
provide the name or IP Address of the host agent.


=item Returns

Id of the of job created To discover the virtual inventory information
managed by the Host Service agent.

=item Examples

    $unsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #register dp agent
    $jobId = $vimsoapObj->authorize_dpagent({
                             HostServiceNameOrId => <hostservices id>});
                        

    #function will set error, if there is an error in registering dp agent.
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: authorize_dpagent()
# Objective:       Authorize Host Agent with the DFM
################################################################################
sub authorize_dpagent {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->HostServiceAuthorize($args);
    $Log->debug( "result :" . $result . "\n" );

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $jobId = $result->get_JobId();
    return $jobId;
} ## end sub authorize_dpagent

=head2 B<unregister_dpagent(I<$args>)>

This function unregisters HS with UM.

=over 4

=item Inputs

=item * $args

=item * DpAgentNameOrId [mandatory]

Provide "DpAgentNameOrId"

=item Returns

This function doesn't return any information.

=item Examples

    $unsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #register dp agent
    $jobId = $vimsoapObj->unregister_dpagent(<DpAgentNameOrId>);

    #function will set error, if there is an error in unregistering dp agent.
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: unregister_dpagent()
# Objective:       Register Host Agent with the DFM
################################################################################
sub unregister_dpagent {
    my ( $self, $dpAgentNameOrId ) = @_;
    $self->_clear_err();
    my $result = $interface->HostServiceUnregister(
        { HostServiceNameOrId => $dpAgentNameOrId } );
    $Log->debug( "result :" . $result . "\n" );

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
} ## end sub unregister_dpagent

=head2 B<discover_dpagent(I<DpAgentNameOrId>)>

This function discovers HS with UM.

=over 4

=item Inputs

=item * DpAgentNameOrId [mandatory]

provide the DpAgentNameOrId.

=item Returns

Id of the of job created to discover resources managed by the Host Service.

=item Examples

    $unsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #register dp agent
    $jobId = $vimsoapObj->discover_dpagent(<DpAgentNameOrId>);

    #function will set error, if there is an error in refreshing dp agents.
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: discover_dpagent()
# Objective:       Get the virtual machines discovered by DFM
################################################################################
sub discover_dpagent {
    my ( $self, $dpAgentId, $trackProgress ) = @_;
    $self->_clear_err();

    unless ( $trackProgress )  { $trackProgress = 0; }
    my $result = $interface->HostServiceDiscover(
        {   HostServiceNameOrId => $dpAgentId,
            TrackProgress   => "$trackProgress",
        }
    );
    $Log->debug( "result :" . $result . "\n" );

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }

    my $jobId = $result->get_JobId();
    return $jobId;
} ## end sub discover_dpagent


=head2 B<configure_dpagent(I<%args>)>

This function for HS configuration

=over 6

=item Inputs

=item * %args

Define the following keys in the hash input.

"DpAgentNameOrId" [mandatory]
provide the Name or Id of Host Service registered with DFM

"DfmServerIpAddress" [optional]
IP address of the DFM Server

"StorageSystemsConfiguration" [optional]
Configuration settings for storage systems that needs to be set on the HS

"VcenterIpAddress" [optional]
IP address of the vcenter server host

"VcenterPassword" [optional]
VCenter user password, required by SMVI plugin to communicate with VCenter

"VcenterUsername" [optional]
VCenter user password, required by SMVI plugin to communicate with VCenter

=item Returns

Id of the of job created to Configure the host service,
Returned only if the Host Service Configuration is successful.

=item Examples

    $unsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #register dp agent
    $jobId = $vimsoapObj->configure_dpagent({
                            DpAgentNameOrId => <NameOrId>
                        });

    #function will set error, if there is an error in registering dp agent.
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: configure_dpagent()
# Objective:       Configure the Host Agent
################################################################################
sub configure_dpagent {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    $Log->debug( "configure_dpagent:" . Dumper($args) );
    my $result = $interface->HostServiceConfigure($args);
    $Log->debug( "result :" . $result . "\n" );

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    } ## end if ( not $result )

    my $jobId = $result->get_JobId();
    return $jobId;
} ## end sub configure_dpagent

=head2 B<get_dpagents(I<%args>)>

This function gets the HS information from UM.

=over 4

=item Inputs

OR

provide boolean value  as "true" for key "IncludeDeleted" to List all
 including deleted storageresources.


=item Returns

Returns a reference to an array.Values in array contains references to hashs
 of the dp agents(refer examples).
HS attributes are I<AdminPort, DpAgentId, DpAgentIpAddress, DpAgentName,
 DpAgentVersion, MgmtPort, UpdateServicePort, UpdateServiceVersion>

Returns I<undef> if there is error in getting datastores.Error will be printed
 by library in logs(refer examples)

=item Examples

    $unsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #get the HS information.
    $dpagents = $vimsoapObj->get_dpagents();


    #listing the elements from the array reference, here is the example for
    #listing dp agent information.
    foreach (@$dpagents) {
        print $_->{AdminPort};
        print $_->{DpAgentId};
        print $_->{DpAgentIpAddress};
        print $_->{DpAgentName};
        print $_->{DpAgentVersion};
        print $_->{MgmtPort};
        print $_->{UpdateServicePort};
        print $_->{UpdateServiceVersion};
    }

    #function will set error, if there is an error in getting dp agents.
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }


=cut

################################################################################
# Subroutine Name: get_dpagents()
# Objective:       Get the dp agents from DFM
################################################################################
sub get_dpagents {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->HostServiceListInfoIterStart($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag           = $result->get_Tag();
    my $max           = $result->get_Records() . int;
    my $resultDpAgent = $interface->HostServiceListInfoIterNext(
        { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultDpAgent . "\n" );

    if ( not $resultDpAgent ) {
        my $error = _check_for_error($resultDpAgent);
        $self->_clear_err();
        $self->_set_err($error);
        return $resultDpAgent;
    } ## end if ( not $resultDpAgent)
    my $DpAgentsInfo = $resultDpAgent->get_HostServices();
    my $DpAgent      = $DpAgentsInfo->get_HostServiceInfo();
    my @dpAgentList;

    foreach (@$DpAgent) {
        my %dpAgentDetails;
        $dpAgentDetails{AdminPort}        = $_->get_AdminPort();
        $dpAgentDetails{DpAgentId}        = $_->get_HostServiceId();
        $dpAgentDetails{DpAgentIpAddress} = $_->get_HostServiceIpAddress();
        $dpAgentDetails{DpAgentName}      = $_->get_HostServiceName();
        $dpAgentDetails{DpAgentVersion}   = $_->get_HostServiceVersion();

        #$dpAgentDetails{InventoryState} = $_->get_InventoryState();
        $dpAgentDetails{MgmtPort}          = $_->get_MgmtPort();
        $dpAgentDetails{UpdateServicePort} = $_->get_UpdateServicePort();
        $dpAgentDetails{UpdateServiceVersion}
            = $_->get_UpdateServiceVersion();
        push( @dpAgentList, \%dpAgentDetails );
    } ## end foreach (@$DpAgent)
    $interface->HostServiceListInfoIterEnd( { Tag => $tag } );
    return \@dpAgentList;
} ## end sub get_dpagents

=head2 B<get_job(I<%args>)>

This function gets the job information from UM.

=over 4

=item Inputs

=item * %args [optional]

provide the "JobId" as key to hash input, the value should be
I<DatacenterId>.


=item Returns

Returns a reference to an array.Values in array contains references to hashs
of the datastores(refer examples).
Datastore attributes are
I< JobContextApplicationResourceNamespace, JobContextApplicationResourcesOnly
   JobContextObjNameOrId, JobId, JobState, JobType, JobTypes, MaxJobs
   ObjectNameOrId, StartedTimestampFrom, StartedTimestampTo>

Returns I<undef> if there is error in getting datastores.
Error will be printed by library in logs(refer examples)

=item Examples

    $vimsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #Specifying jobSourceId
    my $jobs = $soap->get_job({ JobId => $JobId });

    #listing the elements from the array reference, here is the example for
    #listing job details.

	foreach (@$jobs) {
		$Test->nlog( $_->get_JobDescription());
		$Test->nlog( $_->get_JobId());
		$Test->nlog( $_->get_JobOverallStatus());
		$Test->nlog( $_->get_JobState());
	}

    #function will set error, if there is an error in getting datastores
    if ($vimsoap->error) {
        $Test->nlog($vimsoap->errmsg);
        return $TCD::FAIL;
    }

=cut

################################################################################
# Subroutine Name: get_job()
# Objective:       Get the job details for a Job
################################################################################

sub get_job {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    $Log->debug( "Calling get_job:" . Dumper($args) );
    my $result = $interface->DpJobListIterStart($args);
    $Log->comment("Called get_job:");

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag = $result->get_Tag();
    my $max = $result->get_Records() . int;

    #if ($args->{MaxJobs}) { $max = $args->{MaxJobs}; }

    my $resultJobs
        = $interface->DpJobListIterNext( { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultJobs . "\n" );

    if ( not $resultJobs ) {
        my $error = _check_for_error($resultJobs);
        $self->_clear_err();
        $self->_set_err($error);
        return $resultJobs;
    } ## end if ( not $resultJobs )

    my $jobsInfo = $resultJobs->get_Jobs();

    $Log->debug( "resultJobs:" . Dumper($jobsInfo) );
    my $jobs = $jobsInfo->get_DpJobInfo();
    my @jobsList;

    foreach (@$jobs) {
        my %jobsDetails;
        $jobsDetails{JobDescription}   = $_->get_JobDescription();
        $jobsDetails{JobId}            = $_->get_JobId();
        $jobsDetails{JobOverallStatus} = $_->get_JobOverallStatus();
        $jobsDetails{JobState}         = $_->get_JobState();

        push( @jobsList, \%jobsDetails );
    } ## end foreach (@$jobs)
    $interface->DpJobListIterEnd( { Tag => $tag } );
    return \@jobsList;
} ## end sub get_job

=head2 B<get_storageresources(I<%args>)>

This function gets the storage resource information from UM.

=over 4

=item Inputs

=item * %args [optional]

provide the ObjNameOrId as key , the values can be
I<Virtual Center,HyperVisor,storageresource,Virtual Machine>.

OR

provide boolean value  as "true" for key "IncludeDeleted" to List all
 including deleted storageresources.

=item Returns

Returns a reference to an array.Values in array contains references to hashs
 of the storageresources(refer examples).
storageresource attributes are I<StorageResourceId, StorageResourceName, Type>

Returns I<undef> if there is error in getting storageresources.Error will be
 printed by library in logs(refer examples)

=item Examples

    $unsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #Not specifying any arguments.
    $storageresources = $vimsoapObj->get_storageresources();

    #Specifying ObjNameOrId.
    $storageresources = $vimsoapObj->get_storageresources({
                                             ObjNameOrId => <ObjNameOrId>
                                     });

    #List all including deleted storage resources.
    $storageresources = $vimsoapObj->get_storageresources({
                                                 IncludeDeleted => 'true'
                                      });

    #listing the elements from the array reference, here is the example for
    #listing storageresource names.
    foreach (@$storageresources) {
        print $_->{StorageResourceName};
        print $_->{StorageResourceId};
        print $_->{Type};
    }

    #function will set error, if there is an error in getting
    #storage resources.
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: get_storageresources()
# Objective:       Get the storage resources from DFM
################################################################################
sub get_storageresources {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->ViStorageResourceListInfoIterStart($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag                   = $result->get_Tag();
    my $max                   = $result->get_Records() . int;
    my $resultStorageResource = $interface->ViStorageResourceListInfoIterNext(
        { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultStorageResource . "\n" );

    if ( not $resultStorageResource ) {
        my $error = _check_for_error($resultStorageResource);
        $self->_clear_err();
        $self->_set_err($error);
        return $resultStorageResource;
    } ## end if ( not $resultStorageResource)
    my $StorageResourcesInfo = $resultStorageResource->get_StorageResources();
    my $StorageResource      = $StorageResourcesInfo->get_StorageResource();
    my @storageResourceList;

    foreach (@$StorageResource) {
        my %storageResourceDetails;
        $storageResourceDetails{StorageResourceId}
            = $_->get_StorageResourceId();
        $storageResourceDetails{StorageResourceName}
            = $_->get_StorageResourceName();
        $storageResourceDetails{Type} = $_->get_Type();
        push( @storageResourceList, \%storageResourceDetails );
    } ## end foreach (@$StorageResource)
    $interface->ViStorageResourceListInfoIterEnd( { Tag => $tag } );
    return \@storageResourceList;
} ## end sub get_storageresources

=head2 B<get_hypervisors(I<%args>)>

This function gets the hypervisors information from UM.

=over 4

=item Inputs

=item * ObjNameOrId [optional]

=item * VirtualInfrastructureType [optional]

Define the following keys in the hash input.

"ObjNameOrId", the values can be
 I<Virtual Center, Datastore, Data Center, Microsoft HyperV Cluster>.

"VirtualInfrastructureType", the values can be
 I<NetApp.VMwareManagement, NetApp.HyperVManagement>.

OR

provide boolean value  as "true" for key "IncludeDeleted" to List all
including deleted hypervisors.

=item Returns

Returns a reference to an array.Values in array contains references to hashs
 of the hypervisors(refer examples).
hypervisor attributes are
 I<DatacenterName,DomainName, HypervisorId, HypervisorName, IpAddress, >
I<VirtualCenterName, VirtualInfrastructureType>

Returns I<undef> if there is error in getting hypervisors.Error will be
 printed by library in logs(refer examples)

=item Examples

    $unsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #Not specifying any arguments.
    $hypervisors = $vimsoapObj->get_hypervisors();

    #Specifying ObjNameOrId
    $hypervisors = $vimsoapObj->get_hypervisors({
                                ObjNameOrId => <ObjNameOrId>
                   });

    #Specifying VirtualInfrastructureType
    $hypervisors = $vimsoapObj->get_hypervisors({
              VirtualInfrastructureType => <VirtualInfrastructureType>
                   });

    #Specifying both ObjNameOrId and VirtualInfrastructureType
    $hypervisors = $vimsoapObj->get_hypervisors({
                             ObjNameOrId => <ObjNameOrId>,
                   VirtualInfrastructureType => <VirtualInfrastructureType>
                   });

    #List all including deleted hypervisors
    $hypervisors = $vimsoapObj->get_hypervisors({IncludeDeleted => 'true'});

    #listing the elements from the array reference, here is the example for
    #listing hypervisor names.
    foreach (@$hypervisors) {
        print $_->{HypervisorName};
        print $_->{IpAddress};
    }

    #function will set error, if there is an error in getting hypervisors.
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }


=cut

################################################################################
# Subroutine Name: get_hypervisors()
# Objective:       Get the Hypervisors from DFM
################################################################################
sub get_hypervisors {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->ViHypervisorListInfoIterStart($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag      = $result->get_Tag();
    my $max      = $result->get_Records() . int;
    my $resulthv = $interface->ViHypervisorListInfoIterNext(
        { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resulthv . "\n" );

    if ( not $resulthv ) {
        my $error = _check_for_error($resulthv);
        $self->_clear_err();
        $self->_set_err($error);
        return $resulthv;
    } ## end if ( not $resulthv )
    my $hvInfo = $resulthv->get_Hypervisors();
    my $hv     = $hvInfo->get_Hypervisor();
    my @hvList;

    foreach (@$hv) {
        my %hvDetails;
        $hvDetails{DatacenterName}    = $_->get_DatacenterName();
        $hvDetails{DomainName}        = $_->get_DomainName();
        $hvDetails{HypervisorId}      = $_->get_HypervisorId();
        $hvDetails{HypervisorName}    = $_->get_HypervisorName();
        $hvDetails{IpAddress}         = $_->get_IpAddress();
        $hvDetails{VirtualCenterName} = $_->get_VirtualCenterName();
        $hvDetails{VirtualInfrastructureType}
            = $_->get_VirtualInfrastructureType();
        push( @hvList, \%hvDetails );
    } ## end foreach (@$hv)
    $interface->ViHypervisorListInfoIterEnd( { Tag => $tag } );
    return \@hvList;
} ## end sub get_hypervisors

=head2 B<get_virtualdisks(I<%args>)>

This function gets the virtualdisk information from UM.

=over 4

=item Inputs

=item * %args [optional]

provide the "ObjNameOrId" as key , the values can be
 I<Virtual Machine, Datastore>.

OR

provide boolean value  as "true" for key "IncludeDeleted" to List all
including deleted virtualdisks.

=item Returns

Returns a reference to an array.Values in array contains references to
hashs of the virtualdisks(refer examples).
virtualdisk attributes are I<DatastoreId, DatastoreName, VirtualDiskId,
VirtualDiskName, VirtualInfrastructureType>
I<VirtualMachineId, VirtualMachineName>

Returns I<undef> if there is error in getting virtualdisks.Error will be
 printed by library in logs(refer examples)

=item Examples

    $unsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #Not specifying any arguments.
    $virtualdisks = $vimsoapObj->get_virtualdisks();

    #Specifying ObjNameOrId
    $virtualdisks = $vimsoapObj->get_virtualdisks({
                                       ObjNameOrId => <ObjNameOrId>
                                });

    #List all including deleted virtual disks
    $virtualdisks = $vimsoapObj->get_virtualdisks({IncludeDeleted => 'true'});

    #listing the elements from the array reference, here is the example for
    #listing virtualdisk names.
    foreach (@$virtualdisks) {
        print $_->{VirtualDiskName};
        print $_->{VirtualInfrastructureType};
    }

    #function will set error, if there is an error in getting virtual disks.
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: get_virtualdisks()
# Objective:       Get the virtual disks from DFM
################################################################################
sub get_virtualdisks {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->ViVirtualDiskListInfoIterStart($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag           = $result->get_Tag();
    my $max           = $result->get_Records() . int;
    my $resultDiskGet = $interface->ViVirtualDiskListInfoIterNext(
        { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultDiskGet . "\n" );

    if ( not $resultDiskGet ) {
        my $error = _check_for_error($resultDiskGet);
        $self->_clear_err();
        $self->_set_err($error);
        return $resultDiskGet;
    } ## end if ( not $resultDiskGet)
    my $virtualDiskInfo = $resultDiskGet->get_VirtualDisks();
    my $virtualDisk     = $virtualDiskInfo->get_VirtualDisk();
    my @VirtualDiskList;

    foreach (@$virtualDisk) {
        my %VirtualDiskDetails;
        $VirtualDiskDetails{DatastoreId}     = $_->get_DatastoreId();
        $VirtualDiskDetails{DatastoreName}   = $_->get_DatastoreName();
        $VirtualDiskDetails{VirtualDiskId}   = $_->get_VirtualDiskId();
        $VirtualDiskDetails{VirtualDiskName} = $_->get_VirtualDiskName();
        $VirtualDiskDetails{VirtualInfrastructureType}
            = $_->get_VirtualInfrastructureType();
        $VirtualDiskDetails{VirtualMachineId} = $_->get_VirtualMachineId();
        $VirtualDiskDetails{VirtualMachineName}
            = $_->get_VirtualMachineName();
        push( @VirtualDiskList, \%VirtualDiskDetails );
    } ## end foreach (@$virtualDisk)
    $interface->ViVirtualDiskListInfoIterEnd( { Tag => $tag } );
    return \@VirtualDiskList;
} ## end sub get_virtualdisks

=head2 B<get_virtualcenters(I<%args>)>

This function gets the virtualcenter information from UM.

=over 4

=item Inputs

=item * %args[optional]

provide the "ObjNameOrId" as key , the values can be
 I<Virtual Machine, Datastore, Hypervisor, Datacenter>.

OR

provide boolean value  as "true" for key "IncludeDeleted" to
List all including deleted virtual centers.

=item Returns

Returns a reference to an array.Values in array contains references to
hashs of the virtualcenters(refer examples).
virtualcenter attributes are I<VirtualCenterId, VirtualCenterName>

Returns I<undef> if there is error in getting virtualcenters.Error will be
printed by library in logs(refer examples)

=item Examples

    $unsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #get the virtual centers
    $virtualcenters = $vimsoapObj->get_virtualcenters();

    #get the virtual centers for Object Name.
    $virtualcenters = $vimsoapObj->get_virtualcenters({
                                                 ObjNameOrId => <ObjNameOrId>
                      });

    #List all including deleted virtual centers.
     $virtualcenters = $vimsoapObj->get_virtualcenters({
                                                      IncludeDeleted => 'true'
                        });

    #listing the elements from the array reference, here is the example
    #for listing virtualcenter names.
    foreach (@$virtualcenters) {
        print $_->{VirtualCenterName};
        print $_->{VirtualCenterId};
    }

    #function will set error, if there is an error in getting virtual centers.
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }
=cut

################################################################################
# Subroutine Name: get_virtualcenters()
# Objective:       Get the virtual centers from DFM
################################################################################
sub get_virtualcenters {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->ViVirtualCenterListInfoIterStart($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag           = $result->get_Tag();
    my $max           = $result->get_Records() . int;
    my $resultVcenter = $interface->ViVirtualCenterListInfoIterNext(
        { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultVcenter . "\n" );

    if ( not $resultVcenter ) {
        my $error = _check_for_error($resultVcenter);
        $self->_clear_err();
        $self->_set_err($error);
        return $resultVcenter;
    } ## end if ( not $resultVcenter)
    my $VcenterInfo = $resultVcenter->get_VirtualCenters();
    my $Vcenter     = $VcenterInfo->get_VirtualCenter();
    my @vcenterList;

    foreach (@$Vcenter) {
        my %vcenterDetails;
        $vcenterDetails{VirtualCenterId}   = $_->get_VirtualCenterId();
        $vcenterDetails{VirtualCenterName} = $_->get_VirtualCenterName();
        push( @vcenterList, \%vcenterDetails );
    } ## end foreach (@$Vcenter)
    $interface->ViVirtualCenterListInfoIterEnd( { Tag => $tag } );
    return \@vcenterList;
} ## end sub get_virtualcenters

=head2 B<get_datacenters(I<%args>)>

This function gets the datacenter information from UM.

=over 4

=item Inputs

=item * %args [optional]

provide the "ObjNameOrId" as key , the values can be
I<Virtual Machine, Datastore, Hypervisor, Datacenter>.

OR

provide boolean value  as "true" for key "IncludeDeleted" to List all
 including deleted datacenters.

=item Returns

Returns a reference to an array.Values in array contains references
to hashs of the datacenters(refer examples).
datacenter attributes are I<DatacenterId, DatacenterName>

Returns I<undef> if there is error in getting datacenters.Error will be
printed by library in logs(refer examples)

=item Examples

    $unsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #Not specifying any arguments.
    $datacenters = $vimsoapObj->get_datacenters();

    #Specifying ObjNameOrId
    $datacenters = $vimsoapObj->get_datacenters({
                                       ObjNameOrId => <ObjNameOrId>
                               });

    #List all including deleted datacenters.
    $datacenters = $vimsoapObj->get_datacenters({
                                           IncludeDeleted => 'true'
                                       });


    #listing the elements from the array reference, here is the
    #example for listing datacenter names.
    foreach (@$datacenters) {
        print $_->{DatacenterName};
        print $_->{DatacenterId};
    }

    #function will set error, if there is an error in getting datacenters.
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: get_datacenters()
# Objective:       Get the datacenters from DFM
################################################################################
sub get_datacenters {
    my ( $self, $args ) = @_;
    $self->_clear_err();
    my $result = $interface->ViDatacenterListInfoIterStart($args);

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag              = $result->get_Tag();
    my $max              = $result->get_Records() . int;
    my $resultDatacenter = $interface->ViDatacenterListInfoIterNext(
        { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultDatacenter . "\n" );

    if ( not $resultDatacenter ) {
        my $error = _check_for_error($resultDatacenter);
        $self->_clear_err();
        $self->_set_err($error);
        return $resultDatacenter;
    } ## end if ( not $resultDatacenter)
    my $datacenterInfo = $resultDatacenter->get_Datacenters();
    my $datacenter     = $datacenterInfo->get_Datacenter();
    my @datacenterList;

    foreach (@$datacenter) {
        my %datacenterDetails;
        $datacenterDetails{DatacenterId}   = $_->get_DatacenterId();
        $datacenterDetails{DatacenterName} = $_->get_DatacenterName();
        push( @datacenterList, \%datacenterDetails );
    } ## end foreach (@$datacenter)
    $interface->ViDatacenterListInfoIterEnd( { Tag => $tag } );
    return \@datacenterList;
} ## end sub get_datacenters

=head2 B<get_hypervclusters()>

This function gets the hypervcluster information from UM.

=over 4

=item Inputs

No Inputs

=item Returns

Returns a reference to an array.Values in array contains references to hashs
 of the hypervclusters(refer examples).
hypervcluster attributes are I<MsftHypervClusterId, MsftHypervClusterName>

Returns I<undef> if there is error in getting hypervclusters.Error will be
printed by library in logs(refer examples)

=item Examples

    $unsoapObj = new vimsoap(<hostname>,<username>,<password>);

    #get the hypervclusters
    $hypervclusters = $vimsoapObj->get_hypervclusters();

    #listing the elements from the array reference, here is the example for
    #listing hypervcluster names.
    foreach (@$hypervclusters) {
        print $_->{MsftHypervClusterName};
        print $_->{MsftHypervClusterId};
    }

    #function will set error, if there is an error in getting hyperv clusters.
        if ($vimsoap->error) {
            $Test->nlog($vimsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: get_hypervclusters()
# Objective:       Get the hyperv clusters from DFM
################################################################################
sub get_hypervclusters {
    my ($self) = @_;
    $self->_clear_err();
    my $result = $interface->ViMsftHypervClusterListInfoIterStart();

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag        = $result->get_Tag();
    my $max        = $result->get_Records() . int;
    my $resultMsft = $interface->ViMsftHypervClusterListInfoIterNext(
        { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultMsft . "\n" );

    if ( not $resultMsft ) {
        my $error = _check_for_error($resultMsft);
        $self->_clear_err();
        $self->_set_err($error);
        return $resultMsft;
    } ## end if ( not $resultMsft )
    my $MsftInfo = $resultMsft->get_MsftHypervClusters();
    my $Msft     = $MsftInfo->get_MsftHypervCluster();
    my @msftList;

    foreach (@$Msft) {
        my %msftDetails;
        $msftDetails{MsftHypervClusterId}   = $_->get_MsftHypervClusterId();
        $msftDetails{MsftHypervClusterName} = $_->get_MsftHypervClusterName();
        push( @msftList, \%msftDetails );
    } ## end foreach (@$Msft)
    $interface->ViMsftHypervClusterListInfoEnd( { Tag => $tag } );
    return \@msftList;
} ## end sub get_hypervclusters

=head2 B<get_hsplugin_info(I<ObjNameOrId>)>

This function gets the information HS plugins from UM.

=over 4

=item Inputs

=item * ObjNameOrId [mandatory]

provide "DpAgentNameOrId".

=item Returns

Returns a reference hash to plugin details(refer examples).
Plugin attributes are I<PluginType, PluginVersion>

Returns I<undef> if there is error in getting plugin info.
Error will be printed by library in logs(refer examples)

=item Examples

    $unsoapObj = new umsoap(<hostname>,<username>,<password>);

    #if we are specifying DpAgentNameOrId
    $pluginInfo = $umsoapObj->get_hsplugin_info()(<DpAgentNameOrId>);

    #listing the plugin info.

    print $$pluginInfo{PluginType};
    print $$pluginInfo{PluginVersion};

    #function will set error, if there is an error in getting plugins.
        if ($umsoap->error) {
            $Test->nlog($umsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: get_hsplugin_info()
# Objective:       Get plugin details for HS
################################################################################
sub get_hsplugin_info {
    my ( $self, $objName ) = @_;
    $self->_clear_err();
    $Log->debug( "object name :" . $objName . "\n" );
    my $result
        = $interface->HostServiceListInfoIterStart( { ObjNameOrId => $objName } );

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag           = $result->get_Tag();
    my $max           = $result->get_Records() . int;
    my $resultDpAgent = $interface->HostServiceListInfoIterNext(
        { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultDpAgent . "\n" );

    if ( not $resultDpAgent ) {
        my $error = _check_for_error($resultDpAgent);
        $self->_clear_err();
        $self->_set_err($error);
        return $resultDpAgent;
    } ## end if ( not $resultDpAgent)
    my $DpAgentsInfo = $resultDpAgent->get_HostServices()->get_HostServiceInfo()
        ->get_HostServicePlugins()->get_HostServicePluginInfo();
    my %agentPluginDetails;

    foreach (@$DpAgentsInfo) {

        $agentPluginDetails{PluginType}    = $_->get_PluginType();
        $agentPluginDetails{PluginVersion} = $_->get_PluginVersion();
    }
    $interface->HostServiceListInfoIterEnd( { Tag => $tag } );
    return \%agentPluginDetails;

} ## end sub get_hsplugin_info

=head2 B<get_volumes(I<ObjectNameOrId>)>

This function gets the volume details from UM.

=over 4

=item Inputs

=item * ObjectNameOrId

provide virtual objects like "Virtual Machine, vCenter, Datastore, Virtual Disk, Hypervisor".

=item Returns

Returns a reference hash to volume details(refer examples).
Volume attributes are I<VolumeName, VolumeId>

Returns I<undef> if there is error in getting plugin info.
Error will be printed by library in logs(refer examples)

=item Examples

    $umsoapObj = new umsoap(<hostname>,<username>,<password>);

    #if we are specifying ObjectNameOrId
    $volumeInfo = $umsoapObj->get_volumes(<ObjectNameOrId>);

    #listing the volume info.
    foreach (@$volumeInfo ) {
        print $_->{VolumeName};
        print $_->{VolumeId};

    }

    #function will set error, if there is an error in getting volumes.
        if ($umsoap->error) {
            $Test->nlog($umsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: get_volumes()
# Objective:       Get the volumes discovered by DFM
################################################################################

sub get_volumes {
    my ( $self, $objName ) = @_;
    $self->_clear_err();
    $Log->debug( "object name :" . $objName . "\n" );
    my $result = $interface->VolumeListInfoIterStart(
        { ObjectNameOrId => $objName } );

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag          = $result->get_Tag();
    my $max          = $result->get_Records() . int;
    my $resultVolume = $interface->VolumeListInfoIterNext(
        { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultVolume . "\n" );

    if ( not $resultVolume ) {
        my $error = _check_for_error($resultVolume);
        $self->_set_err($error);
        return $resultVolume;
    }
    my $volumes    = $resultVolume->get_Volumes();
    my $volumeInfo = $volumes->get_VolumeInfo();
    my @volumeList;

    foreach (@$volumeInfo) {
        my %volumeDetails;
        $volumeDetails{VolumeName} = $_->get_VolumeName();
        $volumeDetails{VolumeId}   = $_->get_VolumeId();
        push( @volumeList, \%volumeDetails );
    } ## end foreach (@$volumeInfo)
    $interface->VolumeListInfoIterEnd( { Tag => $tag } );
    return \@volumeList;
} ## end sub get_volumes

=head2 B<get_aggregates(I<ObjectNameOrId>)>

This function gets the aggregate details from UM.

=over 4

=item Inputs

=item * ObjectNameOrId

provide virtual objects like "Virtual Machine, vCenter, Datastore, Virtual Disk, Hypervisor".

=item Returns

Returns a reference hash to aggregate details(refer examples).
Aggregate attributes are I<AggregateName, AggregateId>

Returns I<undef> if there is error in getting plugin info.
Error will be printed by library in logs(refer examples)

=item Examples

    $umsoapObj = new umsoap(<hostname>,<username>,<password>);

    #if we are specifying ObjectNameOrId
    $aggrInfo = $umsoapObj->get_aggregates(<ObjectNameOrId>);

    #listing the volume info.
    foreach (@$aggrInfo ) {
        print $_->{AggregateName};
        print $_->{AggregateId};

    }

    #function will set error, if there is an error in getting aggregates.
        if ($umsoap->error) {
            $Test->nlog($umsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: get_aggregates()
# Objective:       Get the aggregates discovered by DFM
################################################################################

sub get_aggregates {
    my ( $self, $objName ) = @_;
    $self->_clear_err();
    $Log->debug( "object name :" . $objName . "\n" );
    my $result = $interface->AggregateListInfoIterStart(
        { ObjectNameOrId => $objName } );

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag        = $result->get_Tag();
    my $max        = $result->get_Records() . int;
    my $resultAggr = $interface->AggregateListInfoIterNext(
        { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultAggr . "\n" );

    if ( not $resultAggr ) {
        my $error = _check_for_error($resultAggr);
        $self->_set_err($error);
        return $resultAggr;
    }
    my $aggrs    = $resultAggr->get_Aggregates();
    my $aggrInfo = $aggrs->get_AggregateInfo();
    my @aggrList;

    foreach (@$aggrInfo) {
        my %aggrDetails;
        $aggrDetails{AggregateName} = $_->get_AggregateName();
        $aggrDetails{AggregateId}   = $_->get_AggregateId();
        push( @aggrList, \%aggrDetails );
    } ## end foreach (@$aggrInfo)
    $interface->AggregateListInfoIterEnd( { Tag => $tag } );
    return \@aggrList;
} ## end sub get_aggregates

=head2 B<get_luns(I<ObjectNameOrId>)>

This function gets the lun details from UM.

=over 4

=item Inputs

=item * ObjectNameOrId

provide virtual objects like "Virtual Machine, vCenter, Datastore, Virtual Disk, Hypervisor".

=item Returns

Returns a reference hash to lun details(refer examples).
Lun attributes are I<LunPath, LunId>

Returns I<undef> if there is error in getting plugin info.
Error will be printed by library in logs(refer examples)

=item Examples

    $umsoapObj = new umsoap(<hostname>,<username>,<password>);

    #if we are specifying ObjectNameOrId
    $lunInfo = $umsoapObj->get_luns(<ObjectNameOrId>);

    #listing the lun info.
    foreach (@$lunInfo ) {
        print $_->{LunPath};
        print $_->{LunId};

    }

    #function will set error, if there is an error in getting luns.
        if ($umsoap->error) {
            $Test->nlog($umsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: get_luns()
# Objective:       Get the luns discovered by DFM
################################################################################

sub get_luns {
    my ( $self, $objName ) = @_;
    $self->_clear_err();
    $Log->debug( "object name :" . $objName . "\n" );
    my $result
        = $interface->LunListInfoIterStart( { ObjectNameOrId => $objName } );

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag = $result->get_Tag();
    my $max = $result->get_Records() . int;
    my $resultLun
        = $interface->LunListInfoIterNext( { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultLun . "\n" );

    if ( not $resultLun ) {
        my $error = _check_for_error($resultLun);
        $self->_set_err($error);
        return $resultLun;
    }
    my $luns    = $resultLun->get_Luns();
    my $lunInfo = $luns->get_LunInfo();
    my @lunList;

    foreach (@$lunInfo) {
        my %lunDetails;
        $lunDetails{LunPath}  = $_->get_LunPath();
        $lunDetails{LunId}    = $_->get_LunId();
        $lunDetails{HostName} = $_->get_HostName();
        push( @lunList, \%lunDetails );
    } ## end foreach (@$lunInfo)
    $interface->LunListInfoIterEnd( { Tag => $tag } );
    return \@lunList;
} ## end sub get_luns

=head2 B<get_hosts(I<ObjectNameOrId>)>

This function gets the filer details from UM.

=over 4

=item Inputs

=item * ObjectNameOrId

provide virtual objects like "Virtual Machine, vCenter, Datastore, Virtual Disk, Hypervisor".

=item Returns

Returns a reference hash to filer details(refer examples).
filer attributes are I<HostName, HostId, HostFqdn>

Returns I<undef> if there is error in getting plugin info.
Error will be printed by library in logs(refer examples)

=item Examples

    $umsoapObj = new umsoap(<hostname>,<username>,<password>);

    #if we are specifying ObjectNameOrId
    $hostInfo = $umsoapObj->get_hosts(<ObjectNameOrId>);

    #listing the lun info.
    foreach (@$hostInfo ) {
        print $_->{HostName};
        print $_->{HostId};
        print $_->{HostFqdn};
    }

    #function will set error, if there is an error in getting filers.
        if ($umsoap->error) {
            $Test->nlog($umsoap->errmsg);
            return $TCD::FAIL;
        }

=cut

################################################################################
# Subroutine Name: get_hosts()
# Objective:       Get the hosts discovered by DFM
################################################################################

sub get_hosts {
    my ( $self, $objName ) = @_;
    $self->_clear_err();
    $Log->debug( "object name :" . $objName . "\n" );
    my $result
        = $interface->HostListInfoIterStart( { ObjectNameOrId => $objName } );

    if ( not $result ) {
        my $error = _check_for_error($result);
        $self->_set_err($error);
        return $result;
    }
    my $tag        = $result->get_Tag();
    my $max        = $result->get_Records() . int;
    my $resultHost = $interface->HostListInfoIterNext(
        { Maximum => $max, Tag => $tag } );
    $Log->debug( "result :" . $resultHost . "\n" );

    if ( not $resultHost ) {
        my $error = _check_for_error($resultHost);
        $self->_set_err($error);
        return $resultHost;
    }
    my $hosts    = $resultHost->get_Hosts();
    my $hostInfo = $hosts->get_HostInfo();
    my @hostList;

    foreach (@$hostInfo) {
        my %hostDetails;
        $hostDetails{HostName} = $_->get_HostName();
        $hostDetails{HostId}   = $_->get_HostId();
        $hostDetails{HostFqdn} = $_->get_HostFqdn();
        push( @hostList, \%hostDetails );
    } ## end foreach (@$hostInfo)
    $interface->HostListInfoIterEnd( { Tag => $tag } );
    return \@hostList;
} ## end sub get_hosts

=head2 B<compare(I<array1, array2>)>

This function compares two arrays.

=over 4

=item Inputs

=item * array1 [mandatory]

provide array1 which is superset, which means, From this array, elements will be
compared with elements of array2.

=item * array2 [mandatory]

provide array2 which is subset of array1.

=item Returns

Returns true, if elements of array1 matches with array2.
else returns false.


=item Examples

    if(compare($vmsVcenter, \@vmsVmware)) {
        return $TCD::PASS;
    }
    return $TCD::FAIL;
=cut

################################################################################
# Subroutine Name: compare()
# Objective:       Compare two arrays
################################################################################
sub compare() {
    $Log->comment("Entering compare function");
    my ( $self, $array1, $array2 ) = @_;
    $Log->debug("Printing array1");
    $Log->debug(@$array1);
    $Log->debug("Printing array2");
    $Log->debug(@$array2);
    my %temp = map { $_, $_ } @$array2;
    my @notFound1;

    foreach (@$array1) {
        if ( not exists $temp{$_} ) {
            push( @notFound1, $_ );
        }
    }
    %temp = map { $_, $_ } @$array1;
    my @notFound2;

    foreach (@$array2) {
        if ( not exists $temp{$_} ) {
            push( @notFound2, $_ );
        }
    }

    if ( scalar @notFound1 > 0 || scalar @notFound2 > 0 ) {
        $Log->comment("The following objects are not found");

        foreach (@notFound1) {
            $Log->comment($_);
        }

        foreach (@notFound2) {
            $Log->comment($_);
        }
        $Log->comment("Exiting compare function with false");
        return 0;
    } else {
        $Log->comment("Exiting compare function with true");
        return 1;
    }
} ## end sub compare

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _check_for_error()
# Objective:       Get the fault details from message response
################################################################################
sub _check_for_error {
    my ($result) = @_;

    if ( not $result ) {
        my $e     = $result->get_faultstring();
        my $error = "Error calling SOAP method:" . _get_fault_string($e);
        return $error;
    }
} ## end sub _check_for_error

################################################################################
# Subroutine Name: _get_fault_detail()
# Objective:       Get the fault details from message response
################################################################################
sub _get_fault_string {
    my ($errorMessage) = @_;
    my $xmlData;

    if ( ($xmlData) = $errorMessage =~ m/Error.*\n.*\n.*\n.*\n(.*\n.*)/ ) {
        ;
    } else {
        return $errorMessage;
    }

    # create object
    my $xml = new XML::Simple;

    # read XML data
    my $data = $xml->XMLin($xmlData);
    return $data->{'SOAP-ENV:Body'}->{'SOAP-ENV:Fault'}->{'faultstring'};
} ## end sub _get_fault_string

1;

