# $Id:
#
# Copyright (c) 2010 NetApp.
# All rights reserved.
#
#########################################################
## @summary Library of functions to get storage objects
##          details from the filer related to vCenter.
##
## @author Satish Hunasikatti (satishh@netapp.com)
##
## @status public
##
## @pod here
#########################################################

# Package Name
##############
package filerutils;

require Exporter;
@ISA = qw/Exporter/;

use Tharn;
use strict;
use NaServer;
use Data::Dumper;

# Package Requirements
######################
use Error;

# Declare global variables
use vars qw(@ISA);

@ISA = qw(Error);

=head NAME

filerutils - functions to get storage objects details from the filer related to vCenter.

=head1 SYNOPSIS

use filerutils;
my $filerutilsObj = new filerutils();

=head1 DESCRIPTION

This package offers the functions to fetch storage objects details from the filer
which are used in vCenter. filerutils object can be created as below,

my $filerutilsObj = new filerutils();

While creating object no need to specify the filer details here. The same can be
passed to each function in this package.

=head1 FUNCTIONS

Details are explained in each function as below.

=cut

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new filerutils object
################################################################################
sub new {
    my ($class) = shift;
    my $self = bless( {}, $class );
    return $self;
}

=head2 B<get_lun_info(I<$filers, $uuid>)>

This function gets the Lun information from filer.

=over 4

=item Inputs

=item * $filers [mandatory]

provide the multiple filer details as reference to array of hashes(refer examples).
filer details attributes are
I<filerName, userName, password>

=item * $uuid [mandatory]

provide the uuid of vmfs datastore from vCenter.

=item Returns

Returns a reference to a hash. Values in hash contains lun details(refer examples).
Hash attributes are
I<filerName,lunPath, volume, aggr>

Returns I<undef> if there is error in getting lun details .Error will be printed
 by library in logs(refer examples)

=item Examples

	my $filerutilsObj = new filerutils();

	my @filerDetails;
    #specify uuid for the vmfs datastore
    my $uuid = <uuid>;
    my $filer1;
    my $filer2;

    #provide filer credentials
    $filer1->{filerName} = "filername";
    $filer1->{userName} = "username";
    $filer1->{password} = "password";

    $filer2->{filerName} = "filername";
    $filer2->{userName} = "username";
    $filer2->{password} = "password";

    push(@filerDetails, $filer1);
    push(@filerDetails, $filer2);
	#pass array of hashes contaning filer credentials and uuid as input to this function.

	my $lunDetails = $filerutilsObj->get_lun_info(\@filerDetails, $uuid);

	#get the lun details.
	my $lunPath =  $lunDetails->{lunPath};
	my $volume = $lunDetails->{volume};
	my $filerName = $lunDetails->{filerName};
	my $aggr = $lunDetails->{aggr};

    #function will set error, if there is an error in getting lun details.
        if ($filerutilsObj->error) {
            $Test->nlog($filerutilsObj->errmsg);
            return $TCD::FAIL;
        }


=cut

################################################################################
# Subroutine Name: get_lun_info()
# Objective:       Get the Lun related information from Lun serial number.
################################################################################
sub get_lun_info {
    my ( $self, $filers, $uuid ) = @_;
    my $lunDetails = undef;
    $self->_clear_err();

    # Characters 18-24 is hex for Netapp Lun serial number
    my $lunsn = pack( "H*", substr( $uuid, 18, 24 ) );
    $Log->debug("uuid: $uuid");
    $Log->debug("Lun serial No: $lunsn");

    foreach (@$filers) {
        my $s = NaServer->new( $_->{filerName}, 1, 1 );
        $s->set_admin_user( $_->{userName}, $_->{password} );
        $Log->debug( $_->{filerName} );
        $Log->debug( $_->{userName} );
        $Log->debug( $_->{password} );
        my $result = $s->invoke( "lun-get-vdisk-attributes", "serial-number",
            $lunsn );

        if ( $result->child_get("path") && $result->child_get("path")->{content} ) {
            my $filerName = $result->child_get("filer-name")->{content};
            $Log->debug("filername from api : $filerName");
            my $lunPath = $result->child_get("path")->{content};
            $Log->debug("lun path from api : $lunPath");
            $lunDetails->{filerName} = $filerName;
            $lunDetails->{lunPath}   = $lunPath;
            $lunDetails->{volume}    = _get_volume($lunPath);
            $Log->debug("volume from api : $lunDetails->{volume}");
            $lunDetails->{aggr}
                = _get_aggr( $self, $s, $lunDetails->{volume} );
            $Log->debug("aggr from api : $lunDetails->{aggr}");
            return $lunDetails;
        } ## end if ( $result->child_get...)
    } ## end foreach (@$filers)
    $self->_set_err("No matching lun in the filers for $uuid");
    $Log->log("No matching lun in the filers for $uuid");
    return $lunDetails;
} ## end sub get_lun_info

=head2 B<get_aggr(I<$filer, $volName>)>

This function gets the aggregate which contains the specified volume.

=over 4

=item Inputs

=item * $args [mandatory]

provide the filer details as reference to array of hashes(refer examples).
filer details attributes are
I<filerName, userName, password>

=item * $volName [mandatory]

provide the volume name present in the specified filer.

=item Returns

Returns aggregate name.

Returns I<undef> if there is error in getting aggregate name .Error will be printed
 by library in logs(refer examples)

=item Examples

	my $filerutilsObj = new filerutils();

	my @filerDetails;
    #specify volume name for the vmfs datastore
    my $volName = <volume name>;
    my $filer1;

    #provide filer credentials
    $filer1->{filerName} = "filername";
    $filer1->{userName} = "username";
    $filer1->{password} = "password";

    push(@filerDetails, $filer1);

	#pass array of hashes contaning filer credentials and uuid as input to this function.
	my $aggrName = $filerutilsObj->get_aggr($filer1, $volName);

    #function will set error, if there is an error in getting lun details.
        if ($filerutilsObj->error) {
            $Test->nlog($filerutilsObj->errmsg);
            return $TCD::FAIL;
        }


=cut

################################################################################
# Subroutine Name: get_aggr()
# Objective:       Get the aggregate which contains the specified volume.
################################################################################
sub get_aggr {
    my ( $self, $filer, $volName ) = @_;
    $self->_clear_err();
    my $s = NaServer->new( $filer->{filerName}, 1, 1 );
    $s->set_admin_user( $filer->{userName}, $filer->{password} );
    return _get_aggr( $self, $s, $volName );
} ## end sub get_aggr

=head2 B<vfiler_status(I<$filer)>

This function checks whether given filer is vFiler or Not.

=over 4

=item Inputs

=item * $filer [mandatory]

provide the filer name or IP(refer examples).

=item Returns

Returns I<true> if  filer is vFiler else returns I<false>.(refer examples).

Returns I<undef> if there is error in getting lun details .Error will be printed
 by library in logs(refer examples)

=item Examples

	my $filerutilsObj = new filerutils();

	my $filer = <filer name or IP>;

	my $isVfiler = $filerutilsObj->vfiler_status($filer);
    if($isVfiler) {
		$Test->nlog("$filer is vFiler");
	} else {
		$Test->nlog("$filer is not a vFiler");
	}

    #function will set error, if there is an error in getting vfiler status.
        if ($filerutilsObj->error) {
            $Test->nlog($filerutilsObj->errmsg);
            return $TCD::FAIL;
        }


=cut

################################################################################
# Subroutine Name: vfiler_status()
# Objective:       Checks whether given host is vFiler or Not.
################################################################################
sub vfiler_status {
    my ( $self, $filer ) = @_;
    $self->_clear_err();
    my $s = NaServer->new( $filer->{filerName}, 1, 1 );
    $s->set_admin_user( $filer->{userName}, $filer->{password} );
	my $result = $s->invoke("system-get-version");

    if ( $result->results_status() eq "failed" ) {
        $self->_set_err( $result->results_reason() );
        $Log->log($result->results_reason());
        return undef;
    }
   my $version = $result->child_get("version")->{content};

   if($version =~ /.*Multistore$/) {
		return 1;
   }
   else {
		return 0;
   }
}

=head2 B<vfiler_to_filer_map(I<$filers>)>

This function retrieves the vfiler to physical filer mapping.

=over 4

=item Inputs

=item * $filers [mandatory]

provide the multiple filer details as reference to array of hashes(refer examples).
filer details attributes are
I<filerName, userName, password>

=item Returns

Returns a reference to a hash. Values in hash contains vfiler to
physical filer mapping(refer examples).

In Hash keys are vfilers and values are corresponding physical filers.

Returns I<undef> if there is error in getting vfiler maping details .Error will be printed
 by library in logs(refer examples)

=item Examples

	my $filerutilsObj = new filerutils();

	my @filerDetails;
    my $filer1;
    my $filer2;

    #provide filer credentials
    $filer1->{filerName} = "filername";
    $filer1->{userName} = "username";
    $filer1->{password} = "password";

    $filer2->{filerName} = "filername";
    $filer2->{userName} = "username";
    $filer2->{password} = "password";

    push(@filerDetails, $filer1);
    push(@filerDetails, $filer2);
	#pass array of hashes contaning filer credentials.

	my $vfilerMaping = $filerutilsObj->vfiler_to_filer_map(\@filerDetails);

	#get the physical filer for vfiler.
	my $vfiler = <vfiler name>;
	my $physicalFiler = $vfilerMaping->{$vfiler};

    #function will set error, if there is an error in getting vfiler maping details.
        if ($filerutilsObj->error) {
            $Test->nlog($filerutilsObj->errmsg);
            return $TCD::FAIL;
        }


=cut

################################################################################
# Subroutine Name: vfiler_to_filer_map()
# Objective:       This function retrieves the vfiler to physical filer mapping.
################################################################################
sub vfiler_to_filer_map {
    my ( $self, $filers ) = @_;
    $self->_clear_err();
    my $vfiler;

    foreach (@$filers) {
        if ( not vfiler_status($self, $_) ) {
            my $s = NaServer->new( $_->{filerName}, 1, 1 );
            $s->set_admin_user( $_->{userName}, $_->{password} );
            my $n = $s->invoke("system-get-info");
            if ( $n->results_status() eq "failed" ) {
                $self->_set_err( $n->results_reason() );
                $Log->log($n->results_reason());
                return undef;
            }

            my $p = $n->child_get("system-info");
            my $systemName = $p->child_get_string("system-name");
            my $result = $s->invoke("vfiler-list-info");

            if ( $result->results_status() eq "failed" ) {
                $self->_set_err( $result->results_reason() );
                $Log->log($result->results_reason());
                return undef;
            }
            my $v      = $result->child_get("vfilers");
            my @info   = $v->children_get();

            foreach (@info) {
                my $vf = $_->child_get_string("name");
                $vfiler->{$vf} = $systemName;
            }
        } ## end if ( not vfiler_status...)
    } ## end foreach (@$filers)

    return $vfiler;
} ## end sub vfiler_to_filer_map

=head2 B<get_hostname(I<$filer,$username,$password>)>

This function retrieves the hostname set on the filer

=over 4

=item Inputs

filer ip address,username & password

=item Returns

Returns the hostname set on the filer
=item Examples

	my $filerutilsObj = new filerutils();

=cut

sub get_hostname {
    my $self   = shift(@_);
    my $filer = shift(@_);
    my $username = shift(@_);
    my $password = shift(@_);
    $self->_clear_err();
	
    my $na = NaServer->new($filer, 1, 1 );
    $na->set_admin_user($username, $password);
    my $result = $na->invoke("system-get-info");
    if ( $result->results_status() eq "failed" ) {
        $self->_set_err( $result->results_reason());
        $Log->log($result->results_reason());
        return undef;
    }

    my $child = $result->child_get("system-info");
    my $name = $child->child_get_string("system-name");
    return $name;	
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _get_volume()
# Objective:       Get the Volume name from the Lun Path
################################################################################

sub _get_volume {
    my $lunPath = shift;

    if ( $lunPath =~ /(\/vol\/\w*)\/\w*/ ) {
        return $1;
    }
} ## end sub _get_volume

################################################################################
# Subroutine Name: _get_aggr()
# Objective:       Get the aggregate for the Volume
################################################################################

sub _get_aggr {
    my ( $self, $s, $volName ) = @_;
    my $result = $s->invoke( "volume-container", "volume", $volName );

    if ( $result->results_status() eq "failed" ) {
        $self->_set_err( $result->results_reason() );
        $Log->log($result->results_reason());
        return undef;
    }
    return $result->child_get("containing-aggregate")->{content};
} ## end sub _get_aggr
