# $Id:$
#
#***********************************************************************
#
# NAME
#   vmware.pm -  programmatic interface to list vCenter information.
#
# SYNOPSIS
#  vCenter api for validation
#
# DESCRIPTION
#   This Perl module provides interfaces to query vCenter.
#
# AUTHOR
#    nachiket ( nachiket@netapp.com)
#
#***********************************************************************

package vmware;

require Exporter;
@ISA    = qw/Exporter/;
@EXPORT = qw/get_virtualmachines
    get_datastores
    get_datacenters
    get_hostsystems
    get_virtualdisks
    get_vmdkfile
    add_virtualdisk
    create_datastore
    delete_datastore
    delete_virtualdisk
    rename_datacenter
    create_datacenter
    delete_datacenter
    vm_power rename_vm
    delete_vm
    update_hostsystem
    get_vigraph
    get_vi_storage_array
    get_vi_storage_graph
    discover_datastore
    logout/;

use Tharn;
use Error;
use VMware::VIRuntime;
use strict;
use warnings;
use filerutils;
use Socket;
use Data::Dumper;
use vars qw(@ISA $vmware);
@ISA = qw(Error);

=head2 new

This subroutine is used to create vmware object.

B<Input:>

url, usrname,password

B<Returns:>

vmware object

B<Usage:>

new vmware("<server name or IP of vCenter>","<username>","<password>");

B<Example:>

my $test = new vmware("<server name or IP of of vCenter>","<username>","<password>");

=cut

sub new {
    my ( $class, $server, $username, $password ) = @_;
    my $self = {};
    bless( $self, $class );

    eval {
        $vmware = Vim::login(
            service_url => "https://$server/sdk/vimService",
            user_name   => $username,
            password    => $password
        );
    };
    if ($@) {
        $Log->comment($@);
        $self->_clear_err();
        $self->_set_err($@);
    }

    return $self;
} ## end sub new

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _get_vi_views()
# Objective:       Gets the list of VI objects based on the type of managed object
################################################################################

sub _get_vi_views() {
    my $self    = shift(@_);
    my $vi_view = shift(@_);
    my $view;
    eval { $view = $vmware->find_entity_views( view_type => $vi_view ); };
    if ($@) {
        $Log->comment($@);
        $self->_clear_err();
        $self->_set_err($@);
    }
    return $view;
} ## end sub _get_vi_views

################################################################################
# Subroutine Name: _get_vi_view()
# Objective:       Gets the VI objects based on the type of managed object and filter
################################################################################

sub _get_vi_view {
    my $self    = shift(@_);
    my $vi_view = shift(@_);
    my $vi_name = shift(@_);
    my $view;
    eval {
        $view = $vmware->find_entity_view(
            view_type => $vi_view,
            filter    => { name => $vi_name }
        );
    };
    if ($@) {
        $Log->comment($@);
        $self->_clear_err();
        $self->_set_err($@);
    }
    return $view;
} ## end sub _get_vi_view

################################################################################
# Subroutine Name: _get_vi_views()
# Objective:       Gets the list of managed object reference
################################################################################

sub _get_moref {
    my $self = shift(@_);
    my $mof  = shift(@_);
    my $view;
    eval { $view = $vmware->get_views( mo_ref_array => $mof ); };
    if ($@) {
        $Log->comment($@);
        $self->_clear_err();
        $self->_set_err($@);
    }
    return $view;
} ## end sub _get_moref

################################################################################
# Subroutine Name: _get_context_view()
# Objective:       Gets the  VI object based on a context
################################################################################

sub _get_context_view {
    my $self         = shift(@_);
    my $vi_view      = shift(@_);
    my $view_context = shift(@_);
    my $view;

    eval {
        $view = $vmware->find_entity_view(
            view_type    => $vi_view,
            begin_entity => $view_context
        );
    };
    if ($@) {
        $Log->comment($@);
        $self->_clear_err();
        $self->_set_err($@);
    }
    return $view;
} ## end sub _get_context_view

################################################################################
# Subroutine Name: _get_context_views()
# Objective:       Gets the  list of VI object based on a context
################################################################################

sub _get_context_views {
    my $self         = shift(@_);
    my $vi_view      = shift(@_);
    my $view_context = shift(@_);
    my $view;

    eval {
        $view = $vmware->find_entity_views(
            view_type    => $vi_view,
            begin_entity => $view_context
        );
    };
    if ($@) {
        $Log->comment($@);
        $self->_clear_err();
        $self->_set_err($@);
    }
    return $view;
} ## end sub _get_context_views

################################################################################
# Subroutine Name: _get_context_view_filter
# Objective:       Gets the VI object based on context and filter
################################################################################

sub _get_context_view_filter {
    my $self         = shift(@_);
    my $vi_view      = shift(@_);
    my $filter       = shift(@_);
    my $view_context = shift(@_);
    my $view;

    eval {
        $view = $vmware->find_entity_view(
            view_type    => $vi_view,
            filter       => { name => $filter },
            begin_entity => $view_context
        );
    };
    if ($@) {
        $Log->comment($@);
        $self->_clear_err();
        $self->_set_err($@);
    }
    return $view;
} ## end sub _get_context_view_filter

################################################################################
### public methods to be used by end user
################################################################################

=head2 get_datastores

This subroutine is used to get a list of datastores.
Datastores can be obtained at vCenter,Datacenter,ESX and Virtual Machine level.

B<Input:>

name of the Datacenter or ESX or Virtual Machine.
no parameters should be passed if you need datastores at vCenter level

B<Returns:>

returns an reference to an array of datastores.
Empty array will be returned if there are no datastores.

B<Usage:>

get_datastore()  (datastores for entire vCenter)
get_datastore("<Datacenter Name>")

B<Example:>

my $datastore = $test->get_datastores;

=cut

sub get_datastores() {
    my $self              = shift(@_);
    my $datacenter_esx_vm = shift(@_);
    my @datastores;
    my $datacenter_view;
    my $hostsystem_view;
    my $vm_view;
    $self->_clear_err();
    if ( !$datacenter_esx_vm ) {
        $datacenter_view = $self->_get_vi_views('Datacenter');
        $Log->debug("Discovering Datastores for entire vCenter");
        foreach (@$datacenter_view) {
            my $ds = $_->datastore;
            if ($ds) {
                my $datacenter_ds = $self->_get_moref( \@$ds );
                foreach (@$datacenter_ds) {
                    push( @datastores, $_->summary->name );
                }
            } ## end if ($ds)
        } ## end foreach (@$datacenter_view)
    } elsif ( $datacenter_view
        = $self->_get_vi_view( 'Datacenter', $datacenter_esx_vm ) )
    {
        $Log->debug(
            "Discovering Datastores for Datacenter : $datacenter_esx_vm");
        my $ds = $datacenter_view->datastore;
        if ($ds) {
            my $datacenter_ds = $self->_get_moref( \@$ds );
            foreach (@$datacenter_ds) {
                push( @datastores, $_->summary->name );
            }
        } ## end if ($ds)
    } elsif ( $hostsystem_view
        = $self->_get_vi_view( "HostSystem", $datacenter_esx_vm ) )
    {
        $Log->debug(
            "Discovering Datastores for HostSystem(ESX) : $datacenter_esx_vm"
        );
        if ( $hostsystem_view->datastore ) {
            my $hostsystem_ds = $hostsystem_view->datastore;
            my $ds            = $self->_get_moref( \@$hostsystem_ds );
            foreach (@$ds) {
                push( @datastores, $_->summary->name );
            }
        } ## end if ( $hostsystem_view->datastore)
    } elsif ( $vm_view
        = $self->_get_vi_view( "VirtualMachine", $datacenter_esx_vm ) )
    {
        $Log->debug(
            "Discovering Datastores for Virtual Machine : $datacenter_esx_vm"
        );
        my $vm_ds        = $vm_view->datastore;
        my $datastore_vm = $self->_get_moref( \@$vm_ds );
        foreach (@$datastore_vm) {
            push( @datastores, $_->summary->name );
        }
    } ## end elsif ( $vm_view = $self->_get_vi_view...)

    $Log->debug("Datastores discovered so far");
    $Log->debug(@datastores);
    return \@datastores;

} ## end sub get_datastores

=head2 get_datacenters

This subroutine gets all the Datacenters managed by vCenter

B<Input:>

none

B<Returns:>

returns an reference to an array of Datacenters names

B<Usage:>

get_datacenters()

B<Example:>

my $datacenters = $test->getDatacenters();

=cut

sub get_datacenters {
    my $self = shift(@_);
    my @datacenters;
    $self->_clear_err();
    my $datacenter_view = $self->_get_vi_views("Datacenter");

    foreach (@$datacenter_view) {
        if ( defined $_->name ) {
            push( @datacenters, $_->name );
        }
    }
    $Log->debug("Datacenters discovered so far");
    $Log->debug( Dumper(@datacenters) );
    return \@datacenters;
} ## end sub get_datacenters

=head2 get_virtualdisks()

This subroutine gets an list of virtual disks at an datastore/virtual machine level.

B<Input:>

name of the datastore or virtual machine

B<Returns:>

returns an reference to an array of virtual disks

B<Usage:>

get_virtualdisks("<datastore name>");
get_virtualdisks("<virtual machine>");

B<Example:>

my $datacenters = $test->get_virtualdisks();

=cut

sub get_virtualdisks {
    my $self         = shift(@_);
    my $vm_datastore = shift(@_);
    my $virtualmachine_view;
    my $datastore_view;
    my @virtualdisks;
    $self->_clear_err();
    if ( $virtualmachine_view
        = $self->_get_vi_view( "VirtualMachine", $vm_datastore ) )
    {
        if ( $virtualmachine_view->name eq $vm_datastore ) {

            my $devices = $virtualmachine_view->config->hardware->device;
            foreach my $device (@$devices) {
                if ( $device->deviceInfo->summary =~ /KB/ ) {
                    push( @virtualdisks, $device->deviceInfo->label );
                }
            }
        } ## end if ( $virtualmachine_view...)
    } elsif ( $datastore_view = $self->_get_vi_views('Datacenter') ) {
        foreach (@$datastore_view) {
            my $ds = $_->datastore;
            if ($ds) {
                my $datacenter_ds = $self->_get_moref( \@$ds );
                foreach (@$datacenter_ds) {
                    if ( $_->summary->name eq $vm_datastore ) {
                        my $vm_view = $self->_get_moref( $_->vm );
                        foreach (@$vm_view) {
                            my $devices = $_->config->hardware->device;
                            foreach my $device (@$devices) {
                                if ( $device->deviceInfo->summary =~ /KB/ ) {
                                    push( @virtualdisks,
                                        $device->deviceInfo->label );
                                }
                            } ## end foreach my $device (@$devices)
                        } ## end foreach (@$vm_view)
                    } ## end if ( $_->summary->name...)
                } ## end foreach (@$datacenter_ds)
            } ## end if ($ds)
        } ## end foreach (@$datastore_view)
    } else {
        $Log->comment("$vm_datastore not found");
        $self->_clear_err();
        $self->_set_err("$vm_datastore not found");
    }

    $Log->debug("Virtual Disks for discovered so far");
    $Log->debug(@virtualdisks);
    return \@virtualdisks;
} ## end sub get_virtualdisks

=head2 get_vmdkfile

This subroutine gets an list of VMDK files at an datastore/virtual machine level.

B<Input:>

name of the datastore or virtual machine

B<Returns:>

returns an reference to an array of virtual disks

B<Usage:>

get_vmdkfile("<datastore name>");
get_vmdkfile("<virtual machine>");

B<Example:>

my $datacenters = $test->get_virtualdisks("<vm name>");

=cut

sub get_vmdkfile {
    my $self         = shift(@_);
    my $vm_datastore = shift(@_);
    my $virtualmachine_view;
    my $datastore_view;
    my @virtualdisks;
    $self->_clear_err();
    if ( $virtualmachine_view
        = $self->_get_vi_view( "VirtualMachine", $vm_datastore ) )
    {
        $Log->debug("Virtual Disks for an VM:$vm_datastore ");
        if ( $virtualmachine_view->name eq $vm_datastore ) {
            my $layout = $virtualmachine_view->layout;
            if ($layout) {
                my $vmdk = $layout->disk;
                foreach (@$vmdk) {
                    my $disk = $_->{'diskFile'};
                    foreach (@$disk) {
                        push( @virtualdisks, $_ );
                    }
                } ## end foreach (@$vmdk)
            } ## end if ($layout)
        } ## end if ( $virtualmachine_view...)
    } else {
        $datastore_view = $self->_get_vi_views('Datacenter');
        $Log->debug("Virtual Disks for an datastore:$vm_datastore");
        foreach (@$datastore_view) {
            my $ds = $_->datastore;
            if ($ds) {
                my $datacenter_ds = $self->_get_moref( \@$ds );
                foreach (@$datacenter_ds) {
                    if ( $_->summary->name eq $vm_datastore ) {
                        my $vm_view = $self->_get_moref( $_->vm );
                        foreach (@$vm_view) {
                            my $layout = $_->layout;
                            if ($layout) {
                                my $vmdk = $layout->disk;
                                foreach (@$vmdk) {
                                    my $disk = $_->{'diskFile'};
                                    foreach (@$disk) {
                                        push( @virtualdisks, $_ );
                                    }
                                } ## end foreach (@$vmdk)
                            } ## end if ($layout)
                        } ## end foreach (@$vm_view)
                    } ## end if ( $_->summary->name...)
                } ## end foreach (@$datacenter_ds)
            } ## end if ($ds)
        } ## end foreach (@$datastore_view)
    } ## end else [ if ( $virtualmachine_view...)]
    $Log->debug("Virtual Disks for discovered so far");
    $Log->debug(@virtualdisks);
    return \@virtualdisks;
} ## end sub get_vmdkfile

=head2 get_virtualmachines()

This subroutine gets an list of virtual machines at an vCenter/datacenter/esx/datastore level

B<Input:>

name of the datastore or esx or datacenter
If no parameter is supplied, virtual machines for the entire vCenter are listed.

B<Returns:>

returns an reference to an array of virtual machines

B<Usage:>

get_virtualmachines()("<datastore name>");
get_virtualmachines()("<esx name>");
get_virtualmachines()("<datacenter name>");

B<Example:>

my $datacenters = $test->get_virtualmachines();

=cut

sub get_virtualmachines {
    my $self                     = shift(@_);
    my $datacenter_esx_datastore = shift(@_);
    my @virtualmachines;
    my $vcenter;
    my $datacenter_view;
    my $hostsystem_view;
    my $datastore_view;
    $self->_clear_err();

    if ( !$datacenter_esx_datastore ) {
        $Log->debug("Virtual Machines for the entire vCenter");
        $vcenter = $self->_get_vi_views('VirtualMachine');
        if ($vcenter) {
            foreach (@$vcenter) {
                push( @virtualmachines, $_->name );
            }
        }
    } elsif ( $datacenter_view
        = $self->_get_vi_view( "Datacenter", $datacenter_esx_datastore ) )
    {
        my $datacenter_vm_views
            = $self->_get_context_views( 'VirtualMachine', $datacenter_view );
        $Log->debug(
            "Virtual Machines for the entire Datacenter : $datacenter_view");
        if ($datacenter_vm_views) {
            foreach (@$datacenter_vm_views) {
                push( @virtualmachines, $_->name );
            }
        }
    } elsif ( $hostsystem_view
        = $self->_get_vi_view( "HostSystem", $datacenter_esx_datastore ) )
    {
        $Log->debug("Virtual Machines for the entire ESX : $datacenter_view");
        if ($hostsystem_view) {
            my $hostsystem_vm_views
                = $self->_get_context_views( 'VirtualMachine',
                $hostsystem_view );
            if ($hostsystem_vm_views) {
                foreach (@$hostsystem_vm_views) {
                    push( @virtualmachines, $_->name );
                }
            }
        } ## end if ($hostsystem_view)
    } else {
        $datastore_view = $self->_get_vi_views('Datacenter');
        $Log->debug("Virtual Machines for an Datastore : $datacenter_view");
        foreach (@$datastore_view) {
            my $ds = $_->datastore;
            if ($ds) {
                my $datacenter_ds = $self->_get_moref( \@$ds );
                foreach (@$datacenter_ds) {
                    if ( $_->summary->name eq $datacenter_esx_datastore ) {
                        my $vm_view = $self->_get_moref( $_->vm );
                        foreach (@$vm_view) {
                            push( @virtualmachines, $_->name );
                        }
                    } ## end if ( $_->summary->name...)
                } ## end foreach (@$datacenter_ds)
            } ## end if ($ds)
        } ## end foreach (@$datastore_view)
    } ## end else [ if ( !$datacenter_esx_datastore)]
    return \@virtualmachines;
} ## end sub get_virtualmachines

=head2 get_hostsystems()

This subroutine gets an list of ESX servers at an vCenter/datacenter level

B<Input:>

name of the datacenter
If no parameter is supplied, ESX for the entire vCenter are listed.

B<Returns:>

returns an reference to an array of ESX

B<Usage:>


get_hostsystems("<datacenter name>");
get_hostsystems();

B<Example:>

my $datacenters = $test->get_hostsystems()();

=cut

sub get_hostsystems() {
    my @hostsystems;
    my $self       = shift(@_);
    my $datacenter = shift(@_);
    my $datacenter_view;
    $self->_clear_err();
    if ( !$datacenter ) {
        $Log->debug("Host Systems (ESX) for the entire vCenter");
        my $hostsystem_view = $self->_get_vi_views('HostSystem');
        foreach (@$hostsystem_view) {
            push( @hostsystems, $_->name );
        }
    } elsif ( $datacenter_view
        = $self->_get_vi_view( "Datacenter", $datacenter ) )
    {
        $Log->debug("Host Systems (ESX) for Datacenter:$datacenter_view");
        my $hostsystem_view
            = $self->_get_context_views( 'HostSystem', $datacenter_view );
        foreach (@$hostsystem_view) {
            push( @hostsystems, $_->name );
        }
    } ## end elsif ( $datacenter_view ...)
    return \@hostsystems;
} ## end sub get_hostsystems

=head2 get_vigraph

This subroutine generates a hash of all the virtual inventory

B<Input:>

none

B<Returns:>

returns an reference to an hash

B<Usage:>

get_vigraph()

B<Example:>

my $datacenters = $test->get_vigraph();

=cut

sub get_vigraph() {
    my $self = shift(@_);
    my %vcenter;
    my $TYPE = 'NFS';
    $self->_clear_err();
    my $datacenters = $self->get_datacenters();

    #update datacenters
    foreach (@$datacenters) {
        $vcenter{$_} = undef;
    }

    my @vidatacenters = keys %vcenter;
    foreach (@vidatacenters) {
        my $datacenter = $_;
        my $esx        = $self->get_hostsystems($_);
        foreach (@$esx) {
            my $hostsystem = $_;
            if ( !$vcenter{$datacenter} ) {
                $vcenter{$datacenter} = { $_ => undef };
            } else {
                my $esx = $vcenter{$datacenter};
                $esx->{$_} = undef;
            }
            my $vm = $self->get_virtualmachines($hostsystem);
            foreach (@$vm) {
                my $vm_name = $_;
                if ( !$vcenter{$datacenter}{$hostsystem} ) {
                    $vcenter{$datacenter}{$hostsystem} = { $_ => undef };
                } else {
                    my $vms = $vcenter{$datacenter}{$hostsystem};
                    $vms->{$_} = undef;
                }
                my $virtualdisks = $self->get_vmdkfile($vm_name);
                foreach (@$virtualdisks) {
                    my $vmdk_name = $_;
                    if ( !$vcenter{$datacenter}{$hostsystem}{$vm_name} ) {
                        $vcenter{$datacenter}{$hostsystem}{$vm_name}
                            = { $_ => undef };
                    } else {
                        my $vmdk
                            = $vcenter{$datacenter}{$hostsystem}{$vm_name};
                        $vmdk->{$_} = undef;
                    }
                    my $datastore_vmdk = $self->_get_datastore($vmdk_name);
                    $vcenter{$datacenter}{$hostsystem}{$vm_name}{$vmdk_name}
                        = $self->discover_datastore( $hostsystem,
                        $datastore_vmdk );
                } ## end foreach (@$virtualdisks)
                my $virtualdisk = $self->get_virtualdisks($vm_name);
                my $vd = $vcenter{$datacenter}{$hostsystem}{$vm_name};
                $vd->{'virtualdisk'} = $virtualdisk;
            } ## end foreach (@$vm)
        } ## end foreach (@$esx)

    } ## end foreach (@vidatacenters)
    return \%vcenter;
} ## end sub get_vigraph

=head2 get_vi_storage_graph()

This subroutine generates a hash of all the virtual inventory with storage information

B<Input:>

An array of hashes which contains filer information

B<Returns:>

returns an reference to an hash

B<Usage:>

get_vigraph(array of hashes)

B<Example:>

my $filer;
my @filerDetails;
$filer->{filerName} = "10.72.233.37";
$filer->{userName} = "root";
$filer->{password} = "csmqa*123";
push(@filerDetails, $filer);


my $hash = $test->get_vi_storage_graph(@filerDetails));

=cut

sub get_vi_storage_graph() {
    my $self   = shift(@_);
    my $filers = shift(@_);
    my %vcenter;
    my $TYPE = 'NFS';
    $self->_clear_err();
    my $filerutils  = new filerutils();
    my $datacenters = $self->get_datacenters();

    if ( !$filers ) {
        $self->_set_err("Filers details are missing");
        return \%vcenter;
    }

    #update with datacenters
    foreach (@$datacenters) {
        $vcenter{$_} = undef;
    }

    my @vidatacenters = keys %vcenter;
    foreach (@vidatacenters) {
        my $datacenter = $_;
        my $esx        = $self->get_hostsystems($_);
        foreach (@$esx) {

            #update with hostsystems
            my $hostsystem = $_;
            if ( !$vcenter{$datacenter} ) {
                $vcenter{$datacenter} = { $_ => undef };
            } else {
                my $esx = $vcenter{$datacenter};
                $esx->{$_} = undef;
            }
            my $vm = $self->get_virtualmachines($hostsystem);
            foreach (@$vm) {

                #update with virtual machines
                my $vm_name = $_;
                if ( !$vcenter{$datacenter}{$hostsystem} ) {
                    $vcenter{$datacenter}{$hostsystem} = { $_ => undef };
                } else {
                    my $vms = $vcenter{$datacenter}{$hostsystem};
                    $vms->{$_} = undef;
                }

                my $ds_vm = $self->get_datastore_esx( $hostsystem, $vm_name );
                foreach (@$ds_vm) {
                    my $ds_name = $_;
                    if ( !$vcenter{$datacenter}{$hostsystem}{$vm_name} ) {
                        $vcenter{$datacenter}{$hostsystem}{$vm_name}
                            = { $_ => undef };
                    } else {
                        my $ds = $vcenter{$datacenter}{$hostsystem}{$vm_name};
                        $ds->{$_} = undef;
                    }
                    my $dsinfo
                        = $self->discover_datastore( $hostsystem, $ds_name );
                    if ( $dsinfo->{'type'} eq 'NFS' ) {

                        #get the filer on the NFS volume is present
                        my $nfs_filer
                            = $self->get_filer( $filers, $dsinfo->{'filer'} );
                        my $hostname
                            = $self->get_hostname( $nfs_filer->{filerName} );

                        #get the aggr on which the volume is present
                        my $info = $filerutils->get_aggr( $nfs_filer,
                            $dsinfo->{'volume'} );
                        $vcenter{$datacenter}{$hostsystem}{$vm_name}{$ds_name}
                            = {
                            $hostname . ":"
                                . $self->get_path( $dsinfo->{'volume'} ) => {
                                      $hostname . ":"
                                    . $info => { $dsinfo->{'filer'} => 'end' }
                                }
                            };
                    } elsif ( $dsinfo->{'type'} eq 'VMFS' ) {
                        my $uuid = $dsinfo->{'uuid'};
                        for (@$uuid) {
                            my $info
                                = $filerutils->get_lun_info( $filers, $_ );
                            if ( defined $info ) {
                                if ( !$vcenter{$datacenter}{$hostsystem}
                                    {$vm_name}{$ds_name} )
                                {
                                    $vcenter{$datacenter}{$hostsystem}
                                        {$vm_name}{$ds_name} = {
                                        $info->{'filerName'} . ":"
                                            . $self->get_path(
                                            $info->{'lunPath'}
                                            ) => {
                                            $info->{'filerName'} . ":"
                                                . $self->get_path(
                                                $info->{'volume'}
                                                ) => {
                                                $info->{'filerName'} . ":"
                                                    . $info->{'aggr'} => {
                                                    $info->{'filerName'} =>
                                                        'end'
                                                    }
                                                }
                                            }
                                        };
                                } else {
                                    my $ds
                                        = $vcenter{$datacenter}{$hostsystem}
                                        {$vm_name}{$ds_name};
                                    $ds->{
                                        $info->{'filerName'} . ":"
                                            . $self->get_path(
                                            $info->{'lunPath'}
                                            )
                                        }
                                        = {
                                        $info->{'filerName'} . ":"
                                            . $self->get_path(
                                            $info->{'volume'}
                                            ) => {
                                            $info->{'filerName'} . ":"
                                                . $info->{'aggr'} => {
                                                $info->{'filerName'} => 'end'
                                                }
                                            }
                                        };
                                } ## end else [ if ( !$vcenter{$datacenter...})]
                            } else {
                                $Log->comment("No storage info");
                                $Log->comment("for Datastore");
                                $Log->comment( $dsinfo->{'datastorename'} );
                                $Log->comment( Dumper($filers) );
                            } ## end else [ if ( defined $info ) ]
                        } ## end for (@$uuid)

                    } ## end elsif ( $dsinfo->{'type'}...)
                } ## end foreach (@$ds_vm)
            } ## end foreach (@$vm)
        } ## end foreach (@$esx)

    } ## end foreach (@vidatacenters)

    return \%vcenter;
} ## end sub get_vi_storage_graph

=head2 get_vi_storage_array()

This subroutine generates a array of hashes of all the virtual inventory with storage information

B<Input:>

An array of hashes which contains filer information

B<Returns:>

returns a reference to an array of hashes

B<Usage:>

get_vi_storage_array(array of hashes)

B<Example:>

my $filer;
my @filerDetails;
$filer->{filerName} = "10.72.233.37";
$filer->{userName} = "root";
$filer->{password} = "csmqa*123";
push(@filerDetails, $filer);


my $arrayofhash = $test->get_vi_storage_array(@filerDetails));

=cut

sub get_vi_storage_array() {
    my $self   = shift(@_);
    my $filers = shift(@_);
    my @vcenter;
    my $TYPE = 'NFS';
    $self->_clear_err();
    my $filerutils = new filerutils();

    if ( !$filers ) {
        $self->_set_err("Filers details are missing");
        return \@vcenter;
    }

    #get list of vfilers.
    my $vfilers = $filerutils->vfiler_to_filer_map($filers);

    #get all datacenters for the vCenter
    my $datacenters = $self->get_datacenters();

    #get datacenters
    foreach (@$datacenters) {
        my $datacenter = $_;
        my $esx        = $self->get_hostsystems($_);
        foreach (@$esx) {

            #get hostsystems
            my $hostsystem = $_;
            my $vm         = $self->get_virtualmachines($hostsystem);
            foreach (@$vm) {

                #update with virtual machines
                my $vm_name = $_;

                my $ds_vm = $self->get_datastore_esx( $hostsystem, $vm_name );
                foreach (@$ds_vm) {
                    my $ds_name = $_;
                    my $dsinfo
                        = $self->discover_datastore( $hostsystem, $ds_name );
                    if ( $dsinfo->{'type'} eq 'NFS' ) {

                        #update details of virtual objects
                        my $details;
                        $details->{"datacenter"}     = $datacenter;
                        $details->{"hostsystem"}     = $hostsystem;
                        $details->{"virtualmachine"} = $vm_name;
                        $details->{"datastore"}      = $ds_name;

                        #get the filer on the NFS volume is present
                        my $nfs_filer
                            = $self->get_filer( $filers, $dsinfo->{'filer'} );
                        if ($nfs_filer) {
                            my $physical;
                            my $hostname = $filerutils->get_hostname(
                                 $nfs_filer->{filerName},$nfs_filer->{userName},
								 $nfs_filer->{password});
                            $Log->comment("$nfs_filer->{filerName} = $hostname");							  
                            #get the aggr on which the volume is present
                            my $info = $filerutils->get_aggr( $nfs_filer,
                                $dsinfo->{'volume'} );
                            $details->{"volume"} = $hostname . ":"
                                . $self->get_path( $dsinfo->{'volume'} );
                            #Check to see if this vfiler.  If its vfiler,
                            #then append physical filer name.
                            if ( $vfilers->{$hostname} ) {
                                $Log->comment("$hostname is an vfiler");
                                $Log->comment("physical filer is $vfilers->{$hostname}");
                                $physical = $vfilers->{$hostname};
                            } else {
                                $physical = $hostname;
                            }
                            $details->{"aggr"}  = $physical . ":" . $info;
                            $details->{"filer"} = $hostname;
                        } else {
                            $Log->comment(
                                "Filer Credentails for $dsinfo->{'filer'} not given!"
                            );
                            $details->{"volume"} = undef;
                            $details->{"aggr"}   = undef;
                            $details->{"filer"}  = undef;
                        } ## end else [ if ($nfs_filer) ]
                        push( @vcenter, $details );

                    } elsif ( $dsinfo->{'type'} eq 'VMFS' ) {
                        my $uuid = $dsinfo->{'uuid'};
                        for (@$uuid) {
                            my $info
                                = $filerutils->get_lun_info( $filers, $_ );
                            if ( defined $info ) {
                                my $details;
                                my $physical;
                                $details->{"datacenter"}     = $datacenter;
                                $details->{"hostsystem"}     = $hostsystem;
                                $details->{"virtualmachine"} = $vm_name;
                                $details->{"datastore"}      = $ds_name;
                                $details->{"volume"}
                                    = $info->{'filerName'} . ":"
                                    . $self->get_path( $info->{'volume'} );
                                $details->{"lun"} = $info->{'filerName'} . ":"
                                    . $self->get_path( $info->{'lunPath'} );

                                my $filerlun = $info->{'filerName'};
                                if ( $vfilers->{$filerlun} ) {
                                $Log->comment("$filerlun is an vfiler");
                                $Log->comment("physical filer is $vfilers->{$filerlun}");
                                    $physical = $vfilers->{$filerlun};
                                } else {
                                    $physical = $info->{'filerName'};
                                }
                                $details->{"aggr"}
                                    = $physical . ":" . $info->{'aggr'};
                                $details->{"filer"} = $info->{'filerName'};
                                push( @vcenter, $details );

                            } else {
                                my $details;
                                $details->{"datacenter"}     = $datacenter;
                                $details->{"hostsystem"}     = $hostsystem;
                                $details->{"virtualmachine"} = $vm_name;
                                $details->{"datastore"}      = $ds_name;
                                $details->{"volume"}         = undef;
                                $details->{"lun"}            = undef;
                                $details->{"aggr"}           = undef;
                                $details->{"filer"}          = undef;
                                push( @vcenter, $details );
                            } ## end else [ if ( defined $info ) ]
                        } ## end for (@$uuid)
                    } ## end elsif ( $dsinfo->{'type'}...)

                } ## end foreach (@$ds_vm)
            } ## end foreach (@$vm)
        } ## end foreach (@$esx)
    } ## end foreach (@$datacenters)
    return \@vcenter;
} ## end sub get_vi_storage_array

sub get_filer {
    my $self       = shift(@_);
    my $filers     = shift(@_);
    my $filer_name = shift(@_);

    foreach (@$filers) {
        if ( $_->{filerName} eq $filer_name ) {
            return $_;
        }
    }
} ## end sub get_filer

=head2 get_hostname()

gets the hostname for an given IP address

B<Input:>

IP address

B<Returns:>

returns the hostname

B<Usage:>

get_hostname(ip address)

B<Example:>

my $string = $test->get_hostname("10.72.225.30") ;

=cut

sub get_hostname {
    my $self = shift(@_);
    my $ip   = shift(@_);

    my $iaddr = inet_aton($ip);
    my $hostname = gethostbyaddr( $iaddr, AF_INET );
    if ( $hostname =~ /\s*(.*?)\./ ) {
        return $1;
    }
} ## end sub get_hostname

sub get_path {
    my $self = shift(@_);
    my $path = shift(@_);
    if ( $path =~ /vol(.*)/ ) {
        return $1;
    }
} ## end sub get_path

=head2 get_datastore_esx()

get datastore for an given esx and virtual machine

B<Input:>

ESX,VM name

B<Returns:>

returns an array of datastores

B<Usage:>

get_datastore_esx(<esx name>,<vm name>)

B<Example:>

my $datastores = $test->get_datastore_esx(<esx name>,<vm name>) ;

=cut

sub get_datastore_esx {
    my $self = shift(@_);
    my $esx  = shift(@_);
    my $vm   = shift(@_);
    my @datastores;

    my $hostsystem = $self->_get_vi_view( "HostSystem", $esx );
    my $vmview = $self->_get_context_view_filter( 'VirtualMachine', $vm,
        $hostsystem );

    if ($vmview) {
        my $vm_ds        = $vmview->datastore;
        my $datastore_vm = $self->_get_moref( \@$vm_ds );
        foreach (@$datastore_vm) {
            push( @datastores, $_->summary->name );
        }
    } ## end if ($vmview)
    return \@datastores;
} ## end sub get_datastore_esx

sub _get_datastore {
    my $self = shift(@_);
    my $vmdk = shift;
    if ( $vmdk =~ /^\[(.*)\]/ ) {
        return $1;
    }
} ## end sub _get_datastore

=head2 discover_datastore()

This subroutine provides information about a given datastore

B<Input:>

hostsystem on which datastore is present, datastore name

B<Returns:>

returns an reference to an hash containing information about datastore.

B<Usage:>

discover_datastore(<hostsystem name(esx name)>,<datastore name>)

B<Example:>

my $datacenters = $test->discover_datastore(<hostsystem name(esx name)>,<datastore name>) ;

=cut

sub discover_datastore {
    my $self = shift(@_);
    my $esx  = shift;
    my $ds   = shift;
    my @uuid;
    my %datastore;
    my $DATASTORE_TYPE = 'NFS';
    my @datastore_luns;
    $self->_clear_err();

    my $hostsystem = $vmware->find_entity_view(
        view_type => "HostSystem",
        filter    => { name => $esx }
    );
    my $storagesys = $vmware->get_view(
        mo_ref => $hostsystem->configManager->storageSystem );
    my $fsmount = $storagesys->fileSystemVolumeInfo->mountInfo;
    foreach (@$fsmount) {

        if ( $_->{'volume'}->{'name'} eq $ds ) {
            $datastore{'datastorename'} = $ds;
            $datastore{'type'}          = $_->{'volume'}->{'type'};
            if ( $_->{'volume'}->{'type'} eq $DATASTORE_TYPE ) {
                $datastore{'filer'}  = $_->{'volume'}->{'remoteHost'};
                $datastore{'volume'} = $_->{'volume'}->{'remotePath'};
            } else {
                my $luns  = $storagesys->storageDeviceInfo->scsiLun;
                my $disks = $_->{'volume'}->{'extent'};
                foreach (@$disks) {
                    my $disk = $_->{'diskName'};
                    foreach (@$luns) {
                        if ( $_->{'canonicalName'} eq $disk ) {
                            push( @datastore_luns, $_->{'deviceName'} );
                            push( @uuid,           $_->uuid );
                        }
                    } ## end foreach (@$luns)
                } ## end foreach (@$disks)
                $datastore{'lun'}  = \@datastore_luns;
                $datastore{'uuid'} = \@uuid;
            } ## end else [ if ( $_->{'volume'}->{...})]
        } ## end if ( $_->{'volume'}->{...})
    } ## end foreach (@$fsmount)
    return \%datastore;
} ## end sub discover_datastore

=head2 get_virtualmachine_info()

This subroutine provides information about a virtual machine.

B<Input:>

virtual machine

B<Returns:>

returns an reference to an hash

B<Usage:>

get_virtualmachine_info(<vm name>)

B<Example:>

my $datacenters = $test->get_virtualmachine_info(<vm name>);

=cut

sub get_virtualmachine_info() {
    my $self = shift(@_);
    my $vm   = shift;
    my %virtualmachine;
    $self->_clear_err();
    my $vmview = $self->_get_vi_view( "VirtualMachine", $vm );

    if ($vmview) {
        my $state = $vmview->runtime->powerState;
        $virtualmachine{'state'} = $state->val;
    }

    return \%virtualmachine;
} ## end sub get_virtualmachine_info

=head2 vm_power()

This subroutine is used for powering on or off a VM.
All VMs(accros ESX/Datacenters) having the same name will get powered off or on.

B<Input:>

vm name, state (on or off)

B<Returns:>

none

B<Usage:>

 
vm_power("esx name","on");

B<Example:>

$test->vm_power("esx name","off");

=cut

sub vm_power {
    my $self  = shift(@_);
    my $vm    = shift(@_);
    my $state = shift(@_);
    my %virtualmachine;
    $self->_clear_err();
    my $vmview = $self->_get_vi_view( "VirtualMachine", $vm );

    if ($vmview) {
        $vmview->update_view_data();
        my $power = $vmview->runtime->powerState->val;
        if ( $state eq 'on' && $power eq 'poweredOff' ) {
            $vmview->PowerOnVM();
        } elsif ( $state eq 'off' && $power eq 'poweredOn' ) {
            $vmview->PowerOffVM();
        }
    } ## end if ($vmview)
} ## end sub vm_power

=head2 create_datastore()

This subroutine is used for creating NAS Datastore.
It does not create VMFS Datastore.
User needs to create a volume on the filer.

B<Input:>

my $param = { "esx" => "10.72.186.80",
                   "datastore_name"     => "yeah",
                   "filer" => "10.72.183.205",
                   "volume" => "/vol/nachi"
};


B<Returns:>

none

B<Usage:>
 
create_datastore($param)

B<Example:>

$test->create_datastore($param);

=cut

sub create_datastore {
    my $self  = shift(@_);
    my $param = shift(@_);
    $self->_clear_err();
    my $hostsystem = $self->_get_vi_view( "HostSystem", $param->{'esx'} );
    my $nas_datastore = HostNasVolumeSpec->new(
        accessMode => 'readWrite',
        localPath  => $param->{'datastore_name'},
        remoteHost => $param->{'filer'},
        remotePath => $param->{'volume'}
    );

    if ($hostsystem) {
        my $dataStoreSystem = $vmware->get_view(
            mo_ref => $hostsystem->configManager->datastoreSystem );
        eval {
            $dataStoreSystem->CreateNasDatastore( spec => $nas_datastore );
        };
        if ($@) {
            $Log->comment($@);
            $self->_set_err($@);
        }
        $Log->comment(
            "successfully created datastore $param->{'datastore_name'}");
    } else {
        $Log->comment("$param->{'esx'} not found");
        $self->_set_err("$param->{'esx'} not found");

    }
} ## end sub create_datastore

=head2 delete_datastore()

This subroutine is used for deleting a NAS Datastore
 

B<Input:>

esx name,datastore name
 
B<Returns:>

none

B<Usage:>

delete_datastore(<name of esx server>,<name of datastore>)

B<Example:>

$test->delete_datastore(<name of esx server>,<name of datastore>);

=cut

sub delete_datastore {
    my $self      = shift(@_);
    my $esx       = shift(@_);
    my $datastore = shift(@_);

    $self->_clear_err();
    my $hostsystem = $self->_get_vi_view( "HostSystem", $esx );

    if ($hostsystem) {
        my $datastore_esx = $hostsystem->datastore;
        if ($datastore_esx) {
            my $datastore_mof = $self->_get_moref( \@$datastore_esx );
            foreach (@$datastore_mof) {
                if ( $_->summary->name eq $datastore ) {
                    my $dataStoreSystem = $vmware->get_view( mo_ref =>
                            $hostsystem->configManager->datastoreSystem );
                    eval {
                        $dataStoreSystem->RemoveDatastore( datastore => $_ );
                    };
                    if ($@) {
                        $Log->comment($@);
                        $self->_set_err($@);
                    }
                } ## end if ( $_->summary->name...)
            } ## end foreach (@$datastore_mof)
        } ## end if ($datastore_esx)
        $Log->comment("successfully deleted datastore $datastore");
    } else {
        $Log->comment("$esx not found");
        $self->_set_err("$esx not found");
    }
} ## end sub delete_datastore

=head2 add_virtualdisk()

This subroutine is used adding a virtual disk to a virtual machines 

B<Input:>

my $param = { "esx" => "10.72.186.80",
                   "vm"     => "nax",
                   "vmdk" => "test.vmdk",
                   "size" => "50"
};

vmdk should have the file extension .vmdk
size is in MB.  
 
B<Returns:>

none

B<Usage:>
 
add_virtualdisk($param)

B<Example:>

$test->add_virtualdisk($param);

=cut

sub add_virtualdisk {

    my $self      = shift(@_);
    my $param     = shift(@_);
    my $datastore = undef;
    my $filename  = undef;
    my $dev       = undef;
    $self->_clear_err();
    my $hostsystem = $self->_get_vi_view( "HostSystem", $param->{'esx'} );

    unless ( defined $hostsystem ) {
        $Log->comment("ESX server $param->{'esx'} not found");
        $self->_set_err("ESX server $param->{'esx'} not found");
        return;
    }

    my $vm
        = $self->_get_context_view_filter( 'VirtualMachine', $param->{'vm'},
        $hostsystem );

    if ($vm) {
        my $path = $vm->config->files->vmPathName;
        if ( $path =~ /^(\[.*\])/ ) {
            $datastore = $1;
        }

        if ($datastore) {
            my $vm_name   = $vm->name;
            my $vmdk_file = $param->{'vmdk'};
            $filename = "$datastore/$vm_name/$vmdk_file";
        } else {
            $Log->comment("Not able to build datastore path");
            $self->_set_err("Not able to build datastore path");
        }

        my $devices = $vm->config->hardware->device;
        foreach my $device (@$devices) {
            $dev = $device
                if ( $device->deviceInfo->label eq 'SCSI controller 0' );
        }

        my $key        = $dev->key;
        my $unitNumber = $#{ $dev->device } + 1;
        my $size       = $param->{'size'} * 1024;

        my $disk_backing_info = VirtualDiskFlatVer2BackingInfo->new(
            diskMode        => 'persistent',
            fileName        => $filename,
            thinProvisioned => 'true'
        );

        my $disk = VirtualDisk->new(
            controllerKey => $key,
            unitNumber    => $unitNumber,
            key           => -1,
            backing       => $disk_backing_info,
            capacityInKB  => $size
        );

        my $devspec = VirtualDeviceConfigSpec->new(
            operation => VirtualDeviceConfigSpecOperation->new('add'),
            device    => $disk,
            fileOperation =>
                VirtualDeviceConfigSpecFileOperation->new('create')
        );

        my $vmspec
            = VirtualMachineConfigSpec->new( deviceChange => [$devspec] );

        eval { $vm->ReconfigVM( spec => $vmspec ); };
        if ($@) {
            $Log->comment($@);
            $self->_set_err($@);
        }
        $Log->comment(
            "successfully added virtual disk $param->{'vmdk'} to $param->{'vm'}"
        );
    } else {
        $Log->comment("virtual machine $param->{'vm'} not found");
        $self->_set_err("virtual machine $param->{'vm'} not found");
    }

} ## end sub add_virtualdisk

=head2 delete_virtualdisk()

This subroutine is used delete a virtual disk for a particular VM

B<Input:>

esx_name,vm name,vmdk file name

B<Returns:>

none

B<Usage:>
 
delete_virtualdisk("<esx name>","<vm name>","<vmdk file name>");

B<Example:>

$test->delete_virtualdisk("<esx name>","<vm name>","<vmdk file name>");

=cut

sub delete_virtualdisk {

    my $self = shift(@_);
    my $esx  = shift(@_);
    my $vm   = shift(@_);
    my $vmdk = shift(@_);
    my $disk = undef;
    $self->_clear_err();
    my $hostsystem = $self->_get_vi_view( "HostSystem", $esx );

    unless ( defined $hostsystem ) {
        $Log->comment("ESX server $esx not found");
        $self->_set_err("ESX server $esx not found");
        return;
    }

    my $virtualmachine
        = $self->_get_context_view_filter( 'VirtualMachine', $vm,
        $hostsystem );

    if ($virtualmachine) {
        my $devices = $virtualmachine->config->hardware->device;

        foreach (@$devices) {
            if ( $_->isa('VirtualDisk') ) {
                if ( $_->backing->isa('VirtualDiskFlatVer2BackingInfo') ) {
                    print $_->backing->fileName, "\n";
                    if ( $_->backing->fileName =~ /$vmdk/ ) {
                        $disk = $_;
                    }
                } ## end if ( $_->backing->isa(...))
            } ## end if ( $_->isa('VirtualDisk'...))
        } ## end foreach (@$devices)

        unless ( defined $disk ) {
            $Log->comment("virtual disk $vmdk not found");
            $self->_set_err("virtual disk $vmdk not found");
            return;
        }

        my $config_spec_operation
            = VirtualDeviceConfigSpecOperation->new('remove');
        my $config_file_operation
            = VirtualDeviceConfigSpecFileOperation->new('destroy');

        my $devspec = VirtualDeviceConfigSpec->new(
            operation     => $config_spec_operation,
            device        => $disk,
            fileOperation => $config_file_operation
        );
        my $vmspec
            = VirtualMachineConfigSpec->new( deviceChange => [$devspec] );
        eval { $virtualmachine->ReconfigVM( spec => $vmspec ); };
        if ($@) {
            $Log->comment($@);
            $self->_set_err($@);
        }
    } else {
        $Log->comment("virtual machine $vm not found");
        $self->_set_err("virtual machine $vm not found");

    }
} ## end sub delete_virtualdisk

=head2 update_hostsystem()

This subroutine is used to the HostSystem View.

B<Input:>

esx name;

B<Returns:>

none

B<Usage:>

update_hostsystem("<esx name>")

B<Example:>

$test->update_hostsystem("<esx name>");

=cut

sub update_hostsystem {
    my $self = shift(@_);
    my $esx  = shift(@_);
    $self->_clear_err();
    my $hostsystem = $self->_get_vi_view( "HostSystem", $esx );

    eval { $hostsystem->update_view_data(); };
    if ($@) {
        $Log->comment($@);
        $self->_set_err($@);
    }

} ## end sub update_hostsystem

=head2 create_datacenter()

This subroutine is creates a datacenter.

B<Input:>

datacenter name

B<Returns:>

none

B<Usage:>

create_datacenter(<name>);

B<Example:>

$test->create_datacenter("test");

=cut

sub create_datacenter {
    my $self            = shift(@_);
    my $datacenter_name = shift(@_);
    $self->_clear_err();
    my $rootfolder = $vmware->get_service_content()->rootFolder;
    my $root = $vmware->get_view( mo_ref => $rootfolder );
    eval { $root->CreateDatacenter( name => $datacenter_name ); };
    if ($@) {
        $Log->comment($@);
        $self->_set_err($@);
    }
} ## end sub create_datacenter

=head2 delete_datacenter()

This subroutine is deletes a datacenter.

B<Input:>

datacenter name

B<Returns:>

none

B<Usage:>

delete_datacenter(<name>);

B<Example:>

$test->delete_datacenter("test");

=cut

sub delete_datacenter {
    my $self            = shift(@_);
    my $datacenter_name = shift(@_);
    $self->_clear_err();
    my $datacenter = $self->_get_vi_view( "Datacenter", $datacenter_name );

    if ($datacenter) {
        eval { $datacenter->Destroy(); };
        if ($@) {
            $Log->comment($@);
            $self->_set_err($@);
        }
    } else {
        $self->_set_err("$datacenter_name not found");
    }

} ## end sub delete_datacenter

=head2 rename_datacenter()

This subroutine is renames a datacenter.

B<Input:>

datacenter name,new name

B<Returns:>

none

B<Usage:>

rename_datacenter(<datacenter name>,<new name>);

B<Example:>

$test->rename_datacenter("test","demo");

=cut

sub rename_datacenter {
    my $self            = shift(@_);
    my $datacenter_name = shift(@_);
    my $newname         = shift(@_);
    $self->_clear_err();
    my $datacenter = $self->_get_vi_view( "Datacenter", $datacenter_name );

    if ($datacenter) {
        eval { $datacenter->Rename( newName => $newname ); };
        if ($@) {
            $Log->comment($@);
            $self->_set_err($@);
        }
    } else {
        $self->_set_err("$datacenter_name not found");
    }

} ## end sub rename_datacenter

=head2 rename_vm()

This subroutine is renames a virtual machine

B<Input:>

esx name,vm name,new name

B<Returns:>

none

B<Usage:>

rename_vm(<esx name>,<name of vm>,<new name of vm>);

B<Example:>

$test->rename_vm(<esx name>,<name of vm>,<new name of vm>);

=cut

sub rename_vm {
    my $self    = shift(@_);
    my $esx     = shift(@_);
    my $vm      = shift(@_);
    my $newname = shift(@_);
    $self->_clear_err();
    my $hostsystem = $self->_get_vi_view( "HostSystem", $esx );
    my $virtualmachine
        = $self->_get_context_view_filter( 'VirtualMachine', $vm,
        $hostsystem );

    if ($virtualmachine) {
        $virtualmachine->Rename( newName => $newname );
    } else {
        $self->_set_err("$vm not found!");
    }

} ## end sub rename_vm

=head2 delete_vm()

This subroutine is deletes a virtual machine

B<Input:>

esx name,vm name

B<Returns:>

none

B<Usage:>

rename_vm(<esx name>,<name of vm>);

B<Example:>

$test->delete_vm(<esx name>,<name of vm>);

=cut

sub delete_vm {
    my $self = shift(@_);
    my $esx  = shift(@_);
    my $vm   = shift(@_);
    $self->_clear_err();
    my $hostsystem = $self->_get_vi_view( "HostSystem", $esx );
    my $virtualmachine
        = $self->_get_context_view_filter( 'VirtualMachine', $vm,
        $hostsystem );

    if ($virtualmachine) {
        $virtualmachine->Destroy();
    } else {
        $self->_set_err("$vm not found!");
    }

} ## end sub delete_vm

=head2 rename_datastore()

This subroutine is used to rename a datastore.

B<Input:>

esx name,datastore name,new name of datastore

B<Returns:>

none

B<Usage:>

rename_datastore(<esx name>,<datastore name>,<new name of datastore>)

B<Example:>

$test->rename_datastore(<esx name>,<datastore name>,<new name of datastore>);

=cut

sub rename_datastore {
    my $self      = shift(@_);
    my $esx       = shift(@_);
    my $datastore = shift(@_);
    my $newname   = shift(@_);
    $self->_clear_err();
    my $hostsystem = $self->_get_vi_view( "HostSystem", $esx );
    if ($hostsystem) {
        my $datastore_esx = $hostsystem->datastore;
        if ($datastore_esx) {
            my $datastore_mof = $self->_get_moref( \@$datastore_esx );
            foreach (@$datastore_mof) {
                if ( $_->summary->name eq $datastore ) {
                    $_->RenameDatastore( newName => $newname );
                }
            }
        } ## end if ($datastore_esx)
    } else {
        $self->_set_err("$datastore not found");
    }

} ## end sub rename_datastore

=head2 logout()

This subroutine is used for logging out from vmware session


B<Input:>

none

B<Returns:>

none

B<Usage:>

logout()

B<Example:>

$test->logout();

=cut

sub logout {
    $vmware->logout();
}

1;
