package Data::Catalog;

use Data::File::Base;
use Data::File::Directory;
use Data::File::ASCII;
use Data::File::BlockDevice;
use Data::File::CharDevice;
use Data::File::Pipe;
use Data::File::HardLink;
use Data::File::SoftLink;
use Data::File::Hole;
use Data::LocalHostInfo;
use Data::Dumper;
use Carp;
use strict;

=head1 NAME

  Data::Catalog - Flatten a Directory Tree into a Hash

=head1 DESCRIPTION

  This creates an internal hash representation of a Data::Tree.  Keys are the 
  filepath and values are the objects.  This is a flattened view of the Tree's 
  internal object graph, not the filesystem.  Keep in mind, as long as a Catalog 
  is in scope, the objects in it will be too.
  
=head1 SYNOPSIS

  use Data::Tree;
  use Data::Catalog;
  use Data::FileTemplate;

  my $f = Data::FileTemplate;
  ...add some files to it

  my $t = Data::Tree->new;
  $t->set_template( $f );
  $t->build;					

  my $c = Data::Catalog->new;
  $c->populate( $t );
  
  # Get a random file or directory object
  my $random_node = $c->random_node;
  
  # Get a random directory object
  my $random_dir = $c->dirs->random_node;
  
  # This is a flat view of the tree, with referents
  print $c->to_string;

=head1 METHODS

=head2 new()

  Constructor.  Takes no arguments.  Returns nothing.

=cut

sub new {
	my ( $class ) = @_;
	return bless {}, $class;
}

=head2 populate() 

  Instance method.  Takes a Data::Tree instance and populates its internal
  hash.  The node paths are the hash keys.  The objects are the values.
  Returns nothing.
  
=cut

sub populate {
	
	my ( $self, $tree ) = @_;
	
	croak "populate() requires an instance of Data::Tree" 
		unless ( ref( $tree ) && $tree->isa( 'Data::Tree' ) );

	my $dir = $tree->get_tree;
	
	$self->add( $dir );	# Add root.
	
	croak "Catalog can't populate itself.  Did you build the tree first?"
		unless defined $dir;

	$self->_parse_dir( $dir );	
}

# Recursively populates this instance with the Tree nodes.

sub _parse_dir {

	my ( $self, $dir ) = @_;
	
	for ( keys %{$dir->{files}} ) {
		$self->add( $dir->{files}{$_} );
	}
	
	# Recurse
	my $dirs = $dir->get_dirs;
	for ( @$dirs ) {
		$self->_parse_dir( $_ );
	}
}

=head2 get_file_by_name()

  Takes a filename.  Returns the corresponding object from the catalog.
  
=cut

sub get_file_by_name {

	my ( $self, $name ) = @_;
	
	croak "No file of that name was found: $name"
		unless $self->{$name};
		
	return $self->{$name};
}

=head2 add()

  Takes a filepath and a Data::File::* object and adds them to the internal 
  hash.  Returns nothing.
  
=cut

sub add {
	my ( $self, $obj ) = @_;
	my $name = $obj->get_name;
	$self->{$name} = $obj;
}

=head2 add_keys()

  Takes a reference to an array of filenames.  Adds the names to the internal
  hash as undef.  Returns nothing.
  
=cut

sub add_keys {
	my ( $self, $names ) = @_;
	
	for ( @$names ) {
		$self->{$_} = undef;
	}
}

=head2 remove()

  Takes a filepath.  Removes the node from the interanl hash.  Returns nothing.
  
=cut

sub remove {
	my ( $self, $obj ) = @_;
	my $name = $obj->get_name;
	delete $self->{$name};
}

=head2 count()

  Takes nothing.  Returns a count of the files in the catalog.
  
=cut

sub count {

	my ( $self ) = @_;
	
	return scalar keys %$self;
}

=head2 random_safe_node()

  Takes nothing.  Returns a random node from the catalog, guaranteed not
  to be root.

=cut

sub random_safe_node {

	my ( $self ) = @_;

	my $node = $self->random_node;
	
	$node = $self->random_safe_node 
		if ( $node->isa( 'Data::File::Directory' ) && $node->is_root );
	
	return $node;
}

=head2 random_node()

  Takes nothing.  Returns a random node from the catalog.

=cut

sub random_node {

	my ( $self ) = @_;
	
	my @nodes = values %$self;
	
	my $index = int rand( @nodes );

	return $nodes[$index];
}

=head2 random_move_pair()

  Takes nothing.  Returns a 'safe' pair of objects: a Data::File::* object to
  move and a Data::File::Directory to which to move it.  'Safe' means a 
  we assure the moving object can legally move to the destination.  For instance,
  a directory can't be moved under itself.

=cut

sub random_move_pair {

	my ( $self ) = @_;
	
	my $from = $self->random_node;			# File to move
	my $to_dir = $self->dirs->random_node;	# Dir to which to move it

	# I need to avoid the case where we're trying to move a directory under itself
	# This includes the case in which the source is the destination.
	
	if ( $from->isa( 'Data::File::Directory' ) ) {	
		if ( $from->is_ancestor( $to_dir ) ) {
			( $from, $to_dir ) = $self->random_move_pair;
		}
	}
	
	return $from, $to_dir;
}

=head2 all()

  Takes nothing.  Return the catalog's hash.
  
=cut

sub all {

	my ( $self ) = @_;
	
	return $self;
}

=head2 sorted_names()

  Takes nothing.  Returns a reference to an asciibetically sorted array of 
  the catalog's names.
  
=cut

sub sorted_names {

	my ( $self ) = @_;
	
	my $names = [];
	for ( sort keys %$self ) {
		push( @$names, $_ );
	}
	
	return $names;
}

=head2 files()

  Takes nothing.  Returns a Data::Catalog containing the files in the catalog,
  exclusive of directories.
  
=cut

sub files {

	my ( $self ) = @_;
		
	my $c = Data::Catalog->new;
	
	for ( sort keys %$self ) {
		next if $self->{$_}->isa( 'Data::File::Directory' );
		$c->add( $self->{$_} );
	}

	return $c;
}

=head2 types()

  Takes a file type (Data::File::*).  Returns a new Data::Catalog containing
  all the files of that type from this Data::Catalog instance.

=cut

sub types {

	my ( $self, $type ) = @_;
	
	my $c = Data::Catalog->new;
	
	for ( keys %$self ) {
		next unless $self->{$_}->isa( $type );
		$c->add( $self->{$_} );
	}
	
	return $c;
}

=head2 dirs()

  Takes nothing.  Returns a Data::Catalog containing the directories in the
  catalog, exclusive of directories.
  
=cut

sub dirs {

	my ( $self ) = @_;
	
	my $c = Data::Catalog->new;
	
	for ( sort keys %$self ) {
		next unless $self->{$_}->isa( 'Data::File::Directory' );
		$c->add( $self->{$_} );
	}

	return $c;
}

### Untested

sub stat_all {

	my ( $self ) = @_;
	
	for ( keys %$self ) {
		$self->{$_}->do_stat;
	}
}

### Untested

sub fetch_total_size {

	my ( $self ) = @_;
	
	$self->stat_all;
	
	my $total = 0;
	
	for ( keys %$self ) {
		$total += $self->{$_}{stat}[7];
	}
	
	return $total;
}

=head2 to_string()

  Takes nothing.  Returns a string representation of the object as a list of 
  filepaths.
  
=cut

sub to_string {

	my ( $self ) = @_;
	
	my $string = "Catalog:\n";
	for ( sort keys %$self ) {
		$string .= $_;
		if ( ref( $self->{$_} ) && $self->{$_}->isa( 'Data::File::Reference' ) ) {
			$string .= " => " . $self->{$_}->get_referent->get_name;
		}
		$string .= "\n";
	}
	
	return $string;
}

=head to_type_string()

  Takes nothing.  Returns a string representation of the objects as a list of 
  filepaths and their object types.
  
=cut

sub to_type_string {

	my ( $self ) = @_;
	
	my $string = "Catalog:\n";
	for ( sort keys %$self ) {
		$string .= $_;
		$string .= " [" . ref( $self->{$_} ) . "]";
		$string .= "\n";
	}
	
	return $string;

}

=head2 to_string()

  Takes nothing.  Returns a string representation of the object as a list of 
  inodes and filepaths.
  
=cut

sub to_inode_string {

	my ( $self ) = @_;

	my $string = "Catalog:\n";
	for ( sort keys %$self ) {
		$self->{$_}->do_stat;
		$string .= $self->{$_}->get_stat->ino . " ";
		$string .= $_;
		if ( $self->{$_}->isa( 'Data::File::Reference' ) ) {
			$string .= " => " . $self->{$_}->get_referent->get_name;
		}
		$string .= "\n";
	}
	
	return $string;	  
}

1;

=head1 TODO

  Make a TODO list.
  
=head1 DEPENDENCIES

  Data::Dumper.pm

=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)
  
=cut
