package Data::File::ASCII;

use Data::File::Base;
our @ISA = qw(Data::File::Base);

use Data::Dumper;
use Carp;
use strict;

=head1 NAME

  Data::File::ASCII - An ASCII File
  
=head1 DESCRIPTION

  Create, Move, or Delete and ASCII file. 
  
=head1 SYNOPSIS

  use Data::File::ASCII;
  use Data::File::Base;
  
  # Create a file
  my $f = Data::File::ASCII->new;
  $f->set_name( './foo' );
  $f->set_size( 200 );		# In bytes
  $f->create;

  print $f->to_string();

  # Create a new file name
  my $d = Data::File::Base->new;
  $d->set_name( './bing' );
 
  # Rename the ASCII file 
  $f->move( $dest );
  
  print $f->to_string();

  # Remove the file
  $f->remove;
  
  # See perldoc for Data::File::Hole.pm for documentation on how to add
  # holes to files.

=head1 METHODS

=head2 set_size()

  Takes an integer representing the number of bytes in a file and sets the 'size'
  instance variable.  Returns nothing.
  
  This is only useful for files that have content.
  
=head2 get_size()

  Takes nothing.  Gets the size of an existing file, if the 'size' instance variable
  isn't already set, by stating the file.  Returns the file size in bytes.
  
=cut

sub set_size {
	my ( $self, $size ) = @_;
	croak "size must be a non-negative integer (bytes): $size" if ( $size =~ /\D/ || $size < 0 );
	$self->{size} = $size;
}
sub get_size { 
	my ( $self ) = @_; 
	$self->set_size( -s $self->get_name ) unless defined $self->{size};
	return $self->{size}; 
}

sub set_fast {
	my ( $self, $fast ) = @_;
	croak "Fast must be 0 or 1." unless ( $fast == 0 || $fast == 1 );
	$self->{fast} = $fast;
}
sub get_fast { my ( $self ) = @_; return $self->{fast} }

=head2 create()

  Takes nothing.  Creates an ASCII file of 'size'.  Verifies that the file 
  was created.  Returns nothing.

=cut

sub create {
	
	my ( $self ) = @_;
	
	croak "You must call set_name() before creating a file." unless defined $self->{name};
	
	if ( $self->get_fast ) {	
		$self->_create_file_fast;
	} else {
		$self->_create_file;
	}
}

# Don't do file checks or generate random characters.

sub _create_file_fast {
	
	my ( $self ) = @_;
		
	my $path = $self->get_name;
	open( FILE, ">$path" ) || croak "Couldn't open file: " . $self->get_name . " for writing.\n";

	for ( 0..( $self->get_size - 1 ) ) {
		print FILE '.';
	}

	close FILE;
}

sub _create_file {

	my ( $self ) = @_;
	
	croak "File exists: " . $self->get_name if $self->file_exists( $self->get_name );
	
	my $path = $self->get_name;
	open( FILE, ">$path" ) || croak "Couldn't open file: " . $self->get_name . " for writing.\n";

	# Create a mapping of holes, so we can skip them when we write the file.
	my %holes = ();
	for my $h ( @{$self->{holes}} ) {
		for ( $h->get_start..$h->get_end ) {
			$holes{$_}++;
		}
	}	
	
	for ( 0..( $self->get_size - 1 ) ) {
		
		seek( FILE, $_, 0 );
		
		# Skip holes
		if ( $holes{$_} ) {
			#print " ";
		} 
		else {
			#print FILE ".";
			print FILE random_char();
		}
		
	}

	close FILE;
	
	croak "File not created: " . $self->get_name unless $self->file_exists( $self->get_name );
}

=head2 remove()

  Takes nothing.  Removes the ASCII file.  Verifies that the file was removed.
  Returns nothing.
  
=cut

sub remove {

	my ( $self ) = @_;

	#$self->_remove_file;

	croak "File doesn't exist." unless -e $self->{name};
	unlink $self->get_name;

	croak "Unable to remove: " . $self->get_name 
		if $self->file_exists( $self->get_name );
}

=head2 move()

  Takes a destination Data::File object.  Renames this instance to that object's
  name.  Verifies that the move occurred.  Returns nothing.

=cut

sub move {

	my ( $self, $path ) = @_;

	croak "Path $path already exists.  Can't move there: " . $self->get_name
		if -e $path;
			
	my $oldname = $self->get_name;

	rename $oldname, $path;
		
	croak "Source still exists: " . $self->get_name 
		if -e $oldname;
}

=head2 random_char()

  Class method.  Takes nothing.  Returns a random ascii char.
  
=cut

sub random_char {
	#return chr( rand( 127 ) );
	return chr( rand( 94 ) + 32 );	# Visible characters only: safer
	#return chr( rand( 126 ) );
}

=head2 add_hole()

  Takes a Hole object.  croak()s if the Hole overlaps an existing Hole.
  Otherwise adds the hole to the internal array.  Returns nothing.

=cut

sub add_hole {

	my ( $self, $hole ) = @_;
	
	croak "add_hole() requires a Hole object" 
		unless ( ref( $hole ) && $hole->isa( 'Data::File::Hole' ) );
	
	croak "You must set_start() on the hole." unless defined $hole->get_start;
	croak "You must set_end() on the hole" unless defined $hole->get_end;
	croak "You must set_eof() on the hole" unless defined $hole->get_eof;
	
	# Check whether this hole overlaps an existing hole.
	for ( @{$self->{holes}} ) {
		
		my $existing = $_;
		
		# The hole starting point is inside an existing hole.
		if ( ( $hole->get_start >= $existing->get_start ) && ( $hole->get_start <= $existing->get_end ) ) {
			croak "The new hole's start overlaps an existing hole.\nNew:\n" . 
			$self->to_string . "\nOld:" . $existing->to_string . "\n";
		}
		
		# The hole ending point is inside an existing hole. 
		elsif ( ( $hole->get_end >= $existing->get_start ) && ( $hole->get_end <= $existing->get_end ) ) {
			croak "The new hole's end overlaps an existing hole.\nNew:\n" . 
			$self->to_string . "\nOld:" . $existing->to_string . "\n";
		}
		
		# The hole start point is before the start of an existing hole AND
		# The hole end point is after the start of an existing hole
		elsif ( ( $hole->get_start < $existing->get_start ) && ( $hole->get_end > $existing->get_end ) ) {
			croak "The new hole entirely overlaps an existing hole.\n" . 
			"Its start is before the start of an existing hole AND its\n" . 
			"End is after the end of an existing hole.\nNew:\n" . 
			$self->to_string . "\nOld:" . $existing->to_string . "\n";
		}
	}

	push( @{$self->{holes}}, $hole );
	
	$self->_sort_holes;
}

# Sort the holes by where they start.  This helps when we write the file.

sub _sort_holes {

	my ( $self ) = @_;
	
	my @holes = ();
	
	for ( sort { $a->get_start <=> $b->get_start } @{$self->{holes}} ) {
		push( @holes, $_ );
	}
	
	@{$self->{holes}} = @holes;
}

1;

=head1 TODO

  Make a TODO list.
  
=head1 DEPENDENCIES

  Data::Dumper.pm
  Data::File::Base.pm

=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)
  
=cut
















