package Data::File::HardLink;

use Data::File::Base;
use Data::File::Reference;
@ISA = qw(Data::File::Base Data::File::Reference);

use Data::Dumper;
use Carp;
use strict;

=head1 NAME

  Data::File::HardLink - A Hard Link
  
=head1 DESCRIPTION

  Create, Move, or Delete a symbolic link.
  
=head1 SYNOPSIS

  use Data::File::HardLink;
  use Data::File::Base;
   
  # Create the referent file
  my $r = Data::File::ASCII->new;
  $r->set_name( './fii' );
  $r->set_size( 200 );
  
  # Create a hardlink file
  my $f = Data::File::HardLink->new;
  $f->set_name( './foo' );
  $f->set_referent( $r );
  $f->create;
  
  print $f->to_string();

  # Create a new file name
  my $d = Data::File::Base->new;
  $d->set_name( './bing' );
  
  # Rename the symlink
  $f->move( $dest );
  
  print $f->to_string();

  # Remove the symlink file
  $f->remove;

=head1 METHODS

=head2 create()

  Takes nothing.  Creates a hard link file.  Verifies that the hardlink was created.  
  Returns nothing.

=cut

sub create {

	my ( $self ) = @_;
	
	croak "You must call set_name() before creating a file." unless defined $self->{name};
	croak "You must call set_referent() before creating a file." 
		unless ( ref( $self->get_referent ) && $self->get_referent->isa( 'Data::File::Base' ) );
	
	$self->_create_file;
}

sub _create_file {

	my ( $self ) = @_;
	
	croak "File exists: " . $self->get_name if $self->file_exists( $self->get_name );
	
	my $referent = $self->get_referent->get_name;
	my $name = $self->get_name;

	#my $cmd = "$Data::File::Base::Sudo /usr/ucb/ln $referent $name";
	my $cmd = "/usr/ucb/ln $referent $name";

	my $result = `$cmd`;
	
	croak "Unable to make soft link: $name" if $result;
	croak "File not created: " . $self->get_name unless $self->file_exists( $self->get_name );
}

=head2 remove()

  Takes nothing.  Removes the hardlink file.  Verifies that the file was removed.
  Returns nothing.
  
=cut

sub remove {

	my ( $self ) = @_;
	
	$self->_remove_file;
}

sub _remove_file {

	my ( $self ) = @_;
		
	croak "File doesn't exist." unless -e $self->{name};
	unlink $self->get_name;

	croak "Unable to remove: " . $self->get_name if $self->file_exists( $self->get_name );
}

=head2 move()

  Takes a destination Data::File object.  Renames this instance to that object's
  name.  Verifies that the move occurred.  Returns nothing.

=cut

sub move {

	my ( $self, $path ) = @_;

	croak "Path $path already exists.  Can't move there: " . $self->get_name
		if -e $path;
			
	my $oldname = $self->get_name;

	rename $oldname, $path;
		
	croak "Source still exists: " . $self->get_name 
		if -e $oldname;
}

1;

=head1 TODO

  Make a TODO list.
  
=head1 DEPENDENCIES

  Data::Dumper.pm
  Data::File::Base.pm

=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)
  
=cut
















