package Data::FileSystemReader;

use strict;

use Data::Dumper;
use Carp;
use File::Find;
use Data::Catalog;
use Data::Tree;
use Data::File::Base;
use Data::File::ASCII;
use Data::File::SoftLink;
use Data::File::Pipe;
use Data::File::BlockDevice;
use Data::File::CharDevice;
#use Data::File::Binary;
#use Data::File::Unknown;

my $Files = [];

=head1 NAME

  Data::FileSystemReader - Comprehend an Existing Filesystem

=head1 DESCRIPTION

  
=head1 SYNOPSIS


=head1 METHODS

=cut

sub new {
	my ( $class ) = @_;
	my $self = bless {}, $class;
	return $self;
}

sub set_root {
	my ( $self, $root ) = @_;
	croak "root doesn't exist: $root" unless -e $root;
	$self->{root} = $root;
}
sub get_root { my ( $self ) = @_; return $self->{root} }

sub set_catalog {
	my ( $self, $catalog ) = @_;
	croak "catalog must be an instance of Data::Catalog"
		unless ( ref( $catalog ) && $catalog->isa( 'Data::Catalog' ) );
	$self->{catalog} = $catalog;
}
sub get_catalog { my ( $self ) = @_; return $self->{catalog} }

sub set_tree {
	my ( $self, $tree ) = @_;
	croak "tree must be an instance of Data::Tree"
		unless ( ref( $tree ) && $tree->isa( 'Data::Tree' ) );
	$self->{tree} = $tree;
}
sub get_tree { my ( $self ) = @_; return $self->{tree} }

=head2 discover()

  Takes nothing.  Searches and comprehends an existing filesystem, 
  populating a the 'catalog' instance variable.  Returns nothing.

=cut

sub discover {

	my ( $self ) = @_;
	
	croak "You must set the root directory before calling discover()."
		unless defined $self->get_root;
			
	# Find everything visible
	find( \&_wanted, $self->get_root );

	# Populate a Catalog
	my $c = Data::Catalog->new;
	$c->add_keys( $Files );
	$self->set_catalog( $c );
	
	for ( @$Files ) {
		my $o = $self->_make_obj( $_ );
		#$o->parse_name( $self->get_root );
		$c->add( $o );
	}

	$self->_build_tree;
	
	# Cleanup the global array (stateful stuff is poo)
	$Files = [];
}

sub _make_obj {

	my ( $self, $path ) = @_;
	
	my $obj;
	if ( -d $path ) {
		$obj = Data::File::Directory->new;
#	} elsif ( -B $path ) {
#		$obj = Data::File::Binary->new;
#	} elsif ( -T $path ) {
#		$obj = Data::File::ASCII->new;
#		$obj->get_size;
	} elsif ( -l $path ) {
		$obj = Data::File::SoftLink->new;
	} elsif ( -p $path ) {
		$obj = Data::File::Pipe->new;
	} elsif ( -c $path ) {
		$obj = Data::File::CharDevice->new;
	} elsif ( -b $path ) {
		$obj = Data::File::BlockDevice->new;
#	} else {
#		$obj = Data::File::Unknown->new;
#	}
	} else {
		$obj = Data::File::ASCII->new;
		$obj->get_size;
	}

	$obj->set_name( $path );
	
	return $obj;
}

sub _wanted {
	
	push( @$Files, $File::Find::name );
}

sub _build_tree {

	my ( $self ) = @_;
	
	my $all = $self->get_catalog->sorted_names;
		
	my $tree = Data::Tree->new;

	$self->set_tree( $tree );
	
	# This depends on the catalog list having already been sorted
	# asciibetically and that asciibetical sorting is the same thing
	# as sorting by depth.  A file's parent should already have been
	# coupled to the tree.

	for ( @$all ) {
		$self->_install_file( $_ );
	}
}

sub _install_file {

	my ( $self, $path ) = @_;
	
	my $o = $self->get_catalog->get_file_by_name( $path );

	# If it's a directory, add it to the tree as root and return
	# $tree->{tree} = $root;
	if ( $path eq $self->get_root ) {
		$self->get_tree->{tree} = $o;
		$o->set_is_root( 1 );
	}

	# Else Figure out its parent dir and add it 
	else {
		my $dir = $self->get_catalog->get_file_by_name( $o->parse_dir_name() );
		$dir->add_file( $o );
		### Still need to handle reference types.
	}
}

=head2 to_string()

  Takes no arguments.  Returns a string of keys/values for this object instance.
  If the instance variable is a string, the values are shown.  If the scalar
  is a reference, the referent's type is shown.

=cut

sub to_string {

	my ( $self ) = @_;

	my $string = "Instance of \"" . ref( $self ) . "\"\n";
	$string .= "=" x ( ( length $string ) - 1 ) . "\n";

	my $longest = '';
	my $i;
	for ( keys %{$self->{files}} ) {
		$longest = $_ unless $i;
		$longest = $_ if ( length $_ >= length $longest );
		$i++;
	}

	for ( sort keys %{$self->{files}} ) {
	
		my $pad = ( length $longest ) - ( length $_ );
		$string .= "$_ " . " " x $pad . ": ";

		# Show the type if it's a reference
		if ( my $type = ref( $self->{files}{$_} ) ) {
			$string .= "$type\n";
		}

		# Show the value
		else {
			$string .= "$self->{files}{$_}\n";
		}
	}
	
	$string .= "\n";

	return $string;
}

1;

=head1 TODO

  Make a TODO list.
  
=head1 DEPENDENCIES

  Data::Dumper.pm

=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)
  
=cut

__END__

File Test Operators:

#                 -r  File is readable by effective uid/gid.
#                 -w  File is writable by effective uid/gid.
#                 -x  File is executable by effective uid/gid.
#                 -o  File is owned by effective uid.
#
#                 -R  File is readable by real uid/gid.
#                 -W  File is writable by real uid/gid.
#                 -X  File is executable by real uid/gid.
#                 -O  File is owned by real uid.
#
#                 -e  File exists.
#                 -z  File has zero size (is empty).
#                 -s  File has nonzero size (returns size in bytes).
#
#                 -f  File is a plain file.
#                 -d  File is a directory.
#                 -l  File is a symbolic link.
#                 -p  File is a named pipe (FIFO), or Filehandle is a pipe.
#                 -S  File is a socket.
#                 -b  File is a block special file.
#                 -c  File is a character special file.
#                 -t  Filehandle is opened to a tty.
#
#                 -u  File has setuid bit set.
#                 -g  File has setgid bit set.
#                 -k  File has sticky bit set.
#
#                 -T  File is an ASCII text file.
#                 -B  File is a "binary" file (opposite of -T).
#
#                 -M  Age of file in days when script started.
#                 -A  Same for access time.
#                 -C  Same for inode change time.
