package Data::TemplateFactory;

use strict;
use Data::Dumper;
use Data::FileTemplate;
use Data::File::ASCII;
use Data::File::BlockDevice;
use Data::File::CharDevice;
use Data::File::HardLink;
use Data::File::Hole;
use Data::File::Pipe;
use Data::File::SoftLink;

=head1 NAME

  Data::TemplateFactory - A Collection of Standard Data Templates

=head1 DESCRIPTION

  This is a static library of standard FileTemplates.  Add methods here to 
  define interesting trees of data sets.  Each method is expected to 
  return a Data::FileTemplate object instance.
  
=head1 SYNOPSIS

  use Data::TemplateFactory;
  use Data::Tree;
  
  # Create a template
  my $template = Data::TemplateFactory::sixty_four_ish();
  
  # Merge templates
  $template->add_template( Data::TemplateFactory::tiny() );

=head1 METHODS

=head2 kitchen_sink()

  Takes nothing.  Returns a Data::FileTemplate comprised of an aggregate
  of the other templates defined in this packages, excluding those 
  containing big files.

=cut

sub kitchen_sink {
	
	my $t = one_of_each();
	$t->add_template( zero_sizes() );
	$t->add_template( NetBackup_block_sizes() );
	$t->add_template( ONTAP_block_sizes() );
	$t->add_template( sixty_four_ish() );
	$t->add_template( sixteen_k_ish() );
	$t->add_template( holey() );
	$t->add_template( five_twelve_byte_holes() );
	$t->add_template( four_kbyte_holes() );
	
	return $t;
}

=head2 one_of_each()

  Takes nothing.  Returns a Data::FileTemplate comprised of: 1 ASCII File,
  1 BlockDevice, 1 CharDevice, 1 Softlink, 1 HardLink, 1 Named Pipe, and 
  1 Holey ASCII File.

=cut

sub one_of_each {

	my $t = Data::FileTemplate->new;

	my $a = Data::File::ASCII->new;
	$a->set_name( 'foo1' );
	$a->set_size( 200 );
	
	my $b = Data::File::BlockDevice->new;
	$b->set_name( 'block1' );
	$b->set_minor( 1 );
	$b->set_major( 1 );
	
	my $c = Data::File::CharDevice->new;
	$c->set_name( 'char1' );
	$c->set_minor( 1 );
	$c->set_major( 1 );
	
	my $s = Data::File::SoftLink->new;
	$s->set_name( 'softlinkfoo1' );
	$s->set_referent( $a );
	
	my $h = Data::File::HardLink->new;
	$h->set_name( 'hardlinkfoo1' );
	$h->set_referent( $a );
	
	my $p = Data::File::Pipe->new;
	$p->set_name( 'pipe1' );
	
	my $g = Data::File::ASCII->new;
	$g->set_name( 'holey1' );
	$g->set_size( 300 );
	
	my $hole = Data::File::Hole->new;
	$hole->set_eof( $g->get_size );
	$hole->set_start( 10 );
	$hole->set_end( 20 );
	
	$g->add_hole( $hole );

	$t->add_files( $a, $b, $c, $s, $h, $p, $g );
	
	return $t;
}

=head2 zero_sizes()

  Takes an optional integer.  Returns a Data::FileTemplate object comprised 
  of files of zero size.  The integer controls how many files to create.

=cut

sub zero_sizes {

	my ( $count ) = @_;
	
	$count ||= 5;
	
	my $t = Data::FileTemplate->new;
	
	# Files of zero size
	for ( 1..$count ) {
	
		my $a = Data::File::ASCII->new;
		$a->set_name( "zero$_" );
		$a->set_size( 0 );
		
		$t->add_files( $a );
	}

	return $t;
}

=head2 NetBackup_block_sizes()

  Takes nothing.  Returns a Data::FileTemplate object containing ASCII file 
  objects of approximately 512 bytes in size.  Intended to stress NetBackup 
  block boundaries.

=cut

sub NetBackup_block_sizes {

	my $t = Data::FileTemplate->new;

	# 512ish for Veritas NetBackup breakpoint
	my $i = 0;
	for ( -3,-2,-1,0,1,2,3 ) {
	
		my $a = Data::File::ASCII->new;
		$a->set_name( "512-$i" );
		$a->set_size( 512 + $_ );
	
		$t->add_files( $a );
	
		$i++;
	}
	
	return $t;
}

=head2 ONTAP_block_sizes()

  Takes nothing.  Returns a Data::FileTemplate object containing ASCII file
  objects of approximately 4096 bytes in size.  Intended to stress ONTAP
  block boundaries.
  
=cut

sub ONTAP_block_sizes {

	my $t = Data::FileTemplate->new;

	# 4096ish for ONTAP breakpoint
	my $i = 0;
	for ( -3,-2,-1,0,1,2,3 ) {
		
		my $a = Data::File::ASCII->new;
		$a->set_name( "4096-$i" );
		$a->set_size( 4096 + $_ );
		
		$t->add_files( $a );
		
		$i++;
	}
	
	return $t;
}

=head2 two_gig_file()

  Takes nothing.  Returns a Data::FileTemplate object containing a single
  ASCII file object exactly 2 gigabytes in size.

=cut

sub two_gig_file {

	my $t = Data::FileTemplate->new;
	
	my $a = Data::File::ASCII->new;
	$a->set_name( "2gig" );
	$a->set_size( 2000000000 );
	
	$t->add_files( $a );
	
	return $t;
}

=head2 five_gig_file()

  Takes nothing.  Returns a Data::FileTemplate object containing a single
  ASCII file object exactly 5 gigabytes in size.

=cut

sub five_gig_file {

	my $t = Data::FileTemplate->new;
	
	my $a = Data::File::ASCII->new;
	$a->set_name( "5gig" );
	$a->set_size( 5000000000 );
	
	$t->add_files( $a );

	return $t;
}

=head2 sixty_four_ish()

  Takes nothing.  Returns a Data::FileTemplate object containing seven ASCII
  file objects approximately 64 bytes in size.

=cut

sub sixty_four_ish {
	
	my $t = Data::FileTemplate->new;
	
	my $i = 0;
	for ( -3,-2,-1,0,1,2,3 ) {
	
		my $a = Data::File::ASCII->new;
		$a->set_name( "64-$i" );
		$a->set_size( 64 + $_ );
		
		$t->add_files( $a );
		
		$i++;
	}
	
	return $t;
}

=head2 tiny()

  Takes an optional integer for the number of files to create.  Returns a 
  Data::FileTemplate object containing ASCII file objects from 0 to 64 bytes
  in size.
  
=cut

sub tiny {

	my ( $count ) = @_;
	
	$count ||= 5;
	
	my $t = Data::FileTemplate->new;
	
	# Files of zero size
	for ( 1..$count ) {
	
		my $a = Data::File::ASCII->new;
		$a->set_name( "tiny$_" );
		$a->set_size( int rand( 64 ) );
		
		$t->add_files( $a );
	}
	
	return $t;
}

=head2 sixteen_k_ish()

  Takes nothing.  Returns a Data::FileTemplate object containing seven 
  ASCII file objects approximately 16 kilobytes in size.
  
=cut

sub sixteen_k_ish {
	
	my $t = Data::FileTemplate->new;
	
	my $i = 0;
	for ( -3,-2,-1,0,1,2,3 ) {
	
		my $a = Data::File::ASCII->new;
		$a->set_name( "16000-$i" );
		$a->set_size( 16000 + $_ );
		
		$t->add_files( $a );
		
		$i++;
	}
	
	return $t;
}

=head2 eight_meg_ish()

  Takes nothing.  Returns a Data::FileTemplate object containing seven
  ASCII file objects approximately 8.16 megabytes in size.
  
=cut

sub eight_meg_ish {

	my $t = Data::FileTemplate->new;
	
	my $i = 0;
	for ( -3,-2,-1,0,1,2,3 ) {
	
		my $a = Data::File::ASCII->new;
		$a->set_name( "8160000-$i" );
		$a->set_size( 8160000 + $_ );
		
		$t->add_files( $a );
		
		$i++;
	}
	
	return $t;
}

=head2 thirty_two_meg_ish()

  Takes nothing.  Returns a Data::FileTemplate object containing seven
  ASCII file objects approximately 32.64 megabytes in size.
  
=cut

sub thirty_two_meg_ish {

	my $t = Data::FileTemplate->new;
	
	my $i = 0;
	for ( -3,-2,-1,0,1,2,3 ) {

		my $a = Data::File::ASCII->new;
		$a->set_name( "32640000-$i" );
		$a->set_size( 32640000 + $_ );
		
		$t->add_files( $a );
		
		$i++;
	}

	return $t;
}

=head2 four_gig_ish()

  Takes nothing.  Returns a Data::FileTemplate object containing seven
  ASCII file objects approximately 4.1616 gigabytes in size.

=cut

sub four_gig_ish {

	my $t = Data::FileTemplate->new;
	
	my $i = 0;
	for ( -3,-2,-1,0,1,2,3 ) {

		my $a = Data::File::ASCII->new;
		$a->set_name( "4161600000-$i" );
		$a->set_size( 4161600000 + $_ );
		
		$t->add_files( $a );
		
		$i++;
	}
	
	return $t;
}

=head2 sixteen_gig_ish()

  Takes nothing.  Returns a Data::FileTemplate object containing seven
  ASCII file objects approximately 16.6464 gigabytes in size.
  
=cut

sub sixteen_gig_ish {

	my $t = Data::FileTemplate->new;
	
	my $i = 0;
	for ( -3,-2,-1,0,1,2,3 ) {

		my $a = Data::File::ASCII->new;
		$a->set_name( "16646400000-$i" );
		$a->set_size( 16646400000 + $_ );
		
		$t->add_files( $a );
		
		$i++;
	}
	
	return $t;
}

=head2 holey()

  Takes nothing.  Returns a Data::FileTemplate object containing various
  holey ASCII file objects.
  
=cut

sub holey {

	my $t = Data::FileTemplate->new;
	
	# Two holes
	my $a = Data::File::ASCII->new;
	$a->set_name( '2holes' );
	$a->set_size( 1000 );
	
	my $ah1 = Data::File::Hole->new;
	$ah1->set_eof( $a->get_size );
	$ah1->set_start( 60 );
	$ah1->set_end( 70 );
	
	$a->add_hole( $ah1 );
	
	my $ah2 = Data::File::Hole->new;
	$ah2->set_eof( $a->get_size );
	$ah2->set_start( 505 );
	$ah2->set_end( 520 );
	
	$a->add_hole( $ah2 );
	$t->add_files( $a );
	
	# Hole at the start
	my $b = Data::File::ASCII->new;
	$b->set_name( 'starthole' );
	$b->set_size( 500 );
	
	my $bh1 = Data::File::Hole->new;
	$bh1->set_eof( $b->get_size );
	$bh1->set_start( 0 );
	$bh1->set_end( 50 );
	
	$b->add_hole( $bh1 );
	$t->add_files( $b );
	
	# Hole at the end
	my $d = Data::File::ASCII->new;
	$d->set_name( 'endhole' );
	$d->set_size( 500 );
	
	my $dh1 = Data::File::Hole->new;
	$dh1->set_eof( $d->get_size );
	$dh1->set_start( 450 );
	$dh1->set_end( 500 );
	
	$d->add_hole( $dh1 );
	$t->add_files( $d );
	
	# 100 Holes
	my $c = Data::File::ASCII->new;
	$c->set_name( '100holes' );
	$c->set_size( 5000 );
	
	for ( 1..100 ) {
		my $h = Data::File::Hole->new;
		$h->set_eof( $c->get_size );
		$h->set_start( $_ * 45 );
		$h->set_end( ( $_ * 45 ) + 20 );
	
		$c->add_hole( $h );	
	}
	
	$t->add_files( $c );

	return $t;
}

=head2 five_twelve_byte_holes()

  Takes an optional integer for the number of files to create.  Returns a
  Data::FileTemplate object containing ASCII file objects with seven holes
  approximately 512 bytes in size.
  
=cut

sub five_twelve_byte_holes {

	my ( $count ) = @_;
	
	$count ||= 1;

	my $t = Data::FileTemplate->new;

	for ( 1..$count ) {

		# ~ 512 byte holes
		my $a = Data::File::ASCII->new;
		$a->set_name( "512holes$_" );
		$a->set_size( 4000 );
		
		my $ah1 = Data::File::Hole->new;
		$ah1->set_eof( $a->get_size );
		$ah1->set_start( 10 );
		$ah1->set_end( 519 );		# 509
		$a->add_hole( $ah1 );
		
		my $ah2 = Data::File::Hole->new;
		$ah2->set_eof( $a->get_size );
		$ah2->set_start( 540 );
		$ah2->set_end( 1050 );		# 510
		$a->add_hole( $ah2 );
		
		my $ah3 = Data::File::Hole->new;
		$ah3->set_eof( $a->get_size );
		$ah3->set_start( 1070 );
		$ah3->set_end( 1581 );		# 511
		$a->add_hole( $ah3 );
		
		my $ah4 = Data::File::Hole->new;
		$ah4->set_eof( $a->get_size );
		$ah4->set_start( 1601 );
		$ah4->set_end( 2113 );		# 512
		$a->add_hole( $ah4 );
		
		my $ah5 = Data::File::Hole->new;
		$ah5->set_eof( $a->get_size );
		$ah5->set_start( 2133 );	
		$ah5->set_end( 2646 );		# 513
		$a->add_hole( $ah5 );
	
		my $ah6 = Data::File::Hole->new;
		$ah6->set_eof( $a->get_size );
		$ah6->set_start( 2666 );
		$ah6->set_end( 3170 );		# 514
		$a->add_hole( $ah6 );
		
		my $ah7 = Data::File::Hole->new;
		$ah7->set_eof( $a->get_size );
		$ah7->set_start( 3190 );	# 515
		$ah7->set_end( 3705 );
		$a->add_hole( $ah7 );
	
		$t->add_files( $a );
	}

	return $t;
}

=head2 four_kbyte_holes()

  Takes an optional integer for the number of files to create.  Returns a
  Data::FileTemplate object containing ASCII file objects with seven holes
  approximately 4096 bytes in size.
  
=cut

sub four_kbyte_holes {

	my ( $count ) = @_;
	
	$count ||= 1;

	my $t = Data::FileTemplate->new;

	for ( 1..$count ) {

		my $a = Data::File::ASCII->new;
		$a->set_name( "4096holes$_" );
		$a->set_size( 30000 );
		
		my $ah1 = Data::File::Hole->new;
		$ah1->set_eof( $a->get_size );
		$ah1->set_start( 10 );
		$ah1->set_end( 5003 );
		$a->add_hole( $ah1 );	# 4093
		
		my $ah2 = Data::File::Hole->new;
		$ah2->set_eof( $a->get_size );
		$ah2->set_start( 5023 );
		$ah2->set_end( 9117 );
		$a->add_hole( $ah2 );	# 4094
		
		my $ah3 = Data::File::Hole->new;
		$ah3->set_eof( $a->get_size );
		$ah3->set_start( 9137 );
		$ah3->set_end( 13232 );
		$a->add_hole( $ah3 );	# 4095
		
		my $ah4 = Data::File::Hole->new;
		$ah4->set_eof( $a->get_size );
		$ah4->set_start( 13252 );
		$ah4->set_end( 17348 );
		$a->add_hole( $ah4 );	# 4096
		
		my $ah5 = Data::File::Hole->new;
		$ah5->set_eof( $a->get_size );
		$ah5->set_start( 17368 );
		$ah5->set_end( 21465 );
		$a->add_hole( $ah5 );	# 4097
		
		my $ah6 = Data::File::Hole->new;
		$ah6->set_eof( $a->get_size );
		$ah6->set_start( 21485 );
		$ah6->set_end( 25583 );
		$a->add_hole( $ah6 );	# 4098
		
		my $ah7 = Data::File::Hole->new;
		$ah7->set_eof( $a->get_size );
		$ah7->set_start( 25603 );
		$ah7->set_end( 29702 );
		$a->add_hole( $ah7 );	# 4099

		$t->add_files( $a );
	}
	
	return $t;
}

=head2 three_gig_hole()

  Takes nothing.  Returns a Data::FileTemplate object containing an ASCII
  file object 3.5 gigabytes in size with a 3.1 gigabyte hole in it.

=cut

sub three_gig_hole {
	
	my $t = Data::FileTemplate->new;

	my $a = Data::File::ASCII->new;
	$a->set_name( '3gighole' );
	$a->set_size( 3500000000 );
		
	my $ah = Data::File::Hole->new;
	$ah->set_eof( $a->get_size );
	$ah->set_start( 100000000 );
	$ah->set_end( 3100000000 );	# 3 gigs
	$a->add_hole( $ah );
	
	$t->add_files( $a );
}

=head2 pixar()

  Takes nothing.  Returns a Data::FileTemplate object containing files ranging
  from 32K to 2M.

=cut

sub pixar {

	my $t = Data::FileTemplate->new;
	
	for ( 16, 32, 64, 128, 512, 1028, 2056 ) {
	
		my $f = Data::File::ASCII->new;
		$f->set_name( "$_" );
		$f->set_size( $_ * 1000 );
		
		$t->add_files( $f );
	}

	return $t;
}

1;

=head1 TODO

  Make a TODO list.
  
=head1 DEPENDENCIES

  Data::Dumper.pm
  Data::FileTemplate.pm
  Data::File::ASCII.pm
  Data::File::BlockDevice.pm
  Data::File::CharDevice.pm
  Data::File::HardLink.pm 
  Data::File::Hole.pm
  Data::File::Pipe.pm
  Data::File::SoftLink.pm

=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)
  
=cut

__END__

0 1
1 2
2 4
3 8
