#Hey emacs, this is basically -*- perl -*-
# Copyright (c) 2008 NetApp, Inc.
# All rights reserved.

## @summary Library to handle data generation and validation.
## @author collinm@netapp.com
## @status public
## @pod here

package DataGV;
use strict;
use Data::Dumper;
use TracePkg;
use IO::File;
use Digest::MD5 qw(md5 md5_hex md5_base64);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);

my $Trace = new TracePkg();

# Class data
my $DEFAULT_VERSION = 0; # Default header encapsulation version
# MD5 length
# Account for the largest integer in string form
# Random seed length
# Header type field
# Header length field
my $FIELD_SIZES_VER0 = {
  "_MD5_size" => 32,
  "_Block_num_size" => 10,
  "_Seed_bytes" => 10,
  "_Type_bytes" => 1,
  "_Len_bytes" => 4
};

# Header type definitions
my $KEY_TO_TYPE_VER0 = { 
  0 => "MD5",
  1 => "Filename",
  2 => "Block Number",
  3 => "User Defined Key",
  4 => "User Defined Value"
};

my $TYPE_TO_KEY_VER0 = {
  "MD5" => 0,
  "Filename" => 1,
  "Block Number" => 2,
  "User Defined Key" => 3,
  "User Defined Value" => 4
};

# Version to type definitions lookup
# Optional improvement:
# A best practice would be to make this a hash of hash rather than a hash with references to other hashes.
my %VALID_VERSIONS = (
  0 => {'key_to_type' => $KEY_TO_TYPE_VER0, 'type_to_key' => $TYPE_TO_KEY_VER0, 'field_sizes' => $FIELD_SIZES_VER0},
);

my @validCifsUTF8 = ();
# -- From RandomData.pm --
# this fills the validCifsUTF8 array with all the UTF8 characters in the Latin-1 Character set
# we know Windows can send us
use bytes; # consider everything in this array a hex value
use utf8;  # consider the bytes as utf8 characters not ascii characters
foreach my $char( 0x41..0x5A, 0x61..0x7A, 0xAA, 0xBA, 0xC0..0xD6, 0xD8..0xF6, 0xF8..0xff,
    0x120..0x121, 0x123..0x129, 0x12b..0x12e, 0x130, 0x13b, 0x13d, 0x140..0x15b,
    0x15d..0x17b, 0x17d..0x17e, 0x1a1..0x1BA, 0x1BB, 0x1BC..0x1BF, 0x1C0..0x1C3,
    0x1C4..0x1ff, 0x223..0x229, 0x22b..0x22e, 0x230..0x233, 0x250..0x25b, 0x25d..0x27b,
    0x27d..0x27e, 0x2a1..0x2AD, 0x2B0..0x2B8, 0x2E0..0x2E4, 0x1e20..0x1e21, 0x1e23..0x1e29,
    0x1e2b..0x1e2e, 0x1e30..0x1e39, 0x1e3b, 0x1e3d, 0x1e40..0x1e5b, 0x1e5d..0x1e7b,
    0x1e7d..0x1e7e, 0x1ea1..0x1EF9, 0x212B, 0xFF21, 0xff23..0xff29, 0xff2b..0xff2e,
    0xff30..0xFF39, 0xFF41..0xFF5A){
    push @validCifsUTF8, chr($char);
}
no utf8;  # turn off utf8 character encoding
no bytes; # turn off byte considerations
# -- End RandomData.pm --

=head1 NAME

DataGV - Library to handle data generation and validation.

For more information, see: http://wikid/w/Spin/QA/projects/DataGV

=head1 DESCRIPTION

C<DataGV> implements an object which is used for generating per block data and for validating correctness of that same data.  Data will be generated based on a specified block size and other key identifiers, then encapsulated in a header and footer.  The header and footer are used to validate the data and detect any corruption.  The key identifiers are used to validate data is placed on disk correcly and that no corruption has occurred.

=head1 SYNOPSIS

    use DataGV;

    my $block = "";
    my $data = "";
    my $block_size = 1024;
    my %opt_headers = (Version => 3.1, Debugging => 2);
    my $library_version = 2;

    my $DataGV = new DataGV(block_size => $block_size, filename => $filename, opt_headers => \%opt_headers, library_version => $library_version);
    $data = $DataGV->generate_data();
    $DataGV->write(data => $data, block_num => $block_number);

    $block = $DataGV->read(block_num => $block_number); # Validates header data
    $DataGV->validate_data(data => $block->{Data}); # Validates random data

    # optional headers, ex: $block->{Version} and $block->{Debugging}, are for own use

=head1 METHODS

=head2 new

     my $DataGV = new DataGV(block_size => $block_size, filename => $filename, opt_headers => \%opt_headers, library_version => $library_version);
     
     or
     
     my $DataGV = new DataGV(read_only => 1, block_size => $block_size, filename => $filename, opt_headers => \%opt_headers, library_version => $library_version);

The C<new> method instantiates a DataGV object.  C<read_only> opens files as RO for validating data on mirrors, C<block_size> determines the number of bytes written and read to and from C<filename>, C<opt_headers> is an optional hash of user-defined headers for the blocks.  C<library_version> is an optional version definition which can be used if the library has been updated but is needed to work the same way it did in previous versions, if not defined, the library will use the default which should be updated to the latest version.  The C<new> method also creates a file handle to C<filename> and computes the remaining block size after accounting for the headers and footer.  The file handle is closed when the object is destroyed.

Returns a DataGV instance.

=cut

sub new {
    my $invocant = shift(@_);
    my $class = ref($invocant) || $invocant;
    my %params = validate (
        @_,
        {
            block_size => {type => SCALAR},
            filename => {type => SCALAR},
            opt_headers => {type => HASHREF, optional => 1},
            library_version => {type => SCALAR, default => $DEFAULT_VERSION},
            read_only => {type => SCALAR, optional => 1, default => 0},
        }
    );
    my $self = {
        _block_size => $params{block_size},
        _filename => $params{filename},
        _opt_headers => $params{opt_headers},
        _lib_version => $params{library_version},
        _read_only => $params{read_only},
        _remaining_bytes => 0,
        _file_handle => undef,
        _key_to_type => undef,
        _type_to_key => undef,
        _field_sizes => undef,
        _last_block => undef,
        _seed => 0
    };
    my $entry = "Initializing instance with, block size: ".$params{block_size}.", filename: ".$params{filename};
    $entry .= ", library_version: ".$params{library_version};
    if (defined $params{opt_headers}) {
        $entry .= ", opt_headers: ".Dumper($params{opt_headers});
    }
    $Trace->entry(msg => $entry);

    # Check if version number is valid, and set type definitions
    $Trace->debug(msg => "Using library version ".$self->{_lib_version});
    if (defined $VALID_VERSIONS{$self->{_lib_version}}) {
        $self->{_key_to_type} = $VALID_VERSIONS{$self->{_lib_version}}->{'key_to_type'};
        $self->{_type_to_key} = $VALID_VERSIONS{$self->{_lib_version}}->{'type_to_key'};
        $self->{_field_sizes} = $VALID_VERSIONS{$self->{_lib_version}}->{'field_sizes'};
    } else {
        my $err_str = "Version ".$self->{_lib_version}." is not a valid DataGV library version!\n";
        $err_str .= "Valid versions are [ ";
        for my $key (keys %VALID_VERSIONS) {
            $err_str .= $key." ";
        }
        $err_str .= "]";
        $Trace->fatal(msg => $err_str, result => "SCRIPT");
    }

    # Determine how many bytes of the block size will remain after removing header + footer.
    &_computeRemainingBytes($self);

    # Open file for reading and writing
    &_openFile($self);

    bless $self, $class;
    $Trace->exit(msg => "Returning DataGV instance.");
    return $self;
}

=head2 write

    $DataGV->write(data => $data, block_num => $block_number);

The C<write> method takes a block of data of size C<get_write_size()> and writes it to C<block_number>.  The data will be encapsulated in a header and footer before being written to disk.

Returns nothing.

=cut 

sub write {
    my $self = shift(@_);
    my %params = validate (
        @_,
        {
            data => {type => SCALAR},
            block_num => {type => SCALAR},
        }
    );
    my $data = $params{data};
    my $block_num = $params{block_num};
    
    $Trace->entry(msg => "Writing data to block number $block_num");
    
    # Error if file was opened in RO mode
    if ($self->{_read_only}) {
        $Trace->fatal(msg => "Cannot write to file ".$self->{_filename}." because it was opened in read-only mode (user-defined in constructor).");
    }
    
    # Determine how many bytes of the block size will remain after removing header + footer.
    &_computeRemainingBytes($self, $self->{_filename});
    # Make sure data + header + footer is not too big for block
    if (length($data) > $self->{_remaining_bytes}) {
        my $err_str = "Cannot write ".length($data)." bytes when only ".$self->{_remaining_bytes}." remaining for block!";
        $Trace->fatal(msg => $err_str, result => "SCRIPT");
    }
    # Make sure data + header + footer is not too small for block
    # This may be updated to allow for smaller write sizes, but the hole between the end of the data and the footer will need to be null'ed out.
    if (length($data) < $self->{_remaining_bytes}) {
        my $err_str = "Block write size must be ".$self->{_remaining_bytes}." bytes, only received ".length($data)." bytes!";
        $Trace->fatal(msg => $err_str, result => "SCRIPT");
    }

    # Encapsulate data in header and footer
    &_encapsulate($self, \$data, $block_num);
    # Move offset by 'block_size' bytes for each block
    my $offset = $block_num * $self->{_block_size};
    # Write 'block_size' bytes starting at 'block_num'
    $Trace->debug(msg => "Writing [$data] at offset $offset");
    &_writeBytes($self, $data, $offset);

    $Trace->exit(msg => "Returning data written.");
    return;
}

=head2 read

    $block = $DataGV->read(block_num => $block_number);
    $data = $block->{Data};
    $md5 = $block->{MD5};
    $version = $block->{Version};
    
    or
    
    $block = $DataGV->read(ignore_filename => 1, block_num => $block_number);
    $data = $block->{Data};
    $md5 = $block->{MD5};
    $version = $block->{Version};

The C<read> method takes a block_number to read, reads the block from disk, and parses out the header and footer.  The parsed header and footer data are stored in a hash along with the user data and returned to the user.  All library defined headers and footers (MD5, filename, block number) are validated before returning.  You can pass the 'ignore_filename' parameter to skip validating the filename for symlinks, etc.

Returns a hash of the data and headers.

Header type definitions:

my $key_to_type = {
  0 => "MD5",
  1 => "Filename",
  2 => "Block Number",
  3 => "User Defined Key",
  4 => "User Defined Value"
};

The hash key for the data will be "Data", and "User Defined Key" will be replaced with the actual key name.

=cut

sub read {
    my $self = shift(@_);
    my %params = validate (
        @_,
        {
            block_num => {type => SCALAR},
            ignore_filename => {type => SCALAR, optional => 1, default => 0},
        }
    );
    my $block_num = $params{block_num};

    $Trace->entry(msg => "Reading data from block number $block_num");
    # Move offset by 'block_size' bytes for each block
    my $offset = $block_num * $self->{_block_size};
    # Read 'block_size' bytes starting at 'block_num'
    my $data = &_readBytes($self, $offset);
    # Extract header, footer, and data
    my $block_data = &_decapsulate($self, $data);
    $self->{_last_block} = $block_data;
    my $hexStr = &_hexString($data);
    $Trace->debug(msg => "Extracted:".&_toString($block_data)."\nfrom block $block_num\nData in hex:\n$hexStr\n");

    # Validate library defined headers
    &_validateBlock($self, $block_data, $block_num, $params{ignore_filename});

    $Trace->exit(msg => "Returning the block hash.");
    return $block_data;
}

=head2 generate_data

    my $data = $DataGV->generate_data();

The C<generate_data> method generates a string of random characters of size C<get_write_size()>.  The random seed used to generate the string will be stored in the beginning of the string (the random data will actually be C<get_write_size()> - length of seed), the seed is stored for regenerating the same random data later.

Returns a string of random data.

=cut

sub generate_data {
    my $self = shift;
    my $data = "";

    $Trace->entry();
    # Calculate a random number seed
    my $seed = (time ^ ($$ + ($$ << 15)));
    $self->{_seed} = $seed;
    $Trace->debug(msg => "Generating ".$self->{_remaining_bytes}." bytes with random seed ".$self->{_seed});

    # Build a random string from the character array
    my $data_size = $self->{_remaining_bytes} - $self->{_field_sizes}->{"_Seed_bytes"};
    &_gen_data($self, \$data, $data_size);
    # Add random seed to beginning of string to reproduce later
    $data = $seed.$data;

    $Trace->exit(msg => "Returning random data.");
    return $data;
}

# Create a 'data_size' random string from a pre-seeded random number generator
sub _gen_data {
    my $self = shift;
    my $data = shift;
    my $data_size = shift;

    $Trace->entry(msg => "Generating a random string of length $data_size");
    # Build a random string from the character array
    my $arr_size = scalar @validCifsUTF8;
    for (my $i = 0; $i < $data_size; $i++) {
        $$data .= $validCifsUTF8[int(&_data_rand($self, $arr_size))];
    }

    $Trace->exit();
}

# Internal random number generator, to avoid variations between different OS's
sub _data_rand {
    my $self = shift;
    my $top = shift || 1.0;

    $self->{_seed} = &_libc_rand($self->{_seed});
    $top * ($self->{_seed} & 32767) / 32768;
}

sub _libc_rand {
    my $seed = shift;
    use integer;
    $seed = (1103515245 * $seed + 12345) & 0x7fffffff;
}

=head2 validate_data

    $DataGV->validate_data();
  or
    $DataGV->validate_data(data => $block->{Data});
  or
    $DataGV->validate_data(expected => $expected_data);
  or
    $DataGV->validate_data(data => $block->{Data}, expected => $expected_data);

The C<validate_data> method takes a string of random data that was generated by the C<generate_data()> method and validates that there is no data corruption, ie. the same data can be regenerated via the stored random seed.  If C<data> is not passed in, the data from the last read block will be used.  Expected data is an optional parameter that can be passed in, which will skip regenerating the random data and will only compare those two pieces of data.  This is not to be confused with the header and footer validation that occurs during the C<read> method.

Returns nothing.

=cut

sub validate_data {
    my $self = shift(@_);
    my %params = validate (
        @_,
        {
            data => {type => SCALAR, default => $self->{_last_block}->{Data}},
            expected => {type => SCALAR, optional => 1},
        }
    );
    my $data = $params{data};
    my $expected = $params{expected};
    my $seed;
    $Trace->entry(msg => "Validating data against expected data");

    if (! defined $expected) {
        # The data has to have at least one character and the seed
        if (length($data) <= $self->{_field_sizes}->{"_Seed_bytes"}) {
            my $err_str = "Data must be larger than the random seed (".$self->{_field_sizes}->{"_Seed_bytes"}.". bytes) to be validated! (Data: $data)";
            $Trace->fatal(msg => $err_str);
        }
        # Extact seed from random data
        $seed = substr($data, 0, $self->{_field_sizes}->{"_Seed_bytes"});
        if ($seed !~ /[0-9]{$self->{_field_sizes}->{"_Seed_bytes"}}/) {
            my $hexStr = &_hexString($data);
            my $err_str = "Could not extract a valid seed from data (Found: [$seed] in [$data])\nData in hex:\n$hexStr\n";
            $Trace->fatal(msg => $err_str);
        }
        $Trace->debug(msg => "Extracted random seed $seed");
        $data = substr($data, $self->{_field_sizes}->{"_Seed_bytes"});

        # Seed the random number generator with same seed used to generate data
        $self->{_seed} = $seed;
        # Determine how many bytes of the block size will remain after removing header + footer.
        &_computeRemainingBytes($self, $self->{_last_block}->{Filename});
        my $data_size = $self->{_remaining_bytes} - $self->{_field_sizes}->{"_Seed_bytes"};
        # Generate the same "random" data
        &_gen_data($self, \$expected, $data_size);
    }

    # Validate that the regenerated data matches the original data
    if ($data ne $expected) {
        my $hexStr = &_hexString($data);
        my $hexStr_expected = &_hexString($expected);
        my $err_str = "The previously generated data does not match the expected data, possible corruption!  See trace comments for data.";
        if (defined $seed) {
            $err_str .= " [Seed: $seed]";
        }
        $Trace->comment(msg => "Original data: $data\n\nExpected data: $expected\n");
        $Trace->comment(msg => "Data in hex:\n$hexStr\nExpected in hex:\n$hexStr_expected\n");
        $Trace->fatal(msg => $err_str);
    }

    $Trace->exit();
}

=head2 get_write_size

    my $data_size = $DataGV->get_write_size();

Returns the block size remaining after the header and footer have been accounted for.

=cut

sub get_write_size {
    my $self = shift;
    return $self->{_remaining_bytes};
}

=head2 close_file

    $DataGV->close_file();

Closes the open filehandle.
Returns nothing.

=cut

sub close_file {
    my $self = shift;

    $Trace->entry(msg => "Closing file ".$self->{_filename});
    # Close the file if it is open
    $self->{_file_handle}->close();
    if ($self->{_file_handle}->opened()) {
        $Trace->warn(msg => "Failed to close file ".$self->{_filename}."$!");
    } else {
        $Trace->debug(msg => "Closed file ".$self->{_filename});
    }
    $Trace->exit();
}

=head2 reopen_file

    $DataGV->reopen_file();

Re-opens the closed filehandle.
Returns nothing.

=cut

sub reopen_file {
    my $self = shift;
    
    $Trace->entry();
    # Open the file
    if ($self->{_file_handle}->opened()) {
        $Trace->warn(msg => "Failed to reopen file ".$self->{_filename}.".  The file is already open!");
    } else {
        $self->_openFile();
    }
    $Trace->exit();
}

# Validates that the headers and footer are not corrupt.
sub _validateBlock {
    my $self = shift;
    my $block_data = shift;
    my $block_num = shift;
    my $ignore_filename = shift;
    my $md5_pass = "";
    my $blocknum_pass = "";
    my $filename_pass = "";

    $Trace->entry(msg => "Validating header and footer of block number $block_num");
    # Validate that fields are defined
    if (! defined $block_data->{"MD5"}) {
        my $block_str = &_toString($block_data);
        my $hexStr = &_hexString($block_data->{Data});
        my $err_str = "Validation failed because MD5 does not exist in read block!  See trace comments for data.";
        $Trace->comment(msg => $block_str."\nData in hex:\n$hexStr\n");
        $Trace->fatal(msg => $err_str);
    } elsif (! defined $block_data->{"Block Number"}) {
        my $block_str = &_toString($block_data);
        my $hexStr = &_hexString($block_data->{Data});
        my $err_str = "Validation failed because Block Number does not exist in read block!  See trace comments for data.";
        $Trace->comment(msg => $block_str."\nData in hex:\n$hexStr\n");
        $Trace->fatal(msg => $err_str);
    } elsif (! defined $block_data->{"Filename"}) {
        my $block_str = &_toString($block_data);
        my $hexStr = &_hexString($block_data->{Data});
        my $err_str = "Validation failed because Filename does not exist in read block!  See trace comments for data.";
        $Trace->comment(msg => $block_str."\nData in hex:\n$hexStr\n");
        $Trace->fatal(msg => $err_str);
    }

    # Validate MD5 checksums
    my $md5 = md5_hex($block_data->{Data});
    my $md5_head = $block_data->{"MD5"};
    my $md5_foot = $block_data->{"MD5 Footer"};
    if ($md5_head ne $md5_foot) { # if header and footer do not match
        if ($md5_head eq $md5) {
            # Header & current match, footer does not
            $md5_pass = "MD5 Footer is corrupt! Does not match the header and current ";
            $md5_pass .= "(Footer: ".$md5_foot.", Header & Current: ".$md5.")\n";
        } elsif ($md5_foot eq $md5) {
            # Footer & current match, header does not
            $md5_pass = "MD5 Header is corrupt! Does not match the footer and current ";
            $md5_pass .= "(Header: ".$md5_head.", Footer & Current: ".$md5.")\n";
        } else {
            # All 3 MD5's mismatch
            $md5_pass = "MD5 checksums don't match! Header, footer and current differ ";
            $md5_pass .= "(Header: ".$md5_head.", Footer: ".$md5_foot.", Current: ".$md5.")\n";
        }
    } elsif ($md5_head ne $md5) {
        # Header & footer match, current does not
        $md5_pass = "MD5 header and footer match each other but the current MD5 does not, may be data corruption! ";
        $md5_pass .= "(Header & Footer: ".$md5_head.", Current: ".$md5.")\n";
    }

    # Validate block numbers
    if ($block_num ne $block_data->{"Block Number"}) {
        $blocknum_pass = "Block number read (".$block_num.") does not match header's block number (";
        $blocknum_pass .= $block_data->{"Block Number"}.")!\n";
    }

    # Validate filenames
    if (! $ignore_filename && $self->{_filename} ne $block_data->{"Filename"}) {
        $filename_pass = "Filename (".$self->{_filename}.") does not match header's filename (";
        $filename_pass .= $block_data->{"Filename"}.")!\n";
    }

    # If any of the validations failed
    if ($md5_pass ne "" || $blocknum_pass ne "" || $filename_pass ne "") {
        my $block_str = &_toString($block_data);
        my $hexStr = &_hexString($block_data->{Data});
        my $err_str = "Validation failed with:\n".$md5_pass.$filename_pass.$blocknum_pass."\nSee trace comments for data.";
        $Trace->comment(msg => $block_str."Data in hex:\n$hexStr\n");
        $Trace->fatal(msg => $err_str);
    }

    $Trace->exit();
}

# Figure out how many bytes the header + footer take up, and compute remaining data bytes
sub _computeRemainingBytes {
    my $self = shift;
    my $filename = shift;
    my $overhead_size = 0;

    $Trace->entry();
    # Add up the length of each optional header's key and value, plus
    # there will be a Type + Length header field for each key and each value
    while ( my ($key, $value) = each(%{$self->{_opt_headers}}) ) {
        $overhead_size += length($key) + length($value) + (2 * $self->{_field_sizes}->{"_Type_bytes"}) + (2 * $self->{_field_sizes}->{"_Len_bytes"});
    }
    # Add up the header lengths for filename
    if (defined($filename)) {
        $overhead_size += length($filename) + $self->{_field_sizes}->{"_Type_bytes"} + $self->{_field_sizes}->{"_Len_bytes"};
    } else {
        $overhead_size += length($self->{_filename}) + $self->{_field_sizes}->{"_Type_bytes"} + $self->{_field_sizes}->{"_Len_bytes"};
    }
    # Add up the MD5 header and footer lengths
    $overhead_size += (2 * $self->{_field_sizes}->{"_MD5_size"}) + $self->{_field_sizes}->{"_Type_bytes"} + $self->{_field_sizes}->{"_Len_bytes"};
    # Add up the header lengths for block_number
    $overhead_size += $self->{_field_sizes}->{"_Block_num_size"} + $self->{_field_sizes}->{"_Type_bytes"} + $self->{_field_sizes}->{"_Len_bytes"};
    # Add the Type + Length fields for the header encapsulation
    $overhead_size += $self->{_field_sizes}->{"_Type_bytes"} + $self->{_field_sizes}->{"_Len_bytes"};

    $self->{_remaining_bytes} = $self->{_block_size} - $overhead_size;
    $Trace->exit(msg => $self->{_remaining_bytes}." remaining bytes from ".$self->{_block_size}." bytes");
}

# Encapsulate the data in a header and footer
sub _encapsulate {
    my $self = shift;
    my $data = shift;
    my $block_num = shift;

    $Trace->entry(msg => "Encapsulating block number $block_num");
    # Create sub headers (format: header type + header length [in hex] + header)
    my $filename_header = $self->{_type_to_key}->{"Filename"};
    $filename_header .= &_toHex(length($self->{_filename}));
    $filename_header .= $self->{_filename};

    my $blocknum_header = $self->{_type_to_key}->{"Block Number"};
    $blocknum_header .= &_toHex($self->{_field_sizes}->{"_Block_num_size"});
    $blocknum_header .= sprintf("%010u", $block_num); # pad with 0's

    my $md5_str = md5_hex($$data);
    $Trace->debug(msg => "Footer encapsulation: ".$md5_str);
    my $md5_head = $self->{_type_to_key}->{"MD5"};
    $md5_head .= &_toHex($self->{_field_sizes}->{"_MD5_size"});
    $md5_head .= $md5_str;

    # Create sub headers from user defined hash
    my $user_header = "";
    my $opt_headers = $self->{_opt_headers};
    # Create a header for each key and for each value
    for my $key (keys %$opt_headers) {
      $user_header .= $self->{_type_to_key}->{"User Defined Key"};
      $user_header .= &_toHex(length($key));
      $user_header .= $key;
      $user_header .= $self->{_type_to_key}->{"User Defined Value"};
      $user_header .= &_toHex(length($opt_headers->{$key}));
      $user_header .= $opt_headers->{$key};
    }

    # Concat sub headers
    my $encap = $md5_head.$filename_header.$blocknum_header.$user_header;
    # Header type + header size + header
    $encap = $self->{_lib_version}.&_toHex(length($encap)).$encap;
    $Trace->debug(msg => "Header encapsulation: ".$encap);
    # Header + data + footer
    $$data = $encap.$$data.$md5_str;

    $Trace->exit();
}

# Strip the header and footer from the data, return a hash.
sub _decapsulate {
    my $self = shift;
    my $block = shift;
    my %block_data = ();
    my $ID_bytes = $self->{_field_sizes}->{"_Type_bytes"} + $self->{_field_sizes}->{"_Len_bytes"};

    $Trace->entry();
    # Find header type
    my $type = substr($block, 0, $self->{_field_sizes}->{"_Type_bytes"});
    if ($type != $self->{_lib_version}) {
        my $err_str = "The block version [".$type."] does not match the library version [".$self->{_lib_version}."]!";
        $Trace->fatal(msg => $err_str);
    }
    # Find length of header data
    my $len = &_toDecimal(substr($block, $self->{_field_sizes}->{"_Type_bytes"}, $self->{_field_sizes}->{"_Len_bytes"}));
    # Grab MD5 footer
    $block_data{"MD5 Footer"} = substr($block, (length($block) - $self->{_field_sizes}->{"_MD5_size"}));
    # Extract data from encapsulation
    my $data = substr($block, ($len + $ID_bytes));
    $block_data{"Data"} = substr($data, 0, (length($data) - $self->{_field_sizes}->{"_MD5_size"}));
    # Grab remaining header
    $block = substr($block, $ID_bytes, $len);
    $block_data{$self->{_key_to_type}->{$type}} = $block;

    # Extract the header's sub headers
    my $tmp_key = "";
    while (length($block) > 0) {
      # Find header type
      $type = substr($block, 0, $self->{_field_sizes}->{"_Type_bytes"});
      if (! defined $self->{_key_to_type}->{$type}) {
        my $block_str = &_toString(\%block_data);
        my $hexStr = &_hexString($block_data{Data});
        my $err_str = "Invalid header type (".$type.") found, header appears to be corrupt!  See trace comments for data.";
        $Trace->comment(msg => "$block_str\nData in hex:\n$hexStr\n");
        $Trace->fatal(msg => $err_str);
      }
      # Find length of header data
      $len = &_toDecimal(substr($block, $self->{_field_sizes}->{"_Type_bytes"}, $self->{_field_sizes}->{"_Len_bytes"}));
      # If user defined headers, associate the value with the key
      if ($self->{_key_to_type}->{$type} eq "User Defined Key") {
        $tmp_key = substr($block, $ID_bytes, $len);
      } elsif ($self->{_key_to_type}->{$type} eq "User Defined Value") {
        $block_data{$tmp_key} = substr($block, $ID_bytes, $len);
      } elsif ($self->{_key_to_type}->{$type} eq "Block Number") {
        # Strip of extra 0's
        $block_data{$self->{_key_to_type}->{$type}} = int(substr($block, $ID_bytes, $len));
      } else {
        $block_data{$self->{_key_to_type}->{$type}} = substr($block, $ID_bytes, $len);
      }
      # Grab remaining header
      $block = substr($block, ($len + $ID_bytes));
    }

    $Trace->exit(msg => "Returning block hash.");
    return \%block_data;
}

# Seek to specified offset and write 'block_size' bytes
sub _writeBytes {
    my $self = shift;
    my $data = shift;
    my $offset = shift;
    my $seek_rc;
    my $write_rc;
    my $fh = $self->{_file_handle};

    $Trace->entry(msg => "Writing bytes to offset $offset of ".$self->{_filename});
    
    # Seek to offset
    seek($fh, $offset, SEEK_SET);
    $seek_rc = tell($fh);
    if ($seek_rc == -1) {
        $Trace->fatal(msg => "Failed seeking to offset $offset in ".$self->{_filename}." because: $!");
    }
    # Write 'block_size' bytes
    $Trace->debug(msg => "Writing to offset $offset of ".$self->{_filename});
    $write_rc = syswrite($fh, $data, $self->{_block_size});
    if (! defined $write_rc) {
        my $hexStr = &_hexString($data);
        my $err_str = "Failed to write block to ".$self->{_filename}." at offset $offset because: $!\nSee trace comments for data.";
        $Trace->comment(msg => "Attempted to write:\n$data\nData in hex:\n$hexStr\n");
        $Trace->fatal(msg => $err_str);
    } elsif ($write_rc != $self->{_block_size}) {
        my $hexStr = &_hexString($data);
        my $err_str = "Only wrote $write_rc of ".$self->{_block_size}." bytes to ".$self->{_filename};
        $err_str .= " at offset $offset\nSee trace comments for data.";
        $Trace->comment(msg => "Attempted to write:\n$data\nData in hex:\n$hexStr\n");
        $Trace->fatal(msg => $err_str);
    }

    $Trace->exit();
}

# Seek to specified offset and read 'block_size' bytes
sub _readBytes {
    my $self = shift;
    my $offset = shift;
    my $seek_rc;
    my $data = "";
    my $fh = $self->{_file_handle};

    $Trace->entry(msg => "Reading bytes from offset $offset of ".$self->{_filename});
    # Seek to offset
    seek($fh, $offset, SEEK_SET);
    $seek_rc = tell($fh);
    if ($seek_rc == -1) {
        $Trace->fatal(msg => "Failed seeking to position $offset of ".$self->{_filename}." because: $!");
    }
    # Read 'block_size' bytes
    $Trace->debug(msg => "Reading from offset $offset of ".$self->{_filename});
    my $bytes = sysread($fh, $data, $self->{_block_size});
    if ($bytes ne $self->{_block_size}) {
        my $err_str = "Failed to read block at offset $offset from ".$self->{_filename}.", only read $bytes of ".$self->{_block_size}." bytes";
        $Trace->fatal(msg => $err_str);
    }

    $Trace->exit(msg => "Returning read data.");
    return $data;
}

# Convert a decimal number to a hexidecimal number, padded with up to 4 0's.
sub _toHex {
    return sprintf("%04x", $_[0]);
}

# Convert a hexidecimal number to a decimal number.
sub _toDecimal {
    return hex($_[0]);
}

# When the instance is destroyed, make sure the filehandle is closed.
sub DESTROY {
    my $self = shift;
    
    # Close the file if it is open
    $self->close_file();
}

# Open a filehandle and store it in the instance
sub _openFile {
    my $self = shift;
    
    my $entry = "Opening file ".$self->{_filename};
    $entry .= ", MODE: read-only" if ($self->{_read_only});
    $Trace->entry(msg => $entry);
    my $fh;
    if ($self->{_read_only}) {
        # Open filename for reading only
        $fh = new IO::File $self->{_filename}, O_RDONLY;
    } else {
        # Open filename for reading and writing, create if DNE
        $fh = new IO::File $self->{_filename}, O_CREAT|O_RDWR, 0777;
    }
    if (! defined $fh) {
        $Trace->fatal(msg => "Failed to open file ".$self->{_filename}.": ".$!);
    }
    $self->{_file_handle} = $fh;
    
    $Trace->exit(msg => "Opened file ".$self->{_filename});
}

# Create a string representation of the block hash
sub _toString {
    my $block_data = shift;
    my $str = "Block:\n";

    $Trace->entry();
    for my $key (keys %$block_data) {
        $str .= $key." => ".$block_data->{$key}."\n";
    }

    $Trace->exit(msg => "Returning string representation of the block hash.");
    return $str."\n";
}

# Create a string representation of the data in hex
sub _hexString {
    my @dArr = split("", shift);
    my $count = 0;
    my $data = "";

    $Trace->entry();
    for (@dArr) {
        $data .= sprintf("%02x ", ord);
#        $data .= "\n" unless ++$count % 20;
    }
    $data .= "\n";
    $Trace->exit(msg => "Returning hex representation of data.");
    return $data;
}

=head1 STATUS

  Shared

=head1 AUTHOR

  Collin McHale (collinm@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/DataGV.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2008, NetApp, Inc. All Rights Reserved

=cut

1;
