# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/DataManip.pm#1 $

## @summary data manipulation NATE wrappers
## @author remi@netapp.com
## @status public
## @pod here

=head1 NAME

DataManip - NATE wrappers for DataManip::local

=head1 DESCRIPTION

The scripts in F</usr/local/test/bin/> are merely wrappers for the
functionality in C<DataManip::local>, calling that functionality
directly from the command-line.  This library is an equivalent set of
wrappers for calling that functionality directly from NATE scripts.

The major subroutines are the same as in C<DataManip::local>, except
that they expect a client hostspec as an additional first argument.
This specifies the client to run the data manipulation routine on.
The trees specified as other arguments must be mounted on and
accessible by the specified client.

The data manipulation routines receive a rather hearty collection of
behavior-modifying options.  The options documented with each function
below are only the options that are specific to these wrappers.  For
descriptions of all the other options, see C<DataManip::local>.  Here,
we also provide a C<env_options()> as a convenient way to import all
options from the environment, without having to modify scripts when
new options are introduced.  See the EXAMPLES section for an
illustration of this.

=head1 RETURN VALUES

The major routines here are wrapped-up to behave like the normal
C<DataManip::local> routines.  When unsuccessful, they return undef
and an error message in C<$@>, instead throwing an exception.  Though,
when successful, their return values are identical to their local
counterparts, since these routines take care of bundling-up and
extracting results from the remote subtests.

However, when running these routines in the background (by setting the
BACKGROUND option to 1), the abstraction breaks.  The calling script
needs to know a little bit about the remote subtests in order to
properly extract the results.  (If you only call these routines in the
foreground, then you do not have to worry about this extra work.)

If you run any of the major routines here in the background, then the
routine will return a subtest object (instead of the normal return
values).  At some point later on, you must then wait for that subtest
to complete by calling C<Subtest::wait_finish>, and then extract the
results with C<FSR::FSRutils::subtestResults>.  The result key names
which need to be passed to C<subtestResults> are named below with each
routine.  Any extracted result which is just a scalar value can then
be interpreted -- lists and hashes, however, are returned in serialized
form (see C<FSR::FSRutils::serialize>) so the returned value must first
be passed to C<eval> to get the list or hash back.  See the EXAMPLES
section for an illustration of this.

=head1 FUNCTIONS

=cut

package DataManip;
use warnings;
use strict;

BEGIN {
    require Exporter;
    our @ISA = qw(Exporter);
    our @EXPORT = qw(env_options
		     populate_opts populate_tree
		     modify_opts modify_tree
		     compare_opts compare_trees
		     delete_opts delete_tree);
    our @EXPORT_OK = qw(manipulate_tree);
}

use Carp;
use File::Spec qw();
use Tharn;
use FSR::FSRutils;
use DataManip::local qw(:DEFAULT legacy_options_compatibility);

# Given an options hash and an option name,
# deletes the option from the hash and returns its former value,
# or zero if the option was not in the hash.
sub extract_option( \%$ ) {
  my $options = shift;
  my ($opt) = @_;

  my $val = 0;
  if (exists $options->{$opt}) {
    if (defined $options->{$opt}) {
      $val = $options->{$opt};
    }
    delete $options->{$opt};
  }
  return $val;
}

# Strips whitespace from a code fragment.
sub strip_whitespace( $ ) {
  $_[0] =~ s/\n\s*/ /g; # eliminate newlines and extra whitespace
  $_[0] =~ s/\A\s*//g;  # strip leading whitespace
  $_[0] =~ s/\s*\z//g;  # strip trailing whitespace
}

=head2 env_options(@options)

Given a list of option names, returns a hash of those options to their
values as defined by environment variables of the same names.

=cut

sub env_options {
  my %opts = ();

  foreach (@_) {

    if (defined $ENV{$_}) {
      $opts{$_} = $ENV{$_};
    }

    # Unfortunately, this doesn't work unless the param()
    # call was used to grab the value from the environment.
#     no strict 'refs';
#     if (defined $ { "::$_" }) {
#       push @params, $_ . "=" . $ { "::$_" };
#     }
#     use strict 'refs';
  }

  return %opts;
}

# Manipulates the subtree on the given client, with the given manipulation.
# Returns the subtest object for the remote subtest.
sub manipulate_subtree( $$$% ) {
  my ($manip, $CLIENT, $tree, %opts) = @_;

  my $code = <<'CODE';
  {
    use DataManip::cmd_utils qw(serialize);
CODE
  $code .= <<"CODE";
    use DataManip::local qw(${manip}_local_tree);
    my (\$seed, \$subpaths, \$subseeds) = eval { &${manip}_local_tree(\@_) };
CODE
  $code .= <<'CODE';
    logresult("FATAL", "EVAL_ERROR=$@") unless ($seed);
    logresult("PASS", "SEED=$seed",
              "SUBPATHS=" . serialize(@$subpaths),
              "SUBSEEDS=" . serialize(@$subseeds));
  }
CODE
  strip_whitespace($code);

  # Forces the options used to get printed in the logs
  $opts{SHOW_OPTS} = 1;

  my $obj = subtest($code,
		    '-bg',
		    '-client', $CLIENT,
		    '-onexit', sub {},
		    "--", $tree, %opts);
  return $obj;
}

# Manipulates the tree on the given client with the given manipulation,
# and similarly manipulates all returned subtrees on the same client,
# if any are returned.
# Logs FATAL/FAIL results if any of the modifications fail.
# Otherwise, logs a PASS result with the SEED, SUBPATHS, and SUBSEEDS keys.
sub manipulate_tree_subtest( $$% ) {
  my ($manip, $CLIENT, $tree, %opts) = @_;

  my $main_obj = manipulate_subtree($manip, $CLIENT, $tree, %opts);
  Subtest::wait_finish(subtests => [$main_obj]);
  my $main_rslts = subtestResults($main_obj, qw(SEED SUBPATHS SUBSEEDS));
  logerror() unless ($main_rslts);

  my $seed = $main_rslts->{SEED};
  my @subpaths = eval $main_rslts->{SUBPATHS};
  logerror() if ($@);
  my @subseeds = eval $main_rslts->{SUBSEEDS};
  logerror() if ($@);
  unless (scalar(@subpaths) == scalar(@subseeds)) {
    logerror(message => "subseed count does not match subpath count");
  }

  my $error = 0;
  if (scalar(@subpaths)) {
    my %sub_opts = %opts;
    $sub_opts{FORK_COUNT} = 0;
    my $depth_base = (defined $sub_opts{DEPTH_BASE} ? $sub_opts{DEPTH_BASE} : 0);
    my @mods = ();
    foreach my $subpath (@subpaths) {
      my $subtree = File::Spec->catdir($tree, $subpath);
      $subpath = File::Spec->canonpath($subpath);
      my @dirs = File::Spec->no_upwards( File::Spec->splitdir($subpath) );
      $sub_opts{DEPTH_BASE} = $depth_base + scalar(@dirs);
      $sub_opts{SEED} = shift @subseeds;
      push @mods, manipulate_subtree($manip, $CLIENT, $subtree, %sub_opts);
    }
    Subtest::wait_finish(subtests => \@mods);

    foreach my $obj (@mods) {
      my $rslts = subtestResults($obj, qw(SEED SUBPATHS SUBSEEDS));
      if ($rslts) {
	my @subpaths = eval $rslts->{SUBPATHS};
	if ($@) {
	  logerror(abort => 0);
	  $error = 1;
	}
	if (scalar(@subpaths)) {
	  logerror(message => "received " . scalar(@subpaths) . " subpaths?",
		   abort => 0);
	  $error = 1;
	}
	my @subseeds = eval $rslts->{SUBSEEDS};
	if ($@) {
	  logerror(abort => 0);
	  $error = 1;
	}
	if (scalar(@subseeds)) {
	  logerror(message => "received " . scalar(@subseeds) . " subseeds?",
		   abort => 0);
	  $error = 1;
	}
      } else {
	logerror(abort => 0);
	$error = 1;
      }
    }
  }

  unless ($error) {
    logresult('PASS', "SEED=$seed",
	      "SUBPATHS=" . serialize(),
	      "SUBSEEDS=" . serialize());
  }
}

sub manipulate_tree( $$$% ) {
  my ($manip, $CLIENT, $tree, %opts) = @_;

  unless ($manip =~ /\A(populate|modify)\z/) {
    die "unsupported manipulation \"$manip\"\n";
  }

  my $bg = extract_option(%opts, 'BACKGROUND');
  my $ret_paths = extract_option(%opts, 'RETURN_PATHS');

  if (extract_option(%opts, 'PRECREATE') && $manip eq 'populate') {
    my $code = <<'CODE';
    {
      use DataManip::local qw(mkdir_path);
      if (&mkdir_path(@_)) {
        logresult("PASS");
      } else {
        logresult("FATAL", "EVAL_ERROR=$!");
      }
    }
CODE
    strip_whitespace($code);

    my $obj = subtest($code,
		      '-client', $CLIENT,
		      '-onexit', sub {},
		      "--", $tree);
    unless (subtestResults($obj)) {
      # Could not create the directory.
      return undef;
    }
  }

  my $obj;
  if ($ret_paths) {
    $obj = manipulate_subtree($manip, $CLIENT, $tree, %opts);
  } else {
    $obj = subtest( sub { manipulate_tree_subtest($manip, $CLIENT, $tree, %opts) },
		    '-bg',
		    '-onexit', sub {});
  }

  if ($bg) {
    return $obj;
  } else {
    logcomment("Waiting...");
    Subtest::wait_finish(subtests => [$obj]);
    logcomment("Finished waiting.");
    my $rslts = subtestResults($obj, qw(SEED SUBPATHS SUBSEEDS));
    if ($rslts) {
      my @subpaths = eval $rslts->{SUBPATHS};
      return undef if ($@);
      my @subseeds = eval $rslts->{SUBSEEDS};
      return undef if ($@);
      return ($rslts->{SEED}, \@subpaths, \@subseeds);
    } else {
      return undef;
    }
  }
}


=head2 populate_tree($CLIENT, $tree, %options)

Populates the specified directory on the given client with a complete
data tree.  Exactly what is created in this data tree is determined by
the options specified.

On success, returns the random seed used, and a reference to a list
of subpaths, and a reference to a list of subseeds.
(See the FORK_COUNT option.)
On failure, returns undef with an error message in C<$@>.

This wrapper makes one small addendum to the functionality of its
local counterpart.  If the FORK_COUNT option is specified and
non-zero, then normally a list of subpaths and subseeds would be
returned, with the expectation that the caller would re-invoke the
routine concurrently on each subpath with the corresponding subseed
and other options the same.
By default, however, this wrapper does that all for you.  If any
subpaths are returned by the initial remote subtest, then the wrapper
automatically starts concurrent remote subtests on each of the
returned paths.  You can prevent this behavior with the RETURN_PATHS
option.

Options:

=over 4

=item BACKGROUND

If specified, runs the operation in the background, and returns a
subtest object instead of the normal return values.  The calling
script will need to extract the following results once the subtest
completes:

=over 4

=item SEED

The random seed used (as a simple scalar).

=item SUBPATHS

A list of returned subpaths (as a serialized list).  This list should
generally be empty unless the FORK_COUNT option was specified and
non-zero.

=item SUBSEEDS

A list of returned subseeds (as a serialized list).  This list should
have the same number of elements as the SUBPATHS.  The seeds here
should be used when populating the corresponding subpaths.

=back

=item RETURN_PATHS

If set, returns the subpaths and subseeds when the FORK_COUNT option
is set, instead of automatically reinvoking this routine on all returned
subpaths concurrently.  This allows the calling routine to manually
parallelize the execution on multiple subpaths, possibly on different
clients.

=item PRECREATE

If specified, then prior to populating the directory checks to see
whether it exists, and if not creates it (along with any necessary
parent directories).

=back

=cut

sub populate_tree( $$% ) {
  my $CLIENT = shift;
  my $tree = shift;
  my %opts = @_;
  if (legacy_options_compatibility(%opts, @_)) {
    logresult(type => 'WARN', msg => "Using list-style options to " .
	      "populate_tree.  This will be a SCRIPT error soon. " .
	      "Fix your script.");
  }
  return manipulate_tree('populate', $CLIENT, $tree, %opts);
}


=head2 modify_tree($CLIENT, $tree, %options)

Walks the specified tree on the given client and modifies its
contents.  The nature and extent of the modifications is determined by
the options specified.

On success, returns the random seed used, and a reference to a list
of subpaths, and a reference to a list of subseeds.
(See the FORK_COUNT option.)
On failure, returns undef with an error message in C<$@>.

This wrapper makes one small addendum to the functionality of its
local counterpart.  If the FORK_COUNT option is specified and
non-zero, then normally a list of subpaths and subseeds would be
returned, with the expectation that the caller would re-invoke the
routine concurrently on each subpath with the corresponding subseed
and other options the same.
By default, however, this wrapper does that all for you.  If any
subpaths are returned by the initial remote subtest, then the wrapper
automatically starts concurrent remote subtests on each of the
returned paths.  You can prevent this behavior with the RETURN_PATHS
option.

Options:

=over 4

=item BACKGROUND

If specified, runs the operation in the background, and returns a
subtest object instead of the normal return values.  The calling
script will need to extract the following results once the subtest
completes:

=over 4

=item SEED

The random seed used (as a simple scalar).

=item SUBPATHS

A list of returned subpaths (as a serialized list).  This list should
generally be empty unless the FORK_COUNT option was specified and
non-zero.

=item SUBSEEDS

A list of returned subseeds (as a serialized list).  This list should
have the same number of elements as the SUBPATHS.  The seeds here
should be used when modifying the corresponding subpaths.

=back

=item RETURN_PATHS

If set, returns the subpaths and subseeds when the FORK_COUNT option
is set, instead of automatically reinvoking this routine on all returned
subpaths concurrently.  This allows the calling routine to manually
parallelize the execution on multiple subpaths, possibly on different
clients.

=back

=cut

sub modify_tree( $$% ) {
  my $CLIENT = shift;
  my $tree = shift;
  my %opts = @_;
  if (legacy_options_compatibility(%opts, @_)) {
    logresult(type => 'WARN', msg => "Using list-style options to " .
	      "modify_tree.  This will be a SCRIPT error soon. " .
	      "Fix your script.");
  }
  return manipulate_tree('modify', $CLIENT, $tree, %opts);
}


=head2 compare_trees($CLIENT, $tree1, $tree2, %options)

Compare the two specified trees on the given client, and their
contents.

On success, returns three references, each of which refers to a list
of paths relative to the base of the trees being compared.

=over 4

=item -

The first is a list of paths in $tree1 but not $tree2.

=item -

The second is a list of paths in $tree2 but not $tree1.

=item -

The third is a list of paths in both trees, but whose contents or
attributes differed.

=back

On failure, returns undef with an error message in $@.

Options:

=over 4

=item BACKGROUND

If specified, runs the operation in the background, and returns a
subtest object instead of the normal return values.  The calling
script will need to extract the following results once the subtest
completes:

=over 4

=item IN1

A list of paths in $tree1 (as a serialized list).

=item IN2

A list of paths in $tree2 (as a serialized list).

=item DIFFER

A list of paths in both trees (as a serialized list), whose contents
or attributes differed.

=back

=back

=cut

sub compare_trees($$$%) {
  my $CLIENT = shift;
  my $tree1 = shift;
  my $tree2 = shift;
  my %opts = @_;
  if (legacy_options_compatibility(%opts, @_)) {
    logresult(type => 'WARN', msg => "Using list-style options to " .
	      "compare_trees.  This will be a SCRIPT error soon. " .
	      "Fix your script.");
  }

  my $code = <<'CODE';
  {
    use DataManip::local qw(compare_local_trees);
    use FSR::FSRutils qw(serialize);
    my ($in1, $in2, $differ, $subpaths) = eval { &compare_local_trees(@_) };
    logresult("FATAL", "EVAL_ERROR=$@") if ($@);
    logresult("PASS",
	      "IN1=" . serialize(@$in1),
	      "IN2=" . serialize(@$in2),
	      "DIFFER=" . serialize(@$differ),
              "SUBPATHS=" . serialize(@$subpaths));
  }
CODE
  strip_whitespace($code);

  # Forces the options used to get printed in the logs
  $opts{SHOW_OPTS} = 1;

  my $bg = extract_option(%opts, 'BACKGROUND');

  my $obj = subtest($code,
		    '-bg',
		    '-client', $CLIENT,
		    '-onexit', sub {},
		    '--', $tree1, $tree2, %opts);

  if ($bg) { # User told us to run in the background.
    return $obj;
  }
  Subtest::wait_finish(subtests => [$obj]);
  my $rslts = subtestResults($obj, qw(IN1 IN2 DIFFER SUBPATHS));
  if ($@) {
    return undef;
  } else {
    my @in1 = eval $rslts->{IN1};
    return undef if ($@);
    my @in2 = eval $rslts->{IN2};
    return undef if ($@);
    my @differ = eval $rslts->{DIFFER};
    return undef if ($@);
    my @subpaths = eval $rslts->{SUBPATHS};
    return undef if ($@);
    return (\@in1, \@in2, \@differ, \@subpaths);
  }
}

=head2 delete_tree($CLIENT, $tree, %options)

Deletes the specified tree on the given client, and everything under
it.

On success, returns 1.
On failure, returns undef with an error message in C<$@>.

Options:

=over 4

=item BACKGROUND

If specified, runs the operation in the background, and returns a
subtest object instead of the normal return values.  The calling
script will need to use C<FSR::FSRutils::subtestResults> to determine
whether it succeded or failed, but there are no named results to
extract.

=back

=cut

sub delete_tree($$%) {
  my $CLIENT = shift;
  my $tree = shift;
  my %opts = @_;
  if (legacy_options_compatibility(%opts, @_)) {
    logresult(type => 'WARN', msg => "Using list-style options to " .
	      "delete_tree.  This will be a SCRIPT error soon. " .
	      "Fix your script.");
  }

  my $code = <<'CODE';
  {
    use DataManip::local qw(delete_local_tree);
    eval { &delete_local_tree(@_) };
    logresult("FATAL", "EVAL_ERROR=$@") if ($@);
    logresult("PASS");
  }
CODE
  strip_whitespace($code);

  # Forces the options used to get printed in the logs
  $opts{SHOW_OPTS} = 1;

  my $bg = extract_option(%opts, 'BACKGROUND');

  my $obj = subtest($code,
		    '-bg',
		    '-client', $CLIENT,
		    '-onexit', sub {},
		    "--", $tree, %opts);

  if ($bg) { # User told us to run in the background.
    return $obj;
  }
  Subtest::wait_finish(subtests => [$obj]);
  my $rslts = subtestResults($obj);
  if ($@) {
    return undef;
  } else {
    return 1;
  }
}

=head2 populate_opts

Returns a list of legal option names which C<populate_tree> accepts.

=cut

sub populate_opts() {
  my @option_names = qw(SEED ENVIRONMENT_TYPE PRECREATE);
  push @option_names, @DataManip::local::populate_options;
  return @option_names;
}

=head2 modify_opts

Returns a list of legal options names which C<modify_tree> accepts.

=cut

sub modify_opts() {
  my @option_names = qw(SEED ENVIRONMENT_TYPE);
  push @option_names, @DataManip::local::modify_options;
  return @option_names;
}

=head2 compare_opts

Returns a list of legal option names which C<compare_trees> accepts.

=cut

sub compare_opts() {
  my @option_names = @DataManip::local::compare_options;
  return @option_names;
}

=head2 delete_opts

Returns a list of legal option names which C<delete_tree> accepts.

=cut

sub delete_opts() {
  my @option_names = @DataManip::local::delete_options;
  return @option_names;
}

=head1 EXAMPLES

Say we have a client hostspec in C<$CLIENT>, and that two data trees
mounted on that client in C<$tree_src> and C<$tree_dst> (presumably
since they are the source and destination for a mirror relationship.

  # First, modify the source tree.
  my %options = env_options(modify_opts());
  unless (modify_tree($CLIENT, $tree_src, %options)) {
      # Modification failed for some reason.
      # Error message is in $@.
      logresult('FATAL', $@);
  }

This code snippet calls the C<modify_opts> routine to get all the
available options for C<modify_tree>.  It then passes that list to
C<env_options> to retrieve any such options defined in the
environment, which would be the case if the user specified them as
parameters to C<ntest>.

Later on, if new options are added to C<modify_tree>, this piece of
code will not need to change.  A user calling this script can just
specify the new option as a parameter to C<ntest>, and this code will
correctly notice it and pass it to C<modify_tree>.

IF we wanted to explicitly set any options to C<modify_tree> (possible
overriding what the user gave as parameters to C<ntest>), then we
would simple set those options in the C<%options> hash before passing
it to C<modify_tree>.

Now, say that the modification of the source tree mounted on our
client succeeds, and we then complete an update of the mirror.  At
this point, the destination tree should match the source tree again.

  # The trees should match again, so compare them.
  my %options = env_options(compare_opts());
  my ($in1, $in2, $differ) =
      compare_trees($CLIENT, $tree_src, $tree_dst, %options);
  if ($@) {
      # Comparison failed for some reason.
      # Error message is in $@.
      logresult('FATAL', $@);
  }
  if (scalar(@$in1) || scalar(@$in2) || scalar(@$differ)) {
      # The trees did not match!
      logresult('FATAL',
                "source and destination are not identical");
      # We could report exactly what the differences were
      # here, by logging the paths in @$in1, @$in2, and
      # @$differ, but this is just an example.
  }

Likewise, this code will not need to change if more options are added
to C<compare_trees>, since the routines will automatically notice and
utilize the new options.

Now, say we knew that the comparison was going to take a long time,
and we wanted to run some other code at the same time.  The following
code snippet does the same comparison as above, but does so in the
background, so that it can run other code in the foreground while the
comparison is active.  This also illustrates how to extract and use
results from routines run in the background.

  # The trees should match again, so compare them.
  my %options = env_options(compare_opts());
  $options{BACKGROUND} = 1;
  my $cmp_obj = compare_trees($CLIENT, $tree_src,
                              $tree_dst, %options);

  # Comparison is running now.  Do other work...
  unless (super_duper_test_code(@args)) {
      logresult('FATAL', "other code failed");
  }

  # OK, we are done running other code.

  # First, wait for the comparison to finish
  # (in case it did not already finish while
  # we were running the other code).
  Subtest::wait_finish(subtests => [ $cmp_obj ]);

  # Next, check whether the comparison succeeded,
  # and extract the results if it did.
  my $rslts = subtestResults($cmp_obj,
                             qw(IN1 IN2 DIFFER));
  if ($@) {
      # Comparison failed for some reason.
      # Error message is in $@.
      logresult('FATAL', $@);
  }

  # Results are in %$rslts, but all the lists are
  # serialized.  Run eval to get back the real lists.
  my @in1 = eval $rslts->{IN1};
  my @in2 = eval $rslts->{IN2};
  my @differ = eval $rslts->{DIFFER};

  # Finally, we can check if the trees matched.
  if (scalar(@in1) || scalar(@in2) || scalar(@differ)) {
      # The trees did not match!
      logresult('FATAL',
                "source and destination are not identical");
      # We could report exactly what the differences were
      # here, by logging the paths in @in1, @in2, and
      # @differ, but this is just an example.
  }

Notice that the serialized lists are returned as real lists, not as
list references.  When run in the foreground, they are returned as
list references.

=head1 BUGS

These routines really ought to provide a way to employ the
C<-username> option to C<subtest()> in order to run the manipulation
subtests as a specific user.  This will enable users to circumvent all
those pesky permissions issues by running manipulations as root (but
only in correctly set-up environments, where being root actually
helps).

We need the somewhat generic ability to manipulate a dataset on multiple
clients concurrently using the FORK_COUNT option.  Basically, what
F<util/manipulate.thpl> does, but as a library call instead of user script.

=cut


1;
