# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/DataManip/cmd_utils.pm#1 $

## @summary Command-line utilities for DataManip scripts
## @author remi@netapp.com
## @status public
## @pod here

=head1 NAME

cmd_utils - Command-line utilities for DataManip scripts

=head1 DESCRIPTION

This module defines routines common amongst the command-line scripts
that employ C<DataManip::local> functionality.  These scripts typically
reside in F</usr/software/test/bin/>.

Unless you are debugging one of these scripts, you probably have no
use for anything in this file.

=cut

package DataManip::cmd_utils;

use strict;

use base qw(Exporter);
BEGIN {
  my @wrapper = qw(args_to_hash pod_usage timer
		   print_pathlist path_sanity_check deviations_sanity_check);
  my @other = qw(usage serialize second_str human_size
		 print_scramble gauss_stats populate_estimates
		 child_pipe find_library pod_render);

  our %EXPORT_TAGS = (wrapper => [ @wrapper ],
		     );
  $EXPORT_TAGS{'all'} = [ @wrapper, @other ];
  Exporter::export_ok_tags('all');
}

use Carp;
use FileHandle;
use Pod::Select qw(podselect);
use File::Temp qw(tempfile);
use File::Spec qw();

=head1 FUNCTIONS

=head2 usage(%options)

This is just a common usage function for the DataManip wrapper scripts
in F</usr/software/test/bin/>.

Options:

=over 4

=item header

Specifies an optional header to printed before the usage string.

=item args

Specifies the usage string to for non-option arguments
(default="<tree>").

=item estimator

Specifies whether the command supports the C<-e> option.

=back

All other options passed are interpreted as additional DataManip-style
options that are specific to the calling script, and the option values
are interpreted as the descriptions of those options.

=cut

sub usage( % ) {
  my %own_opts = @_;
  my $pre = "";
  if (exists $own_opts{header}) {
    $pre = $own_opts{header} . "\n";
    delete $own_opts{header};
  }

  my $args;
  if (exists $own_opts{args}) {
    $args = $own_opts{args};
    delete $own_opts{args};
  } else {
    $args = "<tree>";
  }

  my $estimator = 0;
  if (exists $own_opts{estimator}) {
    $estimator = $own_opts{estimator};
    delete $own_opts{estimator};
  }

  # Format other DataManip-style options and descriptions.
  my $post = "";
  while (my ($k, $v) = each %own_opts) {
    my @lines = ();
    while (($v =~ s/\A\s*(.{0,43}\S)(\s+|\z)//) ||
	   ($v =~ s/\A\s*(\S+)(\s+|\z)//)) {
	push @lines, $1;
    }
    push @lines, $v if ($v);
    $post .= sprintf("   %-14s %s\n", $k, shift @lines);
    while (scalar(@lines)) {
      $post .= sprintf("                  %s\n", shift @lines);
    }
  }
  if ($post) {
    $post = "DataManip OPTIONS (specific to this script):\n" . $post;
  }

  # Reduce the path to the program to just the program name.
  $0 =~ s/\A.*(?:\/|\\)(.*?)\z/$1/;

  die $pre .
      "usage: $0 <options> <tree> [<DataManip OPTIONS>]\n" .
      "Options:\n" .
      "   -h             get help on DataManip OPTIONS\n" .
      "   -h <module>    get help on a specific module\n" .
      "                  (e.g. DataManip::local)\n" .
      "   -q             quiet mode (only print fatal errors)\n" .
      (! $estimator ? "" :
      "   -e <N>         estimate effects within N standard deviations\n" .
      "                  (but do nothing else)\n"
      ) .
      $post;
}


=head2 timer

Records/computes elapsed time and processing times for a section
of code, and any child processes that complete during it.

I<Example>:

  my %t = timer(); # start timer
  code_section();
  %t = timer(%t); # end timer

=cut

sub timer( ;% ) {
  my (%s) = @_;

  my %t = ();
  $t{elapsed} = time;
  ($t{user}, $t{system}, $t{cuser}, $t{csystem}) = times();

  if (scalar(@_)) {
    foreach my $key (qw(elapsed user system cuser)) {
      $t{$key} -= $s{$key};
    }
  }
  return %t;
}


=head2 second_str($seconds)

Converts a number of seconds into a human-readable string.

=cut

sub second_str( $ ) {
  my ($epoch) = @_;
  my $partial = 0;
  croak "Bad time: $epoch\n" unless ($epoch =~ s/\A(\d+)(\.\d+)?\z/$1/);
  $partial = $2 if ($2);

  my $remainder = int($epoch / 60);
  my $hours = int($remainder / 60);
  my $minutes = $remainder % 60;
  my $seconds = $epoch % 60;
  my $msecs = int(($epoch - (($remainder * 60) + $seconds)) * 1000);

  my $short = 0;
  my $ret = "";
  if ($hours >= 24) {
    my $days = int($hours / 24);
    $hours = $hours % 24;
    $ret .= "$days days, ";
    $short = 1;
  }
  if ($short || $hours) {
    $ret .= sprintf("%02d:", $hours);
    $short = 1;
  }
  if ($short || $minutes) {
    $ret .= sprintf("%02d:", $minutes);
    $short = 1;
  }
  $ret .= sprintf("%02d.%03d", $seconds, $msecs);
  return $ret;
}

=head2 human_size($byte_count)

Returns a byte-count in human-readable format (MB, GB, etc.)

=cut

sub human_size( $ ) {
  my ($size) = @_;
  my $type = "B";
  if ($size > 1024) {
    $size /= 1024;
    $type = "KB";
    if ($size > 1024) {
      $size /= 1024;
      $type = "MB";
      if ($size > 1024) {
	$size /= 1024;
	$type = "GB";
	if ($size > 1024) {
	  $size /= 1024;
	  $type = "TB";
	  if ($size > 1024) {
	    $size /= 1024;
	    $type = "PB";
	  }
	}
      }
    }
  }
  return sprintf("%.2f $type", $size);
}

=head2 args_to_hash(@args)

Takes a list of C<KEY=VALUE> strings, and returns
the corresponding hash of C<KEY => VALUE> pairs.
Returns undef with an error in $@ on failure.

=cut

sub args_to_hash {
  my %hash = ();
  my $pat_ident = "[A-Za-z_]\\w*?";
  foreach $_ (@_) {
    croak "Error parsing $_\n"
      unless ($_ =~ /^($pat_ident)=(.*)$/o);
    $hash{$1} = $2;
  }
  return %hash;
}


=head2 serialize

Given a list or a hash, serializes it into a string.  That string
can then be passed to eval to obtain the original list or hash.

=cut

sub serialize {
  my @strings;
  foreach my $arg (@_) {
    my $str = $arg; # copy to avoid modifying originals
    if (defined $str) {
      $str =~ s|\\|\\\\|g;
      $str =~ s|\(|\\\(|g;
      $str =~ s|\)|\\\)|g;
      push @strings, "q($str)";
    } else {
      push @strings, 'undef';
    }
  }
  return "( " . join(", ", @strings) . " )";
}


=head2 print_pathlist($header, $pathlist_ref)

Takes a header string and a reference to a list of paths,
and prints them on multiple lines.  If the header string
contains a %d, it is replaced with the number of paths.

=cut

sub print_pathlist( $$ ) {
  my ($header, $pathlist) = @_;
  printf($header, (($header =~ /%d/) ? (scalar(@$pathlist)): ()));
  print "\n\t" . join("\n\t", @$pathlist) . "\n";
}


=head2 path_sanity_check(@paths)

Takes a list of paths, and prints a warning message to STDERR
if any of the paths do not exist or look suspicious.  This is
useful for checking positional user arguments that may be out
of place.

=cut

sub path_sanity_check {
  foreach my $path (@_) {
    print STDERR "Warning: $path doesn't look like a file or directory.\n"
      if ((! -e $path) && ($path =~ /\=/));
  }
}

=head2 deviations_sanity_check($cnt)

Just a quick sanity check (and informative error messages)
for the deviations argument to '-e'.

=cut

sub deviations_sanity_check( $ ) {
  my ($deviations) = @_;

  unless ($deviations =~ /\A(\d*\.)?\d+\z/) {
    die "Standard deviations argument to '-e' must be numeric.\n";
  }
  unless ($deviations > 0) {
    die "Standard deviations argument to '-e' must be positive.\n";
  }
  unless ($deviations <= 3) {
    die "Standard deviations argument to '-e' cannot be greater than 3.\n";
  }
}


=head2 find_library($symbol)

Takes a complete package name (e.g. MyPackage::Foo::Symbol), and
searches @INC to find either the library specified
(e.g. F</usr/lib/MyPackage/Foo/Symbol.pm>, or a library that exports
the symbol (e.g. F</usr/lib/MyPackage/Foo.pm>, presumably exporting
C<Symbol>).

Returns the path to the library file, the name of the exported symbol
(if any), and the path of the library file relative to @INC.  Returns
undef if the library file could not be located.

=cut

sub find_library( $ ) {
  my ($package) = @_;
  my @parts = split(/::/, $package);
  my $lib1 = File::Spec->catdir(@parts);
  $lib1 .= ".pm" unless ($lib1 =~ /\.pm\z/);
  my $sym = pop @parts;
  my $lib2 = File::Spec->catdir(@parts) . ".pm";

  foreach my $dir (@INC) {
    my $file = File::Spec->catdir($dir, $lib1);
    if (-f $file) {
      return (wantarray ? ($file, undef, $lib1) : $file);
    }
    $file = File::Spec->catdir($dir, $lib2);
    if (-f $file) {
      return (wantarray ? ($file, $sym, $lib2) : $file);
    }
  }
  return undef;
}


=head2 child_pipe($pipe_command, $input_filehandle)

Forks off a child process to run something and produce output which
can be read by the calling process.  On success, returns a FileHandle
which the caller can then read from to get the output of the child.

The arguments are a pipe command and optionally an input FileHandle.
The pipe command must either be a code reference or a string
representing a shell command to run, both of which should print to
STDOUT.  The optional input FileHandle specifies a data stream that
should be provided as input (on STDIN) to the child process running
the pipe command.

=cut

sub child_pipe {
  my ($pipecmd, $input_fh) = @_;

  # Make sure the arguments are right.
  if ($input_fh) {
    my $fh_ref = ref($input_fh);
    unless ($fh_ref eq 'FileHandle' || $fh_ref eq "IO::Handle") {
      croak "invalid FileHandle argument: $fh_ref\n";
    }
  }

  my ($cmd, $cmd_write) = FileHandle::pipe();
  unless (ref($cmd) && ref($cmd_write)) {
    croak "pipe: $!\n";
  }

  my $pid = fork();
  if (! defined $pid) {
    croak "fork: $!\n";
  }
  if ($pid == 0) {
    autoflush STDOUT 1;
    $cmd->close() or croak "close: $!\n";

    # Dup the input filehandle to STDIN.
    if ($input_fh) {
      open(STDIN, "<&" . $input_fh->fileno()) or croak "dup2: $!\n";
    }

    # Dup the write-end of the pipe to STDOUT.
    open(STDOUT, ">&" . $cmd_write->fileno()) or croak "dup2: $!\n";

    # We'll now read from STDIN (really the input filehandle),
    # and produce output to STDOUT (really a pipe back to our parent),
    # so just do our thing and exit.
    if (ref($pipecmd) eq 'CODE') {
      $pipecmd->();
      exit(0);
    } else {
      exec($pipecmd) or die $!;
    }
  }

  $cmd_write->close() or croak "close: $!\n";

  return $cmd;
}

=head2 pod_render($pod_file, %options)

Extracts the relevant pod from the given file, formats it, and sends
it to the PAGER specified in the environment.  This function is the
core of what C<pod_usage>, wrapped up separately for applications that
need to use it separately.

Options:

=over 4

=item title

Title to use on the pod manual page.  Typically, this is the name
of the library from which the pod originates.

=item section

The section of pod to display from the given file.
(I.e. a C<=head1> heading.)

=item subsection

The subsection of pod to display from the given file.
This option is only valid when the section is also specified.
(I.e. a C<=head2> heading.)

=back

=cut

sub pod_render( $% ) {
  my ($file, %opts) = @_;
  unless (-f $file) {
    croak "no such file $file\n";
  }

  # Keep a logical name of the topic we are looking for.
  my $topic = $file;
  # Reduce the file path to just the file name, withouth the .pm extension.
  $topic =~ s/\A.*(?:\/|\\)(.*?)(?:\.pm)?\z/$1/;

  my %podselect_args = ();
  if ($opts{section}) {
    $topic .= "/" . $opts{section};
    my $regexp;
    if ($opts{subsection}) {
      $topic .= "/" . $opts{subsection};
      $regexp = sprintf('%s/%s\s*(?:\z|\().*',
			quotemeta($opts{section}),
			quotemeta($opts{subsection}));
    } else {
      $regexp = quotemeta($opts{section});
    }
    $podselect_args{-sections} = [$regexp];
  }

  # Do not try fancy unsupported pipe stuff on Windows.
  if ($^O eq "MSWin32") {
    print "Relevant POD located in $file.\n";

    # Make a temporary file to hold the POD.
    autoflush STDOUT 1;
    my ($fh, $filename) = tempfile(UNLINK => 1);
    unless ($fh && $filename) {
      croak "cannot create temporary file: $!\n";
    }
    $podselect_args{-output} = $filename;
    podselect(\%podselect_args, $file);

    # Use system() instead of exec() so that
    # we can wait for output to flush.
    my $cmd = 'pod2text';
    my $pid = system(1, $cmd, $filename);
    my $rc = waitpid($pid, 0);
    if ($rc != $pid) {
      croak "could not start or locate $cmd process $pid: $rc\n";
    } elsif ($?) {
      croak "$cmd process $pid exited with error code $?\n";
    }
    exit(0);
  }

  # Set up subroutines for the pipe stream.
  my $podref = sub { podselect(\%podselect_args, $file); };
  my $cntref = sub { my $cnt = 0;
		     while (<STDIN>) { print; $cnt++; }
		     unless ($cnt > 0) {
		       print "=pod\n\nNo POD present for C<$topic>.\n\n=cut\n";
		     }
		   };

  my $pod2man_cmd = "pod2man --section 3";
  if ($opts{title}) {
    $pod2man_cmd .= " --center " . $opts{title};
  }

  # Do a pipe-stream to find the pod and send it to the pager.
  my $pod = child_pipe($podref);
  my $cnt = child_pipe($cntref, $pod);
  my $man = child_pipe($pod2man_cmd, $cnt);
  my $text = child_pipe("nroff -man", $man);
  open(STDIN, "<&" . $text->fileno()) or croak "dup2: $!\n";
  $ENV{PAGER} = 'less' unless (exists $ENV{PAGER});
  exec($ENV{PAGER}) or croak "exec($ENV{PAGER}): $!\n";
}

=head2 pod_usage($function)

Produces a usage message for the library or function specified as a
complete package and symbol name (e.g. Foo::Bar::Symbol), by
searching for and extracting the related POD from the library file.

Given the package name, searches to find either the appropriate
library file, or a library that presumably exports the final symbol in
the package name.  It then searches the library file for all POD (or
just POD regarding the exported symbol), extracts the POD, formats it,
and sends it to the PAGER specified in the environment.

Croaks on any error.
(Otherwise, never returns since it executes the pager.)

=cut

sub pod_usage( $ ) {
  my ($package) = @_;
  my ($file, $symbol, $lib) = find_library($package);
  unless ($file) {
    die "could not locate documentation for $package\n";
  }

  my %opts = (title => $lib);
  if ($symbol) {
    $opts{section} = 'FUNCTIONS';
    $opts{subsection} = $symbol;
  }
  pod_render($file, %opts);
}

=head2 print_scramble($scramble_value, $noun)

Prints an English description of what the given scramble value will do.

=cut

sub print_scramble( $$ ) {
  my ($scramble, $noun) = @_;
  if ($scramble <= 0) {
    print "$noun will be very compressible and very dedupable\n";
  } elsif ($scramble <= 1.0) {
    print "$noun will be very compressible, and dedupable to about " .
      sprintf("%.0f", (1.0 - $scramble) * 100) . "% block sharing\n";
  } else {
    my $comp = ($scramble > 1365 ? "not" :
		($scramble > 1000 ? "a little" :
		 ($scramble > 500 ? "somewhat" :
		  ($scramble > 100 ? "pretty" : "very"))));
    print "$noun will not be dedupable, and $comp compressible\n";
  }
}

# Returns the number of nodes in a tree with the given base and height.
sub tree_nodes( $$ ) {
  my ($base, $height) = @_;

  if ($base == 0) {
    return 1;
  }

  if ($base == 1) {
    return ($base + $height);
  }

  my $cnt = (($base ** ($height + 1)) - 1) / ($base - 1);
  return $cnt;
}

=head2 gauss_stats($mean, $stddev, %options)

Returns a tuple of the average, minimum, and maximum values of a
random variable from a gaussian distribution which is no more than
a certain number of deviations from the mean.

Options:

=over 4

=item deviations

The number of standard deviations from the mean to consider (default=3).

=item fractional

Whether values are assumed to be fractional, e.g. percentages (default=0).
By default, values are assumed to be integral, and hence are rounded.

=back

=cut

sub gauss_stats( $$% ) {
  my ($mean, $dev, %opts) = @_;
  $opts{deviations} = 3 unless (defined $opts{deviations});
  if ($opts{fractional}) {
    my $min = $mean - $opts{deviations} * $dev;
    $min = 0 if ($min <= 0);
    my $max = $mean + $opts{deviations} * $dev;
    return ($mean, $min, $max, $mean);
  } else {
    my $mode = sprintf("%.0f", $mean);
    my $avg = ($dev == 0 ? $mode : $mean);
    my $min = sprintf("%.0f", $avg - $opts{deviations} * $dev);
    $min = 0 if ($min <= 0);
    my $max = sprintf("%.0f", $avg + $opts{deviations} * $dev);
    return ($avg, $min, $max, $mode);
  }
}


# Given the tree depth, subdirectories per directory,
# files per directory, and bytes per file,
# returns a hash with the total number of directories,
# files, file bytes, and directory bytes in the tree.
sub tree_math ( $$$$ ) {
  my ($depth, $dirs, $files, $bytes) = @_;
  my $total = {};
  $total->{dirs} = tree_nodes($dirs, $depth);
  $total->{files} = $total->{dirs} * $files;
  $total->{file_bytes} = $total->{files} * $bytes;
  $total->{dir_bytes} = $total->{dirs} * ($dirs + $files + 2) * 32;
  return $total;
}

=head2 populate_estimates(...)

Given the four main tree parameters (tree depth, subdirectories per directory,
files per directory, and bytes per file) and their corresponding standard
deviations, calculates and returns a list of four hash references
(average, minimum, maximum, and mode).
Each hash describes the parameters at that extreme/level, as well
as the the tree-wide totals for directories, files, file bytes, and
directory bytes.

=cut

sub populate_estimates ( $$$$$$$$;$ ) {
  my ($depth_avg, $depth_dev, $dirs_avg, $dirs_dev,
      $files_avg, $files_dev, $bytes_avg, $bytes_dev,
      $deviations) = @_;
  $deviations = 3 unless (defined $deviations);

  my @depths = gauss_stats($depth_avg, $depth_dev, deviations => $deviations);
  my @dirs = gauss_stats($dirs_avg, $dirs_dev, deviations => $deviations);
  my @files = gauss_stats($files_avg, $files_dev, deviations => $deviations);
  my @bytes = gauss_stats($bytes_avg, $bytes_dev, deviations => $deviations);

  my @hashrefs = ();
  for (my $i = 0; $i < 4; $i++) {
    my $tree = { depth => $depths[$i],
		 dirs => $dirs[$i],
		 files => $files[$i],
		 bytes => $bytes[$i] };
    $tree->{total} = tree_math($tree->{depth}, $tree->{dirs},
			       $tree->{files}, $tree->{bytes});
    push @hashrefs, $tree;
  }
  return @hashrefs;
}


=head1 BUGS

We should probably have routine here to do the general forking/waiting
code for returned subpaths with FORK_COUNT in the command-line scripts.

C<pod_usage> should probably use C<Pod::Man> instead of calling the
shell program C<pod2man>.  (Though it already has to call C<nroff>,
so it is not a big deal.)

=cut


1;
