# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/DataManip/function.pm#1 $

## @summary Function parsing and generation
## @author remi@netapp.com
## @status public
## @pod here

=head1 NAME

function -- Function parsing and generation

=head1 DESCRIPTION

This package defines routines for parsing user-supplied
functions (in the form of strings) and automatically
converting them to executable perl subroutines.

Unlike ordinary perl syntax, consecutive variables or
numbers imply multiplication.

=cut

package DataManip::function;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = ();
our @EXPORT_OK = qw(func_sub func_text tester);

use warnings;
use strict;
use Carp;

my $bin = qr{(?:\*\*|\+|\-|\*|\/|\%)};       # binary operators
my $pfx = qr{(?:\-|sqrt|log)};               # prefix operators
my $id  = qr{[A-Za-z_]\w*};                  # identifier
my $nc  = qr{\d+(?:\.\d+)?};                 # numerical constant
our $np;
   $np  = qr{\((?:(?>[^()]+)|(??{$np}))*\)}; # nested parentheses

=head1 FUNCTIONS

=head2 func_sub($text, @variables)

This takes a text representation of a function and returns an
anonymous subroutine which executes that function.

The list of variables specifies the name and order of the variables
which the function operates on.  All variable names must consist
of alphanumerics and must begin with a letter.

Croaks if the text function syntax is incorrect, contains
unknown variables, or on any other error.

I<Example>:

  my $func = func_sub("((x + 1) y ** 2) / 3.125", qw(x y));

is equivalent to:

  my $func = sub {
    my ($x, $y) = @_;
    return ((($x + 1) * $y ** 2) / 3.125);
  };

=cut

sub func_sub( $@ ) {
  my ($text, @variables) = @_;

  # Check variable names.
  foreach my $v (@variables) {
    croak "invalid variable \"$v\"\n" unless ($v =~ /\A[A-Za-z]\w*\z/);
    croak "variable \"$v\" matches operator" if ($v =~ /\A$pfx\z/o);
  }

  # Check for duplicate variables.
  my %vars = map { $_ => 1 } @variables;
  if (scalar(keys %vars) != scalar(@variables)) {
    croak "duplicate variable names\n";
  }

  # Generate the text expression for the function.
  my ($func_text, @func_vars) = func_text($text);

  # Make sure the returned variables match the ones the user gave.
  foreach my $v (@func_vars) {
    if (exists $vars{$v}) {
      delete $vars{$v};
    } else {
      croak "unrecognized identifier \"$v\"\n";
    }
  }
  my @extra_vars = keys %vars;
  if (scalar(@extra_vars)) {
    croak "found unused variables (" . join(", ", @extra_vars) . ")\n";
  }

  # Build the subroutine expression text.
  my $var_text = join(", ", map { "\$" . $_ } @variables);
  my $code = <<"CODE";
    sub {
      my ($var_text) = \@_;
      return ($func_text);
    }
CODE

  # Evaluate it to get the anonymous subroutine.
  my $func = eval $code;
  croak $@ if ($@);
  return $func;
}


=head2 func_text($text)

This is the recursive meat of func_sub().  It takes the text
representation of a function and returns the text of a perl
expression which can be evaluated to execute that function.

Actually, it returns the text expression followed by a list of the
variables used in the expression.  All those variables need to be
defined when the expression is evaluated.

=cut

sub func_text( $ );

sub func_text( $ ) {
  my ($text) = @_;

  my %variables = ();		# use a hash to eliminate duplicates
  my @tokens = ();

  while ($text =~ s/\A\s*(?:($bin)|($pfx)|($nc)|($id)|($np))\s*//o) {
    my $sym = $1 || $2 || $3;

    if ($sym) {			# operator or number
      push @tokens, $sym;

    } elsif ($4) {		# identifier
      $variables{$4} = 1;
      push @tokens, "\$$4";

    } elsif ($5) {		# parentheses
      my $pexp = $5;
      $pexp =~ s/\A\((.*)\)\z/$1/s or croak "paren strip";
      my ($par_text, @par_vars) = func_text($pexp);
      push @tokens, "($par_text)";
      foreach my $v (@par_vars) {
	$variables{$v} = 1;
      }
    }

  }
  if ($text ne "") {
    croak "invalid syntax \"$text\"\n";
  }
  croak "no tokens" unless (scalar(@tokens));

  # Run forward through the tokens, check syntax of token sequence,
  # record where we need to insert implicit operations, and record
  # where we need to combine prefix operators with their arguments.
  my @inserts = ();
  my @prefix = ();
  # Record whether or not the previous and current tokens are/were operators.
  my ($pre, $cur);
  for (my $i = 0; $i <= (scalar(@tokens) - 1); $i++) {
    if (($tokens[$i] =~ /\A$pfx\z/o) &&
	# '-' is only a prefix operator if it cannot be subtraction,
	# hence if the previous token is nonexistent or an operator
	(($tokens[$i] ne '-') || ((! defined $pre) || $pre))
       ) {
      # Current token is a prefix operator.
      # Grab next token and squash it into this one, to make a non-operator.
      if ($i > $#tokens) {
	croak "no argument to prefix operator \'" . $tokens[$i] . "\'\n";
      }
      $prefix[$i+1] = 1;

    } elsif ($tokens[$i] =~ /\A$bin\z/o) {
      # Current token is a binary operator.
      $cur = 1;
      if (! defined $pre) {
	croak "no first argument to binary operator \'" .
	  $tokens[$i] . "\'\n";
      }
      if ($pre) {
	croak "invalid argument \'" . $tokens[$i-1] .
	  "\' to binary operator \'" . $tokens[$i] . "\'\n";
      }
      if ($i > $#tokens) {
	croak "no second argument to binary operator \'" .
	  $tokens[$i] . "\'\n";
      }

    } else {
      # Current token is a numerical constant, variable,
      # or parenthetical expression.
      $cur = 0;
      if ((defined $pre) && !$pre) {
	# Found consecutive non-operators, i.e. implicit multiplication.
	$inserts[$i] = '*';
      }
    }

    $pre = $cur;
    undef $cur;
  }

  # Run backward through the tokens, insert implicit operations,
  # and combine prefix operators with their arguments.
  # (Note, we could not do it earlier because prefix operators have
  # right-to-left association... Hence "sqrt - log 1" means
  # "sqrt(-(log(1)))", not "(((sqrt)-)log)1".)
  for (my $i = $#tokens; $i >= 0; $i--) {
    if ($prefix[$i]) {
      # Should never prefix first token.
      croak if ($i <= 1);
      # Should never insert between prefix operator and its argument.
      croak if ($inserts[$i]);
      my $token = $tokens[$i - 1] . "(" . $tokens[$i] . ")";
      splice(@tokens, $i - 1, 2, $token);
    } elsif ($inserts[$i]) {
      splice(@tokens, $i, 0, $inserts[$i]);
    }
  }

  my $func = join(" ", @tokens);
  my @vars = keys %variables;

  return ($func, @vars);
}


=head2 tester()

This routine just runs a whole bunch of text strings through
func_text() and func_sub() to test them out.  It should be used
by people making changes to this library before they check them in.

=cut

{
  my @text_tests =
    (
     # $function_text, $expected_result_text, @expected_result_vars
     [ 'x', '$x', [ qw(x) ] ],
     [ 'x y', '$x * $y', [ qw(x y) ] ],
     [ '1 / d + -4', '1 / $d + -(4)', [ qw(d) ] ],
     [ '((2x) / (x**2 - 1)) + 4', '((2 * $x) / ($x**2 - 1)) + 4', [ qw(x) ] ],
    );

  my @func_tests =
    (
     # $function_text, \@vars, map $expected_output => \@inputs
     [ "x", [ qw(x) ],
       {
	1 => [ 1 ],
	-2 => [ -2 ],
       }
     ],
     [ "(x + 5) ** y / 3", [ qw(x y) ],
       {
	12 => [ 1, 2 ],
	9 => [ -2, 3 ],
	-9 => [ -8, 3 ],
       }
     ],
     [ "((2x) / (x**2 - 1)) + 4", [ qw(x) ],
       {
	4.75 => [ 3 ],
	3.25 => [ -3 ],
	4 => [ 0 ],
       }
     ],
    );

  # Compares the text of two functions, ignoring whitespace.
  sub cmp_func_text( $$ ) {
    my ($t1, $t2) = @_;
    $t1 =~ s/\s//g;
    $t2 =~ s/\s//g;
    return ($t1 eq $t2);
  }

  # Compares two lists of variable names, pairwise.
  sub cmp_func_vars( \@\@ ) {
    my ($l1, $l2) = @_;
    my $len = scalar(@$l1);
    return undef unless (scalar(@$l2) == $len);
    for (my $i = 0; $i < $len; $i++) {
      return undef unless ($l1->[$i] eq $l2->[$i]);
    }
    return 1;
  }

  sub tester() {
    # Produce stack traces for all croaks.
    require Carp;
    import Carp qw(verbose);

    foreach my $test (@text_tests) {
      my ($text, $real_result, $real_vars) = @$test;
      my ($result, @vars) = func_text($text);

      unless (cmp_func_text($result, $real_result)) {
	die "\a$text\n\t=> $result\n\t!= $real_result\n";
      }

      unless (cmp_func_vars(@vars, @$real_vars)) {
	die "\a$text\n\t   @ " . join(", ", @vars) . "\n" .
	  "\t!= @ " . join(", ", @$real_vars) . "\n";
      }
    }

    foreach my $test (@func_tests) {
      my ($text, $vars, $map) = @$test;
      my $func = func_sub($text, @$vars);
      while (my ($expected, $args) = each %$map) {
	my $result = $func->(@$args);
	unless ($result == $expected) {
	  die "\a$text\n\t$result != $expected\n";
	}
      }
    }
  }
}

=head1 BUGS

Still does not support:

=over 4

=item *

abs, round, int, exp

=item *

sin, cos, Math::Trig::tan

=item *

POSIX::floor, POSIX::ceil

=item *

S<COND ? EXP1 : EXP2>

=back

Needs more extensive testing.

=cut


1;
