# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/DataManip/local.pm#1 $

## @summary data manipulation routines
## @author remi@netapp.com
## @status public
## @pod here

=head1 NAME

DataManip::local - local client data manipulation routines

=head1 DESCRIPTION

This package defines routines which manipulate datasets, in random yet
predictable and repeatable ways.  The core routines, randomly create
or modify a data tree, in various ways specified by a staggering
number of configurable options.  However, it is all pseudo-random, so
running something from the same client on the same tree with the same
options should yield the same results.

These routines only manipulate data trees directly accessible by the
local client, and do so with whatever permissions the invoking process
has.  Typically they would be invoked on remote clients by the
routines in C<DataManip> NATE library.  However, nothing in this
library depends upon NATE, and as such can be invoked directly by
command-line wrappers in F</usr/software/test/bin/>, namely:

=over 8

=item

F</usr/software/test/bin/populate>

=item

F</usr/software/test/bin/modify>

=item

F</usr/software/test/bin/compare>

=item

F</usr/software/test/bin/delete>

=back

Note that these routines were designed to be generic, cross-platform,
and easy to maintain and extend.  They are not intended to be
high-performance.  (If they had been, we would have written them in
C<C>.)  If your datasets are huge and hence things are running too
slow, then throw more hardware at the problem.  See the FORK_COUNT
option below.

=head1 OPTIONS

The options to these data manipulation routines are numerous and
growing.  As such, the ENVIRONMENT_TYPE option was introduced to
provide an easy way to default all options to some reasonable values.

An "environment type" is basically just a set of default values for
pretty much B<all> options, which is defined in C<DataManip::defaults>.
Specifying the ENVIRONMENT_TYPE automatically sets every option to
the default value for that group, after which point any option can
be further overriden.  The idea is to define sets of default option
values for various common environments.  For example, QA groups might
want datasets with a large amount of variability in tree depth, file
sizes, and file types, in order to test boundary conditions, while
a performance group might be more interested in more regular datasets
containing only normal files that are modified in a tightly-controlled
(yet somewhat random) manner.  In each of these situations, we would
define an environment type (i.e. a set of default option values).

There are a some options, such as those seen as specific to each run,
which are not given default values by an ENVIRONMENT_TYPE.  The other
options that are defaulted by the ENVIRONMENT_TYPE typically come in
pairs or groups (or groups of pairs).  Once common pair is a mean and
a standard deviation from that mean, used to describe a Gaussian
distribution from which random numbers for some parameter are chosen.
Another pair is a "raw" option which controls how a related option
value (typically a mean) is interpreted.

Options specific to each major routine are described with the
routine.  The following are general options that all the major
routines accept:

=over 4

=item SEED

Specifies the seed with which to initialize the random number
generator.  If none is provided, a random seed is chosen by meshing
together the current time and pid.

This option is not controlled by the ENVIRONMENT_TYPE.
(If it was, then there would not be much randomness, would there?)

=item ENVIRONMENT_TYPE

Specifies a group of default option values, as defined in
C<DataManip::defaults>.  Other specified option values override
the default value from the specified ENVIRONMENT_TYPE.

If not specified, and other options are specified, then all
unspecified option values default to zero.  If no options are
specified at all, then the ENVIRONMENT_TYPE defaults to C<filer>.

=item VERBOSE

Specifies a verbosity level (default=0) at which to print debugging
messages (which are useful for tracing execution).

The higher the verbosity level, the more detailed and frequent
the debug messages.  In general:

=over

=item 0

No debug messages.

=item 1

Prints out debug messages that only happen once, i.e. general
information that applies to the entire run.  In particular,
prints out all the options and option values.

=item 2

Prints out debug messages that happen at most once per directory.
This effectively provides a trace of how the directory tree
was traversed.

=item 3

Prints out debug messages that happen multiple times per directory,
e.g. once per file.

=back

=back

=head1 FUNCTIONS

=cut

package DataManip::local;
use warnings;
use strict;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(populate_local_tree
		 modify_local_tree
		 compare_local_trees
		 delete_local_tree);
our @EXPORT_OK = qw(mkdir_path writeFile legacy_options_compatibility
		    set_option_defaults);

use Carp;
use Cwd qw(cwd);
use Errno qw(ENOTEMPTY ENOENT);
use Fcntl qw(:DEFAULT :mode :seek); # O_* S_IF* SEEK_*
use File::Compare;
use File::Temp qw(mktemp);
use File::Spec qw();
use File::stat;
use DataManip::defaults;
use DataManip::manip_log qw(log_manipulation);

use constant UINT4_MAX => 4294967295;
use constant BLOCK_SIZE => 4096;
use constant HALF_BLOCK => 2048;
use constant CURDIR => File::Spec->curdir();
use constant UPDIR => File::Spec->updir();
use constant WINDOWS => "MSWin32";

# These will go away or somehow change when we start using random names.
use constant DIR_PREFIX => "dir";
use constant FILE_PREFIX => "file";

# Prototype these early...
sub mkdir_path( $;$ );
sub writeFile( $$;% );

# Return current time meshed with pid.
sub getSeed () {
  my $seed = 0;
  until ($seed != 0) {
    $seed = time() ^ ($$ + ($$ << 15));
  }
  return $seed;
}

# Return the argument rounded to the nearest integer.
# Return zero if the argument is negative.
sub round {
  return 0 if ($_[0] < 0.0);
  return sprintf("%.0f", $_[0]);
}

# Some notes about normal distributions (aka. Gaussian distrubtions)...
#
# The probability density function for a normal distribution
# with a mean "m" and standard deviation "d" is:
#
#   p(x) = (e ** ((-((x - m) ** 2)) / (2 * (d ** 2)))) / (d * sqrt(2 * pi))
#
# where e = 2.71828182845905 and pi = 3.14159265358979 (approximately).
#
# So, with a mean of zero and standard devation of one, we have:
#
#   p(x) = (e ** ((-(x ** 2)) / 2)) / sqrt(2 * pi)
#
# For a distribution with a given mean and standard deviation,
# finding the probability that a random number generated from
# the distribution falls within a specific range requires
# integrating the probability density function above (for which
# there is no elementary function)...  Thankfully, this is already
# implemented in several forms.
#
# The "error function" erf() is available in Math::Libm library
# (the perl extension for the C math library).  Since that's not
# available, here, we can pull it inline straight from the C library.
#
#   use Inline C => q{ double erf(double); }, ENABLE => "AUTOWRAP";
#   foreach my $x (@ARGV) { print erf($x) . "\n"; }
#
# For a normal distribution with mean zero and standard deviation "d",
# the probability that a random number generated from that distribution
# falls within "x" of the mean (i.e. between "-x" and "x") is:
#
#   f(x) = err(x / (d * sqrt(2)))
#
# Obviously, we can use this repeatedly and take the difference to figure
# out what the probability of a number falling in any given range is.

# Return a random number from a Gaussian distribution centered
# about the given mean with the given standard deviation.
{
  my $spare = undef; # static variable across calls to randGauss()

  sub randGauss {
    my ($mean, $stddev) = @_;
    return $mean unless ($stddev);

    my $r;

    if ($spare) {
      $r = $spare;
      $spare = undef;
    } else {
      # Get two random numbers from the Gaussian distribution
      # centered at zero with a standard deviation of 1.
      my ($x1, $x2, $w, $y1, $y2);
      do {
	$x1 = 2.0 * rand() - 1.0;
	$x2 = 2.0 * rand() - 1.0;
	$w = ($x1 * $x1) + ($x2 * $x2);
      } while ( $w >= 1.0 );
      $w = sqrt( (-2.0 * log($w) ) / $w );
      $y1 = $x1 * $w;
      $y2 = $x2 * $w;

      $r = $y1;
      $spare = $y2;
    }

    # Map the random number into the desired distribution.
    $r = ($r * $stddev) + $mean;
    return $r;
  }
}

# Return the probability that a number picked from a Gaussian distribution
# with the given standard deviation falls within "x" of the mean.
sub probGauss( $$ ) {
  my ($x, $dev) = @_;
  require Inline;
  import Inline C => q{ double erf(double); }, ENABLE => "AUTOWRAP";
  return erf($x / ($dev * 1.4142135623731));
}

# Returns the probability that a number picked from a Gaussian distribution
# with the given mean and standard deviation will fall between $x and $y.
# Either $x or $y can be "undef" to unlimit that side of the range.
sub probGaussRange {
  my ($x, $y, $mean, $dev) = @_;

  $x -= $mean if (defined $x);
  $y -= $mean if (defined $y);

  # All numbers are in the unlimited range.
  if (! defined $x && ! defined $y) {
    return 1.0;
  }

  if (! defined $y) {
    # flip them around
    $y = -$x;
    undef $x;
  }
  if (! defined $x) {
    my $p = probGauss(abs($y), $dev);
    if ($y < 0) {
      return ((1 - $p) / 2);
    } elsif ($y > 0) {
      return (($p / 2) + 0.5);
    } else {
      return 0.5;
    }
  }
  die unless (defined $x && defined $y);

  if ($x == $y) {
    return 0.0;
  }

  if ($y < $x) {
    # Swap them so that $x is smaller.
    my $tmp = $x;
    $x = $y;
    $y = $tmp;
  }
  if ($y <= 0) {
    # flip them around
    my $tmp = $x;
    $x = -$y;
    $y = -$tmp;
  }

  my $py = probGauss($y, $dev);
  if ($x == 0) { # $x is the mean
    return ($py / 2);
  }
  my $px = probGauss(abs($x), $dev);

  if ($x > 0) { # both above the mean
    return (($py - $px) / 2);
  } else { # on either sides of the mean
    return (($px + $py) / 2);
  }
}

# These are the options which do not directly affect how
# data gets modified, but just perform other cosmetic functions.
my @cosmetic_options_1 = qw(SEED ENVIRONMENT_TYPE VERBOSE);

# These are also cosmetic options, as above, except that these
# don't even appear in the environment types, and hence should
# never be defaulted to any value.
my @cosmetic_options_2 = qw(HOLD_FS_SIZE LOG_FILE SHOW_OPTS);

my $cosmetic_opts = join('|', @cosmetic_options_1, @cosmetic_options_2);
my $cosmetic_opts2 = join('|', @cosmetic_options_2);

# Given an options hash and a list of legal option values, fills
# in defaults for unspecified options from the environment type.
sub set_option_defaults( \%@ ) {
  my $options = shift;

  unless (defined $options->{SEED}) {
    $options->{SEED} = getSeed();
  }

  if (defined $options->{ENVIRONMENT_TYPE}) {
    unless (exists $defaults{ $options->{ENVIRONMENT_TYPE} }) {
      croak "Environment type \"" . $options->{ENVIRONMENT_TYPE} .
	"\" does not exist.";
    }
  } else {
    # Figure out whether the user specified any
    # data-modifying options.
    my @mod_opts = grep(!/\A$cosmetic_opts\z/o, @_);
    my $data_mod = grep { defined $options->{$_} } @mod_opts;

    if ($data_mod) {
      # The user specified data-modifying options, so
      # default all other options to zero.
      $options->{ENVIRONMENT_TYPE} = 'nil';
    } else {
      # No data-modifying options specified, so use the
      # "default environment type" for all options.
      $options->{ENVIRONMENT_TYPE} = 'filer';
    }
  }

  # Set all undefined options to the corresponding default
  # values in the environment type.
  my $env_type = $defaults{ $options->{ENVIRONMENT_TYPE} };
  foreach my $opt (@_) {
    next if ($opt =~ /\A$cosmetic_opts2\z/o);
    unless (defined $options->{$opt}) {
      $options->{$opt} = $env_type->{$opt};
    }
  }
}

# Default all unspecified options to zero.
sub define_default_options( \%@ ) {
  my $options = shift;
  foreach my $opt (@_) {
    unless (defined $options->{$opt}) {
      $options->{$opt} = 0;
    }
  }
}

# Given a reference to an options hash, and a list of valid option names,
# simply prints out all the valid option names and values to STDOUT.
sub print_options( \%@ ) {
  my $options = shift;
  foreach my $opt (@_) {
    if (defined $options->{$opt}) {
      print "$opt => '" . $options->{$opt} . "'\n";
    } else {
      print "$opt => undef\n";
    }
  }
}

# Given a hash of argument names to values, simply prints them out.
sub print_args( % ) {
  my %args = @_;
  print_options(%args, keys %args);
}

# Someday, when we decide to deprecate option lists, we can just
# delete this block and all calls to this function.
# Returns whether or not we tried using legacy options.
{
  my $pat_ident = "[A-Za-z_][A-Za-z0-9_]*";

  sub legacy_options_compatibility( \%@ ) {
    my $options = shift;

    if (scalar(@_) == 0) {
      # No options, so no problem.
      return 0;
    }

    # Check whether any options are not "KEY=VALUE" strings.
    my $non_list_options = grep(!/\A($pat_ident)=(.*)$/o, @_);

    if ($non_list_options) {
      # At least one options was not a "KEY=VALUE" string,
      # so do nothing.
      return 0;
    }

    carp "Using list-style options.  This will soon be deprecated.\n";

    # Clear all existing options.
    %$options = ();

    # Parse and set all options.
    foreach (@_) {
      unless ($_ =~ /\A($pat_ident)=(.*)\z/o) {
	croak "internal parse error";
      }
      if (defined $2) {
	$options->{$1} = $2;
      } else {
	$options->{$1} = '';
      }
    }

    return 1;
  }
}

# Given a path to a (non-existent) log file, creates the directory
# for the log file (if necessary) and exclusively creates and opens
# the log file for writing.  Returns the filehandle on success,
# croaks on any error.
sub open_log_file( $ ) {
  my ($log_file_path) = @_;
  my $fh; # empty filehandle reference

  # Undocumented debugging option...
  # Specify '-' to log to STDOUT instead of a file.
  if ($log_file_path eq '-') {
    open($fh, ">-") or croak "error opening STDOUT: $!\n";
    return $fh;
  }

  # Figure out the log file's parent directory, and create it if necessary.
  my ($vol, $dirs, $log_file) = File::Spec->splitpath($log_file_path);
  my $log_dir = File::Spec->catpath($vol, $dirs, "");
  unless (mkdir_path($log_dir)) {
    croak "error creating log directory: $!\n";
  }

  # Create and open the log file, failing if we can't exclusively create it.
  unless (sysopen($fh, $log_file_path, (O_WRONLY|O_CREAT|O_EXCL))) {
    croak "error creating log file: $!\n";
  }
  return $fh;
}

# writeBytes($filehandle, $byte_count, $data_buffer, %options)
#
# Given an open filehandle, a number of bytes to write, and a reference to
# a buffer of bytes, writes the specified number of bytes from the buffer
# to the filehandle (wherever it is currently positioned), and then rotates
# the buffer of bytes by the same number.
#
# Returns success or failure (with error message in $!).
#
# Options:
#
# SCRAMBLE
#
# Percentage of successive blocks to scramble by XORing bytes in each BLOCK_SIZE
# chunk of bytes written.  The higher this is, the more probable that each
# BLOCK_SIZE sequence of bytes (probably) looks different, and therefore this
# basically controls how compressible the data is.
#
sub writeBytes( $$$% ) {
  my ($fh, $cnt, $dataref, %opts) = @_;
  unless (exists $opts{SCRAMBLE} && defined $opts{SCRAMBLE}) {
    $opts{SCRAMBLE} = 0;
  }

  # Figure out how many blocks we'll write.
  my $blocks = int ($cnt / BLOCK_SIZE);
  my $extra = int ($cnt % BLOCK_SIZE);

  # Write out whole blocks while we can.
  if ($opts{SCRAMBLE} > 1) { # scrambling more than one byte per block
    my ($stride, $max_byte_scrambles);
    if ($opts{SCRAMBLE} >= HALF_BLOCK) {
      $stride = 2;
      $max_byte_scrambles = HALF_BLOCK;
    } else {
      $stride = int(BLOCK_SIZE / $opts{SCRAMBLE});
      $max_byte_scrambles = $opts{SCRAMBLE};
    }

    my $s = 0;
    foreach (my $b = 1; $b <= $blocks; $b++) {
      syswrite($fh, pack("C" . BLOCK_SIZE, @$dataref)) or return undef;
      my $stored = $s;
      for (my $r = 0; $r < $max_byte_scrambles; $r++) {
	# Scramble the block and repack each time.
	if ($dataref->[($s - 1)] == 0) {
	  $dataref->[$s] = ~($dataref->[$s]); # flip all the bits
	} else {
	  $dataref->[$s] ^= $dataref->[($s - 1)]; # XOR with previous byte
	}
	$dataref->[$s] &= 0xff; # avoid warnings when packing chars > 255
	$s = ($s + $stride) % BLOCK_SIZE;
      }
      $s = ($stored + 1) % BLOCK_SIZE;
    }

  } elsif ($opts{SCRAMBLE}) { # maybe scrambling one byte per block
    my $s = 0;
    foreach (my $b = 1; $b <= $blocks; $b++) {
      syswrite($fh, pack("C" . BLOCK_SIZE, @$dataref)) or return undef;
      if (rand() <= $opts{SCRAMBLE}) {
	# Scramble the block and repack each time.
	if ($dataref->[($s - 1)] == 0) {
	  $dataref->[$s] = ~($dataref->[$s]); # flip all the bits
	} else {
	  $dataref->[$s] ^= $dataref->[($s - 1)]; # XOR with previous byte
	}
	$dataref->[$s] &= 0xff; # avoid warnings when packing characters > 255
	$s = ($s + 1) % BLOCK_SIZE;
      }
    }
  } else {
    # We're not scrambling between block writes,
    # so just pack the value once and write it out.
    my $bufblock = pack("C" . BLOCK_SIZE, @$dataref);
    foreach (my $b = 1; $b <= $blocks; $b++) {
      syswrite($fh, $bufblock) or return undef;
    }
  }

  # Write out the rest if necessary
  if ($extra) {
    syswrite($fh, pack("C$extra", @$dataref)) or return undef;
    push @$dataref, splice(@$dataref, 0, $extra); # rotate the buffer
  }

  return 1;
}

# writeFragment($file_name, $file_size, $byte_count, $fragment_count, %options)
#
# Writes the specified number of random bytes to the named file,
# broken up into the specified number of fragments, which are
# spread out across the file.
#
# Returns success or failure (with error message in $!).
#
# Options:
#
# FLOOD
#
# If set, random data block written to the file is just a single
# randomly-chosen byte repeated as necessary, rather than a
# randomly-generated block.  (This makes the data pretty uninteresting,
# but it is faster to generate.)
#
# SCRAMBLE
#
# Percentage of successive blocks to scramble by XORing bytes in each BLOCK_SIZE
# chunk of bytes written.  The higher this is, the more probable that each
# BLOCK_SIZE sequence of bytes (probably) looks different, and therefore this
# basically controls how compressible the data is.
#
sub writeFragment( $$$$;% ) {
  my ($file, $file_size, $byte_cnt, $frag_cnt, %opts) = @_;

  # Since we're just doing a lot of system calls, we return an error in $!,
  # not $@.  As such, we need to make sure to return immediately when something
  # goes wrong, so as not to destroy the value of $!.  This means that we can't
  # close the file when something fails.

  my $fh; # empty filehandle reference

  sysopen($fh, $file, (O_WRONLY|O_CREAT)) or return undef;
  binmode $fh;
  if ($byte_cnt <= 0 || $file_size <= 0) { # Just 'touch' the file.
    close($fh) or return undef;
    return 1;
  }

  # Make some random data
  my @data;
  my $max = ($byte_cnt < BLOCK_SIZE) ? $byte_cnt : BLOCK_SIZE;
  if ($opts{FLOOD}) {
    @data = (int rand(256)) x $max;
  } else {
    $#data = ($max - 1);
    foreach (@data) {
      $_ = int rand(256);
    }
  }

  # It's not possible to write a "partial fragment".
  if ($frag_cnt <= 1.0) {
    $frag_cnt = 1;
  } else {
    $frag_cnt = int($frag_cnt);
  }

  # Write out all the fragments.
  my $frag_avg_size = $byte_cnt / $frag_cnt;
  my $frag_range = int($file_size / $frag_cnt);
  my $frag_offset = 0;
  my $written = 0.0; # decimal counter
  my $cnt_written = 0; # always an integer
  foreach (my $frag = 1; $frag <= $frag_cnt; $frag++) {
    # Figure out how big this fragment will be.
    $written = $frag * $frag_avg_size;
    my $cnt = round($written - $cnt_written);
    $cnt_written += $cnt;

    # Pick a spot to write this fragment, and write it.
    my $offset = round(rand() * ($frag_range - $cnt));
    sysseek($fh, ($frag_offset + $offset), SEEK_SET) or return undef;
    writeBytes($fh, $cnt, \@data, %opts) or return undef;

    # Next fragment starts no earlier than $frag_offset.
    $frag_offset += $frag_range;
  }

  close($fh) or return undef;
  return 1;
}

# Randomly modify some given percentage or number of bytes of a given file.
# Returns success or failure (with error message in $!).
sub modify_file( $$;% ) {
  my ($file, $size, %opts) = @_;

  my $byte_cnt = randGauss($opts{MOD_AVG}, $opts{MOD_DEV});
  $byte_cnt *= $size unless ($opts{MOD_RAW});

  if ($byte_cnt >= $size || $byte_cnt <= 0.0) {
    # Modify the whole file || just open/close the file.
    return writeFile($file, $byte_cnt, %opts);
  }

  # Write changed fragments to the file.
  my $frag_cnt = randGauss($opts{FRAGMENT_AVG}, $opts{FRAGMENT_DEV});
  if ($opts{FRAGMENT_RAW}) {
    $frag_cnt = ($frag_cnt == 0) ? $byte_cnt : round($byte_cnt / $frag_cnt);
  }
  return writeFragment($file, $size, $byte_cnt, $frag_cnt, %opts);
}

# Grow the file with random data by the given percentage or number of bytes.
# Returns success or failure (with error message in $!).
sub grow_file( $$;% ) {
  my ($file, $size, %opts) = @_;

  my $byte_cnt = randGauss($opts{GROW_AVG}, $opts{GROW_DEV});
  $byte_cnt *= $size unless ($opts{GROW_RAW});
  return writeFile($file, $byte_cnt, OFFSET => $size, %opts);
}

# Shrink the file by the given percentage or number of bytes.
# Returns success or failure (with error message in $!).
sub shrink_file( $$;% ) {
  my ($file, $size, %opts) = @_;

  my $byte_cnt = randGauss($opts{SHRINK_AVG}, $opts{SHRINK_DEV});
  $byte_cnt *= $size unless ($opts{SHRINK_RAW});

  my $shrunken_size = ($size - $byte_cnt);
  $shrunken_size = 0 if ($shrunken_size < 0);
  return writeFile($file, 0, TRUNC => $shrunken_size, %opts);
}

# Given a path to a file, the stat object for that file, and the hash
# of standard options to modify_tree(), does the core work of modifying
# this single file in the data tree.  Uses the distribution
# parameters to randomly decide which actions to take on this file.
# Returns whether or not anything was actually modified, and throws
# an exception on any error.
sub modify_normal_file( $$;% ) {
  my ($node, $st, %opts) = @_;

  my $modified = 0;
  my $rand_num;

  # Maybe modify this file's contents
  $rand_num = rand();
  if ($rand_num < $opts{TOTAL_MOD}) {
    if (modify_file($node, $st->size, %opts)) {
      $modified = 1;
      print "Modified $node.\n" if ($opts{VERBOSE} >= 3);
    } else {
      croak "Error modifying $node: $!\n";
    }
  }

  # Maybe change this file's size
  $rand_num = rand();
  if ($rand_num < $opts{TOTAL_GROW}) {
    if (grow_file($node, $st->size, %opts)) {
      $modified = 1;
      print "Grew $node.\n" if ($opts{VERBOSE} >= 3);
    } else {
      croak "Error growing $node: $!\n";
    }
  }
  if ($rand_num >= (1.0 - $opts{TOTAL_SHRINK})) {
    if (shrink_file($node, $st->size, %opts)) {
      $modified = 1;
      print "Shrank $node.\n" if ($opts{VERBOSE} >= 3);
    } else {
      croak "Error shrinking $node: $!\n";
    }
  }

  return $modified;
}

############################################################
package dir_bitmap;  # a private sub-package!

use warnings;
use strict;

use Carp;

# This object basically builds a bitmap of directory entries.
# Trying to keep an entire directory in-core just won't scale.
# The method names and some of quirks are a bit particular to
# how DataManip processes directories...
#
# When creating new entries in a directory, we need an easy way
# to pick random names that are not currently used.  Hence,
# we keep an instance of this object for each directory we process.
# As we read the directory, we'll call entry_kept() to set a bit for
# each existing entry -- but we never try to "lookup" whether a bit
# is set for a particular entry.  Then later when we want to create
# entries in the directory, this object provides locations of unset
# bits in the bitmap.
#
# The idea is that we compute a "checksum" of sorts for each directory
# entry name, and that checksum determines which bit we set for the entry.
# We presume that given a "checksum", we can easily generate an entry
# name which has that checksum.  It's all right (though suboptimal) if
# multiple entries have the same checksum -- we just won't generate any
# new entries with that checksum until all those entries get deleted.
#
# The object also remembers specific entry names that were deleted, and
# provides a way to get some of those back.  This allows us to sometimes
# create new entries with the exact same names as previous entries --
# not just the same checksum as previous entries.

our @Fields = qw(kept_cksums deleted_ents free_spot);

use fields @Fields;

use constant BITS_PER_WORD => 32;
use constant ALL_BITS_SET => 4294967295; # ~0 would be bad on 64-bit platforms

# Given an entry name, returns the checksum for the name.
sub entry_cksum {
  my $self = shift;
  my ($entry) = @_;
  my $cksum = undef;

  if ($entry =~ /(\d+)\z/) {
    # This is somewhat of a hack.  Since all the names we create
    # end in numbers, we just use those numbers as our "checksums".
    # With real random names, we'd need to compute checksums of
    # the name data directly.
    $cksum = $1;

  } else {
    # We've got some other name.  Try to compute a checksum based
    # on the name data.  Worst case, give up and return undef.
    # (This is just a hack until we do real random names.)

    # Get 8 bits from each of the last four characters, and pack a checksum.
    my $cnt = length($entry);
    $cnt = 4 if ($cnt > 4);
    my @chars = unpack("C$cnt", substr($entry, -($cnt)));
    return undef unless ($cnt > 0 && scalar(@chars) == $cnt);
    while (scalar(@chars) < $cnt) {
      unshift @chars, 0;
    }
    $cksum = unpack("N", pack("C4", @chars));
  }

  return substr($cksum, -($self->{cksum_length}));
}

sub new {
  my ($obj, $size) = @_;

  my $class = ref $obj;
  if ($class) {
    croak "cannot clone a $class\n" unless ($class eq __PACKAGE__);
  } else {
    croak "cannot create a $obj\n" unless ($obj eq __PACKAGE__);
    $class = $obj;
    undef $obj;
  }

  my $self = {};

  if ($obj) { # clone an existing object

    if (defined $size) {
      croak "cannot pass size to constructor when cloning\n";
    }

    while (my ($k, $v) = each (%$obj)) {
      my $type = ref($v);
      if (!$type) { # not a reference (and possibly undefined)
	$self->{$k} = $v;
      } elsif ($type eq "ARRAY") {
	my @data = @$v;
	$self->{$k} = \@data;
      } elsif ($type eq "HASH") {
	my %data = %$v;
	$self->{$k} = \%data;
      } elsif ($type eq "SCALAR") {
	my $data = $$v;
	$self->{$k} = \$data;
      } else {
	die "unsupported field type $type\n";
      }
    }

  } else { # create a new object

    # The size should be the maximum number of entries/bits
    # which the bitmap is expected to contain.
    if (defined $size) {
      croak "invalid size $size\n" unless ($size =~ /\A\d+\z/);
    } else {
      croak "no size specified\n";
    }

    # We want some extra space in the bitmap, both to make searching easier
    # and also to deal with some eccentricies caused by using numerical
    # suffixes as checksums.
    # Hence, we take the maximum entry count and round it up to the biggest
    # number which uses the same number of digits.  (e.g. 35 rounds to 99,
    # 647 rounds to 999, 1234 rounds to 9999, etc.)

    my $max_cksum_length = length(sprintf("%d", $size));
    my $max_size = "9" x $max_cksum_length;
    my $max_offset = int($max_size / BITS_PER_WORD);
    my @bitmap_data = ();

    # If the number is too big, accessing an offset off the
    # end of the array could cause us to die "Out of memory!",
    # or "Modification of non-creatable array value attempted, subscript -1".
    eval {
      $#bitmap_data = $max_offset;
    };
    if ($@) {
      croak sprintf("out of memory when creating %s with %d offsets: $@\n",
		    __PACKAGE__, $max_offset + 1);
    }

    # The bitmap holds a multiple of BITS_PER_WORD, but we only want
    # checksums up to (and including) $max_size, so preemptively set
    # all the extra bits in the last word of the bitmap.
    my $extra_bit_cnt = (($max_offset + 1) * BITS_PER_WORD) % ($max_size + 1);
    if ($extra_bit_cnt > 0) {
      my $last_word = ~((1 << (BITS_PER_WORD - $extra_bit_cnt)) - 1);
      $last_word &= ALL_BITS_SET; # don't set extra bits on 64-bit platforms
      eval {
	$bitmap_data[$max_offset] = $last_word;
      };
      if ($@) {
	croak sprintf("out of memory when setting extras at offset %d: $@\n",
		      __PACKAGE__, $max_offset);
      }
    }

    $self->{kept_cksums} = \@bitmap_data;
    $self->{max_offset} = $max_offset;
    $self->{bitmap_size} = $max_size + 1;
    $self->{cksum_length} = $max_cksum_length;

    # At the very least, we need to remember the names of entries we deleted,
    # in order to possibly create new entries with the same names.
    $self->{deleted_ents} = [];

    # An internal state kept to speed up repeated calls to get_free_cksum().
    $self->{free_spot} = undef;

  }

  return bless($self, $class);
}

# Declare accessors for all public fields.
foreach my $field (@Fields) {
  my $accessor = sub {
    my ($self, $arg) = @_;
    my $type = ref($self);
    unless ($type && $type eq __PACKAGE__) {
      croak sprintf("%s is not a %s\n", ($type ? $type : $self), __PACKAGE__);
    }
    $self->{$field} = $arg if (defined $arg);
    return $self->{$field};
  };
  no strict 'refs';
  *$field = $accessor;
  use strict 'refs';
}

sub sprint_keepers {
  my $self = shift;
  my $kept_cksums = $self->kept_cksums;
  my $str = "kept_cksums = [\n";
  for (my $i = 0; $i < scalar(@$kept_cksums); $i++) {
    my $bitword = (exists $kept_cksums->[$i] ? $kept_cksums->[$i] : 0);
    my $bitnum = $i * BITS_PER_WORD;
    for (my $j = 0; $j < BITS_PER_WORD; $j++) {
      my $bit = (($bitword & (1 << $j)) ? 1 : 0);
      $str .= sprintf("%d-%d ", $bitnum + $j, $bit);
    }
  }
  $str .= "]\n";
  return $str;
}

# Returns whether the bit for the given checksum is set.
sub cksum_used {
  my $self = shift;
  my ($cksum) = @_;
  my $offset = int($cksum / BITS_PER_WORD);
  my $bitnum = $cksum % BITS_PER_WORD;

  my $kept_cksums = $self->kept_cksums;
  if (exists $kept_cksums->[$offset]) {
    my $bitword = $kept_cksums->[$offset];
    my $bit = (($bitword & (1 << $bitnum)) ? 1 : 0);
    return $bit;
  } else {
    return 0;
  }
}

# Returns whether the bit for the checksum of given directory entry is set.
sub entry_used {
  my $self = shift;
  my $cksum = $self->entry_cksum($_[0]);
  if (defined $cksum) {
    return $self->cksum_used($cksum);
  } else {
    # Be conservative -- if we can't compute the checksum, pretend it's used.
    return 1;
  }
}

# Called with the name of a directory entry to "set as used".
sub entry_kept {
  my $self = shift;

  # Bail unless we can compute a checksum for this entry name.
  # If we can't, then we won't try to create an entry over this one,
  # since at the moment we only create entries that end with numbers.
  # Once we start doing random names, this might not be true anymore,
  # and so we'll need to remember enough to avoid clobbering an entry
  # with *any* name.
  my $cksum = $self->entry_cksum($_[0]);
  return unless (defined $cksum);

  return $self->cksum_kept($cksum);
}

sub cksum_kept {
  my $self = shift;
  my ($cksum) = @_;
  my $offset = int($cksum / BITS_PER_WORD);
  my $bitnum = $cksum % BITS_PER_WORD;

  if ($offset > $self->{max_offset}) {
    croak "cannot set cksum \"$cksum\" at offset $offset\n";
  }

  my $kept_cksums = $self->kept_cksums;
  my $bitword = (exists $kept_cksums->[$offset] ? $kept_cksums->[$offset] : 0);
  $bitword |= (1 << $bitnum);

  # This should never fail, because we preallocated the array.
  eval {
    $kept_cksums->[$offset] = $bitword;
  };
  if ($@) {
    croak sprintf("out of memory when setting %s entry at offset %d: $@\n",
		  __PACKAGE__, $offset);
  }
}

# Returns a checksum which no existing entries have.
sub get_free_cksum {
  my $self = shift;
  my $kept_cksums = $self->kept_cksums;

  my $cksum = $self->free_spot;
  unless (defined $cksum) {
    # Pick a random spot.
    my $cksum_cnt = $self->{bitmap_size};
    $cksum = (($cksum_cnt > 0) ? int(rand $cksum_cnt) : 0);
  }

  # Keep looping until we find a free entry.  In the worst case, we'll
  # loop all the way around and search the entire array before giving up.
  my $is_retry = 0;
 RETRY_ONCE:

  # We can do better than:
  #	while ($self->cksum_used($cksum)) { $cksum++; }
  # by peering into the actual bits and checking a word at a time.

  my $offset = int($cksum / BITS_PER_WORD);
  my $bitnum = $cksum % BITS_PER_WORD;

  my $bitword = undef;
  if (exists $kept_cksums->[$offset]) {
    # Look for a free bit in the rest of this word.
    $bitword = $kept_cksums->[$offset];
    if ($bitnum > 0) {
      # Set all the bits lower that $bitnum.
      my $x = (1 << ($bitnum - 1));
      my $mask = $x ^ (~$x + 1); # perl doesn't do "-$x" correctly
      $bitword |= ~$mask;
    }

    if ($bitword == ALL_BITS_SET) {
      # No free bits in the first word.
      # Find the first word with unset bits.
      $bitnum = 0;
      while (1) {
	undef $bitword;
	$offset++;
	last unless (exists $kept_cksums->[$offset]);
	$bitword = $kept_cksums->[$offset];
	last unless ($bitword == ALL_BITS_SET);
      }
      # At this point, $bitword should be undefined if we ran off the
      # end of the array, else it should contain the word at $offset.
    }
  }

  if (defined $bitword) {
    # Set $bitnum to the position of the first unset bit in $bitword.
    if ($bitword % 2 == 0) {
      $bitnum = 0;
    } else {
      my $x = ~($bitword >> 1);
      my $m = $x ^ (~$x + 1); # perl doesn't do "-$x" correctly
      my $bit = ($m ^ ($m << 1));
      $bitnum = log($bit) / log(2);
    }
  } else { # We ran off the end of the array.
    if ($offset > $self->{max_offset}) {
      # ... and off the end of the allocated array!
      if ($is_retry) {
	# The array completely filled up, which should
	# never happen if we preallocated it correctly.
	croak "no free entries (increase MIN_DIR_SIZE?)\n";
      } else {
	# Wrap around to the beginning and try one more time.
	$is_retry = 1;
	$cksum = 0;
	goto RETRY_ONCE;
      }
    }
  }

  $cksum = ($offset * BITS_PER_WORD) + $bitnum;
  $cksum = sprintf("%.0f", $cksum); # round off, since division uses floats

  $self->cksum_kept($cksum); # Mark this spot taken,
  $self->free_spot($cksum);  # and remember where our search left off.

  return $cksum;
}

sub sprint_deletes {
  my $self = shift;
  return "deleted_ents = (" . join(", ", @{$self->deleted_ents}) . ")\n";
}

# Remeber a deleted entry.
sub entry_deleted {
  my $self = shift;
  push @{$self->deleted_ents}, $_[0];
  # XXX The constructor should take a maximum count of "deleted entries"
  # to remember, to bound the amount of memory we burn.
}

# Given a requested number, returns at most that number of "deleted names".
# Since they are picked randomly and checked for duplicates, it could
# return fewer than the requested number.  However, entry_kept() is
# automatically called for all returned names.
sub deleted_name_subset {
  my $self = shift;
  my ($cnt) = @_;
  $cnt = "undef" unless (defined $cnt);
  croak "non-numeric count $cnt\n" unless ($cnt =~ /\A\d+\z/);

  # A somewhat quirky side-effect of this routine is that it
  # resets the internal state for search with get_free_cksum().
  $self->free_spot(undef);

  my $deleted_ents = $self->deleted_ents;
  my $idx = int(rand() * scalar(@$deleted_ents));
  my @ents = splice(@$deleted_ents, $idx, $cnt);

  # Entries could have been created since the delete was recorded.
  # Do not return any names whose checksums we know were later created.
  my @names = ();
  foreach (@ents) {
    next if ($self->entry_used($_));
    $self->entry_kept($_);
    push @names, $_;
  }

  return @names;
}

############################################################
package DataManip::local; # switch back into the public package

# This routine is the per-directory processing for modify_local_tree().
# It walks the current directory (blatantly assuming $path is the CWD),
# processes all the files, and returns the subpaths to process.
sub modify_directory( $$$% ) {
  my ($path, $depth, $process_error, %opts) = @_;

  print "Depth $depth, Process $path\n" if ($opts{VERBOSE} >= 2);

  my $dirfh; # empty filehandle reference
  unless (opendir($dirfh, CURDIR)) {
    $process_error->("opendir($path): $!\n");
    return ();
  }
  my $dirst = stat(CURDIR);
  unless ($dirst) {
    $process_error->("stat($path): $!\n");
    return ();
  }

  # Figure out how many files and directories we'll create at the end.
  my $file_cnt = round randGauss($opts{CREATE_FILE_AVG},
				 $opts{CREATE_FILE_DEV});
  my $dir_cnt = round randGauss($opts{CREATE_DIR_AVG},
				$opts{CREATE_DIR_DEV} );

  # Estimate the maximum number of entries in this directory,
  # counting the ones we'll create.  This doesn't have to be
  # accurate, but it does have to be an upper-bound.
  #
  # Since WAFL always burns at least 32 bytes per entry, we'll
  # base our guess on that.  This will cause problems if we run
  # on ext3 or ufs, since those can make entries as small as
  # 11-12 bytes each.
  my $dir_bytes = $dirst->size ||
                  ($opts{MIN_DIR_SIZE} * 1024) ||
                  (20940 * 1024);
  my $max_entry_cnt = ($dir_bytes / 32) + 1 + $file_cnt + $dir_cnt;

  # Loop through all the entries in the directory, and remember anything
  # we'll need to know later on when creating names for new entries.
  my $dbm = new dir_bitmap($max_entry_cnt);

  my @subpaths = ();  # our return values go here

  while (my $entry = readdir($dirfh)) {
    next unless (scalar File::Spec->no_upwards($entry));
    my $entry_path = File::Spec->catdir($path, $entry);

    # Stat the node.
    my $st = stat($entry);
    unless ($st) {
      $process_error->("stat($entry_path): $!\n");
      next;
    }

    if (($st->mode & S_IFDIR) == S_IFDIR) {
      # Descend directories to expand the fringe.

      # Special-case the .snapshot directory.
      if ($entry eq ".snapshot") {
	my $fh = undef;
	my $fname = mktemp(File::Spec->catfile($entry, "tmp.XXXXXX"));
	if (open($fh, ">$fname")) {
	  close($fh) or $process_error->("close($fname): $!\n");
	  unlink($fname) or $process_error->("unlink($fname): $!\n");
	} elsif ($!{EROFS}) {
	  # It's read-only, so skip it.
	  next;
	} elsif ($!{EACCES} && -w $entry) {
	  # Some clients return EACCES instead of EROFS.  As long as the entry
	  # is writable (according to the permissions), we will skip it if we
	  # are on one of those "special" clients...

	  # Linux NFS client, running 2.6 kernel or later
	  if ($^O eq "linux") {
	    my $kernel_version = `uname -r`;
	    die "uname: $kernel_version (exited with $?)\n"
	      unless (!$? && $kernel_version =~ /\A(\d+)\.(\d+)/);
	    my ($major, $minor) = ($1, $2);
	    next if ($major > 2 || ($major == 2 && $minor >= 6));
	  }

	  # Windows CIFS
	  next if ($^O eq WINDOWS);
	  # Though, CIFS never even finds the snapshot directory
	  # to begin with, so we really shouldn't even care...

	} else {
	  # Ignore other errors, and just process it normally.
	}
      }

      if (rand() < $opts{TOTAL_SUBTREE_DEL}) {
	# Delete this subtree.
	eval {
	  # XXX This needs to log all the deletes to the manipulation log.
	  delete_local_tree($entry, VERBOSE => $opts{VERBOSE});
	};
	if ($@) {
	  $process_error->($@);
	} else {
	  print "Deleted $entry_path.\n" if ($opts{VERBOSE} >= 4);
	  $dbm->entry_deleted($entry);
	  next;
	}

      } else {
	# Process this subtree later.
	push @subpaths, $entry;
      }

    } else {

      # Modify files immediately (i.e. don't queue them).
      if (rand() < $opts{TOTAL_DEL}) {
	if (!unlink($entry)) {
	  $process_error->("unlink($entry_path): $!\n");
	} else {
	  if ($opts{LOG_FH}) {
	    log_manipulation($opts{LOG_FH}, ACTION => 'delete',
			     TYPE => 'file', PATH => $entry_path);
	  }
	  print "Deleted $entry_path.\n" if ($opts{VERBOSE} >= 4);
	  $dbm->entry_deleted($entry);
	  next;
	}

      } else {
	my $modified = eval { return modify_normal_file($entry, $st, %opts); };
	if ($@) {
	  $process_error->($@);
	} elsif ($modified && $opts{LOG_FH}) {
	  log_manipulation($opts{LOG_FH}, ACTION => 'modify',
			   TYPE => 'file', PATH => $entry_path);
	}
      }
    }

    # We did not delete the entry, and might have modified it.
    # Either way, it still exists, so remember enough about it
    # so that we don't try clobbering it later on.
    $dbm->entry_kept($entry);
  }

  # Having modified/deleted all the entries, we can now create new ones.

  # Maybe create new files...
  if ($file_cnt > 0) {
    print "   create $file_cnt files in $path\n" if ($opts{VERBOSE} >= 3);
    my @ents = $dbm->deleted_name_subset($file_cnt);
    my $cntr = undef;
    for (my $i = 0; $i < $file_cnt; $i++) {
      my $size = round randGauss($opts{CREATE_FILE_SIZE_AVG},
				 $opts{CREATE_FILE_SIZE_DEV});
      my $entry;
      if (scalar(@ents)) {
	$entry = shift @ents;
      } else {
	$entry = FILE_PREFIX . $dbm->get_free_cksum($cntr);
      }
      my $entry_path = File::Spec->catdir($path, $entry);

      if (!writeFile($entry, $size, %opts)) {
	$process_error->("create($entry_path): $!\n");
      } elsif ($opts{LOG_FH}) {
	log_manipulation($opts{LOG_FH}, ACTION => 'create',
			 TYPE => 'file', PATH => $entry_path);
      }
    }
  }

  # Maybe create new subdirectories...
  if ($dir_cnt > 0) {
    print "   create $dir_cnt directories in $path\n" if ($opts{VERBOSE} >= 3);
    my @ents = $dbm->deleted_name_subset($dir_cnt);
    my $cntr = undef;
    for (my $i = 0; $i < $dir_cnt; $i++) {
      my $more = round( randGauss($opts{SUBTREE_DEPTH_AVG},
				  $opts{SUBTREE_DEPTH_DEV}) - $depth );
      next unless ($more);

      my $entry;
      if (scalar(@ents)) {
	$entry = shift @ents;
      } else {
	$entry = DIR_PREFIX . $dbm->get_free_cksum($cntr);
      }
      my $entry_path = File::Spec->catdir($path, $entry);

      if (!mkdir($entry)) {
	$process_error->("mkdir($entry_path): $!\n");
      } else {
	if ($opts{LOG_FH}) {
	  log_manipulation($opts{LOG_FH}, ACTION => 'create',
			   TYPE => 'directory', PATH => $entry_path);
	}
	push @subpaths, $entry;
      }
    }
  }

  croak "closedir($path): $!\n" unless closedir($dirfh);
  print "Processed $path\n" if ($opts{VERBOSE} >= 3);

  return @subpaths;
}


# Sets the unspecified grow and shrink options to try to minimize the amount
# of overall filesystem growth/shrinkage across multiple modifications.
# This has to get called before we fill in defaults, or else we cannot
# distinguish default values from user-provided zeroes.
sub set_unspecified_modify_options( \% ) {
  my ($options) = @_;
  my $verbose = ((defined $options->{VERBOSE}) ? $options->{VERBOSE} : 0);

  # Set unspecified TOTAL_* and *_DEV values
  if ((defined $options->{TOTAL_GROW}) &&
      ($options->{TOTAL_GROW} > 0.0)) {
    unless (defined $options->{TOTAL_SHRINK}) {
      # Set SHRINK options from GROW options
      $options->{TOTAL_SHRINK} = $options->{TOTAL_GROW};
      if (defined $options->{GROW_DEV}) {
	$options->{SHRINK_DEV} = $options->{GROW_DEV}
	  unless (defined $options->{SHRINK_DEV});
      }
    }
  } elsif ((defined $options->{TOTAL_SHRINK}) &&
	    ($options->{TOTAL_SHRINK} > 0.0)) {
    unless (defined $options->{TOTAL_GROW}) {
      # Set GROW options from SHRINK options
      $options->{TOTAL_GROW} = $options->{TOTAL_SHRINK};
      if (defined $options->{SHRINK_DEV}) {
	$options->{GROW_DEV} = $options->{SHRINK_DEV}
	  unless (defined $options->{GROW_DEV});
      }
    }
  } else {
    # We are neither growing nor shrinking, so nothing to do.
    return;
  }

  if ((!!$options->{GROW_RAW}) != (!!$options->{SHRINK_RAW})) {
    print "Cannot fill defaults for mixed types.\n" if ($verbose);
    return;
  }

  # If TOTAL_GROW == TOTAL_SHRINK, then we're growing/shrinking the
  # same percentage of the files, and this code just ends up copying
  # the grow/shrink options to one another.
  # However, if TOTAL_GROW != TOTAL_SHRINK, then percentage of files
  # we're growing doesn't match the percentage we're shrinking, so
  # the amount by which we grow/shrink has to differ accordingly so that
  # the net effect is nil.  That's what this code does (if possible).
  if (defined $options->{GROW_AVG}) {
    if (defined $options->{SHRINK_AVG}) {
      my $change = ($options->{GROW_AVG} * $options->{TOTAL_GROW}) -
	($options->{SHRINK_AVG} * $options->{TOTAL_SHRINK});
      if ($verbose) {
	if ($options->{GROW_RAW}) {
	  printf("Filesystem should %s by approximately %.2f bytes/file.\n",
		 (($change > 0) ? "grow" : "shrink"), abs($change));
	} else {
	  printf("Filesystem should %s by approximately %.2f\%\n",
		 (($change > 0) ? "grow" : "shrink"), abs($change * 100.0));
	}
      }
    } else {
      # Set SHRINK_AVG to balance out GROW_AVG
      my $change = ($options->{GROW_AVG} * $options->{TOTAL_GROW});
      $options->{SHRINK_AVG} = $change / $options->{TOTAL_SHRINK};
    }
  } elsif (defined $options->{SHRINK_AVG}) {
    # Set GROW_AVG to balance out SHRINK_AVG
    my $change = ($options->{SHRINK_AVG} * $options->{TOTAL_SHRINK});
    $options->{GROW_AVG} = $change / $options->{TOTAL_GROW};
  } else {
    # We will neither grow nor shrink any file, so nothing to do.
  }

}


=head2 modify_local_tree($tree, %options)

Walks the specified tree and modifies its contents.  The nature and
extent of the modifications is determined by the options specified.

Croaks on any error.  Returns the random seed used on success.  In a
list context, also returns a reference to a list of subpaths, and a
reference to a list of subseeds.  (See the FORK_COUNT option.)

The net effect is that the routine makes a "set of changes" to the
tree as a whole.  This routine is not intended to provide random
client traffic or stress to the dataset.  (With the FORK_COUNT
parameter, it is often good at doing so, but that is not the goal, so
please do not request traffic-like features.)

Options:

=over 4

=item SEED

Specifies the seed with which to initialize the random number
generator.

=item ENVIRONMENT_TYPE

Specifies a group of default option values.

=item CREATE_FILE_SIZE_AVG

Mean size of newly-created files.

=item CREATE_FILE_SIZE_DEV

Standard deviation from CREATE_FILE_SIZE_AVG.

=item CREATE_FILE_AVG

Mean number of files to create in each directory.

=item CREATE_FILE_DEV

Standard deviation from CREATE_FILE_AVG.

=item CREATE_DIR_AVG

Mean number of subdirectories to create in each directory.

=item CREATE_DIR_DEV

Standard deviation from CREATE_DIR_AVG.

=item SUBTREE_DEPTH_AVG

Mean tree depth to maintain when creating new subdirectory trees.
(i.e. mean depth of terminal branches from the root of the tree).

=item SUBTREE_DEPTH_DEV

Standard deviation from SUBTREE_DEPTH_AVG.

=item DEPTH_BASE

When modifying subtrees (see FORK_COUNT), this specifies the base
depth of the subtree being modified.

=item TOTAL_SUBTREE_DEL

Average percentage of subdirectories to delete.

=item TOTAL_DEL

Average percentage of files to delete.

=item TOTAL_MOD

Average percentage of files to modify.

=item MOD_AVG

Mean percentage by which to modify files which are modified.

=item MOD_DEV

Standard deviation from MOD_AVG.

=item MOD_RAW

If set, treat MOD_AVG as a raw byte count instead of a percentage.

=item TOTAL_GROW

Average percentage of files to grow.

=item GROW_AVG

Mean percentage by which to grow files which are grown.

=item GROW_DEV

Standard deviation from GROW_AVG.

=item GROW_RAW

If set, treat GROW_AVG as a raw byte count instead of a percentage.

=item TOTAL_SHRINK

Average percentage of files to shrink.

=item SHRINK_AVG

Mean percentage by which to shrink files which are shrunk.

=item SHRINK_DEV

Standard deviation from SHRINK_AVG.

=item SHRINK_RAW

If set, treat SHRINK_AVG as a raw byte count instead of a percentage.

=item FRAGMENT_AVG

Mean number of fragments to write to files which are modified.

=item FRAGMENT_DEV

Standard deviation from FRAGMENT_AVG.

=item FRAGMENT_RAW

If set, tread FRAGMENT_AVG as raw byte count instead of a number of
fragments.  Turning this option on effectively makes FRAGMENT_AVG
specify the average write-size (i.e. write-chain length).

=item HOLD_FS_SIZE

If set, automatically set unspecified options to try to maintain the
overall size of the tree being modified.  That is, try to create and
delete the same number of files, try to grow and shrink files by the
same proportion/amount, etc.

=item FLOOD

If set, randomly-generated data blocks are just a randomly-chosen byte
repeated as necessary, rather than a complete randomly-generated
block.  This makes the "random" data pretty uninteresting, but it is
faster to generate.

=item SCRAMBLE

Percentage of successive blocks to scramble by XORing bytes in each BLOCK_SIZE
chunk of bytes written.  The higher this is, the more probable that each
BLOCK_SIZE sequence of bytes (probably) looks different, and therefore this
basically controls how compressible the data is.

=item FORK_COUNT

Specifies a number of tree subpaths to return.  If nonzero, only
modifies the tree until this minimum number of directory subpaths (at
approximately equal depth) are encountered.  A pair of list references
is then immediately returned along with the randomization seed.  The
first is the list of subpaths, and the second is a list of random seeds
to be used on those subpaths.  The idea behind this option is that the
calling process can then re-invoke this routine in multiple concurrent
processes, one for each returned subtree, using the corresponding seed
value to properly keep the randomization properties.

=item LOG_FILE

Specifies a file to which to log all actions performed.  This log file
(or some parsed version of it via C<DataManip::manip_log>) serves as a
input to other scripts which need to know exactly how data was
manipulated.  Beware that the log file specified should not exist at
the time of invokation, as the routine insists on doing an exclusive
create on it.  This is to protect against having separate processes
which are each operatiing on separate subtrees (via FORK_COUNT) from
attempting to concurretly write to the same log file.

=item MIN_DIR_SIZE

When modifying directories, the size of the directory is needed at
the onset in order to do some of the math for entry name generation.
However, there are two unfortunate problems in acquiring this:

=over 4

=item *

Different file systems consume directory space at different rates.
This library assumes a WAFL file system when estimating the number
of entries from the directory size, which might not be correct when
running against a native EXT3 or UFS file system.

=item *

Some popular operating systems (like Windows) do not seem to provide
an interface for retrieving the size of a directory.
On such systems (where the directory size always appears to be zero),
this option is internally defaulted to 20940.

=back

If you run out of memory when setting an entry, for either reason cited
above, try increasing the value of this option.  However, note that
doing so will force every directory modification to consume more memory.

=item PERSEVERE

If specified, causes any errors encountered to be printed without
terminating.  The run continues regardless of any errors.

=back

Finally, a brief philisophical rant (skip this section if you do not
care about how exactly this routine is designed and extended):

There are several ways in which the options for this routine could
have been construed.  The chosen philosophy to make this sane is to
"do what to/with every thing of some type".  The approach says to
manipulate objects individually, rather than as a group.  Feasibility
forces us to avoid manipulations that require over-arching knowledge
about everything in a directory, or everything in a subtree, etc.

One example of this in action is the difference in how we create and
delete files.  We specify how many files to create per directory, but
not how many files to delete per directory, since that would require
knowing how many files are in the directory first.  Creating a file
is something you do to the parent directory, while deleting a file is
something you do to the file.

=cut

# This list contains the subset of options defined in each environment
# type which pertain to modify_local_tree().  Other options which
# modify_local_tree() accepts (but which do not have default values in
# each environment type) are included manually where needed.
our @modify_options =
  qw(CREATE_FILE_AVG  CREATE_DIR_AVG  CREATE_FILE_SIZE_AVG  SUBTREE_DEPTH_AVG
     CREATE_FILE_DEV  CREATE_DIR_DEV  CREATE_FILE_SIZE_DEV  SUBTREE_DEPTH_DEV
     TOTAL_MOD  TOTAL_GROW  TOTAL_SHRINK  TOTAL_DEL   TOTAL_SUBTREE_DEL
     MOD_AVG    GROW_AVG    SHRINK_AVG    FRAGMENT_AVG
     MOD_DEV    GROW_DEV    SHRINK_DEV    FRAGMENT_DEV
     MOD_RAW    GROW_RAW    SHRINK_RAW    FRAGMENT_RAW
     DEPTH_BASE MIN_DIR_SIZE
     HOLD_FS_SIZE
     FLOOD SCRAMBLE
     FORK_COUNT PERSEVERE LOG_FILE VERBOSE);

sub modify_local_tree( $% ) {
  my $tree = shift;
  my %opts = @_;
  legacy_options_compatibility(%opts, @_);

  $tree = File::Spec->canonpath($tree); # chop trailing slashes, etc.

  # The HOLD_FS_SIZE option requires us to set some options depending
  # upon how others are set, but without overriding what is already
  # set.  We need to do this *before* setting all options from the
  # environment type, or else there will not be anything left to set.
  if ((defined $opts{HOLD_FS_SIZE}) && $opts{HOLD_FS_SIZE}) {
    set_unspecified_modify_options(%opts);
  }

  # Define all options using values from the environment type.
  set_option_defaults(%opts, @modify_options);

  # Set up the log file.
  if ($opts{LOG_FILE}) {
    $opts{LOG_FH} = open_log_file($opts{LOG_FILE});
  }

  # Seed the random number generator
  srand($opts{SEED});

  if ($opts{VERBOSE} || $opts{SHOW_OPTS}) {
    print_args(tree => $tree);
    print_options(%opts, 'SEED', 'ENVIRONMENT_TYPE', @modify_options);
  }

  # All tolerable errors will be processed by this routine.
  my $process_error = $opts{PERSEVERE} ?
    sub { print $_[0]; } : sub { croak $_[0]; };

  # Check whether we're working on a single file or a directory tree.
  unless (-d $tree) {
    my $node = $tree;
    my $st = stat($node);
    if ($st) {
      # Just modify the single file.
      my $modified = eval { modify_normal_file($node, $st, %opts); };
      if ($@) {
	$process_error->($@);
      } elsif ($modified && $opts{LOG_FH}) {
	log_manipulation($opts{LOG_FH}, ACTION => 'modify',
			 TYPE => 'file', PATH => $node);
      }

    } else {
      $process_error->("stat($node): $!\n")
    }

    if (wantarray && $opts{FORK_COUNT}) {
      return ($opts{SEED}, []);
    } else {
      return $opts{SEED};
    }
  }

  chdir($tree) or croak "chdir($tree): $!\n";

  # Unfortunately, cwd() prints errors when they happen...
  my $treeroot = cwd() or croak "cwd failed\n";

  # Recursively modify the tree
  my @pathlist;
  if ($opts{FORK_COUNT}) {
    # Breadth-first treewalk...
    # All paths are relative from $treeroot.
    @pathlist = ( [ $opts{DEPTH_BASE}, CURDIR ] );
    while ((scalar(@pathlist) > 0) &&
	   (scalar(@pathlist) < $opts{FORK_COUNT})) {

      my ($depth, $path) = @{shift @pathlist};

      if (chdir($path)) {
	my @subdirs = modify_directory($path, $depth, $process_error, %opts);
	push @pathlist, map {
	  [ $depth + 1, File::Spec->catdir($path, $_) ]
	} @subdirs;

	# Go back to the root of the tree for the next iteration.
	chdir($treeroot) or croak "chdir($treeroot): $!\n";

      } else {
	$process_error->("chdir($path): $!\n");
      }
    }

  } else {
    # Depth-first treewalk...
    # All paths are relative from current directory.
    @pathlist = ( CURDIR );
    my $depth = $opts{DEPTH_BASE} - 1;
    my @entrystack = ();
    while (scalar(@pathlist) > 0) {
      my $entry = pop @pathlist;

      if ($entry eq UPDIR) {
	chdir(UPDIR) or croak "chdir(" . UPDIR . "): $!\n";
	pop @entrystack;
	$depth--;
	next;
      }

      my $path = File::Spec->catdir(@entrystack, $entry);

      if (chdir($entry)) {
	push @entrystack, $entry;
	$depth++;
	push @pathlist, UPDIR;
	push @pathlist, modify_directory($path, $depth, $process_error, %opts);
      } else {
	$process_error->("chdir($path): $!\n");
      }
    }
  }

  if ($opts{LOG_FH}) {
    unless (close($opts{LOG_FH})) {
      croak "error closing log file: $!\n";
    }
    delete $opts{LOG_FH};
  }

  if (wantarray && $opts{FORK_COUNT}) {
    my @paths = map { $_->[1] } @pathlist;
    my @subseeds = map { int(rand() * UINT4_MAX) } @paths;
    return ( $opts{SEED}, \@paths, \@subseeds );
  } else {
    return $opts{SEED};
  }
}


=head2 populate_local_tree($tree, %options)

Populates the specified directory with a complete data tree.  Exactly
what is created in this data tree is determined by the options
specified.

Croaks on any error.  Returns the random seed used on success.  In a
list context, also returns a reference to a list of subpaths, and a
reference to a list of subseeds.  (See the FORK_COUNT option.)

Options:

=over 4

=item SEED

Specifies the seed with which to initialize the random number
generator.

=item ENVIRONMENT_TYPE

Specifies a group of default option values.

=item DIR_AVG

Mean number of subdirectories per directory
(i.e. mean branch factor of the tree).

=item DIR_DEV

Standard deviation from DIR_AVG.

=item FILE_AVG

Mean number of files per directory.

=item FILE_DEV

Standard deviation from FILE_AVG.

=item DEPTH_AVG

Mean tree depth
(i.e. mean depth of terminal branches from the root of the tree).

=item DEPTH_DEV

Standard deviation from DEPTH_AVG.

=item DEPTH_BASE

When populating subtrees (see FORK_COUNT), this specifies the base
depth of the subtree being populated.

=item BREADTH_AVG

Mean file size.

=item BREADTH_DEV

Standard deviation from BREADTH_AVG.

=item FLOOD

If set, randomly-generated data blocks are just a randomly-chosen byte
repeated as necessary, rather than a complete randomly-generated
block.  This makes the "random" data pretty uninteresting, but it is
faster to generate.

=item SCRAMBLE

Percentage of successive blocks to scramble by XORing bytes in each BLOCK_SIZE
chunk of bytes written.  The higher this is, the more probable that each
BLOCK_SIZE sequence of bytes (probably) looks different, and therefore this
basically controls how compressible the data is.

=item FORK_COUNT

Specifies a number of tree subpaths to return.  If nonzero, only
populates the tree until this minimum number of directory subpaths (at
approximately equal depth) are encountered.  A pair of list references
is then immediately returned along with the randomization seed.  The
first is the list of subpaths, and the second is a list of random seeds
to be used on those subpaths.  The idea behind this option is that the
calling process can then re-invoke this routine in multiple concurrent
processes, one for each returned subtree, using the corresponding seed
value to properly keep the randomization properties.

=item LOG_FILE

Specifies a file to which to log all actions performed.  This log file
(or some parsed version of it via C<DataManip::manip_log>) serves as a
input to other scripts which need to know exactly how data was
manipulated.  Beware that the log file specified should not exist at
the time of invokation, as the routine insists on doing an exclusive
create on it.  This is to protect against having separate processes
which are each operatiing on separate subtrees (via FORK_COUNT) from
attempting to concurretly write to the same log file.

=item MIN_DIR_SIZE

When modifying directories, the size of the directory is needed at
the onset in order to do some of the math for entry name generation.
However, there are two unfortunate problems in acquiring this:

=over 4

=item *

Different file systems consume directory space at different rates.
This library assumes a WAFL file system when estimating the number
of entries from the directory size, which might not be correct when
running against a native EXT3 or UFS file system.

=item *

Some popular operating systems (like Windows) do not seem to provide
an interface for retrieving the size of a directory.
On such systems (where the directory size always appears to be zero),
this option is internally defaulted to 20940.

=back

If you run out of memory when setting an entry, for either reason cited
above, try increasing the value of this option.  However, note that
doing so will force every directory modification to consume more memory.

=item PERSEVERE

If specified, causes any errors encountered to be printed without
terminating.  The run continues regardless of any errors.

=back

The names of the population parameters are intentionally distinct from
the modification parameters.  This allows us to define environment
types for gradually (or quickly) mutating datasets.

=cut

# This list contains the subset of options defined in each environment
# type which pertain to populate_local_tree().  Other options which
# populate_local_tree() accepts (but which do not have default values
# in each environment type) are included manually where needed.
our @populate_options =
  qw(DIR_AVG  FILE_AVG  DEPTH_AVG  BREADTH_AVG
     DIR_DEV  FILE_DEV  DEPTH_DEV  BREADTH_DEV
     DEPTH_BASE MIN_DIR_SIZE
     FLOOD SCRAMBLE
     FORK_COUNT PERSEVERE LOG_FILE VERBOSE);

sub populate_local_tree( $% ) {
  my $tree = shift;
  my %opts = @_;
  legacy_options_compatibility(%opts, @_);

  $tree = File::Spec->canonpath($tree); # chop trailing slashes, etc.

  # We will only populate empty directories.
  opendir(DH, $tree) or croak "opendir($tree): $!\n";
  while (my $entry = readdir(DH)) {
    if (scalar File::Spec->no_upwards($entry) && $entry ne ".snapshot") {
      closedir(DH);
      $! = ENOTEMPTY;
      croak "$!\n";
    }
  }
  closedir(DH) or croak "closedir($tree): $!\n";

  # Define all options using values from the environment type.
  set_option_defaults(%opts, @populate_options);

  if ($opts{VERBOSE} || $opts{SHOW_OPTS}) {
    print_args(tree => $tree);
    print_options(%opts, 'SEED', 'ENVIRONMENT_TYPE', @populate_options);
    delete $opts{SHOW_OPTS};
  }

  # Do not specify any modify options unless they are also populate options.
  my %pop_opts = map { $_ => 1 } @populate_options;
  foreach my $mod_opt (@modify_options) {
    delete $opts{$mod_opt} unless (exists $pop_opts{$mod_opt});
  }

  # Set the specific modify options we want.
  $opts{ENVIRONMENT_TYPE} = 'nil';
  $opts{CREATE_DIR_AVG} = $opts{DIR_AVG};
  $opts{CREATE_DIR_DEV} = $opts{DIR_DEV};
  $opts{CREATE_FILE_AVG} = $opts{FILE_AVG};
  $opts{CREATE_FILE_DEV} = $opts{FILE_DEV};
  $opts{SUBTREE_DEPTH_AVG} = $opts{DEPTH_AVG};
  $opts{SUBTREE_DEPTH_DEV} = $opts{DEPTH_DEV};
  $opts{CREATE_FILE_SIZE_AVG} = $opts{BREADTH_AVG};
  $opts{CREATE_FILE_SIZE_DEV} = $opts{BREADTH_DEV};

  return modify_local_tree($tree, %opts);
}


# usage: contains($item, @list)
# Returns whether the given items is in the given list
sub contains( $@ ) {
  my $item = shift;
  foreach (@_) {
    return 1 if ($_ eq $item);
  }
  return 0;
}

# Returns the contents of a directory, without "." or "..".
# Croaks on any errors.
sub ls( $ ) {
  my ($dir) = @_;
  opendir(DIR, $dir) or croak "opendir($dir): $!\n";
  my @contents = readdir DIR;
  closedir(DIR) or croak "closedir($dir): $!\n";
  return File::Spec->no_upwards(@contents);
}

# Given a mode from a stat structure, returns whether or not the mode
# describes some type of "device".  Here, device refers more to
# "some non-comparable, non-symlink, non-normal-file thing".
sub mode_is_dev( $ ) {
  my ($mode) = @_;
  my $isdev = (($mode & S_IFCHR) == S_IFCHR); # is a character device
  unless ($^O eq WINDOWS) {
    # Windows perl does not define these Fnctl symbols.
    $isdev = ($isdev
	      || (($mode & S_IFSOCK) == S_IFSOCK) # is a socket device
	      || (($mode & S_IFBLK) == S_IFBLK) # or block device
	      || (($mode & S_IFIFO) == S_IFIFO) # or named pipe
	     );
  }
  return $isdev;
}

# Given a mode from a stat structure, returns whether or not the mode
# describes a directory.  One might ask why we don't just use the "-d"
# test in perl.  Well, two reasons:  First, we don't know exactly what
# it does on each and every perl platform, and second, we want to make
# sure to only stat each inode exactly once.
sub mode_is_dir( $ ) {
  my ($mode) = @_;
  my $isdev = mode_is_dev($mode);
  my $isdir = ((($mode & S_IFDIR) == S_IFDIR) # is normal directiory
	       && !$isdev); # block devices have the directory bit set
  return $isdir;
}

# Takes the shell command for a checksumming program, and
# the paths to two files to compare via checksums.
# Computes a checksum for each file, and compares the checksums.
# Croaks if something goes awry while computing checksums.
sub cksum_compare( $$$ ) {
  my ($cksum, $path1, $path2) = @_;

  my $cs1 = qx/$cksum $path1/;
  if ($?) {
    chomp $cs1;
    croak "$cs1\n";
  }
  $cs1 =~ s/^\s*(\w+).*$/$1/s;

  my $cs2 = qx/$cksum $path2/;
  if ($?) {
    chomp $cs2;
    croak "$cs2\n";
  }
  $cs2 =~ s/^\s*(\w+).*$/$1/s;

  return ($cs1 ne $cs2);
}

=head2 compare_local_trees($tree1, $tree2, %options)

Compare two trees and their contents.  Croaks on any error.

On success, returns three references, each of which refers
to a list of paths relative to the base of the trees being
compared.

=over 4

=item 1

The first lists all the paths present in $tree1 but not $tree2.

=item 2

The second lists all the paths present in $tree2 but not $tree1.

=item 3

The third lists the paths present in both trees,
but whose contents or attributes differed in some way.

=back

When the FORK_COUNT option is specified, returns an additional
list reference (to a list of subpaths).

Options:

=over 4

=item CKSUM

Specifies a checksumming program, typically C<cksum> or C<md5sum>,
which exist on most unix platforms.  Instead of comparing the contents
of the files, computes the checksum of each with the checksumming
program and then compares the the checksums.  On datasets with very
large files, this will tend to be faster than comparing each file
byte-by-byte.

=item GLANCE

If specified, does not compare the contents of the files.  Instead,
only checks whether files are present.  The third list returned
will be empty.

=item ATTRIBUTES

If specified, compares file and directory attributes, in addition to
file contents.

=item IGNORE_LIST

Specifies a list of paths (relative to the base of each tree) to
ignore when doing the compare.  The actual value passed here should
be a serialized scalar to evaluate.  (See C<FSR::FSRutils::serialize()>.)

=item FORK_COUNT

Specifies a number of tree subpaths to return.  If nonzero, only
compares the trees until this minimum number of common directory subpaths
(at approximately equal depth) are encountered.  A list of the remaining
subpaths is then immediately returned.  The idea behind this option is that
the calling process can then re-invoke this routine in multiple concurrent
processes, one for each returned subtree.

=back

=cut

our @compare_options = qw(CKSUM GLANCE ATTRIBUTES IGNORE_LIST
			  FORK_COUNT VERBOSE);

sub compare_local_trees( $$% ) {
  my $tree1 = shift;
  my $tree2 = shift;
  my %opts = @_;
  legacy_options_compatibility(%opts, @_);

  # Chop trailing slashes, etc.
  $tree1 = File::Spec->canonpath($tree1);
  $tree2 = File::Spec->canonpath($tree2);

  # This routine does not use environment types, so set any
  # unprovided options to zero, to avoid "Uninitialized value"
  # warnings).
  unless (defined $opts{IGNORE_LIST}) {
    # Need to set IGNORE_LIST separately, since zero is not
    # the default "empty" value we want for it.
    $opts{IGNORE_LIST} = '()';
  }
  define_default_options(%opts, @compare_options);

  if ($opts{VERBOSE} || $opts{SHOW_OPTS}) {
    print_args(tree1 => $tree1, tree2 => $tree2);
    print_options(%opts, @compare_options);
  }

  if ($opts{IGNORE_LIST} && $opts{FORK_COUNT}) {
    croak "IGNORE_LIST currently incompatible with FORK_COUNT\n";
    # Really, this is just because DataManip.pm
    # and the CLI scripts don't correctly handle it yet.
  }

  my @ignore = eval $opts{IGNORE_LIST};
  croak $@ if ($@);

  # Just for sanity.
  lstat($tree1) or croak "stat($tree1): $!\n";
  lstat($tree2) or croak "stat($tree2): $!\n";

  my @in1 = ();
  my @in2 = ();
  my @differ = ();

  my @pathlist = ( CURDIR );

  # Recursively compare the two trees
  while (scalar(@pathlist) > 0) {
    my $relpath = pop @pathlist;

    # Skip if on the ignore list
    next if (contains($relpath, @ignore));

    # Check for both, but ignore the error if the inode doesn't exist...
    # ENOTDIR occurs when a parent directory exists but is not a directory.
    # ENOENT occurs when directory entry lookup fails.
    my $path1 = File::Spec->catdir($tree1, $relpath);
    my $path2 = File::Spec->catdir($tree2, $relpath);
    my $errstr = undef;	# Need to save the error string temporarily
    my $st1 = lstat($path1);
    unless ($st1 || $!{ENOENT} || $!{ENOTDIR}) {
      $errstr = $!;
    }
    my $st2 = lstat($path2);
    unless ($st2 || $!{ENOENT} || $!{ENOTDIR}) {
      $errstr = $! unless ($errstr); # keep first error encountered
    }

    croak "stat($relpath): $errstr\n" if ($errstr);

    # Check if it exists in both trees
    if ($st1) {
      unless ($st2) {
	print "Only in $tree1: $relpath\n" if ($opts{VERBOSE} >= 3);
	push @in1, $relpath;
      }
    } elsif ($st2) {
      print "Only in $tree2: $relpath\n" if ($opts{VERBOSE} >= 3);
      push @in2, $relpath;
    } else { # Not present in *either* tree?!?
      $! = ENOENT;
      croak "stat($relpath): $!\n";
    }

    # If both exist, we may need to add this entry to the @differ list below.
    # If one doesn't exist, then we already added this entry to @in1 or @in2.
    # If neither exists, then we should have died above.

    my ($isdir1, $isdir2) = (0, 0);
    $isdir1 = mode_is_dir($st1->mode) if ($st1);
    $isdir2 = mode_is_dir($st2->mode) if ($st2);

    if ($st1 && $st2) { # both exist

      my ($mode1, $mode2) = ($st1->mode, $st2->mode);

      # If we find a difference, set this variable and jump...
      my $differ = undef;

      unless (($mode1 & S_IFMT) == ($mode2 & S_IFMT)) {
	# They're not the same type of file/directory/device/etc.
	print "File mode differs on $relpath\n" if ($opts{VERBOSE} >= 3);
	$differ = 1;
	goto FOUND_DIFF;
      }

      # At this point, the modes should be the same...
      croak "modes are same but only one is a directory?\n"
	unless ($isdir1 == $isdir2);

      my $isdev = mode_is_dev($mode1);

      if ($opts{ATTRIBUTES}) {
	if (($st1->uid != $st2->uid) ||
	    ($st1->gid != $st2->gid) ||
	    (($mode1 & ~S_IFMT) != ($mode2 & ~S_IFMT)) || # permissions
	    ($isdev && ($st1->rdev != $st2->rdev))
	   ) {
	  print "Attributes differ on $relpath\n" if ($opts{VERBOSE} >= 3);
	  $differ = 1;
	  goto FOUND_DIFF;
	}
      }

      # XXX divide "EXTRA" into "TIMESTAMPS" and "LINK_COUNT"
      if ($opts{EXTRA}) {
	if ($st1->mtime != $st2->mtime ||
	    $st1->ctime != $st2->ctime ||
	    $st1->nlink != $st2->nlink) {
	  print "Extras differ on $relpath\n" if ($opts{VERBOSE} >= 3);
	  $differ = 1;
	  goto FOUND_DIFF;
	}
      }

      unless ($isdir1) { # both are some type of file

	if ($st1->size != $st2->size) {
	  # A size mismatch on a non-directory is a guaranteed difference.
	  $differ = 1;

	} elsif (!$opts{GLANCE}) {
	  my $isreg = (($mode1 & S_IFREG) == S_IFREG); # is regular file
	  my $islnk = 0;
	  unless ($^O eq WINDOWS) {
	    # Windows perl does not define this Fnctl symbol.
	    $islnk = (($mode1 & S_IFLNK) == S_IFLNK); # is symbolic link
	  }

	  if ($islnk) {	# both are symlinks -- read and compare manually
	    my $sym1 = readlink($path1) or croak "readlink($path1): $!\n";
	    my $sym2 = readlink($path2) or croak "readlink($path2): $!\n";
	    $differ = ($sym1 ne $sym2);

	  } elsif ($isdev) { # both are "devices"
	    # By our definition in mode_is_dev(), these aren't comparable

	  } elsif ($isreg) { # both are regular files
	    if ($opts{CKSUM}) {
	      # Compare the files by checksumming.
	      $differ = cksum_compare($opts{CKSUM}, $path1, $path2);
	    } else {
	      # Compare the contents of the files byte-by-byte.
	      $differ = compare($path1, $path2);
	      croak "compare($path1, $path2): $!\n" if ($differ < 0);
	    }
	  } else {
	    croak "unknown file mode on $relpath\n";
	  }
	}

	if ($differ) {
	  print "File contents on $relpath differ\n" if ($opts{VERBOSE} >= 3);
	  goto FOUND_DIFF;
	}
      }

    FOUND_DIFF:
      if ($differ) {
	push @differ, $relpath;
      }

    } # end of "both exist" condition

    # Only keep going if at least one of these is a directory.
    next unless ($isdir1 || $isdir2);

    print "Recurse on directory $relpath\n" if ($opts{VERBOSE} >= 5);
    my @dir1 = ();
    @dir1 = ls($path1) if ($isdir1);
    my @dir2 = ();
    @dir2 = ls($path2) if ($isdir2);

    # Merge the two lists
    my @only2 = grep { !contains($_, @dir1) } @dir2;
    my @merged = map { File::Spec->catdir($relpath, $_) } (@dir1, @only2);

    if ($opts{FORK_COUNT}) {
      unshift @pathlist, @merged; # breadth-first
      last if (scalar(@pathlist) < $opts{FORK_COUNT});
    } else {
      push @pathlist, @merged;	# depth-first
    }

  }

  if ($opts{VERBOSE}) {
    printf("Done comparing:  %d in $tree1, %d in $tree2, and %d differed.\n",
	   scalar(@in1), scalar(@in2), scalar(@differ));
  }

  return ( \@in1, \@in2, \@differ, \@pathlist );
}

=head2 delete_local_tree($tree, %options)

Deletes the given tree, and everything under it.

Returns true, or croaks on any error.

Options:

=over 4

=item CONTENTS_ONLY

If set, do not delete tree itself, just everything present within it.

=back

=cut

our @delete_options = qw(CONTENTS_ONLY VERBOSE);

sub delete_local_tree( $% ) {
  my $tree = shift;
  my %opts = @_;
  legacy_options_compatibility(%opts, @_);

  $tree = File::Spec->canonpath($tree); # chop trailing slashes, etc.

  # This routine does not use environment types, so set any
  # unprovided options to zero, to avoid "Uninitialized value"
  # warnings).
  define_default_options(%opts, @delete_options);

  if ($opts{VERBOSE} || $opts{SHOW_OPTS}) {
    print_args(tree => $tree);
    print_options(%opts, @delete_options);
  }

  # Delete $tree if it's a single file.
  unless (-d $tree) {
    unlink($tree) or croak "rm($tree): $!\n";
    return 1;
  }

  # Remove all files iteratively:
  # Try removing everything in this directory.
  # Step into a non-empty sub-directory, if present.
  # Repeat this until everything in the current directory is gone,
  # then step back into to the parent directory.
  # Use depth to track when we're done.

  my $depth = 1;
  print "Depth $depth, Process $tree\n" if ($opts{VERBOSE} >= 2);
  chdir($tree) or croak "chdir: $!\n";

 DIRECTORY: while ($depth > 0) {
    opendir(DIR, CURDIR);

  ENTRY: while (my $entry = readdir(DIR)) {
      # Always skip over "." and ".."
      next ENTRY unless (scalar File::Spec->no_upwards($entry));

      if (! -d $entry) { # It's a file; remove it.

	print "Depth $depth, unlink $entry\n" if ($opts{VERBOSE} >= 3);
	unlink($entry) or croak "unlink($entry): $!\n";

      } else { # It's a directory; step inside if we can't remove it.

	print "Depth $depth, rmdir $entry\n" if ($opts{VERBOSE} >= 4);
	next ENTRY if (rmdir $entry);

	# Skip the ".snapshot" directory.
	if (($depth == 1) && ($entry eq ".snapshot") &&
	    ($!{EACCES} || $!{EPERM})) {
	  next ENTRY;
	}

	croak "rmdir: $!\n"
	  unless ($!{ENOTEMPTY} || $!{EEXIST});

	# Subdirectory is not empty.
	# Step into the subdirectory to delete its entries.
	closedir(DIR) or croak "closedir: $!\n";
	$depth++;
	print "Depth $depth, Process $entry\n" if ($opts{VERBOSE} >= 2);
	chdir($entry) or croak "chdir: $!\n";
	next DIRECTORY;
      }

    }
    closedir(DIR) or croak "closedir: $!\n";

    # Successfully removed everything in this directory.
    print "Depth $depth, directory empty\n" if ($opts{VERBOSE} >= 3);
    chdir(UPDIR) or croak "chdir(" . UPDIR . "): $!\n";
    $depth--;
  }

  # modify_local_tree relies on this function finishing in the parent
  # directory of the deleted tree, so do not change directories anymore

  unless ($opts{CONTENTS_ONLY}) {
    # Delete the tree itself.
    rmdir($tree) or croak "rmdir($tree): $!\n";
  }

  return 1;
}

=head2 writeFile($file_name, $byte_count, %options)

Writes the specified number of random bytes to a file.

Returns success or failure (with error message in $!).

Options:

=over 4

=item OFFSET

Offset at which to write the random data (default=0).

=item TRUNC

If set, truncates the file to length 0 prior to writing any data.

=item FLOOD

If set, random data block written to the file is just a single
randomly-chosen byte repeated as necessary, rather than a
randomly-generated block.  (This makes the data pretty uninteresting,
but it is faster to generate.)

=item SCRAMBLE

Percentage of successive blocks to scramble by XORing bytes in each BLOCK_SIZE
chunk of bytes written.  The higher this is, the more probable that each
BLOCK_SIZE sequence of bytes (probably) looks different, and therefore this
basically controls how compressible the data is.

=back

=cut

sub writeFile( $$;% ) {
  my ($file, $byte_cnt, %opts) = @_;

  # Since we're just doing a lot of system calls, we
  # return an error in $!, not $@.  As such, we need to
  # make sure to return immediately if something goes
  # wrong, so as not to destroy the value of $!.  This
  # means that we can't close the file if something fails.

  my $fh; # empty filehandle reference

  sysopen($fh, $file, (O_WRONLY|O_CREAT)) or return undef;
  binmode $fh;
  if (defined $opts{TRUNC}) {
    truncate($fh, $opts{TRUNC}) or return undef;
  }
  if ($byte_cnt <= 0) {
    close($fh) or return undef;
    return 1;
  }
  if (defined $opts{OFFSET}) {
    sysseek($fh, $opts{OFFSET}, SEEK_SET) or return undef;
  }

  # Make some random data
  my @data;
  my $max = ($byte_cnt < BLOCK_SIZE) ? $byte_cnt : BLOCK_SIZE;
  if ($opts{FLOOD}) {
    @data = (int rand(256)) x $max;
  } else {
    $#data = ($max - 1);
    foreach (@data) {
      $_ = int rand(256);
    }
  }

  writeBytes($fh, $byte_cnt, \@data, %opts) or return undef;

  close($fh) or return undef;
  return 1;
}

=head2 mkdir_path($directory_path, $permissions_mask)

This is essentially just a perl-native version of the C<mkdir -p>
shell command.

Given a directory path, creates the directory (and all parent
directories, if necessary).  The option permissions mask argument is
the same as for C<mkdir>, and defaults to C<0777>.  Does nothing if
the directory already exists.

Returns success or failure (with an error message in $!).

=cut

sub mkdir_path( $;$ ) {
  my ($dirpath, $perms) = @_;
  $perms = 0777 unless (defined $perms);
  $dirpath = File::Spec->canonpath($dirpath);

  return 1 if (-d $dirpath);

  my ($vol, $dirs, $file) = File::Spec->splitpath($dirpath, 1); # 1 -> "no file"
  my $chunk;
  if (File::Spec->file_name_is_absolute($dirpath)) {
    my $root = File::Spec->rootdir();
    $chunk = $vol ? File::Spec->catpath($vol, $root, "") : $root;
    $dirs = File::Spec->abs2rel($dirs, $chunk);
  } else {
    $chunk = CURDIR;
  }

  my @parts = File::Spec->splitdir($dirs);

  # Run through all the parts in order and
  # create the directories.
  while (scalar(@parts) > 0) {
    $chunk = File::Spec->catdir($chunk, shift @parts);
    next if (-d $chunk);
    return undef unless (mkdir($chunk, $perms));
  }

  return 1;
}


=head1 EXAMPLES

These examples invoke the command-line wrappers in
F</usr/software/test/bin/>.  For this to work, F</usr/software/test/bin/>
must be in the PATH environment variable.

First, let us create some random data with C<populate_tree>:

  $ cd /tmp
  $ mkdir datatree
  $ populate /tmp/datatree ENVIRONMENT_TYPE=filer
  Population completed in 00:00:20 (20 seconds).
  Population of /tmp/datatree successful.  (Seed was 1503177889.)
  $ du -sh datatree
  33M	datatree
  $ ls datatree
  data0  data1  dir0  dir1
  $ ls datatree/dir1
  data0  data2  data4  data6  data8  dir1  dir3
  data1  data3  data5  data7  dir0   dir2
  $

OK, that was fun.  Now we have 33 MB of random files and directories.
Remember, all this is randomness is repeatable, if we use the same
options and seed.  So let us create an identical dataset:

  $ mkdir dataflower
  $ populate /tmp/dataflower ENVIRONMENT_TYPE=filer SEED=1503177889
  Population completed in 00:00:20 (20 seconds).
  Population of /tmp/dataflower successful.  (Seed was 1503177889.)
  $ du -sh dataflower
  33M	dataflower
  $ 
  $ compare /tmp/datatree /tmp/dataflower ATTRIBUTES=1 VERBOSE=1
  CKSUM => '0'
  GLANCE => '0'
  ATTRIBUTES => '1'
  IGNORE_LIST => '()'
  VERBOSE => '1'
  Done comparing:  0 in /tmp/datatree, 0 in /tmp/dataflower, and 0 differed.
  Comparison completed in 00:00:10 (10 seconds).
  0 files in /tmp/datatree but not /tmp/dataflower:
  	
  0 files /tmp/dataflower but not /tmp/datatree:
  	
  0 files differed:
  	
  $ 

So now we have two data trees, and C<compare_trees> says that they
are identical (which we were expecting).  Now let us modify 2 percent
of the files in the first tree:

  $ modify /tmp/datatree ENVIRONMENT_TYPE=filer TOTAL_MOD=0.02 TOTAL_GROW=0 TOTAL_SHRINK=0
  Modification completed in 00:00:01 (1 seconds).
  Modification of /tmp/datatree succeeded.  (Seed was 1523783568.)
  $ compare /tmp/datatree /tmp/dataflower ATTRIBUTES=1
  Comparison completed in 00:00:02 (2 seconds).
  0 files in /tmp/datatree but not /tmp/dataflower:
          
  0 files /tmp/dataflower but not /tmp/datatree:
          
  23 files differed:
          dir0/dir0/dir0/dir2/data4
          dir0/dir0/dir2/data3
          dir0/dir2/dir0/dir0/dir1/data6
          dir0/dir2/dir0/dir2/data4
          dir0/dir2/dir1/dir0/dir1/data1
          dir0/dir2/dir1/dir3/dir3/data0
          dir0/dir2/dir1/dir3/dir5/data1
          dir1/dir0/dir0/dir2/dir0/data3
          dir1/dir0/dir0/dir3/data2
          dir1/dir0/dir1/dir0/data2
          dir1/dir0/dir1/dir3/data5
          dir1/dir0/dir3/data2
          dir1/dir1/dir0/dir0/dir4/data4
          dir1/dir1/dir0/data1
          dir1/dir1/dir1/dir0/dir0/data1
          dir1/dir2/dir2/dir0/dir1/dir4/data0
          dir1/dir2/dir3/dir0/data0
          dir1/dir2/data4
          dir1/dir3/dir0/dir2/data5
          dir1/dir3/dir1/dir1/dir1/data2
          dir1/dir3/dir1/data1
          dir1/dir3/dir3/dir0/data2
          dir1/dir3/dir3/dir1/dir2/data3
  $ 

The modification changed a bunch of files, which C<compare_trees> has verified
for us.  Last but not least, let us clean up:

  $ delete /tmp/datatree
  Deletion completed in 00:00:00 (0 seconds).
  Deletion of datatree successful.
  $ ls datatree
  ls: datatree: No such file or directory
  $ delete /tmp/dataflower CONTENTS_ONLY=1 VERBOSE=1
  CONTENTS_ONLY => '1'
  VERBOSE => '1'
  Deletion completed in 00:00:00 (0 seconds).
  Deletion of /tmp/dataflower successful.
  $ ls dataflower
  $ rmdir dataflower
  $ 

Note, the examples above all gave absolute paths to the command-line wrappers.
This is not necessary -- relative paths work too.

=head1 BUGS

Deletes and creates are not always logged to the manipulation log.

HOLD_FS_SIZE still does not do anything for file creates and deletes.

The CKSUM option to C<compare_local_trees()> does not work correctly
on Windows clients, since it does an unsafe quoted execution.

The LOG_FILE parameter doesn't work with forked subpaths.
(Really, this is a bug in DataManip.pm and the CLI scripts.)

The IGNORE_LIST parameter doesn't work with FORK_COUNT.

The code for scrambling more than one byte per block (to make
the data less compressible) really needs to be analyzed further.
Without the "stride" code, the graph of compression ration as
a function of number of bytes scrambled converged nicely
from >100 to 1.0 as the number of bytes scrambled increased up
to 4096.  Unfortunately, that took way too much CPU to get the
compression ratio down into a reasonable level.  The "stride"
code made it converge down to 1.0 much faster (by 2048), however
it now has the strange side-effect that scrambling *more* than
2048 bytes per block will actually increase the compression
ratio.  Actually, it goes all the way back up to something very
compressible as you approach scrambling the entire block!
(This is really, really weird.)
   - Update: This happens because after scrambling half the blocks
(the odd ones), we continued by scrambling the even ones, each one
of which unscrambled the corresponding odd block (since if a ^ b = c,
then b ^ c = a).  The solution:  we need to "stride" with rounded
decimal values, the way writeFragment() currently spreads out the
changed fragments.  If we "stride" with integral values, we're always
going to get discontinuous regions in the compression graph.

=head1 TODO

=over 4

=item *

Add a FORK_RAW option for the breadth-first walk to always return all
the subpaths at the specified depth.

=item *

Deleting subtrees will tend to shrink the overall tree faster than
an equal amount of subtree creation will grow it (i.e. where
S<FILE_CREATE_AVG == TOTAL_SUBTREE_DEL * DIR_AVG>), since the number
of nodes created in a new subtree during modification tends to be
smaller than the number in a newly-populated tree.  Hence, modify
should use the populate parameters when modifying "newly-created
subtrees" (which we can identify via link count and directory size).

=item *

Split verbosity level 3 into 3 and 4, where 3 prints multiple times
per directory but not per file, and 4 prints per file or more.

=item *

Deprecate/remove the FLOOD parameter.

=item *

Make the random generated buffer persist across files, instead of
regenerating it on every call to writeFile() and writeFragment().

=item *

Add the ability to use a text file full of option settings instead
of setting them all on the command line, i.e. ENVIRONMENT_TYPE=/tmp/foo
(and then possibly get rid of defaults.pm?).

=back


=cut


1;
