# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/DataManip/manip_log.pm#1 $

## @summary manipulation log utility routines
## @author remi@netapp.com
## @status public
## @pod here

=head1 NAME

DataManip::manip_log - manipulation log utility routines

=head1 DESCRIPTION

This module defines routines for creating and parsing a log of
filesystem manipulations.  It also defines a log entry object for
passing around log entry information in scripts that interpret such
data.

=head1 FORMAT

This log file is a series of entries, each in the following format:

C<E<lt>timeE<gt> E<lt>actionE<gt> E<lt>typeE<gt> E<lt>pathE<gt>[ E<lt>pathE<gt>]\n>

Where the identifiers above are defined as:

=over 4

=item E<lt>timeE<gt> :=

I<Day Mon DD HH:MM:SS TMZ YYYY>

(e.g. C<Wed Jul 18 19:23:00 GMT 1979>)

=item E<lt>actionE<gt> :=

C<create> | C<delete> | C<modify> | C<attrib> | C<rename> | C<hardlink>

=item E<lt>typeE<gt> :=

C<file> | C<directory> | C<stream> | C<ACL> |
C<symlink> | C<device> | C<FIFO> | C<socket>

=item E<lt>pathE<gt> :=

I<path_length path_to_object>

(e.g. C<23 F</usr/tmp/datatree/data0>>)

=back

Most actions just give the path of the object on which they acted.
In all cases, the path separator B<always> separates the name parts
of the path, even when the object in question is a stream.  This
means that a stream on a file would be presented as C<file/stream>
instead of C<file:stream> (since the latter is ambiguous on Unix
platforms where colons are allowed in file names).

Some action (such as C<rename> and C<hardlink>) also give a second
path, which is where the action originated from.  Also, some
action/type combinations do not make any sense, and should never be
issued (e.g. C<attrib ACL> or C<modify FIFO>).

Note that any entry could span multiple lines.  Directory and file
names can contain newlines (or just about any other character, except
C</>).  As such, the length field for each path specifies how many
characters should be read from the file to get the entire path, not
including the newline at the end of each entry.  Any subsequent data
should be interpreted as a separate entry.

=head1 FUNCTIONS

=cut

package DataManip::ManipEntry;
use warnings;
use strict;

my @Fields = qw(epoch action type path path2);
my @field_names = map { __PACKAGE__ . "::" . $_ } @Fields;
use fields @Fields;

use Carp;
use DataManip::cmd_utils;

=head2 DataManip::ManipEntry::new

This is the constructor for a manipulation log entry object.  The
constructor must be passed a complete, ordered list of all the entry
values, or a serialized version of a manipulation entry object.

External users would typically use the C<parse_manipulation()>
function below to build new ManipEntry objects, or might possibly pass
a serialized object to this constructor.

=cut

sub new {
  my $class = shift;
  my @vals = ();

  if (scalar(@_) == 1) {
    # create from a serialized log entry
    @vals = eval $_[0];
    if ($@) { croak $@; }

  } elsif (scalar(@_) == 4 || scalar(@_) == 5) {
    # create from parsed entry fields
    @vals = @_;

  } else {
    croak "invalid arguments to " . __PACKAGE__ . " constructor\n";
  }

  my $self = fields::phash(\@field_names, \@vals);
  return bless($self, $class);
}

=head2 DataManip::ManipEntry I<accessors>

Each manipulation log entry object has accessor methods for the entry
values C<epoch>, C<action>, C<type>, C<path>, and C<path2>.

=cut

# Create a closure for each accessor method.
foreach my $field (@Fields) {
  my $slot = __PACKAGE__ . "::$field";
  my $method = sub {
    my $self = shift;
    # Just an accessor -- no way to set.
    return $self->{$slot};
  };
  no strict 'refs';
  *$field = $method;
  use strict 'refs';
}

=pod

Since the manipulation log presents paths to streams in a non-standard
way, there are also accessor methods provide to retrieve the more
standard information, namely C<path_without_stream>,
C<path2_without_stream>, C<stream_name>, and C<stream2_name>.  If the
entry is not for a stream, then the path methods simply return the
C<path> unaltered (as the other accessor would), and the stream
methods return C<undef>.  If the entry is for a stream, then the path
methods return the path without the stream at the end, and the stream
methods return the stream name (without a leading colon or path
separator).

=cut

foreach my $field (qw(path_without_stream stream_name
		      path2_without_stream stream2_name)) {
  unless ($field =~ /\A(path|stream)(\d)?_\w+\z/) {
    croak "internal error";
  }
  my $num = ($2 ? $2 : "");
  my $return_path = ($1 eq "path");
  my $method = sub {
    my $self = shift;
    my $path = $self->{__PACKAGE__ . "::path${num}"};
    my $stream = undef;
    if ($self->{__PACKAGE__ . "::type"} eq "stream") {
      unless ($path =~ s/\/(.*)\z//s) {
	croak "invalid path $path\n";
      }
      $stream = $1;
    }
    return ($return_path ? $path : $stream);
  };
  no strict 'refs';
  *$field = $method;
  use strict 'refs';
}

=head2 DataManip::ManipEntry::serialize

Flattens the manipulation log entry object into a scalar string, which
can be passed around and stored more easily.

Later, this serialized version can be passed to a constructor to get
back the original manipulation log entry object.

=cut

sub serialize {
  my $self = shift;

  # Must use map(), not values(), in order to preserve order.
  my @vals = map { $self->{$_} } @field_names;

  my $str = DataManip::cmd_utils::serialize(@vals);
  return $str;
}

=head2 DataManip::ManipEntry::datestr

Returns the date and time of the entry as a human-readable string,
instead of the number of seconds since the epoch.

=cut

sub datestr {
  my $self = shift;

  my $slot = __PACKAGE__ . "::epoch";
  my $datestr = gmtime($self->{$slot});
  unless ($datestr =~ s/(\ \d{4})\z/ GMT$1/) {
    croak "error constructing date string";
  }
  return $datestr;
}

=head2 DataManip::ManipEntry::sprint

Converts the entire entry to a human-readable string and returns it.

=cut

sub sprint {
  my $self = shift;

  my ($epoch, $action, $type, $path, $path2) =
    map { $self->{$_} } @field_names;

  return sprintf("%s %s %s %s%s\n", $self->datestr(), $action, $type,
		  $path, ((defined $path2) ? " $path2" : ""));
}

package DataManip::manip_log;
use warnings;
use strict;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw();
our @EXPORT_OK = qw(log_manipulation parse_datestring read_path_extent
		    parse_manipulation);

use Carp;
use Time::Local;

=head2 log_manipulation($filehandle, %options)

This is the function which routines use to log filesystem
manipulations to an existing (and open) manipulation log, according to
the defined manipulation log format.

Options:

=over 4

=item TIME

The time, specified in seconds since the epoch, for this entry.
If not specified, defaults to the current local time.

=item ACTION

The action taken on the object for this entry.  This is a required
option which must match one of the legal values specified above.

=item TYPE

The object type for this entry.  This is a required option which
must match one of the legal values specified above.

=item PATH

The path to the object for this entry.  This is a required option
which must match one of the legal values specified above.

=item PATH2

The secondary path to the object from which this action originated.
This option must be specified for the actions that use it, and not
for any others.

=back

=cut

{
  my %actions = map { $_ => 1 } qw(create delete modify
				   attrib rename hardlink);
  my %types = map { $_ => 1 } qw(file directory stream ACL
				 symlink device FIFO socket);

  sub log_manipulation( $% ) {
    my ($log_fh, %opts) = @_;

    # Default the time to the current local time.
    $opts{TIME} = time() unless (defined $opts{TIME});

    # Make sure the specified options are legal.
    if (exists $opts{ACTION} && defined $opts{ACTION}) {
      unless ($actions{$opts{ACTION}}) {
	croak "invalid ACTION: " . $opts{ACTION} . "\n";
      }
    } else {
      croak "ACTION must be specified\n";
    }
    if (exists $opts{TYPE} && defined $opts{TYPE}) {
      unless ($types{$opts{TYPE}}) {
	croak "invalid TYPE: " . $opts{TYPE} . "\n";
      }
    } else {
      croak "TYPE must be specified\n";
    }
    unless (exists $opts{PATH} && defined $opts{PATH}) {
      croak "PATH must be specified\n";
    }
    if ($opts{ACTION} eq "rename" || $opts{ACTION} eq "hardlink") {
      unless (exists $opts{PATH2} && defined $opts{PATH2}) {
	croak "PATH2 must be specified\n";
      }
    } else {
      if (exists $opts{PATH2}) {
	croak "PATH2 must not be specified\n";
      }
    }

    # Make sure that logged relative paths start with "./".
    # This probably isn't necessary, and could be removed.
    foreach my $key (qw(PATH PATH2)) {
      next unless (exists $opts{$key});
      next if (File::Spec->file_name_is_absolute($opts{$key}));
      my $dot = File::Spec->curdir();
      my $sep = File::Spec->catpath('', $dot, $dot);
      unless ($sep =~ s/\A\Q$dot\E(.+?)\Q$dot\E\z/$1/) {
	croak "error prepending curdir()\n";
      }
      unless ($opts{$key} =~ /\A\Q$dot$sep\E/) {
	$opts{$key} = $dot . $sep . $opts{$key};
      }
    }

    # Construct the entry.
    # XXX For the moment, always use GMT.
    # XXX Currently, no checking to make sure the specified time is numeric.
    my $dt = gmtime($opts{TIME});
    unless ($dt =~ s/(\d{4})\z/GMT $1/) {
      croak "error formatting time string: $dt\n";
    }
    my $entry = sprintf("%s %s %s %d %s", $dt, $opts{ACTION}, $opts{TYPE},
			length($opts{PATH}), $opts{PATH});
    if (exists $opts{PATH2}) {
      $entry .= sprintf(" %d %s", length($opts{PATH2}), $opts{PATH2});
    }
    $entry .= "\n";

    print $log_fh $entry or croak "$!\n";
  }
}

=head2 parse_datestring($datestr)

Parses a date string of the form C<Day Mon DT HH:MM:SS TMZ YYYY>
and returns the date in terms of epoch time.
Croaks if the provided date string is invalid.

Some timezones are not recognized by the conversion functions,
and so they may not be converted correctly.

=cut

{
  my $remap = sub { my $cnt = 0; return map { $_ => $cnt++ } @_; };
  my %days = $remap->( qw(Sun Mon Tue Wed Thu Fri Sat) );
  my %months = $remap->( qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec) );

  sub parse_datestring( $ ) {
    my ($datestr) = @_;
    unless ($datestr =~ /\A
	    ([A-Za-z]{3})\            # Day
	    ([A-Za-z]{3})\            # Month
	    ([ \d]\d)\                # Date
	    ([ \d]\d):(\d\d):(\d\d)\  # HH:MM:SS
	    ([A-Za-z]{3})\            # Time zone
	    (\d{4})                   # Year
	    \z/x) {
      croak "invalid date string\n"
    }
    my ($day, $month, $date, $hh, $mm, $ss, $tmz, $year) =
      ($1, $2, $3, $4, $5, $6, $7, $8);

    croak "invalid day" unless (exists $days{$day});
    croak "invalid month" unless (exists $months{$month});
    my $mon = $months{$month};

    # Save the arguments to timelocal().
    my @args = ($ss, $mm, $hh, $date, $mon, $year);

    # Convert the time in the proper time zone.
    my $save_tz = $ENV{TZ};
    $ENV{TZ} = $tmz;
    my $epoch = timelocal(@args);

    if (defined $save_tz) {
      $ENV{TZ} = $save_tz;
    } else {
      delete $ENV{TZ};
    }

    # XXX Try to hack around unknown time zones.
    if (($tmz ne 'GMT') && ($tmz ne 'UTC')) {
      my $univ = timegm(@args);
      if ($epoch eq $univ) {
	# Crap... the timezone didn't work, and just gave us GMT.
	# Spew out a warning, and just try using the local timezone.
	carp "timezone $tmz not recognized, using local time\n";
	$epoch = timelocal(@args);
      }
    }

    return $epoch;
  }
}

=head2 read_path_extent($buffer, $filehandle, $length)

Reads a path string of exactly $length bytes from the provided
path buffer and filehandle, reading as many lines as necessary
from the filehandle into the path buffer first.

Removes the path from the buffer and returns it on success.
Croaks on any error.

=cut

sub read_path_extent( $$$ ) {
  my ($entry, $fh, $length) = @_;

  # Append more lines to the entry until it's longer than expected.
  while (length($_[0]) <= $length) {
    my $nextline = <$fh>;
    if ($nextline) {
      $_[0] .= $nextline;
    } else {
      croak "end of file during entry path: $entry\n";
    }
  }

  # Removed the first $length characters of the entry and return it.
  my $path = substr($_[0], 0, $length, "");
  unless (length($path) == $length) { # paranoia
    croak "path length $length mismatch: $path\n";
  }
  return $path;
}

=head2 parse_manipulation($filehandle)

Given an open filehandle for a manipulation log, reads off exactly one
entry and parses it, taking into account the fact that paths can
contain newlines.  However, since every manipulation log entry ends in
a newline (according to the format above), this should read and parse
some integer number of lines from the filehandle, with no extra
characters to spare (i.e. nothing should be ignored).

Returns a false value if EOF is hit while reading from the filehandle.
Otherwise, returns a new ManipEntry object, initialized from the
parsed manipulation log entry.  Croaks on any error.

=cut

sub parse_manipulation( $ ) {
  my ($fh) = @_;

  # Get the next line, but don't immediately chomp it,
  # since the entry path could span lines.
  my $entry = <$fh>;
  unless ($entry) {
    # Get out when we hit EOF.
    return $entry;
  }

  # Strip off and parse the date.
  unless ($entry =~ s/\A(.{28})//) {
    croak "invalid entry date: \"$entry\"\n";
  }
  my $epoch = parse_datestring($1);

  # Grab the action and type.
  unless ($entry =~ s/\A\   # a space
	  ([a-z]+?)\        # action
	  ([a-z]+?)\s+      # type
	  (\d+)\s+          # first path length
	  //x) {
    croak "invalid entry: $entry\n";
  }
  my ($action, $type, $len1) = ($1, $2, $3);

  # Get the path, reading extra lines as necessary.
  # Make sure that the path starts with a './' and is the right length.
  my $path = read_path_extent($entry, $fh, $len1);
  unless ($path =~ /\A\.\//) {
    croak "invalid entry path: $path\n";
  }

  # Read the second path, if the action gives one.
  my $path2 = undef;
  if ($action eq "rename" || $action eq "hardlink") {
    unless ($entry =~ s/\A (\d+)\s+//) {
      croak "invalid second path length: $entry\n";
    }
    my $len2 = $1;
    $path2 = read_path_extent($entry, $fh, $len2);
    unless ($path2 =~ /\A\.\//) {
      croak "invalid entry path: $path\n";
    }
  }

  # Now we can chomp the entry, after which there shouldn't be anything left.
  chomp $entry;
  if (length($entry) > 0) {
    croak "extra cruft at end of entry: $entry\n";
  }

  return new DataManip::ManipEntry ($epoch, $action, $type, $path, $path2);
}

=head1 BUGS

C<log_manipulation()> logs all entries in the GMT timezone, rather
than the local time zone.  Similarly, C<ManipEntry::datestr()> always
returns a string in the GMT timezone.

=cut


1;
