# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/DataStore.pm#1 $

## @summary     Package used to read/write using NATE::IPC.  Assumes functions
## @summary     are being called from subtest.  Provides handshaking layer
## @summary     between parent and subtest for the write_data function.
## @author      dl-afd@netapp.com
## @status      private
## @pod here

package DataStore;

# Compiler directives.
use strict;

#
# This package can be used to pass data between a subtest and the
# parent process using the NATE::IPC mechanism.
#

sub get_data {
    my(
       $dataref,
       $ipc_sock,
       );

    # For now, we are assuming that the thing calling us is a subtest.
    # May want to optimize this in the future so it can be called
    # From non-subtested callers, or make the determination and do
    # the right thing based on whether or not the caller is a subtest.
    $ipc_sock = Subtest::ipc;
    $dataref = $ipc_sock->read_obj;

    return $dataref;
} # end sub get_data


sub write_data {
    my($dataref) = @_;
    my(
       $ipc_sock,
       $msg,
       $rc,
       );

    # Protocol is:
    #   - send data to parent.
    #   - wait for response from parent that the data was received
    #     before continuing.  This prevents a subtest from exiting
    #     before the parent has had a chance to read the message.

    # Refer to get_data comments on assuming this is a subtest.
    $ipc_sock = Subtest::ipc;
    $ipc_sock->write_obj($dataref);

    # Wait for data receipt confirmation from parent.
    $msg = $ipc_sock->read_message;
    $rc = ($msg =~ /^data received/) ? 1 : 0;

    return $rc;
} # end sub write_data

1;

__END__

#################################
#  Pod
#################################

=pod

=head1 NAME

DataStore

=head1 SYNOPSIS

 use DataStore;

 # Get data.
 my $dataref = &DataStore::get_data;

 # Update data.
 $dataref->{newdata} = 'blah';

 # Send data back to parent.
 $rc = &DataStore::write_data($dataref);
 if ($rc) {
     # Data written successfully.
 }
 else {
     # Failure in writing data back to parent.
 }

=head1 DESCRIPTION

DataStore provides a means of retrieving data from and sending modified
data back to a parent process using NATE::IPC. It is a thinlayer around the
NATE::IPC read_object and write_object methods.  It assumes that it is
being called in a subtest context (i.e. uses the Subtest::IPC method to gain
access to the subtest end of an IPC socket).

It provides a handshaking mechanism on the write_data function call to
prevent the subtest from exiting before the parent has the opportunity
to retrieve the modified data.

=head1 FUNCTIONS

=head2 read_data

=over

=item Synopsis

 $dataref = &DataStore::read_data;

=item Arguments

None.

=item Returns

Data reference (could be an object).

=back

=head2 write_data

=over

=item Synopsis

 $rc = &DataStore::write_data(I<$dataref>);
 if ($rc) {
     # successful write - parent acknowledged transfer.
 }
 else {
     # data write did not complete successfully.
 }

=item Arguments

=over

=item $dataref

Data reference or object.  Most likely an updated copy of what was retrieved
using the I<DataStore::read_data> call.

=back

=item Returns

1 if write to parent completed successfully (i.e. write was accepted and
acknowledged by parent), 0 otherwise.

=back

=head1 AUTHOR/MAINTAINER

Kevin Lambright (kevinl@netapp.com)

=cut
