# Perl Module
# * $Id: //depot/prod/test/Rsierranevada/lib/DataStructureStore.pm#1 $ *
# vim: ft=perl
#
# Purpose    Provide common data structure storage routines 
# 
## @summary     Provide common data structure storage routines
##
## @description Provide one common interface for variouse ways
##              to store Perl Data Structures to a file for 
##              QA thpls. Storage is separate from log files
##              but can be stored to same directory location.
##              
##
## @author      parsons@netapp.com
##
## @pod here
# POD at end of file

package DataStructureStore;

#use lib ('/x/eng/kc-users/parsons/p4/test_dev/lib');

use strict;
use Data::Dumper;
use Storable qw(nstore retrieve);
use File::Basename;

use TracePkg;
use Options;
use Params::Common;


BEGIN {
   use Exporter();
   our @ISA           = qw(Exporter);
}


our @DataDumper_tags = qw(
   dss_print
   set_data_dumper_var
);

our @Storable_tags = qw(

);

our @EXPORT  = (
   @DataDumper_tags,
   @Storable_tags
);

our @EXPORT_OK  = qw();
our %EXPORT_TAGS = (
   'datadump'   => [@DataDumper_tags],
   'storable'   => [@Storable_tags],
);

our %MODE2EXT = (
   'dd'    => 'dd',
   'sto'   => 'sto',
   'aoa'   => 'csv',
   'aoh'   => 'csv',
);



my $Trace   = new TracePkg;
my $Options = new Options;
$Options->parameters(definitions=> [
      { 'name'        => 'DSS_MODE',
        'description' => 'data storage mode',
        'default'     => 'dd',
      }, 

      { 'name'        => 'DSS_DIR',
        'description' => 'directory to store data structure',
        'file_tests'  => 'derw'
      },

      { 'name'        => 'DSS_FILE',
        'description' => 'file to store data structure, relative to DSS_DIR ',
      },
      { 'name'        => 'DSS_FILENAME',
        'description' => 'absolute file to store data structure'
      },
   ]
);
$Options->process();
my $options = $Options->get_hash();





#=============================================================================
#
# DataStructureStore Constructor - object holds the info for storing files
# 
# Args
#   'mode'     => <dd|sto>     storage mode
#   'dir'      => <str>        directory
#   'file'     => <str>        relative filename to 'dir'
#   'ext'      => <str>        extension
#   'filename' => <str>        full filename

sub new {

   $Trace->entry( 'msg' => 'Constructor entry');
   my $_proto = shift;
   my $_class = ref($_proto) || $_proto;
   my $_self  = {};
   # bless object
   bless($_self, $_class);

   # get object arguments
   my %_arg = @_;

   my ($_a, 
      $_file, $_dir, $_ext,
   );

   # load args to object
   foreach $_a (keys %_arg) {
      if    ($_a =~ /^-?mode$/i)    { $_self->{'dss_mode'}     = delete($_arg{$_a}); }
      elsif ($_a =~ /^-?dir$/i)     { $_self->{'dss_dir'}      = delete($_arg{$_a}); }
      elsif ($_a =~ /^-?file$/i)    { $_self->{'dss_file'}     = delete($_arg{$_a}); }
      elsif ($_a =~ /^-?ext$/i)     { $_self->{'dss_ext'}      = delete($_arg{$_a}); }
      elsif ($_a =~ /^-?filename$/i){ $_self->{'dss_filename'} = delete($_arg{$_a}); }     
      else {
         $Trace->fatal( 'msg'     => "unknown DataStructureStore argument [$_a]",
                        'result'  => 'CONFIG',
         );
      }
   }

   # Error check and assignment
   #
   # Mode
   #
   # function call has precedence
   $Trace->comment( 'msg' => 'Constructor Check Mode');
   if ($_self->{'dss_mode'}) {
      # dss_mode given as arg
      unless ( dss_check_mode($_self->{'dss_mode'}) ) {
         $Trace->fatal( 'msg'     => "unsupported data storage mode arg: $_self->{'dss_mode'}",
                        'result'  => 'CONFIG',
         );
      }
      $Trace->debug( 'msg' => "mode defined by arg: $_self->{'dss_mode'}");
   } elsif ( defined($options->{'DSS_MODE'}) )  {
      # check param mode, set object
      if ( dss_check_mode($options->{'DSS_MODE'}) ) {
         $_self->{'dss_mode'} = $options->{'DSS_MODE'};
      } else {
         $Trace->fatal( 'msg'     => "unsupported data storage mode param: $options->{'DSS_MODE'}",
                        'result'  => 'CONFIG',
         );
      }
      $Trace->debug( 'msg' => "mode defined by param $options->{'DSS_MODE'}");
   } else {
      # default it, although it already was by params definition
      $_self->{'dss_mode'} = 'dd';
      $Trace->debug( 'msg' => "defaulting mode to $_self->{'dss_mode'}");
   }


   #
   # Filename check
   # 'filename' holds the absolute filename
   # if 'filename is given, populate other object vars
   # if 'filename' is not given make it from dir,file,ext
   $Trace->comment( 'msg' => 'Check filename');
   if ( $_self->{'dss_filename'} ) {
      if ( dss_check_filename($_self->{'dss_filename'}) ) {
         ($_self->{'dss_file'}, $_self->{'dss_dir'}, $_self->{'dss_ext'}) = fileparse($_self->{'dss_filename'}, qr/\.[^.]*/);
         $_self->{'dss_ext'} =~ s/^\.//;
      }
      $Trace->debug( 'msg' => "filename defined by arg: $_self->{'dss_filename'}");
      $Trace->debug( 'msg' => "set filename parts: [$_self->{'dss_dir'}] [$_self->{'dss_file'}] [$_self->{'dss_ext'}]");
   } elsif ( $options->{'DSS_FILENAME'} ) {
      if ( dss_check_filename($options->{'DSS_FILENAME'}) ) {
         $_self->{'dss_filename'} = $options->{'DSS_FILENAME'};
         ($_self->{'dss_file'}, $_self->{'dss_dir'}, $_self->{'dss_ext'}) = fileparse($options->{'DSS_FILENAME'}, qr/\.[^.]*/);
         $_self->{'dss_ext'} =~ s/^\.//;
      }
      $Trace->debug( 'msg' => "filename defined by param: $_self->{'dss_filename'}");
      $Trace->debug( 'msg' => "set filename parts: [$_self->{'dss_dir'}] [$_self->{'dss_file'}] [$_self->{'dss_ext'}]");   

   } else {
      # make the filename
      $Trace->debug( 'msg' => 'make a filename');
      dss_make_filename($_self);
   } 



   $Trace->exit('msg' =>"Constructor exit");
   return $_self;
}
#
#---------------------------------------------------------------------------
#
# change object dss_dir 
# Args
#   $_[0] = ref to obj
#   'mode'     => <dd|sto>     storage mode
#   'dir'      => <str>        directory
#   'file'     => <str>        relative filename to 'dir'
#   'ext'      => <str>        extension
#   'filename' => <str>        full filename
#
sub dss_update_file {

   $Trace->entry('msg' =>"dss_update_file entry ");
   my $_ref = shift;
   my %_arg = @_;
   my $_a;
   foreach $_a (keys %_arg) {
      $Trace->debug('msg' => "update object: $_a with $_arg{$_a}");
      if    ($_a =~ /^-?mode$/i)    { $_ref->{'dss_mode'}     = delete($_arg{$_a}); }
      elsif ($_a =~ /^-?dir$/i)     { $_ref->{'dss_dir'}      = delete($_arg{$_a}); }
      elsif ($_a =~ /^-?file$/i)    { $_ref->{'dss_file'}     = delete($_arg{$_a}); }
      elsif ($_a =~ /^-?ext$/i)     { $_ref->{'dss_ext'}      = delete($_arg{$_a}); }
      elsif ($_a =~ /^-?filename$/i) { 
         $_ref->{'dss_filename'} = delete($_arg{$_a}); 
         ($_ref->{'dss_file'}, $_ref->{'dss_dir'}, $_ref->{'dss_ext'}) = 
                                     fileparse($_ref->{'dss_filename'},  qr/\.[^.]*/);
         $_ref->{'dss_ext'} =~ s/^\.//;
         if ( $_ref->{'dss_ext'} =~ /^(sto|dd)$/ ) {
             $_ref->{'dss_mode'} = $_ref->{'dss_ext'};
         }
         $Trace->comment( 'msg' => "update filename parts: [$_ref->{'dss_dir'}] " . 
                                   "$_ref->{'dss_file'}] [$_ref->{'dss_ext'}]"
         );
      } else {
         $Trace->fatal( 'msg'     => "unknown dss_update_file argument [$_a]",
                        'result'  => 'CONFIG',
         );
      }
   }
   
   if ( !defined($MODE2EXT{$_ref->{'dss_mode'}}) ) {
       $Trace->fatal( 'msg' => "updated mode with unsupported mode: $_ref->{'dss_mode'}",
                      result => 'CONFIG',
       );

   }

   dss_make_filename($_ref);
   $Trace->exit('msg' =>"dss_update_file exit ");
   1;
}
#
#-----------------------------------------------------------------------------
# return filename and and mode info
#
# Args 
#  $_[0] = ref to obj
# Return
#   wantarray ? (filename, mode) : filename
# 
sub dss_filename {
   wantarray ? ( $_[0]->{'dss_filename'}, $_[0]->{'dss_mode'}) : $_[0]->{'dss_filename'};
}

#
#-----------------------------------------------------------------------------
#
# Store given data strcture to objects filenname
#
# Args
#   $_[0] = ref to obj
#   $_[1] = ref to data structure
sub dss_store {

   $Trace->entry('msg' =>"dss_store ");
   my ($_fh, $_a1, $_a2, $_c1, $_c2, %_keys, @_array);
   my $_no_store = 0;  # smoetime used as a flag

   # Data Dumper
   if ($_[0]->{'dss_mode'} eq 'dd') {
      $Trace->comment( 'msg' => "writing data structure with Data::Dumper to $_[0]->{'dss_filename'}");
      open($_fh, ">$_[0]->{'dss_filename'}") || $Trace->fatal( 'msg'    => "dss_store open failed: $!",
                                                              'result' => 'SCRIPT'
      );
      print $_fh Dumper($_[1]);
      close($_fh);
   }

   # Storable
   if ($_[0]->{'dss_mode'} eq 'sto') {
      $Trace->comment( 'msg' => "nstoring data structure with Storable to $_[0]->{'dss_filename'}");
      nstore($_[1], $_[0]->{'dss_filename'}) || $Trace->fatal( 'msg'    => "dss_store nstore failed: $!",
                                                              'result' => 'SCRIPT'
      );
   }

   # AoA to csv
   if ($_[0]->{'dss_mode'} eq 'aoa') {
      # check we have AoA
      $Trace->comment('msg' => 'Verifying AoA structure');
      $_c1 = 0;   # row index
      foreach $_a1 ( @{$_[1]} ) {
          # each 1st level element should be an array ref
          if ( ref($_a1) ne 'ARRAY' ) {
             $Trace->warn( 'msg' => sprintf("AoA row: %u  not an ARRAY [%s]", $_c1, ref($_a1)) );
             $_no_store++;
          }
          # each 2nd level element should not be a reference, thus a value
          $_c2 = 0; # column index
          foreach $_a2 ( @{$_a1} ) {
             if ( ref($_a2) ) {
                $Trace->warn('msg' => sprintf("AoA row: %u col: %u is a ref: %s",  $_c1, $_c2, ref($_a2)));
                $_no_store++;
             } 
             $_c2++;  # incr col index
          }
          $_c1++;  # incr row index
      }
      # check if it should be dumped to csv file
      if ($_no_store) {
         $Trace->warn('msg' => 'AoA not stored due to invalid reference type, see warnings');
      } else {
         # its good, dump to csv file
         $Trace->comment('msg' => "AoA printing to file $_[0]->{'dss_filename'}");
         open($_fh, ">$_[0]->{'dss_filename'}") || $Trace->fatal( 'msg'    => "dss_store open failed: $!",
                                                                 'result' => 'SCRIPT'
         );
         foreach $_a1 ( @{$_[1]} ) { printf $_fh ("%s\n", join(',', @{$_a1}) ); }
         close($_fh);
      } 
   }

   # AoH to csv
   # use the key values in the hash as a column header
   if ($_[0]->{'dss_mode'} eq 'aoh') {
      # check we have AoH
      $Trace->comment('msg' => 'Verifying AoH structure');
      $_c1 = 0;   # row index
      foreach $_a1 ( @{$_[1]} ) {
          # each 1st level element should be a hash ref
          #$Trace->debug('msg' => sprintf("row %u  ref: %s",$_c1, ref($_a1) ));
          if ( ref($_a1) ne 'HASH' ) {
             $Trace->warn( 'msg' => sprintf("AoA row: %u  not a HASH [%s]", $_c1, ref($_a1)) );
             $_no_store++;
          }
          # each 2nd level element should be a key
          foreach $_a2 ( keys %{$_a1} ) {
             if ( ref($_a2) ) {
                $Trace->warn('msg' => sprintf("AoA row: %u has a ref as a key: %s",  $_c1, ref($_a2)));
                $_no_store++;
             }
             $_keys{$_a2}++;    # cache which keys are used
          }
          $_c1++;  # incr row index
      }
      # check if it should be dumped to csv file
      if ($_no_store) {
         $Trace->warn('msg' => 'AoH not stored due to invalid reference type, see warnings');
      } else {
         # its good, dump to csv file
         # use cached keys as a 1st line for column header
         open($_fh, ">$_[0]->{'dss_filename'}") || $Trace->fatal( 'msg'    => "dss_store open failed: $!",
                                                                 'result' => 'SCRIPT'
         );
         # make a consistent list of keys
         @_array = sort keys %_keys;

         # csv header line
         printf $_fh ("%s\n", join(',', @_array) );
         # convert key/value pairs to postional var/values on csv line
         # cycle thru all array elements and each possible keey
         #   if the hash has the current key, dump content, otherwise nothing
         foreach $_a1 ( @{$_[1]} ) {
            foreach $_a2 (@_array) {
               printf $_fh ("%s,",  exists($_a1->{$_a2}) ? $_a1->{$_a2} : undef);
            }
            printf $_fh ("\n");
         }
         close($_fh); 
      } 

   }
   $Trace->exit('msg' =>"dss_store ");
   1;
}
#
#-----------------------------------------------------------------------------
# 
# Load from the file in object, its content into a data structure
#
# Args
#   $_[0] = ref to obj
#   $_[1] = ref to data structure to load to
#
sub dss_load {
   $Trace->entry('msg' =>"dss_load");
   my ($_fh, $_data, $_ref, $_line, @_tokens, %_keys,
       $_c1, $_c2, $_t,
   );
   my $_eval;

   # Auto-assign mode based on file extension
   # aoa and aoh both use csv, so we can't infer format from that extension
   if ( $_[0]->{'dss_filename'} =~ /\.sto$/ ) { 
       if ( $_[0]->{'dss_mode'} ne 'sto' ) {
           $Trace->warn( msg => "The dss_mode specified is ".
                                $_[0]->{'dss_mode'}.", however file extension is \".sto\"  Setting dss_mode=\"sto\"");
           $_[0]->{'dss_mode'} = 'sto'; 
       }
   }

   if ($_[0]->{'dss_filename'} =~  /\.dd$/ ) { 
       if ( $_[0]->{'dss_mode'} ne 'dd' ) {
           $Trace->warn( msg => "The dss_mode specified is ".
                                $_[0]->{'dss_mode'}.", however file extension is \".dd\"  Setting dss_mode=\"dd\"");
           $_[0]->{'dss_mode'} = 'dd'; 
       }
   }

   # Data Dumper
   if ($_[0]->{'dss_mode'} eq 'dd') {
      $Trace->comment( 'msg' => "read and eval data structure from $_[0]->{'dss_filename'}");
      open($_fh, "< $_[0]->{'dss_filename'}") || $Trace->fatal( 'msg'    => "dss_load open failed: $!",
                                                              'result' => 'SCRIPT'
      );
      # slurp the data::dumper file
      {   no strict 'vars';   # account for $VAR1 not declared 
          local $/;
          $_data = <$_fh>;
          $_[1] = eval $_data;
      } 
      close($_fh);

      if ( $@ ){
         $Trace->warn( 'msg' => "eval data structured failed: $@");
      }  else {
         $Trace->debug( 'msg' => sprintf("evaled data string length: %u\n", length($_data))); 
         $Trace->debug( 'msg' => sprintf("evaled data string ref: %s", ref($_[1]) || '?ref?') );
      }
   }

   # Storable
   if ($_[0]->{'dss_mode'} eq 'sto') {
      $Trace->comment( 'msg' => "retrieve data structure with Storable from $_[0]->{'dss_filename'}");
      $_[1] = retrieve($_[0]->{'dss_filename'})
   }

   # AoA  (csv2AoA)
   if ($_[0]->{'dss_mode'} eq 'aoa') {
      $Trace->comment( 'msg' => "read AoA data structure from $_[0]->{'dss_filename'}");
      open($_fh, "< $_[0]->{'dss_filename'}") || $Trace->fatal( 'msg'    => "dss_load open failed: $!",
                                                              'result' => 'SCRIPT'
      );
      $_c1 = 0;
      foreach $_line ( <$_fh> ) {
          next if $_line =~ /^#/;
          chomp($_line);
          @_tokens = split(/,/, $_line);
          $_c2 = 0;
          foreach $_data (@_tokens) {
             $_[1]->[$_c1][$_c2] = $_data;
             #$Trace->comment( 'msg' => "csv2AoA row: $_c1 col: $_c2  [$_data]");
             $_c2++;
          }
          $_c1++;
          $Trace->debug( 'msg' => "csv2AoA added line $_c1 cols: $_c2"); 
      }
      close($_fh);
   }
   # AoH (csv2AoH)
   # assumes first line of csv is header line, meaning use it for hash keys
   if ($_[0]->{'dss_mode'} eq 'aoh') {
      $Trace->comment( 'msg' => "read AoH data structure from $_[0]->{'dss_filename'}");
      open($_fh, "< $_[0]->{'dss_filename'}") || $Trace->fatal( 'msg'    => "dss_load open failed: $!",
                                                               'result' => 'SCRIPT'
      );
      # get 1st line for hash key strings
      # index the keys per array element index
      $_c2 = 0;
      $_line = readline($_fh);
      chomp($_line);
      foreach $_t ( split(/,/, $_line) ) { $_keys{$_c2} = $_t; $_c2++; } 
      # loop thru the remaining lines, split out values
      # match key index to token index to so as to get key for AoH
      $_c1 = 0;
      foreach $_line ( <$_fh> ) {
         chomp($_line);
         @_tokens = split(/,/, $_line);
         $_c2 = 0;
         foreach $_t (@_tokens) {
            # if the token has something, store it in AoH using the index value to lookup hash key
            if ( $_t =~ /./) {
               $_[1]->[$_c1]{$_keys{$_c2}} = $_t; 
            }
            $_c2++;
         }
         $_c1++; 
         $Trace->debug( 'msg' => "csv2AoH added line $_c1 hash\n");
         #$Trace->trace( 'msg' => Dumper($_[1]->[$_c1 -1]));
      }      
      close($_fh);
   }

   $Trace->exit('msg' =>"dss_load");
   1;
}

#
#-----------------------------------------------------------------------------
#
# data dumper short cut
# $_[0] = ref to data structure
# $_[1] = [[type] [message]]
#
#        the first arg is required, its the ref to the data structure to dump
#        next two are optional
#        if second arg is a TrackPkg method name, its treeated as such
#        else its treated as a message string
#        if second arg is not a TrackPkg method name, its treeated as a message string
#
sub dss_print {

   my $_type = 'comment';
   my $_msg  = 'User Supplied Data Structure';

   # check optional args
   if ( $_[1] ) {
       # see if TrackPkg method name
       if ($_[1] =~ /^comment|trace|debug|warn$/) { 
          $_type = $_[1];
          # see if a 3rd arg defined, its the message string
          if ($_[2]) { $_msg = $_[2]; } 
       }  else   {
          # got a second arg thats not a TrackPkg method name, so its a message
          $_msg  = $_[1]; 
       } 
   }

   $Trace->${_type}( 'msg' => "DataDump: $_msg\n" . Dumper $_[0]);
   1;
}
#
#-----------------------------------------------------------------------------
#
# Provide access to Data Dumper Configuration variables
# 
# Args
#   $_[0] = varname
#   $_[1] = value
# Return 
#  0 - unsupported varname
#  1 - set

sub set_data_dumper_var {

   my $_set = 1;
   $Trace->entry( 'msg' => "set_data_dumper_var entry set $_[0] = $_[1]");


   if    ($_[0] eq 'Indent')    { $Data::Dumper::Indent    =  $_[1]; }
   elsif ($_[0] eq 'Purity')    { $Data::Dumper::Purity    =  $_[1]; }
   elsif ($_[0] eq 'Pad')       { $Data::Dumper::Pad       =  $_[1]; }
   elsif ($_[0] eq 'Varname')   { $Data::Dumper::Varname   =  $_[1]; }
   elsif ($_[0] eq 'Useqq')     { $Data::Dumper::Useqq     =  $_[1]; }
   elsif ($_[0] eq 'Terse')     { $Data::Dumper::Terse     =  $_[1]; }
   elsif ($_[0] eq 'Freezer')   { $Data::Dumper::Freezer   =  $_[1]; }
   elsif ($_[0] eq 'Toaster')   { $Data::Dumper::Toaster   =  $_[1]; }
   elsif ($_[0] eq 'Deepcopy')  { $Data::Dumper::Deepcopy  =  $_[1]; }
   elsif ($_[0] eq 'Quotekeys') { $Data::Dumper::Quotekeys =  $_[1]; }
   elsif ($_[0] eq 'Bless')     { $Data::Dumper::Bless     =  $_[1]; }
   elsif ($_[0] eq 'Maxdepth')  { $Data::Dumper::Maxdepth  =  $_[1]; }
   elsif ($_[0] eq 'Pair')      { $Data::Dumper::Pair      =  $_[1]; }
   elsif ($_[0] eq 'Useperl')   { $Data::Dumper::Useperl   =  $_[1]; }
   elsif ($_[0] eq 'Sortkeys')  { $Data::Dumper::Sortkeys  =  $_[1]; }
   elsif ($_[0] eq 'Deparse')   { $Data::Dumper::Deparse   =  $_[1]; }
   else {
      $Trace->warn( 'msg' => "Unknown Data::Dumper var: $_[0]"); 
      $_set = 0;
   }

   $Trace->exit( 'msg' => 'set_data_dumper_var exit');
   $_set;
}





#
#-----------------------------------------------------------------------------
#
# Make a filename from the vars in the object
# Arg
#  $_[0] - ref to object
# Return 
#  nothing meaningful  since working on the object
#
sub dss_make_filename {

   $Trace->entry( 'msg' => 'dss_make_filename entry');

   my $_delimiter;

   # make sure we have a dir
   if ( !defined($_[0]->{'dss_dir'}) ) {
      $_[0]->{'dss_dir'} = $options->{'DSS_DIR'} || $ENV{'LOGDIR'} || $ENV{'PWD'}; 
      $Trace->debug( 'msg' => "set object dss_dir: $_[0]->{'dss_dir'}");
   }
   # make sure we have a file
   if ( !defined($_[0]->{'dss_file'}) ) {
      $_[0]->{'dss_file'} = $options->{'DSS_FILE'} || 'data_struct';
      $Trace->debug( 'msg' => "set object dss_file: $_[0]->{'dss_file'}");
   } 
   # set up the extension
   if ( !defined($_[0]->{'dss_ext'}) ) {
         $_[0]->{'dss_ext'} = $MODE2EXT{$_[0]->{'dss_mode'}};
         $Trace->debug( 'msg' => "set ext based on mode[$_[0]->{'dss_mode'}]: $_[0]->{'dss_ext'} ");
   } else {
      if ( $_[0]->{'dss_ext'} ne $MODE2EXT{$_[0]->{'dss_mode'}} ) {
         $_[0]->{'dss_ext'} = $MODE2EXT{$_[0]->{'dss_mode'}};
         $Trace->debug( 'msg' => "change ext based on mode[$_[0]->{'dss_mode'}]: $_[0]->{'dss_ext'} ");         
      }
   }
    
   # make sure dir does not end with a delimter
   if ($^O =~ /win/i) { $_[0]->{'dss_dir'} =~ s/\\$//; $_delimiter = '\\';}
   else               { $_[0]->{'dss_dir'} =~ s/\/$//; $_delimiter = '/'; }
   $Trace->debug( 'msg' => "$^O delimiter set: $_delimiter");

     
   if ($_[0]->{'dss_ext'}) {
      $_[0]->{'dss_filename'} = sprintf("%s%s%s.%s",  
         $_[0]->{'dss_dir'}, $_delimiter, 
         $_[0]->{'dss_file'}, $_[0]->{'dss_ext'},
      );
   } else {
      $_[0]->{'dss_filename'} = sprintf("%s%s%s",  
         $_[0]->{'dss_dir'}, $_delimiter,
         $_[0]->{'dss_file'},
      );
   }
   $Trace->comment( 'msg' => "made filename: $_[0]->{'dss_filename'}");

   dss_check_filename($_[0]->{'dss_filename'});

   $Trace->exit( 'msg' => 'dss_make_filename exit');
   1;
}

#
#-----------------------------------------------------------------------------
#
# Check the filename for existence,dir,writable
# Args
#   $_[0] = filename
# Return
#  1
# 
sub dss_check_filename {

   $Trace->entry( 'msg' => 'dss_check_filename entry');

   my ($_n,$_p,$_e) = fileparse($_[0], qr/\.[^.]*/);

   # path not exist
   if (!-e $_p) {
      $Trace->fatal( 'msg'     => "filename path not exist: $_p",
                     'result'  => 'CONFIG',
      ); 
   }
   # path dot a dir
   if (!-d $_p) {
      $Trace->fatal( 'msg'     => "filename path not a directory: $_p",
                     'result'  => 'CONFIG',
      );
   }
   # path not writable
   if (!-w $_p) {
      $Trace->fatal( 'msg'     => "filename path not writable: $_p",
                     'result'  => 'CONFIG',
      );
   }
   # if filename exist, check if writable
   if ( -e $_[0] ) {
      if (!-w $_[0] ) {
         $Trace->fatal( 'msg'     => "filename exist, not writable: $_[0]",
                     'result'  => 'CONFIG',
         );
      }
   }
   $Trace->exit( 'msg' => 'dss_check_filename exit');
   1;
}
 
#
#-----------------------------------------------------------------------------
#
# Check the storage mode setting, if defined in %MODE2EXT return 1
# Args
#   $_[0] = storage mode string
# Return
#  0 | 1
#
sub dss_check_mode {
   defined($MODE2EXT{$_[0]}) ? 1 : 0;
}


1;
__END__


=pod

=head1 NAME

DataStructureStore - common interface for Perl data structure storage

=head1 DESCRIPTION

Perl module to provide a common set of tools to store a Perl Data Structure
to a file. Some users may want to store a Perl Data Structure to a file
other than a log file, either for re-use or further examination. Perl offers
a variety of ways to do this and this module was written as an attempt to
consolidate the needed functionality. 

A common practice in NetApp QA is to use Perl's Data::Dumper to dump
the given data structure reference to a file or STDOUT. Data::Dumper has
its pros and cons, so other methods are provided with this module as well as
Data::Dumper. The functionlity currently provided by this module:

=over

=item Perl's Data::Dumper 

dump (write) or load (read) a Perl Data Structure to/from a plain text file.

=item Perl's Storable

nstore (write) or retrieve (read) a Perl Data Structure to/from a file (binary).
Read in file loads directly as data structure.

=item AoA to csv 

Write an Array of Array to csv file, read csv file into
an Array of Array. No header information, also thought of as row/columns

=item AoH to csv 

Write an Array of Hash to csv file, read csv file into
an Array of Hash. Contains header information, also thought of as row/columns.

=back


=head1 SYNOPSIS

    use DataStructureStore;

    # Object constructor - used to maintain file and storage method
    $ds = DataStructureStore->new(
                         [ ['mode'     => <dd|sto|aoa|aoh>],
                           ['dir'      => <str>],        directory
                           ['file'     => <str>],        relative filename to 'dir'
                           ['ext'      => <str>],        extension
                           ['filename' => <str>],        full filename
                         ]
    );

    # Modify object variables
    $ds->dss_update_file(
                         [ ['mode'     => <dd|sto|aoa|aoh>],
                           ['dir'      => <str>],        directory
                           ['file'     => <str>],        relative filename to 'dir'
                           ['ext'      => <str>],        extension
                           ['filename' => <str>],        full filename
                         ]
    );

    # Get filename and storage method values from object
    ($fname, $mode) = $ds->dss_filename;   # array context
    $fname = $ds->dss_filename;            # scalar context

    # Method to store the referenced data structure as defined by object.
    $ds->dss_store(<dat_structure_ref>) 


    # Method to load to  the referenced data structure as defined by object.
    $ds->dss_load(<dat_structure_ref>)


    # Convenience function to call TracePkg and Data:Dumper the given Data Structure reference,
    # so that thpl does not have to use Data::Dumper since DataStructureStore is used.
    dss_print(<data structure ref>, [[<TracePkg method>], [<user defined message>]]

    dss_print($ref, 'debug', 'MyData');  same as $Trace->debug('msg' => 'MyData' . Dumper($ref));

    # Configure Data::Dumper variable
    set_data_dumper_var('Indent', 1);


=head1 PARAMS

Parameters to provide common Data Structure storage. Parameters given allow the object
to build the absolute filename and define the storage method. 

=over

=item DSS_MODE=<dd|sto|aoa|aoh>

Define storage mode

=over

=item dd - store Data Structure with Data::Dumper to DSS_FILENAME

=item sto - store Data Structure with Storable to DSS_FILENAME

=item aoa - Write Array of Array to csv file defined by DSS_FILENAME

=item aoh - Write Array of Array to csv file defined by DSS_FILENAME 

=back

=item DSS_DIR -  Absolute path used to build DSS_FILENAME, defaults to $LOGDIR
if not given or current working directory if LOGDIR=-

=item DSS_FILE - File name given to build DSS_FILENAME, defaults to data_struct

=item DSS_FILENAME - Absolute filename, if not given directly then built with
DSS_DIR, DSS_FILE and extension associated with the storage mode.

=back

=head1 Methods

User needs to first create an object with the new() method. The object is used
to maintain the information on filename and desired storage mode. The objects
values can be changed with dss_update_file() method. This provides flexibility
to change file name or directory name or storage mode as needed, such as
store 20 different data structures in the same directory but with different
filenames.

=head2 new()

    # Object constructor - used to maintain file and storage method
    $ds = DataStructureStore->new(
                         [ ['mode'     => <dd|sto>],
                           ['dir'      => <str>],        directory
                           ['file'     => <str>],        relative filename to 'dir'
                           ['ext'      => <str>],        extension
                           ['filename' => <str>],        full filename
                         ]
    );

=over

=item mode

=over

=item dd - store Data Structure with Data::Dumper to filename

=item sto - store Data Structure with Storable to filename

=item aoa - Write Array of Array to csv file defined by filename

=item aoh - Write Array of Hashes to csv file defined by filename 

=back


=item dir - full directory path used to build filename. Defaults to LOGDIR
or current working directory (CWD) if LOGDIR=-

=item file - file name used to build filename. Defaults to data_struct

=item ext - extension to use to build filename. Dependent on storage mode.

=over

=item dd mode uses .dd

=item sto mode use .sto

=item aoa mode uses .csv

=item aoh uses .csv

=back


=item filename - absolute filename used to store/load data structure
to/from. If not given, built with dir, file and ext arguments. The filename
argument has precedence over arguments given to dir, file or ext options.

=back


=head2 dss_update_file()

    $ds->dss_update_file(
                         [ ['mode'     => <dd|sto|aoa|aoh>],
                           ['dir'      => <str>],        directory
                           ['file'     => <str>],        relative filename to 'dir'
                           ['ext'      => <str>],        extension
                           ['filename' => <str>],        full filename
                         ]
    );

    Option/arguments are same as for new(), see above.


=head2 dss_filename


    ($fname, $mode) = $ds->dss_filename;
    $fname = $ds->dss_filename;

    Accessor method to get filename and storage mode values from object.
    In array context will return both filename and storage mode.
    In scalar context will return filename.


=head2 dss_store()

    $ds->dss_store($ref)

    Method to store Data Structure referenced. Storage location and method
    defined by object.
    

=head2 dss_load()

    $ds->dss_load($ref)

    Method to load Data Structure referenced into memory. Storage location and method
    defined by object.

=head2 dss_print()

    dss_print(<data structure ref>, [[<TracePkg method>], [<user defined message>]]

    Shortcut function to call TracePkg method to dump data structure to log file.
    Provided because if you are using this module then you should have no reason
    to use Data::Dumper in the local thpl.

=over

=item <data structure ref> - reference to data structure to dump to file with Data::Dumper.

=item <TracePkg method> - TracePkg method to define log level, defaults to comment.

=item <user defined message> - TracePkg method msg option argument. Defaults to
'User Supplied Data Structure'

=back


dss_print($ref) same as $Trace->comment('msg' => 'User Supplied Data Structure');

dss_print($ref, 'SomeData); same as $Trace->comment('msg' => 'SomeData' . Dumper($ref));

dss_print($ref, 'debug', 'MyData');  same as $Trace->debug('msg' => 'MyData' . Dumper($ref));

dss_print($ref, 'trace', 'TraceData'); same as $Trace->trace('msg' => 'TraceData' . Dumper($ref));


=head2 set_data_dumper_var()

    set_data_dumper_var(<data dumper variable>, <value>);

    Function to allow user to configure Data::Dumper variables. 
    See Data::Dumper documentation for varaible list and allowed values.
    This function does no testing of the value, it checks for supported
    varaibles and sets $Data::Dumper::<VAR> = value.

    If un unknown variable is passed, the function returns 0, otherwise 1;


=head1 Misc

=head2 Storable file text viewer

If you have a file creted with storable and wish to view its content, this simple
block of code will do

    #!/usr/local/bin/perl
    #
    #
    use Storable qw(retrieve);
    use Data::Dumper;

    my $file    = shift || die "ERROR: no file\n";
    my $db_ref = retrieve($file);

    print Dumper $db_ref;

    exit(0);


=head2 thpl to test this lib:

    # QA thpl
    # vim: ft=perl
    #
    use strict;
    use DataStructureStore;
    
    use TracePkg;
    use Options;
    use Params::Common;
    
    my $Trace   = new TracePkg;
    my $Option  = new Options;
    
    my ($rn, $ra, $rv, $n, $s, $v, $c1, $c2,
        $node, $aggr, $vol,
        %hash,
        @array,
        $load1, $load2,
        @tmp,
        $d2, $f2, $m2,
    );
    
    my @keys = (  'vserver', 'aggr', 'vols', 'nodes', 'files',
                  'nfs', 'cifs', 'disks', 'snapshots', 'qtrees', 'domains',
    );
    
    # thpl PARAMS
    # define params for cmd line control
    #  DSS_TEST - defines which test to run  ds,aoa,aoh
    #  DSS_DIR_2 - 2nd dir for ds test
    #  DSS_FILE_2 - 2nd file for ds test
    #  DSS_MODE_2 - mode to use storing 2nd file
    #
    $Option->parameters( 'definitions' =>
       [   {   'name'         => 'DSS_TEST',
               'description'  => 'which test to run ds|aoa|aoh',
           },
           {   'name'         => 'DSS_DIR_2',
               'description'  => '2nd directory to store data structure',
           },
           {   'name'        => 'DSS_FILE_2',
               'description' => '2nd file to store data structure, relative to DSS_DIR_2 ',
           },
           { 'name'          => 'DSS_MODE_2',
             'description'    => '2nd data structure storage mode'
           },
       ]
    );
    $Option->process();
    my $params_href = $Option->get_hash();
    
    # create data storage object
    # use params or defaults
    my $ds = DataStructureStore->new();
    dss_print($ds, 'trace', "Data Store Object\n");
    
    # set data dumper for minimal indentation
    set_data_dumper_var('Indent', 1);
    
    #
    # Make a HoHoA to store and store it two separate ways,
    # as two separate file and compare them
    #
    if ($params_href->{'DSS_TEST'} eq 'ds') {
       # make a data structure
       $Trace->comment( 'msg' => "Creating HoHoA data structure ...\n");
       $rn = int(rand(10) + 2);
       foreach $n ( 1..$rn ) {
          $node = sprintf("node-%u", $n);
          $ra = int(rand(5) + 3);
          foreach $a ( 1..$ra ) {
             $aggr = sprintf("aggr_%u", $a);
             $rv = int(rand(5) + 5);
             foreach $v ( 1..$rv ) {
                $vol = sprintf("vol_%u", $v);
                push(@{$hash{$node}{$aggr}}, $vol) ;
             }
             $Trace->trace( 'msg' => "data structure content: $node $aggr with vols: $rv\n");
          }
       }
       # 
       # store hash 
       $ds->dss_store(\%hash); sleep 2;
       $ds->dss_load($load1);
    
       #
       # change objects file and mode
       $d2 = $params_href->{'DSS_DIR_2'}  || $params_href->{'DSS_DIR'}  || $ENV{'LOGIIR'} || $ENV{'CWD'};
       $f2 = $params_href->{'DSS_FILE_2'} || $params_href->{'DSS_FILE'} || 'data_file_2';
       $m2 = $params_href->{'DSS_MODE_2'} || 'sto';
       $ds->dss_update_file('dir'  => $d2,
                            'file' => $f2,
                            'mode' => $m2,
       );
       dss_print($ds, 'trace', 'Updated Data Store Object');
       $ds->dss_store(\%hash); sleep 2;
       $ds->dss_load($load2);
    
       # induce error
       $load1->{'node-2'}{'aggr_3'}[3] = 'vol_foo';
    
       # compare data structures
       $Trace->comment( 'msg' => "Compare Structures\n");
       foreach $node (sort keys %{$load1}) {
          $Trace->comment( 'msg' => "node: $node\n");
          foreach $aggr (sort keys %{$load1->{$node}}) {
             $Trace->comment( 'msg' => "   aggr $aggr\n");
             $v = 0;
             foreach $vol  (@{$load1->{$node}{$aggr}}) {
                $vol eq $load2->{$node}{$aggr}[$v] ?
                   $Trace->comment( 'msg' => "      same: $node $aggr [$v]  $vol\n") :
                   $Trace->warn( 'msg' => "      diff: $node $aggr [$v]  $vol $load2->{$node}{$aggr}[$v]\n")
                ;
                $v++;
             }
          }
       }
    }
    #
    # AoA test
    # make AoA, store it, load it, corrupt loaded, compare it
    if ($params_href->{'DSS_TEST'} eq 'aoa' ) {
       $rn = int(rand(10) + 1) * 10;  # home many rows
       $ra = int(rand(10) + 24);      # how many columns
       # create AoA
       @array = ();
       $Trace->comment( 'msg' => "Create AoA: $rn rows $ra columns");
       foreach $n ( 0..$rn-1 ) {
          foreach $a (0..$ra-1) {
             $v = int(rand(1000));
             $array[$n][$a] = $v
          }
       }
       # store AoA
       $ds->dss_update_file('mode' => 'aoa',);
       $ds->dss_store(\@array); sleep 2;
    
       # load AoA back in to other var
       $ds->dss_load($load1);
    
       # insert bad data
       $load1->[6][7] = 5000;
    
       # verify data
       $Trace->comment( 'msg' => "Verify AoA2csv data to csv2AoA");
       foreach $n ( 0..$rn-1 ) {
          foreach $a (0..$ra-1) {
             if ( $array[$n][$a] ne $load1->[$n][$a] ) {
                $Trace->warn( 'msg' => "row $n col $a diff  $array[$n][$a] $load1->[$n][$a] ");
             }
          }
       }
    }
    # AoH test
    # make AoH
    # store it, change mode, store mode
    # corrupt it and check it
    #
    if ($params_href->{'DSS_TEST'} eq 'aoh') {
       $rn = int(rand(10) + 1) * 10;  # home many rows
       # create AoH
       $Trace->comment( 'msg' => "Create AoH: $rn rows");
       foreach $n ( 0..$rn-1 ) {
           # generate hashes with different combination of keys
           # shuffle key list and pick first 6 from list
           @tmp = _shuffle(@keys);
           foreach (1..6) { $array[$n]->{shift(@tmp)} = int(rand(1000));}
       }
       # store AoH 
       $ds->dss_store(\@array);
    
       # store as csv file
       $ds->dss_update_file('mode' => 'aoh',);
       $ds->dss_store(\@array); sleep 2;
    
       # load AoH back in to other var
       $ds->dss_load($load1);
    
       # insert bad data somewhere near end
       foreach $n ( reverse(0..$rn) ) {
          if ( $array[$n]->{'aggr'} ) {
             $Trace->comment( 'msg' => "changing element $n {aggr} to $array[$n]->{'aggr'} -> 9999");
             $array[$n]->{'aggr'} = 9999;
             last;
          }
       }
    
       # verify 
       foreach $n ( 0..$rn-1 ) {
          foreach $a ( keys %{$load1->[$n]} ) {
             if ( $load1->[$n]{$a} ne $array[$n]->{$a} ) {
                $Trace->warn( 'msg' => "row $n key $a diff  $array[$n]->{$a} $load1->[$n]{$a} ");
             }
          }
       }
    }
    
    $Trace->exit( 'msg' => 'This thpl ran and completed');
    
    
    exit(0);
    #.............................................................................
    #
    # shuffle an array
    # from cookbook 4.17
    sub _shuffle {
       my @_array = @_;
       my ($_i, $_j);
       for ($_i = scalar(@_array); --$_i;) {
          $_j = int rand($_i + 1);
          next if $_i == $_j;
          @_array[$_i, $_j] = @_array[$_j, $_i];
       }
       @_array;
    }






=head1 STATUS

   Current 

=head1 AUTHOR

  Scott Parsons (parsons@netapp.com)
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2008, Network Appliance., All Rights Reserved

=cut













