#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/DataUtils.pm#1 $

#***********************************************************************
#
# NAME
#   DataUtils.pm 
#
# SYNOPSIS
#   General Perl Data Structure Utilities
#
# DESCRIPTION
#   This library provides routines to help in manipulate perl data structures.
#
# AUTHOR
#   dl-cluster-qa
#***********************************************************************

package DataUtils;

use Tharn;
use XML::LibXML;

## @name extract_field()
## @description Public method to extract all of the values for a given
##              hash key, in a multi-level hash.
##
## @arg hash   Required: A reference to a hash of hashes.
## @arg array  Required: A reference to an array which will be used to store the values associated
##                       with the keys that match the field option
## @arg field  Required: A string or regexp used to match keys in the hash.
##
## @synopsis
##
##   extract_field( hash => \%hash, array => \@array, field => 'asup_id' );
##
##    For example,
##
##    my (%hash, @states, @asup_ids, @links);
##
##    extract_field( hash => \%hash, array => \@states,   field => 'status_code' );
##    extract_field( hash => \%hash, array => \@asup_ids, field => 'asup_id' );
##    extract_field( hash => \%hash, array => \@links,    field => '(link|asup_search_url)' );
##
## Note: Fields will be populated in the array in the depth order in
##       which they appeared in the hash.  This is useful when a hash
##       contains something like a "status" field with an overall status
##       at a higher level in the hash and sub task statuses deeper in the
##       hash.
##
## Note: Options "depth" and "found_fields" are NOT intended to be called by end
##       users.  These are options passed during recursive calls.
## 
## Note: A known limitation is that only 9999 levels of hash depth are supported.
##       Presumably this should not be a problem.

sub extract_field {
    my $whoami       = (caller(0))[3];
    my $whocalledme  = (caller(1))[3];

    my %options = (
        hash          => undef,
        array         => undef,
        field         => undef,
        depth         => 0,
        found_fields  => undef,
        @_
    );

    if ( !$options{hash} || (ref($options{hash}) !~ /HASH/i) ) {
        die script_result("$whoami must be called with hash option and hash must be a hash reference");
    }

    if ( !$options{array} || (ref ($options{array}) !~ /ARRAY/i) ) {
        die script_result("$whoami must be called with array option and array must be an array reference");
    }

    if ( !$options{field} ) {
        die script_result("$whoami must be called with field option");
    }

    # These options are "private" and are only intended to be passed by extract_field itself
    if ( ($options{depth} || $options{found_fields}) && ($whoami ne $whocalledme) ) {
        die script_result("$whoami options depth and found_fields can not be specified by end-user, only by $whoami");
    }

    my $hash      = $options{hash};
    my $array     = $options{array};
    my $field     = $options{field};
    my $depth     = $options{depth};

    # preface $depth with 0s so that sorts are unique
    my $fmt_depth = sprintf("%04d", $depth);

    my %fhash;
    my $found_fields = ($options->{found_fields}) ? $options->{found_fields} : \%fhash;

    foreach my $key (keys %{$hash}) {
        if ( $key =~ /^$field$/ ) {
            if ( ! exists $found_fields->{$fmt_depth} ) {
                $found_fields->{$fmt_depth} = [];
            }
            push( @{$found_fields->{$fmt_depth}}, $hash->{$key} );
        } elsif ( ref($hash->{$key}) =~ /HASH/i ) {
            extract_field( hash => $hash->{$key}, found_fields => $found_fields, field => $field, depth => ($depth + 1) );
        }
    }

    # Order discovered fields by their depth in the hash
    if ( $depth == 0 ) {
        foreach my $key (sort keys %{$found_fields} ) {
            push( @$array, @{$found_fields->{$key}} );
        }
    }
}

## @name build_hash_from_xml()
## @description Method to convert a XML::LibXML::Document object into
##              a multi-level hash.
##
## @arg xmlnode Required: A reference to a XML::LibXML::Node object
## @arg hash    Required: A reference to a hash which will contain the results of
##                        the conversion
##
## @synopsis
##
##    my %hash;
##    my $xmlNode
##
##    build_hash_from_xml( xmlnode => $xmlNode, hash => \%hash );
##
sub build_hash_from_xml {
    my $whoami  = (caller(0))[3];
    my %options = (
        xmlnode => undef,
        hash    => undef,
        @_
    );

    if ( !$options{xmlnode} || (ref($options{xmlnode}) !~ /LibXML/i) ) {
        die script_result("$whoami must be called with xmlnode option and node must be a XML::LibXML node reference");
    }

    if ( !$options{hash} || (ref ($options{hash}) !~ /HASH/i) ) {
        die script_result("$whoami must be called with hash option and hash must be a hash reference");
    }

    my $node  = $options{xmlnode};
    my $hash  = $options{hash};


    my $name = $node->nodeName();
    my $type = $node->nodeType();
    my ($path, $text);

    if ( ($type == XML_ELEMENT_NODE) ||
         ($type == XML_TEXT_NODE) ||
         ($type == XML_ATTRIBUTE_NODE) ) {
        $path = $node->nodePath();
        $text = $node->textContent();
    }

    if ( ($type == XML_ELEMENT_NODE) || ($type == XML_ATTRIBUTE_NODE) ) {
        my @path = split( /\//, $path );

        # remove any leading ""
        if ( !$path[0] ) {
            shift( @path );
        }

        my @children = $node->childNodes();
        my $textChildren = 0;
        my $elementChildren = 0;

        foreach my $child (@children) {
            my $ctype = $child->nodeType();
            my $content = $child->textContent();
            $content =~ s/\s//g;
            if ( $ctype == XML_ELEMENT_NODE ) {
                $elementChildren++;
            } elsif ( ($ctype == XML_TEXT_NODE) && ($content) ) {
                $textChildren++;
            }
        }

        if ( !$elementChildren || ($textChildren == 1) ) {
            my $lastKey = pop @path;
            $lastKey =~ s/^\@//;

            my $hptr = $hash;
            foreach my $key (@path) {

                if ( !exists $hptr->{$key} ) {
                    $hptr->{$key} = {};
                }
                $hptr = $hptr->{$key};
            }

            $hptr->{$lastKey} = $text;
        }

        if ( $node->hasAttributes() ) {
            foreach my $attr ($node->attributes()) {
                build_hash_from_xml( xmlnode => $attr, hash => $hash );
            }
        }
    }

    if ( ($type == XML_ELEMENT_NODE) ||
         ($type == XML_TEXT_NODE) ||
         ($type == XML_ATTRIBUTE_NODE) ) {
        my @children = $node->childNodes();
        foreach my $child (@children) {
            build_hash_from_xml( xmlnode => $child, hash => $hash );
        }
    }
}

1;
