# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a database utility package for FabricMgr
## @author derricks@netapp.com
## @status private
## @pod here
#
# $Id: //depot/prod/test/Rsierranevada/lib/DbUtil/postgres.pm#1 $
package DbUtil::postgres;

# Compiler directives.
use strict;

# Module import.
use Tharn;
use Error;
use Data::Dumper;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################
###############################################################################
# Method:    insert
# Objective: Insert row into database
# Details:   See POD documentation at the end of this file.
###############################################################################
sub insert {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);
    my $table     = &ParamsMgr::get_arg(-arg => "table",  -required => 1, -default => {});
    my $array_ref = &ParamsMgr::get_arg(-arg => "data",   -required => 1, -default => []);

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    # Make sure $data is a ARRAY
    unless (ref($array_ref) eq "ARRAY") {
        $self->_set_err("Invalid data type for '-data'.  Type must be an array reference");
        return $SanUtil::FALSE;
    }

    # Parse through the array reference.  Each element should be a hash reference of the
    # following layout:
    #
    # $ref = { key   => "key",
    #          value => "value",
    #          type  => "data type"
    #        }
    #
    # We will store each key in an array and each value in a array.
    my @cols;
    my @values;
    foreach my $element (@$array_ref) {
        unless (ref($element) eq "HASH") {
            $self->_set_err("Invalid data type for array element.  Type must be a hash reference");
            return $SanUtil::FALSE;
        }

        # Make sure all of the keys are found
        unless (defined($element->{key})
                and $element->{key} ne ""
                and defined($element->{value})
                and $element->{value} ne ""
                and defined($element->{type})
                and $element->{type} ne "") {
            $self->_set_err("Invalid data specified in hash reference for database insert");
            return $SanUtil::FALSE;
        } ## end unless (defined($element->...

        # Check the data type and create the where clause appropriately.
        if ($element->{type} eq "bool") {
            if ($element->{type} =~ /false/i) {
                push(@cols,   $element->{key});
                push(@values, "FALSE");
            } elsif ($element->{type} =~ /true/i) {
                push(@cols,   $element->{key});
                push(@values, "TRUE");
            } else {
                $self->_set_err("Invalid value specified in hash reference for type of bool");
                return $SanUtil::FALSE;
            }
        } else {
            push(@cols,   $element->{key});
            push(@values, "'$element->{value}'");
        }
    } ## end foreach my $element (@$array_ref)

    # Make sure that the column count and value count are the same
    unless (scalar(@cols) == scalar(@values)) {
        $self->_set_err("Invalid number of key and values pairs specified.");
        return $SanUtil::FALSE;
    }

    # Join the cols and values on commas.
    my $cols   = join(", ", @cols);
    my $values = join(", ", @values);

    # Grab the database handle off of self
    my $dbh = $self->get_handle();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve database handle");
        return [];
    }

    # Do the insert
    my $sth = $dbh->prepare( qq{ INSERT into $table \($cols\) values \($values\) } );

    unless ($sth->execute()) {
        $self->_set_err("Insert into table '$table' failed for columns ($cols) and values ($values):\n" . $dbh->errstr());

        # Disconnect from the database before returning if we persistent is "0"
        if ($self->{persistent} == 0) {
            $self->disconnect_db();
            if ($self->error()) {
                $self->_set_err("There was an error while trying to perform the " .
                                "database disconnect:\n");
            }
        }
        return $SanUtil::FALSE;
    }

    # Disconnect from the database before returning if we persistent is "0"
    if ($self->{persistent} == 0) {
        $self->disconnect_db();
        if ($self->error()) {
        $self->_set_err("There was an error while trying to perform the " .
                        "database disconnect:\n");
        }
    }

    # Log a message about the insert success
    $self->sanlog(-msg => "Insert into table '$table' succeeded for columns ($cols) and values ($values)");

    return $SanUtil::TRUE;
} ## end sub insert

###############################################################################
# Method:    delete
# Objective: Delete a row from the database
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete {
    my ($self, %args) = @_;

    my ($aref);

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);
    my $table     = &ParamsMgr::get_arg(-arg => "table",  -required => 1, -default => {});
    my $array_ref = &ParamsMgr::get_arg(-arg => "data",   -required => 0, -default => []);

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    # Make sure $data is a ARRAY
    unless (ref($array_ref) eq "ARRAY") {
        $self->_set_err("Invalid data type for '-data'.  Type must be an array reference");
        return $SanUtil::FALSE;
    }

    # Parse through the array reference.  Each element should be a hash reference of the
    # following layout:
    #
    # $ref = { key   => "key",
    #          value => "value",
    #          type  => "data type"
    #        }
    #
    # We will store each key/value pair in an array
    my @where = ();
    foreach my $element (@$array_ref) {
        unless (ref($element) eq "HASH") {
            $self->_set_err("Invalid data type for array element.  Type must be a hash reference");
            return $SanUtil::FALSE;
        }

        # Make sure all of the keys are found
        unless (defined($element->{key})
                and $element->{key} ne ""
                and defined($element->{value})
                and $element->{value} ne ""
                and defined($element->{type})
                and $element->{type} ne "") {
            $self->_set_err("Invalid data specified in hash reference for database insert");
            return $SanUtil::FALSE;
        } ## end unless (defined($element->...

        # Check the data type and create the where clause appropriately.
        my $string;
        if ($element->{type} eq "bool") {
            if ($element->{type} =~ /false/i) {
                $string = "$element->{key}=FALSE";
            } elsif ($element->{type} =~ /true/i) {
                $string = "$element->{key}=TRUE";
                push(@where, $string);
            } else {
                $self->_set_err("Invalid value specified in hash reference for type of bool");
                return $SanUtil::FALSE;
            }
        } else {
            $string = "$element->{key}='$element->{value}'";
            push(@where, $string);
        }
    } ## end foreach my $element (@$array_ref)

    if (scalar(@where) == 0) {

        # We weren't given anything to delete so error
        $self->_set_err("No key-value pairs found in data hash");
        return $SanUtil::FALSE;
    } ## end if (scalar(@where) == ...

    # create the where clause
    my $where_clause = join(" and ", @where);
    $where_clause = "where " . $where_clause;

    # Grab the database handle off of self
    my $dbh = $self->get_handle();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve database handle");
        return [];
    }

    # Do the delete
    my $sth = $dbh->prepare( qq{DELETE from $table $where_clause} );

    unless ($sth->execute()) {
        $self->_set_err("Delete from table '$table' failed with where clause of '$where_clause':\n" . $dbh->errstr());

        # Disconnect from the database before returning if we persistent is "0"
        if ($self->{persistent} == 0) {
            $self->disconnect_db();
            if ($self->error()) {
                $self->_set_err("There was an error while trying to perform the " .
                                "database disconnect:\n");
            }
        }
        return $SanUtil::FALSE;
    }

    # Query the database to make sure the row was deleted
    $aref = $self->query(-table => $table, -data => $array_ref, -where_type => "and");
    if ($self->error) {
        $self->_set_err("Unable to query database");

        # Disconnect from the database before returning if we persistent is "0"
        if ($self->{persistent} == 0) {
            $self->disconnect_db();
            if ($self->error()) {
                $self->_set_err("There was an error while trying to perform the " .
                                "database disconnect:\n");
            }
        }
        return $SanUtil::FALSE;
    }

    # Disconnect from the database before returning if we persistent is "0"
    if ($self->{persistent} == 0) {
        $self->disconnect_db();
        if ($self->error()) {
            $self->_set_err("There was an error while trying to perform the " .
                            "database disconnect:\n");
        }
    }

    if (scalar(@$aref) > 0) {
        # Looks like the delete failed
        $self->_set_err("Database delete failed for table '$table' with where clause of '$where_clause'");
        return $SanUtil::FALSE;
    } ## end if (scalar(@$aref) > 0)

    # Log a message about the insert success
    $self->sanlog(-msg => "Delete from table '$table' succeeded with where clause of '$where_clause'");

    return $SanUtil::TRUE;
} ## end sub delete

###############################################################################
# Method:    query
# Objective: Query the database
# Details:   See POD documentation at the end of this file.
###############################################################################
sub query {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);
    my $table      = &ParamsMgr::get_arg(-arg => "table", -required => 1, -default => {});
    my $array_ref  = &ParamsMgr::get_arg(-arg => "data", -required => 0, -default => []);
    my $where_type = &ParamsMgr::get_arg(-arg => "where_type", -required => 0, -default => "and");
    my $all        = &ParamsMgr::get_arg(-arg => "all", -required => 0, -default => 0);

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # Make sure $data is a ARRAY
    unless (ref($array_ref) eq "ARRAY") {
        $self->_set_err("Invalid data type for '-data'.  Type must be an array reference");
        return [];
    }

    # Make sure that where_type is either "and" or "or"
    if ($where_type =~ /and/i) {
        # Reset it to uppercase "and" just in case it was upper case
        $where_type = "AND";
    } elsif ($where_type =~ /or/i) {
        # Reset it to uppercase "or" just in case it was upper case
        $where_type = "OR";
    } else {
        $self->_set_err("Invalid type '$where_type' specified for '-where_type'.  Must be 'or' or 'and'.");
        return [];
    }

    # Parse through the array reference.  Each element should be a hash reference of the
    # following layout:
    #
    # $ref = { key   => "key",
    #          value => "value",
    #          type  => "data type"
    #        }
    #
    # We will store each key/value pair in an array
    my $where_clause = "";
    my @where        = ();
    if ($all == 0) {
        foreach my $element (@$array_ref) {
            unless (ref($element) eq "HASH") {
                $self->_set_err("Invalid data type for array element.  Type must be a hash reference");
                return [];
            }

            # Make sure all of the keys are found
            unless (defined($element->{key})
                and $element->{key} ne ""
                and defined($element->{value})
                and $element->{value} ne ""
                and defined($element->{type})
                and $element->{type} ne "") {
                $self->_set_err("Invalid data specified in hash reference for database insert");
                return [];
            } ## end unless (defined($element->...

            # Check the data type and create the where clause appropriately.
            my $string;
            if ($element->{type} eq "bool") {
                if ($element->{type} =~ /false/i) {
                    $string = "$element->{key}=FALSE";
                    push(@where, $string);
                } elsif ($element->{type} =~ /true/i) {
                    $string = "$element->{key}=TRUE";
                    push(@where, $string);
                } else {
                    $self->_set_err("Invalid value specified in hash reference for type of bool");
                    return [];
                }
            } else {
                $string = "$element->{key}='$element->{value}'";
                push(@where, $string);
            }
        } ## end foreach my $element (@$array_ref)

        if (scalar(@where) == 0) {

            # We weren't given anything to query so error
            $self->_set_err("No key-value pairs found in data hash");
            return [];
        } ## end if (scalar(@where) == ...
        $where_clause = join(" " . $where_type . " ", @where);
        $where_clause = "where " . $where_clause;
    } ## end if ($all == 0)

    # Grab the database handle off of self
    my $dbh = $self->get_handle();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve database handle");
        return [];
    }

    # Do the query
    my $sth = $dbh->prepare( qq{ SELECT * from $table $where_clause } );

    unless ($sth->execute()) {
        $self->_set_err("Select from table '$table' failed with where clause of '$where_clause':\n" . $dbh->errstr());

        # Disconnect from the database before returning if we persistent is "0"
        if ($self->{persistent} == 0) {
            $self->disconnect_db();
            if ($self->error()) {
                $self->_set_err("There was an error while trying to perform the " .
                                "database disconnect:\n");
            }
        }
        return [];
    }

    # Retrieve the results as an arrayref and return
    my $ref = $sth->fetchall_arrayref();

    # Disconnect from the database before returning if we persistent is "0"
    if ($self->{persistent} == 0) {
        $self->disconnect_db();
        if ($self->error()) {
            $self->_set_err("There was an error while trying to perform the " .
                            "database disconnect:\n");
        }
    }

    if (scalar(@$ref) > 0) {
        return $ref;
    } else {
        return [];
    }
} ## end sub query

###############################################################################
# Method:    get_handle
# Objective: Get a database handle off of self or create one
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_handle {
    my ($self) = @_;

    $self->_clear_err();

    unless (defined($self->{handle}) and ref($self->{handle}) =~ /dbi/i) {
        $self->connect_db();
        if ($self->error()) {
            $self->_set_err("Unable to retrieve database handle");
            return {};
        }
    }

    return $self->{handle};
}

###############################################################################
# Method:    connect_db
# Objective: Get a database handle
# Details:   See POD documentation at the end of this file.
###############################################################################
sub connect_db {
    my ($self) = @_;

    $self->_clear_err();

    # Init some variables from self
    my $instance = $self->{instance};
    my $server   = $self->{server};
    my $user     = $self->{username};
    my $pass     = $self->{password};

    # See if we already have a connection active
    if (defined($self->{handle})) {
        if (ref($self->{handle}) =~ /dbi/i) {
            $self->_set_err("Database connection already exists");
        } else {
            $self->_set_err("Database handle is already defined, but the data type is invalid");
        }
        return $SanUtil::FALSE;
    }

    # Try to create the connection via eval.  If it works, store the handle and return TRUE
    my $dbh;
    if (eval {
            require DBI;
            $dbh = DBI->connect("dbi:Pg:dbname=$instance;host=$server", "$user", "$pass", {PrintError => 0});
            }
        ) {

        # Store the handle on self
        $self->{handle} = $dbh;
    } else {
        $self->_set_err("Unable to create the connection for database '$instance' on server '$server'");
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub connect

##############################################################################
# Method:    disconnect_db
# Objective: Disconnect the database
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disconnect_db {
    my ($self) = @_;

    $self->_clear_err();

    # Make sure that $self->{handle} on self is an actual handle
    if (defined($self->{handle})) {
        if (ref($self->{handle}) =~ /dbi/i) {
            unless ($self->{handle}->disconnect()) {
                $self->_set_err("Disconnect of database failed:\n" . $self->{handle}->errstr());
                return $SanUtil::FALSE;
            }

            # Reset the handle to undef
            $self->{handle} = undef;
            return $SanUtil::TRUE;
        } else {
            $self->_set_err("Invalid data type of '" . ref($self->{handle}) . "' found for handle on self");
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
} ## end sub disconnect_db

#############################################################################
#                             Private Methods
#############################################################################

###############################################################################
# Method:    _new
# Objective: Create a new DbUtil::postgres object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($class, %args) = @_;

    my $self = {};

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    # Clear out the errors
    $self->_clear_err();

    # need to tell Error class to handle errors in a special way
    # this will allow us to invoke set err over and over to create an error stack
    # without overwriting the previous error
    $self->{error_dont_overwrite_errors} = 1;

    # Make sure the stack style is normal
    $self->{error_stack_style} = "normal";

    &ParamsMgr::process_args(%args);
    my $server     = &ParamsMgr::get_arg(-arg => "server",   -required => 1, -default => "");
    my $instance   = &ParamsMgr::get_arg(-arg => "instance", -required => 1, -default => "");
    my $username   = &ParamsMgr::get_arg(-arg => "username", -required => 1, -default => "");
    my $password   = &ParamsMgr::get_arg(-arg => "password", -required => 1, -default => "");
    my $persistent = &ParamsMgr::get_arg(-arg => "persistent", -required => 0, -default => 0);

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $self;
    }

    # Store some parameters on the object
    $self->{server}     = $server;
    $self->{instance}   = $instance;
    $self->{username}   = $username;
    $self->{password}   = $password;
    $self->{persistent} = $persistent;

    return $self;
}


#############################################################################
#                             POD Documentation
#############################################################################
=pod

=head1 Package Name

DbUtil::postgres

=head1 Author

Derrick Shultz

=head1 Description

DbUtil::postgres provides an API interface to allow database access to a Postgres database.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Tharn, Error>

=head1 Attributes

=head2 B<server>

String = Database Server to connect to

=head2 B<instance>

String = Database instance

=head2 B<username>

String = Database user name

=head2 B<password>

String = Database user password

=head2 B<persistent>

Boolean = Whether to have a persitent database connection.  True will always leave it connected unless the caller
explicitly disconnects.  False will disconnect the database after every call.  The default is False.

=head1 Public Methods

=head2 B<connect_db()>

Connect to the database and return a database handle.

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $db_obj->connect();

 if ($db_obj->error) {
     print $db_obj->errmsg;
     return 0;
 }

=back

=head2 B<disconnect_db()>

Disconnect from a database

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $db_obj->disconnect_db();
 if ($db_obj->error) {
     print $db_obj->errmsg;
     return 0;
 }

=back

=head2 B<delete()>

Delete something from the database.

=over 4

=item Inputs

=item * I<-table>

Database table to use in the delete

=item * I<-data>

A reference to an array of hashes containing data to delete.

The hash within the array should represent something to be used in the "where" clause of
the delete and should take the following form:

    $ref = { key   => "key",
             value => "value",
             type  => "data type" }

Example:

my @array = (
                { key => "hash",
                  value => $md5_zone_name,
                  type => "varchar"
                },
                { key => "fabric",
                  value => $fabric,
                  type => "varchar"
                },
);

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $db_obj->delete(-table => $table, -data => \@array_data);
 if ($db_obj->error) {
     print $db_obj->errmsg;
     return 0;
 }

=back

=head2 B<insert()>

Insert something into the database.

=over 4

=item Inputs

=item * I<-table>

Database table to use in the delete

=item * I<-data>

A reference to an array of hashes containing data to insert.

The hash within the array should represent something to be used in the "values" clause of
the insert and should take the following form:


my @array = (
                { key => "zone",
                  value => $zone_name,
                  type => "varchar"
                },
                { key => "hash",
                  value => $md5_zone_name,
                  type => "varchar"
                },
                { key => "fabric",
                  value => $fabric,
                  type => "varchar"
                },
                { key => "client",
                  value => $client_name,
                  type => "varchar"
                },
                { key => "hba",
                  value => $hba
                  type => "varchar"
                },
                { key => "cluster",
                  value => $cluster_name,
                  type => "varchar"
                },
                { key => "controller",
                  value => $controller_name
                  type => "varchar"
                },
                { key => "target",
                  value => $target_device
                  type => "varchar"
                },
            );

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $db_obj->insert(-table => $table, -data => \@array_data);
 if ($db_obj->error) {
     print $db_obj->errmsg;
     return 0;
 }

=back

=head2 B<query()>

Query the database

=over 4

=item Inputs

=item * I<-all> (Default: 0)

Boolean.  Indicates whether a "where" clause should used for the query or everything in the table shoudl
be returned.

=item * I<-table>

Database table to use in the query

=item * I<-data>

A reference to an array of hashes containing data to query.

The hash within the array should represent something to be used in the "where" clause of
the insert and should take the following form:

Example:

my @array = (
                { key => "hash",
                  value => $md5_zone_name,
                  type => "varchar"
                },
                { key => "fabric",
                  value => $fabric,
                  type => "varchar"
                },
);

=item * I<-where_type>

The type of operator to use in the where clause.  Should be "or" or "and".  Default is "and".


=item Outputs

Returns a reference to an array containing the results of the query.
If there is a problem, an error is raised and [] is returned.

=item Examples

 # Query the entire database
 my $array_ref = $db_obj->query(-table => $table, -all => 1);
 if ($db_obj->error) {
     print $db_obj->errmsg;
     return 0;
 }

 # Query the specific values
 my $array_ref = $db_obj->query(-table => $table, -data => \@array_data, -where_type => "or");
 if ($db_obj->error) {
     print $db_obj->errmsg;
     return 0;

=back

=cut
