#
{ package Dense::Agnostic;
    use	strict;

    use Data::Dumper;

    use Tharn;
       
    use Dense::Agnostic::7mode  ();
    use Dense::Agnostic::Cmode  ();
    use NACL::APISet::Host::CLI ();

    use vars ('@ISA', '@EXPORT'); 
    require Exporter;
    @ISA       = qw(Exporter);
    @EXPORT    = qw( $Msg );

    our $Msg;

    { package Dense::Agnostic::Node;
        use Data::Dumper;
        use Tharn   qw($Log);
        use Utils   qw(make_host);

        use Dense::Agnostic::Constants;
       

        sub new (@)
        {
            my  ($F, $fh, $class, $mode, %opts);

            $class  = shift();
            %opts   = @_;

            $Log->enter();

            if (!defined($opts{Nodename})) {
                my $emsg = "XXX Dense::Agnostic::Node: require a Nodename arg XXX";
                $Log->comment($emsg);
                $Log->debug(print Dumper(\%opts));
                die($emsg);
            }

            #/Just/ in case someone feels the overwhelming need to override what
            #the tharnhost file believes is the operating mode of the node in
            #question.
            $mode   = delete $opts{Mode};
            if (!defined($mode)) {
                $fh = make_host($opts{Nodename});
                if ($fh->hosttype =~ HOSTTYPE_CMODE) {
                    $mode = 'CMODE';
                } else {
                    $mode = '7MODE';
                }
            }

            if ($mode =~ '7MODE') {
                require Dense::Agnostic::7mode::Messages;
                $Msg = $Dense::Agnostic::7mode::Messages::Msg;
                $F   = Dense::Agnostic::7mode::Node->new(%opts);
            } else {
                require Dense::Agnostic::Cmode::Messages;
                $Msg = $Dense::Agnostic::Cmode::Messages::Msg;
                $F   = Dense::Agnostic::Cmode::Node->new(%opts);
            }

            $Log->exit();

            #XXX note - we consciously do /not/ want to bless $F!
            return $F;
        }
    } #Node

    #Usage
    # setup(Nodename    => <hostname of filer>,
    #       Aggrname    => <aggrname>,
    #       Ndisks      => <no. of disks in aggr>,
    #       Volname     => <volname>,
    #       Volsize     => <size of volume>,
    #       Purge       => <if true, knock off existing aggrs/vvols of given names
    #                       and create afresh>
    #       Reuse       => <the opposite of Purge. Mutually exclusive with Purge>
    #       Client      => <hostname of client>,);
    #
    #   -- OR --
    #
    # setup(cfg  => cfg_spec);
    #
    # Returns a hashref with the following structure:
    #
    # {
    #   Nodes => [
    #     {Node=>node, Partner=>node, Vservers=>[], Aggrs=>[aggr0, ...], Vvols=>[vvol0, ...]},
    #     {Node=>node, Partner=>node, Vservers=>[], Aggrs=>[aggr0, ...], Vvols=>[vvol0, ...]},
    #           ...
    #   ],
    #
    #    Vservers => [
    #    ],
    #
    #   Clients =>  [Client0, Client1, ... ClientN],
    # }
    #
    # NOTE: the first invocation should typically return a hashref populated
    # with exactly one Nodes entry. However, should the filer in question be
    # CFO-paired with another, then setup /should/ return a hashref with two
    # entries, one for each node (with the Partner param set appropriately.
    #
    # The first invocation is dealt with "locally". The second, which is hoped
    # to eventually deprecate the first, is meant to interface with the cluster
    # config tool - which hopefully solves most of our problems.
    #
    # The initial thoughts are that the cfgspec will look like this (and could
    # well be defined in a pm of its own):
    #
    #$cfg_2nodes = {
    #    Nodenames   => [ [$node1, $node2],  #SFO pair 1
    #                     [$node3, $node4],  #SFO pair 2
    #                   ],
    #
    #    Aggrs       => [
    #                     $aggr1 => {
    #                        Nodes       => [$node1, ... ],
    #                        Ndisks      => $N_disks,
    #                        Disklist    => [$disk1, ...],
    #                        }
    #                   ],
    #
    #    Vservers    => [
    #                   ],
    #                    
    #
    #    Vvols       => [
    #                   ],
    #};

    sub setup (@)
    {
        #Setup can operate in one of two modes. In the first mode, it accepts
        #the following args, and returns a "standard configuration" of 1 vvol, 1
        #aggr in 1 node. In the second, it accepts a config spec as an arg, and
        #plugs in with the cluster config tool perhaps, to return the 
        #configuration specified by the config spec.

        my  (%opts, $C);

        %opts   = @_;
        $Log->debug(print "XXX dumping args in Agnostic::setup XXX");
        $Log->debug(print Dumper(\%opts));
        $C   = {};

        #XXX check %opts for valid args

        if (! defined $opts{Cfg_spec}) {
            #Std config, one of each.
            my  ($nodename);
            my  ($vserver, $vs_rootvol, $reuse_vserver);
            my  ($aggr, $ndisks, $reuse_aggr, $aggr_ind_type, $aggr_raid_type);
            my  ($vvol, $volsize, $reuse_vvol);
            my  ($cl_hostname);

            my  ($Vs, $F, $A, $V, $Client);

            #Defaults
            $reuse_aggr     = 1;
            $reuse_vserver  = 1;
            $reuse_vvol     = 0;
	    $aggr_ind_type  = 64;
	    $aggr_raid_type = 'raid_dp';
            $nodename = $opts{Nodename};

            $vserver = $opts{Vserver};
            $vs_rootvol = $opts{Vserver_rootvol};
            if ($opts{ReuseVserver}) {
                $reuse_vserver = $opts{ReuseVserver};
            }

            $aggr = $opts{Aggrname};
            $ndisks = $opts{Ndisks};
            if ($opts{ReuseAggr}) {
                $reuse_aggr = $opts{ReuseAggr};
            }

            $vvol = $opts{Volname};
            $volsize = $opts{Volsize};
            if ($opts{ReuseVol}) {
                $reuse_vvol = $opts{ReuseVol};
            }

	    if ($opts{IndirectType}) {
		$aggr_ind_type = $opts{IndirectType};
	    }

	    if ($opts{Raidtype}) {
		$aggr_raid_type = $opts{Raidtype};
	    }

            $cl_hostname = $opts{Client};


            #Strictly speaking, clients aren't part of the cluster, they are its
            #consumers. Note that $C->{Vservers} will be undefined for 7Mode.
            $C->{Nodes}      = [];
            $C->{Vservers}   = [];
            $C->{Clients}    = [];

            #Populate record for Node #1. Node #2, if we have a partner.
            eval {
                $Log->comment("[1] Creating node instance for <$nodename>");
                $F = Dense::Agnostic::Node->new(Nodename=>$nodename);

	    	# Due to changes in the requirement for the minimum number of disks required to create an aggregate
	    	# on non-debug build filers adding the below check
		if (!$F->check_debug() && $ndisks < 5) {	
		    $ndisks = 5;
		}
 
                $Log->comment("[2] Creating aggr $aggr");
                $A  = $F->aggr_create(
                        Aggrname    =>$aggr,
                        Ndisks      =>$ndisks,
                        Reuse       =>$reuse_aggr,
			IndirectType=>$aggr_ind_type,
			Raidtype    =>$aggr_raid_type,
                      );

                $Log->comment("[3] Creating vserver $vserver");
                $Vs = $F->vserver_create(
                        Vservername => $vserver,
                        Rootvolume  => $vs_rootvol,
                        Aggr        => $A,
                        Reuse       => $reuse_vserver,
                        );

		### voloptions defined or not , we are passing the Vol_opts key additionally here  
		if ( defined ( $opts{Vol_opts} ))
		{
			$Log->comment("[4] Creating vvol $vvol");
	    		$V = $F->vol_create(
    				Vserver => $Vs,
			        Aggr    => $A,
			        Volname => $vvol,
			        Size    => $volsize,
			        Reuse   => $reuse_vvol,
				Vol_opts => $opts{Vol_opts},
    			);
		}else
		{
			
			$Log->comment("[4] Creating vvol $vvol");
	    		$V = $F->vol_create(
    				Vserver => $Vs,
			        Aggr    => $A,
			        Volname => $vvol,
			        Size    => $volsize,
			        Reuse   => $reuse_vvol,
			);
		}

                $Log->comment("[5] Setting up dedup license");
                $F->license_add(License=>'OSFYRVH');

                $C->{Vservers}[0]               = $Vs;

                $C->{Nodes}[0]->{Node}          = $F;

                $C->{Nodes}[0]->{Aggrs}         = [];
                $C->{Nodes}[0]->{Volumes}       = [];

                $C->{Nodes}[0]->{Aggrs}[0]      = $A;
                $C->{Nodes}[0]->{Volumes}[0]    = $V;
            };
            if ($@) {
                die("Setup failed: $@");
            }
        } else {
            #Currently not supported. Need to plugin with the cluster config
            #tool.
            my  ($cspec);

            $cspec = delete $opts{Cfg_spec};
            #Check if $cspec is a valid path, before passing it on to the
            #cluster config tool?

            die("Cluster config tool integration not done yet");
        }

        return($C);
    } #Dense::Agnostic::setup

    return (1);
} #package
