{ package   Dense::Agnostic::7mode::Node;
    use strict;
    use warnings;

    use Data::Dumper;

    use Params::Validate qw( validate SCALAR ARRAYREF HASHREF );

    use Tharn qw($Log);

    use cluster;
    use Dense::Agnostic::Constants;
    use Dense::Agnostic::7mode::Aggr    ();
    use Dense::Agnostic::7mode::Volume  ();
    use boot qw(reboot_to);
    use Dense::Agnostic::7mode::Messages;
    use Dense::Agnostic::7mode::Node::Fmap;
    use Admin::License qw(is_licensed license_raw get_license_code);
    use Wreck qw(wreck);
    use Dense::Upgrade_Revert qw(do_transform get_filer_to_desired_release);
    use LogReader::Sis;
    use Storage::Misc qw(misc_setflag misc_printflag);
    use Utils   qw(make_host make_conn);
    use filer   qw(
                    filer_cli_say
                    filer_cli_test_say
                    filer_version_get
                    filer_service_license
                );
    use Dense::Utils qw(panic);
    use SnapMirror qw(snapmirror_initialize
	                  snapmirror_update
					  snapmirror_break
					  snapmirror_resync);

    sub new (@)
    {
        my  ($F, $class, %opts);

        $class  = shift();
        %opts   = @_;

        $F   = {};

        if (!defined($opts{Nodename})) {
            my  $emsg = "Dense::Agnostic::7mode::Node::new: need Nodename param";
            $Log->comment($emsg);
            die($emsg);       
        }

        $F->{name}   = $opts{Nodename};

        eval {
            $F->{fh}        = make_host($F->{name});
            $F->{fc}        = make_conn($F->{fh});
            $F->{version}   = filer_version_get($F->{fc});
            $F->{mode}      = MODE_7M;
        };
        if ($@) {
            die($@);
        }
        #This is for the param resolution .
        $F->{params}    = {
            fh          => $F->{fh},
            host        => $F->{name},
            filer       => $F->{fh},
            conn        => $F->{fc},
        };

        bless($F, $class);
        return $F;
    }

    #XXX: need to make this understand named params - Cmd, for instance.
    sub say (@)
    {
        my  ($self, %opts, $cmd, $timeout);

        $self = shift();
        %opts = validate (@_, { timeout => { type=>SCALAR, optional=>1, default=>DENSE_CLI_SAY_TIMEOUT },
		       		command => { type=>SCALAR }	});
        
        if (! defined $opts{command}) {
            die("Dense::Agnostic::7mode::Node::say: need 'command' arg.");
        }

        $cmd = $opts{command};
	$timeout = $opts{timeout};

        return filer_cli_test_say($self->{fc}, $cmd, {timeout=>$timeout,retry=>'ok'});
    }

    sub verify_log (@)
    {
        my ($self, %opts, $cmd);

        $self = shift();
        %opts = validate (@_, { tail => {type=>HASHREF},
                verify_string => {type=>SCALAR|ARRAYREF}, });

        my $verify_string = $opts{verify_string};
        my $tail = $opts{tail};
        my $flag = 0;
        $tail->stop;

        my @verify_strings = ();
        my $result = $tail->to_log();
        my @result = split(/\n/, $result);

        if( ref($verify_string) eq "ARRAY" ) {
            @verify_strings = @{$verify_string};
        } else {
            push @verify_strings, $verify_string;
        }
        foreach my $str ( @verify_strings ) {
            if (grep(/$str/i, @result)) {
                print "Have obtained the expected pattern in the log file : @result \n";
                $flag = 1;
            } else {
                die "Did not get the expected output\n";
            }
        }
        if (!$flag) {
            die "Did not get the expected pattern in the log file\n";
        }

        return 1;
    }

    sub start_log (@)
    {
	my ($self, $ret, %opts);

	$self = shift();
	%opts = validate (@_, { file => {type=>SCALAR}, });

	my $fc = $self->{fc};

	if ($opts{file} =~ /sis/) {
	    $ret = LogReader::Sis->new($fc);
	}
	if ($opts{file} =~ /messages/) {
	    $ret = LogReader::Messages->new($fc);
	}

	$ret->start;

	return $ret;
    }
    sub aggr_create (@)
    {
        my  ($F, $A, %opts, %default_opts);

        $F      = shift();
        %opts   = @_;

        if ( (! defined $opts{Aggrname}) ||
             (! defined $opts{Ndisks})      ) {
            my  $emsg;
            $emsg = "Dense::Agnostic::7mode::Node::aggr_create: need Aggrname and Ndisks";
            $Log->comment($emsg);
            $Log->debug(print Dumper(\%opts));
            die($emsg);
        }   

        #Knock off CMode specific args that will only confuse us.

        $opts{Node} = $F;
        $A  = Dense::Agnostic::7mode::Aggr->new(%opts);

        return $A;
    }

    sub vserver_create
    {
        #Dummy function
        return undef;
    }

    sub vol_create  (@)
    {
        my  ($F, $V, %opts, %default_opts);

        $F      = shift();
        %opts   = @_;

        if ( (! defined $opts{Volname})  ||
             (! defined $opts{Aggr}) ||
             (! defined $opts{Size})     ) {
            my $emsg;
            $emsg = "Dense::Agnostic::7mode::Node::vol_create: need Volname, Aggr and Volsize";
            $Log->comment($emsg);
            die($emsg);
        }

        #Knock off CMode specific args that will only confuse us.
        if (exists $opts{Vserver}) {
            delete $opts{Vserver};
        }

        $V = Dense::Agnostic::7mode::Volume->new(%opts);

        return $V;
    }

    sub wreck(@)
    {
      my ($F,$ret);
      $F=shift();
      my ($inj,$pseudobad,$type,$conn,$name);
      my (%opts) = validate(@_,{name => { type=>SCALAR, optional => 0 },
                                inject_script => { type=>SCALAR, optional => 0 },
                                pseudo_bad => {type=>SCALAR, optional => 1,default => 0 },
                                type => { type => SCALAR, optional => 0, regexp => qr/aggr|flex|trad/i,},
                                }
                           );
      $conn = $F->{fc};
      $name = $opts{name};
      $inj  = $opts{inject_script};
      $type = $opts{type};
      $pseudobad = $opts{pseudo_bad};
      $ret = Wreck::wreck($conn,$name,type => $type,inject_script => $inj , pseudo_bad => $pseudobad);
      if (! $ret) {
           die("Unable to inject corruption");
       }

    }

    sub license_add(@)
    {
        my  (%opts);
        my  ($F, $cmd, $ret);

        $F = shift();
        %opts = @_;
            
        my %license_map = (
                FlexClone => "flex_clone",
                );

        if ((! defined $opts{License}) && (! defined $opts{Name})) {
            die("Dense::Agnostic::7mode::Node->license_add: Need License param or Name param");
        }

        if (defined $opts{Name}) {
            my $license = get_license_code($F->{fc}, $opts{Name});
            $Log->comment("License to be installed is $opts{Name} => $license");
            license_raw( $F->{fc}, $license, service => $opts{Name} );
        }  else { 
            $cmd = "license add $opts{License}";
            $ret = $F->say(command => $cmd);
        }
        return $ret;
    }

    sub sync(@)
    {
	my (%opts);
	my ($F, $cmd, $ret);

        my $self = shift;
	%opts = validate(@_, { no_of_syncs => { type=>SCALAR, default=>3, optional=>1 } });
	my $num_of_syncs = $opts{no_of_syncs};

	for(my $i=1; $i<=$num_of_syncs; $i++){
	    $cmd = $cmd."sync;";
	}
	$self->say(command=>$cmd);
    }

    sub filer_panic (%)
    {

        my $F = shift();
        my %opts = validate (@_, {
                delta   => {type=>SCALAR,},
                timeout   => { type => SCALAR, optional => 1, default => 5000,},} );

        #panic the filer
        panic(conn=> $F->{fc},%opts) or die "Could not panic the filer";

    }

    sub wafl_flush_cache ()
    {
        my $self = shift;
        $self->say(command=>"wafl flush cache");
    }
        

=head $F->reboot(state=>'UP')

=item state

state with which you want the filer to come up (UP, LOADER, MAINT)

Failure
fails when filer does not come up successfully to the required state. fails with die.

Success
Returns 1

=cut
	    
    sub reboot(%)
    {
	my $F = shift();
        my %opts = validate (@_, { state => { type=>SCALAR, default=>"UP"},});

	reboot_to( $F->{fc}->{hostp}, $opts{state} ) or die $@;

	return 1;
	
    }

=head $F->setup_version(relase=>$release);

=cut

sub setup_version (%)
{
    my $F = shift();

    my %opts = validate (@_, {release => {type=>SCALAR}, });
    my $host = $F->{params}{'host'};
    my $from_version = $opts{release};
    my %hash = (FILER=>$host, from_release=>$from_version);

    return get_filer_to_desired_release(%hash);
}
    
=head $F->software_transform(from_version=>$from_version, to_version=>$to_version)

=cut

sub software_transform (%)
{
    my $F = shift();

    my %opts = validate (@_, { from_version => { type=>SCALAR, optional=>1},
		    	       to_version => { type=>SCALAR},
			       } );

    my $host = $F->{params}{'host'};
    my $from_version = $opts{from_version} ? $opts{from_version} : $F->filer_version;
    my $to_version = $opts{to_version};

    print "Filer name is $host \n";

    my %hash = (FILER=>$host, from_release=>$from_version, to_release=>$to_version);

    return do_transform(\%hash);
}
    sub filer_version ()
    {
        my $F = shift();
        my $version = Dense::Utils::remove_leading_trailing_spaces($F->say(command=>"version"));
        print "Version returned is $version \n";
        my @version = split(/\s+/, $version);
        $version = $version[2];
        print "Filer ONTAP version is $version \n";
        return $version;
    }

=head $F->cf_takeover(host=>"TAKEOVER", partner=>"DEAD", wait_time=>'300')

=cut

sub cfo_takeover (%)
{
    my $F = shift();
    my %opts = validate (@_, { host => { type=>SCALAR, default=>"TAKEOVER", optional=>"1"},
		    	       partner => { type=>SCALAR, default=>"UNREACH", optional=>"1"},
			       wait_time => { type=>SCALAR, default=>'300', optional=>"1"},
			       } );
    my $src_conn = $F->{fc};
    my $partner = $src_conn->hostp()->part_host();
    my $partner_host = make_host($partner);
    my $partner_conn = make_conn($partner_host);

    cf_takeover($src_conn);

    wait_state( $src_conn, $opts{host}, $opts{wait_time} ) or die "Takeover did not succeed\n";
    eval{
    wait_state( $partner_conn, $opts{partner}, $opts{wait_time} ) or die "Takeover did not succeed\n";
    };
    if($@){
    wait_state( $partner_conn, 'DEAD', $opts{wait_time} ) or die "Takeover did not succeed\n";
    }


    return 1;
}

=head $F->cf_giveback(host=>"TAKEOVER", partner=>"DEAD", wait_time=>'300')

=cut

sub cfo_giveback (%)
{
    my $F = shift();
    my %opts = validate (@_, { host => { type=>SCALAR, default=>"UP", optional=>"1"},
		    	       partner => { type=>SCALAR, default=>"UP", optional=>"1"},
			       wait_time => { type=>SCALAR, default=>'300', optional=>"1"},
			       } );
    my $src_conn = $F->{fc};
    my $partner = $src_conn->hostp()->part_host();
    my $partner_host = make_host($partner);
    my $partner_conn = make_conn($partner_host);

    cf_giveback($src_conn);

    wait_state( $src_conn, $opts{host}, $opts{wait_time} ) or die "Takeover did not succeed\n";;
    wait_state( $partner_conn, $opts{partner}, $opts{wait_time} ) or die "Takeover did not succeed\n";

    return 1;
}
    sub fiji_set(@)
    {
        my ($F, %opts, $cmd);
        
        $F = shift();

        %opts = @_;

        my $name = delete $opts{name};

        $cmd = "fiji filter ".'"'.$opts{command}.'"';

        $F->say(command => $cmd);

    }
=head $F->fiji_clear(slot => "all")

=item slot

slot of which the fiji needs to be cleared

=cut

    sub fiji_clear() {

        my ($F, %opts, $cmd);

        $F = shift();

        (%opts) = validate(@_,
                   {slot => {type => SCALAR, default => "all", optional => 1}});

        my $slot = $opts{slot};

        if (($slot !~ /all/) && ($slot !~ /\d+/)) {
            logcomment("Invalid input to clear fiji");
        }

        $cmd = "fiji clear " . $slot;

        $F->say(command => $cmd);

    } ## end sub fiji_clear()

sub setflag (@)
{
    my $F = shift();
    my (%opts) = validate(@_, { flag => { type => SCALAR },
		    		value => { type => SCALAR },
				privmode => { type=>SCALAR, default => "test", optional => 1 },
				} );

    my $fh = $F->{fh};
    my $fc = $F->{fc};

    misc_setflag(host=>$fh, conntype=>"#rsh", flag=>$opts{flag}, value=>$opts{value}, privmode=>$opts{privmode});

    my $ret = misc_printflag( host=>$fh, conntype=>"#rsh", flag=>$opts{flag}, privmode=>$opts{privmode});
    my $int = hex($ret);
    
    if ($int != $opts{value}) {
	die "Flag is not set to expected value = $opts{value}, obtained value = $int \n";
    } else {
	return 1;
    }
}

## @name        ::snapmirror_initialize
## @arg     $fh   - required: Filer Host
## @arg     $src  - required: source path
## @arg     $dst  - required: destination path
## @description  : Initialize snapmirror relation between
##                 source and destination


          sub snapmirror_initialize ()
		     {
					 my ($F, %opts);
					 $F = shift();
					  %opts = validate( @_, {
							   'destination-path'  => { type => SCALAR },
							   'source-path'  => { type => SCALAR, },
							    });
					  my $fh = $F->{fh};
					  my $dst = $opts{'destination-path'};
					  my $src = $opts{'source-path'};
					  SnapMirror::snapmirror_initialize(host=>$fh, dst => $dst, src=>$src )
					  					or die "Snapmirror initialize failed";
			 }

## @name        ::snapmirror_update
## @arg     $fh   - required: Filer Host
## @arg     $src  - required: source path
## @arg     $dst  - required: destination path
## @description  : Update snapmirror relation between
##                 source and destination


		sub snapmirror_update ()
		  {
				  my ($F, $dst, $src, %opts);
				  $F = shift();
				  %opts = validate( @_, {
						  'destination-path'  => { type => SCALAR },
						   'source-path'  => { type => SCALAR },
						    });
				   my $fh = $F->{fh};
				   $dst = $opts{'destination-path'};
				   $src = $opts{'source-path'};
				   SnapMirror::snapmirror_update(host=>$fh, dst => $dst, src=>$src)
				   					or die "Snapmirror update failed";
		  }

## @name        ::snapmirror_break
## @arg     $fh   - required: Filer Host
## @arg     $dstpath  - required: destination path
## @description  : Break snapmirror relation between
##                 source and destination

		 sub snapmirror_break ()
		 	{
					my ($F, $dstpath,  %opts);
					$F = shift();
					(%opts) = validate( @_, {
							'destination-path' => { type => SCALAR },
							 });
					my $fh = $F->{fh};
					$dstpath = $opts{'destination-path'};
					SnapMirror::snapmirror_break(host=>$fh, dstpath=>$dstpath)
								or die "Snapmirror break failed";
			}

## @name        ::snapmirror_resync
## @arg     $fh   - required: Filer Host
## @arg     $dst  - required: destination path
## @arg     $src  - required: source path
## @description  : Resync snapmirror relation between
##                 source and destination

 
      sub snapmirror_resync ()
	  		{
					my ($F, $dst, $src, %opts);
					$F = shift();
					(%opts) = validate( @_, {
							 'destination-path' => { type => SCALAR },
							 'source-path'  => { type => SCALAR},
					});
					my $fh = $F->{fh};
					$dst = $opts{'destination-path'};
					$src = $opts{'source-path'};
					 SnapMirror::snapmirror_resync(host=>$fh, dst => $dst,src => $src)
					 						or die "Snapmirror resync failed";
			}

 sub AUTOLOAD (@)
    {
        our $AUTOLOAD;
        my  ($self, @req_params, @opt_params, %default_params);
        my  ($in_fn, %in_opts, %exec_opts, $exec_fn);

        $self       = shift();
        %in_opts    = @_;
        $in_fn      = $AUTOLOAD;
        %exec_opts  = %in_opts;

        $Log->enter();
        $in_fn =~ s/.*:://; #Chop off the "Dense::Agnostic::7mode::" prefix.

        #Resolve in_fn to the function that we should actually invoke.
        $exec_fn = $Fmap->{$in_fn}{fname};
        if (!defined $exec_fn) {
            my $emsg = "UNDEFINED: 7mode::Node::Autoload::$in_fn";
            $Log->comment($emsg);
            die($emsg);
        }

        #Resolve $exec_fn's mandatory params
        if (! defined $Fmap->{$in_fn}{req_params}) {
            $Log->debug("7mode::Node::Autoload: no mandatory args for $in_fn\[$exec_fn\]");
        } else {
            @req_params = @{$Fmap->{$in_fn}{req_params}};
            foreach my $param (@req_params) {
                if ((! $exec_opts{$param}) && (! defined $self->{params}{$param})) {
                    my $errmsg = "7mode::Node::Autoload: Can't resolve param\
                                  $param of $in_fn [$exec_fn]";
                    $Log->debug($errmsg);
                    die($errmsg);
                }

                $exec_opts{$param} = $self->{params}{$param};
            } #foreach
        }

        #Resolve $exec_fn's default params
        if (! defined $Fmap->{$in_fn}{default_params}) {
            $Log->debug("7mode::Node::Autoload: no default args for $in_fn\[$exec_fn\]");
        } else {
            %default_params = %{$Fmap->{$in_fn}{default_params}};
            foreach my $param (keys(%default_params)) {
                if (! defined $exec_opts{$param}) {
                    $exec_opts{$param} = $default_params{$param};
                }
            }
        } #else

        #All our params should now be in %exec_opts. Go forth and execute our
        #little function.
        if (%exec_opts) {
            return $exec_fn->(%exec_opts);
        } else {
            return $exec_fn->();
        }

        $Log->exit();

    } #AUTOLOAD


    #XXX TBD - Snapmirror, CFO

    return 1;
} #package Dense::Agnostic::7mode::Node
       
