{   package Dense::Agnostic::Cmode::Node;
    use strict;

    use Data::Dumper;

    use Params::Validate qw( validate SCALAR );
    use Params::Validate qw( validate SCALAR HASHREF ARRAYREF);

    use Dense::Agnostic::Constants;
    use Dense::Agnostic::Cmode::Utils;

    use Tharn   qw($Log);
    use Utils   qw(make_host);
    use Fcntl qw(:DEFAULT);
    use NACL::APISet;
    use NACL::STask::Aggregate;
    use NACL::MTask::VserverMulti;
    use NACL::C::Node;
    use NACL::APISet::Exceptions::ResponseException qw(:try);
    use NACL::Transit;
    use LogReader::Sis;
    use LogReader::EMS;
    use Cmode::License qw(get_license_code);

    sub new    (@)
    {
        my  ($class, $self, %opts, $F, $fh, $timeout);

        $class = shift();
        %opts = validate (@_, { timeout => { type=>SCALAR, optional=>1, default=>DENSE_CLI_SAY_TIMEOUT },
		       		Nodename => { type=>SCALAR }	});

        $F    = $opts{Nodename};
        $self = {};

        $fh = make_host($F);

        $self->{name} = $F;
        $self->{ci} = NACL::C::Node->find();
        $self->{api}  = NACL::APISet->new(
            hostobj   => $fh,
            category  => 'Node',
            interface => 'CLI',
            set       => 'CMode'
        );
        $self->{zapi} = NACL::APISet->new(
            hostobj   => $fh,
            category  => 'Node',
            interface => 'ZAPI',
            set       => 'CMode'
        );

        $timeout = $opts{timeout};
        $self->{api}->set_timeout( "connectrec-timeout" => $timeout );
        my $current_timeout = $self->{api}->get_timeout();
        $Log->comment("The current timeout value is $current_timeout");
	
	my $cluster_show_output = $self->{api}->cluster_identity_show();
	
	$self->{cluster_name} = $cluster_show_output->{'parsed_output'}->[0]->{'name'} ; 
	$Log->comment("The name of the cluster to which $self->{name} belongs is $self->{cluster_name}");

        bless( $self, $class );
        return $self;
    }

    sub say (@)
    {
        my  ($F);
        my  (%opts, %in_opts, $cmd);
        my  ($ret);

        $F  = shift();

        %in_opts = @_;

        if (! defined $in_opts{command}) {
            die("Dense::Agnostic::Cmode::Node::say: need 'command' arg");
        }

        foreach my $k (keys(%in_opts)) {
            if (defined $in_opts{$k}) {
                $opts{$k} = $in_opts{$k};
            }
        }

        $cmd = "\"priv set -q test; $opts{command}\"";
        $opts{command} = $cmd;
        $opts{node} = $F->{name};
        eval {
            $ret = $F->system_node_run(%opts);
        };
        if ($@) {
            die("Dense::Agnostic::Cmode::Node::say -> $@");
        }

        return $ret;
    }

    sub aggr_create (@)
    {
        my  ($F, $A, %opts, %default_opts);

        %default_opts = {
            Ndisks  => 3,
            Reuse   => 1,
        };

        $F = shift();
        %opts = @_;

        if (! defined $opts{Aggrname}) {
            die("Dense::Agnostic::Cmode::Node::aggr_create: need Aggrname arg");
        }

        $opts{Node} = $F;
        for my $k (keys %default_opts) {
            if (! defined $opts{$k}) {
                $opts{$k} = $default_opts{$k};
            }
        }
        eval {
            $A = Dense::Agnostic::Cmode::Aggr->new(%opts);
        };
        if ($@) {
            die("Dense::Agnostic::Cmode::Node::aggr_create: $@");
        }

        return $A;
    }

    sub vserver_create (@)
    {
        my  (%opts);
        my  ($F, $Vs);

        $F = shift();
        %opts   = @_;
        #Mandatory args: Vservername, Rootvol, Aggr
        if ((! defined $opts{Vservername})  ||
            (! defined $opts{Rootvolume})   ||
            (! defined $opts{Aggr}))    {
            die("Dense::Agnostic::Cmode::Node::vserver_create: Need\
                    Vservername/Rootvol/Aggr args");
        }

        $opts{Node} = $F;
        $Vs = Dense::Agnostic::Cmode::Vserver->new(%opts);
	#$Vs->sis_policy_modify( policy => "default", type => "manual", vserver => "$Vs->name");

        return $Vs;
    } #vserver_create

    sub vol_create  (@)
    {
        my  (%opts);
        my  ($F, $Vs, $A, $V);

        $F  = shift();
        %opts = @_;
        
        #Mandatory args: Volname, Vserver, Aggr, Size
        if ((! defined $opts{Volname})  ||
            (! defined $opts{Vserver})  ||
            (! defined $opts{Aggr})     ||
            (! defined $opts{Size}) ) {
            die("Dense::Agnostic::Cmode::Node->vol_create: Need\
                Volname/Vserver/Aggr/Size args");
        }

        $opts{Node} = $F;
        $V = Dense::Agnostic::Cmode::Volume->new(%opts);

        return $V;
    }

    ## @name ::vol_delete
	## @summary Deletes the volume from the filer.
	## 
	## @description
	## Unmounts, off lines, and then destroys the specified volume on the filer.
	##
	## 
	## @return None.
	##
	## @arg Volname Required: Name of volume to be deleted.
	##
	## @arg Vserver Required: Name of vserver where volume is located.
	##
	##
	## @examples
	## vol_delete( Volname => $V->{name},
	##             Vserver => $V->{Vs}->{name});
	## 
	## @author khung@netapp.com
	## @status private (in-progress)
	##
    sub vol_delete  (@)
    {
        my  (%opts);
        my  ($F, $Vs, $V);

        $F  = shift();
        %opts = @_;
        
        #Mandatory args: Volname, Vserver
        if ((! defined $opts{Volname})  ||
            (! defined $opts{Vserver}) ) {
            die("Dense::Agnostic::Cmode::Node->vol_delete: Need\
                Volname/Vserver args");
        }

        $Vs = $opts{Vserver};
        $V = $opts{Volname};

        $F->volume_unmount(vserver => $Vs, volume => $V);
        $F->volume_offline(vserver => $Vs, volume => $V);
        $F->volume_destroy(vserver => $Vs, volume => $V);
    }

	sub license_add (@)
    {
        my  (%opts);
        my  ($F, $cmd, $ret, $code, $err);

        $F = shift();
        %opts = @_;

        if ((! defined $opts{License}) && (! defined $opts{Name})) {
                die("Dense::Agnostic::Cmode::Node->license_add: Need License\
                    param or Name param");
        }
        
        if (defined $opts{Name}) {
            get_license_code(
                    feature => $opts{Name},
                    errmsg  => \$err,
                    code    => \$code) or die("get_license_code() failed, error: $err");
            $F->system_license_add('license-code' => $code);
        } else {
            $cmd = "license add $opts{License}";
            $ret = $F->system_node_run(
                            node=>$F->{name},
                            command=>$cmd,
                            );

        }
        return $ret;
    }

    sub sync (@)
    {
	my (%opts);
    	my ($F, $cmd, $ret);
        my $self = shift();

	%opts = validate(@_, { no_of_syncs => { type=>SCALAR, default=>3, optional=>1 } });
	my $num_of_syncs = $opts{no_of_syncs};

	for(my $i=1; $i<=$num_of_syncs; $i++){
	    $cmd = $cmd."sync;";
	}
        $self->say(command=>$cmd);
    }
	
	sub set_options(@)
    {
        my (%opts);
        my ($F, $cmd, $ret);
        $F = shift();
        %opts = @_;
        if ((! defined $opts{option})) {
                die("Dense::Agnostic::Cmode::Node->set_options: Need Option param");

        }
        if ((! defined $opts{value})) {
                die("Dense::Agnostic::Cmode::Node->set_options: Need Value param");

        }

        $cmd = "options $opts{Option} $opts{Value}";
        $ret = $F->system_node_run(
                node=>$F->{name},
                command=>$cmd,
                );

        return $ret;

    }
    sub fiji_set(@)
    {
        my ($F, %opts, $cmd);

        $F = shift();

        %opts = @_;

        my $name = delete $opts{name};

        $cmd = "fiji filter $opts{command}";

        $F->say(command => $cmd);

    }
=head $F->fiji_clear(slot => "all")

=item slot

slot of which the fiji needs to be cleared

=cut

    sub fiji_clear() {

        my ($F, %opts, $cmd);

        $F = shift();

        (%opts) = validate(@_,
                   {slot => {type => SCALAR, default => "all", optional => 1}});

        my $slot = $opts{slot};

        if (($slot !~ /all/) && ($slot !~ /\d+/)) {
            logcomment("Invalid input to clear fiji");
        }

        $cmd = "fiji clear " . $slot;

        $F->say(command => $cmd);

    } ## end sub fiji_clear()
    


    sub setflag (@)
    {
        my $F = shift();
        my (%opts) = validate(@_, { flag => { type => SCALAR },
                                    value => { type => SCALAR },
                                } );

	$F->say(command=>"setflag $opts{flag} $opts{value}");
	$F->say(command=>"printflag $opts{flag}");
    }
	
   sub set_metafile_visible()
   {

      my ($F,$cmd);

        $F    = shift();

        $cmd = "setflag wafl_metadata_visible 1";

        $F->say(command => $cmd);

    }

    sub filer_version ()
    {
        my $F = shift();
        my $version = Dense::Utils::remove_leading_trailing_spaces($F->say(command=>"version"));
        print "Version returned is $version \n";
        my @version = split(/\s+/, $version);
        $version = $version[2];
        print "Filer ONTAP version is $version \n";
        return $version;
    }
    
    sub software_transform (%)
    {
        my $F = shift();

        my %opts = validate (@_, { to_version => { type=>SCALAR },
                                   from_version => { type=>SCALAR, optional=>1 },
                                } );
        my $from_release = $opts{from_version} ? $opts{from_version} : $F->filer_version;
        my $to_release = $opts{to_version};

        if ($from_release ne $to_release) {
            $Log->comment("Filer current ONTAP version : $from_release");

            $Log->comment("Installing and downloafing the ONTAP release : $to_release");
            my $Node = $F->{ci};
            my $Architecture = Tharn::param("ARCHITECTURE") ? Tharn::param("ARCHITECTURE") : $::ARCHITECTURE;
            my $Package = "/x/eng/rlse/DOT/$to_release/final/bedrock/export/$Architecture.debug/tarball/image.tgz";
            my $update_obj = NACL::STask::SystemNodeImage->update(
                    command_interface => $Node,
                    node              => $Node->hostrec()->hostname(),
                    package           => $Package,
                    );
            $update_obj->wait_for_completion();

            $Log->comment("Trying to Upgrade/Revert the filer from: $from_release ---> $to_release");

            my $boot_type = Filer::Version::version_compare($from_release, $to_release);
            if ($boot_type == -1) {
                $Log->comment("Filer will be upgraded from $from_release ---> $to_release");
                my $Transit_obj = NACL::Transit->new(name => $F->{name});
                $Transit_obj->reboot();
            } elsif ($boot_type == 1) {
                $Log->comment("Filer will be reverted from $from_release ---> $to_release");
                die "Revert automation is not yet available in c-mode. Please setup your filer in $from_release, before executing the script \n";
            } else {
                die "Not a valid upgrade/revert path from $from_release ---> $to_release \n";
            }
        }
    }

    ## @name ::job_exists
	## @summary Verifies existence of the specified job.
	## 
	## @description
	## Checks that the job (specified by name) exists on the filer.
	##
	## 
	## @return None.
	##
	## @arg job Required: Name of job.
	##
	##
	## @examples
	## job_exists( job => $job );
	## 
	## @author khung@netapp.com
	## @status private (in-progress)
	##
    sub job_exists {
        my $F = shift;
        my (%args) = validate (@_, {
            job => { type => SCALAR, }, 
        } );
        my $name = $args{job};
        my $ret;

		# Set privilege to test
		$F->set_privilege( level => "test" );

		try {
			my @output = $F->job_show( 'name' => $name );
			$Log->debug( print Dumper(@output) );

			# Set privilege back to admin
			$F->set_privilege( level => "admin" );

			# Verify that job exists
			if ( ( scalar @output ) ) {
				if ( ( scalar @output ) == 1 ) {

					# Check that correct job was returned
					if ( !( $name eq $output[0]->{name} ) ) {
						$Log->log("Obtained job, $output[0]->{name}, does not match expected job, $name");
						$ret = FALSE;
					}
				}
				else {
					$Log->log("Expecting only one job entry, but multiple entries were returned");
					$ret = FALSE;
				}
			}

            $Log->log("Job $name exists.");
            $ret = TRUE;

		}
		catch NACL::APISet::Exceptions::ResponseException with {
			my ($exception);
            $exception = shift;
            $Log->log("Job $name does not exist.");
            $Log->log( $exception->text() );
            $ret = FALSE;
		};
        $Log->debug("return : $ret");

        return $ret;
	}

    ## @name ::set_privilege
	## @summary Sets the privilege level on the node.
	## 
	## @description
	## Sets the privilege to the specified level.
	##
	## 
	## @return None.
	##
	## @arg level Required: Privilege level (admin, advanced, diagnostic, test)
	##
	## @examples
	## set_privilege( level => "diagnostic" );
	## 
	## @author khung@netapp.com
	## @status private (in-progress)
	##
    sub set_privilege (@) {
        my $F = shift;
        my (%args) = validate (@_, {
            level => { type => SCALAR, }, 
        } );
        my $level = $args{level};

        $Log->log("Setting privilege to $level");
        $F->set( privilege => $level );
    }
    
    sub cfo_takeover (@)
    {
	my ($F);
	$F = shift();

	my $server = new Server();
	my $partner = $F->storage_failover_show(node=>$F->{name}, fields=>'partner-name');
	$partner = $partner->{'partner-name'};
	$F->storage_failover_modify(node=>$F->{name},enabled=>"true");
	my $state_info = $F->say(command=>"cf status");
	$F->storage_failover_takeover(ofnode=>$partner, option=>'normal');
	# Wait 5 minutes for both the nodes to go to takeover state and giveback state respectively
	my $timeout = 600;
	while($timeout > 0) {
		my $state_info = $F->say(command=>"cf status");
		if ($state_info =~ /$F->{name} has taken over $partner/) {
			last;
		} else {
			$timeout--;
			sleep 10;
		}
	}

	if($timeout == 0){
		die "CF takeover did not succeed\n";
	} else {
		$Log->comment("CF takeover was successful");
	}
	return 1;
    }

    sub cfo_giveback (@)
    {
	    my ($F);
	    $F = shift();

	    my $partner = $F->storage_failover_show(node=>$F->{name}, fields=>'partner-name');
	    $partner = $partner->{'partner-name'};
	    $F->storage_failover_modify(node=>$F->{name}, enabled=>"true");
	    my $state_info = $F->say(command=>"cf status");
	    $F->storage_failover_giveback(ofnode=>$partner);
	    # Wait 10 minutes for both the nodes to go to UP state
	    my $timeout = 600;
	while($timeout > 0) {
		my $state_info = $F->say(command=>"cf status");
		if ($state_info =~ /Controller Failover enabled, $partner is up/) {
			last;
		} else {
			$timeout--;
			sleep 10;
		}
	}
	if($timeout == 0){
		die "CF giveback did not succeed\n";
	} else {
		$Log->comment("CF giveback was successful");
	}
	return 1;
    }

    sub verify_log (@)
    {
	    my ($self, %opts, $cmd);

	    $self = shift();
	    %opts = validate (@_, {tail => {type=>HASHREF},
	    	       verify_string => {type=>SCALAR|ARRAYREF}, });

	    my $tail = $opts{tail};
	    my $verify_string = $opts{verify_string};

	    $tail->stop;
	    my $flag = 0;
	    
            my $result = $tail->to_log();
	    my @verify_strings = ();
	    
	    my @result = split(/\n/, $result);
	    if( ref($verify_string) eq "ARRAY" ) {
		@verify_strings = @{$verify_string};
    	    } else {
		push @verify_strings, $verify_string;
	    }
        
            foreach my $str ( @verify_strings ) {
	        if (grep(/$str/, @result)) {
	            $Log->comment("Have obtained the expected pattern in the log 
                        file : $str");
		        $flag = 1;
	        } else {
                print Dumper(\@result);
	            die "Did not get expected log for : $str \n";
	        }
            }
	    if (!$flag) {
		die "Did not get the expected pattern in the log file\n";
	    }

	    return 1;
    }

    sub start_log (@)
    {
	    my ($self, $ret, %opts);

	    $self = shift();
	    %opts = validate (@_, { file => {type=>SCALAR}, });

	    my $server = Server->new();
	    if ($opts{file} =~ /sis/) {
	        $ret = LogReader::Sis->new( $server );
	    }
	    if ($opts{file} =~ /messages/) {
	        $ret = LogReader::Messages->new( $server );
	    }
	    if ($opts{file} =~ /ems/) {
	        $ret = LogReader::EMS->new( $server );
	    }

	    # Start recording sis log file
	    $ret->start;

	    return $ret;
    }
    
    sub reboot(%)
    {
	    my ($F, $Transit_obj, $filer, $ret);

	    $F = shift();
	    $filer = $F->{name};

	    $Transit_obj = NACL::Transit->new(name => $filer);
	    my $curr_state = $Transit_obj->get_state();
	    $Log->debug("Before reboot : $curr_state");
	
	    # AUTOBOOT might be FALSE. To take care of it, first transit to "FIRMWARE" state
	    $Transit_obj->change_state(to=>"FIRMWARE");
	    $Log->comment("Current state is $Transit_obj->get_state()");
	    if ($Transit_obj->get_state() !~ /FIRMWARE/) {
	        die "Filer did not get into expected state \n";
	    }

	    # From "FIRMWARE" state transit to original state
	    $Transit_obj->change_state(to=>$curr_state);
	    my $after_state = $Transit_obj->get_state();
	    $Log->comment("After reboot : $after_state");

	    $F->{api}->api_reconnect_session() if !$F->{api}->api_is_session_alive();

	    if ($after_state !~ $curr_state) {
	        die "Did not reboot properly, Original state : $curr_state. Current state : $after_state \n";
	    }

	    my $prompts = ['.*Do you want to continue.*'=>'y'];
	    my $response = $F->{api}->set('privilege'=>'test','connectrec-match_table' => $prompts);
	    
	    return 1;
    }

    sub registry_get(@) {
        my ( $F, %opts );
        $F = shift();
        %opts = validate( @_, { key => { type => SCALAR } } );

        my ( $key, $cmd, $retstr, $rval );
        $key = $opts{'key'};
        $cmd = "registry get $key";
        $Log->comment("Getting value of $key");
        $retstr = $F->say( command => $cmd );
        $rval = ( split( /=/, $retstr ) )[-1];
        $rval = Dense::remove_leading_trailing_spaces($rval);

        return $rval;
    }

    sub registry_set(@) {
        my ( $F, %opts );
        $F = shift();
        %opts = validate( @_, {
                key    => { type => SCALAR },
                value  => { type => SCALAR },
                verify => { type => SCALAR, optional => 1, default => 1 },
            });

        my ( $key, $sval, $cmd, $rval );
        $key  = $opts{'key'};
        $sval = $opts{'value'};
        $cmd  = "registry set $key $sval";
        $Log->comment("Setting value of $key to $sval");
        $F->say( command => $cmd );
        $F->sync();
        if ( $opts{'verify'} ) {
            $Log->comment( "verifying if the value of $key has been set successfully");
            $rval = $F->registry_get( key => $key );
            if ( "$rval" ne "$sval" ) {
                die("The value of $key has not been set successfully, expected: $sval, actual: $rval");
            } else {
                $Log->comment( "The value of $key has been set successfully");
            }
        }
        return $sval;
    }


    sub get_date (@) {
        my (%opts);
        my ( $F, $Vs, $A, $V, $output, $ret );

        $F    = shift();
        %opts = @_;

        $output = $F->date(%opts);

        if ( ref($output) eq "HASH" ) {
            $Log->comment("Date is a hasref");
            #The below line is meant for backward comatablity
            $ret = $output->{'Node      UTC Date'};
        } else {
            if ( $output =~ /(\w+\s+\w+\s+\d+\s+\d+:\d+:\d+\s+\d+)/ ) {
                $ret = $1;
                $Log->comment("Date parsed: $ret");
            } else {
                $Log->comment( "Date error:\n" . Dumper($output) );
                die('Date could not be parsed as it was not in the expected format');
            }
        }
        return $ret;
    }

############################################
## Description:
##      vserver_peer_capability_enable function is used to enable vserver peer capability on the cluster
## Input params:
##      None
## Return Value
##      None
############################################

    sub vserver_peer_capability_enable (@)
    {
        my  ($F, $Node, @ret);

        $F = shift();
        $Node = $F->{name};

        #Set privilege level to test
        $F->set_privilege( level => "test" );
        # Check capability of the node
        @ret = $F->system_capability_show('node' => $Node, 'capability' => "vserver_peer");
        # For all the nodes in the cluster check whether vserver peer capability is enabled or not
        for my $h (@ret) {
        # If capability is not enabled enable it
            if ($h->{'enabled'} eq "false") {
                $F->system_capability_enable('cap-in' => "vserver_peer", 'all-or-none' => "true", 'cap-out' => "vserver_peer");
                $Log->comment( "vserver peer capability is enabled for node $h->{'node'}");
            }
            elsif ($h->{'enabled'} eq "true") {
                   $Log->comment( "vserver peer capability is already enabled for node $h->{'node'}");
            }
        }
    } #vserver_peer_capability_enable

############################################
## Description:
##      vserver_peer_create_wrapper function is used to create vserver peer on the node
## Input params:
##      Vserver => vserver name
##      VserverPeer => peer vserver name
##      Application => Application type like snapmirror or flexcache
## Return Value
##      None
############################################

    sub vserver_peer_create_wrapper (@)
    {
        my  (%opts);
        my  ($F, $peernode, $src_vserver, $peer_vserver, @ret, $flag);

        $F = shift();
        %opts = validate (@_, { Vserver => { type => SCALAR },
                                VserverPeer => { type => SCALAR },
                                Application => { type => SCALAR } });
        $src_vserver = $opts{Vserver};
        $peer_vserver = $opts{VserverPeer};

        #Check whether vserver peer capability is enabled or not
        $F->vserver_peer_capability_enable();
        @ret = $F->vserver_peer_show_all();

        $flag = 0;
        #if source and peer vservers are same then vserver peer cannot be created, else create vserver peer
        if ($src_vserver eq $peer_vserver) {
            die ("peer vserver & vserver should be different");
        }
        else {
            #if vserver peer is already created don't create else create it
            for my $h (@ret) {
                if (ref ($h) eq "HASH") {
                    if (($h->{'vserver'} eq $src_vserver && $h->{'peer-vserver'} eq $peer_vserver) || ($h->{'vserver'} eq $peer_vserver && $h->{'peer-vserver
'} eq $src_vserver) && $h->{'peer-state'} eq "peered") {
                        $Log->comment( "vserver peer is already created between $src_vserver & $peer_vserver");
                        $flag = 1;
                        last;
                    }
                }
            }
            if ($flag == 0) {
                $F->vserver_peer_create('vserver' => $src_vserver, 'peer-vserver' => $peer_vserver, 'applications' => $opts{Application});
                $Log->comment("vserver peer created between $src_vserver & $peer_vserver");
            }
        }
    } #vserver_peer_create_wrapper

############################################
## Description:
##      vserver_peer_accept_wrapper function is used to accept vserver peer relation on the second node in inter cluster
## Input params:
##      PeerNode => Peer Node object
##      Vserver => vserver name
##      VserverPeer => peer vserver name
## Return Value
##      None
############################################

    sub vserver_peer_accept_wrapper (@)
    {
        my  (%opts);
        my  ($F, $peernode, $cluster_src, $cluster_dst, $src_vserver, $peer_vserver);

        $F = shift();
        %opts = validate (@_, { PeerNode => { type => HASHREF },
                                Vserver => { type => SCALAR },
                                VserverPeer => { type => SCALAR } });
        $peernode = $opts{PeerNode};
        $cluster_src = $F->{cluster_name} ;
        $cluster_dst = $peernode->{cluster_name} ;
        $src_vserver = $opts{Vserver};
        $peer_vserver = $opts{VserverPeer};

        #if source and peer clusters are same then vserver peer accept fails, else do vserver peer accept
        if ($cluster_src eq $cluster_dst) {
            $Log->comment ("vserver peer accept can be done on inter cluster vservers only");
        }
        else {
            $F->vserver_peer_accept('vserver' => $src_vserver, 'peer-vserver' => $peer_vserver);
            $Log->comment ("vserver peer accept between $src_vserver and $peer_vserver is successful");
        }
    } #vserver_peer_accept_wrapper

############################################
## Description:
##      dedupe_lrse_prereq function is used to accept vserver peer relation on the second node in inter cluster
## Input params:
##      PeerNode => Peer Node object
##      Vserver => vserver name
##      VserverPeer => peer vserver name
##      VserverAggr => Aggr for creating root volume of vserver
##      VserverPeerAggr => Aggr on peer node for creating root volume of vserver
##      Application => Application type like snapmirror or flexcache
## Return Value
##      None
############################################

    sub dedupe_lrse_prereq (@)
    {
        my  (%opts);
        my ($node, $peernode);

        $node = shift();
        %opts = validate (@_, { PeerNode => { type => HASHREF },
                                Vserver => { type => SCALAR },
                                VserverPeer => { type => SCALAR },
                                VserverAggr => { type => SCALAR },
                                VserverPeerAggr => { type => SCALAR },
                                Application => { type => SCALAR } });
        $peernode = $opts{PeerNode};

        #Set boot arg snapmirror.sn_enabled on the filer if it is not already set
        #$node->set_lrse_bootarg(PeerNode => $peernode);
        #Create multiple vservers
        $peernode->create_multi_vservers(PeerNode => $node, Vserver => $opts{Vserver}, VserverPeer => $opts{VserverPeer}, VserverAggr => $opts{VserverAggr}, VserverPeerAggr => $opts{VserverPeerAggr});
        #Create vserver peer on one of the node
        $node->vserver_peer_create_wrapper(Vserver => $opts{Vserver}, VserverPeer =>  $opts{VserverPeer}, Application => $opts{Application});
        #Accept vserver peer relationship on another node
        $peernode->vserver_peer_accept_wrapper(PeerNode => $node, Vserver => $opts{Vserver}, VserverPeer =>  $opts{VserverPeer});
    } #dedupe_lrse_prereq


############################################
## Description:
##      create_multi_vservers function is used to multiple vservers on the node
## Input params:
##      PeerNode => Peer Node object
##      Vserver => vserver name
##      VserverPeer => peer vserver name
##      VserverAggr => Aggr for creating root volume of vserver
##      VserverPeerAggr => Aggr on peer node for creating root volume of vserver
## Return Value
##      None
############################################

    sub create_multi_vservers (@)
    {
        my (%opts);
        my ($node, $ci, $F, $peernode, $vsflag, $inaggr, $outaggr, $invs, $iter);

        $node = shift();
        %opts = validate (@_, { PeerNode => { type => HASHREF },
                                Vserver => { type => SCALAR },
                                VserverPeer => { type => SCALAR },
                                VserverAggr => { type => SCALAR },
                                VserverPeerAggr => { type => SCALAR } });
        $ci = NACL::C::Node->find();
        $peernode = $opts{PeerNode};

        if ($opts{Vserver} eq $opts{VserverPeer}) {
            die ("source vserver and destination vserver are same");
        }

        for ($iter = 1; $iter < 3; $iter++) {
            $F = ($iter == 1) ? $node : $peernode;
            $invs = ($iter == 1) ? $opts{Vserver} : $opts{VserverPeer};
            $inaggr = ($iter == 1) ? $opts{VserverAggr} : $opts{VserverPeerAggr};
            $vsflag=0;
            for my $vs ( $F->vserver_show() ) {
            #Check if vserver is existing if not create it
                if ($vs->{vserver} eq $invs) {
                    $Log->comment("Vserver $invs already exists");
                    $vsflag = 1;
                }
            }
            if ($vsflag == 0) {
                #Check if aggr is existing if not create it
                try {
                    $outaggr = NACL::STask::Aggregate->find(
                        command_interface => $ci,
                        filter            => { aggregate => $inaggr },
                    );
                } ## end try
                catch NACL::Exceptions::NoElementsFound with {
                    $F->aggr_create(Aggrname => $inaggr,
                               Ndisks => 3,
                               Reuse => 1);
                };
                #Create vserver
                create_nate_host_file ($invs);
                my @vsvr_names = NACL::MTask::VserverMulti->create_multi(
                    command_interface           => $inaggr->command_interface(),
                    aggregate                   => $inaggr->aggregate(),
                    vserver                     => $invs,
                    nacltask_modify_root_vol    => 1,
                );
            }
        }
    } #create_multi_vservers

############################################
## Description:
##      create_nate_host_file function is used to create nate host file for the vserver
## Input params:
##      vserver name
## Return Value
##      None
############################################

        sub create_nate_host_file
        {
           my ($vsvr_name) = @_;
           my $Nate_Hosts_Dir = "/tmp";
           our $NATE_HOSTS;
           $NATE_HOSTS = $Nate_Hosts_Dir . ';' . $NATE_HOSTS;
           my $fh;
           my $filename = "$Nate_Hosts_Dir/$vsvr_name";

           $Log->comment("Creating nate host file vserver creation");
           if (!sysopen($fh, $filename, O_WRONLY | O_CREAT)) {
                NATE::BaseException->throw("Cannot open NATE host file $filename: $!");
           }

           truncate($fh, 0);
           print $fh 'hosttype=filer-ng-vserver', "\n";
           print $fh 'DATA_IP1=10.0.0.2', "\n";
           print $fh 'DATA_MASK1=255.255.252.0', "\n";
           print $fh 'DATA_PORT1=e0c', "\n";
           print $fh 'default_username=vsadmin', "\n";
           print $fh 'default_password=netapp1!', "\n";
           close($fh);
        } #create_nate_host_file

############################################
## Description:
##      remove_nate_host_file function is used to cleanup nate host file created during
##      the vserver creation
## Input params:
##      vserver name
## Return Value
##      None
############################################

        sub remove_nate_host_file
        {
           my ($vsvr_name) = @_;
           my $Nate_Hosts_Dir = "/tmp";

           my $filename = "$Nate_Hosts_Dir/$vsvr_name";
           $Log->comment("Deleting nate host file created for vserver creation");
           if ( -e $filename ) {
                unlink($filename);
           }
        } #remove_nate_host_file

############################################
## Description:
##      set_lrse_bootarg function is used to set lrse boot arg if not set
## Input params:
##      PeerNode => Peer Node object
## Return Value
##      None
############################################

        sub set_lrse_bootarg
        {
           my (%opts);
           my ($F, $Node, $peernode, $out, $iter, $no_of_iters);

           $Node = shift();
           %opts = validate (@_, { PeerNode => { type => HASHREF } });
           $peernode = $opts{PeerNode};

           #If Node and peer node are same then check the boot arg on only one node
           if ($Node -> {name} eq $peernode -> {name}) {
                $no_of_iters = 1;
           }
           else {
                $no_of_iters = 2;
           }
           for ($iter = 0; $iter < $no_of_iters; $iter++) {
                # In the first instance check node and then check peer node
                $F = ($iter == 0) ? $Node : $peernode;
                # Dump bootargs
                $out = $F -> say(command => 'bootargs get snapmirror.sn_enabled');
                $out =~ s/^\s+|\s+$//g;
                print Dumper ($out);
                #check if snapmirror.sn_enabled bootarg is set on the filer or not
                if ($out eq 'snapmirror.sn_enabled == true') {
                    $Log->comment ("boot arg snapmirror.sn_enabled is already set on the filer");
                }
                else {
                    $out = $F -> say(command => 'bootargs set snapmirror.sn_enabled true');
                    $out =~ s/^\s+|\s+$//g;
                    print Dumper ($out);
                    #check if snapmirror.sn_enabled bootarg is set on the filer or not
                    if ($out eq 'snapmirror.sn_enabled == true') {
                        $Log->comment ("boot arg snapmirror.sn_enabled is set on the filer");
			            $F -> reboot();
                    }
		            else {
                    	die ("Unable to set boot arg snapmirror.sn_enabled on the filer before running LRSE testcases");
		            }
                }
           }
        } #set_lrse_bootarg

 sub check_debug()
 {
	my $self = shift();
        my $Version_manager = NACL::APISet::Version->new(hostobj   => $self->{api}->{hostobj},
                                                     category  => "Node",
                                                     shell_type => 'CMode' 
                                                     );
        return $Version_manager->is_debug(); 
 }
 		
     sub AUTOLOAD (@)
     {
         our $AUTOLOAD;
         my  ($F, $ret, $in_fn, %opts);
 #        my  ($ignore_wantarray);
         my  ($exception_state, $expected_message);
 
         $F = shift();
         %opts = @_;
         $in_fn = $AUTOLOAD;
        $in_fn =~ s/.*:://; #Chop off the Dense::Agnostic::Cmode

        # Process exception parameters
        if ( defined $opts{exception_state} ) {
            $exception_state = delete $opts{exception_state};
        } else {
            $exception_state = DENSE_EXPECT_SUCCESS;
        }
        if ( defined $opts{expected_message} ) {
            $expected_message = delete $opts{expected_message};
		}
        $Log->comment("Cmode::Node::AUTOLOAD: exception_state == $exception_state");
        $Log->comment("Cmode::Node::AUTOLOAD: expected_message == $expected_message");

        $Log->debug(print ">>> Cmode::Node::AUTOLOAD: executing $in_fn <<<\n");
        $Log->debug(print ">>> Cmode::Node::AUTOLOAD:", Dumper(\%opts));

        # Catch exception and do one of the following :
        # 1. Re-throw exception if expected success of command
        # 2. Return success if expected failure of command and exception was caught
        # 3. Compare expected exception message with obtained exception message
        #
        # arg exception_state Optional: Expected command state. Default is DENSE_EXPECT_SUCCESS.
        # arg expected_message Optional: Expected message to compare against obtained message.
        #
        # If exception_state is DENSE_EXPECT_SUCCESS, then any caught exception will be re-thrown to be 
        # handled by the caller.
        # If exception_state is DENSE_IGNORE_FAILURE only, then return true if an exception is caught.
        # If exception_state is DENSE_EXPECT_FAILURE and expected_message is specified, message will be 
        # validated. If messages are the same, return true. Else, return false.
        
        
        try {
            if ( $opts{zapi} ) {
                delete $opts{zapi};
                if ( $opts{vserver} ) {
                    $F->{zapi}->set_vserver( "vserver-name" => $opts{vserver} );
                    delete $opts{vserver};
                }
                $ret = $F->{zapi}->$in_fn(%opts);
            } else {
                $ret = $F->{api}->$in_fn(%opts);
            }

        } catch NACL::APISet::Exceptions::ResponseException with {
              $Log->debug("Inside catch....");
              my $exception = shift;
              if ( $exception_state == DENSE_EXPECT_FAILURE ) {
                  $Log->comment("Cmode::Node::AUTOLOAD: DENSE_EXPECT_FAILURE");
                  $ret = TRUE;
                  if ( defined( $expected_message ) ) {
                      $Log->debug("Cmode::Node::AUTOLOAD: EXPECTED_MESSAGE is \"$expected_message\"");
                      $ret = validate_message($expected_message, $exception->text());
                  }
                  $Log->debug("Result : $ret");
              } 
              elsif ( $exception_state == DENSE_IGNORE_FAILURE ) {
                  $Log->comment("Cmode::Node::AUTOLOAD: DENSE_IGNORE_FAILURE");
                  $ret = TRUE;
                  $Log->debug("Result : $ret");
              }
              else {
                  $Log->comment("Cmode::Node::AUTOLOAD: DENSE_EXPECT_SUCCESS");
                  $exception->throw();
              }
              $Log->debug("Exception Message : $exception->text()");
              goto EXCEPTION;
        };
OUT:
        if ($in_fn eq 'system_node_run') {
            return $ret->{processed_output};
        }

        # Not all commands populate the parsed_output field.
        if ( !( scalar @{$ret->{parsed_output}} ) ) {
            return $ret->{output};
        }

        return(wantarray() ? @{$ret->{parsed_output}} : $ret->{parsed_output}->[0]);

EXCEPTION:
        return $ret;

    } #AUTOLOAD

    return 1;
}
