#
# $Id: //depot/prod/test/Rsierranevada/lib/Dense/Agnostic/Cmode/Vserver/Sis_Policy.pm#1 $
#
# Copyright (c) 2008 NetApp,  Inc.
# All rights reserved.

## @summary Dedupe C-mode volume efficiency policy module
##
## @description
##    This module provides sis policy utility functions
##
## @synopsis
##
## use Dense::Agnostic::Cmode::Vserver::Sis_Policy;
## 
## 
## @author khung@netapp.com
## @status private (in-progress)

{ package   Dense::Agnostic::Cmode::Vserver;
    use strict;
    use warnings;

    use Tharn   qw($Log);

    use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
    use Data::Dumper;

    use Dense::Agnostic::Constants;
    use Dense::Agnostic::Cmode::Utils;
    use NACL::APISet::Exceptions::ResponseException qw(:try);

	## @name ::sis_policy_exists
	## @summary Verifies that specified policy exists
	## 
	## @description
	## Verifies that specified policy exists on filer.
	##
	## 
	## @return TRUE/FALSE.
	##
	## @arg policy Required: Policy name.
	##
	## @arg vserver Required: Vserver name.
	##
	##
	## @examples
	## my $result = sis_policy_exists( $Vs->{name}, $policy );
	## 
	## @author khung@netapp.com
	## @status private (in-progress)
	##
    sub sis_policy_exists {

        my $Vs = shift;
        my (%args) = validate (@_, {
            policy    => { type => SCALAR, }, 
            vserver   => { type => SCALAR },
        } );

        my $policy = $args{policy};
        my $vserver = $args{vserver};
        my $ret = TRUE;

        try {
            my @output = $Vs->sis_policy_show( 'vserver' => $vserver, 'policy' => $policy );
		    $Log->debug( print Dumper(@output) );

		    if ( ( scalar @output ) ) {
                if ( ( scalar @output ) == 1 ) {
                    # Check that correct policy was returned
				    if ( !($policy eq $output[0]->{policy})) {
					    $Log->log("Obtained policy, $output[0]->{policy}, does not match expected policy, $policy");
					    $ret = FALSE;
				    }
			    } 
			    else {
                    $Log->log("Expecting only one policy entry, but multiple entries were returned");
                    $ret = FALSE;
			    }
            }
            $Log->log("Found policy $policy for vserver $vserver.");

        } catch NACL::APISet::Exceptions::ResponseException with {
            my ($exception);
            $exception = shift;
            $Log->log("Policy $policy for vserver $vserver was not found.");
            $Log->log( $exception->text() );
            $ret = FALSE;
        };
		
		return $ret;
	}

    sub sis_policy_sched_create {
        my $Vs = shift;
        my (%args) = validate( @_, {
                policy_name => { type => SCALAR, optional => 1 },
                policy_dur_hr => { type => SCALAR, optional => 1, default => 0, regex => qr/^(\d+)$/ },
                policy_dur_min => { type => SCALAR, optional => 1, default => 0, regex => qr/^(\d+)$/ },
                policy_enabled => { type => SCALAR, optional => 1, default => 'true', regex => qr/^(true|false)$/ },
                sched_name  => { type => SCALAR, optional => 1 },
                min         => { type => SCALAR, optional => 1, default => 0, regex => qr/^(0|1)$/ },
                min_start => { type => SCALAR, optional => 1, default => 0, regex => qr/^(\d+)$/ },
                min_end => { type => SCALAR, optional => 1, default => 60, regex => qr/^(\d+)$/ },
                min_freq => { type => SCALAR, optional => 1, default => 1, regex => qr/^(\d+)$/ },
                min_raw => { type => SCALAR, optional => 1 }, hr => { type => SCALAR, optional => 1, default => 0, regex => qr/^(0|1)$/ },
                hr_start => { type => SCALAR, optional => 1, default => 0, regex => qr/^(\d+)$/ },
                hr_end => { type => SCALAR, optional => 1, default => 23, regex => qr/^(\d+)$/ },
                hr_freq => { type => SCALAR, optional => 1, default => 1, regex => qr/^(\d+)$/ },
                hr_raw => { type => SCALAR, optional => 1 },
                day    => { type => SCALAR, optional => 1, default => 0, regex => qr/^(0|1)$/ },
                day_start => { type => SCALAR, optional => 1, default => 0, regex => qr/^(\d+)$/ },
                day_end => { type => SCALAR, optional => 1, default => 7, regex => qr/^(\d+)$/ },
                day_freq => { type => SCALAR, optional => 1, default => 1, regex => qr/^(\d+)$/ },
                day_raw => { type => SCALAR, optional => 1 },
            });

        my ($policy_name, $sched_name, $min, $hr, $day);
        my ($F, %sched, %policy, %op, $rand);

        $F    = $Vs->{F};
        $rand = int( rand(1000000) );

        if ($args{'policy_name'}) {
            $policy_name = $args{'policy_name'};
        } else {
            $policy_name = 'sis_policy_'.$rand;
        }
        $op{'policy_name'} = $policy_name;

        if ($args{'sched_name'}) {
            $sched_name = $args{'sched_name'};
        } else {
            $sched_name = 'sis_sched_'.$rand;
        }
        $op{'sched_name'} = $sched_name;
        $sched{'name'} = $sched_name;

        if ( $args{'min'} ) {
            if ( $args{'min_raw'} ) {
                $min = $args{'min_raw'};
            } else {
                $min = _get_str_from_range( $args{'min_start'}, $args{'min_end'}, $args{'min_freq'} );
            }
        } else {
            $min = 0;
        }
        $sched{'minute'} = $min;

        if ( $args{'hr'} ) {
            if ( $args{'hr_raw'} ) {
                $hr = $args{'hr_raw'};
            } else {
                $hr = _get_str_from_range( $args{'hr_start'}, $args{'hr_end'}, $args{'hr_freq'} );
            }
            $sched{'hour'} = $hr;
        }

        if ( $args{'day'} ) {
            if ( $args{'day_raw'} ) {
                $day = $args{'day_raw'};
            } else {
                $day = _get_str_from_range( $args{'day_start'}, $args{'day_end'}, $args{'day_freq'} );
            }
            $sched{'day'} = $day;
        }

        my $sched_op = $Vs->job_schedule_cron_create(%sched);
        $Log->debug('sched_op: '.Dumper($sched_op));

        %policy = (
            'vserver'  => $Vs->{'name'},
            'policy'   => $policy_name,
            'schedule' => $sched_name,
            'enabled'  => $args{'policy_enabled'},
        );

        if ( $args{'policy_dur_hr'} ) {
            $policy{'duration'} = $args{'policy_dur_hr'};
        }

        if ( $args{'policy_dur_min'} ) {
            $policy{'duration-mins'} = $args{'policy_dur_min'};
        }

        my $pol_op = $Vs->volume_efficiency_policy_create(%policy);
        $Log->debug('pol_op: '.Dumper($pol_op));

        return \%op;
    }

    sub _get_str_from_range($$$) {
        my ($start, $end, $freq) = @_;
        my ($i, $op);

        for ( $i = $start ; $i < $end ; $i+=$freq ) {
            $op .= $i . ',';
        }
        chop($op);

        return $op;
    }
     
    return 1;
}
