{ package Dense::Clog;

    BEGIN   {
        use strict;
        use warnings;

        use Data::Dumper;

        use base    qw(
                Exporter
                );

        use filer   qw(
                filer_cli_say
                filer_cli_test_say
                );

        use Dense   qw(:all);

        use Dense::Utils    qw(
                dense_new_filer
                dense_new_vol
                dense_destroy_vol
                dense_cleanup_vol
                );

        use Utils::Units    qw(
                convert_to_bytes
                );


        my  @clog_exports = qw(
                printflag
                setflag
                slist
                ls
                is_clog_in_vvol
                get_n_clog_buckets
                get_vvol_fingerprint_count
                );

        our %EXPORT_TAGS    = (
                all    =>  [ @clog_exports ],
                );
        Exporter::export_ok_tags('all');
    } #BEGIN



    use constant    {
        SUCCESS => 1,
        FAILURE => 0,
    };

    use constant    {
        ENOFLAG     => 0xDEADBEEF,
        EBADVAL     => 0xBEADBEAD,
        ENOTDIR     => 0xABBAABBA,
    };


    use constant VVOL_FPDB_REC   => 'Checking fingerprint.vvol\s+...\s+(\d+)\s+records';

    sub _atoi
    {
        my  ($d, $out);

        $out = 0;
        for $d (shift()) {
            $out = $out * 10 + $d;
        }

        return($out);
    } #_atoi

    sub _trim   (@)
    {
        my  ($idx, $string);
        my  (@out);

        @out    = @_;
        for $idx (0..$#out) {
            $string = $out[$idx];
            $string =~ s/^\s+//;
            $string =~ s/\s+$//;
            $out[$idx] = $string;
        }

        return(@out);
    } #_trim

    sub printflag   ($$)
    {
        my  ($fc, $flag);
        my  ($out, $retval);
        my  ($value);

        ($fc, $flag) = @_;

        ($out, $retval) = filer_cli_test_say($fc, "printflag $flag");
        if ($out =~ /no such flag/) {
            $value = ENOFLAG;
        } else {
            ($flag, $value) = _trim(split("=", $out));
        }

        return($value);
    } #printflag

    sub setflag     ($$$)
    {
        my  ($fc, $flag, $value);
        my  ($out, $retval, $checkval);

        ($fc, $flag, $value) = @_;

        $out = printflag($fc, $flag);
        if (&ENOFLAG == $out) {
            return ENOFLAG;
        }

        ($out, $retval) = filer_cli_test_say($fc, "setflag $flag $value");

        $checkval   = printflag($fc, $flag);
        if ($checkval != $value) {
            return EBADVAL;
        }

        return SUCCESS;
    }

    sub slist   ($$)
    {
        my  ($fc, $path);
        my  ($retval, @lines, $out, @keys, @values);
        my  ($slist_info);

        ($fc, $path) = @_;

        ($out, $retval) = filer_cli_test_say($fc, "slist $path");
        if ($out =~ /Invalid path/) {
            return FAILURE;
        }

        @lines  = split(/\n/, $out);
        @values = _trim(split(/\s+/, $lines[-1]));
         #the information of our interest (inum, size, path) is contained in the
         #last line.

        #XXX: $values[0] is '' thanks to the split - no idea why.

        $slist_info->{inum} = _atoi($values[1]);
        $slist_info->{size} = _atoi($values[2]);
        $slist_info->{path} = $values[3];

        return($slist_info);
    } #slist()


    sub ls      ($$)
    {
        my  ($fc, $dirpath);
        my  (@ls_out, $out, $retval);

        ($fc, $dirpath) = @_;

        ($out, $retval) = filer_cli_test_say($fc, "ls $dirpath");
        if ($out =~ /Can't open directory/) {
            return ENOTDIR;
        }

        @ls_out = _trim(split(/\n/, $out));
        
        #knock off the entries for "." and "..". Keep in mind that split
        #introduces a ''. XXX: WHY?
        return(splice(@ls_out, 3));
    }



    sub is_clog_in_vvol ($$)
    {
        my  ($fc, $volname);
        my  ($out, $clog0, $clog1, $slist_info);

        ($fc, $volname) = @_;

        $clog0 = "/vol/$volname/metadir/dense/clogdir.0/bucket0";
        $clog1 = "/vol/$volname/metadir/dense/clogdir.1/bucket0";

        $out = setflag($fc, "wafl_metadata_visible", 1);
        if ($out != SUCCESS) {
            logcomment("unable to setflag wafl_metadata_visible to 1");
            return FAILURE;
        }

        $slist_info = slist($fc, $clog0);
        if ($slist_info == FAILURE) {
            logcomment("$clog0 does not exist");
            return FAILURE;
        }

        $slist_info = slist($fc, $clog1);
        if ($slist_info == FAILURE) {
            logcomment("$clog1 does not exist");
            return FAILURE;
        }

        return SUCCESS;
    } #is_clog_in_vvol

    sub get_n_clog_buckets  ($$)
    {
        my  ($fc, $volname);
        my  ($out, $clogdir0, $clogdir1, $clogdir_contents);
        my  ($n_buckets_0, $n_buckets_1);
        ($fc, $volname) = @_;

        $clogdir0 = "/vol/$volname/metadir/dense/clogdir.0";
        $clogdir1 = "/vol/$volname/metadir/dense/clogdir.1";
        $out = setflag($fc, "wafl_metadata_visible", 1);

        $clogdir_contents = ls($fc, $clogdir0);
        print Dumper $clogdir_content;
        $n_buckets_0 = grep(~/bucket/,$clogdir_contents);

        $clogdir_contents = ls($fc, $clogdir1);
        print Dumper $clogdir_contents;
        $n_buckets_1 = grep(~/bucket/,$clogdir_contents);     


        #XXX: TBD: also look at the output of sis check -c[v], and verify that
        #it reports the same number of buckets that we see here.
        return(($n_buckets_0, $n_buckets_1));
    }

#   TBD:    
#    sub dump_clog_records   ($$)
#    {
#        my  ($fc, $volname);
#        my  (@clog_records);
#
#        ($fc, $volname) = @_;
#
#        #XXX: TBD - run sis check -cv, parse out clog records 
#
#        return(@clog_records);
#    }

    sub get_vvol_fingerprint_count  ($$)
    {
        my  ($filer, $volname);
        my  ($volpath);
        my  ($line, $res);

        ($filer, $volname) = @_;

        $volpath = "/vol/$volname";

        $res = dense_check(
                    host=>$filer,
                    path_name=>$volpath
                );

        for $line (split(/\r*\n/, $res)) {
            if ($line =~ VVOL_FPDB_REC) {
                return($1);
            }
        }

        logcomment("No records found for volume $volname");
        return -1;
    }

} #package Dense::Clog;

return  1;
