# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/Dense/Clone.pm#1 $
#
## @summary 
##    Library routines for SIS Clone
##
## @usage
##	use Dense::Clone;
##
## @author mathan@netapp.com
## @status in-progress
## @pod here

package Dense::Clone;

use strict;
use Switch;
use SPARSE::Util::Logger;
use filer;
use Burt;
use Tharn;
use Volume qw (vol_list vol_status);
use Dense;
use Dense qw(:all);
use Dense::Status; 
use Dense::Utils;
use Dense::Constant::Command;
use Dense::Constant::Message;
use Dense::Constant::Clone_Command;
use Dense::Constant::Clone_Command qw(:all);
use Dense::Constant::Clone_Message;
use Dense::Agnostic::Constants;
use Data::Dumper qw(Dumper);
use Dense::sis_zapi;
use Dense::sis_zapi qw(:all);
use Params::Validate qw(validate validate_pos OBJECT SCALAR HASHREF ARRAYREF UNDEF);

use vars qw($DENSE_ONTAP_VERSION);
use vars qw($CHECK_32K_REFCNT);

BEGIN {
    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            Carp::confess( Tharn::script_result(@msgs) );
        }
    );
}    #BEGIN #1

use base qw(Exporter);

BEGIN	
	{
		my @conf 	= qw ();

		my @action 	= qw (	
					dense_sis_clone_start
					dense_sis_clone_stop
					clone_max_vfiler_ops
					clone_max_ops
					clone_corrupt_header_magic
				     ); 

		my @status 	= qw (
					dense_sis_clone_status	
				     );

		my @state 	= qw ();

		my @util	= qw (
					dense_sis_clone_clear
					dense_clone_fiji
					create_failed_clone_op
					get_clone_metafile_data
					get_aggrshowspace_details
					dense_calculate_exp_vbns
					dense_validate_vbns
					convert_to_block_range_hashlist
					clone_metafile_format
				     );


		my @verify 	= qw (	
					dense_sis_clone_wait_for_finish	
					dense_test_file_cloning
				     );


		our %EXPORT_TAGS= (	conf   => [ @conf ],
							action => [ @action ],
							status => [ @status ],
							state  => [ @state ],
							util   => [ @util ], 
							verify => [ @verify ]
				  );
        our @EXPORT_OK = qw(
          $DENSE_ONTAP_VERSION
          $CHECK_32K_REFCNT
        );

		$EXPORT_TAGS{'all'} = [@conf, @status, @action, @state, @util, @verify,];

		Exporter::export_ok_tags('all');

	}#BEGIN #2

logcomment(__FILE__);

########################################################
### Start sis clone operation
########################################################

=head2 dense_sis_clone_start(conn=> $conn, source=> $src_file,clone => "$clone_file",options => "-n -l",
                             blk_range => "1:2:1 5:10:3", timeout => 3600, partner=> 1)
start clone operation

Return value:

Success	: Return the clone ID
Fail	: Return the clone start command output

=cut

sub dense_sis_clone_start($@)
{

	my (%opts) = validate 	(@_, { 	conn		=> {type=>SCALAR | HASHREF},
									source 		=> {type=>SCALAR},
									clone 		=> {type=>SCALAR, optional=>1},
									options 	=> {type=>SCALAR, optional=>1},
									blk_range 	=> {type=>SCALAR, optional => 1},
									timeout 	=> {type=>SCALAR, default=>DENSE_CLONE_DEFAULT_TIMEOUT, optional=>1},
									partner 	=> {type=>SCALAR, optional=>1, default=>DENSE_CLONE_DEFAULT_PARTNER, regex=>qr/^(0|1)$/},
									vfiler 		=> {type=>SCALAR, optional=>1}
	                              }
	                          );

	### Get clone inputs

	my $conn		= $opts{conn};
	my $source		= $opts{source};
	my $clone		= $opts{clone};
	my $timeout		= $opts{timeout};
	my $options		= $opts{options};
	my ($blk_range)	= $opts{blk_range};
	my $partner		= $opts{partner};
	my $vfiler 		= $opts{vfiler};

	my $cmd_suffix;

	if(defined($options)){
		$options = " " . "$options" . " ";
	}else{
		$options = " ";
	}

	### Set block range
	
	if($blk_range){
		$options 	= "$options" . "-r ";
		$cmd_suffix = "$blk_range" ;
		logcomment("DEBUG: options and cmd_suffix are $options and $cmd_suffix");
	}
		
	### Generate & Execute the Clone command
	my $return_text = ExecuteCloneCmd
		(
			conn		=> $conn,
			src_file	=> "$source",
			clone_file	=> "$clone",
			blk_range	=> "$cmd_suffix",
			options		=> "$options",
			cmd_type	=> "start",
			timeout		=> $timeout,
			partner		=> $partner,
			vfiler		=> $vfiler,
		);

	# Get the Filer name
	my $filer	= $conn->hostp()->hostname();
	logcomment("DEBUG: Filer name is $filer" );
	
	# Check the expected response. Return Clone IC If clone started succussfully
	my $exp_response = GetCloneMessage( message => "DENSE_CLONE_OK", host => $filer);
	logcomment("DEBUG: Clone message is: $exp_response " );

	if($return_text	=~ m/$exp_response/){
		my $reg_exp	= GetCloneMessage( message => "DENSE_CLONE_GETID_REGEXP", host => $filer);
		$return_text =~ m/$reg_exp/;
		my $clone_id_value=$1;
		return(remove_leading_trailing_spaces($clone_id_value));
	}else{ 
		return(remove_leading_trailing_spaces($return_text));
	}
}


########################################################
### Stop sis clone operation
########################################################

=head2 dense_sis_clone_stop(conn=>$conn, vol=>$path_name, id => $id, timeout => 3600, partner=> 1)

Return value:

	Print the clone status after stopping the clone operation
			+
	Success	: Return the clone stop command output
	Fail	: Return the clone stop command output

=cut

sub dense_sis_clone_stop($@)
{
	my (%opts) = validate (@_, { 	conn	=> {type=>SCALAR | HASHREF},
                             		vol		=> {type=>SCALAR},
                             		id 		=> {type=>SCALAR},
                             		timeout => {type=>SCALAR, default=>DENSE_CLONE_DEFAULT_TIMEOUT, optional=>1},
                             		partner => {type=>SCALAR, optional=>1, default=>DENSE_CLONE_DEFAULT_PARTNER, regex=>qr/^(0|1)$/},
									vfiler  => {type=>SCALAR, optional=>1},
                            	    }
                      	       );

	my $conn	= $opts{conn};
	my $vol 	= $opts{vol};
	my $id 		= $opts{id};
	my $timeout = $opts{timeout};
	my $partner	= $opts{partner};
	my $vfiler	= $opts{vfiler};

	### Generate & Execute the Clone command
	my $return_text = ExecuteCloneCmd
		(
			conn		=> $conn,
			cmd_type	=> "stop",
			vol_name	=> "$vol",
			id			=> "$id",
			timeout		=> $timeout,
			partner		=> $partner,
			vfiler		=> $vfiler,
		);
	return ($return_text);
}

########################################################
### Wait for clone operation to complete
########################################################
=head2 dense_sis_clone_wait_for_finish(conn=>$conn, vol=>$path_name, id => $id, timeout => 3600, partner=> 1
					interval_in_sec => 3, max_time_to_wait=>30)

Get the clone ID and check the status in a loop (interval:3sec, max time to wait:30 sec) till
the clone operation completes

Return value:

	Print the clone status after stopping the clone operation
		+
	Success	: Return Success
	Fail	: Return Fail

=cut


sub dense_sis_clone_wait_for_finish($@)
{
	my (%opts) = validate (@_, {conn 			=> {type=>SCALAR | HASHREF},
								vol 			=> {type=>SCALAR},
								id  			=> {type=>SCALAR, regex   => qr/^[1-9]\d*$/},
								interval_in_sec	=> {type=>SCALAR, default =>DENSE_CLONE_INTERVAL, optional =>1},
								max_time_to_wait=> {type=>SCALAR, default =>DENSE_CLONE_MAX_TIME_TO_WAIT, optional =>1},
								timeout 		=> {type=>SCALAR, default=>DENSE_CLONE_DEFAULT_TIMEOUT, optional=>1},
								partner			=> {type=>SCALAR, optional=>1, default=>DENSE_CLONE_DEFAULT_PARTNER, regex=>qr/^(0|1)$/},
								vfiler 			=> {type=>SCALAR, optional=>1},
							}
						);
						

	my $conn			= $opts{conn};
	my $vol				= $opts{vol};
	my $id				= $opts{id};
	my $sleep_time		= $opts{interval_in_sec};
	my $max_time_to_wait= $opts{max_time_to_wait};
	my $timeout			= $opts{timeout};
	my $partner			= $opts{partner};
	my $vfiler			= $opts{vfiler};
	my $interval;
	my %status_info;

	for($interval=$max_time_to_wait; $interval>0; $interval-=$sleep_time) {
		logStep("Get the clone status info");
		if (defined $vfiler) {
		    %status_info = dense_sis_clone_status(conn => $conn, vol => $vol, id =>$id, timeout=>$timeout, partner =>$partner,vfiler=>$vfiler);
		} else {
		    %status_info = dense_sis_clone_status(conn => $conn, vol => $vol, id =>$id, timeout=>$timeout, partner =>$partner);
		}
		return DENSE_CLONE_SUCCESS if(!%status_info);
		logcomment("clone status hash got: ".Dumper(\%status_info));
		if($status_info{$id}{'State'} =~ m/failed/) {
			logcomment("DEBUG: Clone operation is in Failed/Stopped status");
			logcomment("DEBUG: Clone status for vol: $vol ID: $id is $status_info{$id}{'State'}");
			return DENSE_CLONE_FAIL;
		}
		logcomment("time-out: $interval");
		sleep($sleep_time);
	}
	die("Either time-out or wait for clone failed");
}

########################################################
### Check Clone status
########################################################
=head2 dense_sis_clone_status(conn=>$conn, vol=>$path_name, id => $id, timeout => 3600, partner=> 1

Get the clone ID and check the status

Return value:

	Status object

	Clone status		: $status_info{$id}{'State'}
	Source file			: $status_info{$id}{'Source'}
	Destination file	: $status_info{$id}{'Destination'}
	Clone blk range		: $status_info{$id}{'Block ranges'}
	Clone type			: $status_info{$id}{'Type'}

=cut

sub dense_sis_clone_status 
{
        
	my (%opts) = validate (@_,{	conn		=> {type =>SCALAR|HASHREF}, 
								vol			=> {type => SCALAR},
								id 			=> {type=>SCALAR, optional=>1, default=>0},
								timeout 	=> {type=>SCALAR, default=>DENSE_CLONE_DEFAULT_TIMEOUT, optional=>1},
								partner		=> {type=>SCALAR, optional=>1, default=>DENSE_CLONE_DEFAULT_PARTNER, regex=>qr/^(0|1)$/},
								vfiler			=> {type => SCALAR, optional=>1},
							}
						);
			  
	my $conn	= $opts{'conn'};
	my $vol		= $opts{'vol'};
	my $id;
	my $timeout	= $opts{timeout};
	my $partner	= $opts{'partner'};
	my $vfiler	= $opts{'vfiler'};

	if($opts{'id'}){
		$id = $opts{'id'};
	}else{
		$id = " ";
	}

	### Generate & Execute the Clone command
	my $return_text = ExecuteCloneCmd
		(
			conn		=> $conn,
			cmd_type	=> "status",
			vol_name	=> "$vol",
			id			=> "$id",
			timeout		=> $timeout,
			partner		=> $partner,
			vfiler 		=> $vfiler,
		);
	my @to_process = split(/\n/, $return_text);

	my $num = scalar(@to_process);
	my %status_info;
	my $key;
	my $index = 0;

	foreach my $line (@to_process)
	{

		# ID: 21
		# Source: /vol/p3vol/file.0
		# Destination: /vol/p3vol/file.0_clone.15
		# Block ranges: 
		# State: failed (Not enough disk space for the operation)
		# Type: file
		# 
		# ID: 84
		# Source: /vol/temp/lun1
		# Destination: /vol/temp/lun1_clone
		# Block ranges: 
		# State: failed (Not enough disk space for the operation)
		# Type: lun

		# It starts with an empty string
		if ($line =~ m/\w+/)
		{
			if($line =~ /^\s*$/){
				next;
			}
			if($line =~ /^\s*===== $vfiler/){ ## When run on a vfiler context skip the first line
				next;
			}
		
			my($field, $value) = split(":",$line);
			$field = remove_leading_trailing_spaces($field);
			$value = remove_leading_trailing_spaces($value);

			if($field =~ m/ID/){
				$key = $value;
				next;
			}  

			if($field =~ m/State/){
				# split on '(' and remove the trailing ')" using chop
				my @tmp_arr = split(/\(/,$value);
				chop($tmp_arr[1]);
				$status_info{$key}{'State'}		= $tmp_arr[0];
				$status_info{$key}{'Status'}	= $tmp_arr[1];
				next;
			}

			$status_info{$key}{$field} = $value;
		}
        } # end foreach
        
        return %status_info;
}

########################################################
### Clear failed clone ID
########################################################
=head2 dense_sis_clone_clear(conn=>$conn, vol=>$path_name, id => $id, timeout => 3600, partner=> 1

Clear failed clone ID

Return value:

	Print the clone status after stopping the clone operation
			+
	Success	: Return the clone stop command output
	Fail	: Return the clone stop command output

=cut

sub dense_sis_clone_clear
{

	my (%opts) = validate (@_, { 	conn	=> {type=>SCALAR | HASHREF},
                             		vol		=> {type=>SCALAR},
                             		id		=> {type=>SCALAR},
                             		timeout	=> {type=>SCALAR, default=>DENSE_CLONE_DEFAULT_TIMEOUT, optional=>1},
									partner	=> {type=>SCALAR, optional=>1, default=>DENSE_CLONE_DEFAULT_PARTNER, regex=>qr/^(0|1)$/},
									vfiler  => {type=>SCALAR, optional=>1},
				  }
			       );
			       
	my $conn	= $opts{conn};
	my $vol 	= $opts{vol};
	my $id 		= $opts{id};
	my $timeout = $opts{timeout};
	my $partner	= $opts{partner};
	my $vfiler 	= $opts{vfiler};

	### Generate & Execute the Clone command
	my $return_text = ExecuteCloneCmd
		(
			conn		=> $conn,
			cmd_type	=> "clear",
			vol_name	=> "$vol",
			id			=> "$id",
			timeout		=> $timeout,
			partner		=> $partner,
			vfiler		=> $vfiler,
		);
	return ($return_text);
}

########################################################
### Perform file cloning and validate
########################################################
=head2 dense_test_file_cloning(conn=>$conn, vol=>$vol, src_file => $src_file, clone_file=>$clone_file, options =>$options
			   blk_range => $blk_range, check_dense_flag=> $check_dense_flag, check_vbns=> $check_vbns, exp_refcnt => $exp_refcnt,
			   check_clog=> $check_clog, clog_count=> $clog_count, partner=> 1)
Description:
	Perform file cloning and validate the status
	
Inputs for clone operation:
	clone_file	is optional for same file (sub-file) cloning
	options		is optional if we dont want to pass any options (eg. -n, -l)
	blk_range	is Mandatory for sub-file cloning
	
Inputs for clone Validation:

	Validatiog dense flag
		check_dense_flag- if set check the dense flag for volume & source/clone files
	
	Validation fbn,vbn,vvbn,refcnt 
		check_vbns	- Check & compare fbn,vbn,vvbn,refcnt of source & clone file before & after clone
		      Limitation when using 'check_vbns' options:
					1. Cannot validate expected vbn,vvbn,refcnt for "sub-file cloning + multiple block 
					   range with overlapping + with snapshot" -- Bug 361541
					2. If the donor's refcount is 32767. In this case we cannot predict the new receipt's vbn, vvbn. 
		max_refcnt_fbn_range	- To validate Max refcount testcases
	
	Validation Change log
		check_clog	- Check & compare changelog entries before & after clone
		clog_count 	- Its mandatory when validating change log (ie. if we set check_clog)

	compress_src		= set to '1' if the source file is compressed
		
Return value:
	Success	: Return Succuss
	Fail	: Return Fail

=cut

sub dense_test_file_cloning($@)
{

	my (%opts) = validate 	(@_, { 	conn				=> {type=>SCALAR | HASHREF},
									vol 				=> {type=>SCALAR | HASHREF},
									src_file 			=> {type=>SCALAR},
									clone_file			=> {type=>SCALAR, optional=>1},
									options 			=> {type=>SCALAR, optional=>1},
									blk_range 			=> {type=>SCALAR, optional => 1},
									check_dense_flag	=> {type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/},
									check_vbns			=> {type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/},
									exp_refcnt			=> {type=>SCALAR, optional=>1, default=>1, regex => qr/^(0|1)$/},
									max_refcnt_fbn_range=> {type=>SCALAR | ARRAYREF, optional => 1},
									check_clog			=> {type=>SCALAR, optional=>1, default=>0, regex => qr/^(0|1)$/},
									clog_count 			=> {type=>SCALAR, optional=>1},
									compress_src		=> {type=>SCALAR, optional=>1, default=>0, regex => qr/^(0|1)$/},
									interval_in_sec		=> {type=>SCALAR, default =>DENSE_CLONE_INTERVAL, optional =>1},
									max_time_to_wait	=> {type=>SCALAR, default =>DENSE_CLONE_MAX_TIME_TO_WAIT, optional =>1},
									timeout				=> {type=>SCALAR, default=>DENSE_CLONE_DEFAULT_TIMEOUT, optional=>1},
									partner 			=> {type=>SCALAR, optional=>1, default=>DENSE_CLONE_DEFAULT_PARTNER, regex=>qr/^(0|1)$/},
									vfiler          => {type=>SCALAR, optional=>1}
									  }
	                         );
	
	my $conn 				= $opts{'conn'};
	my $vol 				= $opts{'vol'};	
	my $src_file			= $opts{'src_file'};
	my $clone_file			= $opts{'clone_file'};
	my $options				= $opts{options};
	my $blk_range			= $opts{blk_range};
	my $check_dense_flag	= $opts{'check_dense_flag'};
	my $check_vbns			= $opts{'check_vbns'};
	my $exp_refcnt			= $opts{'exp_refcnt'};
	my $max_refcnt_fbn_range= $opts{'max_refcnt_fbn_range'};
	my $check_clog			= $opts{'check_clog'};
	my $clog_count			= $opts{'clog_count'};
	my $compress_src		= $opts{'compress_src'};
	my $interval_in_sec		= $opts{'interval_in_sec'};
	my $max_time_to_wait	= $opts{'max_time_to_wait'};
	my $timeout				= $opts{timeout};
	my $partner				= $opts{partner};
	my $vfiler              = $opts{vfiler};

	my $id1;
	my %vbns_src_bclone;
	my %vbns_dst_bclone;
	my $clone_exp_val;
	my %vbns_src_aclone;
	my %vbns_dst_aclone;
	my %chlog_before_clone;
	my %chlog_after_clone;
	my $log_one_before;
	my $log_two_before;
	my $log_one_after;
	my $log_two_after;
	my $log_one;
	my $log_two;
	my $clog_result;
	my ($size, $blk_cnt);
	my %inode_info;
	my $refcnt_bclone;
	my $exp_blkrange;
	my $s_exp_vbns;
	my $d_exp_vbns;
	my $vbns_src_bclone_ref;
	my $vbns_dst_bclone_ref;
	
	my $vol_name = $vol->{name};
	my $pathname = '/vol/'.$vol_name;
	
	### Get the Filer name from 'conn'
	my $FILER = $conn->hostp()->hostname();
	my $changelog0_name = GetCloneMessage( message => "clog0_name", host => $FILER);
	my $changelog1_name = GetCloneMessage( message => "clog1_name", host => $FILER);
	
	### Get vvbns for source file before clone
	if($check_vbns)
	{
		if ($max_refcnt_fbn_range){
			logStep("Get vbn, vvbn & refcount for max refcount testing before cloning");
			%vbns_src_bclone = dense_parse_showvbns(conn   		=> $conn,
													file_name	=> $src_file,
													fbn_range	=> $max_refcnt_fbn_range,
													partner		=> $partner
													) or die fatal_result("dense_parse_showvbns() Failed") ;
		}else{
			logStep("Get vbn, vvbn & refcount for source file before cloning");
			$refcnt_bclone = $exp_refcnt - 1;
			%vbns_src_bclone = dense_parse_showvbns(conn   		=> $conn,
													file_name	=> $src_file,
													partner		=> $partner
													) or die fatal_result("dense_parse_showvbns() Failed") ;
			if (($clone_file) && ($blk_range)){
				%vbns_dst_bclone = dense_parse_showvbns(conn   	=> $conn,
													file_name	=> $clone_file,
													partner		=> $partner
													) or die fatal_result("dense_parse_showvbns() Failed") ;
				$vbns_dst_bclone_ref = \%vbns_dst_bclone;
			}elsif((! $clone_file) && ($blk_range)){
				### For same file cloning,
				### To calculate the expected vbn, vvbn, & refcnt we need to pass the same hash reference 
				### for source and destination file.
				$vbns_dst_bclone_ref = \%vbns_src_bclone;
			}else{
				$vbns_dst_bclone_ref = \%vbns_dst_bclone;
			}

			### Get the expected vbn, vvbn & refcount for sub-file cloning
			if (! $blk_range){
				my $inode_num = dense_get_inode_num (	conn		=> $conn,
						file_name	=> $src_file,
						timeout		=> $timeout, 
						partner		=> $partner
					)or die fatal_result ("Getting inode for file $src_file Failed");
				my $inode_info = dense_parse_dumpblock (	
						conn	=> $conn,
						vol		=> $vol_name,
						inum	=> $inode_num,
					)or die fatal_result ("Parsing dumpblock dense_parse_dumpblock Failed");
				my $f_size = $inode_info->{size}/4096;
				my $no_blk = POSIX::ceil($f_size);
				$exp_blkrange = "0:0:$no_blk";
			}else{
				$exp_blkrange = $blk_range;
			}
			$vbns_src_bclone_ref = \%vbns_src_bclone;
			
			my %comp_data = ();
			if ($compress_src){
				$comp_data{comp_head} = GetCloneMessage(message => "DENSE_COMPRESSION_HEADER", host => $FILER);
				$comp_data{comp_trial} = GetCloneMessage(message => "DENSE_COMPRESSION_TRAILER", host => $FILER);
			}
			$clone_exp_val = dense_calculate_exp_vbns (
				blkr	=> "$exp_blkrange",
				cmp_data => \%comp_data,
				s_wsv	=> $vbns_src_bclone_ref,
				d_wsv	=> $vbns_dst_bclone_ref,
			);
			$s_exp_vbns = $clone_exp_val->[0];
			$d_exp_vbns = $clone_exp_val->[1];
		}
	}
		
	### Get changelog details
	if($check_clog){
		logStep("Get Change log details before cloning");
		%chlog_before_clone = get_changelog_info(host=>$FILER,path_name=>$pathname);
	}
	
	### Start Clone operation
	if (defined $vfiler) {
		$id1 = dense_sis_clone_start(	conn		=> $conn,
									source		=> $src_file,
									clone  		=> "$clone_file",
									options		=> "$options",
									blk_range	=> "$blk_range",
									timeout		=> $timeout,
									partner		=> $partner,
									vfiler		=> $vfiler
									) or die fatal_result ("dense_sis_clone_start() Failed");
	} else {
                $id1 = dense_sis_clone_start(   conn            => $conn,
                                                                        source          => $src_file,
                                                                        clone           => "$clone_file",
                                                                        options         => "$options",
                                                                        blk_range       => "$blk_range",
                                                                        timeout         => $timeout,
                                                                        partner         => $partner,
                                                                        ) or die fatal_result ("dense_sis_clone_start() Failed");
	}
		
	if ($id1 =~ /^[0-9]+$/){
		logcomment("dense_sis_clone_start Passed Clone ID is $id1");
	}else{
		logresult('FAIL', "Starting Clone operation failed. return value: $id1");
		return  DENSE_CLONE_FAIL; 
	}

	logStep("Wait for clone operation to complete");
	my $clone_wait_status;
	if (defined $vfiler) {
		$clone_wait_status = dense_sis_clone_wait_for_finish (conn			=> $conn,
															vol				=> $vol_name,
															id				=> $id1,
															interval_in_sec	=> $interval_in_sec,
															max_time_to_wait=> $max_time_to_wait,
															partner			=> $partner,
															timeout			=> $timeout,
															vfiler			=> $vfiler
															)or die fatal_result ("Wait for clone ops to finish failed");
	} else {
                $clone_wait_status = dense_sis_clone_wait_for_finish (conn                      => $conn,
                                                                                                                        vol                             => $vol_name,
                                                                                                                        id                              => $id1,
                                                                                                                        interval_in_sec => $interval_in_sec,
                                                                                                                        max_time_to_wait=> $max_time_to_wait,
                                                                                                                        partner                 => $partner,
                                                                                                                        timeout                 => $timeout,
                                                                                                                        )or die fatal_result ("Wait for clone ops to finish failed");
	}
	if ($clone_wait_status eq DENSE_CLONE_SUCCESS){
		logcomment("Clone operation completed as expected");
	}else{
		logresult('FAIL', "dense_sis_clone_wait_for_finish returns $clone_wait_status ");
		return  DENSE_CLONE_FAIL; 
	}
						
	### Get vvbns for source and cloned file after clone
	if($check_vbns){
		logStep("Get vbn, vvbn & refcount for source/clone file after cloning");
		if ($max_refcnt_fbn_range){
			### For sub-file cloning - same file
			if (! $clone_file){
				$clone_file = $src_file;
			}
			%vbns_src_aclone = dense_parse_showvbns(conn   		=> $conn,
													file_name	=> $src_file,
													fbn_range	=> $max_refcnt_fbn_range,
													partner		=> $partner
													) or die fatal_result("dense_parse_showvbns() Failed") ;
			
			%vbns_dst_aclone = dense_parse_showvbns(conn   		=> $conn,
													file_name	=> $clone_file,
													fbn_range	=> $max_refcnt_fbn_range,
													partner		=> $partner
													) or die fatal_result("dense_parse_showvbns() Failed") ;

			### Compare and verify that blk sharing is as expected
			logStep("Verify vbn, vvbn & refcount for source/clone file after cloning");
			my $fbn = 0;
			my $max_refcnt = get_max_supported_ref_count($DENSE_ONTAP_VERSION);
			foreach my $fbn (keys %vbns_src_bclone)
			{
				if(($vbns_src_bclone{$fbn}{vbn} != $vbns_dst_aclone{$fbn}{vbn}) && ($vbns_src_bclone{$fbn}{vvbn} != $vbns_dst_aclone{$fbn}{vvbn}) && ($vbns_src_bclone{$fbn}{refcount} != $vbns_dst_aclone{$fbn}{refcount}) && ($vbns_src_aclone{$fbn}{refcount} != $vbns_dst_aclone{$fbn}{refcount}) && ($vbns_src_bclone{$fbn}{refcount} == $max_refcnt) && ($vbns_src_aclone{$fbn}{refcount} == $max_refcnt) && ($vbns_dst_aclone{$fbn}{refcount} == 0)){
					logcomment("DEBUG: Block $fbn from $src_file was shared as expected");
				}else{
					logresult('FAIL', "Block $fbn from $src_file was not shared as expected values: vbn $vbns_src_bclone{$fbn}{vbn} and $vbns_dst_aclone{$fbn}{vbn}, vvbn $vbns_src_bclone{$fbn}{vvbn}  and $vbns_dst_aclone{$fbn}{vvbn}, refcnt $vbns_src_bclone{$fbn}{refcount} and $vbns_dst_aclone{$fbn}{refcount} ");
					return DENSE_CLONE_FAIL;
				}
			}
		}else{
			%vbns_src_aclone = dense_parse_showvbns(conn   		=> $conn,
													file_name	=> $src_file,
													partner		=> $partner
													) or die fatal_result("dense_parse_showvbns() Failed") ;
			if (($clone_file)){
				%vbns_dst_aclone = dense_parse_showvbns(conn 	=> $conn,
													file_name	=> $clone_file,
													partner		=> $partner
													) or die fatal_result("dense_parse_showvbns() Failed") ;
			}elsif((!$clone_file) && $blk_range){
				%vbns_dst_aclone = dense_parse_showvbns(conn 	=> $conn,
													file_name	=> $src_file,
													partner		=> $partner
													) or die fatal_result("dense_parse_showvbns() Failed") ;			
			}
			
			### Compare and verify that blk sharing is as expected
			logStep("Verify vbn, vvbn & refcount for source/clone file after cloning");
			if ((! $blk_range) || ((!$clone_file) && $blk_range)){
				$d_exp_vbns = $s_exp_vbns;
			}
			
			my $comp_result = dense_compare_wsv (hash1 => \%vbns_src_aclone, hash2 =>$s_exp_vbns);
				if($comp_result ne DENSE_CLONE_SUCCESS){
				logcomment('FAIL', "After clone: Block from $src_file was not shared as expected");
				return DENSE_CLONE_FAIL;
			}
			#foreach my $fbn (keys %vbns_src_aclone){
			#	my $comp_result = dense_compare_hashes (hash1 => $vbns_src_aclone{$fbn}, hash2 =>$s_exp_vbns->{$fbn});
			#	if($comp_result ne DENSE_CLONE_SUCCESS){
			#		logcomment("DEBUG: Obtained Values: fbn:$fbn, vbn: $vbns_src_aclone{$fbn}{vbn}, vvbn:$vbns_src_aclone{$fbn}{vvbn}, refcnt: $vbns_src_aclone{$fbn}{refcount} ");
			#		logcomment("DEBUG: Expected Values: fbn:$fbn, vbn:$s_exp_vbns->{$fbn}->{vbn}, vvbn:$s_exp_vbns->{$fbn}->{vvbn}, refcnt:$s_exp_vbns->{$fbn}->{refcount} ");
			#		logresult('FAIL', "Block $fbn from $src_file was not shared as expected");
			#		return DENSE_CLONE_FAIL;
			#	}
			#}
			$comp_result = dense_compare_wsv (hash1 => \%vbns_dst_aclone, hash2 =>$d_exp_vbns);
				if($comp_result ne DENSE_CLONE_SUCCESS){
				logcomment('FAIL', "After clone: Block from cloned file was not shared as expected");
				return DENSE_CLONE_FAIL;
			}
			#foreach my $fbn (keys %vbns_dst_aclone){
			#	my $comp_result = dense_compare_hashes (hash1 => $vbns_dst_aclone{$fbn}, hash2 =>$d_exp_vbns->{$fbn});
			#	if($comp_result ne DENSE_CLONE_SUCCESS){
			#		logcomment("DEBUG: Obtained Values: fbn:$fbn, vbn:$vbns_dst_aclone{$fbn}{vbn}, vvbn:$vbns_dst_aclone{$fbn}{vvbn}, refcnt:$vbns_dst_aclone{$fbn}{refcount} ");
			#		logcomment("DEBUG: Expected Values: fbn:$fbn, vbn:$d_exp_vbns->{$fbn}->{vbn}, vvbn:$d_exp_vbns->{$fbn}->{vvbn}, refcnt:$d_exp_vbns->{$fbn}->{refcount} ");
			#		logresult('FAIL', "Block $fbn from $src_file was not shared as expected");
			#		return DENSE_CLONE_FAIL;
			#	}
			#}				
		}
	}
	
	if($check_dense_flag)
	{
		### Check dense flag for Volume
		logStep("Check dense flag for Volume after cloning");
		my $vol_dense_flag = is_dense_flag(host=>$FILER, volname=>$vol_name);
		if($vol_dense_flag){
			logcomment("Volume $vol_name has a dense flag");
		}else{
			logresult('FAIL', "Dense flage is not set on volume $vol_name");
			return DENSE_CLONE_FAIL;
		}
		
		
		### Check flag for source file
		logStep("Check dense flag for source file after cloning");
		my $inode_dense_flag = check_inode_dense_flag (	conn		=> $conn,
														vol			=> $vol,
														file_name	=> $src_file,
														timeout		=> $timeout, 
														partner		=> $partner
													)or warn fatal_result ("Getting inode dense flag for file $src_file Failed");
		
		if ($inode_dense_flag){
			logcomment("File $src_file has a dense flag");
		}else{
			logresult('FAIL', "File $src_file does not have a dense flag");
			return DENSE_CLONE_FAIL;
		}
		
		### Check flag for clone file
		if ($clone_file){
			logStep("Check dense flag for clone file after cloning");
			my $inode_dense_flag = check_inode_dense_flag (	conn		=> $conn,
															vol			=> $vol,
															file_name	=> $clone_file,
															timeout		=> $timeout, 
															partner		=> $partner
														)or warn fatal_result ("Getting inode dense flag for file $clone_file Failed");
			
			if ($inode_dense_flag){
				logcomment("File $clone_file has a dense flag");
			}else{
				logresult('FAIL', "File $clone_file does not have a dense flag");
				return DENSE_CLONE_FAIL;
			}
		}
	}
	
	### Verify changelog details
	if($check_clog)
	{
		### Verify clog after cloning
		logStep("Get Change log details after cloning");
		if (! $clog_count){
			logresult('FAIL', "Cannot check the file size. Arg Clog count 'clog_count' is missing");
			return DENSE_CLONE_FAIL;
		}
		
		%chlog_after_clone = get_changelog_info(host=>$FILER,path_name=>$pathname);

		$log_one_before = $chlog_before_clone{$changelog0_name};
		$log_two_before = $chlog_before_clone{$changelog1_name};

		$log_one_after = $chlog_after_clone{$changelog0_name};
		$log_two_after = $chlog_after_clone{$changelog1_name};

		$log_one = $log_one_after - $log_one_before;
		$log_two = $log_two_after - $log_two_before ;
		
		$clog_result = $log_two_after - $log_two_before;
		logStep("Verify Change log details after cloning");
		if (($log_one == $clog_count) || ($log_two = $clog_count)){
			logcomment("DEBUG: Change log counter has an expected value");
		}else{
			logresult('FAIL', "dense_verify_changelog failed $clog_result");
			return DENSE_CLONE_FAIL;
		}
	}
	return  DENSE_CLONE_SUCCESS;
}

########################################################
### Generate Clone fiji command
########################################################
=head2 dense_clone_fiji(volname=>$vol_name, sleep=> $delay_in_sec, action => $slow)

Generate Clone fiji command

Possible values for action
	action	--> 'slow'

Return value:
	Return Clone fiji command based on the input 'action'
	Return default Clone fiji command in invalid action specified (Default fiji is "slow down clone operation")

=cut

sub dense_clone_fiji ($@)
{

	my (%opts) = validate (@_, {volname	=> {type=>SCALAR},
								sleep	=> {type=>SCALAR,optional=>1,default=>DENSE_FIJI_DEFAULT_DELAY, regex=> qr/^[1-9]\d*$/},
								repeat	=> {type=>SCALAR,optional=>1,regex=> qr/^[1-9]\d*$/},
								action	=> {type=>SCALAR}
								}
						   );
			       
	my $volname 	= $opts{volname};
	my $sleep		= $opts{sleep};
	my $action		= $opts{action};
	my $repeat		= $opts{repeat};
	my $clone_fiji;
	
	### Generate Clone fiji command
	### Default fiji is "slow down clone operation"
	if ($repeat) {
		$clone_fiji = "fiji filter \"(repeat $repeat(dense.clone.go.slow(eq \@vol $volname)(sleep $sleep)))\""; 
	} else {
		$clone_fiji = "fiji filter (dense.clone.go.slow(eq \@vol $volname)(sleep $sleep))"; 
	}	
	return $clone_fiji;
}

####################################
### Functions to validate the expected vbn,vvbn & refcount after sub-file clone operation
####################################
=head2 _clone_struct_info

	Create a hash with expected vbn,vvbn & refcount for source and destination file

Inputs:
	1. Hash reference (Wafl showvbn output) for source file
	2. Hash reference (Wafl showvbn output) for the destination file
	3. Block range for sub-file cloning

Note: For same file cloning pass the same Hash reference value for source & destination file

Return value:
	Return Array reference 
		array[0] --> Expected vbn,vvbn & refcount for Source file
		array[1] --> Expected vbn,vvbn & refcount for Cloned file

=cut

sub	_clone_struct_info () {
	my	$x = {
		f2v	=> 	{},
		f2vv=> 	{},
		v2r	=>	{},
	};
	bless($x);
	return $x;
}

sub	dense_calculate_exp_vbns (@)
{
	my	($args, $in_args, $same_file_cloning);	
	my	($s_wsv, $d_wsv, $s_info, $d_info, $cmp_data);
	my $max_rcnt = MAX_REF_COUNT-1;
	$same_file_cloning = 0;
	$args = {
		blkr	=>	"",
		cmp_data =>	{},
		s_wsv	=>	{},
		d_wsv	=>	{},
	};

	$in_args = { @_ };

	map {
		$args->{$_} = $in_args->{$_}
	} keys(%$in_args);

	if ($args->{blkr} eq "") {
		die fatal_result ("Block range is missing");
	}
	my $blk_range = $args->{blkr};
	$cmp_data  = $args->{cmp_data};

	$s_wsv = $args->{s_wsv};
	$d_wsv = $args->{d_wsv};

	### Check Same file cloning
	if ("$s_wsv" eq "$d_wsv"){
		$same_file_cloning = 1;
	}
	
	$s_info = _clone_struct_info();
	$d_info = _clone_struct_info();

	### Construct hashes 'fbn to vbn', 'fbn to vvbn' & 'vbn to refcnt' for source & clone file
	my ($fbn, $vbn, $vvbn, $refc);
	for	my $idx (sort keys(%$s_wsv)) {
		$fbn 	= $s_wsv->{$idx}->{fbn};
		$vbn 	= $s_wsv->{$idx}->{vbn};
		$vvbn 	= $s_wsv->{$idx}->{vvbn};
		$refc 	= $s_wsv->{$idx}->{refcount};

		$s_info->{f2v}->{$fbn} = $vbn;
		$s_info->{f2vv}->{$fbn} = $vvbn;
		$s_info->{v2r }->{$vbn} = $refc;
	}

	for	my $idx (sort keys(%$d_wsv)) {
		$fbn 	= $d_wsv->{$idx}->{fbn};
		$vbn 	= $d_wsv->{$idx}->{vbn};
		$vvbn 	= $d_wsv->{$idx}->{vvbn};
		$refc 	= $d_wsv->{$idx}->{refcount};

		$d_info->{f2v}->{$fbn} = $vbn;
		$d_info->{f2vv}->{$fbn} = $vvbn;
		$d_info->{v2r }->{$vbn} = $refc;
	}
	
	### Calculate expected vbn, vvbn & refcount for each block range
	my @blk_list = split(/ /, $blk_range);
	foreach my $range (@blk_list){
		logcomment("Range: $range");
		my @blk_filed = split(/:/, $range);
		my @s_fbn_range = ($blk_filed[0]..$blk_filed[0]+$blk_filed[2]-1);
		my @d_fbn_range = ($blk_filed[1]..$blk_filed[1]+$blk_filed[2]-1);
		
		logcomment("source fbn range      : @s_fbn_range");
		logcomment("destination fbn _range: @d_fbn_range");
		
		for (my $i=0; $i<$blk_filed[2]; $i++){
			my ($s_fbn, $s_vbn, $s_vvbn, $s_recf, $d_fbn, $d_vbn, $d_vvbn, $d_recf);

			$s_fbn	= $s_fbn_range[$i]; 
			$s_vbn	= $s_info->{f2v}->{$s_fbn};
			$s_vvbn	= $s_info->{f2vv}->{$s_fbn};
			$s_recf	= $s_info->{v2r }->{$s_vbn};
			
			$d_fbn	= $d_fbn_range[$i];
			$d_vbn	= $d_info->{f2v}->{$d_fbn};
			$d_vvbn	= $d_info->{f2vv}->{$d_fbn};
			$d_recf	= $d_info->{v2r }->{$d_vbn};

			### Skip the calcalution if the donor's refcount vbn are same
			if ($s_vbn != $d_vbn){
				### Exit the calcalution if the donor's refcount is 32767
				### We cannot predict the new blocks vbn & vvbn
				if ($s_recf != $max_rcnt){
					### Increment refcount & Copy vbn
					if ($s_vbn != 0){
						### Decrement destination refcount
						if ($d_recf > 0) {
							$s_info->{v2r }->{$d_vbn}--;
							$d_info->{v2r }->{$d_vbn}--;
						}
						### Increment source and destination refcount
						### Do not increment the refcount if the source is a compression file 
						### and vvbn is not a valid number 
						my $incr = 1;
						if ((scalar(keys %$cmp_data)) > 0) {
							foreach my $comp_key (%$cmp_data){
								if ($s_info->{f2v}->{$s_fbn} == $cmp_data->{$comp_key}){
									$incr = 0;
								}
							}
						}
						if ($incr){
							$s_info->{v2r }->{$s_vbn}++;
							#Not required: $d_info->{v2r }->{$s_vbn}++;
						}

						### Copy source vbn to destination
						$d_info->{f2v}->{$d_fbn}	= $s_info->{f2v}->{$s_fbn};
						$d_info->{f2vv}->{$d_fbn}	= $s_info->{f2vv}->{$s_fbn};
						$d_info->{v2r }->{$s_vbn}	= $s_info->{v2r }->{$s_vbn};
					}else{
						if ($d_recf > 0) {
							$s_info->{v2r }->{$d_vbn}--;
							$d_info->{v2r }->{$d_vbn}--;
						}
						$d_info->{f2v}->{$d_fbn}	= $s_info->{f2v}->{$s_fbn};
						$d_info->{f2vv}->{$d_fbn}	= $s_info->{f2vv}->{$s_fbn};
						$d_info->{f2vv}->{$d_fbn}	= $s_info->{f2vv}->{$s_fbn};
						$d_info->{v2r }->{$s_vbn}	= $s_info->{v2r }->{$s_vbn};
					}
				}else{
					### Source block refcount is 32767
					logcomment("Donor block refcount is $max_rcnt");
					die fatal_result ("Cannot predict recipient vbn, vvbn and refcount");
				}
			}
		}
		if ($same_file_cloning){
			### Only for same file cloning
			### Copy the destination hash content to source file hash
			%$s_info = %{Storable::dclone(\%$d_info)};
		}
	}
	my $s_wsv_exp = {};
	my $d_wsv_exp = {};
	
	### Construct a hash with expected vbn, vvbn and refcount values
	for	my $idx (sort keys(%$s_wsv)) {
		$fbn 	= $s_wsv->{$idx}->{fbn};
	
		$s_wsv_exp->{$idx}->{fbn}		= $fbn;
		$s_wsv_exp->{$idx}->{vbn}		= $s_info->{f2v}->{$idx};
		$s_wsv_exp->{$idx}->{vvbn}		= $s_info->{f2vv}->{$idx};
		$s_wsv_exp->{$idx}->{refcount}	= $s_info->{v2r }->{($s_info->{f2v}->{$idx})};
	}	
	for	my $idx (sort keys(%$d_wsv)) {
		$fbn 	= $d_wsv->{$idx}->{fbn};
	
		$d_wsv_exp->{$idx}->{fbn}		= $fbn;
		$d_wsv_exp->{$idx}->{vbn}		= $d_info->{f2v}->{$idx};
		$d_wsv_exp->{$idx}->{vvbn}		= $d_info->{f2vv}->{$idx};
		$d_wsv_exp->{$idx}->{refcount}	= $d_info->{v2r }->{($d_info->{f2v}->{$idx})};
	}

	### Return  an array reference with Expected Source & Destination hash value
	my @exp_value = ($s_wsv_exp, $d_wsv_exp);
	return \@exp_value;
}

########################################################
### Get Clone metafile information
########################################################
=head2 get_clone_metafile_data(filer=>$filer, volname=>$vol_name)

Get Clone metafile information in an hash

Return value:
	Return Clone metafile information in an hash in the following format:
	{	magic				=>	DENSE_CLONE_HDR_MAGIC,
		snap_index			=>	Index in snap_mask to name temporary snapshot taken for cloning.
		entry_index			=>	Index in the persistent storage of clone operation information
		type				=>	Type of cloning operation,
		err					=>	Error associated with clone operation,
		flags				=>	Command line specific flag for change logging, not creating snapshot etc.
		op_id				=>	Unique clone operation ID,
		snapid				=>	Snapid of snapshot taken for cloning,
		snap_cp_cnt			=>	CP count of the snapshot taken for cloning,
		source_fileid		=>	File id of source file, used only when argument has -n option,
		source_generation	=>	Generation count of clone file,
		clone_fileid		=>	File id of clone file,
		clone_generation	=>	Generation count of clone file,
		sfile_len			=>	Length of the source file path,
		cfile_len			=>	Length of the clone file path,
		num_blk_ranges		=>	Total number of block ranges,
	}

=cut

sub get_clone_metafile_data ($@) {
	my (%opts) = validate 	(@_, { 	filer	=> {type=>HASHREF},
									volname	=> {type=>SCALAR},
									clone_id=> {type=>SCALAR, optional=>1},
									timeout	=> {type=>SCALAR, default=>DENSE_CLONE_DEFAULT_TIMEOUT, optional=>1},
									partner => {type=>SCALAR, optional=>1, default=>DENSE_CLONE_DEFAULT_PARTNER, regex=>qr/^(0|1)$/},
								}
						   );
	my $line;
			       
	my $return_text = ExecuteCloneCmd(
		conn	=>	$opts{filer}->{conn},
		cmd_type=>	"print_information",
		vol_name=>	$opts{volname},
		test_mode=>	1,
		timeout	=>	$opts{timeout},
		partner	=>	$opts{partner},
	);
	my %metafile_data;
	$metafile_data{volume} = $opts{volname};
	my @lines=split(/\n/,$return_text);
	$line = shift(@lines);
	if($line=~/(${\DENSE_CLONE_IN_PROGRESS_FIELD}): (\d+)/) {
		$metafile_data{$1} = $2;
	} else {
		logcomment("Expected: ".DENSE_CLONE_IN_PROGRESS_FIELD.", But got $line");
		die fatal_result("Clone Metadata Template does not match for wvdc_dense_clone_inprog");
	}
	$line = shift(@lines);
	die fatal_result("Clone Metadata Template does not match for Metafile Data") if($line!~/Metafile Data/);
	while($line=shift(@lines)) {
		my ($index) = /Entry index (\d+):/;
		$metafile_data{entries}[$index] = parse_metafile_entry(shift(@lines));
		if($opts{clone_id} && ($opts{clone_id} == $metafile_data{entries}[$index]->{op_id})) {
			return($metafile_data{entries}[$index]);
		}
	}
	return(\%metafile_data);
}

########################################################
### Get Clone metafile information
########################################################
=head2 get_clone_metafile_data($entry)
	entry - each entry index lines from clone print_information output

	The entry corresponding to entry index is parsed here to be returned to get_clone_metafile_data. Don't use this directly in your thpls.

=cut

sub parse_metafile_entry {
	my ($entry) = @_;
	my @field_values = split(/\|/, $entry);
	my %metadata;
	#Add 0x at the head of each of metadata field values got as they are with out them
	@field_values = map {$_="0x$_"} @field_values;
	foreach my $field (@clone_metafile_fields) {
		$metadata{$field} = shift(@field_values);
	}
	$metadata{type} = $clone_types[hex($metadata{type})];
	$metadata{op_id} = (hex($metadata{op_id2})<<32)+hex($metadata{op_id1});
	delete $metadata{op_id1};
	delete $metadata{op_id2};
	$metadata{rest} = \@field_values;
	return(\%metadata);
}

########################################################
### Create an failed clone operation on the volume
########################################################
=head2 create_failed_clone_op(filer=>$filer, source=>$src, clone=>$clone, volname=>$vol_name)

Function Steps:
	start the clone operation between source and clone 
	remove clone file
	the clone operation will fail

Comments:
	The source file has to be big enough for creating an failed operation, otherwise set an fiji point before calling this function to create an failed clone operation for sure. It is possible for this function to return an successful cloned operation, if fiji is not used before.

Return Value:
	return clone id and vol uuid info in an hash
	{	clone_id	=> clone id,
		vol_uuid	=> volume uuid,
	}
=cut

sub create_failed_clone_op {
	my (%opts) = validate(@_,
			{   filer   =>  { type => HASHREF },
				source  =>  { type => SCALAR },
				clone   =>  { type => SCALAR },
				volname =>  { type => SCALAR },
			});
	my ($filer, $src, $clone, $volname) = ($opts{filer}, $opts{source}, $opts{clone}, $opts{volname});
	my $result=zapi_sis_clone_start(filer=>$filer, source=>$src, clone=>$clone);

	logcomment("Remove cloned file to fail the Clone operation");
	filer_cli_advanced_say($filer->{conn}, "rm $clone");

	### Wait for clone operation to complete
	dense_sis_clone_wait_for_finish(conn=> $filer->{conn},vol=> $volname,id => $result->{clone_id});

	logStep("Get the clone status info");
	my $clone_status_list=zapi_get_sis_clone_list(filer=>$filer, clone_id_info=>$result);

	die fatal_result("got more than 1 clone status list inspite of giving clone id information") if (@$clone_status_list > 1);

	my %clone_status=%{$clone_status_list->[0]};
	die fatal_result("Could not create failed clone operation in volume $volname") if($clone_status{'clone-state'} !~ /fail/);

	return($result);
}

########################################################
### Convert block range given in string format to hash list
########################################################
=head2 convert_to_block_range_hashlist($blockrange)

Description: 
	utility function to convert an blockrange list specified in string format to an list of hashes

Inputs: block range in string format as "s1:d1:r1 s2:d2:r2"

Return Value: list of hashrefs containing block ranges as follows

[ {
	"source-block-number"		=> s1, 
	"destination-block-number"	=> d1, 
	"block-count"				=> r1,
  },
  {
	"source-block-number"		=> s2, 
	"destination-block-number"	=> d2, 
	"block-count"				=> r2,
  },
]

=cut

sub convert_to_block_range_hashlist($) {
	my @blkranges;
	$_ = shift;
	
	foreach (split) {
		my %blkrange;
		@blkrange{"source-block-number", "destination-block-number", "block-count"} = split(/:/);
		push(@blkranges, \%blkrange);
	}
	return(\@blkranges);
}

########################################################
### Convert vfiler UUID to clone metafile format
########################################################
=head2
=cut

sub clone_metafile_format(@$) {
    my @initial_array = @_;
    foreach my $element (@initial_array) {
        logcomment("The current element is $element");
    }
    my @result;
    my @initial_array_3;
    $result[0] = "0x".$initial_array[0];
    $result[1] = "0x".$initial_array[2].$initial_array[1];
    logcomment("initial_array->3 : $initial_array[3]");
    my $str = $initial_array[3];
    my $arr1 = substr $str, 2, 2;
    logcomment("arr1 -> $arr1");
    my $arr2 = substr $str, 0, 2;
    logcomment("arr2 -> $arr2");
    my $str2 = $initial_array[4];
    my $elem1 = substr $str2, 0, 2;
    my $elem2 = substr $str2, 2, 2;
    $result[2] = "0x".$elem2.$elem1.$arr1.$arr2;
    logcomment("The third part of array -> $result[0] $result[1] $result[2]");
    logcomment("The last digit pair is : int(substr $str2, 10, 2)");
    my $elem = substr $str2, 10, 1;
    if( $elem !~ qr/0/ ) {
        $result[3] = "0x".(substr $str2,10,2).(substr $str2,8,2).(substr $str2,6,2).(substr $str2,4,2);
    } else {
        $result[3] = "0x".(substr $str2,11,1).(substr $str2,8,2).(substr $str2,6,2).(substr $str2,4,2);
    }
    logcomment("The result array is: @result");
    return @result;
}


########################################################
### Validate vbns after clone operation
########################################################
=head2 dense_validate_vbns (conn=>$conn, volname=>$volname, src_file => $src_file, clone_file=>$clone_file, src_vbns_bclone =>$src_wsv, clone_vbns_bclone =>$clone_wsv, blk_range=> $blk_range, timeout=> $timeout, partner=> 1)

Note: 
	1. This function has to be called after clone operation
	2. Use 'Dense::Utils::dense_parse_showvbns' to get 'wafl showvbns' output

Description:
	Gets the 'wafl showvbns' output for source/clone file & block range (for sub-file clone)
	Calculate the expected vbn, vvbn, refcount for all the blocks and compare with the actual values 
	
Inputs:
	conn				= Filer connection
	volname 			= Volume name
	src_file 			= Source file for the clone operation
	clone_file			= Clone file (Not required for same file sub-file cloning)
	src_vbns_bclone		= 'wafl showvbns' output (hash reference) for source file
	clone_vbns_bclone	= 'wafl showvbns' output (hash reference) for clone file (Required for different file sub-file cloning)
	blk_range 			= Block range to clone (Only for sub-file cloning)
	compress_src		= set to '1' if the source file is compressed
	
Return value:
	Success	: Return 1
	Fail	: Return 'FATAL FAIL'

=cut

sub dense_validate_vbns($@)
{

	my (%opts) = validate 	(@_, { 	conn				=> {type=>SCALAR | HASHREF},
									volname 			=> {type=>SCALAR},
									src_file 			=> {type=>SCALAR},
									clone_file			=> {type=>SCALAR, optional=>1},
									src_vbns_bclone		=> {type=>HASHREF},
									clone_vbns_bclone	=> {type=>HASHREF, optional=>1},
									blk_range 			=> {type=>SCALAR, optional => 1},
									compress_src		=> {type=>SCALAR, optional=>1, default=>0, regex => qr/^(0|1)$/},
									timeout				=> {type=>SCALAR, default=>DENSE_CLONE_DEFAULT_TIMEOUT, optional=>1},
									partner 			=> {type=>SCALAR, optional=>1, default=>DENSE_CLONE_DEFAULT_PARTNER, regex=>qr/^(0|1)$/}
									  }
	                         );
	
	my $conn 		= $opts{'conn'};
	my $volname 	= $opts{'volname'};	
	my $src_file	= $opts{'src_file'};
	my $clone_file	= $opts{'clone_file'};
	my $src_wsv		= $opts{'src_vbns_bclone'};
	my $clone_wsv	= $opts{'clone_vbns_bclone'};
	my $blk_range	= $opts{blk_range};
	my $compress_src= $opts{compress_src};
	my $timeout		= $opts{timeout};
	my $partner		= $opts{partner};

	if((!$clone_file) && (!$blk_range)){
		die fatal_result ("One of the input parameter is missing: 'clone file' or ' block range'");
	}
	
	my ($exp_blkrange, $s_exp_vbns, $d_exp_vbns, @comp_result);
	my (%vbns_src_aclone, %vbns_dst_aclone);
	### Get the expected vbn, vvbn & refcount for sub-file cloning
	if (!$blk_range) {
		my $inode_num	= dense_get_inode_num (conn=> $conn, file_name => $src_file,timeout=>$timeout, partner=> $partner) or die fatal_result ("Getting inode for file $src_file Failed");
		my $inode_info	= dense_parse_dumpblock (conn=> $conn,vol=> $volname,inum => $inode_num) or die fatal_result ("Parsing dumpblock dense_parse_dumpblock Failed");
		my $blkcnt		= POSIX::ceil($inode_info->{size}/4096);
		$exp_blkrange	= "0:0:$blkcnt";
	}else{
		if($clone_file && !$clone_wsv) {
			die fatal_result ("For validating sub-file cloning on different files, clone_vbns_bclone is required that is, vbns of destination file before clone start operation is required");
		}
	
		if(!$clone_file && $clone_wsv) {
			die fatal_result ("For validating sub-file cloning on same file, clone_vbns_bclone should not be given as destination file is same as source file");
		}
	
		$exp_blkrange = $blk_range;
		$clone_wsv	= $src_wsv if(!$clone_wsv);
	}
	
	### Get the Filer name from 'conn'
	my $FILER = $conn->hostp()->hostname();
	my %comp_data = ();
	if ($compress_src){
		$comp_data{'comp_head'} = GetCloneMessage(message => "DENSE_COMPRESSION_HEADER", host => $FILER);
		$comp_data{'comp_trial'} = GetCloneMessage(message => "DENSE_COMPRESSION_TRAILER", host => $FILER);
	}
	my $clone_exp_val = dense_calculate_exp_vbns (
		blkr	=> "$exp_blkrange",
		cmp_data => \%comp_data,
		s_wsv   => $src_wsv,
		d_wsv   => $clone_wsv,
	);
	
	$s_exp_vbns = $clone_exp_val->[0];
	$d_exp_vbns = $clone_exp_val->[1];
	
	%vbns_src_aclone = dense_parse_showvbns(conn   		=> $conn,
											file_name	=> $src_file,
											partner		=> $partner
											) or die fatal_result("dense_parse_showvbns() Failed") ;
	if ($clone_file) {
		%vbns_dst_aclone = dense_parse_showvbns(conn 	=> $conn,
											file_name	=> $clone_file,
											partner		=> $partner
											) or die fatal_result("dense_parse_showvbns() Failed") ;
	}elsif((!$clone_file) && $blk_range){
		%vbns_dst_aclone = dense_parse_showvbns(conn 	=> $conn,
											file_name	=> $src_file,
											partner		=> $partner
											) or die fatal_result("dense_parse_showvbns() Failed") ;			
	}
	
	### Compare and verify that blk sharing is as expected
	logStep("Verify vbn, vvbn & refcount for source/clone file after cloning");
	if ((! $blk_range) || ((!$clone_file) && $blk_range)){
		$d_exp_vbns = $s_exp_vbns;
	}
	
	$comp_result[0] = dense_compare_wsv (hash1 => \%vbns_src_aclone, hash2 =>$s_exp_vbns);
		if($comp_result[0] ne DENSE_CLONE_SUCCESS){
			die(fatal_result("After clone: Block from source file was not shared as expected"));
	}
	$comp_result[1] = dense_compare_wsv (hash1 => \%vbns_dst_aclone, hash2 =>$d_exp_vbns);
		if($comp_result[1] ne DENSE_CLONE_SUCCESS){
		die(fatal_result("After clone: Block from cloned file was not shared as expected"));
	}
}

############################################################################ 
### Functions to set or get the clone.max_vfiler_ops value on the filer  ###
############################################################################
=head2 clone_max_vfiler_ops
 
    set or get the clone.max_vfiler_ops value on the filer
 
Inputs:
    1. conn - filer connection
    2. set_value (optional) - value of the option that needs to be set

Return value:
    The value of the options read or set
 
=cut

sub clone_max_vfiler_ops($;$) {
    my ($conn, $set_value) = @_; 
    my ($ret_val, $set_op, $output, $cmd);
    
    $ret_val = 0;
    $cmd = "options clone.max_vfiler_ops"; 

    $output = filer_cli_test_say($conn, $cmd); 
    logcomment("options exec output: ".$output); 
    if ($output =~ /no such option/i) {
	die("This option is not supported on this release"); 
    }
    if (defined($set_value) && ($set_value ne "")) { 
	logcomment("Setting $cmd $set_value"); 
    	$set_op = filer_cli_test_say($conn, $cmd." ".$set_value);
    }
    $output = filer_cli_test_say($conn, $cmd); 
    ($ret_val) = ($output =~ /(\d+)/); 
    logcomment("$cmd = $ret_val");
    if (defined($set_value) && ($set_value ne "")) { 
	if ($set_value != $ret_val) {
	   die($cmd." was not set successfully : ".$set_op); 
    	} else { 
	   logcomment($cmd." was set successfully ".$set_value) 
    	}
    }

    return $ret_val; 
}

############################################################################  
### Functions to set or get the clone.max_ops value on the filer         ###  
############################################################################  
=head2 clone_max_ops

set or get the clone.max_ops value on the filer  

Inputs:  
    1. conn - filer connection
    2. set_value (optional) - value of the option that needs to be set  

Return value:  
    The value of the options read or set  

=cut  

sub clone_max_ops($;$) {  
    my ($conn, $set_value) = @_; 
    my ($ret_val, $set_op, $output, $cmd);
    
    $ret_val = 0; 
    $cmd = "options clone.max_ops";
    
    $output = filer_cli_test_say($conn, $cmd);  
    logcomment("options exec output: ".$output);
    if ($output =~ /no such option/i) {
	die("This option is not supported on this release");
    } 
    if (defined($set_value) && ($set_value ne "")) {  
        logcomment("Setting $cmd $set_value");  
        $set_op = filer_cli_test_say($conn, $cmd." ".$set_value); 
    } 
    $output = filer_cli_test_say($conn, $cmd);  
    ($ret_val) = ($output =~ /(\d+)/); 
    logcomment("$cmd = $ret_val"); 
    if (defined($set_value) && ($set_value ne "")) {  
	if ($set_value != $ret_val) {  
	   die($cmd." was not set successfully : ".$set_op);  
    	} else {  
	   logcomment($cmd." was set successfully ".$set_value)
    	}
    } 
    
    return $ret_val; 
}



############################################################################
###    Functions to corrupt the magic number of a clone operation        ###
############################################################################
=head2 clone_corrupt_header_magic

corrupt the magic number of a clone operation

Inputs:
    1. filer	- filer connection
    2. volname	- volume name
    3. clone_id - clone id for which the magic number needs to be corrupted
    4. value	- value to which magic number need to be corrupted to
    5. timeout	- function timeout
    6. partner	- partner filer
    7. vfiler	- the vfiler name

Return value:
    SUCCESS: 1
    FAILURE: 0

=cut

sub clone_corrupt_header_magic($@) {
    my (%opts) = validate   (@_, {  filer	=> {type=>SCALAR | HASHREF},
				    volname	=> {type=>SCALAR},
				    clone_id	=> {type=>SCALAR},
				    value	=> {type=>SCALAR},
				    timeout	=> {type=>SCALAR, default=>DENSE_CLONE_DEFAULT_TIMEOUT, optional=>1},
				    partner	=> {type=>SCALAR, optional=>1, default=>DENSE_CLONE_DEFAULT_PARTNER, regex=>qr/^(0|1)$/},
				    vfiler	=> {type=>SCALAR , optional=>1},
				 }
			     );

    my $filer	    = $opts{'filer'};
    my $conn	    = $filer->{conn};
    my $volname	    = $opts{'volname'};
    my $clone_id    = $opts{'clone_id'};
    #my $entry_index = $opts{'entry_index'};
    my $offset	    = 0;
    my $value	    = $opts{'value'};
    my $timeout	    = $opts{'timeout'};
    my $partner	    = $opts{'partner'};
    my $vfiler	    = $opts{'vfiler'};
    my $result	    = 0;
    my ($entry_index, $magic_after, $clone_meta_before, $clone_meta_after, $return_text);

    $clone_meta_before = get_clone_metafile_data(filer=>$filer, volname=>$volname, clone_id=>$clone_id);
    logcomment("clone_meta before = ".Dumper($clone_meta_before));

    $entry_index = sprintf("%d", $clone_meta_before->{entry_index}) + 1;
    logcomment("Now corrupting entry index = $entry_index"); 

    $return_text = ExecuteCloneCmd (
	    conn	    => $conn,
	    test_mode   => 1,
	    cmd_type    => "corrupt_header",
	    vol_name    => $volname,
	    entry_index => $entry_index,
	    offset	    => $offset,
	    value	    => $value,
	    timeout	    => $timeout,
	    partner	    => $partner,
	    vfiler	    => $vfiler,
	);	

    $clone_meta_after = get_clone_metafile_data(filer=>$filer, volname=>$volname, clone_id=>$clone_id);
    logcomment("clone_meta after =".Dumper($clone_meta_after));

    $magic_after = sprintf("%d", $clone_meta_after->{magic});

    if ($magic_after == $value) {
	logcomment("Corruption introduced successfully");
	$result = 1;
    } else {
	logcomment("Corruption was not introduced successfully");
    }
    return ($result);

}


1; 
