# -*- perl -*-
## @summary	Library that defines dense messages 
## @description
##	        This package checks for the ontap version running on the filer, and
##          appropriately imports message definitions from Message_<release>.pm.
##          The actual constant definitions are present in the Message_<release>.pm
##          files
##
## @usage
##	use Dense::Constant::Message;
## @author dl-drg-qa@netapp.com
## @status review
## @pod here

=head1 NAME

 Message - Message file for dense operation 

=head1 SYNOPSIS

 Use Dense::Constant::Message qw(define_messages);
 define_messages(version=>$ontap_version);
 # NOTE that ontap_version must be 7.2, 7.3 etc.


=cut

package Dense::Constant::Message;

use	strict;
use	Tharn qw(logcomment host);
use Params::Validate qw(validate validate_pos OBJECT SCALAR HASHREF ARRAYREF UNDEF);
use	Filer::Version;
use	Dense::Constant::Message_gb;
use	Dense::Constant::Message_ic;
use	Dense::Constant::Message_br;
use	Dense::Constant::Message_rr;
use	Dense::Constant::Message_sn;
use     Dense::Constant::Message_fs;

BEGIN {
	use vars ('@ISA', '@EXPORT');
	require Exporter;
	@ISA    = qw(Exporter);
	@EXPORT = qw(
			Dconst
			getwv
			GetMessage
		    );
	@Dense::Constant::Message::EXPORT_OK   = qw($DENSE_ONTAP_VERSION $AGGR_NAME);
}

use vars qw($DENSE_ONTAP_VERSION $AGGR_NAME);

## print "Using ". __FILE__;

#Always return WV_<release>_MIN, depending upon the wafl version
#returned by get_wafl_version.
sub getwv(@)
{
	my (%opts) = validate(@_, {host => {type => Params::Validate::SCALAR|OBJECT|HASHREF}});

		### Return version if 'DENSE_ONTAP_VERSION' is already set
		if ($ENV{DENSE_ONTAP_VERSION}){
			return $ENV{DENSE_ONTAP_VERSION};
		}elsif ($DENSE_ONTAP_VERSION){
			return $DENSE_ONTAP_VERSION;
		}
		
        my ($host);
	my $reftype = ref($opts{host});
        if ( ($reftype eq "OBJECT") || ($reftype eq "HASHREF") ) {
            $host = $opts{host}->hostp()->hostname();
	} else {
            $host = $opts{host};
        }
	
	logcomment("Host = $host");
	my $wv = Filer::Version::get_wafl_version(host($host));

	#The wafl_version returned above can be any number within a range
	#that corresponds to an ontap version. We check for the range and
	#return the minimum version number in that range. IOW, we always
	#return WV_<release>_MIN.
	if (($wv >= Filer::Version::WV_GORDONBIERSCH_MIN) && 
		($wv <= Filer::Version::WV_GORDONBIERSCH_MAX)) {
		return "GB";
	} elsif (($wv >= Filer::Version::WV_IRONCITY_MIN) &&
       		($wv <= Filer::Version::WV_IRONCITY_MAX)) {
		return "IC";
	} elsif (($wv >= Filer::Version::WV_BOILERMAKER_MIN) &&
		($wv <= Filer::Version::WV_BOILERMAKER_MAX)) {
		return "BR";
	} elsif (($wv >= Filer::Version::WV_ROLLINGROCK_MIN) &&
		($wv <= Filer::Version::WV_ROLLINGROCK_MAX)) {
		return "RR";
	} elsif (($wv >= Filer::Version::WV_SIERRANEVADA_MIN) &&
		($wv <= Filer::Version::WV_SIERRANEVADA_MAX)) {
		return "SN";
        } elsif (($wv >= Filer::Version::WV_FULLSTEAM_MIN) &&
                 ($wv <= Filer::Version::WV_FULLSTEAM_MAX)) {
                 return "FS";
        } else {
		die "Dense::Message::getwv doesn't support version $wv";
	}
}

#XXX: NOTE - the definition of @EXPORT below has to contain the union of
# all constants defined in the Message_<release>.pm files! Thankfully Perl
# does not crib if one simply exports a symbol but doesn't define it anywhere.
# This will probably cause some heartburn in the future, but I'm unable to think
# of a more elegant way of implementing conditional constant defs without
# breaking a lot of code.

#We use the constants returned by Filer::Version::get_wafl_version as
#hash keys here. We depend upon Dense::getwv() to return WV_<release>_MIN
#for every version defined here.
our %Dconst = (
	DEFAULT	=> \%Dense::Constant::Message_gb::GBConsts,
	GB	=> \%Dense::Constant::Message_gb::GBConsts,
	IC	=> \%Dense::Constant::Message_ic::ICConsts,
	BR	=> \%Dense::Constant::Message_br::BRConsts,
	RR	=> \%Dense::Constant::Message_rr::RRConsts,
	SN	=> \%Dense::Constant::Message_sn::SNConsts,
        FS      => \%Dense::Constant::Message_fs::FSConsts
);

sub GetMessage 
{
	my ( %opts ) = validate( @_, { 
				code => { type => SCALAR },
				host => { type => SCALAR|HASHREF|OBJECT },
				vol => { type => SCALAR, optional => 1 },
				version => { type => SCALAR, optional => 1 },
				large_aggr => { type => SCALAR, default => 0, regex=>qr/^(0|1)$/ },
			});

	my $version;

	if( defined( $opts{version} ) ) {
		$version = $opts{version};
	} else {
		$version = getwv(host=>$opts{host}); 
	}
	my $code = $opts{code};
        my $msg = "";

        if($code =~ /VOL_SIZE_LIMIT/) {
            $code = "";
            if($opts{large_aggr}) {
               $code = "LA_" if($opts{large_aggr});
            } else {
               my $conn = Utils::make_conn($opts{host}, "#rsh");
               $code = "LA_" if(Dense::Utils::dense_is_larger_aggr(conn=>$conn, aggr_name=>$AGGR_NAME));
            }
            $code.="VOL_SIZE_LIMIT";
        }
	if( $version ) {
		$msg = $Dconst{$version}{$code};
		logcomment("Version=[$version], msg=[$msg]");
	} else {
		$msg = $Dconst{"DEFAULT"}{$code};
		logcomment("Version could not be retrieved: $version, falling back to DEFAULT error message: [$msg]");
	}

	$msg .= ":$opts{vol}" if( defined $opts{vol} );
	return $msg;
}


# end of file 
1;
