# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Dense/DenseCorruptions.pm#1 $
#
## @summary  Dense Corruptions Library
##
## @description
##     Dense Corruptions library, sets different types of corruptions on 
##     destination volume and filer, either by copying the golden dataset
##     or making corruptions manually
##     The corruption data set if used should be only in GB, as there
##     are problems if it is copied from an higher release to GB
##
## @usage
##    use DenseCorrupions;
##    pass the Hash by reference with appropriate to_release and Filer to do_transform routine
##
## @keywords snaplock compliance clock
##
## @dependencies
## SnapMirror
##
## @author dl-dense-qa@netapp.com (DRG QA)
## @author arvindks@netapp.com 
##


package Dense::DenseCorruptions;

use strict;
use Tharn;
use Tharn qw($Log);
use NOTE;
use Dense qw(:action);
use Dense::Utils;
use Aggregate qw(:all);
use Carp qw(croak);
use Utils qw(make_conn);
use Data::Dumper;
use SnapMirror::Helper qw(sm_init sm_cleanup);
use SnapMirror qw(snapmirror_status_vol);
use SnapCmds;
use Volume qw(vol_online vol_root);
use Wreck qw(wreck);
use mount;
use Params::Validate qw(validate validate_pos OBJECT SCALAR BOOLEAN HASHREF CODEREF ARRAYREF UNDEF);

=head2 ::new

Creates an object representing Dense Corruptions

=over

=item Arguments

B<filer> => I<filer name>

I<Required.>
filer containing the Reliant Corruptions data set.

B<aggr> => I<aggregate name>

I<Optional.>
aggregate having the corruption data set

=item Example

=back

	$corr = Dense::SISundoCorruptions->new(filer=>$FILER, aggr=>"dense_aggr");
	
=cut

sub new {
	my $class	= shift;
	my (%args) = validate(@_, {	filer	=> { type=>SCALAR },
								aggr	=> { type=>SCALAR },
								client	=> { type=>SCALAR },
								dstfiler=> { type=>SCALAR, optional=>1 },
								dstaggr	=> { type=>SCALAR, optional=>1 },
								dstvol	=> { type=>SCALAR, optional=>1 },
							});
	my $self = {};
	
	croak "filer argument is mandatory" if(!exists $args{filer});
	
	# Set private variables for object
	$self->{_Filer}	= $args{filer};
	$self->{_aggr}	= $args{aggr};
	$self->{_host}	= host($self->{_Filer});
	$self->{_conn}	= make_conn($self->{_Filer}, "#rsh");
	$self->{_client}= $args{client};
	
	if($args{dstfiler}) {
		$self->{_dstFiler}	= $args{dstfiler};
		$self->{_dstaggr}	= $args{dstaggr};
		$self->{_dstvol}	= $args{dstvol};
		$self->{_dsthost}	= host($args{dstfiler});
		$self->{_dstconn}	= make_conn($args{dstfiler}, "#rsh");
	}
	
	# Bless this object
	bless $self, $class;
	
	return $self;
}

=head2 ::dense_corruptvols_list

Get the list of vols containing the corruption data set

=over

=item Example

=back

	$corrobj = Dense::SISundoCorruptions->new(filer=>$FILER, vol=>"acl_corruption");
	$vollist = $corrobj->dense_corruptvols_list();
	
=cut

sub dense_corruptvols_list {
	my ($self)	= @_;
	my $aggrobj	= aggr_status($self->{_conn}, $self->{_aggr});
	return($aggrobj->volumes());
}

sub _get_undo_msg {
	my ($tail, $count, $interval) = @_;
	my $lines;
	
	foreach (1...$count) {
		sleep($interval);
		$Log->comment("Checking messages: $_");
		$lines	= $tail->get_lines();
		if($tail->triggered()) {
		# get_matched_lines() returns array of arrays matching the patterns in /etc/messages file
			my ($matches) = $tail->get_matched_lines();
			$Log->comment("sis undo returned expected message: ".Dumper($matches));
			return($matches->[0]);
		}
	}

	$Log->comment("sis undo did not return expected message: ".Dumper($lines));
	return();
}

=head2 ::dense_make_corruption

Create the corruption of the specified type on the specified volume

=over

=item Arguments

B<type> => I<Corruption type>

I<Required.>
type of corruption to be created

B<vol> => I<volume name>

I<Required.>
volume to have corruption data set

B<image> => I<hash reference>

I<Required.>
volume and snapshot information on the RELIANT_FILER, where this corruption type can be found
image - {	vol		=>$vol, 
			snapshot=>$snapshot }

=item Example

=back

	$corrobj = Dense::DenseCorruptions->new(filer=>$FILER, aggr=>"dense_aggr");
	$corrobj->dense_make_corruption(type=>ACL_SHARE, vol=>$TEST_VOL, image=>$image) or logresult('FATAL', "Could not make ACL share corruption: $@");
	
=cut

sub dense_make_corruption {
	my $self	= shift;
	my (%args)	= validate(@_, {name	=> { type=>SCALAR },
								type	=> { type=>SCALAR },
								filer	=> { type=>SCALAR },
								vol		=> { type=>SCALAR },
								image	=> { type=>HASHREF },
								code	=> { type=>CODEREF, optional=>1},
								codeargs=> { type=>HASHREF, optional=>1},
								period	=> { type=>SCALAR, optional=>1, default=> 5},
								timeout	=> { type=>SCALAR, optional=>1, default=> 300}
							});
	my (%image)	= validate(@{[%{$args{image}}]},{	vol		=> { type=>SCALAR },
													snapshot=> { type=>SCALAR, optional=>1}
												});
	my $vol		= $args{vol};
	my $dst		= "$args{filer}:$vol";
	my $src		= "$self->{_Filer}:$image{vol}";
	my $host	= host($args{filer});
	my $conn	= make_conn($args{filer});
	my $timeout	= $args{timeout};
	my $snapshot= $image{snapshot};
	my $status;
	my @codeargs;

	$Log->comment("Start VSM transfer of Corruption data set for $args{name}");
	sm_init(srchost=>$self->{_host}, dsthost=>$host, src=>$src,	dst=>$dst) 
		or croak("vsm transfer of $args{name} corruption from $src to $dst failed: $@");

	$Log->comment("Verify snapmirror is completed");
	do {
		$Log->comment("Snapmirror time-out: $timeout");
		sleep($args{period});
		$timeout -= $args{period};
		if($timeout<=0) {
			sm_cleanup(srchost=>$self->{_host},	dsthost=>$host, src=>$src, dst=>$dst);
			croak("snapmirror transfer timed out");
		}
		$status	= snapmirror_status_vol(host=>$host, vol=>$vol);
	} while($status->{'is-transfer-in-progress'} eq "true");
	
	sm_cleanup(srchost=>$self->{_host},	dsthost=>$host, src=>$src, dst=>$dst); 
	
	if($snapshot) {
		$Log->comment("snap restore $vol to the $args{name} corruption data set");
		snaprestore_volume($conn, $vol, $snapshot) or croak("Could not snap restore $vol to $snapshot");
	}
	
	return($snapshot) unless(defined $args{code});
	
	croak("code argument should refer to a code segment or sub-routine") if(ref $args{code} ne "CODE");
	unless(defined $args{codeargs}->{inode}) {
		
	}
	
	@codeargs = ($self, %{$args{codeargs}});
	
	$Log->comment("Executing $args{code} with args @codeargs");
	$args{code}->(@codeargs);
	$snapshot = $image{vol}."_code";
	vol_online($conn, $vol) or croak("Could not online volume $vol after $args{code} because of $@");
	snapshot_create($conn, $vol, $snapshot) or croak("Could not create snapshot ($snapshot) in $vol because of $@");
	return($snapshot);
}

sub get_corrupted_inodes {
	my $self = shift;
	my (%args)	= validate(@_, {type	=> { type=>SCALAR, default=>"log", regex=>qr/^(log|rm)$/},
								corrtype=> { type=>SCALAR },
								});
	my ($filer, $conn, $client, $type, $corrtype) = (@{$self}{qw(_dstFiler _dstconn _client)}, @args{qw(type corrtype)});
	my ($count, $interval) = (10, 2);
	my $sisundo_oper = ($type eq "rm")?"performed":"required";
	my $mountpoint = mount($filer, "/vol/".vol_root($conn), client=>$client) or croak("Could not mount root volume of $filer from $client: $@");
	my $path = "/vol/".$self->{_dstvol};
	my (@lines, @inodes, $wrongshrfile, $expmsg, $tail, $undoret);

	$expmsg = Dense::GetMessage(host=>$filer, code=>(($type eq "rm")?"DENSE_UNDO_RM_WRONG":"DENSE_UNDO_WRONG"));
	$tail	= filer_messages->new($filer, [ $expmsg ]);

	Dense::dense_toggle(host=>$filer, path_name=>$path, state=>"DENSE_OFF_NAME");
	if ($type eq "rm") {
		Dense::dense_undo(host=>$filer, path_name=>$path, rm_sharing=>"wrong");
	} else {
		Dense::dense_undo(host=>$filer, path_name=>$path, log_wrong=>1);
	}
	
	$undoret = _get_undo_msg($tail, $count, $interval);

	if($undoret !~ /$expmsg \'(.*)\'./) {
		croak("undo did not return the expected msg $expmsg, but returned $undoret");
	}	
	$wrongshrfile = $1;
	$Log->comment("wrong share listings can be got from $wrongshrfile");

	open UNDOWRONGLIST, "$mountpoint$wrongshrfile";
	@lines = <UNDOWRONGLIST>;
	pop(@lines);

	foreach(@lines) {
		if(/SIS undo $sisundo_oper on inode: (\d+) of type:(.*)/) {
			if($corrtype eq $2) {
				push(@inodes, $1);
			} else {
				croak("SIS undo found corruption message of type $2, but expected $corrtype");
			}
		} else {
			croak("SIS undo found different message ($_) in $wrongshrfile than corruption messages");
		}
	}
	
	close UNDOWRONGLIST;
	$Log->comment("wrong share inodes are ".Dumper(\@inodes));
	return(\@inodes);
}

sub dense_set_refcnt {
	return sub {
		my $self = shift;
		my (%args)	= validate(@_, {inode	=> { type=>SCALAR, regex=>qr/^\d+$/ },
									blkno	=> { type=>SCALAR, default=>0, regex=>qr/^\d+$/ },
									refcnt	=> { type=>SCALAR, default=>255, regex=>qr/^\d+$/ },
									});	
		$Log->comment("dense_set_refcnt called with @_");
		wreck($self->{_dstconn}, $self->{_dstvol}, type=>"flex", inject_script=>"refcount_set inum=$args{inode} bno=$args{blkno} count=$args{refcnt}") 
			or croak("setting refcount on $self->{_dstvol}'s $args{inode} to $args{refcnt} failed because of $@");
	};
}

1;

