# -*- perl -*-
# $Id:$
## @summary rep utility routines for dense scripts
## @usage
##      use Dense::REP;
## @author dl-dense-qa@netapp.com
## @cf     false
## @status public
## @pod here


=head1 NAME

rep utility routines for dense scripts

=head1 DESCRIPTION

This module defines rep utility routines for dense test scripts use.

=head1 FUNCTIONS

=cut

package Dense::REP;

require Exporter;
our @ISA= qw(Exporter);
our @EXPORT = qw (
		    rep_cleanup
		    get_vvbn_list
		    verify_ref_counts
		    rep_check_verify
		    wreck_pseudo_bad_and_vol_online
		    set_rep_testing_flags
		    get_rep_testing_flags
		    search_raid_error_msg
		    get_exp_bad_blk_count
            wafl_flush_cache
            modify_raid_cv_option
		);

### Use: Start ###
use strict;
use Tharn;
use Params::Validate qw(validate validate_pos OBJECT SCALAR HASHREF ARRAYREF UNDEF);
use filer;
use Volume qw(:all);
use Storage::Misc qw(misc_printflag misc_setflag);
use Wreck;
use Wafliron;
use filecheck;
use Dense qw (:all);
use Dense::Utils;
use Dense::Constants;
use Dense::Constant::Message;
use Dense::Constant::Message qw($DENSE_ONTAP_VERSION);
use Dense::Constant::Command;
use Dense::storagemotion;
use POSIX;
use Data::Dumper;
use Data::Compare;

use vars qw($DENSE_ONTAP_VERSION);
### Use: End ###

### Functions: Start ###

###########################################################################
######  To cleanup possible left overs by the script after rep tests ######
###########################################################################
=head2 rep_cleanup(conn=>$conn, vol=>$vol1, sis_overlap_requests=>"on", 
	    flag_panic_testing_val=>$flag_panic_val, 
	    flag_wafliron_panic_val=>$flag_wafliron_val, client=>$CLIENT)

name : rep_cleanup  

Description:
    To cleanup possible left overs by the script after rep tests
        1. volume destroy after waiting for the sis to complete
        2. sis overlap requests
        3. panic testing flag
        4. wafliron panic flag
        5. Wafliron run on the aggr if failed to destory the volume
 
Return value: 
    Die will be properated for the calling script to handle
=cut 

sub rep_cleanup ($%) {
    my (%opts) = validate(@_, {
            conn                    => {type => HASHREF},
            vol                     => {type => HASHREF | ARRAYREF},
            sis_overlap_requests    => {type => SCALAR, optional => 1, default => "on", regex => qr/^o(n|ff)$/},
            flag_panic_testing_val  => {type => SCALAR, regex => qr/^(0|1)$/},
            flag_wafliron_panic_val => {type => SCALAR, regex => qr/^(0|1)$/},
            client                  => { type => SCALAR },
                        });

    my ($conn, $overlap, $vol, $volname, $volpath, $aggrname, $filer_name, $client, $hostspec);
    my ($flag_panic_testing, $flag_panic_testing_val, $flag_wafliron_panic, $flag_wafliron_panic_val, $vol_list_ref, @sis_status, $each_vol);

    $conn                    = $opts{'conn'};
    $filer_name              = $conn->{'hostp'}->{'hostname'};
    $hostspec                = host($filer_name);
    $overlap                 = $opts{'sis_overlap_requests'};
    $vol                     = $opts{'vol'};
    $flag_panic_testing      = "wafl_inconsistent_panic_testing";
    $flag_wafliron_panic     = "wafl_iron_panic_on_corruption";
    $flag_panic_testing_val  = $opts{'flag_panic_testing_val'};
    $flag_wafliron_panic_val = $opts{'flag_wafliron_panic_val'};
    $client                  = $opts{'client'};
    @sis_status              = ( 'IDLE', 'NOSTATUS' );


    if ( ref($vol) =~ /ARRAY/i ) {
        logcomment("The type of vol input is array ref");
        $vol_list_ref = $vol;
    } else {
        logcomment("The type of vol input is not array ref, hence converting it to array ref");
        push( @$vol_list_ref, $vol );
    }
    
    set_overlap_requests(conn => $conn, state => $overlap);
    foreach $each_vol (@$vol_list_ref) {
        $volname = $each_vol->{'name'};
        $volpath = "/vol/" . $volname;
        if ( is_dense_enabled( host => $filer_name, vol_path => $volpath ) ) {
            dense_stop( host => $filer_name, path_name => $volpath );
            dense_wait_for_status(host => $filer_name, path_name => $volpath, status => \@sis_status);
        } else {
            dense_wait_for_status(host => $filer_name, path_name => $volpath, status => \@sis_status);
        }
        modify_raid_cv_option(conn=>$conn, name=>$volname, option=>'on');
        wafl_flush_cache($conn);
        eval {
            # Putting into eval as non-destruction of this should not lead to script failure
            unmount_and_cleanup_vol( $filer_name, $volpath, $each_vol, $client );
        };
        if ($@) {
            logcomment("ERROR: $@");
        }
    }

    #Run wafliron to clean any corruption that exists
    foreach $each_vol (@$vol_list_ref) {
        $volname  = $each_vol->{'name'};
        $volpath  = "/vol/" . $volname;
        $aggrname = $each_vol->{'aggrName'};

        if ( vol_exists( $conn, $volname ) ) {
            logcomment("Was not able to destroy vol hence running wafliron to clear any corruptions left behind by this test on aggr $aggrname");
            wafliron_start( $conn, $aggrname );
        } else {
            logcomment("Ignoring wafliron run on aggr $aggrname as vol created for this test was successfully destroyed");
        }
    }

    if (misc_setflag(host => $hostspec, flag => $flag_panic_testing, value => $flag_panic_testing_val)) {
        logcomment("Successfully restored default value of flag $flag_panic_testing");
    } else {
        logcomment("ERROR: Unable to restore default value of flag $flag_panic_testing");
    }

    if (misc_setflag(host => $hostspec, flag => $flag_wafliron_panic, value => $flag_wafliron_panic_val)) {
        logcomment("Successfully restored default value of flag $flag_wafliron_panic");
    } else {
        logcomment("ERROR: Unable to restore default value of flag $flag_wafliron_panic");
    }
    
}

################################################################
########   To get the VVBN list for the inputted fbns   ########
################################################################
=head2 get_vvbn_list(conn=>$conn, file_path=>$scr_file_path, blk_list=>\@blk_list)

name : get_vvbn_list  

Description:
    To get the VVBN list for the inputted fbns in the input array list
 
Return value: 
    Array reference to the list of vvbns
 
=cut 

sub get_vvbn_list($%) {
    my (%opts) = validate(@_, {
            conn      => { type => HASHREF },
            file_path => { type => SCALAR },
            blk_list  => { type => ARRAYREF },
            rep_vvbn  => { type => SCALAR, optional => 1, default => 0, regex => qr/^(0|1)$/ },
            vol_name  => { type => SCALAR, optional => 1, default => UNDEF },
        }
    );
    my ($conn, $file_path, $blk_list, $rep_vvbn, %parse_vbns, $blk_no, @vvbn_list, $vol_name);

    $conn      = $opts{'conn'};
    $file_path = $opts{'file_path'};
    $blk_list  = $opts{'blk_list'};
    $rep_vvbn  = $opts{'rep_vvbn'};
    $vol_name  = $opts{'vol_name'};

    if (defined($vol_name)) {
        modify_raid_cv_option(conn=>$conn, name=>$vol_name, option=> 'off');
        wafl_flush_cache($conn);
    }
    %parse_vbns = dense_parse_showvbns (
		    conn        => $conn,
		    file_name   => $file_path,
		    fbn_range   => $blk_list,
		);
    if (defined($vol_name)) {
        modify_raid_cv_option(conn=>$conn, name=>$vol_name, option=> 'on');
        wafl_flush_cache($conn);
    }
    logcomment("parse_vbns = ".Dumper(\%parse_vbns));

    foreach $blk_no (@$blk_list) {
        push( @vvbn_list, $parse_vbns{$blk_no}{'vvbn'} );
        if ($rep_vvbn) {
            for ( my $i = 0 ; $i < $parse_vbns{$blk_no}{'refcount'} ; $i++ ) {
                push( @vvbn_list, $parse_vbns{$blk_no}{'vvbn'} );
            }
        }
    }

    return (\@vvbn_list);
}

################################################################
########        To verify the ref counts of fbns        ########
################################################################
=head2 verify_ref_counts(conn=>$conn, file_path=>$sf_path, block_list=>\@blk_list, expected_ref_counts=>\@ref_cnt)

name : verify_ref_counts  

Description:
    To verify the ref counts of fbns defined by the input expected ref counts

Return value: 
    Success: 1
    Failure: 0

=cut 

sub verify_ref_counts($%) {
    my (%opts) = validate( @_, {
            conn                => { type => HASHREF },
            file_path           => { type => SCALAR },
            block_list          => { type => ARRAYREF },
            expected_ref_counts => { type => ARRAYREF },
            vol_name            => { type => SCALAR, optional => 1, default => UNDEF },
        }
    );

    my ( $conn, $file_path, $blk_list, $expected_ref_counts, %parse_vbns,
        $vol_name, $blk_no, @ref_count_list, $result );

    $result              = 0;
    $conn                = $opts{'conn'};
    $file_path           = $opts{'file_path'};
    $blk_list            = $opts{'block_list'};
    $expected_ref_counts = $opts{'expected_ref_counts'};
    $vol_name            = $opts{'vol_name'};

    if (defined($vol_name)) {
        modify_raid_cv_option(conn=>$conn, name=>$vol_name, option=>'off');
        wafl_flush_cache($conn);
    }

    %parse_vbns = dense_parse_showvbns(
        conn      => $conn,
        file_name => $file_path,
        fbn_range => $blk_list,
    );

    if ( defined($vol_name) ) {
        modify_raid_cv_option(conn => $conn, name => $vol_name, option => 'on');
        wafl_flush_cache($conn);
    }

    logcomment( "parse_vbns = " . Dumper( \%parse_vbns ) );

    foreach $blk_no (@$blk_list) {
        push( @ref_count_list, $parse_vbns{$blk_no}{'refcount'} );
    }

    if ( Compare( $expected_ref_counts, \@ref_count_list ) ) {
        $result = 1;
        logcomment("The ref counts of the corrupted fbns are as expected");
    } else {
        logcomment("Expected ref cnt: \n @$expected_ref_counts, got: \n @ref_count_list");
        die("The ref counts of the corrupted fbns are not as expected");
    }

    return ($result);
}

######################################################################
### To verify the rep errors in a file with the file check utility ###
######################################################################

=head2 rep_check_verify(conn=>$conn, src_file_path=>$sf_path, 
	    expected_bad_cnt=>$bad_count, volname=>$volName, 
	    validate_vvbn=>1, expected_vvbnlist=>vvbn_list)

name : rep_check_verify  

Description:
    To Check that the rep errors are present for particular vvbns in a file 
    using the file check utility

Return value: 
    Success: 1
    Failure: 0

=cut 

sub rep_check_verify($%) {
    my (%opts) = validate (@_, {
		    conn		    => { type=>HASHREF },
		    src_file_path	    => { type=>SCALAR },
		    expected_bad_cnt	    => { type=>SCALAR },
		    volname		    => { type=>SCALAR },
		    validate_vvbn	    => { type=>SCALAR, regex=>qr/^(0|1)$/ },
		    expected_vvbnlist	    => { type=>ARRAYREF, optional=>1 },
		    timeout		    => { type=>SCALAR, optional=>1 ,default=>DENSE_DEFAULT_TIMEOUT },
		    });

    my ($conn, $src_file_path, $expected_bad_cnt, $volname, $validate_vvbn, $expected_vvbnlist, $filer_name, $timeout);
    $conn = $opts{'conn'};
    $src_file_path = $opts{'src_file_path'};
    $expected_bad_cnt = $opts{'expected_bad_cnt'};
    $volname = $opts{'volname'};
    $validate_vvbn = $opts{'validate_vvbn'};
    $expected_vvbnlist = $opts{'expected_vvbnlist'};
    $filer_name = $conn->{'hostp'}->{'hostname'};
    $timeout = $opts{'timeout'};

    my ($fc_error, $fc_result) = file_check(conn=>$conn, args=>$src_file_path);

    logcomment("verifying the expected rep corruption");

    if ($fc_error) {
        die "file check error: $fc_result";
    } else {
        logcomment("File check completed sucessfully, Output can be found at: $fc_result");
    }

    wait_for_scan_done(host=>$filer_name, volname=>$volname, scan=>"file block check", timeout=>$timeout);
    #Waiting for the file to get synced
    filer_sync($conn); filer_sync($conn);

    my ($fc_bad_cnt, $fc_total, $fc_vvbnlist, @fc_snapshot_list) = file_check_read_output(conn=>$conn, path=>$fc_result);

    if ($fc_bad_cnt == $expected_bad_cnt) {
        logcomment("Got the expected reped bad count value of $expected_bad_cnt");
    } else {
        die("Did not get the expected reped bad count value of $expected_bad_cnt");
    }

    logcomment("Acutal VVBN list: ".Dumper($fc_vvbnlist));
    if ($validate_vvbn) {
        if (not defined $expected_vvbnlist) {
            die("Expected VVBN list not given to validate VVBNs");
        }
        @$fc_vvbnlist = sort(@$fc_vvbnlist);
        @$expected_vvbnlist = sort(@$expected_vvbnlist);

        logcomment("Expected VVBN list: ".Dumper($expected_vvbnlist));
        if(Compare($fc_vvbnlist, $expected_vvbnlist)) {
           logcomment("Got the expected list of corrupted VVBN")
        } else {
           die("Did not get the expected list of corrupted VVBN")
        }
    }
}

#######################################################################
## To introduce pseudo_bad corruption using wreck and online the vol ##
#######################################################################

=head2 wreck_pseudo_bad_and_vol_online(conn=>$conn, volname=>$volName, 
voltype=>"flex", inum=>$file_inode_no, block_level=>0,
block_index=>\@index)


name : wreck_pseudo_bad_and_vol_online  

Description:
    To introduce pseudo_bad corruption using wreck and online the vol 
    for the list of fbns

Return value: 
    Die will be propagated for the calling script to handle
=cut 

sub wreck_pseudo_bad_and_vol_online($%) {
    my (%opts) = validate (@_, { 
		conn        => { type=>HASHREF },
		volname	    => { type=>SCALAR },
		voltype	    => { type=>SCALAR, optional=>1, default=>"flex" },
		inum	    => { type=>SCALAR },
		block_level => { type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/ },
		block_index => { type=>ARRAYREF}, 
		timeout	    => { type=>SCALAR, optional=>1, default=>DENSE_DEFAULT_TIMEOUT },
		    });

    my($conn, $volname, $voltype, $inum, $blk_inds, $block_level, $blk_ind, $timeout);
    $conn = $opts{'conn'};
    $volname = $opts{'volname'};
    $voltype = $opts{'voltype'};
    $inum = $opts{'inum'};
    $blk_inds = $opts{'block_index'};
    $block_level = $opts{'block_level'};
    $timeout = $opts{'timeout'};

    logcomment("Introducing pseudo_bad corruption at inode num $inum");
    foreach $blk_ind (@$blk_inds) {
        eval {
            wreck($conn, $volname, type=>$voltype, timeout=>$timeout, inject_script=>"pseudo_bad inum=$inum, level=$block_level, bno=$blk_ind");
        };
        if($@) {
            die("Corruption could not be introduced at blk $blk_ind on inode $inum, Reason : $@");
        } else {
            logcomment("Corruption introduced successfully at blk $blk_ind on inode $inum ");
        }

        if (vol_online($conn, $volname) ) {
            logcomment("Onlined the volume ".$volname);
        } else {
            die("Unable to online volume ".$volname);
        }
    }
}


#############################################################
#####      To set the flags required for rep testing    #####
#############################################################

=head2 set_rep_testing_flags(host=>$host)

name : set_rep_testing_flags  

Description:
    To set the flags required for rep testing like 
    wafl_inconsistent_panic_testing  & "wafl_iron_panic_on_corruption"

Return value: 
    Die will be propagated for the calling script to handle

=cut 

sub set_rep_testing_flags($%) {
     my (%opts) = validate (@_, {
		    host	=> { type=>HASHREF },
		});

    my ($host, $flag_panic_testing, $flag_wafliron_panic);
    $host = $opts{'host'};
    $flag_panic_testing = "wafl_inconsistent_panic_testing";
    $flag_wafliron_panic = "wafl_iron_panic_on_corruption";


    if(misc_setflag(host=>$host, flag=>$flag_panic_testing, value=>"1")) {
        logcomment("Successfully set value of flag $flag_panic_testing");
    } else {
        die("Unable to set value of flag $flag_panic_testing");
    }

    if(misc_setflag(host=>$host, flag=>$flag_wafliron_panic, value=>"0")) {
        logcomment("Successfully set value of flag $flag_wafliron_panic");
    } else {
        die("Unable to set value of flag $flag_wafliron_panic");
    }

}


#############################################################
#####      To get the flags required for rep testing    #####
#############################################################

=head2 get_rep_testing_flags(host=>$host)

name : get_rep_testing_flags

Description:
    To get the value of flags
    wafl_inconsistent_panic_testing  & "wafl_iron_panic_on_corruption"

Return value:
    (wafl_inconsistent_panic_testing, wafl_iron_panic_on_corruption)

=cut

sub get_rep_testing_flags($%) {
     my (%opts) = validate (@_, {
                    host        => { type=>HASHREF },
                });

    my ($host, $flag_panic_testing, $flag_wafliron_panic, $flag_panic_testing_val, $flag_wafliron_panic_val);
    $host = $opts{'host'};
    $flag_panic_testing = "wafl_inconsistent_panic_testing";
    $flag_wafliron_panic = "wafl_iron_panic_on_corruption";

    $flag_panic_testing_val = misc_printflag(host=>$host, flag=>$flag_panic_testing);
    $flag_wafliron_panic_val = misc_printflag(host=>$host, flag=>$flag_wafliron_panic);

    return ($flag_panic_testing_val, $flag_wafliron_panic_val);
}

#################################################################
## To search for the required raid messages in the message log ##
#################################################################

=head2 search_raid_error_msg(err_msgs=>\@err_msgs, volname=>$volName, blk_nos=>\@fbn_blocks)

name : search_raid_error_msg  

Description:
    To search for the required raid messages in the message log

Return value: 
    Die will be propagated for the calling script to handle

=cut 

sub search_raid_error_msg($%) {
    my (%opts) = validate (@_, {
			err_msgs => { type=>ARRAYREF },
    		    	volname => { type=>SCALAR },
    		    	blk_nos=> { type=>ARRAYREF },
		    });
    
    my ($err_msgs, $msg, $vol, $blk_nos, $blk, $raid_error_msg, $raid_error_class, %exp_hash);
    
    $err_msgs = $opts{'err_msgs'};
    $vol = $opts{'volname'};
    $blk_nos = $opts{'blk_nos'};
    $raid_error_msg = "Reading bad block on volume";
    $raid_error_class = "raid.cksum.bad.file.block";
    @exp_hash{@$blk_nos} = (1) x @$blk_nos;
    
    foreach $msg (@$err_msgs) {
	if (($msg =~ /$raid_error_class/i) && ($msg =~ /$raid_error_msg/i) && ($msg =~ /$vol/i)) {
	   logcomment("Got the required error message: $msg");
    	   
    	   if ( $msg =~  m/inode number (\d+), snapid (\d+), file block (\d+), level (\d+)/i ) {
		logcomment("inode number: $1;  snapid: $2; file block: $3; level: $4;");
    	   	delete $exp_hash{$3};
    	   }
    	
    	}
    }
    
    if (scalar(keys %exp_hash) != 0) {
	logcomment("Dint get ems for all the block(s), following are the blocks for missing ems: ".Dumper(\%exp_hash));
    	die("Dint get ems for all the block(s)");
    } else {
	logcomment("Got ems for all the  expected blocks(s)");
    }
}


###############################################################################
######## To get the total expected bad block count from a list of fbns ########
###############################################################################
=head2 get_exp_bad_blk_count(conn=>$conn, file_path=>$scr_file_path, blk_list=>\@blk_list)

name : get_exp_bad_blk_count

Description:
    To get the total expected bad block count from a list of fbns

Return value:
    The total expected bad block count.

=cut


sub get_exp_bad_blk_count($%) {
    my (%opts) = validate (@_, {
		    conn        => { type=>HASHREF },
		    file_path   => { type=>SCALAR },
		    blk_list    => { type=>ARRAYREF },}
		);
        
    my ($conn, $file_path, $blk_list);
    my ($blk_no, $bad_count, %l_show_vbns);
    
    $conn = $opts{'conn'};
    $file_path = $opts{'file_path'};
    $blk_list = $opts{'blk_list'};
    $bad_count = 0;
    
    %l_show_vbns = dense_parse_showvbns (
			conn        => $conn,
			file_name   => $file_path,
			fbn_range   => $blk_list,
		    );
    logcomment("l_show_vbns = ".Dumper(\%l_show_vbns));
    
    foreach $blk_no (@$blk_list) {
    	if ($l_show_vbns{$blk_no}{'vvbn'} != 0) {
	   $bad_count++;
	   $bad_count += $l_show_vbns{$blk_no}{'refcount'};
    	}
    }
    
    return $bad_count;
}


################################################################
########            To flush the WAFL cache             ########
################################################################
=head2 wafl_flush_cache($conn)

name : wafl_flush_cache  

Description:
    To flush the WAFL cache 

Return value: 
    None

=cut 

sub wafl_flush_cache($) {
    my ($conn) = @_;

    filer_cli_test_say($conn, 'wafl flush cache');

    return;
}


################################################################
########               To modify raid cv                ########
################################################################
=head2 modify_raid_cv_option(conn=>$conn, name=>$vol_name, option=>$sf_path)

name : modify_raid_cv_option  

Description:
    To modify raid cv 

Return value: 
    None

=cut 

sub modify_raid_cv_option($%) {
    my (%opts) = validate( @_, {
            conn   => { type => HASHREF },
            name   => { type => SCALAR },
            option => { type => SCALAR, optional => 1, default => 'on', regex => qr/on|off/ },
        }
    );

    my ( $conn, $name, $option );
    $conn   = $opts{'conn'};
    $name   = $opts{'name'};
    $option = $opts{'option'};

    filer_cli_test_say( $conn, "vol options $name raid_cv $option" );

    return;
}


1;
