# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
# $Id:$

## @summary  Upgrade/Revert Library
##
## @description
##     Dense::SAN library, provides API's for common SAN related functions.
##     Functions provided include
##     - Setting up iscsi on filer
##     - Mounting a lun on client
##     - Populating a lun
##     - Cleaning up iscsi setup
##     - Finding out the df of the lun device mounted on client
##
## @usage
##    use  Dense::SAN;
##
## @keywords dense san
##
## @dependencies
##
## @author dl-symphony-qa-automation@netapp.com
##

package Dense::SAN;

use strict;
use POSIX;
use base qw(Exporter);
use Dense::Constants;
use Dense::Constant::Command;
use Utils::Units qw (convert_to_bytes);
use SAN qw(lun_disks mkfs_mount_iscsi);
use Utils qw(make_conn);
use blocks_filer qw( filer_iscsi_start filer_iscsi_security_default_set);
use NOTE;
use SayWrap qw(sayEOF);
use ExportFS qw( export_path );
use Utils::Units qw(:constants);
use Tharn;
use filer;
use FS::Util::Utilities qw(:utils);
use Volume qw(:all);
use Params::Validate qw(validate validate_pos OBJECT SCALAR ARRAYREF HASHREF UNDEF);

BEGIN {
    our @EXPORT = qw(
      dense_populate_lun
      dense_iscsi_cleanup
      get_df
      dense_lun_mount
      dense_filer_iscsi_setup

      );
} ## end BEGIN

##############################################################################
# Function Name : dense_populate_lun()
# Parameters :
#        filer        => $FILER,
#        lun_name    => $LUN_NAME,
#        size         => $size,
#        lun_group    => $LUN_GROUP,
#        vol_name    => $VOL_NAME,
#        timeout    => $TIMEOUT
#
# Return Value : returns 1 on sucess or dies on failure
#
##############################################################################

=head2 dense_populate_lun
This routine populates the data to a lun.

returns 1 on sucess or dies on failure  

=over

=item Example

=back

=for html <hr>

=cut

sub dense_populate_lun {
    my (%opts) = validate( @_,
        {
            filer         => { type => SCALAR },
            lun_name   => { type => SCALAR },
            vol_name => { type => SCALAR },
            size  => { type => SCALAR, optional =>1 },
            mountdir  => { type => SCALAR, optional => 1 },
            disk      => { type => SCALAR , optional => 1},
            client    => { type => SCALAR, optional => 1, default => 'localhost' },
            time_out  => { type => SCALAR, optional => 1, default => DENSE_DEFAULT_TIMEOUT },
        }
    );
    
    my ($filer)   = $opts{filer};
    my ($lunname) = $opts{lun_name};
    my ($size)    = $opts{size};
    my ($volname) = $opts{vol_name};
    my ($mountdir) = $opts{mountdir};
    my ($disk)     = $opts{disk};
    my ($client)   = $opts{client};
    my ($timeout)  = $opts{time_out} || 1800;
    my ($cc, $fs_avail, $data_size, $cmd, $cmd_umount);

    eval {

        #Map the lun
        if (!$mountdir && !$disk) {
            ($mountdir, $disk) =
              dense_lun_mount(filer    => $filer,
                              lun_name => $lunname,
                              vol_name => $volname,
                              client   => $client,
                              timeout  => $timeout);
        }
        $cc = make_conn($client);

        $fs_avail = get_df($cc, $mountdir, $disk, "Avail");
        $data_size = $size;
        $data_size = $fs_avail if !$data_size;
        $data_size = convert_to_bytes($data_size)/1024;

        logcomment("Filling the LUN upto $data_size ...");
        $cmd =
            "/usr/local/bin/sudo dd if=/dev/zero "
          . " of=$mountdir/${lunname} "
          . "bs=1k count=$data_size";
        $timeout = $timeout + $data_size;
        sayEOF($cc, $cmd, timeout => $timeout);

        logcomment("Unmounting $disk after succssful write");
        $cmd_umount = "/usr/local/bin/sudo umount -l $disk";
        sayEOF($cc, $cmd_umount, timeout => $timeout);

    };

    #die if any thing goes wrong here
    die $@ if $@;

    return 1;
} ## end sub dense_populate_lun

##############################################################################
# Function Name : dense_iscsi_cleanup()
# Parameters :
#        conn        => $filer_conn,
#        lun_name    => $LUN_NAME,
#        lun_group    => $LUN_GROUP,
#        vol_name    => $VOL_NAME,
#        timeout    => $TIMEOUT
#
# Return Value : returns 1 on sucess or dies on failure
#
##############################################################################

=head2 dense_iscsi_cleanup
This routine cleans up all iscsi related setup done in iscsi setup. 


returns 1 on sucess or dies on failure  

=over

=item Example

=back

=for html <hr>

=cut

sub dense_iscsi_cleanup {
    my (%opts) = validate( @_,
    {
        conn => { type => HASHREF },
        lun_name   => { type => SCALAR },
        lun_group  => { type => SCALAR },
        vol_name => { type => SCALAR },
        timeout  => { type => SCALAR, optional => 1, default => DENSE_DEFAULT_TIMEOUT },
    }
    );

    my ($filer_conn) = $opts{conn};
    my ($lun_name)   = $opts{lun_name};
    my ($lun_group)  = $opts{lun_group};
    my ($volname)    = $opts{vol_name};
    my ($timeout)    = $opts{timeout};

    my ($cmd_priv, $cmd_iscsi, $output);

    logcomment("unmapping lun $lun_name outof igroup $lun_group");
    $cmd_iscsi = "lun unmap /vol/$volname/$lun_name $lun_group;";
    $output    = filer_cli_test_say($filer_conn, $cmd_iscsi);
    if ($output !~ /result=0/) {
        logcomment("Failed to unmap lun $lun_name.");
    }

    logcomment("Destroying igroup $lun_group");
    $cmd_iscsi = "igroup destroy -f $lun_group;";
    $output    = filer_cli_test_say($filer_conn, $cmd_iscsi );
    if ($output !~ /result=0/) {
        logcomment("Failed to destroy lungroup: $lun_group.");
    }

    logcomment("Stop iscsi service");
    $cmd_iscsi = "iscsi stop;";
    $output    = filer_cli_test_say($filer_conn,$cmd_iscsi );
    if ($output !~ /result=0/) {
        logcomment("Failed to stop iscsi.");
    }
    return 1;

} ## end sub dense_iscsi_cleanup

##############################################################################
# Function Name : get_df()
# Parameters :
#     $cc = client connection
#     $path = path to the device
#     $disk = disk name 
#     $field = name of the feild (Size|Used| Avail)  
#     $timeout = time out in seconds
#
# Return Value : returns requested df field counter on sucess or -1 on failure
#
##############################################################################

=head2 get_df 
Returns requested client df field counter on sucess or -1 on failure 

=over

=item Example

=back

=for html <hr>

=cut

sub get_df($$$$;$) {
    my ($cc, $path, $disk, $field, $timeout) = @_;

    
    my ($cmd, @fields, $diskfree);
    
    $timeout = 300 if (!$timeout);
    
    $cmd = "df $path";
    $diskfree = sayEOF($cc, $cmd, TIMEOUT => $timeout);

    $diskfree =~ /($disk\s+.*$path)/gm;

    @fields = split(/\s+/, $1);
    if ($field eq "Size") {
        return $fields[1];
    }elsif ($field eq "Used") {
        return $fields[2];
    }else {
        return $fields[3];
    }
    return -1;
} ## end sub get_df

##############################################################################
# Function Name : dense_lun_mount
# Parameters :
#        filer        => $FILER,
#        lun_name    => $LUN_NAME,
#        vol_name    => $VOL_NAME,
#        client        => $CLIENT,
#        timeout    => $TIMEOUT
#
# Return Value : returns the "block device" and "mountpoint" where the device is mounted or dies on failure
#
##############################################################################

=head2 dense_lun_mount 
This subroutine will export the lun from filer. Scan the device to find the exported lun device 
from client and format and then mount it.

returns the "block device" and "mountpoint" where the device is mounted or dies on failure 

=over

=item Example

=back

=for html <hr>

=cut

sub dense_lun_mount {
    my (%opts) = validate( @_,
    {
        filer => { type => SCALAR },
        lun_name   => { type => SCALAR },
        vol_name => { type => SCALAR },
        client => { type => SCALAR, optional => 1, default => 'localhost'   },
        timeout  => { type => SCALAR, optional => 1, default => DENSE_DEFAULT_TIMEOUT },
    }
    );

    my ($filer)   = $opts{filer};
    my ($lunname) = $opts{lun_name};
    my ($volname) = $opts{vol_name};
    my ($client)  = $opts{client};
    my ($timeout) = $opts{timeout};


    my $cc         = make_conn($client);
    my $filer_conn = make_conn($filer);

    export_path($filer_conn, "/vol/$volname/$lunname")
      or die "Unable to export /vol/$volname/$lunname";

    my ($cmd_iscsi, $output) ;
    $cmd_iscsi = "/usr/local/bin/sudo /etc/init.d/iscsi stop";
    sayEOF($cc, $cmd_iscsi, TIMEOUT => $timeout);
    $cmd_iscsi = "/usr/local/bin/sudo /etc/init.d/iscsi start";
    sayEOF($cc, $cmd_iscsi, TIMEOUT => $timeout);
                

    # find the lun partition based on lun name in santool
    my @lun_dev = lun_disks($filer, $client, "/vol/$volname/$lunname",
                            sanlun_path => '/opt/netapp/santools');
    my $disk = $lun_dev[0];
    if (!$disk) {
        die("couldn't find lun disk");
    }
    logcomment("lun device is $disk");

    # Nb: mkfs cmd already ensures we aren't formatting any mounted disk
    # make a filesystem.
    logcomment("creating ext2 fs on $disk");
    my $mountdir = mkfs_mount_iscsi($filer, $client, $disk, fstype => "ext2", timeout =>3600)
      or die "Failed to mkfs";

    return ($mountdir, $disk);
} ## end sub dense_lun_mount

##############################################################################
# Function Name : dense_filer_iscsi_setup
# Parameters :
#        conn    => $filer_conn,
#        lun_name    => $LUN_NAME,
#        lun_group    => $LUN_GROUP,
#        lun_platform    => $LUN_PLATFORM,
#        vol_name    => $VOL_NAME,
#        iname        => $INITIATOR_NAME,
#        timeout    => $TIMEOUT
#
# Return Value : returns 1 on sucess or dies on failure
#
##############################################################################

=head2 dense_filer_iscsi_setup 
This routine sets up iscsi on filer. It starts iscsi servide on filer. sets the security.
creates the igroup, and maps the lun to igroup.

returns 1 on sucess or dies on failure  

=over

=item Example

=back

=for html <hr>

=cut

#dense_filer_iscsi_setup  ($conn,)
sub dense_filer_iscsi_setup {

    my (%opts) = validate( @_,
    {
        conn => { type => HASHREF },
        lun_name   => { type => SCALAR },
        lun_group   => { type => SCALAR },
        lun_platform   => { type => SCALAR },
        vol_name => { type => SCALAR },
        iname => { type => SCALAR  },
        timeout  => { type => SCALAR, optional => 1, default => DENSE_DEFAULT_TIMEOUT },
    }
    );


    my ($filer_conn)   = $opts{conn};
    my ($lun_name)     = $opts{lun_name};
    my ($lun_group)    = $opts{lun_group};
    my ($lun_platform) = $opts{lun_platform};
    my ($volname)      = $opts{vol_name};
    my ($iname)        = $opts{iname};
    my ($timeout)      = $opts{timeout};

    my ($cmd_iscsi);

    logcomment("Starting iscsi service");
    filer_iscsi_start($filer_conn);

    # security to default
    filer_iscsi_security_default_set ($filer_conn, "none");
    
    logcomment("Creating igroup $lun_group");
    $cmd_iscsi = "igroup destroy $lun_group; ";
    filer_cli_test_say($filer_conn, $cmd_iscsi );

    $cmd_iscsi = "igroup create -i -t $lun_platform $lun_group $iname;";
    filer_cli_test_say($filer_conn, $cmd_iscsi);

    #lun map
    logcomment("mapping lun $lun_name into igroup $lun_group");
    $cmd_iscsi = "lun map /vol/$volname/$lun_name $lun_group;";
    filer_cli_test_say ($filer_conn, $cmd_iscsi);

    return 1;
} ## end sub dense_filer_iscsi_setup
1;

