# -*- perl -*-
## @summary test data generation routines for dense volume 
## @usage
##	use Dense::TestData;
## @author dl-dpgqa-automation@netapp.com rani@netapp.com abhayjit@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Utility routines for generation of dense test data.

=head1 DESCRIPTION

This module defines routines for dense test scripts use
to generate test data.

=head1 FUNCTIONS

=cut

package Dense::TestData;

## use Tharn qw(connect host say logcomment logfatal logresult param tstlog);
use Tharn;
use filer;
use Utils qw(make_conn);
use Data::Dumper qw(Dumper);
use Dense qw(:all);
use Dense::Utils; 
use Volume qw(vol_size);
use VolCmds qw(diskFree);
use FS::Util::Utilities qw(deltree);
use Carp qw(croak);
use cat qw(do_shell);

require Exporter;
our @ISA= qw(Exporter);
our @EXPORT = qw ( 
		   dense_mk_dataset
		   dense_small_dataset
		   dense_mk_dup_datafile
		   dense_mk_nondup_datafile
		   write_dup_blks
		   write_unique_blks
		   dense_mk_large_mixed_datafile
		   dense_mk_small_datafile
		   populate_vol
		   modify_vol
		   write_to_file
		   );

use SPARSE::Resource::Volume;
use SPARSE::Util::FileIO;
use SPARSE::Util::Logger;
use Params::Validate qw(validate validate_pos SCALAR ARRAYREF HASHREF UNDEF);
BEGIN {
Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
}

use strict;

use vars qw($dir_path $client);
my $BLKSIZE = 4096;
my $MAX_SEGMENT = 16;

=head2 dense_mk_dataset (vol=>$vol)

Make a data set with duplicate data blocks within files and across files.
The new dataset will be created in the 'TEST' directory in the volume.

=over 4

=item vol 

Volume object, required.

=item Returns:

A set of file object

=back

=cut

sub dense_mk_dataset(%)
{
    my (%opts) = validate (@_, { vol => { type=>HASHREF },
				 client => { type=>SCALAR | UNDEF, optional=>1, default=>'localhost' } } );
    
    my ($vol) = $opts{vol};
    my $i;
    my @fs;

    my @files = ("dense_file_0",
		 "dense_file_1", 
		 "dense_file_2", 
		 "dense_file_3",
		 "dense_file_4", 
		 "dense_file_5",
		 "dense_file_6"
		 );

    # $vol->make_testdir();
    Dense::Utils::create_testdir(vol=>$vol, client=>$opts{client});

    # Small file (level 0)
    $fs[0] = fioOpen($vol, $files[0], ">");
    $fs[0]->WriteSeq(32, 0);

    # 4KB file, all data (level 1)
    $fs[1] = fioOpen($vol, $files[1], ">");
    $fs[1]->WriteSeq(4096, 0);

    # 64KB file, all data (level 2)
    $fs[2] = fioOpen($vol, $files[2], ">");
    for ($i = 0; $i < 16; $i++) {
	$fs[2]->WriteSeq(4096, 0);
    }

    # 256KB file, all data 
    $fs[3] = fioOpen($vol, $files[3], ">");
    for ($i = 0; $i < 16; $i++) {
	$fs[3]->WriteSeq(4*4096, 0);
    }

    # 1MB file, all data 
    $fs[4] = fioOpen($vol, $files[4], ">");
    for ($i = 0; $i < 16; $i++) {
	$fs[4]->WriteSeq(16*4096, 0);
    }

    # 32MB file, mostly holes (level 3)
    $fs[5] = fioOpen($vol, $files[5], ">");
    $fs[5]->WriteSeq(16*4096, 0);
    $fs[5]->Seek(16*4096*4096, 1);
    $fs[5]->WriteSeq(16*4096, 0);

    # Large random file, holes and data
    $fs[6] = fioOpen($vol, $files[6], ">");
    for ($i = 0; $i < 16; $i++) {
	srand;
	my $num = int(rand $MAX_SEGMENT);
	$fs[6]->WriteSeq($num*4096, 0);
	$fs[6]->Seek($num*4096*4096, 1);
    }

    dense_mk_large_mixed_datafile(vol=>$vol, filename=>"largefile", blk_cnt=>262144);
    $_->{fh}->sync() foreach(@fs);	
    return @fs;
}

=head2 dense_small_dataset (vol=>$vol)

Make a data set with duplicate data blocks within files and across files.
The new dataset will be created in the 'TEST' directory in the volume.
Creates a total of  data in the volume

=over 4

=item vol

Volume object, required.

=item host

Filer name, required.

=item smoke

Create dataset for smoke environments, optional, set to 0 by defauult.

=item Returns:

A set of file objects. If smoke is set, returns the list of filepaths

=back

=cut

sub dense_small_dataset(%)
{
    my (%opts) = validate (@_, { vol => { type=>HASHREF },
				 host => { type=>SCALAR }, 
				 smoke => { type=>SCALAR, optional=>1, default=>0 }, 
				 client => { type=>SCALAR | UNDEF, optional=>1, default=>'localhost' } } );
    
    my ($vol) = $opts{vol};
    my $host = $opts{host};
    my $conn = make_conn($host, "#rsh");
    my $i;
    my @fs;
    my $bs = 4096;

    my @files = ("dense_sfile_0",
		 "dense_sfile_1", 
		 "dense_sfile_2", 
		 "dense_sfile_3",
		 "dense_sfile_4", 
		 "dense_sfile_5",
		 "dense_sfile_6"
		 );

    # $vol->make_testdir();
    Dense::Utils::create_testdir(vol=>$vol, client=>$opts{client});

    if(!$opts{smoke}) {

        # Small file (level 0)
        $fs[0] = fioOpen($vol, $files[0], ">");
        $fs[0]->WriteSeq(32, 0);
    
        # 4KB file, all data (level 1)
        $fs[1] = fioOpen($vol, $files[1], ">");
        $fs[1]->WriteSeq(4096, 0);
    
        # 64KB file, all data (level 2)
        $fs[2] = fioOpen($vol, $files[2], ">");
        for ($i = 0; $i < 16; $i++) {
    	$fs[2]->WriteSeq(4096, 0);
        }
    
        # 256KB file, all data 
        $fs[3] = fioOpen($vol, $files[3], ">");
        for ($i = 0; $i < 16; $i++) {
    	$fs[3]->WriteSeq(4*4096, 0);
        }
    
        # 1MB file, all data 
        $fs[4] = fioOpen($vol, $files[4], ">");
        for ($i = 0; $i < 16; $i++) {
    	$fs[4]->WriteSeq(16*4096, 0);
        }
    
        # < 2MB file, mostly holes
        $fs[5] = fioOpen($vol, $files[5], ">");
        $fs[5]->WriteSeq(16*4096, 0);
        $fs[5]->Seek(256*4096, 1);
        $fs[5]->WriteSeq(16*4096, 0);
        $fs[5]->Seek(200*4096, 1);
    
        # random file, holes and data < 1M
        $fs[6] = fioOpen($vol, $files[6], ">");
        for ($i = 0; $i < 16; $i++) {
    	srand;
    	my $MAX=8;
    	my $num = int(rand($MAX));
    	$fs[6]->WriteSeq($num*4096, 0);
    	$fs[6]->Seek($num*2*4096, 1);
        }
		
        $_->{fh}->sync() foreach(@fs);	
	logcomment("[dense_small_dataset]:Data creation complete");
	
    }else {
	$i=0;
	foreach (@files) {
 	    $fs[$i] = $vol->{testDirVolName} . "/" . $files[$i];	
	    $i++;
	}
	filer_cli_test_say($conn, "mkfile 32 $fs[0]");	
	filer_cli_test_say($conn, " mkfile 4k $fs[1]");	
	filer_cli_test_say($conn, " mkfile 64k $fs[2]");	
	filer_cli_test_say($conn, " mkfile 256k $fs[3]");	
	filer_cli_test_say($conn, " mkfile 1m $fs[4]");	

	filer_cli_test_say($conn, " mkfile 2m $fs[5]");
	my $offset = 2 * $bs;
	my $length = 495 * $bs;
	filer_cli_test_say($conn,"file punchhole $fs[5] $offset $length");

	filer_cli_test_say($conn, " mkfile 850k $fs[6]");
	$offset = 10 * $bs;
	$length = 50 * $bs;
	filer_cli_test_say($conn,"file punchhole $fs[6] $offset $length");
	$offset = 100 * $bs;
	filer_cli_test_say($conn,"file punchhole $fs[6] $offset $length");
	
	logcomment("[dense_small_dataset]:Data creation complete");
    }
	
    return @fs;
}

=head2 write_dup_blks(dfile=>$dfile, pattern_size=>$pattern_size, rpt_cnt=>$rpt_cnt, offset=>$offset)

Write duplicated data blocks. The unique pattern
size can't exceed 5 blocks (20480 bytes), otherwise
it will start to repeat the pattern from the beginning.

=over 4

=item dfile 

File object of the output file, required.

=item pattern_size 

Byte size of the unique data pattern, must be multiple of 4K. Required.

=item rpt_cnt 

Number of times the unique pattern repeats, required.

=item offset 

Starting offset into the Seq array, to provide some randomness. 
Value range between 0-79, default is 0. Optional.

=back

=cut

sub write_dup_blks(%)
{
    my (%opts) = validate (@_, { dfile => { type=>HASHREF },
				 pattern_size => { type=>SCALAR },
				 rpt_cnt => { type=>SCALAR },
				 offset => { type=>SCALAR, default=>"0", optional=>1 } }
			  );

    my ($dfile) = $opts{dfile};
    my ($pattern_size) = $opts{pattern_size};
    my ($rpt_cnt) = $opts{rpt_cnt};
    my ($offset) = $opts{offset};

    # default will always start from the beginning of the Seq array
    $offset = 0	unless $offset;

    while ($rpt_cnt) {
        $dfile->WriteSeq($pattern_size, $offset);
        $rpt_cnt--;
    }
}

=head2 dense_mk_dup_datafile (vol=>$vol, filename=>$filename, pattern_size=>$pattern_size, rpt_cnt=>$rpt_cnt)

Make a data file with duplicated data blocks. The unique pattern
size can't exceed 5 blocks (20480 bytes) for now.
The new file will be created in the 'TEST' directory in the volume.

=over 4

=item vol 

Volume object, required.

=item filename 

Name of the new file, required.

=item pattern_size 

Size of the unique data pattern, must be multiple of 4K. Required.

=item rpt_cnt 

Number of times the unique pattern repeats, required.

=item Returns:

File object

=back

=cut

sub dense_mk_dup_datafile (%)
{
    my (%opts) = validate (@_, { vol => { type=>HASHREF },
				 filename => { type=>SCALAR },
				 pattern_size => { type=>SCALAR },
				 rpt_cnt => { type=>SCALAR },
				 client => { type=>SCALAR | UNDEF, optional=>1, default=>'localhost' } } 
			  );

    my ($vol) = $opts{vol};
    my ($name) = $opts{filename};
    my ($pattern_size) = $opts{pattern_size};
    my ($rpt_cnt) = $opts{rpt_cnt};

    # $vol->make_testdir();
    Dense::Utils::create_testdir(vol=>$vol, client=>$opts{client});

    my $dfile = fioOpen($vol, $name, ">");
    write_dup_blks(dfile=>$dfile, pattern_size=>$pattern_size, rpt_cnt=>$rpt_cnt);
    $dfile->{fh}->sync();
    return ($dfile);
}

=head2 dense_mk_small_datafile (vol=>$vol, filename=>$filename, filesize=>$filesize, offset=>$offset)

Make a small datafile.
The new file will be created in the 'TEST' directory in the volume.

=over 4

=item vol 

Volume object, required.

=item filename 

Name of the new file, required.

=item filesize 

Size of the data file in bytes, required.

=item offset 

Starting offset into the Seq array, to provide some randomness. 
Value range between 0-79, default is 0. Optional.

=item Returns:

File object

=back

=cut

sub dense_mk_small_datafile (%)
{
    my (%opts) = validate (@_, { vol => { type=>HASHREF },
                                 filename => { type=>SCALAR },
                                 filesize => { type=>SCALAR },
                                 offset => { type=>SCALAR, optional=>1, default=>"0" },
				 client => { type=>SCALAR | UNDEF, optional=>1, default=>'localhost' } }
                          );
    
    my ($vol) = $opts{vol}; 
    my ($name) = $opts{filename}; 
    my ($filesize) = $opts{filesize}; 
    my ($offset) = $opts{offset};

    # default will always start from the beginning of the Seq array
    $offset = 0	unless $offset;

    # $vol->make_testdir();
    Dense::Utils::create_testdir(vol=>$vol, client=>$opts{client});

    my $dfile = fioOpen($vol, $name, ">");
    $dfile->WriteSeq($filesize, $offset);
    return ($dfile);
}


=head2 write_unique_blks(dfile=>$dfile, blk_cnt=>$blk_cnt, offset=>$offset)

Write unique blocks to the specified file.
The max number of unique data blocks is 80 * 4096 = 327680 blocks (~1.28 GB).

=over 4

=item dfile

File handle of the output file, required.

=item blk_cnt

Number of unique data blocks in the file, required.

=item offset

Offset into the seq array. Used as the seed to generate different
data blocks. Optional.

=back

=cut

sub write_unique_blks(%)
{
    my (%opts) = validate (@_, { dfile => { type=>HASHREF },
                                 blk_cnt => { type=>SCALAR },
                                 offset => { type=>SCALAR, optional=>1 } }
                          );
   
    my ($dfile) = $opts{dfile};
    my ($blk_cnt) = $opts{blk_cnt};
    my ($offset) = $opts{offset};

    my $seq_len = 80;	# only 80 different chars in seq
    my $lead_rep = -1;
    my $len = 0;
    my $blk_num = 0;

    $offset = 0	unless $offset;

    while ($blk_num < $blk_cnt) {

	$offset = $blk_num % $seq_len;
	if ($offset == 0) {
	    $lead_rep++;
	    if ($lead_rep > $BLKSIZE) {
		$lead_rep = 0;
	    }
	}

        $dfile->WriteBlk($offset, $lead_rep);
        $blk_num++;
    }
}


=head2 dense_mk_nondup_datafile(vol=>$vol, filename=>$filename, blk_cnt=>$blk_cnt, client=>$client)

Make a data file with non-duplicated blocks. The number of unique
data blocks can't exceed 327680 blocks (~1.28 GB).
The new file will be created in the 'TEST' directory in the volume.

=over 4

=item vol 

Volume object, required.

=item filename 

Name of the new file, required.

=item blk_cnt 

Number of unique data blocks in the file, required.

=item Returns:

File object

=back

=cut

sub dense_mk_nondup_datafile(%)
{
    my (%opts) = validate (@_, { vol => { type=>HASHREF },
				 filename => { type=>SCALAR },
				 blk_cnt => { type=>SCALAR },
				 client => { type=>SCALAR | UNDEF, optional=>1, default=>"localhost" } }
			  );

    my ($vol) = $opts{vol};
    my ($filename) = $opts{filename};
    my ($blk_cnt) = $opts{blk_cnt};

    # $vol->make_testdir();
    Dense::Utils::create_testdir(vol=>$vol, client=>$opts{client});

    my $dfile = fioOpen($vol, $filename, ">");
    write_unique_blks(dfile=>$dfile, blk_cnt=>$blk_cnt);

    return ($dfile);
}

=head2 dense_mk_large_mixed_datafile(vol=>$vol, filename=>$filename, blk_cnt=>$blk_cnt, client=>$client)

Make a large data file with mixed data blocks and holes. 
There are a lot of different duplicated blocks, data blocks with
high share counts, and variable sized holes.
The new file will be created in the 'TEST' directory in the volume.

=over 4

=item vol 

Volume object, required.

=item filename 

Name of the new file, required.

=item blk_cnt 

Number of data blocks in the file, required.

=back

=cut

sub dense_mk_large_mixed_datafile(%)
{
    my (%opts) = validate (@_, { vol => { type=>HASHREF },
				 filename => { type=>SCALAR },
				 blk_cnt => { type=>SCALAR },
				 client => { type=>SCALAR | UNDEF, optional=>1, default=>'localhost' } } 
			  );
    my ($vol) = $opts{vol};
    my ($filename) = $opts{filename};
    my ($blk_cnt) = $opts{blk_cnt};

    logStep("== Creating large data file $filename with $blk_cnt blocks");

    my $uniq_max = 327680;
    my $refcnt_max = 65536;
    my $subs = 4;
    if (!-e $vol->{testDir}) {
	# $vol->make_testdir();
	Dense::Utils::create_testdir(vol=>$vol, client=>$opts{client});
    }
    my $dfile = fioOpen($vol, $filename, ">");
    my $sub_size = int $blk_cnt / $subs;

    # Allocate holes to be inserted in files. Vary in size and location.
    my $hole_alloc = 100;
    my $hole_size = int $sub_size / $hole_alloc;
    if ($hole_size < 1) {
	$hole_size = 1;
    }
    $blk_cnt -= $hole_size * $hole_alloc;
    my $i = 1;
    my $hsize = 0;

    # Adds lots of different shared blocks within the file.
    my $sub_total = $sub_size;
    my $unique_blks = int $sub_size / 2;
    if ($unique_blks > $uniq_max) {
	$unique_blks = $uniq_max;
    }

    while ($sub_total >= $unique_blks) {
	write_unique_blks(dfile=>$dfile, blk_cnt=>$unique_blks);
	$sub_total -= $unique_blks;
	logStep("Created $unique_blks unique blocks");

	# insert a hole here
	$hsize = $hole_size * $i;
	if ($hole_alloc > $i) {
	    $dfile->Seek($hsize * 4096, 1);
	    $hole_alloc -= $i;
	    $i++;
	    logStep("Created $hsize empty blocks");
	}
    }
    if ($sub_total > 0) {
	write_unique_blks(dfile=>$dfile, blk_cnt=>$sub_total);
	logStep("Created $sub_total unique blocks");
    }
    $blk_cnt -= $sub_size;

    # if the total size is too small, just do the simple creation and exit
    if ($sub_size <= int ($refcnt_max * 1.5) ) {
	my $size = int ($sub_size / 2);
	write_dup_blks(dfile=>$dfile, pattern_size=>4096, rpt_cnt=>$size, offset=>39);
	logStep("Created $size duplicated blocks");
	$blk_cnt -= $size;
	write_unique_blks(dfile=>$dfile, blk_cnt=>$size);
	logStep("Created $size unique blocks");
	$blk_cnt -= $size;
	$size = $blk_cnt + $hole_size * $hole_alloc;
	write_dup_blks(dfile=>$dfile, pattern_size=>4096, rpt_cnt=>$size);
	logStep("Created $size duplicated blocks");
        logStep("== File $filename creation completed");

	$dfile->Close();
	return;
    }

    # high share count blocks, within max refcnt
    my $shared_blks = int (($sub_size * 2) / ($refcnt_max * 1.5));
    my $incr = int $refcnt_max / ( 2 * $shared_blks);
    my $save_shared_blks = $shared_blks;
    while ($shared_blks) {
	my $size = (int $refcnt_max / 2) + ($incr * $shared_blks);
	write_dup_blks(dfile=>$dfile, pattern_size=>4096, rpt_cnt=>$size);
	$blk_cnt -= $size;
	$shared_blks--;
        logStep("Created $size duplicated blocks");
    }

    # insert a hole here
    if ($hole_alloc > $i) {
	$hsize = $hole_size * $i;
	$dfile->Seek($hsize * 4096, 1);
	$hole_alloc -= $i;
	$i++;
	logStep("Created $hsize empty blocks");
    }

    # high share count blocks, exceeds max refcnt
    if ($sub_size > $refcnt_max) {
	write_dup_blks(dfile=>$dfile, pattern_size=>4096, rpt_cnt=>$sub_size);
	$blk_cnt -= $sub_size;
        logStep("Created $sub_size duplicated blocks");
    }

    # insert a hole here
    if ($hole_alloc > $i) {
	$hsize = $hole_size * $i;
	$dfile->Seek($hsize * 4096, 1);
	$hole_alloc -= $i;
	$i++;
	logStep("Created $hsize empty blocks");
    }

    # random blocks for the rest of the file
    my $seed = 1;
    while ($blk_cnt > 998) {
	write_unique_blks(dfile=>$dfile, blk_cnt=>499, offset=>$seed);
	write_dup_blks(dfile=>$dfile, pattern_size=>4096, rpt_cnt=>499, offset=>$seed);
	$blk_cnt -= 998;
	$seed++;
	$seed %= 80;
	logStep("Created 998 data blocks");
    }

    # insert last hole here
    $hsize = $hole_size * $hole_alloc;
    $dfile->Seek($hsize * 4096, 1);
    logStep("Created holes with $hsize empty blocks");

    if ($blk_cnt > 0) {
	write_unique_blks(dfile=>$dfile, blk_cnt=>$blk_cnt);
	logStep("Created $blk_cnt unique blocks");
    }

    logStep("== File $filename creation completed");

    $dfile->Close();
}

=head2 write_to_file(file=>$filename, start=>$startbyte, end=>$endbyte, pattern=>$pattern) 

Parameters:

=over 4

=item file

Name of the file, Required - mounted path

=item offset

offset into the file

=item len

Length of pattern

=item pattern

Pattern

=back

=cut

sub write_to_file {

    my (%opts) = validate (@_, { file => { type=>SCALAR },
                                 offset => { type=>SCALAR }, 
                                 len => { type=>SCALAR }, 
                                 pattern => { type=>SCALAR, optional=>1 } }
                          );


    my $filename = $opts{file};
    my $offset = $opts{offset};
    my $length = $opts{len};
    my $pattern = $opts{pattern};
    
    $pattern = "0" if(!defined $pattern);
    
    if(!(-e $filename)) {
	do_shell("sudo touch $filename");
    	logcomment("Setting permissions on the file");
	do_shell("sudo chmod 0777 $filename");
    }

    ## logcomment("Opening file $filename");
    open(FH, "+<$filename") || die("Could not open file $filename: $!");
    
    seek(FH,$offset,0); ## Offset from the beginning of the file
    my $i = 0;
    while($i < $length)
    {
            print FH "$pattern";
            $i++;
    }
    close(FH);

}


=head2 populate_vol(filerhost=>$filerhost, vol_path=>$vol_path,client=>$client, dir_path=>$dir_path, newdata_rate=>$newdata_rate, dir_avg=>$dir_avg, depth_avg=>$depth_avg, file_avg=>$file_avg, breadth_avg=>$breadth_avg, scramble=>$scramble, options=>{overwrite_dir=>1})

This sub invokes populate script with all of the above listed paramters.
a) file_avg, breadth_avg and newdata_rate all cannot be given together.
b) if newdata_rate is given and any one of file_avg and breadth_avg, this function caluculates the other and fills data in volume with reasonable accuracy. If only newdata_rate is given, function assumes file_avg=5 and calulates breadth_avg and populates the data equal to newdata_rate.If newrate is also not supplied 10% is assumed by default.
c) if file_avg and breadth_avg are both defined, volume is filled with that many blocks. Here newdata_rate doesnot come into picture.

Parameters:

=over 4

=item filerhost

Name of filer, required.

=item vol_path

Path of vol, required.

=item client

Name of client,optional

=item dir_path

Dir where data is to be populated, required

=item newdata_rate

Percentage of new data to be added,optional

=item dir_avg

No of subdirs in a dir,optional

=item depth_avg

Depth of root dir,optional

=item file_avg

No of files in a dir,optional

=item breadth_avg

Size of file,optional

=item scramble

Amt of duplicate data (0 being max and 100 being min),optional

=item options->{overwrite_dir}

if overwrite_dir is given, dir will be overwritten. If not function will only proceed if dir is empty. If dir is not there at all function will create it,optional

=back

=cut

sub populate_vol
{
    my (%opts) = validate (@_, { filerhost => { type=>SCALAR },
                                 vol_path => { type=>SCALAR },
                                 client => { type=>SCALAR, optional=>1 , default=>"localhost" },
                                 dir_path => { type=>SCALAR },
                                 newdata_rate => { type=>SCALAR, optional=>1 },
                                 dir_avg => { type=>SCALAR, optional=>1 , default=>2},
                                 depth_avg => { type=>SCALAR, optional=>1 , default=>2},
                                 file_avg => { type=>SCALAR, optional=>1 },
                                 breadth_avg => { type=>SCALAR, optional=>1 },
                                 scramble => { type=>SCALAR, optional=>1, default=>0 },
                                 timeout => { type=>SCALAR, optional=>1 , default=>18000 },
                                 options=> { type=>HASHREF, optional=>1 } }
                          );

    my ($host)=host($opts{filerhost});
    my ($vol_path)=$opts{vol_path};
    local ($dir_path)=$opts{dir_path};
    local ($client)=$opts{client};
    my ($newdata_rate)=$opts{newdata_rate};
    my ($dir_avg)=$opts{dir_avg};
    my ($depth_avg)=$opts{depth_avg};
    my ($file_avg)=$opts{file_avg};
    my ($breadth_avg)=$opts{breadth_avg};
    my ($scramble)=$opts{scramble};
    my ($timeout)=$opts{timeout};
    my ($options)=$opts{options};
    my ($file_avg_flag,$filesize_n_number_flag);
    if(defined($file_avg) && defined($breadth_avg) && defined($newdata_rate))
    {
        croak("file_avg, breadth_avg and newdata_rate cannot be defined together\n");
    }
    elsif(defined($file_avg) && defined($breadth_avg))
    {
        $filesize_n_number_flag=1;
    }
    elsif (!defined($file_avg)&& !defined($breadth_avg))
    {
        $file_avg=5;  #### Default value
        $file_avg_flag=1;
        $newdata_rate=defined($newdata_rate)?$opts{newdata_rate}:10; #### Default value
        logcomment("newdata_rate=$newdata_rate \n");
    }
    elsif (defined($file_avg))
    {
        $file_avg_flag=1;
        $newdata_rate=defined($newdata_rate)?$opts{newdata_rate}:10; #### Default value
        logcomment("newdata_rate=$newdata_rate \n");
    }
    elsif (defined($breadth_avg))
    {
        $file_avg_flag=0;
        $newdata_rate=defined($newdata_rate)?$opts{newdata_rate}:10; #### Default value
        logcomment("newdata_rate=$newdata_rate \n");
    }




#### Check if dir exists, if not create it
   if(!(-e $dir_path))
   {
        logcomment("Dir/File $dir_path does not exist. Creating it. \n");
        makedir() or croak(__PACKAGE__ . " :: $dir_path could not be created : $@ \n") ;
   }
#### Check if dir is a dir and not any other file type
   elsif(-d $dir_path)
   {
        logcomment("Dir $dir_path exists\n");
#### Do rm -rf on dir and re-create it
         if(defined($options->{overwrite_dir}))
        {
            logcomment("overwrite_dir flag on. Deleting and recreating $dir_path \n");
            my $ret=deltree($dir_path);
            $ret=((!($ret)) >> 8);
            if($ret)
            {
                croak(__PACKAGE__ . " :: deletion of $dir_path not successfull, exiting with $ret \n");
            }
        makedir() or croak(__PACKAGE__ . " :: $dir_path not created :$@\n") ;
         }
#### Flag not there, check if dir is empty. If so, proceed else die
        else
        {
             logcomment("overwrite_dir flag off. Checking if empty. \n");
             ## Check if dir is empty
             opendir DIR,$dir_path;
             if(scalar(grep( !/^\.\.?$/, readdir(DIR))))
             {
                closedir DIR;
                croak(__PACKAGE__ . " :: dir $dir_path not empty \n");
             }
             closedir DIR;
             logcomment("dir $dir_path empty\n");
        }
   }

#### Dir is not a dir but some file
   else
   {
        croak(__PACKAGE__ . " :: dir $dir_path is not a dir \n");
   }


#### Get the USED and Free volume space
    host($opts{filerhost});
    my ($filer_conn) = connect('#rsh');
    my ($used_space,$free_space)=diskFree($filer_conn,"/vol/$vol_path") or croak(__PACKAGE__ . " :: used and free space not returned \n");
        logcomment("used space=$used_space KB, free=$free_space KB\n");

## Invoke if both file size and number of files are given
   if($filesize_n_number_flag)
   {
        logcomment("Both file_avg and breadth_avg are defined \n");
        logcomment("DIR_AVG and DEPTH_AVG will both be 0 \n");
# Check Data to populate is less than free space
    my $data_to_populate=($file_avg*$breadth_avg)/1024;
    logcomment("data_to_populate=$data_to_populate KB\n");
    if ($data_to_populate > $free_space)
    {
        croak(__PACKAGE__ . " :: Not enough free space(=$free_space) on $vol_path\n");
    }

# Run populate
    return run_populate(0,0,$file_avg,$breadth_avg,$scramble,$timeout);

   }


#### Get the TOTAL volume size
    my ($vol_size)=(vol_size($filer_conn,$vol_path) or croak(__PACKAGE__ . " :: Size not returned \n"));
    logcomment(" volume size $vol_path = $vol_size KB \n");


############ Amt of data to be populated
    my $data_to_populate=($newdata_rate/100)*$vol_size;
    logcomment("data_to_populate=$data_to_populate KB\n");
    if ($data_to_populate > $free_space)
    {
        croak(__PACKAGE__ . " :: Not enough free space(=$free_space) on $vol_path\n");
    }


############# Calulate number of directories ###########
   my $no_of_dirs=no_of_dirs($dir_avg,$depth_avg);
   logcomment("no_of_dirs is $no_of_dirs \n");


### File size
if ($file_avg_flag)
{
 $breadth_avg=($data_to_populate/(($no_of_dirs)*$file_avg + $file_avg)*1024);
}
### No of Files
else
{
 $file_avg=($data_to_populate/($breadth_avg*(1+$no_of_dirs))*1024);
 if($file_avg < .6)
 {
        croak(__PACKAGE__ . " :: File size is coming out to be =$file_avg which is less than .6, Please either increase the newdata_rate or decrease the file size\n");
 }
}
## Dump all the values
logcomment("host=$host, dir_path=$dir_path, newdata_rate=$newdata_rate, dir_avg=$dir_avg, depth_avg=$depth_avg, file_avg=$file_avg, breadth_avg=$breadth_avg, file_avg_flag=$file_avg_flag \n");


######## Run the populate command
    return run_populate($dir_avg,$depth_avg,$file_avg,$breadth_avg,$scramble,$timeout);
}
sub run_populate
{
   my ($dir_avg,$depth_avg,$file_avg,$breadth_avg,$scramble,$timeout)=@_;
   my $sudo_path = param('SUDO_PATH','-mesg','(NATE is expected to have set this)');
   host($client);
   my $client_conn = connect('#rsh');

## Dump the populate command to run
    logcomment("dir_path=$dir_path DIR_AVG=$dir_avg DEPTH_AVG=$depth_avg FILE_AVG=$file_avg BREADTH_AVG=$breadth_avg SCRAMBLE=$scramble");
    my $ret_msg;
	if($scramble) {
		(undef,undef,$ret_msg) = say($client_conn, "$sudo_path /usr/software/test/noarch/bin/populate $dir_path DIR_AVG=$dir_avg DEPTH_AVG=$depth_avg FILE_AVG=$file_avg BREADTH_AVG=$breadth_avg BREADTH_DEV=0 FILE_DEV=0 DIR_DEV=0 DEPTH_DEV=0 SCRAMBLE=$scramble",-timeout,$timeout);
	} else {
		(undef,undef,$ret_msg) = say($client_conn, "$sudo_path /usr/software/test/noarch/bin/populate $dir_path DIR_AVG=$dir_avg DEPTH_AVG=$depth_avg FILE_AVG=$file_avg BREADTH_AVG=$breadth_avg BREADTH_DEV=0 FILE_DEV=0 DIR_DEV=0 DEPTH_DEV=0",-timeout,$timeout);
	}

if($ret_msg =~ /Population of $dir_path succeeded/)
{
        logcomment("Population of $dir_path succeeded \n");
        return 1;
}
else
{
        logcomment("Population of $dir_path failed: Error=$ret_msg \n");
        return 0;
}
}

=head2 modify_vol(CLIENT=>$client, DIR_PATH=>$dir_path)

Walks the specified tree and modifies its contents. The nature and extent of the modifications is determined by the options specified.
Croaks on any error.  Returns the random seed used on success.  In a list context, also returns a reference to a list of subpaths, and a
reference to a list of subseeds.  (See the FORK_COUNT option.)

Parameters:

=over 4

=item CLIENT 

Required: Name of client,

=item DIR_PATH

Required: Dir where data is to be modified 

=item SEED

Optional: Specifies the seed with which to initialize the random number generator

=item ENVIRONMENT_TYPE

Optional: Specifies a group of default option values.

=item CREATE_FILE_SIZE_AVG

Optional: Mean size of newlycreated files. 

=item CREATE_FILE_SIZE_DEV

Optional: Standard deviation from CREATE_FILE_SIZE_AVG.

=item CREATE_FILE_AVG

Optional: Mean number of files to create in each directory.

=item CREATE_FILE_DEV

Optional: Standard deviation from CREATE_FILE_AVG.

=item CREATE_DIR_AVG

Optional: Mean number of subdirectories to create in each directory.

=item CREATE_DIR_DEV

Optional: Standard deviation from CREATE_FILE_AVG.

=item SUBTREE_DEPTH_AVG

Optional: Mean tree depth to maintain when creating new subdirectory trees.

=item SUBTREE_DEPTH_DEV

Optional: Standard deviation from SUBTREE_DEPTH_AVG.

=item DEPTH_BASE

Optional: When modifying subtrees (see FORK_COUNT), this specifies the base depth of the subtree being modified.

=item TOTAL_SUBTREE_DEL

Optional: Average percentage of subdirectories to delete.

=item TOTAL_DEL

Optional: Average percentage of files to delete.

=item TOTAL_MOD

Optional: Average percentage of files to modify.

=item MOD_AVG

Optional: Mean percentage by which to modify files which are modified.

=item MOD_DEV

Optional: Standard deviation from MOD_AVG.

=item MOD_RAW

Optional: If set, treat MOD_AVG as a raw byte count instead of a percentage.

=item TOTAL_GROW

Optional: Average percentage of files to grow.

=item GROW_AVG

Optional: Mean percentage by which to grow files which are grown.

=item GROW_DEV

Optional: Mean percentage by which to grow files which are grown.

=item GROW_RAW

Optional: If set, treat GROW_AVG as a raw byte count instead of a percentage.

=item TOTAL_SHRINK

Optional: Average percentage of files to shrink.

=item SHRINK_AVG

Optional: Mean percentage by which to shrink files which are shrunk.

=item SHRINK_DEV

Optional: Standard deviation from SHRINK_AVG.

=item SHRINK_RAW

Optional: If set, treat SHRINK_AVG as a raw byte count instead of a percentage

=item FRAGMENT_AVG

Optional: Mean number of fragments to write to files which are modified.

=item FRAGMENT_DEV

Optional: Standard deviation from FRAGMENT_AVG.

=item FRAGMENT_RAW

Optional: If set, tread FRAGMENT_AVG as raw byte count instead of a number of fragments

=item HOLD_FS_SIZE

Optional: If set, automatically set unspecified options to try to maintain the overall size of the 
tree being modified.  That is, try to create and delete the same number of files, try to grow and 
shrink files by the same proportion/amount, etc.

=item FLOOD

Optional: If set, randomlygenerated data blocks are just a randomlychosen byte repeated as necessary,
rather than a complete randomlygener ated block.  This makes the "random" data pretty uninteresting,
but it is faster to generate.

=item SCRAMBLE

Optional: Percentage of successive blocks to scramble by XORing bytes in each BLOCK_SIZE chunk of bytes written.
The higher this is, the more probable that each BLOCK_SIZE sequence of bytes (probably) looks different, 
and therefore this basically controls how compressible the data is.

=item FORK_COUNT

Optional: Specifies a number of tree subpaths to return. 

=item LOG_FILE

Optional: Specifies a file to which to log all actions performed. 

=item MIN_DIR_SIZE

Optional: When modifying directories, the size of the directory is needed at the onset in order to do some of 
the math for entry name generation.

=cut

sub modify_vol {

    my (%opts) = validate (@_, { 
                                 CLIENT => { type=>SCALAR, optional=>1 , default=>"localhost" },
                                 DIR_PATH => { type=>SCALAR },
                                 SEED => { type=>SCALAR, optional=>1 },
                                 ENVIRONMENT_TYPE => { type=>SCALAR, optional=>1 },
                                 CREATE_FILE_SIZE_AVG => { type=>SCALAR, optional=>1 },
                                 CREATE_FILE_AVG => { type=>SCALAR, optional=>1 },
                                 CREATE_FILE_DEV => { type=>SCALAR, optional=>1 },
                                 CREATE_DIR_AVG => { type=>SCALAR, optional=>1 },
                                 CREATE_DIR_DEV => { type=>SCALAR, optional=>1 },
                                 SUBTREE_DEPTH_AVG => { type=>SCALAR, optional=>1 },
                                 SUBTREE_DEPTH_DEV => { type=>SCALAR, optional=>1 },
                                 DEPTH_BASE => { type=>SCALAR, optional=>1 },
                                 TOTAL_SUBTREE_DEL => { type=>SCALAR, optional=>1 },
                                 TOTAL_DEL => { type=>SCALAR, optional=>1 },
                                 TOTAL_MOD => { type=>SCALAR, optional=>1 },
                                 MOD_AVG => { type=>SCALAR, optional=>1 },
                                 MOD_DEV => { type=>SCALAR, optional=>1 },
                                 MOD_RAW => { type=>SCALAR, optional=>1 },
				 TOTAL_GROW => { type=>SCALAR, optional=>1 },
	                         GROW_AVG => { type=>SCALAR, optional=>1 },
                                 GROW_DEV => { type=>SCALAR, optional=>1 },
                                 GROW_RAW => { type=>SCALAR, optional=>1 },
                                 TOTAL_SHRINK => { type=>SCALAR, optional=>1 },
                                 SHRINK_AVG => { type=>SCALAR, optional=>1 },
                                 SHRINK_DEV => { type=>SCALAR, optional=>1 },
                                 SHRINK_RAW => { type=>SCALAR, optional=>1 },
                                 FRAGMENT_AVG => { type=>SCALAR, optional=>1 },
                                 FRAGMENT_DEV => { type=>SCALAR, optional=>1 },
                                 FRAGMENT_RAW => { type=>SCALAR, optional=>1 },
                                 HOLD_FS_SIZE => { type=>SCALAR, optional=>1 },
                                 FLOOD => { type=>SCALAR, optional=>1 },
                                 SCRAMBLE => { type=>SCALAR, optional=>1 },
                                 FORK_COUNT => { type=>SCALAR, optional=>1 },
                                 LOG_FILE => { type=>SCALAR, optional=>1 },
			         MIN_DIR_SIZE => { type=>SCALAR, optional=>1 },
                                 TIMEOUT => { type=>SCALAR, optional=>1 , default=>18000 } }
                          );

    my $dir_path = $opts{DIR_PATH};
    my $client = $opts{CLIENT};
    my $seed = $opts{SEED};
    my $environment_type = $opts{ENVIRONMENT_TYPE};
    my $create_file_size_avg = $opts{CREATE_FILE_SIZE_AVG};
    my $create_file_avg = $opts{CREATE_FILE_AVG};
    my $create_file_dev = $opts{CREATE_FILE_DEV};
    my $create_dir_avg = $opts{CREATE_DIR_AVG};
    my $create_dir_dev = $opts{CREATE_DIR_DEV};
    my $subtree_depth_avg = $opts{SUBTREE_DEPTH_AVG};
    my $subtree_depth_dev = $opts{SUBTREE_DEPTH_DEV};
    my $depth_base = $opts{DEPTH_BASE};
    my $total_subtree_del = $opts{TOTAL_SUBTREE_DEL};
    my $total_del = $opts{TOTAL_DEL};
    my $total_mod = $opts{TOTAL_MOD};
    my $mod_avg = $opts{MOD_AVG};
    my $mod_dev = $opts{MOD_DEV};
    my $mod_raw = $opts{MOD_RAW};
    my $total_grow = $opts{TOTAL_GROW};
    my $grow_avg = $opts{GROW_AVG};
    my $grow_dev = $opts{GROW_DEV};
    my $grow_raw = $opts{GROW_RAW};
    my $total_shrink = $opts{TOTAL_SHRINK};
    my $shrink_avg = $opts{SHRINK_AVG};
    my $shrink_dev = $opts{SHRINK_DEV};
    my $shrink_raw = $opts{SHRINK_RAW};
    my $fragment_avg = $opts{FRAGMENT_AVG};
    my $fragment_dev = $opts{FRAGMENT_DEV};
    my $fragment_raw = $opts{FRAGMENT_RAW};
    my $hold_fs_size = $opts{HOLD_FS_SIZE};
    my $flood = $opts{FLOOD};
    my $scramble = $opts{SCRAMBLE};
    my $fork_count = $opts{FORK_COUNT};
    my $log_file = $opts{LOG_FILE};
    my $min_dir_size = $opts{MIN_DIR_SIZE};
    my $timeout = $opts{TIMEOUT};

    host($client);
    my $client_conn = connect('#rsh');

    my $cmd = "";
    $cmd = $cmd . " SEED=$seed" if ( defined($seed) );
    $cmd = $cmd . " ENVIRONMENT_TYPE=$environment_type" if ( defined($environment_type) );
    $cmd = $cmd . " CREATE_FILE_SIZE_AVG=$create_file_size_avg" if ( defined($create_file_size_avg) );
    $cmd = $cmd . " CREATE_FILE_AVG=$create_file_avg" if ( defined($create_file_avg) );
    $cmd = $cmd . " CREATE_FILE_DEV=$create_file_dev" if ( defined($create_file_dev) );
    $cmd = $cmd . " CREATE_DIR_AVG=$create_dir_avg" if ( defined($create_dir_avg) );
    $cmd = $cmd . " CREATE_DIR_DEV=$create_dir_dev" if ( defined($create_dir_dev) );
    $cmd = $cmd . " SUBTREE_DEPTH_AVG=$subtree_depth_avg" if ( defined($subtree_depth_avg) );
    $cmd = $cmd . " SUBTREE_DEPTH_DEV=$subtree_depth_dev" if ( defined($subtree_depth_dev) );
    $cmd = $cmd . " DEPTH_BASE=$depth_base" if ( defined($depth_base) );
    $cmd = $cmd . " TOTAL_SUBTREE_DEL=$total_subtree_del" if ( defined($total_subtree_del) );
    $cmd = $cmd . " TOTAL_DEL=$total_del" if ( defined($total_del) );   
    $cmd = $cmd . " TOTAL_MOD=$total_mod" if ( defined($total_mod) );
    $cmd = $cmd . " MOD_AVG=$mod_avg" if ( defined($mod_avg) );
    $cmd = $cmd . " MOD_DEV=$mod_dev" if ( defined($mod_dev) );
    $cmd = $cmd . " MOD_RAW=$mod_raw" if ( defined($mod_raw) );
    $cmd = $cmd . " TOTAL_GROW=$total_grow" if ( defined($total_grow) );
    $cmd = $cmd . " GROW_AVG=$grow_avg" if (defined($grow_avg) ); 
    $cmd = $cmd . " GROW_DEV=$grow_dev" if (defined($grow_dev) );
    $cmd = $cmd . " GROW_RAW=$grow_raw" if (defined($grow_raw) );
    $cmd = $cmd . " TOTAL_SHRINK=$total_shrink" if (defined($total_shrink) );
    $cmd = $cmd . " SHRINK_AVG=$shrink_avg" if (defined($shrink_avg) );
    $cmd = $cmd . " SHRINK_DEV=$shrink_dev" if (defined($shrink_dev) );
    $cmd = $cmd . " SHRINK_RAW=$shrink_raw" if (defined($shrink_raw) );
    $cmd = $cmd . " FRAGMENT_AVG=$fragment_avg" if (defined($fragment_avg) );
    $cmd = $cmd . " FRAGMENT_DEV=$fragment_dev" if (defined($fragment_dev) );
    $cmd = $cmd . " FRAGMENT_RAW=$fragment_raw" if (defined($fragment_raw) );
    $cmd = $cmd . " HOLD_FS_SIZE=$hold_fs_size" if (defined($hold_fs_size) );
    $cmd = $cmd . " FLOOD=$flood" if (defined($flood) );
    $cmd = $cmd . " SCRAMBLE=$scramble" if (defined($scramble) );
    $cmd = $cmd . " FORK_COUNT=$fork_count" if (defined($fork_count) );
    $cmd = $cmd . " LOG_FILE=$log_file" if (defined($log_file) );
    $cmd = $cmd . " MIN_DIR_SIZE=$min_dir_size" if (defined($min_dir_size) );
 
    my $sudo_path = param('SUDO_PATH','-mesg','(NATE is expected to have set this)');
    my (undef, undef, $ret_msg) = say($client_conn, "$sudo_path /usr/software/test/noarch/bin/modify $dir_path $cmd", -timeout, $timeout);

    if($ret_msg =~ /Modification of $dir_path succeeded/) {
        logcomment("Modification of $dir_path succeeded \n");
        return 1;
    } else {
        logcomment("Modification of $dir_path failed: Error=$ret_msg \n");
        return 0;
    }
}

##### Function to calulcate $dir_avg**$depth_avg + $dir_avg**($depth_avg-1) + $dir_avg**($depth_avg-2) ... + $dir_avg**1

sub no_of_dirs
{
my ($dir_avg,$depth_avg)=@_;
my $sum=$dir_avg**$depth_avg;
$depth_avg=$depth_avg-1;
if($depth_avg<=0)
{
return $sum;
}
else
{
return ($sum + no_of_dirs($dir_avg,$depth_avg));
}
}

### Function to create dir
sub makedir
{
    host($client);
    my $client_conn = connect('#rsh');
    my $sudo_path = param('SUDO_PATH','-mesg','(NATE is expected to have set this)');
    my (undef,undef,$ret_msg) = say($client_conn, "$sudo_path /bin/mkdir -p $dir_path");
    if ($ret_msg =~ /permission|denied|error|fail|cannot|usage/i)
    {
        $@=$ret_msg;
        return 0;
    }
    else
    {
        return 1;
    }
}

#----------------------------------------------------------------------
# The testme routine to test the exported functions of this library
#
# ntest PKGNAME=Dense::TestData SUBNAME=testme util/test_pkg.thpl
#
# The routine can also be invoked with many optional parameters, such as:
#
# ntest PKGNAME=Dense::TestData SUBNAME=testme FUNCTION_NAME=write_dup_blks util/test_pkg.thpl
#
#----------------------------------------------------------------------

sub testme() {

        package main;

        our ($FILER, $FUNCTION_NAME, $FILER_VOL, $FILER_CONN, $FILER_AGGR);

        param('FILER',       '-mesg',    "Filer to test");
        param('FILER_CONN','-default',"#rsh");
        param('FUNCTION_NAME',   '-default',    "all");
        param('FILER_VOL',   '-default', 'vol0');
        param('FILER_AGGR',   '-default', 'test1');

        package Dense::TestData;

        logcomment("Testing on FILER $FILER, Function $FUNCTION_NAME");
        my $ret;
        my $path_name = "/vol/".$FILER_VOL;

        if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "write_dup_blks"){

                eval {
			my $filer_obj = dense_new_filer(host=>$FILER, nfsAcls=>1, noLicense=>0);
                        my $dvol = dense_setup_test_vvol(filer=>$filer_obj , volName=>$FILER_VOL, volSize=>"5g", purge=>1, aggr=>$FILER_AGGR, aggrPurge=>0);
			my $dfile = fioOpen($dvol, "maxref_data", ">");
			$ret = write_dup_blks(dfile=>$dfile, pattern_size=>"4096", rpt_cnt=>1, offset=>0);
		};
		
		if ($@ =~ /Connection timed out/) {
                        logresult('pass', "Expected fail. Check filer and connection type");
                } elsif ($@) {
                        logresult('fail', "$@");
                } else {
                        logresult('pass', "Got expected result");
                }
	}

	if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "dense_mk_dup_datafile"){

                eval {
                        my $filer_obj = dense_new_filer(host=>$FILER, nfsAcls=>1, noLicense=>0);
                        my $dvol = dense_setup_test_vvol(filer=>$filer_obj , volName=>$FILER_VOL, volSize=>"5g", purge=>1, aggr=>$FILER_AGGR, aggrPurge=>0);
                        $ret = dense_mk_dup_datafile (vol=>$dvol, filename=>"dense_mk_dup_datafile", pattern_size=>4096, rpt_cnt=>2);
                };

                if ($@ =~ /Connection timed out/) {
                        logresult('pass', "Expected fail. Check filer and connection type");
                } elsif ($@) {
                        logresult('fail', "$@");
                } else {
                        logresult('pass', "Got expected result");
                }
        }

	if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "dense_mk_small_datafile"){

                eval {
                        my $filer_obj = dense_new_filer(host=>$FILER, nfsAcls=>1, noLicense=>0);
                        my $dvol = dense_setup_test_vvol(filer=>$filer_obj , volName=>$FILER_VOL, volSize=>"5g", purge=>1, aggr=>$FILER_AGGR, aggrPurge=>0);
                        $ret = dense_mk_small_datafile (vol=>$dvol, filename=>"dense_mk_small_datafile", filesize=>8192, offset=>0);
                };

                if ($@ =~ /Connection timed out/) {
                        logresult('pass', "Expected fail. Check filer and connection type");
                } elsif ($@) {
                        logresult('fail', "$@");
                } else {
                        logresult('pass', "Got expected result");
                }
        }

	if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "write_unique_blks"){

                eval {
                        my $filer_obj = dense_new_filer(host=>$FILER, nfsAcls=>1, noLicense=>0);
                        my $dvol = dense_setup_test_vvol(filer=>$filer_obj , volName=>$FILER_VOL, volSize=>"5g", purge=>1, aggr=>$FILER_AGGR, aggrPurge=>0);
			my $dfile = fioOpen($dvol, "write_unique_blks", ">");
                        $ret = write_unique_blks(dfile=>$dfile, blk_cnt=>10, offset=>0);
		};

                if ($@ =~ /Connection timed out/) {
                        logresult('pass', "Expected fail. Check filer and connection type");
                } elsif ($@) {
                        logresult('fail', "$@");
                } else {
                        logresult('pass', "Got expected result");
                }
        }

	if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "dense_mk_nondup_datafile"){

                eval {
                        my $filer_obj = dense_new_filer(host=>$FILER, nfsAcls=>1, noLicense=>0);
                        my $dvol = dense_setup_test_vvol(filer=>$filer_obj , volName=>$FILER_VOL, volSize=>"5g", purge=>1, aggr=>$FILER_AGGR, aggrPurge=>0);
                        $ret = dense_mk_nondup_datafile(vol=>$dvol, filename=>"dense_mk_nondup_datafile", blk_cnt=>20);
                };

                if ($@ =~ /Connection timed out/) {
                        logresult('pass', "Expected fail. Check filer and connection type");
                } elsif ($@) {
                        logresult('fail', "$@");
                } else {
                        logresult('pass', "Got expected result");
                }
        }

	if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "dense_mk_large_mixed_datafile"){

                eval {
                        my $filer_obj = dense_new_filer(host=>$FILER, nfsAcls=>1, noLicense=>0);
                        my $dvol = dense_setup_test_vvol(filer=>$filer_obj , volName=>$FILER_VOL, volSize=>"5g", purge=>1, aggr=>$FILER_AGGR, aggrPurge=>0);
                        $ret = dense_mk_large_mixed_datafile(vol=>$dvol, filename=>"dense_mk_large_mixed_datafile", blk_cnt=>512);
                };

                if ($@ =~ /Connection timed out/) {
                        logresult('pass', "Expected fail. Check filer and connection type");
                } elsif ($@) {
                        logresult('fail', "$@");
                } else {
                        logresult('pass', "Got expected result");
                }
        }

} ## End of Testme Routine

=head1 SEE ALSO

    Where additional information can be found.

=head1 BUGS

    Description of bugs or what the library does not do and why

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = volume, df, useradmin

=head1 AUTHOR

    Rani Palanisami (rani@netapp.com)
    $Id: //depot/prod/test/Rsierranevada/lib/Dense/TestData.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

1;
