#Copyright (c) 2007 Network Appliance, Inc.
#All rights reserved.
# $Id$

## @summary  Upgrade/Revert Library
##
## @description
##     Upgrade/Revert library,determines whether its upgrade/revert by comparing the current ontap version
##     and the desired ontap version and calls appropriate upgrade/revert routine
##     Includes extra functionality during revert which is not handled by the Rervert Library
##         - Deletes aggregate snapshots
##         - Deletes snaplock_enterprise license
##         - Disables Snapmirror/Snapvault
##	   - Run sis revert_to on all dense volumes
##	   - Online all offline/restricted volumes
##
## @usage
##    use ubs;
##    pass the Hash by reference with appropriate to_release and Filer to do_transform routine
##
## @keywords snaplock compliance clock
##
## @dependencies
##    lib/Filer/Transform.pm
##    lib/Filer/Upgrade.pm
##    lib/Filer/Revert.pm
##
## @author dl-dense-qa@netapp.com (DRG QA)
## @author avinashg@netapp.com 
##


package Dense::Upgrade_Revert;

use strict;
use base qw(Exporter);
use Tharn;
use Filer::Version qw(version_compare codename_to_version);
use Filer::Upgrade;
use Filer::Revert;
use Filer::Transform;
use boot;
use filer;
use FS::Util::Utilities qw(:utils);
use SPARSE::Util::Logger qw( logStep logError logWarn logDbg);
use NOTE;
use cluster;
use TestUtils qw(get_site_info);
use Dense qw(:action :util :verify);
use Dense::Utils;
use Volume qw (vol_list);

BEGIN
{
  our @EXPORT = qw(
	do_transform
	do_upgrade
	do_revert
	do_ndu
	get_filer_to_desired_release
	install_and_download
	enable_telnet
  );
}

##############################################################################
# Function Name : do_transform()
# Parameters : 	FILER			=>	Filer Name
#				from_release	=>  Current release of the filer  	
#				to_release 		=> 	To Upgrade/Downgrade to this release
#				
# Return Value : returns 1 on sucess or dies on failure 
#
##############################################################################

=head2 do_transform

This routine gets the current filer version and if the current version of the 
filer and the to_release version are same then returns 1.  
else calls the version_compare subroutine and based on the return value 
calls do_upgrade routine if return value is -1 
calls do_reverte routine if return value is 1 
dies if return value is 0. 

Deletes the aggregate snapshots before revert
And removes the Snapmirror and Snapvault License

returns 1 on sucess or dies on failure  

Supported for ONTAP 7.0 and above

=over

=item Example

(Required) : 
    my %hash = (FILER => $FILER,
                from_release => '7.2.4X8',
                to_release   => 'RironcityN_071031_2230'
               );
my $install = do_transform(\%$hash);

=back

=for html <hr>

=cut



sub do_transform ($)
{
    my ($input) = @_;
    my ($fh,$fc,$fc_console,$current_release,$to_release,$install);
    $to_release = $input->{"to_release"};
    $fh = host($input->{FILER});
    if ( $fh->hosttype eq 'filer-simulator' ) {
        logresult("NA","cannot run on simulator");
    }
    
    $fc_console = connect('#console', -login, -timeout, 600);
    $input->{fc} = $fc_console;
    $current_release = _get_filer_version($input->{fc});
    $input->{"from_release"} = $current_release;

    ##
    ## Compare the Current Version on Filer and to_release 
    ##
    if( $current_release ne $to_release ) {
        logStep("Filer Current ONTAP release : $current_release");
        logStep("Trying to Upgrade/Revert the Filer from : $current_release ====> $to_release");
	##
        ## Enable NFS incase it was turned off during failed previous runs
        ##
        _enable_nfs($input->{FILER});
        
	$install = version_compare($current_release,$to_release);
        if ($install == -1) { 
            do_upgrade($input); 
        } elsif( $install == 1) {
             do_revert($input); 
        } else { 
			die "Unsupported Upgrade/Revert : $to_release\n";
        }

        ## Compare the current version on filer and to_release
        ## die if the versions are different
        ##
        $current_release = _get_filer_version($input->{fc});
        if($current_release !~ /$to_release/ ) {
	    sleep(30);
	    $current_release = _get_filer_version($input->{fc});
	    if($current_release !~ /$to_release/ ) {
		die "Install Failed,  The filer should have been booted with : $to_release but the filer current ontap version is : $current_release";
	    }
        } else {
            logStep("$install Successful - Filer Current Ontap Version : $current_release ");
        } 
        
    } else {
        logStep("Filer is running the same version as to_release : $to_release");
    }
    return $install;
}


##############################################################################
# Function Name : do_upgrade()
# Parameters :  FILER           =>  Filer Name
#               from_release    =>  Current release of the filer
#               to_release      =>  To Upgrade/Downgrade to this release
#               fc              =>  Filer Connection Object
#
# Return Value : returns 1 on sucess or dies on failure 
#
##############################################################################

=head2 do_upgrade

This routine uses Filer/Upgrade.pm 
Creates Upgrade object 
downloads the kernel
reboot the filer

returns 1 on sucess or dies on failure  

Supported for ONTAP 7.0 and above

=over

=item Example : hash_reference

(Required) : 
    my %hash = (FILER => $FILER,
                from_release => '7.2.4X8',
                to_release   => 'RironcityN_071031_2230',
                fc           => $filer_connection_object
               );
    do_upgrade(\%$hash);

=back

=for html <hr>

=cut

sub do_upgrade (%)
{
    my ($data) = @_;

    my ($fc_console,$upgrade_version,$filer,$old_filer,$upgrade,$site);
    my $kernel = defined $data->{kernel} ? $data->{kernel} : "releases_debug" ;
    $fc_console = $data->{'fc'};
    $upgrade_version = $data->{"to_release"};
    $filer = $data->{'FILER'};
    tstlog('-comment',"Inside Upgrade");

    # Create Upgrade Object 
    $upgrade = Filer::Upgrade->new($fc_console,
                                      to_version => $upgrade_version,
                                      FILER => $filer,
                                     ) or die "cannot create upgrade object";
    $upgrade->do_upgrade( kernel => "$kernel" ) or die "upgrade failed";
    transit_to($fc_console, UP);
    logStep("Filer is rebooted after upgrade");
    return 1;
}



##############################################################################
# Function Name : do_revert()
# Parameters :  FILER           =>  Filer Name
#               from_release    =>  Current release of the filer
#               to_release      =>  To Upgrade/Downgrade to this release
#               fc              =>  Filer Connection Object
#
# Return Value : returns 1 on sucess or dies on failure
#
##############################################################################


=head2 do_revert
This routine uses NOTE & Filer/Revert.pm Libraries

Aggregate snapshots are deleted
snaplock enterprise license will be deleted 
snapmirror and snapvault will be disabled

Creates Revert object 
downloads the kernel
execute revert_to
reboot the filer
Validate whether REVERT is succesfull

returns 1 on sucess or dies on failure  

Supported for ONTAP 7.0 and above

=over

=item Example : hash_reference

(Required) : 
    my %hash = (FILER => $FILER,
                from_release => '7.2.4X8',
                to_release   => 'RironcityN_071031_2230',
                fc           => $filer_connection_object
               );
    do_revert(\%$hash);

=back

=for html <hr>

=cut

sub do_revert (%)
{
    my ($data) = @_;

    my ($fc_console,$revert_to,$revert_from,$filer,$revert,$conn);
 
    $fc_console = $data->{'fc'};
    $revert_from = $data->{"from_release"};
    $revert_to = $data->{"to_release"};
    $filer = $data->{'FILER'};
    host($filer);
    $conn = connect("#rsh");
    my $kernel = defined $data->{kernel} ? $data->{kernel} : "releases_debug" ;
    my $download_kernel = defined $data->{download_kernel} ? $data->{download_kernel} : 1;

    tstlog('-comment',"Inside REVERT, Reverting to : $revert_to");
    
    ##
    ## @step SETUP :
    ## Create a revert object;
    ##
    $revert = Filer::Revert->new($fc_console, 
				    to_version => $revert_to,
                                    FILER => $filer ,
                                   )  or die "Cannot create revert object";

    ## Check whether revert is allowed - Only two previous releases are allowed
    if (! check_revert_version($data->{"from_release"},$data->{"to_release"})) {
	die "Cannot Revert to $revert->to_version";
    }

    my $from_version_short = codename_to_version($revert_from);
    my $from_version = substr("$from_version_short",0,3);    
    my $to_version_short = codename_to_version($revert_to);
    my $version = substr("$to_version_short",0,3);    
    #revert_sis_volumes(host => $filer,revert_to=>$version);
    my @vols = vol_list($conn);
    foreach my $single (@vols){
	if ( ! filer_vol_is_online($conn , $single) ) {
	    filer_vol_online($conn , $single);
	}
	if ( $from_version =~ /8.*/) {
	    if ( is_dense_flag(host=>$filer, volname=>$single) ){
		eval {
		    dense_revert_to( host => $filer ,path_name => "/vol/$single" ,version => $version ,validate => 1, count => 25 );
		    dense_wait_for_status(host=>$filer, path_name=>"/vol/$single", status=>"IDLE");
		};
		if ($@) {
		    dense_off_undo(host=>$filer,path_name=>"/vol/$single");
		    dense_wait_for_status(host=>$filer, path_name=>"/vol/$single", status=>"NOSTATUS");
		}
	    }
	}
    }
    ##
    ## Delete the aggregate snapshots 
    ##
    _aggregate_snapshot_delete($data->{FILER});
    
    ## Check telnet connection
    enable_telnet($filer);
    
    ##
    ## Revert the filer to a supported older version and reboot the filer
    ##
    logDbg("download_kernel=>$download_kernel");
    if ($download_kernel) { ## If we dont want to download then we dont want to install also
	$revert->install_release( kernel => "$kernel" ) or die "Could not initialize revert object";
    }
    
    $revert->do(download_kernel => $download_kernel) or die "could not perform the revert";

    transit_to($fc_console, UP);
   # reboot_verify($fc_console, UP);
    logStep("Filer is rebooted after down grade");

    if ( $revert->validate()) {
        logStep("Revert successful");
    } else {
       die "Revert unsuccessful";
    }
	return 1;
}



##############################################################################
# Function Name : do_ndu()
# Parameters :  FILER           =>  Filer Name
# 		PARTNER		=>  Partner Filer name
#               from_release    =>  Current release of the filer
#               to_release      =>  To Upgrade/Downgrade to this release
#
# Return Value : returns 1 on sucess or dies on failure
#
##############################################################################

sub do_ndu(%)
{
	my (%opts) = @_;
	my $filer= $opts{filer};
	my $partner = $opts{partner};
	my $CF_WAIT_TIME = 300;
	my $type="";	
	my $install;
	if (! exists $opts{type} ) {
		$type="major";
	}
	if (! exists $opts{install} ) {
		$install=1;
	}
	# Install software and download on both the filers parellely	
	if ($install) {
	    my $filer_obj = subtest(sub{install_and_download(filer=>$filer,from_version=>$opts{from_version},to_version=>$opts{to_version})}, -bg);
	    my $partner_obj = subtest(sub{install_and_download(filer=>$partner,from_version=>$opts{from_version},to_version=>$opts{to_version})}, -bg);
	
	    # Wait for download to complete
	    sleep 3 while ($filer_obj->isAlive());
	    sleep 3 while ($partner_obj->isAlive());
	}
	my $host_filer = host($filer);
	my $filer_conn=connect("#rsh");

	my $host_partner = host($partner);
	my $partner_conn=connect("#rsh");
	
	#Enable Cluster 
	cf_enable($filer);
	sleep(10);

	# Takeover the partner filer
	cf_takeover($filer_conn,"cf takeover") or logresult("FAIL", "Could not takeover");

	# Wait for takeover to complete and the partner filer to go down
	wait_state($filer_conn,"TAKEOVER",$CF_WAIT_TIME) or logresult("FATAL", "Could not takeover filer $filer");
	wait_state($partner_conn,"DEAD",$CF_WAIT_TIME) or logresult("FATAL","Could not takeover filer $partner");
	
	# Initiate Giveback so the partner comes with the New version
	cf_giveback($filer_conn) or logresult("FAIL", "Could not giveback");
	transit_to($host_partner,"UP");

	my $ret_version = say($filer_conn,"version");	
	logcomment("Version is $ret_version");
	logcomment("type is $type");

	# Perform Major or Minor NDU depending on the type
	if ( $type eq "major" ) {
		# Initiate "cf takeover -n" from the partner which waits for the filer
		# to Halt
	    say($partner_conn,"cf takeover -n",-noprompt);
	    
		# Halt the filer and wait for boot up. transit_to(UP) will wait till
		# the filer boots with new version and waits for the partner to giveback
		halt($host_filer); 
	    transit_to($host_filer,"UP"); 
	
		# Initiate Giveback from the Partner and wait for the filer to come up
	    cf_giveback($partner_conn) or logresult("FAIL", "Could not giveback");
	    transit_to($host_filer,"UP");
	} else {
	    wait_state($filer_conn,"UP",$CF_WAIT_TIME) or logresult("FATAL", "Could not giveback filer $filer");
	    wait_state($partner_conn,"UP",$CF_WAIT_TIME) or logresult("FATAL","Could not giveback filer $partner");
	    cf_takeover($partner_conn) or logresult("FAIL", "Could not takeover (NDU Minor)");
	    transit_to($host_filer,"UP"); # exit's when wait for givaback
	    cf_giveback($partner_conn) or logresult("FAIL", "Could not giveback");
	    transit_to($host_filer,"UP");
	}
	my $filer_latest_version= _get_filer_version($filer_conn);
	my $partner_latest_version= _get_filer_version($partner_conn);
	logcomment("Filer versions are $filer_latest_version and $partner_latest_version");
	if (($filer_latest_version ne $opts{to_version}) && ($partner_latest_version ne $opts{to_version})) {
	    die "Filer did not boot with the required version";
	}
	return 1;
} 
			





##
## @ Enable nfs on the filer before doing any Upgrade/Revert 
## @ Argument : Filer Name 
##
sub _enable_nfs($)
{
    my ($filer) = @_;
	
    my $filer_obj = NOTE::Filer->connect(host=>$filer);
    $filer_obj->say("nfs on");
    return;
}



##
##
## @ Function Name : _aggregate_snapshot_delete
## @ Argument : FilerName
## @ Connect to the Filer
## @ Get the Aggregates list 
## @ For each aggregate, delete all the snapshots
## @ For each volume, delete all the snapshots
## @ Disable snapmirror/snapvault
##
sub _aggregate_snapshot_delete($)
{
    my ($filer_name) = @_;

    my ($filer,@aggr_objs,$cmd,$full_cmd,@vol_objs);
    	
    $filer = NOTE::Filer->connect(host=>$filer_name);
    eval {
        $filer->say("nfs on");
        $filer->say("options snapmirror.enable off");
        $filer->say("options snapvault.enable off");
        $filer->say("cf disable");
    };

    @aggr_objs = $filer->Aggregates;
    $cmd = 'snap delete -A -a -f '; 
    eval {	
        foreach my $aggr (@aggr_objs) {
            $aggr->online;
            $full_cmd = $cmd.$aggr->name;
            $filer->say($full_cmd);  
        }
    };
    @vol_objs = $filer->Volumes;
    $cmd = 'snap delete -a -f '; 
    eval {	
        foreach my $vol (@vol_objs) {
            $full_cmd = $cmd.$vol->name;
            $filer->say($full_cmd);  
        }
    };
    
    die "Failed while deleting the Aggregate Snapshots $@" if($@) ;

    ## Wait for the summary file cleanup scan completion
    host($filer_name);
    my $conn = connect("#rsh"); 
    set_wafl_scan_speed(conn => $conn, speed => 100000 );
    wait_for_scan_done(host => $filer_name, scan=>"Summary file cleanup");
}



##
## @ name :: get_filer_to_desired_release 
## @ Arguments : Hash Containing 'from_release'
## 
## @ Needs to be Called before starting the Upgrade/Revert test
## @ Calls do_transform to get the filer to desired 'from_release' 

sub get_filer_to_desired_release(%)
{
    my (%data) = @_;
    
    logcomment("Inside get_filer_to_desired_release");    
    my $desired_release = $data{'from_release'};   
    my $filer = $data{'FILER'}; 
    logStep("Filer $filer ONTAP version before performing Upgrade/Revert is : $desired_release ");
    my %inhash = ( 
		  'FILER' => $filer,
		  'to_release' => $desired_release,
                 );
    # Call do_transform with the the constructed hash
    do_transform(\%inhash);
    return 1;
}

##
## @ Get the Current Filer Version
## @ Argument : Filer Connection
##
sub _get_filer_version($)
{
    my ($fc) = @_;
    my $version = filer_version_get($fc);
    my $current_version_short = codename_to_version($version);
    if ( $version =~ m/(.*)<.*>/ ) {
        $version = $1;    # strip last <xx> charactesr
        return $version;
    }
}

##
## @ Convert the ONTAP release code name to version string
## @ Argument : ONTAP release code name
##
sub _filer_version_short($)
{
    my ($version1) = $_;
    my $version = $version1;
    my $version_short = codename_to_version($version);
    if ( $version =~ m/(.*)<.*>/ ) {
        $version = $1;    # strip last <xx> charactesr
        return $version;
    }
    #croak "Not able to resolve the Codename to Version String\n";
}

##
## To check the whether revert is possible with the versions specified.
## Returns 1 if possible else 0.
##


sub check_revert_version($$)
{
    my ($from_version,$to_version) = @_;

    my %version = (
	"7.0"	=> 1,
	"7.1"	=> 2,
	"7.2"	=> 3,
	"7.3"	=> 4,
	"8.0"	=> 5,
	"8.1"	=> 6,
	);
	
    my $from_version_short = codename_to_version($from_version);
    my $to_version_short = codename_to_version($to_version);
    
    my $mapping_from_version = substr("$from_version_short",0,3); ## Chop's off the ".90" from $from_version_short
    my $mapping_to_version = substr("$to_version_short",0,3);
    
    logcomment("Integer Version are from_version = $version{$mapping_from_version} to_version = $version{$mapping_to_version}");

    if (($version{$mapping_from_version} - $version{$mapping_to_version}) > 2) {
		return 0;
    } else {
		return 1;
    }
}

sub install_and_download(%) {
    my (%opts) = @_;
    logcomment("Inside install_and_download()");
    my $filer= $opts{filer};
    my $from_version = $opts{from_version};
    my $to_version = $opts{to_version};
    my $kernel = defined $opts{kernel} ? $opts{kernel} : "releases_debug" ;
    my $host_filer = Tharn::host($filer);
    my $filer_conn = Tharn::connect('#rsh-socket');

    my $Filer_obj = Filer::Upgrade->new(
           $filer_conn,
           to_version => $to_version,
           from_version => $from_version,
           FILER      => $filer,
           );

    $Filer_obj->install_release(kernel=>$kernel);
    $Filer_obj->download_kernel();
}

sub enable_telnet($) {
    my ($filer) = @_;
    my $telnet_conn;
    my $rsh_conn;
    host($filer);
    $rsh_conn = connect("#rsh");
    my $status = filer_options_get($rsh_conn,"telnet.enable");
    if ($status eq "off") {
	filer_options_set($rsh_conn,"telnet.enable","on");
    }
    return 1;
}
	


1;


