# $Id: //depot/prod/test/Rsierranevada/lib/Dense/Vmalign/Vmalign.pm#1 $
#
# Hey emacs, this is basically -*- perl -*-
# vim:filetype=perl
#
# Copyright (c) 2011 Network Appliance, Inc.
# All rights reserved.
#
## @summary	
##      Library for vmalign routines for both 7mode and cmode.
##
## @description
##	This module provides capabilities
##	required for vmalign work flow.
##
## @synopsis
##	use Dense::Vmalign::Vmalign;
##      my $vmalign_obj = Dense::Vmalign::Vmalign->new(test => $test,
##				                       scriptParams => $params);
##
## @author weili@netapp.com


package Dense::Vmalign::Vmalign;

use strict;
use warnings;

use TCD;
use Server;
use Tharn;
use Options;
use Subtest;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF BOOLEAN);

#use SPARSE libraries for client mount/read/write, etc.
use SPARSE::cmode::ClientObject qw(client);
use SPARSE::cmode::ClientOpsObject qw(clientOps);


#NACL libs
use NACL::APISet;
use NACL::C::Node;
use NACL::C::Volume;


# -------------------
# basic functions
# -------------------

## @name ::new
##
## @summary Dense::Vmalign::Vmalign constructor.
##
## @description
## Configure and setup pre-requisites for starting the test for vmalign:
## - Setup SPARSE::cmode::ClientObject
## - Setup NACL::APISet
##
## @synopsis
##	use Dense::Vmalign::Vmalign;
##      use TCD;
##      use Options;
##
##      my $test = TCD->new(-testcases => \@Tests);
##
##      my $options = new Options;
##      $options->parameters( definitions => [
##                      {name => "FILER", description => "Filer name"},
##                      {name => "CLIENT", description => "Client name", spec => "s@", default => "localhost" },
##          ]);
##    
##      $options->test_cases( definitions => {
##        "vmalign_testcase_1"    => {
##            TC_description  => "vmalign basic functional test 1",
##            TEST_TO_RUN     => "vmalign_testcase_1"},});
##    
##      $options->process();
##      my $params = $options->get_hash();
##
##      my $vmalign_obj = Dense::Vmalign::Vmalign->new(test => $test,
##				                       scriptParams => $params);
##
## @arg test requited: TCD object
## 
## @arg scriptParams requited: HASHREF of scripts parameters
##
## @arg filer optional: SCALAR, filer name for create object on
##
## @arg debug Optional: Controls verbosity of this module.
## Defaults to FALSE.
##
## @return Dense::Vmalign::Vmalign object
##

sub new() {
    my $class = shift;
    my %args = validate( @_,
        {
            test         => {type => HASHREF},
            scriptParams => {type => HASHREF},
            filer        => {type => SCALAR, optional => 1,},
            ui           => {type => SCALAR, optional => 1,},
            debug        => {type => BOOLEAN, optional => 1, default => 0},
        });

    my $self = {};
    my $test = $args{test};
    my $scriptParams = $args{scriptParams};
   
    # Client Object Setup 
    my $Clients = $scriptParams->{CLIENT};
    my $Client = @$Clients[0];
    $test->nlog("Clients object is  $Client ".Dumper($Client) ) if ($args{_debug});
    
  
    my $opts={};
    $opts->{test}=$test;
    $opts->{params}=$scriptParams;    
    $opts->{hdl}=host($Client);  
    $opts->{hostname}= $Client; 
    $opts->{os}=host($Client)->hosttype();
   
    my $clientObjHash = client(name=>"Client1", opts => $opts);
    
    my $Vserver = $test->get_param("VSERVER");
    my $Cdef_Path = $test->get_param("CDEF");   
    my $Filers = $scriptParams->{FILER};
    my $Filer = (defined ($args{filer})) ? ($args{filer}) : @$Filers[0];
    my $Host = host($Filer);
    my $Node = NACL::C::Node->find();
    my $Mode = $Node->mode();
    
    my $Ui;
    if (defined ($args{ui})) {
        $Ui = $args{ui};
    } else {
        $Ui= ( defined($test->get_param("FILER_APISET_SHOULD_INTERFACE"))) ? $test->get_param("FILER_APISET_SHOULD_INTERFACE") : 'CLI' ;
    }

    #Create an API object
    my $vserver_host = host("vserver_name;hosttype=filer");
    my $Api_Set_Obj;
    $Api_Set_Obj = NACL::APISet->new(hostobj   => $Host, 
                                     category  => "Node",
                                     interface => $Ui,
                                     set       => $Mode,
                                    );
       
    # Vmalign object member
    $self->{filer} = $Filer;
    $self->{mode} = $Mode;
    $self->{node} = $Node;
    $self->{ui} = $Ui;
    $self->{vserver} = $Vserver;
    $self->{filerip} = $test->get_param("FILER_IP"); 
    $self->{ApiSet} = $Api_Set_Obj;
    $self->{client} = $clientObjHash;
    $self->{test} = $test;
    $self->{_debug} = $args{debug};

    $self->{test}->nlog("class ui is $Ui");
    bless($self, $class);

    return $self;

}


## @name ::volume_create
##
## @summary Create a new volume
##
## @description
## Create a new volume
##
## @return NACL API volume create output.
##
## @examples
##      $vmalign_obj->volume_create(volume =>$vol_name,
##                                  aggregate => 'aggr1',
##                                  vserver => 'vs1',       ## only required for CMode
##                                  "junction-path"=>"/$vol_name",     ## only required for CMode
##                                  "vm-align-sector" => 1,
##                                  "vm-align-suffix" => 'vmdk',
##                                  size=>'500m');
##
## @arg volume required: volume name
##
## @arg aggregate required: aggregate name 
##
## @arg vserver optional: vserver name (for Cmode only, 7mode will ignore this arg)
##  default: $class->{vserver}
##
## @arg "junction-path" optional: junction path (for Cmode only, 7mode will ignore this arg)
##  default: $class->{vserver}
##
## @arg "vm-align-sector"  optional: vm-align-sector parameter for for creating vm align volume
##
## @arg "vm-align-suffix"  optional: vm-align-sector parameter for creating vm align volume
##
## @arg size optional: volume size
##  default: 20m
##
## @arg mount optional: volume will be mounted or not (for 7mode only)
##  default: yes
##

sub volume_create{
    my $class = shift;
    
    my (%args) = validate( @_,
       {
            volume                  => {type => SCALAR},
            aggregate               => {type => SCALAR},
            vserver                 => {type => SCALAR|UNDEF,, optional => 1, default => $class->{vserver}},
            "junction-path"         => {type => SCALAR, optional => 1},
            "vm-align-sector"       => {type => SCALAR, optional => 1,},
            "vm-align-suffix"       => {type => SCALAR, optional => 1,},
            "unix-permissions"      => {type => SCALAR, optional => 1,},
	    size                    => {type => SCALAR, optional => 1, default => '20m'},
            mount                   => {type => BOOLEAN, optional => 1, default => 1 },
            type                    => {type => SCALAR, optional => 1,},
       });
   
    my $ret;
    my $vserver = (defined($args{vserver})) ? $args{vserver} : $class->{vserver};  
   
    if ($class->{mode} eq 'CMode'){
       # my $junction_path = (defined($args{'junction-path'})) ? $args{'junction-path'} : "/".$args{volume};
        my $junction_path = $args{'junction-path'};   
        if ( $class->{ui} eq 'ZAPI'){
            $ret = $class->{ApiSet}->volume_create( volume =>$args{volume},
                                                    "containing-aggr-name" => $args{aggregate}, 
                                                    "junction-path"=>$junction_path,
                                                    "vm-align-sector" => $args{'vm-align-sector'},
                                                    "vm-align-suffix" => $args{'vm-align-suffix'},
                                                    size => $args{size},);
        } else {
		
		if (exists ($args{'unix-permissions'})){
		
            		$ret = $class->{ApiSet}->volume_create( volume =>$args{volume},
                        	                            aggregate => $args{aggregate},  
                                	                    vserver => $args{vserver},  
                                        	            "junction-path" => $junction_path,
                                                	    "vm-align-sector" => $args{'vm-align-sector'},
                                                    	    "vm-align-suffix" => $args{'vm-align-suffix'},
                                             	      	    "unix-permissions" => $args{'unix-permissions'},
							    size => $args{size},
                                                	    type => $args{type},);
		} else {
			 $ret = $class->{ApiSet}->volume_create( volume =>$args{volume},
                                                            aggregate => $args{aggregate},
                                                            vserver => $args{vserver},
                                                            "junction-path" => $junction_path,
                                                            "vm-align-sector" => $args{'vm-align-sector'},
                                                            "vm-align-suffix" => $args{'vm-align-suffix'},
                                                            size => $args{size},
                                                            type => $args{type},);
		}
        }  
    } else {
        if ( $class->{ui} eq 'ZAPI'){
            $ret = $class->{ApiSet}->volume_create( volume =>$args{volume},
                                                    "containing-aggr-name" => $args{aggregate},
                                                    size => $args{size},
                                                    "vm-align-sector"=> $args{'vm-align-sector'},
                                                    "vm-align-suffix"=> $args{'vm-align-suffix'});
        } else {
            $ret = $class->{ApiSet}->vol_create( volume =>$args{volume},
                                                 "containing-aggr-name" => $args{aggregate},
                                                 size => $args{size},
                                                 "vm-align-sector"=> $args{'vm-align-sector'},
                                                 "vm-align-suffix"=> $args{'vm-align-suffix'});
            
            if( $args{mount} ){
                $class->{test}->nlog("volume mount is here : $ret \n"  );
                my $cmd = "exportfs -i -o rw,nosuid,anon=0 /vol/".$args{volume};
                $class->{ApiSet}->execute_raw_command(command => $cmd ) ;
            }
        }
    
    }  if (!$ret) { return $TCD::FAIL; }
    
    $class->{test}->nlog("volume dump is : $ret \n" . Dumper( $ret) ) if ($class->{_debug});

    return $ret;
}


## @name ::volume_show
##
## @summary  volume show (vol status)
##
## @description
## show volume status, specially good for checking 'vm-align-sector' and 'vm-align-suffix'
##  for 7mode, CLI command is: vol status -vC
##  for Cmode, CLI command is: volume show
##
## @return volume status output.
##
## @examples
##  $ret = vmalign_obj->volume_show(volume=>$vol_name, vserver => 'vs1',);
##  the value of 'vmalign_sector' is:
##  $vmalign_sector = @$ret[0]->{options}[0]->{'vmalign_sector'}
##
## @arg volume required: volume name
##
## @arg vserver optional: vserver name (for Cmode only, 7mode will ignore this arg)
##  default: $class->{vserver}
##
## @arg "verbose-list" optional: vol status "verbose-list" (for 7mode only, Cmode will ignore this arg)
##  default: 'vC'
##

sub volume_show{
    my $class = shift;
    my (%args) = validate( @_,
       {
            volume          => {type => SCALAR},
            vserver         => {type => SCALAR|UNDEF, optional => 1, default => $class->{vserver}},
            "verbose-list"  => {type => SCALAR, optional => 1, default=>'vC'},
       });
   
    my $ret;
   
    if ($class->{mode} eq 'CMode'){
        $ret = $class->{ApiSet}->volume_show(volume=>$args{volume}, vserver=>$args{vserver})->get_parsed_output();
        $class->{test}->nlog("volume show parsed output $ret". Dumper($ret)) if ($class->{_debug});
    } else{
        $ret = $class->{ApiSet}->vol_status(volume=>$args{volume},"verbose-list"=>$args{'verbose-list'})->get_parsed_output();
        $class->{test}->nlog("volume show parsed output $ret". Dumper($ret)) if ($class->{_debug});
 
    }
   
    return $ret;
}

## @name ::get_vmalign_attributes
##
## @summary  get vmalign suffix  and sector attributes
##
## @description
## get 'vm-align-suffix' or  'vm-align-sector' value from valume status
##
## @return 'vm-align-suffix' or 'vm-align-sector' value
##
## @examples
##  $vmalign_obj->get_vmalign_attributes(volume=>$vol_name, vserver => $vs, vm_option=>"vm-align-suffix");
##
## @arg volume required: volume name
##
## @arg vserver optional: vserver name (for Cmode only, 7mode will ignore this arg)
##  default: $class->{vserver}
##

sub get_vmalign_attributes{
    my $class = shift;
    my (%args) = validate( @_,
       {
            volume  => {type => SCALAR},
            vserver => {type => SCALAR|UNDEF, optional => 1, default => $class->{vserver}},
       });
   
    my $ret;
    my $vm_options = {};
    if ($class->{ui} eq 'ZAPI'){
        $ret = $class->zapi_volume_list_info(volume=>$args{volume});
        $class->{test}->nlog(" zapi volume list infor is: $ret\n".Dumper ($ret)) if ($class->{_debug});
    } else {   
        $ret = $class->volume_show(volume=>$args{volume}, vserver => $args{vserver},);
        $class->{test}->nlog("CLI volume value : $ret \n" . Dumper( $ret) ) if ($class->{_debug});
    }
   
    if ($class->{mode} eq 'CMode'){
        if ($class->{ui} eq 'ZAPI'){
            $vm_options->{vm_suffix} = $ret->{'volume-vm-align-attributes'}[0]->{'vm-align-suffix'};
            $vm_options->{vm_sector} = $ret->{'volume-vm-align-attributes'}[0]->{'vm-align-sector'};
        } else {
            $vm_options->{vm_suffix} = @$ret[0]->{'vm-align-suffix'};
            $vm_options->{vm_sector} = @$ret[0]->{'vm-align-sector'};
        }
    } else{
           if ($class->{ui} eq 'ZAPI'){
                $vm_options->{vm_suffix} =  @$ret[0]->{volumes}[0]->{'volume-info'}[0]->{'vm-align'}[0]->{'vm-align-info'}[0]->{'vm-align-suffix'};
                $vm_options->{vm_sector} =  @$ret[0]->{volumes}[0]->{'volume-info'}[0]->{'vm-align'}[0]->{'vm-align-info'}[0]->{'vm-align-sector'};
            } else {
                $vm_options->{vm_suffix} = @$ret[0]->{options}[0]->{'vmalign_suffix'};
                $vm_options->{vm_sector} = @$ret[0]->{options}[0]->{'vmalign_sector'};                
            }
    }
   
    $class->{test}->nlog("get_vmalign_suffix/sector: vmalign_suffix/sector is $ret" . Dumper($ret) )
            if ($class->{_debug});
    
    return $vm_options;
 
}


## @name ::testsetattr
##
## @summary testsetattr command
##
## @description
## run testsetattr command
##
## @return NACL API testsetattr output.
##
## @examples
## $ret = $vmalign_obj->testsetattr(path=>$path);
## 
## @arg path required: full path of the file: (example: $path = "/vol/".$vol_name."/".$file_name;)
##
## @arg 'get-attributes' optional : testsetattr parameter '-l'
##

sub testsetattr {
    my $class = shift;
    my (%args) = validate( @_,
       {
            path             => {type => SCALAR},
            'get-attributes' => {type => SCALAR, optional => 1, default => 1},
       });

    my $ret;
    my $path = $args{path};
    if ($class->{mode} eq 'CMode'){
        $ret = $class->{ApiSet}->system_node_run(node => $class->{filer},
                                                 command => "\"priv set test; testsetattr $path -l\"");
    } else {
        $class->{ApiSet}->set_privilege( 'privilege-level' => "test" );
        $ret = $class->{ApiSet}->testsetattr( path => $path,
                                              'get-attributes' => $args{'get-attributes'} );
    }
    
    $class->{test}->nlog("testsetattr: \n $ret\n". Dumper($ret)) if ($class->{_debug});
    
    return $ret;
   
}

## @name ::get_true_flags
##
## @summary get true_flags from testsetattr
##
## @description
## get true_flags from testsetattr
##
## @return true_flags value (for vm align volume, true flag should be '0x4000002')
##
## @examples
## $ret = $vmalign_obj->get_true_flags(path=>$path);
## 
## @arg path required: full path of the file: (example: $path = "/vol/".$vol_name."/".$file_name;)
##

sub get_true_flags {
    my $class = shift;
    my (%args) = validate( @_,
       {
            path => {type => SCALAR},
       });

    my $ret;
    if ($class->{mode} eq 'CMode'){
        $ret = $class->testsetattr(path=>$args{path})->get_processed_output();
        
        my @flags = grep {$_ =~ 'true_flags'} split (/\n/, $ret);
        @flags = split(/:\s+/, @flags[0] );
        $ret = @flags[1];
    } else {
         $ret = $class->testsetattr(path=>$args{path})->get_parsed_output();
         $ret = @$ret[0]->{true_flags};
    }
      
    return $ret;
   
}


## @name ::client_mount
##
## @summary mount the client
##
## @description
## mount the client
##
## @return client mount output.
##
## @examples
## $vmalign_obj->client_mount(volume=>$vol_name,
##                            filerdataip=>$filerDataIp,
##                            clientopsname=>'t1' );
##
## @arg volume required: volume name
##
## @arg filerdataip optional: filer's ip address
##  default: FILER_IP
##
## @arg clientmountpath optiona;: mount path of client
##  default: /t/vol_name
##
## @arg mntopts optional: mount options
##  default: 'noac,hard,intr,proto=tcp,vers=3'
##
## @arg clientopsname optional : client operation name for later client's operation
##  default: 't1'
##
## @arg nfs_version optional : client mount through nfsv3 or nfsv4
##  default: 't1'
##

sub client_mount{
    my $class = shift;
    my (%args) = validate( @_,
       {
           volume           => {type => SCALAR},
           filerdataip      => {type => SCALAR, optional=>1, default=>$class->{filerip} },
           clientmountpath  => {type => SCALAR, optional => 1},
           mntopts          => {type => SCALAR,optional => 1, default => 'noac,hard,intr,proto=tcp,vers=3' },
           clientopsname    => {type => SCALAR, optional=>1, default=>'t1'},
           nfs_version      => {type => SCALAR, optional=>1, default=>'v4'},
       });
   
    my $clientmountpath = (defined($args{clientmountpath})) ? $args{clientmountpath} : "/t/".$args{volume};

    my ($filerjunctionpath, $ret);
    if ($class->{mode} eq 'CMode'){
         $filerjunctionpath = "/".$args{volume};
    } else{
          $filerjunctionpath = "/vol/".$args{volume};
    }
    
    if ($args{nfs_version} eq 'v4') {
        $ret = $class->{client} ->mount_nfsv4({filerdataip       => $args{filerdataip},
                                               filerjunctionpath => $filerjunctionpath,
                                               clientmountpath   => $clientmountpath,
                                               mntopts           => $args{mntopts},
                                               clientopsname     => $args{clientopsname}});
    } else {
        $ret = $class->{client} ->mount_nfsv3({filerdataip       => $args{filerdataip},
                                               filerjunctionpath => $filerjunctionpath,
                                               clientmountpath   => $clientmountpath,
                                               mntopts           => $args{mntopts},
                                               clientopsname     => $args{clientopsname}});
        
    }
    if (!$ret) {  return $TCD::FAIL; }
  
    return $ret;
}


## @name ::volume_cleanup
##
## @summary clean up the volume
##
## @description
## unmount, offline, destroy volume
##
## @examples
## $ret = $vmalign_obj->volume_cleanup(volume=>'testvol1', vserver=>'vs1');
##
## @arg volume required: volume name
##
## @arg vserver optional: vserver name
##  default: $class->{vserver}
##

sub volume_cleanup {
    my $class = shift;
    my (%args) = validate( @_,
    {
        volume          => {type => SCALAR},
        vserver         => {type => SCALAR|UNDEF, optional=>1},
    });
      
    my $vserver = (defined($args{vserver})) ? $args{vserver} : $class->{vserver};
    my $volume  = NACL::C::Volume->new(command_interface => $class->{node},
                                       volume            => $args{volume},
                                       vserver           => $vserver,);
    
    if ($volume->state()->junction_active() =~ qr/true/xi) {
        $volume->unmount();
    }
    
    if ($volume->state()->state() !~ qr/offline/xi) {
        $volume->offline();
    }

    $volume->delete(foreground => 'true');

}

## @name ::dumpblock
##
## @summary dumpblock for file contents
##
## @description
## dumpblock contents
##
## @return dumpblock output
##
## @examples
## $ret = $vmalign_obj->dumpblock(volume=>$vol_name, file_path => $file_name, fbn => 2);
##
## @arg volume required: volume name
##
## @arg file_path required: file path
##  default: $class->{vserver}
##
## @arg fbn optional: fbn
##  default: 0
##

sub dumpblock {
    my $class = shift;
    my (%args) = validate( @_,
       {
            volume    => {type => SCALAR},
            file_path => {type => SCALAR},
            fbn       => {type => SCALAR, optional => 1, default => 0},
       });
     
    # get file handle
    my $fh ;
    my $full_path = "/vol/".$args{volume}."/".$args{file_path};
    
    my $cmd = "showfh $full_path ";
    my $ret = $class->_run_cmd( cmd => $cmd );
    $class->{test}->nlog("showfh: $ret\n". Dumper($ret))  if ($class->{_debug});
    
    if ($ret =~ /(fileid=)(0x\d+)/){
        $fh = $2;
    }
    
    # dumpblock
    $cmd = "dumpblock file $args{volume} $fh $args{fbn} ";
    $ret = $class->_run_cmd( cmd => $cmd );
    $class->{test}->nlog("dumpblock: \n $ret\n". Dumper($ret)) if ($class->{_debug});
    
    return $ret;
}

## @name ::get_16bytes_content
##
## @summary get 16 bytes block contents from dumpblock
##
## @description
## get 16 bytes block contents from dumpblock
##
## @return 16bytes file content
##
## @examples
## $ret = $vmalign_obj->get_16bytes_content(volume=>$vol_name, file_path => $file_name, fbn => 0, byte_num=>'0x0200');
##  0x0200 is from the byte 512
##
## @arg volume required: volume name
##
## @arg file_path required: file path
##  default: $class->{vserver}
##
## @arg byte_num optional: byte_num in hex to retrieve
##
## @arg fbn required: fbn
##  default: 0
##

sub get_16bytes_content {
    my $class = shift;
    my (%args) = validate( @_,
       {
            volume    => {type => SCALAR},
            file_path => {type => SCALAR},
            fbn       => {type => SCALAR, optional => 1, default => 0},
            byte_num  => {type => SCALAR},
       });
    
    my $ret = $class->dumpblock(volume=>$args{volume}, file_path => $args{file_path}, fbn => $args{fbn});
    
    my @bytes = split (/\n/, $ret);
    
    if (my @out = grep {$_ =~ $args{byte_num}} @bytes) {
        $class->{test}->nlog("get_block_type_content: $out[0];;;$args{byte_num} \n". Dumper(@out)) ;
        if ($out[0] =~ /<(\S+)>/){
            $ret = $1;
        }
    }

    return $ret;
}


## @name ::get_buftree
##
## @summary get buftree
##
## @description
## run "wafl buftree check" command
##
## @return NACL API 'wafl buftree check' output.
##
## @examples
## $vmalign_obj->get_buftree(volume=>$vol_name, filename => $file_name);
##
## @arg volume required: volume name
##
## @arg filename optional: file name (including directory path)
##

sub get_buftree {
   
    my $class = shift;
    my %args = validate( @_,
        {  
            volume   => { type => SCALAR },
            filename => { type => SCALAR },
        }
    );
    
    my $ret;
    my $cmd;
    my $inode;
    my $full_node_path = "/vol/".$args{volume};
    my %buftree_data = ();
    my $volume_name  = $args{volume};

    my $filename      = $args{filename};
    my $full_filename = $full_node_path . '/' . $filename;
    my $scope = $volume_name."./".$filename;

    # get file's inode
    if ($class->{mode} eq 'CMode'){
         $ret = $class->{ApiSet}->volume_explore( format => "inode", scope=> $scope)->get_parsed_output();
         $class->{test}->nlog("volume explore inode $ret\n". Dumper($ret)) if ($class->{_debug});
         
         #ignore empty hash ref in output
         my $i;
         for($i = 0; $i <= @$ret; $i++){
            my $hash_ref = @$ret[$i];
            if(!keys %$hash_ref) { next;}
            else {last;}
         }
         $inode = @$ret[$i]->{inum};
    } else {
        $inode = $class->{ApiSet}->slist(path => $full_filename)->get_parsed_output()->[0]->{'inode'};
    }
   
    if ( !$inode ) {
        NATE::Exceptions::Argument->throw("Unable to get inode for file");
    }

    $cmd = "wafl buftree check -n -V $volume_name $inode ";
    my $buftree_data = $class->_run_cmd( cmd => $cmd );
    
    $ret = _buftree_stats_pars($buftree_data);
    $class->{test}->nlog("get buftree results: $ret->{levels}[0]{blks64} \n" . Dumper($ret)) if ($class->{_debug});
    
    return $ret;
} 

## @name ::_run_cmd
##
## @summary run raw command for 7mode or in dblade for cmode
##
## @description
## run raw command
##
## @return NACL API command output.
##
## @examples
## $vmalign_obj->_run_cmd(cmd => $cmd);
##
## @arg cmd required: comamnd to run
##

sub _run_cmd {
    my $class = shift;
    my (%args) = validate( @_, { cmd => { type => SCALAR }, } );
    
    my $ret;
    my $new_cmd = "priv set test; " . $args{cmd};
 
    if ($class->{mode} eq 'CMode'){
        $new_cmd = "\"".$new_cmd."\"";
        $ret = $class->{ApiSet}->system_node_run(node => $class->{filer}, command => $new_cmd)->get_processed_output();
    } else {
        $ret = $class->{ApiSet}->execute_raw_command( command      => $new_cmd,
                                                      'timeout' => 7200
        );
    }
    $Log->comment("response = $ret ". Dumper($ret)) if ($class->{_debug}) ;
    
    return $ret;

} 


## @name ::get_buftree_value
##
## @summary get value from buftree
##
## @description
## get value from buftree
##
## @return value from bufftree
##
## @examples
## $vmalign_obj->get_buftree_value(volume=>$vol_name, filename => $file_name);
##
## @arg volume required: volume name
##
## @arg filename optional: file name (including directory path)
##
## @arg level optional: block level
##  default: 0
##
## @arg value optional: item to get the value from 
##  default: blks64
##  value can be the value of: holes32, holes64, blks64, blks32, total
##

sub get_buftree_value {
   
    my $class = shift;
    my %args = validate( @_,
        {  
            volume   => { type => SCALAR },
            filename => { type => SCALAR },
            level    => { type => SCALAR, optional => 1, default => 0 },
            value    => { type => SCALAR, optional => 1, default => 'blks64' },
        }
    );
    
    my $ret = $class->get_buftree(volume=>$args{volume}, filename => $args{filename});
       
    return $ret->{levels}[$args{level}]{$args{value}};
}


##    # parse this:
##    #
##    # ----- ------------ ------------ ------------ ------------ ------------
##    # level 32-bit blks  32-bit holes 64-bit blks  64-bit holes     total
##    # ----- ------------ ------------ ------------ ------------ ------------
##    #     3            0            0            1            0            1
##    #     2            0            0            2            1            3
##    #     1            0            0            4          509          513
##    #     0            0            0           54       130515       130569
##    # total            0            0           61       131025       131086
##    # ----- ------------ ------------ ------------ ------------ ------------
##    #

sub _buftree_stats_pars {
    my ($str) = @_;
    my @stats_lines = split( /\r?\n/, $str );
    my %hash;
    my $stats = \%hash;

    foreach (@stats_lines) {
        next if /^s+$/;
        next if !/^\s*(\d+|[^\s]+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)$/;
        my $level = $1;    # 0, 1, 2, ... or "total"

        my $levelstats = {
            blks32  => $2,
            holes32 => $3,
            blks64  => $4,
            holes64 => $5,
            total   => $6,
        };

        if ( $level =~ /^\d+$/ ) {
            $stats->{levels}->[$level] = $levelstats;
        } else {
            $stats->{$level} = $levelstats;
        }
    }
    
    $Log->comment("buftree stats = $stats.");
    
    return $stats;
}


## @name ::zapi_volume_list_info
##
## @summary get volume info from zapi
##
## @description
##  get volume info from zapi
##
## @return volume status.
##
## @examples
## $vmalign_obj->->zapi_volume_list_info(volume=>$vol_name);
##
## @arg volume required: volume name
##

sub zapi_volume_list_info {
   
    my $class = shift;
    my %args = validate( @_,
        {  
            volume   => { type => SCALAR },
        }
    );
    
    my $ret;
    if ($class->{mode} eq 'CMode'){
        $ret = $class->{ApiSet}->volume_get_iter()->get_parsed_output();
               
        my $vols = @$ret[0]->{'attributes-list'}[0]->{'volume-attributes'};        
        foreach (@$vols){
            my $name = $_->{'volume-id-attributes'}[0]->{name};
            if ($name eq  $args{volume}){
                $ret = $_;
                last;
            }
        }
        
    } else {
        $ret = $class->{ApiSet}->volume_list_info(volume=>$args{volume})->get_parsed_output();
    }
    
    $class->{test}->nlog("ZAPI volume-list-info: $ret \n" . Dumper($ret))  if ($class->{_debug});
  
    return $ret;
}

1;

