# -*- perl -*-
## @summary wreck, wack and wafliron related library routines
## @usage
##	use Dense::WackIron;
## @author dl-dpgqa-automation@netapp.com
## @author rani@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Wreck error injection, Wack and WaflIron operations for dense volume script.

=head1 DESCRIPTION

This module defines routines for wack and iron that are used in dense 
volumes test scripts.

=head1 FUNCTIONS

=cut

package Dense::WackIron;

require Exporter;
our @ISA= qw(Exporter);
our @EXPORT = qw ( 
		   wack_and_report
		   iron_and_report
		   wack_iron_and_report
		   inject_and_fix_errors
              );

use filer;
use wack;
use Wreck;
use Wafliron;
use Dense qw (:all);
use Dense::Constants;
use Dense::Utils;
use Dense::TestData;
use Tharn;
use Tharn qw(connect host say logcomment logresult param tstlog);
use Utils qw(make_conn);
use Data::Dumper qw(Dumper);
use Volume qw (vol_list vol_status vol_root);
use SPARSE::Resource::Volume;
use SPARSE::Resource::Filer;
use cat qw(do_shell);
Params::Validate::validation_options( allow_extra => 1);
use Params::Validate qw(validate validate_pos OBJECT SCALAR HASHREF ARRAYREF UNDEF);
BEGIN {
Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
}

use strict;

=head2 wack_and_report(filerhost=>$filerHost, aggr=>$aggrName, error_expected=>$error_expected, commit=>$commit, bootimage=>$bootimage, timeout=>$timeout)

Run wack and report errors

=over 4

=item filer host 

Required, filer host, not connection 

=item aggrname

Required, aggrname that wack runs on 

=item error_expected 

Optional, will not die after finding the error on wack if it is 1 
Default is 0, which will die if find error 

=item commit

Optional, will not commit to the wack if it is 0 
Default is 1, which will commit

=item bootimage

Optional, full path to the bootimage of a simulator
No default value

=item timeout 

Optional, time out for wack 

=item Returns:

Success

=over 4

Return null

=back

Fail 

=over 4

Die with error in C<$@>

=back

=back

=cut

sub wack_and_report($@)
{
    my (%opts) = validate (@_, { filerhost => { type=>SCALAR },
				 aggr => { type=>SCALAR },
				 error_expected => { type=>SCALAR, optional=>1, default=>"0", regex=>qr/^(0|1)$/ },
				 commit => { type=>SCALAR, optional=>1, default=>"1", regex=>qr/^(0|1)$/ },
				 bootimage => { type=>SCALAR | UNDEF, optional=>1 },
				 timeout => { type=>SCALAR, optional=>1 } }
			  );

    my $host = host($opts{filerhost});
    my $aggrName = $opts{aggr}; 
    my $conn = connect("#rsh");
    
    my $is_sim = 0;
    my $wack_result;
    my $output = '';

    $opts{error_expected} = 0 unless $opts{error_expected};
    $opts{commit} = 1 unless $opts{commit};

    # check and set the simulator flag
    my $filer = dense_new_filer(host=>$opts{filerhost});
    if ($filer->{hosttype} =~ /simulator/) {
        $is_sim = 1;
    }

    if ($is_sim){
        $wack_result = wack($host, args=>$aggrName, commit=>$opts{commit}, bootimage=>$opts{bootimage},output=>\$output);
    }else{
        $wack_result = wack($host, args=>$aggrName, commit=>$opts{commit}, output=>\$output, timeout=>$opts{timeout});
    }

    if (!$wack_result) {
        die "Wack problem:$@";
    } elsif ($wack_result eq WACK_CLEAN) {
        if ($opts{error_expected} == 1){
            die "File system is unexpectedly clean by Wack\n";
        } else {
	    logcomment("File system is clean as expected by Wack");
	}
    } elsif ($wack_result eq WACK_DIRTY) {
       if ($opts{error_expected} == 0){
           die "File system is unexpectedly dirty by Wack\n";
       } else {
	   logcomment("File system is corrupt and fixed as expected by Wack")
       }
    } else {
        die "Wack is in unknown state: $wack_result\n";
    }

    logcomment("Filer boot after wack");
	reboot_verify($host, timeout=>1200);
}


=head2 iron_and_report(filer=>$filer, aggr=>$aggrName, error_expected=>$error_expected, client=>$client, partner=>1)

Run wafliron and report errors

=over 4

=item filer

Required, filer object

=item aggrname

Required, aggrname that iron runs on 

=item error_expected 

Optional, will not die after finding the error on wack if it is 1 
Default is 0, which will die if find error 

=item partner

0 or 1. default is 0. If set to 1, run iron on partner volumes

=item Returns:

Success

=over 4

Return null

=back

Fail 

=over 4

Die with error in C<$@>

=back

=back

=cut

sub iron_and_report($@)
{
    my (%opts) = validate (@_, { filer => { type=>HASHREF },
                                 aggr => { type=>SCALAR },
	            error_expected => { type=>SCALAR, optional=>1, default=>"0", regex=>qr/^(0|1)$/ }, 
				client => { type=>SCALAR, optional=>1, default=>"localhost" },
	                        speed => { type=>SCALAR, default=>2000, regex=>qr/^\d+$/ }, 
	                        wait => { type=>SCALAR, default=>1, regex=>qr/^(0|1)$/ }, 
	                        waflopts => { type=>HASHREF, optional=>1 }, 
				partner => { type=>SCALAR, optional=>1, default=>0 }  }
				 );

    my ($filer) = $opts{filer};
    my ($aggr) = $opts{aggr};
    my $speed_cmd = "wafl scan speed";
    my $conn = $filer->{conn};
    my $timeout = 3600;
    my ($current_speed, $corruption_found, $corruption_message, %waflopts);
    %waflopts = %{$opts{waflopts}} if(defined $opts{waflopts});
    @waflopts{qw(wait scan_speed)} = @opts{qw(wait speed)};

    $current_speed = filer_cli_test_say($conn, $speed_cmd);
    ($current_speed) = ($current_speed=~/^WAFL scan speed is (\d+)$/);
    logcomment("current speed is $current_speed");
    logcomment("Calling wafliron_start with aggr $aggr and ".Dumper(\%waflopts));
    ($corruption_found, $corruption_message) = wafliron_start($conn, $aggr, %waflopts) or die("Not able to do wafliron: $@");
    logcomment("Corruption Found: <<$corruption_found>>, Corruption Message: <<$corruption_message>>");
    if($opts{wait}) {
       logcomment("Setting scan speed back to the original $current_speed");
       filer_cli_test_say($conn, "$speed_cmd $current_speed");
    }
    return($corruption_found) if(!$opts{wait});
    die($corruption_message) if($corruption_found && !$opts{error_expected});
    die("File system is clean but expected corruption") if(!$corruption_found && $opts{error_expected});
    return($corruption_message) if($corruption_found && $opts{error_expected});
}

=head2 wack_iron_and_report(host=>$host, conn=>$conn, vol=>$vol, list=>$list, options=>$options, image=>$image, client=>$client)

Use wreck to inject errors and run wack or wafliron to fix errors

=over 4

=item host

Required, filer host

=item conn

Required, filer connection

=item vol

Required, volume object

=item list

Required, list of corruption to be injected

=item image

Optional, full path to the bootimage of a simulator
No default value

=item options 

valid options are iron_iron, wack_iron, iron_wack and wack_wack

Optional, specify whether wack or iron will be used for fix errors

=item Returns:

Success

=over 4

Return null

=back

Fail 

=over 4

Die with error in C<$@>

=back

=back

=cut
   
sub wack_iron_and_report($@)
{
    my (%opts) = validate (@_, { host => { type=>SCALAR },
				 conn => { type=>HASHREF },
				 vol => { type=>HASHREF },
				 list => { type=>SCALAR },
				 image => { type=>SCALAR, optional=>1 },
				 client => { type=>SCALAR, optional=>1, default=>"" },
				 options => { type=>SCALAR, optional=>1, regex=>qr/^(iron_iron|wack_iron|iron_wack|wack_wack)$/ } }
			  );
 
    my ($host) = $opts{host};
    my ($conn) = $opts{conn};
    my ($vol) = $opts{vol};
    my ($list) = $opts{list};
    my ($options) = $opts{options};

    my $volName = $vol->{name};
    my $path = '/vol/'.$volName;
    my $filer = $vol->{filer};

    # kick off dense operation and inject errors if specified
    dense_toggle(host=>$host, path_name=>$path, state=>"on");
    dense_start(host=>$host, path_name=>$path);
    dense_wait_for_status(host=>$host, path_name=>$path, status=>"IDLE");

    wreck($conn, $vol->{name}, type=>"flex", inject_script=>$list)
	or die "Problem introducing inconsistencies via wreck:$@";

   ## Online volume where error is injected. WACK will not fix errors in restricted volumes
    ##Temp $vol->online();

    if ($opts{options} eq 'iron_iron') {
	logcomment("Running Iron-Iron on filer $host and aggregate $vol->{aggrName}");

	iron_and_report(filer=>$filer, aggr=>$vol->{aggrName}, error_expected=>1, client=>$opts{client});
	sleep(60); ## Wait till Iron commits changes
	iron_and_report(filer=>$filer,  aggr=>$vol->{aggrName}, client=>$opts{client});

    } elsif ($opts{options} eq 'wack_iron') {
	logcomment("Running WACK-Iron on filer $host and aggregate $vol->{aggrName}");

	wack_and_report(filerhost=>$host, aggr=>$vol->{aggrName}, error_expected=>1, bootimage=>$opts{image});
	iron_and_report(filer=>$filer,  aggr=>$vol->{aggrName}, client=>$opts{client});

    } elsif ($opts{options} eq 'iron_wack') {
	logcomment("Running Iron-WACK on filer $host and aggregate $vol->{aggrName}");

	iron_and_report(filer=>$filer, aggr=>$vol->{aggrName}, error_expected=>1, client=>$opts{client});
	wack_and_report(filerhost=>$host, aggr=>$vol->{aggrName}, bootimage=>$opts{image});

    } elsif ($opts{options} eq 'wack_wack') {
	logcomment("Running WACK-WACK on filer $host and aggregate $vol->{aggrName}");

	wack_and_report(filerhost=>$host, aggr=>$vol->{aggrName}, error_expected=>1, bootimage=>$opts{image});
	wack_and_report(filerhost=>$host, aggr=>$vol->{aggrName}, bootimage=>$opts{image});

    } else {
	$vol->online();
    }
}

=head2 inject_and_fix_errors(host=>$host, conn=>$conn, vol=>$vol, list=>$list, client=>$client)

Use wreck to inject errors and run wack or wafliron to fix errors

=over 4

=item host

Required, filer host

=item conn

Required, filer connection

=item vol

Required, volume object

=item list

Required, list of corruption to be injected

=item Returns:

Success

=over 4

Return null

=back

Fail 

=over 4

Die with error in C<$@>

=back

=back

=cut

sub inject_and_fix_errors($@)
{
    my (%opts) = validate (@_, { host => { type=>SCALAR },
				 conn => { type=>HASHREF },
				 vol => { type=>HASHREF },
				 list => { type=>SCALAR },
				 client => { type=>SCALAR, optional=>1, default=>"" } } 
			  );

    my ($host) = $opts{host};
    my ($conn) = $opts{conn};
    my ($vol) = $opts{vol};
    my ($list) = $opts{list};
    
    my $filer = dense_new_filer(host=>$host);
    if ($main::FAST_MODE) {
	wack_iron_and_report(host=>$host, conn=>$conn, vol=>$vol, list=>$list, client=>$opts{client});
    } elsif ($filer->{hosttype} =~ /simulator/) {
	wack_iron_and_report(host=>$host, conn=>$conn, vol=>$vol, list=>$list, options=>"iron_iron", client=>$opts{client});
    } elsif ($main::CHECK_MODE == 0) {
	wack_iron_and_report(host=>$host, conn=>$conn, vol=>$vol, list=>$list, options=>"iron_iron", client=>$opts{client});
    } elsif ($main::CHECK_MODE == 1) {
	wack_iron_and_report(host=>$host, conn=>$conn, vol=>$vol, list=>$list, options=>"iron_wack", client=>$opts{client});
    } elsif ($main::CHECK_MODE == 2) {
	wack_iron_and_report(host=>$host, conn=>$conn, vol=>$vol, list=>$list, options=>"wack_iron", client=>$opts{client});
    } elsif ($main::CHECK_MODE == 3) {
	wack_iron_and_report(host=>$host, conn=>$conn, vol=>$vol, list=>$list, options=>"wack_wack", client=>$opts{client});
    }
}

        
#----------------------------------------------------------------------
# The testme routine to test the exported functions of this library
#                       
# ntest PKGNAME=Dense::WackIron SUBNAME=testme util/test_pkg.thpl
#                       
# The routine can also be invoked with many optional parameters, such as:
#
# ntest PKGNAME=Dense::WackIron SUBNAME=testme FUNCTION_NAME=wack_and_report util/test_pkg.thpl
#                       
#----------------------------------------------------------------------
                        
sub testme() {  
                        
        package main;
        
        our ($FILER, $FUNCTION_NAME, $FILER_VOL, $FILER_CONN, $FILER_AGGR, $FILE_NAME);
        
        param('FILER',       '-mesg',    "Filer to test");
        param('FILER_CONN','-default',"#rsh");
        param('FUNCTION_NAME',   '-default',    "all");
        param('FILER_VOL',   '-default', 'vol0');
        param('FILER_AGGR',   '-default', 'test1');
        param('FILE_NAME',   '-default', 'fileA');
                        
        package Dense::WackIron;

        logcomment("Testing on FILER $FILER, Function $FUNCTION_NAME");
        my $ret;        
        my $path_name = "/vol/".$FILER_VOL;

        if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "wack_and_report"){

                eval {
                        $ret = wack_and_report(filerhost=>$FILER, aggr=>$FILER_AGGR, error_expected=>0, commit=>1, timeout=>3600);
                };

                if ($@ =~ /Connection timed out/) {
                        logresult('pass', "Expected fail. Check filer and connection type");
                } elsif ($@) {
                        logresult('fail', "$@");
                } else {
                        logresult('pass', "Got expected result");
                }
        }

	if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "iron_and_report"){

                eval {
			my $filer = dense_new_filer(host=>$FILER);
			my $conn = $filer->{conn};
                        $ret = iron_and_report(filer=>$filer, aggr=>$FILER_AGGR, error_expected=>1);
                };

                if ($@ =~ /Connection timed out/) {
                        logresult('pass', "Expected fail. Check filer and connection type");
                } elsif ($@) {
                        logresult('fail', "$@");
                } else {
                        logresult('pass', "Got expected result");
                }
        }

	if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "wack_iron_and_report"){

		logcomment("Running wack_wack on filer $FILER");
                eval {
			my $filer = dense_new_filer(host=>$FILER);
			my $conn = $filer->{conn};
                        my $dvol = dense_setup_test_vvol(filer=>$filer, volName=>$FILER_VOL, volSize=>"3g", purge=>1, aggr=>$FILER_AGGR, aggrPurge=>0);
			$ret = dense_mk_dup_datafile (vol=>$dvol, filename=>$FILE_NAME, pattern_size=>4096, rpt_cnt=>2);
			my $filepath = "path:". "/TEST/". $FILE_NAME;
                        my $list = "inode_flags_remove inum=$filepath flags=0x00008000". "\n";
                        $ret = wack_iron_and_report(host=>$FILER, conn=>$conn, vol=>$dvol, list=>$list, options=>"wack_wack");
                };

                if ($@ =~ /Connection timed out/) {
                        logresult('pass', "Expected fail. Check filer and connection type");
                } elsif ($@) {
                        logresult('fail', "$@");
                } else {
                        logresult('pass', "Got expected result");
                }

		logcomment("Running wack_iron on filer $FILER");
		eval {
                        my $filer = dense_new_filer(host=>$FILER);
                        my $conn = $filer->{conn};
                        my $dvol = dense_setup_test_vvol(filer=>$filer, volName=>$FILER_VOL, volSize=>"3g", purge=>1, aggr=>$FILER_AGGR, aggrPurge=>0);
			$ret = dense_mk_dup_datafile (vol=>$dvol, filename=>$FILE_NAME, pattern_size=>4096, rpt_cnt=>2);
			my $filepath = "path:"."/TEST/". $FILE_NAME;
                        my $list = "inode_flags_remove inum=$filepath flags=0x00008000". "\n";
                        $ret = wack_iron_and_report(host=>$FILER, conn=>$conn, vol=>$dvol, list=>$list, options=>"wack_iron");
                };

                if ($@ =~ /Connection timed out/) {
                        logresult('pass', "Expected fail. Check filer and connection type");
                } elsif ($@) {
                        logresult('fail', "$@");
                } else {
                        logresult('pass', "Got expected result");
                }

		logcomment("Running iron_iron on filer $FILER");
		eval {
                        my $filer = dense_new_filer(host=>$FILER);
                        my $conn = $filer->{conn};
                        my $dvol = dense_setup_test_vvol(filer=>$filer, volName=>$FILER_VOL, volSize=>"3g", purge=>1, aggr=>$FILER_AGGR, aggrPurge=>0);
			$ret = dense_mk_dup_datafile (vol=>$dvol, filename=>$FILE_NAME, pattern_size=>4096, rpt_cnt=>2);
			my $filepath = "path:"."/TEST/". $FILE_NAME;
                        my $list = "inode_flags_remove inum=$filepath flags=0x00008000". "\n";
                        $ret = wack_iron_and_report(host=>$FILER, conn=>$conn, vol=>$dvol, list=>$list, options=>"iron_iron");
                };

                if ($@ =~ /Connection timed out/) {
                        logresult('pass', "Expected fail. Check filer and connection type");
                } elsif ($@) {
                        logresult('fail', "$@");
                } else {
                        logresult('pass', "Got expected result");
                }

		logcomment("Running iron_wack on filer $FILER");
		eval {
                        my $filer = dense_new_filer(host=>$FILER);
                        my $conn = $filer->{conn};
                        my $dvol = dense_setup_test_vvol(filer=>$filer, volName=>$FILER_VOL, volSize=>"3g", purge=>1, aggr=>$FILER_AGGR, aggrPurge=>0);
			$ret = dense_mk_dup_datafile (vol=>$dvol, filename=>$FILE_NAME, pattern_size=>4096, rpt_cnt=>2);
			my $filepath = "path:"."/TEST/". $FILE_NAME;
                        my $list = "inode_flags_remove inum=$filepath flags=0x00008000". "\n";
                        $ret = wack_iron_and_report(host=>$FILER, conn=>$conn, vol=>$dvol, list=>$list, options=>"iron_wack");
                };

                if ($@ =~ /Connection timed out/) {
                        logresult('pass', "Expected fail. Check filer and connection type");
                } elsif ($@) {
                        logresult('fail', "$@");
                } else {
                        logresult('pass', "Got expected result");
                }

        }

	if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "inject_and_fix_errors"){

                eval {
			my $filer = dense_new_filer(host=>$FILER);
                        my $conn = $filer->{conn};
                        my $dvol = dense_setup_test_vvol(filer=>$filer, volName=>$FILER_VOL, volSize=>"3g", purge=>1, aggr=>$FILER_AGGR, aggrPurge=>0);
			$ret = dense_mk_dup_datafile (vol=>$dvol, filename=>$FILE_NAME, pattern_size=>4096, rpt_cnt=>2);
			my $filepath = "path:"."/TEST/". $FILE_NAME;
                        my $list = "inode_flags_remove inum=$filepath flags=0x00008000". "\n";
                        $ret = inject_and_fix_errors(host=>$FILER, conn=>$conn, vol=>$dvol, list=>$list);
                };

                if ($@ =~ /Connection timed out/) {
                        logresult('pass', "Expected fail. Check filer and connection type");
                } elsif ($@) {
                        logresult('fail', "$@");
                } else {
                        logresult('pass', "Got expected result");
                }
        }

} ## End of Testme Routine

=head1 SEE ALSO

    Where additional information can be found.

=head1 BUGS

    Description of bugs or what the library does not do and why

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = volume, df, useradmin

=head1 AUTHOR

    Rani Palanisami (rani@netapp.com)
    $Id: //depot/prod/test/Rsierranevada/lib/Dense/WackIron.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

1;
