# -*- perl -*-
#  $Id: $
#
## @summary
##    Library routines for compression with dense
#### @author dl-dense-qa@netapp.com
## @author dubey@netapp.com
## @status in-progress
## @pod here
=head1 NAME

NDVM - Basic routines for compression with dedupe functionality

=head1 DESCRIPTION

This module defines routines for compression 
=head1 FUNCTIONS 

Compression basic functions

=cut

package Dense::dense_compression;
use strict;
use SPARSE::Util::Logger;
use filer;
use compression;
use Burt;
use Tharn;
use Data::Dumper qw(Dumper);
use Dense qw(:all);
use Aggregate qw(:all);
use Utils qw(make_conn);
use Dense::Status;
use Dense::Constant::Command;
use Compression::Constants qw(:all);
use Dense::Constant::Message;
use Filer::Version;
use Params::Validate qw(validate validate_pos OBJECT SCALAR BOOLEAN HASHREF ARRAYREF UNDEF);
use FSR::FSRutils;
use Date::Manip;
use Time::ParseDate;
use snaplock qw(system_time_set system_time_get);
use VFILER::Vfiler;
use Volume qw(volume_options_list_info);
use Admin::License qw (is_licensed license_raw);

BEGIN {
Params::Validate::validation_options(
    on_fail => sub {
        my @msgs = @_;
        chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))
        }
   );
}

use base qw(Exporter);

BEGIN {
	my @action = qw (
			dense_compress_start
			copy_dataset
			return_expected_savings
			add_compression_license
			wait_for_cp
			 );
	my @status = qw (
			compression_status
			compression_status_wait
			);
	my @verify = qw (
			get_compression_saving
			);
	our %EXPORT_TAGS = ( action => [ @action],
			     status => [ @status],
			     verify => [ @verify]
			   );
	$EXPORT_TAGS{'all'} = [ @action, @status, @verify];
	Exporter::export_ok_tags('all');
}

=head2 dense_compress_start(host=>$host, vol=>$vol, compress_dedupe=>$compress_dedupe, compress_shared=>$compress_shared, delete_checkpoint=>$delete_checkpoint, partner=>$partner)

Function to run vol compress start on the volume. In this function, $host, and $vol are mandatory parameters.
Rest of the parameters are optional.

=over 4

=item host

Filer HostSpec

=item vol

The volume on which compression start has to be run

Failure

=over 4

Return with an error message in C<$@>

=back

=back

=cut

sub dense_compress_start($@) {
        my (%opts) = validate(@_, {
                        host => {type=>SCALAR},
                        vol => {type=>SCALAR},
			options => {type=>SCALAR, optional=>1},
                        partner => {type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/}
                        }
                      );

        my ($host) = $opts{host};
        my ($vol) = $opts{vol};
	my ($options) = $opts{options};
	my $telnet_conn = make_conn( $host, "#telnet" );	
	my $conn =  make_conn( $host, "#rsh");
        my $cmd;

	$cmd = VOL_COMPRESS_START . " $options" . " $vol";
        if ( $opts{partner} ) {
            $cmd = "partner" . " " . VOL_COMPRESS_START . " $options" . " $vol";
        } 

	if($opts{options} ) {
		$options = " " . "$options" . " ";
		my ($return_text, undef, undef, undef) = say($telnet_conn, $cmd, -noprompt, '-wait', "Are you sure you want to proceed" );
		if ($return_text =~ /Are you sure/) {
                    say($telnet_conn, "yes");
		}

	} else {
		$options = " ";
       		my ($return_text, $result_code) = filer_cli_say($conn, $cmd);
        	if (remove_leading_trailing_spaces($return_text)) {
                    die("$return_text\n");
        	}
	}
	disconnect($telnet_conn);
	return 1;
}


=head2 copy_dataset(test_client=>$test_client_path, mount=>$mount_point, client=>$client_conn);

Function to copy the compression dataset to the volume mount location 

=over 4

=item test_client

Testbed/ test client location

=item mount

The nfs mount location of the volume

=item client

The client connection

Failure

=over 4

Return with an error message in C<$@>

=back

=back

=cut

sub copy_dataset($@) {
        my (%opts) = validate(@_, {
                        test_client => {type=>SCALAR},
			mount => {type=>SCALAR},
			client => {type=>SCALAR | HASHREF }
                        }
                      );

        my ($test_client) = $opts{test_client};
	my ($mount) = $opts{mount};
	my ($client) = $opts{client};

	my $sudo = "/usr/local/bin/sudo";
	my $data_loc = $test_client . "/dense/integ/compression/comp_dataset/*" ;
	my $cmd = $sudo . " cp $data_loc $mount";

	say($client, "$cmd && echo _''OK_","_OK_",'-timeout', 40, '-fatal', "Couldn't cp $data_loc to $mount on client!");
}

=head2 return_expected_savings;

Function which defines a hash of the compression and dedupe savings. this will return
the expected compression/dedupe savings based on the datset populated in above function (create_dataset).

=cut

sub return_expected_savings () {
	my %saving_hash = (
		"expected_dedupe_saved_blocks" => 320,
		"expected_compress_saved_blocks" => 316,
		"expected_compress_dedupe_saved_blocks" => 84,
		"expected_dedupe_compress_saved_blocks" => 316,
	);

	my $return_hash = \%saving_hash;
	return $return_hash;
}

=head2 get_compression_saving (host=>$host, vol=>$vol, partner=>$partner)

Function to verify compression saving achieved by compreesion. In this function, $host, and $vol are mandatory parameters.
Rest of the parameters are optional.

=over 4

=item host

Filer HostSpec

=item vol_name

The volume on which compression saving has to be validated

Failure

=over 4

Return with an error message in C<$@>

=back

=back

=cut


sub get_compression_saving($@) {
        my (%opts) = validate(@_, { host => { type=>SCALAR|HASHREF },
                                   vol_name => { type=>SCALAR, optional=>1 },
				   partner => {type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/}}
                             );

	my ($vol_name) = $opts{vol_name};
        my $cfiler;
        if (ref($opts{host})) {
                $cfiler = $opts{host};
        } else {
                my $filer =$opts{host};
                $cfiler = connect("#rsh");
        }
        my $cmd = "df -S $vol_name" ;
	if ( $opts{partner} ) {
	    $cmd = "partner" . " $cmd";
	}

        my ($df_result, $df_exit__code) = filer_cli_say($cfiler,$cmd);

        my %df_result=();
        $df_result{result}=0;

        # Return failure if df command failed.
	return (%df_result)  if ((split(/\n/,$df_result))[0] !~ /^Filesystem\s+used\s+total-saved\s+\%total-saved\s+deduplicated\s+\%deduplicated\s+compressed\s+\%compressed/);
#        return (%df_result)  if ((split(/\n/,$df_result))[0] !~ /^Filesystem\s+used\s+compressed\s+a-sis\s+\%saved/);
        # Test Passed. Fill success bit (1) in $df_result{result} and populate %df_result
        $df_result{result}=1;

        foreach my $line (split(/\n/, $df_result))
        {
            if($line =~ /^\s*(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\%\s+(\d+)\s+(\d+)\%\s+(\d+)\s+(\d+)\%\s*/)
            {
                    $df_result{fs}->{$1}->{'used'}=$2;
                    $df_result{fs}->{$1}->{'total-saved'}=$3;
                    $df_result{fs}->{$1}->{'per-total-saved'}=$4;
                    $df_result{fs}->{$1}->{'deduplicated'}=$5;
		    $df_result{fs}->{$1}->{'per-deduplicated'}=$6;
		    $df_result{fs}->{$1}->{'compressed'}=$7;
		    $df_result{fs}->{$1}->{'per-compressed'}=$8;
            }
        }
        return %df_result;

}

=head2 compression_status_wait (host=>$host, vol=>$vol, partner=>$partner, required_state=>$required_state, timeout=>$timeout)

Function to wait for the completion of vol compress start command.In this function, $host, $vol , $required_state are mandatory parameters.
Rest of the parameters are optional.

=over 4

=item host

Filer HostSpec

=item vol

The volume which needs to be moved. Provide a valid volume name here.

=item required_state

The resquired state denotes the completion of compress start command.

Failure

=over 4

Return with an error message in C<$@>

=back

=back

=cut

sub compression_status_wait($@) {
        my (%opts) = validate(@_, {
                        conn => {type=>SCALAR | HASHREF},
                        vol => {type=>SCALAR},
                        partner => {type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/},
			required_state =>{type=>SCALAR},
			timeout => {type=>SCALAR, optional=>1, default=>DENSE_DEFAULT_TIMEOUT},
			time_interval => {type=>SCALAR, optional=>1}
                        }
                      );
	
        my ($conn) = $opts{conn};
        my ($vol) = $opts{vol};
	my $timeout= $opts{timeout};
	my $time_interval = $opts{time_interval};
	my $cmd;
	my $required_state = $opts{required_state};
	my $output_str;

	while(($output_str !~ /$required_state/) && ($timeout >0)){
		if($opts{partner}){
			$output_str = compression_status(conn=>$conn, vol=>$vol, required_state=>$required_state, partner=>1);
		} else {
			$output_str = compression_status(conn=>$conn, vol=>$vol, required_state=>$required_state);
		}

        sleep $time_interval;
        $timeout--;
	}

        if ($output_str !~ /$required_state/){
                die ("vol compress status did not return the completion status in time");
        } else {
		logcomment("vol compress status completed successfully");
                return $output_str;
        }
}

=head2 compression_status (host=>$host, vol=>$vol, required_state=>$required_state, partner=>$partner)

Function to execute vol compress status command. 

command.In this function, $host, $vol , $required_state are mandatory parameters.
Rest of the parameters are optional.

=over 4

=item host

Filer HostSpec

=item vol

The volume which needs to be moved. Provide a valid volume name here.

=item required_state

The resquired state denotes the completion of compress start command.

Failure

=over 4

Return with an error message in C<$@>

=back

=back

=cut

sub compression_status($@) {
	my (%opts) = validate(@_, {
                        conn => {type=>SCALAR | HASHREF},
                        vol => {type=>SCALAR},
                        partner => {type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/},
                        required_state =>{type=>SCALAR}
                        }
                      );

        my ($conn) = $opts{conn};
        my ($vol) = $opts{vol};
        my $cmd;
	my $required_state = $opts{required_state};

	$cmd = VOL_COMPRESS_STATUS . " $vol";
        if ( $opts{partner} ) {
            $cmd = "partner" . " " . VOL_COMPRESS_STATUS . " $vol";
        } 

        my ($return_text, $result_code) = filer_cli_say($conn, $cmd);
        my $comp_state = ((split(/\s+/,((split(/\n/,$return_text))[0])))[0]);

	if (($return_text =~ /$required_state/) || ($return_text =~ /$comp_state/) ) {
            return $return_text;
	} else {
	   die ("$return_text\n");
	}
}

=head2 add_compression_license (conn => $conn)

Function to add compression license

=over 4

=item conn

filer connection

Failure

=over 4

Return with an error message in C<$@>

=back

=back

=cut

sub add_compression_license($@){
        my (%opts) = validate(@_, {
                         conn=> {type=>SCALAR | HASHREF },
                         }
                     );

        my ($conn) = $opts{conn};
        my $license_code = "ELZRXFF";
        if(!(is_licensed($conn, 'compression'))){
        logcomment("Adding the compression license now" );
        license_raw($conn, $license_code, service=>'compression');
        }
        return ;
}



=head2 wait_for_cp (num_cp=>$num_cp, wait_time=>$wait_time)

Function to wait for specified number of CPs to occur

In this function, parameters are optional.

=over 4

=item num_cp

Number of CPs to wait.

=item wait_time

Extra time to wait, in seconds.

Failure

=over 4

Return with an error message in C<$@>

=back

=back

=cut

sub wait_for_cp {
        my (%opts) = validate(@_, {
			 num_cp    => { type => SCALAR, regex => qr/^\d+$/, optional => 1, default => 1 },
			 wait_time => { type => SCALAR, regex => qr/^\d+$/, optional => 1, default => C_DFLT_EXTRA_WAIT },
                            } );

        my $num_cp = $opts{num_cp};
        my $extra_wait = $opts{wait_time};

       logcomment( "Waiting for $num_cp CP(s) ..." );

        for (my $counter = 0; $counter < $num_cp; $counter++)
        {
	        sleep(ONE_CP_TIME);
        }
        if ($extra_wait > 0)
        {
                logcomment( "Waiting for $extra_wait extra second(s) to elapse..." );
                sleep($extra_wait);
        }
	return ;
}

1;
