# -*- perl -*-
## @summary 
##    Library routines for SIS-Vfiler access
##
## @usage
##    use Dense::dense_vfiler;
##
## @author rani@netapp.com
## @status in-progress
## @pod here

package Dense::dense_vfiler;

use strict;
use Utils;
use filer;
use Tharn;
use Tharn qw(connect host say logcomment logresult param tstlog);
use Volume qw (vol_list vol_status);
use VFILER::Vfiler;
use Dense qw(:all);
use Dense::Constant::Command;
use Dense::Constant::Message;
use Zapi;
use NaElement;
use Data::Dumper;
use SnapMirror;
use snapmirror::SMutils;
use Dense::storagemotion;

use Params::Validate qw(validate validate_pos OBJECT SCALAR HASHREF ARRAYREF UNDEF);

use base qw(Exporter);
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(
            sis_max_active_vfiler_ops
            dense_zapi_vfiler
            vfiler_dense_toggle
            vfiler_dense_get_config
            vfiler_dense_set_config
            vfiler_dense_undo
            vfiler_context_dense_undo
            vfiler_dense_stop
            vfiler_dense_check
            vfiler_dense_start
            vfiler_dense_status
            vfiler_dense_stat
            vfiler_storage_paths
	    dense_vfiler_migrate_start
	    dense_vfiler_migrate_prepare
	    dense_vfiler_status
    	   );

########################################################
# Function to set / modify sis.max_active_vfiler_ops
########################################################

=head2 sis_max_active_vfiler_ops(conn=> $conn, value=> $value, partner=> 1, validate=>$validate)
Set or modify sis.max_active_vfiler_ops if value is supplied, otherwise return current value

=over 4

=item conn

Filer Connection

=item value

Value for sis.max_vfiler_active_ops, optional

=item validate

Check if value is set correctly, optional, defaults to 1

=item partner

Set to 1 if command shd run in partner mode

=item Returns:

Returns Current value on success or dies on error

=back

=cut

sub sis_max_active_vfiler_ops($@)
{

    my (%opts) = validate(@_, {	conn => {type=>HASHREF},
            		value => {type=>SCALAR, optional=>1},
            		validate => {type=>SCALAR, default=>1},
            		partner => {type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/}
            	     }
            	);

    my $conn = $opts{conn};
    my $value = $opts{value};
    my $validate = $opts{validate};
    my $partner = $opts{partner};
    my $current_val;

    # Get the Filer name
    my $filer	= $conn->hostp()->hostname();

    my $DENSE_TEST_CMD = check_test_privilege('host' => $conn);

    my $cmd = "options " . SIS_MAX_ACTIVE_VFILER_OPS;

    if ( $partner ) {
        $cmd = "partner" . " " . $DENSE_TEST_CMD . "partner" . " " .  $cmd;
    }

    my ($return_text, $return_code);

    if ($value) {
    	## Add value to command
    	$cmd .=  " " . $value;
    	logcomment("Setting sis.max_vfiler_active_ops to: $value");
    	($return_text, $return_code) = filer_cli_say($conn, $cmd);
    	$cmd = "options " . SIS_MAX_ACTIVE_VFILER_OPS;
    }else {
    	$validate = 0;
    }

    ($return_text, $return_code) = filer_cli_say($conn, $cmd);
    if ($return_text =~ /sis.max_vfiler_active_ops\s+(\d+)/) {
    	$current_val = $1;
    }else {
    	die("Unexpected output: $return_text \n");
    }

    if($validate) {
    	($value == $current_val)?logcomment("Value set correctly to $value"): die("Current value for max vfiler ops is not set to the value supplied: $value\n");
    }

    return $current_val;
    	 
}

########################################################
# Function to run dense zapi commands on vfiler(s)
########################################################

=head2 dense_zapi_vfiler_run (conn=>$conn, vol=>$path_name, vfilertemplate=>$id, cmd=>$cmd, option=>$option, partner=>1)

Runs the dense zapi command on the vfiler(s) specified in the vfilertemplate
If more than one vfiler matches, the command will be run separately
for each vfiler.


Return value:

Returns the XML output from the ZAPI command, on success
Dies on failure

=cut

sub dense_zapi_vfiler($@)
{
    my (%opts) = validate (@_, {    filer   => {type=>HASHREF},
                                        vfiler  => {type=>SCALAR},
                                        user    => {type=>SCALAR, default=>"root"},
                                        password => {type=>SCALAR, optional=>1},
                                        cmd     => {type=>SCALAR},
                                        parameters     => {type=>HASHREF, optional=>1},
                                    }
                               );

    my $filer = $opts{filer};
    my $cmd = $opts{cmd};

    my $vfiler = $opts{vfiler};
    my $user = $opts{user};
    my $hashref = $opts{parameters};

    my $password = $opts{password};
    
    my %paramhash = %$hashref;

    if(!defined($opts{password})){
    	$password = $filer->{conn}->{hostp}->default_password();
    }
    
    # Create a new ZAPI object
    my $zapi = Zapi->new($filer->{conn}->{hostp}->default_ip(), $user, $password);

    # Set the vfiler where the zapi command is to be run
    $zapi->set_vfiler_server("$vfiler");

    # Build the ZAPI command
    my $elem = new NaElement("$cmd");

    my $args;;
    foreach my $key(keys %paramhash) {
    	$args = new NaElement($key,$paramhash{$key});
    	$elem->child_add($args);
    }

    ## Invoke the ZAPI call
    # ontapi -v vfiler1 172.17.181.56 root  sis-status path /vol/test_vol1
    my $res = $zapi->invoke_elem($elem,"passed");

    my $out = $res->toString();

    logcomment ($res->sprint_pretty());

    if($out =~ /results status=\"passed\"/i) {
   	return $out;
    } else {
    	die("ZAPI call for $cmd failed: $out\n");
    }

}

=head2 vfiler_storage_paths( host=>$host, vf_id=>$vfiler)

Get storage paths of a vfiler

=cut

sub vfiler_storage_paths($@)
{
   my (%opts) = validate(@_, { host => { type=>SCALAR },
                               vf_id => { type=>SCALAR } }
                         );

        my $host = $opts{host};
        my $id = $opts{vf_id};
        my $path;
	my @paths;
        my ($conn) = make_conn($host,"#rsh");

    logcomment("DEBUG: vfiler_Storage_paths()");
    my $pathstat = filer_cli_say($conn, "vfiler status -a $id");
    my @pathstat = split(/\r*\n/, $pathstat);
    for my $entry (@pathstat) {
	if( $entry =~ /Path:/ ){
 	    logcomment("Pattern matched entry : $entry");
	    ($path) = ($entry =~ m#Path: (\S+)#mg);
	    #my @path = split(/ /, $entry);
  	    logcomment("Pattern pushed into the array : $path");
            push( @paths, $path );
        }
    }
    return @paths;
}

=head2 dense_vfiler_status( conn=>$conn, vfiler=>$vfiler )

Returns the status of vfiler passed to it.

=cut

sub dense_vfiler_status($@)
{
    my (%opts) = validate(@_, { conn => { type=>HASHREF },
                               vfiler => { type=>SCALAR } }
			);

    my $conn = $opts{conn};
    my $vfiler = $opts{vfiler};
   
    my $vf_status = filer_cli_say( $conn, "vfiler status $vfiler" );
    my @vf_array = split(/\s{2,}/,$vf_status);
    logcomment("Vfiler status of $vfiler : $vf_array[1]");
    return $vf_array[1];
}


=head2 dense_vfiler_migrate_prepare( dst_host=>$dhost, vfiler=>$vfiler, dst_passwd=>$dpasswd, dst_login=>$dlogin, remote_host=>$rhost, remote_login=>$rlogin, remote_passwd=>$rpasswd, time_out=>$timeout);

Issues semi-sync to vfiler migration SM relationships, wait for it to get "In-sync", and then executes migrate_prepare.

=cut

sub dense_vfiler_migrate_prepare($@)
{
   my (%opts) = validate(@_, { dst_host => { type=>SCALAR },
                               vfiler => { type=>SCALAR },
                               remote_host => { type=>SCALAR },
                               remote_login => { type=>SCALAR },
                               remote_passwd => { type=>SCALAR },
                               dst_login => { type=>SCALAR },
                               dst_passwd => { type=>SCALAR },
                               time_out => { type=>SCALAR, optional=>1, default=>"3600" } }
			);
    my $dhost = $opts{dst_host};
    my $vfiler = $opts{vfiler};
    my $rhost = $opts{remote_host};
    my $rlogin = $opts{remote_login};
    my $rpasswd = $opts{remote_passwd};
    my $dlogin = $opts{dst_login};
    my $dpasswd = $opts{dst_passwd};
    my $timeout = $opts{time_out};

    my $dconn = make_conn( $dhost, "#rsh" );
    my $rconn = make_conn( $rhost, "#rsh" );

    ## Setting snapmirror.conf entry to semi-sync
    logcomment("Setting snapmirror schedule to semi-sync for all the volumes in migration vfiler");
    my @src_paths = Dense::dense_vfiler::vfiler_storage_paths(host=>$rhost, vf_id=>$vfiler);
    for my $entry (@src_paths) {
        my $volname = (split(/\//,$entry))[-1];
        my $src_path = $rhost. ":" .$volname;
        my $dst_path = $dhost. ":" .$volname;
        logcomment("Source : $src_path -> Destination : $dst_path");
        die("EVAL_ERROR=Could not set snapmirror.conf entry.")
            unless setConfEntry($dconn, $dst_path, $src_path, "semi-sync");
    }

    ## Wait for relationships in migration vfiler to get "in-sync"ed
    for my $entry (@src_paths) {
	my $sm_state;
        my $volname = (split(/\//,$entry))[-1];
        my $src_path = $rhost. ":" .$volname;
        my $dst_path = $dhost. ":" .$volname;
        logcomment("Source : $src_path -> Destination : $dst_path");
        my $maxwaittime = time() + $timeout;
        while ( (!($sm_state->{'Status'} =~ /In-sync/)) &&
            ((time() <= $maxwaittime) ))
        {
            sleep(10);
            $sm_state = snapmirror_status_detailed($dconn, $dst_path, $src_path);
        }
        die("Snapmirror in-sync timeout exceeded") if (time() > $maxwaittime);
    }

    ## Issue migrate prepare
    zapi_vfiler_migrate_prepare( $dhost, $dlogin, $dpasswd, remote_vfiler=>$vfiler, remote_filer=>$rhost, remote_login=>$rlogin, remote_passwd=>$rpasswd );

    ## Check for source vfiler status
    my $src_vf_status = dense_vfiler_status( conn=>$rconn, vfiler=>$vfiler );
    if( $src_vf_status !~ /migrate_prepare/ ){
	die "Vfiler migration did not get to prepare state";
    }

}

=head2 dense_vfiler_migrate_start( dst_host=>$dhost, vfiler=>$vfiler, remote_host=>$rhost, remote_login=>$login, remote_passwd=>$passwd, ifnames_ipaddrs_netmasks=>$if_to_use:$ip:$netmask, time_out=>$timeout );

Starts vfiler migrate start and waits for vfiler volumes to go to 'Snapmirrored' state

=cut

sub dense_vfiler_migrate_start($@)
{
   my (%opts) = validate(@_, { dst_host => { type=>SCALAR },
                               vfiler => { type=>SCALAR },
			       remote_host => { type=>SCALAR },
			       remote_login => { type=>SCALAR },
			       remote_passwd => { type=>SCALAR },
			       ifnames => { type=>SCALAR },
			       ipaddrs => { type=>SCALAR },
			       netmasks => { type=>SCALAR },
			       time_out => { type=>SCALAR, optional=>1, default=>"3600" } }
                         );
    my $dhost = $opts{dst_host};
    my $vfiler = $opts{vfiler};
    my $rhost = $opts{remote_host};
    my $rlogin = $opts{remote_login};
    my $rpasswd = $opts{remote_passwd};
    my $ifnames = $opts{ifnames};
    my $ipaddrs = $opts{ipaddrs};
    my $netmasks = $opts{netmasks};
    my $timeout = $opts{time_out};
    my $sm_status;

    my ($dconn) = make_conn($dhost, "#rsh");
    my $dst_hostspec = host($dhost);

    ## Calling vfiler migrate start
    logcomment("vfiler_migrate_start()");
    vfiler_migrate_start( $dconn, $vfiler, $rhost, remote_login=>$rlogin, remote_passwd=>$rpasswd, ifnames_ipaddrs_netmasks=>"$ifnames:$ipaddrs:$netmasks" );

    ## Retrieve the volume list of vfiler
    my @src_paths = Dense::dense_vfiler::vfiler_storage_paths(host=>$rhost, vf_id=>$vfiler);
    for my $entry (@src_paths) {
	my $volname = (split(/\//,$entry))[-1];
        my $src_path = $rhost. ":" .$volname;
	my $dst_path = $dhost. ":" .$volname;
	logcomment("Source : $src_path -> Destination : $dst_path");
	my $maxwaittime = time() + $timeout;
	while(1){
            if( time() < $maxwaittime) {
                $sm_status = SnapMirror::snapmirror_status(host=>$dst_hostspec, dst=>$dst_path, src=>$src_path);
                if ($sm_status->{'state'} eq 'snapmirrored' && $sm_status->{'status'} eq 'idle') {
                    last;
                } else {
                    sleep(10);
                }
            } else {
                die("Destination volume did not become Snapmirrored");
            }
        }
    }
    return 1;
}


 
=head2 vfiler_dense_toggle(host=>$host, vf_id=>$vfiler, path=>$path_name, state=>$state)

Turn dense on or off, depending on the value passed in.

=over 4

=item host

Filer HostSpec

=item vf_id

Vfiler id - can be individual vfiler or *

=item path

complete path  - /vol/volname

=item state

String, either on or off

=item Returns:

Returns Hash - key is vfiler id, value is the status message returned by the command

=back

=cut

sub vfiler_dense_toggle($@)
{
   my (%opts) = validate(@_, { host => { type=>SCALAR },
                               path => { type=>SCALAR },
                               vf_id => { type=>SCALAR },
                               state => { type=>SCALAR } }
                         );

        my $host = $opts{host};
        my $path = $opts{path};
        my $state = $opts{state};
        my $id = $opts{vf_id};


    my ($conn) = make_conn($host,"#telnet");
    ## my $DENSE_TEST_CMD = check_test_privilege('host' => $conn);

    # Construct Command
    my $cmd;

    if ($state =~ "on") {
    	$cmd = DENSE_ON_NAME . "$path";
    } else {
    	$cmd = DENSE_OFF_NAME . "$path";
    }

    # Pass Command and options to vfiler_run
    my @ret = vfiler_run($conn, $id, $cmd);

    my @out = map ( split(/\r*\n/,$_), @ret);

    my (%status, $vfiler);	

    foreach my $line(@out) {
    	next if($line =~ /^\s*$/);
    	next if($line =~ /output/);
    	next if($line =~ /vfiler run/);
    	if($line =~ /===== (vfiler(\d+))/) {
            $vfiler = $1;
    	}else {
            $status{$vfiler} = $line;
    	}
    }

    return %status;

}

=head2 vfiler_dense_get_config(host=>$host, vf_id=>$vfiler, path=>$path_name)

Get the schedule entries and mbs, if applicable for each vfiler or the vfiler id supplied

=over 4

=item host

Filer HostSpec

=item vf_id

Vfiler id - can be individual vfiler or *

=item path

complete path  - /vol/volname

=item Returns:

Returns Hash of Hash - key is vfiler id, then path
If Error, then $config{$key}{Error} gives error message

=back

=cut

sub vfiler_dense_get_config($@)
{
   my (%opts) = validate(@_, { host => { type=>SCALAR },
                               vf_id => { type=>SCALAR },
                               path => { type=>SCALAR, optional=>1 } }
                         );

    my $host = $opts{host};
    my $path = $opts{path};
    my $id = $opts{vf_id};


    my ($conn) = make_conn($host,"#rsh");

    # Construct Command
    my $cmd = DENSE_CONFIG_NAME; 

    if($path) {
    	# Add path_name
    	$cmd = $cmd . " $path";
    }

    # Pass Command and options to vfiler_run
    my @ret = vfiler_run($conn, $id, $cmd);

    my @out = map ( split(/\r*\n/,$_), @ret);

    my (%config, $vfiler);
    my %vfiler_config;
    my $Err = GetMessage( code => "DENSE_CONFIG_NOT_FOUND", host => $host);
    foreach my $line (@out) {
        next if($line =~ /^\s*$/);
        next if($line =~ /(\d+):(\d+):(\d+) GMT \[(\S+):\s(\S+)\]/);
        next if($line =~ /output/);
        next if($line =~ /vfiler run/);
        if ($line =~ /===== (vfiler(\d+))/) {
            $vfiler = $1;
        } elsif($line=~ /$Err/i) {
            $config{$vfiler}{Error} = $line;
        } else {
            push(@{$vfiler_config{$vfiler}}, $line);
        }

    }


    foreach my $key (keys %vfiler_config) {

        my @process = @{$vfiler_config{$key}};
        if (!$config{$key}{Error})
        {
            # Process return value
            my %temp = Dense::Status::dense_parse_config(config_ref=>\@process);
            %{$config{$key}} = %temp;
        }

    }
    return %config;
}

=head2 vfiler_dense_set_config(host=>$host, vf_id=>$vfiler, path=>$path_name, schedule=>$sched, mbs=>$mbs)

Get the schedule entries and mbs, if applicable for each vfiler or the vfiler id supplied

=over 4

=item host

Filer HostSpec

=item vf_id

Vfiler id - can be individual vfiler or *

=item path

complete path  - /vol/volname

=item Schedule

schedule string, Optional, default is "-" - no schedule
 
=item mbs

Mn block share value 

=item Returns:

Returns 1 on success, Error message on failure

=back

=cut

sub vfiler_dense_set_config($@)
{
   my (%opts) = validate(@_, { host => { type=>SCALAR },
                               path => { type=>SCALAR },
                               vf_id => { type=>SCALAR },
                               schedule => { type=>SCALAR, default=>"-", optional=>1 },
                               mbs => { type=>SCALAR, default=>"1", optional=>1 } }
                         );

    my $host = $opts{host};
    my $path = $opts{path};
    my $schedule = $opts{schedule};
    my $mbs = $opts{mbs};
    my $id = $opts{vf_id};


    my ($conn) = make_conn($host,"#telnet");

    my $DENSE_TEST_CMD;
    $DENSE_TEST_CMD = check_test_privilege('host' => $conn);
    if($DENSE_TEST_CMD) {
    	filer_cli_priv($conn, "test");
    } else {
    	logcomment("Warning: Test mode only available on debug builds. Some schedule strings might not work in non debug builds");
    }

    # Construct Command
    my $cmd = DENSE_CONFIG_NAME; 

    if ( defined($opts{schedule}) ) {
    	$cmd = $cmd . " -s " . $schedule;
    } 

    if ( defined($opts{$mbs}) ) {
    	$cmd = $cmd . " -m " . $mbs;
    } 

    if($path) {
    	# Add path_name
    	$cmd = $cmd . " $path";
    }

    # Pass Command and options to vfiler_run
    my @ret = vfiler_run($conn, $id, $cmd);

    my @out = map ( split(/\r*\n/,$_), @ret);

    my ($error, $return_text);
    foreach my $line (@out) {
        next if($line =~ /^\s*$/);
    	next if($line =~ /output/);
    	next if($line =~ /vfiler run/);
    	if ($line =~ /===== vfiler(\d+)/) {
            next;
    	}elsif ($line) {
            $return_text = $line;
            $error = 1;
    	}else {
            die("Unknown error while setting sis schedule or MBS value for $path in vfiler $id\n");
    	}
            
    }

    if($error) {
    	return $return_text;
    } else {
    	return 1;
    }

}

=head2 vfiler_dense_undo(host=>$host, vf_id=>$vfiler, path=>$path_name, force=>$value)

Run sis undo on specified volume and vfiler

=over 4

=item host

Filer HostSpec

=item vf_id

Vfiler id - can be individual vfiler or *

=item path

complete path  - /vol/volname

=item force

force sis undo to run even when there is not enough space, test mode only

=item Returns:

Returns command output

=back

=cut

sub vfiler_dense_undo ($@) {

   my (%opts) = validate(@_, { host => { type=>SCALAR },
                               path => { type=>SCALAR },
                               vf_id => { type=>SCALAR },
                               force => { type=>SCALAR, default=>"0", optional=>1 } }
                         );


    my $host = $opts{host};
    my $path = $opts{path};
    my $id = $opts{vf_id};


    my ($conn) = make_conn($host,"#telnet");

    my $DENSE_TEST_CMD = check_test_privilege('host' => $conn);

    # Construct Command
    my $cmd = DENSE_UNDO_NAME . " $path";

    if($opts{force}) {
        if($DENSE_TEST_CMD) {
            filer_cli_priv($conn, "test");
            $cmd = $cmd . " " . "-f";
        } else {
            logcomment("WARNING: Cannot use -f flag with sis undo on a non-debug build");
        }
    }else {
    	filer_cli_priv($conn, "test");
    }

    # switch filer to vfiler context
    my @ret = vfiler_run($conn, $id, $cmd);

    my @out = map ( split(/\r*\n/,$_), @ret);

    my $return_text;
    foreach my $line (@out) {
        next if($line =~ /^\s*$/);
    	next if($line =~ /(\d+):(\d+):(\d+) GMT \[(\S+):\s(\S+)\]/);
        next if($line =~ /output/);
        next if($line =~ /vfiler run/);
        if ($line =~ /===== vfiler(\d+)/) {
            next;
        } elsif ($line) {
            $return_text .= $line;	
    	}
    }

    return $return_text;

}

=head2 vfiler_context_dense_undo(host=>$host, vf_id=>$vfiler, path=>$path_name, force=>$value)

Run sis undo on specified volume and vfiler

=over 4

=item host

Filer HostSpec

=item vf_id

Vfiler id - can be individual vfiler or *

=item path

complete path  - /vol/volname

=item force

force sis undo to run even when there is not enough space, test mode only

=item Returns:

Returns command output on success, dies on failure

=back

=cut

sub vfiler_context_dense_undo ($@) {

   my (%opts) = validate(@_, { host => { type=>SCALAR },
                               path => { type=>SCALAR },
                   force => ( type=>SCALAR, optional=>1 ),
                               vf_id => { type=>SCALAR } }
                         );

    my $host = $opts{host};
    my $path = $opts{path};
    my $id = $opts{vf_id};


    my ($conn) = make_conn($host,"#telnet");

    my $DENSE_TEST_CMD = check_test_privilege('host' => $conn);

    # Construct Command
    my $cmd = DENSE_UNDO_NAME . " $path";

    if($opts{force}) {
    	if($DENSE_TEST_CMD) {
            filer_cli_priv($conn, "test");
            $cmd = $cmd . " " . "-f";
    	} else {
            logcomment("WARNING: Cannot use -f flag with sis undo on a non-debug build");
    	}
    }else {
    	# Set Priv level on filer  - Need test priv for -f option
    	filer_cli_priv($conn, "diag");
    }

    # switch filer to vfiler context
    my $ret = vfiler_context($conn, $id);

    # Run dense undo command 
    my ($return_text, $result_code) = filer_cli_say($conn, $cmd);

    # check result
    if( $result_code eq cmdOkReturn ){
    	return $return_text;
    }else{
    	die("$return_text\n");
    }

}

=head2 vfiler_dense_stop(host=>$host, vf_id=>$vfiler, path=>$path_name)

Run sis undo on specified volume and vfiler

=over 4

=item host

Filer HostSpec

=item vf_id

Vfiler id - can be individual vfiler or *

=item path

complete path  - /vol/volname

=item Returns:

Returns command output

=back

=cut

sub vfiler_dense_stop ($@) {

   my (%opts) = validate(@_, { host => { type=>SCALAR },
                               path => { type=>SCALAR },
                               vf_id => { type=>SCALAR } }
                         );

    my $host = $opts{host};
    my $path = $opts{path};
    my $id = $opts{vf_id};


    my ($conn) = make_conn($host,"#rsh");

    # Construct Command
    my $cmd = DENSE_STOP_NAME;

    # Add path_name
    $cmd = $cmd . " $path";

    # Pass Command and options to vfiler_run
    my @ret = vfiler_run($conn, $id, $cmd);

    my @out = map ( split(/\r*\n/,$_), @ret);

    my $return_text;
    foreach my $line (@out) {
        next if($line =~ /^\s*$/);
        next if($line =~ /output/);
        next if($line =~ /vfiler run/);
        if ($line =~ /===== vfiler(\d+)/) {
            next;
        }else {
            $return_text .= $line;
    	}
    }

    return $return_text;
}

=head2 vfiler_dense_undo(host=>$host, vf_id=>$vfiler, path=>$path_name)

Run sis undo on specified volume and vfiler

=over 4

=item host

Filer HostSpec

=item vf_id

Vfiler id - can be individual vfiler or *

=item path

complete path  - /vol/volname

=item Returns:

Returns command output

=back

=cut

sub vfiler_dense_check ($@) {

    my (%opts) = validate(@_, { host => { type=>SCALAR },
                               path => { type=>SCALAR },
                               vf_id => { type=>SCALAR } }
                         );

    my $host = $opts{host};
    my $path = $opts{path};
    my $id = $opts{vf_id};


    my ($conn) = make_conn($host,"#telnet");

    # Construct Command
    my $cmd = DENSE_CHECK_NAME;

    # Add path_name
    $cmd = $cmd . " $path";

    # Pass Command and options to vfiler_run
    my @ret = vfiler_run($conn, $id, $cmd);

    my @out = map ( split(/\r*\n/,$_), @ret);

    my ($error, $return_text);
    foreach my $line (@out) {
        next if($line =~ /^\s*$/);
        next if($line =~ /output/);
        next if($line =~ /vfiler run/);
        if ($line =~ /===== vfiler(\d+)/) {
            next;
        }else {
            $return_text .= $line;
        }
    }

    return $return_text;
}

=head2 vfiler_dense_start(host=>$host, vf_id=>$vfiler, path=>$path_name, reuse_chkpt=>1, scan=>$scan, del_chkpt=>$del_chkpt)

Run sis start on specified volume and vfiler

=over 4

=item host

Filer HostSpec

=item vf_id

Vfiler id - can be individual vfiler or *

=item path

complete path  - /vol/volname

=item scan

Runs sis start -sf on the volume

=item reuse_chkpt

Runs sis start -sfp on the volume, forces gatherer to reuse existing chkpt

=item del_chkpt

Runs sis start -d or sis start -sfd on the volume, forces sis to discard existing chpt

=item Returns:

Returns command output

=back

=cut

sub vfiler_dense_start ($@) {

    my (%opts) = validate(@_, { host => { type=>SCALAR },
                               path => { type=>SCALAR },
                               vf_id => { type=>SCALAR },
                               reuse_chkpt => { type=>SCALAR, optional=>1 },
                               scan => { type=>SCALAR, optional=>1 },
                               del_chkpt => { type=>SCALAR, optional=>1 } }
                         );

    my $host = $opts{host};
    my $path = $opts{path};
    my $id = $opts{vf_id};


    my ($conn) = make_conn($host,"#rsh");

    my $cmd = DENSE_START_NAME;

    if ( $opts{scan} ) {
    	$cmd = $cmd . " -s -f ";
    } 
    if ( $opts{scan} && $opts{reuse_chkpt} ) {
    	$cmd = $cmd . " -p " ;
    }
    if ( ($opts{del_chkpt}) && (!$opts{scan}) ) {
    	$cmd = $cmd . " -d -f " ;
    }

    # Add path
    $cmd = $cmd . " " . $path;

    # Pass Command and options to vfiler_run
    my @ret = vfiler_run($conn, $id, $cmd);

    my @out = map ( split(/\r*\n/,$_), @ret);

    my ($error, $return_text);
    foreach my $line (@out) {
        next if($line =~ /^\s*$/);
        next if($line =~ /output/);
        next if($line =~ /vfiler run/);
        if ($line =~ /===== vfiler(\d+)/) {
            next;
        }else {
            $return_text .= $line;
        }
    }

    return $return_text;

}

=head2 vfiler_dense_status(host=>$host, vf_id=>$vfiler, path=>$path_name, long=>$long, test_mode=>$test)

Get status entries for all vfilers/volumes or the ones specified by the parameters

=over 4

=item host

Filer HostSpec

=item vf_id

Vfiler id - can be individual vfiler or *

=item path

complete path  - /vol/volname

=item long

Long output - sis status -l, optional

=item test_mode

Run command in test mode - longer output with more fields, optional

=item Returns:

Returns Hash of Hash - key is vfiler id, then path
If Error, then $status{$key}{Error} gives error message

=back

=cut

sub vfiler_dense_status ($@) {

    my (%opts) = validate(@_, { host => { type=>SCALAR },
                               vf_id => { type=>SCALAR },
                               path => { type=>SCALAR, optional=>1 },
                               conn_type => { type=>SCALAR, default=>"#telnet" },
                               long => { type=>SCALAR, default=>1 },
                               test_mode => { type=>SCALAR, default=>1 } }
                         );

    my $host = $opts{host};
    my $path = $opts{path};
    my $id = $opts{vf_id};
    my $conn_type = $opts{conn_type};

    my ($conn) = make_conn($host, $conn_type);

    # Construct Command
    my $cmd = DENSE_STATUS_NAME;

    if( $opts{long} ) {
    	$cmd = $cmd . " " . "-l";
    }

    my $DENSE_TEST_CMD;
    if( $opts{test_mode} ) {
    	$DENSE_TEST_CMD = check_test_privilege('host' => $conn);
    	if($DENSE_TEST_CMD) {
            filer_cli_priv($conn, "test");
    	} else {
            logcomment("Warning: Test mode only available on debug builds");
    	}
    }

    if($path) {
    	# Add path_name
    	$cmd = $cmd . " $path";
    }

    # Pass Command and options to vfiler_run
    my @ret = vfiler_run($conn, $id, $cmd);

    my @out = map ( split(/\r*\n/,$_), @ret);

    my %status;

    my %vfiler_status;
    my $vfiler;
    my $Err = GetMessage( code => "DENSE_STATUS_NOT_FOUND", host => $host);
    foreach my $line (@out) {
        next if($line =~ /^\s*$/);
    	next if($line =~ /(\d+):(\d+):(\d+) GMT \[(\S+):\s(\S+)\]/);
        next if($line =~ /output/);
        next if($line =~ /vfiler run/);
        if ($line =~ /===== (vfiler(\d+))/) {
            $vfiler = $1;
        } elsif($line=~ /$Err/) {
            $status{$vfiler}{Error} = $line;	
    	} else {
            push(@{$vfiler_status{$vfiler}}, $line);
        }

    }

    foreach my $key (keys %vfiler_status) {
    	my @process = @{$vfiler_status{$key}};
    	if (!$status{$key}{Error})
    	{
            # Process return value
            if ( $opts{long} == 0 ){
            	%{$status{$key}} = Dense::Status::dense_parse_status(status_array=>\@process);
            }else{
            	%{$status{$key}} = Dense::Status::dense_parse_status_detail(status_array=>\@process);
            }
    	}
    }
    return %status;

}

=head2 vfiler_dense_stat(host=>$host, vf_id=>$vfiler, path=>$path_name, long=>$long, test_mode=>$test)

Get the sis stat entries for all vfilers and volumes, or the ones specified

=over 4

=item host

Filer HostSpec

=item vf_id

Vfiler id - can be individual vfiler or *

=item path

complete path  - /vol/volname

=item long

Long output - sis stat -l, optional

=item test_mode

Run command in test mode - longer output with more fields, optional

=item Returns:

Returns Hash of Hash - key is vfiler id, then path
If Error, then $stat{$key}{Error} gives error message

=back

=cut

sub vfiler_dense_stat ($@) {

    my (%opts) = validate(@_, { host => { type=>SCALAR },
                               vf_id => { type=>SCALAR },
                               path => { type=>SCALAR, optional=>1 },
                               long => { type=>SCALAR, default=>1 },
                               test_mode => { type=>SCALAR, default=>1 } }
                         );

    my $host = $opts{host};
    my $path = $opts{path};
    my $id = $opts{vf_id};

    my ($conn) = make_conn($host,"#telnet");

    # Construct Command
    my $cmd = DENSE_STAT_NAME;

    if( $opts{long} ) {
    	$cmd = $cmd . " " . "-l";
    }

    my $DENSE_TEST_CMD;
    if( $opts{test_mode} ) {
    	$DENSE_TEST_CMD = check_test_privilege('host' => $conn);
    	if($DENSE_TEST_CMD) {
            filer_cli_priv($conn, "test");
    	} else {
            logcomment("Warning: Test mode only available on debug builds");
    	}
    }

    if($path) {
    	# Add path_name
    	$cmd = $cmd . " $path";
    }

    # Pass Command and options to vfiler_run
    my @ret = vfiler_run($conn, $id, $cmd);

    my @out = map ( split(/\r*\n/,$_), @ret);

    my %stat;

    my %vfiler_stat;
    my $vfiler;
    my $Err = GetMessage( code => "DENSE_STAT_NOT_FOUND", host => $host);
    foreach my $line (@out) {
        next if($line =~ /^\s*$/);
    	next if($line =~ /(\d+):(\d+):(\d+) GMT \[(\S+):\s(\S+)\]/);
        next if($line =~ /output/);
        next if($line =~ /vfiler run/);
        if ($line =~ /===== (vfiler(\d+))/) {
            $vfiler = $1;
        } elsif($line=~ /$Err/ ) {
            $stat{$vfiler}{Error} = $line;	
    	} else {
            push(@{$vfiler_stat{$vfiler}}, $line);
        }
    }

    foreach my $key (keys %vfiler_stat) {
    	my @process = @{$vfiler_stat{$key}};
    	if (!$stat{$key}{Error})
    	{
            # Process return value
            if ( $opts{long} == 0 ){
            	%{$stat{$key}} = Dense::Status::dense_parse_stat(stat_array=>\@process);
            }else{
            	%{$stat{$key}} = Dense::Status::dense_parse_stat_detail(stat_array=>\@process);
            }
    	}
    }
    return %stat;

}

1;
