# -*- perl -*-
## @summary Filer/Dense Fiji routines for dense volume
## @usage
##      use Dense::fiji;
## @author abhayjit@netapp.com
## @cf     false
## @status public
## @pod here

=head1 NAME

Fiji - Dense Fiji Points

=head1 DESCRIPTION

This module defines Fiji routines for SIS volumes.

=head1 FUNCTIONS

Fiji funtions

=cut


package Dense::fiji;
BEGIN{push(@INC,"/usr/local/src/perl-modules-5.6.1/Params-Validate-0.72/lib")};
use strict;
use Params::Validate qw(validate validate_pos OBJECT SCALAR HASHREF ARRAYREF UNDEF);
use Tharn qw(connect host say logcomment logresult param tstlog);
use filer qw(filer_cli_test_say);
use Dense::Clone qw(dense_clone_fiji);
use Dense qw(check_test_privilege);
use Dense::Utils;
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(
		    fiji_dense_stage 
		    fiji_clear
		    set_dense_clone_fiji
			set_wp1m_delay_fiji
		);
our @EXPORT_OK = qw(
		    fiji_dense_stage 
		    fiji_clear
		    set_dense_clone_fiji
			set_wp1m_delay_fiji
		   );

use constant DENSE_STAGE_NOP => "0";
use constant DENSE_STAGE_LOGGING => "1";
use constant DENSE_STAGE_GATHERING => "2";
use constant DENSE_STAGE_SORTING => "3";
use constant DENSE_STAGE_DEDUPING => "4";
use constant DENSE_STAGE_DEDUPING_P1 => "4";
use constant DENSE_STAGE_DEDUPING_P2 => "5";
use constant DENSE_STAGE_CHECKING => "6";
use constant DENSE_STAGE_CHECKING_P1 => "7";
use constant DENSE_STAGE_CHECKING_P2 => "8";
use constant DENSE_STAGE_DONE => "9";

my @dense_fiji_stages = qw(
                           DENSE_STAGE_NOP
                           DENSE_STAGE_LOGGING
                           DENSE_STAGE_WOC_SORTING
                           DENSE_STAGE_WOC_COMPRESSING
                           DENSE_STAGE_GATHERING
                           DENSE_STAGE_SORTING
                           DENSE_STAGE_DEDUPING_P1
                           DENSE_STAGE_DEDUPING_P2
                           DENSE_STAGE_DEDUPING_SHARING
                           DENSE_STAGE_CHECKING
                           DENSE_STAGE_CHECKING_P0
                           DENSE_STAGE_CHECKING_P1
                           DENSE_STAGE_CHECKING_P2
                           DENSE_STAGE_DONE
                           );
use Dense::Constant::Message qw($DENSE_ONTAP_VERSION getwv);

=head2 fiji_dense_stage(host=>$host, vol_name=>$vol_name, dense_stage=>$dense_stage, wait=>$wait)

Insert the fiji point after the dense_stage is complete.

=over 4

=item host

Filer HostSpec (Required)

=item vol_name

Volume name (Required)

=item dense_stage

Stage of dense operation (Required)

=item wait

Optional argument which specifies how may secs you want to sleep, defaults to 10

=item Returns:

Success

=over 4

Returns 1

=back

Failure

=over 4

Return with an error message in $return_text

=back

=back

=cut

sub fiji_dense_stage {
    my (%opts) = validate(@_, { host => { type=>SCALAR },
                                vol_name => { type=>SCALAR },
                                dense_stage => { type=>SCALAR,regex => qr/^[A-Z_]+\d?$/ },
                                wait => { type=>SCALAR, default=>"10", optional=>1 },
			       	            repeat => { type=>SCALAR, default=>"1", optional=>1 },
			       	            clear => { type => SCALAR, default => "1", optional => 1 },
			    }
                         );

    my ( $host, $vol_name, $wait, $cfiler, $dense_stage, $repeat );
    my ( $cmd, $return_val, $return_text, $result_code );
    $host     = $opts{host};
    $vol_name = $opts{vol_name};
    $wait     = $opts{wait};
    $repeat   = $opts{repeat};
    $cfiler   = connect("$host#rsh");
    
    if(!is_debug(conn => $cfiler)) {
       logcomment("fiji can not be set on promoted build, hence returning");
       return 0;
    }
    $DENSE_ONTAP_VERSION = getwv(host=>$host);
    logcomment("DENSE_ONTAP_VERSION is $DENSE_ONTAP_VERSION");
    if (($DENSE_ONTAP_VERSION ne 'GB') && ($DENSE_ONTAP_VERSION ne 'IC') && ($DENSE_ONTAP_VERSION ne 'BR')) {
        $dense_stage = $opts{dense_stage};
        if ( !grep( /$dense_stage/, @dense_fiji_stages ) ) {
            logcomment("stage not recognized, hence can not set fiji");
            return 0;
        }
        $cmd = "fiji filter \"(repeat $repeat (dense.sleep.stage (&& (eq \@sleep_stage $dense_stage) (eq \@vol $vol_name)) (return $wait)))\"";
    } else {
        $dense_stage = eval $opts{dense_stage};
        $cmd = "fiji filter \"(repeat $repeat (dense.status.stage (&& (eq \@dense_stage $dense_stage) (eq \@vol $vol_name)) (sleep $wait)))\"";
    }
    
    if($opts{clear}) {
        $result_code = filer_cli_test_say( $cfiler, "fiji clear all" );
        if ( $result_code !~ /fiji: filter cleared/ ) { 
            logcomment("Unable to clear the fiji filters set earlier" );
        }
    }
    
    ( $return_text, $result_code ) = filer_cli_test_say( $cfiler, $cmd );
    logcomment("fiji set output: $return_text");
    if ( $return_text =~ /fiji: filter set/i ) {
        logcomment("fiji set sucessfully");
        $return_val = 1;
    } else {
        logcomment("fiji not set");
        $return_val = 0;
    }

    return $return_val;
}

sub fiji_clear ($$) {

    my ($conn,$slot) = @_;
    my ($result, $output);
    $result = 1;
    if (($slot !~ /all/) && ($slot !~ /\d+/)) {
	logcomment("Invalid input to clear fiji");
    	$result = 0;
    	return $result;
    }
    $output = filer_cli_test_say($conn, "fiji clear ".$slot);
    if ( $output !~ /fiji: filter cleared/ ) { 
    logcomment("Unable to clear the fiji filter on ".$slot);
    $result = 0;
    }
    
    return $result;

}

############################################################################ 
############# Functions to set clone fiji's  ###############################
############################################################################ 
=head2 set_dense_clone_fiji 
 
set clone fiji on the filer
 
Inputs:
1. conn - filer connection 
2. volname - volume name
3. sleep (optional) - fiji sleep time 
4. repeat (optional) - the number of times fiji needs to be repeated
5. action - action/type of clone fiji
 
Return value:
Success: 1
Failure: 0
 
=cut

sub set_dense_clone_fiji ($@) {
    my (%opts) = validate (@_, {conn	    => {type=>SCALAR | HASHREF},
				volname     => {type=>SCALAR},
				sleep	    => {type=>SCALAR,optional=>1,default=>30,regex=> qr/^[1-9]\d*$/},
				repeat	    => {type=>SCALAR,optional=>1,default=>1,regex=> qr/^[1-9]\d*$/},
				action	    => {type=>SCALAR},
			       }
			  );
    my ($clone_fiji, $ret_val, $return_text);

    $ret_val = 0;
    $clone_fiji = dense_clone_fiji (volname=>$opts{'volname'}, sleep=>$opts{'sleep'}, action=>$opts{'action'}, repeat=>$opts{'repeat'});
    $return_text = filer_cli_test_say($opts{'conn'}, $clone_fiji);
    logcomment("fiji set output: $return_text");

    if($return_text =~ /fiji: filter set/i) {
	logcomment("fiji set sucessfully"); 
	$ret_val = 1;
    } else {
	logcomment("fiji not set");
	$ret_val = 0;
    }

    return $ret_val;

}


############################################################################
#############   Functions to set wp1m fiji   ###############################
############################################################################
=head2 set_wp1m_delay_fiji

set wafl process one message delay on the filer

Inputs:
1. conn - filer connection
2. volname - volume name
3. sleep (optional) - fiji sleep time
4. msg_type - message type on which fiji should be set 

Return value:
Success: 1
Failure: 0

=cut


sub set_wp1m_delay_fiji {
  my (%opts) = validate(@_, { 	host 		=> { type=>SCALAR },
								vol_name     => { type=>SCALAR},
                                msg_type 	=> { type=>SCALAR },
				repeat		=> { type=>SCALAR, optional=>'1', default=>'1' },
								sleep       => { type=>SCALAR,optional=>1,default=>30,regex=> qr/^[1-9]\d*$/ },
                			}
                    );



    my ($host) = $opts{host};
	my ($vol_name) = $opts{vol_name};
	my ($sleep) = $opts{sleep};
    my ($msg_type) = $opts{msg_type};
    host($host);
    my ($conn) = connect("#rsh");
    my $cmd = "fiji filter \"(repeat $opts{repeat} (wafl.wp1m.delay.mdp (&& (eq \@vol $vol_name) (eq \@type $msg_type)) (return $sleep)))\"";
    my $return_val;
    my ($return_text, $result_code) = filer_cli_test_say($conn, $cmd);

    logcomment("fiji set output: $return_text");
    if($return_text =~ /fiji: filter set/i) {
    logcomment("fiji set sucessfully");
    $return_val = 1;
    } else {
    logcomment("fiji not set");
    $return_val = 0;
    }

    return $return_val;
}
1;
