# -*- perl -*-
#  $Id: $
#
## @summary
##    Library routines for NDVM with dense
#### @author dl-dense-qa@netapp.com
## @author sbhat@netapp.com, dubey@netapp.com
## @status in-progress
## @pod here
=head1 NAME

NDVM - Basic routines for ndvm volume move functionality

=head1 DESCRIPTION

This module defines routines for vol move.
=head1 FUNCTIONS 

NDVM basic functions

=cut

package Dense::ndvm;
use strict;
use SPARSE::Util::Logger;
use filer;
use Burt;
use Tharn qw(connect host say logcomment logresult param tstlog addresult fatal_result);
use Utils qw(make_conn);
use Data::Dumper qw(Dumper);
use Dense qw(:all);
use Aggregate qw(:all);
use Dense::Status;
use Dense::Constant::Command;
use Dense::Constant::Message;
use Filer::Version;
use Params::Validate qw(validate validate_pos OBJECT SCALAR BOOLEAN HASHREF ARRAYREF UNDEF);
use FSR::FSRutils;
use Date::Manip;
use Time::ParseDate;
use snaplock qw(system_time_set system_time_get);
use VFILER::Vfiler;
use Volume qw(volume_options_list_info);

BEGIN {
Params::Validate::validation_options(
    on_fail => sub {
        my @msgs = @_;
        chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))
        }
   );
}

use base qw(Exporter);

BEGIN {
	my @action = qw (
			 vol_move_start
			 vol_move_cutover
			 vol_move_resume
			 vol_move_abort
			 vol_move_pause
			 );
	my @status = qw (
			 vol_move_status
			 vol_move_status_wait
			);
	my @verify = qw (
			 vol_move_verify
			);
	our %EXPORT_TAGS = ( action => [ @action],
			     status => [ @status],
			     verify => [ @verify]
			   );
	$EXPORT_TAGS{'all'} = [ @action, @status, @verify];
	Exporter::export_ok_tags('all');
}

=head2 vol_move_start(host=>$host, src_vol=>$src_vol, dst_aggr=>$dst_aggr, retain_src=>$retain_src,num_cutover_times=>$num_cutover_times, cutover_window=>$cutover_window, partner=>$partner)

Function to do vol move to dst_aggr. In this function, $host, $src_vol and $dst_aggr are mandatory parameters.
Rest of the parameters are optional.

=over 4

=item host

Filer HostSpec

=item src_vol

The volume which needs to be moved. Provide a valid volume name here.

=item dst_aggr

Destination Aggr name: Provide a valid aggr name to which the src_vol needs to be moved.

=back

Failure

=over 4

Return with an error message in C<$@>

=back

=back

=cut
sub vol_move_start($@) {
	my (%opts) = validate(@_, {
                        host => {type=>SCALAR},
                        src_vol => {type=>SCALAR},
			dst_aggr => {type=>SCALAR},
			retain_src => {type=>SCALAR, default=>0, optional=>1},
			no_cutover => {type=>SCALAR, default=>0, optional=>1},
			cutover_window => {type=>SCALAR, default=>0, optional=>1},
			num_cutover_times => {type=>SCALAR, optional=>1},		
                        cmd_opt => {type=>SCALAR, optional=>1},
                        partner => {type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/}
                        }
                      );

        my ($host) = $opts{host};
        my ($src_vol) = $opts{src_vol};
	my ($dst_aggr) = $opts{dst_aggr};
        my ($cmd_opt) = $opts{cmd_opt};
	my ($retain_src) = $opts{retain_src};
	my ($no_cutover) = $opts{no_cutover};
	my ($num_cutover_times) = $opts{num_cutover_times};
        my ($conn) = make_conn($host,"#rsh");	
	my $cmd;
	
	if( $opts{retain_src} && $opts{no_cutover} ) {
        die("Cannot use both retain_src and no_cutover options at the same time\n"); #retaining source option is for retaining src after cutover
    	}

        my $options;
        if( $opts{retain_src} ) {
            $options =  " -k ";
	}    elsif( $opts{no_cutover} ) {
                $options = " -m ";
            } elsif( $opts{cutover_window} ) {
                $options =  " -w ";
            } elsif( $opts{num_cutover_times} ){
            $options =  " -r ";            
	    }
        else {
                $options = " ";
          }

	if ( $opts{partner} ) {
		filer_cli_say($conn, "priv set -q diag; partner exportfs -z /vol/$src_vol ; sync", 300);
        	$cmd = "partner" . " " . "partner" . " " .  "vol move start" . " $src_vol". " $dst_aggr" .  " $options";
    	} else {
		filer_cli_say($conn, "priv set -q diag; exportfs -z /vol/$src_vol ; sync", 300);
        	$cmd =  "vol move start" . " $src_vol" . " $dst_aggr" . " $options";
    	  }

    	my ($return_text, $result_code) = filer_cli_say($conn, $cmd);
	if (remove_leading_trailing_spaces($return_text)) {
		die("$return_text\n");
	}
        return $return_text;
}

=head2 vol_move_cutover(host=>$host, src_vol=>$src_vol, dst_aggr=>$dst_aggr, retain_src=>$retain_src,num_cutover_times=>$num_cutover_times, cutover_window=>$cutover_window, partner=>$partner)

Function to do vol move to dst_aggr. In this function, $host, $src_vol and $dst_aggr are mandatory parameters.
Rest of the parameters are optional.

=over 4

=item host

Filer HostSpec

=item src_vol

The volume which is being moved. Provide a valid volume name here.

=item dst_aggr

=back

Failure

=over 4

Return with an error message in C<$@>

=back

=back

=cut

sub vol_move_cutover(){
	 my (%opts) = validate(@_, {
                        host => {type=>SCALAR},
                        src_vol => {type=>SCALAR},
                        dst_aggr => {type=>SCALAR},
                        retain_src => {type=>SCALAR, optional=>1},
                        cutover_window => {type=>SCALAR, optional=>1},
			num_cutover_times => {type=>SCALAR, optional=>1},
                        partner => {type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/}
                        }
                      );

        my ($host) = $opts{host};
        my ($src_vol) = $opts{src_vol};
        my ($dst_aggr) = $opts{dst_aggr};
        my ($retain_src) = $opts{retain_src};
	my ($cutover_window) = $opts{cutover_window};
	my ($num_cutover_times) = $opts{num_cutover_times};
	my ($conn) = make_conn($host,"#rsh");
	my $cmd;


        my $options;
        if( $opts{retain_src} ) {
            $options = " -k ";
	}	
        if( $opts{cutover_window} ) {
                $options = " -w ";
            }
        if( $opts{num_cutover_times} ){
            $options = " -r ";            
        } else {
                $options = " ";
          }
	if ( $opts{partner} ) {
		filer_cli_say($conn, "priv set -q diag; exportfs -z /vol/$src_vol ; sync", 300);
                $cmd = "partner" . " " . "partner" . " " .  "vol move cutover" . " $src_vol". " $dst_aggr" .  " $options";
        } else {
		filer_cli_say($conn, "priv set -q diag; exportfs -z /vol/$src_vol ; sync", 300);
                $cmd = "vol move cutover" . " $src_vol" . " $dst_aggr" . " $options";
          }

        my ($return_text, $result_code) = filer_cli_say($conn, $cmd);
	if( $result_code eq cmdOkReturn){
        	return $return_text;
	}else{
        die("$return_text\n");
    }
}

=head2 vol_move_verify(host=>$host, src_vol=>$src_vol1_name , dst_aggr=>$dense_aggr_dst);

Function to verify the volumes in the source or destination aggregate after ndvm . In this function, $host, $src_vol and $src_aggr or $dst_aggr 
are mandatory parameters.

=over 4

=item host

Filer HostSpec

=item src_vol

The volume which needs to be moved. Provide a valid volume name here.

=item dst_aggr

Destination Aggr name: Provide a valid aggr name to which the src_vol needs to be moved.

=back

Failure

=over 4

Return with an error message in C<$@>

=back

=back

=cut

sub vol_move_verify($@) {
        my (%opts) = validate(@_, {
                        host => {type=>SCALAR},
                        vol_name => {type=>SCALAR},
			aggr_name => {type=>SCALAR},
                        dst_flag => {type=>SCALAR, default=>1, optional =>1},
			no_cutover_flag => {type=>SCALAR, default=>0, optional =>1}
                        }
                      );

        my ($host) = $opts{host};
        my ($vol_name) = $opts{vol_name};
        my ($aggr_name) = $opts{aggr_name};
	my ($dst_flag) = $opts{dst_flag};
	my ($no_cutover_flag) = $opts{no_cutover_flag};	
        my ($conn) = make_conn($host,"#rsh");
	my $cmd;
	my $vol_match = 0;

        if ((($dst_flag ==1) && (!$no_cutover_flag)) || (($no_cutover_flag ==1) && ($dst_flag == 0))) {

        my $status_object = aggr_status($conn,$aggr_name);
        my $vols_list = $status_object->volumes();
                foreach my $vol_n (@{$vols_list}) {
                        if ($vol_n eq $vol_name) {
                                $vol_match = 1;
                        }
                }
                if ($vol_match == 1) {
                        logcomment ("Volume exist at the source aggregate after ndvm without cutover");
                } else {
                        die("Volume does not exist at the source aggr after ndvm without cutover");
                }
	} elsif (($no_cutover_flag ==1) && ($dst_flag == 1)) {

                my $status_object = aggr_status($conn,$aggr_name);
                my $vols_list = $status_object->volumes();
                my $vol_source = "ndm_dstvol";
                foreach my $vol_n (@{$vols_list}) {
                        if ($vol_n =~ /$vol_source/) {
                        $vol_match = 1;
                        }
                }
                if ($vol_match == 1) {
                        logcomment ("Volume exist at the destination aggregate after ndvm without cutover");
                } else {
                        die("Volume not retained after ndvm without cutover option");
                }
        } else {
                my $status_object = aggr_status($conn,$aggr_name);
                my $vols_list = $status_object->volumes();
                my $vol_source = "$vol_name" . "_old";
                foreach my $vol_n (@{$vols_list}) {
                        if ($vol_n =~ /$vol_source/) {
                        $vol_match = 1;
                        }
                }
                if ($vol_match == 1) {
                        logcomment ("Volume exist at the source aggregate");
                } else {
                        die("Volume not retained after ndvm with option -k ");
                }
        }
}		

=head2 vol_move_status(host=>$host, src_vol=>$src_vol1_name, state_pause=>1, partner=>$partner);

Function to verify the status of ndvm operation . In this function, $host, $src_vol are mandatory parameters. state_pause is optional 
parameter used to verify the status of paused operation.partner specifies if the command is to be run from the cfo context 

=over 4

=item host

Filer HostSpec

=item src_vol

The volume which needs to be moved. Provide a valid volume name here.

=item state_pause

To verify the status of paused operation

=back

Failure

=over 4

Return with an error message in C<$@>

=back

=back

=cut

sub vol_move_status($@){
         my (%opts) = validate(@_, {
                        host => {type=>SCALAR},
                        src_vol => {type=>SCALAR},
                        state_pause => {type=>SCALAR, optional=>1, default=>0},
                        partner => {type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/},
			required_state =>{type=>SCALAR}
                        }
                      );

        my $host = $opts{host};
        my $src_vol = $opts{src_vol};
        my $state_pause = $opts{state_pause};
        my $conn = make_conn($host,"#rsh");
        my $cmd;
		my $required_state = $opts{required_state};
        my $vol_move_state;
        my $options;
		my $NDVM_TEST_CMD = check_test_privilege('host' => $conn);
         if ( $opts{partner} ) {
                $cmd = "partner" . " " . $NDVM_TEST_CMD . "partner" . " " .  "vol move status " . $src_vol;
        } else {
                $cmd = "vol move status" . " $src_vol";
          }

        my ($return_text, $result_code) = filer_cli_say($conn,$cmd);
	if (($state_pause == 0) && ($required_state eq "setup" || $required_state eq "move" || $required_state eq "cutover")) {
                $vol_move_state = ((split(/\s+/,((split(/\n/,$return_text))[1])))[4]);
        } elsif (($state_pause == 0) && ($required_state =~ /vol move: Specified source volume is not being moved/)) {
                $vol_move_state = remove_leading_trailing_spaces($return_text);        
	  } 
	  else {
                $vol_move_state = ((split(/\s+/,((split(/\n/,$return_text))[1])))[5]);        
	  }        
	return $vol_move_state;
}


=head2 vol_move_resume(host=>$host, src_vol=>$src_vol, retain_src=>$retain_src,num_cutover_times=>$num_cutover_times, cutover_window=>$cutover_window, partner=>$partner)

Function to do vol move to dst_aggr. In this function, $host, $src_vol are mandatory parameters.
Rest of the parameters are optional.

=over 4

=item host

Filer HostSpec

=item src_vol

The volume which needs to be moved. Provide a valid volume name here.

=item dst_aggr

Destination Aggr name: Provide a valid aggr name to which the src_vol needs to be moved.

=back

Failure

=over 4

Return with an error message in C<$@>

=back

=back

=cut

sub vol_move_resume($@) {
         my (%opts) = validate(@_, {
                        host => {type=>SCALAR},
                        src_vol => {type=>SCALAR},
			retain_src => {type=>SCALAR, default=>0, optional=>1},
			no_cutover => {type=>SCALAR, default=>0, optional=>1},
			cutover_window => {type=>SCALAR, default=>0, optional=>1},
			num_cutover_times => {type=>SCALAR, optional=>1},
                        partner => {type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/}
                        }
                      );

        my ($host) = $opts{host};
        my ($src_vol) = $opts{src_vol};
	my ($retain_src) = $opts{retain_src};
	my ($no_cutover) = $opts{no_cutover};
        my ($num_cutover_times) = $opts{num_cutover_times};
        my ($conn) = make_conn($host,"#rsh");
        my $cmd;

        if( $opts{retain_src} && $opts{no_cutover} ) {
        die("Cannot use both retain_src and no_cutover options at the same time\n"); #retaining source option is for retaining src after cutover
        }

        my $options;
        if( $opts{retain_src} ) {
            $options =  " -k ";
            if( $opts{no_cutover} ) {
                $options = " -m ";
                }
            if( $opts{cutover_window} ) {
                $options =  " -w ";
            }
        if( $opts{num_cutover_times} ){
            $options =  " -r ";            }
        } else {
                $options = " ";
          }
        if ( $opts{partner} ) {
                $cmd = "partner" . " " . "partner" . " " .  "vol move resume" . " $src_vol".   " $options";
        } else {
                $cmd = "vol move resume" . " $src_vol" . " $options";
          }

        my ($return_text, $result_code) = filer_cli_say($conn, $cmd);
	if (remove_leading_trailing_spaces($return_text)) {
		die("$return_text\n");
	}
	return $return_text;
}


=head2 vol_move_status_wait(host=>$host, src_vol=>$src_vol1_name, state_pause=>1, required_state => $required_state, partner => $partner);

Function to wait for the vol move status . In this function, $host, $src_vol are mandatory parameters.
Rest of the parameters are optional.

=over 4

=item host

Filer HostSpec

=item src_vol

The volume which needs to be moved. Provide a valid volume name here.

=item state_pause

This is to decide whether the vol move state is paused

=item required_state

This is the state string passed for comparison

=item partner

Whether to be executed in cfo context or not 

=back

Failure

=over 4

Return with an error message in C<$@>

=back

=back

=cut

sub vol_move_status_wait($@) {
	my (%opts) = validate(@_, {
                        host => {type=>SCALAR},
                        src_vol => {type=>SCALAR},
                        state_pause => {type=>SCALAR, optional=>1, default=>0},
			no_cutover => {type=>SCALAR, optional=>1, default=>0},
                        partner => {type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/},
			timeout => {type=>SCALAR, optional=>1, default=>300},
			required_state =>{type=>SCALAR} 
                        }
                      );
	my $host = $opts{host};
        my $src_vol = $opts{src_vol};
        my $state_pause = $opts{state_pause};
	my $required_state = $opts{required_state};
        my $cmd;
        my $options;

	my $timeout= $opts{timeout};
	my $output_str;
        # It takes some time to start/resume the vol move. Hence we will sleep a bit before
        # we actually go and start polling for the status. 
        if ($required_state eq "complete"){
                $required_state= "vol move: Specified source volume is not being moved";
        }
        while(($output_str !~ /$required_state/i) && ($timeout >0)){
		if($opts{partner}){
			    $output_str = vol_move_status(host=>$host, src_vol=>$src_vol, state_pause=>$state_pause, required_state=>$required_state,partner=>1);
		}else{
			    $output_str = vol_move_status(host=>$host, src_vol=>$src_vol, state_pause=>$state_pause, required_state=>$required_state);
		}
        sleep 2;
        $timeout--;
	}
					
        if ($output_str !~ /$required_state/i){
                die ("vol move status did not return paused status in time");
        }else{
                return $output_str;
        }

}

sub vol_move_abort($@) {
	my (%opts) = validate(@_, {
			host => {type=>SCALAR},
			src_vol => {type=>SCALAR},
			partner => {type=>SCALAR, optional=>1, default=>0, regex=>qr/^(0|1)$/}
			}
		     );
		my $host = $opts{host};
        my $src_vol = $opts{src_vol};
        my $conn = make_conn($host,"#rsh");
        my $cmd;
		my $NDVM_TEST_CMD = check_test_privilege('host' => $conn);
        if ( $opts{partner} ) {
                $cmd = "partner" . " " . "partner" . " " .  "vol move abort" . " $src_vol";
        } else {
                $cmd = "vol move abort" . " $src_vol";
        }

        my ($return_text, $result_code) = filer_cli_say($conn, $cmd);
        if (remove_leading_trailing_spaces($return_text)) {
                die("$return_text\n");
        }
        
        
}
		
1;
