# -*- perl -*-
#  $Id: //depot/prod/test/main/lib/DevUnitTest.pm#1 $

## @summary Library of Developer Unit Test commands
##
## @description
##    This package defines routines which simplify the creation of Developer
##  unit test scripts.  This will include subroutines to do differing levels of
##  setup/cleanup in a standard way, and many routines to execute filer commands
##  with some post command checking.
##
## @author dl-hk@netapp.com (Serviceablitiy and Volume Operations)
## @cf     false
## @status public

package DevUnitTest;

# Export all public library API's by adding them to this list. This will place 
# them in the scope that they are imported to the test. (Probably ::)
our @ISA = qw(Exporter);
our @EXPORT = qw(clearTraceBuffer 
                 copyTestVol
                 createAggregate
                 createDataSnapshots
                 createDPMirrorSet
                 createEmptySnapshots
                 createLSMirrorSet
                 createVolume 
                 deleteDPRelationship
                 deleteLSRelationship
                 deleteVolumes
                 disableFault
                 enableFault
                 fillVolWithData
                 getClusterId
                 getSnapshotInfoList
                 getSnapshotNameList
                 getSnapshotProperties
                 makeSnapmirrorPath
                 moveTestVol
                 numStripedAggrMem
                 pickAggregates
                 pickMemberAggr
                 pickMemberVol
                 pickVServer
                 searchMgwdLog
                 setPrivilege
                 showTraceBuffer
                 snapshotPromote
                 updateDPMirror
                 updateLSSet
                 volExists
                 volStatus
                 vreport_show
                 vrepair_fix
                 );

use SayWrap;
use Tharn;
use strict;

#use VolCmds qw(volExists);

use TCD;
use Server;
use TracePkg;

use Data::Dumper;

my $Filer;     # Server object pointer
my $Test;      # TCD object pointer
my $trace;     # TracePkg object pointer
my $volPrefix; # String to start the volume names with for this test.
my @Nodes;     # List of nodes in the cluster
my $node;      # A single node, useful sometimes

my $DP_LICENSE_ADDED = 0; # flag to limit multiple additions of license codes.

# @name ::init
# @usage DevUnitTest::init($Filer,$Test,$trace,$volPrefix) 
#
# @Synopsis
#  Initializes commonly needed values for this library. No return value.
# 
# Arguments: Takes several values that provide the needed information for all
#  of the library functions below.
#    $Filer - A Server object.  Use 'new Server()' to instantiate.
#    $Test  - A TCD object. Use 'new TCD(-testcases => \@Test_Cases)' to instantiate.
#    $trace - A TracePkg object. Use 'new TracePkg()' to instantiate.
#    $volPrefix - This is a string name that should uniquely identify this test. It will
#                 be used mostly as a prefix for volume names, but also any other created
#                 resource may use it (i.e. Aggregates).
# 
# This is not EXPORTED. So use the package scope to call it.
sub init 
{
   $Filer = shift;
   $Test = shift;
   $trace = shift;
   $volPrefix = shift;
   @Nodes = $Filer->servers();
   $node = $Nodes[0];
}

# @name ::clearTraceBuffer(\%)
# @usage $result = clearTraceBuffer()
# 
# @synopsis
#  This function will clear a specified mtrace buffer for a process on one or 
# all nodes. 
# 
# Arguments: Takes a hash pointer containing the arguments for the 'debug 
#   mtrace buffer clear' ngsh command. node will default to the list of all 
#   nodes in the cluster ( as set in the ::init() function).  
# 
#ngsh> debug mtrace buffer clear ?
#   [-node] <nodename>           *Node
#   [-process] <process_name>    *Process Name
#   [-buffer] <trace_ring_name>  *Buffer Name
#
#  ngsh options: 
#    process, buffer
# 
#  options with defaults:
#    nodes                   - defaults to global @Nodes
#
#  extra options:
#    nodes - an array of nodes to use of this command
# 
# Returns:
#   If clear succeeds returns $TCD::PASS
#   If clear fails returns $TCD::FAIL
sub clearTraceBuffer 
{
   my $args = shift;
   my $proc = $args->{process};
   my $buf = $args->{buffer};
   my @nodes = (defined $args->{nodes}) ? $args->{nodes} : @Nodes;

   foreach my $node (@nodes) {
      eval {
         $Test->nlog("debug mtrace buffer clear -node $node -process $proc -buffer $buf");
   
         #debug mtrace buffer clear -node * -process mgwd -buffer vopl
         my $output = $Filer->debug_mtrace_buffer_clear(process=>$proc, 
                                                        buffer=>$buf,
                                                        node=>$node);
      };
      if ($@) {
         $Test->nlog("Failed to clear the mtrace<$node:$proc:$buf> buffer: ($@)");
         return $TCD::FAIL;
      }
   }
   return $TCD::PASS;
}

#
# @name ::copyTestVol
# @usage $result = copyTestVol({vserver=>$vserver,
#                               volume=>$volume,
#                               'destination-volume'=>$destVol,
#                               'destination-aggregate'=>$destAggr,
#                               'expectResuts'=>$string})
#
# @synopsis  Creates a copy of a volume to a new name and aggr. Takes a hash of 
#    named options. All options should match the options to the 'volume copy' 
#    ngsh command. Given an 'expectResult' value the command will fail if the
#    completion string does not match.
# 
# volume copy ?
#  [ -vserver <vserver name> ]                 Virtual Server Name (default: vs0)
#   [-volume] <volume name>                    Volume Name
#   [-destination-volume] <volume name>        Destination Volume
#   [-destination-aggregate] <aggregate name>  Destination Aggregate
#  [[-destination-stripe-format] {data|full}]  *Destination Stripe Format
#  [ -foreground [{true|false}] ]              Foreground Process (default: false)
#
#  ngsh options: 
#     vserver, volume, 'destination-volume', 'destination-aggregate',
#     'destination-stripe-format', forground
# 
#  options with defaults:
#    vserver                 - defaults to global symbol ${vserver}
#    volume                  - defaults to global symbol ${volPrefix}
#    'destination-volume'    - defaults to "$args->{volume}_cp"
#    'destination-aggregate' - defaults to global symbol ${dstaggr}
#
#  extra options:
#     expectResults => "string"
#        This option searches the returned progress string (in case of an error)
#        for a match to this "string".  If a match is not found, this function
#        will return $TCD::FAIL
# 
# @result
#   If copy succeeds and expectResult is empty-string, returns $TCD::PASS
#   If copy fails, and expectResult is found in the job status, returns $TCD::PASS
#   If copy succeeds, but expectResult is non-empty, returns $TCD::FAIL
#   If copy fails, and expectResult is empty-string, returns $TCD::FAIL
sub copyTestVol
{
   my $args = shift;
   my $return = $TCD::PASS;

   my @required_args = qw(vserver destination-aggregate destination-volume);
   foreach my $req (@required_args) {
     if (!defined $args->{$req}) {
       $Test->nlog("Call to DevUnitTest::copyTestVol is missing argument '$req'");
       $return = $TCD::SCRIPT;
     }
   }
   return $return unless ($return == $TCD::PASS); #don't return till the whole loop has finished.

   $args->{volume} = $volPrefix unless (defined $args->{volume}); # preset the main default

   my $copy_options = [
      vserver                 => $args->{vserver},
      volume                  => $args->{volume},
      'destination-volume'    => $args->{'destination-volume'},
      'destination-aggregate' => $args->{'destination-aggregate'},
      foreground              => 'true',
      -Timeout                => 600
   ];
   my $expectResult = defined $args->{expect} ? $args->{expect} : undef;

   $Test->nlog("Attempting to copy $args->{volume} to $args->{'destination-volume'}." . 
               (($expectResult) ? " Expect result /$expectResult/" :
                                 "(should succeed)"));

   eval { $Filer->volume_copy(@$copy_options); };

   $Test->nlog("copy result:");
   $Test->nlog( ($@) ? Dumper($@) : "  SUCCESS"); 

   $return = $TCD::PASS;
   if ($@) {
       if (defined($args->{error})) {
           ${$args->{error}} = $@;
       }
       if (!$expectResult || !($@ =~ m/$expectResult/)) {
         $Test->nlog("FAILED to copy the volume $args->{volume}");
         $return = $TCD::FAIL;
      } else {
         $Test->nlog("Failed as expected: Result matches /$expectResult/");
      }
  } elsif ($expectResult) {
      $Test->nlog("Volume copy unexpectedly succeeded: Result did not match /$expectResult/");
      $return = $TCD::FAIL;
  }
   return $return;
}

# @name ::createAggregate
# @usage $result = createAggregate({aggregate=>$aggr,
#                                   diskcount=>$diskcount})
# 
# @synopsis
#  Creates an aggregate. If diskcount is given it will be a striped aggregate.
#@change 08/05/2010 talluri added a for loop to address aggr create timeout issues on vsims.
sub createAggregate 
{
   my $args = shift;

   return $TCD::SCRIPT unless defined $args->{'aggregate'};
   my $create_args = [
      aggregate => $args->{'aggregate'}
   ];
   if (defined $args->{'nodes'}){
	   push( @$create_args, 'nodes' => $args->{'nodes'});
   }
   
   if (defined $args->{'diskcount'}) {
      push( @$create_args, 'diskcount' => $args->{'diskcount'} );
      push( @$create_args, 'volume-style' => 'striped' );
      push( @$create_args, 'voltype' => 'striped' );
   } else {
      push( @$create_args, 'diskcount' => '3');
      push( @$create_args, 'volume-style' => 'flex' );
      push( @$create_args, 'voltype' => 'flex' );
   }
   my $No_entry_msg = "no entries";
   my $Num_retries = 12;
   ## Create the aggr
   $Test->nlog("Creating Aggregate: $args->{'aggregate'}");
   eval {
      my $cmd = $Filer->storage_aggregate_create(@$create_args);
   };
   if($@){
      if ($@ =~ /$No_entry_msg/i) {
         for (my $i = 0; $i < $Num_retries; $i++) {
             sleep 10;
               # This will continue to fail until the aggr is visible
               # Retry some number of times with delay between shows.
                eval {
                    $Filer->storage_aggregate_show(aggregate => "$args->{'aggregate'}");
                };
                if ($@) {
                    next if $@ =~ /$No_entry_msg/i;
                    # Not the expected error.
                    $Test->nlog("Aggr show failed");
                    return $TCD::FAIL;
                } else { 
				        last;
                }
            }
        } else {
            # Otherwise error while creating aggr.
            $Test->nlog("Failed to create aggregate $args->{'aggregate'}': ($@)");
            return $TCD::FAIL;
        }
   }
   return $TCD::PASS;
}

# @name ::createDPMirrorSet
# @usage
#  $return = createDPMirrorSet({
#     source=>$sourceVolume,
#     vserver=>$vserver,
#     'aggregate'=>$srcAggr, 
#     'stripe-count'=>$stripeCount,
#     mirrors=>({'destination-volume'=>$dataProtectionFlexVolume,
#                'destination-aggregate'=>$flexAggr},
#               {'destination-volume'=>$dataProtectionFlexVolume2,
#                'destination-aggregate'=>$flexAggr})
#      });
# 
# @synopsis
#   Creates a DP Source volume and any number of DP Destination volumes
#   The 'mirrors' argument should point to an array of hashes with each array
#   element representing a new destination volume to be created.
# 
# Uses the following commands:
#  volume create -vserver $vserver -volume $src -aggregate $aggr -size $size 
#  license add FEDPEKPJYBFDAA
#  volume create -vserver $vserver -volume $dp -aggregate $aggr -type DP -size $size
#  snapmirror create -destination-path $destPath -source-path $srcPath -vserver $vserver 
#  snapmirror initialize -destination-path $destPath -source-path $srcPath -foreground true
#  snapmirror update -destination-path $destPath -foreground true
#
# Returns $TCD::FAIL immediately if any command fails.
# 
sub createDPMirrorSet
{
   my $args = shift;
   my $return = $TCD::PASS;

   my @required_args = qw(source vserver aggregate mirrors);
   foreach my $req (@required_args) {
     if (!defined $args->{$req}) {
       $Test->nlog("Call to DevUnitTest::createDPMirrorSet is missing argument '$req'");
       $return = $TCD::SCRIPT;
     }
   }
   return $return unless ($return == $TCD::PASS);
   
   my $createArgs = {volume=>$args->{'source'},
                     vserver=>$args->{'vserver'},
                     aggregate=>$args->{'aggregate'}};
   $createArgs->{'size'} = (defined $args->{'size'}) ? $args->{'size'} : '100MB';

   if (defined $args->{'stripe-count'}) {
      $createArgs->{'stripe-count'} = $args->{'stripe-count'};

      # This is a temporary work around to allow the DP mirror to get created 
      # in Boilermaker.0 even though it was officially disabled for the release.
      my $cmd = "set -privilege diag; set -confirmations off; node run -node * -command \"priv set test; setflag paloma_allow_striped_replication 1\"";
      $Test->nlog(" Enabling paloma striped replication.\n  $cmd");
      $Test->nlog( $Filer->shell(cmd=>"ngsh -c '$cmd'") );
   }
   $createArgs->{'size'} = (defined $args->{'size'}) ? $args->{'size'} : '500MB';

   # Create the DP Source volume
   $return = createVolume($createArgs) unless volExists($createArgs->{volume});
   return $return unless ($return == $TCD::PASS);

   if( $DP_LICENSE_ADDED == 0 ) {
      $Test->nlog("Adding the DP License code : NUHYZZGKWVMADF");
      eval { $Filer->license_add('license-code'=>'NUHYZZGKWVMADF'); };# Add the DP license#change:talluri modified the dp license code from 'FEDPEKPJYBFDAA'  to 'NUHYZZGKWVMADF'
      $DP_LICENSE_ADDED++ unless( $@ );
   }

   # make a snapmirror PATH
   my $clusterId = getClusterId();
   my $sourcePath = makeSnapmirrorPath($args->{'source'},
                                       $args->{'vserver'},
                                       $clusterId);

   foreach my $mirror ($args->{mirrors}) {

      # Create a DP destination volume
      $createArgs->{'volume'} = $mirror->{'destination-volume'};
      $createArgs->{'aggregate'} = $mirror->{'destination-aggregate'};
      $createArgs->{'type'} = 'DP';
      $createArgs->{'size'} = $mirror->{'size'} if (defined $args->{'size'}); #else default to last given size.

      $return = createVolume($createArgs);
      return $return unless ($return == $TCD::PASS);

      my $destPath = makeSnapmirrorPath($createArgs->{'volume'},
                                        $createArgs->{'vserver'},
                                        $clusterId);

      # Create the Snapmirror relationship
      $Test->nlog("Creating the dp relationship for '$destPath'");
      eval {
         $Filer->snapmirror_create('destination-path'=>$destPath,
                                   'source-path'=>$sourcePath,
                                   'vserver'=>$createArgs->{'vserver'});
      };
      if ($@) {
         $Test->nlog("Failed to create DP relationship for '$destPath': ($@)");
         return $TCD::FAIL;
      }

      # Initialize the DP destination
      $Test->nlog("Initializing the dp relationship for '$destPath'");
      eval {
         $Filer->snapmirror_initialize('destination-path'=>$destPath,
                                       'source-path'=>$sourcePath,
                                       'vserver'=>$createArgs->{'vserver'},
                                       'foreground'=>'true');
      };
      if ($@) {
         $Test->nlog("Failed to initialize the DP relationship for '$destPath': ($@)");
         return $TCD::FAIL;
      }

      # Perform an update on the DP destination
      $return = updateDPMirror($destPath, $sourcePath, $createArgs->{'vserver'});
      last if ($return != $TCD::PASS);
   }
   return $return;
}

#TODO: implement an equivalent function for striped volumes
#only works on flex vols
# Creates given number of snapshots on the given volume.
# Each snapshot created is of given size. If no size is specified then each 
# snapshot is 100MB in size
sub createDataSnapshots
{

    my $args = shift;
    my $return = $TCD::PASS;

    my @required_args = qw(vserver volume);
    foreach my $req (@required_args) {
     if (!defined $args->{$req}) {
       $Test->nlog("Call to DevUnitTest::createDataSnapshots is missing argument '$req'");
       $return = $TCD::SCRIPT;
     }
    }

    return $return unless ($return == $TCD::PASS); #don't return till the whole loop has finished.

    my $numOfSnaps = 1;
    my $size = "100M";
    if($args->{numOfSnaps} && $args->{numOfSnaps} > 0) {
        $numOfSnaps = $args->{numOfSnaps};
    }

    if($args->{size}) {
        $size = $args->{size};
    }

    my $vserver = $args->{vserver};
    my $volume = $args->{volume};

    my $node = getNodeForFlexVol($vserver,$volume);

    return $node if($node == $TCD::SCRIPT);

    for(my $i=1; $i <= $numOfSnaps; $i++) {

        my $filename = "test123.txt";

        fillVolWithDataHelper({node=>$node,size=>$size,volume=>$volume,filename=>$filename});

        $Test->nlog("Creating snapshot:'snap $i' on volume:$volume ");
        eval {
           $Filer->volume_snapshot_create(vserver=>$vserver,
                                        volume=>$volume,
                                        snapshot=>"devut_data_snap_$i",
                                        foreground=>"true",
                                        comment=>'"snapshot created by DevUnitTest"');
        };
        if ($@) {
           $Test->nlog("Failed to create snapshot 'snap $i' on volume '$volume': ($@)");
           return $TCD::FAIL;
        }

        removeFileFromVol({node=>$node,volume=>$volume,filename=>$filename});

    }

    return $TCD::PASS;
 
}

# @name ::createEmptySnapshots
# @usage $return = createEmptySnapshots($vserver,$vol,$numOfSnaps)
# 
# @synopsis
#   Creates $numOfSnaps snapshot copies on volume $vol
# 
# Returns $TCD::FAIL immediately if a 'snapshot create' command fails.
# 
sub createEmptySnapshots 
{
   my ($vserver,$vol,$numOfSnaps) = @_;
   return $TCD::SCRIPT unless (defined $vserver && 
                               defined $vol &&
                               defined $numOfSnaps );
   $numOfSnaps = 5 unless defined $numOfSnaps;

   for(my $i=1; $i <= $numOfSnaps; $i++) {
      $Test->nlog("Creating snapshot:'snap $i' on volume:$vol ");
      eval {
         $Filer->volume_snapshot_create(vserver=>$vserver,
                                        volume=>$vol,
                                        snapshot=>"devut_snap_$i",
                                        foreground=>"true",
                                        comment=>'"I am a snapshot"');
      };
      if ($@) {
         $Test->nlog("Failed to create snapshot 'snap $i' on volume '$vol': ($@)");
         return $TCD::FAIL;
      }
   }
   return $TCD::PASS;
}

# @name ::createLSMirrorSet
# @usage
#  $return = createLSMirrorSet({
#     source=>$sourceVolume,
#     vserver=>$vserver,
#     'aggregate'=>$srcAggr,
#     mirrors=>({'destination-volume'=>$loadSharingFlexVolume,
#                'destination-aggregate'=>$flexAggr},
#               {'destination-volume'=>$loadSharingFlexVolume2,
#                'destination-aggregate'=>$flexAggr})
#      });
# 
# @synopsis
#   Creates a LS Source volume and any number of LS Destination volumes
#   The 'mirrors' argument should point to an array of hashes with each array
#   element representing a new destination volume to be created.
#
# Uses the following commands: 
#  volume create -vserver $vserver -volume $ls -aggregate $aggr -type LS -size $size -load-sharing-source $src
#  snapmirror create -source-path $srcPath -vserver $vserver
#  snapmirror initialize -source-path $srcPath
#  snapmirror update -source-path $srcPath
#
# Returns $TCD::FAIL immediately if any command fails.
# 
sub createLSMirrorSet
{
   my $args = shift;
   my $return = $TCD::PASS;

   my @required_args = qw(source vserver aggregate mirrors);
   foreach my $req (@required_args) {
     if (!defined $args->{$req}) {
       $Test->nlog("Call to DevUnitTest::createLSMirrorSet is missing argument '$req'");
       $return = $TCD::SCRIPT;
     }
   }
   return $return unless ($return == $TCD::PASS);

   my $createArgs = {volume=>$args->{'source'},
                     vserver=>$args->{'vserver'},
                     aggregate=>$args->{'aggregate'}};
   $createArgs->{'stripe-count'} = $args->{'stripe-count'} if (defined $args->{'stripe-count'});
   $createArgs->{'size'} = (defined $args->{'size'}) ? $args->{'size'} : '500MB';

   # Create the LS Source volume
   $return = createVolume($createArgs) unless volExists($createArgs->{volume});
   return $return unless ($return == $TCD::PASS);

   # make a snapmirror PATH
   my $clusterId = getClusterId();
   my $sourcePath = makeSnapmirrorPath($args->{'source'},
                                       $args->{'vserver'},
                                       $clusterId);

   foreach my $mirror ($args->{mirrors}) {

      # Create a LS destination volume
      $createArgs->{'volume'} = $mirror->{'destination-volume'};
      $createArgs->{'aggregate'} = $mirror->{'destination-aggregate'};
      $createArgs->{'type'} = 'DP';
      $createArgs->{'size'} = $mirror->{'size'} if (defined $args->{'size'}); #else default to last given size.
      #$createArgs->{'load-sharing-source'} = $args->{'source'};

      $return = createVolume($createArgs);
      return $return unless ($return == $TCD::PASS);

      my $destPath = makeSnapmirrorPath($createArgs->{'volume'},
                                        $createArgs->{'vserver'},
                                        $clusterId);

      my $cmd = "set -privilege diagnostic; set -confirmations off; snapmirror show";
      #$Test->nlog(" Using 'raw' command to promote snapshot '$snap' for '$volume'.\n  $cmd");
      $Test->nlog( $Filer->shell(cmd=>"ngsh -c \"$cmd\"") );


      # Create the LS set
      #snapmirror create -source-path qdeux-11234462095://vs0/snapshot_dut_SRC -destination-path qdeux-11234462095://vs0/snapshot_dut_SRC_LS_2 -type LS
      $Test->nlog("Creating a LS relationship for '$destPath' ");
      eval {
         $cmd = "set -privilege diagnostic; set -confirmations off; snapmirror create -type LS -source-path $sourcePath -destination-path $destPath";
         $Test->nlog( $Filer->shell(cmd=>"ngsh -c \"$cmd\"") );

         #$Filer->snapmirror_create('source-path'=>$sourcePath,
         #                          'destination-path'=>$destPath,
         #                          'type'=>'LS');
      };
      if ($@) {
         $Test->nlog("Failed to create the LS relationship for SRC:'$sourcePath' DST:'$destPath': ($@)");
         return $TCD::FAIL;
      }
   }

   return updateLSSet($sourcePath);
}

# @name ::createVolume
# @usage $result = createVolume(\%args)
# 
# @synopsis
# 
#> vol create 
#  (volume create)
#Usage:
#  [-volume] <volume name>                                     Volume Name
#  [-aggregate] <aggregate name>                               Aggregate Name
# [[-size] <size>]                                             Volume Size
# [ -state {online|restricted|offline|force-online} ]          Volume State (default: online)
# [ -type {RW|LS|DP|TMP} ]                                     Volume Type (default: RW)
# [ -policy <text> ]                                           Export Policy (default: default)
# [ -user <user name> ]                                        User ID
# [ -group <group name> ]                                      Group ID
# [ -security-style {unix|ntfs|mixed} ]                        Security Style
# [ -unix-permissions <unix perm> ]                            Unix Permissions (default: ---rwx------)
# [ -junction-path <junction path> ]                           Junction Path
# [ -junction-active [{true|false}] ]                          *Junction Active (default: true)
# [ -vsroot [{true|false}] ]                                   *Virtual Server Root Volume (default: false)
# [ -comment <text> ]                                          Comment
# [ -maxdir-size <size> ]                                      *Maximum Directory Size (default: 10MB)
# [ -space-guarantee|-s {none|volume} ]                        Space Guarantee Style (default: volume)
# [ -snapshot-policy <snapshot policy> ]                       Snapshot Policy (default: default)
# [ -stripe-count {1..254} ]                                   *Striped Data Volume Count
# [ -stripe-width <(256KB..1GB) 4KB increments> ]              *Striped Data Volume Stripe Width
# [ -concurrency {low|medium|high|peak} ]                      *Concurrency Level
# [ -optimize {max-performance|performance|space|max-space} ]  *Optimization Policy
# [ -foreground [{true|false}] ]                               Foreground Process (default: true)
# [ -antivirus-on-access-policy <antivirus policy> ]           Anti-Virus On-Access Policy (default: default)
# [ -stripe-format {data|full} ]                               *Striped Volume Format
# [ -index-dir-enabled [{true|false}] ]                        *Index Directory Enabled
# [ -load-sharing-source <volume name> ]                       Load Sharing Source Volume
# [ -move-target [{true|false}] ]                              *Move Target Volume (default: false)
# [ -max-write-alloc-blocks {0..2048} ]                        *Maximum Write Alloc Blocks (default: 0)
# [ -vserver <vserver name> ]                          Virtual Server Name (default: vs0)
#
# Example ngsh command usages:
#  vol create strip1 stripedaggr_0 -size 10GB -stripe-count 8
# 
# Returns:
#   If create succeeds returns $TCD::PASS
#   If create fails returns $TCD::FAIL
sub createVolume 
{
   my $args = shift;
   my $return = $TCD::PASS;

   my @required_args = qw(volume vserver aggregate);
   foreach my $req (@required_args) {
     if (!defined $args->{$req}) {
       $Test->nlog("Call to DevUnitTest::createVolume is missing argument '$req'");
       $return = $TCD::SCRIPT;
     }
   }
   return $return unless ($return == $TCD::PASS);

   my $create_options = [
      volume    => $args->{volume},
      aggregate => $args->{aggregate},
      vserver   => $args->{vserver},
      -Timeout  => 120
   ];
   push( @$create_options, 'size' => $args->{'size'} ) if (defined $args->{'size'});
   push( @$create_options, 'type' => $args->{'type'} ) if (defined $args->{'type'});
   push( @$create_options, 'policy' => $args->{'policy'} ) if (defined $args->{'policy'});
   push( @$create_options, 'user' => $args->{'user'} ) if (defined $args->{'user'});
   push( @$create_options, 'group' => $args->{'group'} ) if (defined $args->{'group'});
   push( @$create_options, 'security-style' => $args->{'security-style'} ) if (defined $args->{'security-style'});
   push( @$create_options, 'unix-permissions' => $args->{'unix-permissions'} ) if (defined $args->{'unix-permissions'});
   push( @$create_options, 'junction-path' => $args->{'junction-path'} ) if (defined $args->{'junction-path'});
   push( @$create_options, 'junction-active' => $args->{'junction-active'} ) if (defined $args->{'junction-active'});
   push( @$create_options, 'vsroot' => $args->{'vsroot'} ) if (defined $args->{'vsroot'});
   push( @$create_options, 'comment' => $args->{'comment'} ) if (defined $args->{'comment'});
   push( @$create_options, 'maxdir-size' => $args->{'maxdir-size'} ) if (defined $args->{'maxdir-size'});
   push( @$create_options, 'space-guarantee' => $args->{'space-guarantee'} ) if (defined $args->{'space-guarantee'});
   push( @$create_options, 'snapshot-policy' => $args->{'snapshot-policy'} ) if (defined $args->{'snapshot-policy'});
   push( @$create_options, 'stripe-count' => $args->{'stripe-count'} ) if (defined $args->{'stripe-count'});
   push( @$create_options, 'stripe-width' => $args->{'stripe-width'} ) if (defined $args->{'stripe-width'});
   push( @$create_options, 'concurrency' => $args->{'concurrency'} ) if (defined $args->{'concurrency'});
   push( @$create_options, 'optimize' => $args->{'optimize'} ) if (defined $args->{'optimize'});
   push( @$create_options, 'antivirus-on-access-policy' => $args->{'antivirus-on-access-policy'} ) if (defined $args->{'antivirus-on-access-policy'});
   push( @$create_options, 'stripe-format' => $args->{'stripe-format'} ) if (defined $args->{'stripe-format'});
   push( @$create_options, 'index-dir-enabled' => $args->{'index-dir-enabled'} ) if (defined $args->{'index-dir-enabled'});
   push( @$create_options, 'load-sharing-source' => $args->{'load-sharing-source'} ) if (defined $args->{'load-sharing-source'});
   push( @$create_options, 'move-target' => $args->{'move-target'} ) if (defined $args->{'move-target'});
   push( @$create_options, 'max-write-alloc-blocks' => $args->{'max-write-alloc-blocks'} ) if (defined $args->{'max-write-alloc-blocks'});

   $Test->nlog("Creating volume: '$args->{volume}'");
   my $skipSS = 0;
   eval {
      my $output = $Filer->volume_create(@$create_options);
   };
   if ($@) {
      $Test->nlog("Failed to create the source volume '$args->{volume}': ($@)");
      return $TCD::FAIL;
   }

   return $TCD::PASS;
}

# @name ::deleteDPRelationship
# @usage $return = deleteDPRelationship($destPath, $sourcePath)
# 
# @synopsis Deletes a DP mirror relationship between a SRC and DST DataProtection volume.
sub deleteDPRelationship
{
   my $destPath = shift;
   my $sourcePath = shift;

   #snapmirror delete -destination-path qdeux-11234462095://vs0/snapshot_dut_SRC_DP -source-path qdeux-11234462095://vs0/snapshot_dut_SRC
   $Test->nlog("Trying to delete the DP relationship for '$destPath'");
   eval {
      $Filer->snapmirror_delete('destination-path'=>$destPath,
                                'source-path'=>$sourcePath);
   };
   if ($@ =~ /not found/) {
      $Test->nlog("A DP relationship for '$destPath' was not found.");
      return $TCD::FAIL;
   } elsif ($@) {
      $Test->nlog("Failed to delete the DP relationship for '$destPath': ($@)");
      return $TCD::FAIL;
   }
   return $TCD::PASS;
}

# @name ::deleteLSRelationship
# @usage $return = deleteLSRelationship($destPath,$sourcePath)
# 
# @synopsis Deletes a LS mirror relationship between a SRC and DEST LoadShare volume.
sub deleteLSRelationship
{
   my $destPath = shift;
   my $sourcePath = shift;

   #snapmirror delete -destination-path qdeux-11234462095://vs0/snapshot_dut_SRC_LS -source-path qdeux-11234462095://vs0/snapshot_dut_SRC
   $Test->nlog("Trying to delete the LS Relationship for '$destPath'");
   eval {
     $Filer->snapmirror_delete('destination-path'=>$destPath,
                               'source-path'=>$sourcePath);
   };
   if ($@ =~ /not found/) {
      $Test->nlog("A LS Relationship for '$destPath' was not found.");
      return $TCD::FAIL;
   } elsif ($@) {
      $Test->nlog("Failed to delete the LS Relationship for '$destPath': ($@)");
      return $TCD::FAIL;
   }

   return $TCD::PASS;
}

# @name ::deleteVolumes
# @usage deleteVolumes($vserver,$volPrefix,$failureResult,$force)
# 
# @synopsis
#   Deletes any volumes found on $vserver with prefix $volPrefix.
#   The failure result is controllable using the 3rd optional argument. Using the
#   default will cause subsequent tests to not be run.  It is recommended to use
#   the alternative value of $TCD::FAIL if you want the tests to continue. The
#   last argument is an optional force flag. The force value defaults to 'false'.
#   To specify a value for force, you will need to specify the $failureResult.
# 
# Returns $TCD::FAIL immediately if the 'volume delete' command fails.
# 
# Returns:
#   If delete succeeds returns $TCD::PASS
#   If delete fails for any volumes, returns $TCD::FAIL.
# 
# TODOS:
#  - This function could be enhanced to detect SnapMirror relationships and 
#    automatically delete the relationships so that the delete volumes will
#    succeed. Right now that needs to occur before calling this function.
sub deleteVolumes 
{
   my ($vserver,$volPrefix,$failureResult,$force) = @_;
   if (! defined $failureResult) {
      $failureResult = $TCD::SCRIPT;
   }
   if (! defined $force) {
      $force = 'false';
   }

   #delete the volumes that were created by this test
   my $output;
   eval { $output  = $Filer->volume_show(vserver=>$vserver,
                                         volume=>"$volPrefix*",
                                         force=>"$force"); };
   return $TCD::PASS if ($@);
   my $result = $TCD::PASS;

   my $vol_cnt = scalar @{$output};
   $Test->nlog("Found $vol_cnt volumes matching name:'${volPrefix}*'");

   for (my $i = 0; $i < $vol_cnt; $i++)
   {
      my $volume = $$output[$i]{volume};

      $Test->nlog("deleting volume $volume");
      eval { my $del_output = $Filer->volume_purge(volume=>$volume, 
                                                    vserver=>$vserver); };
      if ($@) {
         $Test->nlog("Failed to delete volume '$volume': ($@)");
         $result = $failureResult;
      }
   }
   
   return $result;
}

# @name ::disableFault
# @usage $result = disableFault({'fault-id'=>$fault})
# 
# @synopsis
#   Calls 'debug fault-injection disable-fault $fault to disable the given fault
#  ID.  If the command succeeds the function returns $TCD::PASS, otherwise it 
#  returns $TCD:FAIL.
# 
sub disableFault 
{
   my $args = shift;
   
   # Define which nodes to modify for these commands
   my @nodes = (defined $args->{nodes}) ? $args->{nodes} : @Nodes;

   my $disable_args = [
      'fault-id' => (defined $args->{'fault-id'} ? $args->{'fault-id'} : undef)
   ];

   foreach my $node (@nodes) {
       ## Run the following commands on $node
       $Filer->set_execution_server(-Server => $node);

       ## Disable Fault
       $Test->nlog("Disable fault: $args->{'fault-id'} on node '$node'");
       eval {
          my $cmd = $Filer->debug_fault_injection_disable_fault(@$disable_args);
       };
       if ($@) {
          $Test->nlog("Failed to disable fault $args->{'fault-id'} on node '$node': ($@)");
          return $TCD::FAIL;
       }

   }
   
   return $TCD::PASS;
}

sub doesVolumeExist {

}

# @name ::enableFault
# @usage enableFault({'fault-id'=>$id, 
#                     'fault-mode'=>$mode, 
#                     probability=>$prob,
#                     period=>$period,
#                     'num-calls'=>$num, 
#                     'start-call'=>$start,
#                     'end-call'=>$end})
# 
# @synopsis
#  This function will enable a fault then change its mode as specified by the
# input arguments given. The $Filer must be set to a privilege that allows access
# to this command.  The change will occur on all nodes specified in @Nodes.
# 
# Arguments: Takes a hash pointer containing the arguments for the 'debug 
#   fault-injection set-fault-mode' ngsh command.
# 
#> debug fault-injection set-fault-mode ?
#   [-fault-id] <fid>          *Fault Id
#   [-fault-mode] <fmode>      *Fault Mode
#  [[-probability] <double>]   *Probability for mode GX_FAULT_MODE_RANDOM
#  [ -period <integer> ]       *Time period for mode GX_FAULT_MODE_PERIODIC_TIME
#  [ -num-calls <integer> ]    *Call Interval for mode GX_FAULT_MODE_PERIODIC_CALLS
#  [ -start-call <integer> ]   *Start call for mode GX_FAULT_MODE_SEQUENCE
#  [ -end-call <integer> ]     *End call for mode GX_FAULT_MODE_SEQUENCE
#
#> debug fault-injection set-fault-mode -fault-mode ?
#    GX_FAULT_MODE_ONCE           GX_FAULT_MODE_ALWAYS
#    GX_FAULT_MODE_RANDOM         GX_FAULT_MODE_PERIODIC_TIME
#    GX_FAULT_MODE_PERIODIC_CALLS GX_FAULT_MODE_SEQUENCE
#
#  ngsh options: 
#    'fault-id', 'fault-mode', probability, period, 'num-calls', 'start-call',
#    'end-call'
# 
#  options with defaults:
#    'fault-mode'                   - defaults to "GX_FAULT_MODE_ALWAYS"
#
#  extra options:
#    none
# 
# Returns:
#   If set && enable succeeds returns $TCD::PASS
#   If set || enable fails returns $TCD::FAIL
sub enableFault
{
   my $args = shift;

   # Define which nodes to modify for these commands
   my @nodes = (defined $args->{nodes}) ? $args->{nodes} : @Nodes;

   my $enable_args = [
      'fault-id' => (defined $args->{'fault-id'} ? $args->{'fault-id'} : undef)
   ];

   my $set_mode_args = [
      'fault-id'   => (defined $args->{'fault-id'} ? $args->{'fault-id'} : undef),
      'fault-mode' => (defined $args->{'fault-mode'} ? $args->{'fault-mode'} : 'GX_FAULT_MODE_ALWAYS')
   ];

   push( @$set_mode_args, 'probability' => $args->{'probability'} ) if (defined $args->{'probability'});
   push( @$set_mode_args, 'period' => $args->{'period'} ) if (defined $args->{'period'});
   push( @$set_mode_args, 'num-calls' => $args->{'num-calls'} ) if (defined $args->{'num-calls'});
   push( @$set_mode_args, 'start-call' => $args->{'start-call'} ) if (defined $args->{'start-call'});
   push( @$set_mode_args, 'end-call' => $args->{'end-call'} ) if (defined $args->{'end-call'});

   foreach my $node (@nodes) {
      ## Run the following commands on $node
      $Filer->set_execution_server(-Server => $node);

      ## Enable Fault
      $Test->nlog("Enable fault: $args->{'fault-id'} on node '$node'");
      eval {
         my $cmd = $Filer->debug_fault_injection_enable_fault(@$enable_args);
      };
      if ($@) {
         $Test->nlog("Failed to enable fault '$args->{'fault-id'}' on node '$node': ($@)");
         return $TCD::FAIL;
      }
      
      ## Setting the Fault Mode   
      $Test->nlog("Changing the fault $args->{'fault-id'} to mode $args->{'fault-mode'} on node '$node'");
      eval {
         my $cmd = $Filer->debug_fault_injection_set_fault_mode(@$set_mode_args);
      };
      if ($@) {
         $Test->nlog("Failed to change the fault $args->{'fault-id'} to mode $args->{'fault-mode'} on node '$node': ($@)");
         return $TCD::FAIL;
      }
   }
   return $TCD::PASS;
}

sub getNodeForFlexVol
{

    my ($vserver,$volume) = @_;

    my $output;

    eval { $output = $Filer->volume_show(vserver=>$vserver,
                             volume=>$volume); };

    return $TCD::SCRIPT if($@);

    $Test->nlog("aggregate for volume " . $vserver . " " . $volume . " = " . $$output[0]{aggregate});

    my $output2;
    eval { $output2 = $Filer->storage_aggregate_show(aggregate=>$$output[0]{aggregate}); };

    $Test->nlog("node for aggregate " . $$output[0]{aggregate} . " = " . $$output2[0]{nodes});

    return $TCD::SCRIPT if($@ || scalar @{$output2} != 1);

    return $$output2[0]{nodes};

}

#creates the given file on the given volumes using 
#dbladecli command 'mkfile'
sub fillVolWithDataHelper
{
    my $args = shift;
    my $return = $TCD::PASS;

    my @required_args = qw(node size volume filename);
    foreach my $req (@required_args) {
     if (!defined $args->{$req}) {
       $Test->nlog("Call to DevUnitTest::fillVolWithDataHelper is missing argument '$req'");
       $return = $TCD::SCRIPT;
     }
    }

    return $return unless ($return == $TCD::PASS); #don't return till the whole loop has finished.

    my $node = $args->{node};
    my $size = $args->{size};
    my $volume = $args->{volume};
    my $filename = $args->{filename};

    my $cmd = "set -privilege diag; set -confirmations off; node run -node $node -command \"priv set diag; mkfile $size /vol/$volume/$filename\"";
    $Test->nlog(" Creating file $filename on volume $volume.\n  $cmd");
    $Test->nlog( $Filer->shell(cmd=>"ngsh -c '$cmd'") );

}

#TODO:implement an equivalent function for striped volumes
#only works on flex vols
#Fills volume with given amount of data
#If data size is not specified, the volume is filled with 100MB of data
#
sub fillVolWithData
{

    my $args = shift;
    my $size = "100M";

    my $return = $TCD::PASS;

    my @required_args = qw(vserver volume);
    foreach my $req (@required_args) {
     if (!defined $args->{$req}) {
       $Test->nlog("Call to DevUnitTest::fillVolWithData is missing argument '$req'");
       $return = $TCD::SCRIPT;
     }
    }
    
    return $return unless ($return == $TCD::PASS); #don't return till the whole loop has finished.

    if($args && $args->{size}) {
        $size = $args->{size};
    }

    my $volume = $args->{volume};

    $Test->nlog("Creating a file of size " . $size . " on volume " . $volume);

    my $node = getNodeForFlexVol($args->{vserver},$args->{volume});

    return $node if($node == $TCD::SCRIPT);

    my $filename;

    if($args && $args->{filename}) {
        $filename = $args->{filename};
    }else {
        $filename = "test" . join('',localtime()) . ".txt";
    } 

    fillVolWithDataHelper({node=>$node,size=>$size,volume=>$volume,filename=>$filename});

    return $TCD::PASS;

}
 
# @name ::getClusterId
# @usage $clusterId = getClusterId()
# 
# @synopsis Uses the 'cluster identity show' command to fetch the cluster Id.
#
#  example: qdeux-11231533730
{
    my $clusterid;

    sub getClusterId 
    {
        if (!defined($clusterid)) {
            my $out = $Filer->cluster_identity_show();

            if (exists($out->[0]->{name})) {
                $clusterid = $out->[0]->{name};
                $Test->nlog("Found Cluster ID: $clusterid");
            } else {
                $Test->nlog("Failed to find the 'name'");
            }
        }
        return $clusterid;
    }
}

# @name ::getSnapshotInfoList
# @usage @snapshots = getSnapshotInfoList($volume)
# 
# @synopsis Uses the 'volume snapshot show -volume $volume' command to fetch snapshot data.
#
# Returns: Returns an arrary.  The array will be empty on an error. Otherwise
#   the return will contain a hash for each snapshot containing all the info
#   provided by the 'snapshot show' command.
#
# Example:
#    {
#       'fs-version' => '8.0',
#       'record-owner' => '""',
#       'blocks' => '0%',
#       'crtime' => '"Tue Mar 03 11:41:23 2009"',
#       'msid' => '10737419445',
#       'dsid' => '8589935999',
#       'size' => '40960',
#       'owners' => '-',
#       'create-time' => '"Tue Mar 03 11:41:23 2009"',
#       'physical-snap-id' => '2',
#       'logical-snap-id' => '2',
#       'snapshot' => 'devut_snap_2',
#       'volume' => 'snapshot_dut',
#       'vserver' => 'vs0',
#       'cpcount' => '11',
#       'internal-status' => '-',
#       'description' => '"I am a snapshot"',
#       'tags' => '"VOPL_snapshot_comment=I am a snapshot"',
#       'comment' => '"I am a snapshot"',
#       'usedblocks' => '29%',
#       'busy' => 'false'
#    }
#
sub getSnapshotInfoList {
   my $volume = shift;

   my @snaps = ();
   my $output;
   eval { $output = $Filer->volume_snapshot_show('volume' => $volume); };
   if ($@ =~ /There are no entries matching your query/) {
     return @snaps;
   } elsif ($@) {
     $Test->nlog("Failed to get volume status for '$volume': $@");
     return @snaps;
   }
   my @complex_list = @{$output};
   #$Test->nlog("Got volume snapshot list for '$volume': output=($output) \n" . Dumper(@complex_list));

   return @complex_list; #Returns an array of snapshot info hashes
}

# @name ::getSnapshotNameList
# @usage @snapshots = getSnapshotList($volume)
# 
# @synopsis Uses the 'volume snapshot show -volume $volume' command to fetch snapshot names.
#
sub getSnapshotNameList {
   my $volume = shift;

   my @snaps = ();
   my $output;
   eval { $output = $Filer->volume_snapshot_show('volume' => $volume); };
   if ($@ =~ /There are no entries matching your query/) {
     return @snaps;
   } elsif ($@) {
     $Test->nlog("Failed to get volume status for '$volume': $@");
     return @snaps;
   }
   my @complex_list = @{$output};
   #$Test->nlog("Got volume snapshot list for '$volume': output=($output) \n" . Dumper(@complex_list));

   # Grab just the names of the snapshots
   foreach my $snap (@complex_list) {
      push(@snaps, $snap->{'snapshot'});
   }
   return @snaps;
}

# @name ::getSnapshotProperties
# @usage ($isSnapMirrorRef,$isOwned) = getSnapshotProperties(%snapInfo)
# 
# @synopsis Uses the snapshot Info hash to define a few common 'properties' 
#           about the snapshot.  Returns the properties in a list.
#
# Returns: Returns an arrary.  
#    Will contain scalar values:
#       isSnapMirrorRef -  1 if the snapshot name starts with 'snapmirror'
#       isOwned         -  1 if snapshot onwer tag is used.
#
sub getSnapshotProperties 
{
   my $snapRef = shift;
   my ($isSnapMirrorRef,$isOwned);

   # set the 'isSnapMirrorRef' flag
   $isSnapMirrorRef = 1 if ($snapRef->{'snapshot'} =~ /^snapmirror/);

   # set the 'isOwned' flag
   $isOwned = 1 if ($snapRef->{'record-owner'}); # true if its non-empty

   return ($isSnapMirrorRef,$isOwned);
}

# @name ::makeSnapmirrorPath
# @usage $path = makeSnapmirrorPath($volume,$vserver,$clusterId)
# 
# @synopsis Combines the three pieces of information into the correct format 
#  for use in snapmirror commands. Cluster Id will be fetched if not given.
#
#  example: qdeux-11231533730://vs0/SRC 
sub makeSnapmirrorPath
{
  my ($volume,$vserver,$clusterId) = @_;

  #Fetch the clusterId if it wasn't passed into this function
  unless (defined $clusterId) {
    $clusterId = getClusterId();
  }

  return $clusterId . "://" . $vserver . "/" . $volume;
}

#
# @name ::moveTestVol
# @usage $result = moveTestVol({vserver=>$vserver,
#                               volume=>$volume,
#                               'destination-aggregate'=>$destAggr,
#                               'expectResuts'=>$string,
#                                isMem=>1})
#
# @synopsis  Moves a volume to a new aggregate. Takes a hash of 
#    named options. All options should match the options to the 'volume move' 
#    ngsh command. Given an 'expectResult' value the command will fail if the
#    completion string does not match.
# 
# volume move ?
#  [ -vserver <vserver name> ]                 Virtual Server Name (default: vs0)
#   [-volume] <volume name>                    Volume Name
#   [-destination-aggregate] <aggregate name>  Destination Aggregate
#  [[-destination-stripe-format] {data|full}]  *Destination Stripe Format
#  [ -foreground [{true|false}] ]              Foreground Process (default: false)
#
#  ngsh options: 
#     vserver, volume, 'destination-volume', 'destination-aggregate',
#     'destination-stripe-format', forground
# 
#  options with defaults:
#    vserver                 - defaults to global symbol ${vserver}
#    volume                  - defaults to global symbol ${volPrefix}
#    'destination-aggregate' - defaults to global symbol ${dstaggr}
#
#  extra options:
#     expectResults => "string"
#        This option searches the returned progress string (in case of an error)
#        for a match to this "string".  If a match is not found, this function
#        will return $TCD::FAIL
#
#  If isMem option is specified and is not zero then 'volume member move' command is run
# 
# @result
#   If move succeeds and expectResult is empty-string, returns $TCD::PASS
#   If move fails, and expectResult is found in the job status, returns $TCD::PASS
#   If move succeeds, but expectResult is non-empty, returns $TCD::FAIL
#   If move fails, and expectResult is empty-string, returns $TCD::FAIL
sub moveTestVol
{
   my $args = shift;
   my $return = $TCD::PASS;

   my @required_args = qw(vserver volume destination-aggregate);
   foreach my $req (@required_args) {
     if (!defined $args->{$req}) {
       $Test->nlog("Call to DevUnitTest::moveTestVol is missing argument '$req'");
       $return = $TCD::SCRIPT;
     }
   }
   return $return unless ($return == $TCD::PASS); #don't return till the whole loop has finished.

   $args->{volume} = $volPrefix unless (defined $args->{volume}); # preset the main default

   my $move_options = [
      vserver                 => $args->{vserver},
      volume                  => $args->{volume},
      'destination-aggregate' => $args->{'destination-aggregate'},
      foreground              => 'true',
      -Timeout                => 600
   ];
   my $expectResult = defined $args->{expect} ? $args->{expect} : undef;

   $Test->nlog("Attempting to move $args->{volume} to $args->{'destination-aggregate'}." . 
               (($expectResult) ? " Expect result /$expectResult/" :
                                 "(should succeed)"));

   if($args->{isMem}) {
       eval { $Filer->volume_member_move(@$move_options); };
   }else {
       eval { $Filer->volume_move(@$move_options); };
   }

   $Test->nlog("move result:");
   $Test->nlog( ($@) ? Dumper($@) : "  SUCCESS"); 

   $return = $TCD::PASS;
   if ($@) {
      if (!$expectResult || !($@ =~ m/$expectResult/)) {
         $Test->nlog("FAILED to move the volume $args->{volume}");
         $return = $TCD::FAIL;
      } else {
         $Test->nlog("Failed as expected: Result matches /$expectResult/");
      }
   } elsif ($expectResult) {
      $Test->nlog("Volume move unexpectedly succeeded: Result did not match /$expectResult/");
      $return = $TCD::FAIL;
   }
   return $return;
}

#returns number of member in a given striped aggregate
sub numStripedAggrMem
{

    my $args = shift;
    my $return = $TCD::PASS;

    my $parent = "";
    if($args && $args->{parent}) {
        $parent = $args->{parent} . "*";
    }else {
        $parent = "*";
    }

    my @memaggrs = $Filer->storage_aggregate_member_show(aggregate => $parent);

    return (scalar @memaggrs);

}

# @name ::pickAggregates
# @usage ($srcaggr,$destaggr,$dstStripedAggr) = pickAggregates()
# 
# @synopsis Randomly picks several types of aggregates in the cluster, and
#  returns them in a list.
# 
#  options:
#    none
#
#  extra options:
#    none
# 
# Returns:
#   A source aggregate, this will be a non-striped aggregate other than 'aggr0'.
#   A destination aggregate, this will be a second non-striped aggregate 
#      (not 'aggr0') if one exists. Otherwise it will be equal to $srcaggr.
#   A destination striped aggregate. If a striped aggregate is not found, it will be 'undef'.
sub pickAggregates 
{
   #pick an aggregate
   $Test->nlog("Looking for a aggregates to use.");
   my $output = $Filer->storage_aggregate_show();
   my @complex_list = @{$output};
   $trace->debug(msg=>"complex_list = " . Dumper(@complex_list));
   my @aggr_list = ();
   my @aggr_striped_list = ();
   for (my $i = 0; $i < scalar @complex_list; $i++) {
      $Test->nlog("aggr[$i] '" . $complex_list[$i]{aggregate} . "' has type " . $complex_list[$i]{'voltype'});
      if ( $complex_list[$i]{'voltype'} ne "striped" && $complex_list[$i]{'aggregate'} !~ /aggr0/) {
         push(@aggr_list, $complex_list[$i]{aggregate});
      }
      if ( $complex_list[$i]{'voltype'} eq "striped" && $complex_list[$i]{'aggregate'} !~ /aggr0/) {
         push(@aggr_striped_list, $complex_list[$i]{aggregate});
      }
   }
   $Test->nlog("   flex = @aggr_list (".scalar @aggr_list." entries)");
   $Test->nlog("striped = @aggr_striped_list (".scalar @aggr_striped_list." entries)");
   
   #randomly choose aggregate
   my $srcaggr = splice(@aggr_list, int(rand(scalar @aggr_list)), 1);
   my $dstaggr = splice(@aggr_list, int(rand(scalar @aggr_list)), 1);
   my $dstStripedAggr = undef;
   if (scalar @aggr_striped_list == 1) {
      $dstStripedAggr = $aggr_striped_list[0]; 
   } else {
      $dstStripedAggr = splice(@aggr_striped_list, int(rand(scalar @aggr_list)), 1); 
   }

   $dstaggr = $srcaggr unless $dstaggr;

   $Test->nlog("Using srcaggr:$srcaggr");
   $Test->nlog("Using dstaggr:$dstaggr");
   $Test->nlog("Using dstStripedAggr:$dstStripedAggr");

   return ($srcaggr, $dstaggr, $dstStripedAggr);
}

# pickMemberAggr: returns a member aggr
# if args->{parent} is provided then it will return a member aggr that matches "$args->{parent}*"
# if args->{except} is provided then it will return a member other than the one given in except

sub pickMemberAggr
{

    my $args = shift;
    my $return = $TCD::PASS;

    my $parent = "";
    my $except = "";

    if($args && $args->{parent}) {
        $parent = $args->{parent} . "*";
    }else {
        $parent = "*";
    }

    if($args && $args->{except}) {
        $except = $args->{except};
    }

    my @memaggrs = $Filer->storage_aggregate_member_show(aggregate => $parent);
    
    my @aggrlist = ();

    for (my $i = 0; $i < scalar @memaggrs; $i++) {

        #$Test->nlog("pickMemberAggr: processing aggr " . $memaggrs[$i]{aggregate});

        push(@aggrlist,$memaggrs[$i]{aggregate}) unless ($memaggrs[$i]{aggregate} eq $except);

    }

    return $TCD::SCRIPT if(scalar @aggrlist == 0);

    my $dstaggr = splice(@aggrlist, int(rand(scalar @aggrlist)), 1);

    $Test->nlog("pickMemberAggr: selected aggr = " . $dstaggr);
    
    return $dstaggr;

}

#pickMemberVol:
#picks a member vol of given vserver and given striped volume,
#if selectRep is set then it selects a REP othewise a DV
#returns the selected volume member name and the aggregate it resides in

sub pickMemberVol
{

    my $args = shift;
    my $return = $TCD::PASS;

    my @required_args = qw(vserver volume);
    foreach my $req (@required_args) {
     if (!defined $args->{$req}) {
       $Test->nlog("Call to DevUnitTest::pickMemberVol is missing argument '$req'");
       $return = $TCD::SCRIPT;
     }
    }
  
    return $return unless ($return == $TCD::PASS); #don't return till the whole loop has finished.

    my $vtype = "DV";
    my $volsuffix = "";
    if($args->{selectRep}) {
        $vtype = "REP";
    }else {
        $volsuffix = "_*"
    }

    $Test->nlog("Picking a " . $vtype . " member volume for given striped volume " . $args->{vserver} . " " . $args->{volume} );

    my $parent = $args->{volume};
    
    my @memvols = $Filer->volume_member_show(vserver=>$args->{vserver},
                                              volume=>$parent.$volsuffix);

    if($args->{selectRep}) {

        if(scalar @memvols != 1) {
            return $TCD::SCRIPT;
        }

        $Test->nlog("Found mem vol " . $memvols[0]{volume} . " aggr = " . $memvols[0]{aggregate});

        return ($memvols[0]{volume},$memvols[0]{aggregate});

    }

    #Now we need to select the DV

    # Sanity check
    if(scalar @memvols == 0) {
        $Test->nlog("Could not find a member volume ");
        return $TCD::SCRIPT;
    }

    my $srcvol = splice(@memvols, int(rand(scalar @memvols)), 1);

    $Test->nlog("Found mem vol " . $srcvol->{volume} . " aggr = " . $srcvol->{aggregate});

    return($srcvol->{volume}, $srcvol->{aggregate});     

}
 
# @name ::pickVServer
# @usage $vserver = pickVServer()
# 
# @synopsis Randomly picks a VServer in the cluster, and returns its name.
# 
#  options:
#    none
#
#  extra options:
#    none
# 
# Returns:
#   A vserver name.
sub pickVServer 
{
   $Test->nlog("Looking for a vserver to use.");

   my $output = $Filer->vserver_show();
   my @complex_list = @{$output};
   $trace->debug(msg=>"complex_list = " . Dumper(@complex_list));

   my @vserver_list = ();
   for (my $i = 0; $i < scalar @complex_list; $i++) {
     if( $complex_list[$i]{'role'} eq "cluster") {
       push(@vserver_list, $complex_list[$i]{vserver});
     }
   }
   $trace->trace(msg=>"vserver_list = @vserver_list (".scalar @vserver_list."  entries)");

   #randomly choose vserver
   my $vs = splice(@vserver_list, int(rand(scalar @vserver_list)), 1);
   $Test->nlog("Using vserver:$vs");
   return $vs;
}

# Removes given file from the volume
sub removeFileFromVol
{
    my $args = shift;
    my $return = $TCD::PASS;

    my @required_args = qw(node volume filename);
    foreach my $req (@required_args) {
     if (!defined $args->{$req}) {
       $Test->nlog("Call to DevUnitTest::removeFileFromVol is missing argument '$req'");
       $return = $TCD::SCRIPT;
     }
    }

    return $return unless ($return == $TCD::PASS); #don't return till the whole loop has finished.

    my $node = $args->{node};
    my $volume = $args->{volume};
    my $filename = $args->{filename};

    my $cmd = "set -privilege diag; set -confirmations off; node run -node $node -command \"priv set diag; rm /vol/$volume/$filename\"";
    $Test->nlog(" Removing file $filename from volume $volume.\n  $cmd");
    $Test->nlog( $Filer->shell(cmd=>"ngsh -c '$cmd'") );

}


# @name ::searchMgwdLog
# @usage searchMgwdLog($filter,$verify_str)
# 
# @synopsis
#   Will print the contents of the /mroot/etc/mlog/mgwd.log file and search for 
# a specified string using the test->nlog() routine.  If given a $filter this 
# function will limit the searched results. If given a $verify_str it will
# return $TCD::FAIL if the str can not be found. Otherwise, it returns 
# $TCD::PASS.
#
sub searchMgwdLog
{
   my ($filter,$verify_str) = @_;
   my $return = $TCD::PASS;
   my $output = "";
   my $found_str = 0;

   foreach my $node (@Nodes) {
      ## Run the following commands on $node
      $Filer->set_execution_server(-Server => $node);

      eval { 
         $output = $Filer->shell(cmd=>"grep \"$filter\" /mroot/etc/mlog/mgwd.log");
      };
      if ($@) {
         $Test->nlog("Failed to get the mgwd.log file contents on node $node: ($@)");
         $return = $TCD::FAIL;
      } else {
         $Test->nlog("\ngrep \"$filter\" /mroot/etc/mlog/mgwd.log for node $node:\n$output\n");
         
         # Verify the expected string is found in the filtered data
         $found_str++ if ($output =~ m/$verify_str/);
      }
   }

   if (defined $verify_str) {
      if ($found_str) {
         $Test->nlog( "Found desired string '$verify_str' in mgwd.log $found_str times." );
      } else {
         $Test->nlog( "ERROR: '$verify_str' not found in any mgwd.log files." );
         $return = $TCD::FAIL;
      }
   }
   return $return;
}
                                    
# @name ::setPrivilege
# @usage setPrivilege($level)
# 
# @synopsis Modifies the CLI privilege for a given Filer
# 
#  ngsh options: 
#    level => {admin|advanced|diagnostic|test}
# 
#  options with defaults:
#    level                   - defaults to "test"
#
#  extra options:
#    none
# 
# Returns:
#   If set succeeds returns $TCD::PASS
#   If set fails returns $TCD::FAIL
sub setPrivilege
{
   my $level = shift;
   $level = 'test' unless defined $level;

   foreach my $node (@Nodes) {
      ## Run the following commands on $node
      $Filer->set_execution_server(-Server => $node);

      ## Set privilege level
      $Test->nlog("Set privilege level => '$level' for node '$node'");
      eval {
         my $cmd = $Filer->set(
            'privilege' => $level,
            'confirmations' => 'off'
            );
      };
      if ($@) {
         $Test->nlog("Failed to set privilege level to '$level' for node '$node': ($@)");
         return $TCD::FAIL;
      }
   }
   return $TCD::PASS;
}

# @name ::showTraceBuffer
# @usage showTraceBuffer($proc,$buf,$verify_str)
# 
# @synopsis
#   Will print the contents of the MTRACE buffer for a specified proc and buffer
# using the test->nlog() routine.  If given a @verify_str this function will 
# return $TCD::FAIL if the str can not be found. Otherwise, it returns 
# $TCD::PASS.
#
#  The mtrace contents are printed in a formated table. Columns shown are:
#    'seq-num'
#    'emit-time'
#    'filename'
#    'line-num'
#    'category'
#    'message'
#
sub showTraceBuffer
{
   my ($proc,$buf,$verify_str) = @_;
   my $return = $TCD::PASS;
   my $output = "";
   my $found_str = 0;

   #debug mtrace log show -node * mgwd -buffer vopl -fields message,category 
   eval { 
      $output = $Filer->debug_mtrace_log_show(process=>$proc,
                                              buffer=>$buf,
                                              node=>'*'); 
   };
   if ($@ =~ /There are no entries matching your query/) {
     $Test->nlog("The mtrace buffer is empty.");
   } elsif ($@) {
     $Test->nlog("Failed to get the mtrace buffer: ($@)");
     $return = $TCD::FAIL;
   } else {
      my @mtrace_buffer = @{$output};
      $Test->nlog("\nmtrace buffer <$proc:$buf>:");
      foreach my $buffer (@mtrace_buffer)
      {
         ## FORMAT OF MTRACE HASH
         # {
         #   'emit-time' => '"1/30/2009 09:46:17"',
         #   'flags' => '""',
         #   'buffer' => 'vopl',
         #   'filer' => 'qdeux-1',
         #   'node' => 'qdeux-1',
         #   'message' => '"blah blah blah"',
         #   'emit-ticks' => '68631321265',
         #   'process' => 'mgwd',
         #   'message-len' => '453',
         #   'thread' => 'ami_session',
         #   'seq-num' => '667',
         #   'filename' => 'volumeObject.cc',
         #   'line-num' => '850',
         #   'category' => 'vopl'
         # };
         
         $Test->nlog( sprintf('%6d %-19s %12s:%-6d %-20s %s', 
                              $buffer->{'seq-num'},
                              $buffer->{'emit-time'},
                              $buffer->{'filename'},
                              $buffer->{'line-num'},
                              $buffer->{'category'},
                              $buffer->{'message'}) );
         
         # Verify the copy job hit the FAULT INJ Point
         $found_str++ if ($buffer->{'message'} =~ m/$verify_str/);
      }
   }

   if (defined $verify_str) {
      if ($found_str) {
         $Test->nlog( "Found desired string '$verify_str' in mtrace<$proc:$buf> buffer $found_str times." );
      } else {
         $Test->nlog( "ERROR: '$verify_str' not found in the mtrace<$proc:$buf> buffer." );
         $return = $TCD::FAIL;
      }
   }
   return $return;
}

# @name ::snapshotPromote
# @usage snapshotPromote($volume,$snapshotToPromote,$force,$outputRef)
# 
# @synopsis
#   Will promote a snapshot on a volume to the active file system. If the 
# promote fails it will return $TCD::FAIL. Otherwise, it returns. If given a 
# reference to a scalar, the output of the command will be written to it.
# $TCD::PASS.
# 
#  Arguments 'force' and 'outputRef' are optional.
#
sub snapshotPromote {
   my ($volume,$snap,$force,$outputRef) = @_;
   $Test->nlog("Attempting to promote snapshot '$snap' on '$volume'");

   $force = 'false' if (!defined $force);

   my $return = $TCD::PASS;
   my $output;
   my $outputOK = ($outputRef != undef && 
                   ref($outputRef) == 'REF' &&
                   ref($$outputRef) == 'SCALAR') ? 1 : undef;
   eval { 
      ####  Hack alert  ####
      # Since I'm adding the -force option now, it isn't supported by the 
      # server object.  So I'll have to send it like a raw shell command.
      # I'm only going to do this if the 'force' == true.  
      ######################
      if ($force eq 'false') {
         $output = $Filer->volume_snapshot_promote('volume' => $volume,
                                                   'snapshot' => $snap,
                                                   'force' => $force); 
      } else {
         my $cmd = "set -privilege diagnostic; set -confirmations off; volume snapshot promote -volume $volume -snapshot $snap -force true";
         #$Test->nlog(" Using 'raw' command to promote snapshot '$snap' for '$volume'.\n  $cmd");
         $output = $Filer->shell(cmd=>"ngsh -c \"$cmd\"");
         
         die("\nCommand: $cmd\n$output") unless ($output =~ /Successfully promoted snapshot/);
      }
   };
   if ($@) {
      $Test->nlog("Failed to promote snapshot '$snap' for '$volume': $@\noutput:'$output'") unless ($outputOK);
      $$outputRef = $@ if ($outputOK);
      $return = $TCD::FAIL;
   } else {
      # If given a correct 'reference to a scalar' store the output.
      $$outputRef = $output if ($outputOK);
   }

   return $return;
}

# @name ::updateDPMirror
# @usage $result = updateDPMirror($destPath,$sourcePath,$vserver)
# 
# @synopsis
#   Will return $TCD::PASS if the DP is updated successfully.
#   Otherwise, it returns 'undef'.
#
sub updateDPMirror {
   my ($destPath,$sourcePath,$vserver) = @_;

   # Perform an update on the DP destination
   $Test->nlog("Updating the dp relationship for '$destPath'");
   eval {
      $Filer->snapmirror_update('destination-path'=>$destPath,
                                'source-path'=>$sourcePath,
                                'vserver'=>$vserver,
                                'foreground'=>'true',
                                -Timeout=>240);
   };
   if ($@) {
      $Test->nlog("Failed to update DP relationship for '$destPath': ($@)");
      return $TCD::FAIL;
   }
   return $TCD::PASS;
}

# @name ::updateLSSet
# @usage $result = updateLSSet($sourcePath)
# 
# @synopsis
#   Will return $TCD::PASS if the LS set is updated successfully.
#   Otherwise, it returns 'undef'.
#
sub updateLSSet {
   my $sourcePath = shift;
   my $return = $TCD::PASS;

   # Initialize the LS destination
   $Test->nlog("Initializing LS relationship for '$sourcePath' ");
   eval {
      $Filer->snapmirror_initialize_ls_set('source-path'=>$sourcePath,
                                           'foreground'=>'true',
                                           -Timeout=>240);
   };
   if ($@) {
      $Test->nlog("Failed to initialize the LS relationship for '$sourcePath': ($@)");
      return $TCD::FAIL;
   }

   # Perform an update on the LS destination
   $Test->nlog("Updating LS relationship for '$sourcePath' ");
   eval {
      $Filer->snapmirror_update_ls_set('source-path'=>$sourcePath,
                                       'foreground'=>'true');
   };
   if ($@) {
      $Test->nlog("Failed to update LS relationship for '$sourcePath': ($@)");
      return $TCD::FAIL;
   }

   return $return;
}

# @name ::volExists
# @usage volExists($volume)
# 
# @synopsis
#   Will return 1 if the volume can be seen using the 'volume status' command. 
# Otherwise, it returns 'undef'.
#
sub volExists {
  my $volume = shift;

  # If we can get the status of the volume, then it exists.
  if (volStatus($volume) != undef) {
    $Test->nlog("Volume '$volume' already exists");
    return 1;
  }
  $Test->nlog("Volume '$volume' does not exist");
  return undef;
}

# @name ::volStatus
# @usage volStatus($volume)
# 
# @synopsis
#   Will return the volume information if the volume can be seen using the
# 'volume status' command.  Otherwise, it returns 'undef'.
#
sub volStatus {
  my $volume = shift;

  my $output;
  eval { $output = $Filer->volume_status('volume' => $volume); };
  if ($@ =~ /There are no entries matching your query/) {
    return undef;
  } elsif ($@) {
    $Test->nlog("Failed to get volume status for '$volume': $@");
    return undef;
  }

  my @complex_list = @{$output};
  #$Test->nlog("Got volume status for '$volume': output=($output) \n" . Dumper(@complex_list));

  return $complex_list[0]; # This is a hash reference with all the volume stats
}

# @name ::vreport_show
# @usage vreport_show("$vserver:$flexVolume","in WAFL Only")
# 
# @synopsis
#   Will verify if the volume is found in the vreport output 
#   and returns TCD::PASS if the reason in the vreport output
#   contains the 2nd parameter passed to the function by the testcase.
#   Else it returns TCD::FAIL

sub vreport_show
{
    my($object,$reason,$type) = @_ ;
    my $output;
    $Test->nlog("\n********* Vreport Show for $object *********\n");
    #$Test->nlog("\n\n\n$object\t$reason\n\n\n\n ");
    eval { 
        $output = $Filer->debug_vreport_show( -Server=>"$node");
    };
    return $TCD::PASS if ($@);
    
    my $cmd;
    if(defined($type))
    {
      $cmd = "set -privilege diag; debug vreport show -object $object -type $type";
    }
    else
    {
      $cmd = "set -privilege diag; debug vreport show -object $object ";
    }
    my $result = $Filer->shell(cmd=>"ngsh -c \"$cmd\"");
    #my @complex_list = @{$output};
    #$Test->nlog("debug_vreport_show: " . Dumper(@complex_list));
    
    my $obj_cnt = scalar @{$output};
  #  $Test->nlog("\n$obj_cnt\n");
   $object =~ s/\*/\.\*/g; 
   for (my $i = 0; $i < $obj_cnt; $i++)
   {
        my $objectShow = $$output[$i]{object}; 
        my $reasonShow = $$output[$i]{reason};
        my $typeShow = $$output[$i]{type};
 #       $Test->nlog("\n\n\n$objectShow\t$reasonShow\t$typeShow\n\n\n\n ");
           
        if ( $objectShow =~ /$object/ ) 
        {
         $Test->nlog("PARTIAL PASS : $object discrepancy found \n checking the reason and type\n");
            if(defined($type))
            {
               if( $typeShow =~ /$type/)
               {
                  $Test->nlog("PARTIAL PASS : $object discrepancy type found \n checking the reason\n");
                  if ( $reasonShow =~ /$reason/ )
                  {
                      $Test->nlog("$result");
                      $Test->nlog("SUCCESS: $object  discrepancy with reason \"$reason\" and of type \"$type\" successfully found \n");
                      return $TCD::PASS; 
                  }
                  else 
                  {
                      #$Test->nlog("$result");
                      #$Test->nlog("FAIL: $object discrepancy with reason \"$reason\" and of type \"$type\" not found \n");
                      #return $TCD::FAIL;
                  }
               }
	       else
               {
                     #$Test->nlog("$result");
                     # $Test->nlog("FAIL: $object discrepancy of type \"$type\" not found\n");
                     # return $TCD::FAIL;
               }
         }
         else
         {
            if ( $reasonShow =~ /$reason/ )
            {
              $Test->nlog("$result");
               $Test->nlog("SUCCESS: $object discrepancy with reason \"$reason\"  successfully found\n");
               return $TCD::PASS; 
           }
           else 
           {
               #$Test->nlog("$result");
               #$Test->nlog("FAIL: $object discrepancy with reason \"$reason\" not found\n");
               #return $TCD::FAIL;
           }
         }   
      }
   }
    $Test->nlog("FAIL: $object discrepancy not found\n");
    return $TCD::FAIL ;
}

# @name ::vrepair_fix
# @usage $output = vrepair_fix("volume","$vserver:$flexVolume");
# @usage $output = vrepair_fix("aggregate","$flexAggr");
#@usage $output = vrepair_fix("volume","$vserver:$acroVolume","create");
# 
# @synopsis
#   Takes in the the type of object you are fixing(volume,snapmirror,junction,aggregate)
#   and the object to be fixed.
#   Returns TCD::PASS if vreport fix is successful else returns TCD::FAIL.
#@change 08/05/2010 talluri added support to handle "-action" to vrepair_fix sub routine

sub vrepair_fix
{
    my($type,$object,$action,$node,$aggregate) = @_ ;
    $Test->nlog("\n********* Vreport fix for $object and type \"$type\"*********\n");
    my $cmd = "set -privilege diag; debug vreport fix -type $type -object $object";
    
    
    if ($node)
    {
        $cmd = $cmd." -node $node";
    }
    if($action)
    {
        $cmd = $cmd." -action $action";
    }
    if($aggregate)
    {
        $cmd = $cmd." -aggregate $aggregate";
    }
    
    
    $Test->nlog("$cmd");
    
    my $output;
    eval{
      $output = $Filer->cli(cmd=>$cmd);
    };
    if($@){
      $Test->nlog("FAILED: Unable to fix $object discrepancy for object $object");
      return $TCD::FAIL;
    }
    #elsif ($output =~ /$object/) 
    #{
    #    $Test->nlog("$output");
    #    $Test->nlog("FAILED: Unable to fix $object discrepancy for object $object");
    #    return $TCD::FAIL;
    #}
    #else 
    #{
    #    $Test->nlog("\nSUCCESS: $object  discrepancy fixed successfully ");    
    #}
   $Test->nlog("\nSUCCESS: $object  discrepancy fixed successfully ");
   return $TCD::PASS;
}


1;
