# $Id: //depot/prod/test/Rsierranevada/lib/DfmUtils.pm#19 $

#***********************************************************************
#
# NAME
#   DfmUtils.pm - Utility functions to run DFM commands
#
# SYNOPSIS
#   Provides utility functions to run DFM commands
#
# DESCRIPTION
#   This Perl module is meant to provide utility functions to run DFM commands
#
# AUTHOR
#	dl-sm-nacl-dev
#
#***********************************************************************

## @summary     Programmatic administration of a Server.
## @author sm-nacl-dev (dl-sm-nacl-dev@netapp.com)
## @change 2010-03-17 deepthis:        Original template created - run_monitor code written
## @change 2010-03-24 deepthis:        Review commnets incorporated for run_monitor, _get_monitor_timestamps, Pod written
## @change 2010-04-05 deepthis:        Changes to remove repeated checks of timestamps, configurable value of probing interval, hyphen option names
## @change 2010-06-02 jagadees:        Modified Argument Exception
## @change 2010-11-22 nraju:           Added subroutines is_dfm_alive() and add_host().ran perltidy.
## @change 2010-11-22 nraju:           Added subroutines get_aggrs() , get_vservers(), get_volumes().
## @change 2011-02-02 nraju:           Modified is_dfm_alive() to fix warning messages.
## @change 2011-05-10 deepthis:        Modified the implementation of _get_monitor_timestamps()  
## @change 2011-10-18 revathi:         Added subroutines free_filer_ip()
## @change 2013-10-5 kshashid:         Added subroutines : install_plugin and initialize_DFM
## @change 2013-07-05 sskumar:         Added subroutines : edit_filerVersion()

# The code present in this file provides functions that are needed to run some DFM commands.
# Having an APISet alone for the command isn't quite enough. There is more processing to be done
# in some cases. All those functionalities are present in this file.

package DfmUtils;

#
# Compiler directives.
#
use strict;
use warnings;

#
# Module import.
#
use NACL::APISet;
use NACL::C::Aggregate;
use NACL::C::Node;
use NACL::C::Vserver;
use NACL::C::Volume;
use Data::Dumper;
use HAutils;

#
# Import Log utility
#
use Tharn qw($Log);
use Params::Validate qw(:all);

#
# Import exceptions
#
use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

use vars qw(@ISA @EXPORT );
use Exporter;

@ISA = qw(Exporter OntapSSH);

our @EXPORT_OK = qw(
  run_monitor
  is_dfm_alive
  add_host
  delete_host
  get_aggrs
  get_vservers
  get_volumes
  get_job_status
  free_filer_ip 
  initialize_DFM
  install_plugin
  edit_filerVersion
  increase_vfiler_limit
  );

=pod

=head1 NAME

DfmUtils.pm - Utility functions to run DFM commands

=head1 SYNOPSIS

  use DfmUtils;

  To use any function,  do this:
  DfmUtils::<method>( <arguments> );

=head1 DESCRIPTION

The file provides functions used to provide functionality that cannot be achieved
only by having APISets. These functions internally call APISets but do more processing on the output.

=head1 METHODS

All the functions present in this perform the required operation. If it fails,
then an exception is thrown. It is the responsibility of the calling function to catch the
necessary exceptions.

=head1 AUTHOR

  Deepthi Shetty (deepthis@netapp.com)

=cut

=pod

=head1 GLOBAL VARIABLES

=head2 $DEFAULT_WAIT_TIME

 This is used in the run_monitor module. This is the amount of time,
 for which the module waits for the monitor to get refreshed, if the user does not pass any specific value.
 The value is given in MINUTES. This value is assigned to 10 minutes.
This value is taken considreing most of the monitors need more or less this much time to refresh.

=head2 $DEFAULT_CHECK_INTERVAL

 This is the interval during which a check is done to see if the time-stamps of the required monitors are refreshed.
 The value is given in SECONDS. By default the value is assigned to 60. That is, every 1 minute we check if the timestamps
 of the monitors are updated.

=cut

our $DEFAULT_WAIT_TIME      = 10;
our $DEFAULT_CHECK_INTERVAL = 60;

##################################################################################################
# Method    :  _get_monitor_timestamps
# Objective :  To obtain the timestamps of the list of monitors
# Input     :  Reference to a hash containing: The APISet object, Fullname/IP of the host,
#              monitor list to be checked, a reference to a hash in which the timestamps are put.
# Ouput     :  Reference to a hash containing the timestamps.
# Exceptions : None thrown directly.
###################################################################################################
sub _get_monitor_timestamps {

    #Get the argumemts
    my (%args) = @_;

    #Get the APISet object, fullname/IP of the host, monitor list to check
    my ($api_set_obj, $host_ip, $monitor_check_ref);

    $api_set_obj       = $args{'api_set_obj'};
    $host_ip           = $args{'host_ip'};
    $monitor_check_ref = $args{'monitor_check_ref'};

    my @monitor_list = @{$monitor_check_ref};
    my (%timestamp_hash, %timestamp_hash_one, %timestamp_hash_two);
    
    # Variable to hold the comma separated list of the monitors with "Timestamp" appended to each
    my $comma_list_of_monitors;  

    # Variable to hold the command, response object 
    my ($command,$response); 

    # Variable to hold the parsed output
    my @parsed_output;

    # Hold the host-id value 
    my $host_id;  

    # Command to run to obtain the host ID
    my $list_command = "dfm_host_list";  

    # If "agent" monitor present change to "dfm agent list" 
    if( $monitor_list[0] eq "agent" ) {
	$list_command = "dfm_agent_list"; 
    }  
 
    $response = $api_set_obj->$list_command( "objects" => $host_ip );
    ($host_id) = $response->get_raw_output() =~ /^\s*(\d+)\s+/m;

    # Hold two lists of monitors: One for userQuota and/or xmlQtree if present (List #1)
    # and other for all the other monitors (List #2) 
    my (@monitor_list_one, @monitor_list_two); 

    foreach my $mon ( @monitor_list ) {

        if( $mon =~ /userQuota|xmlQtree|snapshot/i ) {

            push( @monitor_list_one, $mon);

        } else { 

	    push( @monitor_list_two, $mon);
	}
    }

    if( @monitor_list_one ) {

        # Get details of the object in the response object 
        my $response = $api_set_obj->dfm_detail( 'object-id' => $host_id );  

        #Get output in parsed format to take timestamps 
        my $parse_reference = $response->get_parsed_output();
        $Log->debug( "dfm detail response:\n" . Dumper($response) ); 

        # Get time stamps  
        %timestamp_hash_one = map { $_ => $parse_reference->[0]->{$_."Timestamp"} } @monitor_list_one;  

    } #end of if @monitor_list_one 

    if( @monitor_list_two ) {
	    
        # Make a comma separated list of the monitors with "Timestamp" appended to each
        $comma_list_of_monitors = $monitor_list_two[0] . "Timestamp";

        for( my $i = 1; $i < @monitor_list_two; $i++ ){

            $comma_list_of_monitors .= "," . $monitor_list_two[$i] . "Timestamp";
        }

        # Constuct the command to obtain the time stamps
        $command = "dfm query run \"select \"$comma_list_of_monitors\" from objects,hosts where objid=hostid and hostid=$host_id\"";
     
        # Run the command
        $response = $api_set_obj->execute_raw_command( "command" => $command  );
        $Log->debug("Response from dfm query run:\n" . Dumper( $response ) );

        # Split the output lines. First line has the list of monitors and the second has timestamps.

        my @lines = split( "\n", $response);
        my ($monitor_line,$timestamp_line );
        for (my $i=0; $i < @lines; $i++ ) {
            if( $lines[$i] =~ /Timestamp\"/ ) {
                $monitor_line = $lines[$i];
                $timestamp_line = $lines[$i+1];
                last;
            }
        }

        my @list_monitor = split( ",", $monitor_line );
        my @list_timestamp = split( ",", $timestamp_line );

        for( my $i = 0; $i < @list_monitor ; $i++ ) {

            my $monitor_name = $list_monitor[$i];
            $monitor_name =~ s/^\"(.*)\"[\r\n]*$/$1/;

            my $timestamp = $list_timestamp[$i];
            $timestamp =~ s/^\"(.*)\"[\r\n]*$/$1/;

	    # If no timestamp present, put an 'undef' value
	    unless ( $1 ) {
	        $timestamp_hash_two{$monitor_name} = undef;
	    } else {
                $timestamp_hash_two{$monitor_name} = $timestamp;
	    }
        }
 
    } #end of if @monitor_list_two  

    # Combine the two hashes to be returned
    %timestamp_hash = ( %timestamp_hash_one, %timestamp_hash_two);

    $Log->debug("Timestamp hash to be returned: ". Dumper( \%timestamp_hash ) );

    # Return the result
    return \%timestamp_hash;

} ## end sub _get_monitor_timestamps

=head1 run_monitor

	Runs a list of monitors for a particular host and check if the monitors are refreshed.

	Usage:

	DfmUtils::run_monitor( %arguments );

=head2 Parameters:

=over

=item api_set_obj: [Required]; An object of the type NACL::APISet


=item host_ip: [Required]; The full object name or the IP of the host


=item monitor_refresh_ref [Required]; A reference to an array containing the list of monitors to be refreshed


=item monitor_check_ref [Optional]; A reference to an array containing the list of monitors
 whose timestamps need to checked if refresh is completed or not.
 If the $arguments{'monitor_refresh_ref'} points to a list ("all"),
meaning all monitors are to be refreshed, then this list becomes required and not optional.

=item wait_time [Optional]; Value of the number of minutes to wait for the monitors to be refreshed.
If no value is provided, the value $DEFAULT_WAIT_TIME is taken.

=item periodic_check_interval [Optional]; Value of the interval during which a check
is done to see if the time-stamps of the required monitors are refreshed.
If no value is provided, the value $DEFAULT_CHECK_INTERVAL is taken.

=back

=head2 Description

    This function refreshes a list of monitors on a given host.
	After running the 'dfm host discover' command, the function makes sure that the monitors
        are refreshed by checking the monitor timestamp values.
	If all the required monitors are refreshed on time, then success messages are logged.
	If it fails in any of the operation, then exceptions are thrown.
	This function internally calls the following APISets:
	dfm_host_discover() and "dfm query run" to get the timestamps.


=head2 Returns:

     None.

=head2 Exceptions:

    Throws the following exceptions directly:
	NACL::Exceptions::MissingArg, NACL::BaseException

	The exceptions can be thrown by the APISets too.

=cut

sub run_monitor {

    $Log->enter();

    #Get the argumemts
    my (%args) = @_;

    # Get the APISet object, fullname/IP of the host, monitor list to refresh,
    # list of monitor's timestamps to check, time to wait and the time interval to check
    my ($api_set_obj, $host_ip, $monitor_refresh_ref, $monitor_check_ref);
    my ($wait_time, $check_interval);

    $api_set_obj         = $args{'api_set_obj'};
    $host_ip             = $args{'host_ip'};
    $monitor_refresh_ref = $args{'monitor_refresh_ref'};
    $monitor_check_ref   = $args{'monitor_check_ref'};
    $wait_time           = $args{'wait_time'} || $DEFAULT_WAIT_TIME;
    $check_interval      = $args{'periodic_check_interval'}
      || $DEFAULT_CHECK_INTERVAL;


    $Log->debug("\nMonitor refresh list obtained: ". Dumper($monitor_refresh_ref) );
    $Log->debug("\nMonitor check list obtained: ". Dumper($monitor_check_ref) );

    my (@monitor_refresh_list, @monitor_check_list);

    #Copy the array
    if (defined $monitor_refresh_ref) {
        @monitor_refresh_list = @{$monitor_refresh_ref};
    } else {
        @monitor_refresh_list = ();
    }

    if (defined $monitor_check_ref) {
        @monitor_check_list = @{$monitor_check_ref};
    } else {
        @monitor_check_list = ();
    }

    # Variables to store the timestamps of monitors before & after refresh
    my (%monTimestamp);

    # If both monitor refresh list and monitor check list are empty, then throw an error.
    unless (@monitor_check_list or @monitor_refresh_list) {
        NACL::Exceptions::MissingArg->throw(
                          "Both Monitor check list and refresh list are missing"
                            . "Atleast one of them should be present");
    }

    # If monitor refresh list contains "all" and if monitor check list is empty,
    # then throw an error
    unless (@monitor_check_list) {

        if (    (@$monitor_refresh_ref == 1)
            and ($monitor_refresh_ref->[0] eq "all")) {
            NACL::Exceptions::MissingArg->throw(
                                   "Monitor check list is missing");

        } else {

            #Copy the array
            @monitor_check_list = @{$monitor_refresh_ref};
            $args{'monitor_check_ref'} = \@monitor_check_list;
        }
    } ## end unless (@monitor_check_list)

    # Put the keys and map it to null
    %monTimestamp = map { $_ => undef } @monitor_check_list;

    my $temp;

    # The final monitor list
    my $monitor = "";

    if (@monitor_refresh_list) {

        # Check if specific monitors are listed or all have to be refreshed
        if (    ($#monitor_refresh_list == 1)
            and ($monitor_refresh_list[0] eq "all")) {

            $monitor = "all";
        } else {
            my $i;
            $monitor = $monitor_refresh_list[0];

            for ($i = 1 ; $i < @monitor_refresh_list ; $i++) {

                # Make a comma separated list
                $monitor = $monitor . ",$monitor_refresh_list[$i]";
            }
        } ## end else [ if (($#monitor_refresh_list...

        #Response object
        my $response;

        # If specific monitors are to be refreshed, use the '-m' option. Else nothing.
        if ($monitor ne "all") {
            $response =
              $api_set_obj->dfm_host_discover('monitor-name' => $monitor,
                                              objects        => $host_ip);
        } else {
            $response = $api_set_obj->dfm_host_discover(objects => $host_ip);
        }

        $Log->debug("Response get_parsed_output:\n" . Dumper($response));

    } else {

        $Log->comment("\nMonitor refresh list was not provided,"
                 . "\nBut Monitor Checklist has been provided,"
                 . "\nHence will wait for @monitor_check_list timestamps "
                 . " to get refreshed, assuming that host discover has"
                 . " already been performed prior to calling this subroutine.");
    }

    my ($count, $flag);

    $count = 1;
    $flag  = 1;

    while ($count <= $wait_time) {

        #Sleep for $check_interval amount of seconds
        sleep($check_interval);

        %monTimestamp = %{_get_monitor_timestamps(%args)};

        # Check if the timestamps after refresh has null value
        $flag = 1;
        foreach $temp (keys(%monTimestamp)) {

            unless (defined $monTimestamp{$temp}) {
                $flag = 0;
                $Log->comment("Monitor still running for " . $temp);
            } else {
                $Log->comment("Monitor completed for " . $temp);
                delete $monTimestamp{$temp};
            }
        } ## end foreach $temp (keys(%monTimestamp...

        # If all timestamps refreshed, break out
        last if $flag;

        $count++;
    } ## end while ($count <= $wait_time)

    unless ($flag) {
        $Log->comment("Monitors not refreshed in $wait_time minutes");
        NATE::BaseException->throw(
                                "Monitors not refreshed in $wait_time minutes");
    } else {
        $Log->comment("Monitors refreshed in $count minutes");
    }

    $Log->exit();

} ## end sub run_monitor

###############################################################################

=head1 is_dfm_alive

        Checks if all the dfm services are alive or not.
        If "force" is set then the dfm services are forcefully started.

        Usage:
           DfmUtils::is_dfm_alive( %arguments );
        Examples:
           DfmUtils::is_dfm_alive(api_set_obj_dfm => $api_Set_Obj_dfm_DFM) ? undef : return $TCD::CONFIG;
           DfmUtils::is_dfm_alive(api_set_obj_dfm => $api_Set_Obj_dfm_DFM, force => 1);

=head2 Parameters:
=over

=item api_set_obj_dfm: [Required]; An object of the type NACL::APISet.

=item force [Optional]; If "force" is set to 1 then the dfm services are forcefully started.
                       (default value => 0).

=back

=head2 Description

    This function  validates wheather dfm is up and running on a client.
    If "force" is set to 1 then the dfm services are forcefully started.

        If it fails in any of the operation, then exceptions are thrown.
        This function internally calls the following APISets:
        dfm_service_list() and dfm_service_start().

=head2 Returns:

     returns 1 if all dfm services are running succesfully, else returns 0.

=head2 Exceptions:

    Throws the following exceptions directly: NATE::BaseException
    The exceptions can be thrown by the APISets too.

=cut

sub is_dfm_alive {

    #Get the argumemts DFM APISet object and force flag.
    my (%args) = @_;

    my ($response, $parsed_output, @services, $api_Set_Obj_dfm, $force);

    $api_Set_Obj_dfm = $args{'api_set_obj_dfm'};
    $force = $args{'force'} ? $args{'force'} : 0;

    if ($force == 1) {
        $api_Set_Obj_dfm->dfm_service_start();
    }

    $response =
        $api_Set_Obj_dfm->dfm_service_list();
    
    if ($response->get_raw_output() =~ /not\s+started/g) {
            $Log->comment("One or more of the services are in \"not started\" state");
            return 0;
        }
    
    return 1;
} ## end sub is_dfm_alive

###############################################################################

=head1 add_host

        Adds a host to a dfm. (7mode-Cmode release agnostic)

        Usage:
           DfmUtils::add_host( %arguments );

=head2 Parameters:
=over

=item api_set_obj_dfm: [Required]; An object of the type NACL::APISet.

=item api_set_obj_host: [Required]; An object of the type NACL::APISet.

=back

=head2 Description

    This functions adds a host to a dfm.
    If the host is already present then it simply re-uses it.
    In addition, it also makes shure that all snmp and host login credentials are properly set.

        If it fails in any of the operation, then exceptions are thrown.
        This function internally calls the following APISets:
        network_interface_show(), dfm_host_list(), security_snmpusers(),
        security_login_create(), dfm_snmp_list(), dfm_snmp_add(),
        dfm_host_add(), dfm_host_set(), dfm_host_get().

=head2 Returns:

     return 1 on passing, else throws respective exceptions.

=head2 Exceptions:

    Throws the following exceptions directly: NATE::BaseException
    The exceptions can be thrown by the APISets too.

=cut

sub add_host {

    #Get the argumemts DFM,Filer APISet objects.
    my (%args) = @_;

    my ($response,          $parsed_output,   $api_Set_Obj_filer,
        $host_present_flag, $host_ip,         $exception_object,
        $resp,              $str,             $pass,
        $hosttype,          $api_Set_Obj_dfm, $user);

    $api_Set_Obj_dfm   = $args{'api_set_obj_dfm'};
    $api_Set_Obj_filer = $args{'api_set_obj_host'};

    $host_present_flag = 0;

    $user =
        $api_Set_Obj_filer->{hostobj}
      ? $api_Set_Obj_filer->{hostobj}->{default_username}
      : $api_Set_Obj_filer->{connection}->{username};

    $pass =
        $api_Set_Obj_filer->{hostobj}
      ? $api_Set_Obj_filer->{hostobj}->{default_password}
      : $api_Set_Obj_filer->{connection}->{password};

    # Escaping any meta-chars in the password string
    $pass = quotemeta($pass);

    $hosttype =
        $api_Set_Obj_filer->{hostobj}
      ? $api_Set_Obj_filer->{hostobj}->{hosttype}
      : $api_Set_Obj_filer->{connection}->{hostp}->{hosttype};

    # get the ip based on the hosttype parameter defined in tharndoc.

    if ($hosttype =~ /ng/) {
        $response = $api_Set_Obj_filer->network_interface_show();
        $parsed_output = $response->get_parsed_output();
        foreach $str (@$parsed_output) {
            if ($str->{'role'} eq "cluster-mgmt") {
                $host_ip = $str->{'address'};
            }
        }
    } else {
        $host_ip =
            $api_Set_Obj_filer->{hostobj}->{default_ip}
          ? $api_Set_Obj_filer->{hostobj}->{default_ip}
          : $api_Set_Obj_filer->{connection}->{hostp}->{default_ip};
    }

    try {
        $api_Set_Obj_dfm->dfm_host_list(object => $host_ip);
        $Log->comment("The host already exists");
        $host_present_flag = 1;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        $exception_object = shift;
        unless ($exception_object->text =~ /There is no object named/) {
            NATE::BaseException->throw($exception_object);
        }
    };

    if ($hosttype =~ /ng/) {
        $response      = $api_Set_Obj_filer->security_snmpusers();
        $parsed_output = $response->get_parsed_output();
        my $public_flag = 0;
        foreach $str (@$parsed_output) {
            if ($str->{'username'} eq "public") {
                $public_flag = 1;
                last;
            }
        }
        if ($public_flag == 0) {
            $api_Set_Obj_filer->security_login_create(
                 'username'    => "public",
                 'application' => "snmp",
                 'authmethod'  => "community"
            );
        }
    } else {
        $response =
          $api_Set_Obj_filer->execute_raw_command(command => "snmp community");
        unless ($response =~ /public/) {
            $api_Set_Obj_filer->execute_raw_command(
                                     command => "snmp community add ro public");
        }
    }
    try {
        $api_Set_Obj_dfm->dfm_snmp_list("address-or-id" => $host_ip);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        $exception_object = shift;
        $resp             = $exception_object->get_response_object();
        if ($resp->get_raw_output() =~ /Could not find snmp network with address/) {
            $Log->comment("Adding $host_ip to snmp List");
            $api_Set_Obj_dfm->dfm_snmp_add(
                "pref-snmp-version" => 1,
                "community-name"    => "public",
                mask                => 32,
                address             => $host_ip
            );
        } else {
            NATE::BaseException->throw($exception_object);
        }
    };

    unless ($host_present_flag) {
        $api_Set_Obj_dfm->dfm_host_add('objects'       => $host_ip,
                                       'net-discovery' => 1);
    }
    $Log->comment("Setting login and password for Controller or cluster");
    $api_Set_Obj_dfm->dfm_host_set(
        object       => $host_ip,
        hostLogin    => $user,
        hostPassword => $pass
    );
    
    $Log->comment("Setting ndmp login for controller or cluster");
    $api_Set_Obj_dfm->dfm_host_set( 
            object=>$host_ip,
            hostNdmpLogin=>$user,
            hostNdmpPassword => $pass
        );
        
    $Log->comment("Checking set login credentials for the controller or cluster");
    $response = $api_Set_Obj_dfm->dfm_host_get(objects => $host_ip);
    $resp     = $response->get_parsed_output();
    $str      = $resp->[0];
    unless ($str->{'Login'} eq $user) {
        NATE::BaseException->throw("Failed to set host login credentials");
    }
    
    # Ensuring different authentications that need to be passed have passed
    $response = $api_Set_Obj_dfm->dfm_host_diag( object => "$host_ip" );
    unless(${$response->get_parsed_output()}[0]->{output}){
        $Log->comment("dfm host diag output parsing failed");
        NATE::BaseException->throw($response->get_parsed_output());
    }
    my $parsedOutput = ${$response->get_parsed_output()}[0]->{output};
    if( $parsedOutput =~ /SNMPv1\s+Passed/i ){
        $Log->comment("SNMPv1 authentication has passed");
    }else{
        $Log->comment("SNMPv1 authentication FAILED");
        NATE::BaseException->throw($parsedOutput);
    }
    
    if( $parsedOutput =~ /HTTP\s+Passed/i ){
        $Log->comment("HTTP authentication has passed");
    }else{
        $Log->comment("HTTP authentication FAILED");
        NATE::BaseException->throw($parsedOutput);
    }
    
    if( $parsedOutput =~ /SSH\s+Passed/i ){
        $Log->comment("SSH authentication has passed");
    }else{
        $Log->comment("SSH authentication FAILED");
        NATE::BaseException->throw($parsedOutput);
    }
    
    if( $parsedOutput =~ /XML.*Passed/i ){
        $Log->comment("XML authentication has passed");
    }else{
        $Log->comment("XML authentication FAILED");
        NATE::BaseException->throw($parsedOutput);
    }
    
    if( $parsedOutput =~ /perfAdvisorEnabled\s+Passed/i ){
        $Log->comment("perfAdvisorEnabled authentication has passed");
    }else{
        $Log->comment("perfAdvisorEnabled authentication FAILED");
        NATE::BaseException->throw($parsedOutput);
    }
    
    if( $parsedOutput =~ /hostType\s+Passed/i ){
        $Log->comment("hostType authentication has passed");
    }else{
        $Log->comment("hostType authentication FAILED");
        NATE::BaseException->throw($parsedOutput);
    }
    
    if( $parsedOutput =~ /hostRevision\s+Passed/i ){
        $Log->comment("hostRevision authentication has passed");
    }else{
        $Log->comment("hostRevision authentication FAILED");
        NATE::BaseException->throw($parsedOutput);
    }
    
    if( $parsedOutput =~ /hostLogin\s+Passed/i ){
        $Log->comment("hostLogin authentication has passed");
    }else{
        $Log->comment("hostLogin authentication FAILED");
        NATE::BaseException->throw($parsedOutput);
    }
    
    if( $parsedOutput =~ /perfAdvisorTransport\s+Passed|perfAdvisorTransport set to httpOnly/i ){
        $Log->comment("perfAdvisorTransport authentication has passed");
    }else{
        $Log->comment("perfAdvisorTransport authentication FAILED");
        NATE::BaseException->throw($parsedOutput);
    }

    #Validating the presence of the host in DFM
    $api_Set_Obj_dfm->dfm_host_list(object => $host_ip);
    open(FH, "> host_settings.cfg");
    print FH "$host_ip=>$host_present_flag";
    close FH;
    return 1;
    
} ## end sub add_host

###############################################################################


=head1 user_create

        Create a User at Dfm machine. (7mode-Cmode release agnostic)

        Usage:
           DfmUtils::user_create( %arguments );

=head2 Parameters:
=over

=item api_set_obj_dfm: [Required]; An object of the type NACL::APISet.

=item api_set_obj_host: [Required]; An object of the type NACL::APISet.

=back

=head2 Description

    This functions create a user in dfm.

        If it fails in any of the operation, then exceptions are thrown.
        This function internally calls the following APISets:
        network_interface_show(), dfm_host_list(), security_snmpusers(),
        security_login_create(), dfm_snmp_list(), dfm_snmp_add(),
        dfm_host_add(), dfm_host_set(), dfm_host_get().

=head2 Returns:

     return 1 on passing, else throws respective exceptions.

=head2 Exceptions:

    Throws the following exceptions directly: NATE::BaseException
    The exceptions can be thrown by the APISets too.

=cut

sub user_create {

    #Get the argumemts DFM,Filer APISet objects.
    my (%args) = @_;

    my ($response,          $parsed_output,   $api_Set_Obj_filer,
        $host_present_flag, $host_ip,         $exception_object,
        $resp,        $api_Set_Obj_dfm,  $Dfm_user,  $Dfm_pass);

    $api_Set_Obj_dfm	= $args{'api_set_obj_dfm'};
    $Dfm_user 			= $args{'Dfm_user'};
	$Dfm_pass 			= $args{'Dfm_pass'};
    
	if ($api_Set_Obj_dfm->{connection}->{hostp}->{hosttype} =~ /linux|unix/i) {
		$response =
		  $api_Set_Obj_dfm->execute_raw_command(command => "useradd $Dfm_user");
		$response =
		  $api_Set_Obj_dfm->execute_raw_command(
                            command     => "passwd $Dfm_user",
               'connectrec-match_table' =>[
                                "New " => "$Dfm_pass",
                                "Reenter New Password" => "$Dfm_pass",
								"Retype new UNIX password" => "$Dfm_pass",
								"Retype new password" => "$Dfm_pass",
								"New password" => "$Dfm_pass"
                            ]);

		if ($response =~	/passwd:\s+all\s+authentication\s+tokens\s+updated\s+successfully/  
            || $response =~	/Password changed/) {
			$Log->comment(
				 "The password pradee $Dfm_pass is set to the user $Dfm_user");
			return 1;
		}
		else {  
			$Log->comment(
				 "The password $Dfm_pass is set to the user $Dfm_user");
			return 0;
		}
	} elsif ($api_Set_Obj_dfm->{connection}->{hostp}->{hosttype} =~ /windows/i)
	{
		$response = $api_Set_Obj_dfm->execute_raw_command(
							command => "net user $Dfm_user $Dfm_pass /add");
		if ($response !~ /The\s+command\s+completed\s+successfully/) {
			$Log->comment(
				 "The password $Dfm_pass is not set to the user $Dfm_user");
			return 0;
		}
		$api_Set_Obj_dfm->execute_raw_command(
							command => "net localgroup Administrators $Dfm_user /add");
	}
    return 1;
} ## end sub user_create

###############################################################################


=head1 delete_host

        Deletes a host from the dfm. (7mode-Cmode release agnostic)
        This functions deletes the host which was added through add_host() subroutine

        Usage:
           DfmUtils::delete_host( api_set_obj_dfm => $APi_set_obj_DFM);

=head2 Parameters:
=over

=item api_set_obj_dfm: [Required]; An object of the type NACL::APISet.

=back

=head2 Description

    This functions deletes the host which was added through add_host() subroutine.
    If the host was reused in add_host() then the subroutine will not delte it.

        If it fails in any of the operation, then exceptions are thrown.
        This function internally calls the following APISets:
        dfm_host_delete(), dfm_service_start(), dfm_service_stop().

=head2 Returns:

     return 1 on passing, else throws respective exceptions.

=head2 Exceptions:

    Throws the following exceptions directly: NATE::BaseException
    The exceptions can be thrown by the APISets too.

=cut

sub delete_host {

    #Get the argumemts DFM,Filer APISet objects.
    my (%args) = @_;

    my ($api_Set_Obj_dfm, @arr, $host_present_flag, $host_ip);

    $api_Set_Obj_dfm = $args{'api_set_obj_dfm'};

    open(FH, "< host_settings.cfg");
    @arr = <FH>;
    close FH;
    unlink("host_settings.cfg");

    ($host_ip, $host_present_flag) = $arr[0] =~ /([\d.]+)=>(\S+)/;

    unless ($host_present_flag) {
        $api_Set_Obj_dfm->dfm_service_stop();
        $api_Set_Obj_dfm->dfm_service_start('service-names' => "sql");
        $api_Set_Obj_dfm->dfm_host_delete('force'  => 1,
                                          'object' => $host_ip);
        $api_Set_Obj_dfm->dfm_service_start();
    }
} ## end sub delete_host

##############################################################################

=head1 get_aggrs

        Find aggregates from given filer based on filter conditions.
        If enough aggrs re not found then create the remaing.

        Example:

           $Node = NACL::C::Node->find();

           $Flex_aggr_obj = DfmUtils::get_aggrs("count"    => 1,
                                                count_type => 'atmost',
                                                node       => $Node,
                                                filter     => {
                                                           "volume-style" => "flex",
                                                           "state"        => "online",
                                                           "root"         => "false" }
                                               );

          ( or )

           @Flex_aggrs = DfmUtils::get_aggrs("count"    => 3,
                                             count_type => 'atleast',
                                             node       => $Node,
                                             filter     => {
                                                        "volume-style" => "flex",
                                                        "state"        => "online",
                                                        "root"         => "false" }
                                            );


=head2 Parameters:
=over

=item count: [Required]; Total no of Aggregates required.

=item node: [Required]; NACL::C::Component::command_interface object.
                        usually it is a NACL::C::Node object.

=item count_type: [Optional]; Default: atleast
     atmost : if atmost is specified as count type then if get_aggrs() is querried for
              $count(ex: 4) aggrs then 4 aggrs will be returned irrespective of how many are
              found on the filer.
     atleast: if atleast is specified as count type then if get_aggrs() is querried for
              ex: 4 aggrs then minimum of 4 aggrs will be returned .
                  if no of aggrs on filer < 4 then 4 will be retured
                  if no of aggrs on filer > 4 then all the aggrs found will be returned
     force  : if force is specified then aggregates wont be querried.
              $count no of aggrs are forcibly created and returned.

=item filter: [Optional]; Filters which can be applied which querring for aggregates.
                          refer NACL::C::Aggregate.pm for more details.

=back

=head2 Description

    This functions querries for aggregates which fit the filter settings and returns them.
    If the aggrs are already present then it simply re-uses them.
    else it create the aggregates and returns them.

        If it fails in any of the operation, then exceptions are thrown.
        This function internally calls the following Component subroutines.
        NACL::C::Aggregate->find, NACL::C::Aggregate->create
        NACL::C::Aggregate->state->state().

=head2 Returns:

     returns a array or scalar depending on the invoking scirpt.
     refer want_array() from perl docs.

=head2 Exceptions:

    Throws the following exceptions directly: NATE::BaseException
    The exceptions can be thrown by the NACL Componenets too.

=cut

sub get_aggrs {

    my ($aggregate, $time, $aggr_name, $diff, $i, @aggrs);

    my %opts       = @_;
    my $count      = $opts{count};
    my $filter     = $opts{filter};
    my $node       = $opts{node};
    my $count_type = $opts{count_type};
    my $disk_count = 3;

    unless ($count_type eq "force") {
        @aggrs =
          NACL::C::Aggregate->find(command_interface => $node,
                                   filter            => $filter,
                                   allow_empty       => 1);
    }

    if (@aggrs < $count) {
        if ($count_type eq "force") {
            @aggrs = ();
            $diff  = $count;
        } else {
            $diff = $count - @aggrs;
        }
        $i = 0;
        while ($i < $diff) {
            $aggr_name = "_Auto_dfm_aggr" . $$ . $i;
            $aggregate =
              NACL::C::Aggregate->create(
                                     command_interface => $node,
                                     aggregate         => $aggr_name,
                                     'diskcount'       => $disk_count,
                                     'volume-style' => $filter->{"volume-style"}
              );
            push(@aggrs, $aggregate);
            $i++;
        } ## end while ($i < $diff)
        foreach $aggregate (@aggrs) {
            $time = 0;
            while ($aggregate->state()->state() !~ /online/ && $time < 1800) {
                sleep(60);
                $time += 60;
            }

            if ($time >= 1800) {
                NATE::BaseException->throw(
                                 "Script timed out while waiting for aggregate "
                                   . $aggregate->aggregate()
                                   . " to come online ");
            }
        } ## end foreach $aggregate (@aggrs)
    } ## end if (@aggrs < $count)
    $count--;

    if ($count_type eq 'atmost') {
        return (wantarray ? @aggrs[0 .. $count] : $aggrs[0]);
    } else {
        return (wantarray ? @aggrs : $aggrs[0]);
    }

} ## end sub get_aggrs

##############################################################################

=head1 get_vservers

        Find vservers from given filer based on filter conditions.
        If enough vservers are not found then create the remaing.

        Example:

           $Node = NACL::C::Node->find();

           @Vservers = DfmUtils::get_vservers(count      => 2,
                                              node       => $Node,
                                              count_type => 'atmost',
                                              aggr       => $Flex_aggr_obj->aggregate()
                                             );

=head2 Parameters:
=over

=item count: [Required]; Total no of Vservers required.

=item node: [Required]; NACL::C::Component::command_interface object.
                        usually it is a NACL::C::Node object.

=item count_type: [Optional]; Default: atleast
     atmost : if atmost is specified as count type then if get_vservers() is querried for
              $count(ex: 4) vservers then 4 vservers will be returned irrespective of how many are
              found on the filer.
     atleast: if atleast is specified as count type then if get_vservers() is querried for
              ex: 4 vservers then minimum of 4 vservers will be returned .
                  if no of vservers on filer < 4 then 4 will be retured
                  if no of vservers on filer > 4 then all the vservers found will be returned
     force  : if force is specified then vservers wont be querried.
              $count no of vservers are forcibly created and returned.

=item filter: [Optional]; Filters which can be applied which querring for vservers.
                          refer NACL::C::Vserver.pm for more details.

=back

=head2 Description

    This functions querries for vservers which fit the filter settings and returns them.
    If the vservers are already present then it simply re-uses them.
    else it create the vservers and returns them.

        If it fails in any of the operation, then exceptions are thrown.
        This function internally calls the following Component subroutines.
        NACL::C::Vserver->find, NACL::C::Vserver->create

=head2 Returns:

     returns a array or scalar depending on the invoking scirpt.
     refer want_array() from perl docs.

=head2 Exceptions:

    Throws the following exceptions directly: NATE::BaseException
    The exceptions can be thrown by the NACL Componenets too.

=cut

sub get_vservers {

    my ($vserver_name, $root_vol_name, $i, $diff, $vserver, @vservers);

    my %opts       = @_;
    my $count      = $opts{count};
    my $node       = $opts{node};
    my $aggr_name  = $opts{aggr};
    my $count_type = $opts{count_type};

    unless ($count_type eq "force") {
        @vservers =
          NACL::C::Vserver->find(command_interface => $node,
                                 filter            => {type => "cluster"},
                                 allow_empty       => 1);
    }

    if (@vservers < $count) {
        if ($count_type eq "force") {
            $diff     = $count;
            @vservers = ();
        } else {
            $diff = $count - @vservers;
        }
        $i = 0;
        while ($i < $diff) {
            $vserver_name  = "_Auto_dfm_vserver" . $$ . $i;
            $root_vol_name = $vserver_name . "root_vol";

            $vserver =
              NACL::C::Vserver->create(command_interface => $node,
                                       vserver           => $vserver_name,
                                       rootvolume        => $root_vol_name,
                                       aggregate         => $aggr_name,
                                       "ns-switch"       => "file",
                                       "rootvolume-security-style" => "mixed",);
            push(@vservers, $vserver);
            $i++;
        } ## end while ($i < $diff)
    } ## end if (@vservers < $count)
    $count--;

    if ($count_type eq 'atmost') {
        return (wantarray ? @vservers[0 .. $count] : $vservers[0]);
    } else {
        return (wantarray ? @vservers : $vservers[0]);
    }

} ## end sub get_vservers

##############################################################################

=head1 get_volumes

        Find volumes from given filer based on filter conditions.
        If enough volumes are not found then create the remaing.

        Example:

           $Node = NACL::C::Node->find();

           @volumes = DfmUtils::get_volumes(count      => 1,
                                            count_type => 'atleast',
                                            node       => $Node,
                                            filter     => {
                                                     "volume-style" => "flex",
                                                     "state"        => "online",
                                                     vserver => $vserver->vserver()
                                            },
                                            vserver => $vserver->vserver(),
                                            aggr    => $Flex_aggr_obj->aggregate()
                                           );

=head2 Parameters:
=over

=item count: [Required]; Total no of Volumes required.

=item node: [Required]; NACL::C::Component::command_interface object.
                        usually it is a NACL::C::Node object.

=item count_type: [Optional]; Default: atleast
     atmost : if atmost is specified as count type then if get_volumes() is querried for
              $count(ex: 4) volumes then 4 volumes will be returned irrespective of how many are
              found on the filer.
     atleast: if atleast is specified as count type then if get_volumes() is querried for
              ex: 4 volumes then minimum of 4 volumes will be returned .
                  if no of volumes on filer < 4 then 4 will be retured
                  if no of volumes on filer > 4 then all the volumes found will be returned
     force  : if force is specified then volumes wont be querried.
              $count no of volumes are forcibly created and returned.

=item filter: [Optional]; Filters which can be applied which querring for volumes.
                          refer NACL::C::Volume.pm for more details.

=back

=head2 Description

    This functions querries for volumes which fit the filter settings and returns them.
    If thevolumes are already present then it simply re-uses them.
    else it create the volumes and returns them.

        If it fails in any of the operation, then exceptions are thrown.
        This function internally calls the following Component subroutines.
        NACL::C::Volume->find, NACL::C::Volume->create

=head2 Returns:

     returns a array or scalar depending on the invoking scirpt.
     refer want_array() from perl docs.

=head2 Exceptions:

    Throws the following exceptions directly: NATE::BaseException
    The exceptions can be thrown by the NACL Componenets too.

=cut

sub get_volumes {

    my ($vol_name, $i, $diff, $volume, @volumes);

    my %opts       = @_;
    my $count      = $opts{count};
    my $count_type = $opts{count_type};
    my $node       = $opts{node};
    my $filter     = $opts{filter};
    my $vserver    = $opts{vserver};
    my $aggregate  = $opts{aggr};

    unless ($count_type eq "force") {
        @volumes = NACL::C::Volume->find(command_interface => $node,
                                         filter            => $filter,
                                         allow_empty       => 1);
    }

    if (@volumes < $count) {
        if ($count_type eq "force") {
            $diff    = $count;
            @volumes = ();
        } else {
            $diff = $count - @volumes;
        }
        $i = 0;
        while ($i < $diff) {
            my $job;

            $vol_name = "_Auto_dfm_vol" . $filter->{"volume-style"} . $$ . $i;

            if ($filter->{"volume-style"} eq "flex") {
                my $size = "20M";
                $size = $filter->{"size"} if (exists($filter->{"size"}));
                $volume =
                  NACL::C::Volume->create(command_interface => $node,
                                          aggregate         => $aggregate,
                                          vserver           => $vserver,
                                          volume            => $vol_name,
                                          size              => $size,
                                          job_component     => \$job,);
            } else {
                $volume =
                  NACL::C::Volume->create(command_interface => $node,
                                          aggregate         => $aggregate,
                                          vserver           => $vserver,
                                          volume            => $vol_name,
                                          "stripe-format"   => "full",
                                          job_component     => \$job,);
            }
            $job->watch_progress("method-timeout" => 1800);
            push(@volumes, $volume);
            $i++;
        } ## end while ($i < $diff)
    } ## end if (@volumes < $count)
    $count--;

    if ($count_type eq 'atmost') {
        return (wantarray ? @volumes[0 .. $count] : $volumes[0]);
    } else {
        return (wantarray ? @volumes : $volumes[0]);
    }

} ## end sub get_volumes

=head1 get_job_status

	Runs a dfpm job detail cmd for a particular job id and check if
        the job gets completed and returns the values of job_state and job_status respectively .

	Usage:

	DfmUtils::get_job_status( %arguments );

=head2 Parameters:

=over

=item api_set_obj: [Required]; An object of the type NACL::APISet

=item job_id: [Required]; The provisioning job id.

=item wait_time [Optional]; Value of the number of minutes to wait
         for provisioning job to get completed. If no value is provided,
         the localized value 45min of $DEFAULT_WAIT_TIME is taken.

=item periodic_check_interval [Optional];
        Value of the interval during which a check is done to see if the
        job gets completed. If no value is provided,
        the localized value 60Sec of $DEFAULT_CHECK_INTERVAL is taken.
=back

=head2 Description

    This function is to get the provisioined job status.
    This function basically executes 'dfm job detail <job-id>' command periodically,
    untill either job state or job status got values completed and not in_progress respectively,
    and then it returns those values respectively.
    If the expected output values not get updated within the grace period then also
    function returns values of 'job state' and 'job status' respectively.
	This function internally calls the following APISets:
	dfpm_job_detail().

=head2 Returns:

     Returns the values of job_state and job_status respectively.

=head2 Exceptions:

	The exceptions can be thrown by the APISet.

=cut

###############################################################################
# Function to find any ProvMgr job status
###############################################################################
sub get_job_status () {

    # Function call should have DFM server object, job_id, wait_time
    #  in minutes and periodic_check_interval in seconds as arguments
    my (%args) = @_;

    #localizing the global variables.
    local $DEFAULT_WAIT_TIME      = 45;
    local $DEFAULT_CHECK_INTERVAL = 60;

    my $wait_time = $args{'wait_time'} || $DEFAULT_WAIT_TIME;
    my $check_interval = $args{'periodic_check_interval'}
      || $DEFAULT_CHECK_INTERVAL;
    my $api_set_obj = $args{'api_set_obj'};
    my $job_id      = $args{'job_id'};

    my ($response, @parsed_output);
    my $count = 1;

    while ($count <= $wait_time) {

        #executing the job detail command to check the status of the job
        $response = $api_set_obj->dfpm_job_detail("job-id" => $job_id);
        @parsed_output = @{$response->get_parsed_output()};

        # checks either job state or job status got values completed and
        # not in_progress respectively then function returns those values.

	if ( $parsed_output[0]->{"Job State"} =~ /completed/ ) {
		$Log->comment("Job completed in $count minutes");
		return ( $parsed_output[0]->{"Job State"}, $parsed_output[0]->{"Job Status"} );
	}
	sleep($check_interval);
	$count++;
    }
    $Log->comment("Job not completed in $wait_time minutes");
    return ( $parsed_output[0]->{"Job State"}, $parsed_output[0]->{"Job Status"} );
}

##################################################
#Function to get free filer Ip's
##################################################

=head1 free_filer_ip 

        Provide one filer IP to genetrate a set of 40 IP's and check
        their availability to create the vsim..        

        Usage:

        DfmUtils::free_filer_ip( %arguments );

=head2 Parameters:

=over

=item : api_set_obj[Required]; An object of the type NACL::APISet

=item : filer_ip [Required]; The IP of the host

=head2 Description

    This function is used to get free filer ip's.
    This function genertaes 40 different array of filer ip's by incrementing and decrementing 20
    from the given filer ip in the function call.
    This function basically executes 'nslookup filer_ip' command periodically
    to check for the filer ip usage,
    If the particular filer ip is not in use then it pushes the particular filer ip in the 
    array.  

=head2 Returns:

     Returns the array of free filer ip's which is not in use .

=cut

sub free_filer_ip {

    my @array;
    my @free_ips;
    my $new_ip ;

    #Get the argumemts
    my (%args) = @_;
    my $api_set_obj = $args{'api_set_obj'};
    my $filer_ip     = $args{'filer_ip'};

    foreach (-20 .. 20) {
       my @arr = split('\.',$filer_ip);
       $arr[3] += $_;
       #Enetr the filer ip which ends or greater that 128
       if ( ($arr[0] > 255) || ($arr[1] > 255) || ($arr[2] > 255) || ($arr[3] > 255) ||
            ($arr[0] < 1) || ($arr[1] < 1) || ($arr[2] < 1) || ($arr[3] < 128) ) {
           next;
        }
       my $new_ip = join('.',@arr);
       #@array holds 40 different filer ip's generated from the given ip
       push(@array, $new_ip);

    }

    foreach (@array)
    {
         my $response = $api_set_obj->execute_raw_command( "command" => "nslookup $_"  );
         if ( $response =~ /server can\'t find/
              || $response =~ /Non\-existent domain/ ) {
            push(@free_ips,$_);
          }
    }
    return @free_ips;
}

sub initialize_DFM {

    #Get the argumemts DFM APISet object 
    my (%args) = @_;
    my $sqlServiceStatus;
    my $timeout = 0;

    my $api_Set_Obj_dfm = $args{'api_set_obj_dfm'};
    
    if($args{'mode'} =~ /7mode/i){
        $api_Set_Obj_dfm->dfm_option_set( 
            'discoverNetworks'  => '0',
            'discoverHosts'     => '0',
            'discoverAgents'    => '0',
        );    
    }else{
        $api_Set_Obj_dfm->dfm_option_set(
        'discoverNetworks'  => '0',
        'discoverHosts'     => '0',
        'discoverAgents'    => '0',
        'discoverClusters'   => '0'
        );
    }
    
    $api_Set_Obj_dfm->dfm_service_stop();
    
    
    do{
        $timeout++;
        sleep 60;
        $sqlServiceStatus = $api_Set_Obj_dfm->dfm_service_list('s'=>"sql");
        if($timeout == 15){
            $Log->comment("SQL service did not stop in specified time, exiting");
            $timeout = 0;
            exit 1;
        }
    }while($sqlServiceStatus->get_raw_output() !~ /sql: not started/i);
    
    $api_Set_Obj_dfm->dfm_service_start('service-names' => "sql");
    
    do{
        $timeout++;
        sleep(60);
        $sqlServiceStatus = $api_Set_Obj_dfm->dfm_service_list('s'=>"sql");
        if($timeout == 15){
            $Log->comment("SQL service did not start in specified time, exiting");
            $timeout = 0;
            exit 1;
        }
    }while($sqlServiceStatus->get_raw_output() !~ /sql: started/i);
    
    
    $api_Set_Obj_dfm->dfm_network_delete('force'  => "all");
    
    if($args{'mode'} =~ /7mode/i){
        $api_Set_Obj_dfm->execute_raw_command( "command" => "dfm host delete -f all");    
    }else{
        $api_Set_Obj_dfm->execute_raw_command( "command" => "dfm cluster delete -f all");
    }
    
    $api_Set_Obj_dfm->dfm_service_start();
    
    sleep 90;
    
    return 1;
} ## end sub initialize_DFM

sub install_plugin {

    #Get the argumemts DFM APISet object 
    my (%args) = @_;
    my $plugins = 0;
    my $dfmbuild;
    my $api_Set_Obj_dfm = $args{'api_set_obj_dfm'};
    my $filer_ontap_version = $args{'version'};
    my $dfm_plugin_version;
    
    my %codeLine = ( 
          'ironcity'     => '7.3.7',
          'boilermaker'  => '8.0',
          'Rbr'          => '8.0',
          'rollingrock'  => '8.1',
          'rrollingrock' => '8.1',
          'sierranevada' => '8.2',
          'rsierranevada'=> '8.2',
          '8.2'          => '8.2',
          'main_ib_bib'  => '8.2',
          'main'         => '8.2',
          'fullsteam'    => '8.3',
          'longboard'    => '8.4',
          'R8.1'         => '8.1',
          'R8.0'         => '8.0',
          '8.1.4'        => '8.1',
          '8.1.3'        => '8.1',
          '8.1.2'        => '8.1',
          '8.1.1'        => '8.1',
    );
    
    foreach my $release ( keys  %codeLine ){
        $Log->comment("Filer versions compare: ONTAP[$filer_ontap_version] DFM[$release]");
        if( $filer_ontap_version  =~ /$release/i ){
            $dfm_plugin_version = $codeLine{$release};
        }
    }
    return 0 unless($dfm_plugin_version); # If the content of plugin version is null then version detection failed
    $Log->comment("Version of ONTAP found on Controller: [$dfm_plugin_version]");
    my $resp = $api_Set_Obj_dfm->dfm_about(); 
    my @plugins_in_dfm = (keys %{$resp->get_parsed_output()->[0]->{'Installed Plugins'}});
    ($dfmbuild) = $resp->get_parsed_output()->[0]->{'Version'} =~ /[^\(]*\((.[^)]*)\)/gs;
    $Log->comment("DFM build found in dfm_about command is: [$dfmbuild]");
    # Compare plugins in DFM with plugin required for Controller to be discoverd
    foreach (@plugins_in_dfm){ 
        if ($_ =~ /$dfm_plugin_version/){
            $Log->comment("Found plugin needed to discover controller in the OCUM server");
            $plugins = 1;                
        }
    }
   $Log->comment("Result of plugin found in DFM is: [$plugins]");
   if( $plugins eq 0){
    	if( $dfm_plugin_version =~ /7\./ ){
    		if( $resp->get_parsed_output()->[0]->{'Operating System'} =~ /linux/i ){
    			if( $resp->get_parsed_output()->[0]->{'Operating System'} =~/x86_64/i ){  # LINUX 64 bit  7.*.*
    				$api_Set_Obj_dfm->dfm_plugin_add(
    					'path-or-url' =>"http://web.netapp.com/engineering/rlse/ontap/$dfm_plugin_version/prod/common/dfm/Release/Linux64/libfilerconfig/filerconfig_Linux64.zip",
    					'force-upgrade'=> "1",
    				);
    			}else{  # LINUX 32 but 7.*.*
    				$api_Set_Obj_dfm->dfm_plugin_add(
    					'path-or-url' =>"http://web.netapp.com/engineering/rlse/ontap/$dfm_plugin_version/prod/common/dfm/Release/Linux/libfilerconfig/filerconfig_Linux.zip",
    					'force-upgrade'=> "1",
    				);
    			}
    		}else{
    			if( $resp->get_parsed_output()->[0]->{'Operating System'} =~/x86_64/i ){ # WINDOWS 64 bit 7.*.*
    				$api_Set_Obj_dfm->dfm_plugin_add(
    					'path-or-url' =>"http://web.netapp.com/engineering/rlse/ontap/$dfm_plugin_version/prod/common/dfm/Release/Windows64/libfilerconfig/filerconfig_Windows64.zip",
    					'force-upgrade'=> "1",
    				);
    			}else{ # WINDOWS 32 bit 7.*.*
    				$api_Set_Obj_dfm->dfm_plugin_add(
    					'path-or-url' =>"http://web.netapp.com/engineering/rlse/ontap/$dfm_plugin_version/prod/common/dfm/Release/Windows/libfilerconfig/filerconfig_Windows.zip",
    					'force-upgrade'=> "1",
    				);
    			}
    		}
    	}elsif($dfm_plugin_version =~ /8\./){
    		if( $resp->get_parsed_output()->[0]->{'Operating System'} =~ /linux/i ){   
    			if( $resp->get_parsed_output()->[0]->{'Operating System'} =~/x86_64/i ){ # LINUX 64 bit 8.*
    				$api_Set_Obj_dfm->dfm_plugin_add(
    					'path-or-url' =>"http://web.netapp.com/engineering/rlse/DOT/$dfm_plugin_version/ontap/prod/common/dfm/Release/Linux64/libfilerconfig/filerconfig_Linux64.zip",
    					'force-upgrade'=> "1",
    				);
    			}else{   # LINUX 32 bit  8.*
    				$api_Set_Obj_dfm->dfm_plugin_add(
    					'path-or-url' =>"http://web.netapp.com/engineering/rlse/DOT/$dfm_plugin_version/ontap/prod/common/dfm/Release/Linux/libfilerconfig/filerconfig_Linux.zip",
    					'force-upgrade'=> "1",
    				);
    			}
    		}else{
    			if( $resp->get_parsed_output()->[0]->{'Operating System'} =~/x86_64/i ){  # WINDOWS 64 bit 8.*
    				$api_Set_Obj_dfm->dfm_plugin_add(
    					'path-or-url' =>"http://web.netapp.com/engineering/rlse/DOT/$dfm_plugin_version/ontap/prod/common/dfm/Release/Windows64/libfilerconfig/filerconfig_Windows64.zip",
    					'force-upgrade'=> "1",
    				);
    			}else{ #WINDOWS 32 bit 8.*
    				$api_Set_Obj_dfm->dfm_plugin_add(
    					'path-or-url' =>"http://web.netapp.com/engineering/rlse/DOT/$dfm_plugin_version/ontap/prod/common/dfm/Release/Windows/libfilerconfig/filerconfig_Windows.zip",
    					'force-upgrade'=> "1",
    				);
    			}
    		}
    	}else{
    		$Log->comment("ERROR: Could not detect Controller version");
    		print Dumper \$resp;
    		return 0;
    	}
    }else{
    	$Log->comment("Filer plugin already present in DFM");
    }
    return 1;
    
} ## end sub install_plugin

###############################################################################

=head1 edit_filerVersion 

        Edits the /opt/NTAPdfm/conf/filerversions.conf file.

        Usage:
           DfmUtils::edit_filerVersion ( %arguments );

=head2 Parameters:
=over

=item api_set_obj_dfm: [Required]; An object of the type NACL::APISet.

=item api_set_obj_host: [Required]; An object of the type NACL::APISet.

=item filer: [Required]; Filer name or IP.

=back

=head2 Description

    This functions edits the file /opt/NTAPdfm/conf/filerversions.conf in the DFM
      with the filerversion to the corresponding filer version id

        If it fails in any of the operation, then exceptions are thrown.

=head2 Returns:

     return 1 on passing, else throws respective exceptions.

=head2 Exceptions:

    Throws the following exceptions directly: NATE::BaseException
    The exceptions can be thrown by the APISets too.

=cut

sub edit_filerVersion {

    #Get the argumemts DFM,Filer APISet objects.
    my (%args) = @_;

    my ($response,          $parsed_output,   $api_Set_Obj_filer,
        $host_present_flag, $host_ip,         $exception_object,
        $resp,              $str,             $pass,
        $hosttype,          $api_Set_Obj_dfm, $user,
        $hostversion,       $hostid,          $filerversions,
        $ins_dir,           $filer);

    $api_Set_Obj_dfm   = $args{'api_set_obj_dfm'};
    $api_Set_Obj_filer = $args{'api_set_obj_host'};
    $filer = $args{'filer'};

    my %filerType = ( 'ironcity'     => '7.3',
                      'boilermaker'  => '8.0',
                      'Rbr'          => '8.0',
                      'rollingrock'  => '8.1',
                      'sierranevada' => '8.2',
                      'main_ib_bib'  => '8.2',
                      'main'         => '8.2',
                      'fullsteam'    => '8.3',
                      'longboard'    => '8.4',
                    );

    $resp=$api_Set_Obj_filer->execute_raw_command(command => 'version');
    $resp =~ s/(NetApp Release |Data ONTAP Release )//;
    $resp =~ s/:.*//;
    chomp($resp);
    $hostversion = (split /\s+/,$resp)[0];
    chomp($hostversion);
    $Log->log("Filer version with mode : $resp");
    $Log->log("Filer version : $hostversion");
    my $resp1 = $hostversion;

    foreach ( keys ( %filerType) ) {
        if ( $hostversion =~ /$_/i ) {
            $hostversion = $filerType{$_};
        }
    }

    if ( $hostversion =~ /R*([\d.\d]*)\w*.*/) {
        $hostversion = $1;
    }
    if ( $hostversion =~ /(\d.\d).*/) {
        $hostversion = $1;
    }
    $Log->log("Filer version ID : $hostversion");

    $response = $api_Set_Obj_dfm->execute_raw_command(command => "dfm diag");
    ($ins_dir) = $response =~ /Installation\s*Directory\s*([^\n\r]+)/;
    $Log->log("\n\nInstallation Directory : $ins_dir====");

    $filerversions = "$resp1" . "=" . "$hostversion";

    $api_Set_Obj_dfm->execute_raw_command(command =>
        "echo $filerversions >> \"$ins_dir/conf/filerversions.conf\"");

    $response = $api_Set_Obj_dfm->dfm_host_list( "objects" => $filer);
    $parsed_output = $response->get_parsed_output();
    $hostid = $parsed_output->[0]->{'ID'};
    $Log->log("HOST_ID : $hostid");
    $api_Set_Obj_dfm->execute_raw_command(command => "dfm query run -f \"update hosts SET hostOSVersion = $hostversion WHERE hostId=$hostid\"");
    $response = $api_Set_Obj_dfm->execute_raw_command(command => "dfm query run -f \"select hostOSVersion from hosts WHERE hostid=$hostid\"");

return 1;
}


##################################################################################################
# Method    :  increase_vfiler_limit
# Objective :  Increse vfiler creation limit to maximum allowed by the controller
# Input     :  Reference to a hash containing: The APISet object to Controller/Filer, filer instance
#              read using param(FILER)                
# Output    :  0 on SUCCESS 1 on FAILURE
# Exceptions: None thrown directly.
# Call      : increase_vfiler_limit($Api_Set_Obj_FILER,$FILER);
###################################################################################################
sub increase_vfiler_limit{
    
    my ($Api_Set_Obj_FILER,$FILER) = @_;
    my $filer = NACL::C::Node->new(name => $FILER);
    my ($current_limit_vfiler,$current_in_use_vfiler,$max_vfiler_limit,$ping,$platform_hard_limit) = (0,0,0,0,0);
    my $loop = 0;
    my $ping_count;
    my $ping_flag;
    my $response = '';
    my $parsed_output = '';
    $Log->log("Checking validity of input object");
    return 1 unless($Api_Set_Obj_FILER);
    $Log->log("Increasing vfiler count to maximum allowed for controller");
    $response = $Api_Set_Obj_FILER->vfiler_limit();
    $parsed_output = $response->get_parsed_output();
    $current_limit_vfiler = $parsed_output->[0]->{'current_limit'};
    $current_in_use_vfiler = $parsed_output->[0]->{'current_in_use'};
    $platform_hard_limit = $parsed_output->[0]->{'platform_hard_limit'};
    if($platform_hard_limit == $current_in_use_vfiler){
        $Log->log("Error: vFiler current in use limit exceeded beyond platform hard limit. Remove unused vFilers manually and re-run");
        return 1;
    }
    elsif(($current_limit_vfiler == $platform_hard_limit)&&($current_in_use_vfiler < $current_limit_vfiler)){ #Check if Current vfiler limit equals Platfrom Hard limit
        $Log->log("Found vfiler current limit equal to platform limit, so no need of increasing vfiler limit.");
        return 0;
    }
    if (($current_limit_vfiler - $current_in_use_vfiler) >= 0 ){
        $Log->log("Found vfiler limit breach on the controller, vfiler limit now is: $current_limit_vfiler");
        $max_vfiler_limit = $parsed_output->[0]->{'platform_hard_limit'};
        $Log->log("Platform hard limit on controller is: $max_vfiler_limit");
        #increase vfiler limit to create vFilers later
        if ($current_in_use_vfiler == $max_vfiler_limit){
            $Log->log("Error: vFiler current in use limit exceeded beyond platform hard limit. Remove unused vFilers manually and re-run");
            return 1; 
        } elsif ($current_in_use_vfiler < $max_vfiler_limit){
            $Log->log("Increasing vFiler limit to maximum allowed by paltform: $max_vfiler_limit");
            $Api_Set_Obj_FILER->vfiler_limit(maxvfilers => $max_vfiler_limit);
            $Log->log("Setting Autoboot to True");
            my @nodes = NACL::C::Node->find();
            my ($node) = @nodes;
            my $autoboot_state =
            HAutils::set_bootarg_autoboot(node => $node,
                                     autoboot_value => "TRUE");
            if ($autoboot_state != 1) {
               $Log->log("Failed to Set Autoboot to TRUE on node $node");
               return $TCD::FAIL;
            }
            #rebooting filer
            $Log->log("Rebooting controller");
            NACL::C::Node->reboot( command_interface => $filer );
            sleep 100;
            $Log->log("Looping 30 times to check controller is back online or not");
            while ($loop++ < 30){
                $response = `ping -c 2 $FILER`;
                if ($response =~ /\s2\sreceived/ ){
                    $Log->log("Controller has rebooted");
                    $Log->log("Sleeping for 10 seconds for SSH session reactiviation");
                    sleep 10;
                    $ping_flag = 1;
                    last;
                }else{
                    $Log->log("Sleeping for 10 seconds");
                    sleep 10;
                }
            }
            if ( !$ping_flag ){
                $Log->log("$FILER is not reachable");
                $Log->log("Reboot failed");
                return 1;
            }
        }
    }
    return 0;
}

1;
