# $Id: //depot/prod/test/main/lib/EMS.pm#1

#***********************************************************************
#
# NAME
#   EMS.pm - 
#
# SYNOPSIS
#   EMS Library for the invocation and verfication of 7G, 7mode BR, and
#       BR CM EMS.
#
# DESCRIPTION
#   This library is a wrapper for the typical EMS commands that an end user
#   may desire for the configuration, invocation, and verification of ems
#   events.  The current library was designed to use Server.pm.
#
# AUTHOR
#   dl-cluster-qa
#
#
# HISTORY
# 4/1/2009 Initial revision
# 6/17/2009 Added support for 10M (cluster mode).
# 3/10/2010 Added support for Rolling Rock:
#   Additional Routines:
#       configure_destination
#       remove_destination
#       verify_ems_email_reception
#       invoke_event_without_error_checking
#       invoke_kernel_event
#   Modified Routines:
#       invoke_event - when we validate status, we now only filter based on the
#           event name and not the severity as well.
#       _retrieve_events - returns more data about the events
#
#
# TODO
# 1)  For BR and higher builds we should try to load the event catalog via the
#   filer.  These event def files are stored in /etc/ems.  Currently the get
#   catalog routines load the event defs from the build directories at netapp.
# 2)  The invoke and verify routines don't currently support deprecated
#   events.  So the -version tag is useless here.  Perhaps we should add
#   support for this.
# 3) The verify event log routine will always match on the first log entry
#   with that name and parameters.  So if you generate multiples of the same
#   event with the same params, we only match on the first entry encountered.
#   This should be updated to remove matched events from the event log listing.
# 4) The retrieve_log_events routing currently utilizes ems log dump for both
#   modes.  For a c-mode filer, we should try to utilize the event log show.
#
#***********************************************************************

## @summary     EMS library    

package EMS;

#use statements
use Switch;
use Data::Dumper;
use Date::Manip;
use Server;
use vars qw($LOGDIR $LOGDIR_UNIX);
use Tharn qw($Log snooze);
use Options;
use Params::Common;
use Params::Validate qw(:all);
#use XML::Simple qw(:strict);
use Syslog::syslog;
use NACL::APISet;
use NACL::C::Node;
use NATE::BaseException qw(:try);
use NACL::CS::EventStatus;
use NACL::Exceptions::NoElementsFound qw(:try);
#Global variables
our $ontap_version;
my $Server = new Server();

$Options = new Options;
$Options->parameters(definitions=>
                     [
                      {
                       name=>"EVENT_CATALOG",
                       description=>"the event catalog file to use for ems, the
                            full path to the eventCatalog.xml file",
                      },
                      ]
                      );
$Options->set(CONNECTION_MAXBUFSIZE => '15m');
$Options->process();
my $params=$Options->get_hash();
my $mode;

my $version = _get_version();
if ( $version =~ /7M|7G/ ) {
    $mode = "7Mode";
} else {
    $mode = "Nodescope";
}
my $Node = NACL::C::Node->find();
my $Api_Set_Obj = $Node->apiset(
    category  => "Node",
    interface => "CLI",
    set       => "$mode",
);
my $Api_Set_SShell=$Node->apiset(
    interface => "CLI",
    set       => "Systemshell",
);
#Methods

###############################################################################
#Sub: set_autosuppression
#Input Params: Hash
#Return values: None
#Summary: Enables the setting of ems.autosuppression on the filer.
###############################################################################
sub set_autosuppression {
    my %params = validate(
            @_,
            {
                enable=> { type => SCALAR,
                           optional => 0,
                           regex => qr/on|off/i,
                           default => "on"
                         }
            },
    );
    my $version=_get_version();
    if ($version=~/7M|7G/) {
        $Server->EMSAutosuppressEnable(enable=>$params{enable});
    }
    else { #10M
        $Server->event_config_modify(suppression=>$params{enable});
    }
}

###############################################################################
#Sub: set_log_dup_suppression
#Input Params: Hash
#Return values: None
#Summary: Enables the setting of ems.log.dupsuppression on the filer
###############################################################################
sub set_log_dup_suppression { #7G/7M only
    my %params = validate(
            @_,
            {
                enable=> { type => SCALAR,
                           optional => 0,
                           regex => qr/on|off/i,
                           default => "on"
                         }
            },
    );
    my $version=_get_version();
    if ($version=~/7M|7G/) {
        $Server->EMSLogDupSuppressEnable(enable=>$params{enable});
    }
}

###############################################################################
#Sub: invoke_event
#Input Params: Hash
#Return values: Ref to Array of Hashes
#Summary: Function to invoke an ems event on the filer
###############################################################################
sub invoke_event {
    my %params = validate (
        @_,
        {
            event => { type => SCALAR,
                       optional => 0,
                       default => undef
                     },
            log => { type => SCALAR,
                     optional => 1,
                     regex => qr/0|1/i,
                     default => 0
                   },
            trap => { type => SCALAR,
                      optional => 1,
                      regex => qr/0|1/i,
                      default => 0
                    },
            version => { type => SCALAR,
                         optional => 1,
                         regex => qr/\d*/i,
                         default => undef
                       },
            severity => { type => SCALAR,
                          optional => 1,
                          regex => qr/NODE_FAULT|SVC_FAULT|NODE_ERROR|SVC_ERROR|WARNING|NOTICE|INFO|DEBUG/i,
                          default => undef
                        },
            arguments => { type => SCALAR,
                           optional => 1,
                           default => undef
                         },
            verify => { type => SCALAR,
                        optional => 1,
                        regex => qr/both|log|status|none/i,
                        default => "both"
                      },
            suppress_invocation_failures => { type => SCALAR,
                        optional => 1,
                        regex => qr/0|1/i,
                        default => 0
                      },
        }
    );
    my %severity_hash=(NODE_FAULT=>'EMERG',
                       EMERG=>'EMERG',
                       NODE_ERROR=>'CRIT',
                       CRIT=>'CRIT',
                       SVC_FAULT=>'ALERT',
                       ALERT=>'ALERT',
                       SVC_ERROR=>'ERR',
                       ERR=>'ERR',
                       ERROR=>'ERR',
                       WARNING=>'WARN',
                       WARN=>'WARN',
                       NOTICE=>'NOTICE',
                       INFO=>'INFO',
                       DEBUG=>'DEBUG');
    my %event_options;
    my $invoked_events;
    my $pre_event_status;
    my $post_event_status;
    my $event_generation_duration;
    my $version=_get_version();
    my $possible_events;
    my $generated_event; 
    my $count;

    #parse out parameters to determine which ones to pass in to the invoke
    while (my ($key, $value) = each(%params)) {
        if ((defined $value and $value !=0) or (defined $value and $key ne 'log' and $key ne 'trap')) { #we don't parameters that weren't included or log or trap params with 0 values
            if ($key eq 'trap' or $key eq 'log') {
                $value=''; #if the parameter is trap or log, set the value to null as the -l or -t flags are automatically passed in if the tag exists.
            }
            $Log->debug("invoke parameter $key has value $value \n");
            $event_options{$key}=$value;
        }
    }
    
    #get the current time before event generation
    $event_generation_duration=time();
    
    #get all matching events from the event xml database
    if ($version =~ /BR-7M|7G/) {
        $possible_events=_retrieve_events(eventfilter=>$params{event},ignore_ems_eut_events=>1);
    } else {
        $possible_events=_retrieve_events(eventfilter=>$params{event});
    }

    my $number_of_possible_events=@{$possible_events};
    
    #check to ensure there are events to generate
    if ($number_of_possible_events <=0) {
        $Log->warn("No events matching the search criteria were found \n");
        return 0;
    }
     
    #iterate through each event, saving to the invoked array and generating the event    
    foreach my $event (@$possible_events){
        my $current_event;
        $event_options{event}=$event->{name};
        $current_event->{event}=$event->{name};
        
        #set the description if configured
        if ($event->{description} ne "") {
            $current_event->{description}=$event->{description};
        }
        
        #set the corrective-action if configured
        if ($event->{'corrective-action'} ne "") {
            $current_event->{'corrective-action'}=$event->{'corrective-action'};
        }
        
        #set the syslog if configured
        if ($event->{syslog} ne "") {
            $current_event->{syslog}=$event->{syslog};
        }
        
        #set the snmp if configured
        if ($event->{snmp} ne "") {
            $current_event->{snmp}=$event->{snmp}->[0]->{'enterprise-trap'}->[0]->{'string-var'}->[0]->{format}->[0];
        }
        
        #grab trap id, if there is one
        $current_event->{trap_id}=$event->{snmp}->[0]->{'enterprise-trap'}->[0]->{'specific-type'};
        
        #if no arguments were declared, randomly generate them
        if (! $params{arguments}) {
            my $generated_arguments;
            my @parms_v;
            for (my $i=0; $i < $event->{num_params}; $i++) {
                my $type = $event->{param_types}->[$i];
                #$Log->comment("parameter type:$type\n");
                my $randparm=_generate_param_value($type);
                #0x may not appear before a hex number in the log message, 
                #so only use the value for verification 
                my $randparm_v = $randparm;
                if ($type =~ /HEX/i) {
                    $randparm_v =~ s/^0x//;
                }

                #make sure each parameter value does not match with each other.
                if (@parms_v && 
                    (scalar grep ( $_ =~ m"\Q$randparm_v", @parms_v) ||
                    scalar grep ( $randparm_v =~ m"\Q$_", @parms_v))) {
                     redo;
                } else { 
                        push @parms_v, $randparm_v;
                }
                $generated_arguments=$generated_arguments." ".'"'.$randparm.'"';
                $current_event->{parameters}->{$event->{params}->[$i]}=$randparm_v;
            }
            $event_options{arguments}=$generated_arguments;
        }
        
        #otherwise, use the values specified
        else { 
            my @event_params=split(" ",$event_options{arguments});
            if (@event_params > ($event->{num_params}) and $event->{name} !~ /asup.msg|callhome|asup.perf/) {
                $Log->warn("The user specified event contains more parameters than the event definition \n");
                return 0;
            }
            my $i=0;
            foreach my $parameter (@event_params) {
                $current_event->{parameters}->{$event->{params}->[$i]}=$parameter;
                $i++;
            }
        }
        
        #if a value was passed in for the severity convert it and add it to the event
        #we must do this as the cli accepts one value for severity, but converts it
        #to a different severity in ems event status.
        if ($params{severity}) {
            $current_event->{severity}=$severity_hash{$params{severity}};
        }
        
        #else check if we need to generate a severity.  If event severity is VAR
        #it can generate any severity for that event, so let's choose one
        elsif ($event->{severity} eq "VAR") {
            $event_options{severity}="NODE_FAULT";
            $current_event->{severity}="VAR";
        }

        #otherwise just add the message severity to our event after converting the
        #severity.  The event definition contains event severities of both
        #NODE_FAULT and EMERG to signify the EMERG severity.
        else {
            $current_event->{severity}=$severity_hash{$event->{severity}};
        }
        
        #check for special case event
        #this event also requires a process id, so append that to the severity
        if ($current_event->{event} eq "kern.syslog.msg") {
            $current_event->{severity}='rshd:'.$current_event->{severity};
        }
        
        #add the event and parameters to the invoked array
        push (@$invoked_events, $current_event);
        
        #get the current status of the event before generation
        if ($params{verify} eq "status" or $params{verify} eq "both") {
            my $event_status=get_event_status(event=>$current_event->{event});
            push (@$pre_event_status, $event_status->[0]);
        }
         
         
   my  $expected_event = $event->{name};
        #Check if the cmode event already exists in the logs?
   try {
      my $result = NACL::CS::EventStatus->fetch(
        command_interface => $Node,
        filter            => { "messagename" => $expected_event }
    );
      
     $count = $result->last_time_occurred();

     } ## end try
         catch NACL::Exceptions::NoElementsFound with {
         $count = 0;
     };
       # If seqnum == 0 that implies no message found in the logs.     
       #generate the event
       
         $Log->comment("GENERATING EVENT: $event->{name} with $event->{num_params} params, $event_options{arguments} \n");
        my $response=_generate_event(\%event_options);
        delete $event_options{severity};
        
        #if the event generation failed
        if ($response =~ /ERROR/i) {
            $Log->warn("An Error was returned from the EMSInvoke command: $response \n");
            #ignore any event generation errors if suppress flag is set
            if ($params{suppress_invocation_failures}) {
                pop(@$invoked_events);
                pop(@$pre_event_status);
                next;
            }
            #else we failed to generate an event, so fail out of routine
            else {
                return 0;
            }
        }
        
        #get the current status of the event after generation
        if ($params{verify} eq "status" or $params{verify} eq "both") {
            $event_status=get_event_status(event=>$current_event->{event});
            push (@$post_event_status, $event_status->[0]);
        }
    }
    
    $Log->trace("EMS invoke_event routine invoked the following events: ".Dumper($invoked_events)."\n");
    
    #Calculate the amount of time it took to generate all of the events
    $event_generation_duration=time() - $event_generation_duration + 45;
    
    if ( $params{verify} eq "log" or $params{verify} eq "both" ) {
        snooze(40);

#10M has a bug where the "iems log dump" command doesn't return the most recent events
#until either the next event arrives or 30 seconds has passed.
# If this c mode execute the event log show command
        my $Mode = $Node->mode();
        if ( $Mode eq "CMode" ) {
            if ( !_verify_cmode_status( $generated_event, $count ) ) {
                $Log->comment( "Event verification for  "
                        . $generated_event
                        . " failed"
                        . " Failed to find the generated event in the log. \n"
                );
                return 0;
            } ## end if ( !_verify_cmode_eventlog...
        } ## end if ( $Mode eq "CMode" )

        # If this 7 mode then execute event log dump <x> second command
        else {
            if (!_verify_event_log(
                    event => $invoked_events,
                    time  => $event_generation_duration
                )
                )
            {
                $Log->comment( $invoked_events
                        . " event verification failed.  Failed to find the generated event in the log. \n"
                );
                return 0;
            } ## end if ( !_verify_event_log...
        } ## end else [ if ( $Mode eq "CMode" )
    } ## end if ( $params{verify} eq...

     #Verify that the event status has changed for the generated event
    if ($params{verify} eq "status" or $params{verify} eq "both") {
        if (! _verify_event_status(pre_event_status=>$pre_event_status, post_event_status=>$post_event_status, invoked_events=>$invoked_events)) {
            $Log->comment("Event verification failed.  Failed to verify the event status entries changed for the generated event \n");
            return 0;
        }
    }
    
    #event generation and verification were successful, return the events invoked
    return $invoked_events;
    
}

###############################################################################
#Purpose - To verify cmode event status.
###############################################################################
sub _verify_cmode_status {
    my $expected_event   = $_[0];
    my $expected_count =   $_[1];
    my $count;
  $Log->comment("Post event log show");
    try {
        $result = NACL::CS::EventStatus->fetch(
            command_interface => $Node,
            filter            => { "messagename" => $expected_event }
        );
        $count = $result->last_time_occurred();

    } ## end try
    catch NACL::Exceptions::NoElementsFound with {
        $count = 0;
        $Log->comment("No Event was generated". "The testcase now fails");
        return $count;
    };

    if ($count eq $expected_count ) {
      return 0;
      } else {
      return 1;
    }
} ## end sub _verify_cmode_evenStatus
###############################################################################
#Sub: invoke_random_event
#Input Params: Hash
#Return values: Ref to Array of Hashes
#Summary: Invokes a random ems event on the filer
###############################################################################
sub invoke_random_event {
    my %params = validate(
            @_,
            {
            event => { type => SCALAR,
                       optional => 1,
                       default => '*'
                     },
            log => { type => SCALAR,
                     optional => 1,
                     regex => qr/0|1/i,
                     default => 0
                   },
            trap => { type => SCALAR,
                      optional => 1,
                      regex => qr/0|1/i,
                      default => 0
                    },
            verify => { type => SCALAR,
                        optional => 1,
                        regex => qr/both|log|status|none/i,
                        default => "both"
                      },
            ignore_timer_suppressed_events => { type => SCALAR,
                        optional => 1,
                        regex => qr/0|1/i,
                        default => 0
                      },
            }
    );
    my $invoked_event;
    my $possible_events;
    my $number_of_possible_events;
    my $rand_event_id;
    
    #retrieve all events based on event filter specified
    $possible_events=_retrieve_events(eventfilter=>$params{event},ignore_timer_suppressed_events=>$params{ignore_timer_suppressed_events});
    
    #pick a random event to generate
    $number_of_possible_events=@{$possible_events};
    srand(time());
    $rand_event_id=int(rand($number_of_possible_events));
    
    #generate the event
    $invoked_event=invoke_event(event=>$possible_events->[$rand_event_id]->{name}, log=>$params{log}, trap=>$params{trap}, verify=>$params{verify});
    
    #return the invoked event
    return $invoked_event;
}

###############################################################################
#Sub: invoke_random_snmp_event
#Input Params: Hash
#Return values: Ref to Array of Hashes
#Summary: Invokes a random snmp ems event on the filer
###############################################################################
sub invoke_random_snmp_event {
    my %params = validate(
            @_,
            {
            event => { type => SCALAR,
                       optional => 1,
                       default => '*'
                     },
            log => { type => SCALAR,
                     optional => 1,
                     regex => qr/0|1/i,
                     default => 0
                   },
            verify => { type => SCALAR,
                        optional => 1,
                        regex => qr/both|log|status|none/i,
                        default => "both"
                      },
            ignore_timer_suppressed_events => { type => SCALAR,
                        optional => 1,
                        regex => qr/0|1/i,
                        default => 0
                      },
            }
    );
    my $invoked_event;
    my $possible_events;
    my $number_of_possible_events;
    my $rand_event_id;
    
    #retrieve all events based on the event filter specified and the trap parameter
    $possible_events=_retrieve_events(eventfilter=>$params{event},snmp_traps=>1,ignore_timer_suppressed_events=>$params{ignore_timer_suppressed_events});
    
    #pick a random event to generate
    $number_of_possible_events=@{$possible_events};
    srand(time());
    $rand_event_id=int(rand($number_of_possible_events));
    
    #generate the event
    $invoked_event=invoke_event(event=>$possible_events->[$rand_event_id]->{name}, log=>$params{log}, trap=>1, verify=>$params{verify});
    
    #return the invoked event
    return $invoked_event;
}

###############################################################################
#Sub: invoke_random_syslog_event
#Input Params: Hash
#Return values: Ref to Array of Hashes
#Summary: Invokes a random snmp ems event on the filer
###############################################################################
sub invoke_random_syslog_event {
    my %params = validate(
            @_,
            {
            event => { type => SCALAR,
                       optional => 1,
                       default => '*'
                     },
            trap => { type => SCALAR,
                      optional => 1,
                      regex => qr/0|1/i,
                      default => 0
                    },
            verify => { type => SCALAR,
                        optional => 1,
                        regex => qr/both|log|status|none/i,
                        default => "both"
                      },
            ignore_timer_suppressed_events => { type => SCALAR,
                        optional => 1,
                        regex => qr/0|1/i,
                        default => 0
                      },
            }
    );
    my $invoked_event;
    my $possible_events;
    my $number_of_possible_events;
    my $rand_event_id;
    
    #retrieve all events based on the event filter specified and the trap parameter
    $possible_events=_retrieve_events(eventfilter=>$params{event},syslog=>1,ignore_timer_suppressed_events=>$params{ignore_timer_suppressed_events});
    
    #pick a random event to generate
    $number_of_possible_events=@{$possible_events};
    srand(time());
    $rand_event_id=int(rand($number_of_possible_events));
    
    #generate the event
    $invoked_event=invoke_event(event=>$possible_events->[$rand_event_id]->{name}, log=>1, trap=>$params{trap}, verify=>$params{verify});
    
    #return the invoked event
    return $invoked_event;
}

###############################################################################
#Sub: invoke_kernel_event
#Input Params: 
#Return values: 
#Summary: Invoke a kernel level event using a command at the 7M prompt.
###############################################################################
sub invoke_kernel_event {
    my $result = 1;
    try {
        $Api_Set_Obj->syslog(
            "repeat-count" => 1,
            "priority"     => 'crit',
            "facility"     => 'auth',
            "message"      => "Invoke kernel event" . time(),
        );
        snooze(31);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        $Log->comment(
            "Failed to call syslog command
        with: $exception_object->text()\n"
        );
        $result = 0;
    };

    return $result;
} ## end sub invoke_kernel_event

############################################################################
# TestCase: invoke_event_without_error_checking
# Parameters: None
# Returns: TCD Execution Result
# Purpose:  By pass catalog lookup and verification to speed up event generation
#############################################################################
sub invoke_event_without_error_checking {
    my %params = (
            event => "app.log.crit",
            arguments => undef, #User can input specific arguments
            num_of_invokes => 1,
            num_arguments => undef, #The number of arguments to be randomly generated
            @_,
    );
    
    if (!defined($params{arguments}) and !defined($params{num_arguments})) {
        $Log->error("Either arguments or num_arguments are required for 
                     invoke_event_without_error_checking subroutine\n");
        return 0;
    }
    
    my $version=_get_version();
    
    if ($version=~/7M|7G/) {
        my $counter=1;
        while($counter <= $params{num_of_invokes}){
            #Randomly generate events if num_arguments is populated
            eval {
                if (defined($params{num_arguments}) and  $params{num_arguments}== 0 ) {
                    $Server->EMSInvoke(event=>$params{event});
                }
                elsif (defined($params{num_arguments}) and $params{num_arguments} > 0) {
                    my $generated_arguments;
                    for (my $i=0; $i < $params{num_arguments}; $i++) {
                        my $randnumber=int(rand(1000) + 1); #randomly choose value between 1 and 1000
                        $generated_arguments=$generated_arguments." ".$randnumber;
                    }
                    $Server->EMSInvoke(event=>$params{event},arguments=>$generated_arguments);
                }
                else {
                    $Server->EMSInvoke(event=>$params{event},arguments=>$params{arguments});
                }
            };
            if ($@) {
                $Log->error("Failed to invoke the event with error $@...\n");
                return 0;
            }
            $counter++;
        }
    }
    else { #else 10M
        my $counter=1;
        while($counter <= $params{num_of_invokes}){
            eval {
                if (defined($params{num_arguments}) and $params{num_arguments} == 0 ) {
                    $Server->event_generate(messagename=>$params{event});
                }
                elsif (defined($params{num_arguments}) and $params{num_arguments} > 0) {
                    my $generated_arguments;
                    for (my $i=0; $i < $params{num_arguments}; $i++) {
                        my $randnumber=int(rand(1000) + 1); #randomly choose value between 1 and 1000
                        $generated_arguments=$generated_arguments." ".$randnumber;
                    }
                    $Server->event_generate(messagename=>$params{event},values=>$generated_arguments);
                }
                else {
                    $Server->event_generate(messagename=>$params{event},values=>$params{arguments});
                }
            };
            if ($@) {
                $Log->error("Failed to invoke the event with error $@...\n");
                return 0;
            }
            $counter++;
        }
    }
return 1;
} ## end sub invoke_event_without_error_checking

###############################################################################
#Sub: retrieve_log_events
#Input Params: Hash
#Return values: Ref to Array of Hashes
#Summary: Retrieves and returns all of the ems events in the log based upon
#   specified time frame.
###############################################################################
sub retrieve_log_events {
    my %params = validate(
            @_,
            {
            time => { type => SCALAR,
                      optional => 1,
                      regex => qr/\d*ns|\d*nm|\d*nh|\d*nd|\d*/i,
                      default => "5nm"
                    }
            }
    );

    my $emslog_entry;
    my $parsed_event;
    my $log_events;
    my $output;
    
    my $version=_get_version();
    
    #get all of the events based upon the time parameter specified
    if ($version=~/7G|7M/) {
        $output = $Server->EMSLogDump(time=>$params{time});
    }
    else { #10M
        $output = $Server->shell(cmd=>"ngsh -c \"system node run local -command \"ems log dump $params{time}\"\"");
        #$output = $Server->system_node_run(node=>"local",command=>"ems log dump $params{time}");
    }
    my @lines = split(/\n/, $output);
    
    foreach my $line (@lines) {
        #Line begins with NR, so call NR parse and continue
        if ($line =~ /^<NR/) {
            _parseNR_emslog_line($line);
            next;
        }
        #Line starts with LR, so continue reading the line
        elsif ($line =~ /^<LR/) {
            $emslog_entry = $line;
            next;
        }
        #We've reached the end of the LR entry, so lets add the ending and process the message
        elsif ($line =~ /^<\/LR/) {
            $emslog_entry .= $line;
        }
        #The line is part of the LR content, add to the message and continue reading
        else {
            $emslog_entry .= " $line";
            next;
        }
        
        #Now parse the full LR line
        $parsed_event=_parseLR_emslog_line($emslog_entry);
        
        #add the line to the event array
        push (@$log_events,($parsed_event));
        
        #reset the emslog_entry value
        $emslog_entry = "";
    }
    
    $Log->trace("Events returned by the EMS retrieve_log_events: ".Dumper($log_events)."\n");
    
    #return the events
    return $log_events;
}

###############################################################################
#Sub: get_status
#Input Params: total
#Return values: output
#Summary:
#   Current time: 17Mar2009 16:24:37
#   Engine status: total 435 (errs 0), drops 0, suppr (dup 0, timer 46, auto 0)
###############################################################################
sub get_status { #7M/7G only
    my $output= $Server->EMSStatus();
    my %status_hash;
    my @status = split(/\n/, $output);
   
    foreach $line (@status) {
        chomp($line);
    
        if ($line =~ /time:\s(.*\s\d+:\d+:\d+)/) {
            $status_hash{current_date} = $1;
        } elsif ($line =~ /.+total\s(\d+).+errs\s(\d+).+drops\s(\d+).+suppr.+dup\s(\d+).+timer\s(\d+).+auto\s(\d+)/) {
            $status_hash{engine_status}{total} = $1;
            $status_hash{engine_status}{errs} = $2;
            $status_hash{engine_status}{drops} = $3;
            $status_hash{engine_status}{suppression}{dup} = $4;
            $status_hash{engine_status}{suppression}{timer} = $5;
            $status_hash{engine_status}{suppression}{auto} = $6;   
        }    
    }
    return \%status_hash;
}

###############################################################################
# Sub: get_event_status
# Input Params: total
# Return values: output
# Summary:
# Using Server.pm, we issue the command 'ems event status'. Currently, the 
# cli will not allow for advanced queries. This method is intended to 
# enhance the ability for querying against the 'ems event status' output.
# 
###############################################################################
sub get_event_status {
    my %params = validate(
        @_,
        {
            severity => { type  => SCALAR, 
                          optional => 1,
                          regex => qr/EMERG|CRIT|ALERT|ERR|WARN|NOTICE|INFO|DEBUG|VAR/i,
                        },
               event => { type => SCALAR,
                          optional => 1,
                          regex => qr/\w\.|\*\.\w|\*\w|\w.+\*|\w/,
                        },
        sort_by_date => { type => SCALAR,
                          default => 0, # if not specified, will be in alphabetical order by event name 
                          regex => qr/0|1/,
                        },
        },
    );
  
    my $version=_get_version();
      
    if ($version =~ /7G|7M/) {
        my @output_aoh;
        my @return_aoh;
        my @sort_aoh;
        my $output = $Server->EMSEventStatus(-Timeout=>120);
        my @split_array = split("\n",$output);
        # shift off the date and engine stats
        my $date = shift(@split_array);
        my $engine = shift(@split_array);
        # shift off the next two header lines, don't need them
        shift(@split_array);
        shift(@split_array);
    
        # reconstitute a new string with the new array containing
        # only the ems events
        my $join_string = join("\n",@split_array);
        # replace all the newlines, tabs, and carriage returns with a space
        $_ = $join_string;
        s/[\n\t\r]//g;
        # when boot is given as Last, it doesn't allow a position for where 
        # the timestamp is for every other entry. Adding empty quotes 
        # to normalize the entry so that it can be sorted properly.
        s/ boot / boot "" /g;
        # splitting up the normalized string via spaces
        my @output_array = split(/\s+/s,$_);
        # to determine how many times we have to loop through the array
        # dividing the number of indexes by the number of record keys
        # 8 keys: name, date, time, indications, drops, suppr_dup, 
        # suppr_timer, and suppr_auto
        my $indexes = int(@output_array);
        my $num_loop = int(@output_array)/8;
        # shifting off the indexes as they are assigned to the temp hash
        for (my $i = 0;$i<$num_loop;$i++) {
            my %output_hash;
            $output_hash{name} = shift(@output_array);
            $output_hash{date} = shift(@output_array);
            $output_hash{time} = shift(@output_array);
            $output_hash{epoch} = _get_seconds(ems_date=>$output_hash{date},ems_time=>$output_hash{time});
            $output_hash{indications} = shift(@output_array);
            $output_hash{drops} = shift(@output_array);
            $output_hash{suppression}{dup} = shift(@output_array);
            $output_hash{suppression}{timer} = shift(@output_array);
            $output_hash{suppression}{auto} = shift(@output_array);
            # shove it into an array
            push @output_aoh, {%output_hash};
        }
        # filter by severity
        if ($params{severity} || $params{event}) {
            if ($params{severity}) {
                my @sort_aoh;
                foreach my $tmp (@output_aoh) {
                    if ($tmp->{name} =~ $params{severity}) {
                       push @sort_aoh, {%{$tmp}};
                    }
                }
                @output_aoh = @sort_aoh;
            } 
        
            # filter by event: alphanumeric  
            if ($params{event}) {
                my @sort_aoh;
                foreach my $tmp (@output_aoh) {
                    if ($tmp->{name} =~ /$params{event}/) {
                       push @sort_aoh, {%{$tmp}};
                    } elsif ($tmp->{name} eq substr($params{event}, 0, 39)) {
                             #in 7m, event status only show at most 39 chars for name
                             push @sort_aoh, {%{$tmp}}; }
                }
                @output_aoh = @sort_aoh;
            }
        }
        # If the list is empty, return with error string in aoh    
        if (int(@output_aoh) == 0) {
            $output_aoh[0] = "There are no entries matching your query.";
            return \@output_aoh;
        # If sorting by date is chosen, return with sorted aoh
        } elsif ($params{sort_by_date}) {
            my @tmp =  sort { $a->{epoch} <=> $b->{epoch} } @output_aoh;
            return \@tmp;
        # If no options are given, return the unsorted aoh
        } else {
            return \@output_aoh;
        }
    }
    else { #10M
        my $output;
        my @output_aoh;
        if ($params{event}) {
            eval {
                $output=$Server->event_status_show(node=>"local",messagename=>$params{event});
            };
            if ($@){
                $output_aoh[0] = "There are no entries matching your query.";
                return \@output_aoh;
            }
        }
        else {
            eval {
                $output=$Server->event_status_show(node=>"local");
            };
        }
        foreach my $event (@$output) {
            my $newhash;
            my $datetime=$event->{'last-time-occurred'};
            $datetime =~ /(\d+\/\d+\/\d{4})\s*(\d+\:\d+\:\d+)/;
            $newhash->{date}=$1;
            $newhash->{time}=$2;
            $newhash->{name} = $event->{messagename};
            $newhash->{epoch} = _get_seconds(ems_date=>$newhash->{date},ems_time=>$newhash->{time});
            $newhash->{indications} = $event->{indications};
            $newhash->{drops} = $event->{drops};

            # Perform the following only for Rolling Rock and beyond
            if ($version =~ /RR/ || $version =~ /SN/ || $version =~ /FS/) {
                $newhash->{suppression}->{dup} = $event->{'num-dup-suppressed'};
                $newhash->{suppression}->{timer} = $event->{'num-timer-suppressed'};
                $newhash->{suppression}->{auto} = $event->{'num-autosuppressed'};
            } else {
               $newhash->{suppression}->{auto} = $event->{drops};
            }
            push (@output_aoh,$newhash);
        }
        if ($params{sort_by_date}) {
            my @tmp =  sort { $a->{epoch} <=> $b->{epoch} } @output_aoh;
            return \@tmp;
        }
        
        return \@output_aoh;
    }
}

###############################################################################
#Sub: verify_ems_email_reception
#Input Params: Hash
#Return values: 0 (failure) or 1 (success - no save dirs) or array of save dirs
#Summary: Verify the reception of an autosupport via the specified protocol.
#       Returns a 0 on failure, a 1 on success, or the save dirs if successfully
#       verified and asup_dirs was specifed for the server.
###############################################################################
sub verify_ems_email_reception {
    my %params = (
            events => undef, #an event or array of events
            server => undef, #transport server to verify events from
            time_limit => 30, #amount of time to look for the events (in secs)
            @_,
    );
    
    $Log->comment("Verifying events\n\n");
    my $server_messages;
    my $ems_directories=undef;
    my $flag=0; #indicates whether we have found all events or not
    #these are the events we need to look for
    my $generated_events=$params{events};
    
    my $max_time; #time to exit this routine
    
    #if a single event
    if (ref($params{events}) ne "ARRAY") {
        $Log->comment("Verify EMS email reception: Looking for message $params{events}->{event} from node $params{events}->{node} \n");
        #look for the event over a period of 30secs
        $max_time=time()+$params{time_limit};
        while (time() < $max_time and $flag==0) {
            #check to ensure server is still up
            if (!$params{server}->{subtest}->is_alive()) {
                $Log->error("Server failed to start or has terminated...\n");
                return 0;
            }
            #get the messages from the transport server subprocess
            $server_messages=$params{server}->get_all();
            if (@{$server_messages} > 0) {
                $Log->comment("MESSAGES POSTED TO SERVER: ".Dumper($server_messages)."\n");
            }
            else {
                $Log->debug($max_time - time()." seconds left until we stop looking for ems messages!\n");
            }
            #check each message posted from the transport server against our expected events
            foreach my $message (@$server_messages) {
                #if found, add the save directory to the array and break out
                if ($message =~ /$params{events}->{event}/ and $message =~ /$params{events}->{node}/ and $message =~ /Saved to directory\s(.*)/) {
                    $Log->comment("Verify EMS email reception: Found our ems email!  $message \n");
                    if ($1 ne "") {
                        $Log->comment("Verify EMS email reception: EMS SAVE DIR was $1 \n");
                        push (@$ems_directories,$1);
                    }
                    $flag=1;
                }
            }
            #not yet found, look again
        }
    }
    #else if an array of events
    else {
        #set the timeout value to look for events number of events at 150s each
        if ($params{time_limit}==30) {
            $max_time=time()+($params{time_limit} * @$generated_events);
        }
        else {
            $max_time=time()+$params{time_limit};   
        }
        $Log->debug("TIME is ".time()." and max timer is $max_time \n");
        #while there is time, look for the asups
        while (time() < $max_time and $flag==0) {
            #check to ensure server is still up
            if (!$params{server}->{subtest}->is_alive()) {
                $Log->error("Server failed to start or has terminated...\n");
                return 0;
            }
            #get the messages from the transport server subprocess
            $server_messages=$params{server}->get_all();
            if (@{$server_messages} > 0) {
                $Log->comment("MESSAGES POSTED TO SERVER: ".Dumper($server_messages)."\n"); 
            }
            else {
                $Log->debug($max_time - time()." seconds left until we stop looking for ems messages!\n");
            }
            #check each message we got from the tranport server
            foreach my $message (@$server_messages) {
                my $still_looking;
                #$Log->comment("Verify ASUP reception: Looking for event in the message $message \n");
                #look a matching event that we were expecting
                my $match=0;
                foreach my $single_event (@$generated_events) {
                    $Log->debug("Verify EMS email reception: Looking for $single_event->{event} from node $single_event->{node} \n");
                    #if found, add the save directory to the array and break out
                    if ($message =~ /$single_event->{event}/ and $message =~ /$single_event->{node}/ and $match==0 and $message =~ /Saved to directory\s(.*)/) {
                        $Log->comment("Verify EMS email reception: Found an expected ems event!  $message \n");
                        if ($1 ne "") {
                            $Log->comment("Verify EMS email reception: EMS SAVE DIR was $1 \n");
                            push (@$ems_directories,$1);
                        }
                        $match=1;
                    }
                    #else, add the message we are currently looking for back to the queue
                    else {
                        #$Log->comment("Event not found yet, Adding message to queue \n");
                        push (@$still_looking, $single_event);
                    }
                }
                #generated_events is now the array of events we are looking for, minus the ones we found
                $generated_events=$still_looking; #pointer to still_looking
                $still_looking=undef;
                #if there are no events left to look for, break out
                if (!defined $generated_events) {
                    $flag=1;
                    last;
                }
                #else we still have events to look for, so go back and get more messages!
                $Log->comment("Verify EMS email reception: Number of events left to look for: ".@$generated_events."\n");
                $Log->comment("Verify EMS email reception: EVENTS left to find - ".Dumper($generated_events)."\n");
            }
        }
    }
    
    #if we have found all of our expected asup events
    if ($flag) {
        $Log->comment("Verify EMS email reception: Found all of the events \n");
        #return 1 if no ems events were saved to disk
        if (!defined $ems_directories) {
            return 1;
        }
        #else return an array of the directories where the asups are saved
        else {
            return $ems_directories; #returns undef if no directories, otherwise returns a ref to an array of those dirs
        }
    }
    
    #otherwise, we failed to find at least one event so return a failure 0
    $Log->comment("Verify EMS email reception: Failed to find at least one event \n");
    $Log->comment("Verify EMS email reception: Events not found - ".Dumper($generated_events)."\n");
    return 0;
}

###############################################################################
#Sub: configure_destination
#Input Params: Hash
#Return values: 0 (failure) or 1 (success)
#Summary: Configures the destination on the filer.
###############################################################################
sub configure_destination {
    my %params = (
            name => "test_dest_".time(),
            syslog => undef, #an event or array of events
            snmp => undef,
            mail => undef,
            syslog_facility => undef,
            snmp_community => undef,
            mailfrom => 'test_server@localhost.com',
            client => undef, #required for smtp daemon
            messagename => undef, #events to configure for route
            hide-parameters => undef, #Option to exclude parameters
            @_,
    );
    
    #remove undefined params
    while ( my ($key, $value) = each(%params) ) {
        if (!defined($value)) {
            delete $params{$key}; #remove any params that don't have values configured
        }
    }

    my $version=_get_version();
    
    if ($version =~ /7G|7M/) {
        if (defined($params{snmp})) {
            eval {
                $Server->cli(cmd=>"snmp traphost add $params{snmp}");
                $Server->cli(cmd=>"options snmp.enable");
                $Server->cli(cmd=>"snmp init 1");
            };
            if ($@) {
                return $@;
            }
        }
        if (defined($params{syslog})) {
            Syslog::syslog::delete_syslog_conf();
            Syslog::syslog::create_syslog_conf(selector=>'*.*',action=>'@'.$params{syslog});
            Syslog::syslog::restart_syslogd();
        }
        if (defined($params{mail})) {
            $Log->comment("7M filers don't currently support email destinations for ems events, ignoring the mail parameter\n");
        }
    }
    else { #assume 10M
        # check for existance of the destination
        eval {
            $Server->event_destination_show(name=>$params{name});
        };
        if ($@ !~ /There are no entries matching your query/) {
            $Log->comment("Destination already exists, you must delete it first or choose to modify it, not create it. $@\n");
            return 0;
        }
        #check for mail settings, then configure the mailhost
        if (defined($params{mail})) {
            if (!defined($params{mail})) {
                $Log->comment("You asked us EMS.pm to create a smtp destination, but you didn't provide a client for the mailhost\n");
                return 0;
            }
            else {
                $Server->event_config_modify(mailfrom=>$params{mailfrom},mailserver=>$params{client});
            }
        }
        #remove the un-needed keys
        delete $params{mailfrom};
        delete $params{client};
        my $events=$params{messagename};
        delete $params{messagename};
        #create the destination

        eval {
            $Server->event_destination_create(%params,-Verification=>0);
        };
        if ($@) {
            $Log->comment("Error occured during event destination configuration\n");
            return $@;
        }
        #configure the event to route to the destination
        $Server->event_route_add_destinations(messagename=>$events,destinations=>$params{name});
    }
    return 1;
}

###############################################################################
#Sub: remove_destination
#Input Params: Hash
#Return values: 0 (failure) or 1 (success)
#Summary: Unconfigures the snmp destination and|or syslog on the filer.
###############################################################################
sub remove_destination {
    my %params = (
            name => undef,
            protocols => undef,
            snmp => undef,
            @_,
    );
    $Log->comment("Removing destination");
    my $version=_get_version();
    if ($version =~ /7G|7M/) {
        if ($params{snmp} ) {
            eval {
                $Server->cli(cmd=>"snmp traphost delete $params{snmp}");
                $Server->cli(cmd=>"snmp init 0");
            };
            if ($@) {
                return $@;
            }
        # for backward-compatability
        } elsif ($params{protocols} =~ /snmp/) {
            eval {
                $Server->cli(cmd=>"snmp init 0");
            };
            if ($@) {
                return $@;
            }
        }
        if ($params{protocols}=~/syslog/) {
            Syslog::syslog::delete_syslog_conf();
            Syslog::syslog::restart_syslogd();
        }
    }
    else { #assume 10M
        #remove routing
        eval {
            $Server->event_route_remove_destinations(messagename=>'*',destinations=>$params{name});
        };
        eval {
            $Server->event_destination_delete(name=>$params{name});
        };
        if ($@ and $@ !~ /entry doesn\'t exist/) {
            $Log->comment("Destination deletion failed due to $@\n");
            return $@;
        }
    }
    return 1;
}


###############################################################################
#PRIVATE EMS FUNCTIONS
###############################################################################

###############################################################################
#Sub: _retrieve_events
#Input Params: Hash
#Return values: Ref to Array of Hashes
#Summary: Internal function to retrieve the possible events based on filter
###############################################################################
sub _retrieve_events {
    my $version=_get_version();

    my %params;
    #Due to burt 380707, ingnore generating no-parameter events for BR.
    if ($version =~ /BR/i) {
        %params = (
                      eventfilter => '*',
                      snmp_traps=> 0,
                      syslog=>0,
                      ignore_timer_suppressed_events=>0,
                      ignore_asup_events=>0,
                      ignore_ems_eut_events=>1,
                      ignore_no_parameter_events=>1,
                      ems_cit=>0,
                      @_,
                     );
    }else {
        %params = (
                      eventfilter => '*',
                      snmp_traps=> 0,
                      syslog=>0,
                      ignore_timer_suppressed_events=>0,
                      ignore_asup_events=>0,
                      ignore_ems_eut_events=>1,
                      ignore_no_parameter_events=>0,
                      ems_cit=>0,
                      @_,
                     );
    }
    
    #The following events have known issues BURT: 469831 518295,522511 
    #'ses.bridge.fw.mmErr' has a string length limit.
    #qos.monitor cant configure snmp destination
    #kern.rc no syslog message.
    #wafl.vvol events require a valid VLDB entry or mgwd cores BURT: 636987
    my @blackEventList = (
        'wafl.msg.cpuhog',
        'netinet6.nd6.linklevel.alloc2',
        'nfsd.tcp.upcall.not.set',
        'nfsd.tcp.send.recv.upcall.not.set',
        'cifs.copySID.trace.invalidParam',
        'netinet6.prefix.addr.exists',
        'sparse.msg.nonabortable.detail',
        'wafl.delete.restart',
        'wafl.msg.susp.snapshot',
        'wafl.msg.susp.raid.setup',
        'wafl.msg.susp.detail',
        'wafl.msg.susp.detail.vol',
        'wafl.priority.languish',
        'Nblade.invTcpRecordMarker',
        'iscsi.snmp',
        'wafl.vvol',
        'ses.bridge.fw.mmErr',
        'kern.rc',
        'qos.monitor',
        'ems.eut.privar1p',
        'ems.eut.testptr',
        'wafl.msg.susp.raid.AA',
        'wafl.V4State.obj',
        'nfsd.tcp.send.upcall.not.set'
    );

    # The following events are a stable list for CIT 
    my @goodEventList= qw(
        acp.bind.error
        av.avm.trend.licenseExpiring
        cf.rv.nicReset
        cifs.auditfile.lv.enable.on
        cifs.homedir.activateEntries
        coredump.host.spare.set
        disk.unableToRewriteDegraded
        ip.drd.default.route.new
        lib.nvtrace.syslog.relog.info
        lun.map.aluaConflict
        monitor.shutdown.chassisUnderTemp
        Nblade.NfsRaidError
        Nblade.NfsResponseTraceTriggerHourly
        raid.assim.disk.badlabelversion
        raid.disk.timeout.ios.flush.end
        raid.vol.rootSelectMostRecent
        routed.rip.packet
        ses.status.electronicsError
        sis.revert.to.write.Failed
        snapmirror.32BitTo64BitErr
        snapmirror.dst.chkDegradedMirrorErr
        snapmirror.dst.chkSnapInfo
        snapmirror_copy.UpdateTimeIncreased
        vFiler.etcStorage.readOnly
        vFiler.skipped
        vFiler.unable.initModules
    );
    
    my $return_events;  
    #get the build root
    my $catalog=_get_catalog();
    
    $Log->comment("CATALOG: $catalog \n");
    
    my $ems_cit=$params{ems_cit};
  
    #get all of the events from the xml definition file 
    my $catalog_of_events=_get_events_from_catalog(catalog=>$catalog);
    
    #parse the events or filters that we need to look for
    my @eventfilter=split(",",$params{eventfilter});
    
    #iterate through each filter and add the matching events to our event array
    foreach my $filter (@eventfilter) {
        my $cleaned_event_filter = $filter;
        my $events;
        
        #first lets clean up the filter to form a regex
        $cleaned_event_filter=~s/\./\\\./g;
        $cleaned_event_filter=~s/\*/\.\*/g;
        $Log->comment("REVISED EVENT FILTER= $cleaned_event_filter \n");
        
        #now check each event in the catalog to see if it matches our current filter
        foreach my $event (@{$catalog_of_events->{'event-def'}}){
            
            #check if event is in the stable list, if requested
            next if ($ems_cit && ! grep { $event->{name} eq $_} @goodEventList); 
            
            #check if event in the blacklist 
            next if (scalar grep ( $event->{name} =~ /$_/i, @blackEventList));

            #check if event has been deprecated
            next if (defined $event->{deprecated});
            
            #check for snmp (events that generate traps), if requested
            next if ($params{snmp_traps} and !defined $event->{snmp});

            #ignore generic trap, which does not have trap id and varbindings
            next if ($params{snmp_traps} and defined $event->{snmp}->[0]->{'generic-trap'});
            
            #check for syslog (events that generate syslog entries), if requested
            next if ($params{syslog} and !defined $event->{syslog});

            #check if event has a timer suppression flag, and ignore if requested
            next if ($params{ignore_timer_suppressed_events} and defined $event->{timer});
            
            #check if event is an asup, and ignore if requested
            next if ($params{ignore_asup_events} and $event->{name}=~/callhome|asup\.msg/);

            #check if event is an ems_eut event, and ignore if requested
            next if ($params{ignore_ems_eut_events} and $event->{name}=~/ems.eut/);
            next if ($params{ignore_ems_eut_events} and $event->{'debug-build-only'}=~ /true/);
            #check if event is a no-parameter event, and ignore if requested
            next if ($params{ignore_no_parameter_events} and !$event->{param});
            #if current filter starts and ends with a event name
            if (($cleaned_event_filter =~ /^\w/) and ($cleaned_event_filter =~ /\w$/)) {
                if ($event->{name} =~ /^$cleaned_event_filter$/) {
                    my $numparams=@{$event->{param}};
                    if ($numparams < 1) {
                        $numparams=0;
                    }
                    my $params2;
                    my $params_type;
                    foreach my $param (@{$event->{param}}) {
                        push (@$params2, $param->{name});
                        push (@$params_type, $param->{type});
                    }
                    push (@$return_events, {name=>$event->{name}, num_params=>$numparams, params=>$params2, param_types=>$params_type, severity=>$event->{severity}, timer=>$event->{timer}, snmp=>$event->{snmp}, description=>$event->{description}->[0], 'corrective-action'=>$event->{'corrective-action'}->[0], syslog=>$event->{syslog}->[0]->{format}->[0], asup_subject=>$event->{asup}->[0]->{reason}->[0]});
                }
            }
            
            #if current filter starts with a event name
            elsif ($cleaned_event_filter =~ /^\w/) {
                if ($event->{name} =~ /^$cleaned_event_filter/) {
                    my $numparams=@{$event->{param}};
                    if ($numparams < 1) {
                        $numparams=0;
                    }
                    my $params2;
                    my $params_type;
                    foreach my $param (@{$event->{param}}) {
                        push (@$params2, $param->{name});
                        push (@$params_type, $param->{type});
                    }
                    push (@$return_events, {name=>$event->{name}, num_params=>$numparams, params=>$params2, param_types=>$params_type, severity=>$event->{severity}, timer=>$event->{timer}, snmp=>$event->{snmp}, description=>$event->{description}->[0], 'corrective-action'=>$event->{'corrective-action'}->[0], syslog=>$event->{syslog}->[0]->{format}->[0], asup_subject=>$event->{asup}->[0]->{reason}->[0]});
                }
            }
            
            #if current filter ends with a event name
            elsif ($cleaned_event_filter =~ /\w$/) {
                if ($event->{name} =~ /$cleaned_event_filter$/) {
                    my $numparams=@{$event->{param}};
                    if ($numparams < 1) {
                        $numparams=0;
                    }
                    my $params2;
                    my $params_type;
                    foreach my $param (@{$event->{param}}) {
                        push (@$params2, $param->{name});
                        push (@$params_type, $param->{type});
                    }
                    push (@$return_events, {name=>$event->{name}, num_params=>$numparams, params=>$params2, param_types=>$params_type, severity=>$event->{severity}, timer=>$event->{timer}, snmp=>$event->{snmp}, description=>$event->{description}->[0], 'corrective-action'=>$event->{'corrective-action'}->[0], syslog=>$event->{syslog}->[0]->{format}->[0], asup_subject=>$event->{asup}->[0]->{reason}->[0]});
                }
            }
            
            #else, just check for current filter
            else {
                if ($event->{name} =~ /$cleaned_event_filter/) {
                    my $numparams=@{$event->{param}};
                    if ($numparams < 1) {
                        $numparams=0;
                    }
                    my $params2;
                    my $params_type;
                    foreach my $param (@{$event->{param}}) {
                        push (@$params2, $param->{name});
                        push (@$params_type, $param->{type});
                    }
                    push (@$return_events, {name=>$event->{name}, num_params=>$numparams, params=>$params2, param_types=>$params_type, severity=>$event->{severity}, timer=>$event->{timer}, snmp=>$event->{snmp}, description=>$event->{description}->[0], 'corrective-action'=>$event->{'corrective-action'}->[0], syslog=>$event->{syslog}->[0]->{format}->[0], asup_subject=>$event->{asup}->[0]->{reason}->[0]});
                }
            }   
        }
    }
    
    #return the matching events
    return $return_events;
}

###############################################################################
#Sub: _parseLR_emslog_line
#Input Params: $scalar
#Return values: Ref to Hash
#Summary: Parses the ems log file lines that are LR
###############################################################################
sub _parseLR_emslog_line {
    my($line) = @_;
    my($header, $event, $id, $i);
    my(@headerlist, @eventlist);
    my($type, $version);
    my $emslog_event_by_id;
    
    #Process the line, separate the header and event body
    $Log->debug("Processing LR line: $line \n");
    if ($line =~ /^\s*<LR (.*)>\s*<(.*)\/>\s*<\/LR>\s*$/) {
        $header = "$1";
        $event = $2;
    } else {
        $Log->comment("emslog_process_line: mismatch\n");
        return;
    }
    
    # Get the id.
    if ($header =~ /\s*id=\"([^\"]+)\"/) {
        $id = $1;
        $emslog_event_by_id->{id}=$id;
    }

    # Parse the header (LR) section of the event record
    while ($header ne "") {
        if ($header =~ /^\s*([^=]+)=\"([^\"]*)\"\s*(.*)$/) {
            $emslog_event_by_id->{header}->{$1} = $2;
            $header = $3;
        } else {
            $Log->comment("No HEADER \n");
            $header = "";
        }
    }
    
    #Parse for the event and type
    if ($event =~ /^\s*(\S+)\s*(.*)/) {
        $type = $1;
        $event = $2;
    }
    
    if ($type =~ /^(\S+)_([0-9]+)$/) {
        $type = $1;
        $type =~ s/_/\./g;
    }
    
    $emslog_event_by_id->{event} = $type;

    # Parse the event parameters
    while ($event ne "") {
        if ($event =~ /^\s*([^=]+)=\"([^\"]*)\"\s*(.*)$/) {
            $emslog_event_by_id->{parameters}{$1} = $2;
            if ($emslog_event_by_id->{attribute_list} eq "") {
                $emslog_event_by_id->{attribute_list} = $1;
            } else {
                $emslog_event_by_id->{attribute_list} .= ",$1";
            }
            $event = $3;
        } else {
            $Log->comment("No EVENT \n");
            $event = "";
        }
    }
    
    #return the formatted event line
    return $emslog_event_by_id; 
}

###############################################################################
#Sub: _parseNR_emslog_line
#Input Params: $scalar
#Return values: 
#Summary: Parses the ems log file lines that are NR
###############################################################################
sub _parseNR_emslog_line {
    my($line) = @_;
    my($header);
    
    $Log->debug("Processing NR line: $line \n");
    if ($line =~ /^\s*<NR (.*)\/>\s*$/) {
        $header = "$1";
    } else {
        $Log->comment("parseNR mismatch: '$line'\n");
        return;     
    }
    
    # Parse the header (NR) section of the event record
    while ($header ne "") {
        if ($header =~ /^\s*([^=]+)=\"([^\"]*)\"\s*(.*)$/) {
            $emslog_nr{$1} = $2;
            $header = $3;
        } else {
            $header = "";
        }
    }
}

###############################################################################
#Sub: _verify_event_log
#Input Params: Array of Hash
#Return values: 0|1
#Summary: Verifies that the events specified are recorded in the ems log.
###############################################################################
sub _verify_event_log {
    my %params = (
            event => undef, #takes an array of hashes with the events
            time => "2nm",
            @_,
    );
    
    my $log_events=retrieve_log_events(time=>$params{time});
    
    foreach my $verify_events (@{$params{event}}) { #grab each event to verify
        $Log->debug("EMS verify_event_log, Here is the Event we need to verify: ".Dumper($verify_events)."\n");
        my $match=0;
        my $normalized_event=$verify_events->{event};
        $normalized_event=~s/\_/\./g;
        foreach my $logged_event (@$log_events) { #grab each event in the actual log
            $Log->trace("EMS verify_event_log, Comparing Events: $logged_event->{event} and $normalized_event \n");
            if ($logged_event->{event} eq $normalized_event) { #check to see if the events match
                if (defined $verify_events->{parameters}) { #validate the parameters
                    if (keys %{$logged_event->{parameters}} != keys %{$verify_events->{parameters}}) { #check to see if event has same number of params
                        $Log->trace("Number of parameters differ, looking for another possible match... \n");
                        next; #check to see if another version of this event exists
                    }
                    while ( my ($key, $value) = each(%{$verify_events->{parameters}}) ) { #check for each paraemeter and value
                        $Log->debug("EMS verify_event_log, Found matching event, now comparing parameter values.  Looking for $key => $value\n");
                        if ($logged_event->{parameters}->{$key} and $logged_event->{parameters}->{$key}==$value){
                        }
                        else {
                            $Log->trace("EMS verify_event_log, Failed to validate parameter $key for event $verify_events->{event} with current matching event.  Looking for another possible match. \n");
                            next; #check to see if another version of this event exists
                        }
                    }
                }
                #A matching event has been found!
                $Log->comment("Matching event found in ems log and validated: $verify_events->{event}");
                $Log->trace("Matching Event: ".Dumper($logged_event)."\n");
                $match=1;
                last; #exit the search for the event, and look for the next event
            }
        }
        #if no match could be found, tell us what event failed and break out
        if (! $match) {
            $Log->comment("No matching event $verify_events->{event} could be found in the timeframe specified \n");
            return 0;
        }
    }
    return 1;
}

###############################################################################
#Sub: _verify_event_status
#Input Params: Hash
#Return values: 0|1
#Summary: Verify that the events specified are recorded in the event status
###############################################################################
sub _verify_event_status {
    my %params = (
            pre_event_status => undef, #takes an array of hashes with the events
            post_event_status => undef,
            invoked_events => undef,
            @_,
    );
    my $pre_status=$params{pre_event_status};
    my $post_status=$params{post_event_status};
    my $invoked_events=$params{invoked_events};
    my $number_of_events=@{$invoked_events};
    
    if (!defined $params{pre_event_status} or !defined $params{post_event_status}) {
        $Log->comment("EMS _verify_event_status, Status array is empty \n");
        return 0;
    }
    
    #iterate through each element in the post_event_status array
    #verify that:
    #1) the invoked event is in the post event status array
    #2) the epoch time of the invoked event in the post event status array is greater than the epoch time of the event in the pre status array
    #3) the indications of the invoked event in the post event array is greater than the indications of the event in the pre status array
    #the array elements are compared 1 to 1, as the indexes should match up.  If an event did not exist in the pre status array
    #we will get a value of "There are no entries matching your query.", and our indication status should be 1.  We must
    #assume this otherwise we may have multiples of the same event generated and being tested.
    
    for (my $i=0; $i<$number_of_events; $i++) {
        $Log->comment("Verifying status for event $invoked_events->[$i]->{event} \n");
        if ($post_status->[$i]->{name} !~ /$invoked_events->[$i]->{event}/) {
            $Log->comment("Failed to find event $invoked_events->[$i]->{event} in the post event status listing \n");
            return 0;
        }
        if ($pre_status->[$i] eq 'There are no entries matching your query.') {
            if ($post_status->[$i]->{indications} != 1){
                $Log->comment("There is no previous status for the event, but the new indication total for the event is not 1 \n");
                return 0;
            }
        }
        else {
            if ($pre_status->[$i]->{epoch} >= $post_status->[$i]->{epoch}) {
                $Log->comment("Time indicated in event via ems event status has not increased after the recent event generation \n");
                return 0;
            }
            if ($pre_status->[$i]->{indications} >= $post_status->[$i]->{indications}) {
                $Log->comment("Indications for event via ems event status has not increased after the recent event generation \n");
                return 0;
            }
        }
        $Log->comment("EMS _verify_event_status, Event status verified for: $invoked_events->[$i]->{event} \n");    
    }
    
    return 1;
    
}
    

###############################################################################
#Sub: _get_events_from_catalog
#Input Params: 
#Return values: Ref to array of hashes
#Summary: Returns a hash of all of the possible events and their defs
###############################################################################
sub _get_events_from_catalog {
    my %params = (
            catalog => undef,
            @_,
    );
    
    $Log->debug("Using event catalog at: $params{catalog} \n");
    if (!defined $params{catalog} or $params{catalog} eq "") {
        $Log->warn("Failed to find a ems event catalog to load \n");
        return "";
    }
    my $xs = new XML::Simple();
    my $catalog_of_events = $xs->XMLin($params{catalog},ForceArray=>1,KeyAttr=>[]);
    if ($@){
        $Log->comment("Failed to retrieve the event catalog: $@\n");
        return "";
    }
    return $catalog_of_events;
}


###############################################################################
#Sub: _get_catalog
#Input Params: 
#Return values: 
#Summary: Returns the build root for the ONTAP version the filer is running
###############################################################################
sub _get_catalog {
    my $build;
    my @buildstring;
    my $catalogfile;
    
    eval {
        $build=$Server->shell(cmd=>"kenv last-OS-booted-ver");
    };
    if ($@) {
        eval {
            $build = $Server->cli(cmd=>'version');
        };
    }

    $build =~ s/\://g;
    @buildstring=split(/\s/,$build);
    
    #if tricky
    if ($buildstring[2] =~ /R*10\./ and `ls /x/eng/rlse/ontapng/$buildstring[2]/ontap/prod/build/eventCatalog-4.xml`) {
        $catalogfile = "/x/eng/rlse/ontapng/$buildstring[2]/ontap/prod/build/eventCatalog-4.xml";
    }
    #else if 7G
    elsif (`ls /x/eng/rlse/$buildstring[2]/prod/build/eventCatalog-6.xml`) {
        $catalogfile = "/x/eng/rlse/$buildstring[2]/prod/build/eventCatalog-6.xml";
    }
    #BR or higher (includes both 7M and 10M)
    else {
        $catalogfile = "$::LOGDIR/EventCatalog_Full.xml";
    }
    
    return $catalogfile;

}
###############################################################################
#Sub: _get_version
#Input Params: <none>
#Return values: Scalar of the version running
#Summary: Retrieves the current Ontap version running on the node
###############################################################################
sub _get_version {

    use NATE::BaseException qw(:try);
    my %params = (
            force=>0, # 1 = do not use cached value of $ontap_version (re-query filer)
            @_,
    );
    
    if ((!$ontap_version) || ($params{force} eq '1')) {
        
        if (!$ontap_version) {
            $Log->comment("ONTAP version is unknown; obtaining from filer...");
        }
        elsif ($params{force}) {
            $Log->comment("Force parameter specified; obtaining version from filer...");
        }

        my $build;
        my $version;
        my $Node = NACL::C::Node->find();
        my $version_manager = $Node->get_version_manager();
        my $release = $version_manager->get_version_attribute(attribute => 'release');
    
        # This is the NACL equivalent of $Server->get_release(format=>"cooked")
        try {
            $build = lc($version_manager->map_release_to_codename(release => $release));
        } catch NATE::BaseException with {
            $Log->comment("Version manager unable to map release to codename - using release string");
            $build = lc($release);
        };
        
        my $cluster_mode = "false";
        if ($Node->mode() eq 'CMode') {
            $cluster_mode = "true";
        }
        $Log->comment("NACL reports we are running $release ($build), ".$Node->mode());
    
        #if tricky
        if ($build =~ /^10\.|tricky/) {
            $version="GX";
        }
        #else if 7G
        elsif ($build =~ /^7\.|gordonbiersch|ironcity/) {
            $version="7G";
        }
        #else if BR
        elsif ($build =~ /^8\.0|boilermaker/) {
            if ($cluster_mode eq "true") {
                $version="BR-10M";
            }
            else {
                $version="BR-7M";
            }
        }
        #else if RR
        elsif ($build =~ /^8\.1|rollingrock/) {
            if ($cluster_mode eq "true") {
                $version="RR-10M";
            }
            else {
                $version="RR-7M";
            }
        }
        #else if SN 
        elsif ($build =~ /^8\.2|sierranevada/) {
            if ($cluster_mode eq "true") {
                $version="SN-10M";
            }
            else {
                $version="SN-7M";
            }
        }
        #else if FS
        elsif ($build =~ /^8\.3|fullsteam/) {
            if ($cluster_mode eq "true") {
                $version="FS-10M";
            }
            else {
                $version="FS-7M";
            }
        }        
        #otherwise, failed to find path
        else {
            $Log->warn("_get_version: Could not determine the version \n");
            return "";
        }
        $Log->comment("_get_version: Found node is version $version \n");
        $ontap_version = $version;
    }
    else {
        $Log->comment("_get_version: Using cached ONTAP version $ontap_version \n");
    }
    
    return $ontap_version;
}

###############################################################################
# Sub: _get_seconds
# Input Params: ems_date, ems_time
# Return values: seconds 
# Summary:
# Using Perl's Date::Manip module, we are taking the ems formatted
# date : ddmmmyyyy
# time : hh:mm:ss 
# and changing them into the number of seconds since 1970 using
# the method, Date_SecsSince1970() This value is much easier for
# performing comparisons. 
#
###############################################################################

sub _get_seconds {
    %params = (
        ems_date => undef,
        ems_time => undef,
        @_,
    );

    my $seconds;
    if ($params{ems_date} ne "boot"){
        my $day;
        my $mon;
        my $year;
        my $hour;
        my $min;
        my $sec;
        if ($params{ems_date} =~ /(\d+)\/(\d+)\/(\d{4})/) {
            $day = $2;
            $mon = $1;
            $year = $3;
        }
        else {
            my %month_to_int;
            $month_to_int{Jan}  = "01";
            $month_to_int{Feb}  = "02";
            $month_to_int{Mar}  = "03";
            $month_to_int{Apr}  = "04";
            $month_to_int{May}  = "05";
            $month_to_int{June} = "06";
            $month_to_int{July} = "07";
            $month_to_int{Aug}  = "08";
            $month_to_int{Sep}  = "09";
            $month_to_int{Oct}  = "10";
            $month_to_int{Nov}  = "11";
            $month_to_int{Dec}  = "12";
    
            $params{ems_date} =~ /(\d{2})(\w{3})(\d{4})/;
            $day  = $1;
            $mon  = $month_to_int{$2};
            $year = $3;
        }
        $params{ems_time} =~ /(\d+):(\d+):(\d+)/;
        $hour = $1;
        $min  = $2;
        $sec  = $3;
        
        $seconds = Date_SecsSince1970($mon,$day,$year,$hour,$min,$sec);
    } else {
        $seconds = 0;
        }
    
    return $seconds;
}

###############################################################################
# Sub: _generate_event
# Input Params: 
# Return values: 
# Summary:
#
###############################################################################

sub _generate_event {
    my $event_options=shift;
    
    my $version=_get_version();
    if ($version=~/7M|7G/) {
        my $output;
        eval {
            $output=$Server->EMSInvoke(%{$event_options});
        };
        if ($@) {
            return $@;
        }
        if ($output=~/ERROR: insufficient arguments/ or $output=~/ERROR: invalid event id/ or $output=~/ERROR:/) {
            return $output;
        }
    }
    else {
        eval {
            if ($event_options->{arguments} eq "") {
                $Server->event_generate(messagename=>$event_options->{event});
            }
            else {
                $Server->event_generate(messagename=>$event_options->{event},values=>$event_options->{arguments});
            }
        };
        if ($@) {
            return $@;
        }
    }
    return 1;
}

###############################################################################
# Sub: create_EMS_eventcatalog
# Input Params: 
# Return values: 
# Summary: Creates a single event catalog file from the set of .ems files
#
###############################################################################

sub create_EMS_eventcatalog {

    # Variables
    my $ind;
    my $count = 0;
    my $CatalogFileName = "EventCatalog_Full.xml";
    my $catalogFile = "$::LOGDIR/$CatalogFileName";
    my $str1 = "\n</event-catalog>\n";
    my $str2 = "\n</event-definitions>\n";
    my $lastline_tag;
    my $catalogString;
    my $version=_get_version();

    my $tmpVar = `ls $catalogFile`;
    if ($tmpVar =~ m/^$catalogFile$/) {
       $Log->comment("Event Catalog Exists, no need to create another one\n");
       return 1;
    }

    eval {
        $build=$Api_Set_SShell->execute_raw_command(command=>"kenv last-OS-booted-ver");
    };
    if ($@) {
        eval {
            $build = $Server->cli(cmd=>'version');
        };
    }

    $build =~ s/\://g;
    my @buildstring=split(/\s/,$build);

    my $R_R_B_R_BuildDir;
    if ( $build =~ /IBM/ ) {
        $R_R_B_R_BuildDir
            = "/x/eng/rlse/DOT/$buildstring[3]/final/bedrock/export/x86_64/image/etc/ems/";
    } else {
        $R_R_B_R_BuildDir
            = "/x/eng/rlse/DOT/$buildstring[2]/final/bedrock/export/x86_64/image/etc/ems/";
    }

    # If 7M/7G do not create EventCatalog from filer till burt #xxxxxx is fixed
    # since systemshell is not accessible, we won't be able to get to the ems files

    if ($version =~ /7M/) {
       if ($params->{EVENT_CATALOG}) { #if Event catalog path is defined by user
           $R_R_B_R_BuildDir = "$params->{EVENT_CATALOG}";
       }
       else { # Use the catalog files in the build to construct the new one
           $Log->debug("Event Catalog path not specified, using the catalogs from build path for $buildstring[2]\n");
       }
       # copy the ems files into a string
       # For boilermaker 7M, events outside dblade are not recognized. Hence use only the dblade.ems file alone
       if ($buildstring[2] =~ /boilermaker/) {
            $catalogString = `cat $R_R_B_R_BuildDir/dblade_ems.ems`;
       } else {
            $catalogString = `cat $R_R_B_R_BuildDir/*.ems`;
       }
    } elsif ($version =~ /7G/) { # If this is 7G, then just return. The path will be provided in _get_catalog
       return 1;
    } else { # If 10M, we get all details from the filer
       $Api_Set_SShell->set_timeout("connectrec-timeout" =>"180");
       $catalogString=$Api_Set_SShell->execute_raw_command(command=>"cat /etc/ems/*.ems",log_level=> "debug");
    }

    my @catalogArr = split('\n', $catalogString);
    my $fdout = new IO::File( ">>$catalogFile" );

    foreach my $line (@catalogArr) {

        # We need the first three lines of the file so we just write them
        if ($count < 3) {
           if ($count == 1) {
              $line =~ /^([^\s]+)/;
              $lastline_tag = $1;
           }
           print $fdout "$line\n";
           $count++;
           next;
        }

        # Match the following XML tags and remove them, this is done to avoid 
        # parsing errors
        if ($line =~ m/<event-definitions/) {
            $ind = index($line, "<event-definitions");
            substr($line, $ind, length("<event-definitions"), "");
        } 
        if ($line =~ m/<\/event-definitions>/) {
            $ind = index($line, "</event-definitions>");
            substr($line, $ind, length("</event-definitions>"), "");
        } 
        if ($line =~ m/<event-catalog/) {
            $ind = index($line, "<event-catalog");
            substr($line, $ind, length("<event-catalog"), "");
        } 
        if ($line =~ m/<\/event-catalog>/) {
            $ind = index($line, "</event-catalog>");
            substr($line, $ind, length("</event-catalog>"), "");
        } 
        if ($line =~ m/xml version/) {
            $line = " ";
        }

        print $fdout "$line\n";
        $count++;
    }

    # Add the final event-destination/event-catalog tag
    my $str_tmp;
    if ($lastline_tag =~ m/</) {
       $lastline_tag =~ s/<//;
    }
    $str_tmp = "\n</$lastline_tag>\n";

    print $fdout $str_tmp;
    $fdout->close;

    return 1;
}

###############################################################################
# Sub: _generate_param_value
# Input Params: parameter type
# Return values: randomly generated parameter values
# Summary: Genereate random parameter value according to the type
#
###############################################################################

sub _generate_param_value {
    my $me=(caller(0))[3];
    my ($type, $length) = @_;
    $type = uc $type;
    my $maxINT=2147483647;
    my $maxLONGINT=9223372036854775807;
    my @chars = (a..z,A..Z,0..9);
    my $value;

    switch ($type) {
         case /LONGINT/ { #type LONGINT or LONGINTHEX, unsigned
            $value = int(rand($maxLONGINT));
         }
         case /INT/ { #type INT or INTHEX, unsigned.
            $value = int(rand($maxINT));
         }
         case 'STRING' { 
            $length = $length ? $length : int(rand(20) + 5);
            foreach (1..$length) {
                $value.=$chars[int(rand @chars)];
            }
         }
         else { #in case there are new types
            $Log->warn("$me unknown type $type");
            $value = int(rand($maxINT));
         }
    }
    return ($type =~ /HEX/) ? '0x'.sprintf("%0x",$value) : $value;
}

=pod

=head1 NAME

EMS.pm - EMS event generation and verification library for 7G, BR 7mode, and BR CM.


=head1 SYNOPSIS

    use EMS;
    EMS::invoke_event(
            event => "app.log.crit",
            arguments => "1 2 3 4 5 6",
            trap => 1,
            log => 1,
            verify => "both"
            );


=head1 DESCRIPTION

The methods included in this library were developed in order to assist
engineers in generation and verification of ems events.  The current library
support 7G and 7mode filers and invokes communication to the filer via the
Server.pm library.  The EMS.pm library allows the user to generate events,
verify that the event was recorded in the ems log, verify that the event status
is updated, and also allows the user to return a portion of the ems logs.
  
=head1 PARAMETERS

User can set the EVENT_CATALOG parameter via command line.  This parameter will
override the event catalog to use for event lookup and generation.  This should
be the full path the eventCatalog.xml file.
    
=head1 METHODS

    set_autosuppression(enable=>"on|off");
        Description:
            This method will set the options ems.autosuppress.enable option.
        Inputs:
            enable => <on|off> Enable or disable ems suppression.
        Outputs:
            <none>
        Examples:
            see full example below
            
    
    set_log_dup_suppression(enable=>"on|off");
        Description:
            This method will set the options ems.log.dupsuppress.enable option.
        Inputs:
            enable => <on|off> Enable or disable duplicate ems event
                suppression in the log.
        Outputs:
            <none>
        Examples:
            see full example below


    invoke_event(event=>"event");
        Description:
            This method will invoke the event specified with the paramters
            indicated.
        Inputs:
            event => <text> The event you wish to generate.  Filters are also
                acceptable here.
            log => <0|1> Determines whether the event should attempt to be
                recorded in the syslog or not.
            trap => <0|1> Determines whether the event should attempt trap
                generation or not.
            version => <integer> Version of the event to generate
            severity => <NODE_FAULT|SVC_FAULT|NODE_ERROR|SVC_ERROR|WARNING|
                       NOTICE|INFO|DEBUG> Severity to generate
            arguments => <text> Comma separated list of values to pass to the
                event parameters.  If arguments are not specified, the method
                will randomly generate them.  If multiple events are listed
                in the event field and arguments are specified, the same
                arguments will be used for all events.
            verify => <both|log|status|none> Should the event be validated
                in the ems log, ems status, both, or neither.
            suppress_invocation_failures => <0|1> Determines whether the
                routine should fail if a single event invocation fails.  The
                default is 0, to fail when an invocation fails.
        Outputs:
            $invoked_events - returns a ref to an array of hashes containing
                all of the events that have been invoked.
        Examples:
            see full example below
           
            
    invoke_random_event();
        Description:
            This method will invoke a random ems event for the caller.
        Inputs:
            event => <text> The events or filter of events you wish to select
                from.
            log => <0|1> Determines whether the event should attempt to be
                recorded in the syslog or not.
            trap => <0|1> Determines whether the event should attempt trap
                generation or not.
            verify => <both|log|status|none> Should the event be validated
                in the ems log, ems status, both, or neither.
            ignore_timer_suppressed_events => <0|1> Should we not choose a
                random event that may be suppressed by timer suppression.
                Default is 0, meaning it is okay to choose an event that may
                be suppressed.
        Outputs:
            $invoked_event - returns a ref to an array of hashes containing
                the event that has been invoked.
        Examples:
            see full example below
            
            
    invoke_random_snmp_event();
        Description:
            This method will invoke a random ems event that is guaranteed to
            generate a trap for the caller.
        Inputs:
            event => <text> The events or filter of events you wish to select
                from.
            log => <0|1> Determines whether the event should attempt to be
                recorded in the syslog or not.
            verify => <both|log|status|none> Should the event be validated
                in the ems log, ems status, both, or neither.
            ignore_timer_suppressed_events => <0|1> Should we not choose a
                random event that may be suppressed by timer suppression.
                Default is 0, meaning it is okay to choose an event that may
                be suppressed.
        Outputs:
            $invoked_event - returns a ref to an array of hashes containing
                the event that has been invoked.
        Examples:
            see full example below
            
            
    invoke_random_syslog_event();
        Description:
            This method will invoke a random ems event that is guaranteed to
            generate a syslog entry for the caller.
        Inputs:
            event => <text> The events or filter of events you wish to select
                from.
            trap => <0|1> Determines whether the event should attempt to
                generate a trap.
            verify => <both|log|status|none> Should the event be validated
                in the ems log, ems status, both, or neither.
            ignore_timer_suppressed_events => <0|1> Should we not choose a
                random event that may be suppressed by timer suppression.
                Default is 0, meaning it is okay to choose an event that may
                be suppressed.
        Outputs:
            $invoked_event - returns a ref to an array of hashes containing
                the event that has been invoked.
        Examples:
            see full example below


    invoke_kernel_event();
        Description:
            This method will use a command line to generate a kernel level
            event.
        Inputs:
            <none>
        Outputs:
            <none>
        Examples:
            see full example below    


    invoke_event_without_error_checking();
        Description:
            This method will bypass the error checking and catalog loading
            to allow events to be invoked faster.
        Inputs:
            event => The event which the user wished to generate.  Default 
                is "app.log.crit".
            arguments => Parameters to be used for the event generation.
                Default is "1 2 3 4 5 6".
            num_of_invokes => Number of times to invoke the event.  Default 
                is 1.
        Outputs:
            1 or error
        Examples:
            see full example below                    
         
            
    retrieve_log_events();
        Description:
            This method will return all of the ems events in the log based
            upon the time frame passed in.
        Inputs:
            time => <time> Specify the timeframe for retrieving of the events.
                The format is <number>n<units>.  For example, 10nm = 10 mins.
        Outputs:
            $log_events - returns a ref to an array of hashes containing
                the events from the ems log for the specified time frame.
        Examples:
            see full example below
            
            
    get_status();
        Description:
            This method will return the current ems status.
        Inputs:
            <none>
        Outputs:
            $status_hash - returns a ref to an array of hashes containing
                the ems status
        Examples:
            see full example below
            
    
    get_event_status();
        Description:
            This method will return the ems event status based upon the
            filters that were specified.
        Inputs:
            event => <text> The event or events to get status for.
            severity => <NODE_FAULT|SVC_FAULT|NODE_ERROR|SVC_ERROR|WARNING|
                NOTICE|INFO|DEBUG> The event severity to filter for.
            sort_by_date=> <0|1> Should the array of hashes be ordered by time.
                Currently the array is in alphabetical order by event name.
        Outputs:
            $event_status_hash - returns a ref to an array of hashes containing
                the ems event status
        Examples:
            see full example below 
    
    
    configure_destination();
        Description:
            This method will configure a destination on the filer to send
            remote ems messages to.  If a messagename is provided, it will also
            configure that exact message or messages to be sent to the newly
            configured destination.
        Inputs:
            name => <text> The name of the new destination to configure.
            syslog => <text> The name or IP address of the syslog destination.
            snmp => <text> The name or IP address of the snmp trap receiver.
            mail => <text> The name or IP address of the email server.
            syslog_facility => <text> The syslog facility to send messages from.
            snmp_community => <text> The snmp community string to send traps.
            mailfrom => <text> The email addresses to send events to.
            client => <text> Name of the client to invoke the email, syslog, or
                snmp server on.
            messagename => <event> Name of the event you want configured to
                send to the destination.
        Outputs:
            1 (on success) or error
        Examples:
            see full example below
            
    
    remove_destination();
        Description:
            This method will remove the specified destination from any ems
            events configured, then delete the specified destination.  For 7M
            it will simply disable that remote protocol.
        Inputs:
            name => <text> The name of the destination to remove.
            snmp => <text> The name of the SNMP trap destination to remove.
            protocols => <snmp|syslog|smtp> The protocol to disable.
        Outputs:
            1 (on success) or error
        Examples:
            see full example below 

   create_EMS_eventcatalog();
        Description:
            Creates a single event catalog file from the set of .ems files.
            The .ems files are taken from the filer(/etc/ems). 
        Inputs:
            none
        Outputs:
            1 (on success) or error
        Examples:
            see full example below 
           
            
=head1 EXAMPLES

Ex1: Invoke a single specified event with the arguments indicated and
    no event verification.

    EMS::invoke_event(event=>"app.log.crit",arguments=>"1 2 3 4 5 6",
        verify=>"none");

Ex2: Invoke a single specified event.  Let the library randomly choose
    and populate the arguments/parameters.  Then verify the event was
    written to the ems log and the event status table.
    
    EMS::invoke_event(event=>"app.log.crit", verify=>"both");

Ex3: Invoke all of the events in the specified filters with the same
    arguments.
    Verify that the event status is correctly reflected for each event.
    
    EMS::invoke_event(event=>"app.*.crit,app.*.info,*.log.emerg",app.log.debug,
        arguments=> "1 2 3 4 5 6", verify=>"status");
    
Ex4: Invoke a random event and only verify that it was written to the
    ems log.

    my $random_event=EMS::invoke_random_event(verify=>"log");
    
Ex5: Invoke a random event from the filter specified.  Generate traps
    and syslog messages if the event is enabled to do so.
    
    my $random_event=EMS::invoke_random_event(event=>"app.*.crit,app.*.info,
        app.log.debug",log=>1,trap=>1);

Ex6: Randomly invoke a ems event that will generate a trap.
    
    my $random_snmp_event=EMS::invoke_random_snmp_event(log=>1);
    
Ex7: Randomly invoke a ems event that will generate a syslog entry.
    
    my $random_syslog_event=EMS::invoke_random_syslog_event();
    
Ex8: Retrieve the last 1hours worth of ems log events.

    my $events=EMS::retrieve_log_events(time=>"1nh");

Ex9: Get the current event status for the indicate event.

    my $event_status = EMS::get_event_status(event=>"app.log.crit");

Ex10: Get the current event status, ordered by time.

    my $event_status = EMS::get_event_status(sort_by_date=1);
    
Ex11: Get the current event status based upon a filter.

    my $event_status = EMS::get_event_status(
                                 event => "kern*",
                                 severity => "NOTICE",
                                 sort_by_date => 1
                                 );

Ex11: Get the current ems status
    
    my $status = EMS::get_status();
    
Ex12: Generate a kernel event
  
    my $kernel_event = EMS::invoke_kernel_event();

Ex13:  Generate an events quickly by bypassing error checking

    EMS::invoke_event_without_error_checking(
                                 event=>"app.log.crit", 
                                 num_of_invokes=>6
                                 );                                               

    
=cut
