# $Id: //depot/prod/test/Rsierranevada/lib/EnGenX/Controller/Aggregate.pm#2 $

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Aggregate Controller Layer Module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

EnGenX::Controller::Aggregate

=head1 DESCRIPTION

EnGenX::Controller::Aggregate is a derived class of
EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP Aggregates. All methods are static.

=cut

package EnGenX::Controller::Aggregate;

# Compiler directives.
use strict;
use warnings;

# NATE and Standard Module import.
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::STask::Aggregate;
use NACL::C::Aggregate;
use NACL::C::Job;
use EnGenX::SubProcess;
use EnGenX::EnGenXUtil;
use EnGenX::Controller::Volume;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name       => 'aggregate',
    num_disks  => 'diskcount',
    _reuse     => 'if_exists',
    disk_owner => 'nodes',
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (

    wafliron => '',
    state    => '',
    parallel => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface      => '',
    apiset_must            => '',
    apiset_should          => '',
    aggregate              => '',
    force                  => '',
    job_component          => '',
    'ignore-vldb'          => '',
    plex                   => '',
    'dblade-only'          => '',
    'hidden-plex'          => '',
    'preserve-config-data' => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple aggregates serially
or in parallel and return whether the aggregate(s) was created successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Aggregate->create(
            command_interface => $Node,
            aggregate         => $Aggr_name,
            diskcount         => 3,
            job_component     => \$job,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Aggregate->create(
            command_interface => $Node,
            aggregate         => $Aggr_name,
            diskcount         => 3,
            job_component     => \$job,
            'method-timeout'  => 3600,
            count             => 5
            );

  EnGenX::Controller::Aggregate->create(
            command_interfac  => $Node,
            resource_list      => @Aggregates,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Aggregate->create(
            command_interface => $Node,
            resource_list      => @Aggregates,
            );

  EnGenX::Controller::Aggregate->create(
            resource_list      => @Aggregates,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Aggregate->create(
            resource_list      => @Aggregates,
            );

=over

=item Options

Either the resource_list is required or the NACL Aggregate Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the aggregates in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Aggregate Creation Arguments should
NOT be passed in.

=item C<< NACL Aggregate Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::C::Aggregate->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< count => $aggregate_count >>

(Optional)
This arguments is used to create multiple aggregates from the same NACL Aggregate
Creation Arguments.  To create the names for the aggregates, the aggregate name
passed in is appended with the index that it was create at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Aggregate(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "aggregate";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple aggregates serially
or in parallel and return whether the aggregate(s) was deleted successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Aggregate->delete(
            command_interface => $Node,
            aggregate           => $Aggr_name,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Aggregate->delete(
            command_interface => $Node,
            aggregate           => $Aggr_name,
            count             => 5
            );

  EnGenX::Controller::Aggregate->delete(
            command_interface => $Node,
            resource_list      => @Aggregates,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Aggregate->delete(
            command_interface => $Node,
            resource_list      => @Aggregates,
            );

  EnGenX::Controller::Aggregate->delete(
            resource_list      => @Aggregates,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Aggregate->delete(
            resource_list      => @Aggregates,
            );

=over

=item Options

Either the resource_list is required or the NACL Aggregate Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the aggregates in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Aggregate Deletion Arguments should
NOT be passed in.

=item C<< NACL Aggregate Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::C::Aggregate->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< count => $aggregate_count >>

(Optional)
This arguments is used to delete multiple aggregates from the same NACL Aggregate
Deletion Arguments.  To delete the names for the aggregates, the aggregate name
passed in is appended with the index that it was deleted at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Aggregate(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "aggregate";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a aggregate.
# Details:    Makes the NACL::STask::Aggregate->create() call and
#             makes the NACL::STask::Aggregate->wait_for_creation call. Also
#             works around some common non-fatal errors that can be return from
#             aggregate creation and then has a retry mechanism to retry
#             aggregate creation.
# TODO:       Need to update to NACL::STask::Aggregate->create()
# Inputs:     NACL::C::Aggregate->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the aggregate was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;
    # Locals
    my $duplicated = FALSE;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }
    # Default diskcount to 5 if user did not pass a diskcount or disklist.
    $args{diskcount} = 5 unless (exists($args{diskcount}) or exists($args{disklist}));

    try {
        NACL::C::Aggregate->create(%args);
        $Log->comment("Using newly Created Aggregate:" . $args{aggregate});
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if (   $if_exists eq 'reuse'
            && $exception->text() =~ /An aggregate already uses .* as name/)
        {
            $duplicated = TRUE;
            $Log->comment(
                "Using already Created Aggregate:" . $args{aggregate});
        } else {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    };#catch

    if (!$duplicated) {
        my %wait_args = (
            command_interface => $args{command_interface},
            aggregate         => ($args{aggregate}),
            'method-timeout'  => $args{'method-timeout'},
        );
        $wait_args{apiset_should} = $args{apiset_should}
            if exists $args{apiset_should};
        NACL::STask::Aggregate->wait_for_creation(%wait_args);
    }

    $Log->exit();
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>$duplicated) : TRUE;

} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a aggregate.
# Details:    Makes the NACL::STask::Aggregate->delete() calls. Also works around
#             some common non-fatal errors that can be return from aggregate
#             deletion and then has a retry mechanism to retry aggregate deletion.
# Inputs:     NACL::STask::Aggregate->delete arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my $pkg   = shift;
    my %args = @_;
    # Locals
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    my $end_time = time() + $args{'method-timeout'};

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    while (1) {
        my $delete_suceeded = TRUE;
        try {
            NACL::STask::Aggregate->purge(%args, if_purged => 'pass');
            $Log->comment("Aggregate:" . $args{aggregate} . " was deleted");
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $exception = shift;
            if (time() < $end_time
                && $exception->text() =~ /aggregate is currently busy/)
            {
                $delete_suceeded = FALSE;
            } else {
                $Log->exit();

                # Rethrowing caught exception
                $exception->throw();
            }
        };
        last if ($delete_suceeded);
        $Log->trace("Snoozing for 10 seconds");
        Tharn::snooze(10);
    }

    $Log->exit();
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Aggregate->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Aggregate->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
