#
# Copyright (c) 2013 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

EnGenX::Controller::Audit

=head1 DESCRIPTION

EnGenX::Controller::Audit is used to Config, Setup, Cleanup and
Deconfig Audit configurations on a vserver.

=cut

package EnGenX::Controller::Audit;

# Compiler directives
use strict;
use warnings;
use Data::Dumper;

# NATE and Standard Module import
use NATE::Log qw(log_global);
use NATE::Util::DataTypes qw(TRUE FALSE);

# NACL libs
use NACL::C::VserverAudit;

# Exceptions
use NATE::BaseException qw(:try);

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args
            $Log);

$Log = log_global();

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    vserver => ''
);

=head1 METHODS

=head2 create

(Static method) Request to create a cifs service with the following options

  EnGenX::Controller::Audit->create(command_interface => $node,
                                    vserver           => $vserver_name,
                                    destination       => "/myvolume",
                                    %other_opts);

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver to create the audit configuration.

=item C<< destination => $destinatino >>

(Required)
The mount point to store the audit data. This must be a valid junction path that contains an active volume mount.

Example: destination => "/myvolume"

=item C<< %other_opts >>

All other options that can be passed into NACL::C::VserverAudit->create() can be passed into this method.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Audit configuration.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "audit";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit();

    return $parent_answer;
}

=head2 delete

(Static method) Request to delete an Audit configuration with the following options

  EnGenX::Controller::Audit->delete(command_interface => $node,
                                    vserver           => $vserver_name);

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver to create the cifs service for.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the delete of the Audit configuration based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "audit";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit();

    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the audit configuration
# Details:    Makes the NACL::C::VserverAudit->create().
#
# Inputs:     NACL::C::VserverAudit->create arguments
# Outputs:    Returns whether the create was successful
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $reused = FALSE;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 3;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    # create the Audit configuration
    try {
        NACL::C::VserverAudit->create(%args);
    } catch NACL::APISet::Exceptions::ConnectionFailedException with {
        my $e = shift;
        # retry if we are unable to connect
        if ($retry > 0) {
            sleep 5;
            $retry--;
            $pkg->_nacl_create_call(%args,
                                    retry => $retry,
                                    if_exists => $if_exists);
        } else {
            # rethrow the exception
            $Log->exit();
            $e->throw();
        }
    } catch NACL::C::Exceptions::VserverAudit::AlreadyExists with {
        my $e = shift;
        if ($if_exists =~ /reuse/i) {
            NACL::C::VserverAudit->modify(%args);
        } else {
            $Log->exit();
            $e->throw();
        }
    };

    # enable the audit configuration on the vserver
    # XXX todo: might need to also pull in common nacl parameters
    try {
        NACL::C::VserverAudit->enable(command_interface => $args{command_interface},
                                      vserver => $args{vserver});
    } catch NACL::C::Exceptions::VserverAudit::AlreadyEnabled with {
        # vserver audit is already enabled, ignore it
        $Log->comment("Vserver audit is already enabled on vserver $args{vserver}");
    };

    $Log->exit();
    return (wantarray()) ? ('success' => TRUE, 'duplicated' => $reused) : TRUE;
}

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deleting the audit service
# Details:    Makes the NACL::C::VserverAudit->delete().
#
# Inputs:     NACL::C::VserverAudit->delete arguments
# Outputs:    Returns whether the delete was successful
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # disable the audit configuration on the vserver
    my $does_not_exist = 0;
    try {
        NACL::C::VserverAudit->disable(%args);
    } catch NACL::C::Exceptions::VserverAudit::AlreadyDisabled with {
        # ignore if it's already disabled
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        $does_not_exist = 1;
    };

    if ($does_not_exist) {
        $Log->exit();
        return;
    }

    # delete the audit configuration
    my $retry = 10;
    my $success = 0;
    my $e;
    while ($retry > 0) {
        try {
            NACL::C::VserverAudit->delete(%args);
            $retry = 0;
            $success = 1;
        } catch NACL::APISet::Exceptions::CommandFailedException with {
            $e = shift;
            if ($e->text() =~ /Retry after sometime/i) {
                # sleep and retry the audit deletion up to ten times
                $Log->comment("Could not delete Vserver Audit on vserver " .
                              "$args{vserver}, will retry in 60 seconds (count=$retry)");
                sleep 60;
                $retry--;
            } else {
                $Log->exit();
                $e->throw();
            }
        };
    }

    # if we were not successful in deleting the audit configuration, throw the
    # exception
    if (not $success) {
        $Log->exit();
        $e->throw();
    }

    $Log->exit();
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  NOOP for Audit
# Details:
# Inputs:     None
# Outputs:    None
# Exceptions: None
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
