# $Id: //depot/prod/test/main/lib/EnGenX/Controller/CIFS.pm $

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

EnGenX::Controller::CIFS

=head1 DESCRIPTION

EnGenX::Controller::CIFS is used to Config, Setup, Cleanup and
Deconfig CIFS services on a vserver.

=cut

package EnGenX::Controller::CIFS;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL libs
use NACL::C::VserverCifs;

# Exceptions
use NATE::BaseException qw(:try);

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args
            $Log);

$Log = log_global();

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    name => '',
    parallel      => '',
    service_enabled => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    vserver => '',
    username => '',
    password => '',
    'netbios-aliases' => ''
);

=head1 METHODS

=head2 create

(Static method) Request to create a cifs service with the following options

  EnGenX::Controller::CIFS->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            domain                          => $domain,
            username                        => $username,
            password                        => $password
            );

  EnGenX::Controller::CIFS->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            domain                          => $domain,
            username                        => $username,
            password                        => $password,
            'cifs-server'                   => "mycifsserver",
            'status-admin'                  => "up",
            'default-site'                  => $default_site,
            ou                              => $ou
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver to create the cifs service for.

=item C<< domain => $domain >>

(Required)
This is the active directory domain controller to use for this CIFS service.

=item C<< username => $username >>

(Required)
This is the username for the active directory domain controller.

=item C<< password => $password >>

(Required)
This is the password for the active directory domain controller.

=item C<< status-admin => $status_admin >>

(Optional)
The defualt is to have the service up and running. The user can specify
the service be down on creation.

=item C<< cifs-server => $cifs-server >>

(Optional)
The default value of the cifs server is the vserver name.

=item C<< ou => $ou >>

(Optional)

=item C<< default-site => $default_site >>

(Optional)

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the CIFS Targets based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit();

    return $parent_answer;
}

=head2 delete

(Static method) Request to delete a cifs service with the following options

  EnGenX::Controller::CIFS->delete(
            command_interface               => $node,
            vserver                         => $vserver_name,
            username                        => $username,
            password                        => $password
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver to create the cifs service for.

=item C<< username => $username >>

(Required)
This is the username for the active directory domain controller.

=item C<< password => $password >>

(Required)
This is the password for the active directory domain controller.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the delete of the CIFS Targets based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit();

    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the cifs service.
# Details:    Makes the NACL::C::VserverCIFS->create().
#
# Inputs:     NACL::C::VserverCIFS->create arguments
# Outputs:    Returns whether the create was successful
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $reused = TRUE;
    my $if_exists = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 3;
    my $netbios_aliases = exists $args{'netbios-aliases'} ? delete $args{'netbios-aliases'} : undef;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    # default the cifs server name to be the vserver name if it was not provided
    # this is needed because the cifs server name must be unique to the entire cluster.
    if (not $args{'cifs-server'}) {
        $args{'cifs-server'} = $args{'vserver'};
    }

    # create the CIFS service
    try {
        NACL::C::VserverCifs->create(%args);
    } catch NACL::APISet::Exceptions::ConnectionFailedException with {
        my $e = shift;
        # retry if we are unable to connect
        if ($retry > 0) {
            sleep 5;
            $retry--;
            $pkg->_nacl_create_call(%args,
                                    retry => $retry,
                                    if_exists => $if_exists);
        } else {
            # rethrow the exception
            $Log->exit();
            $e->throw();
        }
    } otherwise {
        my $e = shift;
        if ($e->text() =~ /Only one CIFS server is allowed|duplicate entry/i and $if_exists =~ /reuse/i) {
            # lets just reuse the cifs service already there, blindly modify the service that
            # exists
            $reused = TRUE;

            # pull in the arguments used by the modify command
            my %modify_args = ();
            if (exists($args{'default-site'})) {
                $modify_args{'default-site'} = $args{'default-site'};
            }

            if (exists($args{'status-admin'})) {
                $modify_args{'status-admin'} = $args{'status-admin'};
            }
            # only execute the modify command if there are things to modify
            if (scalar(keys %modify_args)) {
                NACL::C::VserverCifs->modify(%modify_args);
            }
        } else {
            # otherwise rethrow the exception
            $Log->exit();
            $e->throw();
        }
    };

    # Add Netbios-aliases if they exist
    if($netbios_aliases ne undef) {
        # Create array ref from list of netbios-aliases
        $args{'netbios-aliases'} = [ split(/,/ , $netbios_aliases) ];
        
        try{
            NACL::C::VserverCifs->add_netbios_aliases(command_interface => $args{command_interface},
                                                      'vserver'         => $args{vserver},
                                                      'netbios-aliases' => $args{'netbios-aliases'});
        
            $Log->comment("Netbios-aliases added for vserver " . $args{vserver});
        } catch NACL::APISet::Exceptions::ConnectionFailedException with {
            my $e = shift;
            # retry if we are unable to connect
            if ($retry > 0) {
                sleep 5;
                $retry--;
                $pkg->_nacl_create_call(%args,
                                        retry => $retry,
                                        if_exists => $if_exists);
            } else {
                # rethrow the exception
                $Log->exit();
                $e->throw();
            }
        } catch NACL::APISet::Exceptions::ResponseException with {
            my $e = shift;
            if ($e->text() =~ /already exists|duplicate entry/i) {
                $Log->comment("Netbios-aliases already exist for vserver " . $args{vserver});
            } else {
                # otherwise rethrow the exception
                $Log->exit();
                $e->throw();
            }
        } otherwise {
            my $e = shift;
           
            # otherwise rethrow the exception
            $Log->exit();
            $e->throw();
        };
    }

    $Log->exit();
    return (wantarray()) ? ('success' => TRUE, 'duplicated' => $reused) : TRUE;
}

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deleting the cifs service
# Details:    Makes the NACL::C::VserverCifs->delete().
#
# Inputs:     NACL::C::VserverCifs->delete arguments
# Outputs:    Returns whether the delete was successful
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my ($pkg, %args) = @_;
    my $retry = exists $args{retry} ? delete $args{retry} : 3;
    my $if_exists = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $netbios_aliases = exists $args{'netbios-aliases'} ? delete $args{'netbios-aliases'} : undef;

   # Remove Netbios-aliases if they exist
    if($netbios_aliases ne undef) {
        # Create array ref from list of netbios-aliases
        $netbios_aliases = [ split(/,/ , $netbios_aliases) ];
        try {
            NACL::C::VserverCifs->remove_netbios_aliases(command_interface => $args{command_interface},
                                                        'vserver'          => $args{vserver},
                                                        'netbios-aliases'  => $netbios_aliases); 
        } catch NACL::APISet::Exceptions::ConnectionFailedException with {
                my $e = shift;
                # retry if we are unable to connect
                if ($retry > 0) {
                    sleep 5;
                    $retry--;
                    $pkg->_nacl_delete_call(%args,
                                            retry => $retry,
                                            if_exists => $if_exists);
                } else {
                    # rethrow the exception
                    $Log->exit();
                    $e->throw();
                }   
            } catch NACL::APISet::Exceptions::ResponseException with {
                my $e = shift;
                if ($e->text() =~ /does not exist|no entry/i) {
                    $Log->comment("Netbios-aliases already removed for vserver " . $args{vserver});
                } else {
                    # otherwise rethrow the exception
                    $Log->exit();
                    $e->throw();
                }   
            } otherwise {
                my $e = shift;

                # otherwise rethrow the exception
                $Log->exit();
                $e->throw();
            }; 
    }

    # delete the cifs service
    try {
        NACL::C::VserverCifs->delete(%args);
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        # no cifs service to delete
        $Log->comment("Vserver $args{vserver} does not have a CIFS service to delete");
    };


    $Log->exit();
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  NOOP for CIFS
# Details:
# Inputs:     None
# Outputs:    None
# Exceptions: None
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
