# $Id$

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

EnGenX::Controller::CIFS::Branchcache

=head1 DESCRIPTION

EnGenX::Controller::CIFS::Branchcache is used to Config, Setup, Cleanup and
Deconfig CIFS Branchcache settings on a vserver

=cut

package EnGenX::Controller::CIFS::Branchcache;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
use NATE::Util::DataTypes qw(TRUE FALSE);

# NACL libs
use NACL::C::VserverCifsBranchcache;

# Exceptions
use NATE::BaseException qw(:try);

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args
            $Log);

$Log = log_global();

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    cifs_service_username =>'',
    cifs_service_password =>''
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    vserver     => '',
    'flush-hashes' => ''
);

=head1 METHODS

=head2 create

(Static method) Request to create a cifs branchcache setting with the following options

  EnGenX::Controller::CIFS::Branchcache->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            'hash-store-path'               => "/myhashstorepath"
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver to use

=item C<< 'hash-store-path' => $path >>

(Required)
The volume junction path to store the hash store.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=back

=cut

sub create {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs_branchcache";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit();

    return $parent_answer;
}

=head2 delete

(Static method) Request to delete a cifs branchcache with the following options

  EnGenX::Controller::CIFS::Branchcache->delete(
            command_interface               => $node,
            vserver                         => $vserver_name,
            'flush-hashes'                  => "true"
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver to use

=item C<< 'flush-hashes' => "{true|false}" >>

(Required)
Boolean string 'true' or 'false', specifies whether to flush the hashes or not.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=back

=cut

sub delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs_branchcache";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit();

    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the cifs branchcache
#
# Inputs:     NACL::C::VserverCifsBranchcache->create arguments.
# Outputs:    Returns whether the create was successful.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $reused = FALSE;
    my $if_exists
        = exists($args{if_exists})
        ? delete($args{if_exists}) : 'reuse';
    my $retry
        = exists $args{retry}
        ? delete $args{retry} : 3;
    $args{'method-timeout'}
        = 3600 unless exists($args{'method-timeout'});
    
    # remove flush-hashes from the args since it is only needed for the delete method
    delete($args{'flush-hashes'});

    # convert 'versions' into an array reference
    if (exists($args{'versions'}) and ref($args{'versions'}) ne "ARRAY") {
        $args{'versions'} = [split(qr/\s*,\s*/, $args{'versions'})];
    }

    try {
        NACL::C::VserverCifsBranchcache->create(%args);
    } catch NACL::APISet::Exceptions::ConnectionFailedException with {
        my $e = shift;
        # retry if we are unable to connect
        if ($retry > 0) {
            sleep 5;
            $retry--;
            $pkg->_nacl_create_call(%args,
                                    retry => $retry,
                                    if_exists => $if_exists);
        } else {
            # rethrow the exception
            $Log->exit();
            $e->throw();
        }
    } catch NACL::APISet::Exceptions::CommandFailedException with {
        my $e = shift;
        if ($e->text() =~ /duplicate entry/i and $if_exists =~ /reuse/i) { 
            # reuse the branch cache entry
            $reused = TRUE;
        } else {
            $Log->exit();
            $e->throw();
        }
    };

    $Log->exit();
    return (wantarray()) ? ('success' => TRUE, 'duplicated' => $reused) : TRUE;
}

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deleting a cifs branchcache
#
# Inputs:     NACL::C::VserverCifsBranchcache->delete arguments.
# Outputs:    Returns whether the delete was successful.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    try {
        NACL::C::VserverCifsBranchcache->delete(%args);
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        # ignore errors when the entry does not exist
    } catch NACL::APISet::Exceptions::CommandFailedException with {
        my $e = shift;
        # catch the exception if the cifs server doesn't exist
        if ($e->text() !~ /Failed to get CIFS server\. Reason: entry doesn't exist\./i) {
            $Log->exit();
            $e->throw();
        }
    };

    $Log->exit();
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  NOOP for CIFS
# Details:
# Inputs:     None
# Outputs:    None
# Exceptions: None
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
