# $Id: //depot/prod/test/main/lib/EnGenX/Controller/CIFS/Local_Group.pm $

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

EnGenX::Controller::CIFS::Local_Group

=head1 DESCRIPTION

EnGenX::Controller::CIFS::Local_Group is used to Config, Setup, Cleanup and
Deconfig CIFS Local Groups on a vserver.

=cut

package EnGenX::Controller::CIFS::Local_Group;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL libs
use NACL::C::VserverCifsUsersAndGroupsLocalGroup;
use NACL::C::VserverCifsUsersAndGroupsPrivilege;

# Exceptions
use NATE::BaseException qw(:try);

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args
            $Log);

$Log = log_global();

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    cifs_service_username =>'',
    cifs_service_password =>''
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    vserver      => '',
    'group-name' => ''
);

=head1 METHODS

=head2 create

(Static method) Request to create a cifs local-group with the following options

  EnGenX::Controller::Local-Group->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            'group-name'                    => $group_name,
            description                     => $description,
            privileges                      => $privileges,
            'member-names'                  => $member_names
            );
=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver of the cifs service for the local-group.

=item C<< 'group-name' => $group_name >>

(Required)
The group name of the local-group.

=item C<< description => $description >>

(Required)
The description of the local-group.

=item C<< privileges => $privileges >>

(Required)
The privileges of the local-group.

=item C<< 'member-names' => $member_names >>

(Required)
The names of the local-users of the local-group.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Local_Group based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs_local_group";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit();

    return $parent_answer;
}

=head2 delete

(Static method) Request to delete a cifs local-group with the following options

  EnGenX::Controller::Local-Group->delete(
            command_interface               => $node,
            vserver                         => $vserver_name,
            'group-name'                    => $group_name
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver of the cifs service for the local-group.

=item C<< 'group-name' => $group_name >>

(Required)
The group-name of the local-group.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the delete of the Local_Group based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs_local_group";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit();

    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the cifs local-group.
# Details:    Makes the NACL::C::VserverCIFSUsersAndGroupsLocalGroup->create().
#
# Inputs:     NACL::C::VserverCIFSUsersAndGroupsLocalGroup->create arguments.
# Outputs:    Returns whether the create was successful.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $reused = TRUE;
    my $if_exists
        = exists($args{if_exists})
        ? delete($args{if_exists}) : 'reuse';
    my $retry
        = exists $args{retry}
        ? delete $args{retry} : 3;
    $args{'method-timeout'}
        = 3600 unless exists($args{'method-timeout'});

    # Arguments for local-group create
    my %local_group_create_args = ();
    $local_group_create_args{'group-name'}
        = exists($args{'group-name'})
        ? delete($args{'group-name'}) : undef;
    $local_group_create_args{description}
        = (exists($args{description})
              and $args{description} =~ /^\s+?\".*?\"\s+$/)
        ? delete($args{description}) : "\"" . delete($args{description}) . "\"";

    # Arguments for add-privilege
    my %add_privilege_args = ();
    $add_privilege_args{'user-or-group-name'}
        = $local_group_create_args{'group-name'};
    my $privileges
        = exists($args{privileges})
        ? delete($args{privileges}) : undef;
    my @privileges;
    if ($privileges) {
        @privileges = split(",", $privileges);
        # privileges must be an ARRAYREF
        $add_privilege_args{privileges} = \@privileges;
    }
    my $reset_privileges
        = exists($args{'reset-privileges'})
        ? delete($args{'reset-privileges'}) : 'false';

    # Arguments for local-group add-members
    my %local_group_add_members_args = ();
    $local_group_add_members_args{'group-name'}
        = $local_group_create_args{'group-name'};
    my $member_names
        = exists($args{'member-names'})
        ? delete($args{'member-names'}) : undef;
    my @member_names;
    if ($member_names) {
        @member_names = split(",", $member_names);
        # member-names must be an ARRAYREF
        $local_group_add_members_args{'member-names'} = \@member_names;
    }
    my $reset_members
        = exists($args{'reset-members'})
        ? delete($args{'reset-members'}) : 'false';

    # create the cifs local group
    try {
        NACL::C::VserverCifsUsersAndGroupsLocalGroup->create(%args,
                                                             %local_group_create_args);
        $Log->comment("Created new CIFS Local-Group: "
                      . $local_group_create_args{'group-name'});
    } catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry > 0) {
            $pkg->_nacl_create_call(%args,
                                    %local_group_create_args,
                                    retry => --$retry,
                                    if_exists => $if_exists
                                );
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    } otherwise {
        my $e = shift;
        if (    $e->text() =~ /The specified name already exists./i
            and $if_exists =~ /reuse/i) {
            # lets just reuse the local-user already there,
            # blindly modify the service that exists
            $reused = TRUE;

            NACL::C::VserverCifsUsersAndGroupsLocalGroup->modify(%args,
                                                                 %local_group_create_args);
            $Log->comment("Modified existing CIFS Local-Group: "
                          . $local_group_create_args{'group-name'});
        } else {
            # otherwise rethrow the exception
            $Log->exit();
            $e->throw();
        }
    };

    # reset privileges for cifs local group
    if ($reset_privileges =~ /^true$/i) {
        NACL::C::VserverCifsUsersAndGroupsPrivilege->reset_privilege(%args,
                                                                     'user-or-group-name' => $add_privilege_args{'user-or-group-name'});
        $Log->comment("Reset privileges for CIFS Local-Group: "
                      . $add_privilege_args{'user-or-group-name'});
    }

    # add privileges to cifs local group
    NACL::C::VserverCifsUsersAndGroupsPrivilege->add_privilege(%args,
                                                               %add_privilege_args);
    $Log->comment("Added new privilege to CIFS Local-Group: "
                  . $add_privilege_args{'user-or-group-name'});

    # reset members for cifs local group
    if ($reset_members =~ /^true$/i) {
            my @members_cs_objs
                = NACL::C::VserverCifsUsersAndGroupsLocalGroup->show_members(
                        command_interface => $args{command_interface},
                        filter            => {'group-name' => $local_group_add_members_args{'group-name'}}
                                                                            );
        foreach my $member_cs_obj (@members_cs_objs) {
           push (@{$local_group_add_members_args{'member-names'}},
                 $member_cs_obj->member());
        }

        NACL::C::VserverCifsUsersAndGroupsLocalGroup->remove_members(%args,
                                                                     %local_group_add_members_args);
        $Log->comment("Removed members for CIFS Local-Group: "
                      . $local_group_add_members_args{'group-name'});
    }

    # add members (local users) to the cifs local group
    try {
        NACL::C::VserverCifsUsersAndGroupsLocalGroup->add_members(%args,
                                                                  %local_group_add_members_args);
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if ($exception->text() =~ /Reason: duplicate entry/) {
            $Log->comment("Adding members one at a time.");

            foreach my $member (@member_names) {
                $local_group_add_members_args{'member-names'} = [$member];
                try {
                    # add members (local users) to the cifs local group
                    NACL::C::VserverCifsUsersAndGroupsLocalGroup->add_members(%args,
                                                                              %local_group_add_members_args);
                } catch NACL::APISet::Exceptions::ResponseException with {
                    my $nested_exception = shift;
                    if (   $if_exists eq 'reuse'
                        && $nested_exception->text()
                        =~ /Reason: duplicate entry/) {
                        $reused = TRUE;
                        $Log->comment("Duplicate entry for member:" . $member);
                    } else {
                        $Log->exit();
                        # Rethrowing caught exception
                        $nested_exception->throw();
                    }
                };
            }
        } else {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    };

    $Log->exit();
    return (wantarray()) ? ('success' => TRUE, 'duplicated' => $reused) : TRUE;
}

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deleting the cifs local-group.
# Details:    Makes the NACL::C::VserverCifsUsersAndGroupsLocalGroup->delete().
#
# Inputs:     NACL::C::VserverCifsUsersAndGroupsLocalGroup->delete arguments.
# Outputs:    Returns whether the delete was successful.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # delete the cifs service
    try {
        NACL::C::VserverCifsUsersAndGroupsLocalGroup->delete(%args);
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        # no cifs local-group to delete
        $Log->comment("CIFS service $args{name} does not have a Local-Group to delete");
    };

    $Log->exit();
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  NOOP for CIFS
# Details:
# Inputs:     None
# Outputs:    None
# Exceptions: None
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
