# $Id: //depot/prod/test/main/lib/EnGenX/Controller/CIFS/Local_User.pm $

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

EnGenX::Controller::CIFS::Local_User

=head1 DESCRIPTION

EnGenX::Controller::CIFS::Local_User is used to Config, Setup, Cleanup and
Deconfig CIFS Local Users on a vserver.

=cut

package EnGenX::Controller::CIFS::Local_User;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL libs
use NACL::C::VserverCifsUsersAndGroupsLocalUser;
use NACL::C::VserverCifsUsersAndGroupsPrivilege;

# Exceptions
use NATE::BaseException qw(:try);

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args
            $Log);

$Log = log_global();

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    cifs_service_username =>'',
    cifs_service_password =>''
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    vserver     => '',
    'user-name' => ''
);

=head1 METHODS

=head2 create

(Static method) Request to create a cifs local-user with the following options

  EnGenX::Controller::Local-User->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            'user-name'                     => $user_name,
            'full-name'                     => $full_name,
            description                     => $description,
            password                        => $password,
            privileges                      => $privileges
            );
=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver of the cifs service for the local-user.

=item C<< 'user-name' => $user_name>>

(Required)
The user-name of the local-user.

=item C<< 'full-name' => $full_name>>

(Required)
The full-name of the local-user.

=item C<< 'description' => $description>>

(Required)
The description of the local-user.

=item C<< password => $password>>

(Required)
The password of the local-user.

=item C<< privileges => $privileges>>

(Required)
The privileges of the local-user.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Local_User based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs_local_user";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit();

    return $parent_answer;
}

=head2 delete

(Static method) Request to delete a cifs local-user with the following options

  EnGenX::Controller::Local-User->delete(
            command_interface               => $node,
            vserver                         => $vserver_name,
            'user-name'                     => $user_name
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver the cifs service is on.

=item C<< 'user-name' => $user_name >>

(Required)
The user-name of the local-user.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the delete of the Local_User based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs_local_user";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit();

    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the cifs local-user.
# Details:    Makes the NACL::C::VserverCIFSUsersAndGroupsLocalUser->create().
#
# Inputs:     NACL::C::VserverCIFSUsersAndGroupsLocalUser->create arguments.
# Outputs:    Returns whether the create was successful.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $reused = TRUE;
    my $if_exists
        = exists($args{if_exists})
        ? delete($args{if_exists}) : 'reuse';
    my $retry
        = exists $args{retry}
        ? delete $args{retry} : 3;
    $args{'method-timeout'}
        = 3600 unless exists($args{'method-timeout'});

    # Arguments for local-user create
    my %local_user_create_args = ();
    $local_user_create_args{'user-name'}
        = exists($args{'user-name'})
        ? delete($args{'user-name'}) : undef;
    $local_user_create_args{password}
        = exists($args{password})
        ? delete($args{password}) : undef;
    $local_user_create_args{'full-name'}
        = exists($args{'full-name'})
        ? delete($args{'full-name'}) : undef;
    $local_user_create_args{description}
        = (exists($args{description})
              and $args{description} =~ /^\s+?\".*?\"\s+$/)
        ? delete($args{description}) : "\"" . delete($args{description}) . "\"";
    $local_user_create_args{'is-account-disabled'}
        = exists($args{'is-account-disabled'})
        ? delete($args{'is-account-disabled'}) : undef;

    # Arguments for add-privilege
    my %add_privilege_args = ();
    $add_privilege_args{'user-or-group-name'}
        = $local_user_create_args{'user-name'};
    my $privileges
        = exists($args{privileges})
        ? delete($args{privileges}) : undef;
    my @privileges;
    if ($privileges) {
        @privileges = split(",", $privileges);
        # privileges must be an ARRAYREF
        $add_privilege_args{privileges} = \@privileges;
    }
    my $reset_privileges
        = exists($args{'reset-privileges'})
        ? delete($args{'reset-privileges'}) : 'false';

    # create the cifs local user
    try {
        NACL::C::VserverCifsUsersAndGroupsLocalUser->create(%args,
                                                            %local_user_create_args);
        $Log->comment("Created new CIFS Local-User: "
                      . $local_user_create_args{'user-name'});
    } catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry > 0) {
            $pkg->_nacl_create_call(%args,
                                    %local_user_create_args,
                                    retry => --$retry,
                                    if_exists => $if_exists);
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    } otherwise {
        my $e = shift;
        if (    $e->text() =~ /The specified name already exists./i
            and $if_exists =~ /reuse/i) {
            # lets just reuse the local-user already there,
            # blindly modify the service that exists
            $reused = TRUE;

            # pull in the arguments used by the modify command
            if (exists($local_user_create_args{password})) {
                delete($local_user_create_args{password});
            }

            # only execute the modify command if there are things to modify
            NACL::C::VserverCifsUsersAndGroupsLocalUser->modify(%args,
                                                                %local_user_create_args);
            $Log->comment("Modified existing CIFS Local-User: "
                          . $local_user_create_args{'user-name'});
        } else {
            # otherwise rethrow the exception
            $Log->exit();
            $e->throw();
        }
    };

    # reset privileges for cifs local user
    if ($reset_privileges =~ /^true$/i) {
        NACL::C::VserverCifsUsersAndGroupsPrivilege->reset_privilege(%args,
                                                                     'user-or-group-name'
                                                                        => $add_privilege_args{'user-or-group-name'});
        $Log->comment("Reset privileges for CIFS Local-User: "
                      . $add_privilege_args{'user-or-group-name'});
    }

    # add privileges to cifs local user
    NACL::C::VserverCifsUsersAndGroupsPrivilege->add_privilege(%args,
                                                               %add_privilege_args);
    $Log->comment("Added new privilege to CIFS Local-User: "
                  . $add_privilege_args{'user-or-group-name'});

    return (wantarray()) ? ('success' => TRUE, 'duplicated' => $reused) : TRUE;
}

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deleting the cifs local-user.
# Details:    Makes the NACL::C::VserverCifsUsersAndGroupsLocalUser->delete().
#
# Inputs:     NACL::C::VserverCifsUsersAndGroupsLocalUser->delete arguments.
# Outputs:    Returns whether the delete was successful.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->comment("CIFS Local-User: " . $args{'user-name'});
    # delete the cifs service
    try {
        NACL::C::VserverCifsUsersAndGroupsLocalUser->delete(%args);
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        # no cifs local-user to delete
        $Log->comment("CIFS service $args{name} does not have a Local-User to delete");
    };

    $Log->exit();
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  NOOP for CIFS
# Details:
# Inputs:     None
# Outputs:    None
# Exceptions: None
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
