# $Id: //depot/prod/test/main/lib/EnGenX/Controller/CIFS/Symlink.pm $

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status public
## @pod here

=head1 NAME

EnGenX::Controller::CIFS::Symlink

=head1 DESCRIPTION

EnGenX::Controller::CIFS::Symlink is used to Config, Setup, Cleanup and
Deconfig CIFS Symlink on a vserver.

=cut

package EnGenX::Controller::CIFS::Symlink;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL libs
use NACL::C::VserverCifsSymlink;

# Exceptions
use NATE::BaseException qw(:try);

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args
            $Log);

$Log = log_global();

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    cifs_service_username =>'',
    cifs_service_password =>''
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    vserver     => '',
    'unix-path' => '',
);

=head1 METHODS

=head2 create

(Static method) Request to create a cifs symlink with the following options

  EnGenX::Controller::CIFS::Symlink->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            'unix-path'                     => $unix_path,
            'share-name'                    => $share_name,
            'cifs-path'                     => $cifs_path,
            'cifs-server'                   => $cifs_server,
            locality                        => $locality,
            'home-directory'                => $home_directory
            );
=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)

=item C<< 'unix-path' => $unix_path >>

(Required)
Specifies the UNIX path prefix to be matched for the mapping.

=item C<< 'share-name' => $share_name >>

(Required)
Specifies the CIFS share name on the destination CIFS server to which the symbolic link is mapped.

=item C<< 'cifs-path' => $cifs_path >>

(Required)
Specifies the CIFS path on the destination to which the symbolic link maps.

=item C<< 'cifs-server' => $cifs_server >>

(Required)
Specifies the destination CIFS server name for the mapping to which the symbolic link is mapped.

=item C<< locality => $locality >>

(Required)
Specifies whether the CIFS symbolic link is a local link or wide link.

=item C<< 'home-directory' => $home_directory >>

(Required)
Specifics the CIFS home directory to which the symbolic link is mapped.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Symlink based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs_symlink";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit();

    return $parent_answer;
}

=head2 delete

(Static method) Request to delete a cifs symlink with the following options

  EnGenX::Controller::CIFS::Symlink->delete(
            command_interface               => $node,
            vserver                         => $vserver_name,
            'unix-path'                     => $unix_path
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)

=item C<< 'unix-path' => $unix-path >>

(Required)
This is the unix-path that will be used during NACL calls.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the delete of the Symlink based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs_symlink";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit();

    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the cifs symlink.
# Details:    Makes the NACL::C::VserverCifsSymlink->create().
#
# Inputs:     NACL::C::VserverCifsSymlink->create arguments.
# Outputs:    Returns whether the create was successful.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $reused = TRUE;
    my $if_exists
        = exists($args{if_exists})
        ? delete($args{if_exists}) : 'reuse';
    my $retry
        = exists $args{retry}
        ? delete $args{retry} : 3;
    $args{'method-timeout'}
        = 3600 unless exists($args{'method-timeout'});

    # check symlink arguments
    my $unix_path = $args{'unix-path'};
    $unix_path .= "/" if $unix_path !~ /\/$/;
    $args{'unix-path'} = $unix_path;

    my $cifs_path = $args{'cifs-path'};
    $cifs_path .= "/" if $cifs_path !~ /\/$/;
    $args{'cifs-path'} = $cifs_path;

    my $cifs_server = (exists($args{'cifs-server'}) and ($args{'cifs-server'} eq ""))
                    ? uc($args{'vserver'}) : $args{'cifs-server'};
    $args{'cifs-server'} = $cifs_server;

    # create the cifs symlink
    try {
        NACL::C::VserverCifsSymlink->create(%args);
        $Log->comment("Created new CIFS Symlink:");
    } catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry > 0) {
            $pkg->_nacl_create_call(%args,
                                    retry => --$retry,
                                    if_exists => $if_exists);
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    } otherwise {
        my $e = shift;
        if (    $e->text() =~ /duplicate entry/i
            and $if_exists =~ /reuse/i) {
            # lets just reuse the symlink already there,
            # blindly modify the service that exists
            $reused = TRUE;

            # only execute the modify command if there are things to modify
            NACL::C::VserverCifsSymlink->modify(%args);
        } else {
            # otherwise rethrow the exception
            $Log->exit();
            $e->throw();
        }
    };

    return (wantarray()) ? ('success' => TRUE, 'duplicated' => $reused) : TRUE;
}

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deleting the cifs symlink.
# Details:    Makes the NACL::C::VserverCifsSymlink->delete().
#
# Inputs:     NACL::C::VserverCifsSymlink->delete arguments.
# Outputs:    Returns whether the delete was successful.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # delete the cifs symlink
    try {
        NACL::C::VserverCifsSymlink->delete(%args);
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        # no cifs symlink to delete
        $Log->comment("CIFS service $args{name} does not have a symlink to delete");
    };

    $Log->exit();
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  NOOP for CIFS
# Details:
# Inputs:     None
# Outputs:    None
# Exceptions: None
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
