# $Id: //depot/prod/test/main/lib/EnGenX/Controller/FailoverGroup.pm $

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

EnGenX::Controller::FailoverGroup

=head1 DESCRIPTION

EnGenX::Controller::FailoverGroup is used to Config, Setup, Cleanup and
Deconfig FailoverGroup services on a vserver.

=cut

package EnGenX::Controller::FailoverGroup;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use EnGenX::SubProcess;
use EnGenX::EnGenXUtil;
use NACL::C::NetworkInterfaceFailoverGroups;
use NACL::C::Job;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::C::Exceptions::NetworkInterfaceFailoverGroups::AlreadyExists;

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name   => 'failover-group',
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    parallel      => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    apiset_must       => '',
    apiset_should     => '',
    'failover-group'    => '',
    node              => '',
    port              => ''
);

=head1 METHODS

=head2 create

(Static method) his method is used to create a new failover group
or add an entry to an existing group.

  NACL::C::NetworkInterfaceFailoverGroups->create(
            command_interface   => $node,
            vserver             => $vserver_name,
            domains             => "gdl.englab.netapp.com,rtp.netapp.com",
            'name-servers'      => ["10.60.248.10","10.60.132.40"]
            );

=over

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< "failover-group" => $failover_group >>

(Required)

Applicable for CMode CLI.

=item C<< "port" => $port >>

(Required)

Applicable for CMode CLI.

=item C<< "node" => $node >>

(Required)

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the FailoverGroup service based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {

    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "FailoverGroup";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
}

=head2 delete

(Static method) This method is used to remove a port from a
failover group.

  NACL::C::NetworkInterfaceFailoverGroups->delete(
            command_interface               => $node,
            vserver                         => $vserver_name);

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< "failover-group" => $failover_group >>

(Required)

Applicable for CMode CLI.

=item C<< "port" => $port >>

(Required)

Applicable for CMode CLI.

=item C<< "node" => $node >>

(Required)

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the FailoverGroup service based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "FailoverGroup";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit();

    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the FailoverGroup service.
# Details:    Makes the NACL::C::NetworkInterfaceFailoverGroups->create().
# Inputs:     NACL::C::NetworkInterfaceFailoverGroups->create arguments
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the FailoverGroup service was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $reused = FALSE;
    my $if_exists = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 3;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    # create the FailoverGroup service
    try {
        NACL::C::NetworkInterfaceFailoverGroups->create(%args);
    } catch NACL::C::Exceptions::NetworkInterfaceFailoverGroups::AlreadyExists with {
        my $e = shift;
        # check if requested to reuse the FailoverGroup service
        if ($if_exists =~ /reuse/i) {
            $reused = TRUE;
            $Log->comment("Reusing FailoverGroup: " . $args{'failover-group'} . "\n" .
                          "node: " . $args{'node'} . "\n" .
                          "port: " . $args{'port'});
        } else {
            # rethrow the exception if the user did not request to reuse the service
            $Log->exit();
            $e->throw();
        }
    };

    $Log->exit();
    return (wantarray()) ? ('success' => TRUE, 'duplicated' => $reused) : TRUE;
}

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for creation of the FailoverGroup service.
# Details:    Makes the NACL::C::NetworkInterfaceFailoverGroups->delete().
# Inputs:     NACL::C::NetworkInterfaceFailoverGroups->delete arguments
# Outputs:    Returns whether the delete was successful or a hash with success
#             and whether the FailoverGroup service was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my ($pkg, %args) = @_;
    try {
        NACL::C::NetworkInterfaceFailoverGroups->delete(%args);
        $Log->comment("FailoverGroup: " . $args{'failover-group'} . "\n" .
                      "node: " . $args{'node'} . "\n" .
                      "port: " . $args{'port'} . "\n" .
                      "was deleted");
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if ($exception->text() =~ /entry doesn't exist/)    # CMode
        {
            $Log->comment(
                "FailoverGroup: " . $args{'failover-group'} . "\n" .
                "node: " . $args{'node'} . "\n" .
                "port: " . $args{'port'} . "\n" .
                "was already deleted");
        } else {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    };

    $Log->exit();
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Vserver->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Vserver->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
