# $Id: //depot/prod/test/main/lib/EnGenX/Controller/FPolicyEnable.pm#4 $

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary FPolicy Controller Layer Module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

EnGenX::Controller::FpolicyEnable

=head1 DESCRIPTION

EnGenX::Controller::FpolicyEnable is a derived class of
EnGenX::Controller::Task. It is used to Create and
Delete ONTAP Vserver Fpolicies. All methods are static.

=cut

package EnGenX::Controller::Fpolicy::FpolicyEnable;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use EnGenX::SubProcess;
use EnGenX::EnGenXUtil;
use NACL::STask::VserverFpolicy;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::ConnectionFailedException;
use NACL::APISet::Exceptions::ResponseException;
use NACL::C::Exceptions::VserverFpolicy::AlreadyEnabled;
use NACL::C::Exceptions::VserverFpolicy::AlreadyDisabled;

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists',
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    parallel => '',
    fpolicy  => ''
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    vserver           => '',
    'policy-name'     => '',
);

=head1 PUBLIC METHODS

=head2 create

(Static method) Request creation of one or multiple vserver fpolicies
serially or in parallel and return whether the vserver fpolicies
were created successfully. This method passes all user arguments through to
EnGenX::Controller::Task and adds specific arguments used by the EnGenX
framework.

  EnGenX::Controller::Fpolicy->create(command_interface => $Node,
                                           vserver                       => $Vserver_name,
                                           name                          => $Policy_name,
                                           events                        => $eventsList
                                           engine                        => $engineName,
                                           'is-mandatory'                => $boolean,
                                           'allow-privileged-access'     => $accessSetting,
                                           'privileged-user-name'        => $username,
                                           'is-passthrough-read-enabled' => $boolean,
                                     );

=over

=item Options

Either the resource_list is required or the NACL Vserver
Quota Policy Creation Arguments are required but both
should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the vserver fpolicies in the resource list
will be created based on the information that is supplied in the resource
list. If the resource list is passed in the NACL Vserver Quota Policy
Creation Arguments should NOT be passed in.

=item C<< NACL Vserver Quota Policy Creation Arguments >>

=item C<< vserver => $Vserver_name >>

(Required)
vserver the name of the vserver that the policy relates to.

=item C<< name => $Policy_name >>

(Required)
name the name of the  fpolicy

(Optional)
These are arguments that can be passed into NACL::C::VolumeFpolicy->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< events => $eventsList >>

(Optional)
List of events for the fpolicy

=item C<< engine => $engineName >>

(Optional)
Enginexi name for the fpolicy

=item C<< is-mandatory => $boolean >>

(Optional)
Is mandatroy screening required for the fpolicy

=item C<< allow-privileged-access => $accessSetting >>

(Optional)
Allow privileged access for the fpolicy

=item C<< privileged-user-name => $username >>

(Optional)
Privileged username for the fpolicy

=item C<< is-passthrough-read-enabled => $boolean >>

(Optional)
Is passthrough read enabled for the fpolicy

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the F Policy based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "fpolicy_enable";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple vserver fpolicies
serially or in parallel and return whether the vserver fpolicies
were deleted successfully. This method passes all user arguments through
to EnGenX::Controller::Task and adds specific arguments used by the EnGenX
framework.

  EnGenX::Controller::Quota_Policy->delete(
            command_interface => $Node,
            vserver           => $Vserver_name,
            name              => $Policy_name);

=over

=item Options

Either the resource_list is required or the NACL Vserver
Quota Policy Deletion Arguments are required but both
should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the vserver fpolicies in the resource list
will be deleted based on the information that is supplied in the resource
list. If the resource list is passed in the NACL Vserver Quota Policy
Deletion Arguments should NOT be passed in.

=item C<< NACL Vserver Quota Policy Deletion Arguments >>

(Required)
vserver the name of the vserver that the name mapping relates to.

(Required)
name the name of the  fpolicy.

(Optional)
These are arguments that can be passed into NACL::C::VolumeFpolicy->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Vserver Quota Policy based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "fpolicy_enable";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete

=head1 PRIVATE METHODS

=head2 _nacl_create_call

=cut

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a  fpolicy.
# Details:    Makes the NACL::STask::VserverFpolicy->enable() call. Also works
#             around some common non-fatal errors that can be return from
#             vserver name-mapping creation and then has a retry mechanism to
#             retry vserver creation.
# Inputs:     NACL::STask::VserverFpolicy->enable() arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the vserver name-mapping was a duplicate if a hash is
#             requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $duplicated = FALSE;
    my $if_exists = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 3;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    $args{nacltask_if_enabled} = $if_exists;

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    try {
        # Enable the Fpolicy
        NACL::STask::VserverFpolicy->enable(%args);
        $Log->comment("Enabled the Fpolicy: " . $args{'policy-name'});
    } catch NACL::C::Exceptions::VserverFpolicy::AlreadyEnabled with {
        $duplicated = TRUE;
        $Log->comment("The fpolicy " . $args{'policy-name'} . " is already enabled.");
    };

    $Log->exit();
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>$duplicated) : TRUE;

} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a  fpolicy.
# Details:    Makes the NACL::STask::VserverFpolicy->purge() calls.
# Inputs:     NACL::STask::VserverFpolicy->purge() arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the delete was successful
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my $pkg   = shift;
    my %args = @_;

    # Locals
    my $retry = exists $args{retry} ? delete $args{retry} : 3;
    my $if_exists = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    try {
        # Delete the FPolicy
        NACL::STask::VserverFpolicy->purge(%args);
        $Log->comment("The Fpolicy " . $args{'policy-name'} . " was disabled.");
    } catch NACL::C::Exceptions::VserverFpolicy::AlreadyDisabled with {
        $Log->comment("The Fpolicy " . $args{'policy-name'} . " has already been disabled.");
    };

    $Log->exit();
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Vserver->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Vserver->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1
