# $Id: //depot/prod/test/main/lib/EnGenX/Controller/Fsecurity/FileDirectoryNTFS.pm $

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

EnGenX::Controller::Fsecurity::FileDirectoryNTFS

=head1 DESCRIPTION

EnGenX::Controller::Fsecurity::FileDirectoryNTFS is used to Config, Setup, Cleanup and
Deconfig File Directory NTFS descriptors on a vserver.

=cut

package EnGenX::Controller::Fsecurity::FileDirectoryNTFS;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL libs
use NACL::STask::VserverSecurityFileDirectoryNtfs;

# Exceptions
use NATE::BaseException qw(:try);
use NACL::C::Exceptions::VserverSecurityFileDirectoryNtfs::AlreadyExists;
use NACL::APISet::Exceptions::ConnectionFailedException;

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args
            $Log);

$Log = log_global();

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists',
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    parallel      =>'',
    file_directory_policy =>''
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    'ntfs-sd' =>'',
    'vserver' =>''
);

=head1 METHODS

=head2 create

(Static method) Request to create a file directory policy with the following options

  EnGenX::Controller::FileDirectoryNTFS->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            owner                           => $owner,
            ntfs-sd                         => $ntfs_sd
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver to create the file directory policy service for.

=item C<< ntfs-sd => $ntfs-sd >>

(Required)
This is the security descriptor.

=item C<< owner => $owner >>

(Optional)
This is the owner of the security descriptor.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the FileDirectoryNTFS Targets based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "file_directory_ntfs";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit();

    return $parent_answer;
}

=head2 delete

(Static method) Request to delete a file directory policy service with the following options

  EnGenX::Controller::FileDirectoryNTFS->delete(
            command_interface               => $node,
            vserver                         => $vserver_name,
            ntfs-sd                         => $ntfs_sd,
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver to create the file directory policy.

=item C<< ntfs-sd => $ntfs_sd >>

(Required)
This is the security descriptor.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the delete of the FileDirectoryNTFS Targets based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "file_directory_ntfs";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit();

    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the file directory policy service.
# Details:    Makes the NACL::STask::VserverFileDirectoryNTFS->create().
#
# Inputs:     NACL::STask::VserverFileDirectoryNTFS->create arguments
# Outputs:    Returns whether the create was successful
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $duplicated = FALSE;
    my $if_exists = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 3;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # create the File Directory Policy service
    try {
        NACL::STask::VserverSecurityFileDirectoryNtfs->create(%args);
    } catch NACL::C::Exceptions::VserverSecurityFileDirectoryNtfs::AlreadyExists with {
        $duplicated = TRUE;
        $Log->comment("File-Directory ntfs descriptor " . $args{'ntfs-sd'} . " already exists. It will be reused.");
    };

    $Log->exit();
    return (wantarray()) ? ('success' => TRUE, 'duplicated' => $duplicated) : TRUE;
}

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deleting the file directory policy
# Details:    Makes the NACL::STask::VserverSecurityFileDirectoryNTFS->delete().
#
# Inputs:     NACL::STask::VserverSecurityFileDirectoryNTFS->delete arguments
# Outputs:    Returns whether the delete was successful
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # delete the file directory policy service
    try {
        NACL::STask::VserverSecurityFileDirectoryNtfs->purge(%args);
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        # no file directory policy service to delete
        $Log->comment("No File-Directory descriptor $args{'ntfs-sd'} to delete.");
    };

    $Log->exit();
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  NOOP for FileDirectoryNTFS
# Details:
# Inputs:     None
# Outputs:    None
# Exceptions: None
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
