# $Id: //depot/prod/test/main/lib/EnGenX/Controller/Fsecurity/PolicyTask.pm.pm $

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

EnGenX::Controller::Fsecurity::PolicyTask.pm

=head1 DESCRIPTION

EnGenX::Controller::Fsecurity::PolicyTask.pm is used to Config, Setup, Cleanup and
Deconfig Policy Tasks  on a vserver.

=cut

package EnGenX::Controller::Fsecurity::PolicyTask;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL libs
use NACL::STask::VserverSecurityFileDirectoryPolicyTask;

# Exceptions
use NATE::BaseException qw(:try);
use NACL::C::Exceptions::VserverSecurityFileDirectoryPolicyTask::AlreadyExists;
use NACL::APISet::Exceptions::ConnectionFailedException;

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args
            $Log);

$Log = log_global();

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists',
    file_directory_policy => 'policy-name',
    'index-num' => 'index'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    parallel      => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
   'policy-name' =>'',
   path          =>'',
   vserver       =>''
);

=head1 METHODS

=head2 create

(Static method) Request to create a policy task with the following options

  EnGenX::Controller::PolicyTask.pm->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            policy-name                     => $policy_name,
            index_num                       => $index_num,
            path                            => $path,
            ntfs-mode                       => $ntfs_mode,
            ntfs_sd                         => $ntfs_sd,
            security-type                   => $security_type
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver to create the policy task service for.

=item C<< policy-name => $policy_name >>

(Required)
This is the policy name to use for this  Policy Task.

=item C<< index-num => $index_num >>

(Required)
This is the target position for adding this  Policy Task.

=item C<< path => $path >>

(Required)
This is the path for this  Policy Task.

=item C<< ntfs_mode => $ntfs_mode >>

(Optional)
This is the propogation mode for ntfs to use for this  Policy Task.

=item C<< ntfs-sd => $ntfs_sd >>

(Optional)
This is the ntfs security descriptor name to use for this  Policy Task.

=item C<< security-type => $security_type >>

(Optional)
This is the policy name to use for this  Policy Task.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the PolicyTask.pm Targets based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "policy_task";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit();

    return $parent_answer;
}

=head2 delete

(Static method) Request to delete a policy task service with the following options

  EnGenX::Controller::PolicyTask.pm->delete(
            command_interface               => $node,
            vserver                         => $vserver_name,
            policy-name                     => $policy_name,
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver to create the policy task.

=item C<< policy-name => $policy_name >>

(Required)
This is the policy-name for the policy task.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the delete of the PolicyTask.pm Targets based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "policy_task";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit();

    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the policy task service.
# Details:    Makes the NACL::STask::VserverSecurityFileDirectoryPolicyTask.pm->create().
#
# Inputs:     NACL::STask::VserverSecurityFileDirectoryPolicyTask.pm->create arguments
# Outputs:    Returns whether the create was successful
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # Locals
    my $reused = TRUE;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 3;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # create the  Policy Task service
    try {
        NACL::STask::VserverSecurityFileDirectoryPolicyTask->add(%args);
    } catch NACL::C::Exceptions::VserverSecurityFileDirectoryPolicyTask::AlreadyExists with {
      $reused = TRUE;

      $Log->comment("Policy task already exists for File-Directory policy " . $args{'policy-name'} . ". It will be reused.");
    };

    $Log->exit();
    return (wantarray()) ? ('success' => TRUE, 'duplicated' => $reused) : TRUE;
}

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deleting the policy task
# Details:    Makes the NACL::STask::VserverSecurityFileDirectoryPolicyTask.pm->delete().
#
# Inputs:     NACL::STask::VserverSecurityFileDirectoryPolicyTask.pm->delete arguments
# Outputs:    Returns whether the delete was successful
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # delete the policy task service
    try {
        NACL::STask::VserverSecurityFileDirectoryPolicyTask->purge(%args);
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        # no policy task to delete
        $Log->comment("Vserver $args{vserver} does not have a  Policy Task $args{'policy-name'} to delete.");
    };

    $Log->exit();
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  NOOP for PolicyTask.pm
# Details:
# Inputs:     None
# Outputs:    None
# Exceptions: None
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my ($pkg, %args) = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
