# $Id: //depot/prod/test/main/lib/EnGenX/Controller/Ifgrp.pm#2 $

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Ifgrp Controller layer module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

EnGenX::Controller::Ifgrp

=head1 DESCRIPTION

EnGenX::Controller::Ifgrp is a derived class of
EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP interface groups. All methods are static.

=cut

package EnGenX::Controller::Ifgrp;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use NATE::STAF::Peer qw($Parent);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::C::NetworkPortIfgrp;
use NACL::STask::NetworkPortIfgrp;
use NACL::C::Job;
use EnGenX::SubProcess;
use EnGenX::EnGenXUtil;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

#Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args
);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name          => 'ifgrp',
    'port-list'   => 'ports',
    _reuse        => 'nacltask_if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
        'port-list'     => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    ifgrp             => '',
    node              => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple Ifgrps serially
or in parallel and return whether the Ifgrp(s) was created successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Ifgrp->create(
            command_interface => $Node,
            ifgrp             => $Ifgrp_name,
            node              => $Node_name,
            'distr-func'      => $Distribution_algorithm
            mode              => $Mode_name
            'port-list'       => $port_list
            'method-timeout'  => 3600,
            );


  EnGenX::Controller::Ifgrp->create(
            command_interface  => $Node,
            resource_list      => @Ifgrps,
            'method-timeout'   => 3600,
            );

=over

=item Options

Either the resource_list is required or the NACL Ifgrp Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the Ifgrps in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Ifgrp Creation Arguments should
NOT be passed in.

=item C<< NACL Ifgrp Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::STask::NetworkPortIfgrp->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Ifgrp(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "ifgrp";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit();
    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple ifgrps serially
or in parallel and return whether the ifgrp(s) was deleted successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Ifgrp->delete(
            command_interface => $Node,
            ifgrp             => $Ifgrp_name,
            node              => $Node_name,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Ifgrp->delete(
            command_interface => $Node,
            resource_list      => @Ifgrps,
            'method-timeout'  => 3600,
            );

=over

=item Options

Either the resource_list is required or the NACL Ifgrp Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the Ifgrps in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Ifgrp Deletion Arguments should
NOT be passed in.

=item C<< NACL Ifgrp Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::STask::NetworkPortIfgrp->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Ifgrp(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "ifgrp";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit();
    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a Ifgrp.
# Details:    Makes the NACL::STask::NetworkPortIfgrp->create() call. Also works
#             around some common non-fatal errors that can be return from
#             Ifgrp creation and then has a retry mechanism to retry Ifgrp
#             creation.
# Inputs:     NACL::STask::NetworkPortIfgrp->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the Ifgrp was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;
    # Locals
    my $duplicated = FALSE;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    # split comma separated string ports into an array for NACL call
    my @port_array = split(',' , $args{ports});
    $args{ports} = \@port_array;

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # Execute Create
    try {
        NACL::STask::NetworkPortIfgrp->create(%args);
        $Log->comment("Using newly created Ifgrp:" . $args{ifgrp});
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 3) {
            sleep 30 * ($retry++);
            $pkg->_nacl_create_call(
                                %args,
                                retry     => $retry,
                                if_exists => $if_exists
                            );
        } else {
            my $exception = shift;
            # Rethrowing caught exception
            $Log->exit();
            $exception->throw();
        }
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if( ($exception->text() =~ /already exists/i)
            && ($if_exists eq 'reuse') )
        {
            $duplicated = TRUE;
            $Log->comment("Using an already created Ifgrp:" . $args{ifgrp});
        } else {
            # Rethrowing caught exception
            $Log->exit();
            $exception->throw();
        }
    };

    $Log->exit();
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>$duplicated) : TRUE;


} ## end sub _nacl_create_call

#==========#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a Ifgrp.
# Details:    Makes the NACL::STask::NetworkPortIfgrp->delete() calls. Also works around
#             some common non-fatal errors that can be return from Ifgrp
#             deletion and then has a retry mechanism to retry Ifgrp deletion.
# Inputs:     NACL::STask::NetworkPortIfgrp->delete arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the delete was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;
    # Locals
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    try {
        NACL::STask::NetworkPortIfgrp->delete(%args,);
        $Log->comment("Ifgrp:" . $args{ifgrp} . " was deleted");
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 3) {
            sleep 30 * ($retry++);
            $pkg->_nacl_delete_call(%args,
                                    retry => $retry,
                                );
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if ($exception->text() =~ /no entries matching your query/
            or $exception->text() =~ /entry doesn't exist/)
        {
            $Log->comment("Ifgrp:" . $args{ifgrp} . " was already deleted");
        } else {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    };

    $Log->exit();
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
