# $Id: //depot/prod/test/main/lib/EnGenX/Controller/Iscsi.pm $

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

EnGenX::Controller

=head1 DESCRIPTION

EnGenX::Controller::FCP is used to Config, Setup, Cleanup and
Deconfig Iscsi services on a vserver.

=cut

package EnGenX::Controller::Iscsi;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use EnGenX::SubProcess;
use EnGenX::EnGenXUtil;
use NACL::C::VserverIscsi;
use NACL::C::Job;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    aggr   => 'aggregate',
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    name   => '',
    parallel      => '',
    service_enabled => ''
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = ();

=head1 METHODS

=head2 create

(Static method) Request to add the iscsi service to a list
of vservers the command interface and vserver name are the
only required perameters.

  EnGenX::Controller::Iscsi->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            target-name                     => $target_name,
            target-alias                    => $target_alias,
            status-admin                    => $status_admin,
            max-error-recovery-level        => $max_error,
            retain-timeout                  => $retain_timeout,
            login-timeout                   => $login_timeout,
            max-conn-per-session            => $max_conn,
            max-ios-per-session             => $max_ios,
            tcp-window-size                 => $tcp_window_size
            );

  EnGenX::Controller::Iscsi->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            );

=over

=item Options

The NACL component only requires the user to specify the
command interface and the vserver. The ontap defaults will
be used for any option that is not stated.

=over

=item C<< target_name => $target_name >>

(Optional)
This is the target IQN that must follow the standard IQN
notation. If none is specified Ontap will use a defualt
IQN.

=item C<< target-alias => $target_alias >>

(Optional)
This argument is used to give a "friendly" name to the
target IQN so it is easy to find.

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< status-admin => $status_admin >>

(Optional)
The defualt is to have the service up and running. The user can specify
the service be down on creation.

=item C<< max-error-recovery-level => $max_error >>

(Optional)

=item C<< retain-timeout => $retain_timeout >>

(Optional)

=item C<< login-timeout => $login_timeout >>

(Optional)

=item C<< max-conn-per-session => $max_conn >>

(Optional)

=item C<< max-ios-per-session => $max_ios >>

(Optional)

=item C<< tcp-window-size => $tcp_window_size >>

(Optional)

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Iscsi Targets based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {

    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "iscsi";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

#==============================================================================
#The delete function is not going to be implemented
#due to the options being deleted upon vserver deletion.
#==============================================================================
sub delete {
    #TODO: no-op for now. May implement in the future
    return;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the fcp service.
# Details:    Makes the NACL::C::VserverIscsi->create(). Also works
#             around some common non-fatal errors that can be return from
#             volume creation and then has a retry mechanism to retry volume
#             creation.
# Inputs:     NACL::C::VserverIscsi->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the volume was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {

    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    my $reused = 0;

    # This creates a NACL command_interface if one is not passed in.
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    try {
        NACL::C::VserverIscsi->create(%args);
        $Log->comment("Enabled Iscsi on vserver " . $args{vserver});
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 3) {
            sleep 30 * ($retry++);
            $pkg->_nacl_create_call(
                %args,
                retry     => $retry,
                if_exists => $if_exists,
            );
        } else {
            my $exception = shift;
            $Log->exit();
            # Rethrowing caught exception
            $exception->throw();
        }
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if (($if_exists eq 'reuse') and ($exception->text() =~ /iSCSI service already exists/)) {
            $reused = 1;
            $Log->comment("The Iscsi service on vserver " . $args{vserver} . " already exists.");
            $Log->comment("Downing Iscsi service on vserver " . $args{vserver} . " to modify settings is needed.");
            try {
                #Stop the iscsi service so that it can be modified
                NACL::C::VserverIscsi->stop( command_interface => $args{command_interface},
                                             vserver => $args{vserver}
                                            );
                $Log->comment("The Iscsi service on vserver " . $args{vserver} . " has been stopped.");
            }
            otherwise {
                my $exception = shift;
                # If the service is already stopped don't worry about it.
                if ($exception->text() =~ /Service, adapter, or operation not started/) {
                    $Log->comment("The Iscsi service on vserver " . $args{vserver} . " was already stopped.");
                } else {
                    $Log->comment("The Iscsi service on vserver " . $args{vserver} . " could not be stopped.");
                    $Log->exit();
                    # Rethrowing caught exception
                    $exception->throw();
                }
            };

            try {
                # Modify the iscsi service if there are new settings
                NACL::C::VserverIscsi->modify(%args);
            }
            otherwise {
                my $exception = shift;
                if ($exception->text() =~ /Specified target name is same as current target name/) {
                    $Log->comment("The Iscsi service target name on vserver " . $args{vserver} .
                                   " could not be modified. It is the existing name.");
                }
                elsif ($exception->text() =~ /You need to provide some values to apply/) {
                    $Log->comment("The Iscsi service target name on vserver " . $args{vserver} .
                                   " had no new information to be added/modified.");
                } else {
                    $Log->warn("The Iscsi service on vserver " . $args{vserver} . " could not be modified.");
                    $Log->exit();
                    # Rethrowing caught exception
                    $exception->throw();
                }
            };

            if ($args{'status-admin'} eq "down") {
                $Log->comment("Leaving the Iscsi service on vserver " . $args{vserver} . " in the admin down state.");
            } elsif ($args{'status-admin'} eq "up") {
                $Log->comment("The Iscsi service on vserver " . $args{vserver} . " has been started by the modify action.");
            }

        } else {
            # Rethrowing caught exception
            $Log->exit();
            $exception->throw();
        }
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        $exception->throw();
    };

    $Log->exit();
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success' => TRUE, 'duplicated' => $reused) : TRUE;

}#sub

#==============================================================================
#The _nacl_delete_call function is not going to be implemented
#due to the options being deleted upon volume deletion.
#==============================================================================
sub _nacl_delete_call {
    #TODO: no-op for now. May implement in the future
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Vserver->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Vserver->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
