# $Id: //depot/prod/test/main/lib/EnGenX/Controller/IscsiSecurity.pm $

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

EnGenX::Controller

=head1 DESCRIPTION

EnGenX::Controller::IscsiSecurity is used to Config, Setup, Cleanup and
Deconfig iSCSI security settings on a vserver.

=cut

package EnGenX::Controller::IscsiSecurity;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use EnGenX::SubProcess;
use EnGenX::EnGenXUtil;
use NACL::C::VserverIscsiSecurity;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::NoMatchingEntriesException;

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    vserver => '',
    'initiator-name' => ''
);

=head1 METHODS

=head2 create

(Static method) Request to create a iscsi security setting for a
vserver.

  # set up bidirectional chap
  EnGenX::Controller::IscsiSecurity->create(
            'command_interface'             => $node_command_interface,
            'vserver'                       => "myvserver",
            'initiator-name'                => "iqn.1970-01.com.company:abcdef",
            'auth-type'                     => "CHAP",
            'chap_mode'                     => bidirectional
            'password'                      => $unidirectional_password,
            'user-name'                     => $unidirectional_username,
            'outbound-password'             => $bidirectional_password,
            'outbound-user-name'            => $bidirectional_username,
            );

  # set up unidirectional chap
  EnGenX::Controller::IscsiSecurity->create(
            'command_interface'             => $node_command_interface,
            'vserver'                       => "myvserver",
            'initiator-name'                => "iqn.1970-01.com.company:abcdef",
            'auth-type'                     => "CHAP",
            'chap_mode'                     => unidirectional
            'password'                      => $unidirectional_password,
            'user-name'                     => $unidirectional_username,
            );

  # this will verify that there is no iscsi security setting for the initiator
  # this is the equivalent of using the delete() api call
  EnGenX::Controller::IscsiSecurity->create(
            'command_interface'             => $node_command_interface,
            'vserver'                       => "myvserver",
            'initiator-name'                => "iqn.1970-01.com.company:abcdef",
            'chap_mode'                     => none
            );

=over

=item Options

The NACL component only requires the user to specify the
command interface and the vserver. The ontap defaults will
be used for any option that is not stated.

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vs_name >>

(Required)
The vserver to operate against

=item C<< chap_mode => unidirectional|bidirectional|none >>

(Default)
none

Specifies what type of CHAP authentication will be used. If none is specified, then
the iscsi security setting specified will be deleted if it exists.

=item C<< %args  >>

All other inputs into NACL::C::VserverIscsiSecurity::create();

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Iscsi Targets based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit();
    return $parent_answer;
} ## end sub create

#==============================================================================
#The delete function is not going to be implemented
#due to the options being deleted upon vserver deletion.
#==============================================================================
sub delete {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit();
    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:
# Details:
# Inputs:     NACL::C::VserverIscsiSecurity->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the volume was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {

    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 3;
    my $reused = FALSE;
    my $chap_mode = exists($args{chap_mode}) ? delete($args{chap_mode}) : 'none';

    # This creates a NACL command_interface if one is not passed in.
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # if the chap mode is set to none, make sure there is no iscsi security setting for this initiator
    if ($chap_mode =~ /none/i) {
        $pkg->_nacl_delete_call(%args);
    } else {
        # create the iscsi security settings with the passed in options
        try {
            NACL::C::VserverIscsiSecurity->create(%args);
        } catch NACL::APISet::Exceptions::ConnectionFailedException with {
            my $e = shift;
            # retry if we are unable to connect
            if ($retry > 0) {
                sleep 5;
                $retry--;
                $pkg->_nacl_create_call(%args,
                                        retry => $retry,
                                        if_exists => $if_exists);
            } else {
                # rethrow the exception
                $Log->exit();
                $e->throw();
            }
        } catch NACL::APISet::Exceptions::CommandFailedException with {
            my $e = shift;
            if ($e->text() =~ /already has a security policy/i and $if_exists =~ /^reuse$/i) {
                # it already has a security settings,
                # modify the setting to accomodate the new values
                $Log->comment("Modifying the existing iSCSI Security settings for initiator" . $args{'initiator-name'} .
                              " on vserver " . $args{'vserver'});

                # clear the outbound settings if doing unidirectional chap
                if ($chap_mode =~ /unidirectional/i) {
                    $args{'clear-outbound'} = "true";
                }

                NACL::C::VserverIscsiSecurity->modify(%args);
                $reused = TRUE;
            } else {
                # rethrow the exception
                $Log->exit();
                $e->throw();
            }
        };
    }

    $Log->exit();

    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success' => TRUE, 'duplicated' => $reused) : TRUE;
}

#==============================================================================
#The _nacl_delete_call function is not going to be implemented
#due to the options being deleted upon volume deletion.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my ($pkg, %args) = @_;

    # if the none of the iscsi security settings were passed in, then
    # we are being told to remove the iscsi security settings for this
    # initiator
    try {
        NACL::C::VserverIscsiSecurity->delete(%args);
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        # ignore errors if this iscsi security setting does not exist
    };

    $Log->exit();
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Vserver->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Vserver->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
