# $Id$

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Lun Controller Layer Module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

EnGenX::Controller::Lun

=head1 DESCRIPTION

EnGenX::Controller::Lun is a derived class of
EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP Luns. All methods are static.

=cut

package EnGenX::Controller::Lun;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::C::Lun;
use NACL::CS::Volume;
use NACL::STask::Lun;
use NACL::C::Job;
use EnGenX::SubProcess;
use EnGenX::EnGenXUtil;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

#Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name   => 'lun',
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    parallel => '',
    map_list => '',
    require_same_lun_id_per_igroup => '',
    'lun-id' => '',
    lun_id_range => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    apiset_must       => '',
    apiset_should     => '',
    lun               => '',
    qtree             => '',
    volume            => '',
    vserver           => '',
    path              => '',
    force             => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple luns serially
or in parallel and return whether the lun(s) was created successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Lun->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            lun                  => $Lun_name,
            vserver           => $Vserver_name,
            size              => "40M",
            ostype              => 'windows',
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Lun->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            lun                  => $Lun_name,
            vserver           => $Vserver_name,
            size              => "40M",
            ostype              => 'windows',
            'method-timeout'  => 3600,
            count             => 5
            );

  EnGenX::Controller::Lun->create(
            command_interfac  => $Node,
            resource_list      => @Luns,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Lun->create(
            command_interface => $Node,
            resource_list      => @Luns,
            );

  EnGenX::Controller::Lun->create(
            resource_list      => @Luns,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Lun->create(
            resource_list      => @Luns,
            );

=over

=item Options

Either the resource_list is required or the NACL Lun Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the luns in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Lun Creation Arguments should
NOT be passed in.

=item C<< NACL Lun Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::C::Lun->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< count => $lun_count >>

(Optional)
This arguments is used to create multiple luns from the same NACL Lun
Creation Arguments.  To create the names for the luns, the lun name
passed in is appended with the index that it was create at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Lun(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "lun";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple luns serially
or in parallel and return whether the lun(s) was deleted successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Lun->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            lun                  => $Lun_name,
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Lun->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            lun                  => $Lun_name,
            vserver           => $Vserver_name,
            count             => 5
            );

  EnGenX::Controller::Lun->delete(
            command_interface => $Node,
            resource_list      => @Luns,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Lun->delete(
            command_interface => $Node,
            resource_list      => @Luns,
            );

  EnGenX::Controller::Lun->delete(
            resource_list      => @Luns,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Lun->delete(
            resource_list      => @Luns,
            );

=over

=item Options

Either the resource_list is required or the NACL Lun Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the luns in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Lun Deletion Arguments should
NOT be passed in.

=item C<< NACL Lun Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::C::Lun->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< count => $lun_count >>

(Optional)
This arguments is used to delete multiple luns from the same NACL Lun
Deletion Arguments.  To delete the names for the luns, the lun name
passed in is appended with the index that it was deleted at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Lun(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "lun";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a lun.
# Details:    Makes the NACL::C::Lun->create() call. Also works
#             around some common non-fatal errors that can be return from
#             lun creation and then has a retry mechanism to retry lun
#             creation.
# Inputs:     NACL::C::Lun->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the lun was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;
    # Locals
    my $duplicated = FALSE;
    my $path = "";
    my $state = exists($args{'state'}) ? delete($args{'state'}) : 'online';
    my $allocation
        = exists($args{'space-allocation'}) ? delete($args{'space-allocation'}) : 'disabled';
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # Build up the lun path.
    if (!defined($args{qtree})) {
        $path = "/vol/$args{volume}/$args{lun}";
    } else {
        $path = "/vol/$args{volume}/$args{qtree}/$args{lun}";
    }

    # Check if size arguments have been defined and calculate percentage if necessary
    $pkg->_compute_size(\%args);

    # Put quotes around the lun  comment, if it exists.
    if (exists $args{comment}) {
        $args{comment} = "\"" . $args{comment} . "\"";
    }

    # Execute Create
    try {
        NACL::STask::Lun->create(%args);
        # Add in space allocation.
        # NOTE the lun modify attirbutes were not in the
        # NACL docs at the time this was writen.
        if (($allocation =~ /enabled/i)) {
            # Modification of the allocation setting or the lun state must be sent seperately (burt#783657)
            # We need to build a path to turn on allocation
            NACL::STask::Lun->modify(command_interface => $args{command_interface},
                                     vserver => $args{vserver},
                                     path => $path,
                                     'space-allocation' => $allocation);
        }
        if (($state eq 'offline')) {

            NACL::STask::Lun->modify(command_interface => $args{command_interface},
                                     vserver => $args{vserver},
                                     path => $path,
                                     'state' => $state,);

        }
        $Log->comment("Comment: Using newly Created Lun:" . $args{lun});
    }
    catch NACL::C::Exceptions::Lun::AlreadyExists with {
        my $exception = shift;
        if ($if_exists eq 'reuse') {
            $duplicated = TRUE;
            $Log->comment("Using already Created Lun:" . $args{lun});
            # Modification of the allocation setting or the lun state must be sent seperately (burt#783657)
            # If the lun is reused then we need to ensure that it is in the
            # state the user requested.
            $Log->comment("Updating already Created Luns State:" . $args{state});

            NACL::STask::Lun->modify(command_interface => $args{command_interface},
                                     vserver => $args{vserver},
                                     path => $path,
                                     'space-allocation' => $allocation);

            NACL::STask::Lun->modify(command_interface => $args{command_interface},
                                     vserver => $args{vserver},
                                     path => $path,
                                     'state' => $state,);
        } else {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 5) {
            sleep 30 * ($retry++);
            # recursively calling _nacl_create_call to retry create
            $pkg->_nacl_create_call(
                                %args,
                                retry     => $retry,
                                if_exists => $if_exists,
                                'state'   => $state,
                                'space-allocation' => $allocation
                            );
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if ($exception->text() =~ /Authorization failed/ && $retry < 5) {
            sleep 30 * ($retry++);
            # recursively calling _nacl_create_call to retry create
            $pkg->_nacl_create_call(
                                %args,
                                retry     => $retry,
                                if_exists => $if_exists,
                                state     => $state,
                                'space-allocation' => $allocation);

        } else {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    };

    $Log->exit();
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>$duplicated) : TRUE;


} ## end sub _nacl_create_call

#========================================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a Lun.
# Details:    Makes the NACL::C::Lun->delete() calls. Also works around
#             some common non-fatal errors that can be return from Lun
#             deletion and then has a retry mechanism to retry Lun deletion.
# Inputs:     NACL::C::Lun->delete arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;
    # Locals
    my $if_exists;
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }
    $args{force} = TRUE unless exists $args{force};

    try {
        NACL::C::Lun->delete(%args,);
        $Log->comment("Lun:" . $args{lun} . " was deleted");
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 3) {
            sleep 30 * ($retry++);
            # recursively calling _nacl_delete_call to retry delete
            $pkg->_nacl_delete_call(%args,
                                    retry =>$retry,
                                );
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;

        # /No such LUN exists/ is for if the lun does not exist
        # /entry doesn't exist/ is for if the volumr does not exist
        # Error: command failed: Volume pressley_vol_0219 does not exist in
        #     Vserver test_io_1. Reason: entry doesn't exist
        if ($exception->text() =~ /Authorization failed/ && $retry < 3) {
            sleep 30 * ($retry++);
            # recursively calling _nacl_delete_call to retry delete
            $pkg->_nacl_delete_call(%args,
                                    retry =>$retry,
                                );
        } elsif ($exception->text() =~ /No such LUN exists/
            or $exception->text() =~ /no entries matching your query/
            or $exception->text() =~ /No such file or directory/
            or $exception->text() =~ /entry doesn't exist|No such file or directory/)
        {
            $Log->comment("Lun:" . $args{lun} . " was already deleted");
        } else {
            # Rethrowing caught exception
            $Log->exit();
            $exception->throw();
        }
    };

    $Log->exit();
    return TRUE;

} ## end sub _nacl_delete_call
#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Lun->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Lun->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}
#==============================================================================
# Method:     _compute_size (args,size_param)
# Objective:  Compute/Process the size parameter if it exists.
# Details:    Retrieves the hash or a reference to one of the following:
# Inputs:     $args, $size_param`
# Outputs:    none
# Exceptions:
#==============================================================================
sub _compute_size {
    $Log->enter();

    my ($pkg, $args) = @_;
    my @size_params = ('size','prefix-size', 'extent-size', 'suffix-size');
    my $modulus;
    my $volume_cs_obj;
    my $percent;

    foreach my $size_param (@size_params) {
        # If size paramter is anything other than 'size' make sure calculated percent is a multiple of 512. 
        # In this case, the size is rounded down to to the nearest multiple of 512.
        if (exists $args->{$size_param} and $args->{$size_param} =~ s/%//) {
            if($volume_cs_obj eq undef) {
                 $volume_cs_obj = NACL::CS::Volume->fetch(command_interface => $args->{command_interface},
                                                          requested_fields => ['size'],
                                                          filter => { vserver => $args->{vserver},
                                                                      volume  => $args->{volume}});
             }

             $percent = $args->{$size_param};
             $args->{$size_param} = ($args->{$size_param}/100)*($volume_cs_obj->size());

             if($size_param ne 'size'){
                 # Remove Decimal
                 $args->{$size_param} =~ s/\.\d+$//;

                 # Make sure the value is a multiple of 512
                 $modulus = $args->{$size_param} % 512;
                 $args->{$size_param} -= $modulus;
             }

             $Log->comment("Setting Lun " . $size_param . " to " . $args->{$size_param} . "b which is " .
                           "$percent\% of volume " . $args->{volume} . " in vserver " .
                           $args->{vserver} . " which is size " . $volume_cs_obj->size() . "b.");
        } 
    }
}
    $Log->exit();

1;
