# $Id: //depot/prod/test/main/lib/EnGenX/Controller/QosPolicyGroup.pm $

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

EnGenX::Controller

=head1 DESCRIPTION

EnGenX::Controller::FCP is used to Config, Setup, Cleanup and
Deconfig QosPolicyGroup services on a vserver.

=cut

package EnGenX::Controller::QosPolicyGroup;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use EnGenX::SubProcess;
use EnGenX::EnGenXUtil;
use NACL::C::QosPolicyGroup;
use NACL::CS::QosPolicyGroup;
use NACL::C::Job;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name   => 'policy-group',
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    name   => '',
    parallel      => '',
);

# Arguments that can be passed into the NACL delete
# command
# Arguments that can be passed into the NACL delete
# command.
%Delete_Args = (
    command_interface       => '',
    apiset_must             => '',
    apiset_should           => '',
    'policy-group'          => '',
    vserver                 => '',
);

=head1 METHODS

=head2 create

(Static method) Request to add the qos policy group service to a list
of vservers the command interface and vserver name are the
only required perameters.

  EnGenX::Controller::QosPolicyGroup->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            target-name                     => $target_name,
            target-alias                    => $target_alias,
            status-admin                    => $status_admin,
            max-error-recovery-level        => $max_error,
            retain-timeout                  => $retain_timeout,
            login-timeout                   => $login_timeout,
            max-conn-per-session            => $max_conn,
            max-ios-per-session             => $max_ios,
            tcp-window-size                 => $tcp_window_size
            );

  EnGenX::Controller::QosPolicyGroup->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            );

=over

=item Options

The NACL component only requires the user to specify the
command interface and the vserver. The ontap defaults will
be used for any option that is not stated.

=over

=item C<< target_name => $target_name >>

(Optional)
This is the target IQN that must follow the standard IQN
notation. If none is specified Ontap will use a defualt
IQN.

=item C<< target-alias => $target_alias >>

(Optional)
This argument is used to give a "friendly" name to the
target IQN so it is easy to find.

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< status-admin => $status_admin >>

(Optional)
The defualt is to have the service up and running. The user can specify
the service be down on creation.

=item C<< max-error-recovery-level => $max_error >>

(Optional)

=item C<< retain-timeout => $retain_timeout >>

(Optional)

=item C<< login-timeout => $login_timeout >>

(Optional)

=item C<< max-conn-per-session => $max_conn >>

(Optional)

=item C<< max-ios-per-session => $max_ios >>

(Optional)

=item C<< tcp-window-size => $tcp_window_size >>

(Optional)

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the QosPolicyGroup Targets based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "qos_policy_group";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

#==============================================================================
#The delete function is not going to be implemented
#due to the options being deleted upon vserver deletion.
#==============================================================================
sub delete {
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # TODO Remove $args{'Name'} from everywhere. And ensure
    # no errors occur.
    $args{'Name'} = "qos_policy_group";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the qos policy group.
# Details:    Makes the NACL::C::VserverQosPolicyGroup->create(). Also works
#             around some common non-fatal errors that can be return from
#             qos policy creation and then has a retry mechanism to retry
#             creation.
# Inputs:     NACL::C::VserverQosPolicyGroup->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the volume was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 3;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    my $assign_to_vserver = exists($args{assign_to_vserver}) ? delete($args{assign_to_vserver}) : FALSE;

    # This creates a NACL command_interface if one is not passed in.
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    my $command_interface = delete $args{command_interface};

    # Removing unneeded parameter after building the name.
    # We should not need these on retry because the name will
    # already be defined and not _default
    delete $args{volume};
    delete $args{qtree};
    delete $args{lun};

    try {
        NACL::C::QosPolicyGroup->create(%args,
                                        command_interface => $command_interface,);
    } catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry > 0) {
            sleep 30 * ($retry--);
            $pkg->_nacl_create_call(
                %args,
                retry     => $retry,
                if_exists => $if_exists,
                assign_to_vserver => $assign_to_vserver,
                command_interface => $command_interface
            );
        } else {
            my $exception = shift;
            $Log->exit();
            # Rethrowing caught exception
            $exception->throw();
        }
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if (($if_exists eq 'reuse') and ($exception->text() =~ /Duplicate .*/i)) {
            try {
                my %fetch_args = (vserver => $args{vserver},
                                  'policy-group' => $args{'policy-group'});
                my $duplicate_qos_obj = NACL::CS::QosPolicyGroup->fetch(
                                        command_interface => $command_interface,
                                        filter => \%fetch_args,
                                        );

                # If we find it then that means that it is really an internal
                # duplicate within this code because we showed all of the
                # qos policy groups because attempting to add more so we could
                # compare the list to what the user wanted to add.

            } otherwise {
                my $exception = shift;
                $Log->exit();
                # Rethrowing caught exception
                $exception->throw("QOS Policy :\n" . Dumper(\%args)
                                . " with this name already exists but it is created"
                                . " under a different vserver");
            };
        } else {
            # Rethrowing caught exception
            $Log->exit();
            $exception->throw();
        }
    } otherwise {
        my $exception = shift;
        $Log->exit();
        $exception->throw();
    };

    if ($assign_to_vserver) {
        try {
            NACL::C::Vserver->modify(command_interface  => $command_interface,
                                     vserver            => $args{vserver},
                                     'qos-policy-group' => $args{'policy-group'});
        } otherwise {
            my $exception = shift;
            $Log->exit();
            # Rethrowing caught exception
            $exception->throw();
        };
    }
    $Log->exit();
    # Returning either a list (hash) or scalar TRUE
    return TRUE;

}#sub

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of the qos policy group.
# Details:    Makes the NACL::C::VserverQosPolicyGroup->delete(). Also works
#             around some common non-fatal errors that can be return from
#             qos policy delete.
# Inputs:     NACL::C::VserverQosPolicyGroup->delete arguments
# Outputs:    Returns whether the delete was successful or a hash with success
#             and whether the qos policy was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    my $vserver = delete $args{vserver};
    # This creates a NACL command_interface if one is not passed in.
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    try {
        NACL::C::QosPolicyGroup->delete(%args);
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if ($exception->text() =~
                /Policy group could not be removed because there are still workloads associated with it/i) {
            # If the qos policy group still has workloads associated with it
            # then comment and leave the policy group.  This could be because the
            # user associated the policy group with the containing vserver or
            # something else that the user created but did not clean up.
            $Log->comment( "Cannot cleanup qos policy group " .
                        $args{'policy-group'} . ". It is still associated" .
                        " with something in vserver: " . $vserver .
                        ". If the vserver is cleaned up then this policy" .
                        " group will also be removed with the deletion" .
                        " of the vserver. Otherwise this policy group" .
                        " will remain.");
        } elsif ($exception->text() !~ /entry doesn't exist/) {
            $Log->exit();
            # Rethrowing caught exception
            $exception->throw();
        }
    }; # end of delete try/catch

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Vserver->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Vserver->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
