# $Id: //depot/prod/test/main/lib/EnGenX/Controller/QuotaRule.pm#4 $

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Quota Policy Rule Controller Layer Module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

EnGenX::Controller::QuotaRule

=head1 DESCRIPTION

EnGenX::Controller::QuotaRule is a derived class of
EnGenX::Controller::Task. It is used to Create and
Delete ONTAP Vserver quota policy rules. All methods are static.

=cut

package EnGenX::Controller::QuotaRule;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use EnGenX::SubProcess;
use EnGenX::EnGenXUtil;
use NACL::C::VolumeQuotaPolicyRule;
use NACL::STask::VolumeQuota;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::ConnectionFailedException;
use NACL::APISet::Exceptions::ResponseException;

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists',
    quota_policy => 'policy-name',
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    parallel => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    apiset_must       => '',
    apiset_should     => '',
    vserver           => '',
    volume            => '',
    'policy-name'     => '',
    type              => '',
    target            => '',
);

=head1 PUBLIC METHODS

=head2 create

(Static method) Request creation of one or multiple quota policy rules
serially or in parallel and return whether the vserver quota policy rules
were created successfully. This method passes all user arguments through to
EnGenX::Controller::Task and adds specific arguments used by the EnGenX
framework.

  EnGenX::Controller::Quota_Rule->create(
            command_interface => $Node,
            vserver => $Vserver_name,
            'policy-name'         => $Policy_name,
            type          => $Rule_type,
            target           => $Rule_target,
            qtree       => $Rule_qtree,
            ...);

=over

=item Options

Either the resource_list is required or the NACL Vserver
quota rule Creation Arguments are required but both
should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the vserver quota policy rules in the resource list
will be created based on the information that is supplied in the resource
list. If the resource list is passed in the NACL Vserver quota rule
Creation Arguments should NOT be passed in.

=item C<< NACL Vserver quota rule Creation Arguments >>

(Required)
vserver the name of the vserver that the quota rule relates to.

(Required)
policy-name the name of the quota policy that the quota rule applies to

(Required)
type the type of quota rule

(Required)
target the target for the quota rule

(Required)
qtree the qtree for the quota rule


(Optional)
These are arguments that can be passed into NACL::C::Vserverquotarule->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Vserver quota rule based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "quota_rule";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple quota policy rules
serially or in parallel and return whether the quota policy rule
was deleted successfully. This method passes all user arguments through
to EnGenX::Controller::Task and adds specific arguments used by the EnGenX
framework.

  EnGenX::Controller::Quota_Rule->delete(
            command_interface => $Node,
            vserver           => $Vserver_name,
            volume            => $Rule_volume,
            'policy-name'     => $Rule_name,
            type              => $Rule_type,
            target            => $Rule_Target,
            ...);

=over

=item Options

Either the resource_list is required or the NACL Vserver
quota rule Deletion Arguments are required but both
should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the vserver quota policy rules in the resource list
will be deleted based on the information that is supplied in the resource
list. If the resource list is passed in the NACL Vserver quota rule
Deletion Arguments should NOT be passed in.

=item C<< NACL Vserver quota rule Deletion Arguments >>

(Required)
vserver the name of the vserver that the quota rule relates to.

(Required)
volume the name of the volume that contains the qtree for the Quota Rule.

(Required)
policy-name the name of the policy for which the quota rule applies.

(Required)
target the target for the quota rule.

(Required)
qtree the name of qtree for which the quota rule applies

(Optional)
These are arguments that can be passed into NACL::C::VolueQuotaPolicyRule->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the quota rule based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete

=head1 PRIVATE METHODS

=head2 _nacl_create_call

=cut

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a vserver quota rule.
# Details:    Makes the NACL::C::VolumeQuotaPolicyRule->create() call. Also works
#             around some common non-fatal errors that can be return from
#             vserver quota rule creation and then has a retry mechanism to
#             retry vserver creation.
# Inputs:     NACL::C::VolumeQuotaPolicyRule->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the vserver quota rule was a duplicate if a hash is
#             requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $duplicated = FALSE;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 3;
    my $policy_enabled = exists($args{policy_enabled}) ? delete($args{policy_enabled}) : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }
    try {
        # Create the policy rule
        NACL::C::VolumeQuotaPolicyRule->create(%args);

        $Log->comment("Created new Volume Quota Policy Rule.");
    } catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry > 0) {
            $pkg->_nacl_create_call(%args,
                                    retry => --$retry,
                                    if_exists => $if_exists,
                                    );
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;

        if (  ($exception->text() =~ /duplicate entry/
            or $exception->text() =~ /The pattern (\S+) already exists/)
            and  $if_exists eq 'reuse') {
            $duplicated = TRUE;
            $Log->comment("Reusing already created Volume Quota Policy Rule for Quota Policy: " . $args{'policy-name'});
        } else {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    };

    $Log->exit();
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>$duplicated) : TRUE;

} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a vserver quota rule.
# Details:    Makes the NACL::C::Vserverquotarule->delete() calls. Also
#             works around some common non-fatal errors that can be return from
#             vserver deletion and then has a retry mechanism to retry vserver
#             deletion.
# TODO:       Need to update so that it uses NACL::STask::Vserverquotarule->purge()
# Inputs:     NACL::C::Vserverquotarule->delete arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my $pkg   = shift;
    my %args = @_;

    # Locals
    my $retry = exists $args{retry} ? delete $args{retry} : 3;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    try {
        NACL::C::VolumeQuotaPolicyRule->delete(%args);
        $Log->comment("Volume Quota Policy Rule was deleted for Quota Policy: " . $args{'policy-name'});
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry > 0) {
            $pkg->_nacl_delete_call(%args,
                                    retry => --$retry,
                                   );
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if ($exception->text() =~ /entry doesn't exist/) {
            $Log->comment(
                "Volume Quota Policy Rule was already deleted.");
        } else {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    };

    $Log->exit();
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Vserver->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Vserver->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1
