# $Id: //depot/prod/test/Rsierranevada/lib/EnGenX/Controller/Task.pm#7 $

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Task Component Module
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

EnGenX::Controller::Task

=head1 DESCRIPTION

c<EnGenX::Controller::Task> is the parent class for the task layer of EnGenX.
This class contains private create and delete methods that are used by all of
the task layer libraries to create and delete their respective objects
singularly, by list, or by count.

=cut

package EnGenX::Controller::Task;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use EnGenX::SubProcess;
use EnGenX::EnGenXUtil;
use NACL::STask::Node;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# Default the CodeString all SubProcesses will be executing.
use constant CODE_STRING=>"{ use EnGenX::Controller::Task; EnGenX::Controller::Task->_subprocess_call();}";

use vars qw($DTD_To_NACL_Translate
    $DTD_To_NACL_Ignore
    $Delete_Args
    $Name
);

#==============================================================================
# Method:     _create
# Objective:  Entry Point to create either one or multiple objects serially or
#             in parallel.
# Details:    Steps: 1)   Check inputs
#                    2.a) if resource_list call _create_multiple_from_list
#                    2.b) if count call _create_multiple
#                    2.c) otherwise call _nacl_create_call
#                             _nacl_create_call is contained in the children
#                             who call this method.
# Inputs:     'resource_list' => @resource_list
#                 If this is passed in the objects in the resource list will
#                 be created based on the information that is supplied in the
#                 resource list and the CodeString that is passed in. If the
#                 resource list is passed in the corresponding NACL Create
#                 Arguments should NOT be passed in.
#             NACL Create Arguments
#                 These are arguments that can be passed into corresponding
#                 NACL create method. If these arguments are passed in
#                 the resource_list argument should NOT be passed in
#             count => integer
#                  This argument is used to create multiple volumes from the
#                  same NACL Volume Creation Arguments.  To create the names
#                  for the volumes, the volume name passed in is appended with
#                  the index that it was create at starting at 1.
#             Name  => string
#                  This is used to set the global $Name that is used by
#                  _create_multiple to define the argument to check for the
#                  name of the Object being replicated.
#             command_interface => NACL_command_interface (optional)
#                 This is the command_interface that will be used during NACL
#                 calls.  If a command_interface is not passed in then one will
#                 be created.  Passing in a command_interface helps performance
#                 because for every command_interface created the version of
#                 the Filer must be determined by creating a connection to the
#                 filer and then sending commands to get the version.
#             method_timeout => NACL_method_timeout (optional)
#                 This is the method_timeout that will be used during NACL
#                 calls. If a method_timeout is not passed in then the
#                 method_timeout will be defaulted to 3600.
# Outputs:    This method returns a True is the method succeeds in the creation
#             of the Volume(s) based on the Options passed in and no exceptions
#             are thrown.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _create {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Globals - Assign Values
    $Name = delete($args{'Name'});

    #Assign Values
    $args{'method-timeout'} = 3600 if (!exists $args{'method-timeout'});

    # Sort
    if (exists($args{resource_list})) {
        $pkg->_create_multiple_from_list(%args);
    } elsif (exists($args{count})) {
        $pkg->_create_multiple(%args);
    } else {
        $pkg->_nacl_create_call(%args);
    }
    $Log->exit();
    return TRUE;
} ## end sub _create

#==============================================================================
# Method:     _create_multiple
# Objective:  Creates as many Objects as specified by the user.
# Details:    Will create N Objects from 1 to 'count', using the current index
#             while looping as a suffix to the Objects Name. Needs to be called
#             by a child because _nacl_create_call is contained in the
#             children of EnGenX::Controller::Task.
# Inputs:     NACL Create Arguments
#                 These are arguments that can be passed into corresponding
#                 NACL create method.
#             count => integer
#                 This arguments is used to create multiple objects from the
#                 same NACL Create Arguments.  To create the names
#                 for the objects, the object name passed in is appended with
#                 the index that it was create at starting at 1.
#             command_interface => NACL_command_interface (optional)
#                 This is the NACL command_interface object that will be used
#                 during NACL calls.  If a command_interface is not passed in
#                 then one will be created.  Passing in a command_interface
#                 helps performance because for every command_interface created
#                 the version of the Filer must be determined by creating a
#                 connection to the filer and then sending commands to get the
#                 version.
#             method_timeout => NACL_method_timeout (optional)
#                 This is the method_timeout that will be used during NACL
#                 calls. If a method_timeout is not passed in then the
#                 method_timeout will be defaulted to 3600.
# Outputs:    This method returns a True or False based on whether the method
#             succeeds or fails in the creation of the Object(s) based on the
#             Options passed in.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _create_multiple {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $cmd_exec = {};
    my $count = exists($args{count}) ? delete($args{count}) : 1;
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    foreach my $id (1 .. $count) {
        # Create a unique name.
        if ($Name =~ /^volume?/ and $id > 99 and $id < 1000) {
            # 99 < $id < 1000 is padded for stripped volume
            #       reserved name error.
            $args{$Name} = $args{$Name} . '0' . $id;
            if ($id == $count) {
                # When creating the last volume, wait for it to complete.
                $args{'wait'} = TRUE;
            }
        }
        else {
            $args{$Name} = $args{$Name} . $id;
        }
        $cmd_exec = $pkg->_nacl_create_call(%args);

        #
        # TODO:
        #
        # We could return an array with the names of the resources that were actually
        # created Or just return an array of names of the resources that were duplicated.
        # Also, could return an array with the complete description of the resources,
        # stating wich one was created and which one is been reused.
    }

    $Log->exit();
    return TRUE;
} ## end sub _create_multiple

#==============================================================================
# Method:     _create_multiple_from_list
# Objective:  Creates all the objects in a resource_list consisting of one
#             object type that is passed in.
# Details:    This method will call private method '_multiple_from_list_helper'
# Inputs:     Will pass them through.
#             Will add argument 'action' with value CREATE to let the method
#             _multiple_from_list_helper know what to do.
# Outputs:    This method returns the output of _multiple_from_list_helper
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _create_multiple_from_list {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;
    # Add the ACTION parameter to let private methods what to do.
    $args{action} = "Create";

    $Log->exit();
    return $pkg->_multiple_from_list_helper(%args);

} ## end sub _create_multiple_from_list

#==============================================================================
# Method:     _delete
# Objective:  Entry Point to delete either one or multiple objects serially or
#             in parallel.
# Details:    Steps: 1)   Check inputs
#                    2.a) if resource_list call delete_multiple_from_list
#                    2.b) if count call _delete_multiple
#                    2.c) otherwise call _nacl_delete_call
#                             _nacl_delete_call is contained in the children
#                             who call this method.
# Inputs:     'resource_list' => @resource_list
#                 If this is passed in the objects in the resource list will
#                 be deleted based on the information that is supplied in the
#                 resource list and the CodeString that is passed in. If the
#                 resource list is passed in the corresponding NACL delete
#                 Arguments should NOT be passed in.
#             NACL delete Arguments
#                 These are arguments that can be passed into corresponding
#                 NACL delete method. If these arguments are passed in
#                 the resource_list argument should NOT be passed in
#             count => integer
#                  This arguments is used to delete multiple volumes from the
#                  same NACL Volume Creation Arguments.  To delete the names
#                  for the volumes, the volume name passed in is appended with
#                  the index that it was delete at starting at 1.
#             Name  => string
#                  This is used to set the global $Name that is used by
#                  _create_multiple to define the argument to check for the
#                  name of the Object being replicated.
#             command_interface => NACL_command_interface (optional)
#                 This is the command_interface that will be used during NACL
#                 calls.  If a command_interface is not passed in then one will
#                 be created.  Passing in a command_interface helps performance
#                 because for every command_interface created the version of
#                 the Filer must be determined by creating a connection to the
#                 filer and then sending commands to get the version.
#             method_timeout => NACL_method_timeout (optional)
#                 This is the method_timeout that will be used during NACL
#                 calls. If a method_timeout is not passed in then the
#                 method_timeout will be defaulted to 3600.
# Outputs:    This method returns a True or False based on whether the method
#             succeeds or fails in the deletion of the Volume(s) based on the
#             Options passed in.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _delete {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Globals - Assign Values
    $Name = delete($args{'Name'});

    #Assign Values
    $args{'method-timeout'} = 3600 if (!exists $args{'method-timeout'});

    # Sort
    if (exists($args{resource_list})) {
        $pkg->_delete_multiple_from_list(%args);
    } elsif (exists($args{count})) {
        $pkg->_delete_multiple(%args);
    } else {
        $pkg->_nacl_delete_call(%args);
    }
    $Log->exit();
    return TRUE;
} ## end sub _delete

#==============================================================================
# Objective:  Deletes as many Objects as specified by the user.
# Details:    Will delete N Objects from 1 to 'count', using the current index
#             while looping as a suffix to the Objects Name. Needs to be called
#             by a child because _nacl_delete_call is contained in the
#             children of EnGenX::Controller::Task.
# Inputs:     NACL Delete Arguments
#                 These are arguments that can be passed into corresponding
#                 NACL delete method.
#             count => integer
#                  This arguments is used to delete multiple objects from the
#                  same NACL Delete Arguments.  To create the names
#                  for the objects, the object name passed in is appended with
#                  the index that it was deleted at starting at 1.
#             command_interface => NACL_command_interface (optional)
#                 This is the command_interface that will be used during NACL
#                 calls.  If a command_interface is not passed in then one will
#                 be created.  Passing in a command_interface helps performance
#                 because for every command_interface created the version of
#                 the Filer must be determined by creating a connection to the
#                 filer and then sending commands to get the version.
#             method_timeout => NACL_method_timeout (optional)
#                 This is the method_timeout that will be used during NACL
#                 calls. If a method_timeout is not passed in then the
#                 method_timeout will be defaulted to 3600.
# Outputs:    This method returns a True or False based on whether the method
#             succeeds or fails in the deletion of the Object(s) based on the
#             Options passed in.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _delete_multiple {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $cmd_exec = {};
    my $count = exists($args{count}) ? delete($args{count}) : 1;
    $args{'command_interface'} = NACL::C::Node->find()
        if (!exists $args{'command_interface'});

    foreach my $id (1 .. $count) {
        # Create a unique name.
        if ($Name =~ /^volume?/ and $id > 99 and $id < 1000) {
            # 99 < $id < 1000 is padded for stripped volume
            #       reserved name error on creation.
            $args{$Name} = $args{$Name} . '0' . $id;
        }
        else {
            $args{$Name} = $args{$Name} . $id;
        }
        $cmd_exec = $pkg->_nacl_delete_call(%args);

  #
  # TODO:
  #
  # We could return an array with the names of the Vols that were actually
  # deleted Or just return an array of names of the vols that were duplicated.
  # Also, could return an array with the complete description of the vols,
  # stating wich one was deleted and which one is been reused.
    }

    $Log->exit();
    return TRUE;
} ## end sub _delete_multiple

#==============================================================================
# Method:     _delete_multiple_from_list
# Objective:  Deletes all the objects in a resource_list consisting of one
#             object type that is passed in.
# Details:    This method will call private method '_multiple_from_list_helper'
# Inputs:     Will pass them through.
#             Will add argument 'action' with value DELETE to let the method
#             _multiple_from_list_helper know what to do.
# Outputs:    This method returns the output of _multiple_from_list_helper
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _delete_multiple_from_list {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;
    # Add the ACTION parameter to let private methods what to do.
    $args{action} = "Delete";

    $Log->exit();
    return $pkg->_multiple_from_list_helper(%args);

} ## end sub _delete_multiple_from_list

#==============================================================================
# Method:     _multiple_from_list_helper
# Objective:  Creates or Deletes objects in a resource_list consisting of one
#             object type that is passed in.
# Details:    Steps: 1)   Sort Resource List by parallel factor.
#                    2)   Loop through the parallel factors and either directly
#                         call the _subprocess_action_call or will create|delete
#                         resources based the parallel factor as the number of
#                         SubProcesses to use.  This is done by calling
#                         _create_parallel_elements
# Inputs:     action => Create|Delete
#             'resource_list' => @resource_list
#                 The objects in the resource list will be either created or
#                 deleted  based on the information that is supplied in the
#                 resource list and the $action.
#             command_interface => NACL_command_interface (optional)
#                 This is the command_interface that will be used during NACL
#                 calls.  If a command_interface is not passed in then one will
#                 be created.  Passing in a command_interface helps performance
#                 because for every command_interface created the version of
#                 the Filer must be determined by creating a connection to the
#                 filer and then sending commands to get the version.
#             method_timeout => NACL_method_timeout (optional)
#                 This is the method_timeout that will be used during NACL
#                 calls. If a method_timeout is not passed in then the
#                 method_timeout will be defaulted to 3600.
# Outputs:    This method returns a resource_list reference that was passed on.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _multiple_from_list_helper {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my @updated_list  = ();
    my $action = $args{'action'};
    my $resource_list = $args{resource_list};
    my $sorted_lists  = EnGenX::EnGenXUtil->list_parallel_sorter(
                                    'resource_list' => $resource_list);

    # COMMAND_INTERFACE value could come from three different sources:
    # 1. The user could provide it, usualy users that are not using EnGenX.
    # 2. EnGenX (as of now), will provide the NODES array to enable load
    #    balacing, so let's get the first node from the array.
    # 3. The user provided niether, so let's find out one.
    unless (defined($args{nacl_nodes})) {
        if (exists($args{command_interface})) {
            my $ci = delete($args{command_interface});
            $args{nacl_nodes} =  [$ci];
        } else {
            $args{nacl_nodes} = [NACL::C::Node->find()];
        }
    }

    $Log->debug("The Before Task $action\_multiple_from_list is:\n" .
                Dumper($resource_list))
        if ($Log->may_debug());


    # Create the Elements without creating a subProcess (PARALLEL=0)
    # if (PARALLEL=0) then creation is performed serial in the same process,
    # if (PARALLEL=1) then a single child process is created and resource
    # creation is performed serially in the created process.
    if (defined($sorted_lists->{'0'})) {
        my $list_no_parallel = delete $sorted_lists->{'0'};
        # Work with each resource
        foreach my $resource (@{$list_no_parallel}) {
            # Execute
            $pkg->_subprocess_action_call(
                             action   => $action,
                             resource => $resource,
                    command_interface => $args{nacl_nodes}->[0],
                );
        }
    }#if parallel=0

    # Work on the rest of the hash (subProcess with PARALLEL>0)
    foreach my $parallel_count (sort keys(%{$sorted_lists})) {
        # SubProcess count will say how many subprocess will be created to
        # execute the list.
        my $rcsr_count = scalar(@{$sorted_lists->{$parallel_count}});
        # This is setting the number of subprocesses we create to the smaller
        # number of the resources to be created or the parallel factor to use
        # when creating the resources.
        my $subproc_count
            = ($parallel_count > $rcsr_count) ? $rcsr_count : $parallel_count;
        my %parallel_args = (
                        'sorted_list'  => $sorted_lists->{$parallel_count},
                        'parallel_index' => $subproc_count,
                        %args
                    );
        # Execute
        $pkg->_create_parallel_elements(%parallel_args);
        # TODO: This could be a good (not the best) place to implement some sort
        #       of logic to safe the resources into a DB or flat file. I think
        #       if will be better if we do that at the Controller level.
    }# foreach


    $Log->debug(
        "The After Task Create_parallel_elements is:\n" . Dumper($resource_list))
        if ($Log->may_debug());

    $Log->exit();
    return $resource_list;

} ## end sub _multiple_from_list_helper


#==============================================================================
# Method:     _create_parallel_elements
# Objective:  Creates the number of SubProcesses needed and then runs the
#             calls the SubProcess management method to manage the processes.
# Details:    Steps: 1)   Create the number of subprocesses based on
#                         parallel factor and CodeString
#                    2)   Call the management method parallel_execution and
#                         pass in the resource_list the SubProcesses will be
#                         using.
# Inputs:     'resource_list' => @resource_list
#                 The resource_list is the list of objects that are passed to
#                 the management method to divide and send to the method in the
#                 global $Codestring to perform the required task
#             parallel_index  => integer
#                 The number of SubProcesses to create and use to perform the
#                 required task.
#             command_interface => NACL_command_interface (optional)
#                 This is the command_interface that will be used during NACL
#                 calls.  If a command_interface is not passed in then one will
#                 be created.  Passing in a command_interface helps performance
#                 because for every command_interface created the version of
#                 the Filer must be determined by creating a connection to the
#                 filer and then sending commands to get the version.
#             method_timeout => NACL_method_timeout (optional)
#                 This is the method_timeout that will be used during NACL
#                 calls. If a method_timeout is not passed in then the
#                 method_timeout will be defaulted to 3600.
# Outputs:    This method returns an updated resource_list or a ref to an
#             updated resource list based on whether the caller wants an array
#             or scalar.
# Exceptions: BaseException
# TODO:       Update to where the command_interface and method_timeout are
#             passed through and used.
# TODO:       Create Specific EnGenX exceptions instead of useing BaseException
#==============================================================================
sub _create_parallel_elements {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # LOCALS
    my $subproc_obj   = undef;
    my $resource_list = $args{'sorted_list'};
    # Create the PARAMS used when creating the SubProcesses
    my %call = ('package'=> $pkg,
                'action' => $args{'action'},
                'nacl_nodes' => $args{nacl_nodes},
            );

    # Parameters that will be use to create the subprocesses.
    my %subproc_params = (
        'code_string' => CODE_STRING,
        'count'       => $args{'parallel_index'},
        'EnGenX'      => $args{'EnGenX'},
    );

    # Create a SubProcess object which will facilitate parallel
    # execution.
    unless ($subproc_obj = EnGenX::SubProcess->new(%subproc_params)) {
        $Log->exit();
        NATE::BaseException->throw(
            "Error while creating a Subprocess EnGenX::SubProcess"
                . Dumper(%subproc_params));
    }

    # Use the Subprocess Object to create resources in parallel.
    $resource_list = $subproc_obj->parallel_execution('call' => \%call,
                                                      'list' => $resource_list
                                                    );

    $Log->exit();
    return (wantarray()) ? @{$resource_list} : \$resource_list;

} ## end sub _create_parallel_elements

#==============================================================================
# Method:     _subprocess_call
# Objective:  Subprocess call to manage communication with a parent process
#             that passes resources to be created or deleted based on the method
#             that called this subrutine.
# Details:    Steps: 1)   Tell Parent waiting on resource
#                    2.a) if parent sends new resource(s) continue to step 3.
#                    2.b) if parent sends 'exit' then create pass result and
#                         die.
#                    3)   Delete lif(s) based on resource(s) sent in and
#                         update resources.
#                    4)   send message to parent with updated resource(s) and
#                         tell parent waiting on resource again.
#                    5)   Repeat steps 2-4 until 2.b occurs or exceptions out.
# Inputs:     Does NOT take inputs. Instead takes arguments via messages from
#             the parent process. The two accepted messages are both of type
#             "WhatToDo". One message is 'exit' which tells this process to
#             exit and return a pass result.  The second message is an
#             Array of lif resources to be deleted.
# Outputs:    Does NOT return anything.  Instead creates a result object based
#             on whether an error occurs while deleting the lifs in the
#             list. If uncaught errors occur then a Fail is made otherwise a
#             Pass is made and on die the result is passed back to the parent
#             process.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _subprocess_call {
    use NATE::STAF::Peer qw($Parent);
    use NATE::Error qw(:try);
    use Data::Dumper;
    # Locals
    # ParamSet will retrieve all the variables that were assigned to the
    # 'params' key when creating the NATE::Process object.
    my $global_params = NATE::ParamSet->get_global();
    my $pkg = $global_params->get('package');
    eval ("require $pkg");
    my $action = $global_params->get('action');
    my $node_obj  = NACL::STask::Node->new(node => $global_params->get('node'));
    my $exit_flag = TRUE;
    my $message = '';
    my $to_die = 0;
    try {
        # Tell parent proc is ready for execution.
        $Log->trace("Started SubProc: '$pkg' @ Task::_subprocess_call:\n".
                    "Parent->messange_put - type[RESPONSE] status[WAITING]");
        $Parent->message_put(
                        type    => "Response",
                        message => {'status' => "waiting"}
                    );
        while ($exit_flag)
        {
            $Log->debug("SubProc: '$pkg' @ Task::_subprocess_call:\n".
                        "Parent->messange_get - TYPE[WhatToDO]");
            $message = $Parent->message_get(
                type    => "WhatToDo",
                timeout => 180
            );

            # THERE ARE THREE POSSIBLE MESSAGES TYPE 'WhatToDo'
            # 1. 'exit'  - MESSAGE TO TERMINATE THE SUB-PROCESS.
            # 2. HASHREF - A REFERENCES TO HASH with KEYs 'wait'_completion'
            #              AND 'resources'. INDICATING TO WAIT_FOR_COMPLETION
            #              OF THE PROVIDED RESROUCE AND THEN EXIT.
            # 3. HASHREF - A REFERENCES TO HASH with KEYs 'rsrc_index' AND
            #              'resources'. MEANING CREAT/DELETE THE PROVIDED
            #              RESOURCE BUCKET THAT HAS THE MENTIONED INDEX.
            if ($message =~ /exit/) {
                # 1. Exit
                $Log->debug("SubProc: '$pkg' @ Task::_subprocess_call:\n".
                            "GOT MESSAGE - type[WhatToDO], mssg[exit]");
                $exit_flag = FALSE;
            } elsif (ref($message) =~ /HASH/) {
                # 2. MODIFY THE ACTION, THE SUB WILL TAKE CARE OF THE'WAIT' CALL
                if (defined ($message->{wait_completion})) {
                    $action = "WaitCompletion";
                }
                $Log->trace("SubProc: '$pkg' @ Task::_subprocess_call:\n".
                            "GOT RESOURCE(S) to execute action[$action]");


                # 3. LOOP THROUGH THE COMPLETE LIST OF RESOURCES
                foreach my $resource (@{$message->{resources}}) {
                    # Locals
                    try {
                        $pkg->_subprocess_action_call(
                                        action   => $action,
                                        resource => $resource,
                                        command_interface => $node_obj,
                                        );
                    }
                    otherwise {
                        my $exception = shift;
                        if ($Log->may_error()) {
                            $Log->error(
                                "Exception Thrown while executing command ".
                                "at EnGenX::Task::_subprocess_call for $pkg->" .
                                "$action on Resource: \n" .
                                Dumper($resource) . " \n Exception: " . ref($exception) . "\n" .
                                $exception->text()
                            );
                        }

                        # Returning unchanged resource because of failure.
                        $Parent->message_put(
                            type    => "Response",
                            message => {
                                status => "failed",
                                %{$message}
                            }
                        );
                        $to_die = 1;
                    };
                    last if ($to_die);
                }
                last if ($to_die);
                # Let the parent know that the _action_call was succesful
                if ($Log->may_trace()) {
                    $Log->trace("Returning message to parent \nMessage: "
                                . Dumper($message));
                }

                # Sending the message
                $Parent->message_put(
                                        type    => "Response",
                                        message => {'status' => "waiting",
                                                    %{$message}
                                                }
                                        );
            } else {
                if ($Log->may_warn()) {
                    $Log->warn("Found a message that does not match known "
                            . "messages.\nMessage REF: "
                            . ref($message)
                            . "\nMessage:\n "
                            . Dumper($message));
                }

                # Get another bucket to work with
                $Parent->message_put(
                    type    => "Response",
                    message => {
                        'status' => "waiting",
                        %{$message}
                    }
                );
            }
        }# end while
    }# end try
    otherwise {
        my $exception = shift;
        if ($Log->may_error()) {
            $Log->error("Exception Thrown in WHILE LOOP at " .
                        "Task::_subprocess_call is: " . Dumper($exception)
                    );
        }
        $Parent->message_put(
            type    => "Response",
            message => {
                'status'    => "failed",
                %{$message}
            }
        );
        $to_die = 1;
    };

    if($to_die) {
        my $result = NATE::Result::Fail->new("Subprocess Fail");
        die $result;
    }
    return;
}    # _subprocess_call

#==============================================================================
# Method:     _subprocess_action_call
# Objective:  Will make the call to each TASK create, delete or wait call and
#             update accordingly
# Details:    .
# Inputs:     'action' => Create|Delete|WaitCompletion
#                 This parameter will tell what method to call with the provided
#                 resource
#             resource  => $resource
#                 The resource to work with.
#             command_interface => NACL_command_interface (optional)
#                 This is the command_interface that will be used during NACL
#                 calls.  If a command_interface is not passed in then one will
#                 be created.  Passing in a command_interface helps performance
#                 because for every command_interface created the version of
#                 the Filer must be determined by creating a connection to the
#                 filer and then sending commands to get the version.
# Outputs:    The update resource.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _subprocess_action_call {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    # Locals
    my $action = $opts{'action'};
    my $resource = $opts{resource};
    my $node = $opts{'command_interface'};
    # Assign Values
    my $translate = $pkg->_get_dtd_nacl_translate();
    my $ignore = $pkg->_get_dtd_nacl_ignore();
    my $delete = $pkg->_get_dtd_nacl_delete();

    # BASED ON THE ACTION EITHER CREATE, DELETE OR WAIT COMPLETION OF THE
    # PROVIDED RESOURCE
    if ($action =~/Create/i) {
        # Move on if no setup was requested or setup has already been done on
        #this resource.
        if( ($resource->{'_setup'} == 0)||($resource->{'_setup_done'} == 1) ){
            $Log->comment("Setup for resource '". $resource->{NAME} .
                          "' was skipped because: " .
                          " _setup=" . $resource->{'_setup'} .
                          " _setup_done=" . $resource->{'_setup_done'}
                        );
            return $resource;
        }

        my %args = EnGenX::EnGenXUtil->strip_and_translate_excess_args(
                    translation_hash => $translate,
                    ignore_hash      => $ignore,
                    %{$resource}
                );

        if ($pkg =~ /EnGenX::Controller::Lif/i){
            %args = $pkg->_get_lif_params(
                                    resource => $resource,
                                    args => \%args,
                                );
        }

        #Add the Command Interface
        $args{command_interface} = $node;
        # EXECUTE COMMAND
        my %cmd_exec = $pkg->_nacl_create_call(%args);

        # UPDATE RESULT
        $resource->{'_setup_done'} = TRUE;
        $resource->{'_cleanup_done'} = FALSE;

        if (%cmd_exec) {
            if ($cmd_exec{'duplicated'}) {
                $resource->{'_reused'} = TRUE;
            } else {
                $resource->{'_reused'} = FALSE;
            }
            # Update the address so we know what address wasused for this resource
            if (defined($cmd_exec{'used_address'})) {
                $resource->{address} = $cmd_exec{'used_address'};
            }
        }
    }
    elsif ($action =~/Delete/i) {
        # Move on if no cleanup was requested, cleanup has already been done on
        # this resource, or cleanup policy tells us not to clean this up.
        if( ($resource->{'_cleanup'} == 0) or
            ($resource->{'_cleanup_done'} == 1) or
            (($resource->{'_cleanup_policy'} eq 'new') && ($resource->{'_reused'} == 1))) {
            $Log->comment("Cleanup for resource '". $resource->{NAME} .
                          "' was skipped because" .
                          " _cleanup=". $resource->{'_cleanup'} .
                          " _cleanup_done=" . $resource->{'_cleanup_done'} .
                          " _cleanup_policy=" . $resource->{'_cleanup_policy'} .
                          " _reused=" . $resource->{'_reused'}
                        );
            return $resource;
        }
        # ARGUMENTS
        # Concatenating the arguments under args key.
        my %args = EnGenX::EnGenXUtil->strip_and_translate_excess_args(
                    translation_hash => $translate,
                    delete_hash      => $delete,
                    %{$resource}
                );
        #Add the Command Interface
        $args{command_interface} = $node;
        # EXECUTE COMMAND
        try {
            $pkg->_nacl_delete_call(%args);
        } catch NACL::APISet::Exceptions::ResponseException with {
            my $exception = shift;

            # Check for common error for vserver not existing. This error
            # is seen by most objects when deleted and the vserver has not
            # been created. Putting here instead of in each individual
            # task.
            # vserver not found is for zapi
            # invalid value for field "-vserver is for cli
            if ($exception->text() =~ /vserver not found|does not exist/
                or $exception->text() =~ /invalid value for field "-vserver/)
            {
                my $type = $pkg;
                $type =~ s/EnGenX::Controller:://;
                $Log->comment("$type:" . $args{lc($type)} . " was already deleted");
            } else {
                # Rethrowing caught exception
                $Log->exit();
                $exception->throw();
            }
        };
        # UPDATE RESULT
        $resource->{'_cleanup_done'} = TRUE;
        $resource->{'_setup_done'} = FALSE;
    }
    elsif ($action =~/WaitCompletion/i) {
        # ARGUMENTS
        my %args = EnGenX::EnGenXUtil->strip_and_translate_excess_args(
                    translation_hash => $translate,
                    ignore_hash      => $ignore,
                    %{$resource}
                );
        #Add the Command Interface
        $args{command_interface} = $node;
        # EXECUTE COMMAND
        $pkg->_nacl_wait_call(%args);
    }


    $Log->exit();
    return $resource;
}## end sub _subprocess_action_call

1;
