# $Id: //depot/prod/test/Rsierranevada/lib/EnGenX/Controller/Volume.pm#9 $

#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Volume Controller Layer Module
## @author dl-san-automation
## @status shared
## @pod here

=head1 NAME

EnGenX::Controller::Volume

=head1 DESCRIPTION

EnGenX::Controller::Volume is a derived class of
EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP Volumes. All methods are static.

=cut

package EnGenX::Controller::Volume;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::CS::StorageAggregate;
use NACL::C::Volume;
use NACL::C::VolumeFlexcache;
use NACL::CS::Volume;
use NACL::STask::Volume;
use NACL::C::Job;
use EnGenX::SubProcess;
use EnGenX::EnGenXUtil;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::C::Exceptions::Volume::AlreadyMounted;
use NACL::C::Exceptions::Volume::AlreadyModified;

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args
);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name   => 'volume',
    aggr   => 'aggregate',
    _reuse => 'if_exists',
    export_policy => 'policy'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods.
%DTD_To_NACL_Ignore = (
    efficiency             => '',
    parallel               => ''
);

# Arguments that can be passed into the NACL delete
# command.
%Delete_Args = (
    command_interface       => '',
    apiset_must             => '',
    apiset_should           => '',
    volume                  => '',
    vserver                 => '',
    foreground              => '',
    'disable-offline-check' => '',
    force                   => '',
    job_component           => '',
    'is-repository'         => '',
    enable_flexcache        => ''
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple volumes serially
or in parallel and return whether the volume(s) was created successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Volume->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            aggregate         => $Aggr_name,
            vserver           => $Vserver_name,
            size              => "20M",
            job_component     => \$job,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            aggregate         => $Aggr_name,
            vserver           => $Vserver_name,
            size              => "20M",
            job_component     => \$job,
            'method-timeout'  => 3600,
            count             => 5,
            );

  EnGenX::Controller::Volume->create(
            command_interface  => $Node,
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume->create(
            command_interface => $Node,
            resource_list      => @Vols,
            );

  EnGenX::Controller::Volume->create(
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume->create(
            resource_list      => @Vols,
            );

=over

=item Options

Either the resource_list is required or the NACL Volume Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the volumes in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Volume Creation Arguments should
NOT be passed in.

=item C<< NACL Volume Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::C::Volume->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< count => $volume_count >>

(Optional)
This arguments is used to create multiple volumes from the same NACL Volume
Creation Arguments.  To create the names for the volumes, the volume name
passed in is appended with the index that it was create at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted to 3600.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Volume(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # TODO Remove $args{'Name'} from everywhere. And ensure
    # no errors occur.
    # NOTE: Can't remove the NAME, it is been used to create objects with a
    # counter. We could get the name from package by asking the for the "caller"
    $args{'Name'} = "volume";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);
    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple volumes serially
or in parallel and return whether the volume(s) was deleted successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Volume->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            count             => 5
            );

  EnGenX::Controller::Volume->delete(
            command_interface => $Node,
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume->delete(
            command_interface => $Node,
            resource_list      => @Vols,
            );

  EnGenX::Controller::Volume->delete(
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume->delete(
            resource_list      => @Vols,
            );

=over

=item Options

Either the resource_list is required or the NACL Volume Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the volumes in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Volume Deletion Arguments should
NOT be passed in.

=item C<< NACL Volume Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::STask::Volume->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< count => $volume_count >>

(Optional)
This arguments is used to delete multiple volumes from the same NACL Volume
Deletion Arguments.  To delete the names for the volumes, the volume name
passed in is appended with the index that it was deleted at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted to 3600.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Volume(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # TODO Remove $args{'Name'} from everywhere. And ensure
    # no errors occur.
    $args{'Name'} = "volume";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a volume.
# Details:    Makes the NACL::STask::Volume->create() call and optionaly could
#             make a the NACL::STask::Volume->wait_for_creation. Also works
#             around some common non-fatal errors that can be return from
#             volume creation and then has a retry mechanism to retry volume
#             creation.
# TODO:       Need to update to NACL::STask::Volume->create()
# Inputs:     NACL::C::Volume->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the volume was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # If snapshot-policy is set for constituent vols delete the attribute.
    if(exists $args{'constituent-role'} and exists $args{'snapshot-policy'}) {
       delete $args{'snapshot-policy'};
       $Log->comment("Snapshot-policy cannot be set when creating a constituent volume" .
                     " The attribute has been removed from one of your [" .
                      $args{'volume'} . "] volumes on the vserver: " .
                      $args{'vserver'} . "./n");
    }

    # Locals
    my $duplicated = FALSE;
    my $is_infinite_vol = FALSE;
    my $is_managed_by_service = FALSE;
    my ($percent, $sizer, $max_size) = "";
    my $wait = exists($args{wait}) ? delete($args{wait}) : FALSE;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    my $files = delete $args{files};
    # account for the depricated XML percent_snapshot_space
    if(defined($args{"percent_snapshot_space"})) {
        $Log->warn("XML option percent_snapshot_space has been deprecated. " .
                   "Use XML option percent-snapshot-space.");
        # set the variables correctly.
        $args{"percent-snapshot-space"} = delete($args{"percent_snapshot_space"});
    }

    # If the size argument has a percent then calculate the size.
    if ($args{size} =~ s/%//) {
        my $aggr_cs_obj = NACL::CS::StorageAggregate->fetch(
                                            command_interface => $args{command_interface},
                                            requested_fields => ['size'],
                                            filter => { aggregate => $args{aggregate},
                                                      },
                                            );
        my $percent = $args{size};
        $args{size} = ($args{size}/100)*($aggr_cs_obj->size());
        $Log->comment("Setting Volume size to " . $args{size} . "b which is " .
                      "$percent\% of aggregate " . $args{aggregate} .
                      " which is size " .     $aggr_cs_obj->size() . "b");
    }

    # Delete the vol autosize-percent option if set and the volume is
    # reused autosize is set to false or the maxsize is set
    if (exists($args{'autosize-percent'})) {
        # remove any percentage signs on the autosize-percent flag
        $args{'autosize-percent'} =~ s/%//g;

        # Delete the autosize-percent if any of these are set
        if (exists($args{'max-autosize'}) or ($args{autosize} =~ /false/i)) {
            delete $args{'autosize-percent'};
        } elsif ($args{autosize} =~ /true/i) {
            # Else we need to do the math and fill in the values.
            # Find the percentage used ex 50% = .5
            $percent = sprintf("%f",.01 * $args{'autosize-percent'});
            delete $args{'autosize-percent'};

            # Look for the sizer
            # The variable $1 holds the last pattern matched.  If there
            # is a sizer, it will hold that sizer (k, m, g, t, p).  If there
            # is not a sizer it will hold "true" from where we checked
            # the autosize in the elsif above.
            $sizer = "";
            if ($args{size} =~ /\d+(\D)$/) {
                $sizer = $1;
            }

            # Do the autosize percent calculations and set the values
            if ($percent <= 1) {
                # the user wants a percentage less than 100

                # set the max size to be the specified size of the volume
                $args{'max-autosize'} = $args{size};

                # remove the unit of measurement from the size value
                # so that we can do the math calculation
                $args{size} =~ s/\D$//;

                # adjust the size to be the percentage specified against
                # the volume size
                $args{size} = $args{size} * $percent;

                # if a unit of measurement was provided, put it back on
                if ($sizer ne "") {
                    $args{size} = $args{size} . $sizer;
                }
            } elsif ($percent > 1) {
                # the user wants the percentage greater than 100

                # remove the unit of measurement from the size value
                # so that we can do the math calculation
                $args{size} =~ s/\D$//;

                # adjust the max size
                $args{'max-autosize'} = $args{size} * $percent;

                # assign the max size with the unit of measurement
                # if it was provided
                if ($sizer ne "") {
                    $args{'max-autosize'} .= $sizer;
                }
            }
        } else {
            # catch every other case and delete autosize-percent
            delete $args{'autosize-percent'};
        }
    }

    if (delete($args{use_deep_junction})) {
        delete($args{'junction-path'});
    }

    # parse out the use default junction path flag
    my $use_default_junction_path = delete($args{use_default_junction_path});

    # if the user specified to use the default junction path,
    # then default the junction path to be the name of the volume
    # only do this if the user did not explicitly provide a junction path
    if ((defined($use_default_junction_path) and $use_default_junction_path eq 1)
        and
        (not exists($args{'junction-path'}) or not length($args{'junction-path'})) ) {
        # set the junction path to be "/$volname"
        $args{'junction-path'} = "/" . $args{volume};
    }

    my $enable_flexcache =
        exists($args{enable_flexcache}) ? delete($args{enable_flexcache}) : '0';

    # This creates a NACL command_interface if one is not passed in.
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # Execute Create
    try {
        NACL::C::Volume->create(%args);
        $Log->comment("Using newly Created Volume:" . $args{volume});
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 3) {
            sleep 30 * ($retry++);
            $pkg->_nacl_create_call(
                            %args,
                            retry     => $retry,
                            if_exists => $if_exists,
                            wait      => $wait
                        );
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    }
    catch NACL::C::Exceptions::Volume::AggregateRequiredForCreate with {
        $Log->comment("Caught AggregateRequiredForCreate exception " .
                      "checking if this volume already exists because " .
                      "infinite volumes dont require aggregates but on reuse " .
                      "have this error.");
        my $vol = NACL::CS::Volume->fetch(
                                        filter => {volume  => $args{volume},
                                                   vserver => $args{vserver}},
                                        command_interface => $args{command_interface},
                                        requested_fields => ['volume','vserver','volume-style']);
        $duplicated = TRUE;
        $Log->comment("Using already Created Volume:" . $args{volume});

        # load up some extra parameters for the online command
        my %online_args = ();
        $online_args{apiset_should} = $args{apiset_should} if (exists $args{apiset_should});

        # need to verify that the volume is online, online it
        try {
            NACL::C::Volume->online(command_interface => $args{command_interface},
                                    vserver => $args{vserver},
                                    volume => $args{volume},
                                    %online_args);
        } catch NACL::APISet::Exceptions::ResponseException with {
            my $e = shift;
            # rethrow the exception if the exception was anything other than "is already online"
            if ($e->text() !~ /is already online/i) {
                $e->throw();
            }
        };

        # Set flag if the existing volume is an infinite volume
        try {
            $is_infinite_vol = $vol->is_infinivol();
        }
        otherwise {
             $Log->comment("Failed to determine if existing volume (" . $args{volume} .") is an infinite volume.");
        };
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if (   ($if_exists eq 'reuse')
            && ($exception->text() =~ /Duplicate volume name .*/)) {
            $duplicated = TRUE;
            $Log->comment("Using already Created Volume:" . $args{volume});

            # load up some extra parameters for the online command
            my %online_args = ();
            $online_args{apiset_should} = $args{apiset_should} if (exists $args{apiset_should});

            # need to verify that the volume is online, online it
            try {
                NACL::C::Volume->online(command_interface => $args{command_interface},
                                        vserver => $args{vserver},
                                        volume => $args{volume},
                                        %online_args);
            } catch NACL::APISet::Exceptions::ResponseException with {
                my $e = shift;
                # rethrow the exception if the exception was anything other than "is already online"
                if ($e->text() !~ /is already online/i) {
                    $e->throw();
                }
            };
        } elsif ($exception->text() =~ /Authorization failed/ && $retry < 3) {
            sleep 30 * ($retry++);
            $pkg->_nacl_create_call(
                            %args,
                            retry     => $retry,
                            if_exists => $if_exists,
                            wait      => $wait
                        );

         # TODO: figure out a way to shorten the aggrlist error code if
         #       possible.
         # A vserver has an aggrlist that tells which aggregate that vserver
         # can do vserver specific commands on.  When running from the vserver
         # if the aggrlist is empty (the default) then you can do anything to
         # any aggregate. If in the vserver context (such as with zapi) or if
         # aggrlist is not empty then the aggregate that the volumes,lun,etc
         # that the commands are being run to must be in the aggrlist for the
         # vserver the commands are being run on.
        } elsif ($exception->text() =~ /not in aggrlist/) {
            my %fetch_args = (
                'requested_fields'  => ['aggr-list'],
                'command_interface' => $args{'command_interface'},
                'filter'            => {('vserver' => $args{'vserver'})},
                'method-timeout'    => $args{'method-timeout'}
            );
            $fetch_args{apiset_should} = $args{apiset_should}
                if exists $args{apiset_should};
            my @aggr_list
                = NACL::CS::Vserver->fetch(%fetch_args)->aggr_list();
            my $is_not_in_list = TRUE;
            foreach (@aggr_list) {
                if ($_ eq $args{aggregate}) {
                    $is_not_in_list = FALSE;
                    last;
                }
            }
            if ($is_not_in_list) {
                if ($aggr_list[0] eq '-') {
                    $aggr_list[0] = $args{aggregate};
                } else {
                    push(@aggr_list, $args{aggregate});
                }
                my %modify_args = (
                    'command_interface' => $args{'command_interface'},
                    'aggr-list'         => \@aggr_list,
                    'vserver'           => $args{'vserver'},
                    'method-timeout'    => $args{'method-timeout'}
                );
                $modify_args{apiset_should} = $args{apiset_should}
                    if exists $args{apiset_should};
                NACL::C::Vserver->modify(%modify_args);
            }
            # Try to create VOL again
            $pkg->_nacl_create_call(
                            %args,
                            retry     => $retry,
                            if_exists => $if_exists,
                            wait      => $wait
                        );
        } else {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    };
    # Create the arguments for the volume to wait for creation if no
    # other volumes are going to be passed in.
    if ( (!$duplicated) && ($wait) ) {
        $pkg->_nacl_wait_call(%args);
    }

    # if a junction path was specified and if the volume was reused,
    # then make sure the volume is mounted on the proper junction-path
    if ($duplicated and $args{'junction-path'}) {
        # an exception will be thrown only if we try to mount the volume on a junction path
        # that is not the same as the one we have now

        # load up some extra parameters for the mount command
        my %mount_args = ();
        $mount_args{apiset_should} = $args{apiset_should} if (exists $args{apiset_should});

        my $mount_attempts = 2;
        my $already_mounted_exception = undef;
        while ($mount_attempts > 0) {
            my $break = 0;
            try {
                # mount the volume
                NACL::C::Volume->mount('command_interface' => $args{'command_interface'},
                                       'junction-path'     => $args{'junction-path'},
                                       'vserver'           => $args{'vserver'},
                                       'volume'            => $args{'volume'},
                                       %mount_args);
                $break = 1;
            } catch NACL::C::Exceptions::Volume::AlreadyMounted with {
                $already_mounted_exception = shift;
                # the volume is already mounted on a different junction path,
                # lets unmount the volume and try again
                NACL::C::Volume->unmount('command_interface' => $args{'command_interface'},
                                         'vserver'           => $args{'vserver'},
                                         'volume'            => $args{'volume'},
                                         %mount_args);
                $mount_attempts--;
            };
            # break out of the loop on success of the mount
            last if ($break);
        }

        # if we $mount_attempts is 0, then we failed to mount,
        if ($mount_attempts <= 0) {
            if (defined($already_mounted_exception)) {
                # rethrow the Already Mounted exception that we may have received
                $Log->exit();
                $already_mounted_exception->throw();
            } else {
                # code should never really get here, if it does then
                # just throw a generic error
                $Log->exit();
                NATE::BaseException->throw("Failed to mount volume " . $args{'volume'} . " for vserver " . $args{'vserver'});
            }
        }
    }

    # Modify the infinite volume with new parameters if it already exists
    if ($duplicated and $is_infinite_vol) {
        if (exists $args{'is-managed-by-service'}){
            $is_managed_by_service = $args{'is-managed-by-service'};
        }

        try{
            NACL::C::Volume->modify('command_interface' => $args{'command_interface'},
                                    'vserver'           => $args{'vserver'},
                                    'volume'            => $args{'volume'},
                                    'is-managed-by-service' => $is_managed_by_service);
        } catch NACL::C::Exceptions::Volume::AlreadyModified with {
            $Log->comment("The volume $args{'volume'} already has its attributes in the desired state.");
        } otherwise {
             my $e = shift;
             $Log->error("Error while attempting to modify existing infinite volume $args{'volume'}:" . $e->text());
        }
    }

    if ($enable_flexcache) {
        try {
            NACL::C::VolumeFlexcache->create('command_interface' => $args{'command_interface'},
                                             'vserver'           => $args{'vserver'},
                                             'origin-volume'     => $args{'volume'});
            $Log->comment("Using newly Created Volume Flexcache: " . $args{volume});
        } otherwise {
            my $e = shift;
            if ($e->text() =~ /Error: command failed: Duplicate volume name $args{'volume'}."/i) {
                $Log->comment("Continuing setup, Volume Flexcache already exists for volume:$args{'volume'}.");
            } else {
                # otherwise rethrow the exception
                $Log->exit();
                $e->throw();
            }
        };
    }

    # If the user defined the files parameter then modify the volume.
    if(defined($files) and length($files)) {
        NACL::C::Volume->modify('command_interface' => $args{'command_interface'},
                                'vserver'           => $args{'vserver'},
                                'volume'            => $args{'volume'},
                                'files'             => $files);
    }

    $Log->exit();
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>$duplicated) : TRUE;

} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a volume.
# Details:    Makes the NACL::C::Volume->unmount()
#                       NACL::C::Volume->offline()
#                       NACL::C::Volume->delete() calls. Also works around
#             some common non-fatal errors that can be return from volume
#             deletion and then has a retry mechanism to retry volume deletion.
# TODO:       Need to update so that it uses NACL::STask::Volume->purge()
# Inputs:     NACL::STask::Volume->delete arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my $pkg   = shift;
    my %args = @_;

    # Locals
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    my $enable_flexcache =
        exists($args{enable_flexcache}) ? delete($args{enable_flexcache}) : '0';

    if ($enable_flexcache) {
        try {
            NACL::C::VolumeFlexcache->delete('command_interface' => $args{'command_interface'},
                                             'vserver'           => $args{'vserver'},
                                             'origin-volume'     => $args{'volume'});
            $Log->comment("Deleted Volume Flexcache: " . $args{'volume'});
        } otherwise {
            my $e = shift;
            if ($e->text() =~ /Error: Vserver $args{'vserver'} does not exist."/i) {
                $Log->comment("Continuing cleanup, Volume Flexcache was already deleted for volume:$args{'volume'}.");
            } else {
                # otherwise rethrow the exception
                $Log->exit();
                $e->throw();
            }
        };
    }

    try {
        NACL::C::Volume->unmount(%args);
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 3) {
            sleep 30 * ($retry++);
            $pkg->_nacl_delete_call(%args,
                                    retry => $retry,
                                );
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if ($exception->text() =~ /Authorization failed/ && $retry < 3) {
            sleep 30 * ($retry++);
            $pkg->_nacl_delete_call(%args,
                                    retry => $retry,
                                );
        } elsif ($exception->text() !~ /entry doesn't exist/
            and $exception->text()
            !~ /Cannot unmount Vserver namespace root volume/)
        {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    };    # end of Unmount try/catch

    try {
        NACL::C::Volume->offline(%args);
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 3) {
            $pkg->_nacl_delete_call(%args,
                                    retry => ++$retry,
                                );
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
        if ($retry < 3) {
            $pkg->_nacl_delete_call(%args,
                                    retry => ++$retry,
                                );
        } else {
            my $exception = shift;
            if ($exception->text() !~ /entry doesn't exist/) {
                $Log->exit();

                # Rethrowing caught exception
                $exception->throw();
            }
        }
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if ($exception->text() =~ /Authorization failed/ && $retry < 3) {
            sleep 30 * ($retry++);
            $pkg->_nacl_delete_call(%args,
                                    retry => $retry,
                                );
        } elsif ($exception->text() !~ /entry doesn't exist|already offline/) {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    };    # end of offline try/catch

    try {
        NACL::C::Volume->delete(%args);
        $Log->comment("Volume:" . $args{volume} . " was deleted");
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 3) {
            $pkg->_nacl_delete_call(%args,
                                    retry => ++$retry,
                                );
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if ($exception->text() =~ /Authorization failed/ && $retry < 3) {
            sleep 30 * ($retry++);
            $pkg->_nacl_delete_call(%args,
                                    retry => $retry,
                                );
        } elsif ($exception->text() =~ /entry doesn't exist/) {
            $Log->comment("Volume:" . $args{volume} . " was already deleted");
        } else {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    };    # end of delete try/catch

    $Log->exit();
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Volume->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Volume->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;
    # Assign values
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # Locals
    my %wait_args = (
            'command_interface' => $args{'command_interface'},
            'volume'            => $args{'volume'},
            'vserver'           => $args{'vserver'},
            'method-timeout'    => $args{'method-timeout'}
        );

    #Special Cases
    $wait_args{apiset_should} = $args{apiset_should}
            if exists $args{apiset_should};

    if ($Log->may_debug()) {
        $Log->debug("Waiting args are being created: " . Dumper(\%wait_args));
    }
    #Call
    NACL::STask::Volume->wait_for_creation(%wait_args);

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
