# $Id: //depot/prod/test/Rsierranevada/lib/EnGenX/Controller/Volume/Mount.pm# $

#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Mount Volume Controller Layer Module
## @author dl-san-automation
## @status shared
## @pod here

=head1 NAME

EnGenX::Controller::Volume::Mount

=head1 DESCRIPTION

EnGenX::Controller::Volume::Mount is a derived class of
EnGenX::Controller::Task.  It is used to Mount and
Unmount ONTAP Volumes using Deep Junction. All methods
# are static.

=cut

package EnGenX::Controller::Volume::Mount;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::CS::StorageAggregate;
use NACL::C::Volume;
use NACL::C::VolumeFlexcache;
use NACL::CS::Volume;
use NACL::STask::Volume;
use NACL::C::Job;
use EnGenX::SubProcess;
use EnGenX::EnGenXUtil;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::C::Exceptions::Volume::AlreadyMounted;

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args
);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name   => 'volume',
    _reuse => 'if_exists',
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods.
%DTD_To_NACL_Ignore = (
    efficiency             => '',
    parallel               => ''
);

# Arguments that can be passed into the NACL delete
# command.
%Delete_Args = (
    command_interface       => '',
    apiset_must             => '',
    apiset_should           => '',
    volumes                 => '',
    vserver                 => '',
);

=head1 METHODS

=head2 create

(Static method) Named create to follow convention, this method actually,
requests mounting of one or multiple volumes serially or in parallel
and return whether the volume(s) was created successfully. This method
passes all user arguments through to EnGenX::Controller::Task and adds
specific arguments used by the EnGenX framework.

  EnGenX::Controller::Volume::Mount->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            junction-path     => $junction_path,
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume::Mount->create(
            command_interfac  => $Node,
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume::Mount->create(
            command_interface => $Node,
            resource_list      => @Vols,
            );

  EnGenX::Controller::Volume::Mount->create(
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume::Mount->create(
            resource_list      => @Vols,
            );

=over

=item Options

Either the resource_list is required or the NACL Volume Mount Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the volumes in the resource list will be mounted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Volume Mount Arguments should
NOT be passed in.

=item C<< NACL Volume Mount Arguments >>

(Optional)
These are arguments that can be passed into NACL::C::Volume->Mount.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted to 3600.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the mounting of the Volume(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    $args{'Name'} = "volume";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);
    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Named delete to follow convention, this method actually
unmounts the mounted Volumes with deep-junction paths enabled.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Volume::Mount->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume::Mount->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            count             => 5
            );

  EnGenX::Controller::Volume::Mount->delete(
            command_interface => $Node,
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume::Mount->delete(
            command_interface => $Node,
            resource_list      => @Vols,
            );

  EnGenX::Controller::Volume::Mount->delete(
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume::Mount->delete(
            resource_list      => @Vols,
            );

=over

=item Options

Either the resource_list is required or the NACL Volume Unmount Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the volumes in the resource list will be unmounted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Volume Unmount Arguments should
NOT be passed in.

=item C<< NACL Volume Unmount Arguments >>

(Required)
These are arguments that can be passed into NACL::STask::Volume->Unmount.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted to 3600.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the unmounting of the Volume(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    $args{'Name'} = "volume";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for mounting of a volume.
# Details:    Makes the NACL::STask::Volume->create() call and optionally could
#             make the NACL::STask::Volume->wait_for_creation. Also works
#             around some common non-fatal errors that can be return from
#             volume mount and then has a retry mechanism to retry volume
#             mounting.
# TODO:       Need to update to NACL::STask::Volume->create()
# Inputs:     NACL::C::Volume->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the mount was successful or a hash with success
#             and whether the volume was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    # This creates a NACL command_interface if one is not passed in.
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    my $volumes_ref = delete($args{volumes});

    foreach my $volume (@{$volumes_ref}) {
        # parse out the use default junction path flag
        my $use_default_junction_path = delete($volume->{use_default_junction_path});
        # skip volumes that are mounting on the default junction path
        next if ($use_default_junction_path);

        # store the command_interface passed in through $args
        my $command_interface = $args{command_interface};

        # if a junction path was specified and if the volume was reused,
        # then make sure the volume is mounted on the proper junction-path
        if ($volume->{'junction-path'}) {
            # an exception will be thrown only if we try to mount the volume on a junction path
            # that is not the same as the one we have now

            # load up some extra parameters for the mount command
            my %mount_args = ();
            $mount_args{apiset_should} = $volume->{apiset_should} if (exists $volume->{apiset_should});

            my $mount_attempts = 2;
            my $already_mounted_exception = undef;

            my $break = 0;
            while ($mount_attempts > 0) {
                try {
                    # mount the volume
                    NACL::C::Volume->mount('command_interface' => $command_interface,
                                           'junction-path'     => $volume->{'junction-path'},
                                           'vserver'           => $volume->{'vserver'},
                                           'volume'            => $volume->{'name'},
                                           %mount_args);
                    $break = 1;
                } catch NACL::C::Exceptions::Volume::AlreadyMounted with {
                    $already_mounted_exception = shift;
                    # the volume is already mounted on a different junction path,
                    # lets unmount the volume and try again
                    NACL::C::Volume->unmount('command_interface' => $volume->{'command_interface'},
                                             'vserver'           => $volume->{'vserver'},
                                             'volume'            => $volume->{'name'},
                                             %mount_args);
                    $mount_attempts--;
                } otherwise {
                    my $e = shift;
                    # otherwise rethrow the exception
                    $Log->exit();
                    $e->throw();
                };

                # break out of the loop on success of the mount
                last if ($break);
            }

            # if we $mount_attempts is 0, then we failed to mount,
            if ($mount_attempts <= 0) {
                if (defined($already_mounted_exception)) {
                    # rethrow the Already Mounted exception that we may have received
                    $Log->exit();
                    $already_mounted_exception->throw();
                } else {
                    # code should never really get here, if it does then
                    # just throw a generic error
                    $Log->exit();
                    NATE::BaseException->throw("Failed to mount volume " . $volume->{'volume'} . " for vserver " . $volume->{'vserver'});
                }
            }
        }
    }
    $Log->exit();
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE) : TRUE;

} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for unmounting of a volume.
# Details:    Makes the NACL::C::Volume->unmount() calls. Also works around
#             some common non-fatal errors that can be return from volume
#             unmount and then has a retry mechanism to retry volume unmount.
# TODO:       Need to update so that it uses NACL::STask::Volume->purge()
# Inputs:     NACL::STask::Volume->unmount arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the unmount was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();

    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    # This creates a NACL command_interface if one is not passed in.
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    my $volumes_ref = delete($args{volumes});

    foreach my $volume (@{$volumes_ref}) {
        # parse out the use default junction path flag
        my $use_default_junction_path = delete($volume->{use_default_junction_path});
        # skip volumes that are mounted on the default junction path
        next if ($use_default_junction_path);

        # store the command_interface passed in through $args
        my $command_interface = $args{command_interface};

        # if a junction path was specified and if the volume was reused,
        # then make sure the volume is unmounted
        if ($volume->{'junction-path'}) {

            # load up some extra parameters for the unmount command
            my %unmount_args = ();
            $unmount_args{apiset_should} = $volume->{apiset_should} if (exists $volume->{apiset_should});

            my $unmount_attempts = 2;

            my $break = 0;
            while ($unmount_attempts > 0) {

                try {
                    # unmount the volume
                    NACL::C::Volume->unmount('command_interface' => $command_interface,
                                           'vserver'           => $volume->{'vserver'},
                                           'volume'            => $volume->{'name'},
                                           %unmount_args);
                    $break = 1;
                } catch NACL::APISet::Exceptions::ConnectionFailedException with {
                    if ($retry < 3) {
                        sleep 30 * ($retry++);
                        $pkg->_nacl_delete_call(%args,
                                                retry => $retry,
                                            );
                    } else {
                        my $exception = shift;
                        $Log->exit();

                        # Rethrowing caught exception
                        $exception->throw();
                    }
                } catch NACL::C::Exceptions::Volume::DoesNotExist with {
                    $Log->comment("Volume " . $volume->{'name'} . " does not exist. " .
                                  "Continuing cleanup.");
                    $break = 1;
                } otherwise {
                    my $e = shift;
                    $Log->exit();
                    $e->throw();
                };

                # break out of the loop on success of the unmount
                last if ($break);
            }
        }
    }
    $Log->exit();
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE) : TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Volume->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Volume->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_wait_call {
    return;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
