# $Id: //depot/prod/test/Rsierranevada/lib/EnGenX/Controller/Vserver.pm#3 $

#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Vserver Controller Layer Module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

EnGenX::Controller::Vserve

=head1 DESCRIPTION

EnGenX::Controller::Vserver is a derived class of
EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP Vservers. All methods are static.

=cut

package EnGenX::Controller::Vserver;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use EnGenX::SubProcess;
use EnGenX::EnGenXUtil;
use NACL::C::Vserver;
use NACL::C::Job;
use EnGenX::Controller::Volume;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name   => 'vserver',
    aggr   => 'aggregate',
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    vs_type                      => '',
    fcp_enabled                  => '',
    iscsi_enabled                => '',
    cifs_enabled                 => '',
    nfs_enabled                  => '',
    dns_enabled                  => '',
    parallel                     => '',
    use_default_qos_policy_group => '',
    use_deep_junction            => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    apiset_must       => '',
    apiset_should     => '',
    vserver           => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple vservers serially
or in parallel and return whether the vserver(s) was created successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Vserver->create(
            command_interface => $Node,
            vserver => $Vserver_name,
            rootvolume => $Vserver_name . '_rootvol',
            aggregate => $Aggr_name,
            'ns-switch'=> ['file'],
            'nm-switch'=> ['file'],
            'rootvolume-security-style' => 'mixed',
            );

  EnGenX::Controller::Vserver->create(
            command_interface => $Node,
            vserver => $Vserver_name,
            rootvolume => $Vserver_name . '_rootvol',
            aggregate => $Aggr_name,
            'ns-switch'=> ['file'],
            'nm-switch'=> ['file'],
            'rootvolume-security-style' => 'mixed',
            count             => 5
            );

  EnGenX::Controller::Vserver->create(
            command_interfac  => $Node,
            resource_list      => @Vservers,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Vserver->create(
            command_interface => $Node,
            resource_list      => @Vservers,
            );

  EnGenX::Controller::Vserver->create(
            resource_list      => @Vservers,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Vserver->create(
            resource_list      => @Vservers,
            );

=over

=item Options

Either the resource_list is required or the NACL Vserver Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the vservers in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Vserver Creation Arguments should
NOT be passed in.

=item C<< NACL Vserver Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::C::Vserver->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< count => $vserver_count >>

(Optional)
This arguments is used to create multiple vservers from the same NACL Vserver
Creation Arguments.  To create the names for the vservers, the vserver name
passed in is appended with the index that it was create at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Vserver(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "vserver";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple vservers serially
or in parallel and return whether the vserver(s) was deleted successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Vserver->delete(
            command_interface => $Node,
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Vserver->delete(
            command_interface => $Node,
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            count             => 5
            );

  EnGenX::Controller::Vserver->delete(
            command_interface => $Node,
            resource_list      => @Vservers,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Vserver->delete(
            command_interface => $Node,
            resource_list      => @Vservers,
            );

  EnGenX::Controller::Vserver->delete(
            resource_list      => @Vservers,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Vserver->delete(
            resource_list      => @Vservers,
            );

=over

=item Options

Either the resource_list is required or the NACL Vserver Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the vservers in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Vserver Deletion Arguments should
NOT be passed in.

=item C<< NACL Vserver Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::C::Vserver->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< count => $vserver_count >>

(Optional)
This arguments is used to delete multiple vservers from the same NACL Vserver
Deletion Arguments.  To delete the names for the vservers, the vserver name
passed in is appended with the index that it was deleted at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Vserver(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "vserver";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a vserver.
# Details:    Makes the NACL::C::Vserver->create() call. Also works
#             around some common non-fatal errors that can be return from
#             vserver creation and then has a retry mechanism to retry vserver
#             creation.
# TODO:       Need to update so that it uses NACL::STask::Vserver->create()
# Inputs:     NACL::C::Vserver->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the vserver was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $duplicated = FALSE;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    my $aggr_list = delete $args{'aggr-list'};
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    #Default the rootvolume
    $args{rootvolume} = $args{vserver} . "_rootvol"
        if (not exists $args{rootvolume});

    #Default the ns-switch
    $args{'ns-switch'} = ['file'] if (not exists $args{'ns-switch'});

    #Default the rootvolume-security-style
    $args{'rootvolume-security-style'} = 'mixed'
        if (not exists $args{'rootvolume-security-style'});
    try {
        NACL::C::Vserver->create(%args);
        $Log->comment("Using newly created Vserver:" . $args{vserver});
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 3) {
            $pkg->_nacl_create_call(%args,
                                    retry => ++$retry,
                                    if_exists => $if_exists
                                );
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if (   $if_exists eq 'reuse'
            && $exception->text()
            =~ /The Vserver name is already used by another Vserver/)
        {
            $duplicated = TRUE;
            $Log->comment("Reusing already created Vserver:" . $args{vserver});
        } else {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    };

    if(defined $aggr_list) {
        $Log->comment("Modifying the vserver $args{vserver} with the specified 'Aggr-list'");
        NACL::C::Vserver->modify(command_interface => $args{command_interface},
                                 vserver           => $args{vserver},
                                 'aggr-list'       => $aggr_list,
                                );
    }

    $Log->exit();
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>$duplicated) : TRUE;

} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a vserver.
# Details:    Makes the NACL::C::Vserver->delete() calls. Also works around
#             some common non-fatal errors that can be return from vserver
#             deletion and then has a retry mechanism to retry vserver deletion.
# TODO:       Need to update so that it uses NACL::STask::Vserver->purge()
# Inputs:     NACL::C::Vserver->delete arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my $pkg   = shift;
    my %args = @_;

    # Locals
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

     my $vserver = NACL::C::Vserver->find(
                        command_interface => $args{command_interface},
                        allow_empty       => 'true',
                        filter            => {vserver => $args{vserver}}
                    );

    if ($vserver and !($vserver->state->rootvolume() eq '-')) {
        EnGenX::Controller::Volume->delete(
                        %args,
                        volume => $vserver->state->rootvolume(),
                    );
    }

    try {
        NACL::C::Vserver->delete(%args,);
        $Log->comment("Vserver:" . $args{vserver} . " was deleted");
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 3) {
            $pkg->_nacl_delete_call(%args,
                                    retry => ++$retry,
                                );
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if ($exception->text() =~ /entry doesn't exist/)    # CMode
        {
            $Log->comment(
                "Vserver:" . $args{vserver} . " was already deleted");
        } else {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    };

    $Log->exit();
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Vserver->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Vserver->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
