# $Id: //depot/prod/test/main/lib/EnGenX/Controller/Security_Trace_Filter.pm#2 $

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Security_Trace_Filter Controller Layer Module
## @author dl-san-automation
## @status public
## @pod here

=head1 NAME

EnGenX::Controller::Security_Trace_Filter

=head1 DESCRIPTION

EnGenX::Controller::Security_Trace_Filter is a derived class of
EnGenX::Controller::Task.  It is used to Set Attributes
of ONTAP Luns. All methods are static.

=cut

package EnGenX::Controller::Vserver::Security_Trace_Filter;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::C::VserverSecurityTraceFilter;

# Exceptions
use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::ResponseException;
use NACL::APISet::Exceptions::ConnectionFailedException;

#Inherit base class
use base qw( EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    parallel => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    apiset_must       => '',
    apiset_should     => '',
    vserver           => '',
    'method-timeout'  => '',
    index_value => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple security trace filters serially
or in parallel and return whether the security trace filter(s) was created successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Security_Trace_Filter->create(
            command_interface   => $Node,
            'index'             => $index,
            'path'              => $path,
            'windows-name'      => $windows_name,
            'unix-name'         => $unix_name,
            'trace-allow'       => $trace_allow,
            'enabled'           => $enabled,
            'time-enabled'      => $time_enabled,
            'method-timeout'    => 3600,
            );

  EnGenX::Controller::Security_Trace_Filter->create(
            command_interfac  => $Node,
            resource_list     => @Security_Trace_Filters,
            'method-timeout'  => 3600,
            );


  EnGenX::Controller::Security_Trace_Filter->create(
            resource_list     => @Security_Trace_Filters,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Security_Trace_Filter->create(
            resource_list      => @Security_Trace_Filters,
            );

=over

=item Options

Either the resource_list is required or the NACL Security_Trace_Filter Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the security trace filters in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Lun Creation Arguments should
NOT be passed in. (This is an array of EnGenX::Resource() objects not
an EnGenX::Resource_List() currently)

=item C<< NACL Security_Trace_Filter Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::C::Security_Trace_Filter->set().
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< index => $index >>

(Optional)
This arguments is to specify the index for the security trace filter.

=item C<< path => $path >>

(Optional)
This arguments is used to specify the path for the security trace filter.

=item C<< windows-name => $windows_name >>

(Optional)
This arguments is used to specify the windows user name

=item C<< unix-name => $unix_name >>

(Optional)
This arguments is used to specify the unix user.

=item C<< trace-allow => $trace_allow >>

(Optional)
This arguments is used to specify id there should be trace allow events.

=item C<< enabled => $enabled >>

(Optional)
This arguments is used to specify if the filter is enabled.

=item C<< time-enabled => $time_enabled >>

(Optional)
This arguments is used to specify the mintues the filter is eanbled.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Lun(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "security_trace_filter";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple luns serially
or in parallel and return whether the security trace filter(s) was deleted successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Lun->delete(
            command_interface => $Node,
            index            =>  $index,
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Lun->delete(
            command_interface => $Node,
            resource_list      => @SecurityTraceFilters,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Lun->delete(
            command_interface => $Node,
            resource_list      => @SecurityTraceFilters,
            );

  EnGenX::Controller::Lun->delete(
            resource_list      => @SecurityTraceFilters,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Lun->delete(
            resource_list      => @SecurityTraceFilters,
            );

=over

=item Options

Either the resource_list is required or the NACL Lun Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the luns in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Vserver Security Trace Filter
Deletion Arguments should NOT be passed in.

=item C<< NACL Lun Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::C::VserverSecurityTraceFilter->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< index => $index >>

(Optional)
This arguments is used to delete multiple security trace fiters from the same NACL Vserver
Security Trace Filter Deletion Arguments.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Lun(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "security_trace_filter";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a security trace filters.
# Details:    Makes the NACL::C::Security_Trace_Filter->set() call. Also works
#             around some common non-fatal errors that can be return from
#             lun creation and then has a retry mechanism to retry lun
#             creation.
# Inputs:     NACL::C::Security_Trace_Filter->set arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the the attribute was set.
#             of the Security_Trace_Filter  was successful or a hash with success
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $wait = exists($args{wait}) ? delete($args{wait}) : FALSE;
    my $if_exists = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    my $duplicated = FALSE;
    my $index_value = undef;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    if (exists $args{index_value}) {
        $index_value = delete $args{index_value};
        $args{index} = $index_value;
    }

    # Create Lun Attribute
    try{
        NACL::C::VserverSecurityTraceFilter->create(%args);

       $Log->comment( "Security Trace filter succesfully created for vserver " . $args{vserver} . " with index of: "  . $args{index} );
    }  catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 3) {
            sleep 30 * ($retry++);
            $pkg->_nacl_create_call(%args,
                                    retry     => $retry,
                                    if_exists => $if_exists,
                                    wait      => $wait);
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;

        if (  ($exception->text() =~ /duplicate entry/
            or $exception->text() =~ /The pattern (\S+) already exists/)
            and  $if_exists eq 'reuse') {
            $duplicated = TRUE;

            try{
                # Make sure policy is set for vserver if the policy already exists
                NACL::C::VserverSecurityTraceFilter->modify(%args);
                $Log->comment("Updating and Reusing already created Security Trace Filter with index of: " . $args{index});
            } catch NACL::APISet::Exceptions::ConnectionFailedException with {
                if ($retry > 0) {
                    $pkg->_nacl_create_call(%args,
                                            retry => --$retry,
                                            if_exists => $if_exists
                                            );
                } else {
                    my $exception = shift;
                    $Log->exit();

                    # Rethrowing caught exception
                    $exception->throw();
                }
            } catch NACL::APISet::Exceptions::ResponseException with {
                my $exception = shift;
                $Log->exit();

                # Rethrowing caught exception
                $exception->throw();
            };
        } else {
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }

    };

    # Rest index_value argument for deletion later in the _nacl_delete_call subroutine.
    $args{index_value} = $index_value;

    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE) : TRUE;
} ## end sub _nacl_create_call

#========================================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a Vserver Security Trace Filter.
# Details:    Makes the NACL::C::VserverSecurityTraceFilter->delete() calls.
# Inputs:     NACL::C::VserverSecurityTraceFilter->delete arguments
#             retry => integer, is the current retry the method is on.
#             This is used to count how many times the _nacl_delete_call
#             has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $if_exists;
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    if( exists $args{index_value}) {
        $args{index} = delete $args{index_value};
    }

    try {
        NACL::C::VserverSecurityTraceFilter->delete(%args);
        $Log->comment("Vserver Security Trace Filter for vserver: " . $args{vserver} . " was deleted");
    } catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 3) {
            sleep 30 * ($retry++);
            # recursively calling _nacl_delete_call to retry delete
            $pkg->_nacl_delete_call(%args,
                                    retry =>$retry,
                                );
        } else {
            my $exception = shift;
            $Log->exit();

            # Rethrowing caught exception
            $exception->throw();
        }
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;

        if ($exception->text() =~ /no entries matching your query/
            or $exception->text() =~ /entry doesn't exist/) {
            $Log->comment("Vserver Security Trace Filter for vserver: " . $args{vserver} . " was already deleted");
        } else {
            # Rethrowing caught exception
            $Log->exit();
            $exception->throw();
        }
    };

    $Log->exit();
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Security_Trace_Filter->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Security_Trace_Filter->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
