# $Id: //depot/prod/test/Rsierranevada/lib/EnGenX/Resource.pm#2 $
#
#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide API's to manage data in  EnGenX Resource objects
## @author dl-san-automation@netapp.com
## @status private
## @pod here

=head1 NAME

Resource

=head1 DESCRIPTION

Resource provides API's to manage data in EnGenX Resource Objects.
API's to create objects, modify data, get data and destory objects
are available.

=cut

package EnGenX::Resource;

# Compiler directives.
use strict;
use warnings;

# NATE and Standard Module import
use Tharn qw($Log);
use Data::Dumper;
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF HASHREF);
use Clone qw(clone);

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

use vars qw(@ISA);
@ISA = qw();

###############################################################################
#                               Public Methods
###############################################################################

=head1 METHODS

=head2 new

Create a new Resource object.

    my $resource = new EnGenX::Resource(\%data);

=over

=item Options

=over

=item C<< $data_hash_ref >>

(Required) A hash ref of data to be stored in the Resource

=back

=item Returns

The Resource object created

=item Exceptions

NATE::Exceptions::Argument - thrown if method arguments fail validation

=cut

########################################################################
sub new {
    my ($proto, $args) = @_;

    # Make sure the user sent us a hash
    if (!defined($args) or ref($args) ne "HASH") {
        NATE::Exceptions::Argument->throw(
            "The new method requires a HASH ref");
    }

    # Clone the provided hash reference to create a unique hash that the
    # object will have unique control
    my $self = clone($args);
    my $class = ref($proto) || $proto;
    bless($self, $class);

    return $self;
} ## end sub new

########################################################################

=head2 field_defined

Check to see if a given field is defined in the resource data

    if ($resource->field_defined($key) == FALSE) {
        $match = FALSE;
    }

=over

=item Options

=over

=item C<< field >>

(Required) A string that tells us which data field to check

=back

=item Returns

TRUE if the field is defined, FALSE otherwise

=item Exceptions

NATE::Exceptions::Argument - thrown if method arguments fail validation

=cut

########################################################################
sub field_defined {
    $Log->enter();
    my ($self, $field) = @_;

    # Makes sure they gave us a field
    if (!defined($field)) {
        $Log->exit();
        NATE::Exceptions::Argument->throw(
                  "The field_defined method requires an argument tell it"
                . " which field to check");
    }
    $Log->exit();
    return TRUE if (defined $self->{$field});
    return FALSE;
} ## end sub field_defined

########################################################################

=head2 get_field

Return the value of a given field

    my $volume_name = $resource->get_field("name");

=over

=item Options

=over

=item C<< field >>

(Required) A string that tells us which data field to return

=back

=item Returns

The value of the field given

=item Exceptions

NATE::Exceptions::Argument - thrown if method arguments fail validation

NATE::BaseException - thrown if field is not defined (has no value)

=cut

########################################################################
sub get_field {
    $Log->enter();
    my ($self, $field) = @_;

    # Makes sure they gave us a field
    if (!defined($field)) {
        $Log->exit();
        NATE::Exceptions::Argument->throw(
                  "The get_field method requires an argument tell it"
                . " which field to retreive");
    }

    # If the field is defined, return the value.  If not throw an exception
    if (defined $self->{$field}) {
        $Log->exit();
        return $self->{$field};
    }

    $Log->exit();
    NATE::BaseException->throw(
              "get_field() was called with a field that is not defined."
            . " Use EnGenX::Resource::field_defined() to check on a "
            . "field that you are not sure about.");
} ## end sub get_field

########################################################################

=head2 set_field

Set the value of a given field

    $resource->set_field("name" => "elvis14");

=over

=item Options

=over

=item C<< field >>

(Required) A string that tells us which data field to set

=back

=item Returns

Nothing

=item Exceptions

NATE::Exceptions::Argument - thrown if method arguments fail validation

=cut

########################################################################
sub set_field {
    $Log->enter();
    my ($self, $field, $value) = @_;

    # Makes sure they gave us a field and a value
    if (!defined($field) or !defined($value)) {
        $Log->exit();
        NATE::Exceptions::Argument->throw(
                  "The set_field method requires an argument tell it"
                . " which field to set and a value: \$r->('count', '84')");
    }

    $self->{$field} = $value;

    $Log->exit();
    return;
} ## end sub set_field

# TODO:
# - methods: delete(field)
# - Implemented the hash clone for the input arguments in NEW. We must be
#   vigilant that the clonning of the hash does not impact the performance or
#   memory footprint.

1;
