#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package EnGenX::Services::VserverClientRelationship;

BEGIN {
    require Exporter;
    our @ISA = qw(Exporter);
    our @EXPORT_OK = qw();
}

use parent qw(EnGenX::Services::Base);

use strict;
use warnings;

use Carp qw(confess);
use Data::Dumper;

use NATE::Messenger qw(messenger_global);
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);

use ParamsMgr;
use San::SanUtil;
use EnGenX::Util;

my $Log = log_global();

sub _install_apis {
    my ($self) = @_;

    # set up listeners for requests
    messenger_global()->listener_add(listener => {
        filter   => sub { shift->type() eq 'get_clients_associated_to_vserver' },
        callback => sub {
            my ($msg, $listener, $msnger) = @_;
            my $request_id = delete($msg->message()->{request_id});
            my $response;
            eval {
                $response = $self->_svc_get_clients_associated_to_vserver($listener, %{$msg->message()});
                $msnger->reply(msg_obj => $msg,
                               type => "response_${request_id}",
                               message => $response);
            };
            if ($@) {
                $msnger->reply(msg_obj => $msg,
                               type => "error_${request_id}",
                               message => $@);
            }
        },
        data     => $self->{_data}
    });
    messenger_global()->listener_add(listener => {
        filter   => sub { shift->type() eq 'get_vservers_associated_to_client' },
        callback => sub {
            my ($msg, $listener, $msnger) = @_;
            my $request_id = delete($msg->message()->{request_id});
            my $response;
            eval {
                $response = $self->_svc_get_vservers_associated_to_client($listener, %{$msg->message()});
                $msnger->reply(msg_obj => $msg,
                               type => "response_${request_id}",
                               message => $response);
            };
            if ($@) {
                $msnger->reply(msg_obj => $msg,
                               type => "error_${request_id}",
                               message => $@);
            }
        },
        data     => $self->{_data}
    });
    messenger_global()->listener_add(listener => {
        filter   => sub { shift->type() eq 'get_luns_associated_to_client' },
        callback => sub {
            my ($msg, $listener, $msnger) = @_;
            my $request_id = delete($msg->message()->{request_id});
            my $response;
            eval {
                $response = $self->_svc_get_luns_associated_to_client($listener, %{$msg->message()});
                $msnger->reply(msg_obj => $msg,
                               type => "response_${request_id}",
                               message => $response);
            };
            if ($@) {
                $msnger->reply(msg_obj => $msg,
                               type => "error_${request_id}",
                               message => $@);
            }
        },
        data     => $self->{_data}
    });
    messenger_global()->listener_add(listener => {
        filter   => sub { shift->type() eq 'get_luns_and_vservers_associated_to_client' },
        callback => sub {
            my ($msg, $listener, $msnger) = @_;
            my $request_id = delete($msg->message()->{request_id});
            my $response;
            eval {
                $response = $self->_svc_get_luns_and_vservers_associated_to_client($listener, %{$msg->message()});
                $msnger->reply(msg_obj => $msg,
                               type => "response_${request_id}",
                               message => $response);
            };
            if ($@) {
                $msnger->reply(msg_obj => $msg,
                               type => "error_${request_id}",
                               message => $@);
            }
        },
        data     => $self->{_data}
    });
    messenger_global()->listener_add(listener => {
        filter   => sub { shift->type() eq 'get_client_relationship_hash' },
        callback => sub {
            my ($msg, $listener, $msnger) = @_;
            my $request_id = delete($msg->message()->{request_id});
            my $response;
            eval {
                $response = $self->_svc_get_client_relationship_hash($listener, %{$msg->message()});
                $msnger->reply(msg_obj => $msg,
                               type => "response_${request_id}",
                               message => $response);
            };
            if ($@) {
                $msnger->reply(msg_obj => $msg,
                               type => "error_${request_id}",
                               message => $@);
            }
        },
        data     => $self->{_data}
    });
    messenger_global()->listener_add(listener => {
        filter   => sub { shift->type() eq 'get_vserver_client_relationship_hash' },
        callback => sub {
            my ($msg, $listener, $msnger) = @_;
            my $request_id = delete($msg->message()->{request_id});
            my $response;
            eval {
                $response = $self->_svc_get_vserver_client_relationship_hash($listener, %{$msg->message()});
                $msnger->reply(msg_obj => $msg,
                               type => "response_${request_id}",
                               message => $response);
            };
            if ($@) {
                $msnger->reply(msg_obj => $msg,
                               type => "error_${request_id}",
                               message => $@);
            }
        },
        data     => $self->{_data}
    });

    return $self->SUPER::_install_apis();
}

sub _post_svc_ready {
    my ($self) = @_;

    $self->{_data} = {};

    # build the hashes after the service has initialized and is serving
    my ($cli_relationship_hash, $vs_relationship_hash, $lun_relationship_hash) = $self->_build_hashes();

    # store the hashes on the cache
    $self->{_data}->{client_relationship_hash} = $cli_relationship_hash;
    $self->{_data}->{vs_relationship_hash}     = $vs_relationship_hash;
    $self->{_data}->{lun_relationship_hash}    = $lun_relationship_hash;


    return $self->SUPER::_post_svc_ready();
}

sub _build_hashes {
    my ($self) = @_;

    my $engenx_obj = $self->{engenx_obj};

    my $build_lun_hash = 1;
    my $load_nas_data = 1;
    my $client_relationship_hash = {};
    my $vserver_relationship_hash = {};
    my $lun_relationship_hash = {};

    my $clusters = &ParamsMgr::search(-search_path => [],
                                      -key => "CLUSTERS",
                                      -key_filter => qr/^CLUSTER\d*$/,
                                      paramset => $engenx_obj->EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        confess &ParamsMgr::get_error();
        return ();
    }

    my @client_objs = ();
    try {
        (@client_objs) = $engenx_obj->get_objects(types => ["client", "hypervisor"]);
    } otherwise {
        # ignore the error here since these objects may not exist
    };
    if ($engenx_obj->error()) {
        # burt550544
        # clear the error here, its possible someone doesn't have clients/hypervisors
        $engenx_obj->_clear_err();
        # dont return here because its possible that code above us can break if we don't
        # build out the hash.
    }

    # iterate over all clusters
    foreach my $cluster (values %{$clusters}) {
        $cluster->{_INDEX} = "" if (not defined($cluster->{_INDEX}));
        my $vservers = &ParamsMgr::search(-search_path => ["CLUSTERS"],
                                          -key => "CLUSTER" . $cluster->{_INDEX},
                                          -key_filter => qr/^VSERVER\d*$/,
                                          paramset => $engenx_obj->EnGenX::Util::get_paramset());
        if (&ParamsMgr::error()) {
            confess &ParamsMgr::get_error();
            return ();
        }

        # iterate over all vservers
        foreach my $vserver (values %{$vservers}) {
            $vserver->{_INDEX} = "" if (not defined($vserver->{_INDEX}));
            $vserver->{COUNT} = 1 if (not defined($vserver->{COUNT}));

            # check if asked to build the lun hash since it may be considered expensive
            my $lun_hash = {};
            if ($build_lun_hash == $SanMgr::TRUE) {
                my $vols = &ParamsMgr::search(-search_path => [],
                                              -exact_key => 1,
                                              -key => "CLUSTERS.CLUSTER" . $cluster->{_INDEX} .
                                                      ".VSERVER" . $vserver->{_INDEX},
                                              -key_filter => qr/^VOL\d*$/,
                                              paramset => $engenx_obj->EnGenX::Util::get_paramset());
                if (&ParamsMgr::error()) {
                    confess &ParamsMgr::get_error();
                    return ();
                }

                # iterate over all the vols
                foreach my $vol (values %{$vols}) {
                    $vol->{_INDEX} = "" if (not defined($vol->{_INDEX}));
                    my $luns = &ParamsMgr::search(-search_path => [],
                                                  -exact_key => 1,
                                                  -key => "CLUSTERS.CLUSTER" . $cluster->{_INDEX} .
                                                          ".VSERVER" . $vserver->{_INDEX} .
                                                          ".VOL" . $vol->{_INDEX},
                                                  -key_filter => qr/^LUN\d*$/,
                                                  paramset => $engenx_obj->EnGenX::Util::get_paramset());
                    if (&ParamsMgr::error()) {
                        confess &ParamsMgr::get_error();
                        return ();
                    }

                    my $qtrees = &ParamsMgr::search(-search_path => [],
                                                    -exact_key => 1,
                                                    -key => "CLUSTERS.CLUSTER" . $cluster->{_INDEX} .
                                                            ".VSERVER" . $vserver->{_INDEX} .
                                                            ".VOL" . $vol->{_INDEX},
                                                    -key_filter => qr/^QTREE\d*$/,
                                                    paramset => $engenx_obj->EnGenX::Util::get_paramset());
                    if (&ParamsMgr::error()) {
                        confess &ParamsMgr::get_error();
                        return ();
                    }

                    # loop through this volume COUNT times
                    my $vol_count = defined($vol->{COUNT}) ? $vol->{COUNT} : 1;
                    for (my $i = 0; $i < $vol_count; $i++)  {
                        my $vol_name = $vol->{NAME};
                        if ($vol_count > 1) {
                            $vol_name .= "_${i}";
                        }

                        # build the process luns routine
                        my $process_luns = sub {
                            my ($process_lun_hash, $process_luns, $process_vol_name, $process_qtree_ds) = @_;

                            my $qtree_string = "";
                            if (defined($process_qtree_ds)) {
                                $qtree_string = ".QTREE" . $process_qtree_ds->{_INDEX};
                            }

                            # iterate over all the luns
                            foreach my $lun (values %{$process_luns}) {
                                # initialize the index field in the lun structure to empty if it doesnt exit
                                $lun->{_INDEX} = "" if (not defined($lun->{_INDEX}));

                                my $lun_map = &ParamsMgr::search(-search_path => [],
                                                                 -exact_key => 1,
                                                                 -key => "CLUSTERS.CLUSTER" . $cluster->{_INDEX} .
                                                                         ".VSERVER" . $vserver->{_INDEX} .
                                                                         ".VOL" . $vol->{_INDEX} . $qtree_string .
                                                                         ".LUN" . $lun->{_INDEX},
                                                                 -key_filter => qr/^LUN_MAP\d*$/,
                                                                 paramset => $engenx_obj->EnGenX::Util::get_paramset());
                                if (&ParamsMgr::error()) {
                                    confess &ParamsMgr::get_error();
                                    return ();
                                }

                                # add map_list entries into this hash
                                if (exists($lun->{MAP_LIST})) {
                                    my @map_list = split(/\s*,\s*/, $lun->{MAP_LIST});
                                    foreach my $map (@map_list) {
                                        $map =~ s/\s+//g;
                                        $lun_map->{$map}->{NAME} = $map;
                                    }
                                }

                                # This lun may have COUNT specified, prep the lun name
                                # and loop count times
                                my $lun_count = defined($lun->{COUNT}) ? $lun->{COUNT} : 1;
                                for (my $j = 0; $j < $lun_count; $j++) {
                                    my $lun_name = $lun->{NAME};
                                    if ($lun_count > 1) {
                                        $lun_name .= "_${j}";
                                    }
                                    #  build the lun path
                                    my $lun_path = "/vol/${process_vol_name}/${lun_name}";

                                    # now we know of all the igroups that this lun is mapped to,
                                    # build a hash to make look up faster
                                    foreach my $map (keys %{$lun_map}) {
                                        my $igroup_name = "";
                                        if (exists($lun_map->{$map}->{NAME})) {
                                            # support for MAP_LIST formats
                                            $igroup_name = $lun_map->{$map}->{NAME};
                                        } elsif (exists($lun_map->{$map}->{IGROUP})) {
                                            # support for LUN_MAP formats
                                            $igroup_name = $lun_map->{$map}->{IGROUP};
                                        } else {
                                            confess "The lun map $map does not have a valid lun map format:" . Dumper($lun_map);
                                            return ();
                                        }
                                        push(@{$process_lun_hash->{$igroup_name}}, $lun_path);
                                    }
                                }
                            }
                        };

                        # process luns, this will handle all luns that are not in qtrees
                        $process_luns->($lun_hash, $luns, $vol_name);

                        # process qtrees and luns inside qtrees
                        # iterate over all the qtrees
                        foreach my $qtree (values %{$qtrees}) {
                            $qtree->{_INDEX} = "" if (not defined($qtree->{_INDEX}));
                            # find all the luns inside this qtree
                            my $qtree_luns = &ParamsMgr::search(-search_path => [],
                                                                -exact_key => 1,
                                                                -key => "CLUSTERS.CLUSTER" . $cluster->{_INDEX} .
                                                                        ".VSERVER" . $vserver->{_INDEX} .
                                                                        ".VOL" . $vol->{_INDEX} .
                                                                        ".QTREE" . $qtree->{_INDEX},
                                                                -key_filter => qr/^LUN\d*$/,
                                                                paramset => $engenx_obj->EnGenX::Util::get_paramset());
                            if (&ParamsMgr::error()) {
                                confess &ParamsMgr::get_error();
                                return ();
                            }

                            # handle the count attribute on this qtree, loop over it
                            my $qtree_count = defined($qtree->{COUNT}) ? $qtree->{COUNT} : 1;
                            for (my $k = 0; $k < $qtree_count; $k++) {
                                my $qtree_name = $qtree->{NAME};
                                if ($qtree_count > 1) {
                                    $qtree_name .= "_${k}";
                                }

                                # update $vol_name to include the qtree name, this way the process_luns routine
                                # will receive this updated value
                                my $new_vol_name = $vol_name;
                                $new_vol_name .= "/${qtree_name}";

                                # process all the luns inside this qtree
                                $process_luns->($lun_hash, $qtree_luns, $new_vol_name, $qtree);
                            }
                        }
                    }
                }
            }

            # Create hash of lifs
            my $lifs = &ParamsMgr::search(
                -search_path => [],
                -exact_key   => 1,
                -key         => "CLUSTERS.CLUSTER"
                    . $cluster->{_INDEX}
                    . ".VSERVER"
                    . $vserver->{_INDEX},
                -key_filter => qr/^LIF\d*$/,
                paramset => $engenx_obj->EnGenX::Util::get_paramset()
            );
            if (&ParamsMgr::error()) {
                confess &ParamsMgr::get_error();
                return ();
            }

            # Make global arrays for iSCSI lifs and FCP lifs
            my @iscsi_lif_list = ();
            my @fcp_lif_list = ();
            my @nfs_lif_list = ();
            my @cifs_lif_list = ();
            my @lif_list = ();
            foreach my $lif (values %{$lifs}) {

                if ($lif->{'DATA-PROTOCOL'} =~ /fcp/) {
                    push(@fcp_lif_list, $lif->{'NAME'});
                } elsif ($lif->{'DATA-PROTOCOL'} =~ /iscsi/) {
                    push(@iscsi_lif_list, $lif->{'NAME'});
                }

                # since a lif can be both nfs and cifs we need to treat them separately
                if ($lif->{'DATA-PROTOCOL'} =~ /nfs/) {
                    push(@nfs_lif_list, $lif->{'NAME'});
                }
                if ($lif->{'DATA-PROTOCOL'} =~ /cifs/) {
                    push(@cifs_lif_list, $lif->{'NAME'});
                }
            }
            # combine fcp and iscsi lif lists into one
            @lif_list = (@fcp_lif_list, @iscsi_lif_list);

            # Create hash of portsets
            my $portsets = &ParamsMgr::search(
                -search_path => [],
                -exact_key   => 1,
                -key         => "CLUSTERS.CLUSTER"
                    . $cluster->{_INDEX}
                    . ".VSERVER"
                    . $vserver->{_INDEX},
                -key_filter => qr/^PORTSET\d*$/,
                paramset => $engenx_obj->EnGenX::Util::get_paramset()
            );
            if (&ParamsMgr::error()) {
                confess &ParamsMgr::get_error();
                return ();
            }

            my $igroups = &ParamsMgr::search(-search_path => [],
                                             -exact_key => 1,
                                             -key => "CLUSTERS.CLUSTER" . $cluster->{_INDEX} .
                                                     ".VSERVER" . $vserver->{_INDEX},
                                             -key_filter => qr/^IGROUP\d*$/,
                                             paramset => $engenx_obj->EnGenX::Util::get_paramset());
            if (&ParamsMgr::error()) {
                confess &ParamsMgr::get_error();
                return ();
            }
            # iterate over all igroups
            foreach my $igroup (values %{$igroups}) {

                my @port_list = ();

                $igroup->{_INDEX} = "" if (not defined($igroup->{_INDEX}));
                my $protocol = $igroup->{'PROTOCOL'};
                my $initiators = &ParamsMgr::search(-search_path => [],
                                                    -exact_key => 1,
                                                    -key => "CLUSTERS.CLUSTER" . $cluster->{_INDEX} .
                                                            ".VSERVER" . $vserver->{_INDEX} .
                                                            ".IGROUP" . $igroup->{_INDEX},
                                                    -key_filter => qr/^INITIATOR\d*$/,
                                                    paramset => $engenx_obj->EnGenX::Util::get_paramset());
                if (&ParamsMgr::error()) {
                    confess &ParamsMgr::get_error();
                    return ();
                }

                # add Initiator_list entries into this hash
                if (exists($igroup->{INITIATOR_LIST})) {
                    my @init_list = split(/\s*,\s*/, $igroup->{INITIATOR_LIST});
                    foreach my $init (@init_list) {
                        $init =~ s/\s+//g;
                        $initiators->{$init}->{NAME} = $init;
                    }
                }

                # check if there is a portset inside of vserver
                # that the igroup we are currently iterating over
                # is bound to
                foreach my $portset (values %{$portsets}) {
                    $portset->{_INDEX} = ""
                        if (not defined($portset->{_INDEX}));

                    my $igroup_binds = &ParamsMgr::search(
                        -search_path => [],
                        -exact_key   => 1,
                        -key         => "CLUSTERS.CLUSTER"
                            . $cluster->{_INDEX}
                            . ".VSERVER"
                            . $vserver->{_INDEX}
                            . ".PORTSET"
                            . $portset->{_INDEX},
                        -key_filter => qr/^IGROUP_BIND\d*$/,
                        paramset => $engenx_obj->EnGenX::Util::get_paramset()
                    );
                    if (&ParamsMgr::error()) {
                        confess &ParamsMgr::get_error();
                        return ();
                    }

                    # for the current igroup and portset check if the igroup_bind
                    # that is being iterated over matches the current igroup and
                    # define the port_list based on the current portset
                    foreach my $igroup_bind (values %{$igroup_binds}) {

                        # if the current igroup_bind matches the current igroup
                        # define the port_list based on the current portset. If
                        # there are multiple igroups append the count index to
                        # find the appropriate igroup
                        if ($igroup->{COUNT} == 1) {
                            if ($igroup->{NAME} eq $igroup_bind->{NAME}) {
                                @port_list = split(',', $portset->{'PORT-NAME'});
                                last;
                            }
                        } else {
                            foreach my $i (0 .. ($igroup->{COUNT} - 1)) {
                                if ($igroup->{NAME}.'_'.$i eq $igroup_bind->{NAME}) {
                                    @port_list = split(',', $portset->{'PORT-NAME'});
                                    last;
                                }
                            }
                        }
                        if (scalar(@port_list) > 0) {
                            last; #if there is something in port_list, break out
                        }
                    }
                    if (scalar(@port_list) > 0) {
                        last; #if there is something in port_list, break out
                    }
                } # end foreach my $portset ( values %{$portsets} )

                # default the iscsi port list to be all the lifs in this vserver
                my @iscsi_port_list = @iscsi_lif_list;
                my $is_iscsi_portset = 0;
                my @fcp_port_list = @fcp_lif_list;
                my $is_fcp_portset = 0;
                if ($protocol =~ /iscsi/i and scalar(@port_list)) {
                    my %iscsi_lif_union = ();
                    my %iscsi_lif_isect = ();
                    foreach my $lif_element (@iscsi_port_list, @port_list) {
                        $iscsi_lif_union{$lif_element}++ and $iscsi_lif_isect{$lif_element}++;
                    }
                    # if there is an intersection between this portset, and the full lif list
                    # then make the list the intersection, otherwise leave it alone
                    if (scalar(keys %iscsi_lif_isect)) {
                        $is_iscsi_portset = 1;
                        @iscsi_port_list = keys %iscsi_lif_isect;
                    }
                }

                if ($protocol =~ /fcp/i and scalar(@port_list)) {
                    my %fcp_lif_union = ();
                    my %fcp_lif_isect = ();
                    foreach my $lif_element (@fcp_port_list, @port_list) {
                        $fcp_lif_union{$lif_element}++ and $fcp_lif_isect{$lif_element}++;
                    }
                    # if there is an intersection between this portset, and the full lif list
                    # then make the list the intersection, otherwise leave it alone
                    if (scalar(keys %fcp_lif_isect)) {
                        $is_fcp_portset = 1;
                        @fcp_port_list = keys %fcp_lif_isect;
                    }
                }

                # iterate over all initiators
                foreach my $initiator (values %{$initiators}) {
                    my $perform_client_lookup = 0;
                    my $proto = "";
                    my $client;

                    # if the initiator matches a WWPN or an IQN iscsi Nodename, then we need to do a reverse
                    # lookup to all the sanclient objects to see who owns the nodename, once we find the
                    # owner, set $client to the owner name
                    if ($initiator->{NAME} =~ /^[a-fA-F0-9]{16}$/ or
                        $initiator->{NAME} =~ /^[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:
                                                [a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}$/x) {
                        # this is an fcp portname, normalize it
                        $initiator->{NAME} = $engenx_obj->normalize_fcp_wwn(-wwn => $initiator->{NAME});
                        if ($engenx_obj->error()) {
                            confess "Could not normalize the FCP WWPN initiator " . $initiator->{NAME};
                            return ();
                        }
                        # set a flag to perform a reverse wwpn client name lookup
                        $perform_client_lookup = 1;
                        $proto = "fcp";
                    } elsif ($initiator->{NAME} =~ /^iqn\.\S+/) {
                        $perform_client_lookup = 1;
                        $proto = "iscsi";
                    } else {
                        $perform_client_lookup = 0;
                    }

                    # search for the owner of this initiator name if requested to
                    if ($perform_client_lookup) {
                        foreach my $client_obj (@client_objs) {
                            my @initiator_objs = $client_obj->get_initiators();
                            if ($client_obj->error()) {
                                confess "Failed to obtain initiators from client " . $client_obj->hostname() . " : " . $client_obj->errmsg();
                                return ();
                            }

                            # search through the initiator objects
                            my $found_it = 0;
                            foreach my $init_obj (@initiator_objs) {
                                my $initiator_protocol = $init_obj->get_protocol();

                                if ($initiator_protocol eq "fcp" and $proto eq "fcp") {
                                    # If it's fcp, compare the portnames
                                    my $portname = $engenx_obj->normalize_fcp_wwn(-wwn => $init_obj->get_portname());
                                    if ($engenx_obj->error()) {
                                        confess "Failed to normalize FCP WWPN " . $init_obj->get_portname();
                                        return ();
                                    }
                                    if ($portname eq $initiator->{NAME}) {
                                        $found_it = 1;
                                        $client = $client_obj->hostname();
                                        last;
                                    }
                                } elsif ($initiator_protocol eq "iscsi" and $proto eq "iscsi") {
                                    # check if the iscsi nodename matches
                                    if (lc($init_obj->get_nodename()) eq lc($initiator->{NAME})) {
                                        $found_it = 1;
                                        $client = $client_obj->hostname();
                                        last;
                                    }
                                }
                            }
                            # break out if we found it
                            last if ($found_it);
                        }

                        # if client is still not defined:
                        if (not defined($client)) {
                            # then this is most likely a dummy initiator OR an initiator that is not included in the XML
                            # we can't error out here, so the only thing we can do at this point
                            # is set the client to be the initiator name.
                            $client = $initiator->{NAME};
                        }
                    } else {
                        # otherwise, the name of the initiator is a client name
                        $client = $initiator->{NAME};
                    }

                    # if this is a single vserver
                    if ($vserver->{COUNT} == 1) {
                        # mark this vserver as related to this client
                        # assign the fcp/iscsi lif relationships for this client
                        $self->_assign_lif_lists_to_client_vserver_hash($client_relationship_hash,
                                                                        $client,
                                                                        $cluster->{NAME},
                                                                        $vserver->{NAME},
                                                                        $is_fcp_portset,
                                                                        $is_iscsi_portset,
                                                                        \@fcp_port_list,
                                                                        \@iscsi_port_list);

                        # this igroup may have COUNT>1
                        for (my $j = 0; $j < $igroup->{COUNT}; $j++) {
                            my $igroup_name;
                            if ($igroup->{COUNT} > 1) {
                                $igroup_name = $igroup->{NAME} . "_${j}";
                            } else {
                                $igroup_name = $igroup->{NAME};
                            }
                            $lun_relationship_hash->{$client}->{$cluster->{NAME}}->{$vserver->{NAME}}->{$igroup_name} = $lun_hash->{$igroup_name};
                        }

                        # vserver lookup
                        $vserver_relationship_hash->{$cluster->{NAME}}->{$vserver->{NAME}}->{$protocol}->{$client} += 1;
                    } else {
                        # handle multiple vservers that are related to this client
                        foreach my $i (0..($vserver->{COUNT} - 1)) {
                            $self->_assign_lif_lists_to_client_vserver_hash($client_relationship_hash,
                                                                            $client,
                                                                            $cluster->{NAME},
                                                                            $vserver->{NAME} . "_${i}",
                                                                            $is_fcp_portset,
                                                                            $is_iscsi_portset,
                                                                            \@fcp_port_list,
                                                                            \@iscsi_port_list);

                            # lun lookup
                            for (my $j = 0; $j < $igroup->{COUNT}; $j++) {
                                my $igroup_name;
                                if ($igroup->{COUNT} > 1) {
                                    $igroup_name = $igroup->{NAME} . "_${j}";
                                } else {
                                    $igroup_name = $igroup->{NAME};
                                }
                                $lun_relationship_hash->{$client}->{$cluster->{NAME}}->{$vserver->{NAME} . "_$i"}->{$igroup_name} = $lun_hash->{$igroup_name};
                            }

                            # vserver lookup
                            $vserver_relationship_hash->{$cluster->{NAME}}->{$vserver->{NAME} . "_$i"}->{$protocol}->{$client} += 1;
                        }
                    }
                }
            }

            if ($load_nas_data) {
                # if this vserver has NAS lifs, associate it to every client that is doing NFS or CIFS
                my $all_clients = $engenx_obj->_get_all_client_paramsets();
                if (scalar(@nfs_lif_list) or scalar(@cifs_lif_list)) {
                    foreach my $client_ds (@{$all_clients}) {
                        my @protos = ();
                        if ($client_ds->{PROTOCOL} =~ /nfs/i) {
                            push(@protos, "nfs");
                        }
                        if ($client_ds->{PROTOCOL} =~ /cifs/i) {
                            push(@protos, "cifs");
                        }

                        next if (not scalar(@protos));

                        if ($vserver->{COUNT} == 1) {
                            foreach my $proto (@protos) {
                                $vserver_relationship_hash->{$cluster->{NAME}}->{$vserver->{NAME}}->{$proto}->{$client_ds->{NAME}} += 1;
                                if ($proto eq "cifs") {
                                    $client_relationship_hash->{$client_ds->{NAME}}->{$cluster->{NAME}}->{$vserver->{NAME}}->{cifs_lifs} = \@cifs_lif_list;
                                }
                                if ($proto eq "nfs") {
                                    $client_relationship_hash->{$client_ds->{NAME}}->{$cluster->{NAME}}->{$vserver->{NAME}}->{nfs_lifs} = \@nfs_lif_list;
                                }
                            }
                        } else {
                            for my $i (0..($vserver->{COUNT} - 1)) {
                                foreach my $proto (@protos) {
                                    $vserver_relationship_hash->{$cluster->{NAME}}->{$vserver->{NAME} . "_$i"}->{$proto}->{$client_ds->{NAME}} += 1;
                                    if ($proto eq "cifs") {
                                        $client_relationship_hash->{$client_ds->{NAME}}->{$cluster->{NAME}}->{$vserver->{NAME} . "_$i"}->{cifs_lifs} = \@cifs_lif_list;
                                    }
                                    if ($proto eq "nfs") {
                                        $client_relationship_hash->{$client_ds->{NAME}}->{$cluster->{NAME}}->{$vserver->{NAME} . "_$i"}->{nfs_lifs} = \@nfs_lif_list;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    return ($client_relationship_hash, $vserver_relationship_hash, $lun_relationship_hash);
}

sub _assign_lif_lists_to_client_vserver_hash {
    my ($self, $hash, $client, $cluster, $vserver, $is_fcp_portset,
        $is_iscsi_portset, $fcp_port_list, $iscsi_port_list) = @_;

    # (A) if the specified lif list is already populated
    #     1 - if its populated with a non portset list, set it to the new list
    #     2 - if its populated with a portset list and the list we currently have is a non portset list, do nothing
    #     3 - if its populated with a portset list and the list we have is a portset list, unionize the groups
    # (B) if list is not populated, then set it to the new lif list
    if (exists($hash->{$client}->{$cluster}->{$vserver}->{fcp_lifs})) {
        my @old_value = @{$hash->{$client}->{$cluster}->{$vserver}->{fcp_lifs}};
        if (not $hash->{$client}->{$cluster}->{$vserver}->{is_fcp_portset}) {
            $hash->{$client}->{$cluster}->{$vserver}->{fcp_lifs} = $fcp_port_list;
        } elsif ($is_fcp_portset) {
            my %fcp_lif_union = ();
            foreach my $lif_element (@{$fcp_port_list}, @old_value) {
                $fcp_lif_union{$lif_element}++;
            }
            $hash->{$client}->{$cluster}->{$vserver}->{fcp_lifs} = [keys %fcp_lif_union];
        }
    } else {
        $hash->{$client}->{$cluster}->{$vserver}->{is_fcp_portset} = $is_fcp_portset;
        $hash->{$client}->{$cluster}->{$vserver}->{fcp_lifs} = $fcp_port_list;
    }

    if (exists($hash->{$client}->{$cluster}->{$vserver}->{iscsi_lifs})) {
        my @old_value = @{$hash->{$client}->{$cluster}->{$vserver}->{iscsi_lifs}};
        if (not $hash->{$client}->{$cluster}->{$vserver}->{is_iscsi_portset}) {
            $hash->{$client}->{$cluster}->{$vserver}->{iscsi_lifs} = $iscsi_port_list;
        } elsif ($is_iscsi_portset) {
            my %iscsi_lif_union = ();
            foreach my $lif_element (@{$iscsi_port_list}, @old_value) {
                $iscsi_lif_union{$lif_element}++;
            }
            $hash->{$client}->{$cluster}->{$vserver}->{iscsi_lifs} = [keys %iscsi_lif_union];
        }
    } else {
        $hash->{$client}->{$cluster}->{$vserver}->{is_iscsi_portset} = $is_iscsi_portset;
        $hash->{$client}->{$cluster}->{$vserver}->{iscsi_lifs} = $iscsi_port_list;
    }
}

sub _svc_get_luns_associated_to_client {
    my ($self) = shift;

    my ($listener, %args) = @_;

    my $client_name = delete($args{client_name}) // confess "Missing required parameter 'client_name'";

    my $hash = $listener->{data}->{lun_relationship_hash};

    # build an array of all the lun paths that are to be mapped to this host,
    # this includes all vservers from all clusters, so duplicate lun names may
    # appear
    my @luns;
    foreach my $cluster_name (keys %{$hash->{$client_name}}) {
        foreach my $vserver_name (keys %{$hash->{$client_name}->{$cluster_name}}) {
            my @vserver_luns;
            foreach my $igroup_name (keys %{$hash->{$client_name}->{$cluster_name}->{$vserver_name}}) {
                push(@vserver_luns, $hash->{$client_name}->{$cluster_name}->{$vserver_name}->{$igroup_name});
            }

            # remove dups from this list
            my %dup_hash;
            foreach my $lun (@vserver_luns) {
                $dup_hash{$lun}++;
            }

            # push all the luns on the main list
            push(@luns, keys %dup_hash);
        }
    }

    return \@luns;
}

sub _svc_get_vservers_associated_to_client {
    my ($self) = shift;

    my ($listener, %args) = @_;

    my $client_name = delete($args{client_name}) // confess "Missing required parameter 'client_name'";

    my $client_map_hash = $listener->{data}->{client_relationship_hash};
    my $vserver_map_hash = $listener->{data}->{vs_relationship_hash};

    # pull out all the vservers that are related to this client
    my @vserver_list;
    foreach my $cluster (keys %{$client_map_hash->{$client_name}}) {
        push(@vserver_list, keys %{$client_map_hash->{$client_name}->{$cluster}});
    }

    return \@vserver_list;
}

sub _svc_get_luns_and_vservers_associated_to_client {
    my ($self) = shift;

    my ($listener, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my $client_name = delete($args{client_name}) // confess "Missing required parameter 'client_name'";

    # generate the relationship hashs
    # vserver_hash in the following format:
    #{
    #    '20:01:00:11:0d:1a:a7:4d' => {
    #        'ssan-6280-1' => {
    #            'vs_8n_mixed3_1' => [
    #                'lf_fcp_n7_p2_13',
    #                'lf_fcp_n7_p1_14',
    #                'lf_fcp_n8_p2_15',
    #                'lf_fcoe_n5_p2_9',
    #                'lf_fcoe_n5_p1_10',
    #                'lf_fcoe_n1_p1_2',
    #                'lf_iscsi_n7_p2_13',
    #                'lf_iscsi_n7_p1_14',
    #                'lf_iscsi_n1_p1_2'
    #            ]
    #        }
    #    }
    #}

    # lun_hash in the following format:
    #{
    #    '20:01:00:11:0d:1a:a7:4d' => {
    #        'ssan-6280-1' => {
    #            'vs_8n_mixed3_1' => {
    #                'ig_SANBLAZE_8_78' => [
    #                    '/vol/vl_1769_78/ln_SANBLAZE_8_vl_1769_78_1',
    #                    '/vol/vl_1769_78/ln_SANBLAZE_8_vl_1769_78_2',
    #                    '/vol/vl_1769_2010/ln_SANBLAZE_8_vl_1769_2010_6',
    #                    '/vol/vl_1769_2010/ln_SANBLAZE_8_vl_1769_2010_7',
    #                    '/vol/vl_1769_2010/ln_SANBLAZE_8_vl_1769_2010_8'
    #                ]
    #            }
    #        }
    #    },
    #}

    my $vserver_hash = $listener->{data}->{vs_relationship_hash};
    my $lun_hash = $listener->{data}->{lun_relationship_hash};

    # build an array of all the lun paths that are to be mapped to this host,
    # this includes all vservers from all clusters, so duplicate lun names may appear
    my @luns;
    foreach my $cluster_name (keys %{$lun_hash->{$client_name}}) {
        foreach my $vserver_name (keys %{$lun_hash->{$client_name}->{$cluster_name}}) {
            my @vserver_luns;
            foreach my $igroup_name (keys %{$lun_hash->{$client_name}->{$cluster_name}->{$vserver_name}}) {
                push(@vserver_luns, $lun_hash->{$client_name}->{$cluster_name}->{$vserver_name}->{$igroup_name});
            }

            # remove dups from this list
            my %dup_hash;
            foreach my $lun (@vserver_luns) {
                $dup_hash{$lun}++;
            }

            # push all the luns on the main list
            push(@luns, keys %dup_hash);
        }
    }

    # pull out all the vservers that are related to this client
    my @vserver_list;
    foreach my $cluster (keys %{$vserver_hash->{$client_name}}) {
        push(@vserver_list, keys %{$vserver_hash->{$client_name}->{$cluster}});
    }

    return [\@vserver_list, \@luns];
}

sub _svc_get_clients_associated_to_vserver {
    my ($self) = shift;

    my ($listener, %args) = @_;

    my $cluster_name  = delete($args{cluster_name})     // confess "Missing required parameter 'cluster_name'";
    my @vserver_names = @{delete($args{vserver_names})};
    if (!scalar(@vserver_names)) {
        @vserver_names = ();
    }
    my @protocols     = @{delete($args{protocols})};
    if (!scalar(@protocols)) {
        @protocols = ("fcp", "iscsi");
    }

    my $client_map_hash = $listener->{data}->{client_relationship_hash};
    my $vserver_map_hash = $listener->{data}->{vs_relationship_hash};

    # pull out all the vservers that are related to this client
    my %client_list;

    # search the vservers provided
    if (scalar(@vserver_names)) {
        foreach my $vserver (@vserver_names) {
            # for each protocol type specified, obtain the clients associated with
            # this vserver
            foreach my $protocol (@protocols) {
                foreach my $client_name (keys %{$vserver_map_hash->{$cluster_name}->{$vserver}->{$protocol}}) {
                    $client_list{$client_name}++;
                }
            }
        }
    } else {
        # walk all vservers
        foreach my $vserver (keys %{$vserver_map_hash->{$cluster_name}}) {
            # for each protocol type specified, obtain the clients associated with
            # this vserver
            foreach my $protocol (@protocols) {
                foreach my $client_name (keys %{$vserver_map_hash->{$cluster_name}->{$vserver}->{$protocol}}) {
                    $client_list{$client_name}++;
                }
            }
        }
    }

    # convert from hash to array to remove duplicates from the list
    my @client_list = keys(%client_list);

    return \@client_list;
}

sub get_luns_associated_to_client {
    my ($self, %args) = @_;

    my $msg = $self->_common_svcapi(%args, _api => "get_luns_associated_to_client");

    return wantarray() ? @{$msg} : $msg;
}

sub get_luns_and_vservers_associated_to_client {
    my ($self, %args) = @_;

    my $msg = $self->_common_svcapi(%args, _api => "get_luns_and_vservers_associated_to_client");

    return wantarray() ? @{$msg} : $msg;
}

sub get_clients_associated_to_vserver {
    my ($self, %args) = @_;

    my $msg = $self->_common_svcapi(%args, _api => "get_clients_associated_to_vserver");

    return wantarray() ? @{$msg} : $msg;
}

sub get_vservers_associated_to_client {
    my ($self, %args) = @_;

    my $msg = $self->_common_svcapi(%args, _api => "get_vservers_associated_to_client");

    return wantarray() ? @{$msg} : $msg;
}

sub get_vserver_client_relationship_hash {
    my ($self, %args) = @_;

    my $data = $self->_common_svcapi(%args, _api => "get_vserver_client_relationship_hash");

    return ($data->{client_relationship_hash},
            $data->{vs_relationship_hash},
            $data->{lun_relationship_hash});
}

sub get_client_relationship_hash {
    my ($self, %args) = @_;

    return $self->_common_svcapi(%args, _api => "get_client_relationship_hash");
}

sub _svc_get_client_relationship_hash {
    my ($self) = shift;

    my ($listener, %args) = @_;

    my $client_name = delete($args{client_name}) // confess "Missing required parameter 'client_name'";

    return $listener->{data}->{client_relationship_hash}->{$client_name};
}

sub _svc_get_vserver_client_relationship_hash {
    my ($self) = shift;

    my ($listener, %args) = @_;

    return $listener->{data};
}

1;
