# $Id: //depot/prod/test/Rsierranevada/lib/EnGenX/SubProcess.pm#4 $

#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Processes Layer Module
## @author dl-san-automation
## @status shared
## @pod here

=head1 NAME

EnGenX::SubProcess

=head1 DESCRIPTION

EnGenX::SubProcess implements many instances of the
NATE::Process class to work in parallel. Each Process class
implementation called SubProcess will execute the provided
code string.

=cut

# Package
package EnGenX::SubProcess;

# Compiler directives.
use strict;

# NetApp's module imports
use Tharn;
use Error;
use ParamsMgr;
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate OBJECT SCALAR SCALARREF ARRAYREF HASHREF UNDEF GLOBREF);
use SubIPC;

# EnGenX Libraries
use EnGenX::EnGenXUtil;
use NATE::Exceptions::Argument qw(:try);
use EnGenX::Exceptions::SubProcessFailure qw(:try);

# Load resource specific modules
use NATE::Process;
use NATE::ParamSet;
use NATE::STAF::Peer;

# Perl's public module imports
use POSIX;
use Data::Dumper;
use DateTime;

# Global Variables.
use vars qw(
    @ISA
    @Resource_List
    $SubProcs_Count
);

#
# Inherit from the Error class
#
@ISA = qw( Error );

# Whenever required the time-out time will be incremented.
my $on_timeout = sub { return 2 * $_[0]; };

=head1 Constructor

=head2 new

(Static method) Request creation of an instance of the SubProcess
class. It will update the provided atributes and return the object
if no error is encountered.

   my $codestring = '{sub func {print "I am a function"}}';
   $subproc_obj = EnGenX::SubProcess->new(
                                         'code_string' => $codestring,
                                         'count' => 4,
                                         'bucket_size' => 10
                                        );
=over

=item C<< 'code_string' => $codestring >>

(Required)
The code string is required to create a SubProcess.
It is the piece of code that will be eval when the Process is
executed. This argument is required to create NATE::Processes



=item C<< count => int >>

(Optional)
This argument is the number of child process that will be created.
These child processes or sub processes are identical, they will all
be executing the same CodeString, but will interact with different
elements of the resource table.
It defaults to 3, so, three process will be created to work the
resource list.
The number must be an integer.

=item C<< bucket_size => int >>

(Optional)
This arguments gives the size of the list of resources that should be
passed to each subprocess.
It defaults to 1, so only one resource at a time will be passed to each
subprocess. Then through messaging other resources will be provided to
the subprocess.
The number must be an integer.


=back

=item Return

Returns the created object.

=item Exceptions

This method does not produces not throws exceptions.
TODO:Implement error handling with Exception
TODO:Implement error handling with Exception

=cut

sub new {
    my ($class, %args) = @_;
    $Log->enter();

    # Class Initialization
    my $self = {};
    bless $self, $class;


    &ParamsMgr::process_args(%args);

    my $EnGenX = &ParamsMgr::get_arg(
        -arg      => "EnGenX",
        -required => 1,
        -default  => ""
    );
    my $codestring
        = &ParamsMgr::get_arg('arg' => "code_string", 'required' => 1);
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => 0,
        'default'  => 3
    );
    my $size = &ParamsMgr::get_arg(
        'arg'      => "bucket_size",
        'required' => 0,
        'default'  => 1
    );


    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return undef;
    }

    # Initialize Global variables
    $self->{'code_string'} = $codestring;
    $self->{'count'}       = $count;
    $self->{'bucket_size'} = $size;
    $self->{'EnGenX'} = $EnGenX;

    $Log->trace("SELF: "
            . "\n CODE_STRING: $codestring "
            . "\n COUNT: $count"
            . "\n BUCKET_SIZE: $size ");


    $Log->exit();
    return $self;

} ## end sub new

=head1 METHODS

=head2 subprocesses_create

Request for the creation of one or multiple NATE::Processes instances.
These instances will return in an array of objects and could return
with the process started if requested by the user.

@subprocesses = EnGenX::SubProcess->subprocesses_create(
                                                    'count'=> 4,
                                                    'start'=> TRUE,
                                                    'call'=> \%call);

=over

=item C<< count => int >>

(Required)
This parameter is the number of subprocesses that will be created and
returned in an array of subprocesses.
The number must be an integer.

=item C<< 'call' => HASHREF >>

(Required)
The call parameter is required to create the SubProcesses, because it's content
will become the 'params' key that the NATE::Process object will be created with.
The keys/values that this hash references needs to hold are:

    'package' => $package_name, # The name of the calling EnGenX::Controller::Task
    'action'  => $action,       # The action for the resource. Create|Delete
    'nacl_nodes'=> \@node       # The array holding (one or more) nodes for
                                # commmand interface that NACL should use.



=item C<< start => TRUE|FALSE >>

(Optional)
If this parameter is passad as TRUE, the method will not only create
N number of processes, but also will start them Subprocess->start().
It expects a boolean value and will default to FALSE.


=back

=item Return

This method returns an array or a reference to an array containing the
subprocess (NATE::Process) objects.

=item Exceptions

This method does not produces nor throws exceptions.
TODO:Implement error handling with Exception

=cut

sub subprocesses_create {
    my ($self, @opts) = @_;
    $Log->enter();
    my %args = ();

    # Validate Argumetns
    try {
        %args = Params::Validate::validate(@opts, {
            call  => {type=>HASHREF,optional => 0},
            start => {type=>SCALAR, default => FALSE,
                                    optional => 1},
            count => {type=>SCALAR, default => $self->{'count'},
                                    optional => 1},
        });
    }
    otherwise {
        my $exception = shift;
        NATE::Exceptions::Argument->throw(
                "SubProcess::subprocesses_create parameters errror :\n" .
                $exception->text());
    };

    # Assign Values
    my $start = $args{start};
    my $count = $args{count};
    my $call = $args{call};

    #Validate all the exepcted key in CALL
    # NODE(S)
    if (!defined($call->{nacl_nodes})) {
        NATE::Exceptions::Argument->throw(
                "SubProcess::subprocesses_create CALL parameters errror :\n" .
                "Missing an array of NODES to be used by NACL as a " .
                "Command_Interface.");
    }
    # PACKAGE
    if (!exists($call->{package})) {
        NATE::Exceptions::Argument->throw(
                "SubProcess::subprocesses_create CALL parameters errror :\n" .
                "Missing the PACKAGE value that is usually is the package of " .
                "the caller of this mehtod");
    }
    # ACTION
    if (!exists($call->{action})) {
        NATE::Exceptions::Argument->throw(
                "SubProcess::subprocesses_create CALL parameters errror :\n" .
                "Missing ACTION to be be performed. Most common values are " .
                "'Create' and 'Delete'.");
    }


    # Locals
    my @subprocesses = ();
    my $node_index = 0;
    my $node = undef;
    # Extracting from the CALLER the TYPE (e.g. lun, vol, etc) to be added to
    # the Run ID to give more information in the Logs.
    my $type = (($call->{package} =~ /.*::(.*).*/i)
                ? $1
                : 'empty')."_".($call->{action});

    # get the runid prefix from subipc
    my $runid_prefix = &SubIPC::_get_next_subtest_count();

    $Log->trace("Will create $count SUB-PROCESSes of Type: '$type'");
    for (my $i = 1; $i <= $count; $i++) {
        # An array of nodes was provided, so distribute the load among them.
        # Reset the index if needed.
        $node_index = 0 if($node_index >= scalar(@{$call->{nacl_nodes}}));
        # Assign a node to this subprocess.
        $node = $call->{nacl_nodes}->[$node_index++];
        # TODO: Create a real load balance logic. Right now, we are only
        #         dividing work evenly based on the number of subprocs.

        $Log->debug("EnGenX::SubProcess::subprocesses_create: "
                      ."Sending to NATE::PROCESS the following PARAMETERS: \n"
                      ."PACKAGE = ". $call->{package} ." \n"
                      ."ACTION  = ". $call->{action} ." \n"
                      ."NODE    = ". $node->{node} ." \n"
                );

        # Create the Sub-Process args
        my %subproc_params = (
            'codespec' => $self->{'code_string'},
            'runid'    => $runid_prefix . "_" . $type . "/" . $count . "_Parallel_SubProcess_" . $i,
            'background' => 1,
            onexit => \&NATE::Process::on_exit_propagate_worst_result,
            # Need to specify the client so that Nate will exec instead of just
            # fork. We do this because when we fork memory leaks for an
            # unknown reason.
            'client'  => 'localhost',
            'params' => {'package' => $call->{package},
                         'action'  => $call->{action},
                         'node'    => $node->{node},
                         $self->_get_tharn_args(),
                        },



            # TODO: Here is where the exception handling implamantation
            #       could start it's implementation with ONEXIT
            #'onexit' => \exit_handler,
        );

        # CREATE THE OBJECTS
        my $subproc_obj = NATE::Process->new(%subproc_params);

        # START THE PROCES IF REQUESTED BY USER
        $subproc_obj->start() if ($start);
        # GATHER THEM IN THE RETURN ARRAY
        push(@subprocesses, $subproc_obj);
    }

    $Log->exit();
    return (wantarray()) ? @subprocesses : \@subprocesses;

}    ## subprocesses_create

=head2 parallel_execution

This method splits the resource list among the one or multiple NATE::Process
instances (subprocesses).
There are several high level steps:
1. Gets the resource list from the input parameters (arguments).
2. Divides the resource list evenly into buckets based on the 'bucket_size'
   parameter (if bocket size is bigger than 1).
3. It calls the subprocesses_create to create all the parallel subprocesses.
4. Monitors messages from each subprocesses:
   There are 3 possible messages of type 'sstatus':
   a) 'waiting' message.
   b) 'failed' message.
   c) 'request_ip' message.
4.a. When found a subprocess in 'waiting' status first we updated the resource
     list (if any work was done) then if needed store the resource index. Then,
     a new resource item or bucket will be assigned to that subProcess and
     continues monitoring other subprocess.
4.b. When found a suprocess with 'failed' status message will updated the
     resource list, then stop, destroy the subprocess and remove it from the
     list so that is no longer monitored.
4.c. When found a suprocess with 'ip_reques' status message will create an stack
    of ip addresses from the {ips} attribute of the resources (only the first
    time). Then it will extract (pop) an IP and will send it back to the
    subprocess that sent the message.
5. When no more items in the resource list are left, it will indicate each
   subprocess to exit, and will be expecting a result from them.
6. After each subprocess has 'died', it will stop and destroy each one of them.
7. Finally returns the updated resource list.



@updated_list
        = EnGenX::SubProcess->parallel_execution('list' => $resource_list,
                                                 'on_error' => "finish"
                                                );


=over

=item list

(Required)
This parameter holds the reference to an array of hashes that is the
resource list to work on.

=item C<< on_error => exit|finish|ignore >>

(Optional)
This parameter will indicate the Parent process what to do in the case a
Child process fails or encounters an error.
There are three options.
1. Exit - The Parent process will throw (die) an exception as soon as
        the child process reports the error.
        All the resources that
        were not porcessed or were in the middle of the processing
        will not be completed. Only the resoureces that were efectively
        processed will be updated in the returning resurce list.

2. Finish - The parent process will terminate(kill) the failed child process,
        will continue processing the provided list with the remaining
        avialble subprocesses and when finished, will throw (die) an
        exception.
        All the resources on the list, except for the one that cause
        the error, will be worked.

3. Ignore - The parent process will terminate (kill) the failed child process
        and will continue processing the provided list with the remaining
        avialble subprocesses. The Parent process will not report the failure
        and continue executing as normal.
        All the resources on the list, except for the one that cause the
        error, will be worked.

The default behavoir will be to wait until FINISH(2) to report the failure.


=back



=item Return

This method returns an array or a reference to an array containing an updated
version of the provided resource list. This resource list is usually a subset
of the complete resource list.

=item Exceptions

This method does not produces nor throws exceptions.
TODO:Implement error handling with Exception
=cut

sub parallel_execution {
    my ($self, @opts) = @_;
    $Log->enter();

    # LOCALS
    my %args = ();
    my @subprocesses  = ();
    my @errors = ();
    my $failed = FALSE;

    # Validate Argumetns
    try {
        %args = Params::Validate::validate(@opts, {
            list     => {type=>ARRAYREF, optional => FALSE},
            call     => {type=>HASHREF,  optional => FALSE},
            on_error => {type=>SCALAR,  default =>"finish", optional => 1},
        });
    }
    otherwise {
        my $exception = shift;
        NATE::Exceptions::Argument->throw(
                "SubProcess::parallel_execution parameters errror :\n" .
                $exception->text());
    };

    # Assign Values
    my $list = $args{list};
    my $on_error = $args{on_error};

    # Verify that the ON_ERROR parameter has a valid value
    if (not ($on_error =~/exit|finish|ignore/i)){
        my $exception_mssg = ("Unvalid value for paramter 'on_error'".
                         "Supported values are: 'exit', 'finish' ".
                         "and 'ignore'");
        $Log->exit();
        NATE::Exceptions::Argument->throw($exception_mssg);
    }

    # Create and Start the SubProcesses
    @subprocesses = $self->subprocesses_create(
        'count' => $self->{'count'},
        'start' => TRUE,
        'call' => $args{call},
    );

    # Variables for the Suprocess While Loop
    my $staf = NATE::STAF::staf_global();
    # A hash containing the last created resource for each process so that
    # we can wait for the last resource to finish when all resources are
    # gone through.
    my %last_rsrcs = ();
    my $stopped_procs_id = 0;
    my $force_exit = FALSE;
    # Will use an array index to know what resource or resources need to be
    # fetched.
    my $rsrc_index = 0;
    my $max_rsrc_index = (scalar(@{$list})-1);
    # The bucket range will go from rsrc_index to rsrc_index+bucket_index
    my $bucket_index = ($self->{bucket_size} > 0) ?
                                ($self->{bucket_size}-1) : $self->{bucket_size};

    # Assign each process a resource to work with
    if ($Log->may_debug()) {
        $Log->debug("The resource list Before Subprocess Loop is:\n"
                . Dumper($list));
    }

while (1) {
        # GET THE FIRST MESSAGE FROM THE "RESPONSE" QUEUE
        my $full_message = $staf->message_get(include_header => TRUE,
                                           type           => "Response",
                                        );
        #LOCALS
        my $message = $full_message->{message};
        my $handle = $full_message->{handle};
        my $handle_name = $full_message->{handleName};
        my %queue_args = (handle   => $handle,
                          endpoint => $full_message->{machine},
                        );

        $Log->trace("SubProcess::parallel_execution: "
                      . "PARENT PROCESS GOT FULL_MESSAGE: "
                      . "Handle[" . $full_message->{handle} .  "], "
                      . "Message->Status[" . $message->{status} . "]\n"
                    );

        #PROCESS THE MESSAGE
        if ($message->{'status'} =~ /waiting/) {
            # UPDATE THE RESOURCE LIST, IF ANY WORK WAS DONE
            my %wait_response = $self->_parallel_execution_wait(
                                                message => $message,
                                                list    => $list,
                                                last_rsrcs    => \%last_rsrcs,
                                                rsrc_index    => $rsrc_index,
                                                bucket_index  => $bucket_index,
                                                max_rsrc_index => $max_rsrc_index,
                                                action => $args{call}->{action},
                                                handle => $handle,
                                                handle_name   => $handle_name,
                                                force_exit    => $force_exit,
                                            );
            # STRIP THE RESPONSE
            my %post_args = %{$wait_response{queue_args}};
            $rsrc_index = $wait_response{rsrc_index};

            # ADD THE ID TO THE STOPPED PROCES
            if($post_args{message} eq "exit") {
                $stopped_procs_id++;
                $Log->trace("SubProcess::parallel_execution: "
                            . "ADDED HANDLE[$handle] TO STOPPED_PORCS_ID: "
                            . $stopped_procs_id
                        );
            }

            # SEND THE MESSAGE
            $staf->queue_post(%queue_args, %post_args);
        } elsif ($message->{'status'} =~ /request_ip/) {
            my %post_args = $self->_parallel_execution_ips(
                                                message     => $message,
                                                handle_name => $handle_name,
                                            );
            # SEND THE MESSAGE
            $staf->queue_post(%queue_args, %post_args);

        } elsif ($message->{'status'} =~ /failed/) {
            # SAVE THE RESOURCES THAT WERE WORKED BEFORE FAILING
            $self->_update_resource_list(message => $message, list=>$list);
            # RECORD ERROR
            my $error_msg = "Error occurred in subprocess [$handle_name]: \n"
                            . $message->{'error_msg'};
            push(@errors, $error_msg);
            # IF USER REQUESTED TO EXIT ON ERROR SET THE FLAG TO GRACEFULLY
            # TERMINATE ALL OTHER SUBPROCESSES
            if ($on_error =~/exit/i){
                $force_exit = TRUE;
            } else {
                # If the user requests to finish, add one the stopped_procs_id to show
                # how many processes have finished. The error will be thrown outside of
                # this loop.
                $stopped_procs_id++;
            }
        } else {
            $Log->warn("Skipped the following message because message status did not match a known status.\n".
                      "This should not happen\n Full Message:".Dumper($full_message));
        }
        # TODO
        # Add logic to keep the subprocess alive so that it can be reused with
        # another resource list that could be of a different type.
        last if ($stopped_procs_id >= $self->{'count'});
    }#end WHILE

    # TERMINATE THE SUB-PROCESSES LEFT IN THE ARRAY
    foreach my $subproc_obj (@subprocesses) {
        my $subproc_id = $subproc_obj->runid();
        if ($subproc_obj->is_running()) {
            $subproc_obj->wait();
        # TODO
        # Uncommnent the following logic when there is logic in place for to
        # keep sub-processes active beyond a particular resource TYPE. Whatever
        # the logic, it should exit (die) the subprocess before stop and destroy.
        # $Log->comment("SENDING EXIT CMD TO SUB-PROCESS: $subproc_id ");
        # $subproc_obj->message_put(type =>"WhatToDo", message=> "exit");
        }

        $Log->trace("Stopping and Destroying SUB-PROCESS(" . $subproc_id . ")");

        # Stop the sub-process
        $subproc_obj->stop();

        # Destroy the sub-process
        $subproc_obj->destroy();
    }

    # RETURN THE UPDATED LIST
    # TODO
    # Logic to compare the scalar of both lists ($list and $worked_list) and if
    # there are any deltas do some logic to overcome.
    if ($Log->may_debug()) {
        $Log->debug("The resource list After Subprocess Loop is:\n"
                . Dumper($list));
    }

    # IF A SUBPROCESS ENCOUNTERED AN ERROR AND ON_ERROR FLAG IS EITHER EXIT OR
    # FINISH
    if (scalar(@errors)>0 && $on_error !~/ignore/i){
        my $exception_msg = "Several (".scalar(@errors). ") SubProcesses failed or ".
                       "encountered errors: \n";
        foreach my $err (@errors){
            $exception_msg .= $err."\n";
        }
        # THROW THE EXCEPTION
        $Log->exit();
        EnGenX::Exceptions::SubProcessFailure->throw(
                                    $exception_msg,
                                    resource => $list
                                );
    }

    $Log->exit();
    return (wantarray()) ? @{$list} : $list;

}    ## parallel_execution

#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
#                PRIVATE METHODS
#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

#==============================================================================
# Method:     _get_tharn_args
# Objective:  Get the tharn args that are to be passed to the children processes
# Details:    Checks if the params are defined and if so then adds them to the
#             hash
# Inputs:   NONE
# Outputs:  Hash of args from tharn that we need.
# Exceptions: No Error handling implemented
# TODO:       Error handling.
#==============================================================================
sub _get_tharn_args {
    my ($self) = @_;
    $Log->enter();
    # Locals
    my %tharn_args = ();

    my $tharn_param = Tharn::param('PROFILING_ENABLED');
    $tharn_args{'PROFILING_ENABLED'} = $tharn_param if (defined($tharn_param));

    return (wantarray()) ? %tharn_args : \%tharn_args;
}

#==============================================================================
# Method:     _update_results_array
# Objective:  Updates the original ResourceList, argument 'list', that is been
#              passed as a reference.
# Details:    Steps: 1) Verify that the message has a resource.
#                    2) Based on the index that is stored within the message
#                       updates the list with the provided resource.
# Inputs:     message
#                 The NATE::STAFF:Peer message that is send from the child
#                 to the parent process to communcate status and retrieve
#                 information. All returning messages must have a resource key
#                 and an index.
#             list
#                 The original resource list references. This list will be
#                 updated acordingly.
# Outputs:    Nothing.
# Exceptions: No Error handling implemented
# TODO:       Error handling.
#==============================================================================
sub _update_resource_list {
    my ($self,%args) = @_;
    $Log->enter();
    #Locals
    my $message = $args{message};
    my $index = $message->{'rsrc_index'};
    # FLAG that will let the caller know if this resource needs to be stored
    # (same as knowing if it was newly created).
    my $last_rsrc_index = undef;

    # IF MESSAGE has a RESOURCE, this message comes from a subproc that did
    # worked a resource, so store the info
    if (defined($message->{'resources'}) && defined($index) ) {
        foreach my $resource (@{$message->{'resources'}}) {
            # Save the index if it was NEWLY created.
            if ((not $resource->{_reused}) and ($resource->{_setup_done})){
                $last_rsrc_index = $index;
            }
            # Update the values of the original reference.
            my $rsrc_ref = $args{list}->[$index++];
            foreach my $key (keys %$resource){
                $rsrc_ref->{$key} = $resource->{$key};
            }
        }
    }


    $Log->exit();
    return $last_rsrc_index;
}    ## _results_array

#==============================================================================
# Method:    _parallel_execution_wait
# Objective:  Extract of the logic inside the while loop on parallel_execution.
#             Holds the logic to be used when the message is of type "waiting".
# Details:   Steps: 1) Update the resources list with the information returned
#                      from the child process, via a subroutine.
#                   2.1) Fetch the next resource list or bucket to the requesting
#                        child process.
#                   2.2) Or sending the exit signal to the child process.
# Inputs:   message
#               The NATE::STAFF:Peer message that is send from the child
#               to the parent process to communcate status and retrieve
#               information. All returning messages must have a resource key
#               and an index.
#           list
#               The original resource list references. This list will be updated
#               accordingly.
#           last_rsrcs
#                The reference to a hash that holds the indexes of the last
#                created resources.
#            rsrc_index
#                The index of the array of the resource that will be fetch.
#            bucket_index
#                The factor of incrementation that will be applied to the resrouce
#                index to know what resource to fetch.
#            max_rsrc_index
#                The max value of the index for the list array.
#            action
#                The action that we are performing (Create, Delete)
#            handle
#                The id of the child process this message belongs to.
#            handle_name
#                The name of the child process this message belongs to
#            force_exit
#                Flag to send the exit message to any child process
# Outputs:  queue_args
#                Hash containing the information for the message to the child
#            rsrc_index
#                Returning the an updated index of the next resource to fetch
# Exceptions: No Error handling implemented
# TODO:       Error handling.
#==============================================================================
sub _parallel_execution_wait {
    my ($self, %args) = @_;
    $Log->enter();
    # Locals
    my $EnGenX_obj = $self->{'EnGenX'};
    my $last_rsrcs = $args{last_rsrcs};
    my $rsrc_index = $args{rsrc_index};
    my $bucket_index = $args{bucket_index};
    my $max_rsrc_index = $args{max_rsrc_index};
    my $action = $args{action};
    my $handle = $args{handle};
    my $handle_name = $args{handle_name};
    my $force_exit = $args{force_exit};
    my $resource = $args{message}->{'resources'}->[0];
    # Return Var
    my %response = ( queue_args => {type => "WhatToDo"},
                     rsrc_index => $rsrc_index,
                );

    # UPDATE THE RESOURCE LIST, IF ANY WORK WAS DONE
    my $last_rsrc_index = $self->_update_resource_list(message => $args{message},
                                                       list    => $args{list},
                                                    );
    # STORE THE LAST RESOURCE INDEX with THE HANDLE (PROCESS ID) AS A
    # THE KEY TO REMEMBER WHO WORKED THE RESOURCE
    $last_rsrcs->{$handle} = $last_rsrc_index if( defined($last_rsrc_index));

    # Update the used ips hash if this was a lif reasoource
    if ($handle_name =~ /Lif_Create/) {

        if($resource->{_setup_done} and
           not $resource->{_reused} and
           not $args{message}->{wait_completion}) {
            my $used_address = $resource->{address};
            if (length $used_address and $used_address =~ /:/) {
                $EnGenX_obj->{used_ips}->{ipv6}->{$used_address} = 1;
            } elsif (length $used_address) {
                $EnGenX_obj->{used_ips}->{ipv4}->{$used_address} = 1;
            }
        } elsif($resource->{_setup_done} and $resource->{_reused}) {
            my $used_address = $resource->{address};
            if (length $used_address and $used_address =~ /:/) {
                delete $EnGenX_obj->{used_ips}->{ipv6}->{$used_address};
            } elsif (length $used_address) {
                delete $EnGenX_obj->{used_ips}->{ipv4}->{$used_address};
            }
        }
    }
    # NOW IF THE RESOURCE LIST STILL HAS RESOURCES TO WORK WITH AND THE
    # FORCE EXIT FLAG HAS not BEEN SET
    if (($rsrc_index <= $max_rsrc_index) && !$force_exit) {
        $Log->trace("SubProcess::_parallel_execution_wait: "
                      ."Sending NEXT available resource to SUB-PROCESS: $handle"
                    );
        # IN ORDER TO GET A SUB-ARRAY FROM THE ORIGINAL LIST, NEED TO KNOW
        # WHERE TO START THE SUB-ARRAY (INDEX) AND WHERE TO FINISH IT
        # (NEXT_INDEX)
        my $next_rsrc_index = $rsrc_index + $bucket_index;
        # IF NEXT_INDEX IS BIGGER THAN THE MAX_INDEX, USE THE LATTER INSTEAD
        $next_rsrc_index = ($next_rsrc_index>=$max_rsrc_index) ?
                                            $max_rsrc_index: $next_rsrc_index;
        # GET THE ARRAY OF RESOURCE(S)
        my @rsrcs = @{$args{list}}[$rsrc_index..$next_rsrc_index];
        $Log->trace("SubProcess::_parallel_execution_wait: "
                      ."ARRAY OF RESOURCES[$rsrc_index] :\n" . Dumper(@rsrcs)
                    );

        # CREATE A HASH THAT HOLDS THE RESOURCE INDED AND RESOURCE OBJECT
        my $rsrc_mssg = {rsrc_index => $rsrc_index,
                         resources  => \@rsrcs
                        };
        # NEED TO INCREMENT THE INDEX TO FETCH THE NEXT RESOURCE OR GROUP OF
        # RESOURCES
        $response{rsrc_index} = $next_rsrc_index + 1;

        # SAVE THE MESSAGE
        $response{queue_args}{message} = $rsrc_mssg;
    }
    # THERE ARE NO MORE RESORCES TO WORK WITH, WE NEED TO EXIT
    else {
        # IF THIS IS A CREATION CALL MAKE THE SUBPROC WAIT FOR
        # COMPLETION OF THE LAST NEWLY CREATED RESOURCE AND THEN EXIT.
        if( ($action =~ /Create/i) and ($last_rsrcs->{$handle}) ) {
            $Log->trace("SubProcess::_parallel_execution_wait: "
                          ."Sending WAIT_COMPLETION cmd to SUB-PROCESS: $handle");
            # ARGS TO WAIT, GET THE RESOURCE FROM THE LIST WITH THE
            # INDEX THAT THE TABLE IS HOLDING. IF
            my $idx = delete($last_rsrcs->{$handle});
            my $wait_mssg = {wait_completion => TRUE,
                             resources => [$args{list}->[$idx]],
                            };

            # SAVE THE MESSAGE
            $response{queue_args}{message} = $wait_mssg;
        }
        # FORCE EXIT OR SUBPROCES CALL IS DELETE
        else {
            # STOPPING THE PROCESS
            $Log->trace("SubProcess::_parallel_execution_wait: "
                        ."Sending EXIT cmd to SUB-PROCESS: $handle");
            # SAVE THE MESSAGE
            $response{queue_args}{message} = "exit";
        }
    } # compete_list AND force_exit

    $Log->exit();
    return (wantarray()) ? %response : \%response;
}## _parallel_execution_wait

#==============================================================================
# Method:     _parallel_execution_ips
# Objective:  Extract of the logic inside the while loop on parallel_execution.
#             Holds the logic to be used when the message is of type "ip_request".
# Details:    Steps: 1) Reads the message to get the ip_range to be used.
#                    2) Get the new (unused) ip address from the stack
#                     3) Adds the new ip address to the used list.
# Inputs:   message
#                 The NATE::STAFF:Peer message that is send from the child
#                 to the parent process to communcate status and retrieve
#                 information. All returning messages must have a resource key
#                 and an index.
#            handle
#                The id of the child process this message belongs to.
#            handle_name
#                The name of the child process this message belongs to
# Outputs:  queue_args
#                Hash containing the information for the message to the child.
#               The ip_address.
# Exceptions: No Error handling implemented
# TODO:       Error handling.
#==============================================================================
sub _parallel_execution_ips {
    my ($self, %args) = @_;
    $Log->enter();
    # Locals
    my $message = $args{message};
    my $handle_name = $args{handle_name};
    my $exception = '';
    # Return Var
    my $queue_post_args = {type => "ip_address"};
    my $EnGenX_obj = $self->{'EnGenX'};
    my $hex_regex = "[A-Fa-f0-9]";
    my $old_address = $message->{old_address};

    if (length $old_address and $old_address =~ /:/) {
        $EnGenX_obj->{used_ips}->{ipv6}->{$old_address} = 1;
    } elsif (length $old_address) {
        $EnGenX_obj->{used_ips}->{ipv4}->{$old_address} = 1;
    }
    # Subprocess is asking for an address from the ip range that is contained
    # inside the resource
    if (defined($message->{'resources'})) {
        my $ip=undef;
        my $lif_rsrc = $message->{'resources'};
        # Ensure that there is an address range in the resource
        if( defined($lif_rsrc->{ips})) {
            # NOW THAT THE THE ARRAY IS IN THE STACK, JUST GET AN ADDRESS
            my $ip_mssg = undef;
            # Look at the address ranges and find an ip that is not use in
            # used_ips. If all the ips in the range have been used then
            # return no ip.
            my @address_ranges = @{$lif_rsrc->{ips}};

            # Loop through all the ranges passed in to find a match
OUT:        foreach my $curr_ip_range (@address_ranges) {
                if (not(defined($curr_ip_range) and length($curr_ip_range))) {
                    next;
                }
                # Check to see whether ipv6
                if ($curr_ip_range =~ /:/) {
                    # Split the address into chunks
                    my @chunks = split(/:/, $curr_ip_range);

                    # There should be 8 chunks since we normalize to x:x:x:x:x:x:x:x
                    # We need to create a hash that stores the information of the
                    # 8 chunks incase we have ranges.
                    my %chunk_hash;
                    # Loop through all the chunks and store the information
                    for (my $i = 0; $i<8; $i++) {
                        if ($chunks[$i] =~ /\[($hex_regex+)-($hex_regex+)\]/) {
                            # This is a range so store the base and max.
                            $chunk_hash{"chunk$i"}->{'base'}  = $1;
                            $chunk_hash{"chunk$i"}->{'max'}   = $2;
                        } else {
                            # This is NOT a range so the base and max are the same.
                            $chunk_hash{"chunk$i"}->{'base'}  = $chunks[$i];
                            $chunk_hash{"chunk$i"}->{'max'}   = $chunks[$i];
                        }

                        # Set the current part of the range to base.
                        $chunk_hash{"chunk$i"}->{'current'}  = $chunk_hash{"chunk$i"}->{'base'};
                    }

                    # This loop will check the addresses within the range to find one
                    # that is free to return.
                    while (1) {
                        # Create the address from the chunk_hash
                        my $addr = $chunk_hash{"chunk0"}->{'current'} . ":" .
                                   $chunk_hash{"chunk1"}->{'current'} . ":" .
                                   $chunk_hash{"chunk2"}->{'current'} . ":" .
                                   $chunk_hash{"chunk3"}->{'current'} . ":" .
                                   $chunk_hash{"chunk4"}->{'current'} . ":" .
                                   $chunk_hash{"chunk5"}->{'current'} . ":" .
                                   $chunk_hash{"chunk6"}->{'current'} . ":" .
                                   $chunk_hash{"chunk7"}->{'current'};

                        # Check if the address is used.
                        # return is of the format
                        # (return_value    =>TRUE,
                        # containing_ip_range => $ip_range)
                        my %return = $self->_is_ipv6_address_used(ip => $addr,
                                                                  EnGenX => $EnGenX_obj);

                        # If it is not used then we found a free address
                        if (not ($return{return_value})) {
                            $ip = $addr;
                            last OUT;
                        }

                        # If we get to here then the address was used. Lets
                        # update the chunk_hash currents based on the highest
                        # used ip from the range we found it in.
                        # Split the ip into chunks.
                        my @containing_ip_range_chunks = split(/:/, $return{containing_ip_range});

                        # Variable to tell whether next chunk needs to have 1 added based on whether
                        # the chunk before was reset. Start out as 1 because last chunk always needs
                        # to add one to get next ip.
                        my $add_one = 1;

                        # Variable to tell which chunks were 2:reset, 1:addone, or 0:stayed the same.
                        my %action_hash = ();

                        # We need to create a hash that stores the information of the
                        # 4 chunks incase we have ranges.
                        my %used_chunk_hash;

                        # Loop through all the chunks and store the information
                        for (my $i = 0; $i<8; $i++) {
                            if ($containing_ip_range_chunks[$i] =~ /\[($hex_regex+)-($hex_regex+)\]/) {
                                # This is a range so store the base and max.
                                $used_chunk_hash{"chunk$i"}->{'base'}  = $1;
                                $used_chunk_hash{"chunk$i"}->{'max'}   = $2;
                            } else {
                                # This is NOT a range so the base and max are the same.
                                $used_chunk_hash{"chunk$i"}->{'base'}  = $containing_ip_range_chunks[$i];
                                $used_chunk_hash{"chunk$i"}->{'max'}   = $containing_ip_range_chunks[$i];
                            }
                        }

                        # Update each of the chunks current values starting with the last
                        for (my $i=7; $i>=0; $i--) {
                            # If add_one is 0 then we know that a previous chunk was updated. We know that
                            # because the used_chunk_range always contains the current ip that if add_one is
                            # zero then rest of the chunks should stay the same.
                            if($add_one == 0) {
                                # Need to update the action_hash based on add_one and
                                # set add_one to 0
                                $action_hash{$i} = $add_one;
                            }
                            # elsif add_one = 1 and the used chunk max + $add_one is greater than the ip
                            # chunk max then we need to reset to base and leave $add_one = 1 to the next chunk
                            elsif(hex($used_chunk_hash{"chunk$i"}->{'max'})+$add_one > hex($chunk_hash{"chunk$i"}->{'max'})) {
                                # Need to update the action_hash with reset and set addone to 1
                                $action_hash{$i} = 2;
                            }
                            # else add_one = 1 and the updated used_ip was within the range so set add_one to zero and
                            # set the action hash to add_one.
                            else {
                                # Need to update the action_hash based on add_one and
                                # set add_one to 0
                                $action_hash{$i} = $add_one;
                                $add_one = 0;
                            }
                        }

                        # A variable to tell whether the rest should be base values.
                        # This occurs when a value is add_one because to add 1 the values
                        # before had to be used up. So all the rest should be bases.
                        my $the_rest_are_base = FALSE;

                        # Loop through and set the current values for the ip
                        for (my $i=0; $i<8; $i++) {
                            if($the_rest_are_base or $action_hash{$i} == 2) {
                                # If rest are base or reset
                                # This range has been exhausted if the first chunk is reset
                                next OUT if($i == 0);

                                # Otherwise set this value to a base because it was reset.
                                $chunk_hash{"chunk$i"}->{'current'} = $chunk_hash{"chunk$i"}->{'base'};
                            }
                            elsif($action_hash{$i} == 0) {
                                # If action is 0
                                # This chunk was not updated so leave it what it was.
                            } elsif($action_hash{$i} == 1) {
                                # If add_one
                                # This was one above the value in containing_ip_range_chunks.
                                $chunk_hash{"chunk$i"}->{'current'} = sprintf("%.4x", (hex($used_chunk_hash{"chunk$i"}->{'max'})+1));
                                # Set the rest are base since this value was updated. we want to start at the bottom of
                                # the new values range.
                                $the_rest_are_base = TRUE;
                            }
                        }
                    }
                } else { # This is an ipv4 address range
                       # Split the address into chunks
                    my @chunks = split(/\./, $curr_ip_range);

                    # There should be 4 chunks x.x.x.x
                    # We need to create a hash that stores the information of the
                    # 4 chunks incase we have ranges.
                    my %chunk_hash;
                    # Loop through all the chunks and store the information
                    for (my $i = 0; $i<4; $i++) {
                        if ($chunks[$i] =~ /\[(\d+)-(\d+)\]/) {
                            # This is a range so store the base and max.
                            $chunk_hash{"chunk$i"}->{'base'}  = $1;
                            $chunk_hash{"chunk$i"}->{'max'}   = $2;
                        } else {
                            # This is NOT a range so the base and max are the same.
                            $chunk_hash{"chunk$i"}->{'base'}  = $chunks[$i];
                            $chunk_hash{"chunk$i"}->{'max'}   = $chunks[$i];
                        }

                        # Set the current part of the range to base.
                        $chunk_hash{"chunk$i"}->{'current'}  = $chunk_hash{"chunk$i"}->{'base'};
                    }

                    # This loop will check the addresses within the range to find one
                    # that is free to return.
                    while (1) {
                        # Create the address from the chunk_hash
                        my $addr = $chunk_hash{"chunk0"}->{'current'} . "." .
                                   $chunk_hash{"chunk1"}->{'current'} . "." .
                                   $chunk_hash{"chunk2"}->{'current'} . "." .
                                   $chunk_hash{"chunk3"}->{'current'};

                        # Check if the address is used.
                        # return is of the format
                        # (return_value    =>TRUE,
                        # containing_ip_range => $ip_range)
                        my %return = $self->_is_ipv4_address_used(ip     => $addr,
                                                                  EnGenX => $EnGenX_obj);

                        # If it is not used then we found a free address
                        if (not ($return{return_value})) {
                            $ip = $addr;
                            last OUT;
                        }

                        # If we get to here then the address was used. Lets
                        # update the chunk_hash currents based on the highest
                        # used ip from the range we found it in.
                        # Split the ip into chunks.
                        my @containing_ip_range_chunks = split(/\./, $return{containing_ip_range});

                        # Variable to tell whether next chunk needs to have 1 added based on whether
                        # the chunk before was reset. Start out as 1 because last chunk always needs
                        # to add one to get next ip.
                        my $add_one = 1;

                        # Variable to tell which chunks were 2:reset, 1:addone, or 0:stayed the same.
                        my %action_hash = ();

                        # We need to create a hash that stores the information of the
                        # 4 chunks incase we have ranges.
                        my %used_chunk_hash;

                        # Loop through all the chunks and store the information
                        for (my $i = 0; $i<4; $i++) {
                            if ($containing_ip_range_chunks[$i] =~ /\[(\d+)-(\d+)\]/) {
                                # This is a range so store the base and max.
                                $used_chunk_hash{"chunk$i"}->{'base'}  = $1;
                                $used_chunk_hash{"chunk$i"}->{'max'}   = $2;
                            } else {
                                # This is NOT a range so the base and max are the same.
                                $used_chunk_hash{"chunk$i"}->{'base'}  = $containing_ip_range_chunks[$i];
                                $used_chunk_hash{"chunk$i"}->{'max'}   = $containing_ip_range_chunks[$i];
                            }
                        }

                        # Update each of the chunks current values starting with the last
                        for (my $i=3; $i>=0; $i--) {
                            # If add_one is 0 then we know that a previous chunk was updated. We know that
                            # because the used_chunk_range always contains the current ip that if add_one is
                            # zero then rest of the chunks should stay the same.
                            if($add_one == 0) {
                                # Need to update the action_hash based on add_one and
                                # set add_one to 0
                                $action_hash{$i} = $add_one;
                            }
                            # elsif add_one = 1 and the used chunk max + $add_one is greater than the ip
                            # chunk max then we need to reset to base and leave $add_one = 1 to the next chunk
                            elsif($used_chunk_hash{"chunk$i"}->{'max'}+$add_one > $chunk_hash{"chunk$i"}->{'max'}) {
                                # Need to update the action_hash with reset and set addone to 1
                                $action_hash{$i} = 2;
                            }
                            # else add_one = 1 and the updated used_ip was within the range so set add_one to zero and
                            # set the action hash to add_one.
                            else {
                                # Need to update the action_hash based on add_one and
                                # set add_one to 0
                                $action_hash{$i} = $add_one;
                                $add_one = 0;
                            }
                        }

                        # A variable to tell whether the rest should be base values.
                        # This occurs when a value is add_one because to add 1 the values
                        # before had to be used up. So all the rest should be bases.
                        my $the_rest_are_base = FALSE;

                        # Loop through and set the current values for the ip.
                        for (my $i=0; $i<4; $i++) {
                            if($the_rest_are_base or $action_hash{$i} == 2) {
                                # If rest are base or reset
                                # This range has been exhausted if the first chunk is reset
                                next OUT if($i == 0);

                                # Otherwise set this value to a base because it was reset.
                                $chunk_hash{"chunk$i"}->{'current'} = $chunk_hash{"chunk$i"}->{'base'};
                            }
                            elsif($action_hash{$i} == 0) {
                                # If action is 0
                                # This chunk was not updated so leave it what it was.
                            } elsif($action_hash{$i} == 1) {
                                # If add_one
                                # This was one above the value in containing_ip_range_chunks.
                                $chunk_hash{"chunk$i"}->{'current'} = $used_chunk_hash{"chunk$i"}->{'max'} +1;
                                # Set the rest are base since
                                $the_rest_are_base = TRUE;
                            }
                        }
                    }
                }
            }
            # We found an unused ip address.
            if (defined $ip) {
                # We found an unused up address so create a return message
                $lif_rsrc->{ADDRESS} = $ip;
                $queue_post_args->{message} = [$lif_rsrc];
            } # End if(defined $ip)
            else {
                # Check to see if there are any unverified ips. If there are then
                # tell the process to wait because one of the ips could still
                # be free.
                foreach my $value (values(%{$EnGenX_obj->{used_ips}->{ipv6}})){
                    if (not ($value)) {
                        $queue_post_args->{message} = "wait";
                        $Log->exit();
                        return (wantarray()) ? %{$queue_post_args} : $queue_post_args;
                    }
                }
                foreach my $value (values(%{$EnGenX_obj->{used_ips}->{ipv4}})){
                    if (not ($value)) {
                        $queue_post_args->{message} = "wait";
                        $Log->exit();
                        return (wantarray()) ? %{$queue_post_args} : $queue_post_args;
                    }
                }
                # If no unused ips could be found in the address range()s provided.
                # Then set the message to the error message.
                $queue_post_args->{message} = "ERROR: Could not find an "
                                             ."IP address for LIF: "
                                             .$lif_rsrc->{NAME} .", on VSERVER: "
                                             .$lif_rsrc->{VSERVER}.", on SUBPROCESS: "
                                             .$handle_name. "\nAll IP addresses from "
                                             ."the provided IP Range have been used." .
                                             Dumper(\@address_ranges) . "\nThe ips this" .
                                             " EnGenX object has used are " .
                                             Dumper($EnGenX_obj->{used_ips});
            } # End else
        } else {
            $exception = "An IP range has not been provide for resource " . Dumper($lif_rsrc);
            $Log->exit();
            NATE::BaseException->throw(
                "Exception caught at SubProcess::parallel_execution_ips:\n $exception");
        }
    } else {
        $exception = "Message has no resource value in it to work with:" . Dumper($message);
        $Log->exit();
        NATE::BaseException->throw(
                "Exception caught at SubProcess::parallel_execution_ips:\n $exception");
    }
    $Log->exit();
    return (wantarray()) ? %{$queue_post_args} : $queue_post_args;
}

#==============================================================================
# Method:     _is_ipv6_address_used
# Objective:  Check whether the ip passed is used or not.
# Details:    Steps: 1) Check whether ip is used.
#                    2) If not add to used list
#                    3) Merge used list ips.
# Inputs:     ip
#                The ip to check.
#             EnGenX
#                The EnGenX object that has the used ips
# Outputs:    Hash with
#                 return_value => whether the ip was used
#                 containing_ip_range => an ip or ip range to determine
#                                        the next ip to try.
# Exceptions: No Error handling implemented
#===============================================================================
sub _is_ipv6_address_used {
    my ($self, @opts) = @_;
    $Log->enter();

    my %args = ();
    # Validate Arguments
    try {
        %args = Params::Validate::validate(@opts, {
            ip         => {type=>SCALAR, optional => FALSE},
            EnGenX     => {type=>OBJECT, optional => FALSE},
        });
    }
    otherwise {
        my $exception = shift;
        NATE::Exceptions::Argument->throw(
                "SubProcess::_is_ipv6_address_used parameters errror :\n" .
                $exception->text());
    };
    # Get the ip pass in.
    my $ip = $args{ip};

    # Store EnGenX object so that we have access to the used_ips hash
    my $EnGenX_obj = $args{EnGenX};

    # Store a hex regex so we can easily use it to find ranges last
    my $hex_regex = "[A-Fa-f0-9]";

    # If there is nothing defined in the ipv6 range then we know that the
    # passed in ip has not been used. Add it to the used ips and return.
    if (not (scalar(keys %{$EnGenX_obj->{used_ips}->{ipv6}}))) {
        $EnGenX_obj->{used_ips}->{ipv6}->{$ip} = '0';
        return (return_value    =>FALSE);
    }

    # Split the ip into chunks, This is a single ip with no ranges.
    my @ip_chunks = split(/:/, $ip);

    # Need to loop through all the used ips so that we can see if the ip
    # passed in has already been used.
TOP:foreach my $ip_range (keys %{$EnGenX_obj->{used_ips}->{ipv6}}) {
        my @used_ip_chunks = split(/:/, $ip_range);
        # There should be 8 chunks because the ips are normalized into x:x:x:x:x:x:x:x

        # We want to loop through and compare the eight to see if we have a match.
        # Any of the 8 could be a range
        for (my $i = 0; $i<8; $i++) {
            my $range_base = undef;
            my $range_max  = undef;
            # Check to see if a range or a single chunk
            if ($used_ip_chunks[$i] =~ /\[($hex_regex+)-($hex_regex+)\]/) {
                # This is a ranged chunk
                $range_base = $1;
                $range_max  = $2;
            } else {
                # This is a single chunk
                $range_base = $used_ip_chunks[$i];
                $range_max  = $used_ip_chunks[$i];
            }

            # If the passed in ip chunk is within the range of this chunk continue
            # otherwise look at the next range.
            if (not (hex($ip_chunks[$i]) >= hex($range_base) and hex($ip_chunks[$i]) <= hex($range_max))) {
                # The ip passed in was not withing the current range so go to
                # the next range.
                next TOP;
            }
        }
        # If we get to here then we found that the ip was used
        # First check to see if there are ips right above the this ip
        # that have not been merged. This is to reduce the overhead of
        # checking ips since when in parallel, x number of
        # unmergable single ips are possible. Where x is the number of
        # of parallel processes
        while(1) {
            # Dont want to compare ranges because this will have same loss
            # as calling this method again. If it is not a range then we check
            # if the last octet + 1 exists and if so then set that as the ip
            if (not ($ip_range =~ /\-/)) {
                # Split the ip into chunks
                @used_ip_chunks = split(/\:/, $ip_range);

                # Build the ip to check in the used ip hash.
                my $compare_ip = $used_ip_chunks[0] . ':' .
                                 $used_ip_chunks[1] . ':' .
                                 $used_ip_chunks[2] . ':' .
                                 $used_ip_chunks[3] . ':' .
                                 $used_ip_chunks[4] . ':' .
                                 $used_ip_chunks[5] . ':' .
                                 $used_ip_chunks[6] . ':' .
                                 sprintf("%.4x", (hex($used_ip_chunks[7]) + 1));

                # If the compare_ip exists the set it do the ip_range so that
                # We can check if there is an ip 1 above that.
                if (exists $EnGenX_obj->{used_ips}->{ipv6}->{$compare_ip}) {
                    $ip_range = $compare_ip;
                } else {
                    return (return_value    =>TRUE,
                            containing_ip_range => $ip_range);
                }
            } else {
                return (return_value    =>TRUE,
                        containing_ip_range => $ip_range);
            }
        }
    }

    # If we get to this point then the ip is new. We need
    # to add it to the used ips before.
    $EnGenX_obj->{used_ips}->{ipv6}->{$ip} = '0';

    # Now loop through and merge the used ips into ranges
    # For now just going to do last chunk ranges.
    my @sorted_used_ips = (sort (keys %{$EnGenX_obj->{used_ips}->{ipv6}}));

    for(my $j = 0; $j < scalar(@sorted_used_ips); $j++) {
        my $temp1_ip = $sorted_used_ips[$j];
        next if (not ($EnGenX_obj->{used_ips}->{ipv6}->{$temp1_ip}));
OUT:    for(my $k = 0; $k < scalar(@sorted_used_ips); $k++) {
            my $temp2_ip = $sorted_used_ips[$k];
            next if (not ($EnGenX_obj->{used_ips}->{ipv6}->{$temp2_ip}));
            # We need to have 2 different ips to compare and merge so if the
            # ips are the same then skip this ip. Since we are pulling from
            # the same array should be able to string compare the two.
            if ($temp1_ip eq $temp2_ip) {
                next OUT;
            }

            # Split both ips into chunks
            my @temp1_chunks = split(/:/, $temp1_ip);
            my @temp2_chunks = split(/:/, $temp2_ip);

            # Variable to tell whether the rest of the chunks should match and
            # not be 1 above. We merge based on one chunk being mergable.
            # Example 1:1:1:1:1:1:1:1-10 merges with 1:1:1:1:1:1:1:11-20
            #         to 1:1:1:1:1:1:1:1-20
            #         1:1:1:1:1:1:1:1-10 merges with 1:1:1:1:1:1:2:1-10
            #         to 1:1:1:1:1:1:1-2:1-10
            #         1:1:1:1:1:1:1-2:1-10 merges with 1:1:1:1:1:1:1-2:11-20
            #         to 1:1:1:1:1:1:1-2:1-20
            #         1:1:1:1:1:1:1:1-10 does not merge with 1:1:1:1:1:1:2:1-9
            my $rest_should_match = FALSE;

            # A variable to hold the new ip
            my $new_ip = '';

            # There should be 8 chunks because the ips x:x:x:x:x:x:x:x
            # We want to loop through and compare the 8 to see if we have a merge
            for (my $i = 0; $i<8; $i++) {
                # Check to see if the chunk of 1 is range or a single address
                my $temp1_base;
                my $temp1_max;
                # If it is a range store the base and max of the range.
                if ($temp1_chunks[$i] =~ /\[($hex_regex+)-($hex_regex+)\]/) {
                    $temp1_base = $1;
                    $temp1_max  = $2;
                } else {
                    # This is a single address so store the address as base and max
                    $temp1_base = $temp1_chunks[$i];
                    $temp1_max  = $temp1_chunks[$i];
                }

                # Check to see if the chunk of 2 is range or a single address
                my $temp2_base;
                my $temp2_max;
                # If it is a range store the base and max of the range.
                if ($temp2_chunks[$i] =~ /\[($hex_regex+)-($hex_regex+)\]/) {
                    $temp2_base = $1;
                    $temp2_max  = $2;
                } else {
                    # This is a single address so store the address as base and max
                    $temp2_base = $temp2_chunks[$i];
                    $temp2_max  = $temp2_chunks[$i];
                }

                # Setup parameter for making the new range.
                my $new_base = undef;
                my $new_max  = undef;

                if (hex($temp1_base) == hex($temp2_base) and hex($temp1_max) == hex($temp2_max)) {
                    # First check if these ranges are matching ranges
                    # Then update the new_base and new_max
                    $new_base = $temp1_base;
                    $new_max  = $temp1_max;
                } elsif ($rest_should_match) {
                    # The rest should match but didn't if we get here
                    # so these ranges cannot be merged. Go to the next ranges.
                    next OUT;
                } elsif ((hex($temp1_base)-1) == hex($temp2_max)) {
                    # Second check if temp1 range is 1 above temp2 range
                    #If yes then we need to merge the two ranges or addresses.

                    # Update the new_base and new_max
                    $new_base = $temp2_base;
                    $new_max  = $temp1_max;

                    # Update $rest_should_match because we can only merge
                    # one different range chunk.
                    $rest_should_match = TRUE;
                } elsif ((hex($temp2_base)-1) == hex($temp1_max)) {
                    # Third Check if temp2 range is 1 above temp1 range
                    #If yes then we need to merge the two ranges or addresses.

                    # Update the new_base and new_max
                    $new_base = $temp1_base;
                    $new_max  = $temp2_max;

                    # Update $rest_should_match because we can only merge
                    # one different range chunk.
                    $rest_should_match = TRUE;
                } else {
                    # If the ranges dont match and are not mergable then go to next ip
                    # to merge
                    next OUT;
                }
                # If we get to here then we should have a new_base and new_max
                # to add to the $new_ip_range we are merging
                if (hex($new_base) == hex($new_max)) {
                    # If they are equal then this is a single chunk
                    $new_ip = $new_ip . $new_base . ":";
                } else {
                    # else they are a range
                    $new_ip = $new_ip . "[" . $new_base .
                                        "-" . $new_max  .
                                        "]:";
                }
            }

            # Once we get here we have a new merged ip range
            # Remove the last . that we added
            $new_ip =~ s/:$//;

            # First remove the old ips first
            delete $EnGenX_obj->{used_ips}->{ipv6}->{$temp1_ip};
            delete $EnGenX_obj->{used_ips}->{ipv6}->{$temp2_ip};

            $EnGenX_obj->{used_ips}->{ipv6}->{$new_ip} = '1';

            # We merge to ranges so update the sorted_used_ips and start the
            # loops over. When we dont merge any range then we are done.
            @sorted_used_ips = (sort (keys %{$EnGenX_obj->{used_ips}->{ipv6}}));
            # setting to -1 so that when it gets updated by the for loop it
            # goes to 0;
            $j = -1;
            $k = 0;
            # Need to continue start the temp1_ip over again.
            last OUT;
        }
    }

    $Log->exit();
    return (return_value => FALSE);
}

#==============================================================================
# Method:     _is_ipv4_address_used
# Objective:  Check whether the ip passed is used or not.
# Details:    Steps: 1) Check whether ip is used.
#                    2) If not add to used list
#                    3) Merge used list ips.
# Inputs:     ip
#                The ip to check.
#             EnGenX
#                The EnGenX object that has the used ips
# Outputs:    Hash with
#                 return_value => whether the ip was used
#                 containing_ip_range => an ip or ip range to determine
#                                        the next ip to try.
# Exceptions: No Error handling implemented
#===============================================================================
sub _is_ipv4_address_used {
    my ($self, @opts) = @_;
    $Log->enter();

    my %args = ();
    # Validate Arguments
    try {
        %args = Params::Validate::validate(@opts, {
            ip         => {type=>SCALAR, optional => FALSE},
            EnGenX     => {type=>OBJECT, optional => FALSE},
        });
    }
    otherwise {
        my $exception = shift;
        NATE::Exceptions::Argument->throw(
                "SubProcess::_is_ipv4_address_used parameters errror :\n" .
                $exception->text());
    };
    # Get the ip pass in.
    my $ip = $args{ip};

    # Store EnGenX object so that we have access to the used_ips hash
    my $EnGenX_obj = $args{EnGenX};

    # If there is nothing defined in the ipv4 range then we know that the
    # passed in ip has not been used. Add it to the used ips and return.
    if (not (scalar(keys %{$EnGenX_obj->{used_ips}->{ipv4}}))) {
        $EnGenX_obj->{used_ips}->{ipv4}->{$ip} = '0';
        return (return_value    =>FALSE);
    }

    # Split the ip into chunks.
    my @ip_chunks = split(/\./, $ip);

    # Need to loop through all the used ips so that we can see if the ip
    # passed in has already been used.
TOP:foreach my $ip_range (keys %{$EnGenX_obj->{used_ips}->{ipv4}}) {
        my @used_ip_chunks = split(/\./, $ip_range);
        # There should be 4 chunks because the ips x.x.x.x

        # We want to loop through and compare the four to see if we have a match.
        # Any of the 4 could be a range
        for (my $i = 0; $i<4; $i++) {
            my $range_base = undef;
            my $range_max  = undef;
            # Check to see if a range or a single chunk
            if ($used_ip_chunks[$i] =~ /\[(\d+)-(\d+)\]/) {
                # This is a ranged chunk
                $range_base = $1;
                $range_max  = $2;
            } else {
                # This is a single chunk
                $range_base = $used_ip_chunks[$i];
                $range_max  = $used_ip_chunks[$i];
            }

            # If the passed in ip chunk is within the range of this chunk continue
            # otherwise look at the next range.
            if (not ($ip_chunks[$i] >= $range_base and $ip_chunks[$i] <= $range_max)) {
                # The ip passed in was not withing the current range so go to
                # the next range.
                next TOP;
            }
        }
        # If we get to here then we found that the ip was used
        # First check to see if there are ips right above the this ip
        # that have not been merged. This is to reduce the overhead of
        # checking ips since when in parallel, x number of
        # unmergable single ips are possible. Where x is the number of
        # of parallel processes
        while(1) {
            # Dont want to compare ranges because this will have same loss
            # as calling this method again. If it is not a range then we check
            # if the last octet + 1 exists and if so then set that as the ip
            if (not ($ip_range =~ /\-/)) {
                # Split the ip into chunks
                @used_ip_chunks = split(/\./, $ip_range);

                # Build the ip to check in the used ip hash.
                my $compare_ip = $used_ip_chunks[0] . '.' .
                                 $used_ip_chunks[1] . '.' .
                                 $used_ip_chunks[2] . '.' .
                                 ($used_ip_chunks[3] + 1);

                # If the compare_ip exists the set it do the ip_range so that
                # We can check if there is an ip 1 above that.
                if (exists $EnGenX_obj->{used_ips}->{ipv4}->{$compare_ip}) {
                    $ip_range = $compare_ip;
                } else {
                    return (return_value    =>TRUE,
                            containing_ip_range => $ip_range);
                }
            } else {
                return (return_value    =>TRUE,
                        containing_ip_range => $ip_range);
            }
        }
    }

    # If we get to this point then the ip is new. We need
    # to add it to the used ips before.
    $EnGenX_obj->{used_ips}->{ipv4}->{$ip} = '0';

    # Now loop through and merge the used ips into ranges
    # For now just going to do last chunk ranges.
    my @sorted_used_ips = (sort (keys %{$EnGenX_obj->{used_ips}->{ipv4}}));

    for(my $j = 0; $j < scalar(@sorted_used_ips); $j++) {
        my $temp1_ip = $sorted_used_ips[$j];
        next if (not ($EnGenX_obj->{used_ips}->{ipv4}->{$temp1_ip}));
OUT:    for(my $k = 0; $k < scalar(@sorted_used_ips); $k++) {
            my $temp2_ip = $sorted_used_ips[$k];
            next if (not ($EnGenX_obj->{used_ips}->{ipv4}->{$temp2_ip}));
            # We need to have 2 different ips to compare and merge so if the
            # ips are the same then skip this ip. Since we are pulling from
            # the same array should be able to string compare the two.
            if ($temp1_ip eq $temp2_ip) {
                next OUT;
            }

            # Split both ips into chunks
            my @temp1_chunks = split(/\./, $temp1_ip);
            my @temp2_chunks = split(/\./, $temp2_ip);

            # Variable to tell whether the rest of the chunks should match and
            # not be 1 above. We merge based on one chunk being mergable.
            # Example 1.1.1.1-10 merges with 1.1.1.11-20 to 1.1.1.1-20
            #         1.1.1.1-10 merges with 1.1.2.1-10 to 1.1.1-2.1-10
            #         1.1.1-2.1-10 merges with 1.1.1-2.11-20 to 1.1.1-2.1-20
            #         1.1.1.1-10 does not merge with 1.1.2.1-9
            my $rest_should_match = FALSE;

            # A variable to hold the new ip
            my $new_ip = '';

            # There should be 4 chunks because the ips x.x.x.x
            # We want to loop through and compare the 4 to see if we have a merge
            for (my $i = 0; $i<4; $i++) {
                # Check to see if the chunk of 1 is range or a single address
                my $temp1_base;
                my $temp1_max;
                # If it is a range store the base and max of the range.
                if ($temp1_chunks[$i] =~ /\[(\d+)-(\d+)\]/) {
                    $temp1_base = $1;
                    $temp1_max  = $2;
                } else {
                    # This is a single address so store the address as base and max
                    $temp1_base = $temp1_chunks[$i];
                    $temp1_max  = $temp1_chunks[$i];
                }

                # Check to see if the chunk of 2 is range or a single address
                my $temp2_base;
                my $temp2_max;
                # If it is a range store the base and max of the range.
                if ($temp2_chunks[$i] =~ /\[(\d+)-(\d+)\]/) {
                    $temp2_base = $1;
                    $temp2_max  = $2;
                } else {
                    # This is a single address so store the address as base and max
                    $temp2_base = $temp2_chunks[$i];
                    $temp2_max  = $temp2_chunks[$i];
                }

                # Setup parameter for making the new range.
                my $new_base = undef;
                my $new_max  = undef;

                if ($temp1_base == $temp2_base and $temp1_max == $temp2_max) {
                    # First check if these ranges are matching ranges
                    # Then update the new_base and new_max
                    $new_base = $temp1_base;
                    $new_max  = $temp1_max;
                } elsif ($rest_should_match) {
                    # The rest should match but didn't if we get here
                    # so these ranges cannot be merged. Go to the next ranges.
                    next OUT;
                } elsif ($temp1_base-1 == $temp2_max) {
                    # Second check if temp1 range is 1 above temp2 range
                    #If yes then we need to merge the two ranges or addresses.

                    # Update the new_base and new_max
                    $new_base = $temp2_base;
                    $new_max  = $temp1_max;

                    # Update $rest_should_match because we can only merge
                    # one different range chunk.
                    $rest_should_match = TRUE;
                } elsif ($temp2_base-1 == $temp1_max) {
                    # Third Check if temp2 range is 1 above temp1 range
                    #If yes then we need to merge the two ranges or addresses.

                    # Update the new_base and new_max
                    $new_base = $temp1_base;
                    $new_max  = $temp2_max;

                    # Update $rest_should_match because we can only merge
                    # one different range chunk.
                    $rest_should_match = TRUE;
                } else {
                    # If the ranges dont match and are not mergable then go to next ip
                    # to merge
                    next OUT;
                }
                # If we get to here then we should have a new_base and new_max
                # to add to the $new_ip_range we are merging
                if ($new_base == $new_max) {
                    # If they are equal then this is a single chunk
                    $new_ip = $new_ip . $new_base . ".";
                } else {
                    # else they are a range
                    $new_ip = $new_ip . "[" . $new_base .
                                        "-" . $new_max  .
                                        "].";
                }
            }

            # Once we get here we have a new merged ip range
            # Remove the last . that we added
            $new_ip =~ s/\.$//;

            # First remove the old ips first
            delete $EnGenX_obj->{used_ips}->{ipv4}->{$temp1_ip};
            delete $EnGenX_obj->{used_ips}->{ipv4}->{$temp2_ip};

            $EnGenX_obj->{used_ips}->{ipv4}->{$new_ip} = '1';

            # We merge to ranges so update the sorted_used_ips and start the
            # loops over. When we dont merge any range then we are done.
            @sorted_used_ips = (sort (keys %{$EnGenX_obj->{used_ips}->{ipv4}}));
            # setting to -1 so that when it gets updated by the for loop it
            # goes to 0;
            $j = -1;
            $k = 0;
            # Need to continue start the temp1_ip over again.
            last OUT;
        }
    }
    $Log->exit();
    return (return_value => FALSE);
}

1;
