# $Id: //depot/prod/test/Rsierranevada/lib/EnGenX/Util.pm#21 $

# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide a utility package for EnGenX libraries
## @author dl-san-automation-core@netapp.com
## @status shared
## @pod here
package EnGenX::Util;

# Compiler directives.
use strict;

# Module import.
use Tharn qw($Log);
use Error;
use Data::Dumper;
use NATE::Util::DataTypes qw(TRUE FALSE);
use NATE::Messenger qw(messenger_global);
use NATE::Events qw(eventloop_drain);
use NATE::ParamSet qw(param_global);
use Params::Validate qw(:all);

# NACL libraries in use
use NACL::C::Cluster;
use NACL::C::Cserver;
use NACL::C::Node;
use NACL::C::StorageAggregate;
use NACL::C::Vserver;
use NACL::C::Volume;
use NACL::C::VolumeQtree;
use NACL::C::Lun;
use NACL::C::LunIgroup;
use NACL::C::NetworkInterface;
use NACL::C::VserverIscsiTpgroup;
use NACL::CS::VserverFcp;
use NACL::CS::VserverIscsi;
use NACL::CS::VserverNfs;
use NACL::CS::NetworkInterface;
use NACL::CS::VolumeQtree;
use NACL::CS::StorageAggregate;
use NACL::CS::Volume;

use San::SanClient::Iscsitarget;
use San::SanClient::Fcptarget;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# San specific modules
use San::ControllerCluster;

use vars qw(@ISA);

@ISA = qw(Error);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################

=head2 get_objects

Get a list of objects type "type" that are under a specific class type
'category' (cluster, client, switch, etc) based on the input arguments.

    my @objs_array = $egx_obj->get_objects('category'=>$cluster_name,
                                           'type'   => "aggr");

=over

=item Options

=over

=item C<< type=>"volume|lun|aggr|vserver|igroup|etc" >>

(Required) The type of resource requested.  The type tag has to match the
tag in the XML.

=item C<< category'=>$cluster_name >>

(Required) The tag where to look for the 'type'. As of now, we need to
provide the name of the cluster where the resources are. You can either
provide the name as it is in the XML or you can provide the name in the
resource list.

=back

=item Return

Returns:   An array or a reference to an array containing all the NACL
Objects.

=item Exceptions

BaseException

=cut

###############################################################################

sub get_objects {
    my ($self, %args) = @_;
    $Log->enter();
    $Log->exit();
    return $self->_get_objs_sorter(%args);
}    # get_objects()

###############################################################################

=head2 get_nacl_objects

Get a list of NACL objects type "type" that are under a specific class type
'category' (cluster, client, switch, etc) based on the input arguments.

    my @objs_array = $egx_obj->get_objects('category'=>$cluster_name,
                                           'type'   => "aggr");

=over

=item Options

=over

=item C<< type=>"volume|lun|aggr|vserver|igroup|etc" >>

(Required) The type of resource requested.  The type tag has to match the
tag in the XML.

=item C<< 'category'=>$cluster_name >>

(Required) The tag where to look for the 'type'. As of now, we need to
provide the name of the cluster where the resources are. You can either
provide the name as it is in the XML or you can provide the name in the
resource list.

=back

=item Return

Returns:   An array or a reference to an array containing all the NACL
Objects.

=item Exceptions

BaseException

=cut

# TODO: Need to implement clients, hypervisor, switches. When
#       they are implemented, more flexibility in the inputs
#       will be needed.  The argument 'key' needs to be renamed to
#       'catagory' and needs to be optional, we can default to the
#       first found cluster.

###############################################################################
sub get_nacl_objects() {
    my ($self, %args) = @_;
    $Log->enter();
    $Log->exit();
    return $self->_get_objs_sorter(%args);
} ## end sub get_nacl_objects()

###############################################################################
# Method:    get_vserver_client_relationship_hash
# Objective: return the relationship hashes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_vserver_client_relationship_hash {
    my ($self, %args) = @_;
    return $self->{vs_cli_relationship_svc}->get_vserver_client_relationship_hash(%args);
}

###############################################################################
# Method:    get_target_list
# Objective: Returns hash of iscsi/fcp targets and details about VServers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_target_list {
    my ($self, %args) = @_;

    $Log->enter();
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                client_name => {
                    type     => SCALAR,
                    optional => 0
                },
                get_iscsi_targets => {
                    type     => BOOLEAN,
                    optional => 1,
                    default  => 1
                },
                get_fcp_targets => {
                    type     => BOOLEAN,
                    optional => 1,
                    default  => 1
                },
                get_nfs_targets => {
                    type     => BOOLEAN,
                    optional => 1,
                    default  => 1
                },
                get_cifs_targets => {
                    type     => BOOLEAN,
                    optional => 1,
                    default  => 1
                },
                filter_iscsi_targets_by_portset => {
                    type     => BOOLEAN,
                    optional => 1,
                    default  => 0
                },
                filter_fcp_targets_by_portset => {
                    type     => BOOLEAN,
                    optional => 1,
                    default  => 0
                },
            }
        );
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        $self->_set_err("Argument Validation Failed:". $exception->text());

        # Cannot Throw Exception Currently because SanMgr
        # relies on errors
        # NATE::Exceptions::Argument->throw($exception->text());
    };
    if ($self->error()) {
        return {};
    }
    my $client_name    = $args{client_name};
    my $get_iscsi_objs = $args{get_iscsi_targets};
    my $get_fcp_objs   = $args{get_fcp_targets};
    my $get_nfs_objs   = $args{get_nfs_targets};
    my $get_cifs_objs   = $args{get_cifs_targets};
    my ($fcp_objs, $iscsi_objs, $lif_objs, $nfs_objs, $cifs_objs);
    my (@FcpTargets, @iSCSITargets, @NFSTargets, @CIFSTargets) = ();
    my $data = {};

    my $hash = $self->get_client_relationship_hash(-client_name => $client_name);

    # obtain the list of vservers that are assiociated with this initiator
    foreach my $cluster_name (keys %{$hash}) {

        my @VServer_objs;
        # Load up the cluster objects
        my ($cluster_command_interfaces) = $self->get_objects('type'      => "cluster",
                                                              'obj_class' => 'nacl',
                                                              'name'      => $cluster_name);
        if ($self->error()) {
            $self->_set_err("Failed to get nacl cluster objects");
            return {};
        }

        # Verify if the node is a vsim, if so return
        if (($args{command_interface}{hostrec}{hosttype} =~ /^vsim-ng/i) and
            ($get_fcp_objs == 1)) {
            $Log->warning("Fcp is not supported on vsims so "
                    . "no FCP Targets are going to be loaded");
            $get_fcp_objs = 0;
        }

        my @vserver_names = keys %{$hash->{$cluster_name}};

        try {
            # Get the fcp objects from the controller sorted by vserver so that
            # we only have to query the controller once and search through the list
            # once.
            if ($get_fcp_objs) {
                $fcp_objs = EnGenX::Util->_get_vserver_sorted_fcp_objs(
                    command_interface => [$cluster_command_interfaces],
                    vservers => \@vserver_names);
            }

            # Get the iscsi objects from the controller sorted by vserver so that
            # we only have to query the controller once and search through the list
            # once.
            if ($get_iscsi_objs) {
                $iscsi_objs = EnGenX::Util->_get_vserver_sorted_iscsi_objs(
                    command_interface => [$cluster_command_interfaces],
                    vservers => \@vserver_names);
            }

            # Get the nfs objects from the controller sorted by vserver so that
            # we only have to query the controller once and search through the list
            # once.
            if ($get_nfs_objs) {
                $nfs_objs = EnGenX::Util->_get_vserver_sorted_nfs_objs(
                    command_interface => [$cluster_command_interfaces],
                    vservers => \@vserver_names);
            }

            # Get the nfs objects from the controller sorted by vserver so that
            # we only have to query the controller once and search through the list
            # once.
            if ($get_cifs_objs) {
                $cifs_objs = EnGenX::Util->_get_vserver_sorted_cifs_objs(
                    command_interface => [$cluster_command_interfaces],
                    vservers => \@vserver_names);
            }

            # Get the lif objects from the controller sorted by vserver and iscsi|fcp
            # so that we only have to query the controller once and search through the
            # list once.
            if ($get_iscsi_objs or $get_fcp_objs or $get_nfs_objs or $get_cifs_objs) {
                $lif_objs = EnGenX::Util->_get_vserver_sorted_lif_objs(
                    command_interface => [$cluster_command_interfaces],
                    vservers => \@vserver_names);
            }
        }
        otherwise {
            my $exception = shift;
            $Log->exit();
            $self->_set_err($exception->text());
            # Cannot Throw Exception Currently because SanMgr
            # relies on errors
            # NATE::Exceptions::Argument->throw($exception->text());
        };
        if ($self->error()) {
            # clean up command interface
            $cluster_command_interfaces->refresh_command_interface();
            return {};
        }

        foreach my $vserver (keys %{$hash->{$cluster_name}}) {

            my $iscsi_lifs = $hash->{$cluster_name}->{$vserver}->{iscsi_lifs};
            my $fcp_lifs   = $hash->{$cluster_name}->{$vserver}->{fcp_lifs};
            my $nfs_lifs   = $hash->{$cluster_name}->{$vserver}->{nfs_lifs};
            my $cifs_lifs  = $hash->{$cluster_name}->{$vserver}->{cifs_lifs};

            # get the controller target objects
            # Ignore any failures that may have occurred.
            # No need to verify that these calls make
            # it back without error because we just don't care.
            if ($get_iscsi_objs and $iscsi_lifs and scalar(@{$iscsi_lifs})) {
                try {
                    push(
                        @iSCSITargets,
                        EnGenX::Util->load_iscsi_targets(
                            cluster_name   => $cluster_name,
                            iscsi_obj      => $iscsi_objs->{$vserver}[0],
                            iscsi_lif_objs => $lif_objs->{$vserver}->{'iscsi'},
                            lif_filter     => $args{filter_iscsi_targets_by_portset} == 0 ?
                                [] : $hash->{$cluster_name}->{$vserver}->{iscsi_lifs}
                        )
                    );
                }
                otherwise {
                    my $exception = shift;
                    $Log->comment("An error occurred while getting iSCSI targets "
                            . "for vserver "
                            . $vserver . "\n"
                            . $exception->text());
                };
            }
            if ($get_fcp_objs and $fcp_lifs and scalar(@{$fcp_lifs})) {
                try {
                    push(
                        @FcpTargets,
                        EnGenX::Util->load_fcp_targets(
                            cluster_name => $cluster_name,
                            fcp_obj      => $fcp_objs->{$vserver}[0],
                            fcp_lif_objs => $lif_objs->{$vserver}->{'fcp'},
                            lif_filter   => $args{filter_fcp_targets_by_portset} == 0 ?
                                [] : $hash->{$cluster_name}->{$vserver}->{fcp_lifs}
                        )
                    );
                }
                otherwise {
                    my $exception = shift;
                    $Log->comment("An error occurred while getting FCP targets "
                            . "for vserver "
                            . $vserver . "\n"
                            . $exception->text());
                };
            }
            if ($get_nfs_objs and $nfs_lifs and scalar(@{$nfs_lifs})) {
                try {
                    push(
                        @NFSTargets,
                        EnGenX::Util->load_nfs_targets(
                            cluster_name   => $cluster_name,
                            nfs_obj      => $nfs_objs->{$vserver}[0],
                            nfs_lif_objs => $lif_objs->{$vserver}->{'nfs'},
                            lif_filter     => []
                        )
                    );
                }
                otherwise {
                    my $exception = shift;
                    $Log->comment("An error occurred while getting NFS targets "
                            . "for vserver "
                            . $vserver . "\n"
                            . $exception->text());
                };
            }
            if ($get_cifs_objs and $cifs_lifs and scalar(@{$cifs_lifs})) {
                try {
                    push(
                        @CIFSTargets,
                        EnGenX::Util->load_cifs_targets(
                            cluster_name   => $cluster_name,
                            cifs_obj      => $cifs_objs->{$vserver}[0],
                            cifs_lif_objs => $lif_objs->{$vserver}->{'cifs'},
                            lif_filter     => []
                        )
                    );
                }
                otherwise {
                    my $exception = shift;
                    $Log->comment("An error occurred while getting CIFS targets "
                            . "for vserver "
                            . $vserver . "\n"
                            . $exception->text());
                };
            }
        }
    }

    $data->{iscsi_targets} = \@iSCSITargets;
    $data->{fcp_targets}   = \@FcpTargets;
    $data->{nfs_targets}   = \@NFSTargets;
    $data->{cifs_targets}  = \@CIFSTargets;

    # Check to verify that iscsi and fcp targets are not empty.
    # If they are, init the hash structures to empty arrays.
    if (not defined($data->{iscsi_targets})) {
        @{$data->{iscsi_targets}} = ();
    }
    if (not defined($data->{fcp_targets})) {
        @{$data->{fcp_targets}} = ();
    }
    if (not defined($data->{nfs_targets})) {
        @{$data->{nfs_targets}} = ();
    }
    if (not defined($data->{cifs_targets})) {
        @{$data->{cifs_targets}} = ();
    }

    $Log->exit();
    return $data;
} ## end sub get_target_list

###############################################################################
# Method Name: _get_vserver_sorted_fcp_objs()
# Objective:   Get Fcp Objs and Sort them in a hash based on vserver
# Details:     See POD documentation at the end of this file.
###############################################################################
sub _get_vserver_sorted_fcp_objs {
    my ($self, %args) = @_;
    $Log->enter();
    %args = &SanUtil::strip_dash(%args);

    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                command_interface => {
                    type     => ARRAYREF,
                    optional => 0
                },
                vservers => {
                    type     => ARRAYREF,
                    optional => 1,
                    default => []
                }
            }
        );
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        NATE::Exceptions::Argument->throw($exception->text());
    };

    my %vs_filter = ();
    if (scalar(@{$args{vservers}})) {
        %vs_filter = (filter => {vserver => join(",", @{$args{vservers}})});
    }

    my $fcp_objs       = {};
    my @fcp_objs_array = ();
    foreach my $cluster (@{$args{command_interface}}) {
        push(@fcp_objs_array,
            NACL::CS::VserverFcp->fetch(
                command_interface => $cluster,
                allow_empty       => 1,
                %vs_filter
            )
        );
    }

    foreach my $fcp_obj (@fcp_objs_array) {
        my $vserver = $fcp_obj->vserver();
        delete $fcp_obj->{command_interface};
        push(@{$fcp_objs->{$vserver}}, $fcp_obj);

    }
    $Log->exit();
    return $fcp_objs;
} ## end sub _get_vserver_sorted_fcp_objs

###############################################################################
# Method Name: _get_vserver_sorted_iscsi_objs()
# Objective:   Get Iscsi Objs and Sort them in a hash based on vserver
# Details:     See POD documentation at the end of this file.
###############################################################################
sub _get_vserver_sorted_iscsi_objs {
    my ($self, %args) = @_;
    $Log->enter();
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                command_interface => {
                    type     => ARRAYREF,
                    optional => 0
                },
                vservers => {
                    type     => ARRAYREF,
                    optional => 1,
                    default => []
                }
            }
        );
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        NATE::Exceptions::Argument->throw($exception->text());
    };

    my %vs_filter = ();
    if (scalar(@{$args{vservers}})) {
        %vs_filter = (filter => {vserver => join(",", @{$args{vservers}})});
    }

    my $iscsi_objs       = {};
    my @iscsi_objs_array = ();
    foreach my $cluster (@{$args{command_interface}}) {
        push(@iscsi_objs_array,
            NACL::CS::VserverIscsi->fetch(
                command_interface => $cluster,
                allow_empty       => 1,
                %vs_filter
            )
        );
    }

    foreach my $iscsi_obj (@iscsi_objs_array) {
        my $vserver = $iscsi_obj->vserver();
        delete $iscsi_obj->{command_interface};
        push(@{$iscsi_objs->{$vserver}}, $iscsi_obj);
    }
    $Log->exit();
    return $iscsi_objs;
} ## end sub _get_vserver_sorted_iscsi_objs

###############################################################################
# Method Name: _get_vserver_sorted_nfs_objs()
# Objective:   Get Nfs Objs and Sort them in a hash based on vserver
# Details:     See POD documentation at the end of this file.
###############################################################################
sub _get_vserver_sorted_nfs_objs {
    my ($self, %args) = @_;
    $Log->enter();
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                command_interface => {
                    type     => ARRAYREF,
                    optional => 0
                },
                vservers => {
                    type     => ARRAYREF,
                    optional => 1,
                    default => []
                }
            }
        );
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        NATE::Exceptions::Argument->throw($exception->text());
    };

    my %vs_filter = ();
    if (scalar(@{$args{vservers}})) {
        %vs_filter = (filter => {vserver => join(",", @{$args{vservers}})});
    }

    my $nfs_objs = {};
    my @nfs_objs_array = ();
    foreach my $cluster (@{$args{command_interface}}) {
        push(@nfs_objs_array,
            NACL::CS::VserverNfs->fetch(
                command_interface => $cluster,
                allow_empty       => 1,
                %vs_filter
            )
        );
    }

    foreach my $nfs_obj (@nfs_objs_array) {
        my $vserver = $nfs_obj->vserver();
        delete $nfs_obj->{command_interface};
        push(@{$nfs_objs->{$vserver}}, $nfs_obj);
    }
    $Log->exit();
    return $nfs_objs;
} ## end sub _get_vserver_sorted_nfs_objs

###############################################################################
# Method Name: _get_vserver_sorted_cifs_objs()
# Objective:   Get Cifs Objs and Sort them in a hash based on vserver
# Details:     See POD documentation at the end of this file.
###############################################################################
sub _get_vserver_sorted_cifs_objs {
    my ($self, %args) = @_;
    $Log->enter();
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                command_interface => {
                    type     => ARRAYREF,
                    optional => 0
                },
                vservers => {
                    type     => ARRAYREF,
                    optional => 1,
                    default => []
                }
            }
        );
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        NATE::Exceptions::Argument->throw($exception->text());
    };

    my %vs_filter = ();
    if (scalar(@{$args{vservers}})) {
        %vs_filter = (filter => {vserver => join(",", @{$args{vservers}})});
    }

    my $cifs_objs       = {};
    my @cifs_objs_array = ();
    foreach my $cluster (@{$args{command_interface}}) {
        push(@cifs_objs_array,
            NACL::CS::VserverCifs->fetch(
                command_interface => $cluster,
                allow_empty       => 1,
                %vs_filter
            )
        );
    }

    foreach my $cifs_obj (@cifs_objs_array) {
        my $vserver = $cifs_obj->vserver();
        delete $cifs_obj->{command_interface};
        push(@{$cifs_objs->{$vserver}}, $cifs_obj);
    }
    $Log->exit();
    return $cifs_objs;
} ## end sub _get_vserver_sorted_cifs_objs

###############################################################################
# Method Name: _get_vserver_sorted_lif_objs()
# Objective:   Get lif Objs and Sort them in a hash based on vserver and
#                  protocol (iscsi|fcp)
# Details:     See POD documentation at the end of this file.
###############################################################################
sub _get_vserver_sorted_lif_objs {
    my ($self, %args) = @_;
    $Log->enter();
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                command_interface => {
                    type     => ARRAYREF,
                    optional => 0
                },
                vservers => {
                    type     => ARRAYREF,
                    optional => 1,
                    default => []
                }
            }
        );
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        NATE::Exceptions::Argument->throw($exception->text());
    };

    my %vs_filter = ();
    if (scalar(@{$args{vservers}})) {
        %vs_filter = (filter => {vserver => join(",", @{$args{vservers}})});
    }

    my $lif_objs       = {};
    my @lif_objs_array = ();
    foreach my $cluster (@{$args{command_interface}}) {
        # need to force CLI here because some fields do not get populated by ZAPI correctly
        push(@lif_objs_array,
            NACL::CS::NetworkInterface->fetch(
                command_interface => $cluster,
                allow_empty       => 1,
                apiset_must => {interface => "CLI"},
                %vs_filter
            )
        );
    }

    foreach my $lif_obj (@lif_objs_array) {
        my $vserver = $lif_obj->vserver();
        delete $lif_obj->{command_interface};
        if (grep(/fcp/i, @{$lif_obj->data_protocol()})) {
            push(@{$lif_objs->{$vserver}->{'fcp'}}, $lif_obj);
        }
        if (grep(/iscsi/i, @{$lif_obj->data_protocol()})) {
            push(@{$lif_objs->{$vserver}->{'iscsi'}}, $lif_obj);
        }
        if (grep(/nfs/i, @{$lif_obj->data_protocol()})) {
            push(@{$lif_objs->{$vserver}->{'nfs'}}, $lif_obj);
        }
        if (grep(/cifs/i, @{$lif_obj->data_protocol()})) {
            push(@{$lif_objs->{$vserver}->{'cifs'}}, $lif_obj);
        }
    }
    $Log->exit();
    return $lif_objs;
} ## end sub _get_vserver_sorted_lif_objs

###############################################################################
# Method:    fetch_nacl_clusters
# Objective: Fetches Unique NACL Clusters
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fetch_nacl_clusters {
    my ($self, %args) = @_;
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                clusters => {
                    type     => ARRAYREF,
                    optional => 1,
                    default  => []
                },
            }
        );
    }
    otherwise {
        my $e = shift;
        $self->_set_err("Exception caught: " . $e->text());
    };
    if ($self->error()) {
        return (wantarray()) ? () : [];
    }

    my @cluster_names = @{$args{clusters}};

    my @cluster_objs;

    # search for unique clusters in this testbed and filter out the list
    # if requested
    try {
        @cluster_objs = NACL::STask::Cluster->find_unique_clusters();

        # filter out the cluster object list with what ever was passed in
        if (scalar(@cluster_names)) {
            @cluster_objs = grep {
                my $cluster_name = $_->name();
                grep(/^$cluster_name$/, @cluster_names);
            } @cluster_objs;
        }
    } otherwise {
        my $e = shift;
        $self->_set_err("Exception caught: " . $e->text());
    };

    # Need this call to do versioning on the cluster objects because if
    # versioning is not done upfront then extra connnections are created
    # during lower levels of EnGenX that can cause connctions failures
    foreach my $cluster_obj (@cluster_objs) {
        $cluster_obj->apiset();
    }

    return (wantarray()) ? @cluster_objs : \@cluster_objs;
}

###############################################################################
# Method Name: load_fcp_targets()
# Objective:   Load important information about vserver fcp targets
# Details:     See POD documentation at the end of this file.
###############################################################################
sub load_fcp_targets {
    my ($self, %args) = @_;

    my ($fcp_obj, @fcp_lif_objs, $lif_obj, $host_obj);
    my ($nodename, $state);
    my ($is_mgmt_port);
    my @target_objs;
    $Log->enter();
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                fcp_obj => {
                    type     => HASHREF|UNDEF,
                    optional => 0
                },
                fcp_lif_objs => {
                    type     => ARRAYREF|UNDEF,
                    optional => 0
                },
                cluster_name => {
                    type     => SCALAR,
                    optional => 0
                },
                lif_filter   => {
                    type     => ARRAYREF,
                    optional => 1,
                    default  => []
                }
            }
        );
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        NATE::Exceptions::Argument->throw($exception->text());
    };

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args         = &SanUtil::strip_dash(%args);
    $nodename     = $args{fcp_obj} ? $args{fcp_obj}->nodename() : "";
    @fcp_lif_objs = $args{fcp_lif_objs} ? @{$args{fcp_lif_objs}} : ();

    # create and load target for each lif
    foreach $lif_obj (@fcp_lif_objs) {
        #go to next lif_obj if $lif_obj is not in lif_filter
        if (scalar(@{$args{lif_filter}}) > 0) {
            my $lif_name = $lif_obj->lif();
            next if (not grep(/^$lif_name$/, @{$args{lif_filter}}));
        }

        $is_mgmt_port = 0;
        $state        = $lif_obj->status_oper();

        # Create fcp target object (if interface is enabled for fcp)
        my $targObj = San::SanClient::Fcptarget->_new(
            -name         => $lif_obj->lif(),
            -state        => $state,
            -is_mgmt_port => $is_mgmt_port,
            -nodename     => $nodename,
            -vserver_name => $lif_obj->vserver(),
            -cluster_name => $args{cluster_name}
        );
        if ($targObj->error()) {
            NATE::BaseException->throw($targObj->errmsg());
        }

        push(@target_objs, $targObj);
    }
    $Log->exit();
    return (wantarray()) ? @target_objs : \@target_objs;
} ## end sub load_fcp_targets

###############################################################################
# Method Name: load_iscsi_targets()
# Objective:   Return important information about filer iscsi targets
# Details:     See POD documentation at the end of this file.
###############################################################################
sub load_iscsi_targets {
    my ($self, %args) = @_;

    my ($iscsi_obj,    @iscsi_lif_objs,      $lif_obj, $host_obj);
    my ($nodename,     $state,               $ipv6,    $ipv4, $tpgn, $tpgt);
    my ($filer_name,   $interface,           $ip,      $netmask);
    my ($ipv4_netmask, $ipv6_netmask_length, $mac,     $is_mgmt_port);
    my @target_objs;
    $Log->enter();
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                iscsi_obj => {
                    type     => HASHREF|UNDEF,
                    optional => 0
                },
                iscsi_lif_objs => {
                    type     => ARRAYREF|UNDEF,
                    optional => 0
                },
                cluster_name => {
                    type     => SCALAR,
                    optional => 0
                },
                lif_filter   => {
                    type     => ARRAYREF,
                    optional => 1,
                    default  => []
                }
            }
        );
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        NATE::Exceptions::Argument->throw($exception->text());
    };

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    $nodename = $args{iscsi_obj} ? $args{iscsi_obj}->target_name() : "";
    @iscsi_lif_objs = $args{iscsi_lif_objs} ? @{$args{iscsi_lif_objs}} : ();

    # create and load target for each lif
    foreach $lif_obj (@iscsi_lif_objs) {
        #go to next lif_obj if $lif_obj is not in lif_filter
        if (scalar(@{$args{lif_filter}}) > 0) {
            my $lif_name = $lif_obj->lif();
            next if (not grep(/^$lif_name$/, @{$args{lif_filter}}));
        }

        $ipv4 = $ipv6 = $tpgt = $tpgn = "not_set";
        $ipv4_netmask = $ipv6_netmask_length = $mac = "not_set";
        $is_mgmt_port = 0;
        my $address = $lif_obj->address();
        if ($address =~ /:/) {
            $ipv6                = $address;
            $ipv6_netmask_length = $lif_obj->netmask_length();
        } else {
            $ipv4         = $address;
            $ipv4_netmask = $lif_obj->netmask();
        }
        $tpgn  = $lif_obj->lif();
        $tpgt  = $lif_obj->numeric_id();
        $state = $lif_obj->status_oper();

        # Create iscsi target object (if interface is enabled for iSCSI)
        my $targObj = San::SanClient::Iscsitarget->_new(
            -name                => $lif_obj->lif(),
            -state               => $state,
            -ip                  => $address,
            -ipv4                => $ipv4,
            -ipv4_netmask        => $ipv4_netmask,
            -ipv6                => $ipv6,
            -ipv6_netmask_length => $ipv6_netmask_length,
            -tpgt                => $tpgt,
            -tpgn                => $tpgn,
            -mac                 => $mac,
            -is_mgmt_port        => $is_mgmt_port,
            -nodename            => $nodename,
            -vserver_name        => $lif_obj->vserver(),
            -cluster_name        => $args{cluster_name}
        );
        if ($targObj->error()) {
            NATE::BaseException->throw($targObj->errmsg());
        }

        push(@target_objs, $targObj);
    }
    $Log->exit();
    return (wantarray()) ? @target_objs : \@target_objs;
} ## end sub load_iscsi_targets

###############################################################################
# Method Name: load_nfs_targets()
# Objective:   Return important information about filer nfs targets
# Details:     See POD documentation at the end of this file.
###############################################################################
sub load_nfs_targets {
    my ($self, %args) = @_;

    my ($nfs_obj,    @nfs_lif_objs,      $lif_obj, $host_obj);
    my ($nodename,     $state,               $ipv6,    $ipv4, $tpgn, $tpgt);
    my ($filer_name,   $interface,           $ip,      $netmask);
    my ($ipv4_netmask, $ipv6_netmask_length, $mac,     $is_mgmt_port);
    my @target_objs;
    $Log->enter();
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                nfs_obj => {
                    type     => HASHREF|UNDEF,
                    optional => 0
                },
                nfs_lif_objs => {
                    type     => ARRAYREF|UNDEF,
                    optional => 0
                },
                cluster_name => {
                    type     => SCALAR,
                    optional => 0
                },
                lif_filter   => {
                    type     => ARRAYREF,
                    optional => 1,
                    default  => []
                }
            }
        );
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        NATE::Exceptions::Argument->throw($exception->text());
    };

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    @nfs_lif_objs = $args{nfs_lif_objs} ? @{$args{nfs_lif_objs}} : ();
    my $hash = {};
    # create and load target for each lif
    foreach $lif_obj (@nfs_lif_objs) {
        #go to next lif_obj if $lif_obj is not in lif_filter
        if (scalar(@{$args{lif_filter}}) > 0) {
            my $lif_name = $lif_obj->lif();
            next if (not grep(/^$lif_name$/, @{$args{lif_filter}}));
        }
        $hash->{cluster_name} = $args{cluster_name};
        $hash->{vserver_name} = $lif_obj->vserver();
        $hash->{nfs_obj} = $args{nfs_obj};
        push(@{$hash->{lif_objs}}, $lif_obj);
    }

    @target_objs = ($hash);

    $Log->exit();
    return (wantarray()) ? @target_objs : \@target_objs;
} ## end sub load_nfs_targets

###############################################################################
# Method Name: load_cifs_targets()
# Objective:   Return important information about filer cifs targets
# Details:     See POD documentation at the end of this file.
###############################################################################
sub load_cifs_targets {
    my ($self, %args) = @_;

    my ($cifs_obj,    @cifs_lif_objs,      $lif_obj, $host_obj);
    my ($nodename,     $state,               $ipv6,    $ipv4, $tpgn, $tpgt);
    my ($filer_name,   $interface,           $ip,      $netmask);
    my ($ipv4_netmask, $ipv6_netmask_length, $mac,     $is_mgmt_port);
    my @target_objs;
    $Log->enter();
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                cifs_obj => {
                    type     => HASHREF|UNDEF,
                    optional => 0
                },
                cifs_lif_objs => {
                    type     => ARRAYREF|UNDEF,
                    optional => 0
                },
                cluster_name => {
                    type     => SCALAR,
                    optional => 0
                },
                lif_filter   => {
                    type     => ARRAYREF,
                    optional => 1,
                    default  => []
                }
            }
        );
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        NATE::Exceptions::Argument->throw($exception->text());
    };

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    @cifs_lif_objs = $args{cifs_lif_objs} ? @{$args{cifs_lif_objs}} : ();

    my $hash = {};
    # create and load target for each lif
    foreach $lif_obj (@cifs_lif_objs) {
        #go to next lif_obj if $lif_obj is not in lif_filter
        if (scalar(@{$args{lif_filter}}) > 0) {
            my $lif_name = $lif_obj->lif();
            next if (not grep(/^$lif_name$/, @{$args{lif_filter}}));
        }
        $hash->{cluster_name} = $args{cluster_name};
        $hash->{vserver_name} = $lif_obj->vserver();
        $hash->{cifs_obj} = $args{cifs_obj};
        push(@{$hash->{lif_objs}}, $lif_obj);
    }

    @target_objs = ($hash);
    $Log->exit();
    return (wantarray()) ? @target_objs : \@target_objs;
} ## end sub load_cifs_targets

###############################################################################
# Method:    debug_level
# Objective: Set SanMgr debug level and return it to caller
# Details:   See POD documentation at the end of this file.
###############################################################################
sub debug_level {
    my ($self, %args) = @_;
    $Log->enter();
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                debug_level => {
                    type     => SCALAR,
                    optional => 1,
                    default  => 0
                },
            }
        );
    }
    otherwise {
        my $xcptn = shift;
        $Log->exit();
        NATE::Exceptions::Argument->throw($xcptn->text());
    };

    my $debug_level = $args{debug_level};
    $self->{debug_level} = $debug_level;
    $Log->comment("Setting debug level to ($debug_level)", -msg_level => 0);

    $Log->exit();
    return $debug_level;
} ## end sub debug_level

###############################################################################
# Method:    get_error
# Objective: Return the errors string on the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_error {
    my ($self) = @_;
    $Log->enter();
    $Log->exit();
    return $self->errmsg();
} ## end sub get_error

###############################################################################
# Method:    clear_error
# Objective: Clear the error string(s) set on the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub clear_error {
    my ($self) = @_;
    $Log->enter();
    $self->_clear_err();
    $Log->exit();
    return TRUE;
} ## end sub clear_error

###############################################################################
# Method:    store_client_dev_list
# Objective: Store client device list on object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub store_client_dev_list {
    my ($self, %args) = @_;
    $Log->enter();
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                device_list => {
                    type     => HASHREF,
                    optional => 0,
                },
            }
        );
    }
    otherwise {
        my $e = shift;
        $self->_set_err("Exception caught: " . $e->text());
    };
    if ($self->error()) {
        return FALSE
    }

    my $device_list = $args{device_list};
    $self->{_client_dev_list} = $device_list;
    $Log->exit();
    return TRUE;
} ## end sub store_client_dev_list

###############################################################################
# Method:    get_client_dev_list
# Objective: Get client device list from internal object.
#            Also allows caller to filter on attributes of client device list.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_client_dev_list {
    my ($self, %args) = @_;
    $Log->enter();
    my ($cdl, $ncdl);
    %args = &SanUtil::strip_dash(%args);
    %args = Params::Validate::validate_with(
        params => \%args,
        spec   => {
            client => {
                type     => SCALAR|UNDEF,
                optional => 1,
                default  => undef
            },
            vserver => {
               type     => SCALAR|UNDEF,
               optional => 1,
               default  => undef
            },
            volume => {
               type     => SCALAR|UNDEF,
               optional => 1,
               default  => undef
            },
                lun => {
                    type     => SCALAR|UNDEF,
                    optional => 1,
                    default  => undef
                },
                protocol => {
                    type     => SCALAR|UNDEF,
                    optional => 1,
                    default  => undef
                },
                serial => {
                    type     => SCALAR|UNDEF,
                    optional => 1,
                    default  => undef
                },
                cdl_format => {
                    type     => SCALAR,
                    optional => 1,
                    default  => FALSE
                },
            }
        );

    my $filter_client = $args{client};
    my $filter_vserver = $args{vserver};
    my $filter_volume = $args{volume};
    my $filter_lun = $args{lun};
    my $filter_protocol = $args{protocol};
    my $filter_serial = $args{serial};
    my $cdl_format = $args{cdl_format};

    if (not defined($self->{_client_dev_list})) {
        # setting errors but also throwing exceptions to deal
        # with pre-EnGenX code.
        $self->_set_err("Client device list is not defined!");
        NATE::BaseException->throw("Client dev list undefined");
    }

    $cdl = $self->{_client_dev_list};

    # Go through each client in the client device list and apply a client
    # filter if requested by caller
    foreach my $client (keys(%{$cdl})) {

        # Check filter_client, if its not empty, process the filter
        if ($filter_client ne "") {

            # If the client matches the filter client, save for returning
            if ($filter_client =~ $client) {
                $ncdl->{$client} = $cdl->{$client};
            }
        } else {

            # No filter specified, save all clients for returning
            $ncdl->{$client} = $cdl->{$client};
        }
    }

    # Filter out the client dev list based on the lun attributes
    $ncdl = $self->_filter_cdl(
        -filter_key => "vserver",
        -filter_str => $filter_vserver,
        -filter_cdl => $ncdl
    );
    $ncdl = $self->_filter_cdl(
        -filter_key => "volume",
        -filter_str => $filter_volume,
        -filter_cdl => $ncdl
    );
    $ncdl = $self->_filter_cdl(
        -filter_key => "lun",
        -filter_str => $filter_lun,
        -filter_cdl => $ncdl
    );
    $ncdl = $self->_filter_cdl(
        -filter_key => "protocol",
        -filter_str => $filter_protocol,
        -filter_cdl => $ncdl
    );
    $ncdl = $self->_filter_cdl(
        -filter_key => "serial",
        -filter_str => $filter_serial,
        -filter_cdl => $ncdl
    );
    if (&ParamsMgr::error()) {
        $self->_set_err("Failed to filter client device list:",
            &ParamsMgr::get_error());
        return {};
    }

    # Check to see if anything is getting returned, if not set an error
    if (scalar(keys(%{$ncdl})) == 0) {
        $self->_set_err(
            "Client device list is empty or filter was not found");
        return {};
    }

    if ($cdl_format == TRUE) {
        my $final_client_dev_list;

        foreach my $client (keys %{$ncdl}) {
            $final_client_dev_list .= $client . ",";
            foreach my $cdl_entry (@{$ncdl->{$client}}) {
                my $path = $cdl_entry->{path};
                $final_client_dev_list .= $path . ",";
            }
            chop($final_client_dev_list);
            $final_client_dev_list .= ";";
        }
        chop($final_client_dev_list);
        $Log->exit();
        return $final_client_dev_list;
    }
    $Log->exit();
    return $ncdl;
} ## end sub get_client_dev_list

###############################################################################
# Method:    _filter_cdl
# Objective: _filter_cdl returns a filter client dev list based on attributes
#            on each device within each client
###############################################################################
sub _filter_cdl {
    my ($self, %args) = @_;
    $Log->enter();
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                filter_key => {
                    type     => SCALAR,
                    optional => 0,
                    default  => ""
                },
                filter_str => {
                    type     => SCALAR|UNDEF,
                    optional => 1,
                    default  => undef
                },
                filter_cdl => {
                    type     => HASHREF,
                    optional => 1,
                    default  => {}
                },
            }
        );
    }
    otherwise {
        my $e = shift;
        $self->_set_err("Exception caught: " . $e->text());
    };
    if ($self->error()) {
        return {};
    }

    my $filter_key = $args{filter_key};
    my $filter_str = $args{filter_str};
    my $filter_cdl = $args{filter_cdl};

    my $fcdl;

    # If the filter string is undefined, just return the cdl
    if (not defined($filter_str)) {
        return $filter_cdl;
    }

# Go through each client in the client device list and apply a vserver filter if
# requested by caller
    foreach my $client (keys(%{$filter_cdl})) {
        foreach my $lun (@{$filter_cdl->{$client}}) {

            # Determine if the filter_key contains any valid data
            if (defined($lun->{$filter_key})) {
                if (ref($lun->{$filter_key}) eq "ARRAY") {

                    # The data is an array, process it
                    my $found = 0;
                    foreach my $item (@{$lun->{$filter_key}}) {
                        if ($item =~ $filter_str) {
                            $found = 1;
                            last;
                        }
                    }
                    if ($found == 1) {
                        push(@{$fcdl->{$client}}, $lun);
                    }
                } else {

                    # The data is a string
                    if ($lun->{$filter_key} =~ $filter_str) {
                        push(@{$fcdl->{$client}}, $lun);
                    }
                }
            }
        }
    }
    $Log->exit();
    return $fcdl;
} ## end sub _filter_cdl

###############################################################################
# Method:    parent_caller
# Objective: parent_caller returns the subroutine of caller of depth 2
# Details:   format for the caller:
#      0         1          2      3            4
#      ($package, $filename, $line, $subroutine, $hasargs,
#      5          6          7            8       9         10
#      $wantarray, $evaltext, $is_require, $hints, $bitmask, $hinthash) = caller($i);
#      $i is the caller depth
###############################################################################
sub parent_caller { (caller(2))[3] }

###############################################################################
# Method:    get_luns_associated_to_client
# Objective: Get the list of luns that are associated to
#            a specified client
# Return:    An array of luns associated to the client passed in.
###############################################################################
sub get_luns_associated_to_client {
    my ($self, %args) = @_;
    return $self->{vs_cli_relationship_svc}->get_luns_associated_to_client(%args);
}

###############################################################################
# Method:    get_luns_and_vserver_associated_to_client
# Objective: Get the list of luns that are associated to
#            a specified client
# Return:    An array of arrays, The first array the vservers, the second
#            array the luns that are all associated to the client passed in
###############################################################################
sub get_luns_and_vservers_associated_to_client {
    my ($self, %args) = @_;
    return $self->{vs_cli_relationship_svc}->get_luns_and_vservers_associated_to_client(%args);
}

###############################################################################
# Method:    get_clients_associated_to_vserver
# Objective: Get the clients that are associated to a vserver/cluster
###############################################################################
sub get_clients_associated_to_vserver {
    my ($self, %args) = @_;
    return $self->{vs_cli_relationship_svc}->get_clients_associated_to_vserver(%args);
}

###############################################################################
# Method:    get_vservers_associated_to_client
# Objective: Get the clients that are associated to a vserver/cluster
###############################################################################
sub get_vservers_associated_to_client {
    my ($self, %args) = @_;
    return $self->{vs_cli_relationship_svc}->get_vservers_associated_to_client(%args);
}

###############################################################################
# Method:    get_clients_associated_to_vserver
# Objective: Get the clients that are associated to a vserver/cluster
###############################################################################
sub get_client_relationship_hash {
    my ($self, %args) = @_;
    return $self->{vs_cli_relationship_svc}->get_client_relationship_hash(%args);
}

###############################################################################
# Method:    get_nfs_target_relationship_hash()
# Objective: Given a client name, generate a relationship hash describing NFS
#            targets that the client can communicate with.
# Return:    An hash datastructure
###############################################################################
sub get_nfs_target_relationship_hash {
    my ($self, %args) = @_;

    my $client_name = delete($args{client_name});
    if (!$client_name) {
        NATE::Exceptions::Argument->throw("Missing required argument 'client_name'");
    }

    # Grab the target datastructures for nfs that this client can talk to
    my ($data) = $self->get_target_list(client_name => $client_name,
                                        get_iscsi_targets => 0,
                                        get_fcp_targets => 0,
                                        get_nfs_targets => 1,
                                        get_cifs_targets => 0);
    if ($self->error()) {
        NATE::BaseException->throw("Failed to obtain target list for  client '$client_name'");
    }

    my @NFSTargets = @{$data->{nfs_targets}};

    # we need to build a hash that will make the next set of code
    # more efficient
    my $nfs_target_hash = {};
    foreach my $nfs_target (@NFSTargets) {
        $nfs_target_hash->{$nfs_target->{cluster_name}}->{vservers}->
            {$nfs_target->{vserver_name}}->{lif_objs} = $nfs_target->{lif_objs};
        $nfs_target_hash->{$nfs_target->{cluster_name}}->{vservers}->
            {$nfs_target->{vserver_name}}->{nfs_obj} =  $nfs_target->{nfs_obj};
    }

    # for each cluster, get the list of volumes that are mountable and associate them in the hash
    # these are the volumes that this client will be allowed to select for mounting
    foreach my $cluster_name (keys %{$nfs_target_hash}) {
        my ($cluster_command_interface) = $self->get_objects('type'      => "cluster",
                                                             'obj_class' => 'nacl',
                                                             'name'      => $cluster_name);
        if ($self->error()) {
            NATE::BaseException->throw("Failed to obtain nacl cluster object for $cluster_name, " . $self->errmsg());
        }
        # find all volumes on these vservers that are mounted, exclude volumes mounted at '/' (rootvols)
        my %vol_filter = (filter => { vserver => join(",", keys %{$nfs_target_hash->{$cluster_name}->{vservers}}),
                                      'junction-path' => '!"-",!"/"' });
        my @vol_objs = ();
        try {
            @vol_objs = NACL::CS::Volume->fetch(command_interface => $cluster_command_interface,
                                                %vol_filter);
        } catch NACL::Exceptions::NoElementsFound with {
            NATE::BaseException->throw("There are no volumes on cluster $cluster_name that we can use " .
                                       "for NFS mounts on client $client_name");
        };
        # store the vol objects in the hash for easy lookup
        foreach my $vol_obj (@vol_objs) {
            $nfs_target_hash->{$cluster_name}->{vservers}->{$vol_obj->vserver()}->{volumes}->{$vol_obj->volume()}->{vol_obj} = $vol_obj;
            $nfs_target_hash->{$cluster_name}->{vservers}->{$vol_obj->vserver()}->{volumes}->{$vol_obj->volume()}->{qtree_objs} = [];
        }

        my %qtree_filter = (filter => { vserver => join(",", keys %{$nfs_target_hash->{$cluster_name}->{vservers}}) });

        my @qtree_objs = NACL::CS::VolumeQtree->fetch(command_interface => $cluster_command_interface,
                                                      %qtree_filter,
                                                      allow_empty => 1);
        # store the qtrees in their respective locations for easy lookup
        foreach my $qtree_obj (@qtree_objs) {
            next if ($qtree_obj->qtree() eq '""');
            next if (not exists($nfs_target_hash->{$cluster_name}->{vservers}->{$qtree_obj->vserver()}->{volumes}->{$qtree_obj->volume()}));
            push(@{$nfs_target_hash->{$cluster_name}->{vservers}->{$qtree_obj->vserver()}->{volumes}->{$qtree_obj->volume()}->{qtree_objs}}, $qtree_obj);
        }

        my @aggr_objs = ();
        try {
            @aggr_objs = NACL::CS::StorageAggregate->fetch(command_interface => $cluster_command_interface);
        } catch NACL::Exceptions::NoElementsFound with {
            NATE::BaseException->throw("Cannot generate NFS Target relationship hash for client $client_name " .
                                       "because there are no aggregates on the cluster $cluster_name");
        };

        foreach my $aggr_obj (@aggr_objs) {
            $nfs_target_hash->{$cluster_name}->{aggr_objs}->{$aggr_obj->aggregate()} = $aggr_obj;
        }
    }

    return $nfs_target_hash;
}

###############################################################################
# Method:    generate_nfs_options()
# Objective: Given a client resource paramset hash, build an array of NFS
#            option hashes that can be used by clients to filter against.
# Return:    An array of nfs option hashes
###############################################################################
sub generate_nfs_options {
    my ($self, %opts) = @_;

    my $resource = delete($opts{resource});
    if (!$resource) {
        NATE::Exceptions::Argument->throw("Missing required parameter 'resource'");
    }

    my @nfs_option_keys;
    # obtain the nfs options from the local nfs_options datastructure
    # if it exists on the client
    foreach my $key (keys %{$resource}) {
        next if ($key !~ /NFS_OPTIONS/);
        push(@nfs_option_keys, $key);
    }

    # if this resource does not have a local nfs options, set a flag and insert a fake one
    # so it enters the for loop
    if (not scalar(@nfs_option_keys)) {
        push(@nfs_option_keys, "global");
    }
    # create a callback to handle searching for the nfs parameter in the
    # global nfs options datastructure
    my $callback = sub {
        my ($args) = @_;
        my ($param, $default) = @{$args};

        my $nfs_options = &ParamsMgr::search(
            search_path => [],
            key         => "ENVIRONMENT",
            key_filter  => qr/^GLOBAL_NFS_OPTIONS$/,
            paramset => $self->EnGenX::Util::get_paramset()
        );
        if (&ParamsMgr::error()) {
            NATE::BaseException->throw("Failed to find global_nfs_options in the environment ParamSet",
                                       &ParamsMgr::get_error());
        }

        # Process the global nfs options element in paramsmgr
        &ParamsMgr::process_args(
            %{$nfs_options->{GLOBAL_NFS_OPTIONS}},
            -CALLER => "global_nfs_options"
        );

        # parse out the parameter in the global nfs options tag
        return &ParamsMgr::get_arg(
            -arg     => $param,
            -default => $default,
            -CALLER  => "global_nfs_options"
        );
    };

    my @nfs_options = ();

    foreach my $key (@nfs_option_keys) {
        if ($key eq 'global') {
            # reset the args that will be processed
            &ParamsMgr::process_args();
        } else {
            &ParamsMgr::process_args(%{$resource->{$key}});
        }

        # load in all the options
        my $nfs_opts = {};
        $nfs_opts->{nfs_version} = &ParamsMgr::get_arg(
            -arg       => "NFS_VERSION",
            -default   => $callback,
            -code_args => ["NFS_VERSION"],
            -eval_code => 1
        );
        $nfs_opts->{transport_protocol} = &ParamsMgr::get_arg(
            -arg       => "TRANSPORT_PROTOCOL",
            -default   => $callback,
            -code_args => ["TRANSPORT_PROTOCOL"],
            -eval_code => 1
        );
        $nfs_opts->{persistent} = &ParamsMgr::get_arg(
            -arg       => "PERSISTENT",
            -default   => $callback,
            -code_args => ["PERSISTENT"],
            -eval_code => 1
        );
        $nfs_opts->{mount_options} = &ParamsMgr::get_arg(
            -arg       => "MOUNT_OPTIONS",
            -default   => $callback,
            -code_args => ["MOUNT_OPTIONS"],
            -eval_code => 1
        );
        $nfs_opts->{mount_qtrees} = eval(&ParamsMgr::get_arg(
            -arg       => "MOUNT_QTREES",
            -default   => $callback,
            -code_args => ["MOUNT_QTREES"],
            -eval_code => 1
        ));
        $nfs_opts->{create_unique_subdir} = eval(&ParamsMgr::get_arg(
            -arg       => "CREATE_UNIQUE_SUBDIR",
            -default   => $callback,
            -code_args => ["CREATE_UNIQUE_SUBDIR"],
            -eval_code => 1
        ));
        $nfs_opts->{cluster_filter} = eval(&ParamsMgr::get_arg(
            -arg       => "CLUSTER_FILTER",
            -default   => $callback,
            -code_args => ["CLUSTER_FILTER"],
            -eval_code => 1
        ));
        $nfs_opts->{vserver_filter} = eval(&ParamsMgr::get_arg(
            -arg       => "VSERVER_FILTER",
            -default   => $callback,
            -code_args => ["VSERVER_FILTER"],
            -eval_code => 1
        ));
        $nfs_opts->{junction_path_filter} = eval(&ParamsMgr::get_arg(
            -arg       => "JUNCTION_PATH_FILTER",
            -default   => $callback,
            -code_args => ["JUNCTION_PATH_FILTER"],
            -eval_code => 1
        ));
        $nfs_opts->{volume_filter} = eval(&ParamsMgr::get_arg(
            -arg       => "VOLUME_FILTER",
            -default   => $callback,
            -code_args => ["VOLUME_FILTER"],
            -eval_code => 1
        ));
        $nfs_opts->{qtree_filter} = eval(&ParamsMgr::get_arg(
            -arg       => "QTREE_FILTER",
            -default   => $callback,
            -code_args => ["QTREE_FILTER"],
            -eval_code => 1
        ));
        $nfs_opts->{lif_address_filter} = eval(&ParamsMgr::get_arg(
            -arg       => "LIF_ADDRESS_FILTER",
            -default   => $callback,
            -code_args => ["LIF_ADDRESS_FILTER"],
            -eval_code => 1
        ));
        $nfs_opts->{lif_name_filter} = eval(&ParamsMgr::get_arg(
            -arg       => "LIF_NAME_FILTER",
            -default   => $callback,
            -code_args => ["LIF_NAME_FILTER"],
            -eval_code => 1
        ));
        $nfs_opts->{volume_lif_affinity} = &ParamsMgr::get_arg(
            -arg       => "VOLUME_LIF_AFFINITY",
            -default   => $callback,
            -code_args => ["VOLUME_LIF_AFFINITY"],
            -eval_code => 1
        );
        push(@nfs_options, $nfs_opts);
    }

    if (not scalar(@nfs_options)) {
        NATE::BaseException->throw("Could not generate NFS options");
    }

    return \@nfs_options;
}

###############################################################################
# Method:    find_nfs_options_match()
# Objective: Given an array of nfs option hashes and a set of conditions to search against,
#            return back the first nfs option hash to match.
# Return:    The nfs options hash that matched the conditions
###############################################################################
sub find_nfs_options_match {
    my ($self, %opts) = @_;

    my $all_nfs_options = delete($opts{all_nfs_options});
    if (!$all_nfs_options) {
        NATE::Exceptions::Argument->throw("Missing required parameter 'all_nfs_options'");
    }

    # parse out the conditions to use when searching
    my $cluster_name = delete($opts{cluster_name});
    my $vserver_name = delete($opts{vserver_name});
    my $lif_address = delete($opts{lif_address});
    my $lif_name = delete($opts{lif_name});
    my $volume_name = delete($opts{volume_name});
    my $qtree_name = delete($opts{qtree_name});
    my $junction_path = delete($opts{junction_path});
    my $volume_lif_affinity = delete($opts{volume_lif_affinity}); # any | local | remote

    # walk each nfs option, and attempt to find matches for each item above
    foreach my $nfs_options (@{$all_nfs_options}) {
        if ($cluster_name) {
            next if ($cluster_name !~ $nfs_options->{cluster_filter});
        }
        if ($vserver_name) {
            next if ($vserver_name !~ $nfs_options->{vserver_filter});
        }
        if ($lif_address) {
            next if ($lif_address !~ $nfs_options->{lif_address_filter});
        }
        if ($lif_name) {
            next if ($lif_name !~ $nfs_options->{lif_name_filter});
        }
        if ($volume_name) {
            next if ($volume_name !~ $nfs_options->{volume_filter});
        }
        if ($qtree_name) {
            next if ($qtree_name !~ $nfs_options->{qtree_filter});
        }
        if ($junction_path) {
            next if ($junction_path !~ $nfs_options->{junction_path_filter});
        }
        if ($volume_lif_affinity) {
            # if the user specified local or remote affinity and this volume
            # lif affinity does not match
            next if ($nfs_options->{volume_lif_affinity} ne 'any' and
                     $volume_lif_affinity ne $nfs_options->{volume_lif_affinity});
        }
        # found a match!
        return $nfs_options;
    }

    return undef;
}

###############################################################################
# Method:    _translate_nfs_options_for_mount()
# Objective: This method will translate an NFS options hash into parameters
#            that can be passed into NACL::C::Client::Mount->create()
# Return:    A translated nfs options hash.
###############################################################################
sub _translate_nfs_options_for_mount {
    my ($self, %opts) = @_;

    my $options = delete($opts{nfs_options});

    my $hash = {};
    my %lookup = (
        mount_options => 'mountopts',
        transport_protocol => 'proto',
        nfs_version => 'vers',
        persistent => 'persistent'
    );
    foreach my $key (keys %{$options}) {
        if (exists $lookup{$key} and $options->{$key}) {
            $hash->{$lookup{$key}} = $options->{$key};
        }
    }

    return $hash;
}

###############################################################################
# Method:    get_vserver_names
# Objective: return a filtered list of all vservers on a sanmgr object
###############################################################################
sub get_vserver_names {
    my ($self, %args) = @_;
    $Log->enter();
    %args = &SanUtil::strip_dash(%args);
    try {
        %args = Params::Validate::validate_with(
            params => \%args,
            spec   => {
                cluster_name => {
                    type     => SCALAR,
                    optional => 1,
                    default  => ""
                },
                type => {
                    type     => SCALAR,
                    optional => 1,
                    default  => "all"
                },
            }
        );
    }
    otherwise {
        my $e = shift;
        $self->_set_err("Exception caught: " . $e->text());
    };
    if ($self->error()) {
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    my $cluster_name = $args{cluster_name};
    my $type = $args{type};

    # Get the hash of clusters from the XML
    my $xml_clusters = &ParamsMgr::search(
        -search_path => [],
        -key         => "CLUSTERS",
        -key_filter  => qr/^CLUSTER\d*$/,
        paramset => $self->EnGenX::Util::get_paramset()
    );
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    # iterate over all clusters
    my @vservers = ();
    foreach my $cluster (values %{$xml_clusters}) {

        # See if the cluster from the xml is one that the
        # caller wanted.
        if ($cluster_name ne "") {
            unless ($cluster->{NAME} eq "$cluster_name") {

                # Looks like the user wanted to scope to a
                # cluster and this cluster doesn't match that
                # name
                next;
            }
        }

        $cluster->{_INDEX} = "" if (not defined($cluster->{_INDEX}));
        my $vservers = &ParamsMgr::search(
            -search_path => ["CLUSTERS"],
            -key         => "CLUSTER" . $cluster->{_INDEX},
            -key_filter  => qr/^VSERVER\d*$/,
            paramset => $self->EnGenX::Util::get_paramset()
        );
        if (&ParamsMgr::error()) {
            $self->_set_err(&ParamsMgr::get_error());
            $Log->exit();
            return (wantarray()) ? () : [];
        }

        # iterate over all vservers
        foreach my $vserver (values %{$vservers}) {

            # filter vserver out if it doesnt match a specified type
            if (defined($type) and $type !~ /^all$|^\s*$/i) {

               # skip this vserver if the type doesnt match the requested type
                next if ($vserver->{VS_TYPE} !~ /^$type$/i);
            }

            # if this is a single vserver
            if ($vserver->{COUNT} == 1) {
                push(@vservers, $vserver->{NAME});
            } else {

                # handle multiple vservers that are related to this client
                foreach my $i (0 .. ($vserver->{COUNT} - 1)) {
                    push(@vservers, $vserver->{NAME} . "_$i");
                }
            }
        }
    }
    $Log->exit();
    return (wantarray()) ? @vservers : \@vservers;
}

sub _init_vs_client_relationship_svc {
    my $self = shift;

    require EnGenX::Services::VserverClientRelationship;

    $self->{vs_cli_relationship_svc} = new EnGenX::Services::VserverClientRelationship(engenx_obj => $self);
    $self->{vs_cli_relationship_svc_endpoint} = $self->{vs_cli_relationship_svc}->init();

    return;
}

sub _get_all_client_paramsets {
    my ($self, %args) = @_;

    my $myclients = $self->_get_all_clients_paramsets();

    my @all_clients;
    foreach my $clients_param (@{$myclients}){
        # find all the CLIENT elements in this CLIENTS element
        foreach my $key (keys %{$clients_param}) {
            next if ($key !~ /^CLIENT/);
            push(@all_clients, $clients_param->{$key});
        }
    }

    return \@all_clients;
}

sub _get_all_clients_paramsets {
    my ($self) = @_;

    my $global_set = $self->EnGenX::Util::get_paramset();

    # walk all the CLIENTS elements
    # CLIENTS elements can appear in two cases:
    # case (1) : CLIENTS
    # case (2) : CLIENTS1, CLIENTS2, ..., CLIENTSn

    my @myclients = ();
    my $clients_elm = $global_set->get("CLIENTS");
    if ($clients_elm and ref($clients_elm)) {
        # case (1)
        push(@myclients, $clients_elm);
    } else {
        # case (2)
        my $i = 1;
        while (1) {
            $clients_elm = $global_set->get("CLIENTS" . $i);
            if ($clients_elm and ref($clients_elm)) {
                push(@myclients, $clients_elm);
                $i++;
            } else {
                # we reached the end, no more CLIENTS elements break out
                last;
            }
        }
    }

    return \@myclients;
}

###############################################################################
# Method:    normalize_fcp_wwn
# Objective: Normalize an FCP WWN (add colons, lowercase it)
###############################################################################
sub normalize_fcp_wwn {
    my ($self, %args) = @_;

    $self->_clear_err();

    &ParamsMgr::process_args(%args);
    my $wwn = &ParamsMgr::get_arg(-arg => "wwn",
                                  -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return undef;
    }

    # Just in case they passed in one with colons, go ahead and strip out any that exist.
    $wwn =~ s/\://g;

    my $count   = 0;
    my $new_wwn = "";
    while ($count < 14) {
        $new_wwn = $new_wwn . substr($wwn, $count, 2) . ":";
        $count += 2;
    }
    $new_wwn = $new_wwn . substr($wwn, $count, 2);

    return lc($new_wwn);
}

###############################################################################
# Method:    assign_sanclient_options
# Objective: Given a sanclient object and it's paramset, assign all sub options
###############################################################################
sub assign_sanclient_options {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $client_obj = &ParamsMgr::get_arg(-arg => "client_obj", required => 1);
    my $client_ds = &ParamsMgr::get_arg(-arg => "datastruct", required => 1);
    my $env_ds = &ParamsMgr::get_arg(-arg => "env_datastruct", required => 1);
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw("Failed to parse args : " . &ParamsMgr::get_error());
    }

    # If initiators are passed in, copy them to the client object
    if (exists($client_ds->{INITIATORS})) {
        $client_obj->{'user_defined_initiators'} = $client_ds->{INITIATORS};
    }

    # obtain the iscsi options from the local iscsi_options datastructure if it exists
    # on the client
    if (exists($client_ds->{ISCSI_OPTIONS})) {
        &ParamsMgr::process_args(%{$client_ds->{ISCSI_OPTIONS}});
    } else {
        # otherwise, reset the process args datastructure in ParamsMgr
        # to prevent any overlap
        &ParamsMgr::process_args();
    }

    # create a callback to handle searching for the iscsi parameter in the
    # global iscsi options datastructure
    my $callback = sub {
        my ($args) = @_;
        my ($param, $default) = @{$args};

        # Process the global iscsi options element in the environtment ds
        &ParamsMgr::process_args(%{$env_ds->{GLOBAL_ISCSI_OPTIONS}},
                                 -CALLER => "global_iscsi_options");

        # parse out the parameter in the global iscsi options tag
        return &ParamsMgr::get_arg(-arg => $param,
                                   -default => $default,
                                   -CALLER => "global_iscsi_options");
    };

    # load in all the options
    my $iscsi_opts = {};
    $iscsi_opts->{CONFIG_ISCSI_DISCOVERY} = &ParamsMgr::get_arg(-arg => "DISCOVERY", -default => $callback, -code_args => ["DISCOVERY"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_DISCOVERY_TYPE} = &ParamsMgr::get_arg(-arg => "DISCOVERY_TYPE", -default => $callback, -code_args => ["DISCOVERY_TYPE"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} = &ParamsMgr::get_arg(-arg => "DISCOVERY_PRECLEANUP", -default => $callback, -code_args => ["DISCOVERY_PRECLEANUP"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_TYPE} = &ParamsMgr::get_arg(-arg => "INITIATOR_TYPE", -default => $callback, -code_args => ["INITIATOR_TYPE"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_SUBNET_MATCHING} = &ParamsMgr::get_arg(-arg => "SUBNET_MATCHING", -default => $callback, -code_args => ["SUBNET_MATCHING"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_USE_MGMT_PORT} = &ParamsMgr::get_arg(-arg => "USE_MGMT_PORT", -default => $callback, -code_args => ["USE_MGMT_PORT"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_HARDWARE_VENDOR} = &ParamsMgr::get_arg(-arg => "HARDWARE_VENDOR", -default => $callback, -code_args => ["HARDWARE_VENDOR"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_CHAP_MODE} = &ParamsMgr::get_arg(-arg => "CHAP_MODE", -default => $callback, -code_args => ["CHAP_MODE"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_CHAP_USERNAME} = &ParamsMgr::get_arg(-arg => "CHAP_IN_USERNAME", -default => $callback, -code_args => ["CHAP_IN_USERNAME"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_CHAP_USERNAME} = &ParamsMgr::get_arg(-arg => "CHAP_OUT_USERNAME", -default => $callback, -code_args => ["CHAP_OUT_USERNAME"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_CHAP_IN_PASSWORD} = &ParamsMgr::get_arg(-arg => "CHAP_IN_PASSWORD", -default => $callback, -code_args => ["CHAP_IN_PASSWORD"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_CHAP_OUT_PASSWORD} = &ParamsMgr::get_arg(-arg => "CHAP_OUT_PASSWORD", -default => $callback, -code_args => ["CHAP_OUT_PASSWORD"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_DATA_DIGEST} = &ParamsMgr::get_arg(-arg => "DATA_DIGEST", -default => $callback, -code_args => ["DATA_DIGEST"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_HEADER_DIGEST} = &ParamsMgr::get_arg(-arg => "HEADER_DIGEST", -default => $callback, -code_args => ["HEADER_DIGEST"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_ERL} = &ParamsMgr::get_arg(-arg => "ERL", -default => $callback, -code_args => ["ERL"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_ISNS_SERVER} = &ParamsMgr::get_arg(-arg => "ISNS_SERVER", -default => $callback, -code_args => ["ISNS_SERVER"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_IPVERSION} = &ParamsMgr::get_arg(-arg => "IP_VERSION", -default => $callback, -code_args => ["IP_VERSION"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_IPV6_EXCLUDE_LINK_LOCAL} = &ParamsMgr::get_arg(-arg => "IPV6_EXCLUDE_LINK_LOCAL", -default => $callback, -code_args => ["IPV6_EXCLUDE_LINK_LOCAL"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_DEFAULT_TIME2WAIT} = &ParamsMgr::get_arg(-arg => "DEFAULT_TIME2WAIT", -default => $callback, -code_args => ["DEFAULT_TIME2WAIT"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_DEFAULT_TIME2RETAIN} = &ParamsMgr::get_arg(-arg => "DEFAULT_TIME2RETAIN", -default => $callback, -code_args => ["DEFAULT_TIME2RETAIN"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_INITIAL_R2T} = &ParamsMgr::get_arg(-arg => "INITIAL_R2T", -default => $callback, -code_args => ["INITIAL_R2T"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_IMMEDIATE_DATA} = &ParamsMgr::get_arg(-arg => "IMMEDIATE_DATA", -default => $callback, -code_args => ["IMMEDIATE_DATA"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_WIN_MPIO} = &ParamsMgr::get_arg(-arg => "WIN_MPIO", -default => $callback, -code_args => ["WIN_MPIO"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_WIN_PERSISTENT} = &ParamsMgr::get_arg(-arg => "WIN_PERSISTENT", -default => $callback, -code_args => ["WIN_PERSISTENT"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_WIN_KEEP_TARGET_PORTALS} = &ParamsMgr::get_arg(-arg => "WIN_KEEP_TARGET_PORTALS", -default => $callback, -code_args => ["WIN_KEEP_TARGET_PORTALS"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_WIN_MAX_CONNECTS} = &ParamsMgr::get_arg(-arg => "WIN_MAX_CONNECTS", -default => $callback, -code_args => ["WIN_MAX_CONNECTS"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_WIN_NUM_SESSIONS} = &ParamsMgr::get_arg(-arg => "WIN_NUM_SESSIONS", -default => $callback, -code_args => ["WIN_NUM_SESSIONS"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_WIN_NUM_CONNS_PER_SESSION} = &ParamsMgr::get_arg(-arg => "WIN_NUM_CONNS_PER_SESSION", -default => $callback, -code_args => ["WIN_NUM_CONNS_PER_SESSION"], -eval_code => 1);
    $iscsi_opts->{CONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET} = &ParamsMgr::get_arg(-arg => "FILTER_TARGETS_BY_PORTSET", -default => $callback, -code_args => ["FILTER_TARGETS_BY_PORTSET"], -eval_code => 1);

    # assign the iscsi options on the sanclient object
    foreach my $key (keys %{$iscsi_opts}) {
        $client_obj->{iscsi}->{$key} = $iscsi_opts->{$key};
    }

    # Obtain FCP options from either the local tag or the environment
    if (exists($client_ds->{FCP_OPTIONS})) {
        &ParamsMgr::process_args(%{$client_ds->{FCP_OPTIONS}});
    } else {
        # otherwise, reset the process args datastructure in ParamsMgr
        # to prevent any overlap
        &ParamsMgr::process_args();
    }

    # create a callback to handle searching for the fcp parameter in the
    # global fcp options datastructure
    $callback = sub {
        my ($args) = @_;
        my ($param, $default) = @{$args};

        # Process the global fcp options element in paramsmgr
        &ParamsMgr::process_args(%{$env_ds->{GLOBAL_FCP_OPTIONS}},
                                 -CALLER => "global_fcp_options");

        # parse out the parameter in the global fcp options tag
        return &ParamsMgr::get_arg(-arg => $param,
                                   -default => $default,
                                   -CALLER => "global_fcp_options");
    };

    # assign the fcp options to the object
    my $fcp_opts = {};
    $fcp_opts->{CONFIG_FCP_DRIVER_TYPE} =
        &ParamsMgr::get_arg(-arg => "DRIVER_TYPE",
                            -default => $callback,
                            -code_args => ["DRIVER_TYPE"],
                            -eval_code => 1);

    $fcp_opts->{CONFIG_FCP_AIX_DYNAMIC_TRACKING} =
        &ParamsMgr::get_arg(-arg => "AIX_DYNAMIC_TRACKING",
                            -default => $callback,
                            -code_args => ["AIX_DYNAMIC_TRACKING"],
                            -eval_code => 1);

    $fcp_opts->{CONFIG_FCP_AIX_NUM_CMD_ELEMS} =
        &ParamsMgr::get_arg(-arg => "AIX_NUM_CMD_ELEMS",
                            -default => $callback,
                            -code_args => ["AIX_NUM_CMD_ELEMS"],
                            -eval_code => 1);

    $fcp_opts->{CONFIG_FCP_AIX_ERROR_RECOVERY} =
        &ParamsMgr::get_arg(-arg => "AIX_ERROR_RECOVERY",
                            -default => $callback,
                            -code_args => ["AIX_ERROR_RECOVERY"],
                            -eval_code => 1);

    $fcp_opts->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET} =
        &ParamsMgr::get_arg(-arg => "FILTER_TARGETS_BY_PORTSET",
                            -default => $callback,
                            -code_args => ["FILTER_TARGETS_BY_PORTSET"],
                            -eval_code => 1);

    $fcp_opts->{CONFIG_FCP_ZONES_PER_NODE_PER_INIT} =
        &ParamsMgr::get_arg(-arg => "ZONES_PER_NODE_PER_INIT",
                            -default => $callback,
                            -code_args => ["ZONES_PER_NODE_PER_INIT"],
                            -eval_code => 1);

    $fcp_opts->{CONFIG_FCP_MULTI_TARGET_ZONING} =
        &ParamsMgr::get_arg(-arg => "MULTI_TARGET_ZONING",
                            -default => $callback,
                            -code_args => ["MULTI_TARGET_ZONING"],
                            -eval_code => 1);

    $fcp_opts->{CONFIG_FCP_PERFORM_ZONING} =
        &ParamsMgr::get_arg(-arg => "PERFORM_ZONING",
                            -default => $callback,
                            -code_args => ["PERFORM_ZONING"],
                            -eval_code => 1);

    $fcp_opts->{CONFIG_FCP_DONT_CREATE_TARGET_ALIASES} =
        &ParamsMgr::get_arg(-arg => "DONT_CREATE_TARGET_ALIASES",
                            -default => $callback,
                            -code_args => ["DONT_CREATE_TARGET_ALIASES"],
                            -eval_code => 1);

    # assign the fcp options on the sanclient object
    foreach my $key (keys %{$fcp_opts}) {
        $client_obj->{fcp}->{$key} = $fcp_opts->{$key};
    }

    # obtain the vx options from the local vx_options datastructure if it exists
    # on the client
    if (exists($client_ds->{VX_OPTIONS})) {
        &ParamsMgr::process_args(%{$client_ds->{VX_OPTIONS}});
    } else {
        # otherwise, reset the process args datastructure in ParamsMgr
        # to prevent any overlap
        &ParamsMgr::process_args();
    }

    # create a callback to handle searching for the vx parameter in the
    # global vx options datastructure
    $callback = sub {
        my ($args) = @_;
        my ($param, $default) = @{$args};

        # Process the global vx options element in paramsmgr
        &ParamsMgr::process_args(%{$env_ds->{GLOBAL_VX_OPTIONS}}, -CALLER => "global_vx_options");

        # parse out the parameter in the global vx options tag
        return &ParamsMgr::get_arg(-arg     => $param,
                                   -default => $default,
                                   -CALLER  => "global_vx_options");
    };

    # assign the fcp options to the object
    my $vrts_opts = {};
    $vrts_opts->{vx_restore_interval} =
        &ParamsMgr::get_arg(-arg => "DMP_RESTORE_INTERVAL",
                            -default => $callback,
                            -code_args => ["DMP_RESTORE_INTERVAL"],
                            -eval_code => 1);

    $vrts_opts->{vx_restore_policy} =
        &ParamsMgr::get_arg(-arg => "DMP_RESTORE_POLICY",
                            -default => $callback,
                            -code_args => ["DMP_RESTORE_POLICY"],
                            -eval_code => 1);

    $vrts_opts->{vx_50_fast_recovery} =
        &ParamsMgr::get_arg(-arg => "DMP_FAST_RECOVERY",
                            -default => $callback,
                            -code_args => ["DMP_FAST_RECOVERY"],
                            -eval_code => 1);

    $vrts_opts->{vx_50_dmp_log_level} =
        &ParamsMgr::get_arg(-arg => "DMP_LOG_LEVEL",
                            -default => $callback,
                            -code_args => ["DMP_LOG_LEVEL"],
                            -eval_code => 1);

    $vrts_opts->{vx_51_lun_retry_timeout} =
        &ParamsMgr::get_arg(-arg => "DMP_LUN_RETRY_TIMEOUT",
                            -default => $callback,
                            -code_args => ["DMP_LUN_RETRY_TIMEOUT"],
                            -eval_code => 1);

    $vrts_opts->{vx_50_monitor_fabric} =
        &ParamsMgr::get_arg(-arg => "DMP_MONITOR_FABRIC",
                            -default => $callback,
                            -code_args => ["DMP_MONITOR_FABRIC"],
                            -eval_code => 1);

    $vrts_opts->{vx_50_dmp_path_age} =
        &ParamsMgr::get_arg(-arg => "DMP_PATH_AGE",
                            -default => $callback,
                            -code_args => ["DMP_PATH_AGE"],
                            -eval_code => 1);

    $vrts_opts->{vx_50_probe_idle_lun} =
        &ParamsMgr::get_arg(-arg => "DMP_PROBE_IDLE_LUN",
                            -default => $callback,
                            -code_args => ["DMP_PROBE_IDLE_LUN"],
                            -eval_code => 1);

    $vrts_opts->{vx_50_queue_depth} =
        &ParamsMgr::get_arg(-arg => "DMP_QUEUE_DEPTH",
                            -default => $callback,
                            -code_args => ["DMP_QUEUE_DEPTH"],
                            -eval_code => 1);

    $vrts_opts->{vx_io_policy} =
        &ParamsMgr::get_arg(-arg => "IO_POLICY",
                            -default => $callback,
                            -code_args => ["IO_POLICY"],
                            -eval_code => 1);

    $vrts_opts->{vx_power_fail_timeout} =
        &ParamsMgr::get_arg(-arg => "POWER_FAIL_TIMEOUT",
                            -default => $callback,
                            -code_args => ["POWER_FAIL_TIMEOUT"],
                            -eval_code => 1);

    $vrts_opts->{vx_50_recovery_option} =
        &ParamsMgr::get_arg(-arg => "RECOVERY_OPTION",
                            -default => $callback,
                            -code_args => ["RECOVERY_OPTION"],
                            -eval_code => 1);

    # Now set the the options on the object.  We can't use the 'options' method since we may have
    # a fake object.  See the SanMgr::Client::load_resources() method for an explanation as to
    # why it creates the fake object.
    foreach my $key (keys %$vrts_opts) {
        $client_obj->{$key} = $vrts_opts->{$key};
    }

    # obtain the Linux dmmp options from the local dmmp_options datastructure if it exists
    # on the client
    if (exists($client_ds->{DMMP_OPTIONS})) {
        &ParamsMgr::process_args(%{$client_ds->{DMMP_OPTIONS}});
    } else {
        # otherwise, reset the process args datastructure in ParamsMgr
        # to prevent any overlap
        &ParamsMgr::process_args();
    }

    # create a callback to handle searching for the dmmp parameter in the
    # global dmmp options datastructure
    $callback = sub {
        my ($args) = @_;
        my ($param, $default) = @{$args};

        # Process the global dmmp options element in paramsmgr
        &ParamsMgr::process_args(%{$env_ds->{GLOBAL_DMMP_OPTIONS}},
                                 -CALLER => "global_dmmp_options");

        # parse out the parameter in the global dmmp options tag
        return &ParamsMgr::get_arg(-arg     => $param,
                                   -default => $default,
                                   -CALLER  => "global_dmmp_options");
    };

    # assign the dmmp options to the object
    my $dmmp_opts = {};
    $dmmp_opts->{config_dmmp_use_user_friendly_names} =
        &ParamsMgr::get_arg(-arg => "USE_USER_FRIENDLY_NAMES",
                            -default => $callback,
                            -code_args => ["USE_USER_FRIENDLY_NAMES"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_max_fds} =
        &ParamsMgr::get_arg(-arg => "MAX_FDS",
                            -default => $callback,
                            -code_args => ["MAX_FDS"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_queue_without_daemon} =
        &ParamsMgr::get_arg(-arg => "QUEUE_WITHOUT_DAEMON",
                            -default => $callback,
                            -code_args => ["QUEUE_WITHOUT_DAEMON"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_flush_on_last_del} =
        &ParamsMgr::get_arg(-arg => "FLUSH_ON_LAST_DEL",
                            -default => $callback,
                            -code_args => ["FLUSH_ON_LAST_DEL"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_pg_prio_calc} =
        &ParamsMgr::get_arg(-arg => "PG_PRIO_CALC",
                            -default => $callback,
                            -code_args => ["PG_PRIO_CALC"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_dev_loss_tmo} =
        &ParamsMgr::get_arg(-arg => "DEV_LOSS_TMO",
                            -default => $callback,
                            -code_args => ["DEV_LOSS_TMO"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_fast_io_fail_tmo} =
        &ParamsMgr::get_arg(-arg => "FAST_IO_FAIL_TMO",
                            -default => $callback,
                            -code_args => ["FAST_IO_FAIL_TMO"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_path_grouping_policy} =
        &ParamsMgr::get_arg(-arg => "PATH_GROUPING_POLICY",
                            -default => $callback,
                            -code_args => ["PATH_GROUPING_POLICY"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_features} =
        &ParamsMgr::get_arg(-arg => "FEATURES",
                            -default => $callback,
                            -code_args => ["FEATURES"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_path_checker} =
        &ParamsMgr::get_arg(-arg => "PATH_CHECKER",
                            -default => $callback,
                            -code_args => ["PATH_CHECKER"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_rr_weight} =
        &ParamsMgr::get_arg(-arg => "RR_WEIGHT",
                            -default => $callback,
                            -code_args => ["RR_WEIGHT"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_rr_min_io} =
        &ParamsMgr::get_arg(-arg => "RR_MIN_IO",
                            -default => $callback,
                            -code_args => ["RR_MIN_IO"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_hardware_handler} =
        &ParamsMgr::get_arg(-arg => "HARDWARE_HANDLER",
                            -default => $callback,
                            -code_args => ["HARDWARE_HANDLER"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_prio_callout} =
        &ParamsMgr::get_arg(-arg => "PRIO_CALLOUT",
                            -default => $callback,
                            -code_args => ["PRIO_CALLOUT"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_prio} =
        &ParamsMgr::get_arg(-arg => "PRIO",
                            -default => $callback,
                            -code_args => ["PRIO"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_failback} =
        &ParamsMgr::get_arg(-arg => "FAILBACK",
                            -default => $callback,
                            -code_args => ["FAILBACK"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_getuid_callout} =
        &ParamsMgr::get_arg(-arg => "GETUID_CALLOUT",
                            -default => $callback,
                            -code_args => ["GETUID_CALLOUT"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_path_selector} =
        &ParamsMgr::get_arg(-arg => "PATH_SELECTOR",
                            -default => $callback,
                            -code_args => ["PATH_SELECTOR"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_product} =
        &ParamsMgr::get_arg(-arg => "PRODUCT",
                            -default => $callback,
                            -code_args => ["PRODUCT"],
                            -eval_code => 1);

    $dmmp_opts->{config_dmmp_vendor} =
        &ParamsMgr::get_arg(-arg => "VENDOR",
                            -default => $callback,
                            -code_args => ["VENDOR"],
                            -eval_code => 1);

    # Now set the the options on the object.  We can't use the 'options' method since we may have
    # a fake object.  See the SanMgr::Client::load_resources() method for an explanation as to
    # why it creates the fake object.
    foreach my $key (keys %$dmmp_opts) {
        $client_obj->{$key} = $dmmp_opts->{$key};
    }

    return TRUE;
}

###############################################################################
# Method:    get_nacl_obj
# Objective: Helper method that create a new instance of a specific NACL obj.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_nacl_obj {
    my ($self, @args) = @_;
    $Log->enter();

    # Locals
    my $obj        = undef;
    my $node       = undef;
    my %valid_args = ();

    # Validate Argumetns
    try {
        %valid_args = Params::Validate::validate(
            @args,
            {   type     => {type => SCALAR,  optional => 0},
                resource => {type => HASHREF, optional => 0},
                node => {type => HASHREF, default => undef, optional => 1},
            }
        );
    }
    otherwise {
        my $exception = shift;
        NATE::Exceptions::Argument->throw($exception->text());
    };

    my $type     = $valid_args{type};
    my $resource = $valid_args{resource};

    if (!defined($node = $valid_args{'node'})) {
        $node = NACL::C::Node->find();
    }

    # Switch/Case implementation
    # ControllerCluster equivalent Componentes
    try {
        if ($type =~ /aggr|aggregate/i) {
            $obj = NACL::C::StorageAggregate->new(
                'command_interface' => $node,
                'aggregate'         => $resource->{'name'}
            );
        } elsif ($type =~ /vserver/i) {
            $obj = NACL::C::Vserver->new(
                'command_interface' => $node,
                'vserver'           => $resource->{'name'}
            );
        } elsif ($type =~ /volume|vol/i) {
            $obj = NACL::C::Volume->new(
                'command_interface' => $node,
                'vserver'           => $resource->{'vserver'},
                'volume'            => $resource->{'name'},
            );
        } elsif ($type =~ /qtree/i) {
            $obj = NACL::C::VolumeQtree->new(
                'command_interface' => $node,
                'vserver'           => $resource->{'vserver'},
                'volume'            => $resource->{'volume'},
                'qtree'             => $resource->{'name'},
            );
        } elsif ($type =~ /lun/) {
            my $volume = $resource->{'volume'};
            my $qtree  = $resource->{'qtree'};
            my $lun    = $resource->{'name'};
            my $path   = '';
            if ($qtree) {
                $path    = "/vol/" . $volume . "/" . $qtree . "/" . $lun;
            } else {
                $path    = "/vol/" . $volume . "/" . $lun;
            }
            $obj = NACL::C::Lun->new(
                'command_interface' => $node,
                'vserver'           => $resource->{'vserver'},
                'path'              => $path
            );
        } elsif ($type =~ /igroup/i) {
            $obj = NACL::C::LunIgroup->new(
                command_interface => $node,
                vserver           => $resource->{'vserver'},
                igroup            => $resource->{'name'}
            );
        } elsif ($type =~ /portset/i) {
            $obj = NACL::C::LunPortset->new(
                'command_interface' => $node,
                'vserver'           => $resource->{'vserver'},
                'portset'           => $resource->{'name'}
            );
        } elsif ($type =~ /lif/i) {
            $obj = NACL::C::NetworkInterface->new(
                'command_interface' => $node,
                'vserver'           => $resource->{'vserver'},
                'lif'               => $resource->{'name'}
            );
        } elsif ($type =~ /tpgroup/i) {
            $obj = NACL::C::VserverIscsiTpgroup->new(
                'command_interface' => $node,
                'vserver'           => $resource->{'vserver'},
                'tpgroup'           => $resource->{'name'}
            );

            # Other equivalent Componentes
        } elsif ($type =~ /node/i) {
            my %args = ('command_interface' => $node);
            $args{'name'} = $resource->{'name'}
                if (defined $resource->{'name'});
            $args{'node'} = $resource->{'node'}
                if (defined $resource->{'node'});
            $args{'hostrec'} = $resource->{'hostrec'}
                if (defined $resource->{'hostrec'});
            $obj = NACL::C::SystemNode->new(%args);
        } elsif ($type =~ /switch/i) {
            my %args = {};
            $args{'name'} = $resource->{'name'}
                if (defined $resource->{'name'});
            $args{'switch'} = $resource->{'switch'}
                if (defined $resource->{'switch'});
            $args{'hostrec'} = $resource->{'hostrec'}
                if (defined $resource->{'hostrec'});
            $obj = NACL::C::Switch->new(%args);
        } elsif ($type =~ /cserver/i) {
            my %args = ('command_interface' => $node);
            $args{'name'} = $resource->{'name'}
                if (defined $resource->{'name'});
            $args{'cserver'} = $resource->{'cserver'}
                if (defined $resource->{'cserver'});
            $args{'hostrec'} = $resource->{'hostrec'}
                if (defined $resource->{'hostrec'});
            $obj = NACL::C::Cserver->new(%args);
        } elsif ($type =~ /cluster/i) {
            my %args = ('command_interface' => $node);
            $args{'name'} = $resource->{'name'}
                if (defined $resource->{'name'});
            $args{'node'} = $resource->{'node'}
                if (defined $resource->{'node'});
            $args{'hostrec'} = $resource->{'hostrec'}
                if (defined $resource->{'hostrec'});
            $obj = NACL::C::Cluster->new(%args);
        } else {
            NATE::BaseException->throw(
                "Type not recognized or supported '$type'");
        }
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        NATE::BaseException->throw(
                  "Exception/Error FOUND while creating a NEW instance"
                . " of NACL::C::$type - Error: \n"
                . $exception);
    };

    $Log->exit();
    return $obj;
} ## end sub get_nacl_obj

#==============================================================================
# Method:    _get_objs_sorter
# Objective:  This method is used to isolate common code between get_objects
#             for in memory objects like clients, swhitches, etc. and Controller
#             cluster objects (that as of now are NACL)
# Details:    Steps: 1) Based on the number of categories, the calls will be
#                       recursively called.
#                    2) Either calls '_get_cached_objs' or '_retrieve_nacl_obj'
# Inputs:     Takes a reference to an array or a scalar for "Category".
# Outputs:    Returns and array with the requested objects.
# Exceptions: This method throws an exceptions when a required argument is not
#             provided.
#             This method throws an exceptions when errors in the calling
#             private subrutines were encountered AND no object was found.
#             NOTE: This logic expects to throw and catch errors. The logic will
#                   use BRUT FORCE to find some objs when multiple CATEGORIES
#                   are provided
#==============================================================================
sub _get_objs_sorter {
    my ($self, @args) = @_;
    $Log->enter();

    # Local variables
    my %valid_args = ();
    my @objs       = ();
    my @errors     = ();
    my $category   = ".*";
    my @types      = ();

    try {
        %valid_args = Params::Validate::validate_with(
            params => \@args,
            spec => {category    => {type => SCALAR,   optional => 1},
                     categories  => {type => ARRAYREF, optional => 1},
                     type        => {type => SCALAR,   optional => 1},
                     types       => {type => ARRAYREF, optional => 1}},
            allow_extra => 1,
            normalize_keys =>
                sub { my $k = shift; $k =~ s/^-//; return lc $k },
        );
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        NATE::Exceptions::Argument->throw($exception->text());
    };

    # Either -category or -categories could be specified.
    # Will consolidate both in the string category and work from there, but we need to
    # verify that 'category' has been defined before converting categories into a string.
    # since get resource uses regexes we are simple putting an or inbetween the different
    # categories  [client,hypervisor] becomes "client|hypervisor"
     if (defined($valid_args{category})){
        $category = delete($valid_args{category});
    } elsif(defined($valid_args{categories} and scalar @{$valid_args{categories}})) {
        $category = join("|", @{delete $valid_args{categories}});
    }

    # Either -type or -types must be specified.
    # Will consolidate both in the array (@types) and work from there, but we need to
    # verify that 'type' has been defined before pushing it to the array. Failing to do so,
    # we could be adding an 'undef' entry in the array that will cause the method to fail
    # Though get resource does except regexes we need to know the type when we create
    # the NACL objects so we will loop through the array instead of sending a regex with
    # all types wanted.
    if (defined($valid_args{type})){
        push(@types, delete($valid_args{type}));
    } else {
        @types = @{delete $valid_args{types}};
        if(not scalar(@types)) {
            my $err_mssg
            = ("-type or -types must be specified");
            $Log->exit();
            NATE::BaseException->throw($err_mssg);
        }
    }

    # If -name is passed in then push it onto the @names array.
    if (defined($valid_args{name})){
        push(@{$valid_args{names}}, delete $valid_args{name});
    }

  #TODO:
  #     All the objects should be defined in the Resources list, independently
  #     if they are clusters, clients or hypervisors. So we should first
  #     verify them in the ResourceList and then retrieve them from
  #     wherever they are (NACL or SanClient or etc.).
  #     If so, something similar to the following code could be implemented.
  #     (Code commented out between break and cut)

=break
    # Locals
    my $resource_list_obj = "";
    # Validate that the resource list is available.
    if (defined($self->{resource_list})) {
        $resource_list_obj = $self->{resource_list};
    }else {
        $Log->exit();
        NATE::BaseException->throw("The resource list has not been set up.");
    }
    # Verify the provided category exists.
    # TODO: This category validation should be inside ResourceList
    my $category_found = FALSE;
    foreach my $rl_catergory ($resource_list_obj->get_categories()){
        if ($rl_catergory =~/$category/i) {
            $category_found = TRUE;
            last;
        }
    }
    if (!$category_found) {
        $Log->exit();
        NATE::BaseException->throw("Could not find the provided name "
                                   . "'$category' as a category (class) key");
    }

    # With the category key, find the type the user is looking for.
    @resources = $resource_list_obj->get_resources('category' => $category
                                                   'type' => $type);

    # With resources Array either get objects from NACL or Memory.

=cut

    # Let's do this as many times as categories the user requested.
    foreach my $type (@types) {
        try {
            # All types are lowercase in EnGenX
            $type = lc($type);

            # Sort base on the type
            if ($type =~ /client|switch|hypervisor|fabric|eventman|node|cluster/i) {
                if($type =~ /node/i) {
                    $type = 'nacl_node';
                } elsif($type =~ /cluster/i) {
                    $type = 'nacl_cluster';
                }
                # Private Method call
                my @local_objs = $self->_get_cached_objs(
                    'type' => $type,
                    %valid_args
                );
                push(@objs, @local_objs);
            } else {
                my @local_objs = $self->_retrieve_nacl_obj(
                    'category' => $category,
                    'type' => $type,
                    %valid_args
                );
                push(@objs, @local_objs);
            }
        }
        otherwise {
            my $exception = shift;
            push(@errors, $exception->text());
        };
    }

    if (!(scalar(@objs)) && scalar(@errors)) {
        my $err_mssg
            = (   "Could not find an object with the provided Category, "
                . "Types or names: "
                . Dumper(\@errors));
        $Log->exit();
        NATE::BaseException->throw($err_mssg);
    }

    $Log->exit();
    return (wantarray()) ? @objs : \@objs;
}    # _get_objs_sorter

#==============================================================================
# Method:    _get_cached_objs
# Objective:  This method is used to retrieve from memorty the objects of type
#             catergory and if provided, that have the name specified.
# Details:    Steps: 1) Will get the EnGenX->{_objects} and will look for the
#                       requested category.
#                    2) With the found EnGenX->{_objects}->{category} will look
#                       for the requested names, if no names were requested,
#                       will return all objects unter the Key combination.
# Inputs:     A scalar specifying the "Category" and an optional array of names.
# Outputs:    Returns and array with the requested objects.
# Exceptions: This method throws an exceptions when a required argument is not
#             provided.
#             This method throws an exceptions when no EnGenx->{_objects} were
#             found.
#             This method throws an exceptions when no EnGenx->{_objects}->
#             {category} were found.
#             This method throws an exceptions when no EnGenx->{_objects}->
#             {category}->{name} were found.
#==============================================================================
sub _get_cached_objs {
    my ($self, %args) = @_;
    $Log->enter();

    # Local variables
    my %valid_args = ();
    my @objs       = ();
    my $data       = undef;

    $valid_args{'type'} = delete $args{type};
    $valid_args{'names'} = delete $args{names};

    if (not defined($valid_args{type})) {
        $Log->exit();
        NATE::Exceptions::Argument->throw("Missing required param 'type'");
    }

    #Assign Values
    my $type = lc($valid_args{'type'});

    # User must provide an array of names, even with only one name.
    my @names
        = (defined($valid_args{'names'}))
        ? @{$valid_args{'names'}}
        : ();

    # Get together all the data needed
    # If method is called from setup or cleanup, the path to objects
    # is slightly different
    # If the parent is not defined then we have the EnGenX object
    if (not defined($self->{parent})) {
        # If objects is defiend then get them.
        if (defined($self->{_objects})) {
            $data = $self->{_objects};
        # If objects is not defined then throw exception.
        } else {
            $Log->exit();
            NATE::BaseException->throw(
                "No Objects have been defined in EnGenX");
        }
    # If parent does exist then we should get the objects from the parent.
    } elsif (defined($self->{parent}->{_objects})) {
        $data = $self->{parent}->{_objects};
    # If the objects do not exist then throw exception.
    } else {
        $Log->exit();
            NATE::BaseException->throw(
                "No Objects have been defined in EnGenX");
    }

    # Check $self to see if it has the list type defined on it
    if (not defined($data->{$type})) {
        $Log->exit();
        NATE::BaseException->throw(
            "Object type[$type] has NOT been defined in EnGenX");
    }

    # If user sent a list of names (one or more), use it, if not, return the
    # entire list
    if (not(scalar(@names))) {
        @names = keys %{$data->{$type}};
    }

    foreach my $name (@names) {
        # Check to see if the object/name exists
        if (defined($data->{$type}->{$name})) {
            push(@objs, $data->{$type}->{$name});
        } else {
            $Log->exit();
            NATE::BaseException->throw("Invalid name specified ($name) for"
                    . " type[$type]");
        }
    }

    $Log->exit();
    return (wantarray()) ? @objs : \@objs;
}    #_get_cached_objs

#==============================================================================
# Method:    _retrieve_nacl_obj
# Objective:  This method is used to create an NACL instance of the requested
#             objects.
# Details:    Steps: 1) Re-Create the resource List.
#                    2) Get the categories from the resource list to verify
#                       the provided one.,
#                    3) Get an arrya of resources from the ResourceList.
#                    4) Create the new NACL objects recursively calling
#                       get_nacl_object.
# Inputs:     A scalar specifying the "Category", another scalar specifying the
#             "Type" (vserver, lun, vol, etc) and an optional array of names.
# Outputs:    Returns an array with the requested objects.
# Exceptions: This method throws an exceptions when a required argument is not
#             provided.
#             This method throws an exceptions when no Catergories in the
#             Resource List obj were found.
#             This method throws an exceptions when no Resources are retuned
#             from the Resource List obj were found.
#             The exceptions thrown in this method are caught by the parent
#             method _get_objs_sorter and to be stored in an error array. If
#             and only if no objs are returned at all does the error array get
#             printed.
#==============================================================================
sub _retrieve_nacl_obj {
    my ($self, @args) = @_;
    $Log->enter();

    # Local variables
    my %valid_args        = ();
    my @objs              = ();
    my $resource_list_obj = "";
    my @resources         = ();

    try {
        %valid_args = Params::Validate::validate_with(
            params => \@args,
            spec   => {
                category => {type => SCALAR,   optional => 0},
                type     => {type => SCALAR,   optional => 0},
                names    => {type => ARRAYREF, optional => 1}
            },
            allow_extra => 1
        );
    }
    otherwise {
        my $exception = shift;
        $Log->exit();
        NATE::Exceptions::Argument->throw($exception->text());
    };

    #Assign Values
    my $category = delete($valid_args{'category'});
    my $type     = delete($valid_args{'type'});

    # User must provide an array of names, even with only one name.
    my @names
        = (defined($valid_args{'names'}))
        ? @{delete($valid_args{'names'})}
        : ();

    # Validate that the resource list is available.
    if (defined($self->{resource_list})) {
        $resource_list_obj = $self->{resource_list};
    } elsif (defined($self->{parent}->{resource_list})) {
        $resource_list_obj = $self->{parent}->{resource_list};
    } else {
        $Log->exit();
        NATE::BaseException->throw("The resource list has not been set up.");
    }

    unless(scalar (grep (/$category/, @{$resource_list_obj->get_categories()}))){
        $Log->exit();
        NATE::BaseException->throw(
            "ResourceList obj does not contain any categories that match: \n" .
            $category);
    }

    # Create get resources params
    # Append eny extra data the user wants to use to find the
    # Also make the type an exact match so that it matches
    # SanMgr and things like Lun/LunMap dont overlap.
    my %params = (
        'category' => $category,
        'type'     => qr/^$type$/,
        %valid_args
    );

    # If user provided names, append them as extra data.
    if (scalar(@names) > 0) {
        $params{data}->{name} = join '|', @names;
        # With the newly created PARAMS, find the type the user is looking for.
        @resources = $resource_list_obj->get_resources(%params);
    } else {
        #No names provided, return all.
        @resources = $resource_list_obj->get_resources(%params);
    }

    # With the array of resources, create the NACL objects
    if (scalar(@resources)) {
        # Get the command interface
        my $node = NACL::C::Node->find();
        foreach my $resource (@resources) {

            # Create NACL objects
            my $obj = $self->get_nacl_obj(
                'type'     => $type,
                'resource' => $resource,
                'node'     => $node
            );
            push(@objs, $obj);
        }
    }  else {
        $Log->exit();
        NATE::BaseException->throw(
            "ResourceList obj returned an empty array of Resources when "
                . "using parameters: \n"
                . Dumper(\%params));
    }

    $Log->exit();
    return (wantarray()) ? @objs : \@objs;
} ## end sub _retrieve_nacl_obj

#==============================================================================
# Method:    _validate_obj_md5_hash
# Objective: Given an object and its previous md5 hash, verify the object
#==============================================================================
sub _validate_obj_md5_hash {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $obj                = &ParamsMgr::get_arg(-arg => "obj",
                                                 -required => 1);
    my $obj_md5_hex_before = &ParamsMgr::get_arg(-arg => "obj_md5_hex_before",
                                                 -required => 1);
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw("Argument Parsing failed", &ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # calculate the new md5 value
    my $obj_md5_hex_after = $self->_calculate_obj_md5_hash($obj);

    # verify the md5
    if ($obj_md5_hex_after ne $obj_md5_hex_before) {
        NATE::BaseException->throw("MD5 verification for " . $obj->hostname() . " did not pass ($obj_md5_hex_before " .
                        "!= $obj_md5_hex_after). Please file a burt against TESTLIB/san to add support for this OS.");
        return $EnGenX::FALSE;
    }

    return $EnGenX::TRUE;
} ## end sub _validate_obj_md5_hash

###############################################################################
# Method:    get_paramset
# Objective: Get the paramset from the EnGenX object if it exists otherwise
#            get the global paramset from paramsmgr
# Returns:   A nate paramset.
###############################################################################
sub get_paramset{
    my ($self, @opts) = @_;

    # If a parent is defined then recurse up till we get the EnGenX object.
    if (defined $self->{parent}) {
        return $self->{parent}->EnGenX::Util::get_paramset();
    } # The fabricmgr parent is in an Args hash. If the parents exists then
    # recurse up till we get the EnGenX object.
    elsif (defined $self->{args} and defined $self->{args}->{parent}) {
        return $self->{args}->{parent}->EnGenX::Util::get_paramset();
    }

    # Once we have the EnGenX object if a paramset exists in it then use
    # that paramset.
    if (defined $self->{nate_paramset}) {
        return $self->{nate_paramset};
    # Otherwise use the global paramset.
    } else {
        return &ParamsMgr::get_global_paramset();
    }
}

#==============================================================================
# Method:    _calculate_obj_md5_hash
# Objective: Given an object, calculate the md5 hash from it
#==============================================================================
sub _calculate_obj_md5_hash {
    my ($self, $obj) = @_;

    my $d_obj = Data::Dumper->new([$obj]);
    $d_obj->Sortkeys(1);
    $d_obj->Maxdepth(0);
    $d_obj->Useqq(0);

    my $obj_serialized = $d_obj->Dump;

    $obj_serialized =~ s/'//g;
    $obj_serialized =~ s/\s+//g;
    $obj_serialized =~ s/\n//g;

    if (param_global()->get("ENGENX_DISABLE_MD5_VERIFICATION")) {
        return "1";
    } else {
        return Digest::MD5::md5_hex($obj_serialized);
    }
}

=pod

=head1 Package Name

EnGenX::Util

=head1 Author

Michael M. Latif

=head1 Description

Provide a common set of functions to be used with SanMgr

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm>

B<Tharn.pm>

=head1 Public Methods

=head2 B<get_objects($self, %args)>

This function is used to retrieve the resource (client, switch, controller,...) objects.
Also can be used to retrieve objects by name.

=over 4

=item Inputs

=item * -self

Self is an SanMgr object

=item * -type

Type can be one of the following:
    client, switch, controller, vserver

=item * I<-name>

Name is optional, this input will filter the type provided to locate only the object with the
name requested.

=item Outputs

Returns back a hash of objects, keyed either by type (if no name provided) OR by name if a
name is provided.

=item Examples

my $client1_object = $self->get_objects(-type => "client", -name => "client1");

OR

my $hash_of_clients = $self->get_objects(-type => "client");

=back

=head2 B<get_target_list($self, %args)>

This function is used to retrieve the target list available from the VServers
defined within the environment/testbed/XML file.

=over 4

=item Inputs

=item * -self

Self is an SanMgr object

=item Outputs

Returns a hash with three keys defined, one is iscsi_targets, this is an array ref of
iscsi targets. The second is fcp_targets, this is an array ref of fcp targets.
The third vserver_objs is an array ref of VServer objects.
    $data->{iscsi_targets} = \@iSCSITargets;
    $data->{fcp_targets} = \@FcpTargets;
    $data->{vserver_objs} = \@VServer_objs;

=item Examples

my $data = $self->get_target_list();
my @iscsi_targets = @{$data->{iscsi_targets}};
my @fcp_targets = @{$data->{fcp_targets}};
my @vserver_objs = @{$data->{vserver_objs}};

=back

=head2 B<get_vserver_names(%args)>

This function is used to retrieve the list of vserver names
defined within the environment/testbed/XML file.

=over 4

=item Inputs

=item * I<-cluster_name>

Cluster is an optional string to narrow the scope of the vserver request

=item * I<-type>

Type of vservers to obtain, valid values are (all|dummy|regular), default: "all"

=item Outputs

Returns an array of vserver names.

=item Examples

my @all_vserver_list = $self->get_vserver_names();
my @cluster_scoped_vserver_list = $self->get_vserver_names(-cluster_name => "cluster1");
my @cluster_scoped_regular_vserver_list = $self->get_vserver_names(-cluster_name => "cluster1",
                                                                   -type         => "regular");
my @cluster_scoped_dummy_vserver_list   = $self->get_vserver_names(-cluster_name => "cluster1",
                                                                   -type         => "dummy");

=back

=head2 B<debug_level($self, %args)>

This function is used to set and get the debug level on SanMgr (or sub objects)

=over 4

=item Inputs

=item * -self

Self is an SanMgr object

=item * I<-debug_level>

Debug level is used to set the internal debug level to what is provided.

=item Outputs

Returns back the set debug level.

=item Examples

Get current debug level
my $debug_level = $self->debug_level();

OR

Set debug level

$self->debug_level(-debug_level => 10);

=back

=head2 B<get_clients_associated_to_vserver(%args)>

Obtain the clients associated to a given vserver/cluster

=over 4

=item Inputs

=item * I<-cluster_name>

Required: The name of the cluster to look up

=item * I<-vserver_names>

Optional: An array reference of vserver names to look up. If not specified or empty,
all vservers will be looked up.
Default: []

=item * I<-protocols>

Optional: An array reference of protocol types to search for.
Default: ["fcp","iscsi"]

=item Outputs

Returns back an array of client names that matched the filters provided,
sets error if an error occurred while looking up the information.

=item Examples

 # obtain all fcp clients associated to cluster "mycluster"
 @clients = $self->get_clients_associated_to_vserver(-cluster_name => "mycluster",
                                                     -protocols => ["fcp"]);
 if ($self->error()) {
    ...
 }

 # obtain all fcp & iscsi clients associated to vservers
 # "vs0" and "vs0" of the cluster "mycluster"
 @clients = $self->get_clients_associated_to_vserver(-cluster_name => "mycluster",
                                                     -vserver_names => ["vs0", "vs1"]);
 if ($self->error()) {
    ...
 }

=back

=head2 B<get_vservers_associated_to_client(%args)>

Obtain the vservers associated to the specified client

=over 4

=item Inputs

=item * I<-client_name>

Required: The name of the client to look up

=item Outputs

Returns back an array of vserver names that are associated to -client_name.
Sets error if an error occurred while looking up the information.

=item Examples

 # obtain all vservers associated to "ibm3550-rtp60"
 @vservers = $self->get_vservers_associated_to_client(-client_name => "ibm3550-rtp60");
 if ($self->error()) {
    ...
 }

=back

=head2 B<get_client_dev_list(%args)>

Get client device list from internal object.
Also allows caller to filter on attributes of client device list.

=over 4

=item Inputs

=item * I<-client>

Optional: A regular expression to filter by clients.

=item * I<-vserver>

Optional:  A regular expression to filter by vservers.

=item * I<-volume>

Optional: A regular expression to filter by volumes.

=item * I<-lun>

Optional: A regular expression to filter by lun.

=item * I<-protocol>

Optional: A regular expression to filter by protocol.

=item * I<-serial>

Optional: A regular expression to filter by LUN serial number.

=item Outputs

Returns back a hash of clients and their associated luns.
If any filters are specified, this hash will reflect it.

=item Examples

  # Obtain all client device lists
  $cdl = $SanMgr_Obj->get_client_dev_list();
  if ($SanMgr_Obj->error()) {
    ...
  }

  # Filtering by client
  $cdl = $SanMgr_Obj->get_client_dev_list(-client => qr/client_01/);
  if ($SanMgr_Obj->error()) {
    ...
  }

  # Filtering by client, vserver, volume, lun and protocol
  $cdl = $SanMgr_Obj->get_client_dev_list(-client => qr/auto-ibm3550-08/,
                                          -vserver => qr/sanmgr_vserv/,
                                          -volume => qr/vol_1/,
                                          -lun => qr/i_3_1/,
                                          -protocol=> qr/iscsi/);
  if ($SanMgr_Obj->error()) {
    ...
  }

=back

=cut

1;
