# -*- perl -*- 
#$Id: //depot/prod/test/main/lib/Err.pm#1 $ 

## @summary Provides some simple Error methods
##
## @owner GX Automation Framework
##
## @author Michael Edmonds (michael.edmonds@netapp.com)
## @author GX Automation Framework (dl-qxautomation@netapp.com)
##
## @change 2007-03-22   medmonds:   Initially created
## @change 2007-04-11   medmonds:   Fixing Up Documentation
## @change 2007-04-11   medmonds:   Fixing Up Documentation ... again
##
#
# Copyright (c) 2007 Network Appliance, Inc., All Rights Reserved
# Permission is granted to use this script on a Network Appliance Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from Network Appliance, Inc.
#
##
## @status public
## @pod here

package Err;
use strict;
use Params::Validate qw(:all);

use TracePkg;

# I know, use contsant would look better, but this is easier to see from an IDE
sub DIE    { return "die";    }
sub RETURN { return "return"; }
sub SCRIPT { return -8;       }
sub CONFIG { return -7;       }
sub NA     { return -6;       }
sub FATAL  { return -5;       }
#use constant UNDEF    => -4;
#use constant DISABLE  => -1;
#use constant PASS     =>  1;
#use constant TRUE     =>  1;
#use constant POSITIVE =>  1;
#use constant FAIL     =>  0;
#use constant FALSE    =>  0;
#use constant NEGATIVE =>  0;

our $Trace;

BEGIN
{
    $Trace = TracePkg->new();
}
# end BEGIN

sub _configure_error
{
    my $self = shift;
    my $hash = {ok  => {optional => 1, type => SCALAR, default => 1},
                type => {optional => 1, type => SCALAR,  default => DIE(),
                         callbacks =>
                         {test=>sub{($_[0] eq DIE()) || ($_[0] eq RETURN())}}}};
    my %params = validate(@_, $hash);

    $self->{error_type}   = $params{type};
    $self->{no_error} = $params{ok};
    return 1;
}
# end _configure_error()


sub _result
{
    my $self = shift;
    my $hash = {rc  => {optional => 1, type => SCALAR, default => $self->{no_error}},
                msg => {optional => 1, type => SCALAR,  default => "NO ERROR"}};
    my %params = validate(@_, $hash);

    # If there is something that was passed in (besides self), fill the result
    if((defined($params{rc})) || (defined($params{msg})))
    {
        $self->{result}->{rc} = $params{rc};
        $self->{result}->{msg} = $params{msg};
        if($self->{result}->{rc} != $self->{no_error})
        {
            $Trace->warn(msg=>"MSG = " . $self->{result}->{msg} . "\n" .
                            "RC  = " . $self->{result}->{rc});

            if(($self->{error_type} eq DIE()) && ($self->{result}->{rc} != $self->NA()))
            {
                die "MSG = " . $self->{result}->{msg} . "\n" .
                    "RC  = " . $self->{result}->{rc};
            }
        }
    }
    else
    {
        # set the existing Error to NO ERROR by default
        $self->{result}->{rc} = $self->{no_error};
        $self->{result}->{msg} = "NO ERROR";
    }
    return $self->{result};
}
# end _result()

sub error_code
{
    my $self = shift;
    return $self->{result}->{rc};
}
# end error_code()

sub ok
{
    my $self = shift;
    return $self->{no_error};
}
# end ok()

sub error_msg
{
    my $self = shift;
    return $self->{result}->{msg};
}
# end error_msg()

sub result
{
    my $self = shift;
    return $self->{result};
}
# end result()

sub clear_result
{
    my $self = shift;
    $self->{result}->{rc} = $self->{no_error};
    $self->{result}->{msg} = "NO ERROR";
}
# end clear_result()

sub is_error
{
    my $self = shift;
    return int(($self->{result}->{rc} != $self->{no_error}));
}
# end is_error()

#sub __destroy_err
#{
#    $Trace->entry();
#    my $self = shift;
#    delete($self->{result});
#    delete($self->{no_error});
#    delete($self->{error_type});
#    $Trace->exit();
#}
# end __destroy_err()

1;

#############################################################################
#                                   POD
#############################################################################

=pod

=head1 Package Name

Err.pm

=head1 Author

Michael Edmonds

=head1 Owner

GX Automation Framework (dl-gxautomation@netapp.com)

=head1 Description

Err is intended to provide some error methods for Classes.

If you are inheriting from Class.pm, you are already inheriting Err.pm

=head1 Synopsis

Here are some examples. 

=over 4

    use Err;
    our @ISA = qw(Err);
    
    # configure the err module to DIE is it is passed anything other than 1 in the rc;
    $self->_configure_error(ok => 1, type => $self->DIE());
    
    # set the result hash to rc => $self->ok() and msg => "NO ERROR"
    $self->_result();

=head1 Freind Methods

=head2 C<_configure_error(%params)>

=over 4

=item Description

configures the Module for a specified return code that is not an error
and
tells the module to either DIE if it gets an error code or let the calling module deal with the result

=item Inputs

=item * ok => <int>
optional
default = 1

in some cases, 0 means no error (when getting the rc back from a system call for instance)
this allows you to set whatever the return code should be on "no error"

=item * type => <Err::DIE() | Err::RETURN()>
optional
default = DIE()

This allows you to either DIE on errors
or
take full responsibility for the Error checking and just return

=item Outputs

1

=back


=head2 C<_result(%params)>

=over 4

=item Description

This is what the parent Lib should call before it exits a function or encounters
an error.
This method sets the result hash AND Dies if ((rc != ok) && (type == DIE()))

=item Inputs

=item * rc => <int>
(Optional)
default = $self->{no_error}

=item * msg => <string>
(Optional)
default = "NO ERROR"

=item Outputs

result HashRef
$result = {rc => <rc>, msg => <msg>};

this is so you "could" do a:
return $self->_result(rc => -1, msg => "error doing stuff");

=item Examples
    
    sub blah
    {
        my ($self, $something) = @_;

        if($something ne "foo")
        {
            $self->_result(rc => -1, msg => "error doing stuff");
            return;
        }

        $self->_result();
    }

=back


=head1 Public Methods

=head2 C<is_error()>

this is what you call to see if there was an rc that was not deamed as "ok"

=over 4 

=item Inputs

None

=item Outputs

Boolean (1 or 0)
if there was an rc that did not match the "ok" rc, then it returns 1.

=item Examples

Inside your own lib:

    $self->_do_stuff();
    if($self->is_error())
    {
        return;
    }
    
--------------------------
using a lib that uses Err.pm:

    $some_obj->do_stuff();
    if($some_obj->is_error())
    {
        $Trace->fatal(msg=>"Error = " . Dumper($some_obj->result()));
        # OR
        if($some_obj->error_msg() ne "blah")
        {
            $Trace->fatal(msg=>"ERROR\nRC = " . $some_obj->error_code() .
                                "\nMSG = " . $some_obj->error_msg());
        }
    }

=back

=head2 C<clear_result()>

this is what you call if the rc is NOT ok, but it's not bad.
It just set the result back to a no error condition

=over 4 

=item Inputs

None

=item Outputs

None

=item Examples

Inside your own lib:

    $self->_do_stuff();
    if($self->is_error())
    {
        if($self->error_code() != $self->NA())
        {
            $Trace->fatal(msg=>"ERROR = " . Dumper($self->result()));
            return;
        }
        $self->clear_result();  # Since this is YOURself, you might as well
                                #   just do $self->_result();
    }
    
--------------------------
using a lib that uses Err.pm:

    $some_obj->do_stuff();
    if($some_obj->is_error())
    {
        if($some_obj->error_code() != $some_obj->NA())
        {
            $Trace->fatal(msg=>"ERROR = " . Dumper($some_obj->result()));
            return;
        }
        $some_obj->clear_result();
    }

=back

=head2 C<error_msg()>

Allows one to check what the message string was

=item Inputs

None

=item Outputs

String
it returns $self->{result}->{msg}

=item Examples

using a lib that uses Err.pm:

    $some_obj->do_stuff();
    if($some_obj->is_error())
    {
        $Trace->fatal(msg=>"Error = " . $some_obj->error_msg());
    }


=back

=head2 C<error_code()>

Allows one to check what the return code was

=item Inputs

None

=item Outputs

Integer
it returns $self->{result}->{rc}

=item Examples

using a lib that uses Err.pm:

    $some_obj->do_stuff();
    if($some_obj->error_code() != 254)
    {
        $Trace->fatal(msg => "RC = "  . $some_obj->error_code() .
                            "\nMSG =" . $some_obj->error_msg());
    }


=back

=head2 C<ok()>

Allows one to check what the "ok" return code is

Not really used (because of is_error() can do this for you), but it's available

=item Inputs

None

=item Outputs

Integer
it returns $self->{no_error}

=item Examples

Inside your own lib:

    $self->_do_stuff();
    if($self->error_code() != $self->ok())
    {
        return;
    }
    
--------------------------
using a lib that uses Err.pm:

    $some_obj->do_stuff();
    if($some_obj->error_code() != $some_obj->ok())
    {
        $Trace->fatal(msg=>"Error = " . Dumper($some_obj->result()));
        # OR
        if($some_obj->error_msg() ne "blah")
        {
            $Trace->fatal(msg=>"ERROR\nRC = " . $some_obj->error_code() .
                                "\nMSG = " . $some_obj->error_msg());
        }
    }


=back

=head2 C<result()>

allows you to get the result hash in one call instead of calling error_code() and error_msg()

=over 4

=item Inputs

None

=item Outputs

result HashRef
$result = {rc => <rc>, msg => <msg>};

=item Examples

using a lib that uses Err.pm:

    $some_obj->do_stuff();
    if($some_obj->is_error())
    {
        my $result = $some_obj->result();
        $Trace->fatal(msg=>"ERROR\nRC = " . $result->{rc} .
                            "\nMSG = " . $result->{msg});
    }

=back

=cut