#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/Error.pm#1 $

# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provides basic error handling methods
## that should be inherited by other object oriented
## packages
##
## @author knopp@netapp.com
## @status Shared
## @pod here

package Error;

use lib();

# Compiler directives.
use strict;

###############################################################################
# Method:    errmsg
# Objective: Return the errmsg component from the object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub errmsg {
    my ($self) = @_;

    unless (ref($self->{errmsg}) eq "ARRAY") {
        return ("NO ERROR MESSAGE");
    }
    return (wantarray()) ? @{$self->{errmsg}} : join("\n", @{$self->{errmsg}});
}    # end sub errmsg

###############################################################################
# Method:    error
# Objective: Return error status of the err object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub error {
    my ($self) = @_;

    return $self->{error};

}    # end sub error

###############################################################################
# Method:    _clear_err
# Objective: Clear error components in the object.
# Inputs:    None.
# Output:    None.
# Affected Data: Error components in the object have been cleared.
###############################################################################
sub _clear_err {
    my ($self) = @_;

    $self->{error}  = 0;
    $self->{errmsg} = '';

    return 1;
}    # end sub _clear_err

###############################################################################
# Method:    _set_err
# Objective: Set error components in the object based on input errmsg.
# Inputs:    errmsg.
# Output:    None.
# Affected Data: Error components in the object have been set.
###############################################################################
sub _set_err {
    my ($self, @tmp_errmsg) = @_;

    my (@errmsg, $parent_obj, $array_index);

    $self->{error} = 1;
    $array_index = 0;

    # if there is no setting for disable caller string, then set it to 0
    if (not defined($self->{error_disable_caller_string})) {
        $self->{error_disable_caller_string} = 0;
    }

    if (not defined($self->{error_stack_style})) {
        $self->{error_stack_style} = "normal";
    }

    foreach my $msg (@tmp_errmsg) {

        # only prepend with the first message in the array
        # indent the remaining messages in the array
        if ($array_index == 0 and $self->{error_disable_caller_string} == 0) {

            # prepend parent caller and line number of the error
            # - ERROR: San::ControllerCluster::Filer10M::Aggregate:: \
            #   create_vol_obj:104 - failed to instantiate volume object
            push @errmsg, "ERROR: " . _parent_caller() . "():" . _parent_lineno() . " - " . $msg;
        } else {

            # indent and append messsage.
            push @errmsg, " " . $msg;
        }
        $array_index++;
    }

    # Check to see if we should include the call stack
    if ((defined($self->{call_stack_on_error})) && ($self->{call_stack_on_error})) {
        my $depth = 1;
        push @errmsg, "-------------------------------------------------------------------";
        push @errmsg, "\nCall Stack in _set_err():";
        push @errmsg, "-------------------------------------------------------------------";
        while ((my (undef, undef, $line, $subroutine) = caller($depth)) && ($depth < 6)) {
            push @errmsg, $depth . " " . $subroutine . ":$line";
            $depth++;
        }
        push @errmsg, "-------------------------------------------------------------------";

        # get top level parent object
        my $current_obj = $self;
        $parent_obj = &_get_top_parent_obj($current_obj);

        # use dumper to dump data and store data structure in errmsg array
        # for display
        my $dump_data = Data::Dumper->Dump([$parent_obj], [ref($parent_obj)]);

        push @errmsg, $dump_data;
    }

    # if there is currently an error on the object, then dont overwrite it,
    # append to it so that an object can keep invoking set err to build an error stack
    if (defined($self->{errmsg}) and
        ref($self->{errmsg}) eq "ARRAY" and
        scalar(@{$self->{errmsg}}) and
        defined($self->{error_dont_overwrite_errors}) and
        $self->{error_dont_overwrite_errors} == 1) {

        # take the current errmsg array, and merge it with the error that is being pushed on
        my @err_array = @{$self->{errmsg}};
        if ($self->{error_stack_style} eq "reverse") {
            push(@err_array, @errmsg);
        } else {
            unshift(@err_array, @errmsg);
        }
        @errmsg = @err_array;
    }

    $self->{errmsg} = \@errmsg;

    return 1;
}    # end sub _set_err

###############################################################################
# Method:    _get_top_parent_obj
# Objective: get the top level parent object
# Inputs:    Caller object
# Return:    Top level object
# Affected Data: Error components in the object have been cleared.
###############################################################################
sub _get_top_parent_obj {
    my ($current_obj) = @_;

    # let's assume current_obj is the top parent object
    my $top_parent = $current_obj;

    # looking for parent key in the object, traverse back until at the top
    # then dump data object
    foreach my $key (keys %{$current_obj}) {

        # if key is parent, set top_parent to parent and make a recursive
        # call to itself until at the top level
        if ($key =~ /-?parent$/) {
            my $parent = $current_obj->{$key};
            $top_parent = &_get_top_parent_obj($parent);
        }
    }

    return $top_parent;
}

# format for the caller:
#  0         1          2      3            4
# ($package, $filename, $line, $subroutine, $hasargs,
#  5          6          7            8       9         10
# $wantarray, $evaltext, $is_require, $hints, $bitmask, $hinthash) = caller($i);
# $i is the caller depth
# _parent_caller returns the subroutine of caller of depth 2
sub _parent_caller { (caller(2))[3] }

# _parent_lineno returns the line number of caller of depth 1.
# this will return the line number within the above subroutine
sub _parent_lineno { (caller(1))[2] }

1;

=pod

=head1 Package Name

Error.pm

=head1 Author

Gary Knopp

=head1 Description

Error provides an API interface for performing basic error handling.  Use @ISA to 
inherit these methods into another package.  Use the private methods 
_clear_err() and _set_err($error_msg) in your package to remove or set error messages.

=over 4

=item Examples

use vars qw( @EXPORT @ISA);

@ISA = qw(Error);

=back

=head1 Public Methods

=head2 B<errmsg()>

This method allows you to get the text of the error message if an error
occurred on the previous method call.

=over 4

=item Inputs

None

=item Outputs

Returns the text of the last error message set.  If no error message is
set, an empty string will be returned.

=item Examples

 # Check for errors.
 if ($myObj->error) {
    $errmsg = $myObj->errmsg;
    print "An error occurred: $errmsg\n";
 }

=back

=head2 B<error()>

This method allows you to determine if any errors occured in the
previous method call.  Note that the error flag is reset at the
beginning of each method, so it should be checked after B<every>
major method call.  If you do not check for errors, your calling
program may fail at some point and you will not know why.

=over 4 

=item Inputs

None.

=item Outputs

Returns 1 if an error occured in the previous call, 0 otherwise.

=item Examples

 # Determine if error occurred...
 if ($myObj->error) {
    print "Error encountered in previous method call\n";
 }

=back

=cut
