# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary An Event Scheduler object
## @author hrivera@netapp.com
## @status public
## @pod here

package EventMan;

# Compiler directives
use strict;
use warnings;

# Globals
use vars qw(@ISA @UTILITY_MODULES @EVENT_MODULES $Log);

# NATE and debug packages
use Error;
use Tharn;
use NATE::STAF::Peer qw($Parent);
use NATE::Events;
use Data::Dumper;
use Data::Dump;
use ParamsMgr;
use Devel::Size qw(size total_size);
use NATE::Log qw(log_global);

# External San packages
use San::SanUtil;

# Internal core packages
use EventMan::Event;
use EventMan::EventSet;
use EventMan::EventSetGroup;

# Internal utility modules
use EventMan::UtilityModules::EventUtil;
use EventMan::UtilityModules::SubtestComm;
use EventMan::UtilityModules::ResMgr;
use EventMan::UtilityModules::PathValidation;
use EventMan::UtilityModules::PortValidation;
use EventMan::UtilityModules::ControllerValidation;
use EventMan::UtilityModules::SysLogging;
use EventMan::UtilityModules::SysStats;
use EventMan::UtilityModules::SRAMDetection;
use EventMan::UtilityModules::ClamValidationUtil;

# Utility module registration list
@UTILITY_MODULES = qw(
                      EventUtil
                      SubtestComm
                      ResMgr
                      PathValidation
                      PortValidation
                      ControllerValidation
                      SysLogging
                      SysStats
                      SRAMDetection
                      ClamValidationUtil
                  );

# Imported event modules
use EventMan::EventModules;
use EventMan::EventModules::LibraryUnitTest;
use EventMan::EventModules::SwitchEvents;
use EventMan::EventModules::FabricEvents;
use EventMan::EventModules::MiscEvents;
use EventMan::EventModules::ControllerEvents;
use EventMan::EventModules::ControllerEvents::CMode;
use EventMan::EventModules::ControllerEvents::CMode::SixtySecondSLA;;
use EventMan::EventModules::ControllerEvents::CMode::ClusterICEvents;
use EventMan::EventModules::ControllerEvents::CMode::SFOEvents;
use EventMan::EventModules::ControllerEvents::CMode::VolumeEvents;
use EventMan::EventModules::ControllerEvents::CMode::LifEvents;
use EventMan::EventModules::ControllerEvents::CMode::DisruptiveNodeEvents;
use EventMan::EventModules::ControllerEvents::CMode::NodeUpgradeEvents;
use EventMan::EventModules::ControllerEvents::CMode::AggrEvents;
use EventMan::EventModules::ClientEvents;
use EventMan::EventModules::ClientEvents::ESXHostEvents;


# Event module registration list
@EVENT_MODULES = qw (
                    EventMan::EventModules
                    EventMan::EventModules::LibraryUnitTest
                    EventMan::EventModules::SwitchEvents
                    EventMan::EventModules::FabricEvents
                    EventMan::EventModules::MiscEvents
                    EventMan::EventModules::ControllerEvents
                    EventMan::EventModules::ControllerEvents::CMode
                    EventMan::EventModules::ControllerEvents::CMode::SixtySecondSLA
                    EventMan::EventModules::ControllerEvents::CMode::ClusterICEvents
                    EventMan::EventModules::ControllerEvents::CMode::SFOEvents
                    EventMan::EventModules::ControllerEvents::CMode::VolumeEvents
                    EventMan::EventModules::ControllerEvents::CMode::LifEvents
                    EventMan::EventModules::ControllerEvents::CMode::DisruptiveNodeEvents
                    EventMan::EventModules::ControllerEvents::CMode::NodeUpgradeEvents
                    EventMan::EventModules::ClientEvents
                    EventMan::EventModules::ClientEvents::ESXHostEvents
                    EventMan::EventModules::ControllerEvents::CMode::AggrEvents
                );

# Internal inherited packages
@ISA = qw(Error);

# Log object
$Log = log_global();

# Possible EventMan states
$EventMan::STATE_NOTSTARTED  = "not_started";
$EventMan::STATE_RUNNING     = "running";
$EventMan::STATE_STOPPED     = "stopped";
$EventMan::STATE_ABORTED     = "aborted";
$EventMan::STATE_TIMEDOUT    = "timedout";

# Timeout States
$EventMan::NO_TIME_LIMIT      = -1;
$EventMan::TIME_LIMIT_REACHED = -2;

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new()
# Objective: EventMan object Constructor
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my $self = {};

    bless($self, $class);

    # need to tell Error class to handle errors in a special way
    # this will allow us to invoke set err over and over to create an error stack
    # without overwriting the previous error
    $self->{error_dont_overwrite_errors} = 1;

    # Make the error stack come out in reverse order
    $self->{error_stack_style} = "reverse";

    # register all event modules
    $self->_register_all_event_modules(\%args);
    if ($self->error()) {
        $self->_set_err("Failed to register all event modules");
        return $self;
    }

    # register all utility modules
    $self->_register_all_utility_modules(\%args);
    if ($self->error()) {
        $self->_set_err("Failed to register all utility modules");
        return $self;
    }

    # store the arguments to the constructor in the args attribute
    $self->{args} = {&SanUtil::strip_dash(%args)};

    # set the current state to not started
    $self->_set_current_state($EventMan::STATE_NOTSTARTED);
    $self->{_initializing_events} = $EventUtil::FALSE;
    $self->{_initialized_events} = $EventUtil::FALSE;
    $self->{_was_set_to_initialize} = $EventUtil::FALSE;

    # initialize the unique event id counters for Event objects and
    # EventSet objects
    $self->{_event_id_counter} = 0;
    $self->{_event_set_id_counter} = 0;

    # create the default event set with all the params passed into the
    # constructor
    $self->create_event_set(-event_set_name => "default_event_set",
                            %args);
    if ($self->error()) {
        $self->_set_err("Failed to create default event set");
        return $self;
    }

    # initialize the default event set sequence
    $self->set_default_event_set_sequence(-event_set_sequence => ["default_event_set"]);
    if ($self->error()) {
        $self->_set_err("Failed to set default event set sequence");
        return $self;
    }

    return $self;
}

###############################################################################
# Method:    create_event_set()
# Objective: Creates an EventSet object and registers it with the EventMan
#            instance.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_event_set {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_set_name = &ParamsMgr::get_arg(-arg => "event_set_name",
                                             -required => 1);
    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # make sure this event set doesnt already exist
    if (exists($self->{event_sets}->{$event_set_name})) {
        $self->_set_err("Invalid event set: $event_set_name already exists");
        return {};
    }

    # create the event set object and store it in the event set list
    my $event_set_obj = new EventMan::EventSet(%args,
                                               -unique_id => $self->{_event_set_id_counter});
    if ($event_set_obj->error()) {
        $self->_set_err($event_set_obj->errmsg());
        return {};
    }

    # store the event set object in the event sets hash
    $self->{event_sets}->{$event_set_name} = $event_set_obj;

    # incrememnt the unique event set id counter
    $self->{_event_set_id_counter}++;

    return $event_set_obj;
}

###############################################################################
# Method:    get_event_set()
# Objective: Retrieves the EventSet object by name.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_event_set {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_set_name = &ParamsMgr::get_arg(-arg => "event_set_name",
                                             -required => 1);
    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    if (exists($self->{event_sets}->{$event_set_name})) {
        return $self->{event_sets}->{$event_set_name};
    } else {
        $self->_set_err("Event set does not exist: '$event_set_name'");
        return {};
    }
}

###############################################################################
# Method:    delete_event_set()
# Objective: Deletes the EventSet object by name.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete_event_set {
    my ($self, %args) = @_;

    my $event_set_obj = $self->get_event_set(%args);
    if ($self->error()) {
        $self->_set_err("Unable to get event set");
        return $EventUtil::FALSE;
    }

    # find each event set group this event set is a member of,
    # and remove this event set from each group
    my @event_set_group_list = $event_set_obj->get_group_list();

    foreach my $event_set_group_name (@event_set_group_list) {
        # get the group object
        my $event_set_group_obj = $self->get_event_set_group(-group_name => $event_set_group_name);
        if ($self->error()) {
            $self->_set_err("Unable to get event set group");
            return $EventUtil::FALSE;
        }

        # remove this event set from this event set group
        $event_set_group_obj->remove_event_set_from_group(-event_set_obj => $event_set_obj);
        if ($event_set_group_obj->error()) {
            $self->_set_err($event_set_group_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # remove this event set from the event set lookup table
    delete($self->{event_sets}->{$event_set_obj->get_name()});

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    add_event()
# Objective: Creates an Event object and adds it to the specified EventSet's
#            event list.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add_event {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_name = &ParamsMgr::get_arg(-arg => "event_name",
                                         -required => 1);
    delete($args{-event_name});

    # if an Eventset isnt defined, assume that the default event set will be used
    my $event_set_name = &ParamsMgr::get_arg(-arg => "event_set_name",
                                             -default => "default_event_set");
    delete($args{-event_set_name});

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # validate that this event is registered with EventMan
    if (not $self->_is_event_registered(-event_name => $event_name)) {
        $self->_set_err("The event: '$event_name' is invalid. It is not registered with any Event Module");
        return {};
    }

    # create this event object
    my $event_obj = new EventMan::Event(-event_name => $event_name,
                                        -event_set_name => $event_set_name,
                                        -unique_id => $self->{_event_id_counter},
                                        -description => $self->{_valid_events}->{$event_name}->{description},
                                        -event_modules => [@{$self->{_valid_events}->{$event_name}->{event_modules}}],
                                        %args);
    if ($event_obj->error()) {
        $self->_set_err($event_obj->errmsg());
        return {};
    }

    # retrieve the event set object by name
    my $event_set_obj = $self->get_event_set(-event_set_name => $event_set_name);
    if ($self->error()) {
        $self->_set_err("Could not obtain EventSet: '$event_set_name'");
        return {};
    }

    # update the current environment with these event set and event objects
    $self->_update_current_event(-event_obj => $event_obj,
                                 -event_set_obj => $event_set_obj);
    if ($self->error()) {
        # reset the current event objects
        $self->_reset_current_event();
        $self->_set_err("Failed to update current event");
        return {};
    }

    # run the validate method for this event keyword (if one exists)
    $self->_validate_event(-event_obj => $event_obj,
                           -event_set_obj => $event_set_obj,
                           %args);
    if ($self->error()) {
        # reset the current event objects
        $self->_reset_current_event();
        $self->_set_err("Event: '$event_name' failed validation");
        return {};
    }

    # reset the current event objects
    $self->_reset_current_event();

    # add this event to this event set
    $event_set_obj->add_event_to_set(-event_obj => $event_obj);
    if ($event_set_obj->error()) {
        $self->_set_err($event_set_obj->errmsg());
        return {};
    }

    # incrememnt the unique event id counter
    $self->{_event_id_counter}++;

    return $event_obj;
}

###############################################################################
# Method:    delete_event()
# Objective: Deletes the specified Event object and removes it from an EventSet
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete_event {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_obj = &ParamsMgr::get_arg(-arg => "event_obj",
                                        -required => 1,
                                        -ref_type => "EventMan::Event");

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # determine which event set this event object belongs to
    my $event_set_name = $event_obj->get_event_set_name();

    # retrieve the event set object by name
    my $event_set_obj = $self->get_event_set(-event_set_name => $event_set_name);
    if ($self->error()) {
        $self->_set_err("Could not obtain EventSet: '$event_set_name'");
        return $EventUtil::FALSE;
    }

    # remove the event from this eventset
    $event_set_obj->remove_event_from_set(-event_obj => $event_obj);
    if ($event_set_obj->error()) {
        $self->_set_err($event_set_obj->errmsg());
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    create_event_set_group()
# Objective: Creates an EventSetGroup object by name and adds the specified
#            EventSet objects to the group.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_event_set_group {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # handle two styles of input, objects directly, or just the names
    my @event_set_names = &ParamsMgr::get_arg(-arg => "event_set_names",
                                              -default => [],
                                              -ref_type => "ARRAY");
    my @event_set_objs = &ParamsMgr::get_arg(-arg => "event_set_objs",
                                             -default => [],
                                             -ref_type => "ARRAY");
    delete($args{-event_set_names});
    delete($args{-event_set_objs});

    my $group_name = &ParamsMgr::get_arg(-arg => "group_name",
                                         -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # if names were provided, fetch the objects, and push onto the obj array
    foreach my $event_set_name (@event_set_names) {
        my $event_set_obj = $self->get_event_set(-event_set_name => $event_set_name);
        if ($self->error()) {
            $self->_set_err("Could not get EventSet: '$event_set_name'");
            return {};
        }
        push(@event_set_objs, $event_set_obj);
    }

    # create the group of the eventset objects
    my $event_set_group_obj = new EventMan::EventSetGroup(%args);
    if ($event_set_group_obj->error()) {
        $self->_set_err($event_set_group_obj->errmsg());
        return {};
    }

    # foreach event set object, add it to the group
    foreach my $event_set_obj (@event_set_objs) {
        $event_set_group_obj->add_event_set_to_group(-event_set_obj => $event_set_obj);
        if ($event_set_group_obj->error()) {
            $self->_set_err($event_set_group_obj->errmsg());
            return {};
        }
    }

    # store the event set group in our event set group lookup table
    $self->{event_set_groups}->{$group_name} = $event_set_group_obj;

    return $event_set_group_obj;
}

###############################################################################
# Method:    get_event_set_group()
# Objective: Obtains the EventSetGroup object by name.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_event_set_group {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $group_name = &ParamsMgr::get_arg(-arg => "group_name",
                                         -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    if (exists($self->{event_set_groups}->{$group_name})) {
        return $self->{event_set_groups}->{$group_name};
    } else {
        $self->_set_err("EventSetGroup does not exist: '$group_name'");
        return {};
    }
}

###############################################################################
# Method:    delete_event_set_group()
# Objective: Deletes the EventSetGroup object by name.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete_event_set_group {
    my ($self, %args) = @_;

    my $event_set_group_obj = $self->get_event_set_group(%args);
    if ($self->error()) {
        $self->_set_err("Unable to get event set group");
        return $EventUtil::FALSE;
    }

    # issue a destroy on this EventSetGroup
    # this will safely remove all dependencies each eventset may have on this EventSetGroup
    $event_set_group_obj->destroy();
    if ($event_set_group_obj->error()) {
        $self->_set_err($event_set_group_obj->errmsg());
        return $EventUtil::FALSE;
    }

    # delete the EventSetGroup from the EventMan EventSetGroup look up table
    delete($self->{event_set_groups}->{$event_set_group_obj->get_group_name()});

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    is_running()
# Objective: Determines if the EventMan subtest is running or not.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_running {
    my ($self) = @_;

    # issue an asynchronous update to the current state
    $self->_update_current_state();
    if ($self->error()) {
        $self->_set_err("Unable to update current state.");
        return $EventUtil::FALSE;
    }

    # determine if the events subtest is currently running
    if ($self->_get_current_state() eq $EventMan::STATE_RUNNING) {
        return $EventUtil::TRUE;
    } else {
        return $EventUtil::FALSE;
    }
}

###############################################################################
# Method:    stop_events()
# Objective: Tells the EventMan subtest to stop all event activities
#            gracefully
# Details:   See POD documentation at the end of this file.
###############################################################################
sub stop_events {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $async = &ParamsMgr::get_arg(-arg => "async",
                                    -default => $EventUtil::FALSE);
    my $graceful = &ParamsMgr::get_arg(-arg => "graceful",
                                       -default => $EventUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;;
    }

    # validate that the events subtest is defined and currently running
    # and that we are not in a terminal state
    if (not defined($self->{_events_subtest})) {
        $self->_set_err("Event subtest does not exist");
        return $EventUtil::FALSE;
    }
    if (not $self->{_events_subtest}->isAlive() or
            $self->_get_current_state() eq $EventMan::STATE_NOTSTARTED) {
        $self->_set_err("Event subtest is not running");
        return $EventUtil::FALSE;
    }
    if ($self->_get_current_state() eq $EventMan::STATE_STOPPED or
        $self->_get_current_state() eq $EventMan::STATE_ABORTED) {
        $self->_set_err("Event subtest is already stopped/aborted");
        return $EventUtil::FALSE;
    }

    # determine if we should gracefully or nongracefully stop the events subtest
    if ($graceful == 0) {
        # issue a non graceful kill of the events subtest
        &SanUtil::nate_log("-comment", "Requested to stop events non-gracefully");
        $self->{_events_subtest}->kill(message => ["-comment", "Requested to stop events non-gracefully"]);
        &SanUtil::nate_wait_finish($self->{_events_subtest});
    } else {
        # order the events subtest to stop, this is a graceful stop, meaning that
        # it will not allow additional events or eventsets to begin, but all
        # "in flight" events will be allowed to finish
        $self->SubtestComm::subtest_stop(-subtest => $self->{_events_subtest});

        # wait for the events subtest to complete if async == $EventUtil::FALSE
        if ($async == $EventUtil::FALSE) {
            $self->wait_for_subtests(-dependence_type => "independent",
                                     -propagate_wait => 0,
                                     -propagate_events => 0,
                                     -check_cmds_from_parent => 0,
                                     -forward_cmds_from_parent => 0,
                                     -subtests => [$self->{_events_subtest}]);

            # change the state to stopped, only in sync mode
            # can't predict when the state will change in async mode
            # a call to "is_running()" will dynamically change the state
            $self->_set_current_state($EventMan::STATE_STOPPED);

            if ($self->error()) {
                $self->_set_err("Could not stop events, the events subtest has failed");
                return $EventUtil::FALSE;
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    abort_events()
# Objective: Tells the EventMan subtest to abort all event activities
#            immediately (or gracefully)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub abort_events {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $async = &ParamsMgr::get_arg(-arg => "async",
                                    -default => $EventUtil::FALSE);
    my $graceful = &ParamsMgr::get_arg(-arg => "graceful",
                                       -default => $EventUtil::FALSE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;;
    }

    # validate that the events subtest is actually running
    # and that we are not in a terminal state
    if (not defined($self->{_events_subtest})) {
        $self->_set_err("Event subtest does not exist");
        return $EventUtil::FALSE;
    }
    if (not $self->{_events_subtest}->isAlive() or
            $self->_get_current_state() eq $EventMan::STATE_NOTSTARTED) {
        $self->_set_err("Event subtest is not running");
        return $EventUtil::FALSE;
    }
    if ($self->_get_current_state() eq $EventMan::STATE_STOPPED or
        $self->_get_current_state() eq $EventMan::STATE_ABORTED) {
        $self->_set_err("Event subtest is already stopped/aborted");
        return $EventUtil::FALSE;
    }

    # determine if we should gracefully or nongracefully abort the events subtest
    if ($graceful == 0) {
        # issue a non graceful kill of the events subtest
        &SanUtil::nate_log("-warn", "Requested to abort events non-gracefully");
        $self->{_events_subtest}->kill(message => ["-warn", "Requested to abort events non-gracefully"]);
        &SanUtil::nate_wait_finish($self->{_events_subtest});
    } else {
        # order the events subtest to do a graceful abort, meaning that
        # it will not allow additional events or eventsets to begin, but all
        # "in flight" events will be allowed to finish, it will then mark the running
        # subtest warn to indicate that it was aborted.
        $self->SubtestComm::subtest_abort(-subtest => $self->{_events_subtest});

        # wait for the events subtest to complete if async == 0
        if ($async == 0) {
            $self->wait_for_subtests(-dependence_type => "independent",
                                     -propagate_wait => 0,
                                     -propagate_events => 0,
                                     -check_cmds_from_parent => 0,
                                     -forward_cmds_from_parent => 0,
                                     -subtests => [$self->{_events_subtest}]);

            # change the state to stopped, only in sync mode
            # can't predict when the state will change in async mode
            # a call to "is_running()" will dynamically change the state
            $self->_set_current_state($EventMan::STATE_ABORTED);

            if ($self->error()) {
                $self->_set_err("Could not abort events, the events subtest has failed");
                return $EventUtil::FALSE;
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    initialize_events()
# Objective: Initializes all the events
# Details:   See POD documentation at the end of this file.
###############################################################################
sub initialize_events {
    my ($self, %args) = @_;

    # by default do not wait for events to initialize
    # prefer wait over -wait, but remove both
    my $wait = delete ($args{-wait}) // $EventUtil::FALSE;
    $wait    = delete ($args{wait})  // $wait;

    # hardcode a couple options so that async is on and that we only intialize
    # events
    $args{async} = $EventUtil::TRUE;
    $args{initialize_only} = $EventUtil::TRUE;

    # set a flag that we are initializing eventman
    $self->{_initializing_events} = $EventUtil::TRUE;
    $self->{_was_set_to_initialize} = $EventUtil::TRUE;

    my $subtest = $self->run_events(%args);
    if ($self->error()) {
        return {};
    }

    # if requested to wait for it to finish initializing, do so here
    if ($wait) {
        while ($self->is_initializing()) {
            # break out if we encounter an error
            if ($self->error()) {
                last;
            }
            sleep 2;
        }
        if ($self->error()) {
            $self->_set_err("Encountered an error while waiting for EventMan to initialize");
            return $subtest;
        }
    }

    return $subtest;
}

###############################################################################
# Method:    is_initializing()
# Objective: Return true/false on whether events are initializing or not
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_initializing {
    my ($self) = @_;

    $self->_clear_err();

    # query the event man subtest for its initializing status
    if ($self->is_running()) {
        $self->{_events_subtest}->message_put(type => 'is_initializing',
                                              message => "");
        my $msg = $self->{_events_subtest}->message_get(type => 'is_initializing_response');
        $self->{_initializing_events} = $msg;
        if (not defined($msg)) {
            $self->_set_err("Received invalid message for query 'is_initializing'");
        }
        return $self->{_initializing_events};
    } else {
        # eventman is not running therefore it can't be initializing
        $self->_set_err("The EventMan object is not running");
        return $EventUtil::FALSE;
    }
}

###############################################################################
# Method:    has_initialized()
# Objective: Return true/false on whether events have finished initialization
# Details:   See POD documentation at the end of this file.
###############################################################################
sub has_initialized {
    my ($self) = @_;

    $self->_clear_err();

    # query the event man subtest for its initializing status
    if ($self->is_running()) {
        $self->{_events_subtest}->message_put(type => 'has_initialized');
        my $msg = $self->{_events_subtest}->message_get(type => 'has_initialized_response');
        $self->{_initialized_events} = $msg;
        if (not defined($msg)) {
            $self->_set_err("Received invalid message for query 'has_initialized'");
        }
        return $self->{_initialized_events};
    } else {
        $self->_set_err("The EventMan object is not running");
        return $EventUtil::FALSE;
    }
}

###############################################################################
# Method:    was_set_to_initialize()
# Objective: Return a true/false on whether the initialize_events() method
#            was invoked.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub was_set_to_initialize {
    my ($self) = @_;

    return $self->{_was_set_to_initialize};
}

###############################################################################
# Method:    run_events()
# Objective: Executes the EventMan subtest with the specified
#            EventSet/EventSetGroup Sequence
# Details:   See POD documentation at the end of this file.
###############################################################################
sub run_events {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # use the default event set if no event set sequence was specified
    my @event_set_sequence = &ParamsMgr::get_arg(-arg => "event_set_sequence",
                                                 -default => $self->get_default_event_set_sequence(),
                                                 -ref_type => "ARRAY");
    my $sequence_style = &ParamsMgr::get_arg(-arg => "sequence_style",
                                             -default => $self->{_default_event_set_sequence_params}->{sequence_style});
    my $sequence_runlength = &ParamsMgr::get_arg(-arg => "sequence_runlength",
                                                 -default => $self->{_default_event_set_sequence_params}->{sequence_runlength});
    my $sequence_count = &ParamsMgr::get_arg(-arg => "sequence_count",
                                             -default => $self->{_default_event_set_sequence_params}->{sequence_count});

    # default the run id of the the events subtest to 'EventMan'
    my $runid = &ParamsMgr::get_arg(-arg => "runid",
                                    -default => $self->{_default_event_set_sequence_params}->{runid});

    # Control whether to refresh resources before running
    my $refresh_resources = &ParamsMgr::get_arg(-arg => "refresh_resources",
                                                -default => $EventUtil::FALSE);

    # default mode is synchronous (blocking)
    my $async = &ParamsMgr::get_arg(-arg => "async",
                                    -default => $EventUtil::FALSE);

    # if initialize_only is set, then load state first and wait to be told to start
    my $initialize_only = &ParamsMgr::get_arg(-arg => "initialize_only",
                                              -default => $EventUtil::FALSE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # abort if intialize_only was set but we were not set to be async
    if ($initialize_only and not $async) {
        $self->_set_err("Cannot use the option 'initialize_only' when 'async' is not set to 1");
        return {};
    }

    # if we are already running, issue a start_events command and return the subtest handle
    if ($self->is_running()) {
        $self->send_passthrough_command(-msg => 'start_events');
        if ($self->error()) {
            $self->_set_err("Failed to send 'start_events' command to the EventMan process");
            return {};
        }
        return $self->{_events_subtest};
    }

    # convert all spaces to underscores in the run id
    $runid =~ s/\s+/_/g;

    # validate that some of the params are proper
    $sequence_count = 1 if ($sequence_count <= 0);
    $sequence_runlength = $EventMan::NO_TIME_LIMIT if ($sequence_runlength < 0);

    # store the start time of the events subtest
    $self->{_start_time} = time();

    # reset the events subtest attribute
    $self->{_events_subtest} = undef;

    # if doing iterations, then set runlength to the No time limit state
    # (this will disable any timeout)
    if ($sequence_style =~ /iterations/i) {
        $sequence_runlength = $EventMan::NO_TIME_LIMIT;
    } elsif ($sequence_style =~ /runlength/i) {
        # if doing runlength, then set the sequence count to an arbitrarily large number
        $sequence_count = 99999999;
    } elsif ($sequence_style =~ /mixed/i) {
        # if doing a mixed sequence style, then allow both values (sequence_count | sequence_runlength) to be set,
        # which ever expires first will determine how long the event sequence will run for
    } else {
        $self->_set_err("Invalid -sequence_style '$sequence_style'");
        return {};
    }

    # verify that -event_set_sequence contains valid event set entries
    # so that we don't fail out half way through events because of an invalid event set
    # in the sequence
    $self->_verify_event_set_sequence(-event_set_sequence => \@event_set_sequence,
                                      -sequence_count => $sequence_count,
                                      -sequence_runlength => $sequence_runlength);
    if ($self->error()) {
        $self->_set_err("The EventSet Sequence -event_set_sequence failed verification");
        return {};
    }

    # refresh the ResMgr if requested to
    if ($refresh_resources == $EventUtil::TRUE) {
        $self->ResMgr::refresh_resources();
        if ($self->error()) {
            $self->_set_err("Failed to refresh resources");
            return {};
        }
    }

    # set current EventMan state to running
    $self->_set_current_state($EventMan::STATE_RUNNING);

    my $subtest = &SanUtil::nate_subtest(
        sub {
            # display a summary of the EventMan execution here
            $self->_display_summary();
            if ($self->error()) {
                $self->SubtestComm::subtest_fail();
                return $EventUtil::FALSE;
            }

            # install the handlers that handle initialization status queries that may
            # come in from the parent process
            my $initialize_query_listener = {
                filter => sub { shift->{type} eq 'is_initializing' },
                callback => sub {
                    $Parent->message_put(type => "is_initializing_response",
                                         message => $self->{_initializing_events});
                }
            };
            my $initialized_query_listener = {
                filter => sub { shift->{type} eq 'has_initialized' },
                callback => sub {
                    $Parent->message_put(type => "has_initialized_response",
                                         message => $self->{_initialized_events});
                }
            };
            NATE::Events::listener_add($initialize_query_listener);
            NATE::Events::listener_add($initialized_query_listener);

            # perform upfront initialization before committing to the event sequence
            $self->_init_event_set_sequence();
            if ($self->error()) {
                $self->SubtestComm::subtest_fail();
                return $EventUtil::FALSE;
            }

            $self->{_initializing_events} = $EventUtil::FALSE;
            $self->{_initialized_events} = $EventUtil::TRUE;

            # if requested to initialize first, then wait here until we are sent a passthrough/start_events
            # message from the parent
            if ($initialize_only) {
                &SanUtil::nate_log("-comment", "-----INITIALIZED, WAITING TO CONTINUE-----");
                while (1) {
                    my $msg_obj = $self->SubtestComm::read_message_obj();

                    if (exists($msg_obj->{type}) and
                        $msg_obj->{type} eq 'passthrough' and
                        $msg_obj->{msg} eq 'start_events') {
                        last;
                    } elsif (exists($msg_obj->{type}) and
                             $msg_obj->{type} eq 'subtest_msg' and
                             $msg_obj->{msg} eq 'abort') {
                        return $self->SubtestComm::subtest_aborted("Requested to abort before starting events.");
                    } elsif (exists($msg_obj->{type}) and
                             $msg_obj->{type} eq 'subtest_msg' and
                             $msg_obj->{msg} eq 'stop') {
                        return $self->SubtestComm::subtest_stopped("Requested to stop before starting events.");
                    }
                    sleep 3;
                }
                &SanUtil::nate_log("-comment", "-----RECEIVED START EVENT COMMAND-----");
            }

            &SanUtil::nate_log("-comment", "------STARTING EVENT SET SEQUENCE------");

            # execute the event set sequence that was pased in
            $self->_run_event_set_sequence(-event_set_sequence => \@event_set_sequence,
                                           -sequence_count => $sequence_count,
                                           -sequence_runlength => $sequence_runlength);
            if ($self->error()) {
                &SanUtil::nate_log("-comment", "------EVENT SET SEQUENCE FAILED------");
                $self->EventUtil::_save_error();

                # uninitialize the event set sequence
                $self->_uninit_event_set_sequence();
                if ($self->error()) {
                    $self->EventUtil::_save_error();
                }

                $self->EventUtil::_restore_error();

                # set an error
                $self->_set_err("Event set sequence has failed");
                $self->SubtestComm::subtest_fail();

                return $EventUtil::FALSE;
            }

            &SanUtil::nate_log("-comment", "------COMPLETED EVENT SET SEQUENCE------");

            # uninitialize the event set sequence
            $self->_uninit_event_set_sequence();
            if ($self->error()) {
                $self->SubtestComm::subtest_fail();
                return $EventUtil::FALSE;
            }

            $self->SubtestComm::subtest_complete();
            return $EventUtil::TRUE;
        },
        -runid, "$runid", -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

    # store the subtest handle on the object
    $self->{_events_subtest} = $subtest;

    if ($async == 1) {
        # let the parent own the subtest handle in asynchronous mode (the background).
        # The parent is responsible for reaping the EventMan subtest
        # and responding to any messages it may be sending up the stack.
        return $subtest;
    } else {
        # if running synchronously (the foreground), reap this subtest and block the caller.
        # the parent should handle any errors there may be on the object
        $self->wait_for_subtests(-dependence_type => "independent",
                                 -propagate_wait => 0,
                                 -propagate_events => 0,
                                 -check_cmds_from_parent => 0,
                                 -forward_cmds_from_parent => 0,
                                 -subtests => [$subtest],
                                 -timeout => $sequence_runlength);

        # change the state to stopped, only in synchronous mode, cannot predict
        # when the state will change in asynchronous mode
        $self->_set_current_state($EventMan::STATE_STOPPED);

        if ($self->error()) {
            $self->_set_err("The events subtest failed");
            return $subtest;
        }

        return $subtest;
    }
}

###############################################################################
# Method:    send_passthrough_command()
# Objective: Broadcasts a passthrough command to all Event subtests that are
#            currently running.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub send_passthrough_command {
    my ($self, %args) = @_;

    # validate that we are currently in a running state
    # and that the {_events_subtest} attribute is defined
    # this method can only be called by the top level thpl
    if (not defined($self->{_events_subtest}) or
        not $self->is_running()) {
        $self->_set_err("Cannot send passthrough commands when the events subtest is not running or " .
                        "if the method is not invoked from the user level (thpl layer)");
        return $EventUtil::FALSE;
    }

    # send the passthrough message to the events subtest
    $self->SubtestComm::subtest_passthrough(-subtest => $self->{_events_subtest},
                                            %args);

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    can_read_event_messages()
# Objective: Determines if any new unread event messages are available
# Details:   See POD documentation at the end of this file.
###############################################################################
sub can_read_event_messages {
    my ($self) = @_;

    # issue an asynchronous update to the current state
    $self->_update_current_state();
    if ($self->error()) {
        $self->_set_err("Unable to update current state");
        return $EventUtil::FALSE;
    }

    # if there are no event messages, then return false
    if (not exists($self->{_event_messages}) or scalar(@{$self->{_event_messages}}) <= 0) {
        return $EventUtil::FALSE;
    }

    # check if there are any unread event messages waiting on the queue
    my @unread_event_messages = grep {
        if ($_->{read} == 0) {
            1;
        }  else {
            0;
        }
    } @{$self->{_event_messages}};

    if (scalar(@unread_event_messages)) {
        return $EventUtil::TRUE;
    } else {
        return $EventUtil::FALSE;
    }
}

###############################################################################
# Method:    read_event_message()
# Objective: Read the earliest unread event message.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub read_event_message {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # by default, mark all event messages as read once this function returns a message to read
    my $dont_mark_read = &ParamsMgr::get_arg(-arg => "dont_mark_read",
                                             -default => $EventUtil::FALSE);

    # by default dont read all unread messages
    my $read_all = &ParamsMgr::get_arg(-arg => "read_all",
                                       -default => $EventUtil::FALSE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return ();
    }

    # if we can't read any event messages, return empty
    if ($self->can_read_event_messages() == $EventUtil::FALSE) {
        return ();
    }

    # filter out the events that have already been read
    my @unread_event_messages = grep {
        if ($_->{read} == 0) {
            1;
        }  else {
            0;
        }
    } $self->get_all_event_messages();

    # by default only read the earliest unread event message
    if ($read_all == 0) {
        if ($dont_mark_read == 0) {
            # take the first element in the unread list, mark it read, and return it,
            $unread_event_messages[0]->{read} = 1;
        }

        # return the earliest unread event
        return ($unread_event_messages[0]);
    } else {
        if ($dont_mark_read == 0) {
            # read every single unread event message if told to mark them read
            foreach my $event (@unread_event_messages) {
                $event->{read} = 1;
            }
        }

        # return all unread messages
        return (@unread_event_messages);
    }
}

###############################################################################
# Method:    get_all_event_messages()
# Objective: Obtain all event messages that are in the stack (read/unread)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_all_event_messages {
    my ($self) = @_;

    return (wantarray()) ? @{$self->{_event_messages}} : $self->{_event_messages};
}

###############################################################################
# Method:    add_resource()
# Objective: Add a resource into the Resource Manager.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add_resource {
    my ($self, %args) = @_;

    $self->ResMgr::add_resource(%args);
    if ($self->error()) {
        $self->_set_err("Unable to register resource");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    remove_resource()
# Objective: Remove a resource from the Resource Manager.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove_resource {
    my ($self, %args) = @_;

    $self->ResMgr::remove_resource(%args);
    if ($self->error()) {
        $self->_set_err("Unable to unregister resource");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    get_resource_event_params()
# Objective: Obtain the event param hash for this resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_resource_event_params {
    my ($self, %args) = @_;

    my $hash = $self->ResMgr::get_resource_event_param_hash(%args);
    if ($self->error()) {
        $self->_set_err("Unable get event param hash for this resource.");
        return {};
    }

    return $hash;
}

###############################################################################
# Method:    get_resource_names_by_type()
# Objective: Obtain the list of resource names by resource_type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_resource_names_by_type {
    my ($self, %args) = @_;

    my @resnames = $self->ResMgr::get_resource_names_by_type(%args);
    if ($self->error()) {
        $self->_set_err("Unable get resource names by type");
        return ();
    }

    return @resnames;
}

###############################################################################
# Method:    set_resource_event_params()
# Objective: Modify the event param hash for this resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_resource_event_params {
    my ($self, %args) = @_;

    $self->ResMgr::set_resource_event_param_hash(%args);
    if ($self->error()) {
        $self->_set_err("Unable set event param hash for this resource.");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    wait_for_subtests()
# Objective: Reaper method for all EventMan subtests
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wait_for_subtests {
    my ($self, %args) = @_;

    my $return_data = [];

    eval {
        $return_data = $self->_wait_for_subtests_to_complete(%args);
    };

    if ($@) {
        # The wait for subtests to complete method threw an exception,
        # set an error on the object.
        $self->_set_err("Exception caught, encountered subtest failure: $@");
        return $return_data;
    }

    return $return_data;
}

###############################################################################
# Method:    set_default_event_set_sequence()
# Objective: Modify the default event set sequence
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_default_event_set_sequence {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @event_set_sequence = &ParamsMgr::get_arg(-arg => "event_set_sequence",
                                                 -required => 1,
                                                 -ref_type => "ARRAY");

    my $sequence_params = &ParamsMgr::get_arg(-arg => "sequence_params",
                                              -default => {},
                                              -ref_type => "HASH");
    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    $self->{_default_event_set_sequence} = \@event_set_sequence;

    # establish the default event set sequence params
    $self->{_default_event_set_sequence_params}->{sequence_runlength} = $EventMan::NO_TIME_LIMIT;
    $self->{_default_event_set_sequence_params}->{sequence_count} = 1;
    $self->{_default_event_set_sequence_params}->{sequence_style} = "iterations";
    $self->{_default_event_set_sequence_params}->{runid} = "EventMan";

    # apply all the overrides for the default parameters of the event set sequence
    foreach my $param (keys %{$sequence_params}) {
        $self->{_default_event_set_sequence_params}->{$param} = $sequence_params->{$param};
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    get_default_event_set_sequence()
# Objective: Obtain the default event set sequence
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_default_event_set_sequence {
    my ($self) = @_;

    return $self->{_default_event_set_sequence};
}

################################################################################
# Subroutine Name: generate_event_param_documentation()
# Objective:       Upon request, generate documentation for all Event
#                  Parameters used by each Utility Module and Event Modules
# Details:   See POD documentation at the end of this file.
################################################################################
sub generate_event_param_documentation {
    my ($self) = @_;

    # generate event param documentation for all modules
    foreach my $module (@UTILITY_MODULES, @EVENT_MODULES) {
        my $document_sub;
        if ($document_sub = $self->can("${module}::__generate_event_param_documentation")) {
            $self->$document_sub();
            if ($self->error()) {
                $self->_set_err("Failed to generate Event Parameter documentation for '${module}'");
                return {};
            }
        }
    }

    # parse out the event module default args hash and store it in the param documentation
    # data structure
    foreach my $module (keys %{$self->{_event_module_default_args}}) {
        foreach my $param (keys %{$self->{_event_module_default_args}->{$module}->{args}}) {
            my $new_mod_name = $module;
            $new_mod_name =~ s/EventMan::EventModules::/EM::/;
            $self->{_event_param_documentation}->{event_module_defaults}->{$new_mod_name}->{param}->{$param}->{value} =
                $self->{_event_module_default_args}->{$module}->{args}->{$param};
        }
    }

    # parse out the valid events hash and store it in the param documentation
    foreach my $key (keys %{$self->{_valid_events}}) {
        my $new_mod_name = $self->{_valid_events}->{$key}->{event_modules}->[0];
        $new_mod_name =~ s/EventMan::EventModules::/EM::/;

        $self->{_event_param_documentation}->{event_module_events}->{$new_mod_name}->{event}->{$key}->{description} =
            $self->{_valid_events}->{$key}->{description};

        $self->{_event_param_documentation}->{event_module_events}->{$new_mod_name}->{event}->{$key}->{required_resource_types} =
            join(",", @{$self->{_valid_events}->{$key}->{required_resource_types}});

        $self->{_event_param_documentation}->{event_module_events}->{$new_mod_name}->{event}->{$key}->{optional_resource_types} =
            join(",", @{$self->{_valid_events}->{$key}->{optional_resource_types}});

        $self->{_event_param_documentation}->{event_module_events}->{$new_mod_name}->{event}->{$key}->{require_at_least_one_resource_types} =
            join(",", @{$self->{_valid_events}->{$key}->{require_at_least_one_resource_types}});
     }

    # remove the event param documentation datastructure from $self and give it to the caller
    my $event_param_documentation = delete($self->{_event_param_documentation});

    return $event_param_documentation;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _register_all_event_modules()
# Objective: Register all Event Modules in the @EVENT_MODULES array.
###############################################################################
sub _register_all_event_modules {

    my ($self, $args) = @_;

    $self->{_valid_events} = {};

    # for every event module in the @EVENT_MODULES array, invoke the __register_event_module() method
    foreach my $module (@EVENT_MODULES) {
        # skip non event module packages
        next if ($module !~ /EventMan::EventModules::/);
        # register this event module
        if ( my $register_subroutine = $self->can("${module}::__register_event_module")) {

            $self->$register_subroutine($args);
            if ($self->error()) {
                $self->_set_err("Failed to register Event Module '${module}'");
                return $EventUtil::FALSE;
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _register_all_utility_modules()
# Objective: Register all Utility Modules in the @UTILITY_MODULES array.
###############################################################################
sub _register_all_utility_modules {
    my ($self, $args) = @_;

    # for every utility module in the @UTILITY_MODULES array, invoke the __register_utility_module() method
    foreach my $module (@UTILITY_MODULES) {
        # register this event module
        my $register_subroutine;
        if ($register_subroutine = $self->can("${module}::__register_utility_module")) {
            $self->$register_subroutine($args);
            if ($self->error()) {
                $self->_set_err("Failed to register Utility Module '${module}'");
                return $EventUtil::FALSE;
            }
        }
    }

    # merge the default event parameters for all utility modules in the EventMan package
    my $all_utility_module_default_args = {};
    foreach my $utility_module (@UTILITY_MODULES) {
        # add this event module's args hash keys into the hash we are building one by one
        while (my ($key, $value) = each %{$self->{_utility_module_default_args}->{$utility_module}->{args}}) {
            $all_utility_module_default_args->{args}->{$key} = $value;
        }
    }

    # now update the utility module default args on the EventMan object with the new hash
    $self->{_utility_module_default_args} = $all_utility_module_default_args;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _register_event()
# Objective: Register an Event with the EventMan package.
###############################################################################
sub _register_event {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # parse all the parameters
    my $event_name = &ParamsMgr::get_arg(-arg => "event_name",
                                         -required => 1);
    my $description = &ParamsMgr::get_arg(-arg => "description",
                                          -required => 1);
    my $enabled = &ParamsMgr::get_arg(-arg => "enabled",
                                      -default => 1);
    my $arg_defaults = &ParamsMgr::get_arg(-arg => "arg_defaults",
                                           -default => {},
                                           -ref_type => "HASH");
    my $arg_hardcoded = &ParamsMgr::get_arg(-arg => "arg_hardcoded",
                                            -default => {},
                                            -ref_type => "HASH");
    my @required_resource_types = &ParamsMgr::get_arg(-arg => "required_resource_types",
                                                      -default => [],
                                                      -ref_type => "ARRAY");
    my @require_at_least_one_resource_types = &ParamsMgr::get_arg(-arg => "require_at_least_one_resource_types",
                                                                  -default => [],
                                                                  -ref_type => "ARRAY");
    my @optional_resource_types = &ParamsMgr::get_arg(-arg => "optional_resource_types",
                                                      -default => [],
                                                      -ref_type => "ARRAY");
    my @event_modules = &ParamsMgr::get_arg(-arg => "event_modules",
                                            -required => 1,
                                            -ref_type => "ARRAY");
    my $resource_counts = &ParamsMgr::get_arg(-arg => "resource_counts",
                                              -default => {},
                                              -ref_type => "HASH");
    my $exec = &ParamsMgr::get_arg(-arg => "exec",
                                   -required => 1,
                                   -ref_type => "CODE");
    my $prejob = &ParamsMgr::get_arg(-arg => "prejob",
                                     -default => undef,
                                     -ref_type => "CODE");
    my $postjob = &ParamsMgr::get_arg(-arg => "postjob",
                                      -default => undef,
                                      -ref_type => "CODE");
    my $cojob = &ParamsMgr::get_arg(-arg => "cojob",
                                    -default => undef,
                                    -ref_type => "CODE");
    my $validate = &ParamsMgr::get_arg(-arg => "validate",
                                       -default => undef,
                                       -ref_type => "CODE");
    my $pre_validate = &ParamsMgr::get_arg(-arg => "pre_validate",
                                       -default => undef,
                                       -ref_type => "CODE");
    my $init = &ParamsMgr::get_arg(-arg => "init",
                                   -default => undef,
                                   -ref_type => "CODE");
    my $uninit = &ParamsMgr::get_arg(-arg => "uninit",
                                     -default => undef,
                                     -ref_type => "CODE");
    my @aliases = &ParamsMgr::get_arg(-arg => "aliases",
                                      -default => [],
                                      -ref_type => "ARRAY");

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # make sure the event hasnt already been registered by another event module
    if ($self->_is_event_registered(-event_name => $event_name)) {
        $self->_set_err("Could not register event $event_name, an Event " .
                        "module has already registered it");
        return $EventUtil::FALSE;
    }

    # register the specified event description in the valid events hash attribute
    $self->{_valid_events}->{$event_name}->{description} = $description;

    # register whether this event is enabled or disabled by the event implementer
    $self->{_valid_events}->{$event_name}->{enabled} = $enabled;

    # register the event parameter defaults for this event
    # strip the dashes on the event params
    %{$arg_defaults} = &SanUtil::strip_dash(%{$arg_defaults});
    $self->{_valid_events}->{$event_name}->{args} = $arg_defaults;

    # register the hardcoded event params for this event
    %{$arg_hardcoded} = &SanUtil::strip_dash(%{$arg_hardcoded});
    $self->{_valid_events}->{$event_name}->{args_hardcoded} = $arg_hardcoded;

    # store the list of required resource types this event can use
    $self->{_valid_events}->{$event_name}->{required_resource_types} = \@required_resource_types;

    # store the list of require one resource types this event can use
    $self->{_valid_events}->{$event_name}->{require_at_least_one_resource_types} = \@require_at_least_one_resource_types;

    # store the list of optional resource types this event can use
    $self->{_valid_events}->{$event_name}->{optional_resource_types} = \@optional_resource_types;

    # store the resource_counts for this event
    $self->{_valid_events}->{$event_name}->{resource_counts} = $resource_counts;

    # store the list of event modules this event is represented by
    $self->{_valid_events}->{$event_name}->{event_modules} = \@event_modules;

    # store the subroutine reference handles for this event
    $self->{_valid_events}->{$event_name}->{exec} = $exec;
    $self->{_valid_events}->{$event_name}->{postjob} = $postjob;
    $self->{_valid_events}->{$event_name}->{prejob} = $prejob;
    $self->{_valid_events}->{$event_name}->{cojob} = $cojob;
    $self->{_valid_events}->{$event_name}->{validate} = $validate;
    $self->{_valid_events}->{$event_name}->{pre_validate} = $pre_validate;
    $self->{_valid_events}->{$event_name}->{init} = $init;
    $self->{_valid_events}->{$event_name}->{uninit} = $uninit;

    # register all aliases to this event
    delete($args{event_name});
    delete($args{-event_name});
    delete($args{-aliases});
    delete($args{aliases});
    foreach my $alias (@aliases) {
        $self->_register_event(-event_name => $alias,
                               %args);
        if ($self->error()) {
            $self->_set_err("Unable to register event alias '$alias' for event '$event_name");
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _register_event_module_default_args()
# Objective: Register the default EventParameter arguments for the specified
#            EventModule.
###############################################################################
sub _register_event_module_default_args {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # parse all the parameters
    my $event_module = &ParamsMgr::get_arg(-arg => "event_module",
                                           -required => 1);
    my $arg_defaults = &ParamsMgr::get_arg(-arg => "arg_defaults",
                                           -ref_type => "HASH",
                                           -default => {});

    my $arg_hardcoded = &ParamsMgr::get_arg(-arg => "arg_hardcoded",
                                            -ref_type => "HASH",
                                            -default => {});

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # strip the dashes on the event params
    %{$arg_defaults} = &SanUtil::strip_dash(%{$arg_defaults});

    $self->{_event_module_default_args}->{$event_module}->{args} = $arg_defaults;

    %{$arg_hardcoded} = &SanUtil::strip_dash(%{$arg_hardcoded});

    $self->{_event_module_hardcoded_args}->{$event_module}->{args} = $arg_hardcoded;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _register_utility_module_default_args()
# Objective: Register the default EventParameter arguments for the specified
#            Utility Module.
###############################################################################
sub _register_utility_module_default_args {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # parse all the parameters
    my $utility_module = &ParamsMgr::get_arg(-arg => "utility_module",
                                             -required => 1);
    my $arg_defaults = &ParamsMgr::get_arg(-arg => "arg_defaults",
                                           -ref_type => "HASH",
                                           -default => {});

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # strip the dashes on the utility params
    %{$arg_defaults} = &SanUtil::strip_dash(%{$arg_defaults});

    $self->{_utility_module_default_args}->{$utility_module}->{args} = $arg_defaults;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _validate_event()
# Objective: Invoke the Validate method for the specified Event object.
###############################################################################
sub _validate_event {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # parse all the parameters
    my $event_obj = &ParamsMgr::get_arg(-arg => "event_obj",
                                        -ref_type => "EventMan::Event",
                                        -required => 1);
    my $event_set_obj = &ParamsMgr::get_arg(-arg => "event_set_obj",
                                            -ref_type => "EventMan::EventSet",
                                            -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $event_name = $event_obj->get_name();

    # run pre_validation routines
    if ($self->{_valid_events}->{$event_name}->{pre_validate}) {
        my $pre_validate_method = $self->{_valid_events}->{$event_name}->{pre_validate};
        if (defined($pre_validate_method)) {
            $self->$pre_validate_method(-event_obj => $event_obj,
                                        -event_set_obj => $event_set_obj);
            if ($self->error()) {
                $self->_set_err("The event '$event_name' failed pre validation");
                return $EventUtil::FALSE;
            }
        }
    }

    # if not requested to perform event validation, return now
    if ($event_obj->is_validation_enabled() == $EventUtil::FALSE) {
        return $EventUtil::TRUE;
    }

    # if this event is not enabled, then set an error
    if ($self->{_valid_events}->{$event_name}->{enabled} != 1) {
        $self->_set_err("The Event '$event_name' is not enabled and thus cannot be used.");
        return $EventUtil::FALSE;
    }

    # validate that the required resources are available for this event
    my @required_resource_types = @{$self->{_valid_events}->{$event_name}->{required_resource_types}};
    foreach my $resource_type (@required_resource_types) {
        my @resources = $self->ResMgr::get_current_resources_by_type(-type => $resource_type);
        if ($self->error()) {
            $self->_set_err("Unable to get resource types for type '$resource_type'");
            return $EventUtil::FALSE;
        }

        # make sure that there are resources of this type
        if (not scalar(@resources)) {
            $self->_set_err("Unable to validate event '$event_name', missing required resource type '$resource_type'");
            return $EventUtil::FALSE;
        }
    }

    # validate that at least one of the @require_at_least_one_resource_types are available
    my @require_at_least_one_resource_types = @{$self->{_valid_events}->{$event_name}->{require_at_least_one_resource_types}};
    if (scalar(@require_at_least_one_resource_types)) {
        my @resources = ();
        foreach my $resource_type (@require_at_least_one_resource_types) {
            push(@resources, $self->ResMgr::get_current_resources_by_type(-type => $resource_type));
            if ($self->error()) {
                $self->_set_err("Unable to get resource types for type '$resource_type'");
                return $EventUtil::FALSE;
            }
        }

        # if none of the resource types were found, then validation failed
        if (not scalar(@resources)) {
            $self->_set_err("Unable to validate event '$event_name', at least one of the " .
                            "resource types '@require_at_least_one_resource_types' must be represented");
            return $EventUtil::FALSE;
        }
    }

    # validate that the proper resource counts are available
    my $resource_counts = $self->{_valid_events}->{$event_name}->{resource_counts};
    if (defined($resource_counts)) {
        # validate that every resource count is represented accurately in this event
        # IE: if the resource count of "sanfilers7m" is 2, then there must be exactly
        # two sanfilers7m resources in this events environment.
        # IE: if the resource count of "sanclients" is 2+, then there must be at least
        # two sanclients resources in this events environment.
        # IE: if the resource count of "sanswitches" is 2-, then there must be no more than
        # two sanswitches resources in this events environment.
        foreach my $res_type (keys %{$resource_counts}) {

            # parse the resource count for this resource type
            my $res_count = $resource_counts->{$res_type};
            if ($res_count =~ /^\s*(\d+)(\+|\-)?(?:,(even|odd))?\s*$/i) {
                my $count = $1;
                my $operator = $2;
                my $even_or_odd = $3;

                # determine how many resources there are of this type
                my @resources = $self->ResMgr::get_current_resources_by_type(-type => $res_type);
                if ($self->error()) {
                    $self->_set_err("Unable to get resources by type '$res_type'");
                    return $EventUtil::FALSE;
                }

                if (defined($operator) and $operator eq "+") {
                    # at least $count must exist
                    if (scalar(@resources) < $count) {
                        $self->_set_err("There must be at least $count resources of type '$res_type' for the event $event_name");
                        return $EventUtil::FALSE;
                    }
                    # check if the resource count is an even or odd number
                    if (defined($even_or_odd)) {
                        if ($even_or_odd =~ /even/i) {
                            if (scalar(@resources) % 2 != 0) {
                                $self->_set_err("There must be an even number of resources of type '$res_type' for the event $event_name");
                                return $EventUtil::FALSE;
                            }
                        } elsif ($even_or_odd =~ /odd/i) {
                            if (scalar(@resources) % 2 == 0) {
                                $self->_set_err("There must be an odd number of resources of type '$res_type' for the event $event_name");
                                return $EventUtil::FALSE;
                            }
                        }
                    }
                } elsif (defined($operator) and $operator eq "-") {
                    # no more than $count must exist
                    if (scalar(@resources) > $count) {
                        $self->_set_err("There must be no more than $count resources of type '$res_type' for the event $event_name");
                        return $EventUtil::FALSE;
                    }
                    # check if the resource count is an even or odd number
                    if (defined($even_or_odd)) {
                        if ($even_or_odd =~ /even/i) {
                            if (scalar(@resources) % 2 != 0) {
                                $self->_set_err("There must be an even number of resources of type '$res_type' for the event $event_name");
                                return $EventUtil::FALSE;
                            }
                        } elsif ($even_or_odd =~ /odd/i) {
                            if (scalar(@resources) % 2 == 0) {
                                $self->_set_err("There must be an odd number of resources of type '$res_type' for the event $event_name");
                                return $EventUtil::FALSE;
                            }
                        }
                    }
                } else {
                    # exactly $count resources must exist
                    if (scalar(@resources) != $count) {
                        $self->_set_err("There must be exactly $count resources of type '$res_type' for the event $event_name");
                        return $EventUtil::FALSE;
                    }
                }
            }
        }
    }

    # execute the validate method for this event if it exists
    if (exists($self->{_valid_events}->{$event_name}->{validate})) {
        my $validate_method = $self->{_valid_events}->{$event_name}->{validate};
        if (defined($validate_method)) {
            $self->$validate_method(-event_obj => $event_obj,
                                    -event_set_obj => $event_set_obj);
            if ($self->error()) {
                $self->_set_err("The event '$event_name' failed validation");
                return $EventUtil::FALSE;
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _is_event_registered()
# Objective: Determine if the specified event is already registered.
###############################################################################
sub _is_event_registered {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # parse all the parameters
    my $event_name = &ParamsMgr::get_arg(-arg => "event_name",
                                         -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # check if the requested event is currently registered in the valid
    # events hash
    if (exists($self->{_valid_events}->{$event_name})) {
        return $EventUtil::TRUE;
    } else {
        return $EventUtil::FALSE;
    }
}

###############################################################################
# Method:    _get_current_state()
# Objective: Obtain the current state of the EventMan subtest
###############################################################################
sub _get_current_state {
    my ($self) = @_;

    return $self->{_current_state};
}

###############################################################################
# Method:    _set_current_state()
# Objective: Change the current state of the EventMan subtest
###############################################################################
sub _set_current_state {
    my ($self, $state) = @_;

    $self->{_current_state} = $state;
}

###############################################################################
# Method:    _update_current_state()
# Objective: Asynchronously update the current state of the EventMan subtest
###############################################################################
sub _update_current_state {
    my ($self) = @_;

    # poke the event loop
    NATE::Events::eventloop_once(timeout => 0);

    # if the current state is in a terminal state, then do nothing
    if ($self->_get_current_state() eq $EventMan::STATE_NOTSTARTED or
        $self->_get_current_state() eq $EventMan::STATE_STOPPED or
        $self->_get_current_state() eq $EventMan::STATE_ABORTED or
        $self->_get_current_state() eq $EventMan::STATE_TIMEDOUT) {
        return $EventUtil::TRUE;
    } elsif (not defined($self->{_events_subtest}) or not $self->{_events_subtest}->isAlive()) {
        # change the state to stop if the events subtest is dead
        $self->_set_current_state($EventMan::STATE_STOPPED);
    } else {
        # otherwise the events subtest is running,
        # asynchronously query for an update on the subtest handle
        # by waiting on it for one iteration
        $self->wait_for_subtests(-dependence_type => "independent",
                                 -propagate_wait => 0,
                                 -propagate_events => 0,
                                 -check_cmds_from_parent => 0,
                                 -forward_cmds_from_parent => 0,
                                 -async => 1,
                                 -subtests => [$self->{_events_subtest}]);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # we need to verify that if the subtest is completed, that we
        # move the eventman state to a terminal state if it is not currently in one
        if ($self->_get_current_state() eq $EventMan::STATE_RUNNING and
            (not defined($self->{_events_subtest}) or
             not $self->{_events_subtest}->isAlive())) {
            # EventMan state is running, but the subtest is not running,
            # move the eventman state to completed because it is possible
            # that the previous call to wait_for_subtests() reaped the
            # subtest and did not change the state
            $self->_set_current_state($EventMan::STATE_STOPPED);
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _verify_event_set_sequence()
# Objective: Verify that the specified event set sequence is error free.
###############################################################################
sub _verify_event_set_sequence {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @event_set_sequence = &ParamsMgr::get_arg(-arg => "event_set_sequence",
                                                 -default => $self->get_default_event_set_sequence(),
                                                 -ref_type => "ARRAY");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # event sets can arrive in several ways
    # -event_set_sequence => ["my event set", $event_set_group_obj, $a_event_set_obj]
    # in this example:
    # (1) "my event set" will be issued an Event set lookup
    # (2) $event_set_group_obj will be discovered to be a fso object and all its event sets
    # will be verified
    # (3) $a_event_set_obj will be verified as an Event set object

    # reset the event modules in use hash attribute
    $self->{_event_modules_in_use} = {};

    # initialize the sequence string, this will be used for display purposes
    my $event_sequence_string = "";

    # verify that each element of the sequence is valid
    foreach my $element (@event_set_sequence) {
        # if this element is an Event set group then verify that this event
        # set group contains valid event set objects
        if (ref($element) eq "EventMan::EventSetGroup") {
            my $event_set_group_obj = $element;

            # add this eventsetgroup to the sequence string
            $event_sequence_string .= $event_set_group_obj->get_group_name() . "::"
                                      . $event_set_group_obj->get_run_style() . "[";

            # make sure this eventsetgroup contains blessed eventset objects
            my $event_set_objs = $event_set_group_obj->get_event_sets();
            if (scalar(@{$event_set_objs}) == 0) {
                # then this eventsegroup is empty, this is not valid
                $self->_set_err("An EventSetGroup in -event_set_sequence does not contain event set objects");
                return $EventUtil::FALSE;
            }

            # make sure each event set in this group is a valid EventSet object
            foreach my $event_set_obj (@{$event_set_objs}) {
                if (ref($event_set_obj) ne "EventMan::EventSet") {
                    # then this is not a valid eventset
                    $self->_set_err("An EventSetGroup in -event_set_sequence contains an " .
                                    "invalid EventSet : " . ref($event_set_obj));
                    return $EventUtil::FALSE;
                }

                # verify the event list on this event set
                my $event_set_name = $event_set_obj->get_name();
                $self->_verify_event_list($event_set_obj);
                if ($self->error()) {
                    $self->_set_err("Failed to verify EventSet '$event_set_name'");
                    return $EventUtil::FALSE;
                }

                # add this event set to the sequence string
                my $event_set_cycle_count = $event_set_obj->get_cycle_count();
                $event_sequence_string .= $event_set_name . "(x${event_set_cycle_count}), ";
            }

            # get rid of the last character on the sequence string and close out the eventset group
            chop($event_sequence_string);
            chop($event_sequence_string);
            $event_sequence_string .= "], ";

        } else {
            # otherwise assume this element is an Event set (by name or by object)
            if (ref($element) eq "") {
                # then this is an Event set name, verify that this event set exists
                my $event_set_obj = $self->get_event_set(-event_set_name => $element);
                if ($self->error()) {
                    $self->_set_err("Invalid member of -event_set_sequence : '$element'");
                    return $EventUtil::FALSE;
                }

                # update the sequence string with this event set
                my $event_set_cycle_count = $event_set_obj->get_cycle_count();
                $event_sequence_string .= $element . "(x${event_set_cycle_count}), ";

                # verify its event list
                $self->_verify_event_list($event_set_obj);
                if ($self->error()) {
                    $self->_set_err("Failed to verify EventSet '$element'");
                    return $EventUtil::FALSE;
                }

            } elsif (ref($element) eq "EventMan::EventSet") {
                # then this element is a valid eventset object
                # verify its event list
                my $event_set_name = $element->get_name();
                $self->_verify_event_list($element);
                if ($self->error()) {
                    $self->_set_err("Failed to verify EventSet '$event_set_name'");
                    return $EventUtil::FALSE;
                }

                # update the sequence string with this event set
                my $event_set_cycle_count = $element->get_cycle_count();
                $event_sequence_string .= $event_set_name . "(x${event_set_cycle_count}), ";

            } else {
                # then this is an invalid EventSet object
                $self->_set_err("Invalid member of -event_set_sequence : " . ref($element));
                return $EventUtil::FALSE;
            }
        }
    }

    # store the event sequence string for display purposes later
    chop($event_sequence_string);
    chop($event_sequence_string);
    $self->{_event_sequence_string} = $event_sequence_string;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _verify_event_list()
# Objective: Verify that the specified event set contains a valid event list
###############################################################################
sub _verify_event_list {
    my ($self, $event_set) = @_;

    my @event_list = $event_set->get_event_list();
    my $event_set_name = $event_set->get_name();

    # make sure the event list is not empty
    if (scalar(@event_list) == 0) {
        $self->_set_err("The EventSet '$event_set_name' has an empty event list");
        return $EventUtil::FALSE;
    }

    # verify each event object in this EventSet
    foreach my $event_obj (@event_list) {
        if (ref($event_obj) ne "EventMan::Event") {
            $self->_set_err("The EventSet '$event_set_name' contains an invalid Event in its list");
            return $EventUtil::FALSE;
        }

        # parse out the event modules and store each on
        # the event_module_in_use attribute hash, this will be used later in the
        # _init_event_set_sequence and _unit_event_set_sequence methods
        # to dynamically determine the proper order in which to
        # initialize and uninitialize event modules
        my @event_modules = $event_obj->get_event_modules();
        my $count = 1;
        foreach my $module (@event_modules) {
            $self->{_event_modules_in_use}->{$module} += $count;
            $count++;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _run_event_set_sequence()
# Objective: Execute the specified event set sequence.
###############################################################################
sub _run_event_set_sequence {
    my ($self, %args) = @_;

    # event sets can arrive in several ways
    # -event_set_sequence => ["my event set", $sequential_event_set_group_obj, $parallel_event_set_group_obj, $a_event_set_obj]
    # in this example:
    # (1) "my event set" will be issued an Event set lookup, and that event set will be run first
    # (2) $sequential_event_set_group_obj will be discovered to be an Eventsetgroup object and all its event set members
    # will be run in the order that they were added.
    # (3) $parallel-event_set_group_obj will be discovered to be a fso object and all its event set members will be run
    # in parallel.
    # (4) $a_event_set_obj will be discovered to be a fs object and it will be run

    # parse out the sequence
    &ParamsMgr::process_args(%args);

    my @event_set_sequence = &ParamsMgr::get_arg(-arg => "event_set_sequence",
                                                 -default => $self->get_default_event_set_sequence(),
                                                 -ref_type => "ARRAY");
    my $sequence_count = &ParamsMgr::get_arg(-arg => "sequence_count",
                                             -default => 1);
    my $sequence_runlength = &ParamsMgr::get_arg(-arg => "sequence_runlength",
                                                 -default => $EventMan::NO_TIME_LIMIT);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $total_event_set_count = 1;

    # run the event set sequence $sequence_count times or until the runlength has been exhausted
    for (my $curr_event_set_seq_count = 1 ; $curr_event_set_seq_count <= $sequence_count ; $curr_event_set_seq_count++) {

        # subtest off each event set (or eventsetgroup) (in sequence)
        foreach my $element (@event_set_sequence) {

            # before we continue to run any eventsets,
            # make sure that we are still in the RUNNING state
            if ($self->_get_current_state() eq $EventMan::STATE_STOPPED) {
                # make this subtest stop now
                $self->SubtestComm::subtest_stopped("Events have stopped gracefully");
                return $EventUtil::TRUE;
            } elsif ($self->_get_current_state() eq $EventMan::STATE_ABORTED) {
                # make this subtest abort now
                $self->SubtestComm::subtest_aborted("Events have aborted gracefully");
                return $EventUtil::TRUE;
            } elsif ($self->_get_current_state() eq $EventMan::STATE_TIMEDOUT) {
                # if we have timedout, make subtest timeout now
                $self->SubtestComm::subtest_timedout();
                return $EventUtil::TRUE;
            }

            # if this element is an EventSetGroup, then handle the subtesting in a different manner
            if (ref($element) eq "EventMan::EventSetGroup") {
                my $event_set_group_obj = $element;
                $self->_run_event_set_group(-event_set_group_obj => $event_set_group_obj,
                                            -runlength => $sequence_runlength,
                                            -curr_event_set_seq_count => $curr_event_set_seq_count,
                                            -total_event_set_count => $total_event_set_count);
                if ($self->error()) {
                    $self->_set_err("Error encountered while running event set group");
                    return $EventUtil::FALSE;
                }
            } else {
                # otherwise this is a single EventSet and it should run by itself
                my $event_set_obj;

                # dynamically determine how to handle $element (as an Event set obj or an Event set name)
                if (ref($element) eq "") {
                    # then this element is the string of an Event set name, convert it to an Event set object
                    $event_set_obj = $self->get_event_set(-event_set_name => $element);
                    if ($self->error()) {
                        $self->_set_err("Event set '$element' does not exist");
                        return $EventUtil::FALSE;
                    }
                } elsif (ref($element) eq "EventMan::EventSet") {
                    # this element is an Event set object,
                    $event_set_obj = $element;
                } else {
                    # this element is not a valid object
                    $self->_set_err("Invalid member of -event_set_sequence '$element'");
                    return $EventUtil::FALSE;
                }

                # execute this event set (not checking for error because run event set doesnt set one)
                my $event_set_subtest = $self->_run_event_set(-event_set_obj => $event_set_obj,
                                                              -curr_event_set_seq_count => $curr_event_set_seq_count,
                                                              -total_event_set_count => $total_event_set_count);

                # wait for this event set to complete
                my ($return_data) = $self->wait_for_subtests(-dependence_type => "independent",
                                                             -propagate_wait => 0,
                                                             -check_cmds_from_parent => 1,
                                                             -forward_cmds_from_parent => 1,
                                                             -subtests => [$event_set_subtest],
                                                             -timeout => $sequence_runlength);
                if ($self->error()) {
                    # save the error away so that we can process event return data
                    $self->EventUtil::_save_error();
                }

                # process the return data from the eventset subtest
                $self->_process_event_return_data(-return_data => $return_data,
                                                  -propagate => $EventUtil::FALSE);
                if ($self->error()) {
                    $self->_set_err("Encountered error when processing return data messages.");
                    return $EventUtil::FALSE;
                }

                # restore any error saved away
                $self->EventUtil::_restore_error();

                # if there is an error after reaping the eventset, bail out
                if ($self->error()) {
                    $self->_set_err("Encountered error while reaping eventset");
                    return $EventUtil::FALSE;
                }
            }

            $total_event_set_count++;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _init_event_set_sequence()
# Objective: Initialize the specified Event sequence, by invoking the init()
#            method for all Utility Modules. Then invoke the init() methods
#            for all Event Modules.
###############################################################################
sub _init_event_set_sequence {
    my ($self) = @_;

    # this function will initialize every Utility Module in @UTILITY_MODULES
    foreach my $utility_module (@UTILITY_MODULES) {
        my $init_subroutine;
        if ($init_subroutine = $self->can("${utility_module}::__init")) {
            &SanUtil::nate_log("-comment", "------INITIALIZING UTILITY MODULE ${utility_module}------");
            $self->$init_subroutine();
            if ($self->error()) {
                $self->_set_err("Failed to initialize Utility Module '${utility_module}'");
                return $EventUtil::FALSE;
            }
        }
    }

    # organize the event modules in inheritence order
    my @event_modules = $self->_sort_event_module_order();

    # invoke the init method in every event module that is
    # represented in the event set sequence that is about to be executed
    # scan through the event set sequence and store all the event modules in a hash
    foreach my $event_module (@event_modules) {
        my $init_subroutine;
        if ($init_subroutine = $self->can("${event_module}::__init")) {
            &SanUtil::nate_log("-comment", "------INITIALIZING EVENT MODULE ${event_module}------");
            $self->$init_subroutine();
            if ($self->error()) {
                $self->_set_err("Failed to initialize Event Module '${event_module}'");
                return $EventUtil::FALSE;
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _sort_event_module_order()
# Objective: Determine the Event Module order based on each one's ref count
###############################################################################
sub _sort_event_module_order {
    my ($self) = @_;

    # convert the event_modules_in_use hash into an array where the elements at the
    # beginning of the array have the highest value
    my $tmp_hash = {};
    foreach my $key (keys %{$self->{_event_modules_in_use}}) {
        push(@{$tmp_hash->{$self->{_event_modules_in_use}->{$key}}}, $key);
    }

    my @modules = ();
    foreach my $key (sort {$a <=> $b} keys %{$tmp_hash}) {
        push(@modules, @{$tmp_hash->{$key}});
    }

    return reverse @modules;
}

###############################################################################
# Method:    _uninit_event_set_sequence()
# Objective: Uninitialize the specified Event sequence, by invoking the uninit()
#            method for all Event Modules. Then invoke the unint() method for
#            all Utility Modules.
###############################################################################
sub _uninit_event_set_sequence {
    my ($self) = @_;

    # organize the event modules in inheritence order
    my @event_modules = $self->_sort_event_module_order();

    # invoke the uninit method in every event module that was
    # represented in the event set sequence that was executed
    foreach my $event_module (reverse @event_modules) {
        my $uninit_subroutine;
        if ($uninit_subroutine = $self->can("${event_module}::__uninit")) {
            &SanUtil::nate_log("-comment", "------UNINITIALIZING EVENT MODULE ${event_module}------");
            $self->$uninit_subroutine();
            if ($self->error()) {
                $self->_set_err("Failed to uninitialize Event Module '${event_module}'");
                return $EventUtil::FALSE;
            }
        }
    }

    # invoke the uninit method on every Utility Module in reverse order
    foreach my $utility_module (reverse @UTILITY_MODULES) {
        my $uninit_subroutine;
        if ($uninit_subroutine = $self->can("${utility_module}::__uninit")) {
            &SanUtil::nate_log("-comment", "------UNINITIALIZING UTILITY MODULE ${utility_module}------");
            $self->$uninit_subroutine();
            if ($self->error()) {
                $self->_set_err("Failed to uninitialize Utility Module '${utility_module}'");
                return $EventUtil::FALSE;
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _run_event_set_group()
# Objective: Execute the specified EventSetGroup object.
###############################################################################
sub _run_event_set_group {
    my ($self, %args) = @_;

    # parse out the sequence
    &ParamsMgr::process_args(%args);

    my $event_set_group_obj = &ParamsMgr::get_arg(-arg => "event_set_group_obj",
                                                  -required => 1,
                                                  -ref_type => "EventMan::EventSetGroup");
    my $curr_event_set_seq_count = &ParamsMgr::get_arg(-arg => "curr_event_set_seq_count",
                                                       -required => 1);
    my $total_event_set_count = &ParamsMgr::get_arg(-arg => "total_event_set_count",
                                                    -required => 1);
    my $runlength = &ParamsMgr::get_arg(-arg => "runlength",
                                        -default => $EventMan::NO_TIME_LIMIT);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # obtain all the event sets that are part of this group
    my @event_sets = @{$event_set_group_obj->get_event_sets()};

    # determine what the run style is of this group (sequential/parallel)
    # and the dependence type and if synchronization is enabled
    my $run_style = $event_set_group_obj->get_run_style();
    my $dependence_type = $event_set_group_obj->get_dependence_type();
    my $enable_sync = $event_set_group_obj->is_sync_enabled();

    # run the event sets in this eventsetgroup in sequential order
    if ($run_style =~ /sequential/i) {
        my @errmsg;

        # execute each event set in sequential order
        while (scalar(@event_sets)) {

            # before we start this eventset, make sure that we were not told to stop by the parent
            if ($self->_get_current_state() eq $EventMan::STATE_STOPPED) {
                # a parent has ordered a stop, stop this subtest now
                $self->SubtestComm::subtest_stopped("Events have stopped gracefully");
                return $EventUtil::FALSE;
            } elsif ($self->_get_current_state() eq $EventMan::STATE_ABORTED) {
                # make this subtest abort now
                $self->SubtestComm::subtest_aborted("Events have aborted gracefully");
                return $EventUtil::FALSE;
            } elsif ($self->_get_current_state() eq $EventMan::STATE_TIMEDOUT) {
                # if we have timedout, make this subtest timeout now
                $self->SubtestComm::subtest_timedout();
                return $EventUtil::TRUE;
            }

            # execute this event set (not checking for error because run event set doesnt set one)
            my $event_set_obj = shift(@event_sets);
            my $event_set_subtest = $self->_run_event_set(-event_set_obj => $event_set_obj,
                                                          -curr_event_set_seq_count => $curr_event_set_seq_count,
                                                          -total_event_set_count => $total_event_set_count);

            # wait for this event set to complete
            my ($return_data) = $self->wait_for_subtests(-dependence_type => $dependence_type,
                                                         -propagate_wait => 0,
                                                         -check_cmds_from_parent => 1,
                                                         -forward_cmds_from_parent => 1,
                                                         -subtests => [$event_set_subtest],
                                                         -timeout => $runlength);

            # if this event set died and the dependence type is "independent", then
            # all other remaining event sets in this event set group should be
            # allowed to keep running
            if ($self->error()) {
                if ($dependence_type =~ /independent/i) {
                    # if there are no more event sets to run, then return the error now
                    if (not scalar(@event_sets)) {
                        # save the error
                        $self->EventUtil::_save_error();

                        # process the return data from the eventset subtest
                        $self->_process_event_return_data(-return_data => $return_data,
                                                          -propagate => $EventUtil::FALSE);
                        if ($self->error()) {
                            $self->_set_err("Encountered error when processing return data messages.");
                            return $EventUtil::FALSE;
                        }

                        # restore the error and return false
                        $self->EventUtil::_restore_error();

                        return $EventUtil::FALSE;
                    } else {
                        # otherwise, there are still event sets in the eventsetgroup
                        # we need to allow them to keep running, so clear the error and
                        # push the error message on @errmsg so that it can be used later
                        $self->_clear_err();
                        push(@errmsg, $self->errmsg());

                        # process the return data from the eventset subtest
                        $self->_process_event_return_data(-return_data => $return_data,
                                                          -propagate => $EventUtil::FALSE);
                        if ($self->error()) {
                            $self->_set_err("Encountered error when processing return data messages.");
                            return $EventUtil::FALSE;
                        }
                    }
                } else {
                    # otherwise there was an error and the eventsets are dependent,

                    # save the error
                    $self->EventUtil::_save_error();

                    # process the return data from the eventset subtest
                    $self->_process_event_return_data(-return_data => $return_data,
                                                      -propagate => $EventUtil::FALSE);
                    if ($self->error()) {
                        $self->_set_err("Encountered error when processing return data messages.");
                        return $EventUtil::FALSE;
                    }

                    # restore the error and return bail out now
                    $self->EventUtil::_restore_error();

                    return $EventUtil::FALSE;
                }
            } else {
                # this eventset subtest completed without error
                # process the return data from the eventset subtest
                $self->_process_event_return_data(-return_data => $return_data,
                                                  -propagate => $EventUtil::FALSE);
                if ($self->error()) {
                    $self->_set_err("Encountered error when processing return data messages.");
                    return $EventUtil::FALSE;
                }
            }
        }

        # if there was errmsg stored, then we need to reassign it back onto the object
        # and return the error
        if (scalar(@errmsg)) {
            $self->{errmsg} = \@errmsg;
            $self->{error} = 1;
            return $EventUtil::FALSE;
        }

    } elsif ($run_style =~ /parallel/i) {
        # execute all the event sets in this eventsetgroup in parallel
        my @event_set_subtests = ();
        foreach my $event_set_obj (@event_sets) {
            # not checking for error because run_event_set does not set one
            my $event_set_subtest = $self->_run_event_set(-event_set_obj => $event_set_obj,
                                                          -enable_sync => $enable_sync,
                                                          -curr_event_set_seq_count => $curr_event_set_seq_count,
                                                          -total_event_set_count => $total_event_set_count);

            push(@event_set_subtests, $event_set_subtest);
        }

        # wait for all event sets to complete in parallel
        my ($return_data) = $self->wait_for_subtests(-dependence_type => $dependence_type,
                                                     -propagate_wait => 0,
                                                     -check_cmds_from_parent => 1,
                                                     -forward_cmds_from_parent => 1,
                                                     -enable_sync => $enable_sync,
                                                     -subtests => [@event_set_subtests],
                                                     -timeout => $runlength);
        if ($self->error()) {
            # save the error so we can process return data
            $self->EventUtil::_save_error();
        }

        # process the return data from the eventset subtests
        $self->_process_event_return_data(-return_data => $return_data,
                                          -propagate => $EventUtil::FALSE);
        if ($self->error()) {
            $self->_set_err("Encountered error when processing return data messages.");
            return $EventUtil::FALSE;
        }

        # restore any error saved away
        $self->EventUtil::_restore_error();

        # if there is an error after reaping the eventset, bail out
        if ($self->error()) {
            $self->_set_err("Encountered error while reaping eventset");
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _run_event_set()
# Objective: Execute the specified EventSet object.
###############################################################################
sub _run_event_set {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $event_set_obj = &ParamsMgr::get_arg(-arg => "event_set_obj",
                                            -required => 1,
                                            -ref_type => "EventMan::EventSet");
    my $enable_sync = &ParamsMgr::get_arg(-arg => "enable_sync",
                                          -default => $EventUtil::FALSE);
    my $curr_event_set_seq_count = &ParamsMgr::get_arg(-arg => "curr_event_set_seq_count",
                                                       -required => 1);
    my $total_event_set_count = &ParamsMgr::get_arg(-arg => "total_event_set_count",
                                                    -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # pad the event_set sequence count to 2 digits and the total event set count to 3 digits
    my $curr_event_set_seq_count_runid = $curr_event_set_seq_count;
    my $total_event_set_count_runid = $total_event_set_count;
    $curr_event_set_seq_count_runid = "0" . $curr_event_set_seq_count if (length($curr_event_set_seq_count) == 1);
    $total_event_set_count_runid = "00" . $total_event_set_count if (length($total_event_set_count) == 1);
    $total_event_set_count_runid = "0" . $total_event_set_count if (length($total_event_set_count) == 2);

    # convert all spaces in the event set name into underscores for the run id
    my $event_set_name = $event_set_obj->get_name();
    my $event_set_runid = $event_set_name;
    $event_set_runid =~ s/\s+/_/g;

    my $event_set_cycle_count = $event_set_obj->get_cycle_count();

    # execute this event set
    my $event_set_subtest = subtest(
        sub {
            # keep a counter of the total number of events executed
            my $total_event_count = 1;

            # display a summary of what this EventSet is going to do
            $event_set_obj->display_summary(-objects => [$self, $event_set_obj],
                                            -event_man_obj => $self);
            if ($event_set_obj->error()) {
                $self->_set_err($event_set_obj->errmsg());
                $self->SubtestComm::subtest_fail();
                return $EventUtil::FALSE;
            }

            # run this event set's list $event_set_cycle_count times
            for (my $curr_cycle = 1; $curr_cycle <= $event_set_cycle_count; $curr_cycle++) {

                # process the event list in this eventset (this will randomize the list if necessary)
                $event_set_obj->process_event_list();

                # execute each event in this event set
                foreach my $event_obj ($event_set_obj->get_event_list()) {

                    # determine how many times this specific event needs to be executed
                    my $event_cycle_count = $event_obj->get_cycle_count();

                    # run this specific event "$event_cycle_count" times in a row
                    for (my $curr_event_cycle = 0; $curr_event_cycle < $event_cycle_count; $curr_event_cycle++) {

                        # before we start this event, verify that we are still in the RUNNING state
                        if ($self->_get_current_state() eq $EventMan::STATE_STOPPED) {
                            # a parent has ordered a stop, make this event set stop now
                            $self->SubtestComm::subtest_stopped("Events have stopped gracefully");
                            return $EventUtil::FALSE;
                        } elsif ($self->_get_current_state() eq $EventMan::STATE_ABORTED) {
                            # make this subtest abort now
                            $self->SubtestComm::subtest_aborted("Events have aborted gracefully");
                            return $EventUtil::FALSE;
                        } elsif ($self->_get_current_state() eq $EventMan::STATE_TIMEDOUT) {
                            # if we have timedout, make this event set timeout now
                            $self->SubtestComm::subtest_timedout();
                            return $EventUtil::TRUE;
                        }

                        # update the current event and eventset objects,
                        # this will allow the "EventMan::EventUtil::get_arg()" method to implicitly
                        # point to the current event and event set objects that are in use
                        $self->_update_current_event(-event_obj => $event_obj,
                                                     -event_set_obj => $event_set_obj,
                                                     -curr_event_set_seq_count => $curr_event_set_seq_count,
                                                     -curr_eventset_cycle => $curr_cycle,
                                                     -total_event_count => $total_event_count,
                                                     -total_event_set_count => $total_event_set_count);
                        if ($self->error()) {
                            $self->_set_err("Failed to update current event");
                            $self->SubtestComm::subtest_fail();
                            return $EventUtil::FALSE;
                        }

                        # log an Event that the event is starting
                        $self->SubtestComm::log_event_message(-event_message => "event_start_" . $event_obj->get_name());

                        &SanUtil::nate_log("-comment", "------STARTING EVENT: " . $event_obj->get_name()  . "------");

                        # execute this event
                        my (@event_subtests) = $self->_run_event(-event_obj => $event_obj,
                                                                 -event_set_obj => $event_set_obj,
                                                                 -curr_cycle => $curr_cycle,
                                                                 -total_event_count => $total_event_count);
                        if ($self->error()) {
                            &SanUtil::nate_log("-comment", "------EVENT FAILED: " . $event_obj->get_name()  . "------");

                            # log an Event that the event failed
                            $self->SubtestComm::log_event_message(-event_message => "event_failed_" . $event_obj->get_name());

                            # reset the current event set and event objects
                            $self->_reset_current_event();

                            $self->SubtestComm::subtest_fail();
                            return $EventUtil::FALSE;
                        }

                        # incremement the total event count
                        $total_event_count++;

                        # wait for this event (and possibly the event cojob) to complete
                        my ($return_data) = $self->wait_for_subtests(-dependence_type => $event_obj->get_cojob_dependence_type(),
                                                                     -propagate_wait => 1,
                                                                     -check_cmds_from_parent => 1,
                                                                     -forward_cmds_from_parent => 1,
                                                                     -enable_sync => $enable_sync,
                                                                     -subtests => \@event_subtests);
                        if ($self->error()) {
                            &SanUtil::nate_log("-comment", "------EVENT FAILED: " . $event_obj->get_name()  . "------");

                            # log an Event message that the event failed
                            $self->SubtestComm::log_event_message(-event_message => "event_failed_" . $event_obj->get_name());

                            # save the error because we still need to process event return data
                            $self->EventUtil::_save_error();
                        } else {
                            # log an Event that the event completed successfully
                            $self->SubtestComm::log_event_message(-event_message => "event_complete_" . $event_obj->get_name());

                            &SanUtil::nate_log("-comment", "------EVENT COMPLETE: " . $event_obj->get_name()  . "------");
                        }

                        # process the return data from this event subtest
                        $self->_process_event_return_data(-return_data => $return_data,
                                                          -propagate => $EventUtil::TRUE);
                        if ($self->error()) {
                            # reset the current event set and event objects
                            $self->_reset_current_event();

                            $self->_set_err("Failed to process event return data");
                            $self->SubtestComm::subtest_fail();
                            return $EventUtil::FALSE;
                        }

                        # reset the current event set and event objects
                        $self->_reset_current_event();

                        # restore any error that may be stored away
                        $self->EventUtil::_restore_error();

                        # if there was an error on the event, fail the subtest
                        if ($self->error()) {
                            $self->SubtestComm::subtest_fail();
                            return $EventUtil::FALSE;
                        }
                    }
                }
            }

            # this event set is now complete
            $self->SubtestComm::subtest_complete();
            return $EventUtil::TRUE;
        }, -runid, "${curr_event_set_seq_count_runid}_${total_event_set_count_runid}_${event_set_runid}",
        -bg, "-onexit" => sub {my ($self) = @_; $self->exithandler(1);});

    # return the subtest of the event set
    return $event_set_subtest;
}

###############################################################################
# Method:    _process_event_return_data()
# Objective: Process return_data subtest messages that may have been
#            propagated by children subtests.
###############################################################################
sub _process_event_return_data {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @return_data = &ParamsMgr::get_arg(-arg => "return_data",
                                          -default => [],
                                          -ref_type => "ARRAY");
    my $propagate = &ParamsMgr::get_arg(-arg => "propagate",
                                        -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # handle all known return_data message types
    foreach my $message_obj (@return_data) {
        if ($message_obj->{msg} eq "set_resource_attribute") {
            # update this resource's specified attribute
            $self->ResMgr::set_resource_attribute(-name => $message_obj->{data}->{name},
                                                  -type => $message_obj->{data}->{type},
                                                  -attribute => $message_obj->{data}->{attribute},
                                                  -value => $message_obj->{data}->{value},
                                                  -timestamp => $message_obj->{data}->{timestamp},
                                                  -propagate => $propagate);
            if ($self->error()) {
                $self->_set_err("Failed to set resource attribute on " . $message_obj->{data}->{name});
                return $EventUtil::FALSE;
            }
        } elsif ($message_obj->{msg} eq "modify_resource_obj") {
            # update this resource's object value
            $self->ResMgr::modify_resource_obj(-name => $message_obj->{data}->{name},
                                               -type => $message_obj->{data}->{type},
                                               -value => $message_obj->{data}->{value},
                                               -serialized => $message_obj->{data}->{serialized},
                                               -reconnect => $EventUtil::TRUE,
                                               -propagate => $propagate);
            if ($self->error()) {
                $self->_set_err("Failed to modify resource obj on " . $message_obj->{data}->{name});
                return $EventUtil::FALSE;
            }
        } elsif ($message_obj->{msg} eq "sram_list_update") {
            # issue an sram list update
            $self->SRAMDetection::sram_list_update(-filer_name => $message_obj->{data}->{name},
                                                   -sram_list => $message_obj->{data}->{sram_list},
                                                   -propagate => $propagate);
            if ($self->error()) {
                $self->_set_err("Failed to update SRAM dump list on resource " . $message_obj->{data}->{name});
                return $EventUtil::FALSE;
            }
        } else {
            # unknown return data messages get thrown away.
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _run_event()
# Objective: Execute the specified Event Object.
###############################################################################
sub _run_event {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_set_obj = &ParamsMgr::get_arg(-arg => "event_set_obj",
                                            -required => 1,
                                            -ref_type => "EventMan::EventSet");
    my $event_obj = &ParamsMgr::get_arg(-arg => "event_obj",
                                        -required => 1,
                                        -ref_type => "EventMan::Event");
    my $curr_cycle = &ParamsMgr::get_arg(-arg => "curr_cycle",
                                         -required => 1);
    my $total_event_count = &ParamsMgr::get_arg(-arg => "total_event_count",
                                                -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return ();
    }

    # pad the cycle count and event count to 3 digits
    $curr_cycle = "0" . $curr_cycle if (length($curr_cycle) == 1);
    $total_event_count = "00" . $total_event_count if (length($total_event_count) == 1);
    $total_event_count = "0" . $total_event_count if (length($total_event_count) == 2);

    # reconnect only to resources that had become faulted and unfaulted
    $self->EventUtil::reconnect_to_previously_faulted_resources();
    if ($self->error()) {
        $self->_set_err("Error when reconnecting to all previously faulted resources");
        return ();
    }

    # get the event job subroutine handle
    my $event_job = $self->_get_event_job(-event_obj => $event_obj,
                                          -job_type => "exec");
    if ($self->error()) {
        $self->_set_err("Unable to obtain the 'exec' method for this event");
        return ();
    }

    # get the event init subroutine handle
    my $event_init = $self->_get_event_job(-event_obj => $event_obj,
                                           -job_type => "init");
    if ($self->error()) {
        $self->_set_err("Unable to obtain the 'init' method for this event");
        return ();
    }

    # get the event prejob subroutine handle
    my $event_prejob = $self->_get_event_job(-event_obj => $event_obj,
                                             -job_type => "prejob");
    if ($self->error()) {
        $self->_set_err("Unable to obtain the 'prejob' method for this event");
        return ();
    }

    # get the event postjob subroutine handle
    my $event_postjob = $self->_get_event_job(-event_obj => $event_obj,
                                              -job_type => "postjob");
    if ($self->error()) {
        $self->_set_err("Unable to obtain the 'postjob' method for this event");
        return ();
    }

    # get the event cojob subroutine handle
    my $event_cojob = $self->_get_event_job(-event_obj => $event_obj,
                                            -job_type => "cojob");
    if ($self->error()) {
        $self->_set_err("Unable to obtain the 'cojob' method for this event");
        return ();
    }

    # get the event uninit subroutine handle
    my $event_uninit = $self->_get_event_job(-event_obj => $event_obj,
                                             -job_type => "uninit");
    if ($self->error()) {
        $self->_set_err("Unable to obtain the 'uninit' method for this event");
        return ();
    }

    my @subtests;

    # dispatch this event subtest
    my $event_subtest = subtest(
        sub {
            # display a summary of what this event is going to do
            $event_obj->display_summary(-objects => [$self, $event_set_obj, $event_obj],
                                        -event_man_obj => $self);
            if ($event_obj->error()) {
                $self->_set_err($event_obj->errmsg());
                $self->SubtestComm::subtest_fail();
                return $EventUtil::FALSE;
            }

            &SanUtil::nate_log('-comment', "EventMan Memory Usage: " . Devel::Size::total_size($self) . " bytes");

            # reconnect to nacl nodes and nacl nodes & clusters
            $self->EventUtil::reconnect_all_nacl_nodes(-faulted => $EventUtil::FALSE,
                                                       -no_version_change => $EventUtil::FALSE);
            if ($self->error()) {
                $self->SubtestComm::subtest_fail();
                return $EventUtil::FALSE;
            }
            $self->EventUtil::reconnect_all_nacl_clusters();
            if ($self->error()) {
                $self->SubtestComm::subtest_fail();
                return $EventUtil::FALSE;
            }

            # execute this event's init method if one exists
            if (defined($event_init)) {
                &SanUtil::nate_log("-comment", "------STARTING EVENT INITIALIZATION------");
                $self->$event_init(-event_set_obj => $event_set_obj,
                                   -event_obj => $event_obj);
                if ($self->error()) {
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }
            }

            # execute this event's prejob method if one exists
            if ($event_obj->is_prejob_enabled() and defined($event_prejob)) {
                &SanUtil::nate_log("-comment", "------STARTING EVENT PREJOB------");

                $self->$event_prejob(-event_set_obj => $event_set_obj,
                                     -event_obj => $event_obj);
                if ($self->error()) {
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }
            }

            # run this event
            &SanUtil::nate_log("-comment", "------STARTING EVENT------");

            $self->$event_job(-event_set_obj => $event_set_obj,
                              -event_obj => $event_obj);
            if ($self->error()) {
                &SanUtil::nate_log("-comment", "------EVENT FAILED------");

                # execute this event's uninit method if one exists (event if the event failed)
                if (defined($event_uninit)) {
                    &SanUtil::nate_log("-comment", "------STARTING EVENT UNINITIALIZATION (event failed)------");
                    $self->$event_uninit(-event_set_obj => $event_set_obj,
                                         -event_obj => $event_obj,
                                         -state => "event_failed");
                }

                $self->SubtestComm::subtest_fail();
                return $EventUtil::FALSE;
            }

            &SanUtil::nate_log("-comment", "------EVENT COMPLETE------");

            # execute this event's postjob method if one exists
            # this will only be run if the event succeeded
            if ($event_obj->is_postjob_enabled() and defined($event_postjob)) {
                &SanUtil::nate_log("-comment", "------STARTING EVENT POSTJOB------");
                $self->$event_postjob(-event_set_obj => $event_set_obj,
                                      -event_obj => $event_obj);
                if ($self->error()) {

                    # execute this event's uninit method if one exists (event if the event failed)
                    if (defined($event_uninit)) {
                        &SanUtil::nate_log("-comment", "------STARTING EVENT UNINITIALIZATION (event postjob failed)------");
                        $self->$event_uninit(-event_set_obj => $event_set_obj,
                                             -event_obj => $event_obj,
                                             -state => "event_failed");
                    }
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }
            }

            # execute this event's uninit method if one exists
            if (defined($event_uninit)) {
                &SanUtil::nate_log("-comment", "------STARTING EVENT UNINITIALIZATION------");
                $self->$event_uninit(-event_set_obj => $event_set_obj,
                                     -event_obj => $event_obj);
                if ($self->error()) {
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }
            }

            # complete this event subtest
            $self->SubtestComm::subtest_complete();
            return $EventUtil::TRUE;
        },
        -runid, "${curr_cycle}_${total_event_count}_" . $event_obj->get_name(),
        -bg,  "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

    push(@subtests, $event_subtest);

    # simultaneously, execute this event's cojob method in it's own subtest if
    # cojob was enabled on this event
    if ($event_obj->is_cojob_enabled() and defined($event_cojob)) {
        my $event_cojob = subtest(
            sub {
                # execute this events cojob method
                $self->$event_cojob(-event_set_obj => $event_set_obj,
                                    -event_obj => $event_obj);
                if ($self->error()) {
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                } else {
                    # complete this cojob
                    $self->SubtestComm::subtest_complete();
                    return $EventUtil::TRUE;
                }
            },
            -runid, "${curr_cycle}_${total_event_count}_" . $event_obj->get_name() . "_cojob",
            -bg,  "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
        push(@subtests, $event_cojob);
    }

    return @subtests;
}

###############################################################################
# Method:    _get_event_job()
# Objective: Obtain the specified event job subroutine from the Event Object
#            passed in.
###############################################################################
sub _get_event_job {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_obj = &ParamsMgr::get_arg(-arg => "event_obj",
                                        -required => 1,
                                        -ref_type => "EventMan::Event");
    my $job_type = &ParamsMgr::get_arg(-arg => "job_type",
                                       -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return undef;
    }

    # attempt to get the prejob/postjob/cojob/init/uninit/exec methods from the
    # event object. Since prejob/postjob/cojob can be overridden by the user of the library,
    # we must fetch them from the event object.
    # exec/init/uninit are hardcoded subroutines defined by the author of the
    # event module so they must be fetched from the valid events hash
    my $return_subroutine;
    if ($job_type =~ /^prejob$/) {
        $return_subroutine = $event_obj->get_prejob();
    } elsif ($job_type =~ /^postjob$/) {
        $return_subroutine = $event_obj->get_postjob();
    } elsif ($job_type =~ /^cojob$/) {
        $return_subroutine = $event_obj->get_cojob();
    } elsif ($job_type =~ /^exec$/) {
        $return_subroutine = $self->{_valid_events}->{$event_obj->get_name()}->{exec};
    } elsif ($job_type =~ /^init$/) {
        $return_subroutine = $self->{_valid_events}->{$event_obj->get_name()}->{init};
    } elsif ($job_type =~ /^uninit$/) {
        $return_subroutine = $self->{_valid_events}->{$event_obj->get_name()}->{uninit};
    } else {
        $self->_set_err("Invalid value for -job_type '$job_type', " .
                        "use 'exec|prejob|postjob|cojob|init|uninit'");
        return undef;
    }

    if ($event_obj->error()) {
        $self->_set_err($event_obj->errmsg());
        return undef;
    }

    # if the event object had a valid value stored in the object, use it
    if (defined($return_subroutine)) {
        return $return_subroutine;
    } else {
        # otherwise default to the value that is in the valid events hash
        return $self->{_valid_events}->{$event_obj->get_name()}->{$job_type};
    }

}

###############################################################################
# Method:    _update_current_event()
# Objective: Update the pointer to the current Event and EventSet object, also
#            prepare the environment by dynamically loading the ISA array
#            with the event modules that will be inherited for the current
#            event. Load in the current EventModule Event Parameters defaults.
###############################################################################
sub _update_current_event {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_set_obj = &ParamsMgr::get_arg(-arg => "event_set_obj",
                                            -required => 1,
                                            -ref_type => "EventMan::EventSet");
    my $event_obj = &ParamsMgr::get_arg(-arg => "event_obj",
                                        -required => 1,
                                        -ref_type => "EventMan::Event");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    delete($args{-event_obj});
    delete($args{-event_set_obj});

    # store the event object and event set object on the current event hash
    $self->{_current_event}->{event_obj} = $event_obj;
    $self->{_current_event}->{event_set_obj} = $event_set_obj;

    # dynamically modify the @ISA array by inheriting the event modules
    # for the current event
    my @event_modules = $event_obj->get_event_modules();
    foreach my $event_module (@event_modules) {
        # add this event module to the @ISA array
        push(@ISA, $event_module);
    }

    # load in the default event parameters for all event modules in the inheritence tree
    # reverse the event module list so that the primary module gets its defaults
    # written into the hash last.
    my $current_event_module_default_args = {};
    foreach my $event_module (reverse @event_modules) {
        # add this event module's args hash keys into the hash we are building one by one
        while (my ($key, $value) = each %{$self->{_event_module_default_args}->{$event_module}->{args}}) {
            $current_event_module_default_args->{args}->{$key} = $value;
        }
    }

    # now its possible that a specific event keyword has established its own
    # event parameter default values, push them onto the event module default args hash so
    # that they overwrite any defaults defined by the modules themselves
    while (my ($key, $value) = each %{$self->{_valid_events}->{$event_obj->get_name()}->{args}}) {
        $current_event_module_default_args->{args}->{$key} = $value;
    }

    # its possible that the event modules have hardcoded parameters load them in as well
    my $current_event_module_hardcoded_args = {};
    foreach my $event_module (reverse @event_modules) {
        # add this event module's args hash keys into the hash we are building one by one
        while (my ($key, $value) = each %{$self->{_event_module_hardcoded_args}->{$event_module}->{args_hardcoded}}) {
            $current_event_module_hardcoded_args->{args}->{$key} = $value;
        }
    }

    # now its possible that a specific event keyword has established its own
    # event parameter hardcoded values, push them onto the event module hardcoded args hash so
    # that they overwrite any hardcoded params defined by the modules themselves
    while (my ($key, $value) = each %{$self->{_valid_events}->{$event_obj->get_name()}->{args_hardcoded}}) {
        $current_event_module_hardcoded_args->{args}->{$key} = $value;
    }

    # now update the current event module default args on the EventMan object
    $self->{_current_event}->{event_module_default_args} = $current_event_module_default_args;
    $self->{_current_event}->{event_module_hardcoded_args} = $current_event_module_hardcoded_args;

    # store any additional event data passed into this method into the current event hash
    %args = &SanUtil::strip_dash(%args);
    $self->{_current_event}->{event_data} = \%args;

    # generate the current event resources,
    # this represents a list of all the resources used by this event
    $self->{_current_event}->{event_data}->{resources} = $self->ResMgr::get_current_resources();
    if ($self->error()) {
        $self->_set_err("Failed to get current resources");
        return $EventUtil::FALSE;
    }
    $self->{_current_event}->{event_data}->{sub_resources} = $self->ResMgr::get_current_sub_resources();
    if ($self->error()) {
        $self->_set_err("Failed to get current sub_resources");
        return $EventUtil::FALSE;
    }

    # store the event name and event set name in the event data hash
    $self->{_current_event}->{event_data}->{event_name} = $event_obj->get_name();
    $self->{_current_event}->{event_data}->{event_set_name} = $event_set_obj->get_name();

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _reset_current_event()
# Objective: Undo everything that was loaded by the _update_current_event()
#            method
###############################################################################
sub _reset_current_event {
    my ($self) = @_;

    # pop all the event modules that were added onto the @ISA array
    # for the current event
    my @event_modules = $self->{_current_event}->{event_obj}->get_event_modules();
    foreach my $event_module (@event_modules) {
        pop(@ISA);
    }

    # reset all the current event/eventset/eventmodule pointers
    $self->{_current_event} = undef;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _get_current_event_data()
# Objective: Obtain the stored event data from the current running event.
###############################################################################
sub _get_current_event_data {
    my ($self) = @_;

    # if there is no event data on the current event, return an empty hash
    if (not defined($self->{_current_event}) or
        not defined($self->{_current_event}->{event_data})) {
        return undef;
    }

    # copy out the event data from the current event and return it
    return $self->{_current_event}->{event_data};
}

###############################################################################
# Method:    _get_current_event()
# Objective: Obtain the current Event Object that is running.
###############################################################################
sub _get_current_event {
    my ($self) = @_;

    return $self->{_current_event}->{event_obj};
}

###############################################################################
# Method:    _get_current_event_set()
# Objective: Obtain the current EventSet Object that is running.
###############################################################################
sub _get_current_event_set {
    my ($self) = @_;

    return $self->{_current_event}->{event_set_obj};
}

################################################################################
# Method:    _wait_for_subtests_to_complete()
# Objective:       This will wait for each subtest and reap failed subtests
#                  without disrupting any other subtest.
################################################################################
sub _wait_for_subtests_to_complete {
    my ($self, %args) = @_;

    # parse all the params
    &ParamsMgr::process_args(%args);

    my $dependence_type = &ParamsMgr::get_arg(-arg => "dependence_type",
                                              -default => "independent");
    my $encountered_failure = &ParamsMgr::get_arg(-arg => "encountered_failure",
                                                  -default => $EventUtil::FALSE);
    my $propagate_wait = &ParamsMgr::get_arg(-arg => "propagate_wait",
                                             -default => $EventUtil::FALSE);
    my $propagate_events = &ParamsMgr::get_arg(-arg => "propagate_events",
                                               -default => $EventUtil::TRUE);
    my $enable_sync = &ParamsMgr::get_arg(-arg => "enable_sync",
                                          -default => $EventUtil::FALSE);
    my $sync_callbacks = &ParamsMgr::get_arg(-arg => "sync_callbacks",
                                             -default => undef,
                                             -ref_type => "ARRAY");
    my $check_cmds_from_parent = &ParamsMgr::get_arg(-arg => "check_cmds_from_parent",
                                                     -default => $EventUtil::FALSE);
    my $forward_cmds_from_parent = &ParamsMgr::get_arg(-arg => "forward_cmds_from_parent",
                                                       -default => $EventUtil::FALSE);
    my $last_error_string = &ParamsMgr::get_arg(-arg => "last_error_string",
                                                -default => "");
    my $async = &ParamsMgr::get_arg(-arg => "async",
                                    -default => $EventUtil::FALSE);
    my $poll_time = &ParamsMgr::get_arg(-arg => "poll_time",
                                        -default => 0.5);
    my @return_data = &ParamsMgr::get_arg(-arg => "return_data",
                                          -default => [],
                                          -ref_type => "ARRAY");
    my $timeout = &ParamsMgr::get_arg(-arg => "timeout",
                                      -default => $EventMan::NO_TIME_LIMIT);
    my @subtests = &ParamsMgr::get_arg(-arg => "subtests",
                                       -required => 1,
                                       -ref_type => "ARRAY");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # validate that there are subtests to work with
    if (not scalar(@subtests)) {
        $self->_set_err("Invalid value for -subtests");
        return (wantarray()) ? () : [];
    }

    # asynchronous mode and subtest synchronization cannot be enabled at the same time, error out if
    # this is the case
    if ($async == $EventUtil::TRUE and $enable_sync == $EventUtil::TRUE) {
        $self->_set_err("Asynchronous mode (-async => 1) and Subtest synchronization " .
                        "(-enable_sync => 1) are mutually exclusive");
        return (wantarray()) ? () : [];
    }

    my $subtests_complete = 0;
    my $count = 0;
    my (@waiting_subtests) = ();
    my (@orig_subtests) = @subtests;
    my (@resmgr_updates) = ();

    # loop over all the subtests and check their status
    while (1) {
        &SanUtil::nate_process_results();
        my $num_subtests_left = scalar(@subtests);
        my $subtest_is_dead = 0;

        # if there are no subtests left, quit
        if ($num_subtests_left < 1) {
            last;
        }

        # pull this subtest off the array
        my $subtest = shift(@subtests);
        my $runid = $subtest->runid();

        # if the subtest is not alive, dont push it back on the array.
        if (not $subtest->isAlive()) {

            # if the exit status was not 512/0 (pass/warn)
            # and the encountered failure flag was not triggered, then its
            # possible that the child subtest terminately non gracefully
            if (($subtest->exit_status() ne "512" and
                 $subtest->exit_status() ne "0" and
                 $subtest->exit_status() ne "64000" and
                 $subtest->exit_status() ne "64768")
                    and $encountered_failure == 0) {
                # enable the encountered failure flag in this case
                $encountered_failure = 1;
                $last_error_string = "The subtest ($runid) terminated abnormally " . $subtest->exit_status();
            }

            # go to the next subtest
            next;
        }

        # while it's alive and we can still read from it, process EventMan messages one by one.
        # dont block on the message get call for more than $poll_time value
        my $message_obj;
        if ($Log->may_debug()) {
            $Log->debug("Checking for messages from ($runid)");
        }
        while ($message_obj = $subtest->message_get(filter => sub { my $m = shift; $m->{type} =~ /EventMan/;},
                                                    timeout => $poll_time,
                                                    on_timeout => sub { return 0; })) {
            if ($Log->may_debug()) {
                $Log->debug("Received message from ($runid): " . Data::Dump::dump($message_obj));
            }

            # test the subtest status
            if (not $subtest->isAlive()) {
                $Log->debug("Subtest $subtest is no longer alive");

                # otherwise this subtest is no longer running, break out
                $subtest_is_dead = 1;

                # if the exit status was not 512 (pass/warn)
                # and the encountered failure flag was not triggered, then its
                # possible that the child subtest terminately non gracefully
                if (($subtest->exit_status() ne "512" and $subtest->exit_status() ne "0")
                        and $encountered_failure == 0) {
                    # enable the encountered failure flag in this case
                    $encountered_failure = $EventUtil::TRUE;
                    $last_error_string = "The subtest ($runid) terminated abnormally " . $subtest->exit_status();
                }

                # drop the message if we got one
                $message_obj = undef;

                # stop reading from this subtest
                last;
            }

            # if this message is of type 'subtest_msg' handle each message type
            if ($message_obj->{type} eq "subtest_msg") {
                my $message = "";
                $message = $message_obj->{msg};

                # if the subtest failed, then terminate the subtest
                if ($message eq "fail") {
                    # in this case, the child subtest encountered a fatal failure that must be treated as an error
                    # the error should eventually be set on the object.
                    $encountered_failure = $EventUtil::TRUE;

                    &SanUtil::nate_log('-warn', "The subtest ($runid) failed");

                    # parse out the error string that the subtest wants to bubble up
                    $last_error_string = $message_obj->{data}->{error_string};

                    # if all sibling subtests are dependent, then if one dies, all siblings should be reaped
                    if ($dependence_type eq "dependent") {
                        &SanUtil::nate_log('-comment', "Terminating all subtests (dependence_type == $dependence_type)");
                        $subtest->kill(message => ["-warn", "The subtest ($runid) failed and is being killed."]);
                        $subtest->wait_finish();
                        foreach my $st (@orig_subtests) {
                            # dont reap the first subtest more than once.
                            if ($st->runid() ne $runid) {
                                $st->kill(message => ["-warn", "A dependent sibling subtest ($runid) failed. Terminating this subtest."]);
                                $st->wait_finish();
                            }
                        }
                        last;
                    } else {
                        # otherwise, just kill off the subtest that died
                        # and allow all siblings to complete
                        $subtest->kill(message => ["-warn", "The subtest ($runid) failed and is being killed."]);
                        $subtest->wait_finish();
                        last;
                    }
                } elsif ($message eq "fail_ignore") {
                    # in this case, this subtest is telling us to reap it, but ignore the failure.
                    # just terminate this subtest with a warning and ignore the error.
                    &SanUtil::nate_log('-warn', "The subtest ($runid) failed (ignoring, since it was a graceful failure)");
                    $subtest->kill(message => ["-warn", "The subtest ($runid) failed and is being killed (ignoring error since it was a graceful failure)."]);
                    $subtest->wait_finish();
                    last;
                } elsif ($message eq "waiting") {
                    # in this case, this subtest is waiting for us to tell it to continue

                    # if synchronization was requested by the caller then move this subtest
                    # into the waiting queue
                    if ($enable_sync == $EventUtil::TRUE) {
                        &SanUtil::nate_log('-comment', "Subtest ($runid) is waiting to continue. Awaiting sync point.");
                        push(@waiting_subtests, $subtest);
                        $subtest_is_dead = 1;

                        # parse the data payload if any
                        if (defined($message_obj->{data_type})) {
                            # if its a resmgr update, push the data payload to the resmgr update array
                            # for later processing
                            if ($message_obj->{data_type} eq "resmgr_update") {
                                push(@resmgr_updates, $message_obj->{data});
                            }
                        }
                        last;
                    } else {
                        # otherwise sync is disabled so tell the subtest to continue on immediately
                        $self->SubtestComm::subtest_continue(-subtest => $subtest);
                    }
                } elsif ($message eq "complete") {
                    # terminate this subtest and mark it pass
                    &SanUtil::nate_log('-comment', "The subtest ($runid) has completed");
                    $subtest->kill(message => ["-comment", "The subtest ($runid) completed."]);
                    $subtest->wait_finish();

                    # stop all sibling subtests if dependence type (stop after one completes)
                    if ($dependence_type eq "stop_after_one_completes") {
                        &SanUtil::nate_log('-comment', "Terminating all sibling subtests of ($runid) (dependence_type == $dependence_type)");
                        foreach my $st (@orig_subtests) {
                            # dont reap the first subtest more than once.
                            if ($st->runid() ne $runid) {
                                $st->kill(message => ["-comment", "A dependent sibling subtest ($runid) completed before this one " .
                                                      "and dependence_type == $dependence_type. Terminating this subtest."]);
                                $st->wait_finish();
                            }
                        }
                    }

                    last;
                } elsif ($message eq "stopped") {
                    # this child is telling us it has entered the stopped state
                    # (due to the parent asking all subtests to stop)
                    # issue a kill pass on the subtest handle
                    &SanUtil::nate_log('-comment', "The subtest ($runid) has stopped");
                    $subtest->kill(message => ["-comment", "The subtest ($runid) has stopped."]);
                    $subtest->wait_finish();
                    last;
                } elsif ($message eq "aborted") {
                    # this child is telling us it has entered the aborted state
                    # (due to the parent asking all subtests to abort gracefully)
                    # issue a kill warn on the subtest handle
                    &SanUtil::nate_log('-comment', "The subtest ($runid) has aborted");
                    $subtest->kill(message => ["-warn", "The subtest ($runid) has aborted."]);
                    $subtest->wait_finish();
                    last;
                } elsif ($message eq "timedout") {
                    # this child is telling us it is in the timedout state
                    # issue a kill pass on the subtest handle
                    &SanUtil::nate_log('-comment', "The subtest ($runid) has timed out");
                    $subtest->kill(message => ["-comment", "The subtest ($runid) has timed out."]);
                    $subtest->wait_finish();
                    last;
                }
            } elsif ($message_obj->{type} eq "event_message") {
                # handle all event messages here

                # if requested to propagate event messagess, then bubble up the message object
                if ($propagate_events == $EventUtil::TRUE) {
                    # propagate this event message up to the parent
                    $Parent->message_put(type => "EventMan/" . $message_obj->{type},
                                         message => $message_obj);
                } else {
                    # otherwise store the event message on the object at this level.
                    push(@{$self->{_event_messages}}, $message_obj);
                }
            } elsif ($message_obj->{type} eq "return_data") {
                # if this is a return data message, then push the object onto the return data array.
                # this array will eventually be returned to the caller of this function
                push(@return_data, $message_obj);
            } else {
                # an unknown message was received, log it as a warning
                &SanUtil::nate_log('-warn', "Received an unknown message '" . $message_obj->{type} .
                                            "' from subtest ($runid)");
            }
        }

        # if the subtest died at any point during the reading phase, then dont push it on the array
        if ($subtest_is_dead == 1) {
            next;
        }

        # at this point the subtest is still alive and running.
        # push it back on the array
        push(@subtests, $subtest);

        $count++;

        # check if the parent has sent any commands for us to process
        if ($check_cmds_from_parent == $EventUtil::TRUE) {
            $Log->debug("Checking for commands from parent");
            while ($self->SubtestComm::can_read()) {
                my $message_obj = $self->SubtestComm::read_message_obj();

                # handle subtest command messages
                if ($message_obj->{type} eq "subtest_msg") {

                    # if the parent wants us to stop, change our current state to stop
                    if ($message_obj->{msg} eq "stop") {
                        &SanUtil::nate_log(-comment, "Parent has requested that we stop gracefully");
                        $self->_set_current_state($EventMan::STATE_STOPPED);

                        # we need to broadcast this stop message down into all children subtests that are
                        # still running/waiting
                        if ($forward_cmds_from_parent == $EventUtil::TRUE) {
                            foreach my $sub (@subtests, @waiting_subtests) {
                                $self->SubtestComm::subtest_stop(-subtest => $sub);
                            }
                        }
                    }

                    # check if the parent is ordering us to abort, we need to tell our
                    # children subtests that are still running to abort
                    elsif ($message_obj->{msg} eq "abort") {
                        &SanUtil::nate_log(-comment, "Parent has requested that we abort gracefully");
                        $self->_set_current_state($EventMan::STATE_ABORTED);

                        # we need to broadcast this abort message down into all children subtests that are
                        # still running
                        if ($forward_cmds_from_parent == $EventUtil::TRUE) {
                            foreach my $sub (@subtests, @waiting_subtests) {
                                $self->SubtestComm::subtest_abort(-subtest => $sub);
                            }
                        }
                    }

                    # check if the parent is telling us to time out and that the timeout
                    # has not been hit already ($timeout > $EventMan::TIME_LIMIT_REACHED)
                    elsif ($message_obj->{msg} eq "timeout" and $timeout > $EventMan::TIME_LIMIT_REACHED) {
                        &SanUtil::nate_log(-comment, "Parent has timed out, telling children to timeout");

                        # change the current state to 'timedout'
                        $self->_set_current_state($EventMan::STATE_TIMEDOUT);

                        # we need to push this timeout message down into all children subtests that are still running
                        if ($forward_cmds_from_parent == $EventUtil::TRUE) {
                            foreach my $sub (@subtests, @waiting_subtests) {
                                $self->SubtestComm::subtest_timeout(-subtest => $sub);
                            }
                        }

                        # disable the timeout so we don't flood the children subtests with timeout commands
                        $timeout = $EventMan::TIME_LIMIT_REACHED;
                    }
                }

                # if the message is a passthrough message then keep forwarding the message to the children subtests
                elsif ($message_obj->{type} eq "passthrough") {
                    # if requested to forward commands from the parent to our children, do so now
                    if ($forward_cmds_from_parent == $EventUtil::TRUE) {
                        foreach my $sub (@subtests, @waiting_subtests) {
                            $sub->message_put(type => "EventMan/" . $message_obj->{type},
                                              message => $message_obj);
                        }
                    } else {
                        # otherwise this message was intended for us, since we have already read the message object
                        # store it on this object's message queue so that it can be processed outside this function
                        if (not $self->SubtestComm::restore_message_obj_on_queue(-message_obj => $message_obj)) {
                            # this means we have seen this message before on the queue, which means we probably
                            # wrapped the message queue, so break out of the while can read loop
                            last;
                        }
                    }
                }
            }
        }

        # if we havnt already timed out or the timeout isnt disabled
        if ($timeout >= 0) {
            # check if we have timed out
            if (time() - $self->{_start_time} >= $timeout) {
                &SanUtil::nate_log(-comment, "Timeout ($timeout seconds) has exceeded, timing out children");

                # we have exceeded the timeout, issue a timeout command to all children
                $self->_set_current_state($EventMan::STATE_TIMEDOUT);
                foreach my $sub (@subtests, @waiting_subtests) {
                    $self->SubtestComm::subtest_timeout(-subtest => $sub);
                }

                # disable the timeout so we don't flood the children subtests with timeout commands
                $timeout = $EventMan::TIME_LIMIT_REACHED;
            }
        }

        # if we've gone through all the subtests, reset the count and sleep for a bit
        if ($num_subtests_left <= $count) {

            # if async mode is on, then break out now, we only want to perform
            # one reaping cycle through the subtest list when -async => 1
            if ($async == $EventUtil::TRUE) {
                last;
            }

            # reset the subtest counter
            $count = 0;
        }

    }

    # if there are any subtests waiting at a synchronization point, tell all subtests to continue, and
    # recursively invoke this function with all the subtests that are waiting.
    if (scalar(@waiting_subtests)) {

        # update the resmgr attrs
        if (scalar(@resmgr_updates)) {
            $self->ResMgr::update_resmgr_attrs(-resmgr_attrs => \@resmgr_updates);
            if ($self->error()) {
                $self->_set_err("Failed to update ResMgr attributes during synchronization");
                return (wantarray()) ? () : [];;
            }

            # reset the resmgr updates
            @resmgr_updates = ();
        }

        # before we can tell the children to continue we have to first obtain permission from the parent
        # unless we were told to not propagate a synchronization any further
        if ($propagate_wait == $EventUtil::TRUE) {
            $self->EventUtil::synchronize();
        }

        # if a sync callback was passed in, invoke it now, pass in the eventman object as the first arg
        if (defined($sync_callbacks)) {
            foreach my $callback_method (@{$sync_callbacks}) {
                if (ref($callback_method) ne "CODE") {
                    $self->_set_err("Invalid CODE reference for sync_callback method");
                    return (wantarray()) ? () : [];;
                }
                $callback_method->($self);
                if ($self->error()) {
                    # abort all the subtests that are waiting
                    foreach my $sub (@waiting_subtests) {
                        $sub->kill(message => ["-warn", "Aborting this subtest because the parent subtest hit an error"]);
                        $sub->wait_finish();
                    }

                    $self->_set_err("Sync callback method has failed");
                    return (wantarray()) ? () : [];;
                }
            }
        }

        # obtain the current resmgr attr state
        my $resmgr_attr_state = $self->ResMgr::get_resmgr_attrs();
        if ($self->error()) {
            $self->_set_err("Failed to get ResMgr attrs");
            return (wantarray()) ? () : [];;
        }

        # tell each child to continue
        foreach my $wait_subtest (@waiting_subtests) {
            $self->SubtestComm::subtest_continue(-subtest => $wait_subtest,
                                                 -data => $resmgr_attr_state,
                                                 -data_type => "update_resmgr_attrs");
        }

        # wait for each child to finish
        &SanUtil::nate_process_results();
        @return_data = $self->_wait_for_subtests_to_complete(-dependence_type => $dependence_type,
                                                             -encountered_failure => $encountered_failure,
                                                             -propagate_wait => $propagate_wait,
                                                             -propagate_events => $propagate_events,
                                                             -enable_sync => $enable_sync,
                                                             -sync_callbacks => $sync_callbacks,
                                                             -last_error_string => $last_error_string,
                                                             -check_cmds_from_parent => $check_cmds_from_parent,
                                                             -forward_cmds_from_parent => $forward_cmds_from_parent,
                                                             -async => $async,
                                                             -return_data => \@return_data,
                                                             -subtests => \@waiting_subtests);
        return (wantarray()) ? @return_data : \@return_data;
    }

    # if async mode is off, then we know all subtests have completed
    # so we can wait on them to complete
    if ($async == $EventUtil::FALSE) {
        &SanUtil::nate_wait_finish(@orig_subtests);
    } else {
        # if async mode is on, then it is possible that a subtest
        # is still running, so we can't wait on them or else we would
        # potentially get dead locked
    }

    # if we encountered a failure in one of the subtests (at the end),
    if ($encountered_failure == $EventUtil::TRUE) {
        # bubble up the last error string we encountered up to the calling function
        # disable the error caller string temporarily while we set this error
        # so that the error string is easier to read.
        $self->{error_disable_caller_string} = 1;
        $self->_set_err("$last_error_string");
        $self->{error_disable_caller_string} = 0;
        return (wantarray()) ? @return_data : \@return_data;
    } else {
        return (wantarray()) ? @return_data : \@return_data;
    }
}

################################################################################
# Subroutine Name: _display_summary()
# Objective:       Display a summary of what Event parameters are set in the
#                  EventMan environment.
################################################################################
sub _display_summary {
    my ($self) = @_;

    my (@summary);

    push(@summary, "EventMan:");

    # display a summary of all the settings that are enabled/disabled
    # for this instance of EventMan
    push(@summary, $self->EventUtil::display_event_parameters(-objects => [$self]));
    if ($self->error()) {
        $self->_set_err("Unable to display event parameters");
        return $EventUtil::FALSE;
    }

    push(@summary, "");
    push(@summary, "\tEventSet Sequence:");
    push(@summary, "\t\t" . $self->{_event_sequence_string});

    my $str = join("\n", @summary);
    &SanUtil::nate_log('-comment', $str);

    # display all the event parameters that are stored per resource
    $self->ResMgr::display_event_parameters_for_current_resources();
    if ($self->error()) {
        $self->SubtestComm::subtest_fail();
        return $EventUtil::FALSE;
    }

    &SanUtil::nate_log('-comment', "EventMan Memory Usage: " . Devel::Size::total_size($self) . " bytes");

    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

EventMan.pm

=head1 Author

Robert Hunter

=head1 Description

Provides an interface to manage Event scheduling.

=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head1 B<Public Methods>

=head2 B<new(%args)>

Creates an EventMan object handle. Will also implicitly create a default EventSet
with all the inputs passed into this method.

=over 4

=item Event Param Capable?

Yes (EventMan and EventSet 'default_event_set' level)

=item Inputs

=item *same inputs into EventMan::EventSet::new()

=item Outputs

Returns an EventMan object

=item Examples

 use EventMan;

 my $sf_obj  = EventMan->new();
 if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
 }

 my $sf_obj2 = EventMan->new(-an_event_param => "some value");
 if ($sf_obj2->error()) {
     &tstlog(-comment, $sf_obj2->errmsg());
     return;
 }

=back

=head2 B<create_event_set(%args)>

Creates an EventSet.pm object and stores it internally on the EventMan instance.
This method accepts event parameters and applies them to the EventMan global
event parameter scope

=over 4

=item Event Param Capable?

Yes (EventSet Level)

=item Inputs

=item *I<-event_set_name>

Required: Name of the event set to create, must be unique.

=item *I<-cycle_count>

Optional: The number of cycles to run this EventSet back to back
Default: 1

=item Outputs

Returns a newly created EventSet object.

=item Examples

 my $fset_obj = $sf_obj->create_event_set(-event_set_name => "event_set_A",
                                          -cycle_count => 2);
 if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
 }

=back

=head2 B<get_event_set(%args)>

Obtain the EventSet object by name.

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-event_set_name>

Required: Name of the event set object to retrieve.

=item Outputs

Returns the EventSet object if found, sets an error if not found.

=item Examples

 my $fset_obj = $sf_obj->get_event_set(-event_set_name => "event_set_A");
 if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
 }

=back

=head2 B<delete_event_set(%args)>

Deletes the EventSet object by name and removes it from any
EventSetGroup that it is a member of.

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-event_set_name>

Required: Name of the event set object to delete.

=item Outputs

Returns $EventUtil::TRUE if successful, $EventUtil::FALSE if not
and sets an error.

=item Examples

 $fset_obj = $sf_obj->delete_event_set(-event_set_name => "event_set_A");
 if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
 }

=back

=head2 B<add_event(%args)>

Creates an Event object by name and adds it to the end of the specified
EventSet's event list.

=over 4

=item Event Param Capable?

Yes (Event level)

=item Inputs

=item *I<-event_name>

Required: Event keyword to use when creating the Event object

=item *I<-event_set_name>

Required: Name of the EventSet to add the Event object to

=item *I<-enable_validation>

Optional: Set to 1 to enable event validation on this Event.
Default: 1 (on)

=item *I<-cycle_count>

Optional: Denotes how many times to run the event back to back.
Default: 1

=item *I<-enable_prejob>

Optional: Enable (1) or Disable the prejob on this event
Default: 1 (enabled)

=item *I<-enable_postjob>

Optional: Enable (1) or Disable the postjob on this event
Default: 1 (enabled)

=item *I<-enable_cojob>

Optional: Enable (1) or Disable the cojob on this event
Default: 0 (disabled)

=item *I<-prejob>

Optional: Override the generic prejob (a subroutine that runs before the event is executed)
 for this event keyword
Type: CODE reference (first argument is the EventMan object)
Default: undef

=item *I<-postjob>

Optional: Override the generic postjob (a subroutine that runs after the event is executed)
for this event keyword
Type: CODE reference (first argument is the EventMan object)
Default: undef

=item *I<-cojob>

Optional: Override the generic cojob (a subroutine that runs in parallel as the event is executed)
for this event keyword
Type: CODE reference (first argument is the EventMan object)
Default: undef

=item *I<-cojob_dependence_type>

Optional: Indicates the dependence attribute between that Event and
its Cojob. (dependent/independent)
Default : 'depenendent'

=item Outputs

Returns the newly create Event object if successful, sets an error if not

=item Examples

 $event_obj = $sf_obj->add_event(-event_name => "some_event",
                                 -cycle_count => 10,
                                 -event_set_name => "event_set_A",
                                 -enable_prejob => 1,
                                 -prejob => sub { print "About to execute 'some_event'";});
 if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
 }

=back

=head2 B<delete_event(%args)>

Deletes the Event object by reference and removes it from any
EventSet that it is a member of.

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-event_obj>

Required: The Event object to destroy

=item Outputs

Returns $EventUtil::TRUE if successful, $EventUtil::FALSE if not
and sets an error.

=item Examples

 $sf_obj->delete_event(-event_obj => $event_obj);
 if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
 }

=back

=head2 B<create_event_set_group(%args)>

Creates an EventSetGroup object with the specified EventSet members

=over 4

=item Event Param Capable?

Yes (EventSetGroup level)

=item Inputs

=item *I<-group_name>

Required: A unique name for the EventSetGroup object

=item *I<-event_set_names> or I<-event_set_objs>

Optional: An array reference of valid EventSet names or of valid EventSet objects

=item *I<-run_style>

Optional: The style to run all the EventSet members ('sequential' or 'parallel')
Default: 'sequential'

=item *I<-dependence_type>

Optional: The dependence relationship between all EventSets in the group ('independent' or 'dependent'),
'Independent' indicates that the outcome of any EventSet will not affect the execution of any
sibling EventSets in the group. 'Dependent' is the opposite.
Default: 'independent'

=item *I<-enable_sync>

Optional: If enabled (1), will turn on Event synchronization between all EventSets in the group.
NOOP if dependence_type == 'dependent'. Events must implement synchronization points for this
to take effect.
Default: 0 (disabled)

=item Outputs

Returns an EventSetGroup object if successful, nothing if not and sets an error.

=item Examples

 my $fsg_obj = $sf_obj->create_event_set_group(-event_set_names => ["fset1", "fset2", "fset3"],
                                               -event_set_objs => [$fset4, $fset5],
                                               -group_name => "my_fset_group",
                                               -run_style => "parallel",
                                               -dependence_type => "independent",
                                               -enable_sync => 1);
 if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
 }

=back

=head2 B<get_event_set_group(%args)>

Obtain the EventSetGroup object by name.

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-group_name>

Required: Name of the EventSetGroup object to retrieve.

=item Outputs

Returns the EventSetGroup object if found, sets an error if not found.

=item Examples

 my $fsg_obj = $sf_obj->get_event_set_group(-group_name => "my_fset_group");
 if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
 }

=back

=head2 B<delete_event_set_group(%args)>

Delete the EventSetGroup object by name and remove any ties to any EventSets that may be in the group

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-group_name>

Required: Name of the EventSetGroup object to delete.

=item Outputs

Returns $EventUtil::TRUE if successful, $EventUtil::FALSE if not
and sets an error.

=item Examples

 $sf_obj->delete_event_set_group(-group_name => "my_fset_group");
 if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
 }

=back

=head2 B<is_running(%args)>

Determines if EventMan is running or not running. Behaves asynchronously (non blocking)

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns $EventUtil::TRUE if running, $EventUtil::FALSE if not running
Sets an error otherwise

=item Examples

 while ($sf_obj->is_running() == $EventUtil::TRUE) {
     # do something
 }
 if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
 }

=back

=head2 B<stop_events(%args)>

Graceful terminates all Events, EventSets, and EventSetGroups
that may be running, marks all EventMan subtests as PASS once
complete with the action.

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-async>

Optional: Indicates whether the action should be asynchronous (1, non blocking)
or synchronous (0, blocking)
Default: 0 (synchronous)

=item Outputs

Returns $EventUtil::TRUE if successful, $EventUtil::FALSE if not and sets an error.

=item Examples

  # synchronously
  $sf_obj->stop_events(-async => 0) {
  if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
  }

  # asynchronously
  $sf_obj->stop_events(-async => 1) {
  if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
  }
  while ($sf_obj->is_running() == $EventUtil::TRUE) {
      # do something
  }
  if ($sf_obj->error()) {
      &tstlog(-comment, $sf_obj->errmsg());
      return;
  }

=back

=head2 B<abort_events(%args)>

Graceful (or non gracefully) terminates all Events, EventSets, and EventSetGroups
that may be running, marks all EventMan subtests as WARN once
complete with the action.

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-async>

Optional: Indicates whether the action should be asynchronous (1, non blocking)
or synchronous (0, blocking)
Default: 0 (synchronous)

=item *I<-graceful>

Optional: Indicates whether the action should be graceful (1) or non graceful (0).
Non graceful will terminate all activities immediately in place. Graceful will allow events
'in flight' to complete the current cycle.
Default: 0 (non graceful)

=item Outputs

Returns $EventUtil::TRUE if successful, $EventUtil::FALSE if not and sets an error.

=item Examples

  # synchronously
  $sf_obj->abort_events(-async => 0) {
  if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
  }

  # asynchronously
  $sf_obj->abort_events(-async => 1) {
  if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
  }
  while ($sf_obj->is_running() == $EventUtil::TRUE) {
      # do something
  }
  if ($sf_obj->error()) {
      &tstlog(-comment, $sf_obj->errmsg());
      return;
  }

=back

=head2 B<initialize_events(%args)>

Initializes Event execution on the given EventSet sequence. This will spawn the eventman process
in the background, initialize all the event modules and utility modules, and wait for the user
to invoke the run_events() method. The purpose of this method is to optimize tests that may
require a long EventMan initialization phase. The options to this method are very similar to
the run_events() method. This method can be used in a blocking and non-blocking fashion.
By default it will be non-blocking. A use case for using it in a non-blocking fashion
would be if you wanted to control from your test exactly when events start, or maybe you
don't want to perform any operations in your test until eventman has fully initialized.

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-wait>

Optional: If set to 1, will wait for events to be fully initialized before returning
Default: 0 (will return immediately with eventman set to initialize in the background)

=item *I<-event_set_sequence>

Optional: An array reference of [EventSet names | EventSet objects | EventSetGroup objects]
to run in sequence
Default: The default behavior is to run the default event set.

=item *I<-sequence_style>

Optional: Indicates whether to run the event set sequence by a given "runlength",
a given "iterations" count, or "mixed" a combination of both runlength and iterations
Default: "iterations"

=item *I<-sequence_count>

Optional: Indicates how many iterations to perform if -sequence_style == "iterations"
Default: 1

=item *I<-sequence_runlength>

Optional: Indicates how many seconds to run the sequence if -sequence_style == "runlength"

=item Outputs

Returns $EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets an error

=item Examples

 # initialize eventman in the background
 $em_obj->initialize_events(-event_set_sequence => [$fset_obj, $fsgroup_obj, "myfset"]);
 if ($em_obj->error()) {
     die "Failed in initialize eventman: " . $em_obj->errmsg();
     return;
 }

 # do other tasks while eventman is initializing
 # ...

 # make sure eventman is still ok
 die "Eventman failed to initialize" if (not $em_obj->is_running());

 # tell eventman to start running events as soon as possible
 $em_obj->run_events();
 if ($em_obj->error()) {
     die "Failed in run events: " . $em_obj->errmsg();
     return;
 }

 while ($em_obj->is_running() == $EventUtil::TRUE) {
     # do other tasks while events are running
     # ...
 }

 if ($em_obj->error()) {
     die "Events failed: " . $em_obj->errmsg();
     return;
 }

=back

=head2 B<is_initializing()>

Determines if eventman is initializing

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns $EventUtil::TRUE if events are still initializing, $EventUtil::FALSE if events are not initializing.

=item Examples

 # initialize eventman in the background
 $em_obj->initialize_events(-event_set_sequence => [$fset_obj, $fsgroup_obj, "myfset"]);
 if ($em_obj->error()) {
     die "Failed in initialize eventman: " . $em_obj->errmsg();
     return;
 }

 while ($em_obj->is_initializing()) {
    # do other tasks while eventman is initializing
    # ...
 }

 if ($em_obj->has_initialized()) {
    # eventman finished initializing!
    # invoke run_events() to tell it to go
 }

=back

=head2 B<has_initialized()>

Determines if eventman has finished initializing.

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns $EventUtil::TRUE if events have been initialized, $EventUtil::FALSE if events have not been initialized yet

=item Examples

 # initialize eventman in the background
 $em_obj->initialize_events(-event_set_sequence => [$fset_obj, $fsgroup_obj, "myfset"]);
 if ($em_obj->error()) {
     die "Failed in initialize eventman: " . $em_obj->errmsg();
     return;
 }

 while ($em_obj->is_initializing()) {
    # do other tasks while eventman is initializing
    # ...
 }

 if ($em_obj->has_initialized()) {
    # eventman finished initializing!
    # invoke run_events() to tell it to go
 }

=back

=head2 B<was_set_to_initialize()>

Determines if eventman was set to initialize via the initialize_events()
method

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns $EventUtil::TRUE if events are have been set to initialize, $EventUtil::FALSE if events have already initialized or
has not been prepared to initialize.

=item Examples

 # initialize eventman in the background
 $em_obj->initialize_events(-event_set_sequence => [$fset_obj, $fsgroup_obj, "myfset"]);
 if ($em_obj->error()) {
     die "Failed in initialize eventman: " . $em_obj->errmsg();
     return;
 }

 if ($em_obj->was_set_to_initialize()) {
     while ($em_obj->is_initializing()) {
        # do other tasks while eventman is initializing
        # ...
     }

     if ($em_obj->has_initialized()) {
        # eventman finished initializing!
        # invoke run_events() to tell it to go
     }
 }

=back

=head2 B<run_events(%args)>

Starts Event execution on the given EventSet sequence

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-event_set_sequence>

Required: An array reference of [EventSet names | EventSet objects | EventSetGroup objects]
to run in sequence

=item *I<-sequence_style>

Optional: Indicates whether to run the event set sequence by a given "runlength",
a given "iterations" count, or "mixed" a combination of both runlength and iterations
Default: "iterations"

=item *I<-sequence_count>

Optional: Indicates how many iterations to perform if -sequence_style == "iterations"
Default: 1

=item *I<-sequence_runlength>

Optional: Indicates how many seconds to run the sequence if -sequence_style == "runlength"

=item *I<-async>

Optional: Indicates whether the action should be asynchronous (1, non blocking)
or synchronous (0, blocking)
Default: 0 (synchronous)

=item *I<-initialize_only>

Optional: Indicates whether EventMan should only initialize in the background and
wait for an additional call to run_events(). Any subsequent call to run_events() will tell
the EventMan process to start as soon as possible. This option only works if -async => 1.
Default: 0 (off)

=item Outputs

Returns $EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets an error

=item Examples

 # run events in the background for 2 hours
 $sf_obj->run_events(-event_set_sequence => [$fset_obj, $fsgroup_obj, "myfset"],
                     -sequence_style => "runlength",
                     -sequence_runlength => 7200,
                     -async => 1);
 if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
 }

 # reap it manually
 while ($sf_obj->is_running() == $EventUtil::TRUE) {
     # do something
 }
 if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
 }

=back

=head2 B<send_passthrough_command(%args)>

Send a passthrough command message to all currently running events.

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-msg>

Required: Specfies the title of the passthrough command

=item *I<-data>

Optional: A hash reference of key => value pairs to send along with the message.

=item Outputs

Returns $EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets an error

=item Examples

 $sf_obj->send_passthrough_command(-msg => "a pass through message",
                                   -data => {key1 => "value1",
                                             key2 => "value2"});
 if ($sf_obj->error()) {
     &tstlog(-comment, $sf_obj->errmsg());
     return;
 }

=back

=head2 B<can_read_event_messages(%args)>

Determine if there are any unread event messages in the event
message queue

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns $EventUtil::TRUE if there are events, $EventUtil::FALSE
if there are no events

=item Examples

 if ($sf_obj->can_read_event_messages() == $EventUtil::TRUE) {
     # do something (like, read an Event message))
 }

=back

=head2 B<read_event_message(%args)>

Read the earliest unread event message in the message queue

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-dont_mark_read>

Optional: If set to 1, will not mark any of the messages
retrieved as read.
Default: 0 (will mark messages as read when returned)

=item *I<-read_all>

Optional: If set to 1, will return all available unread messages
in the queue.
Default: 0 (will only return the earliest unread event message)

=item Outputs

Returns an array of event hash references if there are messages
to read. Nothing otherwise.

=item Examples

 while ($sf_obj->can_read_event_messages() == $EventUtil::TRUE) {
     my (@event_messages) = $sf_obj->read_event_message(-read_all => 1);
     # parse the event messages
 }

=back

=head2 B<get_all_event_messages(%args)>

Obtain every event message in the queue (read / unread)

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns an array of event message hash references if there are events.
Nothing otherwise.

=item Examples

 my (@event_messages) = $sf_obj->get_all_event_messages();

=back

=head2 B<add_resource(%args)>

Registers a resource with the resource manager

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-obj>

Required: A blessed object representing the resource.

=item *I<-type>

Required: The type of resource to associate this object with.
Supported Types:
 sanswitches => San::SanSwitch
 sanclients => San::SanClient
 nacl_clusters_cm => NACL::STask::Cluster
 nacl_nodes_cm => NACL::STask::Node
 controllerclusters => San::ControllerCluster
 sanfilersCm => San::ControllerCluster::Filer10M

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE if not and sets error.

=item Examples

 $sf_obj->add_resource(-name => "mycluster",
                       -type => "nacl_clusters_cm",
                       -obj  => $nacl_cluster_obj);
 if ($sf_obj->error()) {
     ...
 }

=back

=head2 B<remove_resource(%args)>

Unregisters a resource with the resource manager.

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-type>

Required: The type of the resource

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE if not and sets error.

=item Examples

 $sf_obj->remove_resource(-name => "fas3070-rtp07",
                          -type => "nacl_nodes_cm");
 if ($sf_obj->error()) {
     ...
 }

=back

=head2 B<get_resource_event_params(%args)>

Obtain the event parameter hash reference for the specified resource

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-type>

Required: The type of the resource

=item Outputs

A hash reference of event parameter name and value pairs that are associated for this resource.

=item Examples

 my $event_params = $em_obj->get_resource_event_params(-name => "fas3070-rtp07",
                                                       -type => "nacl_nodes_cm");
 if ($em_obj->error()) {
     ...
 }

 $event_params->{sram_detection} = 1;
 $event_params->{enable_syslog_messages} = 0;

 # set the new event params for this resource
 $em_obj->set_resource_event_params(-name => "fas3070-rtp07",
                                    -type => "nacl_nodes_cm",
                                    -args => $event_params);
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<set_resource_event_params(%args)>

Modify the event parameter hash for the specified resource.

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-type>

Required: The type of the resource

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not.

=item Examples

 my $event_params = $em_obj->get_resource_event_params(-name => "fas3070-rtp07",
                                                       -type => "nacl_nodes_cm");
 if ($em_obj->error()) {
     ...
 }

 $event_params->{sram_detection} = 1;
 $event_params->{enable_syslog_messages} = 0;

 # set the new event params for this resource
 $em_obj->set_resource_event_params(-name => "fas3070-rtp07",
                                    -type => "nacl_nodes_cm",
                                    -args => $event_params);
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<get_resource_names_by_type(%args)>

Obtain the resource names registered with eventman given a particular tpe

=over 4

=item Inputs

=item *I<-type>

Required: The type of the resource

=item Outputs

An array or array reference of resource names

=item Examples

 my @vserver_names_registered_with_eventman = $em_obj->get_resource_names_by_type(-type => "vservers");
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<set_default_event_set_sequence(%args)>

Modify the default event set sequence used by EventMan::run_events()

=over 4

=item Inputs

=item *I<-event_set_sequence>

Required: An array reference of EventMan::EventSet objects, EventMan::EventSetGroup objects,
and/or, EventMan::EventSet names.

=item *I<-sequence_params>

Optional: An hash reference of keys(sequence_count, sequence_runlength, sequence_style) and their
default values.

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not.

=item Examples

 $em_obj->set_default_eventset_sequence(-event_set_sequence => ["default_event_set"]);
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<get_default_event_set_sequence(%args)>

Modify the default event set sequence used by EventMan::run_events()

=over 4

=item Inputs

=item *I<-event_set_sequence>

Required: An array reference of EventMan::EventSet objects, EventMan::EventSetGroup objects,
and/or, EventMan::EventSet names.

=item Outputs

An array or array reference of EventSet Objects, EventSetGroup Objects, and/or EventSet names

=item Examples

 my @sequence = $em_obj->get_default_eventset_sequence();
 if ($sf_obj->error()) {
     ...
 }

=back

=head2 B<generate_event_param_documentation()>

Generate a documentation datastructure for all Event Parameters used in this package.

=over 4

=item Inputs

None

=item Outputs

A custom datastructure hash reference containing the documentation.

=item Examples

 my $documentation_hash = $sf_obj->generate_event_param_documentation();
 if ($sf_obj->error()) {
     ...
 }

=back

=head2 B<wait_for_subtests(%args)>

The subtest reaper method for all EventMan subtesting

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-subtests>

Required: An array reference of NATE subtest objects to wait on

=item *I<-dependence_type>

Optional: Specifies what the dependence type is between all subtests in -subtests
Default: "independent"

=item *I<-propagate_wait>

Optional: Specifies whether a child's subtest wait message should be propagated to this subtest's
parent subtest, or whether the wait message should be dealt with at this subtest level.
Default: $EventUtil::FALSE

=item *I<-propagate_events>

Optional: Specifies whether event messagess should be propagated to the parent subtest
or whether they should be stored at the current subtest level.
Default: $EventUtil::TRUE

=item *I<-enable_sync>

Optional: Specifies whether to enable subtest synchronization via subtest_wait() commands
Default: $EventUtil::FALSE

=item *I<-sync_callbacks>

Optional: An array reference of subroutine references to executed during subtest synchronization
points
Default: undef

=item *I<-poll_time>

Optional: The amount of time to wait between polling for incoming messages from children subtests
Default: 0.5 (seconds)

=item *I<-check_cmds_from_parent>

Optional: Specifies whether the method should periodically check for commands
from the parent subtest.
Default: $EventUtil::FALSE

=item *I<-forward_cmds_from_parent>

Optional: Specifies whether the method should forward commands received from the
parent subtest to all children subtests once they are received.
Default: $EventUtil::FALSE

=item *I<-async>

Optional: Specifies whether this method should be asynchronous (non blocking) or synchronous (blocking)
Default: $EventUtil::FALSE

=item Outputs

Returns an array or array reference  of return_data messages if any child
subtest sent them to the current subtest.

=item Examples

 # wait for these two subtests to finish, keep them independent from one
 # another so that if one fails, the other continues to completion.
 $sf_obj->wait_for_subtests(-dependence_type => "independent",
                            -subtests => [$subtest1, $subtest2]);
 if ($sf_obj->error()) {
     ...
 }

=back

=cut
