# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary An object representation of an Event
## @author hrivera@netapp.com
## @status public
## @pod here

package EventMan::Event;

use Error;
use ParamsMgr;

use vars qw(@ISA);
@ISA = qw(Error);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new()
# Objective: Event Object constructor
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my $self = {};

    bless($self, $class);

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);

    # parse out the required event name
    $self->{event_name} = &ParamsMgr::get_arg(-arg => "event_name",
                                              -required => 1);
    delete($args{event_name});

    # parse out the required event set name
    $self->{event_set_name} = &ParamsMgr::get_arg(-arg => "event_set_name",
                                                  -required => 1);
    delete($args{event_set_name});

    # parse out the required event modules array ref
    $self->{event_modules} = &ParamsMgr::get_arg(-arg => "event_modules",
                                                 -required => 1,
                                                 -ref_type => "ARRAY");
    delete($args{event_modules});

    # parse out the optional parameters
    $self->{unique_id} = &ParamsMgr::get_arg(-arg => "unique_id",
                                             -default => 0);
    delete($args{unique_id});

    $self->{enable_validation} = &ParamsMgr::get_arg(-arg => "enable_validation",
                                                     -default => 1);
    delete($args{enable_validation});

    $self->{cycle_count} = &ParamsMgr::get_arg(-arg => "cycle_count",
                                               -default => 1);
    delete($args{cycle_count});

    $self->{description} = &ParamsMgr::get_arg(-arg => "description",
                                               -default => "(none)");
    delete($args{description});

    # prejob
    $self->{enable_prejob} = &ParamsMgr::get_arg(-arg => "enable_prejob",
                                                 -default => "1");
    delete($args{enable_prejob});
    $self->{prejob} = &ParamsMgr::get_arg(-arg => "prejob",
                                          -default => undef,
                                          -ref_type => "CODE");
    delete($args{prejob});

    # postjob
    $self->{enable_postjob} = &ParamsMgr::get_arg(-arg => "enable_postjob",
                                                  -default => "1");
    delete($args{enable_postjob});
    $self->{postjob} = &ParamsMgr::get_arg(-arg => "postjob",
                                           -default => undef,
                                           -ref_type => "CODE");
    delete($args{postjob});

    # cojob
    $self->{enable_cojob} = &ParamsMgr::get_arg(-arg => "enable_cojob",
                                                -default => "0");
    delete($args{enable_cojob});
    $self->{cojob} = &ParamsMgr::get_arg(-arg => "cojob",
                                         -default => undef,
                                         -ref_type => "CODE");
    delete($args{cojob});
    $self->{cojob_dependence_type} = &ParamsMgr::get_arg(-arg => "cojob_dependence_type",
                                                         -default => "dependent");
    delete($args{cojob_dependence_type});

    # initialize the subtest_runid_counter
    $self->{subtest_runid_counter} = sprintf("%03d", "0");

    # store the remaining args in the args datastructure
    $self->{args} = {%args};

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    return $self;
}

###############################################################################
# Method:    set_event_parameter()
# Objective: set an event parameter on the event object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_event_parameter {
    my ($self, $event_param, $value) = @_;

    # fetch the original value
    my $original_value;
    if (exists($self->{args}->{$event_param})) {
        $original_value = $self->{args}->{$event_param};
    }

    # update the parameter value
    $self->{args}->{$event_param} = $value;

    return $original_value;
}

###############################################################################
# Method:    increment_subtest_runid_counter()
# Objective: Increment and return the subtest runid counter
# Details:   See POD documentation at the end of this file.
###############################################################################
sub increment_subtest_runid_counter {
    my ($self) = @_;

    $self->_clear_err();

    $self->{subtest_runid_counter}++;

    return $self->get_current_subtest_runid_counter();
}

###############################################################################
# Method:    get_current_subtest_runid_counter()
# Objective: Get the current subtest runid counter value
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_current_subtest_runid_counter {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{subtest_runid_counter};
}

###############################################################################
# Method:    get_name()
# Objective: Get the event keyword for this event object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{event_name};
}

###############################################################################
# Method:    get_event_modules()
# Objective: Get the list of EventModules this event is associated with
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_event_modules {
    my ($self) = @_;

    $self->_clear_err();

    return @{$self->{event_modules}};
}

###############################################################################
# Method:    get_event_set_name()
# Objective: Get the EventSet name that this object is associated with.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_event_set_name {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{event_set_name};
}

###############################################################################
# Method:    get_unique_id()
# Objective: Get the unique identifier for this Event object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_unique_id {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{unique_id};
}

###############################################################################
# Method:    get_cycle_count()
# Objective: Get the cycle count for this Event
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_cycle_count {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{cycle_count};
}

###############################################################################
# Method:    get_description()
# Objective: Get the description for this event
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_description {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{description};
}

###############################################################################
# Method:    is_prejob_enabled()
# Objective: Determine if the prejob is enabled for this Event
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_prejob_enabled {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{enable_prejob};
}

###############################################################################
# Method:    is_postjob_enabled()
# Objective: Determine if the postjob is enabled for this Event
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_postjob_enabled {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{enable_postjob};
}

###############################################################################
# Method:    is_cojob_enabled()
# Objective: Determine if the cojob is enabled for this Event
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_cojob_enabled {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{enable_cojob};
}

###############################################################################
# Method:    is_validation_enabled()
# Objective: Determine if validation is enabled
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_validation_enabled {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{enable_validation};
}

###############################################################################
# Method:    get_cojob_dependence_type()
# Objective: Get the dependence type between this Events exec and cojob
#            subroutines
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_cojob_dependence_type {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{cojob_dependence_type};
}

###############################################################################
# Method:    get_prejob()
# Objective: Obtain the prejob subroutine reference if it exists
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_prejob {
    my ($self) = @_;

    $self->_clear_err();

    # if a custom prejob is store on the event object, use that instead
    if (defined($self->{prejob})) {
        return $self->{prejob};
    } else {
        return undef;
    }
}

###############################################################################
# Method:    get_postjob()
# Objective: Get the the postjob subroutine reference if it exists
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_postjob {
    my ($self) = @_;

    $self->_clear_err();

    # if a custom postjob is store on the event object, use that instead
    if (defined($self->{postjob})) {
        return $self->{postjob};
    } else {
        return undef;
    }
}

###############################################################################
# Method:    get_cojob()
# Objective: Get the cojob subroutine reference if it exists
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_cojob {
    my ($self) = @_;

    $self->_clear_err();

    # if a custom cojob is store on the event object, use that instead
    if (defined($self->{cojob})) {
        return $self->{cojob};
    } else {
        return undef;
    }
}

###############################################################################
# Method:    display_summary()
# Objective: Print out a summary of this Event and what it will do.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub display_summary {
    my ($self, %args) = @_;

    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # parse out the event man object
    my $event_man_obj = &ParamsMgr::get_arg(-arg => "event_man_obj",
                                            -required => 1);
    delete($args{-event_man_obj});

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # display a summary of what this eventset is going to do
    my @summary = ();

    push(@summary, "Event: " . $self->get_name());
    push(@summary, "\tDescription: " . $self->get_description());
    push(@summary, "");
    push(@summary, "\tEvent Module(s):");
    push(@summary, "\t\t" . join("\n\t\t", $self->get_event_modules()));
    push(@summary, "");
    push(@summary, "\tAttributes:");
    push(@summary, "\t\t-unique_id : " . $self->get_unique_id());
    push(@summary, "\t\t-cycle_count : " . $self->get_cycle_count());
    push(@summary, "\t\t-enable_validation : " . $self->is_validation_enabled());
    push(@summary, "\t\t-enable_prejob : " . $self->is_prejob_enabled());
    push(@summary, "\t\t-enable_cojob : " . $self->is_cojob_enabled());
    push(@summary, "\t\t-enable_postjob : " . $self->is_postjob_enabled());
    push(@summary, "");

    push(@summary, $event_man_obj->EventUtil::display_event_parameters(%args));

    my $str = join("\n", @summary);
    &SanUtil::nate_log('-comment', $str);

    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

Event.pm

=head1 Author

Robert Hunter

=head1 Description

Object interface for an Event Scenario

=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for this event to use.

=head1 B<Public Methods>

=head2 B<new(%args)>

Creates an Event object handle. Do not call this method directly, use
EventMan::add_event() to create an Event object

=over 4

=item Event Param Capable?

Yes (Event level)

=item Inputs

=item *I<-event_name>

Required: A valid event keyword to map to this Event object

=item *I<-event_set_name>

Required: A valid EventSetName

=item *I<-event_modules>

Required: The array of Event modules that will be used when mapping to the
Event keyword

=item *I<-unique_id>

Optional: a unique string to identify this Event object
Default: 0

=item *I<-description>

Optional: A brief description describing the intentions of this event scenario
Default: (no description)

=item *I<-enable_validation>

Optional: Set to 1 to enable event validation on this Event.
Default: 0 (off)

=item *I<-cycle_count>

Optional: Denotes how many times to run the event back to back.
Default: 1

=item *I<-enable_prejob>

Optional: Enable (1) or Disable the prejob on this event
Default: 1 (enabled)

=item *I<-enable_postjob>

Optional: Enable (1) or Disable the postjob on this event
Default: 1 (enabled)

=item *I<-enable_cojob>

Optional: Enable (1) or Disable the cojob on this event
Default: 0 (disabled)

=item *I<-prejob>

Optional: Override the generic prejob (a subroutine that runs before the event is executed)
 for this event keyword
Type: CODE reference (first argument is the sanfaults object)
Default: undef

=item *I<-postjob>

Optional: Override the generic postjob (a subroutine that runs after the event is executed)
for this event keyword
Type: CODE reference (first argument is the sanfaults object)
Default: undef

=item *I<-cojob>

Optional: Override the generic cojob (a subroutine that runs in parallel as the event is executed)
for this event keyword
Type: CODE reference (first argument is the sanfaults object)
Default: undef

=item *I<-cojob_dependence_type>

Optional: Indicates the dependence attribute between that Event and
its Cojob. (dependent/independent)
Default : 'depenendent'

=item Outputs

Returns an Event object

=item Examples

 # see EventMan::add_event()

=back

=head2 B<get_name()>

Event keyword accessor method

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns the name of the event (keyword)

=item Examples

 my $event_name = $event_obj->get_name();

=back

=head2 B<get_current_subtest_runid_counter()>

subtest runid counter accessor method

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns the current subtest runid counter internal to this event (3 digit padded number)

=item Examples

 my $num = $event_obj->get_current_subtest_runid_counter();

=back

=head2 B<increment_subtest_runid_counter()>

Increment and return the subtest runid counter

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

The next subtest runid counter (3 digit padded number)

=item Examples

 my $num = $event_obj->increment_subtest_runid_counter();

=back

=head2 B<set_event_parameter($event_param, $value)>

Set a specific event parameter value on this event object.

=over 4

=item Event Param Capable?

No

=item Inputs

 event_param (scalar)
 
 value (scalar|arrayref|hashref|coderef)

=item Outputs

 The original value of the event parameter

=item Examples

 my $orig_value = $event_obj->set_event_parameter("mycoolparam", 100);

=back

=head2 B<get_event_modules()>

Obtain the list of Event Modules this Event object is using

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns an array of Event Module package strings

=item Examples

 my @event_modules = $event_obj->get_event_modules();

=back

=head2 B<get_event_set_name()>

EventSet name accessor method

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Get the EventSet name this Event is associated to

=item Examples

 my $event_set_name = $event_obj->get_event_set_name();

=back

=head2 B<get_unique_id()>

Accessor method for unique_id

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Get the unique identifier for this event object

=item Examples

 my $id = $event_obj->get_unique_id();

=back

=head2 B<get_cycle_count()>

Accessor method for cycle_count

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Get the cycle_count for this event object

=item Examples

 my $cycle_count = $event_obj->get_cycle_count();

=back

=head2 B<get_description()>

Obtain the Event description

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns the event description

=item Examples

 my $cycle_count = $event_obj->get_description();

=back

=head2 B<is_prejob_enabled()>

is_prejob_enabled Accessor method

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns $EventUtil::TRUE if enabled, $EventUtil::FALSE if not enabled

=item Examples

 my $is_enabled = $event_obj->is_prejob_enabled();

=back


=head2 B<is_postjob_enabled()>

is_postjob_enabled Accessor method

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns $EventUtil::TRUE if enabled, $EventUtil::FALSE if not enabled

=item Examples

 my $is_enabled = $event_obj->is_postjob_enabled();

=back

=head2 B<is_cojob_enabled()>

is_cojob_enabled Accessor method

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns $EventUtil::TRUE if enabled, $EventUtil::FALSE if not enabled

=item Examples

 my $is_enabled = $event_obj->is_cojob_enabled();

=back

=head2 B<get_prejob()>

prejob Accessor method

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns the prejob stored on the Event object

=item Examples

 my $prejob_subroutine = $event_obj->get_prejob();

=back

=head2 B<get_postjob()>

postjob Accessor method

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns the postjob stored on the Event object

=item Examples

 my $postjob_subroutine = $event_obj->get_postjob();

=back

=head2 B<get_cojob()>

cojob Accessor method

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns the cojob stored on the Event object

=item Examples

 my $cojob_subroutine = $event_obj->get_cojob();

=back

=head2 B<get_cojob_dependence_type()>

Cojob dependence type accessor method

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns the cojob dependence type attribute stored on the Event object

=item Examples

 my $cojob_dependence_type = $event_obj->get_cojob_dependence_type();

=back

=cut
